//////////////////////////////////////////////////////////////////////////////
//
// Asura_Nav_Types.cpp
//
// Copyright (C) 2009 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_NEW_NAV_STUFF

// header
#include "AI/Asura_Nav_Types.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "Core/Asura_Mempool.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Debugging/Asura_Logfile.h"

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation
//
//////////////////////////////////////////////////////////////////////////////

void Asura_Navigation::Initialise()
{
	#ifdef ASURA_USE_NEW_NAV_STUFF
	Path::InitialisePathMempools();
	#endif
}

void Asura_Navigation::Shutdown()
{
	#ifdef ASURA_USE_NEW_NAV_STUFF
	Path::ShutdownPathMempools();
	#endif
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::TestParams
//
//////////////////////////////////////////////////////////////////////////////

const Asura_Navigation::TestParams Asura_Navigation::s_xStandardTParams;

Asura_Navigation::TestParams::TestParams()
: m_uPathfinderFlags( 0 )
, m_fActorRadius( 0.25f )
, m_fActorHeight( 2.0f )
, m_bRejectIsolatedWPs( false )
, m_bRejectNonalignedWPs( false )
, m_bAllowLinearSearch( false )
, m_xNormal( 0.0f, -1.0f, 0.0f )
, m_usZone( ASURA_NAVVERTEX_MAX_ZONE_ID )
{
}

bool Asura_Navigation::TestParams::IsValidWaypoint( const Asura_AI_NavVertex& xWP ) const
{
	// If it is not in the correct zone then reject it.
	if ( m_usZone != ASURA_NAVVERTEX_MAX_ZONE_ID && xWP.GetZone() != m_usZone )
	{
		return false;
	}

	if ( m_bRejectIsolatedWPs && xWP.GetNumLinks() == 0 )
	{
		return false;
	}

	if ( m_bRejectNonalignedWPs )
	{
		// If the normal doesn't match our test normal then reject this waypoint.
		// Allow up to 10 difference.
		if ( xWP.GetNormal() * m_xNormal < 0.984807753f )
		{
			return false;
		}
	}

	return true;
}

bool Asura_Navigation::TestParams::IsValidLink( const Asura_AI_NavConnection& xLink ) const
{
	// Generally if a link is blocked we don't want to be using it during nav data querying.
	return ( ( xLink.GetFlags() & ASURA_AI_CONNECTION_FLAGS_BLOCKED ) == 0 );
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::APParams
//
//////////////////////////////////////////////////////////////////////////////

const Asura_Navigation::APParams Asura_Navigation::s_xStandardAPParams;

Asura_Navigation::APParams::APParams()
: m_bSphericalTestOnly( false )
, m_fMaxCoplanarOffset( 0.25f )
, m_fMaxPerpendicularOffset( 0.5f )
, m_bUseNormal( false )
, m_xNormal( 0.0f, -1.0f, 0.0f )
{
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::PathPoint
//
//////////////////////////////////////////////////////////////////////////////

void Asura_Navigation::PathPoint::Initialise()
{
	m_usWPID = ASURA_INVALID_WAYPOINT_ID;
	m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
	m_ucFlags = 0;
	m_xPosition = Asura_Maths::ZeroVector3;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::AStarData
//
//////////////////////////////////////////////////////////////////////////////

Asura_LinkList< Asura_Navigation::AStarData > Asura_Navigation::AStarData::s_xAStarDataList;

void Asura_Navigation::AStarData::InitialiseAStarStorage( const u_int uNumWPs )
{
	// Make sure there is an AStarData structure with enough waypoint storage in it.
	for ( Asura_LinkList_ConstIt< AStarData > xASDIter( s_xAStarDataList ); !xASDIter.Done(); xASDIter.Next() )
	{
		if ( !xASDIter.GetCurrent()->m_bInUse && xASDIter.GetCurrent()->m_uArraySize > uNumWPs )
		{
			return;
		}
	}

	// Will have to add one.
	AStarData xASD;
	s_xAStarDataList.PushBack( xASD );
	s_xAStarDataList.Back()->Data().m_bInUse = false;
	s_xAStarDataList.Back()->Data().Initialise( uNumWPs );
}

void Asura_Navigation::AStarData::ShudownAStarStorage( const u_int uNumWPs )
{
	for ( Asura_LinkList_It< AStarData > xASDIter( s_xAStarDataList ); !xASDIter.Done(); xASDIter.Next() )
	{
		if ( !xASDIter.GetCurrent()->m_bInUse && xASDIter.GetCurrent()->m_uArraySize == uNumWPs )
		{
			s_xAStarDataList.Remove( xASDIter.GetCurrentNode() );
			return;
		}
	}

	// Can't have this in until the nasty Initialise/Reset/Shutdown stuff is fixed in Asura_AI_Navigation_Handler.
	//Asura_Assert( false, "Asura_Navigation::AStarData::ShudownAStarStorage - could not find storage to reset" );
}

Asura_Navigation::AStarData* Asura_Navigation::AStarData::AllocateAStarData( const u_int uNumWPs )
{
	AStarData* pxASD = NULL;

	// Find best fit.
	u_int uBestCountDiff = UINT_MAX;
	for ( Asura_LinkList_It< AStarData > xASDIter( s_xAStarDataList ); !xASDIter.Done(); xASDIter.Next() )
	{
		if ( !xASDIter.GetCurrent()->m_bInUse && xASDIter.GetCurrent()->m_uArraySize >= uNumWPs )
		{
			const u_int uCountDiff = xASDIter.GetCurrent()->m_uArraySize - uNumWPs;
			if ( uCountDiff < uBestCountDiff )
			{
				uBestCountDiff = uCountDiff;
				pxASD = xASDIter.GetCurrent();
			}
		}
	}

	if ( pxASD )
	{
		pxASD->BeginSearch( uNumWPs );
	}

	return pxASD;
}

void Asura_Navigation::AStarData::DeallocateAStarData( AStarData* pxASD )
{
	pxASD->EndSearch();
}

Asura_Navigation::AStarData::AStarData()
: m_pxNodeList( NULL )
, m_pxOpenClosedVector( NULL )
, m_pxOpenList( NULL )
, m_iNumOpenListEntries( 0 )
, m_uArraySize( 0 )
, m_bInUse( false )
{
}

Asura_Navigation::AStarData::AStarData( const AStarData& xRHS )
: m_pxNodeList( NULL )
, m_pxOpenClosedVector( NULL )
, m_pxOpenList( NULL )
, m_iNumOpenListEntries( 0 )
, m_uArraySize( 0 )
, m_bInUse( false )
{
	*this = xRHS;
}

Asura_Navigation::AStarData::~AStarData()
{
	Asura_ResourcePushType( ASURA_RESOURCE_WAYPOINT );
	Asura_ResourceDeallocateArray( m_pxNodeList );
	Asura_ResourceDeallocateArray( m_pxOpenClosedVector );
	Asura_ResourceDeallocateArray( m_pxOpenList );
	Asura_ResourcePopType();
}

void Asura_Navigation::AStarData::AddNodeToOpenList( AStarNode* pxNode )
{
	const float fNewNodeTotalCost = pxNode->m_fCostToHere + pxNode->m_fCostToGoal;

	// Add the new node at the bottom of the min-heap and sift up.
	int iElementIndex = m_iNumOpenListEntries;

	m_pxOpenList[ iElementIndex ].m_fTotalCost = fNewNodeTotalCost;
	m_pxOpenList[ iElementIndex ].m_uIndex     = pxNode->m_uIndex;
	++m_iNumOpenListEntries;

	// While parent is greater, swap the element with its parent.
	int iParentIndex = ( iElementIndex - 1 ) >> 1;
	while ( iParentIndex >= 0 && m_pxOpenList[ iParentIndex ].m_fTotalCost > m_pxOpenList[ iElementIndex ].m_fTotalCost )
	{
		const OpenListElement xSwapNode = m_pxOpenList[ iParentIndex ];
        m_pxOpenList[ iParentIndex ]    = m_pxOpenList[ iElementIndex ];
        m_pxOpenList[ iElementIndex ]   = xSwapNode;
		iElementIndex = iParentIndex;
		iParentIndex  = ( iElementIndex - 1 ) >> 1;
	}
}

Asura_Navigation::AStarData::AStarNode* Asura_Navigation::AStarData::GetOpenListNode()
{
	if ( m_iNumOpenListEntries == 0 ) { return NULL; }

	// Pop the front off the open list, move the back of the heap to the front and sift down.
	int iElementIndex = 0;
	const u_int uBestWPID = m_pxOpenList[ iElementIndex ].m_uIndex;
	m_pxOpenList[ iElementIndex ] = m_pxOpenList[ --m_iNumOpenListEntries ];

	const int iLastEntryIndex = m_iNumOpenListEntries - 1;
	int iFirstChildIndex = ( iElementIndex << 1 ) + 1;
	while ( iFirstChildIndex <= iLastEntryIndex )
	{
		int iMinChildIndex;
		if ( iFirstChildIndex == iLastEntryIndex )
		{
			iMinChildIndex = iFirstChildIndex;
		}
		else if ( m_pxOpenList[ iFirstChildIndex ].m_fTotalCost < m_pxOpenList[ iFirstChildIndex + 1 ].m_fTotalCost )
		{
			iMinChildIndex = iFirstChildIndex;
		}
		else
		{
			iMinChildIndex = iFirstChildIndex + 1;
		}

		if ( m_pxOpenList[ iElementIndex ].m_fTotalCost > m_pxOpenList[ iMinChildIndex ].m_fTotalCost )
		{
			const OpenListElement xSwapNode = m_pxOpenList[ iMinChildIndex ];
            m_pxOpenList[ iMinChildIndex ]  = m_pxOpenList[ iElementIndex ];
            m_pxOpenList[ iElementIndex ]   = xSwapNode;
			iElementIndex    = iMinChildIndex;
			iFirstChildIndex = ( iElementIndex << 1 ) + 1;
		}
		else
		{
			break;
		}
	}

	return &m_pxNodeList[ uBestWPID ];
}

Asura_Navigation::AStarData& Asura_Navigation::AStarData::operator=( const AStarData& xRHS )
{
	Asura_ResourcePushType( ASURA_RESOURCE_WAYPOINT );
	if ( this != &xRHS )
	{
		Asura_ResourceDeallocateArray( m_pxNodeList );
		Asura_ResourceDeallocateArray( m_pxOpenClosedVector );
		Asura_ResourceDeallocateArray( m_pxOpenList );

		m_uArraySize = xRHS.m_uArraySize;
		if ( m_uArraySize )
		{
			Asura_Assert( false, "" );
			Asura_ResourceAllocateArray( m_pxNodeList, m_uArraySize );
			Asura_ResourceAllocateArray( m_pxOpenClosedVector, m_uArraySize );
			Asura_ResourceAllocateArray( m_pxOpenList, m_uArraySize );

			Asura_Mem::Copy( m_pxNodeList,         xRHS.m_pxNodeList,         m_uArraySize * sizeof( AStarNode ) );
			Asura_Mem::Copy( m_pxOpenClosedVector, xRHS.m_pxOpenClosedVector, m_uArraySize * sizeof( OpenClosedElement ) );
			Asura_Mem::Copy( m_pxOpenList,         xRHS.m_pxOpenList,         m_uArraySize * sizeof( OpenListElement ) );
			m_iNumOpenListEntries = xRHS.m_iNumOpenListEntries;
		}
		else
		{
			m_pxNodeList          = NULL;
			m_pxOpenClosedVector  = NULL;
			m_pxOpenList          = NULL;
			m_iNumOpenListEntries = 0;
		}
	}
	Asura_ResourcePopType();

	return *this;
}

bool Asura_Navigation::AStarData::Initialise( const u_int uNumNodes )
{
	if ( uNumNodes == m_uArraySize )
	{
		return true;
	}

	Asura_ResourcePushType( ASURA_RESOURCE_WAYPOINT );
	Asura_ResourceDeallocateArray( m_pxNodeList );
	Asura_ResourceDeallocateArray( m_pxOpenClosedVector );
	Asura_ResourceDeallocateArray( m_pxOpenList );
	m_pxNodeList = NULL;
	m_pxOpenClosedVector = NULL;
	m_pxOpenList = NULL;

	m_uArraySize = uNumNodes;
	Asura_ResourceAllocateArray( m_pxNodeList, m_uArraySize );
	Asura_ResourceAllocateArray( m_pxOpenClosedVector, m_uArraySize );
	Asura_ResourceAllocateArray( m_pxOpenList, m_uArraySize );
	Asura_ResourcePopType();

	// Initialise indices for the A* nodes.
	for ( u_int u = 0; u < m_uArraySize; ++u )
	{
		m_pxNodeList[ u ].m_uIndex = u;
	}

	return ( m_pxNodeList && m_pxOpenClosedVector && m_pxOpenList );
}

void Asura_Navigation::AStarData::BeginSearch( const u_int uNumNodes )
{
	m_bInUse = true;
	// Make all nodes not open and not closed.
	Asura_Mem::Set( m_pxOpenClosedVector, 0, uNumNodes * sizeof( OpenClosedElement ) );
	m_iNumOpenListEntries = 0;
}

void Asura_Navigation::AStarData::EndSearch()
{
	m_bInUse = false;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::SearchFunctor
//
//////////////////////////////////////////////////////////////////////////////

bool Asura_Navigation::SearchFunctor::VerifyInputData()
{
	return ( m_pxWPs != NULL && m_uNumWPs > 0 );
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::AStarFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::AStarFunctor::AStarFunctor()
: m_pxTestParams( &s_xStandardTParams )
, m_pxPathPoints( NULL )
, m_bRefinePath( false )
, m_pxAStarData( NULL )
, m_uCurrentWPID( ASURA_INVALID_WAYPOINT_ID )
{
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::AStarFunctor::Search()
{
	Status eStatus = BeginIterativeSearch();
	while ( eStatus == STATUS_WORKING )
	{
		eStatus = IterateSearch();
	}
	return eStatus;
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::AStarFunctor::BeginIterativeSearch()
{
	if ( !VerifyInputData() )
	{
		Asura_Assert( false, "Asura_Navigation::AStarFunctor::BeginIterativeSearch - bad input data" );
		return STATUS_IN_ERROR;
	}

	const Asura_AI_NavVertex* pxWPs;
	u_int uNumWPs;
	GetInputBuffer( pxWPs, uNumWPs );

	// Simple check first to make sure we're not already at our destination.
	m_uCurrentWPID = GetStartWPID();
	if ( GoalAchieved( m_uCurrentWPID, pxWPs[ m_uCurrentWPID ] ) )
	{
		PathPoint xNewPoint;
		xNewPoint.m_usWPID      = static_cast< u_short >( m_uCurrentWPID );
		xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
		xNewPoint.m_ucFlags     = PathPoint::WAYPOINT_POINT;
		xNewPoint.m_xPosition   = pxWPs[ m_uCurrentWPID ].GetPosition();
		m_pxPathPoints->PushFront( xNewPoint );
		return STATUS_OK;
	}

	m_pxAStarData = Asura_Navigation::AStarData::AllocateAStarData( uNumWPs );
	if ( !m_pxAStarData ) { return STATUS_IN_ERROR; }

	AStarData::AStarNode*		  pxCurrentNode		  = &m_pxAStarData->m_pxNodeList[ m_uCurrentWPID ];
	AStarData::OpenClosedElement* pxCurrentOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ m_uCurrentWPID ];

	// Add first node to open map and open list.
	pxCurrentOpenClosed->m_bOpen = true;
	pxCurrentNode->m_fCostToHere = 0.0f;
	pxCurrentNode->m_fCostToGoal = EstimateCostToGoal( m_uCurrentWPID, pxWPs[ m_uCurrentWPID ] );
	pxCurrentNode->m_pxParent    = NULL;

	m_pxAStarData->AddNodeToOpenList( pxCurrentNode );

	return STATUS_WORKING;
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::AStarFunctor::IterateSearch()
{
	const Asura_AI_NavVertex* pxWPs;
	u_int uNumWPs;
	GetInputBuffer( pxWPs, uNumWPs );

	bool bFinished = false;

	AStarData::AStarNode*		  pxCurrentNode;
	AStarData::OpenClosedElement* pxCurrentOpenClosed;
	if ( ( pxCurrentNode = m_pxAStarData->GetOpenListNode() ) != NULL )
	{
		m_uCurrentWPID      = pxCurrentNode->m_uIndex;
		pxCurrentOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ m_uCurrentWPID ];

		const Asura_AI_NavVertex& xCurrentWP = pxWPs[ m_uCurrentWPID ];
		if ( GoalAchieved( m_uCurrentWPID, xCurrentWP ) )
		{
			// Done.
			bFinished = true;
		}

		// Visit all the neighbours of the current node and decide whether to add them to the open list.
		for ( u_int u = 0; u < xCurrentWP.GetNumLinks(); ++u )
		{
			const Asura_AI_NavConnection& xLink = xCurrentWP.GetLink( u );
			const u_int uNextWPID = xLink.GetIndex();
			const Asura_AI_NavVertex& xNextWP = pxWPs[ uNextWPID ];
			if ( ShouldIgnoreLink( m_uCurrentWPID, xCurrentWP, xLink, uNextWPID, xNextWP ) )
			{
				continue;
			}

			const float fTraversalCost = EstimateTraversalCost( m_uCurrentWPID, xCurrentWP, xLink, uNextWPID, xNextWP );
			// If we get a negative cost then disregard this link.
			if ( fTraversalCost < 0.0f )
			{
				continue;
			}
			const float fNewCostToHere = pxCurrentNode->m_fCostToHere + fTraversalCost;

			// Get a pointer to this neighbour's data.
			AStarData::AStarNode*         pxNewNode       = &m_pxAStarData->m_pxNodeList[ uNextWPID ];
			AStarData::OpenClosedElement* pxNewOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ uNextWPID ];

			// If we've already visited this node before && this cost is no better then move on.
			if ( ( pxNewOpenClosed->m_bOpen || pxNewOpenClosed->m_bClosed ) && pxNewNode->m_fCostToHere <= fNewCostToHere )
			{
				continue;
			}

			pxNewNode->m_pxParent    = pxCurrentNode;
			pxNewNode->m_fCostToHere = fNewCostToHere;
			pxNewNode->m_fCostToGoal = EstimateCostToGoal( uNextWPID, xNextWP );

			pxNewOpenClosed->m_bClosed = false;
			pxNewOpenClosed->m_bOpen   = true;

			m_pxAStarData->AddNodeToOpenList( pxNewNode );
		}

		pxCurrentOpenClosed->m_bClosed = true;
		pxCurrentOpenClosed->m_bOpen   = false;
	}

	if ( bFinished || !pxCurrentNode )
	{
		// Iterate back up the node list and write the route out.
		while ( pxCurrentNode )
		{
			PathPoint xNewPoint;
			xNewPoint.m_usWPID      = static_cast< u_short >( pxCurrentNode->m_uIndex );
			xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
			xNewPoint.m_ucFlags     = PathPoint::WAYPOINT_POINT;
			xNewPoint.m_xPosition   = pxWPs[ pxCurrentNode->m_uIndex ].GetPosition();
			m_pxPathPoints->PushFront( xNewPoint );

			pxCurrentNode = pxCurrentNode->m_pxParent;
		}

		Asura_Navigation::AStarData::DeallocateAStarData( m_pxAStarData );
		m_pxAStarData = NULL;
		return STATUS_OK;
	}
	else
	{
		return STATUS_WORKING;
	}
}

bool Asura_Navigation::AStarFunctor::PostSearchProcessing()
{
	bool bOk = SetLinkIndices();
	bOk &= AddStartAndDestPositions();
	bOk &= GetShouldRefinePath() ? RefinePath() : true;
	bOk &= AnnotatePath();
	return bOk;
}

bool Asura_Navigation::AStarFunctor::VerifyInputData()
{
	if ( !PARENT::VerifyInputData() ) { return false; }

	return ( m_pxTestParams != NULL && m_pxPathPoints != NULL );
}

bool Asura_Navigation::AStarFunctor::ShouldIgnoreLink( const u_int uSourceWPIndex, const Asura_AI_NavVertex& xSourceWP,
													   const Asura_AI_NavConnection& xLink,
													   const u_int uTargetWPIndex, const Asura_AI_NavVertex& xTargetWP )
{
	const u_int uPathfinderFlags = GetTestParams()->m_uPathfinderFlags;
	const u_int uLinkFlags = xLink.GetFlags();

	if ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_BLOCKED )
	{
		// Ignore blocked waypoints.
		return true;
	}
	const bool bCanFallFar = ( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_CANFALLFAR ) ? true : false;
	if ( ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_LONGFALL ) && !bCanFallFar )
	{
		// It's a long fall and we can't long fall - ignore.
		return true;
	}
	const bool bCanFall = ( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_CANFALL ) ? true : false;
	if ( ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_FALL ) && !bCanFall )
	{
		// It's a fall link and we can't fall - ignore.
		return true;
	}

	const bool bCanJump = ( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_CANJUMP ) ? true : false;
	if ( ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_JUMP ) && !bCanJump )
	{
		// It's a jump link and we can't jump - ignore.
		return true;
	}

	return false;
}

bool Asura_Navigation::AStarFunctor::SetLinkIndices()
{
	if ( m_pxPathPoints->Size() == 0 ) { return true; }

	// At this point all path points should have an associated waypoint, so there should be nav links between all path points.

	Asura_ListElt< PathPoint >* pxCurrElt = m_pxPathPoints->Front();
	Asura_ListElt< PathPoint >* pxNextElt = pxCurrElt->Next();
	while ( pxCurrElt && pxNextElt )
	{
		const Asura_AI_NavVertex* pxCurrWP = Asura_AI_Navigation_System::GetWaypoint( pxCurrElt->Data().m_usWPID );
		if ( !pxCurrWP ) { return false; }

		for ( u_char ucLink = 0; ucLink < pxCurrWP->GetNumLinks(); ++ucLink )
		{
			if ( pxCurrWP->GetLink( ucLink ).GetIndex() == pxNextElt->Data().m_usWPID )
			{
				pxCurrElt->Data().m_ucLinkIndex = ucLink;
				break;
			}
		}
		if ( pxCurrElt->Data().m_ucLinkIndex == ASURA_AI_INVALID_LINK_INDEX )
		{
			return false;
		}

		pxCurrElt = pxNextElt;
		pxNextElt = pxNextElt->Next();
	}

	return true;
}

bool Asura_Navigation::AStarFunctor::AddStartAndDestPositions()
{
	if ( m_pxPathPoints->Size() == 0 ) { return true; }

	const Asura_AI_NavVertex* pxStartWP = Asura_AI_Navigation_System::GetWaypoint( m_pxPathPoints->Front()->Data().m_usWPID );
	if ( !pxStartWP ) { return false; }
	const Asura_AI_NavVertex* pxDestWP = Asura_AI_Navigation_System::GetWaypoint( m_pxPathPoints->Back()->Data().m_usWPID );
	if ( !pxDestWP ) { return false; }

	const Asura_Vector_3& xStartPos = GetStartPosition();
	const Asura_Vector_3& xDestPos  = GetDestinationPosition();

	PathPoint xNewPoint;
	xNewPoint.m_usWPID      = pxStartWP->Contains( xStartPos ) ? m_pxPathPoints->Front()->Data().m_usWPID : ASURA_INVALID_WAYPOINT_ID;
	xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
	xNewPoint.m_ucFlags     = 0;
	xNewPoint.m_xPosition   = xStartPos;
	m_pxPathPoints->PushFront( xNewPoint );

	xNewPoint.m_usWPID      = pxDestWP->Contains( xDestPos ) ? m_pxPathPoints->Back()->Data().m_usWPID : ASURA_INVALID_WAYPOINT_ID;
	xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
	xNewPoint.m_ucFlags     = 0;
	xNewPoint.m_xPosition   = xDestPos;
	m_pxPathPoints->PushBack( xNewPoint );

	return true;
}

bool Asura_Navigation::AStarFunctor::RefinePath()
{
	// We know at this point that the first and last points are not associated with a waypoint, but every other point is.
	// This means that we can't test the first waypoint and first and last link, because they are not in the waypoint data.

	// Go through the route removing redundant waypoints.
	// The base element is the element from which we are trying to smooth.
	// The target element is the element we've got to with the smoothing.

	const float fActorHeight = GetTestParams()->m_fActorHeight;
	const float fActorRadius = GetTestParams()->m_fActorRadius;

	Asura_ListElt< PathPoint >* pxBaseElt = m_pxPathPoints->Front();
	while ( pxBaseElt && pxBaseElt->Next() && pxBaseElt->Next()->Next() )
	{
		const Asura_AI_NavVertex* pxBaseWP = Asura_AI_Navigation_System::GetWaypoint( pxBaseElt->Data().m_usWPID );
		const Asura_Vector_3 xBaseNormal = pxBaseWP ? pxBaseWP->GetNormal() : Asura_Vector_3( 0.0f, -1.0f, 0.0f );

		// Iterate away from the base element until we cannot shortcut any more.
		Asura_ListElt< PathPoint >* pxBeginElt = pxBaseElt;
		const bool bIsFirstPoint = ( pxBeginElt->Prev() == NULL );
		const Asura_AI_NavVertex* pxBeginWP = bIsFirstPoint ? NULL : Asura_AI_Navigation_System::GetWaypoint( pxBeginElt->Data().m_usWPID );

		Asura_ListElt< PathPoint >* pxEndElt = pxBeginElt->Next();
		while ( pxEndElt )
		{
			if ( pxBeginWP && !WaypointIsValidForRemoval( pxBeginElt->Data().m_usWPID, *pxBeginWP ) )
			{
				break;
			}

			const bool bIsLastLink = ( pxEndElt->Next() == NULL );
			const Asura_AI_NavConnection* pxLink = ( !bIsFirstPoint && !bIsLastLink && pxBeginWP ) ?
												   pxBeginWP->GetLinkTo( pxEndElt->Data().m_usWPID ) :
												   NULL;
			const Asura_AI_NavVertex* pxEndWP = Asura_AI_Navigation_System::GetWaypoint( pxEndElt->Data().m_usWPID );
			if ( pxLink && pxEndWP && !LinkIsValidForRemoval( pxBeginElt->Data().m_usWPID, *pxBeginWP,
															  *pxLink,
															  pxEndElt->Data().m_usWPID, *pxEndWP ) )
			{
				break;
			}

			// Work out sideways vectors so we can incorporate the character width.
			const Asura_Vector_3 xLineVector = pxEndElt->Data().m_xPosition - pxBaseElt->Data().m_xPosition;

			Asura_Vector_3 xBaseSideways = xBaseNormal ^ xLineVector;
			xBaseSideways.Normalise();
			const Asura_Vector_3 xBaseSidewaysOffset = xBaseSideways * ( fActorRadius * 0.5f );

			const Asura_Vector_3 xEndNormal = pxEndWP ? pxEndWP->GetNormal() : Asura_Vector_3( 0.0f, -1.0f, 0.0f );
			Asura_Vector_3 xEndSideways = xEndNormal ^ xLineVector;
			xEndSideways.Normalise();
			const Asura_Vector_3 xEndSidewaysOffset = xEndSideways * ( fActorRadius * 0.5f );

			// We test two lines, one either side, to ensure the refined route is wide enough for the character.
			if ( !Asura_AI_Navigation_System::LineSegmentIsInWaypointSpace( pxBaseElt->Data().m_xPosition + xBaseSidewaysOffset + xBaseNormal * fActorHeight * 0.1f,
																			pxEndElt->Data().m_xPosition + xEndSidewaysOffset + xEndNormal * fActorHeight * 0.1f ) )
			{
				break;
			}
			if ( !Asura_AI_Navigation_System::LineSegmentIsInWaypointSpace( pxBaseElt->Data().m_xPosition - xBaseSidewaysOffset + xBaseNormal * fActorHeight * 0.1f,
																			pxEndElt->Data().m_xPosition - xEndSidewaysOffset + xEndNormal * fActorHeight * 0.1f ) )
			{
				break;
			}

			// We test another line to ensure the refined route is high enough for the character.
			const Asura_Vector_3 xBaseUpOffset = xBaseNormal * fActorHeight;
			const Asura_Vector_3 xEndUpOffset  = xEndNormal  * fActorHeight;
			if ( !Asura_AI_Navigation_System::LineSegmentIsInWaypointSpace( pxBaseElt->Data().m_xPosition + xBaseUpOffset,
																			pxEndElt->Data().m_xPosition + xEndUpOffset ) )
			{
				break;
			}

			pxBeginElt = pxEndElt;
			pxBeginWP  = pxEndWP;
			pxEndElt   = pxEndElt->Next();
		}

		// Remove the path points between the base and the begin elements.
		if ( pxBeginElt != pxBaseElt )
		{
			// The link from the beginning point no longer corresponds to a link in the nav data.
			pxBeginElt->Data().m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;

			Asura_ListElt< PathPoint >* pxToRemoveElt = pxBaseElt->Next();
			while ( pxToRemoveElt != pxBeginElt )
			{
				Asura_ListElt< PathPoint >* pxNextElt = pxToRemoveElt->Next();
				m_pxPathPoints->Remove( pxToRemoveElt );
				pxToRemoveElt = pxNextElt;
			}
		}

		// Advance the base position.
		pxBaseElt = pxBaseElt->Next();
	}

	return true;
}

bool Asura_Navigation::AStarFunctor::AnnotatePath()
{
	//## Need also to resolve positions to surface?

	// We have already added the start and destination positions in.

	//for every internal waypoint position point (wpp) (not start and destination points)
	//{
	//	vector in:
	//	if first wpp
	//		= previous point position->point position
	//	else
	//		= previous vector out
	//	vector out:
	//		= point position->next point position
	//	
	//	position in:
	//	if first wpp
	//		if start position is outside wp
	//			add a position on vector in
	//	else if link in can be removed
	//		add a position on vector in
	//		
	//	position out:
	//	if last wpp
	//		if destination position is outside wp
	//			add a position on vector out
	//  else if link out can be removed
	//		add a position on vector out
	//
	//	if positions added on link in and out
	//		remove central position
	//}

	if ( m_pxPathPoints->Size() == 0 ) { return true; }

	const float fActorRadius = GetTestParams()->m_fActorRadius;
	const Asura_Vector_3& xStartPos = GetStartPosition();
	const Asura_Vector_3& xDestPos = GetDestinationPosition();

	bool bRemoveLinkIn, bRemoveLinkOut = false;
	Asura_Vector_3 xDirIn, xDirOut = Asura_Vector_3( 0.0f, 0.0f, 1.0f );
	float fLIn, fLOut = 0.0f;
	float fPrevD, fD = 0.0f, fNextD;

	Asura_ListElt< PathPoint >* pxCurrentElt = m_pxPathPoints->Front()->Next();

	const Asura_AI_NavVertex* pxCurrentWP = NULL;
	const Asura_AI_NavVertex* pxNextWP    = Asura_AI_Navigation_System::GetWaypoint( pxCurrentElt->Data().m_usWPID );
	fNextD = pxNextWP ? Asura_Maths::Max( pxNextWP->GetRadius() - fActorRadius, 0.0f ) : 0.0f;

	while ( pxCurrentElt && pxCurrentElt->Next() )
	{
		pxCurrentWP = pxNextWP;
		pxNextWP    = Asura_AI_Navigation_System::GetWaypoint( pxCurrentElt->Next()->Data().m_usWPID );

		fPrevD = fD;
		fD     = fNextD;
		fNextD = pxNextWP ? Asura_Maths::Max( pxNextWP->GetRadius() - fActorRadius, 0.0f ) : 0.0f;

		//	vector in:
		const bool bFirstWPP = ( pxCurrentElt->Prev()->Prev() == NULL );
		if ( bFirstWPP )
		{
			bRemoveLinkIn = true;
            xDirIn        = pxCurrentElt->Data().m_xPosition - pxCurrentElt->Prev()->Data().m_xPosition;
            fLIn          = Asura_Maths::Min( xDirIn.Normalise(), fPrevD + fD );
		}
		else
		{
			bRemoveLinkIn = bRemoveLinkOut;
            xDirIn        = xDirOut;
            fLIn          = fLOut;
		}
		//	vector out:
		{
			const Asura_AI_NavConnection* pxLink = pxCurrentWP ? pxCurrentWP->GetLinkTo( pxCurrentElt->Next()->Data().m_usWPID ) : NULL;
			bRemoveLinkOut = pxLink ?
							 LinkIsValidForRemoval( pxCurrentElt->Data().m_usWPID, *pxCurrentWP,
													*pxLink,
													pxCurrentElt->Next()->Data().m_usWPID, *pxNextWP ) :
							 true;
			xDirOut        = pxCurrentElt->Next()->Data().m_xPosition - pxCurrentElt->Data().m_xPosition;
			fLOut          = Asura_Maths::Min( xDirOut.Normalise(), fD + fNextD );
		}

		//	position in:
		//	if first wpp
		//		if start position is outside wp
		//			add a position on vector in
		//	else if link in can be removed
		//		add a position on vector in
		bool bAddedPositionIn = false;
		if ( bFirstWPP )
		{
			if ( pxCurrentWP && !pxCurrentWP->Contains( xStartPos ) )
			{
				const float fL1 = ( fD / ( fPrevD + fD ) ) * fLIn;

				PathPoint xNewPoint;
				xNewPoint.m_usWPID      = ASURA_INVALID_WAYPOINT_ID;
				xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
				xNewPoint.m_ucFlags     = 0;
				xNewPoint.m_xPosition   = pxCurrentElt->Data().m_xPosition - ( xDirIn * fL1 );
				m_pxPathPoints->InsertBefore( pxCurrentElt, xNewPoint );
			}
			// We have effectively added a position in, regardless of whether we add one here or not - the start position.
			bAddedPositionIn = true;
		}
		else if ( bRemoveLinkIn )
		{
			const float fL1 = ( fD / ( fPrevD + fD ) ) * fLIn;

			PathPoint xNewPoint;
			xNewPoint.m_usWPID      = pxCurrentWP ? pxCurrentElt->Data().m_usWPID : ASURA_INVALID_WAYPOINT_ID;;
			xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
			xNewPoint.m_ucFlags     = 0;
			xNewPoint.m_xPosition   = pxCurrentElt->Data().m_xPosition - ( xDirIn * fL1 );
			m_pxPathPoints->InsertBefore( pxCurrentElt, xNewPoint );
			bAddedPositionIn = true;
		}

		//	position out:
		//	if last wpp
		//		if destination position is outside wp
		//			add a position on vector out
		//  else if link out can be removed
		//		add a position on vector out
		bool bAddedPositionOut = false;
		const bool bLastWPP = ( pxCurrentElt->Next()->Next() == NULL );
		if ( bLastWPP )
		{
			if ( pxCurrentWP && !pxCurrentWP->Contains( xDestPos ) )
			{
				const float fL0 = ( fD / ( fD + fNextD ) ) * fLOut;

				PathPoint xNewPoint;
				xNewPoint.m_usWPID      = ASURA_INVALID_WAYPOINT_ID;
				xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
				xNewPoint.m_ucFlags     = 0;
				xNewPoint.m_xPosition   = pxCurrentElt->Data().m_xPosition + ( xDirOut * fL0 );
				m_pxPathPoints->InsertAfter( pxCurrentElt, xNewPoint );
			}
			// We have effectively added a position out, regardless of whether we add one here or not - the destination position.
			bAddedPositionOut = true;
		}
		else if ( bRemoveLinkOut )
		{
			const float fL0 = ( fD / ( fD + fNextD ) ) * fLOut;

			PathPoint xNewPoint;
			xNewPoint.m_usWPID      = pxCurrentWP ? pxCurrentElt->Data().m_usWPID : ASURA_INVALID_WAYPOINT_ID;
			xNewPoint.m_ucLinkIndex = ASURA_AI_INVALID_LINK_INDEX;
			xNewPoint.m_ucFlags     = 0;
			xNewPoint.m_xPosition   = pxCurrentElt->Data().m_xPosition + ( xDirOut * fL0 );
			m_pxPathPoints->InsertAfter( pxCurrentElt, xNewPoint );
			bAddedPositionOut = true;
		}

		//	if positions added on link in and out
		//		remove central position
		if ( bAddedPositionOut )
		{
			if ( bAddedPositionIn )
			{
				Asura_ListElt< PathPoint >* pxToRemove = pxCurrentElt;
				// We must move past the point that we added.
				pxCurrentElt = pxCurrentElt->Next()->Next();
				m_pxPathPoints->Remove( pxToRemove );
			}
			else
			{
				// We must move past the point that we added.
				pxCurrentElt = pxCurrentElt->Next()->Next();
			}
		}
		else
		{
			pxCurrentElt = pxCurrentElt->Next();
		}
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::FromToPathFindFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::FromToPathFindFunctor::FromToPathFindFunctor()
: m_xFromPos( Asura_Maths::ZeroVector3 )
, m_xToPos( Asura_Maths::ZeroVector3 )
, m_uFromWPID( ASURA_INVALID_WAYPOINT_ID )
, m_uToWPID( ASURA_INVALID_WAYPOINT_ID )
{
}

bool Asura_Navigation::FromToPathFindFunctor::VerifyInputData()
{
	if ( !PARENT::VerifyInputData() ) { return false; }

	const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( m_uFromWPID );
	const Asura_AI_NavVertex* pxToWP   = Asura_AI_Navigation_System::GetWaypoint( m_uToWPID );
	if ( !pxFromWP || !pxToWP ) { return false; }

	// Waypoints must be in the same zone for there to be a valid route between them.
	if ( !( GetTestParams()->m_uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_IGNOREZONES ) && pxFromWP->GetZone() != pxToWP->GetZone() )
	{
		return false;
	}

	return true;
}

float Asura_Navigation::FromToPathFindFunctor::EstimateCostToGoal( const u_int uWPID, const Asura_AI_NavVertex& xWP )
{
	return ( xWP.GetPosition() - m_xToPos ).Magnitude();
}

float Asura_Navigation::FromToPathFindFunctor::EstimateTraversalCost( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP,
																	  const Asura_AI_NavConnection& xLink,
																	  const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
{
	// Cost of the link traversal + cost of the target waypoint (if any).
	return xLink.GetLength() +
		( Asura_AI_Navigation_System::GetWaypointWeight( static_cast< u_short >( uTargetWPID ) ) * Asura_AI_Navigation_System::GetWaypointWeightFactor() );
}

bool Asura_Navigation::FromToPathFindFunctor::GoalAchieved( const u_int uWPID, const Asura_AI_NavVertex& xWP )
{
	return ( uWPID == m_uToWPID );
}

void Asura_Navigation::FromToPathFindFunctor::SetFromToPos( const Asura_Vector_3& xFrom, const Asura_Vector_3& xTo, const TestParams& xTestParams /*= s_xStandardTParams*/ )
{
	m_xFromPos = xFrom;
	m_xToPos   = xTo;
	m_uFromWPID = Asura_AI_Navigation_System::GetNearestWaypointID( m_xFromPos, xTestParams );
	m_uToWPID   = Asura_AI_Navigation_System::GetNearestWaypointID( m_xToPos,   xTestParams );
}

void Asura_Navigation::FromToPathFindFunctor::SetFromToWPIDs( const u_int uFromWPID, const u_int uToWPID )
{
	m_uFromWPID = uFromWPID;
	m_uToWPID   = uToWPID;
	const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( m_uFromWPID );
	if ( pxFromWP )
	{
		m_xFromPos = pxFromWP->GetPosition();
	}
	const Asura_AI_NavVertex* pxToWP = Asura_AI_Navigation_System::GetWaypoint( m_uToWPID );
	if ( pxToWP )
	{
		m_xToPos = pxToWP->GetPosition();
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::FromToWithRangePathFindFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::FromToWithRangePathFindFunctor::FromToWithRangePathFindFunctor()
: m_xBasePos( Asura_Maths::ZeroVector3 )
, m_fIdealRangeSq( 0.0f )
{
}

bool Asura_Navigation::FromToWithRangePathFindFunctor::VerifyInputData()
{
	if ( !PARENT::VerifyInputData() ) { return false; }

	// Can't check base pos, as it may legitimately be the origin.
	return ( m_fIdealRangeSq > 0.0f );
}

float Asura_Navigation::FromToWithRangePathFindFunctor::EstimateTraversalCost( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP,
																			   const Asura_AI_NavConnection& xLink,
																			   const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
{
	const float fParentCost = PARENT::EstimateTraversalCost( uSourceWPID, xSourceWP, xLink, uTargetWPID, xTargetWP );

	// Add a cost for straying from the ideal range.
	Asura_Vector_3 xNearestPoint;
	Asura_Maths::GetNearestPointOnLineSegment( xSourceWP.GetPosition(), xTargetWP.GetPosition(), m_xBasePos, xNearestPoint );
	const float fClosestDistanceSq = ( m_xBasePos - xNearestPoint ).MagnitudeSquared();
	const float fRangeCost = fParentCost * Asura_Maths::Abs( ( m_fIdealRangeSq - fClosestDistanceSq ) / m_fIdealRangeSq );
	return fParentCost + fRangeCost;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::AwayFromPathFindFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::AwayFromPathFindFunctor::AwayFromPathFindFunctor()
: m_uFromWPID( ASURA_INVALID_WAYPOINT_ID )
, m_xFromPos( Asura_Maths::ZeroVector3 )
, m_xAvoidPos( Asura_Maths::ZeroVector3 )
, m_fDist( 0.0f )
, m_fDistSq( 0.0f )
{
}

bool Asura_Navigation::AwayFromPathFindFunctor::VerifyInputData()
{
	if ( !PARENT::VerifyInputData() ) { return false; }

	return ( Asura_AI_Navigation_System::GetWaypoint( m_uFromWPID ) != NULL );
}

float Asura_Navigation::AwayFromPathFindFunctor::EstimateCostToGoal( const u_int uWPID, const Asura_AI_NavVertex& xWP )
{
	const float fDistSq = ( xWP.GetPosition() - m_xAvoidPos ).MagnitudeSquared();
	if ( m_fDistSq > fDistSq )
	{
		return m_fDist - Asura_Maths::Sqrt( fDistSq );
	}
	else
	{
		return 0.0f;
	}
}

float Asura_Navigation::AwayFromPathFindFunctor::EstimateTraversalCost( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP,
																	    const Asura_AI_NavConnection& xLink,
																	    const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
{
	return Asura_AI_Navigation_System::GetWaypointWeight( static_cast< u_short >( uSourceWPID ) ) * Asura_AI_Navigation_System::GetWaypointWeightFactor();
}

bool Asura_Navigation::AwayFromPathFindFunctor::GoalAchieved( const u_int uWPID, const Asura_AI_NavVertex& xWP )
{
	const float fDistSq = ( xWP.GetPosition() - m_xAvoidPos ).MagnitudeSquared();
	return ( fDistSq > m_fDistSq );
}

const Asura_Vector_3& Asura_Navigation::AwayFromPathFindFunctor::GetDestinationPosition() const
{
	if ( GetPathPoints()->Size() > 0 )
	{
		return GetPathPoints()->Back()->Data().m_xPosition;
	}
	else
	{
		Asura_Assert( false, "Asura_Navigation::AwayFromPathFindFunctor::GetDestinationPosition - no path to get destination from" );
		return Asura_Maths::ZeroVector3;
	}
}

void Asura_Navigation::AwayFromPathFindFunctor::SetAvoidPosAndDistance( const Asura_Vector_3& xOrigin, const float fDistance )
{
	m_xAvoidPos = xOrigin;
	m_fDist		= fDistance;
	m_fDistSq	= fDistance * fDistance;
}

void Asura_Navigation::AwayFromPathFindFunctor::SetFromPos( const Asura_Vector_3& xFrom, const TestParams& xTestParams /*= s_xStandardTParams*/ )
{
	m_uFromWPID = Asura_AI_Navigation_System::GetNearestWaypointID( xFrom, xTestParams );
	m_xFromPos  = xFrom;
}

void Asura_Navigation::AwayFromPathFindFunctor::SetFromWPID( const u_int uFromWPID )
{
	m_uFromWPID = uFromWPID;
	const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( m_uFromWPID );
	if ( pxFromWP )
	{
		m_xFromPos = pxFromWP->GetPosition();
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// AwayFromInDirPathFindFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::AwayFromInDirPathFindFunctor::AwayFromInDirPathFindFunctor()
: m_xDirection( Asura_Maths::ZeroVector3 )
{
}

bool Asura_Navigation::AwayFromInDirPathFindFunctor::VerifyInputData()
{
	if ( !PARENT::VerifyInputData() ) { return false; }

	return ( m_xDirection != Asura_Maths::ZeroVector3 );
}

float Asura_Navigation::AwayFromInDirPathFindFunctor::EstimateTraversalCost( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP,
																			 const Asura_AI_NavConnection& xLink,
																			 const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
{
	const float fCost = PARENT::EstimateTraversalCost( uSourceWPID, xSourceWP, xLink, uTargetWPID, xTargetWP );

	Asura_Vector_3 xOriginToToWP = xTargetWP.GetPosition() - GetAvoidPos();
	const float    fOriginToToWPSq = xOriginToToWP.MagnitudeSquared();

	// If we traverse back to the origin position for whatever reason still return a valid cost.
	if ( fOriginToToWPSq < 0.0001f )
	{
		Asura_Vector_3 xFromWPToToWP = xTargetWP.GetPosition() - xSourceWP.GetPosition();
		xFromWPToToWP.Normalise();
		return ( 1.0f - xFromWPToToWP * m_xDirection );
	}

	// Cost for travelling in the incorrect/correct direction.
	xOriginToToWP /= Asura_Maths::Sqrt( fOriginToToWPSq );
	return fCost + ( 1.0f - xOriginToToWP * m_xDirection );
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::BreadthFirstFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::BreadthFirstFunctor::BreadthFirstFunctor()
: m_uStartWPID( ASURA_INVALID_WAYPOINT_ID )
, m_fRange( 0.0f )
, m_uCurrentWPID( ASURA_INVALID_WAYPOINT_ID )
, m_pxAStarData( NULL )
, m_uOpenListFront( 0 )
, m_uOpenListBack( 0 )
{
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::BreadthFirstFunctor::Search()
{
	Status eStatus = BeginIterativeSearch();
	while ( eStatus == STATUS_WORKING )
	{
		eStatus = IterateSearch();
	}
	return eStatus;
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::BreadthFirstFunctor::BeginIterativeSearch()
{
	if ( !VerifyInputData() )
	{
		Asura_Assert( false, "Asura_Navigation::BreadthFirstFunctor::Search - bad input data" );
		return STATUS_IN_ERROR;
	}

	const Asura_AI_NavVertex* pxWPs;
	u_int uNumWPs;
	GetInputBuffer( pxWPs, uNumWPs );

	m_pxAStarData = Asura_Navigation::AStarData::AllocateAStarData( uNumWPs );
	if ( !m_pxAStarData ) { return STATUS_IN_ERROR; }

	m_uCurrentWPID = m_uStartWPID;
	AStarData::AStarNode*         pxCurrentNode       = &m_pxAStarData->m_pxNodeList[ m_uCurrentWPID ];
	AStarData::OpenClosedElement* pxCurrentOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ m_uCurrentWPID ];

	pxCurrentOpenClosed->m_bOpen = true;
	pxCurrentNode->m_fCostToHere = 0.0f;

	// Whilst we're piggy-backing on the AStarData we don't require a sorted open list so we use our own scheme here.
	AddWIPDToOpenList( m_uCurrentWPID );
	return STATUS_WORKING;
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::BreadthFirstFunctor::IterateSearch()
{
	const Asura_AI_NavVertex* pxWPs;
	u_int uNumWPs;
	GetInputBuffer( pxWPs, uNumWPs );

	bool bFinished = false;

	AStarData::AStarNode*         pxCurrentNode;
	AStarData::OpenClosedElement* pxCurrentOpenClosed;
	if ( ( m_uCurrentWPID = GetOpenListWPID() ) != ASURA_INVALID_WAYPOINT_ID )
	{
		pxCurrentNode       = &m_pxAStarData->m_pxNodeList[ m_uCurrentWPID ];
		pxCurrentOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ m_uCurrentWPID ];

		const Asura_AI_NavVertex& xCurrentWP = pxWPs[ m_uCurrentWPID ];
		if ( VisitNode( m_uCurrentWPID, xCurrentWP ) )
		{
			// Done.
			bFinished = true;
		}

		// Visit all the neighbours of the current node and decide whether to add them to the open list.
		for ( u_int u = 0; u < xCurrentWP.GetNumLinks(); ++u )
		{
			const Asura_AI_NavConnection& xLink = xCurrentWP.GetLink( u );
			const u_int uNextWPID = xLink.GetIndex();
			const Asura_AI_NavVertex& xNextWP = pxWPs[ uNextWPID ];
			if ( ShouldIgnoreLink( m_uCurrentWPID, xCurrentWP, xLink, uNextWPID, xNextWP ) )
			{
				continue;
			}

			// Get a pointer to this neighbour's data.
			AStarData::AStarNode*         pxNewNode       = &m_pxAStarData->m_pxNodeList[ uNextWPID ];
			AStarData::OpenClosedElement* pxNewOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ uNextWPID ];

			// If we've already visited this node then move on.
			if ( pxNewOpenClosed->m_bOpen || pxNewOpenClosed->m_bClosed )
			{
				continue;
			}

			const float fNewDistance = pxCurrentNode->m_fCostToHere + xLink.GetLength();
			if ( fNewDistance > m_fRange )
			{
				continue;
			}

			pxNewNode->m_fCostToHere = fNewDistance;

			pxNewOpenClosed->m_bOpen = true;

			AddWIPDToOpenList( pxNewNode->m_uIndex );
		}

		pxCurrentOpenClosed->m_bClosed = true;
		pxCurrentOpenClosed->m_bOpen   = false;
	}

	if ( bFinished || m_uCurrentWPID == ASURA_INVALID_WAYPOINT_ID )
	{
		Asura_Navigation::AStarData::DeallocateAStarData( m_pxAStarData );
		m_pxAStarData = NULL;
		return STATUS_OK;
	}
	else
	{
		return STATUS_WORKING;
	}
}

void Asura_Navigation::BreadthFirstFunctor::SetStartPosition( const Asura_Vector_3& xPos, const TestParams& xTestParams /*= s_xStandardTParams*/ )
{
	m_uStartWPID = Asura_AI_Navigation_System::GetNearestWaypointID( xPos, xTestParams );
}

bool Asura_Navigation::BreadthFirstFunctor::VerifyInputData()
{
	return ( PARENT::VerifyInputData() && m_uStartWPID != ASURA_INVALID_WAYPOINT_ID && Asura_AI_Navigation_System::GetWaypoint( m_uStartWPID ) );
}

void Asura_Navigation::BreadthFirstFunctor::AddWIPDToOpenList( const u_int uWPID )
{
	// Push on the back.
	m_pxAStarData->m_pxOpenList[ m_uOpenListBack++ ].m_uIndex = uWPID;
}

u_int Asura_Navigation::BreadthFirstFunctor::GetOpenListWPID()
{
	// Pop off the front.
	if ( m_uOpenListFront == m_uOpenListBack )
	{
		return ASURA_INVALID_WAYPOINT_ID;
	}
	else
	{
		return m_pxAStarData->m_pxOpenList[ m_uOpenListFront++ ].m_uIndex;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::InDirectionFunctor
//
//////////////////////////////////////////////////////////////////////////////

Asura_Navigation::InDirectionFunctor::InDirectionFunctor()
: m_uStartWPID( ASURA_INVALID_WAYPOINT_ID )
, m_xStartPosition( Asura_Maths::ZeroVector3 )
, m_fRangeSq( 0.0f )
, m_xDirection( Asura_Maths::ZeroVector3 )
, m_uCurrentWPID( ASURA_INVALID_WAYPOINT_ID )
, m_pxAStarData( NULL )
, m_uOpenListFront( 0 )
, m_uOpenListBack( 0 )
{
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::InDirectionFunctor::Search()
{
	Status eStatus = BeginIterativeSearch();
	while ( eStatus == STATUS_WORKING )
	{
		eStatus = IterateSearch();
	}
	return eStatus;
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::InDirectionFunctor::BeginIterativeSearch()
{
	if ( !VerifyInputData() )
	{
		Asura_Assert( false, "Asura_Navigation::InDirectionFunctor::Search - bad input data" );
		return STATUS_IN_ERROR;
	}

	const Asura_AI_NavVertex* pxWPs;
	u_int uNumWPs;
	GetInputBuffer( pxWPs, uNumWPs );

	m_pxAStarData = Asura_Navigation::AStarData::AllocateAStarData( uNumWPs );
	if ( !m_pxAStarData ) { return STATUS_IN_ERROR; }

	m_uCurrentWPID = m_uStartWPID;
	AStarData::AStarNode*         pxCurrentNode       = &m_pxAStarData->m_pxNodeList[ m_uCurrentWPID ];
	AStarData::OpenClosedElement* pxCurrentOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ m_uCurrentWPID ];

	pxCurrentOpenClosed->m_bOpen = true;
	pxCurrentNode->m_fCostToHere = 0.0f;
	pxCurrentNode->m_fCostToGoal = 0.0f;

	m_pxAStarData->AddNodeToOpenList( pxCurrentNode );

	return STATUS_WORKING;
}

Asura_Navigation::SearchFunctor::Status Asura_Navigation::InDirectionFunctor::IterateSearch()
{
	const Asura_AI_NavVertex* pxWPs;
	u_int uNumWPs;
	GetInputBuffer( pxWPs, uNumWPs );

	bool bFinished = false;

	AStarData::AStarNode*         pxCurrentNode;
	AStarData::OpenClosedElement* pxCurrentOpenClosed;
	if ( ( pxCurrentNode = m_pxAStarData->GetOpenListNode() ) != NULL )
	{
		m_uCurrentWPID      = pxCurrentNode->m_uIndex;
		pxCurrentOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ m_uCurrentWPID ];

		const Asura_AI_NavVertex& xCurrentWP = pxWPs[ m_uCurrentWPID ];
		if ( VisitNode( m_uCurrentWPID, xCurrentWP ) )
		{
			// Done.
			bFinished = true;
		}

		// Visit all the neighbours of the current node and decide whether to add them to the open list.
		for ( u_int u = 0; u < xCurrentWP.GetNumLinks(); ++u )
		{
			const Asura_AI_NavConnection& xLink = xCurrentWP.GetLink( u );
			const u_int uNextWPID = xLink.GetIndex();
			const Asura_AI_NavVertex& xNextWP = pxWPs[ uNextWPID ];
			if ( ShouldIgnoreLink( m_uCurrentWPID, xCurrentWP, xLink, uNextWPID, xNextWP ) )
			{
				continue;
			}

			const float fTraversalCost = EstimateTraversalCost( m_uCurrentWPID, xCurrentWP, xLink, uNextWPID, xNextWP );
			// If we get a negative cost then disregard this link.
			if ( fTraversalCost < 0.0f )
			{
				continue;
			}
			const float fNewCostToHere = pxCurrentNode->m_fCostToHere + fTraversalCost;

			// Get a pointer to this neighbour's data.
			AStarData::AStarNode*         pxNewNode       = &m_pxAStarData->m_pxNodeList[ uNextWPID ];
			AStarData::OpenClosedElement* pxNewOpenClosed = &m_pxAStarData->m_pxOpenClosedVector[ uNextWPID ];

			// If we've already visited this node then move on.
			if ( pxNewOpenClosed->m_bOpen || pxNewOpenClosed->m_bClosed )
			{
				continue;
			}

			pxNewNode->m_fCostToHere = fNewCostToHere;
			pxNewNode->m_fCostToGoal = 0.0f;

			pxNewOpenClosed->m_bOpen = true;

			m_pxAStarData->AddNodeToOpenList( pxNewNode );
		}

		pxCurrentOpenClosed->m_bClosed = true;
		pxCurrentOpenClosed->m_bOpen   = false;
	}

	if ( bFinished || !pxCurrentNode )
	{
		Asura_Navigation::AStarData::DeallocateAStarData( m_pxAStarData );
		m_pxAStarData = NULL;
		return STATUS_OK;
	}
	else
	{
		return STATUS_WORKING;
	}
}

void Asura_Navigation::InDirectionFunctor::SetStartPosition( const Asura_Vector_3& xPos, const TestParams& xTestParams /*= s_xStandardTParams*/ )
{
	m_xStartPosition = xPos;
	m_uStartWPID = Asura_AI_Navigation_System::GetNearestWaypointID( xPos, xTestParams );
}

void Asura_Navigation::InDirectionFunctor::SetStartWPID( const u_int uWPID )
{
	m_uStartWPID = uWPID;
	const Asura_AI_NavVertex* pxStartWP = Asura_AI_Navigation_System::GetWaypoint( uWPID );
	if ( pxStartWP )
	{
		m_xStartPosition = pxStartWP->GetPosition();
	}
}

bool Asura_Navigation::InDirectionFunctor::VerifyInputData()
{
	return ( PARENT::VerifyInputData() && m_uStartWPID != ASURA_INVALID_WAYPOINT_ID && Asura_AI_Navigation_System::GetWaypoint( m_uStartWPID ) );
}

bool Asura_Navigation::InDirectionFunctor::ShouldIgnoreLink( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP,
															 const Asura_AI_NavConnection& xLink,
															 const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
{
	Asura_Vector_3 xStartToTargetWP   = xTargetWP.GetPosition() - m_xStartPosition;
	const float    fStartToTargetWPSq = xStartToTargetWP.MagnitudeSquared();

	// Disallow any links that travel over our max range.
	return ( fStartToTargetWPSq > m_fRangeSq );
}

float Asura_Navigation::InDirectionFunctor::EstimateTraversalCost( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP,
																   const Asura_AI_NavConnection& xLink,
																   const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
{
	Asura_Vector_3 xStartToTargetWP   = xTargetWP.GetPosition() - m_xStartPosition;
	const float    fStartToTargetWPSq = xStartToTargetWP.MagnitudeSquared();

	// If we traverse back to the start position for whatever reason still return a valid cost.
	if ( fStartToTargetWPSq < 0.0001f )
	{
		Asura_Vector_3 xSourceWPToTargetWP = xTargetWP.GetPosition() - xSourceWP.GetPosition();
		xSourceWPToTargetWP.Normalise();
		return ( 1.0f - xSourceWPToTargetWP * m_xDirection );
	}

	// Cost for travelling in the incorrect/correct direction.
	xStartToTargetWP /= Asura_Maths::Sqrt( fStartToTargetWPSq );
	return ( 1.0f - xStartToTargetWP * m_xDirection );
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Navigation::Path
//
//////////////////////////////////////////////////////////////////////////////

#ifndef ASURA_NAV_PATH_POINTS_POOL_SIZE
#define ASURA_NAV_PATH_POINTS_POOL_SIZE 146		// 146 * 28b ~= 4K per pool
#endif

Asura_ResizingMempool* Asura_Navigation::Path::s_pxPathPointsMempool = NULL;

Asura_Navigation::Path::Path()
: m_pxAStarFunctor( NULL )
, m_eStatus( STATUS_OK )
, m_xPathPoints()
, m_pxCurrentPointElt( NULL )
, m_pxNextPointElt( NULL )
#ifdef _DEBUG
, m_xOriginalPathPoints()
#endif
{
	m_xPathPoints.SetMemorySupplyFunction( AllocatePathPointNode );
	m_xPathPoints.SetMemoryReturnFunction( DeallocatePathPointNode );
}

Asura_Navigation::Path::~Path()
{
	if ( m_eStatus == STATUS_WORKING )
	{
		StopWork();
	}
	ClearPath();
}

Asura_Navigation::Path::Status Asura_Navigation::Path::FindPath( AStarFunctor& xAStarFunctor )
{
	ClearPath();

	xAStarFunctor.SetInputBuffer( Asura_AI_Navigation_System::GetWaypoints(), Asura_AI_Navigation_System::GetNumberOfWaypoints() );
	xAStarFunctor.SetOutputBuffer( &m_xPathPoints );
	if ( xAStarFunctor.Search() != SearchFunctor::STATUS_OK )
	{
		m_eStatus = STATUS_IN_ERROR;
		return m_eStatus;
	}
	if ( m_xPathPoints.Size() == 0 )
	{
		m_eStatus = STATUS_NO_PATH_FOUND;
		return m_eStatus;
	}

#ifdef _DEBUG
	for ( Asura_LinkList_ConstIt< PathPoint > xPPIter( m_xPathPoints ); !xPPIter.Done(); xPPIter.Next() )
	{
		m_xOriginalPathPoints.PushBack( *xPPIter.GetCurrent() );
	}
#endif

	// Now adjust the path in whatever way necessary.
	// Ordinarily this would include removing redundant waypoints and smoothing the route.
	if ( !xAStarFunctor.PostSearchProcessing() )
	{
		m_eStatus = STATUS_IN_ERROR;
		return m_eStatus;
	}

	m_pxCurrentPointElt = m_xPathPoints.Front();
	m_pxNextPointElt	= m_pxCurrentPointElt->Next();
	m_eStatus = ( m_pxCurrentPointElt && m_pxNextPointElt ) ? STATUS_OK : STATUS_IN_ERROR;
	return m_eStatus;
}

//Asura_Navigation::Path::Status Asura_Navigation::Path::FindPathASync( AStarFunctor& xAStarFunctor )
//{
//	m_pxAStarFunctor = &xAStarFunctor;
//
//	ClearPath();
//
//	xAStarFunctor.SetInputBuffer( Asura_AI_Navigation_System::GetWaypoints(), Asura_AI_Navigation_System::GetNumberOfWaypoints() );
//	xAStarFunctor.SetOutputBuffer( &m_xPathPoints );
//	if ( xAStarFunctor.BeginIterativeSearch() == SearchFunctor::STATUS_WORKING )
//	{
//		m_eStatus = STATUS_WORKING;
//		return m_eStatus;
//	}
//	else
//	{
//		m_eStatus = STATUS_IN_ERROR;
//		return m_eStatus;
//	}
//}

void Asura_Navigation::Path::StopWork()
{
	m_pxAStarFunctor = NULL;
	m_eStatus = STATUS_OK;
}

void Asura_Navigation::Path::ClearPath()
{
	m_eStatus = STATUS_OK;
	while ( m_xPathPoints.Size() )
	{
		m_xPathPoints.Remove( m_xPathPoints.Front() );
	}
	m_pxCurrentPointElt = NULL;
    m_pxNextPointElt    = NULL;

#ifdef _DEBUG
	while ( m_xOriginalPathPoints.Size() )
	{
		m_xOriginalPathPoints.Remove( m_xOriginalPathPoints.Front() );
	}	
#endif
}

float Asura_Navigation::Path::GetPathTotalLength() const
{
	return GetPathLengthAfter( m_xPathPoints.Front() );
}

float Asura_Navigation::Path::GetPathRemainingLength() const
{
	return GetPathLengthAfter( GetCurrentPathPoint() );
}

bool Asura_Navigation::Path::UpdateNearestPositionOnPath( const Asura_Vector_3& xPosition, float fLookAheadDistance,
														  Asura_Vector_3& xPathPosition, bool& bIsDestination, const Asura_Navigation::APParams& xParams /*= s_xStandardAPParams*/ ) const
{
	if ( m_xPathPoints.Size() < 2 )
	{
        xPathPosition  = xPosition;
		bIsDestination = true;
		return false;
	}

	//for ( all segments from current segment until the end of the route )
	//{
	//	get the nearest point on the segment
	//	if ( nearest point distance to us < best nearest point distance to us )
	//	{
	//		best nearest point = nearest point
	//		best segment = current segment
	//	}
	//	else // we're getting further away
	//	{
	//		stop
	//	}
	//}
	//
	//while ( look ahead distance left )
	//{
	//	add on look ahead distance in direction of best segment until used up or segment end
	//	if ( look ahead distance left ) move to next segment
	//}

	float fLeastDistanceSq = Asura_Maths::LargeFloat;
	Asura_ListElt< PathPoint >* pxStartElt = m_pxCurrentPointElt;
	Asura_ListElt< PathPoint >* pxEndElt   = m_pxNextPointElt;
	while ( pxStartElt && pxEndElt )
	{
		Asura_Vector_3 xNearestPosition;
		Asura_Maths::GetNearestPointOnLineSegment( pxStartElt->Data().m_xPosition, pxEndElt->Data().m_xPosition, xPosition, xNearestPosition );
		const float fDistanceSq = ( xNearestPosition - xPosition ).MagnitudeSquared();
		if ( fDistanceSq <= fLeastDistanceSq )
		{
            fLeastDistanceSq    = fDistanceSq;
			xPathPosition		= xNearestPosition;
			m_pxCurrentPointElt = pxStartElt;
            m_pxNextPointElt    = pxEndElt;
		}
		else
		{
			break;
		}

		pxStartElt = pxEndElt;
        pxEndElt   = pxEndElt->Next();
	}

	pxStartElt = m_pxCurrentPointElt;
	pxEndElt   = m_pxNextPointElt;
	while ( pxStartElt && pxEndElt && fLookAheadDistance > 0.0f )
	{
		Asura_Vector_3 xSegmentVector = pxEndElt->Data().m_xPosition - pxStartElt->Data().m_xPosition;
		xSegmentVector.Normalise();
		const float fToSegmentEnd = ( pxEndElt->Data().m_xPosition - xPathPosition ).Magnitude();

		xPathPosition      += xSegmentVector * Asura_Maths::Min( fLookAheadDistance, fToSegmentEnd );
		fLookAheadDistance -= fToSegmentEnd;

		pxStartElt = pxEndElt;
        pxEndElt   = pxEndElt->Next();
	}

	bIsDestination = Asura_AI_Navigation_System::AtPosition( xPosition, xPathPosition, xParams );
	
	return true;
}

#ifdef _DEBUG
void Asura_Navigation::Path::RenderPath( const Asura_LinkList< PathPoint >& xPath, u_int uARGB /*= 0xffffffff*/ )
{
	Asura_ListElt< PathPoint >* pxCurrElt = xPath.Front();
	if ( pxCurrElt )
	{
		Asura_DebugLineRender::AddCross( pxCurrElt->Data().m_xPosition, uARGB );
	}
	else
	{
		return;
	}

	Asura_ListElt< PathPoint >* pxNextElt = pxCurrElt->Next();
	while ( pxNextElt )
	{
		Asura_DebugLineRender::AddArrow( pxCurrElt->Data().m_xPosition, pxNextElt->Data().m_xPosition, uARGB );
		Asura_DebugLineRender::AddCross( pxNextElt->Data().m_xPosition, uARGB );

		pxCurrElt = pxNextElt;
		pxNextElt = pxNextElt->Next();
	}
}

void Asura_Navigation::Path::RenderPathWPs( const Asura_LinkList< PathPoint >& xPath, u_int uARGB /*= 0xffffffff*/ )
{
	Asura_ListElt< PathPoint >* pxCurrElt = xPath.Front();
	while ( pxCurrElt )
	{
		const Asura_AI_NavVertex* pxWP = Asura_AI_Navigation_System::GetWaypoint( pxCurrElt->Data().m_usWPID );
		if ( pxWP )
		{
			Asura_Matrix_3x3 xWPOrientM;
			Asura_Maths::OrientationFromNormal( xWPOrientM, -pxWP->GetNormal() );
			Asura_Quat xWPOrientQ;
			xWPOrientQ.FromMatrix( xWPOrientM );
			Asura_DebugLineRender::AddCylinder( pxWP->GetPosition(), xWPOrientQ, pxWP->GetRadius(), pxWP->GetRadius(), uARGB, 8 );
		}
		pxCurrElt = pxCurrElt->Next();
	}
}
#endif

void Asura_Navigation::Path::InitialisePathMempools()
{
	s_pxPathPointsMempool = new Asura_ResizingMempool( ASURA_NAV_PATH_POINTS_POOL_SIZE,
													   Asura_LinkList< PathPoint >::GetMemoryElementSize(),
													   __FILE__,
													   __LINE__ );
}

void Asura_Navigation::Path::ShutdownPathMempools()
{
	Asura_Assert( s_pxPathPointsMempool->IsEmpty(), "Asura_Navigation::Path::ShutdownPathMempools - s_pxPathPointsMempool not empty" );
	delete s_pxPathPointsMempool;
	s_pxPathPointsMempool = NULL;
}

float Asura_Navigation::Path::GetPathLengthAfter( const Asura_ListElt< PathPoint >* pxElt )
{
	if ( !pxElt ) { return 0.0f; }

	float fLengthSq = 0.0f;
	const Asura_ListElt< PathPoint >* pxNextElt = pxElt->Next();
	while ( pxElt && pxNextElt )
	{
		fLengthSq += ( pxNextElt->Data().m_xPosition - pxElt->Data().m_xPosition ).MagnitudeSquared();

        pxElt     = pxNextElt;
		pxNextElt = pxNextElt->Next();
	}

	return Asura_Maths::Sqrt( fLengthSq );
}

void* Asura_Navigation::Path::AllocatePathPointNode()
{
	return s_pxPathPointsMempool->Allocate();
}

void Asura_Navigation::Path::DeallocatePathPointNode( void* pNode )
{
	s_pxPathPointsMempool->Deallocate( pNode );
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // #ifdef ASURA_USE_NEW_NAV_STUFF
