/**
*
* \file Asura_AnimTable_Handler.cpp
*
* Created by Richard May - 18/06/2002 16:38:21
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/



/************/
/* Includes.*/
/************/

#include "Core/Asura.h"

#ifdef ASURA_DISABLE_ANIM_TABLE_CODE
#else

// This file's header.
#include "Animation/Asura_AnimTable_System.h"

// Engine independent.
#include "Animation/Asura_AnimTable.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Stream.h"

/***************/
/* Static Data */
/***************/

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
Asura_TemplateHashTree<Asura_AnimTable_Handler>	Asura_AnimTable_System::s_xHandlerTree;
Asura_AnimTable_Handler*						Asura_AnimTable_System::s_pxCurrentHandler	= NULL;
#else
Asura_AnimTable_Handler		Asura_AnimTable_System::s_xHandler;
#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)


/*********************************/
/* Basic Init/Shutdown Functions */
/*********************************/

void Asura_AnimTable_Handler::Reset(void)
{
	// Reset the table...
    m_xLoadedTables.Reset();
}



/*******************/
/* Usage Functions */
/*******************/

/*************************************************************
 *
 * Name          : Asura_AnimTable_Handler::AddNewTable
 *
 * Author        : Richard May
 *
 * Description   : Creates & loads the data for a new animation
 *				   table from the file specified
 *
 * Parameters    : void
 *
 * Return value  : void
 *
 *************************************************************/
Asura_AnimTable* Asura_AnimTable_Handler::AddNewTable( Asura_Hash_ID uNameHash, u_int uNoOfPrimaryStates, u_int uNoOfSecondaryStates )
{
	bool bDidExist = true;
	Asura_AnimTable*	pxAnimTable	= m_xLoadedTables.AddOrFindNode(uNameHash,&bDidExist,true);

	// Do we have an anim-table?
	if(pxAnimTable == NULL||bDidExist)
	{
		Asura_Assertf(false,("Can't add animtable %0x%x, already exisits?",uNameHash));
		return(NULL);
	}
	else
	{
		pxAnimTable->SetInitialData(uNameHash, uNoOfPrimaryStates, uNoOfSecondaryStates);
		return(pxAnimTable);
	}
}
/* end of function AddNewTable */


/*************************************************************
 *
 * Name          : Asura_AnimTable_Handler::IsAnimLoaded
 *
 * Author        : Richard May
 *
 * Description   : Checks the list of loaded animations to 
 *				   see whether the specified animation has 
 *				   already been loaded
 *
 * Parameters    : uNameHash - hash id of anim to look for
 *				   addToList - boolean - controls whether or not
 *							   the animation name is added to the
 *							   loaded list if it is found not to be
 *							   present
 *
 * Return value  : bool - true if anim is already loaded
 *
 *************************************************************/
bool Asura_AnimTable_Handler::IsAnimLoaded( Asura_Hash_ID uNameHash, bool bAddToList )
{
	Asura_AnimTable*	pxAnimTable	= m_xLoadedTables.GetDataByHash(uNameHash);

	// Do we need to add an anim-table?
	if((pxAnimTable == NULL) && bAddToList)
	{
		pxAnimTable	= m_xLoadedTables.AddNode(uNameHash);
	}

	// Did we find the table?
	return(pxAnimTable != NULL);
}
/* end of function IsAnimLoaded */



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimTable_System::Initialise
// Author       : Andy Tate
// Date         : 29-Mar-2006
//
// Description  : Initialise the animation-table system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimTable_System::Initialise (void)
{
	// Nothing to initialise...
}

/// End of function 'Asura_AnimTable_System::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimTable_System::Reset
// Author       : Andy Tate
// Date         : 29-Mar-2006
//
// Description  : Resets the animation-table system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimTable_System::Reset (void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	// Reset the current instance...
	GetCurrentHandler().Reset();
#endif
}

/// End of function 'Asura_AnimTable_System::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimTable_System::Shutdown
// Author       : Andy Tate
// Date         : 29-Mar-2006
//
// Description  : Shuts down the animation-system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimTable_System::Shutdown (void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	// Reset the current instance...
	s_xHandler.Reset();
#endif
}

/// End of function 'Asura_AnimTable_System::Shutdown'. ///


#if defined(ASURA_USE_MULTIPLE_HANDLERS)

void Asura_AnimTable_System::ReparentHandler(Asura_AnimTable_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash)
{
	Asura_AnimTable_Handler*	pxParentHandler	= s_xHandlerTree.GetDataByHash(uNewParentHash);

	if(pxParentHandler != NULL)
	{
		xSourceHandler.m_xLoadedTables.SetParent(&(pxParentHandler->m_xLoadedTables));
	}
}

void Asura_AnimTable_System::ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash)
{
	Asura_AnimTable_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uSourceHash);

	if(pxHandler != NULL)
	{
		ReparentHandler(*pxHandler, uNewParentHash);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimTable_System::CreateHandler
// Author       : Andy Tate
// Date         : 29-Mar-2006
//
// Description  : Creates an animation-table handler with the specified ID.
//
// Parameters   : The hash-ID for this anim-table and its parent handler.
//
// Return value : A pointer to the handler created, NULL if failed.
////////////////////////////////////////////////////////////////////////////////

Asura_AnimTable_Handler* Asura_AnimTable_System::CreateHandler (const Asura_Hash_ID uHash,
																const Asura_Hash_ID uParentHandlerHash)
{
	Asura_AnimTable_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHash);

		if(pxHandler != NULL)
		{
			ReparentHandler(*pxHandler, uParentHandlerHash);
		}
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an anim-table handler with this hash-ID.");
	}

	// Return the handler...
	return(pxHandler);
}

/// End of function 'Asura_AnimTable_System::CreateHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimTable_System::DestroyHandler
// Author       : Andy Tate
// Date         : 29-Mar-2006
//
// Description  : Destroys the animation-table handler with the specified ID.
//
// Parameters   : The hash-ID of the anim-table to destroy.
//
// Return value : "true" if successful, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimTable_System::DestroyHandler (const Asura_Hash_ID uHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

/// End of function 'Asura_AnimTable_System::DestroyHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimTable_System::WriteHandlerLog
// Author       : Paul Cardy
// Date         : 20-12-07
//
// Description  : Writes a log of the contents of the supplied handler
//
// Parameters   : The hash-ID of the handler to log and the filename to log to
//
// Return value : None
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimTable_System::WriteHandlerReport(u_int uHandler, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 64;
	char szLogBuffer[uLOGBUFFERSIZE];

	Asura_AnimTable_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHandler);
	if (pxHandler)
	{
		if (!pxHandler->m_xLoadedTables.GetNumberOfNodes())
		{
			xLog << "Anim Table:\t\t\tNone\n";
		}
		else
		{
			Asura_TemplateHashTree_Iterator<Asura_AnimTable> xIt(pxHandler->m_xLoadedTables);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_AnimTable* pxAnimTable = xIt.GetCurrent();
				Asura_PrintfToBfr( szLogBuffer, "Anim Table:\t(0x%08x)\n", uLOGBUFFERSIZE, true, pxAnimTable->GetNameHashID());
				xLog << szLogBuffer;
			}		
		}
	}
	else
	{
		xLog << "Anim Table:\t\t\tNo Handler\n";
	}
}

/// End of function 'Asura_AnimTable_System::WriteHandlerLog'. ///

#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)



/*************************************************************
 *
 * Name          : Asura_AnimTable_System::ConvertStringToAnimState
 *
 * Author        : Richard May
 *
 * Description   : Converts a given string to an animation state
 *				   and returns is as an int (as there are three
 *				   different kinds of states it could be - primary,
 *				   weapon and secondary). This function uses tables
 *				   that must be updated whenever the state lists are
 *				   changed
 *
 * Parameters    : state - The string 'name' to be converted
 *
 * Return value  : int - the value of the corresponding state
 *
 *************************************************************/
int Asura_AnimTable_System::ConvertHashToAnimState( Asura_Hash_ID uID )
{	
	/// Richard May - 20/04/2004 11:53:49 - compare to project-specific states
	int iState = Project_ConvertHashToAnimState(uID);
	if(iState == -1)
	{
		Asura_Assertf( iState != -1, ("Failed to convert anim state. Returning zero") );
		return 0;
	}

	return iState;
}
/* end of function Asura_AnimTable_System::ConvertHashToAnimState */

#ifdef ASURA_TOOLS

int	Asura_AnimTable_System::GetNumTables()
{
	return GetCurrentHandler().m_xLoadedTables.GetNumberOfNodes();
}

const char* Asura_AnimTable_System::GetTableNameByIndex(u_int uIndex)
{
	return GetCurrentHandler().m_xLoadedTables.GetDataByIndex(uIndex)->GetName();
}

Asura_Hash_ID Asura_AnimTable_System::GetTableHashByIndex(u_int uIndex)
{
	return  GetCurrentHandler().m_xLoadedTables.GetDataByIndex(uIndex)->m_uNameHash;
}
#endif // ASURA_TOOLS

#endif// #else not ASURA_DISABLE_ANIM_TABLE_CODE


/****************/
/* End of file. */
/****************/
