//
// Asura_Animation.cpp
//
// Created by Kevin Lea - 12/06/2001 11:13:19
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura.h"

// This file's header.
#include "Animation/Asura_Animation.h"

// Engine independent.
#include "Animation/Asura_AnimationAttachment.h"

#ifdef ASURA_USE_ANIMATION_IK_SOLVER
#include "Animation/IK/Asura_Animation_IK_Solver.h"
#endif // ASURA_USE_ANIMATION_IK_SOLVER

#include "Animation/Asura_Animation_Posable.h"
#ifndef ASURA_DISABLE_ANIM_TABLE_CODE
#include "Animation/Asura_AnimTable_System.h"
#endif
#include "Animation/Asura_AnimationController_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Mempool.h"
#include "Core/Asura_Stream.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"
#include "Light/Asura_GI_Solution.h"
#include "Light/Asura_Lighting_Solution.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_SortedRender.h"
#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
#include "Shadows/Asura_Shadow_Instance.h"
#include "Shadows/Asura_Shadow_System.h"
#endif	// defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
#ifdef ASURA_MAIN
#include "Sound/Asura_Sound.h"
#endif
#ifdef ASURA_USE_DIALOGUE_SYSTEM
#include "Sound/Asura_Dialogue_System.h"
#endif
#include "Sound/Asura_Sound_Event.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "Sound/Asura_Sound_Event_System.h"
#ifdef ASURA_USE_AUDIO_SYSTEM
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Audio_Voice.h"
#endif
#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
#include "Rendering/Asura_Project_Skin_Elements.h"
#endif


#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif //defined(ASURA_USE_STREAMING)

// C O N S T A N T S /////////////////////////////////////////////////////////

// Should be a power of 2.
static const u_int MIN_FRAMES_BEFORE_INTERSECTING_MODULE_LIST_REFRESH = 8;

// Add more delay if we are doing a manual update of the containing zone.
static const u_int NUM_EXTRA_FRAMES_BEFORE_UPDATE_CONTAINING_ZONE = 16;

// G L O B A L S /////////////////////////////////////////////////////////////

// For slightly more helpful assertions when the Asura_HashedFilename_System
// isn't available.
static const char* g_szPreHashNameTemp = NULL;

// Hash-ID used to identify the root group-controller...
static const Asura_Hash_ID	g_uRootGroupHash	= 0x003580E2;	// Asura_GetHashID("Root");

// S T A T I C   D A T A /////////////////////////////////////////////////////

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
Asura_TemplateHashTree<Asura_Animation_Handler>	Asura_Animation_System::s_xHandlerTree;
Asura_Animation_Handler*						Asura_Animation_System::s_pxCurrentHandler	= NULL;
#else
Asura_Animation_Handler		Asura_Animation_System::s_xHandler;
#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)

Asura_Animation::ProjectInitialiseCallback*	Asura_Animation::s_pfnProjectInitialiseCallback	= NULL;
Asura_Animation::ProjectShutdownCallback*	Asura_Animation::s_pfnProjectShutdownCallback	= NULL;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_Animation_Handler::Asura_Animation_Handler(void)
{
}

void Asura_Animation_Handler::ReleaseAll(void)
{	
	m_xAnimTree.Reset();
	
	ReleaseSkins();
}

Asura_Animation_Handler::~Asura_Animation_Handler()
{
	ReleaseAll();
}

void Asura_Animation_Handler::Reset()
{
#if _DEBUG
    Asura_OutputToDebugger(( "Asura_Animation_System::Reset(): %i anims at reset\n", m_xAnimTree.GetNumberOfNodes() ));
#endif

    m_xAnimTree.Reset();
    ReleaseSkins();
}

void DeleteContents( Asura_Hash_ID uHash, Asura_Hierarchy_Skin** ppxSkin )
{
	delete *ppxSkin;
	*ppxSkin = NULL;
}

void Asura_Animation_Handler::ReleaseSkins()
{
	m_xSkinTree.Traverse( DeleteContents );
	m_xSkinTree.Reset();
}

// KJL 02/12/00 17:28:29 - returns true if this is a new animation
// anim gets set to the animation's storage space
bool Asura_Animation_Handler::NewAnimation (const char* szAnimName, Asura_Hierarchy_Anim*& pxAnim)
{
	const Asura_Hash_ID	uHash	= Asura_GetHashID(szAnimName);

	// Do we already have this animation?
	pxAnim	= m_xAnimTree.GetDataByHash(uHash);

	if(pxAnim == NULL)
	{
		// No - create one...
		pxAnim	= m_xAnimTree.AddNode(uHash);
		return(true);
	}
	else
	{
		// Yes - are we able to overwrite it?
#if defined(ASURA_TOOLS)
		if(pxAnim->m_bAllowOverwriteOnLoad)
		{
			//Reset the contents of the existing rpxAnim ,and allow it to be replaced. RAR 14/11/2003
			pxAnim->Reset();
			return(true);
		}
		else
#endif	// defined(ASURA_TOOLS)
		{
			return false;
		}
	}
}

// Kevin Lea - 10/08/2001 15:10:55 - returns true if this is a new skin
bool Asura_Animation_Handler::NewSkin (const char* szSkinName, Asura_Hierarchy_Skin*& pxSkin)
{
	const Asura_Hash_ID	uHash	= Asura_GetHashID(szSkinName);

	// Do we already have this skin?
	pxSkin	= GetSkinByHash( uHash );

	if(pxSkin == NULL)
	{
		// No - allocate new one...		
		Asura_Hierarchy_Skin* pxNewSkin = Asura_Hierarchy_System::Project_AllocateSkin();
		*m_xSkinTree.AddNode(uHash) = pxNewSkin;
		pxSkin	= pxNewSkin;
		
		// setup data
		if( pxSkin )
		{
			pxSkin->Name			= szSkinName;
			pxSkin->m_uNameHashID	= uHash;
		}
	
		return (pxSkin != NULL);
	}
	else
	{
		// Yes - are we able to overwrite it?
#if defined(ASURA_TOOLS)
		if(pxSkin->m_bAllowOverwriteOnLoad)
		{
			//Reset the contents of the existing rpxSkin ,and allow it to be replaced. RAR 14/11/2003
			pxSkin->Reset();
			pxSkin->Name			= szSkinName;
			pxSkin->m_uNameHashID	= uHash;
			return(true);
		}
		else
#endif	// defined(ASURA_TOOLS)
		{
			return false;
		}
	}
}

// accessors
Asura_Hierarchy_Skin* Asura_Animation_Handler::GetSkin(const char* szSkinName)
{ 
	Asura_Hierarchy_Skin** ppxSkin = m_xSkinTree.GetDataByName(szSkinName);
	return ppxSkin ? *ppxSkin : NULL; 
}

Asura_Hierarchy_Skin* Asura_Animation_Handler::GetSkinByHash(Asura_Hash_ID uHash)
{ 
	Asura_Hierarchy_Skin** ppxSkin = m_xSkinTree.GetDataByHash(uHash);
	return ppxSkin ? *ppxSkin : NULL; 
}

#ifdef ASURA_TOOLS
Asura_Hierarchy_Skin* Asura_Animation_Handler::GetSkin(int iIndex)
{
	Asura_Hierarchy_Skin** ppxSkin = m_xSkinTree.GetDataByIndex(iIndex);
	return ppxSkin ? *ppxSkin : NULL; 
}
#endif // ASURA_TOOLS


//////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Asura_Animation_System **
**------------------------*/

Asura_ResizingMempool	Asura_Animation_System::s_xSoundInstanceMempool(32, sizeof(Asura_Animation_SoundInstance), __FILE__, __LINE__);


void Asura_Animation_System::Initialise(void)
{
	// Initialise the related systems...
	Asura_Hierarchy_System::Initialise();
#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::Initialise();
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
	Asura_Platform_Animation_RenderCache::Initialise();
#endif
}

void Asura_Animation_System::Shutdown(void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	Asura_Animation_Handler&	xHandler	= GetCurrentHandler();

	// Reset the current instance...
	xHandler.Reset();
#endif

	// Shutdown the related systems...
	Asura_Hierarchy_System::Shutdown();
#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::Shutdown();
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
	Asura_Platform_Animation_RenderCache::Shutdown();
#endif
}

void Asura_Animation_System::Reset(void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	// Reset the current instance...
	GetCurrentHandler().Reset();
#endif

	// Reset the related systems...
	Asura_Hierarchy_System::Reset();
#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::Reset();
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)

	s_xSoundInstanceMempool.Reset();
}

Asura_Animation_SoundInstance *Asura_Animation_System::AllocateSoundInstance()
{
	Asura_Animation_SoundInstance* pxSoundInstance = static_cast<Asura_Animation_SoundInstance *>(s_xSoundInstanceMempool.Allocate());
	if(pxSoundInstance)
	{
		pxSoundInstance->Initialise();
	}
	return pxSoundInstance;
}

void Asura_Animation_System::DeallocateSoundInstance(Asura_Animation_SoundInstance *pxSoundInstance)
{
	s_xSoundInstanceMempool.Deallocate(pxSoundInstance);
}


#if defined(ASURA_USE_MULTIPLE_HANDLERS)
void Asura_Animation_System::ReparentHandler(Asura_Animation_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash)
{
	Asura_Animation_Handler*	pxParentHandler	= s_xHandlerTree.GetDataByHash(uNewParentHash);

	if(pxParentHandler != NULL)
	{
		xSourceHandler.m_xAnimTree.SetParent(&(pxParentHandler->m_xAnimTree));
		xSourceHandler.m_xSkinTree.SetParent(&(pxParentHandler->m_xSkinTree));
	}
}

void Asura_Animation_System::ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash)
{
	Asura_Animation_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uSourceHash);

	if(pxHandler != NULL)
	{
		ReparentHandler(*pxHandler, uNewParentHash);
	}

#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::ReparentHandler(uSourceHash, uNewParentHash);
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
}

void Asura_Animation_System::SetCurrentHandler(const Asura_Hash_ID uHash)
{
	s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash);

#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::SetCurrentHandler(uHash);
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
}

Asura_Animation_Handler* Asura_Animation_System::CreateHandler (const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)
{
	Asura_Animation_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHash);

		if(pxHandler != NULL)
		{
			ReparentHandler(*pxHandler, uParentHandlerHash);
		}

		// ...and create the sub-system handlers...
//		Asura_Hierarchy_System::CreateHandler(uHash, uParentHandlerHash);
#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
		Asura_AnimTable_System::CreateHandler(uHash, uParentHandlerHash);
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an animation handler with this hash-ID.");
	}

	// Return the handler...
	return(pxHandler);
}

bool Asura_Animation_System::DestroyHandler (const Asura_Hash_ID uHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Destroy the sub-system handlers...
//	Asura_Hierarchy_System::DestroyHandler(uHash);
#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::DestroyHandler(uHash);
#endif	// !defined(ASURA_DISABLE_ANIM_TABLE_CODE)

	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

void Asura_Animation_System::WriteHandlerReport(u_int uHandler, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 96;
	char szLogBuffer[uLOGBUFFERSIZE];

	Asura_Animation_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHandler);
	if (pxHandler)
	{
		if (!pxHandler->m_xAnimTree.GetNumberOfNodes())
		{
			xLog << "Hierarchy Anim:\t\t\tNone\n";
		}
		else
		{
			Asura_TemplateHashTree_Iterator<Asura_Hierarchy_Anim> xIt(pxHandler->m_xAnimTree);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_Hierarchy_Anim* pxAnim = xIt.GetCurrent();
				Asura_PrintfToBfr( szLogBuffer, "Hierarchy Anim:\t(0x%08x)\t%s\n", uLOGBUFFERSIZE, true, pxAnim->GetHashID(), pxAnim->GetName());
				xLog << szLogBuffer;
			}		
		}

		if (!pxHandler->m_xAnimTree.GetNumberOfNodes())
		{
			xLog << "Hierarchy Skin:\t\t\tNone\n";
		}
		else
		{
			Asura_TemplateHashTree_Iterator<Asura_Hierarchy_Skin*> xIt(pxHandler->m_xSkinTree);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_Hierarchy_Skin** ppxSkin = xIt.GetCurrent();
				Asura_PrintfToBfr( szLogBuffer, "Hierarchy Skin:\t(0x%08x)\t%s\n", uLOGBUFFERSIZE, true, (*ppxSkin)->GetHashID(), (*ppxSkin)->GetName());
				xLog << szLogBuffer;
			}		
		}
	}
	else
	{
		xLog << "Hierarchy:\t\t\tNo Handler\n";
	}

#if !defined(ASURA_DISABLE_ANIM_TABLE_CODE)
	Asura_AnimTable_System::WriteHandlerReport(uHandler, xLog);
#endif	//!defined(ASURA_DISABLE_ANIM_TABLE_CODE)
}

void Asura_Animation_System::CreateAndSetHandler (const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)
{
	CreateHandler(uHash, uParentHandlerHash);
	SetCurrentHandler(uHash);
}

#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)
//////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_MAIN //Don't want Asura_Animation and Asura_AnimationStatus in plugins etc RAR 03/10/2006

Asura_Animation::Asura_Animation(const char* szSkinName)
:
#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
  m_uMaterialOverrideMapID(ASURA_HASH_ID_UNSET),
#else
#endif
m_uSoundFrameStamp(0)
, m_uNumTimerUpdates(0)
, m_pxFirstAttachment(NULL)
, m_pProjectData(NULL)
, m_xBSCentre(Asura_Maths::ZeroVector3)
, m_fTightBSRadius(-1.f)
, m_fCoarseBSRadius(-1.f)
, m_fCoarseBSIncEffectsRadius(-1.f)
, m_fPrevCoarseBSIncEffectsRadius(-1.f)
{
	m_uSkinCount = 0;
	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		m_apxSkins[i] = NULL;
	}
	Asura_Mem::Set(m_xLOSFlags, 0, m_xLOSFlags.GetSize());
	Asura_Mem::Set(m_xMergedBBValidFlags, 0, m_xMergedBBValidFlags.GetSize());
	m_pxMergedSectionBB = NULL;


	InvalidateBonesUpdateFrameStamp();

	m_pxRootControllerGroup	= new Asura_AnimationController_Group (g_uRootGroupHash, AAC_FLAG_ACTIVE | AACG_FLAG_PRESERVE_TWEEN_STORE);

	m_pxRootControllerGroup->SetAsuraAnimation(this);
	m_xPrimaryWrapper.SetAnimation(this);

#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
	m_pxRenderCache = Asura_Platform_Animation_RenderCache::Create();
#endif

	SetSkin( szSkinName );
	Initialise();
}

Asura_Animation::Asura_Animation(Asura_Hash_ID uHashSkin)
:
#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
  m_uMaterialOverrideMapID(ASURA_HASH_ID_UNSET),
#endif
 m_uSoundFrameStamp(0)
, m_uNumTimerUpdates(0)
, m_pxFirstAttachment(NULL)
, m_pProjectData(NULL)
, m_xBSCentre(Asura_Maths::ZeroVector3)
, m_fTightBSRadius(-1.f)
, m_fCoarseBSRadius(-1.f)
, m_fCoarseBSIncEffectsRadius(-1.f)
, m_fPrevCoarseBSIncEffectsRadius(-1.f)
{

	m_uSkinCount = 0;
	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		m_apxSkins[i] = NULL;
	}
	Asura_Mem::Set(m_xLOSFlags, 0, m_xLOSFlags.GetSize());
	Asura_Mem::Set(m_xMergedBBValidFlags, 0, m_xMergedBBValidFlags.GetSize());
	m_pxMergedSectionBB = NULL;

	InvalidateBonesUpdateFrameStamp();

	m_pxRootControllerGroup	= new Asura_AnimationController_Group (g_uRootGroupHash, AAC_FLAG_ACTIVE | AACG_FLAG_PRESERVE_TWEEN_STORE);

	m_pxRootControllerGroup->SetAsuraAnimation(this);
	m_xPrimaryWrapper.SetAnimation(this);

#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
	m_pxRenderCache = Asura_Platform_Animation_RenderCache::Create();
#endif

	SetSkin(uHashSkin);
	Initialise();
}

// Copy-constructor-lite for ragdoll
Asura_Animation::Asura_Animation( const Asura_Animation* pxAnim )
:
#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
  m_uMaterialOverrideMapID(ASURA_HASH_ID_UNSET),
#endif
 m_uSoundFrameStamp(0)
, m_uNumTimerUpdates(0)
, m_pxFirstAttachment(NULL)
, m_pProjectData(NULL)
, m_xBSCentre(Asura_Maths::ZeroVector3)
, m_fTightBSRadius(-1.f)
, m_fCoarseBSRadius(-1.f)
, m_fCoarseBSIncEffectsRadius(-1.f)
, m_fPrevCoarseBSIncEffectsRadius(-1.f)
{

	m_uSkinCount = 0;
	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		m_apxSkins[i] = NULL;
	}
	Asura_Mem::Set(m_xLOSFlags, 0, m_xLOSFlags.GetSize());
	Asura_Mem::Set(m_xMergedBBValidFlags, 0, m_xMergedBBValidFlags.GetSize());
	m_pxMergedSectionBB = NULL;

	InvalidateBonesUpdateFrameStamp();

	m_pxRootControllerGroup	= new Asura_AnimationController_Group (g_uRootGroupHash, AAC_FLAG_ACTIVE | AACG_FLAG_PRESERVE_TWEEN_STORE);

	m_pxRootControllerGroup->SetAsuraAnimation(this);
	m_xPrimaryWrapper.SetAnimation(this);

#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
	m_pxRenderCache = Asura_Platform_Animation_RenderCache::Create();
#endif

	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
	    SetSkin( pxAnim->m_apxSkins[i] ? pxAnim->m_apxSkins[i]->Name.GetString() : NULL, i );
	}

	Initialise();
    Asura_Animation::Copy( pxAnim, false, false, true );
}

void Asura_Animation::Copy( const Asura_Animation* pxSrcAnim, bool bCopySkin, bool bCopyOneShotAnims, bool bCopyAttachments )
{
	// Remove all attachments.
    UnlinkAllAnimations();		    

	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		if( m_apxSkins[i] != pxSrcAnim->m_apxSkins[i] )
		{
			if( bCopySkin )
			{
				SetSkin( pxSrcAnim->m_apxSkins[i] ? pxSrcAnim->m_apxSkins[i]->Name.GetString() : NULL );
			}
			else
			{
				Asura_Assert( m_apxSkins[i] == NULL || pxSrcAnim->m_apxSkins[i] == NULL || m_apxSkins[i]->NumberOfBones == pxSrcAnim->m_apxSkins[i]->NumberOfBones, "" );
			}
		}
	}


    FrameNumberStamp           = pxSrcAnim->FrameNumberStamp;
	m_xRGBA                    = pxSrcAnim->m_xRGBA;
    m_iAnimTexID               = pxSrcAnim->m_iAnimTexID;
	m_xTextureAnimController   = pxSrcAnim->m_xTextureAnimController;
	m_xPosition				   = pxSrcAnim->m_xPosition;
	m_xOrient				   = pxSrcAnim->m_xOrient;
    m_usAllBitFlags            = pxSrcAnim->m_usAllBitFlags;

#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
	m_uMaterialOverrideMapID   = pxSrcAnim->m_uMaterialOverrideMapID;
#endif

	m_uNumTimerUpdates		= pxSrcAnim->m_uNumTimerUpdates;
	m_uNumTimerUpdatesAtLastBoneUpdate			= pxSrcAnim->m_uNumTimerUpdatesAtLastBoneUpdate;
	m_uNumTimerUpdatesAtLastBoundingBoxUpdate	= pxSrcAnim->m_uNumTimerUpdatesAtLastBoundingBoxUpdate;

	// Copy the controllers
	DestroyAllControllers();
	m_pxRootControllerGroup	= static_cast<Asura_AnimationController_Group*>(pxSrcAnim->m_pxRootControllerGroup->Duplicate(this));

    // Attachments.
    if( bCopyAttachments )
	{
		CopyAttachments( pxSrcAnim );
	}

#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
	// Copy element visibilities
	if (pxSrcAnim->GetNumSkinElements()>0)
	{	
		if (m_pabSkinElementVisible)
		{
			delete [] m_pabSkinElementVisible;
		}

		m_pabSkinElementVisible = new bool[pxSrcAnim->GetNumSkinElements()];
		for(int i=0;i<pxSrcAnim->GetNumSkinElements();i++)
		{
			m_pabSkinElementVisible[i] = pxSrcAnim->GetSkinElementVisibility(i);
		}
	}
#endif

#ifdef ASURA_USE_DAMAGE_MAPS
	if (m_pafBoneDamage)
	{
		delete [] m_pafBoneDamage;
	}

	m_bIsUsingDamageMap = pxSrcAnim->m_bIsUsingDamageMap;
	m_iDamageMapBoneCount = pxSrcAnim->m_iDamageMapBoneCount;
	m_pafBoneDamage = new float[m_iDamageMapBoneCount];		
	for(int i=0;i<m_iDamageMapBoneCount;i++)
	{
		m_pafBoneDamage[i] = pxSrcAnim->m_pafBoneDamage[i];			
	}		
#endif
	m_xLOSFlags = pxSrcAnim->m_xLOSFlags;

}


void Asura_Animation::CopyAttachments( const Asura_Animation* pxSrcAnim )
{
	if( pxSrcAnim->m_pxFirstAttachment )
	{
		m_pxFirstAttachment = pxSrcAnim->m_pxFirstAttachment->Clone(this);

		//Make all the new attachments delete with the parent.
		Asura_AnimationAttachment *pxAttachment = m_pxFirstAttachment;

		do
		{
			pxAttachment->m_uDeleteAttachedObjectWithParent = true;
			pxAttachment->m_uDeleteWhenUnlinked = true;
			pxAttachment = pxAttachment->m_pxNextAttachment;
		}
		while (pxAttachment);
	}
}


void Asura_Animation::Initialise()
{
    Asura_Assert( m_apxSkins[0], "Asura_Animation::Initialise() called with no skin." );

	FrameNumberStamp = static_cast< u_int >( uINITIAL_FRAMESTAMP );

	m_pxFirstAttachment = NULL;

	// Ensure that the timers don't match...
	m_uNumTimerUpdates	= 0;
	InvalidateBonesUpdateFrameStamp();

	TimersUpdatedBeforeRender = false;

	ParentEntityGuid = ASURA_GUID_UNREGISTERED;
	m_bPosable = false;

	m_xRGBA.m_fA = 1.0f;
    m_iAnimTexID = ASURA_TEXTURE_NONE;

#ifdef ASURA_USE_ANIMATION_IK_SOLVER
	m_pxIKSolver = NULL;
	m_uNumTimerUpdatesAtLastIKUpdate = 0;
	m_bDoIKUpdates = true;
#endif // ASURA_USE_ANIMATION_IK_SOLVER

#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
#if defined(ASURA_USE_STREAMING)
	m_xModuleList.MakeInvalid();
#else
	m_ausIntersectingModules[0] = USHRT_MAX;
#endif //defined(ASURA_USE_STREAMING)
    
    {
        static u_int u = 0;

        m_uNextFrameForFullTest = Asura_Timers::GetFrameNumber() + u;

        u++;
        u &= MIN_FRAMES_BEFORE_INTERSECTING_MODULE_LIST_REFRESH - 1;
    }
#endif // #if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )

	m_pxParent = 0;

	m_xPosition.Zero();
	m_xOrient = Asura_Maths::IdentityQuat;

	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		m_uGroundMaterialResponseHashID = ASURA_HASH_ID_UNSET;
	#endif

	m_sGroundMaterialID = ASURA_MATERIAL_NONE;
	m_xGroundNormal.x = 0.0f;
    m_xGroundNormal.y = -1.0f;
	m_xGroundNormal.z = 0.0f;
	m_uGroundFlagsToTestFor = UINT_MAX;
	m_uGroundFlagsToIgnore  = AsuraPolygonFlag_IgnoreLOS;

    m_usAllBitFlags = 0;

#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
	// If we're using the new shadow-system we need to calculate our bounding-boxes to be tight to
	// the bones in order to get the sharpest possible shadow...
	m_bCalculateBSUsingBones	= true;
#endif	// defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)

	m_pxLightingSolution = 0;

	#ifdef ASURA_TOOLS
	m_uOutlineColour = 0;
	#endif

#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
	m_pabSkinElementVisible = NULL;	
	m_iNumSkinElements = Asura_Project_Skin_Elements::AllocateDefaultVisibilityFlags(this, &m_pabSkinElementVisible);	
#endif

#ifdef ASURA_USE_DAMAGE_MAPS
	m_pafBoneDamage = NULL;
	m_iDamageMapBoneCount = 0;
	m_bIsUsingDamageMap = false;		
#endif

	m_pxFirstSoundInstance = 0;

	if(s_pfnProjectInitialiseCallback)
	{
		 s_pfnProjectInitialiseCallback(this);
	}
}

Asura_Animation::~Asura_Animation()
{
	if(s_pfnProjectShutdownCallback)
	{
		s_pfnProjectShutdownCallback(this);
	}

#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
	m_pxRenderCache->Destroy(); // mark this cache for deletion once render has finished with it
	m_pxRenderCache = NULL;
#endif

	if(m_pxParent)
	{
		//This animation is currently linked to a parent.
		//Cut that link now... RAR 26/03/2004
		m_pxParent->UnlinkAnimation(this);
	}

	delete m_pxLightingSolution;
	DestroyAllAttachments();

#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
	if (m_pabSkinElementVisible)
	{	
		delete [] m_pabSkinElementVisible;
	}
#endif

#ifdef ASURA_USE_DAMAGE_MAPS
	if (m_pafBoneDamage)
	{
		delete [] m_pafBoneDamage;
	}		
#endif

	//Kill any looping sounds
	StopLoopedSounds();

	delete(m_pxRootControllerGroup);

	delete [] m_pxMergedSectionBB;
	m_pxMergedSectionBB = NULL;

}


void Asura_Animation::SetSkin( const char* szSkinName, u_int uSkinIndex /*= 0*/)
{
    g_szPreHashNameTemp = szSkinName;
	SetSkin( Asura_GetHashID(szSkinName), false, uSkinIndex);
    g_szPreHashNameTemp = NULL;
}

	 
void Asura_Animation::SetSkin( Asura_Hash_ID uHashSkin, bool bTemporary /*= false*/, u_int uSkinIndex /*= 0*/ )
{
	if(uSkinIndex != 0 && !m_apxSkins[0])
	{
		Asura_Assert(false, "You must set skin 0 on an animation before setting any others");
		return;
	}

	Asura_Hierarchy_Skin* pxNewSkin = (uHashSkin != ASURA_HASH_ID_UNSET) ? Asura_Animation_System::GetSkinByHash(uHashSkin) : NULL;

    if (!pxNewSkin)
    {
		if (m_apxSkins[uSkinIndex])
		{
			--m_uSkinCount;
		}
        m_apxSkins[uSkinIndex] = NULL;

#if defined(ASURA_ASSERT)
		// Were we trying to set it to nothing?
        if (uHashSkin != ASURA_HASH_ID_UNSET)
        {
            const char* szSkinName = (g_szPreHashNameTemp ? g_szPreHashNameTemp : Asura_HashedFilename_System::GetFilename(uHashSkin, false));
			if (szSkinName == NULL)
			{
// ARC - 12 September 2006 16:21:05 - universal hash look up
				szSkinName = Asura_Universal_HashLookup::GetName(uHashSkin, false);
			}
            Asura_Assertf(m_apxSkins[uSkinIndex], ("Skin '%s' (0x%08x) not found", (szSkinName ? szSkinName : "unknown"), uHashSkin));
        }
#endif
    }
	else
    {
		const Asura_Hierarchy_Skin* pxOldSkin = m_apxSkins[uSkinIndex];
		m_apxSkins[uSkinIndex] = pxNewSkin;

		if (m_apxSkins[uSkinIndex] && !pxOldSkin)
		{
			++m_uSkinCount;
		}

		// re-jig all the attachment marker point indices as the skin's 
		// marker indices won't be the same as the old skin's.
		if( pxOldSkin && m_apxSkins[uSkinIndex] )
		{		
			Asura_AnimationAttachment* pxCurrentAttachment = m_pxFirstAttachment;
			while( pxCurrentAttachment )
			{
				const int	iOldMarkerIndex	= pxCurrentAttachment->GetMarkerPointIndex();
				const u_int	uOldMarkerSkinIndex	= pxCurrentAttachment->GetMarkerPointSkinIndex();

				if(iOldMarkerIndex >= 0)
				{
					// On multiple skins, we use the marker of the lowest indexed skin we can. This means if the marker was previously on the skin we're setting or higher, we could have changed where it should be.
					if (uOldMarkerSkinIndex >= uSkinIndex)
					{
						Asura_Hash_ID uMarkerHash;
						if (uOldMarkerSkinIndex == uSkinIndex)
						{
							uMarkerHash = pxOldSkin->GetMarkerPointHashID( iOldMarkerIndex );
						}
						else
						{
							uMarkerHash = m_apxSkins[uSkinIndex] ? m_apxSkins[uSkinIndex]->GetMarkerPointHashID( iOldMarkerIndex ) : ASURA_HASH_ID_UNSET;
						}
						int iNewMarkerIndex = -1;

						for (u_int u = uSkinIndex; u < ASURA_ANIMATION_SKIN_COUNT; ++u)
						{
							if (m_apxSkins[u] && m_apxSkins[u]->GetMarkerPointIndex( uMarkerHash, iNewMarkerIndex ))
							{
								pxCurrentAttachment->SetMarkerPointIndex( iNewMarkerIndex );
								pxCurrentAttachment->SetMarkerPointSkinIndex( u );
								break;
							}
						}
						Asura_Assertf( iNewMarkerIndex != -1, ("ERROR in SetSkin: New skin ('%s') doesn't have a marker point with hash ID '%x'.\n\n"
							"Attachment ID '%d''s marker point is no longer valid.", m_apxSkins[uSkinIndex]->Name.GetString(), uMarkerHash, pxCurrentAttachment->GetID()) );
					}
				}
				pxCurrentAttachment = pxCurrentAttachment->m_pxNextAttachment;
			}   
		}
		Asura_Assert(uSkinIndex == 0 || m_apxSkins[0]->NumberOfBones == m_apxSkins[uSkinIndex]->NumberOfBones, "All Skins on animation must contain the same number of bones");

		// KOR - 12/01/2007 - trash the framestamps so that if this is done mid-frame, it won't mess up any tween calculations
		InvalidateBonesUpdateFrameStamp();

#if defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
		m_pxRenderCache->Flush(uSkinIndex);
#endif
    }
    
    
	if (m_uSkinCount > 1)
	{
		delete [] m_pxMergedSectionBB;
		m_pxMergedSectionBB = new Asura_Bounding_Box[m_apxSkins[0]->NumberOfBones];

#ifdef ASURA_TOOLS
		for (int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex)
		{
			Asura_Hierarchy_Skin* pxSkin = m_apxSkins[iSkinIndex];

			if( pxSkin && !pxSkin->m_pxSectionBoundingBoxData )
			{
				pxSkin->InitSectionBoundingBoxes();
			}
		}
#endif

		for (int iBoneIndex = 0; iBoneIndex < m_apxSkins[0]->NumberOfBones; ++iBoneIndex)
		{
			m_pxMergedSectionBB[iBoneIndex].Reset();
			for (int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex)
			{
				const Asura_Hierarchy_Skin* pxSkin = m_apxSkins[iSkinIndex];

				if (pxSkin)
				{
					if( pxSkin->SectionData && pxSkin->SectionData[iBoneIndex].m_pxCollisionMesh )
					{
						m_pxMergedSectionBB[iBoneIndex].Set(pxSkin->SectionData[iBoneIndex].m_pxCollisionMesh->GetBoundingBox());
						m_xMergedBBValidFlags[iBoneIndex] = true;
					}
					else if( pxSkin->m_pxSectionBoundingBoxData && pxSkin->m_pxSectionBoundingBoxData[iBoneIndex].m_uValid )
					{
						m_pxMergedSectionBB[iBoneIndex].Set(pxSkin->m_pxSectionBoundingBoxData[iBoneIndex].m_xBB);
						m_xMergedBBValidFlags[iBoneIndex] = true;
					}
				}
			}
		}

	} 
	else 
	{
		delete [] m_pxMergedSectionBB;
		m_pxMergedSectionBB = NULL;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::GetFlagsForLastFrame_Traverse
// Author       : Andy Tate
// Date         : 19-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Animation::GetFlagsForLastFrame_Traverse (Asura_Hash_ID uHash,
													 Asura_AnimationController_Animation* const* ppxAnimation,
													 void* pvCulmulativeFlags)
{
	const Asura_AnimationController_Animation*	pxAnimation	= (ppxAnimation == NULL) ? NULL : *ppxAnimation;

	if((pxAnimation != NULL) && pxAnimation->IsActive())
	{
		int*	piCulmulativeFlags	= reinterpret_cast<int*>(pvCulmulativeFlags);

		// Apply the flags for the last frame of this animation...
		*piCulmulativeFlags	|= pxAnimation->GetFlagsForLastFrame();
	}

	return(true);
}

/// End of function 'Asura_Animation::GetFlagsForLastFrame_Traverse'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::GetFlagsForAllAnimsLastFrame
// Author       : Andy Tate
// Date         : 19-Sep-2006
////////////////////////////////////////////////////////////////////////////////

int Asura_Animation::GetFlagsForAllAnimsLastFrame (void) const
{
	if(m_pxRootControllerGroup->IsActive())
	{
		int		iCulmulativeFlags	= 0;
		void*	pvCulmulativeFlags	= reinterpret_cast<void*>(&iCulmulativeFlags);

		// Traverse the animation tree...
		m_xControllerAnimationContainer.Traverse(GetFlagsForLastFrame_Traverse, pvCulmulativeFlags);

		// Return the flags we've stored...
		return(iCulmulativeFlags);
	}
	else
	{
		return(0);
	}
}

/// End of function 'Asura_Animation::GetFlagsForAllAnimsLastFrame'. ///


void Asura_Animation::InvalidateBonesUpdateFrameStamp(void)
{
	m_uNumTimerUpdatesAtLastBoneUpdate			= 0;
	m_uNumTimerUpdatesAtLastBoundingBoxUpdate	= 0;

	// also invalidate the attachment timestamps
	Asura_AnimationAttachment* pxAttachment = GetFirstAttachment();
	while(pxAttachment)
	{
		pxAttachment->InvalidateFrameStamp();
		pxAttachment = pxAttachment->GetNextAttachment();
	}
}


void Asura_Animation::Update(const Asura_Vector_3& p, const Asura_Matrix_3x3& o,float time /*= ANIMTIME_USEGAMETIME*/,unsigned uRenderFlags /*=0*/)
{
	UpdateStoredPosition(p);
	m_xOrient.FromMatrix(o);

	if(TimersUpdatedBeforeRender)
	{
		UpdateTimers(time);
		Render(p,o,uRenderFlags);
	}
	else
	{
		Render(p,o,uRenderFlags);
		UpdateTimers(time);
	}

	//Play sounds if not done so this frame. RAR 02/03/2005
	if(m_uSoundFrameStamp != Asura_Timers::GetFrameNumber())
	{
		if(!(uRenderFlags & ASURA_HIERARCHY_RENDER_NO_SOUND))
		{
			PlaySoundsForLastFrame();
		}
		m_uSoundFrameStamp = Asura_Timers::GetFrameNumber();
	}
}


void Asura_Animation::Render(const Asura_Vector_3& p, const Asura_Matrix_3x3& o,unsigned uRenderFlags)
{
	//dynamic lighting will need resetting each frame
	ResetDynamicLighting();

	UpdateStoredPosition(p);
	m_xOrient.FromMatrix(o);

	if (m_bNeverHasShadow)
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_SHADOW;
	}

	if( m_bNeverUseFrustumTest )
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_FRUSTUM_TEST;
	}

	Asura_Hierarchy_System::RenderAnimation(this,p,o,uRenderFlags);

	//render any attachment
	Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;

	while(pxAttachment)
	{
		uRenderFlags &= ~ASURA_HIERARCHY_RENDER_BONES_IN_WORLDSPACE;
		pxAttachment->Render(uRenderFlags);
		
		//move onto the next attachment
		pxAttachment = pxAttachment->m_pxNextAttachment;
	}
}


#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
struct Asura_Animation_ShadowData
{
	Asura_Animation *pxAnim;
	uint32			 uRenderFlags;
};

void Asura_Animation::AddShadowCallback(const Asura_ShadowInstance_CreationParams& xParams, const void *pxUserData)
{
	const Asura_Animation_ShadowData *pxData = static_cast<const Asura_Animation_ShadowData*>(pxUserData);
	Asura_Shadow_Instance_Animation* pxNewShadowInstance = new Asura_Shadow_Instance_Animation(xParams, *pxData->pxAnim, pxData->uRenderFlags, false);
	if ( pxNewShadowInstance )
	{
		if ( Asura_Shadow_System::AddShadowInstance( pxNewShadowInstance ) )
		{
			pxNewShadowInstance->CalculateDistance();
			Asura_Matrix_3x3 xOri;
			pxNewShadowInstance->GetLightOrientation( xOri );
			Asura_Vector_3 xShadowBSCentre;
			float fShadowBSRad = pxNewShadowInstance->GetFrustum().GetBoundingSphere( xShadowBSCentre, xOri, pxNewShadowInstance->GetLightPosition() );
			pxData->pxAnim->AddShadowBSToCoarseBS( fShadowBSRad, xShadowBSCentre );
		}
	}
}

#endif


void Asura_Animation::RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri, const u_int uRenderFlags /*= 0*/ )
{
	if((uRenderFlags & ASURA_HIERARCHY_RENDER_NO_SHADOW) == 0)
	{
#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)

		Asura_Animation_ShadowData xData;
		xData.pxAnim	   = this;
		xData.uRenderFlags = uRenderFlags;

		// Need the lighting to be up to date before calling AddShadowInstances
		ResetDynamicLighting();
		UpdateLighting();

		// This function will go through any nearby lights, and call our callback with the details
		// of any shadows which need adding,
		Asura_Shadow_System::AddShadowInstances(xPos, &xData, AddShadowCallback);
		
#endif	// defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::PlaySoundsForLastFrame_Traverse
// Author       : Andy Tate
// Date         : 21-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_Animation::PlaySoundsForLastFrame_Traverse (Asura_Hash_ID uHash,
													   Asura_AnimationController_Animation* const* ppxControllerAnimation)
{
	const Asura_AnimationController_Animation*	pxControllerAnimation	= (ppxControllerAnimation == NULL) ? NULL : *ppxControllerAnimation;

	if((pxControllerAnimation != NULL) && pxControllerAnimation->IsActive())
	{
		// Play the sounds for this animation...
		pxControllerAnimation->PlaySoundsForLastFrame();
	}
}

/// End of function 'Asura_Animation::PlaySoundsForLastFrame_Traverse'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::PlaySound
// Author       : Andy Tate
// Date         : 21-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_Animation::PlaySound (const int iSoundID,
								 const Asura_Hierarchy_Anim& xAnimation,
								 const Asura_Vector_3& xPosition,
								 u_int uSoundMarkerFlags) const
{
	const Asura_Hierarchy_Sound*	pxSoundProp	= GetSoundPropertiesFromAnimSoundID(iSoundID);

	//get the sound properties from the skin , or one of the child skins
	if(pxSoundProp == NULL)
	{
		//see if the animation has anything for this sound then
		pxSoundProp = xAnimation.GetSoundPropertiesFromAnimSoundID(iSoundID);
	}

	if(pxSoundProp != NULL)
	{
		if(pxSoundProp->m_uSoundEvent)
		{
			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::FindEvent(pxSoundProp->m_uSoundEvent);
			if(pxEvent)
			{
				if(pxEvent->IsEndlessPlayback())
				{
					//looping events not allowed since no handle is kept for them
#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA
					Asura_Assertf(0,("Sound event %s used by animation %s is looping.\nThe sound marker must be set up as looping for such sound events to be used.",pxEvent->GetName(),xAnimation.Name.GetString()));
#else
					Asura_Assertf(0,("Sound event used by animation %s is looping.\nThe sound marker must be set up as looping for such sound events to be used.",xAnimation.Name.GetString()));
#endif	
				}
				else
				{
					if(uSoundMarkerFlags & ASURA_ANIMATIONSOUNDFLAG_FOLLOWSKIN)
					{
						//Want to have the sound following the skin , so need to keep track of details for the instance
						//of this sound.
						Asura_Animation_SoundInstance* pxInstance = Asura_Animation_System::AllocateSoundInstance();
						if(pxInstance)
						{

							pxInstance->m_iAnimSoundID = iSoundID;
							pxInstance->m_uEventHash = pxSoundProp->m_uSoundEvent;
							pxInstance->m_uAnimHash = xAnimation.GetHashID();
							pxInstance->m_pxNext = m_pxFirstSoundInstance;
							m_pxFirstSoundInstance = pxInstance;

							Asura_Sound_Event_System::PlayEvent(pxEvent,&pxInstance->m_iHandle,&xPosition, 1.0f, ParentEntityGuid );
						}
					}
					else
					{
						Asura_Sound_Event_System::PlayEvent(pxEvent,0,&xPosition, 1.0f, ParentEntityGuid );
					}
				}
			}
		}

#ifdef ASURA_USE_DIALOGUE_SYSTEM
		if(pxSoundProp->m_uDialogueEvent)
		{
			PlayDialogue(iSoundID, xAnimation, pxSoundProp->m_uDialogueEvent);
		}
#endif


#ifdef ASURA_USE_AUDIO_SYSTEM

		Asura_Hash_ID uSoundHash	= pxSoundProp->m_uSoundHash;
		if( uSoundHash != ASURA_HASH_ID_UNSET )
		{
			Asura_Handle xSoundHandle	= Asura_Audio_System::CreateVoice	( uSoundHash,	false, false );
			Asura_Audio_Voice* pxVoice	= Asura_Audio_System::GetVoice		( xSoundHandle );
			
			Asura_Assert( pxVoice, "Failed to create audio voice" );
			if( pxVoice )
			{
				pxVoice->SetVolume	( pxSoundProp->m_fVolume );
				pxVoice->SetRadii	( pxSoundProp->m_fInnerRadius, pxSoundProp->m_fOuterRadius );
				pxVoice->SetPitch	( pxSoundProp->m_fPitch );
				pxVoice->SetPos		( xPosition );
				pxVoice->Start		();
				pxVoice->Release	();
			}
		}
#else
		Asura_Sound_ID eSound = Asura_Sound_System::GetSoundIDFromHash(pxSoundProp->m_uSoundHash);
		if(eSound != ASURA_SID_NO_SOUND)
		{
			u_int uOldCameraIndex = Asura_Camera::GetCameraIndex();
			Asura_Sound_System::PlayOneShotSound(eSound,&xPosition,pxSoundProp->m_fOuterRadius,pxSoundProp->m_fInnerRadius,pxSoundProp->m_fVolume,pxSoundProp->m_fPitch);
			// (AW - 26.01.2005) - Restore old camera index because sound system might change it.
			Asura_Camera::SetCameraIndex(uOldCameraIndex);
		}
#endif
	}
}

/// End of function 'Asura_Animation::PlaySound'. ///


void Asura_Animation::PlaySoundsForLastFrame() const
{
	//flag all the looped sounds so we can tell which ones are still in use
	//RAR 05/12/2007
	for(Asura_Animation_SoundInstance* pxInstance = m_pxFirstSoundInstance; pxInstance; pxInstance = pxInstance->m_pxNext)
	{
		pxInstance->m_bUpdatedThisFrame = false;
	}


	if(m_apxSkins[0])
	{
		if(m_pxRootControllerGroup->IsActive())
		{
			// Check each animation to see if we need to play sounds...
			m_xControllerAnimationContainer.Traverse(PlaySoundsForLastFrame_Traverse);
		}
	}


	//Update any sounds, start those that have not yet begun and remove those that are no longer needed
	if(m_pxFirstSoundInstance)
	{
		Asura_Animation_SoundInstance** ppxInstance = &m_pxFirstSoundInstance;

		//get the position to play the sounds at
		// Get the centre of the sound from centre of the bounding sphere. This is because, that when we were taking centre of sound
		// from first bone, sometimes the bone was under the ground and sound was muted. So in theory, this could solve this.
		const Asura_Vector_3& xSoundPosition = GetBoundingSphereCentre();

		while(*ppxInstance)
		{
		
			Asura_Animation_SoundInstance* pxInstance = *ppxInstance;
			bool bDelete = false;

			if(pxInstance->m_bDialogue)
			{
				#ifdef ASURA_USE_DIALOGUE_SYSTEM
				if(pxInstance->m_iHandle == ASURA_INVALID_HANDLE)
				{
					//attempt to start the dialogue
					const Asura_Dialogue_Event* pxEvent = Asura_Dialogue_System::FindEvent(pxInstance->m_uEventHash);
					if(pxEvent)
					{
						Asura_Dialogue_System::PlayEvent(pxEvent, xSoundPosition, GetParentEntityGUID(), &pxInstance->m_iHandle);

						if ( pxInstance->m_iHandle == ASURA_INVALID_HANDLE )
						{
							//sounds with low probabilities can deliberately fail, so remove them from the list or this code will retry every frame!
							bDelete = true;
						}
					}
					else
					{
						bDelete = true;
					}
				}
				else if(Asura_Dialogue_System::IsPlaying(pxInstance->m_iHandle))
				{
					//update the position of the dialogue being played
					Asura_Dialogue_System::UpdatePosition(pxInstance->m_iHandle, xSoundPosition);
				}
				else
				{
					//dialogue has finished , so delete the sound instance
					bDelete = true;
				}
				#else
				Asura_Assert(0,"Shouldn't have sound instance marked as dialogue");
				#endif //ASURA_USE_DIALOGUE_SYSTEM
			}
			else if(pxInstance->m_bLoopedSound)
			{
				if(pxInstance->m_bUpdatedThisFrame)
				{
					//If we have a sound handle , make sure the sound hasn't finished
					if(pxInstance->m_iHandle != ASURA_INVALID_HANDLE)
					{
						if(!Asura_Sound_Event_System::IsPlaying(pxInstance->m_iHandle))
						{
							//lose the sound handle so that the sound can be restarted below
							Asura_Sound_Event_System::Stop(pxInstance->m_iHandle);
						}
					}

					if(pxInstance->m_iHandle == ASURA_INVALID_HANDLE)
					{
						//sound not playing , so attempt to restart it
						if(pxInstance->m_uEventHash)
						{
							const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::FindEvent(pxInstance->m_uEventHash);
							if(pxEvent)
							{
								Asura_Sound_Event_System::PlayEvent(pxEvent, &pxInstance->m_iHandle, &xSoundPosition, 1.0f, ParentEntityGuid);
							}
						}					
					}
					else
					{
						//update the position in the existing sound
						Asura_Sound_Event_System::UpdatePosition(pxInstance->m_iHandle, xSoundPosition);
					}
				}
				else
				{
					//deallocate this one and remove from the list
					bDelete = true;
				}
			}
			else
			{
				//Dealing with a non looped sound.
				//This continues until it finishes of its own accord
				if(Asura_Sound_Event_System::IsPlaying(pxInstance->m_iHandle))
				{
					Asura_Sound_Event_System::UpdatePosition(pxInstance->m_iHandle, xSoundPosition);
				}
				else
				{
					Asura_Sound_Event_System::Stop(pxInstance->m_iHandle);
				}

				if(pxInstance->m_iHandle == ASURA_INVALID_HANDLE)
				{
					//suond has finished , so remove this entry
					bDelete = true;
				}
				
			}
			
			if(bDelete)
			{
				*ppxInstance = pxInstance->m_pxNext;

				pxInstance->StopSound();
				Asura_Animation_System::DeallocateSoundInstance(pxInstance);
			}
			else
			{
				ppxInstance = &pxInstance->m_pxNext;
			}
		}
	}


	// Also play any sounds for the attachments
	const Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;

	while(pxAttachment)
	{
		const Asura_Animation* pxAnim = pxAttachment->GetAttachedAnimation();
		if (pxAnim)
		{
			pxAnim->PlaySoundsForLastFrame();
		}

		//move onto the next attachment
		pxAttachment = pxAttachment->m_pxNextAttachment;
	}
}

void Asura_Animation::PlayLoopedSound(int iAnimSoundID, const Asura_Hierarchy_Anim& xAnimation) const
{
	Asura_Hash_ID uAnimHash = xAnimation.GetHashID();

	//See if there is already an instance of this sound in the list
	for(Asura_Animation_SoundInstance* pxInstance = m_pxFirstSoundInstance; pxInstance; pxInstance = pxInstance->m_pxNext)
	{
		if(pxInstance->m_iAnimSoundID == iAnimSoundID)
		{
			if(!pxInstance->m_bUpdatedThisFrame && uAnimHash != pxInstance->m_uAnimHash && pxInstance->m_bLoopedSound)
			{
				//Sound already exists ,but was being triggered by a different animation.
				//Check to see if a different sound event is required now
				const Asura_Hierarchy_Sound* pxSoundProp = GetSoundPropertiesFromAnimSoundID(iAnimSoundID, xAnimation);

				if(pxSoundProp)
				{
					#ifdef ASURA_USE_DIALOGUE_SYSTEM
					if(pxSoundProp->m_uDialogueEvent)
					{
						//Got a dialogue event instead , so head over to the other function
						PlayDialogue(iAnimSoundID, xAnimation,pxSoundProp->m_uDialogueEvent);
						return;
					}
					#endif
					
					if(pxSoundProp->m_uSoundEvent != pxInstance->m_uEventHash)
					{
						//stop the old sound , and set the new sound event to use
						Asura_Sound_Event_System::Stop(pxInstance->m_iHandle);
						pxInstance->m_uEventHash = pxSoundProp->m_uSoundEvent;
					}
				}
				else
				{
					//no sound found ,so just stop the current one.
					Asura_Sound_Event_System::Stop(pxInstance->m_iHandle);
					pxInstance->m_uEventHash = 0;
				}
				pxInstance->m_uAnimHash = uAnimHash;
			}
			
			//mark this sound so that the animation knows that it is still wanted
			pxInstance->m_bUpdatedThisFrame = true;
			return;
		}
	}

	const Asura_Hierarchy_Sound* pxSoundProp = GetSoundPropertiesFromAnimSoundID(iAnimSoundID, xAnimation);
	if(pxSoundProp)
	{
		#ifdef ASURA_USE_DIALOGUE_SYSTEM
		if(pxSoundProp->m_uDialogueEvent)
		{
			//Got a dialogue event instead , so head over to the other function
			PlayDialogue(iAnimSoundID, xAnimation,pxSoundProp->m_uDialogueEvent);
			return;
		}
		#endif

		//allocate a new sound instance and add to the list
		//the sound event will actually be triggered later in the update
		Asura_Animation_SoundInstance* pxInstance = Asura_Animation_System::AllocateSoundInstance();
		if(pxInstance)
		{

			pxInstance->m_iAnimSoundID = iAnimSoundID;
			pxInstance->m_uEventHash = pxSoundProp ? pxSoundProp->m_uSoundEvent : 0;
			pxInstance->m_uAnimHash = uAnimHash;
			pxInstance->m_bUpdatedThisFrame = true;
			pxInstance->m_bLoopedSound = true;
			pxInstance->m_pxNext = m_pxFirstSoundInstance;
			m_pxFirstSoundInstance = pxInstance;
		}
	}
}

#ifdef ASURA_USE_DIALOGUE_SYSTEM
void Asura_Animation::PlayDialogue(int iAnimSoundID, const Asura_Hierarchy_Anim& xAnimation,Asura_Hash_ID uDialogueEvent) const
{
	Asura_Hash_ID uAnimHash = xAnimation.GetHashID();

	//See if there is already an instance of this sound in the list
	for(Asura_Animation_SoundInstance* pxInstance = m_pxFirstSoundInstance; pxInstance; pxInstance = pxInstance->m_pxNext)
	{
		if(pxInstance->m_iAnimSoundID == iAnimSoundID)
		{
			//If old sound was dialogue then it is probably better to leave it going
			//Otherwise if it was a sound event ,stop it , and play the dialogue instead
			if(!pxInstance->m_bDialogue)
			{
				pxInstance->StopSound();
				pxInstance->m_uEventHash = uDialogueEvent;
				pxInstance->m_uAnimHash = uAnimHash;
				pxInstance->m_bDialogue = true;

			}
			pxInstance->m_bUpdatedThisFrame = true;
			return;
		}
	}

	//allocate a new sound instance and add to the list
	//the dialogue event will actually be triggered later in the update
	Asura_Animation_SoundInstance* pxInstance = Asura_Animation_System::AllocateSoundInstance();
	if(pxInstance)
	{

		pxInstance->m_iAnimSoundID = iAnimSoundID;
		pxInstance->m_uEventHash = uDialogueEvent;
		pxInstance->m_uAnimHash = uAnimHash;
		pxInstance->m_bUpdatedThisFrame = true;
		pxInstance->m_bDialogue = true;
		pxInstance->m_pxNext = m_pxFirstSoundInstance;
		m_pxFirstSoundInstance = pxInstance;
	}
	
}
#endif //ASURA_USE_DIALOGUE_SYSTEM

void Asura_Animation::StopLoopedSounds()
{
	while(m_pxFirstSoundInstance)
	{
		Asura_Animation_SoundInstance* pxInstance = m_pxFirstSoundInstance;
		m_pxFirstSoundInstance = m_pxFirstSoundInstance->m_pxNext;
	
		pxInstance->StopSound();
		Asura_Animation_System::DeallocateSoundInstance(pxInstance);

	}
}

void Asura_Animation_SoundInstance::StopSound()
{
	if(m_bDialogue)
	{
		#ifdef ASURA_USE_DIALOGUE_SYSTEM
		Asura_Dialogue_System::Stop(m_iHandle);
		#endif
	}
	else
	{
		Asura_Sound_Event_System::Stop(m_iHandle);
	}
}


bool Asura_Animation::UpdateTimers(float time /*= ANIMTIME_USEGAMETIME*/, bool bIgnoreTimerFrameStamp /*=false*/)
{
	bool bUpdatedThisAnimation = false;
	
	if(time == ANIMTIME_USEGAMETIME)
	{
		time = Asura_Timers::GetGameTimeSlice();
	}
	Asura_Assert(time>=0,"");

	// ART - 26/07/07 - Just because we've been updated, it doesn't mean that our
	// attachments have...
	if (bIgnoreTimerFrameStamp || (Asura_Timers::GetFrameNumber() != FrameNumberStamp))
	{
		FrameNumberStamp = Asura_Timers::GetFrameNumber();
		bUpdatedThisAnimation = true;

		m_xTextureAnimController.UpdateTimer(time);

		// Simply update the timers for the root group...
		if(m_pxRootControllerGroup->IsActive())
		{
			m_pxRootControllerGroup->UpdateTimers(time, bIgnoreTimerFrameStamp);
			IncrementNumTimerUpdates();
		}

		Asura_List_It< Asura_Animation_BoneDataModifier > xListIt( m_xUpdateTimersList );
		while ( !xListIt.Done() )
		{
			Asura_Animation_BoneDataModifier* pxModifier = xListIt.GetCurrent();
			{
				pxModifier->UpdateTimers( time );
			}

			xListIt.Next();
		}
	}

	//update timers for any child animations as well
	Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;

	while(pxAttachment)
	{
		Asura_AnimationAttachment*	pxNextAttachment	= pxAttachment->m_pxNextAttachment;

		pxAttachment->Update(time,bIgnoreTimerFrameStamp);

		if(pxAttachment->NeedsDeleting())
		{
			pxAttachment->UnlinkAttachment(true);
		}

		//move onto the next attachment
		pxAttachment = pxNextAttachment;	
	}

	return bUpdatedThisAnimation;
}

bool Asura_Animation::GetMarkerLocation( int iMarkerIndex, int iSkinIndex, Asura_Vector_3* pxOutputPosition, Asura_Matrix_3x3* pxOutputMatrix, bool bWorldSpace/*=false*/ ) const
{
	Asura_Assert(iSkinIndex >= 0 && iSkinIndex < ASURA_ANIMATION_SKIN_COUNT,"");
	if(!m_apxSkins[iSkinIndex]) return false;
	Asura_Assert(iMarkerIndex >= 0 && iMarkerIndex < m_apxSkins[iSkinIndex]->NumberOfMarkerPoints,"");


	// (MH - 09.08.2002) added code to handle failure
    if (!(iMarkerIndex>=0 && iMarkerIndex < m_apxSkins[iSkinIndex]->NumberOfMarkerPoints))
    {
        return false;
    }

	Asura_Hierarchy_Marker_Point* pxMarker = &m_apxSkins[iSkinIndex]->MarkerPoint[iMarkerIndex];
	
	//Get the tween storage , so we can get the currrent location for the bones
	//(Note : This is fairly dubious as it stands , since I don't think the storage  gets updated at the moment
	//if the character isn't on screen)	

	// If he's never had his bone data calculated then do it now.
	const Asura_Hierarchy_Tween* tweenStorage = GetTweenStore();
	if(!tweenStorage->IsPositionDataValid() && GetSkin() )
	{
		Asura_Animation*		pxNonConstAnimation	= const_cast<Asura_Animation*>(this);
		Asura_Hierarchy_Tween&	xTweenStore			= pxNonConstAnimation->m_pxRootControllerGroup->GetTweenStore();

		// Ensure that the bounding-box is up-to-date...
		if(m_uNumTimerUpdates == 0)
		{
			pxNonConstAnimation->UpdateTimers(0.0f, true);
		}

		// Ensure that the bones have been updated anyway...
		pxNonConstAnimation->UpdateBoneData(0, GetSkin()->NumberOfBones - 1 );

		xTweenStore.SetPositionDataValid(true);
		xTweenStore.SetPositionDataInWorldspace(false);
	}
    
    // (MP - 04.03.2005) Currently not coping with animations from ragdolls
    // (ie. with tween storage in worldspace) here so let's assert about it.
	// AJG - 15/11/05 - Now supporting worldspace bones, although we'll assert if you weren't looking for a worlspace position
    Asura_Assert( !tweenStorage->IsPositionDataInWorldspace() || bWorldSpace, "" );

	Asura_Vector_3 xPosition = pxMarker->Position;
	Asura_Quat xQuat(pxMarker->Matrix);
	
	//work our way up the hierarchy applying the offsets for each section
	int iBoneIndex = pxMarker->BoneIndex;
	Asura_Assertf( iBoneIndex < static_cast<int>(tweenStorage->GetNumberOfBones()), ( "Bad bone index for marker point '%s'", pxMarker->GetName().GetString() ) );

	//If bone index is -1 ,then  marker point is compass relative. RAR 30/03/2006
	if ( ( iBoneIndex != -1) && ( iBoneIndex < static_cast<int>(tweenStorage->GetNumberOfBones()) ) )
	{
		if (tweenStorage->IsPositionDataInWorldspace())
		{
			const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= tweenStorage->GetBoneData(iBoneIndex);
			Asura_Vector_3								xBonePos;
			Asura_Quat									xBoneOri;

			// Only the immediate parent bone is needed here
			xBoneData.GetPosition(xBonePos);
			xBoneData.GetOrientation(xBoneOri);
			xPosition = xBoneOri.RotateVector(xPosition);
			xPosition += xBonePos;	
			xQuat = xBoneOri * xQuat;
		}
		else
		{
			while(iBoneIndex >= 0)
			{
				const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= tweenStorage->GetBoneData(iBoneIndex);
				Asura_Vector_3								xBonePos;
				Asura_Quat									xBoneOri;

				// Only the immediate parent bone is needed here
				xBoneData.GetPosition(xBonePos);
				xBoneData.GetOrientation(xBoneOri);
				xPosition = xBoneOri.RotateVector(xPosition);
				xPosition += xBonePos;	
				xQuat = xBoneOri * xQuat;

				if(!iBoneIndex) break;
				iBoneIndex = m_apxSkins[iSkinIndex]->ParentBone[iBoneIndex];
			}
		}
	}

	// KOR - 6/9/2005 - Now convert to worldspace if requested
	// AJG - 15/11/05 - If we're already in worldspace, there's no need to do anything here
	// RAR  - 30/03/2006 - Compass relative marker points (bone index -1) won't have looked at
	// the tween storage to become worls relative yet.
	if(bWorldSpace && (!tweenStorage->IsPositionDataInWorldspace() || pxMarker->BoneIndex==-1))
	{
		// Rotate position into worldspace
		xPosition = m_xOrient.RotateVector(xPosition);
		// Offset position by animation position
		xPosition += m_xPosition;
		// Rotate orientation into worldspace
		xQuat = m_xOrient * xQuat;
	}

	if(pxOutputPosition)
	{
		*pxOutputPosition = xPosition;
	}

	if(pxOutputMatrix)
	{
		xQuat.ToMatrix(*pxOutputMatrix);
	}

	return true;
}

bool Asura_Animation::GetMarkerLocation(const char* szMarkerName, Asura_Vector_3* pxOutputPosition, Asura_Matrix_3x3* pxOutputMatrix, bool bWorldSpace/*=false*/, int iAttachID/*=Asura_AnimationAttachment::INVALID_ID*/ ) const
{
	return GetMarkerLocationFromHash(Asura_GetHashID(szMarkerName), pxOutputPosition, pxOutputMatrix, bWorldSpace, iAttachID );
}

bool Asura_Animation::GetMarkerLocationFromHash( Asura_Hash_ID uMarkerHash, Asura_Vector_3* pxOutputPosition, Asura_Matrix_3x3* pxOutputMatrix, bool bWorldSpace/*=false*/, int iAttachID/*=Asura_AnimationAttachment::INVALID_ID*/ ) const
{
	if(!m_apxSkins[0]) return false;
	//get the index for the marker
	int iMarkerIndex = -1;
	if(iAttachID == Asura_AnimationAttachment::INVALID_ID)
    {
		for (int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex)
		{
			if (m_apxSkins[iSkinIndex] && m_apxSkins[iSkinIndex]->GetMarkerPointIndex(uMarkerHash, iMarkerIndex))
			{
				return GetMarkerLocation(iMarkerIndex, iSkinIndex, pxOutputPosition, pxOutputMatrix, bWorldSpace);
			}
		}
    }

    // (MH - 21.01.2003) try the attachments
	const Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;
	while( pxAttachment )
	{
		// only consider this attachment if it's the one we've asked for (or we don't care)
		if( (iAttachID == Asura_AnimationAttachment::INVALID_ID) || (pxAttachment->GetID() == iAttachID) )
		{
			const Asura_Animation* pxChildAnimation = pxAttachment->GetAttachedAnimation();
			if( pxChildAnimation )
			{			
				//get the position of the marker point the caller asked for in local space
				Asura_Vector_3 xMarkerPos;
				Asura_Matrix_3x3 xMarkerMat;
				if( pxChildAnimation->GetMarkerLocationFromHash( uMarkerHash, &xMarkerPos, &xMarkerMat, false ) )
				{
					Asura_Vector_3 xParentMarkerPos;
					Asura_Matrix_3x3 xParentMarkerMat;
					if( pxAttachment->GetParentRelativePosition(&xParentMarkerPos, &xParentMarkerMat) )
					{
						if( pxOutputPosition )
						{
							*pxOutputPosition = xMarkerPos * xParentMarkerMat + xParentMarkerPos;
						}
						if( pxOutputMatrix )
						{
							*pxOutputMatrix = xParentMarkerMat * xMarkerMat;	
						}
						
						// KOR - 6/9/2005 - Now convert to worldspace if requested
						if( bWorldSpace )
						{
							if( pxOutputPosition )
							{
								// Rotate position into worldspace
								*pxOutputPosition = m_xOrient.RotateVector(*pxOutputPosition);
								// Offset position by animation position
								*pxOutputPosition += m_xPosition;
							}
							if( pxOutputMatrix )
							{
								// Rotate orientation into worldspace
								Asura_Matrix_3x3 xOri;
								m_xOrient.ToMatrix(xOri);
								*pxOutputMatrix = xOri * *pxOutputMatrix;
							}
						}
						return true;
					}
				}
			}
		}
        
		//move onto the next  attachment
		pxAttachment = pxAttachment->m_pxNextAttachment;
  
	}

    // (MH - 21.01.2003) failed to find marker name in any attachments either
    return false;
}

bool Asura_Animation::GetBoneLocation(int iBoneIndex, Asura_Vector_3 *pxOutputPosition, Asura_Matrix_3x3 *pxOutputMatrix, bool bWorldSpace/* = false*/) const
{
	if (!m_apxSkins[0]) return false;
	Asura_Assert(iBoneIndex >= 0, "Invalid bone requested");
	Asura_Assert(m_apxSkins[0]->NumberOfBones > iBoneIndex, "Invalid bone requested");
	if (iBoneIndex < 0 ||
		iBoneIndex >= m_apxSkins[0]->NumberOfBones)
	{
		return false;
	}
	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
	if(!pxTweenStore->IsPositionDataValid()) 
	{
		return false;
	}

	Asura_Vector_3	xPosition;
	Asura_Quat		xQuat;

	pxTweenStore->GetBoneData(iBoneIndex).GetPosition(xPosition);
	pxTweenStore->GetBoneData(iBoneIndex).GetOrientation(xQuat);

	if (!pxTweenStore->IsPositionDataInWorldspace())
	{
		//work our way up the hierarchy applying the offsets for each section
		while( iBoneIndex > 0 )
		{
			Asura_Vector_3	xBonePos;
			Asura_Quat		xBoneOri;

			iBoneIndex = m_apxSkins[0]->ParentBone[iBoneIndex];

			pxTweenStore->GetBoneData(iBoneIndex).GetPosition(xBonePos);
			pxTweenStore->GetBoneData(iBoneIndex).GetOrientation(xBoneOri);

			xPosition = xBoneOri.RotateVector(xPosition);
			xPosition += xBonePos;	
			xQuat = xBoneOri * xQuat;
		}
	}

	if(pxOutputPosition)
	{
		*pxOutputPosition = xPosition;
	}

	if(pxOutputMatrix)
	{
		xQuat.ToMatrix(*pxOutputMatrix);
	}

	// (AW - 03/11/05) - Now convert to worldspace if requested
	if(bWorldSpace && !pxTweenStore->IsPositionDataInWorldspace())
	{
		if (pxOutputPosition)
		{
			// Rotate position into worldspace
			*pxOutputPosition = m_xOrient.RotateVector(*pxOutputPosition);
			// Offset position by animation position
			*pxOutputPosition += m_xPosition;
		}
		if (pxOutputMatrix)
		{
			// Rotate orientation into worldspace
			Asura_Matrix_3x3 xOri;
			m_xOrient.ToMatrix(xOri);
			*pxOutputMatrix = xOri * *pxOutputMatrix;
		}
	}
	// NY - 16 May 2006 - Or back from worldspace
	else if (!bWorldSpace && pxTweenStore->IsPositionDataInWorldspace())
	{
		if (pxOutputPosition)
		{
			// Offset position by animation position
			*pxOutputPosition -= m_xPosition;
			// Rotate position from worldspace
			Asura_Matrix_3x3 xTrans;
			m_xOrient.ToMatrix(xTrans);
			xTrans.Transpose();
			*pxOutputPosition = *pxOutputPosition * xTrans;
		}
		if (pxOutputMatrix)
		{
			// Rotate orientation from worldspace
			Asura_Matrix_3x3 xOri;
			m_xOrient.ToMatrix(xOri);
			xOri.Transpose();
			*pxOutputMatrix = xOri * *pxOutputMatrix;
		}
	}

	return true;
}


const Asura_Bounding_Box* Asura_Animation::GetBoneBB( int iBoneIdx ) const
{
    // (MP - 07.06.2006) Not currently providing an option of returning the
    // BB in worldspace since we'd end up making a worldspace AABB of the
    // rotated bonespace AABB. This may be pretty inaccurate. It would be
    // straightforward to add this option if it would actually be useful.

    if( !m_apxSkins[0] ) { return NULL; }

	if (m_uSkinCount == 1)
	{
		// This logic mimics the LOS test.
		const Asura_Bounding_Box* pxBB = NULL;
		if( m_apxSkins[0]->SectionData && m_apxSkins[0]->SectionData[iBoneIdx].m_pxCollisionMesh )
		{
			pxBB = &m_apxSkins[0]->SectionData[iBoneIdx].m_pxCollisionMesh->GetBoundingBox();
		}
		else if( m_apxSkins[0]->m_pxSectionBoundingBoxData && m_apxSkins[0]->m_pxSectionBoundingBoxData[iBoneIdx].m_uValid )
		{
			pxBB = &m_apxSkins[0]->m_pxSectionBoundingBoxData[iBoneIdx].m_xBB;
		}

		return pxBB;
	} else {
		return m_pxMergedSectionBB[iBoneIdx].MinX < m_pxMergedSectionBB[iBoneIdx].MaxX ? &m_pxMergedSectionBB[iBoneIdx] : NULL;
	}
}

const Asura_Bounding_Box* Asura_Animation::GetManualBoneBB( int iBoneIdx ) const
{
    if( !m_apxSkins[0] ) { return NULL; }

	//Gets the bounding box from SectionBoundingBoxData.
	//This allows a manually set bounding box to be retrieved even if there is a collision mesh.
	//RAR 01/02/2007
	if (m_uSkinCount == 1)
	{
		const Asura_Bounding_Box* pxBB = NULL;
		if( m_apxSkins[0]->m_pxSectionBoundingBoxData[iBoneIdx].m_uValid )
		{
			pxBB = &m_apxSkins[0]->m_pxSectionBoundingBoxData[iBoneIdx].m_xBB;
		}

		return pxBB;
	} else {
		static Asura_Bounding_Box xBB;
		xBB.Reset();
		for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
		{
			if( m_apxSkins[i] && m_apxSkins[i]->m_pxSectionBoundingBoxData[iBoneIdx].m_uValid )
			{
				xBB.Set(m_apxSkins[i]->m_pxSectionBoundingBoxData[iBoneIdx].m_xBB);
			}
		}
		return xBB.MinX < xBB.MaxX ? &xBB : NULL;
	}
}


bool Asura_Animation::HasMarker( const char* szMarkerName ) const
{
	return HasMarker( Asura_GetHashID(szMarkerName) );
}

bool Asura_Animation::HasMarker( Asura_Hash_ID uMarkerNameHash ) const
{
	int iIgnored;

	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		if( m_apxSkins[i] && m_apxSkins[i]->GetMarkerPointIndex(uMarkerNameHash, iIgnored) ) 
		{ 
			return true; 
		}
	}
    
    // Try attachments.
	for( const Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment; pxAttachment; pxAttachment = pxAttachment->m_pxNextAttachment )
	{
		const Asura_Animation* pxAttachmentAnim = pxAttachment->GetAttachedAnimation();
        if( pxAttachmentAnim && pxAttachmentAnim->HasMarker(uMarkerNameHash) )
        {
            return true;
        }
	}

    return false;
}

bool Asura_Animation::LinkAnimation(Asura_Animation* pxChildAnimation,const char* pszMarkerName,bool bDeleteWithParent,int iID)
{
	Asura_Assert(pszMarkerName,"");
	if(!pszMarkerName) return false;

	return LinkAnimation(pxChildAnimation,Asura_GetHashID(pszMarkerName),bDeleteWithParent,iID);
}

bool Asura_Animation::LinkAnimation(Asura_Animation* pxChildAnimation,Asura_Hash_ID uMarkerHash,bool bDeleteWithParent,int iID)
{
	Asura_Assert( uMarkerHash != ASURA_HASH_ID_UNSET, "" );
	Asura_Assert(pxChildAnimation,"");

    if ( pxChildAnimation )
    {
	    //create the attachment
	    Asura_AnimationAttachment_Animation* pxAttachment = new Asura_AnimationAttachment_Animation(pxChildAnimation,bDeleteWithParent);
	    pxAttachment->SetID(iID);

	    //and add it
	    return pxAttachment->LinkAttachment(this,uMarkerHash,true);
    }
    else
    {
        return false;
    }
}


const Asura_Hierarchy_Sound* Asura_Animation::GetSoundPropertiesFromAnimSoundID(int iAnimSoundID) const
{
	const Asura_Hierarchy_Sound* pxSoundProp = 0;
	if(m_apxSkins[0])
	{
		pxSoundProp = m_apxSkins[0]->GetSoundPropertiesFromAnimSoundID(iAnimSoundID);
	}
	if(!pxSoundProp)
	{
		//check to see if the skins for any attachments define the sound
		const Asura_AnimationAttachment* pxAttachment = GetFirstAttachment();
		while(pxAttachment && !pxSoundProp)
		{
			const Asura_Animation* pxAnimation = pxAttachment->GetAttachedAnimation();
			if(pxAnimation)
			{
				pxSoundProp = pxAnimation->GetSoundPropertiesFromAnimSoundID(iAnimSoundID);	
			}
			pxAttachment = pxAttachment->m_pxNextAttachment;
		}
	}
	return pxSoundProp;
}

const Asura_Hierarchy_Sound* Asura_Animation::GetSoundPropertiesFromAnimSoundID(int iAnimSoundID, const Asura_Hierarchy_Anim& xAnimation) const
{
	//check the skin(s)
	const Asura_Hierarchy_Sound* pxSoundProp = GetSoundPropertiesFromAnimSoundID(iAnimSoundID);
	if(pxSoundProp) return pxSoundProp;

	//check the animation
	return xAnimation.GetSoundPropertiesFromAnimSoundID(iAnimSoundID);
	
}

void Asura_Animation::AddToUpdateTimersList( Asura_Animation_BoneDataModifier* pxModifier )
{
	m_xUpdateTimersList.InsertInList( pxModifier, Asura_Animation_BoneDataModifier::CompareUpdateTimersPriority );
}

void Asura_Animation::AddToUpdateBoneDataList( Asura_Animation_BoneDataModifier* pxModifier )
{
	m_xUpdateBoneDataList.InsertInList( pxModifier, Asura_Animation_BoneDataModifier::CompareUpdateBoneDataPriority );
}

void Asura_Animation::RemoveFromUpdateTimersList( Asura_Animation_BoneDataModifier* pxModifier )
{
	m_xUpdateTimersList.RemoveFromList( pxModifier );
}

void Asura_Animation::RemoveFromUpdateBoneDataList( Asura_Animation_BoneDataModifier* pxModifier )
{
	m_xUpdateBoneDataList.RemoveFromList( pxModifier );
}

////////////////////////////////////////////////////////////////////////////////
//
// LinkAnimation_BoneRelative - uses a bone relative position and orientation
// pair instead of a marker point.
//
////////////////////////////////////////////////////////////////////////////////
bool Asura_Animation::LinkAnimation_BoneRelativeObjectSpace(Asura_Animation* pxChildAnimation, const Asura_Vector_3& xObjectSpacePos, const Asura_Quat& xObjectSpaceOri, int iBoneID, bool bDeleteWithParent, int iID)
{
	Asura_Vector_3 xBoneSpacePosition;
	Asura_Quat xBoneSpaceOri;

	if(!GetBoneRelativePositionAndOrientation(xObjectSpacePos, xObjectSpaceOri, iBoneID, xBoneSpacePosition, xBoneSpaceOri))
	{
		return false;
	}

	return LinkAnimation_BoneRelativeBoneSpace(pxChildAnimation, xBoneSpacePosition, xBoneSpaceOri, iBoneID, bDeleteWithParent, iID);
}


bool Asura_Animation::LinkAnimation_BoneRelativeBoneSpace(Asura_Animation* pxChildAnimation, const Asura_Vector_3& xBoneRelativePos, const Asura_Quat& xBoneRelativeOri, int iBoneID, bool bDeleteWithParent, int iID)
{
	if(pxChildAnimation)
	{
		//Create the attachment.
		Asura_AnimationAttachment_BoneRelativeAnimation *pxAttachment = new Asura_AnimationAttachment_BoneRelativeAnimation(pxChildAnimation, bDeleteWithParent);
		pxAttachment->SetID(iID);

		return pxAttachment->LinkAttachment_BoneRelative(this, xBoneRelativePos, xBoneRelativeOri, iBoneID, true);
	}
	else
	{
		Asura_Assert(false, "Null Child Animation");
		return false;
	}
}

/// Richard May - 22/04/2004 16:27:57 - Now returns true if unlinked successfully
bool Asura_Animation::UnlinkAnimation(Asura_Animation* pxChildAnimation)
{
	if(!pxChildAnimation)
	{
		Asura_Assert(pxChildAnimation,"");
		return false;
	}
	
	//find the attachment for this animation
	Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;
	while(pxAttachment)
	{
		Asura_Animation* pxAttachedAnim = pxAttachment->GetAttachedAnimation();
		if( pxAttachedAnim )
		{
			if( pxAttachedAnim == pxChildAnimation )
			{
				// the child anim belongs to this attachment, so remove it
				pxAttachment->UnlinkAttachment();
				return true;
			}
			// try recursing through any attachments on this attachment's anim
			else if( pxAttachedAnim->UnlinkAnimation( pxChildAnimation ) ) 
			{
				return true;
			}
		}

		pxAttachment = pxAttachment->m_pxNextAttachment;
	}
	
	// failed to find attached anim, return false
	return false;
}




// Temporary storage for bone recursion below.
// Give that this data will have been calculated by RecurseBones() in the renderer
// it might well be quicker to store BonePosition and BoneOrientation along with the
// animation instance.
static Asura_Vector_3 g_xBonePositionWorkspace[ASURA_MAX_NO_OF_BONES_PER_MODEL];
static Asura_Matrix_3x3 g_xBoneOrientationWorkspace[ASURA_MAX_NO_OF_BONES_PER_MODEL];

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::UpdateBoneWorkspace
// Author       : Andy Tate
// Date         : 19-Dec-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_Animation::UpdateBoneWorkspace (const Asura_Hierarchy_Tween& xTweenStore,
										   const Asura_Vector_3& xPosition,
										   const Asura_Matrix_3x3& xOrientation) const
{
	for(int i = 0; i < m_apxSkins[0]->NumberOfBones; i++)
	{
		const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= xTweenStore.GetBoneData(i);

		if(xTweenStore.IsPositionDataInWorldspace())
		{
			xBoneData.GetPosition(g_xBonePositionWorkspace[i]);
			xBoneData.GetOrientation(g_xBoneOrientationWorkspace[i]);
		}
		else
		{
			Asura_Matrix_3x3	xBoneOri;
			Asura_Vector_3		xBonePos;

			xBoneData.GetPosition(xBonePos);
			xBoneData.GetOrientation(xBoneOri);

			if (i == 0)
			{
				// Root
				g_xBonePositionWorkspace[0] = xBonePos * xOrientation + xPosition;
				g_xBoneOrientationWorkspace[0] = xOrientation * xBoneOri;
			}
			else
			{
				// Get parent
				int iParentIndex = m_apxSkins[0]->ParentBone[i];
				Asura_Assert(iParentIndex >= 0 && iParentIndex < i, "");

				g_xBonePositionWorkspace[i] = xBonePos * g_xBoneOrientationWorkspace[iParentIndex] + g_xBonePositionWorkspace[iParentIndex];
				g_xBoneOrientationWorkspace[i] = g_xBoneOrientationWorkspace[iParentIndex] * xBoneOri;
			}
		}
	}
}

/// End of function 'Asura_Animation::UpdateBoneWorkspace'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::GetClosestBone
// Author       : Ste Butcher
// Date         : 1-May-2007
////////////////////////////////////////////////////////////////////////////////

int Asura_Animation::GetClosestBone(const Asura_Vector_3& vPosition)
{
	int iClosestIndex = -1;
	float fClosestDistance = 0.0f;

	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
	if(pxTweenStore)
	{
		Asura_Matrix_3x3 xOrientation;
		Asura_Quat xOrientationQuat = GetOrientQuat();
		xOrientationQuat.ToMatrix(xOrientation);

		Asura_Vector_3 xPosition = GetPosition();

		Asura_Vector_3 xaBonePositionWorkspace[ASURA_MAX_NO_OF_BONES_PER_MODEL];
		Asura_Matrix_3x3 xaBoneOrientationWorkspace[ASURA_MAX_NO_OF_BONES_PER_MODEL];

		int iNumBones = m_apxSkins[0]->NumberOfBones;
		for(int i = 0; i < iNumBones; ++i)
		{
			const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= pxTweenStore->GetBoneData(i);

			// Get bone position in world space
			if(pxTweenStore->IsPositionDataInWorldspace())
			{
				xBoneData.GetPosition(xaBonePositionWorkspace[i]);
				xBoneData.GetOrientation(xaBoneOrientationWorkspace[i]);
			}
			else
			{
				Asura_Matrix_3x3	xBoneOri;
				Asura_Vector_3		xBonePos;

				xBoneData.GetOrientation(xBoneOri);
				xBoneData.GetPosition(xBonePos);

				if (i == 0)
				{
					// Root
					xaBonePositionWorkspace[0] = xBonePos * xOrientation + xPosition;
					xaBoneOrientationWorkspace[0] = xOrientation * xBoneOri;
				}
				else
				{
					// Get parent
					int iParentIndex = m_apxSkins[0]->ParentBone[i];
					Asura_Assert(iParentIndex >= 0 && iParentIndex < i, "");

					xaBonePositionWorkspace[i] = xBonePos * xaBoneOrientationWorkspace[iParentIndex] + xaBonePositionWorkspace[iParentIndex];
					xaBoneOrientationWorkspace[i] = xaBoneOrientationWorkspace[iParentIndex] * xBoneOri;
				}
			}

			// Distance check
			Asura_Vector_3 xVec = xaBonePositionWorkspace[i] - vPosition;
			float fSquaredMag = xVec.MagnitudeSquared();
			if(fSquaredMag < fClosestDistance || iClosestIndex == -1)
			{
				fClosestDistance = fSquaredMag;
				iClosestIndex = i;
			}
		}
	}

	return iClosestIndex;
}

/// End of function 'Asura_Animation::GetClosestBone'. ///



//
// LOS test on section BB or polys (depending on m_bBoundingBoxOnly flag etc).
//
// I'm going to assume that the tester has done an appropriate early rejection
// test on bounding sphere and/or bounding box.
//
// pxOutput->m_fDistanceToIntersection is set on input to be the closest hit
// so far in this LOS test. (Or max dist to test to).
//
// If a closer hit occurs then true is returned and the following are updated in pxOutput:
// m_fDistanceToIntersection, m_iHitBoneID, m_iPolyFlags, m_xNormal
//
bool Asura_Animation::LOSTest(const Asura_Vector_3& xPosition, 
							  const Asura_Matrix_3x3& xOrient,
							  const Asura_LOS_InputData* pxInput,
							  Asura_LOS_OutputData* pxOutput)
{
    if (!m_apxSkins[0])
    {
        return false;
    }

/// Richard May - 02/08/2004 15:09:04 - now only asserts in non-tools builds
/// to prevent assert when LOS testing before a render call. In non-tools
/// builds, the section box data should be calculated beforehand!
#ifdef ASURA_TOOLS
	for (int i = 0; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		if( m_apxSkins[i] && !m_apxSkins[i]->m_pxSectionBoundingBoxData )
		{
			m_apxSkins[i]->InitSectionBoundingBoxes();
		}
	}

#else
    Asura_Assertf(m_apxSkins[0]->m_pxSectionBoundingBoxData, ("no bounding box data for skin '%s'", m_apxSkins[0]->Name.GetString()));
    if (!m_apxSkins[0]->m_pxSectionBoundingBoxData) return false;
#endif // ASURA_TOOLS


	// Get the tween storage, so we can get the currrent location for the bones. If the character is not
    // on screen this will probably be set to the previous pose when the character was rendered. 
	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
	float fDistance;

	if (!pxTweenStore->IsPositionDataValid())
    {
		// Try to set up a valid tween store.
		Update(xPosition, xOrient, ANIMTIME_USEGAMETIME, ASURA_HIERARCHY_RENDER_DONTRENDER | ASURA_HIERARCHY_RENDER_UPDATEBONESALWAYS);
		// if we still don't get a valid store, something strange is going on.
		Asura_Assert(pxTweenStore->IsPositionDataValid(), "Failed to initialise tween store during LOS test.");
		if (!pxTweenStore->IsPositionDataValid())
		{
			return false;
		}
	}

	if ( m_fCoarseBSRadius < 0.0f )
	{
		UpdateCoarseBoundingSphere( xPosition, xOrient );
	}

	bool bLOSHit = false;
	
	// SNH: copied from BHD branched version...
//RCD: since we know that this object has been rendered, it will have a reasonably valid bounding sphere set by the rendering code
// so we use that to save us doing even more tests on the bones.
	// SNH again: can't just return here, since attachments don't seem to be taken into account when calculating fBSRadius.
	//	Just skip the bones check but go through the attachments.
	//if (dbg_HackyLOSCheat) // just for testing purposes
	{
		Asura_Assert(m_fCoarseBSRadius >= 0.f, "Bounding sphere uninitialised?");
		if (Asura_Maths::RayIntersectsSphere(pxInput->m_xOrigin,
										 pxInput->m_xDirection,
										 m_xBSCentre,
										 m_fCoarseBSRadius,
										 &fDistance))
		{
			if(pxInput->m_bBoundingSphereOnly)
			{
				if (fDistance < pxOutput->m_fDistanceToIntersection)
				{
					// Sphere intersection is a closer hit.
					pxOutput->m_iHitBoneID = -1;
					pxOutput->m_iHitAttachmentID = Asura_AnimationAttachment::INVALID_ID;
					const Asura_Vector_3 xSphereHitPosition = pxInput->m_xOrigin + (pxInput->m_xDirection * fDistance);
					pxOutput->m_xNormal = xSphereHitPosition - m_xBSCentre;
					pxOutput->m_xNormal.Normalise();
					pxOutput->m_fDistanceToIntersection = fDistance;
					pxOutput->m_usPolyFlags = 0;

					#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
						pxOutput->m_uMaterialResponseHashID = ASURA_HASH_ID_UNSET;
					#endif
					pxOutput->m_sMaterialID = ASURA_MATERIAL_NONE;

					bLOSHit = true;
				}
			}
			else
			{
				UpdateBoneWorkspace(*pxTweenStore, xPosition, xOrient);

				for (int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex)
				{
					if (m_apxSkins[iSkinIndex])
					{
						for(int i = 0; i < m_apxSkins[0]->NumberOfBones; i++)
						{
							if( m_xLOSFlags[iSkinIndex])
							{
								continue;
							}

							if (m_apxSkins[iSkinIndex]->SectionData && m_apxSkins[iSkinIndex]->SectionData[i].m_pxCollisionMesh)
							{
								if (m_apxSkins[iSkinIndex]->SectionData[i].m_pxCollisionMesh->LOSTest(g_xBonePositionWorkspace[i], g_xBoneOrientationWorkspace[i], pxInput, pxOutput))
								{
									// Hit the collision mesh, closer than previous hits.
									// m_fDistanceToIntersection, normal and poly flags have already been updated.
									pxOutput->m_iHitBoneID = i;
									pxOutput->m_iHitAttachmentID = Asura_AnimationAttachment::INVALID_ID;
									bLOSHit = true;
								}
							}
							else
							{
								if( m_apxSkins[iSkinIndex]->m_pxSectionBoundingBoxData[i].m_uValid )
								{
									// Convert LOS test from world space to bone space
									Asura_Vector_3 xAlpha = pxInput->m_xOrigin;
									Asura_Vector_3 xBeta  = pxInput->m_xDirection;
									xAlpha -= g_xBonePositionWorkspace[i];

									Asura_Matrix_3x3 xMat = g_xBoneOrientationWorkspace[i];
									xMat.Transpose();

									xAlpha = xAlpha * xMat;
									xBeta = xBeta * xMat;

									Asura_Vector_3 xTempNormal;
									float fTempDist;
									bool bStartedInside;
									bool bIntersects;

									if(pxInput->m_bUseBoneScale)
									{
										// Get the bounding box for this bone
										Asura_Bounding_Box	xBoundingBox	= m_apxSkins[iSkinIndex]->m_pxSectionBoundingBoxData[i].m_xBB;

										xBoundingBox.Scale(pxInput->m_xBoneScale);

										bIntersects	= Asura_Maths::RayIntersectsBoundingBox(xAlpha, xBeta, xBoundingBox, &fTempDist, 0, &bStartedInside, &xTempNormal);
									}
									else
									{
										// Get the bounding box for this bone
										const Asura_Bounding_Box& xBoundingBox = m_apxSkins[iSkinIndex]->m_pxSectionBoundingBoxData[i].m_xBB;

										bIntersects	= Asura_Maths::RayIntersectsBoundingBox(xAlpha, xBeta, xBoundingBox, &fTempDist, 0, &bStartedInside, &xTempNormal);
									}

									if (bIntersects)
									{
										if (fTempDist < pxOutput->m_fDistanceToIntersection)
										{
											// Hit the BB closer than last hit. Fix up the return values.
											pxOutput->m_iHitBoneID = i;
											pxOutput->m_iHitAttachmentID = Asura_AnimationAttachment::INVALID_ID;
											pxOutput->m_fDistanceToIntersection = fTempDist;
											pxOutput->m_xNormal = xTempNormal * g_xBoneOrientationWorkspace[i];

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
											pxOutput->m_uMaterialResponseHashID = ASURA_HASH_ID_UNSET;
#endif
											pxOutput->m_sMaterialID = ASURA_MATERIAL_NONE;

											pxOutput->m_usPolyFlags = 0; // TODO set material flags?
											pxOutput->m_bStartedInside = bStartedInside;
											bLOSHit = true;
										}
									}
								}
							}
						}
					}
				}
			}
		}

		// WGP 27/02/2009 - Moved the attachments check inside the BS test as these are now included in the calculation of the coarse bounding sphere

		//RAR 28/02/03 - Check LOS against attachments.
		Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;
		while(pxAttachment)
		{
			if(pxAttachment->CanBeTestedForLOS())
			{
				//get the position of the attachments's marker point
				Asura_Vector_3 xMarkerPosition;
				Asura_Matrix_3x3 xMarkerOrientation;
				if(pxAttachment->GetPosition(&xMarkerPosition,&xMarkerOrientation))
				{
					//convert into world position/orientation
					Asura_Vector_3 xWorldPosition = xMarkerPosition*xOrient + xPosition;
					Asura_Matrix_3x3 xWorldOrientation = xOrient*xMarkerOrientation;

					//and test...
					bool bHitAttachment = pxAttachment->LOSTest(xWorldPosition,xWorldOrientation,pxInput,pxOutput);
					if(bHitAttachment)
					{
						//give the bone ID of the marker point that this is attached to , rather than the ID
						//of the bone in the attached animation.
						pxOutput->m_iHitBoneID = pxAttachment->GetBoneID();
						
						//Only set the attachment hit if it is currently invalid
						//If we hit the immediate child attachment , then the ID should have been set to invalid during the test
						//If an attachment of an attachment was hit , then the ID of that attachment will have been set ,
						//and so should be left alone.
						//RAR 28/02/03
						if( pxOutput->m_iHitAttachmentID == Asura_AnimationAttachment::INVALID_ID )
						{
							pxOutput->m_iHitAttachmentID = pxAttachment->GetID();
						}
						bLOSHit = true;
					}
				}

			}
			//move onto the next attachment
			pxAttachment = pxAttachment->m_pxNextAttachment;
		}
	}


    return bLOSHit;
}


bool Asura_Animation::GetOffsetFromCompassPoint(Asura_Vector_3& xPos) const
{
	// Get the tween storage, so we can get the currrent location for the bones. If the character is not
    // on screen this will probably be set to the previous pose when the character was rendered.
	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
	if (!pxTweenStore->IsPositionDataValid()) 
    {
        return false;
    }

    // (MP - 04.03.2005) Currently not coping with animations from ragdolls
    // (ie. with tween storage in worldspace) here so let's assert about it.
    Asura_Assert( !pxTweenStore->IsPositionDataInWorldspace(), "" );
    
	pxTweenStore->GetBoneData(0).GetPosition(xPos);

	return true;
}

#if defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )

bool Asura_Animation::IsVisible( bool bNoFrustumTest /*= false*/ )
{
	// KJFL - 29 Febrary 2008 - we might be able to calculate a tighter AABB here,
	// but there doesn't seem to be one cached in the anim itself.
	Asura_Bounding_Box xAABB;

	xAABB.SetToPoint( m_xBSCentre );

	Asura_Assert( m_fCoarseBSIncEffectsRadius >= 0.0f, "Coarse bounding sphere not initialised" );
	if (m_fCoarseBSIncEffectsRadius > 0.0f )
	{
		xAABB.Grow( m_fCoarseBSIncEffectsRadius );
	}

	return bNoFrustumTest ? m_xVisibilityQuery.IsAABBVisibleFromCurrentCameraIgnoringDirection( xAABB ) : m_xVisibilityQuery.IsAABBVisibleFromCurrentCamera( xAABB );
}

#else // !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )

#include "Rendering/Asura_Module.h"
#include "Mesh/Asura_Env_System.h"
#include "Mesh/Asura_Env.h"

#if !defined(ASURA_USE_STREAMING)
// Used for non-streaming builds only.
static const Asura_Animation* g_pxCallbackAnim;
static bool                   g_bCallbackFoundVisModule;
static u_int                  g_uCallbackNumModulesFound;
static bool                   g_bCallbackInNonExteriorModule;

void Asura_Animation::IntersectingModuleCallback( const u_int uIdx )
{
	if(Asura_Environment_System::GetEnvironment()->IsSceneryModule(uIdx))
	{
		return;//DWE - ignore scenery modules 
	}
	bool bNonExterior = !Asura_Module_System::GetCurrentHandler().IsExteriorModule(uIdx);

	if(bNonExterior && !g_bCallbackInNonExteriorModule)
	{
		//This is the first non-exterior module that the BB has intersected with.
		//Clearout any previous results. RAR 10/04/2007
		g_bCallbackFoundVisModule = false;
		g_uCallbackNumModulesFound = 0;
		g_bCallbackInNonExteriorModule = true;
	}

	if( bNonExterior == g_bCallbackInNonExteriorModule)
	{
		if( !g_bCallbackFoundVisModule )
    	{
    	    g_bCallbackFoundVisModule = Asura_Module_System::IsModuleVisible( uIdx );
    	}

    	if( UINT_MAX != g_uCallbackNumModulesFound )
    	{
    	    if( g_uCallbackNumModulesFound < MAX_INTERSECTING_MODULES_TO_STORE )
    	    {
    	        g_pxCallbackAnim->m_ausIntersectingModules[g_uCallbackNumModulesFound++] = static_cast<u_short>(uIdx);
    	    }
    	    else
    	    {
    	        g_uCallbackNumModulesFound = UINT_MAX;
    	    }
    	}
	}
}
#endif //!defined(ASURA_USE_STREAMING)

//////////////////////////
// UpdateContainingZone //
//////////////////////////
#if defined(ASURA_USE_STREAMING)
// MGH 6/6/2006: The containing zone is normally updated in IsInModuleVisibleToCurCam, but if the
// module visibility test is disabled then we can call this function instead.
// The containing zone is needed to get the correct GI lighting handler.
void Asura_Animation::UpdateContainingZone(const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient) const
{
	// Reduce the frequency of the updates.
	u_int uTestFrameNumber = m_uNextFrameForFullTest;
	if (m_xModuleList.IsValid())
	{
		uTestFrameNumber += NUM_EXTRA_FRAMES_BEFORE_UPDATE_CONTAINING_ZONE;
	}

	if (Asura_Timers::GetFrameNumber() >= uTestFrameNumber)
	{
		// Update the module list.
		IsInModuleVisibleToCurCam(xPos, xOrient);
	}
}
#endif //defined(ASURA_USE_STREAMING)

///////////////////////////////
// IsInModuleVisibleToCurCam //
///////////////////////////////
#if defined(ASURA_USE_STREAMING)
// Streaming build, needs to take account of animations that span zone boundaries.
bool Asura_Animation::IsInModuleVisibleToCurCam( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient ) const
{
    if ( !GetSkin() ) return false;

    const float POS_DELTA_THRESHOLD  = 0.10f;
	const float POS_DELTA_THRESHOLD2 = POS_DELTA_THRESHOLD * POS_DELTA_THRESHOLD;

    bool bTestModules = false;
	if (!m_xModuleList.IsValid())
    {
		// Invalid, so we need to do a test.
        bTestModules = true;
    }
	else if (!xPos.ApproximatelyEquals(m_xCachedPos, POS_DELTA_THRESHOLD2)) 
	{
		// Animation has moved, recalculate the modules periodically.
		bTestModules = ( Asura_Timers::GetFrameNumber() >= m_uNextFrameForFullTest );
	}
    
	if (!bTestModules && m_xModuleList.IsEmpty())
	{
		// Frame when we last calculated the module list.
		const u_int uLastCalculated = (m_uNextFrameForFullTest-MIN_FRAMES_BEFORE_INTERSECTING_MODULE_LIST_REFRESH);
		if (Asura_Streaming_System::GetFrameNumberLastZoneLoad() > uLastCalculated)
		{
			// A zone has been loaded since the last module test returned no modules.
			// Perhaps we are intersecting this newly loaded zone, try doing a test.
			bTestModules = true;
		}
	}

	if ( bTestModules )
	{
#ifdef ASURA_TOOLS
		if (!Asura_Module_System::IsInitialised()) { return true; }
#endif //ASURA_TOOLS

		Asura_Bounding_Box xUnrotatedBB;
		xUnrotatedBB	= m_pxRootControllerGroup->GetTweenStore().GetBoundingBox();

		Asura_Bounding_Box xBB;
		xBB.FromOBB( xUnrotatedBB, xOrient );
        xBB.Displace( xPos );

		m_xModuleList.CalculateModuleList(xBB);
		if (m_xModuleList.IsValid())
		{
			// Work out next frame to do a full test, if the position changes.
			m_uNextFrameForFullTest = Asura_Timers::GetFrameNumber() + MIN_FRAMES_BEFORE_INTERSECTING_MODULE_LIST_REFRESH;
            m_xCachedPos            = xPos;
		}
		else
		{
			// MGH 31/5/2006: Module list overflowed. Just assume that it is visible.
			// TODO perhaps we should do some more tests here.
			Asura_Assert(0, "Module list overflowed in Asura_Animation");
			return true;
		}
    }
  
	return m_xModuleList.IsVisible();
}
#else
// Non-streaming build. Only a single zone, so a simple array of modules suffices.
bool Asura_Animation::IsInModuleVisibleToCurCam( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient ) const
{
    if( !GetSkin() ) { return false; }

    const float POS_DELTA_THRESHOLD = 0.10f;

    bool bTestModules;
	if( USHRT_MAX == m_ausIntersectingModules[0] )
    {
        bTestModules = true;
    }
    else
    {
        if( xPos.ApproximatelyEquals(m_xCachedPos, POS_DELTA_THRESHOLD * POS_DELTA_THRESHOLD) )
        {
            bTestModules = false;
        }
        else
        {
			bTestModules = ( Asura_Timers::GetFrameNumber() >= m_uNextFrameForFullTest );

			//If the animation is near the camera , then recalculate the current module even if not time
			//to do so.
			//Otherwise if the camera is following an animation , the animation can bink out when crossing module
			//boundaries , since it still thinks its in the module that is now behind the camera.
			//RAR 02/07/2008
			if(!bTestModules && (*Asura_Camera::GetPosition() - xPos).MagnitudeSquared() < 16.0f)
			{
				bTestModules = true;
			}
        }
    }

	if( bTestModules )
    {
        g_pxCallbackAnim           = this;
        g_bCallbackFoundVisModule  = false;
        g_uCallbackNumModulesFound = 0;
		g_bCallbackInNonExteriorModule = false;
    
		Asura_Bounding_Box xAABB;

		xAABB.SetToPoint( m_xBSCentre );

		if (m_fCoarseBSIncEffectsRadius > 0.0f )
		{
			xAABB.Grow( m_fCoarseBSIncEffectsRadius );
		}

		if(Asura_Module_System::IsInitialised())
		{
			Asura_Module_System::TestModuleTree( xAABB, IntersectingModuleCallback );
		}
		else
		{
			return true;
		}

        if( 0 == g_uCallbackNumModulesFound || UINT_MAX == g_uCallbackNumModulesFound )
        {
            m_ausIntersectingModules[0] = USHRT_MAX;
        }
        else
        {
            if( g_uCallbackNumModulesFound < MAX_INTERSECTING_MODULES_TO_STORE )
            {
                m_ausIntersectingModules[g_uCallbackNumModulesFound++] = USHRT_MAX;
            }

            m_uNextFrameForFullTest = Asura_Timers::GetFrameNumber() + MIN_FRAMES_BEFORE_INTERSECTING_MODULE_LIST_REFRESH;
            m_xCachedPos            = xPos;
        }

        return g_bCallbackFoundVisModule;
    }
    else
    {
        for( u_int u = 0; u < MAX_INTERSECTING_MODULES_TO_STORE; ++u )
        {
            u_short usNext = m_ausIntersectingModules[u];
            if( USHRT_MAX == usNext )
            {
                Asura_Assert( u, "Should have tested module tree if we don't have any intersecting module indices stored." );
                break;
            }
            else
            {
                if( Asura_Module_System::IsModuleVisible(usNext) )
                {
                    return true;
                }
            }
        }

        return false;
    }
}
#endif //defined(ASURA_USE_STREAMING)

#endif //#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )



/*************************************************************
 *
 * Name          : Asura_Animation::SetAlpha
 *
 * Author        : Richard May
 *
 * Description   : Sets the alpha of the animation, including any
 *				   attachments it may have if bSetAttachments
 *				   is true (defaults to false)
 *
 *************************************************************/
void Asura_Animation::SetAlpha ( const float fAlpha, bool bSetAttachments/*=false*/ )
{
	m_xRGBA.m_fA = fAlpha;

	if(!bSetAttachments) return;

	Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;
	while(pxAttachment)
	{
		pxAttachment->SetAlpha(fAlpha, bSetAttachments);

		pxAttachment = pxAttachment->m_pxNextAttachment;
	}
}
/* end of function SetAlpha */


void Asura_Animation::SetRGB( float fR, float fG, float fB, bool bPropogateToAttachments )
{
    m_xRGBA.m_fR = fR;
    m_xRGBA.m_fG = fG;
    m_xRGBA.m_fB = fB;

    if( !bPropogateToAttachments ) { return; }

    for( Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment; pxAttachment; pxAttachment = pxAttachment->m_pxNextAttachment )
    {
        pxAttachment->SetRGB( fR, fG, fB, bPropogateToAttachments );
    }
}


void Asura_Animation::UpdateAttachmentLocations() const
{
	if(!m_apxSkins[0]) return;
	if(!m_pxFirstAttachment) return;

	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
	if (!pxTweenStore->IsPositionDataValid()) return;

    Asura_Hierarchy_Bone_RestFrame axBoneWorkspace[ASURA_MAX_NO_OF_BONES_PER_MODEL];

	const Asura_Hierarchy_Bone_RestFrame*	pxBoneData	= axBoneWorkspace;
	if( pxTweenStore->IsPositionDataInWorldspace() )
	{
		for( int i = 0; i < m_apxSkins[0]->NumberOfBones; i++ )
		{
			const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= pxTweenStore->GetBoneData(i);

			xBoneData.GetPosition(axBoneWorkspace[i].Position);
			xBoneData.GetOrientation(axBoneWorkspace[i].Quat);
		}
	}
	else
	{
		for( int i = 0; i < m_apxSkins[0]->NumberOfBones; i++ )
		{
			const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= pxTweenStore->GetBoneData(i);

			if( i == 0 )
			{
				// Root
				xBoneData.GetPosition(axBoneWorkspace[i].Position);
				xBoneData.GetOrientation(axBoneWorkspace[i].Quat);
			}
			else
			{
				Asura_Quat	xBoneOri;

				// Get parent
				int iParentIndex = m_apxSkins[0]->ParentBone[i];
				Asura_Assert(iParentIndex >= 0 && iParentIndex < i, "");

				xBoneData.GetPosition(axBoneWorkspace[i].Position);
				xBoneData.GetOrientation(xBoneOri);
				axBoneWorkspace[i].Position = axBoneWorkspace[iParentIndex].Quat.RotateVector( axBoneWorkspace[i].Position);
				axBoneWorkspace[i].Position += axBoneWorkspace[iParentIndex].Position;

				axBoneWorkspace[i].Quat = axBoneWorkspace[iParentIndex].Quat * xBoneOri;
			}
		}
	}

	//get each attachment , and store the current position/orientation
	const Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;
	while(pxAttachment)
	{
		pxAttachment->UpdateCachedPosition(pxBoneData,m_apxSkins[pxAttachment->GetMarkerPointSkinIndex()],pxTweenStore->IsPositionDataInWorldspace());
		pxAttachment = pxAttachment->m_pxNextAttachment;
	}
}


float Asura_Animation::GetGroundYPosition( const Asura_Vector_3* pxAnimPos, float* pfHeightAboveGround, float fMaxHeight ) const
{
    if( m_pxParent )
    {
        // Animation position specified presumably won't be correct for the
        // parent so don't pass it and hope that the stored position is
        // sufficiently up to date there. This should only happen during
        // rendering anyway (eg. when figuring out the position for the
        // shadow of an attachment) and therefore we'll have updated first.
        return m_pxParent->GetGroundYPosition( NULL, pfHeightAboveGround, fMaxHeight );
    }

    if( pxAnimPos )
    {
        UpdateStoredPosition( *pxAnimPos );
    }
    else
    {
        // Otherwise we'd better be reasonably up to date (see comment above).
        Asura_Assert( Asura_Timers::GetFrameNumber() - FrameNumberStamp <= 1, "Ground data will be out of date / completely wrong" );
    }

    if( !m_bHeightAboveGroundValid )
    {
		// the bounding box of the object may not be in contact with the ground; we'll have to do a line of sight test

        const Asura_Vector_3 xDownwards( 0.0f, 1.0f, 0.0f );

		// (AW - 04/08/05) - Set default max range to 10000m to avoid fp-overflows when we fail to hit any env.
		Asura_LOS_InputData xInputData( m_xPosition, xDownwards, ASURA_GUID_UNREGISTERED, fMaxHeight, false, m_uGroundFlagsToIgnore, m_uGroundFlagsToTestFor );

        // Small fudge to try and ensure we don't start the LOS test below
        // the ground.
        const float fORIGIN_OFFSET = 1.0f;
        xInputData.m_xOrigin.y -= fORIGIN_OFFSET;
		xInputData.m_fInitialMaximumRange += fORIGIN_OFFSET;
		const Asura_LOS_OutputData* pxOutputData = Asura_LOS::TestAgainstEnvironment(&xInputData);
        ASURA_LOS_TRACK_CALL( ASURA_LOS_CALLTYPE_ENV );
		
        if( pxOutputData->m_bHitEnvironment )
        {
		    m_fHeightAboveGround = (xInputData.m_xOrigin.y + pxOutputData->m_fDistanceToIntersection) - m_xPosition.y;
        }
        else
        {
            m_fHeightAboveGround = fASURA_ANIMATION_DEFAULT_MAX_GROUND_HEIGHT;
        }

		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			m_uGroundMaterialResponseHashID = pxOutputData->m_uMaterialResponseHashID;
		#endif

        m_sGroundMaterialID = pxOutputData->m_sMaterialID;
		m_xGroundNormal = pxOutputData->m_xNormal;

        m_bHeightAboveGroundValid = true;
    }

    if( pfHeightAboveGround ) { *pfHeightAboveGround = m_fHeightAboveGround; }

    return m_xPosition.y + m_fHeightAboveGround;
}

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

Asura_Hash_ID Asura_Animation::GetGroundMaterialResponseHashID( const Asura_Vector_3* pxAnimPos ) const
{
    if ( m_pxParent != NULL )
    {
        // See comment in GetGroundYPosition() for why we pass NULL here.
        return m_pxParent->GetGroundMaterialResponseHashID( NULL );
    }
    else
    {
        GetGroundYPosition( pxAnimPos );
		return m_uGroundMaterialResponseHashID;
    }
}

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

short Asura_Animation::GetGroundMaterialID( const Asura_Vector_3* pxAnimPos ) const
{
    if ( m_pxParent != NULL )
    {
        // See comment in GetGroundYPosition() for why we pass NULL here.
        return m_pxParent->GetGroundMaterialID( NULL );
    }
    else
    {
        GetGroundYPosition( pxAnimPos );
        return m_sGroundMaterialID;
    }
}


const Asura_Vector_3& Asura_Animation::GetGroundNormal( const Asura_Vector_3* pxAnimPos ) const
{
	if ( m_pxParent != NULL )
	{
		// See comment in GetGroundYPosition() for why we pass NULL here.
		return m_pxParent->GetGroundNormal( NULL );
	}
	else
	{
		GetGroundYPosition( pxAnimPos );
		return m_xGroundNormal;
	}
}


void Asura_Animation::UpdateLighting(u_int uRenderFlags /* = 0 */) const
{
    //if we have a parent , it can deal with this
    if(m_pxParent)
    {
        m_pxParent->UpdateLighting(uRenderFlags);
        return;
    }

    if( uRenderFlags & ASURA_HIERARCHY_RENDER_FORCE_AMBIENT_COLOUR_ONLY )
	{
        Asura_GI_Solution::s_xSHForCurrentRender.Zero();

		Asura_GI_Solution::s_xSHForCurrentRender.m_afRedCoefficients[0]   = m_xRGBA.m_fR;
		Asura_GI_Solution::s_xSHForCurrentRender.m_afGreenCoefficients[0] = m_xRGBA.m_fG;
		Asura_GI_Solution::s_xSHForCurrentRender.m_afBlueCoefficients[0]  = m_xRGBA.m_fB;

		Asura_GI_Solution::s_xStaticSHForCurrentRender.Zero();
		Asura_GI_Solution::s_xStaticSHForCurrentRender.m_afRedCoefficients[0] = m_xRGBA.m_fR;
		Asura_GI_Solution::s_xStaticSHForCurrentRender.m_afGreenCoefficients[0] = m_xRGBA.m_fG;
		Asura_GI_Solution::s_xStaticSHForCurrentRender.m_afBlueCoefficients[0] = m_xRGBA.m_fB;

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		Asura_GI_Solution::s_xDynamicSHForCurrentRender.Zero();
#endif
        return;
    }

	if(!m_apxSkins[0]) return;

	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
    if (!pxTweenStore->IsPositionDataValid()) return;

    Asura_Bounding_Box xBB = m_apxSkins[0]->GetBB();
	for (int i = 1; i < ASURA_ANIMATION_SKIN_COUNT; ++i)
	{
		if (m_apxSkins[i])
		{
			xBB.Set(m_apxSkins[i]->GetBB());
		}
	}
    Asura_Vector_3 xLightingPos = xBB.CalculateCentre();

	Asura_Vector_3	xBone0Pos;

	pxTweenStore->GetBoneData(0).GetPosition(xBone0Pos);
	//Displace the bounding box's position by the amount it has moved from its rest pose position.
	//This should help with lighting animations that move  away from their compass point. RAR 02/07/03
	if( pxTweenStore->IsPositionDataInWorldspace() )
	{
		xLightingPos += xBone0Pos;
	}
	else
	{
		xBone0Pos = m_xOrient.RotateVector( xBone0Pos );
		xBone0Pos += m_xPosition;
		xLightingPos += xBone0Pos;
	}

#if defined(ASURA_HIERARCHY_QUAD)
	Asura_Vector_3	xPosition = m_apxSkins[0]->m_pxBoneRestPoseTransforms[0].GetTranslation();
	xLightingPos -= xPosition;
#else
    xLightingPos -= m_apxSkins[0]->BoneRestPosition[0];
#endif	// defined(ASURA_HIERARCHY_QUAD)

    UpdateLighting( xLightingPos );
}


void Asura_Animation::UpdateLighting(const Asura_Vector_3& xLightingPosition) const
{
    const bool bJustLightOnce = m_bJustLightOnce;

    bool bUpdate;
	if( m_pxLightingSolution )
    {
        bUpdate = !bJustLightOnce;
    }
    else
    {
        m_pxLightingSolution = new Asura_Lighting_Solution( true, !bJustLightOnce );
        bUpdate = true;
    }

    if( bUpdate )
    {
	    m_pxLightingSolution->SetPosition( xLightingPosition );
    }

	bool bIncludeDynamicLights = !bJustLightOnce;
#if defined( ASURA_TARGET_SM3)
	if( !Asura_SortedRender::IsTransparentPass() )
	{
		bIncludeDynamicLights = false;
	}
#endif
	m_pxLightingSolution->RecalculateLighting( true, bIncludeDynamicLights );
}

void Asura_Animation::ResetDynamicLighting() const
{
	if(m_pxLightingSolution) m_pxLightingSolution->ResetDynamic();
}

bool Asura_Animation::GetBoneRelativePositionAndOrientation(const Asura_Vector_3& xObjectSpacePos, const Asura_Quat& xObjectSpaceOri, int iBoneID, Asura_Vector_3& xBoneRelativePos, Asura_Quat& xBoneRelativeOri) const
{
	//1: Get the bone position corresponding to our ID.
	const Asura_Hierarchy_Skin *pxSkin = GetSkin();

	if(pxSkin)
	{
		if(iBoneID < 0 || iBoneID >= pxSkin->NumberOfBones)
		{
			Asura_Assert(false, "Invalid bone ID");
			return false;
		}
	}
	else
	{
		Asura_Assert(false, "No Skin");
		return false;
	}


	//Code give-or-take copied from GetMarkerLocation.
	const Asura_Hierarchy_Tween* pxTweenStore = GetTweenStore();
	if(!pxTweenStore || !pxTweenStore->IsPositionDataValid()) return false;

	if(pxTweenStore->IsPositionDataInWorldspace())
	{
		const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= pxTweenStore->GetBoneData(iBoneID);
		Asura_Quat									xWorldSpaceOrientation;
		Asura_Vector_3								xWorldSpacePosition;

		xBoneData.GetOrientation(xWorldSpaceOrientation);
		xBoneData.GetPosition(xWorldSpacePosition);

		// Convert bone from world space to object space.
		Asura_Quat xCurrentInverseOri = m_xOrient;
		xCurrentInverseOri.Invert();
		
		// Calculate bone position and orientation in object space.
		Asura_Vector_3 xObjectSpacePosition = xCurrentInverseOri.RotateVector((xWorldSpacePosition - m_xPosition));
		const Asura_Quat xObjectSpaceOrientation = xCurrentInverseOri * xWorldSpaceOrientation;
		
		// Convert input parameters from object space to bone space.
		Asura_Quat xObjectSpaceInverseOrientation = xObjectSpaceOrientation;
		xObjectSpaceInverseOrientation.Invert();

		// Convert (xObjectSpacePos, xObjectSpaceOri) to (xBoneRelativePos, xBoneRelativeOri)
		xBoneRelativePos = xObjectSpaceInverseOrientation.RotateVector((xObjectSpacePos - xObjectSpacePosition));
		xBoneRelativeOri = xObjectSpaceInverseOrientation * xObjectSpaceOri;
	}
	else
	{
		Asura_Vector_3 xPosition(0.f, 0.f, 0.f);

		//Identity quaternion.
		Asura_Quat xQuat(0.f, 0.f, 0.f, 1.f);
		
		//work our way up the hierarchy applying the offsets for each section
		int iBoneIndex = iBoneID;
		while(iBoneIndex >= 0)
		{
			const Asura_Hierarchy_TweenStore_BoneData&	xBoneData	= pxTweenStore->GetBoneData(iBoneIndex);
			Asura_Quat									xBoneOri;
			Asura_Vector_3								xBonePos;

			xBoneData.GetOrientation(xBoneOri);
			xBoneData.GetPosition(xBonePos);

			//Rotate the position so far by the orientation of the child.
			xPosition = xBoneOri.RotateVector(xPosition);
			//Add the position of the child to the running position
			xPosition -= xBonePos;
			//Add the orientation of the child to the running orientation.
			xQuat = xBoneOri * xQuat;

			//If we have got to the base bone, break.
			if(!iBoneIndex) break;

			iBoneIndex = pxSkin->ParentBone[iBoneIndex];
		}

		//2: Get the orientation and position with respect to the parent bone.

		//Invert the quaternion transformation from bone space to object space.
		xQuat.x = -xQuat.x;
		xQuat.y = -xQuat.y;
		xQuat.z = -xQuat.z;

		//Get the bone relative position.
		xBoneRelativePos = xObjectSpacePos + xPosition;

		//Rotate the bone relative position into bone space
		xBoneRelativePos = xQuat.RotateVector(xBoneRelativePos);

		//Transform the object relative orientation into bone space.
		xBoneRelativeOri = xQuat * xObjectSpaceOri;
	}

	return true;
}

//Gets an attachment from a marker index.
Asura_AnimationAttachment* Asura_Animation::GetAttachmentFromMarkerIndex(int iMarkerIndex)
{
	Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;

	while(pxAttachment)
	{
		if( pxAttachment->m_iMarkerPointIndex == iMarkerIndex ) return pxAttachment;

		// check child attachments too
		if( pxAttachment->GetAttachedAnimation() )  
		{
			Asura_AnimationAttachment* pxChildAttachment = pxAttachment->GetAttachedAnimation()->GetAttachmentFromMarkerIndex( iMarkerIndex );
			if( pxChildAttachment )
			{
				return pxChildAttachment;
			}			
		}

		pxAttachment = pxAttachment->m_pxNextAttachment;
	}

	return NULL;
}

// (AAM - 11.07.2005) Const version
const Asura_AnimationAttachment* Asura_Animation::GetAttachmentFromMarkerIndex( int iMarkerIndex ) const
{
	Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment;

	while(pxAttachment)
	{
		if( pxAttachment->m_iMarkerPointIndex == iMarkerIndex ) return pxAttachment;

		// check child attachments too
		if( pxAttachment->GetAttachedAnimation() )  
		{
			Asura_AnimationAttachment* pxChildAttachment = pxAttachment->GetAttachedAnimation()->GetAttachmentFromMarkerIndex( iMarkerIndex );
			if( pxChildAttachment )
			{
				return pxChildAttachment;
			}			
		}

		pxAttachment = pxAttachment->m_pxNextAttachment;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Animation::GetAttachmentByID()
//
// Gets an attachment with the given identifier
//
// Parameters:		iID - Identifier of the attachment
// Return value:	Pointer to the attachment
//
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationAttachment* Asura_Animation::GetAttachmentByID(int iID)
{
	for( Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment; pxAttachment; pxAttachment = pxAttachment->m_pxNextAttachment )
	{
		if( pxAttachment->GetID() == iID ) return pxAttachment;

		// check child attachments
		if( pxAttachment->GetAttachedAnimation() )
		{
			Asura_AnimationAttachment* pxChildAttachment = pxAttachment->GetAttachedAnimation()->GetAttachmentByID( iID );
			if( pxChildAttachment )
			{
				return pxChildAttachment;
			}
		}
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Animation::GetAttachmentByID()
//
// Gets an attachment with the given identifier. Const version
//
// Parameters:		iID - Identifier of the attachment
// Return value:	Pointer to the (constant) attachment
//
////////////////////////////////////////////////////////////////////////////////

const Asura_AnimationAttachment* Asura_Animation::GetAttachmentByID( int iID ) const
{
	for( Asura_AnimationAttachment* pxAttachment = m_pxFirstAttachment; pxAttachment; pxAttachment = pxAttachment->m_pxNextAttachment )
	{
		if( pxAttachment->GetID() == iID ) return pxAttachment;

		// check child attachments
		if( pxAttachment->GetAttachedAnimation() )
		{
			Asura_AnimationAttachment* pxChildAttachment = pxAttachment->GetAttachedAnimation()->GetAttachmentByID( iID );
			if( pxChildAttachment )
			{
				return pxChildAttachment;
			}
		}
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::DestroyAllAttachments
// Author       : Andy Tate
// Date         : 5-Nov-2008
////////////////////////////////////////////////////////////////////////////////

void Asura_Animation::DestroyAllAttachments(void)
{
	//unlink any attachment
	while(m_pxFirstAttachment)
	{
		m_pxFirstAttachment->OnParentDeletion();
		m_pxFirstAttachment->UnlinkAttachment();
	}
}

/// End of function 'Asura_Animation::DestroyAllAttachments'. ///


void Asura_Animation::UpdateStoredPosition( const Asura_Vector_3& p ) const
{  
	if( m_bHeightAboveGroundValid )
	{
		// If the position has changed by more than 1 cm,
		// invalidate our height-above-ground record.
		if( !p.ApproximatelyEquals(m_xPosition, 0.0001f) )
		{
			m_bHeightAboveGroundValid = false;
		}
	}

	m_xPosition = p;
}

/// Richard May - 01/02/2005 10:36:35 - Removes any and all linked animations
void Asura_Animation::UnlinkAllAnimations( bool bDeleteOnUnlink/*=true*/ )
{
	while(m_pxFirstAttachment)
	{
		m_pxFirstAttachment->UnlinkAttachment(bDeleteOnUnlink);
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::CreateControllerGroup
// Author       : Andy Tate
// Date         : 20-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Group* Asura_Animation::CreateControllerGroup (const Asura_Hash_ID uControllerHash,
																		 const u_int uInitialFlags,
																		 const u_int uPriority,
																		 Asura_AnimationBlend* pxBlend,
																		 Asura_AnimationController_Group* pxParentGroup)
{
	if(uControllerHash == ASURA_HASH_ID_UNSET)
	{
		Asura_Assert(false, "Must specify a valid hash-ID when creating animation-controller groups.");
		return(NULL);
	}
	else
	{
		Asura_AnimationController_Group**	ppxNewGroup = NULL;

		// Do we have a parent group?
		if(pxParentGroup == NULL)
		{
			pxParentGroup	= m_pxRootControllerGroup;
		}

		// Create the new controller group...
		Asura_Hash_ID uFinalHash = uControllerHash;
		if (uInitialFlags & AAC_FLAG_ENSURE_UNIQUE_HASH)
		{
			// We've requested a unique hash, so loop on until we reach a value not in the tree.
			do
			{
				// Theoretically, we could loop round to ASURA_HASH_ID_UNSET doing this
				if (uFinalHash != ASURA_HASH_ID_UNSET)
				{
					ppxNewGroup	= m_xControllerGroupContainer.AddNode(uFinalHash);
				}
				if (!ppxNewGroup)
				{
					// That hash exists, increment the value and move on
					++uFinalHash;
				}
			}
			while (!ppxNewGroup);
		}
		else
		{
			// Only add if new.
			ppxNewGroup	= m_xControllerGroupContainer.AddNode(uFinalHash);
		}

		if(ppxNewGroup == NULL)
		{
			Asura_Assert(false, "Cannot create a new group - we already have one with the specified hash-ID.");
			return(NULL);
		}
		else
		{
			Asura_AnimationController_Group*	pxNewGroup	= new Asura_AnimationController_Group(uFinalHash, uInitialFlags, uPriority, pxParentGroup);
			pxNewGroup->SetAsuraAnimation( this );

			*ppxNewGroup	= pxNewGroup;
			if(pxParentGroup)
			{
				pxParentGroup->AddChildController(*pxNewGroup, pxBlend);
			}
			return(pxNewGroup);
		}
	}
}

/// End of function 'Asura_Animation::CreateControllerGroup'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::CreateControllerGroup
// Author       : Andy Tate
// Date         : 20-Jun-2007
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Group* Asura_Animation::CreateControllerGroup(const char* szControllerName,
																		const u_int uInitialFlags,
																		const u_int uPriority,
																		Asura_AnimationBlend* pxBlend,
																		Asura_AnimationController_Group* pxParentGroup)
{
	const Asura_Hash_ID					uControllerHash	= Asura_GetHashID(szControllerName);
	Asura_AnimationController_Group*	pxGroup			= CreateControllerGroup(uControllerHash, uInitialFlags, uPriority, pxBlend, pxParentGroup);

	if(pxGroup != NULL)
	{
		pxGroup->SetName(szControllerName);
	}

	return(pxGroup);
}

/// End of function 'Asura_Animation::CreateControllerGroup'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::CreateControllerAnimation
// Author       : Andy Tate
// Date         : 20-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Animation* Asura_Animation::CreateControllerAnimation (const Asura_Hash_ID uControllerHash,
																				 const u_int uInitialFlags,
																				 const u_int uPriority,
																				 const Asura_Hash_ID uAnimationHash,
																				 Asura_AnimationBlend* pxBlend,
																				 Asura_AnimationController_Group* pxParentGroup)
{
	if(uControllerHash == ASURA_HASH_ID_UNSET)
	{
		Asura_Assert(false, "Must specify a valid hash-ID when creating animation-controller animations.");
		return(NULL);
	}
	else
	{
		Asura_AnimationController_Animation**	ppxNewAnimation = NULL;

		// Do we have a parent group?
		if(pxParentGroup == NULL)
		{
			pxParentGroup	= m_pxRootControllerGroup;
		}

		// Create the new controller group...
		Asura_Hash_ID uFinalHash = uControllerHash;
		if (uInitialFlags & AAC_FLAG_ENSURE_UNIQUE_HASH)
		{
			// We've requested a unique hash, so loop on until we reach a value not in the tree.
			do
			{
				// Theoretically, we could loop round to ASURA_HASH_ID_UNSET doing this
				if (uFinalHash != ASURA_HASH_ID_UNSET)
				{
					ppxNewAnimation	= m_xControllerAnimationContainer.AddNode(uFinalHash);
				}
				if (!ppxNewAnimation)
				{
					// That hash exists, increment the value and move on
					++uFinalHash;
				}
			}
			while (!ppxNewAnimation);
		}
		else
		{
			// Only add if new.
			ppxNewAnimation	= m_xControllerAnimationContainer.AddNode(uFinalHash);
		}

		if(ppxNewAnimation == NULL)
		{
			Asura_Assert(false, "Cannot create a new animation - we already have one with the specified hash-ID.");
			return(NULL);
		}
		else
		{
			Asura_AnimationController_Animation*	pxNewAnimation	= new Asura_AnimationController_Animation(uFinalHash, uInitialFlags, uPriority, uAnimationHash);
			pxNewAnimation->SetAsuraAnimation( this );

			*ppxNewAnimation	= pxNewAnimation;
			pxParentGroup->AddChildController(*pxNewAnimation, pxBlend);
			return(pxNewAnimation);
		}
	}
}

/// End of function 'Asura_Animation::CreateControllerAnimation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::SetControllerAnimationID
// Author       : Richard May
// Date         : 09-Jul-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Animation::SetControllerAnimationID(const Asura_Hash_ID uOldID, const Asura_Hash_ID uNewID)
{
	Asura_AnimationController_Animation* pxAnimCtrl = GetControllerAnimationFromHash( uOldID );
	if( pxAnimCtrl )
	{
		if( m_xControllerAnimationContainer.ReplaceHash( uOldID, uNewID ) )
		{
			pxAnimCtrl->SetControllerHash( uNewID );
			return true;
		}				 		
	}

	return false;
}

/// End of function 'Asura_Animation::SetControllerAnimationID'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::CreateControllerAnimation
// Author       : Andy Tate
// Date         : 20-Jun-2007
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Animation* Asura_Animation::CreateControllerAnimation(const char* szControllerName,
																				const u_int uInitialFlags,
																				const u_int uPriority,
																				const Asura_Hash_ID uAnimationHash,
																				Asura_AnimationBlend* pxBlend,
																				Asura_AnimationController_Group* pxParentGroup)
{
	const Asura_Hash_ID						uControllerHash	= Asura_GetHashID(szControllerName);
	Asura_AnimationController_Animation*	pxAnimation		= CreateControllerAnimation(uControllerHash, uInitialFlags, uPriority, uAnimationHash, pxBlend, pxParentGroup);

	if(pxAnimation != NULL)
	{
		pxAnimation->SetName(szControllerName);
	}

	return(pxAnimation);
}

/// End of function 'Asura_Animation::CreateControllerAnimation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::DestroyController
// Author       : Andy Tate
// Date         : 20-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Animation::DestroyController (const Asura_Hash_ID uControllerHash)
{
	Asura_AnimationController_Group**	ppxGroup	= m_xControllerGroupContainer.GetDataByHash(uControllerHash);
	Asura_AnimationController_Group*	pxGroup		= (ppxGroup == NULL) ? NULL : *ppxGroup;

	// Did we find a group-controller to destroy?
	if(pxGroup == NULL)
	{
		Asura_AnimationController_Animation**	ppxAnimation	= m_xControllerAnimationContainer.GetDataByHash(uControllerHash);
		Asura_AnimationController_Animation*	pxAnimation		= (ppxAnimation == NULL) ? NULL : *ppxAnimation;

		// Did we find an animation-controller to destroy?
		if(pxAnimation == NULL)
		{
			Asura_Assertf(false, ( "Cannot find specified controller (0x%08x) to be destroyed.", uControllerHash ) );
			return(false);
		}
		else
		{
			Asura_AnimationController_Group**	ppxParentGroup	= m_xControllerGroupContainer.GetDataByHash(pxAnimation->GetParentControllerHash());
			Asura_AnimationController_Group*	pxParentGroup	= (ppxParentGroup == NULL) ? m_pxRootControllerGroup : *ppxParentGroup;

			// Do we have a parent group-controller?
			if(pxParentGroup == NULL)
			{
				Asura_Assert(false, "Cannot find parent group-controller to remove controller from.");
				return(false);
			}
			else
			{
				if(pxParentGroup->DestroyChildController(pxAnimation))
				{
					return(true);
				}
				else
				{
					Asura_Assert(false, "Failed to destroy child controller.");
					return(false);
				}
			}
		}
	}
	else
	{
		Asura_AnimationController_Group**	ppxParentGroup	= m_xControllerGroupContainer.GetDataByHash(pxGroup->GetParentControllerHash());
		Asura_AnimationController_Group*	pxParentGroup	= (ppxParentGroup == NULL) ? m_pxRootControllerGroup : *ppxParentGroup;

		// Do we have a parent group-controller?
		if(pxParentGroup == NULL)
		{
			Asura_Assert(false, "Cannot find parent group-controller to remove controller from.");
			return(false);
		}
		else
		{
			if(pxParentGroup->DestroyChildController(pxGroup))
			{
				return(true);
			}
			else
			{
				Asura_Assert(false, "Failed to destroy child controller.");
				return(false);
			}
		}
	}
}

/// End of function 'Asura_Animation::DestroyController'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::DestroyController
// Author       : Andy Tate
// Date         : 20-Jun-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Animation::DestroyController(const char* szControllerName)
{
	const Asura_Hash_ID	uControllerHash	= Asura_GetHashID(szControllerName);

	return(DestroyController(uControllerHash));
}

/// End of function 'Asura_Animation::DestroyController'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Animation::DestroyAllControllers
// Author       : Andy Tate
// Date         : 20-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_Animation::DestroyAllControllers (void)
{
	// Destroy all of the children of the root group...
	delete(m_pxRootControllerGroup);
	m_pxRootControllerGroup	= NULL;
}

/// End of function 'Asura_Animation::DestroyAllControllers'. ///

void Asura_Animation::UpdateTightBoundingSphere( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrient )
{
	Asura_Assert( m_uNumTimerUpdatesAtLastBoneUpdate == m_uNumTimerUpdates && m_uNumTimerUpdates > 0, "Bones aren't up to date" );
	if(m_uNumTimerUpdatesAtLastTightBoundingBoxUpdate != m_uNumTimerUpdates)
	{
		m_uNumTimerUpdatesAtLastTightBoundingBoxUpdate = m_uNumTimerUpdates;
		// Do we need to use section bounding-boxes to recalculate an accurate bounding box?
		if(m_bCalculateBSUsingBones && m_apxSkins[0] != NULL && (m_apxSkins[0]->m_pxSectionBoundingBoxData != NULL))
		{
			const u_int							uNumBones		= m_apxSkins[0]->NumberOfBones;
			u_int								uNumBoneBBsUsed	= 0;

			Asura_Bounding_Box	xBB;

			// Ensure that we have our world-space bone data for this animation...
			UpdateBoneWorkspace(*GetTweenStore(), xPosition, xOrient);

			// Accumulate the bounding-boxes from the bones...
			for(u_int uBone = 0; uBone < uNumBones; uBone++)
			{
				const Asura_Bounding_Box*	pxBoneBB	= GetBoneBB(uBone);

				if(pxBoneBB != NULL)
				{
					Asura_Bounding_Box	xBoneBB;

					xBoneBB.FromOBB(*pxBoneBB, g_xBoneOrientationWorkspace[uBone]);
					xBoneBB.Displace(g_xBonePositionWorkspace[uBone]);

					if(uNumBoneBBsUsed++ == 0)
					{
						xBB	= xBoneBB;
					}
					else
					{
						xBB.Set(xBoneBB);
					}
				}
			}

			// Did we find any bounding-boxes to use?
			Asura_Vector_3 xNewBSCentre;
			if(uNumBoneBBsUsed == 0)
			{
				UpdateBoundingBox();
				m_fTightBSRadius	= m_xObjectSpaceBB.CalculateRadius();
				xNewBSCentre		= m_xObjectSpaceBB.CalculateCentre();
				xNewBSCentre		= (xNewBSCentre * xOrient) + xPosition;
			}
			else
			{
				m_fTightBSRadius	= xBB.CalculateRadius();
				xNewBSCentre		= xBB.CalculateCentre();
			}
			
			// Update the attachments too...
			const Asura_AnimationAttachment*	pxAttachment	= m_pxFirstAttachment;
			while(pxAttachment != NULL)
			{
				pxAttachment->UpdateTightBoundingSphere(xNewBSCentre, m_fTightBSRadius);
				pxAttachment	= pxAttachment->GetNextAttachment();
			}

			float fCentreDisp = (xNewBSCentre - m_xBSCentre).Magnitude();
			m_fCoarseBSRadius += fCentreDisp;
		}
	}
}

void Asura_Animation::UpdateBoneData(const u_int uFirstBone, const u_int uLastBone)
{
	if(uLastBone < uFirstBone)
	{
		// no bones to update
		m_uNumTimerUpdatesAtLastBoneUpdate = m_uNumTimerUpdates;
		return;
	}

	Asura_Assert(m_apxSkins[0], "UpdateBoneData() failed - NULL Skin pointer");
	if (!m_apxSkins[0]) 
	{ 
		return; 
	}

    Asura_Assert( static_cast<int>(uLastBone) < m_apxSkins[0]->NumberOfBones, "" );
	if(m_uNumTimerUpdatesAtLastBoneUpdate != m_uNumTimerUpdates)
	{
		if( IsPosableAnim() )
		{
			Asura_Animation_Posable* pxPosableAnim = static_cast<Asura_Animation_Posable*> (this);
			pxPosableAnim->UpdatePoseAnims();
		}

		if(m_pxRootControllerGroup->IsActive())
		{
			m_pxRootControllerGroup->TweenBoneData(*m_apxSkins[0], uFirstBone, uLastBone, 1.0f, false, NULL);
		}
		m_uNumTimerUpdatesAtLastBoneUpdate	= m_uNumTimerUpdates;

		if( IsPosableAnim() )
		{
			//Do the posable bone update on the update that deals with all the bones , except possibly the first bone
			if((int)uLastBone == (m_apxSkins[0]->NumberOfBones-1))
			{
				Asura_Assert(uFirstBone == 0 || uFirstBone == 1, "changed behaviour");
				
				
				Asura_Animation_Posable* pxPosableAnim = static_cast<Asura_Animation_Posable*> (this);

				#if defined ASURA_ASSERT
					const int *piPosableBoneIndexArray = pxPosableAnim->GetPosableBoneIndexArray();
					if(piPosableBoneIndexArray)
					{
						Asura_Assert((int)uFirstBone <= piPosableBoneIndexArray[ASURA_ANIMATION_POSABLE_BONE_TURRET] && (int)uLastBone >= piPosableBoneIndexArray[ASURA_ANIMATION_POSABLE_BONE_HEAD],
							"unexpected number of bones to update for posable animation");
					}
				#endif

				pxPosableAnim->UpdatePosableBones();
			}
			else
			{
				//Expecting just the first bone to be done ,if not doing all the bones.
				Asura_Assert(uFirstBone == 0 && uLastBone == 0, "changed behaviour");
			}
		}

		Asura_List_It< Asura_Animation_BoneDataModifier > xListIt( m_xUpdateBoneDataList );
		while ( !xListIt.Done() )
		{
			Asura_Animation_BoneDataModifier* pxModifier = xListIt.GetCurrent();
			{
				pxModifier->UpdateBoneData( uFirstBone, uLastBone );
			}

			xListIt.Next();
		}

#ifdef ASURA_USE_ANIMATION_IK_SOLVER
		// Finally update the IK
		if ( m_pxIKSolver != NULL && m_bDoIKUpdates && ((int)uLastBone == (m_apxSkins[0]->NumberOfBones-1)) )
		{
			if ( m_uNumTimerUpdatesAtLastIKUpdate != m_uNumTimerUpdates )
			{
				m_uNumTimerUpdatesAtLastIKUpdate = m_uNumTimerUpdates;
				m_pxIKSolver->UpdateBones( *this );
			}
		}
#endif // ASURA_USE_ANIMATION_IK_SOLVER
	}
}


#ifdef ASURA_SUPPORT_SKIN_ELEMENTS

bool Asura_Animation::GetSkinElementVisibility(int iElement) const
{
	if (m_pabSkinElementVisible)
	{	
		Asura_Assert(iElement>=0 && iElement<=m_iNumSkinElements, "Skin element out of range!");
		return m_pabSkinElementVisible[iElement];
	}
	else
	{ 
		return true;
	}
}

void Asura_Animation::SetSkinElementVisibility(int iElement, bool bVisible)
{
	if (m_pabSkinElementVisible)
	{	
		Asura_Assert(iElement>=0 && iElement<=m_iNumSkinElements, "Skin element out of range!");
		m_pabSkinElementVisible[iElement] = bVisible;
	}
}

#endif


#ifdef ASURA_USE_DAMAGE_MAPS
float Asura_Animation::GetBoneDamageMapValue(int iBone) const
{
	if (iBone>=0 && iBone< m_iDamageMapBoneCount)
	{
		return m_pafBoneDamage[iBone];
	}
	else
	{
		return 0;
	}
}	

void  Asura_Animation::SetBoneDamageMapValue(int iBone, float fValue)
{
	if (m_apxSkins[0])
	{
		// Do we need to re-alloc the damage flags? (if the bone count has changed)
		if (m_apxSkins[0]->NumberOfBones != m_iDamageMapBoneCount)
		{		
			if (m_pafBoneDamage)
			{
				delete [] m_pafBoneDamage;
			}

			m_iDamageMapBoneCount = m_apxSkins[0]->NumberOfBones;
			m_pafBoneDamage = new float[m_iDamageMapBoneCount];
			for(int i=0;i<m_iDamageMapBoneCount;i++)
			{
				m_pafBoneDamage[i] = 0.0f;
			}
		}
	}

	if (iBone>=0 && iBone< m_iDamageMapBoneCount)
	{
		m_pafBoneDamage[iBone] = Asura_Maths::ClampToRange(fValue,0.0f,1.0f);
		m_bIsUsingDamageMap = true;
	}	
}

void  Asura_Animation::ResetDamageMapValues(void)
{
	for(int i=0;i<m_iDamageMapBoneCount;i++)
	{
		m_pafBoneDamage[i] = 0.0f;
	}
	m_bIsUsingDamageMap = false;
}
#endif

#ifdef ASURA_USE_DEBUG_TEXT
void Asura_Animation::PrintDebugInfo( u_int& uCol, u_int& uRow, const bool bRenderControllerInfo /*= false*/ ) const
{
	const Asura_Hierarchy_Skin* pxSkin = GetSkin();
	if( pxSkin )
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "%s", pxSkin->GetName() ));
	}
	else
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "<missing skin>" ));
	}
	
	uCol += 2;
	if ( m_pxRootControllerGroup && bRenderControllerInfo )
	{
		m_pxRootControllerGroup->PrintDebugInfo( uRow, uCol );
	}
	uCol -= 2;
	++uRow;

	const Asura_AnimationAttachment* pxAttach = GetFirstAttachment();
	while( pxAttach )
	{
		const Asura_Animation* pxAnim = pxAttach->GetAttachedAnimation();
		if( pxAnim )
		{
			pxAnim->PrintDebugInfo( uCol, uRow, bRenderControllerInfo );
			++uRow;
		}
		else
		{
		    Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "<not a skin>" ));
        }
        
		pxAttach = pxAttach->GetNextAttachment();
	}
}
#endif

#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)

void Asura_Animation::AddShadowBSToCoarseBS( float fShadowBSRad, const Asura_Vector_3& xShadowBSCentre )
{
	Asura_Vector_3			xDifference			= xShadowBSCentre - m_xBSCentre;
	const float				fCentreDifference	= xDifference.Normalise();
	const float				fMaximumRadius		= fCentreDifference + fShadowBSRad;

	if ( fMaximumRadius > m_fPrevCoarseBSIncEffectsRadius )
	{
		m_fPrevCoarseBSIncEffectsRadius	= fMaximumRadius;
	}
}

#endif

void Asura_Animation::UpdateCoarseBoundingSphere( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation ) 
{
	// Ensure that the bounding-box is up-to-date...
	if(m_uNumTimerUpdates == 0)
	{
		UpdateTimers(0.0f, true);
	}
	UpdateBoundingBox();

	m_fCoarseBSRadius	= m_xObjectSpaceBB.CalculateRadius();
	m_xBSCentre	= m_xObjectSpaceBB.CalculateCentre();
	m_xBSCentre	= (m_xBSCentre * xOrientation) + xPosition;

	// Use the value calculated in last frames deferred rendering (includes shadow info) unless it's not valid in which case use the excluding effects radius
	if ( m_fPrevCoarseBSIncEffectsRadius < 0.0f )
	{
		m_fCoarseBSIncEffectsRadius = m_fCoarseBSRadius;
	}
	else
	{
		m_fCoarseBSIncEffectsRadius = m_fPrevCoarseBSIncEffectsRadius;
	}

	// Initialise all radii to this coarse one for now. (we'll update the tight one later if necessary)
	m_fPrevCoarseBSIncEffectsRadius = m_fTightBSRadius = m_fCoarseBSRadius;

	// Add on attachments (attached anims add to m_fCoarseBSRadius, attached photons/projectors add to m_fCoarseBSIncEffectsRadius)
	const Asura_AnimationAttachment*	pxAttachment	= m_pxFirstAttachment;
	while(pxAttachment != NULL)
	{
		pxAttachment->UpdateCoarseBoundingSpheres( m_xBSCentre, m_fCoarseBSRadius, m_fCoarseBSIncEffectsRadius );
		pxAttachment	= pxAttachment->GetNextAttachment();
	}

#ifdef ASURA_USE_ANIMATION_IK_SOLVER
	if ( m_pxIKSolver && m_pxIKSolver->IsActive() )
	{
		const float fIKDist = m_pxIKSolver->GetRootRootBoneMaxMovement() * m_pxIKSolver->GetLerp();
		m_fCoarseBSRadius += fIKDist;
		m_fCoarseBSIncEffectsRadius += fIKDist;
	}
#endif
}

// KJFL - 15 November 2007  
// This function sets the first parameter to be a coarse world space bounding box covering the
// animation given the input world space position and orientation at which it it being played. 
// This is done by using the bounding sphere radius and centre members of the anim.
//
// This routine will intentionally *only* update the bone/anim data if the bounding data stored
// in the anim is uninitialised (e.g. it's the first frame and the anim has never been updated
// or rendered), or if the "bForceUpdate" parameter is set to true. Otherwise it will not update
// the anim for that frame and instead will use whatever bounding data is already present.
//
// Although that means it can potentially be a frame out of sync wrt the animation, the bounding
// volume is by no means tight on the object and so this should be ok for coarse trivial rejects.
void Asura_Animation::GetCoarseBoundingBox( Asura_Bounding_Box& xAABB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, bool bForceUpdate )
{
	// if the radius doesn't look valid, force an update
	// (the radius is set to -1.0f in the constructor)
	if ( bForceUpdate || m_fCoarseBSRadius < 0.0f )
	{
		UpdateCoarseBoundingSphere( xPosition, xOrientation );
	}

	xAABB.Reset();
	xAABB.Set( m_xBSCentre );

	if (m_fCoarseBSRadius > 0.0f )
	{
		xAABB.Grow( m_fCoarseBSRadius );
	}
}


#endif //ASURA_MAIN (Covers entire Asura_Animation and Asura_AnimationStatus class)


/// End of file. ///
