//
// Asura_AnimationAttachment.h
//
// Created by Richard Rice - 27/01/03
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif


#ifndef __ASURA_ANIMATIONATTACHMENT_H__
#define __ASURA_ANIMATIONATTACHMENT_H__

#include "Core/Asura_Maths.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Network/Asura_Guid.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM
#include "Entity/Container/Asura_ContainerState.h"
#endif //ASURA_USE_CONTAINER_SYSTEM

class Asura_Animation;
class Asura_Shape_Instance;
class Asura_Hierarchy_Bone_RestFrame;
class Asura_Hierarchy_Skin;
class Asura_Shadow;

//////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_AnimationAttachment **
**---------------------------*/

class Asura_AnimationAttachment
{
	friend class Asura_ClientEntity_Ragdoll;
	
	public :

        enum
        {
            INVALID_ID = -1,
            DEFAULT_ID =  0,
        };

		Asura_AnimationAttachment();
				
		/// DWE - 22/09/2003 10:25:43 - Copy constructor lite for use with ragdolls only
		Asura_AnimationAttachment(const Asura_AnimationAttachment* pxAnimAttach);

		virtual ~Asura_AnimationAttachment(){}

		//If bDeleteWhenUnlinked is set then the attachment structure will be deleted when unlinked from the animation.
		//This includes failure to add it in the first place, and when the animation is deleted.
		virtual bool LinkAttachment(Asura_Animation* pxParentAnimation,const char* pszMarkerName,bool bDeleteWhenUnlinked = true);
		virtual bool LinkAttachment(Asura_Animation* pxParentAnimation,Asura_Hash_ID uMarkerHash,bool bDeleteWhenUnlinked = true);
		virtual void UnlinkAttachment();
        void UnlinkAttachment( bool bAndDelete );

		//Makes a bone-relative attachment.
		virtual bool LinkAttachment_BoneRelative(Asura_Animation* pxParentAnimation, const Asura_Vector_3& xBoneRelativePos, const Asura_Quat& xBoneRelativeOri, int iBoneID, bool bDeleteWhenUnlinked = true) 
		{ 
			return false; 
		}

		// Virtual function which allows attachments to flag themselves for deletion. This is needed so that PFX 
		// attachments can tidy themselves up when the associated PFX finishes.
		virtual bool NeedsDeleting (void) const	{ return(false); }

		// Called when the visibility of the parent animation changes so that the change can be propagated each attachment
		virtual void OnParentVisibilityChanged( bool bInvisible ) {}

		//Functions for getting and setting a user set ID.For identification purposes with LOS tests
		void SetID(int iID);
		int GetID() const {return m_iID;}

        void SetCanBeTestedForLOS( const bool bCanBeTested ) { m_uCanBeTestedForLOS = bCanBeTested; }
		//Indicates whether this type of attachment can be tested for LOS. Saves wasting time
		//calculating the position when the LOS function is just going to ignore it anyway
		bool CanBeTestedForLOS() const { return ( m_uCanBeTestedForLOS != 0 ); }

		virtual bool LOSTest(const Asura_Vector_3& xPosition,const Asura_Matrix_3x3 xOrient,const class Asura_LOS_InputData* pxInput,class Asura_LOS_OutputData* pxOutput) const
		{return false;}

		virtual Asura_AnimationAttachment* Clone(Asura_Animation* pxParent);

		void SetDeleteAttachedObjectWithParent( bool bSet ) { m_uDeleteAttachedObjectWithParent = bSet; }
        void SetDeleteWhenUnlinked( bool bSet ) { m_uDeleteWhenUnlinked = bSet; }

		//get the position/orient of the attachment relative to the parent animation
		//(or relative to world for ragdolls , since they do everything world relative)
		bool GetPosition(Asura_Vector_3* pxOutputPosition,Asura_Matrix_3x3* pxOutputOrient) const;

		//Gets the position/orient of the attachment relative to the parent animation.
		bool GetParentRelativePosition(Asura_Vector_3* pxOutputPosition,Asura_Matrix_3x3* pxOutputOrient) const;

		//Gets the position/orient of the attachement relative to the world.
		bool GetWorldRelativePosition(Asura_Vector_3* pxOutputPosition,Asura_Matrix_3x3* pxOutputOrient) const;

		//Sets a world relative orientation override (which is used instead of the orientation
		//from the marker point). This orientation override is kept until cleared.
		void SetWorldRelativeOrientation(const Asura_Matrix_3x3& xOrient);
		void SetWorldRelativeOrientation(const Asura_Quat& xOrient);
		void SetWorldRelativePosition(const Asura_Vector_3& xPos);

		//Clears the orientation override , and causes the marker point's orientation to be used again
		void ClearOrientationOverride();
		void ClearPositionOverride();
		
		//somewhat unfortunate virtual function to deal with some of the old animation attachment code
		virtual Asura_Animation* GetAttachedAnimation() const {return 0;}
		Asura_AnimationAttachment* GetNextAttachment(void) { return m_pxNextAttachment; }
		const Asura_AnimationAttachment* GetNextAttachment(void) const { return m_pxNextAttachment; }
		const Asura_Animation* GetParentAnimation(void) const { return(m_pxParentAnimation); }

		// AJG - 28/05/2004 - made these two functions public, and added the accessor for the marker point index, so subclasses of 
		// Asura_ClientEntity_Ragdoll can still render attachments on their animations.
		// RAR - 13/06/2005 - virtual render function made protected again , instead we
		// have a public non-virtual render that will calculate the position/orient for the other
		virtual void Update(float fTime, bool bIgnoreTimerFrameStamp){}
		void Render(unsigned uRenderFlags);

		void RenderShadow( const u_int uRenderFlags );
		void RenderShadow( Asura_Shadow& xShadow, const u_int uRenderFlags ) const;

		// RJF - 10 May 2005 - Propogating alpha to attachments
		virtual void SetAlpha(float fAlpha, bool bSetAttachments) { }
        virtual void SetRGB( float fR, float fG, float fB, bool bPropogateToAttachments ) { }

		int  GetMarkerPointIndex(void) const   { return m_iMarkerPointIndex; }
		void SetMarkerPointIndex( int iIndex ) { m_iMarkerPointIndex = iIndex; }
		u_int  GetMarkerPointSkinIndex(void) const   { return m_uMarkerPointSkinIndex; }
		void SetMarkerPointSkinIndex( u_int uIndex ) { m_uMarkerPointSkinIndex = uIndex; }

		//Gets the bone that the attachmen's marker point is attached to
		//(or just the bone that the attachment is attached to if bone relative)
		virtual int GetBoneID() const;


		//used by Asura_Animation::UpdateAttachmentLocations to update the positions
		//from the bone data
		virtual bool UpdateCachedPosition(const Asura_Hierarchy_Bone_RestFrame* pxBoneData,const Asura_Hierarchy_Skin* pxSkin,bool bWorldRelative) const;

		// Bounding-sphere update functions...
		void UpdateTightBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
		void UpdateCoarseBoundingSpheres(Asura_Vector_3& xCentre, float& fRadius, float& fRadiusIncEffects) const;

		// AJG - 22/06/06 - Accessor to test if we're already linked to something, to make it easier to avoid attaching the same thing twice.
		bool IsLinked() const { return m_pxParentAnimation != NULL; }

		// Accessor functions when rendering...
		void	SetRenderFlagsToAdd		(const u_int uFlags)	{ m_uRenderFlagsToAdd = uFlags; }
		void	SetRenderFlagsToRemove	(const u_int uFlags)	{ m_uRenderFlagsToRemove = uFlags; }
		u_int	GetRenderFlagsToAdd		(void) const			{ return(m_uRenderFlagsToAdd); }
		u_int	GetRenderFlagsToRemove	(void) const			{ return(m_uRenderFlagsToRemove); }

		void	AddRenderFlags		(const u_int uFlags)		{ m_uRenderFlagsToAdd |= uFlags; }
		void	RemoveRenderFlags	(const u_int uFlags)		{ m_uRenderFlagsToRemove |= uFlags; }

		void	InvalidateFrameStamp();

	protected :
		virtual void Render(const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags){}
		virtual void RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) {}
		virtual void RenderShadow( Asura_Shadow& xShadow, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) const {}

		virtual	bool	GetTightBoundingSphere	(Asura_Vector_3& xCentre, float& fRadius) const	{ return(false); }
		virtual	bool	GetCoarseBoundingSphere	(Asura_Vector_3& xCentre, float& fRadius) const	{ return(false); }
		virtual	bool	GetCoarseBoundingSphereIncludingEffects	(Asura_Vector_3& xCentre, float& fRadius) const	{ return(false); }

		void SetCachedPositionAndOrient(const Asura_Vector_3& xPos,const Asura_Quat& xQuat,bool bWorldRelative) const;
		bool ArePositionAndOrientCached(void) const;

		//Used internally to apply the movement from an attached animation to
		//the cached positions  while they are being updated
		void ApplyAnimationMovementOffset(const Asura_Animation* pxAnim) const;

		//called when the parent animation is deleted
		virtual void OnParentDeletion(){}

		Asura_Animation* m_pxParentAnimation;
		Asura_AnimationAttachment* m_pxNextAttachment;
		int m_iMarkerPointIndex;//marker point in parent that this animation is attached to
		u_int m_uMarkerPointSkinIndex; // which skin in the animations contains this attachment


		int m_iID; //User set ID for this. Not necessarily unique

	private :
		//cached position/orientation data
        // (MP - 26.05.2005) NB: If the parent animation tween storage is in
        // worldspace then this data will be as well.
		// (RAR - 13.06.2005) Cached orient may store a world orientation
		// override instead.
		mutable Asura_Vector_3 m_xCachedPos;
		mutable Asura_Quat m_xCachedOrient;
		mutable u_int m_uCachedFrameStamp;

	protected :


		unsigned int m_uDeleteWhenUnlinked : 1;
		unsigned int m_uDeleteAttachedObjectWithParent : 1;//If true then the animation,corona,shape that this attachment deals with should be deleted when the parent animation is deleted
        unsigned int m_uCanBeTestedForLOS : 1;
		unsigned int m_uWorldOrientOverride : 1; // m_xCachedOrient contains an override orient instead
		unsigned int m_uWorldPosOverride : 1; // m_xCachedPos contains an override position instead
		mutable unsigned int m_uWorldRelative : 1; //true if the root model (and hence this) is using world relative positions and orients(ie ragdoll).

		u_int	m_uRenderFlagsToAdd;		// Flags to add when rendering.
		u_int	m_uRenderFlagsToRemove;		// Flags to remove when rendering.

		friend class Asura_Animation;

};

//////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------------**
** Asura_AnimationAttachment_Animation **
**-------------------------------------*/

class Asura_AnimationAttachment_Animation : public Asura_AnimationAttachment
{
	private :
		typedef Asura_AnimationAttachment PARENT;

	public :
		Asura_AnimationAttachment_Animation(Asura_Animation* pxAnimation,bool bDeleteAnimationWithParent);

		// constructor for use with ragdolls;
		Asura_AnimationAttachment_Animation(const Asura_AnimationAttachment_Animation* pxAnimAttach);

		virtual ~Asura_AnimationAttachment_Animation();

		virtual bool LinkAttachment(Asura_Animation* pxParentAnimation,const char* pszMarkerName,bool bDeleteWhenUnlinked = true);
		virtual bool LinkAttachment(Asura_Animation* pxParentAnimation,Asura_Hash_ID uMarkerHash,bool bDeleteWhenUnlinked = true);
		virtual void UnlinkAttachment();

		virtual bool LOSTest(const Asura_Vector_3& xPosition,const Asura_Matrix_3x3 xOrient,const class Asura_LOS_InputData* pxInput,class Asura_LOS_OutputData* pxOutput) const;
	
		virtual Asura_AnimationAttachment* Clone(Asura_Animation* pxParent);
		virtual Asura_Animation* GetAttachedAnimation() const {return m_pxAnimation;}		

		// AJG - 28/05/2004 - made these two functions public, so subclasses of Asura_ClientEntity_Ragdoll can still render attachments 
		// on their animations.
		virtual void Update(float fTime, bool bIgnoreTimerFrameStamp);

		virtual void SetAlpha( float fAlpha, bool bSetAttachments );
        virtual void SetRGB( float fR, float fG, float fB, bool bPropogateToAttachments );
		virtual void SetSkin( Asura_Hash_ID uSkinID );
		virtual void OnParentVisibilityChanged( bool bInvisible );

		void SetDeleteAnimOnUnlink( bool bDelete );
		void SetUseMovementOffset( bool bUseOffset );
		void SetUpdateTimeOverride( float fTime );				

		Asura_Hash_ID GetSkinHashID( void ) const;

	protected :
		virtual void Render(const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags);
		virtual void RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags );
		virtual void RenderShadow( Asura_Shadow& xShadow, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) const;

		virtual bool GetTightBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
		virtual bool GetCoarseBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;

		virtual void OnParentDeletion();		

		virtual bool UpdateCachedPosition(const Asura_Hierarchy_Bone_RestFrame* pxBoneData,const Asura_Hierarchy_Skin* pxSkin,bool bWorldRelative) const;

		Asura_Animation* m_pxAnimation;
		bool  m_bShouldDeleteAnimation;
		bool  m_bUseMovementOffset;
		float m_fUpdateTimeOverride;
};

//////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_AnimationAttachment_Corona **
**----------------------------------*/

class Asura_AnimationAttachment_Corona : public Asura_AnimationAttachment
{
	public :
		Asura_AnimationAttachment_Corona(Asura_Handle iCoronaHandle,bool bDeleteCoronaWithParent);

		//  AAB:  (08/04/2004 12:13:08):  Copy constructor for ragdoll
		Asura_AnimationAttachment_Corona(const Asura_AnimationAttachment_Corona *pxAnimAttach);

		Asura_Handle GetCoronaHandle() const{return m_iCoronaHandle;}
		Asura_Handle& GetCoronaHandle() {return m_iCoronaHandle;}
	
		virtual Asura_AnimationAttachment* Clone(Asura_Animation* pxParent);


	protected :
		virtual void Render(const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags);
		virtual void OnParentDeletion();

		virtual bool GetCoarseBoundingSphereIncludingEffects(Asura_Vector_3& xCentre, float& fRadius) const;

		Asura_Handle m_iCoronaHandle;
};

//////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_AnimationAttachment_Shape **
**---------------------------------*/

class Asura_AnimationAttachment_Shape : public Asura_AnimationAttachment
{
	public :
		Asura_AnimationAttachment_Shape(Asura_Hash_ID uShapeHashID);
		~Asura_AnimationAttachment_Shape();

		Asura_AnimationAttachment_Shape(const Asura_AnimationAttachment_Shape *pxAnimAttach);

        virtual bool LinkAttachment( Asura_Animation* pxParentAnimation, const char* szMarkerName, bool bDeleteWhenUnlinked = true )
        {
            return Asura_AnimationAttachment::LinkAttachment( pxParentAnimation, szMarkerName, bDeleteWhenUnlinked );
        }
		virtual bool LinkAttachment(Asura_Animation* pxParentAnimation, Asura_Hash_ID uMarkerHash, bool bDeleteWhenUnlinked = true);
		virtual void UnlinkAttachment();

		void SetShape(Asura_Hash_ID uShapeHashID);

		virtual Asura_AnimationAttachment* Clone(Asura_Animation* pxParent);

		virtual bool LOSTest(const Asura_Vector_3& xPosition,const Asura_Matrix_3x3 xOrient,const class Asura_LOS_InputData* pxInput,class Asura_LOS_OutputData* pxOutput) const;


        const Asura_Shape_Instance* GetShapeInstance( void ) const { return m_pxShapeInstance; }
        Asura_Shape_Instance* GetShapeInstance( void ) { return m_pxShapeInstance; }

		virtual void SetAlpha(float fAlpha, bool bSetAttachments);
// (MP - 01.06.2006) Really this ought to be here but we don't pass the
// ambient-only stuff from the render flags onto the shape, and never have
// done. I've left it with this old (possibly broken) behaviour for now.
//      virtual void SetRGB( float fR, float fG, float fB, bool bPropogateToAttachments );
        
	protected :
		virtual void Render(const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags);
		virtual void RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags );
		virtual void RenderShadow( Asura_Shadow& xShadow, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) const;

		virtual bool GetTightBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
		virtual	bool GetCoarseBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
		Asura_Shape_Instance* m_pxShapeInstance;
};

//////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------------**
** Asura_AnimationAttachment_BoneRelative **
**----------------------------------------*/

//Class for attachments that are relative to a bone instead of a marker point.
//Instancs of this shouldn't be created , instead the derived bone relative
//attachments should be used

class Asura_AnimationAttachment_BoneRelative : public Asura_AnimationAttachment
{
	public :
		Asura_AnimationAttachment_BoneRelative(){}
		Asura_AnimationAttachment_BoneRelative(const Asura_AnimationAttachment_BoneRelative* pxAnimAttach);

		virtual bool LinkAttachment_BoneRelative( Asura_Animation* pxParentAnimation, const Asura_Vector_3& xBoneRelativePos, const Asura_Quat& xBoneRelativeOri, int iBoneID, bool bDeleteWhenUnlinked = true);

		virtual const Asura_Vector_3& GetBoneRelativePosition() const { return m_xBoneRelativePosition; }
		virtual const Asura_Quat& GetBoneRelativeOrientation() const { return m_xBoneRelativeOrientation; }
		virtual int GetBoneID() const { return m_iBoneID; }
		virtual void SetBoneRelativePosition(const Asura_Vector_3& xBoneRelativePosition) { m_xBoneRelativePosition = xBoneRelativePosition; }
		virtual void SetBoneRelativeOrientation(const Asura_Quat& xBoneRelativeOrientation) { m_xBoneRelativeOrientation = xBoneRelativeOrientation; }

		//used by Asura_Animation::UpdateAttachmentLocations to update the positions
		//from the bone data
		virtual bool UpdateCachedPosition(const Asura_Hierarchy_Bone_RestFrame* pxBoneData,const Asura_Hierarchy_Skin* pxSkin,bool bWorldRelative) const;

	protected :
		int	m_iBoneID;								//The id of the bone this attachment is attached to.
		Asura_Vector_3 m_xBoneRelativePosition;		//The position relative to the parent bone.
		Asura_Quat m_xBoneRelativeOrientation;		//The orientation relative to the parent bone.
};


//////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------------**
** Asura_AnimationAttachment_BoneRelativeAnimation **
**-------------------------------------------------*/
//This allows for animations that aren't tied to marker points.

class Asura_AnimationAttachment_BoneRelativeAnimation : public Asura_AnimationAttachment_BoneRelative
{
	public :
		Asura_AnimationAttachment_BoneRelativeAnimation( Asura_Animation* pxAnimation, bool bDeleteAnimationWithParent, bool bCanBeTestedForLOS = false );

		// constructor for use with ragdolls;
		Asura_AnimationAttachment_BoneRelativeAnimation( const Asura_AnimationAttachment_BoneRelativeAnimation* pxAnimAttach );

		virtual ~Asura_AnimationAttachment_BoneRelativeAnimation();

		virtual bool LinkAttachment_BoneRelative(Asura_Animation* pxParentAnimation, const Asura_Vector_3& xBoneRelativePos, const Asura_Quat& xBoneRelativeOri, int iBoneID, bool bDeleteWhenUnlinked = true);


		virtual void UnlinkAttachment();

		virtual bool LOSTest( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3 xOrient, const class Asura_LOS_InputData* pxInput, class Asura_LOS_OutputData* pxOutput ) const;
		
		virtual Asura_AnimationAttachment* Clone( Asura_Animation* pxParent );

		virtual void SetAlpha(float fAlpha, bool bSetAttachments);
        virtual void SetRGB( float fR, float fG, float fB, bool bPropogateToAttachments );

	protected :
		virtual void Update(float fTime, bool bIgnoreTimerFrameStamp);
		virtual void Render(const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags);
		virtual void RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags );
		virtual void RenderShadow( Asura_Shadow& xShadow, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) const;

		virtual	bool GetTightBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
		virtual	bool GetCoarseBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;

		virtual void OnParentDeletion();

		virtual Asura_Animation* GetAttachedAnimation() const {return m_pxAnimation;}

		virtual bool UpdateCachedPosition(const Asura_Hierarchy_Bone_RestFrame* pxBoneData,const Asura_Hierarchy_Skin* pxSkin,bool bWorldRelative) const;

		Asura_Animation* m_pxAnimation;
		bool m_bShouldDeleteAnimation;

};


#ifdef ASURA_TARGET_SM3

//////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------**
** Asura_AnimationAttachment_Projector **
**-------------------------------------------*/

// Projectors that are attached at marker points.

class Asura_AnimationAttachment_Projector : public Asura_AnimationAttachment
{
public:

	// IMPORTANT: effect id must be an entity effect!
	Asura_AnimationAttachment_Projector( int iProjectorHandle, bool bDeleteEffectWithParent );
	Asura_AnimationAttachment_Projector( const Asura_AnimationAttachment_Projector* pxOther );

	virtual ~Asura_AnimationAttachment_Projector();

	virtual Asura_AnimationAttachment* Clone( Asura_Animation* pxParent );

	virtual void Update( float fTime, bool bIgnoreTimerFrameStamp );

protected :
	virtual bool GetCoarseBoundingSphereIncludingEffects(Asura_Vector_3& xCentre, float& fRadius) const;

private:

	int m_iProjectorHandle;
};
#endif // ASURA_TARGET_SM3


//////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------**
** Asura_AnimationAttachment_PFX **
**-------------------------------------------*/

// Particle effects that are attached at marker points - identical to 
// Asura_AnimationAttachment_BoneRelativePFX but with different base class.


class Asura_AnimationAttachment_PFX : public Asura_AnimationAttachment
{
	public :

		// IMPORTANT: effect id must be an entity effect!
		Asura_AnimationAttachment_PFX( u_int uEffectID, bool bDeleteEffectWithParent );
		// Alternative constructor that takes an ID of an effect template
		Asura_AnimationAttachment_PFX( Asura_Hash_ID uTemplateID,unsigned short usPFXFlags, Asura_Guid uPFXParentGuid = ASURA_GUID_UNREGISTERED);

		Asura_AnimationAttachment_PFX( const Asura_AnimationAttachment_PFX* pxOther );
		virtual ~Asura_AnimationAttachment_PFX();

		virtual Asura_AnimationAttachment* Clone( Asura_Animation* pxParent );

		virtual void Update( float fTime, bool bIgnoreTimerFrameStamp );
		virtual bool NeedsDeleting (void) const;

		virtual void OnParentVisibilityChanged( bool bInvisible );


		//Determines whether the attachment should be deleted automatically when the PFX finishes - Defaults to true
		void SetDeleteWhenFinished(bool bSet) {m_uDeleteWhenFinished = bSet ? 1 : 0;}

		//Determines whether the attachment should be started automatically when the animation updates - Defaults to false, unless created from a template ID
		void SetStartOnFirstUpdate(bool bSet) {m_uStartOnFirstUpdate = bSet ? 1 : 0;}
		
	protected :
		virtual void Render( const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags );
		
		virtual bool GetCoarseBoundingSphereIncludingEffects(Asura_Vector_3& xCentre, float& fRadius) const;

		u_int m_uEffectGuid;

		u_int m_uDeleteWhenFinished : 1;
		u_int m_uStartOnFirstUpdate : 1;

};

//////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------------------**
** Asura_AnimationAttachment_BoneRelativePFX **
**-------------------------------------------*/

// Particle effects that aren't attached at marker points


class Asura_AnimationAttachment_BoneRelativePFX : public Asura_AnimationAttachment_BoneRelative
{
	public :

		// IMPORTANT: effect id must be an entity effect!
		Asura_AnimationAttachment_BoneRelativePFX( u_int uEffectID, bool bDeleteEffectWithParent );
		// Alternative constructor that takes an ID of an effect template
		Asura_AnimationAttachment_BoneRelativePFX( Asura_Hash_ID uTemplateID,unsigned short usPFXFlags, Asura_Guid uPFXParentGuid = ASURA_GUID_UNREGISTERED );
		Asura_AnimationAttachment_BoneRelativePFX( const Asura_AnimationAttachment_BoneRelativePFX* pxOther );
		virtual ~Asura_AnimationAttachment_BoneRelativePFX();

		virtual Asura_AnimationAttachment* Clone( Asura_Animation* pxParent );

		virtual void Update( float fTime, bool bIgnoreTimerFrameStamp );
		virtual bool NeedsDeleting (void) const;

		virtual void OnParentVisibilityChanged( bool bInvisible );

		//Determines whether the attachment should be deleted automatically when the PFX finishes - Defaults to true
		void SetDeleteWhenFinished(bool bSet) {m_uDeleteWhenFinished = bSet ? 1 : 0;}

		//Determines whether the attachment should be started automatically when the animation updates - Defaults to false, unless created from a template ID
		void SetStartOnFirstUpdate(bool bSet) {m_uStartOnFirstUpdate = bSet ? 1 : 0;}

	protected :

		virtual bool GetCoarseBoundingSphereIncludingEffects(Asura_Vector_3& xCentre, float& fRadius) const;

		u_int m_uEffectGuid;

		u_int m_uDeleteWhenFinished : 1;
		u_int m_uStartOnFirstUpdate : 1;

};


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationAttachment_Sticky
// Author       : Andy Tate
// Date         : 6-Nov-2006
//
// Description  : Class to attach an animation as if it were a ragdoll.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationAttachment_Sticky : public Asura_AnimationAttachment_Animation
{
public:
	// Public functions...
	Asura_AnimationAttachment_Sticky          (Asura_Animation* pxAnimation, const bool bDeleteAnimationWithParent, const u_int uFixedBoneID, const Asura_Vector_3& xWorldSpacePosition);
//	virtual ~Asura_AnimationAttachment_Sticky ();

	// Accessor functions...
	u_int	GetStickyBoneID	(void) const	{ return(m_uStickyBoneID); }

	bool GetRotation (Asura_Quat& xRotation) const
	{
		if(m_bHasRotation)
		{
			xRotation	= m_xRotation;
			return(true);
		}
		else
		{
			return(false);
		}
	}

	bool GetBoneOffset (Asura_Vector_3& xBoneOffset) const
	{
		if(m_bHasBoneOffset)
		{
			xBoneOffset	= m_xBoneOffset;
			return(true);
		}
		else
		{
			return(false);
		}
	}


protected:
	// Protected variables...
	union
	{
		u_int		m_uFlags;				// Single access point for the flags.

		struct
		{
			u_int	m_uStickyBoneID		: 16,	// The ID of the bone to attach to.
					m_bHasRotation		: 1,	// Flag to indicate that we've calculated our rotation.
					m_bHasBoneOffset	: 1,	// Flag to say whether we have calculated our bone-offset or not.
										: 0;	// 14 spare-bits.
		};
	};

	Asura_Vector_3	m_xBoneOffset;
	Asura_Quat		m_xRotation;

	// Protected functions...
	virtual	void	Update					(float fTime, bool bIgnoreTimerFrameStamp);
	virtual	bool	UpdateCachedPosition	(const Asura_Hierarchy_Bone_RestFrame* pxBoneData, const Asura_Hierarchy_Skin* pxSkin, bool bWorldRelative) const;


private:
	// Private typedefs...
	typedef Asura_AnimationAttachment_Animation PARENT;

	// Private variables...

	// Private functions...
	Asura_AnimationAttachment_Sticky  (void);
	Asura_AnimationAttachment_Sticky  (const Asura_AnimationAttachment_Sticky& xOther);

	Asura_AnimationAttachment_Sticky&	operator=	(const Asura_AnimationAttachment_Sticky& xOther);

};   // End of class 'Asura_AnimationAttachment_Sticky'.



//////////////////////////////////////////////////////////
// Asura_AnimationAttachment_ContainerInstance_Internal //
//////////////////////////////////////////////////////////
// Animation attachment used internally by the container instance to manage containers
// that are attached to marker points on meshes.
#ifdef ASURA_USE_CONTAINER_SYSTEM
class Asura_AnimationAttachment_ContainerInstance_Internal : public Asura_AnimationAttachment
{
private:
	Asura_AnimationAttachment_ContainerInstance_Internal(const class Asura_ClientContainerInstance_Attachment* pxContainerInstance)
		: m_pxContainerInstance(pxContainerInstance)
	{
		SetCanBeTestedForLOS(false);
	}

public:
	virtual ~Asura_AnimationAttachment_ContainerInstance_Internal() {}

	virtual Asura_AnimationAttachment* Clone(Asura_Animation* pxParent);

protected:
	virtual void Render(const Asura_Vector_3& xPos,const Asura_Matrix_3x3& xOrient,unsigned uRenderFlags);
	virtual void RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) {}
	virtual void RenderShadow( Asura_Shadow& xShadow, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) const;

	virtual bool GetTightBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
	virtual bool GetCoarseBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
	virtual bool GetCoarseBoundingSphereIncludingEffects(Asura_Vector_3& xCentre, float& fRadius) const;

private:
	// Make these functions private.
	Asura_AnimationAttachment_ContainerInstance_Internal(const Asura_AnimationAttachment_ContainerInstance_Internal&);
	Asura_AnimationAttachment_ContainerInstance_Internal& operator= (const Asura_AnimationAttachment_ContainerInstance_Internal&);

	const class Asura_ClientContainerInstance_Attachment* m_pxContainerInstance;

	// Friends
	friend class Asura_ClientContainerInstance_Mesh;
};


///////////////////////////////////////////////////////
// Asura_AnimationAttachment_ClientContainerInstance //
///////////////////////////////////////////////////////
// Animation attachment which allows things other than the container system to attach container instances
// to their animations. This class will create and destroy a client container instance as needed.

// Note: because there is only a client instance and no server instance then health, damage and
// dynamic physics can't really be used. If you need these things then use a container entity instead.
// Mesh, projectors, photons, coronas, lightshafts, PFX, sound events etc should work fine though.

// Note: LOS tests return m_iHitAttachmentID that will be set using GetID(), not the unique id for the sub-container.

class Asura_AnimationAttachment_ClientContainerInstance : public Asura_AnimationAttachment
{
public:
	Asura_AnimationAttachment_ClientContainerInstance(Asura_Hash_ID uContainerHash,
													  const Asura_Container_InputState& xInputState,
													  Asura_Guid uOwnerGuid = ASURA_GUID_UNREGISTERED);

	Asura_AnimationAttachment_ClientContainerInstance(const Asura_AnimationAttachment_ClientContainerInstance* pxOther);

	virtual ~Asura_AnimationAttachment_ClientContainerInstance();

	virtual Asura_AnimationAttachment* Clone(Asura_Animation* pxParent);
	virtual void Update(float fTime, bool bIgnoreTimerFrameStamp);
		
	virtual void OnParentVisibilityChanged( bool bInvisible );

	// Change container state.
	void ChangeState(const Asura_Container_InputState& xInputState);

	bool IsOff() const { return m_xInputState.IsOff(); }
	const Asura_Container_InputState& GetInputState() const	{ return m_xInputState; }
	Asura_Hash_ID GetContainerHash() const	{ return m_uContainerHash; }
	Asura_Guid GetOwnerGuid() const			{ return m_uOwnerGuid; }

	// Accessors for the client container instance.
	class Asura_ClientContainerInstance* GetClientContainerInstance() { return m_pxClientInstance; }
	const class Asura_ClientContainerInstance* GetClientContainerInstance() const { return m_pxClientInstance; }

	virtual bool LOSTest(const Asura_Vector_3& xPosition,const Asura_Matrix_3x3 xOrient,const class Asura_LOS_InputData* pxInput,class Asura_LOS_OutputData* pxOutput) const;

protected:
	virtual void Render(const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, unsigned uRenderFlags);
	virtual void RenderShadow( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) {}
	virtual void RenderShadow( Asura_Shadow& xShadow, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOrient, const u_int uRenderFlags ) const;

	virtual bool GetTightBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
	virtual bool GetCoarseBoundingSphere(Asura_Vector_3& xCentre, float& fRadius) const;
	virtual bool GetCoarseBoundingSphereIncludingEffects(Asura_Vector_3& xCentre, float& fRadius) const;

	void CreateInstance();
	void DestroyInstance(bool bChangingState);

	void PreStateChange();
	void PostStateChange();

private:
	// Make these functions private.
	Asura_AnimationAttachment_ClientContainerInstance(const Asura_AnimationAttachment_ClientContainerInstance&);
	Asura_AnimationAttachment_ClientContainerInstance& operator= (const Asura_AnimationAttachment_ClientContainerInstance&);

protected:
	class Asura_ClientContainerInstance* m_pxClientInstance;
	Asura_Hash_ID m_uContainerHash;
	Asura_Container_InputState m_xInputState;

	Asura_Guid m_uOwnerGuid;

	// Flags
	union 
	{
		u_int m_uFlags;
		struct
		{
			u_int	m_bTeleportInitialised			: 1,
					m_bShouldCallRender				: 1,
					m_bShouldCallRenderShadow		: 1,
					m_bShouldCallUpdate				: 1,
					m_bParentInvisible				: 1,
					m_bEnableLOSTest				: 1;
		};
	};
};
#endif //ASURA_USE_CONTAINER_SYSTEM


#endif //__ASURA_ANIMATIONATTACHMENT_H__


/// End of file. ///
