
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Animation_Blend.h
// Author : Andy Tate
// Date   : 21-Sep-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_ANIMATION_BLEND_H__
#define __ASURA_ANIMATION_BLEND_H__

////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationController;
class Asura_AnimationController_Group;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationBlend
// Author       : Andy Tate
// Date         : 21-Sep-2006
//
// Description  : Abstract class to define the interface used when blending
//				  between 2 animation controllers.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationBlend
{
public:
	// Public functions...
	Asura_AnimationBlend(const float fWeight, const bool bSyncToNext)	: m_pxNextController(NULL), m_pxPreviousController(NULL), m_fWeight(fWeight), m_uBlendFlags(0) { m_bSyncToNext = bSyncToNext; }
	Asura_AnimationBlend(const float fWeight)							: m_pxNextController(NULL), m_pxPreviousController(NULL), m_fWeight(fWeight), m_uBlendFlags(0) {}
	virtual ~Asura_AnimationBlend ();

	// Virtual update function...
	virtual	void	Update	(const float fTimeStep, Asura_AnimationController& xPreviousController);

	// Accessor functions...
	virtual bool				IsTween					(void) const										{ return false; }

	void						SetNextController		(Asura_AnimationController* pxNextController)		{ m_pxNextController = pxNextController; }
	Asura_AnimationController*	GetNextController		(void) const										{ return(m_pxNextController); }

	void						SetPreviousController	(Asura_AnimationController* pxPreviousController)	{ m_pxPreviousController = pxPreviousController; }
	Asura_AnimationController*	GetPreviousController	(void) const										{ return(m_pxPreviousController); }

	void						SetWeight				(const float fWeight);
	float						GetWeight				(void) const										{ return(m_fWeight); }

	void						SetDeltasUseBlends		(const bool bSet)									{ m_bDeltasUseBlends = bSet; }
	bool						GetDeltasUseBlends		(void) const										{ return(m_bDeltasUseBlends); }

	void						SetSlerpVelocity		(const bool bSet)									{ m_bSlerpVelocity = bSet; }
	bool						GetSlerpVelocity		(void) const										{ return m_bSlerpVelocity; }

	virtual Asura_AnimationBlend* Duplicate() const;

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

protected:
	// Protected variables...
	Asura_AnimationController*	m_pxNextController;		// Link to the next animation controller in the list.
	Asura_AnimationController*	m_pxPreviousController;	// Link to the previous animation controller in the list.

	float						m_fWeight;				// Blend weight to be applied between this controller and the next one in the list.

	union
	{
		u_int					m_uBlendFlags;
		struct  
		{
			u_int						m_bSyncToNext		: 1,	// Flag to say whether we wish to sync our blended anims together.
										m_bDeltasUseBlends	: 1,	// Should we blend delta animations or add them as is the default.
										m_bSlerpVelocity	: 1,	// When blending velocity use a slerp on the direction and lerp the magnitude.
															: 0;
		};
	};
	// Protected functions...


private:
	// Private variables...

	// Private functions...
	Asura_AnimationBlend  (const Asura_AnimationBlend& xOther);

	Asura_AnimationBlend&	operator=	(const Asura_AnimationBlend& xOther);

};   // End of class 'Asura_AnimationBlend'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationBlend_Tween
// Author       : Andy Tate
// Date         : 21-Sep-2006
//
// Description  : The standard tween from one animation into another. NOTE: The
//				  destination animation doesn't begin playing until the tween
//				  has been completed.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationBlend_Tween : public Asura_AnimationBlend
{
public:
	// Public functions...
	Asura_AnimationBlend_Tween          (const float fTweenInDuration, const bool bSyncToNext);
	Asura_AnimationBlend_Tween          (const float fTweenInDuration, const float fTweenOutDuration, const bool bSyncToNext);
//	virtual ~Asura_AnimationBlend_Tween ();

	// Overridden update function...
	virtual	void	Update	(const float fTimeStep, Asura_AnimationController& xPreviousController);

	virtual Asura_AnimationBlend* Duplicate() const;

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

	void ChangeTweenDuration(const float fTweenInDuration, const float fTweenOutDuration = 0.0f);
	void SetPauseOnStateIn( bool bSet ) { m_bPauseOnStateIn = bSet; }
	void SetPlayOnStateIn( bool bSet ) { m_bPlayOnStateIn = bSet; }
	
protected:
	// Protected enumerated types...
	enum State
	{
		STATE_OUT,
		STATE_TWEEN_IN,
		STATE_IN,
		STATE_TWEEN_OUT,

		MAX_STATES
	};

	// Protected variables...
	union
	{
		u_int		m_uTweenFlags;				// Single access point for flags.

		struct
		{
			u_int	m_eState			: 8,	// The current state of the blend.
					m_eNewState			: 8,	// The state to change to.
					m_bHasNewState		: 1,	// Flag to say that we have a new state to change to.
					m_bPauseOnStateIn	: 1,	// Forces anim to come out of tween in paused
					m_bPlayOnStateIn	: 1,	// Forces anim to come out of tween in playing
										: 0;	// 12 spare-bits
		};
	};

	float	m_fTimer;				// The current time within the tween.
	float	m_fTweenInDuration;		// The duration of the tween-in.
	float	m_fTweenOutDuration;	// The duration of the tween-out.

	// Protected functions...

	//Updates the overall weight from the tweening weight
	virtual void UpdateWeightFromTween(float fTweenWeight){m_fWeight = fTweenWeight;}

	// Virtual state functions...
	virtual bool	IsTween			(void) const { return true; }
	virtual bool	ChangeState		(const State eNewState);
	virtual	bool	BeginNewState	(const State eNewState, Asura_AnimationController& xPreviousController);


private:
	// Private variables...

	// Private functions...
	Asura_AnimationBlend_Tween  (void);
	Asura_AnimationBlend_Tween  (const Asura_AnimationBlend_Tween& xOther);

	Asura_AnimationBlend_Tween&	operator=	(const Asura_AnimationBlend_Tween& xOther);

};   // End of class 'Asura_AnimationBlend_Tween'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationBlend_AnimatedTween
// Author       : Andy Tate
// Date         : 21-Sep-2006
//
// Description  : The animated tween from one animation into another. NOTE: The
//				  destination animation begins playing immediately.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationBlend_AnimatedTween : public Asura_AnimationBlend_Tween
{
public:
	// Public functions...
	Asura_AnimationBlend_AnimatedTween          (const float fTweenInDuration, const bool bSyncToNext);
	Asura_AnimationBlend_AnimatedTween          (const float fTweenInDuration, const float fTweenOutDuration, const bool bSyncToNext);
//	virtual ~Asura_AnimationBlend_AnimatedTween ();

	// Overridden update function...
	virtual	void	Update	(const float fTimeStep, Asura_AnimationController& xPreviousController);

	virtual Asura_AnimationBlend* Duplicate() const;

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

protected:
	// Protected variables...

	// Protected functions...
	

	// Overridden state functions...
//	virtual bool	ChangeState		(const State eNewState);
	virtual	bool	BeginNewState	(const State eNewState, Asura_AnimationController& xPreviousController);


private:
	// Private variables...

	// Private functions...
	Asura_AnimationBlend_AnimatedTween  (void);
	Asura_AnimationBlend_AnimatedTween  (const Asura_AnimationBlend_AnimatedTween& xOther);

	Asura_AnimationBlend_AnimatedTween&	operator=	(const Asura_AnimationBlend_AnimatedTween& xOther);
};   // End of class 'Asura_AnimationBlend_AnimatedTween'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationBlend_TweenOut
// Author       : Andy Tate
// Date         : 22-Sep-2006
//
// Description  : This is used primarily for secondary animations using the old
//				  animation system - when the animation finishes playing it is
//				  tweened out over the specified time.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationBlend_TweenOut : public Asura_AnimationBlend
{
public:
	// Public functions...
	Asura_AnimationBlend_TweenOut          (const float fDuration)	: Asura_AnimationBlend(1.0f), m_fTimer(0.0f), m_fDuration(fDuration), m_fInitialWeight(-1.0f)	{}
//	virtual ~Asura_AnimationBlend_TweenOut ();

	// Overridden update function...
	virtual	void	Update	(const float fTimeStep, Asura_AnimationController& xPreviousController);

	virtual Asura_AnimationBlend* Duplicate() const;

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

protected:
	// Protected variables...


	// Protected functions...
	float	m_fTimer;			// The current time within the tween.
	float	m_fDuration;		// The duration of the tween.
	float	m_fInitialWeight;	// The weight before we start tweening-out.


private:
	// Private variables...


	// Private functions...

	Asura_AnimationBlend_TweenOut  (void);
	Asura_AnimationBlend_TweenOut  (const Asura_AnimationBlend_TweenOut& xOther);

	Asura_AnimationBlend_TweenOut&	operator=	(const Asura_AnimationBlend_TweenOut& xOther);

};   // End of class 'Asura_AnimationBlend_TweenOut'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationBlend_Secondary
// Author       : Richard Rice
// Date         : 25-04-Apr
//
// Description  : A blend for simulating the old secondary animations.
//				  This does an animated tween in/out to the max weighting for the secondary
//                Unlike the animated tween it won't remove previous animations once it has finished
//                tweening in,
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationBlend_Secondary : public Asura_AnimationBlend_AnimatedTween
{
public:
	// Public functions...
	Asura_AnimationBlend_Secondary          (const float fWeight, const float fTweenInDuration, const float fTweenOutDuration);

	virtual Asura_AnimationBlend* Duplicate() const;

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

protected:
	// Protected variables...

	// Protected functions...
	virtual void UpdateWeightFromTween(float fTweenWeight){m_fWeight = fTweenWeight*m_fSecondaryWeight;}

	// Overridden state functions...
	virtual	bool	BeginNewState	(const State eNewState, Asura_AnimationController& xPreviousController);


private:
	// Private variables...

	// Private functions...
	Asura_AnimationBlend_Secondary  (void);
	Asura_AnimationBlend_Secondary  (const Asura_AnimationBlend_AnimatedTween& xOther);

	Asura_AnimationBlend_Secondary&	operator=	(const Asura_AnimationBlend_AnimatedTween& xOther);

	//The weighting to use when fully tweened in.
	float m_fSecondaryWeight;

};   // End of class 'Asura_AnimationBlend_AnimatedTween'.




#endif   // ifndef __ASURA_ANIMATION_BLEND_H__


/// End of file. ///
