
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_AnimationController_Animation.h
// Author : Andy Tate
// Date   : 14-Sep-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_ANIMATIONCONTROLLER_INSTANCE_H__
#define __ASURA_ANIMATIONCONTROLLER_INSTANCE_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

// this file's header
#include "Animation/Asura_AnimationController.h"

// asura
#include "Core/Asura_Maths.h"
#include "Animation/Asura_AnimationBlend.h"
#include "Animation/Asura_Hierarchy.h"

////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Cached_KeyFrame;
class Asura_Hierarchy_Bone_RestFrame;


////////////////////////////////////////////////////////////////////////////////
// Defines...
////////////////////////////////////////////////////////////////////////////////

#define AACA_FLAG_LOOPED						0x00010000
#define AACA_FLAG_REVERSED						0x00020000
#define AACA_FLAG_MIRRORED						0x00040000
#define AACA_FLAG_USING_SUBSET_OF_BONES			0x00080000

#define AACA_FLAG_DONT_TWEEN_ROOT_POSITION		0x00100000
#define AACA_FLAG_DONT_TWEEN_ROOT_ORIENTATION	0x00200000

#define AACA_FLAG_CALCULATE_VELOCITY			0x00400000
#define AACA_FLAG_CALCULATE_ANGULAR_VELOCITY	0x00800000
#define AACA_FLAG_CALCULATE_MOVEMENT_OFFSET		0x01000000

#define AACA_FLAG_DONT_TWEEN_VELOCITY			0x02000000
#define AACA_FLAG_DONT_TWEEN_ANGULAR_VELOCITY	0x04000000
#define AACA_FLAG_DONT_TWEEN_MOVEMENT_OFFSET	0x08000000


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationController_Animation
// Author       : Andy Tate
// Date         : 14-Sep-2006
//
// Description  : This class is the interface to the animation data itself, it
//				  provides the ability to play a single animation without needing
//				  to worry about blending to/from other animations. Any blending
//				  is taken care of separately by the "Blend()" function which is
//				  called from the group containing this instance. If no blending
//				  is required then the "Fill()" function is called instead to 
//				  ensure that the tween-store is updated correctly.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationController_Animation : public Asura_AnimationController
{
public:
	// Public functions...
	Asura_AnimationController_Animation          (const Asura_Hash_ID uControllerHash, const u_int uInitialFlags, const u_int uPriority, const Asura_Hash_ID uAnimationHash);
	virtual ~Asura_AnimationController_Animation ();

	// Animation manipulation functions...
	bool	Play	(void)	{ return(ChangeState(PLAYING)); }
	bool	Pause	(void)	{ return(ChangeState(PAUSED)); }
	bool	Stop	(void)	{ return(ChangeState(FINISHED)); }

	// Function to play any sounds which are needed since the last frame...
	void	PlaySoundsForLastFrame	(void) const;

	// Overridden function to get the interval flags which were appropriate during the last frame...
	virtual int		GetFlagsForLastFrame	(void) const;

	// Overridden bone-update function...
	virtual	u_int	TweenBoneData	(const Asura_Hierarchy_Skin& xSkin, const u_int uFirstBone, const u_int uLastBone, const float fWeight, const bool bDeltasUseBlends, Asura_AnimationController_Group* pxParentGroup);

	// Overridden bounding-box functions...
	virtual	u_int	TweenBoundingBox	(Asura_Bounding_Box& xBB, const Asura_Hierarchy_Skin& xSkin, const float fWeight, const bool bDeltasUseBlends, const bool bSlerpVelocity, Asura_AnimationController_Group* pxParentGroup);

	// Overridden message handling functions...
	virtual	bool	HandleMessage	(const u_int uType, void* pvUserData = NULL);

	// Overridden timer update functions...
	virtual	void	UpdateTimers	(const float fTimeStep, bool bIgnoreTimerFrameStamp);

	// Overridden accessor functions...
	virtual	bool	AreAllBonesUsingDeltas	(void) const	{ return(m_bIsDeltaAnimation); }
	virtual	bool	IsUsingSubsetOfBones	(void) const	{ return((m_uControllerFlags & AACA_FLAG_USING_SUBSET_OF_BONES) != 0); }
	virtual	bool	IsFinished				(void) const	{ return(m_eCurrentState == FINISHED); }
	virtual bool	IsReadyForDestruction	(void) const	{ return(m_bReadyForDestruction); }
	virtual	float	GetTimeUntilFinished	(void) const	{ return(m_fTimeUntilFinished); }

	// Accessor functions...
	bool	IsPlaying				(void) const			{ return(m_eCurrentState == PLAYING); }
	bool	IsPaused				(void) const			{ return(m_eCurrentState == PAUSED); }
	bool	IsFinishing				(void) const			{ return(m_eNewState == FINISHED); }
	bool	IsTweening				(void) const			{ return(m_pxBlendToNextController ? m_pxBlendToNextController->IsTween() : false); }

	void	SetLooped				(const bool bLooped)	{ if(bLooped) m_uControllerFlags |= AACA_FLAG_LOOPED; else m_uControllerFlags &= ~AACA_FLAG_LOOPED; }
	bool	IsLooped				(void) const			{ return((m_uControllerFlags & AACA_FLAG_LOOPED) != 0); }

	void	SetMirrored				(const bool bMirrored)	{ if(bMirrored) m_uControllerFlags |= AACA_FLAG_MIRRORED; else m_uControllerFlags &= ~AACA_FLAG_MIRRORED; }
	bool	IsMirrored				(void) const			{ return((m_uControllerFlags & AACA_FLAG_MIRRORED) != 0); }

	void	SetReversed				(const bool bReversed)	{ if(bReversed) m_uControllerFlags |= AACA_FLAG_REVERSED; else m_uControllerFlags &= ~AACA_FLAG_REVERSED; }
	bool	IsReversed				(void) const			{ return((m_uControllerFlags & AACA_FLAG_REVERSED) != 0); }

	void	SetTimeDelay			(const float fTimeDelay);
	void	SetDuration				(const float fDuration);
	void	SetDurationFromSpeed	(const Asura_Hierarchy_Skin* pxSkin,const float fSpeed);
	//The start and end time appear to be more along the lines of start and end loop points. RAR 15/08/2007
	void	SetParametricStartTime	(const float fTime);
	void	SetParametricEndTime	(const float fTime);
	void	SetParametricTimer		(const float fTime, bool bPlayToTime = true);

	void	SetEmptyAnimationAsDelta(void);

	//Sets a time to animate to. The fractional part needs to be between the parametric start and end times , and the
	//overall value should be higher than the current timer.
	//If using this then the loop flag is irrevant
	//RAR 15/08/2007
	void 	SetParametricTargetTime(const float fTime);
	void	ClearParametricTargetTime() {m_bUseParametricTargetTime = false;}


	float	GetDuration				(void) const	{ return(m_fDuration); }
	float	GetTimeDelay			(void) const	{ return(m_fTimeDelay); }

	float	GetParametricStartTime	(void) const	{ return(m_fParametricStartTime); }
	float	GetParametricEndTime	(void) const	{ return(m_fParametricEndTime); }
	float	GetParametricTimer		(void) const	{ return(m_fParametricTimer); }

	bool GetMovementOffset( Asura_Vector_3* pxOutputPos ) const;
	bool GetMovementOffset( Asura_Vector_3* pxOutputPos, Asura_Quat* pxOutputQuat ) const;

	Asura_Hash_ID			GetAnimationHash	(void) const	{ return(m_uAnimationHash); }
	Asura_Hierarchy_Anim*	GetAnimation		(void) const	{ return(m_pxAnimation); }

	virtual Asura_AnimationController* Duplicate(Asura_Animation* pxDestAnimation) const;

	bool HasLoopedThisFrame() const
	{
		if(!IsLooped()) { return false; }
		return IsReversed() ? (m_fOldAnimPlayed < m_fAnimPlayed) : (m_fOldAnimPlayed > m_fAnimPlayed);
	}

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

protected:
	// Protected enumerated types...
	enum State
	{
		INVALID	= 0,

		PAUSED,
		PLAYING,
		FINISHED,

		MAX_STATES
	};

	// Protected variables...
	Asura_Hash_ID			m_uAnimationHash;		// The hash-ID of the animation we are to play.
	Asura_Hierarchy_Anim*	m_pxAnimation;			// A pointer to the animation we're currently playing.

	float					m_fDuration;			// The duration of the animation playing - this may be modified.
	float					m_fOneOverDuration;		// Useful when calculating parametric times.
	float					m_fTimeDelay;			// The delay before starting playing the animation.

	float					m_fParametricStartTime;	// The time to start playing the animation at.
	float					m_fParametricEndTime;	// The time to end playing the animation at.
	float					m_fParametricTimer;		// The time within the animation.
	float					m_fParametricTargetTime;//The time to animate to. If > 1 then it will loop however many times to reach the target

	float					m_fAnimPlayed;			// The amount of the animation we've currently played (from 0.0f to 1.0f inclusive).
	float					m_fOldAnimPlayed;		// The previous amount of the animation we've played.
	float					m_fTimeUntilFinished;	// The number of seconds until this animation has finished playing.

#if defined(ASURA_USE_KEYFRAME_CACHING)
	Asura_Cached_KeyFrame*	m_pxCachedKeyFrames;	// A pointer to the array cached key-frame data on a per-bone basis.
#endif	// defined(ASURA_USE_KEYFRAME_CACHING)


	// Protected functions...
	bool	InitialiseAnimation			(void);
	void	UpdateAnimPlayed			(void);

	bool	ChangeState					(const State eState);

	bool	CalculateBoundingBox		(const Asura_Hierarchy_Skin& xSkin, const Asura_Hierarchy_Tween& xTweenStore, Asura_Bounding_Box& xBB) const;
	bool	CalculateBoneRelativeData	(const Asura_Hierarchy_Skin& xSkin, const u_int uBone, Asura_Hierarchy_Bone_RestFrame& xOutput) const;

	float GetAdjustedAnimPlayed() const;

	// Static functions...
	static	bool	IntervalIntersects	(const float fStart1, const float fEnd1, const float fStart2, const float fEnd2);


private:
	// Private variables...
	union
	{
		u_int		m_uControllerAnimationFlags;		// Single-access point for the flags.

		struct 
		{
			u_int	m_eCurrentState				: 8,	// The current state of the animation we are to play.
					m_eNewState					: 8,	// The state of the animation we are to change to.
					m_bUseParametricTargetTime	: 1,
					m_bIsDeltaAnimation			: 1,	// Flag to say whether this animation is a delta or not.
					m_bWillBeReadyForDestruction: 1,
					m_bReadyForDestruction		: 1,
												: 0;	// 12 spare-bits.
		};
	};


	// Private functions...

	Asura_AnimationController_Animation  (void);
	Asura_AnimationController_Animation  (const Asura_AnimationController_Animation& xOther);

	Asura_AnimationController_Animation&	operator=	(const Asura_AnimationController_Animation& xOther);

};   // End of class 'Asura_AnimationController_Animation'.


#endif   // ifndef __ASURA_ANIMATIONCONTROLLER_INSTANCE_H__


/// End of file. ///
