
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_AnimationController_Group.cpp
// Author : Andy Tate
// Date   : 14-Sep-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Animation/Asura_AnimationController_Group.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationBlend.h"
#include "Animation/Asura_Hierarchy.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"

// Asura platform header-files...

// Game independent header-files...

// Game platform header-files...

// Tools header-files...


////////////////////////////////////////////////////////////////////////////////
// Defines...
////////////////////////////////////////////////////////////////////////////////

#if 1
#define LOG_OUTPUT(s)
#else
#define LOG_OUTPUT(s)	Asura_OutputToDebugger_Retail(s)
#endif

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR u_int	dbg_uGroupNumIndentationChars	= 2;

#if (defined ASURA_LOGFILES_ENABLED && defined ASURA_TOOLS)
static bool s_bBlendErrorLogClear = false;
static const char* dbg_szBlendErrorLogFilename = "logfiles\\AnimBlendErrorLog.txt";
#endif

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::Asura_AnimationController_Group
// Author       : Andy Tate
// Date         : 15-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Group::Asura_AnimationController_Group (const Asura_Hash_ID uControllerHash,
																  const u_int uInitialFlags,
																  const u_int uPriority /*= 0x00004000*/,
																  Asura_AnimationController_Group* pxParentGroup /*= NULL*/)
: Asura_AnimationController(uControllerHash, uInitialFlags, uPriority)
, m_uGroupFlags(0)
, m_pxFirstChildController(NULL)
, m_pxLastChildController(NULL)
, m_uFrameNumberStamp(0)
{
	// if we don't have the inherit tween store flag make a new tween store
	if ( ( uInitialFlags & AACG_FLAG_INHERIT_PARENT_TWEEN_STORE ) == 0 )
	{
		m_bNeedsToDeleteTweenStore = true;
		m_pxTweenStore = new Asura_Hierarchy_Tween;
	}
	else
	{
		Asura_Assert( pxParentGroup, "No parent group specified for group with AACG_FLAG_INHERIT_PARENT_TWEEN_STORE set" );
		if ( pxParentGroup )
		{
			m_pxTweenStore = &(pxParentGroup->GetTweenStore());
		}
	}
	// Nothing more to initialise...
}

/// End of function 'Asura_AnimationController_Group::Asura_AnimationController_Group'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::~Asura_AnimationController_Group
// Author       : Andy Tate
// Date         : 15-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Group::~Asura_AnimationController_Group ()
{
	// Remove from the correct hash-tree...
	m_pxAsuraAnimation->m_xControllerGroupContainer.DeleteNodeByHash(m_uControllerHash);

	if ( m_bNeedsToDeleteTweenStore )
	{
		delete m_pxTweenStore;
	}

	// Destroy any children we may have...
	DestroyAllChildControllers();
}

/// End of function 'Asura_AnimationController_Group::~Asura_AnimationController_Group'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::AddChildController
// Author       : Andy Tate
// Date         : 20-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::AddChildController (Asura_AnimationController& xController,
														  Asura_AnimationBlend* pxBlend)
{
	// Does our controller already have a blend?
	if(xController.GetBlendToNextController() == NULL)
	{
		xController.SetParentControllerHash( GetControllerHash() );

		// Do we actually have a blend?
		if(pxBlend == NULL)
		{
			Asura_AnimationController*	pxCurrentController	= m_pxFirstChildController;

			// No - then remove ALL controllers...
			while(pxCurrentController != NULL)
			{
				Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
				Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

				m_pxAsuraAnimation->DestroyController(pxCurrentController->GetControllerHash());

				pxCurrentController	= pxNextController;
			}

			m_pxFirstChildController	= &xController;
			m_pxLastChildController		= &xController;

			// Increase the count and we're done...
			m_uNumChildControllers++;
			return(true);
		}
		else
		{
			const u_int					uPriority			= xController.GetPriority();
			Asura_AnimationController*	pxCurrentController	= m_pxLastChildController;
			Asura_AnimationBlend*		pxNextBlend			= NULL;

			// Do we have any controllers?
			if(pxCurrentController == NULL)
			{
				// No - create this as the first one...
				m_pxFirstChildController	= &xController;
				m_pxLastChildController		= &xController;

				// We have no need for a blend - remove it...
				delete(pxBlend);

				// Increase the count and we're done...
				m_uNumChildControllers++;
				return(true);
			}
			else
			{
				// Yes - start at the end of the list...
				do
				{
					const u_int	uCurrentPriority	= pxCurrentController->GetPriority();

					// Is it time to add yet?
					if(uPriority <= uCurrentPriority)
					{
						// Yes - add this controller AFTER the current one...
						pxCurrentController->SetBlendToNextController(pxBlend);
						pxBlend->SetNextController(&xController);

						xController.SetBlendToPreviousController(pxBlend);
						pxBlend->SetPreviousController(pxCurrentController);

						if(pxNextBlend == NULL)
						{
							m_pxLastChildController	= &xController;
						}
						else
						{
							xController.SetBlendToNextController(pxNextBlend);
							pxNextBlend->SetPreviousController(&xController);
						}

						// Increase the count and we're done...
						m_uNumChildControllers++;
						return(true);
					}
					else
					{
						Asura_AnimationBlend*		pxPreviousBlend			= pxCurrentController->GetBlendToPreviousController();
						Asura_AnimationController*	pxPreviousController	= (pxPreviousBlend == NULL) ? NULL : pxPreviousBlend->GetPreviousController();

						// No - check the previous controller...
						pxNextBlend			= pxPreviousBlend;
						//pxNextController	= pxCurrentController;
						pxCurrentController	= pxPreviousController;
					}

				} while(pxCurrentController != NULL);

				// This must be the highest priority controller we've got...
				m_pxFirstChildController->SetBlendToPreviousController(pxBlend);
				pxBlend->SetPreviousController(&xController);

				xController.SetBlendToNextController(pxBlend);
				pxBlend->SetNextController(m_pxFirstChildController);

				m_pxFirstChildController	= &xController;

				// Increase the count and we're done...
				m_uNumChildControllers++;
				return(true);
			}
		}
	}
	else
	{
		Asura_Assert(false, "Controller cannot be added to this group because its already linked to something else.");
		return(false);
	}
}

/// End of function 'Asura_AnimationController_Group::AddChildController'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::DestroyChildController
// Author       : Andy Tate
// Date         : 20-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::DestroyChildController (Asura_AnimationController* pxController)
{
	Asura_AnimationBlend*		pxPreviousBlend			= NULL;
	Asura_AnimationController*	pxPreviousController	= NULL;
	Asura_AnimationController*	pxCurrentController		= m_pxFirstChildController;

	// Check each child in-turn...
	while(pxCurrentController != NULL)
	{
		Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
		Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

		// Is this our controller?
		if(pxCurrentController == pxController)
		{
			// Yes - is this the first controller?
			m_uNumChildControllers--;

			if(pxPreviousController == NULL)
			{
				// Yes - make the second controller the head of the list...
				m_pxFirstChildController	= pxNextController;

				if(pxNextController == NULL)
				{
					m_pxLastChildController	= pxPreviousController;
				}
				else
				{
					pxNextBlend->SetNextController(NULL);
					pxNextController->SetBlendToPreviousController(pxPreviousBlend);
				}

				delete(pxController);
				return(true);
			}
			else
			{
				// No - unlink the previous blend and the specified controller from the list...
				pxPreviousController->SetBlendToNextController(pxNextBlend);

				if(pxNextBlend == NULL)
				{
					m_pxLastChildController	= pxPreviousController;
				}
				else
				{
					pxNextBlend->SetPreviousController(pxPreviousController);
				}

				pxPreviousBlend->SetPreviousController(NULL);
				pxCurrentController->SetBlendToNextController(NULL);

				delete(pxPreviousBlend);
				return(true);
			}
		}
		else
		{
			// No - try the next child...
			pxPreviousBlend			= pxNextBlend;
			pxPreviousController	= pxCurrentController;
			pxCurrentController		= pxNextController;
		}
	}

	// We failed to find our controller...
	return(false);
}

/// End of function 'Asura_AnimationController_Group::DestroyChildController'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::DestroyAllChildControllers
// Author       : Andy Tate
// Date         : 11-Apr-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::DestroyAllChildControllers(void)
{
	const bool	bDeltasUseBlends	= m_bDeltasUseBlends;

	// Remove the controllers (and blends)...
	delete(m_pxFirstChildController);

	// Reset the members...
	m_pxFirstChildController	= NULL;
	m_pxLastChildController		= NULL;

	m_uGroupFlags		= 0;
	m_bDeltasUseBlends	= bDeltasUseBlends;

	return(true);
}

/// End of function 'Asura_AnimationController_Group::DestroyAllChildControllers'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::DestroyChildControllers
// Author       : Richard May
// Date         : 3-Jul-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::DestroyChildControllers(bool(*pfnShouldDeleteCallback)(const Asura_AnimationController* pxChild))
{
	Asura_AnimationController* pxCurrentController = m_pxFirstChildController;
	
	while( pxCurrentController )
	{	
		Asura_AnimationController* pxTempController = pxCurrentController;
		pxCurrentController = pxCurrentController->GetBlendToNextController() ? pxCurrentController->GetBlendToNextController()->GetNextController() : NULL;
		
		if( pfnShouldDeleteCallback(pxTempController) )
		{
			DestroyChildController( pxTempController );
		}
	}

	return (true);
}

/// End of function 'Asura_AnimationController_Group::DestroyChildControllers'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::CloneTweenStore
// Author       : Andy Tate
// Date         : 9-Nov-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController_Group::CloneTweenStore (const Asura_AnimationController_Group &xGroup)
{
	// Just copy the tween-store...
	m_pxTweenStore->Copy(&(xGroup.GetTweenStore()));
}

/// End of function 'Asura_AnimationController_Group::CloneTweenStore'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::Deactivate
// Author       : Andy Tate
// Date         : 29-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController_Group::Deactivate (void)
{
	if((m_uControllerFlags & AACG_FLAG_PRESERVE_TWEEN_STORE) == 0)
	{
		m_pxTweenStore->DeleteTweenStore();
	}
	Asura_AnimationController::Deactivate();
}

/// End of function 'Asura_AnimationController_Group::Deactivate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::TweenBoundingBox
// Author       : Andy Tate
// Date         : 21-Sep-2006
////////////////////////////////////////////////////////////////////////////////

u_int Asura_AnimationController_Group::TweenBoundingBox(Asura_Bounding_Box& xBB,
														const Asura_Hierarchy_Skin& xSkin,
														const float fWeight,
														const bool bDeltasUseBlends,
														const bool bSlerpVelocity,
														Asura_AnimationController_Group* pxParentGroup)
{
	const u_int	uNumberOfBones	= xSkin.NumberOfBones;
	u_int		uNumUpdated		= 0;

	// Initialise our tween-stores...
	m_pxTweenStore->SetNumberOfBones(uNumberOfBones);

	// Do we have a child controller to use?
	if(m_pxFirstChildController != NULL)
	{
		float						fPreviousWeight				= 1.0f;
		bool						bPreviousDeltasUseBlends	= false;
		bool						bPreviousSlerpVelocity		= false;
		Asura_AnimationController*	pxCurrentController	= m_pxFirstChildController;

		// Loop, blending in the data from each subsequent controller...
		while(pxCurrentController != NULL)
		{
			Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
			Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

			// Do we need to bother with this controller?
			if(pxCurrentController->IsActive())
			{
				u_int	uNumChildrenUpdated;

				// Yes - apply the data to the current tween-store...
				if(fPreviousWeight > 0.0f)
				{
					const bool	bChildDeltasUseBlends	= bPreviousDeltasUseBlends || m_bDeltasUseBlends;

					uNumChildrenUpdated	= pxCurrentController->TweenBoundingBox(xBB, xSkin, fPreviousWeight, bChildDeltasUseBlends, bPreviousSlerpVelocity, this);
				}
				else
				{
					uNumChildrenUpdated	= 0;
				}

				// Update the pointers...
				uNumUpdated	+= uNumChildrenUpdated;
			}

			// Advance the pointers...
			if(pxNextBlend != NULL)
			{
				fPreviousWeight	= pxNextBlend->GetWeight();
				bPreviousDeltasUseBlends = pxNextBlend->GetDeltasUseBlends();
				bPreviousSlerpVelocity = pxNextBlend->GetSlerpVelocity();
			}

			pxCurrentController	= pxNextController;
		}
	}

	// Did we generate any data?
	if(pxParentGroup == NULL)
	{
#if (defined ASURA_LOGFILES_ENABLED && defined ASURA_TOOLS)
		if ( m_pxTweenStore->IsBlendError() )
		{
			if ( !s_bBlendErrorLogClear )
			{
				Asura_SimpleLogfileClear( dbg_szBlendErrorLogFilename );
				s_bBlendErrorLogClear = true;
			}
			Asura_Assertf( false, ( "There were errors in the anim blends for entity %d, an error file has been generated here: %s\nPlease include this file in the Mantis report.", m_pxAsuraAnimation ? m_pxAsuraAnimation->GetParentEntityGUID() : ASURA_GUID_UNREGISTERED, dbg_szBlendErrorLogFilename ));
			Asura_SimpleLogfileOutput(( dbg_szBlendErrorLogFilename, "Blend Error in TweenBoundingBox for entity %d, frame %d, for the following blend tree:\n\n", m_pxAsuraAnimation ? m_pxAsuraAnimation->GetParentEntityGUID() : ASURA_GUID_UNREGISTERED, Asura_Timers::GetFrameNumber() ));
			u_int uRow = 0;
			u_int uCol = 0;
			PrintDebugInfo( uRow, uCol, NULL, dbg_szBlendErrorLogFilename );
			Asura_SimpleLogfileOutput(( dbg_szBlendErrorLogFilename, "\n\n\n\n" ));
		}
#endif
	}
	else if(uNumUpdated > 0)
	{
		Asura_Hierarchy_Tween&	xToTweenStore			= pxParentGroup->GetTweenStore();
		const bool				bGroupDeltasUseBlends	= bDeltasUseBlends && (m_pxBlendToPreviousController != NULL);
		const bool				bBlendBoneData			= (m_pxTweenStore->IsBoneDataDelta(0))
														? ((xToTweenStore.IsBoneDataDelta(0)) ? bGroupDeltasUseBlends : false)
														: true;

		if(bBlendBoneData)
		{
			// Yes - blend this into the tween-store we've been given...
			m_pxTweenStore->BlendVelocityIntoTweenStore(fWeight, xToTweenStore, bSlerpVelocity);
			m_pxTweenStore->BlendAngularVelocityIntoTweenStore(fWeight, xToTweenStore);
			m_pxTweenStore->BlendMovementOffsetIntoTweenStore(fWeight, xToTweenStore);
		}
		else
		{
			// Yes - add this into the tween-store we've been given...
			m_pxTweenStore->AddVelocityIntoTweenStore(fWeight, xToTweenStore);
			m_pxTweenStore->AddAngularVelocityIntoTweenStore(fWeight, xToTweenStore);
			m_pxTweenStore->AddMovementOffsetIntoTweenStore(fWeight, xToTweenStore);
		}

		if ( m_pxTweenStore->IsBlendError() )
		{
			xToTweenStore.SetBlendError();
		}

		Asura_Assertf( !xToTweenStore.IsBlendError(), ( "Error blending group 0x%08x into group 0x%08x", GetControllerHash(), pxParentGroup->GetControllerHash() ) );
	}

	// Return the number of controllers updated...
	return(uNumUpdated);
}

/// End of function 'Asura_AnimationController_Group::TweenBoundingBox'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::TweenBoneData
// Author       : Andy Tate
// Date         : 15-Sep-2006
////////////////////////////////////////////////////////////////////////////////

u_int Asura_AnimationController_Group::TweenBoneData(const Asura_Hierarchy_Skin& xSkin,
													 const u_int uFirstBone,
													 const u_int uLastBone,
													 const float fWeight,
													 const bool bDeltasUseBlends,
													 Asura_AnimationController_Group* pxParentGroup)
{
	const u_int	uNumberOfBones	= xSkin.NumberOfBones;
	u_int		uNumUpdated		= 0;

	// Initialise our tween-store...
	m_pxTweenStore->SetNumberOfBones(uNumberOfBones);
	LOG_OUTPUT(("...Group::BlendBoneData(): %08X, fWeight\n", m_uControllerHash, fWeight));

	// Do we have a child controller to use?
	if(m_pxFirstChildController != NULL)
	{
		float						fPreviousWeight				= 1.0f;
		bool						bPreviousDeltasUseBlends	= false;	
		Asura_AnimationController*	pxCurrentController	= m_pxFirstChildController;

		// Loop, blending in the data from each subsequent controller...
		while(pxCurrentController != NULL)
		{
			Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
			Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

			// Do we need to bother with this controller?
			if(pxCurrentController->IsActive())
			{
				u_int	uNumChildrenUpdated;

				// Yes - apply the data to the current tween-store...
				if(fPreviousWeight > 0.0f)
				{
					uNumChildrenUpdated	= pxCurrentController->TweenBoneData(xSkin, uFirstBone, uLastBone, fPreviousWeight, bPreviousDeltasUseBlends, this);
				}
				else
				{
					uNumChildrenUpdated	= 0;
				}

				// Update the pointers...
				uNumUpdated	+= uNumChildrenUpdated;
			}

			// Advance the pointers...
			if(pxNextBlend != NULL)
			{
				fPreviousWeight	= pxNextBlend->GetWeight();
				bPreviousDeltasUseBlends = pxNextBlend->GetDeltasUseBlends();
			}

			pxCurrentController	= pxNextController;
		}
	}

	// Did we generate any data?
	if(pxParentGroup == NULL)
	{
		m_pxTweenStore->FillBoneDataWithRestPose(xSkin, uFirstBone, uLastBone);

#if (defined ASURA_LOGFILES_ENABLED && defined ASURA_TOOLS)
		if ( m_pxTweenStore->IsBlendError() )
		{
			if ( !s_bBlendErrorLogClear )
			{
				Asura_SimpleLogfileClear( dbg_szBlendErrorLogFilename );
				s_bBlendErrorLogClear = true;
			}
			Asura_Assertf( false, ( "There were errors in the anim blends for entity %d, an error file has been generated here: %s\nPlease include this file in the Mantis report.", m_pxAsuraAnimation ? m_pxAsuraAnimation->GetParentEntityGUID() : ASURA_GUID_UNREGISTERED, dbg_szBlendErrorLogFilename ));
			Asura_SimpleLogfileOutput(( dbg_szBlendErrorLogFilename, "Blend Error in TweenBoneData for entity %d, frame %d, for the following blend tree:\n\n", m_pxAsuraAnimation ? m_pxAsuraAnimation->GetParentEntityGUID() : ASURA_GUID_UNREGISTERED, Asura_Timers::GetFrameNumber() ));
			u_int uRow = 0;
			u_int uCol = 0;
			PrintDebugInfo( uRow, uCol, NULL, dbg_szBlendErrorLogFilename );
			Asura_SimpleLogfileOutput(( dbg_szBlendErrorLogFilename, "\n\n\n\n" ));
		}
#endif
	}
	else if(uNumUpdated > 0)
	{
		const bool				bGroupDeltasUseBlends	= bDeltasUseBlends && (m_pxBlendToPreviousController != NULL);
		Asura_Hierarchy_Tween&	xToTweenStore			= pxParentGroup->GetTweenStore();

		for(u_int uBone = uFirstBone; uBone <= uLastBone; uBone++)
		{
			const bool	bBlendBoneData	= (m_pxTweenStore->IsBoneDataDelta(uBone))
										? ((xToTweenStore.IsBoneDataDelta(uBone)) ? bGroupDeltasUseBlends : false)
										: true;

			if(bBlendBoneData)
			{
				m_pxTweenStore->BlendBoneDataIntoTweenStore(xSkin, uBone, fWeight, xToTweenStore);
			}
			else
			{
				m_pxTweenStore->AddBoneDataIntoTweenStore(xSkin, uBone, fWeight, xToTweenStore);
			}
		}

		if ( m_pxTweenStore->IsBlendError() )
		{
			xToTweenStore.SetBlendError();
		}

		Asura_Assertf( !xToTweenStore.IsBlendError(), ( "Error blending group 0x%08x into group 0x%08x", GetControllerHash(), pxParentGroup->GetControllerHash() ) );
	}

	// Return the number of controllers updated...
	return(uNumUpdated);
}

/// End of function 'Asura_AnimationController_Group::TweenBoneData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::HandleMessage
// Author       : Andy Tate
// Date         : 21-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::HandleMessage (const u_int uType,
													 void* pvUserData)
{
	// Sometimes we don't want groups to destroy children with messages sent to them
	if (((m_uControllerFlags & AACG_FLAG_DONT_PASS_DESTROY_TO_CHILDREN) != 0) && (uType == DESTROY_MSG || uType == DESTROY_FROM_TWEEN_IN_MSG))
	{
		return true;
	}
	else
	{
		Asura_AnimationController*	pxCurrentController	= m_pxFirstChildController;
		bool						bHandledMessage		= false;

		// Try each child with this message...
		while(pxCurrentController != NULL)
		{
			Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
			Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

			// Let this child handle the message...
			if(pxCurrentController->HandleMessage(uType, pvUserData))
			{
				bHandledMessage	= true;
			}

			// Try the next controller...
			pxCurrentController	= pxNextController;
		}

		// If something has tweened in over us, we need to pass back to any previous controllers.
		if ((uType == DESTROY_FROM_TWEEN_IN_MSG) && m_pxBlendToPreviousController && m_pxBlendToPreviousController->GetPreviousController())
		{
			m_pxBlendToPreviousController->GetPreviousController()->HandleMessage(uType);
		}

		// Return whether we handled the message...
		return(bHandledMessage);
	}
}

/// End of function 'Asura_AnimationController_Group::HandleMessage'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::UpdateTimers
// Author       : Andy Tate
// Date         : 14-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController_Group::UpdateTimers (const float fTimeStep,bool bIgnoreTimerFrameStamp)
{
	const u_int	uCurrentFrameNumber	= Asura_Timers::GetFrameNumber();

	// Debugging...
	LOG_OUTPUT(("...Group::UpdateTimers(): fTimeStep = %.3fs\n", fTimeStep));

	// JEB - 09 February 2007 15:30:05 - Don't update individual groups more than once per frame
	if(bIgnoreTimerFrameStamp || (m_uFrameNumberStamp != uCurrentFrameNumber))
	{
		m_uFrameNumberStamp = Asura_Timers::GetFrameNumber();

		m_pxTweenStore->Invalidate();

		// Do we have a child controller to use?
		if(m_pxFirstChildController != NULL)
		{
			Asura_AnimationController*	pxCurrentController	= m_pxFirstChildController;

			bool						bIsUsingSubsetOfBones;

			// Reset the flags...
			bIsUsingSubsetOfBones	= true;

			// Loop, blending in the data from each subsequent controller...
			while(pxCurrentController != NULL)
			{
				Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
				Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

				// Recurse to update this controller's timers...
				pxCurrentController->UpdateTimers(fTimeStep, bIgnoreTimerFrameStamp);
				bIsUsingSubsetOfBones	&= pxCurrentController->IsUsingSubsetOfBones();

				// Do we have a custom blend to perform?
				if(pxNextBlend != NULL)
				{
					pxNextBlend->Update(fTimeStep, *pxCurrentController);
				}

				// Do we need to remove the current controller?
				if(pxCurrentController->IsReadyForDestruction())
				{
					// NOTE: We check the previous frame's destruction flag to ensure that we don't
					// destroy the animation data on the frame the animation finishes, rather we 
					// perform the destruction on the next frame.

					// WGP 24/03/2009 - We need to destroy groups on the same frame they say they're ready for destruction
					// ( after updating their children ) otherwise we end up with blend errors from blending empty groups
					// So I've moved the logic to delay the animation controller animation destruction into the animation controller animation class
					m_pxAsuraAnimation->DestroyController(pxCurrentController->GetControllerHash());
				}

				// Advance the pointers...
				pxCurrentController	= pxNextController;
			}

			// Update the flag...
			m_bIsUsingSubsetOfBones	= bIsUsingSubsetOfBones;
		}
	}
}

/// End of function 'Asura_AnimationController_Group::UpdateTimers'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::SetBlendWeights
// Author       : Andy Tate
// Date         : 22-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::SetBlendWeights (const float* pfWeights,
													   const u_int uNumWeights)
{
	// Are the number of weights correct?
	if(uNumWeights == GetNumChildBlends())
	{
		// Do we have ANY weights at all?
		if(uNumWeights == 0)
		{
			return(true);
		}
		else
		{
			Asura_AnimationController*	pxCurrentController	= m_pxFirstChildController;

			while(pxCurrentController != NULL)
			{
				Asura_AnimationBlend*		pxNextBlend			= pxCurrentController->GetBlendToNextController();
				Asura_AnimationController*	pxNextController	= (pxNextBlend == NULL) ? NULL : pxNextBlend->GetNextController();

				// Force this blend weight...
				if(pxNextBlend != NULL)
				{
					pxNextBlend->SetWeight(*pfWeights);
					pfWeights++;
				}

				// Advance to the next blend...
				pxCurrentController	= pxNextController;
			}

			return(true);
		}
	}
	else
	{
		Asura_Assert(false, "Mismatch between blend-weight counts for this group.");
		return(false);
	}
}

/// End of function 'Asura_AnimationController_Group::SetBlendWeights'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::SetControllerWeights
// Author       : Andy Tate
// Date         : 22-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::SetControllerWeights (const float* pfWeights,
															const u_int uNumControllers)
{
	// Are the number of weights correct?
	if(uNumControllers == GetNumChildControllers())
	{
		// Do we have ANY weights at all?
		if(uNumControllers < 2)
		{
			return(true);
		}
		else
		{
			Asura_AnimationController*	pxCurrentController	= m_pxLastChildController;
			const float*				pfCurrentWeight		= &pfWeights[uNumControllers - 1];

#if defined(ASURA_ASSERT)
			{
				float	fTotal	= 0.0f;

				// Sum-up the weights provided...
				for(u_int u = 0; u < uNumControllers; u++)
				{
					fTotal	+= pfWeights[u];
				}

				Asura_Assert(Asura_Maths::Abs(fTotal - 1.0f) < 0.0001f, "Specified weights do not add up to 1.0f.");
			}
#endif	// defined(ASURA_ASSERT)

			// Set the weight of each blend in-turn...
			float fOneMinusCumulative = 1.0f;
			while(pxCurrentController != NULL)
			{
				Asura_AnimationBlend*		pxPreviousBlend			= pxCurrentController->GetBlendToPreviousController();
				// Once we reach the first controller, the weightings after it will leave it with the weight it should be,
				// so we can just skip out at that point.
				if (!pxPreviousBlend) { break; }

				Asura_AnimationController*	pxPreviousController	= pxPreviousBlend->GetPreviousController();
				float						fCurrentWeight			= (fOneMinusCumulative <= 0.0f) ? 0.0f : (*pfCurrentWeight / fOneMinusCumulative);
				fOneMinusCumulative -= *pfCurrentWeight;

				// Force this blend weight...
				pxPreviousBlend->SetWeight(fCurrentWeight);

				// Advance to the next blend...
				pxCurrentController	= pxPreviousController;
				pfCurrentWeight--;
			}

			return(true);
		}
	}
	else
	{
		Asura_Assert(false, "Mismatch between blend-weight counts for this group.");
		return(false);
	}
}

/// End of function 'Asura_AnimationController_Group::SetControllerWeights'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::GetAngularVelocity
// Author       : Andy Tate
// Date         : 19-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::GetAngularVelocity (Asura_Vector_3& xOutputAxis,
														  Asura_Angle& fOutputAngle) const
{
	return(m_pxTweenStore->GetAngularVelocity(xOutputAxis, fOutputAngle));
}

/// End of function 'Asura_AnimationController_Group::GetAngularVelocity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::GetMovementOffset
// Author       : Andy Tate
// Date         : 19-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::GetMovementOffset( Asura_Vector_3* pxOutputPos, Asura_Quat* pxOutputQuat ) const
{
	return(m_pxTweenStore->GetMovementOffset(pxOutputPos, pxOutputQuat));
}

/// End of function 'Asura_AnimationController_Group::GetMovementOffset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::GetVelocity
// Author       : Andy Tate
// Date         : 19-Sep-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationController_Group::GetVelocity (Asura_Vector_3& xOutput) const
{
	return(m_pxTweenStore->GetVelocity(xOutput));
}

/// End of function 'Asura_AnimationController_Group::GetVelocity'. ///

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController_Group::PrintDebugInfo
// Author       : Kim Randell
// Date         : 07-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController_Group::PrintDebugInfo(u_int& uRow,
													 u_int& uCol,
													 const Asura_AnimationController_Group* pxParentGroup,
													 const char* szFilename /*= NULL*/) const
{
	Asura_AnimationController* pxController = m_pxFirstChildController;

	if(m_szName == NULL)
	{
		if ( szFilename == NULL )
		{
			Asura_DebugTextPrint(( uCol, uRow, 0xffffffff, 0, "Pri = %d, Flags = %08X, Delta = %s, Group = 0x%08x", m_uPriority, m_uControllerFlags, AreAllBonesUsingDeltas() ? "true" : "false", m_uControllerHash ));
		}
		else
		{
			for ( u_int uItr = 0; uItr < uCol; ++uItr )
			{
				Asura_SimpleLogfileOutput(( szFilename, " " ));
			}
			Asura_SimpleLogfileOutput(( szFilename, "Pri = %d, Flags = %08X, Delta = %s, Group = 0x%08x\n", m_uPriority, m_uControllerFlags, AreAllBonesUsingDeltas() ? "true" : "false", m_uControllerHash ));
		}
	}
	else
	{
		if ( szFilename == NULL )
		{
			Asura_DebugTextPrint(( uCol, uRow, 0xffffffff, 0, "Pri = %d, Flags = %08X, Delta = %s, Group = 0x%08x \"%s\"", m_uPriority, m_uControllerFlags, AreAllBonesUsingDeltas() ? "true" : "false", m_uControllerHash, m_szName ));
		}
		else
		{
			for ( u_int uItr = 0; uItr < uCol; ++uItr )
			{
				Asura_SimpleLogfileOutput(( szFilename, " " ));
			}
			Asura_SimpleLogfileOutput(( szFilename, "Pri = %d, Flags = %08X, Delta = %s, Group = 0x%08x \"%s\"\n", m_uPriority, m_uControllerFlags, AreAllBonesUsingDeltas() ? "true" : "false", m_uControllerHash, m_szName ));
		}
	}

	uCol	+= dbg_uGroupNumIndentationChars;
	uRow	+= 1;

	while(pxController)
	{
		Asura_AnimationBlend*	pxNextBlend	= pxController->GetBlendToNextController();

		pxController->PrintDebugInfo(uRow, uCol, this, szFilename);
		if(pxNextBlend)
		{
			pxNextBlend->PrintDebugInfo(uRow, uCol, this, szFilename);
		}
		pxController = pxNextBlend ? pxNextBlend->GetNextController() : NULL;

	}

	uCol	-= dbg_uGroupNumIndentationChars;
}
/// End of function 'Asura_AnimationController_Group::PrintDebugInfo'. ///
#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController::Duplicate
// Author       : Kim Randell
// Date         : 03/05/2007
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController* Asura_AnimationController_Group::Duplicate(Asura_Animation* pxDestAnimation) const
{
	// Get the parent group
	Asura_AnimationController_Group* pxParentGroup = pxDestAnimation->GetControllerGroupFromHash(m_uParentControllerHash);

	// Create a new blend
	Asura_AnimationBlend* pxBlend = m_pxBlendToPreviousController ? m_pxBlendToPreviousController->Duplicate() : NULL;

	Asura_AnimationController_Group* pxNewGroup =
		pxDestAnimation->CreateControllerGroup(m_uControllerHash, m_uControllerFlags, m_uPriority, pxBlend, pxParentGroup);

	pxNewGroup->CloneTweenStore(*this);

	// Copy members
	pxNewGroup->m_uFrameNumberStamp = m_uFrameNumberStamp;

	// Now iterate through children and duplicate them
	Asura_AnimationController* pxController = m_pxFirstChildController;
	while(pxController)
	{
		pxController->Duplicate(pxDestAnimation);
		const Asura_AnimationBlend*	pxNextBlend	= pxController->GetBlendToNextController();
		pxController = pxNextBlend ? pxNextBlend->GetNextController() : NULL;
	}

	return(pxNewGroup);
}

int Asura_AnimationController_Group::GetFlagsForLastFrame( void ) const
{
	int iFlags = 0;
	const Asura_AnimationController* pxController = m_pxFirstChildController;
	while ( pxController != NULL )
	{
		iFlags |= pxController->GetFlagsForLastFrame();

		const Asura_AnimationBlend* pxNextBlend = pxController->GetBlendToNextController();
		pxController = pxNextBlend != NULL ? pxNextBlend->GetNextController() : NULL;
	}
	return iFlags;
}

/// End of function 'Asura_AnimationController::Duplicate'. ///


/// End of file. ///
