//
// Asura_Hierarchy.cpp
//
// Created by Kevin Lea - 12/06/2001 11:13:40
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Animation/Asura_Hierarchy.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_NameTree.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Rendering/Asura_Render.h"

#ifdef ASURA_TOOLS
#include "Mesh/Asura_Mesh_System.h"
#endif

#ifdef ASURA_TOOLS
bool Asura_Hierarchy_Anim::s_bPreviewCompressedKeyframes = false;
#endif

#ifdef ASURA_TOOLS
Asura_Mesh* Asura_Hierarchy_Skin::GetMesh() const
{
	//find the mesh if we haven't got the pointer
	if(IsSkinnedModel())
	{
		if(!m_pxMesh)
		{
			m_pxMesh = Asura_Mesh_System::GetMesh(Name);
			Asura_Assertf(m_pxMesh,("Failed to find mesh for skin '%s'",Name.GetString()));
		}
	}
	return m_pxMesh;
}
#endif

Asura_Hierarchy_Section::Asura_Hierarchy_Section( void )
{
    m_pxCollisionMesh = NULL;
#ifdef ASURA_TOOLS
    Mesh = NULL;
#endif
}

Asura_Hierarchy_Section::~Asura_Hierarchy_Section( void )
{
    delete m_pxCollisionMesh;
}

Asura_Hierarchy_Skin::Asura_Hierarchy_Skin( void )
{
	Initialise();
}

void Asura_Hierarchy_Skin::Initialise()
{
	VertexData = 0;
	NormalData = 0;
	SectionData = 0;
	ParentBone = 0;
	m_piMirroredBone = 0;
	BoneRestPose = 0;
	#ifdef ASURA_TOOLS
	m_pxInitialRestPose = 0;
	#endif
    m_pxSectionBoundingBoxData = 0;
	#ifdef ASURA_TOOLS
	m_pxManualSectionBoundingBoxData = 0;
	#endif

#ifdef ASURA_HIERARCHY_QUAD
	m_pxBoneRestPoseInverseTransforms = NULL;
	m_pxBoneRestPoseTransforms = NULL;
#else
	BoneRestPosition = NULL;
	BoneRestOrientation = NULL;
#endif
	
	NumberOfVertices = 0;
	NumberOfBones = 0;

	NumberOfMarkerPoints = 0;
	MarkerPoint = 0;

	m_pxNextHigherSkin = 0;
	m_pxNextLowerSkin = 0;
	m_fMinDistance = 0;
	#ifdef ASURA_TOOLS
	m_bManualLOD = false;
	#endif

	m_pxMesh = 0;

	BoneTree = new Asura_NameTree;

    m_fBoundingRadius  = 0.0f;

	m_pxCollisionObject = 0;
	m_bHasCollisionMesh = 0;
	m_bHasCollisionBox = 0;
	m_bHasCustomCollisionBox = 0;

	m_iNumberOfSoundDefns = 0;
	m_pxSoundDefns = 0;

	#ifdef ASURA_TOOLS
	m_bAllowOverwriteOnLoad = false;
	m_xModifiedTime.dwLowDateTime = 0;
	m_xModifiedTime.dwHighDateTime = 0;
	#endif

	#if defined(ASURA_TOOLS) && defined(ASURA_SUPPORT_PLATFORM_HASHED_FILENAMES)
	m_iPlatform = -1;
	#endif


	#ifdef ASURA_TOOLS
	m_piBoneSortKey = 0;
	#endif

}

Asura_Hierarchy_Skin::~Asura_Hierarchy_Skin( void )
{
	Shutdown();
}

void Asura_Hierarchy_Skin::Shutdown()
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		delete m_pxNextLowerSkin;

		Asura_ResourceDeallocateArray(VertexData);
		Asura_ResourceDeallocateArray(NormalData);
		delete BoneTree;

		if (!m_pxNextHigherSkin)
		{
			//These pointers should only be deleted from the highest level
			//(An exception may need to be made for SectionData later when dealing with
			//detail models for non-skinned hierarchies) -RAR 03/09/02
			Asura_ResourceDeallocateArray(SectionData);
			Asura_ResourceDeallocateArray(ParentBone);
			Asura_ResourceDeallocateArray(m_piMirroredBone);
			Asura_ResourceDeallocateArray(BoneRestPose);
			#ifdef ASURA_TOOLS
			Asura_ResourceDeallocateArray(m_pxInitialRestPose);
			#endif

			Asura_ResourceDeallocateArray(m_pxSectionBoundingBoxData);				

#ifdef ASURA_HIERARCHY_QUAD
			Asura_ResourceDeallocateArray(m_pxBoneRestPoseInverseTransforms);
			Asura_ResourceDeallocateArray(m_pxBoneRestPoseTransforms);
#else
			Asura_ResourceDeallocateArray(BoneRestPosition);
			Asura_ResourceDeallocateArray(BoneRestOrientation);
#endif

			Asura_ResourceDeallocateArray(MarkerPoint);
			

		}
#ifdef ASURA_TOOLS
		else if(!IsSkinnedModel())
		{
			//This is an LOD model for an object hierarchy.
			//In tools , we have to delete the array (because of the different mesh pointers)
			//Otherwise it is just a copy of the section data pointer. RAR 18/02/2005
			Asura_ResourceDeallocateArray(SectionData);
		}
#endif

		if(m_bHasCollisionBox) delete (Asura_Collision_Box*)m_pxCollisionObject;
		if(m_bHasCollisionMesh) delete (Asura_CollisionMesh*)m_pxCollisionObject;

		Asura_ResourceDeallocateArray(m_pxSoundDefns);

#ifdef ASURA_TOOLS
		Asura_ResourceDeallocateArray(m_pxManualSectionBoundingBoxData);
#endif

#ifdef ASURA_TOOLS
		Asura_ResourceDeallocateArray(m_piBoneSortKey);
#endif	
	}
	Asura_ResourcePopType();
}

void Asura_Hierarchy_Skin::Reset()
{
	Shutdown();
	Initialise();	
}

int Asura_Hierarchy_Skin::GetParentBoneIndex( int iBoneIndex ) const
{
	if( (iBoneIndex > 0) && (iBoneIndex < NumberOfBones) )
	{
		return ParentBone[iBoneIndex];
	}

	return -1;
}

bool Asura_Hierarchy_Skin::GetBoneIndex( const char* boneName, int& boneIndex )
{
	return BoneTree->FindNodeIndexByName(boneName, false, boneIndex);
}

bool Asura_Hierarchy_Skin::GetMarkerPositionInRestPose(Asura_Hash_ID uMarkerHash, Asura_Vector_3 *pxPosition, Asura_Matrix_3x3 *pxOrientation) const
{
	int iIndex = -1;
	if (GetMarkerPointIndex(uMarkerHash, iIndex))
	{
		const int iBoneIndex = MarkerPoint[iIndex].BoneIndex;
		if(iBoneIndex == -1)
		{
			//RAR 18/05/2006 Compass relative marker point
			if(pxPosition)
			{
				*pxPosition = MarkerPoint[iIndex].Position;
			}
			if(pxOrientation)
			{
				*pxOrientation = MarkerPoint[iIndex].Matrix;
			}
		}
		else
		{
#if defined(ASURA_HIERARCHY_QUAD)
			if(pxPosition)
			{
				Asura_AVector4	xPosition(MarkerPoint[iIndex].Position, 1.0f);

				xPosition	= xPosition * m_pxBoneRestPoseTransforms[iBoneIndex];
				pxPosition->SetX(xPosition.GetX());
				pxPosition->SetY(xPosition.GetY());
				pxPosition->SetZ(xPosition.GetZ());
			}
			if(pxOrientation)
			{
				Asura_Matrix_3x3	xBoneOri;

				m_pxBoneRestPoseTransforms[iBoneIndex].ToMatrix(xBoneOri);
				*pxOrientation	= xBoneOri * MarkerPoint[iIndex].Matrix;
			}
#else
			//RAR 18/05/2006 BoneOrientation in fact stores the transpose of the orientations
			Asura_Matrix_3x3 xBoneOri = BoneRestOrientation[iBoneIndex];
			xBoneOri.Transpose();
			if (pxPosition)
			{
				*pxPosition = BoneRestPosition[iBoneIndex] + (MarkerPoint[iIndex].Position * xBoneOri);
			}
			if (pxOrientation)
			{
				*pxOrientation = xBoneOri * MarkerPoint[iIndex].Matrix;
			}
#endif	// defined(ASURA_HIERARCHY_QUAD)
		}
		return true;
	}
	return false;
}


#if defined(ASURA_HIERARCHY_QUAD)

void Asura_Hierarchy_Skin::CalcRestPoseBone(const int iBone, const Asura_AMatrix4x4& xTransform)
{
	Asura_AMatrix4x4	xBoneTransform(BoneRestPose[iBone].Quat, BoneRestPose[iBone].Position);

	// Calculate the bone matrix...
//	xBoneTransform						= xTransform * xBoneTransform;
	xTransform.Apply(xBoneTransform, xBoneTransform);
	m_pxBoneRestPoseTransforms[iBone]	= xBoneTransform;
	m_pxBoneRestPoseTransforms[iBone].InvertOrthogonal(m_pxBoneRestPoseInverseTransforms[iBone]);

	// Update the child bones...
	for(int i = 1; i < NumberOfBones; i++)
	{
		if(ParentBone[i] == iBone)
		{
			CalcRestPoseBone(i, xBoneTransform);
		}
	}
}

void Asura_Hierarchy_Skin::CalculateRestPose( void )
{
	Asura_Hierarchy_System::s_pxSkin = this;
	CalcRestPoseBone(0, Asura_Maths::IdentityAMatrix4x4);
}

#else

void Asura_Hierarchy_Skin::CalcRestPoseBone(const int iBone, const Asura_Matrix_3x3& xOrientation, const Asura_Vector_3& xPosition )
{
	Asura_Matrix_3x3 boneMat;
	BoneRestPose[iBone].Quat.ToMatrix( boneMat );
	boneMat = xOrientation * boneMat;

	Asura_Vector_3 ps = BoneRestPose[iBone].Position;
	ps = xPosition + (ps * xOrientation);

	BoneRestPosition[iBone] = ps;
	BoneRestOrientation[iBone] = boneMat;
	BoneRestOrientation[iBone].Transpose();

	for( int i = 1; i < NumberOfBones; i++ )
	{
		if( ParentBone[i] == iBone )
		{
			CalcRestPoseBone( i, boneMat, ps );
		}
	}
}

void Asura_Hierarchy_Skin::CalculateRestPose( void )
{
	Asura_Hierarchy_System::s_pxSkin = this;
	CalcRestPoseBone(0, Asura_Maths::IdentityMatrix3x3, Asura_Maths::ZeroVector3);
}

#endif	// defined(ASURA_HIERARCHY_QUAD)

void Asura_Hierarchy_Skin::SetupBB( bool bUseRestPose /*= true*/ )
{
    if( !m_pxSectionBoundingBoxData )
    {
        Asura_Assert( false, "Shouldn't be calling SetupBB() if section BB data hasn't been loaded/generated." );
        return;
    }

    m_xBB.Reset();
	bool bBoxSet = false;
	bool bNoPolys = true;

#ifndef ASURA_HIERARCHY_QUAD
	const Asura_Matrix_3x3* pxOrientations = ( bUseRestPose ? BoneRestOrientation : Asura_Hierarchy_System::s_axBoneOri );
	const Asura_Vector_3*   pxPositions    = ( bUseRestPose ? BoneRestPosition    : Asura_Hierarchy_System::s_axBonePos );
#endif

	Asura_Bounding_Box xTmpBB;
	for( int i = (IsSkinnedModel() ? 1 : 0); i < NumberOfBones; ++i )
	{
		const Asura_Hierarchy_SectionBoundingBox& xNextSectionBB = m_pxSectionBoundingBoxData[i];
		if( !xNextSectionBB.m_uValid ) { continue; }

		// SB: Meshes with no polygons suppress the 'No valid section bounding boxes' assert
		if ( bNoPolys )
		{
			if ( IsSkinnedModel() )
			{
				bNoPolys = false;
			}
			else
			{
				if ( SectionData[i].m_pxCollisionMesh && SectionData[i].m_pxCollisionMesh->GetNumberOfPolys() > 0 )
				{
					bNoPolys = false;
				}
			}
		}

#if defined(ASURA_HIERARCHY_QUAD)
		Asura_Matrix_3x3	xOrientation;
		Asura_Vector_3		xPosition;

		// RJT TODO: Optimise "Asura_Bounding_Box" to use "Asura_AMatrix4x4"...
		if(bUseRestPose)
		{
			m_pxBoneRestPoseTransforms[i].ToMatrix(xOrientation);
			xPosition = m_pxBoneRestPoseTransforms[i].GetTranslation();
		}
		else
		{
			Asura_Hierarchy_System::s_axBoneTransforms[i].ToMatrix(xOrientation);
			xPosition = Asura_Hierarchy_System::s_axBoneTransforms[i].GetTranslation();
		}

		xTmpBB.FromOBB(xNextSectionBB.m_xBB, xOrientation);
		xTmpBB.Displace(xPosition);
#else
		if(bUseRestPose)
		{
			//BoneRestOrientation contains the inverse of the orientations , so we need to transpose to get the desired matrix
			//RAR 14/02/2007
			Asura_Matrix_3x3 xInvMat = pxOrientations[i];
			xInvMat.Transpose();
			xTmpBB.FromOBB( xNextSectionBB.m_xBB, xInvMat );
		}
		else
		{
			xTmpBB.FromOBB( xNextSectionBB.m_xBB, pxOrientations[i] );
		}
		xTmpBB.Displace( pxPositions[i] );
#endif	// defined(ASURA_HIERARCHY_QUAD)

		bBoxSet = true;
		m_xBB.Set( xTmpBB );
	}

	// if we only have one bone, we will never have valid BB data so zero BB without asserting
	// SB - modified this test to check to work for any skeleton with no mesh
	if ( !bBoxSet && bNoPolys )
	{
		m_xBB.Zero();
		bBoxSet = true;
	}

	Asura_Assertf(bBoxSet, ("No valid section bounding boxes found in skin %s. Using zero-size box.", Name.GetString()));

	if (!bBoxSet)
	{
		m_xBB.Zero();
	}

	if( !bUseRestPose )
	{
#ifndef ASURA_HIERARCHY_QUAD
		if( Asura_Hierarchy_System::s_bBonesInWorldspace )
		{
			m_xBB.Displace( -*Asura_Hierarchy_System::s_pxPos );
		}
		else
		{
			m_xBB.FromOBB( m_xBB, *Asura_Hierarchy_System::s_pxOri );
		}
#else
		// RJT TODO: Cleanup/optimise
        if( Asura_Hierarchy_System::s_bBonesInWorldspace )
        {
			Asura_Vector_3	xPos = Asura_Hierarchy_System::s_xBaseTransform.GetTranslation();
			m_xBB.Displace(-xPos);
        }
        else
        {
			Asura_Matrix_3x3 xOri;

			Asura_Hierarchy_System::s_xBaseTransform.ToMatrix(xOri);
			m_xBB.FromOBB( m_xBB, xOri );
        }
#endif
	}

    m_fBoundingRadius = 0.5f * m_xBB.CalculateExtents().Magnitude();


	if(!m_pxCollisionObject)
	{
		m_bHasCollisionBox = true;
		m_pxCollisionObject = new Asura_Collision_Box(m_xBB);
	}
	else if(m_bHasCollisionBox && !m_bHasCustomCollisionBox)
	{
		//Update the existing collision box
		((Asura_Collision_Box*)m_pxCollisionObject)->ChangeBB(m_xBB);
	}
}


Asura_Hierarchy_Skin* Asura_Hierarchy_Skin::GetDetailSkinFromDistance(float fDistance)
{
#if defined(ASURA_USE_SCREENSHOT_SYSTEM)
	if(Asura_Render::IsHiResScreenshotModeEnabled())
	{
		return(this);
	}
	else
#endif	//defined(ASURA_USE_SCREENSHOT_SYSTEM)
	{
		fDistance *= Asura_Camera::GetLODMetric(Asura_Camera::GetCameraIndex()); //Asura_Mesh_System::Project_GetLODMetric(Asura_Camera::GetCameraIndex());

		Asura_Hierarchy_Skin* pxSkin = this;
		while(pxSkin->m_pxNextLowerSkin)
		{
			if(pxSkin->m_pxNextLowerSkin->m_fMinDistance > fDistance) break;
			pxSkin = pxSkin->m_pxNextLowerSkin;
		}
		return pxSkin;
	}
}

const Asura_Hierarchy_Skin* Asura_Hierarchy_Skin::GetDetailSkinFromDistance(float fDistance) const
{
#if defined(ASURA_USE_SCREENSHOT_SYSTEM)
	if(Asura_Render::IsHiResScreenshotModeEnabled())
	{
		return(this);
	}
	else
#endif	//defined(ASURA_USE_SCREENSHOT_SYSTEM)
	{
		fDistance *= Asura_Camera::GetLODMetric(Asura_Camera::GetCameraIndex()); //Asura_Mesh_System::Project_GetLODMetric(Asura_Camera::GetCameraIndex());

		const Asura_Hierarchy_Skin* pxSkin = this;
		while(pxSkin->m_pxNextLowerSkin)
		{
			if(pxSkin->m_pxNextLowerSkin->m_fMinDistance > fDistance) break;
			pxSkin = pxSkin->m_pxNextLowerSkin;
		}
		return pxSkin;
	}
}

const Asura_Hierarchy_Skin* Asura_Hierarchy_Skin::GetRelativeDetailSkin(unsigned int uLevels) const
{
	const Asura_Hierarchy_Skin* pxSkin = this;
	while(pxSkin->m_pxNextLowerSkin && uLevels > 0)
	{
		pxSkin = pxSkin->m_pxNextLowerSkin;
		uLevels--;
	}
	return pxSkin;
}

void Asura_Hierarchy_Skin::CopyFromRootSkin( const Asura_Hierarchy_Skin* pxRootSkin)
{
	//copy data that is shared between detail levels
	NumberOfBones = pxRootSkin->NumberOfBones;
	ParentBone = pxRootSkin->ParentBone;
	m_piMirroredBone = pxRootSkin->m_piMirroredBone;
	BoneRestPose = pxRootSkin->BoneRestPose;
#ifdef ASURA_HIERARCHY_QUAD
	m_pxBoneRestPoseInverseTransforms	= pxRootSkin->m_pxBoneRestPoseInverseTransforms;
	m_pxBoneRestPoseTransforms			= pxRootSkin->m_pxBoneRestPoseTransforms;
#else
	BoneRestPosition	= pxRootSkin->BoneRestPosition;
	BoneRestOrientation	= pxRootSkin->BoneRestOrientation;
#endif
	NumberOfMarkerPoints = pxRootSkin->NumberOfMarkerPoints;
	MarkerPoint = pxRootSkin->MarkerPoint;

	#ifdef ASURA_TOOLS
	if(!pxRootSkin->IsSkinnedModel())
	{
		//in tools we need to make a copy of the section data for hierarchical models ,
		//in order to store the different mesh pointers.
		Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
		{
			Asura_ResourceAllocateArray(SectionData, NumberOfBones);
		}
		Asura_ResourcePopType();
		for(int i=0;i<NumberOfBones;i++)
		{
			SectionData[i].SectionName = pxRootSkin->SectionData[i].SectionName;
			//Mesh pointer will get initialised later by the loaders.
			//Collision mesh pointer should (hopefully) not be required
		}
	}
	else
	#endif
	{
		//For skinned models , or outside of tools , the section data for LODs
		//should be identical RAR 18/02/2005
		SectionData = pxRootSkin->SectionData;
	}
	
}



const Asura_Hierarchy_Bone_Anim& Asura_Hierarchy_Bone_Anim::operator = (const Asura_Hierarchy_Bone_Anim& xAnim)
{
	if(this != &xAnim)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
		{
			Asura_ResourceDeallocateArray(KeyFrames);
			Asura_ResourceDeallocateArray(Positions);
			KeyFrames = 0;
			Positions = 0;

			Initialise(xAnim.NumberOfKeyFrames,xAnim.ConstantOffset);

			for(int i=0;i<NumberOfKeyFrames;i++)
			{
				KeyFrames[i] = xAnim.KeyFrames[i];
				if(!(i && ConstantOffset))
				{
					Positions[i] = xAnim.Positions[i];
				}
			}
		}
		Asura_ResourcePopType();
	}
	return *this;
}

void Asura_Hierarchy_Anim::SetBB(const Asura_Bounding_Box* pxBB)
{
	if(pxBB)
	{
		if(!m_pxBB) m_pxBB = new Asura_Bounding_Box;
		*m_pxBB = *pxBB;
	    m_fBoundingRadius = 0.5f * m_pxBB->CalculateExtents().Magnitude();
	}
	else if(m_pxBB)
	{
		delete m_pxBB;
		m_pxBB = 0;
		m_fBoundingRadius = 0;
	}
}


#ifdef ASURA_TOOLS


int Asura_Hierarchy_Anim::CalculateTotalKeyFrames(void) const
{
	int	iTotal	= 0;

	for(int i=0;i<GetBoneArraySize();i++)
	{		
		iTotal	+= BoneAnims[i].NumberOfKeyFrames;
	}

	return(iTotal);
}

void Asura_Hierarchy_Anim::RemoveUnneededKeyFrames(float fPositionToleranceMetres ,float fAngleToleranceDegrees ,int* pxOutputFramesBefore,int* pxOutputFramesAfter)
{
	if(m_bDontRemoveKeyframes)
	{
		int	iTotalKeyFrames	= CalculateTotalKeyFrames();

		if(pxOutputFramesBefore) *pxOutputFramesBefore = iTotalKeyFrames;
		if(pxOutputFramesAfter) *pxOutputFramesAfter = iTotalKeyFrames;
	}
	else
	{
		if(pxOutputFramesBefore) *pxOutputFramesBefore = 0;
		if(pxOutputFramesAfter) *pxOutputFramesAfter = 0;

		//convert tolerance values into values used by the ApproximatelyEquals functions
		float fPositionTolerance = fPositionToleranceMetres*fPositionToleranceMetres;
		float fQuatTolerance = 1.0f - Asura_Maths::Cosine(Asura_Maths::DegToRad(fAngleToleranceDegrees)/2);

		for(int i=0;i<GetBoneArraySize();i++)
		{
			if(pxOutputFramesBefore) *pxOutputFramesBefore += BoneAnims[i].NumberOfKeyFrames;

			BoneAnims[i].RemoveUnneededKeyFrames(fPositionTolerance,fQuatTolerance);

			if(pxOutputFramesAfter) *pxOutputFramesAfter += BoneAnims[i].NumberOfKeyFrames;
		}
	}
}

void Asura_Hierarchy_Bone_Anim::RemoveUnneededKeyFrames(float fPositionTolerance,float fQuatTolerance)
{
	if (NumberOfKeyFrames>2)
	{
		int s=0;
		int m=1;
		int e=2;

		//int n = NumberOfKeyFrames;
		
		for (s=0; e<NumberOfKeyFrames && m<NumberOfKeyFrames;)
		{
			
			float ts = Time(s);
			float tm = Time(m);
			float te = Time(e);

			float lambda = (tm-ts)/(te-ts);

			bool bOrientMatch = false;
			bool bPositionMatch = false;

			//if the constant offset flag hasn't been set , see if the middle position matches
			//the result from interpolating the other 2
			if(!ConstantOffset)
			{
				Asura_Vector_3 ps = Position(s);
				Asura_Vector_3 pm = Position(m);
				Asura_Vector_3 pe = Position(e);

				Asura_Vector_3 p;

				p = ps + (pe-ps)*lambda;

  				bPositionMatch = p.ApproximatelyEquals(pm,fPositionTolerance);
			}
			else
			{
				bPositionMatch = true;
			}


			//check the orientation
			if(bPositionMatch)
			{
				Asura_Quat qs = Quat(s);
				Asura_Quat qm = Quat(m);
				Asura_Quat qe = Quat(e);

				Asura_Quat q;
				q.Slerp(qs,qe,lambda);

				bOrientMatch = q.ApproximatelyEquals(qm,fQuatTolerance);

				if (bOrientMatch &&
					Asura_Maths::Abs(qs.Dot(qe)) < 0.001f)
				{
					// (AW - 09.07.2003) - The start quaternion is about 180 degrees
					// off from the end one. Once the quaternions have been compressed
					// and decompressed, floating point inaccuracies may result in this 
					// slerp going in different directions on different platforms. Leave
					// the middle frame in, as it will help to get the direction right.
					bOrientMatch = false;
				}
			}
			
			if (bOrientMatch && bPositionMatch)
			{
				//remove this keyframe
				--NumberOfKeyFrames;
				for (int f=m; f<NumberOfKeyFrames; f++)
				{
					Quat(f) = Quat(f+1);
					Time(f) = Time(f+1);
					if(!ConstantOffset)
					{
						Position(f) = Position(f+1);
					}
				}
			}
			else
			{
				s++;
				m++;
				e++;
			}

		}
	}
	
	if (NumberOfKeyFrames==2)
	{
		//down to 2 keyframes... see if they are identical
		Asura_Quat qs = Quat(0);
		Asura_Quat qe = Quat(1);
		
		Asura_Vector_3 ps = Position(0);
		Asura_Vector_3 pe = Position(1);
		if (ps.ApproximatelyEquals(pe,fPositionTolerance) && qs.ApproximatelyEquals(qe,fQuatTolerance))
		{
			//look pretty similar... goodbye frame 2!
			NumberOfKeyFrames = 1;

		}

	}

	if(NumberOfKeyFrames && !ConstantOffset)
	{
		//See if the offsets are constant for this bone
		bool constantOffset = true;
		for(int f=1;f<NumberOfKeyFrames;f++)
		{
			if(!Position(f).ApproximatelyEquals(Position(0),fPositionTolerance))
			{
				constantOffset = false;
				break;
			}
		}
		if(constantOffset)
		{
			ConstantOffset = true;
		}
	}

}



//////////////////
#if 0 // Disabled code for removing frames , which has the snag of needing some hierarchy information. RAR 10/03/03
/*
void Asura_Hierarchy_Anim::RemoveUnneededKeyFrames(const Asura_Hierarchy_Skin* pxSkin,float fPositionToleranceMetres ,float fAngleToleranceDegrees ,int* pxOutputFramesBefore,int* pxOutputFramesAfter)
{
	if(pxOutputFramesBefore) *pxOutputFramesBefore = 0;
	if(pxOutputFramesAfter) *pxOutputFramesAfter = 0;
	
	//convert tolerance values into values used by the ApproximatelyEquals functions
	float fPositionTolerance = fPositionToleranceMetres*fPositionToleranceMetres;
	float fQuatTolerance = 1.0f - Asura_Maths::Cosine(Asura_Maths::DegToRad(fAngleToleranceDegrees)/2);

	for(int i=0;i<NumberOfBones;i++)
	{
		BoneAnims[i].RemoveUnneededKeyFrames(pxSkin,this,i,fPositionTolerance,fQuatTolerance,pxOutputFramesBefore,pxOutputFramesAfter);
	}
}

bool Asura_Hierarchy_Anim::FramesAreApproximatelyEqual(const Asura_Hierarchy_Skin* pxSkin,int iBoneIndex,float fTime,float fPositionTolerance,float fQuatTolerance,Asura_Vector_3 xPos1,const Asura_Quat& xQuat1, const Asura_Vector_3& xPos2 , const Asura_Quat& xQuat2) const
{
		
	if(!xPos1.ApproximatelyEquals(xPos2,fPositionTolerance)) return false;
	if(!xQuat1.ApproximatelyEquals(xQuat2,fQuatTolerance)) return false;

	//check children
	for(int i=iBoneIndex+1;i<NumberOfBones;i++)
	{
		if(pxSkin->ParentBone[i] != iBoneIndex) continue;
		if(!BoneAnims[i].NumberOfKeyFrames) continue;


		//get the relative pos/quat for this bone
		Asura_Quat xQuatRel;
		Asura_Vector_3 xPosRel;

		if (BoneAnims[i].NumberOfKeyFrames==1)
		{
			xQuatRel = BoneAnims[i].KeyFrames[0].Quat;
			xPosRel = BoneAnims[i].Position(0);
		}
		else
		{
			int iFrame;
			for (iFrame = 0; iFrame<BoneAnims[i].NumberOfKeyFrames-2; iFrame++)
			{
				if (BoneAnims[i].KeyFrames[iFrame+1].Time > fTime) break;
			}

			int iNextFrame = iFrame+1;
		
			float fTimeslice = (fTime - BoneAnims[i].KeyFrames[iFrame].Time)/(BoneAnims[i].KeyFrames[iNextFrame].Time-BoneAnims[i].KeyFrames[iFrame].Time);
			if (fTimeslice<0) fTimeslice = 0;
			if (fTimeslice>1) fTimeslice = 1;

			Asura_Quat q1 = BoneAnims[i].KeyFrames[iFrame].Quat;
			Asura_Quat q2 = BoneAnims[i].KeyFrames[iNextFrame].Quat;
		
			xQuatRel.Slerp(q1,q2,fTimeslice);

			Asura_Vector_3 p1 = BoneAnims[i].Position(iFrame);
			Asura_Vector_3 p2 = BoneAnims[i].Position(iNextFrame);

		
			xPosRel = (p1)*(1.0f-fTimeslice) + p2*(fTimeslice);
		}

		//calculate the cumulative position for each of the alternate frames
		Asura_Quat xQuatTotal1 = xQuat1 * xQuatRel;
		Asura_Vector_3 xPosTotal1;
		xQuat1.RotateVector(xPosRel,xPosTotal1);
		xPosTotal1 += xPos1;

		Asura_Quat xQuatTotal2 = xQuat2 * xQuatRel;
		Asura_Vector_3 xPosTotal2;
		xQuat2.RotateVector(xPosRel,xPosTotal2);
		xPosTotal2 += xPos1;
		

		bool bMatch = FramesAreApproximatelyEqual(pxSkin,i,fTime,fPositionTolerance,fQuatTolerance,xPosTotal1,xQuatTotal1,xPosTotal2,xQuatTotal2);
		if(!bMatch) return false;
	
	}
		
	return true;
}

void Asura_Hierarchy_Bone_Anim::RemoveUnneededKeyFrames(const Asura_Hierarchy_Skin* pxSkin,const Asura_Hierarchy_Anim* pxAnim,int iIndex,float fPositionTolerance,float fQuatTolerance,int* pxOutputFramesBefore,int* pxOutputFramesAfter)
{
	if(pxOutputFramesBefore) *pxOutputFramesBefore += NumberOfKeyFrames;
	if (NumberOfKeyFrames>2)
	{
		int s=0;
		int m=1;
		int e=2;

		//int n = NumberOfKeyFrames;
		
		for (s=0; e<NumberOfKeyFrames && m<NumberOfKeyFrames;)
		{
			
			float ts = Time(s);
			float tm = Time(m);
			float te = Time(e);

			float lambda = (tm-ts)/(te-ts);
		
			Asura_Vector_3 ps = Position(s);
			Asura_Vector_3 pm = Position(m);
			Asura_Vector_3 pe = Position(e);

			Asura_Vector_3 p;

			p = ps + (pe-ps)*lambda;



			//check the orientation
			Asura_Quat qs = Quat(s);
			Asura_Quat qm = Quat(m);
			Asura_Quat qe = Quat(e);

			Asura_Quat q;
			q.Slerp(qs,qe,lambda);

			//Check to see if this is okay
			bool bMatch = pxAnim->FramesAreApproximatelyEqual(pxSkin,iIndex,tm,fPositionTolerance,fQuatTolerance,p,q,pm,qm);



			if (bMatch)
			{
				//remove this keyframe
				--NumberOfKeyFrames;
				for (int f=m; f<NumberOfKeyFrames; f++)
				{
					Quat(f) = Quat(f+1);
					Time(f) = Time(f+1);
					if(!ConstantOffset)
					{
						Position(f) = Position(f+1);
					}
				}
			}
			else
			{
				s++;
				m++;
				e++;
			}

		}
	}
	
	if (NumberOfKeyFrames==2)
	{
		//down to 2 keyframes... see if they are identical
		Asura_Quat qs = Quat(0);
		Asura_Quat qe = Quat(1);
		
		Asura_Vector_3 ps = Position(0);
		Asura_Vector_3 pe = Position(1);
		if (ps.ApproximatelyEquals(pe,fPositionTolerance) && qs.ApproximatelyEquals(qe,fQuatTolerance))
		{
			//look pretty similar... goodbye frame 2!
			NumberOfKeyFrames = 1;

		}

	}

	if(NumberOfKeyFrames && !ConstantOffset)
	{
		//See if the offsets are constant for this bone
		bool constantOffset = true;
		for(int f=1;f<NumberOfKeyFrames;f++)
		{
			if(!Position(f).ApproximatelyEquals(Position(0),fPositionTolerance))
			{
				constantOffset = false;
				break;
			}
		}
		if(constantOffset)
		{
			ConstantOffset = true;
		}
	}

	if(pxOutputFramesAfter) *pxOutputFramesAfter += NumberOfKeyFrames;
}

*/
#endif //frame removal test

#endif // ASURA_TOOLS
void Asura_Hierarchy_Anim::MakeCompressedKeyFrames(void)
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		Asura_Assert(!HasCompressedKeyFrames(), "already has compressed keyframes - will cause memory leak");

		// KJL - 18/06/2003 18:26:19 - first lets count the number of keyframes required
		m_uTotalNumberOfCompressedKeyFrames = 0;
		m_uTotalNumberOfCompressedBoneOffsets = 0;
		int iBoneArraySize = GetBoneArraySize();
		for(int u=0; u<iBoneArraySize; u++)
		{
			m_uTotalNumberOfCompressedKeyFrames += BoneAnims[u].NumberOfKeyFrames;
			m_uTotalNumberOfCompressedBoneOffsets += BoneAnims[u].ConstantOffset ? 1 : BoneAnims[u].NumberOfKeyFrames;
		}

		Asura_ResourceAllocateArray(m_pxCompressedQuatKeyFrames, m_uTotalNumberOfCompressedKeyFrames);
		Asura_ResourceAllocateArray(m_pxCompressedTimeKeyFrames, m_uTotalNumberOfCompressedKeyFrames);

		Asura_ResourceAllocateArray(m_pxCompressedBoneOffsets, m_uTotalNumberOfCompressedBoneOffsets);

		Asura_ResourceAllocateArray(m_pxCompressedBoneData, iBoneArraySize);

		u_int w = 0;
		u_int x = 0;
		for(int u=0; u<iBoneArraySize; u++)
		{
			Asura_Assert(BoneAnims[u].NumberOfKeyFrames < 65536, "too many keyframes");

			m_pxCompressedBoneData[u].m_usNumberOfKeyFrames = static_cast<u_short>(BoneAnims[u].NumberOfKeyFrames);

			if (m_pxCompressedBoneData[u].m_usNumberOfKeyFrames)
			{
				Asura_Assertf( w < 65536 && x < 65536, ( "too many anim frames in anim '%s'", GetName() ) );

				m_pxCompressedBoneData[u].m_usFirstKeyFrameIndex = static_cast<u_short>(w);

				m_pxCompressedBoneData[u].m_usFirstBoneOffsetIndex = static_cast<u_short>(x);

				for (int v=0; v<BoneAnims[u].NumberOfKeyFrames; v++)
				{
					m_pxCompressedQuatKeyFrames[w].FromQuat(BoneAnims[u].KeyFrames[v].Quat);
					m_pxCompressedTimeKeyFrames[w] = (u_short)(((float)BoneAnims[u].KeyFrames[v].Time*65535.0f));
					w++;
				}

				if (BoneAnims[u].ConstantOffset)
				{
					m_pxCompressedBoneData[u].m_usNumberOfBoneOffsets = 1;
					m_pxCompressedBoneOffsets[x] = BoneAnims[u].Positions[0];
					x++;
				}
				else
				{
					m_pxCompressedBoneData[u].m_usNumberOfBoneOffsets = static_cast<u_short>(BoneAnims[u].NumberOfKeyFrames);
					Asura_Mem::Copy(&m_pxCompressedBoneOffsets[x],BoneAnims[u].Positions,sizeof(Asura_Vector_3)*BoneAnims[u].NumberOfKeyFrames);
					x+= BoneAnims[u].NumberOfKeyFrames;
				}
			}


			//	Asura_ResourceDeallocateArray(Asura_ResourceAllocator::ANIMATION_RESOURCE, BoneAnims[u].KeyFrames);
			//	BoneAnims[u].KeyFrames = 0;
		}
	}
	Asura_ResourcePopType();
}

// AW - 24/07/03 - Animations that are created in code need compressing, but calling DeleteCompressedKeyFrames and
// MakeCompressedKeyFrames will result in fragmentation, so this function is designed to avoid that.
void Asura_Hierarchy_Anim::RemakeCompressedKeyFrames(void)
{
	Asura_Assert(HasCompressedKeyFrames(), "Should have called MakeCompressedKeyFrames before this");
	if (!HasCompressedKeyFrames())
	{
		return;
	}
	int iBoneArraySize = GetBoneArraySize();
#ifdef ASURA_ASSERT
	// AW - 24/07/03 - Make sure the number of keyframes hasn't changed.
	unsigned uTotalNumberOfCompressedKeyFrames = 0;
	unsigned uTotalNumberOfCompressedBoneOffsets = 0;
	for(int u=0; u<iBoneArraySize; u++)
	{
		uTotalNumberOfCompressedKeyFrames += BoneAnims[u].NumberOfKeyFrames;
		uTotalNumberOfCompressedBoneOffsets += BoneAnims[u].ConstantOffset ? 1 : BoneAnims[u].NumberOfKeyFrames;
	}
	Asura_Assert(m_uTotalNumberOfCompressedKeyFrames == uTotalNumberOfCompressedKeyFrames, "Error - RemakeCompressedKeyFrames only works if the number of key frames hasn't changed");
	Asura_Assert(m_uTotalNumberOfCompressedBoneOffsets == uTotalNumberOfCompressedBoneOffsets, "Error - RemakeCompressedKeyFrames only works if the number of bone offsets hasn't changed");
	if (m_uTotalNumberOfCompressedKeyFrames != uTotalNumberOfCompressedKeyFrames ||
		m_uTotalNumberOfCompressedBoneOffsets != uTotalNumberOfCompressedBoneOffsets)
	{
		return;
	}
#endif //ASURA_ASSERT
	u_int w = 0;
	u_int x = 0;
	for(int u=0; u<iBoneArraySize; u++)
	{
		Asura_Assert(BoneAnims[u].NumberOfKeyFrames < 65536, "too many keyframes");

		m_pxCompressedBoneData[u].m_usNumberOfKeyFrames = static_cast<u_short>(BoneAnims[u].NumberOfKeyFrames);

		if (m_pxCompressedBoneData[u].m_usNumberOfKeyFrames)
		{
			Asura_Assertf( w < 65536 && x < 65536, ( "too many anim frames in anim '%s'", GetName() ) );
			
			m_pxCompressedBoneData[u].m_usFirstKeyFrameIndex = static_cast<u_short>(w);
			
			m_pxCompressedBoneData[u].m_usFirstBoneOffsetIndex = static_cast<u_short>(x);

			for (int v=0; v<BoneAnims[u].NumberOfKeyFrames; v++)
			{
				m_pxCompressedQuatKeyFrames[w].FromQuat(BoneAnims[u].KeyFrames[v].Quat);
				m_pxCompressedTimeKeyFrames[w] = (u_short)(((float)BoneAnims[u].KeyFrames[v].Time*65535.0f));
				w++;
			}

			if (BoneAnims[u].ConstantOffset)
			{
				m_pxCompressedBoneData[u].m_usNumberOfBoneOffsets = 1;
				m_pxCompressedBoneOffsets[x] = BoneAnims[u].Positions[0];
				x++;
			}
			else
			{
				m_pxCompressedBoneData[u].m_usNumberOfBoneOffsets = static_cast<u_short>(BoneAnims[u].NumberOfKeyFrames);
				Asura_Mem::Copy(&m_pxCompressedBoneOffsets[x],BoneAnims[u].Positions,sizeof(Asura_Vector_3)*BoneAnims[u].NumberOfKeyFrames);
				x+= BoneAnims[u].NumberOfKeyFrames;
			}
		}
	}
}


void Asura_Hierarchy_Anim::DetermineMovementScaleBone()
{
	//Determines which bone to look at for comparing bone lengths when scaling the movement offsets for
	//animations that are being played on different sized skins.
	//Normally the zeroth bone is used for this , but in some hierarchies the first bone is on the floor , rather
	//than in the middle of the character. In this case using the next bone is more appropriate.
	//RAR 23/11/2006
	m_iBoneForMovementScale = 0;
	if(!BoneLengths) return;
	if(NumberOfBones < 2) return;
	
	if((BoneLengths[0] < 0.1f) && (BoneLengths[1] > BoneLengths[0]))
	{
		m_iBoneForMovementScale = 1;
	}
}

float Asura_Hierarchy_Anim::GetMovementScale(const Asura_Hierarchy_Skin* pxSkin) const
{

	if(pxSkin && BoneLengths && (BoneLengths[m_iBoneForMovementScale] > 0.01f) && (m_iBoneForMovementScale < pxSkin->NumberOfBones))
	{
		//Scale movement according to the relative length of the first bone in the animation and the skin.
		//RAR  23/11/2006
		return pxSkin->BoneRestPose[m_iBoneForMovementScale].Position.Magnitude() / BoneLengths[m_iBoneForMovementScale];
	}
	return 1.0f;
}

const Asura_Collision_Object* Asura_Hierarchy_Skin::GetCollisionObject() const
{
	#ifdef ASURA_TOOLS
	//The bounding box information isn't calculated immediately in tools , so do
	//so now if needed. 
	//Need to forget about the constness to do this though
	//RAR 03/12/2007
	if(!m_pxSectionBoundingBoxData && !m_pxCollisionObject)
	{
		const_cast<Asura_Hierarchy_Skin*>(this)->InitSectionBoundingBoxes();
	}
	#endif
	return m_pxCollisionObject;
}

const Asura_Bounding_Box& Asura_Hierarchy_Skin::GetCollisionBB() const
{
	#ifdef ASURA_TOOLS
	//The bounding box information isn't calculated immediately in tools , so do
	//so now if needed. 
	//Need to forget about the constness to do this though
	//RAR 03/12/2007
	if(!m_pxSectionBoundingBoxData && !m_pxCollisionObject)
	{
		const_cast<Asura_Hierarchy_Skin*>(this)->InitSectionBoundingBoxes();
	}
	#endif
	return m_pxCollisionObject ? m_pxCollisionObject->GetBoundingBox() : m_xBB;
}

void Asura_Hierarchy_Skin::SetCollisionBB(const Asura_Bounding_Box& xBB)
{
	if(m_bHasCollisionBox) delete (Asura_Collision_Box*)m_pxCollisionObject;
	if(m_bHasCollisionMesh) delete (Asura_CollisionMesh*)m_pxCollisionObject;

	m_bHasCollisionBox = true;
	m_bHasCustomCollisionBox = true;
	m_bHasCollisionMesh = false;
	
	m_pxCollisionObject = new Asura_Collision_Box(xBB);
}

#ifdef ASURA_TOOLS
void Asura_Hierarchy_Skin::ResetCollisionBB()
{
	if(m_bHasCollisionBox) delete (Asura_Collision_Box*)m_pxCollisionObject;
	if(m_bHasCollisionMesh) delete (Asura_CollisionMesh*)m_pxCollisionObject;

	m_bHasCollisionBox = true;
	m_bHasCustomCollisionBox = false;
	m_bHasCollisionMesh = false;
	
	m_pxCollisionObject = new Asura_Collision_Box(GetBB());
}
#endif

const Asura_CollisionMesh* Asura_Hierarchy_Skin::GetCollisionMesh() const
{
    return( m_bHasCollisionMesh ? static_cast<const Asura_CollisionMesh*>(m_pxCollisionObject) : NULL );
}

Asura_CollisionMesh* Asura_Hierarchy_Skin::GetCollisionMesh( void )
{
    return( m_bHasCollisionMesh ? static_cast<Asura_CollisionMesh*>(m_pxCollisionObject) : NULL );
}

void Asura_Hierarchy_Skin::SetCollisionMesh(Asura_CollisionMesh* pxCollMesh)
{
	if(m_bHasCollisionBox) delete (Asura_Collision_Box*)m_pxCollisionObject;
	if(m_bHasCollisionMesh) delete (Asura_CollisionMesh*)m_pxCollisionObject;

	m_bHasCollisionBox = false;
	m_bHasCustomCollisionBox = false;
	m_bHasCollisionMesh = true;

	m_pxCollisionObject = pxCollMesh;
}

static const Asura_Hierarchy_Sound* GetSoundPropertiesFromAnimSoundIDMain(int iNumberOfSounds,const Asura_Hierarchy_Sound* pxSounds,int iAnimSoundID)
{
	if(iNumberOfSounds)
	{
		int iMax = iNumberOfSounds;
		int iMin = 0;
		
		int iCurrent = (iMin+iMax)/2;
		
		do
		{
			if (pxSounds[iCurrent].m_iAnimSoundID == iAnimSoundID)
			{
				break;
			}
			if (pxSounds[iCurrent].m_iAnimSoundID > iAnimSoundID)
			{
				iMax = iCurrent;
			}
			else
			{
				iMin = iCurrent;
			}
			iCurrent = (iMin+iMax)/2;
		}
		while(iCurrent!=iMin);

		if (pxSounds[iCurrent].m_iAnimSoundID == iAnimSoundID)
		{
			int iFirst = iCurrent;
			while(iFirst>0 && pxSounds[iFirst-1].m_iAnimSoundID == iAnimSoundID)
			{
				iFirst--;
			}
			int iLast = iCurrent;
			while(iLast<iNumberOfSounds-1 && pxSounds[iLast+1].m_iAnimSoundID == iAnimSoundID)
			{
				iLast++;
			}

			if(iFirst!=iLast)
			{
				iCurrent = iFirst + Asura_Random::GetIntU() % (1+iLast-iFirst);
			}

			return &pxSounds[iCurrent];
		}
	}
	return 0;
}

const Asura_Hierarchy_Sound* Asura_Hierarchy_Skin::GetSoundPropertiesFromAnimSoundID(int iAnimSoundID) const
{
	if(m_iNumberOfSoundDefns)
	{
		return GetSoundPropertiesFromAnimSoundIDMain(m_iNumberOfSoundDefns,m_pxSoundDefns,iAnimSoundID);
	}
	return 0;
}

const Asura_Hierarchy_Sound* Asura_Hierarchy_Anim::GetSoundPropertiesFromAnimSoundID(int iAnimSoundID) const
{
	if(m_iNumberOfSoundDefns)
	{
		return GetSoundPropertiesFromAnimSoundIDMain(m_iNumberOfSoundDefns,m_pxSoundDefns,iAnimSoundID);
	}
	return 0;
}

#ifdef ASURA_TOOLS
void Asura_Hierarchy_Skin::CalculateMirroredBones()
{
	//allocate and initialise mirrored bone array
	if(!m_piMirroredBone)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
		Asura_ResourceAllocateArray(m_piMirroredBone, NumberOfBones);
		Asura_ResourcePopType();
	}

	for(int i=0;i<NumberOfBones;i++)
	{
		m_piMirroredBone[i] = i;
	}

	for(int i=0;i<NumberOfBones;i++)
	{
		//ignore bones that have already been done
		if(m_piMirroredBone[i] != i) continue;

		for(int j=i+1;j<NumberOfBones;j++)
		{
			if(m_piMirroredBone[j] != j) continue;

			//the parent bones should be equilvalent for this bone pair to be considered to be mirrored
			int iMirroredParent = ParentBone[j];
			if(iMirroredParent >=0 ) iMirroredParent = m_piMirroredBone[iMirroredParent];
			if(iMirroredParent != ParentBone[i]) continue;

			//check the section names
			if(IsMirroredBoneName(SectionData[i].SectionName.GetString(),SectionData[j].SectionName.GetString()))
			{
				m_piMirroredBone[i] = j;
				m_piMirroredBone[j] = i;
				break;
			}
		
		}
	}

}

bool Asura_Hierarchy_Skin::IsMirroredBoneName(const char* szName1,const char* szName2)
{
	//Consider the names to be mirrored versions of each other if they are the same except for an L and an R being swapped.
	//Also accept being same except for differing by left/right. RAR 21/11/2006
	bool bFoundMirroredBit = false;
	const char* pcPos1 = szName1;
	const char* pcPos2 = szName2;

	while(*pcPos1 && *pcPos2)
	{
		char c1 = (char) tolower(*pcPos1++);		
		char c2 = (char) tolower(*pcPos2++);

		if(c1 == c2) continue;

		//found a difference
		//are the characters l and r?
		if((c1 == 'l' && c2== 'r') || (c1 == 'r' && c2== 'l'))
		{
			//are the l and r the beginning of left and right?

			const char*& pcLeftPos = (c1 == 'l') ? pcPos1 : pcPos2;
			const char*& pcRightPos = (c1 == 'l') ? pcPos2 : pcPos1;
			if(!strnicmp(pcLeftPos,"eft",3) && !strnicmp(pcRightPos,"ight",4))
			{
				//skip the rest of left/right
				pcLeftPos += 3;
				pcRightPos += 4;
			}

			
			bFoundMirroredBit = true;
		}
		else
		{
			return false;
		}
	}
	//make sure the strings weren't different lengths
	if(*pcPos1 || !pcPos2) return false;

	return bFoundMirroredBit;
}
#endif //ASURA_TOOLS

void Asura_Hierarchy_Skin::ApplyMirroring(int iBone,bool bDeltaAnim,Asura_Quat& xQuat,Asura_Vector_3& xPos) const
{
	if(m_piMirroredBone && m_piMirroredBone[iBone] == iBone)
	{
		//If the bone doesn't have mirrored equivalent , then only apply mirroring if all parent bones
		//also don't have mirrored versions. RAR 21/11/2006
		int iParent = ParentBone[iBone];
		while(iParent)
		{
			if(iParent != m_piMirroredBone[iParent]) return;
			iParent = ParentBone[iParent];
		}
	}


	xQuat.y = -xQuat.y;
	xQuat.z = -xQuat.z;
	xPos.x = -xPos.x;

}


/// End of file. ///
