//
// Asura_Hierarchy_BodyPart_System.cpp
//
// Created by Richard Rice - 9/10/02
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Animation/Asura_Hierarchy_BodyPart_System.h"

// Engine independent.
#include "Animation/Asura_Hierarchy.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_NameTree.h"
#include "Debugging/Asura_Assert.h"

////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_Hierarchy_BodyPart_System **
**---------------------------------*/
int Asura_Hierarchy_BodyPart_System::s_iNumberOfBodyTypes = 0;
Asura_Hierarchy_BodyType* Asura_Hierarchy_BodyPart_System::s_pxBodyTypes = 0;
Asura_NameTree* Asura_Hierarchy_BodyPart_System::s_pxAnimationTree = 0;

void Asura_Hierarchy_BodyPart_System::Initialise()
{
	s_iNumberOfBodyTypes = 0;
	s_pxBodyTypes = 0;
	s_pxAnimationTree = new Asura_NameTree;
}

void Asura_Hierarchy_BodyPart_System::Shutdown()
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(s_pxBodyTypes);
		s_pxBodyTypes = 0;

		delete(s_pxAnimationTree);
		s_pxAnimationTree = 0;
	}
	Asura_ResourcePopType();
}

void Asura_Hierarchy_BodyPart_System::Reset()
{
// (MH - 15.12.2002) attempting to avoid memory fragmentation
//	Shutdown();
//	Initialise();

	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(s_pxBodyTypes);
		s_pxBodyTypes = 0;
		s_iNumberOfBodyTypes = 0;

		s_pxAnimationTree->Reset();
	}
	Asura_ResourcePopType();
}

void Asura_Hierarchy_BodyPart_System::LoadAnimations()
{
	Asura_Assert(s_pxAnimationTree,"");
	s_pxAnimationTree->Traverse(LoadAnimation);
}

void Asura_Hierarchy_BodyPart_System::LoadAnimation( const Asura_String& xName, int /*iIndex*/ )
{
    Asura_ChunkLoading_System::LoadAsuraFile( xName.GetString() );
}


Asura_Hash_ID Asura_Hierarchy_BodyPart_System::GetDeathAnimation(const Asura_Hierarchy_Skin* pxSkin,const char* pszSectionName,enum AsuraDeathType eDeathType)
{
	Asura_Assert(pxSkin,"");
	Asura_Hash_ID uHashSkin = Asura_GetHashID(pxSkin->Name.GetString());
	return GetDeathAnimation(uHashSkin,pszSectionName,eDeathType);
}

Asura_Hash_ID Asura_Hierarchy_BodyPart_System::GetDeathAnimation(Asura_Hash_ID uHashSkin,const char* pszSectionName,enum AsuraDeathType eDeathType)
{
	Asura_Assert(pszSectionName,"");

	//find the body type for this skin
	const Asura_Hierarchy_BodyType* pxBodyType = GetBodyTypeForSkin(uHashSkin);
	if(!pxBodyType) return ASURA_HASH_ID_UNSET;

	//get the body part from the section name
	Asura_Hash_ID uHashSection = Asura_GetHashID(pszSectionName);
	const Asura_Hierarchy_BodyPart* pxBodyPart = pxBodyType->GetBodyPart(uHashSection);
	if(!pxBodyPart) return ASURA_HASH_ID_UNSET;

	//get the list of animations for this death type
	const Asura_Hierarchy_DeathType* pxDeathType = pxBodyPart->GetDeathType(eDeathType);
	if(!pxDeathType) return ASURA_HASH_ID_UNSET;

	
	//get a random anim
	return pxDeathType->PickAnimation();
}

int Asura_Hierarchy_BodyPart_System::GetBodyPartID(const class Asura_Hierarchy_Skin* pxSkin,const char* pszSectionName)
{
	Asura_Assert(pxSkin,"");
	Asura_Hash_ID uHashSkin = Asura_GetHashID(pxSkin->Name.GetString());
	return GetBodyPartID(uHashSkin,pszSectionName);
}

int Asura_Hierarchy_BodyPart_System::GetBodyPartID(Asura_Hash_ID uHashSkin,const char* pszSectionName)
{
	//find the body type for this skin
	const Asura_Hierarchy_BodyType* pxBodyType = GetBodyTypeForSkin(uHashSkin);
	if(!pxBodyType) return ASURA_BODYPART_INVALID;

	//get the body part from the section name
	Asura_Hash_ID uHashSection = Asura_GetHashID(pszSectionName);
	const Asura_Hierarchy_BodyPart* pxBodyPart = pxBodyType->GetBodyPart(uHashSection);
	if(!pxBodyPart) return ASURA_BODYPART_INVALID;

	return pxBodyPart->GetBodyPartType();
}

const Asura_Hierarchy_BodyType* Asura_Hierarchy_BodyPart_System::GetBodyTypeForSkin(Asura_Hash_ID uHashSkin)
{
	for(int i=0;i<s_iNumberOfBodyTypes;i++)
	{
		if(s_pxBodyTypes[i].UsedBySkin(uHashSkin)) return &s_pxBodyTypes[i];
	}
	return 0;
}

#ifdef ASURA_TOOLS
void Asura_Hierarchy_BodyPart_System::ListAnimationsForSkins(const List<Asura_Hash_ID>& xHashList,List<Asura_String>& xNameList)
{


	List<Asura_Hash_ID> xAnimHashList;
	
	//get the hash values for the anims use by each body type that contains a skin in the list
	for(int i=0;i<s_iNumberOfBodyTypes;i++)
	{
		bool bUsed = false;
		for(CLIF<Asura_Hash_ID> xHashLIF(&xHashList);!xHashLIF.done();xHashLIF.next())
		{
			bUsed = s_pxBodyTypes[i].UsedBySkin(xHashLIF());
			if(bUsed) break;
		}
		if(!bUsed) continue;

		s_pxBodyTypes[i].ListAnimations(xAnimHashList);

	}

	//get a list of all animation names
	int iNumberOfAnims = s_pxAnimationTree->GetNumberOfNodes();
	const Asura_String** ppxFileNames = new const Asura_String*[iNumberOfAnims];
	s_pxAnimationTree->MakeNameList(ppxFileNames);
	
	//ceck each animation to see if the has value matches one in the list
	for(int i=0;i<iNumberOfAnims;i++)
	{
		Asura_String xName;
		ppxFileNames[i]->SplitFilename(0,&xName,0);
		if(xAnimHashList.contains(Asura_GetHashID(xName.GetString())))
		{
			//add this filename to the output list
			xNameList.add_entry(*ppxFileNames[i]);
		}
	}

	
	delete [] ppxFileNames;
}
#endif


////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Asura_Hierarchy_BodyType **
**--------------------------*/

Asura_Hierarchy_BodyType::Asura_Hierarchy_BodyType()
{
	m_iNumberOfSkins = 0;
	m_puSkins = 0;

	m_iNumberOfBodyParts = 0;
	m_pxBodyParts = 0;
}

Asura_Hierarchy_BodyType::~Asura_Hierarchy_BodyType()
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(m_puSkins);
		Asura_ResourceDeallocateArray(m_pxBodyParts);
	}
	Asura_ResourcePopType();
}

const Asura_Hierarchy_BodyPart* Asura_Hierarchy_BodyType::GetBodyPart(Asura_Hash_ID uHashSection) const
{
	for(int i=0;i<m_iNumberOfBodyParts;i++)
	{
		if(m_pxBodyParts[i].ContainsSection(uHashSection))
		{
			return &m_pxBodyParts[i];
		}
	}
	return 0;
}

bool Asura_Hierarchy_BodyType::UsedBySkin(Asura_Hash_ID uHashSkin) const
{
	for(int i=0;i<m_iNumberOfSkins;i++)
	{
		if(m_puSkins[i] == uHashSkin) return true;
	}
	return false;
}

#ifdef ASURA_TOOLS
void Asura_Hierarchy_BodyType::ListAnimations(List<Asura_Hash_ID>& xAnimHashList) const
{
	for(int i=0;i<m_iNumberOfBodyParts;i++)
	{
		m_pxBodyParts[i].ListAnimations(xAnimHashList);
	}
}
#endif


////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Asura_Hierarchy_BodyPart **
**--------------------------*/

Asura_Hierarchy_BodyPart::Asura_Hierarchy_BodyPart()
{
	m_iNumberOfSections = 0;
	m_puSections = 0;
	m_iNumberOfDeathTypes = 0;
	m_pxDeathTypes = 0;
}

Asura_Hierarchy_BodyPart::~Asura_Hierarchy_BodyPart()
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(m_puSections);
		Asura_ResourceDeallocateArray(m_pxDeathTypes);
	}
	Asura_ResourcePopType();
}

const Asura_Hierarchy_DeathType* Asura_Hierarchy_BodyPart::GetDeathType(enum AsuraDeathType eDeathType) const
{
	for(int i=0;i<m_iNumberOfDeathTypes;i++)
	{
		if(m_pxDeathTypes[i].GetDeathType() == eDeathType)
		{
			return &m_pxDeathTypes[i];
		}
	}
	return 0;
}

bool Asura_Hierarchy_BodyPart::ContainsSection(Asura_Hash_ID uHashSection) const
{
	for(int i=0;i<m_iNumberOfSections;i++)
	{
		if(m_puSections[i] == uHashSection)
		{
			return true;
		}
	}
	return false;
}

#ifdef ASURA_TOOLS
void Asura_Hierarchy_BodyPart::ListAnimations(List<Asura_Hash_ID>& xAnimHashList) const
{
	for(int i=0;i<m_iNumberOfDeathTypes;i++)
	{
		m_pxDeathTypes[i].ListAnimations(xAnimHashList);
	}
}
#endif


////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Hierarchy_DeathType **
**---------------------------*/

Asura_Hierarchy_DeathType::Asura_Hierarchy_DeathType()
{
	m_iNumberOfAnimations = 0;
	m_puAnimations = 0;
}

Asura_Hierarchy_DeathType::~Asura_Hierarchy_DeathType()
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(m_puAnimations);
	}
	Asura_ResourcePopType();
}

Asura_Hash_ID Asura_Hierarchy_DeathType::PickAnimation() const
{
	if(!m_iNumberOfAnimations) return ASURA_HASH_ID_UNSET;
	
	return m_puAnimations[Asura_Random::GetIntU()%m_iNumberOfAnimations];
};


#ifdef ASURA_TOOLS
void Asura_Hierarchy_DeathType::ListAnimations(List<Asura_Hash_ID>& xAnimHashList) const
{
	for(int i=0;i<m_iNumberOfAnimations;i++)
	{
		if(!xAnimHashList.contains(m_puAnimations[i]))
		{
			xAnimHashList.add_entry(m_puAnimations[i]);
		}
	}
}
#endif
