
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Animation_IK_Solver.cpp
// Author : Will Plano
// Date   : 08/08/2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_ANIMATION_IK_SOLVER

#include "Core/Asura_Maths.h"

#include "Animation/IK/Asura_Animation_IK_Solver.h"

#include "Animation/Asura_Animation.h"
#include "Animation/IK/Asura_Animation_IK_Chain.h"
#include "Animation/IK/Asura_Animation_IK_Template_System.h"
#include "Animation/IK/Asura_Animation_IK_Template.h"
#include "Camera/Asura_Camera.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Interface/Asura_CommandConsole.h"

DVSVAR bool g_bEnableIK = true;
DVSVAR float g_fDefaultMinDist = ASURA_IK_FADE_NEAR_RANGE;
DVSVAR float g_fDefaultMaxDist = ASURA_IK_FADE_FAR_RANGE;
DVSVAR float g_fRootBoneOffsetSmoothing = 0.2f;

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_DEBUG_VARIABLE_SYSTEM)
void Asura_Animation_IK_Solver::AddConsoleVars()
{
	Asura_CommandConsole::AddVar( L"IK.Enable", &g_bEnableIK, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"IK.MinDist", &g_fDefaultMinDist, 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"IK.MaxDist", &g_fDefaultMaxDist, 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"IK.RootBoneOffsetSmoothing", &g_fRootBoneOffsetSmoothing, 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}
#endif

Asura_Animation_IK_Solver::Asura_Animation_IK_Solver()
: m_fLerp( 1.0f )
, m_fMinDist( g_fDefaultMinDist )
, m_fMaxDist( g_fDefaultMaxDist )
, m_pxFirstChain( NULL )
, m_xSmoothedPositionDelta( Asura_Maths::ZeroVector3 )
, m_uLastKnwonVisibilityNode( 0 )
, m_eRootBoneLerpState( ROOT_BONE_IN )
, m_fRootBoneLerp( 1.0f )
, m_fRootBoneLerpTime( 0.0f )
, m_fRootBoneMaxMovement( 1.0f )
{

}

Asura_Animation_IK_Solver::~Asura_Animation_IK_Solver()
{
	Clear();
}

void Asura_Animation_IK_Solver::Shutdown()
{
	m_xSmoothedPositionDelta = Asura_Maths::ZeroVector3;
	Clear();
}

Asura_Animation_IK_Solver* Asura_Animation_IK_Solver::CreateFromTemplate(Asura_Hash_ID uTemplate, const Asura_Animation* pxAnim)
{
	//make sure the template exists
	const Asura_Animation_IK_Template* pxTemplate = Asura_Animation_IK_Template_System::FindTemplate(uTemplate);
	if(!pxTemplate)
	{
		return 0;
	}

	//allocate and setup a new solver
	Asura_Animation_IK_Solver* pxSolver = new Asura_Animation_IK_Solver;
	pxTemplate->SetupIKSolver(pxSolver, pxAnim);

	return pxSolver;
	
}

void Asura_Animation_IK_Solver::AddChain( Asura_Animation_IK_Chain& xNewChain )
{
	if ( !m_pxFirstChain )
	{
		m_pxFirstChain = &xNewChain;
	}
	else
	{
		Asura_Animation_IK_Chain* pxChain = m_pxFirstChain;
		if ( pxChain == &xNewChain )
		{
			Asura_Assert( false, "Chain already added" );
			return;
		}
		while ( pxChain->GetNextChain() )
		{ 
			if ( pxChain->GetNextChain() == &xNewChain )
			{
				Asura_Assert( false, "Chain already added" );
				return;
			}
			pxChain = pxChain->GetNextChain();
		}
		pxChain->SetNextChain( &xNewChain );
	}
}

void Asura_Animation_IK_Solver::RemoveChain( Asura_Animation_IK_Chain& xChainToRemove )
{
	Asura_Assert(!xChainToRemove.IsOwnedBySolver(), "Not expecting IK chains created by templates to be removed in this way");

	if ( &xChainToRemove == m_pxFirstChain )
	{
		m_pxFirstChain = m_pxFirstChain->GetNextChain();
	}
	else
	{
		Asura_Animation_IK_Chain* pxChain = m_pxFirstChain;
		while ( pxChain && pxChain->GetNextChain() != &xChainToRemove )
		{ 
			pxChain = pxChain->GetNextChain();
		}
		Asura_Assert( pxChain, "Chain not found in list" );
		if ( pxChain )
		{
			pxChain->SetNextChain( xChainToRemove.GetNextChain() );
		}
	}
	xChainToRemove.SetNextChain( NULL );
}

void Asura_Animation_IK_Solver::Clear()
{
	//Clear out the list of IK chains, deleting those that are owned by the solver. RAR 12/05/2008
	Asura_Animation_IK_Chain* pxChain = m_pxFirstChain;
	while(pxChain)
	{
		Asura_Animation_IK_Chain* pxNextChain = pxChain->GetNextChain();
		if(pxChain->IsOwnedBySolver())
		{
			delete pxChain;
		}
		else
		{
			pxChain->Shutdown();
		}
		pxChain = pxNextChain;	
	}

	m_pxFirstChain = NULL;
}

void Asura_Animation_IK_Solver::EnableRootBoneMovement( const float fLerpTime )
{
	m_fRootBoneLerpTime = fLerpTime;
	if ( m_eRootBoneLerpState != ROOT_BONE_IN )
	{
		m_eRootBoneLerpState = ROOT_BONE_LERP_IN;
	}
}

void Asura_Animation_IK_Solver::DisableRootBoneMovement( const float fLerpTime )
{
	m_fRootBoneLerpTime = fLerpTime;
	if ( m_eRootBoneLerpState != ROOT_BONE_OUT )
	{
		m_eRootBoneLerpState = ROOT_BONE_LERP_OUT;
	}
}

void Asura_Animation_IK_Solver::UpdateBones( Asura_Animation& xAnim )
{
	UpdateRootBoneLerp();
	
	const float fLerp = m_fLerp * Asura_Maths::ClampToRange( 1.0f - ( ( xAnim.GetPosition() - *Asura_Camera::GetPosition() ).MagnitudeSquared() - ( m_fMinDist * m_fMinDist ) ) / ( ( m_fMaxDist * m_fMaxDist ) - ( m_fMinDist * m_fMinDist ) ), 0.0f, 1.0f );

	if ( fLerp > 0.0f && g_bEnableIK )
	{
		// Setup the chains from the bone data in the animation.
		for ( Asura_Animation_IK_Chain* pxChain = m_pxFirstChain; pxChain; pxChain = pxChain->GetNextChain() )
		{
			pxChain->GetBoneDataFromAnim( xAnim );
		}

		// Calculate how far we need to move the root bone down to get all chains marked as legs on the floor.
		float fMaxDist = -Asura_Maths::LargeFloat;
		for ( Asura_Animation_IK_Chain* pxChain = m_pxFirstChain; pxChain; pxChain = pxChain->GetNextChain() )
		{
			Asura_AVector3 xTargetPos;
			if ( pxChain->GetTarget( &xTargetPos, NULL, true ) )
			{
				if ( pxChain->IsActive() )
				{
					if ( pxChain->IsRootBoneMover() )
					{
						float fDist = xTargetPos.GetY() - pxChain->GetEndPosition().GetY();
						if ( fDist > fMaxDist )
						{
							fMaxDist = fDist;
						}
					}
				}
			}
			else
			{
				pxChain->Disable();
			}
		}	

		if ( m_xSmoothedPositionDelta.MagnitudeSquared() > 1.0e-10f || ( fMaxDist > 0.01f && fMaxDist < ( 2.0f * m_fRootBoneMaxMovement ) ) )
		{
			// if we are over 1cm above the ground but below 2 * m_fRootBoneMaxMovement move the root bone down.
			if ( fMaxDist > m_fRootBoneMaxMovement )
			{
				// if we are over m_fRootBoneMaxMovement above the ground lerp out the amount we move down by so that falling doesn't snap us onto the floor.
				fMaxDist = ( 2.0f * m_fRootBoneMaxMovement ) - fMaxDist;
			}
			fMaxDist = Asura_Maths::Max( fMaxDist, 0.0f );

			Asura_Vector_3 xPositionDelta = Asura_Vector_3( 0.0f, 1.0f, 0.0f ) * fMaxDist * fLerp * m_fRootBoneLerp;
			m_xSmoothedPositionDelta.Lerp( xPositionDelta, Asura_Maths::ClampToRange( g_fRootBoneOffsetSmoothing * Asura_Timers::GetGameTimeScale(), 0.0f, 1.0f ) );

			Asura_Vector_3 xRootBonePos;
			xAnim.GetTweenStore()->GetBoneData( 0 ).GetPosition( xRootBonePos );
			xRootBonePos += m_xSmoothedPositionDelta;
			xAnim.GetTweenStore()->GetBoneData( 0 ).SetPosition( xRootBonePos );

			xAnim.UpdateAttachmentLocations();

			for ( Asura_Animation_IK_Chain* pxChain = m_pxFirstChain; pxChain; pxChain = pxChain->GetNextChain() )
			{
				Asura_AVector3 xWSPosDelta( m_xSmoothedPositionDelta );
				xWSPosDelta = xAnim.GetOrientQuat().RotateVector( xWSPosDelta);
				pxChain->Enable();
				pxChain->MoveEverything( xWSPosDelta );
			}
		}

		// Loop over the chains and solve them. NB. it is important to input the chains in the right order if they have dependencies.
		// i.e. add hands before fingers.
		for ( Asura_Animation_IK_Chain* pxChain = m_pxFirstChain; pxChain; pxChain = pxChain->GetNextChain() )
		{
			if ( pxChain->GetTarget() )
			{
				if ( pxChain->IsActive() )
				{
					// Solve the chain.
					pxChain->Solve();
					// Update the bone data in the tween store.
					pxChain->PushBoneDataToAnim ( xAnim, fLerp );
				}
			}
			else
			{
				pxChain->Disable();
			}
		}
	}
}

void Asura_Animation_IK_Solver::UpdateRootBoneLerp()
{
	switch ( m_eRootBoneLerpState )
	{
		case ROOT_BONE_LERP_IN:
		{
			if ( m_fRootBoneLerpTime > 0.0f )
			{
				m_fRootBoneLerp += Asura_Timers::GetGameTimeSlice() / m_fRootBoneLerpTime;
				if ( m_fRootBoneLerp > 1.0f )
				{
					m_eRootBoneLerpState = ROOT_BONE_IN;
					m_fRootBoneLerp = 1.0f;
				}
			}
			else
			{
				m_eRootBoneLerpState = ROOT_BONE_IN;
				m_fRootBoneLerp = 1.0f;
			}
		}
		break;

		case ROOT_BONE_LERP_OUT:
		{
			if ( m_fRootBoneLerpTime > 0.0f )
			{
				m_fRootBoneLerp -= Asura_Timers::GetGameTimeSlice() / m_fRootBoneLerpTime;
				if ( m_fRootBoneLerp < 0.0f )
				{
					m_eRootBoneLerpState = ROOT_BONE_OUT;
					m_fRootBoneLerp = 0.0f;
				}
		}
			else
			{
				m_eRootBoneLerpState = ROOT_BONE_OUT;
				m_fRootBoneLerp = 0.0f;
			}
		}
		break;

		default:
			break;
	}
}

#endif // ASURA_USE_ANIMATION_IK_SOLVER
