//////////////////////////////////////////////////////////////////////////////
//
// Asura_Blueprint_System.cpp
//
// Created by NJD - 06 September 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Blueprint/Asura_Blueprint_System.h"

// asura
#include "Blueprint/Asura_Blueprint.h"
#include "Core/Asura_Hash.h"
#include "Core/Memory/Asura_Mem.h"
#include "ChunkSystem/Asura_Chunk_Blueprints.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"

// tools
#ifdef ASURA_TOOLS
#include "Blueprint/Tools_Mode_Blueprint.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Blueprint_System
//
//////////////////////////////////////////////////////////////////////////////

Asura_TemplateHashTree< Asura_BlueprintType > Asura_Blueprint_System::s_xBlueprintTypeContainer;
#ifdef ASURA_TOOLS
Asura_Blueprint_Template* Asura_Blueprint_System::s_pxTemplates = NULL;
#endif

Asura_BlueprintType::Asura_BlueprintType()
: m_xBlueprintHashTree()
#ifdef ASURA_TOOLS
, m_xTypeName()
, m_xFileName()
#endif
{
}

void Asura_Blueprint_System::Initialise()
{
    s_xBlueprintTypeContainer.Reset();
#ifdef ASURA_TOOLS
	s_pxTemplates = NULL;
	Project_InitialiseTemplates();

	Tools_Mode_Blueprint::LoadBlueprints();
#endif
}

void Asura_Blueprint_System::Shutdown()
{
	s_xBlueprintTypeContainer.Reset();
#ifdef ASURA_TOOLS
	// Delete template list.
	Asura_Blueprint_Template* pDel;
	Asura_Blueprint_Template* pTemp = s_pxTemplates;
	while (pTemp)
	{
		pDel = pTemp;
		pTemp = pTemp->m_pNext;
		delete pDel;
	}

    Asura_Parameter::ReleaseGroupNameStorage();
#endif
}

void Asura_Blueprint_System::Reset()
{
	Shutdown();
	Initialise();
}

Asura_Blueprint* Asura_Blueprint_System::GetBlueprintByHashID( const Asura_Hash_ID uTypeHashID, const Asura_Hash_ID uBlueprintHashID )
{
    if ( uTypeHashID == ASURA_HASH_ID_UNSET || uBlueprintHashID == ASURA_HASH_ID_UNSET )
    {
        return NULL;
    }

	for ( Asura_Blueprint_Type_Iterator xIterator; !xIterator.Done(); xIterator.Next() )
	{
        if ( xIterator.GetCurrentHash() == uTypeHashID )
        {
			Asura_BlueprintType* pxBlueprintType = xIterator.GetCurrent();
			return pxBlueprintType->m_xBlueprintHashTree.GetDataByHash( uBlueprintHashID );
        }
    }

    return NULL;
}

int Asura_Blueprint_System::GetNumTypes()
{
	return(s_xBlueprintTypeContainer.GetNumberOfNodes());
}

int Asura_Blueprint_System::GetNumBlueprintsInType( const Asura_Hash_ID uTypeHashID )
{
    Asura_BlueprintType* pxBlueprintType = s_xBlueprintTypeContainer.GetDataByHash( uTypeHashID );
	if ( pxBlueprintType == NULL )
	{
		Asura_Assert( false, "Asura_Blueprint_System::GetNumBlueprintsInType - bad type hash ID" );
		return 0;
	}
	else
	{
		return pxBlueprintType->m_xBlueprintHashTree.GetNumberOfNodes();
	}
}

int Asura_Blueprint_System::GetNumRootBlueprintsInType( const Asura_Hash_ID uTypeHashID )
{
	int iRootBlueprintCount = 0;
	for ( Asura_Blueprint_ConstIterator xBPIter( uTypeHashID ); !xBPIter.Done(); xBPIter.Next() )
	{
		if ( xBPIter.GetCurrent()->GetParentHashID() == ASURA_HASH_ID_UNSET )
		{
			++iRootBlueprintCount;
		}
	}
	return iRootBlueprintCount;
}

Asura_Blueprint* Asura_Blueprint_System::GetRootBlueprint( const Asura_Hash_ID uTypeHashID, const int iIndex )
{
	Asura_Assert( iIndex < GetNumBlueprintsInType( uTypeHashID ), "Asura_Blueprint_System::GetRootBlueprint - bad index" );

	int iRootBlueprintCount = 0;
	for ( Asura_Blueprint_Iterator xBPIter( uTypeHashID ); !xBPIter.Done(); xBPIter.Next() )
	{
		Asura_Blueprint* pxBP = xBPIter.GetCurrent();
		if ( pxBP->GetParentHashID() == ASURA_HASH_ID_UNSET )
		{
			if ( iRootBlueprintCount == iIndex )
			{
				return pxBP;
			}
			++iRootBlueprintCount;
		}
	}
	return NULL;
}

int Asura_Blueprint_System::GetNumChildBlueprints( const Asura_Hash_ID uTypeHashID, const Asura_Hash_ID uParentBlueprintHashID, const bool bOnlyDirectChildren )
{
	int iChildBlueprintCount = 0;
	for ( Asura_Blueprint_ConstIterator xBPIter( uTypeHashID ); !xBPIter.Done(); xBPIter.Next() )
	{
		if ( xBPIter.GetCurrent()->IsChildOf( uTypeHashID, uParentBlueprintHashID, bOnlyDirectChildren ) )
		{
			++iChildBlueprintCount;
		}
	}
	return iChildBlueprintCount;
}

Asura_Blueprint* Asura_Blueprint_System::GetChildBlueprint( const Asura_Hash_ID uTypeHashID, const Asura_Hash_ID uParentBlueprintHashID,
                                                            const bool bOnlyDirectChildren, const int iIndex )
{
	Asura_Assert( iIndex < GetNumBlueprintsInType( uTypeHashID ), "Asura_Blueprint_System::GetChildBlueprint - bad index" );
	
	int iChildBlueprintCount = 0;
	for ( Asura_Blueprint_Iterator xBPIter( uTypeHashID ); !xBPIter.Done(); xBPIter.Next() )
	{
		Asura_Blueprint* pxBP = xBPIter.GetCurrent();
		if ( pxBP->IsChildOf( uTypeHashID, uParentBlueprintHashID, bOnlyDirectChildren ) )
		{
			if ( iChildBlueprintCount == iIndex )
			{
				return pxBP;
			}
			++iChildBlueprintCount;
		}
	}
	return NULL;
}

void Asura_Blueprint_System::EstablishAllLinkage( const Asura_Hash_ID uTypeHashID )
{
	for ( Asura_Blueprint_Iterator xBPIter( uTypeHashID ); !xBPIter.Done(); xBPIter.Next() )
	{
		xBPIter.GetCurrent()->EstablishLinkage( uTypeHashID );
	}
}

#ifdef ASURA_TOOLS
void Asura_Blueprint_System::AddBlueprintType( const Asura_String& xTypeName )
{
	const Asura_Hash_ID	uTypeHashID	= xTypeName.GetHash();

    // We have already checked that this type doesn't already exist.
    Asura_BlueprintType* pxBlueprintType = s_xBlueprintTypeContainer.AddNode(uTypeHashID);
    if ( !pxBlueprintType ) { return; }

    pxBlueprintType->m_xTypeName   = xTypeName;
}

void Asura_Blueprint_System::DeleteBlueprintType( const Asura_String& xTypeName )
{
    const Asura_Hash_ID uTypeHashID = xTypeName.GetHash();

	s_xBlueprintTypeContainer.DeleteNodeByHash(uTypeHashID);
}

bool Asura_Blueprint_System::RenameBlueprintType( const Asura_String& xOldName, const Asura_String& xNewName )
{
	const Asura_Hash_ID	uOldTypeHashID	= xOldName.GetHash();
	const Asura_Hash_ID	uNewTypeHashID	= xNewName.GetHash();

	if(s_xBlueprintTypeContainer.ReplaceHash(uOldTypeHashID, uNewTypeHashID))
	{
		Asura_BlueprintType*	pxBlueprintType	= s_xBlueprintTypeContainer.GetDataByHash(uNewTypeHashID);

		if(pxBlueprintType == NULL)
		{
			return(false);
		}
		else
		{
			// rename
			pxBlueprintType->m_xTypeName   = xNewName;
			return true;
		}
	}
	else
	{
		return(false);
	}
}

bool Asura_Blueprint_System::SetBlueprintTypeFileName( const Asura_Hash_ID uTypeHashID, const Asura_String& xTypeFileName )
{
	Asura_BlueprintType*	pxBlueprintType = s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);
	
	if( pxBlueprintType == NULL )
	{
		return(false);
	}
	else
	{
		pxBlueprintType->m_xFileName = xTypeFileName;
		return true;
	}
}

void Asura_Blueprint_System::AddBlueprint( const Asura_Hash_ID uTypeHashID, const Asura_String& xBlueprintName, const Asura_Hash_ID uParentHashID /*=ASURA_HASH_ID_UNSET*/, const bool bEstablishLinkage /*=true*/ )
{
    // We have already checked we don't have a blueprint of this type with this name.
    if ( uTypeHashID == ASURA_HASH_ID_UNSET )
    {
        Asura_Assert( false, "Asura_Blueprint_System::AddBlueprint - cannot add a blueprint without a specific type" );
        return;
    }

    Asura_BlueprintType* pxBlueprintType = s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);
    if( pxBlueprintType != NULL )
    {
		const Asura_Hash_ID uBlueprintHashID = xBlueprintName.GetHash();
		Asura_Blueprint* pxBP = pxBlueprintType->m_xBlueprintHashTree.AddNode( uBlueprintHashID );
		if ( pxBP )
		{
			pxBP->SetHashID( uBlueprintHashID );
			pxBP->SetParentHashID( uParentHashID );
			pxBP->SetName( xBlueprintName );
		}

        // Establish all of the linkage again.
		if ( bEstablishLinkage )
		{
			EstablishAllLinkage( uTypeHashID );
		}
    }
}

void Asura_Blueprint_System::MarkBlueprintForDeletion( const Asura_Hash_ID uTypeHashID, const Asura_String& xBlueprintName )
{
    if ( uTypeHashID == ASURA_HASH_ID_UNSET )
    {
        Asura_Assert( false, "Asura_Blueprint_System::DeleteBlueprint - cannot delete a blueprint without a specific type" );
        return;
    }

	const Asura_Hash_ID uBlueprintHashID = Asura_GetHashID( xBlueprintName.GetString() );
	Asura_Blueprint* pxBP = GetBlueprintByHashID( uTypeHashID, uBlueprintHashID );
	if ( pxBP )
	{
		pxBP->MarkForDeletion();
	}
	else
	{
		Asura_Assertf( false, ( "Asura_Blueprint_System::DeleteBlueprint - could not find blueprint '%s'", xBlueprintName.GetString() ) );
	}
}

void Asura_Blueprint_System::DeleteAllMarkedBlueprints( const Asura_Hash_ID uTypeHashID )
{
    Asura_BlueprintType* pxBlueprintType = s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);
    if( pxBlueprintType != NULL )
    {
		for ( Asura_TemplateHashTree_Iterator< Asura_Blueprint > xBPIt( pxBlueprintType->m_xBlueprintHashTree ); !xBPIt.Done(); xBPIt.Next() )
		{
			Asura_Blueprint* pxBP = xBPIt.GetCurrent();
			if ( pxBP->IsToBeDeleted() )
			{
				pxBlueprintType->m_xBlueprintHashTree.DeleteNodeByHash( pxBP->GetHashID() );
			}
		}
    }
}

void Asura_Blueprint_System::WriteToChunkStream( Asura_Chunk_Stream& xStream, const Asura_Hash_ID uBlueprintTypeHashID )
{
    xStream << uCURRENT_CHUNK_VERSION;

    // First count the number of blueprint types we're going to output.
	int iNumBlueprintTypes;
	if ( uBlueprintTypeHashID == ASURA_HASH_ID_UNSET )
	{
		iNumBlueprintTypes = s_xBlueprintTypeContainer.GetNumberOfNodes();
	}
	else
	{
		iNumBlueprintTypes = 1;
	}

    // Output the number of blueprint types we're going to write.
    xStream << iNumBlueprintTypes;

    // Output the blueprint type lists themselves.
	for ( Asura_Blueprint_Type_ConstIterator xIterator; !xIterator.Done(); xIterator.Next() )
    {
		const Asura_Hash_ID	uTypeHashID	= xIterator.GetCurrentHash();

        if ( uBlueprintTypeHashID == ASURA_HASH_ID_UNSET || uTypeHashID == uBlueprintTypeHashID )
        {
			const Asura_BlueprintType*	pxBlueprintType	= xIterator.GetCurrent();

			xStream << uTypeHashID;
			xStream << pxBlueprintType->m_xTypeName;

			const int iNumBlueprints = pxBlueprintType->m_xBlueprintHashTree.GetNumberOfNodes();
			xStream << iNumBlueprints;
			int iBadBlueprintCount = 0;
			for ( Asura_TemplateHashTree_SaveIterator xIt( iNumBlueprints ); !xIt.Done(); xIt.Next() )
			{
				Asura_Blueprint* pxBP = pxBlueprintType->m_xBlueprintHashTree.GetDataByIndex( xIt.GetCurrent() );
				if ( pxBP )
				{
					xStream << pxBP->GetHashID();
					pxBP->WriteToChunkStream( xStream );
				}
				else
				{
					Asura_Assert( false, "Asura_Blueprint_System::WriteToChunkStream - cannot get blueprint" );

					// Write out a blank blueprint so that the chunk data doesn't get messed up.
					Asura_String xBadBlueprintName;
					Asura_StringUtil::ToString( iBadBlueprintCount++, xBadBlueprintName );
					xBadBlueprintName += " Bad Blueprint - please delete me.";

					Asura_Blueprint xBlueprint;
					xBlueprint.SetName( xBadBlueprintName );
					xBlueprint.SetHashID( Asura_GetHashID( xBadBlueprintName.GetString() ) );
					xStream << xBlueprint.GetHashID();
					xBlueprint.WriteToChunkStream( xStream );
				}
			}
        }
    }
}

const Asura_String* Asura_Blueprint_System::GetTypeName( Asura_Hash_ID uTypeHashID )
{
	const Asura_BlueprintType* pxBlueprintType	= s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);

	if(pxBlueprintType == NULL)
	{
		return(NULL);
	}
	else
	{
		return &pxBlueprintType->m_xTypeName;
	}
}

const Asura_String* Asura_Blueprint_System::GetTypeFileName( Asura_Hash_ID uTypeHashID )
{
	const Asura_BlueprintType* pxBlueprintType	= s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);

	if(pxBlueprintType == NULL)
	{
		return(NULL);
	}
	else
	{
		return &pxBlueprintType->m_xFileName;
	}
}

void Asura_Blueprint_System::StartExport()
{
	// Probably doesn't need to do anything here...
}

void Asura_Blueprint_System::EndExport()
{
	for ( Asura_Blueprint_Iterator xBPIter; !xBPIter.Done(); xBPIter.Next() )
	{
		xBPIter.GetCurrent()->ResetExportFlags();
	}
}

#endif //ASURA_TOOLS

void Asura_Blueprint_System::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
    u_int uVersion;
    xStream >> uVersion;
    if ( uVersion == uCURRENT_CHUNK_VERSION )
	{
        int iNumBlueprintTypes;
        xStream >> iNumBlueprintTypes;

        for ( int iListIndex = 0; iListIndex < iNumBlueprintTypes; ++iListIndex )
        {
			Asura_Hash_ID	uTypeHashID;
#if defined(ASURA_TOOLS)
			Asura_String	xTypeName;
			Asura_String	xFilename;
#endif	//defined(ASURA_TOOLS)

            xStream >> uTypeHashID;
#if defined(ASURA_TOOLS)
			xStream >> xTypeName;
			Asura_Chunk_Blueprints::GetCurrentBlueprintFileName( xFilename );
#else
			Asura_String xDummyString;
			xStream >> xDummyString;
#endif	//defined(ASURA_TOOLS)

			const Asura_BlueprintType*	pxExistingType	= s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);
			Asura_BlueprintType*		pxBlueprintType	= (pxExistingType == NULL)
														? s_xBlueprintTypeContainer.AddNode(uTypeHashID)
														: new Asura_BlueprintType;

			if(pxBlueprintType == NULL)
			{
				return;
			}

#ifdef ASURA_TOOLS
            pxBlueprintType->m_xTypeName	= xTypeName;
			pxBlueprintType->m_xFileName	= xFilename;
#endif

            // Read the blueprints in.
			int iNumBlueprints;
			xStream >> iNumBlueprints;
			for ( int i = 0; i < iNumBlueprints; ++i )
			{
				Asura_Hash_ID uBPHashID;
				xStream >> uBPHashID;
				Asura_Blueprint* pxBP = pxBlueprintType->m_xBlueprintHashTree.AddNode( uBPHashID );
				if ( pxBP )
				{
					pxBP->ReadFromChunkStream( xStream );
				} 
				else
				{
					// Need to read into a dummy blueprint to stop the chunk getting messed up.
					Asura_Blueprint xDummy;
					xDummy.ReadFromChunkStream( xStream );
				}
			}			

			// Check for loading duplicate blueprints.
			if ( pxExistingType )
			{
#ifdef ASURA_TOOLS
				Asura_Assertf( false,
					( "Asura_Blueprint_System::ReadFromChunkStream - already read type '%s' from file '%s', now found in file '%s'",
					pxExistingType->m_xTypeName.GetString(), pxExistingType->m_xFileName.GetString(), pxBlueprintType->m_xFileName.GetString() ) );
#else
				Asura_Assertf( false, ( "Asura_Blueprint_System::ReadFromChunkStream - already read type %u", uTypeHashID ) );
#endif

				delete pxBlueprintType;
			}
			else
			{
				// Make sure all the blueprints know who their parent is.
				EstablishAllLinkage( uTypeHashID );
			}
        }
	}
	else if ( uVersion == 0 )
	{
		int iNumBlueprintTypes;
		xStream >> iNumBlueprintTypes;

		for ( int iListIndex = 0; iListIndex < iNumBlueprintTypes; ++iListIndex )
		{
			int				iNumBlueprints;
			Asura_Hash_ID	uTypeHashID;
#if defined(ASURA_TOOLS)
			Asura_String	xTypeName;
			Asura_String	xFilename;
#endif	//defined(ASURA_TOOLS)

			xStream >> uTypeHashID;
			xStream >> iNumBlueprints;
#if defined(ASURA_TOOLS)
			xStream >> xTypeName;
			Asura_Chunk_Blueprints::GetCurrentBlueprintFileName( xFilename );
#else
			Asura_String xDummyString;
			xStream >> xDummyString;
#endif	//defined(ASURA_TOOLS)

			const Asura_BlueprintType*	pxExistingType	= s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID);
			Asura_BlueprintType*		pxBlueprintType = (pxExistingType == NULL)
														? s_xBlueprintTypeContainer.AddNode(uTypeHashID)
														: new Asura_BlueprintType;

			if(pxBlueprintType == NULL)
			{
				return;
			}

#ifdef ASURA_TOOLS
			pxBlueprintType->m_xTypeName	= xTypeName;
			pxBlueprintType->m_xFileName	= xFilename;
#endif

			// Read the blueprints in.
			for ( int i = 0; i < iNumBlueprints; ++i )
			{
				Asura_Blueprint xWorkingBlueprint;
				xWorkingBlueprint.ReadFromChunkStream( xStream );
				pxBlueprintType->m_xBlueprintHashTree.AddNode( xWorkingBlueprint.GetHashID(), xWorkingBlueprint );
			}			

			// Check for loading duplicate blueprints.
			if ( pxExistingType )
			{
#ifdef ASURA_TOOLS
				Asura_Assertf( false,
					( "Asura_Blueprint_System::ReadFromChunkStream - already read type '%s' from file '%s', now found in file '%s'",
					pxExistingType->m_xTypeName.GetString(), pxExistingType->m_xFileName.GetString(), pxBlueprintType->m_xFileName.GetString() ) );
#else
				Asura_Assertf( false, ( "Asura_Blueprint_System::ReadFromChunkStream - already read type %u", uTypeHashID ) );
#endif

				delete pxBlueprintType;
			}
			else
			{
				// Make sure all the blueprints know who their parent is.
				EstablishAllLinkage( uTypeHashID );
			}
		}
	}
    else
    {
        Asura_Assertf( false, ( "Asura_Blueprint_System::ReadFromChunkStream - unhandled chunk version (%u)", uVersion ) );
    }

#ifdef ASURA_TOOLS
    Asura_Parameter::BuildGroupNameTreeAndList();
#endif
}

#ifdef ASURA_TOOLS
Asura_Blueprint_Template* Asura_Blueprint_System::RegisterTemplate( Asura_Char* pszTemplateName, Asura_Hash_ID uTemplateHash, Asura_Value::VALUE_TYPE eType ) 
{
	// Add a project defined template to our lists.

	Asura_Blueprint_Template* pTemp = new Asura_Blueprint_Template();
	pTemp->m_pNext = s_pxTemplates;
	s_pxTemplates = pTemp;

	pTemp->m_xName = pszTemplateName;
	pTemp->m_uHash = uTemplateHash;
	pTemp->m_eType = eType;

	return pTemp;
}

const int Asura_Blueprint_System::GetTemplateList( ListSelection_Overlay& xList, Asura_Value::VALUE_TYPE eType ) 
{
	Asura_Blueprint_Template* pTemp = s_pxTemplates;
	int iCount = 0;

	while(pTemp)
	{
		if(pTemp->m_eType == eType)
		{
			xList.AddItem(pTemp->m_xName.GetString(), static_cast<int>(pTemp->m_uHash) );
			iCount++;
		}
		pTemp = pTemp->m_pNext;
	}
	return iCount;
}

Asura_Blueprint_Template* Asura_Blueprint_System::GetTemplate( const Asura_Hash_ID uBlueprintTemplateHashID ) 
{
	Asura_Blueprint_Template* pTemp = s_pxTemplates;

	while(pTemp)
	{
		if(pTemp->m_uHash == uBlueprintTemplateHashID)
		{
			return pTemp;
		}
		pTemp = pTemp->m_pNext;
	}
	return NULL;
}
#endif

Asura_BlueprintType* Asura_Blueprint_System::GetType( const Asura_Hash_ID uTypeHashID )
{
	return(s_xBlueprintTypeContainer.GetDataByHash(uTypeHashID));
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
