
/**
*
* \file Asura_Botanicals_InstanceManager.cpp
*
* Created by Dave Reed - 10/11/2005
*
* Copyright (C) 2005 Rebellion (Derby), All Rights Reserved.
*
* When not in Botanicals mode, all the botanical instances are stored by this 
* instance manager. 
* 
* This handles loading/saving the instances
*
* There's also a horribly inefficient render function, that just draws everything using
* Shape_Instances. It's just a placeholder so we can see this stuff when in other edit modes. 
*
* This module should be completely replaced when the proper botanicals system goes in -
* this version just stores a flat list of all instanes, without any batching by type/location,
* and without any LOD...
*
**/

#include "Core/Asura.h"

#ifdef ASURA_USE_BOTANICALS

#include "Core/Asura_Hash.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_CPUProfile.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#include "Shapes/Asura_Shape_System.h"
#include "Rendering/Asura_Render.h"

#include "Botanicals/Asura_Botanicals_TemplateManager.h"
#include "Botanicals/Asura_Botanicals_InstanceManager.h"

#ifdef ASURA_TOOLS

#include "Shapes/Asura_Shapes.h"
#include "Shapes/Asura_Shape_System.h"

#include "Camera/Asura_Camera.h"
#include "Tools_System.h"
#include "Shapes/Tools_Shapes.h"
#include "Objects/Tools_Object_System.h"
#include "Botanicals/Tools_Botanicals_Object.h"
#include "Botanical/Asura_Direct3D_Botanicals_Game_Render.h"
#include "Botanical/Asura_Direct3D_Botanicals_Game_Render_Manager.h"
#include "BotanicalCreation/Tools_BotanicalCreation_BotanicalShape.h"
#include "Objects/Tools_Object.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Misc/Tools_LOS.h"
#include "Mesh/Asura_Mesh.h"

#include "Misc/Tools_Render_Options.h"
#include "Rendering/Asura_Module.h"

Asura_LinkList<Asura_Botanicals_Instance*> Asura_Botanicals_InstanceManager::s_xInstanceList;
Asura_Vector_3 Asura_Botanicals_InstanceManager::s_xCamPos=Asura_Vector_3(0,0,0);

u_int Asura_Botanicals_InstanceManager::s_uNextInstanceID = 1;
int Asura_Botanicals_InstanceManager::s_iModuleIndex = -1;
bool (*Asura_Botanicals_InstanceManager::s_pfnModuleFilterCallback)(int iModule);

#endif // #ifdef ASURA_TOOLS

float Asura_Botanicals_InstanceManager::s_fFade = 1.0f;
float Asura_Botanicals_InstanceManager::s_fDefaultHorizFoV = 1.0f;

void Asura_Botanicals_InstanceManager::Initialise()
{
	Platform_Initialise();
}

void Asura_Botanicals_InstanceManager::Shutdown()
{
	Platform_Shutdown();
}

void Asura_Botanicals_InstanceManager::Render()
{
	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_BOTANICALS))
	{
		return;
	}

#if 0 && defined( ASURA_TOOLS)
	// We don't have support for wireframe botanical rendering
	if (Tools_Render_Options::GetWireframeMode()==TOOLS_WIREFRAME_ON)
	{
		return;
	}
#endif

	ASURA_CPU_PROFILE__START_LABEL( ASURA_CPU_PROFILE__BOTANICALS_RENDER, "Botanicals Render" );

	const float fFade = GetFade();
	if ( fFade > 0.0f )
	{
		Platform_Render( fFade );
	}

	ASURA_CPU_PROFILE__STOP( ASURA_CPU_PROFILE__BOTANICALS_RENDER );

}

void Asura_Botanicals_InstanceManager::Reset()
{
#ifdef ASURA_TOOLS
	// Clear the instance list
	 for (Asura_ListElt<Asura_Botanicals_Instance*> *pxElt = s_xInstanceList.Front();
		  pxElt; pxElt = pxElt->Next())
	 {
		 Asura_Botanicals_Instance *pxInst=pxElt->Data();
		 
		 if (pxInst->pxShapeInstance)
		 {
			 delete pxInst->pxShapeInstance;
		 }
		 
		 delete pxInst;
	 }

	 s_xInstanceList.Clear();

	s_uNextInstanceID = 1;
#endif // #ifdef ASURA_TOOLS
	Platform_Reset();
}

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
void Asura_Botanicals_InstanceManager::AddConsoleCommands( void )
{
	Asura_CommandConsole::AddCmd( L"BotanicalsFade", ConsoleCmd_SetFade, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );	
}

bool Asura_Botanicals_InstanceManager::ConsoleCmd_SetFade( float fFade, bool bInterrogate, bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( L"Sets the fade value 0.0->1.0" );
		Asura_CommandConsole::Printf( L"Only 0.0 & 1.0 currently supported" );
	}
	else if ( bInterrogate )
	{
		const float fCurrentFade = GetFade();
		Asura_CommandConsole::Printf( L"Current fade is %f", Asura_Maths::FtoD( fCurrentFade ) );
	}
	else
	{
		if ( ( fFade >= 0.0f ) && ( fFade <= 1.0f ) )
		{
			SetFade( fFade );
			Asura_CommandConsole::Printf( L"Fade set %f", Asura_Maths::FtoD( fFade ) );
			ConsoleCmd_SetFade( 0.0f, true, false );
			return true;
		}
		else
		{
			Asura_CommandConsole::Printf( L"failed to set fade to %f, allowed range is 0.0f -> 1.0f", Asura_Maths::FtoD( fFade ) );
			return false;
		}
	}
	return true;
}
#endif // #if defined(ASURA_SUPPORT_COMMAND_CONSOLE)

#ifdef ASURA_TOOLS

int  Asura_Botanicals_InstanceManager::GetNumInstances()
{
	int iNumInstances=0;

	for (Asura_ListElt<Asura_Botanicals_Instance*> *pxInst = s_xInstanceList.Front();
		pxInst; pxInst = pxInst->Next())
	{
		iNumInstances++;
	}

	return iNumInstances;
}


int  Asura_Botanicals_InstanceManager::GetNumInstancesOfTemplate(int iTemplate)
{
	int iNumInstances=0;

	for (Asura_ListElt<Asura_Botanicals_Instance*> *pxInst = s_xInstanceList.Front();
		pxInst; pxInst = pxInst->Next())
	{
		if (pxInst->Data()->iTemplateIdx == iTemplate)
		{		
			iNumInstances++;
		}
	}

	return iNumInstances;
}


int  Asura_Botanicals_InstanceManager::GetNumInstancesOfTemplateByUID(int iTemplateUID)
{
	int iNumInstances=0;

	for (Asura_ListElt<Asura_Botanicals_Instance*> *pxInst = s_xInstanceList.Front();
		pxInst; pxInst = pxInst->Next())
	{
		if (pxInst->Data()->iTemplateUID == iTemplateUID)
		{		
			iNumInstances++;
		}
	}

	return iNumInstances;
}


void Asura_Botanicals_InstanceManager::RemapTemplates()
{
	// For everything in the instance list, find it's template index.
	// If the template no longer exists, delete the instance
}

void Asura_Botanicals_InstanceManager::UpdateToolsObjectSet(Asura_Vector_3 *pxCamPos, float fRange, bool bForce)
{
	// If the camera has moved far enough, do an update

	Asura_Vector_3 vMoved=s_xCamPos-*pxCamPos;

	if (vMoved.Magnitude()>fRange*0.2f || bForce)
	{	
		s_xCamPos=*pxCamPos;
		ConvertFromToolObjects(true, pxCamPos, fRange);
		ConvertToToolObjects(true, pxCamPos, fRange);
	}
}


void Asura_Botanicals_InstanceManager::ConvertFromToolObjects(bool bLimitRange, Asura_Vector_3 *pxCamPos, float fRange)
{
	Asura_LinkList<Tools_Botanicals_Object*> xToolsObjsToRemove;

	// Deselect all tool objects, or it'll crash if we delete selected objects!
	Tools_Object_System::DeselectAllObjects();

	// Find all botanicals tool_objects, and convert them to Asura_Botanicals_Instances
	for(LIF<Tools_Object*> olif(&Tools_Object_System::ObjectList);!olif.done();olif.next())
	{
		if (olif()->Type == TOT_Botanicals)
		{	
			bool bConvert=true;
			Tools_Botanicals_Object *pxToolsObject = (Tools_Botanicals_Object *)olif();
			
			// Locate the current template index from the template ID
			int iIdx=Asura_Botanicals_TemplateManager::GetTemplateIndexFromUID(pxToolsObject->m_iTemplateID);

			// If the template no longer exists, we can't create the instance
			// (maybe we need a default template?)

			if (bLimitRange)
			{
				float fDistSquared= (pxToolsObject->Position-*pxCamPos).MagnitudeSquared();

				if (fDistSquared<=fRange*fRange)
				{			
					bConvert=false;
				}
			}

			if (bConvert)
			{		
				if (iIdx>=0)
				{
					Asura_Botanicals_Instance *pxInst=new Asura_Botanicals_Instance();
				
					// Store the instance params
					pxInst->xOrientation	= pxToolsObject->Orientation;
					pxInst->xPosition		= pxToolsObject->Position;
					pxInst->fScale			= pxToolsObject->GetScale();				
					pxInst->fBrightness     = pxToolsObject->GetBrightness();
					pxInst->iTemplateUID	= pxToolsObject->m_iTemplateID;									
					pxInst->iTemplateIdx	= iIdx;
					pxInst->uFlags			= pxToolsObject->GetBotanicalFlags();
					pxInst->uInstanceID		= pxToolsObject->GetInstanceID();
					
					pxInst->pxShapeInstance = NULL;

					if( 0 && pxToolsObject->m_iModuleIdx < 0)
					{
						pxInst->iModuleIdx = Asura_Botanicals_InstanceManager::GetModuleIndexFromPosition(pxToolsObject->Position, pxToolsObject->Orientation);
					}
					else
					{
						pxInst->iModuleIdx = pxToolsObject->m_iModuleIdx;
					}

					// Add it to the list
					s_xInstanceList.Push_Back(pxInst);
				}

				xToolsObjsToRemove.Push_Back(pxToolsObject);
			}
		}
	}


	// Remove the toolsobjects
	for (Asura_ListElt<Tools_Botanicals_Object*> *pxInst = xToolsObjsToRemove.Front();
		pxInst; pxInst = pxInst->Next())
	{
		delete pxInst->Data();
	}

	xToolsObjsToRemove.Clear();
}

void Asura_Botanicals_InstanceManager::ConvertToToolObjects(bool bLimitRange, Asura_Vector_3 *pxCamPos, float fRange)
{
	Asura_LinkList<Asura_Botanicals_Instance*> xInstsToRemove;

	// Create tool_objects from everything in the instance list	
	for (Asura_ListElt<Asura_Botanicals_Instance*> *pxListElt = s_xInstanceList.Front();
		pxListElt; pxListElt = pxListElt->Next())
	{
		Asura_Botanicals_Instance *pxInst=pxListElt->Data();
		bool bConvert=true;

		// Find the template from it's UID
		int iIdx=Asura_Botanicals_TemplateManager::GetTemplateIndexFromUID(pxInst->iTemplateUID);

		if (bLimitRange)
		{
			float fDistSquared= (pxInst->xPosition-*pxCamPos).MagnitudeSquared();

			if (fDistSquared>fRange*fRange)
			{			
				bConvert=false;
			}
		}

		if (bConvert)
		{		
			// If the template exists, create the toolsobject
			if (iIdx>=0)
			{		
				Tools_Botanicals_Object *pxToolsObject = new Tools_Botanicals_Object(pxInst->xPosition,iIdx,pxInst->uInstanceID);

				pxToolsObject->Orientation				= pxInst->xOrientation;
				pxToolsObject->m_fInstanceScale			= pxInst->fScale;
				pxToolsObject->m_fInstanceBrightness	= pxInst->fBrightness;
				pxToolsObject->m_iModuleIdx				= pxInst->iModuleIdx;
				pxToolsObject->SetBotanicalFlags(pxInst->uFlags);
				
			}			

			xInstsToRemove.Push_Back(pxInst);
		}
	}

// Then clear the list	
//	Reset();

	// Remove any deleted instances
	for (Asura_ListElt<Asura_Botanicals_Instance*> *pxInst = xInstsToRemove.Front();pxInst; pxInst = pxInst->Next())
	{
		Asura_Botanicals_Instance* pxBotanicalInstance = pxInst->Data();
		
		s_xInstanceList.Remove( s_xInstanceList.Find( pxBotanicalInstance));
		
		delete pxBotanicalInstance;
	}
	
#if defined( ASURA_BOTANICALS_BGR_IN_EDIT)
	Asura_Botanicals_InstanceManager::ConnectToolObjectsToBGR();
#endif
}

void Asura_Botanicals_InstanceManager::TestModuleTreeRayCallback( const u_int uIndex )
{
	// Should it be the 1st module or the last module that it hits?
	// Should we choose the bbox the point is closest to?
	// This is only called if the ray didn't actually hit any polygons e.g. an LoS test

	// Take the 1st module!
	if ( s_iModuleIndex == -1 )
	{
		s_iModuleIndex = uIndex;
	}
}

int Asura_Botanicals_InstanceManager::GetModuleIndexFromPosition( const Asura_Vector_3& xPos , const Asura_Matrix_3x3 &xOrientation , bool bSmallObject)
{
	Asura_Vector_3 xDirection( 0.0, +1.0f, 0.0f );

	// Move upwards 10cm so we find the ground
	Asura_Vector_3 xCentre( xPos );
	xCentre.y -= 0.1f;

	int iModuleBelow = -1;
	int iPolyBelow   = 0;
	float fDistBelow = 0;

	// Use the up vector of the instance
	for(int i=0;i<3;i++)
	{	
		xDirection[i]=xOrientation[i][1];
	}

	// There shouldn't be a scale in xOrientation, but do this just in case
	xDirection.Normalise();
	
	// Cast downwards ray, from 2.5m above the botanical origin, using the botanical orientation
	Tools_LOS::SetPolygonFlagsToIgnore( AsuraPolygonFlag_DontRender );

	if (bSmallObject)
	{
		Tools_LOS::SetModuleFilterCallback(s_pfnModuleFilterCallback);
		Tools_LOS::FindPolyInLOS(xCentre-(xDirection*0.25f),  xDirection, iPolyBelow, iModuleBelow, fDistBelow, true );	
		Tools_LOS::SetModuleFilterCallback(0);

		if ( iModuleBelow != -1 && fDistBelow<0.5f)
		{
			return iModuleBelow;
		}
	}	
	else
	{
		Tools_LOS::SetModuleFilterCallback(s_pfnModuleFilterCallback);
		Tools_LOS::FindPolyInLOS(xCentre-(xDirection*2.5f),  xDirection, iPolyBelow, iModuleBelow, fDistBelow, true );	
		Tools_LOS::SetModuleFilterCallback(0);

		Tools_LOS::SetPolygonFlagsToIgnore( 0 );

		if ( iModuleBelow != -1 && fDistBelow<5.0f)
		{
			return iModuleBelow;
		}
	}
	
	


	if (!bSmallObject)
	{	
		// Cast downwards ray, from 2.5m the botanical origin, using the world up vector.
		// This is really a special case for canopy instances, where using the orientation my put it in the
		// wrong module, as it's floating high above the collision mesh
		xDirection = Asura_Vector_3( 0.0, -1.0f, 0.0f );
		iModuleBelow = -1;
		Tools_LOS::SetPolygonFlagsToIgnore( AsuraPolygonFlag_DontRender );
		Tools_LOS::SetModuleFilterCallback(s_pfnModuleFilterCallback);
		Tools_LOS::FindPolyInLOS(xCentre,  xDirection, iPolyBelow, iModuleBelow, fDistBelow, true );	
		Tools_LOS::SetModuleFilterCallback(0);
		Tools_LOS::SetPolygonFlagsToIgnore( 0 );
		
		if ( iModuleBelow != -1)
		{
			return iModuleBelow;
		}
	}

	// Note the module linking needs to be loaded for these functions to work
	// An LoS ray failed so just find the 1st module under the instance
	int iModuleIndex = Asura_Module_System::GetModuleIndexFromPosition( xPos );
	if ( iModuleIndex < 0 )
	{
		// So not within a module bbox
		// Cast a ray downwards to see the 1st module that it hits and put it in that module
		const Asura_Vector_3 xEndPos = xPos + xDirection * 10000.0f;
		Asura_Bounding_Box xRayBB;
		xRayBB.Set( xPos );
		xRayBB.Set( xEndPos );
		s_iModuleIndex = -1;
		Asura_Module_System::TestModuleTree( xRayBB, TestModuleTreeRayCallback );
		iModuleIndex = s_iModuleIndex;
	}
	return iModuleIndex;
}


void Asura_Botanicals_InstanceManager::ResetModuleLinkingData(void)
{
	// Setting the module index to -2 flags it as needing recomputing next time the instance is rendered
	for (Asura_ListElt<Asura_Botanicals_Instance*> *pxElt = s_xInstanceList.Front(); pxElt; pxElt = pxElt->Next())
	{
		Asura_Botanicals_Instance *pxInst=pxElt->Data();
		pxInst->iModuleIdx = -2;		
	}
}

#if defined( ASURA_BOTANICALS_BGR_IN_EDIT)
void Asura_Botanicals_InstanceManager::ConnectToolObjectsToBGR()
{
	Asura_Direct3D_BGR_Manager& xBGRManager = Asura_Direct3D_Botanicals_Game_Render::GetDirect3DBGRManager();
	
	// Find all botanicals tool_objects, and connect them to BGR Instances
	for(LIF<Tools_Object*> olif(&Tools_Object_System::ObjectList);!olif.done();olif.next())
	{
		Tools_Botanicals_Object *pxToolsObject = static_cast<Tools_Botanicals_Object *>( olif());
		if( TOT_Botanicals != pxToolsObject->Type)
		{
			continue;	
		}
		
		// DLH - 12 October 2006 18:10:22 - Find the BGR Instance
		xBGRManager.FindExistingInstance( pxToolsObject);
	}
}
#endif

#if defined( ASURA_BOTANICALS_BGR_IN_EDIT)
void Asura_Botanicals_InstanceManager::ConnectToolObjectToBGR( Tools_Botanicals_Object *pxToolsObject)
{
	Asura_Direct3D_BGR_Manager& xBGRManager = Asura_Direct3D_Botanicals_Game_Render::GetDirect3DBGRManager();
	
	// DLH - 12 October 2006 18:10:22 - Find the BGR Instance
	xBGRManager.FindExistingInstance( pxToolsObject);
}
#endif

#if defined( ASURA_BOTANICALS_BGR_IN_EDIT)
void Asura_Botanicals_InstanceManager::CreateToolObjectInBGR( Tools_Botanicals_Object *pxToolsObject)
{
	Asura_Direct3D_BGR_Manager& xBGRManager = Asura_Direct3D_Botanicals_Game_Render::GetDirect3DBGRManager();
	
	xBGRManager.CreateToolObjectInBGR( pxToolsObject);
}
#endif




#endif // ASURA_TOOLS

#endif //ASURA_USE_BOTANICALS
