//////////////////////////////////////////////////////////////////////////////
//
// Asura_FlyCam.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     Basic camera controlled by analogue sticks or buttons. Intended only
//     for debugging and prototyping purposes. 
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_FLYCAM

// This file's header.
#include "Camera/Asura_FlyCam.h"

// Asura includes
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#if defined(ASURA_USE_CONTROLMAP_MANAGER)
#include "Input/Asura_ControlManager.h"
#elif defined(ASURA_SUPPORT_COMMAND_CONSOLE)
#include "Interface/Asura_CommandConsole.h"
#endif

DEBUGVAR float dbg_fGimbalLimit = Asura_Maths::DegToRad( 85.0f );

// S T A T  I C   D E C L A R A T  I O N S ///////////////////////////////////
Asura_ControlMap	Asura_FlyCam::s_xControlMap(MAX_MAPPINGS);

bool		Asura_FlyCam::s_bInvertY				= false;
Asura_Angle	Asura_FlyCam::s_fCameraTurnSpeed		= 2.0f;
float		Asura_FlyCam::s_fCameraMoveSpeedSlow	= 2.0f;
float		Asura_FlyCam::s_fCameraMoveSpeedNormal	= 5.0f;
float		Asura_FlyCam::s_fCameraMoveSpeedFast	= 20.0f;


// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::Initialise
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::Initialise(void)
{
	Platform_Initialise();
}

/// End of function 'Asura_FlyCam::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::Shutdown
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::Shutdown(void)
{
	Platform_Shutdown();
}

/// End of function 'Asura_FlyCam::Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::Update
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::Update(void)
{
#if defined(ASURA_USE_CONTROLMAP_MANAGER)
	const bool	bAllowUpdate	= !Asura_ControlManager::HasOverrideControls();
#elif defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	const bool	bAllowUpdate	= !Asura_CommandConsole::IsConsoleDown();
#else
	const bool	bAllowUpdate	= true;
#endif	// defined(ASURA_USE_CONTROLMAP_MANAGER)

	if(bAllowUpdate)
	{
		const float		fFrameTime		= Asura_Timers::GetFrameTime();

		float		fMovementSpeed	= s_fCameraMoveSpeedNormal * fFrameTime;

		if(s_xControlMap.GetValue(EXTRA_FAST) > 0.0f)
		{
			fMovementSpeed = 2*s_fCameraMoveSpeedFast * fFrameTime;
		}
		else if(s_xControlMap.GetValue(FAST) > 0.0f)
		{
			fMovementSpeed = s_fCameraMoveSpeedFast * fFrameTime;
		}
		else if(s_xControlMap.GetValue(SLOW) != 0.0f)
		{
			fMovementSpeed = s_fCameraMoveSpeedSlow * fFrameTime;
		}

		const float		fTurnSpeed		= s_fCameraTurnSpeed * fFrameTime;

		const float		fRoll			= s_xControlMap.GetValue(ROLL);
		const float		fPitch			= s_xControlMap.GetValue(PITCH);
		const float		fYaw			= s_xControlMap.GetValue(YAW);

		Asura_Vector_3	xMovement(s_xControlMap.GetValue(MOVE_LEFT_RIGHT), s_xControlMap.GetValue(MOVE_UP_DOWN), s_xControlMap.GetValue(MOVE_FORWARDS_BACKWARDS));

		Asura_Euler_3	xEuler(*Asura_Camera::GetOriEuler());
		Asura_Vector_3	xPosition(*Asura_Camera::GetPosition());
		Asura_Quat		xOrientation(*Asura_Camera::GetOrientation());

		// Update the control-map...
		s_xControlMap.Update();

		// Apply any movement...
		xMovement	*= fMovementSpeed;
		xMovement = xOrientation.RotateVector(xMovement);
		xPosition	+= xMovement;

		// Apply any rotation...
		if(s_bInvertY)
		{
			xEuler.x	-= (fPitch * fTurnSpeed);
		}
		else
		{
			xEuler.x	+= (fPitch * fTurnSpeed);
		}

		xEuler.x = Asura_Maths::ClampToRange( xEuler.x, -dbg_fGimbalLimit, dbg_fGimbalLimit );
		xEuler.y	+= (fYaw * fTurnSpeed);

		if(s_xControlMap.GetValue(RESET_ROLL) > 0.0f)
		{
			xEuler.z	= 0.0f;
		}
		else
		{
			xEuler.z	+= (fRoll * fTurnSpeed);
		}

		xOrientation.FromEuler(xEuler);

		// Handle inverting...
		if(s_xControlMap.GetValue(INVERT_Y) > 0.0f)
		{
			s_bInvertY	= !s_bInvertY;
		}

		// Finally, update the camera position...
		Asura_Camera::SetPosition(xPosition);
		Asura_Camera::SetOrientation(xOrientation);
	}
}

/// End of function 'Asura_FlyCam::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddMoveForwardsBackwardsMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddMoveForwardsBackwardsMapping(const int iAxis,
												   const float fScale)
{
	s_xControlMap.AddAxisMapping(MOVE_FORWARDS_BACKWARDS, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddMoveForwardsBackwardsMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddMoveForwardsBackwardsMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddMoveForwardsBackwardsMapping(const int iForwardsKey,
												   const int iBackwardsKey)
{
	s_xControlMap.AddKeyMapping(MOVE_FORWARDS_BACKWARDS, iForwardsKey, Asura_ControlMapping::ON_PRESS, 1.0f);
	s_xControlMap.AddKeyMapping(MOVE_FORWARDS_BACKWARDS, iBackwardsKey, Asura_ControlMapping::ON_PRESS, -1.0f);
}

/// End of function 'Asura_FlyCam::AddMoveForwardsBackwardsMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddMoveLeftRightMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddMoveLeftRightMapping(const int iAxis,
										   const float fScale)
{
	s_xControlMap.AddAxisMapping(MOVE_LEFT_RIGHT, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddMoveLeftRightMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddMoveLeftRightMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddMoveLeftRightMapping(const int iLeftKey,
										   const int iRightKey)
{
	s_xControlMap.AddKeyMapping(MOVE_LEFT_RIGHT, iLeftKey, Asura_ControlMapping::ON_PRESS, -1.0f);
	s_xControlMap.AddKeyMapping(MOVE_LEFT_RIGHT, iRightKey, Asura_ControlMapping::ON_PRESS, 1.0f);
}

/// End of function 'Asura_FlyCam::AddMoveLeftRightMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddMoveUpDownMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddMoveUpDownMapping(const int iAxis,
										const float fScale)
{
	s_xControlMap.AddAxisMapping(MOVE_UP_DOWN, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddMoveUpDownMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddMoveUpDownMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddMoveUpDownMapping(const int iUpKey,
										const int iDownKey)
{
	s_xControlMap.AddKeyMapping(MOVE_UP_DOWN, iUpKey, Asura_ControlMapping::ON_PRESS, -1.0f);
	s_xControlMap.AddKeyMapping(MOVE_UP_DOWN, iDownKey, Asura_ControlMapping::ON_PRESS, 1.0f);
}

/// End of function 'Asura_FlyCam::AddMoveUpDownMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddRollMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddRollMapping(const int iAxis,
								  const float fScale)
{
	s_xControlMap.AddAxisMapping(ROLL, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddRollMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddRollMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddRollMapping(const int iAntiClockwiseKey,
								  const int iClockwiseKey)
{
	s_xControlMap.AddKeyMapping(ROLL, iAntiClockwiseKey, Asura_ControlMapping::ON_PRESS, -1.0f);
	s_xControlMap.AddKeyMapping(ROLL, iClockwiseKey, Asura_ControlMapping::ON_PRESS, 1.0f);
}

/// End of function 'Asura_FlyCam::AddRollMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddPitchMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddPitchMapping(const int iAxis,
								   const float fScale)
{
	s_xControlMap.AddAxisMapping(PITCH, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddPitchMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddPitchMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddPitchMapping(const int iAntiClockwiseKey,
								   const int iClockwiseKey)
{
	s_xControlMap.AddKeyMapping(PITCH, iAntiClockwiseKey, Asura_ControlMapping::ON_PRESS, -1.0f);
	s_xControlMap.AddKeyMapping(PITCH, iClockwiseKey, Asura_ControlMapping::ON_PRESS, 1.0f);
}

/// End of function 'Asura_FlyCam::AddPitchMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddYawMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddYawMapping(const int iAxis,
								 const float fScale)
{
	s_xControlMap.AddAxisMapping(YAW, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddYawMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddYawMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddYawMapping(const int iAntiClockwiseKey,
								 const int iClockwiseKey)
{
	s_xControlMap.AddKeyMapping(YAW, iAntiClockwiseKey, Asura_ControlMapping::ON_PRESS, -1.0f);
	s_xControlMap.AddKeyMapping(YAW, iClockwiseKey, Asura_ControlMapping::ON_PRESS, 1.0f);
}

/// End of function 'Asura_FlyCam::AddYawMapping'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddFastMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddFastMapping(const int iKey1, const int iKey2)
{
	s_xControlMap.AddKeyMapping(FAST, iKey1, Asura_ControlMapping::ON_PRESS);
	s_xControlMap.AddKeyMapping(FAST, iKey2, Asura_ControlMapping::ON_PRESS);
	s_xControlMap.AddCombinationMapping(EXTRA_FAST,	new Asura_ControlMapping_Combination(new Asura_ControlMapping_Key( iKey1, Asura_ControlMapping::ON_PRESS, 1.0f ),
																						 new Asura_ControlMapping_Key( iKey2, Asura_ControlMapping::ON_PRESS, 1.0f ), 1.0f));
}

void Asura_FlyCam::AddFastMapping(const int iKey)
{
	s_xControlMap.AddKeyMapping(FAST, iKey, Asura_ControlMapping::ON_PRESS);
}

/// End of function 'Asura_FlyCam::AddFastMapping'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddSlowMapping
// Author       : Francois Remacle
// Date         : 6-Mar-2009
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddSlowMapping(const int iKey)
{
	s_xControlMap.AddKeyMapping(SLOW, iKey, Asura_ControlMapping::ON_PRESS);
}

void Asura_FlyCam::AddSlowMapping(const int iAxis, const float fScale)
{
	s_xControlMap.AddAxisMapping(SLOW, iAxis, fScale);
}

/// End of function 'Asura_FlyCam::AddSlowMapping'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddInvertYMapping
// Author       : Andy Tate
// Date         : 1-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddInvertYMapping(const int iKey)
{
	s_xControlMap.AddKeyMapping(INVERT_Y, iKey, Asura_ControlMapping::ON_PRESS);
}

/// End of function 'Asura_FlyCam::AddInvertYMapping'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FlyCam::AddResetRollMapping
// Author       : Francois Remacle
// Date         : 6-Mar-2009
////////////////////////////////////////////////////////////////////////////////

void Asura_FlyCam::AddResetRollMapping(const int iKey)
{
	s_xControlMap.AddKeyMapping(RESET_ROLL, iKey, Asura_ControlMapping::ON_PRESS);
}

/// End of function 'Asura_FlyCam::AddInvertYMapping'. ///

#endif // #ifdef ASURA_SUPPORT_FLYCAM


/// End of file. ///
