
/**
*
* \file Asura_Chunk_Botanicals_Instances.cpp
*
* Created by Dave Reed - 11/11/2005
*
* Copyright (C) 2002 Rebellion (Derby), All Rights Reserved.
*
**/

#include "Core/Asura.h"

#ifdef ASURA_USE_BOTANICALS

#ifdef ASURA_TOOLS

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureFlags.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"

#include "ChunkSystem/Asura_Chunk_Botanicals_Instances.h"
#include "Botanicals/Asura_Botanicals_TemplateManager.h"
#include "Botanicals/Asura_Botanicals_InstanceManager.h"

#define ASURA_CHUNK_BOTANICAL_INSTANCES_CURRENT_VERSION 5

bool Asura_Chunk_Botanicals_Instances::Create(char* szFileName)
{
	int iNumberOfInstances=Asura_Botanicals_InstanceManager::GetNumInstances();

	Asura_SaveBlock* pxSaveBlock = new Asura_SaveBlock();
	Asura_Assert(pxSaveBlock, "Unable to create saveblock");

	if(pxSaveBlock)
	{
		Asura_Chunk_SaveBlock_Stream	xStream (pxSaveBlock);

		xStream.WriteChunk_Start	(ASURA_CHUNK_BOTANICALS_INSTANCES, ASURA_CHUNK_BOTANICAL_INSTANCES_CURRENT_VERSION, 0);

		xStream << iNumberOfInstances;

		// Store instance data
		Asura_LinkList<Asura_Botanicals_Instance*>& pxList=Asura_Botanicals_InstanceManager::GetInstanceList();

		for (Asura_ListElt<Asura_Botanicals_Instance*>* pxElt= pxList.Front();
			pxElt; pxElt = pxElt->Next())
		{
			Asura_Botanicals_Instance* pxInst=pxElt->Data();

			xStream << pxInst->iTemplateUID;
			xStream << pxInst->xPosition;
			xStream << pxInst->xOrientation;
			xStream << pxInst->fScale;
			xStream << pxInst->uFlags;
			xStream << pxInst->uInstanceID;
			xStream << pxInst->iModuleIdx;
			xStream << pxInst->fBrightness;
		}
		
		xStream.WriteChunk_End();
		xStream.WriteAsuraTerminator();

		bool bSuccess=pxSaveBlock->SaveFile(szFileName, false );

		delete pxSaveBlock;

		if (bSuccess)
		{
			return true;
		}
	}

	return true;
}



bool Asura_Chunk_Botanicals_Instances::Process(Asura_Handle iFileHandle) const
{
	//Pre-Version 2 chunks aren't dealt with. RAR 17/03/2006
	if (Version<2 || Version>ASURA_CHUNK_BOTANICAL_INSTANCES_CURRENT_VERSION)
	{
		Asura_Assertf(0,("Unhandled version of botanical instance chunk (%d). Any botanicals will be lost.",Version));
		return false;
	}

	Asura_Botanicals_Instance xInst;
	int iNumberOfInstances=0;
	

	Asura_Chunk_File_Stream	xStream(iFileHandle);

	//ASURA_CHUNK_FILEEXTRACT(iFileHandle, iNumberOfInstances, int);
	xStream >> iNumberOfInstances;

	// Clear any existing botanical instances
	Asura_Botanicals_InstanceManager::Reset();

	Asura_LinkList<Asura_Botanicals_Instance*>& pxList=Asura_Botanicals_InstanceManager::GetInstanceList();

	
	for(int i=0;i<iNumberOfInstances;i++)
	{
		xStream >> xInst.iTemplateUID;
		xStream >> xInst.xPosition;
		xStream >> xInst.xOrientation;
		xStream >> xInst.fScale;

		
		
		// Check for fubar matrices, and reset them. Somehow this has happened with an instance
		// in SS2 M08 - which has NaNs in the matrix. This will automatically reset these to default values, so the
		// offending instance(s) can be fixed/removed. Not sure of the original cause of this, though :(

		bool bMatrixInvalid = false;
		for(int i=0; i<3; i++)
			for(int j=0; j<3; j++)
			{				
			    bMatrixInvalid |=  ( xInst.xOrientation.Element[i][j]!=xInst.xOrientation.Element[i][j] ||
									 xInst.xOrientation.Element[i][j] > 1.01f ||
									 xInst.xOrientation.Element[i][j] < -1.01f);
			}
		
		if ( xInst.fScale != xInst.fScale || xInst.fScale<0.1f || xInst.fScale>4.0f )
		{			
			Asura_Assertf(false, ( "Botanical instance with template ID %d at position (%.4f, %.4f, %.4f) with invalid scale detected... Resetting to defaults!", xInst.iTemplateUID, xInst.xPosition.GetX(), xInst.xPosition.GetY(), xInst.xPosition.GetZ() ) );

			xInst.fScale = 1.0f;
		}
		if ( bMatrixInvalid )
		{
			Asura_Assertf(false, ( "Botanical instance with template ID %d at position (%.4f, %.4f, %.4f) with invalid orientation detected... Resetting to defaults!", xInst.iTemplateUID, xInst.xPosition.GetX(), xInst.xPosition.GetY(), xInst.xPosition.GetZ() ) );

			xInst.xOrientation.Identity();
		}

		if(Version >= 3) //Flags and instance ID added with version 3.
		{
			xStream >> xInst.uFlags;
			xStream >> xInst.uInstanceID;
			Asura_Botanicals_InstanceManager::RegisterInstanceID(xInst.uInstanceID);
		}
		else
		{
			xInst.uFlags = 0;
			//Old instances won't have an ID ,so get one now. RAR 17/03/2006
			xInst.uInstanceID = Asura_Botanicals_InstanceManager::GenerateInstanceID();
		}

		// We now have a module index. Shouldn't need to save it, but it might be useful
		// for now on PS2?  In tools builds, it's recomputed a few lines below - DMR 07/04/06
		xInst.iModuleIdx  =-1;

		if (Version>= 4)
		{
			xStream >> xInst.iModuleIdx;
		}

		if (Version>= 5)
		{
			xStream >> xInst.fBrightness;
		}
		else
		{
			xInst.fBrightness = 1.0f;
		}

		int iIdx= Asura_Botanicals_TemplateManager::GetTemplateIndexFromUID(xInst.iTemplateUID);

		
		// Only create the instance if we can find the template
		if (iIdx>=0)
		{
			Asura_Botanicals_Instance* pxInstToAdd = new Asura_Botanicals_Instance;
			*pxInstToAdd = xInst;
			pxInstToAdd->iTemplateIdx = iIdx;
			pxInstToAdd->pxShapeInstance = NULL;		

			// Assign it to a module - DMR 07/04/06
			// JAT 11/04/06
			// GetModuleIndexFromPosition can fail here because the module links haven't been loaded yet
			// So set it to -2 to mark as couldn't get a valid module index
			pxInstToAdd->iModuleIdx = -2;
			pxList.Push_Back(pxInstToAdd);
		}
	}

	return true;
}

#endif // #ifdef ASURA_TOOLS

#endif	// ASURA_USE_BOTANICALS
