///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_Chunk_Container.cpp
//
// Created by Martin G.Harris (13/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "ChunkSystem/Asura_Chunk_Container.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Entity/Container/Asura_Container.h"
#include "Entity/Container/Asura_Container_System.h"
#include "Entity/Container/Asura_ContainerRegistry.h"


// Tools
#ifdef ASURA_TOOLS
#include "Entity/Container/Tools_Container.h"
#include "Entity/Container/Tools_ContainerFile.h"
#include "Entity/Container/Tools_Container_System.h"
#include "Entity/Container/Tools_ContainerRegistry.h"
#endif //ASURA_TOOLS


// Constants
static const int iASURA_CHUNK_ASURA_CONTAINER_VERSION = 0;
#ifdef ASURA_TOOLS
static const int iASURA_CHUNK_TOOLS_CONTAINER_FILE_VERSION = 0;
static const int iASURA_CHUNK_TOOLS_CONTAINER_GROUP_VERSION = 1;
static const int iASURA_CHUNK_TOOLS_CONTAINER_VERSION = 0;
#endif

///////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_Container
///////////////////////////////////////////////////////////////////////////////

bool Asura_Chunk_Container::Process(Asura_Handle iFileHandle) const
{
	Asura_Chunk_File_Stream	xStream(iFileHandle);
	return Process(xStream);
}

bool Asura_Chunk_Container::Process(Asura_Chunk_Stream& xStream) const
{
	if (Version < 0 || Version > iASURA_CHUNK_ASURA_CONTAINER_VERSION)
	{
		Asura_Assertf(false, ("Unknown chunk version %d for Asura_Container", Version));
		return false;
	}

	u_int uClassification = AsuraContainerClass_NotSet;
	xStream >> uClassification;

	// Create container of given classification.
	Asura_Container* pxContainer = Asura_ContainerRegistry::CreateAsuraClass(uClassification);
	if (!pxContainer) return false;

	pxContainer->ReadFromChunkStream(xStream);

	if (!Asura_Container_System::AddContainer(pxContainer))
	{
		// Couldn't add to container tree, delete it to avoid a memory leak.
		// Perhaps we read the chunk twice or something.
		delete pxContainer;
		pxContainer  = 0;
	}

	return true;
}


void Asura_Chunk_Container::Create(const Asura_Container* pxContainer, Asura_Chunk_Stream& xStream)
{
	Asura_Assert(pxContainer, "Null pointer");
	if (!pxContainer) return;

	u_int uChunkFlags = 0;

	xStream.WriteChunk_Start(ASURA_CHUNK_CONTAINER, iASURA_CHUNK_ASURA_CONTAINER_VERSION, uChunkFlags);

	// Write the classification
	xStream << pxContainer->GetClassification();

	// Write the rest of the container.
	pxContainer->WriteToChunkStream(xStream);

	xStream.WriteChunk_End();
}

void Asura_Chunk_Container::Create(const Asura_Container* pxContainer, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream( pxSaveBlock );
	Create(pxContainer, xStream);
}


///////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_ToolsContainerFile
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS

bool Asura_Chunk_ToolsContainerFile::Process(Asura_Handle iFileHandle) const
{
	Asura_Chunk_File_Stream	xStream(iFileHandle);
	return Process(xStream);
}


bool Asura_Chunk_ToolsContainerFile::Process(Asura_Chunk_Stream& xStream) const
{
	if (Version < 0 || Version > iASURA_CHUNK_TOOLS_CONTAINER_FILE_VERSION)
	{
		Asura_Assertf(false, ("Unknown chunk version %d for Tools_ContainerFile", Version));
		return false;
	}

	Asura_String xName;
	xStream >> xName;

	// Check that we don't already have this loaded.
	const Asura_Hash_ID uFileHash = xName.GetHash();
	Tools_ContainerFile* pxExistingFile = Tools_Container_System::FindFileByHash(uFileHash);
	if (pxExistingFile)
	{
		Asura_Assertf(0,("Ignoring container file '%s' which already exists.", xName.GetString()));
		return true;
	}

	Tools_ContainerFile xFile;
	xFile.m_xCreationShortName = xName;

	Tools_Container_System::RegisterFile(xFile, false);

	return true;
}

void Asura_Chunk_ToolsContainerFile::Create(const Tools_ContainerFile* pxFile, Asura_Chunk_Stream& xStream)
{	
	Asura_Assert(pxFile, "Null pointer");
	if (!pxFile) return;

	u_int uChunkFlags = 0;

	xStream.WriteChunk_Start(TOOLS_CHUNK_CONTAINER_FILE, iASURA_CHUNK_TOOLS_CONTAINER_FILE_VERSION, uChunkFlags);

	xStream << pxFile->m_xCreationShortName;

	xStream.WriteChunk_End();

}

void Asura_Chunk_ToolsContainerFile::Create(const Tools_ContainerFile* pxFile, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream( pxSaveBlock );
	Create(pxFile, xStream);
}

#endif // #ifdef ASURA_TOOLS




///////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_ToolsContainerGroup
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS

bool Asura_Chunk_ToolsContainerGroup::Process(Asura_Handle iFileHandle) const
{
	Asura_Chunk_File_Stream	xStream(iFileHandle);
	return Process(xStream);
}


bool Asura_Chunk_ToolsContainerGroup::Process(Asura_Chunk_Stream& xStream) const
{
	if (Version < 0 || Version > iASURA_CHUNK_TOOLS_CONTAINER_GROUP_VERSION)
	{
		Asura_Assertf(false, ("Unknown chunk version %d for Tools_ContainerGroup", Version));
		return false;
	}

	Asura_Hash_ID uFileHash;
	Asura_String xCreationGroupName;
	Asura_String xGroupName;
	u_int uAutoGenNumber = 0;
	
	xStream >> uFileHash;
	xStream >> xCreationGroupName;
	xStream >> xGroupName;

	if (Version >= 1)
	{
		xStream >> uAutoGenNumber;
	}

	// Check that we don't already have this loaded.
	const Asura_Hash_ID uGroupHash = xCreationGroupName.GetHash();
	Tools_ContainerGroup* pxExistingGroup = Tools_Container_System::FindGroupByHash(uGroupHash);
	if (pxExistingGroup)
	{
		Asura_Assertf(0,("Ignoring container group '%s' which already exists.", xGroupName.GetString()));
		return true;
	}

	Tools_ContainerGroup xGroup;
	xGroup.m_uFileHash = uFileHash;
	xGroup.m_xCreationGroupName = xCreationGroupName;
	xGroup.m_xGroupName = xGroupName;
	xGroup.m_uAutoGenNumber = uAutoGenNumber;

	Tools_Container_System::RegisterGroup(xGroup);

	return true;
}

void Asura_Chunk_ToolsContainerGroup::Create(const Tools_ContainerGroup* pxGroup, Asura_Chunk_Stream& xStream)
{	
	Asura_Assert(pxGroup, "Null pointer");
	if (!pxGroup) return;

	u_int uChunkFlags = 0;

	xStream.WriteChunk_Start(TOOLS_CHUNK_CONTAINER_GROUP, iASURA_CHUNK_TOOLS_CONTAINER_GROUP_VERSION, uChunkFlags);

	xStream << pxGroup->m_uFileHash;
	xStream << pxGroup->m_xCreationGroupName;
	xStream << pxGroup->m_xGroupName;
	xStream << pxGroup->m_uAutoGenNumber;

	xStream.WriteChunk_End();
}

void Asura_Chunk_ToolsContainerGroup::Create(const Tools_ContainerGroup* pxGroup, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream( pxSaveBlock );
	Create(pxGroup, xStream);
}

#endif // #ifdef ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_ToolsContainer
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS

bool Asura_Chunk_ToolsContainer::Process(Asura_Handle iFileHandle) const
{
	Asura_Chunk_File_Stream	xStream(iFileHandle);
	return Process(xStream);
}

bool Asura_Chunk_ToolsContainer::Process(Asura_Chunk_Stream& xStream) const
{
	if (Version < 0 || Version > iASURA_CHUNK_TOOLS_CONTAINER_VERSION)
	{
		Asura_Assertf(false, ("Unknown chunk version %d for Tools_Container", Version));
		return false;
	}

	u_int uClassification = AsuraContainerClass_NotSet;
	xStream >> uClassification;

	// Create container of given classification.
	Tools_Container* pxContainer = Tools_ContainerRegistry::CreateToolsClass(uClassification);
	if (!pxContainer) return false;

	pxContainer->ReadFromChunkStream(xStream);

	Tools_Container_System::RegisterContainer(pxContainer);

	return true;
}


void Asura_Chunk_ToolsContainer::Create(const Tools_Container* pxContainer, Asura_Chunk_Stream& xStream)
{
	Asura_Assert(pxContainer, "Null pointer");
	if (!pxContainer) return;

	u_int uChunkFlags = 0;

	xStream.WriteChunk_Start(TOOLS_CHUNK_CONTAINER, iASURA_CHUNK_TOOLS_CONTAINER_VERSION, uChunkFlags);

	// Write the classification
	xStream << pxContainer->GetClassification();

	// Write the rest of the container.
	pxContainer->WriteToChunkStream(xStream);

	xStream.WriteChunk_End();
}

void Asura_Chunk_ToolsContainer::Create(const Tools_Container* pxContainer, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream( pxSaveBlock );
	Create(pxContainer, xStream);
}

#endif // #ifdef ASURA_TOOLS


#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

