#include "Core/Asura.h"

#include "ChunkSystem/Asura_Chunk_Cutscene.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Cutscene/Asura_Cutscene_Actor.h"
#include "Cutscene/Asura_Cutscene_Attachment.h"
#include "Cutscene/Asura_Cutscene_Event.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Cutscene/Asura_Cutscene_Track.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Network.h"

#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING)

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Chunk_Cutscene **
**----------------------*/

#define ASURA_CHUNK_CUTSCENE_VERSION (22)
// RAR 23/09/03 Version 12 - Set default value for skippable flag
// MP  02/12/04         13 - Bumped for nefarious Rogue-specific purposes.
//                           Data is unchanged.
// RAR 10/02/05 Version 14 - Added sound track volume
// RAR 02/09/05 Version 15 - Cutscene start time.
// RAR 11/01/06 Version 16 - FOV moved to camera actor
// MP  14/03/06         17 - Another bump to fix Rogue issues, this time due
//                           to our chunk data diverging slightly in our
//                           branch.
// RDM 30/05/07 Version 18 - Following in Morgan's footsteps, a bump for BFF's nefarious purposes.
//							 Again, data is unchanged. We need a better way of doing this...
// RAR 04/04/08 Version 19 - A proper version increase this time! Added light set flags.
// RAR 27/06/08 Version 20 - Added mixer preset.
// TY  11/07/08 Version 21 - added SoundTrackContainsMusic flag
// ESC 27/03/09 Version 22 - Bumped to allow future projects to increment their own versions

//static members
Asura_Cutscene* Asura_Chunk_Cutscene::CurrentCutscene = 0;
int Asura_Chunk_Cutscene::NextActorIndex = 0;
int Asura_Chunk_Cutscene::NextEventIndex = 0;

DEBUGVAR bool dbg_bForceServerCutscenes = false;

#ifdef ASURA_TOOLS
bool Asura_Chunk_Cutscene::s_bCutscenesLoadedFromEntityFile = false;
#endif

bool Asura_Chunk_Cutscene::Process(Asura_Handle fileHandle)
{
	#ifdef ASURA_TOOLS
	//See if the cutscene is being loaded from an entity file.
	//This is needed so that tools can warn about the neeed to have both cutscene and entity files
	//available upon save. RAR 28/03/03
	const char* pszFilename = Asura_FileAccess::GetFilename(fileHandle);
	const char* pszExtension = strrchr(pszFilename,'.');
	if(pszExtension && !stricmp(pszExtension,".ent"))
	{
		s_bCutscenesLoadedFromEntityFile = true;
	}
	#endif

	
	Asura_Assert(Version<=ASURA_CHUNK_CUTSCENE_VERSION,"Unknown chunk version");
	Asura_Assert(Version!=7,"This cutscene version no longer supported. (unless really , really , really needed).");

	const char* name = Asura_FileAccess::ReadString(fileHandle);
	if(Asura_Cutscene_System::GetCutscene(name))
	{
		Asura_Assert(0,"This cutscene already exists");
		CurrentCutscene = 0;

		return false;
	}
	
	//get the cutscene type
	int cutsceneType = ACT_StandardCutscene;
	if(Version>4)
	{
		ASURA_CHUNK_FILEEXTRACT(fileHandle,cutsceneType,int);
	}

    // (MH - 22.07.2002) check whether this is a server-side cutscene
    int playsOnServer = false;
    if (Version>5)
    {
        ASURA_CHUNK_FILEEXTRACT(fileHandle,playsOnServer,int);
    }
	
	//create the cutscene
    const Asura_String xName(name);
	Asura_Cutscene *cutscene = Asura_Cutscene::AllocateCutscene(cutsceneType,xName);
	Asura_Assert(cutscene,"");

	
    if (cutscene)
    {
        // (MH - 04.10.2002) disable this until all transitions etc. have the right properties.
        // Normal cutscenes don't play on the server so this shouldn't affect anything other
        // than GF2.
        //cutscene->PlaysOnServer = playsOnServer ? true : false;

        Asura_Cutscene_System::AddCutscene(cutscene);

	    cutscene->LoadFromChunk(this,fileHandle);

	    //store the current cutscene being loaded for later chunk loads
	    CurrentCutscene = cutscene;
	    NextActorIndex = 0;
	    NextEventIndex = 0;
    }
    else
    {
        // set CurrentCutscene to NULL so further related chunks are skipped.
        CurrentCutscene = 0;

		return false;
    }
	
	return true;
}


void Asura_Cutscene::LoadFromChunk(class Asura_Chunk_Cutscene* chunk,Asura_Handle fileHandle)
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
#ifndef ASURA_TOOLS
		// No longer bothering to support versions < 14 outside tools - they're
		// all ancient now.
		Asura_Assert( chunk->Version >= 14, "" );
#endif

#ifdef ASURA_TOOLS
		if(chunk->Version>1)
#endif
		{
			ASURA_CHUNK_FILEEXTRACT_STRING(fileHandle,CutsceneDirectory)
		}

		/// Kevin Lea - 25/04/2002 17:49:28 - new for version 3, soundtrack name
#ifdef ASURA_TOOLS
		if( chunk->Version < 3 )
		{
			SoundTrackName.MakeBlank();
			UseSoundTrack = false;
		}
		else
#endif
		{
			ASURA_CHUNK_FILEEXTRACT_STRING(fileHandle,SoundTrackName)
				UseSoundTrack = !SoundTrackName.IsBlank();
		}

		//Version 4 - CutsceneEndTime
#ifdef ASURA_TOOLS
		if(chunk->Version>3)
#endif
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle,CutsceneEndTime,float)
		}

		// Version 7 - name lists
#ifdef ASURA_TOOLS
		if (chunk->Version > 6)
#endif
		{
			SkinNames.LoadFromChunk(fileHandle);
			SkinFileNames.LoadFromChunk(fileHandle);
			SkinFileNames.TidyFilenames();
			AnimationNames.LoadFromChunk(fileHandle);
			AnimationFileNames.LoadFromChunk(fileHandle);
			AnimationFileNames.TidyFilenames();
		}

#ifdef ASURA_TOOLS
		if(chunk->Version>8)
#endif
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle,m_iCutsceneFlags,int);
			// (MH - 24.08.2004) this force widescreen was done in November 2002, probably not needed now.
			//m_iCutsceneFlags |= AsuraCutsceneFlag_Widescreen;
			unsigned int temp;
			ASURA_CHUNK_FILEEXTRACT(fileHandle,temp,unsigned int);
			m_usEndCutsceneMessageBlock = (unsigned short) temp;

#ifdef ASURA_TOOLS
			//Make sure the message block is in the cutscene message block set. RAR 28/03/03
			if(m_usEndCutsceneMessageBlock != ASURA_STATICMESSAGE_EMPTYBLOCK)
			{
				int iIndex;
				Asura_StaticMessageSetIndex eSet;
				Asura_StaticMessage_System::ConvertIDToIndex(m_usEndCutsceneMessageBlock, &iIndex, &eSet);

				if(eSet != ASURA_STATICMESSAGE_CUTSCENESET)
				{
			 		Asura_Assertf(0,("The end of cutscene messages for '%s' have been added to the entity file instead of the cutscene file.\n"
									"Please delete them all and then add them again , with both the entity and cutscene files checked out",
									GetName().GetString()));
				}
			}
			//m_usEndCutsceneMessageBlock = Asura_StaticMessage_System::TransferMessageBlock(m_usEndCutsceneMessageBlock,ASURA_STATICMESSAGE_CUTSCENESET);
#endif

#ifdef ASURA_TOOLS
			if(chunk->Version<12)
			{
				//skippable flag added with version 12. should be on by default for old cutscenes
				m_iCutsceneFlags |= AsuraCutsceneFlag_Skippable;
			}
#endif
		}

		if(chunk->Version>9 && chunk->Version<16)
		{
			//FOV now lives in the camera actor
#ifdef ASURA_TOOLS
			ASURA_CHUNK_FILEEXTRACT(fileHandle,m_fInitialFOV_UsedForUpdatingOldCutscenes,Asura_Angle);
#else
			Asura_FileAccess::SkipData(fileHandle,sizeof(Asura_Angle));
#endif
		}

#ifdef ASURA_TOOLS
		if(chunk->Version>10)
#endif
		{
			//subtitle array now obsolete.. so skip it. RAR 22/08/2005
			int iNumberOfSubtitles;
			ASURA_CHUNK_FILEEXTRACT(fileHandle,iNumberOfSubtitles,int);
			if(iNumberOfSubtitles)
			{
				Asura_FileAccess::SkipData(fileHandle,sizeof(Asura_Hash_ID)*iNumberOfSubtitles);
			}
		}

#ifdef ASURA_TOOLS
		if(chunk->Version>13)
#endif
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle,m_fSoundTrackVolume,float);
		}

		if(chunk->Version>14)
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle,m_fCutsceneStartTime,float);
		}

		if(chunk->Version>18)
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle,m_uLightSets,u_int);
		}
		
		if(chunk->Version>19)
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle,m_uMixerPreset,Asura_Hash_ID);
		}

		m_bSoundTrackContainsMusic = false;
		if( chunk->Version > 20 )
		{		
			int iSoundTrackContainsMusic = 0;
			ASURA_CHUNK_FILEEXTRACT( fileHandle, iSoundTrackContainsMusic, int );
			m_bSoundTrackContainsMusic = ( iSoundTrackContainsMusic != 0 ) ? true : false;
		}

		//allocate and initalise actor pointers
		NumberOfActors = chunk->NumberOfActors;
		if(NumberOfActors)
		{
			Asura_ResourceAllocateArray(Actors, NumberOfActors);
			int i;	
			for(i=0;i<NumberOfActors;i++)
			{
				Actors[i] = 0;
			}
		}

		//allocate and initialise event pointers
		NumberOfEvents = chunk->NumberOfEvents;
		if(NumberOfEvents)
		{
			Asura_ResourceAllocateArray(Events, NumberOfEvents);
			int i;	
			for(i=0;i<NumberOfEvents;i++)
			{
				Events[i] = 0;
			}
		}
		Position = chunk->Position;	
	}
	Asura_ResourcePopType();
}

#ifdef ASURA_TOOLS
void Asura_Chunk_Cutscene::Create(Asura_Cutscene* cutscene,Asura_SaveBlock& saveBlock)
{
	// Do we need to save the chunk?

    if (cutscene->ShouldSave())
	{
		Asura_Chunk_SaveBlock_Stream	xStream (&saveBlock);


		// Yes - set the current cutscene...

		CurrentCutscene = cutscene;


#if 1
		int chunkSize = cutscene->GetChunkSize();
		
		Asura_Chunk_Cutscene *chunk = (Asura_Chunk_Cutscene*)saveBlock.GetAccessPointer(chunkSize);
			
		
		chunk->ID = ASURA_CHUNK_CUTSCENE;
		ASURA_CHUNK_VERIFY(chunk->ID, int);
		chunk->Size = chunkSize;
		ASURA_CHUNK_VERIFY(chunk->Size, int);
		chunk->Version = ASURA_CHUNK_CUTSCENE_VERSION;
		ASURA_CHUNK_VERIFY(chunk->Version, int);
		chunk->Flags = 0;
		ASURA_CHUNK_VERIFY(chunk->Flags, int);

		chunk->NumberOfActors = cutscene->NumberOfActors;
		ASURA_CHUNK_VERIFY(chunk->NumberOfActors, int);
		chunk->NumberOfEvents = cutscene->NumberOfEvents;
		ASURA_CHUNK_VERIFY(chunk->NumberOfEvents, int);
		chunk->Position = cutscene->Position;
		ASURA_CHUNK_ARRAY_VERIFY(chunk->Position, float, 3);

		char* chunkBuffer = (char*)(chunk+1);
		cutscene->SaveToChunk(chunkBuffer);
		
		//create chunks for each actor
		int i;
		for(i=0;i<cutscene->NumberOfActors;i++)
		{
			Asura_Chunk_Actor::Create(cutscene->Actors[i],saveBlock);
		}

		//and for each event
		for(i=0;i<cutscene->NumberOfEvents;i++)
		{
			Asura_Chunk_CutsceneEvent::Create(cutscene->Events[i],saveBlock);
		}
#else
		// Write the data...

		xStream.WriteChunk_Start	(ASURA_CHUNK_CUTSCENE, ASURA_CHUNK_CUTSCENE_VERSION, 0);

		xStream << cutscene->NumberOfActors;
		xStream << cutscene->NumberOfEvents;
		xStream << cutscene->Position;

		cutscene->WriteToChunkStream	(xStream);
		xStream.WriteChunk_End			();


		// Ccreate chunks for each actor..

		for (int i = 0; i < cutscene->NumberOfActors; i++)
		{
			Asura_Chunk_Actor::Create	(cutscene->Actors [i], xStream);
		}


		// ...and for each event...

		for (int i = 0; i < cutscene->NumberOfEvents; i++)
		{
			Asura_Chunk_CutsceneEvent::Create	(cutscene->Events [i], xStream);
		}
#endif
	}
}
#endif

int Asura_Cutscene::GetChunkSize()
{
	if (!UseSoundTrack || SoundTrackName.IsNull()) 
	{
		SoundTrackName = "";
	}

	int chunkSize = sizeof(Asura_Chunk_Cutscene);
	ASURA_CHUNK_SIZE_STRING(chunkSize,CutsceneName)
	chunkSize += sizeof(int);//for cutsceneType
    chunkSize += sizeof(int);//for PlaysOnServer (added in version 6)
	ASURA_CHUNK_SIZE_STRING(chunkSize,CutsceneDirectory)
	ASURA_CHUNK_SIZE_STRING(chunkSize,SoundTrackName)
	ASURA_CHUNK_SIZE(chunkSize,CutsceneEndTime,float)
    chunkSize += SkinNames.GetChunkSize();
    chunkSize += SkinFileNames.GetChunkSize();
    chunkSize += AnimationNames.GetChunkSize();
    chunkSize += AnimationFileNames.GetChunkSize();
	chunkSize += sizeof(int)+sizeof(unsigned int); // V9 Flags and message block
	chunkSize += sizeof(int);//V11 array of subtitle ids (obsolete , just write 0)
	chunkSize += sizeof(float); // V14 sound track volume
	chunkSize += sizeof(float); // V15 cutscene start time
	chunkSize += sizeof(u_int); // V19 light sets
	chunkSize += sizeof(Asura_Hash_ID); // V20 mixer preset
	chunkSize += sizeof(int);  // V21 SoundTrackContainsMusic

	return chunkSize;
}

void Asura_Cutscene::SaveToChunk(char*& chunkBuffer)
{
	ASURA_CHUNK_FILL_STRING(chunkBuffer,CutsceneName)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,GetCutsceneType(),int)
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,PlaysOnServer,int) // (MH - 22.07.2002) added in version 6
	ASURA_CHUNK_FILL_STRING(chunkBuffer,CutsceneDirectory)
	/// Kevin Lea - 25/04/2002 17:32:06 - added for version 3 - soundtrack name
	ASURA_CHUNK_FILL_STRING(chunkBuffer,SoundTrackName)
	//version 4 , cutscene end time
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,CutsceneEndTime,float)

    // Version 7 - name lists
    SkinNames.SaveToChunk(chunkBuffer);
    SkinFileNames.SaveToChunk(chunkBuffer);
    AnimationNames.SaveToChunk(chunkBuffer);
    AnimationFileNames.SaveToChunk(chunkBuffer);

	// Version 9 - Flags and end cutscene message block
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_iCutsceneFlags,int);
	unsigned int temp = m_usEndCutsceneMessageBlock;
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,temp,unsigned int);

	//Version 11 - Subtitles (obsoloete
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,0,int);

	//Version 14 - Sound track volume
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_fSoundTrackVolume,float);

	//Version 15 - Start time
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_fCutsceneStartTime,float);

	//Version 19 - lightsets
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_uLightSets,u_int);

	//Version 20 - mixer preset
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_uMixerPreset,Asura_Hash_ID);

	//Version 21 - SoundTrackCountainsMusic
	ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, ( m_bSoundTrackContainsMusic ) ? 1 : 0, int );

}

void Asura_Cutscene::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	int		iType						= GetCutsceneType ();
	int		iPlaysOnServer				= PlaysOnServer;
	u_int	uEndCutsceneMessageBlock	= m_usEndCutsceneMessageBlock;


	// Write the chunk-data...

	xStream << CutsceneName;
	xStream << iType;
	xStream << iPlaysOnServer;
	xStream << CutsceneDirectory;

	/// Kevin Lea - 25/04/2002 17:32:06 - added for version 3 - soundtrack name
	xStream << SoundTrackName;

	//version 4 , cutscene end time
	xStream << CutsceneEndTime;

	// Version 7 - name lists
    SkinNames.WriteToChunkStream			(xStream);
    SkinFileNames.WriteToChunkStream		(xStream);
    AnimationNames.WriteToChunkStream		(xStream);
    AnimationFileNames.WriteToChunkStream	(xStream);

	// Version 9 - Flags and end cutscene message block
	xStream << m_iCutsceneFlags;
	xStream << uEndCutsceneMessageBlock;


	//Version 11 - Subtitles (obsolete)
	int iZero = 0;
	xStream << iZero;

	//Version 14 - Sound track volume
	xStream << m_fSoundTrackVolume;

	//Version 15 - Start time
	xStream << m_fCutsceneStartTime;

	//Version 19 - lightsets
	xStream << m_uLightSets;

	//Version 20 - mixer preset
	xStream << m_uMixerPreset;

	//Version 21 - SoundTrackCountainsMusic
	const int iSoundTrackContainsMusic = ( m_bSoundTrackContainsMusic ) ? 1 : 0;
	xStream << iSoundTrackContainsMusic;

}

Asura_Cutscene_Actor*& Asura_Chunk_Cutscene::GetNextActor()
{
	Asura_Assert(CurrentCutscene,"");
	Asura_Assert(NextActorIndex<CurrentCutscene->NumberOfActors,"");
	return CurrentCutscene->Actors[NextActorIndex++];
}

Asura_Cutscene_Actor* Asura_Chunk_Cutscene::GetActor(const Asura_String& name)
{
	Asura_Assert(CurrentCutscene,"");
	for(int i=0;i<CurrentCutscene->NumberOfActors;i++)
	{
		if(name == CurrentCutscene->Actors[i]->ActorName)
		{
			return CurrentCutscene->Actors[i];
		}
	}
	return 0;
}

Asura_Cutscene_Event*& Asura_Chunk_Cutscene::GetNextEvent()
{
	Asura_Assert(CurrentCutscene,"");
	Asura_Assert(NextEventIndex<CurrentCutscene->NumberOfEvents,"");
	return CurrentCutscene->Events[NextEventIndex++];
}



////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------**
** Asura_Chunk_Actor **
**-------------------*/

#define ASURA_CHUNK_ACTOR_VERSION 22
//Version 10 - Added flags to shape actor. RAR 07/04/2004
//Version 11 - Added alternate texture from hierarchies. RAR 14/01/2005
//Version 12 - Actor flags , and multiple camera support. RAR 14/10/2005
//Version 13 - Added initial FOV. RAR 11/01/2006
//Version 14 - Added m_uGuidToReplace. RAR 31/03/2006
//Version 15 - Added handheld options to camera. JDS 21/03/2007
//Version 16 - Upper version to stop load corruption of unversioned child DF 24/05/2007
//static members
//Version 17 - Added player camera control options. KOR 10/03/2008
//Version 18 - Added IK template ID. RAR 13/05/2008
//Version 19 - Support for using multiple skins. RAR 30/09/2008
//Version 20 - Tools sort order. RAR 06/11/2008
//Version 21 - Face ID. RAR 27/11/2008
//Version 22 - Near/Far clip for camera. RAR 27/02/2009
Asura_Cutscene_Actor* Asura_Chunk_Actor::CurrentActor = 0;
int Asura_Chunk_Actor::NextTrackIndex;


bool Asura_Chunk_Actor::Process(Asura_Handle fileHandle)
{
	Asura_Assert(Version<=ASURA_CHUNK_ACTOR_VERSION,"Unknown chunk version");

	if(!Asura_Chunk_Cutscene::CurrentCutscene)
	{
		return false;
	}

	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		//get the actor
		Asura_Cutscene_Actor*& actor = Asura_Chunk_Cutscene::GetNextActor();

		bool bServer = dbg_bForceServerCutscenes || Asura_Chunk_Cutscene::CurrentCutscene->PlaysOnServer;
		actor = Asura_Cutscene_Actor::AllocateActor(Type,bServer,Asura_Chunk_Cutscene::CurrentCutscene);

		Asura_Assert(actor,"");

		actor->LoadFromChunk(this,fileHandle);


		//initialise track pointers
		actor->NumberOfTracks = NumberOfTracks;
		if(NumberOfTracks)
		{
			Asura_ResourceAllocateArray(actor->Tracks, NumberOfTracks);
			int i;	
			for(i=0;i<NumberOfTracks;i++)
			{
				actor->Tracks[i] = 0;
			}
		}

		CurrentActor = actor;
		NextTrackIndex = 0;
	}
	Asura_ResourcePopType();
	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_Actor::Create(Asura_Cutscene_Actor* actor,Asura_SaveBlock& saveBlock)
{
	int chunkSize = actor->GetChunkSize() + sizeof(Asura_Chunk_Actor);
	
	Asura_Chunk_Actor *chunk = (Asura_Chunk_Actor*)saveBlock.GetAccessPointer(chunkSize);

	chunk->ID = ASURA_CHUNK_CUTSCENEACTOR;
	ASURA_CHUNK_VERIFY(chunk->ID, int);
	chunk->Size = chunkSize;
	ASURA_CHUNK_VERIFY(chunk->Size, int);
	chunk->Version = ASURA_CHUNK_ACTOR_VERSION;
	ASURA_CHUNK_VERIFY(chunk->Version, int);
	chunk->Flags = 0;
	ASURA_CHUNK_VERIFY(chunk->Flags, int);

	chunk->Type = (int) actor->Type;
	ASURA_CHUNK_VERIFY(chunk->Type, int);
	chunk->NumberOfTracks = actor->NumberOfTracks;
	ASURA_CHUNK_VERIFY(chunk->NumberOfTracks, int);

	char* chunkBuffer = (char*)(chunk+1);
	actor->SaveToChunk(chunkBuffer);

	//create chunks for the tracks
	for(int i=0;i<actor->NumberOfTracks;i++)
	{
		Asura_Chunk_CutsceneTrack::Create(actor->Tracks[i],saveBlock);
	}

	//create chunks for attachments , if a hierarchy actor
	if(actor->IsHierarchyActor())
	{
		Asura_Cutscene_Actor_Hierarchy* pxHierarchyActor = (Asura_Cutscene_Actor_Hierarchy*) actor;
		for(int i=0;i<pxHierarchyActor->m_iNumberOfAttachments;i++)
		{
			Asura_Chunk_CutsceneAttachment::Create(pxHierarchyActor->m_ppxAttachments[i],saveBlock);
		}
	}
}

void Asura_Chunk_Actor::Create (Asura_Cutscene_Actor* actor, Asura_Chunk_Stream& xStream)
{
	// Write the data...

	xStream.WriteChunk_Start	(ASURA_CHUNK_CUTSCENEACTOR, ASURA_CHUNK_ACTOR_VERSION, 0);

	xStream << actor->Type;
	xStream << actor->NumberOfTracks;

	actor->WriteToChunkStream	(xStream);
	xStream.WriteChunk_End		();


	// Create chunks for the tracks...

	for (int i = 0; i < actor->NumberOfTracks; i++)
	{
		Asura_Chunk_CutsceneTrack::Create	(actor->Tracks [i], xStream);
	}


	// Is it a hierarchy actor?

	if (actor->IsHierarchyActor())
	{
		Asura_Cutscene_Actor_Hierarchy*	pxHierarchyActor	= static_cast<Asura_Cutscene_Actor_Hierarchy*>(actor);


		// Yes - create chunks for attachments...

		for (int i = 0; i < pxHierarchyActor->m_iNumberOfAttachments; i++)
		{
			Asura_Chunk_CutsceneAttachment::Create	(pxHierarchyActor->m_ppxAttachments [i], xStream);
		}
	}
}
#endif

Asura_Cutscene_Track*& Asura_Chunk_Actor::GetNextTrack()
{
	Asura_Assert(CurrentActor,"");
	Asura_Assert(NextTrackIndex<CurrentActor->NumberOfTracks,"");
	return CurrentActor->Tracks[NextTrackIndex++];
}

Asura_Cutscene_Attachment*& Asura_Chunk_Actor::GetNextAttachment()
{
	Asura_Assert(CurrentActor,"");
	Asura_Assert(CurrentActor->IsHierarchyActor(),"");
	Asura_Cutscene_Actor_Hierarchy* pxHierarchyActor = (Asura_Cutscene_Actor_Hierarchy*) CurrentActor;

	//find the next unallocated one
	for(int i=0;i<pxHierarchyActor->m_iNumberOfAttachments;i++)
	{
		if(!pxHierarchyActor->m_ppxAttachments[i]) return pxHierarchyActor->m_ppxAttachments[i];
	}

	Asura_Assert(0,"Error finding next attachment");
	return pxHierarchyActor->m_ppxAttachments[0];
}

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Cutscene_Actor **
**----------------------*/


int Asura_Cutscene_Actor::GetChunkSize()
{
	int chunkSize = 0;//sizeof(Asura_Chunk_Actor);
	ASURA_CHUNK_SIZE_STRING(chunkSize,ActorName)
	ASURA_CHUNK_SIZE(chunkSize,m_uGuid,Asura_Guid)
	ASURA_CHUNK_SIZE(chunkSize,m_uGuidToReplace,Asura_Guid)
	ASURA_CHUNK_SIZE(chunkSize,m_uActorFlags,u_int)
	chunkSize += sizeof(u_int);//m_uToolsSortOrder;

	return chunkSize;
}

void Asura_Cutscene_Actor::SaveToChunk(char*& chunkBuffer)
{
	ASURA_CHUNK_FILL_STRING(chunkBuffer,ActorName);
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_uGuid,Asura_Guid);
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_uGuidToReplace,Asura_Guid);
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_uActorFlags,u_int);
	#ifdef ASURA_TOOLS
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,m_uToolsSortOrder,u_int);
	#else
	u_int uZero = 0;
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,uZero,u_int);
	#endif
}

void Asura_Cutscene_Actor::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	xStream << ActorName;
	xStream << m_uGuid;
	xStream << m_uGuidToReplace;
	xStream << m_uActorFlags;
	#ifdef ASURA_TOOLS
	xStream << m_uToolsSortOrder;
	#else
	xStream << (u_int)0;
	#endif
}

void Asura_Cutscene_Actor::LoadFromChunk(class Asura_Chunk_Actor* chunk,Asura_Handle fileHandle)
{
	ASURA_CHUNK_FILEEXTRACT_STRING(fileHandle,ActorName)
	if(chunk->Version>=5)
	{
		ASURA_CHUNK_FILEEXTRACT(fileHandle,m_uGuid,Asura_Guid)
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
		m_uGuid = Asura_Streaming_System::MakeCurrentMasterZoneGuid(m_uGuid);
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
	}

	Asura_Assert(Cutscene,"");
#ifdef ASURA_TOOLS
	if(m_uGuid!=ASURA_GUID_UNREGISTERED)
	{
		//If the guid isn't in the range for cuscene guids , get a new one. RAR 28/03/03
		if(!Cutscene->IsValidActorGuid(m_uGuid))
		{
			m_uGuid = Cutscene->AssignActorGuid();
		}
		else
		{
			Cutscene->RegisterActorGuid(m_uGuid);
		}
	}
#endif
	if(chunk->Version>=14)
	{
		ASURA_CHUNK_FILEEXTRACT(fileHandle,m_uGuidToReplace,Asura_Guid)
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
		m_uGuidToReplace = Asura_Streaming_System::MakeCurrentMasterZoneGuid(m_uGuidToReplace);
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
	}

	if(chunk->Version>=12)
	{
		ASURA_CHUNK_FILEEXTRACT(fileHandle,m_uActorFlags,u_int)
	}
	if(chunk->Version>=20)
	{
		#ifdef ASURA_TOOLS
		ASURA_CHUNK_FILEEXTRACT(fileHandle,m_uToolsSortOrder,u_int);
		#else
		Asura_FileAccess::SkipData(fileHandle,sizeof(u_int));
		#endif
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_Cutscene_Actor_Hierarchy **
**--------------------------------*/

int Asura_Cutscene_Actor_Hierarchy::GetChunkSize()
{
	int chunkSize = Asura_Cutscene_Actor::GetChunkSize();
    chunkSize += 5 * sizeof(int) + sizeof(float) + sizeof(Asura_Hash_ID)*3;
	chunkSize += sizeof(int) * ASURA_ANIMATION_SKIN_COUNT;

	return chunkSize;
}

void Asura_Cutscene_Actor_Hierarchy::SaveToChunk(char*& chunkBuffer)
{
	Asura_Cutscene_Actor::SaveToChunk(chunkBuffer);

	u_int uHierarchyFlagsToSave = m_uHierarchyFlags;
	#ifdef  ASURA_TOOLS
	if(m_aiSkinID[0] != -1)
	{
		//Automatically flag particle_bloke as not rendered. RAR 24/01/2006
		if(!Cutscene->SkinNames.GetName(m_aiSkinID[0]).Compare("particle_bloke"))
		{
			uHierarchyFlagsToSave |= ASURA_HIERARCHYACTORFLAG_INITIALLY_HIDDEN;
		}
	}
	#endif


	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, ASURA_ANIMATION_SKIN_COUNT, int);
	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_aiSkinID[i], int);
	}
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, AnimID, int);
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iAnimationFlags, int);
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fAnimationDelay, float);
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iNumberOfAttachments, int);
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, uHierarchyFlagsToSave, u_int);
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_uTextureHash, Asura_Hash_ID);
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_uIKTemplateID, Asura_Hash_ID);
	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_uFaceID, Asura_Hash_ID);
	#else
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, 0, Asura_Hash_ID);
	#endif

}

void Asura_Cutscene_Actor_Hierarchy::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Actor::WriteToChunkStream	(xStream);

	u_int uHierarchyFlagsToSave = m_uHierarchyFlags;
	#ifdef  ASURA_TOOLS
	if(m_aiSkinID[0] != -1)
	{
		//Automatically flag particle_bloke as not rendered. RAR 24/01/2006
		if(!Cutscene->SkinNames.GetName(m_aiSkinID[0]).Compare("particle_bloke"))
		{
			uHierarchyFlagsToSave |= ASURA_HIERARCHYACTORFLAG_INITIALLY_HIDDEN;
		}
	}
	#endif

	xStream << (int)ASURA_ANIMATION_SKIN_COUNT;
	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		xStream << m_aiSkinID[i];
	}
    xStream << AnimID;
    xStream << m_iAnimationFlags;
    xStream << m_fAnimationDelay;
    xStream << m_iNumberOfAttachments;
    xStream << uHierarchyFlagsToSave;
	xStream << m_uTextureHash;
	xStream << m_uIKTemplateID;
	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	xStream << m_uFaceID;
	#else
	xStream << (Asura_Hash_ID)0;
	#endif
}

void Asura_Cutscene_Actor_Hierarchy::LoadFromChunk(class Asura_Chunk_Actor* chunk,Asura_Handle fileHandle)
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
#ifndef ASURA_TOOLS
		// No longer bothering to support versions < 9 outside tools - they're
		// ancient.
		//And no longer in tools either , since they are even more ancient now. RAR 30/09/2008
		Asura_Assert( chunk->Version >= 9, "Attempting to load an ancient cutscene actor chunk , which is no longer supported" );
#endif

		Asura_Cutscene_Actor::LoadFromChunk(chunk,fileHandle);

		//See how many skin IDs were saved with the actor
		int iNumSavedSkins;
		if(chunk->Version<19)
		{
			iNumSavedSkins = 1;
		}
		else
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle, iNumSavedSkins, int);
		}
		int iNumSkinsToLoad = Asura_Maths::Min(iNumSavedSkins, ASURA_ANIMATION_SKIN_COUNT);

		for(int i=0; i<iNumSkinsToLoad; i++)
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle, m_aiSkinID[i], int);
		}

		//if the maximum number of skins has gone down , then any excess will need to be skipped
		for(int i=iNumSkinsToLoad; i<iNumSavedSkins; i++)
		{
			int iTempSkinID;
			ASURA_CHUNK_FILEEXTRACT(fileHandle, iTempSkinID, int);
			Asura_Assert(iTempSkinID == -1, "Cutscene actor was saved out using more skins than is currently allowed , and one of the excess skins was being used.");
		}

		ASURA_CHUNK_FILEEXTRACT(fileHandle, AnimID, int);


		ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iAnimationFlags, int);
		ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fAnimationDelay, float);

		ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iNumberOfAttachments, int);
		if(m_iNumberOfAttachments)
		{
			Asura_ResourceAllocateArray(m_ppxAttachments, m_iNumberOfAttachments);
			for(int i=0;i<m_iNumberOfAttachments;i++)
			{
				m_ppxAttachments[i] = 0;
			}
		}

		ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uHierarchyFlags, u_int);

		if(chunk->Version>=11)
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uTextureHash, Asura_Hash_ID);
		}

		if(chunk->Version>=18)
		{
			ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uIKTemplateID, Asura_Hash_ID);
		}
		if(chunk->Version>=21)
		{
			#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
			ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uFaceID, Asura_Hash_ID);
			#else
			Asura_FileAccess::SkipData(fileHandle, 4);
			#endif
		}
	}
	Asura_ResourcePopType();
}

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Asura_Cutscene_Actor_Shape **
**----------------------------*/

int Asura_Cutscene_Actor_Shape::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Actor::GetChunkSize();
	iChunkSize += sizeof(Asura_Hash_ID) + sizeof(u_int);

	return iChunkSize;
}

void Asura_Cutscene_Actor_Shape::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Actor::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_uShapeID, Asura_Hash_ID);
    ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_uShapeFlags, u_int);
}

void Asura_Cutscene_Actor_Shape::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Actor::WriteToChunkStream	(xStream);

	xStream << m_uShapeID;
    xStream << m_uShapeFlags;
}

void Asura_Cutscene_Actor_Shape::LoadFromChunk(class Asura_Chunk_Actor* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Actor::LoadFromChunk(pxChunk,iFileHandle);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_uShapeID, Asura_Hash_ID);

	if(pxChunk->Version>=10)
	{
        ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_uShapeFlags, u_int);
	}
}
////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------**
** Asura_Cutscene_Actor_Camera **
**-----------------------------*/
int Asura_Cutscene_Actor_Camera::GetChunkSize()
{
	Asura_Chunk_Dummy_Stream xStream;
	WriteToChunkStream(xStream);
	return xStream.GetPosition();
}

void Asura_Cutscene_Actor_Camera::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Actor::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILL_STRING(pcChunkBuffer, m_xTargetName);
	ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_fInitialFOV, Asura_Angle );

	m_xInitialShake.SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_uCameraFlags, u_int);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_fInitialPlayerControlYawLimit, float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_fInitialPlayerControlPitchLimit, float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_fNearClip, float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_fFarClip, float);
}

void Asura_Cutscene_Actor_Camera::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Actor::WriteToChunkStream	(xStream);

	xStream << m_xTargetName;
	xStream << m_fInitialFOV;

	m_xInitialShake.WriteToChunkStream(xStream);

	xStream << m_uCameraFlags;
	xStream << m_fInitialPlayerControlYawLimit;
	xStream << m_fInitialPlayerControlPitchLimit;
	xStream << m_fNearClip;
	xStream << m_fFarClip;
}

void Asura_Cutscene_Actor_Camera::LoadFromChunk(class Asura_Chunk_Actor* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Actor::LoadFromChunk(pxChunk,iFileHandle);

	if(pxChunk->Version>=12)
	{
        ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle, m_xTargetName);
	}
	else
	{
		//pre multiple camera support... so camera points at the lone camera target
        // (MP - 31.03.2006) Can't necessarily assume the target's name. For
        // example, in Rogue our kill move code happened to assign a
        // different one. Also I think artists have always been free to
        // rename them anyway. So instead we'll search, like the cutscene
        // code used to elsewhere. Note that we can assume that the target
        // will have been loaded before the camera, since that's the way
        // tools code sorts the actors.
        int i;
        const Asura_Cutscene* pxParentCutscene = this->Cutscene;
        const int iNumActors = pxParentCutscene->NumberOfActors;
        for( i = 0; i < iNumActors; ++i )
        {
            const Asura_Cutscene_Actor* pxNextActor = pxParentCutscene->Actors[i];
            if( pxNextActor && pxNextActor != this && pxNextActor->Type == ACAT_CameraTarget )
            {
                m_xTargetName = pxNextActor->ActorName;
                break;
            }
        }
        
        // Fall back to guessing the name otherwise.
        if( i == iNumActors )
        {
		    m_xTargetName = "Camera Target";
        }
    }

	if(pxChunk->Version>=13)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_fInitialFOV,float);
	}
	#ifdef ASURA_TOOLS
	else
	{
		//FOV with in cutscene for old versions , so copy from there
		m_fInitialFOV = Cutscene->m_fInitialFOV_UsedForUpdatingOldCutscenes;
	}
	#endif

	if(pxChunk->Version>=15)
	{
		m_xInitialShake.LoadFromChunk(iFileHandle);
	}

	if(pxChunk->Version >= 17)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_uCameraFlags, float);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_fInitialPlayerControlYawLimit, float);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_fInitialPlayerControlPitchLimit, float);
	}
	
	if(pxChunk->Version >= 22)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_fNearClip, float);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_fFarClip, float);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Cutscene_Track **
**----------------------*/

// (MH - 05.11.2002) the work is now done by track member functions.
int Asura_Cutscene_Track::GetChunkBufferSize()
{
	int iNumberOfPoints = GetNumberOfPoints();

	//determine size of chunk
	int iChunkSize = 0;
	iChunkSize += sizeof(int); // version
	iChunkSize += sizeof(int); // number of points
	iChunkSize += TrackName.GetSizeInInts()*4;
	iChunkSize += sizeof(Asura_Vector_3) * iNumberOfPoints;
	iChunkSize += sizeof(float) * iNumberOfPoints;
//	iChunkSize += sizeof(Asura_Matrix_3x3); //matrix no longer used , instead of array of quats
	iChunkSize += sizeof(Asura_Quat) * iNumberOfPoints;
	iChunkSize += sizeof(int) * iNumberOfPoints;
	iChunkSize += sizeof(unsigned);//for m_uTrackFlags
	iChunkSize += sizeof(int); // number of point durations
	if(PointDurations) iChunkSize += sizeof(float) * iNumberOfPoints;

    return iChunkSize;
}


void Asura_Cutscene_Track::FillInChunkBuffer( char*& pcChunkBuffer )
{
	Asura_Vector_3* points = GetPoints();
	int iNumberOfPoints = GetNumberOfPoints();

	// write out version number
	ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, ASURA_CHUNK_CUTSCENETRACK_VERSION, int );
	ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, iNumberOfPoints, int );

	pcChunkBuffer = (char*) TrackName.IntCopy((int*)(pcChunkBuffer));

	#ifdef ASURA_TOOLS
	//RAR 08/03/2006 For recentreing during level exports :
	const Asura_Vector_3& xPositionOffset = Tools_Level_Export::GetExportCentre();
	#else
	const Asura_Vector_3& xPositionOffset = Asura_Maths::ZeroVector3;
	#endif	

	//copy points
	int i;
	for(i=0;i<iNumberOfPoints;i++)
	{
		*(Asura_Vector_3*)pcChunkBuffer = points[i] - xPositionOffset;
		ASURA_CHUNK_ARRAY_VERIFY(*pcChunkBuffer, float, 3);
		pcChunkBuffer += sizeof(Asura_Vector_3);
	}

	//copy times
	for(i=0;i<iNumberOfPoints;i++)
	{
		*(float*)pcChunkBuffer = PointArrivalTimes[i];
		ASURA_CHUNK_VERIFY(*pcChunkBuffer, float);
		pcChunkBuffer += sizeof(float);
	}

	//copy flags
	for(i=0;i<iNumberOfPoints;i++)
	{
		*(int*)pcChunkBuffer = PointFlags[i];
		ASURA_CHUNK_VERIFY(*pcChunkBuffer, int);
		pcChunkBuffer += sizeof(int);
	}

	//copy point durations
	if(PointDurations)
	{
		// write out the number of durations to save
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, iNumberOfPoints, int );
		
		for(i=0;i<iNumberOfPoints;i++)
		{
			*(float*)pcChunkBuffer = PointDurations[i];
			ASURA_CHUNK_VERIFY(*pcChunkBuffer, float);
			pcChunkBuffer += sizeof(float);
		}
	}
	else
	{
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, 0, int );
	}

	*(unsigned*)pcChunkBuffer = m_uTrackFlags;
	ASURA_CHUNK_VERIFY(*pcChunkBuffer, unsigned);
	pcChunkBuffer += sizeof(unsigned);

	//copy orientations
	for(i=0;i<iNumberOfPoints;i++)
	{
		*(Asura_Quat*)pcChunkBuffer = m_pxPointOrientations[i];
		ASURA_CHUNK_ARRAY_VERIFY(*pcChunkBuffer, float, 4);
		pcChunkBuffer += sizeof(Asura_Quat);
	}
}

void Asura_Cutscene_Track::WriteToChunkStream ( Asura_Chunk_Stream& xStream ) const
{
	Asura_Vector_3*	pxPoints		= GetPoints ();
    int				iNumberOfPoints	= GetNumberOfPoints ();

	// Write the version number
	xStream << ASURA_CHUNK_CUTSCENETRACK_VERSION;

	// Write the number of points
	xStream << iNumberOfPoints;	

	// Write the name...
	xStream << TrackName;

	// Write the points...
	for (int i = 0; i < iNumberOfPoints; i++)
	{
		xStream << pxPoints [i];
	}

	// Write the times...
	for (int i = 0; i < iNumberOfPoints; i++)
	{
		xStream << PointArrivalTimes [i];
	}


	// Write the flags...
	for (int i = 0; i < iNumberOfPoints; i++)
	{
		xStream << PointFlags [i];
	}


	// Do we have any point durations?	
	if( PointDurations )
	{
		// Write of number of durations
		xStream << iNumberOfPoints; 

		// Write the durations...
		for( int i = 0; i < iNumberOfPoints; i++)
		{
			xStream << PointDurations [i];
		}
	}
	else
	{
		// no point durations	
		xStream << 0; 
	}


	// Write the track-flags...
	xStream << m_uTrackFlags;


	// Write the orientations...
	for (int i = 0; i < iNumberOfPoints; i++)
	{
		xStream << m_pxPointOrientations [i];
	}
}

void Asura_Cutscene_Track::InitialiseFromChunkBuffer( const char*& pcChunkBuffer )
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		int iVersion = 0;
		ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion, int );

		if( iVersion >= 7 )
		{			
			int iNumberOfPoints = 0;
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, iNumberOfPoints, int );

			//read in the name
			ASURA_CHUNK_EXTRACT_STRING( pcChunkBuffer, TrackName );

			Asura_ResourcePushType(ASURA_RESOURCE_TEMPORARY);
			{
				//and the points
				Asura_Vector_3* pxPoints;
				Asura_ResourceAllocateArray( pxPoints, iNumberOfPoints );

				for( int i=0; i<iNumberOfPoints; ++i )
				{
					ASURA_CHUNK_EXTRACT( pcChunkBuffer, pxPoints[i], Asura_Vector_3 );
				}

				SetPoints( iNumberOfPoints, pxPoints );
				Asura_ResourceDeallocateArray( pxPoints );
			}
			Asura_ResourcePopType();

			// now that SetPoints has allocated the data arrays, initialise the rest of the point data			
			for( int i=0; i<iNumberOfPoints; ++i )
			{
				//get the times
				ASURA_CHUNK_EXTRACT( pcChunkBuffer, PointArrivalTimes[i], float );
			}			

			for( int i=0; i<iNumberOfPoints; ++i )
			{
				//get the flags
				ASURA_CHUNK_EXTRACT( pcChunkBuffer, PointFlags[i], int );
			}			

			//get the durations
			int iNumberOfPointDurations = 0;
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, iNumberOfPointDurations, int );
			if( iNumberOfPointDurations )
			{
				Asura_Assert( iNumberOfPointDurations == iNumberOfPoints, "Number of durations should match number of points!" );

				Asura_ResourceAllocateArray( PointDurations, iNumberOfPointDurations );				
				for( int i=0; i<iNumberOfPointDurations; ++i )
				{
					ASURA_CHUNK_EXTRACT( pcChunkBuffer, PointDurations[i], float );
				}				
			}

			//get the flags
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_uTrackFlags, u_int );

			//get the orientations
			for( int i=0; i<iNumberOfPoints; ++i )
			{
				ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_pxPointOrientations[i], Asura_Quat );
			}
		}
		else
		{
			Asura_Assert(iVersion <= ASURA_CHUNK_CUTSCENETRACK_VERSION, "incompatible cutscene track chunk version");
		}

		//calculate the section lengths
		CalculateLengths();
	}
	Asura_ResourcePopType();	
}

void Asura_Cutscene_Track::LoadFromChunkBuffer(Asura_Handle iFileHandle, int iNumberOfPoints, int iVersion, int iChunkFlags)
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		Asura_Assert(iVersion <= ASURA_CHUNK_CUTSCENETRACK_VERSION, "incompatible cutscene track chunk version");

#ifndef ASURA_TOOLS
		// No longer bothering to support versions < 6 outside tools - they're
		// ancient.
		Asura_Assert( iVersion >= 6, "" );
#endif

		if( iVersion >= 7 )
		{
			// RDM - 18/04/06
			// read in a version number and the number of points
			// although we already know this data from the chunk header, storing it directly in the track's chunk 
			// allows us to save and load cutscene tracks from within other entities, etc.
			ASURA_CHUNK_FILEEXTRACT( iFileHandle, iVersion, int );
			ASURA_CHUNK_FILEEXTRACT( iFileHandle, iNumberOfPoints, int );
		}

		//read in the name
		TrackName = Asura_FileAccess::ReadString(iFileHandle);
		
		Asura_ResourcePushType(ASURA_RESOURCE_TEMPORARY);
		{
			//and the points
			Asura_Vector_3* points;
			Asura_ResourceAllocateArray(points, iNumberOfPoints);
			Asura_FileAccess::ReadData(iFileHandle,points,iNumberOfPoints*sizeof(Asura_Vector_3));

			SetPoints(iNumberOfPoints,points);
			Asura_ResourceDeallocateArray(points);
		}
		Asura_ResourcePopType();

#ifdef ASURA_TOOLS
		if(iVersion<2) goto FinishTrackInitialisation;
#endif

		//get the times
		Asura_FileAccess::ReadData(iFileHandle,PointArrivalTimes,iNumberOfPoints*sizeof(float));

#ifdef ASURA_TOOLS
		if(iVersion<3) goto FinishTrackInitialisation;
#endif

#ifdef ASURA_TOOLS
		//get the orientation
		if(iVersion<6)
		{
			//copy matrix to first point's orientation
			Asura_Matrix_3x3 xMat;
			ASURA_CHUNK_FILEEXTRACT(iFileHandle,xMat,Asura_Matrix_3x3);
			SetOrientationAtPoint(0,xMat);

		}
#endif

#ifdef ASURA_TOOLS
		if(iVersion<4) goto FinishTrackInitialisation;
#endif

		//get the flags
		Asura_FileAccess::ReadData(iFileHandle,PointFlags,iNumberOfPoints*sizeof(int));

		//get the durations
		if( iVersion >= 7 )
		{
			int iNumberOfPointDurations = 0;
			ASURA_CHUNK_FILEEXTRACT( iFileHandle, iNumberOfPointDurations, int );

			if( iNumberOfPointDurations > 0 )
			{
				Asura_Assert( iNumberOfPointDurations == iNumberOfPoints, "Number of point durations should match number of points!" );
				Asura_ResourceAllocateArray(PointDurations, iNumberOfPointDurations);
				Asura_FileAccess::ReadData(iFileHandle,PointDurations,iNumberOfPointDurations*sizeof(float));			
			}
		}
		else
		{
			// old versions used a chunk flag bitfield 
			if( iChunkFlags & Asura_Chunk_CutsceneTrack::ChunkFlag_HasDurationArray )
			{
				Asura_ResourceAllocateArray(PointDurations, iNumberOfPoints);
				Asura_FileAccess::ReadData(iFileHandle,PointDurations,iNumberOfPoints*sizeof(float));
			}
		}

#ifdef ASURA_TOOLS
		if(iVersion<5) goto FinishTrackInitialisation;
#endif

		//get the flags
		ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uTrackFlags,unsigned);

#ifdef ASURA_TOOLS
		if(iVersion<6)	goto FinishTrackInitialisation;
#endif

		//get the orientations
		Asura_FileAccess::ReadData(iFileHandle,m_pxPointOrientations,iNumberOfPoints*sizeof(Asura_Quat));

		//calculate the section lengths
#ifdef ASURA_TOOLS
FinishTrackInitialisation :
#endif

		CalculateLengths();
	}
	Asura_ResourcePopType();
}

////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Chunk_CutsceneTrack **
**---------------------------*/

bool Asura_Chunk_CutsceneTrack::Process(Asura_Handle iFileHandle)
{
	if(!Asura_Chunk_Cutscene::CurrentCutscene)
	{
		return false;
	}
	
	//get the track
	Asura_Cutscene_Track*& track = Asura_Chunk_Actor::GetNextTrack();

	//allocate
	track = new Asura_Cutscene_Track;

    track->LoadFromChunkBuffer(iFileHandle, NumberOfPoints, Version, Flags);

	if(Version<6 && Asura_Chunk_Actor::CurrentActor)
	{
		//need to set defaults for the new track flags
		track->m_uTrackFlags |= Asura_Chunk_Actor::CurrentActor->GetDefaultTrackFlags();	
	}

	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_CutsceneTrack::Create(Asura_Cutscene_Track* pxTrack,Asura_SaveBlock& saveBlock)
{
    int iChunkSize = sizeof(Asura_Chunk_CutsceneTrack) + pxTrack->GetChunkBufferSize();

	//allocate space for chunk
	Asura_Chunk_CutsceneTrack *pxChunk = (Asura_Chunk_CutsceneTrack*)saveBlock.GetAccessPointer(iChunkSize);

	//fill in base stuff
	pxChunk->ID = ASURA_CHUNK_CUTSCENETRACK;
	ASURA_CHUNK_VERIFY(pxChunk->ID, int);
	pxChunk->Size = iChunkSize;
	ASURA_CHUNK_VERIFY(pxChunk->Size, int);
	pxChunk->Version = ASURA_CHUNK_CUTSCENETRACK_VERSION;
	ASURA_CHUNK_VERIFY(pxChunk->Version, int);
    pxChunk->Flags = pxTrack->PointDurations ? ChunkFlag_HasDurationArray : 0;
	ASURA_CHUNK_VERIFY(pxChunk->Flags, int);

	// RDM - 18/04/06
	// this data is now redundant as far as the track itself is concerned, 
	// but is used by old-style chunk loading code...
    int iNumberOfPoints = pxTrack->GetNumberOfPoints();
	pxChunk->NumberOfPoints = iNumberOfPoints;
	ASURA_CHUNK_VERIFY(pxChunk->NumberOfPoints, int);

	//copy name
    char* pcChunkBuffer = reinterpret_cast<char*>(pxChunk + 1);
    pxTrack->FillInChunkBuffer( pcChunkBuffer );

    // check we wrote the right amount of data
    Asura_Assert(int(pcChunkBuffer - (char*)(pxChunk)) == iChunkSize,
        "GetChunkBufferSize returned incorrect value?");
}

void Asura_Chunk_CutsceneTrack::Create (Asura_Cutscene_Track* pxTrack, Asura_Chunk_Stream& xStream)
{
	int	iFlags			= (pxTrack->PointDurations) ? ChunkFlag_HasDurationArray : 0;
    int	iNumberOfPoints = pxTrack->GetNumberOfPoints ();


	// Write the chunk-data...

	xStream.WriteChunk_Start	(ASURA_CHUNK_CUTSCENETRACK, ASURA_CHUNK_CUTSCENETRACK_VERSION, iFlags);

	xStream << iNumberOfPoints;
	xStream << iFlags;

	pxTrack->WriteToChunkStream	( xStream );
	xStream.WriteChunk_End		();
}
#endif
////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_Chunk_CutsceneAttachment **
**--------------------------------*/
#define ASURA_CHUNK_CUTSCENEATTACHMENT_VERSION 2

bool Asura_Chunk_CutsceneAttachment::Process(Asura_Handle iFileHandle)
{
	Asura_Assert(Version<=ASURA_CHUNK_CUTSCENEATTACHMENT_VERSION,"Unknown chunk version");
	if(!Asura_Chunk_Cutscene::CurrentCutscene)
	{
		return false;
	}

	//get the attachment
	Asura_Cutscene_Attachment*& pxAttachment = Asura_Chunk_Actor::GetNextAttachment();

	//allocate
	pxAttachment = Asura_Cutscene_Attachment::AllocateAttachment((Asura_Cutscene_Actor_Hierarchy*)Asura_Chunk_Actor::CurrentActor,m_iType);
	Asura_Assert(pxAttachment,"");

	pxAttachment->LoadFromChunk(this,iFileHandle);
	
	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_CutsceneAttachment::Create(Asura_Cutscene_Attachment* pxAttachment,Asura_SaveBlock& xSaveBlock)
{
	//determine size of chunk
	int iChunkSize = sizeof(Asura_Chunk_CutsceneAttachment);
	iChunkSize += pxAttachment->GetChunkSize();

	//allocate chunk
	Asura_Chunk_CutsceneAttachment* pxChunk = (Asura_Chunk_CutsceneAttachment*) xSaveBlock.GetAccessPointer(iChunkSize);

	pxChunk->ID = ASURA_CHUNK_CUTSCENEATTACHMENT;
	ASURA_CHUNK_VERIFY(pxChunk->ID, int);
	pxChunk->Size = iChunkSize;
	ASURA_CHUNK_VERIFY(pxChunk->Size, int);
	pxChunk->Version = ASURA_CHUNK_CUTSCENEATTACHMENT_VERSION;
	ASURA_CHUNK_VERIFY(pxChunk->Version, int);
	pxChunk->Flags = 0;
	ASURA_CHUNK_VERIFY(pxChunk->Flags, int);
	pxChunk->m_iType = pxAttachment->GetType();
	ASURA_CHUNK_VERIFY(pxChunk->m_iType, int);

	char* pcChunkBuffer = (char*)(pxChunk+1);
	pxAttachment->SaveToChunk(pcChunkBuffer);
}

void Asura_Chunk_CutsceneAttachment::Create (Asura_Cutscene_Attachment* pxAttachment, Asura_Chunk_Stream& xStream)
{
	int	iType	= pxAttachment->GetType ();


	// Write the chunk-data...

	xStream.WriteChunk_Start	(ASURA_CHUNK_CUTSCENEATTACHMENT, ASURA_CHUNK_CUTSCENEATTACHMENT_VERSION, 0);

	xStream << iType;

	pxAttachment->WriteToChunkStream	(xStream);
	xStream.WriteChunk_End				();
}
#endif
////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Cutscene_Attachment **
**---------------------------*/
int Asura_Cutscene_Attachment::GetChunkSize()
{
	if(m_xMarkerName.IsNull()) m_xMarkerName = "";
	if(m_xName.IsNull()) m_xName = "";

	int iChunkSize = m_xMarkerName.GetSizeInInts()*4;
	iChunkSize += m_xName.GetSizeInInts()*4;
	iChunkSize += sizeof(u_int)*2;//for m_uID,m_uFlags
	return iChunkSize;
}

void Asura_Cutscene_Attachment::SaveToChunk(char*& pcChunkBuffer)
{
	ASURA_CHUNK_FILL_STRING(pcChunkBuffer,m_xMarkerName);
	ASURA_CHUNK_FILL_STRING(pcChunkBuffer,m_xName);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uID,u_int);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uFlags,u_int);
}

void Asura_Cutscene_Attachment::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	xStream << m_xMarkerName;
	xStream << m_xName;
	xStream << m_uID;
	xStream << m_uFlags;
}

void Asura_Cutscene_Attachment::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle,m_xMarkerName);
	if(pxChunk->Version<1) return;

	ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle,m_xName);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uID,u_int);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uFlags,u_int);

}
////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------------**
** Asura_Cutscene_Attachment_Hierarchy **
**-------------------------------------*/

int Asura_Cutscene_Attachment_Hierarchy::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Attachment::GetChunkSize();
	iChunkSize += 2 * sizeof(int)
				+ ASURA_ANIMATION_SKIN_COUNT*sizeof(int);
	return iChunkSize;
}

void Asura_Cutscene_Attachment_Hierarchy::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Attachment::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,ASURA_ANIMATION_SKIN_COUNT,int);
	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_aiSkinID[i],int);
	}
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_iAnimID,int);
}

void Asura_Cutscene_Attachment_Hierarchy::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Attachment::WriteToChunkStream	(xStream);

	xStream << (int) ASURA_ANIMATION_SKIN_COUNT;
	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		xStream << m_aiSkinID[i];
	}
	xStream << m_iAnimID;
}

void Asura_Cutscene_Attachment_Hierarchy::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Attachment::LoadFromChunk(pxChunk,iFileHandle);

	//See how many skin IDs were saved with the attachment
	int iNumSavedSkins = 1;
	if(pxChunk->Version >= 2)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, iNumSavedSkins, int);
	}
	int iNumSkinsToLoad = Asura_Maths::Min(iNumSavedSkins, ASURA_ANIMATION_SKIN_COUNT);

	for(int i=0; i<iNumSkinsToLoad; i++)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_aiSkinID[i], int);
	}

	//if the maximum number of skins has gone down , then any excess will need to be skipped
	for(int i=iNumSkinsToLoad; i<iNumSavedSkins; i++)
	{
		int iTempSkinID;
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, iTempSkinID, int);
		Asura_Assert(iTempSkinID == -1, "Cutscene attachment was saved out using more skins than is currently allowed , and one of the excess skins was being used.");
	}

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_iAnimID,int);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_Cutscene_Attachment_Shape **
**---------------------------------*/

int Asura_Cutscene_Attachment_Shape::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Attachment::GetChunkSize();
	iChunkSize += sizeof(Asura_Hash_ID);
	return iChunkSize;
}

void Asura_Cutscene_Attachment_Shape::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Attachment::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xShapeInstance.GetShapeHashID(),Asura_Hash_ID);
}

void Asura_Cutscene_Attachment_Shape::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Hash_ID	xHashID	= m_xShapeInstance.GetShapeHashID ();


	Asura_Cutscene_Attachment::WriteToChunkStream	(xStream);

	xStream << xHashID;
}

void Asura_Cutscene_Attachment_Shape::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Attachment::LoadFromChunk(pxChunk,iFileHandle);

	Asura_Hash_ID uID;
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,uID,Asura_Hash_ID);

	m_xShapeInstance.SetShape(uID);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_Cutscene_Attachment_Actor **
**---------------------------------*/

int Asura_Cutscene_Attachment_Actor::GetChunkSize()
{
	if(m_xAttachedActorName.IsNull()) m_xAttachedActorName = "";

	int iChunkSize = Asura_Cutscene_Attachment::GetChunkSize();
	ASURA_CHUNK_SIZE_STRING(iChunkSize,m_xAttachedActorName);

	return iChunkSize;
}

void Asura_Cutscene_Attachment_Actor::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Attachment::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILL_STRING(pcChunkBuffer,m_xAttachedActorName);
}

void Asura_Cutscene_Attachment_Actor::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Attachment::WriteToChunkStream	(xStream);

	xStream << m_xAttachedActorName;
}

void Asura_Cutscene_Attachment_Actor::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Attachment::LoadFromChunk(pxChunk,iFileHandle);

	ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle,m_xAttachedActorName);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_Cutscene_Attachment_Entity **
**----------------------------------*/

int Asura_Cutscene_Attachment_Entity::GetChunkSize()
{
	int iChunkSize = PARENT::GetChunkSize();

	ASURA_CHUNK_SIZE( iChunkSize, m_uEntityGUID, Asura_Guid );

	return iChunkSize;
}

void Asura_Cutscene_Attachment_Entity::SaveToChunk(char*& pcChunkBuffer)
{
	PARENT::SaveToChunk( pcChunkBuffer );

	ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_uEntityGUID, Asura_Guid );
}

void Asura_Cutscene_Attachment_Entity::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	PARENT::WriteToChunkStream	(xStream);

	xStream << m_uEntityGUID;
}

void Asura_Cutscene_Attachment_Entity::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	PARENT::LoadFromChunk( pxChunk, iFileHandle );

	ASURA_CHUNK_FILEEXTRACT( iFileHandle, m_uEntityGUID, Asura_Guid );
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
	m_uEntityGUID	= Asura_Streaming_System::MakeCurrentMasterZoneGuid(m_uEntityGUID);
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
}

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_Cutscene_Attachment_Corona **
**----------------------------------*/

int Asura_Cutscene_Attachment_Corona::GetChunkSize()
{

	int iChunkSize = Asura_Cutscene_Attachment::GetChunkSize();
	iChunkSize += 4*sizeof(float) + sizeof(int) + sizeof(unsigned);

	return iChunkSize;
}

void Asura_Cutscene_Attachment_Corona::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Attachment::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fRed,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fGreen,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fBlue,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fSize,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uCoronaFlags,unsigned);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_iTextureIDForSave,int);
}

void Asura_Cutscene_Attachment_Corona::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Attachment::WriteToChunkStream	(xStream);

	xStream << m_fRed;
	xStream << m_fGreen;
	xStream << m_fBlue;
	xStream << m_fSize;
	xStream << m_uCoronaFlags;
	xStream << m_iTextureIDForSave;
}

void Asura_Cutscene_Attachment_Corona::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Attachment::LoadFromChunk(pxChunk,iFileHandle);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fRed,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fGreen,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fBlue,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fSize,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uCoronaFlags,unsigned);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_iTextureID,int);

	//convert the texture ID
	if(m_iTextureID!=ASURA_TEXTURE_NONE)
	{
		m_iTextureID = Asura_Texture_Management::ConvertIndex( m_iTextureID );
	}
	
}


////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Cutscene_Attachment_PFX **
**--------------------------------*/

int Asura_Cutscene_Attachment_PFX::GetChunkSize()
{

	int iChunkSize = Asura_Cutscene_Attachment::GetChunkSize();
	iChunkSize += sizeof(Asura_Hash_ID);

	return iChunkSize;
}

void Asura_Cutscene_Attachment_PFX::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Attachment::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uEffectTemplateHash,Asura_Hash_ID);
}

void Asura_Cutscene_Attachment_PFX::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Attachment::WriteToChunkStream	(xStream);

	xStream << m_uEffectTemplateHash;
}

void Asura_Cutscene_Attachment_PFX::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Attachment::LoadFromChunk(pxChunk,iFileHandle);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uEffectTemplateHash,Asura_Hash_ID);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_Cutscene_Attachment_SoundEvent **
**--------------------------------------*/

int Asura_Cutscene_Attachment_SoundEvent::GetChunkSize()
{

	int iChunkSize = Asura_Cutscene_Attachment::GetChunkSize();
	iChunkSize += sizeof(Asura_Hash_ID);

	return iChunkSize;
}

void Asura_Cutscene_Attachment_SoundEvent::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Attachment::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uSoundEvent,Asura_Hash_ID);
}

void Asura_Cutscene_Attachment_SoundEvent::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Attachment::WriteToChunkStream	(xStream);

	xStream << m_uSoundEvent;
}

void Asura_Cutscene_Attachment_SoundEvent::LoadFromChunk(const Asura_Chunk_CutsceneAttachment* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Attachment::LoadFromChunk(pxChunk,iFileHandle);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uSoundEvent,Asura_Hash_ID);
}



////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Chunk_CutsceneEvent **
**---------------------------*/

// MH  17/06/05 Version 12 - Added template event and wrapper name (Tools data)
// RAR 04/11/08 Version 13 - Added anim delay to animation change events
// RAR 28/04/09 Version 14 - Added flags for facial animation event
#define ASURA_CHUNK_CUTSCENEEVENT_VERSION 14

bool Asura_Chunk_CutsceneEvent::Process(Asura_Handle fileHandle)
{
	Asura_Assert(Version<=ASURA_CHUNK_CUTSCENEEVENT_VERSION,"Unknown chunk version");

	if(!Asura_Chunk_Cutscene::CurrentCutscene)
	{
		return false;
	}

	if(Type == ACET_Obsolete_NextScriptLine || Type == ACET_Obsolete_DepthOfField)
	{
		//This event is no longer supported , so just skip it
		//Any events that refer to other by index may be messed up by this. But this event hasn't been available for 3+ years,
		//so there shouldn't be any instances to worry about.
		//RAR 22/01/2009
		//Also doing this for the old depth of field system. This has never worked on PS3/360, so again there hopefully won't
		//be any instances to worry about in current levels.
		//AAM 03/03/2009
		Asura_Assert(0,"Obsolete cutscene event being removed. This could possibly mess up LoopBack events in the same cutscene");
		Asura_Chunk_Cutscene::CurrentCutscene->NumberOfEvents--;
		return false;
	}
	
	//get the event
	Asura_Cutscene_Event*& event = Asura_Chunk_Cutscene::GetNextEvent();

	event = Asura_Cutscene_Event::AllocateEvent(Type,Asura_Chunk_Cutscene::CurrentCutscene);
	Asura_Assert(event,"");

	event->m_fTriggerTime = TriggerTime;
		
	event->LoadFromChunk(this,fileHandle);
	
	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_CutsceneEvent::Create(Asura_Cutscene_Event* event,Asura_SaveBlock& saveBlock)
{
	int chunkSize = event->GetChunkSize();

	Asura_Chunk_CutsceneEvent *chunk = (Asura_Chunk_CutsceneEvent*)saveBlock.GetAccessPointer(chunkSize);

	chunk->ID = ASURA_CHUNK_CUTSCENEEVENT;
	ASURA_CHUNK_VERIFY(chunk->ID, int);
	chunk->Size = chunkSize;
	ASURA_CHUNK_VERIFY(chunk->Size, int);
	chunk->Version = ASURA_CHUNK_CUTSCENEEVENT_VERSION;
	ASURA_CHUNK_VERIFY(chunk->Version, int);
	chunk->Flags = 0;
	ASURA_CHUNK_VERIFY(chunk->Flags, int);

	chunk->Type = (int) event->m_iType;
	ASURA_CHUNK_VERIFY(chunk->Type, int);
	chunk->TriggerTime = event->m_fTriggerTime;
	ASURA_CHUNK_VERIFY(chunk->TriggerTime, float);

	char* chunkBuffer = (char*)(chunk+1);

	event->SaveToChunk(chunkBuffer);
}

void Asura_Chunk_CutsceneEvent::Create (Asura_Cutscene_Event* event, Asura_Chunk_Stream& xStream)
{
	// Write the chunk-data...

	xStream.WriteChunk_Start	(ASURA_CHUNK_CUTSCENEEVENT, ASURA_CHUNK_CUTSCENEEVENT_VERSION, 0);

	xStream << event->m_iType;
	xStream << event->m_fTriggerTime;

	event->WriteToChunkStream	(xStream);
	xStream.WriteChunk_End		();
}
#endif

///////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Cutscene_Event **
**----------------------*/
int Asura_Cutscene_Event::GetChunkSize()
{
	int chunkSize = sizeof(Asura_Chunk_CutsceneEvent);
	ASURA_CHUNK_SIZE_STRING(chunkSize, m_xTargetActorName)
	ASURA_CHUNK_SIZE(chunkSize, m_iParentEventIndex, int)
	ASURA_CHUNK_SIZE_STRING(chunkSize, m_xAnimationFlagActorName)
	ASURA_CHUNK_SIZE(chunkSize, m_iAnimationFlag, int)
	ASURA_CHUNK_SIZE(chunkSize,m_uEventFlags,u_int)
	ASURA_CHUNK_SIZE(chunkSize,m_iEnableID,int)
	chunkSize += sizeof(int);//for tools data size
	#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())
	{
		chunkSize += GetToolsDataSize();
	}
	#endif

	return chunkSize;
}

#ifdef ASURA_TOOLS
int Asura_Cutscene_Event::GetToolsDataSize() const
{
	int iChunkSize = sizeof(Asura_Vector_3);
	ASURA_CHUNK_SIZE_STRING(iChunkSize,m_xName);
	ASURA_CHUNK_SIZE_STRING(iChunkSize,m_xComment);
	ASURA_CHUNK_SIZE_STRING(iChunkSize,m_xNameOfSourceTemplate);
	ASURA_CHUNK_SIZE_STRING(iChunkSize,m_xNameOfEventWrapper);
	return iChunkSize;
}
#endif

void Asura_Cutscene_Event::SaveToChunk(char*& chunkBuffer)
{
	ASURA_CHUNK_FILL_STRING(chunkBuffer, m_xTargetActorName)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iParentEventIndex, int)
	ASURA_CHUNK_FILL_STRING(chunkBuffer, m_xAnimationFlagActorName)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iAnimationFlag, int)
	ASURA_CHUNK_BITFIELD_FILLANDVERIFY(chunkBuffer,m_uEventFlags,u_int)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iEnableID, int)

	#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())
	{
		//if not a level export , save all the tools data
		int iToolsDataSize = GetToolsDataSize();

		ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,iToolsDataSize,int);
		ASURA_CHUNK_FILL_STRING(chunkBuffer,m_xName);
		ASURA_CHUNK_FILL_STRING(chunkBuffer,m_xComment);
		ASURA_CHUNK_ARRAY_FILLANDVERIFY(chunkBuffer,m_xPosition,Asura_Vector_3, float, 3);
		ASURA_CHUNK_FILL_STRING(chunkBuffer,m_xNameOfSourceTemplate);
		ASURA_CHUNK_FILL_STRING(chunkBuffer,m_xNameOfEventWrapper);
	}
	else
	#endif
	{
		//In level export , or not in tools build (probably wouldn't be saved in that case anyway) ,
		//so just fill in with no tools data.
		ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,0,int);
	}
}

void Asura_Cutscene_Event::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xFlagsBitfield (m_uEventFlags);


	// Create the bitfield...

	xFlagsBitfield.Add	(m_bIgnoreWhenAdvancingToTime,	1);
	xFlagsBitfield.Add	(m_bDisabled,	1);


	// Write the data...

	xStream << m_xTargetActorName;
	xStream << m_iParentEventIndex;
	xStream << m_xAnimationFlagActorName;
	xStream << m_iAnimationFlag;
	xStream << xFlagsBitfield;
	xStream << m_iEnableID;

#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())
	{
		const int iToolsDataSize = GetToolsDataSize();

		//if not a level export , save all the tools data
		xStream << iToolsDataSize;
		xStream << m_xName;
		xStream << m_xComment;
		xStream << m_xPosition;
		xStream << m_xNameOfSourceTemplate;
		xStream << m_xNameOfEventWrapper;
	}
	else
#endif
	{
		int	iZero	= 0;

		//In level export , or not in tools build (probably wouldn't be saved in that case anyway) ,
		//so just fill in with no tools data.
		xStream << iZero;
	}
}

void Asura_Cutscene_Event::ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Chunk_Stream& xStream)
{
	xStream >> m_xTargetActorName;
	if(pxChunk->Version>=4)
	{
		xStream >> m_iParentEventIndex;
	}
	if (pxChunk->Version >= 8)
	{
		xStream >> m_xAnimationFlagActorName;
		xStream >> m_iAnimationFlag;
	}
	if(pxChunk->Version >= 10)
	{
		xStream >> m_uEventFlags;
	}
	if(pxChunk->Version >= 11)
	{
		xStream >> m_iEnableID;
	}
	//find the actors from the name
	if(m_xTargetActorName.GetLength())
	{
		m_pxTargetActor = Asura_Chunk_Cutscene::GetActor(m_xTargetActorName);
	}
	if(m_xAnimationFlagActorName.GetLength())
	{
		m_pxAnimationFlagActor = Asura_Chunk_Cutscene::GetActor(m_xAnimationFlagActorName);
	}
	if(pxChunk->Version >= 9)
	{
		int iToolsDataSize;
		xStream >> iToolsDataSize;
		if(iToolsDataSize)
		{
#ifdef ASURA_TOOLS
			//read in tools data
			xStream >> m_xName;
			xStream >> m_xComment;
			xStream >> m_xPosition;
			if (pxChunk->Version >= 12)
			{
				xStream >> m_xNameOfSourceTemplate;
				xStream >> m_xNameOfEventWrapper;
			}
#else
			xStream.Skip(iToolsDataSize);
#endif
		}
	}

}

void Asura_Cutscene_Event::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle)
{
	Asura_Chunk_File_Stream xStream(iFileHandle);
	ReadFromChunkStream(pxChunk, xStream);
}

///////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_Cutscene_Event_AnimationChange **
**--------------------------------------*/
int Asura_Cutscene_Event_AnimationChange::GetChunkSize()
{
//	if(AnimationFileName.IsNull()) AnimationFileName = "";

	int chunkSize = Asura_Cutscene_Event::GetChunkSize();


	ASURA_CHUNK_SIZE(chunkSize, m_iAnimationFlags, int)	
	ASURA_CHUNK_SIZE(chunkSize, m_fTweenDuration, float)	
    ASURA_CHUNK_SIZE(chunkSize, m_iAnimationIndex, int)
    ASURA_CHUNK_SIZE(chunkSize, m_fStartTimer, float)
    ASURA_CHUNK_SIZE(chunkSize, m_fEndTimer, float)
    ASURA_CHUNK_SIZE(chunkSize, m_fAnimDelay, float)

	return chunkSize;
}

void Asura_Cutscene_Event_AnimationChange::SaveToChunk(char*& chunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(chunkBuffer);
								 
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iAnimationFlags, int)	
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fTweenDuration, float)	
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iAnimationIndex, int)	
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fStartTimer, float)
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fEndTimer, float)
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fAnimDelay, float)

}

void Asura_Cutscene_Event_AnimationChange::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);
								 
	xStream << m_iAnimationFlags;	
	xStream << m_fTweenDuration;
	xStream << m_iAnimationIndex;
    xStream << m_fStartTimer;
    xStream << m_fEndTimer;
	xStream << m_fAnimDelay;
}

void Asura_Cutscene_Event_AnimationChange::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(chunk,fileHandle);

	if (chunk->Version < 5)
    {
	    Asura_String AnimationName;
        Asura_String AnimationFileName = "";

        ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iAnimationFlags, int)	
        ASURA_CHUNK_FILEEXTRACT_STRING(fileHandle,AnimationName)	

	    if(chunk->Version>=2)
		{
	    	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fTweenDuration, float)	
		}
 	    if(chunk->Version>=3)
		{
	    	ASURA_CHUNK_FILEEXTRACT_STRING(fileHandle,AnimationFileName)	
		}

		//add the animation name/filename to the lists
		if(!AnimationName.IsBlank())
		{
        	m_iAnimationIndex = Asura_Chunk_Cutscene::CurrentCutscene->AnimationNames.AddName(AnimationName);
		   	#ifdef ASURA_ASSERT
			int filenameindex = 
			#endif
			Asura_Chunk_Cutscene::CurrentCutscene->AnimationFileNames.AddName(AnimationFileName);
        	Asura_Assert(filenameindex == m_iAnimationIndex, "");
		}


    }
    else
    {
	    ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iAnimationFlags, int)	
	    ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fTweenDuration, float)	
	    ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iAnimationIndex, int)	
    }

    if(chunk->Version>=7)
	{
    	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fStartTimer, float)
    	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fEndTimer, float)
	}
	if(chunk->Version>=13)
	{
    	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fAnimDelay, float)
	}

}




///////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------------**
** Asura_Cutscene_Event_SecondaryAnimation **
**-----------------------------------------*/

int Asura_Cutscene_Event_SecondaryAnimation::GetChunkSize()
{
	int chunkSize = Asura_Cutscene_Event_AnimationChange::GetChunkSize();

	ASURA_CHUNK_SIZE(chunkSize, m_fWeight, float)
	ASURA_CHUNK_SIZE(chunkSize, m_fDuration, float)

	return chunkSize;
}

void Asura_Cutscene_Event_SecondaryAnimation::SaveToChunk(char*& chunkBuffer)
{
  	Asura_Cutscene_Event_AnimationChange::SaveToChunk(chunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fWeight, float)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fDuration, float)
}

void Asura_Cutscene_Event_SecondaryAnimation::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event_AnimationChange::WriteToChunkStream	(xStream);

	xStream << m_fWeight;
	xStream << m_fDuration;
}

void Asura_Cutscene_Event_SecondaryAnimation::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle)
{
	Asura_Cutscene_Event_AnimationChange::LoadFromChunk(chunk,fileHandle);
	
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fWeight, float)
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fDuration, float)

}

///////////////////////////////////////////////////////////////////////////////////
/*---------------------------------------**
** Asura_Cutscene_Event_TextureAnimation **
**---------------------------------------*/
int Asura_Cutscene_Event_TextureAnimation::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize();


	return iChunkSize + sizeof(int) + sizeof(float) + sizeof(u_int);
}

void Asura_Cutscene_Event_TextureAnimation::SaveToChunk(char*& chunkBuffer)
{
  	Asura_Cutscene_Event::SaveToChunk(chunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iSequence, int)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fSpeed, float)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_uTextureAnimFlags, u_int)
}

void Asura_Cutscene_Event_TextureAnimation::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_iSequence;
	xStream << m_fSpeed;
	xStream << m_uTextureAnimFlags;
}

void Asura_Cutscene_Event_TextureAnimation::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(chunk,fileHandle);
	
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iSequence, int)
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fSpeed, float)
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uTextureAnimFlags, u_int)

}


///////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_Cutscene_Event_TimeScale **
**--------------------------------*/
int Asura_Cutscene_Event_TimeScale::GetChunkSize()
{
	int chunkSize = Asura_Cutscene_Event::GetChunkSize();

	ASURA_CHUNK_SIZE(chunkSize, m_fTimeScale, float)

	return chunkSize;
}

void Asura_Cutscene_Event_TimeScale::SaveToChunk(char*& chunkBuffer)
{
  	Asura_Cutscene_Event::SaveToChunk(chunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fTimeScale, float)
}

void Asura_Cutscene_Event_TimeScale::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_fTimeScale;
}

void Asura_Cutscene_Event_TimeScale::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(chunk,fileHandle);
	
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fTimeScale, float)
}


///////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_Cutscene_Event_FieldOfView **
**----------------------------------*/
int Asura_Cutscene_Event_FieldOfView::GetChunkSize()
{
	int chunkSize = Asura_Cutscene_Event::GetChunkSize();

	ASURA_CHUNK_SIZE(chunkSize, m_fDuration, float)
	ASURA_CHUNK_SIZE(chunkSize, m_fFinalFOV, float)

	return chunkSize;
}

void Asura_Cutscene_Event_FieldOfView::SaveToChunk(char*& chunkBuffer)
{
  	Asura_Cutscene_Event::SaveToChunk(chunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fDuration, float)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fFinalFOV, float)
}

void Asura_Cutscene_Event_FieldOfView::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_fDuration;
	xStream << m_fFinalFOV;
}

void Asura_Cutscene_Event_FieldOfView::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(chunk,fileHandle);
	
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fDuration, float)
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fFinalFOV, float)
}
///////////////////////////////////////////////////////////////////////////////////
/*------------------------------**
** Asura_Cutscene_Event_Trigger **
**------------------------------*/

//for those bits that aren't nice multiples of 4 bytes
struct TriggerEventSaveBlock
{
	unsigned short MessageBlock;
	unsigned short Pad;
};

int Asura_Cutscene_Event_Trigger::GetChunkSize()
{
	int chunkSize = Asura_Cutscene_Event::GetChunkSize();

	chunkSize+= sizeof(TriggerEventSaveBlock);

	return chunkSize;
}

void Asura_Cutscene_Event_Trigger::SaveToChunk(char*& chunkBuffer)
{
  	Asura_Cutscene_Event::SaveToChunk(chunkBuffer);

	TriggerEventSaveBlock saveBlock;
	memset(&saveBlock,0,sizeof(saveBlock));
	saveBlock.MessageBlock = m_sMessageBlock;

	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,saveBlock.MessageBlock,unsigned short)
	ASURA_CHUNK_FILLANDVERIFY(chunkBuffer,saveBlock.Pad,unsigned short)
}

void Asura_Cutscene_Event_Trigger::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	TriggerEventSaveBlock saveBlock;
	memset(&saveBlock,0,sizeof(saveBlock));
	saveBlock.MessageBlock = m_sMessageBlock;

	xStream << saveBlock.MessageBlock;
	xStream << saveBlock.Pad;
}

void Asura_Cutscene_Event_Trigger::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(chunk,fileHandle);
	
	TriggerEventSaveBlock saveBlock;
	ASURA_CHUNK_FILEEXTRACT(fileHandle,saveBlock,TriggerEventSaveBlock)

	m_sMessageBlock = saveBlock.MessageBlock;

	#ifdef ASURA_TOOLS
	//Make sure the message block is in the cutscene message block set. RAR 28/03/03
	m_sMessageBlock = Asura_StaticMessage_System::TransferMessageBlock(m_sMessageBlock,ASURA_STATICMESSAGE_CUTSCENESET);
	#endif
}

///////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Cutscene_Event_LoopBack **
**-------------------------------*/
int Asura_Cutscene_Event_LoopBack::GetChunkSize(void)
{
	int chunkSize = Asura_Cutscene_Event::GetChunkSize();

	ASURA_CHUNK_SIZE(chunkSize, m_fNumSecondsToLoopBack, float);
	ASURA_CHUNK_SIZE(chunkSize, m_iEventToLoopBackTo, int);

	return chunkSize;
}

void Asura_Cutscene_Event_LoopBack::SaveToChunk(char*& chunkBuffer)
{
  	Asura_Cutscene_Event::SaveToChunk(chunkBuffer);
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_fNumSecondsToLoopBack, float);
    ASURA_CHUNK_FILLANDVERIFY(chunkBuffer, m_iEventToLoopBackTo, int);
}

void Asura_Cutscene_Event_LoopBack::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

    xStream << m_fNumSecondsToLoopBack;
    xStream << m_iEventToLoopBackTo;
}

void Asura_Cutscene_Event_LoopBack::LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk, Asura_Handle fileHandle)
{
    Asura_Cutscene_Event::LoadFromChunk(chunk,fileHandle);
    ASURA_CHUNK_FILEEXTRACT(fileHandle, m_fNumSecondsToLoopBack, float);
	if(chunk->Version>=6)
	{
    	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iEventToLoopBackTo, int);
	}
}
////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_Cutscene_Event_Attachment **
**---------------------------------*/
int Asura_Cutscene_Event_Attachment::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() + sizeof(u_int);
	return iChunkSize;
}

void Asura_Cutscene_Event_Attachment::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uAttachmentID,u_int);
}

void Asura_Cutscene_Event_Attachment::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_uAttachmentID;
}

void Asura_Cutscene_Event_Attachment::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uAttachmentID,u_int);
}


////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------------**
** Asura_Cutscene_Event_AttachmentControl **
**----------------------------------------*/
int Asura_Cutscene_Event_AttachmentControl::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize();
	iChunkSize += sizeof(u_int); // num attachments
	//filter out attachments which are going to be unchanged
	u_int uNumToSave = 0;
	for(u_int i=0;i<m_uNumAttachments;i++)
	{
		if(m_pxAttachments[i].m_uType)
		{
			uNumToSave++;
		}
	}
	iChunkSize += sizeof(u_int)*2*uNumToSave;
	return iChunkSize;
}

void Asura_Cutscene_Event_AttachmentControl::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	//filter out attachments which are going to be unchanged
	u_int uNumToSave = 0;
	for(u_int i=0;i<m_uNumAttachments;i++)
	{
		if(m_pxAttachments[i].m_uType)
		{
			uNumToSave++;
		}
	}

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,uNumToSave,u_int);
	for(u_int i=0;i<m_uNumAttachments;i++)
	{
		if(m_pxAttachments[i].m_uType)
		{
			ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_pxAttachments[i].m_uID,u_int);
			ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_pxAttachments[i].m_uType,u_int);
		}
	}
}

void Asura_Cutscene_Event_AttachmentControl::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	//filter out attachments which are going to be unchanged
	u_int uNumToSave = 0;
	for(u_int i=0;i<m_uNumAttachments;i++)
	{
		if(m_pxAttachments[i].m_uType)
		{
			uNumToSave++;
		}
	}
	xStream << uNumToSave;

	for(u_int i=0;i<m_uNumAttachments;i++)
	{
		if(m_pxAttachments[i].m_uType)
		{
			xStream << m_pxAttachments[i].m_uID;
			xStream << m_pxAttachments[i].m_uType;
		}
	}

}

void Asura_Cutscene_Event_AttachmentControl::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uNumAttachments,u_int);

	m_pxAttachments = m_uNumAttachments ? new Asura_Cutscene_Event_AttachmentControl_Attachment[m_uNumAttachments] : 0;
	for(u_int i=0;i<m_uNumAttachments;i++)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_pxAttachments[i].m_uID,u_int);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_pxAttachments[i].m_uType,u_int);
	}
}


////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Cutscene_Event_Beam **
**---------------------------*/
int Asura_Cutscene_Event_Beam::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() 
				   + 2*sizeof(Asura_Vector_3)//start/end poaition
				   + sizeof(int) //beam type
				   + 5*sizeof(float)   //2 radii  , 2 speeds , duration
				   + 4*4*sizeof(float); //4 colours
	return iChunkSize;
}

void Asura_Cutscene_Event_Beam::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	#ifdef ASURA_TOOLS
	//RAR 08/03/2006 For recentreing during level exports :
	const Asura_Vector_3& xPositionOffset = Tools_Level_Export::GetExportCentre();
	#else
	const Asura_Vector_3& xPositionOffset = Asura_Maths::ZeroVector3;
	#endif

	ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xInitialPosition-xPositionOffset,Asura_Vector_3, float, 3);
	ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xEndPosition-xPositionOffset,Asura_Vector_3, float, 3);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.uTileHashID,u_int);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.fInitialRadius,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.fEndRadius,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.fTailSpeed,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.fHeadSpeed,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_fDuration,float);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[0],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[1],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[2],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[3],float);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[0],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[1],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[2],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[3],float);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[0],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[1],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[2],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[3],float);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[0],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[1],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[2],float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[3],float);
}

void Asura_Cutscene_Event_Beam::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_xParam.m_xInitialPosition;
	xStream << m_xParam.m_xEndPosition;

	xStream << m_xParam.m_xTemplate.uTileHashID;
	xStream << m_xParam.m_xTemplate.fInitialRadius;
	xStream << m_xParam.m_xTemplate.fEndRadius;
	xStream << m_xParam.m_xTemplate.fTailSpeed;
	xStream << m_xParam.m_xTemplate.fHeadSpeed;
	xStream << m_xParam.m_fDuration;

	xStream << m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[0];
	xStream << m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[1];
	xStream << m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[2];
	xStream << m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[3];

	xStream << m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[0];
	xStream << m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[1];
	xStream << m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[2];
	xStream << m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[3];

	xStream << m_xParam.m_xTemplate.xEndTailColour.m_afChannels[0];
	xStream << m_xParam.m_xTemplate.xEndTailColour.m_afChannels[1];
	xStream << m_xParam.m_xTemplate.xEndTailColour.m_afChannels[2];
	xStream << m_xParam.m_xTemplate.xEndTailColour.m_afChannels[3];

	xStream << m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[0];
	xStream << m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[1];
	xStream << m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[2];
	xStream << m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[3];
}

void Asura_Cutscene_Event_Beam::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xInitialPosition,Asura_Vector_3);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xEndPosition,Asura_Vector_3);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.uTileHashID,u_int);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.fInitialRadius,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.fEndRadius,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.fTailSpeed,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.fHeadSpeed,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_fDuration,float);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[0],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[1],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[2],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialTailColour.m_afChannels[3],float);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[0],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[1],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[2],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xInitialHeadColour.m_afChannels[3],float);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[0],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[1],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[2],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndTailColour.m_afChannels[3],float);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[0],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[1],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[2],float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_xParam.m_xTemplate.xEndHeadColour.m_afChannels[3],float);

}

////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Cutscene_Event_Fade **
**---------------------------*/
int Asura_Cutscene_Event_Fade::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() 
				   + sizeof(u_int) //flags
				   + 4*sizeof(float);   //red,green,blue,duration
	return iChunkSize;
}

void Asura_Cutscene_Event_Fade::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fRed,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fGreen,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fBlue,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fDuration,float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uFlags,u_int);
}

void Asura_Cutscene_Event_Fade::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_fRed;
	xStream << m_fGreen;
	xStream << m_fBlue;
	xStream << m_fDuration;
	xStream << m_uFlags;
}

void Asura_Cutscene_Event_Fade::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);


	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fRed,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fGreen,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fBlue,float);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fDuration,float);

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uFlags,u_int);

}

////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Cutscene_Event_Subtitle **
**-------------------------------*/
int Asura_Cutscene_Event_Subtitle::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() 
				   + sizeof(Asura_Hash_ID) //text hash
				   + sizeof(float);   //duration
	return iChunkSize;
}

void Asura_Cutscene_Event_Subtitle::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uTextHash,Asura_Hash_ID);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_fDuration,float);
}

void Asura_Cutscene_Event_Subtitle::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_uTextHash;
	xStream << m_fDuration;
}

void Asura_Cutscene_Event_Subtitle::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);


	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uTextHash,Asura_Hash_ID);
	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_fDuration,float);

}

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Asura_Cutscene_Event_Sound **
**----------------------------*/
int Asura_Cutscene_Event_Sound::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() 
				   + m_xSoundName.GetSizeInInts()*4;
	return iChunkSize;
}

void Asura_Cutscene_Event_Sound::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILL_STRING(pcChunkBuffer,m_xSoundName);
}

void Asura_Cutscene_Event_Sound::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_xSoundName;
}

void Asura_Cutscene_Event_Sound::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);


	ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle,m_xSoundName);

}

////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------------**
** Asura_Cutscene_Event_ChangeCameraTarget **
**-----------------------------------------*/
int Asura_Cutscene_Event_ChangeCameraTarget::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize();
	ASURA_CHUNK_SIZE_STRING(iChunkSize,m_xCameraTargetName);
	return iChunkSize;
}

void Asura_Cutscene_Event_ChangeCameraTarget::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILL_STRING(pcChunkBuffer,m_xCameraTargetName);
}

void Asura_Cutscene_Event_ChangeCameraTarget::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_xCameraTargetName;
}

void Asura_Cutscene_Event_ChangeCameraTarget::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);


	ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle,m_xCameraTargetName);

}

////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------------**
** Asura_Cutscene_Event_ChangeCameraShake **
**-----------------------------------------*/

int Asura_Cutscene_Event_ChangeCameraShake::GetChunkSize()
{
	Asura_Chunk_Dummy_Stream xStream;
	WriteToChunkStream(xStream);
	return sizeof(Asura_Chunk_CutsceneEvent) + xStream.GetPosition();
}

void Asura_Cutscene_Event_ChangeCameraShake::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	WriteToChunkStream(xStream);
}

void Asura_Cutscene_Event_ChangeCameraShake::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	xStream << CHUNK_VERSION;

	Asura_Cutscene_Event::WriteToChunkStream(xStream);

	m_xShake.WriteToChunkStream(xStream);
	xStream << m_fDuration;
}

void Asura_Cutscene_Event_ChangeCameraShake::ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert(uVersion <= CHUNK_VERSION, "Asura_Cutscene_Event_ChangeCameraShake::ReadFromChunkStream - Unrecognised chunk version" );

	Asura_Cutscene_Event::ReadFromChunkStream(pxChunk, xStream);

	m_xShake.ReadFromChunkStream(xStream);
	xStream >> m_fDuration;
}

void Asura_Cutscene_Event_ChangeCameraShake::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle)
{
	Asura_Chunk_File_Stream xStream(iFileHandle);
	ReadFromChunkStream(pxChunk, xStream);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Cutscene_Event_SetAlpha **
**-------------------------------*/

int Asura_Cutscene_Event_SetAlpha::GetChunkSize()
{
	Asura_Chunk_Dummy_Stream xStream;
	WriteToChunkStream(xStream);
	return sizeof(Asura_Chunk_CutsceneEvent) + xStream.GetPosition();
}

void Asura_Cutscene_Event_SetAlpha::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	WriteToChunkStream(xStream);
}

void Asura_Cutscene_Event_SetAlpha::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	xStream << CHUNK_VERSION;

	PARENT::WriteToChunkStream(xStream);

	xStream << m_fDuration;

	bool bIsExport = false;
#ifdef ASURA_TOOLS
	bIsExport = Tools_Level_Export::IsExport() && !Tools_Level_Export::IsDoingFakeLevelExport();
#endif

	m_xAlpha.WriteToChunkStream(xStream,bIsExport);
}

void Asura_Cutscene_Event_SetAlpha::ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert(uVersion <= CHUNK_VERSION, "Asura_Cutscene_Event_SetAlpha::ReadFromChunkStream - Unrecognised chunk version" );

	PARENT::ReadFromChunkStream(pxChunk, xStream);

	xStream >> m_fDuration;
	m_xAlpha.ReadFromChunkStream(xStream);
}

void Asura_Cutscene_Event_SetAlpha::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle)
{
	Asura_Chunk_File_Stream xStream(iFileHandle);
	ReadFromChunkStream(pxChunk, xStream);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------------------------**
** Asura_Cutscene_Event_ChangePlayerCameraControl **
**------------------------------------------------*/

int Asura_Cutscene_Event_ChangePlayerCameraControl::GetChunkSize()
{
	Asura_Chunk_Dummy_Stream xStream;
	WriteToChunkStream(xStream);
	return sizeof(Asura_Chunk_CutsceneEvent) + xStream.GetPosition();
}

void Asura_Cutscene_Event_ChangePlayerCameraControl::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	WriteToChunkStream(xStream);
}

void Asura_Cutscene_Event_ChangePlayerCameraControl::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	xStream << CHUNK_VERSION;

	PARENT::WriteToChunkStream(xStream);

	xStream << m_fPitchLimit;
	xStream << m_fYawLimit;
	xStream << m_fDuration;
}

void Asura_Cutscene_Event_ChangePlayerCameraControl::ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert(uVersion <= CHUNK_VERSION, "Asura_Cutscene_Event_ChangePlayerCameraControl::ReadFromChunkStream - Unrecognised chunk version" );

	PARENT::ReadFromChunkStream(pxChunk, xStream);

	xStream >> m_fPitchLimit;
	xStream >> m_fYawLimit;
	xStream >> m_fDuration;
}

void Asura_Cutscene_Event_ChangePlayerCameraControl::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle)
{
	Asura_Chunk_File_Stream xStream(iFileHandle);
	ReadFromChunkStream(pxChunk, xStream);
}

////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_Cutscene_Event_FacialAnimation **
**--------------------------------------*/
#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
int Asura_Cutscene_Event_FacialAnimation::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() 
				   + sizeof(Asura_Hash_ID) //anim hash
				   + sizeof(u_int);//flags
	return iChunkSize;
}

void Asura_Cutscene_Event_FacialAnimation::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uAnimID,Asura_Hash_ID);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uAnimationFlags, u_int);
}

void Asura_Cutscene_Event_FacialAnimation::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_uAnimID;
	xStream << m_uAnimationFlags;
}

void Asura_Cutscene_Event_FacialAnimation::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);


	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uAnimID,Asura_Hash_ID);
	if(pxChunk->Version >= 14)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uAnimationFlags,u_int);
	}
}
#endif //ASURA_USE_FACIAL_ANIMATION_SYSTEM

////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_Cutscene_Event_MixerPreset **
**----------------------------------*/
int Asura_Cutscene_Event_MixerPreset::GetChunkSize()
{
	int iChunkSize = Asura_Cutscene_Event::GetChunkSize() 
				   + sizeof(Asura_Hash_ID); //mixer hash
	return iChunkSize;
}

void Asura_Cutscene_Event_MixerPreset::SaveToChunk(char*& pcChunkBuffer)
{
	Asura_Cutscene_Event::SaveToChunk(pcChunkBuffer);

	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,m_uMixerPreset,Asura_Hash_ID);
}

void Asura_Cutscene_Event_MixerPreset::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_Cutscene_Event::WriteToChunkStream	(xStream);

	xStream << m_uMixerPreset;
}

void Asura_Cutscene_Event_MixerPreset::LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle)
{
	Asura_Cutscene_Event::LoadFromChunk(pxChunk,iFileHandle);


	ASURA_CHUNK_FILEEXTRACT(iFileHandle,m_uMixerPreset,Asura_Hash_ID);
}

