//
// Asura_Chunk_Mesh.cpp
//
// Created by Kevin Lea - 12/06/2001 11:19:27
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Mesh.h"

#include <float.h>


#include "Core/Asura_TextureManagement.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "ChunkSystem/Asura_Chunk_FileInfo.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Mesh/Asura_Mesh_CollMeshGenData.h"
#ifdef ASURA_MAIN
#include "Shapes/Asura_Shape_System.h"
#endif

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#include "Materials/Tools_MaterialResponse_System.h"
#endif

#include "PC/FileSystem/Tools_FileSystem.h"

//This bool gets set when a duplicate mesh is found , to prevent the remaining chunks for that mesh from being processed.
//It relies on the fact that every mesh has a vertex chunk , which will appear before all the other chunks for it
//This is only checked for when loading from a file handle , to prevent it from intefereing with some tools stuff
//that processes the mesh chunks 'by hand'
//RAR 25/01/2007
static bool g_bSkipChunksForCurrentMesh = false;

bool Asura_Chunk_Mesh_Vertices::Process(Asura_Handle fileHandle) const
{
	const Asura_Char* meshname = Asura_FileAccess::ReadString(fileHandle);

	Asura_Mesh *meshPtr = Asura_Mesh_System::GetMesh(meshname,true);

	meshPtr->m_iSourceFileType = Asura_Chunk_FileInfo::GetLastLoadedFileType();
	
	if (meshPtr->VertexDataValid)
	{
		
		
		#if ASURA_LOG_FILE_CONTENTS
		Asura_SimpleLogfileOutput((ASURA_FILE_CONTENTS_LOGFILE,"Mesh : %s - Already Exists!\n",meshname));
		#endif
		
		
		// Kevin Lea - 13/09/2001 15:52:55 - mesh already exists
		// this could be a naming problem (duplicates)
		// otherwise if this happens regularly we'll have to read off
		// the remainder of the chunk to keep things in sync
		Asura_Assertf(0, ("Mesh: '%s'  - Already exists", meshname));
        Asura_FileAccess::SkipData(fileHandle,NumberOfVertices*sizeof(Asura_Vector_3));
		//Skip any other mesh chunks until we next come across a vertex chunk		
		g_bSkipChunksForCurrentMesh = true;
		return true;
	}
	g_bSkipChunksForCurrentMesh = false;

	#if ASURA_LOG_FILE_CONTENTS
	Asura_SimpleLogfileOutput((ASURA_FILE_CONTENTS_LOGFILE,"Mesh : %s\n",meshname));
	#endif

	//Get a checksum for the mesh based on the file modified time , and the meshname. RAR 20/03/2006
	const char* szFilename = Asura_FileAccess::GetFilename(fileHandle);
	FILETIME xModifiedTime;
	if(Tools_FileSystem::GetModifiedTime(szFilename ,&xModifiedTime))
	{
		meshPtr->m_uChecksum = xModifiedTime.dwLowDateTime;
		meshPtr->m_uChecksum += Asura_GetHashID(meshPtr->Name.GetString());
	}
	else
	{
		meshPtr->m_uChecksum = 0;
	}

	//store the name of the file that this mesh came from
	meshPtr->m_xStrippedFilename = Tools_FileSystem::StripPath(szFilename);	
	

	Asura_Chunk_File_Stream xStream(fileHandle);
	return Process(xStream,meshPtr,true);
}

bool Asura_Chunk_Mesh_Vertices::Process(Asura_Chunk_Stream& xStream, Asura_Mesh* pxMesh, bool bAlreadyReadName) const
{

	pxMesh->NumberOfVertices = NumberOfVertices;
	pxMesh->VertexDataPtr = new Asura_Mesh_Vertex[NumberOfVertices];
	
	xStream.Read(NumberOfVertices*sizeof(Asura_Vector_3),pxMesh->VertexDataPtr);

	pxMesh->VertexDataValid = true;

	return true;
}


#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_Vertices::Create(const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_Vertices::Create(const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	u_int uChunkVersion = 2;
	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_VERTICES, uChunkVersion, uChunkFlags);

	xStream << pxMesh->NumberOfVertices;
	xStream << pxMesh->Name;

	Asura_Mesh_Vertex* pxSrc = pxMesh->VertexDataPtr;
	for(int i=0; i<pxMesh->NumberOfVertices; i++,pxSrc++)
	{
		xStream << pxSrc->Position;
	}
	xStream.WriteChunk_End();
}

#endif

bool Asura_Chunk_Mesh_Polygons::Process(Asura_Handle fileHandle) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	const Asura_Char* meshname = Asura_FileAccess::ReadString(fileHandle);
	
	Asura_Mesh *meshPtr = Asura_Mesh_System::GetMesh(meshname,true);

	meshPtr->m_iSourceFileType = Asura_Chunk_FileInfo::GetLastLoadedFileType();
	
	if (meshPtr->PolygonDataValid)
	{
		// Kevin Lea - 13/09/2001 15:52:55 - mesh already exists
		// this could be a naming problem (duplicates)
		// otherwise if this happens regularly we'll have to read off
		// the remainder of the chunk to keep things in sync
		Asura_Assertf(0, ("Mesh: '%s'  - Already exists", meshname));
        Asura_FileAccess::SkipData(fileHandle,sizeof(Asura_Mesh_Polygon)*NumberOfPolygons);
		return true;
	}

	Asura_Chunk_File_Stream xStream(fileHandle);
	return Process(xStream,meshPtr,true);
}

bool Asura_Chunk_Mesh_Polygons::Process(Asura_Chunk_Stream& xStream, Asura_Mesh* pxMesh, bool bAlreadyReadName) const
{
	if(!bAlreadyReadName) xStream.SkipString();


	pxMesh->NumberOfPolygons = NumberOfPolygons;
	pxMesh->NumberOfTriangles = NumberOfTriangles;
	pxMesh->PolygonDataPtr = new Asura_Mesh_Polygon[NumberOfPolygons];

	xStream.Read(sizeof(Asura_Mesh_Polygon)*NumberOfPolygons,pxMesh->PolygonDataPtr);

	pxMesh->PolygonDataValid = true;

	// JGP - Painted colours constructed here in case there is no chunk for
	// painted colours...

	pxMesh->m_pxPaintedColourDataPtr = new Asura_Mesh_PaintedColours[ NumberOfPolygons ];
	pxMesh->m_bPaintedColourDataValid = true;
	for ( int i = 0; i < NumberOfPolygons; i++ )
	{
		for ( int j = 0; j < pxMesh->PolygonDataPtr[ i ].NumberOfVertices(); j++ )
		{
			pxMesh->m_pxPaintedColourDataPtr[ i ].Colour[ j ] = 0x808080;
		}

	}
	return true;

}

#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_Polygons::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh, xStream);
}

void Asura_Chunk_Mesh_Polygons::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{

	u_int uChunkVersion = 2;
	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_POLYGONS, uChunkVersion, uChunkFlags);
	
	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->NumberOfTriangles;
	xStream << pxMesh->Name;

	const Asura_Mesh_Polygon* pxSrc = pxMesh->PolygonDataPtr;

	for(int i=0; i<pxMesh->NumberOfPolygons; i++, pxSrc++)
	{
		xStream << pxSrc->VertexID[0];					
		xStream << pxSrc->VertexID[1];					
		xStream << pxSrc->VertexID[2];					
		xStream << pxSrc->VertexID[3];					
	}

	xStream.WriteChunk_End();
}

#endif

bool Asura_Chunk_Mesh_PaintedColours::Process( Asura_Handle hFile ) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	const Asura_Char* szMeshName = Asura_FileAccess::ReadString( hFile );

	Asura_Mesh *pxMesh = Asura_Mesh_System::GetMesh( szMeshName, true );
	Asura_Assert( NumberOfPolygons == pxMesh->NumberOfPolygons, "Polygon count mismatch" );

	// Read into painted colours
	Asura_FileAccess::ReadData( hFile, pxMesh->m_pxPaintedColourDataPtr, NumberOfPolygons * sizeof( Asura_Mesh_PaintedColours ) );

	return true;
}

#ifdef ASURA_TOOLS

void Asura_Chunk_Mesh_PaintedColours::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_PaintedColours::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	Asura_Assert(pxMesh, "No mesh");
	if ( !pxMesh ) 
	{
		return;
	}

	Asura_Assert( pxMesh->m_bPaintedColourDataValid && pxMesh->m_pxPaintedColourDataPtr, "Painted data is not valid" );
	if ( ( !pxMesh->m_bPaintedColourDataValid ) || ( !pxMesh->m_pxPaintedColourDataPtr ) )
	{
		return;
	}

	u_int uChunkVersion = 0;
	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_PAINTEDCOLOURS, uChunkVersion, uChunkFlags);
	
	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->Name;

	const Asura_Mesh_PaintedColours* pxSrc = pxMesh->m_pxPaintedColourDataPtr;

	for(int i=0; i<pxMesh->NumberOfPolygons; i++, pxSrc++)
	{
		xStream << pxSrc->Colour[0];
		xStream << pxSrc->Colour[1];
		xStream << pxSrc->Colour[2];
		xStream << pxSrc->Colour[3];
	}

	xStream.WriteChunk_End();
}


#endif//ASURA_TOOLS

#ifdef ASURA_USE_SAVED_MESH_COLOURS

bool Asura_Chunk_Mesh_Colours::Process( Asura_Handle hFile ) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	const Asura_Char* szMeshName = Asura_FileAccess::ReadString(hFile);
	// Get mesh by name
	Asura_Mesh *pxMesh = Asura_Mesh_System::GetMesh(szMeshName,true);

	// Check for existing colour data
	if (pxMesh->PrelightingDataValid)
	{
		#if ASURA_LOG_FILE_CONTENTS
		Asura_SimpleLogfileOutput((ASURA_FILE_CONTENTS_LOGFILE,"Mesh : %s - Already has colour information!\n",szMeshName));
		#endif
		Asura_Assertf(0, ("Mesh: '%s'  - Already has colour information", szMeshName));
        Asura_FileAccess::SkipData(hFile,NumberOfPolygons * sizeof(Asura_Mesh_Prelighting));
		return true;
	}
	Asura_Assert(NumberOfPolygons == pxMesh->NumberOfPolygons, "Polygon count mismatch");
	// Allocate prelighting pointer
	pxMesh->PrelightingDataPtr = new Asura_Mesh_Prelighting[NumberOfPolygons];
	// Read into colours
	Asura_FileAccess::ReadData(hFile, pxMesh->PrelightingDataPtr, NumberOfPolygons * sizeof(Asura_Mesh_Prelighting));

	pxMesh->PrelightingDataValid = true;

	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_Colours::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_Colours::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	Asura_Assert(pxMesh, "No mesh");
	if(!pxMesh) return;
	Asura_Assert(pxMesh->PrelightingDataValid && pxMesh->PrelightingDataPtr, "Prelighting data is not valid");
	if(!pxMesh->PrelightingDataValid || !pxMesh->PrelightingDataPtr)	return;

	u_int uChunkVersion = 0;
	u_int uChunkFlags = 0;	
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_COLOURS, uChunkVersion, uChunkFlags);

	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->Name;

	const Asura_Mesh_Prelighting* pxSrc = pxMesh->PrelightingDataPtr;

	for(int i=0; i<pxMesh->NumberOfPolygons; i++, pxSrc++)
	{
		xStream << pxSrc->Colour[0];		
		xStream << pxSrc->Colour[1];		
		xStream << pxSrc->Colour[2];		
		xStream << pxSrc->Colour[3];		
	}

	xStream.WriteChunk_End();
}

#endif//ASURA_TOOLS

#endif//ASURA_USE_SAVED_MESH_COLOURS


bool Asura_Chunk_Mesh_Normals::Process( Asura_Handle hFile ) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

    return Platform_Process( hFile );
}

#ifdef ASURA_TOOLS

void Asura_Chunk_Mesh_Normals::Create(const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_Normals::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	u_int uChunkVersion = 0;
	u_int uChunkFlags = 0;
	if(pxMesh->m_bSmoothedAcrossModules) uChunkFlags |= ChunkFlag_SmoothedAcrossModules;

	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_NORMALS, uChunkVersion, uChunkFlags);

	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->Name;

	const Asura_Mesh_Normals* pxSrc = pxMesh->NormalsDataPtr;
	for( int i = 0; i < pxMesh->NumberOfPolygons; ++i )
	{
		xStream << pxSrc->FaceNormal;

        Asura_Assert( pxSrc->VertexNormal, "" );
        for( u_int u = 0; u < 4; ++u )
        {
            xStream << pxSrc->VertexNormal[u];
        }
        ++pxSrc;
	}
	xStream.WriteChunk_End();

}
#endif


bool Asura_Chunk_Mesh_TextureInfo::Process(Asura_Chunk_Stream& xStream) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	Asura_Assertf(Version<=ASURA_CHUNK_MESH_TEXTUREINFO_VERSION,("Unrecognised chunk version %d",Version));
	//chunk code would need some versioning stuff done if these defines change. RAR 13/06/2007
	ASURA_STATIC_ASSERT(ASURA_MATERIAL_MAXTEXTURES==2);
	ASURA_STATIC_ASSERT(TOOLS_NUM_MATERIALRESPONSE_PER_POLY==2);
	
	char szMeshName[512];
	xStream.ReadString(szMeshName,sizeof(szMeshName));

	Asura_Mesh* pxMesh = Asura_Mesh_System::GetMesh( szMeshName, true );

	if( pxMesh->TextureDataValid )
	{
		// Kevin Lea - 13/09/2001 15:52:55 - mesh already exists
		// this could be a naming problem (duplicates)
		// otherwise if this happens regularly we'll have to read off
		// the remainder of the chunk to keep things in sync
		return false;
	}

	pxMesh->TextureDataPtr = new Asura_Mesh_TextureInfo[NumberOfPolygons];

	#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM) && defined(ASURA_TOOLS)
	// KJFL - 15 April 2008 - we need to load materials used by the mesh. Rather than querying the material per poly,
	// let's make a separate tree of the ones we come across and make sure they're loaded at the end. 
	// (This should be faster than checking against all the materials in the project per poly).
	Asura_HashTree xUsedMaterials;
	#endif

	u_int uNumSavedMaterialResponse = 0;
	if(Version==4) uNumSavedMaterialResponse = 1;
	else if(Version>=5) uNumSavedMaterialResponse = 2;
	
	Asura_Mesh_TextureInfo* pxDestPtr = pxMesh->TextureDataPtr;
	for( int i = 0; i < NumberOfPolygons; ++pxDestPtr, ++i )
	{
        // Read material ID.
		xStream >> pxDestPtr->m_iMaterialID;

        // Read UVs.
		if(Version<3)
		{
			xStream.Read(4 * sizeof(float), pxDestPtr->U[0]);
			xStream.Read(4 * sizeof(float), pxDestPtr->V[0]);
			//Copy the first textures UVs to the other channels
			for(int iChan=1;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				memcpy(pxDestPtr->U[iChan],pxDestPtr->U[0],4*sizeof(float));
				memcpy(pxDestPtr->V[iChan],pxDestPtr->V[0],4*sizeof(float));
			}
			//initialise vertex alpha
			for(int j=0;j<4;j++)
			{
				pxDestPtr->m_afAlpha[j] = 1.0f;
			}

		}
		else
		{
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				xStream.Read(4 * sizeof(float), pxDestPtr->U[iChan]);
				xStream.Read(4 * sizeof(float), pxDestPtr->V[iChan]);
			}
			#ifdef ASURA_DISABLE_VERTEX_ALPHA
			//skip alpha components , and initialise to defaults instead
			xStream.Skip(4 * sizeof(float));

			for(int j=0;j<4;j++)
			{
				pxDestPtr->m_afAlpha[j] = 1.0f;
			}

			#else
			xStream.Read(4 * sizeof(float), pxDestPtr->m_afAlpha);
			#endif
		}

		// Read material response hash ID
		for(u_int i=0;i< uNumSavedMaterialResponse; i++)
		{
			xStream >> pxDestPtr->m_auMaterialResponseHashID[i];
		}

		#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM) && defined(ASURA_MAIN)
		if((uNumSavedMaterialResponse==1 || (Flags & ChunkFlag_MaterialResponse_NotYetConvertedToDualMaterial)) && pxDestPtr->m_auMaterialResponseHashID[0])
		{
			//find the second material to use on this polygon
			const Tools_MaterialResponse* pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID(pxDestPtr->m_auMaterialResponseHashID[0]);
			if(pxMaterial)
			{
				pxDestPtr->m_auMaterialResponseHashID[1] = pxMaterial->GetSecondMaterialID();
			}
		}
		#endif

		
		#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM) && defined(ASURA_TOOLS)
		// KJFL - 15 April 2008 - store the materials we've used in a tree for later
		for ( u_int u=0; u<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; ++u )
		{
			if ( pxDestPtr->m_auMaterialResponseHashID[u] != ASURA_HASH_ID_UNSET )
			{
				int iDummy;
				xUsedMaterials.FindNodeIndexByHash( pxDestPtr->m_auMaterialResponseHashID[u], true, iDummy );
			}
		}
		#endif


		//Seem to be getting some nans in the texturing data from somewhere (gun in particular) , so
		//going to check for them here. RAR 13/03/2006
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			for(int i=0;i<4;i++)
			{
				if(_isnan(pxDestPtr->U[iChan][i]))
				{
					pxDestPtr->U[iChan][i] = iChan ? pxDestPtr->U[0][i] : 0.0f;
				}
				if(_isnan(pxDestPtr->V[iChan][i]))
				{
					pxDestPtr->V[iChan][i] = iChan ? pxDestPtr->V[0][i] : 0.0f;
				}
			}
		}
		for(int i=0;i<4;i++)
		{
			if(_isnan(pxDestPtr->m_afAlpha[i]))
			{
				pxDestPtr->m_afAlpha[i] = 1.0f;
			}
		}

        // Convert material ID if necessary.
		if( -1 != pxDestPtr->m_iMaterialID )
		{
			pxDestPtr->m_iMaterialID = Asura_Material_System::ConvertIndex( pxDestPtr->m_iMaterialID );
		}
	}

	#if !defined(ASURA_MAIN)
	//Avoid doing the material conversions in the max plugins , since it is less certain that things will be setup correctly
	//for loading them.
	//Instead just set a flag so that the mesh will get saved back out as the earlier version during the export process.
	if(Version <= 4 || (Flags & ChunkFlag_MaterialResponse_NotYetConvertedToDualMaterial))
	{
		pxMesh->m_bMaterialResponse_NotYetConvertedToDualMaterial = true;	
	}
	#endif

	#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM) && defined(ASURA_TOOLS)
	// KJFL - 15 April 2008 - make a list of all the materials and check that they're loaded
	int iNumMaterialResponse = xUsedMaterials.GetNumberOfNodes();

	Asura_Hash_ID* puMaterialIDs = new Asura_Hash_ID[iNumMaterialResponse];
	xUsedMaterials.MakeHashList(puMaterialIDs);

	for(int i=0; i<iNumMaterialResponse; i++)
	{
		// merely accessing the material forces a load on demand
		Tools_MaterialResponse_System::GetMaterialByHashID(puMaterialIDs[i]);
	}

	delete [] puMaterialIDs;
	#endif
	


	pxMesh->TextureDataValid = true;
	return true;
}

bool Asura_Chunk_Mesh_TextureInfo::Process( Asura_Handle fileHandle ) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	Asura_Chunk_File_Stream	xStream(fileHandle);
	return Process(xStream);
}

#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_TextureInfo::Create(const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream,const int* piMaterialIDToIndexConversion)
{
	u_int uChunkFlags = 0;

	#ifndef ASURA_MAIN
	if(pxMesh->m_bMaterialResponse_NotYetConvertedToDualMaterial)
	{
		uChunkFlags |= ChunkFlag_MaterialResponse_NotYetConvertedToDualMaterial;
	}
	#endif

	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_TEXTUREINFO, ASURA_CHUNK_MESH_TEXTUREINFO_VERSION, uChunkFlags);

	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->Name;

	const Asura_Mesh_TextureInfo* pxSrc = pxMesh->TextureDataPtr;

	for( int i = 0; i < pxMesh->NumberOfPolygons; ++i )
	{
        xStream << Asura_Material_System::ConvertIDToIndex(pxSrc->m_iMaterialID, piMaterialIDToIndexConversion);

        for(int j=0;j<ASURA_MATERIAL_MAXTEXTURES;j++)
		{
        	Asura_Assert( pxSrc->U[j] && pxSrc->V[j], "" );
        	
        	for( u_int u = 0; u < 4; ++u )
        	{
			    xStream << pxSrc->U[j][u];
        	}
        	for( u_int u = 0; u < 4; ++u )
        	{
			    xStream << pxSrc->V[j][u];
        	}
		}

		for(int j=0;j<4;j++)
		{
		    xStream << pxSrc->m_afAlpha[j];
		}

		for(int j=0;j<TOOLS_NUM_MATERIALRESPONSE_PER_POLY;j++)
		{
			xStream << pxSrc->m_auMaterialResponseHashID[j];
		}

        ++pxSrc;
	}

	xStream.WriteChunk_End();
}

void Asura_Chunk_Mesh_TextureInfo::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock,const int* piMaterialIDToIndexConversion)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create( pxMesh, xStream, piMaterialIDToIndexConversion);
}
#endif

bool Asura_Chunk_Mesh_SmoothingGroup::Process(Asura_Handle fileHandle) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	const Asura_Char* meshname = Asura_FileAccess::ReadString(fileHandle);

	Asura_Mesh *meshPtr = Asura_Mesh_System::GetMesh(meshname,true);

	if (meshPtr->SmoothingGroupDataValid)
	{
		// Kevin Lea - 13/09/2001 15:52:55 - mesh already exists
		// this could be a naming problem (duplicates)
		// otherwise if this happens regularly we'll have to read off
		// the remainder of the chunk to keep things in sync
		Asura_Assertf(0, ("Mesh: '%s'  - Already exists", meshname));
        Asura_FileAccess::SkipData(fileHandle,NumberOfPolygons*sizeof(int));
		return true;
	}

	meshPtr->SmoothingGroupDataPtr = new int[NumberOfPolygons];
	
	Asura_FileAccess::ReadData(fileHandle,meshPtr->SmoothingGroupDataPtr,NumberOfPolygons*sizeof(int));

	meshPtr->SmoothingGroupDataValid = true;
	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_SmoothingGroup::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_SmoothingGroup::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	u_int uChunkVersion = 1;
	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_SMOOTHINGGROUP, uChunkVersion, uChunkFlags);
	

	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->Name;

	const int* piSrc = pxMesh->SmoothingGroupDataPtr;

	for(int i=0; i<pxMesh->NumberOfPolygons; i++, piSrc++)
	{
		xStream << (*piSrc);
	}
	xStream.WriteChunk_End();
}

#endif

//////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Chunk_Mesh_PolygonFlags **
**-------------------------------*/



bool Asura_Chunk_Mesh_PolygonFlags::Process(Asura_Handle fileHandle) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	Asura_Assert(Version<=ASURA_CHUNK_MESH_POLYGONFLAGS_VERSION,"Unknown chunk version for polygon flags");

	const Asura_Char* meshname = Asura_FileAccess::ReadString(fileHandle);

	Asura_Mesh *meshPtr = Asura_Mesh_System::GetMesh(meshname,true);

	if (meshPtr->PolygonFlagDataValid)
	{
		// Kevin Lea - 13/09/2001 15:52:55 - mesh already exists
		// this could be a naming problem (duplicates)
		// otherwise if this happens regularly we'll have to read off
		// the remainder of the chunk to keep things in sync
		Asura_Assertf(0, ("Mesh: '%s'  - Already exists", meshname));
        Asura_FileAccess::SkipData(fileHandle,NumberOfPolygons*sizeof(Asura_Mesh_PolygonFlags));
		return true;
	}

	Asura_Chunk_File_Stream xStream(fileHandle);
	return Process(xStream,meshPtr,true);
}

bool Asura_Chunk_Mesh_PolygonFlags::Process(Asura_Chunk_Stream& xStream, Asura_Mesh* pxMesh, bool bAlreadyReadName) const
{
	Asura_Assert(Version<=ASURA_CHUNK_MESH_POLYGONFLAGS_VERSION,"Unknown chunk version for polygon flags");
	Asura_Assert(NumberOfPolygons == pxMesh->NumberOfPolygons,"Incorrect number of polygons in polygon flags chunk");

	if(!bAlreadyReadName) xStream.SkipString();

	pxMesh->PolygonFlagDataPtr = new Asura_Mesh_PolygonFlags[NumberOfPolygons];
	
	xStream.Read(NumberOfPolygons*sizeof(Asura_Mesh_PolygonFlags),pxMesh->PolygonFlagDataPtr);
	pxMesh->PolygonFlagDataValid = true;

	if(Version<2)
	{
		//polygon flag values need to be converted for earlier versions.
		pxMesh->ConvertPolygonFlags();
	}
    else if( Version == 2 )
    {
        // (MP - 19.07.2004) Nasty fix for flag mix up in Rogue. Fortunately
        // no other project had its own flags yet...
        for( int i = 0;i < NumberOfPolygons; ++i )
        {
            pxMesh->PolygonFlagDataPtr[i].Flags |= (pxMesh->PolygonFlagDataPtr[i].Flags & 0x000000F0) << 4;
        }
    }
	
	if(Version<5)
	{
		pxMesh->ConvertPolygonFlags_AddIgnoreLOS();
	}
	if(Version<6)
	{
		pxMesh->ConvertPolygonFlags_AddWalkableSurface();
	}

	return true;
}


#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_PolygonFlags::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_PolygonFlags::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	//don't bother creating the chunk if there aren't any flags to save
	if(!pxMesh->PolygonFlagDataValid) return;

	//also don't bother if no flags have been set
	bool bAnySet = false;
	for(int i=0; i<pxMesh->NumberOfPolygons && !bAnySet; i++)
	{
		if(pxMesh->PolygonFlagDataPtr[i].Flags)
		{
			bAnySet = true;
			break;
		}
	}
	if(!bAnySet) return;

	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_POLYGONFLAGS, ASURA_CHUNK_MESH_POLYGONFLAGS_VERSION, uChunkFlags);
	
	

	xStream << pxMesh->NumberOfPolygons;
	xStream << pxMesh->Name;

	const Asura_Mesh_PolygonFlags* pxSrc = pxMesh->PolygonFlagDataPtr;

	for(int i=0; i<pxMesh->NumberOfPolygons; i++, pxSrc++)
	{
		xStream << pxSrc->Flags;
	}

	xStream.WriteChunk_End();
}

#endif

////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Chunk_Mesh_LOD **
**----------------------*/
#define ASURA_CHUNK_MESH_LOD_VERSION 0

bool Asura_Chunk_Mesh_LOD::Process(Asura_Handle iFileHandle) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	Asura_Assert(Version<=ASURA_CHUNK_MESH_LOD_VERSION,"Unknown chunk version");

	const Asura_Char* szMeshName = Asura_FileAccess::ReadString(iFileHandle);
	Asura_Mesh *pxMesh = Asura_Mesh_System::GetMesh(szMeshName);
	if(!pxMesh || pxMesh->IsDetailMesh())
	{
		Asura_Assert(pxMesh,"Failed to find LOD mesh");
		Asura_Assert(!pxMesh->IsDetailMesh(),"LOD data for this mesh has already been loaded");
		return true;
	}

	//determine the name for the main mesh , and get it
	if(!pxMesh->HasDetailMeshName())
	{
		Asura_Assert(0,"Unexpected name for LOD mesh");
		return true;
	}
	const Asura_Char* szParentMeshName = szMeshName+3;
	Asura_Mesh *pxParentMesh = Asura_Mesh_System::GetMesh( szParentMeshName );
	if(!pxParentMesh)
	{
		Asura_Assert(pxParentMesh,"Failed to find main mesh for LOD");
		return true;
	}

	pxMesh->m_fMinDistance = m_fMinDistance;

	//add this mesh to the linked list on the main mesh
	while(pxParentMesh->m_pxNextLowerMesh)
	{
		if(pxParentMesh->m_pxNextLowerMesh->m_fMinDistance>m_fMinDistance)
		{
			break;
		}
		pxParentMesh = pxParentMesh->m_pxNextLowerMesh;
	}
	pxMesh->m_pxNextLowerMesh = pxParentMesh->m_pxNextLowerMesh;
	pxParentMesh->m_pxNextLowerMesh = pxMesh;

	return true;

}

#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_LOD::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	if(!pxMesh->IsDetailMesh()) return;

	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh,xStream);
}

void Asura_Chunk_Mesh_LOD::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream)
{
	if(!pxMesh->IsDetailMesh()) return;

	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_LOD, ASURA_CHUNK_MESH_LOD_VERSION, uChunkFlags);

	xStream << pxMesh->m_fMinDistance;
	xStream << pxMesh->Name;

	xStream.WriteChunk_End();
}

#endif

////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Asura_Chunk_Mesh_Misc **
**-----------------------*/
#define ASURA_CHUNK_MESH_MISC_VERSION 12

bool Asura_Chunk_Mesh_Misc::Process(Asura_Handle iFileHandle) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	Asura_Assert(Version<=ASURA_CHUNK_MESH_MISC_VERSION,"Unknown chunk version");

	const Asura_Char* pszMeshName = Asura_FileAccess::ReadString(iFileHandle);
	Asura_Mesh *pxMesh = Asura_Mesh_System::GetMesh(pszMeshName);
	if(!pxMesh)
	{
		Asura_Assertf(pxMesh,("Failed to find mesh %s",pszMeshName));
		return false;
	}

	Asura_Chunk_File_Stream xStream(iFileHandle);
	return Process(xStream,pxMesh,true,true);
}

bool Asura_Chunk_Mesh_Misc::Process(Asura_Chunk_Stream& xStream, Asura_Mesh* pxMesh, bool bAlreadyReadName, bool bInMeshSystem) const
{
	Asura_Assert(Version<=ASURA_CHUNK_MESH_MISC_VERSION,"Unknown chunk version");

	if(!bAlreadyReadName) xStream.SkipString();

	xStream >> pxMesh->m_uFragmentHash;

	if(Flags & ChunkFlag_CollisionMesh)
	{
		//read in the size of the collision mesh data
		int iCollisionChunkSize = 0;
		xStream >> iCollisionChunkSize;

		//create the collision mesh from the chunk buffer
		if(!pxMesh->m_pxCollisionMesh)
		{
			pxMesh->m_pxCollisionMesh = new Asura_CollisionMesh();
			pxMesh->m_pxCollisionMesh->ReadFromChunkStream(xStream);
		}
		else
		{
			xStream.Skip(iCollisionChunkSize);
		}
		
	}

	if(Flags & ChunkFlag_AddToShapeTable)
	{
		pxMesh->m_bAddToShapeTable = true;

		#ifdef ASURA_MAIN
		if(bInMeshSystem)
		{
			const Asura_Hash_ID	uHash	= Asura_GetHashID(pxMesh->Name.GetString());
			Asura_Shape*		pxShape	= Asura_Shape_System::NewShape(uHash);

			// Initialise the shape...
			pxShape->Initialise(pxMesh);
		}
		#endif
	}

	if(Flags & ChunkFlag_PolygonGuids)
	{
		int iNumPolygons;
		xStream >> iNumPolygons;
		
		Asura_Assert(!pxMesh->m_pxPolygonGuids,"Polygon guids already allocated");
 		Asura_Assert(iNumPolygons==pxMesh->NumberOfPolygons,"Number of polygons doesn't match");

		pxMesh->m_pxPolygonGuids = new Asura_Guid[iNumPolygons];
		xStream.Read(sizeof(Asura_Guid)*iNumPolygons, pxMesh->m_pxPolygonGuids);
	}

	if(Flags & ChunkFlag_MergedEntities)
	{
		xStream >> pxMesh->m_iNumberOfMergedEntities;

		pxMesh->m_pxMergedEntities = new Asura_Mesh_MergedEntity[pxMesh->m_iNumberOfMergedEntities];
		xStream.Read(sizeof(Asura_Mesh_MergedEntity)*pxMesh->m_iNumberOfMergedEntities, pxMesh->m_pxMergedEntities);
	}

	if(Flags & ChunkFlag_MergedEntityCollisionMesh)
	{
		//read in the size of the collision mesh data
		int iCollisionChunkSize = 0;
		xStream >> iCollisionChunkSize;

		//create the collision mesh from the chunk buffer
		if(!pxMesh->m_pxMergedEntityCollisionMesh)
		{
			pxMesh->m_pxMergedEntityCollisionMesh = new Asura_CollisionMesh();
			pxMesh->m_pxMergedEntityCollisionMesh->ReadFromChunkStream(xStream);
		}
		else
		{
			xStream.Skip(iCollisionChunkSize);
		}
	}

	if(Flags & ChunkFlag_CollPolygonGuids)
	{
		int iNumPolygons;
		xStream >> iNumPolygons;
		
		Asura_Assert(!pxMesh->m_pxCollPolygonGuids,"Collision polygon guids already allocated");
		Asura_Assert(pxMesh->m_pxMergedEntityCollisionMesh,"Merged entity collision mesh should exist");
 		Asura_Assert(iNumPolygons==(int)pxMesh->m_pxMergedEntityCollisionMesh->GetNumberOfPolys(),"Number of polygons doesn't match");

		if( iNumPolygons > 0 )
		{
			pxMesh->m_pxCollPolygonGuids = new Asura_Guid[iNumPolygons];
			xStream.Read(sizeof(Asura_Guid)*iNumPolygons, pxMesh->m_pxCollPolygonGuids);
		}
	}

	if(Flags & ChunkFlag_PolygonLightData)
	{
		//load light data
		xStream >> pxMesh->m_iNumberOfPolygonLights;
		pxMesh->m_pxPolygonLights = (pxMesh->m_iNumberOfPolygonLights) ? new Asura_Mesh_PolygonLight[pxMesh->m_iNumberOfPolygonLights] : 0;

		Asura_Mesh_PolygonLight* pxLightData = pxMesh->m_pxPolygonLights;
		for(int i=0;i<pxMesh->m_iNumberOfPolygonLights;i++,pxLightData++)
		{
			xStream >> pxLightData->m_uLightID;
			xStream >> pxLightData->m_xColour;
			xStream >> pxLightData->m_fPower;
			xStream >> pxLightData->m_fAngle;
			xStream >> pxLightData->m_uAccuracy;


			u_int uBleached;
			xStream >> uBleached;
			pxLightData->m_bBleached = uBleached  ? true : false;
			if(Version<3)
			{
				pxLightData->m_bShortRange = false;
			}
			else
			{
				u_int uShortRange;
				xStream >> uShortRange;
				pxLightData->m_bShortRange = uShortRange  ? true : false;
			}
			if(Version<4)
			{
				pxLightData->m_bUseRange = false;
				pxLightData->m_fInnerRange = 0.0f;
				pxLightData->m_fOuterRange = 0.0f;
			}
			else
			{
				u_int uUseRange;
				xStream >> uUseRange;
				pxLightData->m_bUseRange = uUseRange  ? true : false;

				xStream >> pxLightData->m_fInnerRange;
				xStream >> pxLightData->m_fOuterRange;
			}
			if(Version<6)
			{
				pxLightData->m_fReflectionChance = 0.25f;
			}
			else
			{
				xStream >> pxLightData->m_fReflectionChance;
			}
			if(Version<7)
			{
				pxLightData->m_uLightSets = 1;
			}
			else
			{
				xStream >> pxLightData->m_uLightSets;
			}

		}

		//and the polygon light IDs
		pxMesh->m_puPolygonLightIDs = new u_int[pxMesh->NumberOfPolygons];
		xStream.Read(sizeof(u_int)*pxMesh->NumberOfPolygons, pxMesh->m_puPolygonLightIDs);
	}

	if(Flags & ChunkFlag_MaxXRef)
	{
		//load Max XRef data
		xStream >> pxMesh->m_xMeshPosition;
		xStream >> pxMesh->m_xMeshOrient;
		xStream >> pxMesh->m_xXRefFilename;
		xStream >> pxMesh->m_xXRefObjectName;

	}

	if(Version >= 5)
	{
		xStream >> pxMesh->m_uMaxWireframeColour;
	}

	// JGP - Version 8 introduces layer
	// and external data
	if ( Version >= 8 )
	{
		xStream >> pxMesh->m_xLayer;
		xStream >> pxMesh->m_xExternalData;
	}

	// JGP - Polygon Indices

	if ( ( Version >= 9 ) && ( Flags & ChunkFlag_PolygonIndices ) )
	{
		Asura_Assertf( false == pxMesh->m_bPolygonIndexDataValid && NULL == pxMesh->m_piPolygonIndexDataPtr, ("Loading Polygon Index data over the top of existing data for mesh '%s'", pxMesh->Name.GetString()));

		// ARC - 25 May 2007 12:01:49 - Prevent a memory leak if loading over the top of existing data
		if (pxMesh->m_piPolygonIndexDataPtr != NULL)
		{
			delete[] pxMesh->m_piPolygonIndexDataPtr;
		}

		pxMesh->m_bPolygonIndexDataValid = true;
		pxMesh->m_piPolygonIndexDataPtr = new int[ pxMesh->NumberOfPolygons ];

		xStream.Read(sizeof(int)*pxMesh->NumberOfPolygons, pxMesh->m_piPolygonIndexDataPtr);

	}

	// Project polygon data
	if( Version >= 10 && (Flags & ChunkFlag_ProjectPolygonData))
	{
		// ARC - 25 May 2007 12:01:49 - Prevent a memory leak if loading over the top of existing data
		if (pxMesh->m_pxProjectPolygonData != NULL)
		{
			delete[] pxMesh->m_pxProjectPolygonData;
		}

		pxMesh->m_pxProjectPolygonData = new Asura_Mesh_ProjectPolygonData[pxMesh->NumberOfPolygons];

		xStream.Read(sizeof(u_int)*pxMesh->NumberOfPolygons, pxMesh->m_pxProjectPolygonData);
	}

	// Vertex flags
	if( Version >= 10 && (Flags & ChunkFlag_VertexFlags))
	{
		// ARC - 25 May 2007 12:01:49 - Prevent a memory leak if loading over the top of existing data
		if (pxMesh->m_pxVertexFlags != NULL)
		{
			delete[] pxMesh->m_pxVertexFlags;
		}

		pxMesh->m_pxVertexFlags = new Asura_Mesh_VertexFlags[pxMesh->NumberOfVertices];

		xStream.Read(sizeof(u_int)*pxMesh->NumberOfVertices, pxMesh->m_pxVertexFlags);
	}

	//Platform specific mesh
	if( Version >= 11 && (Flags & ChunkFlag_PlatformMesh))
	{
		char szDefaultMeshName[256];
		xStream.ReadString(szDefaultMeshName,sizeof(szDefaultMeshName));
		
		if(bInMeshSystem)
		{
			Asura_Mesh *pxDefaultMesh = Asura_Mesh_System::GetMesh(szDefaultMeshName);

			Asura_Assertf(pxDefaultMesh,("Failed to find default mesh '%s' for platform specific mesh %s",szDefaultMeshName,pxMesh->Name.GetString()));
			if(pxDefaultMesh)
			{
				pxDefaultMesh->AddPlatformMesh(pxMesh);
			}
		}

		xStream >> pxMesh->m_uPlatforms;
	}

	//3dsmax selection sets
	if( Version >= 12 && (Flags & ChunkFlag_SelectionSets))
	{
		xStream >> pxMesh->m_uNumSelectionSets;
		pxMesh->m_pxSelectionSets = pxMesh->m_uNumSelectionSets ? new Asura_String[pxMesh->m_uNumSelectionSets] : 0;
		for(u_int i=0; i<pxMesh->m_uNumSelectionSets; i++)
		{
			xStream >> pxMesh->m_pxSelectionSets[i];
		}
	}

	return true;

}


#ifdef ASURA_TOOLS
void Asura_Chunk_Mesh_Misc::Create( const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock,const int* piMaterialIDToIndexConversion)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxMesh, xStream, piMaterialIDToIndexConversion);
}

void Asura_Chunk_Mesh_Misc::Create( const Asura_Mesh* pxMesh, Asura_Chunk_Stream& xStream,const int* piMaterialIDToIndexConversion)
{
	bool bPolygonGuids = false;
	if(pxMesh->m_pxPolygonGuids)
	{
		//check for existence of some valid guids
		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			if(pxMesh->m_pxPolygonGuids[i] != ASURA_GUID_UNREGISTERED)
			{
				bPolygonGuids = true;
				break;
			}
		}
	}

	bool bMergedEntities = pxMesh->m_iNumberOfMergedEntities ? true : false;

	bool bMergedCollMesh = pxMesh->m_pxMergedEntityCollisionMesh ? true : false;
	bool bCollPolygonGuids = pxMesh->m_pxCollPolygonGuids ? true : false;

	bool bLightData = pxMesh->m_iNumberOfPolygonLights && pxMesh->m_puPolygonLightIDs;

	bool bMaxXRef = !pxMesh->m_xXRefFilename.IsBlank();

	bool bPolygonIndices = pxMesh->m_bPolygonIndexDataValid;

	bool bPlatformMesh = pxMesh->IsPlatformSpecificMesh();
	bool bSelectionSets = (pxMesh->m_uNumSelectionSets > 0);

	//project polygon data needed ?
	bool bProjectPolygonData = false;
	if(pxMesh->m_pxProjectPolygonData)
	{
		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			if(pxMesh->m_pxProjectPolygonData[i].m_uData)
			{
				bProjectPolygonData = true;
				break;
			}
		}
	}

	//vertex flags needed ?
	bool bVertexFlags = false;
	if(pxMesh->m_pxVertexFlags)
	{
		for(int i=0;i<pxMesh->NumberOfVertices;i++)
		{
			if(pxMesh->m_pxVertexFlags[i].m_uFlags)
			{
				bVertexFlags = true;
				break;
			}
		}
	}

	int iVersion = ASURA_CHUNK_MESH_MISC_VERSION;

	//Write as previous version if possible to avoid asserts from tools/plugins being updated at different times
	//RAR 23/10/2006
	if(iVersion == 12 && !bSelectionSets)
	{
		iVersion = 11;
	}
	if(iVersion == 11 && !bPlatformMesh)
	{
		iVersion = 10;
	}
	if(iVersion == 10 && !bProjectPolygonData && !bVertexFlags)
	{
		iVersion = 9;	
	}


	u_int uChunkFlags = 0;
	if(pxMesh->m_bAddToShapeTable) uChunkFlags |= ChunkFlag_AddToShapeTable;
	if(pxMesh->m_pxCollisionMesh) uChunkFlags |= ChunkFlag_CollisionMesh;
	if(bPolygonGuids) uChunkFlags |= ChunkFlag_PolygonGuids;
	if(bMergedEntities) uChunkFlags |= ChunkFlag_MergedEntities;
	if(bMergedCollMesh) uChunkFlags |= ChunkFlag_MergedEntityCollisionMesh;
	if(bCollPolygonGuids) uChunkFlags |= ChunkFlag_CollPolygonGuids;
	if(bLightData) uChunkFlags |= ChunkFlag_PolygonLightData;
	if(bMaxXRef) uChunkFlags |= ChunkFlag_MaxXRef;
	if( bPolygonIndices ) uChunkFlags |= ChunkFlag_PolygonIndices;
	if(bProjectPolygonData) uChunkFlags |= ChunkFlag_ProjectPolygonData;
	if(bVertexFlags) uChunkFlags |= ChunkFlag_VertexFlags;
	if(bPlatformMesh) uChunkFlags |= ChunkFlag_PlatformMesh;
	if(bSelectionSets) uChunkFlags |= ChunkFlag_SelectionSets;

	xStream.WriteChunk_Start(ASURA_CHUNK_MESH_MISC, iVersion, uChunkFlags);


	xStream << pxMesh->Name;
	xStream << pxMesh->m_uFragmentHash;

	if(pxMesh->m_pxCollisionMesh)
	{
		pxMesh->m_pxCollisionMesh->WriteToChunkStream(xStream, true, piMaterialIDToIndexConversion);
	}

	if(bPolygonGuids)
	{
		xStream << pxMesh->NumberOfPolygons;
		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			xStream << pxMesh->m_pxPolygonGuids[i];
		}
	}

	if(bMergedEntities)
	{
		xStream << pxMesh->m_iNumberOfMergedEntities;
		for(int i=0;i<pxMesh->m_iNumberOfMergedEntities;i++)
		{
			xStream << pxMesh->m_pxMergedEntities[i].m_uGuid;
			xStream << pxMesh->m_pxMergedEntities[i].m_uChecksum;
		}
	}

	if(bMergedCollMesh)
	{
		pxMesh->m_pxMergedEntityCollisionMesh->WriteToChunkStream(xStream, true, piMaterialIDToIndexConversion);
	}

	if(bCollPolygonGuids)
	{
		int iNumPolys = (int)pxMesh->m_pxMergedEntityCollisionMesh->GetNumberOfPolys();

		xStream << iNumPolys;
		for(int i=0;i<iNumPolys;i++)
		{
			xStream << pxMesh->m_pxCollPolygonGuids[i];
		}
	}

	if(bLightData)
	{
		//save the light data
		xStream << pxMesh->m_iNumberOfPolygonLights;
		Asura_Mesh_PolygonLight* pxLightData = pxMesh->m_pxPolygonLights;
		for(int i=0;i<pxMesh->m_iNumberOfPolygonLights;i++,pxLightData++)
		{
			xStream << pxLightData->m_uLightID;
			xStream << pxLightData->m_xColour;
			xStream << pxLightData->m_fPower;
			xStream << pxLightData->m_fAngle;
			xStream << pxLightData->m_uAccuracy;
			
			u_int uTemp = pxLightData->m_bBleached ? 1 : 0;
			xStream << uTemp;

			uTemp = pxLightData->m_bShortRange ? 1 : 0;
			xStream << uTemp;
				
			uTemp = pxLightData->m_bUseRange ? 1 : 0;
			xStream << uTemp;

			xStream << pxLightData->m_fInnerRange;
			xStream << pxLightData->m_fOuterRange;
			xStream << pxLightData->m_fReflectionChance;
			xStream << pxLightData->m_uLightSets;
		}

		//and the polygon light IDs
		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			xStream << pxMesh->m_puPolygonLightIDs[i];
		}
	}

	if(bMaxXRef)
	{
		xStream << pxMesh->m_xMeshPosition;
		xStream << pxMesh->m_xMeshOrient;
		xStream << pxMesh->m_xXRefFilename;
		xStream << pxMesh->m_xXRefObjectName;
	}

	xStream << pxMesh->m_uMaxWireframeColour;

	// JGP - new data
	xStream << pxMesh->m_xLayer;
	xStream << pxMesh->m_xExternalData;

	// JGP - Polygon Indices

  	if ( bPolygonIndices )
	{
		for( int i = 0; i < pxMesh->NumberOfPolygons; i++ )
		{
			xStream << pxMesh->m_piPolygonIndexDataPtr[ i ];
		}
	}

	// Project polygon data
	if(bProjectPolygonData)
	{
		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			xStream << pxMesh->m_pxProjectPolygonData[i].m_uData;
		}	
	}

	// Vertex flags
	if(bVertexFlags)
	{
		for(int i=0;i<pxMesh->NumberOfVertices;i++)
		{
			xStream << pxMesh->m_pxVertexFlags[i].m_uFlags;
		}
	}

	//Platform specific mesh data
	if(bPlatformMesh)
	{
		const Asura_Mesh* pxMainMesh = pxMesh->GetDefaultPlatformMesh();
		xStream << pxMainMesh->Name;
		xStream << pxMesh->m_uPlatforms;
	}

	//3dsmax selection sets
	if(bSelectionSets)
	{
		xStream << pxMesh->m_uNumSelectionSets;
		for(u_int i=0; i<pxMesh->m_uNumSelectionSets; i++)
		{
			xStream << pxMesh->m_pxSelectionSets[i];
		}
	}

	xStream.WriteChunk_End();
}

#endif

/////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Asura_Chunk_Mesh_CollMeshGenData **
**-----------------------------------*/

#define ASURA_CHUNK_MESH_COLLMESHGENDATA_VERSION 0

bool Asura_Chunk_Mesh_CollMeshGenData::Process(Asura_Handle iFileHandle) const
{
	if(g_bSkipChunksForCurrentMesh) return false;

	if(Version>ASURA_CHUNK_MESH_COLLMESHGENDATA_VERSION)
	{
		Asura_Assert(Version<=ASURA_CHUNK_MESH_COLLMESHGENDATA_VERSION,"Unknown chunk version");
		return false;
	}

	const Asura_Char* pszMeshName = Asura_FileAccess::ReadString(iFileHandle);
	Asura_Mesh *pxMesh = Asura_Mesh_System::GetMesh(pszMeshName);
	if(!pxMesh)
	{
		Asura_Assertf(pxMesh,("Failed to find mesh %s",pszMeshName));
		return false;
	}

	if(pxMesh->m_pxCollMeshGenData)
	{
		return false;//already loaded
	}
	pxMesh->m_pxCollMeshGenData = new Asura_Mesh_CollMeshGenData;
	Asura_Mesh_CollMeshGenData* pxCollMeshGenData = pxMesh->m_pxCollMeshGenData;

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,pxCollMeshGenData->m_iNumberOfPolygons,int);
	if(pxCollMeshGenData->m_iNumberOfPolygons)
	{
		pxCollMeshGenData->m_pxPolygons = new Asura_Mesh_CollPolyInfo[pxCollMeshGenData->m_iNumberOfPolygons];
	
		for(int i=0;i<pxCollMeshGenData->m_iNumberOfPolygons;i++)
		{
			ASURA_CHUNK_FILEEXTRACT(iFileHandle,pxCollMeshGenData->m_pxPolygons[i],Asura_Mesh_CollPolyInfo);
		}
	}

	ASURA_CHUNK_FILEEXTRACT(iFileHandle,pxCollMeshGenData->m_uNumberOfBoxes,u_int);
	if(pxCollMeshGenData->m_uNumberOfBoxes)
	{
		pxCollMeshGenData->m_pxBoxes = new Asura_Mesh_CollBoxInfo[pxCollMeshGenData->m_uNumberOfBoxes];

		for(u_int i=0;i<pxCollMeshGenData->m_uNumberOfBoxes;i++)
		{
			ASURA_CHUNK_FILEEXTRACT(iFileHandle,pxCollMeshGenData->m_pxBoxes[i],Asura_Mesh_CollBoxInfo);
		}
	}


	return true;
}

void Asura_Chunk_Mesh_CollMeshGenData::Create(const Asura_Mesh* pxMesh, Asura_SaveBlock* pxSaveBlock)
{
	//Is this chunk needed for this mesh?
	if(!pxMesh->m_pxCollMeshGenData) return;
	const Asura_Mesh_CollMeshGenData* pxCollMeshGenData = pxMesh->m_pxCollMeshGenData;

	int iNamelength = pxMesh->Name.GetSizeInInts();
	int iChunkSize 	= sizeof(Asura_Chunk_Mesh_CollMeshGenData) 
					+ iNamelength*4 
					+ 2*sizeof(int)//num polygons , num boxes
					+ pxCollMeshGenData->m_iNumberOfPolygons * sizeof(Asura_Mesh_CollPolyInfo)
					+ pxCollMeshGenData->m_uNumberOfBoxes * sizeof(Asura_Mesh_CollBoxInfo);


	Asura_Chunk_Mesh_CollMeshGenData* pxChunk = (Asura_Chunk_Mesh_CollMeshGenData*)pxSaveBlock->GetAccessPointer(iChunkSize);

	pxChunk->ID = ASURA_CHUNK_MESH_COLLMESHGENDATA;
	pxChunk->Size = iChunkSize;
	pxChunk->Version = ASURA_CHUNK_MESH_COLLMESHGENDATA_VERSION;
	pxChunk->Flags = 0;


	char* pcBuffer = (char*)(pxChunk+1);

	ASURA_CHUNK_FILL_STRING(pcBuffer,pxMesh->Name);

	ASURA_CHUNK_FILL(pcBuffer,pxCollMeshGenData->m_iNumberOfPolygons,int);
	for(int i=0;i<pxCollMeshGenData->m_iNumberOfPolygons;i++)
	{
		ASURA_CHUNK_FILL(pcBuffer,pxCollMeshGenData->m_pxPolygons[i],Asura_Mesh_CollPolyInfo);
	}

	ASURA_CHUNK_FILL(pcBuffer,pxCollMeshGenData->m_uNumberOfBoxes,u_int);
	for(u_int i=0;i<pxCollMeshGenData->m_uNumberOfBoxes;i++)
	{
		ASURA_CHUNK_FILL(pcBuffer,pxCollMeshGenData->m_pxBoxes[i],Asura_Mesh_CollBoxInfo);
	}

	Asura_Assert((int)pcBuffer-(int)pxChunk==iChunkSize,"Chunk size incorrect");

}

#endif // #ifdef ASURA_TOOLS
