#include "Core/Asura.h"

#include "ChunkSystem/Asura_Chunk_ModuleLinking.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#endif
#if defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)
#include "ChunkSystem/Asura_Chunk_StreamingModuleNames.h"
#endif
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Module.h"
#ifdef ASURA_TOOLS
#include "Mesh/Asura_Mesh.h"
#endif

bool Asura_Chunk_ModuleLinking::Process(Asura_Handle fileHandle) const
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		Asura_Module_Handler&	xHandler	= Asura_Module_System::GetCurrentHandler();

		// Initialise the data...
		xHandler.PreLoad();

		if(Version==1)
		{
			//check number of modules with the module index conversion thing
			Asura_ModuleIndex_Conversion::CheckNumberOfModules(m_iNumberOfModules,xHandler.GetNumberOfModules());

			//load module linking in new format
			for (int i=0; i<m_iNumberOfModules; i++)
			{
				//read in number of visible modules
				int numberOfVisibleModules;
				Asura_FileAccess::ReadData(fileHandle,&numberOfVisibleModules,sizeof(int));
				if(!numberOfVisibleModules) continue;

				//convert module index
				int moduleIndex = Asura_ModuleIndex_Conversion::ConvertIndex(i);

				if(moduleIndex<0)
				{
					//dud module , skip it
					Asura_FileAccess::SkipData(fileHandle,sizeof(int)*numberOfVisibleModules);
				}
				else
				{
					Asura_Module&	xModule	= xHandler.m_pxModuleStorage[moduleIndex];
					xModule.NumberOfVisibleModules = 0;

					//allocate memory
					xModule.VisibleModulesArraySize = numberOfVisibleModules;
					Asura_ResourceAllocateArray(xModule.VisibleModules, xModule.VisibleModulesArraySize);

					for(int m=0;m<numberOfVisibleModules;m++)
					{
						//read and convert the visible module index
						int index;
						Asura_FileAccess::ReadData(fileHandle,&index,sizeof(int));
						index = Asura_ModuleIndex_Conversion::ConvertIndex(index);
						if(index<0) continue;

						xModule.VisibleModules[xModule.NumberOfVisibleModules++] = index;
					}

				}

				Asura_ChunkLoading_System::YieldBackgroundProcessing();
			}

		}
		else if(Version==0)
		{
#ifdef ASURA_TOOLS
			//load using the old format

			Asura_ResourcePushType(ASURA_RESOURCE_TEMPORARY);
			{
				//read in the module linking
				int dataSize = ((m_iNumberOfModules*m_iNumberOfModules)+3)&~3;
				char* buffer;
				Asura_ResourceAllocateMemoryBlock(buffer, dataSize, ASURA_ALIGN_4_BYTE);
				Asura_FileAccess::ReadData(fileHandle,buffer,dataSize);

				//check number of modules is correct
				if(m_iNumberOfModules != xHandler.m_iNumberOfModules)
				{
					Asura_Assert(0,"Number of modules changed!");
				}
				else
				{
					char* visData = buffer;
					for (int i=0; i<m_iNumberOfModules; i++)
					{
						for(int m=0; m<m_iNumberOfModules; ++m,visData++)
						{
							if(*visData)
							{
								Asura_Module_System::SetModuleLink(i,m,0,true);
							}
						}
					}
				}
				Asura_ResourceDeallocateMemoryBlock(buffer);
			}
			Asura_ResourcePopType();
#else
			Asura_Assert( false, "Module linking format is out of date. Please re-export this level." );
#endif
		}
		else
		{
			Asura_Assert(0,"Unknown version number");
		}

		xHandler.m_bModuleLinkingLoaded = true;
	}
	Asura_ResourcePopType();
	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_ModuleLinking::Create(Asura_SaveBlock* saveBlock)
{
	const Asura_Module_Handler&		xHandler	= Asura_Module_System::GetCurrentHandler();
	Asura_Chunk_SaveBlock_Stream	xStream (saveBlock);

	// Write the data...

	xStream.WriteChunk_Start	(ASURA_CHUNK_MODULELINKING, 1, 0);

	xStream << xHandler.m_iNumberOfModules;

	for (int i = 0; i < xHandler.m_iNumberOfModules; i++)
	{
		const Asura_Module&						xModule	= xHandler.m_pxModuleStorage [i];

		Asura_WriteToStream_Array_Wrapper<int>	xVisibleModulesArray (xModule.NumberOfVisibleModules, xModule.VisibleModules);


		// Write the data for this module...

		xStream << xModule.NumberOfVisibleModules;
		xStream << xVisibleModulesArray;
	}

	xStream.WriteChunk_End	();
}

void Asura_Module_Handler::Save( Asura_SaveBlock* pxSaveBlock )
{	
	#ifdef ASURA_USE_STREAMING
	Asura_Chunk_StreamingModuleNames::Create( pxSaveBlock );
	#endif

	Asura_Chunk_ModuleNames::Create( pxSaveBlock );
	Asura_Chunk_ModuleLinking::Create( pxSaveBlock );
}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS

bool Asura_Chunk_ModuleNames::Process(Asura_Handle fileHandle) const
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		const Asura_Module_Handler&	xHandler	= Asura_Module_System::GetCurrentHandler();

		Asura_ModuleIndex_Conversion::ClearArray();

		//allocate conversion array
		Asura_ModuleIndex_Conversion::m_iNumberOfModules = m_iNumberOfModules;
		Asura_ResourceAllocateArray(Asura_ModuleIndex_Conversion::ConversionArray, m_iNumberOfModules);

		int* piInverseConversionArray = new int[xHandler.m_iNumberOfModules];
		for(int i=0;i<xHandler.m_iNumberOfModules;i++)
		{
			piInverseConversionArray[i] = -1;
		}

		//read in each module name , and find the current index
		for(int i=0;i<m_iNumberOfModules;i++)
		{
			Asura_ModuleIndex_Conversion::ConversionArray[i] = -1;

			const Asura_Char* moduleName = Asura_FileAccess::ReadString(fileHandle);

			//check the old index first
			if(i<xHandler.m_iNumberOfModules)
			{
				if( !xHandler.m_pxModuleStorage[i].Mesh->Name.Compare(moduleName) )
				{
					Asura_Assert(piInverseConversionArray[i] == -1, "Error building conversion array.Found two 'old' module indeces that are being converted to the same current module index.");
					if(piInverseConversionArray[i] == -1)
					{
						Asura_ModuleIndex_Conversion::ConversionArray[i] = i;
						piInverseConversionArray[i] = i;
					}
					continue;
				}
			}

			//check each module in turn then
			for(int m=0;m<xHandler.m_iNumberOfModules;m++)
			{
				if( !xHandler.m_pxModuleStorage[m].Mesh->Name.Compare(moduleName) )
				{
					Asura_Assert(piInverseConversionArray[m] == -1, "Error building conversion array.Found two 'old' module indeces that are being converted to the same current module index.");
					if(piInverseConversionArray[m] == -1)
					{
						Asura_ModuleIndex_Conversion::ConversionArray[i] = m;
						piInverseConversionArray[m] = i;
					}
					break;
				}
			}
		}

		delete [] piInverseConversionArray;
	}
	Asura_ResourcePopType();
	return true;
}

void Asura_Chunk_ModuleNames::Create(Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create(xStream);
}

void Asura_Chunk_ModuleNames::Create(Asura_Chunk_Stream& xStream)
{
	const Asura_Module_Handler&	xHandler	= Asura_Module_System::GetCurrentHandler();

	u_int uChunkVersion = 0;
	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_MODULENAMES, uChunkVersion, uChunkFlags);

	xStream <<xHandler.m_iNumberOfModules;
	for (int i=0; i<xHandler.m_iNumberOfModules; i++)
	{
		xStream << xHandler.m_pxModuleStorage[i].Mesh->Name;
	}

	xStream.WriteChunk_End();
}

#endif // #ifdef ASURA_TOOLS

///////////////////////////////////////////////////////////////////////////////////////////////

int Asura_ModuleIndex_Conversion::m_iNumberOfModules = 0;
int* Asura_ModuleIndex_Conversion::ConversionArray = 0;
bool Asura_ModuleIndex_Conversion::ConversionValid = false;
#ifdef ASURA_TOOLS
bool Asura_ModuleIndex_Conversion::s_bNumModulesChanged = false;
#endif

void Asura_ModuleIndex_Conversion::ClearArray()
{
	if(ConversionArray)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
		{
			m_iNumberOfModules = 0;
			Asura_ResourceDeallocateArray(ConversionArray);
			ConversionArray = 0;
			ConversionValid = false;
		}
		Asura_ResourcePopType();
	}
}

void Asura_ModuleIndex_Conversion::CheckNumberOfModules(int oldNumberOfModules,int currentNumberOfModules)
{
#ifdef ASURA_TOOLS
    s_bNumModulesChanged = false;
#endif
    
	if(m_iNumberOfModules)
	{
		//module nams chunk was loaded , check against the number of modules given by that
		ConversionValid = (m_iNumberOfModules == oldNumberOfModules);

		//If this assert fires then the number of modules in the module names chunk (which is used to setup the
		//conversion), didn't match the number of modules known about by the chunk in the same file that is needing to do the
		//conversion.
		Asura_Assert(ConversionValid,"Module index conversion data not valid.");

#ifdef ASURA_TOOLS
        if( m_iNumberOfModules != Asura_Module_System::GetNumberOfModules() )
        {
            s_bNumModulesChanged = true;
        }
#endif
	}
	else
	{
		//no module names chunk found... so conversion is only good if the number of modules hasn't changed
		ConversionValid = (oldNumberOfModules == currentNumberOfModules);
		#ifndef ASURA_TOOLS
		//Retail builds should be using level exports , which should have everything agreeing about
		//the number of modules. RAR 22/02/2008
		Asura_Assert(ConversionValid, "Unexpected number of modules");
		#endif
	}

}



int Asura_ModuleIndex_Conversion::ConvertIndex(int moduleIndex)
{

	if(!ConversionValid) return -1;

	//Don't adjust module indices that contain connection indices as well. RAR 11/04/2006
	if(moduleIndex & 0xffff0000) return moduleIndex;

	if(m_iNumberOfModules)
	{
		Asura_Assert(moduleIndex<m_iNumberOfModules,"Module index out of range");
		return ConversionArray[moduleIndex];
	}
	return moduleIndex;

}
