//
// Asura_Chunk_Navigation.cpp
//
// Created by Kevin Lea - 25/07/2001 17:07:57
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "ChunkSystem/Asura_Chunk_Navigation.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#endif
#include "Core/Asura_AABB_Tree.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"

// For memset
#include <string.h>

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_DISABLE_AI_CODE
bool Asura_Chunk_Navigation::Process( Asura_Handle hFile ) const
{
    //Asura_Assert(false, "navigation code is disabled");
    return false;
}
#ifdef ASURA_TOOLS
bool Asura_Chunk_Navigation::Create( Asura_SaveBlock& saveBlock )
{
    return false;
}
#endif//ASURA_TOOLS
#else//not ASURA_DISABLE_AI_CODE

// C O N S T A N T S /////////////////////////////////////////////////////////

// Kevin Lea - 11/10/2002 11:26:50 - Chunk version history
//
//                     1 - Original file format, no longer supported
// (KJL - 26.06.2002)  2 - Contains Waypoints and Coverpoints data
// (KJL - 11.10.2002)  3 - Added AABB tree of waypoints
// (DWE - 31.10.2002)  4 - Resized Asura_AI_Coverpoint.
// (MP  - 30.11.2002)  5 - Added m_uFlags to Asura_AI_NavVertex.
// (AW  - 04.02.2003)  6 - Added m_usFlags to Asura_AI_NavConnection and reduced Index to unsigned short.
// (KJL - 17/05/2003)  7 - Changed the size of AABB nodes (32->12 bytes)
// (KOR - 07/07/2003)  8 - Reduced number of allocates
// (NJD - 22/10/2004)  9 - Added zone to nav vertex
// (NJD - 29/10/2004) 10 - retyping of nav vertex/connection and coverpoint to save space
// (ART - 18/08/2005) 11 - necessary 4-byte padding at the end of the u_short array
// (RAR - 25/04/2006) 12 - Interzone waypoints for streaming system
// (TI  - 15/02/2007) 13 - Flags added to user defined waypoints.
// (NJD - 15/08/2007) 14 - Added compressed waypoint normal.
// (NJD - 12/12/2007) 15 - Added orientations to used defined waypoints.
// (NJD - 26/03/2008) 16 - Restructured Asura_AI_NavVertex to save space and add more zones.
// (NJD - 31/03/2009) 17 - User defined waypoints are now project-overridable. Nav system is responsible for chunking these in.

static const int CUR_CHUNK_VERSION = 17;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

class Asura_AI_Coverpoint_PreV10
{
public:
    Asura_Vector_3 Centre;
    unsigned int   m_uFlags;
    unsigned char  m_ucShortExits;
    unsigned char  m_ucLongExits;
};

class Asura_AI_NavVertex_V9
{
public:
    Asura_AI_NavVertex_V9()
    : Position( Asura_Maths::ZeroVector3 ),
      Distance( 0.0f ),
      Connections( NULL ),
      NumberOfConnections( 0 ),
      Coverpoints( NULL ),
      NumberOfCoverpoints( 0 ),
      m_uFlags( 0 ),
      m_usZone( 0 )
    {}

public:
    Asura_Vector_3          Position;
    float                   Distance;
    Asura_AI_NavConnection* Connections;
    int                     NumberOfConnections;
    unsigned short*         Coverpoints;
    int                     NumberOfCoverpoints;
    u_int                   m_uFlags;
    unsigned short          m_usZone;
};

class Asura_AI_NavVertex_V5_V8
{
public:
    Asura_AI_NavVertex_V5_V8()
    : Position( Asura_Maths::ZeroVector3 ),
      Distance( 0.0f ),
      Connections( NULL ),
      NumberOfConnections( 0 ),
      Coverpoints( NULL ),
      NumberOfCoverpoints( 0 ),
      m_uFlags( 0 )
    {}

public:
    Asura_Vector_3          Position;
    float                   Distance;
    Asura_AI_NavConnection* Connections;
    int                     NumberOfConnections;
    unsigned short*         Coverpoints;
    int                     NumberOfCoverpoints;
    u_int                   m_uFlags;
};

class Asura_AI_NavConnection_Pre_V6
{
public:
    int   Index;
    float Range;
};

class Asura_AI_NavVertex_PreV5
{
public:
    Asura_AI_NavVertex_PreV5()
    : Position( Asura_Maths::ZeroVector3 ),
      Distance( 0.0f ),
      Connections( NULL ),
      NumberOfConnections( 0 ),
      Coverpoints( NULL ),
      NumberOfCoverpoints( 0 )
    {}

    Asura_Vector_3                 Position;
    float                          Distance;
    Asura_AI_NavConnection_Pre_V6* Connections;
    int                            NumberOfConnections;
    unsigned short*                Coverpoints;
    int                            NumberOfCoverpoints;
};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

bool Asura_Chunk_Navigation::Process( Asura_Chunk_Stream& xStream ) const
{
	Asura_AI_NavVertex*					pxWaypointStorage					= Asura_AI_Navigation_System::s_pxWaypointStorage;
	Asura_AI_NavConnection*				pxNavConnectionStorage				= Asura_AI_Navigation_System::s_pxNavConnectionStorage;
	u_char*								pucWaypointWeights					= Asura_AI_Navigation_System::s_pucWaypointWeights;
	Asura_AI_PerVertexCoverpointData*	pxPerVertexCoverpointData			= Asura_AI_Navigation_System::s_pxPerVertexCoverpointData;
	u_short*							pusCoverpointIndexStorage			= Asura_AI_Navigation_System::s_pusCoverpointIndexStorage;
	Asura_AI_Coverpoint*				pxCoverpointStorage					= Asura_AI_Navigation_System::s_pxCoverpointStorage;
	Asura_AABB_Tree*					pxWaypointTree						= Asura_AI_Navigation_System::s_pxWaypointTree;

	int	iNastyPointer;	// This data should never have been written and is not compatible with the PS3.

    if ( pxWaypointStorage != NULL )
    {
        // if we allocate all the storage again we'll leak memory
        Asura_Assert( false, "Asura_Chunk_Navigation::Process - already loaded navigation data" );
        return false;
    }
    
    if( Version > CUR_CHUNK_VERSION )
    {
        Asura_Assert( false, "Asura_Chunk_Navigation::Process - navigation chunk version is too new. You need a new executable." );
        return false;
    }

#ifdef ASURA_TOOLS
    if ( Version < 4 )
    {
        Asura_Assert( false, "Asura_Chunk_Navigation::Process - ancient chunk version - no longer handled!" );
        return false;
    }
    else if ( Version < CUR_CHUNK_VERSION )
    {
        Asura_OutputToDebugger(( "Asura_Chunk_Navigation::Process - loading nav chunk v%i (current version is %i.)\n", Version, CUR_CHUNK_VERSION ));
    }
#else
    // Don't bother supporting ancient chunk versions outside of tools. For
    // now, this means we only need to cope with the current one, since the
    // previous version is getting on for a year old.
    Asura_Assert( Version >= 10, "Unsupported version." );
#endif

	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
	{
		u_int	uNumConnections			= 0;
		u_int	uNumCoverpointIndices	= 0;

		if ( m_iNumWaypoints )
		{
			// waypoints
			Asura_ResourceAllocateArray(pxWaypointStorage, m_iNumWaypoints);

			for (int w = 0; w < m_iNumWaypoints; w++)
			{
				Asura_AI_NavVertex&	xNV	= pxWaypointStorage [w];

				// Read this waypoint...
				xStream >> xNV.m_xPosition;
				float fRadius;
				xStream >> fRadius;
				xNV.SetRadius( fRadius );

				if (11 > Version)
				{
					xStream >> iNastyPointer;	// xNV.m_pxConnections;
				}

				xStream >> xNV.m_ucNumConnections;
				if ( Version >= 16 )
				{
					xStream >> xNV.m_usZone;
				}
				else
				{
					u_char ucZone;
					xStream >> ucZone;
					xNV.m_usZone = ucZone;
				}
				xStream >> xNV.m_usFlags;

                if (Version >= 14)
                {
                    xStream >> xNV.m_ucNormalIndex;
                }

				Asura_ChunkLoading_System::YieldBackgroundProcessing();
			}


			// connections
			xStream >> uNumConnections;

			// (MH - 28.10.2005) sanity check on num waypoint connections. 
			// Max waypoints is 65536 and typical num connections is about 6.
			// Allow well over this (32) which would come to 16MB of Connections data.
			// More than this is likely an error in the chunk data.
			Asura_Assert(uNumConnections < 65536 * 32, "num connections looks too big, aborting");
			if (uNumConnections >= 65536 * 32)
			{
				Asura_ResourcePopType();
				return false;
			}

			Asura_ResourceAllocateArray(pxNavConnectionStorage, uNumConnections);

			for (u_int c = 0; c < uNumConnections; c++)
			{
				Asura_AI_NavConnection&	xNC	= pxNavConnectionStorage[c];

				// Read this connection...
				xStream >> xNC.m_usIndex;
				xStream >> xNC.m_usFlags;
				float fLength;
				xStream >> fLength;
				xNC.SetLength( fLength );

                // Clear the blocked flag, as this information will get updated by the physical objects on the fly.
                // It seems that when the physical object changes went in some information got saved out and never cleared.
                // This should fix that. It's debatable whether we should ever save this flag, as it should always get reset when
                // physical objects are loaded, streaming project or not.
                xNC.m_usFlags &= ~ASURA_AI_CONNECTION_FLAGS_BLOCKED;

				Asura_ChunkLoading_System::YieldBackgroundProcessing();
			}


			// waypoint weights
			Asura_ResourceAllocateArray(pucWaypointWeights, m_iNumWaypoints);
			memset( pucWaypointWeights, 0, m_iNumWaypoints * sizeof( unsigned char ) );
			Asura_ChunkLoading_System::YieldBackgroundProcessing();
		}

		if ( m_iNumCoverpoints )
		{
			// per vertex coverpoint data
			Asura_ResourceAllocateArray(pxPerVertexCoverpointData, m_iNumWaypoints);

			for (int w = 0; w < m_iNumWaypoints; w++)
			{
				Asura_AI_PerVertexCoverpointData& xPVCD = pxPerVertexCoverpointData[w];

				// Read this data...
				if (11 > Version)
				{
					xStream >> iNastyPointer;	// xPVCD.m_pusCoverpoints;
				}

				xStream >> xPVCD.m_usNumCoverpoints;

				xStream.Skip4BytePadding();

				Asura_ChunkLoading_System::YieldBackgroundProcessing();
			}


			// coverpoint indices
			xStream >> uNumCoverpointIndices;
			Asura_ResourceAllocateArray(pusCoverpointIndexStorage, uNumCoverpointIndices);
			Asura_ReadFromStream_Array_Wrapper<u_short>	xCoverpointIndexArray(uNumCoverpointIndices, pusCoverpointIndexStorage);

			if (11 <= Version)
			{
				xStream.Skip4BytePadding();
			}

			xStream >> xCoverpointIndexArray;

			// coverpoints
			Asura_ResourceAllocateArray(pxCoverpointStorage, m_iNumCoverpoints);

			for (int c = 0; c < m_iNumCoverpoints; c++)
			{
				Asura_AI_Coverpoint& xCP = pxCoverpointStorage[c];

				// Read the data...
				xStream >> xCP.m_xCentre;
				xStream >> xCP.m_usFlags;
				xStream >> xCP.m_ucShortExits;
				xStream >> xCP.m_ucLongExits;

				Asura_ChunkLoading_System::YieldBackgroundProcessing();
			}
		}

		// go through and fix up the waypoint and per vertex coverpoint data pointers
		unsigned int uCurrentNavConnection   = 0;
		unsigned int uCurrentCoverpointIndex = 0;
		for ( int i = 0; i < m_iNumWaypoints; ++i )
		{
			Asura_AI_NavVertex& xWaypoint = pxWaypointStorage[ i ];
			if ( xWaypoint.m_ucNumConnections )
			{
				xWaypoint.m_pxConnections = pxNavConnectionStorage + uCurrentNavConnection;
				uCurrentNavConnection += xWaypoint.m_ucNumConnections;
			}
			else
			{
				xWaypoint.m_pxConnections = NULL;
			}

			if ( m_iNumCoverpoints )
			{
				Asura_AI_PerVertexCoverpointData& xPVCD = pxPerVertexCoverpointData[ i ];
				if ( xPVCD.m_usNumCoverpoints )
				{
					xPVCD.m_pusCoverpoints = pusCoverpointIndexStorage + uCurrentCoverpointIndex;
					uCurrentCoverpointIndex += xPVCD.m_usNumCoverpoints;
				}
				else
				{
					xPVCD.m_pusCoverpoints = NULL;
				}
			}

			Asura_ChunkLoading_System::YieldBackgroundProcessing();
		}

		// create, or load, the waypoint tree
		if ( m_iNumWaypoints )
		{
			if ( pxWaypointTree )
			{
				delete pxWaypointTree;
			}

			pxWaypointTree = new Asura_AABB_Tree;

			/// Kevin Lea - 11/10/2002 11:12:54 - first is the root bounding box of the tree
			xStream >> pxWaypointTree->RootBoundingBox;
			Asura_ChunkLoading_System::YieldBackgroundProcessing();

			/// Kevin Lea - 11/10/2002 11:11:16 - fairly bizarre I know, but we have to check to see if there is only
			/// one waypoint - in this case only the RootBoundingBox is used and Nodes is left as null.
			if ( m_iNumWaypoints > 1 )
			{
				Asura_ResourcePushType(ASURA_RESOURCE_AABB_TREE);
				{
					Asura_ResourceAllocateArray(pxWaypointTree->Nodes, m_iNumWaypoints - 1);
				}
				Asura_ResourcePopType();

				for (int w = 0; w < (m_iNumWaypoints - 1); w++)
				{
					pxWaypointTree->Nodes[w].ReadFromChunkStream(xStream);
					Asura_ChunkLoading_System::YieldBackgroundProcessing();
				}
			}
		}

		// Load in the user defined waypoints if they were included in the chunk.
		if ( Flags & ChunkFlag_IncludeUserDefinedWaypoints )
		{
			if ( Version >= 17 )
			{
				Asura_AI_Navigation_System::GetInstance()->ReadUDWDataFromChunkStream( xStream );
			}
			else
			{
				u_int uNumUDWs;
				xStream >> uNumUDWs;

				if ( uNumUDWs > 0 )
				{
					Asura_Vector_3* pxUDWPositions = NULL;
                    Asura_Quat*     pxUDWOrientQs  = NULL;
                    u_int*          puUDWFlags     = NULL;

					Asura_ResourceAllocateArray( pxUDWPositions, uNumUDWs );
					Asura_ResourceAllocateArray( pxUDWOrientQs,  uNumUDWs );
					Asura_ResourceAllocateArray( puUDWFlags,     uNumUDWs );
					if ( pxUDWPositions && pxUDWOrientQs && puUDWFlags )
					{
						for ( u_int u = 0; u < uNumUDWs; ++u )
						{
							xStream >> pxUDWPositions[ u ];
						}
						if ( Version >= 15 )
						{
							for ( u_int u = 0; u < uNumUDWs; ++u )
							{
								xStream >> pxUDWOrientQs[ u ];
							}
						}
						else
						{
							// Previous versions did not have orientations.
							for ( u_int u = 0; u < uNumUDWs; ++u )
							{
								pxUDWOrientQs[ u ] = Asura_Maths::IdentityQuat;
							}
						}
						if ( Version >= 13 )
						{                        
							for ( u_int u = 0; u < uNumUDWs; ++u )
							{
								xStream >> puUDWFlags[ u ];
							}
						}                        
						else
						{
							// Previous versions did not have flags.
							for ( u_int u = 0; u < uNumUDWs; ++u )
							{
								puUDWFlags[ u ] = 0;
							}
						}
						Asura_ChunkLoading_System::YieldBackgroundProcessing();

						for ( u_int u = 0; u < uNumUDWs; ++u )
						{
							Asura_AI_Navigation_System::UserDefinedWaypoint* pxUDW = Asura_AI_Navigation_System::GetInstance()->CreateUserDefinedWaypoint();
							pxUDW->AquireID();
							pxUDW->m_xPosition = pxUDWPositions[ u ];
							pxUDW->m_xOrientQ  = pxUDWOrientQs[ u ];
							pxUDW->m_uFlags    = puUDWFlags[ u ];

							#ifdef ASURA_TOOLS
							// In tools we're interested in these.
							Asura_AI_Navigation_System::s_xUDWPList.Push_Back( pxUDW );
							#else
							delete pxUDW;
							#endif
						}

						Asura_ResourceDeallocateArray( pxUDWPositions );
						Asura_ResourceDeallocateArray( pxUDWOrientQs );
						Asura_ResourceDeallocateArray( puUDWFlags );
					}
				}
			}
		}

		//load in asura interzone waypoint stuff
		if( Flags & ChunkFlag_IncludeAsuraInterZoneWaypoints)
		{
#ifdef ASURA_USE_STREAMING
			xStream >> uNumInterZoneWaypoints;
			Asura_ResourceAllocateArray(pxInterZoneWaypoints, uNumInterZoneWaypoints);
			for(u_int u=0;u<uNumInterZoneWaypoints;u++)
			{
				xStream >> pxInterZoneWaypoints[u].m_uWaypointIndex;
				xStream >> pxInterZoneWaypoints[u].m_uZoneHash;
			}
			Asura_ChunkLoading_System::YieldBackgroundProcessing();

#else
			//ignore if this exists outside of streaming
			u_int uNum;
			xStream >> uNum;
			xStream.Skip(uNum * sizeof(u_int) * 2);
#endif
		}

		//load in tools interzone waypoint stuff
		if( Flags & ChunkFlag_IncludeToolsInterZoneWaypoints)
		{
#if defined(ASURA_USE_STREAMING) && defined(ASURA_TOOLS)

			xStream >> uNumInterZoneWaypointPairs;
			Asura_ResourceAllocateArray(pxInterZoneWaypointPairs, uNumInterZoneWaypointPairs);
			for(u_int u=0;u<uNumInterZoneWaypointPairs;u++)
			{
				xStream >> pxInterZoneWaypointPairs[u].m_xThisZonePosition;
				xStream >> pxInterZoneWaypointPairs[u].m_xOtherZonePosition;
			}

			Asura_ChunkLoading_System::YieldBackgroundProcessing();

#else
			//ignore if this exists outside of streaming and tools
			u_int uNum;
			xStream >> uNum;
			xStream.Skip(uNum * sizeof(Asura_Vector_3) * 2);
#endif
		}

		// ART - 26/07/07 - These are now assigned here because it's possible to
		// process chunks in parts and we don't want to use any chunk data until
		// we have processed the entire chunk.
		Asura_AI_Navigation_System::s_usNumWaypoints			= static_cast<unsigned short>(m_iNumWaypoints);
		Asura_AI_Navigation_System::s_uNumConnections			= uNumConnections;
		Asura_AI_Navigation_System::s_uNumCoverpointIndices		= uNumCoverpointIndices;
		Asura_AI_Navigation_System::s_iNumCoverpoints			= m_iNumCoverpoints;

		Asura_AI_Navigation_System::s_pxWaypointStorage			= pxWaypointStorage;	
		Asura_AI_Navigation_System::s_pxNavConnectionStorage	= pxNavConnectionStorage;
		Asura_AI_Navigation_System::s_pucWaypointWeights		= pucWaypointWeights;	
		Asura_AI_Navigation_System::s_pxPerVertexCoverpointData	= pxPerVertexCoverpointData;
		Asura_AI_Navigation_System::s_pusCoverpointIndexStorage	= pusCoverpointIndexStorage;
		Asura_AI_Navigation_System::s_pxCoverpointStorage		= pxCoverpointStorage;
		Asura_AI_Navigation_System::s_pxWaypointTree			= pxWaypointTree;

		if ( m_iNumWaypoints > 0 )
		{
			Asura_AI_Navigation_System::s_bWaypointsValid = true;
			Asura_AI_Navigation_System::InitialiseAStarStorage();
		}
	}
	Asura_ResourcePopType();

    return true;
}

bool Asura_Chunk_Navigation::Process( Asura_Handle fileHandle ) const
{
#if defined(ASURA_TOOLS)
	// Check the version...
    if (10 > Version)
    {
        return ProcessOldVersions( fileHandle );
    }
	else
#endif	// defined(ASURA_TOOLS)
	{
    	Asura_Chunk_File_Stream	xStream(fileHandle);

		// Process the stream...
		return	(Process(xStream));
	}
}

#ifdef ASURA_TOOLS
bool Asura_Chunk_Navigation::Create( Asura_Chunk_Stream& xStream )
{
	const int   iNumWaypoints         	= Asura_AI_Navigation_System::s_usNumWaypoints;
	const u_int uNumConnections       	= Asura_AI_Navigation_System::s_uNumConnections;
	const int   iNumCoverpoints       	= Asura_AI_Navigation_System::s_iNumCoverpoints;
	const u_int uNumCoverpointIndices 	= Asura_AI_Navigation_System::s_uNumCoverpointIndices;

	const int	iNastyPointer			= 0;	// This data should never have be written and is not compatible with the PS3.

	// Write the data...
	const bool bIncludeUserDefinedWaypoints = !Tools_Level_Export::IsExport();
	u_int uFlags = bIncludeUserDefinedWaypoints ? ChunkFlag_IncludeUserDefinedWaypoints : 0;
#ifdef ASURA_USE_STREAMING
	uFlags |= ChunkFlag_IncludeAsuraInterZoneWaypoints;
	if(!Tools_Level_Export::IsExport()) uFlags |= ChunkFlag_IncludeToolsInterZoneWaypoints;
#endif

	xStream.WriteChunk_Start( ASURA_CHUNK_NAVIGATION, CUR_CHUNK_VERSION, uFlags );

	xStream << iNumWaypoints;
	xStream << iNumCoverpoints;

	//RAR 08/03/2006 For recentreing during level exports :
	const Asura_Vector_3& xPositionOffset = Tools_Level_Export::GetExportCentre();

	// Do we have any waypoints?
	if ( iNumWaypoints > 0 )
	{
		// Yes - write each waypoint in-turn...
		for (int i = 0; i < iNumWaypoints; ++i)
		{
			const Asura_AI_NavVertex&	xNV	= Asura_AI_Navigation_System::s_pxWaypointStorage[i];

			// Write this waypoint...
			xStream << xNV.m_xPosition - xPositionOffset;
			xStream << xNV.GetRadius();

			if (11 > CUR_CHUNK_VERSION)
			{
				xStream << iNastyPointer;		// xNV.m_pxConnections;
			}

			xStream << xNV.m_ucNumConnections;
			xStream << xNV.m_usZone;
			xStream << xNV.m_usFlags;
            xStream << xNV.m_ucNormalIndex;
        }

		// Write the connections in-turn..
		xStream << uNumConnections;

		for (u_int u = 0; u < uNumConnections; ++u)
		{
			const Asura_AI_NavConnection&	xNC	= Asura_AI_Navigation_System::s_pxNavConnectionStorage[u];

			// Write this connection...
			xStream << xNC.m_usIndex;
			xStream << xNC.m_usFlags;
			xStream << xNC.GetLength();
		}
	}

	// Do we have any coverpoints?
	if (iNumCoverpoints > 0)
	{
		// Yes - write the coverpoint data for each waypoint in-turn...
		for (int i = 0; i < iNumWaypoints; ++i)
		{
			const Asura_AI_PerVertexCoverpointData&	xPVCD	= Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[i];

			// Write the coverpoint data...
			if (11 > CUR_CHUNK_VERSION)
			{
				xStream << iNastyPointer;		// xPVCD.m_pusCoverpoints;
			}

			xStream << xPVCD.m_usNumCoverpoints;
			xStream.Write4BytePadding();
		}


		// Write the coverpoint indices...
		xStream << uNumCoverpointIndices;
		xStream << Asura_WriteToStream_Array_Wrapper<u_short>(uNumCoverpointIndices, Asura_AI_Navigation_System::s_pusCoverpointIndexStorage);

		if (11 > CUR_CHUNK_VERSION)
		{
			xStream.Skip4BytePadding();
		}

		// Write the coverpoints in-turn...
		for (int i = 0; i < iNumCoverpoints; ++i)
		{
			const Asura_AI_Coverpoint&	xCP	= Asura_AI_Navigation_System::s_pxCoverpointStorage[i];

			// Write the coverpoint data...
			xStream << xCP.m_xCentre - xPositionOffset;
			xStream << xCP.m_usFlags;
			xStream << xCP.m_ucShortExits;
			xStream << xCP.m_ucLongExits;
		}
	}

	// Write the waypoint tree...
	if (iNumWaypoints > 0)
	{
		// Write the bounding-box...
		//RAR 08/03/2006 I think displacing the root bounding box should take care of the whole tree...
		Asura_Bounding_Box xBB = Asura_AI_Navigation_System::s_pxWaypointTree->RootBoundingBox;
		xBB.Displace(-xPositionOffset);
		xStream << xBB;

		// Write the waypoints...
		for (int i = 0; i < (iNumWaypoints - 1); ++i)
		{
			const Asura_AABB_Tree_Node&	xAABBTN	= Asura_AI_Navigation_System::s_pxWaypointTree->Nodes[i];

			// Write the AABB tree-node...
			xAABBTN.WriteToChunkStream(xStream);
		}
	}

	// save the user defined waypoint position array if necessary
	if ( bIncludeUserDefinedWaypoints )
	{
		Asura_AI_Navigation_System::GetInstance()->WriteUDWDataToChunkStream( xStream );
	}

	// Finish with this chunk...
	xStream.WriteChunk_End();

	return true;
}

bool Asura_Chunk_Navigation::ProcessOldVersions( Asura_Handle fileHandle ) const
{
	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
	{
		Asura_AI_Navigation_System::s_usNumWaypoints    = static_cast< unsigned short >( m_iNumWaypoints );
		Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pxWaypointStorage, m_iNumWaypoints);

		const bool bCreateCoverpointData = ( m_iNumCoverpoints > 0 );

		if ( bCreateCoverpointData )
		{
			Asura_AI_Navigation_System::s_iNumCoverpoints           = m_iNumCoverpoints;
			Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pxCoverpointStorage, m_iNumCoverpoints);
			Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pxPerVertexCoverpointData, m_iNumWaypoints);
		}

		if ( m_iNumWaypoints )
		{
			// waypoint weights
			Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pucWaypointWeights, m_iNumWaypoints);
			memset( Asura_AI_Navigation_System::s_pucWaypointWeights, 0, m_iNumWaypoints * sizeof( unsigned char ) );

			if ( Version == 9 )
			{
				Asura_AI_NavVertex_V9 xV9NavVertex;

				for( int i = 0; i < m_iNumWaypoints; ++i )
				{
					Asura_FileAccess::ReadData( fileHandle, &xV9NavVertex, sizeof( Asura_AI_NavVertex_V9 ) );

					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_xPosition        = xV9NavVertex.Position;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].SetRadius( xV9NavVertex.Distance );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_pxConnections    = NULL;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections = static_cast< unsigned char >( xV9NavVertex.NumberOfConnections );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_usFlags          = static_cast< unsigned short >( xV9NavVertex.m_uFlags );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_usZone           = xV9NavVertex.m_usZone;

					if ( bCreateCoverpointData )
					{
						Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ].m_pusCoverpoints   = NULL;
						Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ].m_usNumCoverpoints = static_cast< unsigned short >( xV9NavVertex.NumberOfCoverpoints );
					}
				}
			}
			else if ( Version >= 5 )
			{
				Asura_AI_NavVertex_V5_V8 xV5V8NavVertex;

				for( int i = 0; i < m_iNumWaypoints; ++i )
				{
					Asura_FileAccess::ReadData( fileHandle, &xV5V8NavVertex, sizeof( Asura_AI_NavVertex_V5_V8 ) );

					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_xPosition        = xV5V8NavVertex.Position;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].SetRadius( xV5V8NavVertex.Distance );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_pxConnections    = NULL;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections = static_cast< unsigned char >( xV5V8NavVertex.NumberOfConnections );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_usFlags          = static_cast< unsigned short >( xV5V8NavVertex.m_uFlags );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_usZone           = 0;

					if ( bCreateCoverpointData )
					{
						Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ].m_pusCoverpoints   = NULL;
						Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ].m_usNumCoverpoints = static_cast< unsigned short >( xV5V8NavVertex.NumberOfCoverpoints );
					}
				}
			}
			else
			{
				Asura_AI_NavVertex_PreV5 xPreV5NavVertex;

				for( int i = 0; i < m_iNumWaypoints; ++i )
				{
					Asura_FileAccess::ReadData( fileHandle, &xPreV5NavVertex, sizeof( Asura_AI_NavVertex_PreV5 ) );

					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_xPosition        = xPreV5NavVertex.Position;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].SetRadius( xPreV5NavVertex.Distance );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_pxConnections    = NULL;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections = static_cast< unsigned char >( xPreV5NavVertex.NumberOfConnections );
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_usFlags          = 0;
					Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_usZone           = 0;

					if ( bCreateCoverpointData )
					{
						Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ].m_pusCoverpoints   = NULL;
						Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ].m_usNumCoverpoints = static_cast< unsigned short >( xPreV5NavVertex.NumberOfCoverpoints );
					}
				}
			}
		}

		if ( m_iNumCoverpoints )
		{
			Asura_AI_Coverpoint_PreV10 xPreV10Coverpoint;

			for ( int i = 0; i < m_iNumCoverpoints; ++i )
			{
				Asura_FileAccess::ReadData( fileHandle, &xPreV10Coverpoint, sizeof( Asura_AI_Coverpoint_PreV10 ) );

				Asura_AI_Navigation_System::s_pxCoverpointStorage[ i ].m_xCentre      = xPreV10Coverpoint.Centre;
				Asura_AI_Navigation_System::s_pxCoverpointStorage[ i ].m_usFlags      = static_cast< unsigned short >( xPreV10Coverpoint.m_uFlags );
				Asura_AI_Navigation_System::s_pxCoverpointStorage[ i ].m_ucShortExits = xPreV10Coverpoint.m_ucShortExits;
				Asura_AI_Navigation_System::s_pxCoverpointStorage[ i ].m_ucLongExits  = xPreV10Coverpoint.m_ucLongExits;
			}
		}

		unsigned int uTotalNavConnections    = 0;
		unsigned int uTotalCoverpointIndices = 0;
		if ( Version >= 8 )
		{
			Asura_FileAccess::ReadData( fileHandle, &uTotalNavConnections,    sizeof( unsigned int ) );
			Asura_FileAccess::ReadData( fileHandle, &uTotalCoverpointIndices, sizeof( unsigned int ) );

			Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pxNavConnectionStorage, uTotalNavConnections);
			Asura_AI_Navigation_System::s_uNumConnections        = uTotalNavConnections;

			if ( bCreateCoverpointData )
			{
				Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pusCoverpointIndexStorage, uTotalCoverpointIndices);
				Asura_AI_Navigation_System::s_uNumCoverpointIndices     = uTotalCoverpointIndices;
			}
		}

		unsigned int uCurrentNavConnection   = 0;
		unsigned int uCurrentCoverpointIndex = 0;
		for ( int i = 0; i < m_iNumWaypoints; ++i )
		{
			Asura_AI_NavVertex& xWaypoint = Asura_AI_Navigation_System::s_pxWaypointStorage[ i ];
			if ( xWaypoint.m_ucNumConnections )
			{
				if ( Version >= 8 )
				{
					xWaypoint.m_pxConnections = Asura_AI_Navigation_System::s_pxNavConnectionStorage + uCurrentNavConnection;
					uCurrentNavConnection += xWaypoint.m_ucNumConnections;
				}
				else
				{
					Asura_ResourceAllocateArray(xWaypoint.m_pxConnections, xWaypoint.m_ucNumConnections);
				}
				if ( Version <= 5 )
				{
					for ( int c = 0; c < xWaypoint.m_ucNumConnections; ++c )
					{
						Asura_AI_NavConnection_Pre_V6 xOldConnection;
						Asura_FileAccess::ReadData( fileHandle, &xOldConnection, sizeof( Asura_AI_NavConnection_Pre_V6 ) );

						xWaypoint.m_pxConnections[ c ].m_usFlags = 0;
						xWaypoint.m_pxConnections[ c ].SetLength( xOldConnection.Range );
						xWaypoint.m_pxConnections[ c ].m_usIndex = static_cast< unsigned short >( xOldConnection.Index );
					}
				}
				else
				{
					Asura_FileAccess::ReadData( fileHandle, xWaypoint.m_pxConnections,
						sizeof( Asura_AI_NavConnection ) * xWaypoint.m_ucNumConnections );
				}
			}
			else
			{
				xWaypoint.m_pxConnections = NULL;
			}

			if ( bCreateCoverpointData )
			{
				Asura_AI_PerVertexCoverpointData& xPVCD = Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ];
				if ( xPVCD.m_usNumCoverpoints )
				{
					if ( Version >= 8 )
					{
						xPVCD.m_pusCoverpoints = Asura_AI_Navigation_System::s_pusCoverpointIndexStorage + uCurrentCoverpointIndex;
						uCurrentCoverpointIndex += xPVCD.m_usNumCoverpoints;
					}
					else
					{
						xPVCD.m_pusCoverpoints = new unsigned short[ xPVCD.m_usNumCoverpoints ];
					}
					Asura_FileAccess::ReadData( fileHandle, xPVCD.m_pusCoverpoints,
						sizeof( unsigned short ) * xPVCD.m_usNumCoverpoints );
				}
				else
				{
					xPVCD.m_pusCoverpoints = NULL;
				}
			}
		}

		if ( Version >= 8 )
		{
			Asura_Assert( uCurrentNavConnection   == uTotalNavConnections,    "Nav connection count mismatch." );
			Asura_Assert( uCurrentCoverpointIndex == uTotalCoverpointIndices, "Coverpoint index count mismatch." );
		}
		else
		{
			// take all of the connection and coverpoint data back out of the individual waypoints, and
			// put it all in the large arrays, so the data in the Asura_AI_Navigation_System is consistent
			// across chunk versions

			int iWPIndex;

			// count the total number of connections and coverpoints
			uTotalNavConnections    = 0;
			uTotalCoverpointIndices = 0;
			for ( iWPIndex = 0; iWPIndex < Asura_AI_Navigation_System::s_usNumWaypoints; ++iWPIndex )
			{
				uTotalNavConnections +=
					Asura_AI_Navigation_System::s_pxWaypointStorage[ iWPIndex ].m_ucNumConnections;
				uTotalCoverpointIndices +=
					Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ iWPIndex ].m_usNumCoverpoints;
			}

			// allocate the storage
			Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pxNavConnectionStorage, uTotalNavConnections);
			Asura_AI_Navigation_System::s_uNumConnections           = uTotalNavConnections;
			Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pusCoverpointIndexStorage, uTotalCoverpointIndices);
			Asura_AI_Navigation_System::s_uNumCoverpointIndices     = uTotalCoverpointIndices;

			// go through the waypoints, copying over the information from the connection and coverpoint
			// storage, deleting the individual storage, and fixing up the pointers
			Asura_AI_NavVertex*               pxWaypoints      = Asura_AI_Navigation_System::s_pxWaypointStorage;
			Asura_AI_PerVertexCoverpointData* pxPVCD           = Asura_AI_Navigation_System::s_pxPerVertexCoverpointData;
			Asura_AI_NavConnection*           pxConnectionPtr  = Asura_AI_Navigation_System::s_pxNavConnectionStorage;
			unsigned short*                   pusCoverpointPtr = Asura_AI_Navigation_System::s_pusCoverpointIndexStorage;
			for ( iWPIndex = 0; iWPIndex < Asura_AI_Navigation_System::s_usNumWaypoints; ++iWPIndex )
			{
				if ( pxWaypoints[ iWPIndex ].m_ucNumConnections )
				{
					for ( int c = 0; c < pxWaypoints[ iWPIndex ].m_ucNumConnections; ++c )
					{
						*pxConnectionPtr++ = pxWaypoints[ iWPIndex ].m_pxConnections[ c ];
					}

					Asura_ResourceDeallocateArray(pxWaypoints[ iWPIndex ].m_pxConnections);
				}

				if ( bCreateCoverpointData && pxPVCD[ iWPIndex ].m_usNumCoverpoints )
				{
					for ( int c = 0; c < pxPVCD[ iWPIndex ].m_usNumCoverpoints; ++c )
					{
						*pusCoverpointPtr++ = pxPVCD[ iWPIndex ].m_pusCoverpoints[ c ];
					}

					delete [] pxPVCD[ iWPIndex ].m_pusCoverpoints;
				}
			}
		}

		// create, or load, the waypoint tree
		if( m_iNumWaypoints )
		{
			if (Asura_AI_Navigation_System::s_pxWaypointTree) delete Asura_AI_Navigation_System::s_pxWaypointTree;

			if (Version<7)
			{
				// KJL - 17/05/2003 12:24:54 - older chunks have an AABB tree node of 32 bytes rather than 12 bytes.

				// KJL - 16/05/2003 16:43:20 - skip old data
				Asura_FileAccess::SkipData(fileHandle,sizeof(Asura_Bounding_Box));
				if( m_iNumWaypoints > 1 )
				{
					// 32 bytes was the old size of Asura_AABB_Tree_Node
					Asura_FileAccess::SkipData(fileHandle,32*(m_iNumWaypoints-1));
				}

				// KJL - 17/05/2003 12:25:38 - and make new data on the fly
				Asura_AI_Navigation_System::CreateWaypointTree();
			}
			else if ( Version == 9 )
			{
				//## todo - remove this

				// create the tree from scratch and get rid of the saved data
				Asura_AI_Navigation_System::CreateWaypointTree();

				Asura_Bounding_Box xDummy;
				Asura_FileAccess::ReadData( fileHandle, &xDummy, sizeof( Asura_Bounding_Box ) );
				if ( m_iNumWaypoints > 1 )
				{
					unsigned char ucDummy;
					for ( unsigned int u = 0; u < sizeof( Asura_AABB_Tree_Node ) * ( m_iNumWaypoints - 1 ); ++u )
					{
						Asura_FileAccess::ReadData( fileHandle, &ucDummy, sizeof( unsigned char ) );
					}
				}
			}
			else
			{
				// KJL - 17/05/2003 12:26:12 - version 7 or above: we can load in the AABB tree using the same code as we used before the node size change

				Asura_AI_Navigation_System::s_pxWaypointTree = new Asura_AABB_Tree;

				/// Kevin Lea - 11/10/2002 11:12:54 - first is the root bounding box of the tree
				Asura_FileAccess::ReadData(fileHandle,&(Asura_AI_Navigation_System::s_pxWaypointTree->RootBoundingBox),sizeof(Asura_Bounding_Box));

				/// Kevin Lea - 11/10/2002 11:11:16 - fairly bizarre I know, but we have to check to see if there is only
				/// one waypoint - in this case only the RootBoundingBox is used and Nodes is left as null.
				if( m_iNumWaypoints > 1 )
				{
					Asura_ResourcePushType(ASURA_RESOURCE_AABB_TREE);
					{
						Asura_ResourceAllocateArray(Asura_AI_Navigation_System::s_pxWaypointTree->Nodes, m_iNumWaypoints - 1);
						Asura_FileAccess::ReadData(fileHandle,(Asura_AI_Navigation_System::s_pxWaypointTree->Nodes),sizeof(Asura_AABB_Tree_Node)*(m_iNumWaypoints-1));
					}
					Asura_ResourcePopType();
				}
			}
		}

		Asura_AI_Navigation_System::s_bWaypointsValid = true;
		Asura_AI_Navigation_System::InitialiseAStarStorage();
	}
	Asura_ResourcePopType();
    return true;
}

#endif // #ifdef ASURA_TOOLS

#endif//#else not ASURA_DISABLE_AI_CODE

