//////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_ReverbPresets.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

#include "ChunkSystem/Asura_Chunk_ReverbPresets.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Resources/Asura_ResourceSet.h"
#include "Core/Asura_HashTree.h"
#include "Sound/Asura_Audio_System.h"

// Tools
#if defined ASURA_TOOLS
	#include "Misc/Tools_Timed_Messages.h"
#endif

/******************************************************************************/
/* Asura_Chunk_ReverbPresets Static Class Data                                */
/******************************************************************************/

const int Asura_Chunk_ReverbPresets::iCURRENT_VERSION = 0;

/******************************************************************************/
/* Asura_Chunk_ReverbPresets Class Functions                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_ReverbPresets::Process()
//
// Processes a chunk containing reverb presets from a file
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Chunk_ReverbPresets::Process( const Asura_Handle iFileHandle )
{
	Asura_Chunk_File_Stream xStream( iFileHandle );

	return Process( xStream );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_ReverbPresets::Process()
//
// Processes a chunk containing reverb presets from a chunk stream
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Chunk_ReverbPresets::Process( Asura_Chunk_Stream& xStream )
{
	if( Version > iCURRENT_VERSION )
	{
		Asura_Assertf( false,
			("Reverb presets chunk version (%d) is too new. You may need new code.", Version) );

		return false;
	}

	u_int uNumReverbs;
	xStream >> uNumReverbs;

	for( u_int u = 0; u < uNumReverbs; ++u )
	{
		Asura_String xName;
		xStream >> xName;
		
		bool bAlreadyExists;
		Asura_Audio_Reverb* const pxReverb = Asura_Audio_System::AddReverb( xName.GetString(), bAlreadyExists );
		if( bAlreadyExists || !pxReverb )
		{
			xStream.Skip(44);//skip 4 ints and 7 floats
		}
		else
		{
			xStream >> pxReverb->m_iRoom;
			xStream >> pxReverb->m_iRoomHF;
			xStream >> pxReverb->m_fRoomRolloffFactor;
			xStream >> pxReverb->m_fDecayTime;
			xStream >> pxReverb->m_fDecayHFRatio;
			xStream >> pxReverb->m_iReflections;
			xStream >> pxReverb->m_fReflectionsDelay;
			xStream >> pxReverb->m_iReverb;
			xStream >> pxReverb->m_fDiffusion;
			xStream >> pxReverb->m_fDensity;
			xStream >> pxReverb->m_fHFReference;
		}
	}

	return true;
}

/******************************************************************************/
/* Asura_Chunk_ReverbPresets Static Class Functions                           */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_ReverbPresets::Create()
//
// Creates a chunk containing reverb presets and saves it to a chunk stream.
// Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

#if defined ASURA_TOOLS

bool Asura_Chunk_ReverbPresets::Create( Asura_Chunk_Stream& xStream )
{
	xStream.WriteChunk_Start( ASURA_CHUNK_REVERB_PRESETS, iCURRENT_VERSION, 0 );

	const u_int uNumReverbs = Asura_Audio_System::GetNumReverbs();
	xStream << uNumReverbs;

	const Asura_ResourceSet< Asura_Audio_Reverb >& xSet = Asura_Audio_System::s_xReverbs;

	for( Asura_ResourceSet_ConstIterator< Asura_Audio_Reverb > xIter( xSet );
		!xIter.Done(); xIter.Next() )
	{
		const Asura_Audio_Reverb* const pxReverb = xIter.GetCurrent();
		if( !pxReverb ) return false;

		pxReverb->WriteToChunkStream( xStream );
	}

	xStream.WriteChunk_End();

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_ReverbPresets::Create()
//
// As above, but this version only save the reverbs in the given hash tree. Used
// by the level export process
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Chunk_ReverbPresets::Create( Asura_Chunk_Stream& xStream, const Asura_HashTree& xTree )
{
	xStream.WriteChunk_Start( ASURA_CHUNK_REVERB_PRESETS, iCURRENT_VERSION, 0 );

	const u_int uNumReverbs = xTree.GetNumberOfNodes();

	// First check all the reverbs are valid. If some are missing, need to decrement the number in
	// the chunk, otherwise loading on the retail platform will read beyond the chunk and the load
	// will be completely broken

	u_int uNumValidReverbs = uNumReverbs;

	for( u_int u = 0; u < uNumReverbs; ++u )
	{
		if( !Asura_Audio_System::GetReverb( xTree.FindHashIDByIndex(u), false ) )
		{
			--uNumValidReverbs;
		}
	}

	xStream << uNumValidReverbs;

	for( u_int u = 0; u < uNumReverbs; ++u )
	{
		const Asura_Hash_ID uHash = xTree.FindHashIDByIndex(u);

		const Asura_Audio_Reverb* const pxReverb = Asura_Audio_System::GetReverb( uHash, false );

		if( pxReverb )
		{
			pxReverb->WriteToChunkStream( xStream );
		}
		else
		{
			Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR,
				"Trying to export reverb 0x%08X, which doesn't exist.", uHash );
		}
	}

	xStream.WriteChunk_End();

	return false;
}

#endif // defined ASURA_TOOLS

#endif // defined ASURA_USE_AUDIO_SYSTEM
