//////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_TiledTexture.cpp
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Implementation of chunk for a tiled texture
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "ChunkSystem/Asura_Chunk_TiledTexture.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "Core/Asura_Hash.h"
#include "FileAccess/Asura_FileAccess.h"

#ifdef ASURA_TOOLS
// Tools
#include "TextureTiler/Tools_Mode_TextureTiler.h"
#include "TextureTiler/Tools_TextureTiler_Export.h"
#endif //ASURA_TOOLS


// Chunk Version
//
// [2] added hash ID's for each tile
const int ASURA_CHUNK_TILEDTEXTURE_VERSION = 2;

bool Asura_Chunk_TiledTexture::Process(Asura_Handle xFileHandle) const
{
	Asura_Chunk_File_Stream xStream( xFileHandle );

	return Process( xStream );
}

bool Asura_Chunk_TiledTexture::Process( Asura_Chunk_Stream& xStream, Asura_String* const pxTileNameBegin /* = NULL */, Asura_String* const pxTileNameEnd /* = NULL */ ) const
{
	Asura_Assert( Version<=ASURA_CHUNK_TILEDTEXTURE_VERSION, "Tiled Texture - Unsupported Chunk version!" );

	Asura_TiledTextureInfo	xTileToAdd;
	Asura_String			xNameOfParentTexture;
	u_int					uTextureHashID;

	// Get the name of the parent texture to which all the tiles reffer
	xStream >> xNameOfParentTexture; 

	if( Version > 1 )
	{
		xStream >> uTextureHashID;
	}
	else
	{
		uTextureHashID = Asura_GetHashID_IgnoreExtension( xNameOfParentTexture.GetString() );
	}

	
	#ifdef ASURA_TOOLS	
	const int iAsuraFileNameIndex( Asura_TiledTextureManagement::AddAsuraFileName( Asura_FileAccess::GetFilename( xStream.GetFileHandle() ) ) );

	Asura_String* pxTileNameItr(pxTileNameBegin);
	#endif
	
	// iterate over all tiles in chunk (m_iNTiles read in on instanciation (usually done by recasting a block of memory :~ - NDB 181007) )
	for( int i=0;i<m_iNTiles;++i )
	{
		// Read in the tiles and add them to the manager
		xStream >> xTileToAdd;

		xTileToAdd.m_uTextureHashID = uTextureHashID;

		if( Version > 1 )
		{
			#ifdef ASURA_TOOLS
			xStream >> xTileToAdd.m_xNameOfTextureUsed;	
			#else
			xStream.SkipString();
			#endif

			// read hash from chunk file
			xStream >> xTileToAdd.m_uTileHashID;

			// version 2 and above place the texture resource chunk before the tile chunk. This means that the texture is 
			// probably already loaded into texture management and texture index can be set here - NDB 011107
			xTileToAdd.m_iIndexOfTextureUsed = Asura_Texture_Management::GetIndexFromHashID( xTileToAdd.m_uTextureHashID );
		}
		else
		{	
			// find hash from texture file name
			#ifdef ASURA_TOOLS
			xStream >> xTileToAdd.m_xNameOfTextureUsed;	
			xTileToAdd.m_uTileHashID = Asura_GetHashID( xTileToAdd.m_xNameOfTextureUsed.GetString() );
			#else
			Asura_String xTileTextureName;
			xStream >> xTileTextureName;
			xTileToAdd.m_uTileHashID = Asura_GetHashID( xTileTextureName.GetString() );
			#endif

			// version 1 chunks place the texture resource chunk after the tile chunk. This means that the texture has not 
			// been loaded at this stage and we have to rely on a call to SetTextureHandlesFromHash|Names() - NDB 011107
			xTileToAdd.m_iIndexOfTextureUsed = ASURA_TEXTURE_NONE;
		}


		#ifdef ASURA_TOOLS
		// copy original name of texture used into collection provided (if required, probably only in Tools - NDB 181007)
		if( (pxTileNameItr) && (pxTileNameEnd) && (pxTileNameItr!=pxTileNameEnd) )
		{
			// older version 1 chunks made with stand alone texture tiler did not have preceeding \ so it gets added here, 
			// however, version 1 chunks made with texture tiler mode do have preceeding '\'  
			if( !xTileToAdd.m_xNameOfTextureUsed.StartsWith( "\\" ) )
			{
				*pxTileNameItr = "\\";
			}
			else
			{
				*pxTileNameItr = "";
			}


			*pxTileNameItr += xTileToAdd.m_xNameOfTextureUsed;
			++pxTileNameItr;
		}
		#endif
		
		xTileToAdd.m_xNameOfTextureUsed = xNameOfParentTexture; // this is legacy information, made obsolete by storing texture hash in each tile - NDB 181007
		
#ifdef ASURA_TOOLS
		// log name of asura file from which this tile has been read 
		Asura_Handle iTileHandle = Asura_TiledTextureManagement::AddTile( xTileToAdd );
		if( iAsuraFileNameIndex != Asura_TiledTextureManagement::ASURA_TILEDTEXTURE_UNKNOWN_FILE )
		{
			Asura_TiledTextureManagement::LogAsuraFileName( iTileHandle,iAsuraFileNameIndex );
		}
#else
		Asura_TiledTextureManagement::AddTile( xTileToAdd );
#endif
	}

	return true;
}

// read version 1 tiled texture chunk
const Asura_TiledTextureInfo& operator>>( Asura_Chunk_Stream& xStream,Asura_TiledTextureInfo& xOut )
{
	xStream >> xOut.m_fu1;
	xStream >> xOut.m_fv1;
	xStream >> xOut.m_fu2;
	xStream >> xOut.m_fv2;
	xStream >> xOut.m_TiledToWidth; 
	xStream >> xOut.m_TiledToHeight; 

	return xOut;
}

// AJG - 07/07/04 - Removed ASURA_TOOLS requirement as Texture_Tiler needs this, but also doesn't want ASURA_TOOLS defined as that
//					opens up needing everything relating to the hashed filename system included in the project.
//					TODO - maybe find another way to decide whether to include this/whether to define ASURA_ENABLE_HASH_TO_FILENAME_LOOKUPS.

#ifdef ASURA_TOOLS
void Asura_Chunk_TiledTexture::Create(const Asura_String& xTiledTextureName, int iNTiles, 
									  const Asura_TiledTextureInfo* pUVInfoList, Asura_Chunk_Stream& xStream)
{
	// Write the data...
	xStream.WriteChunk_Start(ASURA_CHUNK_TILEDTEXTURE, ASURA_CHUNK_TILEDTEXTURE_VERSION, 0);

	xStream << iNTiles;
	xStream << xTiledTextureName;
	xStream << static_cast<u_int>( Asura_GetHashID_IgnoreExtension( xTiledTextureName.GetString() ));

	const char* pcTileHashString(NULL);

	for(int i=0; i<iNTiles; ++i)
	{
		xStream << pUVInfoList[i].m_fu1;
		xStream << pUVInfoList[i].m_fv1;
		xStream << pUVInfoList[i].m_fu2;
		xStream << pUVInfoList[i].m_fv2;
		xStream << pUVInfoList[i].m_TiledToWidth;
		xStream << pUVInfoList[i].m_TiledToHeight;
		xStream << pUVInfoList[i].m_xNameOfTextureUsed;

		// calcualte tile hash id
		pcTileHashString = Tools_TT_StripPath( pUVInfoList[i].m_xNameOfTextureUsed );
		xStream << static_cast<u_int>( Asura_GetHashID( pcTileHashString ) ); // create hash id from texture file name
	}

	xStream.WriteChunk_End();
}

void Asura_Chunk_TiledTexture::Create(const Asura_String& xTiledTextureName, int iNTiles, 
									  const Asura_TiledTextureInfo* pUVInfoList, Asura_SaveBlock& xSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(&xSaveBlock);

	// Call the stream-version...
	Create( xTiledTextureName,iNTiles,pUVInfoList,xStream );
}

const char* Tools_TT_StripPath( const Asura_String& xPath )
{
	const char* pcPositionOfLastSlash( strrchr( xPath.GetString(),'\\' ) );
	
	if( !pcPositionOfLastSlash )
	{
		// if slash not found then return pointer to begining of string
		pcPositionOfLastSlash = xPath.GetString();	
	}
	else
	{
		// if slash is found then return pointer to first character after the last '\' character
		pcPositionOfLastSlash += 1;
	}

	return pcPositionOfLastSlash;
}


#endif


/// End of file. ///
