//////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_WeatherSystem.cpp
//
// Copyright (c) 2002-2005 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_Chunk_WeatherSystem.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Colour.h"
#include "Shadows/Asura_CSM.h"
#include "SpecialFX/Asura_Glow_System.h"
#include "SpecialFX/Asura_Weather_System.h"

// T Y P E S /////////////////////////////////////////////////////////////////

struct OldVersionExtraHeaderData
{
    u_int uNumShelters;
    u_int auPad[5];
};

// C O N S T A N T S /////////////////////////////////////////////////////////

//
// Version history:
//
//     (RDM - 07.12.2002) -  0 - Initial version.
//
//     (RDM - 08.12.2002) -  1 - Added fog and additional rain data.
//
//     (RDM - 12.12.2002) -  2 - Added IDs to rain shelter data.
//
//     (MP  - 21.05.2003) -  3 - Add sun glare data.
//
//     (RAR - 15.07.2003) -  4 - Added group number to rainshelters
//
//     (AW  - 21.01.2004) -  5 - Changed fog parameters to use PS2 palettized fog.
//
//     (AW  - 04.02.2004) -  6 - Reused redundant parameter m_fFogColourRed as m_fFogLevelForSkybox.
//
//     (MP  - 12.05.2004) -  7 - Removed some fog-related data and replaced
//                               some with padding.
//
//     (RJF - 15.09.2004) -  8 - Added glow input color data.
//
//     (MH  - 22.10.2004) -  9 - Added horizon glare data
//
//     (MP  - 24.06.2005) - 10 - Removed rain-related data.
//
//     (AAM - 06.01.2009) - 11 - Added cascade shadow map data.
//
//     (AAM - 09.02.2009) - 12 - Added CSM split planes
//
//     (AAM - 13.02.2009) - 13 - Set strength per-split instead of globally
//
static const int iCHUNK_VERSION = 13;

#ifdef ASURA_TOOLS
static const u_int uPALETTISED_FOG_DATA_SIZE    = 48;
static const u_int uRAIN_SHELTER_SIZE_VERSION_1     = 28;
static const u_int uRAIN_SHELTER_SIZE_VERSION_3     = 32;
#endif
static const u_int uRAIN_SHELTER_SIZE_VERSION_OTHER = 36;


// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

bool Asura_Chunk_WeatherSystem::Process ( Asura_Handle iFileHandle ) const
{
    // delete existing weather system storage
    Asura_Weather_System::Reset();
    
#ifdef ASURA_TOOLS
    Asura_Assert( (Version!=0), "Unsupported Chunk version!" );
#else
    Asura_Assert( Version >= 9, "Chunk version is no longer supported outside of tools. Please re-export." );
#endif

    Asura_Assert( Version <= iCHUNK_VERSION, "Chunk version is newer than current build supports. You need new code." );

    Asura_Chunk_File_Stream xStream( iFileHandle );

    // Skip old header and rain shelter data if required.
    if( Version <= 9 )
    {
        OldVersionExtraHeaderData xExtraHdrData;
        xStream.Read( sizeof( OldVersionExtraHeaderData ), &xExtraHdrData );

        // Skip any rain shelter data.
        if( xExtraHdrData.uNumShelters )
        {
#ifdef ASURA_TOOLS
            if( 1 == Version )
            {
                xStream.Skip( xExtraHdrData.uNumShelters * uRAIN_SHELTER_SIZE_VERSION_1 );
            }
            else if( Version <= 3 )
            {
                xStream.Skip( xExtraHdrData.uNumShelters * uRAIN_SHELTER_SIZE_VERSION_3 );
            }
            else
#endif
            {
                xStream.Skip( xExtraHdrData.uNumShelters * uRAIN_SHELTER_SIZE_VERSION_OTHER );
            }
        }
    }

    // Sun glare data.
#ifdef ASURA_TOOLS
    if( Version < 3 )
    {
        // No data available so just set default values.
        Asura_Weather_System::s_uSunFlags   = 0;
        Asura_Weather_System::s_xSunPos     = Asura_Maths::ZeroVector3;
        Asura_Weather_System::s_fSunColourR = 0.25f;
        Asura_Weather_System::s_fSunColourG = 0.1875f;
        Asura_Weather_System::s_fSunColourB = 0.15625f;
    }
    else
#endif
    {
        xStream >> Asura_Weather_System::s_uSunFlags; // This isn't a bitfield
        xStream >> Asura_Weather_System::s_xSunPos;
        xStream >> Asura_Weather_System::s_fSunColourR;
        xStream >> Asura_Weather_System::s_fSunColourG;
        xStream >> Asura_Weather_System::s_fSunColourB;
    }

    // Skip old palettised fog data if required.
#ifdef ASURA_TOOLS
    if( Version == 5 || Version == 6 )
    {
        xStream.Skip( uPALETTISED_FOG_DATA_SIZE );
    }
#endif

    // Glow parameters.
#ifdef ASURA_TOOLS
    if( Version < 8 )
    {
        // No data available so just set default values.
        Asura_Glow_System::SetGlowTint( 1.0f, 1.0f, 1.0f );
        Asura_Glow_System::SetGlowLevel( 0.605f );
    }
    else
#endif
    {
        u_int uColour;
        Asura_RGBA xColour;
        float fLevel;

        xStream >> uColour;
        xStream >> fLevel;

        Asura_Glow_System::SetGlowTint(xColour.FromUIntARGB(uColour));
        Asura_Glow_System::SetGlowLevel(fLevel);
    }

    // Horizon glare parameters.
#ifdef ASURA_TOOLS
    if( Version < 9 )
    {
        // No data available so just set default values.
        Asura_Weather_System::s_fHorizonGlareSizeDown = 0.0f;
        Asura_Weather_System::s_fHorizonGlareSizeUp = 0.25f;
    }
    else
#endif
    {
        xStream >> Asura_Weather_System::s_fHorizonGlareSizeDown;
        xStream >> Asura_Weather_System::s_fHorizonGlareSizeUp;
    }

	// Cascaded shadow map parameters.
	if( Version >= 11 )
	{
		Asura_Vector_3 xShadowDir;
		bool bEnabled;

		xStream >> xShadowDir;

		if( Version >= 13 )
		{
			// One strength per map

			u_int uNumStrengths;
			xStream >> uNumStrengths;

			float* const pfStrengths = new float[ uNumStrengths ];
			for( u_int u = 0; u < uNumStrengths; ++u )
			{
				xStream >> pfStrengths[u];
			}

			#ifdef ASURA_USE_CSM
				const u_int uMaxMaps = ASURA_CSM_MAX_MAPS;
				const u_int uMin = Asura_Maths::Min( uNumStrengths, uMaxMaps );

				u_int u = 0;
				for( u = 0; u < uMin; ++u )
				{
					Asura_CSM::SetStrength( u, pfStrengths[u] );
				}

				// Set any remaining ones. Will happen if MAX_MAPS increased after chunk saved
				for( ; u < ASURA_CSM_MAX_MAPS; ++u )
				{
					Asura_CSM::SetStrength( u, 1.0f );
				}
			#endif

			delete[] pfStrengths;
		}
		else
		{
			// Just one strength value

			float fStrength;
			xStream >> fStrength;

			#ifdef ASURA_USE_CSM
				for( u_int u = 0; u < ASURA_CSM_MAX_MAPS; ++u )
				{
					Asura_CSM::SetStrength( u, fStrength );
				}
			#endif
		}

		xStream >> bEnabled;

#ifdef ASURA_USE_CSM
		Asura_CSM::SetShadowDir( xShadowDir );
		Asura_CSM::SetEnabled( bEnabled );
#endif
	}

	// CSM splits
	if( Version >= 12 )
	{
		u_int uNumEntries;
		xStream >> uNumEntries;

		float* const pfSplits = new float[ uNumEntries ];
		for( u_int u = 0; u < uNumEntries; ++u )
		{
			xStream >> pfSplits[u];
		}

#ifdef ASURA_USE_CSM
		const u_int uMaxMapsPlus1 = static_cast< u_int >( ASURA_CSM_MAX_MAPS+1 );
		const u_int uMin = Asura_Maths::Min( uNumEntries, uMaxMapsPlus1 );

		u_int u;
		for( u = 0; u < uMin; ++u )
		{
			Asura_CSM::SetSplit( u, pfSplits[u] );
		}

		// Set any remaining ones. Will happen if MAX_MAPS has been increased after chunk saved
		for( ; u <= ASURA_CSM_MAX_MAPS; ++u )
		{
			Asura_CSM::SetSplit( u, Asura_Camera::GetFarClip() );
		}
#endif

		delete[] pfSplits;
	}

    return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_WeatherSystem::Create( Asura_SaveBlock_Base* pxSaveBlock )
{
    Asura_Chunk_SaveBlock_Stream xStream( pxSaveBlock );

    xStream.WriteChunk_Start( ASURA_CHUNK_WEATHERSYSTEM, iCHUNK_VERSION, 0 );

    xStream << Asura_Weather_System::s_uSunFlags; // This isn't a bitfield
    xStream << Asura_Weather_System::s_xSunPos;
    xStream << Asura_Weather_System::s_fSunColourR;
    xStream << Asura_Weather_System::s_fSunColourG;
    xStream << Asura_Weather_System::s_fSunColourB;

    // RJF - 15 September 2004 - Glow input values
    u_int uColour = Asura_Glow_System::GetGlowTint().ToUIntARGB();
    float fLevel = Asura_Glow_System::GetGlowLevel();
    xStream << uColour;
    xStream << fLevel;

    // (MH - 22.10.2004) horizon glare values
    xStream << Asura_Weather_System::s_fHorizonGlareSizeDown;
    xStream << Asura_Weather_System::s_fHorizonGlareSizeUp;

	// (AAM - 06.01.2009) cascaded shadow map values
#ifdef ASURA_USE_CSM

	xStream << Asura_CSM::GetShadowDir();

	const u_int uNumStrengths = ASURA_CSM_MAX_MAPS;
	xStream << uNumStrengths;

	for( u_int u = 0; u < uNumStrengths; ++u )
	{
		xStream << Asura_CSM::GetStrength(u);
	}

	xStream << Asura_CSM::IsEnabled();

	const u_int uNumEntries = ASURA_CSM_MAX_MAPS+1;
	xStream << uNumEntries;

	for( u_int u = 0; u < uNumEntries; ++u )
	{
		xStream << Asura_CSM::GetSplit(u);
	}

#else

	xStream << Asura_Vector_3( 0.7f, 0.7f, 0.7f ); // Direction
	xStream << u_int(0);                           // Num strengths
	xStream << false;                              // Is enabled
	xStream << u_int(0);                           // Num splits

#endif

    xStream.WriteChunk_End();
}
#endif // #ifdef ASURA_TOOLS
