//////////////////////////////////////////////////////////////////////////////
//
// Asura_SaveBlock.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_SaveBlock.h"

// Stdlib independent.
#include <stdio.h>
#include <string.h>

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Huffman.h"
#ifdef ASURA_SAVEBLOCK_USE_ZLIB
#include "ChunkSystem/Asura_Zlib.h"
#endif
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"

// Tools
#if defined(ASURA_TOOLS) && defined(ASURA_WRITE_ZLIB)
#include "Load&Save/Tools_Zlib.h"
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Asura_SaveBlock_Base.
//////////////////////////////////////////////////////////////////////////

bool Asura_SaveBlock_Base::AddAsuraTerminator( void )
{
	int aiZeros[4] = { 0, 0, 0, 0 };
	return AddData( aiZeros, sizeof(aiZeros) );
}


u_int Asura_SaveBlock_Base::GetSpaceUsed( void ) const 
{ 
	return GetBytesWritten(); 
}

//////////////////////////////////////////////////////////////////////////
// Asura_ProtoSaveBlock.
//////////////////////////////////////////////////////////////////////////

Asura_ProtoSaveBlock::Asura_ProtoSaveBlock()
: m_pcStart(NULL)
, m_pcCurrentPos(NULL)
, m_pcCurrentReadPos(NULL)
, m_uSize(0)
, m_uSpaceLeft(0)
, m_uSpaceUsed(0)
, m_uBytesRead(0)
, m_uBytesLeftToRead(0)
{
}

bool Asura_ProtoSaveBlock::AddData( const void* pSrc, size_t uSize )
{
#ifdef ASURA_SAVEBLOCK_ALLOWACCESSPOINTER
    // Note that this method copes with a zero size but we'll still
    // assert about it because it's unlikely to be intentional.
    Asura_Assert( uSize > 0, "Invalid data size." );

    void* pDest = GetAccessPointer( uSize );
    if( !pDest ) { return false; }

	Asura_Mem::Copy( pDest, pSrc, uSize );
    return true;
#else
	Asura_Assert(false, "");
	return false;
#endif
}

// writes 4 bytes at the specified position.
bool Asura_ProtoSaveBlock::AddPatchData( const void* pData, u_int uNumberOfBytes, u_int uPos )
{
	Asura_Assert(uNumberOfBytes == 4, "only expecting this to be used for 4-byte values");
	// Could handle this if required (by adding appropriate amount of data), but easiest just to assert at the moment.
	Asura_Assert(reinterpret_cast<const char*>(GetCurrentPosition()) >= (m_pcStart + uPos + uNumberOfBytes),
		"Cannot insert here as it will over-run the buffer");
	Asura_Mem::Copy(reinterpret_cast<void*>(m_pcStart + uPos), pData, uNumberOfBytes);
	return true;
}

size_t Asura_ProtoSaveBlock::ReadData( void* pDest, size_t uSize ) const
{
	Asura_Assert(m_pcCurrentReadPos, "didn't set read pos in save block");
	size_t uBytesToRead = Asura_Maths::Min( uSize, m_uBytesLeftToRead );
	Asura_Mem::Copy( pDest, m_pcCurrentReadPos, uBytesToRead );
	m_pcCurrentReadPos += uBytesToRead;
	m_uBytesRead += uBytesToRead;
	m_uBytesLeftToRead -= uBytesToRead;
	return uBytesToRead;
}

size_t Asura_ProtoSaveBlock::SkipData( size_t uSize ) const
{
	Asura_Assert(m_pcCurrentReadPos, "didn't set read pos in save block");
	size_t uBytesToSkip = Asura_Maths::Min( uSize, m_uBytesLeftToRead );
	m_pcCurrentReadPos += uBytesToSkip;
	m_uBytesRead += uBytesToSkip;
	m_uBytesLeftToRead -= uBytesToSkip;
	return uBytesToSkip;
}

// virtual function - read counter
size_t Asura_ProtoSaveBlock::GetBytesRead() const
{
	return m_uBytesRead;
}

// virtual function - write counter
size_t Asura_ProtoSaveBlock::GetBytesWritten() const
{
	return reinterpret_cast<const char*>(GetCurrentPosition()) - GetStartPtr();
}

void Asura_ProtoSaveBlock::RestartRead() const
{
	m_pcCurrentReadPos = m_pcStart;
	m_uBytesRead = 0;
	m_uBytesLeftToRead = m_uSpaceUsed;
}

void Asura_ProtoSaveBlock::Clear( bool bWriteFileID )
{
    m_pcCurrentPos = m_pcStart;
    m_uSpaceLeft   = m_uSize;
    m_uSpaceUsed   = 0;

    if( bWriteFileID )
    {
        AddData( ASURA_UNCOMPRESSED_FILE_ID, ASURA_FILE_ID_NUM_CHARS );
    }
}

#if defined(ASURA_TOOLS) || !defined(ASURA_MAIN)
bool Asura_ProtoSaveBlock::SaveFile( const char* szFilename, bool bUseCompression ) const
{
	#if defined(ASURA_TOOLS) && defined(ASURA_WRITE_ZLIB)
	if(bUseCompression)
	{
		return Tools_Zlib::SaveCompressedFile(GetStartPtr(), GetSize(), szFilename, false);
	}
	#endif
    
    FILE* pxFile = fopen( szFilename, "wb" );
    if( !pxFile ) { return false; }

    const char* pcBfr = GetStartPtr();
    if( !pcBfr ) { return false; }

    // KJL 08/08/00 16:33:15 - save out, with compression if required.
    if( !bUseCompression )
    {
        fwrite( pcBfr, GetSize(), 1, pxFile );
    }
    else
    {
        Asura_Huffman_CompressedFileHdr* pxHdr;
        u_int uTotalSize;

        if( !Asura_Huffman_Encoder::AllocateAndCompressForASR(pcBfr, GetSize(), &pxHdr, &uTotalSize) )
        {
            return false;
        }

        fwrite( pxHdr, uTotalSize, 1, pxFile );

        Asura_Deallocate( pxHdr );
    }

    return( !fclose(pxFile) );
}
#endif


#if defined(ASURA_TOOLS) && defined(ASURA_MAIN)
int Asura_ProtoSaveBlock::WriteToFile( HANDLE hFile )
{
	u_int uDataSize = GetSize() - ASURA_FILE_ID_NUM_CHARS;
	DWORD uBytesWritten;
	if(!WriteFile( hFile, GetStartPtr() + ASURA_FILE_ID_NUM_CHARS, uDataSize, &uBytesWritten, NULL ))
	{
		Asura_Assert(0, "Did not write to file successfully.");
		return -1;
	}
	if(uBytesWritten != uDataSize)
	{
		Asura_Assert(0, "Wrote incorrect number of bytes.");
		return -1;
	}
	return uBytesWritten;
}
#endif

void Asura_ProtoSaveBlock::Truncate(u_int uLength)
{
	if(uLength > m_uSpaceUsed)
	{
		Asura_Assert(0, "Can't truncate to larger size.");
		return;
	}
	const u_int uDiff = m_uSpaceUsed - uLength;
	m_uSpaceUsed -= uDiff;
	m_uSpaceLeft += uDiff;
	m_pcCurrentPos -= uDiff;
}

//
// Asura_SaveBlock.
//

Asura_SaveBlock::Asura_SaveBlock( void )
{
    m_pcStart		= NULL;
    m_pcCurrentPos	= NULL;
    m_uSize			= 0;
    m_uSpaceLeft	= 0;
    m_uSpaceUsed	= 0;
	m_uSmallestSize	= 128 * 1024;

    AddData( ASURA_UNCOMPRESSED_FILE_ID, ASURA_FILE_ID_NUM_CHARS );
}
Asura_SaveBlock::Asura_SaveBlock( const u_int uSmallestSize )
{
	m_pcStart		= NULL;
	m_pcCurrentPos	= NULL;
	m_uSize			= 0;
	m_uSpaceLeft	= 0;
	m_uSpaceUsed	= 0;
	m_uSmallestSize	= uSmallestSize;

	AddData( ASURA_UNCOMPRESSED_FILE_ID, ASURA_FILE_ID_NUM_CHARS );
}

Asura_SaveBlock::~Asura_SaveBlock( void )
{
	// how many reallocations did we need
	//Asura_Assert( m_uSize <= ( m_uSmallestSize * 64 ), "Asura_SaveBlock::GetAccessPointer, needs larger default size" );

	delete[] m_pcStart;
}

#ifdef ASURA_SAVEBLOCK_ALLOWACCESSPOINTER
void* Asura_SaveBlock::GetAccessPointer( int iNumBytesRequired )
{
    // Note that this method copes with a zero size but we'll still
    // assert about it because it's unlikely to be intentional.
    Asura_Assert( iNumBytesRequired > 0, "Invalid data size." );
    if( iNumBytesRequired < 0 ) { return NULL; }

    // NRM - Make sure that it's aligned to 4 byte boundaries
//  iNumBytesRequired = ((iNumBytesRequired + 3) / 4) * 4;

    //see if we need to enlarge the buffer to allow for this block
    if( static_cast<u_int>(iNumBytesRequired) > m_uSpaceLeft )
    {
        //need more space
        u_int uSpaceNeeded = m_uSpaceUsed + iNumBytesRequired;

        //// (MH - 13.11.2007) changed way buffer is resized - now larger sizes
		//// use larger allocations to avoid the resizing being a bottleneck.
		//u_int uBlockSize;
		//const u_int uMegabyte = 1024*1024;
		//if (uSpaceNeeded < uMegabyte)
		//{
		//	// up to 1MB increase in 100000 byte chunks (old behaviour)
		//	uBlockSize = 100000;
		//}
		//else if (uSpaceNeeded < 10 * uMegabyte)
		//{
		//	// up to 10MB increase in 1MB chunks
		//	uBlockSize = uMegabyte;
		//}
		//else
		//{
		//	// allocating 10MB+, increase in 10MB blocks
		//	uBlockSize = 10 * uMegabyte;
		//}
		//u_int uNewSize = ((uSpaceNeeded / uBlockSize) + 1) * uBlockSize;


		// (PAZ - 20.02.2008) try doubling the currently allocated size - to hopefully reduce memory fragmentation and reallocations
		u_int uNextPowerTwo = 1;
		while( ( uNextPowerTwo < uSpaceNeeded ) && ( uNextPowerTwo != 0 ) )
		{
			uNextPowerTwo *= 2;
		}
		u_int uNewSize = uNextPowerTwo;
		while( ( uNewSize < m_uSmallestSize ) && ( uNewSize != 0 ) )
		{
			uNewSize *= 2;
		}
		if( uNewSize < uSpaceNeeded )
		{
			// must be > 2GB?
			Asura_Assert( false, "More than 2GB allocated - will need to update Asura_SaveBlock to use size_t." );
			uNewSize = uSpaceNeeded;
		}

        //allocate the memory
        // (MP - 15.05.2003) Changed this to use new and delete
        // instead of realloc. Our memory manager doesn't track
        // reallocs so we miss leaks. Also, realloc would probably
        // fail to extend the block in most cases anyway.
        char* pcNewStart = new char[uNewSize];
        if( !pcNewStart ) { return NULL; }
        
        if( m_pcStart )
        {
            Asura_Mem::Copy( pcNewStart, m_pcStart, m_uSpaceUsed );
            delete[] m_pcStart;
        }

        //correct the position etc.
        m_pcStart = pcNewStart;
        m_pcCurrentPos = m_pcStart + m_uSpaceUsed;
        m_uSpaceLeft += (uNewSize - m_uSize);
        m_uSize = uNewSize;
    }
    
    void* pcRet = m_pcCurrentPos;

    m_uSpaceUsed   += iNumBytesRequired;
    m_uSpaceLeft   -= iNumBytesRequired;
    m_pcCurrentPos += iNumBytesRequired;

    return pcRet;
}
#endif // #ifdef ASURA_SAVEBLOCK_ALLOWACCESSPOINTER

//
// Asura_StaticSaveBlock.
//

Asura_StaticSaveBlock::Asura_StaticSaveBlock( int iBufferSize )
{
    m_pcStart      = (char*)Asura_AllocateAlign(iBufferSize,ASURA_ALIGN_16_BYTE);
    m_pcCurrentPos = m_pcStart;
    m_uSize        = iBufferSize;
    m_uSpaceLeft   = iBufferSize;
    m_uSpaceUsed   = 0;

    AddData( ASURA_UNCOMPRESSED_FILE_ID, ASURA_FILE_ID_NUM_CHARS ); 			
}

Asura_StaticSaveBlock::~Asura_StaticSaveBlock()
{
	Asura_DeallocateAlign(m_pcStart);
	m_pcStart = NULL; // note this will stop the Asura_SaveBlock destructor from trying to free the buffer again
}

#ifdef ASURA_SAVEBLOCK_ALLOWACCESSPOINTER
void* Asura_StaticSaveBlock::GetAccessPointer( int iNumBytesRequired )
{
    // Note that this method copes with a zero size but we'll still
    // assert about it because it's unlikely to be intentional.
    Asura_Assert( iNumBytesRequired > 0, "Invalid data size." );
    if( iNumBytesRequired < 0 ) { return NULL; }

	// DLH - 16 April 2007 12:08:46 - Commented out the asserts as streaming produces allowed non-aligned data which fires the asserts. JakeT game permission to do this change.
#if 0
    // Some platforms get upset if we try and address word-aligned data that's
    // not on a word boundary, eg. the PS2. Chunk stuff is generally all word-
    // aligned anyway so let's just require that.
	if (4 <= iNumBytesRequired)
	{
	    Asura_Assert( (0 == (m_uSpaceUsed & 3)), "Large block requests should be 4-byte aligned within the save block." );
	}

	else if (2 == iNumBytesRequired)
	{
	    Asura_Assert( (0 == (m_uSpaceUsed & 1)), "2 byte block requests should be 2-byte aligned within the save block." );
	}
#endif

    //see if we've run out of space
    if( static_cast<u_int>(iNumBytesRequired) > m_uSpaceLeft )
    {
		Asura_Assert( false, "No more free space in save block!" );
		return NULL;
    }
    
    void* pcRet = m_pcCurrentPos;

    m_uSpaceUsed   += iNumBytesRequired;
    m_uSpaceLeft   -= iNumBytesRequired;
    m_pcCurrentPos += iNumBytesRequired;

    return pcRet;
}
#endif // #ifdef ASURA_SAVEBLOCK_ALLOWACCESSPOINTER


// virtual override of AddData() method
bool Asura_StaticSaveBlock::AddData( const void* pSrc, size_t uSize )
{
	// Note that this method copes with a zero size but we'll still
	// assert about it because it's unlikely to be intentional.
	Asura_Assert( uSize > 0, "Invalid data size." );
	if( uSize > m_uSpaceLeft )
	{
		Asura_Assert( false, "No more free space in save block!" );
		return false;
	}

	Asura_Mem::Copy( m_pcCurrentPos, pSrc, uSize );

	m_uSpaceUsed   += uSize;
	m_uSpaceLeft   -= uSize;
	m_pcCurrentPos += uSize;

	return true;
}


//////////////////////////////////////////////////////////////////////////
// Asura_CompressedSaveBlock
//////////////////////////////////////////////////////////////////////////
const u_int uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE = 32768;
const u_int uMAX_COMPRESSION_RATIO = 10;
#ifdef ASURA_SAVEBLOCK_USE_ZLIB
u_int Asura_CompressedSaveBlock::s_uZLibWindowBits = 10;
u_int Asura_CompressedSaveBlock::s_uZlibMemLevel = 4;
#endif


Asura_CompressedSaveBlock::Asura_CompressedSaveBlock( u_int uCompressedSize )
{
	m_bOwnsData = true;
	m_pcCompressedData = new char[uCompressedSize];
	m_pcUncompressedBuffer = new char[uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE];
	m_uCompressedSize = uCompressedSize;
	// estimate max patches that will be needed. This should be no more than the number of uncompressed
	// blocks used, so assume a high compression ratio (10:1) to guess the max uncompressed size.
	u_int uMaxPatches = uMAX_COMPRESSION_RATIO * uCompressedSize / uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE;
	m_pxPatches = new SaveBlockPatch[uMaxPatches];
	m_uMaxPatches = uMaxPatches;
	m_uPatchesUsed = 0;

	m_uWritePosInUncompressedBuffer = 0;
	m_uWritePosInCompressedBuffer = 0;
	m_uBytesCompressed = 0;

	m_uReadPosInUncompressedBuffer = 0;
	m_uReadPosInCompressedBuffer = 0;
	m_uBytesDecompressed = 0;
	m_uNextPatch = 0;

	m_uBytesWritten = 0;
	m_uBytesRead = 0;

#ifdef ASURA_SAVEBLOCK_USE_ZLIB
	m_xZlibStream.zalloc = Z_NULL;
	m_xZlibStream.zfree = Z_NULL;
	m_xZlibStream.opaque = Z_NULL;
#endif
}

Asura_CompressedSaveBlock::Asura_CompressedSaveBlock()
{
	m_bOwnsData = false;
//	m_pcCompressedData = new char[uCompressedSize];
	m_pcUncompressedBuffer = new char[uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE];
	m_uCompressedSize = 0;
	// estimate max patches that will be needed. This should be no more than the number of uncompressed
	// blocks used, so assume a high compression ratio (10:1) to guess the max uncompressed size.

	m_uWritePosInUncompressedBuffer = 0;
	m_uWritePosInCompressedBuffer = 0;
	m_uBytesCompressed = 0;

	m_uReadPosInUncompressedBuffer = 0;
	m_uReadPosInCompressedBuffer = 0;
	m_uBytesDecompressed = 0;
	m_uNextPatch = 0;

	m_uBytesWritten = 0;
	m_uBytesRead = 0;

#ifdef ASURA_SAVEBLOCK_USE_ZLIB
	m_xZlibStream.zalloc = Z_NULL;
	m_xZlibStream.zfree = Z_NULL;
	m_xZlibStream.opaque = Z_NULL;
#endif
}

void Asura_CompressedSaveBlock::SetCompressedDataPointer( char* pxDataBuffer, u_int uSize )
{
	Asura_Assert(m_uCompressedSize == 0, "Attempting to alter the compressedsaveblock");
	m_pcCompressedData = pxDataBuffer;
	m_uCompressedSize = uSize;

	u_int uMaxPatches = uMAX_COMPRESSION_RATIO * uSize / uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE;
	m_pxPatches = new SaveBlockPatch[uMaxPatches];
	m_uMaxPatches = uMaxPatches;
	m_uPatchesUsed = 0;

}
void Asura_CompressedSaveBlock::UnsetCompressedDataPointer()
{
	m_pcCompressedData = 0x00;
	m_uCompressedSize = 0;
	m_uMaxPatches = 0;
	m_uPatchesUsed = 0;
	delete[] m_pxPatches;

}
Asura_CompressedSaveBlock::~Asura_CompressedSaveBlock()
{
	delete[] m_pxPatches;
	delete[] m_pcUncompressedBuffer;

	if( m_bOwnsData )
	{
		delete[] m_pcCompressedData;
	}
}

// private functions for manipulating data - may be inlined which is why they're above all other functions
void Asura_CompressedSaveBlock::WriteDataToUncompressedBuffer( const void* pSrc, size_t uSize )
{
	Asura_Mem::Copy( m_pcUncompressedBuffer + m_uWritePosInUncompressedBuffer, pSrc, uSize );
	m_uWritePosInUncompressedBuffer += uSize;
	m_uBytesWritten += uSize;
	Asura_Assert(m_uWritePosInUncompressedBuffer <= uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE, "buffer overrun");
}

void Asura_CompressedSaveBlock::ReadDataFromUncompressedBuffer( void* pDest, size_t uSize ) const
{
	Asura_Mem::Copy( pDest, m_pcUncompressedBuffer + m_uReadPosInUncompressedBuffer, uSize );
	m_uReadPosInUncompressedBuffer += uSize;
	m_uBytesRead += uSize;
	Asura_Assert(m_uReadPosInUncompressedBuffer <= uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE, "buffer overrun");
}

void Asura_CompressedSaveBlock::SkipDataInUncompressedBuffer( size_t uSize ) const
{
	m_uReadPosInUncompressedBuffer += uSize;
	m_uBytesRead += uSize;
	Asura_Assert(m_uReadPosInUncompressedBuffer <= uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE, "buffer overrun");
}

bool Asura_CompressedSaveBlock::AddData( const void* pSrc, size_t uSize )
{
	u_int uBytesToWrite = uSize;
	const char* pcSource = reinterpret_cast<const char*>(pSrc);

	while (uBytesToWrite)
	{
		u_int uBytesAvailable = uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE - m_uWritePosInUncompressedBuffer;
		if (uBytesAvailable >= uBytesToWrite)
		{
			// can fit all the data in the uncompressed buffer
			WriteDataToUncompressedBuffer(pcSource, uBytesToWrite);
			return true;
		}
		else
		{
			// fill up the uncompressed buffer
			WriteDataToUncompressedBuffer( pcSource, uBytesAvailable );

			// update input parameters
			pcSource += uBytesAvailable;
			uBytesToWrite -= uBytesAvailable;

			// compress, move & clear uncompressed buffer
			if (!Flush())
			{
				return false;
			}
			Asura_Assert(uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE - m_uWritePosInUncompressedBuffer, "no bytes available after Flush()");
		}
	}

	return true;
}

// writes 4 bytes at the specified position.
bool Asura_CompressedSaveBlock::AddPatchData( const void* pData, u_int uNumberOfBytes, u_int uPos )
{
	Asura_Assert( uPos + uNumberOfBytes <= m_uBytesWritten, "patching data that hasn't been written yet");
	Asura_Assert(uNumberOfBytes == 4, "only expecting this to be used for 4-byte values");

	// for any data still in the uncompressed buffer, just change the data.
	// Otherwise we need to store a patch.
	if (uPos >= m_uBytesCompressed)
	{
		u_int uOffset = uPos - m_uBytesCompressed;
		Asura_Assert(uOffset < uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE, "buffer overrun");
		Asura_Mem::Copy(m_pcUncompressedBuffer + uOffset, pData, uNumberOfBytes);
	}
	else
	{
		// get a new patch to use
		if (m_uPatchesUsed >= m_uMaxPatches)
		{
			Asura_Assert(false, "ran out of patches for compressed saveblock. increase uMAX_COMPRESSION_RATIO");
			return false;
		}
		SaveBlockPatch& xPatch = m_pxPatches[m_uPatchesUsed++];

		// check how many bytes go in the compressed buffer - these are stored in the patch
		u_int uBytesInPatch = Asura_Maths::Min( uNumberOfBytes, m_uBytesCompressed - uPos );
		xPatch.m_uPos = uPos;
		xPatch.m_uNumBytes = uBytesInPatch;
		Asura_Mem::Copy(xPatch.m_acValue, pData, uBytesInPatch);
		
		// put any remainder at the start of the uncompressed buffer
		uNumberOfBytes -= uBytesInPatch;
		if (uNumberOfBytes)
		{
			pData = reinterpret_cast<const char*>(pData) + uBytesInPatch;
			Asura_Mem::Copy(m_pcUncompressedBuffer, pData, uNumberOfBytes);
		}
	}
	return true;
}

size_t Asura_CompressedSaveBlock::ReadData( void* pDest, size_t uSize ) const
{
	u_int uBytesRead = 0;
	while (uSize)
	{
		Asura_Assert(m_uBytesRead <= m_uBytesWritten, "read more bytes than were written");
		// check if there are enough bytes in the uncompressed buffer to satisfy the request (and clamp to the valid bytes in the save block)
		u_int uBytesAvailable = Asura_Maths::Min(uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE - m_uReadPosInUncompressedBuffer, m_uBytesWritten - m_uBytesRead);
		if (uBytesAvailable >= uSize)
		{
			ReadDataFromUncompressedBuffer(pDest, uSize);
			uBytesRead += uSize;
			return uBytesRead;
		}
		else
		{
			// read what's in the uncompressed buffer
			ReadDataFromUncompressedBuffer(pDest, uBytesAvailable);
			uBytesRead += uBytesAvailable;

			// update input parameters
			pDest = reinterpret_cast<char*>(pDest) + uBytesAvailable;
			uSize -= uBytesAvailable;

			// refill the uncompressed buffer
			Decompress();
		}
	}
	return uBytesRead;
}

size_t Asura_CompressedSaveBlock::SkipData( size_t uSize ) const
{
	u_int uBytesRead = 0;
	while (uSize)
	{
		Asura_Assert(m_uBytesRead <= m_uBytesWritten, "read more bytes than were written");
		// check if there are enough bytes in the uncompressed buffer to satisfy the request (and clamp to the valid bytes in the save block)
		u_int uBytesAvailable = Asura_Maths::Min(uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE - m_uReadPosInUncompressedBuffer, m_uBytesWritten - m_uBytesRead);
		if (uBytesAvailable >= uSize)
		{
			SkipDataInUncompressedBuffer(uSize);
			uBytesRead += uSize;
			return uBytesRead;
		}
		else
		{
			// read what's in the uncompressed buffer
			SkipDataInUncompressedBuffer(uBytesAvailable);
			uBytesRead += uBytesAvailable;

			// update input parameters
			uSize -= uBytesAvailable;

			// refill the uncompressed buffer
			Decompress();
		}
	}
	return uBytesRead;
}

// virtual function - read counter
size_t Asura_CompressedSaveBlock::GetBytesRead() const
{
	return m_uBytesRead;
}

// virtual function - write counter
size_t Asura_CompressedSaveBlock::GetBytesWritten() const
{
	return m_uBytesWritten;
}

void Asura_CompressedSaveBlock::RestartRead() const
{
	Asura_Assert(m_uWritePosInUncompressedBuffer == 0, "data remains in uncompressed buffer - Flush() was not called");

	m_uReadPosInUncompressedBuffer = 0;
	m_uReadPosInCompressedBuffer = 0;
	m_uBytesDecompressed = 0;
	m_uNextPatch = 0;

	m_uBytesRead = 0;

#ifdef ASURA_SAVEBLOCK_USE_ZLIB
	BeginZlibDecompression();
#endif

	// prime the decompressed buffer
	Decompress();
}

void Asura_CompressedSaveBlock::Clear( bool bWriteFileID )
{
	m_uPatchesUsed = 0;

	m_uWritePosInUncompressedBuffer = 0;
	m_uWritePosInCompressedBuffer = 0;
	m_uBytesCompressed = 0;

	m_uReadPosInUncompressedBuffer = 0;
	m_uReadPosInCompressedBuffer = 0;
	m_uBytesDecompressed = 0;
	m_uNextPatch = 0;

	m_uBytesWritten = 0;
	m_uBytesRead = 0;

#ifdef ASURA_SAVEBLOCK_USE_ZLIB
	BeginZlibCompression();
#endif
	if( bWriteFileID )
	{
		AddData( ASURA_UNCOMPRESSED_FILE_ID, ASURA_FILE_ID_NUM_CHARS );
	}
}

// override of AddAsuraTerminator() to force a Flush() at the end of the block.
bool Asura_CompressedSaveBlock::AddAsuraTerminator( void )
{
	bool bOK = Asura_SaveBlock_Base::AddAsuraTerminator();
	if (bOK)
	{
		bOK = Flush();
	}
#ifdef ASURA_SAVEBLOCK_USE_ZLIB
	EndZlibCompression();
#endif
	return bOK;
}

DEBUGVAR bool dbg_bUncompressed = false;
// Flush() compresses and moves the data from the uncompressed buffer into the compressed buffer
bool Asura_CompressedSaveBlock::Flush( void )
{
	Asura_Assert(m_uWritePosInUncompressedBuffer <= uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE, "buffer overrun");
	bool bOK = true;

	char* pcData = m_pcUncompressedBuffer;
	u_int uDataSize = m_uWritePosInUncompressedBuffer;
	char* pcBuffer = m_pcCompressedData + m_uWritePosInCompressedBuffer;
	Asura_Assert(m_uCompressedSize > m_uWritePosInCompressedBuffer, "Compressed buffer overrun");
	u_int uBufferSize = m_uCompressedSize - m_uWritePosInCompressedBuffer;

	if(dbg_bUncompressed)
	{
		// to test the code - uncompressed version
		Asura_Mem::Copy( pcBuffer, pcData, uDataSize);
		m_uWritePosInCompressedBuffer += uDataSize;
	}
	else
	{
#ifdef ASURA_SAVEBLOCK_USE_ZLIB
		// zlib compressed version
		m_xZlibStream.next_in = (Bytef*)pcData;
		m_xZlibStream.avail_in = uDataSize;

		const int iRet = deflate(&m_xZlibStream, Z_NO_FLUSH);
		if(iRet != Z_OK)
		{
			if(!m_xZlibStream.avail_out)
			{
				Asura_Assert(0, "Error compressing saveblock: out of space");
			}
			else if(m_xZlibStream.msg)
			{
				Asura_Assertf(0, ("Error compressing saveblock: %s", m_xZlibStream.msg));
			}
			else
			{
				Asura_Assert(0, "Unknown error compressing saveblock");
			}
			return false;
		}
#else//ASURA_SAVEBLOCK_USE_ZLIB
		// huffman-compressed version
		Asura_Chunk_Memory_Stream xMemoryStream(uBufferSize, pcBuffer);
#ifdef ASURA_TOOLS
		// prevent assert in tools.
		xMemoryStream.SetSwapEndian( false );
#endif
		u_int uCompressedSize;

		bOK = Asura_Huffman_Encoder::CompressToBfr( pcData, uDataSize, &uCompressedSize, xMemoryStream );

		// compressed size seems to include the optional asura file header - remove that here.
		m_uWritePosInCompressedBuffer += (uCompressedSize - ASURA_FILE_ID_NUM_CHARS);
		Asura_Assert(m_uWritePosInCompressedBuffer <= m_uCompressedSize, "Over-ran compressed buffer");
#endif
	}

	// update the write variables appropriately
	m_uBytesCompressed += m_uWritePosInUncompressedBuffer;
	m_uWritePosInUncompressedBuffer = 0;

	return bOK;
}

// Decompress() - function to fill up the uncompressed buffer from the compressed one while reading
void Asura_CompressedSaveBlock::Decompress() const
{
	Asura_Assert(m_uBytesDecompressed < m_uBytesCompressed, "shouldn't have decompressed more data than we compressed");
	u_int uDecompressedBytes = Asura_Maths::Min(m_uBytesCompressed - m_uBytesDecompressed, uASURA_COMPRESSEDSAVEBLOCK_BUFFERSIZE);

	const char* pcData = m_pcCompressedData + m_uReadPosInCompressedBuffer;
	char* pcBuffer = m_pcUncompressedBuffer;
	u_int uBufferSize = uDecompressedBytes;
	if(dbg_bUncompressed)
	{
		// to test the code - uncompressed version
		Asura_Mem::Copy( pcBuffer, pcData, uBufferSize );
		m_uReadPosInCompressedBuffer += uDecompressedBytes;
	}
	else
	{
#ifdef ASURA_SAVEBLOCK_USE_ZLIB
		m_xZlibStream.avail_out = uBufferSize;
		m_xZlibStream.next_out = (Bytef*)pcBuffer;

		const int iRet = inflate(&m_xZlibStream, Z_NO_FLUSH);
		Asura_Assert(m_xZlibStream.total_out - m_uBytesDecompressed == uDecompressedBytes, "Decompressed unexpected number of bytes");
		
		if(iRet == Z_STREAM_END)
		{
			EndZlibDecompression();
		}
		else if(iRet != Z_OK)
		{
			Asura_Assert(0,"Error decompressing");
			return;
		}
#else
		// huffman-compressed version
		Asura_Huffman_Decoder xDecoder;
		const Asura_Huffman_Package* pxPackage = reinterpret_cast<const Asura_Huffman_Package*>(pcData);
		xDecoder.Begin(pxPackage);
#ifdef ASURA_ASSERT
		u_int uBytesFromHuffman = 
#endif
			xDecoder.Decompress(pcBuffer, uBufferSize);
		Asura_Assert(uBytesFromHuffman == uBufferSize, "decoded unexpected number of bytes");
		xDecoder.End();
		// move pointer along in the compressed buffer.
		m_uReadPosInCompressedBuffer += (sizeof(Asura_Huffman_Package) + pxPackage->m_uCompressedDataSize);
#endif
	}

	// update read variables appropriately
	u_int uPrevBytesDecompressed = m_uBytesDecompressed;
	m_uReadPosInUncompressedBuffer = 0;
	m_uBytesDecompressed = uPrevBytesDecompressed + uDecompressedBytes;

	// need to apply patch?
	if (m_uNextPatch < m_uPatchesUsed)
	{
		const SaveBlockPatch& xPatch = m_pxPatches[m_uNextPatch];
		u_int uPatchPos = xPatch.m_uPos;
		if (uPrevBytesDecompressed <= uPatchPos && uPatchPos < m_uBytesDecompressed)
		{
			// yes, apply it here
			u_int uOffset = uPatchPos - uPrevBytesDecompressed;
			Asura_Mem::Copy(m_pcUncompressedBuffer + uOffset, xPatch.m_acValue, xPatch.m_uNumBytes);

			// done with this patch
			++m_uNextPatch;
		}
	}
}

#ifdef ASURA_SAVEBLOCK_USE_ZLIB

bool Asura_CompressedSaveBlock::BeginZlibCompression()
{
	//initialise the compression stream
	const int iRet = deflateInit2(&m_xZlibStream, Z_DEFAULT_COMPRESSION, Z_DEFLATED, s_uZLibWindowBits, s_uZlibMemLevel, Z_DEFAULT_STRATEGY);
	if(iRet != Z_OK)
	{
		Asura_Assert(0, "Error opening zlib stream");
		return false;
	}

	m_xZlibStream.avail_out = m_uCompressedSize;
	m_xZlibStream.next_out = (Bytef*)m_pcCompressedData;

	return true;
}

bool Asura_CompressedSaveBlock::BeginZlibDecompression() const
{
	//initialise the compression stream
	const int iRet = inflateInit2(&m_xZlibStream, s_uZLibWindowBits);
	if(iRet != Z_OK)
	{
		Asura_Assert(0, "Error opening zlib stream");
		return false;
	}

	m_xZlibStream.avail_in = m_uWritePosInCompressedBuffer;
	m_xZlibStream.next_in = (Bytef*)m_pcCompressedData;

	return true;
}

bool Asura_CompressedSaveBlock::EndZlibCompression()
{
	const int iRet = deflate(&m_xZlibStream, Z_FINISH);
	if(iRet != Z_STREAM_END)
	{
		Asura_Assert(0, "Error finishing zlib stream");
		return false;
	}
	m_uWritePosInCompressedBuffer = m_xZlibStream.total_out;
	const int iEndRet = deflateEnd(&m_xZlibStream);
	if(iEndRet != Z_OK)
	{
		Asura_Assert(0, "Error closing zlib stream");
		return false;
	}
	return true;
}

bool Asura_CompressedSaveBlock::EndZlibDecompression() const
{
	m_uReadPosInCompressedBuffer = m_xZlibStream.total_in;
	const int iEndRet = inflateEnd(&m_xZlibStream);
	if(iEndRet != Z_OK)
	{
		Asura_Assert(0, "Error closing zlib stream");
		return false;
	}
	return true;
}

void Asura_CompressedSaveBlock::SetMemLevel(u_int uMemLevel)
{
	Asura_Assert(1 <= uMemLevel && uMemLevel <= 9, "Mem level out of range");
	s_uZlibMemLevel = Asura_Maths::ClampToRange(uMemLevel, 1u, 9u);
}

void Asura_CompressedSaveBlock::SetWindowBits(u_int uWindowBits)
{
	Asura_Assert(8 <= uWindowBits && uWindowBits <= 15, "Window bits out of range");
	s_uZLibWindowBits = Asura_Maths::ClampToRange(uWindowBits, 8u, 15u);
}

#endif//ASURA_SAVEBLOCK_USE_ZLIB


// eof
