/**
*
* \file Asura_AABB_Tree.cpp
*
* Created by Kevin Lea - 21/06/2002 14:41:25
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "Core/Asura_AABB_Tree.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Macros.h"
#endif
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"

static const float g_fCompressionScale = 128.0f;
static const float g_fInvCompressionScale = 1.0f/128.0f;

Asura_AABB_Tree::Asura_AABB_Tree()
{
	Nodes = 0;
}

Asura_AABB_Tree::~Asura_AABB_Tree()
{
	Asura_ResourcePushType(ASURA_RESOURCE_AABB_TREE);
	{
		Asura_ResourceDeallocateArray(Nodes);
	}
	Asura_ResourcePopType();
}

void Asura_AABB_Tree::TestTree( const Asura_Bounding_Box& xBB, bool(*pfnOutputCallback)(u_short usIdx) )
{
	OutputCallback = pfnOutputCallback;
	TestBB         = &xBB;

	if (!Nodes)
	{
		// KJL - 16/05/2002 17:33:22 -  if there is only one object, the node test will
		// always return an index of zero. Though I must say it's a complete waste of
		// time and space to use this data structure with such sparse information. I've
		// put this in to make sure all cases (e.g. only one module in the environment, a 
		// module with just one triangle) are catered for - even though one would hope 
		// such situations don't occur in a released game.
		if (TestBB->Intersects(RootBoundingBox))
		{
			OutputCallback(0);
		}
		return;
	}

	TestNode(0,this->RootBoundingBox);
}

void Asura_AABB_Tree::TestTree( const Asura_Bounding_Box& xBB, Asura_AABB_Tree_Searcher& xSearchClass ) const
{
	if ( Nodes )
	{
		TestNode( 0, RootBoundingBox, xBB, xSearchClass );
	}
	else
	{
		// KJL - 16/05/2002 17:33:22 -  if there is only one object, the node test will
		// always return an index of zero. Though I must say it's a complete waste of
		// time and space to use this data structure with such sparse information. I've
		// put this in to make sure all cases (e.g. only one module in the environment, a 
		// module with just one triangle) are catered for - even though one would hope 
		// such situations don't occur in a released game.
		if ( xBB.Intersects( RootBoundingBox ) )
		{
			xSearchClass.VisitNode( 0 );
		}
	}
}

void Asura_AABB_Tree::SetBoundingBoxes(const Asura_AABB_Tree_Node& xNode, const Asura_Bounding_Box& xParentBB, Asura_Bounding_Box& leftBB, Asura_Bounding_Box& rightBB)
{
	const float fParentBBWidth  = (xParentBB.MaxX-xParentBB.MinX) * g_fInvCompressionScale;
	const float fParentBBHeight = (xParentBB.MaxY-xParentBB.MinY) * g_fInvCompressionScale;
	const float fParentBBDepth  = (xParentBB.MaxZ-xParentBB.MinZ) * g_fInvCompressionScale;

	const float fParentBBMinX = xParentBB.MinX;
	const float bbMinX = fParentBBMinX + ((float)xNode.uChildMinX) * fParentBBWidth;
	if (xNode.MinXBelongsToLeft)
	{
		leftBB.MinX = bbMinX;
		rightBB.MinX = fParentBBMinX;
	}
	else
	{
		leftBB.MinX = fParentBBMinX;
		rightBB.MinX = bbMinX;
	}

	const float fParentBBMaxX = xParentBB.MaxX;
	const float bbMaxX = fParentBBMaxX - ((float)xNode.uChildMaxX) * fParentBBWidth;
	if (xNode.MaxXBelongsToLeft)
	{
		leftBB.MaxX = bbMaxX;
		rightBB.MaxX = fParentBBMaxX;
	}
	else
	{
		leftBB.MaxX = fParentBBMaxX;
		rightBB.MaxX = bbMaxX;
	}
	

	const float fParentBBMinY = xParentBB.MinY;
	const float bbMinY = fParentBBMinY + ((float)xNode.uChildMinY) * fParentBBHeight;
	if (xNode.MinYBelongsToLeft)
	{
		leftBB.MinY = bbMinY;
		rightBB.MinY = fParentBBMinY;
	}
	else
	{
		leftBB.MinY = fParentBBMinY;
		rightBB.MinY = bbMinY;
	}

	const float fParentBBMaxY = xParentBB.MaxY;
	const float bbMaxY = fParentBBMaxY - ((float)xNode.uChildMaxY) * fParentBBHeight;
	if (xNode.MaxYBelongsToLeft)
	{
		leftBB.MaxY = bbMaxY;
		rightBB.MaxY = fParentBBMaxY;
	}
	else
	{
		leftBB.MaxY = fParentBBMaxY;
		rightBB.MaxY = bbMaxY;
	}
	

	const float fParentBBMinZ = xParentBB.MinZ;
	const float bbMinZ = fParentBBMinZ + ((float)xNode.uChildMinZ) * fParentBBDepth;
	if (xNode.MinZBelongsToLeft)
	{
		leftBB.MinZ = bbMinZ;
		rightBB.MinZ = fParentBBMinZ;
	}
	else
	{
		leftBB.MinZ = fParentBBMinZ;
		rightBB.MinZ = bbMinZ;
	}

	const float fParentBBMaxZ = xParentBB.MaxZ;
	const float bbMaxZ = fParentBBMaxZ - ((float)xNode.uChildMaxZ) * fParentBBDepth;
	if (xNode.MaxZBelongsToLeft)
	{
		leftBB.MaxZ = bbMaxZ;
		rightBB.MaxZ = fParentBBMaxZ;
	}
	else
	{
		leftBB.MaxZ = fParentBBMaxZ;
		rightBB.MaxZ = bbMaxZ;
	}
}


// Returns true iff the search should continue.
bool Asura_AABB_Tree::TestNode( int nodeIndex, const Asura_Bounding_Box& xParentBB ) const 
{
	Asura_Bounding_Box leftBB;
	Asura_Bounding_Box rightBB;

	const Asura_AABB_Tree_Node& node = this->Nodes[nodeIndex];
	
	SetBoundingBoxes(node,xParentBB,leftBB,rightBB);

	/// Kevin Lea - 14/05/2002 11:03:41 - if left is leaf, output index if overlap occurs
	if (TestBB->Intersects(leftBB))
	{
		if (node.LeftIsLeaf)
		{
			/// Kevin Lea - 14/05/2002 11:05:36 - output
            if( !OutputCallback(node.LeftIndex) ) { return false; }
		}
		else 
		{
			/// Kevin Lea - 14/05/2002 10:58:52 - if test bounding box surrounds a node completely, 
			/// output all the contents of that node
			if(TestBB->Contains(leftBB))
			{	
				/// Kevin Lea - 14/05/2002 11:05:41 - output all
				if( !OutputAll(node.LeftIndex) ) { return false; }
			}
			else
			{
				/// Kevin Lea - 14/05/2002 11:06:38 - test further
				if( !TestNode(node.LeftIndex,leftBB) ) { return false; }
			}
		}
	}

	/// Kevin Lea - 14/05/2002 11:03:41 - if Right is leaf, output index if overlap occurs
	if (TestBB->Intersects(rightBB))
	{
		if (node.RightIsLeaf)
		{
			/// Kevin Lea - 14/05/2002 11:05:36 - output
            if( !OutputCallback(node.RightIndex) ) { return false; }
		}
		else
		{
			/// Kevin Lea - 14/05/2002 10:58:52 - if test bounding box surrounds a node completely, 
			/// output all the contents of that node
			if(TestBB->Contains(rightBB))
			{	
				/// Kevin Lea - 14/05/2002 11:05:41 - output all
				if( !OutputAll(node.RightIndex) ) { return false; }
			}
			else
			{
				/// Kevin Lea - 14/05/2002 11:06:38 - test further
				if( !TestNode(node.RightIndex,rightBB) ) { return false; }
			}
		}
	}

    // Keep searching.
    return true;
}

bool Asura_AABB_Tree::TestNode( int iNodeIndex, const Asura_Bounding_Box& xParentBB, const Asura_Bounding_Box& xTestBB, Asura_AABB_Tree_Searcher& xSearchClass ) const
{
	Asura_Bounding_Box xLeftBB;
	Asura_Bounding_Box xRightBB;

	const Asura_AABB_Tree_Node& xNode = Nodes[ iNodeIndex ];

	SetBoundingBoxes( xNode, xParentBB, xLeftBB, xRightBB );

	/// Kevin Lea - 14/05/2002 11:03:41 - if left is leaf, output index if overlap occurs
	if (xTestBB.Intersects(xLeftBB))
	{
		if (xNode.LeftIsLeaf)
		{
			/// Kevin Lea - 14/05/2002 11:05:36 - output
			if( !xSearchClass.VisitNode(xNode.LeftIndex) ) { return false; }
		}
		else 
		{
			/// Kevin Lea - 14/05/2002 10:58:52 - if test bounding box surrounds a node completely, 
			/// output all the contents of that node
			if(xTestBB.Contains(xLeftBB))
			{
				/// Kevin Lea - 14/05/2002 11:05:41 - output all
				if( !OutputAll(xNode.LeftIndex, xSearchClass) ) { return false; }
			}
			else
			{
				/// Kevin Lea - 14/05/2002 11:06:38 - test further
				if( !TestNode(xNode.LeftIndex, xLeftBB, xTestBB, xSearchClass) ) { return false; }
			}
		}
	}

	/// Kevin Lea - 14/05/2002 11:03:41 - if Right is leaf, output index if overlap occurs
	if (xTestBB.Intersects(xRightBB))
	{
		if (xNode.RightIsLeaf)
		{
			/// Kevin Lea - 14/05/2002 11:05:36 - output
			if( !xSearchClass.VisitNode(xNode.RightIndex) ) { return false; }
		}
		else
		{
			/// Kevin Lea - 14/05/2002 10:58:52 - if test bounding box surrounds a node completely, 
			/// output all the contents of that node
			if(xTestBB.Contains(xRightBB))
			{	
				/// Kevin Lea - 14/05/2002 11:05:41 - output all
				if( !OutputAll(xNode.RightIndex, xSearchClass) ) { return false; }
			}
			else
			{
				/// Kevin Lea - 14/05/2002 11:06:38 - test further
				if( !TestNode(xNode.RightIndex, xRightBB, xTestBB, xSearchClass) ) { return false; }
			}
		}
	}

	// Keep searching.
	return true;
}

void Asura_AABB_Tree::TestTreeWithRay( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, float* pfDistance, bool(*pfnRayTestCallback)( const u_int uIndex ) )
{
	m_pfMaxDistance = pfDistance;

	m_pxRayOrigin = &xRayOrigin;
	m_pxRayDirection = &xRayDirection;
	m_pfnRayTestCallback = pfnRayTestCallback;

	const float fEPSILON = 0.000001f;
	m_xOneOverRayDirection.x = ( Asura_Maths::Abs( xRayDirection.x ) < fEPSILON ) ? 0.0f : ( 1.0f / xRayDirection.x );
	m_xOneOverRayDirection.y = ( Asura_Maths::Abs( xRayDirection.y ) < fEPSILON ) ? 0.0f : ( 1.0f / xRayDirection.y );
	m_xOneOverRayDirection.z = ( Asura_Maths::Abs( xRayDirection.z ) < fEPSILON ) ? 0.0f : ( 1.0f / xRayDirection.z );

	if (!Nodes)
	{
		// KJL - 16/05/2002 17:33:22 -  if there is only one object, the node test will
		// always return an index of zero. Though I must say it's a complete waste of
		// time and space to use this data structure with such sparse information. I've
		// put this in to make sure all cases (e.g. only one module in the environment, a 
		// module with just one triangle) are catered for - even though one would hope 
		// such situations don't occur in a released game.
		float fDistance = *m_pfMaxDistance;
		if ( RayIntersectsAABB( *m_pxRayOrigin, m_xOneOverRayDirection, RootBoundingBox, fDistance ) )
		{
			m_pfnRayTestCallback( 0 );
		}
		return;
	}

	TestNodeWithRay( 0, this->RootBoundingBox );
}

// Returns true iff the search should continue.
bool Asura_AABB_Tree::TestNodeWithRay( const u_int uNodeIndex, const Asura_Bounding_Box& xParentBB )
{
	Asura_Bounding_Box xLeftBB;
	Asura_Bounding_Box xRightBB;
	
	const Asura_AABB_Tree_Node& xNode = this->Nodes[uNodeIndex];

	// reconstruct left and right bounding boxes
	SetBoundingBoxes(xNode,xParentBB,xLeftBB,xRightBB);

	float fLeftDistance = *m_pfMaxDistance;
	bool bLeftHit = RayIntersectsAABB( *m_pxRayOrigin, m_xOneOverRayDirection, xLeftBB, fLeftDistance );

	float fRightDistance = *m_pfMaxDistance;
	bool bRightHit = RayIntersectsAABB( *m_pxRayOrigin, m_xOneOverRayDirection, xRightBB, fRightDistance );
	
	bool bLeftFirst = false;
	if (bLeftHit && bRightHit)
	{
		if (fLeftDistance<fRightDistance)
		{
			bLeftFirst = true;
		}
	}

	if (bLeftHit && bLeftFirst)
	{
		// if left is leaf, output index if overlap occurs
		if (xNode.LeftIsLeaf)
		{
			// output
			if (!m_pfnRayTestCallback(xNode.LeftIndex))
			{
				return false;
			}
		}
		else 
		{
			// test further
			if (!TestNodeWithRay( xNode.LeftIndex, xLeftBB ))
			{
				return false;
			}
		}
	}

	// If the distance has been improved upon re-evaluate if we need to process the right hit.
	if (bRightHit && (*m_pfMaxDistance < fRightDistance)) bRightHit = false;

	if (bRightHit)
	{
		// if right is leaf, output index if overlap occurs
		if (xNode.RightIsLeaf)
		{
			// output
			if (!m_pfnRayTestCallback(xNode.RightIndex))
			{
				return false;
			}
		}
		else
		{
			// test further
			if (!TestNodeWithRay( xNode.RightIndex, xRightBB ))
			{
				return false;
			}
		}
	}

	// If the distance has been improved upon re-evaluate if we need to process the left hit.
	if (bLeftHit && (*m_pfMaxDistance < fLeftDistance)) bLeftHit = false;

	if (bLeftHit && !bLeftFirst)
	{
		// if left is leaf, output index if overlap occurs
		if (xNode.LeftIsLeaf)
		{
			// output
			if (!m_pfnRayTestCallback(xNode.LeftIndex))
			{
				return false;
			}
		}
		else 
		{
			// test further
			if (!TestNodeWithRay( xNode.LeftIndex, xLeftBB ))
			{
				return false;
			}
		}
	}

	return true;
}


// Returns true iff the search should continue.
bool Asura_AABB_Tree::OutputAll(int nodeIndex) const
{
	Asura_AABB_Tree_Node& node = this->Nodes[nodeIndex];

	// KJL - 16/05/2002 16:14:11 - merely recurse through and output every leaf
	if (node.LeftIsLeaf)
	{
        if( !OutputCallback(node.LeftIndex) ) { return false; }
	}
	else
	{
        if( !OutputAll(node.LeftIndex) ) { return false; }
	}

	if (node.RightIsLeaf)
	{
        if( !OutputCallback(node.RightIndex) ) { return false; }
	}
	else
	{
        if( !OutputAll(node.RightIndex) ) { return false; }
	}

    // Keep searching.
    return true;
}

bool Asura_AABB_Tree::OutputAll( int iNodeIndex, Asura_AABB_Tree_Searcher& xSearchClass ) const
{
	const Asura_AABB_Tree_Node& xNode = Nodes[ iNodeIndex ];

	// KJL - 16/05/2002 16:14:11 - merely recurse through and output every leaf
	if (xNode.LeftIsLeaf)
	{
		if( !xSearchClass.VisitNode(xNode.LeftIndex) ) { return false; }
	}
	else
	{
		if( !OutputAll(xNode.LeftIndex, xSearchClass) ) { return false; }
	}

	if (xNode.RightIsLeaf)
	{
		if( !xSearchClass.VisitNode(xNode.RightIndex) ) { return false; }
	}
	else
	{
		if( !OutputAll(xNode.RightIndex, xSearchClass) ) { return false; }
	}

	// Keep searching.
	return true;
}

void Asura_AABB_Tree::Traverse( bool ( *pfnBBCallback )( const Asura_Bounding_Box& xBB ) )
{
    m_pfnBBCallback = pfnBBCallback;

    if ( !Nodes )
    {
        m_pfnBBCallback( RootBoundingBox );
    }
    else
    {
        OutputAll( 0, RootBoundingBox );
    }
}

bool Asura_AABB_Tree::OutputAll(int iNodeIndex, const Asura_Bounding_Box& xParentBB) const
{
    const Asura_AABB_Tree_Node& xNode = Nodes[ iNodeIndex ];

    // reconstruct left and right bounding boxes
    Asura_Bounding_Box xLeftBB;
    Asura_Bounding_Box xRightBB;
    SetBoundingBoxes( xNode, xParentBB, xLeftBB, xRightBB );

    // KJL - 16/05/2002 16:14:11 - merely recurse through and output every leaf
    if (xNode.LeftIsLeaf)
    {
        if( !m_pfnBBCallback( xLeftBB ) ) { return false; }
    }
    else
    {
        if( !OutputAll( xNode.LeftIndex, xLeftBB ) ) { return false; }
    }

    if (xNode.RightIsLeaf)
    {
        if( !m_pfnBBCallback( xRightBB ) ) { return false; }
    }
    else
    {
        if( !OutputAll( xNode.RightIndex, xRightBB ) ) { return false; }
    }

    // Keep searching.
    return true;
}

// RJF - 05 October 2004 - Common save/load code
// Schema versions are included to allow the tree format to change in future
// Schema version:				1	Original serialized format

static const int CURRENT_AABB_SCHEMA_VERSION = 1;

Asura_AABB_Tree *Asura_AABB_Tree::LoadTree(int schemaVersion, Asura_Handle hFileHandle, int iNodeCount)
{
	if(schemaVersion > CURRENT_AABB_SCHEMA_VERSION)
	{
		Asura_Assert(false, "Trying to load an AABB tree using a schema that is too new. You probably need a new build.");
		return 0;
	}

	Asura_Assert(iNodeCount, "Trying to create a tree with no nodes!");
	if(!iNodeCount) return 0;

	Asura_AABB_Tree *tree = new Asura_AABB_Tree();
	Asura_Assert(tree, "Could not create new AABB tree!");

	Asura_FileAccess::ReadData(hFileHandle, &tree->RootBoundingBox, sizeof(Asura_Bounding_Box));

	// Read in iNodeCount nodes
	if(iNodeCount==1)
	{
		// Tree with only a root node - still valid
		return tree;
	}

	Asura_ResourcePushType(ASURA_RESOURCE_AABB_TREE);
	{
		Asura_ResourceAllocateArray(tree->Nodes, iNodeCount-1);
		Asura_FileAccess::ReadData(hFileHandle, tree->Nodes, sizeof(Asura_AABB_Tree_Node) * (iNodeCount-1));
	}
	Asura_ResourcePopType();
	return tree;
}

#ifdef ASURA_TOOLS
int Asura_AABB_Tree::GetTreeSize(int schemaVersion, const Asura_AABB_Tree *pxTree, int iNodeCount)
{
	if(schemaVersion > CURRENT_AABB_SCHEMA_VERSION)
	{
		Asura_Assert(false, "Trying to size an AABB tree using a schema that is too new. You probably need a new build.");
		return 0;
	}

	Asura_Assert(pxTree, "Null tree passed to GetTreeSize - you should have checked for a null tree already.");
	Asura_Assert((iNodeCount == 1) || (pxTree->Nodes), "Claiming a >1 number of nodes when pxTree->Nodes is NULL...");

	return sizeof(Asura_Bounding_Box) + sizeof(Asura_AABB_Tree_Node) * (iNodeCount-1);
}

int Asura_AABB_Tree::SaveTree(int schemaVersion, const Asura_AABB_Tree *pxTree, int iNodeCount, char *pxDest)
{
	if(schemaVersion > CURRENT_AABB_SCHEMA_VERSION)
	{
		Asura_Assert(false, "Trying to save an AABB tree using a schema that is too new. You probably need a new build.");
		return 0;
	}

	char *pxStart = pxDest;

	Asura_Assert(pxTree, "Null tree passed to GetTreeSize - you should have checked for a null tree already.");
	Asura_Assert((iNodeCount == 1) || (pxTree->Nodes), "Claiming a >1 number of nodes when pxTree->Nodes is NULL...");

	// bounding box
	Asura_Mem::Copy(pxDest, &pxTree->RootBoundingBox, sizeof(Asura_Bounding_Box));
	ASURA_CHUNK_ARRAY_VERIFY(*pxDest, float, 6);
	pxDest += sizeof(Asura_Bounding_Box);

	// nodes
	if(!pxTree->Nodes)return pxDest - pxStart; // only one node

	for ( int i = 0; i < iNodeCount-1; ++i)
	{
		Asura_AABB_Tree_Node xTempNode = pxTree->Nodes[i];

		ASURA_CHUNK_VERIFY(xTempNode.uChildMinX, u_char);
		ASURA_CHUNK_VERIFY(xTempNode.uChildMinY, u_char);
		ASURA_CHUNK_VERIFY(xTempNode.uChildMinZ, u_char);
		ASURA_CHUNK_VERIFY(xTempNode.uChildMaxX, u_char);
		ASURA_CHUNK_VERIFY(xTempNode.uChildMaxY, u_char);
		ASURA_CHUNK_VERIFY(xTempNode.uChildMaxZ, u_char);

		ASURA_CHUNK_VERIFY(xTempNode.LeftIndex, unsigned short);
		ASURA_CHUNK_VERIFY(xTempNode.RightIndex, unsigned short);
		ASURA_CHUNK_BITFIELD_VERIFY(xTempNode.m_ucFlags, unsigned char);
		Asura_Mem::Copy(pxDest, &xTempNode, sizeof(Asura_AABB_Tree_Node));
		pxDest += sizeof(Asura_AABB_Tree_Node);
	}
	return pxDest - pxStart;
}

void Asura_AABB_Tree::SaveTree(int schemaVersion, const Asura_AABB_Tree *pxTree, int iNodeCount, Asura_Chunk_Stream& xStream)
{
	if(schemaVersion > CURRENT_AABB_SCHEMA_VERSION)
	{
		Asura_Assert(false, "Trying to save an AABB tree using a schema that is too new. You probably need a new build.");
	}

	else
	{
		Asura_Assert(pxTree, "Null tree passed to GetTreeSize - you should have checked for a null tree already.");
		Asura_Assert((iNodeCount == 1) || (pxTree->Nodes), "Claiming a >1 number of nodes when pxTree->Nodes is NULL...");

		// bounding box
		xStream << pxTree->RootBoundingBox;

		// nodes
		for ( int i = 0; i < iNodeCount-1; ++i)
		{
			pxTree->Nodes [i].WriteToChunkStream(xStream);
		}
	}
}
#endif

// KJL - 12/05/2002 16:03:43 - Tree creator - only really needs to be called in pre-processing step
unsigned short*				Asura_AABB_Tree_Creator::OldObjectSet;
unsigned short*				Asura_AABB_Tree_Creator::NewObjectSet;
unsigned char*				Asura_AABB_Tree_Creator::ToRightBitfield;
const Asura_Vector_3*		Asura_AABB_Tree_Creator::LeafCentre;
const Asura_Bounding_Box*	Asura_AABB_Tree_Creator::LeafAABB;
Asura_AABB_Tree*			Asura_AABB_Tree_Creator::Tree;
unsigned short				Asura_AABB_Tree_Creator::NodeIndex;

Asura_AABB_Tree* Asura_AABB_Tree_Creator::Create( int iNumObjs, bool(*pfnPopulateArraysCallback)(u_int, Asura_Bounding_Box*, Asura_Vector_3*) )
{
	Asura_ResourcePushType(ASURA_RESOURCE_AABB_TREE);
	{
		Asura_Assert( iNumObjs > 0 && iNumObjs < 65536, "invalid number of objects" );

		// Allocate these bits first, since they're all we'll keep around. This
		// will help avoid memory fragmentation.
		Tree        = new Asura_AABB_Tree;
		if(iNumObjs>1)
		{
			Asura_ResourceAllocateArray(Tree->Nodes, iNumObjs - 1);
		}

		// Allocate and fill temporary BB and object centre arrays.
		Asura_Bounding_Box* pxBBs     = new Asura_Bounding_Box[iNumObjs];
		Asura_Vector_3*     pxCentres = new Asura_Vector_3[iNumObjs];
		bool bFilledInCentres = pfnPopulateArraysCallback( iNumObjs, pxBBs, pxCentres );

		// Calculate the centres ourselves if the callback didn't bother.
		if( !bFilledInCentres )
		{
			for( int i = 0; i < iNumObjs; ++i )
			{
				pxCentres[i] = pxBBs[i].CalculateCentre();
			}
		}

		// Store ptrs for recursion.
		LeafAABB   = pxBBs;
		LeafCentre = pxCentres;

		// KJL - 16/05/2002 17:33:22 -  if there is only one object, the node test will
		// always return an index of zero. Though I must say it's a complete waste of
		// time and space to use this data structure with such sparse information. I've
		// put this in to make sure all cases (e.g. only one module in the environment, a 
		// module with just one triangle) are catered for - even though one would hope 
		// such situations don't occur in a released game.
		if( 1 == iNumObjs )
		{
			Tree->RootBoundingBox = LeafAABB[0];
			// KJL - 16/05/2002 17:38:04 - Nodes is set to NULL by the constructor, we
			// can use that to test for this case later
		}
		else
		{
			Tree->RootBoundingBox.Reset();

			for( int i = 0; i < iNumObjs; ++i )
			{
				Tree->RootBoundingBox.Set(LeafAABB[i]);
			}

			OldObjectSet = new u_short[iNumObjs];
			NewObjectSet = new u_short[iNumObjs];
			ToRightBitfield = new unsigned char[iNumObjs];
			int start = 0;
			int end = iNumObjs;

			for(int i=start; i<end; i++)
			{
				OldObjectSet[i] = (unsigned short)i;
			}

			NodeIndex = 0;
			SortObjectsIntoNodes(start,end,Tree->RootBoundingBox,Tree->RootBoundingBox);

			delete[] ToRightBitfield;
			delete[] NewObjectSet;
			delete[] OldObjectSet;
		}
		delete[] pxCentres;
		delete[] pxBBs;
	}
	Asura_ResourcePopType();
	return Tree;
}


void Asura_AABB_Tree_Creator::SortObjectsIntoNodes(int start, int end, const Asura_Bounding_Box& xParentBB, const Asura_Bounding_Box& xEstimatedBB)
{
	Asura_AABB_Tree_Node& node = (Tree->Nodes[NodeIndex]);

	// KJL - 12/05/2002 17:05:44 - we want to split the objects into two nodes, which we'll call
	// 'left' and 'right'. Ideally we'd like a pretty even distribution between the two nodes,
	// and the bounding volumes of the 'left' and 'right' sets to overlap as little as possible.
	// It may be that a better splitting heuristic would give noticable speed gains in game-time
	// code.
	//
	// Okay, I'm going to go through each axis of the xParentBB in turn, calculating a midpoint
	// value to sort the objects' centroids against. (Intuitively one would expect the longest axis
	// to give the best results, but there may be cases like long thin objects that counter this.)
	// I'll then pick an axis based on the most even split of objects (e.g. minimum sum of the
	// squares of the objects in each node), weighted by the sum of the nodes' volumes. Obviously
	// we could split the parent volume in any way, but this should be a fast way of getting
	// reasonable results.

	/*
		Changed so that the left/right assignments for each axis are stored in the temporary
		ToRightBitfield array. This is to avoid problems with the PC release version sometimes 
		getting different results the second time round , once the axis has been chosen.
		RAR 12/10/2005
	*/

	float score[3];
	int objectsToTheRight[3];
	int objectsToTheLeft[3];
	Asura_Bounding_Box leftBB[3],rightBB[3];
	
	for( int i = 0; i < 3; ++i )
	{
		unsigned char ucAxisBit = (unsigned char)(1<<i);
		
		
		float midpoint;
		switch(i)
		{
			default:
			case 0:
			{
				midpoint = (xParentBB.MaxX+xParentBB.MinX)*0.5f;
				break;
			}
			case 1:
			{
				midpoint = (xParentBB.MaxY+xParentBB.MinY)*0.5f;
				break;
			}
			case 2:
			{
				midpoint = (xParentBB.MaxZ+xParentBB.MinZ)*0.5f;
				break;
			}
		}

		objectsToTheRight[i] = 0;
		objectsToTheLeft[i] = 0;

		leftBB[i].Reset();
		rightBB[i].Reset();

		for( int j = start; j < end; ++j )
		{
			int objectIndex = OldObjectSet[j];
			if (LeafCentre[objectIndex].GetElement(i) > midpoint)
			{
				// KJL - 12/05/2002 16:19:44 - in 'right' node
				++objectsToTheRight[i];
				rightBB[i].Set(LeafAABB[objectIndex]);
				//Store the side for later use. RAR 12/10/2005
				ToRightBitfield[j] |= ucAxisBit;

			}
			else
			{
				// KJL - 12/05/2002 16:19:57 - in 'left' node
				++objectsToTheLeft[i];
				leftBB[i].Set(LeafAABB[objectIndex]);
				//Store the side for later use. RAR 12/10/2005
				ToRightBitfield[j] &=~ ucAxisBit;
	
			}
		}

		float leftVolume=0;
		float rightVolume=0;

		if (objectsToTheRight[i])
		{
			rightVolume = rightBB[i].GetVolume();
		}
		if (objectsToTheLeft[i])
		{
			leftVolume = leftBB[i].GetVolume();
		}

		score[i] = (float)((float)objectsToTheLeft[i]*(float)objectsToTheLeft[i]+(float)objectsToTheRight[i]*(float)objectsToTheRight[i])*100.0f
				 + leftVolume + rightVolume;
	}

	// KJL - 12/05/2002 17:46:29 - pick the best axis based on the calculated scores
	int axis = 2;
	if (score[0]<score[1])
	{
		if (score[0]<score[2])
		{
			axis = 0;
		}
	}
	else
	{
		if (score[1]<score[2])
		{
			axis = 1;
		}
	}

	if (objectsToTheLeft[axis] == 0 || objectsToTheRight[axis] == 0)
	{
		// KJL - 12/05/2002 17:51:18 - after all that, we've ended up with an empty node...
		// let's take the first object and arbitrarily put that in the left node on its
		// own to make a leaf node, and put the rest in the right node. (In this situation
		// I expect that we're down to a small number of objects anyway.)
	
		for (int i=start; i<end; i++)
		{
			NewObjectSet[i] = OldObjectSet[i];
		}
		
		objectsToTheLeft[axis] = 1;
		leftBB[axis].Reset();
		leftBB[axis].Set(LeafAABB[NewObjectSet[start]]);
	
		objectsToTheRight[axis] = end-start-1;
		rightBB[axis].Reset();
		for (int i=start+1; i<end; i++)
		{
			unsigned short objectIndex = NewObjectSet[i];
			rightBB[axis].Set(LeafAABB[objectIndex]);
		}
	}
	else
	{
		// KJL - 12/05/2002 17:52:13 - okay, let's sort the objects into two sets
		unsigned char ucAxisBit = (unsigned char)(1<<axis);

		int right = end;
		int left = start;

		for (int i=start; i<end; i++)
		{
			unsigned short objectIndex = OldObjectSet[i];
			if (ToRightBitfield[i] & ucAxisBit)
			{
				// KJL - 12/05/2002 16:19:44 - in 'right' node
				--right;
				NewObjectSet[right] = objectIndex;
			}
			else
			{
				// KJL - 12/05/2002 16:19:57 - in 'left' node
				NewObjectSet[left] = objectIndex;
				++left;
			}
		}
	}
	for (int i=start; i<end; i++)
	{
		OldObjectSet[i] = NewObjectSet[i];
	}
	

	Asura_Assert(leftBB[axis].MaxX <= xParentBB.MaxX + 0.00001f, "");
	Asura_Assert(leftBB[axis].MaxY <= xParentBB.MaxY + 0.00001f, "");
	Asura_Assert(leftBB[axis].MaxZ <= xParentBB.MaxZ + 0.00001f, "");
	Asura_Assert(leftBB[axis].MinX >= xParentBB.MinX - 0.00001f, "");
	Asura_Assert(leftBB[axis].MinY >= xParentBB.MinY - 0.00001f, "");
	Asura_Assert(leftBB[axis].MinZ >= xParentBB.MinZ - 0.00001f, "");

	Asura_Assert(rightBB[axis].MaxX <= xParentBB.MaxX + 0.00001f, "");
	Asura_Assert(rightBB[axis].MaxY <= xParentBB.MaxY + 0.00001f, "");
	Asura_Assert(rightBB[axis].MaxZ <= xParentBB.MaxZ + 0.00001f, "");
	Asura_Assert(rightBB[axis].MinX >= xParentBB.MinX - 0.00001f, "");
	Asura_Assert(rightBB[axis].MinY >= xParentBB.MinY - 0.00001f, "");
	Asura_Assert(rightBB[axis].MinZ >= xParentBB.MinZ - 0.00001f, "");

	float fLengthX = xEstimatedBB.MaxX - xEstimatedBB.MinX;
	float fLengthY = xEstimatedBB.MaxY - xEstimatedBB.MinY;
	float fLengthZ = xEstimatedBB.MaxZ - xEstimatedBB.MinZ;

	if (fLengthX<=0.0f) fLengthX = 1e10f;
	if (fLengthY<=0.0f) fLengthY = 1e10f;
	if (fLengthZ<=0.0f) fLengthZ = 1e10f;

	if (leftBB[axis].MinX != xParentBB.MinX)
	{
		node.MinXBelongsToLeft = true;
		node.uChildMinX = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (leftBB[axis].MinX - xEstimatedBB.MinX) / (fLengthX)) );
	}
	else
	{
		node.MinXBelongsToLeft = false;
		node.uChildMinX = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (rightBB[axis].MinX - xEstimatedBB.MinX) / (fLengthX)) );
	}

	if (leftBB[axis].MaxX != xParentBB.MaxX)
	{
		node.MaxXBelongsToLeft = true;
		node.uChildMaxX = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (xEstimatedBB.MaxX - leftBB[axis].MaxX) / (fLengthX)) );
	}
	else
	{
		node.MaxXBelongsToLeft = false;
		node.uChildMaxX = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (xEstimatedBB.MaxX - rightBB[axis].MaxX) / (fLengthX)) );
	}
	
	if (leftBB[axis].MinY != xParentBB.MinY)
	{
		node.MinYBelongsToLeft = true;
		node.uChildMinY = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (leftBB[axis].MinY - xEstimatedBB.MinY) / (fLengthY)) );
	}
	else
	{
		node.MinYBelongsToLeft = false;
		node.uChildMinY = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (rightBB[axis].MinY - xEstimatedBB.MinY) / (fLengthY)) );
	}

	if (leftBB[axis].MaxY != xParentBB.MaxY)
	{
		node.MaxYBelongsToLeft = true;
		node.uChildMaxY = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (xEstimatedBB.MaxY - leftBB[axis].MaxY) / (fLengthY)) );
	}
	else
	{
		node.MaxYBelongsToLeft = false;
		node.uChildMaxY = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (xEstimatedBB.MaxY - rightBB[axis].MaxY) / (fLengthY)) );
	}

		
	if (leftBB[axis].MinZ != xParentBB.MinZ)
	{
		node.MinZBelongsToLeft = true;
		node.uChildMinZ = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (leftBB[axis].MinZ - xEstimatedBB.MinZ) / (fLengthZ)) );
	}
	else
	{
		node.MinZBelongsToLeft = false;
		node.uChildMinZ = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (rightBB[axis].MinZ - xEstimatedBB.MinZ) / (fLengthZ)) );
	}

	if (leftBB[axis].MaxZ != xParentBB.MaxZ)
	{
		node.MaxZBelongsToLeft = true;
		node.uChildMaxZ = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (xEstimatedBB.MaxZ - leftBB[axis].MaxZ) / (fLengthZ)) );
	}
	else
	{
		node.MaxZBelongsToLeft = false;
		node.uChildMaxZ = (u_char)Asura_Maths::FloatToInt_Chop( (g_fCompressionScale * (xEstimatedBB.MaxZ - rightBB[axis].MaxZ) / (fLengthZ)) );
	}

	Asura_Bounding_Box xEstimatedLeftBB;
	Asura_Bounding_Box xEstimatedRightBB;

	Asura_AABB_Tree::SetBoundingBoxes(node,xEstimatedBB,xEstimatedLeftBB,xEstimatedRightBB);



	if (objectsToTheLeft[axis]==1)
	{
		node.LeftIsLeaf = true;
		node.LeftIndex = NewObjectSet[start];
	}
	else
	{
		node.LeftIsLeaf = false;
		node.LeftIndex = ++NodeIndex;
		SortObjectsIntoNodes(start,start+objectsToTheLeft[axis],leftBB[axis], xEstimatedLeftBB);
	}

	if (objectsToTheRight[axis]==1)
	{
		node.RightIsLeaf = true;
		node.RightIndex = NewObjectSet[end-1];
	}
	else
	{
		node.RightIsLeaf = false;
		node.RightIndex = ++NodeIndex;
		SortObjectsIntoNodes(start+objectsToTheLeft[axis],end,rightBB[axis], xEstimatedRightBB);
	}

}



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AABB_Tree_Node::ReadFromChunkStream
// Author       : Andy Tate
// Date         : 17-Aug-2005
//
// Description  : Reads "this" tree-node from the specified chunk-stream.
//
// Parameters   : A reference to the chunk-stream to read from.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_AABB_Tree_Node::ReadFromChunkStream (Asura_Chunk_Stream& xStream)
{
	Asura_ReadFromStream_Bitfield_Wrapper<u_char> xFlagsBitfield(m_ucFlags);


	// Write the data...

	xStream	>> uChildMinX;
	xStream	>> uChildMinY;
	xStream	>> uChildMinZ;
	xStream	>> uChildMaxX;
    xStream	>> uChildMaxY;
	xStream	>> uChildMaxZ;

	xStream	>> LeftIndex;
	xStream	>> RightIndex;

	xStream >> xFlagsBitfield;

	xStream.Skip2BytePadding();
}

/// End of function 'Asura_AABB_Tree_Node::ReadFromChunkStream'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AABB_Tree_Node::WriteToChunkStream
// Author       : Andy Tate
// Date         : 17-Aug-2005
//
// Description  : Writes "this" tree-node to the specified chunk-stream.
//
// Parameters   : A reference to the chunk-stream to write to.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_AABB_Tree_Node::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_WriteToStream_Bitfield_Wrapper<u_char> xFlagsBitfield(m_ucFlags);


	// Create the bitfield...

	xFlagsBitfield.Add(MinXBelongsToLeft,	1);
	xFlagsBitfield.Add(MinYBelongsToLeft,	1);
	xFlagsBitfield.Add(MinZBelongsToLeft,	1);
	xFlagsBitfield.Add(MaxXBelongsToLeft,	1);
	xFlagsBitfield.Add(MaxYBelongsToLeft,	1);
	xFlagsBitfield.Add(MaxZBelongsToLeft,	1);
	xFlagsBitfield.Add(LeftIsLeaf,		1);
	xFlagsBitfield.Add(RightIsLeaf,		1);


	// Write the data...

	xStream	<< uChildMinX;
	xStream	<< uChildMinY;
	xStream	<< uChildMinZ;
	xStream	<< uChildMaxX;
    xStream	<< uChildMaxY;
	xStream	<< uChildMaxZ;

	xStream	<< LeftIndex;
	xStream	<< RightIndex;

	xStream << xFlagsBitfield;

	xStream.Write2BytePadding();
}

/// End of function 'Asura_AABB_Tree_Node::WriteToChunkStream'. ///

bool Asura_AABB_Tree::RayIntersectsAABB( const Asura_Vector_3& xAlpha, const Asura_Vector_3& xOneOverBeta, const Asura_Bounding_Box& xAABB, float& fDistance ) const
{
	float fMax = fDistance;	// max distance ray can travel 
	fDistance = 0.0f;		// want to return 0 if ray origin inside AABB

	// KJFL - 2 November 2007 - TODO
	// change format for bounding box to be these vectors, methinks?
	const Asura_Vector_3 xMin( xAABB.MinX, xAABB.MinY, xAABB.MinZ );
	const Asura_Vector_3 xMax( xAABB.MaxX, xAABB.MaxY, xAABB.MaxZ );

	// For all three slabs
	for (int i = 0; i < 3; i++)
	{
		const float fOneOverBeta = xOneOverBeta[i];

		if ( fOneOverBeta )		
		{
			// Compute intersection values of ray with near and far plane of slab
			float fDist1 = (xMin[i] - xAlpha[i]) * fOneOverBeta;
			float fDist2 = (xMax[i] - xAlpha[i]) * fOneOverBeta;

			// Make fDist1 be intersection with near plane, fDist2 with far plane
			Asura_Maths::Sort( fDist1, fDist2 );

			fDistance = Asura_Maths::Max( fDist1, fDistance );
			fMax = Asura_Maths::Min( fDist2, fMax );

			// If the interval inverts, there is no intersection
			if (fDistance > fMax)
			{
				return false;
			}
		}
		// Ray is parallel to slab. No hit if origin not within slab
		else if (xAlpha[i] < xMin[i] || xAlpha[i] > xMax[i])
		{
			return false;
		}
	}

	return true;
}


/// End of file. ///

