/**
*
* \file Asura_AABB_Tree.h
*
* Created by Kevin Lea - 21/06/2002 14:41:38
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

// (MH - 31.05.2004) note this file has been changed in some PS2 projects
// to allow PS2 optimisations. At some point the changes should be merged.

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_AABB_TREE_H__
#define __ASURA_AABB_TREE_H__

#include "Core/Asura_Handle.h"
#include "Core/Asura_Maths.h"

class Asura_Chunk_Stream;

struct Asura_AABB_Tree_Node
{
	// KJL - 08/05/2002 19:37:27 - Child Extents
	//
	// These extents are shared between the left and right nodes,
	// such that the bounding box of a child node is constructed
	// by a mixture of the parent's extents and these new ones,
	// as determined by the flags below. This halves the space
	// required.
	//
	// Further optimisations could be made by compressing this
	// data - e.g. using unsigned chars which indicate the scale
	// of the new extents relative to the parent bounding box,
	// rather than just using floats.


	u_char uChildMinX;
	u_char uChildMinY;
	u_char uChildMinZ;
	u_char uChildMaxX;
	u_char uChildMaxY;
	u_char uChildMaxZ;


	// KJL - 08/05/2002 19:41:25 - node index numbers
	// if node is a leaf then index is to a triangle/module/etc,
	// otherwise it's an index to its child node
	unsigned short LeftIndex;
	unsigned short RightIndex;


	union
	{
		unsigned char m_ucFlags;
		struct 
		{
			// NOTE: These bits are written out, adding any new bits means
			// ~~~~~ that you will need to add those bits to the bitfield
			//       preparation code in "WriteToChunkStream".
			unsigned char MinXBelongsToLeft :1;
			unsigned char MinYBelongsToLeft :1;
			unsigned char MinZBelongsToLeft :1;
			unsigned char MaxXBelongsToLeft :1;
			unsigned char MaxYBelongsToLeft :1;
			unsigned char MaxZBelongsToLeft :1;
			unsigned char LeftIsLeaf		:1;
			unsigned char RightIsLeaf		:1;
		};
	};
	// KJL - 16/05/2002 16:12:41 - hmm, this structure may actually
	// be irritating to some processors since it's an odd number of
	// bytes.

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
};

class Asura_AABB_Tree_Searcher
{
public:
	virtual ~Asura_AABB_Tree_Searcher() {}
	// Return true if the search should stop.
	virtual bool VisitNode( const u_short usNodeIndex ) { return false; }
};

class Asura_AABB_Tree
{
public:
	Asura_AABB_Tree();
	~Asura_AABB_Tree();

    // Output callbacks should return true iff they want the search to continue.
	void TestTree( const Asura_Bounding_Box& xBB, bool(*pfnOutputCallback)(u_short usIdx) );
	void TestTree( const Asura_Bounding_Box& xBB, Asura_AABB_Tree_Searcher& xSearchClass ) const;
	void TestTreeWithRay( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, float* pfDistance, bool(*pfnRayTestCallback)( const u_int uIndex ) );
    void Traverse( bool ( *pfnBBCallback )( const Asura_Bounding_Box& xBB ) );

	/// Kevin Lea - 11/10/2002 11:14:58 - originally I had the Nodes and RootBoundingBox private, but they need
	/// to be set by the chunk loaders so I've made them public. (Any interface would have to do something like
	/// return a pointer to Nodes or involve copying large chunks of memory so its just not worth it.)
	Asura_AABB_Tree_Node* Nodes;
	Asura_Bounding_Box RootBoundingBox;

	// RJF - 05 October 2004 - Common code save/load system
	static Asura_AABB_Tree* LoadTree(int schemaVersion, Asura_Handle hFileHandle, int iNodeCount);
#ifdef ASURA_TOOLS
	// Include the root node when passing iNodeCount to these functions
	static int GetTreeSize(int schemaVersion, const Asura_AABB_Tree *pxTree, int iNodeCount);
	// Note: SaveTree returns the number of bytes written
	static int SaveTree(int schemaVersion, const Asura_AABB_Tree *pxTree, int iNodeCount, char *pxDest);
	static void SaveTree(int schemaVersion, const Asura_AABB_Tree *pxTree, int iNodeCount, Asura_Chunk_Stream& xStream);
#endif

private:
	friend class Asura_AABB_Tree_Creator;

	bool TestNode(int nodeIndex, const Asura_Bounding_Box& xParentBB) const ASURA_CODE_SECTION("testnode");
	bool TestNode( int iNodeIndex, const Asura_Bounding_Box& xParentBB, const Asura_Bounding_Box& xTestBB, Asura_AABB_Tree_Searcher& xSearchClass ) const ASURA_CODE_SECTION("testnode");
	bool OutputAll(int nodeIndex) const ASURA_CODE_SECTION("testnode");
	bool OutputAll( int iNodeIndex, Asura_AABB_Tree_Searcher& xSearchClass ) const ASURA_CODE_SECTION("testnode");
    bool OutputAll(int iNodeIndex, const Asura_Bounding_Box& xParentBB) const ASURA_CODE_SECTION("testnode");

	// Returns true iff the search should continue.
	bool TestNodeWithRay( const u_int uNodeIndex, const Asura_Bounding_Box& xParentBB );

	bool RayIntersectsAABB( const Asura_Vector_3& xAlpha, const Asura_Vector_3& xOneOverBeta, const Asura_Bounding_Box& xAABB, float& fDistance) const;

	static void SetBoundingBoxes(const Asura_AABB_Tree_Node& xNode, const Asura_Bounding_Box& xParentBB, Asura_Bounding_Box& leftBB, Asura_Bounding_Box& rightBB);

	/// Kevin Lea - 15/05/2002 17:27:06 - data used in testing
	/// I originally thought to make this static i.e. shared between
	/// all trees, but there are occasions where multiple tests
	/// of different trees (e.g. tree within tree) are needed.
	bool(*OutputCallback)(u_short usIdx);
	const Asura_Bounding_Box*  TestBB;

	// Ray test data.
	float*					m_pfMaxDistance;
	const Asura_Vector_3*	m_pxRayOrigin;
	const Asura_Vector_3*	m_pxRayDirection;
	Asura_Vector_3			m_xOneOverRayDirection;
	bool					(*m_pfnRayTestCallback)( const u_int uIndex );

    // Traverse callback.
    bool ( *m_pfnBBCallback )( const Asura_Bounding_Box& xBB );
};

class Asura_AABB_Tree_Creator
{
public: 
    static Asura_AABB_Tree* Create( int iNumObjs, bool(*pfnPopulateArraysCallback)(u_int, Asura_Bounding_Box*, Asura_Vector_3*) );

protected:
	static void SortObjectsIntoNodes(int start, int end, const Asura_Bounding_Box& xParentBB, const Asura_Bounding_Box& xEstimatedBB);
		
	static unsigned short*				OldObjectSet;
	static unsigned short*				NewObjectSet;

	static unsigned char*				ToRightBitfield;

	static const Asura_Vector_3*		LeafCentre;
	static const Asura_Bounding_Box*	LeafAABB;

	static Asura_AABB_Tree*				Tree;
	static unsigned short				NodeIndex;
};

#endif//__ASURA_AABB_TREE_H__
