//////////////////////////////////////////////////////////////////////////////
//
// Asura_Callback.h
//
// Created by NJD - 14 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_CALLBACK_H__
#define __ASURA_CALLBACK_H__ 1

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// asura
// Turning the memory debug stuff off allows us to use placement new.
#include "Core/Asura_MemoryManagement_DebugOff.h"
#include "Debugging/Asura_Assert.h"


//////////////////////////////////////////////////////////////////////////////
//
// Callback classes.
//
// Usage:
//
// ----The caller of the callback----
//
// The system that is asking for a callback should request an instance of one of
// the two base callback classes.
// It should use Asura_Callback_Base if it doesn't need to provide any of its own arguments
// when calling it, or Asura_Callback_Base_1Arg if the system will be providing one argument
//
// eg.
//
// void RegisterCallback(const Asura_Callback_Base<int>*)
// could be used if the system wants a callback which takes no arguments and returns an int
//
//
// ---- The provider of the callback----
//
// The code that will be providing the callback should make use of the derived classes
// for providing the callback
//
// If the function is a member of a class then use Asura_Callback_MemberFunction
// eg.
//
// class SomeClass
// {
// public:
//     int SomeFunction();
// } xSomeClass;
//
// Asura_Callback_MemberFunction<int, SomeClass > xMyCallback( &xSomeClass, &SomeClass::SomeFunction );
// RegisterCallback( xMyCallback );
//
//
// If the function is a static/global function that wants to take some user data to identify the context
// then use Asura_Callback_UserData
// eg.
//
// static int SomeFunction(float fUserData);
//
// float fUserData;
// Asura_Callback_UserData<int, float> xMyCallback( SomeFunction, fUserData );
// RegisterCallback( xMyCallback );
//
// If the function is a static/global function that doesn't need any such arguments then use Asura_Callback_Void
// eg.
//
// static int SomeFunction();
//
// Asura_Callback_Void<int> xMyCallback( SomeFunction );
// RegisterCallback( xMyCallback);
//
// The 1Arg variants work in much the same way.
//
//
// ----The wrapper classes----
//
// Asura_Callback_Wrapper and Asura_Callback_Wrapper_1Arg are two wrapper classes for storing a pointer to a callback.
// They just exist to avoid the hassle of having to clone and delete the callbacks when copying things around
// in the overlay code
//
// RAR 13/05/2008
//////////////////////////////////////////////////////////////////////////////

// Forward declarations.

template <typename ReturnType,typename ArgType> class Asura_Callback_Base_1Arg;
template <typename ReturnType,typename ArgType> class Asura_Callback_Wrapper_1Arg;

//////////////////////////////////////////////////////////////////////////////
//Asura_Callback_Base and derived classes

//Base callback that will take no arguments from the caller
template <typename ReturnType>
class Asura_Callback_Base
{
public :
	virtual ~Asura_Callback_Base() { }

	//Allocates a copy of this callback
	virtual Asura_Callback_Base* Clone() const = 0;
	//Allocates a copy of this callback using the provided memory
    virtual Asura_Callback_Base* Clone( void* pMemory ) const = 0;

	//Calls the callback function
	virtual ReturnType Execute() const = 0;

	//Returns true if the callback function is null
	virtual bool IsNull() const = 0;

	//Gets the size of the class
	virtual u_int GetSize() const = 0;

	#ifdef ASURA_TOOLS
	//Gets the function pointer if possible.
	//Used by Asura_Function_Dialog to compare against the previous used function
	//Doesn't work for member functions.
	virtual const void* GetFunctionPointer() const {return 0;}
	#endif
};

//Wrapper for storing an instance of the callback
template <typename ReturnType>
class Asura_Callback_Wrapper
{
public :
	Asura_Callback_Wrapper()
	{
		m_pxCallback = 0;
	}
	Asura_Callback_Wrapper(const Asura_Callback_Base<ReturnType>* pxCallback)
	{
		m_pxCallback = 0;
		SetCallback(pxCallback);		
	}
	Asura_Callback_Wrapper(const Asura_Callback_Wrapper<ReturnType>& xWrapper)
	{
		m_pxCallback = 0;
		SetCallback(xWrapper.GetCallback());
	}
	// These constructors convert a 1 argument callback into a 0 argument callback with user data.
	template< typename ArgType > Asura_Callback_Wrapper( const Asura_Callback_Base_1Arg< ReturnType, ArgType >* pxCallback1Arg, ArgType xUserData );
	template< typename ArgType > Asura_Callback_Wrapper( const Asura_Callback_Wrapper_1Arg< ReturnType, ArgType >& xWrapper1Arg, ArgType xUserData );

	const Asura_Callback_Wrapper<ReturnType>& operator = (const Asura_Callback_Wrapper<ReturnType>& xWrapper)
	{
		if(&xWrapper != this)
		{
			SetCallback(xWrapper.GetCallback());
		}
		return *this;
	}

	~Asura_Callback_Wrapper()
	{
		delete m_pxCallback;
	}

	//Changes the callback stored in the wrapper
	void SetCallback(const Asura_Callback_Base<ReturnType>* pxCallback)
	{
		MakeNull();		
		m_pxCallback = pxCallback && !pxCallback->IsNull() ? pxCallback->Clone() : 0;		
	}
	const Asura_Callback_Base<ReturnType>* GetCallback() const {return m_pxCallback;}
	
	//Returns true if there is a non-null function
	bool IsValid() const {return m_pxCallback && !m_pxCallback->IsNull();}
	//Returns true if there is no function or if it is null
	bool IsNull() const {return !m_pxCallback || m_pxCallback->IsNull();}

	#ifdef ASURA_TOOLS
	const void* GetFunctionPointer() const {return m_pxCallback ? m_pxCallback->GetFunctionPointer() : 0;}
	#endif

	//Deletes the callback in the wrapper
	void MakeNull()
	{
		delete m_pxCallback;
		m_pxCallback = 0;
	}

	//calls the callback
	ReturnType Execute() const 
	{
		return m_pxCallback->Execute();
	}

private :
	Asura_Callback_Base<ReturnType>* m_pxCallback;
};

//Callback that takes no arguments from the caller , and is a member function of a class
template <typename ReturnType, typename ClassType>
class Asura_Callback_MemberFunction : public Asura_Callback_Base<ReturnType>
{
    typedef ReturnType ( ClassType::*CallbackFnPtr )();
public :
    Asura_Callback_MemberFunction() : m_pxObject( NULL ), m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_MemberFunction( ClassType* pxObject, CallbackFnPtr pFnPtr ) : m_pxObject( pxObject ), m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( ClassType* pxObject, CallbackFnPtr pFnPtr ) { m_pxObject = pxObject; m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base<ReturnType>* Clone() const { return new Asura_Callback_MemberFunction<ReturnType, ClassType>( m_pxObject, m_pCallbackFnPtr ); }
    virtual Asura_Callback_Base<ReturnType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_MemberFunction<ReturnType, ClassType>( m_pxObject, m_pCallbackFnPtr ); }
    
    virtual ReturnType Execute() const { return ( m_pxObject->*m_pCallbackFnPtr )(); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}

	virtual u_int GetSize() const {return sizeof(*this);}
private:
    ClassType*         m_pxObject;
    CallbackFnPtr m_pCallbackFnPtr;
};

//Callback that takes no arguments from the caller , and is a member function of a class , and passes back some user data that is provided on construction 
template <typename ReturnType, typename ClassType, typename UserDataType>
class Asura_Callback_MemberFunctionWithUserData : public Asura_Callback_Base<ReturnType>
{
    typedef ReturnType ( ClassType::*CallbackFnPtr )(UserDataType);
public :
    Asura_Callback_MemberFunctionWithUserData() : m_pxObject( NULL ), m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_MemberFunctionWithUserData( ClassType* pxObject, CallbackFnPtr pFnPtr, UserDataType xUserData ) : m_pxObject( pxObject ), m_pCallbackFnPtr( pFnPtr ), m_xUserData(xUserData) {}
    void SetCallback( ClassType* pxObject, CallbackFnPtr pFnPtr, UserDataType xUserData ) { m_pxObject = pxObject; m_pCallbackFnPtr = pFnPtr; m_xUserData = xUserData;}

    virtual Asura_Callback_Base<ReturnType>* Clone() const { return new Asura_Callback_MemberFunctionWithUserData<ReturnType, ClassType, UserDataType>( m_pxObject, m_pCallbackFnPtr, m_xUserData ); }
    virtual Asura_Callback_Base<ReturnType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_MemberFunctionWithUserData<ReturnType, ClassType, UserDataType>( m_pxObject, m_pCallbackFnPtr, m_xUserData ); }
    
    virtual ReturnType Execute() const { return ( m_pxObject->*m_pCallbackFnPtr )(m_xUserData); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}

	virtual u_int GetSize() const {return sizeof(*this);}
private:
    ClassType*         m_pxObject;
    CallbackFnPtr m_pCallbackFnPtr;
	UserDataType m_xUserData;
};

//Callback that takes no arguments from the caller, and passes back some user data that is provided on construction
template <typename ReturnType, typename UserDataType>
class Asura_Callback_UserData : public Asura_Callback_Base<ReturnType>
{
    typedef ReturnType ( *CallbackFnPtr )(UserDataType);
public :
    Asura_Callback_UserData() : m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_UserData(CallbackFnPtr pFnPtr, UserDataType xUserData ) : m_xUserData( xUserData ), m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( CallbackFnPtr pFnPtr, UserDataType xUserData ) { m_xUserData = xUserData; m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base<ReturnType>* Clone() const { return new Asura_Callback_UserData<ReturnType,UserDataType>( m_pCallbackFnPtr, m_xUserData ); }
    virtual Asura_Callback_Base<ReturnType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_UserData<ReturnType,UserDataType>(m_pCallbackFnPtr, m_xUserData ); }
    
    virtual ReturnType Execute() const { return ( m_pCallbackFnPtr )(m_xUserData); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}

	#ifdef ASURA_TOOLS
	virtual const void* GetFunctionPointer() const {return m_pCallbackFnPtr;}
	#endif

	virtual u_int GetSize() const {return sizeof(*this);}

private:
	UserDataType m_xUserData;
    CallbackFnPtr m_pCallbackFnPtr;
};

//Callback that will take no arguments from the caller
template <typename ReturnType>
class Asura_Callback_Void : public Asura_Callback_Base<ReturnType>
{
	typedef ReturnType ( *CallbackFnPtr )();
public :
    Asura_Callback_Void() : m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_Void( CallbackFnPtr pFnPtr ) : m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( CallbackFnPtr pFnPtr ) { m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base<ReturnType>* Clone() const { return new Asura_Callback_Void<ReturnType>( m_pCallbackFnPtr); }
    virtual Asura_Callback_Base<ReturnType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_Void<ReturnType>(m_pCallbackFnPtr); }
    
    virtual ReturnType Execute() const { return ( m_pCallbackFnPtr )(); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}
	
	#ifdef ASURA_TOOLS
	virtual const void* GetFunctionPointer() const {return m_pCallbackFnPtr;}
	#endif

	virtual u_int GetSize() const {return sizeof(*this);}
private:
    CallbackFnPtr m_pCallbackFnPtr;
};

//////////////////////////////////////////////////////////////////////////////
//Asura_Callback_Base_1Arg and derived classes

//Base callback that will take one argument from the caller
template <typename ReturnType,typename ArgType> 
class Asura_Callback_Base_1Arg
{
public :
	virtual ~Asura_Callback_Base_1Arg() { }
	//Allocates a copy of this callback
    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone() const = 0;
	//Allocates a copy of this callback using the provided memory
    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone( void* pMemory ) const = 0;
	// Converts a callback that takes one argument into a user data callback.
	virtual Asura_Callback_Base<ReturnType>* CreateUserDataCallback( ArgType xUserData ) const = 0;

	//Calls the callback function
	virtual ReturnType Execute(ArgType) const = 0;
	
	//Returns true if the callback function is null
	virtual bool IsNull() const = 0;
	
	//Gets the size of the class
	virtual u_int GetSize() const = 0;
};

// This has to be defined after Asura_Callback_Base_1Arg.
template< typename ReturnType >
template< typename ArgType >
Asura_Callback_Wrapper< ReturnType >::Asura_Callback_Wrapper( const Asura_Callback_Base_1Arg< ReturnType, ArgType >* pxCallback1Arg, ArgType xUserData )
{
	m_pxCallback = 0;
	SetCallback( pxCallback1Arg->CreateUserDataCallback( xUserData ) );
}

//Wrapper for storing an instance of the callback
template <typename ReturnType, typename ArgType>
class Asura_Callback_Wrapper_1Arg
{
public :
	Asura_Callback_Wrapper_1Arg()
	{
		m_pxCallback = 0;
	}
	Asura_Callback_Wrapper_1Arg(const Asura_Callback_Base_1Arg<ReturnType,ArgType>* pxCallback)
	{
		m_pxCallback = 0;
		SetCallback(pxCallback);
	}

	Asura_Callback_Wrapper_1Arg(const Asura_Callback_Wrapper_1Arg<ReturnType,ArgType>& xWrapper)
	{
		m_pxCallback = 0;
		SetCallback(xWrapper.GetCallback());
	}

	const Asura_Callback_Wrapper_1Arg<ReturnType,ArgType>& operator = (const Asura_Callback_Wrapper_1Arg<ReturnType,ArgType>& xWrapper)
	{
		if(&xWrapper != this)
		{
			SetCallback(xWrapper.GetCallback());
		}
		return *this;
	}

	~Asura_Callback_Wrapper_1Arg()
	{
		delete m_pxCallback;
	}

	//Changes the callback stored in the wrapper
	void SetCallback(const Asura_Callback_Base_1Arg<ReturnType,ArgType>* pxCallback)
	{
		MakeNull();		
		m_pxCallback = pxCallback && !pxCallback->IsNull() ? pxCallback->Clone() : 0;		
	}
	const Asura_Callback_Base_1Arg<ReturnType,ArgType>* GetCallback() const {return m_pxCallback;}
	
	//Returns true if there is a non-null function
	bool IsValid() const {return m_pxCallback && !m_pxCallback->IsNull();}
	//Returns true if there is no function or if it is null
	bool IsNull() const {return !m_pxCallback || m_pxCallback->IsNull();}

	//Deletes the callback in the wrapper
	void MakeNull()
	{
		delete m_pxCallback;
		m_pxCallback = 0;
	}

	//calls the callback
	ReturnType Execute(ArgType xArg) const 
	{
		return m_pxCallback->Execute(xArg);
	}

private :
	Asura_Callback_Base_1Arg<ReturnType,ArgType>* m_pxCallback;
};

// This has to be defined after Asura_Callback_Wrapper_1Arg.
template< typename ReturnType >
template< typename ArgType >
Asura_Callback_Wrapper< ReturnType >::Asura_Callback_Wrapper( const Asura_Callback_Wrapper_1Arg< ReturnType, ArgType >& xWrapper1Arg, ArgType xUserData )
{
	m_pxCallback = 0;
	SetCallback( xWrapper1Arg.GetCallback()->CreateUserDataCallback( xUserData ) );
}

//Callback that takes one argument from the caller , and is a member function of a class
template <typename ReturnType, typename ArgType, typename ClassType>
class Asura_Callback_MemberFunction_1Arg : public Asura_Callback_Base_1Arg<ReturnType, ArgType>
{
    typedef ReturnType ( ClassType::*CallbackFnPtr )(ArgType);
public :
    Asura_Callback_MemberFunction_1Arg() : m_pxObject( NULL ), m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_MemberFunction_1Arg( ClassType* pxObject, CallbackFnPtr pFnPtr ) : m_pxObject( pxObject ), m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( ClassType* pxObject, CallbackFnPtr pFnPtr ) { m_pxObject = pxObject; m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone() const { return new Asura_Callback_MemberFunction_1Arg<ReturnType, ArgType, ClassType>( m_pxObject, m_pCallbackFnPtr ); }
    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_MemberFunction_1Arg<ReturnType, ArgType, ClassType>( m_pxObject, m_pCallbackFnPtr ); }
	virtual Asura_Callback_Base<ReturnType>* CreateUserDataCallback( ArgType xUserData ) const
	{
		return new Asura_Callback_MemberFunctionWithUserData<ReturnType,ClassType,ArgType>( m_pxObject, m_pCallbackFnPtr, xUserData );
	}

    virtual ReturnType Execute(ArgType xArg) const { return ( m_pxObject->*m_pCallbackFnPtr )(xArg); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}

	virtual u_int GetSize() const {return sizeof(*this);}
private:
    ClassType*    m_pxObject;
    CallbackFnPtr m_pCallbackFnPtr;
};


//Callback that takes one argument from the caller, and passes back some user data that is provided on construction
template <typename ReturnType, typename ArgType, typename UserDataType>
class Asura_Callback_UserData_1Arg : public Asura_Callback_Base_1Arg<ReturnType, ArgType>
{
    typedef ReturnType ( *CallbackFnPtr )(ArgType, UserDataType);
public :
    Asura_Callback_UserData_1Arg() : m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_UserData_1Arg(CallbackFnPtr pFnPtr, UserDataType xUserData ) : m_xUserData( xUserData ), m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( CallbackFnPtr pFnPtr, UserDataType xUserData ) { m_xUserData = xUserData; m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone() const { return new Asura_Callback_UserData_1Arg<ReturnType, ArgType, UserDataType>( m_pCallbackFnPtr, m_xUserData ); }
    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_UserData_1Arg<ReturnType, ArgType, UserDataType>(m_pCallbackFnPtr, m_xUserData ); }
	// We have to assume that the caller wants to use this new user data, rather than the already stored user data.
	virtual Asura_Callback_Base<ReturnType>* CreateUserDataCallback( ArgType xUserData ) const { return new Asura_Callback_UserData<ReturnType, ArgType>( m_pCallbackFnPtr, xUserData ); }

    virtual ReturnType Execute(ArgType xArg) const { return ( m_pCallbackFnPtr )(xArg, m_xUserData); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}
	
	virtual u_int GetSize() const {return sizeof(*this);}
private:
	UserDataType m_xUserData;
    CallbackFnPtr m_pCallbackFnPtr;
};

//Callback that will take one argument from the caller
template <typename ReturnType, typename ArgType>
class Asura_Callback_Void_1Arg : public Asura_Callback_Base_1Arg<ReturnType, ArgType>
{
	typedef ReturnType ( *CallbackFnPtr )(ArgType);
public :
    Asura_Callback_Void_1Arg() : m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_Void_1Arg( CallbackFnPtr pFnPtr ) : m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( CallbackFnPtr pFnPtr ) { m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone() const { return new Asura_Callback_Void_1Arg<ReturnType, ArgType>( m_pCallbackFnPtr); }
    virtual Asura_Callback_Base_1Arg<ReturnType,ArgType>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_Void_1Arg<ReturnType, ArgType>(m_pCallbackFnPtr); }
	virtual Asura_Callback_Base<ReturnType>* CreateUserDataCallback( ArgType xUserData ) const { return new Asura_Callback_UserData<ReturnType, ArgType>( m_pCallbackFnPtr, xUserData ); }

    virtual ReturnType Execute(ArgType xArg) const { return ( m_pCallbackFnPtr )(xArg); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}
	
	virtual u_int GetSize() const {return sizeof(*this);}

private:
    CallbackFnPtr m_pCallbackFnPtr;
};

//////////////////////////////////////////////////////////////////////////////
//Asura_Callback_Base_2Arg and derived classes

//Base callback that will take one argument from the caller
template <typename ReturnType,typename Arg1Type,typename Arg2Type> 
class Asura_Callback_Base_2Arg
{
public :
	virtual ~Asura_Callback_Base_2Arg() { }
	//Allocates a copy of this callback
    virtual Asura_Callback_Base_2Arg* Clone() const = 0;
	//Allocates a copy of this callback using the provided memory
    virtual Asura_Callback_Base_2Arg* Clone( void* pMemory ) const = 0;

	//Calls the callback function
	virtual ReturnType Execute(Arg1Type, Arg2Type) const = 0;
	
	//Returns true if the callback function is null
	virtual bool IsNull() const = 0;
	
	//Gets the size of the class
	virtual u_int GetSize() const = 0;
};

//Wrapper for storing an instance of the callback
template <typename ReturnType, typename Arg1Type, typename Arg2Type>
class Asura_Callback_Wrapper_2Arg
{
public :
	Asura_Callback_Wrapper_2Arg()
	{
		m_pxCallback = 0;
	}
	Asura_Callback_Wrapper_2Arg(const Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* pxCallback)
	{
		m_pxCallback = 0;
		SetCallback(pxCallback);
	}

	Asura_Callback_Wrapper_2Arg(const Asura_Callback_Wrapper_2Arg<ReturnType,Arg1Type,Arg2Type>& xWrapper)
	{
		m_pxCallback = 0;
		SetCallback(xWrapper.GetCallback());
	}

	const Asura_Callback_Wrapper_2Arg<ReturnType,Arg1Type,Arg2Type>& operator = (const Asura_Callback_Wrapper_2Arg<ReturnType,Arg1Type,Arg2Type>& xWrapper)
	{
		if(&xWrapper != this)
		{
			SetCallback(xWrapper.GetCallback());
		}
		return *this;
	}

	~Asura_Callback_Wrapper_2Arg()
	{
		delete m_pxCallback;
	}

	//Changes the callback stored in the wrapper
	void SetCallback(const Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* pxCallback)
	{
		MakeNull();		
		m_pxCallback = pxCallback && !pxCallback->IsNull() ? pxCallback->Clone() : 0;		
	}
	const Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* GetCallback() const {return m_pxCallback;}
	
	//Returns true if there is a non-null function
	bool IsValid() const {return m_pxCallback && !m_pxCallback->IsNull();}
	//Returns true if there is no function or if it is null
	bool IsNull() const {return !m_pxCallback || m_pxCallback->IsNull();}

	//Deletes the callback in the wrapper
	void MakeNull()
	{
		delete m_pxCallback;
		m_pxCallback = 0;
	}

	//calls the callback
	ReturnType Execute(Arg1Type xArg1, Arg2Type xArg2) const 
	{
		return m_pxCallback->Execute(xArg1, xArg2);
	}

private :
	Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* m_pxCallback;
};



//Callback that takes one argument from the caller , and is a member function of a class
template <typename ReturnType, typename Arg1Type, typename ClassType, typename Arg2Type>
class Asura_Callback_MemberFunction_2Arg : public Asura_Callback_Base_2Arg<ReturnType, Arg1Type, Arg2Type>
{
    typedef ReturnType ( ClassType::*CallbackFnPtr )(Arg1Type);
public :
    Asura_Callback_MemberFunction_2Arg() : m_pxObject( NULL ), m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_MemberFunction_2Arg( ClassType* pxObject, CallbackFnPtr pFnPtr ) : m_pxObject( pxObject ), m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( ClassType* pxObject, CallbackFnPtr pFnPtr ) { m_pxObject = pxObject; m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* Clone() const { return new Asura_Callback_MemberFunction_2Arg<ReturnType, Arg1Type, Arg2Type, ClassType>( m_pxObject, m_pCallbackFnPtr ); }
    virtual Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_MemberFunction_2Arg<ReturnType, Arg1Type, Arg2Type, ClassType>( m_pxObject, m_pCallbackFnPtr ); }

    virtual ReturnType Execute(Arg1Type xArg1, Arg2Type xArg2) const { return ( m_pxObject->*m_pCallbackFnPtr )(xArg1, xArg2); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}

	virtual u_int GetSize() const {return sizeof(*this);}
private:
    ClassType*         m_pxObject;
    CallbackFnPtr m_pCallbackFnPtr;
};


//Callback that takes one argument from the caller, and passes back some user data that is provided on construction
template <typename ReturnType, typename Arg1Type, typename Arg2Type,typename UserDataType>
class Asura_Callback_UserData_2Arg : public Asura_Callback_Base_2Arg<ReturnType, Arg1Type, Arg2Type>
{
    typedef ReturnType ( *CallbackFnPtr )(Arg1Type, Arg2Type, UserDataType);
public :
    Asura_Callback_UserData_2Arg() : m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_UserData_2Arg(CallbackFnPtr pFnPtr, UserDataType xUserData ) : m_xUserData( xUserData ), m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( CallbackFnPtr pFnPtr, UserDataType xUserData ) { m_xUserData = xUserData; m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* Clone() const { return new Asura_Callback_UserData_2Arg<ReturnType, Arg1Type, Arg2Type, UserDataType>( m_pCallbackFnPtr, m_xUserData ); }
    virtual Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_UserData_2Arg<ReturnType, Arg1Type, Arg2Type, UserDataType>(m_pCallbackFnPtr, m_xUserData ); }

    virtual ReturnType Execute(Arg1Type xArg1, Arg2Type xArg2) const { return ( m_pCallbackFnPtr )(xArg1, xArg2, m_xUserData); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}
	
	virtual u_int GetSize() const {return sizeof(*this);}
private:
	UserDataType m_xUserData;
    CallbackFnPtr m_pCallbackFnPtr;
};

//Callback that will take one argument from the caller
template <typename ReturnType, typename Arg1Type, typename Arg2Type>
class Asura_Callback_Void_2Arg : public Asura_Callback_Base_2Arg<ReturnType, Arg1Type, Arg2Type>
{
	typedef ReturnType ( *CallbackFnPtr )(Arg1Type,Arg2Type);
public :
    Asura_Callback_Void_2Arg() : m_pCallbackFnPtr( NULL ) {}
    Asura_Callback_Void_2Arg( CallbackFnPtr pFnPtr ) : m_pCallbackFnPtr( pFnPtr ) {}
    void SetCallback( CallbackFnPtr pFnPtr ) { m_pCallbackFnPtr = pFnPtr; }

    virtual Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* Clone() const { return new Asura_Callback_Void_2Arg<ReturnType, Arg1Type, Arg2Type>( m_pCallbackFnPtr); }
    virtual Asura_Callback_Base_2Arg<ReturnType,Arg1Type,Arg2Type>* Clone( void* pMemory ) const { return new ( pMemory ) Asura_Callback_Void_2Arg<ReturnType, Arg1Type, Arg2Type>(m_pCallbackFnPtr); }

    virtual ReturnType Execute(Arg1Type xArg1, Arg2Type xArg2) const { return ( m_pCallbackFnPtr )(xArg1, xArg2); }

	virtual bool IsNull() const {return m_pCallbackFnPtr == 0;}
	
	virtual u_int GetSize() const {return sizeof(*this);}
private:
    CallbackFnPtr m_pCallbackFnPtr;
};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_MemoryManagement_DebugOn.h"

#endif //__ASURA_CALLBACK_H__
