//////////////////////////////////////////////////////////////////////////////
//
// File: Asura_LOS_Penetration.h
//
// Created by MGH 13/12/2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//
// Description: Class does LOS tests for penetrating projectiles (eg bullets).
//              Given stopping powers of various materials, it returns
//              multiple hits.
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_LOS_PENETRATION_H__
#define __ASURA_LOS_PENETRATION_H__

//////////////////////////
// Forward declarations //
//////////////////////////
class Asura_LOS_InputData;
class Asura_LOS_OutputData;

//////////////
// Includes //
//////////////
#include "Network/Asura_Guid.h"

///////////////
// Constants //
///////////////
// Projectile will stop when the total stopping power reaches this number.
const float fASURA_MAX_STOPPING_POWER = 1.0f;
const int iASURA_IGNORE_LOS_HIT_LIMIT = -1;

///////////////////////
// Asura_Penetration //
///////////////////////
// Function object. Derive from this for each type of bullet to provide
// the stopping powers of different materials and entities.
// Returns 0.0 (doesn't slow bullet) to 1.0 (stops bullet completely).
class Asura_Penetration
{
public:
	Asura_Penetration() {}
	virtual ~Asura_Penetration() {}

	// Override these functions in derived classes.
	
	// Called when an environmental polygon is hit.
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	virtual float GetStoppingPower(Asura_Hash_ID uMaterialResponseHashID) const
	{
		return fASURA_MAX_STOPPING_POWER;
	}

#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	virtual float GetStoppingPower(short sMaterialID) const
	{
		return fASURA_MAX_STOPPING_POWER;
	}

#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	// Called when an entity might have been hit.
	// Provides guid, bone/attachment etc in output data and also the classification.
	// Note: This function needs to work on both remote clients (with no server entities)
	// and dedicated servers (with no client entities).
	virtual float GetStoppingPower( const Asura_LOS_OutputData& xOutput,
									unsigned short usClassification) const
	{
		return fASURA_MAX_STOPPING_POWER;
	}

	// (AW - 30.03.2005) - The next 2 functions address the problem that this penetration
	// model treats thin walls exactly the same as thick walls. Any correct way to handle this
	// would involve making invalid assumptions about the way meshes are built. Instead, the
	// model will only allow an LOS to carry on for x metres once it has received environment 
	// impacts adding up to x stopping power. 
	virtual float GetThresholdBeforeMaxDistanceKicksIn() const
	{
		return fASURA_MAX_STOPPING_POWER;
	}

	virtual float GetMaxDistanceAfterEnvironmentImpact() const
	{
		return 999999999.9f;
	}

	//TY 04.12.07 - Return the maximum LoS hits to register before we stop checking for more
	//              set to -1 to ignore (no maximum)
	virtual int GetMaxEnvironmentLoSHits () const
	{
		return iASURA_IGNORE_LOS_HIT_LIMIT;
	}

	virtual int GetMaxEntityLoSHits () const
	{
		return iASURA_IGNORE_LOS_HIT_LIMIT;
	}

};


///////////////////////////
// Asura_LOS_Penetration //
///////////////////////////
class Asura_LOS_Penetration
{
public:
	// Do a LOS test against everything, using the given input data.
	// A functor provides callbacks for the stopping power of different materials.
	static void TestAgainstEverything(const Asura_LOS_InputData* const pxData,
									  const Asura_Penetration& xFunctor);
	
	// Test only against the environment.
	static void TestAgainstEnvironment(const Asura_LOS_InputData* const pxData,
									   const Asura_Penetration& xFunctor);

	//
	// Results of LOS test.
	//

	// Number of hits returned by the LOS test.
	// Note: this will be zero if nothing was hit (eg ray pointing at skybox).
	static int GetNumberOfHits() { return s_iNumHits; }

	// Gives the hits sorted by distance (0=closest).
	static const Asura_LOS_OutputData* GetHit(int iHitNumber);

	// Gives the cumulative stopping power up to and including the specified hit,
	// hits are sorted by distance as for GetHit().
	// Returns 0.0f (not slowed) to 1.0f (stopped).
	// Note: passing in iHitNumber equal to -1 is allowed and will return 0.0f.
	static float GetTotalStoppingPower(int iHitNumber);

	// Distance of furthest hit (or maximum range given in the input data if the
	// projectile wasn't stopped).
	static float GetMaxDistance() { return s_fMaxDistance; }

	// Search through the impacts to see whether a specific entity was hit.
	// Useful when we want to check whether a bullet hit its intended target.
	static bool HasHitEntity(Asura_Guid uGuid);

#ifdef _DEBUG
	// Print out contents of arrays, for debugging purposes.
	static void Dump();
#endif //_DEBUG

private:
	// Register a possible hit on an entity.
	// Return false if the array is full and no further hits can be stored.
	static bool RegisterHit(const Asura_Penetration& xFunctor,
							unsigned short usClassification);

private:
	static int s_iNumHits;
	static float s_fMaxDistance;


	// Friends
	friend class Asura_DynamicBVH_Penetration_Functor;
};

#endif //__ASURA_LOS_PENETRATION_H__
