//////////////////////////////////////////////////////////////////////////////
//
// Asura_Maths.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Colour.h"

// G L O B A L S /////////////////////////////////////////////////////////////

const Asura_IntVector_2	Asura_Maths::ZeroIntVector2(0, 0);
const Asura_Euler_3		Asura_Maths::ZeroEuler3( 0.0f, 0.0f, 0.0f );
const Asura_AVector3	Asura_Maths::ZeroAVector3( 0.0f, 0.0f, 0.0f );
const Asura_AVector4	Asura_Maths::ZeroAVector4W0( 0.0f, 0.0f, 0.0f, 0.0f);
const Asura_AVector4	Asura_Maths::ZeroAVector4W1( 0.0f, 0.0f, 0.0f, 1.0f);
const Asura_AMatrix3x3	Asura_Maths::IdentityAMatrix3x3( 1.0f, 0.0f, 0.0f, 
														 0.0f, 1.0f, 0.0f, 
														 0.0f, 0.0f, 1.0f );
const Asura_AMatrix4x4	Asura_Maths::IdentityAMatrix4x4( 1.0f, 0.0f, 0.0f, 0.0f,
														 0.0f, 1.0f, 0.0f, 0.0f,
														 0.0f, 0.0f, 1.0f, 0.0f,
														 0.0f, 0.0f, 0.0f, 1.0f );

Asura_RandomGenerator Asura_Random::s_xSharedGenerator;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Maths.
//

void Asura_Maths::OrientationFromNormal( Asura_Matrix_3x3& xOrientation, const Asura_Vector_3& xNormal )
{
    Asura_Vector_3 xForwards(0.0f, 0.0f, 1.0f);
    Asura_Vector_3 xLeft(1.0f, 0.0f, 0.0f);
    Asura_Vector_3 xDown(-xNormal);

	Asura_Assertf( Asura_Maths::Abs( 1.0f - xNormal.MagnitudeSquared() ) < 1.0e-3, ( "Normal is not normalised, mag sq. is %f", xNormal.MagnitudeSquared() ) );
    const float	fDotProduct = xDown * xForwards;

    // Decide whether which potential side-vector we're furthest away from...
    if ( Asura_Maths::Abs( fDotProduct ) > 0.707f )
    {
		xForwards = xLeft ^ xDown;
		xForwards.Normalise();
		xLeft	  = xDown ^ xForwards;
    }
    else
    {
		xLeft	  = xDown ^ xForwards;
		xLeft.Normalise();
        xForwards = xLeft ^ xDown;
    }

    // Re-create the orientation matrix...
    xOrientation.e11 = xLeft[ 0 ];
    xOrientation.e21 = xLeft[ 1 ];
    xOrientation.e31 = xLeft[ 2 ];
    xOrientation.e12 = xDown[ 0 ];
    xOrientation.e22 = xDown[ 1 ];
    xOrientation.e32 = xDown[ 2 ];
    xOrientation.e13 = xForwards[ 0 ];
    xOrientation.e23 = xForwards[ 1 ];
    xOrientation.e33 = xForwards[ 2 ];
}

void Asura_Maths::GetNearestPointOnLineSegment( const Asura_AVector3& xLineStart, const Asura_AVector3& xLineEnd, const Asura_AVector3& xPoint, Asura_AVector3& xNearestPoint )
{
    const Asura_AVector3 xLine              = xLineEnd - xLineStart;
    const float          fLineLengthSquared = xLine.MagnitudeSquared();

    // Find the nearest point on the segment. If the edge length is zero,
    // just take the start of the segment to be the nearest point. Otherwise
    // work out the nearest point on an infinite line, and then clamp that
    // to be on the segment (so it gets clamped to the start or end points).
    float fLambda = 0.0f;
    if ( fLineLengthSquared > 0.0f )
    {
        fLambda = Asura_Maths::ClampToRange( ( ( xPoint - xLineStart ) * xLine ) / fLineLengthSquared, 0.0f, 1.0f );
    }

    xNearestPoint = xLineStart + xLine * fLambda;
}

void Asura_Maths::GetNearestPointOnLineSegment( const Asura_Vector_3& xLineStart, const Asura_Vector_3& xLineEnd, const Asura_Vector_3& xPoint, Asura_Vector_3& xNearestPoint )
{
    const Asura_Vector_3 xLine              = xLineEnd - xLineStart;
    const float          fLineLengthSquared = xLine.MagnitudeSquared();

    // Find the nearest point on the segment. If the edge length is zero,
    // just take the start of the segment to be the nearest point. Otherwise
    // work out the nearest point on an infinite line, and then clamp that
    // to be on the segment (so it gets clamped to the start or end points).
    float fLambda = 0.0f;
    if ( fLineLengthSquared > 0.0f )
    {
        fLambda = Asura_Maths::ClampToRange( ( ( xPoint - xLineStart ) * xLine ) / fLineLengthSquared, 0.0f, 1.0f );
    }

    xNearestPoint = xLineStart + xLine * fLambda;
}

//
// Asura_Random.
//

void Asura_Random::Initialise( void )
{
	// check that Platform_Initialise() has called Asura_Random::SetSeed() with something
	// appropriate such as time().
#ifdef ASURA_ASSERT
	int iHigh;
	int iLow;
	s_xSharedGenerator.GetCurrentState(iLow, iHigh);
	Asura_Assert(iLow != 0 || iHigh != ~0, "random number generator not seeded");
#endif
}

void Asura_Random::Shutdown( void )
{
}

//
// Asura_Matrix_3x3.
//

bool Asura_Matrix_3x3::ApproximatelyEquals(const Asura_Matrix_3x3& xMat, float fEpsilon) const
{
	return	(Asura_Maths::Abs(this->e11 - xMat.e11) < fEpsilon) && 
			(Asura_Maths::Abs(this->e12 - xMat.e12) < fEpsilon) && 
			(Asura_Maths::Abs(this->e13 - xMat.e13) < fEpsilon) &&
			(Asura_Maths::Abs(this->e21 - xMat.e21) < fEpsilon) && 
			(Asura_Maths::Abs(this->e22 - xMat.e22) < fEpsilon) && 
			(Asura_Maths::Abs(this->e23 - xMat.e23) < fEpsilon) &&
			(Asura_Maths::Abs(this->e31 - xMat.e31) < fEpsilon) && 
			(Asura_Maths::Abs(this->e32 - xMat.e32) < fEpsilon) && 
			(Asura_Maths::Abs(this->e33 - xMat.e33) < fEpsilon);
}

#ifdef ASURA_TOOLS
Asura_Matrix_3x3 Asura_Matrix_3x3::GetAxiallyAlignedMatrix() const
{
	//extract the x,y and z vectors from the matrix
	Asura_Vector_3 x(e11,e21,e31);
	Asura_Vector_3 y(e12,e22,e32);
	Asura_Vector_3 z(e13,e23,e33);

	float Asura_Vector_3::*pfAxis1 = &Asura_Vector_3::x;
	float Asura_Vector_3::*pfAxis2 = &Asura_Vector_3::y;
	float Asura_Vector_3::*pfAxis3 = &Asura_Vector_3::z;

	//found the  highest axis of the x component
	if(Asura_Maths::Abs(x.*pfAxis2) > Asura_Maths::Abs(x.*pfAxis1))
	{
		float Asura_Vector_3::*pfTemp = pfAxis1;
		pfAxis1 = pfAxis2;
		pfAxis2 = pfTemp;
	}
	if(Asura_Maths::Abs(x.*pfAxis3) > Asura_Maths::Abs(x.*pfAxis1))
	{
		float Asura_Vector_3::*pfTemp = pfAxis1;
		pfAxis1 = pfAxis3;
		pfAxis3 = pfTemp;
	}

	//get the highest out of pfAxis2 and axis 3 for y
	if(Asura_Maths::Abs(y.*pfAxis3) > Asura_Maths::Abs(y.*pfAxis2))
	{
		float Asura_Vector_3::*pfTemp = pfAxis2;
		pfAxis2 = pfAxis3;
		pfAxis3 = pfTemp;
	}


	x.*pfAxis1 = (x.*pfAxis1 >0) ? 1.0f : -1.0f;
	x.*pfAxis2 = 0;
	x.*pfAxis3 = 0;

	y.*pfAxis1 = 0;
	y.*pfAxis2 = (y.*pfAxis2 >0) ? 1.0f : -1.0f;
	y.*pfAxis3 = 0;

	z = x^y;

	//build the matrix to return
	Asura_Matrix_3x3 xMat;
	xMat.e11 = x.x;
	xMat.e21 = x.y;
	xMat.e31 = x.z;
	xMat.e12 = y.x;
	xMat.e22 = y.y;
	xMat.e32 = y.z;
	xMat.e13 = z.x;
	xMat.e23 = z.y;
	xMat.e33 = z.z;

	return xMat;
}
#endif //ASURA_TOOLS

Asura_Angle Asura_Quat::ToRotation(Asura_Vector_3& xAxis) const
{
	xAxis.x	 = x;
	xAxis.y	 = y;
	xAxis.z	 = z;

	if ( xAxis.ApproximatelyEquals( Asura_Maths::ZeroVector3 ) )
	{
		xAxis.y	= 1.0f;
		return(0.0f);
	}
	else
	{
        // http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToAngle/index.htm
        xAxis.Normalise();
        const float fAngle = 2.0f * Asura_Maths::ArcCosine( w );
        return ( fAngle > Asura_Maths::Deg180 ) ? ( fAngle - Asura_Maths::Deg360 ) : fAngle;
	}
}

Asura_Angle Asura_Quat::ToRotation(Asura_AVector3& xAxis) const
{
	xAxis.SetX(x);
	xAxis.SetY(y);
	xAxis.SetZ(z);

	if ( xAxis.ApproximatelyEquals( Asura_Maths::ZeroAVector3 ) )
	{
		xAxis.SetY(1.0f);
		return(0.0f);
	}
	else
	{
        // http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToAngle/index.htm
        xAxis.Normalise();
        const float fAngle = 2.0f * Asura_Maths::ArcCosine( w );
        return ( fAngle > Asura_Maths::Deg180 ) ? ( fAngle - Asura_Maths::Deg360 ) : fAngle;
	}
}

// Make a quaternion that represents the rotation from direction A to direction B.
// NOTE: xA and xB must be normalised.
void Asura_Quat::RotationFromAToB(const Asura_Vector_3& xA, const Asura_Vector_3& xB)
{
	const float fDot = xA * xB;

	if(fDot < -0.999f)
	{
		// ART - 29/05/07 - If our 2 vectors are 180 degrees apart, we cannot use the cross-product
		// to get the perpendicular axis, we need to do something else. So we'll pick an arbitrary
		// axis to use...
		if(Asura_Maths::Abs(xA.GetY()) > 0.999f)	// ~2.5 degrees
		{
			x	= -xA.y;
			y	= 0.0f;
			z	= 0.0f;
			w	= 0.0f;
		}
		else
		{
			x	= xA.z;
			y	= 0.0f;
			z	= -xA.x;
			w	= 0.0f;
		}
	}
	else
	{
		const Asura_Vector_3 xAxis = (xA^xB);

		// KJFL - Friday, December 10, 2004 
		//
		// The actual maths leads to the following expression: 
		//
		//  xResult = Asura_Quat(xAxis.x*0.5f, xAxis.y*0.5f, xAxis.z*0.5f, (fDot + 1.0f) * 0.5f);
		//
		// but the 0.5f constants all disappear when we normalise, so they aren't actually required.

		x = xAxis.x;
		y = xAxis.y;
		z = xAxis.z;
		w = (fDot + 1.0f);
	}

	// KJFL - Friday, December 10, 2004 - in theory the vector component of the quat should be
	// a unit vector at this point, so one might be able to speed up the normalise operation
	// if required - but you'd have to take into account cases where the directions are parallel
	// and so xAxis is actually a zero vector.

	Normalise();
}

// Make a quaternion that represents the rotation from direction A to direction B.
// NOTE: xA and xB must be normalised.
void Asura_Quat::RotationFromAToB(const Asura_AVector3& xA, const Asura_AVector3& xB)
{
	// Find the angle between the two directions, and an axis perpendicular to both directions
	// that's all we need to make our quaternion.

	const float fDot = xA * xB;

	// KJFL - Friday, December 10, 2004 
	//
	// The actual maths leads to the following expression: 
	//
	//  xResult = Asura_Quat(xAxis.m_fX*0.5f, xAxis.m_fY*0.5f, xAxis.m_fZ*0.5f, (fDot + 1.0f) * 0.5f);
	//
	// but the 0.5f constants all disappear when we normalise, so they aren't actually required.

	if(fDot == -1.0f)
	{
		// ART - 29/05/07 - If our 2 vectors are 180 degrees apart, we cannot use the cross-product
		// to get the perpendicular axis, we need to do something else. So we'll pick an arbitrary
		// axis to use...
		if(Asura_Maths::Abs(xA.GetY()) > 0.999f)	// ~2.5 degrees
		{
			x	= -xA.GetY();
			y	= 0.0f;
			z	= 0.0f;
			w	= 0.0f;
		}
		else
		{
			x	= xA.GetZ();
			y	= 0.0f;
			z	= -xA.GetX();
			w	= 0.0f;
		}
	}
	else
	{
		const Asura_AVector3 xAxis(xA^xB);

		x = xAxis.GetX();
		y = xAxis.GetY();
		z = xAxis.GetZ();
		w = (fDot + 1.0f);
	}

	// KJFL - Friday, December 10, 2004 - in theory the vector component of the quat should be
	// a unit vector at this point, so one might be able to speed up the normalise operation
	// if required - but you'd have to take into account cases where the directions are parallel
	// and so xAxis is actually a zero vector.

	Normalise();
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_SphericalHarmonics_Order2_RGB
//
//////////////////////////////////////////////////////////////////////////////

void Asura_SphericalHarmonics_Order2_RGB::EvaluateInDirection( const class Asura_Vector_3& xNormal, class Asura_RGBA& xValue ) const 
{
    xValue.m_fR  = m_afRedCoefficients[0];
    xValue.m_fR += m_afRedCoefficients[1] * xNormal.x;
    xValue.m_fR += m_afRedCoefficients[2] * xNormal.y;
    xValue.m_fR += m_afRedCoefficients[3] * xNormal.z;
    xValue.m_fR = Asura_Maths::Max( xValue.m_fR, 0.0f );

    xValue.m_fG  = m_afGreenCoefficients[0];
    xValue.m_fG += m_afGreenCoefficients[1] * xNormal.x;
    xValue.m_fG += m_afGreenCoefficients[2] * xNormal.y;
    xValue.m_fG += m_afGreenCoefficients[3] * xNormal.z;
    xValue.m_fG = Asura_Maths::Max( xValue.m_fG, 0.0f );

    xValue.m_fB  = m_afBlueCoefficients[0];
    xValue.m_fB += m_afBlueCoefficients[1] * xNormal.x;
    xValue.m_fB += m_afBlueCoefficients[2] * xNormal.y;
    xValue.m_fB += m_afBlueCoefficients[3] * xNormal.z;
    xValue.m_fB = Asura_Maths::Max( xValue.m_fB, 0.0f );

    xValue.m_fA = 1.0f;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_OBB::GetSupportFeature
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

u_int Asura_OBB::GetSupportFeature(Asura_AVector3* pxSupportPoints,
								   const Asura_AVector3& xAxis) const
{
	const float fX = -xAxis * m_xOrientation.GetColumn0(); 
	const float fY = -xAxis * m_xOrientation.GetColumn1(); 
	const float fZ = -xAxis * m_xOrientation.GetColumn2(); 

	const Asura_AVector3 xXOffset = m_xOrientation.GetColumn0() * m_xHalfExtents.GetX();
	const Asura_AVector3 xYOffset = m_xOrientation.GetColumn1() * m_xHalfExtents.GetY();
	const Asura_AVector3 xZOffset = m_xOrientation.GetColumn2() * m_xHalfExtents.GetZ();

	// first check to see if we have supporting faces - this happens when xAxis is very nearly parallel with an edge.
	const float fFaceDotThreshold = (0.9999f);

	if ( Asura_Maths::Abs(fX) > fFaceDotThreshold ) // support feature is an X face
	{
		Asura_AVector3 xCentreOfFace = m_xCentre + xXOffset * Asura_Maths::Sign( fX );

		pxSupportPoints[0] = xCentreOfFace + xYOffset + xZOffset;
		pxSupportPoints[1] = xCentreOfFace + xYOffset - xZOffset;
		pxSupportPoints[2] = xCentreOfFace - xYOffset - xZOffset;
		pxSupportPoints[3] = xCentreOfFace - xYOffset + xZOffset;

		return 4;
	}
	else if ( Asura_Maths::Abs(fY) > fFaceDotThreshold ) // support feature is an Y face
	{
		Asura_AVector3 xCentreOfFace = m_xCentre + xYOffset * Asura_Maths::Sign( fY );

		pxSupportPoints[0] = xCentreOfFace + xXOffset + xZOffset;
		pxSupportPoints[1] = xCentreOfFace + xXOffset - xZOffset;
		pxSupportPoints[2] = xCentreOfFace - xXOffset - xZOffset;
		pxSupportPoints[3] = xCentreOfFace - xXOffset + xZOffset;

		return 4;
	}
	else if ( Asura_Maths::Abs(fZ) > fFaceDotThreshold ) // support feature is an Z face
	{
		Asura_AVector3 xCentreOfFace = m_xCentre + xZOffset * Asura_Maths::Sign( fZ );

		pxSupportPoints[0] = xCentreOfFace + xYOffset + xXOffset;
		pxSupportPoints[1] = xCentreOfFace + xYOffset - xXOffset;
		pxSupportPoints[2] = xCentreOfFace - xYOffset - xXOffset;
		pxSupportPoints[3] = xCentreOfFace - xYOffset + xXOffset;

		return 4;
	}

	// ok, what about any supporting edges - this happens when xAxis is perpendicular to an edge, put not parallel to one
	// (because that would be the face case).
	const float fEdgeDotThreshold = 2.6e-3f;

	if ( Asura_Maths::Abs(fX) < fEdgeDotThreshold ) // edge crosses X axis
	{
		Asura_AVector3 xCentreOfEdge = m_xCentre + xYOffset * Asura_Maths::Sign(fY) + xZOffset * Asura_Maths::Sign(fZ);

		pxSupportPoints[0] = xCentreOfEdge + xXOffset;
		pxSupportPoints[1] = xCentreOfEdge - xXOffset;

		return 2;
	}
	else if ( Asura_Maths::Abs(fY) < fEdgeDotThreshold ) // edge crosses Y axis
	{
		Asura_AVector3 xCentreOfEdge = m_xCentre + xXOffset * Asura_Maths::Sign(fX) + xZOffset * Asura_Maths::Sign(fZ);

		pxSupportPoints[0] = xCentreOfEdge + xYOffset;
		pxSupportPoints[1] = xCentreOfEdge - xYOffset;

		return 2;
	}
	else if ( Asura_Maths::Abs(fZ) < fEdgeDotThreshold ) // edge crosses Z axis
	{
		Asura_AVector3 xCentreOfEdge = m_xCentre + xYOffset * Asura_Maths::Sign(fY) + xXOffset * Asura_Maths::Sign(fX);

		pxSupportPoints[0] = xCentreOfEdge + xZOffset;
		pxSupportPoints[1] = xCentreOfEdge - xZOffset;

		return 2;
	}

	// then it must be a vert...
	pxSupportPoints[0] = m_xCentre +  xXOffset * Asura_Maths::Sign(fX) + xYOffset * Asura_Maths::Sign(fY) + xZOffset * Asura_Maths::Sign(fZ);

	return 1;
}

/// End of function 'Asura_OBB::GetSupportFeature'. ///


/// End of file. ///
