// Asura_Memheap.h
// Copyright (c) 2002 Rebellion. All rights reserved.
// TY - 30.10.07
// Rewritten to be based off of Doug Lee malloc (DLM)

// Asura_MemHeap Notes:							For design documentation see: http://gee.cs.oswego.edu/dl/html/malloc.html
// --------------------
// This class provides memory heap functionality, designed to avoid fragmentation and increase speed over the regular new/malloc
// The class may be created without any arguments, in which case the default memory allocation routines and options will be used, 
// (these utlise Asura_Allocate and Asura_Deallocate) however you may wish to provide your own memory allocator which can be done 
// via the callbacks in Asura_Memheap_Core_Details. Asura_Memheap_Core_Details is passed in as an argument to the heap and allows 
// you to specify various options such as the malloc page size and maximum allowable memory. As a simple example, if you wish to
// create a heap with a maximum allowable size of 10Mb, allocated in 1Mb pages, using the default memory allocator you would 
// simply pass in a Asura_Memheap_Core_Details class with m_uPageSize = 1024*1024 and m_iMaxPages = 10. You can also create 
// subheaps by passing another Asura_MemHeap into the constructor, options can be set by passing in an optional 
// Asura_Subheap_Core_Details class. Finally you can pass in a fixed blob of memory which will be used, along with its size.

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_MEMHEAP__H___
#define ___ASURA_MEMHEAP__H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura_List.h"
#include "Core/Asura_Multithread.h"
#include "Core/Memory/Asura_Mem.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// See comment in Asura_MemHeap::SetNoncontiguous().
#define ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES (0)

// T Y P E D E F S ///////////////////////////////////////////////////////////

typedef void*(*Asura_Memheap_AllocFunc)(size_t);			// [size_t IN]: number of bytes to allocate [RETURN]: address of allocated memory, zero on failure
typedef void*(*Asura_Memheap_TrimFunc )(size_t);			// [IN]: number of bytes to trim [RETURN]: address that has been trimmed to, may be unchanged on failure

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

class Asura_Memheap_Malloc_State;
class Asura_MallocInfo;
class Asura_Memheap_Malloc_Chunk;
class Asura_Memheap_Core_Details;
#if ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
class Asura_Subheap_Core_Details;
#endif

// C O N S T A N T S   1 ////////////////////////////////////////////////////////

const unsigned int uASURA_MEMHEAP_PAGESIZE							= 4096;			//we will request/release memory to boundries of this size
const unsigned int uASURA_MEMHEAP_DEFAULT_TRIM_THRESHOLD			= 256 * 1024;	//Trim memory when we have this much free at the top
const unsigned int uASURA_MEMHEAP_DEFAULT_TOP_PAD					= 0;			//always leave this much memory allocated
const unsigned int uASURA_MEMHEAP_ANYCHUNKS_BIT						= 1;	//Bit fields...
const unsigned int uASURA_MEMHEAP_FASTCHUNKS_BIT					= 2;
const unsigned int uASURA_MEMHEAP_PREV_INUSE						= 1;
const unsigned int uASURA_MEMHEAP_CONTIGUOUS_BIT					= 1;
const unsigned int uASURA_MEMHEAP_SIZE_BITS							= (uASURA_MEMHEAP_PREV_INUSE); //bits to exclude when checking chunk size
const unsigned int uASURA_MEMHEAP_MIN_LARGE_SIZE					= 256;	//smaller allocations than this use 'smallbins'
const unsigned int uASURA_MEMHEAP_NSMALLBINS						= 32;	//Number of small bins
const unsigned int uASURA_MEMHEAP_DEFAULT_MXFAST					= 64;   //maximum allocation size for a fastbin
const unsigned int uASURA_MEMHEAP_NBINS								= 96;	//Number of memory bins
const unsigned int uASURA_MEMHEAP_BINMAPSHIFT						= 5;	//Bit shift for bin maps
const unsigned int uASURA_MEMHEAP_SMALLBIN_WIDTH					= 8;	//size of smallbin area
const unsigned int uASURA_MEMHEAP_MAX_FAST_SIZE						= 80;
const unsigned int uASURA_MEMHEAP_MALLOC_ALIGNMENT					= 16;	//We expect our memory provider to align to this (This must be at least 2*sizeof(size_t))
const size_t       uASURA_MEMHEAP_MALLOC_ALIGN_MASK					= (uASURA_MEMHEAP_MALLOC_ALIGNMENT - 1); //mask for the above value
const unsigned int uASURA_MEMHEAP_FASTBIN_CONSOLIDATION_THRESHOLD 	= (uASURA_MEMHEAP_DEFAULT_TRIM_THRESHOLD >> 1);	//if we have consolidated chunks beyond this threshold, consolidat fastchunks too
const unsigned int uASURA_MEMHEAP_BINMAPSIZE						= (uASURA_MEMHEAP_NBINS / (1U << uASURA_MEMHEAP_BINMAPSHIFT)); //Memory allocated for binmaps
const unsigned int uASURA_MEMHEAP_NFASTBINS							= (((((uASURA_MEMHEAP_MAX_FAST_SIZE + sizeof(size_t) + uASURA_MEMHEAP_MALLOC_ALIGN_MASK) & ~uASURA_MEMHEAP_MALLOC_ALIGN_MASK)+1) >> 3) - 2);	//Rather ugly equation for the number of  fastbins

// C L A S S   D E C L A R A T I O N S   1 ///////////////////////////////////////

class Asura_Memheap_Malloc_Chunk 
{
public:

	//_______________ Functions ______________

	Asura_Memheap_Malloc_Chunk()
	{
		Asura_Mem::Set(this,0,sizeof(Asura_Memheap_Malloc_Chunk));			
	}

	//_______________ Variables ______________

	size_t						m_uPrevSize;			// Size of previous chunk (if free).  
	size_t						m_uSize;				// Size in bytes, including overhead. 

	Asura_Memheap_Malloc_Chunk* m_pxNextChunk;				// double links -- used only if free. 
	Asura_Memheap_Malloc_Chunk* m_pxPrevChunk;
};

// C O N S T A N T S   2 ////////////////////////////////////////////////////////

const unsigned int uASURA_MEMHEAP_MINSIZE = static_cast<size_t>(((sizeof(Asura_Memheap_Malloc_Chunk) + uASURA_MEMHEAP_MALLOC_ALIGN_MASK) & ~uASURA_MEMHEAP_MALLOC_ALIGN_MASK));	//The smallest allocatable size

// C L A S S   D E C L A R A T I O N S   2 ///////////////////////////////////////

class Asura_Memheap_Malloc_State 
{
public:

	//_______________ Functions ______________

	Asura_Memheap_Malloc_State()
	{
		//there is a reliance on this data being zeroed
		Asura_Mem::Set(this,0,sizeof(Asura_Memheap_Malloc_State));
	}

	//_______________ Variables ______________

	size_t							m_uMaxFastSize;							// The maximum chunk size to be eligible for fastbin (low 2 bits used as flags)

	Asura_Memheap_Malloc_Chunk* 	m_xFastbins[uASURA_MEMHEAP_NFASTBINS];	// Fastbins 
	Asura_Memheap_Malloc_Chunk* 	m_xTopChunk;							// Base of the topmost chunk -- not otherwise kept in a bin 
	Asura_Memheap_Malloc_Chunk* 	m_xRemainderChunk;						// The remainder from the most recent split of a small request 
	Asura_Memheap_Malloc_Chunk*		m_xMemoryBins[uASURA_MEMHEAP_NBINS * 2];// Normal bins

	unsigned int					m_uBinmap[uASURA_MEMHEAP_BINMAPSIZE + 1];// Bitmap of bins. Trailing zero map handles cases of largest binned size

	size_t							m_uTrimThresh;							// Tunable parameters
	size_t							m_uTopPad;		

	unsigned int					m_uPagesize;							// amount we allocate at a time
	int								m_iMaxPages;

	bool							m_bContiguous;							// Using contiguous memory?

	size_t  						m_uSystemAllocated;						// Statistics 
	size_t  						m_uMaxSystemAllocated;

	char*							m_pcEndOfMem;							//'end' of memory
};


class Asura_MallocInfo 
{
public:

	//_______________ Functions ______________

	Asura_MallocInfo()
	{
		Asura_Mem::Set(this,0,sizeof(Asura_MallocInfo));
	}

	//_______________ Variables ______________

	int		m_iSystemAllocated;    			// space allocated from system 
	int		m_iFreeChunks;  				// number of free chunks 
	int		m_iFastbinBlocks;   			// number of fastbin blocks 
	int		m_iMaxSystemAllocated;  		// highest amount of space allocated from system
	int		m_iSpaceInFreeFast;  			// space available in freed fastbin blocks 
	int		m_iTotalAllocated; 				// total allocated space 
	int		m_iTotalFree; 					// total free space 
	int		m_iKeepCost; 					// top-most, releasable (via malloc_trim) space 
	u_int	m_uRequestedAllocations;		// Total number of calls to Alloc
	u_int	m_uSystemMemCalls;				// Calls to request more system memory
};


class Asura_MemHeap
{
	// Internal Memory Allocation Function Reference
	// ---------------------------------------------
	//                     < (internal)
	//                            |           
	// Alloc(a)     Alloc(a,b)                               essentially the doug lee malloc / mallign functions
	//   \_____________/          |                      
	// - - - - - |- - - - - - - - - - ^ (public)
	//      SystemAlloc           |                           essentially the doug lee sysalloc function (wraps requests for more memory)
	//           |
	//       MoreMemory ----------|----> MemoryCallback       MoreMemory wraps the callbacks provided to Asura_Memheap for more memory
	
	public:
		
		//_______________ Functions ______________
		
		Asura_MemHeap						(void);							                                                    // for master heap, using default allocator
		Asura_MemHeap						(void* pMemory, const size_t uSize);											    // for master heap, using a fixed area of memory
		Asura_MemHeap						(const Asura_Memheap_Core_Details& xCoreDetails);							        // for master heap, using a memory allocator
#if ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
  		Asura_MemHeap						(Asura_MemHeap* pxParent, const Asura_Subheap_Core_Details* pCoreDetails = NULL);	// for sub heaps
#endif
		~Asura_MemHeap						();

		void*	Alloc						(size_t uBytes);						// allocate memory
		void*   Alloc						(size_t uBytes, size_t alignment);		// allocate memory aligned to a specific value
		void    Free						(void* pAddress);						// free allocated memory 		

		void*	Realloc						(void *pvMemory, size_t uBytes);		// reallocate memory with adjusted size

		size_t	GetFreeAllocatedSpace		() const;								//How many bytes are free to the user in the current allocated memory
		
		void	PrintMallocInfo				() const;								//Print out the malloc info to the debugger 
																					//NOTE: do NOT use this to check if you have outstanding allocations, memheap may be using part of the allocated memory
																					//for its own internal bookeeping and will not account for it when reporting the 'free space'

        Asura_MallocInfo GetMallocInfo( void ) const;

        bool	IsContiguous				() const;	

		#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
		int	GetNumIntegrityFailures			();										//Read and reset the number of integrity failures recorded (for unit tests)
		#endif

		void	Wipe						();

		void	EnableThreadSafety			();
		void	DisableThreadSafety			();
		void	DisableThreadSafetyWarning	();										// if you only call memheap functions from within your own mutex then you are threadsafe 
																					// without enabling the internal mutex, so call this to disable the warning.

	private:
    
		//_______________ Functions ______________

		void					InitData					();	

		void					SetCoreDetails				(const Asura_Memheap_Core_Details* pxDetails) ;
#if ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
  		void					SetCoreDetails				(const Asura_Subheap_Core_Details* pxDetails) ;
#endif
		
		void					SetPageSize					(unsigned int	uPageSize);		//How much memory is requested/trimmed at a time
		void					SetMaxPages					(int			iMaxPages);		//The maximum number of pages to allow requests for
		void					SetTrimThreshold			(size_t			uTrim);			//The amount of free space available before we will try trimming
		void					SetTopPad					(size_t			uPad);			//The amount of memory to always keep allocated (i.e. not trim) within the memheap

		void					MallocInitState				();
		void					MallocConsolidate			();
		
		void					InternalAlloc				(u_int uBytes);

		void*					SystemAlloc					(size_t nb);
		int						SystemTrim					(size_t pad);

		void					UpdateMallinfo				() const;

		inline u_int			GetSizeFromRequest			(size_t uRequest) const;

		inline	void*			MoreMemory					(size_t uSize);	//Get more memory for the heap

		#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
		void					Checks_InUseChunk			(Asura_Memheap_Malloc_Chunk* pChunk) const;
		void					Checks_MallocState			() const;
		void					Checks_RemallocedChunk		(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize) const;
		void					Checks_MallocedChunk		(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize) const;
		void					Checks_Chunk				(Asura_Memheap_Malloc_Chunk* pChunk) const;
		void					Check_FreeChunk				(Asura_Memheap_Malloc_Chunk* pChunk) const;

		mutable int				m_iIntegrityFailures;
		#endif

		inline void 			SetFastchunks				();
		inline bool 			HaveFastchunks				() const;
		inline bool 			HaveAnyChunks				() const;
		inline void 			SetAnyChunks				();
		inline void 			ClearFastChunks				();
		inline void				SetInuseBitAtOffset			(Asura_Memheap_Malloc_Chunk* pChunk, size_t uOffset);
		inline int				GetInuseBitAtOffset			(Asura_Memheap_Malloc_Chunk* pChunk, size_t uOffset) const;
		inline void				SetMaxFast					(size_t s);
		inline bool				InSmallbinRange				(size_t u) const;
		inline bool				ChunkInUse					(Asura_Memheap_Malloc_Chunk* p) const;
		inline void				UnlinkChunk					(const Asura_Memheap_Malloc_Chunk* pxChunk);
		
		inline void				SetNoncontiguous			();
		inline void				SetContiguous				();

		inline Asura_Memheap_Malloc_Chunk*	NextChunk				(Asura_Memheap_Malloc_Chunk* pChunk) const;
		inline Asura_Memheap_Malloc_Chunk*	PrevChunk				(Asura_Memheap_Malloc_Chunk* pChunk) const;
		inline Asura_Memheap_Malloc_Chunk*	GetChunkFromMemory		(void* pMem) const;
		inline void*						GetMemoryFromChunk		(void* pBase) const;
		inline size_t						ChunkSize				(Asura_Memheap_Malloc_Chunk* pChunk) const;
		inline bool							PrevChunkInUse			(Asura_Memheap_Malloc_Chunk* pChunk) const;
		inline Asura_Memheap_Malloc_Chunk*	GetChunkAtOffset		(Asura_Memheap_Malloc_Chunk* pChunk, size_t uOffset) const;
		inline Asura_Memheap_Malloc_Chunk*	GetUnsortedChunks		() const;
		
		inline u_int						BlockFromIndex			(u_int u) const;
		inline u_int						BitFromIndex			(u_int u) const;

		inline Asura_Memheap_Malloc_Chunk*	GetNextBin				(Asura_Memheap_Malloc_Chunk* b) const;
		inline Asura_Memheap_Malloc_Chunk*	GetBinAt				(u_int u) const;
		inline u_int						GetBinIndex				(size_t u) const;
		inline u_int						GetBinmap				(u_int u) const;
		inline void							MarkBin					(size_t u);
		inline u_int						GetFastbinIndex			(size_t u) const;
		inline u_int						GetLargebinIndex		(size_t u) const;
		inline u_int						GetSmallbinIndex		(size_t u) const;

		inline void				SetHeadSize					(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize);
		inline void				SetHead						(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize);
		inline void				SetFoot						(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize);

		inline bool				IsAligned					(void* pMem) const;

		inline size_t			GetMaxFastSize				() const;

		void* DefaultMemoryAllocator						(size_t uSize);				//our default memory allocator
		
		inline void				MutexAcquire				() const;
		inline void				MutexRelease				() const;

		//_______________ Variables ______________

		char*					    m_pcBaseMem;			//Our memory pointer if we were constructed from memory
		size_t						m_uBaseMemSize;

  		Asura_MemHeap*				m_pxParent;				//Our parent if we were constructed from a heap

		Asura_Memheap_AllocFunc		m_pfnAllocFunction;		//Our allocation functions if we were constructed with functions
		Asura_Memheap_TrimFunc		m_pfnTrimFunction;

		Asura_Memheap_Malloc_State	m_xMallocState;
		mutable Asura_MallocInfo	m_xMallinfo;	

		Asura_List<char>			m_xDefaultAllocs;

		bool						m_bThreadSafetyEnabled;
		bool						m_bThreadSafetyWarningDisabled;
		mutable Asura_Mutex			m_xMutex;

		mutable bool				m_bUpdateMallinfo;		//mallinfo needs updating?

		Asura_ThreadID				m_uCreationThread;		//the thread we were created in
};

class Asura_Memheap_Core_Details
{
	//The memheap core is the mechanism by which Asura_Memheap gets more system memory;
	//there are defaults provided but if possible use platform/game specific functions.
	//If at all possible have the allocator return contiguous memory as the memheap operates
	//much more efficiently that way.
	public:

        Asura_Memheap_Core_Details( void )
        {
            m_pfnAllocator  = NULL;
            m_pfnTrimmer    = NULL;
            m_bIsContiguous = true;
            m_uPageSize		= uASURA_MEMHEAP_PAGESIZE;
            m_iMaxPages		= -1;
            m_uTrim			= uASURA_MEMHEAP_DEFAULT_TRIM_THRESHOLD;
            m_uPad			= uASURA_MEMHEAP_DEFAULT_TOP_PAD;
        }

        Asura_Memheap_Core_Details( Asura_Memheap_AllocFunc pfnAllocator,
                                    Asura_Memheap_TrimFunc pfnTrimmer,
                                    bool bIsContiguous,
                                    u_int uPageSize,
                                    int iMaxPages,
                                    size_t uTrim,
                                    size_t uPad )
        {
            m_pfnAllocator  = pfnAllocator;
            m_pfnTrimmer    = pfnTrimmer,
            m_bIsContiguous = bIsContiguous;
            m_uPageSize     = uPageSize;
            m_iMaxPages     = iMaxPages;
            m_uTrim         = uTrim;	
            m_uPad          = uPad;
        }
                                    
		Asura_Memheap_AllocFunc	m_pfnAllocator;		//pointer to function for allocating memory (pass NULL to use default)							
		Asura_Memheap_TrimFunc	m_pfnTrimmer;		//pointer to function for trimming memory							

		bool	m_bIsContiguous;					//Set to true if your functions will provide contiguous memory (dlm can technically handle this being misreported but it is wasteful)

		u_int	m_uPageSize;						//Allocate this much at a time from the allocation function
		int		m_iMaxPages;						//Only allow the allocation of this many pages (-1 no limit)
		size_t	m_uTrim;							//This much space must be free before we trim
		size_t	m_uPad;								//always have this much space allocated
};

#if ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
class Asura_Subheap_Core_Details
{
  	//As above for subheaps
  	public:

  		Asura_Subheap_Core_Details()
  		{
  			m_uPageSize = uASURA_MEMHEAP_PAGESIZE;
  			m_iMaxPages = -1;
  			m_uPad		= uASURA_MEMHEAP_DEFAULT_TOP_PAD;
  		}

  		u_int	m_uPageSize;						//Allocate this much at a time from the allocation function
  		int		m_iMaxPages;						//Only allow the allocation of this many pages (-1 no limit)
  		size_t	m_uPad;								//always have this much space allocated
};
#endif

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_MEMHEAP_H___

