//////////////////////////////////////////////////////////////////////////////
//
// Asura_Multithread.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_MULTITHREAD_H___
#define ___ASURA_MULTITHREAD_H___ 1

#include "Core/Asura_Platform_MultithreadTypes.h"

#include "Debugging/Asura_Assert.h"

// E N U M E R A T I O N S ///////////////////////////////////////////////////

enum ASURA_MULTITHREAD_STACK_SIZE
{
	ASURA_MULTITHREAD_STACK_SIZE_SMALL = 0,
	ASURA_MULTITHREAD_STACK_SIZE_MEDIUM,
	ASURA_MULTITHREAD_STACK_SIZE_LARGE,

	ASURA_MULTITHREAD_STACK_NUM_SIZES
};

// T Y P E D E F S ///////////////////////////////////////////////////////////
typedef void (*Asura_ThreadFunc)(const void*);
// each platform should provide a typedef Asura_ThreadID, and that type is
// assumed to allow the == and != operators.

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////
class Asura_Thread_System
{
public:
    // Returns the number of threads the hardware can support independently.
    // Creating more software threads than this will not provide a greater work rate.
    static u_int GetNumHardwareThreads() { return Platform_GetNumHardwareThreads(); }

	// returns true iff thread was successfully created.
	static bool CreateThread(	const char* szThreadName
							,	Asura_ThreadFunc pfnThreadFunc
							,	const void* pUserData
							,	const ASURA_MULTITHREAD_STACK_SIZE eStackSize = ASURA_MULTITHREAD_STACK_SIZE_LARGE)
	{
		return Platform_CreateThread( szThreadName, pfnThreadFunc, pUserData, eStackSize );
	}

	static void YieldCurrentThread()													{ Platform_YieldCurrentThread(); }
	static void DelayCurrentThread(float fTimeInSeconds, bool bAllowCallbacks = true)	{ Platform_DelayCurrentThread(fTimeInSeconds, bAllowCallbacks); }

	enum 
	{
		DELAY_DONT_ALLOW_CALLBACKS = 0,
		DELAY_ALLOW_CALLBACKS
	};

	static Asura_ThreadID GetCurrentThreadID() { return Platform_GetCurrentThreadID(); }
	static Asura_ThreadID GetInvalidThreadID() { return Platform_GetInvalidThreadID(); }

	//This function is provided to allow projects to override the default stack size enumerations. Sensible defaults have been provided for
	//each platform, however if a project wishes to use particularily large or small stacks this function can be called to override them. Bear
	//in mind this is platform specific, and naturally a value which is too small may result in stack overflows so be careful.
	static void RedefineStackSize (const ASURA_MULTITHREAD_STACK_SIZE eSizeEnumeration, const u_int uOverrideValue)
	{
		Asura_Assert(eSizeEnumeration != ASURA_MULTITHREAD_STACK_NUM_SIZES, "Invalid stack size enumeration passed to RedefineStackSize");

		s_auStackSizes[eSizeEnumeration] = uOverrideValue;
	}
	
private:

    static u_int Platform_GetNumHardwareThreads();
	static bool Platform_CreateThread( const char* szThreadName, Asura_ThreadFunc pfnThreadFunc, const void* pUserData, const ASURA_MULTITHREAD_STACK_SIZE eStackSize );
	static void Platform_YieldCurrentThread();
	static void Platform_DelayCurrentThread(float fTimeInSeconds, bool ballowCallbacks); 
	static Asura_ThreadID Platform_GetCurrentThreadID();
	static Asura_ThreadID Platform_GetInvalidThreadID();
	
	static u_int s_auStackSizes[ASURA_MULTITHREAD_STACK_NUM_SIZES];	//This must be defined in Asura_Platform_Multithread
};


// Recursive mutex. each call to Lock() must be matched with a call to Unlock().
// Only the first Lock() and last Unlock() cause the platform _Lock and _Unlock()
// functions to be called for the same thread so the platform implementation doesn't 
// have to handle recursive mutexes (this helps on PSP and PS2).
class Asura_Mutex : public Asura_Platform_MutexData
{
public:

	Asura_Mutex()
	{
		m_uLockCount = 0;
		m_xLockOwnerThreadID = Asura_Thread_System::GetInvalidThreadID();
	}

	void Lock() 
	{ 
		const Asura_ThreadID xThisThreadID = Asura_Thread_System::GetCurrentThreadID();
		
		if ( m_xLockOwnerThreadID == xThisThreadID )
		{
			++m_uLockCount;
		}
		else
		{
			Platform_Lock();
			m_xLockOwnerThreadID = xThisThreadID;
			++m_uLockCount;
		}
	}
	
	void Unlock()
	{
		Asura_Assert(m_uLockCount, "mutex not locked, mismatched lock/unlock calls?");
		if(!m_uLockCount)
		{
			return;
		}

		Asura_Assert(m_xLockOwnerThreadID == Asura_Thread_System::GetCurrentThreadID(), "mutex unlock called by different thread to locking thread");

		--m_uLockCount;
		if (!m_uLockCount)
		{
			m_xLockOwnerThreadID = Asura_Thread_System::GetInvalidThreadID();
			Platform_Unlock();
		}		
	}

	// Returns the number of unlocks to free the mutex
	int MultiUnlock()
	{
		int iNumUnlocks=0;
		while(m_uLockCount)
		{
			Unlock();
			iNumUnlocks++;
		}
		return iNumUnlocks;
	}

	// Performs the given number of locks
	void MultiLock(int iNumLocks)
	{
		while(iNumLocks)
		{
			Lock();
			iNumLocks--;
		}
	}


private:

	// private functions
	void Platform_Lock();
	void Platform_Unlock();

	// private data (only modifed between calls to Platform _Lock and _Unlock)
	volatile Asura_ThreadID m_xLockOwnerThreadID;
	volatile u_int m_uLockCount;
};


// MGH 29/1/2008: Lock the specified mutex and automatically unlock it when we go out of scope.
class Asura_ScopedMutex
{
public:
	Asura_ScopedMutex(Asura_Mutex& xMutex) : m_xMutex(xMutex) { m_xMutex.Lock(); }
	~Asura_ScopedMutex() { m_xMutex.Unlock(); }

private:
	// Prevent copying.
	Asura_ScopedMutex(const Asura_ScopedMutex& xOther);
	Asura_ScopedMutex& operator= (const Asura_ScopedMutex& xOther);

private:
	Asura_Mutex& m_xMutex;
};


// Asura_SyncEvent allows threads to wait for another thread to tell them its OK 
// to proceed with something. If the event if signalled it's OK to carry on, if
// it isn't then the thread will sleep until the event is signalled.
class Asura_SyncEvent : public Asura_Platform_SyncEvent
{
public:

	Asura_SyncEvent( const char* szName, bool bStartsSignalled )
	: Asura_Platform_SyncEvent( szName, bStartsSignalled )
	{
		m_bSignalled = bStartsSignalled;
	}

	bool IsSignalled() const { return m_bSignalled; }
	// KOR - 12/09/2006 - made WaitForSignal non-const, as
	// GameCube uses direct data access instead of handles,
	// so the platform data does get altered.
	void WaitForSignal() { return Platform_WaitForSignal(); }
	void SetSignal() 
	{ 
		Asura_Assert(!m_bSignalled, "Signal is already set - should not be set a second time.");
		m_bSignalled = true; 
		Platform_SetSignal(); 
	}
	void ClearSignal()
	{ 
		Asura_Assert(m_bSignalled, "Signal is already cleared - should not be cleared a second time.");
		Platform_ClearSignal(); 
		m_bSignalled = false;
	}

private:

	// private functions
	void Platform_WaitForSignal();
	void Platform_SetSignal();
	void Platform_ClearSignal();

	// private data
	bool m_bSignalled;
};


// a semaphore is signaled when its count is greater than zero and non-signaled when it is zero.
//#define ASURA_DISABLE_SEMAPHORE 1 // e.g. to force a unit test to break
class Asura_Semaphore : public Asura_Platform_Semaphore
{
public:

	Asura_Semaphore( const char* szName, u_int uInitialCount, u_int uMaxCount )
		: Asura_Platform_Semaphore( szName, uInitialCount, uMaxCount )
	{
	}

	// Wait() waits for semaphore to be non-zero, and decreases it.
	void Wait()
	{
#ifndef ASURA_DISABLE_SEMAPHORE
		Platform_Wait();
#endif
	}

	// TryWait() if the semaphore is non-zero, if returns true and decreases it otherwise returns false.
	bool TryWait()
	{
#ifdef ASURA_DISABLE_SEMAPHORE
		return false;
#else
		return Platform_TryWait();
#endif
	}

	// Signal() increments the semaphore. Returns false if max count would be exceeded.
	bool Signal()
	{
#ifdef ASURA_DISABLE_SEMAPHORE
		return false;
#else
		return Platform_Signal();
#endif
	}

private:

	void Platform_Wait();
	bool Platform_TryWait();
	bool Platform_Signal();

};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_MULTITHREAD_H___
