//////////////////////////////////////////////////////////////////////////////
//
// Asura_NameList.h
//
// Created by Mike Healey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Manages a simple list of Asura_Strings, allowing duplicates to avoid 
// reference counting and dynamically resizing as appropriate. Written 
// specifically for use by cutscene code and the slow bits should only be 
// used by the Tools.
//
// Be aware of these issues before using this for game code.
//
// The whole thing needs sorting out at some point, preferably a reference-
// counted NameTree or some such which will avoid duplicates and therefore
// save us some memory.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_NAMELIST_H___
#define ___ASURA_NAMELIST_H___ 1

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

#define ASURA_NAMELIST_NUM_INITIAL_ENTRIES 8

class Asura_NameList
{
public:
    Asura_NameList(const bool bFilenameList = true)
    {
        m_iNumEntries = 0;
        m_iNumEntriesAllocated = ASURA_NAMELIST_NUM_INITIAL_ENTRIES;
		m_bFilenameList = bFilenameList;
        m_pxList = new Asura_String[ASURA_NAMELIST_NUM_INITIAL_ENTRIES];
    }

    ~Asura_NameList()
    {
        delete [] m_pxList;
    }

    void Clear()
    {
        for ( int i = 0; i < m_iNumEntriesAllocated; ++i )
        {
            if ( !m_pxList[ i ].IsNull() )
            {
                m_pxList[ i ].MakeNull();
            }
        }
        m_iNumEntries = 0;
    }

    // (MH - 23.07.2002) slow functions - for tools use
    int AddName(const Asura_String& xName)
    {
		if(xName.IsNull()) return -1;

        if (m_iNumEntries < m_iNumEntriesAllocated)
        {
            //search for a free slot
			for(int i=0;i<m_iNumEntriesAllocated;i++)
			{
				if(m_pxList[i].IsNull())
				{
					m_iNumEntries++;
					m_pxList[i] = xName;
					return i;
				}
			}
			Asura_Assert(0,"Failed to find free slot in name list");
			return -1;
        }
        else
        {
            // reallocate the list and copy the old list across
            m_iNumEntriesAllocated *= 2;
            Asura_String * NewList = new Asura_String[m_iNumEntriesAllocated];
            for (int i = 0; i < m_iNumEntries; i++)
            {
                NewList[i] = m_pxList[i];
            }
            NewList[m_iNumEntries++] = xName;
            delete [] m_pxList;
            m_pxList = NewList;
            return m_iNumEntries - 1;
        }
    }

    void RemoveName(int index)
    {
        Asura_Assert(0 <= index && index < m_iNumEntriesAllocated, "invalid index");
		if (0 <= index && index < m_iNumEntriesAllocated)
        {
            Asura_Assert(!m_pxList[index].IsNull(),"");
            if (!m_pxList[index].IsNull())
            {
                m_pxList[index].MakeNull();
		        m_iNumEntries--;
            }
        }
    }

    Asura_String& GetName(int index)
    {
        Asura_Assert(0 <= index && index < m_iNumEntriesAllocated, "invalid index, using first name found");
        if (0 <= index && index < m_iNumEntriesAllocated)
        {
            return m_pxList[index];
        }
        else
        {
            return m_pxList[0];
        }
    }
    const Asura_String& GetName( int index ) const
    {
        Asura_Assert(0 <= index && index < m_iNumEntriesAllocated, "invalid index, using first name found");
        if (0 <= index && index < m_iNumEntriesAllocated)
        {
            return m_pxList[index];
        }
        else
        {
            return m_pxList[0];
        }
    }

    int GetNumEntries( void ) const
    {
        // (MH - 22.10.2002) changed this from NumEntries, since callers are expected
        // to run through all entries skipping the blank ones.
        return m_iNumEntriesAllocated;
    }
    
    int GetNumUsedEntries( void ) const
    {
        return m_iNumEntries;
    }
    
    // (MH - 27.09.2002) slow search for entry in list, returns -1 if not found
    int GetIndexOfName(const Asura_String& xName)
    {
		if(m_bFilenameList)
		{
			for (int i = 0; i < m_iNumEntriesAllocated; i++)
			{
				if(!m_pxList[i].IsNull())
				{
					// ignore case
					if (GetName(i).EqualsIgnoringExtension(xName))
					{
						return i;
					}
				}
			}
		}
		else
		{
			for (int i = 0; i < m_iNumEntriesAllocated; i++)
			{
				if(!m_pxList[i].IsNull())
				{
					// ignore case
					if (GetName(i) == xName)
					{
						return i;
					}
				}
			}
		}
        return -1;
    }

    int GetChunkSize()
    {
        int chunk_size = sizeof(int)*2;
        for (int i = 0; i < m_iNumEntriesAllocated; i++)
        {
             if(!m_pxList[i].IsNull())
			 {
             	chunk_size += sizeof(int);
             	ASURA_CHUNK_SIZE_STRING(chunk_size, GetName(i));
			 }
        }
        return chunk_size;
    }

    void SaveToChunk(char*& pcChunkBuffer)
	{
		//get the index of the highest string
		int iMaxEntry;
		int iNumOutput = 0;
		for(iMaxEntry = m_iNumEntriesAllocated-1;iMaxEntry>=0;iMaxEntry--)
		{
			if(!m_pxList[iMaxEntry].IsNull()) break;
		}
	    
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_iNumEntries, int);
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iMaxEntry, int);
		for (int i = 0; i < m_iNumEntriesAllocated; i++)
		{
			if(!m_pxList[i].IsNull())
			{
				ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,i,int);
				ASURA_CHUNK_FILL_STRING(pcChunkBuffer, GetName(i))
				iNumOutput++;
			}
		}
		Asura_Assert(iNumOutput==m_iNumEntries,"");
	}

	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const
	{
		int iMaxEntry;
		int iNumOutput = 0;

		//get the index of the highest string
		for(iMaxEntry = m_iNumEntriesAllocated-1;iMaxEntry>=0;iMaxEntry--)
		{
			if(!m_pxList[iMaxEntry].IsNull()) break;
		}
	    
		xStream << m_iNumEntries;
		xStream << iMaxEntry;
		for (int i = 0; i < m_iNumEntriesAllocated; i++)
		{
			if(!m_pxList[i].IsNull())
			{
				xStream << i;
				xStream << GetName(i);
				iNumOutput++;
			}
		}
		Asura_Assert(iNumOutput==m_iNumEntries,"");
	}

    void LoadFromChunk(Asura_Handle iFileHandle)
    {
        int maxEntry;
		int index;
        ASURA_CHUNK_FILEEXTRACT(iFileHandle, m_iNumEntries, int);
        ASURA_CHUNK_FILEEXTRACT(iFileHandle, maxEntry, int);
		if(maxEntry>=m_iNumEntriesAllocated)
		{
			m_iNumEntriesAllocated = maxEntry+1;
			delete [] m_pxList;
			m_pxList = new Asura_String[m_iNumEntriesAllocated];
		}

		for(int i=0;i<m_iNumEntries;i++)
        {
           	ASURA_CHUNK_FILEEXTRACT(iFileHandle,index,int);
			Asura_Assert(index>=0 && index<m_iNumEntriesAllocated,"");
            ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle, m_pxList[index]);
       	}
    }

    static void SkipChunk(Asura_Handle iFileHandle)
    {
        int numEntries,maxEntry,index;
        ASURA_CHUNK_FILEEXTRACT(iFileHandle, numEntries, int);
        ASURA_CHUNK_FILEEXTRACT(iFileHandle, maxEntry, int);
        for (int i = 0; i < numEntries; i++)
        {
           	ASURA_CHUNK_FILEEXTRACT(iFileHandle,index,int);
            Asura_FileAccess::ReadString(iFileHandle);
        }
    }

    void TidyFilenames()
    {
		// JAT 04/10/07 - loop over all the entries and fix up all the names
        for (int i = 0; i < GetNumEntries(); i++)
        {
             if( true == m_pxList[i].IsNull())
			 {
				continue;
			 }
			 
			 Asura_String xName = m_pxList[i];
			 xName.TidyFilename();
			 // JAT 04/10/07 remove any leading ./ or .\ as well
			 if ( xName.StartsWith( "./" ) )
			 {
				 m_pxList[i] = xName.GetString() + 2;
			 }
			 if ( xName.StartsWith( ".\\" ) )
			 {
				 m_pxList[i] = xName.GetString() + 2;
			 }
        }
    }

private:
    Asura_String* m_pxList;
    int           m_iNumEntries;
    int           m_iNumEntriesAllocated;
	bool          m_bFilenameList;
};

#endif // #ifndef ___ASURA_NAMELIST_H___
