//
// Asura_OrientedVolume.cpp
//
// Created by Richard Rice - 25/01/05
// 
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Core/Asura_OrientedVolume.h"

bool Asura_OrientedVolume::Contains(const Asura_Vector_3& xPosition) const
{
	//see if the orientation is being used
	if(m_xOrient.w == 1.0f)
	{
		//nope , can just do simple test
		return m_xBox.Contains(xPosition);
	}
	else
	{
		//need to get position relative to volume
		Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
		Asura_Vector_3 xRelative = xPosition - xCentre;

		//and rotate into the space of the volume
		Asura_Quat xOrientInverse = m_xOrient;
        xOrientInverse.Invert();
		xRelative = xOrientInverse.RotateVector(xRelative);
		
		return m_xBox.Contains(xRelative+xCentre);
	}
}

float Asura_OrientedVolume::GetSqdMinDistFromPoint(const Asura_Vector_3& xPosition) const
{
	if (m_xOrient.w == 1.0f)
	{
		return m_xBox.GetSqdMinDistFromPoint(xPosition);
	}
	else
	{
		//need to get position relative to volume
		Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
		Asura_Vector_3 xRelative = xPosition - xCentre;

		//and rotate into the space of the volume
		Asura_Quat xOrientInverse = m_xOrient;
        xOrientInverse.Invert();
		xRelative = xOrientInverse.RotateVector(xRelative);
		
		return m_xBox.GetSqdMinDistFromPoint(xRelative+xCentre);
	}
}

void Asura_OrientedVolume::CalcMinVectorIntoVolume( const Asura_Vector_3& xPosition, Asura_Vector_3& xVector ) const
{
    if ( m_xOrient.w == 1.0f )
    {
        m_xBox.CalcMinVectorIntoBox( xPosition, xVector );
    }
    else
    {
		// need to get position relative to volume
		Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
		Asura_Vector_3 xRelative = xPosition - xCentre;

		// and rotate into the space of the volume
		Asura_Quat xOrientInverse = m_xOrient;
        xOrientInverse.Invert();
		xRelative = xOrientInverse.RotateVector(xRelative);

        m_xBox.CalcMinVectorIntoBox( xRelative+xCentre, xVector );

        // and finally rotate the result back into world space
        xVector = m_xOrient.RotateVector( xVector);
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool Asura_OBBIntersectTest(const Asura_Vector_3 &xHalfLengthsForA, const Asura_OBB_AdjustedForIntersectionTest &xB)
{
	// (AW - 09.02.2004) - For an explanation of this algorithm, see ch11.12.7 in Geometric Tools
	// for Computer Graphics by Schneider and Eberly.
	// (AW - 09.02.2004) - First, test along the 3 edges of A.
	for (int i=0;i<3;i++)
	{
		float fRa = xHalfLengthsForA[i];
		float fRb = xB.xHalfLengths[0] * Asura_Maths::Abs(xB.axOrientationUnitVectors[0][i]) +
					xB.xHalfLengths[1] * Asura_Maths::Abs(xB.axOrientationUnitVectors[1][i]) +
					xB.xHalfLengths[2] * Asura_Maths::Abs(xB.axOrientationUnitVectors[2][i]);
		if (Asura_Maths::Abs(xB.xCentre[i]) > fRa + fRb)
		{
			return false;
		}
	}
	// (AW - 09.02.2004) - Now, test along the 3 edges of B.
	for (int i=0;i<3;i++)
	{
		float fRa = xHalfLengthsForA.x * Asura_Maths::Abs(xB.axOrientationUnitVectors[i].x) + 
					xHalfLengthsForA.y * Asura_Maths::Abs(xB.axOrientationUnitVectors[i].y) + 
					xHalfLengthsForA.z * Asura_Maths::Abs(xB.axOrientationUnitVectors[i].z);
		float fRb = xB.xHalfLengths[i];
		if (Asura_Maths::Abs(xB.xCentre.x * xB.axOrientationUnitVectors[i].x +
							 xB.xCentre.y * xB.axOrientationUnitVectors[i].y + 
							 xB.xCentre.z * xB.axOrientationUnitVectors[i].z) > fRa + fRb)
		{
			return false;
		}
	}
	// (AW - 09.02.2004) - Now the 9 combinations of edges between A and B.
	// (AW - 09.02.2004) - There are potential optimisations to be made by unwrapping these loops,
	// because we know that the first BB is axis-aligned. If this function proves too slow, 
	// then this is where the most gains could be made.
    // (MP - 23.05.2006) Unrolled inner loop to eliminate use of matrix[]
    // returning vector.
	for (int i=0;i<3;i++)
    {
        Asura_Vector_3 xAxisToTest;
        float fRa, fRb;

        xAxisToTest = Asura_Vector_3(1.0f, 0.0f, 0.0f) ^ xB.axOrientationUnitVectors[0];
        fRa = xHalfLengthsForA.x * Asura_Maths::Abs(Asura_Vector_3(1.0f, 0.0f, 0.0f) * xAxisToTest) + 
              xHalfLengthsForA.y * Asura_Maths::Abs(Asura_Vector_3(0.0f, 1.0f, 0.0f) * xAxisToTest) +
              xHalfLengthsForA.z * Asura_Maths::Abs(Asura_Vector_3(0.0f, 0.0f, 1.0f) * xAxisToTest);
        fRb = xB.xHalfLengths.x * Asura_Maths::Abs(xB.axOrientationUnitVectors[0] * xAxisToTest) +
              xB.xHalfLengths.y * Asura_Maths::Abs(xB.axOrientationUnitVectors[1] * xAxisToTest) +
              xB.xHalfLengths.z * Asura_Maths::Abs(xB.axOrientationUnitVectors[2] * xAxisToTest);
        if (xB.xCentre * xAxisToTest > fRa + fRb)
        {
            return false;
        }

        xAxisToTest = Asura_Vector_3(0.0f, 1.0f, 0.0f) ^ xB.axOrientationUnitVectors[1];
        fRa = xHalfLengthsForA.x * Asura_Maths::Abs(Asura_Vector_3(1.0f, 0.0f, 0.0f) * xAxisToTest) + 
              xHalfLengthsForA.y * Asura_Maths::Abs(Asura_Vector_3(0.0f, 1.0f, 0.0f) * xAxisToTest) +
              xHalfLengthsForA.z * Asura_Maths::Abs(Asura_Vector_3(0.0f, 0.0f, 1.0f) * xAxisToTest);
        fRb = xB.xHalfLengths.x * Asura_Maths::Abs(xB.axOrientationUnitVectors[0] * xAxisToTest) +
              xB.xHalfLengths.y * Asura_Maths::Abs(xB.axOrientationUnitVectors[1] * xAxisToTest) +
              xB.xHalfLengths.z * Asura_Maths::Abs(xB.axOrientationUnitVectors[2] * xAxisToTest);
        if (xB.xCentre * xAxisToTest > fRa + fRb)
        {
            return false;
        }

        xAxisToTest = Asura_Vector_3(0.0f, 0.0f, 1.0f) ^ xB.axOrientationUnitVectors[2];
        fRa = xHalfLengthsForA.x * Asura_Maths::Abs(Asura_Vector_3(1.0f, 0.0f, 0.0f) * xAxisToTest) + 
              xHalfLengthsForA.y * Asura_Maths::Abs(Asura_Vector_3(0.0f, 1.0f, 0.0f) * xAxisToTest) +
              xHalfLengthsForA.z * Asura_Maths::Abs(Asura_Vector_3(0.0f, 0.0f, 1.0f) * xAxisToTest);
        fRb = xB.xHalfLengths.x * Asura_Maths::Abs(xB.axOrientationUnitVectors[0] * xAxisToTest) +
              xB.xHalfLengths.y * Asura_Maths::Abs(xB.axOrientationUnitVectors[1] * xAxisToTest) +
              xB.xHalfLengths.z * Asura_Maths::Abs(xB.axOrientationUnitVectors[2] * xAxisToTest);
        if (xB.xCentre * xAxisToTest > fRa + fRb)
        {
            return false;
        }
    }
	return true;
}

bool Asura_OrientedVolume::Overlaps(const Asura_Quat& xOrient, const Asura_Bounding_Box xOBB, const Asura_Bounding_Box& xAABB)
{
	if(xOrient.w == 1.0f)
	{
		//just do a straight AABB intersection test
		return xOBB.Intersects(xAABB);
	}


	// (AW - 09.02.2004) - Convert the OBB into the form our intersection test would like.
	Asura_Vector_3 xCentre = xAABB.CalculateCentre();
	Asura_Vector_3 xBoundingBoxAHalfLengths(xAABB.MaxX - xCentre.x, xAABB.MaxY - xCentre.y, xAABB.MaxZ - xCentre.z);
	// (AW - 09.02.2004) - Remember, the orientation and position of BB2 has to be relative to
	// BB1. (this is slightly easier because BB1 is already axis-aligned.)
	Asura_OBB_AdjustedForIntersectionTest xB;
	xB.xCentre = xOBB.CalculateCentre();
	xB.xHalfLengths.x = xOBB.MaxX - xB.xCentre.x;
	xB.xHalfLengths.y = xOBB.MaxY - xB.xCentre.y;
	xB.xHalfLengths.z = xOBB.MaxZ - xB.xCentre.z;
	xB.xCentre -= xCentre;
	xB.axOrientationUnitVectors[0] = xOrient.RotateVector(Asura_Vector_3(1.0f, 0.0f, 0.0f));
	xB.axOrientationUnitVectors[1] = xOrient.RotateVector(Asura_Vector_3(0.0f, 1.0f, 0.0f));
	xB.axOrientationUnitVectors[2] = xOrient.RotateVector(Asura_Vector_3(0.0f, 0.0f, 1.0f));
	return Asura_OBBIntersectTest(xBoundingBoxAHalfLengths, xB);
}
