
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Reference_VectorMaths.h
// Author : Andy Tate
// Date   : 28-Jun-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////
//
// NOTE: This code is the reference code which should be used when starting
// ~~~~~ writing the maths libraries on new platforms. It contains the "correct"
// versions of the aligned-maths classes using standard C++ instead of the 
// platform-specific intrinsics/assembler.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_REFERENCE_VECTORMATHS_H__
#define __ASURA_REFERENCE_VECTORMATHS_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

// Asura Platform-independent
#include "Core/Asura_HalfFloat.h"

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

// Temporary define for toggling platform implementation. 
// Platforms with their own implementation should define this 
#if !defined(ASURA_USE_PLATFORM_VECTORS)

// Asura_AVector3...
#define Asura_AVector3_InParam		const Asura_AVector3&
#define Asura_AVector3_OutParam		Asura_AVector3&
#define Asura_AVector3_InOutParam	Asura_AVector3&

// Asura_AMatrix3x3...
#define Asura_AMatrix3x3_InParam	const Asura_AMatrix3x3&
#define Asura_AMatrix3x3_OutParam	Asura_AMatrix3x3&
#define Asura_AMatrix3x3_InOutParam	Asura_AMatrix3x3&

// Asura_AVector4...
#define Asura_AVector4_InParam		const Asura_AVector4&
#define Asura_AVector4_OutParam		Asura_AVector4&
#define Asura_AVector4_InOutParam	Asura_AVector4&

// Asura_AMatrix4x4...
#define Asura_AMatrix4x4_InParam	const Asura_AMatrix4x4&
#define Asura_AMatrix4x4_OutParam	Asura_AMatrix4x4&
#define Asura_AMatrix4x4_InOutParam	Asura_AMatrix4x4&

class Asura_AVector3
{
public:
	Asura_AVector3(void);
	Asura_AVector3(const float fX, const float fY, const float fZ);
	explicit Asura_AVector3(const Asura_Vector_3& xOther);

	// Load/store functions...
	void LoadAligned(const float* pfFloats);
	void LoadUnaligned(const float* pfFloats);
	void StoreAligned(float* pfFloats) const;
	void StoreUnaligned(float* pfFloats) const;
	void StoreAlignedHalf(Asura_HalfFloat* pxHalves) const;

	// Selection functions...
	void SelectGE(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS, Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse);
	void SelectGT(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS, Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse);
	void SelectLE(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS, Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse);
	void SelectLT(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS, Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse);
	void SelectEQ(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS, Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse);
	void SelectNE(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS, Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse);

	// Manipulation functions...
	void Zero(void);

	float MagnitudeSquared(void) const;
	float Magnitude(void) const;

	void Normalise(void);
	float NormaliseReturningMagnitude(void);

	void ApproximateNormalise(void);
	float ApproximateNormaliseReturningMagnitude(void);

	void Lerp(Asura_AVector3_InParam xTo, const float fWeight);
	void Lerp(Asura_AVector3_InParam xFrom, Asura_AVector3_InParam xTo, const float fWeight);

	// Minimum/maximum functions...
	void Max(Asura_AVector3_InParam xV1, Asura_AVector3_InParam xV2);
	void Min(Asura_AVector3_InParam xV1, Asura_AVector3_InParam xV2);

	// Assignment operators...
	void operator=(const Asura_Vector_3& xOther);
	void operator=(Asura_AVector3_InParam xOther);
	void operator+=(Asura_AVector3_InParam xOther);
	void operator-=(Asura_AVector3_InParam xOther);
	void operator*=(const float fScale);
	void operator/=(const float fScale);
	void operator*=(Asura_AVector3_InParam xOther);

	// Unary operators...
	Asura_AVector3 operator-(void) const;

	// Binary operators...
	Asura_AVector3 operator+(Asura_AVector3_InParam xOther) const;
	Asura_AVector3 operator-(Asura_AVector3_InParam xOther) const;
	Asura_AVector3 operator*(const float fScale) const;
	Asura_AVector3 operator/(const float fScale) const;
	Asura_AVector3 operator^(Asura_AVector3_InParam xOther) const;

	float operator*(Asura_AVector3_InParam xOther) const;

	// Equivalence operators...
	bool ApproximatelyEquals(Asura_AVector3_InParam xOther, const float fEpsilon = 1.0e-6f) const;

	bool operator==(Asura_AVector3_InParam xOther) const;
	bool operator!=(Asura_AVector3_InParam xOther) const;

	bool operator>(Asura_AVector3_InParam xOther) const;
	bool operator<(Asura_AVector3_InParam xOther) const;
	bool operator>=(Asura_AVector3_InParam xOther) const;
	bool operator<=(Asura_AVector3_InParam xOther) const;

	// Element accessors...
	float GetX(void) const;
	float GetY(void) const;
	float GetZ(void) const;

	void SetX(const float fX);
	void SetY(const float fY);
	void SetZ(const float fZ);

	float GetElement(const u_int uIndex) const;
	void SetElement(const u_int uIndex, const float fValue);

private:
	union
	{
		float	m_afElements[3];

		struct 
		{
			float m_fX;
			float m_fY;
			float m_fZ;
		};
	};
};


inline void Asura_Vector_3:: operator = ( const Asura_AVector3& xV4)
{
	x = xV4.GetX();
	y = xV4.GetY();
	z = xV4.GetZ();
}

// explicit constructor
inline Asura_Vector_3::Asura_Vector_3( const Asura_AVector3& xV4 )
{
	x = xV4.GetX();
	y = xV4.GetY();
	z = xV4.GetZ();
}


// Note that in Asura_Matrix_3x3 the data is stored as 3 rows. For efficiency when using 
// vector instructions it's better for us to store 3 columns. Since the old 3x3 matrix
// use e11 to e33 I'xMat keeping the convention of starting named columns at 1 rather than 0.
class Asura_AMatrix3x3
{
public:
	// Constructors...
	Asura_AMatrix3x3(void);
	Asura_AMatrix3x3(const float fE11, const float fE12, const float fE13,
					 const float fE21, const float fE22, const float fE23,
					 const float fE31, const float fE32, const float fE33);
	Asura_AMatrix3x3(const Asura_AVector3& xCol0, const Asura_AVector3& xCol1, const Asura_AVector3& xCol2);
	explicit Asura_AMatrix3x3(const Asura_Matrix_3x3& xMat3);

	void Zero(void);
	void Identity(void);
	void ToMat3(Asura_Matrix_3x3& xMat3) const;
	void FromMat3(const Asura_Matrix_3x3& xMat3);

	Asura_AVector3 GetColumn0(void) const;
	Asura_AVector3 GetColumn1(void) const;
	Asura_AVector3 GetColumn2(void) const;
	Asura_AVector3 GetColumn(const u_int uIndex) const;

	Asura_AVector3 GetRow0(void) const;
	Asura_AVector3 GetRow1(void) const;
	Asura_AVector3 GetRow2(void) const;
	Asura_AVector3 GetRow(const u_int uIndex) const;

	float GetElement(const u_int uRow, const u_int uCol) const;

	void SetColumn0(Asura_AVector3_InParam xCol0);
	void SetColumn1(Asura_AVector3_InParam xCol1);
	void SetColumn2(Asura_AVector3_InParam xCol2);
	void SetColumn(const u_int uIndex, Asura_AVector3_InParam xColumn);

	void SetRow0(Asura_AVector3_InParam xRow0);
	void SetRow1(Asura_AVector3_InParam xRow1);
	void SetRow2(Asura_AVector3_InParam xRow2);
	void SetRow(const u_int uIndex, Asura_AVector3_InParam xRow);

	void SetElement(const u_int uRow, const u_int uCol, const float fValue);

	// Tilde operator to make a skew-symmetric matrix from a vector
	void Tilde(Asura_AVector3_InParam xVector);

	// Try to make the matrix orthogonal by forcing the axes represented by the matrix
	// columns to be a) unit-length and b) perpendicular.
	void Orthogonalise(void);

	// matrix-matrix per component addition - used in integration
	void Add(const Asura_AMatrix3x3& xM);

	Asura_AVector3 Apply(const Asura_AVector3& xInput) const;
	void Apply(Asura_AMatrix3x3_InParam xInput, Asura_AMatrix3x3_OutParam xOutput) const;

	void Transpose(void);
	Asura_AVector3 ApplyTranspose(Asura_AVector3_InParam xVec) const;

	Asura_AMatrix3x3 operator*(const float fScale) const;

	bool operator==(Asura_AMatrix3x3_InParam xMat) const;
	bool operator!=(Asura_AMatrix3x3_InParam xMat) const;

private:
	Asura_AVector3 m_xCol0;
	Asura_AVector3 m_xCol1;
	Asura_AVector3 m_xCol2;
};

inline void Asura_Quat::ToMatrix(Asura_AMatrix3x3_OutParam xMat) const
{
	/* Implementation from Graphics Gems IV */

	float s = 2.0f;///Magnitude();      

	float xs = x*s;
	float ys = y*s;
	float zs = z*s;

	float wx = w*xs;
	float wy = w*ys;
	float wz = w*zs;

	float xx = x*xs;
	float xy = x*ys;
	float xz = x*zs;

	float yy = y*ys;
	float yz = y*zs;
	float zz = z*zs;

	xMat.SetColumn0(Asura_AVector3( 1.0f - (yy+zz), xy + wz, xz - wy ));
	xMat.SetColumn1(Asura_AVector3( xy - wz, 1.0f - (xx + zz), yz + wx ));
	xMat.SetColumn2(Asura_AVector3( xz + wy, yz - wx, 1.0f - (xx+yy) ));
}

inline void Asura_Quat::FromMatrix(Asura_AMatrix3x3_InParam xM4)
{
	Asura_Matrix_3x3 xM3;
	xM4.ToMat3(xM3);
	FromMatrix(xM3);
}

inline Asura_AVector3 Asura_Quat::RotateVector(Asura_AVector3_InParam p) const
{
	//-----------------------------------------------------
	// 
	// This function rotates a vector p by the quaternion q
	// The formula is:
	//                  p' = (q)(p)(q*)
	//
	// but we calculate it using a vector formula which is an
	// expansion of this that Kev worked out.
	// for q = (xOther,w)
	// xOther' = 2w(xOther^p) + (2(xOther.p)xOther + (w^2-xOther^2)p
	//
	//--------------------------------------------------------

	const Asura_AVector3 xOther(x,y,z); // the vector part of q
	Asura_AVector3 section1, section2, section3;

	float vMag = xOther.Magnitude();

	section1 = xOther^p;
	section1 *= (2*w);

	section2 = xOther;
	section2 *= (2*(xOther*p));

	section3 = p;
	section3 *= (w*w - vMag*vMag);

	return section1+section2+section3;
}

// creates a quaternion from an angle and axis of rotation
inline void Asura_Quat::FromRotation(Asura_Angle angle, Asura_AVector3_InParam axis)
{
	Asura_Assert(Asura_Maths::Abs(axis.MagnitudeSquared() - 1.0f) < 0.0001f, "axis should be normalised");

	angle *= 0.5f;
	float s;
	Asura_Maths::SinCos(angle, s, w);
	x = axis.GetX() * s;
	y = axis.GetY() * s;
	z = axis.GetZ() * s;
}

class Asura_SoA_Float
{
	float m_afValues[4];

public:

	Asura_SoA_Float() {}
	Asura_SoA_Float( float fSplat )
	{
		m_afValues[0] = fSplat;
		m_afValues[1] = fSplat;
		m_afValues[2] = fSplat;
		m_afValues[3] = fSplat;
	}

	Asura_SoA_Float( float f0, float f1, float f2, float f3 )
	{
		m_afValues[0] = f0;
		m_afValues[1] = f1;
		m_afValues[2] = f2;
		m_afValues[3] = f3;
	}

	inline void operator = ( const Asura_SoA_Float& xRHS )
	{
		m_afValues[0] = xRHS.m_afValues[0];
		m_afValues[1] = xRHS.m_afValues[1];
		m_afValues[2] = xRHS.m_afValues[2];
		m_afValues[3] = xRHS.m_afValues[3];
	}

	Asura_SoA_Float( const Asura_SoA_Float& xRHS )
	{
		m_afValues[0] = xRHS.m_afValues[0];
		m_afValues[1] = xRHS.m_afValues[1];
		m_afValues[2] = xRHS.m_afValues[2];
		m_afValues[3] = xRHS.m_afValues[3];
	}

	inline Asura_SoA_Float operator + ( const Asura_SoA_Float& xRHS ) const
	{
		return Asura_SoA_Float( m_afValues[0] + xRHS.m_afValues[0],
			m_afValues[1] + xRHS.m_afValues[1],
			m_afValues[2] + xRHS.m_afValues[2],
			m_afValues[3] + xRHS.m_afValues[3] );
	}

	inline Asura_SoA_Float operator * ( const Asura_SoA_Float& xRHS ) const
	{
		return Asura_SoA_Float( m_afValues[0] * xRHS.m_afValues[0],
			m_afValues[1] * xRHS.m_afValues[1],
			m_afValues[2] * xRHS.m_afValues[2],
			m_afValues[3] * xRHS.m_afValues[3] );
	}

	inline bool operator == ( const Asura_SoA_Float& xRHS ) const
	{
		return	m_afValues[0] == xRHS.m_afValues[0] && 
			m_afValues[1] == xRHS.m_afValues[1] && 
			m_afValues[2] == xRHS.m_afValues[2] && 
			m_afValues[3] == xRHS.m_afValues[3];
	}
	inline bool operator != ( const Asura_SoA_Float& xRHS ) const
	{
		return	m_afValues[0] != xRHS.m_afValues[0] || 
			m_afValues[1] != xRHS.m_afValues[1] || 
			m_afValues[2] != xRHS.m_afValues[2] || 
			m_afValues[3] != xRHS.m_afValues[3];
	}

	friend class Asura_SoA_Vector_3;
};

class Asura_SoA_Vector_3
{
	Asura_SoA_Float m_xX;
	Asura_SoA_Float m_xY;
	Asura_SoA_Float m_xZ;

public:

	Asura_SoA_Vector_3() {}
	Asura_SoA_Vector_3( const Asura_SoA_Float& xX, const Asura_SoA_Float& xY, const Asura_SoA_Float& xZ ) 
	{
		m_xX = xX;
		m_xY = xY;
		m_xZ = xZ;
	}
	Asura_SoA_Vector_3( const Asura_SoA_Vector_3& xRHS)
	{
		m_xX = xRHS.m_xX;
		m_xY = xRHS.m_xY;
		m_xZ = xRHS.m_xZ;
	}
	Asura_SoA_Vector_3& operator = (const Asura_SoA_Vector_3& xRHS)
	{
		m_xX = xRHS.m_xX;
		m_xY = xRHS.m_xY;
		m_xZ = xRHS.m_xZ;
		return *this;
	}

	// create new SoA floats from members
	inline Asura_SoA_Float GetXXXX() const { return m_xX; }
	inline Asura_SoA_Float GetYYYY() const { return m_xY; }
	inline Asura_SoA_Float GetZZZZ() const { return m_xZ; }

	// dot product
	inline Asura_SoA_Float operator * ( const Asura_SoA_Vector_3& xRHS ) const
	{
		//		return	m_xX * xRHS.m_xX + 
		//				m_xY * xRHS.m_xY + 
		//				m_xZ * xRHS.m_xZ;
		float fOut0, fOut1, fOut2, fOut3;

		float fX0 = m_xX.m_afValues[0];
		float fX1 = xRHS.m_xX.m_afValues[0];
		float fY0 = m_xY.m_afValues[0];
		float fY1 = xRHS.m_xY.m_afValues[0];
		float fZ0 = m_xZ.m_afValues[0];
		float fZ1 = xRHS.m_xZ.m_afValues[0];
		float fX2 = m_xX.m_afValues[1];
		float fX3 = xRHS.m_xX.m_afValues[1];
		float fY2 = m_xY.m_afValues[1];
		float fY3 = xRHS.m_xY.m_afValues[1];
		float fZ2 = m_xZ.m_afValues[1];
		float fZ3 = xRHS.m_xZ.m_afValues[1];

		fOut0 = fX0 * fX1 + fY0 * fY1 + fZ0 * fZ1;
		fOut1 = fX2 * fX3 + fY2 * fY3 + fZ2 * fZ3;

		fX0 = m_xX.m_afValues[2];
		fX1 = xRHS.m_xX.m_afValues[2];
		fY0 = m_xY.m_afValues[2];
		fY1 = xRHS.m_xY.m_afValues[2];
		fZ0 = m_xZ.m_afValues[2];
		fZ1 = xRHS.m_xZ.m_afValues[2];
		fX2 = m_xX.m_afValues[3];
		fX3 = xRHS.m_xX.m_afValues[3];
		fY2 = m_xY.m_afValues[3];
		fY3 = xRHS.m_xY.m_afValues[3];
		fZ2 = m_xZ.m_afValues[3];
		fZ3 = xRHS.m_xZ.m_afValues[3];

		fOut2 = fX0 * fX1 + fY0 * fY1 + fZ0 * fZ1;
		fOut3 = fX2 * fX3 + fY2 * fY3 + fZ2 * fZ3;

		return Asura_SoA_Float( fOut0, fOut1, fOut2, fOut3 );
	}

	// transform by matrix
	inline const Asura_SoA_Vector_3 operator * ( Asura_AMatrix3x3_InParam xMat ) const
	{
		const Asura_AVector3 xCol0 = xMat.GetColumn0();
		// splat M11 to all elements of xM11, etc.
		Asura_SoA_Float xM11(xCol0.GetX());
		Asura_SoA_Float xM21(xCol0.GetY());
		Asura_SoA_Float xM31(xCol0.GetZ());

		Asura_SoA_Float xOutX = m_xX * xM11;
		Asura_SoA_Float xOutY = m_xX * xM21;
		Asura_SoA_Float xOutZ = m_xX * xM31;

		const Asura_AVector3 xCol1 = xMat.GetColumn1();
		Asura_SoA_Float xM12(xCol1.GetX());
		Asura_SoA_Float xM22(xCol1.GetY());
		Asura_SoA_Float xM32(xCol1.GetZ());

		xOutX = m_xY * xM12 + xOutX;
		xOutY = m_xY * xM22 + xOutY;
		xOutZ = m_xY * xM32 + xOutZ;

		const Asura_AVector3 xCol2 = xMat.GetColumn2();
		Asura_SoA_Float xM13(xCol2.GetX());
		Asura_SoA_Float xM23(xCol2.GetY());
		Asura_SoA_Float xM33(xCol2.GetZ());

		xOutX = m_xZ * xM13 + xOutX;
		xOutY = m_xZ * xM23 + xOutY;
		xOutZ = m_xZ * xM33 + xOutZ;

		return Asura_SoA_Vector_3( xOutX, xOutY, xOutZ );
	}
};


//////////////////////////////////////////////////////////////////////////
// Asura_AVector4
//////////////////////////////////////////////////////////////////////////
class Asura_AVector4
{
public:
	// Constructors...
	Asura_AVector4(void);
	Asura_AVector4(const float fX, const float fY, const float fZ, const float fW);
	Asura_AVector4(const Asura_AVector4& xOther);
	Asura_AVector4(const Asura_AVector3& xOther, const float fW);
	Asura_AVector4(const Asura_Vector_3& xOther, const float fW);

	// Zero functions...
	void ZeroW0(void);
	void ZeroW1(void);

	// Selection functions...
	void SelectGE(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS, Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse);
	void SelectGT(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS, Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse);
	void SelectLE(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS, Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse);
	void SelectLT(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS, Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse);
	void SelectEQ(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS, Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse);
	void SelectNE(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS, Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse);

	// Load/store functions...
	void LoadAligned(const float* pfFloats);
	void LoadUnaligned(const float* pfFloats);
	void StoreAligned(float* pfFloats) const;
	void StoreUnaligned(float* pfFloats) const;
	void StoreAlignedHalf(Asura_HalfFloat* pxHalves) const;

	// Manipulation functions...
	Asura_AVector4 ComponentwiseProduct(Asura_AVector4_InParam xOther) const;
	float DotProduct3(Asura_AVector4_InParam xOther) const;
	float DotProduct4(Asura_AVector4_InParam xOther) const;

	float MagnitudeSquared(void) const;
	float Magnitude(void) const;

	void Normalise(void);
	float NormaliseReturningMagnitude(void);
	void ApproximateNormalise(void);
	float ApproximateNormaliseReturningMagnitude(void);

	// Assignment operators...
	void operator=(Asura_AVector4_InParam xOther);

	void operator+=(Asura_AVector4_InParam xOther);
	void operator-=(Asura_AVector4_InParam xOther);

	void operator*=(const float fScale);
	void operator/=(const float fScale);

	// Unary operators...
	Asura_AVector4 operator-(void) const;

	// Binary operators...
	Asura_AVector4 operator+(Asura_AVector4_InParam xOther) const;
	Asura_AVector4 operator-(Asura_AVector4_InParam xOther) const;
	Asura_AVector4 operator^(Asura_AVector4_InParam xOther) const;

	float operator*(Asura_AVector4_InParam xOther) const;

	Asura_AVector4 operator*(const float fScale) const;
	Asura_AVector4 operator/(const float fScale) const;

	// Comparison operators...
	bool operator==(Asura_AVector4_InParam xOther) const;
	bool ApproximatelyEquals(Asura_AVector4_InParam xOther, const float fEpsilon = 1.0e-6f) const;
	bool operator!=(Asura_AVector4_InParam xOther) const;

	// Minimum/maximum functions...
	void Max(Asura_AVector4_InParam xV1, Asura_AVector4_InParam xV2);
	void Min(Asura_AVector4_InParam xV1, Asura_AVector4_InParam xV2);

	// Properties...
	float GetX(void) const;
	float GetY(void) const;
	float GetZ(void) const;
	float GetW(void) const;

	void SetX(const float fX);
	void SetY(const float fY);
	void SetZ(const float fZ);
	void SetW(const float fW);

	float GetElement(const u_int uIndex) const;
	void SetElement(const u_int uIndex, const float fValue);
	void SetElements(const float fX, const float fY, const float fZ, const float fW);

private:
	union
	{
		float	m_afElements[4];

		struct 
		{
			float m_fX;
			float m_fY;
			float m_fZ;
			float m_fW;
		};
	};
};


//////////////////////////////////////////////////////////////////////////
// Asura_AMatrix4x4
//////////////////////////////////////////////////////////////////////////
class Asura_AMatrix4x4
{
public:
	// Constructors/initializers
	Asura_AMatrix4x4(void);
	Asura_AMatrix4x4(const float fE11, const float fE12, const float fE13, const float fE14,
					 const float fE21, const float fE22, const float fE23, const float fE24,
					 const float fE31, const float fE32, const float fE33, const float fE34,
					 const float fE41, const float fE42, const float fE43, const float fE44);
	Asura_AMatrix4x4(const Asura_AMatrix4x4& xOther);
	Asura_AMatrix4x4(const Asura_Matrix_3x3& xOrientation, const Asura_Vector_3& xTranslation);
	Asura_AMatrix4x4(const Asura_Quat& xOrientation, const Asura_Vector_3& xTranslation);

	// Constant functions...
	void Identity(void);
	void Zero(void);

	// Orientation/translation functions...
	void ToMatrix(Asura_Matrix_3x3& xOrientation) const;
	void FromMatrix(const Asura_Matrix_3x3& xOrientation);

	Asura_Vector_3 GetTranslation() const;
	void SetTranslation(const Asura_Vector_3& xTranslation);

	Asura_AVector3 GetATranslation() const;
	void SetTranslation(Asura_AVector3_InParam xTranslation);

	Asura_AVector4 GetA4Translation() const;
	void SetTranslation(Asura_AVector4_InParam xTranslation);

	// Manipulation functions...
	Asura_AVector4 Apply(Asura_AVector4_InParam xInput) const;
	void ApplyToArray(const Asura_AVector4* pxInput, const u_int uNumElements, Asura_AVector4* pxOutput) const;

	void Apply(Asura_AMatrix4x4_InParam xInput, Asura_AMatrix4x4_OutParam xOutput) const;
	void ApplyToArray(const Asura_AMatrix4x4* pxInput, const u_int uNumElements, Asura_AMatrix4x4* pxOutput) const;

	float Cofactor(const u_int uRow, const u_int uCol) const;
	float Determinant(void) const;

	void Invert(void);
	void InvertOrthogonal(void);
	void Transpose(void);

	void Invert(Asura_AMatrix4x4_OutParam xOther) const;
	void InvertOrthogonal(Asura_AMatrix4x4_OutParam xOther) const;
	void Transpose(Asura_AMatrix4x4_OutParam xOther) const;

	// Unary operators
	Asura_AMatrix4x4 operator-(void) const;

	// Binary operators
	Asura_AMatrix4x4 operator*(const float fScale) const;
	Asura_AMatrix4x4 operator/(const float fScale) const;

	// Assignment operators
	void operator=(Asura_AMatrix4x4_InParam xOther);
	void operator*=(const float fScale);
	void operator/=(const float fScale);

	// Conditional operators
	bool operator==(Asura_AMatrix4x4_InParam xOther) const;
	bool ApproximatelyEquals(Asura_AMatrix4x4_InParam, const float fEpsilon = 1.0e-6f) const;
	bool operator!=(Asura_AMatrix4x4_InParam xOther) const;

	// Elementary properties...
	float GetElement(const u_int uRow, const u_int uCol) const;
	void SetElement(const u_int uRow, const u_int uCol, const float fValue);

	Asura_AVector4 GetColumn(const u_int uColumnIndex) const;
	Asura_AVector4 GetColumn0() const;
	Asura_AVector4 GetColumn1() const;
	Asura_AVector4 GetColumn2() const;
	Asura_AVector4 GetColumn3() const;

	void SetColumn(const u_int uColumnIndex, Asura_AVector4_InParam xColumn);
	void SetColumn0(Asura_AVector4_InParam xCol0);
	void SetColumn1(Asura_AVector4_InParam xCol1);
	void SetColumn2(Asura_AVector4_InParam xCol2);
	void SetColumn3(Asura_AVector4_InParam xCol3);

	Asura_AVector4 GetRow(const u_int uRowIndex) const;
	Asura_AVector4 GetRow0() const;
	Asura_AVector4 GetRow1() const;
	Asura_AVector4 GetRow2() const;
	Asura_AVector4 GetRow3() const;

	void SetRow(const u_int uRowIndex, Asura_AVector4_InParam xRow);
	void SetRow0(Asura_AVector4_InParam xRow0);
	void SetRow1(Asura_AVector4_InParam xRow1);
	void SetRow2(Asura_AVector4_InParam xRow2);
	void SetRow3(Asura_AVector4_InParam xRow3);

	void SetColumns(Asura_AVector4_InParam xCol0, Asura_AVector4_InParam xCol1, Asura_AVector4_InParam xCol2, Asura_AVector4_InParam xCol3);
	void SetRows(Asura_AVector4_InParam xRow0, Asura_AVector4_InParam xRow1, Asura_AVector4_InParam xRow2, Asura_AVector4_InParam xRow3);

	// Load/store functions...
	void LoadColsAligned(const float* pfFloats);
	void LoadRowsAligned(const float* pfFloats);
	void LoadColsUnaligned(const float* pfFloats);
	void LoadRowsUnaligned(const float* pfFloats);

	void StoreColsAligned(float* pfFloats) const;
	void StoreRowsAligned(float* pfFloats) const;
	void StoreColsUnaligned(float* pfFloats) const;
	void StoreRowsUnaligned(float* pfFloats) const;

private:
	Asura_AVector4 m_axColumns[4];
};


//////////////////////////////////////////////////////////////////////////
// Inline Function Definitions
//////////////////////////////////////////////////////////////////////////

inline Asura_AVector3 operator*(Asura_AVector3_InParam xVec, Asura_AMatrix3x3_InParam xMat)
{
	return xMat.Apply(xVec);
}

inline Asura_AVector3 operator*(Asura_AVector3_InParam xVec, const Asura_Matrix_3x3& xMat)
{
	const Asura_AMatrix3x3	xTemp(xMat);
	return xTemp.Apply(xVec);
}

inline Asura_AMatrix3x3 operator*(Asura_AMatrix3x3_InParam xM1, Asura_AMatrix3x3_InParam xM2)
{
	Asura_AMatrix3x3	xResult;

	xM2.Apply(xM1, xResult);
	return(xResult);
}

// scalar * 4x1 Vector - to match the member function in Asura_AVector4
inline Asura_AVector4 operator*(const float fScale, Asura_AVector4_InParam xOther)
{
	return(xOther * fScale);
}

inline Asura_AVector4 operator*(Asura_AVector4_InParam xVec, Asura_AMatrix4x4_InParam xMat)
{
	return xMat.Apply(xVec);
}

// To match the backwards vector-matrix multiply, we do a backwards matrix-vector multiply too.
inline Asura_AMatrix4x4 operator*(Asura_AMatrix4x4_InParam xM1, Asura_AMatrix4x4_InParam xM2)
{
	Asura_AMatrix4x4	xResult;

	xM2.Apply(xM1, xResult);
	return(xResult);
}


//////////////////////////////////////////////////////////////////////////
// Asura_AVector3
//////////////////////////////////////////////////////////////////////////
inline Asura_AVector3::Asura_AVector3(void)
{
}

inline Asura_AVector3::Asura_AVector3(const float fX, const float fY, const float fZ)
{
	m_fX = fX;
	m_fY = fY;
	m_fZ = fZ;
}

inline Asura_AVector3::Asura_AVector3(const Asura_Vector_3& xOther)
{
	m_fX = xOther.x;
	m_fY = xOther.y;
	m_fZ = xOther.z;
}

inline void	Asura_AVector3::LoadAligned(const float* pfFloats)
{
	m_fX = pfFloats[0];
	m_fY = pfFloats[1];
	m_fZ = pfFloats[2];
}

inline void	Asura_AVector3::LoadUnaligned(const float* pfFloats)
{
	m_fX = pfFloats[0];
	m_fY = pfFloats[1];
	m_fZ = pfFloats[2];
}

inline void	Asura_AVector3::StoreAligned(float* pfFloats) const 
{
	pfFloats[0] = m_fX;
	pfFloats[1] = m_fY;
	pfFloats[2] = m_fZ;
}

inline void	Asura_AVector3::StoreUnaligned(float* pfFloats) const
{
	pfFloats[0] = m_fX;
	pfFloats[1] = m_fY;
	pfFloats[2] = m_fZ;
}

inline void Asura_AVector3::StoreAlignedHalf(Asura_HalfFloat* pxHalves) const
{
	pxHalves[0] = m_fX;
	pxHalves[1] = m_fY;
	pxHalves[2] = m_fZ;
}

inline void Asura_AVector3::operator=(const Asura_Vector_3& xOther) 
{
	m_fX = xOther.x;
	m_fY = xOther.y;
	m_fZ = xOther.z;
}

inline void Asura_AVector3::operator=(Asura_AVector3_InParam xOther) 
{
	m_fX = xOther.m_fX;
	m_fY = xOther.m_fY;
	m_fZ = xOther.m_fZ;
}

inline void Asura_AVector3::operator+=(Asura_AVector3_InParam xOther)
{
	m_fX += xOther.m_fX;
	m_fY += xOther.m_fY;
	m_fZ += xOther.m_fZ;
}

inline void Asura_AVector3::operator-=(Asura_AVector3_InParam xOther)
{
	m_fX -= xOther.m_fX;
	m_fY -= xOther.m_fY;
	m_fZ -= xOther.m_fZ;
}

inline void Asura_AVector3::operator*=(const float fScale)
{
	m_fX *= fScale;
	m_fY *= fScale;
	m_fZ *= fScale;
}

inline void Asura_AVector3::operator/=(const float fScale)
{
	float fInv = 1.0f / fScale;
	m_fX *= fInv;
	m_fY *= fInv;
	m_fZ *= fInv;
}

inline void Asura_AVector3::operator*=(Asura_AVector3_InParam xOther)
{
	m_fX *=	xOther.m_fX;
	m_fY *=	xOther.m_fY;
	m_fZ *=	xOther.m_fZ;
}

inline Asura_AVector3 Asura_AVector3::operator-(void) const
{
	return Asura_AVector3(-m_fX, -m_fY, -m_fZ );
}

inline Asura_AVector3 Asura_AVector3::operator+(Asura_AVector3_InParam xOther) const
{
	return Asura_AVector3(m_fX + xOther.m_fX, m_fY + xOther.m_fY, m_fZ + xOther.m_fZ );
}

inline Asura_AVector3 Asura_AVector3::operator-(Asura_AVector3_InParam xOther) const
{
	return Asura_AVector3(m_fX - xOther.m_fX, m_fY - xOther.m_fY, m_fZ - xOther.m_fZ );
}

inline Asura_AVector3 Asura_AVector3::operator*(const float fScale) const
{
	return Asura_AVector3(m_fX * fScale, m_fY * fScale, m_fZ * fScale );
}

inline Asura_AVector3 Asura_AVector3::operator/(const float fScale) const
{
	const float	fInv	= 1.0f / fScale;

	return Asura_AVector3(m_fX * fInv, m_fY * fInv, m_fZ * fInv );
}                                               

inline Asura_AVector3 Asura_AVector3::operator^(Asura_AVector3_InParam xOther) const
{
	return Asura_AVector3(m_fY*xOther.m_fZ - m_fZ*xOther.m_fY, m_fZ*xOther.m_fX - m_fX*xOther.m_fZ, m_fX*xOther.m_fY - m_fY*xOther.m_fX );
}

inline float Asura_AVector3::operator*(Asura_AVector3_InParam xOther) const
{
	return (m_fX*xOther.m_fX + m_fY*xOther.m_fY + m_fZ*xOther.m_fZ);
}

inline bool Asura_AVector3::operator==(const Asura_AVector3& xOther) const
{
	return m_fX == xOther.m_fX && m_fY == xOther.m_fY && m_fZ == xOther.m_fZ;
}

inline bool Asura_AVector3::ApproximatelyEquals(Asura_AVector3_InParam xOther, const float fEpsilon) const
{
	Asura_AVector3 delta;

	delta.m_fX = m_fX-xOther.m_fX;
	delta.m_fY = m_fY-xOther.m_fY;
	delta.m_fZ = m_fZ-xOther.m_fZ;

	return ((delta.m_fX*delta.m_fX+delta.m_fY*delta.m_fY+delta.m_fZ*delta.m_fZ)<fEpsilon);
}

inline bool Asura_AVector3::operator!=(Asura_AVector3_InParam xOther) const
{
	return m_fX != xOther.m_fX || m_fY != xOther.m_fY || m_fZ != xOther.m_fZ;
}

inline bool	Asura_AVector3::operator>(Asura_AVector3_InParam xOther) const
{
	return (m_fX > xOther.m_fX) && (m_fY > xOther.m_fY) && (m_fZ > xOther.m_fZ);
}

inline bool	Asura_AVector3::operator<(Asura_AVector3_InParam xOther) const
{
	return (m_fX < xOther.m_fX) && (m_fY < xOther.m_fY) && (m_fZ < xOther.m_fZ);
}

inline bool	Asura_AVector3::operator>=(Asura_AVector3_InParam xOther) const
{
	return (m_fX >= xOther.m_fX) && (m_fY >= xOther.m_fY) && (m_fZ >= xOther.m_fZ);
}

inline bool	Asura_AVector3::operator<=(Asura_AVector3_InParam xOther) const
{
	return (m_fX <= xOther.m_fX) && (m_fY <= xOther.m_fY) && (m_fZ <= xOther.m_fZ);
}

inline float Asura_AVector3::MagnitudeSquared(void) const
{   
	return ((m_fX*m_fX) + (m_fY*m_fY) + (m_fZ*m_fZ));
}

inline float Asura_AVector3::Magnitude(void) const
{   
	return Asura_Maths::Sqrt((m_fX*m_fX) + (m_fY*m_fY) + (m_fZ*m_fZ));
}

inline void Asura_AVector3::Normalise(void)
{
	const float fMag = Magnitude();
	Asura_Assert(fMag != 0.0f, "trying to Normalise Zero vector");
	if (fMag != 0.0f)
	{
		const float fInvMag = 1.0f/fMag;
		m_fX *= fInvMag;
		m_fY *= fInvMag;
		m_fZ *= fInvMag;
	}
}       

inline float Asura_AVector3::NormaliseReturningMagnitude(void)
{
	const float fMag = Magnitude();

	if (fMag==0.0f) return fMag;

	const float fInvMag = 1.0f/fMag;
	m_fX *= fInvMag;
	m_fY *= fInvMag;
	m_fZ *= fInvMag;

	return fMag;
}       

inline void Asura_AVector3::ApproximateNormalise(void)
{
	return(Normalise());
}       

inline float Asura_AVector3::ApproximateNormaliseReturningMagnitude(void)
{
	return(NormaliseReturningMagnitude());
}       

inline void Asura_AVector3::Lerp(Asura_AVector3_InParam xTo, const float fWeight)
{
	const float	fFromWeight	= 1.0f - fWeight;

	m_fX	= (m_fX * fFromWeight) + (xTo.m_fX * fWeight);
	m_fY	= (m_fY * fFromWeight) + (xTo.m_fY * fWeight);
	m_fZ	= (m_fZ * fFromWeight) + (xTo.m_fZ * fWeight);
}

inline void Asura_AVector3::Lerp(Asura_AVector3_InParam xFrom, Asura_AVector3_InParam xTo, const float fWeight)
{
	const float	fFromWeight	= 1.0f - fWeight;

	m_fX	= (xFrom.m_fX * fFromWeight) + (xTo.m_fX * fWeight);
	m_fY	= (xFrom.m_fY * fFromWeight) + (xTo.m_fY * fWeight);
	m_fZ	= (xFrom.m_fZ * fFromWeight) + (xTo.m_fZ * fWeight);
}

inline void Asura_AVector3::Max(Asura_AVector3_InParam xV1, Asura_AVector3_InParam xV2)
{
	m_fX	= Asura_Maths::Max(xV1.m_fX, xV2.m_fX);
	m_fY	= Asura_Maths::Max(xV1.m_fY, xV2.m_fY);
	m_fZ	= Asura_Maths::Max(xV1.m_fZ, xV2.m_fZ);
}
inline void Asura_AVector3::Min(Asura_AVector3_InParam xV1, Asura_AVector3_InParam xV2)
{
	m_fX	= Asura_Maths::Min(xV1.m_fX, xV2.m_fX);
	m_fY	= Asura_Maths::Min(xV1.m_fY, xV2.m_fY);
	m_fZ	= Asura_Maths::Min(xV1.m_fZ, xV2.m_fZ);
}

inline void Asura_AVector3::Zero(void) 
{
	m_fX = m_fY = m_fZ = 0.0f;
}

inline void	Asura_AVector3::SelectGE(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS,
									 Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse)
{
	m_fX = xLHS.m_fX >= xRHS.m_fX ? xTrue.m_fX : xFalse.m_fX;
	m_fY = xLHS.m_fY >= xRHS.m_fY ? xTrue.m_fY : xFalse.m_fY;
	m_fZ = xLHS.m_fZ >= xRHS.m_fZ ? xTrue.m_fZ : xFalse.m_fZ;
}

inline void	Asura_AVector3::SelectGT(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS,
									 Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse)
{
	m_fX = xLHS.m_fX > xRHS.m_fX ? xTrue.m_fX : xFalse.m_fX;
	m_fY = xLHS.m_fY > xRHS.m_fY ? xTrue.m_fY : xFalse.m_fY;
	m_fZ = xLHS.m_fZ > xRHS.m_fZ ? xTrue.m_fZ : xFalse.m_fZ;
}
inline void	Asura_AVector3::SelectLE(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS,
									 Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse)
{
	m_fX = xLHS.m_fX <= xRHS.m_fX ? xTrue.m_fX : xFalse.m_fX;
	m_fY = xLHS.m_fY <= xRHS.m_fY ? xTrue.m_fY : xFalse.m_fY;
	m_fZ = xLHS.m_fZ <= xRHS.m_fZ ? xTrue.m_fZ : xFalse.m_fZ;
}
inline void	Asura_AVector3::SelectLT(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS,
									 Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse)
{
	m_fX = xLHS.m_fX < xRHS.m_fX ? xTrue.m_fX : xFalse.m_fX;
	m_fY = xLHS.m_fY < xRHS.m_fY ? xTrue.m_fY : xFalse.m_fY;
	m_fZ = xLHS.m_fZ < xRHS.m_fZ ? xTrue.m_fZ : xFalse.m_fZ;
}
inline void	Asura_AVector3::SelectEQ(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS,
									 Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse)
{
	m_fX = xLHS.m_fX == xRHS.m_fX ? xTrue.m_fX : xFalse.m_fX;
	m_fY = xLHS.m_fY == xRHS.m_fY ? xTrue.m_fY : xFalse.m_fY;
	m_fZ = xLHS.m_fZ == xRHS.m_fZ ? xTrue.m_fZ : xFalse.m_fZ;
}
inline void	Asura_AVector3::SelectNE(Asura_AVector3_InParam xLHS, Asura_AVector3_InParam xRHS,
									 Asura_AVector3_InParam xTrue, Asura_AVector3_InParam xFalse)
{
	m_fX = xLHS.m_fX != xRHS.m_fX ? xTrue.m_fX : xFalse.m_fX;
	m_fY = xLHS.m_fY != xRHS.m_fY ? xTrue.m_fY : xFalse.m_fY;
	m_fZ = xLHS.m_fZ != xRHS.m_fZ ? xTrue.m_fZ : xFalse.m_fZ;
}

inline float Asura_AVector3::GetX(void) const
{
	return m_fX;
}

inline float Asura_AVector3::GetY(void) const
{
	return m_fY;
}

inline float Asura_AVector3::GetZ(void) const
{
	return m_fZ;
}

inline void Asura_AVector3::SetX(const float fX)
{
	m_fX = fX;
}

inline void Asura_AVector3::SetY(const float fY)
{
	m_fY = fY;
}

inline void Asura_AVector3::SetZ(const float fZ)
{
	m_fZ = fZ;
}

inline float Asura_AVector3::GetElement(const u_int uIndex) const
{
	return(m_afElements[uIndex]);
}

inline void Asura_AVector3::SetElement(const u_int uIndex, const float fValue)
{
	m_afElements[uIndex]	= fValue;
}


//////////////////////////////////////////////////////////////////////////
// Asura_AMatrix3x3
//////////////////////////////////////////////////////////////////////////
inline Asura_AMatrix3x3::Asura_AMatrix3x3(void)
{
}

inline Asura_AMatrix3x3::Asura_AMatrix3x3(const float fE11, const float fE12, const float fE13,
										  const float fE21, const float fE22, const float fE23,
										  const float fE31, const float fE32, const float fE33 )
{
	m_xCol0 = Asura_AVector3(fE11, fE21, fE31);
	m_xCol1 = Asura_AVector3(fE12, fE22, fE32);
	m_xCol2 = Asura_AVector3(fE13, fE23, fE33);
}

inline Asura_AMatrix3x3::Asura_AMatrix3x3(const Asura_AVector3& xCol0, const Asura_AVector3& xCol1, const Asura_AVector3& xCol2)
{
	m_xCol0 = xCol0;
	m_xCol1 = xCol1;
	m_xCol2 = xCol2;
}

inline Asura_AMatrix3x3::Asura_AMatrix3x3(const Asura_Matrix_3x3& xMat3)
{
	FromMat3(xMat3);
}

inline void Asura_AMatrix3x3::ToMat3(Asura_Matrix_3x3& xMat3) const
{
	xMat3.e11 = GetElement(0,0);
	xMat3.e12 = GetElement(0,1);
	xMat3.e13 = GetElement(0,2);
	xMat3.e21 = GetElement(1,0);
	xMat3.e22 = GetElement(1,1);
	xMat3.e23 = GetElement(1,2);
	xMat3.e31 = GetElement(2,0);
	xMat3.e32 = GetElement(2,1);
	xMat3.e33 = GetElement(2,2);
}

inline void Asura_AMatrix3x3::FromMat3(const Asura_Matrix_3x3& xMat3)
{
	m_xCol0 = Asura_AVector3(xMat3.e11, xMat3.e21, xMat3.e31);
	m_xCol1 = Asura_AVector3(xMat3.e12, xMat3.e22, xMat3.e32);
	m_xCol2 = Asura_AVector3(xMat3.e13, xMat3.e23, xMat3.e33);
}

inline Asura_AVector3 Asura_AMatrix3x3::GetColumn0(void) const
{
	return m_xCol0;
}

inline Asura_AVector3 Asura_AMatrix3x3::GetColumn1(void) const
{
	return m_xCol1;
}

inline Asura_AVector3 Asura_AMatrix3x3::GetColumn2(void) const
{
	return m_xCol2;
}

inline Asura_AVector3 Asura_AMatrix3x3::GetColumn(const u_int uIndex) const
{
	return *(&m_xCol0 + uIndex);
}

inline Asura_AVector3 Asura_AMatrix3x3::GetRow0(void) const
{
	return(Asura_AVector3(m_xCol0.GetX(), m_xCol1.GetX(), m_xCol2.GetX()));
}

inline Asura_AVector3 Asura_AMatrix3x3::GetRow1(void) const
{
	return(Asura_AVector3(m_xCol0.GetY(), m_xCol1.GetY(), m_xCol2.GetY()));
}

inline Asura_AVector3 Asura_AMatrix3x3::GetRow2(void) const
{
	return(Asura_AVector3(m_xCol0.GetZ(), m_xCol1.GetZ(), m_xCol2.GetZ()));
}

inline Asura_AVector3 Asura_AMatrix3x3::GetRow(const u_int uIndex) const
{
	return(Asura_AVector3(m_xCol0.GetElement(uIndex), m_xCol1.GetElement(uIndex), m_xCol2.GetElement(uIndex)));
}

inline float Asura_AMatrix3x3::GetElement(const u_int uRow, const u_int uCol) const
{
	return GetColumn(uCol).GetElement(uRow);
}

inline void Asura_AMatrix3x3::SetColumn0(Asura_AVector3_InParam xCol0)
{
	m_xCol0 = xCol0;
}

inline void Asura_AMatrix3x3::SetColumn1(Asura_AVector3_InParam xCol1)
{
	m_xCol1 = xCol1;
}

inline void Asura_AMatrix3x3::SetColumn2(Asura_AVector3_InParam xCol2)
{
	m_xCol2 = xCol2;
}

inline void Asura_AMatrix3x3::SetColumn(const u_int uIndex, Asura_AVector3_InParam xColumn)
{
	switch(uIndex)
	{
		case 0:
		{
			m_xCol0	= xColumn;
			break;
		}

		case 1:
		{
			m_xCol1	= xColumn;
			break;
		}

		case 2:
		{
			m_xCol2	= xColumn;
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected column-index specified");
			break;
		}
	}		// End of switch.
}

inline void Asura_AMatrix3x3::SetRow0(Asura_AVector3_InParam xRow0)
{
	m_xCol0.SetX(xRow0.GetX());
	m_xCol1.SetX(xRow0.GetY());
	m_xCol2.SetX(xRow0.GetZ());
}

inline void Asura_AMatrix3x3::SetRow1(Asura_AVector3_InParam xRow1)
{
	m_xCol0.SetY(xRow1.GetX());
	m_xCol1.SetY(xRow1.GetY());
	m_xCol2.SetY(xRow1.GetZ());
}

inline void Asura_AMatrix3x3::SetRow2(Asura_AVector3_InParam xRow2)
{
	m_xCol0.SetZ(xRow2.GetX());
	m_xCol1.SetZ(xRow2.GetY());
	m_xCol2.SetZ(xRow2.GetZ());
}

inline void Asura_AMatrix3x3::SetRow(const u_int uIndex, Asura_AVector3_InParam xRow)
{
	m_xCol0.SetElement(uIndex, xRow.GetX());
	m_xCol1.SetElement(uIndex, xRow.GetY());
	m_xCol2.SetElement(uIndex, xRow.GetZ());
}

inline void Asura_AMatrix3x3::SetElement(const u_int uRow, const u_int uCol, const float fValue)
{
	switch(uCol)
	{
		case 0:
		{
			m_xCol0.SetElement(uRow, fValue);
			break;
		}

		case 1:
		{
			m_xCol1.SetElement(uRow, fValue);
			break;
		}

		case 2:
		{
			m_xCol2.SetElement(uRow, fValue);
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected column-index specified");
			break;
		}
	}		// End of switch.
}

inline void Asura_AMatrix3x3::Tilde(Asura_AVector3_InParam xVector)
{
	m_xCol0 = Asura_AVector3( 0.0f,				+xVector.GetZ(),	-xVector.GetY() );
	m_xCol1 = Asura_AVector3( -xVector.GetZ(),	0.0f,				+xVector.GetX() );
	m_xCol2 = Asura_AVector3( +xVector.GetY(),	-xVector.GetX(),	0.0f			);
}

inline void Asura_AMatrix3x3::Orthogonalise(void)
{
	// normalise "down" column
	m_xCol1.Normalise();

	// make "forwards" column perpendicular to this, and normalise
	m_xCol2 -= m_xCol1 * ( m_xCol2 * m_xCol1 );
	m_xCol2.Normalise();

	// and now the "right" column must be perpendicular to both of these: x = + y^z
	m_xCol0 = m_xCol1 ^ m_xCol2;
}

inline void Asura_AMatrix3x3::Add(Asura_AMatrix3x3_InParam xM)
{
	m_xCol0 += xM.m_xCol0;
	m_xCol1 += xM.m_xCol1;
	m_xCol2 += xM.m_xCol2;
}

inline Asura_AVector3 Asura_AMatrix3x3::Apply(Asura_AVector3_InParam xInput) const
{
	return Asura_AVector3((GetElement(0,0) * xInput.GetX()) + (GetElement(0,1) * xInput.GetY()) + (GetElement(0,2) * xInput.GetZ()),
						  (GetElement(1,0) * xInput.GetX()) + (GetElement(1,1) * xInput.GetY()) + (GetElement(1,2) * xInput.GetZ()),
						  (GetElement(2,0) * xInput.GetX()) + (GetElement(2,1) * xInput.GetY()) + (GetElement(2,2) * xInput.GetZ()));
}

inline void Asura_AMatrix3x3::Apply(Asura_AMatrix3x3_InParam xInput, Asura_AMatrix3x3_OutParam xOutput) const
{
	xOutput	= Asura_AMatrix3x3(xInput.m_xCol0 * *this, xInput.m_xCol1 * *this, xInput.m_xCol2 * *this);
}

inline void Asura_AMatrix3x3::Transpose(void)
{
	float f = m_xCol0.GetY();
	m_xCol0.SetY( m_xCol1.GetX() );
	m_xCol1.SetX( f );

	f = m_xCol0.GetZ();
	m_xCol0.SetZ( m_xCol2.GetX() );
	m_xCol2.SetX( f );

	f = m_xCol1.GetZ();
	m_xCol1.SetZ( m_xCol2.GetY() );
	m_xCol2.SetY( f );
}

inline Asura_AVector3 Asura_AMatrix3x3::ApplyTranspose(Asura_AVector3_InParam xVec) const
{
	Asura_AMatrix3x3 xTemp(*this);
	xTemp.Transpose();
	return xVec * xTemp;
}

inline void Asura_AMatrix3x3::Identity(void)
{
	m_xCol0 = Asura_AVector3( 1.0f, 0.0f, 0.0f );
	m_xCol1 = Asura_AVector3( 0.0f, 1.0f, 0.0f );
	m_xCol2 = Asura_AVector3( 0.0f, 0.0f, 1.0f );
}

inline Asura_AMatrix3x3 Asura_AMatrix3x3::operator*(const float s) const
{
	return Asura_AMatrix3x3(m_xCol0 * s, m_xCol1 * s, m_xCol2 * s);
}

inline void Asura_AMatrix3x3::Zero(void)
{
	m_xCol0.Zero();
	m_xCol1.Zero();
	m_xCol2.Zero();
}

inline bool Asura_AMatrix3x3::operator==(Asura_AMatrix3x3_InParam xMat) const
{
	return	m_xCol0 == xMat.m_xCol0 &&
		m_xCol1 == xMat.m_xCol1 &&
		m_xCol2 == xMat.m_xCol2;
}

inline bool Asura_AMatrix3x3::operator!=(Asura_AMatrix3x3_InParam xMat) const
{
	return ! (*this==xMat);
}


//////////////////////////////////////////////////////////////////////////
// Asura_AVector4
//////////////////////////////////////////////////////////////////////////
inline Asura_AVector4::Asura_AVector4(void)
{
}

inline Asura_AVector4::Asura_AVector4(const float fX, const float fY, const float fZ, const float fW)
: m_fX(fX)
, m_fY(fY)
, m_fZ(fZ)
, m_fW(fW)
{
}

inline Asura_AVector4::Asura_AVector4(Asura_AVector4_InParam xOther)
: m_fX(xOther.m_fX)
, m_fY(xOther.m_fY)
, m_fZ(xOther.m_fZ)
, m_fW(xOther.m_fW)
{
}

inline Asura_AVector4::Asura_AVector4(const Asura_AVector3& xOther, const float fW)
: m_fX(xOther.GetX())
, m_fY(xOther.GetY())
, m_fZ(xOther.GetZ())
, m_fW(fW)
{
}

inline Asura_AVector4::Asura_AVector4(const Asura_Vector_3& xOther, const float fW)
: m_fX(xOther.x)
, m_fY(xOther.y)
, m_fZ(xOther.z)
, m_fW(fW)
{
}

inline void Asura_AVector4::ZeroW0(void) 
{
	m_fX = m_fY = m_fZ = m_fW = 0.0f;
}

inline void Asura_AVector4::ZeroW1(void) 
{
	m_fX = m_fY = m_fZ = 0.0f;
	m_fW = 1.0f;
}

inline void	Asura_AVector4::SelectGE(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS,
									 Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse)
{
	m_fX 	= (xLHS.m_fX >= xRHS.m_fX) ? xTrue.m_fX : xFalse.m_fX;
	m_fY 	= (xLHS.m_fY >= xRHS.m_fY) ? xTrue.m_fY : xFalse.m_fY;
	m_fZ 	= (xLHS.m_fZ >= xRHS.m_fZ) ? xTrue.m_fZ : xFalse.m_fZ;
	m_fW 	= (xLHS.m_fW >= xRHS.m_fW) ? xTrue.m_fW : xFalse.m_fW;
}

inline void	Asura_AVector4::SelectGT(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS,
									 Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse)
{
	m_fX 	= (xLHS.m_fX > xRHS.m_fX) ? xTrue.m_fX : xFalse.m_fX;
	m_fY 	= (xLHS.m_fY > xRHS.m_fY) ? xTrue.m_fY : xFalse.m_fY;
	m_fZ 	= (xLHS.m_fZ > xRHS.m_fZ) ? xTrue.m_fZ : xFalse.m_fZ;
	m_fW 	= (xLHS.m_fW > xRHS.m_fW) ? xTrue.m_fW : xFalse.m_fW;
}

inline void	Asura_AVector4::SelectLE(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS,
									 Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse)
{
	m_fX 	= (xLHS.m_fX <= xRHS.m_fX) ? xTrue.m_fX : xFalse.m_fX;
	m_fY 	= (xLHS.m_fY <= xRHS.m_fY) ? xTrue.m_fY : xFalse.m_fY;
	m_fZ 	= (xLHS.m_fZ <= xRHS.m_fZ) ? xTrue.m_fZ : xFalse.m_fZ;
	m_fW 	= (xLHS.m_fW <= xRHS.m_fW) ? xTrue.m_fW : xFalse.m_fW;
}

inline void	Asura_AVector4::SelectLT(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS,
									 Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse)
{
	m_fX 	= (xLHS.m_fX < xRHS.m_fX) ? xTrue.m_fX : xFalse.m_fX;
	m_fY 	= (xLHS.m_fY < xRHS.m_fY) ? xTrue.m_fY : xFalse.m_fY;
	m_fZ 	= (xLHS.m_fZ < xRHS.m_fZ) ? xTrue.m_fZ : xFalse.m_fZ;
	m_fW 	= (xLHS.m_fW < xRHS.m_fW) ? xTrue.m_fW : xFalse.m_fW;
}

inline void	Asura_AVector4::SelectEQ(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS,
									 Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse)
{
	m_fX 	= (xLHS.m_fX == xRHS.m_fX) ? xTrue.m_fX : xFalse.m_fX;
	m_fY 	= (xLHS.m_fY == xRHS.m_fY) ? xTrue.m_fY : xFalse.m_fY;
	m_fZ 	= (xLHS.m_fZ == xRHS.m_fZ) ? xTrue.m_fZ : xFalse.m_fZ;
	m_fW 	= (xLHS.m_fW == xRHS.m_fW) ? xTrue.m_fW : xFalse.m_fW;
}

inline void	Asura_AVector4::SelectNE(Asura_AVector4_InParam xLHS, Asura_AVector4_InParam xRHS,
									 Asura_AVector4_InParam xTrue, Asura_AVector4_InParam xFalse)
{
	m_fX 	= (xLHS.m_fX != xRHS.m_fX) ? xTrue.m_fX : xFalse.m_fX;
	m_fY 	= (xLHS.m_fY != xRHS.m_fY) ? xTrue.m_fY : xFalse.m_fY;
	m_fZ 	= (xLHS.m_fZ != xRHS.m_fZ) ? xTrue.m_fZ : xFalse.m_fZ;
	m_fW 	= (xLHS.m_fW != xRHS.m_fW) ? xTrue.m_fW : xFalse.m_fW;
}

inline void Asura_AVector4::operator=(Asura_AVector4_InParam xOther)
{
	m_fX	= xOther.m_fX;
	m_fY	= xOther.m_fY;
	m_fZ	= xOther.m_fZ;
	m_fW	= xOther.m_fW;
}

inline void Asura_AVector4::LoadAligned(const float* pfFloats)
{
	m_fX	= pfFloats[0];
	m_fY	= pfFloats[1];
	m_fZ	= pfFloats[2];
	m_fW	= pfFloats[3];
}

inline void Asura_AVector4::LoadUnaligned(const float* pfFloats)
{
	m_fX	= pfFloats[0];
	m_fY	= pfFloats[1];
	m_fZ	= pfFloats[2];
	m_fW	= pfFloats[3];
}

inline void Asura_AVector4::StoreAligned(float* pfFloats) const
{
	pfFloats[0]	= m_fX;
	pfFloats[1]	= m_fY;
	pfFloats[2]	= m_fZ;
	pfFloats[3]	= m_fW;
}

inline void Asura_AVector4::StoreUnaligned(float* pfFloats) const
{
	pfFloats[0]	= m_fX;
	pfFloats[1]	= m_fY;
	pfFloats[2]	= m_fZ;
	pfFloats[3]	= m_fW;
}

inline void Asura_AVector4::StoreAlignedHalf(Asura_HalfFloat* pxHalves) const
{
	pxHalves[0]	= m_fX;
	pxHalves[1]	= m_fY;
	pxHalves[2]	= m_fZ;
	pxHalves[3]	= m_fW;
}

inline Asura_AVector4 Asura_AVector4::ComponentwiseProduct(Asura_AVector4_InParam xOther) const
{
	return(Asura_AVector4((m_fX * xOther.m_fX), (m_fY * xOther.m_fY), (m_fZ * xOther.m_fZ), (m_fW * xOther.m_fW)));
}

inline float Asura_AVector4::DotProduct3(Asura_AVector4_InParam xOther) const
{
	return((m_fX * xOther.m_fX) + (m_fY * xOther.m_fY) + (m_fZ * xOther.m_fZ));
}

inline float Asura_AVector4::DotProduct4(Asura_AVector4_InParam xOther) const
{
	return((m_fX * xOther.m_fX) + (m_fY * xOther.m_fY) + (m_fZ * xOther.m_fZ) + (m_fW * xOther.m_fW));
}

inline float Asura_AVector4::MagnitudeSquared(void) const 
{
	return((m_fX * m_fX) + (m_fY * m_fY) + (m_fZ * m_fZ) + (m_fW * m_fW));
}

inline float Asura_AVector4::Magnitude(void) const 
{
	return(Asura_Maths::Sqrt(MagnitudeSquared()));
}

inline void Asura_AVector4::Normalise(void)
{
	const float	fMagnitude	= Magnitude();

	*this	/= fMagnitude;
}

inline float Asura_AVector4::NormaliseReturningMagnitude(void) 
{
	const float	fMagnitude	= Magnitude();

	if(fMagnitude == 0.0f)
	{
		return(0.0f);
	}
	else
	{
		*this	/= fMagnitude;
		return(fMagnitude);
	}
}

inline void Asura_AVector4::ApproximateNormalise(void)
{
	Normalise();
}

inline float Asura_AVector4::ApproximateNormaliseReturningMagnitude(void) 
{
	return(NormaliseReturningMagnitude());
}

inline void Asura_AVector4::operator+=(Asura_AVector4_InParam xOther)
{
	m_fX	+= xOther.m_fX;
	m_fY	+= xOther.m_fY;
	m_fZ	+= xOther.m_fZ;
	m_fW	+= xOther.m_fW;
}

inline void Asura_AVector4::operator-=(Asura_AVector4_InParam xOther)
{
	m_fX	-= xOther.m_fX;
	m_fY	-= xOther.m_fY;
	m_fZ	-= xOther.m_fZ;
	m_fW	-= xOther.m_fW;
}

inline void Asura_AVector4::operator*=(const float fScale) 
{
	m_fX	*= fScale;
	m_fY	*= fScale;
	m_fZ	*= fScale;
	m_fW	*= fScale;
}

inline void Asura_AVector4::operator/=(const float fScale) 
{
	m_fX	/= fScale;
	m_fY	/= fScale;
	m_fZ	/= fScale;
	m_fW	/= fScale;
}

inline Asura_AVector4 Asura_AVector4::operator-(void) const 
{
	return Asura_AVector4(-m_fX, -m_fY, -m_fZ, -m_fW);
}

inline Asura_AVector4 Asura_AVector4::operator+(Asura_AVector4_InParam xOther) const
{
	return(Asura_AVector4((m_fX + xOther.m_fX), (m_fY + xOther.m_fY), (m_fZ + xOther.m_fZ), (m_fW + xOther.m_fW)));
}

inline Asura_AVector4 Asura_AVector4::operator-( Asura_AVector4_InParam xOther) const
{
	return(Asura_AVector4((m_fX - xOther.m_fX), (m_fY - xOther.m_fY), (m_fZ - xOther.m_fZ), (m_fW - xOther.m_fW)));
}

inline Asura_AVector4 Asura_AVector4::operator^(Asura_AVector4_InParam xOther) const
{
	return(Asura_AVector4((m_fY * xOther.m_fZ) - (m_fZ * xOther.m_fY), (m_fZ * xOther.m_fX) - (m_fX * xOther.m_fZ), (m_fX * xOther.m_fY) - (m_fY * xOther.m_fX), 0.0f));
}

inline float Asura_AVector4::operator*(Asura_AVector4_InParam xOther) const
{
	return(DotProduct4(xOther));
}

inline Asura_AVector4 Asura_AVector4::operator*(const float fScale) const 
{
	return(Asura_AVector4((m_fX * fScale), (m_fY * fScale), (m_fZ * fScale), (m_fW * fScale)));
}

inline Asura_AVector4 Asura_AVector4::operator/(const float fScale) const 
{
	return(Asura_AVector4((m_fX / fScale), (m_fY / fScale), (m_fZ / fScale), (m_fW / fScale)));
}

inline bool Asura_AVector4::operator==(Asura_AVector4_InParam xOther) const
{
	return((m_fX == xOther.m_fX) && (m_fY == xOther.m_fY) && (m_fZ == xOther.m_fZ) && (m_fW == xOther.m_fW));
}

inline bool Asura_AVector4::ApproximatelyEquals(Asura_AVector4_InParam xOther, float fEpsilon) const
{
	Asura_AVector4	xDifference(*this - xOther);

	return(xDifference.MagnitudeSquared() < fEpsilon);
}

inline bool Asura_AVector4::operator!=(Asura_AVector4_InParam xOther) const
{
	return((m_fX != xOther.m_fX) || (m_fY != xOther.m_fY) || (m_fZ != xOther.m_fZ) || (m_fW != xOther.m_fW));
}

inline void Asura_AVector4::Max(Asura_AVector4_InParam xV1, Asura_AVector4_InParam xV2)
{
	m_fX	= Asura_Maths::Max(xV1.m_fX, xV2.m_fX);
	m_fY	= Asura_Maths::Max(xV1.m_fY, xV2.m_fY);
	m_fZ	= Asura_Maths::Max(xV1.m_fZ, xV2.m_fZ);
	m_fW	= Asura_Maths::Max(xV1.m_fW, xV2.m_fW);
}
inline void Asura_AVector4::Min(Asura_AVector4_InParam xV1, Asura_AVector4_InParam xV2)
{
	m_fX	= Asura_Maths::Min(xV1.m_fX, xV2.m_fX);
	m_fY	= Asura_Maths::Min(xV1.m_fY, xV2.m_fY);
	m_fZ	= Asura_Maths::Min(xV1.m_fZ, xV2.m_fZ);
	m_fW	= Asura_Maths::Min(xV1.m_fW, xV2.m_fW);
}

inline float Asura_AVector4::GetX(void) const
{
	return(m_fX);
}

inline float Asura_AVector4::GetY(void) const
{ 
	return(m_fY);
}

inline float Asura_AVector4::GetZ(void) const
{
	return(m_fZ);
}

inline float Asura_AVector4::GetW(void) const
{
	return(m_fW);
}

inline void Asura_AVector4::SetX(const float fX)
{
	m_fX	= fX;
}

inline void Asura_AVector4::SetY(const float fY)
{
	m_fY	= fY;
}

inline void Asura_AVector4::SetZ(const float fZ)
{
	m_fZ	= fZ;
}

inline void Asura_AVector4::SetW(const float fW)
{
	m_fW	= fW;
}

inline float Asura_AVector4::GetElement(const u_int uIndex) const
{ 
	return(m_afElements[uIndex]);
}

inline void Asura_AVector4::SetElement(const u_int uIndex, const float fValue) 
{
	m_afElements[uIndex]	= fValue;
}		

inline void Asura_AVector4::SetElements(const float fX, const float fY, const float fZ, const float fW)
{
	m_fX	= fX;
	m_fY	= fY;
	m_fZ	= fZ;
	m_fW	= fW;
}


//////////////////////////////////////////////////////////////////////////
// Asura_AMatrix4x4
//////////////////////////////////////////////////////////////////////////
inline Asura_AMatrix4x4::Asura_AMatrix4x4(void)
{
}

inline Asura_AMatrix4x4::Asura_AMatrix4x4(const float fE11, const float fE12, const float fE13, const float fE14,
										  const float fE21, const float fE22, const float fE23, const float fE24,
										  const float fE31, const float fE32, const float fE33, const float fE34,
										  const float fE41, const float fE42, const float fE43, const float fE44)
{
	m_axColumns[0].SetElements(fE11, fE21, fE31, fE41);
	m_axColumns[1].SetElements(fE12, fE22, fE32, fE42);
	m_axColumns[2].SetElements(fE13, fE23, fE33, fE43);
	m_axColumns[3].SetElements(fE14, fE24, fE34, fE44);
}

inline Asura_AMatrix4x4::Asura_AMatrix4x4(const Asura_AMatrix4x4& xOther)
{
	m_axColumns[0]	= xOther.m_axColumns[0];
	m_axColumns[1]	= xOther.m_axColumns[1];
	m_axColumns[2]	= xOther.m_axColumns[2];
	m_axColumns[3]	= xOther.m_axColumns[3];
}

inline Asura_AMatrix4x4::Asura_AMatrix4x4(const Asura_Matrix_3x3& xRotation, const Asura_Vector_3& xPosition)
{
	m_axColumns[0].SetElements(xRotation.e11,		xRotation.e21,		xRotation.e31,		0.0f);
	m_axColumns[1].SetElements(xRotation.e12,		xRotation.e22,		xRotation.e32,		0.0f);
	m_axColumns[2].SetElements(xRotation.e13,		xRotation.e23,		xRotation.e33,		0.0f);
	m_axColumns[3].SetElements(xPosition.GetX(),	xPosition.GetY(),	xPosition.GetZ(),	1.0f);
}

inline Asura_AMatrix4x4::Asura_AMatrix4x4(const Asura_Quat& xRotation, const Asura_Vector_3& xPosition)
{
	const float	fS	= 2.0f;	//xRotation.Magnitude();

	const float	fXS	= xRotation.x * fS;
	const float	fYS	= xRotation.y * fS;
	const float	fZS	= xRotation.z * fS;

	const float	fWX	= xRotation.w * fXS;
	const float	fWY	= xRotation.w * fYS;
	const float	fWZ	= xRotation.w * fZS;

	const float	fXX	= xRotation.x * fXS;
	const float	fXY	= xRotation.x * fYS;
	const float	fXZ	= xRotation.x * fZS;

	const float	fYY	= xRotation.y * fYS;
	const float	fYZ	= xRotation.y * fZS;
	const float	fZZ	= xRotation.z * fZS;

	// Implementation from Graphics Gems IV...
	m_axColumns[0].SetElements(1.0f - (fYY + fZZ),	fXY + fWZ,			fXZ - fWY,			0.0f);
	m_axColumns[1].SetElements(fXY - fWZ,			1.0f - (fXX + fZZ),	fYZ + fWX,			0.0f);
	m_axColumns[2].SetElements(fXZ + fWY,			fYZ - fWX,			1.0f - (fXX + fYY),	0.0f);
	m_axColumns[3].SetElements(xPosition.GetX(),	xPosition.GetY(),	xPosition.GetZ(),	1.0f);
}

inline void Asura_AMatrix4x4::SetColumns(Asura_AVector4_InParam xCol0, Asura_AVector4_InParam xCol1, Asura_AVector4_InParam xCol2, Asura_AVector4_InParam xCol3)
{
	m_axColumns[0]	= xCol0;
	m_axColumns[1]	= xCol1;
	m_axColumns[2]	= xCol2;
	m_axColumns[3]	= xCol3;
}

inline void Asura_AMatrix4x4::SetRows(Asura_AVector4_InParam xRow0, Asura_AVector4_InParam xRow1, Asura_AVector4_InParam xRow2, Asura_AVector4_InParam xRow3)
{
	SetRow0(xRow0);
	SetRow1(xRow1);
	SetRow2(xRow2);
	SetRow3(xRow3);
}

inline void Asura_AMatrix4x4::Identity(void)
{
	m_axColumns[0].SetElements(1.0f, 0.0f, 0.0f, 0.0f);
	m_axColumns[1].SetElements(0.0f, 1.0f, 0.0f, 0.0f);
	m_axColumns[2].SetElements(0.0f, 0.0f, 1.0f, 0.0f);
	m_axColumns[3].ZeroW1();
}

inline void Asura_AMatrix4x4::Zero(void)
{
	m_axColumns[0].ZeroW0();
	m_axColumns[1].ZeroW0();
	m_axColumns[2].ZeroW0();
	m_axColumns[3].ZeroW0();
}

inline void Asura_AMatrix4x4::ToMatrix(Asura_Matrix_3x3& xOrientation) const
{
	xOrientation.e11	= m_axColumns[0].GetX();
	xOrientation.e21	= m_axColumns[0].GetY();
	xOrientation.e31	= m_axColumns[0].GetZ();

	xOrientation.e12	= m_axColumns[1].GetX();
	xOrientation.e22	= m_axColumns[1].GetY();
	xOrientation.e32	= m_axColumns[1].GetZ();

	xOrientation.e13	= m_axColumns[2].GetX();
	xOrientation.e23	= m_axColumns[2].GetY();
	xOrientation.e33	= m_axColumns[2].GetZ();
}

inline void Asura_AMatrix4x4::FromMatrix(const Asura_Matrix_3x3& xOrientation)
{
	m_axColumns[0].SetElements(xOrientation.e11, xOrientation.e21, xOrientation.e31, 0.0f);
	m_axColumns[1].SetElements(xOrientation.e12, xOrientation.e22, xOrientation.e32, 0.0f);
	m_axColumns[2].SetElements(xOrientation.e13, xOrientation.e23, xOrientation.e33, 0.0f);
	m_axColumns[3].ZeroW1();
}

inline Asura_Vector_3 Asura_AMatrix4x4::GetTranslation() const
{
	return Asura_Vector_3(m_axColumns[3].GetX(),m_axColumns[3].GetY(),m_axColumns[3].GetZ());
}

inline void Asura_AMatrix4x4::SetTranslation(const Asura_Vector_3& xTranslation)
{
	m_axColumns[3]	= Asura_AVector4(xTranslation, 1.0f);
}

inline Asura_AVector3 Asura_AMatrix4x4::GetATranslation() const
{
	return Asura_AVector3(m_axColumns[3].GetX(),m_axColumns[3].GetY(),m_axColumns[3].GetZ());
}

inline void Asura_AMatrix4x4::SetTranslation(Asura_AVector3_InParam xTranslation)
{
	m_axColumns[3].SetElements(xTranslation.GetX(), xTranslation.GetY(), xTranslation.GetZ(), 1.0f);
}

inline Asura_AVector4 Asura_AMatrix4x4::GetA4Translation() const
{
	return m_axColumns[3];
}

inline void Asura_AMatrix4x4::SetTranslation(Asura_AVector4_InParam xTranslation)
{
	m_axColumns[3]	= xTranslation;
}

inline Asura_AMatrix4x4 Asura_AMatrix4x4::operator-(void) const
{
	Asura_AMatrix4x4	xTemp;

	xTemp.SetColumns(-m_axColumns[0], -m_axColumns[1], -m_axColumns[2], -m_axColumns[3]);
	return(xTemp);
}

inline Asura_AVector4 Asura_AMatrix4x4::Apply(Asura_AVector4_InParam xInput) const
{
	Asura_AVector4	axRows[4];

	// Extract the data...
	axRows[0] = GetRow0();
	axRows[1] = GetRow1();
	axRows[2] = GetRow2();
	axRows[3] = GetRow3();

	// Set each element of the result in turn...
	return Asura_AVector4((axRows[0] * xInput), (axRows[1] * xInput), (axRows[2] * xInput), (axRows[3] * xInput));
}

inline void Asura_AMatrix4x4::ApplyToArray(const Asura_AVector4* pxInput, const u_int uNumElements, Asura_AVector4* pxOutput) const
{
	Asura_AVector4	axRows[4];

	// Extract the data...
	axRows[0] = GetRow0();
	axRows[1] = GetRow1();
	axRows[2] = GetRow2();
	axRows[3] = GetRow3();

	for(u_int u = 0; u < uNumElements; u++, pxInput++, pxOutput++)
	{
		// Set each element of the result in turn...
		pxOutput->SetElements((axRows[0] * *pxInput), (axRows[1] * *pxInput), (axRows[2] * *pxInput), (axRows[3] * *pxInput));
	}
}

inline void Asura_AMatrix4x4::Apply(Asura_AMatrix4x4_InParam xInput, Asura_AMatrix4x4_OutParam xOutput) const
{
	Asura_AVector4	axRows[4];
	Asura_AVector4	axColumns[4];

	// Extract the data...
	axRows[0] = GetRow0();
	axRows[1] = GetRow1();
	axRows[2] = GetRow2();
	axRows[3] = GetRow3();
	axColumns[0] = xInput.GetColumn0();
	axColumns[1] = xInput.GetColumn1();
	axColumns[2] = xInput.GetColumn2();
	axColumns[3] = xInput.GetColumn3();

	// Set each element of the result in turn...
	xOutput.SetColumn0(Asura_AVector4(axColumns[0] * axRows[0], axColumns[0] * axRows[1], axColumns[0] * axRows[2], axColumns[0] * axRows[3]));
	xOutput.SetColumn1(Asura_AVector4(axColumns[1] * axRows[0], axColumns[1] * axRows[1], axColumns[1] * axRows[2], axColumns[1] * axRows[3]));
	xOutput.SetColumn2(Asura_AVector4(axColumns[2] * axRows[0], axColumns[2] * axRows[1], axColumns[2] * axRows[2], axColumns[2] * axRows[3]));
	xOutput.SetColumn3(Asura_AVector4(axColumns[3] * axRows[0], axColumns[3] * axRows[1], axColumns[3] * axRows[2], axColumns[3] * axRows[3]));
}

inline void Asura_AMatrix4x4::ApplyToArray(const Asura_AMatrix4x4* pxInput, const u_int uNumElements, Asura_AMatrix4x4* pxOutput) const
{
	Asura_AVector4	axRows[4];

	// Extract the row-data...
	axRows[0] = GetRow0();
	axRows[1] = GetRow1();
	axRows[2] = GetRow2();
	axRows[3] = GetRow3();

	for(u_int u = 0; u < uNumElements; u++, pxInput++, pxOutput++)
	{
		Asura_AVector4	axColumns[4];

		// Extract the column-data...
		axColumns[0] = pxInput->GetColumn0();
		axColumns[1] = pxInput->GetColumn1();
		axColumns[2] = pxInput->GetColumn2();
		axColumns[3] = pxInput->GetColumn3();

		// Set each element of the result in turn...
		pxOutput->SetColumn0(Asura_AVector4(axColumns[0] * axRows[0], axColumns[0] * axRows[1], axColumns[0] * axRows[2], axColumns[0] * axRows[3]));
		pxOutput->SetColumn1(Asura_AVector4(axColumns[1] * axRows[0], axColumns[1] * axRows[1], axColumns[1] * axRows[2], axColumns[1] * axRows[3]));
		pxOutput->SetColumn2(Asura_AVector4(axColumns[2] * axRows[0], axColumns[2] * axRows[1], axColumns[2] * axRows[2], axColumns[2] * axRows[3]));
		pxOutput->SetColumn3(Asura_AVector4(axColumns[3] * axRows[0], axColumns[3] * axRows[1], axColumns[3] * axRows[2], axColumns[3] * axRows[3]));
	}
}

inline float Asura_AMatrix4x4::Cofactor(const u_int uRow, const u_int uCol) const
{
	static const u_int	aauIndices[4][3]	= { {1, 2, 3}, {0, 2, 3}, {0, 1, 3}, {0, 1, 2} };
	const u_int*		puRowsToUse			= aauIndices[uRow];
	const u_int*		puColsToUse			= aauIndices[uCol];

	// Calculate the determinant of the sub-matrix produced by ignoring row uRow and column uCol
	const float	fA	= GetElement(puRowsToUse[0], puColsToUse[0]);
	const float	fB	= GetElement(puRowsToUse[0], puColsToUse[1]);
	const float	fC	= GetElement(puRowsToUse[0], puColsToUse[2]);

	const float	fD	= GetElement(puRowsToUse[1], puColsToUse[0]);
	const float	fE	= GetElement(puRowsToUse[1], puColsToUse[1]);
	const float	fF	= GetElement(puRowsToUse[1], puColsToUse[2]);

	const float	fG	= GetElement(puRowsToUse[2], puColsToUse[0]);
	const float	fH	= GetElement(puRowsToUse[2], puColsToUse[1]);
	const float	fI	= GetElement(puRowsToUse[2], puColsToUse[2]);

	return( ((fA * fE * fI) + (fD * fH * fC) + (fG * fB * fF)) -
			((fA * fH * fF) + (fD * fB * fI) + (fG * fE * fC)) );
}

inline float Asura_AMatrix4x4::Determinant(void) const
{
	const float	fElement00	= GetElement(0, 0);
	const float	fElement01	= GetElement(0, 1);
	const float	fElement02	= GetElement(0, 2);
	const float	fElement03	= GetElement(0, 3);

	const float fCofactor00	= Cofactor(0, 0);
	const float fCofactor01	= Cofactor(0, 1);
	const float fCofactor02	= Cofactor(0, 2);
	const float fCofactor03	= Cofactor(0, 3);

	return((fElement00 * fCofactor00) - (fElement01 * fCofactor01) + (fElement02 * fCofactor02) - (fElement03 * fCofactor03));
}

inline void Asura_AMatrix4x4::Invert(Asura_AMatrix4x4_OutParam xOther) const
{
	const float			fDeterminant		= Determinant();
	const float			fOneOverDeterminant	= 1.0f / fDeterminant;
	Asura_AMatrix4x4	xResult;

	// Calculate using the adjoint matrix approach...
	Asura_Assert(fDeterminant != 0.f, "Trying to calculate the inverse of a non-invertable matrix");
	for(int iRow = 0; iRow < 4; ++iRow)
	{
		for(int iCol = 0; iCol < 4; ++iCol)
		{
			const float	fSign		= (((iRow + iCol) & 1) == 0) ? 1.0f : -1.0f;
			const float	fCofactor	= Cofactor(iRow, iCol);	// note flipped uCol/uRow

			xOther.SetElement(iCol, iRow, fSign * fCofactor * fOneOverDeterminant);
		}
	}
}

inline void Asura_AMatrix4x4::Invert(void)
{
	const Asura_AMatrix4x4	xTemp(*this);

	xTemp.Invert(*this);
}

inline void Asura_AMatrix4x4::InvertOrthogonal(Asura_AMatrix4x4_OutParam xOther) const
{
	// Create the transposed orientation matrix...
	xOther.m_axColumns[0].SetElements(m_axColumns[0].GetX(), m_axColumns[1].GetX(), m_axColumns[2].GetX(), 0.0f);
	xOther.m_axColumns[1].SetElements(m_axColumns[0].GetY(), m_axColumns[1].GetY(), m_axColumns[2].GetY(), 0.0f);
	xOther.m_axColumns[2].SetElements(m_axColumns[0].GetZ(), m_axColumns[1].GetZ(), m_axColumns[2].GetZ(), 0.0f);
	xOther.m_axColumns[3].ZeroW1();

	// ...apply the transpose to the position and subtract this from the final translation...
	xOther.m_axColumns[3]	-= m_axColumns[3] * xOther;
	xOther.m_axColumns[3].SetW(1.0f);
}

inline void Asura_AMatrix4x4::InvertOrthogonal(void)
{
	const Asura_AMatrix4x4	xTemp(*this);

	xTemp.InvertOrthogonal(*this);
}

inline void Asura_AMatrix4x4::Transpose(Asura_AMatrix4x4_OutParam xOther) const
{
	xOther.SetRows(m_axColumns[0], m_axColumns[1], m_axColumns[2], m_axColumns[3]);
}

inline void Asura_AMatrix4x4::Transpose(void)
{
	const Asura_AMatrix4x4	xTemp(*this);

	xTemp.Transpose(*this);
}

inline Asura_AMatrix4x4 Asura_AMatrix4x4::operator*(const float fScale) const
{
	Asura_AMatrix4x4	xTemp(*this);

	xTemp.m_axColumns[0]	*= fScale;
	xTemp.m_axColumns[1]	*= fScale;
	xTemp.m_axColumns[2]	*= fScale;
	xTemp.m_axColumns[3]	*= fScale;
	return(xTemp);
}

inline Asura_AMatrix4x4 Asura_AMatrix4x4::operator/(const float fScale) const
{
	const float			fOneOverScale	= 1.0f / fScale;
	Asura_AMatrix4x4	xTemp(*this);

	xTemp.m_axColumns[0]	*= fOneOverScale;
	xTemp.m_axColumns[1]	*= fOneOverScale;
	xTemp.m_axColumns[2]	*= fOneOverScale;
	xTemp.m_axColumns[3]	*= fOneOverScale;
	return(xTemp);
}

inline void Asura_AMatrix4x4::operator=(Asura_AMatrix4x4_InParam xM)
{
	m_axColumns[0]	= xM.m_axColumns[0];
	m_axColumns[1]	= xM.m_axColumns[1];
	m_axColumns[2]	= xM.m_axColumns[2];
	m_axColumns[3]	= xM.m_axColumns[3];
}

inline void Asura_AMatrix4x4::operator*=(const float fScale)
{
	m_axColumns[0]	*= fScale;
	m_axColumns[1]	*= fScale;
	m_axColumns[2]	*= fScale;
	m_axColumns[3]	*= fScale;
}

inline void Asura_AMatrix4x4::operator/=(const float fScale)
{
	const float	fOneOverScale	= 1.0f / fScale;

	m_axColumns[0]	*= fOneOverScale;
	m_axColumns[1]	*= fOneOverScale;
	m_axColumns[2]	*= fOneOverScale;
	m_axColumns[3]	*= fOneOverScale;
}

inline bool Asura_AMatrix4x4::operator==(Asura_AMatrix4x4_InParam xOther) const
{
	return (m_axColumns[0] == xOther.m_axColumns[0])
		&& (m_axColumns[1] == xOther.m_axColumns[1])
		&& (m_axColumns[2] == xOther.m_axColumns[2])
		&& (m_axColumns[3] == xOther.m_axColumns[3]);
}

inline bool Asura_AMatrix4x4::ApproximatelyEquals(Asura_AMatrix4x4_InParam xOther, const float fEpsilon) const
{
	return (m_axColumns[0].ApproximatelyEquals(xOther.m_axColumns[0], fEpsilon))
		&& (m_axColumns[1].ApproximatelyEquals(xOther.m_axColumns[1], fEpsilon))
		&& (m_axColumns[2].ApproximatelyEquals(xOther.m_axColumns[2], fEpsilon))
		&& (m_axColumns[3].ApproximatelyEquals(xOther.m_axColumns[3], fEpsilon));
}

inline bool Asura_AMatrix4x4::operator!=(Asura_AMatrix4x4_InParam xOther) const
{
	return (m_axColumns[0] != xOther.m_axColumns[0])
		|| (m_axColumns[1] != xOther.m_axColumns[1])
		|| (m_axColumns[2] != xOther.m_axColumns[2])
		|| (m_axColumns[3] != xOther.m_axColumns[3]);
}

inline float Asura_AMatrix4x4::GetElement(const u_int uRow, const u_int uCol) const
{
	Asura_Assert(uRow <= 3, "Invalid matrix row index passed to Asura_AMatrix4x4::GetElement");
	Asura_Assert(uCol <= 3, "Invalid matrix column index passed to Asura_AMatrix4x4::GetElement");
	return(m_axColumns[uCol].GetElement(uRow));
}

inline void Asura_AMatrix4x4::SetElement(const u_int uRow, const u_int uCol, const float fValue)
{ 
	Asura_Assert(uRow <= 3, "Invalid matrix row index passed to Asura_AMatrix4x4::SetElement");
	Asura_Assert(uCol <= 3, "Invalid matrix column index passed to Asura_AMatrix4x4::SetElement");
	m_axColumns[uCol].SetElement(uRow, fValue);
}

inline Asura_AVector4 Asura_AMatrix4x4::GetColumn(const u_int uColumnIndex) const
{
	Asura_Assert(uColumnIndex <= 3, "Invalid matrix column index passed to Asura_AMatrix4x4::GetColumn");
	return m_axColumns[uColumnIndex];
}

inline Asura_AVector4 Asura_AMatrix4x4::GetColumn0() const
{
	return m_axColumns[0];
}

inline Asura_AVector4 Asura_AMatrix4x4::GetColumn1() const
{
	return m_axColumns[1];
}

inline Asura_AVector4 Asura_AMatrix4x4::GetColumn2() const
{
	return m_axColumns[2];
}

inline Asura_AVector4 Asura_AMatrix4x4::GetColumn3() const
{
	return m_axColumns[3];
}
inline void Asura_AMatrix4x4::SetColumn(const u_int uColumnIndex, Asura_AVector4_InParam xColumn)
{
	Asura_Assert(uColumnIndex <= 3, "Invalid matrix column index passed to Asura_AMatrix4x4::SetColumn");
	m_axColumns[uColumnIndex] = xColumn;
}

inline void Asura_AMatrix4x4::SetColumn0(Asura_AVector4_InParam xCol0)
{
	m_axColumns[0] = xCol0;
}

inline void Asura_AMatrix4x4::SetColumn1(Asura_AVector4_InParam xCol1)
{
	m_axColumns[1] = xCol1;
}

inline void Asura_AMatrix4x4::SetColumn2(Asura_AVector4_InParam xCol2)
{
	m_axColumns[2] = xCol2;
}

inline void Asura_AMatrix4x4::SetColumn3(Asura_AVector4_InParam xCol3)
{
	m_axColumns[3] = xCol3;
}

inline Asura_AVector4 Asura_AMatrix4x4::GetRow(const u_int uRowIndex) const
{
	Asura_Assert(uRowIndex <= 3, "Invalid matrix row index passed to Asura_AMatrix4x4::GetRow");
	return Asura_AVector4(m_axColumns[0].GetElement(uRowIndex), m_axColumns[1].GetElement(uRowIndex),
						  m_axColumns[2].GetElement(uRowIndex), m_axColumns[3].GetElement(uRowIndex));
}

inline Asura_AVector4 Asura_AMatrix4x4::GetRow0() const
{
	return Asura_AVector4(m_axColumns[0].GetElement(0), m_axColumns[1].GetElement(0),
						  m_axColumns[2].GetElement(0), m_axColumns[3].GetElement(0));
}

inline Asura_AVector4 Asura_AMatrix4x4::GetRow1() const
{
	return Asura_AVector4(m_axColumns[0].GetElement(1), m_axColumns[1].GetElement(1),
						  m_axColumns[2].GetElement(1), m_axColumns[3].GetElement(1));
}

inline Asura_AVector4 Asura_AMatrix4x4::GetRow2() const
{
	return Asura_AVector4(m_axColumns[0].GetElement(2), m_axColumns[1].GetElement(2),
						  m_axColumns[2].GetElement(2), m_axColumns[3].GetElement(2));
}

inline Asura_AVector4 Asura_AMatrix4x4::GetRow3() const
{
	return Asura_AVector4(m_axColumns[0].GetElement(3), m_axColumns[1].GetElement(3),
						  m_axColumns[2].GetElement(3), m_axColumns[3].GetElement(3));
}

inline void Asura_AMatrix4x4::SetRow(const u_int uRowIndex, Asura_AVector4_InParam xRow)
{
	Asura_Assert(uRowIndex <= 3, "Invalid matrix row index passed to Asura_AMatrix4x4::SetRow");
	m_axColumns[0].SetElement(uRowIndex, xRow.GetElement(0));
	m_axColumns[1].SetElement(uRowIndex, xRow.GetElement(1));
	m_axColumns[2].SetElement(uRowIndex, xRow.GetElement(2));
	m_axColumns[3].SetElement(uRowIndex, xRow.GetElement(3));
}

inline void Asura_AMatrix4x4::SetRow0(Asura_AVector4_InParam xRow0)
{
	m_axColumns[0].SetElement(0, xRow0.GetElement(0));
	m_axColumns[1].SetElement(0, xRow0.GetElement(1));
	m_axColumns[2].SetElement(0, xRow0.GetElement(2));
	m_axColumns[3].SetElement(0, xRow0.GetElement(3));
}

inline void Asura_AMatrix4x4::SetRow1(Asura_AVector4_InParam xRow1)
{
	m_axColumns[0].SetElement(1, xRow1.GetElement(0));
	m_axColumns[1].SetElement(1, xRow1.GetElement(1));
	m_axColumns[2].SetElement(1, xRow1.GetElement(2));
	m_axColumns[3].SetElement(1, xRow1.GetElement(3));
}

inline void Asura_AMatrix4x4::SetRow2(Asura_AVector4_InParam xRow2)
{
	m_axColumns[0].SetElement(2, xRow2.GetElement(0));
	m_axColumns[1].SetElement(2, xRow2.GetElement(1));
	m_axColumns[2].SetElement(2, xRow2.GetElement(2));
	m_axColumns[3].SetElement(2, xRow2.GetElement(3));
}

inline void Asura_AMatrix4x4::SetRow3(Asura_AVector4_InParam xRow3)
{
	m_axColumns[0].SetElement(3, xRow3.GetElement(0));
	m_axColumns[1].SetElement(3, xRow3.GetElement(1));
	m_axColumns[2].SetElement(3, xRow3.GetElement(2));
	m_axColumns[3].SetElement(3, xRow3.GetElement(3));
}

inline void Asura_AMatrix4x4::LoadColsAligned(const float* pfFloats)
{
	m_axColumns[0].LoadAligned(&pfFloats[0]);
	m_axColumns[1].LoadAligned(&pfFloats[4]);
	m_axColumns[2].LoadAligned(&pfFloats[8]);
	m_axColumns[3].LoadAligned(&pfFloats[12]);
}

inline void Asura_AMatrix4x4::LoadRowsAligned(const float* pfFloats) 
{
	LoadColsAligned(pfFloats);
	Transpose();
}

inline void Asura_AMatrix4x4::LoadColsUnaligned(const float* pfFloats)
{
	m_axColumns[0].LoadUnaligned(&pfFloats[0]);
	m_axColumns[1].LoadUnaligned(&pfFloats[4]);
	m_axColumns[2].LoadUnaligned(&pfFloats[8]);
	m_axColumns[3].LoadUnaligned(&pfFloats[12]);
}

inline void Asura_AMatrix4x4::LoadRowsUnaligned(const float* pfFloats)
{
	LoadColsUnaligned(pfFloats);
	Transpose();
}

inline void Asura_AMatrix4x4::StoreColsAligned(float* pfFloats) const
{
	m_axColumns[0].StoreAligned(&pfFloats[0]);
	m_axColumns[1].StoreAligned(&pfFloats[4]);
	m_axColumns[2].StoreAligned(&pfFloats[8]);
	m_axColumns[3].StoreAligned(&pfFloats[12]);
}

inline void Asura_AMatrix4x4::StoreRowsAligned(float* pfFloats) const
{
	Asura_AMatrix4x4	xTemp;

	Transpose(xTemp);
	xTemp.StoreColsAligned(pfFloats);
}

inline void Asura_AMatrix4x4::StoreColsUnaligned(float* pfFloats) const
{
	m_axColumns[0].StoreUnaligned(&pfFloats[0]);
	m_axColumns[1].StoreUnaligned(&pfFloats[4]);
	m_axColumns[2].StoreUnaligned(&pfFloats[8]);
	m_axColumns[3].StoreUnaligned(&pfFloats[12]);
}

inline void Asura_AMatrix4x4::StoreRowsUnaligned(float* pfFloats) const
{
	Asura_AMatrix4x4	xTemp;

	Transpose(xTemp);
	xTemp.StoreColsUnaligned(pfFloats);
}

#endif	// !defined(ASURA_USE_PLATFORM_VECTORS)

#endif   // ifndef __ASURA_REFERENCE_VECTORMATHS_H__


/// End of file. ///
