//////////////////////////////////////////////////////////////////////////////
//
// Asura_Scheduler.cpp
//
// Created by NJD - 14 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Core/Asura_Scheduler.h"

// asura
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Font_System.h"

#if ASURA_SCHEDULER_DEBUG == 2
    #define ADD_TO_MY_SCHEDULER_LOG_LIGHT( args ) { this->AddToSchedulerLog args ; }
    #define ADD_TO_MY_SCHEDULER_LOG( args )       { this->AddToSchedulerLog args ; }
#elif ASURA_SCHEDULER_DEBUG == 1
    #define ADD_TO_MY_SCHEDULER_LOG_LIGHT( args ) { this->AddToSchedulerLog args ; }
    #define ADD_TO_MY_SCHEDULER_LOG( args )       { }
#else
    #define ADD_TO_MY_SCHEDULER_LOG_LIGHT( args ) { }
    #define ADD_TO_MY_SCHEDULER_LOG( args )       { }
#endif

// We have no useful data for new callbacks as to how long they're likely to
// take so base it on some proportion of the scheduler's max per-frame time.
// (MP - 13.11.2008) NB: Previously we just used the callback's desired max
// update period so this is almost certainly always going to be better than
// that at least. Other options: (i) use an invalid value for the average
// exec time initially and special-case it elsewhere; (ii) the caller could
// provide an estimate.
DEBUGVAR float dbg_fSchedulerNewCallbackDurationEstimateFactor = 0.5f;

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Scheduler
//
//////////////////////////////////////////////////////////////////////////////

Asura_ResizingMempool Asura_Scheduler::s_xQueueListNodeMempool( 16, Asura_List< QueueEntry >::GetMemoryElementSize(), __FILE__, __LINE__ );
Asura_ResizingMempool Asura_Scheduler::s_xQueueEntryMempool( 16, sizeof( QueueEntry ), __FILE__, __LINE__ );
// We take advantage of the fact that all callbacks, no matter what type, are the same size.
// Note that the template argument is just a dummy, and has no meaning.
// RAR 13/05/2008 They may not all be the same size now,  so will need to assert to make sure we don't get one that is too large
Asura_ResizingMempool Asura_Scheduler::s_xCallbackMempool( 16, sizeof( Asura_Callback_MemberFunction_1Arg< void, float, Asura_Scheduler > ), __FILE__, __LINE__ );

Asura_Scheduler::Asura_Scheduler()
: m_fMaxPerFrameTime( 0.0075f )
, m_iNextHandle( 0 )
, m_bMidUpdate( false )
, m_xQueue()
, m_xToBeRegistered()
, m_xToBeUnregistered()
#if ASURA_SCHEDULER_DEBUG
, m_fAvExecutionPeriodFraction( 0.0f )
, m_fAvPerFrameTime( 0.0f )
, m_fAvTasksPerFrame( 0.0f )
, m_fAvPreemptiveExecutionsPerFrame( 0.0f )
, m_fAvOverExecutionsPerFrame( 0.0f )
, m_szPerFrameLogFile( NULL )
, m_szAccumulatedLog( NULL )
, m_uAccumulatedLogLength( 0 )
, m_bNeedToClearLogFile( true )
#endif
, m_fPrevScaledTime( Asura_Timers::GetLocalScaledTime() )
, m_fAvTimeBetweenUpdates( 0.0f )
{
    // InsertInList and RemoveFromList are called during the Update loop.
    // These will hammer new and delete unless we specify our own memory supply and return functions.
    m_xQueue.SetMemorySupplyFunction( AllocateFromQueueListNodeMempool );
    m_xQueue.SetMemoryReturnFunction( DeallocateToQueueListNodeMempool );
}

Asura_Scheduler::~Asura_Scheduler()
{
    // Make sure we free all of our memory.
    Asura_List_It< QueueEntry > xQueueIter( &m_xQueue );
    while ( !xQueueIter.Done() )
    {
        QueueEntry* pxEntry = xQueueIter.GetCurrent();
        xQueueIter.DeleteCurrent();
        s_xCallbackMempool.Deallocate( pxEntry->m_pxCallback );
        s_xQueueEntryMempool.Deallocate( pxEntry );
    }

    Asura_List_It< QueueEntry > xToBeRegisteredIter( &m_xToBeRegistered );
    while ( !xToBeRegisteredIter.Done() )
    {
        QueueEntry* pxEntry = xToBeRegisteredIter.GetCurrent();
        xToBeRegisteredIter.DeleteCurrent();
        s_xCallbackMempool.Deallocate( pxEntry->m_pxCallback );
        s_xQueueEntryMempool.Deallocate( pxEntry );
    }

    Asura_List_It< Asura_Handle > xToBeUnregisteredIter( &m_xToBeUnregistered );
    while ( !xToBeUnregisteredIter.Done() )
    {
        Asura_Handle* piToBeUnregistered = xToBeUnregisteredIter.GetCurrent();
        xToBeUnregisteredIter.DeleteCurrent();
        delete piToBeUnregistered;
    }

    Asura_Assert( s_xQueueEntryMempool.IsEmpty(), "Asura_Scheduler::~Asura_Scheduler - there are queue entries still allocated" );
    Asura_Assert( s_xCallbackMempool.IsEmpty(),   "Asura_Scheduler::~Asura_Scheduler - there are callbacks still allocated" );

#if ASURA_SCHEDULER_DEBUG
    FlushSchedulerLog( true );
    delete[] m_szAccumulatedLog;
#endif
}


void Asura_Scheduler::Update()
{
    m_bMidUpdate = true;

    const float fScaledTimeAtStartOfFrame = Asura_Timers::GetLocalScaledTime();
    const float fInitialTimeInFrame = Asura_Timers::GetTimeInFrame();
    const float fInitialLocalScaledTime = fScaledTimeAtStartOfFrame + fInitialTimeInFrame;

    m_fAvTimeBetweenUpdates += ((fInitialLocalScaledTime - m_fPrevScaledTime) - m_fAvTimeBetweenUpdates) * 0.01f;
    m_fPrevScaledTime = fInitialLocalScaledTime;

#if ASURA_SCHEDULER_DEBUG
    int iNumEntriesExecuted = 0;
    int iNumEntriesPreemptivelyExecuted = 0;
    int iNumEntriesOverExecuted = 0;
#endif

    ADD_TO_MY_SCHEDULER_LOG(( "\n"
                              "frame %i (time: %fs)\n"
                              "queue: ",
                              Asura_Timers::GetFrameNumber(), fInitialLocalScaledTime ));

    for( Asura_List_ConstIt<QueueEntry> xIt(&m_xQueue); !xIt.Done(); xIt.Next() )
    {
        ADD_TO_MY_SCHEDULER_LOG(( "0x%03X ", xIt.GetCurrent()->m_iHandle ));
    }

    ADD_TO_MY_SCHEDULER_LOG(( "\n" ));
    ADD_TO_MY_SCHEDULER_LOG_LIGHT(( "avg task exec period %age: %.2f%%\n", 100.0f * m_fAvExecutionPeriodFraction ) );
    ADD_TO_MY_SCHEDULER_LOG(( "avg time between scheduler updates: %fs\n", m_fAvTimeBetweenUpdates ));

    // We go through the queue executing tasks that are ready and have not been executed this frame already.
    // We always execute at least one task, so that the queue is moved along even if we haven't been allocated enough time.
    if( !m_xQueue.IsEmpty() )
    {   
        QueueEntry* pxCurrentEntry = m_xQueue.GetHeadObject();
		while( true )
		{
			const float fInitialTimeInFrameThisIteration = Asura_Timers::GetTimeInFrame();
			const float fInitialLocalScaledTimeThisIteration = fScaledTimeAtStartOfFrame + fInitialTimeInFrameThisIteration;

			// if our last execution time is greater than equal to the initial time this frame we have most likely already exectued this entry and so we need to stop for this frame.
			if ( pxCurrentEntry->m_fLastExecutionTime >= fInitialLocalScaledTime )
			{
				break;
			}

			ADD_TO_MY_SCHEDULER_LOG(( "   iteration @ %fs\n", fInitialLocalScaledTimeThisIteration ));

            // Stop if next task is not ready yet.
            if( pxCurrentEntry->m_fNextExecutionTime >= fInitialLocalScaledTimeThisIteration )
            {
                ADD_TO_MY_SCHEDULER_LOG(( "      task 0x%03X not due until %fs (%fs away) - postponing\n",
                                          pxCurrentEntry->m_iHandle, pxCurrentEntry->m_fNextExecutionTime, pxCurrentEntry->m_fNextExecutionTime - fInitialLocalScaledTimeThisIteration ));

                DEBUGVAR bool dbg_bSchedulerPreemptiveExecutionEnabled = true;
                DEBUGVAR float dbg_fSchedulerPreemptiveExecutionDurationOverestimate = 1.5f;
                DEBUGVAR float dbg_fSchedulerPreemptiveExecutionUpdatePeriodErrorFactor = 0.333333f;

                if( !dbg_bSchedulerPreemptiveExecutionEnabled ) { break; }

                // If we have plenty of execution time left to cover this task...
                float fRemainingExecutionTime = m_fMaxPerFrameTime - (fInitialTimeInFrameThisIteration - fInitialTimeInFrame);
                if( fRemainingExecutionTime < dbg_fSchedulerPreemptiveExecutionDurationOverestimate * pxCurrentEntry->m_fAvExecutionTime )
                {
                    break;
                }

                // ... and the deviation from the requested update period for
                // this task would be (much) smaller if we updated it now
                // rather than waiting until next time...
                float fTimeUntilDue = pxCurrentEntry->m_fNextExecutionTime - fInitialLocalScaledTimeThisIteration;
                float fExpectedTimeUntilNextUpdate = (fInitialLocalScaledTime + m_fAvTimeBetweenUpdates) - fInitialLocalScaledTimeThisIteration;
                if( fTimeUntilDue >= dbg_fSchedulerPreemptiveExecutionUpdatePeriodErrorFactor * fExpectedTimeUntilNextUpdate )
                {
                    break;
                }

                // ... then actually go ahead and execute it anyway.
#if ASURA_SCHEDULER_DEBUG
                ++iNumEntriesPreemptivelyExecuted;
#endif
                ADD_TO_MY_SCHEDULER_LOG(( "      ... making an exception because:\n"
                                          "      got %fs left and avg exec time for task is %fs\n"
                                          "      and only %fs left until due (vs %fs update period and %fs expected until next opportunity\n",
                                          fRemainingExecutionTime, pxCurrentEntry->m_fAvExecutionTime,
                                          fTimeUntilDue, pxCurrentEntry->m_fUpdatePeriod, fExpectedTimeUntilNextUpdate ));
            }
            else
            {
                ADD_TO_MY_SCHEDULER_LOG(( "      task 0x%03X due @ %fs (%fs ago) - proceeding\n",
                                          pxCurrentEntry->m_iHandle, pxCurrentEntry->m_fNextExecutionTime, fInitialLocalScaledTimeThisIteration - pxCurrentEntry->m_fNextExecutionTime ));
            }

            float fActualUpdatePeriod = fInitialLocalScaledTimeThisIteration - pxCurrentEntry->m_fLastExecutionTime;
            pxCurrentEntry->m_pxCallback->Execute( fActualUpdatePeriod );

            float fDurationOfThisUpdate = Asura_Timers::GetTimeInFrame() - fInitialTimeInFrameThisIteration;
            ADD_TO_MY_SCHEDULER_LOG(( "      update complete after %fs (vs avg of %fs)\n", fDurationOfThisUpdate, pxCurrentEntry->m_fAvExecutionTime ));

            // If asserts fire or breakpoints are hit the update time can be massive. To avoid these erroneous times interfering with the scheduling we tweak them.
            if( fDurationOfThisUpdate > m_fMaxPerFrameTime )
            {
                fDurationOfThisUpdate = pxCurrentEntry->m_fAvExecutionTime;
                ADD_TO_MY_SCHEDULER_LOG(( "      ... but clamping update duration to %fs\n", m_fMaxPerFrameTime ));
            }
            // The last term in the calculation below is to make the value converge faster if the update period is long.
            pxCurrentEntry->m_fAvExecutionTime  += ( fDurationOfThisUpdate - pxCurrentEntry->m_fAvExecutionTime ) * Asura_Maths::Min( 0.1f + fActualUpdatePeriod, 0.9f );
            pxCurrentEntry->m_fLastExecutionTime = fInitialLocalScaledTimeThisIteration;

#if ASURA_SCHEDULER_DEBUG
            float fExecutionPeriodFraction;
            if( pxCurrentEntry->m_bUpdatePeriodHasChanged )
            {
                // This could be a new entry, or the period could have been
                // modified mid-update, or the period could have been
                // significantly reduced. Don't allow this to affect the
                // statistics (eg. in the latter case it would seem like the
                // task is overdue).
                fExecutionPeriodFraction = 1.0f;
                pxCurrentEntry->m_bUpdatePeriodHasChanged = false;
            }
            else
            {
                fExecutionPeriodFraction = fActualUpdatePeriod / pxCurrentEntry->m_fUpdatePeriod;
            }

            m_fAvExecutionPeriodFraction += ( fExecutionPeriodFraction - m_fAvExecutionPeriodFraction ) * Asura_Maths::Min( 0.1f + fActualUpdatePeriod, 0.9f );

            ++iNumEntriesExecuted;
#endif

            // Reschedule the entry after updating its details.
            m_xQueue.RemoveFromList( pxCurrentEntry );
            pxCurrentEntry->m_fNextExecutionTime = fInitialLocalScaledTimeThisIteration + pxCurrentEntry->m_fUpdatePeriod;
            m_xQueue.InsertInList( pxCurrentEntry, QueueEntry::Compare );
            ADD_TO_MY_SCHEDULER_LOG(( "      scheduled next update for %fs\n", pxCurrentEntry->m_fNextExecutionTime ));

            // Get the new most urgent task.
            pxCurrentEntry = m_xQueue.GetHeadObject();

            // If the new task will take us over our allowed execution time we must stop.
            float fEndTimeInFrameThisIteration = Asura_Timers::GetTimeInFrame();
            float fRemainingExecutionTime = m_fMaxPerFrameTime - (fEndTimeInFrameThisIteration - fInitialTimeInFrame);
            if( fRemainingExecutionTime <= pxCurrentEntry->m_fAvExecutionTime )
            {
                ADD_TO_MY_SCHEDULER_LOG(( "   finishing updates; %fs of allotted time left (%.1f%%), vs %fs avg exec time of next task\n"
                                          "   (time now: %fs; next task due @ %fs)\n",
                                          fRemainingExecutionTime, 100.0f * (fRemainingExecutionTime / m_fMaxPerFrameTime), pxCurrentEntry->m_fAvExecutionTime,
                                          fScaledTimeAtStartOfFrame + fEndTimeInFrameThisIteration, pxCurrentEntry->m_fNextExecutionTime ));
                break;
            }
        }
    }

#if ASURA_SCHEDULER_DEBUG
    m_fAvPerFrameTime  += ( ( Asura_Timers::GetTimeInFrame() - fInitialTimeInFrame ) - m_fAvPerFrameTime ) * 0.01f;
    m_fAvTasksPerFrame += ( static_cast<float>(iNumEntriesExecuted) - m_fAvTasksPerFrame ) * 0.01f;
    m_fAvPreemptiveExecutionsPerFrame += (static_cast<float>(iNumEntriesPreemptivelyExecuted) - m_fAvPreemptiveExecutionsPerFrame) * 0.01f;
    m_fAvOverExecutionsPerFrame += (static_cast<float>(iNumEntriesOverExecuted) - m_fAvOverExecutionsPerFrame) * 0.01f;
#endif

    m_bMidUpdate = false;

    // If we have queued QueueEntrys to be registered then register them now.
    if ( m_xToBeRegistered.GetNumEntries() )
    {
        Asura_List_It< QueueEntry > xToBeRegisteredIter( &m_xToBeRegistered );
        while ( !xToBeRegisteredIter.Done() )
        {
            QueueEntry* pxToBeRegistered = xToBeRegisteredIter.GetCurrent();
            xToBeRegisteredIter.DeleteCurrent();
            m_xQueue.InsertInList( pxToBeRegistered, &QueueEntry::Compare );
        }
    }

    // If we have queued handles to be unregistered then unregister them now.
    if ( m_xToBeUnregistered.GetNumEntries() )
    {
        Asura_List_It< Asura_Handle > xToBeUnregisteredIter( &m_xToBeUnregistered );
        while ( !xToBeUnregisteredIter.Done() )
        {
            Asura_Handle* piToBeUnregistered = xToBeUnregisteredIter.GetCurrent();
            xToBeUnregisteredIter.DeleteCurrent();
            UnregisterCallback( *piToBeUnregistered );
            delete piToBeUnregistered;
        }
    }
}

Asura_Handle Asura_Scheduler::RegisterCallback( const float fUpdatePeriod, const Asura_Callback_Base_1Arg< void, float >* pxCallback )
{
	if ( pxCallback->GetSize() > s_xQueueEntryMempool.GetSizeOfElement() )
	{
		Asura_Assert( false, "Asura_Scheduler::RegisterCallback - callback is larger than the element size allowed by the mempool" );
		return ASURA_INVALID_HANDLE;
	}

    QueueEntry* pxNewEntry = static_cast< QueueEntry* >( s_xQueueEntryMempool.Allocate() );
    if ( !pxNewEntry ) { return ASURA_INVALID_HANDLE; }

    pxNewEntry->m_iHandle            = m_iNextHandle++;
    pxNewEntry->m_pxCallback         = pxCallback->Clone( s_xCallbackMempool.Allocate() );
    pxNewEntry->m_fUpdatePeriod      = fUpdatePeriod;
    pxNewEntry->m_fNextExecutionTime = Asura_Timers::GetLocalScaledTime() + Asura_Random::GetFloat( 0.0f, fUpdatePeriod );
    pxNewEntry->m_fLastExecutionTime = pxNewEntry->m_fNextExecutionTime - fUpdatePeriod;
    pxNewEntry->m_fAvExecutionTime   = dbg_fSchedulerNewCallbackDurationEstimateFactor * m_fMaxPerFrameTime;

#if ASURA_SCHEDULER_DEBUG
    pxNewEntry->m_bUpdatePeriodHasChanged = true;
#endif

    if ( m_bMidUpdate )
    {
        m_xToBeRegistered.AddToList( pxNewEntry );
    }
    else
    {
        m_xQueue.InsertInList( pxNewEntry, &QueueEntry::Compare );
    }

    return pxNewEntry->m_iHandle;
}


void Asura_Scheduler::SetUpdatePeriodForCallback( const Asura_Handle iHandle, const float fUpdatePeriod )
{
    if ( iHandle == ASURA_INVALID_HANDLE ) { return; }

    Asura_List_It< QueueEntry > xQueueIter( &m_xQueue );
    while ( !xQueueIter.Done() )
    {
        QueueEntry* pxEntry = xQueueIter.GetCurrent();
        if ( pxEntry->m_iHandle == iHandle )
        {
            pxEntry->m_fUpdatePeriod = fUpdatePeriod;
#if ASURA_SCHEDULER_DEBUG
            pxEntry->m_bUpdatePeriodHasChanged = true;
#endif

            // If the max update period has been reduced then we should at
            // least ensure it's not going to wait longer from now than the
            // new period. (We could bring it down to (last exec time +
            // update period) instead but doing anything like that may make
            // it immediately overdue and isn't necessarily any more correct.)
            pxEntry->m_fNextExecutionTime = Asura_Maths::Min( pxEntry->m_fNextExecutionTime,
                                                              Asura_Timers::GetLocalScaledTime() + Asura_Timers::GetTimeInFrame() + fUpdatePeriod );

            return;
        }
        xQueueIter.Next();
    }

    Asura_List_It< QueueEntry > xToBeRegisteredIter( &m_xToBeRegistered );
    while ( !xToBeRegisteredIter.Done() )
    {
        QueueEntry* pxEntry = xToBeRegisteredIter.GetCurrent();
        if ( pxEntry->m_iHandle == iHandle )
        {
            pxEntry->m_fUpdatePeriod = fUpdatePeriod;
#if ASURA_SCHEDULER_DEBUG
            pxEntry->m_bUpdatePeriodHasChanged = true;
#endif
            return;
        }
        xToBeRegisteredIter.Next();
    }

    Asura_Assert( false, "Asura_Scheduler::SetUpdatePeriodForCallback() - could not find entry" );
}

void Asura_Scheduler::UnregisterCallback( const Asura_Handle iHandle )
{
    if ( m_bMidUpdate )
    {
        m_xToBeUnregistered.AddToList( new Asura_Handle( iHandle ) );
        return;
    }
    else
    {
        Asura_List_It< QueueEntry > xQueueIter( &m_xQueue );
        while ( !xQueueIter.Done() )
        {
            QueueEntry* pxEntry = xQueueIter.GetCurrent();
            if ( pxEntry->m_iHandle == iHandle )
            {
                xQueueIter.DeleteCurrent();
                s_xCallbackMempool.Deallocate( pxEntry->m_pxCallback );
                s_xQueueEntryMempool.Deallocate( pxEntry );

                return;
            }
            else
            {
                xQueueIter.Next();
            }
        }
    }
}

#if ASURA_SCHEDULER_DEBUG

void Asura_Scheduler::RenderDebugInfo( Asura_Vector_2& xTextPosition )
{
    unsigned int uColour = 0xffffffff;
    Asura_Font_System::Print( xTextPosition.x, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
        "Number of tasks: %d", m_xQueue.GetNumEntries() );
    xTextPosition.y += 10.0f;

	Asura_Font_System::Print( xTextPosition.x, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
        "Average tasks executed per frame: %f", Asura_Maths::FtoD(m_fAvTasksPerFrame) );
    xTextPosition.y += 10.0f;

    uColour = ( m_fAvPerFrameTime > m_fMaxPerFrameTime * 1.0f ) ? 0xffff0000 : 0xff00ff00;
	Asura_Font_System::Print( xTextPosition.x, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
        "Average execution time: %f (%f%% of max)",
        Asura_Maths::FtoD(m_fAvPerFrameTime),
        Asura_Maths::FtoD(( m_fAvPerFrameTime / m_fMaxPerFrameTime ) * 100.0f) );
    xTextPosition.y += 10.0f;

    uColour = ( m_fAvExecutionPeriodFraction > 1.1f ) ? 0xffff0000 : 0xff00ff00;
	Asura_Font_System::Print( xTextPosition.x, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
        "Average execution period: %f of max", Asura_Maths::FtoD(m_fAvExecutionPeriodFraction * 100.0f) );
    xTextPosition.y += 10.0f;

    uColour = 0xffffffff;
    Asura_Font_System::Print( xTextPosition.x, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
        "Average time between scheduler updates: %f", Asura_Maths::FtoD(m_fAvTimeBetweenUpdates) );
    xTextPosition.y += 10.0f;

    //Asura_Font_System::Print( xTextPosition.x, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
    //    "Task average execution times:" );
    //xTextPosition.y += 10.0f;
    //for ( Asura_List_ConstIt< QueueEntry > xQueueIt( &m_xQueue ); !xQueueIt.Done(); xQueueIt.Next() )
    //{
    //    const QueueEntry* pxQueueEntry = xQueueIt.GetCurrent();
    //    Asura_Font_System::Print( pxQueueEntry->m_iHandle * 30.0f, xTextPosition.y, 0.5f, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
    //        "%f", pxQueueEntry->m_fAvExecutionTime );
    //}
    //xTextPosition.y += 10.0f;
}

#include <stdarg.h>
#include "Core/Asura_String.h"

#if ASURA_SCHEDULER_DEBUG == 2
const u_int uSCHEDULER_LOG_CAPACITY = 128 * 1024;
#else
const u_int uSCHEDULER_LOG_CAPACITY = 32 * 1024;
#endif

const u_int uSCHEDULER_LOG_MAX_CHARS_PER_ADD = 256;

void Asura_Scheduler::AddToSchedulerLog( const char* szFormat, ... )
{
    if( !m_szPerFrameLogFile ) { return; }

    if( m_uAccumulatedLogLength + uSCHEDULER_LOG_MAX_CHARS_PER_ADD >= uSCHEDULER_LOG_CAPACITY )
    {
        FlushSchedulerLog( false );
    }

	va_list xArgs;
	va_start( xArgs, szFormat );
    m_uAccumulatedLogLength += Asura_PrintfToBfr( m_szAccumulatedLog + m_uAccumulatedLogLength, szFormat, xArgs, uSCHEDULER_LOG_MAX_CHARS_PER_ADD, true );
	va_end( xArgs );
}


void Asura_Scheduler::FlushSchedulerLog( bool bLeaveEmpty )
{
    if( m_szPerFrameLogFile && m_uAccumulatedLogLength )
    {
        if( m_bNeedToClearLogFile )
        {
            Asura_SimpleLogfileClear(( m_szPerFrameLogFile ));
            m_bNeedToClearLogFile = false;
        }

        Asura_SimpleLogfileOutput_PreFormatted( m_szPerFrameLogFile, m_szAccumulatedLog, m_uAccumulatedLogLength );
        m_uAccumulatedLogLength = 0;

        if( !bLeaveEmpty )
        {
            // Make it obvious in the log where this has happened. It can
            // take quite a long time on some platforms and so it may affect
            // scheduling for a short period afterwards (ie. tasks end up
            // waiting long beyond their max update period because of a
            // blocking logfile output here).
            const char* szFLUSH_MSG = "\n"
                                      "************\n"
                                      "LOG FLUSHED!\n"
                                      "************\n"
                                      "\n";
            strcpy( m_szAccumulatedLog, szFLUSH_MSG );
            m_uAccumulatedLogLength = strlen( m_szAccumulatedLog );
        }
    }
}


void Asura_Scheduler::SetPerFrameLogFile( const char* szPerFrameLogFile )
{
    if( szPerFrameLogFile == m_szPerFrameLogFile )
    {
        return;
    }

    FlushSchedulerLog( true );

    m_szPerFrameLogFile = szPerFrameLogFile;

    if( m_szPerFrameLogFile )
    {
        if( !m_szAccumulatedLog )
        {
            m_szAccumulatedLog = new char[uSCHEDULER_LOG_CAPACITY];
            m_uAccumulatedLogLength = 0;
        }
    }
    else
    {
        if( m_szAccumulatedLog )
        {
            delete[] m_szAccumulatedLog;
            m_szAccumulatedLog = NULL;
            m_uAccumulatedLogLength = 0;
        }
    }
}

#endif // #if ASURA_SCHEDULER_DEBUG
