//////////////////////////////////////////////////////////////////////////////
//
// Asura_Scheduler.h
//
// Created by NJD - 14 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_SCHEDULER_H__
#define __ASURA_SCHEDULER_H__ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// asura
#include "Core/Asura_Callback.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_List.h"
#include "Core/Asura_Mempool.h"

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Scheduler
//
//////////////////////////////////////////////////////////////////////////////

// Logging level implied by this #define:
//    0 = off
//    1 = light
//    2 = heavy
//
// 1 is useful for being able to view a cheap log of average execution period
// fractions. 2 allows more in-depth analysis but will probably impact upon
// scheduling on some platforms (since the overhead of creating the logging
// information consumes the scheduler's allotted time).
#define ASURA_SCHEDULER_DEBUG (1)

class Asura_Scheduler
{
public:
    Asura_Scheduler();
    ~Asura_Scheduler();
    void Update();

    void         SetMaxPerFrameTime( const float fTime ) { m_fMaxPerFrameTime = fTime; }
    Asura_Handle RegisterCallback( const float fUpdatePeriod, const Asura_Callback_Base_1Arg< void, float >* pxCallback );
    void         SetUpdatePeriodForCallback( const Asura_Handle iHandle, const float fUpdatePeriod );
    void         UnregisterCallback( const Asura_Handle iHandle );

#if ASURA_SCHEDULER_DEBUG
    void RenderDebugInfo( class Asura_Vector_2& xTextPosition );
    void SetPerFrameLogFile( const char* szPerFrameLogFile );
#else
    inline void RenderDebugInfo( class Asura_Vector_2& xTextPosition ) { }
    inline void SetPerFrameLogFile( const char* szPerFrameLogFile ) { }
#endif

private:
    static void* AllocateFromQueueListNodeMempool()                { return s_xQueueListNodeMempool.Allocate(); }
    static void  DeallocateToQueueListNodeMempool( void* pMemory ) { s_xQueueListNodeMempool.Deallocate( pMemory ); }

private:
    float        m_fMaxPerFrameTime;
    Asura_Handle m_iNextHandle;
    bool         m_bMidUpdate;

    struct QueueEntry
    {
        QueueEntry()
        : m_iHandle( ASURA_INVALID_HANDLE )
        , m_pxCallback( NULL )
        , m_fUpdatePeriod( 0.0f )
#if ASURA_SCHEDULER_DEBUG
        , m_bUpdatePeriodHasChanged( false )
#endif
        , m_fLastExecutionTime( 0.0f )
        , m_fNextExecutionTime( 0.0f )
        , m_fAvExecutionTime( 0.0f )
        {}
        static int Compare( const QueueEntry& xOne, const QueueEntry& xTwo )
        {
            if ( xOne.m_fNextExecutionTime < xTwo.m_fNextExecutionTime ) { return -1; }
            else { return 1; }
        }

        Asura_Handle         m_iHandle;
        Asura_Callback_Base_1Arg< void, float >* m_pxCallback;
        // How often we want to be called.
        float                m_fUpdatePeriod;
#if ASURA_SCHEDULER_DEBUG
        bool                 m_bUpdatePeriodHasChanged;
#endif
        // Local scaled time of when we were last, and when we should next be, called.
        float                m_fLastExecutionTime;
        float                m_fNextExecutionTime;
        // Average value of how long we take to execute.
        float                m_fAvExecutionTime;
    };
    Asura_List< QueueEntry >   m_xQueue;
    Asura_List< QueueEntry >   m_xToBeRegistered;
    Asura_List< Asura_Handle > m_xToBeUnregistered;

#if ASURA_SCHEDULER_DEBUG
    // The actual execution period of the callbacks as a percentage of their max execution period.
    // If this goes much over 100% then we're falling behind on our tasks (we've not been allocated enough time per frame).
    float m_fAvExecutionPeriodFraction;
    float m_fAvPerFrameTime;
    float m_fAvTasksPerFrame;

    float m_fAvPreemptiveExecutionsPerFrame;
    float m_fAvOverExecutionsPerFrame;

    const char* m_szPerFrameLogFile;
    char* m_szAccumulatedLog;
    u_int m_uAccumulatedLogLength;
    bool m_bNeedToClearLogFile;

    void AddToSchedulerLog( const char* szFormat, ... );
    void FlushSchedulerLog( bool bLeaveEmpty );
#endif

    float m_fPrevScaledTime;
    float m_fAvTimeBetweenUpdates;

    static Asura_ResizingMempool s_xQueueListNodeMempool;
    static Asura_ResizingMempool s_xQueueEntryMempool;
    static Asura_ResizingMempool s_xCallbackMempool;
};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif //__ASURA_SCHEDULER_H__
