//////////////////////////////////////////////////////////////////////////////
//
// Asura_SortAndSearch.h
//
// Created by NJD - 19 September 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_SORTANDSEARCH_H__
#define __ASURA_SORTANDSEARCH_H__ 1

// Asura includes
#include "Core/Asura_PlatformBasics.h" // for stricmp on PS3

// library includes
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////////////
//
// Basic sorting and searching algorithms.
//
//////////////////////////////////////////////////////////////////////////////

class Asura_SortAndSearch
{
public:
    // Wrapper for qsort
    // CompareElements must return:
    // < 0      for xDataOne < xDataTwo
    // > 0      for xDataOne > xDataTwo
    // 0        for xDataOne == xDataTwo
    // In order to sort ascending.
	// Note, I used to have SortAscending<T> as a default argument but PS2 GCC 2.95.3 won't compile it, 
	// and the 360 compiler seemed to produce incorrect code with it.
    template < typename T >
    static void Quicksort( T axData[], u_int uNumElements,
							int ( *CompareElements )( const void* pData1, const void* pData2 ) );

	// non-templatised wrapper for familiarity.
	static void Qsort( void* pData, u_int uNumElements, u_int uElementSize, 
						int ( *CompareElements )( const void* pData1, const void* pData2 ) );

	// Search: TODO look into using stdlib's bsearch() instead.
    // This is a binary search for now, but could become different algorithms based on the data size.
    // Note that to use a binary search the data must already be sorted, normally into ascending order.
    // CompareToKey must return:
    // < 0      for Key < Data
    // > 0      for Key > Data
    // 0        for Key == Data
    template < typename T, typename KeyType >
    static T* Search( T axData[], const KeyType& xKey, int iNumElements, int ( *CompareToKey )( const KeyType& xKey, const T& xData ) );
    template < typename T, typename KeyType >
    static T* Search( T axData[], const KeyType& xKey, int iLeftIndex, int iRightIndex, int ( *CompareToKey )( const KeyType& xKey, const T& xData ) );

	// helper functions for types which implement <, > and =
	template < typename T >
	static int SortAscending( const void* pData1, const void* pData2 )
	{
		const T& xData1 = *reinterpret_cast<const T*>(pData1);
		const T& xData2 = *reinterpret_cast<const T*>(pData2);
		if (xData1 < xData2)
		{
			return -1;
		}
		if (xData1 > xData2)
		{
			return 1;
		}
		return 0;
	}

	template < typename T >
	static int SortDescending( const void* pData1, const void* pData2 )
	{
		const T& xData1 = *reinterpret_cast<const T*>(pData1);
		const T& xData2 = *reinterpret_cast<const T*>(pData2);
		if (xData1 < xData2)
		{
			return 1;
		}
		if (xData1 > xData2)
		{
			return -1;
		}
		return 0;
	}

	// helper for pointer types which have GetHashID() member function
	template < typename T >
	static int SortAscendingByHashID( const void* pData1, const void* pData2 )
	{
		const T& xData1 = *reinterpret_cast<const T*>(pData1);
		const T& xData2 = *reinterpret_cast<const T*>(pData2);
		if (xData1->GetHashID() < xData2->GetHashID())
		{
			return -1;
		}
		if (xData1->GetHashID() > xData2->GetHashID())
		{
			return 1;
		}
		return 0;
	}

	// string comparison helpers
	static int SortStringsCaseInsensitive( const void* p1, const void* p2 )
	{
		const char* szOne = *reinterpret_cast<const char* const *>(p1);
		const char* szTwo = *reinterpret_cast<const char* const *>(p2);
		return stricmp(szOne, szTwo);
	}

	static int SortStringsCaseSensitive( const void* p1, const void* p2 )
	{
		const char* szOne = *reinterpret_cast<const char* const *>(p1);
		const char* szTwo = *reinterpret_cast<const char* const *>(p2);
		return strcmp(szOne, szTwo);
	}


	// non-templatised versions of the template functions to help things compile on PS2.
	// When we no longer need to support broken compilers like GCC 2.95.3 these can go.
	static int SortAscendingInt( const void* pData1, const void* pData2 );
	static int SortDescendingInt( const void* pData1, const void* pData2 );
	static int SortAscendingFloat( const void* pData1, const void* pData2 );
	static int SortDescendingFloat( const void* pData1, const void* pData2 );
	static int SortAscendingUint( const void* pData1, const void* pData2 );
	static int SortDescendingUint( const void* pData1, const void* pData2 );

};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

template < typename T >
void Asura_SortAndSearch::Quicksort( T axData[], u_int uNumElements,
									int ( *CompareElements )( const void* pData1, const void* pData2 ) /* = SortAscending<T> */ )
{
	qsort( axData, uNumElements, sizeof(T), CompareElements );
}


template < typename T, typename KeyType >
T* Asura_SortAndSearch::Search( T axData[], const KeyType& xKey, int iNumElements, int ( *CompareToKey )( const KeyType& xKey, const T& xData ) )
{
    return Search< T, KeyType >( axData, xKey, 0, iNumElements - 1, CompareToKey );
}

template < typename T, typename KeyType >
T* Asura_SortAndSearch::Search( T axData[], const KeyType& xKey, int iLeftIndex, int iRightIndex, int ( *CompareToKey )( const KeyType& xKey, const T& xData ) )
{
    int iMidIndex;
    while ( iLeftIndex <= iRightIndex )
    {
        iMidIndex = ( iLeftIndex + iRightIndex ) / 2;
        if ( CompareToKey( xKey, axData[ iMidIndex ] ) > 0 )
        {
            iLeftIndex = iMidIndex + 1;
        }
        else if ( CompareToKey( xKey, axData[ iMidIndex ] ) < 0 )
        {
            iRightIndex = iMidIndex - 1;
        }
        else
        {
            return &axData[ iMidIndex ];
        }
    }

    return NULL;
}

// some specialisation for for signed types which have fast - operator (i.e. int, float)
template <>
int Asura_SortAndSearch::SortAscending<int>( const void* pData1, const void* pData2 );
template <>
int Asura_SortAndSearch::SortDescending<int>( const void* pData1, const void* pData2 );
template <>
int Asura_SortAndSearch::SortAscending<float>( const void* pData1, const void* pData2 );
template <>
int Asura_SortAndSearch::SortDescending<float>( const void* pData1, const void* pData2 );


//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif //__ASURA_SORTANDSEARCH_H__
