
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Stream.h
// Author : Andy Tate
// Date   : 9-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_STREAM_H__
#define __ASURA_STREAM_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include <string.h>

#include "Core/Asura_Endian.h"

#include "Core/Asura_PlatformBasics.h"

////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

class Asura_Stream
{
public:
	// Public enumerated types...

	enum OpenType
	{
		READ_DATA,
		WRITE_DATA,
	};


	// Public functions...

	// Abstract functions...
	virtual	bool			Open				(const OpenType eOpenType)						= 0;

	virtual	u_int			ReadData			(const u_int uNumberOfBytes, void* pData)		= 0;
	virtual	u_int			SkipData			(const u_int uNumberOfBytes)					= 0;
	virtual	u_int			WriteData			(const u_int uNumberOfBytes, const void* pData)	= 0;

	// NOTE: These 2 functions should be abstract but that would mean that they 
	// need implementing on all platforms and I don't have time at the mo - ART...
	virtual	u_int			GetCurrentPosition	(void) const									{ return (0); }
	virtual	u_int			GetFileSize			(void) const									{ return (0); }

	virtual	bool			Close				(void)											= 0;

	virtual bool			SetCurrentPosition(u_int uNumberOfBytes)
	{
		Asura_Assert(false, "SetCurrentPosition not implemented for this stream type.");
		return false;
	}

	// Stream operators for basic types
	Asura_Stream& operator<< (const int iValue)
	{
		int	iTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(iValue) : iValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(iTemp), reinterpret_cast<void*>(&iTemp));
		return		(*this);
	}

	Asura_Stream& operator<< (const u_int uValue)
	{
		u_int	uTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(uValue) : uValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(uTemp), reinterpret_cast<void*>(&uTemp));
		return		(*this);
	}

	Asura_Stream& operator<< (const float fValue)
	{
		float	fTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(fValue) : fValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(fTemp), reinterpret_cast<void*>(&fTemp));
		return		(*this);
	}

	Asura_Stream& operator<< (const short sValue)
	{
		short	sTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(sValue) : sValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(sTemp), reinterpret_cast<void*>(&sTemp));
		return		(*this);
	}

	Asura_Stream& operator<< (const u_short usValue)
	{
		u_short	usTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(usValue) : usValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(usTemp), reinterpret_cast<void*>(&usTemp));
		return		(*this);
	}

	Asura_Stream& operator<< (const char cValue)
	{
		char	cTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(cValue) : cValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(cTemp), reinterpret_cast<void*>(&cTemp));
		return		(*this);
	}

	Asura_Stream& operator<< (const u_char ucValue)
	{
		u_char	ucTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(ucValue) : ucValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(ucTemp), reinterpret_cast<void*>(&ucTemp));
		return		(*this);
	}

#if defined(ASURA_UNICODE_CHAR_DISTINCT_TYPE)
	Asura_Stream& operator<< (const Asura_UnicodeChar wcValue)
	{
		Asura_UnicodeChar	wcTemp	= (m_bSwapEndian) ? Asura_Endian::SwapBytes(wcValue) : wcValue;

		// Write the appropriate number of bytes...
		WriteData	(sizeof(wcTemp), reinterpret_cast<void*>(&wcTemp));
		return		(*this);
	}
#endif	// defined(ASURA_UNICODE_CHAR_DISTINCT_TYPE)

	// Stream operator functions...
	template<typename T> Asura_Stream& operator>> (T& xValue)
	{
		// Read the appropriate number of bytes...
		ReadData	(sizeof(xValue), reinterpret_cast<void*>(&xValue));
		return		(*this);
	}

	template<typename T> Asura_Stream& operator<< (const T& xValue)
	{
		// Write the appropriate number of bytes...
		WriteData	(sizeof(xValue), reinterpret_cast<const void*>(&xValue));
		return		(*this);
	}

	Asura_Stream& operator<< (char* szText)
	{
		// Write the string specified...
		WriteData	(strlen(szText), reinterpret_cast<const void*>(szText));
		return		(*this);
	}

	Asura_Stream& operator<< (const char* szText)
	{
		// Write the string specified...
		WriteData	(strlen(szText), reinterpret_cast<const void*>(szText));
		return		(*this);
	}

	// Accessor functions...
			bool			IsReading		(void) const		{ return (m_bIsOpen && (READ_DATA == m_eOpenType)); }
			bool			IsWriting		(void) const		{ return (m_bIsOpen && (WRITE_DATA == m_eOpenType)); }

			bool			IsOpen			(void) const		{ return (m_bIsOpen); }

			void			SetSwapEndian	(bool bSwapEndian)	{ m_bSwapEndian = bSwapEndian; }
			bool			IsSetSwapEndian	(void) const		{ return (m_bSwapEndian); }


protected:
	// Protected variables...

	// Flag to say whether the stream is currently open...
			bool		m_bIsOpen;

	// Flag to say what sort of endian-ness we are to write...
			bool		m_bSwapEndian;

	// The type of stream-open performed...
			OpenType	m_eOpenType;


	// Protected functions...

	Asura_Stream          (void)	: m_bIsOpen(false), m_bSwapEndian(false)	{}
	virtual ~Asura_Stream ()													{}


private:
	// Private variables...


	// Private functions...

	Asura_Stream  (const Asura_Stream& xOther);

			Asura_Stream&	operator=	(const Asura_Stream& xOther);

};   // End of class 'Asura_Stream'.


#endif   // ifndef __ASURA_STREAM_H__


/// End of file. ///
