//*******************************************************************
//
// File: Asura_StringBuffer.h                  
//
// Created: 12th May 2007
//
// Description: Fixed length string type.
//
//*******************************************************************
//
// Class:       Asura_StringBuffer
//
// Description: Fixed length string buffer. Supports most standard string
//              functions, but cannot exceed a specified length, there is no extra memory overhead for storing these strings
//
//              Has an interface like a dynamic string object but without dynamic memory allocation
//				So it is useful for game run-time applications instead of tools
//
//				A template class based on Asura_Char could do a wide string version based on Asura_UnicodeChar
//
//				Would be interesting to see if can have a base template class which takes the type (Asura_Char or Asura_UnicodeChar)
//
//				The interface is intended to be similar to the Asura_String interface
//
//*******************************************************************

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_STRINGBUFFER_H__
#define __ASURA_STRINGBUFFER_H__

#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"

template<u_int uLength>
class Asura_StringBuffer
{
public:
	Asura_StringBuffer();
	Asura_StringBuffer( const Asura_Char* const szFormat, ... );

	const Asura_Char* GetString( void ) const; // For functions like printf

	int GetLength( void ) const;
	u_int GetMaxLength( void ) const;

	Asura_Hash_ID GetHash( void ) const;
	Asura_Hash_ID GetHash_IgnoreExtension( void ) const;

	bool IsBlank( void ) const;
	void MakeBlank( void ); 

	Asura_Char& operator [] ( const int iIndex );
	const Asura_Char& operator [] ( const int iIndex ) const;

	void operator = ( const Asura_Char* const szOther );
	void operator +=( const Asura_Char* const szSource );

	int CompareIgnoringExtension( const Asura_Char* const szStringBuffer, const bool bIgnoreCase = true ) const;
	bool EqualsIgnoringExtension( const Asura_Char* const szOther, const bool bIgnoreCase = true ) const;
	bool LessThanIgnoringExtension( const Asura_Char* const szOther, const bool bIgnoreCase = true ) const;
	bool GreaterThanIgnoringExtension( const Asura_Char* const szOther, const bool bIgnoreCase = true ) const;

	int Compare( const Asura_Char* szStringBuffer, const bool bIgnoreCase = true ) const;

	// These operators are case-insensitive by default - use Compare for case-sensitive functions
	// This is to stick with the convention in Asura_String
    bool operator < ( const Asura_Char* const szBfr ) const;
    bool operator > ( const Asura_Char* const szBfr ) const;
    bool operator ==( const Asura_Char* const szBfr ) const;
    bool operator !=( const Asura_Char* const szBfr ) const;

	bool StartsWith( const Asura_Char* const szPrefix, const bool bIgnoreCase = true ) const;
    bool EndsWith( const Asura_Char* const szSuffix, const bool bIgnoreCase = true ) const;

	bool Contains( const Asura_Char* const szString, int* piPositionInString = NULL ) const;

    void MakeLowercase( void );
    void MakeUppercase( void );

	bool SplitFilename( Asura_StringBuffer<uLength>* const pxPath, Asura_StringBuffer<uLength>* const pxFilename, Asura_StringBuffer<uLength>* const pxExtension ) const;

	void Truncate( const u_int uMaxLength );
    void StripWhitespace( const u_int uStripFlags );
	void ReplaceAllChars( const Asura_Char cSearch, const Asura_Char cReplace );

	void Copy( const Asura_Char* const szFrom );
	void NCopy( const Asura_Char* const szFrom, const u_int uMaxLength );

	void Cat( const Asura_Char* const szString );
	void Cat( const Asura_Char cChar );

	bool GetBoolValue( bool* const pbValue ) const;
	bool GetFloatValue( float* const pfValue ) const;
	bool GetIntValue( int* const piValue, const bool bForceHex = false ) const;

	int FindChar( const Asura_Char cChar ) const;
	int ReverseFindChar( const Asura_Char cChar ) const;

	int VPrint( const Asura_Char* const szFormat, va_list& xArgs );
	int Print( const Asura_Char* const szFormat, ... );

	// formatted print to the end of the string
	int CatPrint( const Asura_Char* const szFormat, ... );

private:
	Asura_Char m_szData[uLength]; // Character string buffer.
};

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Asura_StringBuffer
//
// Description: Constructor. Empty null terminated string.
//-------------------------------------------------------------------
template<u_int uLength> 
inline Asura_StringBuffer<uLength>::Asura_StringBuffer()
{
	m_szData[0] = 0;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Asura_StringBuffer
//
// Description: Constructor. Creates string from formatted parameters.
//
// Parameters:  szFormat        - The string format.
//              ...             - The variable argument list.
//-------------------------------------------------------------------
template<u_int uLength>
inline Asura_StringBuffer<uLength>::Asura_StringBuffer( const Asura_Char* const szFormat, ...)
{
	va_list xArgs;
	va_start( xArgs, szFormat );
	Asura_PrintfToBfr( m_szData, szFormat, xArgs, uLength, true );
	va_end( xArgs );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetString
//
// Description: Gets the string as a constant character pointer. Needed
//              when a CoreStringBuffer is passed as an argument to printf
//              style functions.
//
// Returns:     The character string.
//-------------------------------------------------------------------
template<u_int uLength>
inline const Asura_Char* Asura_StringBuffer<uLength>::GetString( void ) const
{
	return m_szData;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetLength
//
// Description: Get the current length of the string.
//
// Returns:     The length of the string.
//-------------------------------------------------------------------
template<u_int uLength>
inline int Asura_StringBuffer<uLength>::GetLength( void ) const
{
	return (int)strlen( m_szData );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetMaxLength
//
// Description: Get the maximum length (including null terminator) of this
//              string buffer.
//
// Returns:     The maximum length.
//-------------------------------------------------------------------
template<u_int uLength>
inline u_int Asura_StringBuffer<uLength>::GetMaxLength( void ) const
{
	return uLength;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetHash
//
// Description: Convert the string to its hash value
//-------------------------------------------------------------------
template<u_int uLength>
inline Asura_Hash_ID Asura_StringBuffer<uLength>::GetHash( void ) const
{
	return( Asura_GetHashID( m_szData ) );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetHash_IgnoreExtension
//
// Description: Convert the string to its hash value ignoring any extensions
//-------------------------------------------------------------------
template<u_int uLength>
inline Asura_Hash_ID Asura_StringBuffer<uLength>::GetHash_IgnoreExtension( void ) const
{
	return( Asura_GetHashID_IgnoreExtension( m_szData ) );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::IsBlank
//
// Description: is the string empty
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::IsBlank( void ) const
{
	return ( m_szData[0] == 0 );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::MakeBlank
//
// Description: make the string empty
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::MakeBlank( void )
{
	m_szData[0] = 0;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator []
//
// Description: return the n'th character in the string (non-const reference)
//-------------------------------------------------------------------
template<u_int uLength> 
inline Asura_Char& Asura_StringBuffer<uLength>::operator [] ( const int iIndex ) 
{ 
	Asura_Assert( ( ( iIndex >= 0 ) && ( iIndex < (int)uLength ) ), "" );
	return m_szData[iIndex];
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator []
//
// Description: return the n'th character in the string (constant reference)
//-------------------------------------------------------------------
template<u_int uLength> 
inline const Asura_Char& Asura_StringBuffer<uLength>::operator [] ( const int iIndex )  const
{ 
	Asura_Assert( ( (iIndex >= 0 ) && ( iIndex < (int)uLength ) ), "" );
	return m_szData[iIndex]; 
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator =
//
// Description: Copy another string into the string buffer. Will be truncated
//              and null terminated if too long to fit in the buffer.
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::operator = ( const Asura_Char* const szFrom )
{
	Asura_PrintfToBfr( m_szData, "%s", uLength, true, szFrom );
	// Zero-terminate the string
	m_szData[uLength-1] = '\0';
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator +=
//
// Description: Cat another string into the string buffer. Will be truncated
//              and null terminated if too long to fit in the buffer.
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::operator +=( const Asura_Char* const szSource )
{
	const int iLen = GetLength();
	Asura_PrintfToBfr( m_szData+iLen, "%s", uLength-iLen, true, szSource );
	// Zero-terminate the string
	m_szData[uLength-1] = '\0';
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::CompareIgnoringExtension 
//
// Description: returns int strcmp value of this string compared to the other string ignoring this string's extension
//				Copied behaviour from Asura_String so the second string needs to have an extension i.e.
//				"jake.asr" compared to "jake" will fail 
//				"jake.asr" compared to "jake." will pass
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::CompareIgnoringExtension( const Asura_Char* const szStringBuffer, const bool bIgnoreCase /* = true */ ) const
{
	const int iLen = GetLength();
	for ( int i = iLen - 1 ; i >= 0; --i )
	{
		if ( m_szData[i] == '.')
		{
			if ( bIgnoreCase )
			{
				return strnicmp( m_szData, szStringBuffer, i+1 );
			}
			else
			{
				return strncmp( m_szData, szStringBuffer, i+1 );
			}
		}
	}
		
	if ( bIgnoreCase )
	{
		return stricmp( m_szData, szStringBuffer );
	}
	else
	{
		return strcmp( m_szData, szStringBuffer );
	}
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::EqualsIgnoringExtension 
//
// Description: reutrns true if this string equals the other string ignoring this string's extension
//				Copied behaviour from Asura_String so the second string needs to have an extension i.e.
//				"jake.asr" compared to "jake" will fail 
//				"jake.asr" compared to "jake." will pass
//				"jake" compared to "jake.asr" will pass
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::EqualsIgnoringExtension( const Asura_Char* const szOther, const bool bIgnoreCase ) const
{
	return ( CompareIgnoringExtension( szOther, bIgnoreCase ) == 0 );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::LessThanIgnoringExtension 
//
// Description: reutrns true if this string is less than the other string ignoring this string's extension
//				Copied behaviour from Asura_String so the second string needs to have an extension i.e.
//				"jake.asr" compared to "jake" will fail 
//				"jake.asr" compared to "jake." will pass
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::LessThanIgnoringExtension( const Asura_Char* const szOther, const bool bIgnoreCase ) const
{
	return ( CompareIgnoringExtension( szOther, bIgnoreCase ) < 0 );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GreaterThanIgnoringExtension 
//
// Description: reutrns true if this string is greater than the other string ignoring this string's extension
//				Copied behaviour from Asura_String so the second string needs to have an extension i.e.
//				"jake.asr" compared to "jake" will fail 
//				"jake.asr" compared to "jake." will pass
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::GreaterThanIgnoringExtension( const Asura_Char* const szOther, const bool bIgnoreCase ) const
{
	return ( CompareIgnoringExtension( szOther, bIgnoreCase ) > 0 );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Compare 
//
// Description: reutrns int strcmp value of this string compared to the other string
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::Compare( const Asura_Char* szStringBuffer, const bool bIgnoreCase ) const
{
	// check for empty string cases first
	if ( m_szData[0] == 0 )
	{
		if ( !szStringBuffer || !*szStringBuffer )
		{
			return 0;
		}
		else
		{
			return -1;
		}
	}
	else
	{
		if ( !szStringBuffer || !*szStringBuffer )
		{
			return 1;
		}
	}

	if ( bIgnoreCase )
	{
		return stricmp( m_szData, szStringBuffer );				
	}
	else
	{
		return strcmp( m_szData, szStringBuffer );
	}
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator >
//
// Description: reutrns true if this string is greater than the other string (case-insensitive)
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::operator > ( const Asura_Char* const szBfr ) const 
{ 
	return( Compare( szBfr ) > 0 ); 
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator >
//
// Description: reutrns true if this string is less than the other string (case-insensitive)
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::operator < ( const Asura_Char* const szBfr ) const
{ 
	return( Compare( szBfr ) < 0 ); 
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator ==
//
// Description: reutrns true if this string equals the other string (case-insensitive)
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::operator ==( const Asura_Char* const szBfr ) const 
{ 
	return( Compare( szBfr ) ==  0 ); 
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::operator !=
//
// Description: reutrns true if this string does not equal the other string (case-insensitive)
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::operator !=( const Asura_Char* const szBfr ) const 
{ 
	return( Compare( szBfr ) !=  0 ); 
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::StartsWith
//
// Description: returns true if this string starts with the specified string 
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::StartsWith( const Asura_Char* const szPrefix, const bool bIgnoreCase ) const
{
	return Asura_StringUtil::StartsWith( m_szData, szPrefix, bIgnoreCase );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::EndsWith
//
// Description: returns true if this string ends with the specified string 
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::EndsWith( const Asura_Char* const szSuffix, const bool bIgnoreCase ) const
{
	return Asura_StringUtil::EndsWith( m_szData, szSuffix, bIgnoreCase );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Contains
//
// Description: returns true if this string contains the other string 
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::Contains( const Asura_Char* const szString, int* piPositionInString ) const
{
	return Asura_StringUtil::Contains( m_szData, szString, piPositionInString );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::MakeLowercase
//
// Description: makes this string lowercase
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::MakeLowercase( void )
{
	return Asura_StringUtil::MakeLowercase( m_szData );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::MakeUppercase
//
// Description: makes this string lowercase
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::MakeUppercase( void )
{
	return Asura_StringUtil::MakeUppercase( m_szData );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::SplitFilename
//
// Description: puts the path, filename, extension into the input string buffers (which must match the template size of this string)
//-------------------------------------------------------------------
template<u_int uLength> 
bool Asura_StringBuffer<uLength>::SplitFilename( Asura_StringBuffer<uLength>* const pxPath, Asura_StringBuffer<uLength>* const pxFilename, Asura_StringBuffer<uLength>* const pxExtension ) const
{
	if ( IsBlank() ) 
	{
		return false; // don't split if there's nothing to split
	}

	Asura_StringBuffer<uLength> xWorkspace;
	xWorkspace = *this;
	int i;
	i = xWorkspace.GetLength();
	// grab extension
	for ( ; ( xWorkspace[i] != '.' && ( i != 0 ) ) ; i-- ) { }

	if ( xWorkspace[i]=='.')
	{
		xWorkspace[i] = 0;

		if ( pxExtension )
		{
			*pxExtension = ( &xWorkspace[i+1] );
		}
	}
	else if ( pxExtension )
	{
		pxExtension->MakeBlank();
	}

	// grab filename
	i = xWorkspace.GetLength();
    for ( ; ( ( xWorkspace[i] != '\\' && xWorkspace[i] != '/' ) && ( i != 0 ) ); i-- ) { }

	if ( pxFilename )
	{
		// (AAM - 15.11.2006) Used not to check for the slashes here, but that resulted in
		// inputs of the form "\\foo.bar" producing a filename of "\\foo" rather than "foo"

		if ( i == 0 && xWorkspace[i] != '\\' && xWorkspace[i] != '/')
		{
			*pxFilename = ( &xWorkspace[0] );
		}
		else
		{
			*pxFilename = ( &xWorkspace[i+1] );
		}
	}

	if ( ( xWorkspace[i]=='\\' ) || ( xWorkspace[i]=='/' ) )
	{
		// there should be a path too
		if ( pxPath )
		{		
			xWorkspace[i] = 0;
			*pxPath = ( &xWorkspace[0] );
		}
	}
			
	return true;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Truncate
//
// Description: reduce the length of string to iMaxLength
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::Truncate( const u_int uMaxLength )
{
	const int iNewLength = Asura_Maths::Min( (uLength-1), uMaxLength );
	m_szData[iNewLength] = '\0';
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::StripWhitespace
//
// Description: use strip flag values from Asura_StringUtil.
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::StripWhitespace( const u_int uStripFlags )
{
	u_int uNewLen;
	// Returns the new ptr and the new length
	const Asura_Char* szNewBfr = Asura_StringUtil::StripWhitespace( m_szData, uStripFlags, &uNewLen );
	// Should never get NULL back
	Asura_Assert( ( szNewBfr != NULL ), "" );

	Asura_Assert( ( uNewLen < uLength ), "" );

	// So the new string is at szNewBfr and it is uNewLen characters long
	// Do the copy - the copy is safe even source & destination are the same memory because source > destination
	Asura_PrintfToBfr( m_szData, "%s", uNewLen+1, true, szNewBfr );
	// Null terminate
	m_szData[uNewLen] = '\0';
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::ReplaceAllChars
//
// Description: replace all char instances with another char e.g. replace '\' with '/'
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::ReplaceAllChars( const Asura_Char cSearch, const Asura_Char cReplace )
{
	Asura_StringUtil::ReplaceAllChars( m_szData, cSearch, cReplace );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Copy
//
// Description: copy from the specified string (same as assignment operator)
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::Copy( const Asura_Char* const szFrom )
{
	Asura_StringBuffer<uLength>::operator=( szFrom );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::NCopy
//
// Description: copy n characters from target string
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::NCopy( const Asura_Char* const szFrom, const u_int uMaxLength )
{
	const u_int uSize = Asura_Maths::Min( uMaxLength+1, uLength );
	Asura_PrintfToBfr( m_szData, "%s", uSize, true, szFrom );
	// Zero-terminate the string
	m_szData[uLength-1] = '\0';
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Cat
//
// Description: append a character onto this string ( same as += operator )
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::Cat( const Asura_Char* const szString )
{
	Asura_StringBuffer<uLength>::operator += ( szString );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::Cat
//
// Description: append a character onto this string 
//-------------------------------------------------------------------
template<u_int uLength> 
inline void Asura_StringBuffer<uLength>::Cat( const Asura_Char cChar )
{
	Asura_Char szString[2];
	szString[0] = cChar;
	szString[1] = 0;
	Asura_StringBuffer<uLength>::operator += ( szString );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetBoolValue
//
// Description: convert this string into a bool : return false if failed
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::GetBoolValue( bool* const pbValue ) const
{
	return Asura_StringUtil::GetBoolValue( m_szData, pbValue );
}
//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetFloatValue
//
// Description: convert this string into a float : return false if failed
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::GetFloatValue( float* const pfValue ) const
{
	return Asura_StringUtil::GetFloatValue( m_szData, pfValue );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::GetIntValue
//
// Description: convert this string into a int : return false if failed
//-------------------------------------------------------------------
template<u_int uLength> 
inline bool Asura_StringBuffer<uLength>::GetIntValue( int* const piValue, const bool bForceHex ) const
{
	return Asura_StringUtil::GetIntValue( m_szData, piValue, bForceHex );
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::FindChar
//
// Description: return index in the string where this character has been found
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::FindChar( const Asura_Char cChar ) const
{
	for ( u_int i = 0; i < uLength; i++ )
	{
		if ( m_szData[i] == cChar )
		{
			return i;
		}
		if ( m_szData[i] == 0 )
		{
			return -1;
		} 
	}
	return -1;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::ReverseFindChar
//
// Description: return index in the string where this character has been found (reverse search)
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::ReverseFindChar( const Asura_Char cChar ) const
{
	const int iLen = GetLength();
	if ( iLen == 0 ) 
	{
		return -1;
	}
	for ( int i = iLen-1; i >= 0; i-- )
	{
		if ( m_szData[i] == cChar )
		{
			return i;
		}
	}
	return -1;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::VPrint
//
// Description: Write a formatted string into the buffer
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::VPrint( const Asura_Char* const szFormat, va_list& xArgs )
{
	const int iLen = Asura_PrintfToBfr( m_szData, szFormat, xArgs, uLength, true );
	return iLen;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::VPrint
//
// Description: Write a formatted string into the buffer
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::Print( const Asura_Char* const szFormat, ... )
{
	va_list xArgs;
	va_start( xArgs, szFormat );
	const int iLen = VPrint( szFormat, xArgs );
	va_end( xArgs );
	return iLen;
}

//-------------------------------------------------------------------
// Function:    Asura_StringBuffer<uLength>::CatPrint
//
// Description: append a formatted string onto the buffer
//-------------------------------------------------------------------
template<u_int uLength> 
inline int Asura_StringBuffer<uLength>::CatPrint( const Asura_Char* const szFormat, ... )
{
	va_list xArgs;
	va_start( xArgs, szFormat );
	const int iCurrentLength = GetLength();
	const int iNewLength = Asura_PrintfToBfr( m_szData+iCurrentLength, szFormat, xArgs, ( uLength - iCurrentLength ), true );
	va_end( xArgs );

	return iNewLength;

}

//-------------------------------------------------------------------
// Function:    operator <<
//
// Description: Inserts an Asura_StringBuffer into an Asura_Chunk_Stream
//-------------------------------------------------------------------
template<u_int uLength>
inline Asura_Chunk_Stream& operator <<(Asura_Chunk_Stream& xStream, const Asura_StringBuffer<uLength>& xStringBuffer)
{
	ASURA_STATIC_ASSERT( sizeof(Asura_Char) == sizeof(char) );

	if( !xStringBuffer.IsBlank() )
	{
		int iZero = 0;
		xStream.Write((xStringBuffer.GetLength() + 1) * sizeof(Asura_Char), xStringBuffer.GetString());
		const int iNumPadBytes = 4 - (((xStringBuffer.GetLength() + 1) * sizeof(Asura_Char)) % 4);
		if(iNumPadBytes < 4) xStream.Write(iNumPadBytes, reinterpret_cast<void*>(&iZero));
	}
	else
	{
		return xStream << Asura_String::NULL_STRING_CHUNK;
	}
	return xStream;
}

//-------------------------------------------------------------------
// Function:    operator >>
//
// Description: Extracts an Asura_StringBuffer from an Asura_Chunk_Stream
//-------------------------------------------------------------------
template<u_int uLength>
inline Asura_Chunk_Stream& operator >>(Asura_Chunk_Stream& xStream, Asura_StringBuffer<uLength>& xStringBuffer)
{
	ASURA_STATIC_ASSERT( sizeof(Asura_Char) == sizeof(char) );

    // This will break horribly if uLength < 4 since ReadString() below
    // reads in 4 byte chunks.
    ASURA_STATIC_ASSERT( uLength >= 4 );

	char szReadStringBuffer[uLength];
	xStream.ReadString(szReadStringBuffer, uLength);

	//check for the null string identifier
	if(*(int*)&szReadStringBuffer[0] == Asura_String::NULL_STRING_CHUNK)
	{
		xStringBuffer.MakeBlank();
	}
	else
	{
		xStringBuffer = szReadStringBuffer;
	}
	return xStream;
}

// 64 character Asura_StringBuffer
typedef Asura_StringBuffer<64> Asura_String64;

// 256 character Asura_StringBuffer
typedef Asura_StringBuffer<256> Asura_String256;

// 2048 character Asura_StringBuffer
typedef Asura_StringBuffer<2048> Asura_String2048;

#endif // #ifndef __ASURA_STRINGBUFFER_H__
