//
// Asura_Texture_Streaming.cpp
//
// Created by KJFL - 10 March 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_TEXTURE_STREAMING

// This file's header.
#include "Core/Asura_Texture_Streaming.h"

#include "Core/Asura_Atomic.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_DebugText.h"
#include "Rendering/Asura_Environment_Renderer.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif

u_int					Asura_Texture_Streaming_System::s_uNumberOfStreamingTextures;
Asura_Streaming_Texture	Asura_Texture_Streaming_System::s_axStreamingTextures[ASURA_MAX_NO_OF_TEXTURES];
short					Asura_Texture_Streaming_System::s_sTextureIndexToStreamingConversion[ASURA_MAX_NO_OF_TEXTURES];

Asura_Semaphore*		Asura_Texture_Streaming_System::s_pxLoadingThreadTerminated;
Asura_Semaphore*		Asura_Texture_Streaming_System::s_pxTextureLoadRequired;
Asura_Semaphore*		Asura_Texture_Streaming_System::s_pxLoadingThreadReady;

Asura_StringBuffer<128> Asura_Texture_Streaming_System::s_xTextureStreamingFilename;

u_int					Asura_Texture_Streaming_System::s_uMaxNumberOfLoadedTextures = ASURA_TEXTURE_STREAMING_DEFAULT_MAX_LOADED_TEXTURES;
u_int					Asura_Texture_Streaming_System::s_uNumberOfLoadedTextures;

u_int					Asura_Texture_Streaming_System::s_uTextureMemoryUsedInBytes;


DVSVAR float	g_fLoadingDelay		= 0.0f;
DVSVAR bool		g_bDisplayDebugInfo	= false;

static bool g_bTextureStreamingThreadActive = false;
static Asura_Streaming_Texture* volatile g_pxTextureToLoad = 0;

// KJFL - 26 March 2008 - not happy with this mutex but it makes Reset()
// completely robust.
static Asura_Mutex g_xLoadingMutex;

static  u_int g_uLowestLoadedPriority = 0xffffffff;
static  u_int g_uHighestUnloadedPriority = 0;

void Asura_Texture_Streaming_System::TextureStreamingThread( const void* pParam )
{
	
	while (g_bTextureStreamingThreadActive)
	{
		s_pxLoadingThreadReady->Signal();
		s_pxTextureLoadRequired->Wait();

		g_xLoadingMutex.Lock();

		if ( g_bTextureStreamingThreadActive )
		{
			Asura_Streaming_Texture* pxTextureToLoad = g_pxTextureToLoad;

			if ( pxTextureToLoad )
			{
				pxTextureToLoad->SetStatus(Asura_Streaming_Texture::AST_LOADING);
			
				if ( g_fLoadingDelay > 0.01f )
				{
					Asura_Thread_System::DelayCurrentThread( g_fLoadingDelay );
				}

				Platform_StreamTexture( *pxTextureToLoad );
			}
		}

		g_xLoadingMutex.Unlock();
	}
	
	s_pxLoadingThreadTerminated->Signal();
}


bool Asura_Texture_Streaming_System::ConsoleCommand_UnloadAll( bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( L"Throws away all loaded textures." );
    }
    else
    {
		for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
		{
			if ( s_axStreamingTextures[u].GetStatus() == Asura_Streaming_Texture::AST_LOADED )
			{
					s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_FREEIN3FRAMES );
			}
		}
    }

    return true;
}

#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
bool Asura_Texture_Streaming_System::ConsoleCommand_RemoveFromCache( bool bHelp )
{
	if( bHelp )
	{
		Asura_CommandConsole::Printf( L"Throws away all loaded textures in the cache file." );
	}
	else
	{
		for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
		{
			s_axStreamingTextures[u].SetIsInCache( false );
		}
	}

	return true;
}
#endif

void Asura_Texture_Streaming_System::Initialise( void )
{
	Platform_Initialise();

	// initialise the conversion lookup table
	for (u_int u=0; u<ASURA_MAX_NO_OF_TEXTURES; ++u)
	{
		s_sTextureIndexToStreamingConversion[ u ] = -1;
	}

	// Need these event in place before the loading thread is created.
	s_pxTextureLoadRequired = new Asura_Semaphore( "ATSS_TextureLoadRequired", 0 , 2 );
	s_pxLoadingThreadTerminated = new Asura_Semaphore( "ATSS_LoadingThreadTerminated", 0, 1 );
	s_pxLoadingThreadReady = new Asura_Semaphore( "ATSS_LoadingThreadReady", 0, 1 );

	s_uNumberOfStreamingTextures = 0;
	s_uNumberOfLoadedTextures = 0;

	s_uTextureMemoryUsedInBytes = 0;

	g_bTextureStreamingThreadActive = true;

	Asura_Thread_System::CreateThread( "TextureStreamingThread", TextureStreamingThread, 0 );

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar( L"TextureStreaming.LoadingDelay", &g_fLoadingDelay, 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"TextureStreaming.DisplayDebugInfo", &g_bDisplayDebugInfo, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
    Asura_CommandConsole::AddCmd( L"TextureStreaming.UnloadAll", ConsoleCommand_UnloadAll, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
		Asura_CommandConsole::AddCmd( L"TextureStreaming.RemoveAllTexturesFromCache", ConsoleCommand_RemoveFromCache, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif
#endif
}

void Asura_Texture_Streaming_System::Reset( void )
{
	// Tell texture streaming thread it needs to stop loading things
	g_pxTextureToLoad = 0;

	// Wait until the loading thread has paused. Currently doing this
	// via the loading mutex, since this will guarantee the loading thread
	// is now waiting for the "s_pxTextureLoadRequired" semaphore
	g_xLoadingMutex.Lock();
	g_xLoadingMutex.Unlock();

	// TODO - do we need to wait a frame?
	
	// release any textures we may have loaded
	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		Asura_Streaming_Texture::AST_STATUS eStatus = s_axStreamingTextures[u].GetStatus();
		
		Asura_Assert( eStatus != Asura_Streaming_Texture::AST_LOADING, "Code bug: streaming texture is still loading" );
	
		if ( eStatus != Asura_Streaming_Texture::AST_NOTLOADED )
		{
			int iIndex	= s_axStreamingTextures[u].GetTextureIndex();

			if(iIndex >= 0)
			{
				Platform_UnloadTexture( iIndex );
			}
			s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_NOTLOADED );
		}
	}

	s_uNumberOfStreamingTextures = 0;
	s_uNumberOfLoadedTextures = 0;
	
	s_uTextureMemoryUsedInBytes = 0;

	// tell the platform to close the streaming file or whatever else it needs to do
	Platform_EndStreaming();
	
	// initialise the conversion lookup table
	for (u_int u=0; u<ASURA_MAX_NO_OF_TEXTURES; ++u)
	{
		s_sTextureIndexToStreamingConversion[ u ] = -1;
	}
}

void Asura_Texture_Streaming_System::Shutdown( void )
{
	// indicate the streaming thread should exit
	g_pxTextureToLoad = 0;
	g_bTextureStreamingThreadActive = false;

	// need to wake up the thread - there's not really textures to load
	s_pxTextureLoadRequired->Signal();
	
	// waiting until the loading thread has truly finished
	s_pxLoadingThreadTerminated->Wait();

	// TODO - do we need to wait a frame?

	// release any textures we may have loaded
	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		if ( s_axStreamingTextures[u].GetStatus() == Asura_Streaming_Texture::AST_LOADED )
		{
			int	iIndex	= s_axStreamingTextures[u].GetTextureIndex();

			if(iIndex >= 0)
			{
				Platform_UnloadTexture( iIndex );
			}
			s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_NOTLOADED );
		}
	}

#ifdef ASURA_ASSERT
	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		Asura_Assert( s_axStreamingTextures[u].GetStatus() != Asura_Streaming_Texture::AST_LOADING, "Code bug: streaming texture is loading" );
		Asura_Assert( s_axStreamingTextures[u].GetStatus() != Asura_Streaming_Texture::AST_LOADED, "Code bug: streaming texture is loaded" );
	}
#endif // ASURA_ASSERT
	
	delete s_pxLoadingThreadTerminated;
	delete s_pxTextureLoadRequired;
	delete s_pxLoadingThreadReady;

	Platform_Shutdown();
}



void Asura_Texture_Streaming_System::Update( void )
{
	ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__TEXTURE_STREAMING_UPDATE, "Texture-streaming Update");

	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		// KJFL - TODO - we don't need to do this every frame
		int iIndex = s_axStreamingTextures[u].GetTextureIndex();

		Asura_Assert( iIndex>=0, "Code bug: texture not found in level" );
		if ( iIndex >= 0 )
		{
			s_sTextureIndexToStreamingConversion[ iIndex ] = static_cast<short>( u );
		}

		switch (s_axStreamingTextures[u].GetStatus())
		{
			default:
			{
				break;
			}
			case Asura_Streaming_Texture::AST_FREEIN3FRAMES:
			{
				s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_FREEIN2FRAMES );
				break;
			}
			case Asura_Streaming_Texture::AST_FREEIN2FRAMES:
			{
				s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_FREENEXTFRAME );
				break;
			}
			case Asura_Streaming_Texture::AST_FREENEXTFRAME:
			{
				if(iIndex >= 0)
				{
					Platform_UnloadTexture( iIndex );
				}
				s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_NOTLOADED );
				break;
			}
		}
	}

	if(g_bDisplayDebugInfo)
	{
#ifdef ASURA_USE_DEBUG_TEXT
		Asura_DebugText::AddTextBlock( 20, 20, 0xffffff00, 0, "Total textures in world: %d", s_uNumberOfStreamingTextures );
		Asura_DebugText::AddTextBlock( 20, 21, 0xffffff00, 0, "Textures allowed loaded: %d", s_uMaxNumberOfLoadedTextures );
		Asura_DebugText::AddTextBlock( 20, 22, 0xffffff00, 0, "Textures currently loaded: %d", s_uNumberOfLoadedTextures );
		if (ASURA_TEXTURE_STREAMING_MEMORY_ALLOWED_MB == 0)
		{
			Asura_DebugText::AddTextBlock( 20, 23, 0xffffff00, 0, "Texture mem allowed: Unlimited");
		}
		else
		{
			Asura_DebugText::AddTextBlock( 20, 23, 0xffffff00, 0, "Texture mem allowed: %d kb", ASURA_TEXTURE_STREAMING_MEMORY_ALLOWED_MB * 1024 );
		}
		Asura_DebugText::AddTextBlock( 20, 24, 0xffffff00, 0, "Texture mem used: %d kb", s_uTextureMemoryUsedInBytes / 1024 );

		Asura_DebugText::AddTextBlock( 20, 26, 0xffffff00, 0, "Highest unloaded priority %d", g_uHighestUnloadedPriority );
		Asura_DebugText::AddTextBlock( 20, 27, 0xffffff00, 0, "Lowest loaded priority %d", g_uLowestLoadedPriority );
#endif
	}

	// KJFL - 28 August 2008 - if loading thread isn't ready for a new texture, there's no point
	// looking for the best one to load. We'll just leave it until next frame (when the best
	// texture to load might be different anyway). If the loading thread is flagged as ready,
	// we either have to give it work or signal it's still ready - if this TryWait() is successful
	// it will clear the ready status.
	if ( !s_pxLoadingThreadReady->TryWait() )
	{
		ASURA_CPU_PROFILE__END( ASURA_CPU_PROFILE__TEXTURE_STREAMING_UPDATE );
		return;
	}

	Asura_Environment_Renderer::SetTexturePriorities();

	const u_int uTimeStamp = Asura_Timers::GetFrameNumber();

	s_uNumberOfLoadedTextures= 0;
	s_uTextureMemoryUsedInBytes = 0;

	g_uLowestLoadedPriority = 0xffffffff;
	u_int g_uLowestLoadedPriorityTexture = 0;
	g_uHighestUnloadedPriority = 0;
	u_int g_uHighestUnloadedPriorityTexture = 0;

	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		u_int uPriority = s_axStreamingTextures[u].GetPriority();

		if ( s_axStreamingTextures[u].GetStatus() == Asura_Streaming_Texture::AST_LOADED )
		{
			++s_uNumberOfLoadedTextures;
			s_uTextureMemoryUsedInBytes += s_axStreamingTextures[u].GetFileSize();

		
			if ( uPriority==0 && s_axStreamingTextures[u].GetLastUsedFrameStamp() + 100 < uTimeStamp )
			{
				s_axStreamingTextures[u].SetStatus( Asura_Streaming_Texture::AST_FREEIN3FRAMES );
			}
			else if ( uPriority < g_uLowestLoadedPriority )
			{
				g_uLowestLoadedPriority = uPriority;
				g_uLowestLoadedPriorityTexture = u;
			}
		}
		else if ( s_axStreamingTextures[u].GetStatus() == Asura_Streaming_Texture::AST_NOTLOADED )
		{
			if ( uPriority > g_uHighestUnloadedPriority )
			{
				g_uHighestUnloadedPriority = uPriority;
				g_uHighestUnloadedPriorityTexture = u;
			}
		}
	}

	u_int uTextureMemoryAllowedInBytes = (ASURA_TEXTURE_STREAMING_MEMORY_ALLOWED_MB == 0) ? UINT_MAX : (ASURA_TEXTURE_STREAMING_MEMORY_ALLOWED_MB * 1048576);

	bool bRemoveATexture = false;

	// If we're over memory, or over the allowed number of textures, get rid of the lowest priority one.
	// This should only really happen if we're reducing the upper limits on the fly (e.g. via the DVS).
	if ( ( s_uNumberOfLoadedTextures > s_uMaxNumberOfLoadedTextures ) || (s_uTextureMemoryUsedInBytes > uTextureMemoryAllowedInBytes) )
	{
		bRemoveATexture = true;
	}
	// If we have an unloaded texture which is a higher priority than a currently loaded one, 
	// check to see if we need to remove the loaded texture to make room.
	else if ( g_uHighestUnloadedPriority > g_uLowestLoadedPriority )
	{
		u_int uTextureMemRequired = s_uTextureMemoryUsedInBytes + s_axStreamingTextures[g_uHighestUnloadedPriorityTexture].GetFileSize();

		// if we've ran out of textures, or if we've ran out of memory, the lowest priority texture must go
		if ( (s_uNumberOfLoadedTextures == s_uMaxNumberOfLoadedTextures) || ( uTextureMemRequired > uTextureMemoryAllowedInBytes ) ) 
		{
			bRemoveATexture = true;
		}
	}

	if ( bRemoveATexture )
	{
		Asura_Assert( s_axStreamingTextures[g_uLowestLoadedPriorityTexture].GetStatus() == Asura_Streaming_Texture::AST_LOADED, "Code bug: texture should be loaded" );

		s_axStreamingTextures[g_uLowestLoadedPriorityTexture].SetStatus( Asura_Streaming_Texture::AST_FREEIN3FRAMES );
	}
	
	// KJFL - 28 August 2008 - we must either signal there is work, or that the loading thread is ready for work,
	// so we can come round again next frame.
	if ( (g_uHighestUnloadedPriority > 0) && (s_uNumberOfLoadedTextures < s_uMaxNumberOfLoadedTextures)
		&& (uTextureMemoryAllowedInBytes > s_uTextureMemoryUsedInBytes + s_axStreamingTextures[g_uHighestUnloadedPriorityTexture].GetFileSize()) )
	{
		// if there is space, load the texture
		g_pxTextureToLoad = &s_axStreamingTextures[g_uHighestUnloadedPriorityTexture];
		s_pxTextureLoadRequired->Signal();
	}
	else
	{
		// otherwise signal we're still ready for next frame
		s_pxLoadingThreadReady->Signal();
	}

	ASURA_CPU_PROFILE__END(ASURA_CPU_PROFILE__TEXTURE_STREAMING_UPDATE);
}

bool Asura_Texture_Streaming_System::IsTextureLoaded( int iTextureIndex )
{
	if (iTextureIndex<0) return false;

	int iStreamingTexture = s_sTextureIndexToStreamingConversion[ iTextureIndex ];

	if ( iStreamingTexture >= 0 )
	{
		return ( s_axStreamingTextures[iStreamingTexture].GetStatus() == Asura_Streaming_Texture::AST_LOADED );
	}

	return false;
}


void Asura_Texture_Streaming_System::SetTexturePriority( int iTextureIndex, u_int uPriority )
{
	if (iTextureIndex<0) return;

	int iStreamingTexture = s_sTextureIndexToStreamingConversion[ iTextureIndex ];

	Asura_Assert( iStreamingTexture>=0, "Code bug: matching streaming texture not found" );

	if ( iStreamingTexture >= 0 )
	{
		s_axStreamingTextures[iStreamingTexture].SetPriority( uPriority );
	}
}

Asura_Streaming_Texture* Asura_Texture_Streaming_System::GetBestTextureToLoad( void )
{
	if (!g_bTextureStreamingThreadActive) return 0;

	if ( s_uMaxNumberOfLoadedTextures <=  s_uNumberOfLoadedTextures )
	{
		return 0;
	}

	// TODO - cache the best as we go along? lazy sort of some kind? 
	// Possibly it doesn't have to be *the* best texture, just one of top ones.
	u_int uHighestPriority = 0;
	u_int uTexture = 0;

	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		if ( s_axStreamingTextures[u].GetStatus() == Asura_Streaming_Texture::AST_NOTLOADED
		 &&  s_axStreamingTextures[u].GetPriority() > uHighestPriority ) 
		{
			uHighestPriority = s_axStreamingTextures[u].GetPriority();
			uTexture = u;
		}
	}
	
	if ( uHighestPriority )
	{
		return &s_axStreamingTextures[uTexture];
	}
	else
	{
		return 0;
	}
}


bool Asura_Texture_Streaming_System::ShouldUpdatePriorities( void )
{
	return true;
}

Asura_Streaming_Texture* Asura_Texture_Streaming_System::GetStreamingTexture( u_int u )
{
	return &s_axStreamingTextures[u];
}

#ifdef ASURA_TOOLS

void Asura_Texture_Streaming_System::Tools_RegisterTexture( int iTextureIndex, int iTextureUsage )
{
	if (iTextureIndex<0) return;

	// TODO - locks? atomic adds?
	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		if ( s_axStreamingTextures[u].GetTextureIndex() == iTextureIndex )
		{
			return;
		}
	}
	
	s_axStreamingTextures[ s_uNumberOfStreamingTextures ].Initialise( iTextureIndex, iTextureUsage );

	++s_uNumberOfStreamingTextures;
}

bool Asura_Texture_Streaming_System::Tools_IsTextureRegistered( int iTextureIndex )
{
	if ( iTextureIndex<0 ) return false;

	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		if ( s_axStreamingTextures[u].GetTextureIndex() == iTextureIndex ) return true;
	}

	return false;
}

void Asura_Texture_Streaming_System::Tools_FindStreamingTextures( void )
{
	Asura_Environment_Renderer::GetTexturesUsed( Tools_RegisterTexture );

	//Go through the list of streaming textures , and filter out any that are unsuitable for streaming.
	//Initially applies to the fmv texture , and to textures that are very small.
	//RAR 23/03/2009
	for ( u_int u=0; u<s_uNumberOfStreamingTextures; ++u )
	{
		if(!Tools_Level_Export::IsTextureOkayForStreaming(s_axStreamingTextures[u].GetTextureIndex()))
		{
			s_axStreamingTextures[u] = s_axStreamingTextures[s_uNumberOfStreamingTextures-1];
			s_uNumberOfStreamingTextures--;
			u--;
		}
	}
}

#endif

void Asura_Texture_Streaming_System::Test_SetLoadingDelay(float fDelay)
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_Assert(fDelay >= 0.0f, "Cannot set a negative delay.");
	g_fLoadingDelay = Asura_Maths::Max(fDelay, 0.0f);
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
}

void Asura_Texture_Streaming_System::Test_UnloadAll()
{
	// Should be fine to just go through the console function.
	ConsoleCommand_UnloadAll(false);
}

#endif // #ifdef ASURA_USE_TEXTURE_STREAMING


