//
// Asura_Timers.h
//
// Created by Kevin Lea - 12/06/2001 11:01:59
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_TIMERS_H__
#define __ASURA_TIMERS_H__

#include "Core/Asura_Platform_Timers.h"	// for Asura_TickCount

class Asura_Timers
{
	public:
        
        static void Initialise( void );
		static void Reset( void );
        static void Shutdown( void );
        
		static void HandlePrimaryTimer(void);
		
		static inline float GetFrameRate(void)
		{	
			return s_fFrameRate;
		}

        static inline float GetSmoothedFrameRate( void ) { return s_fSmoothedFrameRate; }

		static inline float GetFrameTime(void)
		{
			return s_fFrameTime;
		}

        // returns the amount of time elapsed since this frame started capped to the minimum framerate
        static float GetTimeInFrame(void);
		// returns the amount of time elapsed since this frame started
		static float GetTimeInFrame_Uncapped(void);

		static inline float GetGameTimeSlice(void)
		{
			return s_fGameTime;
		}
		static inline float GetPreviousFrameTime(void)
		{
			return s_fPreviousFrameTime;
		}


		static inline float GetLocalUnscaledTime()
		{
			return s_fLocalUnscaledTime;
		}
		
		static inline float GetLocalScaledTime()
		{
			return s_fLocalScaledTime;
		}

		#ifdef ASURA_TOOLS
		//Gets the overall timer in a similar way to GetPhase() but without the errors caused by capping the
		//frame times to 0.2 seconds. RAR 30/09/2003
		static inline float GetUncappedPhase(void)
		{	
			return s_fUncappedPhase;
		}
		#endif

		static void SetGameTimeScale(float fScale);

		static inline float GetGameTimeScale(void)
		{
			return s_bSystemPause ? 0.f : s_fGameTimeScale;
		}

		static inline unsigned int GetFrameNumber(void)
		{
			return s_uFrameNumber;
		}

		// server absolute time is a timer which is synchronised across the network.
		// the client updates it every frame and corrects it over a few frames
		// when a time update is received from the server.
		static inline float GetServerAbsoluteTime(void)
		{
			return s_fServerAbsoluteTime;
		}

		// (MH - 25.11.2003) GetGlobalGameTime is now deprecated. Use GetServerAbsoluteTime() instead.
		static inline float GetGlobalGameTime(void)
		{
			return GetServerAbsoluteTime();
		}

		// UpdateServerAbsoluteTime is called when receiving a message from the server 
		// latency variations from frame-to-frame may be averaged out. bInstant 
		// can be used to force an immediate update otherwise update will happen 
		// over several frames.
		static void UpdateServerAbsoluteTime(float fNewServerTime, float fCurrentLatency, bool bInstant = false);
		static void NeedToResyncServerAbsoluteTime();
		static bool HasServerAbsoluteTimeBeenSynchronised();

		#ifdef ASURA_TOOLS
		//If fixed game time is set to a non-zero value , then this value will be used
		//for updating s_fGameTimeScale and s_fGamePhase instead of the actual time.
		//Used for getting a fixed frame rate during avi recording
		//RAR 03/02/03
		static void SetFixedGameTime(float fTime);
		static float GetFixedGameTime();
		#endif

		// GetMSecsSinceStart() - returns milliseconds since game was started. Note that it
		// is only updated with each call to HandlePrimaryTimer().
		static inline unsigned int GetMSecsSinceStart() { return s_uMilliSecondsSinceStart; }
		// GetMSecsSinceStart_Continuous() - like GetMSecsSinceStart(), but is updated continuously.
		// Note that on PS2 it will only handle one timer wrap so HandlePrimaryTimer() must 
		// still be called regularly.
		static unsigned int GetMSecsSinceStart_Continuous();

		// AJG - 04/07/05 - Fixed frame rate functions. See comments above variables for explanation of meanings.
		static void SetFixedFrameRate(float fDesiredFrameRate)
		{
			if (fDesiredFrameRate <= 0.0f)
			{
				s_fFixedFrameRate = 0.0f;
				s_fFixedFrameTime = 0.0f;
			}
			else
			{
				s_fFixedFrameRate = fDesiredFrameRate;
				s_fFixedFrameTime = 1.0f / fDesiredFrameRate;
			}
		}
		static void SetFixedFrameTime(float fDesiredFrameTime)
		{
			if (fDesiredFrameTime <= 0.0f)
			{
				s_fFixedFrameRate = 0.0f;
				s_fFixedFrameTime = 0.0f;
			}
			else
			{
				s_fFixedFrameRate = 1.0f / fDesiredFrameTime;
				s_fFixedFrameTime = fDesiredFrameTime;
			}
		}
		static float GetFixedFrameRate(void)
		{
			return(s_fFixedFrameRate);
		}
		static float GetFixedFrameTime(void)
		{
			return(s_fFixedFrameTime);
		}

		static float GetActualFrameRate()
		{
			return s_fActualFrameRate;
		}
		static float GetActualFrameTime()
		{
			return s_fActualFrameTime;
		}

        // (MH - 10.01.2007) SetMaxFrameRate() - this differs from SetFixedFrameTime
		// in that it doesn't slow things down when the frame rate isn't reached.
		static void SetMaxFrameRate(float fMaxFrameRate) { s_fMaxFrameRate = fMaxFrameRate; }
		static float GetMaxFrameRate() { return s_fMaxFrameRate; }

		// Interface for getting platform-dependent "ticks" for timing purposes. 
		// These could be u_int, int64, uint64 etc. and can have any frequency 
		// (but are expected to be high-resolution). 
		static Asura_TickCount GetCurTicks()
		{
			return Platform_GetCurTicks();
		}
		static Asura_TickCount GetElapsedTicks(Asura_TickCount xStartTicks)
		{
			return Platform_GetElapsedTicks(xStartTicks);
		}
		static float TicksToSeconds(Asura_TickCount uTicks)
		{
			return Platform_TicksToSeconds(uTicks);
		}
		static Asura_TickCount SecondsToTicks(float fSeconds)
		{
			return Platform_SecondsToTicks(fSeconds);
		}
		// Used whenever the system needs to over-ride the timescale, e.g. Wii home button menu or PSP sleep mode
		static void SetSystemPause(bool bPause)	{ s_bSystemPause = bPause; }
		static bool GetSystemPause() { return s_bSystemPause; }

	private:

        // Checks the fixed and max frame rates and determines which, if
        // either, should be used. 0 => no clamping required.
        static float GetEffectiveMinFrameTime( void );

		static void Platform_Initialise();
		static void Platform_Reset();
		static void Platform_Shutdown();
		static void Platform_Update();
		static u_int Platform_GetTicksThisFrame();
        static u_int Platform_GetTicksInFrame();
		static float Platform_GetTickFrequency();
		static float Platform_GetDefaultFrameRate();

		static Asura_TickCount Platform_GetCurTicks();
		static Asura_TickCount Platform_GetElapsedTicks(Asura_TickCount xStartTicks);
		static float Platform_TicksToSeconds(Asura_TickCount xTicks);
		static Asura_TickCount Platform_SecondsToTicks(float fSeconds);
			
		static float s_fPreviousFrameTime;
		static float s_fFrameTime;
		static float s_fFrameRate;
        static float s_fSmoothedFrameRate;
		static float s_fGameTime;
		static float s_fLocalUnscaledTime;
		static float s_fLocalScaledTime;
		static float s_fGameTimeScale;
		static float s_fTickFrequency;
		#ifdef ASURA_TOOLS
		static float s_fUncappedPhase;
		#endif

		// AJG - 04/07/05 - Values for fixed frame rate.
		// Initially, this is being used so we can record inputs at a fixed rate, and play them back at the same fixed rate.
		// If our actual frame rate is too high, we'll block to avoid things happening too fast for the user to keep up with.
		// If our actual frame rate is too low, time slows down to compensate (although as far as the exterior code is concerned, it's playing as fast as normal).
		// Obviously, the latter of these changes is what differentiates this from a purely capped frame rate.
		// These first two variables are 0 if we aren't using a fixed rate.
		static float s_fFixedFrameRate;
		static float s_fFixedFrameTime;
		// If the fixed frame rate is 0, these are the same as s_fFrameRate and s_fFrameTime respectively, otherwise they match what those would have been without the fixed value.
		// These still get capped when the rate is too high, it's just for when the rate is too low
		static float s_fActualFrameRate;
		static float s_fActualFrameTime;

        // (MH - 10.01.2007) s_fMaxFrameRate differs from s_fFixedFrameRate in that it 
		// only clamps one end.
		static float s_fMaxFrameRate;
	
		static unsigned int s_uFrameNumber;

		static float s_fTicksPerMilliSecond;
		static float s_fTicksPerMicroSecond;
		static unsigned int s_uMilliSecondsSinceStart;
		static unsigned int s_uMicroSecondsRemainder;

		static float s_fServerAbsoluteTime;
		static bool s_bSystemPause;
};

#endif//__ASURA_TIMERS_H__
