/**
*
* \file Asura_UserInput.cpp
*
* Created by Kevin Lea - 10/01/2002 15:31:50
*
* Copyright (C) 2001 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "Core/Asura_UserInput.h"

#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"

#include "Debugging/Asura_Debug_Timers.h"

#include <wchar.h>

unsigned char	Asura_UserInput::PreviousInput[MAX_NUMBER_OF_ASURA_KEYIDS];		
unsigned char	Asura_UserInput::CurrentInput[MAX_NUMBER_OF_ASURA_KEYIDS];		
bool			Asura_UserInput::AnyKeyPressedHasBeenEvaluated;
bool			Asura_UserInput::AnyKeyHasBeenPressed;
bool            Asura_UserInput::s_bDebouncedAnyKeyPressedHasBeenEvaluated;
bool            Asura_UserInput::s_bDebouncedAnyKeyHasBeenPressed;
float			Asura_UserInput::AxisValue[ASURA_MAX_NUMBER_OF_AXES];
bool            Asura_UserInput::s_bInitialised = false;
bool            Asura_UserInput::s_bAnyAxisMovedHasBeenEvaluated;
bool            Asura_UserInput::s_bAnyAxisHasBeenMoved;

u_int Asura_UserInput::s_uInputBufferSize = 0;
const Asura_UnicodeChar* Asura_UserInput::s_wszInputBuffer = NULL;


#define ASURA_USER_INPUT__HAS_AXIS_MOVED__DEAD_ZONE		(0.01f)

void Asura_UserInput::Initialise(void)
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_UserInput::Initialise");
	
	AnyKeyPressedHasBeenEvaluated = false;
    s_bDebouncedAnyKeyPressedHasBeenEvaluated = false;
	s_bAnyAxisMovedHasBeenEvaluated	= false;

	for(int i=0; i<MAX_NUMBER_OF_ASURA_KEYIDS; i++)
	{
		PreviousInput[i] = 0;
		CurrentInput[i] = 0;
	}
	for(int i=0; i<ASURA_MAX_NUMBER_OF_AXES; i++)
	{
		AxisValue[i] = 0.0f;
	}

	Platform_Initialise();
	PlatformProject_Initialise();
    s_bInitialised = true;
}
void Asura_UserInput::Shutdown(void)
{
	Platform_Shutdown();
	PlatformProject_Shutdown();
    s_bInitialised = false;
}
void Asura_UserInput::Reset(void)
{
	for(int i=0; i<MAX_NUMBER_OF_ASURA_KEYIDS; i++)
	{
		PreviousInput[i] = 0;
		CurrentInput[i] = 0;
	}
	for(int i=0; i<ASURA_MAX_NUMBER_OF_AXES; i++)
	{
		AxisValue[i] = 0.0f;
	}

	Platform_Reset();
	PlatformProject_Reset();
}

void Asura_UserInput::Update(void)
{
	AnyKeyPressedHasBeenEvaluated = false;
    s_bDebouncedAnyKeyPressedHasBeenEvaluated = false;
	s_bAnyAxisMovedHasBeenEvaluated	= false;
	
	Asura_Mem::Copy( PreviousInput, CurrentInput, sizeof( CurrentInput ) );
	Asura_Mem::Set( CurrentInput, 0, sizeof( CurrentInput ) );
	Asura_Mem::Set( AxisValue, 0, sizeof( AxisValue ) );

	Platform_Update();
	PlatformProject_Update();
	
	// SNH: reset 'void' key status: this was being incorrectly set by the PS2 USB keyboard handler
	//	for any keys it didn't recognise (printscrn).
	PreviousInput[ASURA_KEYID_VOID] = 0;
	CurrentInput[ASURA_KEYID_VOID] = 0;
}

bool Asura_UserInput::HasAnyKeyBeenPressed(void)
{
	if (!AnyKeyPressedHasBeenEvaluated)
	{
		AnyKeyPressedHasBeenEvaluated=true;
		AnyKeyHasBeenPressed = false;
		{
			int i = MAX_NUMBER_OF_ASURA_KEYIDS;
		
			while(--i)
			{
				if (CurrentInput[i])
				{
					AnyKeyHasBeenPressed = true;
					break;
				}
			}
		}
	}
	return AnyKeyHasBeenPressed;
}

bool Asura_UserInput::HasAnyAxisBeenMoved(void)
{
	if (!s_bAnyAxisMovedHasBeenEvaluated)
	{
		s_bAnyAxisMovedHasBeenEvaluated = true;
		s_bAnyAxisHasBeenMoved = false;

		int i = ASURA_MAX_NUMBER_OF_AXES;
		while(--i)
		{
			if ( Asura_Maths::Abs( AxisValue[i] ) > ASURA_USER_INPUT__HAS_AXIS_MOVED__DEAD_ZONE )
			{
				s_bAnyAxisHasBeenMoved = true;
				break;
			}
		}
	}
	return s_bAnyAxisHasBeenMoved;
}


bool Asura_UserInput::DebouncedHasAnyKeyBeenPressed( void )
{
	if( !s_bDebouncedAnyKeyPressedHasBeenEvaluated )
	{
		s_bDebouncedAnyKeyHasBeenPressed = false;
        for( int i = MAX_NUMBER_OF_ASURA_KEYIDS; --i; )
        {
            if( CurrentInput[i] && !PreviousInput[i] )
            {
                s_bDebouncedAnyKeyHasBeenPressed = true;
                break;
            }
        }

        s_bDebouncedAnyKeyPressedHasBeenEvaluated = true;
    }

	return s_bDebouncedAnyKeyHasBeenPressed;
}


void Asura_UserInput::HandleKeyInput( ASURA_KEYID eKey, Asura_UnicodeChar* wszBuffer, u_int uMaxChars, u_int& uInputCursorPos )
{
    switch( eKey )
    {
        case ASURA_KEYID_HOME:
            uInputCursorPos = 0;
            break;

        case ASURA_KEYID_END:
            uInputCursorPos = wcslen( wszBuffer );
            break;

        case ASURA_KEYID_LEFT:
            if( uInputCursorPos >= uMaxChars )
            {
                uInputCursorPos = wcslen( wszBuffer ) - 1;
            }
            else if( uInputCursorPos )
            {
                --uInputCursorPos;
            }
            break;

        case ASURA_KEYID_RIGHT:
            if( uInputCursorPos < wcslen(wszBuffer) )
            {
                ++uInputCursorPos;
            }
            break;

        case ASURA_KEYID_DEL:
            if( uInputCursorPos >= uMaxChars )
            {
                // this indicates 'clear on first printable input'
                uInputCursorPos = 0;
                wszBuffer[0] = L'\0';
            }
            else
            {
                // Get length of input.
                u_int uLen = wcslen( wszBuffer );
                // Nothing to do if it's empty or if the cursor is at the
                // end.
                if( uLen && uInputCursorPos < uLen )
                {
                    // Move all characters after the current cursor
                    // position down one.
                    for( u_int u = uInputCursorPos + 1; u <= uLen; ++u )
                    {
                        wszBuffer[u - 1] = wszBuffer[u];
                    }
                }
            }
            break;

        case ASURA_KEYID_BACKSPACE:
            if( uInputCursorPos >= uMaxChars )
            {
                // this indicates 'clear on first printable input'
                uInputCursorPos = 0;
                wszBuffer[0] = L'\0';
            }
            else if( uInputCursorPos )
            {
                // Get length of input.
                u_int uLen = wcslen( wszBuffer );

                // Move the characters down from the cursor position
                // onwards. Note that this also deals with the case
                // where the cursor is at the end of the input
                // because it'll just move the nul terminator.
                for( u_int u = uInputCursorPos; u <= uLen; ++u )
                {
                    wszBuffer[u - 1] = wszBuffer[u];
                }
            
                // Update the cursor position.
                --uInputCursorPos;

            }
            break;

        default:
        {
            int iKeyPressed;

            if( eKey >= ASURA_KEYID_NUMPAD0 && eKey <= ASURA_KEYID_NUMPAD9 )
            {
                iKeyPressed = '0' + (eKey - ASURA_KEYID_NUMPAD0);
            }
            else if( eKey >= ASURA_KEYID_SPACE && eKey < ASURA_NUMBER_OF_ASCII_KEYS )
            {
                iKeyPressed = eKey;
            }    
            else
            {
                // Ignore anything else.
                break;
            }

            const u_int uLen = wcslen( wszBuffer );

            if( uInputCursorPos >= uMaxChars )
            {
                // this indicates 'clear on first printable input'
                uInputCursorPos = 0;
                wszBuffer[uInputCursorPos++] = static_cast<Asura_UnicodeChar>( iKeyPressed );
                wszBuffer[uInputCursorPos]   = '\0';
            }
            else if( uLen + 2 <= uMaxChars ) // +2 because of new character and nul terminator.
            {
                if( uLen == uInputCursorPos )
                {
                    wszBuffer[uInputCursorPos++] = static_cast<Asura_UnicodeChar>( iKeyPressed );
                    wszBuffer[uInputCursorPos]   = '\0';
                }
                else
                {
                    // Move all characters from the cursor position onwards
                    // up one position. Note that we want to move the nul
                    // terminator while we're at it. We need to start at the
                    // end because otherwise we'll overwrite stuff!
                    for( u_int u = uLen + 1; u-- > uInputCursorPos; )
                    {
                        wszBuffer[u + 1] = wszBuffer[u];
                    }

                    // And insert the new character, moving the cursor on
                    // while we're at it.
                    wszBuffer[uInputCursorPos++] = static_cast<Asura_UnicodeChar>( iKeyPressed );
                }
            }

            break;
        }
    }
    // End switch( eKey ).
}


void Asura_UserInput::UpdateInputBuffer( Asura_UnicodeChar* wszBuffer,
                                         u_int uMaxChars, 
                                         u_int& uInputCursorPos,
                                         bool (*pfnHandleInputOverride)(ASURA_KEYID eKey),
                                         const Asura_UnicodeChar* wszInputBuffer,
                                         u_int uInputBufferSize )
{
// (MP - 09.03.2004) Hmm, actually, this just seemed to be a bug in the
// PC input handling code that meant it didn't insert other interesting
// key codes into the input buffer. It resulted in duplicate input on the
// PS2.
//  if (wszInputBuffer == s_wszInputBuffer)
//  {
//      // some keys won't appear in the standard input buffer, check for them separately
//      const ASURA_KEYID aeSpecialKeys[] = 
//      {
//          ASURA_KEYID_HOME, ASURA_KEYID_END, ASURA_KEYID_LEFT, ASURA_KEYID_RIGHT, ASURA_KEYID_UP, ASURA_KEYID_DOWN, 
//              ASURA_KEYID_DEL, ASURA_KEYID_PAGEUP, ASURA_KEYID_PAGEDOWN
//      };
//
//      u_int u = sizeof(aeSpecialKeys) / sizeof(aeSpecialKeys[0]);
//      for( const ASURA_KEYID* pxKey = aeSpecialKeys; u--; ++pxKey )
//      {
//          if( Asura_UserInput::DebouncedKeyPressed(*pxKey) && (!pfnHandleInputOverride || !pfnHandleInputOverride(*pxKey)) )
//          {
//              HandleKeyInput( *pxKey, wszBuffer, uMaxChars, uInputCursorPos );
//          }
//      }
//  }

	/// Richard May - 18/03/2004 10:11:44 - added this function to replicate what's
	/// commented out above but on a PC only basis
	Platform_UpdateInputBuffer(wszBuffer,uMaxChars,uInputCursorPos,pfnHandleInputOverride,wszInputBuffer,uInputBufferSize);

    for( const Asura_UnicodeChar* wszCurrBufferChar = wszInputBuffer; uInputBufferSize--; ++wszCurrBufferChar )
    {
        if( !pfnHandleInputOverride || !pfnHandleInputOverride(static_cast<ASURA_KEYID>(*wszCurrBufferChar)) )
        {
            HandleKeyInput( static_cast<ASURA_KEYID>(*wszCurrBufferChar), wszBuffer, uMaxChars, uInputCursorPos );
        }
    }
}
