
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ResourceSet.h
// Author : Andy Tate
// Date   : 17-Jun-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_RESOURCESET_H__
#define __ASURA_RESOURCESET_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_String.h"
#include "Core/Resources/Asura_ResourceSet_ResourceBank.h"
#include "Core/Resources/Asura_Resource_System.h"
#include "Core/Resources/Asura_ResourceSet_UnifiedResourceContainer.h"
#include "Debugging/Asura_Assert.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_Iterator;
template <class T> class Asura_ResourceSet_ConstIterator;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_Base
// Author       : Andy Tate
// Date         : 18-Jun-2008
//
// Description  : Abstract class used to allow storage of resource-sets without
//				  needing them to be bound to a type.
////////////////////////////////////////////////////////////////////////////////

class Asura_ResourceSet_Base
{
public:
	// Public functions...
	virtual ~Asura_ResourceSet_Base()	{}

	// Reset functions...
	void	Reset(void)									{ DestroyAllResourceBanks(); }
	void	ResetBank(const Asura_Hash_ID uBankHash)	{ DestroyResourceBank(uBankHash); }

	// Accessor functions...
	const Asura_String&	GetName(void) const	{ return(m_xName); }
	Asura_Hash_ID		GetHash(void) const	{ return(m_uHash); }


protected:
	// Protected variables...

	// Protected functions...
	Asura_ResourceSet_Base(const char* szName)	: m_xName(szName), m_uHash(m_xName.GetHash())	{ Asura_Resource_System::AddResourceSet(*this); }

	// Abstract functions...
	virtual bool	CreateResourceBank(const Asura_Hash_ID uBankHash)	= 0;
	virtual bool	DestroyResourceBank(const Asura_Hash_ID uBankHash)	= 0;

	virtual void	DestroyAllResourceBanks(void)						= 0;


private:
	// Private variables...
	Asura_String	m_xName;	// The name of this resource-set.
	Asura_Hash_ID	m_uHash;	// The hash-ID of this resource-set.

	// Private functions...
	Asura_ResourceSet_Base(const Asura_ResourceSet_Base& xOther);

	Asura_ResourceSet_Base&	operator=(const Asura_ResourceSet_Base& xOther);

};   // End of class 'Asura_ResourceSet_Base'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Template class use to hold Asura resource-data.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet : public Asura_ResourceSet_Base
{
	// Friend classes...
	friend class Asura_ResourceSet_Iterator<T>;
	friend class Asura_ResourceSet_ConstIterator<T>;

public:

	// Public functions...
	Asura_ResourceSet(const char* szName)	: PARENT(szName)	{ Asura_Resource_System::AddResourceSet(*this); }
	virtual ~Asura_ResourceSet()								{ DestroyAllResourceBanks(); Asura_Resource_System::RemoveResourceSet(*this); }

	// Resource manipulation functions...
	bool	AddResource(const Asura_Hash_ID uResourceHash, T& xData, bool bAssertIfAlreadyInBank = true);
#if defined(ASURA_TOOLS)
	bool	DeleteResource(const Asura_Hash_ID uResourceHash);
	bool	RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash);

	T*			GetResourceByIndex(const u_int uIndex)			{ return(m_xUnifiedResourceContainer.GetResourceByIndex(uIndex)); }
	const T*	GetResourceByIndex(const u_int uIndex) const	{ return(m_xUnifiedResourceContainer.GetResourceByIndex(uIndex)); }
#endif	//defined(ASURA_TOOLS)

	// Accessor functions...
	T*			GetResource(const Asura_Hash_ID uResourceHash)			{ return(m_xUnifiedResourceContainer.GetResource(uResourceHash)); }
	const T*	GetResource(const Asura_Hash_ID uResourceHash) const	{ return(m_xUnifiedResourceContainer.GetResource(uResourceHash)); }

	u_int	GetNumResources(void) const	{ return(m_xUnifiedResourceContainer.GetNumResources()); }


protected:
	// Protected variables...

	// Protected functions...

	// Overridden creation/destruction functions...
	virtual bool	CreateResourceBank(const Asura_Hash_ID uBankHash);
	virtual bool	DestroyResourceBank(const Asura_Hash_ID uBankHash);

	virtual void	DestroyAllResourceBanks(void);


private:
	// Public typedefs...
	typedef Asura_ResourceSet_Base PARENT;

	typedef Asura_ResourceSet_ResourceBank<T> ResourceBank;
	typedef Asura_TemplateHashTree<ResourceBank> ResourceBankContainer;
	typedef Asura_TemplateHashTree_Iterator<ResourceBank> ResourceBankIterator;
	typedef Asura_TemplateHashTree_ConstIterator<ResourceBank> ResourceBankConstIterator;

	typedef Asura_ResourceSet_UnifiedResourceContainer<T> UnifiedResourceContainer;
	typedef Asura_ResourceSet_UnifiedResourceContainer_Iterator<T> UnifiedResourceIterator;
	typedef Asura_ResourceSet_UnifiedResourceContainer_ConstIterator<T> UnifiedResourceConstIterator;

	// Private variables...
	UnifiedResourceContainer	m_xUnifiedResourceContainer;	// The data-container used to store our unified resources in.
	ResourceBankContainer		m_xResourceBankContainer;		// The data-container used to stored our banks of resources in.

	// Private functions...
	Asura_ResourceSet(const Asura_ResourceSet& xOther);

	Asura_ResourceSet&	operator=(const Asura_ResourceSet& xOther);

};   // End of class 'Asura_ResourceSet'.



////////////////////////////////////////////////////////////////////////////////
// Inlined functions...
////////////////////////////////////////////////////////////////////////////////

template <class T> bool Asura_ResourceSet<T>::CreateResourceBank(const Asura_Hash_ID uBankHash)
{
	ResourceBank*	pxBank	= m_xResourceBankContainer.GetDataByHash(uBankHash);

	if(pxBank == NULL)
	{
		pxBank	= m_xResourceBankContainer.AddNode(uBankHash);

		return(pxBank != NULL);
	}
	else
	{
		Asura_Assert(false, "We already have a resource-bank with this hash-ID.");
		return(false);
	}
}

template <class T> bool Asura_ResourceSet<T>::DestroyResourceBank(const Asura_Hash_ID uBankHash)
{
	ResourceBank*	pxBank	= m_xResourceBankContainer.GetDataByHash(uBankHash);

	if(pxBank == NULL)
	{
		return(false);
	}
	else
	{
		Asura_ResourceSet_ResourceBank_Iterator<T>	xIterator(*pxBank);

		while(!xIterator.Done())
		{
			const Asura_Hash_ID	uResourceHash	= xIterator.GetCurrentHash();

			m_xUnifiedResourceContainer.RemoveLinkToResource(uResourceHash, uBankHash, true);
			xIterator.Next();
		}

		return(m_xResourceBankContainer.DeleteNodeByHash(uBankHash));
	}
}

template <class T> void Asura_ResourceSet<T>::DestroyAllResourceBanks(void)
{
	//Clear out the contents of all the banks.
	for(ResourceBankIterator xIt(m_xResourceBankContainer); !xIt.Done(); xIt.Next())
	{
		DestroyResourceBank(xIt.GetCurrentHash());
	}
	Asura_Assert(m_xUnifiedResourceContainer.GetNumResources() == 0, "All resources should have been deleted as a result of the banks being deleted.");
	//These should be empty by now , so resetting may be a little redundant
	m_xUnifiedResourceContainer.Reset();
	m_xResourceBankContainer.Reset();
}

template <class T> bool Asura_ResourceSet<T>::AddResource(const Asura_Hash_ID uResourceHash, T& xData, bool bAssertIfAlreadyInBank /*= true*/)
{
	const Asura_Hash_ID	uBankHash	= Asura_Resource_System::GetCurrentResourceBank();

	// Ensure that we've got the relevant banks create...
	if(Asura_Resource_System::HasResourceBank(uBankHash))
	{
		ResourceBank*	pxBank	= m_xResourceBankContainer.GetDataByHash(uBankHash);

		// If this is the first time for this bank, ensure that we've created the data-container...
		if((pxBank == NULL) && CreateResourceBank(uBankHash))
		{
			pxBank	= m_xResourceBankContainer.GetDataByHash(uBankHash);

			if(pxBank == NULL)
			{
				return(false);
			}
		}

		// Add this resource to the appropriate places...
		if(pxBank == NULL)
		{
			Asura_Assert(false, "Failed to find the specified resource-bank to add this resource to.");
			return(false);
		}
		else
		{
			if(pxBank->AddLinkToResource(uResourceHash, xData, bAssertIfAlreadyInBank))
			{
				return(m_xUnifiedResourceContainer.AddLinkToResource(uResourceHash, uBankHash, xData));
			}
			else
			{
				return(false);
			}
		}
	}
	else
	{
		Asura_Assert(false, "The resource-bank to add this resource to doesn't exist - something's gone very wrong.");
		return(false);
	}
}

#if defined(ASURA_TOOLS)

template <class T> bool Asura_ResourceSet<T>::DeleteResource(const Asura_Hash_ID uResourceHash)
{
	const Asura_Hash_ID	uBankHash	= Asura_Resource_System::GetCurrentResourceBank();
	ResourceBank*		pxBank		= m_xResourceBankContainer.GetDataByHash(uBankHash);

	if(pxBank == NULL)
	{
		Asura_Assert(false, "Failed to find the specified resource-bank to delete this resource from.");
		return(false);
	}
	else
	{
		if(pxBank->RemoveLinkToResource(uResourceHash))
		{
			if(m_xUnifiedResourceContainer.RemoveLinkToResource(uResourceHash, uBankHash, true))
			{
				return(true);
			}
			else
			{
				return(false);
			}
		}
		else
		{
			return(false);
		}
	}
}

template <class T> bool Asura_ResourceSet<T>::RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash)
{
	const Asura_Hash_ID	uBankHash	= Asura_Resource_System::GetCurrentResourceBank();
	ResourceBank*		pxBank		= m_xResourceBankContainer.GetDataByHash(uBankHash);

	if(pxBank == NULL)
	{
		Asura_Assert(false, "Failed to find the specified resource-bank to delete this resource from.");
		return(false);
	}
	else
	{
		if(pxBank->RelocateResource(uOldResourceHash, uNewResourceHash))
		{
			return(m_xUnifiedResourceContainer.RelocateResource(uOldResourceHash, uNewResourceHash, uBankHash));
		}
		else
		{
			return(false);
		}
	}
}

#endif	//defined(ASURA_TOOLS)



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_Iterator
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to iterate through all of the resources
//				  held by the associated resource-set.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_Iterator
{
public:
	// Public functions...
	Asura_ResourceSet_Iterator(Asura_ResourceSet<T>& xResourceSet)
	: m_xUnifiedResourceIterator(xResourceSet.m_xUnifiedResourceContainer)
	{
	}

	~Asura_ResourceSet_Iterator()	{}

	// Manipulation functions...
	void			Begin(void)					{ m_xUnifiedResourceIterator.Begin(); }
	void			Next(void)					{ m_xUnifiedResourceIterator.Next(); }
	void			End(void)					{ m_xUnifiedResourceIterator.End(); }

	bool			Done(void) const			{ return(m_xUnifiedResourceIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xUnifiedResourceIterator.GetCurrentHash()); }
	T*				GetCurrent(void)			{ return(m_xUnifiedResourceIterator.GetCurrent()); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_ResourceSet_UnifiedResourceContainer_Iterator<T> UnifiedResourceIterator;

	// Private variables...
	UnifiedResourceIterator	m_xUnifiedResourceIterator;	// Iterator for the unified data.

	// Private functions...
	Asura_ResourceSet_Iterator(void);
	Asura_ResourceSet_Iterator(const Asura_ResourceSet_Iterator& xOther);

	Asura_ResourceSet_Iterator&	operator=(const Asura_ResourceSet_Iterator& xOther);

};   // End of class 'Asura_ResourceSet_Iterator'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_ConstIterator
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to iterate through all of the resources
//				  held by the associated resource-set.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_ConstIterator
{
public:
	// Public functions...
	Asura_ResourceSet_ConstIterator(const Asura_ResourceSet<T>& xResourceSet)
	: m_xUnifiedResourceConstIterator(xResourceSet.m_xUnifiedResourceContainer)
	{
	}

	~Asura_ResourceSet_ConstIterator()	{}

	// Manipulation functions...
	void			Begin(void)					{ m_xUnifiedResourceConstIterator.Begin(); }
	void			Next(void)					{ m_xUnifiedResourceConstIterator.Next(); }
	void			End(void)					{ m_xUnifiedResourceConstIterator.End(); }

	bool			Done(void) const			{ return(m_xUnifiedResourceConstIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xUnifiedResourceConstIterator.GetCurrentHash()); }
	const T*		GetCurrent(void) const		{ return(m_xUnifiedResourceConstIterator.GetCurrent()); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_ResourceSet_UnifiedResourceContainer_ConstIterator<T>	UnifiedResourceConstIterator;

	// Private variables...
	UnifiedResourceConstIterator	m_xUnifiedResourceConstIterator;	// ConstIterator for the unified data.

	// Private functions...
	Asura_ResourceSet_ConstIterator(void);
	Asura_ResourceSet_ConstIterator(const Asura_ResourceSet_ConstIterator& xOther);

	Asura_ResourceSet_ConstIterator&	operator=(const Asura_ResourceSet_ConstIterator& xOther);

};   // End of class 'Asura_ResourceSet_ConstIterator'.


#endif   // ifndef __ASURA_RESOURCESET_H__


/// End of file. ///
