
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ResourceSet_ResourceBank.h
// Author : Andy Tate
// Date   : 17-Jun-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_RESOURCESET_RESOURCEBANK_H__
#define __ASURA_RESOURCESET_RESOURCEBANK_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_HashTree.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_ResourceBank_Iterator;
template <class T> class Asura_ResourceSet_ResourceBank_ConstIterator;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_ResourceBank
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to store all of the resources for a 
//				  particular "bank".
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_ResourceBank
{
	friend class Asura_ResourceSet_ResourceBank_Iterator<T>;
	friend class Asura_ResourceSet_ResourceBank_ConstIterator<T>;

public:
	// Public functions...
	Asura_ResourceSet_ResourceBank(void)	{}
	//~Asura_ResourceSet_ResourceBank()		{}

	// Resource manipulation functions...
	bool	AddLinkToResource(const Asura_Hash_ID uResourceHash, T& xData, bool bAssertIfAlreadyInBank);
	bool	RemoveLinkToResource(const Asura_Hash_ID uResourceHash);
#if defined(ASURA_TOOLS)
	bool	RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash);
#endif	// defined(ASURA_TOOLS)

	

protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_TemplateHashTree<T*> ResourceContainer;
	typedef Asura_TemplateHashTree_Iterator<T*> ResourceIterator;

	// Private variables...
	ResourceContainer	m_xResourceContainer;	// The data-container for the resource held by this bank.

	// Private functions...
	Asura_ResourceSet_ResourceBank(const Asura_ResourceSet_ResourceBank& xOther);

	Asura_ResourceSet_ResourceBank&	operator=(const Asura_ResourceSet_ResourceBank& xOther);

};   // End of class 'Asura_ResourceSet_ResourceBank'.



////////////////////////////////////////////////////////////////////////////////
// Inlined functions...
////////////////////////////////////////////////////////////////////////////////

template <class T> bool Asura_ResourceSet_ResourceBank<T>::AddLinkToResource(const Asura_Hash_ID uResourceHash, T& xData, bool bAssertIfAlreadyInBank)
{
	T**	ppxData	=	m_xResourceContainer.AddNode(uResourceHash, &xData);

	if(ppxData == NULL)
	{
		Asura_Assert(!bAssertIfAlreadyInBank, "A resource with this hash-ID has already been added to this resource-bank.");
		return(false);
	}
	else
	{
		return(true);
	}
}

template <class T> bool Asura_ResourceSet_ResourceBank<T>::RemoveLinkToResource(const Asura_Hash_ID uResourceHash)
{
	return(m_xResourceContainer.DeleteNodeByHash(uResourceHash));
}

#if defined(ASURA_TOOLS)

template <class T> bool Asura_ResourceSet_ResourceBank<T>::RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash)
{
	return(m_xResourceContainer.ReplaceHash(uOldResourceHash, uNewResourceHash));
}

#endif	//defined(ASURA_TOOLS)





////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_ResourceBank_Iterator
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to iterate through all of the resources
//				  held by this resource-bank.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_ResourceBank_Iterator
{
public:
	// Public functions...
	Asura_ResourceSet_ResourceBank_Iterator(Asura_ResourceSet_ResourceBank<T>& xResourceBank)
	: m_xResourceIterator(xResourceBank.m_xResourceContainer)
	{
	}

	~Asura_ResourceSet_ResourceBank_Iterator()	{}

	// Manipulation functions...
	void			Begin(void)					{ m_xResourceIterator.Begin(); }
	void			Next(void)					{ m_xResourceIterator.Next(); }
	void			End(void)					{ m_xResourceIterator.End(); }

	bool			Done(void) const			{ return(m_xResourceIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xResourceIterator.GetCurrentHash()); }
	T*				GetCurrent(void)			{ T**	ppxData	= m_xResourceIterator.GetCurrent(); return((ppxData == NULL) ? NULL : *ppxData); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_TemplateHashTree_Iterator<T*> ResourceIterator;

	// Private variables...
	ResourceIterator	m_xResourceIterator;	// Iterator for the resources within the resource-bank.

	// Private functions...
	Asura_ResourceSet_ResourceBank_Iterator(void);
	Asura_ResourceSet_ResourceBank_Iterator(const Asura_ResourceSet_ResourceBank_Iterator& xOther);

	Asura_ResourceSet_ResourceBank_Iterator&	operator=(const Asura_ResourceSet_ResourceBank_Iterator& xOther);

};   // End of class 'Asura_ResourceSet_ResourceBank_Iterator'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_ResourceBank_ConstIterator
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to iterate through all of the resources
//				  held by this resource-bank.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_ResourceBank_ConstIterator
{
public:
	// Public functions...
	Asura_ResourceSet_ResourceBank_ConstIterator(const Asura_ResourceSet_ResourceBank<T>& xResourceBank)
	: m_xResourceConstIterator(xResourceBank.m_xResourceContainer)
	{
	}

	~Asura_ResourceSet_ResourceBank_ConstIterator()	{}

	// Manipulation functions...
	void			Begin(void)					{ m_xResourceConstIterator.Begin(); }
	void			Next(void)					{ m_xResourceConstIterator.Next(); }
	void			End(void)					{ m_xResourceConstIterator.End(); }

	bool			Done(void) const			{ return(m_xResourceConstIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xResourceConstIterator.GetCurrentHash()); }
	const T*		GetCurrent(void) const		{ const T**	ppxData	= m_xResourceConstIterator.GetCurrent(); return((ppxData == NULL) ? NULL : *ppxData); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_TemplateHashTree_ConstIterator<T*> ResourceConstIterator;

	// Private variables...
	ResourceConstIterator	m_xResourceConstIterator;	// Constant-iterator for the resources within the resource-bank.

	// Private functions...
	Asura_ResourceSet_ResourceBank_ConstIterator(void);
	Asura_ResourceSet_ResourceBank_ConstIterator(const Asura_ResourceSet_ResourceBank_ConstIterator& xOther);

	Asura_ResourceSet_ResourceBank_ConstIterator&	operator=(const Asura_ResourceSet_ResourceBank_ConstIterator& xOther);

};   // End of class 'Asura_ResourceSet_ResourceBank_ConstIterator'.


#endif   // ifndef __ASURA_RESOURCESET_RESOURCEBANK_H__


/// End of file. ///
