
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ResourceSet_UnifiedResourceContainer.h
// Author : Andy Tate
// Date   : 17-Jun-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_RESOURCESET_UNIFIEDRESOURCECONTAINER_H__
#define __ASURA_RESOURCESET_UNIFIEDRESOURCECONTAINER_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_HashTree.h"
#include "Core/Resources/Asura_ResourceSet_UnifiedResource.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_UnifiedResourceContainer_Iterator;
template <class T> class Asura_ResourceSet_UnifiedResourceContainer_ConstIterator;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_UnifiedResourceContainer
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Template class used to store multiple sets of resource-data
//				  with the same hash-ID.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_UnifiedResourceContainer
{
	// Friend classes...
	friend class Asura_ResourceSet_UnifiedResourceContainer_Iterator<T>;
	friend class Asura_ResourceSet_UnifiedResourceContainer_ConstIterator<T>;

public:

	// Public typedefs...
	typedef Asura_ResourceSet_UnifiedResource<T> UnifiedResource;

	// Public functions...
	Asura_ResourceSet_UnifiedResourceContainer(void)	{}
	~Asura_ResourceSet_UnifiedResourceContainer()		{}

	void	Reset(void)	{ m_xUnifiedResourceContainer.Reset(); }

	// Resource manipulation functions...
	bool	AddLinkToResource(const Asura_Hash_ID uResourceHash, const Asura_Hash_ID uBankHash, T& xData);
	bool	RemoveLinkToResource(const Asura_Hash_ID uResourceHash, const Asura_Hash_ID uBankHash, bool bDeleteResourceIfNoRemainingReferences);
#if defined(ASURA_TOOLS)
	bool	RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash, const Asura_Hash_ID uBankHash);

	T*			GetResourceByIndex(const u_int uIndex);
	const T*	GetResourceByIndex(const u_int uIndex) const;
#endif	// defined(ASURA_TOOLS)

	// Accessor functions...
#if defined(ASURA_TOOLS)
	T*	GetResource(const Asura_Hash_ID uResourceHash, const Asura_Hash_ID uBankHash);
#endif	//defined(ASURA_TOOLS)

	T*			GetResource(const Asura_Hash_ID uResourceHash);
	const T*	GetResource(const Asura_Hash_ID uResourceHash) const;

	u_int	GetNumResources(void) const	{ return(m_xUnifiedResourceContainer.GetNumberOfNodes()); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_TemplateHashTree<UnifiedResource> UnifiedResourceContainer;

	// Private variables...
	UnifiedResourceContainer	m_xUnifiedResourceContainer;	// The data-container used to stored our unified-resources in.

	// Private functions...
	Asura_ResourceSet_UnifiedResourceContainer(const Asura_ResourceSet_UnifiedResourceContainer& xOther);

	Asura_ResourceSet_UnifiedResourceContainer&	operator=(const Asura_ResourceSet_UnifiedResourceContainer& xOther);

	

};   // End of class 'Asura_ResourceSet_UnifiedResourceContainer'.



////////////////////////////////////////////////////////////////////////////////
// Inlined functions...
////////////////////////////////////////////////////////////////////////////////

template <class T> bool Asura_ResourceSet_UnifiedResourceContainer<T>::AddLinkToResource(const Asura_Hash_ID uResourceHash, const Asura_Hash_ID uBankHash, T& xData)
{
	UnifiedResource*	pxInitialResource	= m_xUnifiedResourceContainer.GetDataByHash(uResourceHash);

	if(pxInitialResource == NULL)
	{
		pxInitialResource	= m_xUnifiedResourceContainer.AddNode(uResourceHash);

		if(pxInitialResource == NULL)
		{
			return(false);
		}
		else
		{
			pxInitialResource->Initialise(uBankHash, xData);
			return(true);
		}
	}
	else
	{
		//If an entry for the resource already exists ,make sure we are adding the same data again. RAR 17/02/2009
		Asura_Assert(pxInitialResource->GetData() == &xData, "Different pointer being used while adding an existing resource. This will cause memory leaks.");
				
		UnifiedResource*	pxNextResource	= pxInitialResource->CreateNextResource(uBankHash, xData);

		return(pxNextResource != NULL);
	}
}

template <class T> bool Asura_ResourceSet_UnifiedResourceContainer<T>::RemoveLinkToResource(const Asura_Hash_ID uResourceHash, const Asura_Hash_ID uBankHash, bool bDeleteResourceIfNoRemainingReferences)
{
	UnifiedResource*	pxInitialResource	= m_xUnifiedResourceContainer.GetDataByHash(uResourceHash);

	if(pxInitialResource == NULL)
	{
		Asura_Assert(false, "Failed to find resource to unlink.");
		return(false);
	}
	else
	{
		UnifiedResource*	pxNextResource	= pxInitialResource->GetNext();

		// Is the first resource the one we want to delete?
		if(pxInitialResource->GetBankHash() == uBankHash)
		{
			// Yes - is this our only node?
			if(pxNextResource == NULL)
			{
				//Last reference to the resource , so delete the underlying data and delete the node from
				//the resource container
				if(bDeleteResourceIfNoRemainingReferences)
				{
					delete pxInitialResource->GetData();
				}
				m_xUnifiedResourceContainer.DeleteNodeByHash(uResourceHash);
				return(true);
			}
			else
			{
				// No - copy the second resource into the first one...
				pxInitialResource->CopyData(*pxNextResource);
				pxInitialResource->DestroyNextResource();
				return(true);
			}
		}
		else
		{
			UnifiedResource*	pxCurrentResource	= pxInitialResource;

			// No - search through our list to find the one we want...
			while(pxNextResource != NULL)
			{
				if(pxNextResource->GetBankHash() == uBankHash)
				{
					pxCurrentResource->DestroyNextResource();
					return(true);
				}
				else
				{
					pxCurrentResource	= pxNextResource;
					pxNextResource		= pxCurrentResource->GetNext();
				}
			}

			Asura_Assert(false, "Unexpected bank-hash found when attempting to unlink the last resource in the unified-container.");
			return(false);
		}
	}
}

#if defined(ASURA_TOOLS)

template <class T> bool Asura_ResourceSet_UnifiedResourceContainer<T>::RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash, const Asura_Hash_ID uBankHash)
{
	T*	pxResource	= GetResource(uOldResourceHash);

	if(pxResource == NULL)
	{
		Asura_Assert(false, "Failed to find resource to be relocated for this unified-resource-container.");
		return(false);
	}
	else
	{
		if(AddLinkToResource(uNewResourceHash, uBankHash, *pxResource))
		{
			return(RemoveLinkToResource(uOldResourceHash, uBankHash, false));
		}
		else
		{
			Asura_Assert(false, "Failed to add the new resource (during relocation) to this unified-resource-container.");
			return(false);
		}
	}
}

template <class T> T* Asura_ResourceSet_UnifiedResourceContainer<T>::GetResource(const Asura_Hash_ID uResourceHash, const Asura_Hash_ID uBankHash)
{
	Asura_ResourceSet_UnifiedResource<T>*	pxUnifiedResource	= m_xUnifiedResourceContainer.GetDataByHash(uResourceHash);

	if(pxUnifiedResource == NULL)
	{
		return(NULL);
	}
	else
	{
		return(pxUnifiedResource->GetData(uBankHash));
	}
}

template <class T> T* Asura_ResourceSet_UnifiedResourceContainer<T>::GetResourceByIndex(const u_int uIndex)
{
	Asura_ResourceSet_UnifiedResource<T>*	pxUnifiedResource	= m_xUnifiedResourceContainer.GetDataByIndex(uIndex);

	if(pxUnifiedResource == NULL)
	{
		return(NULL);
	}
	else
	{
		return(pxUnifiedResource->GetData());
	}
}

template <class T> const T* Asura_ResourceSet_UnifiedResourceContainer<T>::GetResourceByIndex(const u_int uIndex) const
{
	const Asura_ResourceSet_UnifiedResource<T>*	pxUnifiedResource	= m_xUnifiedResourceContainer.GetDataByIndex(uIndex);

	if(pxUnifiedResource == NULL)
	{
		return(NULL);
	}
	else
	{
		return(pxUnifiedResource->GetData());
	}
}

#endif	//defined(ASURA_TOOLS)

template <class T> T* Asura_ResourceSet_UnifiedResourceContainer<T>::GetResource(const Asura_Hash_ID uResourceHash)
{
	Asura_ResourceSet_UnifiedResource<T>*	pxUnifiedResource	= m_xUnifiedResourceContainer.GetDataByHash(uResourceHash);

	if(pxUnifiedResource == NULL)
	{
		return(NULL);
	}
	else
	{
		return(pxUnifiedResource->GetData());
	}
}

template <class T> const T* Asura_ResourceSet_UnifiedResourceContainer<T>::GetResource(const Asura_Hash_ID uResourceHash) const
{
	const Asura_ResourceSet_UnifiedResource<T>*	pxUnifiedResource	= m_xUnifiedResourceContainer.GetDataByHash(uResourceHash);

	if(pxUnifiedResource == NULL)
	{
		return(NULL);
	}
	else
	{
		return(pxUnifiedResource->GetData());
	}
}




////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_UnifiedResourceContainer_Iterator
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to iterate through all of the resources
//				  held by the associated resource-set.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_UnifiedResourceContainer_Iterator
{
public:
	// Public functions...
	Asura_ResourceSet_UnifiedResourceContainer_Iterator(Asura_ResourceSet_UnifiedResourceContainer<T>& xContainer)
	: m_xUnifiedResourceIterator(xContainer.m_xUnifiedResourceContainer)
	{
	}

	~Asura_ResourceSet_UnifiedResourceContainer_Iterator()	{}

	// Manipulation functions...
	void			Begin(void)					{ m_xUnifiedResourceIterator.Begin(); }
	void			Next(void)					{ m_xUnifiedResourceIterator.Next(); }
	void			End(void)					{ m_xUnifiedResourceIterator.End(); }

	bool			Done(void) const			{ return(m_xUnifiedResourceIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xUnifiedResourceIterator.GetCurrentHash()); }

	T* GetCurrent(void)
	{
		UnifiedResource*	pxUnifiedResource	= m_xUnifiedResourceIterator.GetCurrent();

		if(pxUnifiedResource == NULL)
		{
			return(NULL);
		}
		else
		{
			return(pxUnifiedResource->GetData());
		}
	}


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_ResourceSet_UnifiedResource<T> UnifiedResource;
	typedef Asura_TemplateHashTree_Iterator<UnifiedResource> UnifiedResourceIterator;

	// Private variables...
	UnifiedResourceIterator	m_xUnifiedResourceIterator;	// Iterator for the unified-resource data.

	// Private functions...
	Asura_ResourceSet_UnifiedResourceContainer_Iterator(void);
	Asura_ResourceSet_UnifiedResourceContainer_Iterator(const Asura_ResourceSet_UnifiedResourceContainer_Iterator& xOther);

	Asura_ResourceSet_UnifiedResourceContainer_Iterator&	operator=(const Asura_ResourceSet_UnifiedResourceContainer_Iterator& xOther);

};   // End of class 'Asura_ResourceSet_UnifiedResourceContainer_Iterator'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_UnifiedResourceContainer_ConstIterator
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Templified class used to iterate through all of the resources
//				  held by the associated resource-set.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_UnifiedResourceContainer_ConstIterator
{
public:
	// Public functions...
	Asura_ResourceSet_UnifiedResourceContainer_ConstIterator(const Asura_ResourceSet_UnifiedResourceContainer<T>& xContainer)
	: m_xUnifiedResourceConstIterator(xContainer.m_xUnifiedResourceContainer)
	{
	}
	~Asura_ResourceSet_UnifiedResourceContainer_ConstIterator()	{}

	// Manipulation functions...
	void			Begin(void)					{ m_xUnifiedResourceConstIterator.Begin(); }
	void			Next(void)					{ m_xUnifiedResourceConstIterator.Next(); }
	void			End(void)					{ m_xUnifiedResourceConstIterator.End(); }

	bool			Done(void) const			{ return(m_xUnifiedResourceConstIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xUnifiedResourceConstIterator.GetCurrentHash()); }

	const T* GetCurrent(void) const
	{
		const UnifiedResource*	pxUnifiedResource	= m_xUnifiedResourceConstIterator.GetCurrent();

		if(pxUnifiedResource == NULL)
		{
			return(NULL);
		}
		else
		{
			return(pxUnifiedResource->GetData());
		}
	}


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_ResourceSet_UnifiedResource<T> UnifiedResource;
	typedef Asura_TemplateHashTree_ConstIterator<UnifiedResource> UnifiedResourceConstIterator;

	// Private variables...
	UnifiedResourceConstIterator	m_xUnifiedResourceConstIterator;	// Constant iterator for the unified-resource data.

	// Private functions...
	Asura_ResourceSet_UnifiedResourceContainer_ConstIterator(void);
	Asura_ResourceSet_UnifiedResourceContainer_ConstIterator(const Asura_ResourceSet_UnifiedResourceContainer_ConstIterator& xOther);

	Asura_ResourceSet_UnifiedResourceContainer_ConstIterator&	operator=(const Asura_ResourceSet_UnifiedResourceContainer_ConstIterator& xOther);

};   // End of class 'Asura_ResourceSet_UnifiedResourceContainer_ConstIterator'.


#endif   // ifndef __ASURA_RESOURCESET_UNIFIEDRESOURCECONTAINER_H__


/// End of file. ///
