#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _ASURA_CUTSCENE_EVENT_H_
#define _ASURA_CUTSCENE_EVENT_H_

#include "Camera/Asura_Camera.h"
#include "Core/Asura_Graph.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Cutscene/Asura_Cutscene_Camera_Shake.h"
#include "BluePrint/Asura_BluePrint.h"
#include "GUIMenu/Asura_GUIMenu_CommandConsole.h"
#include "Network/Asura_Guid.h"
#include "SpecialFX/Asura_Beam_System.h"

class Asura_Chunk_Stream;
class Asura_Cutscene;
class Asura_Cutscene_Actor;

#ifdef ASURA_TOOLS
#include "Misc/list_tem.hpp"
#endif


//////////////////////////////////////////////////////////////////////////////////////////
//types of event
enum Asura_Cutscene_Event_Type
{
	ACET_AnimationChange,
	ACET_TimeScale,
	ACET_FieldOfView,
	ACET_SecondaryAnimation,
	ACET_Trigger,
    ACET_LoopBack,
	ACET_ActivateAttachment,
	ACET_DeactivateAttachment,
	ACET_Obsolete_NextScriptLine,//Old subtitle event
	ACET_Beam,
	ACET_Fade,
	ACET_ReactivateAttachment,
	ACET_Obsolete_DepthOfField,//Old depth of field system
	ACET_TemplateWrapper,	// special event to package up a bunch of events (only used in tools).
	ACET_TextureAnimation,
	ACET_Subtitle,
	ACET_Sound,
	ACET_ChangeCamera,
	ACET_ChangeCameraTarget,
	ACET_AttachmentControl,//handles multiple attachment activations/deactivations at once
	ACET_Hide,
	ACET_Unhide,
	ACET_ChangeCameraShake,
	ACET_SetAlpha,
	ACET_ChangePlayerCameraControl,
    ACET_GUIMenu,
    ACET_ConsoleVar,
    ACET_HDRSettings,
	ACET_FacialAnimation,
	ACET_MixerPreset,

	ACET_FirstProjectSpecific = 1000,//starting point for project specfic event types
};


#ifdef ASURA_TOOLS
//flag defines for the tickbox dialogs
#define AsuraCutsceneEventFlag_IgnoreWhenAdvancingToTime 0x00000001
#define AsuraCutsceneEventFlag_Disabled 0x00000002
#endif

class Asura_Cutscene_Event
{
//    friend class Asura_Chunk_CutsceneEvent;
//    friend class Tools_CutsceneEvent_Object;
//    friend class Tools_Cutscene_Event;
//    friend class Tools_Cutscene_Event_AnimationChange;

public :

	Asura_Cutscene_Event(int iEventType,Asura_Cutscene* pxCutscene);
	virtual ~Asura_Cutscene_Event(){}

	#ifdef ASURA_TOOLS
	//Used as part of the cutscene copying code , should only copy to a newly allocated event
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	//Gets the addresses of all stored event indeces , so they can be remapped to equivalent events after copying
	virtual void GetEventIndexPointers(List<int*>& xIndexList);

	//These two are for moving the members of a cutscene as the cutscene object is moved
	virtual void CalculateRelativePosition(const Asura_Matrix_3x3& xCutsceneOrientInverse,const Asura_Vector_3 xCutscenePosition);
	virtual void UpdateRelativeToPosition(const Asura_Matrix_3x3& xCutsceneOrient,const Asura_Vector_3 xCutscenePosition);
	#endif

	static Asura_Cutscene_Event* AllocateEvent(int iEventType,Asura_Cutscene* pxCutscene);
	static Asura_Cutscene_Event* Project_AllocateEvent(int iEventType,Asura_Cutscene* pxCutscene);

	//Tools use.. remove any links this event may have elsewhere before nuking it
	virtual void PreDestructionCleanup(Asura_Cutscene*){};
	
	virtual void Initialise();
	virtual void Finish(){}
	virtual void Trigger(Asura_Cutscene*);//activates this event
	virtual void UpdateAfterTrigger(Asura_Cutscene*);
	virtual void RenderAfterTrigger(Asura_Cutscene*){};
	virtual void LoadFiles(){}

	void Update(Asura_Cutscene* pxCutscene);//triggers event if needed , and maintains after trigger
	// Ali Jakamy 11-07-08
	// Some events need to render independtly from in game used objects like GUIMenu event
	void Render(Asura_Cutscene* pxCutscene);


    bool HasBeenTriggered() const;
    bool IsTimeToTrigger(float fCurCutsceneTime);
    void HandleCutsceneLoopBack(float fLoopBackTriggerTime, float fTimeToLoopBackTo);
    bool AffectsActor(const Asura_Cutscene_Actor* pActor) const { return m_pxTargetActor == pActor; }
    int  GetType() const { return m_iType; }

	bool ShouldSkipWhenAdvancingTime(){return m_bIgnoreWhenAdvancingToTime;}

	bool IsEnabled() const;

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	virtual int GetChunkSize();
	virtual void SaveToChunk(char*& chunkBuffer);
	virtual void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);
	virtual void ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	#ifdef ASURA_TOOLS
	//Gets the chunk size for the data that is only used in tools (name , comment , position)
	int GetToolsDataSize() const;
	#endif

#ifdef ASURA_TOOLS
    // (MH - 01.11.2002) code to track down unused animations
    virtual bool UsesAnim(int iAnimIndex) {return false;}// note this is an index into the NameList NOT a hashID
#endif

	#ifdef ASURA_TOOLS
	//Resets any guids and static messages so as to make cutscene
	//safe for exporting to other levels.
	virtual void ResetGuidsAndMessages(){};
	#endif

	#ifdef ASURA_TOOLS
	//called during level export to add extra asura files to export list
	virtual void AddExtraFilesToList( List<Asura_String>& xNameList ) const {}
	#endif
	

// (MH - 05.08.2002) this should be protected, but loads of tools classes access this data.

	//common event data
    Asura_Cutscene * m_pxCutscene;
	int m_iType;
	float m_fTriggerTime;//time into the cutscene that this event occurs
	Asura_Cutscene_Actor* m_pxTargetActor; //actor to affect , if relevant
	Asura_String m_xTargetActorName;//name of that actor
	int m_iParentEventIndex;
	// (AW - 31.05.2004) - Data for animation flag triggered events.
	int m_iAnimationFlag;
	Asura_Cutscene_Actor *m_pxAnimationFlagActor;
	Asura_String m_xAnimationFlagActorName;
	// NB - if using this system then m_fTriggerTime is treated as a retrigger delay.

	//ID value which is 0 if unused , or 1 to 32 if valid. Used to compare
	//against a bitfield of enabled events.(Not necessarily unique)
	int m_iEnableID;

	union
	{
		u_int m_uEventFlags;
		struct
		{
			// NOTE: These bit's are written out, adding any new bits means
			// ~~~~~ that you will need to add those bits to the bitfield
			//       preparation code in "WriteToChunkStream".
			u_int m_bIgnoreWhenAdvancingToTime : 1;
			u_int m_bDisabled : 1;
		};
	};


    
    //cutscene playback data
	bool m_bTriggered;
    float m_fNextTriggerTime;

#ifdef ASURA_TOOLS
	Asura_String m_xName;//For overiding the name displayed in tools
	Asura_String m_xComment;//Optional comment for use in tools
	Asura_Vector_3 m_xPosition;//For overiding the event object's position
	Asura_String m_xNameOfSourceTemplate;//if this event belongs to a template, template event's name is here.
	Asura_String m_xNameOfEventWrapper;//if this event belongs to a template, wrapper event name is here.

	//relative position for temporary use while moving the cutscene object
	Asura_Vector_3 m_xRelativePosition;
#endif

};

//////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_Cutscene_Event_AnimationChange **
**--------------------------------------*/

//As it turns out , these flags are just a subset of those in Asura_Animation.h. If this changes
//then care will need to be taken with the start animation functions. RAR 13/08/02
#define AnimationChangeFlag_Loop  0x00000001
#define AnimationChangeFlag_Tween 0x00000002
#define AnimationChangeFlag_Reversed 0x00000004

//And now a flag which isn't the same as the standard animation flags...
#define AnimationChangeFlag_UseCameraTarget 0x20000000 //For camera animations.. determines whether camera target should still be used to determine the base orientation
#define AnimationChangeFlag_IgnoreMovement 0x40000000 //prevents use of movement data stored in anim. (otherwise it may get used , according to whether there is any track movement)
#define AnimationChangeFlag_ToPosition 0x80000000

//flags that are just used for events , and aren't valid animation system flags
#define AnimationChangeFlag_EventSpecificFlags  (AnimationChangeFlag_IgnoreMovement|AnimationChangeFlag_ToPosition)

class Asura_Cutscene_Event_AnimationChange : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_AnimationChange(Asura_Cutscene* _Cutscene);
	virtual void PreDestructionCleanup(Asura_Cutscene*);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif


	void Trigger(Asura_Cutscene*);
	void LoadFiles();
	
	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

#ifdef ASURA_TOOLS
    // (MH - 01.11.2002) code to track down unused animations.
    virtual bool UsesAnim(int iAnimIndex) {return m_iAnimationIndex == iAnimIndex;}// note this is an index into the NameList NOT a hashID
#endif

    //event data
// (MH - 23.07.2002) 
//	Asura_String AnimationName;
//	Asura_String AnimationFileName;
	int m_iAnimationIndex;
	int m_iAnimationFlags;
	float m_fTweenDuration;
	float m_fAnimDelay;//how long to hold on the first frame

	//start/end position used if the ToPosition flag is set
	float m_fStartTimer;
	float m_fEndTimer;
};

//////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------------**
** Asura_Cutscene_Event_SecondaryAnimation **
**-----------------------------------------*/

class Asura_Cutscene_Event_SecondaryAnimation : public Asura_Cutscene_Event_AnimationChange
{
public :
	Asura_Cutscene_Event_SecondaryAnimation(Asura_Cutscene* _Cutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	void Trigger(Asura_Cutscene*);
	void UpdateAfterTrigger(Asura_Cutscene*);
	void Initialise();
	void Finish();

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	//event data
	float m_fWeight;
	float m_fDuration;//zero = play forever , or one loop according to loop flag

	//cutscene playback data
	Asura_Handle m_iAnimHandle;
};

//////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------------**
** Asura_Cutscene_Event_TextureAnimation **
**---------------------------------------*/

#define ASURA_CUTSCENEEVENT_TEXTUREANIMATION_CHANGESEQUENCE 0x00000001
#define ASURA_CUTSCENEEVENT_TEXTUREANIMATION_CHANGESPEED 0x00000002

class Asura_Cutscene_Event_TextureAnimation : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_TextureAnimation(Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	//event data
	int m_iSequence;
	float m_fSpeed;
	u_int m_uTextureAnimFlags;
};


//////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_Cutscene_Event_TimeScale **
**--------------------------------*/

class Asura_Cutscene_Event_TimeScale : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_TimeScale(Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	virtual void Initialise();
	virtual void Finish();
	void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	//event data
	float m_fTimeScale;

	float m_fPreCutsceneTimeScale;
};

//////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_Cutscene_Event_FieldOfView **
**----------------------------------*/

class Asura_Cutscene_Event_FieldOfView : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_FieldOfView(Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	void Trigger(Asura_Cutscene*);
	virtual void UpdateAfterTrigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	//event data
	float m_fDuration;//time over which the change happens
	float m_fFinalFOV;//final field of view (in degrees)

	//temporary playback stuff
	float m_fInitialFOV;
	bool m_bFinished;
};

//////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------**
** Asura_Cutscene_Event_Trigger **
**------------------------------*/

class Asura_Cutscene_Event_Trigger : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Trigger(Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	virtual void PreDestructionCleanup(Asura_Cutscene*);

	void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	#ifdef ASURA_TOOLS
	virtual void ResetGuidsAndMessages();
	#endif


	
	//event data
	unsigned short m_sMessageBlock;

};

//////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Cutscene_Event_LoopBack **
**-------------------------------*/
class Asura_Cutscene_Event_LoopBack: public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_LoopBack(Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	virtual void GetEventIndexPointers(List<int*>& xIndexList);
	#endif

	void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

    //event data
    float m_fNumSecondsToLoopBack;
	int m_iEventToLoopBackTo;//index of event to go back to (if -1 , just use the time value above)
};

//////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_Cutscene_Event_Attachment **
**---------------------------------*/

class Asura_Cutscene_Event_Attachment : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Attachment(int iEventType,Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	u_int m_uAttachmentID;

};

//////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------------**
** Asura_Cutscene_Event_AttachmentControl **
**----------------------------------------*/



class Asura_Cutscene_Event_AttachmentControl_Attachment
{
public :
	enum Asura_Cutscene_AttachmentControlType
	{
		NONE,
		ACTIVATE,
		DEACTIVATE,
		REACTIVATE,
	};

	u_int m_uID;
	u_int m_uType;
};

class Asura_Cutscene_Event_AttachmentControl : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_AttachmentControl(Asura_Cutscene* pxCutscene);
	~Asura_Cutscene_Event_AttachmentControl();

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* chunk,Asura_Handle fileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	u_int m_uNumAttachments;
	Asura_Cutscene_Event_AttachmentControl_Attachment* m_pxAttachments;

};


//////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Cutscene_Event_Beam **
**---------------------------*/

//putting all the data for the beam event in a seperate structure to make copying it around during
//editing less tedious.
struct Asura_Cutscene_BeamParameters
{
	Asura_Vector_3 m_xInitialPosition;
	Asura_Vector_3 m_xEndPosition;

	float m_fDuration;

	Asura_Beam_Template m_xTemplate;
};

class Asura_Cutscene_Event_Beam : public  Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Beam(Asura_Cutscene* pxCutscene);

	virtual void Trigger(Asura_Cutscene*);
	virtual void UpdateAfterTrigger(Asura_Cutscene*);
	virtual void Finish();

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	virtual void CalculateRelativePosition(const Asura_Matrix_3x3& xCutsceneOrientInverse,const Asura_Vector_3 xCutscenePosition);
	virtual void UpdateRelativeToPosition(const Asura_Matrix_3x3& xCutsceneOrient,const Asura_Vector_3 xCutscenePosition);
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_Cutscene_BeamParameters m_xParam;

	//temporary playback data
	u_int m_uBeamID;
	float m_fLastTimer;

#ifdef ASURA_TOOLS
	//relative position for temporary use while moving the cutscene object
	Asura_Vector_3 m_xRelativeInitialPosition;
	Asura_Vector_3 m_xRelativeEndPosition;
#endif


};

//////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Cutscene_Event_Fade **
**---------------------------*/

#define ASURACUTSCENEFADEFLAG_FADETO 0x00000001		//fade to colour (instead of fade from)
#define ASURACUTSCENEFADEFLAG_CANCELAFTERCUTSCENE 0x00000002 //cancel any fade effect when cutscene finished

class Asura_Cutscene_Event_Fade : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Fade(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);
	virtual void Finish();

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	float m_fRed;
	float m_fGreen;
	float m_fBlue;
	float m_fDuration;
	u_int m_uFlags;
		

};

//////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_Cutscene_Event_TemplateWrapper **
**--------------------------------------*/
// this class exists to provide a handle in Tools for modifying/deleting blocks of events
// that belong to an event template. 
class Asura_Cutscene_Event_TemplateWrapper : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_TemplateWrapper(Asura_Cutscene* pxCutscene)
	:Asura_Cutscene_Event(ACET_TemplateWrapper,pxCutscene)
	{}
};

//////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_Cutscene_Event_Subtitle **
**-------------------------------*/

class Asura_Cutscene_Event_Subtitle : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Subtitle(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_Hash_ID m_uTextHash;
	float m_fDuration;
};

//////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Asura_Cutscene_Event_Sound **
**----------------------------*/

/*
Sound event is currently intented as a tools-only event for testing speech using 
individual wav files (as a replacement for the NextScriptLine functionality).
It may or may not get updated for in-game use later. RAR 25/08/2005
*/
class Asura_Cutscene_Event_Sound : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Sound(Asura_Cutscene* pxCutscene);

	#ifdef ASURA_TOOLS
	virtual void Initialise();
	virtual void Finish();
	virtual void Trigger(Asura_Cutscene*);
	virtual void UpdateAfterTrigger(Asura_Cutscene*);
	#endif

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_String m_xSoundName;

	#ifdef ASURA_TOOLS
	//palayback stuff
	#ifdef ASURA_USE_AUDIO_SYSTEM
		Asura_Hash_ID m_uWave;
	#else
		int m_iSoundID;
	#endif

	Asura_Handle m_iSoundHandle;
	#endif

};

//////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Asura_Cutscene_Event_ChangeCamera **
**-----------------------------------*/

class Asura_Cutscene_Event_ChangeCamera : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_ChangeCamera(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);

	#ifdef ASURA_TOOLS
	//virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	//event data
	//(no extra needed at the moment)
};

//////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------------**
** Asura_Cutscene_Event_ChangeCameraTarget **
**-----------------------------------------*/

class Asura_Cutscene_Event_ChangeCameraTarget : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_ChangeCameraTarget(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_String m_xCameraTargetName;//name of new camera target
};

class Asura_Cutscene_Event_Hide : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Hide(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);
	//event data
	//(no extra needed at the moment)
};

class Asura_Cutscene_Event_Unhide : public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_Unhide(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);
	//event data
	//(no extra needed at the moment)
};

class Asura_Cutscene_Event_ChangeCameraShake : public Asura_Cutscene_Event
{
public:
	Asura_Cutscene_Event_ChangeCameraShake(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);
	void ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_Cutscene_Camera_Shake m_xShake;
	float m_fDuration;
private:
	enum { CHUNK_VERSION = 0 };
};
 
class Asura_Cutscene_Event_SetAlpha : public Asura_Cutscene_Event
{
	typedef Asura_Cutscene_Event PARENT;

public :
	Asura_Cutscene_Event_SetAlpha(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);
	virtual void UpdateAfterTrigger(Asura_Cutscene*);

#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
#endif
	
	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);
	void ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	//event data
	Asura_Graph<float> m_xAlpha;
	float m_fDuration;

private:
	enum { CHUNK_VERSION = 0 };
};

class Asura_Cutscene_Event_ChangePlayerCameraControl : public Asura_Cutscene_Event
{
	typedef Asura_Cutscene_Event PARENT;
public:
	Asura_Cutscene_Event_ChangePlayerCameraControl(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene* pxCutscene);
	virtual void UpdateAfterTrigger(Asura_Cutscene* pxCutscene);

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);
	void ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	float m_fPitchLimit;
	float m_fYawLimit;
	float m_fDuration;

	// temp data
	float m_fInitialPitchLimit;
	float m_fInitialYawLimit;
private:
	enum { CHUNK_VERSION = 0 };
};

#if defined(ASURA_USE_GUI_MENU_SYSTEM)
// Asura_Cutscene_Event_GUIMenu
class Asura_Cutscene_Event_GUIMenu : public Asura_Cutscene_Event
{
	friend class Tools_Cutscene_Event_GUIMenu;

public :

	Asura_Cutscene_Event_GUIMenu(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene* pxCutscene);
	virtual void UpdateAfterTrigger(Asura_Cutscene* pxCutscene);
	virtual void RenderAfterTrigger(Asura_Cutscene* pxCutscene);

#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	virtual int  GetChunkSize();
	virtual void SaveToChunk(char*& pcChunkBuffer);
	virtual void LoadFromChunk(Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);

	// Setters
	void SetMenuHash( Asura_Hash_ID uGUIMenuHash) { m_uGUIMenuHash = uGUIMenuHash; }
	void SetDuration( float fDuration) { m_fDuration = fDuration; }

private :
    
	enum
	{
		CURRENT_CHUNK_VERSION = 0,
	};

	Asura_Hash_ID m_uGUIMenuHash;	    // Menu Hash
	float         m_fDuration;			// Duration of the event
	float         m_fLastUpdateTime;    // Time between update

    typedef Asura_Cutscene_Event PARENT;
};

// Asura_Cutscene_Event_ConsoleVar
class Asura_Cutscene_Event_ConsoleVar : public Asura_Cutscene_Event
{
	friend class Tools_Cutscene_Event_ConsoleVar;

public :

	Asura_Cutscene_Event_ConsoleVar(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene* pxCutscene);
	virtual void Finish();
	virtual void Initialise();

#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	virtual int  GetChunkSize();
	virtual void SaveToChunk(char*& pcChunkBuffer);
	virtual void LoadFromChunk(Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);

	// Setters
	void SetValue(const Asura_String& xValue) { m_xValue = xValue; }
	void SetConsoleVar(const Asura_GUIMenu_ConsoleVar& xConsoleVar) { m_xConsoleVar.Copy(&xConsoleVar); }

private :

    enum
	{
		CURRENT_CHUNK_VERSION = 0,
	};

	// Types for console variable
	enum EValueType
	{
        CONSOLE_VAR_INVALID = -1,
		CONSOLE_VAR_BOOL,
		CONSOLE_VAR_INT,
		CONSOLE_VAR_FLOAT,
	};

	Asura_String			 m_xValue;          // console var value
	Asura_GUIMenu_ConsoleVar m_xConsoleVar;     // console var
    Asura_Value				 m_xOriginalValue;  // original console var value
    EValueType				 m_eValueType;      // value type

    typedef Asura_Cutscene_Event PARENT;
};
#endif	//defined(ASURA_USE_GUI_MENU_SYSTEM)

// Asura_Cutscene_Event_HDRSettings
#ifdef ASURA_USE_HDR
class Asura_Cutscene_Event_HDRSettings : public Asura_Cutscene_Event
{
	friend class Tools_Cutscene_Event_HDRSettings;

public :

	Asura_Cutscene_Event_HDRSettings(Asura_Cutscene* pxCutscene);
	virtual void Finish();
	virtual void Trigger(Asura_Cutscene* pxCutscene);
	virtual void UpdateAfterTrigger(Asura_Cutscene* pxCutscene);

#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	virtual int  GetChunkSize();
	virtual void SaveToChunk(char*& pcChunkBuffer);
	virtual void LoadFromChunk(Asura_Chunk_CutsceneEvent* pxChunk, Asura_Handle iFileHandle);
	virtual void ReadFromChunkStream(class Asura_Chunk_CutsceneEvent* pxChunk, Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	// Setters
	void SetHDRSettings( Asura_CameraData::HDRSettings& xHDRSettings) { s_xHDRSettings = xHDRSettings; }
	void SetDuration( float fDuration) { m_fDuration = fDuration; }

#ifdef ASURA_TOOLS
	static bool IsUsingFixedExposure() { return !s_xHDRSettings.m_bDynamicExposure; }
	static bool IsUsingDynamicExposure() { return s_xHDRSettings.m_bDynamicExposure; }
#endif

private :
    
	enum
	{
		CURRENT_CHUNK_VERSION = 1,
	};

	float  m_fDuration;
	static Asura_CameraData::HDRSettings s_xHDRSettings;
	static Asura_CameraData::HDRSettings s_xOriginalHDRSettings;
	static bool s_bHDRSet;



    typedef Asura_Cutscene_Event PARENT;
};
#endif	//defined ASURA_USE_HDR

#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
#define ASURA_CUTSCENE_EVENT_FACIALANIMATION_LOOPED 0x00000001

class Asura_Cutscene_Event_FacialAnimation: public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_FacialAnimation(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	virtual void LoadFiles();
	virtual void AddExtraFilesToList( List<Asura_String>& xNameList ) const;
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_Hash_ID m_uAnimID;
	u_int m_uAnimationFlags;
};
#endif //ASURA_USE_FACIAL_ANIMATION_SYSTEM

class Asura_Cutscene_Event_MixerPreset: public Asura_Cutscene_Event
{
public :
	Asura_Cutscene_Event_MixerPreset(Asura_Cutscene* pxCutscene);
	virtual void Trigger(Asura_Cutscene*);

	#ifdef ASURA_TOOLS
	virtual void Copy(Asura_Cutscene_Event* pxEvent);
	#endif

	//Load/save functions , defined in Asura_Chunk_Cutscene.cpp
	int GetChunkSize();
	void SaveToChunk(char*& chunkBuffer);
	void LoadFromChunk(class Asura_Chunk_CutsceneEvent* pxChunk,Asura_Handle iFileHandle);
//	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	//event data
	Asura_Hash_ID m_uMixerPreset;
};



#endif // _ASURA_CUTSCENE_EVENT_H_
