//////////////////////////////////////////////////////////////////////////////
//
// Asura_CPUProfile_Renderer.cpp
//
// Created by: Mike Healey 15/08/2005
// Extended by: Duncan Hopkins 22/05/2007
//
// Copyright (c) 2004-2005 Rebellion. All rights reserved.
//
// Basic profiling bar renderer so new platforms can get up and running 
// quickly with a profiling system that isn't as efficient as it could be, 
// but does the job reasonably well. It is very much based on Anthony 
// Gregson's work for PSP, which never became particularly PSP-specific.
//
// I have added code to time the total frame length, so in platform code, 
// before calling the render function stop the total frame timer, and start it
// again after the VSync.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined( ASURA_USE_PROFILING_SYSTEM)

#include "Debugging/Asura_CPUProfile_Renderer.h"
#include "Debugging/Asura_CPUProfile_Multithreaded_Render.h"

#include "Core/Asura_Atomic.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#ifdef ASURA_USE_DEBUG_TEXT
#include "Debugging/Asura_DebugText.h"
#endif
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

#include "../../../Project/Tools/Misc/Tools_CyclicBuffer.h"

#ifdef ASURA_ASSERT
#include <string.h>
#endif

// G L O B A L S /////////////////////////////////////////////////////////////
//static Asura_TickCount g_auTimerStartTicks[ASURA_CPU_PROFILE__MAX_TIMERS] = {0};
//static Asura_TickCount g_auTimerTicks[ASURA_CPU_PROFILE__MAX_TIMERS] = {0};

// colours should match those in Asura_CPUProfile.h, i.e. 
// grey, blue, green, red, cyan, yellow, magenta, black (1st row)
// white, blue, green, red, cyan, yellow, magenta, black (other rows)
static const u_int g_auProfileColours[ASURA_CPU_PROFILE__MAX_TIMERS] =
{
	0xff604020, 0xff2222ff, 0xff22ff22, 0xffff2222, 0xff22ffff, 0xffffff22, 0xffff22ff, 0xff222222,
	0xffffffff, 0xff5555dd, 0xff55dd55, 0xffdd5555, 0xff55dddd, 0xffdddd55, 0xffdd55dd, 0xff555555,
	0xffdddddd, 0xff2222ff, 0xff22ff22, 0xffff2222, 0xff22ffff, 0xffffff22, 0xffff22ff, 0xff222222,
	0xffffffff, 0xff5555dd, 0xff55dd55, 0xffdd5555, 0xff55dddd, 0xffdddd55, 0xffdd55dd, 0xff555555,
	0xff303030, 0xff000050, 0xff500000, 0xff005000, 0xff882288, 0xff552255
};

//static const char* g_aszProfileNames[ASURA_CPU_PROFILE__MAX_TIMERS];

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////

#if defined( ASURA_TOOLS)
bool Asura_CPUProfile_Renderer::s_bBarsEnabled = false;
#else
bool Asura_CPUProfile_Renderer::s_bBarsEnabled = true;
#endif

#if ASURA_USE_DEBUG_TEXT
bool Asura_CPUProfile_Renderer::s_bTextEnabled = false;
#endif

bool Asura_CPUProfile_Renderer::s_bPaused;

static bool g_bShowOnlyPrimaryThread = false;
static volatile bool g_bLockListAccess = false;
static volatile u_int g_uListAccessCount = 0;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_CPUProfile_Renderer::Initialise()
{
	Asura_CPUProfile_Renderer_ThreadData::Platform_Initialise();
	
	ASURA_CPU_PROFILE__START_THREAD( "Main");
	
// JAT 27/02/07 add the variable to command console & DVS for easy toggle of the bars & text on/off
#if defined( ASURA_SUPPORT_COMMAND_CONSOLE )
	Asura_CommandConsole::AddVar( L"Debug.Profile.CPU.Bars", &s_bBarsEnabled, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Debug.Profile.CPU.OnlyPrimaryThread", &g_bShowOnlyPrimaryThread, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#if ASURA_USE_DEBUG_TEXT
	Asura_CommandConsole::AddVar( L"Debug.Profile.CPU.Text", &s_bTextEnabled, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif // #if ASURA_USE_DEBUG_TEXT
#endif // #if defined( ASURA_USE_COMMANDCONSOLE )


}

void Asura_CPUProfile_Renderer::Shutdown()
{
	ASURA_CPU_PROFILE__STOP_THREAD();
	
	Asura_CPUProfile_Renderer_ThreadData::Platform_Shutdown();
}

void Asura_CPUProfile_Renderer::StartProfile(int iProfileType)
{
	if (s_bPaused)
	{
		return;
	}

	while( true == g_bLockListAccess) {}
	
	Asura_Atomic::Increment( g_uListAccessCount);
	
	Asura_CPUProfile_Renderer_ThreadData* pxThreadData = Asura_CPUProfile_Renderer_ThreadData::Platform_GetThreadData();
	if( NULL == pxThreadData)
	{
		Asura_Atomic::Decrement( g_uListAccessCount);
		return;
	}
	
//	Asura_CPUProfile_Renderer_ThreadData::Platform_GetAndLockThreadList();

	Asura_TickCount uStartTicks = Asura_Timers::GetCurTicks();
	
	Asura_Assert(iProfileType >= 0 && iProfileType < ASURA_CPU_PROFILE__MAX_TIMERS, "bad timer index");
	Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ iProfileType];
	if (iProfileType)
	{
		pxThreadData->m_uInProfileDepth += 1;
		if( 1 == pxThreadData->m_uInProfileDepth)
		{
			pxThreadData->m_uInProfile = iProfileType;
		}
		else
		{
			// DLH - 24 May 2007 10:46:26 - If already profiling then stack if the same type. Else ignor.
			Asura_Assertf( static_cast< u_int>( iProfileType) == pxThreadData->m_uInProfile, ("profile '%s' depth is already %d. already profiling type %d. Trying to profile %d.", xProfile.m_pszProfileName, pxThreadData->m_uInProfileDepth, pxThreadData->m_uInProfile, iProfileType));
			
//			Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
			Asura_Atomic::Decrement( g_uListAccessCount);
			return;
		}
	}

	Asura_Assertf( 0 == xProfile.m_uInProfileDepth, ("profile '%s' depth is already %d. already profiling type %d. Trying to profile %d.", xProfile.m_pszProfileName, xProfile.m_uInProfileDepth, pxThreadData->m_uInProfile, iProfileType));
	xProfile.m_uTimerStartTicks = uStartTicks;
	xProfile.m_uInProfileDepth += 1;
	
//	Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
	Asura_Atomic::Decrement( g_uListAccessCount);
}

void Asura_CPUProfile_Renderer::StopProfile(int iProfileType)
{
	if (s_bPaused)
	{
		return;
	}

	while( true == g_bLockListAccess) {}
	
	Asura_Atomic::Increment( g_uListAccessCount);
	
	Asura_CPUProfile_Renderer_ThreadData* pxThreadData = Asura_CPUProfile_Renderer_ThreadData::Platform_GetThreadData();
	if( NULL == pxThreadData)
	{
		Asura_Atomic::Decrement( g_uListAccessCount);
		return;
	}
	
//	Asura_CPUProfile_Renderer_ThreadData::Platform_GetAndLockThreadList();

	Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ iProfileType];
	
	Asura_TickCount uEndTicks = Asura_Timers::GetCurTicks();
	
	Asura_Assert(iProfileType >= 0 && iProfileType < ASURA_CPU_PROFILE__MAX_TIMERS, "bad timer index");
	if (iProfileType)
	{
		Asura_Assertf( 0 < pxThreadData->m_uInProfileDepth, ("profiling of type %d not started so cannot leave.", iProfileType));
		if(1 < pxThreadData->m_uInProfileDepth)
		{
			pxThreadData->m_uInProfileDepth -= 1;
			
			// DLH - 24 May 2007 10:53:54 - Already profiling so check for wrong profile type
			Asura_Assertf( static_cast< u_int>( iProfileType) == pxThreadData->m_uInProfile, ("profile '%s' depth is already %d. already profiling type %d. Trying to profile %d.", xProfile.m_pszProfileName, pxThreadData->m_uInProfileDepth, pxThreadData->m_uInProfile, iProfileType));

//			Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
			Asura_Atomic::Decrement( g_uListAccessCount);
			return;
		}
		
		pxThreadData->m_uInProfile = 0;
		pxThreadData->m_uInProfileDepth = 0;
	}

	xProfile.m_uTimerEndTicks = uEndTicks;
	
	Asura_Assertf( 1 == xProfile.m_uInProfileDepth, ("profile '%s' depth is already %d. already profiling type %d. Trying to profile %d.", xProfile.m_pszProfileName, xProfile.m_uInProfileDepth, pxThreadData->m_uInProfile, iProfileType));

	Asura_TickCount uElapsedTicks = uEndTicks - xProfile.m_uTimerStartTicks;
	xProfile.m_uTimerTicks += uElapsedTicks;
	xProfile.m_uInProfileDepth -= 1;
	
//	Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
	Asura_Atomic::Decrement( g_uListAccessCount);
}

void Asura_CPUProfile_Renderer::SetName(int iProfileType, const char* szName)
{
	Asura_CPUProfile_Renderer_ThreadData* pxThreadData = Asura_CPUProfile_Renderer_ThreadData::Platform_GetThreadData();
	if( NULL == pxThreadData)
	{
		return;
	}
	
	Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ iProfileType];
	
#ifdef ASURA_ASSERT
	Asura_Assert(iProfileType >= 0 && iProfileType < ASURA_CPU_PROFILE__MAX_TIMERS, "bad timer index");
	{
		const char* paszProfileName = xProfile.m_pszProfileName;
		if ( paszProfileName)
		{
			Asura_Assertf(!strcmp( paszProfileName, szName), ("same timer, different name. From '%s' to '%s'", paszProfileName, szName));
		}
	}
#endif
	xProfile.m_pszProfileName = szName;
}

static const float fASURA_PROF_TIMING_BAR_SECTION = 1.0f/60.0f;
static const float fASURA_PROF_ONE_OVER_TIMING_BAR_SECTION = 60.0f;
static const int iASURA_PROF_TIMING_MAX_DIVISIONS = 12;	// i.e. 5fps

DEBUGVAR float dbg_fProfileBarsHBorder = 0.004f;
// TODO: max divisions
DEBUGVAR float dbg_fFrameSmoothingTime = 2.0f;		// time to keep rendering more divisions than required.
DEBUGVAR u_int dbg_uProfBackColour0 = 0xff333333;	// colours for alternating background divisions.
DEBUGVAR u_int dbg_uProfBackColour1 = 0xff666666;

DEBUGVAR float dbg_fProfileBarX = 0.1f;
DEBUGVAR float dbg_fProfileBarY = 0.05f;
DEBUGVAR float dbg_fProfileBarW = 0.8f;

DEBUGVAR u_int dbg_uProfTextX = 8;

static void RenderSection(float fTimeStart, float fTimeWidth, float fY, float fH, u_int uColour)
{
	float fX = (dbg_fProfileBarX + (dbg_fProfileBarW * fTimeStart)) * Asura_Viewport::GetWidth();
	float fW = (dbg_fProfileBarW * fTimeWidth) * Asura_Viewport::GetWidth();
	Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
	Asura_Render::Overlay.Colour = uColour;
	Asura_Render::Overlay.Render();
}

void Asura_CPUProfile_Renderer::Render(bool bIncludeText)
{
	if (s_bPaused)
	{
		return;
	}

	if (
		false == s_bBarsEnabled
#if defined( ASURA_USE_DEBUG_TEXT)
	&&
		false == s_bTextEnabled
#endif
	)
	{
		return;
	}

	Asura_Render::FlushDisplay();

	g_bLockListAccess = true;
	while( 0 < g_uListAccessCount) {}

#if defined ASURA_USE_DEBUG_TEXT
	float fStart = Asura_Timers::GetTimeInFrame();
#endif

	Asura_LinkList< Asura_CPUProfile_Renderer_ThreadData*>& xList_pxThreadData = Asura_CPUProfile_Renderer_ThreadData::Platform_GetAndLockThreadList();
	Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList0 = xList_pxThreadData.Front();
	Asura_CPUProfile_Renderer_ThreadData* pxThreadData0 = NULL;
	if( NULL != pxIteratorList0)
	{
		pxThreadData0 = pxIteratorList0->Data();
	}
	if( NULL == pxThreadData0)
	{
		Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
		g_bLockListAccess = false;
		return;
	}
//	const Asura_CPUProfile_Renderer_ThreadData::Profile& xThread0ProfileTotalTime = pxThreadData0->m_axProfiles[ ASURA_CPU_PROFILE__TOTALFRAME];
	
//	Asura_TickCount uTicksStart = xThread0ProfileTotalTime.m_uTimerStartTicks;
//	Asura_TickCount uTicksEnd = xThread0ProfileTotalTime.m_uTimerEndTicks;
	const Asura_TickCount uTicksStart = pxThreadData0->m_uTimerCleared;
	const Asura_TickCount uTicksEnd = Asura_Timers::GetCurTicks();
	pxThreadData0->m_uTimerCleared = uTicksEnd;
	const Asura_TickCount uTicksDelta = uTicksEnd - uTicksStart;

	float fTotalTime = Asura_Timers::TicksToSeconds( uTicksDelta);
	Asura_Assert( fTotalTime, "didn't set total frame?");

	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;

	float fBarOffsetY = (dbg_fProfileBarY * Asura_Viewport::GetHeight());

	u_int uThreadsToShow = 1;
	if( false == g_bShowOnlyPrimaryThread)
	{
		uThreadsToShow = xList_pxThreadData.Size();
	}
	
	{
		// DLH - 23 May 2007 16:35:28 - Clamp any profiles that are presently profiling
		for( Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList = xList_pxThreadData.Front(); NULL != pxIteratorList; pxIteratorList = pxIteratorList->Next())
		{
			Asura_CPUProfile_Renderer_ThreadData* pxThreadData = pxIteratorList->Data();
			
			for (int i = 0; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
			{
				Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ i];
				Asura_TickCount uTimerTicks = xProfile.m_uTimerTicks;
				if( 0 < xProfile.m_uInProfileDepth)
				{
					xProfile.m_uTimerEndTicks = uTicksEnd;
					if( (uTicksEnd - uTicksStart) < (uTicksEnd - xProfile.m_uTimerStartTicks))
					{
						xProfile.m_uTimerStartTicks = uTicksStart;
					}
					uTimerTicks = uTicksEnd - xProfile.m_uTimerStartTicks;
					xProfile.m_uTimerTicks = uTimerTicks;
				}
			}
		}
	}
	
	if (s_bBarsEnabled)
	{
		int iDivisions = Asura_Maths::FloatToInt_Chop(fTotalTime / fASURA_PROF_TIMING_BAR_SECTION) + 1;
		static int ls_iLastDivisions = iDivisions;
		static float ls_fClampTime = 0.0f;
		// if we're going to draw fewer divisions than we did last time use the previous value for a few frames
		if (iDivisions < ls_iLastDivisions)
		{
			ls_fClampTime += Asura_Timers::GetFrameTime();
			if (ls_fClampTime < dbg_fFrameSmoothingTime)
			{
				iDivisions = ls_iLastDivisions;
			}
		}
		else
		{
			ls_fClampTime = 0.0f;
		}
		// don't draw more than max (12) divisions.
		iDivisions = Asura_Maths::Min(iDivisions, iASURA_PROF_TIMING_MAX_DIVISIONS);
		ls_iLastDivisions = iDivisions;

		const float fOneOverDivisions = 1.0f / static_cast<float>(iDivisions);
		const float fDivisionW = dbg_fProfileBarW * fOneOverDivisions * Asura_Viewport::GetWidth();
		const float fDivisionH = 0.9f * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
		const float dbg_fProfileBarsHBorder = 0.1f;
		
		u_int uThreadDataIndex = 0;
		for( Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList = xList_pxThreadData.Front(); NULL != pxIteratorList; pxIteratorList = pxIteratorList->Next(), uThreadDataIndex += 1)
		{
			if( uThreadsToShow <= uThreadDataIndex)
			{
				continue;
			}

			Asura_CPUProfile_Renderer_ThreadData* pxThreadData = pxIteratorList->Data();
			
			float fRunningThreadTotal = 0.0f;
			float fRunningThreadTotalTime = 0.0f;

			const float fDivisionXBase = dbg_fProfileBarX * Asura_Viewport::GetWidth();
			const float fDivisionY = fBarOffsetY;
			
			{
				const char* pszThreadName = pxThreadData->m_szThreadName;
				u_int uPosX = static_cast<u_int>( fDivisionXBase / ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
				u_int uPosY = static_cast<u_int>( (fDivisionY + (0.5f * fDivisionH)) / ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
				Asura_DebugText::AddTextBlock( uPosX, uPosY, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_RIGHTJUSTIFIED, 
							"%s", pszThreadName);
			}
			
			// render background divisions (alternating colours for handy scale)
			{
				for (int i=0; i<iDivisions; i++)
				{
					const float fX = fDivisionXBase + (fDivisionW * static_cast<float>(i));
					Asura_Render::Overlay.Colour = (i & 1) ? dbg_uProfBackColour0 : dbg_uProfBackColour1;
					Asura_Render::Overlay.SetRectXY(fX, fDivisionY, fDivisionW, fDivisionH);
  					Asura_Render::Overlay.Render();
				}
			}

			float fY = fBarOffsetY + (dbg_fProfileBarsHBorder * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
			float fH = fDivisionH * ( 1.0f - (2.0f * dbg_fProfileBarsHBorder));

#if 0			
			{
				// assuming that timer 0 is used for total frame
				for (int i = 1; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
				{
					Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ i];
					Asura_TickCount uTimerTicks = xProfile.m_uTimerTicks;
					if( uTimerTicks)
					{
						float fTimeStart = Asura_Timers::TicksToSeconds( xProfile.m_uTimerStartTicks - uTicksStart);
						float fTimeEnd = Asura_Timers::TicksToSeconds( xProfile.m_uTimerEndTicks - uTicksStart);
						float fValueStart = fTimeStart * fASURA_PROF_ONE_OVER_TIMING_BAR_SECTION * fOneOverDivisions;
						float fValueEnd = fTimeEnd * fASURA_PROF_ONE_OVER_TIMING_BAR_SECTION * fOneOverDivisions;
						RenderSection( fValueStart, fValueEnd - fValueStart, fY + (0.6f * fH), 0.4f * fH, g_auProfileColours[i]);
					}
				}
			}
#endif
			
			// assuming that timer 0 is used for total frame
			for (int i = 1; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
			{
				Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ i];
				Asura_TickCount uTimerTicks = xProfile.m_uTimerTicks;
				if( uTimerTicks)
				{
					float fTime = Asura_Timers::TicksToSeconds( uTimerTicks);
					float fValue = fTime * fASURA_PROF_ONE_OVER_TIMING_BAR_SECTION * fOneOverDivisions;
					//RenderSection( fRunningThreadTotal, fValue, fY, 0.4f * fH, g_auProfileColours[i]);
					RenderSection( fRunningThreadTotal, fValue, fY, fH, g_auProfileColours[i]);
					fRunningThreadTotal += fValue;
					fRunningThreadTotalTime += fTime;
				}
			}
			// and the remainder for timer 0
			{
				float fTime = fTotalTime - fRunningThreadTotalTime;
				// clamp at 0 to account for floating point addition inaccuracy
				fTime = Asura_Maths::Max( fTime, 0.0f );
				const float fValue = fTime * fASURA_PROF_ONE_OVER_TIMING_BAR_SECTION * fOneOverDivisions;
				RenderSection( fRunningThreadTotal, fValue, fY, fH, g_auProfileColours[0]);
			}
			
			fBarOffsetY += ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
		}
		
		fBarOffsetY += ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
	}


#if 1 && defined( ASURA_USE_DEBUG_TEXT)
	if (s_bTextEnabled)
	{
		u_int uTextLineIndex = static_cast< u_int>( fBarOffsetY / ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
		
		{
			const u_int uTotalTime = Asura_Maths::FloatToInt( fTotalTime * 1000000.0f);
			
			static Tools_CyclicBuffer< u_int> g_xAverageTotalTime( 60);

			g_xAverageTotalTime.ValueAdd( uTotalTime);
			u_int uAverageTotalTime = g_xAverageTotalTime.GetAverage();
			Asura_DebugText::AddTextBlock( dbg_uProfTextX, uTextLineIndex, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
						"Total %6d (%6d) us", uTotalTime, uAverageTotalTime);
						
			uTextLineIndex += 1;
		}
		
		
		u_int uThreadDataIndex = 0;
		for( Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList = xList_pxThreadData.Front(); NULL != pxIteratorList; pxIteratorList = pxIteratorList->Next(), uThreadDataIndex += 1)
		{
			if( uThreadsToShow <= uThreadDataIndex)
			{
				continue;
			}

			Asura_CPUProfile_Renderer_ThreadData* pxThreadData = pxIteratorList->Data();
			
			float fRunningThreadTotalTime = 0.0f;
			
			float fOffsetX = 0;
			float fOffsetY = 0;
			
			{
				const char* pszThreadName = pxThreadData->m_szThreadName;
				Asura_DebugText::AddTextBlock( dbg_uProfTextX, uTextLineIndex, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
							"%s", pszThreadName);
							
				uTextLineIndex += 1;
			}
			
			u_int uNumRendered = 0;
			for (int i = 1; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
			{
				const Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ i];
				const Asura_TickCount xTimerTicks = xProfile.m_uTimerTicks;
				if ( xTimerTicks)
				{
					float fTime = Asura_Timers::TicksToSeconds( xTimerTicks);
					if (fTime > 0.0f)
					{
						fRunningThreadTotalTime += fTime;
						
						Asura_Assertf(fTime <= fTotalTime, ("timed event (%f) took longer than entire frame! (%f)", fTime, fTotalTime));
						const char* pszName = xProfile.m_pszProfileName;
						if( NULL == pszName)
						{
							pszName = "???";
						}
						Asura_DebugText::AddTextBlock( dbg_uProfTextX, uTextLineIndex, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
							"%2d %6.02f%% %s: (%3d us)", i, 100.0f * fTime / fTotalTime, pszName, Asura_Maths::FloatToInt(fTime * 1000000.0f));

						// render a colour coded square
						float fX = fOffsetX + (static_cast<float>(dbg_uProfTextX - 1) * ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
						float fY = fOffsetY + (static_cast<float>(uTextLineIndex) * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
						float fW = ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH - 2.0f;
						float fH = ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT - 2.0f;
						Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
						Asura_Render::Overlay.Colour = g_auProfileColours[i];
						Asura_Render::Overlay.Render();
						
						++uNumRendered;
						uTextLineIndex += 1;
					}
				}
			}

			{
				// render a colour coded square for the untimed section
				float fX = fOffsetX + (static_cast<float>(dbg_uProfTextX - 1) * ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
				float fY = fOffsetY + (static_cast<float>(uTextLineIndex) * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
				float fW = ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH - 2.0f;
				float fH = ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT - 2.0f;
				Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
				Asura_Render::Overlay.Colour = g_auProfileColours[0];
				Asura_Render::Overlay.Render();
				// and the text...
				const float fTime = fTotalTime - fRunningThreadTotalTime;
				Asura_DebugText::AddTextBlock( dbg_uProfTextX, uTextLineIndex, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
					"   %6.02f%% unprofiled: (%3d us)", 100.0f * fTime / fTotalTime, Asura_Maths::FloatToInt(fTime * 1000000.0f));

				uTextLineIndex += 1;

			}
			
			uTextLineIndex += 1;
		}
		
		{
			Asura_DebugText::AddTextBlock( dbg_uProfTextX, uTextLineIndex, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
				"rendering this lot: (%d us)", Asura_Maths::FloatToInt((Asura_Timers::GetTimeInFrame() - fStart) * 1000000.0f));
				
			uTextLineIndex += 1;
		}
	}
#endif // #ifdef ASURA_USE_DEBUG_TEXT

	g_bLockListAccess = false;
	Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();

	Asura_Render::FlushDisplay();
}

void Asura_CPUProfile_Renderer::ClearValues()
{
	g_bLockListAccess = true;
	while( 0 < g_uListAccessCount) {}

	Asura_LinkList< Asura_CPUProfile_Renderer_ThreadData*>& xList_pxThreadData = Asura_CPUProfile_Renderer_ThreadData::Platform_GetAndLockThreadList();
	
	for( Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList = xList_pxThreadData.Front(); NULL != pxIteratorList; pxIteratorList = pxIteratorList->Next())
	{
		Asura_CPUProfile_Renderer_ThreadData* pxThreadData = pxIteratorList->Data();
		for (int i=0; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
		{
			Asura_CPUProfile_Renderer_ThreadData::Profile& xProfile = pxThreadData->m_axProfiles[ i];
			
			xProfile.m_uTimerTicks = 0;
		}
	}
	
	Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
	
	g_bLockListAccess = false;
}

void Asura_CPUProfile_Renderer::SetPaused( bool bPaused )
{
	Asura_Assert( bPaused != s_bPaused, "unmatched pause/unpause calls?");
	s_bPaused = bPaused;
}

void Asura_CPUProfile_Renderer::Reset()
{
	// reset all the names to avoid asserts.

	Asura_LinkList< Asura_CPUProfile_Renderer_ThreadData*>& xList_pxThreadData = Asura_CPUProfile_Renderer_ThreadData::Platform_GetAndLockThreadList();
	
	for( Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList = xList_pxThreadData.Front(); NULL != pxIteratorList; pxIteratorList = pxIteratorList->Next())
	{
		Asura_CPUProfile_Renderer_ThreadData* pxThreadData = pxIteratorList->Data();
		for (int i=0; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
		{
			pxThreadData->m_axProfiles[ i].m_pszProfileName = NULL;;
		}
	}

	Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList();
}

Asura_CPUProfile_Renderer_ThreadData::Asura_CPUProfile_Renderer_ThreadData()
{
	this->m_szThreadName[0] = '\0';
	this->m_uTimerCleared = Asura_Timers::GetCurTicks();
	
	this->m_uInProfileDepth = 0;
	this->m_uInProfile = 0;
	
	for (int i=0; i<ASURA_CPU_PROFILE__MAX_TIMERS; i++)
	{
		Profile& xProfile = this->m_axProfiles[ i];
		
		xProfile.m_uTimerStartTicks = 0;
		xProfile.m_uTimerTicks = 0;
		xProfile.m_pszProfileName = NULL;
		xProfile.m_uInProfileDepth = 0;
	}
}

#if !defined( ASURA_USE_PROFILING_SYSTEM_MULTITHREADED)
static Asura_CPUProfile_Renderer_ThreadData g_xProfileRenderThreadData;
static Asura_LinkList< Asura_CPUProfile_Renderer_ThreadData*> g_xList_pxProfileRenderThreadData;

void Asura_CPUProfile_Renderer_ThreadData::Platform_Initialise()
{
	g_xList_pxProfileRenderThreadData.Push_Back( &g_xProfileRenderThreadData);
}

void Asura_CPUProfile_Renderer_ThreadData::Platform_Shutdown()
{
	// DLH - 22 May 2007 14:54:14 - Clear the list of thread data
	for( Asura_ListElt< Asura_CPUProfile_Renderer_ThreadData*>* pxIteratorList = g_xList_pxProfileRenderThreadData.Front(); NULL != pxIteratorList; pxIteratorList = g_xList_pxProfileRenderThreadData.Front())
	{
		g_xList_pxProfileRenderThreadData.Remove( pxIteratorList);
	}
}

Asura_CPUProfile_Renderer_ThreadData* Asura_CPUProfile_Renderer_ThreadData::Platform_GetThreadData()
{
	Asura_CPUProfile_Renderer_ThreadData* pxThreadData = &g_xProfileRenderThreadData;
	
	return( pxThreadData);
}

Asura_LinkList< Asura_CPUProfile_Renderer_ThreadData*>& Asura_CPUProfile_Renderer_ThreadData::Platform_GetAndLockThreadList()
{
	return( g_xList_pxProfileRenderThreadData);
}

void Asura_CPUProfile_Renderer_ThreadData::Platform_ReturnAndUnlockThreadList()
{
}

#endif // defined( ASURA_USE_PROFILING_SYSTEM_MULTITHREADED)

#endif //#ifdef ASURA_USE_PROFILING_SYSTEM
