//////////////////////////////////////////////////////////////////////////////
//
// Asura_CPUProfile_Renderer.h
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     Basic profiling bar renderer so new platforms can get up and running
//     quickly with a profiling system that isn't as efficient as it could
//     be, but does the job reasonably well. It is entirely based on Anthony
//     Gregson's work for PSP, which never became particularly PSP-specific.
//
// Author:
//
//     Mike Healey, with credit to Anthony Gregson.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_CPUPROFILE_RENDERER_H___
#define ___ASURA_CPUPROFILE_RENDERER_H___ 1

class Asura_CPUProfile_Renderer
{
public:
	static void Initialise();
	static void Shutdown();
	static void Reset(); // reset clears all the name assignments.

#if defined( ASURA_USE_PROFILING_SYSTEM_MULTITHREADED)
	static void StartThread( const char* pszName) { Platform_StartThread( pszName); }
	static void StopThread( void) { Platform_StopThread(); }
#endif

	static void StartProfile(int iProfileType);
	static void StopProfile(int iProfileType);
	static void SetName(int iProfileType, const char* szName);

	static void Render(bool bIncludeText);
	static void ClearValues();

	static void EnableBars( bool bEnable ) { s_bBarsEnabled = bEnable; }
	static bool IsBarsEnabled( void ) { return s_bBarsEnabled; }
#ifdef ASURA_USE_DEBUG_TEXT
	static void EnableText( bool bEnable ) { s_bTextEnabled = bEnable; }
	static bool IsTextEnabled( void ) { return s_bTextEnabled; }
#endif
	static void SetPaused( bool bPaused );// to help avoid nested profiling, e.g. tools modal overlay.
	static bool IsPaused( void ) { return s_bPaused; }
    static float GetYOffset( void ) { return s_fScaledYOffset / fY_OFFSET_SCALE; }

private:
#if defined( ASURA_USE_PROFILING_SYSTEM_MULTITHREADED)
	static void Platform_StartThread( const char* pszName);
	static void Platform_StopThread( void);
#endif

	static bool s_bBarsEnabled;
#ifdef ASURA_USE_DEBUG_TEXT
	static bool s_bFPSEnabled;
	static bool s_bTextEnabled;
#endif
	static bool s_bPaused;	// to help avoid nested profiling, e.g. tools modal overlay.

    static const float fY_OFFSET_SCALE; // So that default DVS increment / decrement is some use.
    static float s_fScaledYOffset;

};


#ifdef ASURA_USE_PROFILING_SYSTEM

//Macro for disabling profiling in a function , to avoid asserts cause by nested profiles
//RAR 26/10/2007
#define ASURA_CPUPROFILE_SCOPED_PAUSE Asura_CPUProfile_Scoped_Pause xScopedPauseProfile##__LINE__;

class Asura_CPUProfile_Scoped_Pause
{
public :
	Asura_CPUProfile_Scoped_Pause()
	{
		//Keep track of whether the system is already paused , since there are a few cases
		//where nested pauses take place.
		m_bWasPaused = Asura_CPUProfile_Renderer::IsPaused();
		if(!m_bWasPaused) Asura_CPUProfile_Renderer::SetPaused(true);
	}
	~Asura_CPUProfile_Scoped_Pause()
	{
		if(!m_bWasPaused) Asura_CPUProfile_Renderer::SetPaused(false);
	}
private :
	bool m_bWasPaused;
};

#else //!defined(ASURA_USE_PROFILING_SYSTEM)
#define ASURA_CPUPROFILE_SCOPED_PAUSE
#endif //!defined(ASURA_USE_PROFILING_SYSTEM)


// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_CPUPROFILE_RENDERER_H___
