//
// Asura_Entity.cpp
//
// Created by Kevin Lea - 19/06/2001 10:05:15
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Entity/Asura_Entity.h"

#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_Entity_PeriodicClientUpdateList.h"
#include "Entity/Asura_Entity_System.h"
#include "Entity/Asura_Entity_Event.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "LoadAndSave/Asura_LoadAndSave.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"

// KOR - this must come after the Asura_Network.h include
// since that is where ASURA_NETWORK_LOGGING is defined
#if ASURA_NETWORK_LOGGING
#include "Core/Asura_String.h"
#endif

#ifdef ASURA_SUPPORT_NETMSG_SYSTEM
#include "Network/Asura_NetMsg.h"
#endif

#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#include "Streaming/Asura_Streaming_ZoneEntities.h"
#endif

#ifdef ASURA_TOOLS
#include "Entity/Tools_Entity_Data.h"
#include "Entity/Tools_Entity_Object.h"
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#include "Lighting/Tools_LightSets.h"
#include "Mesh/Tools_PlatformMesh_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity
//
//////////////////////////////////////////////////////////////////////////////

Asura_Entity::Asura_Entity()
{
	Classification			= AsuraEntityClass_NotSet;
	m_bReadyToBeDestroyed	= false;
	m_bEntityActive			= true;
	m_bTemplateInstanceActive = true;
	m_bActiveStatusInitialised = false;
	m_bHasBeenShutdown = false;
	Guid					= ASURA_GUID_UNREGISTERED;

    m_uBlueprintHashID     = ASURA_HASH_ID_UNSET;
    m_uBlueprintTypeHashID = ASURA_HASH_ID_UNSET;

#ifdef ASURA_TOOLS
	ToolsData = 0;
#endif

#if defined(ASURA_USE_STREAMING)
	m_uZoneHashID	= ASURA_HASH_ID_UNSET;
	m_bAllowZoneChange = false;
#endif
}

Asura_Entity::~Asura_Entity()
{
#ifdef ASURA_TOOLS
	if (ToolsData)
	{
		if (ToolsData->ToolsObject)
			ToolsData->ToolsObject->ClearEntity();

		delete ToolsData;
		ToolsData = NULL;
	}
#endif

    // Remove from to-be-destroyed list if necessary.
	Asura_Entity_System::RemoveFromDestroyList(this);
}

void Asura_Entity::SetAsDestroyed(void)
{
	if(!m_bReadyToBeDestroyed)
	{
		m_bReadyToBeDestroyed = true;
		Asura_Entity_System::AddToDestroyList( this );
	}
}

void Asura_Entity::PreDestructionShutdown()
{
	Asura_Assertf( !m_bHasBeenShutdown, ( "Entity %d is being Shutdown twice", GetGuid() ) );
}

void Asura_Entity::Send( Asura_Message* pxMessage, Asura_Guid uTo ) const
{
	Asura_Assert(Guid != ASURA_GUID_UNREGISTERED,"Unregistered entity!");

	if (uTo==ASURA_GUID_UNREGISTERED)
	{
		/// Kevin Lea - 29/11/2001 16:37:55 - send to self
		uTo = GetGuid();
	}
	if (pxMessage->GetEntityClassification() == AsuraEntityClass_NotSet)
	{
		pxMessage->SetEntityClassification(GetClassification());
	}

	pxMessage->SetToGuid(uTo);
	pxMessage->SetFromGuid(GetGuid());
	
	Asura_Network::Send(pxMessage);
}

void Asura_Entity::SetEntityActive( bool bActive )
{
	// activate or deactivate if it's the first time or if the status changed.
	// Only do the OnXXXMessage() stuff if the active status was initialised.
	bool bWasActive = IsEntityActive();
	m_bEntityActive = bActive;
	if ((!m_bActiveStatusInitialised) || (bWasActive != IsEntityActive()))
	{
		ActivityChanged();
	}
	m_bActiveStatusInitialised = true;
}

void Asura_Entity::ActivityChanged()
{
	if (IsEntityActive())
	{
		// activate first then do the side-effects
		OnEntityActivate();
		if (m_bActiveStatusInitialised)
		{
			OnEntityActivateMessage();
		}
	}
	else
	{
		// do the side-effects first, then deactivate
		if (m_bActiveStatusInitialised)
		{
			OnEntityDeactivateMessage();
		}
		OnEntityDeactivate();
	}
}

void Asura_Entity::SetTemplateInstanceActive( bool bActive )
{
	// only do ActivityChanged if m_bActiveStatusInitialised is true so 
	// it doesn't matter if this is called before or after PostLoadInitialise()
	// for this entity.
	bool bWasActive = IsEntityActive();
	m_bTemplateInstanceActive = bActive;
	if (m_bActiveStatusInitialised && bWasActive != IsEntityActive())
	{
		ActivityChanged();
	}
}

bool Asura_Entity::HandleMessage(const Asura_Message* pxMessage)
{
    Asura_Assert(pxMessage, "");

	if (NULL != pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			case ASURA_MSG_DESTRUCT:
			{	
				SetAsDestroyed();
				return true;
			}
			case ASURA_MSG_ACTIVATE:
			{
				SetEntityActive(true);
				return true;
			}
			case ASURA_MSG_DEACTIVATE:
			{
				SetEntityActive(false);
				return true;
			}
			default:
			{
				break;
			}
		}
	}

#if ASURA_NETWORK_LOGGING
	char acGuidFrom[32];
	char acGuidTo[32];
	const char* szGuidTo = NULL;
	const char* szGuidFrom = NULL;
	const char* szMsgID = NULL;
	Asura_Network::GetMsgIDDesc(pxMessage->GetMsgID(), &szMsgID);
	Asura_Network::GetGuidDesc(pxMessage->GetToGuid(), &szGuidTo);
	Asura_Network::GetGuidDesc(pxMessage->GetFromGuid(), &szGuidFrom);
	if (!szGuidFrom)
	{
		Asura_PrintfToBfr(acGuidFrom, "%u", 32, true, pxMessage->GetFromGuid());
		szGuidFrom = acGuidFrom;
	}
	if (!szGuidTo)
	{
		Asura_PrintfToBfr(acGuidTo, "%u", 32, true, pxMessage->GetToGuid());
		szGuidTo = acGuidTo;
	}
	Asura_Assertf(false,("Unhandled message (To: %s From: %s ID:%s)", szGuidTo, szGuidFrom, szMsgID));
#else
	Asura_Assertf(0,("Unhandled message (To: %d From: %d ID:%d)",pxMessage->To,pxMessage->From,pxMessage->GetMsgID()));
#endif
	return false;
}

void Asura_Entity::Update(void)
{
#ifdef ASURA_USE_STREAMING
	if (m_bAllowZoneChange)
	{
		Asura_Hash_ID uNewZoneHash = CheckForZoneChange();
		if (uNewZoneHash != m_uZoneHashID)
		{
			MigrateZone(uNewZoneHash);
		}
	}
#endif
}

#ifdef ASURA_USE_STREAMING
void Asura_Entity::MigrateZone(Asura_Hash_ID uNewZoneHash)
{
	Asura_Streaming_Zone* pxOldZone = Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_Zone* pxNewZone = Asura_Streaming_System::FindMasterZone(uNewZoneHash);
	if (pxOldZone && pxNewZone)
	{
		Asura_Streaming_ZoneEntities* pxOldEntities = pxOldZone->GetEntities();
		Asura_Streaming_ZoneEntities* pxNewEntities = pxNewZone->GetEntities();
		if (pxOldEntities && pxNewEntities)
		{
			MigrateZone(pxOldEntities, pxNewEntities);
			m_uZoneHashID = uNewZoneHash;
		}
	}
}
#endif

void Asura_Entity::InitFromBlueprint()
{
    Asura_Blueprint* pxBlueprint = GetBlueprint();
    if ( pxBlueprint )
    {
        InitialiseFromBlueprint( *pxBlueprint );
    }
}

Asura_Blueprint* Asura_Entity::GetBlueprint() const
{
    if ( m_uBlueprintHashID == ASURA_HASH_ID_UNSET || m_uBlueprintTypeHashID == ASURA_HASH_ID_UNSET )
    {
        return NULL;
    }

    Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( m_uBlueprintTypeHashID, m_uBlueprintHashID );
#ifdef ASURA_TOOLS
	Asura_Assertf( pxBlueprint, ( "Asura_Entity::GetBlueprint - Entity %s (%d): couldn't find blueprint 0x%0X in type 0x%0X", (ToolsData ? ToolsData->Name.GetString() : "UNKNOWN"), GetGuid(), m_uBlueprintHashID, m_uBlueprintTypeHashID ) );
#else
	Asura_Assertf( pxBlueprint, ( "Asura_Entity::GetBlueprint - Entity %d couldn't find blueprint 0x%0X in type 0x%0X", GetGuid(), m_uBlueprintHashID, m_uBlueprintTypeHashID ) );	
#endif
    return pxBlueprint;
}

#ifdef ASURA_TOOLS

bool Asura_Entity::ShouldIncludeInLevelExport() const
{
	//If this being called during a level export ,then check to see if the entity is intended
	//for the current platform. RAR 07/09/2005
	//Also check whether it should be included for the current light set.
	//RAR 12/06/2006
	//Add IsShuttingDown() as an optional condition, to allow this to be called from within Tools_Entity_Object::DeleteUnwantedEntitiesForQuitToGame
	//Get platform from current mesh when not exporting - PJC 23/01/2009

	if(ToolsData && (Tools_System::IsShuttingDown() || Tools_Level_Export::IsExport()))
	{
		if(!ToolsData->IsMemberOfLightSets(Tools_LightSets::GetCurrentLightSets())) return false;

		int iExportType = (Tools_Level_Export::IsExport()) ? Tools_Level_Export::s_eExportType : Tools_PlatformMesh_System::GetPlatform(false);
		if (ToolsData->ShouldExportToPlatform(iExportType))
		{
			return	(true);
		}
		else
		{
			return	(false);
		}
	}
	else
	{
		return true;
	}
}
#endif

Asura_Entity_PeriodicClientUpdateList* Asura_Entity::GetPeriodicClientUpdateList() const
{
#if defined(ASURA_USE_STREAMING)
	// Find the relevant list.
	Asura_Streaming_Zone* pxZone = Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	if (pxZone)
	{
		Asura_Streaming_ZoneEntities* pxEntities = pxZone->GetEntities();
		if (pxEntities)
		{
			return pxEntities->GetPeriodicClientUpdateList();
		}
	}
	Asura_Assert(0, "Error: Couldn't find periodic update list.");
	return 0;
#else
	return Asura_Entity_System::GetPeriodicClientUpdateList();
#endif
}

Asura_Hashed_List<Asura_Entity_Server>* Asura_Entity::GetConstantClientUpdateList() const
{
#if defined(ASURA_USE_STREAMING)
	// Find the relevant list.
	Asura_Streaming_Zone* pxZone = Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	if (pxZone)
	{
		Asura_Streaming_ZoneEntities* pxEntities = pxZone->GetEntities();
		if (pxEntities)
		{
			return pxEntities->GetConstantClientUpdateList();
		}
	}
	Asura_Assert(0, "Error: Couldn't find constant update list.");
	return 0;
#else
	return Asura_Entity_System::GetConstantClientUpdateList();
#endif
}

#if defined(ASURA_USE_STREAMING)
Asura_Hash_ID Asura_Entity::UpdateContainingZone(const Asura_Vector_3& xPosition)
{
	m_xZoneCache.Update(xPosition);
	return(m_xZoneCache.GetZoneHash());
}
#endif

#ifdef ASURA_USE_STREAMING
Asura_Hash_ID Asura_Entity::CheckForZoneChange( void )
{
	// Cannot do anything at this level, as we have no position to update with.
    Asura_Assert(false, "Cannot change zones without a position");
    return m_uZoneHashID;
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity_Server
//
//////////////////////////////////////////////////////////////////////////////

Asura_Combined_List<Asura_Entity_Server> Asura_Entity_Server::s_xList( 5 );
Asura_Classified_List<Asura_Entity_Server> Asura_Entity_Server::s_xUpdateableList;

Asura_Entity_Server::Asura_Entity_Server(Asura_Guid uGuid)
: m_iNumberOfEvents( 0 )
, m_pxEvents( 0 )
, m_bShouldAddToUpdateListWhenActivated( 0 )
, m_bShouldAddToClientUpdateListWhenActivated( 0 )
, m_bAlwaysUpdateClient( 0 )
#ifdef ASURA_ASSERT
, m_bSentBaseEventMessage( 0 )
#endif
{
	/// DWE - 31/07/2003 16:50:32 - The glass server entity is allowed to be created on the client
	/// so that we can get all the initialisation data from it (e.g. texture name)
	/// SO I'm preventing this assert from firing.
	/// KOR - 15/03/2004 15:30:16 - we might not have started the network in Tools
#ifndef ASURA_TOOLS
	Asura_Assert(Asura_Network::IsServer() || Classification == AsuraEntityClass_Glass, "creating server entity on non-server machine");
#endif
	if (ASURA_GUID_UNREGISTERED==uGuid)
	{
		Asura_Entity_System::RegisterServerEntity(this);
	}
	else
	{
		Guid = uGuid;
	}

	AddToList();
    GetPeriodicClientUpdateList()->AddToList( this );
	AddToUpdateableList();
}

Asura_Entity_Server::~Asura_Entity_Server()
{
#ifdef ASURA_SUPPORT_NETMSG_SYSTEM
	Asura_NetMsg_System::EntityDestroyed( GetGuid() );
#endif

    // Try removing from periodic list.
    if( !GetPeriodicClientUpdateList()->RemoveFromList(this) )
    {
        // Ok, maybe it's on the constant update list. (Otherwise it must
        // have had SetNeverUpdateClient() called on it.)
        GetConstantClientUpdateList()->RemoveFromListIfPresent( this );
    }

	if(m_pxEvents) delete [] m_pxEvents;

	RemoveFromUpdateableList(false);
	RemoveFromList();
}

bool Asura_Entity_Server::HandleMessage(const Asura_Message* pxMessage)
{
    Asura_Assert(pxMessage, "");

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
		case ASURA_MSG_DESTRUCT:
			SetThisAndClientsAsDestroyed();
			return true;

		case ASURA_MSG_CLIENT_REQUESTCREATEMSG:
			// By default, use the create message if we have one.
			if (GetHasCreateMessage())
			{
				// Send guaranteed if we aren't sending client update messages as, chances are, there's not going to be repeated create requests if we aren't updating.
				MakeCreateMessage(pxMessage->GetFromGuid(), GetClientUpdateMode() < 0);
				return true;
			}
			// Without one, don't handle the message.
			break;

		default:
			break;
		}
	}

	// Pass on to parent.
   return PARENT::HandleMessage(pxMessage);
}

void Asura_Entity_Server::PostLoadInitialise()
{
	PARENT::PostLoadInitialise();
	SetEntityActive( GetEntityActiveIgnoringTemplate() );
}

void Asura_Entity_Server::SetAlwaysUpdateClient( bool bEnable )
{
    // Shouldn't be using these functions if you've previously called
    // SetUpdateClientEveryFrame(true). It's necessary to call
    // SetUpdateClientEveryFrame(false) first.
    Asura_Assert( !GetConstantClientUpdateList()->FindEntity(GetGuid()), "" );

    GetPeriodicClientUpdateList()->SetAlwaysUpdate( this, bEnable );
}

void Asura_Entity_Server::SetNeverUpdateClient()
{
    // Shouldn't be using these functions if you've previously called
    // SetUpdateClientEveryFrame(true). It's necessary to call
    // SetUpdateClientEveryFrame(false) first.
    Asura_Assert( !GetConstantClientUpdateList()->FindEntity(GetGuid()), "" );

	GetPeriodicClientUpdateList()->RemoveFromList( this );
}

void Asura_Entity_Server::RequestClientUpdate()
{
    // Shouldn't be using these functions if you've previously called
    // SetUpdateClientEveryFrame(true). It's necessary to call
    // SetUpdateClientEveryFrame(false) first.
    Asura_Assert( !GetConstantClientUpdateList()->FindEntity(GetGuid()), "" );

    GetPeriodicClientUpdateList()->UpdateNeeded( this );
}

bool Asura_Entity_Server::IsInClientUpdateLists() const
{
	return (GetConstantClientUpdateList()->FindEntity(GetGuid()) || GetPeriodicClientUpdateList()->IsInList(GetGuid()));
}

int Asura_Entity_Server::GetClientUpdateMode() const
{
	return GetPeriodicClientUpdateList()->GetUpdateMode( GetGuid() );
}

void Asura_Entity_Server::SetUpdateClientEveryFrame( bool bEnable )
{
    Asura_Entity_PeriodicClientUpdateList*  const pxPeriodicList = GetPeriodicClientUpdateList();
    Asura_Hashed_List<Asura_Entity_Server>* const pxConstantList = GetConstantClientUpdateList();

    // An entity should never end up in both lists...
    Asura_Assert( !(pxPeriodicList->IsInList(GetGuid()) && pxConstantList->FindEntity(GetGuid())), "" );

    if( bEnable )
    {
        // Don't do anything if the entity is already set to constantly
        // update its client since otherwise we'll just end up with it in
        // the list twice.
        if( !pxConstantList->FindEntity(GetGuid()) )
        {
            pxPeriodicList->RemoveFromList( this );
            pxConstantList->AddToList( this );
        }
    }
    else
    {
        // See comment in other branch.
        if( !pxPeriodicList->IsInList(GetGuid()) )
        {
            // Remove from constant update list and reset behaviour to
            // default.
            pxConstantList->RemoveFromList( this );
            SetAlwaysUpdateClient( true );
        }
    }
}

void Asura_Entity_Server::SetNeverUpdates(bool bNeverUpdates)
{
	if(bNeverUpdates)
	{
		RemoveFromUpdateableList(false);
	}
	else
	{
		// Make sure we don't get duplicates - classified list add doesn't check for this.
		if(!s_xUpdateableList.FindEntity(GetGuid()))
		{
			AddToUpdateableList();
		}
	}
}

bool Asura_Entity_Server::IsUpdateable()
{
	return (s_xUpdateableList.FindEntity(GetGuid()) != NULL);
}

void Asura_Entity_Server::SetThisAndClientsAsDestroyed()
{
    SetAsDestroyed();
    Asura_Message * pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DESTRUCT,0);
	if (pxMessage)
    {
        pxMessage->UseGuaranteedDelivery();
    	Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
    }
}

void Asura_Entity_Server::AddToList(void)
{
	// Add to the standard list...
	s_xList.AddToList(this);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone* pxZone = Asura_Streaming_System::GetZoneToBeCreatedIn();
	Asura_Assertf( pxZone, ("Can't find a zone to create entity %d in. Death is imminent.", Guid) );

	Asura_Streaming_ZoneEntities* pxEntities = pxZone->GetEntities();
	Asura_Assertf( pxEntities, ("Can't find a zone entities object to create entity %d in. Death is imminent.", Guid) );

	// Set the hash-ID of the zone which created us...
	m_uZoneHashID	= pxZone->GetHashID();
	// MGH 22/5/2006: If this is a zone rather than a mission then we can set the initial containing zone.
	m_xZoneCache.Reset(Asura_Streaming_System::GetContainingZoneHash(m_uZoneHashID));

	// Register with the streaming system...
	pxEntities->AddEntityToList(*this);
#endif
}

void Asura_Entity_Server::RemoveFromList(void)
{
	// Remove from the standard li st...
	s_xList.RemoveFromList(this);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone*			pxZone		= Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_ZoneEntities*	pxEntities	= pxZone->GetEntities();

	// Reset the zone hash-ID...
	m_uZoneHashID	= ASURA_HASH_ID_UNSET;

	// Register with the streaming system...
	pxEntities->RemoveEntityFromList(*this);
#endif
}

void Asura_Entity_Server::AddToUpdateableList(void)
{
	// Add to the standard list...
	s_xUpdateableList.AddToList(this);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone*			pxZone		= Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_ZoneEntities*	pxEntities	= pxZone->GetEntities();

	// Register with the streaming system...
	pxEntities->AddEntityToUpdateableList(*this);
#endif
}

void Asura_Entity_Server::RemoveFromUpdateableList(const bool bShouldAlwaysBePresent)
{
	// Remove from the standard list...
	s_xUpdateableList.RemoveFromList(this, bShouldAlwaysBePresent);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone*			pxZone		= Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_ZoneEntities*	pxEntities	= pxZone->GetEntities();

	// Register with the streaming system...
	pxEntities->RemoveEntityFromUpdateableList(*this, bShouldAlwaysBePresent);
#endif
}

#ifdef ASURA_USE_STREAMING
void Asura_Entity_Server::MigrateZone(Asura_Streaming_ZoneEntities* pxOldEntities, Asura_Streaming_ZoneEntities* pxNewEntities)
{
	Asura_Streaming_ZoneEntities::MigrateEntity(this, pxOldEntities, pxNewEntities);
}
#endif

// (MH - 31.03.2004) this used to be pure virtual. Now that entities without 
// client updates shouldn't get added to the update list, I'm removing empty
// versions of this function from derived classes, so this will get called.
void Asura_Entity_Server::MakeClientUpdateMessage( void )
{
#ifndef ASURA_UPDATE_CLIENTENTITIES_EVERY_FRAME
	Asura_Assert(false, "shouldn't get here");
#endif
}

int Asura_Entity_Server::GetChunkBufferSize()
{
	Asura_Chunk_Dummy_Stream xStream;
	// must only call our own version of WriteToChunkStream not the overridden one.
	Asura_Entity_Server::WriteToChunkStream(xStream);
	return xStream.GetPosition();
}

void Asura_Entity_Server::InitialiseFromChunkBuffer(const char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream( pcChunkBuffer );
	// must only call our own version of ReadFromChunkStream not the overridden one.
	Asura_Entity_Server::ReadFromChunkStream(xStream);
}


//For dealing with old physical objects that store this data elsewhere
void Asura_Entity_Server::ReadEventsFromChunkBuffer(const char*& pcChunkBuffer)
{
	ASURA_CHUNK_EXTRACT(pcChunkBuffer,m_iNumberOfEvents,int);
	if(m_iNumberOfEvents)
	{
		// MGH 28/3/2005: DTR calls InitialiseFromChunkBuffer more than once,
		// that is when resetting some entities. This prevents a memory leak.
		if (m_pxEvents)
		{
			delete [] m_pxEvents;
		}
		m_pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents];
		for(int i=0;i<m_iNumberOfEvents;i++)
		{
			ASURA_CHUNK_EXTRACT(pcChunkBuffer,m_pxEvents[i],Asura_EventAndMessage);
		}
	}
}

void Asura_Entity_Server::SkipDataInChunkBuffer(const char*& pcChunkBuffer, SkippedServerData* puSkippedServerData /* = NULL*/)
{
	Asura_Chunk_ChunkBuffer_Stream xStream( pcChunkBuffer );
	Asura_Entity_Server::SkipDataInChunkStream(xStream, puSkippedServerData);
}

void Asura_Entity_Server::SkipDataInChunkStream(Asura_Chunk_Stream& xStream, SkippedServerData* puSkippedServerData /* = NULL*/)
{
	int iVersion;
	xStream >> iVersion;

	Asura_Hash_ID uBlueprintHashID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uBlueprintTypeHashID = ASURA_HASH_ID_UNSET;
	bool bActive = true;
	bool bTemplateActive = true;

	// the reason each chunk version duplicates code from previous versions is to make
	// it simpler to retire old versions for non-tools builds.
	if ( iVersion & SERVER_PARENT_DATA_INCREMENT )
	{
		PARENT::SkipDataInChunkStream( xStream );
	}

	if ( iVersion >= 3 )
	{
        // Skip blueprint hashes.
        xStream >> uBlueprintHashID;
        xStream >> uBlueprintTypeHashID;

		// Skip flags (not using bitfield for these flags as we don't load/save all flags in entity bitfield)
		xStream >> bActive;
		xStream >> bTemplateActive;
		// maintain 4-byte alignment with 2 spare bools.
		bool bUnused1;
		bool bUnused2;
		xStream >> bUnused1;
		xStream >> bUnused2;

        // Skip events.
        int iNumberOfEvents;
        xStream >> iNumberOfEvents;
        xStream.Skip(iNumberOfEvents * 2 * sizeof(unsigned short));
    }
	else if ( iVersion == 2 )
    {
        // Skip blueprint hashes.
        xStream >> uBlueprintHashID;
        xStream >> uBlueprintTypeHashID;

		// Skip activity
		int iActive;
		xStream >> iActive;
		bActive = (iActive != 0);

        // Skip events.
        int iNumberOfEvents;
        xStream >> iNumberOfEvents;
        xStream.Skip(iNumberOfEvents * 2 * sizeof(unsigned short));
    }
#ifdef ASURA_TOOLS
	else if ( iVersion == 1 )
    {
        // Skip blueprint hashes.
        xStream >> uBlueprintHashID;
        xStream >> uBlueprintTypeHashID;

        // Skip events.
        int iNumberOfEvents;
        xStream >> iNumberOfEvents;
        xStream.Skip(iNumberOfEvents * 2 * sizeof(unsigned short));
    }
    else if ( iVersion == 0 )
    {
        // Skip events.
        int iNumberOfEvents;
        xStream >> iNumberOfEvents;
        xStream.Skip(iNumberOfEvents * 2 * sizeof(unsigned short));
    }
#endif
	else
    {
        Asura_Assertf( false, ( "Unknown chunk version %d for Asura_Entity_Server", iVersion ) );
    }

	if (puSkippedServerData)
	{
		puSkippedServerData->m_uBlueprintHashID = uBlueprintHashID;
		puSkippedServerData->m_uBlueprintTypeHashID = uBlueprintTypeHashID;
		puSkippedServerData->m_bActive = bActive;
		puSkippedServerData->m_bTemplateActive = bTemplateActive;
	}
}

void Asura_Entity_Server::FillInChunkBuffer(char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream( pcChunkBuffer );
	Asura_Entity_Server::WriteToChunkStream(xStream);
}


void Asura_Entity_Server::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	int iVersion;
	xStream >> iVersion;

	if ( iVersion & SERVER_PARENT_DATA_INCREMENT )
	{
		PARENT::ReadFromChunkStream(xStream);
	}

    if ( iVersion >= 3 )
    {
        // Read in the blueprint hash IDs.
        Asura_Hash_ID uHashID;
        xStream >> uHashID;
        SetBlueprintHashID( uHashID );
        xStream >> uHashID;
        SetBlueprintTypeHashID( uHashID );

		// version 3 - read in flags
		// (not using bitfield for these flags as we don't load/save all flags in entity bitfield)
		bool bActive;
		xStream >> bActive;
		SetEntityActiveFromLoad( bActive );

		bool bTemplateActive;
		xStream >> bTemplateActive;
		SetTemplateInstanceActiveFromLoad( bTemplateActive );

		// maintain 4-byte alignment with 2 spare bools.
		bool bUnused1;
		bool bUnused2;
		xStream >> bUnused1;
		xStream >> bUnused2;

        if ( Asura_LoadAndSave::IsLoadingFromBuffer() )
        {
            // We are loading from a buffer not a file (== loading a checkpoint or similar).
            // Initialise from the blueprints first and then continue loading the entity data, so that the entity is initialised
            // with the most recent blueprint data and then updates itself from its chunk data, so that data that has changed in game takes precedence.
            // This contrasts with the method for loading from a file (== loading an exported level), where we first load all of the chunk
            // data then we initialise from the blueprint data, so that any new blueprint data is applied to already exported levels.
            InitFromBlueprint();
        }

        // Read in events.
        xStream >> m_iNumberOfEvents;
        if(m_iNumberOfEvents)
        {
            // MGH 28/3/2005: DTR calls InitialiseFromChunkBuffer more than once,
            // that is when resetting some entities. This prevents a memory leak.
            if (m_pxEvents)
            {
                delete [] m_pxEvents;
            }
            m_pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents];
            for(int i=0;i<m_iNumberOfEvents;i++)
            {
                xStream >> m_pxEvents[i].m_usEventID;
                xStream >> m_pxEvents[i].m_usMessageBlockID;
            }
        }
    }
    else if ( iVersion == 2 )
    {
        // Read in the blueprint hash IDs.
        Asura_Hash_ID uHashID;
        xStream >> uHashID;
        SetBlueprintHashID( uHashID );
        xStream >> uHashID;
        SetBlueprintTypeHashID( uHashID );

		// version 2 - read activity status. Use int to avoid alignment issues.
		int iActive;
		xStream >> iActive;
		SetEntityActiveFromLoad(iActive != 0);

        if ( Asura_LoadAndSave::IsLoadingFromBuffer() )
        {
            // We are loading from a buffer not a file (== loading a checkpoint or similar).
            // Initialise from the blueprints first and then continue loading the entity data, so that the entity is initialised
            // with the most recent blueprint data and then updates itself from its chunk data, so that data that has changed in game takes precedence.
            // This contrasts with the method for loading from a file (== loading an exported level), where we first load all of the chunk
            // data then we initialise from the blueprint data, so that any new blueprint data is applied to already exported levels.
            InitFromBlueprint();
        }

        // Read in events.
        xStream >> m_iNumberOfEvents;
        if(m_iNumberOfEvents)
        {
            // MGH 28/3/2005: DTR calls InitialiseFromChunkBuffer more than once,
            // that is when resetting some entities. This prevents a memory leak.
            if (m_pxEvents)
            {
                delete [] m_pxEvents;
            }
            m_pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents];
            for(int i=0;i<m_iNumberOfEvents;i++)
            {
                xStream >> m_pxEvents[i].m_usEventID;
                xStream >> m_pxEvents[i].m_usMessageBlockID;
            }
        }
    }
#ifdef ASURA_TOOLS
    else if ( iVersion == 1 )
    {
        // Read in the blueprint hash IDs.
        Asura_Hash_ID uHashID;
        xStream >> uHashID;
        SetBlueprintHashID( uHashID );
        xStream >> uHashID;
        SetBlueprintTypeHashID( uHashID );

        if ( Asura_LoadAndSave::IsLoadingFromBuffer() )
        {
            // We are loading from a buffer not a file (== loading a checkpoint or similar).
            // Initialise from the blueprints first and then continue loading the entity data, so that the entity is initialised
            // with the most recent blueprint data and then updates itself from its chunk data, so that data that has changed in game takes precedence.
            // This contrasts with the method for loading from a file (== loading an exported level), where we first load all of the chunk
            // data then we initialise from the blueprint data, so that any new blueprint data is applied to already exported levels.
            InitFromBlueprint();
        }

        // Read in events.
        xStream >> m_iNumberOfEvents;
        if(m_iNumberOfEvents)
        {
            // MGH 28/3/2005: DTR calls InitialiseFromChunkBuffer more than once,
            // that is when resetting some entities. This prevents a memory leak.
            if (m_pxEvents)
            {
                delete [] m_pxEvents;
            }
            m_pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents];
            for(int i=0;i<m_iNumberOfEvents;i++)
            {
                xStream >> m_pxEvents[i].m_usEventID;
                xStream >> m_pxEvents[i].m_usMessageBlockID;
            }
        }
    }    
	else if ( iVersion == 0 )
    {
        // Read in events.
        xStream >> m_iNumberOfEvents;
        if(m_iNumberOfEvents)
        {
            // MGH 28/3/2005: DTR calls InitialiseFromChunkBuffer more than once,
            // that is when resetting some entities. This prevents a memory leak.
            if (m_pxEvents)
            {
                delete [] m_pxEvents;
            }
            m_pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents];
            for(int i=0;i<m_iNumberOfEvents;i++)
            {
                xStream >> m_pxEvents[i].m_usEventID;
                xStream >> m_pxEvents[i].m_usMessageBlockID;
            }
        }
    }
#endif // #ifdef ASURA_TOOLS
    else
    {
        Asura_Assertf( false, ( "Unknown chunk version %d for Asura_Entity_Server", iVersion ) );
    }
}

void Asura_Entity_Server::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// Write version number.
	xStream << CURRENT_CHUNK_VERSION;

	PARENT::WriteToChunkStream(xStream);

    // Write blueprint hash IDs.
    xStream << GetBlueprintHashID();
    xStream << GetBlueprintTypeHashID();

	// version 3 - write out flags
	// (not using bitfield for these flags as we don't load/save all flags in entity bitfield)
	bool bActive = GetEntityActiveIgnoringTemplate();
	xStream << bActive;

	bool bTemplateActive = GetTemplateInstanceActive();

	#ifdef ASURA_TOOLS

		if ( ToolsData )
		{
			if ( !ToolsData->IsMemberOfTemplate() )
			{
				bTemplateActive = true;
			}
		}

	#endif

	xStream << bTemplateActive;

	// maintain 4-byte alignment with 2 spare bools. These are guaranteed to be false
	// so they can be brought into use without bumping the chunk version.
	bool bUnused1 = false;
	bool bUnused2 = false;
	xStream << bUnused1;
	xStream << bUnused2;

	// Write events.
	xStream << m_iNumberOfEvents;
    for ( int i = 0; i < m_iNumberOfEvents; ++i )
    {
        xStream << m_pxEvents[ i ].m_usEventID;
        xStream << m_pxEvents[ i ].m_usMessageBlockID;
    }
}

void Asura_Entity_Server::SendMessagesForEvent(int iEventID)
{
	for(int i=0;i<m_iNumberOfEvents;i++)
	{
		if((int)m_pxEvents[i].m_usEventID == iEventID)
		{
			if(m_pxEvents[i].m_usMessageBlockID != ASURA_STATICMESSAGE_EMPTYBLOCK)
			{
#ifdef ASURA_ASSERT
				if (m_pxEvents[i].m_usEventID == ASURA_ENTITY_EVENT_ONACTIVATION ||
					m_pxEvents[i].m_usEventID == ASURA_ENTITY_EVENT_ONDEACTIVATION)
				{
					m_bSentBaseEventMessage = true;
				}
#endif
				Asura_StaticMessage_System::SendMessages(m_pxEvents[i].m_usMessageBlockID,GetGuid());
			}
		}
	}
}

bool Asura_Entity_Server::SetMessageBlockForEvent( int iEventID, u_short usMsgBlockID )
{
	// check there isn't already a message block for this event
	for(int i=0;i<m_iNumberOfEvents;i++)
	{
		if(m_pxEvents[i].m_usEventID == (unsigned short)iEventID)
		{
			Asura_Assert( false, "Entity already has a message block for this event!" );
			return false;
		}
	}

	// if we've reached here, there can't already be an event entry, so make a new one
	
	//need to add a new entry for this event then.
	//allocate a larger array , and copy over the old ones
	Asura_EventAndMessage* pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents+1];
	for(int i=0;i<m_iNumberOfEvents;i++)
	{
		pxEvents[i] = m_pxEvents[i];
	}

	//initialise the new event
	int iEventIndex = m_iNumberOfEvents++;
	pxEvents[iEventIndex].m_usEventID = (unsigned short) iEventID;
	pxEvents[iEventIndex].m_usMessageBlockID = usMsgBlockID;

	delete [] m_pxEvents;
	m_pxEvents = pxEvents;

	return true;
}

#ifdef ASURA_TOOLS
void Asura_Entity_Server::AddMessageForEvent(int iEventID,Asura_Guid uGuid,int iMessageID)
{
	//check for an existing message block for this event
	int iEventIndex = -1;
	for(int i=0;i<m_iNumberOfEvents;i++)
	{
		if(m_pxEvents[i].m_usEventID == (unsigned short)iEventID)
		{
			iEventIndex = i;
			break;
		}
	}
	if(iEventIndex == -1)
	{
		//need to add a new entry for this event then.
		//allocate a larget array , and copy over the old ones
		Asura_EventAndMessage* pxEvents = new Asura_EventAndMessage[m_iNumberOfEvents+1];
		for(int i=0;i<m_iNumberOfEvents;i++)
		{
			pxEvents[i] = m_pxEvents[i];
		}
		//initialise the new event
		iEventIndex = m_iNumberOfEvents++;
		pxEvents[iEventIndex].m_usEventID = (unsigned short) iEventID;
		pxEvents[iEventIndex].m_usMessageBlockID = ASURA_STATICMESSAGE_EMPTYBLOCK;

		delete [] m_pxEvents;
		m_pxEvents = pxEvents;
	}

	//create the message to add
	Asura_StaticMessage xMessage;
	xMessage.m_usMsgID = (unsigned short) iMessageID;
	xMessage.m_uSendTo = uGuid;	

	//and add it to the event
	Asura_StaticMessage_System::AddMessage(ASURA_STATICMESSAGE_ENTITYSET,xMessage,m_pxEvents[iEventIndex].m_usMessageBlockID);
}
#endif

#ifdef ASURA_TOOLS
bool Asura_Entity_Server::HasMessageBlockForEvent( int iEventID ) const
{
	//check for an existing message block for this event
	for(int i=0;i<m_iNumberOfEvents;i++)
	{
		if( m_pxEvents[i].m_usEventID == (u_short)iEventID ) return true;
	}

	return false;
}
#endif

#ifdef ASURA_TOOLS
void Asura_Entity_Server::CopyMessagesForEvent( const Asura_Entity_Server* pxEntity, int iEventID )
{
	int iToCopy = -1;
	int i;

	for( i = 0; i < pxEntity->m_iNumberOfEvents; i++)
	{
		if( pxEntity->m_pxEvents[i].m_usEventID == static_cast<unsigned short>( iEventID ) )
		{
			iToCopy = i;
		}
	}

	for( i = 0; i < m_iNumberOfEvents; i++)
	{
		// check there isn't already a message block for this event	
		if( m_pxEvents[i].m_usEventID == static_cast<unsigned short>( iEventID ) )
		{
			Asura_Assert( false, "Entity already has a message block for this event!" );
			return;
		}
	}

	if( iToCopy == -1 )
	{
		// Nothing to copy
		return;
	}

	const unsigned short usNewMsgBlock = Asura_StaticMessage_System::CopyMessageBlock( pxEntity->m_pxEvents[iToCopy].m_usMessageBlockID, pxEntity->GetGuid(), GetGuid() );
	
	SetMessageBlockForEvent( iEventID, usNewMsgBlock );
}
#endif

#ifdef ASURA_TOOLS
void Asura_Entity_Server::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	PARENT::ApplyChangesForTemplateInstance(pxInstance);

	//update message blocks for events
	for(int i=0;i<m_iNumberOfEvents;i++)
	{
		m_pxEvents[i].m_usMessageBlockID = pxInstance->ConvertMessageBlock(m_pxEvents[i].m_usMessageBlockID);	
	}
}
#endif

#ifdef ASURA_TOOLS
int Asura_Entity_Server::GetCustomEventMessageSet(int iEventIndex) const
{
	return ASURA_STATICMESSAGE_ENTITYSET;
}
#endif

void Asura_Entity_Server::MakeCreateMessage(Asura_Guid uRequestingClient, bool bSendGuaranteed/* = false*/)
{
	Asura_Assertf(false, ("Entity %d (classification %d) doesn't have a create message.", Guid, Classification));
}

// assert-only code to check event messages are sent correctly
#ifdef ASURA_ASSERT
bool Asura_Entity_Server::ExpectingBaseEventMessage( u_short usMsgID )
{
	// check if activate/deactivate message should be sent but isn't
	if (usMsgID == ASURA_MSG_ACTIVATE || usMsgID == ASURA_MSG_DEACTIVATE)
	{
		int iEventID = (usMsgID == ASURA_MSG_ACTIVATE) ? ASURA_ENTITY_EVENT_ONACTIVATION : ASURA_ENTITY_EVENT_ONDEACTIVATION;

		for(int i=0;i<m_iNumberOfEvents;i++)
		{
			if(static_cast<int>(m_pxEvents[i].m_usEventID) == iEventID)
			{
				if(m_pxEvents[i].m_usMessageBlockID != ASURA_STATICMESSAGE_EMPTYBLOCK)
				{
					m_bSentBaseEventMessage = 0;
					return true;
				}
			}
		}
	}
	return false;
}
#endif // ASURA_ASSERT


// overrides of OnEntityActivate()/OnEntityDeactivate() - called when active status changes or is initialised. 
// Derived classes should always call parent versions if they override them. 
// We remove the entity from the update lists as appropriate on deactivation and put it back in on activation.
void Asura_Entity_Server::OnEntityActivate(void) 
{
	PARENT::OnEntityActivate();

	if (m_bShouldAddToUpdateListWhenActivated)
	{
		SetNeverUpdates(false);
	}

	if (m_bShouldAddToClientUpdateListWhenActivated)
	{
		SetAlwaysUpdateClient(m_bAlwaysUpdateClient);
	}
}

void Asura_Entity_Server::OnEntityDeactivate(void) 
{
	PARENT::OnEntityDeactivate();

	if (IsUpdateable())
	{
		// set flag so we know this entity should go back in the list when activated.
		m_bShouldAddToUpdateListWhenActivated = true;
		SetNeverUpdates(true);
	}
	else
	{
		m_bShouldAddToUpdateListWhenActivated = false;
	}

	if (IsInClientUpdateLists())
	{
		// (MH - 18.10.2007) haven't got a suitable test case for entities using the "update client every frame" setting.
		// Rather than implement it without testing, I'm leaving it until we can. It should be a simple flag to record the
		// fact and remove it first then carry on as normal, and check the flag on activating to restore the setting.
		Asura_Assert(!GetConstantClientUpdateList()->FindEntity(GetGuid()), "code path not supported yet");

		// set flag so we know this entity should go back in the list when activated.
		m_bShouldAddToClientUpdateListWhenActivated = true;
		// remember the update mode (always or burst/heartbeat)
		m_bAlwaysUpdateClient = (GetClientUpdateMode() == ClientUpdateList_UpdateMode_Always);
		SetNeverUpdateClient();
	}
	else
	{
		m_bShouldAddToClientUpdateListWhenActivated = false;
	}
}

// overrides of OnEntityActivateMessage()/OnEntityDeactivateMessage() - called when active status changes as a result 
// of a message being received as opposed to on creation/loading. Used for things that have side-effects
// such as sending messages to other entities.
// Derived classes should always call parent versions if they override them. 
// We send a message to client entities about the change of state, if we're updating a client entity.
void Asura_Entity_Server::OnEntityActivateMessage(void) 
{
	PARENT::OnEntityActivateMessage();

	SendMessagesForEvent(ASURA_ENTITY_EVENT_ONACTIVATION);

	if (HasClientEntities())
	{
#ifndef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		// If we have a create message, our client entities should be destroyed on deactivation, so send a create message instead.
		if (GetHasCreateMessage())
		{
			MakeCreateMessage(ASURA_GUID_UNREGISTERED, GetClientUpdateMode() < 0);
		}
		else
#endif //#ifndef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_ACTIVATE );
			if( pxMessage )
			{
				pxMessage->UseGuaranteedDelivery();
				Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
			}
		}
	}
}

void Asura_Entity_Server::OnEntityDeactivateMessage(void) 
{
	PARENT::OnEntityDeactivateMessage();

	SendMessagesForEvent(ASURA_ENTITY_EVENT_ONDEACTIVATION);

	if (HasClientEntities())
	{
		u_short usMsgID;
#ifndef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		if (GetHasCreateMessage())
		{
			// If this entity can be recreated later, destroy its client on deactivation.
			usMsgID = ASURA_MSG_DESTRUCT;
		}
		else
#endif //#ifndef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		{
			usMsgID = ASURA_MSG_DEACTIVATE;
		}

		Asura_Message* pxMessage = Asura_Network::CreateMessage(usMsgID);
		if( pxMessage )
		{
			pxMessage->UseGuaranteedDelivery();
			Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
		}
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity_Client
//
//////////////////////////////////////////////////////////////////////////////

Asura_Combined_List<Asura_Entity_Client> Asura_Entity_Client::s_xList( 5 );
Asura_Classified_List<Asura_Entity_Client> Asura_Entity_Client::s_xUpdateableList;

Asura_Entity_Client::Asura_Entity_Client()
: m_uClientFlags( 0 )
{
	Asura_Entity_System::RegisterClientEntity(this); // set guid

	AddToList();

	InitialiseClientEntityActivity();
}

Asura_Entity_Client::Asura_Entity_Client(Asura_Guid uGuid)
: m_uClientFlags( 0 )
{
	Asura_Assert(uGuid!=ASURA_GUID_UNREGISTERED,"");
	Guid = uGuid;

	AddToList();

	InitialiseClientEntityActivity();
}

Asura_Entity_Client::~Asura_Entity_Client()
{
	RemoveFromList();
	// (MH - 24.10.2007) might as well remove from updateable list so derived classes don't all have to.
	RemoveFromUpdateableList(false);

#ifdef ASURA_SUPPORT_NETMSG_SYSTEM
	Asura_NetMsg_System::EntityDestroyed(GetGuid());
#endif

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	// (MH - 02.09.2003) I'm assuming that this network flag will not be 
	// cleared until after all entities have been destroyed.
	if (Asura_Network::IsExternalNetworkActive())
	{
		Asura_RecentlyDestroyedEntity_System::AddDestroyedEntity(GetGuid());
	}
#endif
}

void Asura_Entity_Client::SetAsDestroyed(void)
{
	PARENT::SetAsDestroyed();
		
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	if (Asura_Network::IsExternalNetworkActive())
	{
		Asura_RecentlyDestroyedEntity_System::AddDestroyedEntity(GetGuid());
	}
#endif
}

bool Asura_Entity_Client::HasRecentlyBeenDestroyed(Asura_Guid uGuid)
{
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	// (MH - 02.09.2003) I'm assuming that this network flag will not be cleared until after all entities have been destroyed.
	if (Asura_Network::IsExternalNetworkActive())
	{
		return Asura_RecentlyDestroyedEntity_System::HasRecentlyBeenDestroyed(uGuid);
	}
#endif
	return false;
}

void Asura_Entity_Client::AddToList(void)
{
	// Add to the standard list...
	s_xList.AddToList(this);

#if defined(ASURA_USE_STREAMING)
	// MGH 2/5/2006: Look up the server entity to get the zone hash.
	// This only works on the server machine, but then streaming only works
	// in single player at the moment.
	Asura_Streaming_Zone* pxZone = 0;

	const Asura_Entity* const pxServer = Asura_Entity_Server::FindEntity(Guid);
	if (pxServer)
	{
		// Copy the zone hash from the server entity.
		m_uZoneHashID = pxServer->GetZoneHashID();
		pxZone = Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
		Asura_Assert(pxZone, "Null zone!");
		// MGH 22/5/2006: Copy the containing zone too.
		m_xZoneCache = pxServer->GetZoneCache();
	}

	if (!pxZone)
	{
		// Just use the current master zone. Usually this will return the mission.
		pxZone = Asura_Streaming_System::GetZoneToBeCreatedIn();
		if (pxZone)
		{
			m_uZoneHashID = pxZone->GetHashID();
		}
		Asura_Assert(pxZone, "Error: Null zone!");
		// MGH 22/5/2006: If this is a zone rather than a mission then we can set the initial containing zone.
		m_xZoneCache.Reset(Asura_Streaming_System::GetContainingZoneHash(m_uZoneHashID));
	}

	if (pxZone)
	{
		Asura_Streaming_ZoneEntities* const pxEntities = pxZone->GetEntities();
		Asura_Assert(pxEntities, "Error: Null zone entities!");
		if (pxEntities)
		{
			// Register with the streaming system...
			pxEntities->AddEntityToList(*this);
		}
	}
#endif
}

void Asura_Entity_Client::RemoveFromList(void)
{
	// Remove from the standard list...
	s_xList.RemoveFromList(this);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone*			pxZone		= Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_ZoneEntities*	pxEntities	= pxZone->GetEntities();

	// Register with the streaming system...
	pxEntities->RemoveEntityFromList(*this);
#endif
}

void Asura_Entity_Client::AddToUpdateableList(void)
{
	// Add to the standard list...
	s_xUpdateableList.AddToList(this);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone*			pxZone		= Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_ZoneEntities*	pxEntities	= pxZone->GetEntities();

	// Register with the streaming system...
	pxEntities->AddEntityToUpdateableList(*this);
#endif
}

void Asura_Entity_Client::RemoveFromUpdateableList(const bool bShouldAlwaysBePresent)
{
	// Remove from the standard list...
	s_xUpdateableList.RemoveFromList(this, bShouldAlwaysBePresent);

#if defined(ASURA_USE_STREAMING)
	Asura_Streaming_Zone*			pxZone		= Asura_Streaming_System::FindMasterZone(m_uZoneHashID);
	Asura_Streaming_ZoneEntities*	pxEntities	= pxZone->GetEntities();

	// Register with the streaming system...
	pxEntities->RemoveEntityFromUpdateableList(*this, bShouldAlwaysBePresent);
#endif
}

#ifdef ASURA_USE_STREAMING
void Asura_Entity_Client::MigrateZone(Asura_Streaming_ZoneEntities* pxOldEntities, Asura_Streaming_ZoneEntities* pxNewEntities)
{
	Asura_Streaming_ZoneEntities::MigrateEntity(this, pxOldEntities, pxNewEntities);
}
#endif

// overrides of OnEntityActivate()/OnEntityDeactivate() - called when active status changes or is initialised. 
// Derived classes should always call parent versions if they override them. 
// We remove the entity from the update lists as appropriate on deactivation and put it back in on activation.
void Asura_Entity_Client::OnEntityActivate(void) 
{
	PARENT::OnEntityActivate();

	if (m_bShouldAddToUpdateListWhenActivated)
	{
		AddToUpdateableList();
	}
}

void Asura_Entity_Client::OnEntityDeactivate(void) 
{
	PARENT::OnEntityDeactivate();

	if (IsUpdateable())
	{
		// set flag so we know this entity should go back in the list when activated.
		m_bShouldAddToUpdateListWhenActivated = true;
		RemoveFromUpdateableList(true);
	}
	else
	{
		m_bShouldAddToUpdateListWhenActivated = false;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
