//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity_Renderable.h
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_ENTITY_RENDERABLE_H___
#define ___ASURA_ENTITY_RENDERABLE_H___ 1

#include "Core/Asura_Handle.h"
#include "Core/Asura_Combined_List.h"
#include "Core/Asura_Maths.h"
#include "Entity/Asura_Entity.h"

/// Kevin Lea - 27/11/2001 16:03:05 - example of a derived entity, which is
/// itself designed to be used as a base entity type. Anything derived from
/// Asura_Renderable_Entity will be added to the 'Renderable' entity list,
/// and will have its Render method called each frame. If an entity is not
/// derived from this class, then it won't be drawn automatically.
class Asura_Entity_Renderable : public Asura_Entity_Client
{
	typedef Asura_Entity_Client PARENT;

	friend class Asura_Entity_System;
	friend class Asura_LOS_Penetration;

	public:

		enum TimeLagType
		{
			TimeLag_LocalPlayer,
			TimeLag_NetworkEntity,
			TimeLag_Projectile,
			NumTimeLagTypes
		};

		Asura_Entity_Renderable();
		virtual ~Asura_Entity_Renderable();

		virtual const Asura_Vector_3& GetCurrentPosition( void ) const;
        virtual const Asura_Matrix_3x3& GetCurrentOrientation( void ) const;
		virtual void GetCurrentVelocity( Asura_Vector_3& xVelocity ) const;
		void GetCurrentAngularVelocity( Asura_Euler_3& xEuler ) const;
		void SetTimeLagType( TimeLagType eType );

		virtual bool GetPositionForDialogue(Asura_Vector_3& xPosition) const;

        // Move and reorient the entity without any interpolation.
		virtual void TeleportTo( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation );

		//Returns true if any of the invisible flags are set
		bool IsInvisible() const { return m_ucInvisibleFlags!=0; }
		// KOR - 23/03/2006 - Now adds or removes from the renderable list as appropriate,
		// to avoid even calling the render function.
		// RAR 03/04/2005 - General purpose invisibility flag
		void SetInvisibleFlag(bool bInvisible);
		bool GetInvisibleFlag() const {return m_bInvisibleGeneral!=0;}
		// RAR 31/03/2006 - Makes invisible as a result of the artist set physical object flag
		void SetInvisibleByPhysicalFlag(bool bInvisible);
		bool GetInvisibleByPhysicalFlag() const { return m_bInvisibleByPhysicalFlag!=0;}
		// RAR 31/03/2006 - Makes invisible as a result of cutscenes
		void SetInvisibleByCutscene(bool bInvisible);
		bool GetInvisibleByCutscene() const {return m_bInvisibleByCutscene!=0;}
		// (MH - 24.10.2007) - makes invisible as a result of being deactivated
		void SetInvisibleByDeactivation(bool bInvisible);
		bool GetInvisibleByDeactivation() const {return m_bInvisibleByDeactivation!=0;}

		//This gets called when the entity switches between being visible/invisible
		virtual void OnVisibilityChanged() {}

		// RDM - 09/02/07 - Made these accessible publicly to allow for removal/add without messing up the invisible flag
		void AddToRenderableList(void);
		void RemoveFromRenderableList(const bool bShouldAlwaysBePresent);

		virtual void RenderShadow() {}

		inline void SetEntityAlwaysMoves(const bool bAlwaysMoves) { m_bEntityAlwaysMoves = bAlwaysMoves; }

		// Returns the conservative axially-aligned WORLD SPACE box covering the entity. 
		// Asserts if the derived class does not implement a version of the method. 
		virtual void GetLineOfSightAABB( Asura_Bounding_Box& xAABB );

		// KJFL - 6 November 2007 - LOSTest is now public so the new BVH doesn't need to be a friend
		virtual bool LOSTest(const class Asura_LOS_InputData*, class Asura_LOS_OutputData*)
		{
			return false;
		}

	protected:


		virtual void Render()=0;

		void AddToLineOfSightList( void );
		
		void RemoveFromLineOfSightList( void );

		// overrides of OnEntityActivate()/OnEntityDeactivate() - called when active status changes or is initialised. 
		// Derived classes should always call parent versions if they override them. 
		virtual	void OnEntityActivate(void);
		virtual	void OnEntityDeactivate(void);

#ifdef ASURA_USE_STREAMING
		virtual Asura_Hash_ID CheckForZoneChange();
	private:
		virtual void MigrateZone(Asura_Streaming_ZoneEntities* pxOldEntities, Asura_Streaming_ZoneEntities* pxNewEntities);
#endif //#ifdef ASURA_USE_STREAMING

	private:
		void SetInvisibleInternal( bool bInvisible, bool bWasInvisible );

	public:
        // (MH - 15.07.2002) force users to pass an orientation now
        void ProcessNewPositionAndOrientationData(const Asura_Vector_3& xPosition, 
                                                  const Asura_Quat& xOrientation,
                                                  float fServerTimestamp);
	protected:
#ifndef ASURA_NO_NETWORK_LAG_COMPENSATION
		// (MH - 28.08.2003) attempt to improve projectiles
		void InitialiseMovingObject(const Asura_Vector_3& xPosition, 
									const Asura_Quat& xOrientQuat,
									const Asura_Vector_3& xVelocity,
									float fServerTime);
#endif
		
		const Asura_Vector_3& GetLastReportedPosition() const;
		const Asura_Quat& GetLastReportedOrientation() const;
		void GetLastReportedOrientationMatrix(Asura_Matrix_3x3& xMat) const;

		// Functions to help culling
		virtual void SetEntityCurrentlyMoving(bool bMoving) {};

			
		// Kevin Lea - 30/01/2002 15:14:10 - positional data
#ifdef ASURA_NO_NETWORK_LAG_COMPENSATION
		float			    m_fLastReportedServerTimestamp;
		Asura_Euler_3		m_xCurrentAngularVelocity;
#else
		// (MH - 07.05.2003) position smoothing is more complex now so it's in its own class.
		class Asura_PosAndOriInterpolator* m_pxPosAndOriInterpolator;

        // (MP - 26.11.2002) Made these mutable so that GetCurrentPosition()
        // and GetCurrentOrientation() can be made const.
        mutable unsigned int    m_uCurrentPositionFrameStamp;
        mutable unsigned int    m_uCurrentOrientationFrameStamp;
#endif //ASURA_NO_NETWORK_LAG_COMPENSATION
        mutable Asura_Vector_3   m_xCurrentPosition;
        mutable Asura_Matrix_3x3 m_xCurrentOrientation;
		mutable Asura_Vector_3	 m_xCurrentVelocity;

        // Richard May - 05/08/2002 14:24:47 - bit field for rendering flags
        //  Using this bit-field for culling purposes too, now.aw.
        u_char m_bIgnoreVisibilityChecks              	 : 1, // ignores the module visibility when rendering (so will always render)
               m_bIsEntityCurrentlyMoving             	 : 1, // Stores whether entity is moving or still. For culling.
               m_bEntityAlwaysMoves                   	 : 1, // Make true if the entity will always be moving (e.g bullet).
               m_bRemoveFromRenderableListIfInvisible 	 : 1, // Determines whether or not the SetInvisible...() methods add/remove this entity from the renderable list as required.
                                                         : 0;

private :
		union
		{
			//Invisibility flags... object is invisible if any of them are set. RAR 31/03/2006
			unsigned char m_ucInvisibleFlags;
			struct
			{
				unsigned char	m_bInvisibleGeneral:1, //The old generaly purpose invisibility flag
								m_bInvisibleByPhysicalFlag:1,//Hidden as a result of the physical object flag (that can be set by artists)
								m_bInvisibleByCutscene:1,//Hidden because a cutscene has said so.
								m_bInvisibleByDeactivation:1,//hidden because entity was deactivated
								:0;
			};
		};
		
public:
	/* static members */
	static Asura_Entity_Renderable* FindEntity(Asura_Guid uGuid) { return s_xList.FindEntity(uGuid); }
	static Asura_Classified_List<Asura_Entity_Renderable>* GetList() { return s_xList.GetList(); }
	static Asura_Classified_List<Asura_Entity_Renderable>* GetRenderableList() { return &s_xRenderableList; }
	
	// accessor to the line of sight list. Adding and removing from the list is done via the protected member fns.
	static Asura_Classified_List<Asura_Entity_Renderable>* GetLineOfSightList() { return &s_xLineOfSightList; } 
	
	// Each time an entity is added to / removed from the LoS list, this update stamp is incremented. 
	// This provides a simple way for any other systems relying on the list (e.g. spatial tree of
	// objects) to know whether there have been any changes without breaking OO behaviour (e.g. the 
	// list code setting a flag in a different system to say it's changed, or vice versa).
	static u_int GetLineOfSightListUpdateStamp( void ) { return s_uLineOfSightUpdateStamp; }

	static u_int GetNumRenderable(void)
	{
		return(s_xRenderableList.GetNumEntries());
	}

	
private:
	/* static data */
	static Asura_Combined_List<Asura_Entity_Renderable> s_xList;

	// KOR - 23/03/2006 - only entities in this list have Render() called on them.
	// Mostly just iterated through, so use the classified rather than combined list.
	static Asura_Classified_List<Asura_Entity_Renderable> s_xRenderableList;

	// KJFL - 5 November 2007 - only entities in this list will be queried in LoS 
	// tests versus entities. Entities are added/removed from the list by the 
	// protected methods in this class.
	static Asura_Classified_List<Asura_Entity_Renderable> s_xLineOfSightList;

	// see accessor comment above
	static u_int s_uLineOfSightUpdateStamp;


	// ART - 18/04/06 - List manipulation functions are now used to hide the
	// extra work the streaming system does with entities...
	void AddToList(void);
	void RemoveFromList(void);
};


// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_ENTITY_RENDERABLE_H___
