//////////////////////////////////////////////////////////////////////////////
//
// Asura_PosAndOriInterpolator.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     class to manage interpolating position and orientation for entities.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Asura_PosAndOriInterpolator.h"

#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Dynamics/Asura_Collision_Object.h"
#include "Network/Asura_Network.h"
#include "Mesh/Asura_Mesh_PolygonFlags.h"
#include "Physics/Asura_Physics_MovementQuery.h"

// (MH - 07.05.2003) About the time lag: the idea is to make all client 
// entities run a short time (e.g. 0.1s) behind the latest server update. This 
// helps reduce issues of latency, and ensures that interpolated positions 
// are between positions which were validated on the server. See Bernier's 
// article on Half-Life's networking in GDC 2001 p.73. There are useful 
// articles in 'Networked Virtual Environments' (Singhal/Zyda) and in 'Massively 
// Multiplayer Game Development' (ed. Alexander, pub. Charles River Media).

DEBUGVAR bool dbg_bPrintPosInput = false;
DEBUGVAR bool dbg_bPrintPosOutput = false;
//DEBUGVAR bool dbg_bPrintPosInput = false;
u_int g_uCurGuidForTest = 0;
DEBUGVAR u_int dbg_uTestGuid = ASURA_GUID_UNREGISTERED;

// if more time than dbg_fMaxTimeBetweenServerUpdates goes by, entity is 
// assumed to have stopped moving. Although updates are sent every 0.1s, some
// do not arrive due to unreliable UDP messaging.
DEBUGVAR float dbg_fMaxTimeBetweenServerUpdates = 0.5f;
DEBUGVAR float dbg_fMaxExtrapolationTimeNoDynamics = 0.1f;	// time allowed beyond last validated pos if we have no dynamics information
DEBUGVAR float dbg_fMaxExtrapolationTimeWithDynamics = 0.5f;// time allowed beyond last validated pos if we are able to do collisions
DEBUGVAR float dbg_fMaxExtrapolationTimeProjectile = 0.5f;
DEBUGVAR float dbg_fPosLookaheadTime = 0.2f;
DEBUGVAR float dbg_fRecentlyUsedPositionTime = 0.2f;	// if position was used within this period, splines try to interpolate from the position.
DEBUGVAR bool dbg_bPrintExtrapolateWarning = false;
DEBUGVAR bool dbg_bDisableInternetPosPrediction = false;

static const int iNO_POSITIONS = -2;
static const int iONE_POSITION = -1;

static bool UseInternetPositionPrediction()
{
	if (dbg_bDisableInternetPosPrediction)
	{
		return false;
	}
	return Asura_Network::IsPlayingOnInternet();
}

static void OutputFrameAndTime()
{
	Asura_OutputToDebugger(("frame %d: time %.2f. ", Asura_Timers::GetFrameNumber(), Asura_Maths::FtoD(Asura_Timers::GetServerAbsoluteTime())));
}

static bool ShouldPrintInput()
{
	return (dbg_bPrintPosInput && (dbg_uTestGuid == g_uCurGuidForTest));
}

static bool ShouldPrintOutput()
{
	return (dbg_bPrintPosOutput && (dbg_uTestGuid == g_uCurGuidForTest));
}

// TODO - move to Asura_Maths.h



#if _DEBUG
u_int Asura_PosAndOriInterpolator::s_uNumPosAndOriInterpolators = 0;
#endif

Asura_PosAndOriInterpolator::Asura_PosAndOriInterpolator()
: m_iLatestEntry(iNO_POSITIONS)
, m_fProcessNewDataIncrement(fASURA_NETWORK_TIME_BETWEEN_ENTITY_UPDATES)
, m_xCurrentPos(Asura_Maths::ZeroVector3)
, m_xCurrentVel(Asura_Maths::ZeroVector3)
, m_fCurPosTime(0.f)
, m_eTimeLagType(Asura_Entity_Renderable::TimeLag_NetworkEntity)
, m_pxSpline(NULL)
, m_fSplineStart(0.f)
, m_fSplineEnd(0.f)
, m_pxCollisionObject(NULL)
, m_iDynamicsHandleToIgnore(ASURA_INVALID_HANDLE)
{
	// JCT - init sample list
	for (u_int uSampleNo = 0; uSampleNo < uTOTALENTRIES; uSampleNo++)
	{
		m_axEntries[uSampleNo].m_xPosition.Zero();
		m_axEntries[uSampleNo].m_xQuat.Identity();
		m_axEntries[uSampleNo].m_fServerTime = 0.0f;
	}
#if _DEBUG
	++s_uNumPosAndOriInterpolators;
#endif
}


Asura_PosAndOriInterpolator::~Asura_PosAndOriInterpolator()
{
#if _DEBUG
	--s_uNumPosAndOriInterpolators;
#endif

	if (m_pxSpline)
	{
		delete m_pxSpline;
	}
}

void Asura_PosAndOriInterpolator::GetCurrentPosition(Asura_Vector_3& xOutputPos, Asura_Vector_3& xOutputVel) const
{
	// check if there is any data to interpolate between
	Asura_Assert(m_iLatestEntry != iNO_POSITIONS, "no positions stored yet");
	if (m_iLatestEntry == iONE_POSITION)
	{
		m_xCurrentPos = m_axEntries[0].m_xPosition;
		m_xCurrentVel.Zero();
		m_fCurPosTime = Asura_Timers::GetServerAbsoluteTime();
		xOutputPos = m_xCurrentPos;
		xOutputVel = m_xCurrentVel;
		if (dbg_bPrintPosOutput && (g_uCurGuidForTest == dbg_uTestGuid))
		{
			Asura_OutputToDebugger(("%d: pos (%.2f,%.2f,%.2f), vel %.2f m/s (%.2f,%.2f,%.2f). single entry.\n", Asura_Timers::GetFrameNumber(), Asura_Maths::FtoD(xOutputPos.x), Asura_Maths::FtoD(xOutputPos.y), Asura_Maths::FtoD(xOutputPos.z), Asura_Maths::FtoD(xOutputVel.Magnitude()), Asura_Maths::FtoD(xOutputVel.x), Asura_Maths::FtoD(xOutputVel.y), Asura_Maths::FtoD(xOutputVel.z)));
		}
		return;
	}

	float fMaxExtrapolationTime;
	if (!UseInternetPositionPrediction())// not on internet - might as well use low lag.
	{
		fMaxExtrapolationTime = dbg_fMaxExtrapolationTimeNoDynamics;
	}
	else
	{
		switch (m_eTimeLagType)
		{
			case Asura_Entity_Renderable::TimeLag_LocalPlayer:
			{
				fMaxExtrapolationTime = m_pxCollisionObject ? dbg_fMaxExtrapolationTimeWithDynamics : dbg_fMaxExtrapolationTimeNoDynamics;
				// small lag so can update at less than framerate and interpolate.
				break;
			}
			
			default:
				Asura_Assert(false, "unknown time lag type");
				//fallthrough
			case Asura_Entity_Renderable::TimeLag_NetworkEntity:
			{
				fMaxExtrapolationTime = m_pxCollisionObject ? dbg_fMaxExtrapolationTimeWithDynamics : dbg_fMaxExtrapolationTimeNoDynamics;
				break;
			}

			case Asura_Entity_Renderable::TimeLag_Projectile:
			{
				fMaxExtrapolationTime = dbg_fMaxExtrapolationTimeProjectile;
				break;
			}
		}
	}

	float fCurrentTime = Asura_Timers::GetServerAbsoluteTime();

	if ( !UseInternetPositionPrediction() )
	{
		GetPosAndVel(m_xCurrentPos, m_xCurrentVel, fCurrentTime, fMaxExtrapolationTime);
		m_fCurPosTime = Asura_Timers::GetServerAbsoluteTime();
	}
	else
	{
		bool bDone = false;
		if (m_pxSpline)
		{
			if (m_fSplineStart <= fCurrentTime && fCurrentTime <= m_fSplineEnd)
			{
				float fSplineLength = m_fSplineEnd - m_fSplineStart;
				float fSplineTime = (fCurrentTime - m_fSplineStart) / fSplineLength;
				m_pxSpline->GetPosition(fSplineTime, m_xCurrentPos);
				m_pxSpline->GetVelocity(fSplineTime, m_xCurrentVel);
				// scale velocity back into normal time
				m_xCurrentVel /= dbg_fPosLookaheadTime;
				// (MH - 25.06.2004) compensate for game time scale.
				float fGameTimeScale = Asura_Timers::GetGameTimeScale();
				if (fGameTimeScale > 0.001f)
				{
					m_xCurrentVel *= 1.0f / fGameTimeScale;
				}
				m_fCurPosTime = Asura_Timers::GetServerAbsoluteTime();
				bDone = true;

				if (ShouldPrintOutput())
				{
					OutputFrameAndTime();
					Asura_OutputToDebugger(("output from spline (%.2f): pos = { %.2f, %.2f, %.2f }, speed = %.2f, vel = { %.2f, %.2f, %.2f }.\n", 
						Asura_Maths::FtoD(fSplineTime),
						Asura_Maths::FtoD(m_xCurrentPos.x), Asura_Maths::FtoD(m_xCurrentPos.y), Asura_Maths::FtoD(m_xCurrentPos.z), 
						Asura_Maths::FtoD(m_xCurrentVel.Magnitude()), 
						Asura_Maths::FtoD(m_xCurrentVel.x), Asura_Maths::FtoD(m_xCurrentVel.y), Asura_Maths::FtoD(m_xCurrentVel.z) ));
				}
			}
		}

		if (!bDone)
		{
			GetPosAndVel(m_xCurrentPos, m_xCurrentVel, fCurrentTime, fMaxExtrapolationTime);
			m_fCurPosTime = Asura_Timers::GetServerAbsoluteTime();
		}
	}

	xOutputPos = m_xCurrentPos;
	xOutputVel = m_xCurrentVel;
}

// returns whether we are moving. If we are and pxVelOutput is not NULL also write in the velocity.
void Asura_PosAndOriInterpolator::GetPosAndVel(Asura_Vector_3& xPosOutput, Asura_Vector_3& xVelOutput, float fTime, float fMaxExtrapolation) const
{
	Asura_Assert(m_iLatestEntry != iNO_POSITIONS, "no positions stored yet");
	if (m_iLatestEntry < 0)
	{
		xPosOutput = m_axEntries[0].m_xPosition;
		xVelOutput.Zero();
		return;
	}

	// find a suitable set of stored data, ideally we want times either side of the current time.
	bool bExtrapolating = false;	// set to true if we find that we are. Do some basic collisions if so.

	const Asura_PosAndOriArrayEntry& xLatest = m_axEntries[m_iLatestEntry];
	const Asura_PosAndOriArrayEntry& xPrev = m_axEntries[1 - m_iLatestEntry];

	const Asura_Vector_3& xLastReportedPosition = xLatest.m_xPosition;
	const Asura_Vector_3& xPrevReportedPosition = xPrev.m_xPosition;
	float fLastReportedServerTimestamp = xLatest.m_fServerTime;
	float fPrevReportedServerTimestamp = xPrev.m_fServerTime;

	if (fLastReportedServerTimestamp < fTime)
	{
		// we're extrapolating from the end of the data - clamp how far below.
		bExtrapolating = true;
	}
	else if (fPrevReportedServerTimestamp <= fTime)
	{
		// these times are either side of the required time - good!
	}
	else
	{
		// asked for a time that predates the entry history.
		// should we limit how much back-extrapolation is done?
		if (ShouldPrintOutput() || ShouldPrintInput())
		{
			OutputFrameAndTime();
			Asura_OutputToDebugger(("back extrapolation: %.2f < %.2f\n", Asura_Maths::FtoD(fTime), Asura_Maths::FtoD(fPrevReportedServerTimestamp)));
		}
	}

	// fTimeDelta2 is the time between the last update and the one before. If 
	// nothing changed, no update would have been sent after the first frame we 
	// stopped moving. If that update was lost but we haven't had another one 
	// for a while we can guess that we stopped moving.
	float fTimeDelta2 = fLastReportedServerTimestamp - fPrevReportedServerTimestamp;
	if (fTimeDelta2 > dbg_fMaxTimeBetweenServerUpdates)
	{
		xPosOutput = xLastReportedPosition;
		xVelOutput.Zero();
		return;
	}

	// fTimeDelta is the time elapsed since the last update was sent (approx, 
	// since m_fLastReportedClientTimestamp is calculated using latency / 2).
	// Then the time lag is removed to get the time the data is valid for.
	float fTimeDelta = fTime - fLastReportedServerTimestamp;

	// If fTimeDelta < 0 we are interpolating between the last 2 reported positions.
	// If fTimeDelta == 0, we are at the time the last server update was intended for.
	// If fTimeDelta > 0 we are extrapolating the position into the future.
	if (fTimeDelta > fMaxExtrapolation) 
	{
		// (MH - 10.03.2004) want to see how often this fires...
		if (dbg_bPrintExtrapolateWarning || (ShouldPrintOutput()))
		{
			Asura_OutputToDebugger(("extrapolating too far (%.2f) into future.\n", Asura_Maths::FtoD(fTimeDelta - fMaxExtrapolation)));
		}

		fTimeDelta = fMaxExtrapolation;
	}


	if (fTimeDelta2 > 0.0f)
	{
		const Asura_Vector_3 xPrevToLast(xLastReportedPosition - xPrevReportedPosition);
		float fTimeRatio = fTimeDelta / fTimeDelta2;
		Asura_Vector_3 xDelta = xPrevToLast * fTimeRatio;

		// (MH - 16.04.2004) this used to be computed from actual distance moved, but it's 
		// not reliable to divide by fTimeDelta since it could be near (and reportedly 
		// sometimes equal to) 0. So we compute it from the positions we're using.
		xVelOutput = xPrevToLast / fTimeDelta2;

		// (MH - 25.06.2004) compensate for game time scale.
		float fGameTimeScale = Asura_Timers::GetGameTimeScale();
		if (fGameTimeScale > 0.001f)
		{
			xVelOutput *= 1.0f / fGameTimeScale;
		}

		// if extrapolating, check new pos is valid, or move to a valid one.
		if (bExtrapolating && m_pxCollisionObject)
		{
			if (xDelta.MagnitudeSquared() != 0.0f)
			{
				const Asura_AVector3 xStartPos( xLastReportedPosition );
				const Asura_AVector3 xDisplacement( xDelta );
				const Asura_OBB xOBB( m_pxCollisionObject->GetBoundingBox() );
	
				Asura_Physics_MovementQuery xQuery( xStartPos, xDisplacement, xOBB, Asura_Physics_MovementQuery::ASURA_PMQCP_OBB, (u_short)AsuraPolygonFlag_NoCollisions, false );
				Asura_Physics_MovementQuery_Results xResult( xQuery );
				
				xPosOutput = xResult.GetEndPosition();
			}
		}
		else
		{
			xPosOutput = xLastReportedPosition + xDelta;
		}

		if  (ShouldPrintOutput())
		{
			OutputFrameAndTime();
			Asura_OutputToDebugger(("output: last = {%.2f,%.2f,%.2f}, prev = {%.2f,%.2f,%.2f}, speed = %.2f, td1 = %.2f, td2 = %.2f.\n", Asura_Maths::FtoD(xLastReportedPosition.x), Asura_Maths::FtoD(xLastReportedPosition.y), Asura_Maths::FtoD(xLastReportedPosition.z), Asura_Maths::FtoD(xPrevReportedPosition.x), Asura_Maths::FtoD(xPrevReportedPosition.y), Asura_Maths::FtoD(xPrevReportedPosition.z), Asura_Maths::FtoD(xVelOutput.Magnitude()), Asura_Maths::FtoD(fTimeDelta), Asura_Maths::FtoD(fTimeDelta2)));
		}
	}
	else
	{
		Asura_Assert(false, "out-of-order or duplicate entry");
		xPosOutput = xLastReportedPosition;
		xVelOutput.Zero();
	}
}


void Asura_PosAndOriInterpolator::GetCurrentOrientation(Asura_Matrix_3x3& xOutputMat) const
{
	Asura_Assert(m_iLatestEntry != iNO_POSITIONS, "no positions stored yet");
	if (m_iLatestEntry < 0)
	{
		m_axEntries[0].m_xQuat.ToMatrix(xOutputMat);
		return;
	}

	// find a suitable set of stored data, ideally we want times either side of the current time.
	float fCurrentTime = Asura_Timers::GetServerAbsoluteTime();	// find a suitable set of stored data, ideally we want times either side of the current time.

	const Asura_PosAndOriArrayEntry& xLatest = m_axEntries[m_iLatestEntry];
	const Asura_PosAndOriArrayEntry& xPrev = m_axEntries[1 - m_iLatestEntry];

	// check if timelag is bigger than entry history. no timelag in orientation 
	// so shouldn't happen, but variations in global time mean we can receive
	// packets timestamped in advance of our estimate of global time..
	// (MH - 15.07.2004) disabled this code, it dates back from when we had a bigger 
	// entry history and seems to be causing discontinuities now. This is the same case
	// as back-extrapolation in the position calculation.
//	if (xPrev.m_fServerTime > fCurrentTime)
//	{
//		// use the oldest known orientation.
//		xPrev.m_xQuat.ToMatrix(xOutputMat);
//		return;		
//	}

	const Asura_Quat& xLastReportedOrientation = xLatest.m_xQuat;
	const Asura_Quat& xPrevReportedOrientation = xPrev.m_xQuat;
	float fLastReportedServerTimestamp = xLatest.m_fServerTime;
	float fPrevReportedServerTimestamp = xPrev.m_fServerTime;

	// fTimeDelta is the time elapsed since the last update was sent (approx, 
	// since m_fLastReportedClientTimestamp is calculated using latency / 2).
	// Then the time lag is removed to get the time the data is valid for.
	float fTimeDelta = fCurrentTime - fLastReportedServerTimestamp;

	// fTimeDelta2 is the time between the last update and the one before. If 
	// nothing changed, no update would have been sent after the first frame we 
	// stopped moving. If that update was lost but we haven't had another one 
	// for a while we can guess that we stopped moving.
	float fTimeDelta2 = fLastReportedServerTimestamp - fPrevReportedServerTimestamp;
	if (fTimeDelta2 > dbg_fMaxTimeBetweenServerUpdates)
	{
		xLastReportedOrientation.ToMatrix(xOutputMat);
		return;
	}
	
	// If fTimeDelta < 0 we are interpolating between the last 2 reported positions.
	// If fTimeDelta == 0, we are at the time the last server update was intended for.
	// If fTimeDelta > 0 we are extrapolating the position into the future.
	// (MH - 23.08.2003) for orientation, use 0
	if (fTimeDelta > 0) 
	{
		fTimeDelta = 0;
	}

	Asura_Quat xCurrentOrientQuat = xLastReportedOrientation;

	Asura_Assert(fTimeDelta2 > 0.0f, "out-of-order or duplicate entry");

	if (fTimeDelta2 > 0.0f)
	{
		float fTimeRatio = fTimeDelta / fTimeDelta2;
		xCurrentOrientQuat.Slerp(xPrevReportedOrientation, xLastReportedOrientation, 1.0f+fTimeRatio);
	}

	xCurrentOrientQuat.ToMatrix(xOutputMat);
}


void Asura_PosAndOriInterpolator::GetCurrentAngularVelocity(Asura_Euler_3& xOutputEuler) const
{
	if (m_iLatestEntry < 0)
	{
		// less than 2 valid entries.
		xOutputEuler.Zero();
	}
	else
	{
		// get the 2 previous entries - 0 is most recent, 1 is next-most recent
		const Asura_PosAndOriArrayEntry* pxEntry0 = &m_axEntries[m_iLatestEntry];
		const Asura_PosAndOriArrayEntry* pxEntry1 = &m_axEntries[1 - m_iLatestEntry];
		float fTimeDifference = pxEntry0->m_fServerTime - pxEntry1->m_fServerTime;
		Asura_Assert(fTimeDifference != 0.0f, "2 orientation entries with same timestamp");
		Asura_Assert(fTimeDifference > 0.0f, "2 orientation entries with out-of-order timestamps");

		if (fTimeDifference > dbg_fMaxTimeBetweenServerUpdates)
		{
			xOutputEuler.Zero();
		}
		else
		{
			Asura_Euler_3 xEuler0;
			Asura_Euler_3 xEuler1;

			pxEntry0->m_xQuat.ToEuler(xEuler0);
			pxEntry1->m_xQuat.ToEuler(xEuler1);
			
			xOutputEuler = (xEuler0 - xEuler1);
			// assume velocity is within +/- 180 deg per frame, so shift into range
			if (xOutputEuler.x < -Asura_Maths::Deg180) xOutputEuler.x += Asura_Maths::Deg360;
			else if (xOutputEuler.x > Asura_Maths::Deg180) xOutputEuler.x -= Asura_Maths::Deg360;
			if (xOutputEuler.y < -Asura_Maths::Deg180) xOutputEuler.y += Asura_Maths::Deg360;
			else if (xOutputEuler.y > Asura_Maths::Deg180) xOutputEuler.y -= Asura_Maths::Deg360;
			if (xOutputEuler.z < -Asura_Maths::Deg180) xOutputEuler.z += Asura_Maths::Deg360;
			else if (xOutputEuler.z > Asura_Maths::Deg180) xOutputEuler.z -= Asura_Maths::Deg360;

			xOutputEuler /= fTimeDifference;

			// (MH - 25.06.2004) compensate for game time scale.
			float fGameTimeScale = Asura_Timers::GetGameTimeScale();
			if (fGameTimeScale > 0.001f)
			{
				xOutputEuler *= 1.0f / fGameTimeScale;
			}

		}
	}
}


void Asura_PosAndOriInterpolator::ProcessNewData(const Asura_Vector_3& xPosition, 
												const Asura_Quat& xOrientation,
												float fServerTimestamp,
												const Asura_Vector_3* pxActualVel /* = NULL */)
{
	// add time lag so the time we use this data for is the time the next update is due.
	fServerTimestamp += m_fProcessNewDataIncrement;

	// log to debugger?
	if (ShouldPrintInput())
	{
		OutputFrameAndTime();
		Asura_OutputToDebugger(("input for %.2f: { %.2f, %.2f, %.2f }\n", Asura_Maths::FtoD(fServerTimestamp), Asura_Maths::FtoD(xPosition.x), Asura_Maths::FtoD(xPosition.y), Asura_Maths::FtoD(xPosition.z)));
	}

	if (m_iLatestEntry == iNO_POSITIONS)
	{
		m_axEntries[0].m_xPosition = xPosition;
		m_axEntries[0].m_xQuat = xOrientation;
		m_axEntries[0].m_fServerTime = fServerTimestamp;
		m_iLatestEntry = iONE_POSITION;

		// ART - 01/08/2007 - We're getting no-normalised quaternions on PSP
		// due to the shear number of quat->matrix conversions...
		m_axEntries[0].m_xQuat.Normalise();
	}
	else if (m_iLatestEntry == iONE_POSITION)
	{
		// check this update is newer
		if (m_axEntries[0].m_fServerTime < fServerTimestamp)
		{
			// validate the previously entry so the code below will add the new one
			m_iLatestEntry = 0;
		}
	}

	if (m_iLatestEntry >= 0)
	{

		if (fServerTimestamp <= m_axEntries[m_iLatestEntry].m_fServerTime)
		{
			// probably duplicate or out-of-order packet.
			// could also be flaw in entity update code which is sending 2 positions in the same frame
			// ignore that case for now.
			if (ShouldPrintInput())
			{
				Asura_OutputToDebugger(("out of order packet %.2f <= %.2f\n", Asura_Maths::FtoD(fServerTimestamp), Asura_Maths::FtoD(m_axEntries[m_iLatestEntry].m_fServerTime)));
			}
			return;
		}

		m_iLatestEntry = 1 - m_iLatestEntry;

		Asura_PosAndOriArrayEntry& xLatest = m_axEntries[m_iLatestEntry];
		xLatest.m_xPosition = xPosition;
		xLatest.m_xQuat = xOrientation;
		xLatest.m_fServerTime = fServerTimestamp;

		// ART - 01/08/2007 - We're getting no-normalised quaternions on PSP
		// due to the shear number of quat->matrix conversions...
		xLatest.m_xQuat.Normalise();

		const Asura_PosAndOriArrayEntry& xPrev = m_axEntries[1 - m_iLatestEntry];
		float fPrevTimeStamp = xPrev.m_fServerTime;
		// does it look like we're moving? check for a long time between this and the
		// last update, and compare the values as well.
		float fTimeDelta = fServerTimestamp - fPrevTimeStamp;
		if ( (fTimeDelta > dbg_fMaxTimeBetweenServerUpdates) || 
			 (xLatest.m_xPosition.ApproximatelyEquals(xPrev.m_xPosition) && xLatest.m_xQuat.ApproximatelyEquals(xPrev.m_xQuat) ) )
		{
			// probably not moving
			m_axEntries[0] = xLatest;
			m_iLatestEntry = iONE_POSITION;
			if (ShouldPrintInput())
			{
				if (fTimeDelta > dbg_fMaxTimeBetweenServerUpdates)
				{
					Asura_OutputToDebugger(("long time between updates (%.2fs)\n", Asura_Maths::FtoD(fTimeDelta)));
				}
				else
				{
					Asura_OutputToDebugger(("not moving.\n"));
				}
			}
			return;
		}

		// we're moving, set up a cubic spline
		// TODO: store the parameters and compute the spline only when it's needed.
		// TODO: don't do this in single-player?
		if (IsUsingCubicSplineInterpolation())
		{
			// current pos and velocity are the last used values.
			// if they weren't used recently we can bink to the correct position.
			if ( (Asura_Timers::GetServerAbsoluteTime() - m_fCurPosTime) > dbg_fRecentlyUsedPositionTime)
			{
				m_axEntries[0] = xLatest;
				m_iLatestEntry = iONE_POSITION;
				if (ShouldPrintInput())
				{
					Asura_OutputToDebugger(("bink to position.\n"));
				}
			}
			else
			{
				
				// Within lookahead time (0.2s), get us on the correct path. To do this, we create 
				// a spline starting at the current position and ending at the position we 
				// would predict using just the most recent data from the server. This means
				// we use the current velocity (if known, or calculated from the last 2 points)
				// as the basis to move from the latest reported position to the position we will
				// be at in 0.2s. i.e. moving at current velocity for the difference between that 
				// time and the reported time.
				Asura_Vector_3 xPredictedVel;
				if (pxActualVel)
				{
					xPredictedVel = *pxActualVel;
				}
				else
				{
					Asura_Assert(fTimeDelta > 0.0f, "2 updates from same server time");
					xPredictedVel = (xPosition - xPrev.m_xPosition) / fTimeDelta;
				}
				
				float fCurrentTime = m_fCurPosTime;
				float fPredictedTime = fCurrentTime + dbg_fPosLookaheadTime;	// 0.2 sec in future
				Asura_Vector_3 xPredictedPos;
				float fPredictionTimeDelta = fPredictedTime - fServerTimestamp;
				//Asura_Assert(fPredictionTimeDelta > 0.0f && fPredictionTimeDelta < dbg_fMaxTimeBetweenServerUpdates, "prediction error");
				if (fPredictionTimeDelta > 0.0f && fPredictionTimeDelta < dbg_fMaxTimeBetweenServerUpdates)
				{
					xPredictedPos = xPosition + xPredictedVel * fPredictionTimeDelta;
				}
				else
				{
                    if (dbg_bPrintExtrapolateWarning)
					{
						Asura_OutputToDebugger(( "prediction time delta out of range: %.2f.\n", Asura_Maths::FtoD(fPredictionTimeDelta) ));
					}
					xPredictedPos = xPosition;
				}
				// must pass velocity in the spline's time-space where time goes from 0-0.2
				CalculateCubicSpline(m_xCurrentPos, xPredictedPos, m_xCurrentVel * dbg_fPosLookaheadTime, xPredictedVel * dbg_fPosLookaheadTime, fCurrentTime, fPredictedTime);

				if (ShouldPrintInput())
				{
					Asura_OutputToDebugger(("  spline for %.2f->%.2f from pos0 = { %.2f, %.2f, %.2f }, pos1 = { %.2f, %.2f, %.2f }, vel0 = { %.2f, %.2f, %.2f }, vel1 = { %.2f, %.2f, %.2f }. \n", 
						Asura_Maths::FtoD(fCurrentTime), Asura_Maths::FtoD(fPredictedTime),
						Asura_Maths::FtoD(m_xCurrentPos.x), Asura_Maths::FtoD(m_xCurrentPos.y), Asura_Maths::FtoD(m_xCurrentPos.z),
						Asura_Maths::FtoD(xPredictedPos.x), Asura_Maths::FtoD(xPredictedPos.y), Asura_Maths::FtoD(xPredictedPos.z),
						Asura_Maths::FtoD(m_xCurrentVel.x), Asura_Maths::FtoD(m_xCurrentVel.y), Asura_Maths::FtoD(m_xCurrentVel.z),
						Asura_Maths::FtoD(xPredictedVel.x), Asura_Maths::FtoD(xPredictedVel.y), Asura_Maths::FtoD(xPredictedVel.z) ));
				}
			}
		}
	}
}

// moving if last reported positions or orientations differ. This is used e.g. 
// by the rendering code to see whether a shape instance needs updating.
bool Asura_PosAndOriInterpolator::IsMoving() const
{
	if (m_iLatestEntry>= 0)
	{
		if (m_axEntries[0].m_xPosition != m_axEntries[1].m_xPosition)
		{
			return true;
		}
		
		if( !m_axEntries[0].m_xQuat.ApproximatelyEquals(m_axEntries[1].m_xQuat,0.0f) )
		{
			return true;
		}
		
		return false;
	}
	// fewer than 2 valid entries, assume moving... 
	// (MH - 02.04.2004) not sure why, but don't want to risk changing it at the moment.
	return true;
}


void Asura_PosAndOriInterpolator::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	InvalidateCache();
	float fFakeServerTime = Asura_Timers::GetServerAbsoluteTime();
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientation);
	ProcessNewData(xPosition, xQuat, fFakeServerTime);
}


const Asura_Vector_3& Asura_PosAndOriInterpolator::GetLastReportedPosition() const
{
	if (m_iLatestEntry == iNO_POSITIONS)
	{
		Asura_Assert(false, "no positions stored");
		return Asura_Maths::ZeroVector3;
	}
	else if (m_iLatestEntry == iONE_POSITION)
	{
		return m_axEntries[0].m_xPosition;
	}
	else
	{
		return m_axEntries[m_iLatestEntry].m_xPosition;
	}
}


const Asura_Quat& Asura_PosAndOriInterpolator::GetLastReportedOrientation() const
{
	if (m_iLatestEntry == iNO_POSITIONS)
	{
		Asura_Assert(false, "no orientations stored");
		return Asura_Maths::IdentityQuat;
	}
	else if (m_iLatestEntry == iONE_POSITION)
	{
		return m_axEntries[0].m_xQuat;
	}
	else
	{
		return m_axEntries[m_iLatestEntry].m_xQuat;
	}
}

float Asura_PosAndOriInterpolator::GetLastReportedServerTime() const
{
	if (m_iLatestEntry == iNO_POSITIONS)
	{
		return 0.0f;
	}
	else if (m_iLatestEntry == iONE_POSITION)
	{
		return m_axEntries[0].m_fServerTime;
	}
	else
	{
		return m_axEntries[m_iLatestEntry].m_fServerTime;
	}
}


void Asura_PosAndOriInterpolator::InitialiseMovingObject(const Asura_Vector_3& xPosition, 
														 const Asura_Quat& xOrientQuat, 
														 const Asura_Vector_3& xVelocity,
														 float fServerTime)
{
	InvalidateCache();

	float fTimeBetweenUpdates = ASURA_NETWORK_TIME_BETWEEN_ENTITY_UPDATES;
	float fServerTime0 = fServerTime - fTimeBetweenUpdates;
	float fServerTime1 = fServerTime;
	
	// (MH - 08.06.2004) now correcting for game time scale
	Asura_Vector_3 xPrevPos = xPosition - (xVelocity * (fTimeBetweenUpdates * Asura_Timers::GetGameTimeScale()));
	ProcessNewData(xPrevPos, xOrientQuat, fServerTime0);
	ProcessNewData(xPosition, xOrientQuat, fServerTime1);
}


void Asura_PosAndOriInterpolator::SetUseCubicSplineInterpolation(bool bEnable)
{
	if (bEnable)
	{
		if (!m_pxSpline)
		{
			m_pxSpline = new Asura_HermiteSpline;
		}
	}
	else
	{
		if (m_pxSpline)
		{
			delete m_pxSpline;
			m_pxSpline = NULL;
		}
	}
}

void Asura_PosAndOriInterpolator::CalculateCubicSpline(const Asura_Vector_3& xPos0, 
													   const Asura_Vector_3& xPos1, 
													   const Asura_Vector_3& xVel0, 
													   const Asura_Vector_3& xVel1, 
													   float fStartTime, 
													   float fEndTime)
{
	Asura_Assert(m_pxSpline, "spline not created yet");

	m_pxSpline->Recalculate(xPos0, xPos1, xVel0, xVel1);

	m_fSplineStart = fStartTime;
	m_fSplineEnd = fEndTime;

	//TestSpline();
}


void Asura_PosAndOriInterpolator::InvalidateCache()
{
	m_iLatestEntry = iNO_POSITIONS;
	m_fSplineStart = 0.0f;
	m_fSplineEnd = 0.0f;
	m_fCurPosTime = 0.0f;
}
