//
// Asura_StaticMessageBlock.h
//
// Created by Richard Rice - 15/11/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//


#include "Core/Asura.h"
#include "Entity/Asura_StaticMessageBlock.h"

//asura
#include "ChunkSystem/Asura_Chunk_Entity.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Entity/Asura_StaticMessage_System.h"

//tools
#ifdef ASURA_TOOLS
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#endif


static const int iASURA_STATICMESSAGEBLOCK_VERSION = 0;
static const u_int uASURA_STATICMESSAGEBLOCK_CHUNKFLAG_HASTOOLSDATA = 0x00000001;

Asura_StaticMessageBlock::Asura_StaticMessageBlock()
{
	m_uNumMessages = 0;
	m_pxMessages = 0;
}


Asura_StaticMessageBlock::~Asura_StaticMessageBlock()
{
	Asura_ResourcePushType(ASURA_RESOURCE_STATIC_MESSAGE);
	for(u_int i=0;i<m_uNumMessages;i++)
	{
		m_pxMessages[i].DeallocateMessageData();	
	}

	Asura_ResourceDeallocateArray(m_pxMessages);
	Asura_ResourcePopType();
}

#ifdef ASURA_TOOLS
Asura_StaticMessageBlock& Asura_StaticMessageBlock::operator = (const Asura_StaticMessageBlock& xBlock)
{
	if(&xBlock != this)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_STATIC_MESSAGE);
		//delete existing messages
		for(u_int i=0;i<m_uNumMessages;i++)
		{
			m_pxMessages[i].DeallocateMessageData();	
		}
		Asura_ResourceDeallocateArray(m_pxMessages);

		//copy messages from other block
		m_uNumMessages = xBlock.m_uNumMessages;
		Asura_ResourceAllocateArray(m_pxMessages,m_uNumMessages);

		for(u_int i=0;i<m_uNumMessages;i++)
		{
			m_pxMessages[i].Copy(&xBlock.m_pxMessages[i]);
		}


		Asura_ResourcePopType();
	}
	return *this;
}
#endif //ASURA_TOOLS


void Asura_StaticMessageBlock::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	int iVersion;
	xStream >> iVersion;
	Asura_Assertf(iVersion <= iASURA_STATICMESSAGEBLOCK_VERSION,("Unexpected version %d for static message block.",iVersion));

	int iMessageVersion;
	xStream >> iMessageVersion;

	u_int uChunkFlags;
	xStream >> uChunkFlags;

	xStream >> m_uNumMessages;

	if(m_uNumMessages)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_STATIC_MESSAGE);
		Asura_ResourceAllocateArray(m_pxMessages,m_uNumMessages);
		Asura_ResourcePopType();
	}

	bool bHasToolsData = (uChunkFlags & uASURA_STATICMESSAGEBLOCK_CHUNKFLAG_HASTOOLSDATA) ? true : false;

	for(u_int i=0;i<m_uNumMessages;i++)
	{
		Asura_Chunk_StaticMessages::ReadSingleMessageFromChunkStream(iMessageVersion,xStream,&m_pxMessages[i],bHasToolsData);
	}


}

#ifdef ASURA_TOOLS
void Asura_StaticMessageBlock::SkipInChunkStream(Asura_Chunk_Stream& xStream)
{
	//Could possibly simplify this by reading into a dummy static message block and then ditching it.
	//But doing it this way saves some unecessary allocations when skipping through the chunk
	//RAR 17/04/2009
	int iVersion;
	xStream >> iVersion;
	Asura_Assertf(iVersion <= iASURA_STATICMESSAGEBLOCK_VERSION,("Unexpected version %d for static message block.",iVersion));

	int iMessageVersion;
	xStream >> iMessageVersion;

	u_int uChunkFlags;
	xStream >> uChunkFlags;

	u_int uNumMessages;
	xStream >> uNumMessages;
	
	bool bHasToolsData = (uChunkFlags & uASURA_STATICMESSAGEBLOCK_CHUNKFLAG_HASTOOLSDATA) ? true : false;

	for(u_int i=0;i<uNumMessages;i++)
	{
		Asura_Chunk_StaticMessages::SkipSingleMessageInChunkStream(iMessageVersion, xStream, bHasToolsData);
	}
}
#endif //ASURA_TOOLS



void Asura_StaticMessageBlock::ReadOldStaticMessageChunkFromChunkStream(Asura_Chunk_Stream& xStream, u_int uVersion, bool bHasToolsData)
{
	Asura_StaticMessage* pxMessage = m_pxMessages;

	Asura_ResourcePushType(ASURA_RESOURCE_STATIC_MESSAGE);
	for(u_int j=0;j<m_uNumMessages;j++,pxMessage++)
	{
		xStream	>> pxMessage->m_usMsgID;
		xStream >> pxMessage->m_usDataSize;
		xStream >> pxMessage->m_uSendTo;
		#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
		pxMessage->m_uSendTo = Asura_Streaming_System::MakeCurrentMasterZoneGuid(pxMessage->m_uSendTo);
		#endif

		if(uVersion >= 3)
		{
			xStream >> pxMessage->m_fDelay;
		}

		if(uVersion >= 5)
		{
			xStream	>> pxMessage->m_fProbability;
		}

	    if( uVersion >= 6 )
	    {
			xStream	>> pxMessage->m_uFlags;
	    }

		if(bHasToolsData && uVersion >= 4)
		{
			#ifdef ASURA_TOOLS
			xStream	>> pxMessage->m_uToolsFlags;
			#else
			//Not really expecting tools data in a non-tools build , since it is exlcuded from
			//level exports. But skip it just in case
			xStream.Skip( sizeof( u_int ) );
			#endif
		}


		if(pxMessage->m_usDataSize)
		{
			u_int uSizeInChunk = (pxMessage->m_usDataSize+3)&~3;
			Asura_ResourceAllocateMemoryBlock(pxMessage->m_pData, uSizeInChunk, ASURA_ALIGN_4_BYTE);
			xStream.Read( uSizeInChunk, pxMessage->m_pData );

			#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
			// This is necessary to ensure that unique GUIDs are available when streaming...
			FixGuidsForStreaming(*pxMessage);
			#endif
		}
		else
		{
			pxMessage->m_pData = 0;
		}
		#ifdef ASURA_TOOLS
		// Fix up the data for old versions
		Asura_Chunk_StaticMessages::FixData(pxMessage);
		#endif
	}
	Asura_ResourcePopType();

}

#ifdef ASURA_TOOLS
void Asura_StaticMessageBlock::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	u_int uChunkFlags = 0;
	bool bIncludeToolsData = false;
	if(!Tools_Level_Export::IsExport())
	{
		uChunkFlags |= uASURA_STATICMESSAGEBLOCK_CHUNKFLAG_HASTOOLSDATA;
		bIncludeToolsData = true;
	}


	//two version numbers , one for thr block , and one for the messages
	xStream << iASURA_STATICMESSAGEBLOCK_VERSION;
	xStream << Asura_Chunk_StaticMessages::GetCurrentMaskedVersion();
	xStream << uChunkFlags;

	u_int uNumMessagesToSave = m_uNumMessages;
	if(!bIncludeToolsData)
	{
		//Not saving disabled messages in exports , so determine how many messages will be saved
		for(u_int i=0;i<m_uNumMessages;i++)
		{
			if(m_pxMessages[i].IsDisabled()) uNumMessagesToSave--;
		}
	}
	
	xStream << uNumMessagesToSave;


	for(u_int i=0;i<m_uNumMessages;i++)
	{
		if(!bIncludeToolsData)
		{
			//skip disabled messages
			if(m_pxMessages[i].IsDisabled()) continue;
		}
		
		Asura_Chunk_StaticMessages::WriteSingleMessageToChunkStream(xStream,&m_pxMessages[i],bIncludeToolsData);
	}

}
#endif //ASURA_TOOLS

#if 0//test
void Asura_StaticMessageBlock::CheckSaveAndLoad() const
{
	for(u_int i=0;i<m_uNumMessages;i++)
	{
		//if(m_pxMessages[i].m_usMsgID >= ASURA_MSG_END_OF_STATIC_MESSAGES) continue;
		if(m_pxMessages[i].m_usMsgID == 108 || m_pxMessages[i].m_usMsgID == 32801) continue;
		if(m_pxMessages[i].m_usMsgID == 32777) continue;
		
		Asura_Chunk_Memory_Stream xStream;
		xStream.SetSwapEndian(false);

		Asura_Chunk_StaticMessages::WriteSingleMessageToChunkStream(xStream,&m_pxMessages[i],true);
		xStream.SetPosition(0);

		Asura_StaticMessage xLoadedMessage;
		Asura_Chunk_StaticMessages::ReadSingleMessageFromChunkStream(Asura_Chunk_StaticMessages::GetCurrentMaskedVersion(),xStream,&xLoadedMessage,true);

		
		Asura_Assertf(m_pxMessages[i] == xLoadedMessage, ("Difference after save and load\nMsg %d",xLoadedMessage.m_usMsgID));
		xLoadedMessage.DeallocateMessageData();
	}
}
#endif


Asura_StaticMessage* Asura_StaticMessageBlock::GetMessagePointer(u_int uIndex)
{
	Asura_Assert(uIndex<m_uNumMessages,"");
	return &m_pxMessages[uIndex];
}

const Asura_StaticMessage* Asura_StaticMessageBlock::GetMessagePointer(u_int uIndex) const
{
	Asura_Assert(uIndex<m_uNumMessages,"");
	return &m_pxMessages[uIndex];
}


void Asura_StaticMessageBlock::AddMessage(const Asura_StaticMessage& xMessage)
{
	Asura_ResourcePushType(ASURA_RESOURCE_STATIC_MESSAGE);
	{
		//increase array size
		Asura_StaticMessage* pxMessages;
		Asura_ResourceAllocateArray(pxMessages,m_uNumMessages+1);

		//copy old messages
		for(u_int i=0;i<m_uNumMessages;i++)
		{
			pxMessages[i] = m_pxMessages[i];
		}
		pxMessages[m_uNumMessages] = xMessage;

		//replace old array
		Asura_ResourceDeallocateArray(m_pxMessages);
		m_pxMessages = pxMessages;
		m_uNumMessages++;
	}
	Asura_ResourcePopType();

}

void Asura_StaticMessageBlock::DeleteMessage(u_int uIndex)
{
	Asura_Assert(uIndex < m_uNumMessages,"");

	
	m_pxMessages[uIndex].DeallocateMessageData();
	
	//move the messages above this one down to fill the slot
	for(u_int i=uIndex;i<m_uNumMessages-1;i++)
	{
		m_pxMessages[i] = m_pxMessages[i+1];
	}

	m_uNumMessages--;
}


void Asura_StaticMessageBlock::AllocateMessages(u_int uNumMessages)
{
	Asura_Assert(!m_uNumMessages, "Not expecting AllocateMessages() to be called on a message block that already has messages");

	m_uNumMessages = uNumMessages;
	
	if(m_uNumMessages)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_STATIC_MESSAGE);
		Asura_ResourceAllocateArray(m_pxMessages,m_uNumMessages);
		Asura_ResourcePopType();
	}
	else
	{
		m_pxMessages = 0;
	}
}


#ifdef ASURA_USE_DIALOGUE_SYSTEM
void Asura_StaticMessageBlock::SendMessagesForDialogueSystem(Asura_Guid uSpeaker,unsigned short usExtraMessageBlock, Asura_StaticMessage_Handler* pxHandler) const
{
	const Asura_StaticMessage* pxStaticMessage = m_pxMessages;

	for(u_int i=0;i<m_uNumMessages;i++,pxStaticMessage++)
	{

        if( !Asura_StaticMessage_System::ShouldSendMessage(pxStaticMessage) )
        {
            continue;
        }
		
		const Asura_StaticMessage* pxMessageToSend = pxStaticMessage;

		if(pxStaticMessage->m_usMsgID == ASURA_MSG_DIALOGUE_FROMINSTANCE)
		{
			if(!pxHandler) continue;
			//taking a message from the extra message block
			int iIndex = *(const u_int*)pxStaticMessage->m_pData;
			int iMessages = pxHandler->GetNumberOfMessagesInBlock(usExtraMessageBlock);
			if(iIndex >= iMessages) continue;

			pxMessageToSend = pxHandler->GetMessagePointer(usExtraMessageBlock,iIndex);
			Asura_Assert(pxMessageToSend,"Failed to get static message");

            if( !Asura_StaticMessage_System::ShouldSendMessage(pxMessageToSend) )
            {
                continue;
            }
		}

		Asura_Assert(pxHandler || pxMessageToSend->m_fDelay <= 0.0f, "Can't deal with delayed messages without a message handler");

		if(pxMessageToSend->m_uSendTo == ASURA_GUID_DIALOGUE_SPEAKER)
		{
			if(uSpeaker == ASURA_GUID_UNREGISTERED) continue;
			//substitute in the speaker
			Asura_StaticMessage xMessageCopy = *pxMessageToSend;
			xMessageCopy.m_uSendTo = uSpeaker;

			if(xMessageCopy.m_fDelay > 0.0f && pxHandler)
			{
				pxHandler->QueueMessage(&xMessageCopy, ASURA_GUID_UNREGISTERED, i, 0.0f);
			}
			else
			{
				Asura_StaticMessage_Handler::SendMessage(&xMessageCopy, ASURA_GUID_UNREGISTERED);
			}
		}
		else
		{
			if(pxMessageToSend->m_fDelay > 0.0f && pxHandler)
			{
				pxHandler->QueueMessage(pxMessageToSend, ASURA_GUID_UNREGISTERED, i, 0.0f);
			}
			else
			{
				Asura_StaticMessage_Handler::SendMessage(pxMessageToSend, ASURA_GUID_UNREGISTERED);
			}
		}

	}
}
#endif //ASURA_USE_DIALOGUE_SYSTEM

void Asura_StaticMessageBlock::SendMessages(Asura_Guid uFrom, float fDelay, Asura_StaticMessage_Handler* pxHandler)
{
	Asura_Assert(pxHandler, "No static message handler?");
	if(!pxHandler) pxHandler = &Asura_StaticMessage_System::GetCurrentHandler();

	const Asura_StaticMessage* pxStaticMessage = m_pxMessages;

	for(u_int i=0;i<m_uNumMessages;i++,pxStaticMessage++)
	{
	    if( !Asura_StaticMessage_System::ShouldSendMessage(pxStaticMessage) )
	    {
	        continue;
	    }

		//send or queue the message as required
		if((pxStaticMessage->m_fDelay + fDelay) > 0.0f)
		{
			pxHandler->QueueMessage(pxStaticMessage,uFrom, i, fDelay);
		}
		else
		{
			pxHandler->SendMessage(pxStaticMessage, uFrom);
		}
	}
}
