/**
*
* \file Asura_StaticMessage_System.h
*
* Created by Kevin Lea - 17/04/2002 15:57:27
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/
#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_STATICMESSAGE_SYSTEM_H__
#define __ASURA_STATICMESSAGE_SYSTEM_H__

#include "Core/Asura_HashTree.h"
#include "Entity/Asura_StaticMessageBlock.h"
#include "Network/Asura_Guid.h"
#include "Network/Asura_Message.h"



//Maximum number of delayed messages that can be queued up
#ifndef ASURA_STATICMESSAGE_MAXDELAYEDMESSAGES
	#define ASURA_STATICMESSAGE_MAXDELAYEDMESSAGES 100
#endif

const u_int uASURA_STATICMESSAGE_FLAG_ENGINE_MASK  = 0x0000FFFF;
const u_int uASURA_STATICMESSAGE_FLAG_PROJECT_MASK = 0xFFFF0000;

#ifdef ASURA_TOOLS
const u_int uTOOLS_STATICMESSAGE_FLAG_DISABLED					= 0x00000001; // Disables message for testing purposes. (Message won't get included in exports)
const u_int uTOOLS_STATICMESSAGE_FLAG_IS_TEST_MESSAGE			= 0x00000002; // Temporary test message created in entity simulation mode.
const u_int uTOOLS_STATICMESSAGE_FLAG_MESSAGEEDITOR_SELECTED	= 0x00000004; // Used by the message editor to restore selection on table recreation.
#endif

struct Asura_StaticMessage
{
	public:

        Asura_StaticMessage(){Initialise();}

        u_short     m_usMsgID;
        u_short     m_usDataSize;
		Asura_Guid	m_uSendTo;
		void*		m_pData;
		float		m_fDelay;
		float		m_fProbability;
        u_int       m_uFlags;

		#ifdef ASURA_TOOLS
		//Some tools only data
		u_int 		m_uToolsFlags;
		#endif// ASURA_TOOLS

		void Initialise()
		{
			m_usMsgID=0;
			m_usDataSize=0;
			m_uSendTo=0;
			m_pData=0;
			m_fDelay=0.0f;
			m_fProbability = 1.0f;
            m_uFlags = 0;
			#ifdef ASURA_TOOLS
			m_uToolsFlags = 0;
			#endif
		}

		//Allocates and fills in m_pData and m_usDataSize.
		//Zeroes memory if required
		void AllocateMessageData(u_int uSize,bool bZero);

		//Deallocates m_pData ,and resets it and the data size to zero
		void DeallocateMessageData();

		bool IsRandom() const {return m_fProbability != 1.0f;}

		#ifdef ASURA_TOOLS
		bool operator == ( const Asura_StaticMessage& xMessage) const;
		bool operator != ( const Asura_StaticMessage& xMessage) const {return !(xMessage == *this);}

		bool IsDisabled() const {return (m_uToolsFlags & uTOOLS_STATICMESSAGE_FLAG_DISABLED)!=0;}
		#endif

		void Copy(const Asura_StaticMessage* pxMessageToCopy);
};



/// Kevin Lea - 17/04/2002 16:45:00 - Block numbers are restricted to unsigned short - this may have to change to
/// unsigned int if there are more than 65,535 message blocks... something to watch out for during development.
/// At the time of writing there are 4 trigger entity states, and so as a worse-case scenario each trigger entity would 
/// require 4 message blocks (incidentally with a limit of 255 messages per block) which comes out to 16,000 entities.
/// Should be enough for one level, I'd think. Famous last words, no doubt. ;)

#define ASURA_STATICMESSAGE_EMPTYBLOCK 0xffff // max unsigned short value


enum Asura_StaticMessageSetIndex
{
	ASURA_STATICMESSAGE_ENTITYSET,
	ASURA_STATICMESSAGE_CUTSCENESET,
	ASURA_STATICMESSAGE_NUMSETS,
};

class Asura_StaticMessageSet
{
	public :
		void Initialise();
		void Shutdown();
		void SetupStorageSpace(u_short usNumberOfMessageBlocks);
		void AddMessage(const Asura_StaticMessage& xMessage, u_short& usBlockIndex);
		void DeleteMessage(u_short usBlockIndex,int iMessageIndex);
		void DeleteMessageBlock(u_short usBlockIndex);
		//Gets index of an empty message block , increasing array size if needed
		u_short AllocateMessageBlock();
        
        int GetNumberOfMessageBlocks(void) { return m_usNumberOfMessageBlocks; }
		int GetNumberOfMessagesInBlock(int iBlockIndex);
		Asura_StaticMessage* GetMessagePointer(int iBlockIndex,int iMessageIndex);
		Asura_StaticMessageBlock* GetMessageBlock(int iBlockIndex);
		bool IsMessageBlockValid(int iBlockIndex);

		u_short CopyMessageBlock(Asura_StaticMessageSet* pxSourceSet,u_short usBlockIndex,Asura_Guid uFromGuid = ASURA_GUID_UNREGISTERED,Asura_Guid uToGuid = ASURA_GUID_UNREGISTERED);
		u_short CopyMessageBlock(const Asura_StaticMessageBlock* pxSourceBlock, Asura_Guid uFromGuid = ASURA_GUID_UNREGISTERED, Asura_Guid uToGuid = ASURA_GUID_UNREGISTERED);
	
	private :
		u_short m_usNumberOfMessageBlocks;
		//Asura_StaticMessage** m_ppxMessagesInBlock;
		//u_char* m_pucNumberOfMessagesInBlock;
		Asura_StaticMessageBlock** m_ppxMessageBlocks;

		friend class Asura_Chunk_StaticMessages;

};

class Asura_DelayedStaticMessage
{
public :
	void SetMessage(const Asura_StaticMessage* pxStaticMessage);

	void DeallocateMessage();

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	Asura_StaticMessage* m_pxStaticMessage;
	u_int m_uMessageIndex;

	float m_fDelay;
	Asura_Guid m_uFromGuid;
	#ifdef ASURA_TOOLS
	float m_fInitialDelay;
	#endif
};

/////////////////////////////////
// Asura_StaticMessage_Handler //
/////////////////////////////////
class Asura_StaticMessage_Handler
{
public:
	Asura_StaticMessage_Handler();
	~Asura_StaticMessage_Handler();

	void Initialise();
	void Shutdown();
	void Reset();
	
	/// Kevin Lea - 15/04/2002 17:38:08 - this method is called to set up storage space for the messages,
	/// and will typically be called when loading level data
	void SetupStorageSpace(enum Asura_StaticMessageSetIndex eSet, u_short usNumberOfMessageBlocks);

	//Clears out queue of delayed messages. Needs to be done when restarting from a quicksave , or similar
	void ResetDelayedMessages();

	//per-frame update for dealing with delayed messages
	void Update();


	void SendMessages(int iBlockID, Asura_Guid uFrom,float fDelay = 0.0f);


	void AddMessage(const Asura_StaticMessage& xMessage, u_short& usBlockID);

	/// Kevin Lea - 15/04/2002 17:36:51 - the AddMessage method only get used by the Tools, not in-game,
	/// and as such it can be cavalier when it comes to re-allocating more memory
	void AddMessage(enum Asura_StaticMessageSetIndex eSet, const Asura_StaticMessage& xMessage, u_short& usBlockID);

	// RAR - Deletes a message , shuffling the remaining messages down to fill the slot
	void DeleteMessage(u_short usBlockID, int iMessageIndex);

	//Deletes a message , and the message block if now empty.
	//Sets the block ID to ASURA_STATICMESSAGE_EMPTYBLOCK if deleted
	void DeleteMessage(u_short* pusBlockID, int iMessageIndex);

	//Deletes all the messages in a message block		
	void DeleteMessageBlock(u_short usBlockID);

	//Some access functions for editing
    int GetNumberOfMessageBlocks(enum Asura_StaticMessageSetIndex eSet);
	int GetNumberOfMessagesInBlock(enum Asura_StaticMessageSetIndex eSet, int iBlockIndex);
	int GetNumberOfMessagesInBlock(int iBlockID);
	Asura_StaticMessage* GetMessagePointer(enum Asura_StaticMessageSetIndex eSet, int iBlockIndex, int iMessageIndex);
	Asura_StaticMessage* GetMessagePointer(int iBlockID, int iMessageIndex);


	int GetNumDelayedMessages() const { return m_iNumDelayedMessages; }
	int GetNumDelayedMessagesBeingSent() const { return m_iNumDelayedMessagesBeingSent; }

	//returns true if index is in range , and there are some messages associated with the slot
	bool IsMessageBlockValid(int iBlockID);

	//creates a duplicate of a message block ,and returns the index of the copy
	//uFromGuid and uToGuid are used for redirecting message blocks for messages that are being
	//sent to the owner of the message.
	u_short CopyMessageBlock(u_short usBlockID, Asura_Guid uFromGuid = ASURA_GUID_UNREGISTERED, Asura_Guid uToGuid = ASURA_GUID_UNREGISTERED);
	u_short CopyMessageBlock( const Asura_StaticMessageBlock* pxBlock, const Asura_StaticMessageSetIndex eSet, Asura_Guid uFromGuid /*= ASURA_GUID_UNREGISTERED*/, Asura_Guid uToGuid /*= ASURA_GUID_UNREGISTERED*/);

	//moves a message block from one message set to another.
	u_short TransferMessageBlock(u_short usBlockID, enum Asura_StaticMessageSetIndex eDestinationSet);

	u_short ConvertIndexToID(enum Asura_StaticMessageSetIndex eSet, int iBlockIndex);
	bool ConvertIDToIndex(int iBlockID, int* piOutputBlockIndex, enum Asura_StaticMessageSetIndex* peOutputSet);

	#ifdef ASURA_TOOLS
	static void SendMessage(const Asura_StaticMessage* pxStaticMessage, Asura_Guid uFrom, float fDelayUsed = 0.0f);
	#else
	static void SendMessage(const Asura_StaticMessage* pxStaticMessage, Asura_Guid uFrom);
	#endif

	void QueueMessage(const Asura_StaticMessage* pxStaticMessage, Asura_Guid uFrom, u_int uMessageIndex,float fExtraDelay);
private:
		

	// Callback for qsort
	static int DelayedMessageSort(const void* p1, const void* p2);

	Asura_StaticMessageSet m_xMessageSet[ASURA_STATICMESSAGE_NUMSETS];

	int m_iNumDelayedMessages;
	int m_iNumDelayedMessagesBeingSent;
	class Asura_DelayedStaticMessage m_xDelayedMessages[ASURA_STATICMESSAGE_MAXDELAYEDMESSAGES];


	friend class Asura_Chunk_StaticMessages;
	friend class Asura_Chunk_StaticMessages_InGame;
};

/*
If a function uses BlockIndex as a paramter then it is using the zero-based index for the message in
that message set.

Functions using BlockID are using a global ID value unique across all message sets.

In both cases ASURA_STATICMESSAGE_EMPTYBLOCK is used to indicate an empty block.

RAR 31/03/03

*/
class Asura_StaticMessage_System
{
public:
	static void Initialise(void);

	/// Kevin Lea - 15/04/2002 17:39:08 - free-up storage space
	static void Shutdown(void);

	static void Reset();

	//Clears out queue of delayed messages. Needs to be done when restarting from a quicksave , or similar. The streaming version will do this for all handlers, not just the current one.
	static void ResetDelayedMessages();

	/// Kevin Lea - 15/04/2002 17:38:08 - this method is called to set up storage space for the messages,
	/// and will typically be called when loading level data
	static void SetupStorageSpace(enum Asura_StaticMessageSetIndex eSet, u_short usNumberOfMessageBlocks)
	{
		GetCurrentHandler().SetupStorageSpace(eSet, usNumberOfMessageBlocks);
	}

	//per-frame update for dealing with delayed messages
	static void Update();

#if defined(ASURA_USE_STREAMING)
	static void SendMessages(int iBlockID, Asura_Guid uFrom,float fDelay = 0.0f);
#else
	static void SendMessages(int iBlockID, Asura_Guid uFrom,float fDelay = 0.0f)
	{
		GetCurrentHandler().SendMessages(iBlockID, uFrom, fDelay);
	}
#endif


	static void AddMessage(const Asura_StaticMessage& xMessage, u_short& usBlockID)
	{
		GetCurrentHandler().AddMessage(xMessage, usBlockID);
	}

	/// Kevin Lea - 15/04/2002 17:36:51 - the AddMessage method only get used by the Tools, not in-game,
	/// and as such it can be cavalier when it comes to re-allocating more memory
	static void AddMessage(enum Asura_StaticMessageSetIndex eSet, const Asura_StaticMessage& xMessage, u_short& usBlockID)
	{
		GetCurrentHandler().AddMessage(eSet, xMessage, usBlockID);
	}

	// RAR - Deletes a message , shuffling the remaining messages down to fill the slot
	static void DeleteMessage(u_short usBlockID, int iMessageIndex)
	{
		GetCurrentHandler().DeleteMessage(usBlockID, iMessageIndex);
	}

	//Deletes a message , and the message block if now empty.
	//Sets the block ID to ASURA_STATICMESSAGE_EMPTYBLOCK if deleted
	static void DeleteMessage(u_short* pusBlockID, int iMessageIndex)
	{
		GetCurrentHandler().DeleteMessage(pusBlockID, iMessageIndex);
	}


	//Deletes all the messages in a message block		
	static void DeleteMessageBlock(u_short usBlockID)
	{
		GetCurrentHandler().DeleteMessageBlock(usBlockID);
	}

	//Some access functions for editing
	static int GetNumberOfMessageBlocks(enum Asura_StaticMessageSetIndex eSet)													{ return GetCurrentHandler().GetNumberOfMessageBlocks(eSet); }
	static int GetNumberOfMessagesInBlock(enum Asura_StaticMessageSetIndex eSet,int iBlockIndex)								{ return GetCurrentHandler().GetNumberOfMessagesInBlock(eSet, iBlockIndex); }
	static int GetNumberOfMessagesInBlock(int iBlockID)																			{ return GetCurrentHandler().GetNumberOfMessagesInBlock(iBlockID); }
	static Asura_StaticMessage* GetMessagePointer(enum Asura_StaticMessageSetIndex eSet, int iBlockIndex, int iMessageIndex)	{ return GetCurrentHandler().GetMessagePointer(eSet, iBlockIndex, iMessageIndex); }
	static Asura_StaticMessage* GetMessagePointer(int iBlockID, int iMessageIndex)												{ return GetCurrentHandler().GetMessagePointer(iBlockID, iMessageIndex); }

	//returns true if index is in range , and there are some messages associated with the slot
	static bool IsMessageBlockValid(int iBlockID)
	{
		return GetCurrentHandler().IsMessageBlockValid(iBlockID);
	}

	//creates a duplicate of a message block ,and returns the index of the copy
	//uFromGuid and uToGuid are used for redirecting message blocks for messages that are being
	//sent to the owner of the message.
	static u_short CopyMessageBlock(u_short usBlockID, Asura_Guid uFromGuid = ASURA_GUID_UNREGISTERED, Asura_Guid uToGuid = ASURA_GUID_UNREGISTERED)
	{
		return GetCurrentHandler().CopyMessageBlock(usBlockID, uFromGuid, uToGuid);
	}

	static u_short CopyMessageBlock(const Asura_StaticMessageBlock* pxBlock, const Asura_StaticMessageSetIndex eSet, Asura_Guid uFromGuid = ASURA_GUID_UNREGISTERED, Asura_Guid uToGuid = ASURA_GUID_UNREGISTERED)
	{
		return GetCurrentHandler().CopyMessageBlock( pxBlock, eSet, uFromGuid, uToGuid );
	}

	//moves a message block from one message set to another.
	static u_short TransferMessageBlock(u_short usBlockID, enum Asura_StaticMessageSetIndex eDestinationSet)
	{
		return GetCurrentHandler().TransferMessageBlock(usBlockID, eDestinationSet);
	}

	static u_short ConvertIndexToID(enum Asura_StaticMessageSetIndex eSet, int iBlockIndex)
	{
		return GetCurrentHandler().ConvertIndexToID(eSet, iBlockIndex);
	}
	static bool ConvertIDToIndex(int iBlockID, int* piOutputBlockIndex, enum Asura_StaticMessageSetIndex* peOutputSet) { return GetCurrentHandler().ConvertIDToIndex(iBlockID, piOutputBlockIndex, peOutputSet); }

#if defined(ASURA_USE_STREAMING)
	// Handler functions...
	static	Asura_StaticMessage_Handler*	CreateHandler(const Asura_Hash_ID uHash);
	static	bool							DestroyHandler(const Asura_Hash_ID uHash);

	static	void							CreateAndSetHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = CreateHandler(uHash); }
	static	void							SetCurrentHandler(Asura_StaticMessage_Handler* pxHandler) { s_pxCurrentHandler = pxHandler; }
	static	void							SetCurrentHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static	Asura_StaticMessage_Handler&	GetCurrentHandler(void) { return *s_pxCurrentHandler; }
	static	Asura_StaticMessage_Handler*	GetHandler(const Asura_Hash_ID uHash) { return s_xHandlerTree.GetDataByHash(uHash); }
#else
	// Handler functions...
	static	Asura_StaticMessage_Handler&	GetCurrentHandler(void) { return s_xHandler; }
#endif	// defined(ASURA_USE_STREAMING)

    static bool ShouldSendMessage( const Asura_StaticMessage* pxMessage );
    static void SetMessageFilterCallback( bool(*pfnCallback)(const Asura_StaticMessage*) );

	#ifdef ASURA_TOOLS
	//Used by cutscene mode to intercept any messages being sent to be handled by itself
	static void SetSendMessageCallback(void(*pfnCallback)(const Asura_StaticMessage*, Asura_Guid uFrom));
	#endif

private:
	
#ifdef ASURA_USE_STREAMING
	// Callbacks for tree traversals
	static void Update_Callback(Asura_Hash_ID uHash, class Asura_StaticMessage_Handler* pxData)	{ if (pxData) pxData->Update(); }
	static void Reset_Callback(Asura_Hash_ID uHash, class Asura_StaticMessage_Handler* pxData);
	static void ResetDelayedMessages_Callback(Asura_Hash_ID uHash, class Asura_StaticMessage_Handler* pxData);

	static Asura_StaticMessage_Handler*							s_pxCurrentHandler;		// Currently active handler.
	static Asura_TemplateHashTree<Asura_StaticMessage_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.
#else //#ifdef ASURA_USE_STREAMING
	static Asura_StaticMessage_Handler s_xHandler;				// Single static-instance of a zone.
#endif //#else //#ifdef ASURA_USE_STREAMING

#ifdef ASURA_TOOLS
	static void(*s_pfnSendMessageCallback)(const Asura_StaticMessage*, Asura_Guid uFrom);
#endif	

	friend class Asura_Chunk_StaticMessages;
	friend class Asura_Chunk_StaticMessages_InGame;
};


#endif//__ASURA_STATICMESSAGE_SYSTEM_H__
