///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_ContainerState.cpp
//
// Created by Martin G.Harris (4/7/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Asura_ContainerState.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"

// Asura
#include "Entity/Container/Asura_Container.h"
#include "Entity/Container/Asura_Container_System.h"

// Invariant states "on" and "off". There is no need to load/save these.
// We can also share them between all of the collections saving some memory.
const Asura_ContainerState s_xInvariantOn ( Asura_ContainerState::GetOnStateHash() );
const Asura_ContainerState s_xInvariantOff( Asura_ContainerState::GetOffStateHash() );

///////////////////////////////////////////////////////////////////////////////
// Asura_ContainerState
///////////////////////////////////////////////////////////////////////////////
Asura_ContainerState::Asura_ContainerState(Asura_Hash_ID uStateHash)
: m_uStateHash(uStateHash)
, m_uNumChildStates(0)
, m_pxChildState(0)
, m_uNumDamageOverrides(0)
, m_pxDamageOverride(0)
{
}

Asura_ContainerState::Asura_ContainerState()
: m_uStateHash(ASURA_HASH_ID_UNSET)
, m_uNumChildStates(0)
, m_pxChildState(0)
, m_uNumDamageOverrides(0)
, m_pxDamageOverride(0)
{
}

Asura_ContainerState::~Asura_ContainerState()
{
	FreeAll();
}

void Asura_ContainerState::FreeAll()
{
	delete [] m_pxChildState;
	m_pxChildState = 0;
	m_uNumChildStates = 0;

	delete [] m_pxDamageOverride;
	m_pxDamageOverride = 0;
	m_uNumDamageOverrides = 0;
}

// Chunkstream code
const int iCHUNK_VERSION_ASURA_CONTAINER_STATE = 2;

void Asura_ContainerState::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	FreeAll();

	int iVersion = -1;
	xStream >> iVersion;

	if (iVersion < 0 || iVersion > iCHUNK_VERSION_ASURA_CONTAINER_STATE)
	{
		Asura_Assertf(0, ("Unknown chunk version (%d) in Asura_ContainerState", iVersion));
		return;
	}

	xStream >> m_uStateHash;
	xStream >> m_uNumChildStates;

	if (m_uNumChildStates > 0)
	{
		m_pxChildState = new Asura_ContainerStateEntry[m_uNumChildStates];
		
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			Asura_ContainerStateEntry& xEntry = m_pxChildState[u];
			xStream >> xEntry.m_uContainerHash;
			xStream >> xEntry.m_uStateHash;

			if (iVersion >= 2)
			{
				xStream >> xEntry.m_uInputIndex;
			}
		}
	}

	if (iVersion >= 1)
	{
		xStream >> m_uNumDamageOverrides;

		if (m_uNumDamageOverrides > 0)
		{
			m_pxDamageOverride = new Asura_ContainerState_DamageOverride[m_uNumDamageOverrides];

			for (u_int u=0; u<m_uNumDamageOverrides; ++u)
			{
				Asura_ContainerState_DamageOverride& xEntry = m_pxDamageOverride[u];
				xStream >> xEntry.m_uOverrideStateHash;
				xStream >> xEntry.m_fMaxHealthPercent;
			}
		}
	}
}

void Asura_ContainerState::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// There is no need to write out any invariant states (i.e. "on"/"off").
	Asura_Assert(!IsInvariant(), "Writing an invariant state");

	xStream << iCHUNK_VERSION_ASURA_CONTAINER_STATE;
	xStream << m_uStateHash;
	xStream << m_uNumChildStates;

	for (u_int u=0; u<m_uNumChildStates; ++u)
	{
		const Asura_ContainerStateEntry& xEntry = m_pxChildState[u];
		xStream << xEntry.m_uContainerHash;
		xStream << xEntry.m_uStateHash;
		xStream << xEntry.m_uInputIndex;
	}

	xStream << m_uNumDamageOverrides;
	for (u_int u=0; u<m_uNumDamageOverrides; ++u)
	{
		const Asura_ContainerState_DamageOverride& xEntry = m_pxDamageOverride[u];
		xStream << xEntry.m_uOverrideStateHash;
		xStream << xEntry.m_fMaxHealthPercent;
	}
}


///////////////////
// GetChildState //
///////////////////
// Get child state, returns "off" state hash if child is disabled or unknown.
// Note: Linear search is fine provided there aren't too many children switched on for each state.
// TODO: However if we have states with large numbers of children enabled then we may need to rethink.
Asura_Hash_ID Asura_ContainerState::GetChildState(u_int uInputIndex, Asura_Hash_ID uChildContainerHash) const
{
	// On/off states can't be modified. Their children inherit the same state.
	if (IsInvariant()) return m_uStateHash;

	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			if (m_pxChildState[u].m_uContainerHash == uChildContainerHash &&
				m_pxChildState[u].m_uInputIndex == uInputIndex)
			{
				// Found the correct container and checked that it is for the right input slot.
				return m_pxChildState[u].m_uStateHash;
			}
		}
	}
	return GetOffStateHash();
}

////////////////////////////
// GetDamageOverrideState //
////////////////////////////
// Get damage override state.
Asura_Hash_ID Asura_ContainerState::GetDamageOverrideState(float fHealthPercent) const
{
	// On/off states can't be modified.
	if (IsInvariant()) return m_uStateHash;

	Asura_Assertf(fHealthPercent >= 0.0f && fHealthPercent <= 100.0f, ("Out-of-range health percentage (%f)", fHealthPercent));

	// Note: 100% health always returns the original state.
	if (m_pxDamageOverride && fHealthPercent<100.0f)
	{
		for (u_int u=0; u<m_uNumDamageOverrides; ++u)
		{
			if (fHealthPercent <= m_pxDamageOverride[u].m_fMaxHealthPercent)
			{
				return m_pxDamageOverride[u].m_uOverrideStateHash;
			}
		}
	}

	return m_uStateHash;
}

//////////////////////////////
// HasDamageOverrideChanged //
//////////////////////////////
// Has damage override changed?
bool Asura_ContainerState::HasDamageOverrideChanged(float fOldHealthPercent, float fNewHealthPercent) const
{
	if (IsInvariant()) return false;
	if (!m_pxDamageOverride)  return false;

	return GetDamageOverrideState(fOldHealthPercent) != GetDamageOverrideState(fNewHealthPercent);
}


#ifdef ASURA_TOOLS
// Allocate space from Tools.
void Asura_ContainerState::Allocate(Asura_Hash_ID uStateHash, u_int uNumChildStates, u_int uNumDamageOverrides)
{
	FreeAll();

	m_uStateHash = uStateHash;
	m_uNumChildStates = uNumChildStates;
	m_uNumDamageOverrides = uNumDamageOverrides;

	if (m_uNumChildStates > 0)
	{
		m_pxChildState = new Asura_ContainerStateEntry[m_uNumChildStates];
	}

	if (m_uNumDamageOverrides > 0)
	{
		m_pxDamageOverride = new Asura_ContainerState_DamageOverride[m_uNumDamageOverrides];
	}
}


// Add entry from Tools.
void Asura_ContainerState::SetStateChild(u_int uChildIndex, const Asura_ContainerStateEntry* pxEntry)
{
	if (m_pxChildState && uChildIndex < m_uNumChildStates)
	{
		Asura_Assert(pxEntry, "Null pointer");
		if (pxEntry)
		{
			m_pxChildState[uChildIndex] = *pxEntry;
		}
		return;
	}
	Asura_Assert(0, "Failed to set state child");
}

// Add damage override from Tools.
void Asura_ContainerState::SetDamageOverride(u_int uOverrideIndex, Asura_Hash_ID uOverrideStateHash, float fMaxHealthPercent)
{
	if (m_pxDamageOverride && uOverrideIndex < m_uNumDamageOverrides)
	{
		m_pxDamageOverride[uOverrideIndex].m_uOverrideStateHash = uOverrideStateHash;
		m_pxDamageOverride[uOverrideIndex].m_fMaxHealthPercent  = fMaxHealthPercent;
	}
}
#endif //ASURA_TOOLS

///////////////////////////////////////////////////////////////////////////////
// Asura_ContainerStateMachine
///////////////////////////////////////////////////////////////////////////////
Asura_ContainerStateMachine::Asura_ContainerStateMachine()
: m_uNumStates(0)
, m_pxState(0)
, m_uNumInputs(1)
{
}

Asura_ContainerStateMachine::~Asura_ContainerStateMachine()
{
	FreeAll();
}

/////////////
// FreeAll //
/////////////
void Asura_ContainerStateMachine::FreeAll()
{
	delete [] m_pxState;
	m_pxState = 0;
	m_uNumStates = 0;
}

// Chunkstream code
const int iCHUNK_VERSION_ASURA_CONTAINER_STATE_MACHINE = 1;


void Asura_ContainerStateMachine::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	FreeAll();

	int iVersion = -1;
	xStream >> iVersion;

	if (iVersion < 0 || iVersion > iCHUNK_VERSION_ASURA_CONTAINER_STATE_MACHINE)
	{
		Asura_Assertf(0, ("Unknown chunk version (%d) in Asura_ContainerStateMachine", iVersion));
		return;
	}

	xStream >> m_uNumStates;

	if (iVersion >= 1)
	{
		xStream >> m_uNumInputs;
	}

	if (m_uNumStates > 0)
	{
		m_pxState = new Asura_ContainerState[m_uNumStates];
		Asura_Assert(m_pxState, "Null pointer");
		if (!m_pxState)
		{
			m_uNumStates = 0;
			return;
		}

		for (u_int u=0; u<m_uNumStates; ++u)
		{
			m_pxState[u].ReadFromChunkStream(xStream);
		}
	}
}

void Asura_ContainerStateMachine::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << iCHUNK_VERSION_ASURA_CONTAINER_STATE_MACHINE;
	xStream << m_uNumStates;
	xStream << m_uNumInputs;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		m_pxState[u].WriteToChunkStream(xStream);
	}
}

///////////////
// FindState //
///////////////
// Note: linear search is fine if there are only a handful of states, might
// need something more efficient if we start defining larger numbers of states.
const Asura_ContainerState& Asura_ContainerStateMachine::FindState(Asura_Hash_ID uStateHash) const
{
	// Return invariant states.
	if (uStateHash == Asura_ContainerState::GetOffStateHash())
	{
		return s_xInvariantOff;
	}

	if (uStateHash == Asura_ContainerState::GetOnStateHash())
	{
		return s_xInvariantOn;
	}

	// Search for the state.
	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_pxState[u].GetStateHash() == uStateHash)
		{
			return m_pxState[u];
		}
	}

	// MGH 12/11/2008: When states are renamed, it takes a while for all of the entity files
	// in all of the levels to be automatically fixed-up. I've #ifdef'ed this out for tools as
	// people keep reporting the assert, even though it'll probably fix itself pretty quickly.
	// For retail builds though it indicates that new exports are needed at the very least.
	// Deleting states will also cause this to trigger and this case needs to be fixed-up manually.
#ifndef ASURA_TOOLS
	Asura_Assertf(0, ("Container state (0x%08x) not found. Please use the Verify option on the 'Container File Overlay' to find and fix the error. If the state has been renamed then going in to object mode and saving out the entity file will automatically fix the problem.", uStateHash));
#endif //#ifndef ASURA_TOOLS

	// State not found. Return off state.
	return s_xInvariantOff;
}


//////////////////////////////
// HasDamageOverrideChanged //
//////////////////////////////
// Return true if the damage override state will change as a result of the change in health.
bool Asura_ContainerStateMachine::HasDamageOverrideChanged(const Asura_Container_InputState& xInputState,
														   float fOldHealthPercent,
														   float fNewHealthPercent) const
{
	// Loop over the input states checking whether any of them will be modified by the health change.
	for (u_int u=0; u<xInputState.GetNumInputs(); ++u)
	{
		const Asura_Hash_ID uInputStateHash = xInputState.GetStateHash(u);
		const Asura_ContainerState& xState = FindState(uInputStateHash);
		if (xState.HasDamageOverrideChanged(fOldHealthPercent, fNewHealthPercent))
		{
			return true;
		}
	}
	return false;
}

//////////////////
// ConvertState //
//////////////////
// Convert input state to an output state taking into account any damage overrides.
void Asura_ContainerStateMachine::ConvertState(const Asura_Container_InputState& xInputState,
											   float fHealthPercent,
											   Asura_Container_OutputState& xOutputState) const
{
	const u_int uNumInputs = xInputState.GetNumInputs();
	xOutputState.SetNumInputs(uNumInputs);

	for (u_int u=0; u<uNumInputs; ++u)
	{
		const Asura_Hash_ID uInputStateHash = xInputState.GetStateHash(u);
		const Asura_ContainerState& xState = FindState(uInputStateHash);
		const Asura_Hash_ID uDamageOverrideStateHash = xState.GetDamageOverrideState(fHealthPercent);

		if (uDamageOverrideStateHash == uInputStateHash)
		{
			// No damage override, return the requested state.
			xOutputState.m_apxState[u] = &xState;
		}
		else
		{
			// Substitute a damaged state.
			xOutputState.m_apxState[u] = &FindState(uDamageOverrideStateHash);
		}
	}
}


#ifdef ASURA_TOOLS
void Asura_ContainerStateMachine::SetNumInputs(u_int uNumInputs)
{
	if (m_uNumInputs >= 1 && m_uNumInputs <= uASURA_MAX_NUM_CONTAINER_INPUTS)
	{
		m_uNumInputs = uNumInputs;
	}
	else
	{
		m_uNumInputs = 1;
		Asura_Assertf(0, ("Invalid number of inputs (%u)", uNumInputs));
	}
}

// Allocate space for the states.
void Asura_ContainerStateMachine::AllocateStates(u_int uNumStates)
{
	FreeAll();

	m_uNumStates = uNumStates;
	if (m_uNumStates > 0)
	{
		m_pxState = new Asura_ContainerState[m_uNumStates];
	}
}

// Set a state and allocate space for the child states.
void Asura_ContainerStateMachine::SetState(u_int uIndex, Asura_Hash_ID uStateHash, u_int uNumChildStates, u_int uNumDamageOverrides)
{
	if (m_pxState && uIndex<m_uNumStates)
	{
		m_pxState[uIndex].Allocate(uStateHash, uNumChildStates, uNumDamageOverrides);
		return;
	}
	Asura_Assert(0, "Failed to set state!");
}

// Set a child state.
void Asura_ContainerStateMachine::SetStateChild(u_int uStateIndex, u_int uChildIndex, const Asura_ContainerStateEntry* pxEntry)
{
	if (m_pxState && uStateIndex<m_uNumStates)
	{
		m_pxState[uStateIndex].SetStateChild(uChildIndex, pxEntry);
		return;
	}
	Asura_Assert(0, "Failed to set state child!");
}

// Set a damage override
void Asura_ContainerStateMachine::SetStateDamageOverride(u_int uStateIndex, u_int uOverrideIndex,
														 Asura_Hash_ID uOverrideStateHash, float fMaxHealthPercent)
{
	if (m_pxState && uStateIndex<m_uNumStates)
	{
		m_pxState[uStateIndex].SetDamageOverride(uOverrideIndex, uOverrideStateHash, fMaxHealthPercent);
		return;
	}
	Asura_Assert(0, "Failed to set damage override!");
}
#endif //ASURA_TOOLS

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_InputState
///////////////////////////////////////////////////////////////////////////////
const Asura_Container_InputState Asura_Container_InputState::s_xOffState(Asura_ContainerState::GetOffStateHash());
const Asura_Container_InputState Asura_Container_InputState::s_xOnState(Asura_ContainerState::GetOnStateHash());


///////////////////////////////////////////////////////////////////////////////
// Asura_Container_OutputState
///////////////////////////////////////////////////////////////////////////////
void Asura_Container_OutputState::SetNumInputs(u_int uNumInputs)
{
	m_uNumInputs = uNumInputs;
	for (u_int u=0; u<m_uNumInputs; ++u)
	{
		m_apxState[u] = 0;
	}
}

// Returns the input state for the specified child.
void Asura_Container_OutputState::GetChildInputState(const Asura_Hash_ID uChildContainerHash, Asura_Container_InputState& xInput)
{
	const Asura_Container* const pxContainer = Asura_Container_System::FindContainer(uChildContainerHash);
	Asura_Assertf(pxContainer, ("Container (0x%08x) not found", uChildContainerHash));
	if (!pxContainer)
	{
		xInput = Asura_Container_InputState::Off();
		return;
	}

	// Set child inputs "off" by default.
	const u_int uNumChildInputs = pxContainer->GetNumInputs();
	xInput.SetNumInputs(uNumChildInputs);
	xInput.SetOff();

	// Loop over the inputs and find the relevant child states for each child input.
	// The child input state is basically a union of the all the inputs; it is an error for
	// two or more different states to be written to the same child input.
	for (u_int u=0; u<m_uNumInputs; ++u)
	{
		const Asura_ContainerState* pxState = m_apxState[u];
		Asura_Assert(pxState, "Null state pointer");
		if (pxState)
		{
			for (u_int v=0; v<uNumChildInputs; ++v)
			{
				// Check whether the child container has an input set for the given input index.
				const Asura_Hash_ID uChildStateHash = pxState->GetChildState(v, uChildContainerHash);

				if (Asura_ContainerState::GetOffStateHash() != uChildStateHash)
				{
					// If the state is already set to "on" then don't overwrite it, nor assert.
					const Asura_Hash_ID uPreviousStateHash = xInput.GetStateHash(v);
					if (uPreviousStateHash != Asura_ContainerState::GetOnStateHash())
					{
						// If we are writing a state to a child input, then either:
						// 1) that input should not have been written to yet, or
						// 2) it should already have the value that is about to be written to it, or
						// 3) one of the states (previous/new) is "on" and that overrides the other (new/previous).
						// Any other case means that the input is not well-defined.
#ifndef ASURA_TOOLS
						Asura_Assert(uPreviousStateHash == Asura_ContainerState::GetOffStateHash() || uPreviousStateHash == uChildStateHash || uChildStateHash == Asura_ContainerState::GetOnStateHash(), "Tried to write two different states to container child input. Result is undefined.");
#endif //ASURA_TOOLS

						xInput.SetStateHash(v, uChildStateHash);
					}
				}
			}
		}	
	}	
}

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM
