///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_ContainerState.h
//
// Created by Martin G.Harris (4/7/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_CONTAINER_STATE_H__
#define __ASURA_CONTAINER_STATE_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_Maths.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_FixedArray.h"

// Forward declarations
class Asura_Chunk_Stream;
class Asura_Container_InputState;
class Asura_Container_OutputState;

// The maximum number of container inputs.
// Increase if necessary, but not too much as it will waste quite a bit of memory and some bandwidth.
const u_int uASURA_MAX_NUM_CONTAINER_INPUTS = 4;

///////////////////////////////
// Asura_ContainerStateEntry //
///////////////////////////////
// Associate a container and a state; used as an entry for a child in an overall state definition.
// The specified child container, has the specified state, for the specified input slot.
class Asura_ContainerStateEntry
{
public:
	Asura_ContainerStateEntry()
		: m_uContainerHash(ASURA_HASH_ID_UNSET)
		, m_uStateHash(ASURA_HASH_ID_UNSET)
		, m_uInputIndex(0)
	{}

	Asura_Hash_ID m_uContainerHash;	// Hash of child container.
	Asura_Hash_ID m_uStateHash;		// Input state of child container.
	u_int m_uInputIndex;			// Input slot that is affected for the child.
};


/////////////////////////////////////////
// Asura_ContainerState_DamageOverride //
/////////////////////////////////////////
// When a state (e.g. "light on") has been damaged (i.e. health is less than 100%)
// it should be replaced by a substitute state (e.g. "light flicker" or "destroyed").
// This class contains the substitute state hash and the corresponding maximum health percentage.
class Asura_ContainerState_DamageOverride
{
public:
	Asura_ContainerState_DamageOverride()
		: m_uOverrideStateHash(ASURA_HASH_ID_UNSET)
		, m_fMaxHealthPercent(0.0f)
	{}

	Asura_Hash_ID m_uOverrideStateHash;
	float m_fMaxHealthPercent;	// Max health percentage inclusive.
};


//////////////////////////
// Asura_ContainerState //
//////////////////////////
// Object representing a container state definition belonging to a specific collection.
// Each state stores information about the states of any child containers.
class Asura_ContainerState
{
public:
	Asura_ContainerState(Asura_Hash_ID uStateHash);
	Asura_ContainerState();
	~Asura_ContainerState();

	Asura_Hash_ID GetStateHash() const { return m_uStateHash; }

	// Get child state, returns "off" state hash if child is disabled or unknown.
	// uChildContainerHash specifies the child container.
	// uInputIndex specifies which input slot of the child we are interested in (usually zero).
	Asura_Hash_ID GetChildState(u_int uInputIndex, Asura_Hash_ID uChildContainerHash) const;

	// Get damage override state.
	Asura_Hash_ID GetDamageOverrideState(float fHealthPercent) const;

	// Has damage override changed?
	bool HasDamageOverrideChanged(float fOldHealthPercent, float fNewHealthPercent) const;

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	// On/off states can't be modified.
	bool IsInvariant() const
	{
		return (m_uStateHash == GetOnStateHash() || m_uStateHash == GetOffStateHash());
	}

	// Note: these must match the corresponding definition in Tools_ContainerState.
	// Hash representing the "on" state.
	static Asura_Hash_ID GetOnStateHash() { return 0x00000ddf; } // Asura_GetHashID("on");
	
	// Hash representing the "off" state.
	static Asura_Hash_ID GetOffStateHash() { return 0x0001ad6f; } // Asura_GetHashID("off");

	// Helper. Does the state correspond to off?
	static bool IsOff(const Asura_Hash_ID uStateHash)
	{
		return (uStateHash == GetOffStateHash()) || (uStateHash == ASURA_HASH_ID_UNSET);
	}

#ifdef ASURA_TOOLS
	void Allocate(Asura_Hash_ID uStateHash, u_int uNumChildStates, u_int uNumDamageOverrides);
	void SetStateChild(u_int uChildIndex, const Asura_ContainerStateEntry* pxEntry);
	void SetDamageOverride(u_int uOverrideIndex, Asura_Hash_ID uOverrideStateHash, float fMaxHealthPercent);
#endif //ASURA_TOOLS

protected:
	void FreeAll();


protected:
	Asura_Hash_ID m_uStateHash;		// Hash of the state name.

	// States of child containers.
	u_int m_uNumChildStates;
	Asura_ContainerStateEntry* m_pxChildState;

	// Damage overrides
	u_int m_uNumDamageOverrides;
	Asura_ContainerState_DamageOverride* m_pxDamageOverride;
};


/////////////////////////////////
// Asura_ContainerStateMachine //
/////////////////////////////////
// Set of state definitions comprising a state machine.
// Each container collection has its own state machine.
class Asura_ContainerStateMachine
{
public:
	Asura_ContainerStateMachine();
	~Asura_ContainerStateMachine();

	u_int GetNumStates() const { return m_uNumStates; }
	u_int GetNumInputs() const { return m_uNumInputs; }

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	// Return true if the damage override state will change as a result of the change in health.
	bool HasDamageOverrideChanged(const Asura_Container_InputState& xInputState,
								  float fOldHealthPercent,
								  float fNewHealthPercent)const;

	// Convert input state to an output state taking into account any damage overrides.
	void ConvertState(const Asura_Container_InputState& xInputState,
					  float fHealthPercent,
					  Asura_Container_OutputState& xOutputState) const;


#ifdef ASURA_TOOLS
	void SetNumInputs(u_int uNumInputs);
	void AllocateStates(u_int uNumStates);
	void SetState(u_int uIndex, Asura_Hash_ID uStateHash, u_int uNumChildStates, u_int uNumDamageOverrides);
	void SetStateChild( u_int uStateIndex, u_int uChildIndex, const Asura_ContainerStateEntry* pxEntry);
	void SetStateDamageOverride(u_int uStateIndex, u_int uOverrideIndex,
								Asura_Hash_ID uOverrideStateHash, float fMaxHealthPercent);
#endif //ASURA_TOOLS

protected:
	void FreeAll();
	const Asura_ContainerState& FindState(Asura_Hash_ID uStateHash) const;

protected:
	// States
	u_int m_uNumStates;
	Asura_ContainerState* m_pxState;

	u_int m_uNumInputs;
};



////////////////////////////////
// Asura_Container_InputState //
////////////////////////////////
// Input state for a collection or other container instance.
// For now there is only a single input state, but that will be changed soon.
class Asura_Container_InputState
{
private:
	explicit Asura_Container_InputState(Asura_Hash_ID uStateHash) : m_uNumInputs(1) { m_auStateHash[0] = uStateHash; }

public:
	// Constructors
	Asura_Container_InputState() : m_uNumInputs(1) { m_auStateHash[0] = ASURA_HASH_ID_UNSET; }

	Asura_Container_InputState(const Asura_Container_InputState& xOther) { *this = xOther; }
	
	// Reset
	void Reset()
	{
		for (u_int u=0; u<uASURA_MAX_NUM_CONTAINER_INPUTS; ++u)
		{
			m_auStateHash[u] = ASURA_HASH_ID_UNSET;
		}
	}

	// Set
	void SetNumInputs(const u_int uNumber)
	{
		m_uNumInputs = Asura_Maths::ClampToRange(uNumber, 1u, uASURA_MAX_NUM_CONTAINER_INPUTS);
	}

	void SetStateHash(u_int uIndex, Asura_Hash_ID uStateHash) { m_auStateHash[uIndex] = uStateHash; }

	void Set(const u_int uNumInputs, const Asura_Hash_ID* puStateHash)
	{
		SetNumInputs(uNumInputs);
		for (u_int u=0; u<m_uNumInputs; ++u)
		{
			m_auStateHash[u] = (puStateHash ? puStateHash[u] : ASURA_HASH_ID_UNSET);
		}
	}

	// Set all inputs to be off.
	void SetOff()
	{
		for (u_int u=0; u<m_uNumInputs; ++u)
		{
			m_auStateHash[u] = Asura_ContainerState::GetOffStateHash();
		}
	}

	// Get
	u_int GetNumInputs() const { return m_uNumInputs; }
	Asura_Hash_ID GetStateHash(u_int uIndex) const { return m_auStateHash[uIndex]; }

	const Asura_Hash_ID* GetStateHashArray() const { return &m_auStateHash[0]; }
	      Asura_Hash_ID* GetStateHashArray()       { return &m_auStateHash[0]; }


	// Operators
	Asura_Container_InputState& operator=(const Asura_Container_InputState& xOther)
	{
		m_uNumInputs = xOther.m_uNumInputs;

		for (u_int u=0; u<m_uNumInputs; ++u)
		{
			m_auStateHash[u] = xOther.m_auStateHash[u];
		}
		return *this;
	}

	bool operator==(const Asura_Container_InputState& xOther) const
	{
		if (m_uNumInputs != xOther.m_uNumInputs) return false;

		for (u_int u=0; u<m_uNumInputs; ++u)
		{
			if (m_auStateHash[u] != xOther.m_auStateHash[u])
			{
				return false;
			}
		}
		return true;
	}

	bool operator!=(const Asura_Container_InputState& xOther) const { return !(*this == xOther); }
	
	// Are all the inputs "off"?
	bool IsOff() const
	{
		for (u_int u=0; u<m_uNumInputs; ++u)
		{
			if (!Asura_ContainerState::IsOff(m_auStateHash[u]))
			{
				return false;
			}
		}
		return true;
	}

private:
	u_int m_uNumInputs;
	Asura_FixedArray<Asura_Hash_ID, uASURA_MAX_NUM_CONTAINER_INPUTS> m_auStateHash;

	// Static members
public:
	// Pre-defined on/off states with a single input.
	static const Asura_Container_InputState& Off() { return s_xOffState; }
	static const Asura_Container_InputState& On()  { return s_xOnState; }

private:
	static const Asura_Container_InputState s_xOffState;
	static const Asura_Container_InputState s_xOnState;
};

/////////////////////////////////
// Asura_Container_OutputState //
/////////////////////////////////
class Asura_Container_OutputState
{
public:
	Asura_Container_OutputState() : m_uNumInputs(1) { m_apxState[0] = 0; }

	// Returns the input state for the specified child.
	void GetChildInputState(const Asura_Hash_ID uChildContainerHash, Asura_Container_InputState& xInput);

private:
	// Make private.
	Asura_Container_OutputState(const Asura_Container_OutputState& xOther);
	Asura_Container_OutputState& operator=(const Asura_Container_OutputState& xOther);

	void SetNumInputs(u_int uNumInputs);

private:
	u_int m_uNumInputs;
	Asura_FixedArray<const Asura_ContainerState*, uASURA_MAX_NUM_CONTAINER_INPUTS> m_apxState;

	// Friend
	friend class Asura_ContainerStateMachine;
};

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__ASURA_CONTAINER_STATE_H__

