///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_ServerContainerInstance.h
//
// Created by Martin G.Harris (21/08/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// Instance of a container owned by a server entity.
// Contains any data that differs between instances, which needs to be serialised
// and sent to all the clients. E.g. current health.
//
// The tree is always made ignoring the current state hash (i.e. as if all the
// elements were enabled), but only those elements needed are included
// (i.e. collections and health); attachments are just skipped over.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_SERVER_CONTAINER_INSTANCE_H__
#define __ASURA_SERVER_CONTAINER_INSTANCE_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Entity/Container/Asura_Container.h"
#include "Debugging/Asura_Assert.h"

// Forward declarations
class Asura_Container_FloatParam;
class Asura_Container_PhysicsParam;
class Asura_Container_PosOrientVel;
class Asura_Physics_Object;
class Asura_ServerContainerInstance;

///////////////////////////////////
// Asura_Container_ServerFunctor //
///////////////////////////////////
// Base class for functor that traverses over the server container instance.
class Asura_Container_ServerFunctor
{
public:
	Asura_Container_ServerFunctor() {}
	virtual ~Asura_Container_ServerFunctor() {}

	virtual void Process(const Asura_ServerContainerInstance* pxEntry) = 0;
};


//////////////////////////////////////////////
// Asura_ServerContainerInstance_Properties //
//////////////////////////////////////////////
class Asura_ServerContainerInstance_Properties
{
public:
	Asura_ServerContainerInstance_Properties()
		: m_uNumFloatParam(0)
		, m_bBlockWaypointLinks(false)
	{}

	int m_uNumFloatParam;
	// Set if any of the meshes are flagged as blocking links.
	bool m_bBlockWaypointLinks;
};

///////////////////////////////////
// Asura_ServerContainerInstance //
///////////////////////////////////
class Asura_ServerContainerInstance
{
public:
	Asura_ServerContainerInstance(const Asura_Container* pxContainer);
	virtual ~Asura_ServerContainerInstance() {}

	u_int GetClassification() const;		// Returns classification of container.
	Asura_Hash_ID GetContainerHash() const;	// Returns hash of container.

	void SetUniqueID(u_int uUniqueID) { m_uUniqueID = uUniqueID; }
	u_int GetUniqueID() const { return m_uUniqueID; }

	virtual void Initialise(Asura_ServerContainerInstance_Properties& xData) {}

	// Fill out the physics parameters.
	virtual void FillPhysicsParam(Asura_Container_PhysicsParam*& pxPhysicsData, u_int& uNumLeftToWrite) const {}

	// Fill out the float parameters.
	virtual void FillFloatParam(Asura_Container_FloatParam*& pxFloatData, u_int& uNumLeftToWrite) const {}

	// Store the physics data. Return true if added new data.
	virtual bool StorePhysicsParam(const Asura_Physics_Object* pxPhysics)
	{
		Asura_Assert(0, "StorePhysicsParam() not implemented for this class!");
		return false;
	}

	virtual void GetNumPhysicsParam(u_int& uNumber) const {}

	// Look up the stored physics velocities for the given container instance.
	// Return true if found, in which case xLinVelocity and xAngVelocity are set with the output data.
	virtual bool GetPhysicsVelocity(Asura_Vector_3& xLinVelocity, Asura_Vector_3& xAngVelocity) const { return false; }


	// Find instance from unique ID.
	virtual Asura_ServerContainerInstance* FindFromUniqueID_Safe(u_int uUniqueID, Asura_Hash_ID uHash, u_int uClass)
	{
		if (uUniqueID == 0 && uHash == GetContainerHash() && uClass == GetClassification())
		{
			return this;
		}
		return 0;
	}

	virtual Asura_ServerContainerInstance* FindFromUniqueID(u_int uUniqueID)
	{
		return (uUniqueID == 0) ? this : 0;
	}

	// Change state.
	virtual void ChangeState(const Asura_Container_InputState& xInputState) { m_xInputState = xInputState; }
	const Asura_Container_InputState& GetInputState() const { return m_xInputState; }
	bool IsOff() const { return m_xInputState.IsOff(); }

	// Use the unique ID to find the relevant parent collection.
	virtual class Asura_ServerContainerInstance_Collection* FindCollectionFromUniqueID(u_int uUniqueID);

	// Read/write all of serialisable data for the server instance containers.
	static void ReadAllData (Asura_Chunk_Stream& xStream, Asura_ServerContainerInstance* pxInstance);
	static void WriteAllData(Asura_Chunk_Stream& xStream, Asura_ServerContainerInstance* pxInstance);

	// Chunkstream functions.
	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream, u_short usVersion) {}
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const {}

	// Traverse the tree with callback functor.
	virtual void Traverse(Asura_Container_ServerFunctor& xFunctor) const { xFunctor.Process(this); }

protected:
	// Write out a header.
	void WriteHeader(Asura_Chunk_Stream& xStream, u_int uSize, u_short usVersion) const;
	
	// Skip the header and return the size of the remaining data.
	static u_int SkipHeader(Asura_Chunk_Stream& xStream);

	// Skip all of the rest of the data.
	static void SkipAllData(Asura_Chunk_Stream& xStream, int iVersion);

protected:
	const Asura_Container* const m_pxContainer;
	u_int m_uUniqueID; // Identifier for use by LOS and serialisation code.
	
	// Current state of container (for collections this is the input state before applying damage overrides).
	Asura_Container_InputState m_xInputState;
};

//////////////////////////////////////////////
// Asura_ServerContainerInstance_Collection //
//////////////////////////////////////////////
class Asura_ServerContainerInstance_Collection : public Asura_ServerContainerInstance
{
	typedef Asura_ServerContainerInstance PARENT;

public:
	Asura_ServerContainerInstance_Collection(const Asura_Container_Collection* pxContainer);
	virtual ~Asura_ServerContainerInstance_Collection();

	virtual void Initialise(Asura_ServerContainerInstance_Properties& xData);

	// Fill out the physics parameters.
	virtual void FillPhysicsParam(Asura_Container_PhysicsParam*& pxPhysicsData, u_int& uNumLeftToWrite) const;

	// Fill out the float parameters.
	virtual void FillFloatParam(Asura_Container_FloatParam*& pxFloatData, u_int& uNumLeftToWrite) const;

	// Store the physics data. Return true if added new data.
	virtual bool StorePhysicsParam(const Asura_Physics_Object* pxPhysics);

	virtual void GetNumPhysicsParam(u_int& uNumber) const;

	// Look up the stored physics velocities for the given container instance.
	// Return true if found, in which case xLinVelocity and xAngVelocity are set with the output data.
	virtual bool GetPhysicsVelocity(Asura_Vector_3& xLinVelocity, Asura_Vector_3& xAngVelocity) const;


	// Find instance from unique ID.
	virtual Asura_ServerContainerInstance* FindFromUniqueID_Safe(u_int uUniqueID, Asura_Hash_ID uHash, u_int uClass);
	virtual Asura_ServerContainerInstance* FindFromUniqueID(u_int uUniqueID);
	
	// Change state.
	virtual void ChangeState(const Asura_Container_InputState& xInputState);

	// Use the unique ID to find the relevant parent collection.
	virtual Asura_ServerContainerInstance_Collection* FindCollectionFromUniqueID(u_int uUniqueID);

	// Chunkstream functions.
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream, u_short usVersion);
	// Traverse tree, writing out instances as needed.
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	// Find the relevant health container (for ApplyDamage we only consider enabled health containers).
	class Asura_ServerContainerInstance_Health* FindHealth(bool bEnabledOnly = true);

	// Check whether the damage override has changed.
	bool HasDamageOverrideChanged(float fOldHealthPercent, float fNewHealthPercent) const;

	// Traverse the tree with callback functor.
	virtual void Traverse(Asura_Container_ServerFunctor& xFunctor) const;

protected:
	void DeleteAllChildren();

protected:
	u_int m_uNumChildren;
	Asura_ServerContainerInstance** m_ppxChildren;
	Asura_Container_PosOrientVel* m_pxPosOrientData;
};


//////////////////////////////////////////
// Asura_ServerContainerInstance_Health //
//////////////////////////////////////////
class Asura_ServerContainerInstance_Health : public Asura_ServerContainerInstance
{
	typedef Asura_ServerContainerInstance PARENT;

public:
	Asura_ServerContainerInstance_Health(const Asura_Container_Health* pxContainer);

	virtual void Initialise(Asura_ServerContainerInstance_Properties& xData) { ++xData.m_uNumFloatParam; }

	// Fill out the float parameters.
	virtual void FillFloatParam(Asura_Container_FloatParam*& pxFloatData, u_int& uNumLeftToWrite) const;

	// Chunkstream functions.
	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream, u_short usVersion);
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	// Get the health as a percentage of the maximum value.
	float GetHealthPercent() const;

	// Get/set the health values.
	float GetHealth() const { return m_fHealth; }
	void SetHealth(float fHealth);

	// Reduce health by an amount fDamage (or increase it if fDamage is negative).
	// Returns true if the health was changed.
	bool ApplyDamage(float fDamage);

protected:
	void ClampToRange();

protected:
	float m_fHealth;
};

////////////////////////////////////////
// Asura_ServerContainerInstance_Mesh //
////////////////////////////////////////
class Asura_ServerContainerInstance_Mesh : public Asura_ServerContainerInstance
{
	typedef Asura_ServerContainerInstance PARENT;

public:
	Asura_ServerContainerInstance_Mesh(const Asura_Container_Mesh* pxContainer);
	virtual ~Asura_ServerContainerInstance_Mesh();

	virtual void Initialise(Asura_ServerContainerInstance_Properties& xData);

	// Fill out the physics parameters.
	virtual void FillPhysicsParam(Asura_Container_PhysicsParam*& pxPhysicsData, u_int& uNumLeftToWrite) const;

	// Store the physics data. Return true if added new data.
	virtual bool StorePhysicsParam(const Asura_Physics_Object* pxPhysics);

	virtual void GetNumPhysicsParam(u_int& uNumber) const;
	
	// Look up the stored physics velocities for the given container instance.
	// Return true if found, in which case xLinVelocity and xAngVelocity are set with the output data.
	virtual bool GetPhysicsVelocity(Asura_Vector_3& xLinVelocity, Asura_Vector_3& xAngVelocity) const;


	// Chunkstream functions.
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream, u_short usVersion);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

protected:
	Asura_Container_PosOrientVel* m_pxPosOrientData;
};


#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__ASURA_SERVER_CONTAINER_INSTANCE_H__

