/**
*
* \file Asura_ServerEntity_CutsceneController.cpp
*
* Created by Kevin Lea - 15/04/2002 12:13:56
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "Entity/Cutscene/Asura_ServerEntity_CutsceneController.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Network/Asura_Message.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Cutscene/Asura_Cutscene_Event.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Cutscene/Asura_ClientEntity_CutsceneController.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Network.h"

#ifdef ASURA_TOOLS
#include "Entity/Tools_Entity_CutsceneController_Object.h"
#endif

#ifdef ASURA_USE_STREAMING
#include "Streaming/Asura_Streaming_System.h"
#endif //ASURA_USE_STREAMING


#ifdef ASURA_TOOLS
Asura_Hashed_List<Asura_ServerEntity_CutsceneController> Asura_ServerEntity_CutsceneController::s_xList;
#endif


static const u_int CHUNK_VERSION = 7;

Asura_ServerEntity_CutsceneController::Asura_ServerEntity_CutsceneController(const Asura_String& cutsceneName,Asura_Guid guid /*= ASURA_GUID_UNREGISTERED*/)
: PARENT(guid)
{
	Classification = AsuraEntityClass_CutsceneController;

	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif

	CutsceneName = cutsceneName;
	m_uStateFlags = 0;
	m_uCheckStartActiveFlag = 1;
	m_fTimer = 0;
	m_uEnabledEvents = 0xffffffff;
	// NY - 20 June 2008 - this used to leave itself on the update client list, but stub it's make update function out
	SetNeverUpdateClient();
}

Asura_ServerEntity_CutsceneController::Asura_ServerEntity_CutsceneController(Asura_Guid guid)
: PARENT(guid)
, m_uStateFlags(0)
, m_uEnabledEvents(0xffffffff)
, m_fTimer(0)
{
	Classification = AsuraEntityClass_CutsceneController;

	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif
	
	m_uCheckStartActiveFlag = 1;
	// NY - 20 June 2008 - this used to leave itself on the update client list, but stub it's make update function out
	SetNeverUpdateClient();
}

Asura_ServerEntity_CutsceneController::Asura_ServerEntity_CutsceneController(Asura_Guid guid,const char*& chunkBuffer)
:PARENT(guid)
{
	Classification = AsuraEntityClass_CutsceneController;

	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif

	//check the version
	u_int uVersion;
	ASURA_CHUNK_EXTRACT( chunkBuffer, uVersion, u_int );
	Asura_Assert( CHUNK_VERSION >= uVersion, "Chunk version is too new; your code is out of date wrt this data." );

	//read in parent class data if new enough
	if(uVersion >= 7)
	{
		Asura_Entity_Server::InitialiseFromChunkBuffer(chunkBuffer);
	}

	//read in the name
	ASURA_CHUNK_EXTRACT_STRING(chunkBuffer, CutsceneName);

    m_uStateFlags = 0;
    if( uVersion >= 1 )
    {
        ASURA_CHUNK_EXTRACT( chunkBuffer, m_uStateFlags, u_int );
		
		if(uVersion==1)
		{
			//make sure the unused bits are zeroes
			bool bStartActive = m_uCheckStartActiveFlag;
			m_uStateFlags = 0;
	        m_uCheckStartActiveFlag = bStartActive;
		}	
		else if(uVersion==2)
		{
			// zero the previously unset m_uManuallyDisabled flag
			m_uManuallyDisabled = 0;
		}
    }
    else
    {
        m_uCheckStartActiveFlag = 1;
    }

	if(uVersion>=4)
	{
		ASURA_CHUNK_EXTRACT(chunkBuffer, m_fTimer, float);
	}
	else
	{
		m_fTimer = 0;
	}
	if(uVersion>=5)
	{
		ASURA_CHUNK_EXTRACT(chunkBuffer, m_uEnabledEvents, u_int);
	}
	if(uVersion<6)
	{
		//version 5 may contain junk here.
		m_uEnabledEvents = 0xffffffff;
	}

	//Update the enabled events status for the associated cutscene.
	//This is needed so that  cutscenes that don't get started via their cutscene controller 
	//get set correctly. (Namely cutscenes that are attachments).
	//This will only work correctly for server side cutscenes , but its only server cutscenes that
	//have this issue at the moment. RAR 27/07/2005
	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
	if(pxCutscene)
	{
		pxCutscene->m_uEnabledEvents = m_uEnabledEvents;

#if defined(ASURA_USE_STREAMING)
		pxCutscene->m_uGuid = GetGuid();
#endif //defined(ASURA_USE_STREAMING)
	}

	m_uCheckForCutsceneRestartAfterLoad = m_uPlaying;
	m_uSkipMessagesAfterLoad = false;
	// NY - 20 June 2008 - this used to leave itself on the update client list, but stub it's make update function out
	SetNeverUpdateClient();
}

Asura_ServerEntity_CutsceneController::~Asura_ServerEntity_CutsceneController()
{
	//If cutscene is being played on server (and still exists) , then we need to stop it
	//RAR 01/07/04
	if(m_uPlaying)
	{
		Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
		if(pxCutscene && pxCutscene->PlaysOnServer && pxCutscene->IsPlaying())
		{
			pxCutscene->Finish();
		}
	}

	#ifdef ASURA_TOOLS
	s_xList.RemoveFromList(this);
	#endif

}

void Asura_ServerEntity_CutsceneController::CreateFromChunkBuffer(Asura_Guid uGuid, const char*& pcChunkBuffer)
{
#ifdef TEST_ENTITY_CHUNK_STREAM
	TestStream<Asura_ServerEntity_CutsceneController>(uGuid, pcChunkBuffer);
#endif
	new Asura_ServerEntity_CutsceneController(uGuid, pcChunkBuffer);
}

bool Asura_ServerEntity_CutsceneController::HandleMessage(const Asura_Message* pxMessage)
{
	Asura_Assert(pxMessage,"");
	if (!pxMessage)	return false;
	
	switch(pxMessage->GetMsgID())
	{
        case ASURA_MSG_CUTSCENECONTROLLER_REQ_PROFILE_CUTSCENE:
        {
			//get the hash for this cutscene
			Asura_Hash_ID uCutsceneHash = Asura_GetHashID(CutsceneName.GetString());
            Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByHash(uCutsceneHash);
			Asura_Assert(pxCutscene,"Failed to find cutscene");
			if(!pxCutscene) break;//giveup

			// (MH - 03.10.2002) see if this needs to start a cutscene up on the client.
            if (pxCutscene && pxCutscene->PlaysOnServer)
            {
                Asura_Assert( false, "Not implemented." );
            }
            else
            {
			    Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_CUTSCENECONTROLLER_PROFILE_CUTSCENE, sizeof(Asura_Msg_ActivateCutscene) );
                if( !pxMsg ) { break; }

				Asura_Msg_ActivateCutscene* pxData = static_cast<Asura_Msg_ActivateCutscene*>(pxMsg->GetData());
				if (pxData)
				{
					pxData->m_uHash = uCutsceneHash;
					pxData->m_ucFlags = 0;
					
					if ((pxCutscene->m_iCutsceneFlags & AsuraCutsceneFlag_StartsActive) != 0)
					{
						pxData->m_ucFlags |= Asura_Msg_ActivateCutscene::FLAG_STARTS_ACTIVE;
					}

					pxData->m_fTimer = 0.0f; // Ignored
					pxData->m_uEnabledEvents = pxCutscene->m_uEnabledEvents;
				}

		 	    pxMsg->To   = ASURA_GUID_MY_CLIENT_ENTITIES;
			    pxMsg->From = this->Guid;

			    Asura_Network::Send( pxMsg );
            }
            return true;
        }
		case ASURA_MSG_ACTIVATECUTSCENEWITHTIME:
		{
			const Asura_Message_ActivateCutsceneWithTime* pxCutsceneMessage= static_cast<const Asura_Message_ActivateCutsceneWithTime*>(pxMessage->GetData());
			if(!m_uDisabled)
			{
				UpdatePlayingFlag();
				if(!m_uPlaying)
				{
					Activate( false, pxCutsceneMessage->m_fStartTime );
				}
			}
			return true;
		}
        
		case ASURA_MSG_ACTIVATE:
		{
			if(!m_uDisabled)
			{
				UpdatePlayingFlag();
				if(!m_uPlaying)
				{
					Activate( false, 0 );
				}
			}
			return true;
		}
		case ASURA_MSG_DEACTIVATE:
		{
			if(!m_uDisabled)
			{
				// (MH - 03.10.2002) see if this needs to send a deactivate message to the client.
            	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
            	Asura_Assert(pxCutscene, "failed to find cutscene");
            	if (pxCutscene && pxCutscene->PlaysOnServer)
            	{
            	    pxCutscene->Finish();
            	}
            	else
            	{
				    Asura_Message* pxMsg = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
				    Send(pxMsg, ASURA_GUID_MY_CLIENT_ENTITIES);
            	}
				m_uPlaying = false;
			}
			return true;
		}

		case ASURA_MSG_CUTSCENE_ENABLEEVENT :
		case ASURA_MSG_CUTSCENE_DISABLEEVENT :
		{
			const Asura_Message_CutsceneEventID* pxCutsceneMessage = static_cast<const Asura_Message_CutsceneEventID*>(pxMessage->GetData());
			u_int uEventBit = 1<<(pxCutsceneMessage->m_iEventID-1);

			//update local copy of enabled events
			if(pxMessage->GetMsgID() == ASURA_MSG_CUTSCENE_ENABLEEVENT)
				m_uEnabledEvents |= uEventBit;
			else
				m_uEnabledEvents &= ~uEventBit;

			//update the cutscene as well
            Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
            Asura_Assert(pxCutscene, "failed to find cutscene");

			if(pxCutscene &&pxCutscene->PlaysOnServer)
			{
				//server cutscenes can be done directly
				pxCutscene->m_uEnabledEvents = m_uEnabledEvents;
			}
			else
			{
				//for client cutscenes we need to send a message with the current state
			    Asura_Message* pxCurrentStateMessage = Asura_Network::CreateMessage( ASURA_MSG_CUTSCENECONTROLLER_UPDATEEVENTSTATE, sizeof(Asura_Message_CutsceneEventState));
				Asura_Message_CutsceneEventState* pxData = (Asura_Message_CutsceneEventState*)pxCurrentStateMessage->GetData();
				pxData->m_uEnabledEvents = m_uEnabledEvents;

			    pxCurrentStateMessage->To = ASURA_GUID_MY_CLIENT_ENTITIES;
			    pxCurrentStateMessage->From = Guid;

				Asura_Network::Send(pxCurrentStateMessage);
			}
			return true;
		}

		case ASURA_MSG_CUTSCENE_SETCAMERA :
		{
			const Asura_Message_Cutscene_SetCamera* pxCutsceneMessage = static_cast<const Asura_Message_Cutscene_SetCamera*>(pxMessage->GetData());

            Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
            Asura_Assert(pxCutscene, "failed to find cutscene");

			if(pxCutscene && pxCutscene->PlaysOnServer)
			{
				//server cutscene , so can update directly
				if(pxCutscene->IsPlaying() && pxCutsceneMessage->m_uCameraHash)
				{
					pxCutscene->SetActiveCameraByHash(pxCutsceneMessage->m_uCameraHash);
				}
			}
			else
			{
				//client cutscene , so need to forward the message to the client entity instead
				Asura_Network::Forward(pxMessage,ASURA_GUID_MY_CLIENT_ENTITIES,GetGuid());
			}
			return true;
		}

		default:
		{
			break;
		}
	}
	
	Asura_Assert(0,"Unhandled message");
	return false;

	// Pass to parent?
	//return PARENT::HandleMessage(pxMessage);
}


#ifdef ASURA_TOOLS
unsigned short* Asura_ServerEntity_CutsceneController::GetCustomEvent(int iIndex)
{
	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
	if(!pxCutscene) return 0;

	if(iIndex == 0)
	{
		return &pxCutscene->m_usEndCutsceneMessageBlock;
	}

	//search for the appropriate trigger event
	int iTrigger = 0;
	for(int i=0; i<pxCutscene->NumberOfEvents; i++)
	{
		if(pxCutscene->Events[i]->GetType() == ACET_Trigger)
		{
			iTrigger++;
			if(iTrigger == iIndex)
			{
				return &(((Asura_Cutscene_Event_Trigger*) pxCutscene->Events[i])->m_sMessageBlock);
			}
		}
	}
	return NULL;
}

int Asura_ServerEntity_CutsceneController::GetNumCustomEvents() const
{
	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
	if(!pxCutscene) return 0;

	int iNumCustomEvents = 1;//first is for the end of cutscene event

	//search for any trigger events in the cutscene
	for(int i=0; i<pxCutscene->NumberOfEvents; i++)
	{
		if(pxCutscene->Events[i]->GetType() == ACET_Trigger) iNumCustomEvents++;
	}
	return iNumCustomEvents;
}

int Asura_ServerEntity_CutsceneController::GetCustomEventMessageSet(int iEventIndex) const
{
	return ASURA_STATICMESSAGE_CUTSCENESET;
}

void Asura_ServerEntity_CutsceneController::GetCustomEventName(int iEventIndex, Asura_String& xEventName) 
{ 
	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
	if(!pxCutscene) return;
	
	if(iEventIndex == 0)
	{
		xEventName = "End Cutscene"; 
	}
	else
	{
		//search for the appropriate trigger event
		int iTrigger = 0;
		for(int i=0; i<pxCutscene->NumberOfEvents; i++)
		{
			if(pxCutscene->Events[i]->GetType() == ACET_Trigger)
			{
				iTrigger++;
				if(iTrigger == iEventIndex)
				{
					Asura_Cutscene_Event* pxEvent = pxCutscene->Events[i];

					xEventName = pxEvent->m_xName.GetString();
					if(xEventName.IsBlank())
					{
						//the trigger event doesn't have a name , so just give one based on the time instead
						char acName[100];
						Asura_PrintfToBfr(acName, "%.3f Trigger", sizeof(acName), true, pxEvent->m_fTriggerTime);
						xEventName = acName;
						break;
					}
				}
			}
		}
	}
}

#endif

void Asura_ServerEntity_CutsceneController::Activate( bool bCutsceneStartsActive, float fTimer )
{
	if(m_uDisabled) return;

	//get the index for this cutscene
	Asura_Hash_ID uCutsceneHash = Asura_GetHashID(CutsceneName.GetString());

	// (MH - 03.10.2002) see if this needs to start a cutscene up on the client.
	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByHash(uCutsceneHash);
	Asura_Assert(pxCutscene, "failed to find cutscene");
	if (!pxCutscene) { return; }

	if(m_uTriggerMessagesOnly)
	{
		if(pxCutscene)
		{
			//trigger the messages
			pxCutscene->m_uEnabledEvents = m_uEnabledEvents;
			pxCutscene->SendAllMessages();
		}
	}
	else
	{
		if (pxCutscene && pxCutscene->PlaysOnServer)
		{
		    bool bPlaying = pxCutscene->IsPlaying();
		    Asura_Assertf(!bPlaying, ("cutscene '%s' already playing when activated", pxCutscene->CutsceneName.GetString()));
		    if (!bPlaying)
		    {
// (MP - 10.03.2003) We don't want to do this outside of tools. The level
// exports should contain the necessary data.
#ifdef ASURA_TOOLS
		        // (MH - 03.10.2002) LoadFiles and LoadTextures shouldn't be needed.
		        // Most of the time it will do nothing anyway.
		        pxCutscene->LoadFiles();
		        Asura_Texture_Management::LoadTextures();
#endif
                
		        pxCutscene->Initialise();
				pxCutscene->m_uEnabledEvents = m_uEnabledEvents;


                if( fTimer )
                {
                    pxCutscene->AdvanceToTime( fTimer );
                }
		    }
		}
		else
		{
		    //fill in the index in the data part of the message
		    Asura_Message* m = Asura_Network::CreateMessage( ASURA_MSG_CUTSCENECONTROLLER_ACTIVATE, sizeof(Asura_Msg_ActivateCutscene));
		    Asura_Msg_ActivateCutscene* pxData = (Asura_Msg_ActivateCutscene*)m->GetData();

	        pxData->m_uHash = uCutsceneHash;
			pxData->m_ucFlags = 0;
			
			if (bCutsceneStartsActive) 
			{
				pxData->m_ucFlags |= Asura_Msg_ActivateCutscene::FLAG_STARTS_ACTIVE;
			}
			if (m_uSkipMessagesAfterLoad)
			{
				pxData->m_ucFlags |= Asura_Msg_ActivateCutscene::FLAG_SKIP_MESSSAGES_AFTER_LOAD;
			}
			
			pxData->m_fTimer = fTimer;
			pxData->m_uEnabledEvents = m_uEnabledEvents;

		    m->To = ASURA_GUID_MY_CLIENT_ENTITIES;
		    m->From = Guid;

		    Asura_Network::Send(m);
		}

		m_uPlaying = true;
		// NY - 20 June 2008 - add us to the update list
		SetNeverUpdates(false);

		if(pxCutscene)
		{
			//initialse timer from cutscene , in case it had a start time
			m_fTimer = pxCutscene->Timer;
		}
		else
		{
			m_fTimer = fTimer;
		}
	}
	m_uCheckForCutsceneRestartAfterLoad = false;
	m_uSkipMessagesAfterLoad = false;
}

void Asura_ServerEntity_CutsceneController::Update()
{
	if(m_uPlaying)
	{
		m_fTimer += Asura_Timers::GetGameTimeSlice();
	}

    // server cutscene update used to be called here, but now it is done
    // by Asura_Cutscene_System::Update();
	if( m_uCheckStartActiveFlag )
	{
		m_uCheckStartActiveFlag = 0;
		Asura_Hash_ID uCutsceneHash = Asura_GetHashID(CutsceneName.GetString());

		Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByHash(uCutsceneHash);
		Asura_Assert(pxCutscene, "failed to find cutscene");
		if (!pxCutscene) { return; }

		if(pxCutscene->m_iCutsceneFlags & AsuraCutsceneFlag_StartsActive)
		{
			Activate( true, 0 );
		}
	}

	if(m_uCheckForCutsceneRestartAfterLoad)
	{
		//First frame after loading.
		//If the cutscene was flagged as playing when saved , then we need to reactivate
		//them at the appropriate point. RAR 09/06/03
		m_uCheckForCutsceneRestartAfterLoad = false;
		if(m_uPlaying)
		{
			Activate( false, m_fTimer);
		}
		m_uSkipMessagesAfterLoad = false;
	}

	if (!m_uPlaying)
	{
		//take us off the update list
		SetNeverUpdates(true);
	}
}


#ifdef ASURA_TOOLS
// (MH - 25.11.2002) static function - called from tools
void Asura_ServerEntity_CutsceneController::CutsceneRenamed(const Asura_String& xOldCutsceneName, const char* szNewCutsceneName)
{
    Asura_Classified_List_It<Asura_Entity_Server> xIt(Asura_Entity_Server::GetList());
    while (!xIt.Done())
    {
        Asura_Entity_Server* pxEntity = xIt.GetCurrent();
        if (pxEntity->GetClassification() == AsuraEntityClass_CutsceneController)
        {
            Asura_ServerEntity_CutsceneController* pxController = reinterpret_cast
                <Asura_ServerEntity_CutsceneController*>(pxEntity);
            if (pxController->CutsceneName == xOldCutsceneName)
            {
                pxController->CutsceneName = szNewCutsceneName;
            }
        }
        xIt.Next();
    }
}
#endif

int Asura_ServerEntity_CutsceneController::GetChunkBufferSize()
{
	int iChunkSize = Asura_Entity_Server::GetChunkBufferSize();

    ASURA_CHUNK_SIZE( iChunkSize, CHUNK_VERSION, u_int );
    ASURA_CHUNK_SIZE_STRING( iChunkSize, CutsceneName );
    ASURA_CHUNK_SIZE( iChunkSize, m_uStateFlags, u_int );
    ASURA_CHUNK_SIZE( iChunkSize, m_fTimer, float );
    ASURA_CHUNK_SIZE( iChunkSize, m_uEnabledEvents, u_int );

    return iChunkSize;
}

void Asura_ServerEntity_CutsceneController::FillInChunkBuffer( char*& pcChunkBuffer )
{
	UpdatePlayingFlag();

    ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, CHUNK_VERSION, u_int );
	Asura_Entity_Server::FillInChunkBuffer(pcChunkBuffer);

    ASURA_CHUNK_FILL_STRING( pcChunkBuffer, CutsceneName );
    ASURA_CHUNK_BITFIELD_FILLANDVERIFY( pcChunkBuffer, m_uStateFlags, u_int );
    ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_fTimer, float );
    ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_uEnabledEvents, u_int );
}

void Asura_ServerEntity_CutsceneController::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	Asura_ServerEntity_CutsceneController*		pxHack			= const_cast<Asura_ServerEntity_CutsceneController*>(this);
	const u_int									uChunkVersion	= CHUNK_VERSION;


	// Ooh, nasty...

	pxHack->UpdatePlayingFlag();

	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xStateFlagsBitfield (m_uStateFlags);


	// Create the bitfields...

	xStateFlagsBitfield.Add	(m_uCheckStartActiveFlag,				1);
	xStateFlagsBitfield.Add	(m_uDisabled,							1);
	xStateFlagsBitfield.Add	(m_uManuallyDisabled,					1);
	xStateFlagsBitfield.Add	(m_uPlaying,							1);
	xStateFlagsBitfield.Add	(m_uCheckForCutsceneRestartAfterLoad,	1);
	xStateFlagsBitfield.Add	(m_uTriggerMessagesOnly,				1);
	xStateFlagsBitfield.Add (m_uSkipMessagesAfterLoad,				1);


	// Write the version
    xStream << uChunkVersion;

	// Write the parent data
	Asura_Entity_Server::WriteToChunkStream(xStream);


	// Write the data...
    xStream << CutsceneName;
    xStream << xStateFlagsBitfield;
    xStream << m_fTimer;
    xStream << m_uEnabledEvents;
}

void Asura_ServerEntity_CutsceneController::ReadFromChunkStream (Asura_Chunk_Stream& xStream)
{
	//check the version
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert( CHUNK_VERSION >= uVersion, "Chunk version is too new; your code is out of date wrt this data." );

	//read in parent class data if new enough
	if(uVersion >= 7)
	{
		Asura_Entity_Server::ReadFromChunkStream(xStream);
	}

	xStream >> CutsceneName;

	m_uStateFlags = 0;
	if(uVersion >= 1)
	{
		xStream >> m_uStateFlags;

		if(uVersion == 1)
		{
			//make sure the unused bits are zeroes
			bool bStartActive = m_uCheckStartActiveFlag;
			m_uStateFlags = 0;
			m_uCheckStartActiveFlag = bStartActive;
		}
		else if(uVersion == 2)
		{
			// zero the previously unset m_uManuallyDisabled flag
			m_uManuallyDisabled = 0;
		}
	}
	else
	{
		m_uCheckStartActiveFlag = 1;
	}

	if(uVersion >= 4)
	{
		xStream >> m_fTimer;
	}
	else
	{
		m_fTimer = 0;
	}
	if(uVersion >= 5)
	{
		xStream >> m_uEnabledEvents;
	}
	if(uVersion < 6)
	{
		//version 5 may contain junk here.
		m_uEnabledEvents = 0xffffffff;
	}
	//Update the enabled events status for the associated cutscene.
	//This is needed so that  cutscenes that don't get started via their cutscene controller 
	//get set correctly. (Namely cutscenes that are attachments).
	//This will only work correctly for server side cutscenes , but its only server cutscenes that
	//have this issue at the moment. RAR 27/07/2005
	Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
	if(pxCutscene)
	{
		pxCutscene->m_uEnabledEvents = m_uEnabledEvents;

#if defined(ASURA_USE_STREAMING)
		pxCutscene->m_uGuid = GetGuid();
#endif //defined(ASURA_USE_STREAMING)
	}

	m_uCheckForCutsceneRestartAfterLoad = m_uPlaying;
	m_uSkipMessagesAfterLoad = false;
}

void Asura_ServerEntity_CutsceneController::SetDisabled( bool bDisable )
{
	/// Richard May - 03/06/2003 15:01:23 - can only be enabled if the cutscene was
	/// disabled manually (in code, rather than on export) 
	if( !bDisable && m_uManuallyDisabled )
	{
		m_uManuallyDisabled = bDisable;
		m_uDisabled = bDisable;
	}
	else if(bDisable) // otherwise, disable manually
	{
		m_uManuallyDisabled = bDisable;
		m_uDisabled = bDisable;
	}	
}

void Asura_ServerEntity_CutsceneController::UpdatePlayingFlag()
{
	if(m_uPlaying && !m_uCheckForCutsceneRestartAfterLoad)
	{
		//get the cutscene
		Asura_Hash_ID uCutsceneHash = Asura_GetHashID(CutsceneName.GetString());

		Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByHash(uCutsceneHash);
		Asura_Assert(pxCutscene, "failed to find cutscene");
		if (!pxCutscene) { return; }

		if(pxCutscene->PlaysOnServer)
		{
			//If the cutscene plays on the server , we can just check it directly to see if it is playing
			m_uPlaying = pxCutscene->IsPlaying();

			// (AAM - 28.07.2005) If cutscene is available, sync timer too
			if( pxCutscene->IsPlaying() )
			{
				m_fTimer = pxCutscene->Timer;
			}
		}
		else if(Asura_Network::IsClient())
		{
			//This is a client cutscene , and we have a client on this machine.
			//Check the cutscene directly (this is to avoid problems with cutscenes that
			//have been skipped)
			m_uPlaying = pxCutscene->IsPlaying();

			// (AAM - 28.07.2005) If cutscene is available, sync timer too
			if( pxCutscene->IsPlaying() )
			{
				m_fTimer = pxCutscene->Timer;
			}
		}
		else
		{
			//Just guess the playing state of the cutscene based on the time since it started
			if(!pxCutscene->WouldCutsceneBePlayingAtThisTime(m_fTimer))
			{
				m_uPlaying = false;
			}
		}
		if(!m_uPlaying) m_fTimer = 0;
	}
}

#ifdef ASURA_TOOLS
bool Asura_ServerEntity_CutsceneController::ShouldIncludeInLevelExport( void ) const
{
    if( !PARENT::ShouldIncludeInLevelExport() ) { return false; }

    // Check if our cutscene is being saved.
	const Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutscene( CutsceneName.GetString() );
	if( pxCutscene && !pxCutscene->ShouldSave() )
    {
        return false;
    }

    return true;
}
#endif

//////////////////
// ReInitialise //
//////////////////

// MGH 16/1/2007: HP5 is using this function to restore the state of the cutscenes after a streaming zone is reloaded.

// Re-initialise the cutscene controller, skipping forward to the specified time if necessary.
// Don't send any static messages that would otherwise be triggered when advancing forward through time.
void Asura_ServerEntity_CutsceneController::ReInitialise(float fTimer, u_int uEnabledEvents, bool bPlaying, bool bDisabled,
														 bool bCheckStartActiveFlag /*= false*/)
{
	#ifdef ASURA_USE_STREAMING
	// Handlers aren't all set correctly, so we need to queue the messsage to the client until later.
	const bool bOldShouldQueue = Asura_Streaming_System::GetShouldQueueMessages();
	Asura_Streaming_System::SetShouldQueueMessages(true);

	// We still need to set the handler for the cutscene system as we need to look up the cutscene.
	Asura_Cutscene_Handler& xOldHandler = Asura_Cutscene_System::GetCurrentHandler();
	Asura_Cutscene_System::SetCurrentHandler(GetZoneHashID());
	if (&Asura_Cutscene_System::GetCurrentHandler() == 0)
	{
		Asura_Assert(0, "Failed to set the cutscene handler");
		Asura_Cutscene_System::SetCurrentHandler(&xOldHandler);
	}
	#endif //ASURA_USE_STREAMING

	Asura_Cutscene* const pxCutscene = Asura_Cutscene_System::GetCutscene(CutsceneName.GetString());
	if (pxCutscene)
	{
		// Stop the cutscene if it is currently playing.
		if (m_uPlaying && pxCutscene->IsPlaying())
		{	
			Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
			if (pxMessage)
			{
				pxMessage->SetFromGuid(Guid);
				pxMessage->SetToGuid(Guid);
				pxMessage->SentTime = Asura_Timers::GetServerAbsoluteTime();
				HandleMessage(pxMessage);
				Asura_Network::DestroyMessage(pxMessage);
			}
		}

		if (bCheckStartActiveFlag && (pxCutscene->m_iCutsceneFlags & AsuraCutsceneFlag_StartsActive))
		{
			// Start the cutscene.
			bPlaying = true;
		}
	}

	#ifdef ASURA_USE_STREAMING
	// Restore the settings and handlers.
	Asura_Streaming_System::SetShouldQueueMessages(bOldShouldQueue);
	Asura_Cutscene_System::SetCurrentHandler(&xOldHandler);
	#endif //ASURA_USE_STREAMING

	m_uCheckStartActiveFlag = 0;
	
	m_uCheckForCutsceneRestartAfterLoad = 0;
	m_uEnabledEvents = uEnabledEvents;

	m_fTimer = fTimer;
	m_uPlaying = bPlaying ? 1 : 0;
	m_uDisabled = bDisabled ? 1 : 0;

	if (bPlaying && !bDisabled)
	{
		// MGH 16/1/2007: Restart the cutscene at the specified time, but don't resend any static messages.
		// Basically we want to fast forward through the cutscene to the given time replaying any
		// events, but skipping the sending of any static messages to other entities,
		// because these messages have already been sent before we saved and we don't want to send them
		// again upon loading.
		m_uCheckForCutsceneRestartAfterLoad = true;
		m_uSkipMessagesAfterLoad = true;
	}
}
