//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_DynamicLight.cpp
//
// Copyright (c) 2007 Rebellion (Derby). All rights reserved.
//
// Description:
//
// The all-new dynamic light entity
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_TARGET_SM3

// This file's header.
#include "Entity/DynamicLight/Asura_ServerEntity_DynamicLight.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/DynamicLight/Asura_ClientEntity_DynamicLight.h"
#include "Network/Asura_Message.h"


// asura tools
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#include "Tools_System.h"
#endif


Asura_Hashed_List<Asura_ServerEntity_DynamicLight> Asura_ServerEntity_DynamicLight::s_xList;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#ifdef ASURA_TOOLS
Asura_ServerEntity_DynamicLight* Asura_ServerEntity_DynamicLight::CreateFromTools( Asura_Guid uGuid , const Asura_Vector_3& xCentre)
{
	Asura_ServerEntity_DynamicLight* pxNew = new Asura_ServerEntity_DynamicLight(uGuid);
	pxNew->m_xParams.m_xOrientation = *Asura_Camera::GetOrientation();
	pxNew->m_xParams.m_xPosition = xCentre;
	return pxNew;


}
#endif

Asura_ServerEntity_DynamicLight::Asura_ServerEntity_DynamicLight( Asura_Guid uGuid)
: PARENT( uGuid )
, m_xParams()
{
	Classification = AsuraEntityClass_DynamicLight;
	SetNeverUpdateClient();

	s_xList.AddToList( this );
}

Asura_ServerEntity_DynamicLight::~Asura_ServerEntity_DynamicLight( void )
{
	s_xList.RemoveFromList( this );
}

void Asura_ServerEntity_DynamicLight::CreateFromChunkStream( Asura_Guid uGuid, Asura_Chunk_Stream &xStream )
{
	Asura_ServerEntity_DynamicLight* pxNew = new Asura_ServerEntity_DynamicLight( uGuid);
	pxNew->ReadFromChunkStream(xStream);
}

void Asura_ServerEntity_DynamicLight::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert( uVersion <= DYNLIGHT_CHUNK_VERSION, "Unhandled Asura_ServerEntity_DynamicLight chunk version" );

	if (uVersion >= 11)
	{
		PARENT::ReadFromChunkStream(xStream);
	}

	ExtractParamsFromChunkStream(m_xParams, xStream, uVersion );

	if (uVersion >= 12)
	{
		bool bHasClientState;
		xStream >> bHasClientState;

		if (bHasClientState)
		{
			// Read the client state data, but the server doesn't care about it - it's only the client that needs it
			Asura_ClientEntity_DynamicLight_State xState;
			xState.ExtractFromChunkStream(xStream, uVersion);
		}		
	}

	SetNeverUpdates(true);
}

void Asura_ServerEntity_DynamicLight::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << DYNLIGHT_CHUNK_VERSION;

	PARENT::WriteToChunkStream(xStream);


	// Write constant parameters to the chunk stream
	WriteParamsToChunkStream(m_xParams, xStream);

	// If we can, write the current state (from the client) to the chunk stream
	Asura_ClientEntity_DynamicLight *pxClient = NULL;
	// only bother saving client state for in-game (checkpoint) save
#ifdef ASURA_TOOLS
	if (!Tools_System::IsActive())
#endif
	{
		pxClient = Asura_ClientEntity_DynamicLight::FindEntity(GetGuid());
	}

	bool bHasClientState = pxClient?true:false;
	xStream << bHasClientState;

	if (bHasClientState)
	{
		Asura_ClientEntity_DynamicLight_State *pxState = pxClient->GetState();
		pxState->WriteToChunkStream(xStream);
	}	

}


void Asura_ServerEntity_DynamicLight::ExtractParamsFromChunkStream( Asura_DynamicLight_Parameters &xParams , Asura_Chunk_Stream &xStream, u_int uVersion )
{
	// V1 was the tools UI prototype, saving an empty chunk. V2 is the first real version of the chunk
	if (uVersion > 1)
	{
		int iType;
		xStream >> iType;
		xParams.m_eType = static_cast<ASURA_DYNAMICLIGHT_TYPE>(iType);

		xStream >> xParams.m_uFlags;
		xStream >> xParams.m_xPosition;
		xStream >> xParams.m_xOrientation;
		xStream >> xParams.m_fFadeoutStart;
		xStream >> xParams.m_fFadeoutDist;

		if (uVersion >= 11)
		{
			u_int uBoolVals;
			xStream >> uBoolVals;
			for( u_int u = 0; u < 3; ++u )
			{
				xParams.m_abSwing[u] = ((uBoolVals & (1 << u)) != 0)?true:false;
			}
			xStream >> xParams.m_xRotSpeed;
			xStream >> xParams.m_xSwingDist;
		}
		else
		{
			//fixup old data
			float fTemp;
			for( u_int u = 0; u < 3; ++u )
			{
				int iBoolVal;
				xStream >> iBoolVal;
				xParams.m_abSwing[u] = (iBoolVal!=0)?true:false;
				xStream >> fTemp;
				xParams.m_xRotSpeed[u] = Asura_Maths::DegToRad(fTemp);
				xStream >> fTemp;
				xParams.m_xSwingDist[u] = Asura_Maths::DegToRad(fTemp);
			}
		}
		xStream >> xParams.m_fRotationOffset;

		xStream >> xParams.m_fPhotonRange;

		if (uVersion >= 11)
		{
			xStream >> xParams.m_fLimitingAngleInner;
			xStream >> xParams.m_fLimitingAngleOuter;
		}
		else if (uVersion>7)
		{
			float fTemp;
			xStream >> fTemp;
			xParams.m_fLimitingAngleInner = Asura_Maths::DegToRad(fTemp);
			xStream >> fTemp;
			xParams.m_fLimitingAngleOuter = Asura_Maths::DegToRad(fTemp);
		}


		if (uVersion>8)
		{
#ifdef ASURA_TOOLS
			xStream >> xParams.m_xProjCubeTextureName;
#else // #ifdef ASURA_TOOLS
			Asura_String xDummy;
			xStream >> xDummy;
#endif // #ifdef ASURA_TOOLS
			xStream >> xParams.m_xProjCubeTextureID;
		}

		if (uVersion>9)
		{
#ifdef ASURA_TOOLS
			xStream >> xParams.m_xProjCubeBlurredTextureName;
#else // #ifdef ASURA_TOOLS
			Asura_String xDummy;
			xStream >> xDummy;
#endif // #ifdef ASURA_TOOLS
			xStream >> xParams.m_xProjCubeBlurredTextureID;
			xStream >> xParams.m_fProjCubeLerpDist;
		}

		if (uVersion>2)
		{
			// Oops.. forgot to store this in v2!
			xStream >> xParams.m_xLightVolume;
		}

		xStream >> xParams.m_xLightVolumeOrient;
		xStream >> xParams.m_uTextureHash;
		xStream >> xParams.m_uDetailTextureHash;
		xStream >> xParams.m_fDetailTextureScale;
		xStream >> xParams.m_fMinRange;
		xStream >> xParams.m_fMaxRange;
		xStream >> xParams.m_iBlendMode;

		if (uVersion >= 11)
		{
			xStream >> xParams.m_fFoV;
			xStream >> xParams.m_fInnerFoV;
		}
		else
		{
			float fTemp;
			xStream >> fTemp;
			xParams.m_fFoV = Asura_Maths::DegToRad(fTemp);
			xStream >> fTemp;
			xParams.m_fInnerFoV = Asura_Maths::DegToRad(fTemp);
		}

		xStream >> xParams.m_fScale;
		xStream >> xParams.m_fAspect;
		xStream >> xParams.m_uMasterEntity;
		xStream >> xParams.m_xOffsetFromMaster;
		xStream >> xParams.m_uMasterMarker;

		if (uVersion>6)
		{
			xStream >> xParams.m_uAttachedLightShaft;
		}

		xStream >> xParams.m_iShadowMode;
		xStream >> xParams.m_fCoronaRadius;
		xStream >> xParams.m_uCoronaTextureHash;

		if (uVersion>5)
		{
			xStream >> xParams.m_fCoronaVolumetricDepthScale;
		}
		else
		{
			xParams.m_fCoronaVolumetricDepthScale = Asura_Corona_System::GetDefaultVolumetricDepthScale();
		}

		xStream >> xParams.m_fFalloffInnerRatio;
		xStream >> xParams.m_fFalloffPower;
		xStream >> xParams.m_fBaseIntensity;
		xStream >> xParams.m_fBaseR;
		xStream >> xParams.m_fBaseG;
		xStream >> xParams.m_fBaseB;
		xStream >> xParams.m_fGraphDuration;
		xStream >> xParams.m_fGraphTimerOffset;
		xStream >> xParams.m_fStrobeMinTimeOff;
		xStream >> xParams.m_fStrobeMaxTimeOff;
		xStream >> xParams.m_fStrobeMinTimeOn;
		xStream >> xParams.m_fStrobeMaxTimeOn;

		if (uVersion>3)
		{
			xStream >> xParams.m_fUScrollSpeed;
			xStream >> xParams.m_fVScrollSpeed;
			xStream >> xParams.m_fUScrollRange;
			xStream >> xParams.m_fVScrollRange;
		}
		else
		{
			xParams.m_fUScrollSpeed = xParams.m_fVScrollSpeed = 0.0f;
			xParams.m_fUScrollRange = xParams.m_fVScrollRange = 0.0f;
		}

		if (uVersion>4)
		{
			xStream >> xParams.m_xLimitingPlaneDirection;
			xStream >> xParams.m_xLimitingPlanePosition;
		}
		else
		{
			xParams.m_xLimitingPlaneDirection = Asura_Vector_3(0,-1,0);
			xParams.m_xLimitingPlanePosition  = Asura_Vector_3(0,0,0);
		}

		// Texture names
#ifdef ASURA_TOOLS
		xStream >> xParams.m_xTextureName;
		xStream >> xParams.m_xDetailTextureName;
		xStream >> xParams.m_xCoronaTextureName;
#else // #ifdef ASURA_TOOLS
		Asura_String xDummy;
		xStream >> xDummy;
		xStream >> xDummy;
		xStream >> xDummy;
#endif // #ifdef ASURA_TOOLS

		// Graphs
		xParams.m_xColourRangeGraph.ReadFromChunkStream(xStream);
		xParams.m_xFlickerGraph.ReadFromChunkStream(xStream);


		// Shadow params
		xParams.m_xShadowParams.ReadFromChunkStream(xStream);
	}
}




void Asura_ServerEntity_DynamicLight::WriteParamsToChunkStream( const Asura_DynamicLight_Parameters &xParams , Asura_Chunk_Stream& xStream)
{
	Asura_String xNullString;
	xStream << xParams.m_eType;
	xStream << xParams.m_uFlags;
	xStream << xParams.m_xPosition;
	xStream << xParams.m_xOrientation;
	xStream << xParams.m_fFadeoutStart;
	xStream << xParams.m_fFadeoutDist;

	u_int uBoolVals = 0;
	for( u_int u = 0; u < 3; ++u )
	{
		if (xParams.m_abSwing[u])
		{
			uBoolVals += 1 << u;
		}
	}
	xStream << uBoolVals;
	xStream << xParams.m_xRotSpeed;
	xStream << xParams.m_xSwingDist;
	xStream << xParams.m_fRotationOffset;

	xStream << xParams.m_fPhotonRange;
	xStream << xParams.m_fLimitingAngleInner;
	xStream << xParams.m_fLimitingAngleOuter;
#ifdef ASURA_TOOLS
	xStream << xParams.m_xProjCubeTextureName;
#else // #ifdef ASURA_TOOLS
	xStream << xNullString;
#endif // #ifdef ASURA_TOOLS
	xStream << xParams.m_xProjCubeTextureID;
#ifdef ASURA_TOOLS
	xStream << xParams.m_xProjCubeBlurredTextureName;
#else // #ifdef ASURA_TOOLS
	xStream << xNullString;
#endif // #ifdef ASURA_TOOLS
	xStream << xParams.m_xProjCubeBlurredTextureID;
	xStream << xParams.m_fProjCubeLerpDist;
	xStream << xParams.m_xLightVolume;
	xStream << xParams.m_xLightVolumeOrient;
	xStream << xParams.m_uTextureHash;
	xStream << xParams.m_uDetailTextureHash;
	xStream << xParams.m_fDetailTextureScale;
	xStream << xParams.m_fMinRange;
	xStream << xParams.m_fMaxRange;
	xStream << xParams.m_iBlendMode;
	xStream << xParams.m_fFoV;
	xStream << xParams.m_fInnerFoV;
	xStream << xParams.m_fScale;
	xStream << xParams.m_fAspect;
	xStream << xParams.m_uMasterEntity;
	xStream << xParams.m_xOffsetFromMaster;
	xStream << xParams.m_uMasterMarker;
	xStream << xParams.m_uAttachedLightShaft;
	xStream << xParams.m_iShadowMode;
	xStream << xParams.m_fCoronaRadius;
	xStream << xParams.m_uCoronaTextureHash;
	xStream << xParams.m_fCoronaVolumetricDepthScale;
	xStream << xParams.m_fFalloffInnerRatio;
	xStream << xParams.m_fFalloffPower;
	xStream << xParams.m_fBaseIntensity;
	xStream << xParams.m_fBaseR;
	xStream << xParams.m_fBaseG;
	xStream << xParams.m_fBaseB;
	xStream << xParams.m_fGraphDuration;
	xStream << xParams.m_fGraphTimerOffset;
	xStream << xParams.m_fStrobeMinTimeOff;
	xStream << xParams.m_fStrobeMaxTimeOff;
	xStream << xParams.m_fStrobeMinTimeOn;
	xStream << xParams.m_fStrobeMaxTimeOn;

	xStream << xParams.m_fUScrollSpeed;
	xStream << xParams.m_fVScrollSpeed;
	xStream << xParams.m_fUScrollRange;
	xStream << xParams.m_fVScrollRange;

	xStream << xParams.m_xLimitingPlaneDirection;
	xStream << xParams.m_xLimitingPlanePosition;

	// Texture names
#ifdef ASURA_TOOLS
	xStream << xParams.m_xTextureName;
	xStream << xParams.m_xDetailTextureName;
	xStream << xParams.m_xCoronaTextureName;
#else // #ifdef ASURA_TOOLS
	xStream << xNullString;
	xStream << xNullString;
	xStream << xNullString;
#endif // #ifdef ASURA_TOOLS

	// Graphs
	bool bSamplesOnly = true;
#ifdef ASURA_TOOLS
	bSamplesOnly = Tools_Level_Export::IsLevelExport();
#endif

	xParams.m_xColourRangeGraph.WriteToChunkStream(xStream, bSamplesOnly);
	xParams.m_xFlickerGraph.WriteToChunkStream(xStream, bSamplesOnly);

	// Shadow params
	xParams.m_xShadowParams.WriteToChunkStream(xStream);
}






bool Asura_ServerEntity_DynamicLight::HandleMessage(const Asura_Message* pxMessage)
{

	switch (pxMessage->GetMsgID())
	{

		case ASURA_MSG_TRIGGER:
		case ASURA_MSG_UNTRIGGER:
		case ASURA_MSG_DYNAMICLIGHT_FADE:
			// The client handles the light effects. Might need to update this to track this
			// state in the server, too, so we can save the light state, if it's needed by savegames?
			Asura_Network::Forward(pxMessage,ASURA_GUID_MY_CLIENT_ENTITIES,GetGuid());
			return true;

		default:
			return PARENT::HandleMessage(pxMessage);

	}

}



#ifdef ASURA_TOOLS
void Asura_ServerEntity_DynamicLight::SetParams( const Asura_DynamicLight_Parameters& xNewParams )
{
    m_xParams = xNewParams;

	Asura_ClientEntity_DynamicLight *pxClient = static_cast<Asura_ClientEntity_DynamicLight*>(Asura_ClientEntity_DynamicLight::FindEntity(GetGuid()));
	if (pxClient)
	{
		pxClient->SetParams( m_xParams );

		//NY - inelegant but effective way of keeping the omnipresent client entity up to date...
		if (IsEntityActive())
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_ACTIVATE );
			if( pxMessage )
			{
				pxMessage->UseGuaranteedDelivery();
				Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
			}
		}
		else
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_DEACTIVATE );
			if( pxMessage )
			{
				pxMessage->UseGuaranteedDelivery();
				Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
			}
		}
	}
}
#endif


#ifdef ASURA_TOOLS

void Asura_ServerEntity_DynamicLight::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	PARENT::ApplyChangesForTemplateInstance(pxInstance);

	m_xParams.m_xPosition			= pxInstance->TransformPosition(m_xParams.m_xPosition);
	m_xParams.m_xOrientation		= pxInstance->TransformMatrix(m_xParams.m_xOrientation);
	m_xParams.m_uMasterEntity		= pxInstance->ConvertGuid(m_xParams.m_uMasterEntity);
	m_xParams.m_uAttachedLightShaft	= pxInstance->ConvertGuid(m_xParams.m_uAttachedLightShaft);
	m_xParams.m_xOffsetFromMaster	= m_xParams.m_xOffsetFromMaster * pxInstance->GetRelativeMatrix();
}

void Asura_ServerEntity_DynamicLight::TraverseUsedTextures( void (*pfnCallback)( int iTextureID,
	void* pUserData ), void* pUserData ) const
{
	PARENT::TraverseUsedTextures( pfnCallback, pUserData );

	int iTexID = Asura_Texture_Management::GetIndexFromHashID( m_xParams.m_uTextureHash );
	pfnCallback( iTexID, pUserData );

	iTexID = Asura_Texture_Management::GetIndexFromHashID( m_xParams.m_uDetailTextureHash );
	pfnCallback( iTexID, pUserData );

	iTexID = Asura_Texture_Management::GetIndexFromHashID( m_xParams.m_uCoronaTextureHash ),
	pfnCallback( iTexID, pUserData );

	iTexID = Asura_Texture_Management::GetIndexFromHashID( m_xParams.m_xProjCubeTextureID ),
	pfnCallback( iTexID, pUserData );

	iTexID = Asura_Texture_Management::GetIndexFromHashID( m_xParams.m_xProjCubeBlurredTextureID ),
	pfnCallback( iTexID, pUserData );
}

#endif // ASURA_TOOLS


#endif //#ifdef ASURA_TARGET_SM3

