//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity_Ephemeron.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     Base class for short-lived client-only objects with a shape instance.
//
// History:
//
//     (MH  - 28.11.2002) - Created. (Ephemeron is from the Greek for mayfly). 
//
//	   SJL - 13.4.2007 - 
//
// Added these accessors:
//
// const float& GetVelocityDecayRate() const { return m_fVelocityDecayRate; }
// void  SetVelocityDecayRate( float fVelocityDecayRate ) { m_fVelocityDecayRate = fVelocityDecayRate; }
// const Asura_Vector_3& GetAcceleration() const { return m_xAcceleration; } 
// void  SetAcceleration(const Asura_Vector_3& xAcceleration ) { m_xAcceleration = m_xAcceleration; }
// const Asura_Euler_3& GetSpinRate() const { return m_xSpinRate; }
// void  SetSpinRate( const Asura_Euler_3& xSpinRate) { m_xSpinRate = xSpinRate; }
//
// Added  these members:
//
// Asura_Vector_3 m_xAcceleration;
// float m_fVelocityDecayRate;
//
// Made small changes to constructor to initliase these  to the old hardcoded values that where used in the update loop previously.
//
// Added a clamp to where decay was applied in case of v low framerate spikes that could potentaily cause ephemera to move backwards.
//
//
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura.h"
#include "Entity/Ephemera/Asura_Entity_Ephemeron.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Physics/Asura_Physics.h"

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR float dbg_fEphemeronElasticity = 0.45f;
DEBUGVAR float dbg_fEphemeronVelocityChangeFudge = 0.3f;
DEBUGVAR float dbg_fEphemeronVelocityDecayRate = 2.0f;
DEBUGVAR float dbg_fEphemeronMaxVelocityY = -0.8f;  // if velocity greater than this at impact, hit the floor and stop.
DEBUGVAR float dbg_fEphemeronFadeoutTime = 1.0f;

// C O N S T R U C T O R   &   D E S T R U C T O R   /////////////////////////
Asura_Entity_Ephemeron::Asura_Entity_Ephemeron(Asura_Hash_ID uShapeHashID, 
                                               const Asura_Vector_3& xPos, 
                                               const Asura_Matrix_3x3& xOrient, 
                                               const Asura_Vector_3& xInitialVelocity, 
                                               float fMaximumSpinRate, 
                                               float fLifeTime) 
{
	Classification = AsuraEntityClass_Ephemeron;

	// setup position, orientation and spin
    m_xCurrentPosition = xPos;
    m_xCurrentOrientation = xOrient;

	fMaximumSpinRate *= 2.0f;
    m_xSpinRate = Asura_Euler_3( (Asura_Random::GetFloat()-0.5f) * fMaximumSpinRate, 
                                 (Asura_Random::GetFloat()-0.5f) * fMaximumSpinRate, 
                                 (Asura_Random::GetFloat()-0.5f) * fMaximumSpinRate );
	m_xSpinAngle.FromMatrix(xOrient);

	m_xVelocity = xInitialVelocity;
	m_xAcceleration = Asura_Vector_3(0.0f,Asura_Physics_System::GetGravityScalar(),0.0f);

    m_fLifeTimer = fLifeTime;

	// setup the shape instance
	m_xShapeInstance.SetShape(uShapeHashID);

    m_bActive = true;
	m_usFlagsOfPolysToIgnore = AsuraPolygonFlag_NoCollisions;
	
	m_fVelocityDecayRate = dbg_fEphemeronVelocityDecayRate;
	// add to the correct lists
	AddToUpdateableClientList(this);	
}

Asura_Entity_Ephemeron::~Asura_Entity_Ephemeron()
{
	RemoveFromUpdateableClientList(this);
}


/****************************************************************************
 *
 * Name          : Asura_Entity_Ephemeron::Update
 *
 * Author        : Mike Healey, based on Richard May's shell casing code
 *
 * Description   : Main update function (called every frame)
 *
 ****************************************************************************/
void Asura_Entity_Ephemeron::Update ( void )
{
	float fTimeSlice = Asura_Timers::GetGameTimeSlice();	

	if(m_bActive) 
	{		
		// apply gravity/acceleration
		m_xVelocity += m_xAcceleration*fTimeSlice;

		// update object's position with velocity
		Asura_Vector_3 xPrevVelocity = m_xVelocity;
		Asura_Physics_System::MoveParticle( m_xCurrentPosition, m_xVelocity, dbg_fEphemeronElasticity, m_usFlagsOfPolysToIgnore );	

		// scale rate of spin by velocity magnitude
		float fMag = m_xVelocity.Magnitude();
		Asura_Euler_3 xScaledSpin;
		
		xScaledSpin.x = m_xSpinRate.x * fMag; 
		xScaledSpin.y = m_xSpinRate.y * fMag; 
		xScaledSpin.z = m_xSpinRate.z * fMag; 

		// basic impact test
        // if the velocity has changed drastically, assume we've had some kind of impact
        if( !m_xVelocity.ApproximatelyEquals( xPrevVelocity, dbg_fEphemeronVelocityChangeFudge ) )
		{
            DealWithImpact();
		}

		// Decay velocity, but we may not lose more velocity than we have so clamp this below or at 1.0
		float fFractionOfVelocityToLose = Asura_Maths::Min(1.0f, fTimeSlice * GetVelocityDecayRate());
		m_xVelocity -= m_xVelocity * fFractionOfVelocityToLose;        

		// apply the spin
		m_xSpinAngle.x += xScaledSpin.x * fTimeSlice;
		m_xSpinAngle.y += xScaledSpin.y * fTimeSlice;
		m_xSpinAngle.z += xScaledSpin.z * fTimeSlice;
		
		m_xSpinAngle.ToMatrix(m_xCurrentOrientation);

#ifdef ASURA_NO_NETWORK_LAG_COMPENSATION
		m_xCurrentVelocity = m_xVelocity;
#endif //ASURA_NO_NETWORK_LAG_COMPENSATION
	}
	
	// life timer
	m_fLifeTimer -= fTimeSlice;
	if( m_fLifeTimer <= 0.0f )
	{
		SetAsDestroyed();
	}
    else
    {
        if( m_fLifeTimer < dbg_fEphemeronFadeoutTime )
        {
            float fAlpha = m_fLifeTimer / dbg_fEphemeronFadeoutTime;
            m_xShapeInstance.SetAlpha(fAlpha);
        }
    }
}
/* end of function Asura_Entity_Ephemeron::Update */

void Asura_Entity_Ephemeron::DealWithImpact()
{
    MakeImpactSound();

    // if the upward y velocity is less than a small amount then just drop us on the floor
    if (m_xVelocity.y > dbg_fEphemeronMaxVelocityY) 
    {
	    m_xSpinRate.Zero();
	    
	    m_bActive = false;
	    m_xVelocity.Zero();
    }
}


/*************************************************************
 *
 * Name          : Asura_Entity_Ephemeron::Render
 *
 * Author        : Mike Healey, based on Richard May's code
 *
 * Description   : Main display function
 *
 *************************************************************/
void Asura_Entity_Ephemeron::Render ( void )
{
	if (m_bIsEntityCurrentlyMoving)
	{
		// If the entity is moving, then update its position.
		m_xShapeInstance.SetNewPosAndOri(m_xCurrentPosition, m_xCurrentOrientation);
	}

	m_xShapeInstance.Render();
}
/* end of function Render */

// E N D   O F   F I L E /////////////////////////////////////////////////////
