
////////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_FMVController.cpp
// Author    : Jason Turner
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/FMVController/Asura_FMVController_Messages.h"
#include "Entity/FMVController/Asura_ServerEntity_FMVController.h"
#include "Entity/FMVController/Asura_ServerEntity_FMVSpeaker.h"
#include "FMV/Asura_VideoToTexture.h"
#include "Network/Asura_Message.h"

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#endif

Asura_Hashed_List< Asura_ServerEntity_FMVController > Asura_ServerEntity_FMVController::s_xList;

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

Asura_ServerEntity_FMVController::Asura_ServerEntity_FMVController(Asura_Guid uGuid)
: Asura_ServerEntity_Activatable(uGuid)
, m_uFlags(0)
, m_xLinkedSpeakerGuids(GuidCompareFunction)
{
	Classification = AsuraEntityClass_FMVController;
	s_xList.AddToList( this );
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

Asura_ServerEntity_FMVController::~Asura_ServerEntity_FMVController()
{
	Asura_TemplateTree_ConstIterator<Asura_Guid, Asura_Guid> xIterator(m_xLinkedSpeakerGuids);

	while(!xIterator.Done())
	{
		const Asura_Guid* puSpeakerGuid = xIterator.GetCurrent();

		if (puSpeakerGuid)
		{
			// Check the speaker actually exists
			Asura_ServerEntity_FMVSpeaker* pxSpeaker = Asura_ServerEntity_FMVSpeaker::FindEntity(*puSpeakerGuid);
			if(pxSpeaker)
			{
				pxSpeaker->RemoveLinkedController(GetGuid());
			}
		}

		xIterator.Next();
	}

	s_xList.RemoveFromList( this );
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ServerEntity_FMVController::CreateFromChunkStream( Asura_Guid uGuid, Asura_Chunk_Stream &xStream )
{
	Asura_ServerEntity_FMVController* pxNew = new Asura_ServerEntity_FMVController(uGuid);
	if (pxNew)
	{
		pxNew->ReadFromChunkStream(xStream);
	}
}

////////////////////////////////////////////////////////////////////////////////
// WriteToChunkStream
////////////////////////////////////////////////////////////////////////////////

void Asura_ServerEntity_FMVController::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	// Version
	xStream << static_cast<int>( CURRENT_CHUNK_VERSION );

	PARENT::WriteToChunkStream(xStream);

	//
	// Write our data...
	//

	// We're using bit flags (ASURA_FMV_FLAGS) to access m_uFlags,
	//		so don't use Asura_WriteToStream_Bitfield_Wrapper (which endian swaps).
	xStream << m_uFlags;
	xStream << m_xFMVName;

	xStream << m_xLinkedSpeakerGuids.GetNumberOfNodes();

	Asura_TemplateTree_ConstIterator<Asura_Guid, Asura_Guid> xIterator(m_xLinkedSpeakerGuids);

	while(!xIterator.Done())
	{
		const Asura_Guid* puSpeakerGuid = xIterator.GetCurrent();

		if (puSpeakerGuid)
		{
			// Check the speaker actually exists
			Asura_ServerEntity_FMVSpeaker* pxSpeaker = Asura_ServerEntity_FMVSpeaker::FindEntity(*puSpeakerGuid);
			if(pxSpeaker)
			{
				xStream << *puSpeakerGuid;
			}
			else
			{
				Asura_Assertf(false, ("FMV Controller linked to non-existent Speaker Guid [%d]", *puSpeakerGuid));
			}
		}

		xIterator.Next();
	}
}

////////////////////////////////////////////////////////////////////////////////
// ReadFromChunkStream
// Must cope with all versions up to CURRENT_CHUNK_VERSION (be backwards compatible)
////////////////////////////////////////////////////////////////////////////////

void Asura_ServerEntity_FMVController::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	// Version.
	int iVersion;
	xStream >> iVersion;

	// Parent data.
	PARENT::ReadFromChunkStream( xStream );    

	//
	// Read our data...
	//

	Asura_Assert((iVersion >= 0) && (iVersion <= CURRENT_CHUNK_VERSION), "Asura_ServerEntity_FMVController: Invalid chunk version");

	if ( iVersion >= 0 )
	{
		// This purposely doesn't use Asura_ReadFromStream_Bitfield_Wrapper for m_uFlags
		xStream >> m_uFlags;
		xStream >> m_xFMVName;

		if ( iVersion >= 1 )
		{
			//
			// Read in linked speaker Guid list
			//
			unsigned int uTotalGuids;
			xStream >> uTotalGuids;

			for(unsigned int uGuidCount = 0; uGuidCount < uTotalGuids; uGuidCount++)
			{
				unsigned int uThisGuid;
				xStream >> uThisGuid;
				AddLinkedSpeaker(uThisGuid);
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_ServerEntity_FMVController::HandleMessage(const Asura_Message* pxMessage)
{
	if( IsDestroyed() ) { return true; } // ignore all further messages?

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			case ASURA_MSG_FMV_CONTROLLER_PLAY:
			{
				if (m_xFMVName.IsBlank())
				{
					Asura_Assert (false, "No FMV selected");
					return true;
				}
				else
				{
					Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, GetGuid());
					return true;
				}
			}

			case ASURA_MSG_FMV_CONTROLLER_PAUSE:
			{
				Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, GetGuid());
				return true;
			}

			default:
			{
				Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, GetGuid());
				return PARENT::HandleMessage(pxMessage);
			}
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ServerEntity_FMVController::AddLinkedSpeaker(Asura_Guid uGuid)
{
	Asura_Guid* pxGuid = m_xLinkedSpeakerGuids.AddNode(uGuid);
	Asura_Assertf(pxGuid, ("Unable to add FMV Speaker Guid %d", uGuid));
	if (pxGuid)
	{
		*pxGuid = uGuid;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ServerEntity_FMVController::RemoveLinkedSpeaker(Asura_Guid uGuid)
{
	m_xLinkedSpeakerGuids.DeleteNode(uGuid);
}

#endif // ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM
