///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_ClientEntity_ConsoleVar.cpp
//
// Created by Martin G.Harris (20/6/2007)
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Entity allowing designers to add a run-time console variable in tools and modify
// the value using static messages. This allows easy hooking up of game logic to
// (for example) the HUD system.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Need command console for the console variables.
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#include "Entity/Intangible/Asura_ClientEntity_ConsoleVar.h"
#include "Entity/Intangible/Asura_ServerEntity_ConsoleVar.h"


#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Network/Asura_Message.h"


// Constants
const u_int uBUFFER_LENGTH = 256;

// Static data
Asura_Hashed_List<Asura_ClientEntity_ConsoleVar> Asura_ClientEntity_ConsoleVar::s_xList(4);

/////////////////
// ConvertType //
/////////////////
// Helper functions to convert between the types. This is just in case someone renumbers the
// enum in the Asura_CommandConsole_VarRepository; we can't renumber our enum as it is saved out in chunks.
// Also we haven't implemented string types (yet).
Asura_ClientEntity_ConsoleVar::TYPE Asura_ClientEntity_ConsoleVar::ConvertType(Asura_CommandConsole_VarRepository::Var::TYPE eVarType)
{
	switch (eVarType)
	{
	case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL:
		return Asura_ClientEntity_ConsoleVar::TYPE_BOOL;

	case Asura_CommandConsole_VarRepository::Var::TYPE_INT:
		return Asura_ClientEntity_ConsoleVar::TYPE_INT;

	case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT:
		return Asura_ClientEntity_ConsoleVar::TYPE_FLOAT;

	default:
		break;
	}

	return Asura_ClientEntity_ConsoleVar::TYPE_INVALID;
}

Asura_CommandConsole_VarRepository::Var::TYPE Asura_ClientEntity_ConsoleVar::ConvertType(Asura_ClientEntity_ConsoleVar::TYPE eType)
{
	switch (eType)
	{
	case Asura_ClientEntity_ConsoleVar::TYPE_BOOL:
		return Asura_CommandConsole_VarRepository::Var::TYPE_BOOL;

	case Asura_ClientEntity_ConsoleVar::TYPE_INT:
		return Asura_CommandConsole_VarRepository::Var::TYPE_INT;

	case Asura_ClientEntity_ConsoleVar::TYPE_FLOAT:
		return Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT;

	default:
		break;
	}

	return Asura_CommandConsole_VarRepository::Var::TYPE_INVALID;
}


///////////////////////////////////
// Asura_ClientEntity_ConsoleVar //
///////////////////////////////////
Asura_ClientEntity_ConsoleVar::Asura_ClientEntity_ConsoleVar(Asura_Guid uGuid)
: Asura_Entity_Client(uGuid)
, m_szVarName(0)
, m_uVarNameHash(ASURA_HASH_ID_UNSET)
, m_bOwnsConsoleVar(false)
, m_eType(TYPE_INVALID)
, m_iData(0)
, m_uFlags(0)
{
	Classification = AsuraEntityClass_ConsoleVar;

	s_xList.AddToList(this);
}

Asura_ClientEntity_ConsoleVar::~Asura_ClientEntity_ConsoleVar()
{
	RemoveConsoleVar();

	s_xList.RemoveFromList(this);
}

///////////////////
// ChangeVarName //
///////////////////
// Initialise or change variable name. Adding/Removing console variables as necessary.
void Asura_ClientEntity_ConsoleVar::ChangeVarName(const Asura_Char* szVarName, const TYPE eType)
{
	// Remove the old console var if necessary.
	RemoveConsoleVar();

	if (szVarName == 0 || *szVarName == 0)
	{
		// Passing in a null pointer indicates that we just wanted to remove the old variable.
		return;
	}

	// Copy name and type.
	m_szVarName = Asura_StringUtil::Duplicate(szVarName);
	m_eType = eType;
	m_uVarNameHash = Asura_GetHashID(szVarName);

	// Add new console var.
	AddConsoleVar();
}

//////////////////////
// RemoveConsoleVar //
//////////////////////
// Removes console variable if necessary.
void Asura_ClientEntity_ConsoleVar::RemoveConsoleVar()
{
	if (m_szVarName)
	{
		// Save hash.
		const Asura_Hash_ID uHash = m_uVarNameHash;
		// Reset hash so we don't transfer to ourself.
		m_uVarNameHash = ASURA_HASH_ID_UNSET;

		if (m_bOwnsConsoleVar)
		{
			m_bOwnsConsoleVar = false;

			// If this console variable is being shared then transfer it.
			const bool bTransfered = TransferConsoleVar(uHash, m_iData);
			if (!bTransfered)
			{
				// Didn't transfer so we can delete the variable.
				Asura_UnicodeChar wszBuffer[uBUFFER_LENGTH];
				Asura_PrintfToBfr(wszBuffer, L"%S", uBUFFER_LENGTH, true, m_szVarName);
				//TY 31/10/07: We have to call remove on the var repository directly because the RemoveVar function of the Command console might be #defined out
				Asura_CommandConsole_VarRepository::Remove( wszBuffer );
			}
		}

		delete [] m_szVarName;
		m_szVarName = 0;

		m_eType = TYPE_INVALID;
	}
}

//////////////////////////////
// TransferConsoleVarToCode //
//////////////////////////////
// Try to transfer the ownership of the console variable from an entity to the code.
// This is called if a code-defined variable is set up after a level has finished loading,
// and that level contains a ConsoleVar entity refering to the variable.
bool Asura_ClientEntity_ConsoleVar::TransferConsoleVarToCode(const Asura_UnicodeChar* wszVarName, 
															Asura_CommandConsole_VarRepository::Var::TYPE eType, 
															const Asura_CommandConsole_VarRepository::Var::Data& xData, 
															Asura_CommandConsole::REMOVE eRemove)
{
	if (!wszVarName || !*wszVarName)
	{
		return false;
	}

	if (eType == Asura_CommandConsole_VarRepository::Var::TYPE_INVALID)
	{
		return false;
	}

	Asura_Char szBuffer[uBUFFER_LENGTH];
	Asura_PrintfToBfr(szBuffer, "%S", uBUFFER_LENGTH, true, wszVarName);

	const Asura_Hash_ID uVarNameHash = Asura_GetHashID(szBuffer);
	if (uVarNameHash == ASURA_HASH_ID_UNSET)
	{
		return false;
	}

	// Loop over the ConsoleVars looking for an owner.
	for (Asura_Hashed_List_It<Asura_ClientEntity_ConsoleVar> xIt(GetList());
		!xIt.Done();
		xIt.Next())
	{
		Asura_ClientEntity_ConsoleVar* pxCurrent = xIt.GetCurrent();
		if (!pxCurrent) continue;
		
		if (pxCurrent->m_uVarNameHash == uVarNameHash &&
			pxCurrent->m_bOwnsConsoleVar &&
			pxCurrent->m_eType != TYPE_INVALID)
		{
			// Found owner!
			if (ConvertType(pxCurrent->m_eType) != eType)
			{
				// Types differ! Help!
				Asura_Assertf(0, ("Type differs between code and tools, for console variable (%s).", szBuffer));
				// Code still takes priority even though the types are different...
			}
			
			Asura_CommandConsole_VarRepository::Var* pxVar = pxCurrent->GetVar();
			if (pxVar)
			{
				// Copy the data
				pxVar->m_eRemove = eRemove;
				pxVar->m_eType = eType;
				pxVar->m_xData = xData;

				// Remove ownership
				pxCurrent->m_bOwnsConsoleVar = false;
				return true;
			}

			// This entity owns the variable, but it doesn't exist; something has gone wrong.
			Asura_Assert(0, "Couldn't find ConsoleVar entity's console variable.");
			return false;
		}
	}

	// Not found.
	return false;
}


////////////////////////
// TransferConsoleVar //
////////////////////////
// If this console variable is being shared by several ConsoleVar entities then transfer it.
bool Asura_ClientEntity_ConsoleVar::TransferConsoleVar(Asura_Hash_ID uHash, int iData)
{
	if (uHash != ASURA_HASH_ID_UNSET)
	{
		for (Asura_Hashed_List_It<Asura_ClientEntity_ConsoleVar> xIt(GetList());
			!xIt.Done();
			xIt.Next())
		{
			Asura_ClientEntity_ConsoleVar* pxCurrent = xIt.GetCurrent();
			if (!pxCurrent) continue;
			
			if (!pxCurrent->IsDestroyed() &&
				pxCurrent->m_uVarNameHash == uHash &&
				!pxCurrent->m_bOwnsConsoleVar &&
				pxCurrent->m_szVarName &&
				pxCurrent->m_eType != TYPE_INVALID)
			{
				// Transfer ownership and data.
				pxCurrent->m_bOwnsConsoleVar = true;
				pxCurrent->m_iData = iData; // Assumes union of all data types is 4 bytes.

				// Adjust pointer.
				Asura_CommandConsole_VarRepository::Var* pxVar = pxCurrent->GetVar();
				if (pxVar)
				{
					Asura_Assert(ConvertType(pxVar->m_eType) == pxCurrent->m_eType, "Transfered Console Var, but type is different!");
					// NOTE: Switch is not really necessary as all the addresses are the same, due to union.
					switch (pxCurrent->m_eType)
					{
					case TYPE_BOOL:
						pxVar->m_xData.pbData = &pxCurrent->m_bData;
						break;

					case TYPE_INT:
						pxVar->m_xData.piData = &pxCurrent->m_iData;
						break;

					case TYPE_FLOAT:
					default:
						pxVar->m_xData.pfData = &pxCurrent->m_fData;
						break;
					}
				}

				return true;
			}
		}
	}
	
	return false;
}

///////////////////////////
// GetNumberOfReferences //
///////////////////////////
// Count the number of ConsoleVar entities refering to a specified variable.
// Return true if one of these entities owns the console variable.
bool Asura_ClientEntity_ConsoleVar::GetNumberOfReferences(Asura_Hash_ID uVarNameHash, int& iNumber)
{
	bool bOwner = false;
	iNumber = 0;

	if (uVarNameHash != ASURA_HASH_ID_UNSET)
	{
		for (Asura_Hashed_List_It<Asura_ClientEntity_ConsoleVar> xIt(GetList());
			!xIt.Done();
			xIt.Next())
		{
			Asura_ClientEntity_ConsoleVar* pxCurrent = xIt.GetCurrent();
			if (!pxCurrent) continue;
			
			if (pxCurrent->m_uVarNameHash == uVarNameHash)
			{
				++iNumber;
				if (pxCurrent->m_bOwnsConsoleVar)
				{
					bOwner = true;
				}
			}
		}
	}

	return bOwner;
}

////////////////
// ChangeType //
////////////////
// Update the type of all entities pointing to the specified variable.
// Only really used by tools code.
#ifdef ASURA_TOOLS
void Asura_ClientEntity_ConsoleVar::ChangeType(Asura_Hash_ID uVarNameHash, TYPE eNewType)
{
	if (uVarNameHash != ASURA_HASH_ID_UNSET)
	{
		for (Asura_Hashed_List_It<Asura_ClientEntity_ConsoleVar> xIt(GetList());
			!xIt.Done();
			xIt.Next())
		{
			Asura_ClientEntity_ConsoleVar* pxCurrent = xIt.GetCurrent();
			if (!pxCurrent) continue;
			
			if (pxCurrent->m_uVarNameHash == uVarNameHash)
			{
				// Fix up the type.
				pxCurrent->m_eType = eNewType;

				Asura_ServerEntity_ConsoleVar* pxServer = Asura_ServerEntity_ConsoleVar::FindEntity(pxCurrent->GetGuid());
				Asura_Assert(pxServer, "Null server entity");
				if (pxServer)
				{
					pxServer->m_eType = eNewType;
				}

				// Fix up the console variable.
				if (pxCurrent->m_bOwnsConsoleVar)
				{
					Asura_CommandConsole_VarRepository::Var* pxVar = pxCurrent->GetVar();
					if (pxVar)
					{
						// Update type, limits etc.
						pxVar->m_eType = ConvertType(eNewType);
						pxVar->m_xData = pxCurrent->GetDefaultData(eNewType);
					}
				}
			}
		}
	}
}
#endif // ASURA_TOOLS

///////////////////
// AddConsoleVar //
///////////////////
// Adds console variable if necessary,
void Asura_ClientEntity_ConsoleVar::AddConsoleVar()
{
	if (m_eType == TYPE_INVALID)
	{
		return;
	}

	if (m_szVarName)
	{
		Asura_CommandConsole_VarRepository::Var* pxVar = GetVar();
		if (pxVar)
		{
			// The variable already exists, so we don't need to add it (eg it may have been added in code,
			// for example if it needs to persist between levels).

			// Check that the type is correct.
			if (ConvertType(pxVar->m_eType) != m_eType)
			{
				// Type is wrong!
				 // Try to fix up the type locally, will work OK if the types are similar.
				m_eType = ConvertType(pxVar->m_eType);
				Asura_Assertf(0, ("Console variable (%s) already exists and has different type!", m_szVarName));
			}

			CopyValueToThis();
		}
		else
		{
			// Variable doesn't exist, so add it.
			const Asura_CommandConsole_VarRepository::Var::Data xData = GetDefaultData(m_eType);
			
			Asura_UnicodeChar wszBuffer[uBUFFER_LENGTH];
			Asura_PrintfToBfr(wszBuffer, L"%S", uBUFFER_LENGTH, true, m_szVarName);
			
			const bool bAdded = Asura_CommandConsole_VarRepository::Add(wszBuffer, ConvertType(m_eType), xData, Asura_CommandConsole::REMOVE_AT_SHUTDOWN, true);
			
			Asura_Assert(bAdded, "Failed to add console variable");
			if (!bAdded)
			{
				m_eType = TYPE_INVALID; // Mark it as invalid.
			}

			// Since we added it we can remove it. In an ideal world we would do reference counting on these objects.
			m_bOwnsConsoleVar = bAdded;
		}
	}
}

////////////////////
// GetDefaultData //
////////////////////
// Generate default data for specified type.
Asura_CommandConsole_VarRepository::Var::Data Asura_ClientEntity_ConsoleVar::GetDefaultData(TYPE eNewType)
{
	Asura_CommandConsole_VarRepository::Var::Data xData;
	xData.uRestrictions = 0;

	if (eNewType == TYPE_BOOL)
	{
		xData.pbData = &m_bData;
		xData.iLowerLimit = 0;
		xData.iUpperLimit = 1;
		m_bData = false;
	}
	else if (eNewType == TYPE_INT)
	{
		xData.piData = &m_iData;
		xData.iLowerLimit = INT_MIN;
		xData.iUpperLimit = INT_MAX;
		m_iData = 0;
	}
	else if (eNewType == TYPE_FLOAT)
	{
		Asura_Assert(eNewType == TYPE_FLOAT, "Error unknown console var type!");

		xData.pfData = &m_fData;
		xData.fLowerLimit = -FLT_MAX;
		xData.fUpperLimit = FLT_MAX;
		m_fData = 0.0f;
	}

	return xData;
}


/////////////////////
// CopyValueToThis //
/////////////////////
// If we don't own the console variable, then we want to copy it's value to this entity.
void Asura_ClientEntity_ConsoleVar::CopyValueToThis()
{
	if (m_bOwnsConsoleVar)
	{
		// We own the variable, so our data is the actual value of the variable.
		return;
	}

	Asura_CommandConsole_VarRepository::Var* pxVar = GetVar();
	if (pxVar)
	{
		// Copy data to this entity.
		const Asura_CommandConsole_VarRepository::Var::Data& xData = pxVar->m_xData;
		switch (m_eType)
		{
		case TYPE_BOOL:
			m_bData = (xData.pbData ? *xData.pbData : false);
			break;

		case TYPE_INT:
			m_iData = (xData.piData ? *xData.piData : 0);
			break;

		case TYPE_FLOAT:
			m_fData = (xData.pfData ? *xData.pfData : 0.0f);
			break;

		default:
			break;
		}
	}
}

///////////////////
// HandleMessage //
///////////////////
bool Asura_ClientEntity_ConsoleVar::HandleMessage( const Asura_Message* pxMessage )
{
	if (pxMessage && pxMessage->GetMsgID() == ASURA_MSG_CHANGE_CONSOLEVAR)
	{
		const Asura_Message_Change_ConsoleVar* pxData = static_cast<const Asura_Message_Change_ConsoleVar*>(pxMessage->GetData());
		if (pxData)
		{
			HandleMessage_Change(pxData);
		}
		return true;
	}

	return PARENT::HandleMessage(pxMessage);
}

//////////////////////////
// HandleMessage_Change //
//////////////////////////
void Asura_ClientEntity_ConsoleVar::HandleMessage_Change( const Asura_Message_Change_ConsoleVar* pxData )
{
	Asura_Assert(m_eType != TYPE_INVALID, "Sent message to an invalid console variable! Perhaps wrong type or a string?");
	if (m_eType == TYPE_INVALID)
	{
		return;
	}

	// Variable exists?
	Asura_CommandConsole_VarRepository::Var* pxVar = GetVar();
	if (!pxVar)
	{
		Asura_Assert(0, "HandleMessage_Change: Failed to find console variable.");
		return;
	}
	
	// Read only?
	if (pxVar->m_bReadOnly)
	{
		Asura_Assert(0, "HandleMessage_Change: console variable is read only!");
		return;
	}

	// Restricted?
	const u_int uRestrictions = pxVar->m_xData.uRestrictions;
	if( uRestrictions )
    {
		if (!Asura_CommandConsole::Project_CheckRestrictions(uRestrictions, false))
		{
			Asura_Assert(0, "HandleMessage_Change: console variable is restricted!");
			return;
		}
	}

	// Check that the type is what we are expecting.
	Asura_Assert(ConvertType(pxVar->m_eType) == m_eType, "Console Variable has unexpected type!");

	const int iOperation = (pxData->m_iFlags & Asura_Message_Change_ConsoleVar::FLAG_OPERATION_MASK);
	const float fData = pxData->m_fData;

	switch (iOperation)
	{
	case Asura_Message_Change_ConsoleVar::FLAG_OPERATION_SET:
		DoSetOperation(pxVar, fData);
		break;

	case Asura_Message_Change_ConsoleVar::FLAG_OPERATION_ADD:
		DoAddOperation(pxVar, fData);
		break;

	case Asura_Message_Change_ConsoleVar::FLAG_OPERATION_MULTIPLY:
		DoMultiplyOperation(pxVar, fData);
		break;

	case Asura_Message_Change_ConsoleVar::FLAG_OPERATION_TOGGLE_BOOLEAN:
		DoToggleBoolean(pxVar);
		break;

	default:
		Asura_Assert(0, "Asura_ClientEntity_ConsoleVar: Unknown operation!");
		break;
	}
}


////////////////
// SetBoolean //
////////////////
// Set the variable to be the new value, applying clamping as necessary.
void Asura_ClientEntity_ConsoleVar::SetBoolean(Asura_CommandConsole_VarRepository::Var* pxVar, bool bNewValue)
{
	if (pxVar &&
		pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_BOOL &&
		pxVar->m_xData.pbData)
	{
		const bool bOld = *pxVar->m_xData.pbData;
		if (bOld != bNewValue)
		{
			*pxVar->m_xData.pbData = bNewValue;
			if (pxVar->m_pbModified)
			{
				*pxVar->m_pbModified = true;
			}
		}
	}
	else
	{
		Asura_Assert(0, "Failed to set boolean console variable!");
	}
}

////////////////
// SetInteger //
////////////////
// Set the variable to be the new value, applying clamping as necessary.
void Asura_ClientEntity_ConsoleVar::SetInteger(Asura_CommandConsole_VarRepository::Var* pxVar, int iNewValue)
{
	if (pxVar &&
		pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_INT &&
		pxVar->m_xData.piData)
	{
		const int iOld = *pxVar->m_xData.piData;
		const int iNew = Asura_Maths::ClampToRange(iNewValue, pxVar->m_xData.iLowerLimit, pxVar->m_xData.iUpperLimit);
		if (iOld != iNew)
		{
			*pxVar->m_xData.piData = iNew;
			if (pxVar->m_pbModified)
			{
				*pxVar->m_pbModified = true;
			}
		}
	}
	else
	{
		Asura_Assert(0, "Failed to set integer console variable!");
	}
}


//////////////
// SetFloat //
//////////////
// Set the variable to be the new value, applying clamping as necessary.
void Asura_ClientEntity_ConsoleVar::SetFloat(Asura_CommandConsole_VarRepository::Var* pxVar, float fNewValue)
{
	if (pxVar &&
		pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT &&
		pxVar->m_xData.pfData)
	{
		const float fOld = *pxVar->m_xData.pfData;
		const float fNew = Asura_Maths::ClampToRange(fNewValue, pxVar->m_xData.fLowerLimit, pxVar->m_xData.fUpperLimit);
		if (fOld != fNew)
		{
			*pxVar->m_xData.pfData = fNew;
			if (pxVar->m_pbModified)
			{
				*pxVar->m_pbModified = true;
			}
		}
	}
	else
	{
		Asura_Assert(0, "Failed to set float console variable!");
	}
}

////////////////////
// DoSetOperation //
////////////////////
void Asura_ClientEntity_ConsoleVar::DoSetOperation(Asura_CommandConsole_VarRepository::Var* pxVar, float fData)
{
	if (!pxVar) return;

	// Use actual type rather than the expected type.
	switch (ConvertType(pxVar->m_eType))
	{
	default:
	case TYPE_INVALID:
		Asura_Assert(0, "DoSetOperation: Unknown or invalid console variable type!");
		break;

	case TYPE_BOOL:
		SetBoolean(pxVar, ((fData == 0.0f) ? false : true));
		break;

	case TYPE_INT:
		SetInteger(pxVar, Asura_Maths::FloatToInt_Near(fData));
		break;

	case TYPE_FLOAT:
		SetFloat(pxVar, fData);
		break;
	}
}

////////////////////
// DoAddOperation //
////////////////////
void Asura_ClientEntity_ConsoleVar::DoAddOperation(Asura_CommandConsole_VarRepository::Var* pxVar, float fData)
{
	if (!pxVar) return;

	// Use actual type rather than the expected type.
	switch (ConvertType(pxVar->m_eType))
	{
	default:
	case TYPE_INVALID:
		Asura_Assert(0, "DoAddOperation: Unknown or invalid console variable type!");
		break;

	case TYPE_BOOL:
		Asura_Assert(0, "DoAddOperation: Add is invalid operation for boolean type.");
		break;

	case TYPE_INT:
		{
			const int iOldValue = (pxVar->m_xData.piData) ? *(pxVar->m_xData.piData) : 0;
			SetInteger(pxVar, iOldValue + Asura_Maths::FloatToInt_Near(fData));
		}
		break;

	case TYPE_FLOAT:
		{
			const float fOldValue = (pxVar->m_xData.pfData) ? *(pxVar->m_xData.pfData) : 0.0f;
			SetFloat(pxVar, fOldValue + fData);
		}
		break;
	}
}

/////////////////////////
// DoMultiplyOperation //
/////////////////////////
void Asura_ClientEntity_ConsoleVar::DoMultiplyOperation(Asura_CommandConsole_VarRepository::Var* pxVar, float fData)
{
	if (!pxVar) return;

	// Use actual type rather than the expected type.
	switch (ConvertType(pxVar->m_eType))
	{
	default:
	case TYPE_INVALID:
		Asura_Assert(0, "DoMultiplyOperation: Unknown or invalid console variable type!");
		break;

	case TYPE_BOOL:
		Asura_Assert(0, "DoMultiplyOperation: Multiply is invalid operation for boolean type.");
		break;

	case TYPE_INT:
		{
			// NOTE: we are allowing a float multiply here.
			const int iOldValue = (pxVar->m_xData.piData) ? *(pxVar->m_xData.piData) : 0;
			SetInteger(pxVar, Asura_Maths::FloatToInt_Near(iOldValue * fData));
		}
		break;

	case TYPE_FLOAT:
		{
			const float fOldValue = (pxVar->m_xData.pfData) ? *(pxVar->m_xData.pfData) : 0.0f;
			SetFloat(pxVar, fOldValue * fData);
		}
		break;
	}
}

/////////////////////
// DoToggleBoolean //
/////////////////////
void Asura_ClientEntity_ConsoleVar::DoToggleBoolean(Asura_CommandConsole_VarRepository::Var* pxVar)
{
	if (!pxVar) return;

	// Use actual type rather than the expected type.
	switch (ConvertType(pxVar->m_eType))
	{
	default:
	case TYPE_INVALID:
		Asura_Assert(0, "DoToggleBoolean: Unknown or invalid console variable type!");
		break;

	case TYPE_BOOL:
		{
			const bool bOldValue = (pxVar->m_xData.pbData) ? *(pxVar->m_xData.pbData) : false;
			SetBoolean(pxVar, !bOldValue);
		}
		break;

	case TYPE_INT:
		{
			//Asura_Assert(0, "DoToggleBoolean: invalid operation for integer type.");
			const int iOldValue = (pxVar->m_xData.piData) ? *(pxVar->m_xData.piData) : 0;
			SetInteger( pxVar, ((iOldValue==0) ? 1 : 0) );
		}
		break;

	case TYPE_FLOAT:
		{
			//Asura_Assert(0, "DoToggleBoolean: invalid operation for float type.");
			const float fOldValue = (pxVar->m_xData.pfData) ? *(pxVar->m_xData.pfData) : 0.0f;
			SetFloat( pxVar, ((fOldValue==0.0f) ? 1.0f : 0.0f) );
		}
		break;
	}
}

//////////////////////////////////////////////////////////////////////////
//	Static create functions
//////////////////////////////////////////////////////////////////////////
void Asura_ClientEntity_ConsoleVar::Create( Asura_Guid uGuid, Asura_Chunk_Stream& xStream )
{
	Asura_ClientEntity_ConsoleVar* pxEntity = new Asura_ClientEntity_ConsoleVar(uGuid);
	if (pxEntity)
	{
		pxEntity->ReadFromChunkStream( xStream );
	}
}

/////////////////////////
// ReadFromChunkStream //
/////////////////////////
// NOTE: If you change this you *must* update the corresponding function in Asura_ServerEntity_ConsoleVar.
void Asura_ClientEntity_ConsoleVar::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	// Get version number
	int iVersion = -1;
	xStream >> iVersion;
	
	// Skip server data.
	Asura_Entity_Server::SkipDataInChunkStream(xStream);
 
    // Our data.
	Asura_Assert(iVersion>=0 && iVersion<=Asura_ServerEntity_ConsoleVar::CURRENT_CHUNK_VERSION, "Asura_ClientEntity_ConsoleVar: Invalid chunk version");
    
	if ( iVersion >= 0 )
    {
		// Type
		int iType = -1;
		xStream >> iType;
		m_eType = static_cast<TYPE>(iType);

		// Console Variable name
		u_int uLength = 0; // Length excluding the null terminator (number of charcters written).
		xStream >> uLength;

		// Any string longer than this is probably an error in reading the chunks.
		Asura_Assert(uLength<256, "Asura_ClientEntity_ConsoleVar: Error: String too long!");
		Asura_Assert(!m_szVarName, "Asura_ClientEntity_ConsoleVar: Memory leak!");

		if (uLength>0)
		{
			m_szVarName = new Asura_Char[uLength+1];
			for (u_int u=0; u<uLength; ++u)
			{
				xStream >> m_szVarName[u];
			}

			// May as well align.
			xStream.Skip4BytePadding();

			// Null terminator.
			m_szVarName[uLength] = 0;

			m_uVarNameHash = Asura_GetHashID(m_szVarName);
		}
	}

	if (iVersion >= 1)
	{
		Asura_Guid uEntityToTrackGuid;
		xStream >> uEntityToTrackGuid; // skip
		xStream >> m_uFlags;
	}

	// Create the console variable.
	AddConsoleVar();
}


#endif //ASURA_SUPPORT_COMMAND_CONSOLE
