//////////////////////////////////////////////////////////////////////////////
//
// Asura_ClientEntity_DestructibleLight.cpp
//
// Created by Mike Healey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifndef ASURA_TARGET_SM3

#include "Entity/PhysicalObject/Asura_ClientEntity_DestructibleLight.h"

#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_DestructibleLight.h"
#include "Light/Asura_Photon.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "SpecialFX/Asura_Corona_System.h"


Asura_Hashed_List<Asura_ClientEntity_DestructibleLight> Asura_ClientEntity_DestructibleLight::s_xDestructibleLightList;

//////////////////////////////////////////////////////////////////////////////////////////
// Asura_ClientEntity_DestructibleLight                                                 //
//////////////////////////////////////////////////////////////////////////////////////////
Asura_ClientEntity_DestructibleLight::Asura_ClientEntity_DestructibleLight()
: m_pxPhoton(0) , m_iCoronaHandle(ASURA_INVALID_HANDLE)
{
	Classification = AsuraEntityClass_DestructibleLight;
	s_xDestructibleLightList.AddToList(this);
	AddToUpdateableClientList(this);	
}

Asura_ClientEntity_DestructibleLight::~Asura_ClientEntity_DestructibleLight()
{
    if (m_pxPhoton)
    {
        Asura_Photon_System::DestroyPhoton(m_pxPhoton);
		m_pxPhoton = 0;
    }
	if(m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Corona_System::DestroyCorona(m_iCoronaHandle);
		m_iCoronaHandle = ASURA_INVALID_HANDLE;
	}
	s_xDestructibleLightList.RemoveFromList(this);
	RemoveFromUpdateableClientList(this, false);	// could have been removed from updateable list at any time.
}


///////////////////
// HandleMessage //
///////////////////
bool Asura_ClientEntity_DestructibleLight::HandleMessage(const Asura_Message* pxMessage)
{
	// Process a message from the server
	if (pxMessage && pxMessage->GetMsgID() == ASURA_MSG_CLIENTUPDATE)
	{
		const Asura_Message_DestructibleLight_ClientUpdate* pxData = 
			static_cast<const Asura_Message_DestructibleLight_ClientUpdate*>(pxMessage->GetData());
		if (pxData)
		{
            ProcessClientUpdate(pxData, pxMessage->GetSentTime());

            if(!IsDead())
			{
				if (!m_pxPhoton)
				{
					m_pxPhoton = Asura_Photon_System::CreatePhoton();
					if (m_pxPhoton)
					{
						m_pxPhoton->SetPosition( pxData->m_xPosition );
						m_pxPhoton->SetRange( pxData->m_fRange );
						float fScale = (pxData->m_fBrightness / 255.0f) / 2.0f;
						m_pxPhoton->SetColours( pxData->m_fR * fScale, pxData->m_fG * fScale, pxData->m_fB * fScale);
						m_pxPhoton->SetUseBoundingBox( pxData->m_bUseBoundingBox);
						m_pxPhoton->SetBoundingBox( pxData->m_xBoundingBox );
					}
				}
				else
				{
					// update position and colour
					m_pxPhoton->SetPosition(pxData->m_xPosition);
					m_pxPhoton->SetRange(pxData->m_fRange);
					float fScale = (pxData->m_fBrightness / 255.0f) / 2.0f;
					m_pxPhoton->SetColours( pxData->m_fR * fScale, pxData->m_fG * fScale, pxData->m_fB * fScale);

					// check everything is as expected
	//              Asura_Assert(m_pxPhoton->GetPosition() == pxData->m_xPosition, "");
	//              Asura_Assert(m_pxPhoton->GetRange() == pxData->m_fRange, "");
	//              Asura_Assert(m_pxPhoton->GetRedComponent() == pxData->m_fR, "");
	//              Asura_Assert(m_pxPhoton->GetGreenComponent() == pxData->m_fG, "");
	//              Asura_Assert(m_pxPhoton->GetBlueComponent() == pxData->m_fB, "");
				}
				if(m_pxPhoton)
				{
					m_pxPhoton->SetDontLightEnv(pxData->m_bDontLightEnv);
				}

				UpdateCorona(pxData);

			}
		}

		return true;
	}
	
	return false; // Must not pass to the parent
}


Asura_ClientEntity_DestructibleLight* Asura_ClientEntity_DestructibleLight::FindEntity(Asura_Guid uGuid)
{
	return s_xDestructibleLightList.FindEntity(uGuid);
}

void Asura_ClientEntity_DestructibleLight::Update()
{
	//Make sure the light is off , if this is destroyed. RAR 09/04/03
	if(IsDead())
	{
		if (m_pxPhoton)
		{
			Asura_Photon_System::DestroyPhoton(m_pxPhoton);
			m_pxPhoton = 0;
		}
		if(m_iCoronaHandle != ASURA_INVALID_HANDLE)
		{
			Asura_Corona_System::DestroyCorona(m_iCoronaHandle);
			m_iCoronaHandle = ASURA_INVALID_HANDLE;
		}
		
	}
}

void Asura_ClientEntity_DestructibleLight::UpdateCorona(const Asura_Message_DestructibleLight_ClientUpdate* pxData)
{
	if(pxData->m_bHasCorona)
	{
		if(m_iCoronaHandle == ASURA_INVALID_HANDLE)
		{
			m_iCoronaHandle = Asura_Corona_System::CreateCorona();
		}
		int iTextureIndex = Asura_Texture_Management::GetIndexFromHashID(pxData->m_uCoronaTexture);
		//update the corona values
		Asura_Corona_System::SetPosition(m_iCoronaHandle,pxData->m_xPosition);
		Asura_Corona_System::SetRadius(m_iCoronaHandle,pxData->m_fCoronaRadius);
		Asura_Corona_System::SetTextureIndex(m_iCoronaHandle,iTextureIndex);
		Asura_Corona_System::SetFlag_DontFadeWithDistance(m_iCoronaHandle,pxData->m_bCorona_DontFadeWithDistance);
		
		float fScale = (pxData->m_fBrightness / 255.0f);
		float fRed = pxData->m_fR*fScale;
		if(fRed>1.0f) fRed = 1.0f;
		float fGreen = pxData->m_fG*fScale;
		if(fGreen>1.0f) fGreen = 1.0f;
		float fBlue = pxData->m_fB*fScale;
		if(fBlue>1.0f) fBlue = 1.0f;

		Asura_Corona_System::SetColour(m_iCoronaHandle,fRed,fGreen,fBlue);


	}
	else if(m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Corona_System::DestroyCorona(m_iCoronaHandle);
		m_iCoronaHandle = ASURA_INVALID_HANDLE;
	}
}

void Asura_ClientEntity_DestructibleLight::SetPhotonAndCoronaColour(float fRed,float fGreen,float fBlue)
{
	if(m_pxPhoton)
	{
		m_pxPhoton->SetColours(fRed,fGreen,fBlue);
	}
	if(m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		float fCoronaRed = fRed * 2.0f;
		if(fCoronaRed>1.0f) fCoronaRed = 1.0f;
		float fCoronaGreen = fGreen * 2.0f;
		if(fCoronaGreen>1.0f) fCoronaGreen = 1.0f;
		float fCoronaBlue = fBlue * 2.0f;
		if(fCoronaBlue>1.0f) fCoronaBlue = 1.0f;

		Asura_Corona_System::SetColour(m_iCoronaHandle,fCoronaRed,fCoronaGreen,fCoronaBlue);
	}
}

#endif // #ifndef ASURA_TARGET_SM3
