//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_Lift.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     Lift Entity - External control physics object moving along a spline
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/PhysicalObject/Asura_ServerEntity_Lift.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_Entity_Event.h"
#include "Entity/Locations/Spline/LiftSpline/Asura_ServerNode_LiftSpline.h"
#include "Entity/Locations/OrientedNode/LiftNode/Asura_ServerNode_LiftNode.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_Broadphase.h"
#include "Physics/Asura_Physics_RigidBody.h"

//Tools
#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_ServerEntity_Lift
//

Asura_Hashed_List<Asura_ServerEntity_Lift> Asura_ServerEntity_Lift::s_xList( 5 );

Asura_ServerEntity_Lift::Asura_ServerEntity_Lift()
: Asura_Entity_Server(ASURA_GUID_UNREGISTERED) // ignored
, m_uLiftFlags(0)
, m_fProgress(0.0f)
, m_uSplineGuid(ASURA_GUID_UNREGISTERED)
, m_uGotoNodeGuid(ASURA_GUID_UNREGISTERED)
, m_uStartSoundEventHash(ASURA_HASH_ID_UNSET)
, m_uStopSoundEventHash(ASURA_HASH_ID_UNSET)
{
	Classification = AsuraEntityClass_Lift;

	s_xList.AddToList( this );

	m_pxDynamicObject = Asura_Physics_Object::Create( ASURA_PHYSICS_OBJECT_EXTERNALCONTROL );
	m_pxDynamicObject->SetAssociatedEntityGuid(GetGuid());
}


Asura_ServerEntity_Lift::~Asura_ServerEntity_Lift( void )
{
	BlockLiftNodeLinks( false );
	s_xList.RemoveFromList( this );
}

void Asura_ServerEntity_Lift::SetRunning( const bool bRunning )
{
#ifdef ASURA_TOOLS
	if( Tools_System::IsActive() )
	{
		m_bSendStartSound = bRunning;
	}
	else
#endif //ASURA_TOOLS
	{
		const bool bIsRunning = m_bRunning; //m_bRunning is one bit of a u_int, can't compare to a bool directly
		if( bIsRunning != bRunning )
		{
			if( bRunning )
			{
				m_bSendStartSound = true;
			}
			else
			{
				m_bSendStopSound = true;
			}
			// Ensure the sounds get updated in the next network update.
			RequestClientUpdate();
		}
	}

	if ( bRunning )
	{
		// Remove any links we may have blocked.
		BlockLiftNodeLinks( false );
	}

	if (bRunning && !m_bRunning)
	{
		m_bJustStarted = true;//flag that we've just started, so we can make sure we clear current node
		SendMessagesForEvent(ASURA_ENTITY_EVENT_ONSTARTED);
	}
	else if (!bRunning && m_bRunning)
	{
		SendMessagesForEvent(ASURA_ENTITY_EVENT_ONSTOPPED);
	}

	m_bRunning = bRunning;
}

void Asura_ServerEntity_Lift::SetReversing(bool bReversing)
{
	if (static_cast<bool>(m_bReversing) != bReversing)
	{
		SendMessagesForEvent(ASURA_ENTITY_EVENT_ONREVERSED);
	}
	m_bReversing = bReversing;
}

bool Asura_ServerEntity_Lift::HandleMessage(const Asura_Message* pxMessage)
{
	if (!pxMessage) return false;

	switch(pxMessage->GetMsgID())
	{
		case ASURA_MSG_LIFT_START:
		{
			// NY - 15 November 2006 - just enable lift running, even when inactive - this means that an activate and start can be sent together,
			// and their order of arrival will be immaterial
			SetRunning(true);

			return true;
		}
		case ASURA_MSG_LIFT_STOP:
		{
			// NY - 15 November 2006 - just disable lift running, even when inactive - this means that an deactivate and stop can be sent together,
			// and their order of arrival will be immaterial
			SetRunning(false);

			return true;
		}
		case ASURA_MSG_LIFT_REVERSE:
		{
			SetReversing(!GetReversing());
			return true;
		}
		case ASURA_MSG_LIFT_GOTO_NODE:
		{
			Asura_Assert(pxMessage->GetData(),"message has no data");
			if (pxMessage->GetData())
			{
				const Asura_MsgData_Lift_Goto_Node* pxLiftMessage = static_cast< const Asura_MsgData_Lift_Goto_Node* >( pxMessage->GetData() );
				if (pxLiftMessage->GetTeleportToNode())
				{
					TeleportToNode(pxLiftMessage->GetNodeGuid());
				}
				else
				{
					GotoNode(pxLiftMessage->GetNodeGuid());
				}
			}
			return true;
		}
		
		case ASURA_MSG_LIFT_SETFLAGS:
		{
			Asura_Assert(pxMessage->GetData(),"message has no data");
			if (pxMessage->GetData())
			{
				const Asura_Message_Lift_SetFlags* pxLiftMessage = static_cast< const Asura_Message_Lift_SetFlags* >( pxMessage->GetData() );
				HandleLiftFlagsMessage(pxLiftMessage);
			}
			return true;
		}
		default: break;
	}

	return false;
}

void Asura_ServerEntity_Lift::HandleLiftFlagsMessage(const Asura_Message_Lift_SetFlags* pxLiftMessage)
{
	if (pxLiftMessage->GetAutoReverseOff())
	{
		SetAutoReverse(false);
	}
	else if (pxLiftMessage->GetAutoReverseOn())
	{
		SetAutoReverse(true);
	}
	if (pxLiftMessage->GetPauseOnCollideOff())
	{
		SetPauseOnCollide(false);
	}
	else if (pxLiftMessage->GetPauseOnCollideOn())
	{
		SetPauseOnCollide(true);
	}
	if (pxLiftMessage->GetCollideEnvOff())
	{
		SetCollideEnv(false);
	}
	else if (pxLiftMessage->GetCollideEnvOn())
	{
		SetCollideEnv(true);
	}
	if (pxLiftMessage->GetCollideStaticOff())
	{
		SetCollideStatic(false);
	}
	else if (pxLiftMessage->GetCollideStaticOn())
	{
		SetCollideStatic(true);
	}
	if (pxLiftMessage->GetCollideDynamicOff())
	{
		SetCollideDynamic(false);
	}
	else if (pxLiftMessage->GetCollideDynamicOn())
	{
		SetCollideDynamic(true);
	}
}

void Asura_ServerEntity_Lift::Update( void )
{
	if (m_bRunning)
	{
		float fTimeSlice  = Asura_Timers::GetGameTimeSlice();

		Asura_Vector_3 xPos(GetPosition());
		Asura_Matrix_3x3 xOri(GetOrientation());
		// NY - 09 November 2006 - since m_bReversing is actually one bit of a u_int, it can't be passed as a bool&
		// use a temp local to pass instead.....
		// use locals for all the variables so we can ignore the changes
		bool bReverse(m_bReversing);
		Asura_Guid uSpline(m_uSplineGuid);
		float fProgress(m_fProgress);
		Asura_ServerNode_LiftNode* pxNodeArrivedAt = NULL;
		bool bFinished = UpdateLift(uSpline, fProgress, bReverse, m_bAutoReverse, xPos, xOri, m_uGotoNodeGuid, pxNodeArrivedAt, m_bJustStarted);
		//bool bHasJustStarted = m_bJustStarted;
		m_bJustStarted = false;
		bool bCollided = false;
		if (m_bCollideEnv || m_bCollideDynamic || m_bCollideStatic )
		{
			if (GetCollisionObject())
			{
				if (m_bCollideEnv)
				{
					bCollided = !Asura_Physics_System::CanPlaceObjectAt(xPos, xOri, *GetCollisionObject(), AsuraPolygonFlag_NoCollisions, GetPhysicsObjectHandle(), 65535, m_bCollideStatic, m_bCollideDynamic);
				}
				else
				{
					if ( m_bCollideDynamic || m_bCollideStatic )
					{
						Asura_AMatrix3x3 xOrient;
						
						xOrient.FromMat3( xOri );

						Asura_OBB xOBB( GetCollisionObject()->GetBoundingBox(), Asura_AVector3(xPos), xOrient );

						// AJG - 09/08/07 - The old check was insufficient, as you could block a lift just by standing on it if colliding with dynamics was turned on.
						//					Now, create a broadphase query, then check against the objects found individually.
						Asura_Bounding_Box xEnclosingBB;
						xOBB.CalculateEnclosingAABB(xEnclosingBB);

						Asura_Physics_Broadphase_Query xQuery;
						xQuery.MakeQuery(xEnclosingBB, m_bCollideStatic, m_bCollideDynamic, m_bCollideDynamic, m_bCollideDynamic);

						const Asura_Physics_Broadphase_Entry* pxCurrentEntry = xQuery.GetFirstEntry();

						while (pxCurrentEntry)
						{
							bCollided |= IsInstanceBlockingMovement(pxCurrentEntry->GetInstance(), xPos, xOri);
							pxCurrentEntry = pxCurrentEntry->GetNextEntry();

							if (bCollided) { break; }
						}
					}
				}
			}
			else
			{
				Asura_Assert(0, "no collision object");
			}
		}
		if (bCollided)
		{
			bFinished = false;
			if (!m_bPauseOnCollide)
			{
				if (m_bAutoReverse)
				{
					SetReversing(!m_bReversing);
				}
				else
				{
					bFinished = true;
				}
			}
		}
		else
		{
			if(fTimeSlice > 0.0f)
			{
				//Check for any messages that should be sent as a result of the lift's movement
				const Asura_ServerNode_LiftSpline* pxStartSpline = Asura_ServerNode_LiftSpline::FindEntity(m_uSplineGuid);
				const Asura_ServerNode_LiftSpline* pxEndSpline = Asura_ServerNode_LiftSpline::FindEntity(uSpline);

				TriggerMessagesFromMovement(pxStartSpline, m_fProgress, m_bReversing,
											pxEndSpline, fProgress, bReverse);
			}
			

			SetReversing(bReverse);
			m_uSplineGuid = uSpline;
			m_fProgress = fProgress;
			SetPosition(xPos);
			SetOrientation(xOri);
		}

		if (bFinished)
		{
			m_uGotoNodeGuid = ASURA_GUID_UNREGISTERED;
			SetRunning(false);
			if ( pxNodeArrivedAt && pxNodeArrivedAt->GetBlockWhenOccupied() )
			{
				BlockLiftNodeLinks( true );
			}
		}
	}
}


void Asura_ServerEntity_Lift::TriggerMessagesFromMovement(const Asura_ServerNode_LiftSpline* pxStartSpline, float fStartProgress, bool bStartReversed,
															const Asura_ServerNode_LiftSpline* pxEndSpline, float fEndProgress, bool bEndReversed)
{
	if(!pxStartSpline) return;
	if(!pxEndSpline) return;

	const Asura_ServerNode_LiftSpline* pxSpline = pxStartSpline;
	float fProgress = fStartProgress;
	bool bReversed = bStartReversed;

	while(true)
	{
		bool bDone = false;
		
		float fDestProgress;

		//Work out which interval to trigger messages for on this stage.
		if(bReversed == bEndReversed && pxSpline == pxEndSpline)
		{
			//On the final spline and in the correct direction , so just go up to the end position , and we are done.
			fDestProgress = fEndProgress;
			bDone = true;
		}
		else
		{
			//Not on the corrent spline/direction yet , so trigger messages up to the end/start of this spline
			fDestProgress = bReversed ? 0.0f : 1.0f;
		}

		if(fProgress != fDestProgress)
		{
			if(bReversed)
			{
				if(fProgress >= 1.0f && fDestProgress < 1.0f)
				{
					//Leaving the end point on the spline.
					//Trigger departure message
					Asura_ServerNode_LiftNode* pxNode = Asura_ServerNode_LiftNode::FindEntity( pxSpline->GetTargetGuid() );
					if(pxNode) pxNode->SendOnDepartureMessages();
					SendMessagesForEvent(ASURA_ENTITY_EVENT_ONDEPARTEDFROM);

				}
				//deal with any messages along the spline
				u_int uNumSplineEvents = pxSpline->GetNumSplineEvents();
				if(uNumSplineEvents)
				{
					float fStartDist = pxSpline->GetTotalLength() * fProgress;
					float fEndDist = pxSpline->GetTotalLength() * fDestProgress;
					
					for(int i=uNumSplineEvents-1; i>=0; i--)
					{
						if(!pxSpline->GetSplineEventFlag_TriggerWhenReversed(i)) continue;

						float fEventDist = pxSpline->GetSplineEventDistance(i);
						//Only trigger for events that are at the start of the current interval if they are
						//also at the end of the spline , to avoid the possibility of being triggered
						//on consecutive frames.
						if((fStartDist > fEventDist && fEventDist >= fEndDist) ||
							(fStartDist == fEventDist && fProgress == 1.0f))
						{
							pxSpline->SendMessagesForSplineEvent(i);
						}
					} 
				}

				if(fProgress > 0.0f && fDestProgress <= 0.0f)
				{
					//Reached start point on the spline.
					//Trigger arrival message
					Asura_ServerNode_LiftNode* pxNode = Asura_ServerNode_LiftNode::FindEntity( pxSpline->GetSourceGuid() );
					if(pxNode) pxNode->SendOnArrivalMessages();
					SendMessagesForEvent(ASURA_ENTITY_EVENT_ONARRIVEDAT);
				}
			}
			else
			{
				if(fProgress <= 0.0f && fDestProgress > 0.0f)
				{
					//Leaving the start point on the spline.
					//Trigger departure message
					Asura_ServerNode_LiftNode* pxNode = Asura_ServerNode_LiftNode::FindEntity( pxSpline->GetSourceGuid() );
					if(pxNode) pxNode->SendOnDepartureMessages();
					SendMessagesForEvent(ASURA_ENTITY_EVENT_ONDEPARTEDFROM);

				}
				//deal with any messages along the spline
				u_int uNumSplineEvents = pxSpline->GetNumSplineEvents();
				if(uNumSplineEvents)
				{
					float fStartDist = pxSpline->GetTotalLength() * fProgress;
					float fEndDist = pxSpline->GetTotalLength() * fDestProgress;
					
					for(u_int i=0; i<uNumSplineEvents; i++)
					{
						if(pxSpline->GetSplineEventFlag_TriggerWhenReversed(i)) continue;

						float fEventDist = pxSpline->GetSplineEventDistance(i);
						//Only trigger for events that are at the start of the current interval if they are
						//also at the start of the spline , to avoid the possibility of being triggered
						//on consecutive frames.
						if((fStartDist < fEventDist && fEventDist <= fEndDist) ||
							(fStartDist == fEventDist && fProgress == 0.0f))
						{
							pxSpline->SendMessagesForSplineEvent(i);
						}
					} 
				}

				if(fProgress < 1.0f && fDestProgress >= 1.0f)
				{
					//Reached end point on the spline.
					//Trigger arrival message
					Asura_ServerNode_LiftNode* pxNode = Asura_ServerNode_LiftNode::FindEntity( pxSpline->GetTargetGuid() );
					if(pxNode) pxNode->SendOnArrivalMessages();
					SendMessagesForEvent(ASURA_ENTITY_EVENT_ONARRIVEDAT);
				}
			}
		}
		if(bDone) break;

		//Determine the next spline/direction to continue from in the next loop
		const Asura_ServerNode_LiftNode* pxNode = Asura_ServerNode_LiftNode::FindEntity( bReversed ? pxSpline->GetSourceGuid() : pxSpline->GetTargetGuid() );
		const Asura_ServerNode_LiftSpline* pxNextSpline = GetSplineFromNode(pxNode, bReversed);

		if(pxNextSpline)
		{
			//There is another spline to go to , so continue from the start/end of that one
			pxSpline = pxNextSpline;
			fProgress = bReversed ? 1.0f : 0.0f;
		}
		else
		{
			//No new spline , so reverse direction on the current spline
			bReversed = !bReversed;
			fProgress = bReversed ? 1.0f : 0.0f;
		}

	}
}

bool Asura_ServerEntity_Lift::UpdateLift( Asura_Guid& uSpline, float& fProgress, bool& bReverse, const bool bAutoReverse,
										  Asura_Vector_3& xPos, Asura_Matrix_3x3& xOri, const Asura_Guid uGotoNodeGuid, 
										  Asura_ServerNode_LiftNode*& pxNodeArrivedAt, const bool bJustStarted )
{
	const Asura_ServerNode_LiftSpline* pxSpline = Asura_ServerNode_LiftSpline::FindEntity( uSpline );
	if ( !pxSpline )
	{
		return true;
	}

	float fSplineTime = pxSpline->GetTotalLinkTime();
	float fTimeSlice  = Asura_Timers::GetGameTimeSlice();
	bool  bFinished   = false;
	//while loop to use up timeslice, so that movement flows across spline joins
	while (fTimeSlice > 0.0f && !bFinished)
	{
		bool bOverflowed = false;

		fProgress += bReverse ? ( -fTimeSlice / fSplineTime ) : ( fTimeSlice / fSplineTime );
		if ( fProgress < 0.0f )
		{
			//we've run out of spline - correct overflow and set timeslice to remainder
			fTimeSlice = -fProgress * fSplineTime;
			fProgress = 0.0f;
			bOverflowed = true;
		}
		else if ( fProgress > 1.0f )
		{
			//we've run out of spline - correct overflow and set timeslice to remainder
			fTimeSlice = (fProgress - 1.0f) * fSplineTime;
			fProgress = 1.0f;
			bOverflowed = true;
		}
		else
		{
			//mark timeslice as used up
			fTimeSlice = 0.0f;
		}

		if ( bOverflowed )
		{
			// Try and switch onto the next spline. Reverse or stop if there isn't one.
			pxNodeArrivedAt = Asura_ServerNode_LiftNode::FindEntity( bReverse ? pxSpline->GetSourceGuid() : pxSpline->GetTargetGuid() );

			// ignore stop here flag if we've just been told to start, to make sure we pull away from the node we'd stopped at
			if ( pxNodeArrivedAt->GetStopHere() && !bJustStarted )
			{
				bFinished = true;
			}
			if ( uGotoNodeGuid == pxNodeArrivedAt->GetGuid() )
			{
				bFinished = true;
			}

			// If we should continue then look for the next spline.
			if ( !bFinished )
			{
				bool bFoundNext = false;
				const Asura_ServerNode_LiftSpline* pxNextSpline = GetSplineFromNode(pxNodeArrivedAt, bReverse);
				if (pxNextSpline)
				{
					pxSpline = pxNextSpline;
					fSplineTime = pxSpline->GetTotalLinkTime();
					fProgress += bReverse ? 1.0f : -1.0f;
					uSpline = pxSpline->GetGuid();
					bFoundNext = true;
				}

				if ( !bFoundNext )
				{
					//we didn't find one, so we've reached a dead end
					if (bAutoReverse)
					{
						//stay on current spline and reverse course
						bReverse = !bReverse;
					}
					else
					{
						//nowhere to go, so stop
						bFinished = true;
					}
				}
			}
		}
	}

	Asura_Quat xOriQuat;
	xOriQuat.FromMatrix(xOri);
	pxSpline->GetPositionFromTime(fProgress * fSplineTime, xPos, &xOriQuat);
	xOriQuat.ToMatrix(xOri);

	return bFinished;
}

const Asura_ServerNode_LiftSpline* Asura_ServerEntity_Lift::GetSplineFromNode(const Asura_ServerNode_LiftNode* pxNodeArrivedAt, const bool bReverse)
{
	if (!pxNodeArrivedAt)
	{
		Asura_Assert(0, "no node to find from");
		return 0;
	}
	const Asura_ServerNode_LiftSpline* pxSpline = 0;
	if ( bReverse )
	{
		//loop through splines, searching for any that end on this node (note early out on pxSpline as we just use first for now)
		for ( Asura_Hashed_List_ConstIt< Asura_ServerNode_LiftSpline > xIt( Asura_ServerNode_LiftSpline::GetList() );
			!xIt.Done() && !pxSpline;
			xIt.Next() )
		{
			const Asura_ServerNode_LiftSpline* pxSearchSpline = xIt.GetCurrent();
			if ( !pxSearchSpline || !pxSearchSpline->IsEntityActive() ) { continue; }

			if ( pxSearchSpline->GetTargetGuid() == pxNodeArrivedAt->GetGuid() )
			{
				//use first active spline found
				pxSpline = pxSearchSpline;
			}
		}
	}
	else
	{
		for ( u_int u = 0; u < pxNodeArrivedAt->GetNumberLinks() && !pxSpline; ++u )
		{
			const Asura_ServerNode_LiftSpline* pxNextSpline = Asura_ServerNode_LiftSpline::FindEntity( pxNodeArrivedAt->GetLink( u ) );
			if ( !pxNextSpline && !pxNextSpline->IsEntityActive() ) { continue; }

			//use the first active spline found
			pxSpline = pxNextSpline;
		}
	}
	return pxSpline;
}

void Asura_ServerEntity_Lift::TeleportToNode( Asura_Guid uGuid )
{
	const Asura_ServerNode_LiftNode* pxNode = Asura_ServerNode_LiftNode::FindEntity( uGuid );

	if (pxNode)
	{
		const Asura_ServerNode_LiftSpline* pxNextSpline = GetSplineFromNode(pxNode, m_bReversing);
		if (pxNextSpline)
		{
			m_uSplineGuid = pxNextSpline->GetGuid();
			float fSplineTime = pxNextSpline->GetTotalLinkTime();
			m_fProgress = m_bReversing ? 1.0f : 0.0f;

			Asura_Vector_3 xPos;
			Asura_Matrix_3x3 xOri;
			Asura_Quat xOriQuat;
			pxNextSpline->GetPositionFromTime(m_fProgress * fSplineTime, xPos, &xOriQuat);
			xOriQuat.ToMatrix(xOri);
			TeleportTo(xPos, xOri, true);

			m_uGotoNodeGuid = ASURA_GUID_UNREGISTERED;
		}
		else
		{
			// OK, that failed, try the opposite approach but set up the spline values as if we're ending what we find instead of starting it.
			const Asura_ServerNode_LiftSpline* pxPreviousSpline = GetSplineFromNode(pxNode, !m_bReversing);
			if (pxPreviousSpline)
			{
				m_uSplineGuid = pxPreviousSpline->GetGuid();
				float fSplineTime = pxPreviousSpline->GetTotalLinkTime();
				m_fProgress = m_bReversing ? 0.0f : 1.0f;

				Asura_Vector_3 xPos;
				Asura_Matrix_3x3 xOri;
				Asura_Quat xOriQuat;
				pxPreviousSpline->GetPositionFromTime(m_fProgress * fSplineTime, xPos, &xOriQuat);
				xOriQuat.ToMatrix(xOri);
				TeleportTo(xPos, xOri, true);

				m_uGotoNodeGuid = ASURA_GUID_UNREGISTERED;
			}
		}
	}
}

void Asura_ServerEntity_Lift::GotoNode( Asura_Guid uGuid )
{
	m_uGotoNodeGuid = uGuid;

	if (m_uSplineGuid != ASURA_GUID_UNREGISTERED)
	{
		// Get our current spline
		Asura_ServerNode_LiftSpline* pxCurrentSpline = Asura_ServerNode_LiftSpline::FindEntity(m_uSplineGuid);
		if (pxCurrentSpline)
		{
			if (pxCurrentSpline->GetSourceGuid() == uGuid)
			{
				// Go to the start of our current spline
				if (m_fProgress <= 0.0f)
    			{
					SetRunning(false);		// Already there.
				}
				else
				{
					SetRunning(true);
					SetReversing(true);
				}
				return;
			}

			if (pxCurrentSpline->GetTargetGuid() == uGuid)
			{
				// Go to the end of our current spline
				if (m_fProgress >= 1.0f)
				{
					SetRunning(false);		// Already there.
				}
				else
				{
					SetRunning(true);
					SetReversing(false);
				}
				return;
			}

			// Ok. the target node is not on our current spline, so which way do we need to go?
			// Try looking forwards first
			const Asura_ServerNode_LiftSpline* pxSearchSpline = pxCurrentSpline;
			while (pxSearchSpline)
			{
				if (pxSearchSpline->GetTargetGuid() == uGuid)
				{
					// Found it... Need to go forwards along our spline list
					SetRunning(true);
					SetReversing(false);
					return;
				}
				pxSearchSpline = GetNextSpline(pxSearchSpline, true);
			}

			// Right then. I guess the target must be backwards along our spline list
			pxSearchSpline = pxCurrentSpline;
			while (pxSearchSpline)
			{
				if (pxSearchSpline->GetSourceGuid() == uGuid)
				{
					// Found it... Need to go backwards along our spline list
					SetRunning(true);
					SetReversing(true);
					return;
				}
				pxSearchSpline = GetNextSpline(pxSearchSpline, false);
			}
		}
	}
	Asura_Assertf(false, ("Lift unable to goto node (GUID %d)", uGuid));
}

#ifdef ASURA_TOOLS
void Asura_ServerEntity_Lift::ApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance )
{
	m_uSplineGuid = pxInstance->ConvertGuid(m_uSplineGuid);
}

void Asura_ServerEntity_Lift::SetInitialNode( const int iNode )
{
	m_bAtNode   = true;
	m_fProgress = iNode ? 1.0f : 0.0f;
}
#endif

const Asura_ServerNode_LiftSpline* Asura_ServerEntity_Lift::GetNextSpline(const Asura_ServerNode_LiftSpline* pxSpline, bool bForwards)
{
	// Find the next spline connected to this one. (forwards or backwards)
	for ( Asura_Hashed_List_ConstIt< Asura_ServerNode_LiftSpline > xIt( Asura_ServerNode_LiftSpline::GetList() ); !xIt.Done(); xIt.Next() )
	{
		const Asura_ServerNode_LiftSpline* pxSearchSpline = xIt.GetCurrent();
		if (pxSearchSpline)
		{
			if (bForwards)
			{
				if (pxSearchSpline->GetSourceGuid() == pxSpline->GetTargetGuid() && pxSearchSpline->IsEntityActive())
				{
					return pxSearchSpline;
				}
			}
			else
			{
				if (pxSearchSpline->GetTargetGuid() == pxSpline->GetSourceGuid() && pxSearchSpline->IsEntityActive())
				{
					return pxSearchSpline;
				}
			}
		}
	}

	return NULL;
}

void Asura_ServerEntity_Lift::BlockLiftNodeLinks( const bool bBlock )
{
	if ( bBlock )
	{
		InitialiseWaypointLinkBlockCalculations();
		SetBlockWaypointLinksFlag( true );
		CalculateWaypointLinkBlocks();
		SetBlockWaypointLinksFlag( false );
		ShutdownWaypointLinkBlockCalculations();
		BlockLinks( true );
	}
	else
	{
		BlockLinks( false );
	}
}

void Asura_ServerEntity_Lift::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
    xStream << uCHUNK_VERSION;

	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xFlagsBitfield( m_uLiftFlags );
	xFlagsBitfield.Add(m_bOldActive, 1);//need to do this to pad correctly
	xFlagsBitfield.Add(m_bReversing, 1);
	xFlagsBitfield.Add(m_bAutoReverse, 1);
	xFlagsBitfield.Add(m_bPauseOnCollide, 1);
	xFlagsBitfield.Add(m_bCollideEnv, 1);
	xFlagsBitfield.Add(m_bCollideStatic, 1);
	xFlagsBitfield.Add(m_bCollideDynamic, 1);
	xFlagsBitfield.Add(m_bRunning, 1);
	xFlagsBitfield.Add(m_bJustStarted, 1);
	xFlagsBitfield.Add(m_bSendStartSound, 1);
	xFlagsBitfield.Add(m_bSendStopSound, 1);
	xFlagsBitfield.Add(m_bAtNode, 1);
	xStream << xFlagsBitfield;
	xStream << m_fProgress;
	xStream << m_uSplineGuid;
	xStream << m_uGotoNodeGuid;
	const int iCollisionProxy = GetCollisionProxy();
	xStream << iCollisionProxy;
	xStream << m_uStartSoundEventHash;
	xStream << m_uStopSoundEventHash;
}

void Asura_ServerEntity_Lift::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;

	if ( uVersion == 4 )
	{
		Asura_ReadFromStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uLiftFlags );
		xStream >> xFlagsBitfield;
		xStream >> m_fProgress;
		xStream >> m_uSplineGuid;
		xStream >> m_uGotoNodeGuid;

		int iCollisionProxy;
		xStream >> iCollisionProxy;
		SetCollisionProxy( static_cast<Asura_Physics_Object::COLLISION_PROXY>(iCollisionProxy) );

		xStream >> m_uStartSoundEventHash;
		xStream >> m_uStopSoundEventHash;
	}
	else if ( uVersion == 3 )
	{
		Asura_ReadFromStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uLiftFlags );
		xStream >> xFlagsBitfield;
		SetEntityActiveFromLoad(m_bOldActive);
		m_bOldActive = 0;
		xStream >> m_fProgress;
		xStream >> m_uSplineGuid;
		xStream >> m_uGotoNodeGuid;

		int iCollisionProxy;
		xStream >> iCollisionProxy;
		SetCollisionProxy( static_cast<Asura_Physics_Object::COLLISION_PROXY>(iCollisionProxy) );

		xStream >> m_uStartSoundEventHash;
		xStream >> m_uStopSoundEventHash;
	}
	else if ( uVersion == 2 )
	{
		Asura_ReadFromStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uLiftFlags );
		xStream >> xFlagsBitfield;
		xStream >> m_fProgress;
		xStream >> m_uSplineGuid;
		xStream >> m_uGotoNodeGuid;

		int iCollisionProxy;
		xStream >> iCollisionProxy;
		SetCollisionProxy( static_cast<Asura_Physics_Object::COLLISION_PROXY>(iCollisionProxy) );
	}
#ifdef ASURA_TOOLS
	else if ( uVersion == 1 )
	{
		Asura_ReadFromStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uLiftFlags );
		xStream >> xFlagsBitfield;
		xStream >> m_fProgress;
		xStream >> m_uSplineGuid;
		xStream >> m_uGotoNodeGuid;
	}
	else if ( uVersion == 0 )
	{
		Asura_ReadFromStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uLiftFlags );
		xStream >> xFlagsBitfield;
		xStream >> m_fProgress;
		xStream >> m_uSplineGuid;
	}
#endif
	else
	{
		Asura_Assertf( false, ( "Asura_ServerEntity_Lift::ReadFromChunkStream - unsupported chunk version (%u)", uVersion ) );
	}

	if(m_bRunning)
	{
		m_bSendStartSound = true;
	}
}

void Asura_ServerEntity_Lift::InitialiseFromChunkBuffer( const char*& pcChunkBfr )
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBfr);
	Asura_ServerEntity_Lift::ReadFromChunkStream(xStream);
}


void Asura_ServerEntity_Lift::MakeClientUpdateMessage( void )
{
	Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_CLIENTUPDATE, sizeof(Asura_Message_PhysicalObject_ClientUpdate), GetClassification() );
	if( !pxMsg ) { return; }

	Asura_Message_PhysicalObject_ClientUpdate* pxMsgData = static_cast<Asura_Message_PhysicalObject_ClientUpdate*>( pxMsg->GetData() );
	pxMsgData->FillPhysical( this );

	Send( pxMsg, ASURA_GUID_MY_CLIENT_ENTITIES );

	if( m_bSendStartSound )
	{
		//play lift start sound, if we have one
		if( m_uStartSoundEventHash )
		{
			Asura_Message* pxStartMsg = Asura_Network::CreateMessage( ASURA_MSG_PLAYSOUNDEVENT, sizeof(Asura_MsgData_PlaySoundEvent) );
			if( pxStartMsg )
			{
				Asura_MsgData_PlaySoundEvent* const pxSoundEventMsgData = static_cast<Asura_MsgData_PlaySoundEvent*> (pxStartMsg->GetData());
				pxSoundEventMsgData->m_uSoundEventHash = m_uStartSoundEventHash ;
				Send( pxStartMsg, ASURA_GUID_MY_CLIENT_ENTITIES );
			}
		}
		m_bSendStartSound = false;
	}

	if( m_bSendStopSound )
	{
		//send message when lift has stopped, regardless of whether we have a sound - we still need to stop the travelling sound from looping
		Asura_Message* pxStopMsg = Asura_Network::CreateMessage( ASURA_MSG_PLAYSOUNDEVENT, sizeof(Asura_MsgData_PlaySoundEvent) );
		if( pxStopMsg )
		{	
			Asura_MsgData_PlaySoundEvent* const pxSoundEventMsgData = static_cast<Asura_MsgData_PlaySoundEvent*> (pxStopMsg->GetData());
			pxSoundEventMsgData->m_uSoundEventHash = m_uStopSoundEventHash;
			Send( pxStopMsg, ASURA_GUID_MY_CLIENT_ENTITIES );
		}
		
		m_bSendStopSound = false;
	}
}

void Asura_ServerEntity_Lift::PostLoadInitialise()
{
	PARENT::PostLoadInitialise();

	if (m_pxDynamicObject)
	{
		m_pxDynamicObject->SetUseAsRestFrame(true);
	}

	// Block links if we're supposed to.
	if ( m_bAtNode )
	{
		const Asura_ServerNode_LiftSpline* pxSpline = Asura_ServerNode_LiftSpline::FindEntity( GetSplineGuid() );
		if ( pxSpline )
		{
			const Asura_ServerNode_LiftNode* pxNode =
				Asura_ServerNode_LiftNode::FindEntity( ( GetProgress() < 0.000001f ) ? pxSpline->GetSourceGuid() : pxSpline->GetTargetGuid() );
			if ( pxNode && pxNode->GetBlockWhenOccupied() )
			{
				BlockLiftNodeLinks( true );
			}
		}
	}
}

bool Asura_ServerEntity_Lift::IsInstanceBlockingMovement(const Asura_Physics_RigidBodyInstance* pxInstance, const Asura_Vector_3& xLiftPos, const Asura_Matrix_3x3& xLiftOri) const
{
	if (!m_pxDynamicObject || !pxInstance) { return false; }

	if (pxInstance->GetHandleToIgnore() == m_pxDynamicObject->GetHandle()
		|| m_pxDynamicObject->GetHandleToIgnore() == pxInstance->GetHandle())
	{
		return false;
	}
	// Ignore objects resting on us
	if (pxInstance->HasMatchingRestFrame(m_pxDynamicObject->GetOwnedRestFrame())) { return false; }

	const Asura_AVector3 xAVectorLiftPos(xLiftPos);

	// AJG - 09/08/07 - First version is just equivalent to the original blocking test - i.e. using an OBB regardless of whether that's what the lift itself uses or not.
	Asura_OBB xLiftOBB( GetCollisionObject()->GetBoundingBox(), xAVectorLiftPos, Asura_AMatrix3x3(xLiftOri) );
	Asura_OBB xInstanceOBB( pxInstance->GetHalfExtents(), pxInstance->GetCollisionMeshCentre(), pxInstance->GetOrientation() );
	if ( xInstanceOBB.Intersects( xLiftOBB ) )
	{
		// AJG - 09/08/07 - The above isn't really accurate enough, particularly with polygonal lifts.
		//					The next test fixes a few cases, but is a little hacky and still leaves a number of problem cases.
		//					The test gets the offset to the nearest point on the object's box from the centre of the lift, and checks if it's approximately in the direction of travel.

		// The test doesn't really work with orientation-change-only lifts, so skip it for them.
		const Asura_AVector3 xTravel = xAVectorLiftPos - Asura_AVector3(GetPosition());

		if (xTravel.ApproximatelyEquals(Asura_Maths::ZeroAVector3)) { return true; }

		Asura_AVector3 xOffset = xAVectorLiftPos - pxInstance->GetCollisionMeshCentre();
		pxInstance->GetOrientation().ApplyTranspose(xOffset);
		const Asura_Vector_3 xVectorOffset(xOffset);
		Asura_Bounding_Box xUnrotatedBox;
		xUnrotatedBox.Reset();
		const Asura_Vector_3 xHalfExtents(pxInstance->GetHalfExtents());
		xUnrotatedBox.Set(xHalfExtents);
		xUnrotatedBox.Set(-xHalfExtents);

		Asura_Vector_3 xResult;
		xUnrotatedBox.CalcMinVectorIntoBox(xVectorOffset, xResult);

		xOffset = xResult;
		xOffset = xOffset * pxInstance->GetOrientation();

		if ((xTravel * xOffset) > 0.0f)
		{
			return true;
		}
	}

	return false;
}

void Asura_Message_Lift_SetFlags::FillInChunkBuffer(char*& pcChunkBuffer) const
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	WriteToChunkStream(xStream);
}

void Asura_Message_Lift_SetFlags::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	Asura_WriteToStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uFlags );
	xFlagsBitfield.Add(m_bAutoReverseOn, 1);
	xFlagsBitfield.Add(m_bAutoReverseOff, 1);
	xFlagsBitfield.Add(m_bPauseOnCollideOn, 1);
	xFlagsBitfield.Add(m_bPauseOnCollideOff, 1);
	xFlagsBitfield.Add(m_bCollideEnvOn, 1);
	xFlagsBitfield.Add(m_bCollideEnvOff, 1);
	xFlagsBitfield.Add(m_bCollideStaticOn, 1);
	xFlagsBitfield.Add(m_bCollideStaticOff, 1);
	xFlagsBitfield.Add(m_bCollideDynamicOn, 1);
	xFlagsBitfield.Add(m_bCollideDynamicOff, 1);
	xStream << xFlagsBitfield;
}

void Asura_MsgData_Lift_Goto_Node::FillInChunkBuffer(char*& pcChunkBuffer) const
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	WriteToChunkStream(xStream);
}

void Asura_MsgData_Lift_Goto_Node::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	Asura_WriteToStream_Bitfield_Wrapper<u_int> xFlagsBitfield( m_uFlags );
	xFlagsBitfield.Add(m_bTeleportToNode, 1);
	xStream << xFlagsBitfield;
	xStream << m_uNodeGuid;
}
