
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_MoveData_Animation.h
// Author : Andy Tate
// Date   : 7-Sep-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_MOVEDATA_ANIMATION_H__
#define __ASURA_MOVEDATA_ANIMATION_H__

#if defined(ASURA_USE_MOVEMENT_SYSTEM)

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Entity/PhysicalObject/Moves/Asura_MoveData.h"
#include "Core/Asura_Maths.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation;
class Asura_AnimationController_Animation;
class Asura_Hierarchy_Anim;
class Asura_Hierarchy_Skin;
class Asura_AVector3;
class Asura_Quat;

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_MoveData_Animation
// Author       : Andy Tate
// Date         : 5-Sep-2007
//
// Description  : Class to hold the animation data required during a move.
////////////////////////////////////////////////////////////////////////////////

class Asura_MoveData_Animation : public Asura_MoveData
{
public:
	// Public functions...
	Asura_MoveData_Animation(const Asura_Hash_ID uAnimHash, const u_int uFlags = 0, const u_int uPriority = 0x00008000, const Asura_Hash_ID uControllerGroupHash = ASURA_HASH_ID_UNSET);
	virtual ~Asura_MoveData_Animation() {}

	// Manipulation functions...
	virtual void	Reset(void);

	virtual bool	Start(const bool bLoopAtEnd, Asura_Animation& xAnimation);
	virtual bool	Update(const float fTimeStep);
	virtual void	Finish(void);

	virtual void	StopLoopedAnims();

	// Animation query functions...
	virtual bool	CalculateMovementOffset(const float fParametricTimer, Asura_AVector3* pxPosition, Asura_Quat* pxOrientation) const;
	virtual bool	CalculateVelocity(const float fParametricTimer, Asura_AVector3& xVelocity) const;
	virtual bool	CalculateAngularVelocity(const float fParametricTimer, Asura_AVector3& xAxis, Asura_Angle& fAngle) const;

	bool	CalculateMovementOffset(Asura_AVector3* pxPosition, Asura_Quat* pxOrientation) const	{ return(CalculateMovementOffset(GetParametricTimer(), pxPosition, pxOrientation)); }
	bool	CalculateVelocity(Asura_AVector3& xVelocity) const										{ return(CalculateVelocity(GetParametricTimer(), xVelocity)); }
	bool	CalculateAngularVelocity(Asura_AVector3& xAxis, Asura_Angle& fAngle) const				{ return(CalculateAngularVelocity(GetParametricTimer(), xAxis, fAngle)); }

	virtual Asura_Hash_ID	GetAnimationControllerHash(void) const { return m_uControllerAnimationHash; }
	virtual const Asura_AnimationController_Animation*	GetAnimationController(void) const;
	virtual Asura_AnimationController_Animation*		GetAnimationController(void);

	float										GetAnimationDuration(void) const;
	
	virtual const Asura_Hierarchy_Anim*					GetHierarchyAnim(void) const;
	
	const Asura_Hierarchy_Skin*					GetHierarchySkin(void) const;

	// Accessor functions...
	virtual void	SetAnimHash(const Asura_Hash_ID uHash)				{ m_uAnimHash = uHash; }
	void	SetAnimFlags(const u_int uFlagsToSet)				{ m_uAnimFlags |= uFlagsToSet; }
	void	ClearAnimFlags(const u_int uFlagsToClear)			{ m_uAnimFlags &= ~uFlagsToClear; }
    void	SetAnimPriority(const u_int uPriority)				{ m_uAnimPriority = uPriority; }
    void    SetControllerGroupHash(const Asura_Hash_ID uHash)	{ m_uControllerGroupHash = uHash; }
    void    SetLeaveAnimAtEnd(const bool bValue)				{ m_bLeaveAnimAtEnd = bValue; }
	void    SetWaitForCompletion(const bool bValue)				{ m_bWaitForCompletion = bValue; }
	void	SetDontCalculateVelocity(const bool bValue)			{ m_bDontCalculateVelocity = bValue; }
	void	SetDontCalculateAngularVelocity(const bool bValue)	{ m_bDontCalculateAngularVelocity = bValue; }

	void	SetTweenData(const float fTweenInDuration, const float fTweenOutDuration, const bool bAnimatedTween);
	void	SetParametricTimes(const float fStartTime, const float fEndTime);

	bool	IsLooped(void) const;
	bool	IsWaitingForCompletion(void) const	{ return(m_bWaitForCompletion); }
	bool	GetLeaveAnimAtEnd() const { return m_bLeaveAnimAtEnd; }

	bool	GetUseAnimatedTween() const { return m_bAnimatedTween; }
	u_int	GetAnimPriority() const { return m_uAnimPriority; }
	Asura_Hash_ID GetControllerGroupHash() const { return m_uControllerGroupHash; }

	virtual bool	HasValidAnimation(void) const	{ return(m_uAnimHash != ASURA_HASH_ID_UNSET); }

	float	GetParametricTimer(void) const;

	// Call this if you've changed your tween data mid-move, and want to make sure these changes get applied to the blend.
	void UpdateTweenData();

	float GetTweenInDuration() const { return m_fTweenInDuration; }
	float GetTweenOutDuration() const { return m_fTweenOutDuration; }

protected:
	// Protected variables...
	Asura_Hash_ID		m_uAnimHash;				// The hash-ID of the animation to play.
	u_int				m_uAnimFlags;				// The flags for the animation to play.
	u_int				m_uAnimPriority;			// The priority of the animation to play.
	Asura_Hash_ID		m_uControllerGroupHash;		// The hash-ID of the animation-group in which we are to create our animation.

	Asura_Animation*	m_pxAnimation;				// A pointer to the animation we're using.
	Asura_Hash_ID		m_uControllerAnimationHash;	// The hash-ID of the animation we are actually playing.

	float				m_fTweenInDuration;			// The tween-in duration for the animation to play.
	float				m_fTweenOutDuration;		// The tween-in duration for the animation to play.

	float				m_fParametricStartTime;		// The parametric start-time within the animation to play.
	float				m_fParametricEndTime;		// The parametric end-time within the animation to play.

	union
	{
		u_int		m_uFlags;						// Single-access point for the flags for this animation.

		struct 
		{
			u_int	m_bAnimatedTween				: 1,	// Flag to say whether our blend is to use an animated-tween.
					m_bWaitForCompletion			: 1,	// Flag to say that we want to wait for this animation completing before finishing the move.
                    m_bLeaveAnimAtEnd				: 1,    // Keep our animation around once we're finished, even if reset.
					m_bDontCalculateVelocity		: 1,	// Flags to prevent the automatic adding of the animation movement data calculations.
					m_bDontCalculateAngularVelocity	: 1,
													: 0;	// 27 spare-bits.
		};
	};

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_MoveData PARENT;

	// Private variables...

	// Private functions...
	Asura_MoveData_Animation(void);
	Asura_MoveData_Animation(const Asura_MoveData_Animation& xOther);

	Asura_MoveData_Animation&	operator=(const Asura_MoveData_Animation& xOther);

};   // End of class 'Asura_MoveData_Animation'.


////////////////////////////////////////////////////////////////////////////////
// Inlined functions...
////////////////////////////////////////////////////////////////////////////////
inline void Asura_MoveData_Animation::SetTweenData(const float fTweenInDuration, const float fTweenOutDuration, const bool bAnimatedTween)
{
	m_fTweenInDuration	= fTweenInDuration;
	m_fTweenOutDuration	= fTweenOutDuration;
	m_bAnimatedTween	= bAnimatedTween;
}

inline void Asura_MoveData_Animation::SetParametricTimes(const float fStartTime, const float fEndTime)
{
	m_fParametricStartTime	= fStartTime;
	m_fParametricEndTime	= fEndTime;
}

#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)

#endif   // ifndef __ASURA_MOVEDATA_ANIMATION_H__


/// End of file. ///
