
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_MoveData_OrientationInterpolation.cpp
// Author : Andy Tate
// Date   : 10-Sep-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(ASURA_USE_MOVEMENT_SYSTEM)

// This file's header...
#include "Entity/PhysicalObject/Moves/Asura_MoveData_OrientationInterpolation.h"

// Asura independent header-files...
#include "Debugging/Asura_Assert.h"
#include "Entity/PhysicalObject/Moves/Asura_Moves.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float	dbg_fSingularityAngleThreshold	= 3.140f;	// Close to PI, but not exact.


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::CalculateWSSlerp
// Author       : Andy Tate
// Date         : 16-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_OrientationInterpolation::CalculateWSSlerp(const float fParametricTimer,
															   const Asura_Quat& xFinalOrientation,
															   Asura_Quat& xCurrentOrientation) const
{
	if((m_fSingularityAngle != 0.0f) && m_bHasSingularityHintAxis)
	{
		const Asura_Angle	fRotationRequired	= m_fSingularityAngle * fParametricTimer;

		Asura_Quat			xRotation;

		// We've got a singularity - use our hint-axis and rotation...
		xRotation.FromRotation(fRotationRequired, m_xSingularityHintAxis);
		xCurrentOrientation	= m_xInitial * xRotation;
	}
	else
	{
		xCurrentOrientation.Slerp(m_xInitial, xFinalOrientation, fParametricTimer);
	}
}

/// End of function 'Asura_MoveData_OrientationInterpolation::CalculateWSSlerp'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::UpdateDurationBased
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_OrientationInterpolation::UpdateDurationBased(void)
{
	if(m_fTimer < m_fDuration)
	{
		const float	fParametricTimer	= m_fTimer / m_fDuration;

		// Duration based simply implies interpolation from initial -> final...
		m_xCurrent.Slerp(m_xInitial, m_xFinal, fParametricTimer);
		m_xActual	= m_xCurrent;
		return(true);
	}
	else
	{
		if(m_bSnapToFinal)
		{
			m_xCurrent	= m_xFinal;
			m_xActual	= m_xCurrent;
		}

		return(false);
	}
}

/// End of function 'Asura_MoveData_OrientationInterpolation::UpdateDurationBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::UpdateDurationBased
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_OrientationInterpolation::UpdateDurationBased(const float fParametricTimer,
																  const Asura_Quat& xLSMOCurrentOrientation,
																  const Asura_Quat& xLSMOFinalOrientation)
{
	// We're still updating with movement-offsets - calculate the current values...
	switch(m_eMovementOffsetOperation)
	{
		case IGNORE_MOVEMENT_OFFSET:
		{
			Asura_Assert(false, "Shouldn't be here if we don't have a movement-offset operation.");
			return(false);
		}

		case ADD_MOVEMENT_OFFSET:
		{
			// Simply add the movement-offset onto the interpolated position...
			m_xMovementOffsetFinal	= m_xFinal * xLSMOFinalOrientation;

			CalculateWSSlerp(fParametricTimer, m_xMovementOffsetFinal, m_xMovementOffsetCurrent);
			CalculateWSSlerp(fParametricTimer, m_xFinal, m_xCurrent);

			m_xActual	= m_xCurrent * xLSMOCurrentOrientation;
			return(true);
		}

		case SUBTRACT_MOVEMENT_OFFSET:
		{
			Asura_Quat	xInverseLSMOFinalOrientation(xLSMOFinalOrientation);

			// Simply subtract the movement-offset onto the interpolated position...
			xInverseLSMOFinalOrientation.Invert();
			m_xMovementOffsetFinal	= m_xFinal * xInverseLSMOFinalOrientation;

			CalculateWSSlerp(fParametricTimer, m_xMovementOffsetFinal, m_xMovementOffsetCurrent);
			CalculateWSSlerp(fParametricTimer, m_xFinal, m_xCurrent);

			m_xActual	= m_xMovementOffsetCurrent * xLSMOCurrentOrientation;
			return(true);
		}

		case STRETCH_MOVEMENT_OFFSET:
		{
			Asura_AVector3		xCurrentAxis;
			Asura_AVector3		xFinalAxis;

			const Asura_Angle	fCurrentRotation	= xLSMOCurrentOrientation.ToRotation(xCurrentAxis);
			const Asura_Angle	fFinalRotation		= xLSMOFinalOrientation.ToRotation(xFinalAxis);
			const float			fAxisDotProduct		= xCurrentAxis * xFinalAxis;

			// Are we able to stretch?
			if( (Asura_Maths::Abs(fAxisDotProduct) < 0.999f) || (fFinalRotation == 0.0f) )
			{
				Asura_Quat	xInverseLSMOFinalOrientation(xLSMOFinalOrientation);

				// Simply subtract the movement-offset onto the interpolated position...
				xInverseLSMOFinalOrientation.Invert();
				m_xMovementOffsetFinal	= m_xFinal * xInverseLSMOFinalOrientation;

				CalculateWSSlerp(fParametricTimer, m_xMovementOffsetFinal, m_xMovementOffsetCurrent);
				CalculateWSSlerp(fParametricTimer, m_xFinal, m_xCurrent);

				m_xActual	= m_xMovementOffsetCurrent * xLSMOCurrentOrientation;
				return(true);
			}
			else
			{
				const float	fParametric	= fCurrentRotation / fFinalRotation;

				m_xMovementOffsetFinal		= m_xInitial * xLSMOFinalOrientation;
				m_xMovementOffsetCurrent	= m_xInitial * xLSMOCurrentOrientation;

				CalculateWSSlerp(fParametricTimer, m_xFinal, m_xCurrent);
				CalculateWSSlerp(Asura_Maths::Abs(fParametric), m_xFinal, m_xActual);
				return(true);
			}
		}
		
		default:
		{
			Asura_Assert(false, "Unexpected movement-offset operation when updating position-interpolation.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'Asura_MoveData_OrientationInterpolation::UpdateDurationBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::UpdateDurationBased
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_OrientationInterpolation::UpdateDurationBased(const Asura_Quat& xLSMOCurrentOrientation,
																  const Asura_Quat& xLSMOFinalOrientation)
{
	if(m_fTimer < m_fDuration)
	{
		const float	fParametricTimer	= m_fTimer / m_fDuration;

		return(UpdateDurationBased(fParametricTimer, xLSMOCurrentOrientation, xLSMOFinalOrientation));
	}
	else
	{
		// We've finished updating...
		if(m_bSnapToFinal)
		{
			UpdateDurationBased(1.0f, xLSMOCurrentOrientation, xLSMOFinalOrientation);
			return(false);
		}
		else
		{
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_OrientationInterpolation::UpdateDurationBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::Asura_MoveData_OrientationInterpolation
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

Asura_MoveData_OrientationInterpolation::Asura_MoveData_OrientationInterpolation(void)
: m_fReachedFinalEpsilon(1.0e-6f)
, m_fTimer(0.0f)
, m_fDuration(0.0f)
, m_fSpeed(0.0f)
, m_xSingularityHintAxis(Asura_Maths::ZeroAVector3)
, m_fSingularityAngle(0.0f)
, m_uFlags(0)
{
	m_bSnapToFinal	= true;
}

/// End of function 'Asura_MoveData_PositionInterpolation::Asura_MoveData_PositionInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::Reset
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_OrientationInterpolation::Reset(void)
{
	PARENT::Reset();
	ChangeState(WAITING);

	m_xInitial.Identity();
	m_xCurrent.Identity();
	m_xFinal.Identity();

	m_xMovementOffsetCurrent.Identity();
	m_xMovementOffsetFinal.Identity();

	m_xActual.Identity();

	m_fTimer			= 0.0f;
	m_fSingularityAngle	= 0.0f;

	m_bInitialSet				= false;
	m_bFinalSet					= false;
	m_bHasSingularityHintAxis	= false;
}

/// End of function 'Asura_MoveData_OrientationInterpolation::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::Start
// Author       : Andy Tate
// Date         : 12-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_OrientationInterpolation::Start(const Asura_Quat& xInitial, const Asura_Move& xOwnerMove)
{
	Asura_Quat		xInverseInitial(xInitial);

	Asura_Quat		xDifference;
	Asura_AVector3	xDifferenceAxis;
	Asura_Angle		fDifferenceAngle;

	SetInitial(xInitial);
	ChangeState(STARTING);

	// Time to check for a singularity...
	xInverseInitial.Invert();
	xDifference			= m_xFinal * xInverseInitial;
	fDifferenceAngle	= xDifference.ToRotation(xDifferenceAxis);

	if(Asura_Maths::Abs(fDifferenceAngle) >= dbg_fSingularityAngleThreshold)
	{
		// We have our singularity - do we have a hint to use?
		if(m_bHasSingularityHintAxis)
		{
			m_fSingularityAngle	= Asura_Maths::Abs(fDifferenceAngle);
		}
		else
		{
			Asura_Assertf(false, ("No hint axis provided for singularity in move 0x%08x - unexpected rotations may appear.", xOwnerMove.GetHash()));
		}
	}

	return(true);
}

/// End of function 'Asura_MoveData_OrientationInterpolation::Start'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::Update
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_OrientationInterpolation::Update(const float fTimeStep)
{
	if(m_bInitialSet)
	{
		if(m_bFinalSet)
		{
			// Increment the timer...
			m_fTimer	+= fTimeStep;

			// Is our move duration-based?
			if(m_bDurationBased)
			{
				return(UpdateDurationBased());
			}
			else
			{
				Asura_Assert(false, "Position-interpolation not fully initialised.");
				return(false);
			}
		}
		else
		{
			Asura_Assert(false, "Initial value is set but not the final value for move orientation-interpolation - cannot update.");
			return(false);
		}
	}
	else
	{
		if(m_bFinalSet)
		{
			Asura_Assert(false, "Final value is set but not the initial value for move orientation-interpolation - cannot update.");
			return(false);
		}
		else
		{
			Asura_Assert(false, "Neither the initial nor final value is set for move orientation-interpolation - cannot update.");
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_OrientationInterpolation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::Update
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_OrientationInterpolation::Update(const float fTimeStep,
													 const Asura_Quat& xLSMOCurrentOrientation,
													 const Asura_Quat& xLSMOFinalOrientation)
{
	if(m_bInitialSet)
	{
		if(m_bFinalSet)
		{
			// Increment the timer...
			m_fTimer	+= fTimeStep;

			// Is our move duration-based?
			if(m_bDurationBased)
			{
				return(UpdateDurationBased(xLSMOCurrentOrientation, xLSMOFinalOrientation));
			}
			else
			{
				Asura_Assert(false, "Position-interpolation not fully initialised.");
				return(false);
			}
		}
		else
		{
			Asura_Assert(false, "Initial value is set but not the final value for move orientation-interpolation - cannot update.");
			return(false);
		}
	}
	else
	{
		if(m_bFinalSet)
		{
			Asura_Assert(false, "Final value is set but not the initial value for move orientation-interpolation - cannot update.");
			return(false);
		}
		else
		{
			Asura_Assert(false, "Neither the initial nor final value is set for move orientation-interpolation - cannot update.");
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_OrientationInterpolation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_OrientationInterpolation::Finish
// Author       : Andy Tate
// Date         : 12-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_OrientationInterpolation::Finish(void)
{
	ChangeState(FINISHED);
}

/// End of function 'Asura_MoveData_OrientationInterpolation::Finish'. ///


#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)


/// End of file. ///
