
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Asura_Movement_System.cpp
//
// Created By Chris Wallace - 09-March-2007 (from Simp_Movement_System by Anthony Gregson / Richard May)
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////

// I N C L U D E   F I L E S /////////////////////////////////////////////////

// Main Asura header
#include "Core/Asura.h"

#if defined(ASURA_USE_MOVEMENT_SYSTEM)

// Local header
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"

// Asura
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_DebugText.h"
#include "Entity/PhysicalObject/Moves/Asura_Moves.h"
#include "Entity/PhysicalObject/Asura_ClientEntity_PhysicalObject.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_Player.h" 
#include "Physics/Asura_Physics_Object.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

////////////////////
// Asura_Msg_Move //
////////////////////

void Asura_Msg_Move::SetIgnoredByOwner(const bool bInput)
{
	if (bInput)
	{
		m_usAsuraFlags |= ASURA_MSG_MOVE_FLAG_IGNOREDBYOWNER;
	}
	else
	{
		m_usAsuraFlags &= ~ASURA_MSG_MOVE_FLAG_IGNOREDBYOWNER;
	}
}

void Asura_Msg_Move::SetForceAbort(const bool bInput)
{
	if (bInput)
	{
		m_usAsuraFlags |= ASURA_MSG_MOVE_START_FLAG_FORCEABORT;
	}
	else
	{
		m_usAsuraFlags &= ~ASURA_MSG_MOVE_START_FLAG_FORCEABORT;
	}
}


// Asura_Movement_System ////////////////////////////////////////////////////////////////////////
Asura_TemplateHashTree<Asura_Movement_System*> Asura_Movement_System::s_xDeferredUpdateTree;
bool Asura_Movement_System::s_bUseDeferredUpdate = false;
bool Asura_Movement_System::s_bRunningDeferredUpdate = false;

Asura_Movement_System::Asura_Movement_System( Asura_ClientEntity_PhysicalObject& xOwner )
: m_xOwner(xOwner)
, m_uOwnerGuid(xOwner.GetGuid())
, m_pxPendingMoveMessage(NULL)
, m_pxCurrentMove(NULL)
, m_xCurrentMovePosition(Asura_Maths::ZeroVector3)
, m_xCurrentMoveVelocity(Asura_Maths::ZeroVector3)
, m_xCurrentMoveOrientation(Asura_Maths::IdentityQuat)
, m_xPreviousMovePosition(Asura_Maths::ZeroAVector3)
, m_xPreviousMoveOrientation(Asura_Maths::IdentityQuat)
, m_xLSMOInitialPosition(Asura_Maths::ZeroAVector3)
, m_xLSMOInitialOrientation(Asura_Maths::IdentityQuat)
, m_xLSMOPreviousPosition(Asura_Maths::ZeroAVector3)
, m_xLSMOPreviousOrientation(Asura_Maths::IdentityQuat)
, m_xLSMOCurrentPosition(Asura_Maths::ZeroAVector3)
, m_xLSMOCurrentOrientation(Asura_Maths::IdentityQuat)
, m_xLSMOFinalPosition(Asura_Maths::ZeroAVector3)
, m_xLSMOFinalOrientation(Asura_Maths::IdentityQuat)
, m_fStationaryTimer(0.0f)
, m_uFirstMoveInChain(ASURA_HASH_ID_UNSET)
, m_uFlags(0)
{
}

// traversal function for destructor
void Traverse_DeleteAllMoves( Asura_Hash_ID uMoveHashID, Asura_Move** ppxMoveNode )
{
	// delete the move inside this node (nodes get freed anyway on destruction)
	if( ppxMoveNode && *ppxMoveNode )
	{
		delete *ppxMoveNode;
		*ppxMoveNode = NULL;
	}
}   


Asura_Movement_System::~Asura_Movement_System()
{
	// traverse the master move list and delete the moves
	m_xMasterMoveTree.Traverse( Traverse_DeleteAllMoves );

	if( m_pxPendingMoveMessage ) 
	{
		Asura_Network::DestroyMessage( m_pxPendingMoveMessage );
		m_pxPendingMoveMessage = NULL;
	}

	// Make sure we don't end up with a dead pointer in the update tree.
	s_xDeferredUpdateTree.DeleteNodeByHash(GetOwnerGuid());
}


////////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::Update
// Author        : Richard May
// Description   : Updates the current move or checks to see
//			       if a new move should start
////////////////////////////////////////////////////////////////////////

void Asura_Movement_System::Update( void )
{
	if (s_bUseDeferredUpdate && !s_bRunningDeferredUpdate)
	{
		// Just add ourselves to the tree awaiting update.
		s_xDeferredUpdateTree.AddNode(GetOwnerGuid(), this);
		return;
	}

	const Asura_Move* pxActiveMove = m_pxCurrentMove;

    // If the owner is suppressing moves then just clean up and then don't
    // allow new ones to start.
    if(m_xOwner.CanCurrentlyUseMoves())
	{
		m_bUpdatingCurrentMove = true;

		while(m_bUpdatingCurrentMove)
		{
			const Asura_Move*	pxOldMove	= m_pxCurrentMove;

			// If there's now a currently active move, update it	
			m_pxCurrentMove = m_pxCurrentMove ? m_pxCurrentMove->Update() : NULL;
			m_bUpdatingCurrentMove = false;

			// If, during the update above, we received a message then it cannot be handled immediately. 
			// It's therefore stored and handled here, instead.
			if( m_pxPendingMoveMessage )
			{
				HandleMessage( m_pxPendingMoveMessage );
				Asura_Network::DestroyMessage( m_pxPendingMoveMessage );
				m_pxPendingMoveMessage = NULL;
			}

			if( m_pxCurrentMove )
			{
				m_bUpdatingCurrentMove	= (pxOldMove != m_pxCurrentMove);
			}
			else 
			{
				// Check to see whether a move should start

				//if( !Asura_Game_Core::IsPaused() )
				{
					Asura_Move* pxMoveToStart = NULL;

					for( Asura_List_It<Asura_Move> xIt(&m_xInitialMoveList); !xIt.Done(); xIt.Next() )
					{
						Asura_Move* pxMove = xIt.GetCurrent();
						if( pxMove->ShouldStart() )
						{
							pxMoveToStart = pxMove;
							break;
						}
					}

					if( pxMoveToStart )
					{
						m_bUpdatingCurrentMove	= StartMove( *pxMoveToStart, NULL );
					}
				}
			}
		}
	}
	else
    {      
        if( m_pxCurrentMove )
        {
            m_pxCurrentMove->Abort();
            m_pxCurrentMove = NULL;
        }

        // If we've somehow stored up a move message, best to just disregard
        // it at this point.
        if( m_pxPendingMoveMessage )
        {
		    Asura_Network::DestroyMessage( m_pxPendingMoveMessage );
		    m_pxPendingMoveMessage = NULL;
        }
    }

	if (pxActiveMove && !m_pxCurrentMove)
	{
		// The last move in the chain has ended, send a message to tell people about this.
		SendChainCompleteMessage(pxActiveMove->GetHash());
	}
}

/// end of function Update ///

/////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::Render
// Author        : Richard May
// Description   : Renders debug info to the screen
/////////////////////////////////////////////////////////////////

void Asura_Movement_System::Render( void ) const
{
	if(m_pxCurrentMove != NULL)
	{
		m_pxCurrentMove->Render();
	}
}   

/// end of function Render ///

/////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::HandleMessage
// Author        : Richard May
// Description   : Returns the final end position of this animation
//                 from its movement offsets, etc.
/////////////////////////////////////////////////////////////////////

bool Asura_Movement_System::HandleMessage( const Asura_Message* pxMessage )
{
	switch( pxMessage->GetMsgID() )
	{
		case ASURA_MSG_MOVE_START:
		case ASURA_MSG_MOVE_ABORT:
		{
			// If this is the 'master' system and not one of the 'slaves' (if it belongs to a local player or an AI on the server)
            // ignore any messages sent from ourselves.
			// We only want to handle messages sent to us directly by the server, not ones bounced back to us from the server.
			{
				const Asura_Msg_Move* pxData = static_cast<const Asura_Msg_Move*>(pxMessage->GetData());
				if ( pxData && pxData->GetIgnoredByOwner() )
                {
					const Asura_Guid	uGuid	= pxMessage->GetFromGuid();
					
                    if ( IsControlledByLocalPlayer( uGuid ) )
                    {
                        return true;
                    }
                    else if ( !IsControlledByAnyPlayer( uGuid ) && Asura_Network::IsServer() )
                    {
                        return true;
                    }
                }
                
			}

			break;
		}

		default:
		{
			return(false); // not handled by the system, return false
		}
	}

	// If we're in the middle of updating the current move, we can't process messages.
	// Store the message and handle it later.
	if( m_bUpdatingCurrentMove )
	{
		Asura_Assert( !m_pxPendingMoveMessage, "Already have a move message pending. Currently pending message will be lost." );
		if( m_pxPendingMoveMessage )
		{
			Asura_Network::DestroyMessage( m_pxPendingMoveMessage );
			m_pxPendingMoveMessage = NULL;
		}

		m_pxPendingMoveMessage = Asura_Network::DuplicateMessage( pxMessage, false );
		return true;
	}
	else
	{	
		// safe to handle message
		switch( pxMessage->GetMsgID() )
		{
			case ASURA_MSG_MOVE_START:
			{
                // Ignore requests to start moves if our owner actor is
                // currently dictating that we can't.
                if( !m_xOwner.CanCurrentlyUseMoves() ) { return true; }

				const Asura_Msg_Move* pxData = static_cast<const Asura_Msg_Move*>(pxMessage->GetData());

				Asura_Move* pxMove = GetMove( pxData->GetMoveID() );

				if(pxMove)
				{
					// if there's a move in progress, try to abort that one first 
					if( m_pxCurrentMove )
					{
						if( pxData->GetForceAbort() || m_pxCurrentMove->CanBeAborted() )
						{
							bool bSuppressAbortMessage = m_pxCurrentMove->GetSuppressAbortMessage(); 
							m_pxCurrentMove->SetSuppressAbortMessage(true); 
							m_pxCurrentMove->Abort();
							m_pxCurrentMove->SetSuppressAbortMessage(bSuppressAbortMessage); 
						}
						else
						{
							// can't abort this move, ignore the message
							return true;
						}
					}	

					StartMove( *pxMove, pxData );
					return true;
				}
				else
				{
					Asura_Assertf( false, ("Could not start move from message: failed to find move 0x%08X.", pxData->GetMoveID() ));
					return false;
				}
			} 

			case ASURA_MSG_MOVE_ABORT:
			{
				if( m_pxCurrentMove )
				{
					Asura_Move* pxMove = m_pxCurrentMove;
					m_pxCurrentMove = NULL;

					const bool bSuppressAbortMessage = pxMove->GetSuppressAbortMessage();
					pxMove->SetSuppressAbortMessage(true); 
					pxMove->Abort();
					pxMove->SetSuppressAbortMessage(bSuppressAbortMessage); 
				}
				
				return true;
			} 

			default: 
			{
				break;
			}
		}
	}

	return false;
}

/// End of function 'HandleMessage' ///

//////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::Reset
// Author        : 
// Description   : Resets the movement system
//////////////////////////////////////////////////////////////////////

void Asura_Movement_System::Reset()
{
	if( m_pxCurrentMove != NULL )
	{
		// Don't allow anything to call this recursively.
		Asura_Move* pxMove = m_pxCurrentMove;
		m_pxCurrentMove = NULL;

		// We don't send an abort message out as we assume reset will be getting called on all machines automatically.
		const bool bSuppressAbortMessage = pxMove->GetSuppressAbortMessage();
		pxMove->SetSuppressAbortMessage(true); 
		pxMove->Abort();
		pxMove->SetSuppressAbortMessage(bSuppressAbortMessage); 
	}
}

/// end of function Reset ///

///////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::AddInitialMove
// Author        : Richard May
// Description   : Adds a move to the system's initial move list
///////////////////////////////////////////////////////////////////////

void Asura_Movement_System::AddInitialMove( Asura_Move* pxMove )
{
    Asura_Assert( pxMove, "" );

    // Callers should be ensuring these aren't added twice.
    Asura_Assert( !m_xInitialMoveList.FindObject(pxMove), "" );

    m_xInitialMoveList.AddToList( pxMove );
}

/// end of function AddInitialMove ///

//////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::AbortCurrentMove
// Author        : Andy McCabe
// Description   : Returns whether it aborted the current move
//////////////////////////////////////////////////////////////////////

bool Asura_Movement_System::AbortCurrentMove( bool bForceAbort/* = false*/, bool bSuppressAbortMessage/* = false*/ )
{
	if (m_pxCurrentMove && (bForceAbort || m_pxCurrentMove->CanBeAborted()) )
	{
		if (bSuppressAbortMessage)
		{
			bool bStoredSuppress = m_pxCurrentMove->GetSuppressAbortMessage();
			m_pxCurrentMove->SetSuppressAbortMessage(true);
			m_pxCurrentMove->Abort();
			m_pxCurrentMove->SetSuppressAbortMessage(bStoredSuppress);
		}
		else
		{
			m_pxCurrentMove->Abort();
		}

		// We need to send a chain complete message too, since this isn't called from the movement system update.
		SendChainCompleteMessage(m_pxCurrentMove->GetHash());

		m_pxCurrentMove = NULL;

		return true;
	}

	// Current move cannot be aborted
	return false;
}

/// end of function AbortCurrentMove ///

////////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::AddMove
// Author        : Richard May
// Description   : Adds the move to the master move tree
////////////////////////////////////////////////////////////////////////

void Asura_Movement_System::AddMove( Asura_Move* pxMove )
{
	if( !pxMove ) return;

	if( m_xMasterMoveTree.GetDataByHash( pxMove->m_uHash ) )
	{
		Asura_Assert(false, "Move is already registered with movement-system");
	}
	else
	{
		Asura_Move** ppxNewMove = m_xMasterMoveTree.AddNode( pxMove->m_uHash );
		if( ppxNewMove )
		{
			*ppxNewMove = pxMove;
		}
	}
}

/// end of function AddToMasterList ///

///////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::StartMove
// Author        : Richard May
// Description   : Starts the given move
///////////////////////////////////////////////////////////////////////////////

bool Asura_Movement_System::StartMove( Asura_Move& xMove, const Asura_Msg_Move* pxStartMsg )
{
	if(xMove.Start(pxStartMsg))
	{
		// KOR - filched from BFF. I assume it's correct.
		// if this isn't a manually/remotely started move, send a start message to other clients
		if( !pxStartMsg )
		{
			xMove.SendStartMsg();			
		}

		if (!m_pxCurrentMove)
		{
			// This is the first move from a new chain. Set the appropriate variable.
			m_uFirstMoveInChain = xMove.GetHash();
		}

		m_pxCurrentMove			= &xMove;
		m_xPreviousMovePosition	= m_xCurrentMovePosition;

		m_xPreviousMoveOrientation = m_xCurrentMoveOrientation;
		return(true);
	}
	else
	{
		xMove.Abort();
		m_pxCurrentMove	= NULL;
		return(false);	
	}
}

/// end of function StartMove ///

/////////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::GetMove
// Author        : Richard May
// Description   : Returns a pointer to the move with the given hash ID
/////////////////////////////////////////////////////////////////////////////////////

const Asura_Move* Asura_Movement_System::GetMove( const Asura_Hash_ID uHashID ) const
{
	Asura_Move** ppxMoveNode = m_xMasterMoveTree.GetDataByHash( uHashID );
	if( ppxMoveNode )
	{
		return *ppxMoveNode;
	}

	return NULL;
}


Asura_Move* Asura_Movement_System::GetMove( const Asura_Hash_ID uHashID )
{
	Asura_Move** ppxMoveNode = m_xMasterMoveTree.GetDataByHash( uHashID );
	if( ppxMoveNode )
	{
		return *ppxMoveNode;
	}

	return NULL;
}

/// end of function GetMove ///


/////////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Movement_System::DeferredUpdate
// Author        : Anthony Gregson
// Description   : Runs the update on all movement systems added to the tree, and then resets the tree ready for next frame.
/////////////////////////////////////////////////////////////////////////////////////

void Asura_Movement_System::DeferredUpdate()
{
	if (s_xDeferredUpdateTree.GetNumberOfNodes() <= 0) { return; }

	s_bRunningDeferredUpdate = true;

	Asura_TemplateHashTree_Iterator<Asura_Movement_System*> xIt(s_xDeferredUpdateTree);
	while (!xIt.Done())
	{
		Asura_Movement_System** ppxCurrent = xIt.GetCurrent();
		xIt.Next();

		if (ppxCurrent && *ppxCurrent)
		{
			(*ppxCurrent)->Update();
		}
	}
	s_xDeferredUpdateTree.Reset();

	s_bRunningDeferredUpdate = false;
}



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Movement_System::UpdateMoveVelocityFromPosition
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Movement_System::UpdateMoveVelocityFromPosition(const float fTimeStep,
														   const Asura_AVector3& xPosition,
														   const bool bWorldSpace)
{
	Asura_AVector3	xWorldSpacePosition;
	Asura_AVector3	xWorldSpaceVelocity;

	// Store the position (in world-space)...
	if(bWorldSpace)
	{
		xWorldSpacePosition	= xPosition;
	}
	else
	{
		xWorldSpacePosition = m_xCurrentMoveOrientation.RotateVector(xPosition);
		xWorldSpacePosition	+= Asura_AVector3(m_xCurrentMovePosition);
	}

	SetCurrentMovePosition(xWorldSpacePosition);

	// Once we've updated the position we can also update the velocity...
	if(fTimeStep > 0.0f)
	{
		xWorldSpaceVelocity	 = xWorldSpacePosition - m_xPreviousMovePosition;
		xWorldSpaceVelocity	/= fTimeStep;
	}
	else
	{
		xWorldSpaceVelocity.Zero();
	}

	// Update the velocity...
	SetCurrentMoveVelocity(xWorldSpaceVelocity);
}

/// End of function 'Asura_Movement_System::UpdateMoveVelocityFromPosition'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Movement_System::UpdateMovePositionFromVelocity
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Movement_System::UpdateMovePositionFromVelocity(const float fTimeStep,
														   const Asura_AVector3& xVelocity,
														   const bool bWorldSpace)
{
	Asura_AVector3	xWorldSpacePosition(m_xCurrentMovePosition);
	Asura_AVector3	xWorldSpaceVelocity;

	// Store the velocity (in world-space)...
	if(bWorldSpace)
	{
		xWorldSpaceVelocity	= xVelocity;
	}
	else
	{
		xWorldSpaceVelocity = m_xCurrentMoveOrientation.RotateVector(xVelocity);
	}

	SetCurrentMoveVelocity(xWorldSpaceVelocity);

	// Update the position...
	xWorldSpacePosition	+= (xWorldSpaceVelocity * fTimeStep);
	SetCurrentMovePosition(xWorldSpacePosition);
}

/// End of function 'Asura_Movement_System::UpdateMovePositionFromVelocity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Movement_System::UpdateMoveOrientationFromAngularVelocity
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Movement_System::UpdateMoveOrientationFromAngularVelocity(const float fTimerStep,
																	 const Asura_AVector3& xAxis,
																	 const Asura_Angle fAngle,
																	 const bool bWorldSpace)
{
	Asura_Quat	xCurrentOrientation(m_xCurrentMoveOrientation);

	Asura_Quat	xRotation;

	// Calculate the amount to rotate this frame (in local-space)...
	xRotation.FromRotation(fAngle * fTimerStep, xAxis);

	if(bWorldSpace)
	{
		Asura_Quat	xInverseCurrentOrientation(xCurrentOrientation);

		xInverseCurrentOrientation.Invert();
		xRotation	= xInverseCurrentOrientation * xRotation;
	}

	// Update the orientation accordingly...
	xCurrentOrientation	= xCurrentOrientation * xRotation;
	SetCurrentMoveOrientation(xCurrentOrientation);
}

/// End of function 'Asura_Movement_System::UpdateMoveOrientationFromAngularVelocity'. ///

void Asura_Movement_System::SendChainCompleteMessage(Asura_Hash_ID uFinalMoveID) const
{
	if (!m_bSendChainCompleteMessages) { return; }

	Asura_Message*	pxMessage	= Asura_Network::CreateMessage(ASURA_MSG_MOVECHAINCOMPLETE, sizeof(Asura_Msg_MoveChainComplete));

	if(pxMessage != NULL)
	{
		Asura_Msg_MoveChainComplete* pxData = static_cast<Asura_Msg_MoveChainComplete*>(pxMessage->GetData());

		if(pxData != NULL)
		{
			pxData->SetInitialMoveID(m_uFirstMoveInChain);
			pxData->SetFinalMoveID(uFinalMoveID);
		}

		pxMessage->SetToGuid(GetOwnerGuid());
		pxMessage->SetFromGuid(GetOwnerGuid());
		pxMessage->UseGuaranteedDelivery();

		Asura_Network::Send(pxMessage);
	}
}


#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)


/// End of file. ///
