
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Moves.cpp
// Author : Andy Tate
// Date   : 7-Sep-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(ASURA_USE_MOVEMENT_SYSTEM)

// This file's header...
#include "Entity/PhysicalObject/Moves/Asura_Moves.h"

// Asura Independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Animation/Asura_Animation_Posable.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#if !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Debugging/Asura_DebugLineRender.h"
#if defined(ASURA_USE_DEBUG_TEXT)
#include "Debugging/Asura_DebugText.h"
#endif	// defined(ASURA_USE_DEBUG_TEXT)
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/PhysicalObject/Asura_ClientEntity_PhysicalObject.h"
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_Player.h"
#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_Object.h"
#include "Shapes/Asura_Shapes.h"
#include "Sound/Asura_Sound_Event_System.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#define ASURA_MOVE_MAX_DEBUG_RENDER_GUIDS	2

// Debug variables...
#if defined(_DEBUG)
DEBUGVAR float	dbg_fMoveTimeScale	= -1.0f;
#endif	// defined(_DEBUG)

DEBUGVAR u_int	dbg_uNumSingularityAxisChecks			= 8;
DEBUGVAR float	dbg_fSingularityAxisMinAngleThreshold	= 0.001f;
DEBUGVAR float	dbg_fSingularityAxisMaxAngleThreshold	= 3.140f;

DEBUGVAR float	dbg_fMoveDebugRenderPositionAxisSize		= 0.375f;
DEBUGVAR float	dbg_fMoveDebugRenderPositionCrosshairSize	= 0.250f;
DEBUGVAR float	dbg_fMoveDebugRenderOrientationArrowSize	= 0.500f;

DEBUGVAR bool	dbg_bMoveDebugRenderAnimation					= true;	// false;
DEBUGVAR bool	dbg_bMoveDebugRenderOrientationInterpolation	= true;	// false;
DEBUGVAR bool	dbg_bMoveDebugRenderPositionInterpolation		= true;	// false;

DEBUGVAR Asura_Guid	dbg_auMoveDebugRenderGuids[ASURA_MOVE_MAX_DEBUG_RENDER_GUIDS] =
{
	ASURA_GUID_UNREGISTERED,
	ASURA_GUID_UNREGISTERED,
};

DEBUGVAR u_int dbg_auMoveDebugRenderDebugTextInitialColumns[ASURA_MOVE_MAX_DEBUG_RENDER_GUIDS] =
{
	5,
	10,
};

DEBUGVAR u_int dbg_auMoveDebugRenderDebugTextInitialRows[ASURA_MOVE_MAX_DEBUG_RENDER_GUIDS] =
{
	1,
	1,
};


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::StartAnimation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::StartAnimation()
{
	// Start the animation if necessary...
	if(m_pxAnimationData == NULL)
	{
		return(false);
	}
	else if(IsReadyToStart(*m_pxAnimationData))
	{
		Asura_ClientEntity_PhysicalObject& xOwner = m_xParentSystem.GetOwner();

		Asura_Shape_Instance&	xShapeInstance	= xOwner.GetShapeInstance();
		Asura_Animation*		pxAnimation		= xShapeInstance.GetAnimation();

		if(pxAnimation == NULL)
		{
			Asura_Assert(false, "Cannot find the animation for the owner of this move.");
			return(false);
		}
		else
		{
			// Initialise posable data (if we have any)...
			if(!m_bKeepPosing && pxAnimation->IsPosableAnim())
			{
				Asura_Animation_Posable*	pxPosable	= static_cast<Asura_Animation_Posable*>(pxAnimation);

				pxPosable->ResetAllOrientations();
			}

			// Time to create (and play) the animation...
			if(m_pxAnimationData->Start(m_bLoopAtEnd, *pxAnimation))
			{
				Asura_Quat		xLSMOInitialOrientation;
				Asura_AVector3	xLSMOInitialPosition;

				Asura_Quat		xLSMOFinalOrientation;
				Asura_AVector3	xLSMOFinalPosition;

				// Decide whether to calculate the movement-offset data...
				if(!m_pxAnimationData->CalculateMovementOffset(0.0f, &xLSMOInitialPosition, &xLSMOInitialOrientation))
				{
					xLSMOInitialOrientation.Identity();
					xLSMOInitialPosition.Zero();
				}

				if(!m_pxAnimationData->CalculateMovementOffset(1.0f, &xLSMOFinalPosition, &xLSMOFinalOrientation))
				{
					xLSMOFinalOrientation.Identity();
					xLSMOFinalPosition.Zero();
				}

				m_xParentSystem.SetLSMOInitialOrientation(xLSMOInitialOrientation);
				m_xParentSystem.SetLSMOInitialPosition(xLSMOInitialPosition);

				m_xParentSystem.SetLSMOFinalOrientation(xLSMOFinalOrientation);
				m_xParentSystem.SetLSMOFinalPosition(xLSMOFinalPosition);
				return(true);
			}
			else
			{
				Abort();
				return(false);
			}
		}
	}
	else
	{
		return(false);
	}
}

/// End of function 'Asura_Move::StartAnimation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::UpdateAnimation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::UpdateAnimation(const float fTimeStep)
{
	// We're at our destination, have we done everything we need to?
	if(m_pxAnimationData == NULL)
	{
		return(false);
	}
	else
	{
		const Asura_MoveData::State	eCurrentState	= m_pxAnimationData->GetState();

		switch(eCurrentState)
		{
			case Asura_MoveData::WAITING:
			{
				return(true);
			}

			case Asura_MoveData::STARTING:
			case Asura_MoveData::UPDATING:
			{
				if(m_pxAnimationData->Update(fTimeStep))
				{
					Asura_AVector3	xLSMOPosition;
					Asura_Quat		xLSMOOrientation;

					Asura_AVector3*	pxLSMOPosition		= ((m_pxPositionInterpolationData == NULL) || (m_pxPositionInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_PositionInterpolation::IGNORE_MOVEMENT_OFFSET))
														? NULL
														: &xLSMOPosition;
					Asura_Quat*		pxLSMOOrientation	= ((m_pxOrientationInterpolationData == NULL) || (m_pxOrientationInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_OrientationInterpolation::IGNORE_MOVEMENT_OFFSET))
														? NULL
														: &xLSMOOrientation;

					// Decide whether to calculate the movement-offset data...
					if((pxLSMOOrientation != NULL) || (pxLSMOPosition != NULL))
					{
						const float	fParametricTime	= m_pxAnimationData->GetParametricTimer();

						if(m_pxAnimationData->CalculateMovementOffset(fParametricTime, pxLSMOPosition, pxLSMOOrientation))
						{
							if(pxLSMOOrientation != NULL)
							{
								m_xParentSystem.SetLSMOCurrentOrientation(xLSMOOrientation);
							}

							if(pxLSMOPosition != NULL)
							{
								m_xParentSystem.SetLSMOCurrentPosition(xLSMOPosition);
							}
						}
					}

					return(true);
				}
				else
				{
					return(false);
				}
			}

			case Asura_MoveData::FINISHED:
			{
				return(false);
			}

			default:
			{
				Asura_Assert(false, "Unexpected state for the animation in this move.");
				return(false);
			}
		}		// End of switch.
	}
}

/// End of function 'Asura_Move::UpdateAnimation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::FinishAnimation
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::FinishAnimation()
{
	if(m_pxAnimationData != NULL)
	{
		Asura_AVector3	xLSMOPosition;
		Asura_Quat		xLSMOOrientation;

		Asura_AVector3*	pxLSMOPosition		= ((m_pxPositionInterpolationData == NULL) || (m_pxPositionInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_PositionInterpolation::IGNORE_MOVEMENT_OFFSET))
											? NULL
											: &xLSMOPosition;
		Asura_Quat*		pxLSMOOrientation	= ((m_pxOrientationInterpolationData == NULL) || (m_pxOrientationInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_OrientationInterpolation::IGNORE_MOVEMENT_OFFSET))
											? NULL
											: &xLSMOOrientation;

		m_pxAnimationData->Finish();

		if(pxLSMOOrientation != NULL)
		{
			m_xParentSystem.GetLSMOFinalOrientation(xLSMOOrientation);
			m_xParentSystem.SetLSMOCurrentOrientation(xLSMOOrientation);
		}

		if(pxLSMOPosition != NULL)
		{
			m_xParentSystem.GetLSMOFinalPosition(xLSMOPosition);
			m_xParentSystem.SetLSMOCurrentPosition(xLSMOPosition);
		}
	}
}

/// End of function 'Asura_Move::FinishAnimation'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::ResetAnimation
// Author       : Anthony Gregson
// Date         : 17/03/08
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::ResetAnimation()
{
	if(m_pxAnimationData != NULL)
	{
		m_pxAnimationData->Reset();
		m_pxAnimationData = NULL;
	}
}

/// End of function 'Asura_Move::ResetAnimation'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::StartOrientationInterpolation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::StartOrientationInterpolation()
{
	if(m_pxOrientationInterpolationData == NULL)
	{
		return(false);
	}
	else if(IsReadyToStart(*m_pxOrientationInterpolationData))
	{
		Asura_Quat	xCurrentOrientation;
		
		m_xParentSystem.GetCurrentMoveOrientation(xCurrentOrientation);
		return(m_pxOrientationInterpolationData->Start(xCurrentOrientation, *this));
	}
	else
	{
		return(false);
	}
}

/// End of function 'Asura_Move::StartOrientationInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::UpdateOrientationInterpolation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::UpdateOrientationInterpolation(const float fTimeStep)
{
	if(m_pxOrientationInterpolationData == NULL)
	{
		if(m_pxAnimationData != NULL)
		{
			Asura_AVector3	xAxis;
			Asura_Angle		fAngle;

			if(m_pxAnimationData->CalculateAngularVelocity(xAxis, fAngle))
			{
				m_xParentSystem.UpdateMoveOrientationFromAngularVelocity(fTimeStep, xAxis, fAngle, false);
			}
		}

		return(false);
	}
	else
	{
		const Asura_MoveData::State	eCurrentState	= m_pxOrientationInterpolationData->GetState();

		switch(eCurrentState)
		{
			case Asura_MoveData::WAITING:
			{
				return(true);
				
			}

			case Asura_MoveData::STARTING:
			case Asura_MoveData::UPDATING:
			case Asura_MoveData::FINISHED:
			{
				Asura_Quat	xCurrentOrientation;

				const bool	bUpdating	= CalculateOrientation(fTimeStep, xCurrentOrientation);

				m_xParentSystem.SetCurrentMoveOrientation(xCurrentOrientation);
				return(bUpdating);
			}

			default:
			{
				Asura_Assert(false, "Unexpected state for the orientation-interpolation in this move.");
				return(false);
			}
		}		// End of switch.
	}
}

/// End of function 'Asura_Move::UpdateOrientationInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::FinishOrientationInterpolation
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::FinishOrientationInterpolation()
{
	if(m_pxOrientationInterpolationData != NULL)
	{
		m_pxOrientationInterpolationData->Finish();
	}
}

/// End of function 'Asura_Move::FinishOrientationInterpolation'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::ResetOrientationInterpolation
// Author       : Anthony Gregson
// Date         : 17/03/08
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::ResetOrientationInterpolation()
{
	if(m_pxOrientationInterpolationData != NULL)
	{
		SetOrientationUsage(ORIENTATION_NOT_USED);
		m_pxOrientationInterpolationData->Reset();
		m_pxOrientationInterpolationData = NULL;
	}
}

/// End of function 'Asura_Move::ResetOrientationInterpolation'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::StartPositionInterpolation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::StartPositionInterpolation()
{
	if(m_pxPositionInterpolationData == NULL)
	{
		return(false);
	}
	else if(IsReadyToStart(*m_pxPositionInterpolationData))
	{
		Asura_AVector3	xCurrentPosition;

		m_xParentSystem.GetCurrentMovePosition(xCurrentPosition);
		return(m_pxPositionInterpolationData->Start(xCurrentPosition));
	}
	else
	{
		return(false);
	}
}

/// End of function 'Asura_Move::StartPositionInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::UpdatePositionInterpolation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::UpdatePositionInterpolation(const float fTimeStep)
{
	if(m_pxPositionInterpolationData == NULL)
	{
		// If we have no position interpolation, just calculate a velocity. Still use the CalculateVelocity function though, so moves can override it.
		Asura_AVector3 xCurrentVelocity;

		const bool bUpdating = CalculateVelocity(fTimeStep, xCurrentVelocity);

		m_xParentSystem.UpdateMovePositionFromVelocity(fTimeStep, xCurrentVelocity, true);
		return(bUpdating);
	}
	else
	{
		const Asura_MoveData::State	eCurrentState	= m_pxPositionInterpolationData->GetState();

		switch(eCurrentState)
		{
			case Asura_MoveData::WAITING:
			{
				return(true);
			}

			case Asura_MoveData::STARTING:
			case Asura_MoveData::UPDATING:
			case Asura_MoveData::FINISHED:
			{
				if(GetPositionUsage() == POSITION_SET_FROM_VELOCITY)
				{
					Asura_AVector3	xCurrentVelocity;

					const bool		bUpdating	= CalculateVelocity(fTimeStep, xCurrentVelocity);

					m_xParentSystem.UpdateMovePositionFromVelocity(fTimeStep, xCurrentVelocity, true);
					return(bUpdating);
				}
				else
				{
					Asura_AVector3	xCurrentPosition;

					const bool		bUpdating	= CalculatePosition(fTimeStep, xCurrentPosition);

					m_xParentSystem.UpdateMoveVelocityFromPosition(fTimeStep, xCurrentPosition, true);
					return(bUpdating);
				}
			}

			default:
			{
				Asura_Assert(false, "Unexpected state for the position-interpolation in this move.");
				return(false);
			}
		}		// End of switch.
	}
}

/// End of function 'Asura_Move::UpdatePositionInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::FinishPositionInterpolation
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::FinishPositionInterpolation()
{
	if((m_pxPositionInterpolationData != NULL) && m_pxPositionInterpolationData->IsUpdating())
	{
		m_pxPositionInterpolationData->Finish();
	}
}

/// End of function 'Asura_Move::FinishPositionInterpolation'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::ResetPositionInterpolation
// Author       : Anthony Gregson
// Date         : 17/03/08
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::ResetPositionInterpolation()
{
	if(m_pxPositionInterpolationData != NULL)
	{
		SetPositionUsage(POSITION_NOT_USED);
		m_pxPositionInterpolationData->Reset();
		m_pxPositionInterpolationData = NULL;
	}
}

/// End of function 'Asura_Move::ResetPositionInterpolation'. ///


void Asura_Move::SetParentSystemLSMOFinalOrientation( const Asura_Quat& xLSMOFinalOri )
{
	m_xParentSystem.SetLSMOFinalOrientation( xLSMOFinalOri );
}

void Asura_Move::SetParentSystemLSMOFinalPosition( const Asura_AVector3& xLSMOFinalPos )
{
	m_xParentSystem.SetLSMOFinalPosition( xLSMOFinalPos );
}

////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::InitialiseAnimation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::InitialiseAnimation(Asura_MoveData_Animation& xAnimationData,
									 Asura_AVector3* pxSingularityAxis)
{
	// Initialise the data...
	m_pxAnimationData	= &xAnimationData;
	xAnimationData.SetParentMove(this);

	// Do we need to calculate a hint-axis for singularities?
	if(pxSingularityAxis != NULL)
	{
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= xAnimationData.GetHierarchyAnim();

		if(pxHierarchyAnim == NULL)
		{
			Asura_Assert(false, "Cannot find animation to calculate hint-axis for.");
		}
		else
		{
			const int	iMovementBoneIndex		= pxHierarchyAnim->GetMovementBoneIndex();
			float		fParamtericDifference	= 0.500f;
			float		fParametricTime			= 1.000f;

			Asura_Angle	fAngle;
			Asura_Quat	xInitialOrientation;
			Asura_Quat	xInverseInitialOrientation;

			// Get the initial orientation...
			pxHierarchyAnim->GetRelativeBonePosition(iMovementBoneIndex, 0.000f, NULL, &xInitialOrientation);

			xInverseInitialOrientation	= xInitialOrientation;
			xInverseInitialOrientation.Invert();

			// Time to do a binary-chop...
			for(u_int u = 0; u < dbg_uNumSingularityAxisChecks; u++, fParamtericDifference *= 0.500f)
			{
				Asura_Quat	xCurrentOrientation;
				Asura_Quat	xOrientationDifference;

				// Check the orientation at this point in time...
				pxHierarchyAnim->GetRelativeBonePosition(iMovementBoneIndex, fParametricTime, NULL, &xCurrentOrientation);

				xOrientationDifference	= xCurrentOrientation * xInverseInitialOrientation;
				fAngle					= xOrientationDifference.ToRotation(*pxSingularityAxis);

				if(Asura_Maths::Abs(fAngle) <= dbg_fSingularityAxisMinAngleThreshold)
				{
					fParametricTime	+= fParamtericDifference;
				}
				else if(Asura_Maths::Abs(fAngle) >= dbg_fSingularityAxisMaxAngleThreshold)
				{
					fParametricTime	-= fParamtericDifference;
				}
				else
				{
					// We've found our axis...
					return;
				}
			}

			// No axis found...
			Asura_Assert(false, "Cannot calculate a singularity-axis for this animation - initialising to zero.");
			pxSingularityAxis->Zero();
		}
	}
}

/// End of function 'Asura_Move::InitialiseAnimation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::InitialiseOrientationInterpolation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::InitialiseOrientationInterpolation(Asura_MoveData_OrientationInterpolation& xOrientationInterpolationData)
{
	// Initialise the data...
	m_pxOrientationInterpolationData	= &xOrientationInterpolationData;
	xOrientationInterpolationData.SetParentMove(this);

	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

/// End of function 'Asura_Move::InitialiseOrientationInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::InitialisePositionInterpolation
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::InitialisePositionInterpolation(Asura_MoveData_PositionInterpolation& xPositionInterpolationData)
{
	// Initialise the data...
	m_pxPositionInterpolationData	= &xPositionInterpolationData;
	xPositionInterpolationData.SetParentMove(this);

	SetPositionUsage(POSITION_SET_DIRECTLY);
}

/// End of function 'Asura_Move::InitialisePositionInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::CheckStatesToStart
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::CheckStatesToStart()
{
	// Decide whether its time to start the appropriate parts of the move...
	StartAnimation();
	StartOrientationInterpolation();
	StartPositionInterpolation();
}

/// End of function 'Asura_Move::CheckStatesToStart'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::UpdateStates
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::UpdateStates(const float fTimeStep)
{
	////////////////////////////////////////////////////////////////////////////////
	// NOTE: Order is important here - animation, orientation and finally position...
	////////////////////////////////////////////////////////////////////////////////
	if(!UpdateAnimation(fTimeStep))
	{
		FinishAnimation();
	}

	if(!UpdateOrientationInterpolation(fTimeStep))
	{
		FinishOrientationInterpolation();
	}

	if(!UpdatePositionInterpolation(fTimeStep))
	{
		FinishPositionInterpolation();
	}
}

/// End of function 'Asura_Move::UpdateStates'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::Reset
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::Reset()
{
	// Reset the remaining data...
	ResetAnimation();

	ResetOrientationInterpolation();

	ResetPositionInterpolation();
}

/// End of function 'Asura_Move::Reset'. ///



//////////////////////////////////////////////////////////////////
// Name          : Asura_Move::Start
// Author        : Richard May
// Description   : Base-class start function
//////////////////////////////////////////////////////////////////

bool Asura_Move::Start(const Asura_Msg_Move* pxStartMsg)
{
	Asura_ClientEntity_PhysicalObject& xOwner = m_xParentSystem.GetOwner();

	// Reset the local-space movement-offset data needed...
	m_xParentSystem.SetLSMOInitialOrientation(Asura_Maths::IdentityQuat);
	m_xParentSystem.SetLSMOFinalOrientation(Asura_Maths::IdentityQuat);

	m_xParentSystem.SetLSMOInitialPosition(Asura_Maths::ZeroAVector3);
	m_xParentSystem.SetLSMOFinalPosition(Asura_Maths::ZeroAVector3);

	// Initialise the data...
	m_pxPrevMove	= m_xParentSystem.GetCurrentMove();

	// Do we need to initialised our data from the entity?
	if(m_pxPrevMove) // If the previous move wants us to ignore it's pos and ori, get them from the entity
	{
		if(m_pxPrevMove->ShouldNextMoveIgnoreMovePos())
		{
			const Asura_Vector_3&	xCurrentPosition	= xOwner.GetCurrentPosition();
			Asura_Vector_3			xCurrentVelocity;
			xOwner.GetCurrentVelocity(xCurrentVelocity);

			m_xParentSystem.SetCurrentMovePosition(xCurrentPosition);
			m_xParentSystem.SetCurrentMoveVelocity(xCurrentVelocity);
		}

		if(m_pxPrevMove->ShouldNextMoveIgnoreMoveOri())
		{
			const Asura_Matrix_3x3&	xCurrentOrientation	= xOwner.GetCurrentOrientation();

			m_xParentSystem.SetCurrentMoveOrientation(xCurrentOrientation);
		}
	}
	else // If there is no move, get details from entity
	{
		const Asura_Matrix_3x3&	xCurrentOrientation	= xOwner.GetCurrentOrientation();
		const Asura_Vector_3&	xCurrentPosition	= xOwner.GetCurrentPosition();

		Asura_Vector_3			xCurrentVelocity;

		// Initialise the positional-data...
		xOwner.GetCurrentVelocity(xCurrentVelocity);

		m_xParentSystem.SetCurrentMoveOrientation(xCurrentOrientation);
		m_xParentSystem.SetCurrentMovePosition(xCurrentPosition);
		m_xParentSystem.SetCurrentMoveVelocity(xCurrentVelocity);
	}

	return(true);
}

/// End of function "Asura_Move::Start". ///



///////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::SendStartMsg
// Author        : Richard May
// Description   : Basic function to send a message to our server
//                 entity. This should then forward the message to other
//                 client's movement systems.
///////////////////////////////////////////////////////////////////////////////////

void Asura_Move::SendStartMsg()
{
	if(!m_bSuppressStartMessage)
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(Asura_Msg_Move));

		if(pxMessage != NULL)
		{
			Asura_Msg_Move*	pxData		= static_cast<Asura_Msg_Move*>(pxMessage->GetData());
			const Asura_Guid		uOwnerGuid	= m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(true);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}
}

/// End of function "Asura_Move::SendStartMsg". ///



//////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::Update
// Author        : Richard May
// Description   : Updates the move. Returns a pointer to 
//                 itself if the move hasn't finished. If it has,
//                 returns the next move or NULL
//////////////////////////////////////////////////////////////////////////////

Asura_Move* Asura_Move::Update()
{
	const bool	bIsAnimationLooped	= (m_pxAnimationData != NULL) && m_pxAnimationData->IsLooped();
	const float	fTimeStep			= Asura_Timers::GetGameTimeSlice();
	const bool	bWasFinished		= HasFinished();

	bool		bHasFinished;

#if defined(_DEBUG)
	// Useful debug code - allows the time-scale to be altered during a move...
	if(dbg_fMoveTimeScale >= 0.0f)
	{
		Asura_Timers::SetGameTimeScale(dbg_fMoveTimeScale);
	}
#endif	// defined(_DEBUG)

	// Update the time this move has taken so far...
	CheckStatesToStart();
	UpdateStates(fTimeStep);

	bHasFinished	= HasFinished();

	// Do we need to check whether to finish this move?
	if(bIsAnimationLooped || m_bAlwaysCheckNextMoves || bHasFinished)
	{
		Asura_Move*	pxNextMove	= GetNextMove();

		// Have we only just finished this move?
		if(!bWasFinished && bHasFinished)
		{
			Finish();
		}

		// Check to see if any of the next moves should start...
		if(pxNextMove == NULL)
		{
			if(m_bLoopAtEnd || (m_bAlwaysCheckNextMoves && !bHasFinished))
			{
				// looping moves return
				return(this);
			}
			else
			{
				// finished moves return NULL
				return(NULL);
			}
		}
		else
		{
			if(m_bLoopAtEnd && bHasFinished)
			{
				// if loop at end is set, and we've already finished, 
				// we need to abort the move before starting the next
				Abort();
			}
			else if(m_bAlwaysCheckNextMoves)
			{
				// We need to ensure that we finish what we're doing...
				Finish();
			}

			// Try to start the next move...
			if(pxNextMove->Start(NULL))
			{				
				pxNextMove->SendStartMsg();
				return(pxNextMove);
			}
			else
			{
				Asura_Assert(false, "Failed to start new move.");
				return(NULL);
			}
		}
	}
	else
	{
		// Time to check whether we've moved at enough this frame...
		if(m_fStationaryTimer > 0.0f)
		{
			Asura_AVector3	xPreviousPosition;
			Asura_AVector3	xCurrentPosition;
			Asura_AVector3	xCurrentVelocity;

			m_xParentSystem.GetCurrentMoveVelocity(xCurrentVelocity);
			m_xParentSystem.GetCurrentMovePosition(xCurrentPosition);
			m_xParentSystem.GetPreviousMovePosition(xPreviousPosition);

			if(xCurrentPosition.ApproximatelyEquals(xPreviousPosition, m_fStationaryEpsilon) && !xCurrentVelocity.ApproximatelyEquals(Asura_Maths::ZeroAVector3))
			{
				if(m_xParentSystem.UpdateStationaryTimer(fTimeStep, m_fStationaryTimer))
				{
					return(this);
				}
				else
				{
					Abort();
					return(NULL);
				}
			}
			else
			{
				m_xParentSystem.ResetStationaryTimer();
				return(this);
			}
		}
		else
		{
			return(this);
		}
	}
}

/// End of function "Asura_Move::Update". ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::Render
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Move::Render() const
{
#if !defined(ASURA_NO_DEBUG_OUTPUT)
	Asura_ClientEntity_PhysicalObject& xOwner = m_xParentSystem.GetOwner();

	const Asura_Guid	uOwnerGuid				= m_xParentSystem.GetOwnerGuid();
	int					iDebugGuidToRenderIndex	= -1;

	for(int i = 0; i < ASURA_MOVE_MAX_DEBUG_RENDER_GUIDS; i++)
	{
		if(uOwnerGuid == dbg_auMoveDebugRenderGuids[i])
		{
			iDebugGuidToRenderIndex	= i;
		}
	}

	if(iDebugGuidToRenderIndex >= 0)
	{
#if defined(ASURA_USE_DEBUG_TEXT)
		u_int				uRow	= dbg_auMoveDebugRenderDebugTextInitialRows[iDebugGuidToRenderIndex];
		u_int				uColumn	= dbg_auMoveDebugRenderDebugTextInitialColumns[iDebugGuidToRenderIndex];
#endif	// defined(ASURA_USE_DEBUG_TEXT)

		Asura_Vector_3		xWSInitialPosition;
		Asura_Vector_3		xWSCurrentPosition;
		Asura_Vector_3		xWSFinalPosition;
		Asura_Vector_3		xWSMOCurrentPosition;
		Asura_Vector_3		xWSMOFinalPosition;
		Asura_Vector_3		xWSActualPosition;

		Asura_Matrix_3x3	xWSInitialOrientation;
		Asura_Matrix_3x3	xWSCurrentOrientation;
		Asura_Matrix_3x3	xWSFinalOrientation;
		Asura_Matrix_3x3	xWSMOCurrentOrientation;
		Asura_Matrix_3x3	xWSMOFinalOrientation;
		Asura_Matrix_3x3	xWSActualOrientation;

#if defined(ASURA_USE_DEBUG_TEXT)
		// Display the name of this move...
		Asura_DebugText::AddTextBlock(uRow, uColumn, 0xFFFFFFFF, 0, "GUID: %d, move: \"%s\"", m_xParentSystem.GetOwnerGuid(), GetName());
		uColumn++;
#endif	// defined(ASURA_USE_DEBUG_TEXT)

		// Debug render the animation data...
		if(dbg_bMoveDebugRenderAnimation && (m_pxAnimationData != NULL))
		{
#if defined(ASURA_USE_DEBUG_TEXT)
			const Asura_AnimationController_Animation*	pxAnimationController	= m_pxAnimationData->GetAnimationController();

			if(pxAnimationController != NULL)
			{
				const Asura_MoveData::State	eState				= m_pxAnimationData->GetState();
				const float					fDuration			= pxAnimationController->GetDuration();
				const float					fParametricTimer	= pxAnimationController->GetParametricTimer();
				const float 				fTimer				= fParametricTimer * fDuration;

				Asura_DebugText::AddTextBlock(uRow, uColumn++, 0x80FFFFFF, 0, "-> Anm: State = %d, Timer = %.3f/%.3fs (%.3f)", eState, fTimer, fDuration, fParametricTimer);
			}
#endif	// defined(ASURA_USE_DEBUG_TEXT)
		}

		// Initialise position-interpolation data...
		if(m_pxPositionInterpolationData == NULL)
		{
			xWSInitialPosition	= xOwner.GetCurrentPosition();
			xWSCurrentPosition	= xWSInitialPosition;
			xWSFinalPosition	= xWSInitialPosition;

			xWSMOCurrentPosition	= xWSInitialPosition;
			xWSMOFinalPosition		= xWSInitialPosition;

			xWSActualPosition	= xWSInitialPosition;
		}
		else
		{
			xWSInitialPosition	= m_pxPositionInterpolationData->GetInitial();
			xWSCurrentPosition	= m_pxPositionInterpolationData->GetCurrent();
			xWSFinalPosition	= m_pxPositionInterpolationData->GetFinal();

			xWSMOCurrentPosition	= m_pxPositionInterpolationData->GetMovementOffsetCurrent();
			xWSMOFinalPosition		= m_pxPositionInterpolationData->GetMovementOffsetFinal();

			xWSActualPosition	= m_pxPositionInterpolationData->GetActual();
		}

		// Initialise orientation-interpolation data...
		if(m_pxOrientationInterpolationData == NULL)
		{
			xWSInitialOrientation	= xOwner.GetCurrentOrientation();
			xWSCurrentOrientation	= xWSInitialOrientation;
			xWSFinalOrientation		= xWSInitialOrientation;

			xWSMOCurrentOrientation	= xWSInitialOrientation;
			xWSMOFinalOrientation	= xWSInitialOrientation;

			xWSActualOrientation	= xWSInitialOrientation;
		}
		else
		{
			m_pxOrientationInterpolationData->GetInitial().ToMatrix(xWSInitialOrientation);
			m_pxOrientationInterpolationData->GetCurrent().ToMatrix(xWSCurrentOrientation);
			m_pxOrientationInterpolationData->GetFinal().ToMatrix(xWSFinalOrientation);

			m_pxOrientationInterpolationData->GetMovementOffsetCurrent().ToMatrix(xWSMOCurrentOrientation);
			m_pxOrientationInterpolationData->GetMovementOffsetFinal().ToMatrix(xWSMOFinalOrientation);

			m_pxOrientationInterpolationData->GetActual().ToMatrix(xWSActualOrientation);
		}

		// Debug render the position-interpolation data...
		if(dbg_bMoveDebugRenderPositionInterpolation && (m_pxPositionInterpolationData != NULL))
		{
#if defined(ASURA_USE_DEBUG_TEXT)
			const Asura_MoveData::State	eState	= m_pxPositionInterpolationData->GetState();
#endif	// defined(ASURA_USE_DEBUG_TEXT)

			Asura_DebugLineRender::AddArrow(xWSInitialPosition, xWSFinalPosition, 0x40FFFF00);
			Asura_DebugLineRender::AddAxis(xWSInitialPosition, xWSInitialOrientation, dbg_fMoveDebugRenderPositionAxisSize, 0.250f);
			Asura_DebugLineRender::AddCross(xWSInitialPosition, 0x80FF0000, dbg_fMoveDebugRenderPositionCrosshairSize);
			Asura_DebugLineRender::AddAxis(xWSCurrentPosition, xWSCurrentOrientation, dbg_fMoveDebugRenderPositionAxisSize, 0.500f);
			Asura_DebugLineRender::AddCross(xWSCurrentPosition, 0x80FFFF00, dbg_fMoveDebugRenderPositionCrosshairSize);
			Asura_DebugLineRender::AddAxis(xWSFinalPosition, xWSFinalOrientation, dbg_fMoveDebugRenderPositionAxisSize, 0.250f);
			Asura_DebugLineRender::AddCross(xWSFinalPosition, 0x8000FF00, dbg_fMoveDebugRenderPositionCrosshairSize);

			if(m_pxPositionInterpolationData->GetMovementOffsetOperation() != Asura_MoveData_PositionInterpolation::IGNORE_MOVEMENT_OFFSET)
			{
				Asura_DebugLineRender::AddArrow(xWSInitialPosition, xWSMOFinalPosition, 0x40FF00FF);
				Asura_DebugLineRender::AddAxis(xWSMOCurrentPosition, xWSMOCurrentOrientation, dbg_fMoveDebugRenderPositionAxisSize, 0.500f);
				Asura_DebugLineRender::AddCross(xWSMOCurrentPosition, 0x80FF00FF, dbg_fMoveDebugRenderPositionCrosshairSize);
				Asura_DebugLineRender::AddAxis(xWSMOFinalPosition, xWSMOFinalOrientation, dbg_fMoveDebugRenderPositionAxisSize, 0.250f);
				Asura_DebugLineRender::AddCross(xWSMOFinalPosition, 0x800000FF, dbg_fMoveDebugRenderPositionCrosshairSize);
			}

#if defined(ASURA_USE_DEBUG_TEXT)
			if(m_pxPositionInterpolationData->IsDurationBased())
			{
				const float	fTimer				= m_pxPositionInterpolationData->GetTimer();
				const float	fDuration			= m_pxPositionInterpolationData->GetDuration();
				const float	fParametricTimer	= m_pxPositionInterpolationData->GetParametricTimer();

				Asura_DebugText::AddTextBlock(uRow, uColumn++, 0x80FFFFFF, 0, "-> Pos: State = %d, Timer = %.3f/%.3fs (%.3f)", eState, fTimer, fDuration, fParametricTimer);
			}
			else if(m_pxPositionInterpolationData->IsSpeedBased())
			{
				const float	fSpeed	= m_pxPositionInterpolationData->GetSpeed();

				Asura_DebugText::AddTextBlock(uRow, uColumn++, 0x80FFFFFF, 0, "-> Pos: State = %d, Speed = %.3f", eState, fSpeed);
			}
#endif	// defined(ASURA_USE_DEBUG_TEXT)
		}

		// Debug render the orientation-interpolation data...
		if(dbg_bMoveDebugRenderOrientationInterpolation && (m_pxOrientationInterpolationData != NULL))
		{
#if defined(ASURA_USE_DEBUG_TEXT)
			const Asura_MoveData::State	eState	= m_pxOrientationInterpolationData->GetState();
#endif	// defined(ASURA_USE_DEBUG_TEXT)
			const Asura_Vector_3	xWSInitialDirection(xWSInitialOrientation.e13, xWSInitialOrientation.e23, xWSInitialOrientation.e33);
			const Asura_Vector_3	xWSCurrentDirection(xWSCurrentOrientation.e13, xWSCurrentOrientation.e23, xWSCurrentOrientation.e33);
			const Asura_Vector_3	xWSFinalDirection(xWSFinalOrientation.e13, xWSFinalOrientation.e23, xWSFinalOrientation.e33);

			Asura_DebugLineRender::AddArrow(xWSInitialPosition, Asura_Vector_3(xWSInitialPosition + (xWSInitialDirection * dbg_fMoveDebugRenderOrientationArrowSize)), 0x80FF0000);
			Asura_DebugLineRender::AddArrow(xWSCurrentPosition, Asura_Vector_3(xWSCurrentPosition + (xWSCurrentDirection * dbg_fMoveDebugRenderOrientationArrowSize)), 0x80FFFF00);
			Asura_DebugLineRender::AddArrow(xWSFinalPosition, Asura_Vector_3(xWSFinalPosition + (xWSFinalDirection * dbg_fMoveDebugRenderOrientationArrowSize)), 0x8000FF00);

			if(m_pxOrientationInterpolationData->GetMovementOffsetOperation() != Asura_MoveData_OrientationInterpolation::IGNORE_MOVEMENT_OFFSET)
			{
				const Asura_Vector_3	xWSMOCurrentDirection(xWSMOCurrentOrientation.e13, xWSMOCurrentOrientation.e23, xWSMOCurrentOrientation.e33);
				const Asura_Vector_3	xWSMOFinalDirection(xWSMOFinalOrientation.e13, xWSMOFinalOrientation.e23, xWSMOFinalOrientation.e33);

				Asura_DebugLineRender::AddArrow(xWSMOCurrentPosition, Asura_Vector_3(xWSMOCurrentPosition + (xWSMOCurrentDirection * dbg_fMoveDebugRenderOrientationArrowSize)), 0x80FF00FF);
				Asura_DebugLineRender::AddArrow(xWSMOFinalPosition, Asura_Vector_3(xWSMOFinalPosition + (xWSMOFinalDirection * dbg_fMoveDebugRenderOrientationArrowSize)), 0x800000FF);
			}

#if defined(ASURA_USE_DEBUG_TEXT)
			if(m_pxOrientationInterpolationData->IsDurationBased())
			{
				const float	fTimer				= m_pxOrientationInterpolationData->GetTimer();
				const float	fDuration			= m_pxOrientationInterpolationData->GetDuration();
				const float	fParametricTimer	= m_pxOrientationInterpolationData->GetParametricTimer();

				Asura_DebugText::AddTextBlock(uRow, uColumn++, 0x80FFFFFF, 0, "-> Ori: State = %d, Timer = %.3f/%.3fs (%.3f)", eState, fTimer, fDuration, fParametricTimer);
			}
#endif	// defined(ASURA_USE_DEBUG_TEXT)
		}

		// Finally, debug render the actual position...
		Asura_DebugLineRender::AddAxis(xWSActualPosition, xWSActualOrientation, dbg_fMoveDebugRenderPositionAxisSize);
	}
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)
}

/// End of function 'Asura_Move::Render'. ///



//////////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::Abort
// Author        : Richard May
// Description   : Premature abort of move
//////////////////////////////////////////////////////////////////////////////////////

void Asura_Move::Abort()
{
	// Only send this message if this is local player...
	if(!m_bSuppressAbortMessage && m_xParentSystem.IsControlledByLocalPlayer())
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_ABORT, sizeof(Asura_Msg_Move));

		if(pxMessage != NULL)
		{
			Asura_Msg_Move*		pxData		= static_cast<Asura_Msg_Move*>(pxMessage->GetData());
			const Asura_Guid	uOwnerGuid	= m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(true);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}

	if ( m_pxAnimationData )
	{
		m_pxAnimationData->StopLoopedAnims();
	}
	
	// Reset the variables...
	Reset();
}

/// End of function "Asura_Move::Abort". ///



//////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::Finish
// Author        : Richard May
// Description   : Called at the end of move. 
//////////////////////////////////////////////////////////////////////////////

void Asura_Move::Finish()
{
	// Only send this message if this is local player...
	if(!m_bSuppressFinishMessage && m_xParentSystem.IsControlledByLocalPlayer())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(ASURA_MSG_MOVE_FINISHED, sizeof(Asura_Msg_Move));

		if(pxMessage != NULL)
		{
			Asura_Msg_Move*	pxData		= static_cast<Asura_Msg_Move*>(pxMessage->GetData());
			const Asura_Guid		uOwnerGuid	= m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(true);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}

	if ( m_pxAnimationData )
	{
		m_pxAnimationData->StopLoopedAnims();
	}

	// Reset the variables...
	Reset();
}

/// End of function "Asura_Move::Finish". ///



//////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::CalculateOrientation
// Author        : Richard May
// Description   : Calculate the current orientation using the interpolator.
//////////////////////////////////////////////////////////////////////////////

bool Asura_Move::CalculateOrientation(const float fTimeStep,
									  Asura_Quat& xOrientation) const
{
	bool	bUpdating;

	if((m_pxAnimationData == NULL) || (m_pxOrientationInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_OrientationInterpolation::IGNORE_MOVEMENT_OFFSET))
	{
		// No animation - interpolate normally...
		bUpdating	= m_pxOrientationInterpolationData->Update(fTimeStep);
	}
	else
	{
		Asura_Quat	xWSCurrentOrientation;
		Asura_Quat	xWSFinalOrientation;

		Asura_Quat	xLSMOCurrentOrientation;
		Asura_Quat	xLSMOFinalOrientation;

		// Calculate the orientations from the movement-offset data...
		m_xParentSystem.GetLSMOCurrentOrientation(xLSMOCurrentOrientation);
		m_xParentSystem.GetLSMOFinalOrientation(xLSMOFinalOrientation);

		// Use the animation AND local-space movement-offset data for interpolation...
		bUpdating	= m_pxOrientationInterpolationData->Update(fTimeStep, xLSMOCurrentOrientation, xLSMOFinalOrientation);
	}

	xOrientation	= m_pxOrientationInterpolationData->GetActual();
	return(bUpdating);
}

/// End of function "Asura_Move::CalculateOrientation". ///



/////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::CalculatePosition
// Author        : Richard May
// Description   : Gets the desired position of the move from
//                 the animation's data (assuming we're already at the 
//                 target position)
/////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::CalculatePosition(const float fTimeStep,
								   Asura_AVector3_OutParam xPosition) const
{
	bool	bUpdating;

	if((m_pxAnimationData == NULL) || (m_pxPositionInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_PositionInterpolation::IGNORE_MOVEMENT_OFFSET))
	{
		// No animation - interpolate normally...
		bUpdating	= m_pxPositionInterpolationData->Update(fTimeStep);
	}
	else
	{
		Asura_Quat		xWSCurrentOrientation;
		Asura_Quat		xWSFinalOrientation;

		Asura_AVector3	xLSMOCurrentPosition;
		Asura_AVector3	xLSMOFinalPosition;

		// Calculate the positions from the movement-offset data...
		m_xParentSystem.GetLSMOCurrentPosition(xLSMOCurrentPosition);
		m_xParentSystem.GetLSMOFinalPosition(xLSMOFinalPosition);

		if(m_pxOrientationInterpolationData == NULL)
		{
			m_xParentSystem.GetCurrentMoveOrientation(xWSCurrentOrientation);

			xWSFinalOrientation	= xWSCurrentOrientation;
		}
		else
		{
			xWSCurrentOrientation	= m_pxOrientationInterpolationData->GetCurrent();
			xWSFinalOrientation		= m_pxOrientationInterpolationData->GetFinal();
		}

		// Use the animation AND movement-offset data for interpolation...
		bUpdating	= m_pxPositionInterpolationData->Update(fTimeStep, xLSMOCurrentPosition, xLSMOFinalPosition, xWSCurrentOrientation, xWSFinalOrientation);
	}

	xPosition	= m_pxPositionInterpolationData->GetActual();
	return(bUpdating);
}

/// End of function "Asura_Move::CalculatePosition". ///



/////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::CalculateVelocity
// Author        : Richard May
// Description   : Gets the desired velocity of the move from
//                 the animation's data (assuming we're already at the 
//                 target position)
/////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::CalculateVelocity(const float fTimeStep,
								   Asura_AVector3_OutParam xVelocity) const
{
	bool	bUpdating;

	xVelocity.Zero();

	// The no-position-interpolation case now handled here, so subclasses can override the handling.
	if (!m_pxPositionInterpolationData)
	{
		if (m_pxAnimationData)
		{
			if(m_pxAnimationData->CalculateVelocity(xVelocity))
			{
				// We first need to eliminate any orientation changes that have been introduced since the animation began.
				Asura_Quat xOrientation;
				if ( m_pxAnimationData->CalculateMovementOffset( NULL, &xOrientation ) )
				{
					xOrientation.Invert();
					xVelocity = xOrientation.RotateVector( xVelocity);
				}

				// We can now resolve to world space.
				xVelocity = xVelocity * m_xParentSystem.GetOwner().GetCurrentOrientation();
			}
		}

		return false;
	}

	if(m_pxAnimationData == NULL)
	{
		// No animation - interpolate normally...
		bUpdating	= m_pxPositionInterpolationData->UpdateUsingVelocity(fTimeStep, Asura_Maths::ZeroAVector3);
	}
	else if(m_pxPositionInterpolationData->GetMovementOffsetOperation() == Asura_MoveData_PositionInterpolation::IGNORE_MOVEMENT_OFFSET)
	{
		// Get the velocity from the animation...
		if ((m_pxPositionInterpolationData->GetVelocityOperation() != Asura_MoveData_PositionInterpolation::IGNORE_VELOCITY) && m_pxAnimationData->CalculateVelocity(xVelocity))
		{
			// We first need to eliminate any orientation changes that have been introduced since the animation began.
			Asura_Quat xOrientation;
			if ( m_pxAnimationData->CalculateMovementOffset( NULL, &xOrientation ) )
			{
				xOrientation.Invert();
				xVelocity = xOrientation.RotateVector( xVelocity);
			}

			// We can now resolve to world space.
			xVelocity = xVelocity * m_xParentSystem.GetOwner().GetCurrentOrientation();

			bUpdating = m_pxPositionInterpolationData->UpdateUsingVelocity(fTimeStep, xVelocity);
		}
		else
		{
			// Calculate the velocity using the distance to travel and the duration
			Asura_Assert(m_pxPositionInterpolationData->GetDuration() > 0.0f, "Cannot work out a velocity without a duration");
			if (m_pxPositionInterpolationData->GetDuration() > 0.0001f)
			{
				xVelocity = m_pxPositionInterpolationData->GetFinal() - m_pxPositionInterpolationData->GetInitial();
				xVelocity /= m_pxPositionInterpolationData->GetDuration();

				bUpdating	= m_pxPositionInterpolationData->UpdateUsingVelocity(fTimeStep, xVelocity);
			}
			else
			{
				bUpdating	= m_pxPositionInterpolationData->UpdateUsingVelocity(fTimeStep, Asura_Maths::ZeroAVector3);
			}
		}
	}
	else
	{
		Asura_Assert(false, "Unexpected movement-offset interpolation operation when calculating the velocity for a move.");
		bUpdating	= false;
	}

	return(bUpdating);
}

/// End of function "Asura_Move::CalculateVelocity". ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::IsReadyToStart
// Author       : Andy Tate
// Date         : 7-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::IsReadyToStart(const Asura_MoveData& xMoveData) const
{
	if(xMoveData.IsWaiting())
	{
		const Asura_MoveData::StartType	eStartType	= xMoveData.GetStartType();

		switch(eStartType)
		{
			case Asura_MoveData::START_IMMEDIATLY:
			{
				return(true);
			}

			case Asura_MoveData::START_WHEN_ORIENTATION_COMPLETES:
			{
				return((m_pxOrientationInterpolationData == NULL) || m_pxOrientationInterpolationData->IsFinished());
			}

			case Asura_MoveData::START_WHEN_POSITION_COMPLETES:
			{
				return((m_pxPositionInterpolationData == NULL) || m_pxPositionInterpolationData->IsFinished());
			}

			case Asura_MoveData::START_WHEN_ANIMATION_COMPLETES:
			{
				return((m_pxAnimationData == NULL) || m_pxAnimationData->IsFinished());
			}

			default:
			{
				Asura_Assert(false, "Unexpected start-type for this move.");
				return(false);
			}
		}		// End of switch.
	}
	else
	{
		return(false);
	}
}

/// End of function 'Asura_Move::IsReadyToStart'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::Asura_Move
// Author       : Andy Tate
// Date         : 6-Sep-2007
////////////////////////////////////////////////////////////////////////////////

Asura_Move::Asura_Move(Asura_Movement_System& xParentSystem,
					   const Asura_Hash_ID uHash)
: m_xParentSystem(xParentSystem)
, m_uHash(uHash)
, m_pxPrevMove(NULL)
, m_pxAnimationData(NULL)
, m_pxPositionInterpolationData(NULL)
, m_pxOrientationInterpolationData(NULL)
, m_fStationaryTimer(0.0f)
, m_fStationaryEpsilon(1.0e-6f)
, m_uFlags(0)
{
	// Ensure that this move is part of the movement-system...
	m_xParentSystem.AddMove(this);
}

/// End of function 'Asura_Move::Asura_Move'. ///



/////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::AddNextMove
// Author        : Richard May
// Description   : Adds a potential next move to the chain
/////////////////////////////////////////////////////////////////////////////////

void Asura_Move::AddNextMove(Asura_Move* pxNext)
{
	Asura_Move** ppxNewNode = m_xNextMoveTree.AddNode(pxNext->GetHash());

	if(ppxNewNode)
	{
		*ppxNewNode = pxNext;			
	}
	else
	{
		Asura_Assertf(false, ("Failed to add next move with ID '%x' to move '%x'.", pxNext->GetHash(), GetHash()));
	}
}

/// End of function "Asura_Move::AddNextMove". ///



///////////////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::AddNextMove
// Author        : Andy Tate
// Description   : Adds a potential next move to the chain from a HashID
///////////////////////////////////////////////////////////////////////////////////

void Asura_Move::AddNextMove(const Asura_Hash_ID uHash)
{
	Asura_Move*	pxMove	= m_xParentSystem.GetMove(uHash);

	if(pxMove == NULL)
	{
		Asura_Assertf(false, ("Failed to find move with hash '%08X'.", uHash));
	}
	else
	{
		Asura_Move** ppxNewNode	= m_xNextMoveTree.AddNode(uHash);

		if(ppxNewNode == NULL)
		{
			Asura_Assertf(false, ("Failed to add next move with ID '%08X' to move '%08X'.", uHash, m_uHash));
		}
		else
		{
			*ppxNewNode	= pxMove;
		}
	}
}

/// End of function "Asura_Move::AddNextMove". ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Traverse_GetNextMove
// Author       : Andy Tate
// Date         : 6-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::Traverse_GetNextMove(Asura_Hash_ID uMoveHashID,
									  Asura_Move** ppxMove,
									  void* pUserData)
{
	Asura_Move**	ppxNextMove	= static_cast<Asura_Move**>(pUserData);
	Asura_Move*		pxMove		= (ppxMove == NULL) ? NULL : *ppxMove;

	// Should this move start?
	if((pxMove != NULL) && pxMove->ShouldStart())
	{
		// found move, stop
		*ppxNextMove	= pxMove;
		return(false);
	}
	else
	{
		// otherwise continue
		return(true);
	}
}

/// End of function "Asura_Move::Traverse_GetNextMove". ///



//////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::GetNextMove
// Author        : Richard May
// Description   : Traverses the next move tree and calls 
//  			   ShouldStart. Returns the first move to return true.
//////////////////////////////////////////////////////////////////////////

Asura_Move* Asura_Move::GetNextMove()
{
	Asura_Move*	pxNextMove	= NULL;

	m_xNextMoveTree.Traverse(Traverse_GetNextMove, &pxNextMove);

	return(pxNextMove);
}

/// End of function "Asura_Move::GetNextMove". ///



//////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::SetNextMoves
// Author        : Chris Wallace
// Description   : Adds Next moves from an argument list of moves
//////////////////////////////////////////////////////////////////////////

void Asura_Move::SetNextMoves(u_int uCount, ...)
{
	if(m_xNextMoveTree.GetNumberOfNodes() == 0)
	{
		if(uCount > 0)
		{
			va_list xArgs;
			va_start(xArgs, uCount);
			for(u_int u = 0; u < uCount; ++u)
			{
				Asura_Move* pxMove = va_arg(xArgs, Asura_Move*);
				Asura_Assert(pxMove, "");
				Asura_Assert(pxMove != this, "Re-entrant moves not currently supported.");
				AddNextMove(pxMove);
			}
			va_end(xArgs);
		}
	}
	else
	{
		Asura_Assert(false, "You should manually clear the next move list if you're intending to create a new one.");
	}
}

/// End of function "Asura_Move::SetNextMoves". ///



//////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::SetNextMovesByHash
// Author        : Chris Wallace
// Description   : Adds Next moves from an argument list of hashes
//////////////////////////////////////////////////////////////////////////

void Asura_Move::SetNextMovesByHash(u_int uCount, ...)
{
	if(m_xNextMoveTree.GetNumberOfNodes() == 0)
	{
		if(uCount > 0)
		{
			va_list xArgs;
			va_start(xArgs, uCount);
			for(u_int u = 0; u < uCount; ++u)
			{
				Asura_Hash_ID uMoveHashID = va_arg(xArgs, Asura_Hash_ID);
				Asura_Move* pxMove = m_xParentSystem.GetMove(uMoveHashID);
				Asura_Assertf(pxMove, ("Failed to find move with hash 0x%08X.", uMoveHashID));
				Asura_Assert(pxMove != this, "Re-entrant moves not currently supported.");
				AddNextMove(pxMove);
			}
			va_end(xArgs);
		}
	}
	else
	{
		Asura_Assert(false, "You should manually clear the next move list if you're intending to create a new one.");
	}
}

/// End of function "Asura_Move::SetNextMovesByHash". ///


//////////////////////////////////////////////////////////////////////////
// Name          : Asura_Move::ResetNextMove
// Author        : Sharon Galtzur
// Description   : Remove all the Next moves for animation trees that are build dynamically
//////////////////////////////////////////////////////////////////////////

void Asura_Move::ResetNextMoves()
{
	m_xNextMoveTree.Reset();
}



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::HasFinished
// Author       : Andy Tate
// Date         : 12-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::HasFinished(void) const
{
	const bool	bAnimationFinished					= ((m_pxAnimationData == NULL) || m_pxAnimationData->IsFinished());
	const bool	bOrientationInterpolationFinished	= ((m_pxOrientationInterpolationData == NULL) || m_pxOrientationInterpolationData->IsFinished());
	const bool	bPositionInterpolationFinished		= ((m_pxPositionInterpolationData == NULL) || m_pxPositionInterpolationData->IsFinished());

	return(bAnimationFinished && bOrientationInterpolationFinished && bPositionInterpolationFinished);
}

/// End of function 'Asura_Move::HasFinished'. ///

///////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::GetAnimFlagsForLastFrame
// Author       : Chris Wallace
// Date         : 25-Jan-2008
// Description	: Returns the flags for last frame for ONLY this moves animation
///////////////////////////////////////////////////////////////////////////////////

u_int Asura_Move::GetAnimFlagsForLastFrame( void ) const
{
	if(m_pxAnimationData)
	{
		Asura_AnimationController_Animation* pxAnimController = m_pxAnimationData->GetAnimationController();
		if(pxAnimController)
		{
			return(pxAnimController->GetFlagsForLastFrame());
		}
	}	
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Move::IsPlayingAnimation
// Author       : Andy McCabe
// Date         : 04-Nov-2008
// Description	: Returns whether this move is playing an animation
////////////////////////////////////////////////////////////////////////////////

bool Asura_Move::IsPlayingAnimation() const
{
	return (m_pxAnimationData && m_pxAnimationData->IsUpdating());
}

#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)


/// End of file. ///
