//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity_NetworkFPSPlayer.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Player/Asura_Entity_NetworkFPSPlayer.h"

#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Dynamics/Asura_Dynamics.h"
#include "Entity/Asura_PosAndOriInterpolator.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uASURA_PLAYER_MAX_STORED_UPDATES = 32;

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR float dbg_fOrientXClampDead = 1.4f;
DEBUGVAR float dbg_fOrientXClampAlive = 1.3f;
DEBUGVAR float dbg_fSpeedModBackward = 0.8f;	// multiplier for speed going backwards.
DEBUGVAR float dbg_fSpeedModStrafe = 0.8f;		// multiplier for speed going sideways
DEBUGVAR float dbg_fMaxSpeedStanding = 6.0f;			// TODO: set this in project code.
DEBUGVAR float dbg_fMaxSpeedCrouching = 2.5f;			// TODO: set this in project code.
DEBUGVAR float dbg_fMaxSpeedProne = 0.4f;			// TODO: set this in project code.

// 1 / max time between updates sent to server
DEBUGVAR float dbg_fPlayerCommandMaxSendRate = 10.0f;

// 1 / max time between input polls
DEBUGVAR float dbg_fPlayerCommandMaxUpdateRate = 30.0f;

DEBUGVAR bool dbg_bDebugPacketLoss = true;	// print messages indicating recovery status when we detect missing packets from client
DEBUGVAR bool dbg_bPrintPlayerUpdatesOnServer = false;	// print position updates as they are calculated by server entity
DEBUGVAR bool dbg_bPrintPlayerUpdatesOnClient = false;	// print position updates as they are calculated by client entity
DEBUGVAR bool dbg_bPrintPlayerPosDiscrepancies = true;	// print messages when client's position is corrected

DEBUGVAR float dbg_fMaxTimeBetweenPlayerPositionValidations = 2.0f;	// if longer than this elapsed, stop moving and display message.

DEBUGVAR float dbg_fLocalPlayerPosTimerFix = -0.06f;	// this is added to the current time to compensate for the 0.1s added by interpolator

// S T A T I C   F U N C T I O N S ///////////////////////////////////////////


// static function that can be called by client and server code to apply commands to 
static void ApplyPlayerCommandsToDynamicsBlock(const Asura_Msg_PlayerMovementCommands* pxCommands, class Asura_Dynamic_Object* pxDyn)
{
	Asura_Assert(pxCommands, "We should have a command message");

	if (!pxCommands)
	{
		return;
	}

	// Update player orientation/velocity
	if (pxDyn)
	{
		// Orientation
		Asura_Matrix_3x3 xOri;
		Asura_Euler_3 xEuler(pxCommands->GetOrientX(), pxCommands->GetOrientY(), 0.0f);
		xEuler.ToMatrix(xOri);
		pxDyn->SetOrientation(xOri);

		// Movement
		Asura_Vector_3 xDelta (xOri.e13, 0.0f, xOri.e33);
		Asura_Vector_3 xDeltah(xOri.e11, 0.0f, xOri.e31);

		xDelta.Normalise();
		xDeltah.Normalise();
	
		Asura_Vector_3 xLinearTargetVelocity;

		float fWalkAxis = pxCommands->GetWalkAxis();
		float fStrafeAxis = pxCommands->GetStrafeAxis();

		// if moving backwards or sideways, reduce speed slightly
		if (fWalkAxis < 0.0f)
		{
			fWalkAxis *= dbg_fSpeedModBackward;
		}
		fStrafeAxis *= dbg_fSpeedModStrafe;

		float fMaxSpeed;
		u_short usFlags = pxCommands->GetFlags();
		if (usFlags & ASURA_MOVEMENTFLAG_CROUCHING)
		{
			fMaxSpeed = dbg_fMaxSpeedCrouching;
		}
		else if (usFlags & ASURA_MOVEMENTFLAG_PRONE)
		{
			fMaxSpeed = dbg_fMaxSpeedProne;
		}
		else
		{
			fMaxSpeed = dbg_fMaxSpeedStanding;
		}

		xLinearTargetVelocity = xDelta*(fMaxSpeed * fWalkAxis);
		xLinearTargetVelocity += xDeltah*(fMaxSpeed * fStrafeAxis);
	
		pxDyn->SetLinVelocity(xLinearTargetVelocity);
	}

	//EJD 13/01/2003 Jump code added (from Dredd)
	if(pxCommands->GetFlags() & ASURA_MOVEMENTFLAG_JUMP)
	{
		if (pxDyn->IsInContactWithFloor() || pxDyn->IsInContactWithNearlyFloor())
		{
			//Get current impulse
			Asura_Vector_3 vImpulse = pxDyn->GetLinImpulse();
			//Add jump component
			vImpulse.y = Asura_Maths::Max(-3.5f, vImpulse.y - 3.5f);
			//Set new impulse
			pxDyn->SetLinImpulse(vImpulse);
		}
	}
// (MH - 25.06.2004) since this is effectively a debug command I think if it is wanted it should
// go in its own debug update message rather than consume bandwidth in the main update message.
//	// (EJD - 06.03.2003) Teleport added from SAS 
//	if(pxCommands->uCommands & BHD_CONTROLLER_COMMAND_TELEPORT)
//	{
//		Asura_Camera::SetCameraIndex(0);
//
//		//Get the position and orientation
//		const Asura_Vector_3 xPosition = *(Asura_Camera::GetPosition());
//		const Asura_Matrix_3x3* xOrientation = Asura_Camera::GetOrientation();
//		//Convert orientation to a vector to feed to LOS test
//		Asura_Vector_3 xVectOrientation;
//		xVectOrientation.x = xOrientation->e13;
//		xVectOrientation.y = xOrientation->e23;
//		xVectOrientation.z = xOrientation->e33;
//
//		Asura_LOS_InputData xLOSInput(xPosition, xVectOrientation, GetGuid());
//		const Asura_LOS_OutputData* pxLOSOutput = Asura_LOS::TestAgainstEverything(&xLOSInput);
//		Asura_Assert(!(pxLOSOutput->m_bHitEnvironment && pxLOSOutput->m_uHitEntity), "");
//		if (pxLOSOutput->m_bHitEnvironment || pxLOSOutput->m_uHitEntity)
//		{
//			Asura_Vector_3 xNewPos = xPosition +(xVectOrientation * pxLOSOutput->m_fDistanceToIntersection);
//			xNewPos.y -= 0.25f; 
//			Asura_Vector_3 displacement = xNewPos - pxDyn->GetPosition();
//			pxDyn->SetPosition(xNewPos);
//			Asura_Bounding_Box bb =  pxDyn->GetBoundingBox();
//			bb.Displace(displacement);
//		}
//	// End teleport code
//	}
}


// P L A Y E R   C L I E N T   C O D E ///////////////////////////////////////
Asura_Entity_NetworkFPSPlayer_Client::Asura_Entity_NetworkFPSPlayer_Client()
: m_ePosture(ASURA_FPS_PLAYERPOSTURE_STANDING)
, m_fClampY(Asura_Maths::LargeFloat)
, m_bReceivedFirstUpdate(false)
, m_fOrientX(0.0f)
, m_fOrientY(0.0f)
, m_pxUpdatePackageStorage(NULL)
, m_pxDynamicObject(NULL)
{
	Reset();
}

Asura_Entity_NetworkFPSPlayer_Client::~Asura_Entity_NetworkFPSPlayer_Client()
{
	Shutdown();
}


void Asura_Entity_NetworkFPSPlayer_Client::Initialise()
{
	Shutdown();

	if (ShouldUseMultipleCommands())
	{
		m_pxUpdatePackageStorage = new Asura_PlayerUpdatePackage[uASURA_PLAYER_MAX_STORED_UPDATES];
		Asura_Dynamics_System::AllocateDynamicsBlock(m_pxDynamicObject, ASURA_DYNAMIC_BLOCK_DYNAMIC);
		m_pxDynamicObject->SetExcludeFromMainUpdate(true);
		m_pxDynamicObject->SetCanClimbStairs(true);
	}


}

void Asura_Entity_NetworkFPSPlayer_Client::Shutdown()
{
	if (m_pxUpdatePackageStorage)
		delete[] m_pxUpdatePackageStorage;
	
	Asura_Dynamics_System::ReleaseDynamicsBlock(m_pxDynamicObject);

	m_pxUpdatePackageStorage = NULL;
	m_pxDynamicObject		 = NULL;
}

void Asura_Entity_NetworkFPSPlayer_Client::Reset()
{
	Initialise();

	if (ShouldUseMultipleCommands())
	{
		// ignore the server dynamics block, if it exists on this machine
		// This is done here so that if the server entity was destroyed and recreated, calling
		// Reset() sorts it out.
		const Asura_ServerEntity_PhysicalObject* pxServerPlayer = Asura_ServerEntity_PhysicalObject::FindEntity(GetGuid());

		if (pxServerPlayer)
		{
			const Asura_Dynamic_Object* pxServerDyn = pxServerPlayer->GetDynamicsBlock();
			Asura_Assert(pxServerDyn, "server player has no dyn block");
			if (pxServerDyn)
			{
				m_pxDynamicObject->SetHandleToIgnore(pxServerDyn->GetHandle());
			}
		}
	}

	m_bReceivedFirstUpdate = false;

	m_fTimeToNextCommandUpdate					= 1.0f / dbg_fPlayerCommandMaxUpdateRate;
	m_fGameTimeAtLastCommandUpdate				= Asura_Timers::GetLocalScaledTime();
	m_fTimeToNextCommandSend					= 1.0f / dbg_fPlayerCommandMaxSendRate;
	m_uCurrentUpdatePackage						= 0;
	m_fLastServerPositionConfirmationTimestamp	= 0.0f;
	m_fTimeOfLastPositionConfirmation			= Asura_Timers::GetLocalScaledTime();
}


void Asura_Entity_NetworkFPSPlayer_Client::MakeCommandsMessageForServer( bool bUsingMouse )
{
	if (!HasReceivedFirstUpdate())
	{
		return;
	}
	
	// build the current command packet every frame. This may OR in various flags to ensure 
	// that all button presses are captured, if we're not acting on the packet every frame.
	BuildCommandPacket(bUsingMouse);
	
	if (ShouldUseMultipleCommands())
	{
		UpdateTimers();
		
		if (IsTimeToActOnCommandPacket())
		{
			// update the local dynamics block
			ActOnCommandPacket();
		
			if (IsTimeToSendCommandPacket())
			{
				// send the updates we've done since the last send, and include some previous ones
				SendMultipleCommandPacket();
			}
		}
	}
	else
	{
		SendSingleCommandPacket();
	}
}


bool Asura_Entity_NetworkFPSPlayer_Client::ShouldUseMultipleCommands( void ) const
{
	return Asura_Network::IsPlayingOnInternet();
}


void Asura_Entity_NetworkFPSPlayer_Client::UpdateTimers( void )
{
	m_fTimeToNextCommandSend -= Asura_Timers::GetFrameTime();
	m_fTimeToNextCommandUpdate -= Asura_Timers::GetFrameTime();
}


bool Asura_Entity_NetworkFPSPlayer_Client::IsTimeToActOnCommandPacket( void ) const
{
	return (m_fTimeToNextCommandUpdate <= 0.0f);
}


bool Asura_Entity_NetworkFPSPlayer_Client::IsTimeToSendCommandPacket( void ) const
{
	return (m_fTimeToNextCommandSend <= 0.0f);
}


void Asura_Entity_NetworkFPSPlayer_Client::ActOnCommandPacket( void )
{
	// reset timer
	m_fTimeToNextCommandUpdate = 1.0f / dbg_fPlayerCommandMaxUpdateRate;

	// (MH - 28.06.2004) TODO: don't want to have to search for this every frame
	Asura_Entity_Renderable* pxActor = Asura_Entity_Renderable::FindEntity(GetGuid());
	Asura_Assert(pxActor, "no actor");
	if (!pxActor) return;

//	// if we're doing client dynamics (for net play) check stuff here
//	if (pxDyn)
//	{
//		if (m_bCrouching && !bWasCrouching)
//		{
//			// can always crouch from standing
//			pxActor->SetBBIndex(DREDD_BB_HUMAN_CROUCHED);
//		}
//		else if (bWasCrouching && !m_bCrouching)
//		{
//			// check if we think we can stand up
//			pxActor->SetBBIndex(DREDD_BB_HUMAN_STANDING);
//			if ( !Asura_Dynamics_System::CanPlaceDynamicObjectAt(pxDyn, pxDyn->GetPosition(),*(pxDyn->GetCollisionObject())) )
//			{
//				// nope, stay crouched
//				m_bCrouching = true;
//				pxActor->SetBBIndex(DREDD_BB_HUMAN_CROUCHED);
//			}
//		}
//	}
	
	float fUpdateTimeSlice = Asura_Timers::GetLocalScaledTime() - m_fGameTimeAtLastCommandUpdate;

	m_fGameTimeAtLastCommandUpdate = Asura_Timers::GetLocalScaledTime();
		
	Asura_PlayerUpdatePackage* pxCurUpdatePackage = &m_pxUpdatePackageStorage[m_uCurrentUpdatePackage];

	Asura_Assert(!pxCurUpdatePackage->m_bHasBeenSent, "");

	// add to queue. check there is room, if not should perhaps increase update rate rather than have
	// player performance suffer. This could happen if client FPS is very high.
	if (pxCurUpdatePackage->m_usNumCommands >= uASURA_PLAYER_MAX_COMMANDS_IN_UPDATE_PACKAGE)
	{
		Asura_Assert(false, "more client updates per second than available storage. Need to reduce framerate or increase update rate");
	}
	else
	{
		Asura_Msg_PlayerMovementCommands* pxCommandData = &pxCurUpdatePackage->m_pxCommands[pxCurUpdatePackage->m_usNumCommands++];

		// write current commands into store and clear them
		*pxCommandData = m_xCurrentCommands;
		m_xCurrentCommands.SetFlags(0);
		pxCommandData->SetFrameTime(fUpdateTimeSlice);

		if (dbg_bPrintPlayerUpdatesOnClient)
		{
			Asura_Dynamic_Object* pxDyn = m_pxDynamicObject;

			Asura_OutputToDebugger(("cl[%u], P0(%.4f,%.4f), V0(%.4f), (%.4f, %.4f), ", 
				pxCurUpdatePackage->m_usUpdateID, 
				Asura_Maths::FtoD(pxDyn->GetPosition().x), 
				Asura_Maths::FtoD(pxDyn->GetPosition().z),
				Asura_Maths::FtoD(pxDyn->GetLinVelocity().Magnitude()),
				Asura_Maths::FtoD(pxCommandData->GetWalkAxis()), 
				Asura_Maths::FtoD(pxCommandData->GetStrafeAxis())));
		}

		// (MH - 18.08.2003) run the server dynamics simulation on our local dynamics block.
		// First check we've been validated recently, if not don't move.
		float fTimeSinceLastServerPositionValidation = Asura_Timers::GetLocalScaledTime() - m_fTimeOfLastPositionConfirmation;
		Asura_Assert(fTimeSinceLastServerPositionValidation >= 0.0f, "negative time elapsed?");
		if (fTimeSinceLastServerPositionValidation > dbg_fMaxTimeBetweenPlayerPositionValidations)
		{
			// TODO: display "connection problems"
			Asura_OutputToDebugger(("connection problems...\n"));
		}
		else
		{
			// if we're alive, do the dynamics update
			if (!IsDead())
			{
				//shouldn't need to do posture here as this is using our own data which is already set up.
				//SetPostureFromCommands(pxCommandData);
				ApplyPlayerCommandsToDynamicsBlock(pxCommandData, m_pxDynamicObject);
				
				#ifdef ASURA_DEBUG_PLAYER_DYNAMICS
				extern int g_iDoingPlayerDynamics;
				g_iDoingPlayerDynamics = 2;
				#endif
				
				Asura_Dynamics_System::UpdateDynamicObject(reinterpret_cast<Asura_Dynamic_DynamicObject*>(m_pxDynamicObject), pxCommandData->GetFrameTime());

				if (dbg_bPrintPlayerUpdatesOnClient)
				{
					Asura_Dynamic_Object* pxDyn = m_pxDynamicObject;

					Asura_OutputToDebugger(("P1(%.4f,%.4f), V1(%.4f) t(%.4f).\n", 
						Asura_Maths::FtoD(pxDyn->GetPosition().x), 
						Asura_Maths::FtoD(pxDyn->GetPosition().z),
						Asura_Maths::FtoD(pxDyn->GetLinVelocity().Magnitude()), 
						Asura_Maths::FtoD(Asura_Timers::GetServerAbsoluteTime()))); 
				}

				Asura_Quat xQuat;
				xQuat.FromMatrix(m_pxDynamicObject->GetOrientation());

				pxActor->ProcessNewPositionAndOrientationData(m_pxDynamicObject->GetPosition(), xQuat, Asura_Timers::GetServerAbsoluteTime() + dbg_fLocalPlayerPosTimerFix);
			}
		}
	}
}


void Asura_Entity_NetworkFPSPlayer_Client::SendMultipleCommandPacket( void )
{
	m_fTimeToNextCommandSend = 1.0f / dbg_fPlayerCommandMaxSendRate;


	Asura_PlayerUpdatePackage* pxCurUpdatePackage = &m_pxUpdatePackageStorage[m_uCurrentUpdatePackage];
		
	// fill out package
	pxCurUpdatePackage->m_bHasBeenSent		= true;
	pxCurUpdatePackage->m_xPositionWhenSent = m_pxDynamicObject->GetPosition();

	u_int uPrev1 = (m_uCurrentUpdatePackage - 1 + uASURA_PLAYER_MAX_STORED_UPDATES) % uASURA_PLAYER_MAX_STORED_UPDATES;
	u_int uPrev2 = (m_uCurrentUpdatePackage - 2 + uASURA_PLAYER_MAX_STORED_UPDATES) % uASURA_PLAYER_MAX_STORED_UPDATES;
	const Asura_PlayerUpdatePackage* pxPrev1 = &m_pxUpdatePackageStorage[uPrev1];
	const Asura_PlayerUpdatePackage* pxPrev2 = &m_pxUpdatePackageStorage[uPrev2];
	
	u_int uMessageDataSize = Asura_Msg_MultiplePlayerCommands_Header::GetSizeRequired(pxCurUpdatePackage, pxPrev1, pxPrev2);

	Asura_Message* m = Asura_Network::CreateMessage(ASURA_MSG_PLAYER_MULTIPLECOMMANDS, uMessageDataSize);

	if(m)
	{
		u_int uBytesWritten = Asura_Msg_MultiplePlayerCommands_Header::CreateMessageHeaderAndData(m->GetData(), pxCurUpdatePackage, pxPrev1, pxPrev2);
		Asura_Assert(uBytesWritten == uMessageDataSize, "incorrect number of bytes written");
		Send(m,Guid);
	}

	// new update package
	m_uCurrentUpdatePackage = (m_uCurrentUpdatePackage + 1) % uASURA_PLAYER_MAX_STORED_UPDATES;
	m_pxUpdatePackageStorage[m_uCurrentUpdatePackage].m_usUpdateID		= static_cast<u_short>(pxCurUpdatePackage->m_usUpdateID + 1);
	m_pxUpdatePackageStorage[m_uCurrentUpdatePackage].m_bHasBeenSent	= false;
	m_pxUpdatePackageStorage[m_uCurrentUpdatePackage].m_usNumCommands	= 0;
	m_pxUpdatePackageStorage[m_uCurrentUpdatePackage].m_fCreationTime	= Asura_Timers::GetServerAbsoluteTime();
}


void Asura_Entity_NetworkFPSPlayer_Client::BuildCommandPacket( bool bUsingMouse )
{	
	Asura_Msg_PlayerMovementCommands* pxData = &m_xCurrentCommands;
	bool bRequestCrouch = (m_ePosture == ASURA_FPS_PLAYERPOSTURE_CROUCHING);
	bool bRequestProne = (m_ePosture == ASURA_FPS_PLAYERPOSTURE_PRONE);

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	if (Asura_CommandConsole::IsConsoleDown())
	{
		// stop movement; look and flags will not be updated. Need to set posture though.
		pxData->SetWalkAxis(0.0f);
		pxData->SetStrafeAxis(0.0f);
		if (bRequestCrouch)
		{
			Asura_Assert(!bRequestProne, "can't be crouched and prone at the same time");
			pxData->SetFlags(ASURA_MOVEMENTFLAG_CROUCHING);
		}
		else if (bRequestProne)
		{
			pxData->SetFlags(ASURA_MOVEMENTFLAG_PRONE);
		}
		return;
	}
#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE

	float fInput_WalkAxis;
	float fInput_StrafeAxis;
	float fInput_ViewHoriz;
	float fInput_ViewVert;
	GetWalkAxes(fInput_WalkAxis, fInput_StrafeAxis);
	GetLookAxes(fInput_ViewHoriz, fInput_ViewVert);

	bool bRequestJump = IsJumpRequested();

	// OR in commands since we last sent
	u_int uCommands = pxData->GetFlags();

	//-----------------------------------------------
	// Calculate velocity of thumb stick, as well as 
	// its acceleration.
	//-----------------------------------------------
	float fMag = (fInput_WalkAxis * fInput_WalkAxis) + (fInput_StrafeAxis * fInput_StrafeAxis);

	if(fMag > 1.0f)
	{
		fMag			   = Asura_Maths::Sqrt(fMag);
		fInput_WalkAxis   /= fMag;
		fInput_StrafeAxis /= fMag;
	}

	// set walk and strafe
	pxData->SetWalkAxis(fInput_WalkAxis);
	pxData->SetStrafeAxis(fInput_StrafeAxis);

	// set command flags
	if (bRequestJump &&
		m_ePosture != ASURA_FPS_PLAYERPOSTURE_PRONE) // (AW - 12.08.2004) - Ignore jump requests when prone.
	{
		uCommands |= ASURA_MOVEMENTFLAG_JUMP;
	}
	if (bRequestCrouch)
	{
		Asura_Assert(!bRequestProne, "can't be crouched and prone at the same time");
		uCommands |= ASURA_MOVEMENTFLAG_CROUCHING;
		uCommands &= ~ASURA_MOVEMENTFLAG_PRONE;
	}
	else if (bRequestProne)
	{
		uCommands |= ASURA_MOVEMENTFLAG_PRONE;
		uCommands &= ~ASURA_MOVEMENTFLAG_CROUCHING;
	}

	//------------------------------------------------------------------------------------------
	//
	// UPDATE STANDARD AIM
	//
	//------------------------------------------------------------------------------------------
	{
		float fDeltaVert;
		float fDeltaHoriz;
		if (IsDead())
		{
			fDeltaVert = 0.0f;
			fDeltaHoriz = 0.0f;
		}
		else
		{
			if (IsYAxisInverted())
			{
				fInput_ViewVert *= -1.0f;
			}

			CalculateViewDirection(	fDeltaVert,		
									fDeltaHoriz, 
									fInput_ViewVert, 
									fInput_ViewHoriz, 
									bUsingMouse);
		}

		m_fOrientX += fDeltaVert;
		m_fOrientY += fDeltaHoriz;

		// Clamp x orient
		float fClampX = IsDead() ? dbg_fOrientXClampDead : dbg_fOrientXClampAlive;
		m_fOrientX = Asura_Maths::ClampToRange(m_fOrientX, -fClampX, fClampX);

		// Clamp y orient
		m_fOrientY = Asura_Maths::ClampToRange(m_fOrientY, -m_fClampY, m_fClampY);

			// (MH - 28.06.2004) force into range +/- deg180
		if (m_fOrientY < -Asura_Maths::Deg180)
		{
			m_fOrientY += Asura_Maths::Deg360;
		}
		else if (m_fOrientY > Asura_Maths::Deg180)
		{
			m_fOrientY -= Asura_Maths::Deg360;
		}
	}

	pxData->SetOrientX(m_fOrientX);
	pxData->SetOrientY(m_fOrientY);

	// set commands
	pxData->SetFlags(uCommands);
}


void Asura_Entity_NetworkFPSPlayer_Client::SendSingleCommandPacket( void )
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_CLIENTUPDATE, sizeof(Asura_Msg_PlayerMovementCommands), 0, &m_xCurrentCommands);
	Send(pxMessage, GetGuid());
	m_xCurrentCommands.SetFlags(0);
}


void Asura_Entity_NetworkFPSPlayer_Client::HandleMessage_PositionValidated(float fSentTime)
{
	if (fSentTime <= m_fLastServerPositionConfirmationTimestamp)
	{
		return;
	}

	m_fLastServerPositionConfirmationTimestamp	= fSentTime;
	m_fTimeOfLastPositionConfirmation			= Asura_Timers::GetLocalScaledTime();
	return;
}


void Asura_Entity_NetworkFPSPlayer_Client::HandleMessage_PositionCorrected(const Asura_MsgData_PlayerPositionCorrection* pxData, float fSentTime)
{
	if (fSentTime <= m_fLastServerPositionConfirmationTimestamp)
	{
		return;
	}

	m_fLastServerPositionConfirmationTimestamp = fSentTime;
	m_fTimeOfLastPositionConfirmation = Asura_Timers::GetLocalScaledTime();

	// At the time of this message, this is where the server says we were.
	// We must now go back and run the dynamics since then, if the positions
	// don't match.
	// find update with ID m_usUpdateID
	u_short usUpdateID = pxData->m_usUpdateID;
	Asura_PlayerUpdatePackage* pxUpdatePackage = NULL;
	u_int uPackageIndex;
	for (uPackageIndex = 0; uPackageIndex < uASURA_PLAYER_MAX_STORED_UPDATES; ++uPackageIndex)
	{
		if (m_pxUpdatePackageStorage[uPackageIndex].m_usUpdateID == usUpdateID)
		{
			Asura_Assert(m_pxUpdatePackageStorage[uPackageIndex].m_bHasBeenSent, "");
			pxUpdatePackage = &m_pxUpdatePackageStorage[uPackageIndex];
			break;
		}
	}

	Asura_Assert(pxUpdatePackage, "");
	if (dbg_bPrintPlayerPosDiscrepancies && !pxUpdatePackage)
	{
		Asura_OutputToDebugger(("position validated but update too old.\n"));

		// will teleport to the position. TODO: should perhaps display "connection problems".
	}
	if (pxUpdatePackage)
	{
		if (!pxData->m_xPosition.ApproximatelyEquals(pxUpdatePackage->m_xPositionWhenSent))
		{
			if (dbg_bPrintPlayerPosDiscrepancies)
			{
				Asura_Vector_3 xDiff(pxUpdatePackage->m_xPositionWhenSent - pxData->m_xPosition);
				Asura_OutputToDebugger(("player out of position, diff = %.6f, %.6f, %.6f.\n", 
					Asura_Maths::FtoD(xDiff.x), 
					Asura_Maths::FtoD(xDiff.y), 
					Asura_Maths::FtoD(xDiff.z) ));
			}
			
			Asura_Dynamic_Object* pxDyn = m_pxDynamicObject;
			Asura_Assert(pxDyn, "no dyn block");
			if (!pxDyn) return;

			// set dynblock to server's values.
			pxDyn->SetPosition(pxData->m_xPosition);
			pxDyn->SetLinImpulse(pxData->m_xLinImpulse);
			pxDyn->SetIsInContactWithFloor(pxData->m_bIsInContactWithFloor);

//			Asura_Quat xQuat;
//			xQuat.FromMatrix(pxDyn->GetOrientation());
//			pxActor->ProcessNewPositionAndOrientationData(pxDyn->GetPosition(), xQuat, pxUpdatePackage->m_fCreationTime);

			float fTimeOfThisUpdate = pxUpdatePackage->m_fCreationTime;
			// run commands which we have executed since then.
			for (u_int u = uPackageIndex + 1; u < uPackageIndex + uASURA_PLAYER_MAX_STORED_UPDATES - 1; ++u)
			{
				u_int uActualPackageIndex = u % uASURA_PLAYER_MAX_STORED_UPDATES;
				Asura_PlayerUpdatePackage* pxPackage = &m_pxUpdatePackageStorage[uActualPackageIndex];
				if (pxPackage->m_fCreationTime <= fTimeOfThisUpdate)
				{
					// this package predates the update, we must have wrapped around the list.
					break;
				}
				Asura_Assert(pxPackage->m_usNumCommands <= uASURA_PLAYER_MAX_COMMANDS_IN_UPDATE_PACKAGE, "");
				for (u_int uCommand = 0; uCommand < pxPackage->m_usNumCommands; ++uCommand)
				{
					// (MH - 18.08.2003) run the server dynamics simulation on our local dynamics block
					const Asura_Msg_PlayerMovementCommands* pxCommandData = &pxPackage->m_pxCommands[uCommand];
					float fUpdateTimeSlice = pxCommandData->GetFrameTime();
					if (!IsDead())
					{
						SetPostureFromCommands(pxCommandData);
						ApplyPlayerCommandsToDynamicsBlock(pxCommandData, m_pxDynamicObject);

						Asura_Dynamics_System::UpdateDynamicObject(reinterpret_cast<Asura_Dynamic_DynamicObject*>(pxDyn), fUpdateTimeSlice);
					}

					// TODO: print warning "position corrected"?

#ifdef ASURA_DEBUG_PLAYER_DYNAMICS
					extern int g_iDoingPlayerDynamics;
					g_iDoingPlayerDynamics = 3;
#endif
				}
			}
		}
	}
	else
	{
		// this position validation update predates the command history, so teleport.
		// TODO: should display "Warning: connection problem".

		Teleport(&pxData->m_xPosition);
	}
	return;
}


// protected function which may be overridden by derived class. In that case this should still 
// be called.
void Asura_Entity_NetworkFPSPlayer_Client::Teleport( const class Asura_Vector_3* pxPosition /* = NULL */, 
													 const class Asura_Matrix_3x3* pxOrientation /* = NULL */)
{
	if (m_pxDynamicObject)
	{
		if (pxPosition)
		{
			m_pxDynamicObject->SetPosition(*pxPosition);
		}

		if (pxOrientation)
		{
			m_pxDynamicObject->SetOrientation(*pxOrientation);
		}
	}

	// also teleport the actor

	// (MH - 28.06.2004) TODO: don't want to have to search for this every frame
	Asura_Entity_Renderable* pxActor = Asura_Entity_Renderable::FindEntity(GetGuid());
	Asura_Assert(pxActor, "no actor");
	if (!pxActor) return;

	pxActor->TeleportTo(pxPosition ? *pxPosition : pxActor->GetCurrentPosition(), 
		pxOrientation ? *pxOrientation : pxActor->GetCurrentOrientation());
}


void Asura_Entity_NetworkFPSPlayer_Client::SetPostureFromCommands(const Asura_Msg_PlayerMovementCommands* pxCommandData)
{
	// TODO - do a dynamics check before changing posture? On the client we only do this if we're 
	// processing a server correction, so the best way to do it would probably be to include the 
	// server's posture in the correction message and use that without checking.
	if (pxCommandData->GetFlags() & ASURA_MOVEMENTFLAG_CROUCHING)
	{
		m_ePosture = ASURA_FPS_PLAYERPOSTURE_CROUCHING;
	}
	else if (pxCommandData->GetFlags() & ASURA_MOVEMENTFLAG_PRONE)
	{
		m_ePosture = ASURA_FPS_PLAYERPOSTURE_PRONE;
	}
	else
	{
		m_ePosture = ASURA_FPS_PLAYERPOSTURE_STANDING;
	}
}


// P L A Y E R   S E R V E R   C O D E ///////////////////////////////////////
Asura_Entity_NetworkFPSPlayer_Server::Asura_Entity_NetworkFPSPlayer_Server()
: Asura_Entity_Server(ASURA_GUID_UNREGISTERED)
, m_eFPSPlayerPosture(ASURA_FPS_PLAYERPOSTURE_STANDING)
, m_pxPosPredictor(NULL)
{
	if (Asura_Network::IsPlayingOnInternet())
	{
		m_pxPosPredictor = new Asura_PosAndOriInterpolator();
	}

	// init this data
	m_bNeedActorUpdate			= true;
	m_bReceivedAnyCommands		= false;
	m_fLastUpdatePacketTime		= 0.0f;
	m_usLastUpdatePacketID		= 0xffff;
	m_uTimeOfLastActorUpdate	= Asura_Timers::GetMSecsSinceStart();
}

Asura_Entity_NetworkFPSPlayer_Server::~Asura_Entity_NetworkFPSPlayer_Server()
{
	delete m_pxPosPredictor;
}

void Asura_Entity_NetworkFPSPlayer_Server::HandleMessage_MultipleCommands(const void* pxCommands, float fSentTime)
{
	Asura_Assert(IsInternetPlay(), "multiple commands only work with internet play");
	
	Asura_Assert(m_pxDynamicObject, "no dyn block for server player");
	if (!m_pxDynamicObject) return;
	Asura_Assert(m_pxDynamicObject->IsExcludedFromMainUpdate() && m_pxDynamicObject->CanClimbStairs(), "not right kind of dyn block");
	if (!m_pxDynamicObject->IsExcludedFromMainUpdate())
	{
		m_pxDynamicObject->SetExcludeFromMainUpdate(true);
	}
	if (!m_pxDynamicObject->CanClimbStairs())
	{
		m_pxDynamicObject->SetCanClimbStairs(true);
	}
		

	float fTimeBetweenThisAndLastUpdate = fSentTime - m_fLastUpdatePacketTime;
	
	if (fTimeBetweenThisAndLastUpdate <= 0.0f)
	{
		// must have missed the previous one.
#if _DEBUG
		const Asura_Msg_MultiplePlayerCommands_Header* pxHeader = reinterpret_cast
			<const Asura_Msg_MultiplePlayerCommands_Header*>(pxCommands);
		Asura_OutputToDebugger(("received old update (%d) from player.\n", pxHeader->m_usUpdateID));
#endif
		return;
	}
	m_fLastUpdatePacketTime = fSentTime;

	if (fTimeBetweenThisAndLastUpdate >= 0.5f)
	{
		Asura_OutputToDebugger(("irregular updates (%.2f between) received from player.\n", Asura_Maths::FtoD(fTimeBetweenThisAndLastUpdate)));
	}

	const Asura_Msg_MultiplePlayerCommands_Header* pxHeader = reinterpret_cast
		<const Asura_Msg_MultiplePlayerCommands_Header*>(pxCommands);

	// if last packet was fairly recent, see if we missed any expected packets
	// and can compensate at all...
	u_int uPacketsBetween = (pxHeader->m_usUpdateID - m_usLastUpdatePacketID);
	bool bApplyPrev1 = false;
	bool bApplyPrev2 = false;
	if (uPacketsBetween < 10)
	{
		if (uPacketsBetween > 1)
		{
			bApplyPrev1 = true;
			if (dbg_bDebugPacketLoss)
			{
				Asura_OutputToDebugger(("prev1.\n"));
			}
		}
		if (uPacketsBetween > 2)
		{
			bApplyPrev2 = true;
			if (dbg_bDebugPacketLoss)
			{
				Asura_OutputToDebugger(("AND prev2!.\n"));
			}
		}
		if (uPacketsBetween > 3)
		{
			if (dbg_bDebugPacketLoss)
			{
				Asura_OutputToDebugger(("and *STILL* not enough!!!\n"));
			}
		}
	}

	Asura_Msg_PlayerMovementCommands axData[uASURA_PLAYER_MAX_COMMANDS_IN_UPDATE_PACKAGE];
	if (bApplyPrev2)
	{
		pxHeader->GetCommands(axData, 2);
		const Asura_Msg_PlayerMovementCommands* pxData = axData;
		for (u_int u = 0; u < pxHeader->m_uNumCommandsPrev2; ++u)
		{
			ProcessCommands(pxData);
			
			// move dynamics block according to that packet
#ifdef ASURA_DEBUG_PLAYER_DYNAMICS
			extern int g_iDoingPlayerDynamics;
			g_iDoingPlayerDynamics = 1;
#endif
			Asura_Dynamics_System::UpdateDynamicObject(reinterpret_cast<Asura_Dynamic_DynamicObject*>(m_pxDynamicObject), pxData->GetFrameTime());
			// get next message data
			pxData++;
		}
	}
	if (bApplyPrev1)
	{
		pxHeader->GetCommands(axData, 1);
		const Asura_Msg_PlayerMovementCommands* pxData = axData;
		for (u_int u = 0; u < pxHeader->m_uNumCommandsPrev1; ++u)
		{
			ProcessCommands(pxData);
			// move dynamics block according to that packet
#ifdef ASURA_DEBUG_PLAYER_DYNAMICS
			extern int g_iDoingPlayerDynamics;
			g_iDoingPlayerDynamics = 1;
#endif
			Asura_Dynamics_System::UpdateDynamicObject(reinterpret_cast<Asura_Dynamic_DynamicObject*>(m_pxDynamicObject), pxData->GetFrameTime());
			// get next message data
			pxData++;
		}
	}

	// set ID of this packet as most recent
	m_usLastUpdatePacketID = pxHeader->m_usUpdateID;

	pxHeader->GetCommands(axData, 0);
	const Asura_Msg_PlayerMovementCommands* pxData = axData;
	for (u_int u = 0; u < pxHeader->m_uNumCommands0; ++u)
	{
		if (dbg_bPrintPlayerUpdatesOnServer)
		{
			Asura_OutputToDebugger(("sv[%u], P0(%.4f,%.4f), V0(%.4f), (%.4f, %.4f), ", 
			m_usLastUpdatePacketID, 
			Asura_Maths::FtoD(m_pxDynamicObject->GetPosition().x), 
			Asura_Maths::FtoD(m_pxDynamicObject->GetPosition().z),
			Asura_Maths::FtoD(m_pxDynamicObject->GetLinVelocity().Magnitude()),
			Asura_Maths::FtoD(pxData->GetWalkAxis()), 
			Asura_Maths::FtoD(pxData->GetStrafeAxis())));
		}

		ProcessCommands(pxData);
		// move dynamics block according to that packet
#ifdef ASURA_DEBUG_PLAYER_DYNAMICS
		extern int g_iDoingPlayerDynamics;
		g_iDoingPlayerDynamics = 1;
#endif
		Asura_Dynamics_System::UpdateDynamicObject(reinterpret_cast<Asura_Dynamic_DynamicObject*>(m_pxDynamicObject), pxData->GetFrameTime());

#if _DEBUG
		if (dbg_bPrintPlayerUpdatesOnServer)
		{
			Asura_Dynamic_Object* pxDyn = m_pxDynamicObject;
			Asura_OutputToDebugger(("P1(%.4f,%.4f), V1(%.4f) upd(%.4f).\n", 
				Asura_Maths::FtoD(pxDyn->GetPosition().x), 
				Asura_Maths::FtoD(pxDyn->GetPosition().z),
				Asura_Maths::FtoD(pxDyn->GetLinVelocity().Magnitude()), 
				Asura_Maths::FtoD(pxData->GetFrameTime())));
		}
#endif

		// get next message data
		pxData++;
	}

//			// update falling damage, done here instead of in main update to keep in sync with client
//			if (m_pxDynamicObject->IsInContactWithFloor() && 
//				((m_fPrevVerticalImpulse - s_fFallingDamageThreshold) > 0.0f) )
//			{
//				TakeFallingDamage( (m_fPrevVerticalImpulse - s_fFallingDamageThreshold) * s_fFallingDamageScale );
//			}
//			m_fPrevVerticalImpulse = m_pxDynamicObject->GetLinImpulse().y;
	
	// tell the player where he was at this time, or send message confirming he was right
	if (m_pxDynamicObject->GetPosition().ApproximatelyEquals(pxHeader->m_xPredictedPosition))
	{
		Asura_Message* pxReplyMsg = Asura_Network::CreateMessage(ASURA_MSG_PLAYER_POSITIONVALIDATED);
		if (pxReplyMsg)
		{
			pxReplyMsg->To   = ASURA_GUID_PLAYER;
			pxReplyMsg->From = GetGuid();

			Asura_Network::Send(pxReplyMsg);
		}
	}
	else
	{
		Asura_Message* pxReplyMsg = Asura_Network::CreateMessage(ASURA_MSG_PLAYER_POSITIONCORRECTED, sizeof(Asura_MsgData_PlayerPositionCorrection));
		if (pxReplyMsg)
		{
			pxReplyMsg->To   = ASURA_GUID_PLAYER;
			pxReplyMsg->From = GetGuid();

			Asura_MsgData_PlayerPositionCorrection* pxData = reinterpret_cast
				<Asura_MsgData_PlayerPositionCorrection*>(pxReplyMsg->GetData());
			pxData->m_xPosition = m_pxDynamicObject->GetPosition();
			pxData->m_xLinImpulse = m_pxDynamicObject->GetLinImpulse();
			pxData->m_bIsInContactWithFloor = m_pxDynamicObject->IsInContactWithFloor();
			pxData->m_usUpdateID = pxHeader->m_usUpdateID;

			Asura_Network::Send(pxReplyMsg);
		}
	}

	// add to list of stored positions
	if (m_pxPosPredictor)
	{
		Asura_Quat xQuat;
		xQuat.FromMatrix(m_pxDynamicObject->GetOrientation());
		float fLatency = static_cast<float>(Asura_Network::GetLatencyForPlayer(GetGuid())) * 0.001f;
		m_pxPosPredictor->ProcessNewData(m_pxDynamicObject->GetPosition(), xQuat, fSentTime + fLatency);
	}
	m_bNeedActorUpdate = true;

	return;
}


void Asura_Entity_NetworkFPSPlayer_Server::ResetInternetPlay( void )
{
	Asura_Assert(IsInternetPlay(), "ResetInternetPlay called but not internet play");
	
	m_bNeedActorUpdate			= true;
	m_bReceivedAnyCommands		= false;
	m_fLastUpdatePacketTime		= 0.0f;
	m_usLastUpdatePacketID		= 0xffff;
	m_uTimeOfLastActorUpdate	= Asura_Timers::GetMSecsSinceStart();

	if (m_pxDynamicObject)
	{
		m_pxDynamicObject->SetExcludeFromMainUpdate( true );
	}

	Asura_Assert(m_pxPosPredictor, "no interpolator");
	if (m_pxPosPredictor)
	{
		m_pxPosPredictor->TeleportTo(GetPosition(), GetOrientation());
	}
}

void Asura_Entity_NetworkFPSPlayer_Server::ProcessCommands(const Asura_Msg_PlayerMovementCommands* pxCommands)
{
	SetPostureFromCommands(pxCommands);
	ApplyPlayerCommandsToDynamicsBlock(pxCommands, m_pxDynamicObject);
}

void Asura_Entity_NetworkFPSPlayer_Server::GetInternetPlayPosition( Asura_Vector_3& xPos ) const
{
	Asura_Vector_3 xVel;
	m_pxPosPredictor->GetCurrentPosition(xPos, xVel);
}

void Asura_Entity_NetworkFPSPlayer_Server::InternetPlay_ResetPosition(const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri)
{
	m_pxPosPredictor->TeleportTo(xPos, xOri);
}

// send update to clients every update cycle until commands have been received, and then 
// only when the position has actually changed. Also send at least every 5 seconds.
bool Asura_Entity_NetworkFPSPlayer_Server::ShouldSendActorUpdate( void ) const
{
	if (!IsInternetPlay())
	{
		return true;
	}
	else
	{
		return m_bNeedActorUpdate || (!m_bReceivedAnyCommands) || 
			((Asura_Timers::GetMSecsSinceStart() - m_uTimeOfLastActorUpdate) > 5000);
	}
}

// call ActorUpdated when the client update package is sent out.
void Asura_Entity_NetworkFPSPlayer_Server::ActorUpdated( void )
{
	m_bNeedActorUpdate = false;
	m_uTimeOfLastActorUpdate = Asura_Timers::GetMSecsSinceStart();
}


void Asura_Entity_NetworkFPSPlayer_Server::SetPostureFromCommands(const Asura_Msg_PlayerMovementCommands* pxCommandData)
{
	Asura_NetworkFPSPlayer_Posture eDesiredPosture;
	if (pxCommandData->GetFlags() & ASURA_MOVEMENTFLAG_CROUCHING)
	{
		eDesiredPosture = ASURA_FPS_PLAYERPOSTURE_CROUCHING;
	}
	else if (pxCommandData->GetFlags() & ASURA_MOVEMENTFLAG_PRONE)
	{
		eDesiredPosture = ASURA_FPS_PLAYERPOSTURE_PRONE;
	}
	else
	{
		eDesiredPosture = ASURA_FPS_PLAYERPOSTURE_STANDING;
	}

	if (CanChangePosture(eDesiredPosture))
	{
		ChangePosture(eDesiredPosture);
		m_eFPSPlayerPosture = eDesiredPosture;
	}
}


Asura_PlayerUpdatePackage::Asura_PlayerUpdatePackage()
{
	m_fCreationTime		= Asura_Timers::GetServerAbsoluteTime();
	m_usNumCommands		= 0;
	m_usUpdateID		= 0;
	m_bHasBeenSent		= false;
	m_pxCommands		= new Asura_Msg_PlayerMovementCommands[uASURA_PLAYER_MAX_COMMANDS_IN_UPDATE_PACKAGE];
}

Asura_PlayerUpdatePackage::~Asura_PlayerUpdatePackage()
{
	delete[] m_pxCommands;
}
