//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_CountedVolumeTrigger.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// includes
//
//////////////////////////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Entity/Triggers/Asura_ServerEntity_CountedVolumeTrigger.h"

// asura
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Message.h"

#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_CountedVolumeTrigger
//
//////////////////////////////////////////////////////////////////////////////

Asura_ServerEntity_CountedVolumeTrigger::Asura_ServerEntity_CountedVolumeTrigger( const Asura_Bounding_Box& xBB, bool bInitiallyActive /*=false*/ )
: PARENT( bInitiallyActive ),
  m_iListFlags( 0 ),
  m_iNumVolumes( 1 ),
  m_pxVolumes( new Asura_OrientedVolume[ 1 ] ),
  m_uIncrementTriggerMultiple( 0 ),
  m_uDecrementTriggerMultiple( 0 ),
  m_uTriggerCount( 0 ),
  m_uFlags( 0 ),
  m_uPrevCount( 0 ),
  m_uCurrentCount( 0 ),
  m_uTriggerType( 0 )
{
    Classification = AsuraEntityClass_CountedVolumeTrigger;

    m_pxVolumes[ 0 ].m_xBox    = xBB;
    m_pxVolumes[ 0 ].m_xOrient = Asura_Maths::IdentityQuat;

    for ( int i = 0; i < NUM_TRIGGER_TYPES; ++i )
    {
        m_ausMultipleTriggerMessageBlocks[ i ] = ASURA_STATICMESSAGE_EMPTYBLOCK;
    }
}

Asura_ServerEntity_CountedVolumeTrigger::Asura_ServerEntity_CountedVolumeTrigger( Asura_Guid uGuid, const char*& pcChunkBuffer )
: PARENT( uGuid, pcChunkBuffer ),
  m_iListFlags( 0 ),
  m_iNumVolumes( 0 ),
  m_pxVolumes( NULL ),
  m_uIncrementTriggerMultiple( 0 ),
  m_uDecrementTriggerMultiple( 0 ),
  m_uTriggerCount( 0 ),
  m_uFlags( 0 ),
  m_uPrevCount( 0 ),
  m_uCurrentCount( 0 ),
  m_uTriggerType( 0 )
{
    Classification = AsuraEntityClass_CountedVolumeTrigger;

    for ( int i = 0; i < NUM_TRIGGER_TYPES; ++i )
    {
        m_ausMultipleTriggerMessageBlocks[ i ] = ASURA_STATICMESSAGE_EMPTYBLOCK;
    }

    // initialise from the chunk buffer
    int iVersion;
    ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion, int );

    if ( iVersion == ChunkData::CURRENT_VERSION )
    {
        ChunkData xChunkData;
        ASURA_CHUNK_EXTRACT( pcChunkBuffer, xChunkData, ChunkData );

        m_iListFlags                = xChunkData.m_iListFlags;
        m_iNumVolumes               = xChunkData.m_iNumVolumes;
        m_uIncrementTriggerMultiple = xChunkData.m_uIncrementTriggerMultiple;
        m_uDecrementTriggerMultiple = xChunkData.m_uDecrementTriggerMultiple;
        m_uTriggerCount             = xChunkData.m_uTriggerCount;
        m_uFlags                    = xChunkData.m_uFlags;
        m_uPrevCount                = xChunkData.m_uPrevCount;
        m_uCurrentCount             = xChunkData.m_uCurrentCount;

        int i;
        m_pxVolumes = new Asura_OrientedVolume[ m_iNumVolumes ];
        for ( i = 0; i < m_iNumVolumes; ++i )
        {
            ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_pxVolumes[ i ].m_xBox,    Asura_Bounding_Box );
            ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_pxVolumes[ i ].m_xOrient, Asura_Quat         );
        }

        int iNumMessageBlocks = xChunkData.m_iNumMessageBlocks;
        for ( i = 0; i < iNumMessageBlocks && i < NUM_TRIGGER_TYPES; ++i )
        {
            ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_ausMultipleTriggerMessageBlocks[ i ], unsigned short );
        }

        unsigned short usDummy;
        ASURA_CHUNK_EXTRACT( pcChunkBuffer, usDummy, unsigned short );
		usDummy = usDummy; // This is done to prevent an warning when compiling for PSP
    }
    else
    {
        Asura_Assertf( false, ( "Asura_ServerEntity_CountedVolumeTrigger - unhandled chunk version, %d", iVersion ) );
    }
}

Asura_ServerEntity_CountedVolumeTrigger::~Asura_ServerEntity_CountedVolumeTrigger()
{
    delete [] m_pxVolumes;
}

bool Asura_ServerEntity_CountedVolumeTrigger::Contains( const Asura_Vector_3& xPosition ) const
{
    for ( int i = 0; i < m_iNumVolumes; ++i )
    {
        if ( m_pxVolumes[ i ].Contains( xPosition ) )
        {
            return true;
        }
    }

    return false;
}

#ifdef ASURA_TOOLS
void Asura_ServerEntity_CountedVolumeTrigger::ApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance )
{
    PARENT::ApplyChangesForTemplateInstance( pxInstance );

    for ( int i = 0; i < m_iNumVolumes; ++i )
    {
        Asura_Vector_3 xOldCentre = m_pxVolumes[ i ].m_xBox.CalculateCentre();
        Asura_Vector_3 xNewCentre = pxInstance->TransformPosition( xOldCentre );
        m_pxVolumes[ i ].m_xBox.Displace( xNewCentre - xOldCentre );

        m_pxVolumes[ i ].m_xOrient = pxInstance->TransformQuat( m_pxVolumes[ i ].m_xOrient );
    }
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
//
//
//////////////////////////////////////////////////////////////////////////////

int Asura_ServerEntity_CountedVolumeTrigger::GetChunkBufferSize()
{
    return PARENT::GetChunkBufferSize() +
           sizeof( int ) +
           sizeof( ChunkData ) +
           m_iNumVolumes * sizeof( Asura_OrientedVolume ) +
           NUM_TRIGGER_TYPES * sizeof( unsigned short ) +
           sizeof( unsigned short );
}

void Asura_ServerEntity_CountedVolumeTrigger::FillInChunkBuffer( char*& pcChunkBuffer )
{
    // This is not the accepted best practice for filling chunks any more.
    // The correct way is (our version) then (parent chunk) then (our chunk).
    // We have to do it the old way here because of A_SE_Trigger's constructor.

    // parent
    PARENT::FillInChunkBuffer( pcChunkBuffer );

    // version
    ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, ChunkData::CURRENT_VERSION, int );

    // chunk data
    ChunkData xChunkData;
    xChunkData.m_iListFlags = m_iListFlags;
    // bitmask, not bitfield
    ASURA_CHUNK_VERIFY( xChunkData.m_iListFlags, int );
    xChunkData.m_iNumVolumes = m_iNumVolumes;
    ASURA_CHUNK_VERIFY( xChunkData.m_iNumVolumes, int );
    xChunkData.m_iNumMessageBlocks = NUM_TRIGGER_TYPES;
    ASURA_CHUNK_VERIFY( xChunkData.m_iNumMessageBlocks, int );
    xChunkData.m_uIncrementTriggerMultiple = m_uIncrementTriggerMultiple;
    ASURA_CHUNK_VERIFY( xChunkData.m_uIncrementTriggerMultiple, unsigned int );
    xChunkData.m_uDecrementTriggerMultiple = m_uDecrementTriggerMultiple;
    ASURA_CHUNK_VERIFY( xChunkData.m_uDecrementTriggerMultiple, unsigned int );
    xChunkData.m_uTriggerCount = m_uTriggerCount;
    ASURA_CHUNK_VERIFY( xChunkData.m_uTriggerCount, unsigned int );
    xChunkData.m_uFlags = m_uFlags;
    ASURA_CHUNK_VERIFY( xChunkData.m_uFlags, unsigned int );
    xChunkData.m_uPrevCount = m_uPrevCount;
    ASURA_CHUNK_VERIFY( xChunkData.m_uPrevCount, unsigned int );
    xChunkData.m_uCurrentCount = m_uCurrentCount;
    ASURA_CHUNK_VERIFY( xChunkData.m_uCurrentCount, unsigned int );

    ASURA_CHUNK_FILL( pcChunkBuffer, xChunkData, ChunkData );

    int i;
    for ( i = 0; i < m_iNumVolumes; ++i )
    {
        ASURA_CHUNK_ARRAY_FILLANDVERIFY( pcChunkBuffer, m_pxVolumes[ i ].m_xBox,    Asura_Bounding_Box, float, 6 );
        ASURA_CHUNK_ARRAY_FILLANDVERIFY( pcChunkBuffer, m_pxVolumes[ i ].m_xOrient, Asura_Quat,         float, 4 );
    }

    for ( i = 0; i < NUM_TRIGGER_TYPES; ++i )
    {
        ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_ausMultipleTriggerMessageBlocks[ i ], unsigned short );
    }

    ASURA_CHUNK_FILL( pcChunkBuffer, 0, unsigned short );
}

void Asura_ServerEntity_CountedVolumeTrigger::Update()
{
    m_uPrevCount    = m_uCurrentCount;
    m_uCurrentCount = GetTriggerObjectCount();

    PARENT::Update();
}

bool Asura_ServerEntity_CountedVolumeTrigger::IsTriggered()
{
    m_uTriggerType = 0;

    if ( m_uCurrentCount > m_uPrevCount )
    {
        if ( m_uIncrementTriggerMultiple != 0 && m_uCurrentCount % m_uIncrementTriggerMultiple == 0 )
        {
            m_uTriggerType |= INCREMENT_TRIGGER;
        }
    }
    else if ( m_uCurrentCount < m_uPrevCount )
    {
        if ( m_uDecrementTriggerMultiple != 0 && m_uCurrentCount % m_uDecrementTriggerMultiple == 0 )
        {
            m_uTriggerType |=  DECREMENT_TRIGGER;
        }
    }
    if ( m_uCurrentCount != m_uPrevCount && m_uCurrentCount == m_uTriggerCount )
    {
        m_uTriggerType |= LIMIT_TRIGGER;
    }

    return ( m_uTriggerType != 0 );
}

bool Asura_ServerEntity_CountedVolumeTrigger::IsUnTriggered()
{
    if ( !m_bHasTriggered )
    {
        return false;
    }

    if ( m_uIncrementTriggerMultiple != 0 || m_uDecrementTriggerMultiple != 0 )
    {
        m_bHasTriggered = false;
        return true;
    }
    if ( m_uCurrentCount != m_uPrevCount && m_uCurrentCount != m_uTriggerCount )
    {
        m_bHasTriggered = false;
        return true;
    }

    return false;
}

void Asura_ServerEntity_CountedVolumeTrigger::OnTrigger()
{
    if ( m_uTriggerType & INCREMENT_TRIGGER )
    {
        if ( m_ausMultipleTriggerMessageBlocks[ INCREMENT_TRIGGER_MULTIPLE_INDEX ] != ASURA_STATICMESSAGE_EMPTYBLOCK )
        {
            Asura_StaticMessage_System::SendMessages( m_ausMultipleTriggerMessageBlocks[ INCREMENT_TRIGGER_MULTIPLE_INDEX ], Guid );
        }
    }
    else if ( m_uTriggerType & DECREMENT_TRIGGER )
    {
        if ( m_ausMultipleTriggerMessageBlocks[ DECREMENT_TRIGGER_MULTIPLE_INDEX ] != ASURA_STATICMESSAGE_EMPTYBLOCK )
        {
            Asura_StaticMessage_System::SendMessages( m_ausMultipleTriggerMessageBlocks[ DECREMENT_TRIGGER_MULTIPLE_INDEX ], Guid );
        }
    }
    if ( m_uTriggerType & LIMIT_TRIGGER )
    {
        if ( m_ausMultipleTriggerMessageBlocks[ LIMIT_TRIGGER_INDEX ] != ASURA_STATICMESSAGE_EMPTYBLOCK )
        {
            Asura_StaticMessage_System::SendMessages( m_ausMultipleTriggerMessageBlocks[ LIMIT_TRIGGER_INDEX ], Guid );
        }
    }

    m_bHasTriggered = true;
    m_uTriggerType  = 0;
}

void Asura_ServerEntity_CountedVolumeTrigger::OnUnTrigger()
{
    Asura_ServerEntity_Trigger::OnUnTrigger();
    if ( m_bTriggerOnce )
    {
        // deactivate this trigger so it doesn't get triggered again
        OnDeactivation();
    }
}

//////////////////////////////////////////////////////////////////////////////
//
//
//
//////////////////////////////////////////////////////////////////////////////
