//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_DebugMessageTrigger.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     trigger that displays a debug message on screen when it it triggered.
//     The system is incredibly simple, which is all we need at the moment. 
//     It can easily be modified if more functionality is necessary.
//
// History:
//
//     (JCS  - 08.12.2005) - Created as a copy of Asura_ServerEntity_CountedTrigger.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_DEBUG_MESSAGE_TRIGGER

#define COMMAND_DEBUGMESSAGETRIGGERENABLE		L"DebugMessageTriggerEnable"

// This file's header.
#include "Entity/Triggers/Asura_ServerEntity_DebugMessageTrigger.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#if defined(ASURA_TOOLS)
#include "Entity/Tools_Entity_Data.h"
#endif // ASURA_TOOLS
#include "Camera/Asura_Camera.h"
#ifdef ASURA_SUPPORT_ENTITYWATCH
#include "Entity/Asura_EntityWatch_System.h"
#endif
#include "Interface/Asura_CommandConsole.h"
#include "Interface/Asura_CommandConsole_IO.h"
#include "Network/Asura_Message.h"
#include "Core/Asura_Timers.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Colour.h"

static const int ASURA_SERVERENTITY_DEBUGMESSAGETRIGGER_VERSION_NUMBER = 5;

bool Asura_ServerEntity_DebugMessageTrigger::s_bEnableMessages = true;


Asura_ServerEntity_DebugMessageTrigger::Asura_ServerEntity_DebugMessageTrigger( void )
:Asura_ServerEntity_Trigger()
{
	Asura_DebugMessageTrigger_Messages::Reset();

	Classification = AsuraEntityClass_DebugMessageTrigger;
	InitData();
}

u_int Asura_ServerEntity_DebugMessageTrigger::m_uStringDisplayCount = 0;

Asura_ServerEntity_DebugMessageTrigger::Asura_ServerEntity_DebugMessageTrigger( Asura_Guid uGuid ) : Asura_ServerEntity_Trigger( uGuid )
{
	Asura_DebugMessageTrigger_Messages::Reset();

	Classification = AsuraEntityClass_DebugMessageTrigger;
	InitData();
}

void Asura_ServerEntity_DebugMessageTrigger::InitData(void)
{
	m_fR = 255.0f;
	m_fG = 255.0f;
	m_fB = 255.0f;

	m_xStringToDisplay = "Text";
	m_fTimeToDisplayString = 10.0f;
	m_fPreDisplayTime = 0.0f;

	m_bDisplayingString = false;
	m_fPreDisplayTimer = 0.0f;
	m_fTimer = 0.0f;
	m_uDisplayOffset = 0;
	m_iDrawInWorld = 0;
}

bool Asura_ServerEntity_DebugMessageTrigger::HandleMessage(const Asura_Message* pxMessage)
{
	if( IsDestroyed() ) { return true; } // ignore all further messages?

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
		case ASURA_MSG_RESET:
			{
				m_fTimer = 0.0f;
				m_bDisplayingString = false;
				return true;
			}
		}
	}

	return PARENT::HandleMessage(pxMessage);
}


bool Asura_ServerEntity_DebugMessageTrigger::IsTriggered()
{
	if (m_bDisplayingString)
	{
		return true;
	}

	return false;
}

void Asura_ServerEntity_DebugMessageTrigger::OnTrigger()
{
// Set up the message's variables for display:
	m_bDisplayingString = true;
	m_fTimer = 0.0f;
	m_fPreDisplayTimer = 0.0f;

	// Loop round the variable slots if >10 have been / are being displayed
	m_uStringDisplayCount++;
	if(m_uStringDisplayCount > 10)
	{
		m_uStringDisplayCount = 0;
	}
	m_uDisplayOffset = m_uStringDisplayCount;

	PARENT::OnTrigger();
}

void Asura_ServerEntity_DebugMessageTrigger::OnUnTrigger()
{
	// Turn off the message
	m_bDisplayingString = false;
	m_fTimer = 0.0f;
	m_fPreDisplayTimer = 0.0f;

	PARENT::OnUnTrigger();
}



//virtual function overides for saving

void Asura_ServerEntity_DebugMessageTrigger::Create( Asura_Guid uGuid, Asura_Chunk_Stream& xStream )
{
	Asura_ServerEntity_DebugMessageTrigger* pxTrigger = new Asura_ServerEntity_DebugMessageTrigger( uGuid );
	if( pxTrigger )
	{
		pxTrigger->ReadFromChunkStream(xStream);
	}
}

void Asura_ServerEntity_DebugMessageTrigger::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	PARENT::ReadFromChunkStream( xStream );

	int iVersion;
	xStream >> iVersion;
	Asura_Assert( iVersion <= ASURA_SERVERENTITY_DEBUGMESSAGETRIGGER_VERSION_NUMBER, "" );

	if ( iVersion >= 1 )
	{
		xStream >> m_fTimeToDisplayString;
		xStream >> m_xStringToDisplay;
	}
	if ( iVersion >= 2 )
	{
		xStream >> m_fPreDisplayTime;
	}
	if ( iVersion >= 3 )
	{
		xStream >> m_fR;
		xStream >> m_fG;
		xStream >> m_fB;
	}
	if ( iVersion >= 4 )
	{
		xStream >> m_iDrawInWorld;
	}
	if ( iVersion >= 5 )
	{
		xStream >> m_uDisplayOffset;
		xStream >> m_fTimer;
		xStream >> m_fPreDisplayTimer;
		xStream >> m_bDisplayingString;
	}
}

void Asura_ServerEntity_DebugMessageTrigger::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	PARENT::WriteToChunkStream( xStream );

	const int iVersion = ASURA_SERVERENTITY_DEBUGMESSAGETRIGGER_VERSION_NUMBER;
	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/

	xStream << iVersion;
	
	xStream << m_fTimeToDisplayString;
	xStream << m_xStringToDisplay;
	xStream << m_fPreDisplayTime;
	xStream << m_fR;
	xStream << m_fG;
	xStream << m_fB;
	xStream << m_iDrawInWorld;
	xStream << m_uDisplayOffset;
	xStream << m_fTimer;
	xStream << m_fPreDisplayTimer;
	xStream << m_bDisplayingString;

}

void Asura_ServerEntity_DebugMessageTrigger::Update()
{
	// note that we don't call the parent update function here as we have many different triggers
	if ( IsDestroyed() )
	{
		return;
	}

	if(m_bDisplayingString)
	{
		if(m_fPreDisplayTimer < m_fPreDisplayTime)
		{
			m_fPreDisplayTimer += Asura_Timers::GetGameTimeSlice();
		}
		else
		{
			m_fTimer += Asura_Timers::GetGameTimeSlice();
			if ( m_fTimer >= m_fTimeToDisplayString )
			{
				m_bDisplayingString = false;
			}
			// JAT 030506 if m_fTimeToDisplayString is < 0 then display string forever
			if ( m_fTimeToDisplayString < 0.0f )
			{
				m_bDisplayingString = true;
			}

			if(s_bEnableMessages)
			{
				u_int uColour = 0;
				Asura_RGBA xColour( m_fR/255.0f, m_fG/255.0f, m_fB/255.0f, 1.0f );
				uColour = xColour.ToUIntARGB();

				// DMR 01/09/06
				// Can't do this - got to render them late, after the FSFX - about when the HUD is renderered
				// We were getting lucky before, that the text wasn't flushed until later, but that isn't 
				// always the case
				//
				// Asura_Font_System::Print( 100,100 + (float)(20 * m_uDisplayOffset),1.0f, uColour,0,m_xStringToDisplay.GetString());
#ifdef ASURA_TOOLS
				Asura_Vector_3 xWorldPos = ToolsData->Position;
				xWorldPos.SetY(xWorldPos.GetY() - 0.5f);	// Move Text to middle of tool representation.
				Asura_Camera::ProjectWorldCoordsToScreenSpace(xWorldPos, true);
				if(m_iDrawInWorld)
				{	
					if(xWorldPos.GetZ() >= 0.0f)
					Asura_DebugMessageTrigger_Messages::BufferDebugMessage(xWorldPos.GetX(), xWorldPos.GetY(), uColour,m_xStringToDisplay.GetString());
				}
				else
#endif //ASURA_TOOLS
				{
					Asura_DebugMessageTrigger_Messages::BufferDebugMessage(100, 100 + (float)(20 * m_uDisplayOffset), uColour,m_xStringToDisplay.GetString());
				}
			}
		}
	}
}

#if defined ASURA_SUPPORT_COMMAND_CONSOLE

void Asura_ServerEntity_DebugMessageTrigger::AddConsoleCommands( void )
{
	Asura_CommandConsole::AddCmd( COMMAND_DEBUGMESSAGETRIGGERENABLE, ConsoleCmd_DebugMessageTriggerEnable, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}

bool Asura_ServerEntity_DebugMessageTrigger::ConsoleCmd_DebugMessageTriggerEnable( bool bEnable, bool bInterrogate, bool bHelp )
{
	//Enable/Disable displaying of the text
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( L"Sets whether debug message trigger text should be displayed" );
		Asura_CommandConsole::Printf( L"Syntax: " COMMAND_DEBUGMESSAGETRIGGERENABLE L" <true | false>" );
	}
	else
	{
		if ( !bInterrogate )
		{
			s_bEnableMessages = bEnable;
		}
		Asura_CommandConsole::Printf( L"Debug Message Trigger: %ls.", s_bEnableMessages ? L"enabled" : L"disabled" );
	}
	return true;
}

#endif // defined ASURA_SUPPORT_COMMAND_CONSOLE

////////////////////////////////////////
// Asura_DebugMessageTrigger_Messages //
///////////////////////////////////////

// Debug messages now buffered, and rendered later
int								  Asura_DebugMessageTrigger_Messages::s_iNumDebugMessages = 0;
Asura_DebugMessageTrigger_Message Asura_DebugMessageTrigger_Messages::s_axDebugMessages[ASURA_DEBUGMESSAGETRIGGER_MAXMESSAGES];
int								  Asura_DebugMessageTrigger_Messages::s_iFontHandle = 0;


void Asura_DebugMessageTrigger_Messages::Reset()
{
	s_iNumDebugMessages = 0;
}

void Asura_DebugMessageTrigger_Messages::Render( int iFontHandle )
{
	SetFont( iFontHandle );
	Render();
}

void Asura_DebugMessageTrigger_Messages::Render()
{
	if (s_iNumDebugMessages>0)
	{	
		Asura_Font_System::SetFont(s_iFontHandle);
		
		for (int i=0;i<s_iNumDebugMessages;i++)
		{
			Asura_DebugMessageTrigger_Message *pxMsg = &s_axDebugMessages[i];
			Asura_Font_System::Print( pxMsg->fX, pxMsg->fY, 1.0f, pxMsg->uColor,0,pxMsg->szString);		
		}

		Asura_Render::FlushDisplay();
		s_iNumDebugMessages = 0;
	}
}


void Asura_DebugMessageTrigger_Messages::BufferDebugMessage(float fX, float fY, uint32 uColor, char *szString)
{
	if (s_iNumDebugMessages<ASURA_DEBUGMESSAGETRIGGER_MAXMESSAGES)
	{
		Asura_DebugMessageTrigger_Message *pxMsg = &s_axDebugMessages[s_iNumDebugMessages];
		pxMsg->fX = fX;
		pxMsg->fY = fY;
		pxMsg->uColor = uColor;
		pxMsg->szString = szString;
		s_iNumDebugMessages++;
	}	
}


#endif //ASURA_USE_DEBUG_MESSAGE_TRIGGER



