//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_TimelineTrigger.cpp
//
// Copyright (C) 2008 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Entity/Triggers/Asura_ServerEntity_TimelineTrigger.h"

// asura
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#ifdef ASURA_SUPPORT_ENTITYWATCH
#include "Entity/Asura_EntityWatch_System.h"
#endif
#include "Entity/Asura_StaticMessage_System.h"
#include "Entity/Triggers/Asura_ServerEntity_TimeTrigger.h"

// tools
#ifdef ASURA_TOOLS
#include "Entity/Tools_Message_Editing.h"
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_TimelineTrigger_MessageBlock
//
//////////////////////////////////////////////////////////////////////////////

void Asura_TimelineTrigger_MessageBlock::SetDefaultValues()
{
	m_uHashID = 0;
	m_fTime = 0;
	m_usMessageBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;
	m_usFlags = 0;
	m_bActive = 1;

	#ifdef ASURA_TOOLS
	m_xName = "";
	m_bSelected = false;
	#endif
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_TimelineTrigger
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS
Asura_Hashed_List<Asura_ServerEntity_TimelineTrigger> Asura_ServerEntity_TimelineTrigger::s_xList(5);
#endif

Asura_ServerEntity_TimelineTrigger::Asura_ServerEntity_TimelineTrigger()
: Asura_ServerEntity_Activatable()
, m_uNumberOfMessageBlocks( 0 )
, m_pxMessageBlocks( NULL )
, m_fTimer( 0.0f )
, m_uTimelineTriggerFlags( 0 )
, m_fMaxTriggerTime( -1.0f )
{
	Classification = AsuraEntityClass_TimelineTrigger;

	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif
}

Asura_ServerEntity_TimelineTrigger::Asura_ServerEntity_TimelineTrigger( const Asura_Guid uGuid )
: Asura_ServerEntity_Activatable( uGuid )
, m_uNumberOfMessageBlocks( 0 )
, m_pxMessageBlocks( NULL )
, m_fTimer( 0.0f )
, m_uTimelineTriggerFlags( 0 )
, m_fMaxTriggerTime( -1.0f )
{
	Classification = AsuraEntityClass_TimelineTrigger;

	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif
}

Asura_ServerEntity_TimelineTrigger::~Asura_ServerEntity_TimelineTrigger()
{
	#ifdef ASURA_TOOLS
	s_xList.RemoveFromList(this);
	#endif
	delete [] m_pxMessageBlocks;
}

#ifdef ASURA_TOOLS
int Asura_ServerEntity_TimelineTrigger::GetToolsDataSize() const
{
	int iSize = 0;
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		iSize += m_pxMessageBlocks[i].m_xName.GetSizeInInts()*4;
	}
	return iSize;
}
#endif //ASURA_TOOLS

void Asura_ServerEntity_TimelineTrigger::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	int iVersion = xStream.PeekInt();
	// From v1000 onwards we have corrected the order of version and parent.
	if ( iVersion < 1000 )
	{
		Asura_ServerEntity_Activatable::ReadFromChunkStream( xStream );
		xStream >> iVersion;
	}
	else
	{
		xStream >> iVersion;
		Asura_ServerEntity_Activatable::ReadFromChunkStream( xStream );
	}

	if ( iVersion == CURRENT_CHUNK_VERSION || iVersion == 1 || iVersion == 0 )
	{
		xStream >> m_fTimer;
		xStream >> m_uNumberOfMessageBlocks;

		m_pxMessageBlocks = NULL;
		if ( m_uNumberOfMessageBlocks )
		{
			m_pxMessageBlocks = new Asura_TimelineTrigger_MessageBlock[m_uNumberOfMessageBlocks];
			for(unsigned i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				xStream >> m_pxMessageBlocks[i].m_uHashID;
				xStream >> m_pxMessageBlocks[i].m_fTime;
				xStream >> m_pxMessageBlocks[i].m_usMessageBlock;
				xStream >> m_pxMessageBlocks[i].m_usFlags;
				#ifdef ASURA_TOOLS
				m_pxMessageBlocks[i].m_bSelected = false;
				#endif
			}
		}
		InvalidateMaxTriggerTime();

		if(iVersion >= 1)
		{
			xStream >> m_uTimelineTriggerFlags;
		}
		else
		{
			m_uTimelineTriggerFlags = 0;
		}

		int iToolsDataSize;
		xStream >> iToolsDataSize;

		#ifdef ASURA_TOOLS
		if(iToolsDataSize)
		{
			for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				xStream >> m_pxMessageBlocks[i].m_xName;
			}
		}
		else
		{
			for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				m_pxMessageBlocks[i].m_xName = "";
			}
		}
		#else
		//not in tools , so just skip this
		xStream.Skip( iToolsDataSize );
		#endif // ASURA_TOOLS	
	}
	else
	{
		Asura_Assertf( false, ( "Asura_ServerEntity_TimelineTrigger::ReadFromChunkStream - unhandled chunk version (%d)", iVersion ) );
	}
}

void Asura_ServerEntity_TimelineTrigger::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	/*
	This entity type can be used in template files , so any changes to WriteToChunkStream may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/

	xStream << CURRENT_CHUNK_VERSION;

	Asura_ServerEntity_Activatable::WriteToChunkStream( xStream );

	xStream << m_fTimer;
	xStream << m_uNumberOfMessageBlocks;

	for(unsigned i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		xStream << m_pxMessageBlocks[i].m_uHashID;
		xStream << m_pxMessageBlocks[i].m_fTime;
		xStream << m_pxMessageBlocks[i].m_usMessageBlock;
		
		Asura_WriteToStream_Bitfield_Wrapper<unsigned short>	xFlagsBitfield ( m_pxMessageBlocks[i].m_usFlags );
		xFlagsBitfield.Add( m_pxMessageBlocks[i].m_bActive,		1 );
		xFlagsBitfield.Add( m_pxMessageBlocks[i].m_usSpareBits, 15);

		xStream << xFlagsBitfield;
	}

	xStream << m_uTimelineTriggerFlags;

	#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())//exported levels won't want the tools data
	{
		//save out data only needed by tools
		xStream << GetToolsDataSize();

		for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
		{
			xStream << m_pxMessageBlocks[i].m_xName;
		}
	}
	else
	#endif
	{
		//Either not in tools , or doing a level export.
		//In either case , we don't want the tools data.
		xStream << 0;
	}
}

void Asura_ServerEntity_TimelineTrigger::Update()
{
	if(IsActive() && m_uNumberOfMessageBlocks)
	{
		SetTimer(m_fTimer + Asura_Timers::GetGameTimeSlice());
	}

	Asura_ServerEntity_Activatable::Update();
}

void Asura_ServerEntity_TimelineTrigger::SetTimer(float fTimer)
{
	float fOldTime = m_fTimer;
	m_fTimer = fTimer;

	//Trigger any messages beteen the previous timer and the new time ,
	//if active , and the timer is moved forwards
	if(IsActive() && m_uNumberOfMessageBlocks && fTimer > fOldTime)
	{
		if(!IsLooped())
		{
			if(m_fTimer > GetMaxTriggerTime())
			{
				//If timer is past last trigger ,then deactivate now.
				//Doing this before the messages are processed in case the trigger wants to reset itself
				SetInternalActiveFlag(false);
			}
		}
		
		//trigger any messages
		TriggerMessages(fOldTime,m_fTimer);

		float fMaxTriggerTime = GetMaxTriggerTime();
		if(IsLooped() && IsActive() && m_fTimer > fMaxTriggerTime)
		{
			//wrap the timer
			if(fMaxTriggerTime > 0.0f)
			{
				m_fTimer = Asura_Maths::FloatMod(m_fTimer,fMaxTriggerTime);
				//trigger messages at the start of the loop as needed
				TriggerMessages(0.0f,m_fTimer);
			}
			else
			{
				m_fTimer = 0.0f;
			}
		}
	}
}

void Asura_ServerEntity_TimelineTrigger::TriggerMessages(float fStartTime,float fEndTime)
{
	Asura_TimelineTrigger_MessageBlock* pxMessage = m_pxMessageBlocks;
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++,pxMessage++)
	{
		if(pxMessage->m_fTime < fStartTime) continue;
		
		if(pxMessage->m_fTime < fEndTime)
		{
			//this block was triggered this frame.
			if(pxMessage->m_usMessageBlock != ASURA_STATICMESSAGE_EMPTYBLOCK && pxMessage->m_bActive)
			{
				Asura_StaticMessage_System::SendMessages(pxMessage->m_usMessageBlock,Guid);
			}
		}
	}
}


float Asura_ServerEntity_TimelineTrigger::GetMaxTriggerTime() const
{
	if(m_fMaxTriggerTime < 0.0f)
	{
		//max trigger time not calculated yet
		m_fMaxTriggerTime = 0.0f;
		const Asura_TimelineTrigger_MessageBlock* pxMessage = m_pxMessageBlocks;
		for(u_int i=0;i<m_uNumberOfMessageBlocks;i++,pxMessage++)
		{
			m_fMaxTriggerTime = Asura_Maths::Max(m_fMaxTriggerTime,pxMessage->m_fTime);
		}
	}
	return m_fMaxTriggerTime;
}

void Asura_ServerEntity_TimelineTrigger::SetTriggerTime(u_int uOutput,float fTime)
{
	Asura_Assert(uOutput < m_uNumberOfMessageBlocks,"");
	Asura_TimelineTrigger_MessageBlock* pxMessage = &m_pxMessageBlocks[uOutput];

	//set the new trigger time
	float fOldTriggerTime = pxMessage->m_fTime;
	pxMessage->m_fTime = fTime;
	InvalidateMaxTriggerTime();

	if(IsActive())
	{
		//trigger the messages if the time has been brought before the current timer
		if(fTime < m_fTimer && m_fTimer <= fOldTriggerTime)
		{
	  		if(pxMessage->m_usMessageBlock != ASURA_STATICMESSAGE_EMPTYBLOCK && pxMessage->m_bActive)
	  		{
	  			Asura_StaticMessage_System::SendMessages(pxMessage->m_usMessageBlock,Guid);
	  		}
		
		}
	}
}





bool Asura_ServerEntity_TimelineTrigger::HandleMessage(const Asura_Message* pxMessage)
{
	if (!pxMessage) return false;

	switch(pxMessage->GetMsgID())
	{

		//these use the same message structure  so deal with them at the same time
		case ASURA_MSG_TIMELINETRIGGER_ENABLEMESSAGE:
		case ASURA_MSG_TIMELINETRIGGER_DISABLEMESSAGE :
		{
			const Asura_Message_TimelineTrigger* pxMessageData = reinterpret_cast
				<const Asura_Message_TimelineTrigger*>(pxMessage->GetData());
			
			//search for a matching message
			for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				if(m_pxMessageBlocks[i].m_uHashID == pxMessageData->m_uMessageHash)
				{
					m_pxMessageBlocks[i].m_bActive = (pxMessage->GetMsgID()==ASURA_MSG_TIMELINETRIGGER_ENABLEMESSAGE);

					break;
				}
			}

			return true;
		}

		case ASURA_MSG_SETTIMER :
		{
			const Asura_Message_SetTimer* pxMessageData = reinterpret_cast
				<const Asura_Message_SetTimer*>(pxMessage->GetData());

			SetTimer(pxMessageData->m_fTimer);
			return true;
				
		}

		case ASURA_MSG_TIMELINETRIGGER_SETTRIGGERTIME :
		{
			const Asura_Message_TimelineTrigger_SetTriggerTime* pxMessageData = reinterpret_cast
				<const Asura_Message_TimelineTrigger_SetTriggerTime*>(pxMessage->GetData());

			//search for a matching message
			for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				if(m_pxMessageBlocks[i].m_uHashID == pxMessageData->m_uMessageHash)
				{
					SetTriggerTime(i,pxMessageData->m_fTriggerTime);
					break;
				}
			}

			return true;
		}

		case ASURA_MSG_PAUSETIMER :
		{
			//pause by deactivating the timer
			SetInternalActiveFlag(false);
			return true;
		}

		case ASURA_MSG_UNPAUSETIMER :
		{
			//unpause by deactivating the timer
			SetInternalActiveFlag(true);
			return true;
		}

		
		case ASURA_MSG_RESET:
			m_fTimer = 0;
			SetInternalActiveFlag(true);
			return true;

		default :
			break;
	}

	// Pass to parent.
	return Asura_ServerEntity_Activatable::HandleMessage(pxMessage);
}



#ifdef ASURA_TOOLS
void Asura_ServerEntity_TimelineTrigger::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	Asura_ServerEntity_Activatable::ApplyChangesForTemplateInstance(pxInstance);

	//update the message blocks
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		m_pxMessageBlocks[i].m_usMessageBlock = pxInstance->ConvertMessageBlock(m_pxMessageBlocks[i].m_usMessageBlock);
	}

}
#endif // ASURA_TOOLS



#ifdef ASURA_TOOLS
const char* Asura_ServerEntity_TimelineTrigger::GetMessageDescription(const Asura_StaticMessage* pxMessage)
{
	if(pxMessage->m_usMsgID != ASURA_MSG_TIMELINETRIGGER_ENABLEMESSAGE &&
		pxMessage->m_usMsgID != ASURA_MSG_TIMELINETRIGGER_DISABLEMESSAGE &&
		pxMessage->m_usMsgID != ASURA_MSG_TIMELINETRIGGER_SETTRIGGERTIME) return 0;

	const Asura_Message_TimelineTrigger* pxData = (const Asura_Message_TimelineTrigger*) pxMessage->m_pData;


	//Find the entity being triggered
	Asura_ServerEntity_TimelineTrigger* pxEntity = FindEntity(Tools_Message_Editing::GetFinalDestinationGuid(pxMessage));
	if(!pxEntity) return 0;

	static char szText[100];
	for(u_int i=0;i<pxEntity->m_uNumberOfMessageBlocks;i++)
	{
		if(pxData->m_uMessageHash == pxEntity->m_pxMessageBlocks[i].m_uHashID)
		{
			if(pxMessage->m_usMsgID == ASURA_MSG_TIMELINETRIGGER_ENABLEMESSAGE)
				Asura_PrintfToBfr(szText,"Enable Msg(%s)",100,true,pxEntity->m_pxMessageBlocks[i].m_xName.GetString());
			else if(pxMessage->m_usMsgID == ASURA_MSG_TIMELINETRIGGER_DISABLEMESSAGE)
				Asura_PrintfToBfr(szText,"Disable Msg(%s)",100,true,pxEntity->m_pxMessageBlocks[i].m_xName.GetString());
			else
				Asura_PrintfToBfr(szText,"Set Msg Trigger Time(%s)",100,true,pxEntity->m_pxMessageBlocks[i].m_xName.GetString());
			return szText;
		}
	}
	return 0;
}
#endif //ASURA_TOOLS

#ifdef ASURA_SUPPORT_ENTITYWATCH
void Asura_ServerEntity_TimelineTrigger::DisplayEntityWatchText()
{
	Asura_ServerEntity_Activatable::DisplayEntityWatchText();
	Asura_EntityWatch_System::PrintText(16,"Timer : %.3f",m_fTimer);
	Asura_EntityWatch_System::NextLine();

	Asura_TimelineTrigger_MessageBlock* pxMessage = m_pxMessageBlocks;
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++,pxMessage++)
	{
		Asura_EntityWatch_System::PrintText(0,"Msg %d (%.3f) %s ",i,pxMessage->m_fTime,pxMessage->m_bActive ? "On " : "Off");
	}
}
#endif

#ifdef ASURA_TOOLS
void Asura_ServerEntity_TimelineTrigger::GetCustomEventName(int iEventIndex, Asura_String& xEventName)
{
	xEventName = "";
}
#endif

void Asura_ServerEntity_TimelineTrigger::CreateFromChunkStream( const Asura_Guid uGuid, Asura_Chunk_Stream& xStream )
{
	Asura_ServerEntity_TimelineTrigger* pxTT = new Asura_ServerEntity_TimelineTrigger( uGuid );
	if ( pxTT )
	{
		pxTT->ReadFromChunkStream( xStream );
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
