//
// Asura_FSFX_Component.cpp
//
// Created by JGP 02/02/06
// 
// Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_FSFX

#include "Asura_FSFX_Component.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"

#include "Core/Asura_MaterialFlags.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_TextureManagement.h"


#ifdef ASURA_TOOLS
#include "Texturing/Tools_PickTexture.h"
#endif

#ifdef ASURA_TOOLS
// all supported platforms enabled by default
#define SET_PLATFORM_ACTIVE_FLAGS() m_iPlatformsActive = GetPlatformsSupportedID(); 
#else
// all platforms...
#define SET_PLATFORM_ACTIVE_FLAGS() m_iPlatformsActive = ( 1 << NUM_PLATFORMS ) - 1; 
#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Create of type...
///////////////////////////////////////////////////////////////////////////////////////


/*
	Given a type field, create a component of the correct
	corrisponding class
*/
Asura_FSFX_Component* Asura_FSFX_Component::CreateOfType( const Type eType )
{
	switch( eType )
	{
		case FSFX_TEST:
		{
			return new Asura_FSFX_Component_Test;
		}
		case FSFX_BRIGHTNESS_AND_CONTRAST:
		{
			return new Asura_FSFX_Component_BrightnessAndContrast;
		}
		case FSFX_NOISE:
		{
			return new Asura_FSFX_Component_Noise;
		}
		case FSFX_GAUSSIAN_BLUR:
		{
			return new Asura_FSFX_Component_GaussianBlur;
		}
		case FSFX_RADIAL_BLUR:
		{
			return new Asura_FSFX_Component_RadialBlur;
		}
		case FSFX_DEPTH_OF_FIELD:
		{
			return new Asura_FSFX_Component_DepthOfField;
		}
		case FSFX_BLOOM:
		{
			return new Asura_FSFX_Component_Bloom;
		}
		case FSFX_GAMMA:
		{
			return new Asura_FSFX_Component_Gamma;
		}
		case FSFX_RGB_GAMMA:
		{
			return new Asura_FSFX_Component_RGBGamma;
		}
		case FSFX_DESATURATE_AND_TINT:
		{
			return( new Asura_FSFX_Component_DesaturateAndTint);
		}
		case FSFX_SM30_BLOOM:
		{
			return( new Asura_FSFX_Component_SM30_Bloom);
		}
		case FSFX_COLOUR_MAP:
		{
			return new Asura_FSFX_Component_ColourMap;
		}
		case FSFX_SM30_DEPTH_OF_FIELD:
		{
			return( new Asura_FSFX_Component_SM30_DepthOfField);
		}
		case FSFX_SM30_FISHEYE:
		{
			return( new Asura_FSFX_Component_SM30_FishEye);
		}
		case FSFX_SM30_DOUBLEVISION:
		{
			return( new Asura_FSFX_Component_SM30_DoubleVision);
		}
		case FSFX_SM30_MOTION_BLUR:
		{
			return( new Asura_FSFX_Component_SM30_MotionBlur);
		}

		default:
		{
			Asura_FSFX_Component* const pxNewComponent = Project_CreateOfType( eType );
			Asura_Assert( pxNewComponent, "Create rule not set for this type in Asura_FSFX_Component::CreateOfType" );

			return pxNewComponent;
		}
	}
}


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component::Asura_FSFX_Component()
{
#ifdef ASURA_TOOLS
	m_bExpanded = false; 
#endif // ASURA_TOOLS
}	


/*
	Get whether the effect is valid and active
	for the given platform
*/
bool Asura_FSFX_Component::IsPlatformEnabled( const Platform m_ePlatform )
{
#ifdef ASURA_TOOLS
	// on tools build check a platform is supported as well as enabled
	return ( ( GetPlatformsSupported() & m_iPlatformsActive & ( 1 << m_ePlatform ) ) != 0 );
#else
	// otherwise just check it's enabled
	return ( ( m_iPlatformsActive & ( 1 << m_ePlatform ) ) != 0 );
#endif // ASURA_TOOLS
}


/*
	The standard header is at the start of each 
	component. This gives a version number, a 
	size and a platform active flags. The platform
	active flags are simply set in the class, everything
	else is returned to the calling function for 
	processing (eg for loading out of date data and 
	checking size)
*/
void Asura_FSFX_Component::ReadStandardHeader( Asura_Chunk_Stream& xStream, int& iVersion, int& iSize )
{
	xStream >> iVersion;
	xStream >> iSize;
	xStream >> m_iPlatformsActive;
}


#ifdef ASURA_TOOLS

void Asura_FSFX_Component::WriteStandardHeader( Asura_Chunk_Stream& xStream, const int iSize ) const
{
	xStream << GetVersion();
	xStream << iSize;
	xStream << m_iPlatformsActive;
}

#endif // ASURA_TOOLS

u_int Asura_FSFX_Component::GetTextureCount( void) const
{
	return( 0 );
}

int Asura_FSFX_Component::GetTextureIndex( const u_int uTemplateTextureIndex) const
{
#ifdef ASURA_TOOLS
	// DLH - 05 April 2006 14:34:18 
	Asura_Assertf( false, ("FSFX '%s' does not support textures, so do not ask for them.", this->GetTypeDesc()));
#endif

	return( ASURA_TEXTURE_INVALID);
}

const char* Asura_FSFX_Component::GetTextureName( const u_int uTemplateTextureIndex) const
{
#ifdef ASURA_TOOLS
	Asura_Assertf( false, ("FSFX '%s' does not support textures, so do not ask for them.", this->GetTypeDesc()));
#endif

	return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_Test functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_Test::Asura_FSFX_Component_Test() 
{ 
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fRed = m_fGreen = m_fBlue = 0.0f; 
	m_bSubtractive = false; 
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_Test::CreateCopy() 
{
	Asura_FSFX_Component_Test* pxComp = new Asura_FSFX_Component_Test;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_Test::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_bSubtractive;
		xStream >> m_fRed;
		xStream >> m_fGreen;
		xStream >> m_fBlue;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_Test version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_Test::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_Test::Serialize( Asura_Chunk_Stream& xStream ) const
{

	int iSize = sizeof( bool ) + 3 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_bSubtractive;
	xStream << m_fRed;
	xStream << m_fGreen;
	xStream << m_fBlue;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_Test::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_BrightnessAndContrast functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_BrightnessAndContrast::Asura_FSFX_Component_BrightnessAndContrast() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fBrightness = m_fContrast = 0.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_BrightnessAndContrast::CreateCopy() 
{
	Asura_FSFX_Component_BrightnessAndContrast* pxComp = new Asura_FSFX_Component_BrightnessAndContrast;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_BrightnessAndContrast::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fBrightness;
		xStream >> m_fContrast;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_BrightnessAndContrast version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_BrightnessAndContrast::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_BrightnessAndContrast::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 2 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << m_fBrightness;
	xStream << m_fContrast;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_BrightnessAndContrast::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_DesaturateAndTint functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_DesaturateAndTint::Asura_FSFX_Component_DesaturateAndTint() 
{
	SET_PLATFORM_ACTIVE_FLAGS();
	
	this->m_fStrength = 50.0f;
	this->m_fIntensityR = 100.0f;
	this->m_fIntensityG = 100.0f;
	this->m_fIntensityB = 100.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_DesaturateAndTint::CreateCopy() 
{
	Asura_FSFX_Component_DesaturateAndTint* pxComp = new Asura_FSFX_Component_DesaturateAndTint;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_DesaturateAndTint::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> this->m_fStrength;
		xStream >> this->m_fIntensityR;
		xStream >> this->m_fIntensityG;
		xStream >> this->m_fIntensityB;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_DesaturateAndTint version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_DesaturateAndTint::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_DesaturateAndTint::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 4 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << this->m_fStrength;
	xStream << this->m_fIntensityR;
	xStream << this->m_fIntensityG;
	xStream << this->m_fIntensityB;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_BrightnessAndContrast::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_Noise functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_Noise::Asura_FSFX_Component_Noise() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fAmount = 0.0f;
	
	this->m_uTextureHash = ASURA_HASH_ID_UNSET;
	this->m_iTextureID = ASURA_TEXTURE_NONE;
	this->m_xTextureFilename = "";
	this->m_uTextureTileX = 32;
	this->m_uTextureTileY = 32;
}

/*
	SetTexture
*/
void Asura_FSFX_Component_Noise::SetTexture()
{
#if defined( ASURA_TOOLS )
	bool bWasNew = false;
	Asura_Assert( this->m_xTextureFilename.GetLength() > 0, "Invalid texture name on FSFX noise component" );
	if ( ASURA_TEXTURE_INVALID == this->m_iTextureID )
	{
		const int iPrevTexResetGrp = Asura_Texture_Management::GetTextureResetGroup();
		Asura_Texture_Management::SetTextureResetGroup( ASURA_TEXTURE_LOADED_PERMANENT );

		this->m_iTextureID = Asura_Texture_Management::AddTexture( this->m_xTextureFilename, &bWasNew );
		// Make the hash value 
		this->m_uTextureHash = Asura_GetHashID_IgnoreExtension( this->m_xTextureFilename.GetString() );
		
	    Asura_Texture_Management::SetTextureResetGroup( iPrevTexResetGrp );
	}
    
	if ( ASURA_TEXTURE_NONE == this->m_iTextureID )
	{
		this->m_uTextureHash = ASURA_HASH_ID_UNSET;
		return;
	}

	if ( ASURA_TEXTURE_INVALID == this->m_iTextureID )
	{
		this->m_uTextureHash = ASURA_HASH_ID_UNSET;
		return;
	}
	
	if( true == bWasNew )
	{
		Asura_Texture_Management::LoadTextures();
	}
#else
	if ( this->m_uTextureHash == ASURA_HASH_ID_UNSET )
	{
		this->m_iTextureID = ASURA_TEXTURE_NONE;
		return;
	}
	this->m_iTextureID = Asura_Texture_Management::GetIndexFromHashID( this->m_uTextureHash );
	Asura_Assert( this->m_iTextureID != ASURA_TEXTURE_INVALID, "" );
#endif // #if defined( ASURA_TOOLS )
}

u_int Asura_FSFX_Component_Noise::GetTextureCount( void ) const
{
	return ( 1 );
}

int Asura_FSFX_Component_Noise::GetTextureIndex( const u_int uTemplateTextureIndex ) const
{
#if defined( ASURA_TOOLS)
	Asura_Assertf( uTemplateTextureIndex < this->GetTextureCount(), ("FSFX '%s' only supports %d textures. %d is invalid.", (const_cast<Asura_FSFX_Component_Noise*>( this))->GetTypeDesc(), this->GetTextureCount(), uTemplateTextureIndex ) );
#endif

	return( this->m_iTextureID );
}

const char* Asura_FSFX_Component_Noise::GetTextureName( const u_int uTemplateTextureIndex ) const
{
#if defined( ASURA_TOOLS)
	Asura_Assertf( uTemplateTextureIndex < this->GetTextureCount(), ("FSFX '%s' only supports %d textures. %d is invalid.", (const_cast<Asura_FSFX_Component_Noise*>( this))->GetTypeDesc(), this->GetTextureCount(), uTemplateTextureIndex ) );
#endif

	return( this->m_xTextureFilename.GetString() );
}

/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_Noise::CreateCopy() 
{
	Asura_FSFX_Component_Noise* pxComp = new Asura_FSFX_Component_Noise;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_Noise::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	this->m_iTextureID = ASURA_TEXTURE_INVALID;
	this->m_xTextureFilename = "";

	switch( iVersion )
	{
		case 3:
		{
			xStream >> this->m_uTextureHash;
			xStream >> this->m_fAmount;
			xStream >> this->m_uTextureTileX;
			xStream >> this->m_uTextureTileY;
			// For Version 3 only load the string name in tools build (for non tools the texture hash will have been output)
#if defined( ASURA_TOOLS )
			xStream >> this->m_xTextureFilename;
#else // #if defined( ASURA_TOOLS )
			xStream.SkipString();
#endif // #if defined( ASURA_TOOLS )
			break;
		}
		case 2:
		{
			xStream >> this->m_fAmount;
			xStream >> this->m_uTextureTileX;
			xStream >> this->m_uTextureTileY;
			xStream >> this->m_xTextureFilename;
			// For Version 2 make the hash value 
			this->m_uTextureHash = Asura_GetHashID_IgnoreExtension( this->m_xTextureFilename.GetString() );
#if !defined( ASURA_TOOLS )
			// For non-tools build release the string name
			this->m_xTextureFilename.MakeNull();
#endif // #if !defined( ASURA_TOOLS )
			break;
		}
		case 1:
		{
			xStream >> m_fAmount;
			break;
		}
		default:
		{
			Asura_Assert( 0, "Asura_FSFX_Component_Noise version not supported. Skipping" );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assertf( xStream.GetPosition() == (uPos + iSize), ("Size missmatch in Asura_FSFX_Component_Noise::Deserialize. File = 0x%08X Calculated = 0x%08X" , xStream.GetPosition(), (uPos + iSize)));

	this->SetTexture();
}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_Noise::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = (1 * sizeof( float )) + (3 * sizeof( uint32)) + (4 * this->m_xTextureFilename.GetSizeInInts());
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	Asura_Assert( this->m_uTextureHash == Asura_GetHashID_IgnoreExtension( this->m_xTextureFilename.GetString() ), "" );
	xStream << this->m_uTextureHash;
	xStream << this->m_fAmount;
	xStream << this->m_uTextureTileX;
	xStream << this->m_uTextureTileY;
	xStream << this->m_xTextureFilename;

	Asura_Assertf( xStream.GetPosition() == (uPos + iSize), ("Size missmatch in Asura_FSFX_Component_Noise::Serialize. File = 0x%08X Calculated = 0x%08X" , xStream.GetPosition(), (uPos + iSize)));
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_GaussianBlur functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_GaussianBlur::Asura_FSFX_Component_GaussianBlur() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fSize = 4.0f;
	m_fStrength = 100.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_GaussianBlur::CreateCopy() 
{
	Asura_FSFX_Component_GaussianBlur* pxComp = new Asura_FSFX_Component_GaussianBlur;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_GaussianBlur::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fSize;
		xStream >> m_fStrength;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_GaussianBlur version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_GaussianBlur::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_GaussianBlur::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 2 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << m_fSize;
	xStream << m_fStrength;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_GaussianBlur::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_RadialBlur functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_RadialBlur::Asura_FSFX_Component_RadialBlur() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fSize = 1.0f;
	m_fStrength = 0.0f;
	m_fAngle = 0.0f;
	m_fCentreX = 0.0f;
	m_fCentreY = 0.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_RadialBlur::CreateCopy() 
{
	Asura_FSFX_Component_RadialBlur* pxComp = new Asura_FSFX_Component_RadialBlur;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_RadialBlur::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fSize;
		xStream >> m_fStrength;
		xStream >> m_fAngle;
		xStream >> m_fCentreX;
		xStream >> m_fCentreY;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_RadialBlur version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_RadialBlur::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_RadialBlur::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 5 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << m_fSize;
	xStream << m_fStrength;
	xStream << m_fAngle;
	xStream << m_fCentreX;
	xStream << m_fCentreY;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_RadialBlur::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_DepthOfField functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_DepthOfField::Asura_FSFX_Component_DepthOfField() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fSize = 4.0f;
	m_fStrength = 100.0f;
	m_fNearBlurZ = 0.0f;
	m_fFarBlurZ = 0.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_DepthOfField::CreateCopy() 
{
	Asura_FSFX_Component_DepthOfField* pxComp = new Asura_FSFX_Component_DepthOfField;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_DepthOfField::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	switch( iVersion)
	{
		case 2:
		{
			xStream >> this->m_fSize;
			xStream >> this->m_fStrength;
			xStream >> this->m_fNearBlurZ;
			xStream >> this->m_fFarBlurZ;
			break;
		}
		case 1:
		{
			xStream >> m_fSize;
			xStream >> m_fStrength;
			xStream >> m_fFarBlurZ;
			this->m_fNearBlurZ = 0.0f;
			break;
		}
		default:
		{
			Asura_Assertf( 0, ("Asura_FSFX_Component_DepthOfField version not supported. Skipping. File Version %d. Current Version %d", iVersion, this->GetVersion()) );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_DepthOfField::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_DepthOfField::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 4 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << this->m_fSize;
	xStream << this->m_fStrength;
	xStream << this->m_fNearBlurZ;
	xStream << this->m_fFarBlurZ;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_DepthOfField::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_Bloom functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_Bloom::Asura_FSFX_Component_Bloom() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fThreshold = 0.5f;
	m_fIntensityR = 0.5f;
	m_fIntensityG = 0.5f;
	m_fIntensityB = 0.5f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_Bloom::CreateCopy() 
{
	Asura_FSFX_Component_Bloom* pxComp = new Asura_FSFX_Component_Bloom;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_Bloom::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	switch( iVersion)
	{
		case 2:
		{
			xStream >> this->m_fThreshold;
			xStream >> this->m_fIntensityR;
			xStream >> this->m_fIntensityG;
			xStream >> this->m_fIntensityB;
			break;
		}
		case 1:
		{
			float fDummy;
			
			xStream >> this->m_fThreshold;
			xStream >> this->m_fIntensityR;
			xStream >> fDummy;
			
			this->m_fIntensityG = this->m_fIntensityR;
			this->m_fIntensityB = this->m_fIntensityR;
			break;
		}
		default:
		{
			Asura_Assert( 0, "Asura_FSFX_Component_Bloom version not supported. Skipping" );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_Bloom::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_Bloom::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 4 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << this->m_fThreshold;
	xStream << this->m_fIntensityR;
	xStream << this->m_fIntensityG;
	xStream << this->m_fIntensityB;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_Bloom::Serialize" );
}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_Gamma functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_Gamma::Asura_FSFX_Component_Gamma() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fAmount = 1.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_Gamma::CreateCopy() 
{
	Asura_FSFX_Component_Gamma* pxComp = new Asura_FSFX_Component_Gamma;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_Gamma::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fAmount;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_Gamma version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_Gamma::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_Gamma::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 1 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << m_fAmount;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_Gamma::Serialize" );
}

#endif // ASURA_TOOLS



///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_RGBGamma functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_RGBGamma::Asura_FSFX_Component_RGBGamma() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fAmountR = 1.0f;
	m_fAmountG = 1.0f;
	m_fAmountB = 1.0f;
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_RGBGamma::CreateCopy() 
{
	Asura_FSFX_Component_RGBGamma* pxComp = new Asura_FSFX_Component_RGBGamma;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_RGBGamma::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fAmountR;
		xStream >> m_fAmountG;
		xStream >> m_fAmountB;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_RGBGamma version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_RGBGamma::Deserialize" );

}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_RGBGamma::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 3 * sizeof( float );
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();
	
	xStream << m_fAmountR;
	xStream << m_fAmountG;
	xStream << m_fAmountB;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size missmatch in Asura_FSFX_Component_RGBGamma::Serialize" );
}

#endif // ASURA_TOOLS



///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_SM30_Bloom functions
///////////////////////////////////////////////////////////////////////////////////////

/*
Constructor
*/

Asura_FSFX_Component_SM30_Bloom::Asura_FSFX_Component_SM30_Bloom() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fRBloomThreshold = 0.5f;
	m_fGBloomThreshold = 0.5f;
	m_fBBloomThreshold = 0.5f;

	m_fROutputCol = 1.0f;
	m_fGOutputCol = 0.6f;
	m_fBOutputCol = 0.2f;

	m_fBlurRadius = 2.0f;	
	m_iBlurPasses = 2;
	m_fSaturation = 0.5f;

	m_bAccumulate = true;
	m_fAccumFade  = 0.3f;

	m_bShowBlurBuffer = false;
	m_fInputScale = 0.5f;

}


/*
Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_SM30_Bloom::CreateCopy() 
{
	Asura_FSFX_Component_SM30_Bloom* pxComp = new Asura_FSFX_Component_SM30_Bloom;
	*pxComp = *this;
	return pxComp;
}


/*
Load function
*/
void Asura_FSFX_Component_SM30_Bloom::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	unsigned int uPos = xStream.GetPosition();
#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fRBloomThreshold;
		xStream >> m_fGBloomThreshold;
		xStream >> m_fBBloomThreshold;
		xStream >> m_fROutputCol;
		xStream >> m_fGOutputCol;
		xStream >> m_fBOutputCol;
		xStream >> m_fBlurRadius;		
		xStream >> m_fSaturation;
		xStream >> m_bAccumulate;
		xStream >> m_fAccumFade;
		xStream >> m_iBlurPasses;
		xStream >> m_bShowBlurBuffer;
		xStream >> m_fInputScale;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_SM30_Bloom version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_SM30_Bloom::Deserialize" );
}


#ifdef ASURA_TOOLS
/*
Save function - only needed on tools
*/
void Asura_FSFX_Component_SM30_Bloom::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 11 * sizeof( float ) + 2; // bools in streams are output as 1 byte now?
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_fRBloomThreshold;
	xStream << m_fGBloomThreshold;
	xStream << m_fBBloomThreshold;
	xStream << m_fROutputCol;
	xStream << m_fGOutputCol;
	xStream << m_fBOutputCol;
	xStream << m_fBlurRadius;
	xStream << m_fSaturation;
	xStream << m_bAccumulate;
	xStream << m_fAccumFade;
	xStream << m_iBlurPasses;
	xStream << m_bShowBlurBuffer;
	xStream << m_fInputScale;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_SM30_Bloom::Serialize" );
}

#endif // ASURA_TOOLS

#ifndef ASURA_TARGET_SM3
void Asura_FSFX_Component_SM30_Bloom::Platform_Render( const float fIntensity )
{
	// There's a separate source file for this effect if we are targeting SM3
	// It's never used on non-SM3 platforms
}
#endif //ASURA_TARGET_SM3


///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_SM30_DepthOfField functions
///////////////////////////////////////////////////////////////////////////////////////

/*
Constructor
*/

Asura_FSFX_Component_SM30_DepthOfField::Asura_FSFX_Component_SM30_DepthOfField() 
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fBlurRadius = 2.0f;	
	m_iBlurPasses = 2;

	m_fBlurDesaturate = 0.0f;

	m_bShowBlurBuffer = false;

	m_afDistances[0] = 0.0f;
	m_afDistances[1] = 10.0f;
	m_afDistances[2] = 50.0f;
	m_afDistances[3] = 60.0f;

	m_fBleedScale = 1.25f;
}


/*
Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_SM30_DepthOfField::CreateCopy() 
{
	Asura_FSFX_Component_SM30_DepthOfField* pxComp = new Asura_FSFX_Component_SM30_DepthOfField;
	*pxComp = *this;
	return pxComp;
}


/*
Load function
*/
void Asura_FSFX_Component_SM30_DepthOfField::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	unsigned int uPos = xStream.GetPosition();
#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fBlurRadius;	
		xStream >> m_iBlurPasses;

		xStream >> m_bShowBlurBuffer;
		xStream >> m_fBlurDesaturate;

		xStream >> m_afDistances[0];
		xStream >> m_afDistances[1];
		xStream >> m_afDistances[2];
		xStream >> m_afDistances[3];

		xStream >> m_fBleedScale;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_SM30_DepthOfField version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_SM30_Bloom::Deserialize" );
}


#ifdef ASURA_TOOLS
/*
Save function - only needed on tools
*/
void Asura_FSFX_Component_SM30_DepthOfField::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = (8 * 4) + 1; // bools in streams are output as 1 byte now?
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_fBlurRadius;	
	xStream << m_iBlurPasses;

	xStream << m_bShowBlurBuffer;
	xStream << m_fBlurDesaturate;

	xStream << m_afDistances[0];
	xStream << m_afDistances[1];
	xStream << m_afDistances[2];
	xStream << m_afDistances[3];

	xStream << m_fBleedScale;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_DepthOfField::Serialize" );
}

#endif // ASURA_TOOLS

#ifndef ASURA_TARGET_SM3
void Asura_FSFX_Component_SM30_DepthOfField::Platform_Render( const float fIntensity )
{
	// There's a separate source file for this effect if we are targeting SM3
	// It's never used on non-SM3 platforms
}
#endif //ASURA_TARGET_SM3



///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_SM30_MotionBlur functions
///////////////////////////////////////////////////////////////////////////////////////

/*
Constructor
*/

Asura_FSFX_Component_SM30_MotionBlur::Asura_FSFX_Component_SM30_MotionBlur() 
{
	SET_PLATFORM_ACTIVE_FLAGS();
	m_fBlurAmount				= 0.5f;
	m_fIgnoreMovementBlur		= 0.0f;
}


/*
Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_SM30_MotionBlur::CreateCopy() 
{
	Asura_FSFX_Component_SM30_MotionBlur* pxComp = new Asura_FSFX_Component_SM30_MotionBlur;
	*pxComp = *this;
	return pxComp;
}


/*
Load function
*/
void Asura_FSFX_Component_SM30_MotionBlur::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	unsigned int uPos = xStream.GetPosition();
#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_fBlurAmount;
		xStream >> m_fIgnoreMovementBlur;
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_SM30_MotionBlur version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_SM30_MotionBlur::Deserialize" );
}


#ifdef ASURA_TOOLS
/*
Save function - only needed on tools
*/
void Asura_FSFX_Component_SM30_MotionBlur::Serialize( Asura_Chunk_Stream& xStream ) const
{
	int iSize = 2 * 4;
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_fBlurAmount;
	xStream << m_fIgnoreMovementBlur;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_MotionBlur::Serialize" );
}

#endif // ASURA_TOOLS

#ifndef ASURA_TARGET_SM3
void Asura_FSFX_Component_SM30_MotionBlur::Platform_Render( const float fIntensity )
{
	// There's a separate source file for this effect if we are targeting SM3
	// It's never used on non-SM3 platforms
}
#endif //ASURA_TARGET_SM3



///////////////////////////////////////////////////////////////////////////////////////
// Asura_FSFX_Component_ColourMap functions
///////////////////////////////////////////////////////////////////////////////////////


/*
	Constructor
*/
Asura_FSFX_Component_ColourMap::Asura_FSFX_Component_ColourMap()
	: m_iColourMapIndex( ASURA_TEXTURE_NONE )
{
	SET_PLATFORM_ACTIVE_FLAGS();
}


/*
	Copy function
*/
Asura_FSFX_Component* Asura_FSFX_Component_ColourMap::CreateCopy()
{
	Asura_FSFX_Component_ColourMap* pxComp = new Asura_FSFX_Component_ColourMap;
	*pxComp = *this;
	return pxComp;
}


/*
	Load function
*/
void Asura_FSFX_Component_ColourMap::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		unsigned int uPos = xStream.GetPosition();
	#endif

	if ( iVersion == GetVersion() )
	{
		xStream >> m_xColourMapName;

		bool bWasNew = false;
		m_iColourMapIndex = Asura_Texture_Management::AddTexture( m_xColourMapName.GetString(),
																  &bWasNew );
		if( bWasNew )
		{
			Asura_Texture_Management::LoadTextures();
		}
	}
	else
	{
		Asura_Assert( 0, "Asura_FSFX_Component_ColourMap version not supported. Skipping" );
		xStream.Skip( iSize );
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_ColourMap::Deserialize" );
}


#ifdef ASURA_TOOLS

/*
	Save function - only needed on tools
*/
void Asura_FSFX_Component_ColourMap::Serialize( Asura_Chunk_Stream& xStream ) const
{
	const int iSizeInBytes = m_xColourMapName.GetSizeInBytes();
	const int iPadding = m_xColourMapName.GetSizeInInts() * 4 - iSizeInBytes;

	int iSize = iSizeInBytes + iPadding;
	WriteStandardHeader( xStream, iSize );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_xColourMapName;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_ColourMap::Serialize" );
}

#endif // ASURA_TOOLS


#ifndef ASURA_TARGET_SM3

/*
	Render function - only needed here for non-SM3 platforms
*/
void Asura_FSFX_Component_ColourMap::Platform_Render( const float fIntensity )
{
}

#endif // ASURA_TARGET_SM3

/******************************************************************************/
/* AVP_FSFX_Component_DoubleVision Class Functions                              */
/******************************************************************************/

Asura_FSFX_Component_SM30_DoubleVision::Asura_FSFX_Component_SM30_DoubleVision()
:	m_fSpeed1(0.0f)
,	m_fPhase1(0.0f)
,	m_fSpeed2(0.0f)
,	m_fPhase2(0.0f)
,	m_fOrigXOffset1(0.0f)
,	m_fOrigYOffset1(0.0f)
,	m_fOrigXScale1(0.0f)
,	m_fOrigYScale1(0.0f)
,	m_fOrigXOffset2(0.0f)
,	m_fOrigYOffset2(0.0f)
,	m_fOrigXScale2(0.0f)
,	m_fOrigYScale2(0.0f)
{
	SET_PLATFORM_ACTIVE_FLAGS();
}

Asura_FSFX_Component* Asura_FSFX_Component_SM30_DoubleVision::CreateCopy()
{
	Asura_FSFX_Component_SM30_DoubleVision* pxComp = new Asura_FSFX_Component_SM30_DoubleVision;
	*pxComp = *this;
	return pxComp;
}

void Asura_FSFX_Component_SM30_DoubleVision::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	u_int uPos = xStream.GetPosition();
#endif

	switch( iVersion )
	{
	case 1:
		{
			xStream >> m_fOrigXOffset1;
			xStream >> m_fOrigYOffset1;
			xStream >> m_fOrigXScale1;
			xStream >> m_fOrigYScale1;
			xStream >> m_fSpeed1;
			xStream >> m_fPhase1;

			xStream >> m_fOrigXOffset2;
			xStream >> m_fOrigYOffset2;
			xStream >> m_fOrigXScale2;
			xStream >> m_fOrigYScale2;
			xStream >> m_fSpeed2;
			xStream >> m_fPhase2;
			break;
		}

	default:
		{
			Asura_Assert( false, "Asura_FSFX_Component_SM30_DoubleVision version not supported. Skipping." );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_SM30_DoubleVision::Deserialize." );
}

#if defined ASURA_TOOLS

void Asura_FSFX_Component_SM30_DoubleVision::Serialize( Asura_Chunk_Stream& xStream ) const
{
	const int iSizeInBytes = 48;
	WriteStandardHeader( xStream, iSizeInBytes );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_fOrigXOffset1;
	xStream << m_fOrigYOffset1;
	xStream << m_fOrigXScale1;
	xStream << m_fOrigYScale1;
	xStream << m_fSpeed1;
	xStream << m_fPhase1;

	xStream << m_fOrigXOffset2;
	xStream << m_fOrigYOffset2;
	xStream << m_fOrigXScale2;
	xStream << m_fOrigYScale2;
	xStream << m_fSpeed2;
	xStream << m_fPhase2;

	Asura_Assert( xStream.GetPosition() == uPos + iSizeInBytes, "Size mismatch in Asura_FSFX_Component_SM30_DoubleVision::Serialize." );
}

#endif // defined ASURA_TOOLS

#ifndef ASURA_TARGET_SM3
void Asura_FSFX_Component_SM30_DoubleVision::Platform_Render( const float fIntensity )
{
	// There's a separate source file for this effect if we are targeting SM3
	// It's never used on non-SM3 platforms
}
#endif //ASURA_TARGET_SM3

/******************************************************************************/
/* AVP_FSFX_Component_FishEye Class Functions                              */
/******************************************************************************/

Asura_FSFX_Component_SM30_FishEye::Asura_FSFX_Component_SM30_FishEye()
: m_fXScale( 1.0f ), m_fYScale( 1.0f ), m_fFov( Asura_Maths::Pi / 2 ), m_fMultiplier(1.0f), m_eType(Linear), m_uFlags(0), m_fMinMultiplier(0.0f), m_fMaxMultiplier(1.0f), m_fSpeed(1.0f), m_fSineClip(-1.0f)
{
	SET_PLATFORM_ACTIVE_FLAGS();
}

Asura_FSFX_Component* Asura_FSFX_Component_SM30_FishEye::CreateCopy()
{
	Asura_FSFX_Component_SM30_FishEye* pxComp = new Asura_FSFX_Component_SM30_FishEye;
	*pxComp = *this;
	return pxComp;
}

void Asura_FSFX_Component_SM30_FishEye::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	u_int uPos = xStream.GetPosition();
#endif

	switch( iVersion )
	{
	case 1:
		{
			xStream >> m_fXScale;
			xStream >> m_fYScale;
			xStream >> m_fFov;
			xStream >> m_fMultiplier;
			xStream >> m_eType;
			break;
		}
	case 2:
		{
			xStream >> m_fXScale;
			xStream >> m_fYScale;
			xStream >> m_fFov;
			xStream >> m_fMultiplier;
			xStream >> m_eType;
			xStream >> m_uFlags;
			xStream >> m_fMinMultiplier;
			xStream >> m_fMaxMultiplier;
			xStream >> m_fSpeed;
			break;
		}
	case 3:
		{
			xStream >> m_fXScale;
			xStream >> m_fYScale;
			xStream >> m_fFov;
			xStream >> m_fMultiplier;
			xStream >> m_eType;
			xStream >> m_uFlags;
			xStream >> m_fMinMultiplier;
			xStream >> m_fMaxMultiplier;
			xStream >> m_fSpeed;
			xStream >> m_fSineClip;
			break;
		}

	default:
		{
			Asura_Assert( false, "Asura_FSFX_Component_SM30_FishEye version not supported. Skipping." );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in Asura_FSFX_Component_SM30_FishEye::Deserialize." );
}

#if defined ASURA_TOOLS

void Asura_FSFX_Component_SM30_FishEye::Serialize( Asura_Chunk_Stream& xStream ) const
{
	const int iSizeInBytes = 40;
	WriteStandardHeader( xStream, iSizeInBytes );

	unsigned int uPos = xStream.GetPosition();

	xStream << m_fXScale;
	xStream << m_fYScale;
	xStream << m_fFov;
	xStream << m_fMultiplier;
	xStream << m_eType;
	xStream << m_uFlags;
	xStream << m_fMinMultiplier;
	xStream << m_fMaxMultiplier;
	xStream << m_fSpeed;
	xStream << m_fSineClip;

	Asura_Assert( xStream.GetPosition() == uPos + iSizeInBytes, "Size mismatch in Asura_FSFX_Component_SM30_FishEye::Serialize." );
}

#endif // defined ASURA_TOOLS

#ifndef ASURA_TARGET_SM3
void Asura_FSFX_Component_SM30_FishEye::Platform_Render( const float fIntensity )
{
	// There's a separate source file for this effect if we are targeting SM3
	// It's never used on non-SM3 platforms
}
#endif //ASURA_TARGET_SM3



#endif // ASURA_USE_FSFX
