//
// Asura_FileAccess.cpp
//
// Created by Kevin Lea - 04/09/2001 15:06:14
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "FileAccess/Asura_FileAccess.h"

// Engine independent.
#include "ChunkSystem/Asura_Huffman.h"
#include "ChunkSystem/Asura_Zlib.h"
#include "Core/Asura_Checksum.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#if defined(ASURA_USE_STREAMING)
#include "FileAccess/Asura_FileCache.h"
#endif	// defined(ASURA_USE_STREAMING)

#ifdef ASURA_TOOLS
#include "FileAccess/Tools_LZX_Decoder.h"
#endif

//wrapper for the two possible decoders that may be in use
//RAR 21/08/2007
class Asura_DecoderWrapper
{
public :
	Asura_DecoderWrapper()
	: m_eCompression(ASURA_COMPRESSION_NOTCOMPRESSED)
	{}

	bool IsCompressed() const {return m_eCompression != ASURA_COMPRESSION_NOTCOMPRESSED;}
	ASURA_COMPRESSION_TYPE GetCompressionType() const {return m_eCompression;}

	void SetCompressed(ASURA_COMPRESSION_TYPE eType) {m_eCompression = eType;}

    u_int Decompress( void* pvDest, u_int uNumBytes )
	{
		switch(m_eCompression)
		{
			#ifdef ASURA_USE_ZLIB
			case ASURA_COMPRESSION_ZLIB :
				return m_xZlibDecoder.Decompress(pvDest, uNumBytes);
			#endif

			case ASURA_COMPRESSION_HUFFMAN :
				return m_xHuffmanDecoder.Decompress(pvDest, uNumBytes);

			#ifdef ASURA_TOOLS
			case ASURA_COMPRESSION_LZX :
				return m_xLZXDecoder.Decompress(pvDest, uNumBytes);
			#endif

			default :
				Asura_Assert(0,"Error in Asura_FileAccess");
				return 0;
		}
	}

    void Begin( Asura_Handle iFileHandle )
	{
		switch(m_eCompression)
		{
			#ifdef ASURA_USE_ZLIB
			case ASURA_COMPRESSION_ZLIB :
				m_xZlibDecoder.Begin(iFileHandle);
				break;
			#endif

			case ASURA_COMPRESSION_HUFFMAN :
				m_xHuffmanDecoder.Begin(iFileHandle);
				break;

			#ifdef ASURA_TOOLS
			case ASURA_COMPRESSION_LZX :
				m_xLZXDecoder.Begin(iFileHandle);
				break;
			#endif

			default :
				Asura_Assert(0,"Error in Asura_FileAccess");
				break;
		}
	}

	void End()
	{
		switch(m_eCompression)
		{
			#ifdef ASURA_USE_ZLIB
			case ASURA_COMPRESSION_ZLIB :
				m_xZlibDecoder.End();
				break;
			#endif

			case ASURA_COMPRESSION_HUFFMAN :
				m_xHuffmanDecoder.End();
				break;

			#ifdef ASURA_TOOLS
			case ASURA_COMPRESSION_LZX :
				m_xLZXDecoder.End();
				break;
			#endif

			default :
				Asura_Assert(0,"Error in Asura_FileAccess");
				break;
		}
		m_eCompression = ASURA_COMPRESSION_NOTCOMPRESSED;
	}

	u_int GetDecompressedSize()
	{
		switch(m_eCompression)
		{
			#ifdef ASURA_USE_ZLIB
			case ASURA_COMPRESSION_ZLIB :
				return m_xZlibDecoder.GetDecompressedSize();
			#endif

			case ASURA_COMPRESSION_HUFFMAN :
				return m_xHuffmanDecoder.GetDecompressedSize();

			#ifdef ASURA_TOOLS
			case ASURA_COMPRESSION_LZX :
				return m_xLZXDecoder.GetDecompressedSize();
			#endif

			default :
				Asura_Assert(0,"Error in Asura_FileAccess");
				return 0;
		}
	}

	Asura_Huffman_Decoder& GetHuffmanDecoder() {return m_xHuffmanDecoder;}

private:
	ASURA_COMPRESSION_TYPE m_eCompression;
	
	Asura_Huffman_Decoder m_xHuffmanDecoder;
	#ifdef ASURA_USE_ZLIB
	Asura_Zlib_Decoder m_xZlibDecoder;
	#endif
	#ifdef ASURA_TOOLS
	Tools_LZX_Decoder m_xLZXDecoder;
	#endif
};


// Kevin Lea - 04/09/2001 15:16:47 - static data private to Asura_FileAccess
Asura_Handle_List* Asura_FileAccess::HandleList;
bool Asura_FileAccess::s_abFileIsRaw[ASURA_MAX_NO_OF_OPEN_FILES];

static unsigned g_auFilePosition[ASURA_MAX_NO_OF_OPEN_FILES];
static Asura_DecoderWrapper g_axDecompressors[ASURA_MAX_NO_OF_OPEN_FILES];


#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
static bool g_abComputeChecksum[ASURA_MAX_NO_OF_OPEN_FILES];
static Asura_Checksum g_axChecksumInstance[ASURA_MAX_NO_OF_OPEN_FILES];
#endif // #ifdef ASURA_SUPPORT_FILE_CHECKSUMS

#if 1//_DEBUG || defined(ASURA_TOOLS)
static const int ASURA_MAX_FILENAME_LENGTH = 256;
static char g_aszFilenames[ASURA_MAX_NO_OF_OPEN_FILES][ASURA_MAX_FILENAME_LENGTH] = {""};
#endif

#if defined(ASURA_USE_STREAMING)
// The handle to be used for a streaming file - we can only have one of these...
static const Asura_Handle	g_iStreamingFileHandle	= ASURA_MAX_NO_OF_OPEN_FILES;

// File-cache used when streaming...
Asura_FileCache*	Asura_FileAccess::s_pxStreamingFileCache	= NULL;
#endif	// defined(ASURA_USE_STREAMING)

// (AAM - 20.06.2006) On Xbox 360, texture packages can be about 256 MB.
// Allow projects to override the suspicion limit
#if !defined ASURA_FILEACCESS_SUSPICIOUS_BYTE_COUNT
	#define ASURA_FILEACCESS_SUSPICIOUS_BYTE_COUNT (128 * 1024 * 1024)
#endif

void Asura_FileAccess::Initialise(void)
{
	HandleList = new Asura_Handle_List(ASURA_MAX_NO_OF_OPEN_FILES);

    Platform_Initialise();
}

void Asura_FileAccess::Shutdown(void)
{
    Platform_Shutdown();

	delete HandleList;
	HandleList	= NULL;
}

Asura_Handle Asura_FileAccess::Open( const char* szFilename, bool bComputeChecksum /* = false */, bool bFailureHandledExternally /* = false */ )
{
	Asura_Handle iFileHandle;

	bool success = HandleList->GetNewHandle(iFileHandle);

	if (!success)
	{
		Asura_Assert(bFailureHandledExternally,"Too many files open");
        Asura_OutputToDebugger_Retail(( "Unable to open file \"%s\" (too many files open.)\n", szFilename ));
		return ASURA_INVALID_HANDLE;
	}

	s_abFileIsRaw[iFileHandle] = false;
	success = Platform_Open(iFileHandle,szFilename);

	if (!success)
	{
		// Kevin Lea - 05/09/2001 15:01:10 - unable to open file
#if 0
		char buffer[256];
		sprintf(buffer,"Platform unable to open file\n<%s>",szFilename);
		Asura_Assert(0,buffer);
#endif
		
		// Kevin Lea - 05/09/2001 15:02:52 - release the file handle, since it's not needed now
		HandleList->ReleaseHandle(iFileHandle);
		return ASURA_INVALID_HANDLE;
	}

#if 1//_DEBUG || defined(ASURA_TOOLS)
	// Kevin Lea - 03/10/2001 18:01:30 - filename stored for later query, as per Morgan's request
	// (MH - 17.12.2002) evil global.
    //g_axFilenames[iFileHandle] = filename;
    Asura_PrintfToBfr(g_aszFilenames[iFileHandle], "%s", ASURA_MAX_FILENAME_LENGTH, true, szFilename);
#endif

    // Assume it's not compressed until we're told differently.
    g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_NOTCOMPRESSED);
	g_auFilePosition[iFileHandle] = 0;

#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
	g_abComputeChecksum[iFileHandle] = bComputeChecksum;
	if (bComputeChecksum)
	{
		g_axChecksumInstance[iFileHandle].Initialise();
	}
#endif

	return iFileHandle;
}

Asura_Handle Asura_FileAccess::OpenRaw( const char* szFilename, bool bFailureHandledExternally /* = false */ )
{
	Asura_Handle iFileHandle;

	bool success = HandleList->GetNewHandle(iFileHandle);

	if (!success)
	{
		Asura_Assert(bFailureHandledExternally,"Too many files open");
		Asura_OutputToDebugger_Retail(( "Unable to open file \"%s\" (too many files open.)\n", szFilename ));
		return ASURA_INVALID_HANDLE;
	}

	s_abFileIsRaw[iFileHandle] = true;
	success = Platform_Open(iFileHandle,szFilename);

	if (!success)
	{
		// Kevin Lea - 05/09/2001 15:01:10 - unable to open file
#if 0
		char buffer[256];
		sprintf(buffer,"Platform unable to open file\n<%s>",szFilename);
		Asura_Assert(0,buffer);
#endif

		// Kevin Lea - 05/09/2001 15:02:52 - release the file handle, since it's not needed now
		HandleList->ReleaseHandle(iFileHandle);
		return ASURA_INVALID_HANDLE;
	}

#if 1//_DEBUG || defined(ASURA_TOOLS)
	// Kevin Lea - 03/10/2001 18:01:30 - filename stored for later query, as per Morgan's request
	// (MH - 17.12.2002) evil global.
	//g_axFilenames[iFileHandle] = filename;
	Asura_PrintfToBfr(g_aszFilenames[iFileHandle], "%s", ASURA_MAX_FILENAME_LENGTH, true, szFilename);
#endif

	// Assume it's not compressed until we're told differently.
	g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_NOTCOMPRESSED);
	g_auFilePosition[iFileHandle] = 0;

#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
	g_abComputeChecksum[iFileHandle] = false;
#endif

	return iFileHandle;
}

// function to allow loading a file directly into memory, and then using file access
// functions to read it. Allows use of temp storage on PS3
Asura_Handle Asura_FileAccess::OpenFromMemory( const void* pMem, u_int uSize, bool bFailureHandledExternally /* = false */ )
{
	Asura_Handle iFileHandle;
	bool bSuccess = HandleList->GetNewHandle(iFileHandle);

	if (!bSuccess)
	{
		Asura_Assert(bFailureHandledExternally,"Too many files open");
		return ASURA_INVALID_HANDLE;
	}

	s_abFileIsRaw[iFileHandle] = false;
	if (!Platform_OpenFromMemory(iFileHandle, pMem, uSize))
	{
		Asura_Assert(bFailureHandledExternally, "platform can't open from memory");
		HandleList->ReleaseHandle(iFileHandle);
		return ASURA_INVALID_HANDLE;
	}

	Asura_PrintfToBfr(g_aszFilenames[iFileHandle], "%s", ASURA_MAX_FILENAME_LENGTH, true, "memory file");

    // Assume it's not compressed until we're told differently.
	g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_NOTCOMPRESSED);
	g_auFilePosition[iFileHandle] = 0;

	return iFileHandle;
}

#if defined(ASURA_USE_STREAMING)

Asura_Handle Asura_FileAccess::OpenForStreaming (const char* szFilename)
{
	if(s_pxStreamingFileCache == NULL)
	{
		Asura_Assert(false, "No file-cache to use for streaming.");
		return(ASURA_INVALID_HANDLE);
	}
	else if(s_pxStreamingFileCache->OpenFile(szFilename))
	{
		return(g_iStreamingFileHandle);
	}
	else
	{
		Asura_Assertf(false, ("Cannot open \"%s\" for streaming.", szFilename));
		return(ASURA_INVALID_HANDLE);
	}
}

#endif	// defined(ASURA_USE_STREAMING)

Asura_Handle Asura_FileAccess::LoadAsynchronous(const char* szFilename)
{
	Asura_Handle iFileHandle;

	bool success = HandleList->GetNewHandle(iFileHandle);

	if (!success)
	{
		Asura_Assert(0,"Too many files open");
		return ASURA_INVALID_HANDLE;
	}

	s_abFileIsRaw[iFileHandle] = false;
	success = Platform_LoadAsynchronous(iFileHandle,szFilename);

	if (!success)
	{
		
		// Kevin Lea - 05/09/2001 15:02:52 - release the file handle, since it's not needed now
		HandleList->ReleaseHandle(iFileHandle);
		return ASURA_INVALID_HANDLE;
	}

#if 1//_DEBUG || defined(ASURA_TOOLS)
	// Kevin Lea - 03/10/2001 18:01:30 - filename stored for later query, as per Morgan's request
	// (MH - 17.12.2002) evil global.
    //g_axFilenames[iFileHandle] = filename;
    Asura_PrintfToBfr(g_aszFilenames[iFileHandle], "%s", ASURA_MAX_FILENAME_LENGTH, true, szFilename);
#endif

    // Assume it's not compressed until we're told differently.
	g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_NOTCOMPRESSED);
    
	g_auFilePosition[iFileHandle] = 0;
#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
	g_abComputeChecksum[iFileHandle] = 0;
#endif

	return iFileHandle;
}

bool Asura_FileAccess::SeekFile( Asura_Handle iFileHandle, u_int uFilePosition, bool bFailureHandledExternally /* = false */ )
{
	if(IsFileRaw(iFileHandle))
	{
		Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );

		if(Platform_SeekFile(iFileHandle, uFilePosition))
		{
			g_auFilePosition[iFileHandle]	= uFilePosition;
			return(true);
		}
		else
		{
			return(false);
		}
	}
	else
	{
#if defined(ASURA_USE_STREAMING)
		if(iFileHandle == g_iStreamingFileHandle)
		{
			Asura_Assert(false, "Cannot seek in a file-cache.");
			return(false);
		}
		else
#endif	// defined(ASURA_USE_STREAMING)
		{
			Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
			if(Platform_SeekFile(iFileHandle, uFilePosition))
			{
				g_auFilePosition[iFileHandle]	= uFilePosition;
				return(true);
			}
			else
			{
				return(false);
			}
		}
	}
}

bool Asura_FileAccess::CanBeRead(Asura_Handle iFileHandle)
{
#if defined(ASURA_USE_STREAMING)
	// Are we trying to read the streaming file?
	if(iFileHandle == g_iStreamingFileHandle)
	{
		return(s_pxStreamingFileCache->CanBeRead());
	}
	else
#endif	// defined(ASURA_USE_STREAMING)
	{
		return(iFileHandle != ASURA_INVALID_HANDLE);
	}
}

bool Asura_FileAccess::ReadData( Asura_Handle iFileHandle, void* pvDest, int iNumBytes, bool bFailureHandledExternally )
{
	Asura_Assert( bFailureHandledExternally || iNumBytes > 0, "Invalid number of bytes requested." );

	// (MP - 13.06.2006) Catch chunk parsing errors, etc., a bit earlier.
	// Nothing reads close to this amount of data in one go at the moment,
	// even on PS3/360.
	Asura_Assert( bFailureHandledExternally || iNumBytes < ASURA_FILEACCESS_SUSPICIOUS_BYTE_COUNT, "Suspiciously large number of bytes requested" );

	if(IsFileRaw(iFileHandle))
	{
		if(iNumBytes <= 0)
		{ 
			return false; 
		} 
		else
		{
			Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );

			if( Platform_ReadData(iFileHandle, pvDest, iNumBytes) )
			{
				g_auFilePosition[iFileHandle] += iNumBytes;
				return true;
			}
			else
			{
				Asura_Assert( bFailureHandledExternally, "Platform_ReadData() failed" );
				return false;
			}
		}
	}
	else
	{
		if( !iNumBytes ) // RAR - Decompressor doesn't behave correctly when reading 0 bytes
		{ 
			return false; 
		} 
#if defined(ASURA_USE_STREAMING)
		else if(iFileHandle == g_iStreamingFileHandle)
		{
			const u_int	uNumBytes		= static_cast<u_int>(iNumBytes);
			const u_int	uNumBytesRead	= s_pxStreamingFileCache->ReadData(uNumBytes, static_cast<u_char*>(pvDest));

			Asura_Assert(uNumBytesRead == uNumBytes, "streaming file cache ReadData() failed");
			// Use the streaming file-cache...
			return(uNumBytesRead == uNumBytes);
		}
#endif	// defined(ASURA_USE_STREAMING)
		else
		{
			Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
			if( g_axDecompressors[iFileHandle].IsCompressed() )
			{
				int iNumBytesDecompressed = g_axDecompressors[iFileHandle].Decompress( reinterpret_cast<u_char*>(pvDest), iNumBytes );
				g_auFilePosition[iFileHandle] += iNumBytesDecompressed;
#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
				if (g_abComputeChecksum[iFileHandle])
				{
					g_axChecksumInstance[iFileHandle].Update(reinterpret_cast<u_char*>(pvDest), iNumBytesDecompressed);
				}
#endif
				Asura_Assert(bFailureHandledExternally || iNumBytesDecompressed == iNumBytes, "decompressor ReadData() failed");
				return( iNumBytesDecompressed == iNumBytes );
			}
			else
			{
				if( Platform_ReadData(iFileHandle, pvDest, iNumBytes) )
				{
					g_auFilePosition[iFileHandle] += iNumBytes;
#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
					if (g_abComputeChecksum[iFileHandle])
					{
						g_axChecksumInstance[iFileHandle].Update(reinterpret_cast<u_char*>(pvDest), iNumBytes);
					}
#endif
					return true;
				}

				Asura_Assert( bFailureHandledExternally, "Platform_ReadData() failed" );
				return false;
			}
		}
	}
}

bool Asura_FileAccess::SkipData( Asura_Handle iFileHandle, int iNumBytes, bool bFailureHandledExternally /* = false */ )
{
	Asura_Assert( bFailureHandledExternally || iNumBytes >= 0, "Invalid number of bytes specified." );

	if(IsFileRaw(iFileHandle))
	{
		Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );

		if(Platform_SkipUncompressedData(iFileHandle, iNumBytes))
		{
			g_auFilePosition[iFileHandle] += iNumBytes;
			return true;
		}
		else
		{
			return false;
		}
	}
	else
	{
#if defined(ASURA_USE_STREAMING)
		if(iFileHandle == g_iStreamingFileHandle)
		{
			const u_int	uNumBytes			= static_cast<u_int>(iNumBytes);
			const u_int	uNumBytesSkipped	= s_pxStreamingFileCache->SkipData(uNumBytes);

			// Use the streaming file-cache...
			return(uNumBytesSkipped == uNumBytes);
		}
		else
#endif	// defined(ASURA_USE_STREAMING)
		{
			Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );

			bool bReadDataToTempBuffer = g_axDecompressors[iFileHandle].IsCompressed();
#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
			bReadDataToTempBuffer = bReadDataToTempBuffer || g_abComputeChecksum[iFileHandle];
#endif
			if (bReadDataToTempBuffer)
			{
				// Kevin Lea - 14/09/2001 16:44:01 - ugh, this is awkward with compressed data sources
				// for now let's do something a bit inefficient: load the data into a dummy area and then trash it
				char dummyBuffer[1024];

				for(int i=0; i<iNumBytes; i+=1024)
				{
					int bytesLeft = iNumBytes-i;
					int ammountToRead;

					if (bytesLeft>=1024)
					{
						ammountToRead = 1024;
					}
					else
					{
						ammountToRead = bytesLeft;
					}

					if( !ReadData(iFileHandle,dummyBuffer,ammountToRead) )
					{
						return false;
					}
				}
			}
			else
			{
				if( !Platform_SkipUncompressedData(iFileHandle, iNumBytes) )
				{
					return false;
				}

				g_auFilePosition[iFileHandle] += iNumBytes;
			}

			return true;
		}
	}
}

static const u_int READ_STRING_BUFFER_SIZE = 512; // Must be a multiple of 4.
static char g_szReadStringBuffer[READ_STRING_BUFFER_SIZE] = "";
static u_int g_uReadStringLength = 0;

const char* Asura_FileAccess::ReadString( Asura_Handle iFileHandle, u_int* puLength /*= NULL*/, u_int* puSizeInInts /*= NULL*/ )
{
	char* pcDest = g_szReadStringBuffer;
	
    bool bEndFound = false;
    g_uReadStringLength = 0;
	do
	{
		ReadData( iFileHandle, pcDest, 4 );
        
        for( u_int u = 4; u--; ++g_uReadStringLength )
        {
            if( !*pcDest++ )
            {
                bEndFound = true;
                break;
            }
        }
	}
	while( !bEndFound && g_uReadStringLength < READ_STRING_BUFFER_SIZE );

    // Ensure it's null terminated if the string was too long.
    if( g_uReadStringLength >= READ_STRING_BUFFER_SIZE )
    {
        g_uReadStringLength = READ_STRING_BUFFER_SIZE;
        g_szReadStringBuffer[g_uReadStringLength - 1] = '\0';
    }

    if( puLength )
    {
        *puLength = g_uReadStringLength;
    }

    if( puSizeInInts )
    {
        // +4 rather than +3 to account for nul terminator.
        *puSizeInInts = ((g_uReadStringLength + 4) >> 2);
    }

    return (*(int*)&g_szReadStringBuffer[0] == Asura_String::NULL_STRING_CHUNK) ? 0 : g_szReadStringBuffer;
}

u_int Asura_FileAccess::ReadLine( Asura_Handle iFileHandle, char* szBfr, u_int uBfrSize )
{
    Asura_Assert( uBfrSize, "" );
    
    char* szOut = szBfr;
    while( uBfrSize > 1 )
    {
        bool bRead = ReadData( iFileHandle, szOut, 1, true );

        // Stop if EOF, nul terminator or newline.
        if( !bRead || !*szOut || '\n' == *szOut ) { break; }

        // Ignore \r characters. May encounter them before a new line (\n)
        // and I don't think we should include them in the output.
        if( '\r' == *szOut ) { continue; }

        // Otherwise, keep it and continue to next char.
        ++szOut;
        --uBfrSize;
    }

    *szOut = '\0';
    return( szOut - szBfr );
}

const char* Asura_FileAccess::GetLastReadString( u_int* puLength /*= NULL*/, u_int* puSizeInInts /*= NULL*/ )
{
    if( puLength )
    {
        *puLength = g_uReadStringLength;
    }

    if( puSizeInInts )
    {
        *puSizeInInts = ((g_uReadStringLength + 3) >> 2);
    }

    return (*(int*)&g_szReadStringBuffer[0] == Asura_String::NULL_STRING_CHUNK) ? 0 : g_szReadStringBuffer;
}

void Asura_FileAccess::Close( Asura_Handle iFileHandle, u_char* pucChecksumOutput /* = NULL */, bool bFailureHandledExternally /* = false */ )
{
	if( !HandleList )
	{
		Asura_Assert( bFailureHandledExternally, "Trying to close a file after Asura_FileAccess has been shut down." );
		return;
	}
	if(iFileHandle == ASURA_INVALID_HANDLE)
	{
		//Not sure if this should be handled silently.. asserting for now. RAR 22/01/2009
		Asura_Assert(iFileHandle != ASURA_INVALID_HANDLE, "Calling Asura_FileAccess::Close() with an invalid handle");
		return;
	}

	if(IsFileRaw(iFileHandle))
	{
		Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );

		Platform_Close( iFileHandle );
		HandleList->ReleaseHandle( iFileHandle );
	}
	else
	{
#if defined(ASURA_USE_STREAMING)
		if(iFileHandle == g_iStreamingFileHandle)
		{
			// Use the streaming file-cache...
			s_pxStreamingFileCache->CloseFile();
		}
		else
#endif	// defined(ASURA_USE_STREAMING)
		{
			Asura_Assert( bFailureHandledExternally || HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );

			if( g_axDecompressors[iFileHandle].IsCompressed() )
			{
				g_axDecompressors[iFileHandle].End();
			}

#ifdef ASURA_SUPPORT_FILE_CHECKSUMS
			if (g_abComputeChecksum[iFileHandle])
			{
				Asura_Assert(pucChecksumOutput, "no checksum output array");
				if (pucChecksumOutput)
				{
					g_axChecksumInstance[iFileHandle].Finalise(pucChecksumOutput);
				}
			}
#endif

			Platform_Close( iFileHandle );
			HandleList->ReleaseHandle( iFileHandle );
		}
	}
}

const char* Asura_FileAccess::GetFilename(Asura_Handle iFileHandle)
{
#if defined(ASURA_USE_STREAMING)
	if(iFileHandle == g_iStreamingFileHandle)
	{
		return(s_pxStreamingFileCache->GetFilename());
	}
	else
#endif	// defined(ASURA_USE_STREAMING)
	{
		Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
	    
		// (MH - 17.12.2002) evil global
		//return g_axFilenames[iFileHandle].GetString();
		return g_aszFilenames[iFileHandle];
	}
}

u_int Asura_FileAccess::GetNumOpenCompressedFiles( void )
{
    u_int uCount = 0;

    for( int i = 0; i < ASURA_MAX_NO_OF_OPEN_FILES; ++i )
    {
        if( HandleList->IsValidHandle(i) && g_axDecompressors[i].IsCompressed() )
        {
            ++uCount;
        }
    }
    
    return uCount;
}

unsigned Asura_FileAccess::GetFilePosition( Asura_Handle iFileHandle )
{
#if defined(ASURA_USE_STREAMING)
	if(iFileHandle == g_iStreamingFileHandle)
	{
		return(s_pxStreamingFileCache->GetFilePosition());
	}
	else
#endif	// defined(ASURA_USE_STREAMING)
	{
		Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
	    
		return g_auFilePosition[iFileHandle];
	}
}

Asura_Huffman_Decoder& Asura_FileAccess::SetCompressed( Asura_Handle iFileHandle )
{
    Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
    Asura_Assert( !g_axDecompressors[iFileHandle].IsCompressed(), "You've already called SetCompressed() for this file." );

	g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_HUFFMAN);

	g_axDecompressors[iFileHandle].Begin( iFileHandle );

    return g_axDecompressors[iFileHandle].GetHuffmanDecoder();
}

#ifdef ASURA_USE_ZLIB
void Asura_FileAccess::SetZlibCompressed( Asura_Handle iFileHandle )
{
    Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
    Asura_Assert( !g_axDecompressors[iFileHandle].IsCompressed(), "You've already called SetCompressed() for this file." );

	g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_ZLIB);

	g_axDecompressors[iFileHandle].Begin( iFileHandle );
}
#endif //ASURA_USE_ZLIB

#ifdef ASURA_TOOLS
void Asura_FileAccess::SetLZXCompressed( Asura_Handle iFileHandle)
{
    Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
    Asura_Assert( !g_axDecompressors[iFileHandle].IsCompressed(), "You've already called SetCompressed() for this file." );

	g_axDecompressors[iFileHandle].SetCompressed(ASURA_COMPRESSION_LZX);

	g_axDecompressors[iFileHandle].Begin( iFileHandle);
}
#endif //ASURA_USE_ZLIB


ASURA_COMPRESSION_TYPE Asura_FileAccess::GetCompressionType( Asura_Handle iFileHandle )
{
    Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
	return g_axDecompressors[iFileHandle].GetCompressionType();
}

void Asura_FileAccess::SetFilePosition( Asura_Handle iFileHandle, u_int uFilePos )
{
    Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
    g_auFilePosition[iFileHandle] = uFilePos;
}


enum ASURA_ASYNCLOADING_STATUS Asura_FileAccess::GetAsynchronousLoadingStatus(Asura_Handle iFileHandle)
{
    Asura_Assert( HandleList->IsValidHandle(iFileHandle), "Invalid file handle." );
	return Platform_GetAsynchronousLoadingStatus(iFileHandle);
}

u_int Asura_FileAccess::GetDecompressedFileSize(const Asura_Handle iFileHandle)
{
	if(iFileHandle == ASURA_INVALID_HANDLE)
	{
		Asura_Assert(iFileHandle != ASURA_INVALID_HANDLE,"");
		return 0;
	}

	if(g_axDecompressors[iFileHandle].IsCompressed())
	{
		return g_axDecompressors[iFileHandle].GetDecompressedSize();
	}
	else
	{
		return GetFileSize(iFileHandle);
	}
}


bool Asura_FileAccess::ProcessAsuraFileHeader( Asura_Handle iFileHandle, const bool bFailureHandledExternally )
{
	if(iFileHandle == ASURA_INVALID_HANDLE)
	{
		Asura_Assert(iFileHandle != ASURA_INVALID_HANDLE,"");
		return false;
	}

	char acHeader[ASURA_FILE_ID_NUM_CHARS];

	// Kevin Lea - 07/09/2001 11:56:01 - check id

	if(ReadData(iFileHandle, acHeader, ASURA_FILE_ID_NUM_CHARS))
	{
		bool bCompressed = false;
		//Check against one of the compressed file header ids
		#ifdef ASURA_USE_ZLIB
		if( !strnicmp(acHeader, ASURA_ZLIB_COMPRESSED_FILE_ID, ASURA_ZLIB_COMPRESSED_FILE_ID_ACTUAL_CHARS) )
		{
			SetZlibCompressed( iFileHandle );
			bCompressed = true;
		}
		else
		#endif
		if( !strnicmp(acHeader, ASURA_COMPRESSED_FILE_ID, ASURA_COMPRESSED_FILE_ID_ACTUAL_CHARS) )
		{
			SetCompressed( iFileHandle );
			bCompressed = true;
		}
		#ifdef ASURA_TOOLS
		else if( *(u_int*)acHeader == 0xed12f50f || *(u_int*)acHeader == 0xee12f50f)
		{
			SetLZXCompressed( iFileHandle);
			bCompressed = true;
		}
		#endif

		if(bCompressed)
		{
			//The Asura id exists in the compressed file too, so read that in and make sure it is coorect
			//RAR 21/08/2007
			g_axDecompressors[iFileHandle].Decompress(acHeader,	ASURA_FILE_ID_NUM_CHARS);

			if(strnicmp(acHeader, ASURA_UNCOMPRESSED_FILE_ID, ASURA_UNCOMPRESSED_FILE_ID_ACTUAL_CHARS) )
			{
				if( !bFailureHandledExternally )
				{
					Asura_Assertf(false, ("Incorrect file header ID in compressed data in '%s'",GetFilename(iFileHandle)));
				}
				return false;
			}
			return true;
		}

		
		if( !strnicmp(acHeader, ASURA_UNCOMPRESSED_FILE_ID, ASURA_UNCOMPRESSED_FILE_ID_ACTUAL_CHARS) )
		{
			// DLH - a normal uncompressed file
			return( true);
		}

		if( !bFailureHandledExternally )
		{
			if( strnicmp(acHeader, ASURA_UNCOMPRESSED_FILE_ID, ASURA_ALL_FILE_ID_ACTUAL_CHARS) )
			{
				Asura_Assertf( false, ("\"%s\" isn't an Asura file.", GetFilename(iFileHandle)) );
			}
			else
			{
				Asura_Assertf( false, ("Unknown Compression Type in asura file '%s'", GetFilename(iFileHandle)));
			}
		}
		return(false);
	}
	else
	{
		return false;
	}
}


/// End of file. ///
