
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_FileCache.h
// Author : Andy Tate
// Date   : 26-Apr-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_FILECACHE_H__
#define __ASURA_FILECACHE_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Multithread.h"
#include "Core/Asura_String.h"
#include "FileAccess/Asura_FileAccess.h"


////////////////////////////////////////////////////////////////////////////////
// Constants
////////////////////////////////////////////////////////////////////////////////
static const u_int uASURA_FILECACHE_MAX_FILENAME_LENGTH = 260;	// arbitrary limit
static const u_int uASURA_FILECACHE_MAX_NAME_LENGTH = 32;		// arbitrary limit

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_FileCache
// Author       : Andy Tate
// Date         : 26-Apr-2006
//
// Description  : Cyclic buffer which is filled up when a specified file is 
//				  opened this buffer is read-from.
////////////////////////////////////////////////////////////////////////////////

class Asura_FileCache
{
public:
	// Public functions...
	Asura_FileCache  (const char* szName, const u_int uBufferSize, const u_int uNumSections, const bool bAllowBlockingRead);
	Asura_FileCache  (const char* szName, const u_int uBufferSize, u_char* pucBuffer, const u_int uNumSections, const bool bAllowBlockingRead);
	~Asura_FileCache ();

	// Memory-assignment...
	void				ResetMemory		(void);
	void				SetMemory		(const u_int uNumBytes, u_char* pucDest);

	// File-access functions...
	bool				OpenFile		(const char* const szFilename, const bool bForceBlocking = false, const u_int uStartOffset = 0 );
	u_int				ReadData		(const u_int uNumBytes, u_char* pucDest);
	u_int				SkipData		(const u_int uNumBytes);
	bool				CloseFile		(const bool bForceBlocking = false);

	// Cache manipulation functions...
	void				Invalidate		(void)							{ m_bDataValid	= false; }

	// Accessor functions...
	const char*			GetFilename		(void) const					{ return(Asura_FileAccess::GetFilename(m_iFileHandle)); }
	u_int				GetFilePosition	(void) const					{ return(m_uFilePosition); }
	u_int				GetFileSize		(void) const					{ return(m_uFileSize); }

	bool				CanBeRead		(void) const					{ return(m_bCanRead && m_bDataValid && !m_xUpdateReadyEvent.IsSignalled()); }
	bool				CanRead			(const u_int uNumBytes) const	{ return(uNumBytes <= m_uValidSizeRemaining); }
// ART - 21/02/07 - Replaced the following code with the line above. None of these functions
// should access more than one member variable since they're all volatile and can cause race-
// conditions. However, the code must have been added for a reason - hopefully, this will be
// rectified when the file-cache is stripped back and re-written.
//	bool				CanRead			(const u_int uNumBytes) const	{ return(m_bCanRead && m_bDataValid && (uNumBytes <= m_uValidSizeRemaining)); }

	bool				IsOpening		(void) const					{ return(m_eState == OPENING); }
	bool				IsOpen			(void) const					{ return(m_iFileHandle >= 0); }
	bool				IsAtEnd			(void) const					{ return(m_uFilePosition >= m_uFileSize); }
	bool				IsClosing		(void) const					{ return(m_eState == CLOSING); }

	void				SetCanRead(bool bCanRead) { m_bCanRead = bCanRead; }

protected:
	// Protected variables...

	// Protected functions...


private:
	// Private enumerated type...
	enum State
	{
		IDLE,
		OPENING,
		SEEKING,
		READING,
		CLOSING,

		MAX_STATES
	};

	// Private variables...
	char					m_acName[uASURA_FILECACHE_MAX_NAME_LENGTH];				// The name of this file-cache.
	char					m_acFilename[uASURA_FILECACHE_MAX_FILENAME_LENGTH];		// The name of the current file in this cache.

	Asura_Mutex				m_xMutex;				// Mutex used to allow the data-cache to be used when multi-threading.

	Asura_SyncEvent			m_xUpdateReadyEvent;	// Event used to synchronise update operations.

	volatile bool			m_bAllocated;			// Flag to say whether we allocated this buffer.
	u_char* volatile		m_pucBuffer;			// A pointer to the start of the buffer.
	volatile u_int			m_uBufferSize;			// The total size of the buffer.
	volatile u_int			m_uNumSections;			// The number of sections the cache is split into.
	volatile u_int			m_uSectionSize;			// The size of data we're actually going to cache.

	bool					m_bAllowBlockingRead;	// Flag to say whether we allow reads to block until the cache contains the necessary data.

	volatile State			m_eState;				// The current state of reading from this cache.
	volatile State			m_eStateFollowingError;	// Replaced bool to allow open to fail and be retried as well as read.
	volatile bool			m_bInRefresh;			// Used to prevent refresh-within-refresh situations

	volatile Asura_Handle	m_iFileHandle;			// The file-handle for the file currently open.
	volatile u_int			m_uFilePosition;		// The current position we've actually read up to in the currently open file.
	volatile u_int			m_uFileSize;			// The size of the file we're currently reading.
	volatile u_int			m_uStartOffset;			// the offset in the file where to start reading from (usually 0)

	volatile bool			m_bDataValid;			// Flag to say whether the data we currently contain is valid.
	volatile bool			m_bCanRead;				// Flag to prevent reading while we're in the automated read loop.
	volatile u_int			m_uFirstValidOffset;	// The offset into the buffer of the first byte of valid data.
	volatile u_int			m_uLastValidOffset;		// The offset into the buffer of the last byte of valid data.
	volatile u_int			m_uValidSizeRemaining;	// The number of bytes left within the cache which are valid.
	volatile u_int			m_uCurrentOffset;		// The current offset to read from within the buffer.
	volatile u_int			m_uBytesLeftToCache;	// The number of bytes we've got left in the currently opened file.


	// Private functions...
	Asura_FileCache  (void);
	Asura_FileCache  (const Asura_FileCache& xOther);

	Asura_FileCache&	operator=			(const Asura_FileCache& xOther);

	// Refresh functions...
	static	void				UpdateThread		(const void* pUserData);

	int					GetSectionToRefresh	(void);
	bool				ReadDuringRefresh	(const u_int uNumBytes, u_char* pucDest);
	
#ifdef ASURA_USE_DEBUG_REFRESH
	bool				Refresh				(const char* szFunction, int iLine);
#else
	bool				Refresh				(void);
#endif

	// Internal file-access functions...
	bool				OpenFile_Internal	(void);
	bool				SeekFile_Internal	(void);
	bool				CloseFile_Internal	(void);	

};   // End of class 'Asura_FileCache'.


#endif   // ifndef __ASURA_FILECACHE_H__


/// End of file. ///
