//////////////////////////////////////////////////////////////////////////////
//
// Asura_GUIMenu_Animator.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     classes to manage animation of GUIMenu elements and widgets, e.g.
//     changing position, height, colour etc
//     Some code is based on Asura_TextureAnimation, but has enough 
//     differences to make sharing the same code impractical.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
// This file's header.
#include "GUIMenu/Asura_GUIMenu_Animator.h"

//Asura
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "GUIMenu/Asura_GUIMenu_Element_Base.h"
#include "GUIMenu/Asura_GUIMenu_Root.h"
#include "GUIMenu/Asura_GUIMenu_System.h"

#ifdef ASURA_TOOLS
//Tools
#include "GUIMenu/Tools_GUIMenu_AnimEdit.h"
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif


// G L O B A L S /////////////////////////////////////////////////////////////
static Asura_TemplateHashTree<Asura_GUIMenu_Animator_Sequence*> g_xSequenceTree;
static bool g_bShuttingDown = false;

float Asura_GUIMenu_Animator_Sequence::s_fGlobalAnimationTimer = 0.0f;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// Asura_GUIMenu_Animator_Sequence
//////////////////////////////////////////////////////////////////////////////
Asura_GUIMenu_Animator_Sequence::Asura_GUIMenu_Animator_Sequence( const char* szName )
:m_uStateFlags(0)
,m_uModifyType(ASURA_GUIMENU_MODIFY_TYPE_INVALID)
,m_eDataType(ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INVALID)
,m_uNumberOfFrames(0)
,m_pfFrameTimes(NULL)
,m_pFrameData(NULL)
,m_fDuration(0.0f)
,m_uFlags(0)
#ifdef ASURA_TOOLS
,m_xName(szName)
,m_bUseGlobalTime(false)
,m_bSmoothConsoleVar(false)
,m_fSmoothFrameTime(0.0f)
,m_eEaseType(ASURA_GUIMENU_ANIMATOR_EASE_NONE)
#endif
{
	m_uHashID = Asura_GetHashID( szName );

	Asura_GUIMenu_Animator_Storage::AddSequenceToTree( *this );
}

Asura_GUIMenu_Animator_Sequence::~Asura_GUIMenu_Animator_Sequence()
{
	Asura_GUIMenu_Animator_Storage::RemoveSequenceFromTree( *this );
	delete[] m_pfFrameTimes;
	Asura_Deallocate(m_pFrameData);
}

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Animator_Sequence::Copy(const Asura_GUIMenu_Animator_Sequence* pxSeq)
{
	if(this == pxSeq) return;

	//clear out old data
	delete[] m_pfFrameTimes;
	m_pfFrameTimes = 0;
	Asura_Deallocate(m_pFrameData);
	m_pFrameData = 0;

	//and copy in the new
	m_uStateFlags = pxSeq->m_uStateFlags;
	m_uModifyType = pxSeq->m_uModifyType;
	m_eDataType = pxSeq->m_eDataType;

	m_uNumberOfFrames = pxSeq->m_uNumberOfFrames;
	m_pfFrameTimes = new float[m_uNumberOfFrames];
	for(u_int i=0; i<m_uNumberOfFrames; i++)
	{
		m_pfFrameTimes[i] = pxSeq->m_pfFrameTimes[i];
	}

	m_pFrameData = AllocateFrameData( m_uNumberOfFrames, m_eDataType );
	Asura_Mem::Copy(m_pFrameData, pxSeq->m_pFrameData, GetFrameDataSize(m_eDataType)*m_uNumberOfFrames);
	
	m_fDuration = pxSeq->m_fDuration;
	m_uFlags = pxSeq->m_uFlags;

	// Copy the console var
	m_xConsoleVar.Copy(&pxSeq->m_xConsoleVar);

	// Copy the global timer flag
	m_bUseGlobalTime = pxSeq->m_bUseGlobalTime;

	m_bSmoothConsoleVar = pxSeq->m_bSmoothConsoleVar;

	m_fSmoothFrameTime = pxSeq->m_fSmoothFrameTime;

	m_eEaseType = pxSeq->m_eEaseType;
}
#endif //ASURA_TOOLS


void Asura_GUIMenu_Animator_Sequence::Update()
{
	// Increment the global timer for this frame.
	s_fGlobalAnimationTimer += Asura_Timers::GetFrameTime();
}

float Asura_GUIMenu_Animator_Sequence::GetFrameTime( u_int u ) const
{
	Asura_Assert(u < m_uNumberOfFrames, "invalid frame");
	return m_pfFrameTimes[u];
}

void* Asura_GUIMenu_Animator_Sequence::GetFrameData( u_int u ) const
{
	Asura_Assert(u < m_uNumberOfFrames, "invalid frame");
	return static_cast<char*>(m_pFrameData) + (u * GetFrameDataSize(m_eDataType));
}

// static function to deal with variable frame data size
u_int Asura_GUIMenu_Animator_Sequence::GetFrameDataSize( ASURA_GUIMENU_ANIMATOR_DATA_TYPE eDataType )
{
	switch (eDataType)
	{
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT:
		{
			return sizeof(float);
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT:
		{
			return sizeof(int);
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2:
		{
			return sizeof(Asura_Vector_2);
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET:
		{
			return sizeof(Asura_GUIMenu_UV_Set);
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA:
		{
			return sizeof(Asura_RGBA);
		}
		default:
		{
			Asura_Assert(false, "invalid data type");
		}
	}
	return 0;
}

// static function to allocate frame data
void* Asura_GUIMenu_Animator_Sequence::AllocateFrameData( u_int uNumFrames, 
															ASURA_GUIMENU_ANIMATOR_DATA_TYPE eDataType )
{
	u_int uTypeSize = GetFrameDataSize(eDataType);
	if (!uTypeSize)
	{
		return NULL;
	}

	u_int uSize = uTypeSize * uNumFrames;
	void* pFrameData = Asura_Allocate(uSize);
	Asura_Mem::Set( pFrameData, 0, uSize );
	return pFrameData;
}

// static function to deal with variable frame size
void Asura_GUIMenu_Animator_Sequence::CopyFrameData( void* pDestArray, u_int uDestFrame, const void* pSrcArray, 
													u_int uSrcFrame, ASURA_GUIMENU_ANIMATOR_DATA_TYPE eDataType )
{
	switch (eDataType)
	{
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT:
		{
			float* pxDest = &static_cast<float*>(pDestArray)[uDestFrame];
			const float* pxSrc = &static_cast<const float*>(pSrcArray)[uSrcFrame];
			*pxDest = *pxSrc;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT:
		{
			int* pxDest = &static_cast<int*>(pDestArray)[uDestFrame];
			const int* pxSrc = &static_cast<const int*>(pSrcArray)[uSrcFrame];
			*pxDest = *pxSrc;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2:
		{
			Asura_Vector_2* pxDest = &static_cast<Asura_Vector_2*>(pDestArray)[uDestFrame];
			const Asura_Vector_2* pxSrc = &static_cast<const Asura_Vector_2*>(pSrcArray)[uSrcFrame];
			*pxDest = *pxSrc;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET:
		{
			Asura_GUIMenu_UV_Set* pxDest = &static_cast<Asura_GUIMenu_UV_Set*>(pDestArray)[uDestFrame];
			const Asura_GUIMenu_UV_Set* pxSrc = &static_cast<const Asura_GUIMenu_UV_Set*>(pSrcArray)[uSrcFrame];
			*pxDest = *pxSrc;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA:
		{
			Asura_RGBA* pxDest = &static_cast<Asura_RGBA*>(pDestArray)[uDestFrame];
			const Asura_RGBA* pxSrc = &static_cast<const Asura_RGBA*>(pSrcArray)[uSrcFrame];
			*pxDest = *pxSrc;
			break;
		}
		default:
		{
			Asura_Assert(false, "invalid data type");
			break;
		}
	}
}

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Animator_Sequence::Initialise( ASURA_GUIMENU_MODIFY_TYPE eModifyType )
{
	Asura_Assert(!m_uNumberOfFrames,"This should be called on a newly created sequence");
	Asura_Assert(eModifyType != ASURA_GUIMENU_MODIFY_TYPE_INVALID, "invalid modify type");
	switch (eModifyType)
	{
		case ASURA_GUIMENU_MODIFY_POSITION:
		case ASURA_GUIMENU_MODIFY_POSITION_RELATIVE:
		case ASURA_GUIMENU_MODIFY_SIZE_SCALE:
		{
			m_eDataType = ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2;
			break;
		}

		case ASURA_GUIMENU_MODIFY_TEXT_OFFSET_X :
		case ASURA_GUIMENU_MODIFY_TEXT_OFFSET_Y :
		case ASURA_GUIMENU_MODIFY_TEXT_SCROLL_RATE :
		case ASURA_GUIMENU_MODIFY_ALPHA_SCALED:
		case ASURA_GUIMENU_MODIFY_WIDTH_SCALED:
		case ASURA_GUIMENU_MODIFY_HEIGHT_SCALED:
		case ASURA_GUIMENU_MODIFY_FONT_SCALE:
		case ASURA_GUIMENU_MODIFY_IMAGE_ROTATE:
			m_eDataType = ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT;
			break;

		case ASURA_GUIMENU_MODIFY_UVS :
			m_eDataType = ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET;
			break;

		case ASURA_GUIMENU_MODIFY_COLOUR :
			m_eDataType = ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA;
			break;

		default:
		{
			Asura_Assert(false, "unhandled type of data to modify. Need to add code.");
			return;
		}
	}

	m_uModifyType = eModifyType;

	m_fDuration = 1.0f;

	m_uNumberOfFrames = 1;
	m_pfFrameTimes = new float[1];
	*m_pfFrameTimes = 0.0f;
	m_pFrameData = AllocateFrameData(1, m_eDataType);

	
	
	//Set default data for new frame - only needed for cases where zero isn't suitable.
	//RAR 25/07/2007
	switch(m_eDataType)
	{
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA :
			SetFrameData(0, Asura_RGBA(1.0f,1.0f,1.0f,1.0f));
			break;
		default :
			break;
	}

}

void Asura_GUIMenu_Animator_Sequence::AddFrameAfter(int iPrevFrame)
{
	Asura_Assert(iPrevFrame>=-1 && iPrevFrame<(int)m_uNumberOfFrames,"");
	float* pfFrameTimes = new float[m_uNumberOfFrames+1];
	void* pFrameData = AllocateFrameData(m_uNumberOfFrames+1, m_eDataType);

	//copy over the old frames
	u_int uDestFrame = 0;
	for(u_int i=0;i<m_uNumberOfFrames;i++,uDestFrame++)
	{
		if((int)i==iPrevFrame+1) uDestFrame++;

		pfFrameTimes[uDestFrame] = m_pfFrameTimes[i];
		CopyFrameData( pFrameData, uDestFrame, m_pFrameData, i, m_eDataType );
	}
	//initialise the new one
	int iNewFrame = iPrevFrame+1;
	if(iNewFrame)
	{
		pfFrameTimes[iNewFrame] = m_pfFrameTimes[iNewFrame-1];
		CopyFrameData( pFrameData, iNewFrame, pFrameData, iNewFrame-1, m_eDataType );
	}

	//and replace
	delete[] m_pfFrameTimes;
	Asura_Deallocate(m_pFrameData);
	m_pfFrameTimes = pfFrameTimes;
	m_pFrameData = pFrameData;
	m_uNumberOfFrames++;
}

void Asura_GUIMenu_Animator_Sequence::DeleteFrame(u_int uFrame)
{
	Asura_Assert(m_uNumberOfFrames>1,"");
	Asura_Assert(uFrame<m_uNumberOfFrames,"");

	//shuffle the frames down to replace the deleted one
	m_uNumberOfFrames--;
	for(u_int i=uFrame;i<m_uNumberOfFrames-1;i++)
	{
		m_pfFrameTimes[i] = m_pfFrameTimes[i+1];
		CopyFrameData(m_pFrameData, i, m_pFrameData, i+1, m_eDataType);
	}
	//make sure frame 0 still starts at time 0
	m_pfFrameTimes[0] = 0;

}

void Asura_GUIMenu_Animator_Sequence::SetFrameTime( u_int uFrame, float fTime )
{
	Asura_Assert(uFrame < m_uNumberOfFrames, "invalid frame");
	m_pfFrameTimes[uFrame] = fTime;
}

void Asura_GUIMenu_Animator_Sequence::SetFrameData( u_int uFrame, const Asura_Vector_2& xVec )
{
	Asura_Assert( m_eDataType == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2, "incorrect data type");
	Asura_Vector_2* pxVec = static_cast<Asura_Vector_2*>(GetFrameData(uFrame));
	*pxVec = xVec;
}

void Asura_GUIMenu_Animator_Sequence::SetFrameData( u_int uFrame, float fValue )
{
	Asura_Assert( m_eDataType == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT, "incorrect data type");
	float* pfData = static_cast<float*>(GetFrameData(uFrame));
	*pfData = fValue;
}

void Asura_GUIMenu_Animator_Sequence::SetFrameData( u_int uFrame, int iValue )
{
	Asura_Assert( m_eDataType == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT, "incorrect data type");
	int* piData = static_cast<int*>(GetFrameData(uFrame));
	*piData = iValue;
}

void Asura_GUIMenu_Animator_Sequence::SetFrameData( u_int uFrame, const Asura_GUIMenu_UV_Set& xUV )
{
	Asura_Assert( m_eDataType == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET, "incorrect data type");
	Asura_GUIMenu_UV_Set* pxUV = static_cast<Asura_GUIMenu_UV_Set*>(GetFrameData(uFrame));
	*pxUV = xUV;
}

void Asura_GUIMenu_Animator_Sequence::SetFrameData( u_int uFrame, const Asura_RGBA& xColour )
{
	Asura_Assert( m_eDataType == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA, "incorrect data type");
	Asura_RGBA* pxColour = static_cast<Asura_RGBA*>(GetFrameData(uFrame));
	*pxColour = xColour;
}



void Asura_GUIMenu_Animator_Sequence::Rename( const char* szNewName )
{ 
	m_xName = szNewName; 
	m_uHashID = Asura_GetHashID( szNewName );
}
#endif //ASURA_TOOLS

// chunk load/save/skip
bool Asura_GUIMenu_Animator_Sequence::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;
	if (uVersion > uASURA_GUIMENU_ANIMATOR_SEQUENCE_CHUNKVERSION)
	{
		Asura_Assertf(false, ("GUIMenu Animator Sequence version unknown (%u > %u)", 
			uVersion, uASURA_GUIMENU_ANIMATOR_SEQUENCE_CHUNKVERSION));
		return false;
	}

	Asura_Assert(!m_uNumberOfFrames, "already set up");

	xStream >> m_uStateFlags;
	xStream >> m_uModifyType;

	Asura_ReadFromStream_Enum_Wrapper<ASURA_GUIMENU_ANIMATOR_DATA_TYPE> xDataTypeEnumWrapper(m_eDataType);
	xStream >> xDataTypeEnumWrapper;
	
	xStream >> m_uNumberOfFrames;
	Asura_Assert(m_uNumberOfFrames < 1024, "suspiciously-large number of frames");
	m_pfFrameTimes = new float[m_uNumberOfFrames];
	m_pFrameData = AllocateFrameData( m_uNumberOfFrames, m_eDataType );

	Asura_ReadFromStream_Array_Wrapper<float> xArrayTimesWrapper(m_uNumberOfFrames, m_pfFrameTimes);
	xStream >> xArrayTimesWrapper;

	switch(m_eDataType)
	{
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT:
		{
			Asura_ReadFromStream_Array_Wrapper<float> xArrayWrapper(m_uNumberOfFrames, static_cast<float*>(m_pFrameData));
			xStream >> xArrayWrapper;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT:
		{
			Asura_ReadFromStream_Array_Wrapper<int>xArrayWrapper(m_uNumberOfFrames, static_cast<int*>(m_pFrameData));
			xStream >> xArrayWrapper;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2:
		{
			Asura_ReadFromStream_Array_Wrapper<Asura_Vector_2>xArrayWrapper(m_uNumberOfFrames, static_cast<Asura_Vector_2*>(m_pFrameData));
			xStream >> xArrayWrapper;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET:
		{
			Asura_ReadFromStream_Array_Wrapper<float>xArrayWrapper(m_uNumberOfFrames*4, static_cast<float*>(m_pFrameData));
			xStream >> xArrayWrapper;
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA:
		{
			Asura_ReadFromStream_Array_Wrapper<float>xArrayWrapper(m_uNumberOfFrames*4, static_cast<float*>(m_pFrameData));
			xStream >> xArrayWrapper;
			break;
		}
		default:
		{
			Asura_Assert(false, "invalid data type");
			break;
		}
	}
	xStream >> m_fDuration;
	
	xStream >> m_uFlags;
	if (uVersion < 1) m_bPlayToEnd = false;

	// Read the console var and global time flag
	if (uVersion >= 2 )
	{
		m_xConsoleVar.ReadFromChunkStream(xStream);

		xStream >> m_bUseGlobalTime;
	}

	// Read the smooth console var flag and smooth time
	if (uVersion >= 3 )
	{
		xStream >> m_bSmoothConsoleVar;

		xStream >> m_fSmoothFrameTime;
	}

	if (uVersion >= 4 )
	{
		// read the ease setting
		u_int uEaseType;
		xStream >> uEaseType;
		m_eEaseType = static_cast<ASURA_GUIMENU_ANIMATOR_EASE_TYPE>(uEaseType);
	}

	return true;
}

// static Skip function
bool Asura_GUIMenu_Animator_Sequence::SkipChunkData( Asura_Chunk_Stream& xStream )
{
  	u_int uVersion;
  	xStream >> uVersion;
  	if (uVersion > uASURA_GUIMENU_ANIMATOR_SEQUENCE_CHUNKVERSION)
  	{
  		Asura_Assertf(false, ("GUIMenu Animator Sequence version unknown (%u > %u)", 
  			uVersion, uASURA_GUIMENU_ANIMATOR_SEQUENCE_CHUNKVERSION));
  		return false;
  	}

  	u_int								uStateFlags;
  	u_int								uDataToModify;
  	ASURA_GUIMENU_ANIMATOR_DATA_TYPE	eDataType;
  	u_int								uNumberOfFrames;
  	float								fDuration;
  	u_int								uFlags;
	Asura_GUIMenu_ConsoleVar			xConsoleVar;
	bool								bUseGlobalTime;
	bool								bSmoothVar;
	float								fSmoothFrameTime;
	u_int								iEaseType;

  	xStream >> uStateFlags;
  	xStream >> uDataToModify;

  	Asura_ReadFromStream_Enum_Wrapper<ASURA_GUIMENU_ANIMATOR_DATA_TYPE> xDataTypeEnumWrapper(eDataType);
  	xStream >> xDataTypeEnumWrapper;
  	
  	xStream >> uNumberOfFrames;
  	Asura_Assert(uNumberOfFrames < 1024, "suspiciously-large number of frames");

  	xStream.Skip( sizeof(float) * uNumberOfFrames );
  	xStream.Skip( GetFrameDataSize(eDataType) * uNumberOfFrames );

  	xStream >> fDuration;
  	xStream >> uFlags;

	// Skip over the console var and global time flag if this sequence is already loaded
	if (uVersion >= 2 )
	{
		xConsoleVar.ReadFromChunkStream(xStream);

		xStream >> bUseGlobalTime;
	}

	// Skip over the smooth console var flag and smooth time
	if (uVersion >= 3 )
	{
		xStream >> bSmoothVar;

		xStream >> fSmoothFrameTime;
	}

	if (uVersion >= 4 )
	{
		// read the ease setting
		xStream >> iEaseType;
	}

	return true;
}

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Animator_Sequence::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << uASURA_GUIMENU_ANIMATOR_SEQUENCE_CHUNKVERSION;

	xStream << m_uStateFlags;
	xStream << m_uModifyType;

	Asura_WriteToStream_Enum_Wrapper<ASURA_GUIMENU_ANIMATOR_DATA_TYPE> xDataTypeEnumWrapper(m_eDataType);
	xStream << xDataTypeEnumWrapper;
	
	xStream << m_uNumberOfFrames;
	xStream << Asura_WriteToStream_Array_Wrapper<float>(m_uNumberOfFrames, m_pfFrameTimes);
	switch(m_eDataType)
	{
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT:
		{
			xStream << Asura_WriteToStream_Array_Wrapper<float>(m_uNumberOfFrames, static_cast<float*>(m_pFrameData));
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT:
		{
			xStream << Asura_WriteToStream_Array_Wrapper<int>(m_uNumberOfFrames, static_cast<int*>(m_pFrameData));
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2:
		{
			xStream << Asura_WriteToStream_Array_Wrapper<Asura_Vector_2>(m_uNumberOfFrames, static_cast<Asura_Vector_2*>(m_pFrameData));
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET:
		{
			xStream << Asura_WriteToStream_Array_Wrapper<float>(m_uNumberOfFrames*4, static_cast<float*>(m_pFrameData));
			break;
		}
		case ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA:
		{
			xStream << Asura_WriteToStream_Array_Wrapper<float>(m_uNumberOfFrames*4, static_cast<float*>(m_pFrameData));
			break;
		}
		default:
		{
			Asura_Assert(false, "invalid data type");
			break;
		}
	}
	xStream << m_fDuration;

	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xFlagsBitfield (m_uFlags);
	xFlagsBitfield.Add(m_bLoop, 1);
	xFlagsBitfield.Add(m_bInterpolate, 1);
	xFlagsBitfield.Add(m_bLoopBackAndForth, 1); 
	xFlagsBitfield.Add(m_bPlayToEnd, 1);
	
	xStream << xFlagsBitfield;

	// Write the console var
	m_xConsoleVar.WriteToChunkStream( xStream);

	// Write the global time flag
	xStream << m_bUseGlobalTime;

	// Write the smooth console var flag 
	xStream << m_bSmoothConsoleVar;

	// Write the smooth frame time
	xStream << m_fSmoothFrameTime;

	// Write the ease setting
	u_int uEaseType = m_eEaseType;
	xStream << uEaseType;
}
#endif

//////////////////////////////////////////////////////////////////////////////
// Asura_GUIMenu_Animator
//////////////////////////////////////////////////////////////////////////////

Asura_GUIMenu_Animator::Asura_GUIMenu_Animator()
: m_pxSequence(NULL)
, m_fTimer(0.0f)
, m_fSpeed(1.0f)
, m_uFrameStamp(0xffffffff)
, m_fDelay(0.0f)
, m_fDelayTimer(0.0f)
#ifdef ASURA_TOOLS
, m_uToolsFlags(0)
#endif
, m_bShouldStartActive(true)       // default setting maintains old behaviour (chunk version < 2) - NDB 070208
, m_bLastKnownIsStartConditionMet(false)
, m_bInitialiseToEnd(false)
{
}

Asura_GUIMenu_Animator::Asura_GUIMenu_Animator(const Asura_GUIMenu_Animator& xOther)
{
	*this = xOther;
}

Asura_GUIMenu_Animator& Asura_GUIMenu_Animator::operator=(const Asura_GUIMenu_Animator& xOther)
{
	if (this == &xOther) 
	{
		Asura_Assert(false, "unexpected copy-to-self");
		return *this;
	}

#ifdef ASURA_TOOLS
	m_xSequenceName = (xOther.m_xSequenceName);
#endif

	m_xStartConditions.Copy( &(xOther.m_xStartConditions) );

	m_pxSequence	= (xOther.m_pxSequence);
	m_fTimer		= (0.0f);
	m_fSpeed		= (xOther.m_fSpeed);
	m_fDelay		= (xOther.m_fDelay);
	m_fDelayTimer	= (xOther.m_fDelay);//I guess to be consistent with timer copying , this should be copied as the initial state
	m_uFrameStamp	= (0xffffffff);

#ifdef ASURA_TOOLS
	m_uToolsFlags	= (xOther.m_uToolsFlags);
#endif

	m_bShouldStartActive		= xOther.m_bShouldStartActive;
	m_bLastKnownIsStartConditionMet  = xOther.m_bLastKnownIsStartConditionMet;
	
	m_bInitialiseToEnd	= xOther.m_bInitialiseToEnd;

	return *this;
}

void Asura_GUIMenu_Animator::Initialise(bool bInitMenuPage /*= false*/)
{
	// If the animator is being initialised when the menu page is being initialised
	// then this is an indication that the menu page is being opened.
	// In this case the animation may need to initialised to its end point.
	// This is needed so the animation doesn't play from the beginning when the 
	// menu page opens and is set so it appears to have already played.
	// This stops the animation from running as though its a transition animation, but
	// will still let the animation run if it get reset while the page is open.
	if ( bInitMenuPage == true )
	{
		if ( m_bInitialiseToEnd == true)
		{
			Finish();
			m_uFrameStamp = Asura_Timers::GetFrameNumber();
			return;
		}
	}

	if( m_bShouldStartActive )
	{
		Start();
	}
	else
	{
		Finish();
	}
}

bool Asura_GUIMenu_Animator::ShouldUseAnimator(u_int uStateFlags) const
{
	#ifdef ASURA_TOOLS
	//check whether this is being overridden by tools
	if(Tools_GUIMenu_AnimEdit::IsToolsOverride(m_pxSequence)) return true;
	#endif

	return ((uStateFlags & GetStateFlags())!=0 || IsPlayingToEnd());
}

void Asura_GUIMenu_Animator::UpdateStartCondition()
{
	// if there are no condition vars then do nothing
	if( m_xStartConditions.m_uNumConditionVars )
	{
		// if condition has changed from false to true then restart animator
		const bool bConditionMet( IsStartConditionMet() );

		if( bConditionMet && !m_bLastKnownIsStartConditionMet )
		{
			Start();
		}
		
		m_bLastKnownIsStartConditionMet = bConditionMet;
	}
}

void Asura_GUIMenu_Animator::Finish()
{
	m_fTimer = GetDuration();
	m_fDelayTimer = 0.0f;
}

// returns widget state this animator applies to
u_int Asura_GUIMenu_Animator::GetStateFlags() const
{
	Asura_Assert( m_pxSequence, "no sequence");
	if (m_pxSequence)
	{
		return m_pxSequence->GetStateFlags();
	}
	return 0;
}

// returns what is being changed (e.g. position i.e. ASURA_GUIMENU_MODIFY_POSITION)
u_int Asura_GUIMenu_Animator::GetModifyType() const
{
	Asura_Assert( m_pxSequence, "no sequence");
	if (m_pxSequence)
	{
		return m_pxSequence->GetModifyType();
	}
	return 0;
}

bool Asura_GUIMenu_Animator::IsLooping() const
{
	Asura_Assert( m_pxSequence, "no sequence");
	if (m_pxSequence)
	{
		return m_pxSequence->GetLoop() || m_pxSequence->GetLoopBackAndForth();
	}
	return false;
}

float Asura_GUIMenu_Animator::GetDuration() const
{
	Asura_Assert( m_pxSequence, "no sequence");
	if (m_pxSequence)
	{
		if (GetSpeed() != 0.0f)
		{
			return Asura_Maths::Abs(m_pxSequence->GetDuration() / GetSpeed());
		}
	}
	return 0.0f;
}

Asura_Vector_2 Asura_GUIMenu_Animator::GetVector2Data()
{
	Asura_Assert( m_pxSequence && m_pxSequence->GetDataType() == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2, "incorrect data type");

	u_int uFrame;
	u_int uNextFrame;
	float fLerp;
	if (Update(uFrame, uNextFrame, fLerp))
	{
		const Asura_Vector_2* pxFrameVec2 = static_cast<Asura_Vector_2*>(m_pxSequence->GetFrameData(uFrame));
		if (fLerp != 0.0f)
		{
			const Asura_Vector_2* pxNextFrameVec2 = static_cast<Asura_Vector_2*>(m_pxSequence->GetFrameData(uNextFrame));
			return *pxFrameVec2 + (*pxNextFrameVec2 - *pxFrameVec2) * fLerp;
		}
		else
		{
			return *pxFrameVec2;
		}
	}
	return Asura_Maths::ZeroVector2;
}

float Asura_GUIMenu_Animator::GetFloatData()
{
	Asura_Assert( m_pxSequence && m_pxSequence->GetDataType() == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT, "incorrect data type");

	u_int uFrame;
	u_int uNextFrame;
	float fLerp;
	if (Update(uFrame, uNextFrame, fLerp))
	{
		const float* pfFrame = static_cast<float*>(m_pxSequence->GetFrameData(uFrame));
		if (fLerp != 0.0f)
		{
			const float* pfNextFrame = static_cast<float*>(m_pxSequence->GetFrameData(uNextFrame));
			return *pfFrame + (*pfNextFrame - *pfFrame) * fLerp;
		}
		else
		{
			return *pfFrame;
		}
	}

	return 0.0f;
}

int Asura_GUIMenu_Animator::GetIntData()
{
	Asura_Assert( m_pxSequence && m_pxSequence->GetDataType() == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT, "incorrect data type");

	u_int uFrame;
	u_int uNextFrame;
	float fLerp;
	if (Update(uFrame, uNextFrame, fLerp))
	{
		const int* piFrame = static_cast<int*>(m_pxSequence->GetFrameData(uFrame));
		if (fLerp != 0.0f)
		{
			const int* piNextFrame = static_cast<int*>(m_pxSequence->GetFrameData(uNextFrame));
			return *piFrame + Asura_Maths::FloatToInt(static_cast<float>(*piNextFrame - *piFrame) * fLerp);
		}
		else
		{
			return *piFrame;
		}
	}

	return 0;
}

Asura_GUIMenu_UV_Set Asura_GUIMenu_Animator::GetUVSetData()
{
	Asura_Assert( m_pxSequence && m_pxSequence->GetDataType() == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET, "incorrect data type");

	u_int uFrame;
	u_int uNextFrame;
	float fLerp;
	if (Update(uFrame, uNextFrame, fLerp))
	{
		const Asura_GUIMenu_UV_Set* pxFrameUVSet = static_cast<Asura_GUIMenu_UV_Set*>(m_pxSequence->GetFrameData(uFrame));
		if (fLerp != 0.0f)
		{
			const Asura_GUIMenu_UV_Set* pxNextFrameUVSet = static_cast<Asura_GUIMenu_UV_Set*>(m_pxSequence->GetFrameData(uNextFrame));

			Asura_GUIMenu_UV_Set xUVSet;

			xUVSet.xUV0.u = pxFrameUVSet->xUV0.u + (pxNextFrameUVSet->xUV0.u - pxFrameUVSet->xUV0.u) * fLerp;
			xUVSet.xUV0.v = pxFrameUVSet->xUV0.v + (pxNextFrameUVSet->xUV0.v - pxFrameUVSet->xUV0.v) * fLerp;
			xUVSet.xUV1.u = pxFrameUVSet->xUV1.u + (pxNextFrameUVSet->xUV1.u - pxFrameUVSet->xUV1.u) * fLerp;
			xUVSet.xUV1.v = pxFrameUVSet->xUV1.v + (pxNextFrameUVSet->xUV1.v - pxFrameUVSet->xUV1.v) * fLerp;
			
			return xUVSet;
		}
		else
		{
			return *pxFrameUVSet;
		}
	}
	Asura_GUIMenu_UV_Set xZero;
	xZero.xUV0.u = 0.0f;
	xZero.xUV0.v = 0.0f;
	xZero.xUV1.u = 0.0f;
	xZero.xUV1.v = 0.0f;
	return xZero;
}

Asura_RGBA Asura_GUIMenu_Animator::GetRGBAData()
{
	Asura_Assert( m_pxSequence && m_pxSequence->GetDataType() == ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA, "incorrect data type");

	u_int uFrame;
	u_int uNextFrame;
	float fLerp;
	if (Update(uFrame, uNextFrame, fLerp))
	{
		const Asura_RGBA* pxFrameRGBA = static_cast<Asura_RGBA*>(m_pxSequence->GetFrameData(uFrame));
		if (fLerp != 0.0f)
		{
			const Asura_RGBA* pxNextFrameRGBA = static_cast<Asura_RGBA*>(m_pxSequence->GetFrameData(uNextFrame));

			Asura_RGBA xRGBA;

			xRGBA.m_fR = pxFrameRGBA->m_fR + (pxNextFrameRGBA->m_fR - pxFrameRGBA->m_fR) * fLerp;
			xRGBA.m_fG = pxFrameRGBA->m_fG + (pxNextFrameRGBA->m_fG - pxFrameRGBA->m_fG) * fLerp;
			xRGBA.m_fB = pxFrameRGBA->m_fB + (pxNextFrameRGBA->m_fB - pxFrameRGBA->m_fB) * fLerp;
			xRGBA.m_fA = pxFrameRGBA->m_fA + (pxNextFrameRGBA->m_fA - pxFrameRGBA->m_fA) * fLerp;
			
			return xRGBA;
		}
		else
		{
			return *pxFrameRGBA;
		}
	}
	return Asura_RGBA(1.0f,1.0f,1.0f,1.0f);;
}



#ifdef ASURA_TOOLS
void Asura_GUIMenu_Animator::SetSequence( const char* szName )
{
	m_xSequenceName = szName;
	m_pxSequence = Asura_GUIMenu_Animator_Storage::FindSequence( Asura_GetHashID(szName) );
	Asura_Assertf( m_pxSequence, ("failed to find GUIMenu animator sequence %s", szName));
}
#endif // #ifdef ASURA_TOOLS

bool Asura_GUIMenu_Animator::IsPlayingToEnd( void ) const
{
	return (m_pxSequence && m_pxSequence->GetPlayToEnd() && (m_fTimer != m_pxSequence->GetDuration()) && (m_fTimer != 0.0f) );
}

void Asura_GUIMenu_Animator::Start()	
{ 
	m_fTimer = 0.0f; 
	m_fDelayTimer = m_fDelay;

	if ( m_pxSequence != NULL &&
		 m_pxSequence->GetConsoleVar().GetHash() != ASURA_HASH_ID_UNSET )
	{
		m_pxSequence->GetConsoleVar().GetFloatValue(m_fCurrentConsoleVar);
	}
}


bool Asura_GUIMenu_Animator::Update( u_int& uFrameOut, 
									 u_int& uNextFrameOut,
									 float& fLerpOut )
{
	const Asura_GUIMenu_Animator_Sequence* pxSeq = m_pxSequence;

	Asura_Assert(pxSeq, "no sequence for menu animator");
	if (!pxSeq) return false;

	float fDuration = pxSeq->GetDuration();

	#ifdef ASURA_TOOLS
	//Tools overriding the state of this anim? RAR 17/04/2007
	bool bToolsOverride = Tools_GUIMenu_AnimEdit::IsToolsOverride(m_pxSequence);
	if(bToolsOverride)
	{
		//Is the frame being set directly?
		u_int uFrame = Tools_GUIMenu_AnimEdit::GetFrameOverride();
		if(uFrame < pxSeq->GetNumberOfFrames())
		{
			uFrameOut = uFrame;
			uNextFrameOut = uFrame;
			fLerpOut = 0.0f;
			return true;
		}
	}
	#endif //ASURA_TOOLS

	//get a copy of the timer
	float fTimer = m_fTimer;

	// Has the animation been updated this frame.
	// Update is called every time the value needs to be calculated. Therefore it
	// can be called many times in one frame. Only update the timer once per frame.
	bool bUpdateTimer = false;

	if(m_uFrameStamp != Asura_Timers::GetFrameNumber())
	{
		bUpdateTimer = true;

		//Restart the animator if it wasn't updated last frame. RAR 13/0/2008
		if(m_uFrameStamp+1 != Asura_Timers::GetFrameNumber())
		{
			Initialise(false); // take account of should-start-when-page-opens flag - NDB 170608
			fTimer = m_fTimer;
		}

		float fFrameTime = Asura_GUIMenu_System::GetFrameTime();

		//If there is any delay left , then use up some/all of the frame time on that
		if(m_fDelayTimer > 0.0f)
		{
			if(m_fDelayTimer > fFrameTime)
			{
				m_fDelayTimer -= fFrameTime;
				fFrameTime = 0.0f;
			}
			else
			{
				fFrameTime -= m_fDelayTimer;
				m_fDelayTimer = 0.0f;
			}
		}

		//update the anim timer
		if(GetSpeed() != 0.0f)
		{
			fTimer += fFrameTime * GetSpeed();
		}
		m_uFrameStamp = Asura_Timers::GetFrameNumber();
	}

	#ifdef ASURA_TOOLS
	//tools override for timer?
	if(bToolsOverride)
	{
		fTimer = Tools_GUIMenu_AnimEdit::GetTimerOverride();
	}
	#endif 

	if ( pxSeq->GetUseGlobalTime() == true )
	{
		// Don't use the local timer. Use the global timer.
		fTimer = Asura_GUIMenu_Animator_Sequence::GetGlobalTimer();
	}

	//double the duration for the purposes of looping if going back and forth
	float fModifiedDuration = pxSeq->GetLoopBackAndForth() ? (fDuration * 2.0f) : fDuration;

	//make sure its in range 
	if(fTimer > fModifiedDuration)
	{
		//end of anim.. loop or clamp as needed
		if(pxSeq->GetLoop() || pxSeq->GetLoopBackAndForth())
		{
			if(fModifiedDuration)
			{
				fTimer -= Asura_Maths::FloatToInt_Chop(fTimer/fModifiedDuration) * fModifiedDuration;
				fTimer = Asura_Maths::ClampToRange(fTimer, 0.0f, fModifiedDuration);
			}
			else
			{
				fTimer = 0.0f;
			}
		}
		else
		{
			fTimer = fModifiedDuration;
		}
	}
	else if(fTimer < 0.0f) //in case speed is negative
	{
		if( fModifiedDuration && (pxSeq->GetLoop() || pxSeq->GetLoopBackAndForth()) )
		{
			fTimer -= ( Asura_Maths::FloatToInt_Chop(fTimer/fModifiedDuration) - 1.0f ) * fModifiedDuration;
			fTimer = Asura_Maths::ClampToRange(fTimer, 0.0f, fModifiedDuration);
		}
		else
		{
			fTimer = 0.0f;
		}
	}

	// update timer with computed value
	m_fTimer = fTimer;

	if(pxSeq->GetLoopBackAndForth())
	{

		//adjust the timer if its in the back part of its journey
		if(fTimer > fDuration)
		{
			fTimer = fModifiedDuration - fTimer;
			fTimer = Asura_Maths::ClampToRange(fTimer, 0.0f, fDuration);
		}
	}

	Asura_Assert(fDuration, "dividing by zero");
	float fProportionalTime = fTimer / fDuration;

#ifdef ASURA_TOOLS
	//If tools override is not set the apply the console var
	//Tools override is not set when in preview mode.
	//The override is set when editing in tools mode
	if(bToolsOverride==false)
	{
#endif
		// If the animation sequence has a console var use that to calculate the 
		// proportional time
		if ( pxSeq->GetConsoleVar().GetHash() != ASURA_HASH_ID_UNSET )
		{
			Asura_CommandConsole_VarRepository::Var* pxVar = Asura_CommandConsole_VarRepository::Find( pxSeq->GetConsoleVar().GetName().GetString());
			float fConsoleValue;

			if ( pxVar != NULL)
			{
				pxSeq->GetConsoleVar().GetFloatValue(fConsoleValue);
				fProportionalTime = (fConsoleValue - pxVar->m_xData.fLowerLimit) / (pxVar->m_xData.fUpperLimit - pxVar->m_xData.fLowerLimit);

				if ( pxSeq->GetSmoothConsoleVar() == true && 
					 pxSeq->GetSmoothFrameTime() > 0.0f)
				{
					// Smooth out console var over a specified time
					// e.g. move from 0-1 over 5 seconds.
					// Don't update the smoothing value if it has already been updated this frame.
					if ( bUpdateTimer == true )
					{
						float deltaMove = (Asura_GUIMenu_System::GetFrameTime() * (1.0f / pxSeq->GetSmoothFrameTime()));

						if ( m_fCurrentConsoleVar < fProportionalTime )
						{
							m_fCurrentConsoleVar += deltaMove;
							// Don't allow animation to run ahead of where it needs to be.
							if ( m_fCurrentConsoleVar > fProportionalTime)
							{
								m_fCurrentConsoleVar = fProportionalTime;
							}
						}
						else if ( m_fCurrentConsoleVar > fProportionalTime )
						{
							m_fCurrentConsoleVar -= deltaMove;
							// Don't allow animation to run ahead of where it needs to be.
							if ( m_fCurrentConsoleVar < fProportionalTime)
							{
								m_fCurrentConsoleVar = fProportionalTime;
							}
						}
					}

					fProportionalTime = m_fCurrentConsoleVar;
				}
			}
		}
#ifdef ASURA_TOOLS
	}
#endif

	// Apply ease in/out flags. Apply a curve to proportionTime (0.0f - 1.0f).
	if ( pxSeq->GetEaseType() == ASURA_GUIMENU_ANIMATOR_EASE_IN_OUT )
	{
		// Apply ease in/out curve
		// Accelerate up then slow down.
		fProportionalTime = fProportionalTime * fProportionalTime * ( 3.0f - 2.0f * fProportionalTime);
	}
	if ( pxSeq->GetEaseType() == ASURA_GUIMENU_ANIMATOR_EASE_OUT_IN )
	{
		// Apply ease out/in curve. Invert the in/out curve.
		// start fast, decelerate then accelerate to finish fast.
		float fEaseCurve = fProportionalTime * fProportionalTime * ( 3.0f - 2.0f * fProportionalTime);
		fProportionalTime = fEaseCurve + ( fProportionalTime - fEaseCurve) * 2.0f;
	}
	else if ( pxSeq->GetEaseType() == ASURA_GUIMENU_ANIMATOR_EASE_IN )
	{
		// Start slow and accelerate
		// y = x*x
		fProportionalTime = fProportionalTime * fProportionalTime;
	}
	else if ( pxSeq->GetEaseType() == ASURA_GUIMENU_ANIMATOR_EASE_OUT )
	{
		// Start fast and decelerate
		float fPropTimeSq = fProportionalTime * fProportionalTime;
		fProportionalTime = fPropTimeSq + (fProportionalTime - fPropTimeSq) * 2.0f;
	}

	//get the current frame
	u_int uFrame;
	for(uFrame = pxSeq->GetNumberOfFrames() - 1; uFrame; uFrame--)
	{
		if(pxSeq->GetFrameTime(uFrame) <= fProportionalTime) 
		{
			break;
		}
	}

	if(pxSeq->GetInterpolate())
	{
		//interpolate to next frame
		if (uFrame != (pxSeq->GetNumberOfFrames() - 1))
		{
			float fFrameTime = pxSeq->GetFrameTime(uFrame);
			float fNextFrameTime = pxSeq->GetFrameTime(uFrame + 1);
			float fFrameDuration = fNextFrameTime - fFrameTime;

			Asura_Assert(fFrameDuration >= 0, "invalid frame duration");

			//get the proportion through the frame
			float fLerp = (fProportionalTime - fFrameTime) / fFrameDuration;
			
			// fill out the return data
			fLerpOut = Asura_Maths::ClampToRange(fLerp, 0.0f, 1.0f);
			uFrameOut = uFrame;
			uNextFrameOut = uFrame + 1;
		}
		else
		{
			//On last frame. Not going to interpolate back to first frame by default , since
			//this may well end up with interpolating the wrong way for scrolly effects.
			//(And it can always be achieved if desired by making the last frame the same as the first)
			fLerpOut = 0.0f;
			uFrameOut = uFrame;
			uNextFrameOut = uFrame;
		}
	}
	else
	{
		//just take the frame as is
		fLerpOut = 0.0f;
		uFrameOut = uFrame;
		uNextFrameOut = uFrame;
	}

	return true;
}

// chunk load/save
bool Asura_GUIMenu_Animator::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;

	if (uVersion > uASURA_GUIMENU_ANIMATOR_CHUNKVERSION)
	{
		Asura_Assertf(false, ("GUIMenu Animator version unknown (%u > %u)", 
			uVersion, uASURA_GUIMENU_ANIMATOR_CHUNKVERSION));
		return false;
	}

	char acName[128];
	xStream.ReadString(acName,sizeof(acName));

#ifdef ASURA_TOOLS
	m_xSequenceName = acName;
#endif

	// find sequence by hash ID
	m_pxSequence = Asura_GUIMenu_Animator_Storage::FindSequence( Asura_GetHashID(acName) );
	Asura_Assertf(m_pxSequence, ("sequence %s for GUIMenu animator not found", acName));

	// read in other data
	if( uVersion >= 2 )
	{
		m_xStartConditions.ReadFromChunkStream( xStream ); // Activation conditions added in chunk version 2
				
		xStream >> m_bShouldStartActive;
	}

	xStream >> m_fSpeed;

	if(uVersion >= 1)
	{
		#ifdef ASURA_TOOLS
		xStream >> m_uToolsFlags;
		#else
		xStream.Skip(sizeof(u_int));
		#endif
	}

	if(uVersion >= 3)
	{
		xStream >> m_fDelay;
		m_fDelayTimer = m_fDelay;
	}

	if(uVersion >= 4)
	{
		xStream >> m_bInitialiseToEnd;
	}

	return true;
}

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Animator::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	if (m_pxSequence)
	{
		Asura_Assert(m_xSequenceName == m_pxSequence->GetNameAsString(), "sequence name doesn't match stored value");
	}

	xStream << uASURA_GUIMENU_ANIMATOR_CHUNKVERSION;

	// write sequence name and speed
	xStream << m_xSequenceName;

	m_xStartConditions.WriteToChunkStream( xStream ); // Restart conditions (added in chunk version 2)
	
	xStream << m_bShouldStartActive; // 'Starts active' bool (added in chunk version 2)

	xStream << m_fSpeed;
	xStream << m_uToolsFlags; //(Not bothering to use the bitfield wrapper for this , since it is only loaded in tools)
	xStream << m_fDelay;
	xStream << m_bInitialiseToEnd;
}

void Asura_GUIMenu_Animator::ExportToXML( Tools_GUIMenu_XMLExporter* pxExporter ) const
{
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer("<Animator name=\"");
	pxExporter->AddToBufferEscaped( GetSequenceName() );
	pxExporter->AddToBuffer("\" >\n");
	pxExporter->IncTabs();

	// Restart conditions (added in chunk version 2)
	m_xStartConditions.ExportToXML( pxExporter ); 

	// 'Starts active' bool (added in chunk version 2)
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<starts_active>" );
	pxExporter->AddToBuffer( m_bShouldStartActive );
	pxExporter->AddToBuffer( "</starts_active>\n" );

	if (m_pxSequence)
	{
		Asura_Assert(m_xSequenceName == m_pxSequence->GetNameAsString(), "sequence name doesn't match stored value");
	}

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<sequence_name>" );
	pxExporter->AddToBufferEscaped( m_xSequenceName.GetString() );
	pxExporter->AddToBuffer( "</sequence_name>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<speed>" );
	pxExporter->AddToBuffer( m_fSpeed );
	pxExporter->AddToBuffer( "</speed>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<tools_flags>" );
	pxExporter->AddToBufferHex( m_uToolsFlags );
	pxExporter->AddToBuffer( "</tools_flags>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<delay>" );
	pxExporter->AddToBuffer( m_fSpeed );
	pxExporter->AddToBuffer( "</delay>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<init_to_end>" );
	pxExporter->AddToBuffer( m_bInitialiseToEnd );
	pxExporter->AddToBuffer( "</init_to_end>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer("</Animator>\n");
}
#endif

Asura_Hash_ID Asura_GUIMenu_Animator::GetSequenceHash() const
{
	return m_pxSequence ? m_pxSequence->GetHashID() : 0;
}


//////////////////////////////////////////////////////////////////////////////
// Asura_GUIMenu_Animator_Storage
//////////////////////////////////////////////////////////////////////////////
void Asura_GUIMenu_Animator_Storage::Initialise()
{
}

void Asura_GUIMenu_Animator_Storage::Reset()
{
	// everything is deleted at the moment
	Shutdown();

#if 0 && defined (ASURA_TOOLS) && defined( _DEBUG)
	// hack - create some default sequences
	Asura_GUIMenu_Animator_Sequence* pxNewSeq = new Asura_GUIMenu_Animator_Sequence("test-animator");
	pxNewSeq->Initialise(ASURA_GUIMENU_MODIFY_POSITION);
	pxNewSeq->AddFrameAfter(0);
	pxNewSeq->AddFrameAfter(1);
	pxNewSeq->AddFrameAfter(2);
	pxNewSeq->AddFrameAfter(3);
	pxNewSeq->SetDuration(2.0f);
	pxNewSeq->SetInterpolate(true);
	pxNewSeq->SetLoop(true);
	pxNewSeq->SetStateFlags(ASURA_GUIMENU_ACTION_FLAG_IS_HILIGHTED);

	pxNewSeq->SetFrameTime(1, 0.25f);
	pxNewSeq->SetFrameTime(2, 0.50f);
	pxNewSeq->SetFrameTime(3, 0.75f);
	pxNewSeq->SetFrameTime(4, 1.00f);

	pxNewSeq->SetFrameData(0, Asura_Vector_2(-10.0f, -10.0f));
	pxNewSeq->SetFrameData(1, Asura_Vector_2(+10.0f, -10.0f));
	pxNewSeq->SetFrameData(2, Asura_Vector_2(+10.0f, +10.0f));
	pxNewSeq->SetFrameData(3, Asura_Vector_2(-10.0f, +10.0f));
	pxNewSeq->SetFrameData(4, Asura_Vector_2(-10.0f, -10.0f));

	pxNewSeq = new Asura_GUIMenu_Animator_Sequence("test-animator2");
	pxNewSeq->Initialise(ASURA_GUIMENU_MODIFY_POSITION);
	pxNewSeq->AddFrameAfter(0);
	pxNewSeq->SetDuration(1.0f);
	pxNewSeq->SetInterpolate(true);
	pxNewSeq->SetLoop(true);
	pxNewSeq->SetLoopBackAndForth(true);
	pxNewSeq->SetStateFlags(ASURA_GUIMENU_ACTION_FLAG_IS_HILIGHTED);

	pxNewSeq->SetFrameTime(1, 1.0f);

	pxNewSeq->SetFrameData(0, Asura_Vector_2(-50.0f, 0.0f));
	pxNewSeq->SetFrameData(1, Asura_Vector_2(+50.0f, 0.0f));

	pxNewSeq = new Asura_GUIMenu_Animator_Sequence("test-animator3");
	pxNewSeq->Initialise(ASURA_GUIMENU_MODIFY_POSITION);
	pxNewSeq->AddFrameAfter(0);
	pxNewSeq->SetDuration(1.0f);
	pxNewSeq->SetInterpolate(true);
	pxNewSeq->SetLoopBackAndForth(true);
	pxNewSeq->SetStateFlags(ASURA_GUIMENU_ACTION_FLAG_IS_HILIGHTED);

	pxNewSeq->SetFrameTime(1, 1.0f);

	pxNewSeq->SetFrameData(0, Asura_Vector_2(-50.0f, 0.0f));
	pxNewSeq->SetFrameData(1, Asura_Vector_2(+50.0f, 0.0f));

#endif // test others
#if 0 && defined (ASURA_TOOLS) && defined( _DEBUG)
	Asura_GUIMenu_Animator_Sequence* pxNewSeq = new Asura_GUIMenu_Animator_Sequence("test-animator-flyon");
	pxNewSeq->Initialise(ASURA_GUIMENU_MODIFY_POSITION);
	pxNewSeq->AddFrameAfter(0);
	pxNewSeq->SetDuration(1.0f);
	pxNewSeq->SetInterpolate(true);
	pxNewSeq->SetLoop(false);
	pxNewSeq->SetLoopBackAndForth(false);
	pxNewSeq->SetStateFlags(ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_IN);

	pxNewSeq->SetFrameTime(1, 1.0f);

	pxNewSeq->SetFrameData(0, Asura_Vector_2(0.0f, -272.0f));
	pxNewSeq->SetFrameData(1, Asura_Vector_2(0.0f, 0.0f));

	pxNewSeq = new Asura_GUIMenu_Animator_Sequence("test-animator-flyoff");
	pxNewSeq->Initialise(ASURA_GUIMENU_MODIFY_POSITION);
	pxNewSeq->AddFrameAfter(0);
	pxNewSeq->SetDuration(1.0f);
	pxNewSeq->SetInterpolate(true);
	pxNewSeq->SetLoop(false);
	pxNewSeq->SetLoopBackAndForth(false);
	pxNewSeq->SetStateFlags(ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT);

	pxNewSeq->SetFrameTime(1, 1.0f);

	pxNewSeq->SetFrameData(0, Asura_Vector_2(0.0f, 0.0f));
	pxNewSeq->SetFrameData(1, Asura_Vector_2(0.0f, 272.0f));
#endif // test fly on
}

void Asura_GUIMenu_Animator_Storage::Shutdown()
{
	g_bShuttingDown = true;
	g_xSequenceTree.Traverse(DeleteSequence);
	g_xSequenceTree.Reset();
	g_bShuttingDown = false;
}


const Asura_GUIMenu_Animator_Sequence* Asura_GUIMenu_Animator_Storage::FindSequence( Asura_Hash_ID uHashID )
{
	Asura_GUIMenu_Animator_Sequence** pxResult = g_xSequenceTree.GetDataByHash(uHashID);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}

#ifdef ASURA_TOOLS
//Tools accessors for the sequences
//Note that new sequences aren't added to the end of the 'indexed' list , so
//it isn't safe to store an index value for any length of time.
u_int Asura_GUIMenu_Animator_Storage::GetNumberOfSequences()
{
	return g_xSequenceTree.GetNumberOfNodes();
}

Asura_GUIMenu_Animator_Sequence* Asura_GUIMenu_Animator_Storage::GetSequenceByHash(Asura_Hash_ID uHashID)
{
	Asura_GUIMenu_Animator_Sequence** pxResult = g_xSequenceTree.GetDataByHash(uHashID);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}

Asura_GUIMenu_Animator_Sequence* Asura_GUIMenu_Animator_Storage::GetSequenceByIndex(u_int uIndex)
{
	Asura_GUIMenu_Animator_Sequence** pxResult = g_xSequenceTree.GetDataByIndex(uIndex);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}
bool Asura_GUIMenu_Animator_Storage::RenameSequence( const char* szOldName, const char* szNewName )
{
	Asura_Hash_ID uOldHash = Asura_GetHashID(szOldName);
	Asura_Hash_ID uNewHash = Asura_GetHashID(szNewName);

	//make sure the sequence exists
	Asura_GUIMenu_Animator_Sequence* pxSequence = GetSequenceByHash(uOldHash);
	if(!pxSequence) return false;

	//see if the hash replacement is okay
	if(uOldHash != uNewHash)
	{
		if(!g_xSequenceTree.ReplaceHash(uOldHash,uNewHash)) return false;
	}

	pxSequence->Rename(szNewName);
	return true;
}
#endif

//tree traversal callbacks
void Asura_GUIMenu_Animator_Storage::DeleteSequence( Asura_Hash_ID uHash, Asura_GUIMenu_Animator_Sequence** ppxSeq )
{
	delete *ppxSeq;
}

// called by contructors/destructors
void Asura_GUIMenu_Animator_Storage::AddSequenceToTree( Asura_GUIMenu_Animator_Sequence& xSequence )
{
	Asura_Assert(!FindSequence(xSequence.GetHashID()), "Sequence already exists in tree");
	g_xSequenceTree.AddNode(xSequence.GetHashID(), &xSequence);
}

void Asura_GUIMenu_Animator_Storage::RemoveSequenceFromTree( const Asura_GUIMenu_Animator_Sequence& xSequence )
{
	//Don't bother removing entries from the tree while shutting down , since the tree will be fully reset afterwards.
	if(g_bShuttingDown) return;
#ifdef ASURA_ASSERT
	bool bOK = 
#endif
	g_xSequenceTree.DeleteNodeByHash( xSequence.GetHashID() );
	Asura_Assert(bOK, "failed to delete node from hash tree");
}

#endif // #ifdef ASURA_USE_GUI_MENU_SYSTEM
