//////////////////////////////////////////////////////////////////////////////
//
// Asura_GUIMenu_Animator.h
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     classes to manage animation of GUIMenu elements and widgets, e.g.
//     changing position, height, colour etc
//     Some code is based on Asura_TextureAnimation, but has enough 
//     differences to make sharing the same code impractical.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_GUIMENU_ANIMATOR_H___
#define ASURA_GUIMENU_ANIMATOR_H___ 1

#ifdef ASURA_USE_GUI_MENU_SYSTEM
// I N C L U D E S ///////////////////////////////////////////////////////////
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "GUIMenu/Asura_GUIMenu_Modifier.h"	// for ModifyType enum

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_GUIMenu_Animator_Sequence;
class Tools_GUIMenu_XMLExporter;

// E N U M S /////////////////////////////////////////////////////////////////
enum ASURA_GUIMENU_ANIMATOR_DATA_TYPE
{
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INVALID	= 0,
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE_FLOAT		= 1,
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE_INT		= 2,
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE_VECTOR2	= 3,
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE_UVSET		= 4,
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE_RGBA		= 5,
};

// Update g_apcAnimatorsEaseNames for Tools when changing this enum
enum ASURA_GUIMENU_ANIMATOR_EASE_TYPE
{
	ASURA_GUIMENU_ANIMATOR_EASE_NONE	= 0,
	ASURA_GUIMENU_ANIMATOR_EASE_IN_OUT	= 1,
	ASURA_GUIMENU_ANIMATOR_EASE_OUT_IN	= 2,
	ASURA_GUIMENU_ANIMATOR_EASE_IN		= 3,
	ASURA_GUIMENU_ANIMATOR_EASE_OUT		= 4,

	ASURA_GUIMENU_ANIMATOR_MAX_EASE_TYPES, // Must be last in list	
};

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uASURA_GUIMENU_ANIMATOR_CHUNKVERSION = 4;
// Version history:
// 0 - initial version (MH - 03.04.2007) 
// 1 - Added tools flags for marking style instances (RAR - 16.04.2007)
// 2 - Added restart condition vars and starts-active bool (NDB - 28.01.08)
// 3 - Added delay time (RAR - 27.03.08)
// 4 - Added m_bInitialiseToEnd flag. Start the animation at the end point when initialising the menu page

static const u_int uASURA_GUIMENU_ANIMATOR_SEQUENCE_CHUNKVERSION = 4;
// Version history:
// 0 - initial version (MH - 03.04.2007) 
// 1 - Add ability for animators to play to end, regardless of state changes (PJC 19.05.08)
// 2 - Add Console Var and global timer. Use console var for animation progression.
//		Global timer use to keep different animations in sync.
// 3 - Add Smoothing console var settings
// 4 - Add ease in/out combination flags  19/3/09
// C L A S S E S /////////////////////////////////////////////////////////////


// Asura_GUIMenu_Animator is a sequence instance.
class Asura_GUIMenu_Animator
{
public:

	Asura_GUIMenu_Animator();
	Asura_GUIMenu_Animator(const Asura_GUIMenu_Animator& xOther);
	Asura_GUIMenu_Animator& operator=(const Asura_GUIMenu_Animator& xOther);

	// Applys ShouldStartActive condition
	void Initialise(bool bInitMenuPage = false);

	//returns true if the animator should be used in the current state.
	//while in tools this will also return for the animator sequence that is currently being edited.
	bool ShouldUseAnimator(u_int uStateFlags) const;

	inline bool IsStartConditionMet() const { return m_xStartConditions.IsConditionMet(); }
	
	// Calls Start() if IsConditionMet() is true and was false on the last call to this method
	void UpdateStartCondition();

	bool ShouldStartActive() { return m_bShouldStartActive; }

	u_int GetStateFlags() const;		// returns widget state this animator applies to
	u_int GetModifyType() const;		// returns what is being changed (e.g. position i.e. ASURA_GUIMENU_MODIFY_POSITION)
	bool IsLooping() const;
	float GetDuration() const;
	float GetSpeed() const {return m_fSpeed;}
	float GetDelay() const {return m_fDelay;}

	// GetXXXData() functions also update the timer
	Asura_Vector_2 GetVector2Data();
	float GetFloatData();
	int GetIntData();
	Asura_GUIMenu_UV_Set GetUVSetData();
	Asura_RGBA GetRGBAData();

	void Start();
	void Restart()	{ Start(); }
	void Finish();	// move timer to the end of the animation loop. Does not stop a looping animation 

	Asura_Hash_ID GetSequenceHash() const;

#ifdef ASURA_TOOLS
	// tool functions to set up animator
	void SetSequence( const char* szName );
	void SetSpeed( float fSpeed ) { m_fSpeed = fSpeed; }
	void SetDelay( float fDelay ) { m_fDelay = fDelay; }

	const char* GetSequenceName() const {return m_xSequenceName.GetString();}

	bool IsGeneratedByStyle() const {return m_bToolsFlag_GeneratedByStyle;}
	void SetGeneratedByStyle(bool bSet) {m_bToolsFlag_GeneratedByStyle = bSet;}

	// Tools accessors for restart conditions
	void SetShouldStartActive( const bool bShouldStartActive ) { m_bShouldStartActive = bShouldStartActive; }

	void SetStartConditions( Asura_GUIMenu_ConditionVarManager* pxConditions ) { m_xStartConditions.Copy(pxConditions); }
	Asura_GUIMenu_ConditionVarManager*  GetStartConditions() { return &m_xStartConditions; }

	void SetInitialiseToEnd( const bool bInitialiseToEnd ) { m_bInitialiseToEnd = bInitialiseToEnd; }
	bool GetInitialiseToEnd() { return m_bInitialiseToEnd; }

	const Asura_GUIMenu_Animator_Sequence* GetSequence() { return m_pxSequence; }
#endif

	// chunk load/save
	bool ReadFromChunkStream( Asura_Chunk_Stream& xStream );
#ifdef ASURA_TOOLS
	void WriteToChunkStream( Asura_Chunk_Stream& xStream ) const;
	void ExportToXML( Tools_GUIMenu_XMLExporter* pxExporter ) const;
#endif

private:
	bool IsPlayingToEnd( void ) const;
	bool Update(u_int& uFrameOut, u_int& uNextFrameOut, float& fLerpOut);

	Asura_GUIMenu_ConditionVarManager m_xStartConditions; // If specified, Restart() is called each time this change to true

#ifdef ASURA_TOOLS
	Asura_String m_xSequenceName;
#endif
	const Asura_GUIMenu_Animator_Sequence* m_pxSequence;
	float m_fTimer;
	float m_fSpeed;
	u_int m_uFrameStamp;
	float m_fDelay;//delay before animation should start
	float m_fDelayTimer;//time left on delay
	float m_fCurrentConsoleVar;

#ifdef ASURA_TOOLS
	union
	{
		u_int m_uToolsFlags;
		struct
		{
			u_int m_bToolsFlag_GeneratedByStyle : 1;//Marks animators that have been generated as a result of being part of a style instance
		};
	};
#endif	


	bool m_bShouldStartActive;       // if true, animation timer will be initialised to zero
	bool m_bLastKnownIsStartConditionMet;
	bool m_bInitialiseToEnd;

public:
};

// Asura_GUIMenu_Animator_Sequence is collection of frames, duration, and flags (interpolate, loop).
class Asura_GUIMenu_Animator_Sequence
{
	friend class Tools_GUIMenu_AnimEdit;
public:

	Asura_GUIMenu_Animator_Sequence( const char* szName );
	~Asura_GUIMenu_Animator_Sequence();

	ASURA_GUIMENU_ANIMATOR_DATA_TYPE GetDataType() const { return m_eDataType; }
	float GetDuration() const { return m_fDuration; }

	float GetFrameTime( u_int u ) const;
	void* GetFrameData( u_int u ) const;
	u_int GetNumberOfFrames() const { return m_uNumberOfFrames; }
	u_int GetStateFlags() const { return m_uStateFlags; }
	u_int GetModifyType() const { return m_uModifyType; }

	bool GetLoop() const { return m_bLoop; }
	bool GetInterpolate() const { return m_bInterpolate; }
	bool GetLoopBackAndForth() const { return m_bLoopBackAndForth; }
	bool GetPlayToEnd() const { return m_bPlayToEnd; }

	// Get and set the global timer flag.
	bool GetUseGlobalTime() const { return m_bUseGlobalTime; }
	void SetUseGlobalTime( bool bUseGlobalTime ) { m_bUseGlobalTime = bUseGlobalTime; }

	// Get the console var instance.
	const Asura_GUIMenu_ConsoleVar& GetConsoleVar() const { return m_xConsoleVar; }

	// Flag used to decide if the console var value needs to be smoothed.
	bool GetSmoothConsoleVar() const { return m_bSmoothConsoleVar; }
	void SetSmoothConsoleVar( bool bSmoothConsoleVar ) { m_bSmoothConsoleVar = bSmoothConsoleVar; }

	// Time over which the console var is smoothed.
	// Time represents the length of time, in seconds, it takes for the var to go between it min and max values.
	float GetSmoothFrameTime() const { return m_fSmoothFrameTime; }
	void SetSmoothFrameTime( float fSmoothFrameTime ) { m_fSmoothFrameTime = fSmoothFrameTime; }

	// Set animation ease type
	ASURA_GUIMENU_ANIMATOR_EASE_TYPE GetEaseType() const { return m_eEaseType; }
	void SetEaseType( ASURA_GUIMENU_ANIMATOR_EASE_TYPE eEaseType ) { m_eEaseType = eEaseType; }

#ifdef ASURA_TOOLS
	//Initialise a new sequence with a default frame
	void Initialise( ASURA_GUIMENU_MODIFY_TYPE eModifyType );
	void AddFrameAfter(int iPrevFrame);
	void DeleteFrame(u_int uFrame);
	void SetFrameTime( u_int uFrame, float fTime );
	void SetFrameData( u_int uFrame, const Asura_Vector_2& xVec );
	void SetFrameData( u_int uFrame, float fValue );
	void SetFrameData( u_int uFrame, int iValue );
	void SetFrameData( u_int uFrame, const Asura_GUIMenu_UV_Set& xUV );
	void SetFrameData( u_int uFrame, const Asura_RGBA& xColour );
	void SetLoop( bool bLoop ) { m_bLoop = bLoop; }
	void SetInterpolate( bool bInterpolate ) { m_bInterpolate = bInterpolate; }
	void SetLoopBackAndForth( bool bLoopBackAndForth ) { m_bLoopBackAndForth = bLoopBackAndForth; }
	void SetPlayToEnd( bool bPlayToEnd ) { m_bPlayToEnd = bPlayToEnd; }
	void SetDuration( float fDuration ) { m_fDuration = fDuration; }
	void SetStateFlags( u_int uStateFlags ) { m_uStateFlags = uStateFlags; }
	const Asura_String& GetNameAsString() const { return m_xName; }
	const char* GetName() const { return m_xName.GetString(); }
	void Rename( const char* szNewName );
	void Copy(const Asura_GUIMenu_Animator_Sequence* pxSeq);
#endif
	Asura_Hash_ID GetHashID() const { return m_uHashID; }

	// chunk load/save
	bool ReadFromChunkStream( Asura_Chunk_Stream& xStream );
	static bool SkipChunkData( Asura_Chunk_Stream& xStream );
#ifdef ASURA_TOOLS
	void WriteToChunkStream( Asura_Chunk_Stream& xStream ) const;
#endif

	// Update the sequence system.
	static void Update();

	// Get the value of the global timer.
	static float GetGlobalTimer() { return s_fGlobalAnimationTimer; }

private:

	// non-copiable (for now)
	Asura_GUIMenu_Animator_Sequence& operator=(const Asura_GUIMenu_Animator_Sequence&);
	Asura_GUIMenu_Animator_Sequence(const Asura_GUIMenu_Animator_Sequence&);

	// static functions to deal with variable frame size
	static u_int GetFrameDataSize( ASURA_GUIMENU_ANIMATOR_DATA_TYPE eDataType );
	static void* AllocateFrameData( u_int uNumFrames, ASURA_GUIMENU_ANIMATOR_DATA_TYPE eDataType );
	static void CopyFrameData( void* pDestArray, u_int uDestFrame, const void* pSrcArray, 
								u_int uSrcFrame, ASURA_GUIMENU_ANIMATOR_DATA_TYPE eDataType );

	// data members
	u_int								m_uStateFlags;		// bit flags from ASURA_GUIMENU_ACTION_FLAG_IS_DISABLED etc.
	u_int								m_uModifyType;		// what is being changed (e.g. position i.e. ASURA_GUIMENU_MODIFY_POSITION)
	ASURA_GUIMENU_ANIMATOR_DATA_TYPE	m_eDataType;		// indicates what type the frames should be cast to.
	u_int								m_uNumberOfFrames;
	float*								m_pfFrameTimes;
	void*								m_pFrameData;
	float								m_fDuration;	// in seconds
	
	Asura_GUIMenu_ConsoleVar			m_xConsoleVar;
	bool								m_bUseGlobalTime;

	bool								m_bSmoothConsoleVar; // Smooth var flag
	float								m_fSmoothFrameTime; // Smoothing time in seconds.

	ASURA_GUIMENU_ANIMATOR_EASE_TYPE	m_eEaseType;

	union
	{
		u_int m_uFlags;
		struct
		{
			u_int m_bLoop : 1;
			u_int m_bInterpolate : 1;//interpolates UV offsets between frames
			u_int m_bLoopBackAndForth : 1; //animation reverses upon reaching the end , instead of returning to the start
			u_int m_bPlayToEnd : 1; //animation should finish playing even when it's no longer in a valid state
		};
	};

#ifdef ASURA_TOOLS
	Asura_String m_xName;
#endif
	Asura_Hash_ID m_uHashID;

	static float s_fGlobalAnimationTimer; // Time in seconds
};

// class to deal with managing all the loaded sequences.
class Asura_GUIMenu_Animator_Storage
{
	friend class Asura_GUIMenu_Animator_Sequence;

public:

	static void Initialise();
	static void Reset();
	static void Shutdown();

	static const Asura_GUIMenu_Animator_Sequence* FindSequence( Asura_Hash_ID uHashID );

#ifdef ASURA_TOOLS
	//Tools accessors for the sequences
	//Note that new sequences aren't added to the end of the 'indexed' list , so
	//it isn't safe to store an index value for any length of time.
	static u_int GetNumberOfSequences();
	static Asura_GUIMenu_Animator_Sequence* GetSequenceByHash(Asura_Hash_ID uHashID);
	static Asura_GUIMenu_Animator_Sequence* GetSequenceByIndex(u_int uIndex);
	static bool RenameSequence( const char* szOldName, const char* szNewName );
#endif

private:

	//tree traversal callbacks
	static void DeleteSequence( Asura_Hash_ID uHash, Asura_GUIMenu_Animator_Sequence** ppxSeq );

	// called by contructors/destructors
	static void AddSequenceToTree( Asura_GUIMenu_Animator_Sequence& xSeq );
	static void RemoveSequenceFromTree( const Asura_GUIMenu_Animator_Sequence& xSeq );

};

#endif // ASURA_USE_GUI_MENU_SYSTEM
// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ASURA_GUIMENU_ANIMATOR_H___
