// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Element_Base.cpp												//
//																						//
//	Description: base class for all elements											//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "Asura_GUIMenu_Element_Base.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
#include "Asura_GUIMenu_System.h"
#include "Asura_GUIMenu_Menu_Page.h"
#include "Asura_GUIMenu_Widget_Base.h"

#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Element_Base::Asura_GUIMenu_Element_Base()
: m_uStateOverride(0)
{
	// by default, make elements appear in all states.
	m_uFlags = ASURA_GUIMENU_ACTION_FLAG_ALL;
}

// ---------------------------------------------------------------------------------------
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Element_Base::~Asura_GUIMenu_Element_Base()
{}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Base::Initialise()
{
	PARENT::Initialise();
	m_uStateOverride = 0;
}

void Asura_GUIMenu_Element_Base::Update( float fElapsedTime)	//, unsigned int uButtons, POINT ptMousePos)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_BASE);

	PARENT::Update( fElapsedTime);

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Base::Render( u_int uState)
{
	if (IsRenderingDisabled())
	{
		return;
	}

	Render( uState | m_uStateOverride, GetScaledWorldBounds());
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Element_Base::Whisper( const Asura_Message_GUIMenu_Notify* pxNotify)
{
	bool bProcessed = false;

	const u_int uContainerCount = GetContainerCount();
	for( u_int u=0; u<uContainerCount; u++)
	{
		bProcessed = m_ppxContainerParts[u]->Whisper( pxNotify);

		if( bProcessed == true)
		{	break;	}
	}

	return bProcessed;	
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_Base::UpdateLocalDataFromStyle()
{
	if( m_pxStyleObject != NULL)
	{
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_BASE_LOCAL_FLAG_WIDTH) == 0)
		{
			Asura_GUIMenu_Element_Base* pxStyleBase = static_cast<Asura_GUIMenu_Element_Base*>(m_pxStyleObject);

			m_xUnscaledLocalBounds.MaxX = pxStyleBase->GetUnscaledLocalBounds().GetWidth() + m_xUnscaledLocalBounds.MinX;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_BASE_LOCAL_FLAG_HEIGHT) == 0)
		{
			Asura_GUIMenu_Element_Base* pxStyleBase = static_cast<Asura_GUIMenu_Element_Base*>(m_pxStyleObject);

			m_xUnscaledLocalBounds.MaxY = pxStyleBase->GetUnscaledLocalBounds().GetHeight() + m_xUnscaledLocalBounds.MinY;
		}
	}

	PARENT::UpdateLocalDataFromStyle();
}
#endif

// --------------------------------------------------------------------------------------- 


// --------------------------------------------------------------------------------------- 
// -- Protected Functions ---------------------------------------------------------------- 


// --------------------------------------------------------------------------------------- 
// -- Overide tool helper function ------------------------------------------------------------

void Asura_GUIMenu_Element_Base::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	u_int uLocalChunkVersion;

	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	uLocalChunkVersion = ReadVersionNumber( xStream);

	if( CheckVersionNumber( uLocalChunkVersion, ASURA_GUIMENU_CHUNK_TYPE_ELEMENT_BASE,xStream.GetFileHandle() ) == true)
	{
		if (uLocalChunkVersion < 4)
		{
			Asura_Vector_2 xDims;

			xStream >> xDims.x;
			xStream >> xDims.y;

			m_xUnscaledLocalBounds.MaxX = xDims.x + m_xUnscaledLocalBounds.MinX;
			m_xUnscaledLocalBounds.MaxY = xDims.y + m_xUnscaledLocalBounds.MinY;
		}

		if( uLocalChunkVersion >= 1)
		{
			if( uLocalChunkVersion < 3)
			{
				u_int uScalingFlags(0);
				xStream >> uScalingFlags;

				if (uLocalChunkVersion < 2)
				{
					if (uScalingFlags & 1)
					{
						uScalingFlags = ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX | ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY;
					}
				}

				SetScalingFlags( uScalingFlags );
			
				// Fix for old "scale to menu page" behaviour. If this element includes any 
				// scaling flags an its parent is not the menu page then force its parent to 
				// scale with the menu page.
				if( uScalingFlags && (pxParentObject->GetType() != Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_MENUPAGE) )
				{
					pxParentObject->SetScalingFlags( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX | ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY );
				}

				float fRedundentValue(0.0f);
				xStream >> fRedundentValue; // was m_xSizeAsPercentage.width;
				xStream >> fRedundentValue; // was m_xSizeAsPercentage.height;
			}
		}

	}
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_Base::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	// -- Write out parent data - Asura_GUIMenu_Root --
	PARENT::WriteToChunkStream( xStream);

	WriteVersionNumber( ASURA_CHUNK_ELEMENT_BASE_VERSION, xStream);
}


void Asura_GUIMenu_Element_Base::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<ElementBase_data>\n" );
	pxExporter->IncTabs();

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</ElementBase_data>\n" );

	
}
#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------ 


#endif // ASURA_USE_GUI_MENU_SYSTEM
