// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Element_StatusText.cpp											//
//																						//
//	Description: status text element class 												//
//				 - the class stores a list of text id's matched to a specific int value	//
//				 - a command console var will be used to match against an int value		//
//																						//
//	Created: Steve Keen - 26/04/2007													//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
#include "Asura_GUIMenu_Element_StatusText.h"

#include <wchar.h>

#include "Core/Asura_Hash.h"
#include "Core/Asura_HashedLocalisedText.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"

#include "GUIMenu/Asura_GUIMenu_Animator.h"
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"
#include "GUIMenu/Asura_GUIMenu_Modifier.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"

#include "Rendering/Asura_Font_System.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#include "GUIMenu/Tools_Mode_GUIMenu.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Element_StatusText::Asura_GUIMenu_Element_StatusText(void)
: m_iCurrentConsoleVarValue( 0)
, m_iFontID( -1)
, m_uFontNameHash( ASURA_HASH_ID_UNSET)
, m_fOffsetX( 0.0f)
, m_fOffsetY( 0.0f)
, m_fFontScale( 1.0f)
, m_uTextRenderFlags( ASURA_GUIMENU_STATUSTEXT_JUSTIFY_DEFAULT)
, m_uTextInfoEntryCount( 0)
, m_uMaxTextInfoEntryCount( 0)
, m_pxTextInfoEntries( NULL)
, m_uActiveTextPage(ASURA_HASH_ID_UNSET)
, m_uActiveTextHash(ASURA_HASH_ID_UNSET)
, m_xActiveTextColour( 0x00000000, 0 )
, m_pxSpecialEffectsSettings(NULL)
{
	m_uType = ASURA_GUIMENU_TYPEID_ELEMENT_STATUSTEXT;
#ifdef ASURA_TOOLS
	AllocateStatusTextArraySpace( ASURA_GUIMENU_STATUSTEXT_ADDITION_GRADIENT);
#endif

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_STATUSTEXT);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Element_StatusText::~Asura_GUIMenu_Element_StatusText(void)
{
	if( m_pxTextInfoEntries != NULL)
	{
		delete []m_pxTextInfoEntries;
		m_pxTextInfoEntries = NULL;

		m_uTextInfoEntryCount = 0;
		m_uMaxTextInfoEntryCount = 0;
	}

	if ( m_pxSpecialEffectsSettings != NULL )
	{
		delete m_pxSpecialEffectsSettings;
	}

	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_STATUSTEXT);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::SetEnableSpecialEffects(bool bEnable)
{
	if ( bEnable == true && m_pxSpecialEffectsSettings == NULL )
	{
		m_pxSpecialEffectsSettings = new Asura_Font_Special_Effects;
	}
	else if ( bEnable == false && m_pxSpecialEffectsSettings != NULL )
	{
		delete m_pxSpecialEffectsSettings;
		m_pxSpecialEffectsSettings = NULL;
	}
}

void Asura_GUIMenu_Element_StatusText::Render( u_int uState)
{
	PARENT::Render( uState);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::Render( u_int uState, const Asura_Bounding_Rectangle& xBounds)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_STATUSTEXT);

	Asura_Bounding_Rectangle xRect(xBounds);
	const Asura_UnicodeChar* pwszText = NULL;
	float fFontScale = GetFontScale();
	if (GetTextRenderFlag(ASURA_GUIMENU_STATUSTEXT_SCALE_ON_PARENT_RESIZE))
	{
		if (m_xUnscaledLocalBounds.GetWidth() && m_xUnscaledLocalBounds.GetHeight())
		{
			//scale by smaller of parent dimension scales
			float fParentScale = m_xScaledWorldBounds.GetWidth() / m_xUnscaledLocalBounds.GetWidth();
			fParentScale = Asura_Maths::Min(fParentScale, m_xScaledWorldBounds.GetHeight() / m_xUnscaledLocalBounds.GetHeight());
			fFontScale *= fParentScale;
		}
	}
	float fOffsetX = GetOffsetX();
	float fOffsetY = GetOffsetY();
	u_int uRenderFlags = GetTextRenderFlags();
	int iFontID;

	if( m_iFontID == -1)
	{
		if( m_uFontNameHash != ASURA_HASH_ID_UNSET)
		{
			//			m_iFontID = Asura_Font_System::GetFontHandle( GetFontName() );
			m_iFontID = Asura_Font_System::GetFontHandle( m_uFontNameHash);
		}

		//if( GetFontName() )
		//{
		//	m_iFontID = Asura_Font_System::GetFontHandle( GetFontName() );
		//}
	}

	Asura_Assert( m_iFontID != -1, "Asura_GUIMenu_Element_StatusText::Render - Font ID == -1");


	xRect.MinX = (float)Asura_Maths::FloatToInt_Chop(xRect.MinX);
	xRect.MinY = (float)Asura_Maths::FloatToInt_Chop(xRect.MinY);

	u_int uCol = m_xActiveTextColour.ToUIntARGB();

	ASURA_TRANSLUCENCY_MODE eTransMode = GetTranslucencyMode();

	if( m_uActiveTextPage != ASURA_HASH_ID_UNSET && m_uActiveTextHash != ASURA_HASH_ID_UNSET)
	{
		pwszText = Asura_HashedLocalisedText::GetString(m_uActiveTextPage, m_uActiveTextHash);
	}

	iFontID = m_iFontID;

	ApplyModifierChanges( uState, xRect, fOffsetX, fOffsetY, fFontScale, uCol, eTransMode, uRenderFlags, &pwszText, iFontID);

#ifdef ASURA_TOOLS
	if(Tools_Mode_GUIMenu::ShouldUseOffsetAndZoom())
	{
		//Apply offset to coords from menu mode
		xRect.MinX = (xRect.MinX + Tools_Mode_GUIMenu::GetOffsetX())*Tools_Mode_GUIMenu::GetZoom();
		xRect.MaxX = (xRect.MaxX + Tools_Mode_GUIMenu::GetOffsetX())*Tools_Mode_GUIMenu::GetZoom();
		xRect.MinY = (xRect.MinY + Tools_Mode_GUIMenu::GetOffsetY())*Tools_Mode_GUIMenu::GetZoom();
		xRect.MaxY = (xRect.MaxY + Tools_Mode_GUIMenu::GetOffsetY())*Tools_Mode_GUIMenu::GetZoom();

		fOffsetX *= Tools_Mode_GUIMenu::GetZoom(); 
		fOffsetY *= Tools_Mode_GUIMenu::GetZoom(); 

		fFontScale *= Tools_Mode_GUIMenu::GetZoom();
	}
#endif


	ASURA_TRANSLUCENCY_MODE eOldMode = Asura_Font_System::SetTranslucencyMode( eTransMode);

	Asura_Font_System::SetFont( iFontID);

	if(pwszText && fFontScale)
	{
		if (fOffsetX != 0.0f)
		{
			// no wordwrap or offset X doesn't do anything.
			uRenderFlags |= ASURA_TEXTFORMAT_RELATIVETORECT;
		}
		else
		{
			uRenderFlags |= ASURA_TEXTFORMAT_RELATIVETORECT | ASURA_TEXTFORMAT_WORDWRAP;
		}

		if ( m_pxSpecialEffectsSettings != NULL )
		{
			if ( m_pxSpecialEffectsSettings->GetLastFrameNum()+1 < Asura_Timers::GetFrameNumber() )
			{
				m_pxSpecialEffectsSettings->Start();
			}

			m_pxSpecialEffectsSettings->Update();
		}

		if ( m_pxSpecialEffectsSettings != NULL )
		{
			Asura_Font_System::PrintSpecialEffectsString( 
				pwszText, 
				&xRect, 
				fOffsetX, 
				fOffsetY, 
				1.0f, 
				fFontScale, 
				uCol, 
				uRenderFlags, 
				m_pxSpecialEffectsSettings );
		}
		else
		{
			Asura_Font_System::PrintLongString(pwszText, &xRect, fOffsetX, fOffsetY,
				1.0f, fFontScale, uCol, uRenderFlags);
		}

	}

	Asura_Font_System::SetTranslucencyMode(eOldMode);

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::Update( float fElapsedTime)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_STATUSTEXT);

	PARENT::Update( fElapsedTime);

	const u_int uCount = GetTextEntryCount();

	GetConsoleVar();

	// I've made this so the first entry that meets it's display condition is the one that
	// is displayed.  At the moment, i don't see the point of allowing multiple text strings
	// being drawn over each other.

	// (MH - 23.05.2007) clear the flag so that when the value changes the displayed text is cleared.
	m_uActiveTextPage = m_uActiveTextHash = ASURA_HASH_ID_UNSET;

	for( u_int u = 0; u < uCount; ++u)
	{
		Asura_GUIMenu_StatusText_Info* pxEntry = GetTextEntry( u);
		if( IsStatusConditionMet( pxEntry->m_xDisplayParams.m_eConditionCompare,
								  pxEntry->m_xDisplayParams.m_iDisplayConditionThreshold,
								  m_iCurrentConsoleVarValue))
		{
			// condition has been met, so set it as the active text
			m_uActiveTextPage = pxEntry->m_uTextPage;
			m_uActiveTextHash = pxEntry->m_uTextID;
			m_xActiveTextColour = pxEntry->m_xColour;

			break;
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::Copy( const Asura_GUIMenu_Root* pxSource)
{
	PARENT::Copy( pxSource);

	const Asura_GUIMenu_Element_StatusText* pxSourceText = static_cast<const Asura_GUIMenu_Element_StatusText*> (pxSource);
	
	//copy asura data
#ifdef ASURA_TOOLS
	SetFontByName(pxSourceText->GetFontName());
#else
	m_uFontNameHash = pxSourceText->GetFontNameHash();
#endif

	SetFontScale(pxSourceText->GetFontScale());
	SetTextRenderFlags( pxSourceText->GetTextRenderFlags());

	m_xConsoleVar.Copy(&pxSourceText->m_xConsoleVar);
	
	const u_int uTextEntryCount = pxSourceText->GetTextEntryCount();

	// -- this check and allocation is to make sure that the TextArraySpace is allocated if it hasn't been already.
	if( m_pxTextInfoEntries == NULL)
	{
		AllocateStatusTextArraySpace( uTextEntryCount);
	}

	for( u_int u = 0; u < uTextEntryCount; ++u)
	{
		const Asura_GUIMenu_StatusText_Info* pxTextEntry = pxSourceText->GetTextEntry( u);
		if( pxTextEntry)
		{
			Asura_RGBA xCol = pxTextEntry->m_xColour;
			AddTextInfo( pxTextEntry->m_uTextPage, pxTextEntry->m_uTextID,
						 pxTextEntry->m_xDisplayParams.m_eConditionCompare,
						 pxTextEntry->m_xDisplayParams.m_iDisplayConditionThreshold,
						 xCol);
		}
	}

	SetEnableSpecialEffects(pxSourceText->AreSpecialEffectsEnabled());

	if ( pxSourceText->GetSpecialEffectsSettings() != NULL )
	{
		m_pxSpecialEffectsSettings->Copy(pxSourceText->GetSpecialEffectsSettings());
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Element_StatusText::IsStatusConditionMet( Asura_GUIMenu_Display_Condition eCondition, int iHeldValue, int iValidate)
{
	bool bCondition = true;

	switch( eCondition)
	{
		case ASURA_GUIMENU_CONDITION_EQUAL:
			bCondition = (iValidate == iHeldValue);
			break;

		case ASURA_GUIMENU_CONDITION_NOT_EQUAL:
			bCondition = (iValidate != iHeldValue);
			break;

		case ASURA_GUIMENU_CONDITION_GREATER:
			bCondition = (iValidate > iHeldValue);
			break;

		case ASURA_GUIMENU_CONDITION_GREATER_EQUAL:
			bCondition = (iValidate >= iHeldValue);
			break;

		case ASURA_GUIMENU_CONDITION_LESS:
			bCondition = (iValidate < iHeldValue);
			break;

		case ASURA_GUIMENU_CONDITION_LESS_EQUAL:
			bCondition = (iValidate <= iHeldValue);
			break;

		case ASURA_GUIMENU_CONDITION_NONE:
		default:
			bCondition = true;
			break;
	}


	return bCondition;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::SetFontByName( const char* szFontName )
{
#ifdef ASURA_TOOLS
	m_xFontName = szFontName;
#endif
	m_iFontID = -1;

	m_uFontNameHash = Asura_GetHashID( szFontName);
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
const char* Asura_GUIMenu_Element_StatusText::GetFontName( void) const
{
	return m_xFontName.GetString();
}
#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 

Asura_Hash_ID Asura_GUIMenu_Element_StatusText::GetFontNameHash( void ) const
{
	return m_uFontNameHash;
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Element_StatusText::GetFontScale( void)const
{
	return m_fFontScale;
}

// --------------------------------------------------------------------------------------- 

float* Asura_GUIMenu_Element_StatusText::GetFontScalePointer( void)
{
	return &m_fFontScale;
}

// --------------------------------------------------------------------------------------- 

int Asura_GUIMenu_Element_StatusText::GetFontID( void) const
{
	return m_iFontID;
}

// --------------------------------------------------------------------------------------- 
float Asura_GUIMenu_Element_StatusText::GetOffsetX( void)
{
	return m_fOffsetX;
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Element_StatusText::GetOffsetY( void)	
{
	return m_fOffsetY;
}

// ----------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_StatusText::UpdateLocalDataFromStyle( void)
{
	Asura_GUIMenu_Element_StatusText* pxStyleText;
	Asura_String xString;

	PARENT::UpdateLocalDataFromStyle();

	if( m_pxStyleObject != NULL)
	{
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_STATUSTEXT_LOCAL_FLAG_FONT_SCALE) == 0)
		{
			pxStyleText = static_cast<Asura_GUIMenu_Element_StatusText*>(m_pxStyleObject);
			SetFontScale( pxStyleText->GetFontScale());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_STATUSTEXT_LOCAL_FLAG_FONT_NAME) == 0)
		{
			pxStyleText = static_cast<Asura_GUIMenu_Element_StatusText*>(m_pxStyleObject);
			xString = pxStyleText->GetFontName();
			SetFontByName( xString);
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_STATUSTEXT_LOCAL_FLAG_OFFSET_X) == 0)
		{
			pxStyleText = static_cast<Asura_GUIMenu_Element_StatusText*>(m_pxStyleObject);
			SetOffsetX( pxStyleText->GetOffsetX());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_STATUSTEXT_LOCAL_FLAG_OFFSET_Y) == 0)
		{
			pxStyleText = static_cast<Asura_GUIMenu_Element_StatusText*>(m_pxStyleObject);
			SetOffsetY( pxStyleText->GetOffsetY());
		}

		pxStyleText = static_cast<Asura_GUIMenu_Element_StatusText*>(m_pxStyleObject);

		SetEnableSpecialEffects(pxStyleText->AreSpecialEffectsEnabled());

		if ( GetSpecialEffectsSettings() != NULL )
		{
			GetSpecialEffectsSettings()->Copy(pxStyleText->GetSpecialEffectsSettings());
		}
	}
}
#endif

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Element_StatusText::GetTextRenderFlag( u_int uFlag) const
{
	return ((m_uTextRenderFlags&uFlag) != 0);
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Element_StatusText::GetTextRenderFlags( void) const
{
	return m_uTextRenderFlags;
}


// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::SetTextRenderFlag( u_int uNewFlag, bool bFlagSet)
{
	if( bFlagSet)
	{
		m_uTextRenderFlags |= uNewFlag;
	}
	else
	{
		m_uTextRenderFlags &= (0xffffffff - uNewFlag);
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::SwapTextEntries( u_int uIndex1, u_int uIndex2)
{
	Asura_GUIMenu_StatusText_Info xEntry;

	Asura_Assert( uIndex1 < GetTextEntryCount(), "Asura_GUIMenu_Element_StatusText::SwapTextEntries - Index 1 given is greater than amount contained.");
	Asura_Assert( uIndex2 < GetTextEntryCount(), "Asura_GUIMenu_Element_StatusText::SwapTextEntries - Index 2 given is greater than amount contained.");

	if( (uIndex1 < GetTextEntryCount()) && (uIndex2 < GetTextEntryCount()))
	{
		xEntry = m_pxTextInfoEntries[uIndex1];

		m_pxTextInfoEntries[uIndex1] = m_pxTextInfoEntries[uIndex2];
		m_pxTextInfoEntries[uIndex2] = xEntry;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::AddTextInfo( Asura_Hash_ID uTextPage, Asura_Hash_ID uTextID, Asura_GUIMenu_Display_Condition eCondition, int iDisplayThreshold, Asura_RGBA& xColour)
{
	if( m_uTextInfoEntryCount == m_uMaxTextInfoEntryCount)
	{
		ExpandStatusTextArraySpace();
	}

	m_pxTextInfoEntries[m_uTextInfoEntryCount].m_uTextPage = uTextPage;
	m_pxTextInfoEntries[m_uTextInfoEntryCount].m_uTextID = uTextID;

	m_pxTextInfoEntries[m_uTextInfoEntryCount].m_xDisplayParams.m_eConditionCompare = eCondition;
	m_pxTextInfoEntries[m_uTextInfoEntryCount].m_xDisplayParams.m_iDisplayConditionThreshold = iDisplayThreshold;

	m_pxTextInfoEntries[m_uTextInfoEntryCount].m_xColour = xColour;

	m_uTextInfoEntryCount++;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::RemoveTextInfo( u_int uIndex)
{
	const u_int uCount = GetTextEntryCount();

	Asura_Assert( uIndex < uCount, "Asura_GUIMenu_Element_StatusText::RemoveTextInfo - Index given is greater than amount contained.");

	if( uIndex < uCount)
	{
		for( u_int u = uIndex; u < uCount-1; ++u)
		{
			m_pxTextInfoEntries[u] = m_pxTextInfoEntries[u+1];
		}

		Asura_Mem::Set( &m_pxTextInfoEntries[uCount-1], 0, sizeof( Asura_GUIMenu_StatusText_Info));

		m_uTextInfoEntryCount--;
	}
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_StatusText_Info* Asura_GUIMenu_Element_StatusText::GetTextEntry( u_int uIndex) const
{
	Asura_GUIMenu_StatusText_Info* pxEntry = NULL;

	Asura_Assert( uIndex < GetTextEntryCount(), "Asura_GUIMenu_Element_StatusText::GetTextEntry - Index given is greater than amount contained.");

	if( uIndex < GetTextEntryCount())
	{
		pxEntry = &m_pxTextInfoEntries[uIndex];	
	}

	return pxEntry;
}

// --------------------------------------------------------------------------------------- 
// -- Inheritance tool helper function -------------------------------------------------------

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_StatusText::ClearTextRenderFlags( void)
{
	m_uTextRenderFlags = 0;
}
#endif


// --------------------------------------------------------------------------------------- 
// -- Chunk ------------------------------------------------------------------------------

void Asura_GUIMenu_Element_StatusText::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	char acName[1024];
	u_int uStatusTextCount;
	
	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

#ifdef ASURA_TOOLS
	xStream.ReadString( acName, sizeof(acName));
	Asura_String xFontName = acName;
	SetFontByName( xFontName);
#else
	if( uChunkVersion >= 2)
	{
		xStream >> m_uFontNameHash;
	}
	else
	{
		xStream.ReadString( acName, sizeof(acName));
		m_uFontNameHash = Asura_GetHashID( acName);
	}
#endif


	//xStream.ReadString( acName, sizeof(acName));
	//Asura_String xFontName = acName;
	//SetFontByName( xFontName);

	xStream >> m_fFontScale;

	xStream >> m_uTextRenderFlags;

	m_xConsoleVar.ReadFromChunkStream(xStream);

	xStream >> uStatusTextCount;

#ifndef ASURA_TOOLS
	AllocateStatusTextArraySpace( uStatusTextCount);
#endif

	for( u_int u = 0; u < uStatusTextCount; ++u)
	{
		u_int uDispCompare;
		int iThreshold;
		Asura_RGBA xColour;
		Asura_Hash_ID uTextPage, uTextID;

		xStream >> uDispCompare;
		xStream >> iThreshold;
		if(uChunkVersion >= 1)
		{
			xStream >> uTextPage;
		}
		else
		{
			uTextPage = uASURA_HASHEDTEXTPAGE_MENU;
		}
		xStream >> uTextID;
		xStream >> xColour.m_fR;
		xStream >> xColour.m_fG;
		xStream >> xColour.m_fB;
		xStream >> xColour.m_fA;

		AddTextInfo( uTextPage, uTextID, (Asura_GUIMenu_Display_Condition)uDispCompare, iThreshold, xColour);
	}

	// -- read in the container info
	ReadAllChildObjects( xStream, this);

	if(uChunkVersion >= 3)
	{
		bool bSpecialEffects;

		xStream >> bSpecialEffects;

		SetEnableSpecialEffects(bSpecialEffects);

		if ( bSpecialEffects && m_pxSpecialEffectsSettings != NULL )
		{
			m_pxSpecialEffectsSettings->ReadFromChunkStream(xStream);
		}
	}
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_StatusText::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	u_int u, uDispCompare;

	bool bIsExporting = Tools_Level_Export::IsGUIMenuExport() | Tools_Level_Export::IsLevelExport();

	WriteLocalChunkHeader( xStream, ASURA_CHUNK_ELEMENT_STATUSTEXT_VERSION, GetType());

	PARENT::WriteToChunkStream( xStream);

//	xStream << m_xFontName;
	if( bIsExporting)
	{
		xStream << m_uFontNameHash;
	}
	else
	{
		xStream << m_xFontName;
	}

	xStream << m_fFontScale;
	xStream << m_uTextRenderFlags;

	m_xConsoleVar.WriteToChunkStream(xStream);

	xStream << m_uTextInfoEntryCount;

	for( u=0; u<m_uTextInfoEntryCount; u++)
	{
		uDispCompare = m_pxTextInfoEntries[u].m_xDisplayParams.m_eConditionCompare;
		xStream << uDispCompare;

		xStream << m_pxTextInfoEntries[u].m_xDisplayParams.m_iDisplayConditionThreshold;
		xStream << m_pxTextInfoEntries[u].m_uTextPage;
		xStream << m_pxTextInfoEntries[u].m_uTextID;

		xStream << m_pxTextInfoEntries[u].m_xColour.m_fR;
		xStream << m_pxTextInfoEntries[u].m_xColour.m_fG;
		xStream << m_pxTextInfoEntries[u].m_xColour.m_fB;
		xStream << m_pxTextInfoEntries[u].m_xColour.m_fA;
	}

	// -- write out any children --
	WriteChildObjectData( xStream);

	bool bSpecialEffects = AreSpecialEffectsEnabled();
	xStream << bSpecialEffects;
	if ( bSpecialEffects )
	{
		m_pxSpecialEffectsSettings->WriteToChunkStream(xStream);
	}
}

void Asura_GUIMenu_Element_StatusText::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<StatusText_data>\n" );
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<font_name>" );
	pxExporter->AddToBufferEscaped( m_xFontName.GetString() );
	pxExporter->AddToBuffer( "</font_name>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<font_scale>" );
	pxExporter->AddToBuffer( m_fFontScale );
	pxExporter->AddToBuffer( "</font_scale>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<text_render_flags>" );
	pxExporter->AddToBufferHex( m_uTextRenderFlags );
	pxExporter->AddToBuffer( "</text_render_flags>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</StatusText_data>\n" );
}

#endif // ASURA_TOOLS

// ----------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------


// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------

void Asura_GUIMenu_Element_StatusText::AllocateStatusTextArraySpace( u_int uMaxStatusTextCount)
{
	u_int u;

	Asura_Assert( m_pxTextInfoEntries == NULL, "Asura_GUIMenu_Element_StatusText::AllocateStatusTextArraySpace - container array is already allocated.  Ignoring new request.");

	if( (uMaxStatusTextCount > 0) && (m_pxTextInfoEntries == NULL))
	{
		m_uMaxTextInfoEntryCount = uMaxStatusTextCount;

		m_pxTextInfoEntries = new Asura_GUIMenu_StatusText_Info[m_uMaxTextInfoEntryCount];

		m_uTextInfoEntryCount = 0;

		for( u=0; u<m_uMaxTextInfoEntryCount; u++)
		{
			Asura_Mem::Set( &m_pxTextInfoEntries[u], 0, sizeof( Asura_GUIMenu_StatusText_Info));
		}
	}
}

 //--------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::ExpandStatusTextArraySpace( void)
{
	u_int u, uNewSize;

	if( m_uTextInfoEntryCount == m_uMaxTextInfoEntryCount)
	{
		uNewSize = m_uMaxTextInfoEntryCount + ASURA_GUIMENU_STATUSTEXT_ADDITION_GRADIENT;
		// Time to grow the array
		Asura_GUIMenu_StatusText_Info* pxNewStatusTextInfo = new Asura_GUIMenu_StatusText_Info[uNewSize];
		Asura_Assert( pxNewStatusTextInfo != NULL, "Out of memory!");

		if( pxNewStatusTextInfo != NULL)
		{
			for( u=0; u<uNewSize; u++)
			{
				Asura_Mem::Set( &pxNewStatusTextInfo[u], 0, sizeof( Asura_GUIMenu_StatusText_Info));
			}

			if( m_pxTextInfoEntries != NULL)
			{
				Asura_Mem::Copy( pxNewStatusTextInfo, m_pxTextInfoEntries, m_uMaxTextInfoEntryCount * sizeof( Asura_GUIMenu_StatusText_Info));

				// Get rid of the old array...
				delete []m_pxTextInfoEntries;
			}

			m_uMaxTextInfoEntryCount = uNewSize;

			// ...and use the new one
			m_pxTextInfoEntries = pxNewStatusTextInfo;
		}
	}
}


// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::GetConsoleVar( void)
{
	m_xConsoleVar.GetIntValue( m_iCurrentConsoleVarValue);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::SetConsoleVar( void) const
{
	m_xConsoleVar.SetIntValue( m_iCurrentConsoleVarValue);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_StatusText::ApplyModifierChanges( u_int uStateToMatch, Asura_Bounding_Rectangle& xRect, float& fOffsetX, float& fOffsetY, float& fFontScale, u_int& uColour, ASURA_TRANSLUCENCY_MODE& eTranslucencyMode, u_int& uRenderFlags, const Asura_UnicodeChar** ppwszText, int& iFontID)
{
	u_int u, uModifierCount;
	u_int uModifierState;
	
	Asura_GUIMenu_Modifier* pxMod;
	
	uModifierCount = GetModifierCount();

	for( u=0; u<uModifierCount; u++)
	{
		pxMod = GetModifier( u);
		if( pxMod)
		{
			uModifierState = pxMod->GetStateBeingModified();

			if( ((uModifierState&uStateToMatch) != 0) && pxMod->IsActivationConditionMet() )
			{
				switch (pxMod->GetModifyType())
				{
					case ASURA_GUIMENU_MODIFY_POSITION:
					{
						Asura_Vector_2 xDeltaPos = pxMod->GetVector2Data();

						xRect.MinX += xDeltaPos.x;
						xRect.MinY += xDeltaPos.y;
						xRect.MaxX += xDeltaPos.x;
						xRect.MaxY += xDeltaPos.y;
						break;
					}

					case ASURA_GUIMENU_MODIFY_POSITION_RELATIVE:
						{
							Asura_Vector_2 xDims(Asura_Maths::ZeroVector2);
							Asura_GUIMenu_Menu_Page* pxPage = FindOwnerMenu();

							Asura_Vector_2 xDeltaPos = pxMod->GetVector2Data();

							if( pxPage)
							{
								xDims = pxPage->GetScaledWorldBounds().GetDimension();
							}

							xRect.MinX += xDeltaPos.x * xDims.x;
							xRect.MinY += xDeltaPos.y * xDims.y;
							xRect.MaxX += xDeltaPos.x * xDims.x;
							xRect.MaxY += xDeltaPos.y * xDims.y;
							break;
						}

					case ASURA_GUIMENU_MODIFY_COLOUR:
					{
						uColour = pxMod->GetColourData().ToUIntARGB();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TRANSLUCENCY_MODE:
					{
						eTranslucencyMode = (ASURA_TRANSLUCENCY_MODE)pxMod->GetIntData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_WIDTH:
					{
						xRect.MaxX = xRect.MinX + pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_HEIGHT:
					{
						xRect.MaxY = xRect.MinY + pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TEXT_STRING:
					{
						*ppwszText = pxMod->GetUnicodeCharStringData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TEXT_OFFSET_X:
					{
						fOffsetX = pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TEXT_OFFSET_Y:
					{
						fOffsetY = pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TEXT_RENDER_FLAGS:
					{
						uRenderFlags = pxMod->GetUIntData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TEXT_ID:
					{
						*ppwszText = Asura_HashedLocalisedText::GetString( m_uActiveTextPage, pxMod->GetUIntData());
						break;
					}

					case ASURA_GUIMENU_MODIFY_FONT_SCALE:
					{
						fFontScale = pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_FONT_NAME:
					{
						iFontID = Asura_Font_System::GetFontHandle( pxMod->GetCharStringData());
						break;
					}

					case ASURA_GUIMENU_MODIFY_ALPHA_SCALED:
					{
						int iAlpha = uColour >> 24;
						float fAlphaScale = pxMod->GetFloatData();
						Asura_Assert(fAlphaScale >= 0.0f && fAlphaScale <= 1.0f, "alpha scale out of range");
						float fNewAlpha = fAlphaScale * static_cast<float>(iAlpha);
						iAlpha = Asura_Maths::FloatToInt( fNewAlpha );
						Asura_Assert(iAlpha >= 0 && iAlpha <= 255, "alpha out of range");
						u_int uAlpha = iAlpha << 24;
						uColour = (uColour & 0x00ffffff) | uAlpha;
						break;
					}


					default:
					{
						// possibly just ignore with no assert?
						Asura_Assert(false, "unhandled modify type");
						break;
					}
				}// end switch( pxMod->GetModifyType())
			}
		}
	}

	// apply animators from this object and all parent objects
	const Asura_GUIMenu_Root* pxElementItr = this;
	while ( pxElementItr )
	{
		for( u=0; u<pxElementItr->GetAnimatorCount(); u++)
		{
			Asura_GUIMenu_Animator* pxAnim = pxElementItr->GetAnimator( u);
			Asura_Assert( pxAnim, "shouldn't fail");
			ApplyAnimatorChanges( pxAnim, uStateToMatch, xRect, fOffsetX, fOffsetY, fFontScale, uColour, eTranslucencyMode, uRenderFlags, ppwszText, iFontID );
		}// end loop on animators

		pxElementItr = pxElementItr->GetParentObject();
	}
}

void Asura_GUIMenu_Element_StatusText::ApplyAnimatorChanges( Asura_GUIMenu_Animator* pxAnim, u_int uStateToMatch, Asura_Bounding_Rectangle& xRect, float& fOffsetX, float& fOffsetY, float& fFontScale, u_int& uColour, ASURA_TRANSLUCENCY_MODE& eTranslucencyMode,u_int& uRenderFlags, const Asura_UnicodeChar** ppwszText, int& iFontID)
{
	Asura_Assert(pxAnim, "null pointer");
	if(pxAnim->ShouldUseAnimator(uStateToMatch))
	{
		switch (pxAnim->GetModifyType())
		{
			case ASURA_GUIMENU_MODIFY_POSITION:
			{
				Asura_Vector_2 xPos = pxAnim->GetVector2Data();

				xRect.MinX += xPos.x;
				xRect.MinY += xPos.y;

				xRect.MaxX += xPos.x;
				xRect.MaxY += xPos.y;
				break;
			}

			case ASURA_GUIMENU_MODIFY_POSITION_RELATIVE:
				{
					Asura_Vector_2 xDims(Asura_Maths::ZeroVector2);
					Asura_GUIMenu_Menu_Page* pxPage = FindOwnerMenu();

					Asura_Vector_2 xPos = pxAnim->GetVector2Data();

					if( pxPage)
					{
						xDims = pxPage->GetScaledWorldBounds().GetDimension();
					}

					xRect.MinX += xPos.x * xDims.x;
					xRect.MinY += xPos.y * xDims.y;

					xRect.MaxX += xPos.x * xDims.x;
					xRect.MaxY += xPos.y * xDims.y;
					break;
				}

			case ASURA_GUIMENU_MODIFY_COLOUR:
			{
				u_int uAnimColour = pxAnim->GetRGBAData().ToUIntARGB();
				//Ignore the alpha component for now I think , since that wlll probably be more useful.
				//May need to add a flag later to determine which components to use.
				//RAR 25/07/2007
				uColour  = (uColour & 0xff000000) | (uAnimColour & 0x00ffffff);
				break;
			}

			case ASURA_GUIMENU_MODIFY_ALPHA_SCALED:
			{
				int iAlpha = uColour >> 24;
				float fAlphaScale = pxAnim->GetFloatData();
				Asura_Assert(fAlphaScale >= 0.0f && fAlphaScale <= 1.0f, "alpha scale out of range");
				float fNewAlpha = fAlphaScale * static_cast<float>(iAlpha);
				iAlpha = Asura_Maths::FloatToInt( fNewAlpha );
				Asura_Assert(iAlpha >= 0 && iAlpha <= 255, "alpha out of range");
				u_int uAlpha = iAlpha << 24;
				uColour = (uColour & 0x00ffffff) | uAlpha;
				break;
			}

			case ASURA_GUIMENU_MODIFY_WIDTH_SCALED:
			{
				float fWidth = xRect.MaxX - xRect.MinX;
				fWidth *= pxAnim->GetFloatData();
				xRect.MaxX = xRect.MinX + fWidth;
				break;
			}

			case ASURA_GUIMENU_MODIFY_HEIGHT_SCALED:
			{
				float fHeight = xRect.MaxY - xRect.MinY;
				fHeight *= pxAnim->GetFloatData();
				xRect.MaxY = xRect.MinY + fHeight;
				break;
			}

			case ASURA_GUIMENU_MODIFY_TEXT_OFFSET_X:
			{
				fOffsetX = pxAnim->GetFloatData();
				break;
			}

			case ASURA_GUIMENU_MODIFY_TEXT_OFFSET_Y:
			{
				fOffsetY = pxAnim->GetFloatData();
				break;
			}


			default:
			{
				//Not sure its worth having an assert here , since if an animator is applied to a widget , then
				//the child elements will get some types fed to it that can't be handled.
				//RAR 17/04/2007
				//Asura_Assert(false, "unhandled data type");
				break;
			}
		} // end switch
	}// end if (state is affected)
	else
	{
		pxAnim->Start();
	}
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_StatusText::ListUsedTextPages(Asura_HashTree& xTextPageTree) const
{
	for(u_int i=0 ; i<m_uTextInfoEntryCount; i++)
	{
		if(m_pxTextInfoEntries[i].m_uTextPage)
		{
			int iJunkIndex;
			xTextPageTree.FindNodeIndexByHash(m_pxTextInfoEntries[i].m_uTextPage, true, iJunkIndex);
		}
	}
	
	PARENT::ListUsedTextPages(xTextPageTree);
}
#endif //ASURA_TOOLS

#endif // ASURA_USE_GUI_MENU_SYSTEM
