// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Widget_Menu_Page.cpp											//
//																						//
//	Description: menu page class														//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"

// This file's header
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
// Asura
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"

#include "Debugging/Asura_Logfile.h"

#include "GUIMenu/Asura_GUIMenu_Audio.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Base.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Button.h"
#include "GUIMenu/Asura_GUIMenu_Message.h"
#include "GUIMenu/Asura_GUIMenu_MessagePump.h"
#include "GUIMenu/Asura_GUIMenu_System.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"

#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Render.h"


#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_PreviewOverlay.h"
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#include "GUIMenu/Tools_Mode_GUIMenu.h"
#include "Interface/Asura_Overlay_System.h"
#endif //ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
// -- Constants --------------------------------------------------------------------------
const float fASURA_MAX_MENU_TRANSITION_TIME = 1.0f;

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Menu_Page::Asura_GUIMenu_Menu_Page( void)
//: m_uWidgetCount( 0)
: m_pxActiveWidget( NULL)
, m_uHashStartupWidget( ASURA_HASH_ID_UNSET)
, m_uStartupWidgetOverride( ASURA_HASH_ID_UNSET)
, m_pxBackButton( NULL )
, m_uHashBackMenuID( ASURA_HASH_ID_UNSET)
, m_uHashBackWidgetID( ASURA_HASH_ID_UNSET)
, m_uMenuPageFlags( ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_BACK_INTO_THIS | 
				   ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_BACK_FROM_THIS | 
				   ASURA_GUIMENU_MENUPAGE_FLAG_CLOSE_THIS_MENU_ON_BACK |
				   ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_KEY_PRESS_BACK_FROM_THIS)	// CLF (07.11.08)
, m_uState( ASURA_GUIMENU_ACTION_NONE )
, m_fStateTimer( 0.0f )
, m_fFocusTimer( 0.0f )
, m_uRefCount( 0)
{
	m_xVirtualSize.Zero();
	m_uType = ASURA_GUIMENU_TYPEID_MENUPAGE;

	#ifdef ASURA_TOOLS
	//Assign all menus to the 'Level' file by default for now
	m_xFilename = "Level";
	#endif	

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__GUIMENU_PAGE);
}

// ---------------------------------------------------------------------------------------
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Menu_Page::~Asura_GUIMenu_Menu_Page( void)
{
#ifdef ASURA_GUIMENU_STORE_NAMES
	m_xName.MakeNull();
#endif
	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__GUIMENU_PAGE);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::Update( float fElapsedTime)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__GUIMENU_PAGE);

#ifdef ASURA_GUIMENU_ENABLE_METRICS
	Asura_GUIMenu_System::s_xMetrics.uUpdateCalls_MenuPages++;
#endif

	if(Asura_GUIMenu_System::GetFocusedMenuHashID() == GetID())
	{
		//Update the record of the current active widget.
		//RAR 19/04/2007
		Asura_GUIMenu_System::SetActiveWidgetConsoleVar(m_uHashActiveWidget);
	}

	// call baseclass update for display condition functionality
	// (MH - 10.07.2007) don't carry on updating the display conditions
	// and contained items once page is fading out.
	if (m_uState != ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT)
	{
		Asura_GUIMenu_Root::Update( fElapsedTime );
		GetOnUpdateCommand()->Execute( NULL);
	}

	switch (m_uState)
	{
		case ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_IN:
		{
			// wait for previous menu to finish fading out?
			if (Asura_GUIMenu_System::IsMenuFadingOut())
			{
				ASURA_GUIMENU_PROFILE_END_TASK();
				return;
			}
			m_fStateTimer -= fElapsedTime;
			if (m_fStateTimer <= 0.0f)
			{
				m_fStateTimer = 0.0f;
				m_uState = ASURA_GUIMENU_ACTION_FLAG_IS_MENU_ACTIVE;
			}
			break;
		}
		case ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT:
		{
			m_fStateTimer -= fElapsedTime;
			if (m_fStateTimer <= 0.0f)
			{
				Asura_GUIMenu_System::EndMenuFadingOut( this );
				m_fStateTimer = 0.0f;
				m_uState = ASURA_GUIMENU_ACTION_NONE;
			}
			break;
		}
		case ASURA_GUIMENU_ACTION_FLAG_IS_MENU_ACTIVE:
		{
			m_fStateTimer += fElapsedTime;
			break;
		}
		case ASURA_GUIMENU_ACTION_NONE:
		{
			// inactive
			//Asura_Assertf(false, ("shouldn't have Update() called in this state (menu %s)", GetNameAsChar()));
			// can't assert here as OnInit() only gets called in the next update loop, but menu is flagged as
			// active immediately, so we might get an update even though the state hasn't changed yet.
			break;
		}
		default:
		{
			Asura_Assert(false, "unhandled case");
			break;
		}
	}
	if(Asura_GUIMenu_System::GetFocusedMenuHashID() == GetID())
	{
		m_fFocusTimer += fElapsedTime;
	}
	else
	{
		m_fFocusTimer = 0.0f;
	}

	// (MH - 09.03.2007) check active widget is selectable
	if(m_pxActiveWidget != NULL)
	{
		u_int uType = m_pxActiveWidget->GetType();
		if (uType >= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BEGIN && 
			uType <= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_WIDGET_END)
		{
			// it's a widget
			Asura_GUIMenu_Widget_Base* pxWidget = static_cast<Asura_GUIMenu_Widget_Base*>(m_pxActiveWidget);
			if (!pxWidget->IsSelectable())
			{
				Asura_GUIMenu_MessagePump::PostMessage( m_uHashActiveWidget, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT, 0);
				m_pxActiveWidget = NULL;
				m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
			}
		}
		else
		{
			Asura_Assert(false, "m_pxActiveWidget is not a widget");
			m_pxActiveWidget = NULL;
			m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();

}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::Copy( const Asura_GUIMenu_Root* pxSource )
{
	PARENT::Copy( pxSource );

	const Asura_GUIMenu_Menu_Page* pxSourceMenu = static_cast<const Asura_GUIMenu_Menu_Page*> (pxSource);

	SetStartupWidgetID(pxSourceMenu->GetStartupWidgetID());
	m_xOnInitCommand.Copy(&pxSourceMenu->m_xOnInitCommand);
	m_xOnUpdateCommand.Copy(&pxSourceMenu->m_xOnUpdateCommand);
	m_xOnDeInitCommand.Copy(&pxSourceMenu->m_xOnDeInitCommand);
	SetMenuPageFlags(pxSourceMenu->GetMenuPageFlags());
	m_xVirtualSize = pxSourceMenu->GetVirtualSize();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::Resize( const float fNewWidth, const float fNewHeight)
{
	Asura_Bounding_Rectangle xNewBounds( GetUnscaledLocalBounds() );

	xNewBounds.MaxX = xNewBounds.MinX + fNewWidth;
	xNewBounds.MaxY = xNewBounds.MinY + fNewHeight;

	Resize(xNewBounds);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::GetUnscaledVirtualBounds(Asura_Bounding_Rectangle& xBounds) const
{
	xBounds = GetUnscaledLocalBounds();
	if (HasVirtualSize()
		&& xBounds.GetWidth()
		&& xBounds.GetHeight())
	{
		// NY - 09 April 2009 - if we have a virtual size, convert bounds into virtual space
		float fXScale = m_xVirtualSize.x / xBounds.GetWidth();
		float fYScale = m_xVirtualSize.y / xBounds.GetHeight();
		xBounds.MinX *= fXScale;
		xBounds.MaxX *= fXScale;
		xBounds.MinY *= fYScale;
		xBounds.MaxY *= fYScale;
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::HasVirtualSize() const
{
	return m_xVirtualSize.x >= 1.0f
		&& m_xVirtualSize.y >= 1.0f;
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Widget_Base* Asura_GUIMenu_Menu_Page::FindBestWidget() const
{
	// (MH - 09.03.2007) to help avoid phantom menu pages, find an active widget.
	// Prioritise buttons, and of those prioritise the one with the most navigation links.
	Asura_GUIMenu_Widget_Base* pxBestWidget = NULL;
	
	const u_int uContainerCount = GetContainerCount();
	for( u_int i=0; i<uContainerCount; i++)
	{
		u_int uType = m_ppxContainerParts[i]->GetType();
		if (uType >= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BEGIN && 
			uType <= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_WIDGET_END)
		{
			// it's a widget
			Asura_GUIMenu_Widget_Base* pxWidget = static_cast<Asura_GUIMenu_Widget_Base*>(m_ppxContainerParts[i]);
			if (pxWidget->IsSelectable())
			{
				if (pxBestWidget)
				{
					if (pxWidget->GetType() == ASURA_GUIMENU_TYPEID_WIDGET_BUTTON)
					{
						if (pxBestWidget->GetType() == ASURA_GUIMENU_TYPEID_WIDGET_BUTTON)
						{
							// both are buttons, which has most links?
							u_int uNumLinksBestWidget = 0;
							u_int uNumLinksNewWidget = 0;
							for (u_int u = 0; u < ASURA_GUIMENU_UI_NAV_EXIT_MAX; ++u)
							{
								if (pxBestWidget->GetNavExitID( static_cast<Asura_GUIMenu_Navigation_Exit>(u) ) != ASURA_HASH_ID_UNSET)
								{
									++uNumLinksBestWidget;
								}
								if (pxWidget->GetNavExitID( static_cast<Asura_GUIMenu_Navigation_Exit>(u) ) != ASURA_HASH_ID_UNSET)
								{
									++uNumLinksNewWidget;
								}
							}
							if (uNumLinksNewWidget > uNumLinksBestWidget)
							{
								pxBestWidget = pxWidget;
							}
						}
						else
						{
							pxBestWidget = pxWidget;
						}
					}
				}
				else
				{
					pxBestWidget = pxWidget;
				}

				// if the current selectable widget is the startup widget then return this as the best widget.
				if( pxBestWidget->GetID() == m_uHashStartupWidget || pxBestWidget->GetID() == m_uStartupWidgetOverride )
				{
					return pxBestWidget;
				}
			}
		}
	}
	return pxBestWidget;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::RemoveActiveWidget()
{
	m_pxActiveWidget = NULL;
	m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Menu_Page::ProcessInputActivity( const Asura_GUIMenu_Input& xInput)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__INPUT_ACTIVITY);

#ifdef ASURA_TOOLS
	// in tools preview, if entering text or otherwise focused on a gadget, ignore input.
	if (Tools_Mode_GUIMenu::IsPreview())
	{
		if (Asura_Overlay_System::GetFocusGadget())
		{
			ASURA_GUIMENU_PROFILE_END_TASK();
			return 0;
		}
	}
#endif

	u_int uButtonsTrapped = 0;
	u_int uButtonPresses = xInput.m_uButtonPresses;

	if(xInput.m_bUseCursor)
	{
		//Check for selection changing as a result of the cursor moving over another widget.
		//Don't do this while the mouse button (or similar) is being held down.
		//RAR 01/08/2007
		bool bButtonHeld = (uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK_HELD) && !(uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK);
		if(!bButtonHeld)
		{
			Asura_GUIMenu_Widget_Base* pxCursorWidget = FindSelectableWidgetAtPos(xInput.m_xCursorPos);

			//Change the selected widget if the cursor is over a different one.
			//Can't really deselect when not over any widget since that would break navigation via keypresses , so
			//just the leave the previous selection in that case
			//RAR 01/08/2007
			if((pxCursorWidget || Asura_GUIMenu_System::GetAllowCursorDeselect())
				 && pxCursorWidget != m_pxActiveWidget)
			{
				
				// -- notify the next item to be selected
				if (pxCursorWidget)
				{
					const Asura_Message_GUIMenu_Notify& xMsg = Asura_GUIMenu_MessagePump::CreateMessage( pxCursorWidget->GetID(), GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_IN, 0);
					pxCursorWidget->Whisper(&xMsg);
				}

				// -- notify the current item, to be deselected --
				if(m_pxActiveWidget)
				{
					const Asura_Message_GUIMenu_Notify& xMsg = Asura_GUIMenu_MessagePump::CreateMessage( m_pxActiveWidget->GetID(), GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT, 0);
					m_pxActiveWidget->Whisper(&xMsg);
				}

				//Play sound for changing selection by cursor				
				Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_CURSOR_NAVIGATE);

				m_pxActiveWidget = pxCursorWidget;
				if (m_pxActiveWidget)
				{
					m_uHashActiveWidget = m_pxActiveWidget->GetID();
				}
				else
				{
					m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
				}
			}
		}
	}

	if( CanKeyPressBackOutOfThisMenu() && Asura_GUIMenu_System::Platform_ProcessMenuBackControls( uButtonPresses, uButtonsTrapped))
	{
		// (MH - 11.05.2007) if a selectable back button with a valid hotlink exists prefer it over the menu back.
		// (MXH - 27.02.2009) Also if there is a button set to trap the back press and it has commands to run
		if (m_pxBackButton && m_pxBackButton->IsSelectable() &&
			 ( m_pxBackButton->GetLinkToHash() != ASURA_HASH_ID_UNSET ||
			 m_pxBackButton->DoesButtonHaveActionCommands() == true ))
		{
			// (03.07.2007 - CJW) SOUND REMOVED - Back button sound effects get played from the button 
			// widget to which the processing of input activity is passed on.
			
			// tell the button accept was pressed.
			Asura_GUIMenu_Input xAccept;
			xAccept.m_uButtonPresses = ASURA_GUIMENU_ACTION_BUTTON_ACCEPT;
			m_pxBackButton->ProcessInputActivity( xAccept );
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CANCEL;
		}
		else if(CanBackOutOfThisMenu())
		{
			bool bIsCodeUpdate = false;

			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CANCEL;
			// -- play the cancel sound
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_CANCEL);

			// -- returning back to the last page so don't set a backlink.
			Asura_Hash_ID uMenuHash, uWidgetHash;
			GetBackMenuAndWidgetHash( uMenuHash, uWidgetHash);

			if (uMenuHash)
			{
				Asura_GUIMenu_Menu_Page* pxPageToBackInto = Asura_GUIMenu_System::GetMenuPageByID( uMenuHash);
				if( pxPageToBackInto)
				{
					if( (pxPageToBackInto->GetMenuPageFlags()&ASURA_GUIMENU_MENUPAGE_FLAG_CODE_UPDATE_AND_RENDER_ONLY) != 0)
					{
						bIsCodeUpdate = true;
					}
				}

				if( bIsCodeUpdate)
				{
					Asura_GUIMenu_System::InitMenuPageData( uMenuHash, true);
				}
				else
				{
					Asura_GUIMenu_System::OpenMenuPageOnWidget( uMenuHash, uWidgetHash, true, false);
				}
			}

			ClearBackMenuHash();

			if( ShouldCloseMenuOnBackAction())
			{
				if( (m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_CODE_UPDATE_AND_RENDER_ONLY) != 0)
				{
					Asura_GUIMenu_System::DeInitMenuPageData( GetID());
				}
				else
				{
					Asura_GUIMenu_System::CloseMenuPage( GetID());
				}
			}
		}
	}
	else if(m_pxActiveWidget != NULL)
	{
		uButtonsTrapped = m_pxActiveWidget->ProcessInputActivity( xInput);
	}

	if (m_pxActiveWidget == NULL)
	{
		if (!Asura_GUIMenu_System::GetAllowCursorDeselect())
		{
			// find the best widget
			Asura_GUIMenu_Widget_Base* pxBestWidget = Asura_GUIMenu_Menu_Page::FindBestWidget();
			if (pxBestWidget)
			{
				m_pxActiveWidget = pxBestWidget;
				m_uHashActiveWidget = m_pxActiveWidget->GetID();
				Asura_GUIMenu_MessagePump::PostMessage( m_uHashActiveWidget, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_IN, 0);
			}
			else
			{
				// this can happen in status dialogs while waiting for a condition.
				m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
			}
		}
		else
		{
			m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();

	return uButtonsTrapped;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::Render( u_int uState)
{
	//Prevent the menu from being rendered before OnInit() gets called. RAR 17/12/2008
	if(m_uState == ASURA_GUIMENU_ACTION_NONE) return;
	
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__GUIMENU_PAGE);

	if( !IsRenderingDisabled() )
	{
		if( (m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_RENDER_TO_FULLSCREEN) != 0)
		{
			Asura_Render::AllowFullScreenRendering();
		}

		// render widgets and standalone elements
		const u_int uContainerCount = GetContainerCount();
		for( u_int i=0; i<uContainerCount; i++)
		{
			if (!m_ppxContainerParts[i]->IsRenderingDisabled())
			{
				// render standalone elements using all state flags; widgets get passed this menu's state.
				if (m_ppxContainerParts[i]->IsWidget())
				{
					m_ppxContainerParts[i]->Render( GetState() );
				}
				else
				{
					m_ppxContainerParts[i]->Render( ASURA_GUIMENU_ACTION_FLAG_ALL | GetState());

					// force draw order to be correct. RAR 28/06/2007
					Asura_Render::FlushDisplay();
				}
			}
		}

		if ( DoesMenuRenderWidescreenBars() == true )
		{
			Asura_GUIMenu_System::EnableWidescreenBarsThisFrame();
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::IsWidgetOnMenu( Asura_Hash_ID uHashId)
{
	return FindWidget( uHashId) != 0;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::HandleMessage( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bRet = false;
	u_int uMessageType = ASURA_GUIMENU_GET_MAIN_MESSAGE( pxMessage->uActionType);


	if( uMessageType != ASURA_GUIMENU_UI_MSG_NO_TYPE)
	{
		switch( uMessageType)
		{
			case ASURA_GUIMENU_UI_MSG_ON_INIT:
				bRet = OnInit( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ON_DEINIT:
				bRet = OnDeInit( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE:
				bRet = OnDataChange( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ON_DATA_REQUEST:
				bRet = OnDataRequest( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ON_CLICKED:
				bRet = OnClicked( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ON_ROLL_IN:
				// if the to id isn't for the menu and it's not for everything,
				// it's for a widget that's contained on this page.  The parent menu
				// must keep track of which widget has the current focus -- 
				if( (pxMessage->uToID != ASURA_GUIMENU_ID_ALL) && (pxMessage->uToID != GetID()))
				{
					m_uHashActiveWidget = pxMessage->uToID;
					m_pxActiveWidget = FindWidget( m_uHashActiveWidget);

					if (m_pxActiveWidget)
					{
						Asura_Assertf(m_pxActiveWidget->IsWidget(), ("initial widget %s on page %s is not widget type", 
							m_pxActiveWidget->GetNameAsChar(), GetNameAsChar()));
						if (!m_pxActiveWidget->IsWidget())
						{
							m_pxActiveWidget = NULL;
							m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
						}
					}
				}

				bRet = OnRollIn( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT:
				bRet = OnRollOut( pxMessage);
				break;

			case ASURA_GUIMENU_UI_MSG_ENABLE_DISABLE:
				bRet = OnEnable( pxMessage);
				break;

			default:
				break;
		}
	}

	return bRet;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::AddToThisContainer( Asura_GUIMenu_Root* pxObject)
{
	bool bRet = false;
	Asura_Assert( pxObject != NULL, "Asura_GUIMenu_Menu_Page::AddToThisContainer - Asura_GUIMenu_Root* pxObject == NULL");

//#ifdef ASURA_TOOLS
	CheckAndExpandContainerIfNecessary();
//#endif

	if( m_usContainerPartsCount < m_usMaxContainerPartsCount )
	{
		m_ppxContainerParts[m_usContainerPartsCount++] = pxObject;
		pxObject->SetParentObject( this);

		bRet = true;
	}

	return bRet;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::Whisper( const Asura_Message_GUIMenu_Notify* pxNotify)
{
	bool bMessageHandled = false;
	Asura_GUIMenu_Widget_Base* pxWidget = NULL;
	bool bIsOwnerOfWidgetHash = false;

	Asura_Assert( pxNotify != NULL, "Asura_GUIMenu_Menu_Page::Whisper - Asura_Message_GUIMenu_Notify* pxNotify == NULL");
	// -- Is message meant for this menu page? 
	// -- GUIMENU_ID_ALL signified that the message is meant for everything --

	if( (pxNotify->uToID == GetID()) || (pxNotify->uToID == ASURA_GUIMENU_ID_ALL))
	{
		bMessageHandled = HandleMessage( pxNotify);
	}
	else
	{
		pxWidget = FindWidget( pxNotify->uToID);
		bIsOwnerOfWidgetHash = (pxWidget != 0);
	}

	if( pxWidget == NULL)
	{
		if( (bIsOwnerOfWidgetHash) || (pxNotify->uToID == ASURA_GUIMENU_ID_ALL))
		{
			if( (bMessageHandled == false))
			{
				for(u_int u = 0; u < GetContainerCount(); ++u)
				{
					bMessageHandled = m_ppxContainerParts[u]->Whisper( pxNotify);

					if( (bMessageHandled == true) && (pxNotify->uToID != ASURA_GUIMENU_ID_ALL))
					{	break;	}
				}
			}	
		}
	}
	else
	{
		bMessageHandled = HandleMessage( pxNotify);
		bMessageHandled = pxWidget->Whisper( pxNotify);
	}

	return bMessageHandled;
}


//bool Asura_GUIMenu_Menu_Page::Whisper( const Asura_Message_GUIMenu_Notify* pxNotify)
//{
//	u_int i;
//	bool bMessageHandled = false;
//
//	Asura_Assert( pxNotify != NULL, "Asura_GUIMenu_Menu_Page::Whisper - Asura_Message_GUIMenu_Notify* pxNotify == NULL");
//
//	// -- Is message meant for this menu page? 
//	// -- GUIMENU_ID_ALL signified that the message is meant for everything --
//
//
//
//	bool bIsOwnerOfWidgetHash = IsWidgetOnMenu( pxNotify->uToID);
//
//	if( (pxNotify->uToID == GetID()) || (pxNotify->uToID == ASURA_GUIMENU_ID_ALL) || (bIsOwnerOfWidgetHash == true))
//	{
//		bMessageHandled = HandleMessage( pxNotify);
//	}
//
//	if( bIsOwnerOfWidgetHash)
//	{
//		if( (bMessageHandled == false) || (pxNotify->uToID == ASURA_GUIMENU_ID_ALL))
//		{
//			for( i=0; i<m_uContainerPartsCount; i++)
//			{
//				bMessageHandled = m_ppxContainerParts[i]->Whisper( pxNotify);
//
//				if( (bMessageHandled == true) && (pxNotify->uToID != ASURA_GUIMENU_ID_ALL))
//				{	break;	}
//			}
//		}	
//	}
//
//	return bMessageHandled;
//}

// --------------------------------------------------------------------------------------- 

Asura_Hash_ID Asura_GUIMenu_Menu_Page::GetStartupWidgetID( void) const
{
	return m_uHashStartupWidget;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::SetStartupWidgetID( Asura_Hash_ID uHashStartupWidgetID)
{
	const Asura_GUIMenu_Root* pxItem = FindContainerItemByID(uHashStartupWidgetID);
	
	if(pxItem && pxItem->IsWidget())
	{
		m_uHashStartupWidget = uHashStartupWidgetID;
	}
}

// --------------------------------------------------------------------------------------- 

Asura_Hash_ID Asura_GUIMenu_Menu_Page::GetStartupOverrideWidgetID( void) const
{
	return m_uStartupWidgetOverride;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::SetStartupOverrideWidgetID( Asura_Hash_ID uHashStartupOverrideWidgetID)
{
	if( FindContainerObject( uHashStartupOverrideWidgetID) != NULL)
	{
		m_uStartupWidgetOverride = uHashStartupOverrideWidgetID;
	}
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Menu_Page::GetMenuPageFlags( void) const
{
	return m_uMenuPageFlags;	
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Menu_Page::UpdateLocalDataFromStyle( void)
{
	Asura_GUIMenu_Root::UpdateLocalDataFromStyle();

	if( m_pxStyleObject != NULL)
	{

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_WIDTH) == 0)
		{
			Asura_GUIMenu_Menu_Page* pxStyleMenuPage = static_cast<Asura_GUIMenu_Menu_Page*>(m_pxStyleObject);
			
			m_xUnscaledLocalBounds.MaxX = pxStyleMenuPage->GetUnscaledLocalBounds().GetWidth() + m_xUnscaledLocalBounds.MinX;
			m_xVirtualSize.x = pxStyleMenuPage->GetVirtualSize().x;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_HEIGHT) == 0)
		{
			Asura_GUIMenu_Menu_Page* pxStyleMenuPage = static_cast<Asura_GUIMenu_Menu_Page*>(m_pxStyleObject);

			m_xUnscaledLocalBounds.MaxY = pxStyleMenuPage->GetUnscaledLocalBounds().GetHeight() + m_xUnscaledLocalBounds.MinY;
			m_xVirtualSize.y = pxStyleMenuPage->GetVirtualSize().y;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_MENUPAGE_FLAGS) == 0)
		{
			Asura_GUIMenu_Menu_Page* pxStyleMenuPage = static_cast<Asura_GUIMenu_Menu_Page*>(m_pxStyleObject);
			m_uMenuPageFlags = pxStyleMenuPage->m_uMenuPageFlags;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_ONINIT) == 0)
		{
			Asura_GUIMenu_Menu_Page* pxStyleMenuPage = static_cast<Asura_GUIMenu_Menu_Page*>(m_pxStyleObject);
			m_xOnInitCommand.Copy(&pxStyleMenuPage->m_xOnInitCommand);
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_ONUPDATE) == 0)
		{
			Asura_GUIMenu_Menu_Page* pxStyleMenuPage = static_cast<Asura_GUIMenu_Menu_Page*>(m_pxStyleObject);
			m_xOnUpdateCommand.Copy(&pxStyleMenuPage->m_xOnUpdateCommand);
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_ONDEINIT) == 0)
		{
			Asura_GUIMenu_Menu_Page* pxStyleMenuPage = static_cast<Asura_GUIMenu_Menu_Page*>(m_pxStyleObject);
			m_xOnDeInitCommand.Copy(&pxStyleMenuPage->m_xOnDeInitCommand);
		}
	}

	// SK 01.03.07 - i don't think needs to be here, i'll check and remove later.
	const u_int uContainerCount = GetContainerCount();
	for(u_int u = 0; u < uContainerCount; ++u)
	{
		m_ppxContainerParts[u]->UpdateLocalDataFromStyle();
	}
}
#endif
// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
//	u_int uContainerCount, i;
	u_int uStartupWidget;

	// SK - when reading in the root info, the call doesn't require a chunk version number or parent object
	Asura_GUIMenu_Root::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	if(uChunkVersion < 5)
	{
		#ifdef ASURA_TOOLS
		//Need to set defaults for OnInit, OnDeinit and OnUpdate local data flags on old menus. RAR 23/04/2007
		const u_int uFlagsToAdd = ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_ONINIT|ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_ONDEINIT|ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_ONUPDATE;
		
		if(m_uHashStyleObject)
		{
			m_uStyleUseLocalDataFlags |= uFlagsToAdd;
		}

		if(m_uToolsFlags & ASURA_GUIMENU_TOOLS_FLAG_ISSTYLE)
		{
			m_uStyleChildEditableFlags |= uFlagsToAdd;
		}
		#endif
	}
	if(uChunkVersion < 8)
	{
		// -- read in the width & height --
		float fWidth, fHeight;
		xStream >> fWidth;
		xStream >> fHeight;
		m_xUnscaledLocalBounds.MaxX = fWidth + m_xUnscaledLocalBounds.MinX;
		m_xUnscaledLocalBounds.MaxY = fHeight + m_xUnscaledLocalBounds.MinY;
	}

	Asura_GUIMenu_System::SetLoadingMenu(this);
	ReadAllChildObjects( xStream, this);
	Asura_GUIMenu_System::SetLoadingMenu(0);
	if (uChunkVersion < 8)
	{
		// NY - 31 July 2008 - bounds weren't chunked at a base level; make sure they're set up.
		for(u_int u = 0; u < m_usContainerPartsCount; ++u)
		{
			Asura_GUIMenu_Root* pxRootItem = m_ppxContainerParts[u];
			pxRootItem->FixupBoundsFromChildren();
		}
	}

	// -- Setup the starting widget for this menu --
	xStream >> uStartupWidget;

	SetStartupWidgetID( uStartupWidget);

#ifdef ASURA_TOOLS
	// set filename from either the currently open file or the data we saved in the chunk
	Asura_String xFilename;
	
	if(uChunkVersion >= 1)
	{
		char szFile[128];
		xStream.ReadString(szFile,sizeof(szFile));
		xFilename = szFile;
	}

	// want to keep 'level' as a filename as it's a 'special' one.
	if( (xFilename.Compare("Level") != 0) 
	 && (xStream.GetFileHandle() != ASURA_INVALID_HANDLE) )
	{
		// otherwise, if we can, just use the filename that the stream is currently 
		// loading from. This allows us to rename .gui files without having to mash the chunk data.
		Asura_String xExt;
		Asura_String xPath;
		Asura_String xName( Asura_FileAccess::GetFilename( xStream.GetFileHandle() ) ); 
		xName.SplitFilename( &xPath, &xName, &xExt );

		// want to preserve any \guimenu-relative path information
		int iSlashPos = 0;
		xPath.ReplaceAllChars( '/', '\\' );
		xPath.MakeLowercase();
		xPath.Contains("guimenu",&iSlashPos);
		iSlashPos += strlen("guimenu");
		xPath = xPath.SubString( (iSlashPos+1), xPath.GetLength() ); // skip the trailing slash
		
		if( xExt.Compare("gui") == 0 )
		{			 
			if( xPath.GetLength() > 0 )
			{
				xFilename = xPath;
				xFilename += "\\";
				xFilename += xName;
			}
			else
			{
				xFilename = xName;
			}
		}
	}

	SetFilename( xFilename.GetString() );

	if(uChunkVersion >= 2 && uChunkVersion<=6)
	{
		int iPlatform;
		xStream >> iPlatform;
		if(iPlatform >= 0)
		{
			SetPlatforms (1<<iPlatform);
		}

		//trasfer style flag for platform as well
		if(m_uStyleUseLocalDataFlags &  OLD_ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_PLATFORM)
		{
			m_uStyleUseLocalDataFlags &=~  OLD_ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_PLATFORM;
			m_uStyleUseLocalDataFlags |= ASURA_GUIMENU_ROOT_LOCAL_FLAG_PLATFORMS;
		}
		if(m_uStyleChildEditableFlags &  OLD_ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_PLATFORM)
		{
			m_uStyleChildEditableFlags &=~  OLD_ASURA_GUIMENU_MENU_PAGE_LOCAL_FLAG_PLATFORM;
			m_uStyleChildEditableFlags |= ASURA_GUIMENU_ROOT_LOCAL_FLAG_PLATFORMS;
		}
	}
#else
	//skip filename
	if(uChunkVersion >= 1)
	{
		xStream.SkipString();
	}

	if(uChunkVersion >= 2 && uChunkVersion<=6)
	{
		int iPlatform;
		xStream >> iPlatform;
		// do nothing with this on exported data
	}

#endif

	if( uChunkVersion >= 3)
	{
		m_xOnInitCommand.ReadFromChunkStream( xStream);
		m_xOnDeInitCommand.ReadFromChunkStream( xStream);
		if(uChunkVersion >= 6)
		{
			m_xOnUpdateCommand.ReadFromChunkStream( xStream);
		}
	}

	if( uChunkVersion >= 4)
	{
		xStream >> m_uMenuPageFlags;
	}

	// CLF (07.11.08) - Support for new flag allowing key press back functionality to be disabled
	if (uChunkVersion < 9)
	{
		if ((GetMenuPageFlags() & ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_BACK_FROM_THIS) != 0)
		{
			//Enable key press if back from this menu is enabled
			m_uMenuPageFlags |= ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_KEY_PRESS_BACK_FROM_THIS;
		}
	}

	if (uChunkVersion < 11)
	{
		// Make sure widescreen bars flag is not set.
		m_uMenuPageFlags &= ~ASURA_GUIMENU_MENUPAGE_FLAG_RENDER_WIDESCREEN_BARS;
	}

	if( uChunkVersion >= 10)
	{
		xStream >> m_xVirtualSize;
	}
	else
	{
		// NY - 20 November 2008 - only wii menus to deal with at present; ntsc doesn't need virtual size, so just PAL to check for
		if (GetUnscaledLocalBounds().GetWidth() == 640.0f
			&& GetUnscaledLocalBounds().GetHeight() == 528.0f)
		{
			m_xVirtualSize.x = 640.0f;
			m_xVirtualSize.y = 480.0f;
		}
	}

	// -- SK 12.03.07 - This needs to be here because the system allocator isn't used to create menu pages 
	//  when loading from a chunk.
	// if this returns false the add failed, and it may well have deleted us
	if (Asura_GUIMenu_System::AddToMenuList( this))
	{
		Resize(GetUnscaledLocalBounds());//this will populate the scaled world bounds for the menu and all children
	}


	// -- now added to Asura_GUIMenu_Root::ReadFromChunkStream function.
	//if( pxParentObject != NULL)
	//{	pxParentObject->AddToThisContainer( this);	}
}


#ifdef ASURA_TOOLS
void Asura_GUIMenu_Menu_Page::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	// -- write out menu page version --
	//	WriteVersionNumber( ASURA_CHUNK_GUIMENU_MENUPAGE_VERSION, xStream);
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_GUIMENU_MENUPAGE_VERSION, GetType());

	// -- write out Asura_GUIMenu_Root base info --
	Asura_GUIMenu_Root::WriteToChunkStream( xStream);

	// -- output menu specific data --

	// -- write out any children --
	WriteChildObjectData( xStream);

	// -- output the default active widget for this menu page --
	xStream << m_uHashStartupWidget;

	//tools data
	Asura_String xTemp = m_xFilename;
	xStream << xTemp;

	m_xOnInitCommand.WriteToChunkStream( xStream);
	m_xOnDeInitCommand.WriteToChunkStream( xStream);
	m_xOnUpdateCommand.WriteToChunkStream( xStream);

	xStream << m_uMenuPageFlags;

	xStream << m_xVirtualSize;
}

void Asura_GUIMenu_Menu_Page::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<MenuPage_data>\n" );
	pxExporter->IncTabs();

	// -- output menu specific data --

	// -- output the default active widget for this menu page --

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<startup_widget_hash>" );
	pxExporter->AddToBufferHex( m_uHashStartupWidget );
	pxExporter->AddToBuffer( "</startup_widget_hash>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<filename>" );
	pxExporter->AddToBufferEscaped( m_xFilename.GetString() );
	pxExporter->AddToBuffer( "</filename>\n" );


	if( ASURA_CHUNK_GUIMENU_MENUPAGE_VERSION >= 3)
	{
		if(!m_xOnInitCommand.IsBlank())
		{
			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "<on_init>\n" );
			pxExporter->IncTabs();
			m_xOnInitCommand.ExportToXML( pxExporter );
			pxExporter->DecTabs();
			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "</on_init>\n" );
		}
		if(!m_xOnDeInitCommand.IsBlank())
		{
			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "<de_init>\n" );
			pxExporter->IncTabs();
			m_xOnDeInitCommand.ExportToXML( pxExporter );
			pxExporter->DecTabs();
			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "</de_init>\n" );
		}
		if( ASURA_CHUNK_GUIMENU_MENUPAGE_VERSION >= 6)
		{
			if(!m_xOnUpdateCommand.IsBlank())
			{
				pxExporter->AddToBuffer( pxExporter->GetTabs() );
				pxExporter->AddToBuffer( "<on_update>\n" );
				pxExporter->IncTabs();
				m_xOnUpdateCommand.ExportToXML( pxExporter );
				pxExporter->DecTabs();
				pxExporter->AddToBuffer( pxExporter->GetTabs() );
				pxExporter->AddToBuffer( "</on_update>\n" );
			}
		}
	}

	if( ASURA_CHUNK_GUIMENU_MENUPAGE_VERSION >= 4)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<menu_page_flags>" );
		pxExporter->AddToBufferHex( m_uMenuPageFlags );
		pxExporter->AddToBuffer( "</menu_page_flags>\n" );
	}

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<virtual_width>" );
	pxExporter->AddToBuffer( m_xVirtualSize.x );
	pxExporter->AddToBuffer( "</virtual_width>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<virtual_height>" );
	pxExporter->AddToBuffer( m_xVirtualSize.y );
	pxExporter->AddToBuffer( "</virtual_height>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</MenuPage_data>\n" );
}


// --------------------------------------------------------------------------------------- 

const char* Asura_GUIMenu_Menu_Page::GetFilename() const
{
	return m_xFilename.GetString();
}
#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::AutoResize()
{
	//get the menu's raw bounds in virtual space
	Asura_Bounding_Rectangle xBounds;
	GetUnscaledVirtualBounds(xBounds);

	//get the screen bounds in screen space
	Asura_Bounding_Rectangle xScreenBounds;
	if( Asura_GUIMenu_System::GetAutoMenuSizesSet())
	{
		xScreenBounds.SetTopLeft(Asura_Maths::ZeroVector2);
		xScreenBounds.SetBottomRight(Asura_GUIMenu_System::GetAutoMenuSizes());
		xScreenBounds.Displace( Asura_GUIMenu_System::GetAutoMenuOffsets() );
	}
	else
	{
		xScreenBounds.SetTopLeft(Asura_Maths::ZeroVector2);
		xScreenBounds.SetBottomRight(Asura_Vector_2( static_cast<float>( Asura_Viewport::GetWidth() ), static_cast<float>( Asura_Viewport::GetHeight() ) ));
	}

	//convert screen bounds to virtual space
	if (Asura_GUIMenu_System::HasVirtualResolution())
	{
		xScreenBounds.MinX /= Asura_GUIMenu_System::GetVirtualScreenScale().x;
		xScreenBounds.MaxX /= Asura_GUIMenu_System::GetVirtualScreenScale().x;
		xScreenBounds.MinY /= Asura_GUIMenu_System::GetVirtualScreenScale().y;
		xScreenBounds.MaxY /= Asura_GUIMenu_System::GetVirtualScreenScale().y;
	}

	if( CanAutoResize())
	{
		//scale menu to screen bounds
		xBounds = xScreenBounds;
	}
	else if( m_uMenuPageFlags & ASURA_GUIMENU_MENUPAGE_FLAG_CENTER_MENU_ON_SCREEN)
	{
		//centre menu within screen bounds
		xBounds.Displace(xScreenBounds.GetDimension() * 0.5f - xBounds.GetDimension() * 0.5f - ( xBounds.GetTopLeft() - xScreenBounds.GetTopLeft() ));
	}

	// new menu bounds in virtual space
	Resize(xBounds);
}

// --------------------------------------------------------------------------------------- 

// CW - Sending the message alone would not set the active widget on the page, added function to do both.

void Asura_GUIMenu_Menu_Page::DeselectActiveWidget()
{
	Asura_GUIMenu_MessagePump::PostMessage( m_uHashActiveWidget, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT, 0);
	RemoveActiveWidget();
}

void Asura_GUIMenu_Menu_Page::SelectWidget( Asura_Hash_ID uWidgetHash)
{
	DeselectActiveWidget();

	Asura_GUIMenu_Widget_Base* pxWidget = FindWidget(uWidgetHash);
	if(pxWidget)
	{
		Asura_GUIMenu_MessagePump::PostMessage( uWidgetHash, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_IN, 0);
	}
}

// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// -- Protected Functions ---------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::OnInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bHandledMessage = false;
//	Asura_GUIMenu_Root* pxItem;

	// -- 
	if( pxMessage->uToID == GetID())
	{
		// initialise the state to fade in.
		if (m_uState == ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT)
		{
			Asura_GUIMenu_System::EndMenuFadingOut( this );
		}
		if( (GetMenuPageFlags() & ASURA_GUIMENU_MENUPAGE_FLAG_CODE_UPDATE_AND_RENDER_ONLY) == 0)
		{
			Asura_GUIMenu_System::SetMenuPageActiveByID( GetID(), true);
		}
		// Initialise the animatiors. Make sure the animatiors know they are
		// being initialised because the menu page is being initialised.
		RecurseInitialiseAnimators(true);
		m_fStateTimer = Asura_Maths::Min( fASURA_MAX_MENU_TRANSITION_TIME, GetLongestAnimatorDuration( ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_IN) );
		m_fFocusTimer = 0.0f;
		m_uState = ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_IN;

		// -- this call will set the inherited data on this menu page and all contained widgets and elements if necessary.
//		UpdateInheritedValues ( true);
		
		GetOnInitCommand()->Execute( NULL);
		
		// -- init message for menu page, so change message to GUIMENU_ID_ALL, this means the widgets need to be woken up --
		const u_int uContainerCount = GetContainerCount();
		for( u_int u=0; u<uContainerCount; u++)
		{
			if( m_ppxContainerParts[u]->IsWidget() )
			{
				const Asura_Message_GUIMenu_Notify& xMsg =  Asura_GUIMenu_MessagePump::CreateMessage( m_ppxContainerParts[u]->GetID(), GetID(), ASURA_GUIMENU_UI_MSG_ON_INIT, 0);
				m_ppxContainerParts[u]->Whisper(&xMsg);
			}
		}

		if( m_uStartupWidgetOverride != ASURA_HASH_ID_UNSET)
		{
			// the override hash is set so use this instead.
			m_uHashActiveWidget = m_uStartupWidgetOverride;
			m_uStartupWidgetOverride = ASURA_HASH_ID_UNSET;
		}
		else
		{
			m_uHashActiveWidget = m_uHashStartupWidget;
		}


		if( m_uHashActiveWidget != ASURA_HASH_ID_UNSET)
		{
			// set active widget
			m_pxActiveWidget = FindWidget( m_uHashActiveWidget);

			// check it's a widget, clear it if not.
			if (m_pxActiveWidget)
			{
				Asura_Assertf(m_pxActiveWidget->IsWidget(), ("initial widget %s on page %s is not widget type", m_pxActiveWidget->GetNameAsChar(), GetNameAsChar()));
				
				if (m_pxActiveWidget->IsWidget())
				{
					Asura_GUIMenu_Widget_Base* pxWidget = static_cast<Asura_GUIMenu_Widget_Base*>(m_pxActiveWidget);
					if(!pxWidget->IsSelectable())
					{
						RemoveActiveWidget();
					}
				}
				else
				{
					m_pxActiveWidget = NULL;
					m_uHashActiveWidget = ASURA_HASH_ID_UNSET;
				}
			}

			// CJW (03/12/08) - Previously we called FindBestWidget here, even when we had an active widget. 
			// This ensured that we caught the case where the InitialWidget had no links out of it, but others 
			// on the page did. However this caused a large pause on entering widget-heavy pages. As such I 
			// have moved it to only be called if the active widget was un-selectable or did not exist.
			
			if( !m_pxActiveWidget ) 
			{
				// find the best widget
				Asura_GUIMenu_Widget_Base* pxBestWidget = Asura_GUIMenu_Menu_Page::FindBestWidget();

				if (pxBestWidget)
				{
					m_pxActiveWidget = pxBestWidget;
					m_uHashActiveWidget = pxBestWidget->GetID();
				}
			}

			if (m_pxActiveWidget)
			{
				Asura_Assert( m_pxActiveWidget->GetID() == m_uHashActiveWidget, "forgot to change m_uHashActiveWidget");
				
				// whisper roll in message to active widget
				const Asura_Message_GUIMenu_Notify& xMsg = Asura_GUIMenu_MessagePump::CreateMessage( m_uHashActiveWidget, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_IN, 0);
				m_pxActiveWidget->Whisper(&xMsg);
			}
		}

		AutoResize();

		bHandledMessage = true;
	}


	return bHandledMessage;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::OnDeInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	Asura_GUIMenu_Root* pxItem;
	bool bHandledMessage = false;

	// -- 
	if( pxMessage->uToID == GetID())
	{
		// initialise the state to fade out.
		m_fStateTimer = Asura_Maths::Min( fASURA_MAX_MENU_TRANSITION_TIME, GetLongestAnimatorDuration( ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT ) );
		m_uState = ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT;
		Asura_GUIMenu_System::StartMenuFadingOut( this );
		

		GetOnDeInitCommand()->Execute( NULL);

		const u_int uContainerCount = GetContainerCount();
		for( u_int u=0; u<uContainerCount; u++)
		{
			pxItem = m_ppxContainerParts[u];

			if( pxItem && pxItem->IsWidget() )
			{
				const Asura_Message_GUIMenu_Notify& xMsg = Asura_GUIMenu_MessagePump::CreateMessage( pxItem->GetID(), GetID(), ASURA_GUIMENU_UI_MSG_ON_DEINIT, 0);
				pxItem->Whisper(&xMsg);
			}
		}

		Asura_Hash_ID uFocused = Asura_GUIMenu_System::GetFocusedMenuHashID();//store as deactivating will stop this working
		if( (GetMenuPageFlags() & ASURA_GUIMENU_MENUPAGE_FLAG_CODE_UPDATE_AND_RENDER_ONLY) == 0)
		{
			Asura_GUIMenu_System::SetMenuPageActiveByID( GetID(), false);
		}
		
		if(uFocused == GetID())
		{
			// -- this will cause the controlmap to be popped off the control stack --
			Asura_GUIMenu_System::SetFocusedMenuByID( 0);
		}

		bHandledMessage = true;
	}

	return bHandledMessage;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::OnDataRequest( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bHandledMessage = false;

	// -- 
	if( pxMessage->uToID == GetID())
	{
		// -- init message for menu page, so change message to GUIMENU_ID_ALL, this means the widgets need to be woken up --
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_ID_ALL, pxMessage->uFromID, ASURA_GUIMENU_UI_MSG_ON_DATA_REQUEST, 0);

		bHandledMessage = true;
	}

	return bHandledMessage;
}

// --------------------------------------------------------------------------------------- 

// GetLongestAnimatorDuration() is to help decide transition lengths. Ignoring looping animators, 
// find the longest duration of animators in this object matching the given state(s), and recurse 
// into the contained objects.
float Asura_GUIMenu_Menu_Page::GetLongestAnimatorDuration( u_int uStateFlags ) const
{
	float fLongestDuration = PARENT::GetLongestAnimatorDuration( uStateFlags );

	const u_int uContainerCount = GetContainerCount();
	for( u_int u=0; u<uContainerCount; u++)
	{
		fLongestDuration = Asura_Maths::Max( fLongestDuration, m_ppxContainerParts[u]->GetLongestAnimatorDuration( uStateFlags ) );
	}

	return fLongestDuration;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::DoesMenuRenderWidescreenBars( void) const
{
	return ((m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_RENDER_WIDESCREEN_BARS) != 0);
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::CanBackIntoThisMenu( void) const
{
	return ((m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_BACK_INTO_THIS) != 0);
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::CanBackOutOfThisMenu( void) const
{
	return ((m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_BACK_FROM_THIS) != 0);
}

bool Asura_GUIMenu_Menu_Page::CanKeyPressBackOutOfThisMenu( void) const
{
	return ((m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_KEY_PRESS_BACK_FROM_THIS) != 0);
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Menu_Page::ShouldCloseMenuOnBackAction( void) const
{
	return ((m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_CLOSE_THIS_MENU_ON_BACK) != 0);
}


void Asura_GUIMenu_Menu_Page::SetBackButton( Asura_GUIMenu_Widget_Button* pxButton )
{
	Asura_Assert(!m_pxBackButton, "already have a back button set");
	m_pxBackButton = pxButton;
}

void Asura_GUIMenu_Menu_Page::ClearBackButton( Asura_GUIMenu_Widget_Button* pxButton )
{
	Asura_Assert(m_pxBackButton == pxButton, "back button not set correctly");
	m_pxBackButton = NULL;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::GetBackMenuAndWidgetHash( Asura_Hash_ID& uOutMenuHash, Asura_Hash_ID& uOutWidgetHash, bool bGetLocalOnly)
{
	uOutMenuHash = m_uHashBackMenuID;
	uOutWidgetHash = m_uHashBackWidgetID;

	if( bGetLocalOnly == false)
	{
		Asura_GUIMenu_Menu_Page* pxMenuPage = Asura_GUIMenu_System::GetMenuPageByID( m_uHashBackMenuID);
		while( pxMenuPage != NULL)
		{
			if( pxMenuPage->CanBackIntoThisMenu())
			{
				uOutMenuHash = pxMenuPage->GetID();
				break;
			}
			else
			{
				// -- cannot go 'back' into this page, so get the next one in the chain if there is one.
				Asura_Hash_ID uHashMenuTemp, uHashWidgetTemp;
				
				pxMenuPage->GetBackMenuAndWidgetHash( uHashMenuTemp, uHashWidgetTemp);
				// -- clear out the existing 'back' has from this menu as it will not be used.
				// -- if this isn't done it will also assert if we enter this menu again.
				pxMenuPage->ClearBackMenuHash();

				pxMenuPage = Asura_GUIMenu_System::GetMenuPageByID( uHashMenuTemp);

				uOutMenuHash = uHashMenuTemp;
				uOutWidgetHash = uHashWidgetTemp;
			}
		}
	}

//	return uHashBackID;	
}

// --------------------------------------------------------------------------------------- 

Asura_Hash_ID Asura_GUIMenu_Menu_Page::GetBackMenuHash( bool bGetLocalOnly)
{
	Asura_Hash_ID uHashBackID= m_uHashBackMenuID;
	if( bGetLocalOnly == false)
	{
		Asura_GUIMenu_Menu_Page* pxMenuPage = Asura_GUIMenu_System::GetMenuPageByID( m_uHashBackMenuID);
		while( pxMenuPage != NULL)
		{
			if( pxMenuPage->CanBackIntoThisMenu())
			{
				uHashBackID = pxMenuPage->GetID();
				break;
			}
			else
			{
				// -- cannot go 'back' into this page, so get the next one in the chain if there is one.
				Asura_Hash_ID uHashMenuTemp;

				uHashMenuTemp = pxMenuPage->GetBackMenuHash();
				// -- clear out the existing 'back' has from this menu as it will not be used.
				// -- if this isn't done it will also assert if we enter this menu again.
				pxMenuPage->ClearBackMenuHash();

				pxMenuPage = Asura_GUIMenu_System::GetMenuPageByID( uHashMenuTemp);

				uHashBackID = uHashMenuTemp;
			}
		}
	}

	return uHashBackID;	
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::SetBackMenuHash( Asura_Hash_ID uBackMenuHash, Asura_Hash_ID uBackWidgetHash)
{
    // (MH - 26.04.2007) only assert if we have a back menu set and it's not the same one. You sometimes get a
	// situtation where the back hash is not cleared when menus jump around a bit, but as long as it's not 
	// changing there shouldn't be any problems.
	Asura_Assert( m_uHashBackMenuID == ASURA_HASH_ID_UNSET || uBackMenuHash == m_uHashBackMenuID, 
		"Asura_GUIMenu_Menu_Page::SetBackMenuHash - This menu already has a back menu hash set.  Overwriting this value could cause cyclic menu linking problems.");

	m_uHashBackMenuID = uBackMenuHash;
	m_uHashBackWidgetID = uBackWidgetHash;
}

// ---------------------------------------------------------------------------------------

bool Asura_GUIMenu_Menu_Page::CanAutoResize( void)
{
	return ((m_uMenuPageFlags&ASURA_GUIMENU_MENUPAGE_FLAG_ALLOW_AUTO_RESIZE) != 0);
}

// ---------------------------------------------------------------------------------------

#ifdef ASURA_GUIMENU_STORE_NAMES
void Asura_GUIMenu_Menu_Page::ForceGenerateAllSubIDs( void)
{
	GenerateHashID();

	const u_int uContainerCount = GetContainerCount();
	for( u_int i=0; i<uContainerCount; i++)
	{
		m_ppxContainerParts[i]->ForceGenerateAllSubIDs();	
	}

}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Menu_Page::OutputGUIMenuIDInfo( const Asura_String& xOutputFileName)
{
	Asura_GUIMenu_Root* pxParent = GetParentObject();
	Asura_String xActual = "";
	Asura_String xTemp;
	Asura_Hash_ID uHashID;

	while( pxParent != NULL)
	{
		xTemp = xActual;

		xActual = pxParent->GetNameAsChar();
		xActual += "/";

		xActual += xTemp;

		pxParent = pxParent->GetParentObject();
	}

	xActual += GetNameAsChar();//m_strName;

	uHashID = Asura_GetHashID( xActual.GetString());

	Asura_SimpleLogfileOutput(( xOutputFileName.GetString(), "0x%08x - %s\n", uHashID, xActual.GetString()));

	const u_int uContainerCount = GetContainerCount();
	for( u_int i=0; i<uContainerCount; i++)
	{
		m_ppxContainerParts[i]->OutputGUIMenuIDInfo( xOutputFileName);	
	}

}
#endif // ASURA_GUIMENU_STORE_NAMES

// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------ 

Asura_GUIMenu_Widget_Base* Asura_GUIMenu_Menu_Page::FindWidget( const Asura_Hash_ID uHashWidgetID) const
{
	Asura_GUIMenu_Widget_Base* pxFoundWidget = NULL;

	Asura_GUIMenu_Menu_Widget_Match* pxMatchFound = Asura_GUIMenu_System::FastSearchForWidget( uHashWidgetID);
	if( pxMatchFound)
	{
		if( (pxMatchFound->pxMenuPage->GetID() == GetID()) &&
			(pxMatchFound->pxWidgetBase->GetID() == uHashWidgetID))	// second check probably isn't necessary, it's more of a paranoia check.
		{
			pxFoundWidget = pxMatchFound->pxWidgetBase;
		}
	}

	if( pxFoundWidget == NULL)
	{
		for(u_int u = 0; u < GetContainerCount(); ++u)
		{
			if( m_ppxContainerParts[u]->GetID() == uHashWidgetID)
			{
				if( (m_ppxContainerParts[u]->GetType() >= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BEGIN) &&
					(m_ppxContainerParts[u]->GetType() <= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_END))
				{
					pxFoundWidget = static_cast<Asura_GUIMenu_Widget_Base*>(m_ppxContainerParts[u]);
					Asura_GUIMenu_System::AddWidgetToSearchCache( pxFoundWidget, this);
					break;
				}
			}

			Asura_GUIMenu_Root* pxFoundRoot = m_ppxContainerParts[u]->FindContainerItemByID( uHashWidgetID);
			if( pxFoundRoot != NULL)
			{
				if( (pxFoundRoot->GetType() >= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BEGIN) &&
					(pxFoundRoot->GetType() <= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_END))
				{
					pxFoundWidget = static_cast<Asura_GUIMenu_Widget_Base*>(pxFoundRoot);
					Asura_GUIMenu_System::AddWidgetToSearchCache( pxFoundWidget, this);
					break;
				}
			}
			if( pxFoundWidget != NULL)
			{
				break;
			}
		}
	}

	return pxFoundWidget;
}


// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Menu_Page::PostMenuPreviewCleanup()
{
	PARENT::PostMenuPreviewCleanup();

	ClearBackMenuHash();
}
#endif //ASURA_TOOLS

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_ConsoleCmd* Asura_GUIMenu_Menu_Page::GetOnInitCommand() const
{
	return &m_xOnInitCommand;
}

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_ConsoleCmd* Asura_GUIMenu_Menu_Page::GetOnDeInitCommand() const
{
	return &m_xOnDeInitCommand;
}

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_ConsoleCmd* Asura_GUIMenu_Menu_Page::GetOnUpdateCommand() const
{
	return &m_xOnUpdateCommand;
}

// --------------------------------------------------------------------------------------- 


float Asura_GUIMenu_Menu_Page::GetTimeActive() const
{
	if (m_uState == ASURA_GUIMENU_ACTION_FLAG_IS_MENU_ACTIVE)
	{
		return m_fStateTimer;
	}
	else
	{
		return 0.0f;
	}
}

float Asura_GUIMenu_Menu_Page::GetTimeFocused() const
{
	if (m_uState == ASURA_GUIMENU_ACTION_FLAG_IS_MENU_ACTIVE
		&& Asura_GUIMenu_System::GetFocusedMenuHashID() == GetID())
	{
		return m_fFocusTimer;
	}
	else
	{
		return 0.0f;
	}
}

#endif // ASURA_USE_GUI_MENU_SYSTEM
