// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_MessagePump.cpp													//
//																						//
//	Description: Header file for the GUI Messaging system, all internal messages come	//
//				 through here. 															//
//																						//
//	Created: Steve Keen - 17/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "Asura_GUIMenu_MessagePump.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
#include "Core/Asura_Hash.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"

#include "GUIMenu/Asura_GUIMenu_Message.h"

#ifdef ASURA_GUIMENU_ENABLE_METRICS
#include "Core/Asura_Timers.h"
#include "GUIMenu/Asura_GUIMenu_System.h"
#endif // ASURA_GUIMENU_ENABLE_METRICS

// ---------------------------------------------------------------------------------------

Asura_Message_GUIMenu_Notify*	Asura_GUIMenu_MessagePump::m_pucMessage = NULL;
u_int Asura_GUIMenu_MessagePump::m_uMessageIndexToProcess;
u_int Asura_GUIMenu_MessagePump::m_uMessageIndexCanPostTo;

// ---------------------------------------------------------------------------------------

void Asura_GUIMenu_MessagePump::Initialise( void)
{	
	Reset();

	m_pucMessage = new Asura_Message_GUIMenu_Notify[ASURA_GUIMENU_MAX_MESSAGEPUMP_MESSAGES];
}

// ---------------------------------------------------------------------------------------

void Asura_GUIMenu_MessagePump::Reset( void)
{
	m_uMessageIndexToProcess = 0;
	m_uMessageIndexCanPostTo = 0;
}

// ---------------------------------------------------------------------------------------

void Asura_GUIMenu_MessagePump::Shutdown( void)
{
	if( m_pucMessage != NULL)
	{
		delete []m_pucMessage;
		m_pucMessage = NULL;
	}
}

// ---------------------------------------------------------------------------------------

const Asura_Message_GUIMenu_Notify* Asura_GUIMenu_MessagePump::GetNextMessage( void)
{
	Asura_Message_GUIMenu_Notify* pxMessage = NULL;
	
	if( m_uMessageIndexToProcess < m_uMessageIndexCanPostTo)
	{
		pxMessage = &m_pucMessage[m_uMessageIndexToProcess];
		m_uMessageIndexToProcess++;
	}
	else
	{
		// this means that there are no messages to process, so reset the counters
		m_uMessageIndexToProcess = 0;
		m_uMessageIndexCanPostTo = 0;
	}

	return pxMessage;
}

// ---------------------------------------------------------------------------------------

bool Asura_GUIMenu_MessagePump::PostMessage( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bOnQueue = false;

	Asura_Assert( m_pucMessage != NULL, "Asura_GUIMenu_MessagePump::PostMessage - m_pucMessage == NULL");

	Asura_Assert( pxMessage != NULL, "Asura_GUIMenu_MessagePump::PostMessage - const Asura_Message_GUIMenu_Notify* pxMessage == NULL");
	Asura_Assert( m_uMessageIndexCanPostTo < ASURA_GUIMENU_MAX_MESSAGEPUMP_MESSAGES, "Asura_GUIMenu_MessagePump::PostMessage - m_uMessageIndexCanPostTo >= MAX_MESSAGEPUMP_MESSAGES");
	
	if( (m_uMessageIndexCanPostTo < ASURA_GUIMENU_MAX_MESSAGEPUMP_MESSAGES) && (m_pucMessage != NULL))
	{
		m_pucMessage[m_uMessageIndexCanPostTo].uToID = pxMessage->uToID;
		m_pucMessage[m_uMessageIndexCanPostTo].uFromID = pxMessage->uFromID;

		m_pucMessage[m_uMessageIndexCanPostTo].uActionType = pxMessage->uActionType;
		m_pucMessage[m_uMessageIndexCanPostTo].WidgetType = pxMessage->WidgetType;

		m_pucMessage[m_uMessageIndexCanPostTo].uData = pxMessage->uData;
	
		m_uMessageIndexCanPostTo++;
		bOnQueue = true;
	}
#ifdef ASURA_GUIMENU_ENABLE_METRICS
	static float fStatUpdateTimer = 1.0f;

	fStatUpdateTimer -= Asura_Timers::GetFrameTime();
	if( fStatUpdateTimer >= 0.0f)
	{
		if( m_uMessageIndexCanPostTo >= Asura_GUIMenu_System::s_xMetrics.uQueuePeakPerSecond_MessagePump)
		{
			Asura_GUIMenu_System::s_xMetrics.uQueuePeakPerSecond_MessagePump = m_uMessageIndexCanPostTo;
		}
	}
	else
	{
		Asura_GUIMenu_System::s_xMetrics.uQueuePeakPerSecond_MessagePump = 0;
		fStatUpdateTimer = 1.0f;
	}

	if( m_uMessageIndexCanPostTo >= Asura_GUIMenu_System::s_xMetrics.uQueuePeak_MessagePump)
	{
		Asura_GUIMenu_System::s_xMetrics.uQueuePeak_MessagePump = m_uMessageIndexCanPostTo;
	}
#endif // ASURA_GUIMENU_ENABLE_METRICS
	return bOnQueue;
}

// ---------------------------------------------------------------------------------------

bool Asura_GUIMenu_MessagePump::PostMessage( Asura_Hash_ID uTo, Asura_Hash_ID uFrom,
											 u_int uActionType, u_int uWidgetType,
											 u_int uData)
{
	Asura_Message_GUIMenu_Notify xMessage;

	xMessage.uToID = uTo;
	xMessage.uFromID = uFrom;
	xMessage.uActionType = uActionType;
	xMessage.WidgetType = uWidgetType;
	xMessage.uData = uData;
	
	return PostMessage( &xMessage);
}

// ---------------------------------------------------------------------------------------

const Asura_Message_GUIMenu_Notify Asura_GUIMenu_MessagePump::CreateMessage( Asura_Hash_ID uTo, Asura_Hash_ID uFrom,
											u_int uActionType, u_int uWidgetType,
											u_int uData)
{
	Asura_Message_GUIMenu_Notify xMessage;

	xMessage.uToID = uTo;
	xMessage.uFromID = uFrom;
	xMessage.uActionType = uActionType;
	xMessage.WidgetType = uWidgetType;
	xMessage.uData = uData;

	return xMessage;
}

// ---------------------------------------------------------------------------------------

void Asura_GUIMenu_MessagePump::DeleteMessagesForMenu( Asura_Hash_ID uMenuID)
{
	u_int u, v;

	if( m_pucMessage != NULL)
	{
		for( u=0; u<m_uMessageIndexCanPostTo; u++)
		{
			if( m_pucMessage[u].uToID == uMenuID)
			{
				for( v=u; v<m_uMessageIndexCanPostTo-1; v++)
				{
					Asura_Mem::Copy( &m_pucMessage[v], &m_pucMessage[v+1], sizeof(Asura_Message_GUIMenu_Notify));
					Asura_Mem::Set( &m_pucMessage[v+1], 0, sizeof(Asura_Message_GUIMenu_Notify));
					if( m_uMessageIndexCanPostTo < m_uMessageIndexToProcess)
					{
						m_uMessageIndexToProcess = m_uMessageIndexCanPostTo;
					}
				}

				m_uMessageIndexCanPostTo--;
				u=0;
			}
		}
	}
}
#endif // ASURA_USE_GUI_MENU_SYSTEM
