// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Root.h															//
//																						//
//	Description: Root class for all widgets, elements and menus							//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_GUIMENU_ROOT_H__
#define __ASURA_GUIMENU_ROOT_H__

#ifdef ASURA_USE_GUI_MENU_SYSTEM

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_List.h"
#include "Core/Asura_String.h"
#include "GUIMenu/Asura_GUIMenu_CommandConsole.h"
#include "GUIMenu/Asura_GUIMenu_Modifier.h"
#include "Network/Asura_Message.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_RenderStates.h"

// -- Forward declarations ------------------------------------------------------------
class Asura_GUIMenu_Animator;
class Asura_GUIMenu_Menu_Page;
class Asura_GUIMenu_Widget_Base;
struct Asura_GUIMenu_UV_Set;
class Tools_GUIMenu_XMLExporter;
class Asura_Subwindow_Dialog;

// -- Typedef -------------------------------------------------------------------------

// Only store text names in the GUI elements in tools, assert or debug builds
#if defined(ASURA_TOOLS) || defined(ASURA_ASSERT) || defined(_DEBUG)
	#define ASURA_GUIMENU_STORE_NAMES (1)
#endif

// -- Enums ---------------------------------------------------------------------------

enum Asura_GUIMenu_Data_Message_Type
{
	ASURA_GUIMENU_UI_MSG_NO_TYPE = 0,
	ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE,
	ASURA_GUIMENU_UI_MSG_ON_DATA_REQUEST,
	ASURA_GUIMENU_UI_MSG_ON_CLICKED,		// data is now passed to this message, it's an index of which accept type was triggered.
											// 0 - ACCEPT, 1 - ALT_ACCEPT1, 2 - ALT_ACCEPT2, 3 - ALT_ACCEPT3
	ASURA_GUIMENU_UI_MSG_ON_ROLL_IN,
	ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT,
	ASURA_GUIMENU_UI_MSG_ENABLE_DISABLE,
	ASURA_GUIMENU_UI_MSG_ON_INIT,
	ASURA_GUIMENU_UI_MSG_ON_DEINIT
};

// -- Defines -------------------------------------------------------------------------
#define ASURA_GUIMENU_ACTION_NONE	0

// These flags are saved out to chunk data, so changing existing ones would be bad.
// Some of the action flags should be renamed to ASURA_GUIMENU_ROOT_FLAG_RENDER_WHEN_XXX
// and others to ASURA_GUIMENU_ROOT_FLAG_NOTIFY_XXX. the ones with ACTION in them are 
// properties. Others are runtime flags.
// The action flags are (re)used as widget states, and the flags are used to indicate 
// which states elements should be displayed in. Also modifiers and animators use the 
// flags to see when they should be active. There are also now some menu page states, 
// but they are currently only used for animators (i.e. elements and modifiers are 
// active in all menu states).
#define ASURA_GUIMENU_ACTION_FLAG_IS_DISABLED						(1<<0)	// property
#define ASURA_GUIMENU_ACTION_FLAG_IS_DEFAULT						(1<<1)	// property: normal unhighlighted item
#define ASURA_GUIMENU_ACTION_FLAG_IS_CLICKED						(1<<2)	// property
#define ASURA_GUIMENU_ACTION_FLAG_IS_ROLL_ON						(1<<3)	// property
#define ASURA_GUIMENU_ACTION_FLAG_IS_ROLL_OFF						(1<<4)	// property
#define ASURA_GUIMENU_ACTION_FLAG_IS_HILIGHTED						(1<<5)	// property: highlighted item
#define ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_ON_EXIT		(1<<6)	// property - not fully implemented
#define ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE		(1<<7)	// property - not fully implemented
#define ASURA_GUIMENU_ROOT_FLAG_RENDERING_DISABLED					(1<<8)	// runtime: not rendered
#define ASURA_GUIMENU_ROOT_FLAG_IS_HIDDEN_BY_COND_VAR				(1<<9)	// runtime: has a condition variable which caused item to be hidden.
#define ASURA_GUIMENU_ACTION_FLAG_IS_EDITING						(1<<10)	// property: in edit mode
#define ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_IN					(1<<11)	// property: menu is fading in
#define ASURA_GUIMENU_ACTION_FLAG_IS_MENU_FADE_OUT					(1<<12)	// property: menu is fading out
#define ASURA_GUIMENU_ACTION_FLAG_IS_MENU_ACTIVE					(1<<13)	// property: menu is active
#define ASURA_GUIMENU_ROOT_FLAG_USE_RELATIVE_POSITIONING			(1<<14) // property: determines whether positions are fixed or use 0-1 (actual position is calculated based on the menu page size)
#define ASURA_GUIMENU_ROOT_FLAG_USE_ANCHOR_POSITIONING				(1<<15)	// property: can only be used with fixed positioning.  When a menu page is resized, using this flag will keep the object a set pixel distance from the edges specified.
// These flags are saved out to chunk data, so changing existing ones would be bad.

// ASURA_GUIMENU_ACTION_FLAG_ALL is used for the flags corresponding to widget states
#define ASURA_GUIMENU_ACTION_FLAG_ALL	(ASURA_GUIMENU_ACTION_FLAG_IS_DISABLED | ASURA_GUIMENU_ACTION_FLAG_IS_DEFAULT | ASURA_GUIMENU_ACTION_FLAG_IS_CLICKED | ASURA_GUIMENU_ACTION_FLAG_IS_ROLL_ON | ASURA_GUIMENU_ACTION_FLAG_IS_ROLL_OFF | ASURA_GUIMENU_ACTION_FLAG_IS_HILIGHTED | ASURA_GUIMENU_ACTION_FLAG_IS_EDITING)

// flags that shouldn't be saved in chunk data (pretty much any runtime flag), plus obsoleted ones
#define ASURA_GUIMENU_ROOT_FLAGS_NOTSAVED (ASURA_GUIMENU_ROOT_FLAG_RENDERING_DISABLED | ASURA_GUIMENU_ROOT_FLAG_IS_HIDDEN_BY_COND_VAR | ASURA_GUIMENU_ROOT_FLAG_USE_RELATIVE_POSITIONING | ASURA_GUIMENU_ROOT_FLAG_USE_ANCHOR_POSITIONING)

#define ASURA_GUIMENU_TOOLS_FLAG_ISSTYLE				(1<<0) //Tells the tools mode that this menu item is a style , and so should be added to the list of placeable styles.
#define ASURA_GUIMENU_TOOLS_FLAG_GENERATEDBYSTYLE		(1<<1) //Marks automatically children of style instances
#define ASURA_GUIMENU_TOOLS_FLAG_STYLEINSTANCE			(1<<2) //Marks the root item in an instance of a style
#define ASURA_GUIMENU_TOOLS_FLAG_STYLECONTAINER			(1<<3) //Used for marking menus that contain the master copies of styles , which means that they shouldn't be included in exports.
#define ASURA_GUIMENU_TOOLS_FLAG_LOCKUVTODIMS			(1<<4) //Keeps UV coordinates synced with the width/height while editing. Only relevant to image elements , but hopefully there should be more than enough tools flag space.
#define ASURA_GUIMENU_TOOLS_FLAG_TILEDTEXTURE			(1<<5) // If set, indicates that an image element's m_uHashTextureName reffers to a tile, and not a texture
#define ASURA_GUIMENU_TOOLS_FLAG_ISUNCLAMPED			(1 << 6) // If set then this object is constrained to the bounding box of it's parent
#define ASURA_GUIMENU_TOOLS_FLAG_AUTOSHRINK				(1 << 7) // If set then this object will shrink to fit it's children
#define ASURA_GUIMENU_TOOLS_FLAG_GROWONCHILDMOVE		(1 << 8) // If set then this object can grow to allow children to leave it's old borders
#define ASURA_GUIMENU_TOOLS_FLAG_EXPORTRESOURCES		(1<<9)	// If set, texture or tiled texture files refferenced by this element will be added to the GUI menu file

// -- Local messaging id's
#define ASURA_GUIMENU_ID_ALL			(ASURA_HASH_ID_UNSET)

// -- these can be the same, it's just for clarity in code :o)
#define ASURA_GUIMENU_EXTERNAL_REQUEST (0xffffffff)
#define ASURA_GUIMENU_EXTERNAL_NOTIFY	 (0xffffffff)

//#define ASURA_GUIMENU_MAX_CONTAINER_PARTS	32
#define ASURA_GUIMENU_CONTAINER_ADDITION_GRADIENT	5

#define ASURA_GUIMENU_INIT_TOOL_CONTAINER_PARTS	5


#define ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS	2

#define ASURA_GUIMENU_ROOT_LOCAL_FLAG_START	0

// -- Data Chunk Version --
// Versions: ASURA_CHUNK_ROOT_VERSION
// 0:	Initial data - SK 5.12.06
// 1:	Changed display threshold to allow multiple(2) conditions. - SK
// 2:	Added Inheritance flags and hash id - SK 15.02.07
// 3:	Added tools flags and tools style flags - RAR 26/02/2007
// 4:	Added Style flags and hash id - SK 05.03.07
// 5:	Added condition var
// 6:   Added modifier data
// 7:   Bumped up version number to trap any versions of data that are less than No.7
//		This is to make sure that if the hilighted flag is set for this object, the edit flag should be set on loading too.
//		This is to fix instances of data where the edit flag wasn't available before this version.
// 8:	Added animator data
// 9:   Removed inheritance data
// 10:	Allowed for multiple condition vars. RAR 21/05/2007
// 11:  Version number increase to fix errorneously duplicated animators. RAR 29/06/2007
// 12:	Added comment & username strings, only to be outputted as a save and NOT exported (exported version is 2 x u_int to represent string terminators only). SK 21/11/07
// 13:  Changed the comments system to allow for multiple entries.
// 14:  Added the anchor data
// 15:  Revised the anchor data
// 16:	Moved scaling flags from element into root - NDB 13/05/08
// 17:  Changed meaning of scaling flags, adding position x|y scaling to parent x|y flags. No new data - NDB 27/05/08
// 18:  Reduced size of various data members in an attempt to reduce memory consumption. RDM 06/06/08
// 19:  Added flags for indicating which platforms the menu item is for. RAR 23/06/2008
// 20:  moved all size data up to root, and tidied
#define ASURA_CHUNK_ROOT_VERSION 20


// Versions: ASURA_CHUNK_CONDITION_VAR_VERSION
// 0:	Initial data
// 1:   Addition of console cmd to condition var
#define ASURA_CHUNK_CONDITION_VAR_VERSION 1

// 0:	Initial data
#define ASURA_CHUNK_CONDITION_VAR_MANAGER_VERSION 0



// -- Structures ----------------------------------------------------------------------
struct Asura_Message_GUIMenu_Notify;


// -- Classes -------------------------------------------------------------------------


//structure for ProcessInputActivity()
class Asura_GUIMenu_Input
{
public :
	Asura_GUIMenu_Input()
	: m_uButtonPresses(0)
	, m_xCursorPos(0.0f,0.0f)
	, m_bUseCursor(false)
	{}

	u_int m_uButtonPresses;
	Asura_Vector_2 m_xCursorPos;
	bool m_bUseCursor;
};

// class to handle comment data.
class Asura_GUIMenu_ToolComment
{
public:
	Asura_GUIMenu_ToolComment()	{}
	~Asura_GUIMenu_ToolComment() {}

	void ReadFromChunkStream( Asura_Chunk_Stream& xStream);

#ifdef ASURA_TOOLS
	void Set( const Asura_Char* pszName, const Asura_Char* pszComment);
	void Set( const Asura_Char* pszName, const Asura_Char* pszDate, const Asura_Char* pszComment);

	void AddOverlayControls( Asura_Subwindow_Dialog* pxSubWindow, Asura_Vector_2& xOffset);


	const Asura_Char*	GetDateString()	{	return m_xDate.GetString();	}
	void				SetDateString( const Asura_Char* pszNewDate)	{	m_xDate = pszNewDate;	}

	const Asura_Char*	GetNameString()	{	return m_xName.GetString();	}
	void				SetNameString( const Asura_Char* pszNewName)	{	m_xName = pszNewName;	}

	const Asura_Char*	GetCommentString()	{	return m_xComment.GetString();	}
	void				SetCommentString( const Asura_Char* pszNewComment)	{	m_xComment = pszNewComment;	}

	bool				IsValidComment()	{	return !m_xComment.IsBlank();	}

	void				WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
private:

	Asura_String m_xDate;
	Asura_String m_xName;
	Asura_String m_xComment;

#endif // ASURA_TOOLS
};

class Asura_GUIMenu_Root
{
#ifdef ASURA_TOOLS
	friend class Tools_GUIMenu_Element_Base;
	friend class Tools_GUIMenu_Root;
	friend class Tools_GUIMenu_Widget_Base;
	friend class Tools_GUIMenu_Widget_Group;
#endif

	// non-copiable - private and unimplemented copy ctor and operator=
	Asura_GUIMenu_Root(const Asura_GUIMenu_Root&);
	Asura_GUIMenu_Root& operator=(const Asura_GUIMenu_Root&);

// -- Functions -----------------------------------------------------------------------
public:

	enum Asura_GUIMenu_Object_Type
	{
		// HE - 18 May 2007 - Please keep Tools_GUIMenu_XMLExporter::MapObjectTypesToName() updated if you're adding or removing new elements/widgets!
		ASURA_GUIMENU_TYPEID_NONE = 0,
		ASURA_GUIMENU_TYPEID_MENUPAGE = 1,
		
		ASURA_GUIMENU_TYPEID_ELEMENT_BEGIN			= 1000,
		ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE			= 1001,
		ASURA_GUIMENU_TYPEID_ELEMENT_TEXT			= 1002,
		ASURA_GUIMENU_TYPEID_ELEMENT_PATH			= 1003,
		ASURA_GUIMENU_TYPEID_ELEMENT_ANIMATION		= 1004,
		ASURA_GUIMENU_TYPEID_ELEMENT_STATUSTEXT		= 1005,
		ASURA_GUIMENU_TYPEID_ELEMENT_FMV			= 1006,
		ASURA_GUIMENU_TYPEID_ELEMENT_END			= 1999,

		ASURA_GUIMENU_TYPEID_WIDGET_BEGIN			= 2000,
		ASURA_GUIMENU_TYPEID_WIDGET_BUTTON			= 2001,
		ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX		= 2002,
		ASURA_GUIMENU_TYPEID_WIDGET_SLIDER			= 2003,
		ASURA_GUIMENU_TYPEID_WIDGET_TEXTBOX			= 2004,
		ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX		= 2005,
		ASURA_GUIMENU_TYPEID_WIDGET_GROUP			= 2006,
		ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE	= 2007,
		ASURA_GUIMENU_TYPEID_WIDGET_PROGRESSBAR		= 2008,
		ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC			= 2009,
		ASURA_GUIMENU_TYPEID_WIDGET_LISTBOX			= 2010,
		ASURA_GUIMENU_TYPEID_WIDGET_LISTBOXENTRY	= 2011,
		ASURA_GUIMENU_TYPEID_WIDGET_CAROUSEL		= 2012,
		ASURA_GUIMENU_TYPEID_WIDGET_END				= 2999,

		ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_WIDGET_BEGIN	= 3000,
		ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_WIDGET_END	= 3999,
		ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_ELEMENT_BEGIN = 4000,
		ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_ELEMENT_END	= 4999,

		ASURA_GUIMENU_CHUNK_TYPE_ROOT				= 10001,
		ASURA_GUIMENU_CHUNK_TYPE_ELEMENT_BASE		= 10002,
		ASURA_GUIMENU_CHUNK_TYPE_WIDGET_BASE		= 10003,
	};

	enum Asura_GUIMenu_Inherit_Flags_Root_Type
	{
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_POSITION				= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+0)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_COLOUR				= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+1)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_DISPLAY_CONDITIONS	= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+2)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_TRANSLUCENCY			= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+3)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_FLAGS					= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+4)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_ANCHOR				= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+5)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_SCALING				= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+6)),
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_PLATFORMS				= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+7)),

		//ASURA_GUIMENU_ROOT_OVERIDE_MAX is here to stop another programmer adding past bit flag 7
		ASURA_GUIMENU_ROOT_LOCAL_FLAG_MAX					= (1<<(ASURA_GUIMENU_ROOT_LOCAL_FLAG_START+8))
	};

	// anchoring flag definitions
	enum Asura_GUIMenu_Anchoring_Flag
	{
		POS_NORMAL = 0,//left/top
		POS_CENTRE,
		POS_FAR,//right/bottom
		POS_RELATIVE,//X for x, y for y
		POS_RELATIVE_ALTAXIS,// y for x, x for y

		NUM_POS_MODES
	};

	// scaling flag definitions
	enum Asura_GUIMenu_Scaling_Flag
	{
		ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX	= (1<<0),
		ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTY	= (1<<1),
		ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTX	= (1<<2),
		ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY	= (1<<3),
	};

	Asura_GUIMenu_Root();
	virtual ~Asura_GUIMenu_Root();

	virtual void				Render( u_int uState) = 0;
	virtual void				Initialise() { RecurseInitialiseAnimators(); }
	virtual void				Update( float fElapsedTime);
	virtual void				Copy( const Asura_GUIMenu_Root* pxSource );
	virtual u_int				ProcessInputActivity( const Asura_GUIMenu_Input& xInput) { return 0; }
	virtual bool				HandleMessage( const Asura_Message_GUIMenu_Notify* pxMessage)	{	return false;	}

	void ApplyDisplayConditionSettings();

	u_int						GetType() const;
	bool						IsWidget() const;		// not very OO, but useful for debugging.

	//get/set unscaled position of top left corner relative to top left of parent
	Asura_Vector_2				GetUnscaledLocalPos() const; 
	void 						SetUnscaledLocalPos( const Asura_Vector_2& xPos );

	//get world (screenspace/render) coordinate of top left corner
	Asura_Vector_2				GetScaledWorldPos() const;
	//vector to offset children by
	virtual Asura_Vector_2		GetChildPosOffset() const { return Asura_Maths::ZeroVector2;}

	//get world (screenspace/render) rectangle
	virtual const Asura_Bounding_Rectangle& GetScaledWorldBounds() const { return m_xScaledWorldBounds; }

	//get/set unscaled rectangle relative to top left of parent
	virtual void				SetUnscaledLocalBounds( const Asura_Bounding_Rectangle& xBounds );
	const Asura_Bounding_Rectangle&	GetUnscaledLocalBounds() const { return m_xUnscaledLocalBounds; }

	//set world (screenspace/render) rectangle, and recursively update children accordingly. coordinates in virtual resolution if in use
	virtual void				Resize( const Asura_Bounding_Rectangle& xBounds );

	//set world (screenspace/render) rectangle's top left corner, and recursively update children accordingly
	virtual void				Reposition( const Asura_Vector_2& xPosition ) { Translate( xPosition - GetScaledWorldPos() ); }
	void						Reposition( const float fX,const float fY )	  { Reposition( Asura_Vector_2(fX,fY) ); }

	// move world (screenspace/render) rectangle's top left corner, and recursively update children accordingly
	virtual void				Translate( const Asura_Vector_2& xOffset );

	//set world (screenspace/render) rectangle, and recursively update children accordingly. coordinates in true screen space
	void				SetScaledWorldBounds( const Asura_Bounding_Rectangle& xBounds );
	
	// Called in Resize(). Can be overriden to allow resizing of child elements not contained in Root
	virtual Asura_GUIMenu_Root* GetResizeChild( const u_int u ) { return GetContainerItem(u); }
	virtual u_int GetResizeChildCount() const { return GetContainerCount(); }

	//data generation/legacy fixup function; stretches bounds to include child elements
	void						FixupBoundsFromChildren();

	//Does a recursive search for a selectable widget at the given location
	virtual Asura_GUIMenu_Widget_Base*  FindSelectableWidgetAtPos(const Asura_Vector_2& xPos);

	//Checks for the specified position being contained in this or any child item.
	//RAR 01/08/2007
	virtual bool				ContainsPos(const Asura_Vector_2& xPos) const;

	virtual bool				GetCropRectForChildren(Asura_Bounding_Rectangle& xBounds) const;

	// -- Colour functions --
	const Asura_RGBA&			GetColour() const;

	void 						SetColour( float fRed, float fGreen, float fBlue, float fAlpha);
	void 						SetColour( const Asura_RGBA& xColour);
	void 						SetColour( const Asura_RGBA* pxColour);
	void						SetColour( u_int uRGBAColourValue);
	void						SetColourFromRGBA( u_int uRGBAColourValue);
	void						SetColourFromARGB( u_int uARGBColourValue);

	void						SetParentObject( Asura_GUIMenu_Root* pxParent)	{	m_pxParent = pxParent;	}
	Asura_GUIMenu_Root*			GetParentObject() const	{	return m_pxParent;	}
	virtual Asura_GUIMenu_Root*	GetParentObjectForIDGeneration() const	{	return m_pxParent;	}
	Asura_GUIMenu_Menu_Page*	FindOwnerMenu() const;	// returns parent menu (or NULL).

	// -- Flag Functions --
	void						ChangeFlag( u_int uNewFlag, bool bTF);	// change local (not inherited flag)
	bool						GetFlag( u_int uFlag) const;
	void						ToggleFlag( u_int uFlag);
	u_int						GetFlags() const;	// may return inherited object flags
	u_int*						GetFlagsPtr()	{	return &m_uFlags;	}
	void						SetFlags( u_int uFlags) {	m_uFlags = uFlags;	}
	
	// Following accessors are for runtime flags, so they operate on local (not inherited flags)
	#ifdef ASURA_TOOLS
	bool						IsRenderingDisabled() const		{ return (m_uFlags & ASURA_GUIMENU_ROOT_FLAG_RENDERING_DISABLED) || !IsFlaggedForCurrentPlatform();}
	#else
	bool						IsRenderingDisabled() const		{ return 0 != (m_uFlags & ASURA_GUIMENU_ROOT_FLAG_RENDERING_DISABLED);	}
	#endif
	void						SetRenderingDisabledFlag( bool bTF)		{ ChangeFlag(ASURA_GUIMENU_ROOT_FLAG_RENDERING_DISABLED, bTF);	}
	bool						IsHiddenByConditionVar()	const	{ return 0 != (m_uFlags & ASURA_GUIMENU_ROOT_FLAG_IS_HIDDEN_BY_COND_VAR);	}
	virtual void				SetHiddenByConditionVar( bool bTF)		{ ChangeFlag(ASURA_GUIMENU_ROOT_FLAG_IS_HIDDEN_BY_COND_VAR, bTF);	}

	virtual void				ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion);
	u_int						ReadVersionNumber( Asura_Chunk_Stream& xStream);
	u_int						ReadLocalChunkHeader( Asura_Chunk_Stream& xStream, u_int& uType);
	bool						CheckVersionNumber( const u_int uChunkVersion, const Asura_GUIMenu_Object_Type eType,const Asura_Handle iFileHandle = ASURA_INVALID_HANDLE );
	bool						ReadAllChildObjects( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParent);
	bool						ReadChildObject( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParent);
	#ifdef ASURA_TOOLS
	void						SetToolsFlags(u_int uFlags) {m_uToolsFlags = uFlags;}
	u_int						GetToolsFlags() const {return m_uToolsFlags;}
	#endif



#ifdef ASURA_TOOLS
	Asura_String*				GetCommentString()	{	return &m_xComments;	}
	void						SetCommentString( const Asura_Char* pcComment)	{	m_xComments = pcComment;	}

	const Asura_Char*			GetCommentUserName()	{	return m_xCommentUser.GetString();	}
	void						SetCommentUserName( const Asura_Char* pcCommentUser)	{	m_xCommentUser = pcCommentUser;	}

	virtual void 				UpdateLocalDataFromStyle();

	//Calls the  callback for each texture used by this and any child items. Used for determining textures to export etc
	virtual	void				TraverseUsedTextures(void (*pfnCallback)(int iTextureID,void* pUserData),void* pUserData);
	virtual void				AddFilesToExportList();

	virtual	void				ListUsedTextPages(Asura_HashTree& xTextPageTree) const;


	// Inheritance tool helper functions --
	virtual void				WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	void 						WriteVersionNumber( u_int uVersion, Asura_Chunk_Stream& xStream) const;
	void						WriteLocalChunkHeader( Asura_Chunk_Stream& xStream, u_int uVersion, u_int uType) const;
	void						WriteChildObjectData( Asura_Chunk_Stream& xStream) const;

	// XML export functions --
	void						ExportToXMLStart(Tools_GUIMenu_XMLExporter* pxExporter) const;	//the start point for an XML export
	virtual void				ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const; //exports an object's properties
	void						WriteChildXMLData(Tools_GUIMenu_XMLExporter* pxExporter) const;
	void						WriteXMLOpenTag(Tools_GUIMenu_XMLExporter* pxExporter) const;
	void						WriteXMLCloseTag(Tools_GUIMenu_XMLExporter* pxExporter) const;

#endif // ASURA_TOOLS
	bool						IsActionEnabled( u_int uFlags);

	inline Asura_Hash_ID		GetID() const	{	return m_uHashID;	}
	void						SetID( Asura_Hash_ID uHashID );
	virtual void				ChildIDChanged( Asura_Hash_ID uOldID, Asura_Hash_ID uNewID ) {};

	// -- Name of GUI part --
	virtual void				SetName(const Asura_String& xName);
	virtual void				SetName( const char* szName);
	Asura_Hash_ID				GetNameHash() const { return m_uHashName; }
	void						SetNameHash( const Asura_Hash_ID uHash ) { m_uHashName = uHash; }
	void						GenerateHashID();
#ifdef ASURA_GUIMENU_STORE_NAMES
	virtual void				ForceGenerateAllSubIDs();

	const Asura_Char*			GetNameAsChar() const	{	return m_xName.GetString();	}
	const Asura_String*			GetNamePointer() const	{	return &m_xName;	}
	bool						HasName()	{	return (!m_xName.IsBlank() && !m_xName.IsNull());	}	

	// return the string used to generate the items hash id. This includes all parent item names
	const Asura_String			GenerateFullyQualifiedName() const;

	virtual void				OutputGUIMenuIDInfo( const Asura_String& xOutputFileName);
#endif // ASURA_GUIMENU_STORE_NAMES

	// message handling stuff
	virtual bool				Whisper( const Asura_Message_GUIMenu_Notify* pxNotify) {	return false;	}
	virtual void				TextInputEnded();

	bool						IsDisplayConditionMet( const int iValidate);
	bool						SetInternalDisplayCondition( const int iThreshold, const Asura_GUIMenu_Display_Condition eCondition);
	bool						AddInternalDisplayCondition( const int iThreshold, const Asura_GUIMenu_Display_Condition eCondition);
	bool						ReplaceInternalDisplayCondition( const u_int uIndex, const int iThreshold, const Asura_GUIMenu_Display_Condition eCondition);
	void						CopyInternalDisplayConditions( const Asura_GUIMenu_Root& xSource );

	void						RemoveAllInternalDisplayConditions();//Clears current list of display conditions

	int							GetInternalDisplayConditionThreshold() const { return GetInternalDisplayConditionThreshold( 0);	}
	Asura_GUIMenu_Display_Condition	GetInternalDisplayConditionCompare() const { return GetInternalDisplayConditionCompare( 0);	}

	int												GetInternalDisplayConditionThreshold( const u_int uIndex) const;
	Asura_GUIMenu_Display_Condition					GetInternalDisplayConditionCompare( const u_int uIndex) const;
	const Asura_GUIMenu_Display_Condition_Params*	GetInternalDisplayCondition( const u_int uIndex) const;
	u_int											GetInternalDisplayConditionCount() const;//	{ return m_uInternalDisplayConditionCount;	}

#ifdef ASURA_TOOLS
	// sets any console var name matching the 1st arg to the 2nd arg
	virtual const bool			FindAndReplaceConsoleVarName( const Asura_UnicodeChar* wszName,const Asura_UnicodeChar* wszReplacementName );
#endif // ASURA_TOOLS

	Asura_GUIMenu_Root*			FindContainerObject( const Asura_Hash_ID uHashID) const;

	virtual bool				AddToThisContainer( Asura_GUIMenu_Root* pxObject);
	u_short						GetContainerCount() const	{	return m_usContainerPartsCount;	}
	Asura_GUIMenu_Root*			GetContainerItem( u_int uIndex) const;
	virtual bool				RemoveFromThisContainer( Asura_GUIMenu_Root* pxObject);
	virtual bool				IsContainerFull();
	void						CheckAndExpandContainerIfNecessary();
	virtual bool				IsContainerEmpty();

	Asura_GUIMenu_Root*			FindContainerItemByID( Asura_Hash_ID uHashID) const;
	Asura_GUIMenu_Root*			FindContainerItemByLocalNameHash( Asura_Hash_ID uHashLocalName);

#ifdef ASURA_TOOLS
	Asura_GUIMenu_Root*			FindContainerItemByLocalName( const Asura_String& xItemName);
#endif // ASURA_TOOLS	

	virtual ASURA_TRANSLUCENCY_MODE		GetTranslucencyMode() const;
	virtual void 				SetTranslucencyMode( ASURA_TRANSLUCENCY_MODE eTransMode);

	#ifdef ASURA_TOOLS
	inline u_int				GetStyleUseLocalDataFlags() const	{ return m_uStyleUseLocalDataFlags;	}
	inline void					SetStyleUseLocalDataFlags( u_int uFlags)	{	m_uStyleUseLocalDataFlags = uFlags;	}
	void						ChangeStyleUseLocalDataFlag( u_int uNewFlag, bool bTF);
	bool						GetStyleUseLocalDataFlag( u_int uFlag) const;

	inline Asura_Hash_ID		GetStyleObjectHash() const { return m_uHashStyleObject;	}
	void						SetStyleObjectHash( Asura_Hash_ID uHashObjectID);
	void						SetStyleObject(Asura_GUIMenu_Root* pxStyleObject);

	inline u_int				GetStyleChildEditableDataFlags() const{ return m_uStyleChildEditableFlags;	}
	inline void					SetStyleChildEditableDataFlags( u_int uFlags){ m_uStyleChildEditableFlags = uFlags;	}
	#endif

//	void						SetConditionVar( const Asura_GUIMenu_ConditionVar& xCondVar ) { m_xDisplayConditionVar = xCondVar; }

	#ifdef ASURA_TOOLS
	//Resets visibility etc after finishing with the preview mode in tools.
	virtual void				PostMenuPreviewCleanup();
	#endif

	// functions for modifiers
	int							CreateModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType );
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, u_int uData);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, int iData);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, float fData);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_RGBA& xColourData);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_Vector_2& xVector2);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_UnicodeChar* pwszUnicodeString);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_Char* pcCharString);
	bool						AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_GUIMenu_UV_Set& xUVSet);
	
	bool						CopyModifier( Asura_GUIMenu_Modifier* pxModifier);

#ifdef ASURA_TOOLS
	bool						DeleteModifier( u_int uListIndex);	
#endif

	u_char						GetModifierCount() const;
	Asura_GUIMenu_Modifier*		GetModifier( u_int uIndex) const;
	void						CopyModifiers( const Asura_GUIMenu_Root& xSource );

	// functions for animators
	u_char						GetAnimatorCount( void ) const { return m_ucAnimatorCount; }
	Asura_GUIMenu_Animator*		GetAnimator( u_int uIndex ) const;

	//find an animator using a given sequence
	const Asura_GUIMenu_Animator* GetAnimatorBySequenceHash(Asura_Hash_ID uHash) const;
	Asura_GUIMenu_Animator* GetAnimatorBySequenceHash(Asura_Hash_ID uHash);
	
	// GetLongestAnimatorDuration() is to help decide transition lengths. Ignoring looping animators, 
	// find the longest duration of animators in this object matching the given state(s), and recurse 
	// into the contained objects.
	virtual float				GetLongestAnimatorDuration( u_int uStateFlags ) const;
	
	//Resets the timers for all animators for this and all children
	void RecurseInitialiseAnimators(bool bInitMenuPage = false);

#ifdef ASURA_TOOLS
	Asura_GUIMenu_Animator*		AddAnimator( const char* szSequenceName );
	void						DeleteAnimator( u_int uListIndex );	

	Asura_List<Asura_GUIMenu_ToolComment>*	GetCommentListPtr()	{	return &m_xCommentList;	}
	void						AddComment( const Asura_Char* pszName, const Asura_Char* pszComment);
	u_int						GetCommentCount()	{	return m_xCommentList.GetNumEntries(); }
#endif

	Asura_GUIMenu_Animator*		AddAnimator( const Asura_GUIMenu_Animator& xAnimator );
	void						CopyAnimators( const Asura_GUIMenu_Root& xSource );


	void SetPosMode(const u_int uAxis,const Asura_GUIMenu_Anchoring_Flag eFlag ) { m_aePosMode[uAxis] = eFlag; }
	Asura_GUIMenu_Anchoring_Flag GetPosMode(const u_int uAxis) const	{ return m_aePosMode[uAxis]; }
	const Asura_Vector_2& GetPosOrigin() const { return m_xPosOrigin; }

	// takes the dimension scale factor that has been applied to the parent and modifiys it 
	// (accounting for scale width|height to parent width|height flags etc) so that it can be 
	// applied to this object.
	virtual Asura_Vector_2 GetRelativeDimensionScale( const Asura_Vector_2& xParentScale ) const;

	virtual void ApplyScalingFlagsToChildren();
	virtual void ApplyAnchoringFlagsToChildren();

	bool						GetScalingFlag( const Asura_GUIMenu_Scaling_Flag uFlag) const;
	void						SetScalingFlag( const Asura_GUIMenu_Scaling_Flag uFlagsToSet, const bool bFlagSet);
	u_int						GetScalingFlags() const					{ return m_uScalingFlags; }
	void						SetScalingFlags( const u_int uFlags)	{ m_uScalingFlags = uFlags; }
	u_int*						GetScalingFlagsPtr()					{ return &m_uScalingFlags; }

	void						RenderDebugBounds( void);

#ifdef ASURA_TOOLS
	bool ShouldExportToPlatform(int iPlatform) const {return (m_uPlatforms & (1<<iPlatform)) ? true : false;}
	void SetPlatforms(u_int uPlatforms) {m_uPlatforms = uPlatforms;}
	bool IsFlaggedForCurrentPlatform() const;
	bool AreThisAndParentsFlaggedForPlatform(int iPlatform) const;
#endif


protected:
	
	void						StoreBoundsToTempList();
	bool						IsChildObjectBeingScaled();

	u_int						GetChunkVersionNumber( const Asura_GUIMenu_Object_Type eType ) const;
	const char*					GetTypeName( const Asura_GUIMenu_Object_Type eType ) const;

private:

	void						AllocateContainerArraySpace( u_short usMaxContainerCount);
	void						ExpandContainerArraySpace();

	void						AllocateModifierArraySpace( u_char ucMaxModifierCount );	
	void						AllocateAnimatorArraySpace( u_char ucMaxAnimatorCount );	

// -- Variables -----------------------------------------------------------------------

protected:

	// anchoring types
	enum ASURA_GUIMENU_POSITION_MODE_AXIS
	{
		ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL,
		ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE,
		
		ASURA_GUIMENU_POSITION_MODE_AXIS_MAX
	};


	u_int				m_uType;	
	Asura_RGBA			m_xColour;

	Asura_GUIMenu_Root**	m_ppxContainerParts;
	u_short					m_usContainerPartsCount;
	u_short					m_usMaxContainerPartsCount;

	Asura_GUIMenu_Display_Condition_Params	m_axInternalDisplayConditions[ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS];
	u_int				m_uInternalDisplayConditionCount;

	Asura_Hash_ID		m_uHashName;
#ifdef ASURA_GUIMENU_STORE_NAMES
	Asura_String		m_xName;
#endif

	ASURA_TRANSLUCENCY_MODE m_eTranslucencyMode;

#ifdef ASURA_TOOLS
	u_int					m_uStyleChildEditableFlags; //Used by styles to indicate the default UseLocalDataFlags that should be set on instances of the style.

	u_int					m_uStyleUseLocalDataFlags;		// this tells the object items need to have their own styles used locally. if a bit = 1, then it doesn't use style data. 0 = use the style if it's available.
	Asura_Hash_ID			m_uHashStyleObject;				// this is the hash of the object being used to get style information from.
	Asura_GUIMenu_Root*		m_pxStyleObject;
#endif

	// Display conditions are used for things like tickboxes where the state of a widget affects which elements are displayed.
	// Condition vars are used for controlling selectability & display based on value of a console variable
	Asura_GUIMenu_ConditionVarManager m_xDisplayConditions;


	u_int m_uFlags;
#ifdef ASURA_TOOLS
	u_int m_uToolsFlags;///extra flags that are only needed while in tools
	Asura_String	m_xComments;
	Asura_String	m_xCommentUser;

	Asura_List<Asura_GUIMenu_ToolComment> m_xCommentList;
#endif // ASURA_TOOLS

	Asura_GUIMenu_Anchoring_Flag  m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_MAX];

	Asura_Vector_2 m_xPosOrigin;
	Asura_Bounding_Rectangle m_xScaledWorldBounds;
	Asura_Bounding_Rectangle m_xUnscaledLocalBounds;


private:

	Asura_GUIMenu_Modifier** m_ppxModifiers;
	u_char					 m_ucModifierCount;
	u_char					 m_ucMaxModifierCount;

	Asura_GUIMenu_Animator** m_ppxAnimators;
	u_char					 m_ucAnimatorCount;
	u_char					 m_ucMaxAnimatorCount;

	Asura_Hash_ID		m_uHashID;

	Asura_GUIMenu_Root*	m_pxParent;

	u_int					 m_uScalingFlags;

#ifdef ASURA_TOOLS
	u_int m_uPlatforms; //Bitfield of platforms that this should exist for
#endif

};
#endif // ASURA_USE_GUI_MENU_SYSTEM
#endif // __ASURA_GUIMENU_ROOT__
