// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Widget_AnimatingImage.cpp										//
//																						//
//	Description: Animating image widgets class											//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "Asura_GUIMenu_Widget_AnimatingImage.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM

#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureManagement.h"

#include "GUIMenu/Asura_Profiling_GUIMenu.h"

#include "Rendering/Asura_Overlay_Quad.h"

#include "Asura_GUIMenu_Root.h"
#include "Asura_GUIMenu_Element_Image.h"

#include "Asura_GUIMenu_Message.h"
#include "Asura_GUIMenu_MessagePump.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Widget_AnimatingImage::Asura_GUIMenu_Widget_AnimatingImage(void)
: m_fFPSScale( 1.0f)
, m_fTimeLapse( 0.0f)
, m_uFrameCount( 0)
, m_uCurrentFrame( 0)
, m_uAnimMode( ASURA_GUIMENU_ANIMMODE_NONE)
, m_iPingPongDir( 0)
, m_uStoredRepeatCount( 0)
, m_uCurrentRepeatCount( 0)	// 0xffff means infinite
, m_iAnimDirection( 1)		// This is the increment/decrement value when animating.  If reverse is set then this value will be -1.
, m_iCurrentAnimSequence( -1)
{
	u_int u;

	m_uType = ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE;

	for( u=0; u<ASURA_GUIMENU_MAX_ANIMIMAGE_FRAMES; u++)
	{
		m_axFrameInfo[u].uTexureID = 0xffffffff;

		m_axFrameInfo[u].axUV[0].u = m_axFrameInfo[u].axUV[0].v = 0.0f;
		m_axFrameInfo[u].axUV[1].u = m_axFrameInfo[u].axUV[1].v = 1.0f;

		m_axFrameInfo[u].uHashTextureName = ASURA_HASH_ID_UNSET;
		
		// SK 02.03.07 - Commented out because the Asura_String constructor clears out any data
		// m_axFrameInfo[u].xTextureName.MakeBlank();
	}

	for( u=0; u<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; u++)
	{
		m_axAnimSequence[u].fFPS = 0.0f;
		m_axAnimSequence[u].puSequenceFrameId = NULL;
		m_axAnimSequence[u].uMaxNumberFrames = 0;
		m_axAnimSequence[u].uCurrentNumFrames = 0;
	}

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_ANIMATINGIMAGE);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Widget_AnimatingImage::~Asura_GUIMenu_Widget_AnimatingImage(void)
{
	u_int i;

	for( i=0; i<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; i++)
	{
		DestroySequence( i);
	}

	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_ANIMATINGIMAGE);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::Render( u_int uState)
{
	// -- we definately need to over-ride this function.
	// -- this draws an image based on which anim frame needs to be displayed.
	Asura_Overlay_Quad xQuad;

	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_ANIMATINGIMAGE);

	if( IsValidAnimSequence( m_iCurrentAnimSequence) == true)
	{
		const ANIMIMAGE_FRAME* pxFrame = GetFrame( m_uCurrentFrame);

		Asura_Vector_2 xPos = GetScaledWorldPos();
		Asura_RGBA col = GetColour();
		xQuad.Colour = col.ToUIntARGB();

		xQuad.m_bNonRotated = true;
		xQuad.m_bGouraudShaded = false;
		xQuad.TextureIndex = pxFrame->uTexureID;
		xQuad.TranslucencyMode = GetTranslucencyMode();

		xQuad.SetRectXY( xPos.x + pxFrame->xRect.MinX-0.5f,
						 xPos.y + pxFrame->xRect.MinY-0.5f,
						 pxFrame->xRect.MaxX,
						 pxFrame->xRect.MaxY);

		// -- top-left
		xQuad.U[0] = pxFrame->axUV[0].u;
		xQuad.V[0] = pxFrame->axUV[0].v;

		// -- top-right
		xQuad.U[1] = pxFrame->axUV[1].u;
		xQuad.V[1] = pxFrame->axUV[0].v;

		// -- bottom-right
		xQuad.U[2] = pxFrame->axUV[1].u;
		xQuad.V[2] = pxFrame->axUV[1].v;

		// -- bottom-left
		xQuad.U[3] = pxFrame->axUV[0].u;
		xQuad.V[3] = pxFrame->axUV[1].v;

		xQuad.Render();			
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::Update( float fElapsedTime)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_ANIMATINGIMAGE);

	bool bLooped = false;
	ANIMIMAGE_SEQUENCE* pxSeq = NULL;
	float fFPSScale = GetFPSScale();

	PARENT::Update( fElapsedTime);

	if( (fFPSScale > 0.0f) && (IsValidAnimSequence( m_iCurrentAnimSequence) == true))
	{
		pxSeq = &m_axAnimSequence[m_iCurrentAnimSequence];

		float fTimeSwap = 1.0f/(pxSeq->fFPS * fFPSScale);

		m_fTimeLapse += fElapsedTime;

		while( m_fTimeLapse > fTimeSwap)
		{
			m_fTimeLapse -= fTimeSwap;

			switch( m_uAnimMode)
			{
				case ASURA_GUIMENU_ANIMMODE_ONESHOT:
					m_uCurrentFrame += m_iAnimDirection;
					
					if( m_iAnimDirection > 0)
					{
						if( m_uCurrentFrame >= pxSeq->uCurrentNumFrames)
						{
							m_uCurrentFrame = pxSeq->uCurrentNumFrames - 1; //m_uFrameCount-1;
							m_uAnimMode = ASURA_GUIMENU_ANIMMODE_NONE;	
						}
					}
					else
					{
						if( m_uCurrentFrame == 0)
						{
							m_uCurrentFrame = 0;
							m_uAnimMode = ASURA_GUIMENU_ANIMMODE_NONE;	
						}
					}
					break;

				case ASURA_GUIMENU_ANIMMODE_LOOP:
					if( m_uCurrentRepeatCount != 0)
					{
						m_uCurrentFrame += m_iAnimDirection;

						if( m_uCurrentFrame >= pxSeq->uCurrentNumFrames)
						{
							bLooped = true;	
						}

						if( (m_uCurrentRepeatCount != 0xffff) && (bLooped == true))
						{
							m_uCurrentRepeatCount--;	
							if( m_uCurrentRepeatCount == 0)
							{
								m_uAnimMode = ASURA_GUIMENU_ANIMMODE_NONE;	

								if( m_iAnimDirection > 0)
								{
									m_uCurrentFrame = pxSeq->uCurrentNumFrames - 1;	
								}
								else
								{
									m_uCurrentFrame = 0;	
								}

								break;
							}
						}
						m_uCurrentFrame %= pxSeq->uCurrentNumFrames;
					}
					break;

				case ASURA_GUIMENU_ANIMMODE_PINGPONG:
					if( m_uCurrentRepeatCount != 0)
					{
						m_uCurrentFrame += (m_iPingPongDir * m_iAnimDirection);
						

						if( m_uCurrentFrame == (u_int)pxSeq->uCurrentNumFrames - 1)
						{	
							m_iPingPongDir *= -1;	
						
							if( m_iAnimDirection == -1)
							{
								bLooped = true;	
							}
						}

						if( m_uCurrentFrame == 0)
						{
							m_iPingPongDir *= -1;

							if( m_iAnimDirection == 1)
							{
								bLooped = true;	
							}
						}

						if( (m_uCurrentRepeatCount != 0xffff) && (bLooped == true))
						{
							m_uCurrentRepeatCount--;
							if( m_uCurrentRepeatCount == 0)
							{
								m_uAnimMode = ASURA_GUIMENU_ANIMMODE_NONE;	
							}
						}
					}
					
					break;

				case ASURA_GUIMENU_ANIMMODE_NONE:
				default:
					break;
			}
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::Copy( const Asura_GUIMenu_Root* pxSource )
{
	PARENT::Copy( pxSource );

	//const Asura_GUIMenu_Widget_AnimatingImage* pxSourceAnimImage = static_cast<const Asura_GUIMenu_Widget_AnimatingImage*> (pxSource);

	Asura_Assert( false, "Copying of animated image elements not implemented yet!" );
}

// --------------------------------------------------------------------------------------- 
// -- UI Message handler functions -------------------------------------------------------

bool Asura_GUIMenu_Widget_AnimatingImage::OnInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	PARENT::OnInit( pxMessage);

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnDeInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnDataChange( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	u_short uSequence;
	u_short uType;
	u_short uRepeat;

	if( pxMessage->uFromID != GetID())
	{
		// -- something external has changed the data --
		if( pxMessage->WidgetType == ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE)
		{
			uSequence = (u_short)ASURA_GUIMENU_GET_ANIM_SEQUENCE( pxMessage->uData);
			uType = (u_short)ASURA_GUIMENU_GET_ANIM_TYPE( pxMessage->uData);
			uRepeat = (u_short)ASURA_GUIMENU_GET_ANIM_REPEAT( pxMessage->uData);

			switch( ASURA_GUIMENU_GET_SUB_MESSAGE( pxMessage->uActionType))
			{
				case ASURA_GUIMENU_ANIMIMAGE_PLAY:
					StartAnim( uSequence, uType, uRepeat);
					break;

				case ASURA_GUIMENU_ANIMIMAGE_PLAY_BACKWARDS:
					StartAnimReversed( uSequence, uType, uRepeat);
					break;

				case ASURA_GUIMENU_ANIMIMAGE_STOP:
					StopAnim();
					break;

				case ASURA_GUIMENU_ANIMIMAGE_NONE:
				default:
					break;
			}
		}
	}
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnDataRequest( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnClicked( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnRollIn( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnRollOut( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::OnEnable( const Asura_Message_GUIMenu_Notify* pxMessage)
{

	return true;
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_AnimatingImage::AddFrame( u_int uTexID, float u0, float v0, float u1, float v1, Asura_Bounding_Rectangle xRect)
{
	bool bTextureContained = false;
	Asura_GUIMenu_Element_Base* pxObj;
	u_int uRetSlot = 0xffff;
	Asura_GUIMenu_Element_Image* pxImageUsed = NULL;

	Asura_Assert( m_uFrameCount < ASURA_GUIMENU_MAX_ANIMIMAGE_FRAMES, "Asura_GUIMenu_Widget_AnimatingImage::AddFrame - There are no more frame slots available");

	if( m_uFrameCount < ASURA_GUIMENU_MAX_ANIMIMAGE_FRAMES)
	{
		const u_int uContainerCount = GetContainerCount();
		for( u_int i=0; i<uContainerCount; i++)
		{
			pxObj = static_cast<Asura_GUIMenu_Element_Base*>(m_ppxContainerParts[i]);
			if( pxObj->GetType() == ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE)
			{
				Asura_GUIMenu_Element_Image* pxImage = static_cast<Asura_GUIMenu_Element_Image*>(pxObj);
				if( pxImage->GetTexture() == static_cast<int>(uTexID))
				{
					bTextureContained = true;
					pxImageUsed = pxImage;
					break;
				}
			}

		}
		
		Asura_Assert( bTextureContained == true, "Asura_GUIMenu_Widget_AnimatingImage::AddFrame - uTexID not found");

		if( bTextureContained == true)
		{
			m_axFrameInfo[m_uFrameCount].uTexureID = uTexID;

			m_axFrameInfo[m_uFrameCount].uHashTextureName = pxImageUsed->GetTextureNameHash();
#ifdef ASURA_GUIMENU_STORE_NAMES
			m_axFrameInfo[m_uFrameCount].xTextureName = pxImageUsed->GetTextureName();
#endif 

			m_axFrameInfo[m_uFrameCount].axUV[0].u = u0;
			m_axFrameInfo[m_uFrameCount].axUV[0].v = v0;

			m_axFrameInfo[m_uFrameCount].axUV[1].u = u1;
			m_axFrameInfo[m_uFrameCount].axUV[1].v = v1;

			m_axFrameInfo[m_uFrameCount].xRect = xRect;

			uRetSlot = m_uFrameCount;

			m_uFrameCount++;
		}
	}

	return uRetSlot;
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_AnimatingImage::CreateSequence( u_short uMaxNumFrames, float fFPS)
{
	u_int uRetSeq = 0xffffffff;
	u_int i, j;

	for( j=0; j<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; j++)
	{
		if( m_axAnimSequence[j].puSequenceFrameId == NULL)
		{
			m_axAnimSequence[j].fFPS = fFPS;

			m_axAnimSequence[j].puSequenceFrameId = new u_int[uMaxNumFrames];

			m_axAnimSequence[j].uMaxNumberFrames = uMaxNumFrames;

			for( i=0; i<uMaxNumFrames; i++)
			{	m_axAnimSequence[j].puSequenceFrameId[i] = 0xffffffff;	}

			m_axAnimSequence[j].uCurrentNumFrames = 0;

			uRetSeq = j;
			break;
		}
	}

	return uRetSeq;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::DestroySequence( u_int uSequenceIndex)
{
	bool bRet = false;

	if( uSequenceIndex < ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES)
	{
		if( m_axAnimSequence[uSequenceIndex].puSequenceFrameId != NULL)
		{
			delete []m_axAnimSequence[uSequenceIndex].puSequenceFrameId;
			m_axAnimSequence[uSequenceIndex].puSequenceFrameId = NULL;

			m_axAnimSequence[uSequenceIndex].fFPS = 0.0f;
			m_axAnimSequence[uSequenceIndex].uCurrentNumFrames = 0;
			m_axAnimSequence[uSequenceIndex].uMaxNumberFrames = 0;
			bRet = true;
		}
	}
	return bRet;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::AddFrameToSequence( u_int uSequence, u_int uFrameIndex)
{
	bool bRet = false;

	if( (uSequence < ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES) && (m_axAnimSequence[uSequence].uCurrentNumFrames < m_axAnimSequence[uSequence].uMaxNumberFrames))
	{
		m_axAnimSequence[uSequence].puSequenceFrameId[m_axAnimSequence[uSequence].uCurrentNumFrames++] = uFrameIndex;
		bRet = true;
	}

	return bRet;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::IsValidAnimSequence( short iSequence) const
{
	bool bRet = false;

	if( (iSequence != -1) && 
		(iSequence >= 0) &&
		(iSequence < ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES))
	{

		if( m_axAnimSequence[iSequence].puSequenceFrameId != NULL)
		{
			bRet = true;
		}
	}

	return bRet;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::SetFPSScale( float fFPSScale)
{
	Asura_Assert( fFPSScale >= 0.0f, "Asura_GUIMenu_Widget_AnimatingImage::SetFPSScale - fFPS is < 0.0");

	if( fFPSScale >= 0.0f)
	{
		m_fFPSScale = fFPSScale;	
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::StartAnim( u_int uSequence, u_short uAnimMode, u_short uRepeatCount)
{
	if( IsValidAnimSequence( (short)uSequence) == true)
	{
		m_uAnimMode = uAnimMode;
		m_uStoredRepeatCount = uRepeatCount;

		m_uCurrentRepeatCount = m_uStoredRepeatCount;

		if( m_uAnimMode == ASURA_GUIMENU_ANIMMODE_PINGPONG)
		{	m_iPingPongDir = 1;	}

		m_iCurrentAnimSequence = (short)uSequence;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::StartAnimReversed( u_int uSequence, u_short uAnimMode, u_short uRepeatCount)
{
	if( IsValidAnimSequence( (short)uSequence) == true)
	{
		m_uAnimMode = uAnimMode;
		m_uStoredRepeatCount = uRepeatCount;

		m_uCurrentRepeatCount = m_uStoredRepeatCount;

		SetReverseDirection( true);

		m_uCurrentFrame = GetFrameCount()-1;

		if( m_uAnimMode == ASURA_GUIMENU_ANIMMODE_PINGPONG)
		{	m_iPingPongDir = 1;	}

		m_iCurrentAnimSequence = (short)uSequence;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::StopAnim( void)
{
	m_uAnimMode = ASURA_GUIMENU_ANIMMODE_NONE;
	m_uStoredRepeatCount = 0;
	m_uCurrentRepeatCount = 0;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_AnimatingImage::SetReverseDirection( bool bIsReversed)
{
	if( bIsReversed == true)
	{	m_iAnimDirection = -1;	}
	else
	{	m_iAnimDirection = 1;	}
}

// --------------------------------------------------------------------------------------- 

const ANIMIMAGE_FRAME* Asura_GUIMenu_Widget_AnimatingImage::GetFrame( u_int uIndex) const
{
	return &m_axFrameInfo[uIndex];
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_AnimatingImage::GetFrameCount( void) const
{
	return m_uFrameCount;
}

// --------------------------------------------------------------------------------------- 

const ANIMIMAGE_SEQUENCE* Asura_GUIMenu_Widget_AnimatingImage::GetSequence( short iSequence) const
{
	return &m_axAnimSequence[iSequence];
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Widget_AnimatingImage::GetFPSScale( void)
{
	return m_fFPSScale;
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_AnimatingImage::UpdateLocalDataFromStyle( void)
{
	u_int u, v, uNewSequenceIndex;
	short i;

	Asura_GUIMenu_Widget_AnimatingImage* pxStyleObject;
	const ANIMIMAGE_FRAME* pxAnimImageFrame;
	const ANIMIMAGE_SEQUENCE* pxAnimImageSequence;

	PARENT::UpdateLocalDataFromStyle();

	if( m_pxStyleObject != NULL)
	{
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_ANIMIMAGE_LOCAL_FLAG_FPS_SCALE) == 0)
		{
			pxStyleObject = static_cast<Asura_GUIMenu_Widget_AnimatingImage*>(m_pxStyleObject);
			
			SetFPSScale( pxStyleObject->GetFPSScale());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_ANIMIMAGE_LOCAL_FLAG_FRAME_DATA) == 0)
		{
			pxStyleObject = static_cast<Asura_GUIMenu_Widget_AnimatingImage*>(m_pxStyleObject);

			for( u=0; u<pxStyleObject->GetFrameCount(); u++)
			{
				pxAnimImageFrame = pxStyleObject->GetFrame( u);
				AddFrame( pxAnimImageFrame->uTexureID,
							pxAnimImageFrame->axUV[0].u, pxAnimImageFrame->axUV[0].v,
							pxAnimImageFrame->axUV[1].u, pxAnimImageFrame->axUV[1].v,
							pxAnimImageFrame->xRect);
			}
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_ANIMIMAGE_LOCAL_FLAG_SEQUENCE_DATA) == 0)
		{
			pxStyleObject = static_cast<Asura_GUIMenu_Widget_AnimatingImage*>(m_pxStyleObject);
		
			for( i=0; i<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; i++)
			{
				if( pxStyleObject->IsValidAnimSequence( i) == true)
				{
					pxAnimImageSequence = pxStyleObject->GetSequence( i);
					uNewSequenceIndex = CreateSequence( pxAnimImageSequence->uMaxNumberFrames, pxAnimImageSequence->fFPS);
					if( uNewSequenceIndex != 0xffffffff)
					{
						for( v=0; v<pxAnimImageSequence->uCurrentNumFrames; v++)
						{
							AddFrameToSequence( uNewSequenceIndex, pxAnimImageSequence->puSequenceFrameId[v]);
						}
					}
				}
			}
		}
	}
}
#endif

// --------------------------------------------------------------------------------------- 
	
void Asura_GUIMenu_Widget_AnimatingImage::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
//	u_int uContainerCount;
	u_int i, j;
	Asura_String xTextureName;
	Asura_Hash_ID uHashTextureName;
	Asura_Bounding_Rectangle xFrameRect;
	float u0, v0, u1, v1;
	char acName[256];
	//u_int u;

	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	// -- read in the container info
	ReadAllChildObjects( xStream, this);

	// read in animating image specific data --
	// -- write out the frame definitions --
	u_int uFrameCount;
	u_int uSeqCount;

	xStream >>uFrameCount;

	for( i=0; i<uFrameCount; i++)
	{
		xStream >> uHashTextureName;
		xStream.ReadString( acName, sizeof(acName));
		xTextureName = acName;

		xStream >> xFrameRect.MinX;
		xStream >> xFrameRect.MinY;
		xStream >> xFrameRect.MaxX;
		xStream >> xFrameRect.MaxY;

		xStream >> u0;
		xStream >> v0;
		xStream >> u1;
		xStream >> v1;

		int iTexIndex = Asura_Texture_Management::GetIndexFromHashID( uHashTextureName);
		AddFrame( iTexIndex, u0, v0, u1, v1, xFrameRect);
	}

	// -- now read in the sequences -- 
	u_short uMaxFrames, uNumFrames;
	u_int uSeqFrame;
	u_int uSeqIndex;
	float fFPS;

	xStream >> uSeqCount;

	for( i=0; i<uSeqCount; i++)
	{
		xStream >> fFPS;
		xStream >> uMaxFrames;
		xStream >> uNumFrames;

		uSeqIndex = CreateSequence( uMaxFrames, fFPS);

		for( j=0; j<uNumFrames; j++)
		{
			xStream >> uSeqFrame;
			AddFrameToSequence( uSeqIndex, uSeqFrame);
		}
	}
}

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_AnimatingImage::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	u_int i;
	short j;
	int iCounter;

	WriteLocalChunkHeader( xStream, ASURA_CHUNK_WIDGET_ANIMIMAGE_VERSION, GetType());

	PARENT::WriteToChunkStream( xStream);

	// -- write out any children --
	WriteChildObjectData( xStream);

	// -- write out the animating image class info --
	// -- the reason it's written out after the child objects is because it may need information stored in a child --

	// -- write out the frame definitions --
	xStream << m_uFrameCount;

	for( i=0; i<m_uFrameCount; i++)
	{
		xStream << m_axFrameInfo[i].uHashTextureName;
		xStream << m_axFrameInfo[i].xTextureName;

		xStream << m_axFrameInfo[i].xRect.MinX;
		xStream << m_axFrameInfo[i].xRect.MinY;
		xStream << m_axFrameInfo[i].xRect.MaxX;
		xStream << m_axFrameInfo[i].xRect.MaxY;

		xStream << m_axFrameInfo[i].axUV[0].u;
		xStream << m_axFrameInfo[i].axUV[0].v;
		xStream << m_axFrameInfo[i].axUV[1].u;
		xStream << m_axFrameInfo[i].axUV[1].v;
	}

	// -- write out the sequences --
	iCounter = 0;
	for( j=0; j<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; j++)
	{
		if( IsValidAnimSequence( j) == true)
		{
			iCounter++;
		}
	}

	// -- amount of valid sequences available --
	xStream << iCounter;

	for( j=0; j<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; j++)
	{
		if( IsValidAnimSequence( j) == true)
		{
			xStream << m_axAnimSequence[j].fFPS;
			xStream << m_axAnimSequence[j].uMaxNumberFrames;
			xStream << m_axAnimSequence[j].uCurrentNumFrames;

			for( i=0; i<m_axAnimSequence[j].uCurrentNumFrames; i++)
			{
				xStream << m_axAnimSequence[j].puSequenceFrameId[i];
			}
		}
	}
}

void Asura_GUIMenu_Widget_AnimatingImage::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML(pxExporter);
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<AnimatingImage_data>\n" );
	pxExporter->IncTabs();

	// -- write out the frame definitions --
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<frame_count>" );
	pxExporter->AddToBuffer( m_uFrameCount );
	pxExporter->AddToBuffer( "</frame_count>\n" );

	for(u_int i=0; i<m_uFrameCount; i++)
	{

		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<frame num='" ); pxExporter->AddToBuffer( i ); pxExporter->AddToBuffer( "' >\n" );
		pxExporter->IncTabs();

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<hash_texture_name>" );
		pxExporter->AddToBufferHex( m_axFrameInfo[i].uHashTextureName ); pxExporter->AddToBuffer( "</hash_texture_name>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<texture_name>" );
		pxExporter->AddToBufferEscaped( m_axFrameInfo[i].xTextureName.GetString() ); pxExporter->AddToBuffer( "</texture_name>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<min_x>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].xRect.MinX ); pxExporter->AddToBuffer( "</min_x>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<min_y>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].xRect.MinY ); pxExporter->AddToBuffer( "</min_y>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<max_x>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].xRect.MaxX ); pxExporter->AddToBuffer( "</max_x>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<max_y>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].xRect.MaxY ); pxExporter->AddToBuffer( "</max_y>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv0_u>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].axUV[0].u ); pxExporter->AddToBuffer( "</uv0_u>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv0_v>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].axUV[0].v ); pxExporter->AddToBuffer( "</uv0_v>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv1_u>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].axUV[1].u ); pxExporter->AddToBuffer( "</uv1_u>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv1_v>" );
		pxExporter->AddToBuffer( m_axFrameInfo[i].axUV[1].v ); pxExporter->AddToBuffer( "</uv1_v>\n" );
		
		pxExporter->DecTabs();
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "</frame>\n" );

	}

	// -- write out the sequences --
	int iCounter = 0;
	for( short j=0; j<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; j++)
	{
		if( IsValidAnimSequence(j) )
		{
			iCounter++;
		}
	}

	// -- amount of valid sequences available --
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<valid_sequences>" );
	pxExporter->AddToBuffer( iCounter );
	pxExporter->AddToBuffer( "</valid_sequences>\n" );

	for( short j=0; j<ASURA_GUIMENU_MAX_ANIMIMAGE_SEQUENCES; j++)
	{
		if( IsValidAnimSequence(j) )
		{
			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "<anim_sequence num='" ); pxExporter->AddToBuffer( j ); pxExporter->AddToBuffer( "' >\n" );
			pxExporter->IncTabs();

			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "<fps>" );
			pxExporter->AddToBuffer( m_axAnimSequence[j].fFPS );
			pxExporter->AddToBuffer( "</fps>\n" );

			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "<max_num_frames>" );
			pxExporter->AddToBuffer( m_axAnimSequence[j].uMaxNumberFrames );
			pxExporter->AddToBuffer( "</max_num_frames>\n" );

			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "<current_num_frames>" );
			pxExporter->AddToBuffer( m_axAnimSequence[j].uCurrentNumFrames );
			pxExporter->AddToBuffer( "</current_num_frames>\n" );

			for( int k=0; k<m_axAnimSequence[j].uCurrentNumFrames; ++k)
			{
				pxExporter->AddToBuffer( pxExporter->GetTabs() );
				pxExporter->AddToBuffer( "<sequence_frame num='" ); pxExporter->AddToBuffer( k ); pxExporter->AddToBuffer( "' id='" );
				pxExporter->AddToBuffer( m_axAnimSequence[j].puSequenceFrameId[k] ); pxExporter->AddToBuffer( "' />\n" );
			}

			pxExporter->DecTabs();
			pxExporter->AddToBuffer( pxExporter->GetTabs() );
			pxExporter->AddToBuffer( "</anim_sequence>\n" );
		}
	}

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</AnimatingImage_data>\n" );
}

#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------

void Asura_GUIMenu_Widget_AnimatingImage::SetupDefaultElements( void)
{

}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_AnimatingImage::AddToThisContainer( Asura_GUIMenu_Root* pxObject)
{
	bool bRet = false;

	if( pxObject->GetType() == Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE)
	{
		PARENT::AddToThisContainer( pxObject);
		bRet = true;
	}
	else
	{
		Asura_Assert( 0, "Animating image widgets will only accept image elements at the moment.");
	}
	return bRet;
}

// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------


// --------------------------------------------------------------------------------------- 

#endif // ASURA_USE_GUI_MENU_SYSTEM
