// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Widget_CheckBox.cpp												//
//																						//
//	Description: CheckBox widgets class													//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "GUIMenu/Asura_GUIMenu_Widget_CheckBox.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"

#include "GUIMenu/Asura_GUIMenu_Audio.h"
#include "GUIMenu/Asura_GUIMenu_Element_Base.h"
#include "GUIMenu/Asura_GUIMenu_Message.h"
#include "GUIMenu/Asura_GUIMenu_MessagePump.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"

#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Widget_CheckBox::Asura_GUIMenu_Widget_CheckBox(void)
: m_bIsOn(false)
, m_uCheckboxFlags( 0)
{
	m_uType = ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX;

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_CHECKBOX);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Widget_CheckBox::~Asura_GUIMenu_Widget_CheckBox(void)
{
	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_CHECKBOX);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_CheckBox::Render( u_int uState)
{
	if (IsRenderingDisabled())
	{
		return;
	}

	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_CHECKBOX);

	const u_int uContainerCount = GetContainerCount();
	for( u_int u=0; u<uContainerCount; u++)
	{
		if( (m_ppxContainerParts[u]->IsActionEnabled( GetCurrentState()) == true) && 
			(m_ppxContainerParts[u]->IsDisplayConditionMet( (int)m_bIsOn) == true))
		{
			// uState should contain menu state flags but no widget state flags
			Asura_Assert(!(uState & GetCurrentState()), "widget state overlaps with parent");

			m_ppxContainerParts[u]->Render( uState | GetCurrentState());
			Asura_Render::FlushDisplay();
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_CheckBox::Update( float fElapsedTime)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_CHECKBOX);

	Asura_GUIMenu_Root::Update( fElapsedTime);

	switch( GetCurrentState())
	{
		case ASURA_GUIMENU_WIDGET_STATE_DEFAULT:
			break;

		case ASURA_GUIMENU_WIDGET_STATE_ROLL_ON:
			SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_HILIGHTED);
			break;

		case ASURA_GUIMENU_WIDGET_STATE_HILIGHTED:
			break;

		case ASURA_GUIMENU_WIDGET_STATE_CLICKED:
			SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_HILIGHTED);
			break;

		case ASURA_GUIMENU_WIDGET_STATE_ROLL_OFF:
			SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_DEFAULT);
			break;

		case ASURA_GUIMENU_WIDGET_STATE_EDITING:
			break;

		case ASURA_GUIMENU_WIDGET_STATE_DISABLED:
		default:
			break;
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_CheckBox::Copy( const Asura_GUIMenu_Root* pxSource)
{
	PARENT::Copy( pxSource);

	const Asura_GUIMenu_Widget_CheckBox* pxSourceCheckbox = static_cast<const Asura_GUIMenu_Widget_CheckBox*> (pxSource);
	m_xConsoleVar.Copy(&pxSourceCheckbox->m_xConsoleVar);

	m_uCheckboxFlags = pxSourceCheckbox->GetCheckboxFlags();
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_CheckBox::ProcessInputActivity( const Asura_GUIMenu_Input& xInput)
{
	u_int uButtonsTrapped = 0;
	const u_int uButtonPresses = xInput.m_uButtonPresses;

	bool bNavigateChange = false;
	bool bDataChange = false;

	Asura_Hash_ID uHashExitID = ASURA_HASH_ID_UNSET;

	if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_UP) != 0)
	{
		uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_UP);
		uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_UP;
		bNavigateChange = true;
	}
	else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_DOWN) != 0)
	{
		uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_DOWN);
		uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_DOWN;
		bNavigateChange = true;
	}
	else if( IsCheckboxFlagSet( ASURA_GUIMENU_WIDGET_CHECKBOX_FLAGS_USE_LEFT_RIGHT_TOGGLE_CONTROL))
	{
		if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_LEFT) != 0)
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_LEFT;
			bDataChange = true;
		}
		else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_RIGHT) != 0)
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_RIGHT;
			bDataChange = true;
		}
	}
	else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_LEFT) != 0)
	{
		uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_LEFT);
		uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_LEFT;
		bNavigateChange = true;
	}
	else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_RIGHT) != 0)
	{
		uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_RIGHT);
		uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_RIGHT;
		bNavigateChange = true;
	}
	//check for clicking on the widget
	if(xInput.m_bUseCursor && ContainsPos(xInput.m_xCursorPos))
	{
		if(uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK)
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
			bDataChange = true;
		}
	}

	// -- we have an exit hash
	if( uHashExitID != ASURA_HASH_ID_UNSET)
	{
		// -- notify the next item to be selected
		Asura_GUIMenu_MessagePump::PostMessage( uHashExitID, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_IN, 0);

		// -- notify the current item, to be deselected --
		Asura_GUIMenu_MessagePump::PostMessage( GetID(), GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT, 0);
	}

	if( bNavigateChange)
	{
		if( uHashExitID != ASURA_HASH_ID_UNSET)
		{
			// -- change to a different widget --
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_NAVIGATE);
		}
		else
		{
			// -- notifies the player that they can't go in this direction.
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_FAILURE);
		}
	}

	if( bDataChange)
	{
		m_bIsOn ^= 1;
		Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_VALUE_CHANGE);

		u_int uFlags = GetFlags();
		uFlags |= ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE;
		SetFlags( uFlags);

		if( (GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE) == true))
		{
			Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_NOTIFY, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX, m_bIsOn);
			SetConsoleVar();
		}
	}
	//CLF (13.02.2009) - Ensure that the accept button can be used to toggle the check box state
	if(uButtonsTrapped == 0)
	{
		uButtonsTrapped = PARENT::ProcessInputActivity(xInput);
	}


	return uButtonsTrapped;
}

// --------------------------------------------------------------------------------------- 
// -- flag functions ---------------------------------------------------------------------

void Asura_GUIMenu_Widget_CheckBox::SetCheckboxFlags( u_int uNumericFlags)
{
	m_uCheckboxFlags = uNumericFlags;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_CheckBox::SetCheckboxFlag( u_int uNumericFlag, bool bTrueOrFalse)
{
	if( bTrueOrFalse)
	{
		m_uCheckboxFlags |= uNumericFlag;
	}
	else
	{
		m_uCheckboxFlags &= (0xffffffff - uNumericFlag);
	}
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_CheckBox::GetCheckboxFlags( void) const
{
	return m_uCheckboxFlags;
}

// --------------------------------------------------------------------------------------- 

u_int* Asura_GUIMenu_Widget_CheckBox::GetCheckboxFlagsAsPtr( void)
{
	return &m_uCheckboxFlags;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::IsCheckboxFlagSet( u_int uCheckboxFlag) const
{
	return ((m_uCheckboxFlags&uCheckboxFlag) != 0);
}

// --------------------------------------------------------------------------------------- 
// -- UI Message handler functions -------------------------------------------------------

bool Asura_GUIMenu_Widget_CheckBox::OnInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	PARENT::OnInit( pxMessage);

	// set internal data from value of console var
	InitFromConsoleVar();

	Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_REQUEST, ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX);

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnDeInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	if( GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_ON_EXIT) == true)
	{
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_NOTIFY, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX, m_bIsOn);
		SetConsoleVar();
	}

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnDataChange( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bHandled = false;

	if( pxMessage->uFromID != GetID())
	{
		// -- something external has changed the data --
		if( pxMessage->WidgetType == ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX)
		{
			m_bIsOn = ( pxMessage->uData != 0);
			SetConsoleVar();
		}

		bHandled = true;
	}
	return bHandled;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnDataRequest( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bHandled = false;

	if( pxMessage->uFromID != GetID())
	{
		// -- something external has requested the data --
		Asura_GUIMenu_MessagePump::PostMessage( pxMessage->uFromID, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX, m_bIsOn);		
		
		bHandled = true;
	}
	return bHandled;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnClicked( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	// -- Toggle the switch --
	// dont process if the left/right toggle control is enabled.
	bool bToggle = false;

	if( (pxMessage->uData == ASURA_GUIMENU_ACTION_BUTTON_ACCEPT) &&
		(!IsCheckboxFlagSet(ASURA_GUIMENU_WIDGET_CHECKBOX_FLAGS_USE_LEFT_RIGHT_TOGGLE_CONTROL)))
	{
		bToggle = true;
	}
	else if(pxMessage->uData == ASURA_GUIMENU_ACTION_BUTTON_CLICK)
	{
		bToggle = true;
	}

	if(bToggle)
	{
		m_bIsOn ^= 1;

		// -- play the accept sound, we're leaving edit mode --
		Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_ACCEPT);
	}


	// SK - TEMP FIX UNTIL IMMEDIATE NOTIFY FLAG IS EXPOSED IN TOOLS
	u_int uFlags = GetFlags();
	uFlags |= ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE;
	SetFlags( uFlags);

	if( (GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE) == true))
	{
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_NOTIFY, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX, m_bIsOn);
		SetConsoleVar();
	}
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnRollIn( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_ON);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnRollOut( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_OFF);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_CheckBox::OnEnable( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	Asura_GUIMenu_Widget_State eNewState;

	if( pxMessage->uData != 0)
	{	
		if( IsFocusedObject())
		{	eNewState = ASURA_GUIMENU_WIDGET_STATE_HILIGHTED;	}
		else
		{	eNewState = ASURA_GUIMENU_WIDGET_STATE_DEFAULT;	}
	}
	else
	{
		eNewState = ASURA_GUIMENU_WIDGET_STATE_DISABLED;	
	}

	SetCurrentState( eNewState);

	return true;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_CheckBox::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
//	u_int u, uContainerCount;

	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	ReadAllChildObjects( xStream, this);

	// read in checkbox data
	if (uChunkVersion >= 1)
	{
		// console var
		m_xConsoleVar.ReadFromChunkStream(xStream);
	}

	if (uChunkVersion >= 2)
	{
		// checkbox flags
		xStream >> m_uCheckboxFlags;
	}
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_CheckBox::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{	
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_WIDGET_CHECKBOX_VERSION, GetType());
	
	PARENT::WriteToChunkStream( xStream);

	// -- write out any children --
	WriteChildObjectData( xStream);

	// write out checkbox data
	m_xConsoleVar.WriteToChunkStream( xStream );

	// write out the checkbox specific flags
	xStream << m_uCheckboxFlags;
}

void Asura_GUIMenu_Widget_CheckBox::ExportToXML( Tools_GUIMenu_XMLExporter* pxExporter ) const
{	
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<CheckBox_data>\n" );
	pxExporter->IncTabs();

	// write out checkbox data
	m_xConsoleVar.ExportToXML( pxExporter );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<checkbox_flags>" );
	pxExporter->AddToBuffer( m_uCheckboxFlags );
	pxExporter->AddToBuffer( "</checkbox_flags>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</CheckBox_data>\n" );
}

#endif

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_CheckBox::UpdateLocalDataFromStyle( void)
{
	Asura_GUIMenu_Widget_CheckBox* pxStyleCheckbox;

	PARENT::UpdateLocalDataFromStyle();

	if( m_pxStyleObject != NULL)
	{
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_CHECKBOX_LOCAL_FLAG_FLAGS) == 0)
		{
			pxStyleCheckbox = static_cast<Asura_GUIMenu_Widget_CheckBox*>(m_pxStyleObject);
			SetCheckboxFlags( pxStyleCheckbox->GetCheckboxFlags());
		}
	}
}
#endif

// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------


// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_CheckBox::SetupDefaultElements( void)
{

}

// --------------------------------------------------------------------------------------- 

// helpers to get/set the console var
void Asura_GUIMenu_Widget_CheckBox::InitFromConsoleVar()
{
	m_xConsoleVar.GetBoolValue( m_bIsOn );
}

void Asura_GUIMenu_Widget_CheckBox::SetConsoleVar() const
{
	m_xConsoleVar.SetBoolValue( m_bIsOn );
}

// --------------------------------------------------------------------------------------- 

#endif // ASURA_USE_GUI_MENU_SYSTEM
