// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Widget_TextBox.cpp												//
//																						//
//	Description: TextBox widgets class													//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "Asura_GUIMenu_Widget_TextBox.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM

#include <wchar.h>

#include "Core/Asura_Hash.h"
#include "Core/Asura_HashedLocalisedText.h"
#include "Core/Asura_String.h"
#include "Core/Memory/Asura_Mem.h"

#include "GUIMenu/Asura_GUIMenu_Audio.h"
#include "GUIMenu/Asura_GUIMenu_Element_Base.h"
#include "GUIMenu/Asura_GUIMenu_Element_Text.h"
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"
#include "GUIMenu/Asura_GUIMenu_Message.h"
#include "GUIMenu/Asura_GUIMenu_MessagePump.h"
#include "GUIMenu/Asura_GUIMenu_System.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"
#include "Rendering/Asura_Font_System.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Widget_TextBox::Asura_GUIMenu_Widget_TextBox(void)
: m_fOffsetYLimit( 0.0f)
, m_pxActiveTextElement( NULL)
, m_uActiveTextFlagBackup( 0)
, m_wszTextEntryString( NULL)
, m_uTextEntryMaxChars( 0)
, m_bIsPasswordField(false )
, m_bAllowBlankString(false)
{
	m_uType = ASURA_GUIMENU_TYPEID_WIDGET_TEXTBOX;
	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_TEXTBOX);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Widget_TextBox::~Asura_GUIMenu_Widget_TextBox(void)
{
	delete[] m_wszTextEntryString;
	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_TEXTBOX);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::Render( u_int uState)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_TEXTBOX);

	// if doing text edit, set the text element text to be our buffer
	if (IsTextEntryBox() && m_wszTextEntryString)
	{
		Asura_Assertf(m_pxActiveTextElement, ("no active element in %s on %s", GetNameAsChar(), FindOwnerMenu() ? GetParentObject()->GetNameAsChar() : "unknown") );
		if (m_pxActiveTextElement)
		{
			if (m_bIsPasswordField)
			{
				// are we currently doing the input?
				if (Asura_GUIMenu_System::GetIDOfTextInputWidget() == GetID())
				{
					// use correct length
					const u_int uMAX_PASSWORD_LENGTH = 63;
					Asura_Assert(m_uTextEntryMaxChars <= uMAX_PASSWORD_LENGTH, "assuming limit on password length");
					Asura_UnicodeChar awcBuffer[uMAX_PASSWORD_LENGTH+1];
					u_int uLen = wcslen(m_wszTextEntryString);
					uLen = Asura_Maths::ClampToRange( uLen, 0u, uMAX_PASSWORD_LENGTH );
					for (u_int u = 0; u < uLen; ++u)
					{
						awcBuffer[u] = L'*';
					}
					awcBuffer[uLen] = 0;
					m_pxActiveTextElement->SetText(awcBuffer);
				}
				else if (m_wszTextEntryString[0])
				{
					// has some data - use fixed length of 8
					m_pxActiveTextElement->SetText(L"********");
				}
				else
				{
					// no data - leave blank
					m_pxActiveTextElement->SetText(L"");
				}
			}
			else
			{
				if (m_wszTextEntryString[0] || m_bAllowBlankString)
				{
					m_pxActiveTextElement->SetText(m_wszTextEntryString);
				}
				else
				{
					m_pxActiveTextElement->SetTextID(m_pxActiveTextElement->GetTextID());
				}
			}
		}
	}
	PARENT::Render( uState);

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::Update(float fElapsedTime)	//, u_int uButtons, POINT ptMousePos)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_TEXTBOX);

	PARENT::Update( fElapsedTime);

	BuildDisplayParams();

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::Copy( const Asura_GUIMenu_Root* pxSource)
{
	PARENT::Copy( pxSource);

	const Asura_GUIMenu_Widget_TextBox* pxSourceTextBox = static_cast<const Asura_GUIMenu_Widget_TextBox*> (pxSource);

	SetTextEntryMaxChars( pxSourceTextBox->GetTextEntryMaxChars() );
	m_bIsPasswordField = pxSourceTextBox->IsPasswordField();
	m_xConsoleVar.Copy(&pxSourceTextBox->GetConsoleVar());
}

// --------------------------------------------------------------------------------------- 
// -- UI Message handler functions -------------------------------------------------------

bool Asura_GUIMenu_Widget_TextBox::OnInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	// Init value from var
	if (IsTextEntryBox())
	{
		u_int uBufferSize = m_xConsoleVar.GetStringBufferSize();
		if (uBufferSize > 1 && uBufferSize < m_uTextEntryMaxChars)
		{
			SetTextEntryMaxChars( uBufferSize - 1);
		}
		m_xConsoleVar.GetStringValue( m_wszTextEntryString, m_uTextEntryMaxChars );
	}

	PARENT::OnInit( pxMessage);

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnDeInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	if (IsTextEntryBox() && !m_bAllowBlankString && !m_wszTextEntryString[0])
	{
		Asura_Assert(m_pxActiveTextElement, "no active element");

		if (m_pxActiveTextElement)
		{
			Asura_Assert(m_pxActiveTextElement->GetTextUsingTextID() && m_pxActiveTextElement->GetTextUsingTextID()[0], "no text");
			m_xConsoleVar.SetStringValue( m_pxActiveTextElement->GetTextUsingTextID() );
		}
	}
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnDataChange( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnDataRequest( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnClicked( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnRollIn( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_ON);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnRollOut( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_OFF);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::OnEnable( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	Asura_GUIMenu_Widget_State eNewState;

	if( pxMessage->uData != 0)
	{	
		if( IsFocusedObject())
		{	
			eNewState = ASURA_GUIMENU_WIDGET_STATE_HILIGHTED;	
		}
		else
		{	
			eNewState = ASURA_GUIMENU_WIDGET_STATE_DEFAULT;	
		}
	}
	else
	{
		eNewState = ASURA_GUIMENU_WIDGET_STATE_DISABLED;	
	}

	SetCurrentState( eNewState);

	return true;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::TextInputEnded()
{
	// set var if necessary.
	m_xConsoleVar.SetStringValue( m_wszTextEntryString );
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_TextBox::ProcessInputActivity( const Asura_GUIMenu_Input& xInput)
{
	u_int uButtonsTrapped = 0;
	const u_int uButtonPresses = xInput.m_uButtonPresses;
//	Asura_Hash_ID uHashExitID = ASURA_HASH_ID_UNSET;
	float fOffsetY = 0.0f;

	if( m_pxActiveTextElement != NULL)
	{
		// if held (but wasn't just pressed down) do scrolling
		if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT_HELD) != 0 && 
			(uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT) == 0 )
		{
			if( GetCurrentState() != ASURA_GUIMENU_WIDGET_STATE_DISABLED)
			{
				fOffsetY = m_pxActiveTextElement->GetOffsetY();

				if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_UP_HELD) != 0)
				{
					uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_UP_HELD;

					fOffsetY += 1.0f;

					if( fOffsetY > 0.0f)
					{	
						fOffsetY = 0.0f;	
					}
				}
				else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_DOWN_HELD) != 0)
				{
					uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_DOWN_HELD;

					fOffsetY -= 1.0f;

					if( fOffsetY <= m_fOffsetYLimit)
					{	
						fOffsetY = m_fOffsetYLimit;	
					}
				}
			
				m_pxActiveTextElement->SetOffsetY( fOffsetY);
			}
		}
		else
		{
			bool bStartEdit = false;
			if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT) != 0)
			{
				uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_ACCEPT;
				bStartEdit = true;
			}
			else if(xInput.m_bUseCursor && (uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK) && ContainsPos(xInput.m_xCursorPos))
			{
				//clicked within widget , so start editing
				uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
				bStartEdit = true;
			}

			if(bStartEdit)
			{
				Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_ACCEPT);

				if (IsTextEntryBox())
				{
					const Asura_UnicodeChar* wszMessage = m_pxActiveTextElement->GetTextUsingTextID();
					Asura_GUIMenu_System::StartTextInput( GetID(), m_wszTextEntryString, m_uTextEntryMaxChars, wszMessage );
				}
			}

			uButtonsTrapped |= ProcessNavigationInput( uButtonPresses);
		}
	}

	return uButtonsTrapped;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextBox::SetActiveText( Asura_Hash_ID uHashTextElement)
{
	if(uHashTextElement == ASURA_HASH_ID_UNSET) 
	{
		m_pxActiveTextElement = NULL;	
		return true;
	}

	u_int i, uContainerCount;
	bool bRet = false;
	Asura_GUIMenu_Root* pxItem;
	
	uContainerCount = GetContainerCount();

	for( i=0; i<uContainerCount; i++)
	{
		pxItem = GetContainerItem( i);
//		if( m_apxContainerParts[i]->GetID() == uHashTextElement)
		if( pxItem->GetID() == uHashTextElement)
		{
//			m_pxActiveTextElement = static_cast<Asura_GUIMenu_Element_Text*>(m_apxContainerParts[i]);
			if( m_pxActiveTextElement != NULL)
			{
				// -- restore the display settings of the previous active text element --
//				m_pxActiveTextElement->ChangeFlag( ASURA_GUIMENU_ACTION_FLAG_ALL, false);
				m_pxActiveTextElement->ChangeFlag( m_uActiveTextFlagBackup, false);
			}

			if(pxItem->GetType() == ASURA_GUIMENU_TYPEID_ELEMENT_TEXT)
			{
				m_pxActiveTextElement = static_cast<Asura_GUIMenu_Element_Text*>(pxItem);
			
				// -- this backs up the flags, so we can restore them if another activetext element is set.
				m_uActiveTextFlagBackup = m_pxActiveTextElement->GetFlags();

				m_pxActiveTextElement->ChangeFlag( ASURA_GUIMENU_ACTION_FLAG_ALL, true);
				
				// (AMM 30/10/08) Mantis 62879
				// Calling BuildDisplayParamas here causes an assert whilst trying to reference a font that is not yet loaded.
				// As BuildDisplayParams is called whenever the menu is updated I think that this call is unnecessary here.
				// It may be that we could try to ensure that projects loads the fonts before the menus but this would require
				// all projects to examine their initialisation process, when I think its OK to just remove this call.
				// BuildDisplayParams();
				
				bRet = true;
				break;
			}
		}
	}

	return bRet;
}

// --------------------------------------------------------------------------------------- 

Asura_Hash_ID Asura_GUIMenu_Widget_TextBox::GetActiveTextHashID( void) const
{
	Asura_Hash_ID uHashID = ASURA_HASH_ID_UNSET;

	if( m_pxActiveTextElement != NULL)
	{
		uHashID = m_pxActiveTextElement->GetID();
	}

	return uHashID;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::SetTextEntryMaxChars( u_int uMaxChars )
{
	if (m_wszTextEntryString)
	{
		delete[] m_wszTextEntryString;
	}
	m_wszTextEntryString = new Asura_UnicodeChar[uMaxChars + 2];	// +1 for NULL and extra + 1 in case there are any off-by-one errors anywhere.
	m_wszTextEntryString[0] = 0;
	m_uTextEntryMaxChars = uMaxChars;
}


// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_TextBox::UpdateLocalDataFromStyle( void)
{
	Asura_GUIMenu_Widget_TextBox* pxStyleTextBox;

	PARENT::UpdateLocalDataFromStyle();

  	if( m_pxStyleObject != NULL)
  	{
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTBOX_LOCAL_FLAG_TEXTENTRY) == 0)
		{
			pxStyleTextBox = static_cast<Asura_GUIMenu_Widget_TextBox*>(m_pxStyleObject);
			SetTextEntryMaxChars( pxStyleTextBox->GetTextEntryMaxChars());
		}
  	}

}
#endif
// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
//	u_int u, uContainerCount;

	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	ReadAllChildObjects( xStream, this);

	// read in TextBox specific data --
	Asura_Hash_ID uHashActiveText;
	xStream >> uHashActiveText;
	if (uChunkVersion >= 1)
	{
		u_int uMaxTextEntryChars;
		xStream >> uMaxTextEntryChars;
		SetTextEntryMaxChars(uMaxTextEntryChars);
	}
	if (uChunkVersion >= 3)
	{
		xStream >> m_bIsPasswordField;
	}
	if (uChunkVersion >= 4)
	{
		xStream >> m_bAllowBlankString;
	}
	if (uChunkVersion >= 2)
	{
		// console var
		GetConsoleVar().ReadFromChunkStream(xStream);
	}

	SetActiveText( uHashActiveText);
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_TextBox::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_WIDGET_TEXTBOX_VERSION, GetType());

	PARENT::WriteToChunkStream( xStream);

	// -- write out any children --
	WriteChildObjectData( xStream);

	// -- write out textbox specific data --
	Asura_Hash_ID uHash = ASURA_HASH_ID_UNSET;

	if( m_pxActiveTextElement != NULL)
	{
		uHash = m_pxActiveTextElement->GetID();
	}

	xStream << uHash;

	// text entry
	xStream << m_uTextEntryMaxChars;
	// version 3
	xStream << m_bIsPasswordField;
	// version 4
	xStream << m_bAllowBlankString;

	// console var
	m_xConsoleVar.WriteToChunkStream( xStream );
}

void Asura_GUIMenu_Widget_TextBox::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<TextBox_data>\n" );
	pxExporter->IncTabs();

	// -- write out textbox specific data --
	if( m_pxActiveTextElement != NULL)
	{
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<active_text_element>" );
	pxExporter->AddToBufferHex( m_pxActiveTextElement->GetID() );
	pxExporter->AddToBuffer( "</active_text_element>\n" );
	}

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<text_entry_max_chars>" );
	pxExporter->AddToBuffer( m_uTextEntryMaxChars );
	pxExporter->AddToBuffer( "</text_entry_max_chars>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<is_password_field>" );
	pxExporter->AddToBuffer( m_bIsPasswordField );
	pxExporter->AddToBuffer( "</is_password_field>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<allow_blank_string>" );
	pxExporter->AddToBuffer( m_bAllowBlankString );
	pxExporter->AddToBuffer( "</allow_blank_string>\n" );

	// console var
	m_xConsoleVar.ExportToXML( pxExporter ); 

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</TextBox_data>\n" );

}
#endif
// --------------------------------------------------------------------------------------- 
bool Asura_GUIMenu_Widget_TextBox::RemoveFromThisContainer( Asura_GUIMenu_Root* pxObject)
{
	if(m_pxActiveTextElement == pxObject) m_pxActiveTextElement = 0;

	return PARENT::RemoveFromThisContainer(pxObject);
}


// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------

void Asura_GUIMenu_Widget_TextBox::SetupDefaultElements( void)
{
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextBox::BuildDisplayParams( void)
{
	const Asura_UnicodeChar* wszText = NULL;

	if( (m_pxActiveTextElement != NULL) && ( m_fOffsetYLimit >= 0.0f))
	{

		Asura_Bounding_Rectangle xTextBBox(m_pxActiveTextElement->GetUnscaledLocalBounds());

		const Asura_Hash_ID uHashTextID = m_pxActiveTextElement->GetTextID();
		const Asura_Hash_ID uTextPage = m_pxActiveTextElement->GetTextPageHash();
		if (uHashTextID)
		{
			wszText = Asura_HashedLocalisedText::GetString(uTextPage, uHashTextID);
			Asura_Assertf( wszText, ( "missing text for %s on %s", GetNameAsChar(), FindOwnerMenu() ? FindOwnerMenu()->GetNameAsChar() : "unknown") );
		}
		else
		{
			wszText = m_pxActiveTextElement->GetText();
		}

		if( wszText)
		{
			int iFontID = m_pxActiveTextElement->GetFontID();
			if( iFontID != -1)
			{
				Asura_Font_System::SetFont( iFontID);

				m_fOffsetYLimit = -Asura_Font_System::GetHeightOfText( wszText, &xTextBBox, 1.0f, 1.0f, ASURA_TEXTFORMAT_RELATIVETORECT | ASURA_TEXTFORMAT_WORDWRAP);
				if( m_fOffsetYLimit != 0)
				{
					m_fOffsetYLimit += (xTextBBox.MaxY - xTextBBox.MinY);
				}
			}
		}
	}
}

// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------

// --------------------------------------------------------------------------------------- 

#endif // ASURA_USE_GUI_MENU_SYSTEM
