// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Widget_TextListBox.cpp												//
//																						//
//	Description: ListBox widgets class													//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"

// This file's header
#include "Asura_GUIMenu_Widget_TextListBox.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM

// Asura
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "GUIMenu/Asura_GUIMenu_Audio.h"
#include "GUIMenu/Asura_GUIMenu_Element_Base.h"
#include "GUIMenu/Asura_GUIMenu_Element_StatusText.h"
#include "GUIMenu/Asura_GUIMenu_Element_Text.h"
#include "GUIMenu/Asura_GUIMenu_Element_Image.h"
#include "GUIMenu/Asura_GUIMenu_Element_Path.h"
#include "GUIMenu/Asura_GUIMenu_Message.h"
#include "GUIMenu/Asura_GUIMenu_MessagePump.h"
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"
#include "GUIMenu/Asura_GUIMenu_System.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// Tools
#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_PreviewOverlay.h"
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#include "GUIMenu/Tools_Mode_GUIMenu.h"
#endif //ASURA_TOOLS

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_GUIMenu_Widget_TextListBox* Asura_GUIMenu_Widget_TextListBox::s_pxThis;	// for callback function

static const float fASURA_GUIMENU_HIGHLIGHTFADEMAX = 1.0f;	// 100% Alpha
static const float fASURA_GUIMENU_HIGHLIGHTFADEMIN = 0.5f;	// 50%  Alpha
static const float fASURA_GUIMENU_FADERATE = 1.0f;
static const float fASURA_GUIMENU_FADEPERSECOND	= ((fASURA_GUIMENU_HIGHLIGHTFADEMAX - fASURA_GUIMENU_HIGHLIGHTFADEMIN) / fASURA_GUIMENU_FADERATE);

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Widget_TextListBox::Asura_GUIMenu_Widget_TextListBox(void)
: m_fElapsedTime( 0.0f)
, m_uButtonRepeatPhase( 0)
, m_fCurrentRepeatDelay( 0)
, m_fInitialRepeatDelay( fASURA_GUIMENU_WIDGET_TEXTLISTBOX_CONTROL_INITIAL_REPEAT_DELAY)
, m_fSecondaryRepeatDelay( fASURA_GUIMENU_WIDGET_TEXTLISTBOX_CONTROL_SECONDARY_REPEAT_DELAY)
, m_pxEntries( NULL)
, m_pxBoundingRects( NULL)
, m_pxDecImage( NULL)
, m_pxIncImage( NULL)
, m_pxScrollBar( NULL)
, m_bDragging(false)
, m_fDecFlashTimerCountdown( 0.0f)
, m_fIncFlashTimerCountdown( 0.0f)
, m_fFlashTimerStartVal( fASURA_GUIMENU_WIDGET_TEXTLISTBOX_FLASH_TIME_DEFAULT)
, m_uEntryCount( 0)
, m_uSelectedIndex( 0xffffffff)
, m_uOldSelectedIndex( 0xffffffff)
, m_fDisplayListOffsetY( 0.0f)
, m_fDisplayTargetOffsetY( 0.0f)
, m_fMoveDir( 0.0f)
, m_uMaxAllowedEntries( 0)
, m_uMaxDisplayedEntries( 1)
, m_fEntryGapY( 0.0f)
, m_fFontScale( 1.0f)
, m_fTextScrollRate( 0.0f)
, m_uTextRenderFlags( 0)
, m_bDirtyRect( false)
, m_bFadeUp( false)
, m_fFadeAmount( 1.0f)
{
	m_uType = ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX;

	// SK - 02.03.07 - doesn't need this as Asura_String clears itself out on construction 
	//	m_xFontName.MakeBlank();

	m_xFlashColour.Set( 1.0f, 1.0f, 1.0f, 1.0f);

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_TEXTLISTBOX);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Widget_TextListBox::~Asura_GUIMenu_Widget_TextListBox(void)
{
	DeAllocateEntryList();
	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_TEXTLISTBOX);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::Render( u_int uState)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_TEXTLISTBOX);

	Asura_RGBA xTempColour;
	Asura_RGBA xOldColour;
	u_int uDecColBackup = 0xffffffff;
	u_int uIncColBackup = 0xffffffff;

	if( m_pxDecImage)
	{
		uDecColBackup = m_pxDecImage->GetColour().ToUIntRGBA();

		if( m_fDecFlashTimerCountdown > 0.0f)
		{
			Asura_RGBA xColour = GetFlashColour();
			m_pxDecImage->SetColour( &xColour);
		}
	}

	if( m_pxIncImage)
	{
		uIncColBackup = m_pxIncImage->GetColour().ToUIntRGBA();

		if( m_fIncFlashTimerCountdown > 0.0f)
		{
			Asura_RGBA xColour = GetFlashColour();
			m_pxIncImage->SetColour( &xColour);
		}
	}

	PARENT::Render( uState);
	Asura_Render::FlushDisplay();

	if( m_pxDecImage)
	{
		m_pxDecImage->SetColour( uDecColBackup);
	}

	if( m_pxIncImage)
	{
		m_pxIncImage->SetColour( uIncColBackup);
	}

	for (u_int u = 0; u < m_uEntryCount; ++u)
	{
		// change the colour of the selected entry
		//CLF (10.07.07) Only if listbox is highlighted
		if( u == m_uSelectedIndex && GetCurrentState() == ASURA_GUIMENU_WIDGET_STATE_HILIGHTED)
		{
			xTempColour = m_pxEntries[u].GetColour();
			xOldColour = xTempColour;

			Asura_RGBA xNewColour = m_xFlashColour;
			xNewColour.m_fA = m_fFadeAmount;
			m_pxEntries[u].SetColour( xNewColour );

			//Control fade when highlighted
  			if(GetCurrentState() == ASURA_GUIMENU_WIDGET_STATE_HILIGHTED)
  			{
  				if(m_bFadeUp)
  				{
  					m_fFadeAmount += fASURA_GUIMENU_FADEPERSECOND * Asura_Timers::GetFrameTime();
  					if(m_fFadeAmount >= fASURA_GUIMENU_HIGHLIGHTFADEMAX)	
  					{
  						m_fFadeAmount = fASURA_GUIMENU_HIGHLIGHTFADEMAX;
  						m_bFadeUp = false;
  					}
  				}
  				else
  				{
  					m_fFadeAmount -= fASURA_GUIMENU_FADEPERSECOND * Asura_Timers::GetFrameTime();
  					if(m_fFadeAmount <= fASURA_GUIMENU_HIGHLIGHTFADEMIN)
  					{
  						m_fFadeAmount = fASURA_GUIMENU_HIGHLIGHTFADEMIN;
  						m_bFadeUp = true;
  					}
  				}
  			}
		}

		m_pxEntries[u].Render( uState );
		
		// change the colour of the selected entry back to what it was
		//CLF (10.07.07) Only if listbox is highlighted
		if( u == m_uSelectedIndex && GetCurrentState() == ASURA_GUIMENU_WIDGET_STATE_HILIGHTED)
		{
			m_pxEntries[u].SetColour( &xOldColour);	
		}
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::Update( float fElapsedTime)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_TEXTLISTBOX);

	m_fElapsedTime = fElapsedTime;
	// SK 14.03.07 - the state of this widget is managed in the WidgetBase::Update() function.
	// This should now mean that this widget will handle states correctly, in case of modifiers.
	PARENT::Update( fElapsedTime);

	// repopulate entire list
	m_uOldSelectedIndex = m_uSelectedIndex;

	ClearAllEntries();
	s_pxThis = this;
	char acBuffer[12];
	for (u_int u = 0; u < m_uMaxAllowedEntries; ++u)
	{
#ifdef ASURA_TOOLS
		// in tools preview, populate using the specified text as population functions are unlikely to work.
		if (Tools_Mode_GUIMenu::IsPreview())
		{
			Asura_UnicodeChar awcText[64];
			Asura_PrintfToBfr( awcText, L"%S", 64, true, Tools_GUIMenu_PreviewOverlay::GetPopulationText() );
			AddTextElementCallback(awcText);
		}
		else
#endif // ASURA_TOOLS
		{
			Asura_PrintfToBfr( acBuffer, "%u", 12, true, u);
			m_xPopulationFunction.Execute( acBuffer, AddTextElementCallback );
		}
	}
	s_pxThis = NULL;

	// TODO: try and keep the same item selected (by text comparison) even if the index changes.
	if (m_uEntryCount && (m_uOldSelectedIndex != 0xffffffff))
	{
		m_uSelectedIndex = Asura_Maths::Min(m_uOldSelectedIndex, m_uEntryCount - 1);
	}


	if( GetCurrentState() != ASURA_GUIMENU_WIDGET_STATE_DISABLED)
	{
		UpdateEntryObjectBounds();

		if( m_fMoveDir != 0.0f)
		{
			m_bDirtyRect = true;
			m_fDisplayTargetOffsetY += 7.0f * fElapsedTime * m_fMoveDir;

			if( m_fMoveDir > 0.0f && m_fDisplayTargetOffsetY > m_fDisplayListOffsetY)
			{
				m_fDisplayTargetOffsetY = m_fDisplayListOffsetY;
				m_fMoveDir = 0.0f;
			}

			if( m_fMoveDir < 0.0f && m_fDisplayTargetOffsetY < m_fDisplayListOffsetY)
			{
				m_fDisplayTargetOffsetY = m_fDisplayListOffsetY;
				m_fMoveDir = 0.0f;
			}
		}
	}

	UpdateIcons();
	UpdateFlashingElements();

	ASURA_GUIMENU_PROFILE_END_TASK();
}


// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Widget_Base* Asura_GUIMenu_Widget_TextListBox::FindSelectableWidgetAtPos(const Asura_Vector_2& xPos)
{
	if(IsSelectable())
	{
		for(u_int u = 0; u < m_uEntryCount; ++u)
		{
			if ((m_fDisplayTargetOffsetY - 0.5f) < static_cast<float>(u)
				&& (m_fDisplayTargetOffsetY - 0.5f + static_cast<float>(m_uMaxDisplayedEntries)) > static_cast<float>(u))
			{
				if (m_pxEntries[u].ContainsPos(xPos - m_pxBoundingRects[u].GetTopLeft()))
				{
					return this;
				}
			}
		}
	}
	return PARENT::FindSelectableWidgetAtPos(xPos);
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Root* Asura_GUIMenu_Widget_TextListBox::GetResizeChild( const u_int u ) 
{
	Asura_GUIMenu_Root* pxRtn(NULL);

	if( u < PARENT::GetResizeChildCount() )
	{
		pxRtn = PARENT::GetResizeChild(u);
	}
	else if( u < m_uEntryCount + PARENT::GetResizeChildCount() )
	{
		pxRtn = &m_pxEntries[u - PARENT::GetResizeChildCount()];
	}

	return pxRtn;
}

u_int Asura_GUIMenu_Widget_TextListBox::GetResizeChildCount() const 
{ 
	return PARENT::GetResizeChildCount() + m_uEntryCount; 
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::Copy( const Asura_GUIMenu_Root* pxSource)
{
	PARENT::Copy( pxSource);

	const Asura_GUIMenu_Widget_TextListBox* pxSourceTextListBox = static_cast<const Asura_GUIMenu_Widget_TextListBox*> (pxSource);

	//update asura data
	SetInitialControlRepeatDelay( pxSourceTextListBox->GetInitialControlRepeatDelay());
	SetSecondaryControlRepeatDelay( pxSourceTextListBox->GetSecondaryControlRepeatDelay());
	SetTextScrollRate( pxSourceTextListBox->GetTextScrollRate());
	SetMaxDisplayedEntries( pxSourceTextListBox->GetMaxDisplayedEntries() );
	SetFontByName( pxSourceTextListBox->GetFontName().GetString());
	SetFontScale( pxSourceTextListBox->GetFontScale());

	SetTextRenderFlags( pxSourceTextListBox->GetTextRenderFlags());

	m_xCommandManager.Copy( &pxSourceTextListBox->m_xCommandManager);
	m_xPopulationFunction.Copy( &pxSourceTextListBox->m_xPopulationFunction);

	SetFlashColour( pxSourceTextListBox->m_xFlashColour.ToUIntRGBA());
	SetFlashTime( pxSourceTextListBox->m_fFlashTimerStartVal);

	m_xInitValueVar.Copy(&pxSourceTextListBox->m_xInitValueVar);

	// This method must be called last as it assumes the other data members have already been set - NDB 070408
	SetMaxAllowedEntries( pxSourceTextListBox->GetMaxAllowedEntries());
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::Resize( const Asura_Bounding_Rectangle& xBounds )
{	
	// resize this
	PARENT::Resize( xBounds );

	// resize text entries 
	for (u_int u = 0; u < m_uMaxAllowedEntries; ++u)
	{
		m_pxEntries[u].SetTextRenderFlag( ASURA_GUIMENU_TEXT_SCALE_ON_PARENT_RESIZE, GetTextRenderFlag( ASURA_GUIMENU_TEXTLISTBOX_SCALE_ON_PARENT_RESIZE ) );
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::UpdateIcons( void)
{
	bool bDisableDisplayDec = false;
	bool bDisableDisplayInc = false;

	if (m_pxScrollBar)
	{
		if (m_fDisplayTargetOffsetY <= 0.0f)
		{
			bDisableDisplayDec = true;
		}
		if( m_fDisplayTargetOffsetY > static_cast<float>(m_uEntryCount) - static_cast<float>(m_uMaxDisplayedEntries + 1))
		{
			bDisableDisplayInc = true;
		}
		float fVal = 0.0f;
		float fMax = static_cast<float>(m_uEntryCount) - static_cast<float>(m_uMaxDisplayedEntries);
		if (fMax >= 1.0f)
		{
			fVal = m_fDisplayTargetOffsetY / fMax;
		}
		m_pxScrollBar->SetCurrentPathValue(fVal);
	}
	else
	{
		if( m_uSelectedIndex == 0)
		{
			bDisableDisplayDec = true;
		}
		if( m_uSelectedIndex >= m_uMaxAllowedEntries - 1)
		{
			bDisableDisplayInc = true;
		}
	}

	if( m_pxDecImage)
	{
		m_pxDecImage->SetRenderingDisabledFlag( bDisableDisplayDec);
	}

	if( m_pxIncImage)
	{
		m_pxIncImage->SetRenderingDisabledFlag( bDisableDisplayInc);
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::UpdateFlashingElements( void)
{
	if( m_fDecFlashTimerCountdown > 0.0f)
	{
		m_fDecFlashTimerCountdown -= m_fElapsedTime;
	}

	if( m_fIncFlashTimerCountdown > 0.0f)
	{
		m_fIncFlashTimerCountdown -= m_fElapsedTime;
	}

}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::StartDecrementFlash( void)
{
	m_fDecFlashTimerCountdown = GetFlashTime();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::StartIncrementFlash( void)
{
	m_fIncFlashTimerCountdown = GetFlashTime();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::UpdateEntryObjectBounds( void)
{
	if( m_bDirtyRect == true)
	{
		// -- now calc and clip the elements bounding rectangles
		float fOffsetY = -m_fDisplayTargetOffsetY * m_fEntryGapY;
		
		float fTextHeightSum = 0.0f;
		Asura_Bounding_Rectangle xParentBounds(GetUnscaledLocalBounds());
		xParentBounds.Displace(xParentBounds.GetTopLeft() * -1.0f);//shunt parent bounds into own local space

		for (u_int u = 0; u < m_uEntryCount; ++u)
		{
			m_pxBoundingRects[u].MinX = 0.0f;
			m_pxBoundingRects[u].MinY = fOffsetY;
			m_pxBoundingRects[u].MaxX = GetUnscaledLocalBounds().GetWidth();
			m_pxBoundingRects[u].MaxY = fOffsetY + m_fEntryGapY;

			if( m_pxEntries[u].GetType() == ASURA_GUIMENU_TYPEID_ELEMENT_TEXT)
			{
				Asura_GUIMenu_Element_Text* pxTextElement = &m_pxEntries[u];

				// -- we need to clear the verticle text render flags for now.
				// -- the clipping changes mess up the justify code on vertical adjustments.
				pxTextElement->SetTextRenderFlag( ASURA_GUIMENU_TEXT_JUSTIFY_VERTICAL_BOTTOM, false);
				pxTextElement->SetTextRenderFlag( ASURA_GUIMENU_TEXT_JUSTIFY_VERTICAL_CENTRE, false);

				if( m_pxBoundingRects[u].MinY < 0.0f)
				{	
					pxTextElement->SetOffsetY( m_pxBoundingRects[u].MinY);	
				}
				else
				{	
					pxTextElement->SetOffsetY( 0.0f);
				}
			}

			CheckAndClipObjectBounds( xParentBounds, m_pxBoundingRects[u]);

			
			m_pxEntries[u].SetUnscaledLocalBounds(m_pxBoundingRects[u]);

			fOffsetY += m_fEntryGapY;

			if( !m_pxEntries[u].IsBlank() )
			{
				fTextHeightSum += m_fEntryGapY;
			}
		}

		// -- apply vertical text justification (only if text height < box height) - NDB
		if( (GetTextRenderFlags() & (ASURA_GUIMENU_TEXT_JUSTIFY_VERTICAL_CENTRE|ASURA_GUIMENU_TEXT_JUSTIFY_VERTICAL_BOTTOM)) &&
			(GetUnscaledLocalBounds().GetHeight() > fTextHeightSum) )
		{
			Asura_Vector_2 xOffset(0.0f, GetUnscaledLocalBounds().GetHeight() - fTextHeightSum);

			if( GetTextRenderFlag( ASURA_GUIMENU_TEXT_JUSTIFY_VERTICAL_CENTRE) )
			{
				xOffset.y *= 0.5f;
			}
			
			for (u_int u = 0; u < m_uEntryCount; ++u)
			{
				m_pxBoundingRects[u].Displace(xOffset);
			}
		}

		SetScaledWorldBounds(GetScaledWorldBounds());

		m_bDirtyRect = false;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::CheckAndClipObjectBounds( const Asura_Bounding_Rectangle& xParentBounds, Asura_Bounding_Rectangle& xChildBounds)
{
	// -- check is within parent bounds --
	if( (xChildBounds.MinY > xParentBounds.MaxY) || 
		(xChildBounds.MaxY < xParentBounds.MinY) ||
		(xChildBounds.MinX > xParentBounds.MaxX) || 
		(xChildBounds.MaxX < xParentBounds.MinX))
	{
		// -- child is out of bounds
		xChildBounds.MinX = xChildBounds.MaxX = xChildBounds.MinY = xChildBounds.MaxY = 0.0f;
		return;
	}

	// -- is dropping off bottom of listbox
	if( xChildBounds.MaxY > xParentBounds.MaxY)
	{
		xChildBounds.MaxY = xParentBounds.MaxY;
	}

	// -- is going of top of listbox
	if( xChildBounds.MinY < xParentBounds.MinY)
	{
		xChildBounds.MinY = xParentBounds.MinY;
	}

	// -- is dropping off right of listbox
	if( xChildBounds.MaxX > xParentBounds.MaxX)
	{
		xChildBounds.MaxX = xParentBounds.MaxX;
	}

	// -- is going of left of listbox
	if( xChildBounds.MinX < xParentBounds.MinX)
	{
		xChildBounds.MinX = xParentBounds.MinX;
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::HandleInputAutoRepeat( bool bHeldValue)
{
	bool bRetVal = bHeldValue;

	if( bHeldValue == true)
	{
		switch( m_uButtonRepeatPhase)
		{
		case 1:
			m_fCurrentRepeatDelay -= m_fElapsedTime;
			if( m_fCurrentRepeatDelay < 0.0f)
			{
				m_fCurrentRepeatDelay += m_fSecondaryRepeatDelay;
				m_uButtonRepeatPhase++;
			}
			else
			{
				bRetVal = false;
			}
			break;

		case 2:
			m_fCurrentRepeatDelay -= m_fElapsedTime;
			if( m_fCurrentRepeatDelay < 0.0f)
			{
				m_fCurrentRepeatDelay += m_fSecondaryRepeatDelay;
			}
			else
			{
				bRetVal = false;		
			}
			break;

		default:
			// -- setup case 1: details
			m_uButtonRepeatPhase++;
			m_fCurrentRepeatDelay = m_fInitialRepeatDelay;	
			break;
		}
	}
	else
	{
		m_uButtonRepeatPhase = 0;
		m_fCurrentRepeatDelay = 0.0f;
	}

	return bRetVal;
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_TextListBox::ProcessInputActivity( const Asura_GUIMenu_Input& xInput)
{
	u_int uButtonsTrapped = 0;
	const u_int uButtonPresses = xInput.m_uButtonPresses;

	//u_int uMaxAllowed = GetMaxAllowedEntries();
	u_int uMaxDisplayed = GetMaxDisplayedEntries();


	bool bUpPressed = ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_UP_HELD) != 0);
	bool bDownPressed = ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_DOWN_HELD) != 0);

	// CW - This is making the assumption that the only way to be pressing
	// the up key without holding it as well is with the scrollwheel....
	bool bUpScrolled = ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_UP) != 0) && !bUpPressed;
	bool bDownScrolled = ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_DOWN) != 0) && !bDownPressed;

	if( bUpPressed)
	{
		uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_UP_HELD;
	}
	if( bDownPressed)
	{
		uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_DOWN_HELD;
	}

	bool bPressed = bUpPressed | bDownPressed;

	if( HandleInputAutoRepeat( bPressed) == false)
	{
		bUpPressed = false;
		bDownPressed = false;
	}

	if(xInput.m_bUseCursor && ContainsPos(xInput.m_xCursorPos))
	{
		for(u_int u = 0; u < m_uEntryCount; ++u)
		{
			if ((m_fDisplayTargetOffsetY - 0.5f) < static_cast<float>(u)
				&& (m_fDisplayTargetOffsetY - 0.5f + static_cast<float>(m_uMaxDisplayedEntries)) > static_cast<float>(u))
			{
				if (m_pxEntries[u].ContainsPos(xInput.m_xCursorPos))
				{
					m_uSelectedIndex = u;
					//FocusOnSelected();
					if (uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK)
					{
						uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
						m_xCommandManager.Execute( ASURA_GUIMENU_ACTION_BUTTON_ACCEPT, m_uSelectedIndex, true);
					}
					break;
				}
			}
		}
	}

	if(xInput.m_bUseCursor
		&& (uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK)
		&& ((uButtonsTrapped & ASURA_GUIMENU_ACTION_BUTTON_CLICK) == 0))
	{
		//start editing if the cursor was clicked within the path element
		if(m_pxScrollBar && m_pxScrollBar->ContainsPos(xInput.m_xCursorPos))
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
			m_bDragging = true;
		}

		if(m_pxDecImage && m_pxDecImage->ContainsPos(xInput.m_xCursorPos))
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
			if (m_fDisplayListOffsetY >= 1.0f)
			{
				m_fDisplayListOffsetY -= 1.0f;
				m_fMoveDir = -1.0f;
			}
		}

		if(m_pxIncImage && m_pxIncImage->ContainsPos(xInput.m_xCursorPos))
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
			float fMax = static_cast<float>(m_uEntryCount) - static_cast<float>(m_uMaxDisplayedEntries);
			if (m_fDisplayListOffsetY < fMax)
			{
				m_fDisplayListOffsetY += 1.0f;
				m_fMoveDir = 1.0f;
			}
		}
	}

	if(m_bDragging)
	{
		if(xInput.m_bUseCursor && (uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK_HELD) && m_pxScrollBar)
		{
			float fPosition = m_pxScrollBar->GetPathValueNearestPosition(xInput.m_xCursorPos);
			float fMax = Asura_Maths::Max(0.0f, static_cast<float>(m_uEntryCount) - static_cast<float>(m_uMaxDisplayedEntries));
			m_fDisplayListOffsetY = static_cast<float>(Asura_Maths::FloatToInt_Near(fPosition * fMax));
			m_fMoveDir = m_fDisplayListOffsetY - m_fDisplayTargetOffsetY;
		}
		else
		{
			m_bDragging = false;
		}
	}

	if( bUpPressed || bUpScrolled)
	{
		if( m_uSelectedIndex > 0)
		{
			if(!bUpScrolled) // CW - Only when not scrolling with the mouse
			{
				m_uSelectedIndex--;
			}
			
			// -- play the navigate sound, we're moving up the list --
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_NAVIGATE);

			StartDecrementFlash();

			if( GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE) == true)
			{
				Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX, m_uSelectedIndex);
			}
			
			if( m_uSelectedIndex < static_cast<u_int>(Asura_Maths::FloatToInt_Near(m_fDisplayListOffsetY)) || bUpScrolled)
			{
				m_bDirtyRect = true;

				//m_fDisplayTargetOffsetY -= 1.0f;
				m_fDisplayListOffsetY -= 1.0f;
				m_fMoveDir = -1.0f;
				if( m_fDisplayListOffsetY < 0)
				{
					m_fDisplayListOffsetY = 0;
				}
				if( m_fDisplayTargetOffsetY < 0)
				{
					m_fDisplayTargetOffsetY = 0;
					m_fMoveDir = 0.0f;
				}
			}
		}
		else
		{
			// -- play the failed sound, we're at the top of the list --
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_FAILURE);
		}
	}
	else if( bDownPressed || bDownScrolled)
	{
		if( m_uSelectedIndex < m_uEntryCount-1)
		{
			if(!bDownScrolled)
			{
				m_uSelectedIndex++;
			}
			
			// -- play the navigate sound, we're moving up the list --
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_NAVIGATE);

			StartIncrementFlash();

			if( GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE) == true)
			{
				Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX, m_uSelectedIndex);
			}

			if( m_uSelectedIndex >= (static_cast<u_int>(Asura_Maths::FloatToInt_Near(m_fDisplayListOffsetY)) + GetMaxDisplayedEntries()) || bDownScrolled)
			{
				m_bDirtyRect = true;
				m_fDisplayListOffsetY += 1.0f;
				m_fMoveDir = 1.0f;

				// CW - Added as when less than full, was able to scroll
				if((GetEntryCount() < uMaxDisplayed))
				{
					m_fDisplayTargetOffsetY = 0.0f;
					m_fDisplayListOffsetY = 0.0f;
					m_fMoveDir = 0.0f;
				}

				if( m_fDisplayListOffsetY > (GetEntryCount() - uMaxDisplayed))
				{
					m_fDisplayListOffsetY = (float)(GetEntryCount() - uMaxDisplayed);
				}
				if( m_fDisplayTargetOffsetY > (GetEntryCount() - uMaxDisplayed))
				{
					m_fDisplayTargetOffsetY = (float)(GetEntryCount() - uMaxDisplayed);
					m_fMoveDir = 0.0f;
				}
			}
		}
		else
		{
			// -- play the failed sound, we're at the bottom of the list --
			Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_FAILURE);
		}

	}
	else
	{
		u_int uParentButtonsTrapped = PARENT::ProcessInputActivity( xInput);

		uButtonsTrapped |= uParentButtonsTrapped;
	}

	return uButtonsTrapped;
}

// --------------------------------------------------------------------------------------- 
// -- UI Message handler functions -------------------------------------------------------

bool Asura_GUIMenu_Widget_TextListBox::OnInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	PARENT::OnInit( pxMessage);

	if(!m_xInitValueVar.IsBlank())
	{
		int iInitVal;
		m_xInitValueVar.GetIntValue(iInitVal);
		m_uSelectedIndex = Asura_Maths::Min(static_cast<u_int>(iInitVal), m_uEntryCount - 1);
		FocusOnSelected();
	}

	Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_REQUEST, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnDeInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	if( GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_ON_EXIT) == true)
	{
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX, m_uSelectedIndex);
	}
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnDataChange( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	if( pxMessage->uFromID != GetID())
	{
		// -- something external has changed the data --
		if( pxMessage->WidgetType == ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX)
		{
			switch( ASURA_GUIMENU_GET_SUB_MESSAGE( pxMessage->uActionType))
			{
				case ASURA_GUIMENU_TEXTLISTBOX_CLEAR:
					ClearAllEntries();
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_GET_SELECTION_INDEX:
					Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX, m_uSelectedIndex);
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_SET_SELECTION_INDEX:
					if( pxMessage->uData < m_uEntryCount)
					{
						m_uSelectedIndex = pxMessage->uData;	
						FocusOnSelected();
					}
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_ADD_TEXT:
					AddTextElement( (Asura_UnicodeChar*)pxMessage->uData);
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_GET_TEXT:
					break;		

				case ASURA_GUIMENU_TEXTLISTBOX_ADD_TEXTHASH:
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_GET_TEXTHASH:
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_SET_IMAGE:
					break;		

				case ASURA_GUIMENU_TEXTLISTBOX_GET_IMAGE:
					break;

				case ASURA_GUIMENU_TEXTLISTBOX_NONE:
				default:
					break;
			}
		}
	}
	return true;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::FocusOnSelected( void)
{
	u_int uMaxDispEntries = GetMaxDisplayedEntries();

	m_bDirtyRect = true;
	m_fDisplayListOffsetY = (float)m_uSelectedIndex;

	if( m_uSelectedIndex > m_uEntryCount - uMaxDispEntries)
	{
		m_fDisplayListOffsetY = (float)(m_uEntryCount - uMaxDispEntries);
	}

	if( m_uSelectedIndex < uMaxDispEntries)
	{
		m_fDisplayListOffsetY = 0;
	}


	m_fDisplayTargetOffsetY = m_fDisplayListOffsetY;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnDataRequest( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	if( pxMessage->uFromID != GetID())
	{
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX, m_uSelectedIndex);
	}
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnClicked( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	u_int uButton = pxMessage->uData;

	//Converting 'click' to accept for the purposes of determining what to do.
	if(uButton | ASURA_GUIMENU_ACTION_BUTTON_CLICK)
	{
		uButton &= ~ASURA_GUIMENU_ACTION_BUTTON_CLICK;
		uButton |= ASURA_GUIMENU_ACTION_BUTTON_ACCEPT;
	}

	if( (GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_IMMEDIATE) == true) &&
		(uButton & ASURA_GUIMENU_ACTION_BUTTON_ACCEPT))
	{
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX, m_uSelectedIndex);

		// -- play the navigate sound, we're moving up the list --
		Asura_GUIMenu_Audio::Play( ASURA_GUIMENU_AUDIO_TYPE_ON_NAVIGATE);

	}

	// -- uData is the accept button type
	m_xCommandManager.Execute( uButton, m_uSelectedIndex, true);

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnRollIn( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	//Reset fade when entering listbox
	m_fFadeAmount = 1.0f;
	m_bFadeUp = false;	

	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_ON);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnRollOut( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	//Reset fade when exiting listbox
	m_fFadeAmount = 1.0f;
	m_bFadeUp = false;	

	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_OFF);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::OnEnable( const Asura_Message_GUIMenu_Notify* pxMessage)
{
Asura_GUIMenu_Widget_State eNewState;

	if( pxMessage->uData != 0)
	{	
		if( IsFocusedObject())
		{	eNewState = ASURA_GUIMENU_WIDGET_STATE_HILIGHTED;	}
		else
		{	eNewState = ASURA_GUIMENU_WIDGET_STATE_DEFAULT;	}
	}
	else
	{
		eNewState = ASURA_GUIMENU_WIDGET_STATE_DISABLED;	
	}

	SetCurrentState( eNewState);

	return true;

}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetMaxAllowedEntries( u_int uMaxEntries)
{
	AllocateEntryList( uMaxEntries, GetMaxAllowedEntries());

	m_uMaxAllowedEntries = uMaxEntries;	

	m_bDirtyRect = true;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetSelectedIndex( u_int uSelected, bool bFocus)
{
	m_uSelectedIndex = uSelected;
	m_uOldSelectedIndex = uSelected;

	if(bFocus)
	{
		FocusOnSelected();
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetMaxDisplayedEntries( u_int uMaxDisplayedEntries)
{
	m_uMaxDisplayedEntries = uMaxDisplayedEntries;	
	
	SetEntryHeight( GetUnscaledLocalBounds().GetHeight() / m_uMaxDisplayedEntries);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetFontByName( const char* szFontName)
{
	m_xFontName = szFontName;
}

// --------------------------------------------------------------------------------------- 

const Asura_String Asura_GUIMenu_Widget_TextListBox::GetFontName( void) const
{
	return m_xFontName;
}

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_Element_Image* Asura_GUIMenu_Widget_TextListBox::GetDecImage( void) const
{
	return m_pxDecImage;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetDecImage( Asura_GUIMenu_Element_Image* pxDecImage)
{
	m_pxDecImage = pxDecImage;
}

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_Element_Image* Asura_GUIMenu_Widget_TextListBox::GetIncImage( void) const
{
	return m_pxIncImage;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetIncImage( Asura_GUIMenu_Element_Image* pxIncImage)
{
	m_pxIncImage = pxIncImage;
}

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_Element_Path* Asura_GUIMenu_Widget_TextListBox::GetScrollBar( void) const
{
	return m_pxScrollBar;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetScrollBar( Asura_GUIMenu_Element_Path* pxScrollBar)
{
	m_pxScrollBar = pxScrollBar;
}

// --------------------------------------------------------------------------------------- 

const Asura_RGBA Asura_GUIMenu_Widget_TextListBox::GetFlashColour( void) const
{
	return m_xFlashColour;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetFlashColour( float fRed, float fGreen, float fBlue, float fAlpha)
{
	m_xFlashColour.m_fR = fRed;
	m_xFlashColour.m_fG = fGreen;
	m_xFlashColour.m_fB = fBlue;
	m_xFlashColour.m_fA = fAlpha;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetFlashColour( u_int uRGBAColourValue)
{
	m_xFlashColour.FromUIntRGBA( uRGBAColourValue);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetFlashTime( float fFlashTime)
{
	m_fFlashTimerStartVal = fFlashTime;
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Widget_TextListBox::GetFlashTime( void) const
{
	return m_fFlashTimerStartVal;
}

// --------------------------------------------------------------------------------------- 
u_int Asura_GUIMenu_Widget_TextListBox::GetMaxAllowedEntries( void) const
{
	return m_uMaxAllowedEntries;	
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_TextListBox::GetMaxDisplayedEntries( void) const
{
	return m_uMaxDisplayedEntries;	
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetUnscaledLocalBounds( const Asura_Bounding_Rectangle& xBounds )
{
	PARENT::SetUnscaledLocalBounds(xBounds);

	SetEntryHeight( GetUnscaledLocalBounds().GetHeight() / m_uMaxDisplayedEntries);
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_TextListBox::UpdateLocalDataFromStyle( void)
{
	PARENT::UpdateLocalDataFromStyle();

	if( m_pxStyleObject != NULL)
	{
		Asura_GUIMenu_Widget_TextListBox* pxStyleListBox = static_cast<Asura_GUIMenu_Widget_TextListBox*>(m_pxStyleObject);
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_WIDTH) == 0)
		{
			m_xUnscaledLocalBounds.MaxX = pxStyleListBox->GetUnscaledLocalBounds().GetWidth() + m_xUnscaledLocalBounds.MinX;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_HEIGHT) == 0)
		{
			m_xUnscaledLocalBounds.MaxY = pxStyleListBox->GetUnscaledLocalBounds().GetHeight() + m_xUnscaledLocalBounds.MinY;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_MAX_DISPLAY_ENTRIES) == 0)
		{
			SetMaxDisplayedEntries( pxStyleListBox->GetMaxDisplayedEntries());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_MAX_ALLOWED_ENTRIES) == 0)
		{
			SetMaxAllowedEntries( pxStyleListBox->GetMaxAllowedEntries());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_RENDER_FLAGS) == 0)
		{
			SetTextRenderFlags( pxStyleListBox->GetTextRenderFlags());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_FLASH_DETAILS) == 0)
		{
			m_xFlashColour = pxStyleListBox->m_xFlashColour;
			m_fFlashTimerStartVal = pxStyleListBox->m_fFlashTimerStartVal;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_CONTROL_REPEAT) == 0)
		{
			m_fInitialRepeatDelay = pxStyleListBox->m_fInitialRepeatDelay;
			m_fSecondaryRepeatDelay = pxStyleListBox->m_fSecondaryRepeatDelay;
		}
	
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_FONT) == 0)
		{
			SetFontByName( pxStyleListBox->GetFontName() );
			SetFontScale( pxStyleListBox->GetFontScale() );
			SetTextScrollRate( pxStyleListBox->GetTextScrollRate() );
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_POPULATION_FUNCTION) == 0)
		{
			m_xPopulationFunction.Copy( &(pxStyleListBox->m_xPopulationFunction) );
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_TEXTLISTBOX_LOCAL_FLAG_COMMAND_MANAGER) == 0)
		{
			m_xCommandManager.Copy( &(pxStyleListBox->m_xCommandManager) );
		}
	}
}
#endif

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	float fWidth( 0.0f);
	float fHeight( 0.0f);
	float fFontScale( 1.0f);

	u_int uMaxAllowed, uMaxDisplayed;
	char acName[256];

	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	ReadAllChildObjects( xStream, this);

	// read in ListBox specific data --
	if (uChunkVersion < 10)
	{
		xStream >> fWidth;
		xStream >> fHeight;
	}

	xStream >> uMaxAllowed;
	xStream >> uMaxDisplayed;

	xStream.ReadString( acName, sizeof(acName));
	SetFontByName( acName);

	if( uChunkVersion >= 7)
	{
		xStream >> fFontScale;
	}

	if( uChunkVersion >= 1)
	{
		xStream >> m_uTextRenderFlags;
	}

	if( uChunkVersion >= 2)
	{
		// V2 - read console commands through the manager system
		m_xCommandManager.ReadFromChunkStream( xStream );
	}

	if (uChunkVersion >= 3)
	{
		m_xPopulationFunction.ReadFromChunkStream( xStream );
	}


	if ( uChunkVersion >= 4)
	{
		u_int uHashDecID, uHashIncID;

		xStream >> uHashDecID;
		xStream >> uHashIncID;

		Asura_GUIMenu_Root* pxTempRoot = FindContainerItemByID( uHashDecID);
		SetDecImage( static_cast<Asura_GUIMenu_Element_Image*>(pxTempRoot));

		pxTempRoot = FindContainerItemByID( uHashIncID);
		SetIncImage( static_cast<Asura_GUIMenu_Element_Image*>(pxTempRoot));

		xStream >> m_xFlashColour.m_fR;
		xStream >> m_xFlashColour.m_fG;
		xStream >> m_xFlashColour.m_fB;
		xStream >> m_xFlashColour.m_fA;

		xStream >> m_fFlashTimerStartVal;
	}

	if( uChunkVersion >= 5)
	{
		xStream >> m_fInitialRepeatDelay;
		xStream >> m_fSecondaryRepeatDelay;
	}

	if( uChunkVersion >= 6)
	{
		xStream >> m_fTextScrollRate;
	}

	if (uChunkVersion >= 8)
	{
		// console var
		m_xInitValueVar.ReadFromChunkStream(xStream);
	}

	if (uChunkVersion >= 9)
	{
		u_int uHash;

		xStream >> uHash;

		Asura_GUIMenu_Root* pxTempRoot = FindContainerItemByID( uHash);
		SetScrollBar( static_cast<Asura_GUIMenu_Element_Path*>(pxTempRoot));
	}

	// init based on settings
	if (uChunkVersion < 10)
	{
		m_xUnscaledLocalBounds.MaxX = fWidth + m_xUnscaledLocalBounds.MinX;
		m_xUnscaledLocalBounds.MaxY = fHeight + m_xUnscaledLocalBounds.MinY;
	}

	SetMaxDisplayedEntries( uMaxDisplayed);
	SetMaxAllowedEntries( uMaxAllowed);
	SetFontScale( fFontScale);
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_TextListBox::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_WIDGET_TEXTLISTBOX_VERSION, GetType());

	PARENT::WriteToChunkStream( xStream);

	// -- write out any children --
	WriteChildObjectData( xStream);

	// -- write out the listbox specific data --
	xStream << m_uMaxAllowedEntries;
	xStream << m_uMaxDisplayedEntries;
	xStream << m_xFontName;
	xStream << m_fFontScale;
	xStream << m_uTextRenderFlags;

	m_xCommandManager.WriteToChunkStream( xStream);
	m_xPopulationFunction.WriteToChunkStream(xStream);

	u_int uHashDecID = ASURA_HASH_ID_UNSET;
	u_int uHashIncID = ASURA_HASH_ID_UNSET;

	if( m_pxDecImage)
	{
		uHashDecID = m_pxDecImage->GetID();
	}
	if( m_pxIncImage)
	{
		uHashIncID = m_pxIncImage->GetID();
	}

	xStream << uHashDecID;
	xStream << uHashIncID;

	xStream << m_xFlashColour.m_fR;
	xStream << m_xFlashColour.m_fG;
	xStream << m_xFlashColour.m_fB;
	xStream << m_xFlashColour.m_fA;

	xStream << m_fFlashTimerStartVal;
	xStream << m_fInitialRepeatDelay;
	xStream << m_fSecondaryRepeatDelay;
	xStream << m_fTextScrollRate;

	// console var
	m_xInitValueVar.WriteToChunkStream(xStream);

	u_int uHashScrollBar = ASURA_HASH_ID_UNSET;

	if( m_pxScrollBar)
	{
		uHashScrollBar = m_pxScrollBar->GetID();
	}

	xStream << uHashScrollBar;
}

void Asura_GUIMenu_Widget_TextListBox::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<TextListBox_data>\n" );
	pxExporter->IncTabs();

	// -- write out the listbox specific data --
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<max_allowed_entries>" );
	pxExporter->AddToBuffer( m_uMaxAllowedEntries );
	pxExporter->AddToBuffer( "</max_allowed_entries>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<max_displayed_entries>" );
	pxExporter->AddToBuffer( m_uMaxDisplayedEntries );
	pxExporter->AddToBuffer( "</max_displayed_entries>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<font_name>" );
	pxExporter->AddToBufferEscaped( m_xFontName.GetString() );
	pxExporter->AddToBuffer( "</font_name>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<font_scale>" );
	pxExporter->AddToBuffer( GetFontScale() );
	pxExporter->AddToBuffer( "</font_scale>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<text_render_flags>" );
	pxExporter->AddToBufferHex( m_uTextRenderFlags );
	pxExporter->AddToBuffer( "</text_render_flags>\n" );

	m_xCommandManager.ExportToXML( pxExporter ); 

	m_xPopulationFunction.ExportToXML( pxExporter ); 

	if( m_pxDecImage)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<dec_image>" );
		pxExporter->AddToBufferHex( m_pxDecImage->GetID() );
		pxExporter->AddToBuffer( "</dec_image>\n" );
	}
	if( m_pxIncImage)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<inc_image>" );
		pxExporter->AddToBufferHex( m_pxIncImage->GetID() );
		pxExporter->AddToBuffer( "</inc_image>\n" );
	}

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<flash_colour r=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fR );
	pxExporter->AddToBuffer( "\" g=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fG );
	pxExporter->AddToBuffer( "\" b=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fB );
	pxExporter->AddToBuffer( "\" a=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fA );
	pxExporter->AddToBuffer("\" />\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<flash_timer_start_value>" );
	pxExporter->AddToBuffer( m_fFlashTimerStartVal );
	pxExporter->AddToBuffer( "</flash_timer_start_value>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<initial_repeat_delay>" );
	pxExporter->AddToBuffer( m_fInitialRepeatDelay );
	pxExporter->AddToBuffer( "</initial_repeat_delay>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<secondary_repeat_delay>" );
	pxExporter->AddToBuffer( m_fSecondaryRepeatDelay );
	pxExporter->AddToBuffer( "</secondary_repeat_delay>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<text_scroll_rate>" );
	pxExporter->AddToBuffer( m_fTextScrollRate );
	pxExporter->AddToBuffer( "</text_scroll_rate>\n" );

	// console var
	pxExporter->AddToBuffer( "<init_value_var>" );
	m_xInitValueVar.ExportToXML(pxExporter);
	pxExporter->AddToBuffer( "</init_value_var>" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</TextListBox_data>\n" );

	if( m_pxScrollBar)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<scroll_bar>" );
		pxExporter->AddToBufferHex( m_pxScrollBar->GetID() );
		pxExporter->AddToBuffer( "</scroll_bar>\n" );
	}

}

#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
bool Asura_GUIMenu_Widget_TextListBox::RemoveFromThisContainer( Asura_GUIMenu_Root* pxObject)
{
	//Check for one of the children that we have a pointer to being removed.
	if(m_pxDecImage == pxObject) m_pxDecImage = 0;
	if(m_pxIncImage == pxObject) m_pxIncImage = 0;
	if(m_pxScrollBar == pxObject) m_pxScrollBar = 0;

	return PARENT::RemoveFromThisContainer(pxObject);
}


// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------

void Asura_GUIMenu_Widget_TextListBox::ClearAllEntries( void)
{
	m_uSelectedIndex = 0xffffffff;
	m_uEntryCount = 0;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::AddTextElement( const Asura_UnicodeChar* pwszTextString )
{
	Asura_Assert( !m_xFontName.IsBlank(), "Asura_GUIMenu_Widget_TextListBox::AddTextElement - No Font name specified.");
	Asura_Assert( GetMaxAllowedEntries() > 0, "Asura_GUIMenu_Widget_TextListBox::AddTextElement - Max entries is zero.");

	if( m_uEntryCount < GetMaxAllowedEntries())
	{
		Asura_GUIMenu_Element_Text* pxTextElement = &m_pxEntries[m_uEntryCount++];
		pxTextElement->SetText(pwszTextString);
		// should be already initialised: pxTextElement->SetFontByName(m_xFontName);
		// should be already initialised: pxTextElement->SetTextRenderFlags( GetTextRenderFlags());
		// should be already initialised: pxTextElement->SetColour( GetColour() );

		// -- this is a safeguard to ensure that if this item is the first, it will set it as selected
		if( m_uSelectedIndex == 0xffffffff)
		{	
			m_uSelectedIndex = 0;	
		}

		m_bDirtyRect = true;
	}
}

// --------------------------------------------------------------------------------------- 

// static callback function
void Asura_GUIMenu_Widget_TextListBox::AddTextElementCallback( const Asura_UnicodeChar* pwszTextString )
{
	Asura_Assert(s_pxThis, "didn't set this pointer for callback");
	s_pxThis->AddTextElement( pwszTextString );
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_TextListBox::GetTextRenderFlag( u_int uFlag) const
{
	return ((m_uTextRenderFlags&uFlag) != 0);
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_TextListBox::GetTextRenderFlags( void) const
{
	return m_uTextRenderFlags;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetupDefaultElements( void)
{
}

// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------

void Asura_GUIMenu_Widget_TextListBox::AllocateEntryList( u_int uNewListSize, u_int uOldListSize)
{
	Asura_GUIMenu_Element_Text* pxList = Asura_GUIMenu_System::CreateTextElementArray( uNewListSize );
	Asura_Assert( pxList != NULL, "Asura_GUIMenu_Widget_TextListBox::AllocateEntryList - Failed to allocate text elements");

	Asura_Bounding_Rectangle* pxRects = new Asura_Bounding_Rectangle[uNewListSize];
	Asura_Assert( pxRects != NULL, "Asura_GUIMenu_Widget_TextListBox::AllocateEntryList - Failed to allocate bounds list space");

	// set bounding rects to 0,0,0,0
	Asura_Mem::Set( pxRects, 0, uNewListSize * sizeof( Asura_Bounding_Rectangle));
	
	// free the old data
	DeAllocateEntryList();

	// set all the text elements from the listbox settings
	Asura_Assert(m_fEntryGapY > 0.0f, "invalid height - was SetMaxAllowedEntries() called before SetMaxDisplayedEntries()?");

	for( u_int u = 0; u < uNewListSize; u++)
	{
		pxList[u].SetFontByName(m_xFontName);
		pxList[u].SetTextRenderFlags( GetTextRenderFlags());
		pxList[u].SetColour( GetColour() );
		pxList[u].SetScrollRate(m_fTextScrollRate);
		pxList[u].SetFontScale( GetFontScale() );
		Asura_Bounding_Rectangle xBounds = pxList[u].GetUnscaledLocalBounds();
		xBounds.MaxX = xBounds.MinX + GetUnscaledLocalBounds().GetWidth();
		xBounds.MaxY = xBounds.MinY + m_fEntryGapY;
		pxList[u].SetUnscaledLocalBounds(xBounds);

		if (GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX))
		{
			pxList[u].SetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX, true);
			pxList[u].SetPosMode(0, POS_RELATIVE);
		}
		else if (GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTY))
		{
			pxList[u].SetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTY, true);
			pxList[u].SetPosMode(0, POS_RELATIVE_ALTAXIS);
		}

		if (GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY))
		{
			pxList[u].SetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY, true);
			pxList[u].SetPosMode(1, POS_RELATIVE);
		}
		else if (GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTX))
		{
			pxList[u].SetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTX, true);
			pxList[u].SetPosMode(1, POS_RELATIVE_ALTAXIS);
		}
	}

	// set our pointers to the new data
	m_pxEntries = pxList;
	m_pxBoundingRects = pxRects;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::DeAllocateEntryList( void)
{
	if( (m_pxEntries != NULL) && (m_pxBoundingRects != NULL))
	{
		ClearAllEntries();

		delete []m_pxEntries;
		m_pxEntries = NULL;

		delete []m_pxBoundingRects;
		m_pxBoundingRects = NULL;
	}
}


// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_TextListBox::SetTextRenderFlag( u_int uFlagsToSet, bool bFlagSet)
{
	if( bFlagSet)
	{
		m_uTextRenderFlags |= uFlagsToSet;
	}
	else
	{
		m_uTextRenderFlags &=~ uFlagsToSet;
	}
}

#endif // ASURA_USE_GUI_MENU_SYSTEM
