/**
*
* \file Asura_Rumble_System.h
*
* Created by James Paulin - 14/08/2002 16:46:12
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
* Reorganisation - AG 17/01/2003
* A game can add a platform rumble (which can take into account the number and type of motors available
* since it's platform specific) to the system, getting a handle if it may need modifying/stopping early,
* e.g. machinery that the player moves closer to or further away from. A platform rumble should contain
* one rumble for each motor, and works out the final rumble for each individually. Unless set on infinite
* repeat, they will be automatically removed upon completion.
* Some basic rumbles will be declared here, but projects should be able to create their own or redefine
* these on a project-specific basis if required.
*
**/
#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_RUMBLE_SYSTEM_H__
#define __ASURA_RUMBLE_SYSTEM_H__

#include "Core/Asura_Maths.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Mempool.h"
#include "Input/Asura_Platform_Rumble.h"
#include "Network/Asura_Guid.h"

enum RUMBLE_BASE_EFFECT
{
//Note the first two are much simpler functions than the others, so use them as much as possible.
	RUMBLE_BASE_CONSTANT,		//Constant rumble
	RUMBLE_BASE_LINEAR,			//Linearly increasing rumble
	RUMBLE_BASE_INCCURVE,		//Curve with increasing gradient (may use square between 0 and 1)
	RUMBLE_BASE_DECCURVE,		//Curve with decreasing gradient (may use sin between 0 and pi/2)
	RUMBLE_BASE_SYMWAVE			//A symmetrical wave (may use sin between 0 and pi, or maybe (cos+1)/2 between -pi and pi)
};

enum RUMBLE_ADJUST
{
	RUMBLE_ADJUST_NONE = 0,
	RUMBLE_ADJUST_REVERSE = 1,	//Use the effect in reverse
	RUMBLE_ADJUST_FLIP = 2		//Use the vertical reflection of the effect
};

//Class storing the basic step used (individually or concatenated together) to make up a single-motor rumble
class Asura_Rumble_Step
{
public:
	RUMBLE_BASE_EFFECT eBaseEffect;
	RUMBLE_ADJUST eAdjustment;
	float fStepMaxMagnitude;	//Top rumble rate for this step.
	float fStepMinMagnitude;	//Base rumble rate for this step. Must be less than fStepMaxMagnitude. The effect will adjust so this is the lowest level reached. Will not affect RUMBLE_BASE_CONSTANT.
	float fStepDecay;			//Decay time for this step. This step will last 1/fStepDecay seconds
	int iStepEffectPlays;		//Number of times this particular step repeats before continuing to the next step in the sequence e.g. 1 will play the step once, 2 twice and so on. 0 will signify an infinite loop
	Asura_Rumble_Step* pxNextStep;	//Next step in the sequence
};

//Structure storing the current position through a rumble of an active rumble.
struct RumblePosition
{
	float fStepPosition;		//Will always be between 0 and 1
	int iStepEffectPlaysRemaining;
	Asura_Rumble_Step* pxCurrentStep;
};

//Structure for registering players using the rumble system
struct CurrentRumblers
{
	Asura_Guid uGuid;
	int iController;
    float fRumbleMultiplier;    // in the range [0.f - 1.f]
};

//Class for single rumble for one motor only. See Asura_Platform_Rumble for use of multiple motors.
class Asura_Rumble
{
	friend class Asura_Platform_Rumble;
	friend class Asura_ActiveRumble;
	friend class Asura_Rumble_System;

	public:
		//Note that for now the total time for each step is 1/stepdecay seconds, so if there are 5 steps
		//each with step decay of 1 the total time for the rumble with overall decay of 1 is 5 seconds
		//this may be changed later so all rumbles have a fixed base length.

        float m_fMotorMaxMagnitude; //Modifier to all steps in the sequence.
        float m_fMotorMinMagnitude; //Modifier to all steps in the sequence. Must be less than fMaxMagnitude. Will not affect RUMBLE_BASE_CONSTANT.
        float m_fMotorDecay;        //Modifier to all decays in sequence so whole thing can be sped up/slowed down

		static Asura_Handle CreateAsuraRumble(RUMBLE_BASE_EFFECT eBase, RUMBLE_ADJUST eAdjust, float fMaxMagnitude,
					float fMinMagnitude, float fDecay, int iPlays);

		static void AddRumbleStep(Asura_Handle iRumbleHandle, RUMBLE_BASE_EFFECT eBase, RUMBLE_ADJUST eAdjust, float fMaxMagnitude,
					float fMinMagnitude, float fDecay, int iPlays);

		static Asura_Rumble* GetRumblePointer(Asura_Handle iHandle)
		{
			return s_pxDynamicRumbleList->GetPointer(iHandle);
		};

	private:
		static Asura_TrackingHandlePool<Asura_Rumble>* s_pxDynamicRumbleList;
		static Asura_TrackingHandlePool<Asura_Rumble_Step>* s_pxDynamicRumbleStepList;

		Asura_Rumble_Step* m_pxFirstStep;
};

//Class containing a set of rumbles making up a rumble effect on the current platform
class Asura_Platform_Rumble
{
	public:
		Asura_Rumble* m_apxMotorRumbles[PLATFORM_NUM_RUMBLE_MOTORS];
};

enum ASURA_ACTIVE_RUMBLE_TYPE
{
	ASURA_HUD_RUMBLE,
	ASURA_WORLD_RUMBLE,
};

//Class for any rumble currently running in the game world, either for specific players or generically in the world.
class Asura_ActiveRumble
{
	public:
		Asura_Platform_Rumble* m_pxActiveRumble;
	    float m_fMaxMagnitude;			//Modifier to all motors.
	    float m_fMinMagnitude;			//Modifier to all motors. Must be less than fMaxMagnitude. Will not affect RUMBLE_BASE_CONSTANT.
	    float m_fDecay;					//Modifier to all decays in each sequence    
		bool m_bPaused;
		bool m_bStarted;
		int  m_iEffectPlaysRemaining;	//Number of times these sequences will repeat. e.g. 1 will play the effect once, 2 twice and so on. 0 will signify an infinite loop
										//Current plan is that sequences will wait for each other before repeating - maybe should be optional whether do this or overlap loops?

		Asura_Handle m_iMyHandle;
// This function enables updates to the maximum, while adjusting the min by the same relative amount. Could be useful for fixed-point continuous rumble sources.
		inline void ChangeMaxMinRelativeMagnitude(float fNewMaxMagnitude)
		{
			if (fNewMaxMagnitude <= 0.0f)
			{
				m_fMaxMagnitude = 0.0f;
				m_fMinMagnitude = 0.0f;
			}
			else
			{
				float fMultiplier = fNewMaxMagnitude / m_fMaxMagnitude;
				m_fMaxMagnitude = fNewMaxMagnitude;
				m_fMinMagnitude *= fMultiplier;
				SanityCheck();
			}
		};

//These two functions are inherently more dangerous than the above, as you can easily try to set invalid values for either by not keeping track of both. Probably best used only on initialising.
		inline void ChangeMaxMagnitude(float fNewMaxMagnitude)
		{
			m_fMaxMagnitude = fNewMaxMagnitude;
			SanityCheck();
		};

		inline void ChangeMinMagnitude(float fNewMinMagnitude)
		{
			m_fMinMagnitude = fNewMinMagnitude;
			SanityCheck();
		};

		inline void ChangeDecay(float fNewDecay)
		{
			if (fNewDecay < 0.0f)
			{
				Asura_Assert(0, "Attempted to set negative rate of decay, using 0 (no decay)");
				fNewDecay = 0.0f;
			}
			m_fDecay = fNewDecay;
		};

//Update will return the current magnitude for the given motor of this active rumble.
//A negative value will signify that the rumble has passed the end of a loop, which the calling function can use to continue to the next loop, or delete the rumble as required.
		float Update(float fTimeSlice, int iMotor);

        inline void ResetPositions()
        {
	        for (int i=0; i<PLATFORM_NUM_RUMBLE_MOTORS; i++)
	        {
		        m_axMotorEffectPositions[i].fStepPosition = 0.0f;

                if(m_pxActiveRumble &&
                    m_pxActiveRumble->m_apxMotorRumbles[i])   // (NRM - 28.03.2003) check for NULL pointer
                {
                    m_axMotorEffectPositions[i].pxCurrentStep = m_pxActiveRumble->m_apxMotorRumbles[i]->m_pxFirstStep;
			        m_axMotorEffectPositions[i].iStepEffectPlaysRemaining = m_axMotorEffectPositions[i].pxCurrentStep->iStepEffectPlays;
                }
                else
                {
                    m_axMotorEffectPositions[i].iStepEffectPlaysRemaining = 0;
                }
	        }
        }
		ASURA_ACTIVE_RUMBLE_TYPE m_eMyType;
	private:
		inline void SanityCheck()
		{
			if (m_fMaxMagnitude > 1.0f)
			{
				Asura_Assert(0, "Maximum magnitude greater than 1.0");
				m_fMaxMagnitude = 1.0f;
			}

			if (m_fMinMagnitude < 0.0f)
			{
				Asura_Assert(0, "Minimum magnitude less than 0.0");
				m_fMinMagnitude = 0.0f;
			}

// Arbitrarily the max is brought up to the minimum if it is less than it. Could easily be switched so the min comes down to the max.
			if (m_fMaxMagnitude < m_fMinMagnitude)
			{
				Asura_Assert(0, "Maximum magnitude less than minimum magnitude");
				m_fMaxMagnitude = m_fMinMagnitude;
			}
		};

	protected:
		RumblePosition m_axMotorEffectPositions[PLATFORM_NUM_RUMBLE_MOTORS];
};

//Class for any rumble that exists solely for a given player on this system (using given controller)
class Asura_HUDRumble : public Asura_ActiveRumble
{
	public:
		int m_iController;
};

//Class for any rumble that eminates from a point in the world. Players feel them when they get sufficiently close (within the radius, greater magnitude the closer they are)
class Asura_WorldRumble : public Asura_ActiveRumble
{
	public:
		Asura_Vector_3 m_xPosition;
		float m_fOuterRadius;
		float m_fInnerRadius;

		inline void ChangePosition(Asura_Vector_3& xPosition)
		{
			m_xPosition = xPosition;
		};

		inline void ChangeOuterRadius(float fRadius)
		{
			m_fOuterRadius = fRadius;
		};

		inline void ChangeInnerRadius(float fRadius)
		{
			m_fInnerRadius = fRadius;
		}
};

// Templates
class Asura_Rumble_Template
{
    public:
	    float m_fMaxMagnitude;			//Modifier to all motors.
	    float m_fMinMagnitude;			//Modifier to all motors. Must be less than fMaxMagnitude. Will not affect RUMBLE_BASE_CONSTANT.
	    float m_fDecay;					//Modifier to all decays in each sequence    
        int   m_iEffectPlays;           //Number of times effect plays - no the same a the remaining effects below
        Asura_Platform_Rumble* m_pxRumble;
};

class Asura_HUDRumble_Template : public Asura_Rumble_Template
{
    public:
        void Create(Asura_Platform_Rumble* pxRumble, float fMaxMagnitude,
                    float MinMagnitude, float Decay,
                    int   iEffectPlays);
};

class Asura_WorldRumble_Template : public Asura_Rumble_Template
{
    public:
        void Create(Asura_Platform_Rumble* pxRumble, float fMaxMagnitude,
                    float MinMagnitude, float fDecay,
                    int   EffectPlays,  float fOuterRadius, float fInnerRadius);

		float m_fOuterRadius;
		float m_fInnerRadius;
};


// A function to allow projects to inspect their own settings or other project-specific arrangements to
//  determine if the given controllerID should rumble
typedef bool ProjectIsRumbleActiveFunction(const u_int uControllerID);

// A function to allow projects to get the position a given rumbler GUID should be treated as at, in the
//  event that the rumbler entity is not actually a renderable object
typedef bool ProjectGetRumblerPositionFunction(const Asura_Guid uGuid, Asura_Vector_3& xPosition);


class Asura_Rumble_System
{
	public:
		static bool IsInitialised() { return s_bIsInitialised; }

		static void Initialise();
		static void Shutdown();
		static void Reset();
		static void Update();

		static void PauseAllRumbling()
		{
			s_bPaused = true;
//			s_pxHUDRumbleList->Traverse(PauseRumbleCallback);
		};
		static void UnpauseAllRumbling()
		{
			s_bPaused = false;
//			s_pxHUDRumbleList->Traverse(UnpauseRumbleCallback);
//AG - 20/01/2003 - These two commented lines probably aren't needed, but are left here should I be wrong about this.
//			s_fCurrentTimeSlice = 0.0f;
//			Update2();
		};
		static void TogglePauseRumble()
		{
			s_bPaused = !s_bPaused;
		};

		static void PauseRumble(Asura_Handle iHandle);
		static void UnpauseRumble(Asura_Handle iHandle);
		static bool IsHUDRumbleActive(Asura_Handle iHandle);
		
        static Asura_Handle AddHUDRumble(Asura_Platform_Rumble* pxRumble, int iControllerNumber, 
                                         int iEffectPlays = 1,            float fMaxMagnitude = 1.0f, 
                                         float fMinMagnitude = 0.0f,      float fDecay = 1.0f, 
                                         bool bPause = false);

        static Asura_Handle AddHUDRumble(int   iControllerNumber, 
                                         const Asura_HUDRumble_Template& xRumbleTemplate,
                                         bool  bPause = false);

		static Asura_Handle AddWorldRumble(Asura_Platform_Rumble* pxRumble, const Asura_Vector_3& xPosition, 
                                           float fOuterRadius,              float fInnerRadius = 1.0f, 
                                           int   iEffectPlays = 1,          float fMaxMagnitude = 1.0f, 
                                           float fMinMagnitude = 0.0f,      float fDecay = 1.0f,
                                           bool bPause = false);

        static Asura_Handle AddWorldRumble(const Asura_Vector_3& xPosition, 
                                           Asura_WorldRumble_Template& xRumbleTemplate,
                                           bool  bPause = false);

		static void ChangeRumbleMaxMinRelativeMagnitude(Asura_Handle iHandle, float fNewMaxMagnitude);
		static void ChangeRumbleMaxMagnitude(Asura_Handle iHandle, float fNewMaxMagnitude);
		static void ChangeRumbleMinMagnitude(Asura_Handle iHandle, float fNewMinMagnitude);
		static void ChangeRumbleDecay(Asura_Handle iHandle, float fNewDecay);

		static void StopRumble(Asura_Handle iHandle);	// These functions are for removing rumble effects permanently.
		static void StopAllRumbling();					// Use Pause to remove rumbles temporarily.
        static void StopPlayerRumbles(int iControllerNumber);
		static void StopHUDRumbles();
		static void StopHUDRumble(Asura_Handle iHandle);	// These functions are for removing rumble effects permanently.
		
		static Asura_Platform_Rumble* ASURA_STANDARD_RUMBLE_EXPLOSION;
		static Asura_Platform_Rumble* ASURA_STANDARD_RUMBLE_SINGLESHOT;
		static Asura_Platform_Rumble* ASURA_STANDARD_RUMBLE_HEARTBEAT;
		static Asura_Platform_Rumble* ASURA_STANDARD_RUMBLE_HIT;

		static void AddPlayer(int iPlayerID, Asura_Guid uPlayer);
		static void RemovePlayer(int iPlayerID);
		static void ResetPlayers();

		// To allow project-specific ways to determine if a rumble should be played on a given controller
		static void SetProjectIsRumbleActiveFunction(ProjectIsRumbleActiveFunction* const pfnProjectIsRumbleActive)
			{ s_pfnProjectIsRumbleActive = pfnProjectIsRumbleActive; }

		// To cater for the situation (found on Simpsons) where the GUID which corresponds to the player is not actually
		// the renderable entity controlled by said player. This returns the position of the associated renderable.
		static void SetProjectGetRumblerPositionFunction(ProjectGetRumblerPositionFunction* const pfnProjectGetRumblerPosition)
			{ s_pfnProjectGetRumblerPosition = pfnProjectGetRumblerPosition; }

        // controller rumble multiplier: 0.f is off, 1.f is full
        static void SetPlayerRumbleMultiplier(u_int uPlayerID, float fMultiplier);
        static float GetPlayerRumbleMultiplier(u_int uPlayerID);

        static Asura_TrackingHandlePool<Asura_Platform_Rumble>* GetPlatformRumbleList(void) 
        {
            return s_pxDynamicPlatformRumbleList;
        }

		static float GetBaseEffect(float fTime, RUMBLE_BASE_EFFECT eBaseEffect, RUMBLE_ADJUST eAdjustment)
		{
			if (fTime > 1.0f)
			{
				Asura_Assert(0, "Time is too large for rumble effect (must be <= 1)");
				fTime = 1.0f;
			}
			else if (fTime < 0.0f)
			{
				Asura_Assert(0, "Attempted to get rumble function on negative time");
				fTime = 0.0f;
			}
			
			if (eAdjustment & RUMBLE_ADJUST_REVERSE)
			{
				fTime = 1.0f-fTime;
			}

			float result;
			switch (eBaseEffect)
			{
            default:
                Asura_Assert( false, "Invalid rumble base effect." );
                // Fall through to valid case so result gets initialised
                // and the compiler doesn't complain.

			case RUMBLE_BASE_CONSTANT:
				result = 1.0f;
				break;
			case RUMBLE_BASE_LINEAR:
				result = fTime;
				break;
			case RUMBLE_BASE_INCCURVE:
				result = (fTime * fTime);
				break;
			case RUMBLE_BASE_DECCURVE:
				result = Asura_Maths::Sine(Asura_Maths::Deg90 * fTime);
				break;
			case RUMBLE_BASE_SYMWAVE:
				result = (Asura_Maths::Cosine(Asura_Maths::Deg180 * (fTime-0.5f)) + 1.0f)/2.0f;
				break;
			}

			if (eAdjustment & RUMBLE_ADJUST_FLIP)
			{
				result = 1.0f-result;
			}
			
			return result;
		};

		static inline float GetCurrentMotorMagnitude(u_int iControllerIndex, u_int eMotorIndex)
		{
			return s_fCurrentMotorMagnitudes[iControllerIndex][eMotorIndex];
		};


	private:
		static Asura_TrackingHandlePool<Asura_HUDRumble>* s_pxHUDRumbleList;
		static Asura_TrackingHandlePool<Asura_WorldRumble>* s_pxWorldRumbleList;
		static Asura_TrackingHandlePool<Asura_Platform_Rumble>* s_pxDynamicPlatformRumbleList;

		static void Update2();

		static void AddRumbleCommon(Asura_ActiveRumble* pxNewRumble, Asura_Handle iNewHandle, 
                                    Asura_Platform_Rumble* pxActiveRumble, int iEffectPlays, 
			float fMaxMagnitude, float fMinMagnitude, float fDecay, bool bPause, ASURA_ACTIVE_RUMBLE_TYPE eType);
		
		static void Platform_Initialise();
		static void Platform_Shutdown();
		static void Platform_Reset();

		static void Project_Initialise();
		static void Project_Shutdown();
		static void Project_Reset();

		static void PlatformProject_Initialise();
		static void PlatformProject_Shutdown();
		static void PlatformProject_Reset();

		static void Platform_StopAllRumbles();

		static float s_fCurrentMotorMagnitudes[PLATFORM_NUM_CONTROLLERS][PLATFORM_NUM_RUMBLE_MOTORS];

		static void CheckHUDRumbleCallback(Asura_HUDRumble* pxThisRumble);
		static void CheckWorldRumbleCallback(Asura_WorldRumble* pxThisRumble);

		static void PauseRumbleCallback(Asura_WorldRumble* pxThisRumble);
		static void UnpauseRumbleCallback(Asura_WorldRumble* pxThisRumble);

		static void Platform_PlayCurrentRumbles();

		static float s_fCurrentTimeSlice;

		static bool s_bPaused;
		static bool s_bIsInitialised;

		static inline void SanityCheck()
		{
			for (int i=0; i<PLATFORM_NUM_CONTROLLERS; i++)
			{
				for (int j=0; j<PLATFORM_NUM_RUMBLE_MOTORS; j++)
				{
					if (s_fCurrentMotorMagnitudes[i][j] > 1.0f)
					{
						s_fCurrentMotorMagnitudes[i][j] = 1.0f;
					}
					else if (s_fCurrentMotorMagnitudes[i][j] < 0.0f)
					{
						s_fCurrentMotorMagnitudes[i][j] = 0.0f;
					}
				}
			}
		};

		static CurrentRumblers s_axCurrentUsers[PLATFORM_NUM_CONTROLLERS];
		static ProjectIsRumbleActiveFunction* s_pfnProjectIsRumbleActive;
		static ProjectGetRumblerPositionFunction* s_pfnProjectGetRumblerPosition;
};

#endif//__ASURA_RUMBLE_SYSTEM_H__
