//////////////////////////////////////////////////////////////////////////////
//
// Asura_CommandConsole_Remote.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Interface/Asura_CommandConsole_Remote.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

// Stdlib independent.
#include <wchar.h>

// Engine independent.
#include "Core/Asura_Checksum.h"
#include "Core/Asura_LinkList.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_Entity.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_Player.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// Macros to allow compile-time concatenation.
#define ACCESS_DENIED_MSG         L"Access denied."
#define CONSOLE_CMDNAME_REMOTECMD L"RemoteCmd"
#define CONSOLE_VARNAME_REMOTEPW  L"con_szRemoteCmdPassword"

// T Y P E S /////////////////////////////////////////////////////////////////

struct ClientCmdRecord
{
    Asura_UnicodeChar wszCmd[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];
    float             fTimeIssued;
};

struct Msg_ExecCmd
{
    u_char aucMD5[Asura_Checksum::CHECKSUM_LENGTH];
    // Followed by cmd string.
};

struct Msg_Nonce
{
    int iNonceValue;
};

struct ServerNonceRecord
{
    int        iNonceValue;
    Asura_Guid uPlayerGUID;
    float      fTimeCreated;
};

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int PASSWORD_BFR_SIZE = 16;
static const float TIMEOUT           = 2.0f;

// G L O B A L S /////////////////////////////////////////////////////////////

// Used by both client and server.
static Asura_UnicodeChar g_wszPassword[PASSWORD_BFR_SIZE];

// Client-side.
static ClientCmdRecord g_xPendingCmd;

// Server-side.
static Asura_LinkList<ServerNonceRecord> g_xNonceRecordList;
static Asura_Guid                        g_uPlayerGUIDForCmdResultMsgs;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static bool ConsoleCmd_RemoteCmd( const Asura_UnicodeChar* wszCmd, bool bInterrogate, bool bHelp );
static void GenerateMD5( u_char aucMD5[Asura_Checksum::CHECKSUM_LENGTH], int iNonceValue, const Asura_UnicodeChar* wszCmd );
static Asura_Guid GetLocalPlayerGUID( void );
static void ProcessExecCmdMsg( const Asura_Message* pxMsg );
static void ProcessExecCmdResultMsg( const Asura_Message* pxMsg );
static void ProcessNonce( const Asura_Message* pxMsg );
static void ProcessNonceRequest( const Asura_Message* pxMsg );
static void SendAccessDeniedMsg( Asura_Guid uPlayerGUID );
static void SendExecCmdResultMsg( const Asura_UnicodeChar* wszMsgFromConsole );
static void StoreCmdAndSendNonceReq( const Asura_UnicodeChar* wszCmd );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_CommandConsole_Remote::Initialise( void )
{
    Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_REMOTECMD, ConsoleCmd_RemoteCmd, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );   
    Asura_CommandConsole::AddVar( CONSOLE_VARNAME_REMOTEPW, g_wszPassword, PASSWORD_BFR_SIZE, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

    Reset();
}

void Asura_CommandConsole_Remote::ProcessMessage( const Asura_Message* pxMsg )
{
    switch( pxMsg->GetMsgID() )
    {
        case ASURA_MSG_COMMANDCONSOLE_REMOTE_NONCE_REQ:
            ProcessNonceRequest( pxMsg );
            break;

        case ASURA_MSG_COMMANDCONSOLE_REMOTE_NONCE:
            ProcessNonce( pxMsg );
            break;

        case ASURA_MSG_COMMANDCONSOLE_REMOTE_EXEC_CMD:
            ProcessExecCmdMsg( pxMsg );
            break;

        case ASURA_MSG_COMMANDCONSOLE_REMOTE_EXEC_CMD_RESULT:
            ProcessExecCmdResultMsg( pxMsg );
            break;

        default:
            Asura_Assert( false, "Why's this msg been routed to here then?" );
            break;
    }
}

void Asura_CommandConsole_Remote::Update( void )
{
    const float fCurTime = Asura_Timers::GetLocalUnscaledTime();
    
    if( Asura_Network::IsClient() && g_xPendingCmd.wszCmd[0] )
    {
        // If it's been ages since we started trying to issue the command,
        // give up and notify the user. Also disregard commands with
        // timestamps later than the current time, in case timers have been
        // reset or something.
        bool bRemove = true;
        if( fCurTime > g_xPendingCmd.fTimeIssued + TIMEOUT )
        {
            Asura_CommandConsole::Printf( L"Remote command execution timed out." );
        }
        else if( g_xPendingCmd.fTimeIssued <= fCurTime )
        {
            bRemove = false;
        }

		if (bRemove)
		{
			g_xPendingCmd.wszCmd[0] = '\0';
		}
    }
    
    if( Asura_Network::IsServer() )
    {
        // Remove expired nonces.
        for( Asura_ListElt<ServerNonceRecord>* pxCurElt = g_xNonceRecordList.Front(); pxCurElt; )
        {
            Asura_ListElt<ServerNonceRecord>* pxNext = pxCurElt->Next();

            // Remove if timed out or if newer than current time. The reason
            // for the latter is the same as in the client case above.
            if( fCurTime > pxCurElt->Data().fTimeCreated + TIMEOUT || pxCurElt->Data().fTimeCreated > fCurTime )
            {
                g_xNonceRecordList.Remove( pxCurElt );
            }

            pxCurElt = pxNext;
        }
    }
}

void Asura_CommandConsole_Remote::Reset( void )
{
    g_xPendingCmd.wszCmd[0] = '\0';
    g_xNonceRecordList.Clear();
}

void Asura_CommandConsole_Remote::Shutdown( void )
{
    Reset();
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static bool ConsoleCmd_RemoteCmd( const Asura_UnicodeChar* wszCmd, bool bInterrogate, bool bHelp )
{
    if( bInterrogate || bHelp )
    {
        Asura_CommandConsole::Printf( L"Executes the command on the server, provided you have set the" );
        Asura_CommandConsole::Printf( CONSOLE_VARNAME_REMOTEPW L" variable first." );
        return true;
    }
    else
    {
        if( Asura_Network::IsClient() )
        {
            if( Asura_Network::IsExternalNetworkActive() && Asura_Network::IsConnectionOkay() )
            {
                if( g_wszPassword[0] )
                {
                    if( !g_xPendingCmd.wszCmd[0] )
                    {
                        StoreCmdAndSendNonceReq( wszCmd );
                        return true;
                    }
                    else
                    {
                        Asura_CommandConsole::Printf( L"Still executing previous command." );
                    }
                }
                else
                {
                    Asura_CommandConsole::Printf( L"Need to set " CONSOLE_VARNAME_REMOTEPW L" first." );
                }
            }
            else
            {
                Asura_CommandConsole::Printf( L"Not connected." );
            }
        }
        else
        {
            Asura_CommandConsole::Printf( L"This command can only be used from a client." );
        }
    }
    return false;
}

static void GenerateMD5( u_char* pucMD5, int iNonceValue, const Asura_UnicodeChar* wszCmd )
{
    const u_int uBfrSize = 11 + (PASSWORD_BFR_SIZE - 1) + (ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE - 1) + 1;
    Asura_UnicodeChar wszAuthString[uBfrSize];

    Asura_PrintfToBfr( wszAuthString, L"%i%s%s", uBfrSize, true, iNonceValue, g_wszPassword, wszCmd );

    Asura_Checksum::EncodeFromBuffer( reinterpret_cast<const u_char*>(wszAuthString), wcslen(wszAuthString) * sizeof(Asura_UnicodeChar), pucMD5 );
}

static Asura_Guid GetLocalPlayerGUID( void )
{
    Asura_Guid uPlayerGUID = ASURA_GUID_UNREGISTERED;

    // If there's more than one on this machine, just use the first. We can't
    // distinguish between which one of them's using the console so they might
    // as well both be authenticated.
    Asura_Hashed_List_It<Asura_Entity_Player> xIt( Asura_Entity_Player::GetList() );
    if( !xIt.Done() )
    {
        uPlayerGUID = xIt.GetCurrent()->GetGuid();
    }

    return uPlayerGUID;
}

static void ProcessExecCmdMsg( const Asura_Message* pxMsg )
{
    // This should never happen since only the client message handler accepts
    // this message.
    Asura_Assert( Asura_Network::IsServer(), "" );

    const Msg_ExecCmd* pxMsgData = reinterpret_cast<const Msg_ExecCmd*>( pxMsg->GetData() );
    if( !pxMsgData ) { return; }

    const Asura_Guid uPlayerGUID = pxMsg->From;

    // Find nonce record corresponding to this player.
    Asura_ListElt<ServerNonceRecord>* pxElt = g_xNonceRecordList.Front();
    for( ; pxElt; pxElt = pxElt->Next() )
    {
        if( uPlayerGUID == pxElt->Data().uPlayerGUID ) { break; }
    }
    if( !pxElt ) { return; }

    // Only process cmd if player is still connected.
	// (MH - 30.09.2003) also only if player has a name - in Dredd, players 
	// only get names when they are actually in the game. This prevents players
	// from changing maps repeatedly, as once the server has changed the map 
	// the player will not longer be in the game, but is still actually connected.
    if( Asura_Network::IsPlayerConnected(uPlayerGUID, false) && Asura_Network::GetPlayerName(uPlayerGUID) )
    {
        const Asura_UnicodeChar* wszCmd = reinterpret_cast<const Asura_UnicodeChar*>( pxMsgData + 1 );

        u_char aucExpectedMD5[Asura_Checksum::CHECKSUM_LENGTH];
        GenerateMD5( aucExpectedMD5, pxElt->Data().iNonceValue, wszCmd );

        if( Asura_Checksum::Compare(aucExpectedMD5, pxMsgData->aucMD5) )
        {
            // Ok, execute cmd.
            g_uPlayerGUIDForCmdResultMsgs = uPlayerGUID;
            Asura_CommandConsole::Execute( wszCmd, NULL, SendExecCmdResultMsg );

            // Echo it to server console, unless we're also the client that
            // will be receiving the result messages.
            if( 0 != Asura_Network::GetMachineForPlayer(uPlayerGUID, false) )
            {
                Asura_CommandConsole::Printf( L"[R %s] %s", Asura_Network::GetPlayerName(uPlayerGUID), wszCmd );
            }
        }
        else
        {
            SendAccessDeniedMsg( uPlayerGUID );
        }
    }

    // Remove from nonce records.
    g_xNonceRecordList.Remove( pxElt );
}

static void ProcessExecCmdResultMsg( const Asura_Message* pxMsg )
{
    // This should never happen since only the client message handler accepts
    // this message.
    Asura_Assert( Asura_Network::IsClient(), "" );

    // Get result msg.
    const Asura_UnicodeChar* wszResultStr = reinterpret_cast<const Asura_UnicodeChar*>( pxMsg->GetData() );
    if( !wszResultStr ) { return; }

    // And just echo it to local console.
    Asura_CommandConsole::Printf( L"[R] %s", wszResultStr );

    // Clear pending cmd record. We may receive this message in response to
    // the initial nonce request if remote command access is disabled.
    g_xPendingCmd.wszCmd[0] = '\0';
}

static void ProcessNonce( const Asura_Message* pxMsg )
{
    // This should never happen since only the client message handler accepts
    // this message.
    Asura_Assert( Asura_Network::IsClient(), "" );

    // Disregard if we don't have a pending cmd.
    if( !g_xPendingCmd.wszCmd[0] ) { return; }

    const Msg_Nonce* pxData = reinterpret_cast<const Msg_Nonce*>( pxMsg->GetData() );
    if( !pxData ) { return; }

    const u_int uReplyDataSize = sizeof(Msg_ExecCmd) + (wcslen(g_xPendingCmd.wszCmd) + 1) * sizeof(Asura_UnicodeChar);
    Asura_Message* pxReply = Asura_Network::CreateMessage( ASURA_MSG_COMMANDCONSOLE_REMOTE_EXEC_CMD, uReplyDataSize );
    if( pxReply )
    {
        pxReply->To   = ASURA_GUID_SERVER;
        pxReply->From = GetLocalPlayerGUID();

        Msg_ExecCmd* pxReplyData = reinterpret_cast<Msg_ExecCmd*>( pxReply->GetData() );
        GenerateMD5( pxReplyData->aucMD5, pxData->iNonceValue, g_xPendingCmd.wszCmd );
        wcscpy( reinterpret_cast<Asura_UnicodeChar*>(pxReplyData + 1), g_xPendingCmd.wszCmd );

        pxReply->UseGuaranteedDelivery();
        Asura_Network::Send( pxReply );
    }
    
    // Clear the pending cmd record.
    g_xPendingCmd.wszCmd[0] = '\0';
}

static void ProcessNonceRequest( const Asura_Message* pxMsg )
{
    // This should never happen since only the server message handler accepts
    // this message.
    Asura_Assert( Asura_Network::IsServer(), "" );

    // Who's it from?
    const Asura_Guid uPlayerGUID = pxMsg->From;
    
    // Ignore requests if there's no password set since that's how you
    // disable remote command execution.
    if( !g_wszPassword[0] )
    {
        SendAccessDeniedMsg( uPlayerGUID );
        return;
    }

    // If we've already got a request from them, we'll overwrite it with a
    // new one. This can only happen if somehow our response didn't reach the
    // client and so they've sent a new request.
    Asura_ListElt<ServerNonceRecord>* pxElt = g_xNonceRecordList.Front();
    for( ; pxElt; pxElt = pxElt->Next() )
    {
        if( uPlayerGUID == pxElt->Data().uPlayerGUID ) { break; }
    }

    // Check the GUID is valid, and that the player has a name.
    if( (!Asura_Network::IsPlayerConnected(uPlayerGUID, false)) || (!Asura_Network::GetPlayerName(uPlayerGUID)) )
    {
        if( pxElt ) { g_xNonceRecordList.Remove( pxElt ); }
        return;
    }

    // If not, create a new entry in the list.
    if( !pxElt )
    {
        pxElt = g_xNonceRecordList.Push_Back( ServerNonceRecord() );
    }

    const float fPhase = Asura_Timers::GetLocalUnscaledTime();

    // Generate nonce.
    int iNonce = Asura_Random::GetIntU();
    // Since our unseeded random number generator won't necessarily be all
    // that random, include the time as well. Simply use the int
    // representation of the (floating point) phase since minor fluctuations
    // in load times, etc., should guarantee at least a few bits will be
    // different.
    iNonce += reinterpret_cast<const int&>( fPhase );

    // Fill in list entry.
    ServerNonceRecord& xRecord = pxElt->Data();
    xRecord.iNonceValue  = iNonce;
    xRecord.uPlayerGUID  = uPlayerGUID;
    xRecord.fTimeCreated = fPhase;
    
    // Send reply.
    Asura_Message* pxReply = Asura_Network::CreateMessage( ASURA_MSG_COMMANDCONSOLE_REMOTE_NONCE, sizeof(Msg_Nonce) );
    if( pxReply )
    {
        pxReply->To   = ASURA_GUID_PLAYERS_CLIENTMACHINE;
        pxReply->From = uPlayerGUID;

        Msg_Nonce* pxReplyData = reinterpret_cast<Msg_Nonce*>( pxReply->GetData() );
        pxReplyData->iNonceValue = iNonce;

        pxReply->UseGuaranteedDelivery();
        Asura_Network::Send( pxReply );
    }
}

static void SendAccessDeniedMsg( Asura_Guid uPlayerGUID )
{
    g_uPlayerGUIDForCmdResultMsgs = uPlayerGUID;
    SendExecCmdResultMsg( ACCESS_DENIED_MSG );

    // Report failed access attempt in server console, unless we're
    // also the client that will be receiving the result messages.
    if( 0 != Asura_Network::GetMachineForPlayer(uPlayerGUID, false) )
    {
        Asura_CommandConsole::Printf( L"[R %s] " ACCESS_DENIED_MSG, Asura_Network::GetPlayerName(uPlayerGUID) );
    }
}

static void SendExecCmdResultMsg( const Asura_UnicodeChar* wszMsgFromConsole )
{
    const u_int uStrSize = (wcslen(wszMsgFromConsole) + 1) * sizeof(Asura_UnicodeChar);
    Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_COMMANDCONSOLE_REMOTE_EXEC_CMD_RESULT, uStrSize );
    if( pxMsg )
    {
        pxMsg->To   = ASURA_GUID_PLAYERS_CLIENTMACHINE;
        pxMsg->From = g_uPlayerGUIDForCmdResultMsgs;

        Asura_UnicodeChar* wszData = reinterpret_cast<Asura_UnicodeChar*>( pxMsg->GetData() );
        wcscpy( wszData, wszMsgFromConsole );

        pxMsg->UseGuaranteedDelivery();
        Asura_Network::Send( pxMsg );
    }
}

static void StoreCmdAndSendNonceReq( const Asura_UnicodeChar* wszCmd )
{
    // This shouldn't happen since the command console code will limit the
    // input at under this length anyway.
    Asura_Assert( wcslen(wszCmd) < ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE, "" );

    // Store.
    wcscpy( g_xPendingCmd.wszCmd, wszCmd );
    g_xPendingCmd.fTimeIssued = Asura_Timers::GetLocalUnscaledTime();

    // Send nonce request.
    Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_COMMANDCONSOLE_REMOTE_NONCE_REQ );
    if( pxMsg )
    {
        pxMsg->To = ASURA_GUID_SERVER;
        pxMsg->From = GetLocalPlayerGUID();
        pxMsg->UseGuaranteedDelivery();
        Asura_Network::Send( pxMsg );
    }
}

#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
