
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Table_Dialog.h
// Author : Andy Tate
// Date   : 26-Apr-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_TABLE_DIALOG_H__
#define __ASURA_TABLE_DIALOG_H__

#if defined(ASURA_TOOLS)

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Interface/Asura_Dialog.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Table_Dialog;
class Asura_Scrollbar_Gadget;


////////////////////////////////////////////////////////////////////////////////
// Defines...
////////////////////////////////////////////////////////////////////////////////

#define ASURA_TABLE_DIALOG_ROW_SELECTED_COLOUR			0x00FFFF00
#define ASURA_TABLE_DIALOG_ROW_HIGHLIGHT_COLOUR			0x0000FF00
#define ASURA_TABLE_DIALOG_ROW_SORT_HIGHLIGHT_COLOUR	0x0000FFFF

#define ASURA_TABLE_DIALOG_COLUMN_SELECTED_COLOUR		0x00FFFF00
#define ASURA_TABLE_DIALOG_COLUMN_HIGHLIGHT_COLOUR		0x0000FF00
#define ASURA_TABLE_DIALOG_COLUMN_SORT_HIGHLIGHT_COLOUR	0x0000FFFF


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_TableCell_Dialog
// Author       : Andy Tate
// Date         : 27-Apr-2007
//
// Description  : Class to handle the displaying of a particular cell in the 
//				  table dialog.
////////////////////////////////////////////////////////////////////////////////

class Asura_TableCell_Dialog : public Asura_Panel_Dialog
{
public:
	Asura_TableCell_Dialog(Asura_Table_Dialog& xTableDialog, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const int iRow, const int iColumn, const u_int uColour = 0x00000000)
    : PARENT(xOffset, xMinCellSize, uColour)
	, m_xTableDialog(xTableDialog)
    , m_iRow(iRow)
    , m_iColumn(iColumn)
	, m_bClickToSort(false)
    {
    }

	virtual int	Compare(const Asura_TableCell_Dialog& xOther) const	= 0;
	virtual bool ShouldClip( void ) const	{ return !CanRenderOutsideOfParentSubwindow(); }

	// Accessor functions...
	bool	IsRowHeading(void) const	{ return(m_iColumn < 0); }
	bool	IsColumnHeading(void) const	{ return(m_iRow < 0); }
	bool	IsHeading(void) const		{ return((m_iRow < 0) || (m_iColumn < 0)); }

	void	SetClickToSort(const bool bSet)	{ m_bClickToSort = bSet; }
	bool	IsSortedOnClick(void) const		{ return(m_bClickToSort); }

	int GetRow() const { return m_iRow; }
	int GetColumn() const { return m_iColumn; }

protected:
	Asura_Table_Dialog&	m_xTableDialog;		// A reference to the parent table-dialogue.
	int					m_iRow;				// The index of the row this cell belongs to (-1 meaning it a column-heading).
	int					m_iColumn;			// The index of the column this cell belongs to (-1 meaning its a row-heading).
	bool				m_bClickToSort;		// Flag to say that clicking on this cell will force a sort (only applicable for headings).

private:
	typedef Asura_Panel_Dialog PARENT;

	Asura_TableCell_Dialog(void);
	Asura_TableCell_Dialog(const Asura_TableCell_Dialog& xOther);

	Asura_TableCell_Dialog&	operator=(const Asura_TableCell_Dialog& xOther);
};

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_TableCell_TextDialog
// Author       : Andy Tate
// Date         : 23-May-2008
//
// Description  : Class used to display a cell in a table as simply (non-selectable)
//				  text - can be used for row/column headings.
////////////////////////////////////////////////////////////////////////////////

class Asura_TableCell_TextDialog : public Asura_TableCell_Dialog
{
public:
	// Public functions...
	static Asura_TableCell_TextDialog*	Create(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const int iRow, const int iColumn);

	// Overridden functions...
	int Compare(const Asura_TableCell_Dialog& xOther) const;

	// Accessor functions...
	const Asura_String&	GetName(void) const	{ return(m_xName); }

protected:
	// Protected functions...
	Asura_TableCell_TextDialog(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const int iRow, const int iColumn)
	: PARENT(xTableDialog, xOffset, xMinCellSize, iRow, iColumn)
	, m_xName(szName)
	{
	}

private:
	// Private typedefs...
	typedef Asura_TableCell_Dialog PARENT;

	// Private variables...
	Asura_String	m_xName;	// We have to keep our own copy of our name because if we set it on the base Asura_Dialog it'll render underneath any dialogues we have.

	// Private functions...
	Asura_TableCell_TextDialog(void);
	Asura_TableCell_TextDialog(const Asura_TableCell_TextDialog& xOther);

	Asura_TableCell_TextDialog&	operator=(const Asura_TableCell_TextDialog& xOther);

};   // End of class 'Asura_TableCell_TextDialog'.


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_TableColumnHeading_Dialog
// Author       : Andy Tate
// Date         : 27-Apr-2007
//
// Description  : Class to handle the displaying of an individual column in a 
//				  table.
////////////////////////////////////////////////////////////////////////////////

class Asura_TableColumnHeading_Dialog : public Asura_TableCell_TextDialog
{
public:
	Asura_TableColumnHeading_Dialog(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const u_int uColumn)
    : PARENT(xTableDialog, szName, xOffset, xMinCellSize, -1, uColumn)
    {
    }

	// So we can change the text colour.
	virtual unsigned GetColour( void ) const;

private:
	typedef Asura_TableCell_TextDialog PARENT;

	Asura_TableColumnHeading_Dialog(void);
	Asura_TableColumnHeading_Dialog(const Asura_TableColumnHeading_Dialog& xOther);
	Asura_TableColumnHeading_Dialog& operator=(const Asura_TableColumnHeading_Dialog& xOther);
};

//////////////////////////////////////////////////////////////////////////////
//
// Asura_TableColumnHeading_StandardDialog
//
// Standard column heading dialogue. Simply displays the name of the column.
//
//////////////////////////////////////////////////////////////////////////////

class Asura_TableColumnHeading_StandardDialog : public Asura_TableColumnHeading_Dialog
{
public:
    static Asura_TableColumnHeading_StandardDialog*	Create(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const u_int uColumn);

protected:
    Asura_TableColumnHeading_StandardDialog(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const u_int uColumn);


private:
	typedef Asura_TableColumnHeading_Dialog PARENT;

	Asura_TableColumnHeading_StandardDialog(void);
	Asura_TableColumnHeading_StandardDialog(const Asura_TableColumnHeading_Dialog& xOther);
    Asura_TableColumnHeading_StandardDialog& operator=(const Asura_TableColumnHeading_StandardDialog& xOther);
};

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_TableRowHeading_Dialog
// Author       : Andy Tate
// Date         : 27-Apr-2007
//
// Description  : Class to handle the displaying of an individual row in a 
//				  table.
////////////////////////////////////////////////////////////////////////////////

class Asura_TableRowHeading_Dialog : public Asura_TableCell_TextDialog
{
public:
	Asura_TableRowHeading_Dialog(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const u_int uRow)
    : PARENT(xTableDialog, szName, xOffset, xMinCellSize, uRow, -1)
    {
    }

	// So we can change the text colour.
	virtual unsigned GetColour( void ) const;

private:
	typedef Asura_TableCell_TextDialog PARENT;

	Asura_TableRowHeading_Dialog(void);
	Asura_TableRowHeading_Dialog(const Asura_TableRowHeading_Dialog& xOther);
	Asura_TableRowHeading_Dialog& operator=(const Asura_TableRowHeading_Dialog& xOther);
};

//////////////////////////////////////////////////////////////////////////////
//
// Asura_TableRowHeading_StandardDialog
//
// Standard row heading dialogue. Simply displays the name of the row.
//
//////////////////////////////////////////////////////////////////////////////

class Asura_TableRowHeading_StandardDialog : public Asura_TableRowHeading_Dialog
{
public:
    static Asura_TableRowHeading_StandardDialog*	Create(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const u_int uRow);

protected:
    Asura_TableRowHeading_StandardDialog(Asura_Table_Dialog& xTableDialog, const char* szName, const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize, const u_int uRow);


private:
    typedef Asura_TableRowHeading_Dialog PARENT;

    Asura_TableRowHeading_StandardDialog(void);
    Asura_TableRowHeading_StandardDialog(const Asura_TableRowHeading_StandardDialog& xOther);
    Asura_TableRowHeading_StandardDialog& operator=(const Asura_TableRowHeading_StandardDialog& xOther);
};

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Table_Dialog
// Author       : Andy Tate
// Date         : 26-Apr-2007
//
// Description  : Class to handle displaying of a 2D table with an arbitrary
//				  number of rows and columns.
////////////////////////////////////////////////////////////////////////////////


class Asura_Table_Dialog : public Asura_Dialog
{
public:
	// Public typedefs...
	typedef Asura_TableCell_Dialog* CreateCell( Asura_Table_Dialog& xTableDialog,
												const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize,
                                                const u_int uRow, const u_int uColumn, void* const pxUserData );
    typedef Asura_TableColumnHeading_Dialog* CreateColumnHeading( Asura_Table_Dialog& xTableDialog, const char* szName,
                                                                  const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize,
																  const u_int uColumn,
                                                                  void* const pxUserData );
    typedef Asura_TableRowHeading_Dialog* CreateRowHeading( Asura_Table_Dialog& xTableDialog, const char* szName,
                                                            const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize,
															const u_int uRow,
                                                            void* const pxUserData );

	// Public functions...
	Asura_Table_Dialog(const Asura_Vector_2& xOffset, const float fWidth, const float fHeight, const bool bResizeAsNecessary);
	virtual ~Asura_Table_Dialog();
	
	// Manipulation functions...
	void CreateTable( const Asura_Vector_2& xMinCellSize, const u_int uNumRows, const u_int uNumColumns,
                      CreateCell* pfnCreateCell, void* const pxCreateCellUserData,
                      const char* const* pszRowNames = NULL, const char* const* pszColumnNames = NULL,
                      CreateRowHeading* pfnCreateRowHeading = NULL, void* const pxCreateRowHeadingUserData = NULL,
                      CreateColumnHeading* pfnCreateColumnHeading = NULL, void* const pxCreateColumnHeadingUserData = NULL );
	void CreateTable( const Asura_Vector_2& xMinCellSize, const u_int uNumRows, const u_int uNumColumns,
                      CreateCell* pfnCreateCell, void* const pxCreateCellUserData,
                      const Asura_String* pxRowNames = NULL, const Asura_String* pxColumnNames = NULL,  
                      CreateRowHeading* pfnCreateRowHeading = NULL, void* const pxCreateRowHeadingUserData = NULL,
                      CreateColumnHeading* pfnCreateColumnHeading = NULL, void* const pxCreateColumnHeadingUserData = NULL );
	void DestroyTable(void);

	// Sort functions...
	void	EnableClickRowHeadingsToSort(void)		{ SetClickToSortForAllRowHeadings(true); SetClickToSortForAllColumnHeadings(false); }
	void	EnableClickColumnHeadingsToSort(void)	{ SetClickToSortForAllRowHeadings(false); SetClickToSortForAllColumnHeadings(true); }
	void	DisableClickHeadingsToSort(void)		{ SetClickToSortForAllRowHeadings(false); SetClickToSortForAllColumnHeadings(false); }
	void    GetSortParameters( u_int& uIndex, bool& bBackwards ) { uIndex = m_uSortIndex; bBackwards = m_bSortBackwards; }

	void	SortRowsUsingHeadings(const bool bBackwards);
	void	SortRowsUsingColumn(const u_int uColumn, const bool bBackwards);

	void	SortColumnsUsingHeadings(const bool bBackwards);
	void	SortColumnsUsingRow(const u_int uRow, const bool bBackwards);

	// Accessor functions...
	u_int	GetNumColumns(void) const	{ return(m_uNumColumns); }
	u_int	GetNumRows(void)    const	{ return(m_uNumRows); }

    // Returns the size of the table, which can change if this table can resize itself, or will be the maximum size if it can't.
    Asura_Vector_2 GetTableSize() const { return m_xViewSizeWithoutScrollBars; } 

	Asura_TableCell_Dialog* GetCell( const u_int uRow, const u_int uColumn ) const;
	Asura_TableRowHeading_Dialog* GetRowHeading( const u_int uRow, const bool bCheckCells = false ) const;
	Asura_TableColumnHeading_Dialog* GetColumnHeading( const u_int uColumn, const bool bCheckCells = false ) const;

	void	DisableRowSelection(void)					{ m_bRowSelectionEnabled = false; }
	void	EnableRowSelection(const bool bHeadingOnly)	{ m_bRowSelectionEnabled = true; m_bSelectRowFromHeadingOnly = bHeadingOnly; }
	void	SetCurrentlySelectedRow(const int iRow);
	int		GetCurrentlySelectedRow(void) const			{ return(m_iCurrentlySelectedRow); }
	int		GetRowUnderCursor(void) const				{ return(m_iRowUnderCursor); }
	int		GetRowHeadingUnderCursor(void) const		{ return(m_iRowHeadingUnderCursor); }
	void	SetRowSelectionFlag( bool* pbFlag )			{ m_pbRowSelectionFlag = pbFlag; }

	void	DisableColumnSelection(void)					{ m_bColumnSelectionEnabled = false; }
	void	EnableColumnSelection(const bool bHeadingOnly)	{ m_bColumnSelectionEnabled = true; m_bSelectColumnFromHeadingOnly = bHeadingOnly; }
	void	SetCurrentlySelectedColumn(const int iColumn)	{ m_iCurrentlySelectedColumn = iColumn; }
	int		GetCurrentlySelectedColumn(void) const			{ return(m_iCurrentlySelectedColumn); }
	int		GetColumnUnderCursor(void) const				{ return(m_iColumnUnderCursor); }
	int		GetColumnHeadingUnderCursor(void) const			{ return(m_iColumnHeadingUnderCursor); }
	void	SetColumnSelectionFlag( bool* pbFlag )			{ m_pbColumnSelectionFlag = pbFlag; }

	// PAZ : used to enable / disable rows, columns and cells in the table
	void SetIsRowEnabledCallBack( bool (*pfnIsRowEnabled)( const u_int uRow, void* const pxUserData ), void* const pxUserData )
	{
		m_pfnIsRowEnabled = ( pfnIsRowEnabled == NULL ) ? DefaultIsRowEnabled : pfnIsRowEnabled;
		m_pxIsRowEnabledUserData = pxUserData;
	}
	void SetIsColumnEnabledCallBack( bool (*pfnIsColumnEnabled)( const u_int uColumn, void* const pxUserData ), void* const pxUserData )
	{
		m_pfnIsColumnEnabled = ( pfnIsColumnEnabled == NULL ) ? DefaultIsColumnEnabled : pfnIsColumnEnabled;
		m_pxIsColumnEnabledUserData = pxUserData;
	}
	void SetIsCellEnabledCallBack( bool (*pfnIsCellEnabled)( const u_int uRow, const u_int uColumn, void* const pxUserData ), void* const pxUserData )
	{
		m_pfnIsCellEnabled = ( pfnIsCellEnabled == NULL ) ? DefaultIsCellEnabled : pfnIsCellEnabled;
		m_pxIsCellEnabledUserData = pxUserData;
	}

	void SetGetColourCallBack( unsigned (*pfnGetColour)( const Asura_Table_Dialog& xTableDialog, void* const pxUserData, const Asura_Dialog* const pxDialog, const unsigned uColour, const u_int uRow, const u_int uColumn ), void* const pxUserData )
	{
		m_pfnGetColour = ( pfnGetColour == NULL ) ? DefaultGetColour : pfnGetColour;
		m_pxGetColourUserData = pxUserData;
	}

	// enable / disable
	bool AreLinesEnabled( void )							const	{ return( m_bEnableLines ); }
	void EnableLines( const bool bEnable )							{ m_bEnableLines = bEnable; }

	// the spacing between cells
	Asura_Vector_2 GetCellSpacing( void )					const	{ return( m_xCellSpacing ); }
	void SetCellSpacing( const Asura_Vector_2 xSpacing )			{ m_xCellSpacing = xSpacing; }
	Asura_Vector_2 GetHeadingsSpacing( void )				const	{ return( m_xHeadingsSpacing ); }
	void SetHeadingsSpacing( const Asura_Vector_2 xSpacing )		{ m_xHeadingsSpacing = xSpacing; }

	// cell / headings offsets - useful for adjusting due to drawing the lines
	Asura_Vector_2 GetCellOffset( void )					const	{ return( m_xCellOffset ); }
	void SetCellOffset( const Asura_Vector_2 xOffset )				{ m_xCellOffset = xOffset; }
	Asura_Vector_2 GetRowHeadingOffset( void )				const	{ return( m_xRowHeadingOffset ); }
	void SetRowHeadingOffset( const Asura_Vector_2 xOffset )		{ m_xRowHeadingOffset = xOffset; }
	Asura_Vector_2 GetColumnHeadingOffset( void )			const	{ return( m_xColumnHeadingOffset ); }
	void SetColumnHeadingOffset( const Asura_Vector_2 xOffset )		{ m_xColumnHeadingOffset = xOffset; }

    const Asura_Vector_2& GetScrollBarValue() const { return m_xScrollBarValue; }
    void SetScrollBarValue( const Asura_Vector_2& xValue ) { m_xScrollBarValue = xValue; }

    void UpdateAll( void );


	//Allows control to prepare for the overlay being resized
	virtual void StartResize(u_int uResizeFlags);
	//Handles user resizing the overlay. xDelta is the cumulative delta from the start of the operation
	virtual void UpdateResize(u_int uResizeFlags, const Asura_Vector_2& xDelta);
	//Gets the direction(s) in which this dialog can be resized
	virtual u_int GetAllowedResizeFlags() const;

	float GetRowHeight(u_int uRow) { return m_pfRowHeights[uRow]; }

protected:
	// Protected variables...

	// Protected functions...
	virtual int  UpdateList(const Asura_Vector_2& xOrigin, int iPriority);
	virtual void Update(const Asura_Vector_2& xOrigin);
    virtual void PerFrameUpdate();
	virtual void Render(const Asura_Vector_2& xOrigin);
	virtual void OnHiddenStateChange();
	virtual bool IsPointInDialog(const Asura_Vector_2& xOrigin,const Asura_Vector_2& xPos) const;
    virtual void UpdateSize();

	// Update functions...
	void	UpdateRowAndColumnDimensions( const Asura_Vector_2* pxOrigin );
	void	UpdateRowAndColumnHeadingDimensions( const Asura_Vector_2* pxOrigin );
	void	UpdateRowAndColumnHeadingPositions( const Asura_Vector_2* pxOrigin );
	void	UpdateCellPositions( const Asura_Vector_2* pxOrigin );
	void	UpdateScrollbars( const Asura_Vector_2* pxOrigin );
	void	UpdateKeys( const Asura_Vector_2* pxOrigin );

	// Sort functions...
    static bool ShouldSortBackwards() { return s_bSortBackwards; }

	static int	SortByRowHeadingDialogues(const void* pElement1, const void* pElement2);
	static int	SortByCellDialoguesInRow(const void* pElement1, const void* pElement2);
	static int	SortByColumnHeadingDialogues(const void* pElement1, const void* pElement2);
	static int	SortByCellDialoguesInColumn(const void* pElement1, const void* pElement2);

	void	SetClickToSortForAllRowHeadings(const bool bSet);
	void	SetClickToSortForAllColumnHeadings(const bool bSet);

	// Accessor functions...
	static bool IsInView(const float fOffset, const float fSize, const float fMin, const float fMax)
	{
		return((fOffset >= (fMin - fSize)) && (fOffset < fMax));
	}


private:
	// Private typedefs...
	typedef Asura_Dialog PARENT;

	// Private variables...
	Asura_Vector_2						m_xTableSize;					// The size of the whole table.
	Asura_Vector_2						m_xViewSize;					// The size available to the view itself.
    Asura_Vector_2                      m_xMaxViewSize;                 // The max size available to the view itself.
	Asura_Vector_2						m_xViewSizeWithoutScrollBars;	// The size available to the view itself, without the scroll bars
	Asura_Vector_2						m_xMaxHeadingSize;				// The size taken up horizontally by the row headings and vertically by the column headings.
	Asura_Vector_2						m_xScrollBarValue;				// The current value for the scroll-bar.
    bool                                m_bResizeAsNecessary;           // Whether the table will always display at the max view size, or resize if it can be smaller.

    enum ScrollBarOrientations
    {
        SBO_HORIZONTAL,
        SBO_VERTICAL,
        NUM_SCROLL_BAR_ORIENTATIONS
    };
	Asura_Scrollbar_Gadget*				m_apxScrollBarGadgets[ NUM_SCROLL_BAR_ORIENTATIONS ];   // A pointer to the scroll-bar gadgets.	0 = horizontal, 1 = vertical
	bool								m_bScrollBarNeeded[ NUM_SCROLL_BAR_ORIENTATIONS ];	    // Are the scroll bars required?
	Asura_TableCell_Dialog***			m_pppxCellDialogues;			// A pointer to an array of array of cell dialogs.
	Asura_TableRowHeading_Dialog**		m_ppxRowHeadingDialogues;		// An array of pointers to the dialogs to be used for the row headings.
	Asura_TableColumnHeading_Dialog**	m_ppxColumnHeadingDialogues;	// An array of pointers to the dialogs to be used for the column headings.

	u_int								m_uNumRows;						// The number of rows in the table.
	float*								m_pfRowHeights;					// Array of heights for each row.
	u_int*								m_puRowDisplayIndices;			// A pointer to the display order to be used for the rows.
	bool								m_bRowSelectionEnabled;			// Flag to allow the current row to be selected (and hence highlighted).
	bool								m_bSelectRowFromHeadingOnly;	// Flay to allow a row to be selected from just the heading OR anywhere along that row.
	int									m_iRowUnderCursor;				// The row under the current cursor position.
	int									m_iRowHeadingUnderCursor;		// The row-heading under the current cursor position.
	int									m_iCurrentlySelectedRow;		// The row we have currently selected.
	bool*								m_pbRowSelectionFlag;			// Set to true when the row selected is changed;

	u_int								m_uNumColumns;					// The number of columns in the table.
	float*								m_pfColumnWidths;				// Array of widths for each column.
	u_int*								m_puColumnDisplayIndices;		// A pointer to the display order to be used for the columns.
	bool								m_bColumnSelectionEnabled;		// Flag to allow the current column to be selected (and hence highlighted).
	bool								m_bSelectColumnFromHeadingOnly;	// Flay to allow a column to be selected from just the heading OR anywhere along that column.
	int									m_iColumnUnderCursor;			// The column under the current cursor position.
	int									m_iColumnHeadingUnderCursor;	// The column-heading under the current cursor position.
	int									m_iCurrentlySelectedColumn;		// The column we have currently selected.
	bool*								m_pbColumnSelectionFlag;		// Set to true when the column selected is changed;

	bool								m_bEnableLines;					// Draw the lines in the tables
	Asura_Vector_2						m_xCellSpacing;					// Spacing between the cells
	Asura_Vector_2						m_xHeadingsSpacing;				// Spacing for the headings ( .x = rows, .y = columns )
	Asura_Vector_2						m_xCellOffset;					// Cell offset
	Asura_Vector_2						m_xRowHeadingOffset;			// Row heading offset
	Asura_Vector_2						m_xColumnHeadingOffset;			// Column heading offset

	Asura_Vector_2						m_xMinCellSize;					// The minimum size allowed for cell in the table.

	Asura_Vector_2						m_xPreResizeViewSize;			// Size at start of resize operation

	bool								m_bSortBackwards;				// A local copy of the sort parameters so they are remembered.
	u_int								m_uSortIndex;					// A local copy of the sort parameters so they are remembered.

	static bool							s_bSortBackwards;				// Static flag allowing sorts to be performed backwards.
	static u_int						s_uSortIndex;					// Static index to be used when sorting.
	static Asura_Table_Dialog*			s_pxSortTableDialogue;			// Static pointer to the table-dialogue currently being sorted.

	// Private functions...
	Asura_Table_Dialog  (void);
	Asura_Table_Dialog  (const Asura_Table_Dialog& xOther);

	Asura_Table_Dialog&	operator=	(const Asura_Table_Dialog& xOther);

	// PAZ : used to enable / disable rows, columns and cells in the table
	bool (*m_pfnIsRowEnabled)( const u_int uRow, void* const pxUserData );
	void* m_pxIsRowEnabledUserData;
	bool IsRowEnabled( const u_int uRow )										{ return( m_pfnIsRowEnabled( uRow, m_pxIsRowEnabledUserData ) ); }
	static bool DefaultIsRowEnabled( const u_int, void* const )					{ return( true ); }

	bool (*m_pfnIsColumnEnabled)( const u_int uColumn, void* const pxUserData );
	void* m_pxIsColumnEnabledUserData;
	bool IsColumnEnabled( const u_int uColumn )									{ return( m_pfnIsColumnEnabled( uColumn, m_pxIsColumnEnabledUserData ) ); }
	static bool DefaultIsColumnEnabled( const u_int, void* const)				{ return( true ); }

	bool (*m_pfnIsCellEnabled)( const u_int uRow, const u_int uColumn, void* const pxUserData );
	void* m_pxIsCellEnabledUserData;
	bool IsCellEnabled( const u_int uRow, const u_int uColumn )					{ return( m_pfnIsCellEnabled( uRow, uColumn, m_pxIsCellEnabledUserData ) ); }
	static bool DefaultIsCellEnabled( const u_int, const u_int, void* const )	{ return( true ); }

	unsigned ( *m_pfnGetColour )( const Asura_Table_Dialog& xTableDialog, void* const pxUserData, const Asura_Dialog* const pxDialog, const unsigned uColour, const u_int uRow, const u_int uColumn );
	void* m_pxGetColourUserData;
	unsigned GetColour( const Asura_Dialog* const pxDialog, const unsigned uColour, const u_int uRow, const u_int uColumn )	const	{ return( m_pfnGetColour( *this, m_pxGetColourUserData, pxDialog, uColour, uRow, uColumn ) ); }
	static unsigned DefaultGetColour( const Asura_Table_Dialog&, void* const, const Asura_Dialog* const, const unsigned uColour, const u_int, const u_int )	{ return( uColour ); }

	// The size available to the view itself - takes account of any scroll-bars.
	const Asura_Vector_2 GetViewSize( void )								const;

	friend unsigned Asura_TableRowHeading_Dialog::GetColour		( void )	const;
	friend unsigned Asura_TableColumnHeading_Dialog::GetColour	( void )	const;

    static Asura_TableColumnHeading_Dialog* CreateStandardColumnHeading( Asura_Table_Dialog& xTableDialog, const char* szName,
                                                                         const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize,
																		 const u_int uColumn,
                                                                         void* const pxUserData );
    static Asura_TableRowHeading_Dialog* CreateStandardRowHeading( Asura_Table_Dialog& xTableDialog, const char* szName,
                                                                   const Asura_Vector_2& xOffset, const Asura_Vector_2& xMinCellSize,
																   const u_int uRow,
                                                                   void* const pxUserData );

	// use helper function, saves updating two virtually identical functions
	void CreateTable( const Asura_Vector_2& xMinCellSize, const u_int uNumRows, const u_int uNumColumns,
                      CreateCell* pfnCreateCell, void* const pxCreateCellUserData,
                      const char* const* pszRowNames, const char* const* pszColumnNames,
                      const Asura_String* pxRowNames, const Asura_String* pxColumnNames,
                      CreateRowHeading* pfnCreateRowHeading, void* const pxCreateRowHeadingUserData,
                      CreateColumnHeading* pfnCreateColumnHeading, void* const pxCreateColumnHeadingUserData );

};   // End of class 'Asura_Table_Dialog'.


#endif	// defined(ASURA_TOOLS)

#endif   // ifndef __ASURA_TABLE_DIALOG_H__


/// End of file. ///
