//////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_TOOLS

#include "Interface/Asura_TexturePicker_Dialog.h"

// Asura Platform-independent
#include "Core/Asura_UserInput.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Viewport.h"

// Asura PC
#include "Rendering/Asura_Direct3D_Effects_System.h"

// Tools
#include "Misc/Tools_Dialog.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Texturing/Tools_PickTexture.h"

/******************************************************************************/
/* Asura_TexturePicker_Gadget Class Functions                                */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::Asura_TexturePicker_Gadget()
//
// Constructor
//
// Parameters:		xOffset - Offset within the parent overlay, measured in
//							  Asura_Text_Output units
//					piTextureID - Pointer to texture ID to track. May be NULL
//					puTextureHash - Pointer to name hash ID to track. May be NULL
//					pxTextureName - Pointer to texture name to track. May be NULL
//					fWidth - Display width for the texture thumbnail, in pixels
//					fHeight - Display width for the texture thumbnail, in pixels
//					pbFlag - Pointer to flag to be set when the texture is
//							 changed. Defaults to NULL
//					eTextureOp - Texture op to use when rendering the image.
//								 Defaults to ASURA_TEXTUREOP_SINGLE_PASS
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_TexturePicker_Gadget::Asura_TexturePicker_Gadget( const Asura_Vector_2& xOffset,
													    int* const piTextureID,
														Asura_Hash_ID* const puTextureHash,
														Asura_String* const pxTextureName,
														const float fWidth,
														const float fHeight,
														bool* const pbFlag /*= NULL*/,
														const ASURA_TEXTUREOP_ID eTextureOp /*= ASURA_TEXTUREOP_SINGLE_PASS*/ )
	: m_piTextureID( piTextureID )
	, m_puTextureHash( puTextureHash )
	, m_pxTextureName( pxTextureName )
	, m_pxSharedTextureID(0)
	, m_pxSharedTextureHash(0)
	, m_pxSharedTextureName(0)
	, m_pbFlag( pbFlag )
	, m_uPickerFlags( PICKTEXTURE_ALLTEXTURES )
	, m_eTextureOp( eTextureOp )
	, m_bCubemap( false )
{
	m_xOffset = xOffset;
	m_fWidth = fWidth / ASURA_TEXT_OUTPUT_FONT_WIDTH;
	m_fHeight = fHeight / ASURA_TEXT_OUTPUT_FONT_HEIGHT;
}

Asura_TexturePicker_Gadget::Asura_TexturePicker_Gadget( const Asura_Vector_2& xOffset,
														Asura_Dialog_SharedValue<int>* const pxSharedTextureID,
														Asura_Dialog_SharedValue<Asura_Hash_ID>* const pxSharedTextureHash,
														Asura_Dialog_SharedValue<Asura_String>* const pxSharedTextureName,
														const float fWidth,
														const float fHeight,
														bool* const pbFlag /*= NULL*/,
														const ASURA_TEXTUREOP_ID eTextureOp /*= ASURA_TEXTUREOP_SINGLE_PASS*/ )
	: m_piTextureID( 0 )
	, m_puTextureHash( 0 )
	, m_pxTextureName( 0 )
	, m_pxSharedTextureID(pxSharedTextureID)
	, m_pxSharedTextureHash(pxSharedTextureHash)
	, m_pxSharedTextureName(pxSharedTextureName)
	, m_pbFlag( pbFlag )
	, m_uPickerFlags( PICKTEXTURE_ALLTEXTURES )
	, m_eTextureOp( eTextureOp )
	, m_bCubemap( false )
{
	m_xOffset = xOffset;
	m_fWidth = fWidth / ASURA_TEXT_OUTPUT_FONT_WIDTH;
	m_fHeight = fHeight / ASURA_TEXT_OUTPUT_FONT_HEIGHT;
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::Update()
//
// Per-frame update of a texture picker gadget
//
// Parameters:		xOrigin - Screen Position of the parent overlay, measured in
//							  Asura_Text_Output units
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsClickable && IsMouseOver() )
	{
		if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_MOUSEBUTTON1 ) )
		{
			Asura_Callback_MemberFunction<void,Asura_TexturePicker_Gadget> xCallback(this, &Asura_TexturePicker_Gadget::PickTexture);
			Asura_Overlay_System::SetCallbackFunctionToCall(xCallback);
			m_bClickedThisFrame = true;
		}

		if( Asura_UserInput::DebouncedKeyPressed( 'r' ) )
		{
			Reload();
		}
	}
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::Render()
//
// Renders a texture picker gadget
//
// Parameters:		xOrigin - Screen position of the parent overlay, measured in
//							  Asura_Text_Output units
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	int iTextureID;
	GetTextureIDAndHash(&iTextureID, 0);

	Asura_Vector_2 xTopLeft = xOrigin;
	xTopLeft += m_xOffset;
	xTopLeft.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xTopLeft.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	Asura_Vector_2 xDimensions( m_fWidth, m_fHeight );
	xDimensions.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xDimensions.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	extern u_int Asura_OverlayBrightness;

	Asura_Render::Overlay.Colour = Asura_OverlayBrightness | GetColour( IsClickedThisFrame() );
	Asura_Render::Overlay.m_bGouraudShaded = false;
	Asura_Render::Overlay.m_bNonRotated = true;
	Asura_Render::Overlay.TextureIndex = iTextureID;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
	Asura_Render::Overlay.m_eTextureOp = m_eTextureOp;

	#if defined ASURA_TARGET_SM3

		if( Asura_Texture_Management::IsTextureCubemap( iTextureID ) )
		{
			Asura_Direct3D_Effect* const pxEffect = Asura_Direct3D_Effects_System::GetEffect( "OverlayQuad_Cubemap" );
			if( pxEffect )
			{
				// Since override effects might use vertex shaders, we have to pass in
				// untransformed vertices. That means taking them from screen-space to
				// [-1,1] clip space
				{
					const float fViewportWidth = static_cast< float >( Asura_Viewport::GetWidth() );
					const float fViewportHeight = static_cast< float >( Asura_Viewport::GetHeight() );

					xTopLeft.x = (xTopLeft.x / fViewportWidth) * 2.0f - 1.0f;
					xTopLeft.y = 1.0f - (xTopLeft.y / fViewportHeight) * 2.0f;
					xDimensions.x = (xDimensions.x / fViewportWidth) * 2.0f;
					xDimensions.y = -(xDimensions.y / fViewportHeight) * 2.0f;
				}

				Asura_Render::Overlay.SetRectXY( xTopLeft.x, xTopLeft.y, xDimensions.x, xDimensions.y );
				Asura_Render::Overlay.SetRectUV( 1.0f, 1.0f, -2.0f, -2.0f );
				Asura_Direct3D::AddQuadToBatch( &Asura_Render::Overlay, pxEffect );
			}
			else
			{
				Asura_Assert( false, "Could not retrieve cubemap quad effect." );
			}
		}
		else

	#endif // defined ASURA_TARGET_SM3

	{
		Asura_Render::Overlay.SetRectXY( xTopLeft.x, xTopLeft.y, xDimensions.x, xDimensions.y );
		Asura_Render::Overlay.SetRectUV( 0.0f, 0.0f, 1.0f, 1.0f );
		Asura_Render::Overlay.Render();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::Clear()
//
// Clears the texture bound to this gadget and sets the modified flag
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Gadget::Clear()
{
	if( m_piTextureID )
	{
		*m_piTextureID = ASURA_TEXTURE_NONE;
	}

	if( m_puTextureHash )
	{
		*m_puTextureHash = ASURA_HASH_ID_UNSET;
	}

	if( m_pxTextureName )
	{
		m_pxTextureName->MakeNull();
	}

	if(m_pxSharedTextureID)
	{
		m_pxSharedTextureID->SetValue(ASURA_TEXTURE_NONE);
	}

	if(m_pxSharedTextureHash)
	{
		m_pxSharedTextureHash->SetValue(ASURA_HASH_ID_UNSET);
	}

	if(m_pxSharedTextureName)
	{
		Asura_String xNullString;
		m_pxSharedTextureName->SetValue(xNullString);
	}

	if( m_pbFlag )
	{
		*m_pbFlag = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::Reload()
//
// Reloads the texture bound to this gadget
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Gadget::Reload()
{
	int iTextureID;
	GetTextureIDAndHash(&iTextureID, 0);


	if( iTextureID != ASURA_TEXTURE_NONE )
	{
		const Asura_String* const pxTextureName = Asura_Texture_Management::GetTextureName( iTextureID );

		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_INFO, "Reloading %s",
												pxTextureName ? pxTextureName->GetString() : "<Unknown>" );

		Asura_Texture_Management::ReloadTexture( iTextureID );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::GetTextureIDAndHash()
//
// Determines the texture ID from the million and one variables that can be used , and possibly the texture hash
// if one of the texture hash options is in use 
//
// Parameters:		Some
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Gadget::GetTextureIDAndHash(int* piTextureID, Asura_Hash_ID* puTextureHash)
{
	//try the non-shared values
	int iTextureID = m_piTextureID ? *m_piTextureID
					: m_puTextureHash ? Asura_Texture_Management::GetIndexFromHashID( *m_puTextureHash )
					: m_pxTextureName ? Asura_Texture_Management::GetIndexFromHashID( m_pxTextureName->GetHash_IgnoreExtension() )
					: ASURA_TEXTURE_NONE;
	
	Asura_Hash_ID uTextureHash = m_puTextureHash ? *m_puTextureHash : ASURA_HASH_ID_UNSET;
	
	
	
	//and the shared values
	if(m_pxSharedTextureID)
	{
		m_pxSharedTextureID->GetValue(&iTextureID,0);
	}
	else if(m_pxSharedTextureHash)
	{
		m_pxSharedTextureHash->GetValue(&uTextureHash,0);
		iTextureID = Asura_Texture_Management::GetIndexFromHashID( uTextureHash );

	}
	else if(m_pxSharedTextureName)
	{
		Asura_String xName;
		m_pxSharedTextureName->GetValue(&xName, 0);
		iTextureID = Asura_Texture_Management::GetIndexFromHashID( xName.GetHash_IgnoreExtension() );
	}

	if(piTextureID) *piTextureID = iTextureID;
	if(puTextureHash) *puTextureHash = uTextureHash;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Gadget::PickTexture()
//
// Callback for dealing with the texture selection at the end of the overlay system update
//
// Parameters:		None
//					
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Gadget::PickTexture()
{
	int iOldTextureID = m_piTextureID ? *m_piTextureID
						: m_puTextureHash ? Asura_Texture_Management::GetIndexFromHashID( *m_puTextureHash )
						: m_pxTextureName ? Asura_Texture_Management::GetIndexFromHashID( m_pxTextureName->GetHash_IgnoreExtension() )
						: ASURA_TEXTURE_NONE;
	Asura_Hash_ID uOldHash = m_puTextureHash ? *m_puTextureHash : ASURA_HASH_ID_UNSET;


	if(m_pxSharedTextureID)
	{
		m_pxSharedTextureID->GetValue(&iOldTextureID,0);
	}
	else if(m_pxSharedTextureHash)
	{
		m_pxSharedTextureHash->GetValue(&uOldHash,0);
		iOldTextureID = Asura_Texture_Management::GetIndexFromHashID( uOldHash );

	}
	else if(m_pxSharedTextureName)
	{
		Asura_String xName;
		m_pxSharedTextureName->GetValue(&xName, 0);
		iOldTextureID = Asura_Texture_Management::GetIndexFromHashID( xName.GetHash_IgnoreExtension() );
	}

	Asura_Hash_ID uNewHash = uOldHash;

	//If m_puTextureHash has been provided this must be passed in instad of &uNewHash , so that the texture gets converted
	//correctly within PickTexture(). Otherwise the query to find out how the texture is used will get the wrong results , since
	//it won't have been set yet
	//RAR 15/11/2007
	const int iNewTextureID = Tools_PickTexture::PickTexture( "Select texture",
															  iOldTextureID,
															  m_uPickerFlags, 0,
															  m_puTextureHash ? m_puTextureHash : &uNewHash );

	if( iNewTextureID != ASURA_TEXTURE_NONE )
	{
		const bool bCubemap = Asura_Texture_Management::IsTextureCubemap( iNewTextureID );

		if( GetValidateCubemap() && !bCubemap )
		{
			Message_Overlay xMessage( "The selected texture isn't a cubemap." );
			xMessage.DoModal();

			Clear();
		}
		else
		{
			if( m_piTextureID )
			{
				*m_piTextureID = iNewTextureID;
			}

			
			if( m_pxTextureName )
			{
				*m_pxTextureName = *Asura_Texture_Management::GetTextureName( iNewTextureID );
			}

			if( m_pxSharedTextureID )
			{
				m_pxSharedTextureID->SetValue(iNewTextureID);
			}

			if(	m_pxSharedTextureHash)
			{
				m_pxSharedTextureHash->SetValue(m_puTextureHash ? *m_puTextureHash : uNewHash);
			}
			
			if( m_pxSharedTextureName )
			{
				m_pxSharedTextureName->SetValue(*Asura_Texture_Management::GetTextureName( iNewTextureID ));
			}


			if( m_pbFlag )
			{
				*m_pbFlag = true;
			}
		}
	}
}



/******************************************************************************/
/* Asura_TexturePicker_Dialog Class Functions                                 */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::Asura_TexturePicker_Dialog()
//
// Constructor
//
// Parameters:		szName - Display name of the control
//					xOffset - Offset within the parent overlay, measured in
//							  Asura_Text_Output units
//					piTextureID - Pointer to texture ID to track. May be NULL
//					puTextureHash - Pointer to name hash ID to track. May be NULL
//					pxTextureName - Pointer to texture name to track. May be NULL
//					fWidth - Display width for the texture thumbnail, in pixels
//					fHeight - Display width for the texture thumbnail, in pixels
//					pbFlag - Pointer to flag to be set when the texture is
//							 changed. Defaults to NULL
//					eTextureOp - Texture op to use when rendering the image.
//								 Defaults to ASURA_TEXTUREOP_SINGLE_PASS
//
////////////////////////////////////////////////////////////////////////////////

Asura_TexturePicker_Dialog::Asura_TexturePicker_Dialog( const char* const szName,
													    const Asura_Vector_2& xOffset,
														int* const piTextureID,
														Asura_Hash_ID* const puTextureHash,
														Asura_String* const pxTextureName,
														const float fWidth,
														const float fHeight,
														bool* const pbFlag /*= NULL*/,
														const ASURA_TEXTUREOP_ID eTextureOp /*= ASURA_TEXTUREOP_SINGLE_PASS*/ )
	: Asura_Dialog( szName, xOffset )
{
	Asura_Vector_2 xGadgetOffset = xOffset;
	xGadgetOffset.y += 1.5f;

	// Texture picker gadget
	m_pxTextureGadget = new Asura_TexturePicker_Gadget( xGadgetOffset, piTextureID, puTextureHash, pxTextureName, fWidth, fHeight, pbFlag, eTextureOp );
	AddGadget( m_pxTextureGadget );
	xGadgetOffset.x += (fWidth + 2.0f) / ASURA_TEXT_OUTPUT_FONT_WIDTH;

	// Clear gadget
	m_pxClearButton = new Asura_Icon_Gadget( xGadgetOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_DELETE );
	m_pxClearButton->SetClickedFunction( OnClear, reinterpret_cast< int >( this ) );
	AddGadget( m_pxClearButton );
	xGadgetOffset.y += 1.3f;

	// Reload gadget
	m_pxReloadButton = new Asura_Icon_Gadget( xGadgetOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_REFRESH );
	m_pxReloadButton->SetClickedFunction( OnReload, reinterpret_cast< int >( this ) );
	AddGadget( m_pxReloadButton );
	xGadgetOffset.y += 1.3f;

	// Have to set help text to something non-blank, otherwise the tooltip function won't get called
	SetHelpText( " " );
}

Asura_TexturePicker_Dialog::Asura_TexturePicker_Dialog( const char* const szName,
													    const Asura_Vector_2& xOffset,
														Asura_Dialog_SharedValue<int>* const pxSharedTextureID,
														Asura_Dialog_SharedValue<Asura_Hash_ID>* const pxSharedTextureHash,
														Asura_Dialog_SharedValue<Asura_String>* const pxSharedTextureName,
														const float fWidth,
														const float fHeight,
														bool* const pbFlag /*= NULL*/,
														const ASURA_TEXTUREOP_ID eTextureOp /*= ASURA_TEXTUREOP_SINGLE_PASS*/ )
	: Asura_Dialog( szName, xOffset )
{
	Asura_Vector_2 xGadgetOffset = xOffset;
	xGadgetOffset.y += 1.5f;

	// Texture picker gadget
	m_pxTextureGadget = new Asura_TexturePicker_Gadget( xGadgetOffset, pxSharedTextureID, pxSharedTextureHash, pxSharedTextureName, fWidth, fHeight, pbFlag, eTextureOp );
	AddGadget( m_pxTextureGadget );
	xGadgetOffset.x += (fWidth + 2.0f) / ASURA_TEXT_OUTPUT_FONT_WIDTH;

	// Clear gadget
	m_pxClearButton = new Asura_Icon_Gadget( xGadgetOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_DELETE );
	m_pxClearButton->SetClickedFunction( OnClear, reinterpret_cast< int >( this ) );
	AddGadget( m_pxClearButton );
	xGadgetOffset.y += 1.3f;

	// Reload gadget
	m_pxReloadButton = new Asura_Icon_Gadget( xGadgetOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_REFRESH );
	m_pxReloadButton->SetClickedFunction( OnReload, reinterpret_cast< int >( this ) );
	AddGadget( m_pxReloadButton );
	xGadgetOffset.y += 1.3f;

	// Have to set help text to something non-blank, otherwise the tooltip function won't get called
	SetHelpText( " " );
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::ShowHelpText()
//
// Shows the tooltip for a texture picker dialog
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Dialog::ShowHelpText()
{
	if( m_pxClearButton && m_pxClearButton->IsMouseOver() )
	{
		Asura_Overlay_System::SetHelpText( "Clears the texture." );
	}
	else if( m_pxReloadButton && m_pxReloadButton->IsMouseOver() )
	{
		Asura_Overlay_System::SetHelpText( "Reloads the selected texture. (R)" );
	}
	else if( m_pxTextureGadget && m_pxTextureGadget->IsMouseOver() )
	{
		const int* const piGadgetID = m_pxTextureGadget->GetID();
		const Asura_Hash_ID* const puGadgetHash = m_pxTextureGadget->GetHash();
		const Asura_String* const pxGadgetString = m_pxTextureGadget->GetName();

		const int iTextureID = piGadgetID ? *piGadgetID
							 : puGadgetHash ? Asura_Texture_Management::GetIndexFromHashID( *puGadgetHash )
							 : ASURA_TEXTURE_NONE;

		const Asura_String* const pxTextureName = pxGadgetString && !pxGadgetString->IsBlank()
												? pxGadgetString
												: Asura_Texture_Management::GetTextureName( iTextureID );

		const char* const szTooltip = pxTextureName ? pxTextureName->GetString() : "No texture selected";

		Asura_Overlay_System::SetHelpText( szTooltip );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::SetPickerFlags()
//
// Sets the flags for use in Tools_PickTexture when this picker's clicked
//
// Parameters:		uFlags - Flags to use. See Tools_PickTexture.h
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Dialog::SetPickerFlags( const u_int uFlags )
{
	if( m_pxTextureGadget )
	{
		m_pxTextureGadget->SetPickerFlags( uFlags );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::GetPickerFlags()
//
// Queries this picker's Tools_PickTexture flags
//
// Parameters:		None
// Return value:	Current flags. See Tools_PickTexture.h
//
////////////////////////////////////////////////////////////////////////////////

u_int Asura_TexturePicker_Dialog::GetPickerFlags() const
{
	return m_pxTextureGadget ? m_pxTextureGadget->GetPickerFlags() : 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::SetValidateCubemap()
//
// Enables validation that a picked texture is a cubemap
//
// Parameters:		true iff texture must be a cubemap
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Dialog::SetValidateCubemap( const bool bCubemap )
{
	if( m_pxTextureGadget )
	{
		m_pxTextureGadget->SetValidateCubemap( bCubemap );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::GetValidateCubemap()
//
// Queries whether this picker validates the selection is a cubemap
//
// Parameters:		None
// Return value:	true iff cubemap validation is enabled
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_TexturePicker_Dialog::GetValidateCubemap() const
{
	return m_pxTextureGadget ? m_pxTextureGadget->GetValidateCubemap() : 0;
}

/******************************************************************************/
/* Asura_TexturePicker_Dialog Static Class Functions                          */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::OnClear()
//
// Called when the 'clear' gadget is clicked
//
// Parameters:		iData - Pointer to dialog, since this has to be a static fn
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Dialog::OnClear( int iData )
{
	Asura_TexturePicker_Dialog* const pxDialog = reinterpret_cast< Asura_TexturePicker_Dialog* >( iData );

	if( pxDialog && pxDialog->m_pxTextureGadget )
	{
		pxDialog->m_pxTextureGadget->Clear();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_TexturePicker_Dialog::OnReload()
//
// Called when the 'reload' gadget is clicked
//
// Parameters:		iData - Pointer to dialog, since this has to be a static fn
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_TexturePicker_Dialog::OnReload( int iData )
{
	Asura_TexturePicker_Dialog* const pxDialog = reinterpret_cast< Asura_TexturePicker_Dialog* >( iData );

	if( pxDialog && pxDialog->m_pxTextureGadget )
	{
		pxDialog->m_pxTextureGadget->Reload();
	}
}

#endif // defined ASURA_TOOLS
