//////////////////////////////////////////////////////////////////////////////
//
// Asura_Job_PFX.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

// This file's header.
#include "Jobs/Asura_Job_PFX.h"

// asura includes
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Jobs/Asura_Job.h"
#include "SpecialFX/PFX/Asura_PFX_Classes.h"
#include "SpecialFX/PFX/Asura_PFX_Effect.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

#ifdef _XBOX
#include "Rendering/Asura_Xbox360_Graphics.h"
#endif

#ifdef ASURA_USE_FASTPATH_PFX
#include "SpecialFX/Asura_Platform_PFX_System_FastPath.h"
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////
#if ASURA_MAX_NO_OF_CAMERAS != MAX_CAMERAS_FOR_PFX_JOB
#error mismatch in max number of cameras
#endif

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Job_Queue* Asura_Job_PFX_Update::s_pxQueue;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_Job_PFX_Update::InitialiseData(Asura_PFX_Particle* pxParticles, u_int uNumParticles, u_int uNumParticlesToUpdate, 
    			    const Asura_PFX_Source_Template* pxSourceTemplate, 
					bool bUseCollisionPlane, const Asura_Vector_3& xCollisionPosition, const Asura_Vector_3& xCollisionDirection, 
					bool bUsePointGravity, const Asura_Vector_3& xGravityPosition, float fSizeScale, void* pRenderData,
					const Asura_RGBA& xLightColour, const Asura_Matrix_3x3& xEffectOrientation, u_int uCameraFlags, const Asura_Spline<Asura_Vector_3>* pxSpline )
{
	m_pxParticles					= pxParticles;

//	Asura_OutputToDebugger(("InitJob %x\n",pxParticles));

	m_uNumParticles 				= uNumParticles;
	m_uNumParticlesToUpdate			= uNumParticlesToUpdate;
	m_pxTemplate					= pxSourceTemplate;
#ifdef ASURA_PFX_TEMPLATES_SUPPORT_DATABLOCK
	const Asura_PFX_Particle_Template* pxPcleTemp = Asura_PFX_System::GetParticleTempByHash(pxSourceTemplate->m_uParticleTemplateHash);
	Asura_Assert(pxPcleTemp, "No particle template");

	m_pSourceTemplateDataBlock		= pxSourceTemplate->GetDataBlock();
	m_uSourceTemplateDataBlockSize	= pxSourceTemplate->GetDataBlockSize();
	m_pParticleTemplateDataBlock	= pxPcleTemp->GetDataBlock();
	m_uParticleTemplateDataBlockSize= pxPcleTemp->GetDataBlockSize();
#endif
	m_bUseCollisionPlane			= bUseCollisionPlane;
	m_bUsePointGravity				= bUsePointGravity;
	m_xCollisionPosition			= xCollisionPosition;
	m_xCollisionDirection			= xCollisionDirection;
	m_xGravityPosition				= xGravityPosition;
	m_fSizeScale					= fSizeScale;
	m_fTimeSlice					= Asura_Timers::GetGameTimeSlice();
	m_fTurbulencePhase				= Asura_Timers::GetLocalUnscaledTime();
	m_pxSpline						= pxSpline;

	
	// extra data for rendering
	m_pPlatformRenderData			= pRenderData;

	m_xLightColour					= xLightColour;
	m_xEffectOrientation			= xEffectOrientation;

	int iNumCameras = Asura_Camera::GetNumberOfCameras();
	for (int i = 0; i < iNumCameras; ++i)
	{
		m_axCameraPosition[i] = *Asura_Camera::GetPosition();

		const Asura_Matrix_3x3*	pxCameraOrientation	= Asura_Camera::GetOrientation ();

		// Store the X and Y axes from the camera...
		m_axCameraDeltaX[i].x	= pxCameraOrientation->e11;
		m_axCameraDeltaX[i].y	= pxCameraOrientation->e21;
		m_axCameraDeltaX[i].z	= pxCameraOrientation->e31;

		m_axCameraDeltaY[i].x	= pxCameraOrientation->e12;
		m_axCameraDeltaY[i].y	= pxCameraOrientation->e22;
		m_axCameraDeltaY[i].z	= pxCameraOrientation->e32;
	}

	m_iNumberOfCameras	= iNumCameras;
	m_uCameraFlags		= uCameraFlags;
}



void Asura_Job_PFX_Update::Execute()
{

//	Asura_OutputToDebugger(("Exec: %x\n",m_pxParticles));

	// DMR 14/11/06 - Yes, this is a bit nasty, but there's no platform-specific part of this code to
	// put the profile calls in....
#ifdef _XBOX
		SCOPEDPIXEVENT((0xff00ff00,  "Asura_Job_PFX_Update::Execute"));
#endif


	const Asura_PFX_Particle_Template* pxPcleTemp = Asura_PFX_System::GetParticleTempByHash(m_pxTemplate->m_uParticleTemplateHash);
	Asura_PFX_Effect::UpdateParticles(m_pxParticles,
									  m_uNumParticlesToUpdate,
									  m_pxTemplate,
									  m_bUseCollisionPlane,
									  m_xCollisionPosition,
									  m_xCollisionDirection,
									  m_bUsePointGravity,
									  m_xGravityPosition,
									  m_fSizeScale,
									  m_fTimeSlice,
									  m_fTurbulencePhase,
									  pxPcleTemp,
									  m_pxSpline);


 #ifdef ASURA_USE_FASTPATH_PFX
	// On the 360, we fill in the instance buffer here, too
	if (m_pPlatformRenderData)
	{		
		//InitialiseData
		Asura_Platform_PFX_System_FastPath::UpdateInstanceBuffer(pxPcleTemp,m_pxParticles, (Asura_Platform_PFXInstanceData *)m_pPlatformRenderData,m_uNumParticles);
	}
#endif

}

// for hack
DEBUGVAR bool dbg_bTestMultithreadedPFX = false;
#include "Debugging/Asura_Assert.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

#ifdef ASURA_ASSERT

static bool ArePFXApproximatelyEqual( const Asura_PFX_Particle* pxParticles1, const Asura_PFX_Particle* pxParticles2, u_int uNumParticles )
{
	// for each pair of particles compare the first 8 floats within a slight error.
	for (u_int u = 0; u < uNumParticles; ++u)
	{
		const float* pfFloat1 = reinterpret_cast<const float*>(pxParticles1++);
		const float* pfFloat2 = reinterpret_cast<const float*>(pxParticles2++);
		for (u_int uFloat = 0; uFloat < 8; ++uFloat)
		{
			if (fabsf(*pfFloat1++ - *pfFloat2++) > 0.0001f)
			{
				return false;
			}
		}
	}
	return true;
}

#endif

void Asura_Job_PFX_Update::AddToQueue()
{
	bool bDoTest = dbg_bTestMultithreadedPFX;
	Asura_PFX_Particle* pxParticlesTest = NULL;
	if (bDoTest)
	{
  		// hack - execute job to test we get the same result
  		pxParticlesTest = new Asura_PFX_Particle[m_uNumParticles];
  		memcpy(pxParticlesTest, m_pxParticles, sizeof(Asura_PFX_Particle) * m_uNumParticles);
  		Asura_PFX_Particle* const pxParticlesPtrBackup = m_pxParticles;
  		m_pxParticles = pxParticlesTest;
  		Execute();
  		m_pxParticles = pxParticlesPtrBackup;
	}

//	Asura_OutputToDebugger(("push job %u particles, EA = 0x%x\n", m_uNumParticles, m_pxParticles));

	s_pxQueue->Push(this);

  	if (bDoTest)
	{
		FinishAllJobs();
  		Asura_Assert( ArePFXApproximatelyEqual( m_pxParticles, pxParticlesTest, m_uNumParticles ), "pfx update difference" );
		delete[] pxParticlesTest;
	}
}


/*DEBUGVAR */bool dbg_bDoJobsWhileWaiting = false;
/*DEBUGVAR */bool dbg_bPrintJobsDoneWhileWaiting = false;//!dbg_bDoJobsWhileWaiting;

void Asura_Job_PFX_Update::FinishAllJobs()
{

//	Asura_OutputToDebugger(("--------------\n"));
	if(!s_pxQueue) return;

#ifdef SN_TARGET_PS3
	if (dbg_bDoJobsWhileWaiting)
	{
#endif
		// DMR 14/11/06
		// We always allow this on PC (also on 360 for now). There's often going to be 
		// just one worker thread (on dual-core/HT PCs), so sitting idle here is bad.
		u_int uRemainingJobs = s_pxQueue->GetNumEntries();
		while (uRemainingJobs)
		{
			Asura_Job_PFX_Update xJob;
			bool bGotJob = s_pxQueue->TryPop( &xJob );
			if (bGotJob)
			{
				xJob.Execute();
				s_pxQueue->JobFinished();
				if (dbg_bPrintJobsDoneWhileWaiting)
				{
					Asura_OutputToDebugger_Retail(("done job while waiting for %u PFX jobs\n", uRemainingJobs));
				}
			}

			// If there's no jobs left to do, this becomes a spin-wait. Could probably do something better here,
			// like add a 's_pxQueue->WaitUntilEmpty()', and use platform-specific methods to sleep/wake 
			// this main thread?
			uRemainingJobs = s_pxQueue->GetNumEntries();			
		}
#ifdef SN_TARGET_PS3
	}
#endif

	u_int uNumUnfinishedJobs = s_pxQueue->GetNumUnfinishedJobs();
	bool bAnyWaiting = uNumUnfinishedJobs != 0;
	while (uNumUnfinishedJobs)
	{
		static u_int ls_uPrevNumUnfinishedJobs = 0xffffffff;
		if (dbg_bPrintJobsDoneWhileWaiting && (ls_uPrevNumUnfinishedJobs != uNumUnfinishedJobs))
		{
			Asura_OutputToDebugger_Retail(("waiting for %u PFX jobs\n", uNumUnfinishedJobs));
		}
		ls_uPrevNumUnfinishedJobs = uNumUnfinishedJobs;
		uNumUnfinishedJobs = s_pxQueue->GetNumUnfinishedJobs();
	}
	if (dbg_bPrintJobsDoneWhileWaiting && bAnyWaiting)
	{
		Asura_OutputToDebugger_Retail(("FinishAllJobs complete\n"));
	}
}

void Asura_Job_PFX_Update::InitialiseAllocations()
{
	s_pxQueue = new Asura_Job_Queue(128, sizeof(Asura_Job_PFX_Update), ExecuteJob);
	s_pxQueue->SetThreadName( "PFX Updates");
}

void Asura_Job_PFX_Update::ResetAllocations()
{
	s_pxQueue->Clear();
}

void Asura_Job_PFX_Update::FreeAllocations()
{
	delete s_pxQueue;
	s_pxQueue = NULL;
}


bool Asura_Job_PFX_Update::ExecuteJob(void *pxJob)
{
	ASURA_CPU_PROFILE__START_MULTITHREADED_LABEL(ASURA_CPU_PROFILE__SPECIALFX_UPDATE, "SpecialFX Update");
	
	Asura_Job_PFX_Update* pxPFXJob = reinterpret_cast<Asura_Job_PFX_Update*>(pxJob);

	bool bReturnValue;
	if (pxPFXJob->IsValid())
	{
		pxPFXJob->Execute();
		bReturnValue = true;
	}
	else
	{
		bReturnValue = false;
	}
	
	ASURA_CPU_PROFILE__STOP_MULTITHREADED(ASURA_CPU_PROFILE__SPECIALFX_UPDATE);
	
	return( bReturnValue);
}


#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
