//
// Asura_Light.h
//
// Created by Kevin Lea - 12/06/2001 10:56:19
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_LIGHT_H__
#define __ASURA_LIGHT_H__

#include "Core/Asura_Maths.h"

#define MAX_NO_OF_LIGHTS 2048
#define MAX_NO_OF_LOCAL_LIGHTS (MAX_NO_OF_LIGHTS)

//Defines for accessing the bitfield values from m_uFlags
//(mainly so that the FlagTickbox in tools can access them)
//RAR 30/01/03
#define AsuraLightFlag_HasCorona			 0x00000001
#define AsuraLightFlag_InfiniteLight		 0x00000002
#define AsuraLightFlag_AffectsEntities		 0x00000004
#define AsuraLightFlag_AffectsEnvironment	 0x00000008
#define AsuraLightFlag_IsVolumetric			 0x00000010
#define AsuraLightFlag_UseBoundingBox 		 0x00000020
#define AsuraLightFlag_IsShadowVolume		 0x00000040

// ART 26/09/05 - This is nasty, the value is hooked to the number of
// platforms (defined in "Tools_Level_Export.h"...
#define ASURA_NUM_LIGHT_SCALE_VALUES 8


class Asura_Light
{
	public:

	// name?

	Asura_Vector_3 Position;
	Asura_Vector_3 Direction;

	//RAR 20/05/03 - Colour values in range 0 to 255
	float R;
	float G;
	float B;

	float Brightness; // RAR - Brightness is straight multiplier (1.0f equsls normal brightness)
	float Range;
	float m_fInnerRange;//inner range at which prelighting is full strength
	float Angle;

	float ShadowStrength;

	Asura_Bounding_Box m_xBoundingBox;//cut off bounding box for light , used if flag is set below


	/// Kevin Lea - 26/09/2002 11:28:36 - added a union so that all the
	/// light flags can be copied just by using the 'm_uFlags' member
	union
	{
		struct
		{
			u_int	HasCorona:1,
					InfiniteLight:1,
					AffectsEntities:1,
					AffectsEnvironment:1,
					IsVolumetric:1,
					UseBoundingBox:1,
					m_bIsShadowVolume:1,
					SpareBits:25;
		};

		u_int		m_uFlags;
	};

	float BrightnessOverRange;
	Asura_Vector_3 OldPosition;
	float OldRange;
	bool HasChanged;

	//Sets everything to some reasonable default non-junky value
	void SetToDefault();
	
	void Set(const Asura_Light& newSettings)
	{
		OldPosition = Position;
		OldRange = Range;

		Position = newSettings.Position;
		Direction = newSettings.Direction;
		R = newSettings.R;
		G = newSettings.G;
		B = newSettings.B;
		Brightness = newSettings.Brightness;
		Range = newSettings.Range;
		m_fInnerRange = newSettings.m_fInnerRange;
		Angle = newSettings.Angle;
		ShadowStrength = newSettings.ShadowStrength;
		m_xBoundingBox = newSettings.m_xBoundingBox;
	
		/// Kevin Lea - 26/09/2002 11:36:07 - copy all the flags in one go
		m_uFlags = newSettings.m_uFlags;

		HasChanged = true;
	}
    
	// (MH - 09.01.2006) this operator= is now required for the PS2 compiler.
	Asura_Light& operator= ( const Asura_Light& l) 
	{
		Position = l.Position;
		Direction = l.Direction;
		R = l.R;
		G = l.G;
		B = l.B;

		Brightness = l.Brightness;
		Range = l.Range;
		m_fInnerRange = l.m_fInnerRange;
		Angle = l.Angle;

		ShadowStrength = l.ShadowStrength;

		m_xBoundingBox = l.m_xBoundingBox;

		m_uFlags = l.m_uFlags;

		BrightnessOverRange = l.BrightnessOverRange;
		OldPosition = l.OldPosition;
		OldRange = l.OldRange;
		HasChanged = l.HasChanged;
		return *this;
	}

	inline bool operator == ( const Asura_Light& l) const
	{
	    return (Position==l.Position) && (Direction==l.Direction) && (Range==l.Range) && (m_uFlags == l.m_uFlags)
		 && (Brightness==l.Brightness) && (R==l.R) && (G==l.G) && (B==l.B) && (Angle==l.Angle) && (ShadowStrength==l.ShadowStrength)
		 && (m_xBoundingBox==l.m_xBoundingBox) && (m_fInnerRange == l.m_fInnerRange);
	}

    inline bool operator != ( const Asura_Light& xOther ) const { return( !(*this == xOther) ); }
};

class Asura_Light_Visibility
{
	public:

	bool DataValid;
	int NumberOfVisibleVertices;
	int NumberAllocated;
	int* Indices;

	float MaxRange;
	
	void AddIndex(int index, const Asura_Light* light)
	{
		if(NumberOfVisibleVertices+1>NumberAllocated)
		{
			IncreaseAllocation(light);
		}
		Indices[NumberOfVisibleVertices++] = index;

	}

	void MakeVoid(void)
	{
		DataValid = false;
		//free(Indices);
		Asura_Deallocate(Indices);
		Indices = 0;
		NumberAllocated = 0;
		NumberOfVisibleVertices = 0;
		MaxRange=0;
	}

	void IncreaseAllocation(const Asura_Light* light);

	void Sort(void);

	static int GetVisibilityIndex(int iModuleIndex,int iVertexIndex)
	{
		return iModuleIndex*100000+iVertexIndex;
	}
	static void GetModuleAndVertexIndex(int iVisIndex,int& iModuleIndex,int& iVertexIndex)
	{
		iModuleIndex = iVisIndex/100000;
		iVertexIndex = iVisIndex%100000;
	}

	#ifdef ASURA_TOOLS
	void DeleteModule(int iModuleIndex);
	#endif
	
	
};

class Asura_Lighting_System
{
	friend class Asura_Photon_System;
	friend class Asura_Chunk_Lights;
	friend class Tools_Mode_Lighting;
	public:
		static void Initialise(void);
		static void Update(void);
		static void Reset(void);
		static void Shutdown(void);

		static Asura_Light* AllocateLight(void);
		static void DeallocateLight(Asura_Light *lightPtr);

		static void MakeStaticPhotonsList(void);

#ifdef ASURA_TOOLS
		static void CalculatePrelighting(class Asura_Mesh& mesh, const Asura_Vector_3& position,int iModuleIndex);
		static void CalculatePreLightValue(const Asura_Vector_3& position,const Asura_Vector_3& xPositionForLOS, const Asura_Vector_3& normal, unsigned int& colour, int iPolygonFlags);
		//For doing prelighting on subdivided mesh.. in Tools_SubdividedMesh.cpp
		static void CalculatePrelighting(class Tools_SubdividedMesh* pxSubdividedMesh);

		static bool MakeLightListForObject(const Asura_Bounding_Box& boundingBox);
		static bool IsUsingEntityLights();
		static void ToggleUsingEntityLights();

		static void DeleteModule(int iModuleIndex);
#endif
			
	//private:
		
		static int NumberOfActiveLights;

        static Asura_Light * LightStorage;
		static Asura_Light_Visibility * Visibility;
		
		static int NumberOfLocalLights;
        static Asura_Light** LocalLights;
	
		static bool LightsHaveChanged;
		static bool ForceLightingUpdate;

		static bool ShowLightRadius;

		static float AmbientR;
		static float AmbientG;
		static float AmbientB;

		static float s_fBottomAmbientR;
		static float s_fBottomAmbientG;
		static float s_fBottomAmbientB;

	private:
		static float s_fEntityAmbientR;
		static float s_fEntityAmbientG;
		static float s_fEntityAmbientB;

	public:
		static u_int s_uEntityAmbientRGB; // Precalculated RGB value.

		static bool s_bUseHemisphereAmbient;

		static float SkyRed;
		static float SkyGreen;
		static float SkyBlue;
		static bool s_bRenderUsingEntityLights;

		//tools callback function for finding out how much each light influences a point
		static void (*s_pfnLightInfluenceCallback)(int iIndex,float fRed,float fGreen,float fBlue);
	
//		static Asura_Vector_3 AverageLightPosition;

#ifdef ASURA_TOOLS
		//light scale values for each platform. RAR 16/02/04
		static float s_afPrelightScale[ASURA_NUM_LIGHT_SCALE_VALUES];
		static float s_afRuntimeLightScale[ASURA_NUM_LIGHT_SCALE_VALUES];
#endif

	private:

#ifdef ASURA_TOOLS
		static void CalculatePreLightValueForEntityLights(const Asura_Vector_3& position, const Asura_Vector_3& normal, unsigned int& colour, int objectIndex, int vertexIndex, int iPolygonFlags);
		static void CalculatePreLightValue(const Asura_Vector_3& position,const Asura_Vector_3& xPositionForLOS, const Asura_Vector_3& normal, unsigned int& colour, int objectIndex, int vertexIndex, int iPolygonFlags);
#endif

};

#endif//__ASURA_LIGHT_H__


