//
// Asura_Photon.cpp
//
// Created by Kevin Lea - 12/06/2001 10:53:54
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Light/Asura_Photon.h"
#include "Light/Asura_Light.h"

#include "Core/Asura_AABB_Tree.h"
#include "Core/Asura_DynamicBVH.h"
#include "Core/Asura_Mempool.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"

#ifdef ASURA_TARGET_SM3
#include "Camera/Asura_Camera.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_DeferredShading.h"
#endif

#include <stdlib.h>

static Asura_DynamicBVH<Asura_Photon>* g_pxDynamicPhotonBVH;
static bool g_bForceRecreateOnDynamicPhotonBVH;
static u_int g_uNumDynamicPhotonBVHRefits;
static u_int g_uDynamicPhotonBVHRefitFrameStamp;
static const u_int uMAX_NUMBER_OF_DYNAMIC_PHOTON_BVH_REFITS = 100;

class DynamicPhotonBVH_BBRetrievalWrapper
{
public:
    static inline void GetAABB( const Asura_Photon& xPhoton, Asura_Bounding_Box& xAABB )
    {
        if( xPhoton.GetUseBoundingBox() )
        {
            xAABB = xPhoton.GetBoundingBox();
        }
        else
        {
            xAABB.SetToPoint( xPhoton.GetPosition() );
            xAABB.Grow( xPhoton.GetRange() );
        }
    }
};

u_int							Asura_Photon_System::NumberOfLocalPhotons;
Asura_Local_Photon				Asura_Photon_System::LocalPhotons[ASURA_MAX_NO_OF_LOCAL_PHOTONS];


Asura_TrackingHandlePool<Asura_Photon>*		Asura_Photon_System::s_pxDynamicPhotonList;

class Asura_AABB_Tree*						Asura_Photon_System::s_pxStaticPhotonTree;
Asura_Photon*								Asura_Photon_System::s_pxStaticPhotonStorage;
u_int										Asura_Photon_System::s_uNumberOfStaticPhotons;

Asura_Photon*								Asura_Photon_System::s_pxTemporaryAveragePhoton;
const Asura_Vector_3*						Asura_Photon_System::s_pxProximityTestCentre;
const Asura_Bounding_Box*					Asura_Photon_System::s_pxProximityTestBoundingBox;
float										Asura_Photon_System::s_fProximityTestMaxInfluence;

Asura_Photon_Sampling_Result				Asura_Photon_System::s_xPhotonResult;
float										Asura_Photon_System::s_fShadowVolumeScale;

bool										Asura_Photon_System::s_bGettingPhotonsForEnv;

void Asura_Photon_System::Initialise(void)
{
    s_pxTemporaryAveragePhoton = new Asura_Photon;

	s_pxDynamicPhotonList = new Asura_TrackingHandlePool<Asura_Photon>(16);

	s_pxStaticPhotonStorage = 0;
	s_pxStaticPhotonTree = 0;
	s_uNumberOfStaticPhotons = 0;

    g_pxDynamicPhotonBVH = new Asura_DynamicBVH<Asura_Photon>();
    g_bForceRecreateOnDynamicPhotonBVH = true;
}

void Asura_Photon_System::Reset(void)
{
	Asura_Assert(s_pxDynamicPhotonList, "Reset called when not Initialised - crash imminent.");
	s_pxDynamicPhotonList->Reset();

    // (MH - 15.12.2002) these should have been deleted by the lighting reset
	Asura_Assert(!s_uNumberOfStaticPhotons, "shouldn't be any static photons when Reset() is called");
    Asura_Assert(!s_pxStaticPhotonStorage, "shouldn't be any static photons when Reset() is called");
    Asura_Assert(!s_pxStaticPhotonTree, "shouldn't be any static photons when Reset() is called");

    g_pxDynamicPhotonBVH->Reset();
    g_bForceRecreateOnDynamicPhotonBVH = true;
}

void Asura_Photon_System::Update(void)
{

}

void Asura_Photon_System::Shutdown(void)
{
	delete s_pxTemporaryAveragePhoton;
	s_pxTemporaryAveragePhoton = NULL;

	delete s_pxDynamicPhotonList;
	s_pxDynamicPhotonList = NULL;

	delete [] s_pxStaticPhotonStorage;
	s_pxStaticPhotonStorage = NULL;

	delete s_pxStaticPhotonTree;
	s_pxStaticPhotonTree = NULL;

	s_uNumberOfStaticPhotons = 0;

    delete g_pxDynamicPhotonBVH;
    g_pxDynamicPhotonBVH = NULL;
}

Asura_Photon* Asura_Photon_System::CreatePhoton(void)
{
	Asura_Handle iHandle = s_pxDynamicPhotonList->Allocate();

	if (iHandle == ASURA_INVALID_HANDLE)
	{
		return NULL;
	}

	Asura_Photon* pxPhoton = s_pxDynamicPhotonList->GetPointer(iHandle);

	pxPhoton->m_uFlags = 0;
	pxPhoton->m_bActive = true;
	pxPhoton->m_iHandle = iHandle;

	Asura_ShadowParameters xNewParams;
	pxPhoton->m_xShadowParams = xNewParams;

#if defined (ASURA_TARGET_SM3)
	pxPhoton->m_fFalloffScale			= 1.0f;
	pxPhoton->m_fFalloffPower			= 1.0f;	
	pxPhoton->m_bUsingCustomLightVolume = false;	
	pxPhoton->m_bIsDirectional			= false;
	pxPhoton->m_fDistanceFadeStart		= 0.0f;
	pxPhoton->m_fDistanceFadeDist		= 0.0f;
	pxPhoton->m_bDistanceFadeEnable		= false;
	pxPhoton->m_fInnerLimitAngle		= 0.0f;
	pxPhoton->m_fOuterLimitAngle		= 0.0f;
	pxPhoton->m_iProjCubeTextureIndex   = ASURA_TEXTURE_NONE;
	pxPhoton->m_iProjCubeBlurredTextureIndex   = ASURA_TEXTURE_NONE;
	pxPhoton->m_fProjCubeLerpDist			   = 0.0f;

#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM)
	pxPhoton->m_iLastKnownModule		= -1;
#endif

#endif

    g_bForceRecreateOnDynamicPhotonBVH = true;

	return pxPhoton;
}

void Asura_Photon_System::DestroyPhoton(const Asura_Photon* pxPhoton)
{
    // The delete operator, etc., tolerates NULL ptrs so it kind of makes
    // sense that we should here as well.
    if( !pxPhoton ) { return; }
	
#ifdef ASURA_ASSERT
	Asura_Handle iHandle = pxPhoton->m_iHandle;
	
	Asura_Assert( iHandle != ASURA_INVALID_HANDLE, "invalid handle" );

	Asura_Photon* pxPhotonTest = s_pxDynamicPhotonList->GetPointer(iHandle);

	Asura_Assert( pxPhotonTest == pxPhoton, "pointer mismatch" );
#endif

	s_pxDynamicPhotonList->Deallocate( pxPhoton->m_iHandle );

    g_bForceRecreateOnDynamicPhotonBVH = true;
}

Asura_Photon* Asura_Photon_System::GetPhotonPointer(Asura_Handle iHandle)
{
	Asura_Assert( iHandle!=ASURA_INVALID_HANDLE, "Invalid handle!" );
	
	return s_pxDynamicPhotonList->GetPointer(iHandle);
}

int Asura_Photon_System::PhotonSortSub(const void *cmp1, const void *cmp2)
{
	const Asura_Local_Photon *p1 = (const Asura_Local_Photon*)cmp1;
	const Asura_Local_Photon *p2 = (const Asura_Local_Photon*)cmp2;

	if (p1->Influence < p2->Influence)
        return  1;
    if (p1->Influence > p2->Influence)
        return -1;

    return 0;
}


void Asura_Photon_System::DynamicPhotonCallback( const Asura_Photon& xPhoton )
{
	if (!xPhoton.m_bActive) return;

	if(s_bGettingPhotonsForEnv && xPhoton.m_bDontLightEnv) return;
    
	/// Kevin Lea - 03/10/2002 15:06:47 - obviously want to avoid any photons with range == 0 (or -ve, for that matter)
	const float fPhotonRange = xPhoton.m_fRange;
	if (! (fPhotonRange > 0.0f) ) return;

// (MP - 24.03.2009) Already dealt with by the BVH test.
//	//Check against the photon's bounding box if it has one. RAR 10/03/03
//	if( xPhoton.m_bUseBoundingBox )
//	{
//		if(!s_pxProximityTestBoundingBox->Intersects(xPhoton.m_xBoundingBox))
//		{
//			return;
//		}
//	}

	bool intersects = s_pxProximityTestBoundingBox->IntersectsSphere(xPhoton.m_xPosition,fPhotonRange);
	if (intersects)
	{
		float distance = (xPhoton.m_xPosition-*s_pxProximityTestCentre).Magnitude();

		float range = fPhotonRange;
		float oorange = distance/range;
		float influence = 1.0f - oorange;
		if (influence<0.0f) influence=0.0f;
		
		influence *= (xPhoton.m_fR + xPhoton.m_fG + xPhoton.m_fB);

        u_int uIdx;
		if (NumberOfLocalPhotons==ASURA_MAX_NO_OF_LOCAL_PHOTONS)
		{
			int min = 0;
			for (u_int p=1; p<NumberOfLocalPhotons; p++)
			{
				if (LocalPhotons[min].Influence>LocalPhotons[p].Influence)
				{
					min = p;
				}
			}

            if( influence <= LocalPhotons[min].Influence ) { return; }

			s_fProximityTestMaxInfluence-=LocalPhotons[min].Influence;
            uIdx = min;
		}
		else
		{
            uIdx = NumberOfLocalPhotons++;
		}

		LocalPhotons[uIdx].Influence   = influence;
		LocalPhotons[uIdx].Photon      = &xPhoton;

		s_fProximityTestMaxInfluence += influence;       
	}
}


bool Asura_Photon_System::StaticPhotonCallback(unsigned short index)
{
	Asura_Assert( index < s_uNumberOfStaticPhotons, "index out of range" );
	Asura_Photon* pxPhoton = &s_pxStaticPhotonStorage[index];

	// If the test bounding box doesn't intersect this photon, carry on searching
	// (Due to the conservative compression of the AABB data, this function may be called
	// even though the real bounding boxes don't overlap.)
	if(pxPhoton->m_bUseBoundingBox)
	{
		if (!s_pxProximityTestBoundingBox->Intersects(pxPhoton->GetBoundingBox()) )
		{
			return true;
		}
	}
	
	float distance = (pxPhoton->m_xPosition-*s_pxProximityTestCentre).Magnitude();

	float range = pxPhoton->m_fRange;
	float oorange = distance/range;
	float influence = 1.0f - oorange;
	if (influence<0.0f) influence=0.0f;

	if (pxPhoton->IsShadowVolume())
	{
		float fShadow = 1.0f - pxPhoton->m_fR * influence;
		
		if (fShadow < 0.0f) fShadow = 0.0f;

		if (fShadow < s_fShadowVolumeScale ) s_fShadowVolumeScale = fShadow;
	}
	else
	{
		influence *= (pxPhoton->m_fR + pxPhoton->m_fG + pxPhoton->m_fB);

		u_int uIdx;
		if (NumberOfLocalPhotons==ASURA_MAX_NO_OF_LOCAL_PHOTONS)
		{
			int min = 0;
			for (u_int p=1; p<NumberOfLocalPhotons; p++)
			{
				if (LocalPhotons[min].Influence>LocalPhotons[p].Influence)
				{
					min = p;
				}
			}

			if( influence <= LocalPhotons[min].Influence )
			{
				return true;
			}

			s_fProximityTestMaxInfluence-=LocalPhotons[min].Influence;
			uIdx = min;
		}
		else
		{
			uIdx = NumberOfLocalPhotons++;
		}

		LocalPhotons[uIdx].Influence   = influence;
		LocalPhotons[uIdx].Photon      = pxPhoton;
//		LocalPhotons[uIdx].m_fDistance = distance; 

		s_fProximityTestMaxInfluence += influence;
	}
			
    return true;
}


const Asura_DynamicBVH<Asura_Photon>& Asura_Photon_System::GetDynamicPhotonBVH( void )
{
    // Refresh first as necessary.
    {
        const u_int uCurFrame = Asura_Timers::GetFrameNumber();
        if( g_bForceRecreateOnDynamicPhotonBVH
            ||
            ((g_uDynamicPhotonBVHRefitFrameStamp != uCurFrame) && (g_uNumDynamicPhotonBVHRefits >= uMAX_NUMBER_OF_DYNAMIC_PHOTON_BVH_REFITS)) )
        {
            Asura_TrackingHandlePool_Iterator<Asura_Photon> xIt( s_pxDynamicPhotonList );
            g_pxDynamicPhotonBVH->Create<DynamicPhotonBVH_BBRetrievalWrapper>( xIt );

            g_bForceRecreateOnDynamicPhotonBVH = false;
            g_uNumDynamicPhotonBVHRefits = 0;
            g_uDynamicPhotonBVHRefitFrameStamp = uCurFrame;
        }
        else if( g_uDynamicPhotonBVHRefitFrameStamp != uCurFrame )
        {
            g_pxDynamicPhotonBVH->Refit<DynamicPhotonBVH_BBRetrievalWrapper>();

            ++g_uNumDynamicPhotonBVHRefits;
            g_uDynamicPhotonBVHRefitFrameStamp = uCurFrame;
        }
    }

    return *g_pxDynamicPhotonBVH;
}


DEBUGVAR float dbg_fCheckBoundingBoxCentreTolerance = 0.0f;
void Asura_Photon_System::MakePhotonListForBoundingBox( const Asura_Bounding_Box& xBoundingBox,
													    const Asura_Vector_3& xBoundingBoxCentre,
                                                        const bool bIncludeStaticPhotons,
                                                        const bool bIncludeDynamicPhotons,
                                                        u_int uMaxNumberOfPhotons,
                                                        bool bResetLocalPhotons )
{
	if(bResetLocalPhotons)
	{
		NumberOfLocalPhotons = 0;
	}

	Asura_Assert(bIncludeDynamicPhotons || bIncludeStaticPhotons, "must request at least one type of photon");
	if (dbg_fCheckBoundingBoxCentreTolerance != 0.0f)
	{
		Asura_Assert(xBoundingBox.CalculateCentre().ApproximatelyEquals(xBoundingBoxCentre, dbg_fCheckBoundingBoxCentreTolerance), 
			"BB centre is not central");
	}

	s_fProximityTestMaxInfluence = 0.0f;
	s_pxProximityTestCentre = &xBoundingBoxCentre;
	s_pxProximityTestBoundingBox = &xBoundingBox;

	//Currently the environment uses just dynamic lights , and everything else asks for both static and dynamic.
	//RAR 21/09/2005
	s_bGettingPhotonsForEnv = !bIncludeStaticPhotons;

	if (bIncludeDynamicPhotons)
	{
        GetDynamicPhotonBVH().TestAABB( xBoundingBox, DynamicPhotonCallback );
	}

	if (bIncludeStaticPhotons && s_uNumberOfStaticPhotons)
	{
		Asura_Assert( s_pxStaticPhotonTree!=NULL, "no tree!" );

		s_fShadowVolumeScale = 1.0f;
		s_pxStaticPhotonTree->TestTree(xBoundingBox,StaticPhotonCallback);
	}
	s_bGettingPhotonsForEnv = false;


	qsort(LocalPhotons, NumberOfLocalPhotons, sizeof(Asura_Local_Photon), PhotonSortSub);

	if (uMaxNumberOfPhotons < NumberOfLocalPhotons)
	{
		NumberOfLocalPhotons = uMaxNumberOfPhotons;
		/// Kevin Lea - 03/10/2002 18:38:41 - TODO: use the last photon to represent an average of the 
		/// ones which have been ignored.
	}
}


void Asura_Photon_System::CalculateAverageLocalLightPos(Asura_Vector_3& xPos)
{
	xPos.Zero();

	if (s_fProximityTestMaxInfluence>0.0f)
	{
		for (u_int u = 0; u < NumberOfLocalPhotons; u++)
		{
			xPos += LocalPhotons[u].Photon->m_xPosition * LocalPhotons[u].Influence;
		}
		xPos/=s_fProximityTestMaxInfluence;
	}
	else
	{
		xPos.y = -10000.0f;
	}
}
	
const Asura_Photon_Sampling_Result* Asura_Photon_System::SamplePhotonList(const Asura_Vector_3& xPosition)
{
	s_xPhotonResult.m_fDiffuseR = 0.0f;
	s_xPhotonResult.m_fDiffuseG = 0.0f;
	s_xPhotonResult.m_fDiffuseB = 0.0f;

	float fMaxInfluence = 0;
	const Asura_Photon* pxMostInfluentialPhoton = 0;
	float fMostInfluentialPhotonsScale = 0;
	
	for (u_int u=0; u<Asura_Photon_System::NumberOfLocalPhotons; u++)
	{
		const Asura_Photon* pxPhoton = LocalPhotons[u].Photon;
	
		float fRange = pxPhoton->GetRange();

		Asura_Vector_3 xLightToPosition = xPosition-pxPhoton->GetPosition();

		float fDistance = xLightToPosition.Magnitude();

		if (fDistance<fRange)
		{
			Asura_Assert( fRange>0.0f, "Range is <= zero!" );

			float fScale = (fRange - fDistance) / fRange;
			
			s_xPhotonResult.m_fDiffuseR += fScale * pxPhoton->GetRedComponent();
			s_xPhotonResult.m_fDiffuseG += fScale * pxPhoton->GetGreenComponent();
			s_xPhotonResult.m_fDiffuseB += fScale * pxPhoton->GetBlueComponent();

			if (Asura_Photon_System::LocalPhotons[u].Influence > fMaxInfluence)
			{
				fMaxInfluence = Asura_Photon_System::LocalPhotons[u].Influence;
				pxMostInfluentialPhoton = pxPhoton;
				fMostInfluentialPhotonsScale = fScale;
			}
		}
		
	}

	if (pxMostInfluentialPhoton)
	{
		s_xPhotonResult.m_fSpecularR = pxMostInfluentialPhoton->m_fR * fMostInfluentialPhotonsScale;
		s_xPhotonResult.m_fSpecularG = pxMostInfluentialPhoton->m_fG * fMostInfluentialPhotonsScale;
		s_xPhotonResult.m_fSpecularB = pxMostInfluentialPhoton->m_fB * fMostInfluentialPhotonsScale;
	}
	else
	{
		s_xPhotonResult.m_fSpecularR = 0.0f;
		s_xPhotonResult.m_fSpecularG = 0.0f;
		s_xPhotonResult.m_fSpecularB = 0.0f;
	}

	return &s_xPhotonResult;
}

// Loop over all photons calling the callback function.
void Asura_Photon_System::TraverseAll(	void(*pfnCallback)(Asura_Photon* pxPhoton),
										const bool bIncludeStaticPhotons,
										const bool bIncludeDynamicPhotons )
{
	if (bIncludeDynamicPhotons)
	{
		s_pxDynamicPhotonList->Traverse(pfnCallback);
	}

	if (bIncludeStaticPhotons)
	{
		for (unsigned int u=0; u<s_uNumberOfStaticPhotons; ++u)
		{
			(*pfnCallback)(&s_pxStaticPhotonStorage[u]);
		}
	}
}

void Asura_Photon_System::TraverseAll(	void(*pfnCallback)(Asura_Photon* pxPhoton, void* pData),
										void* pData,
										const bool bIncludeStaticPhotons,
										const bool bIncludeDynamicPhotons )
{
	if (bIncludeDynamicPhotons)
	{
		s_pxDynamicPhotonList->Traverse(pfnCallback, pData);
	}

	if (bIncludeStaticPhotons)
	{
		for (unsigned int u=0; u<s_uNumberOfStaticPhotons; ++u)
		{
			(*pfnCallback)(&s_pxStaticPhotonStorage[u], pData);
		}
	}
}

Asura_Photon_Iterator Asura_Photon_System::GetDynamicIterator(void)
{
	return GetDynamicPhotonList()->GetIterator();
}


bool Asura_Photon_System::AABBTreeCreateCallback( u_int uNumStaticPhotons, Asura_Bounding_Box* pxBBs, Asura_Vector_3* )
{
    Asura_Assert( uNumStaticPhotons == s_uNumberOfStaticPhotons, "" );
    
    for( u_int u = 0; u < uNumStaticPhotons; ++u )
    {
        const Asura_Photon& xPhoton = s_pxStaticPhotonStorage[u];

		pxBBs[u].Reset();

		{
		   float fRange = xPhoton.GetRange();
			Asura_Vector_3 xSize( fRange, fRange, fRange );

			Asura_Vector_3 xPosition = xPhoton.GetPosition() - xSize;
			pxBBs[u].Set(xPosition);

			xPosition = xPhoton.GetPosition() + xSize;
			pxBBs[u].Set(xPosition);
		}

		if( xPhoton.GetUseBoundingBox() )
		{
			const Asura_Bounding_Box& xPhotonBB = xPhoton.GetBoundingBox();

			//the light has its own cutoff bounding box... get the intersection of this and the
			//range bounding box
			if( pxBBs[u].MinX < xPhotonBB.MinX) pxBBs[u].MinX = xPhotonBB.MinX;
			if( pxBBs[u].MinY < xPhotonBB.MinY) pxBBs[u].MinY = xPhotonBB.MinY;
			if( pxBBs[u].MinZ < xPhotonBB.MinZ) pxBBs[u].MinZ = xPhotonBB.MinZ;
			if( pxBBs[u].MaxX > xPhotonBB.MaxX) pxBBs[u].MaxX = xPhotonBB.MaxX;
			if( pxBBs[u].MaxY > xPhotonBB.MaxY) pxBBs[u].MaxY = xPhotonBB.MaxY;
			if( pxBBs[u].MaxZ > xPhotonBB.MaxZ) pxBBs[u].MaxZ = xPhotonBB.MaxZ;

			//make sure this hasn't left any of the Maxs less than the Mins
			if( pxBBs[u].MaxX < pxBBs[u].MinX ) pxBBs[u].MaxX = pxBBs[u].MinX;
			if( pxBBs[u].MaxY < pxBBs[u].MinY ) pxBBs[u].MaxY = pxBBs[u].MinY;
			if( pxBBs[u].MaxZ < pxBBs[u].MinZ ) pxBBs[u].MaxZ = pxBBs[u].MinZ;
		}        

	}

    // Let tree creator class calculate centres for us.
    return false;
}

float Asura_Photon_System::GetEntityAmbientRed()
{
	return s_fShadowVolumeScale * Asura_Lighting_System::s_fEntityAmbientR;
}
float Asura_Photon_System::GetEntityAmbientGreen()
{
	return s_fShadowVolumeScale * Asura_Lighting_System::s_fEntityAmbientG;
}
float Asura_Photon_System::GetEntityAmbientBlue()
{
	return s_fShadowVolumeScale * Asura_Lighting_System::s_fEntityAmbientB;
}

/*
void Asura_Photon_System::GetSHRadiance(const Asura_Vector_3& xCentre, const float fRadius, Asura_SHRadiance_Result& xResult)
{
	Asura_Vector_3 xOffset(fRadius, fRadius, fRadius);

	Asura_Bounding_Box xBB;
	xBB.Reset();
	xBB.Set(xCentre+xOffset);
	xBB.Set(xCentre-xOffset);

	Asura_Photon_System::MakePhotonListForBoundingBox(xBB, xCentre, true, true);

	Asura_Vector_3 xSH[9];

	for (u_int u=0; u<9; u++)
	{
		xSH[u].Zero();
	}

	const float fConst1 = 0.235294465f;
	const float fConst2 = 0.470589175497f;
	const float fConst3 = 0.882352245676f;
	const float fConst4 = 0.073529614476f;
	const float fConst5 = 0.220588061419f;

	for (u_int p=0; p<Asura_Photon_System::NumberOfLocalPhotons; p++)
	{
		const Asura_Photon& xPhoton = Asura_Photon_System::LocalPhotons[p].Photon;

		Asura_Vector_3 xDirection = xPhoton->GetPosition() - position;
		
		float xDistance = xDirection.Magnitude() - fRadius;

		float fLightScale;
		Asura_Vector_3 xColour;
		
		xColour.x = Asura_Photon_System::LocalPhotons[p].Photon->GetRedComponent();
		xColour.y = Asura_Photon_System::LocalPhotons[p].Photon->GetGreenComponent();
		xColour.z = Asura_Photon_System::LocalPhotons[p].Photon->GetBlueComponent();

		if (xDistance>0.0f)
		{
			// KJL - 07/11/2003 15:47:38 - lots of division by zero possibilities here. Need to give Photons a GetOneOverRange() accessor too.
			float fScale = (xPhoton->GetRange() - xDistance) / xPhoton->GetRange();

			xColour *= fScale;
		}

		xDirection.Normalise();

		xSH[0] += xColour * fConst1;
		xSH[1] += xColour * fConst2 * xDirection.x;
		xSH[2] += xColour * fConst2 * xDirection.y;
		xSH[3] += xColour * fConst2 * xDirection.z;
		xSH[4] += xColour * fConst3 * (xDirection.x*xDirection.z);
		xSH[5] += xColour * fConst3 * (xDirection.z*xDirection.y);
		xSH[6] += xColour * fConst3 * (xDirection.y*xDirection.x);
		xSH[7] += xColour * fConst4 * (3.0f*xDirection.z*xDirection.z-1.0f);
		xSH[8] += xColour * fConst5 * (xDirection.x*xDirection.x - xDirection.y*xDirection.y);
	}

	// KJL - 07/11/2003 15:50:38 - now rearrange SH terms into the 7 4-vectors we'll use for rendering 

	xResult.m_xSHConstants[0].m_fX = xSH[1].x;
	xResult.m_xSHConstants[0].m_fY = xSH[2].x;
	xResult.m_xSHConstants[0].m_fZ = xSH[3].x;
	xResult.m_xSHConstants[0].m_fW = xSH[0].x - xSH[7].x;

	xResult.m_xSHConstants[1].m_fX = xSH[1].y;
	xResult.m_xSHConstants[1].m_fY = xSH[2].y;
	xResult.m_xSHConstants[1].m_fZ = xSH[3].y;
	xResult.m_xSHConstants[1].m_fW = xSH[0].y - xSH[7].y;
	
	xResult.m_xSHConstants[2].m_fX = xSH[1].z;
	xResult.m_xSHConstants[2].m_fY = xSH[2].z;
	xResult.m_xSHConstants[2].m_fZ = xSH[3].z;
	xResult.m_xSHConstants[2].m_fW = xSH[0].z - xSH[7].z;
	
	xResult.m_xSHConstants[3].m_fX = xSH[4].x;
	xResult.m_xSHConstants[3].m_fY = xSH[5].x;
	xResult.m_xSHConstants[3].m_fZ = xSH[6].x;
	xResult.m_xSHConstants[3].m_fW = 3.0f * xSH[7].x;
	
	xResult.m_xSHConstants[4].m_fX = xSH[4].y;
	xResult.m_xSHConstants[4].m_fY = xSH[5].y;
	xResult.m_xSHConstants[4].m_fZ = xSH[6].y;
	xResult.m_xSHConstants[4].m_fW = 3.0f * xSH[7].y;
	
	xResult.m_xSHConstants[5].m_fX = xSH[4].z;
	xResult.m_xSHConstants[5].m_fY = xSH[5].z;
	xResult.m_xSHConstants[5].m_fZ = xSH[6].z;
	xResult.m_xSHConstants[5].m_fW = 3.0f * xSH[7].z;
	
	xResult.m_xSHConstants[6].m_fX = xSH[8].x;
	xResult.m_xSHConstants[6].m_fY = xSH[8].y;
	xResult.m_xSHConstants[6].m_fZ = xSH[8].z;
	xResult.m_xSHConstants[6].m_fW = 0.0f;  // this last float isn't necessarily used, so it could store a useful constant etc.
}
*/



#ifdef ASURA_TARGET_SM3

void Asura_Photon::RenderLightVolume(void)
{
	
	bool bPrev = Asura_Render::GetEnableBatchedLineRendering();
	Asura_Render::SetEnableBatchedLineRendering(false);

	for (int iPass=0; iPass<2; iPass++)
	{
		// Pass 0 - Translucent bits - behind scenery
		// Pass 1 - Opaque visible bits
		if (iPass == 0)
		{
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
			Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_GREATER);
		}
		else
		{
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
			Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
		}		

		uint32 uCol;

		float fDistSquared = GetDistanceFadeStart()+GetDistanceFadeDist();
		fDistSquared*=fDistSquared;
		Asura_Vector_3 xVec = *Asura_Camera::GetPosition() - GetPosition();

		if (xVec.MagnitudeSquared() > fDistSquared && GetDistanceFadeEnable())			
		{
			if (iPass == 0)
				uCol = 0x18888888;							
			else
				uCol = 0x80888888;			
		}
		else
		{
			if (iPass == 0)
				uCol = 0x20ffff00;			
			else
				uCol = 0xffffff00;			
		}

		Asura_Bounding_Box  xBB;
		Asura_Matrix_3x3    xOri;
		Asura_Vector_3		xPos;

		if (GetCustomLightVolumeEnable())
		{	
			// Custom light volume - always rendered as a cuboid
			xBB					= GetCustomLightVolume();
			Asura_Vector_3 xCtr	= xBB.CalculateCentre();
			xPos				= xCtr; 
			xBB.Displace(-xCtr);			
			GetCustomLightVolumeOrient().ToMatrix(xOri);
			Asura_Render::RenderBB(xBB, xPos,xOri, uCol);		
		}
		else
		{				
			
			
			if (Asura_DeferredShading::GetUseCubesForPhotons())
			{
				// Render it as a cube			
				xOri.Identity();		
				xBB.Reset();
				float fRange = GetRange();
				xBB.Set(GetPosition()+Asura_Vector_3(fRange,fRange,fRange));
				xBB.Set(GetPosition()-Asura_Vector_3(fRange,fRange,fRange));
				xPos = Asura_Vector_3(0,0,0);
				Asura_Render::RenderBB(xBB, xPos,xOri, uCol);		
			}
			else
			{
				// Render it as a sphere			
				xPos = GetPosition();		

				float fRadius = GetRange();
				Asura_Vector_3 xUp = Asura_Vector_3(0,1,0);
				Asura_Vector_3 xRight = Asura_Vector_3(1,0,0);
				Asura_Vector_3 xForwards = Asura_Vector_3(0,0,1);
					
				// Render the photon as a circle
				for(int i=0;i<16;i++)
				{
					float fAngle  = (Asura_Maths::Pi/8.0f) * (float)(i);
					float fAngle2 = (Asura_Maths::Pi/8.0f) * (float)(i+1);
					Asura_Vector_3 xA = xPos+ (xUp*(fRadius*Asura_Maths::Sine(fAngle)))+(xRight*(fRadius*Asura_Maths::Cosine(fAngle)));
					Asura_Vector_3 xB = xPos+ (xUp*(fRadius*Asura_Maths::Sine(fAngle2))) +(xRight*(fRadius*Asura_Maths::Cosine(fAngle2)));
					Asura_Render::RenderLine(xA,xB,uCol);

					xA = xPos+ (xUp*(fRadius*Asura_Maths::Sine(fAngle)))+(xForwards*(fRadius*Asura_Maths::Cosine(fAngle)));
					xB = xPos+ (xUp*(fRadius*Asura_Maths::Sine(fAngle2))) +(xForwards*(fRadius*Asura_Maths::Cosine(fAngle2)));
					Asura_Render::RenderLine(xA,xB,uCol);

					xA = xPos+ (xForwards*(fRadius*Asura_Maths::Sine(fAngle)))+(xRight*(fRadius*Asura_Maths::Cosine(fAngle)));
					xB = xPos+ (xForwards*(fRadius*Asura_Maths::Sine(fAngle2))) +(xRight*(fRadius*Asura_Maths::Cosine(fAngle2)));
					Asura_Render::RenderLine(xA,xB,uCol);
				}			
			}
		}


	}

	Asura_Render::SetEnableBatchedLineRendering(bPrev);
}

#if defined (ASURA_USE_NEW_VISIBILITY_SYSTEM)
bool Asura_Photon::IsVisible( void )
{
	const Asura_Vector_3& xPosition = GetPosition();		
	const float fRadius = GetRange();
		
	{
		Asura_Bounding_Box xBB;
		xBB.SetToPoint( xPosition );
		xBB.Grow( fRadius );

		if ( !m_xVisibilityQuery.IsAABBVisibleFromCurrentCamera( xBB ) ) 
		{
			return false;
		}
	}

	if ( !Asura_Camera::IsSphereInFrustum( xPosition, fRadius ) )
	{
		return false;
	}

	return true;
}
#endif

#endif
