//////////////////////////////////////////////////////////////////////////////
//
// Asura_CollisionMesh.h
//
// Created by Mike Healey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Data structure for dynamics and LOS
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_COLLISIONMESH_H___
#define ___ASURA_COLLISIONMESH_H___ 1

//////////////
// Includes //
//////////////
#include "Core/Asura_Hash.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Material_System.h"
#include "Debugging/Asura_Assert.h"
#include "Dynamics/Asura_Collision_Object.h"
#include "Mesh/Asura_Mesh.h"


//////////////////////////
// Forward declarations //
//////////////////////////
class Asura_Mesh;
class Asura_Chunk_Stream;


////////////////////////////////////////////////
// Asura_CollisionMesh_Poly: triangle or quad //
////////////////////////////////////////////////
class Asura_CollisionMesh_Poly
{
public:
	inline bool IsQuad() const { return m_usVertexID[3] != 0xffff; }

	u_short m_usVertexID[4]; // Fourth entry is 0xffff for triangles.
};


/////////////////////////
// Asura_CollisionMesh //
/////////////////////////
class Asura_CollisionMesh : public Asura_Collision_Object
{
public:
	//constructor for an empty collision mesh
	Asura_CollisionMesh();
	// Constructor for chunk loader as part of shape, hierarchy or environment module chunks.
    Asura_CollisionMesh(const char*& pcChunkBuffer);
    virtual ~Asura_CollisionMesh();

	//Reads in from chunk stream to an empty newly created collision mesh
	void ReadFromChunkStream(Asura_Chunk_Stream& xStream, bool bConvertMaterialIDs = true, int* piOutputVersion = 0);

#ifdef ASURA_TOOLS
	// For access by stripifier in case collision vertices are a superset of render vertices
	friend class Tools_EnvironmentProcessor_PosMesh;

	// Typedef...
	typedef bool fnProjectExcludePolygonFromExport (const Asura_Mesh& xMesh, const u_int uPolygonIndex) const;

    // Saving as part of a chunk.
	int GetChunkBufferSize() const;
	void FillInChunkBuffer(char*& pcChunkBuffer,const int* piMaterialIDToIndexConversion = 0) const;

	//Saves as a chunk stream. 
	//If bPrefixWithSize is true then the first 4 bytes written will be the size of the rest of the collision mesh data
	void WriteToChunkStream(Asura_Chunk_Stream& xStream, bool bPrefixWithSize,const int* piMaterialIDToIndexConversion = 0) const;


    // Constructor for tools generation from mesh.
	// piMaterialIndexToIDConversion is used to convert the material indeces back into IDs ,if
	// the asura meshes have been converted for saving.
    Asura_CollisionMesh(class Asura_Mesh * pxAsuraMesh);

	//copy constructor
	Asura_CollisionMesh(const Asura_CollisionMesh * pxCollisionMesh);

	// Destroy the polygons, but leaves the BB...
	void DestroyPolygons(void);

	// Recalculate the collision mesh from an Asura_Mesh
	void UpdateFromMesh(class Asura_Mesh * pxAsuraMesh);

	//copy from another collision mesh
	void Copy(const Asura_CollisionMesh * pxCollisionMesh);

    // Removes any polys flagged as 'no collisions'.
    void RemoveNonCollisionPolys( void );

	// Reorders the collision mesh into an order convenient for LBAABB trees.
	void ReorderCollisionMesh(const u_int* puConversionArray);
#endif //ASURA_TOOLS

	// LOS test (per poly)
    bool LOSTestPolygons(const Asura_Vector_3& xPosition, 
						 const Asura_Matrix_3x3& xOrient,
						 const Asura_LOS_InputData* pxInput,
						 Asura_LOS_OutputData* pxOutput) const;


	// LOS test (uses m_bBoundingBoxOnly flag to decide which test to call)
	virtual bool LOSTest(const Asura_Vector_3& xPosition, 
						 const Asura_Matrix_3x3& xOrient,
						 const Asura_LOS_InputData* pxInput,
						 Asura_LOS_OutputData* pxOutput) const
	{
		if(pxInput->m_bBoundingSphereOnly)
		{
			return LOSTestSphere(xPosition, pxInput, pxOutput);
		}
		else if(pxInput->m_bBoundingBoxOnly)
		{
			return LOSTestRotatedBB(xPosition, xOrient, pxInput, pxOutput);
		}
		else
		{
			return LOSTestPolygons(xPosition, xOrient, pxInput, pxOutput);
		}
	}


	// MGH 7/10/2004: Return a pointer to the mesh.
	virtual const Asura_CollisionMesh* GetMesh() const { return this; }

    // Accessor functions
	unsigned int GetNumberOfPolys() const { return m_uNumberOfPolys; }
	unsigned int GetNumberOfVertices() const {return m_uNumberOfVertices; }

	int GetNumVerticesInPoly(u_int uPolyIndex) const 
	{
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		return m_pxPolys[uPolyIndex].IsQuad() ? 4 : 3;
	}

	bool IsQuad(u_int uPolyIndex) const 
	{
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		return m_pxPolys[uPolyIndex].IsQuad();
	}

	// Vertices
	void GetVertices(Asura_Vector_3 axVertexArray[4], u_int uPolyIndex) const
    {
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		const Asura_CollisionMesh_Poly& xPoly = m_pxPolys[uPolyIndex];
        axVertexArray[0] = m_pxVertices[xPoly.m_usVertexID[0]];
        axVertexArray[1] = m_pxVertices[xPoly.m_usVertexID[1]];
        axVertexArray[2] = m_pxVertices[xPoly.m_usVertexID[2]];

		if (xPoly.IsQuad())
		{
			axVertexArray[3] = m_pxVertices[xPoly.m_usVertexID[3]];
		}
    }

	void GetVertices( Asura_AVector3 axVertexArray[3], u_int uPolyIndex, bool bGetFirstTriOfQuad ) const
	{
		Asura_Assert( uPolyIndex<m_uNumberOfPolys, "Code bug: invalid poly index" );

		const Asura_CollisionMesh_Poly& xPoly = m_pxPolys[uPolyIndex];
		
		if ( bGetFirstTriOfQuad )
		{
			axVertexArray[0] = m_pxVertices[xPoly.m_usVertexID[0]];
			axVertexArray[1] = m_pxVertices[xPoly.m_usVertexID[1]];
			axVertexArray[2] = m_pxVertices[xPoly.m_usVertexID[2]];
		}
		else
		{
			Asura_Assert( xPoly.IsQuad(), "Code bug: poly isn't a quad" );
			axVertexArray[0] = m_pxVertices[xPoly.m_usVertexID[2]];
			axVertexArray[1] = m_pxVertices[xPoly.m_usVertexID[3]];
			axVertexArray[2] = m_pxVertices[xPoly.m_usVertexID[0]];
		}
	}


	const Asura_Vector_3& GetVertex(u_int uIndex) const
	{
		Asura_Assert(uIndex<m_uNumberOfVertices,"Invalid index");
		return m_pxVertices[uIndex];
	}

	u_int GetVertexIndex(u_int uPolyIndex,u_int uVertex) const
	{
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		const Asura_CollisionMesh_Poly& xPoly = m_pxPolys[uPolyIndex];
		return xPoly.m_usVertexID[uVertex];
	}

	// Normals
	void GetNormal(unsigned int uPolyIndex, Asura_Vector_3& xOutput) const
    {
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		const Asura_Vector_3& xV0 = m_pxVertices[m_pxPolys[uPolyIndex].m_usVertexID[0]];
		const Asura_Vector_3& xV1 = m_pxVertices[m_pxPolys[uPolyIndex].m_usVertexID[1]];
		const Asura_Vector_3& xV2 = m_pxVertices[m_pxPolys[uPolyIndex].m_usVertexID[2]];

	    xOutput = ((xV1 - xV0)^(xV2 - xV0));
		xOutput.Normalise();
    }

	// Flags
	short GetFlags(unsigned int uPolyIndex) const
	{
		u_short usFlags = m_pusPolygonFlags ? m_pusPolygonFlags[uPolyIndex] : m_usOverallFlags;
		//Check for any material flags that need to be transferred to the polygon flags
		//(For polygons with texture anims)
		//RAR 05/06/2006
		short sMaterial = GetMaterialID(uPolyIndex);
		u_int uMatFlags = Asura_Material_System::GetMaterialFlags(sMaterial);
		if(uMatFlags)
		{
			if(uMatFlags & AsuraMaterialFlag_DontRender) usFlags |= AsuraPolygonFlag_DontRender;
			if(uMatFlags & AsuraMaterialFlag_NoCollisions) usFlags |= AsuraPolygonFlag_NoCollisions;
			if(uMatFlags & AsuraMaterialFlag_IgnoreLOS) usFlags |= AsuraPolygonFlag_IgnoreLOS;
		}
		return usFlags;
	}
    //Gets the polygon flags ignoring the effects of any texture animation
	short GetFlagsIgnoringAnim(unsigned int uPolyIndex) const
	{
		return m_pusPolygonFlags ? m_pusPolygonFlags[uPolyIndex] : m_usOverallFlags;
	}
    u_short GetOverallFlags() const
    {
        return m_usOverallFlags;
    }
#ifdef ASURA_TOOLS
    u_int Tools_GetFlags(unsigned int uPolyIndex) const
    {
        u_int uFlags = m_puToolsPolygonFlags ? m_puToolsPolygonFlags[uPolyIndex] : m_uToolsOverallFlags;
        //Check for any material flags that need to be transferred to the polygon flags
        //(For polygons with texture anims)
        //RAR 05/06/2006
        short sMaterial = GetMaterialID(uPolyIndex);
        u_int uMatFlags = Asura_Material_System::GetMaterialFlags(sMaterial);
        if(uMatFlags)
        {
            if(uMatFlags & AsuraMaterialFlag_DontRender)   uFlags |= AsuraPolygonFlag_DontRender;
            if(uMatFlags & AsuraMaterialFlag_NoCollisions) uFlags |= AsuraPolygonFlag_NoCollisions;
            if(uMatFlags & AsuraMaterialFlag_IgnoreLOS)    uFlags |= AsuraPolygonFlag_IgnoreLOS;
        }
        return uFlags;
    }
    u_int Tools_GetFlagsIgnoringAnim(unsigned int uPolyIndex) const
    {
        return m_puToolsPolygonFlags ? m_puToolsPolygonFlags[uPolyIndex] : m_uToolsOverallFlags;
    }
    u_int Tools_GetOverallFlags() const
    {
        return m_uToolsOverallFlags;
    }
#endif

	short GetOverallMaterial() const
	{
		return m_sOverallMaterial;
	}

	short GetMaterialID(unsigned int uPolyIndex) const
	{
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		return m_psMaterialIDs ? m_psMaterialIDs[uPolyIndex] : m_sOverallMaterial;
	}

	Asura_Hash_ID GetOverallMaterialResponse() const
	{
		return m_uOverallMaterialResponse;
	}

	Asura_Hash_ID GetMaterialResponseHashID(unsigned int uPolyIndex) const
	{
		Asura_Assert(uPolyIndex<m_uNumberOfPolys, "Invalid index");
		return m_puMaterialResponseHashIDs ? m_puMaterialResponseHashIDs[uPolyIndex] : m_uOverallMaterialResponse;
	}

	// Fill in collision poly array
	virtual int FillPolyArray(  int iMaxNumber,
								Asura_Collision_Poly*& pxPolyPtr,
								const Asura_Vector_3& xPosition,
								const Asura_Matrix_3x3* pxOrient) const;

#ifdef ASURA_TOOLS
    // for waypointing only - return the flags as well
	virtual int FillPolyArray(  int iMaxNumber,
								Asura_Collision_Poly*& pxPolyPtr,
                                u_short*& pusFlagsPtr,
								const Asura_Vector_3& xPosition,
								const Asura_Matrix_3x3* pxOrient) const;
#endif

	void SetAllMaterialIDs(short iMaterialID);
	void SetAllMaterialResponseHashIDs(Asura_Hash_ID uMaterialResponseHashID);

#ifdef ASURA_TOOLS
	void SetOverallMaterialID(short sMaterialID) {m_sOverallMaterial = sMaterialID;}

	bool HasPerPolyMaterialID() const {return m_psMaterialIDs != NULL;}

	void SetMaterialID(u_int uPoly, short sMaterialID);
#endif

#ifdef ASURA_TOOLS
	bool IsValid(){return m_bValid;}
#endif

#ifdef ASURA_TOOLS
	static void SetIncludeDontExportPolys(bool bSet){s_bIncludeDontExportPolys = bSet;}
#endif

#ifdef ASURA_TOOLS
	//Converts materials to and from indecs for saving , using conversion arrays generated
	//by Tools_SaveAsura::PrepareMeshMaterialConversionLists
	void ConvertMaterialIDsToIndeces(const int* piConversion);
	void ConvertMaterialIndecesToIDs(const int* piConversion);

	//reset the old style materials to ASURA_MATERIAL_NONE
	void ResetMaterialIDs();
#endif

#ifdef ASURA_TOOLS
	//Delete the polygons from the collision mesh that are flagged 'true' in the array
	void DeletePolygons(const bool* pbDeleteArray);

	//merge another collision mesh into this one
	void AddCollisionMesh(const Asura_CollisionMesh* pxCollMesh,const Asura_Matrix_3x3& xMat,const Asura_Vector_3& xOffset,float fScale = 1.0f);
	void AddCollisionMesh(const Asura_CollisionMesh* pxCollMesh,const Asura_Matrix_3x3& xMat,const Asura_Vector_3& xOffset,const Asura_Vector_3& xScale,u_int uPolygonFlagsToAdd);

	//recaculate the bounding box and radius from the verts
	//RAR 20/03/2006 If bResetExisting is false then the current bounding box is expanded to include
	//the vertices in the mesh , instead of being calculated from scratch
	void CalculateBoundingBox(bool bResetExisting = true);

	// Set the function to check whether collision polys need excluding from export...
	static	void	SetProjectExcludePolygonFromExport(fnProjectExcludePolygonFromExport* pfnFunction)	{ s_pfnProjectExcludePolygonFromExport	= pfnFunction; }

	//applies a rotate to all the vertices in the collision mesh
	void Rotate(const Asura_Quat& xQuat);
#endif

	const Asura_Vector_3* GetVertexArray() { return m_pxVertices; }

private:
    // Vertex data
	u_int m_uNumberOfVertices;
	Asura_Vector_3* m_pxVertices;

	// Polys and flags
	u_int m_uNumberOfPolys;
	Asura_CollisionMesh_Poly* m_pxPolys;
	unsigned short* m_pusPolygonFlags; // Per polygon flags (or NULL if all the same)
#ifdef ASURA_TOOLS
    unsigned int*   m_puToolsPolygonFlags; // Per polygon flags (or NULL if all the same)
#endif

	// Per polygon material IDs (or NULL if all the same)
	Asura_Hash_ID* m_puMaterialResponseHashIDs;
	short* m_psMaterialIDs;

#ifdef ASURA_TOOLS
	bool m_bValid;//used to mark collision meshes that weren't loaded correctly

	static bool s_bIncludeDontExportPolys;

	// Static function pointer used for project-specific code which
	// checks whether a given set of flags allows the specified
	// collision polygon to be excluded from the exported mesh.
	static	fnProjectExcludePolygonFromExport*	s_pfnProjectExcludePolygonFromExport;
#endif

	// Private functions...
	void	UpdatePolygonFromMesh(const Asura_Mesh* pxAsuraMesh, const u_int uMeshPoly, const u_int uCollPoly, const bool bFlipPolygon,
								  int* piVertConv, bool& bCanDiscardFlags, bool& bCanDiscardMaterialIDs);
};

#endif // #ifndef ___ASURA_COLLISIONMESH_H___
