//////////////////////////////////////////////////////////////////////////////
//
// Asura_InternetSimulation.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     functionality for introducing latency and packet loss to network
//     traffic
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Asura_InternetSimulation.h"

#ifdef ASURA_ENABLE_INTERNET_SIMULATION

// asura includes
#include "Core/Asura_Maths.h"
#include "Core/Asura_MemHeap.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// D E B U G V A R S /////////////////////////////////////////////////////////

// I N T E R N A L S   S T R U C T U R E S ///////////////////////////////////
struct Asura_NetSim_PacketHeader
{
	u_int m_uDeliveryTime;
	int m_iLength;
	Asura_NetAddress m_xAddress;
	u_short m_usPort;
	u_short m_usPadding;

	Asura_NetSim_PacketHeader* m_pxNext;
	Asura_NetSim_PacketHeader* m_pxPrev;
};

// C O N S T A N T S   A N D   G L O B A L S /////////////////////////////////
const u_int uASURA_NETSIM_BUFFERSIZE = 128*1024; // in bytes
static Asura_MemHeap* g_pxMemHeap = NULL;
static void* g_pBuffer = NULL;
static Asura_NetSim_PacketHeader* g_pxPacketHead = NULL;
static bool g_bInUpdate = false;

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_InternetSimulation::s_bEnabled = false;
Asura_NetSim_Settings Asura_InternetSimulation::s_xCurrentSettings;

// C L A S S   M E T H O D S /////////////////////////////////////////////////
void Asura_InternetSimulation::Initialise()
{
	// set defaults
	s_bEnabled = false;
	s_xCurrentSettings.m_uLatencyMean = 30;
	s_xCurrentSettings.m_uLatencyStddev = 3;
	s_xCurrentSettings.m_fDropPercent = 5.0f;

	Asura_Assert(!g_pBuffer, "already initialised");
	g_pBuffer = Asura_Allocate( uASURA_NETSIM_BUFFERSIZE );
	if (g_pBuffer)
	{
		g_pxMemHeap = new Asura_MemHeap( g_pBuffer, uASURA_NETSIM_BUFFERSIZE );
	}
	g_pxPacketHead = NULL;
	g_bInUpdate = false;
}

void Asura_InternetSimulation::Shutdown()
{
	// no need to iterate throught the linked list: all memory was allocated from our own heap
	if (g_pxMemHeap)
	{
		g_pxMemHeap->Wipe();
	}
	delete g_pxMemHeap;
	g_pxMemHeap = NULL;
	Asura_Deallocate(g_pBuffer);
	g_pBuffer = NULL;

	g_pxPacketHead = NULL;
}

void Asura_InternetSimulation::SetLatencyMean( u_int uLatencyMean )
{
	s_xCurrentSettings.m_uLatencyMean = uLatencyMean;
}

void Asura_InternetSimulation::SetLatencyStandardDeviation( u_int uLatencyStddev )
{
	s_xCurrentSettings.m_uLatencyStddev = uLatencyStddev;
}

void Asura_InternetSimulation::SetDropPercent( float fDropPercent )
{
	Asura_Assert( fDropPercent >= 0.0f && fDropPercent <= 100.0f, "invalid percentage" );
	s_xCurrentSettings.m_fDropPercent = fDropPercent;
}

void Asura_InternetSimulation::SetSettings( const Asura_NetSim_Settings& xSettings )
{ 
	Asura_Assert( xSettings.m_fDropPercent >= 0.0f && xSettings.m_fDropPercent <= 100.0f, "invalid percentage" );
	s_xCurrentSettings = xSettings;
}

bool Asura_InternetSimulation::AddToBuffer( const char* pcMessage, int iLength, Asura_NetAddress xAddress, unsigned short usPort )
{
	if (!s_bEnabled)
	{
		return false;
	}

	if (ShouldDrop())
	{
		// drop, so claim to have added it but do nothing
		return true;
	}

	if (!g_pxMemHeap) 
	{
		// not initialised
		return false;
	}

	if (g_bInUpdate)
	{
		// don't want to add to the linked list whilst we're iterating through it
		Asura_Assert(false, "unexpected: AddToBuffer() called while doing Update()");
		return false;
	}

	u_int uAddedLatency = GetAddedLatency();
	if (!uAddedLatency)
	{
		// deliver immediately
		return false;
	}

	void* pMem = g_pxMemHeap->Alloc( iLength + sizeof(Asura_NetSim_PacketHeader) );
	if (!pMem)
	{
		// out of room to store packets. don't drop.
		return false;
	}

	// data goes straight after the header
	Asura_NetSim_PacketHeader* pxNewHdr = static_cast<Asura_NetSim_PacketHeader*>(pMem);
	Asura_Mem::Copy( pxNewHdr + 1, pcMessage, iLength );

	u_int uDeliveryTime = Asura_Timers::GetMSecsSinceStart_Continuous() + uAddedLatency;

	// insert in delivery order
	{
		Asura_NetSim_PacketHeader* pxNextHdr = g_pxPacketHead;
		Asura_NetSim_PacketHeader* pxPrevHdr = NULL;

		while (pxNextHdr)
		{

			// note that we check for greater-than so that messages which arrive at the same
			// game time are added and then delivered in the order they arrive.
			if ( pxNextHdr->m_uDeliveryTime > uDeliveryTime )
			{
				// insert here
				break;
			}
			else
			{
				pxPrevHdr = pxNextHdr;
				pxNextHdr = pxNextHdr->m_pxNext;
			}
		}

		pxNewHdr->m_pxPrev = pxPrevHdr;
		pxNewHdr->m_pxNext = pxNextHdr;
		if (pxPrevHdr)
		{
			pxPrevHdr->m_pxNext = pxNewHdr;
		}
		if (pxNextHdr)
		{
			pxNextHdr->m_pxPrev = pxNewHdr;
		}
		if (pxNextHdr == g_pxPacketHead)
		{
			g_pxPacketHead = pxNewHdr;
		}
	}

	// set data in the header
	pxNewHdr->m_uDeliveryTime = uDeliveryTime;
	pxNewHdr->m_iLength = iLength;
	pxNewHdr->m_xAddress = xAddress;
	pxNewHdr->m_usPort = usPort;
	
	return true;
}

void Asura_InternetSimulation::Update( void* pInstance, Asura_InternetSimulation_Callback pfCallback )
{
	if (!s_bEnabled)
	{
		return;
	}

	if (!g_pxMemHeap) 
	{
		// not initialised
		return;
	}

	g_bInUpdate = true;

	// keep popping from the head of the list until delivery time of head message is later than current time.
	u_int uCurrentTime = Asura_Timers::GetMSecsSinceStart_Continuous();

	Asura_NetSim_PacketHeader* pxCurHdr = g_pxPacketHead;

	while (pxCurHdr)
	{
		if (pxCurHdr->m_uDeliveryTime > uCurrentTime)
		{
			break;
		}

		char* pcData = reinterpret_cast<char*>(pxCurHdr + 1);
		pfCallback( pInstance, pcData, pxCurHdr->m_iLength, pxCurHdr->m_xAddress, pxCurHdr->m_usPort );

		// get the next header
		Asura_NetSim_PacketHeader* pxNextHdr = pxCurHdr->m_pxNext;

		// free the memory
		g_pxMemHeap->Free( pxCurHdr );

		// move on to next header
		pxCurHdr = pxNextHdr;
	}

	if (pxCurHdr)
	{
		pxCurHdr->m_pxPrev = NULL;
	}
	g_pxPacketHead = pxCurHdr;

	g_bInUpdate = false;
}

u_int Asura_InternetSimulation::GetAddedLatency()
{
	// based this on the source code for NetTool (see Kirmse's article in Game Programming Gems 3)
	/* See Knuth TAOCP Section 3.4.1 Algorithm P */
	float s;
	float x1;
	float v1;
	while (true)
	{
		v1 = Asura_Random::GetFloat(-1.0f, 1.0f);
		float v2 = Asura_Random::GetFloat(-1.0f, 1.0f);
		s = v1 * v1 + v2 * v2;
		if (s <= 1.0f)
		{
			break;
		}
	}

	x1 = v1 * Asura_Maths::Sqrt(-2.0f * logf(s) / s);

	int iLatency = Asura_Maths::FloatToInt(static_cast<float>(s_xCurrentSettings.m_uLatencyMean) + 
		(x1 * static_cast<float>(s_xCurrentSettings.m_uLatencyStddev)));
	if (iLatency < 0)
	{
	   iLatency = 0;
	}
	return static_cast<u_int>(iLatency);
}

bool Asura_InternetSimulation::ShouldDrop()
{
	return (Asura_Random::GetFloat() * 99.9999f) < s_xCurrentSettings.m_fDropPercent;
}

#endif // #ifdef ASURA_ENABLE_INTERNET_SIMULATION
