//////////////////////////////////////////////////////////////////////////////
//
// Asura_LobbyChat.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     helper class for chat SDKs which buffers up messages and provides
//     functionality for the frontend to get the messages for rendering of
//     chat windows
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "Network/Asura_LobbyChat.h"

#ifdef ASURA_SUPPORT_MULTIPLAYER_COMMS

// asura includes
#include "Core/Asura_String.h"
#include "Interface/Asura_CommandConsole.h"
#include "Multiplayer/Asura_Multiplayer_Comms_System.h"

// stdlib includes
#include <wchar.h>

// C O N S T A N T S /////////////////////////////////////////////////////////
const u_int uASURA_CHAT_HISTORY_MAX_MESSAGES = 16;
const u_int uASURA_CHAT_MAX_USERS = 64;


// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_LobbyChat::s_bInitialised = false;
u_int Asura_LobbyChat::s_uNextHistoryEntry = 0;
bool Asura_LobbyChat::s_bHistoryHasWrapped = false;
Asura_UnicodeChar** Asura_LobbyChat::s_awszChatHistory;
u_int Asura_LobbyChat::s_uNumUsers = 0;
Asura_UnicodeChar** Asura_LobbyChat::s_awszChatUserNames;
bool* Asura_LobbyChat::s_abChatUserMuted;
int* Asura_LobbyChat::s_aiChatUserID;
bool (*Asura_LobbyChat::s_pfnAddUser)(const Asura_UnicodeChar*, bool*, int*);

	
void Asura_LobbyChat::Initialise(bool (*pfnAddUser)(const Asura_UnicodeChar*, bool*, int*)/* = 0*/)
{
	Asura_Assert(!s_bInitialised, "already initialised");
	if (!s_bInitialised)
	{
		if (pfnAddUser)
		{
			s_pfnAddUser = pfnAddUser;
		}
		// allocate buffers used by this class
		s_uNextHistoryEntry = 0;
		s_bHistoryHasWrapped = false;
		s_awszChatHistory = new Asura_UnicodeChar*[uASURA_CHAT_HISTORY_MAX_MESSAGES];
		for (u_int u = 0; u < uASURA_CHAT_HISTORY_MAX_MESSAGES; ++u)
		{
			s_awszChatHistory[u] = new Asura_UnicodeChar[uASURA_CHAT_MAX_MESSAGE_SIZE];
			*s_awszChatHistory[u] = L'\0';
		}
		
		
		s_uNumUsers = 0;
		s_awszChatUserNames = new Asura_UnicodeChar*[uASURA_CHAT_MAX_USERS];
		for (u_int u = 0; u < uASURA_CHAT_MAX_USERS; ++u)
		{
			s_awszChatUserNames[u] = new Asura_UnicodeChar[uASURA_CHAT_MAX_USERNAME_LEN + 1];
			*s_awszChatUserNames[u] = L'\0';
		}
		s_abChatUserMuted = new bool[uASURA_CHAT_MAX_USERS];
		for (u_int u = 0; u < uASURA_CHAT_MAX_USERS; ++u)
		{
			s_abChatUserMuted[u] = false;
		}
		s_aiChatUserID = new int[uASURA_CHAT_MAX_USERS];
		for (u_int u = 0; u < uASURA_CHAT_MAX_USERS; ++u)
		{
			s_aiChatUserID[u] = 0;
		}

		s_bInitialised = true;
	}
}


void Asura_LobbyChat::Shutdown( void )
{
	Asura_Assert(s_bInitialised, "chat not intialised");
	if (s_bInitialised)
	{
		if (s_awszChatHistory)
		{
			for (u_int u = 0; u < uASURA_CHAT_HISTORY_MAX_MESSAGES; ++u)
			{
				delete[] s_awszChatHistory[u];
			}
			delete[] s_awszChatHistory;
			s_awszChatHistory = NULL;
		}

		if (s_awszChatUserNames)
		{
			for (u_int u = 0; u < uASURA_CHAT_MAX_USERS; ++u)
			{
				delete[] s_awszChatUserNames[u];
			}
			delete[] s_awszChatUserNames;
			s_awszChatUserNames = NULL;
		}

		if (s_abChatUserMuted)
		{
			delete[] s_abChatUserMuted;
			s_abChatUserMuted = NULL;
		}

		if (s_aiChatUserID)
		{
			delete[] s_aiChatUserID;
			s_aiChatUserID = NULL;
		}


		s_bInitialised = false;

		s_uNextHistoryEntry = 0;
		s_bHistoryHasWrapped = false;
		s_uNumUsers = 0;
	}
}


void Asura_LobbyChat::Reset( void )
{
	s_uNextHistoryEntry = 0;
	s_bHistoryHasWrapped = false;
	s_uNumUsers = 0;
}


bool Asura_LobbyChat::AddUser(const Asura_UnicodeChar* wszUser)
{
	Asura_Assert(s_bInitialised, "chat not intialised");
	if (s_bInitialised)
	{
		bool bExists = false;
		for (u_int u = 0; u < s_uNumUsers; ++u)
		{
			if (!wcscmp(s_awszChatUserNames[u], wszUser))
			{
				Asura_Assert(false, "user already exists - will not add");
				bExists = true;
				break;
			}
		}

		if (!bExists)
		{
			if (s_uNumUsers < uASURA_CHAT_MAX_USERS)
			{
				bool bMuted = false;
				int iUserId = 0;
				if (!s_pfnAddUser || s_pfnAddUser(wszUser, &bMuted, &iUserId))
				{
					s_abChatUserMuted[s_uNumUsers] = bMuted;
					s_aiChatUserID[s_uNumUsers] = iUserId;
					Asura_PrintfToBfr(s_awszChatUserNames[s_uNumUsers++], L"%s", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, wszUser);
					return true;
				}
			}
		}
	}
	return false;
}


bool Asura_LobbyChat::RemoveUser(const Asura_UnicodeChar* wszUser)
{
	Asura_Assert(s_bInitialised, "chat not intialised");
	if (s_bInitialised)
	{
		for (u_int u = 0; u < s_uNumUsers; ++u)
		{
			if (!wcscmp(s_awszChatUserNames[u], wszUser))
			{
				// copy last entry over this one (unless it's already the last one)
				u_int uLastEntry = --s_uNumUsers;
				if (uLastEntry != u)
				{
					wcscpy(s_awszChatUserNames[u], s_awszChatUserNames[uLastEntry]);
					s_abChatUserMuted[u] = s_abChatUserMuted[uLastEntry];
					s_aiChatUserID[u] = s_aiChatUserID[uLastEntry];
				}
				return true;
			}
		}
	}
	return false;
}


void Asura_LobbyChat::AddMessage(const Asura_UnicodeChar* wszUser, const Asura_UnicodeChar* wszMessage, bool bPrivate /*= false*/)
{
	Asura_Assert(s_bInitialised, "chat not intialised");
	if (s_bInitialised)
	{
		// create a copy which is then checked for bad language
		Asura_UnicodeChar wszFilteredMessage[uASURA_CHAT_MAX_MESSAGE_SIZE];

		Asura_PrintfToBfr(wszFilteredMessage, L"%s", uASURA_CHAT_MAX_MESSAGE_SIZE, true, wszMessage);
		Asura_Multiplayer_Comms_System::FilterProfanities(wszFilteredMessage);

		if (wszUser[0])
		{
			u_int uUser = s_uNumUsers;
			for (u_int u = 0; u < s_uNumUsers; ++u)
			{
				if (!wcscmp(s_awszChatUserNames[u], wszUser))
				{
					uUser = u;
					break;
				}
			}

			if (uUser == s_uNumUsers)
			{
				//user not found - assume this is a corrupt or bogus message
				Asura_Assert(false, "chat message from unlisted user");
				return;
			}

			if (IsUserMuted(uUser))
			{
				//user is muted
				return;
			}
		}

		u_int uHistoryEntry = s_uNextHistoryEntry++;
		if (s_uNextHistoryEntry >= uASURA_CHAT_HISTORY_MAX_MESSAGES)
		{
			s_uNextHistoryEntry = 0;
			s_bHistoryHasWrapped = true;
		}

		// add to the buffer
		// if private (sent only to us) add a + symbol before the sender's name.
		// This can either be displayed as is, or filtered out by the renderer which
		// could e.g. use a different colour, or insert localised text.
		Asura_PrintfToBfr(s_awszChatHistory[uHistoryEntry], L"%s%s: %s", uASURA_CHAT_MAX_MESSAGE_SIZE, true, bPrivate ? L"+" : L"", wszUser, wszFilteredMessage);
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
		// echo to console
		Asura_CommandConsole::Printf( L"%s", s_awszChatHistory[uHistoryEntry] );
#endif

	}
}



u_int Asura_LobbyChat::GetNumMessages( void )
{
	if (s_bHistoryHasWrapped)
	{
		return uASURA_CHAT_HISTORY_MAX_MESSAGES;
	}
	else
	{
		return s_uNextHistoryEntry;
	}
}

const Asura_UnicodeChar* Asura_LobbyChat::GetMessage(u_int uMessageNumber)
{
	Asura_Assert(s_bInitialised, "chat not intialised");
	if (s_bInitialised)
	{
		Asura_Assert(uMessageNumber < GetNumMessages(), "invalid message number");

		int iHistoryIndex = s_uNextHistoryEntry - (1 + uMessageNumber);
		if (iHistoryIndex < 0)
		{
			Asura_Assert(s_bHistoryHasWrapped, "");
			iHistoryIndex += uASURA_CHAT_HISTORY_MAX_MESSAGES;
		}
		Asura_Assert(iHistoryIndex >= 0 && iHistoryIndex < (int)uASURA_CHAT_HISTORY_MAX_MESSAGES, "invalid history index");
		
		return s_awszChatHistory[iHistoryIndex];
	}
	return L"";
}

u_int Asura_LobbyChat::GetNumUsers( void )
{
	return s_uNumUsers;
}

const Asura_UnicodeChar* Asura_LobbyChat::GetUserName( u_int uUserNumber )
{
	if (uUserNumber < s_uNumUsers)
	{
		return s_awszChatUserNames[uUserNumber];
	}
	else
	{
		Asura_Assert(false, "invalid user number");
		return L"????";
	}
}
void Asura_LobbyChat::MuteUser( u_int uUserNumber )
{
	if (uUserNumber < s_uNumUsers)
	{
		s_abChatUserMuted[uUserNumber] = true;
	}
	else
	{
		Asura_Assert(false, "invalid user number");
	}
}

void Asura_LobbyChat::UnmuteUser( u_int uUserNumber )
{
	if (uUserNumber < s_uNumUsers)
	{
		s_abChatUserMuted[uUserNumber] = false;
	}
	else
	{
		Asura_Assert(false, "invalid user number");
	}
}

const bool Asura_LobbyChat::IsUserMuted( u_int uUserNumber )
{
	if (uUserNumber < s_uNumUsers)
	{
		return s_abChatUserMuted[uUserNumber];
	}
	else
	{
		Asura_Assert(false, "invalid user number");
		return false;
	}
}

const int Asura_LobbyChat::GetUserID( u_int uUserNumber )
{
	if (uUserNumber < s_uNumUsers)
	{
		return s_aiChatUserID[uUserNumber];
	}
	else
	{
		Asura_Assert(false, "invalid user number");
		return 0;
	}
}

void Asura_LobbyChat::SetUserID( u_int uUserNumber, int iChatUserID )
{
	if (uUserNumber < s_uNumUsers)
	{
		s_aiChatUserID[uUserNumber] = iChatUserID;
	}
	else
	{
		return;
	}
}

// temporary overloaded versions to simplify integration with non-unicode GameSpy SDKs
bool Asura_LobbyChat::AddUser(const char* szUser)
{
	Asura_UnicodeChar awcUser[uASURA_CHAT_MAX_USERNAME_LEN + 1];
	Asura_PrintfToBfr(awcUser, L"%S", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, szUser);
	return AddUser(awcUser);
}

bool Asura_LobbyChat::RemoveUser(const char* szUser)
{
	Asura_UnicodeChar awcUser[uASURA_CHAT_MAX_USERNAME_LEN + 1];
	Asura_PrintfToBfr(awcUser, L"%S", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, szUser);
	return RemoveUser(awcUser);
}

void Asura_LobbyChat::AddMessage(const char* szUser, const char* szMessage, bool bPrivate /*= false*/)
{
	Asura_UnicodeChar awcUser[uASURA_CHAT_MAX_USERNAME_LEN + 1];
	Asura_UnicodeChar awcMessage[uASURA_CHAT_MAX_MESSAGE_SIZE];
	Asura_PrintfToBfr(awcUser, L"%S", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, szUser);
	Asura_PrintfToBfr(awcMessage, L"%S", uASURA_CHAT_MAX_MESSAGE_SIZE, true, szMessage);
	AddMessage(awcUser, awcMessage, bPrivate);
}

#endif //ASURA_SUPPORT_MULTIPLAYER_COMMS
