//////////////////////////////////////////////////////////////////////////////
//
// Asura_MessageHeader.h
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     message data used when sending messages over external network
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_MESSAGEHEADER_H___
#define ___ASURA_MESSAGEHEADER_H___ 1

#include "Network/Asura_Guid.h"
#include "Network/Asura_Message.h"

// this class is used for sending messages over the network. It used to be 
// in Asura_DirectPlay.cpp, but it is generally useful.

const u_int uASURA_MESSAGE_HEADER_DATA_SIZE_MASK = 0x7fffffff;
const u_int uASURA_MESSAGE_HEADER_GUARANTEED_BIT = 0x80000000;

class Asura_MessageHeader
{
	public:

		// accessor functions
		u_short GetMsgID() const { return m_usMsgID; }
		u_short GetEntityClassification() const { return m_usEntityClassification; }
		Asura_Guid GetToGuid() const { return m_uToGuid; }
		Asura_Guid GetFromGuid() const { return m_uFromGuid; }
		float GetSentTime() const { return m_fSentTime; }
		u_int GetDataSize() const { return (m_uDataSizeAndGuaranteedBit & uASURA_MESSAGE_HEADER_DATA_SIZE_MASK); }
		bool IsGuaranteed() const { return (m_uDataSizeAndGuaranteedBit & uASURA_MESSAGE_HEADER_GUARANTEED_BIT) != 0; }

		void SetDataSize(u_int uDataSize)
		{
			Asura_Assertf(uDataSize <= uASURA_MESSAGE_HEADER_DATA_SIZE_MASK, ("That's a rather large message... %u bytes", uDataSize));
			m_uDataSizeAndGuaranteedBit = (m_uDataSizeAndGuaranteedBit & uASURA_MESSAGE_HEADER_GUARANTEED_BIT) | uDataSize;
		}
		void SetGuaranteed(bool bGuaranteed)
		{
			m_uDataSizeAndGuaranteedBit = (m_uDataSizeAndGuaranteedBit & uASURA_MESSAGE_HEADER_DATA_SIZE_MASK) | (bGuaranteed ? uASURA_MESSAGE_HEADER_GUARANTEED_BIT : 0);
		}

		void ChangeByteOrder();

		// legacy public access to data members
		unsigned short  m_usMsgID;
        unsigned short  m_usEntityClassification;
		Asura_Guid		m_uToGuid;
		Asura_Guid		m_uFromGuid;
		float			m_fSentTime;

		// operator= to convert from Asura_Message
		inline const Asura_MessageHeader& operator = ( const Asura_Message& m ) 
		{
		    m_usMsgID = m.m_usMsgID;
            m_usEntityClassification = m.m_usEntityClassification;
			m_uToGuid = m.To;
			m_uFromGuid = m.From;
			m_fSentTime = m.SentTime;
			SetDataSize(m.m_uDataSize);
			SetGuaranteed(m.IsUsingGuaranteedDelivery());
			return *this;
		}

		bool operator== (const Asura_MessageHeader& xRHS) const 
		{
			return ( (xRHS.m_usMsgID == m_usMsgID) && 
					(xRHS.m_usEntityClassification == m_usEntityClassification) && 
					(xRHS.m_uToGuid == m_uToGuid) && 
					(xRHS.m_uFromGuid == m_uFromGuid) && 
					(xRHS.m_fSentTime == m_fSentTime) && 
					(xRHS.GetDataSize() == GetDataSize()) && 
					(xRHS.IsGuaranteed() == IsGuaranteed()));
		}
	private:
		u_int m_uDataSizeAndGuaranteedBit;
};

class Asura_MessagePacking
{
	public:

		// Create() function takes buffer ptr and size, and asura message and 
		// creates Asura_CompressedMessageHeaderDesc, header fields and copies
		// the message data (if any) into the buffer as well.
		// It returns the number of bytes used. This is zero if there is not enough room.
		static u_int PackMessageToBuffer( u_char* pucBuffer, u_int uBufferSize, const Asura_Message& xMsg );

		// UncompressToMessageHeader takes the compressed buffer and fills out the Asura_MessageHeader.
		// It returns a ptr to the first byte after the packed header, which is the message data block
		// (if any) or it could be the start of the next packed header. If the data itself was packed, 
		// uSizeOfPackedData will contain the size of the packed data packet (0 if the data is unpacked)
		// while xMsgHeader.GetDataSize() will always be the unpacked data size.
		static const void* UnpackBufferToMessageHeader( const void* pData, bool bGuaranteed, 
														Asura_MessageHeader& xMsgHeader, u_int& uSizeOfPackedData );

		// GetBytesRequiredForMessage returns the number of bytes required to pack the message.
		static u_int GetBytesRequiredForMessage( const Asura_MessageHeader& xMsgHeader );
		static u_int GetBytesRequiredForMessage( u_int uEntityClassification, 
			u_int uToGuid, u_int uFromGuid, float fSentTime, u_int uDataSize);

		// functions to pack msgID, entity classification and guid in/out of a BitPacker.
		// Num bits used is currently 8 for msgID, 7 for entity class, and 3-24 for guid depending on range.
		static void PackMsgID( u_short usMsgID, class Asura_BitPacker* pxPacker );
		static u_short UnpackMsgID( class Asura_BitUnpacker* pxUnpacker );
		static u_int GetBitsRequiredForMsgID( u_short usMsgID );
		static void PackEntityClassification( u_short usEntityClassification, class Asura_BitPacker* pxPacker );
		static u_short UnpackEntityClassification( class Asura_BitUnpacker* pxUnpacker );
		static u_int GetBitsRequiredForEntityClassification( u_short usEntityClassification );
		static void PackGuid( u_int uGuid, class Asura_BitPacker* pxPacker );
		static u_int UnpackGuid( class Asura_BitUnpacker* pxUnpacker );
		static u_int GetBitsRequiredForGuid( Asura_Guid uGuid );

	private:

		static u_int GetGuidRange(Asura_Guid uGuid);
};


// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_MESSAGEHEADER_H___
