//////////////////////////////////////////////////////////////////////////////
//
// Asura_NetLinkManager.h
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     class to handle connection, disconnection and link status of internet
//     and LAN connections (including Wifi infrastructure and adhoc)
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_NETLINKMANAGER_H___
#define ASURA_NETLINKMANAGER_H___ 1


// E N U M S /////////////////////////////////////////////////////////////////
enum ASURA_NETLINK_TYPE
{
	ASURA_NETLINK_INTERNET	= 0,		// e.g. WiFi infrastructure or PC internet connection
	ASURA_NETLINK_LAN		= 1,		// e.g. WiFi adhoc
	ASURA_NETLINK_NUM_TYPES
};

// ASURA_NETLINK_STATUS may be referred to by GUIMenu condition vars so don't 
// change any existing values. Update the var help text if you add any.
// Note that a link may still exist even though status is reported as 
// disconnected, e.g. on PC your internet connection is always on, but as far
// as that game is concerned it can't use it until Connect() has been called 
// and the status is reported as Connected. This allows for a consistent 
// interface across platforms, and for extra things to be done like initialising
// GameSpy.
enum ASURA_NETLINK_STATUS
{
	ASURA_NETLINK_DISCONNECTED								= 0,
	ASURA_NETLINK_CONNECTING								= 1,
	ASURA_NETLINK_CONNECTED									= 2,
	ASURA_NETLINK_CONNECTION_FAILED							= 3,
	ASURA_NETLINK_DISCONNECTING								= 4,
	ASURA_NETLINK_CONNECTION_DROPPING						= 5,	// transition to DROPPED.
	ASURA_NETLINK_CONNECTION_DROPPED						= 6,	// e.g. WLAN switch off, router down
#ifdef ASURA_SUPPORT_GAMESPY
	ASURA_NETLINK_CHECKING_GAMESPY							= 7,
	ASURA_NETLINK_FAILED_GAMESPY_UNAVAILABLE				= 8,
	ASURA_NETLINK_FAILED_GAMESPY_TEMPORARILY_UNAVAILABLE	= 9,
	ASURA_NETLINK_FAILED_GAMESPY_DNS_FAILED					= 10,
#endif
	ASURA_NETLINK_NUM_STATUS_ENTRIES
};
// ASURA_NETLINK_STATUS may be referred to by GUIMenu condition vars so don't 
// change any existing values.

// C L A S S E S /////////////////////////////////////////////////////////////
class Asura_NetLinkManager
{
public:
	static void Initialise();
	static void Update();
	static void Update_PostRender();	// call this after EndFrame(). For PSP Utility library.
	static void Shutdown();

	// help params are for the console. iType is cast to ASURA_NETLINK_TYPE
	static bool Connect( int iType, bool bHelp = false, bool bInterrogate = false );
	static bool Disconnect( bool bHelp = false );
	static bool ResetDroppedStatus( bool bHelp = false );
	static ASURA_NETLINK_STATUS GetStatus() { return static_cast<ASURA_NETLINK_STATUS>(s_uStatus); }
	static ASURA_NETLINK_TYPE GetType() { return static_cast<ASURA_NETLINK_TYPE>(s_uType); }
	static bool IsConnected();
	static bool IsDisconnected();

	static bool IsWLANSwitchedOff() { return s_bWLANSwitchedOff; }		// set by PSP code, other platforms leave it off.

	static bool DropConnection();	// accessor to allow external code to force network disconnect for an error condition.

	// Get approx bandwidth in bits per second.
	// Returns zero if the bandwidth hasn't been measured.
	static u_int GetApproxUploadBandwidth()   { return s_uUploadBitsPerSec; }
	static u_int GetApproxDownloadBandwidth() { return s_uDownloadBitsPerSec; }

	static void SetApproxUploadBandwidth(u_int uUploadBitsPerSec)		{ s_uUploadBitsPerSec   = uUploadBitsPerSec; }
	static void SetApproxDownloadBandwidth(u_int uDownloadBitsPerSec)	{ s_uDownloadBitsPerSec = uDownloadBitsPerSec; }

private:

	// Platform_UpdateXXX() functions should call FinishedPlatformXXX() when finished.
	static void Platform_UpdateConnecting();
	static void Platform_UpdateConnecting_PostRender();	// PSP needs to do this after EndFrame()
	static void Platform_UpdateDisconnecting();

	static void Platform_UpdateConnected(); // could call ConnectionDropped() if link goes down.
	
	// Platform_InitiateConnection() function should return true iff requested operation was started.
	static bool Platform_InitiateConnection();
	static void Platform_InitiateDisconnection();

	static void FinishedPlatformConnecting(bool bSuccess);
	static void FinishedPlatformDisconnecting();
	static void ConnectionDropped();

	// data
	static u_int s_uStatus;
	static u_int s_uType;
	static bool s_bWLANSwitchedOff;
	static int s_iStateProgress;

	// Approx bandwidth data
	static u_int s_uUploadBitsPerSec;
	static u_int s_uDownloadBitsPerSec;
};



// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ASURA_NETLINKMANAGER_H___
