//////////////////////////////////////////////////////////////////////////////
//
// Asura_Network_Session.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     LAN session management - for clients.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Asura_Network_Session.h"

// asura includes
#include "Debugging/Asura_Assert.h"

#ifdef ASURA_SUPPORT_GAMESPY
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Network/GameSpy/Asura_GameSpy_ServerBrowsing.h"
#include "Network/Asura_NetTransport_System.h"
#endif

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK


// G L O B A L S /////////////////////////////////////////////////////////////
#ifdef ASURA_SUPPORT_GAMESPY
static u_int g_uTimeOfNextGSRefresh;
DEBUGVAR u_int dbg_uTimeBetweenGSRefreshes = 30000; // msecs
#endif

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Network_Session* Asura_Network_Session_System::s_pxSessions = 0;
u_int Asura_Network_Session_System::s_uNumberOfSessions = 0;
bool Asura_Network_Session_System::s_bLan = true;



// C L A S S   F U N C T I O N S /////////////////////////////////////////////
void Asura_Network_Session_System::Initialise( bool bLan /* = true */ )
{
	Asura_Assert(!s_pxSessions, "already initialised: memory leak");
	s_pxSessions = new Asura_Network_Session[uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS];
	s_uNumberOfSessions = 0;
	s_bLan = bLan;

	if (bLan)
	{
		Platform_Initialise();
	}
#ifdef ASURA_SUPPORT_GAMESPY
	else
	{
		Asura_GameSpy_ServerBrowsingSDK::Initialise( bLan, 10, uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS );
	}
#endif
}


void Asura_Network_Session_System::Update( void )
{
	Asura_Assert(s_pxSessions, "not initialised");

	if (s_bLan)
	{
		Platform_Update();
	}
#ifdef ASURA_SUPPORT_GAMESPY
	else
	{
		Asura_GameSpy_ServerBrowsingSDK::Update();

		const u_int uCurTime = Asura_Timers::GetMSecsSinceStart();
		if (g_uTimeOfNextGSRefresh < uCurTime)
		{
			g_uTimeOfNextGSRefresh = uCurTime + dbg_uTimeBetweenGSRefreshes;
			Asura_GameSpy_ServerBrowsingSDK::Refresh();
		}

		{
			u_int uProjectSessionDataSize = Project_GetSessionDataSize();
            void* pvProjectSessionData = new char[uProjectSessionDataSize];
			
			s_uNumberOfSessions = Asura_GameSpy_ServerBrowsing::GetNumServers();
			Asura_Assert(s_uNumberOfSessions < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS, "too many servers");
			s_uNumberOfSessions = Asura_Maths::Min(s_uNumberOfSessions, uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS);
			for (u_int u = 0; u < s_uNumberOfSessions; ++u)
			{
				Asura_UnicodeChar awcServerName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
				Asura_GameSpy_ServerBrowsing::GetServerName(u, awcServerName, uASURA_NETWORK_MAX_SESSION_NAME + 1);
				u_int uServerSessionID = 0; // hack.
				Project_FillSessionDataFromGameSpyServer( u, pvProjectSessionData );

				s_pxSessions[u].Set(
					Asura_NetAddress(Asura_GameSpy_ServerBrowsing::GetServerInetAddress(u)),
					Asura_GameSpy_ServerBrowsing::GetServerHostPort(u),
					Asura_GameSpy_ServerBrowsing::GetNumPlayers(u),
					Asura_GameSpy_ServerBrowsing::GetMaxPlayers(u),
					awcServerName,
					Asura_GameSpy_ServerBrowsing::GetPing(u),
					uServerSessionID,
					&u,			// platform data - store server num
					sizeof(u),	// platform data - store server num
					pvProjectSessionData,
					uProjectSessionDataSize 
				);

			}

			delete[] static_cast<char*>(pvProjectSessionData);
		}
	}
#endif
}


void Asura_Network_Session_System::Shutdown( void )
{
	if (s_bLan)
	{
		Platform_Shutdown();
	}
#ifdef ASURA_SUPPORT_GAMESPY
	else
	{
		Asura_GameSpy_ServerBrowsingSDK::Shutdown();
	}
#endif

	delete[] s_pxSessions;
	s_pxSessions = NULL;
	s_uNumberOfSessions = 0;
}
	

void Asura_Network_Session_System::ForceRefresh( void )
{
#ifdef ASURA_SUPPORT_GAMESPY
	g_uTimeOfNextGSRefresh = 0;
#endif
}


const Asura_Network_Session* Asura_Network_Session_System::GetSession(u_int uSessionIndex)
{
    Asura_Assert(s_pxSessions, "session system not initialised");
	Asura_Assert(uSessionIndex < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS, "invalid session ID");
	Asura_Assert(s_pxSessions[uSessionIndex].IsValid(), "invalid session requested");
	
	return &s_pxSessions[uSessionIndex];
}

void Asura_Network_Session_System::SetSession(	u_int uSessionIndex,
												Asura_NetAddress xAddress,
												u_short usPort,
												u_int uCurPlayers,
												u_int uMaxPlayers,
												const Asura_UnicodeChar* wszSessionName,
												u_int uRoundTripLatency,
												u_int uServerSessionID,
												const void* pvPlatformSessionData,
												u_int uPlatformSessionDataSize, 
												const void* pvProjectSessionData,
												u_int uProjectSessionDataSize )
{
    Asura_Assert(s_pxSessions, "session system not initialised");
	Asura_Assert(uSessionIndex < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS, "invalid session ID");
	
	s_pxSessions[uSessionIndex].Set(xAddress,
									usPort,
									uCurPlayers,
									uMaxPlayers,
									wszSessionName,
									uRoundTripLatency,
									uServerSessionID,
									pvPlatformSessionData,
									uPlatformSessionDataSize,
									pvProjectSessionData,
									uProjectSessionDataSize);
}

void Asura_Network_Session_System::ClearSession( u_int uSessionIndex )
{
    Asura_Assert(s_pxSessions, "session system not initialised");
	Asura_Assert(uSessionIndex < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS, "invalid session ID");
	
	s_pxSessions[uSessionIndex].Reset();
}


void Asura_Network_Session::PrepareConnection() const
{
	//AS 20/12/2007 Call the platform specific function
	// On XBOX360 this handles setting up the secure connection
	Platform_PrepareConnection();

#ifdef ASURA_SUPPORT_GAMESPY
	u_int uServer = *reinterpret_cast<u_int*>(m_pPlatformSessionData);
	if (Asura_GameSpy_ServerBrowsing::NeedToUseNatNeg(uServer))
	{
		int iCookie = Asura_Random::GetSeededInt();
		Asura_NetTransport_System::SetUseNatNeg( true, iCookie );
		Asura_GameSpy_ServerBrowsing::SendNatNegCookie(uServer, iCookie);
	}
	else
	{
		Asura_NetTransport_System::SetUseNatNeg(false, 0);
	}
#endif
}

#endif //#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
