/**
*
* \file Asura_Network_StatLogging.cpp
*
* Created by Richard May - 25/11/2003 15:15:52
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/



/************/
/* Includes.*/
/************/

// main asura header
#include "Core/Asura.h"

// local header
#include "Asura_Network_StatLogging.h"

// asura
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"

#ifdef ASURA_USE_NETWORK_STATLOGGING

/*********/
/* Data. */
/*********/

float Asura_Network_StatLogging::s_fSnapshotInterval = 0.0f;
float Asura_Network_StatLogging::s_fIntervalTimer = s_fSnapshotInterval;
bool  Asura_Network_StatLogging::s_bSessionOpen = false;
u_int Asura_Network_StatLogging::s_uSnapshotModeFlags = ASURA_STATLOG_SNAPSHOTFLAG_ONEND;
class Asura_StatLogging_Stat Asura_Network_StatLogging::s_axVariableData[ASURA_STATLOG_MAXSTATS];
Asura_HashTree Asura_Network_StatLogging::s_xVariableTree;


/**************/
/* Functions. */
/**************/

/*** Main Functions ***/

/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::Initialise
 *
 * Author        : Richard May
 *
 * Description   : Main initialisation function. Calls platform
 *				   specific intialisers.
 *
 *************************************************************/
void Asura_Network_StatLogging::Initialise ( void )
{
	s_bSessionOpen = false;
	s_fSnapshotInterval = 0.0f;
	s_fIntervalTimer = s_fSnapshotInterval;
	SetSnapshotMode(ASURA_STATLOG_SNAPSHOTFLAG_ONEND);

	ClearAllData(); // clear hash tree and reset variable data

	Platform_Initialise();
}
/* end of function Initialise */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::Shutdown
 *
 * Author        : Richard May
 *
 * Description   : Main shutdown function. Calls platform specific
 *				   shutdown functions
 *
 *************************************************************/
void Asura_Network_StatLogging::Shutdown ( void )
{
	if( IsSessionInProgress() )
	{
		EndSession();
	}

	Platform_Shutdown();
}
/* end of function Shutdown */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::Reset
 *
 * Author        : Richard May
 *
 * Description   : Main reset function
 *
 *************************************************************/
void Asura_Network_StatLogging::Reset ( void )
{
	Shutdown();
	Initialise();
}
/* end of function Reset */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::Update
 *
 * Author        : Richard May
 *
 * Description   : Main update function
 *
 *************************************************************/
void Asura_Network_StatLogging::Update ( void )
{
	if(!IsSessionInProgress()) return;

	// update periodic snapshot timer
	if( s_uSnapshotModeFlags & ASURA_STATLOG_SNAPSHOTFLAG_ONINTERVAL )
	{
		s_fIntervalTimer -= Asura_Timers::GetFrameTime();
		if( s_fIntervalTimer <= 0.0f )
		{
			s_fIntervalTimer = s_fSnapshotInterval;
			SendSnapshot();
		}
	}
}
/* end of function Update */



/*** Session Management ***/

/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::StartSession
 *
 * Author        : Richard May
 *
 * Description   : Starts a logging session. Returns true if 
 *				   successful.
 *
 *************************************************************/
bool Asura_Network_StatLogging::StartSession ( void )
{
	if( IsSessionInProgress() ) return false; // session already in progress

	s_xVariableTree.Reset();

	return Platform_StartSession();
}
/* end of function StartSession */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::EndSession
 *
 * Author        : Richard May
 *
 * Description   : Finishes a logging session. If the snapshot
 *				   mode is set to 'ONEND', it will trigger the
 *				   sending of the final snapshot data
 *
 *************************************************************/
void Asura_Network_StatLogging::EndSession ( void )
{
	//if( !IsSessionInProgress() ) return; // no session to end

	if( s_uSnapshotModeFlags & ASURA_STATLOG_SNAPSHOTFLAG_ONEND )
	{
		SendSnapshot(true);
	}
	
	Asura_Network_StatLogging::Platform_EndSession();
}
/* end of function EndSession */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::IsSessionInProgress
 *
 * Author        : Richard May
 *
 * Description   : Returns true if session is currently running
 *				   and logging is enabled
 *
 *************************************************************/
bool Asura_Network_StatLogging::IsSessionInProgress ( void )
{
	return s_bSessionOpen;
}
/* end of function IsSessionInProgress */



/*** Snapshot setup and control ***/

/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::SetSnapshotMode
 *
 * Author        : Richard May
 *
 * Description   : This version of the SetSnapshotMode function 
 *				   allows you to set the snapshot mode flags 
 *				   directly to whatever value you choose
 *
 *************************************************************/
void Asura_Network_StatLogging::SetSnapshotMode ( u_int uSnapshotFlags )
{
	s_uSnapshotModeFlags = uSnapshotFlags;
}
/* end of function SetSnapshotMode */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::SetSnapshotMode
 *
 * Author        : Richard May
 *
 * Description   : This version allows you to set the individual
 *				   components of the snapshot flags separately
 *
 *************************************************************/
void Asura_Network_StatLogging::SetSnapshotMode ( ASURA_STATLOG_SNAPSHOT_MODE eMode, bool bValue )
{
	if(bValue)
	{
		s_uSnapshotModeFlags |= (1<<eMode);  // set flag
	}
	else
	{
		s_uSnapshotModeFlags &= ~(1<<eMode); // unset flag
	}
	
}
/* end of function SetSnapshotMode */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::SetSnapshotInterval
 *
 * Author        : Richard May
 *
 * Description   : Allows you to set the time between which snapshots
 *				   are sent (in seconds). This will only have an 
 *				   effect if the SnapShot mode has been set to ONINTERVAL
 *
 *************************************************************/
void Asura_Network_StatLogging::SetSnapshotInterval ( float fTimeBetweenSnapshots )
{
	s_fSnapshotInterval = fTimeBetweenSnapshots;
	s_fIntervalTimer = s_fSnapshotInterval;
}
/* end of function SetSnapshotInterval */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::TriggerSnapshot
 *
 * Author        : Richard May
 *
 * Description   : Allows you to manually trigger a snapshot send
 *				   Returns true if snapshot send was successful
 *
 *************************************************************/
bool Asura_Network_StatLogging::TriggerSnapshot ( void )
{
	// check the snapshot mode allows manual triggering
	if( !(s_uSnapshotModeFlags & ASURA_STATLOG_SNAPSHOTFLAG_ONTRIGGER) ) return false;

	return SendSnapshot();
}
/* end of function TriggerSnapshot */



/*** Stat data management ***/

/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::ClearAllData
 *
 * Author        : Richard May
 *
 * Description   : Clears the stat storage 
 *
 *************************************************************/
void Asura_Network_StatLogging::ClearAllData ( void )
{
	s_xVariableTree.Reset();
	for( u_int u=0; u<ASURA_STATLOG_MAXSTATS; ++u )
	{
		s_axVariableData[u].m_eType = STAT_NONE;
	}
}
/* end of function ClearAllData */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::RegisterVariable
 *
 * Author        : Richard May
 *
 * Description   : This function and its various overrides allow
 *				   the user to register a variable with the stat
 *				   logging system. Whenever a snapshot send is triggered
 *				   the current value of the variable will be included
 *				   in the snapshot data
 *				 
 *				   To associate this variable with a particular 
 *				   player or team, supply a Guid (for players) or
 *				   a team number (for teams)  
 *
 *				   Note that the variable name should be no longer
 *				   than eight characters
 *
 *************************************************************/
bool Asura_Network_StatLogging::RegisterVariable ( const char* szName, const bool* pbVar, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );

		s_axVariableData[iNewIndex].m_eType   = STAT_PTR_BOOL; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_pbValue = pbVar;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
		return true;
	}

	return false;
}

bool Asura_Network_StatLogging::RegisterVariable ( const char* szName, const int* piVar, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_PTR_INT; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_piValue = piVar;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
		return true;
	}
	
	return false;
}

bool Asura_Network_StatLogging::RegisterVariable ( const char* szName, const u_int* puVar, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_PTR_UINT; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_puValue = puVar;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
		return true;
	}
	
	return false;
}

bool Asura_Network_StatLogging::RegisterVariable ( const char* szName, const float* pfVar, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_PTR_FLOAT; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_pfValue = pfVar;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
		return true;
	}
	
	return false;
}

bool Asura_Network_StatLogging::RegisterVariable ( const char* szName, const char* szVar, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_STRING; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_szValue = szVar;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
		return true;
	}
	
	return false;
}
/* end of function RegisterVariable */

/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::AddData
 *
 * Author        : Richard May
 *
 * Description   : This function and its various overrides allow
 *				   the user to register a variable with the stat
 *				   logging system. Whenever a snapshot send is triggered
 *				   the current value of the variable will be included
 *				   in the snapshot data
 *				 
 *				   To associate this variable with a particular 
 *				   player or team, supply a Guid (for players) or
 *				   a team number (for teams)  
 *
 *				   Note that the variable name should be no longer
 *				   than eight characters
 *
 *************************************************************/
void Asura_Network_StatLogging::AddData( const char* szName, bool  bValue, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_BOOL; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_bValue  = bValue;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
	}
}

void Asura_Network_StatLogging::AddData( const char* szName, float fValue, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_FLOAT; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_fValue  = fValue;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
	}
}

void Asura_Network_StatLogging::AddData( const char* szName, int   iValue, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_INT; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_iValue  = iValue;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
	}
}

void Asura_Network_StatLogging::AddData( const char* szName, u_int uValue, Asura_Guid uPlayer/*=ASURA_GUID_UNREGISTERED*/, int iTeam/*=ASURA_STATLOG_NOTEAM*/ )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, uPlayer, iTeam );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_UINT; 
		s_axVariableData[iNewIndex].m_uOwner  = uPlayer;
		s_axVariableData[iNewIndex].m_iTeamID = iTeam;
		s_axVariableData[iNewIndex].m_uValue  = uValue;
		s_axVariableData[iNewIndex].m_bIsGuid = true;
	}
}

/* end of function AddData */



/* end of function RegisterVariable */

// interger-based ID version
void Asura_Network_StatLogging::AddData( const char* szName, bool  bValue, int iPlayerID )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, iPlayerID, ASURA_STATLOG_NOTEAM );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_BOOL; 
		s_axVariableData[iNewIndex].m_uOwner  = (u_int)iPlayerID;		
		s_axVariableData[iNewIndex].m_bValue  = bValue;
		s_axVariableData[iNewIndex].m_bIsGuid = false;
	}
}

void Asura_Network_StatLogging::AddData( const char* szName, float fValue, int iPlayerID )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, iPlayerID, ASURA_STATLOG_NOTEAM );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_FLOAT; 
		s_axVariableData[iNewIndex].m_uOwner  = (u_int)iPlayerID;		
		s_axVariableData[iNewIndex].m_fValue  = fValue;
		s_axVariableData[iNewIndex].m_bIsGuid = false;
	}
}

void Asura_Network_StatLogging::AddData( const char* szName, int   iValue, int iPlayerID )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, iPlayerID, ASURA_STATLOG_NOTEAM );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_INT; 
		s_axVariableData[iNewIndex].m_uOwner  = (u_int)iPlayerID;		
		s_axVariableData[iNewIndex].m_iValue  = iValue;
		s_axVariableData[iNewIndex].m_bIsGuid = false;
	}
}

void Asura_Network_StatLogging::AddData( const char* szName, u_int uValue, int iPlayerID )
{
	char szTempKey[ASURA_STATLOG_MAXKEYSIZE];
	GenerateVariableKey( szTempKey, szName, iPlayerID, ASURA_STATLOG_NOTEAM );
	
	int iNewIndex = 0;
	if( !s_xVariableTree.FindNodeIndexByName( szTempKey, true, iNewIndex ) )
	{
		strcpy( s_axVariableData[iNewIndex].m_szName, szTempKey );
		
		s_axVariableData[iNewIndex].m_eType   = STAT_UINT; 
		s_axVariableData[iNewIndex].m_uOwner  = (u_int)iPlayerID;	
		s_axVariableData[iNewIndex].m_uValue  = uValue;
		s_axVariableData[iNewIndex].m_bIsGuid = false;
	}
}

/* end of function AddData */



/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::GenerateVariableKey
 *
 * Author        : Richard May
 *
 * Description   : Generates a unique name for this variable from
 *				   the given text name and Player/Team IDs
 *
 *************************************************************/
void Asura_Network_StatLogging::GenerateVariableKey( char* szNewKey, const char* szName, Asura_Guid uPlayer, int iTeam )
{
	if(!szNewKey) return;

	// setup header (null is default)
	szNewKey[0] = 0;
	if( uPlayer != ASURA_GUID_UNREGISTERED ) // players use 'p<GUID>_'
	{				
		char szID[16];
		Asura_PrintfToBfr( szID, "p%d_", 16, true, uPlayer );
		Asura_PrintfToBfr( szNewKey, "%s%s", ASURA_STATLOG_MAXKEYSIZE, true, szID, szName );

		int iKeyLength = strlen(szName) + strlen(szID);
		Asura_Assert( (iKeyLength < ASURA_STATLOG_MAXKEYSIZE), "Warning - variable key truncated!" );
	}
	else if( iTeam != ASURA_STATLOG_NOTEAM ) // teams use 't<ID>_'
	{
		char szID[16];
		Asura_PrintfToBfr( szID, "t%d_", 16, true, iTeam );
		Asura_PrintfToBfr( szNewKey, "%s%s", ASURA_STATLOG_MAXKEYSIZE, true, szID, szName );

		int iKeyLength = strlen(szName) + strlen(szID);
		Asura_Assert( (iKeyLength < ASURA_STATLOG_MAXKEYSIZE), "Warning - variable key truncated!" );
	}
	else
	{
		Asura_PrintfToBfr( szNewKey, "%s", ASURA_STATLOG_MAXKEYSIZE, true, szName );
		Asura_Assert( (strlen(szNewKey) < ASURA_STATLOG_MAXKEYSIZE), "Warning - variable key truncated!" );
	}
}
/* end of function GenerateVariableKey */


/*************************************************************
 *
 * Name          : Asura_Network_StatLogging::SendSnapshot
 *
 * Author        : Richard May
 *
 * Description   : Sends all the currently collected data to 
 *				   the logging server. Returns true if successful
 *				   bGameEnd should be true if the game has just finished
 *
 *************************************************************/
bool Asura_Network_StatLogging::SendSnapshot ( bool bGameEnd )
{
	return Platform_SendSnapshot( bGameEnd );
}
/* end of function SendSnapshot */

#endif // ASURA_USE_NETWORK_STATLOGGING

/****************/
/* End of file. */
/****************/
