///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_Qos_System.h
//
// Created by Martin G.Harris (9/4/2009)
// 
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
// Description: System for measuring the upload bandwidth of machines in the
//				lobby of online games. This information can then be used to
//              help determine which machine would make the best game server.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_QOS_SYSTEM_H__
#define ASURA_QOS_SYSTEM_H__ 1

#ifdef ASURA_ENABLE_QOS_SYSTEM

// Includes
#include "Core/Asura_Timers.h"
#include "Network/Asura_Platform_NetSocket.h"


//////////////////////////
// Forward declarations //
//////////////////////////
class Asura_Qos_Message;
class Asura_Qos_Listener;
class Asura_Qos_NetConnection;
class Asura_QosSocket;

/////////////
// Defines //
/////////////
// Enable logging to logfile and console/debugger.
// NOTE: Do *not* enable this for submission builds!
//#define ASURA_QOS_ENABLE_LOGGING 1

// Size of the receive buffer.
// The largest message that we will send is 1024 bytes, so this should be enough.
#define ASURA_QOS_RECEIVE_BUFFER_SIZE 2048

///////////////
// Constants //
///////////////
// Maximum number of connections that the server can maintain.
const u_int uASURA_QOS_MAX_NUM_CONNECTIONS = 32;

// Maximum number of tests to do per connection.
const u_int uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION = 16;

// Number of successful tests needed before finishing testing (in a given direction).
const u_int uASURA_QOS_NUM_TEST_RESULTS_NEEDED = 3;

// Number of large packets to send for the large packet test.
const u_int uASURA_QOS_NUM_LARGE_PACKETS = 2;

// Maximum transmission unit (MTU) is limited by ethernet to typically 1500.
// Xbox360 limits TCP data size to 1264 bytes.
// Since we are only using packets of max size 1024 bytes, that value will be enough.
const u_int uASURA_QOS_MAX_PACKET_LENGTH = 1024;

// Maximum number of data vars in the message.
const u_int uASURA_QOS_MAX_NUM_DATA_VARS = uASURA_QOS_MAX_PACKET_LENGTH/sizeof(u_int);

// Sleep time (in seconds) for the QoS thread, when nothing much is happening.
const float fASURA_QOS_THREAD_SLEEP_TIME = 0.025f; // 25 ms

///////////
// Enums //
///////////
// Message ids
enum ASURA_QOS_MSGID
{
	// Connection
	ASURA_QOS_MSGID_MSG_CONNECTION_REQUEST,		// client to server
	ASURA_QOS_MSGID_MSG_CONNECTION_ACCEPT,		// server to client
	ASURA_QOS_MSGID_MSG_CONNECTION_REJECT,		// server to client
	ASURA_QOS_MSGID_MSG_CONNECTION_CLOSE,		// server to client, or client to server.
	ASURA_QOS_MSGID_MSG_CONNECTION_KEEPALIVE,	// keep alive message

	// Testing
	ASURA_QOS_MSGID_START_TEST,					// instruction to start a test
	ASURA_QOS_MSGID_PACKET,						// test packet
	ASURA_QOS_MSGID_RECV_TIMING,				// timing results returned by receiver
	ASURA_QOS_MSGID_RESULT,						// bandwidth result returned to the receiver
};

// QoS connection states
enum ASURA_QOS_CONNECTION_STATE
{
	ASURA_QOS_STATE_NULL,						// no connection attempted
	ASURA_QOS_STATE_CONNECTED,					// connected
	ASURA_QOS_STATE_CONNECTED_PROBLEMS,			// connected (but no response for a while)
	ASURA_QOS_STATE_CLOSED,						// closed connection
};

// QoS test states
enum ASURA_QOS_TESTSTATE
{
	ASURA_QOS_TESTSTATE_NULL,					// no test in progress

	// Sending
	ASURA_QOS_TESTSTATE_START,					// start sending large packets
	ASURA_QOS_TESTSTATE_SENDING,				// sending large packets
	ASURA_QOS_TESTSTATE_POST_TEST,				// waiting for results

	// Receiving
	ASURA_QOS_TESTSTATE_WAITING,				// waiting for next test to start.
	ASURA_QOS_TESTSTATE_TESTING,				// test is in progress.
};


// Results of a QoS test.
enum ASURA_QOS_RESULT_FLAGS
{
	ASURA_QOS_RESULT_FLAG_REJECT					= 0x00000001,	// Rejected

	ASURA_QOS_RESULT_FLAG_OUTOFORDER				= 0x00000002,	// Large packets were out of order (but data is usable).
	ASURA_QOS_RESULT_FLAG_LOSTPACKETS				= 0x00000004,	// Lost some large packets (data may be usable).
	ASURA_QOS_RESULT_FLAG_LATE						= 0x00000008,	// Result was returned late (data may be usable).

	ASURA_QOS_RESULT_FLAG_ZEROTH_OUTOFORDER			= 0x00000010,	// Zeroth packet arrived after the large packets. Reject.
	ASURA_QOS_RESULT_FLAG_TIMEOUT					= 0x00000020,	// Had no reply from client in time. Reject.
	ASURA_QOS_RESULT_FLAG_TOOFAST					= 0x00000040,	// Received bandwidth greater than sent bandwidth. Reject.
	ASURA_QOS_RESULT_FLAG_QUEUING					= 0x00000080,	// Zeroth/first packet were queued up. Reject.

	ASURA_QOS_RESULT_FLAG_DOWNLOAD_FROM_REMOTE		= 0x00000100,	// This result is for a download from the remote computer,
																	// (otherwise it is an upload to the remote one).
};


//////////////////////
// Asura_Qos_System //
//////////////////////
class Asura_Qos_System
{
public:
	static void Initialise();
	static void Shutdown();

	// Called to indicate state changes in the lobby system,.
	static void CreateLobby(bool bIsHost);
	static void LobbyTerminated();
	static void StartingGame();
	static void EndingGame();
	static void SetServerAddress(Asura_NetAddress xServerAddress);

	static inline float GetTimeSlice() { return s_fTimeSlice; }

	// Is the Qos system asleep?
	// The server thread remains extant as long as the lobby exists,
	// but mostly sleeps whilst the game is playing.
	static bool IsAsleep() { return s_bThreadShouldSleep; }

	// Is the thread idle? (i.e. no testing is currently in progress)
	static bool IsIdle() { return s_bThreadIdle; }
	static void SetIdle(bool bIdle) { s_bThreadIdle = bIdle; }

	// Store the result of the testing.
	static void StoreResult(int iUploadBandwidthKbps, int iDownloadBandwidthKbps);

protected:
	// Create/destroy listener socket.
	static void CreateListener(u_short usLocalPort, bool bIsHost);
	static void DestroyListener(bool bCloseConnections = false);

	static bool IsSocketValid();
	static bool IsDone();			// Finished testing (or gave up due to error)?

	// Tell the client which server to connect to.
	static void AddServerConnection(Asura_NetAddress xServerAddress, u_short usServerPort);
	static void RequestServerConnect();

	static void UpdateThread(const void* pUserData);
	static void Update();

	static void UpdateTimeSlice();

	// Platform functions
	static void Platform_Initialise();
	static void Platform_Shutdown();

public:
	static Asura_QosSocket* Platform_CreateSocket(bool bIsHost);

protected:
	static Asura_Qos_Listener* s_pxListener;
	static bool s_bInitialised;

	static u_int volatile s_uNumberOfActiveThreads;
	static bool volatile s_bTerminateThreads;
	static bool volatile s_bThreadShouldSleep;
	static bool volatile s_bThreadIdle;

private:
	// Timer variables
	static float s_fTimeSlice;
	static Asura_TickCount s_xLastTickCount;
	
	// Server address.
	static Asura_NetAddress s_xServerAddress;

	static int s_iNumStoredResults;
};


/////////////////////
// Asura_QosSocket //
/////////////////////
class Asura_QosSocket
{
public:
	Asura_QosSocket() : m_bCreatedSocket(false) {}
	virtual ~Asura_QosSocket() {}

	virtual void CreateSocket(Asura_Qos_Listener* pxListener, u_short usLocalPort) = 0;
	virtual void CloseSocket()  = 0;

	// Return true if succeeded in sending the data.
	virtual bool SendTo(const Asura_Qos_NetConnection* pxDestination, const char* pcData, int iDataLength) = 0;

	// Return false if an error occurred.
	virtual bool Receive(Asura_Qos_Listener* pxListener, u_int uWaitTimeSeconds = 0) = 0;

	bool HasCreated() const { return m_bCreatedSocket; }

	// Return true if any low-level handshaking is finished.
	virtual bool IsReady() const { return true; }

protected:
	bool m_bCreatedSocket;
};

//////////////////////
// Asura_Qos_Result //
//////////////////////
// Result of a single large packet test.
class Asura_Qos_Result
{
public:
	void Reset()
	{
		m_uTestIDNumber = 0;
		m_uFlags = 0;

		m_iSendBandKbps = 0;
		m_iRecvBandKbps = 0;
		m_iAdjustedRTT  = 0;
	}

	bool IsEmpty()	  const { return m_uTestIDNumber == 0; }
	bool IsValid()	  const { return (!IsRejected() && !IsEmpty()); }
	bool IsRejected() const { return 0 != ( m_uFlags &ASURA_QOS_RESULT_FLAG_REJECT ); }	
	
	bool IsOutOfOrder()			const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_OUTOFORDER ); }
	bool HasLostPackets()		const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_LOSTPACKETS ); }
	bool IsLate()				const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_LATE ); }
	bool IsZerothOutOfOrder()	const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_ZEROTH_OUTOFORDER ); }
	bool HasTimedOut()			const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_TIMEOUT ); }
	bool IsTooFast()			const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_TOOFAST ); }
	bool IsQueued()				const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_QUEUING ); }

	bool IsDownloadFromRemote()	const { return 0 != ( m_uFlags & ASURA_QOS_RESULT_FLAG_DOWNLOAD_FROM_REMOTE ); }

	void PrintToBuffer(char* szBuffer, u_int uBufferSize) const;

public:
	u_int m_uTestIDNumber;
	u_int m_uFlags;

	int m_iSendBandKbps;	// Sent bandwidth (kbps)
	int m_iRecvBandKbps;	// Received bandwidth (kbps) - this is what we are measuring
	int m_iAdjustedRTT;		// Estimated round trip time for a small packet (microseconds)
};

/////////////////////////
// Asura_Qos_TestTicks //
/////////////////////////
// Data relating to an on-going test.
class Asura_Qos_TestTicks
{
public:
	void Reset();

	// Tick count after each packet sent/received;
	Asura_TickCount m_axTicks[uASURA_QOS_NUM_LARGE_PACKETS+1];
};

////////////////////
// Asura_Qos_Test //
////////////////////
class Asura_Qos_Test
{
public:
	void Reset();

	// Set functions
	void SetNull()
	{
		m_eTestState = ASURA_QOS_TESTSTATE_NULL;
		m_fDelayTimer = 0.0f;
	}

	void SetRequestedDelay(float fRequestedDelay)
	{
		m_fRequestedDelay = fRequestedDelay;
		m_fDelayTimer     = fRequestedDelay;
	}

	void SetLatePacketsFlags()
	{
		m_uReceivedFlags |= ASURA_QOS_RESULT_FLAG_LATE;
	}
	
	// Get functions
	float GetRequestedDelay()  const { return m_fRequestedDelay; }
	
	ASURA_QOS_TESTSTATE GetState() const { return m_eTestState; }
	
	bool IsStateNull() const
	{
		return	m_eTestState == ASURA_QOS_TESTSTATE_NULL;
	}
	
	bool IsStateReceiving() const
	{
		return	m_eTestState == ASURA_QOS_TESTSTATE_WAITING ||
				m_eTestState == ASURA_QOS_TESTSTATE_TESTING;
	}

	bool IsStateSending() const
	{
		return	m_eTestState == ASURA_QOS_TESTSTATE_START ||
				m_eTestState == ASURA_QOS_TESTSTATE_SENDING ||
				m_eTestState == ASURA_QOS_TESTSTATE_POST_TEST;
	}

	bool CanStartTest() const
	{
		return m_eTestState == ASURA_QOS_TESTSTATE_NULL ||
			   m_eTestState == ASURA_QOS_TESTSTATE_WAITING;
	}

	u_int GetNumTestResults() const { return m_uNumTestResults; }
	bool HasFinishedTests() const { return m_bHasFinished || (m_uNumTestResults == uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION); }

	u_int GetNumValidUploadResults()   const { return m_usNumValidUploadResults; }
	u_int GetNumValidDownloadResults() const { return m_usNumValidDownloadResults; }

	u_short GetPacketSeqID() const { return m_usPacketSeqID; }
	u_int GetTestID() const { return m_usPacketSeqID>>8; }
	u_int GetSubPacketID() const { return (m_usPacketSeqID & 0xff); }

	u_int GetReceivedFlags() const { return m_uReceivedFlags; }
	u_int GetReceivedSubIndex() const { return m_uReceivedSubIndex; }

	const Asura_TickCount* GetTicks(bool bPreviousTest) const
	{
		return (bPreviousTest ? m_xPrevTest.m_axTicks : m_xCurrentTest.m_axTicks);
	}

	const Asura_Qos_Result& GetResult(u_int uIndex) const { return m_axResult[uIndex]; }

	// Requested direction of current/next test.
	bool IsDownloadFromRemote() const { return m_bDownloadFromRemote; }

	// Store the result of a test.
	// Return result if added, or null otherwise.
	const Asura_Qos_Result* AddResult( u_int uTestIDNumber,
									   u_int uFlags,
									   int iSendBandKbps,
									   int iRecvBandKbps,
									   int iAdjustedRTT );

	// Sender/receiver
	void UpdateTest(Asura_Qos_NetConnection* pxConnection);
	void HandleStartTest();

	// Sender
	void UpdateDelayTimer();
	void SendTestLargePacketStart(Asura_Qos_NetConnection* pxConnection);
	bool SendTestLargePacket(Asura_Qos_NetConnection* pxConnection);
	void SummarizeResults(Asura_Qos_NetConnection* pxConnection);
	void HandleRecvTiming(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage);
	void HandleResult(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage);
	void AnalyseData(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage, bool bPreviousTest);
	void PrintResult(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Result* pxResult, const int* piSent, const int* piRecv);
	void HandlePostTestEnd(Asura_Qos_NetConnection* pxConnection);

	// Receiver
	void HandlePacket(Asura_Qos_NetConnection* pxConnection, u_short usPacketSeqID);
	void WaitForNextTest();
	void StartingTest();

protected:
	//
	// Sender/receiver
	//

	ASURA_QOS_TESTSTATE m_eTestState;	// State of test
	u_short m_usPacketSeqID;			// Packet seq ID (last id sent, or greatest received).
	
	// Tick count after each packet sent/received for current test.
	Asura_Qos_TestTicks m_xCurrentTest;

	//
	// Sender only
	//
	
	// Tick count after each packet sent for previous test.
	// Saved in case the received timings are sent back late.
	Asura_Qos_TestTicks m_xPrevTest;

	// Time until next state (seconds).
	float m_fDelayTimer;

	// Requested delay for the next packet (seconds).
	float m_fRequestedDelay;

	// Direction (client upload or server upload?).
	bool m_bDownloadFromRemote;
	// Have we got enough results?
	bool m_bHasFinished;
	// Have we summarized the results yet?
	bool m_bHasSummarized;

	//
	// Sender/receiver
	//
	// Previous test results.
	u_int m_uNumTestResults;
	Asura_Qos_Result m_axResult[uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION];
	
	u_short m_usNumValidUploadResults;
	u_short m_usNumValidDownloadResults;

	//
	// Receiver only
	//
	u_int m_uReceivedPacketFlags;		// Bit flags for received sub-packets.
	u_int m_uReceivedFlags;				// Result flags.
	u_int m_uReceivedSubIndex;			// Index for tickcount storage.
};

/////////////////////////////////
// Asura_Qos_NetConnectionInfo //
/////////////////////////////////
struct Asura_Qos_NetConnectionInfo
{
	Asura_Qos_NetConnectionInfo() { Reset(); }

	Asura_Qos_NetConnectionInfo(const Asura_NetAddress& xAddress, u_short usPort)
		: m_xAddress(xAddress), m_usPort(usPort)
	{
#ifdef ASURA_PC_SUPPORT_STEAMWORKS
		m_xSocket  = 0;
#endif //ASURA_PC_SUPPORT_STEAMWORKS	
	}

	void Reset()
	{
		m_xAddress = 0; 
		m_usPort   = 0;

#ifdef ASURA_PC_SUPPORT_STEAMWORKS
		m_xSocket  = 0;
#endif //ASURA_PC_SUPPORT_STEAMWORKS
	}

	bool operator==(const Asura_Qos_NetConnectionInfo& xOther) const
	{
		return (m_xAddress == xOther.m_xAddress) && (m_usPort == xOther.m_usPort);
	}

	bool operator!=(const Asura_Qos_NetConnectionInfo& xOther) const
	{
		return ! (*this == xOther);
	}

	Asura_NetAddress m_xAddress;	// Remote address IPv4 (network byte order for BSD) 
	u_short m_usPort;				// Remote port (host byte order)

#ifdef ASURA_PC_SUPPORT_STEAMWORKS
	SNetSocket_t m_xSocket;			// Relevant Steamworks socket for the remote address
#endif //ASURA_PC_SUPPORT_STEAMWORKS
};

/////////////////////////////
// Asura_Qos_NetConnection //
/////////////////////////////
class Asura_Qos_NetConnection
{
public:
	Asura_Qos_NetConnection()
		: m_pxListener(0)
	{
		Reset();
	}

	void Reset();

	// Parent listener
	void SetListener(Asura_Qos_Listener* pxListener) { m_pxListener = pxListener; }
	Asura_Qos_Listener* GetListener() { return m_pxListener; }

	// Remote address IPv4 (network byte order for BSD) 
	const Asura_NetAddress& GetRemoteAddress() const { return m_xRemoteInfo.m_xAddress; }
	// Remote port (host byte order)
	u_short GetRemotePort() const { return m_xRemoteInfo.m_usPort; }
	// Get connection info (IP address, port, Steam ID etc).
	const Asura_Qos_NetConnectionInfo& GetConnectionInfo() const { return m_xRemoteInfo; }
	// Set the connection info.
	void SetConnectionInfo(const Asura_Qos_NetConnectionInfo& xRemoteInfo) { m_xRemoteInfo = xRemoteInfo; }

	// Connection functions
	bool IsConnected() const { return (m_eState != ASURA_QOS_STATE_NULL && m_eState != ASURA_QOS_STATE_CLOSED); }
	bool HasConnectionProblems() const { return m_eState == ASURA_QOS_STATE_CONNECTED_PROBLEMS; }
	bool IsNullConnection()      const { return m_eState == ASURA_QOS_STATE_NULL; }
	bool IsClosedConnection()    const { return m_eState == ASURA_QOS_STATE_CLOSED; }

	void SetConnected() { m_eState = ASURA_QOS_STATE_CONNECTED; m_xTest.SetNull(); }
	void SetClosed()    { m_eState = ASURA_QOS_STATE_CLOSED;    m_xTest.SetNull(); }
	void CloseConnection();
	
	// Test data functions
	const Asura_TickCount* GetTicks(bool bPreviousTest) const { return m_xTest.GetTicks(bPreviousTest); }
	u_int GetReceivedFlags() const { return m_xTest.GetReceivedFlags(); }
	u_int GetReceivedSubIndex() const { return m_xTest.GetReceivedSubIndex(); }
	u_int GetTestID() const { return m_xTest.GetTestID(); }

	ASURA_QOS_TESTSTATE GetTestState() const { return m_xTest.GetState(); }
	
	bool IsTestStateNull() const	  { return m_xTest.IsStateNull(); }
	bool IsTestStateReceiving() const { return m_xTest.IsStateReceiving(); }
	bool IsTestStateSending() const	  { return m_xTest.IsStateSending(); }
	bool CanStartTest() const		  { return m_xTest.CanStartTest(); }
	bool HasFinishedTests() const	  { return m_xTest.HasFinishedTests(); }

	// Requested direction of current/next test.
	bool IsDownloadFromRemote() const { return m_xTest.IsDownloadFromRemote(); }

	// Send keep alive messages as necessary.
	void Update(const Asura_TickCount& xNow);

	// Send message to the remote machine corresponding to this connection.
	// Send a specific length packet.
	void SendPacketMessageTo( u_short usPacketSeqID,
							  u_int uDesiredPacketLength );


	// Send a message with data.
	void SendDataMessageTo( ASURA_QOS_MSGID eMsgID,
							u_short usPacketSeqID,
							u_int uNumDataVars );

	// Set the message data.
	void SetMessageData(int iData, u_int uIndex) const;

	// Send a message with no data.
	void SendMessageTo( ASURA_QOS_MSGID eMsgID );


	// Sender/receiver
	void UpdateDelayTimer()  { m_xTest.UpdateDelayTimer(); }
	void UpdateTest()		 { m_xTest.UpdateTest(this); }
	void HandleStartTest()	 { m_xTest.HandleStartTest(); }
	void SendStartTest();
	void HandlePostTestEnd() { m_xTest.HandlePostTestEnd(this); }

	// Sender
	void HandleRecvTiming(const Asura_Qos_Message* pxMessage) { m_xTest.HandleRecvTiming(this, pxMessage); }

	// Receiver
	void SendRecvTimingMessage(bool bZeroResults, u_short usPacketSeqID);
	void HandlePacket(u_short usPacketSeqID) { m_xTest.HandlePacket(this, usPacketSeqID); }
	void HandleResult(const Asura_Qos_Message* pxMessage) { m_xTest.HandleResult(this, pxMessage); }
	void SendRecvTimingMessagePartial();

protected:
	Asura_Qos_Listener* m_pxListener;			// Parent QoS listener

	Asura_Qos_NetConnectionInfo m_xRemoteInfo;	// Remote address, port (and steam ID if relevant).
	ASURA_QOS_CONNECTION_STATE m_eState;		// State of the connection

public:
	Asura_TickCount m_xLastMessageRecvTicks;
	Asura_TickCount m_xLastMessageSentTicks;

protected:
	Asura_Qos_Test m_xTest;
};


/////////////////////////////
// Asura_Qos_MessageHeader //
/////////////////////////////
struct Asura_Qos_MessageHeader
{
	Asura_Qos_MessageHeader();
	void Set(ASURA_QOS_MSGID eMsgID, u_short usPacketSeqID);

	bool IsPasswordValid() const;

	u_int   m_uNBOPassword;		// password
	u_short m_usNBOMsgID;		// QoS message ID
	u_short m_usNBOPacketSeqID;	// Packet sequence ID
};

///////////////////////
// Asura_Qos_Message //
///////////////////////
// Message that is big enough to represent a large packet (1024 bytes).
class Asura_Qos_Message
{
public:
	Asura_Qos_Message();

	u_int GetMsgID() const; // Get msg ID in host byte order.
	u_short GetPacketSeqID() const; // Get packet seq ID in host byte order.

	bool IsPasswordValid() const { return m_xHeader.IsPasswordValid(); }

	// Get integer data.
	int GetData_Int(u_int uIndex) const;

	// Set message data.
	void SetData(int iData, u_int uIndex);

	void FillData(); // Fill with dummy data.

public:
	Asura_Qos_MessageHeader m_xHeader;
	int m_aiData[uASURA_QOS_MAX_NUM_DATA_VARS]; // Data, mostly padding
};


// Message padded out to be bigger than the maximum number of bytes that can be read from a socket.
class Asura_Qos_Message_Padded : public Asura_Qos_Message
{
public:
	Asura_Qos_Message_Padded() {}

private:
	char m_acBuffer[ASURA_QOS_RECEIVE_BUFFER_SIZE - sizeof(Asura_Qos_Message)];
};


////////////////////////
// Asura_Qos_Listener //
////////////////////////
class Asura_Qos_Listener
{
public:
	Asura_Qos_Listener(u_int uMaxNumConnections, u_short usLocalPort, bool bIsHost);
	virtual ~Asura_Qos_Listener();

	bool IsDone() const { return m_bDone; }
	bool IsHost() const { return m_bIsHost; }

	bool IsSocketValid() const
	{
		return m_pxSocket && m_pxSocket->HasCreated();
	}

	// Steamworks sockets take a while to finish their handshaking and establish a connection.
	// Returns true if the low-level handshaking is finished, so that the QoS system can
	// start doing its own handshaking.
	bool IsSocketReady() const
	{
		return m_pxSocket && m_pxSocket->IsReady();
	}

	bool CreateSocket();
	void CloseSocket();

	virtual void Update() = 0;
	virtual void RequestConnect() {}

	virtual void HandleMessage(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage) = 0;

	void CloseConnections();

	// Set message data.
	void SetMessageData(int iData, u_int uIndex)
	{
		m_xMessage.SetData(iData, uIndex);
	}

	// Send a specific length packet.
	void SendPacketMessageTo( Asura_Qos_NetConnection* pxDestination,
							  u_short usPacketSeqID,
							  u_int uDesiredPacketLength );

	// Send a message with data.
	void SendDataMessageTo( Asura_Qos_NetConnection* pxDestination,
							ASURA_QOS_MSGID eMsgID,
							u_short usPacketSeqID,
							u_int uNumDataVars );

	// Send a message with no data.
	void SendMessageTo( Asura_Qos_NetConnection* pxDestination,
						ASURA_QOS_MSGID eMsgID );

	void ReceiveMessages(u_int uWaitTimeSeconds = 0);

	// Parameters in host byte order. Find the connection.
	// Return 0 if not found.
	Asura_Qos_NetConnection* FindConnection(const Asura_Qos_NetConnectionInfo& xRemoteInfo);

	// Parameters in host byte order. Add new connection if existing one not found.
	// Return 0 if couldn't add (i.e. reached maximum number of connections).
	Asura_Qos_NetConnection* AddConnection(const Asura_Qos_NetConnectionInfo& xRemoteInfo);

	// For clients the zeroth connection is the server.
	const Asura_Qos_NetConnection* GetServerConnection() const { return (m_bIsHost ? 0 : m_pxConnection); }

	void HandleMessageFrom(const Asura_Qos_NetConnectionInfo& xRemoteInfo, const Asura_Qos_Message* pxMessage);

protected:
	void SocketSendTo( Asura_Qos_NetConnection* pxDestination, int iDataLength);

protected:
	u_int m_uMaxNumConnections;	// Maximum number of active connections.
	u_short m_usLocalPort;		// Local port for the Qos listener.

	bool m_bDone;				// Finished?
	bool m_bIsHost;				// Is this the lobby host?

	Asura_Qos_NetConnection* m_pxConnection; // Connections.
	Asura_QosSocket* m_pxSocket; // Listener socket.

	Asura_Qos_Message m_xMessage; // Outgoing message
};

///////////////////////////////
// Asura_Qos_Listener_Server //
///////////////////////////////
class Asura_Qos_Listener_Server : public Asura_Qos_Listener
{
public:
	Asura_Qos_Listener_Server(u_short usLocalPort);
	virtual ~Asura_Qos_Listener_Server() {}

	virtual void Update();
	virtual void HandleMessage(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage);

protected:
	void UpdateTestTimer(bool bTesting);
	void UpdateActiveConnection(Asura_Qos_NetConnection& xConnection, u_int uIndex, bool& bTesting);


	float m_fTimeToNextTest;		// Time until next test (seconds).
	float m_fTimeOfCurrentTest;		// Duration of current test (seconds).

	u_int m_uLastConnectionTested;
};

///////////////////////////////
// Asura_Qos_Listener_Client //
///////////////////////////////
class Asura_Qos_Listener_Client : public Asura_Qos_Listener
{
public:
	Asura_Qos_Listener_Client(u_short usLocalPort)
		: Asura_Qos_Listener(1, usLocalPort, false)
	{}

	virtual ~Asura_Qos_Listener_Client() {}

	virtual void Update();
	virtual void RequestConnect();
	virtual void HandleMessage(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage);
};


#endif //#ifdef ASURA_ENABLE_QOS_SYSTEM

#endif //ASURA_QOS_SYSTEM_H__


