//////////////////////////////////////////////////////////////////////////////
//
// Asura_ReliableSocket_TCP.cpp
//
// Copyright (c) 2009 Rebellion. All rights reserved.
//
// Description:
//
//     TCP socket implementations of Asura_ReliableSocket classes
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// This file's header.
#include "Network/Asura_ReliableSocket_TCP.h"

// Asura includes
#include "Network/Asura_NetSocket.h"

// S T A T I C   F U N C T I O N S ///////////////////////////////////////////
static ASURA_TCP_DISCONNECT_REASON ConvertDisconnectReason(ASURA_RELIABLE_SOCKET_DISCONNECT_REASON eReason)
{
	switch(eReason)
	{
		case ASURA_RS_DISCONNECT_NOT_DISCONNECTED:
		{
			return ASURA_TCP_DISCONNECT_NOT_DISCONNECTED;
		}
		case ASURA_RS_DISCONNECT_SHUTDOWN: 
		{
			return ASURA_TCP_DISCONNECT_SHUTDOWN; 
		}
		case ASURA_RS_DISCONNECT_REJECTED: 
		{
			return ASURA_TCP_DISCONNECT_CONNECTION_REJECTED; 
		}
		case ASURA_RS_DISCONNECT_UNKNOWN:
		{
			return ASURA_TCP_DISCONNECT_UNKNOWN;
		}
		default: 
		{
			Asura_Assert(false, "unexpected disconnect reason");
			return ASURA_TCP_DISCONNECT_UNKNOWN;
		}
	}
}

static ASURA_RELIABLE_SOCKET_DISCONNECT_REASON ConvertDisconnectReason(ASURA_TCP_DISCONNECT_REASON eReason)
{
	switch(eReason)
	{
		case ASURA_TCP_DISCONNECT_NOT_DISCONNECTED:
		{
			return ASURA_RS_DISCONNECT_NOT_DISCONNECTED;
		}
		case ASURA_TCP_DISCONNECT_UNKNOWN:
		{
			return ASURA_RS_DISCONNECT_UNKNOWN;
		}
		case ASURA_TCP_DISCONNECT_SHUTDOWN: 
		{
			return ASURA_RS_DISCONNECT_SHUTDOWN; 
		}
		case ASURA_TCP_DISCONNECT_CONNECTION_REJECTED: 
		{
			return ASURA_RS_DISCONNECT_REJECTED; 
		}
		default: 
		{
			Asura_Assert(false, "unexpected disconnect reason");
			return ASURA_RS_DISCONNECT_UNKNOWN;
		}
	}
}

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
// constructor - init the TCP socket member with our static callback function and 'this' pointer
Asura_ReliableSocket_Client_TCP::Asura_ReliableSocket_Client_TCP()
: m_xTCPSocket(MessageCallback, NULL, DisconnectCallback)
{
	m_xTCPSocket.SetCallbackUserData(this);
}

bool Asura_ReliableSocket_Client_TCP::Connect( const Asura_NetAddress& xAddress, 
											  Asura_NetPort& xPort, 
											  const void* pConnectData, 
											  const u_short usConnectDataSize )
{
	const u_int uAddress = Asura_NetSocket::Platform_GetINetAddress(xAddress);
	return m_xTCPSocket.Connect( uAddress, xPort, usConnectDataSize, pConnectData );
}

void Asura_ReliableSocket_Client_TCP::Disconnect( const bool bImmediate, const ASURA_RELIABLE_SOCKET_DISCONNECT_REASON eReason )
{
	m_xTCPSocket.Disconnect(bImmediate, ConvertDisconnectReason(eReason));
}

void Asura_ReliableSocket_Client_TCP::Update()
{
	m_xTCPSocket.Update();
}

bool Asura_ReliableSocket_Client_TCP::SendMessage( const u_short usMsgID, const u_short usMsgBodySize, const void* pMsgBody )
{
	return m_xTCPSocket.SendMessage( usMsgID, usMsgBodySize, pMsgBody );
}

ASURA_RELIABLE_SOCKET_CLIENT_STATE Asura_ReliableSocket_Client_TCP::GetState()
{
	if (m_xTCPSocket.IsConnected())
	{
		return ASURA_RS_CLIENT_STATE_CONNECTED;
	}
	else if (m_xTCPSocket.IsActive())
	{
		return ASURA_RS_CLIENT_STATE_CONNECTING;
	}
	else
	{
		return ASURA_RS_CLIENT_STATE_INACTIVE;
	}
}

void Asura_ReliableSocket_Client_TCP::MessageCallback(Asura_ReliableConnectionID xConnectionID, u_short usMsgID, u_short usMsgBodySize, const char* pcMsgBody, void* pUserData)
{
	Asura_ReliableSocket_Client_TCP* pxThis = reinterpret_cast<Asura_ReliableSocket_Client_TCP*>(pUserData);
	if (pxThis && pxThis->m_pfnCallback_Message)
	{
		pxThis->m_pfnCallback_Message(usMsgID, usMsgBodySize, pcMsgBody);
	}
}

void Asura_ReliableSocket_Client_TCP::DisconnectCallback(Asura_ReliableConnectionID xConnectionID, ASURA_TCP_DISCONNECT_REASON eReason, void* pUserData)
{
	Asura_ReliableSocket_Client_TCP* pxThis = reinterpret_cast<Asura_ReliableSocket_Client_TCP*>(pUserData);
	if (pxThis && pxThis->m_pfnCallback_Disconnection)
	{
		pxThis->m_pfnCallback_Disconnection(ConvertDisconnectReason(eReason));
	}
}



//////////////////////////////////////////////////////////////////////////
// Listener
//////////////////////////////////////////////////////////////////////////

Asura_ReliableSocket_Listener_TCP::Asura_ReliableSocket_Listener_TCP()
: m_xTCPSocket(MessageCallback, ConnectCallback, DisconnectCallback)
{
	m_xTCPSocket.SetCallbackUserData(this);
}

bool Asura_ReliableSocket_Listener_TCP::Listen( const Asura_NetPort& xPort )
{
	return m_xTCPSocket.Listen(xPort);
}

bool Asura_ReliableSocket_Listener_TCP::Listen( const Asura_NetAddress& xAddress, const Asura_NetPort& xPort )
{
	const u_int uAddress = Asura_NetSocket::Platform_GetINetAddress(xAddress);
	return m_xTCPSocket.Listen(uAddress, xPort);
}

void Asura_ReliableSocket_Listener_TCP::Close()
{
	return m_xTCPSocket.Close();
}

void Asura_ReliableSocket_Listener_TCP::Disconnect( const Asura_ReliableConnectionID& xConnectionID, 
												   const ASURA_RELIABLE_SOCKET_DISCONNECT_REASON eReason )
{
	m_xTCPSocket.Disconnect(xConnectionID, ConvertDisconnectReason(eReason));
}

void Asura_ReliableSocket_Listener_TCP::Update()
{
	m_xTCPSocket.Update();
}

bool Asura_ReliableSocket_Listener_TCP::SendMessage( const Asura_ReliableConnectionID& xConnectionID, 
													const u_short usMsgID, 
													const u_short usMsgBodySize, 
													const void* pMsgBody )
{
	return m_xTCPSocket.SendMessage( xConnectionID, usMsgID, usMsgBodySize, pMsgBody);
}

bool Asura_ReliableSocket_Listener_TCP::SendMessageToAll( const u_short usMsgID, const u_short usMsgBodySize, const void* pMsgBody )
{
	return m_xTCPSocket.SendToAll( usMsgID, usMsgBodySize, pMsgBody );
}

bool Asura_ReliableSocket_Listener_TCP::SendMessageToAllExcept( const u_short usMsgID, 
															const u_short usMsgBodySize, 
															const void* pMsgBody, 
															const Asura_ReliableConnectionID& xException )
{
	return m_xTCPSocket.SendToAllExcept( usMsgID, usMsgBodySize, pMsgBody, xException );
}

u_short Asura_ReliableSocket_Listener_TCP::GetListenPort() const
{
	return m_xTCPSocket.GetListenPort();
}

u_int Asura_ReliableSocket_Listener_TCP::GetNumConnections() const
{
	return m_xTCPSocket.GetNumConnections();
}

ASURA_RELIABLE_SOCKET_LISTENER_STATE Asura_ReliableSocket_Listener_TCP::GetState()
{
	if (m_xTCPSocket.IsActive())
	{
		return ASURA_RS_LISTENER_STATE_LISTENING;
	}
	else
	{
		return ASURA_RS_LISTENER_STATE_INACTIVE;
	}
}

bool Asura_ReliableSocket_Listener_TCP::ConnectCallback(Asura_ReliableConnectionID xConnectionID, 
														u_short usMsgSize, const char* pcMsgBody, 
														void* pUserData)
{
	Asura_ReliableSocket_Listener_TCP* pxThis = reinterpret_cast<Asura_ReliableSocket_Listener_TCP*>(pUserData);
	return pxThis->m_pfnCallback_Connection(xConnectionID, usMsgSize, pcMsgBody);
}

void Asura_ReliableSocket_Listener_TCP::DisconnectCallback(Asura_ReliableConnectionID xConnectionID, 
														   ASURA_TCP_DISCONNECT_REASON eReason, 
														   void* pUserData)
{
	Asura_ReliableSocket_Listener_TCP* pxThis = reinterpret_cast<Asura_ReliableSocket_Listener_TCP*>(pUserData);
	pxThis->m_pfnCallback_Disconnection(xConnectionID, ConvertDisconnectReason(eReason));
}

void Asura_ReliableSocket_Listener_TCP::MessageCallback(Asura_ReliableConnectionID xConnectionID, 
														u_short usMsgID, u_short usMsgBodySize, 
														const char* pcMsgBody, void* pUserData)
{
	Asura_ReliableSocket_Listener_TCP* pxThis = reinterpret_cast<Asura_ReliableSocket_Listener_TCP*>(pUserData);
	pxThis->m_pfnCallback_Message(xConnectionID, usMsgID, usMsgBodySize, pcMsgBody);
}

#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
