//////////////////////////////////////////////////////////////////////////////
//
// Asura_GameSpy_GT2.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     Asura_GameSpy_GT2 class to provide Asura interface to GT2
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// required asura header
#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// this file's header
#include "Asura_GameSpy_GT2.h"

// middleware
extern "C" {
#include "GameSpy/gt2/gt2Socket.h"
#ifdef ASURA_SUPPORT_GAMESPY
#include "GameSpy/qr2/qr2.h"
#include "GameSpy/serverbrowsing/sb_serverbrowsing.h"
#endif // #ifdef ASURA_SUPPORT_GAMESPY
};

// asura independent includes
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_MessageHeader.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_Network_Session.h"
#include "Network/GameSpy/Asura_GameSpy.h"
#ifdef ASURA_SUPPORT_GAMESPY
#include "Network/GameSpy/Asura_GameSpy_QR2.h"
#include "Network/GameSpy/Asura_GameSpy_Peer.h"
#endif // #ifdef ASURA_SUPPORT_GAMESPY

// I N T E R N A L   C L A S S   D E F S /////////////////////////////////////

// C O N S T A N T S /////////////////////////////////////////////////////////
// for data received which doesn't come from an established connection, check 
// the first 2 bytes to see if they're magic. The QR2 sdk uses 0xFEFD for this
// and the GT2 sdk uses 0xFEFE in its connection protocol.
const u_short ASURA_GAMESPY_GT2_ENUMHOSTS_MAGICID = 0xFEFC;
const u_short ASURA_GAMESPY_GT2_HOSTREPLY_MAGICID = 0xFEFB;

// constants to help compute average ping
const u_int ASURA_GAMESPY_GT2_NUM_PING_SAMPLES_SHIFT = 2;
const u_int ASURA_GAMESPY_GT2_NUM_PING_SAMPLES = (1 << ASURA_GAMESPY_GT2_NUM_PING_SAMPLES_SHIFT);
const u_int ASURA_GAMESPY_GT2_NUM_PING_SAMPLES_MASK = ASURA_GAMESPY_GT2_NUM_PING_SAMPLES - 1;

#ifdef ASURA_SUPPORT_GAMESPY
const u_int ASURA_GAMESPY_GT2_NATNEG_CLIENT = 0;
const u_int ASURA_GAMESPY_GT2_NATNEG_SERVER = 1;
#endif // #ifdef ASURA_SUPPORT_GAMESPY

class Asura_GameSpy_GT2_MachineInfo
{
	public:

		Asura_GameSpy_GT2_MachineInfo()
			: m_pxConnectionHandle(NULL)
			, m_uPing(0)
			, m_uPingTableEntry(0)
			, m_uTimeOfLastPing(0)
			, m_uSentBytes(0)
			, m_uRecvBytes(0)
		{
		}

		void SetConnectionHandle( GT2Connection pxConnection )
		{
			Asura_Assert(!IsValid(), "overwriting valid connection");
			m_pxConnectionHandle = pxConnection;
			gt2SetConnectionData(pxConnection, this);
		}

		GT2Connection GetConnectionHandle( void ) { return m_pxConnectionHandle; }

		void ClearConnection( void )
		{
			Asura_Assert(IsValid(), "clearing invalid connection");
			gt2SetConnectionData(m_pxConnectionHandle, 0);
			m_pxConnectionHandle = 0;
			m_uPing = 0;
			m_uPingTableEntry = 0;
			m_uSentBytes = 0;
			m_uRecvBytes = 0;
		}
			
		bool IsValid( void ) const { return m_pxConnectionHandle != NULL; }
		
		// ping stuff
		u_int GetPing( void ) const
		{
			Asura_Assert(IsValid(), "ping of invalid connection requested");
			return m_uPing;
		}
		u_int GetTimeOfLastPing( void ) const { return m_uTimeOfLastPing; }
		void SetTimeOfLastPing( u_int uTime ) { m_uTimeOfLastPing = uTime; }
		void InitPing( u_int uPing )
		{
			m_uPing = uPing;
			for (u_int u = 0; u < ASURA_GAMESPY_GT2_NUM_PING_SAMPLES; ++u)
			{
				m_auPingTable[u] = uPing;
			}
		}
		void SetPing( u_int uPing )
		{ 
			m_auPingTable[m_uPingTableEntry++] = uPing;
			m_uPingTableEntry &= ASURA_GAMESPY_GT2_NUM_PING_SAMPLES_MASK;
			u_int uPingSum = 0;
			for (u_int u = 0; u < ASURA_GAMESPY_GT2_NUM_PING_SAMPLES; ++u)
			{
				uPingSum += m_auPingTable[u];
			}
			m_uPing = uPingSum >> ASURA_GAMESPY_GT2_NUM_PING_SAMPLES_SHIFT;
		}

		// bytes sent stuff - changed to clear after getting to avoid errors
		// when machine disconnects.
		void AddSentBytes( u_int uNumBytes ) { m_uSentBytes += uNumBytes; }
		void AddReceivedBytes( u_int uNumBytes ) { m_uRecvBytes += uNumBytes; }
		u_int GetBytesSent( void ) 
		{
			u_int uReturn = m_uSentBytes;
			m_uSentBytes = 0;
			return uReturn; 
		}
		u_int GetBytesReceived( void ) 
		{
			u_int uReturn = m_uRecvBytes;
			m_uRecvBytes = 0;
			return uReturn;
		}

	private:
		
		GT2Connection m_pxConnectionHandle;
		
		// ping stuff
		u_int m_uPing;
		u_int m_auPingTable[ASURA_GAMESPY_GT2_NUM_PING_SAMPLES];
		u_int m_uPingTableEntry;
		u_int m_uTimeOfLastPing;

		// throughput stuff
		u_int m_uSentBytes;
		u_int m_uRecvBytes;
};


class Asura_GameSpy_GT2_ConnectData
{

	public:

		// should include password hash, game version, and, if supported, CD key
		u_short m_usGameVersion;
		u_short m_usNumPlayers;
		u_int m_uPasswordHash;
};


class Asura_GameSpy_GT2_ConnectionRejectedData
{

	public:
	
		u_int m_uReason;
};


class Asura_GameSpy_GT2_LANBroadcastServerRequest
{

	public:

		u_short m_usMagicID;		// ASURA_GAMESPY_GT2_ENUMHOSTS_MAGICID = 0xfefc
		u_short m_usGameVersion;
		u_int m_uApplicationGuid;
};


class Asura_GameSpy_GT2_LANBroadcastServerReplyHeader
{
	public:

		u_short m_usMagicID;		// ASURA_GAMESPY_GT2_HOSTREPLY_MAGICID = 0xfefb
		u_short m_usGameVersion;
		u_int m_uApplicationGuid;

		u_short m_usDataSize;		// this data is followed by this many bytes of user data
		Asura_UnicodeChar m_wszSessionName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
		u_char m_ucNumPlayers;
		u_char m_ucMaxPlayers;
};


class Asura_GameSpy_GT2_SessionInfo
{
	public:

		Asura_GameSpy_GT2_SessionInfo(u_int uHostAddr, u_short usHostPort, 
			u_int uPing, const u_char* pcSessionData, u_int uSessionDataSize, 
			const Asura_UnicodeChar* pszSessionName, u_char ucNumPlayers, u_char ucMaxPlayers );
		~Asura_GameSpy_GT2_SessionInfo();
		
		bool IsSameHost(u_int uHostAddr, u_short usHostPort) const 
		{ 
			return (uHostAddr == m_uHostAddr) && (usHostPort == m_usHostPort); 
		}
		void Refresh(u_int uPing, const u_char* pcSessionData, u_int uSessionDataSize, 
			const Asura_UnicodeChar* pszSessionName, u_char ucNumPlayers, u_char ucMaxPlayers);
		
		u_int GetHostAddr( void ) const { return m_uHostAddr; }
		u_short GetHostPort( void ) const { return m_usHostPort; }
		u_int GetNumPlayers( void ) const { return m_ucNumPlayers; }
		u_int GetMaxPlayers( void ) const { return m_ucMaxPlayers; }
		const Asura_UnicodeChar* GetSessionName( void ) const { return m_wszSessionName; };
		const void* GetSessionData( void ) const { return reinterpret_cast<const void*>(m_pucSessionData); }
		u_int GetSessionDataSize( void ) const { return m_uSessionDataSize; }
		u_int GetPing( void ) const { return m_uPing; }
		float GetLastUpdateTime( void ) const { return m_fLastUpdateTime; }
		void GetAddress( u_int* puIP, u_short* pusPort ) const { *puIP = m_uHostAddr; *pusPort = m_usHostPort; }

	private:
		
		// private and unimplemented copy construct and assignment operator
		Asura_GameSpy_GT2_SessionInfo(const Asura_GameSpy_GT2_SessionInfo& xFrom);
		Asura_GameSpy_GT2_SessionInfo& operator=(const Asura_GameSpy_GT2_SessionInfo& xRHS);

		u_int m_uHostAddr;
		u_short m_usHostPort;

		float m_fLastUpdateTime;
		u_int m_uPing;

		u_char* m_pucSessionData;
		u_int m_uSessionDataSize;

		u_char m_ucNumPlayers;
		u_char m_ucMaxPlayers;
		Asura_UnicodeChar m_wszSessionName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
};


// D E F I N E S /////////////////////////////////////////////////////////////
#define ASURA_GAMESPY_GT2_BUFFERSIZE (2048)	// max amount of data (in bytes) that can be sent in one go.
#define ASURA_GAMESPY_GT2_SOCKET_INCOMING_BUFFERSIZE (1024)	// bytes for messages buffered for in-order delivery
#define ASURA_GAMESPY_GT2_SOCKET_OUTGOING_BUFFERSIZE (32*1024)	// bytes for messages buffered for guaranteed delivery

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR u_int dbg_uGameSpyGT2PingInterval = 2000;	// 2 seconds
DEBUGVAR float dbg_fGameSpyGT2EnumHostsRefreshTime = 1.0f;
DEBUGVAR float dbg_fGameSpyGT2EnumHostsExpiryTime = 2.0f;
DEBUGVAR float dbg_fTimeBetweenBandwidthCalculations = 5.0f;
// if dbg_bPrintAllNetworkTraffic is true, every sent or received packet is 
// printed in the debugger with To/From IP:port, num bytes and hex data.
DEBUGVAR bool dbg_bPrintAllNetworkTraffic = false;

// S T A T I C   D A T A /////////////////////////////////////////////////////

// general
GT2Socket Asura_GameSpy_GT2::s_pxSocket = 0;
u_short Asura_GameSpy_GT2::s_usPort = ASURA_GAMESPY_GT2_DEFAULT_PORT;
u_int Asura_GameSpy_GT2::s_uPasswordHash = 0;
u_char* Asura_GameSpy_GT2::s_pucSessionData = NULL;
u_int Asura_GameSpy_GT2::s_uSessionDataSize = 0;
float Asura_GameSpy_GT2::s_fTimeSinceLastBandwidthCalculation;
u_int Asura_GameSpy_GT2::s_uSendBPS;
u_int Asura_GameSpy_GT2::s_uReceiveBPS;
u_int Asura_GameSpy_GT2::s_uPeakSendBPS;
u_int Asura_GameSpy_GT2::s_uPeakReceiveBPS;


Asura_GameSpy_GT2_MachineInfo* Asura_GameSpy_GT2::s_pxMachineInfo = 0;
u_int Asura_GameSpy_GT2::s_uMaxNumberOfMachines = 0;
Asura_GameSpy_GT2_NetworkErrorType Asura_GameSpy_GT2::s_eLastNetworkError = ASURA_GT2_NETERROR_NONE;

// static data - server
bool Asura_GameSpy_GT2::s_bServerLanReportingOn = false;
// static data - client
bool Asura_GameSpy_GT2::s_bUseNatNeg = false;
int Asura_GameSpy_GT2::s_iNatNegCookie;
Asura_GameSpy_GT2::ClientConnectionState Asura_GameSpy_GT2::s_eClientConnectionState = Asura_GameSpy_GT2::CCS_NotConnected;
char Asura_GameSpy_GT2::s_acServerAddress[32];
u_int Asura_GameSpy_GT2::s_uNumberOfActiveSessions;
Asura_GameSpy_GT2_SessionInfo** Asura_GameSpy_GT2::s_ppxSessionsStorage;
float Asura_GameSpy_GT2::s_fLastSessionRequestBroadcastTime;

// G L O B A L S /////////////////////////////////////////////////////////////
static u_char g_aucBuffer[ASURA_GAMESPY_GT2_BUFFERSIZE];

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_GameSpy_GT2::Initialise( void )
{
	Platform_Initialise();

	if (!s_pxSocket)
	{
		GetSocket();
	}

	Asura_Assert(!s_pxMachineInfo, "GT2 already initialised");
	if (Asura_Network::IsServer())
	{
		s_uMaxNumberOfMachines = Asura_Network::GetMaxNumberOfMachines();
	}
	else
	{
		s_uMaxNumberOfMachines = 1;
	}
	s_pxMachineInfo = new Asura_GameSpy_GT2_MachineInfo[s_uMaxNumberOfMachines];


	s_pucSessionData = NULL;
	s_uSessionDataSize = 0;

	s_fTimeSinceLastBandwidthCalculation = 0.0f;
	s_uPeakSendBPS = 0;
	s_uPeakReceiveBPS = 0;
}


SOCKET Asura_GameSpy_GT2::GetSocket( void )
{
	if (!s_pxSocket)
	{
		char ucPort[16];
		Asura_PrintfToBfr(ucPort, ":%d", 16, true, s_usPort);
		
		GT2Result eResult = gt2CreateSocket(
			&s_pxSocket,		// GT2Socket * socket,
			ucPort,				// const char * localAddress, This is the address to bind to locally.  Typically of the form ":<port>", e.g., ":7777". Can be NULL or "".
			ASURA_GAMESPY_GT2_SOCKET_OUTGOING_BUFFERSIZE,// int outgoingBufferSize, This is the byte size of the buffer for reliable outgoing messages. This is a per-connection buffer.  Can be 0 to use the internal default.
			ASURA_GAMESPY_GT2_SOCKET_INCOMING_BUFFERSIZE,// int incomingBufferSize, This is the byte size of the buffer for out-of-order reliable incoming messages. This is a per-connection buffer.  Can be 0 to use the internal default.
			SocketErrorCallback	// gt2SocketErrorCallback callback
		);
		Asura_Assert(eResult == GT2Success, "failed to create local socket");
		if (eResult == GT2Success)
		{
			gt2SetUnrecognizedMessageCallback(s_pxSocket, UnrecognizedMessageCallback);

			gt2SetSendDump(s_pxSocket, SendDumpCallback);
			gt2SetReceiveDump(s_pxSocket, ReceiveDumpCallback);
		}
		else
		{
			s_pxSocket = 0;
		}
	}
	return gt2GetSocketSOCKET(s_pxSocket);
}


void Asura_GameSpy_GT2::Update( void )
{
	for (u_int u = 0; u < s_uMaxNumberOfMachines; ++u)
	{
		Asura_GameSpy_GT2_MachineInfo& xMachineInfo = s_pxMachineInfo[u];
		if (xMachineInfo.IsValid())
		{
			u_int uCurrentTime = Asura_Timers::GetMSecsSinceStart();
			u_int uTimeSinceLastPing = uCurrentTime - xMachineInfo.GetTimeOfLastPing();
			if (uTimeSinceLastPing > dbg_uGameSpyGT2PingInterval)
			{
				gt2Ping(xMachineInfo.GetConnectionHandle());
				xMachineInfo.SetTimeOfLastPing(uCurrentTime);
			}
		}
	}

#ifdef ASURA_SUPPORT_GAMESPY
	// update natneg - no harm in doing it every frame
	NNThink();
#endif // #ifdef ASURA_SUPPORT_GAMESPY

	// if client connection state machine active, update it
	switch (s_eClientConnectionState)
	{
		case CCS_WaitingForNatNeg:
		{
			break;
		}
		case CCS_ReadyToConnect:
		{
			ConnectToHost(s_acServerAddress);
			s_eClientConnectionState = CCS_WaitingForConnection;
			break;
		}
		default:
		{
			// not a client, or nothing to do
			break;
		}
	}

	//Asura_Assert(s_pxSocket, "no socket");
	if (s_pxSocket)
	{
		gt2Think(s_pxSocket);
	}

	UpdateBandwidthStats();
}


void Asura_GameSpy_GT2::EndOfFrame( void )
{
	if (s_pxSocket)
	{
		gt2Think(s_pxSocket);
	}
}


void Asura_GameSpy_GT2::Shutdown( void )
{
	Asura_Assert(s_pxSocket, "no socket");
	if (s_pxSocket)
	{
		gt2CloseSocket(s_pxSocket);
		s_pxSocket = NULL;
	}

	if (s_pxMachineInfo)
	{
		delete[] s_pxMachineInfo;
		s_pxMachineInfo = NULL;
		s_uMaxNumberOfMachines = 0;

		s_uSessionDataSize = 0;
		delete[] s_pucSessionData;
		s_pucSessionData = NULL;
	}

	s_bUseNatNeg = false;
	s_eClientConnectionState = CCS_NotConnected;

	Platform_Shutdown();
}


// server code
bool Asura_GameSpy_GT2::StartServerAndHostSession( bool bLanOnly )
{
	Asura_Assert(s_pxSocket, "no socket");
	if (s_pxSocket)
	{
		gt2Listen(s_pxSocket, ServerCallback_Connection);
#ifdef ASURA_SUPPORT_GAMESPY
		Asura_GameSpy_QR2::Initialise(bLanOnly, gt2GetSocketSOCKET(s_pxSocket), s_usPort);
#endif // #ifdef ASURA_SUPPORT_GAMESPY
		return true;
	}
	return false;
}


void Asura_GameSpy_GT2::StopServer( void )
{
	Asura_Assert(s_pxSocket, "no socket");
	if (s_pxSocket)
	{
		gt2Listen(s_pxSocket, NULL);
	}
	s_bServerLanReportingOn = false;
}


void Asura_GameSpy_GT2::SetServerLANReporting( bool bEnable )
{
	s_bServerLanReportingOn = bEnable;
}


void Asura_GameSpy_GT2::SendMessageToClient( u_int uMachine, const Asura_Message* pxMessage )
{
	Asura_GameSpy_GT2_MachineInfo* pxMachineInfo = GetMachineInfoFromIndex(uMachine);
	if (pxMachineInfo->IsValid())
	{
		SendAsuraMessage(pxMachineInfo->GetConnectionHandle(), pxMessage);
	}
}


void Asura_GameSpy_GT2::SendMessageToAllClients( const Asura_Message* pxMessage )
{
	for (u_int u = 0; u < s_uMaxNumberOfMachines; ++u)
	{
		if (s_pxMachineInfo[u].IsValid())
		{
			SendMessageToClient(u, pxMessage);
		}
	}
}


void Asura_GameSpy_GT2::SetServerName( const Asura_UnicodeChar* wszServerName )
{
	// TODO - LAN broadcast stuff
}


u_int Asura_GameSpy_GT2::GetLatencyForMachine( u_int uMachine )
{
	Asura_Assert(uMachine < s_uMaxNumberOfMachines, "invalid machine index");
	const Asura_GameSpy_GT2_MachineInfo* pxInfo = &s_pxMachineInfo[uMachine];
	return pxInfo->GetPing();	
}

#ifdef ASURA_SUPPORT_GAMESPY
void Asura_GameSpy_GT2::ServerBeginNatNeg(int iCookie)
{
	NegotiateError eNatNegError = NNBeginNegotiationWithSocket(gt2GetSocketSOCKET(s_pxSocket), iCookie, ASURA_GAMESPY_GT2_NATNEG_SERVER, NatNegProgress_ServerCallback, NatNegComplete_ServerCallback, NULL);
	Asura_Assert(ne_noerror == eNatNegError, "natneg error");
}
#endif // #ifdef ASURA_SUPPORT_GAMESPY

// client code
void Asura_GameSpy_GT2::StartClient( const char* szIPAddressToConnectTo )
{
	Asura_Assert(s_pxSocket, "no socket");
	if (!s_pxSocket) return;

	Asura_Assert(szIPAddressToConnectTo, "no server to connect to!");
	if (szIPAddressToConnectTo)
	{
		// if the address doesn't have the port in it, add our default port
		if (!strstr(szIPAddressToConnectTo, ":"))
		{
			Asura_PrintfToBfr(s_acServerAddress, "%s:%d", 32, true, szIPAddressToConnectTo, s_usPort);
		}
		else 
		{
			Asura_PrintfToBfr(s_acServerAddress, "%s", 32, true, szIPAddressToConnectTo);
		}
		
		if (s_bUseNatNeg)
		{
#ifdef ASURA_SUPPORT_GAMESPY
			s_eClientConnectionState = CCS_WaitingForNatNeg;

			NegotiateError eNatNegError = NNBeginNegotiationWithSocket(gt2GetSocketSOCKET(s_pxSocket), s_iNatNegCookie, ASURA_GAMESPY_GT2_NATNEG_CLIENT, NatNegProgress_ClientCallback, NatNegComplete_ClientCallback, NULL);
			Asura_Assert(ne_noerror == eNatNegError, "natneg error");
#else  // not #ifdef ASURA_SUPPORT_GAMESPY
			Asura_Assert(false, "NatNeg not supported");
			s_eClientConnectionState = CCS_ReadyToConnect;
#endif // end #else  // not #ifdef ASURA_SUPPORT_GAMESPY
		}
		else
		{
			// connect directly
			s_eClientConnectionState = CCS_ReadyToConnect;
		}
	}
}


void Asura_GameSpy_GT2::StopClient( void )
{
	if (s_pxMachineInfo->IsValid())
	{
		gt2CloseConnection(s_pxMachineInfo->GetConnectionHandle());	
	}
}


void Asura_GameSpy_GT2::StartSessionManager( void )
{
	Asura_Assert(s_pxSocket, "no socket");
	if (!s_pxSocket) return;

	// init client stuff
	s_uNumberOfActiveSessions = 0;
	s_ppxSessionsStorage = new Asura_GameSpy_GT2_SessionInfo*[uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS];
	
	if (Platform_CanUseLANBroadcast())
	{
		// enable broadcasts on the socket
		SOCKET xSock = gt2GetSocketSOCKET(s_pxSocket);
		int iRcode;
		int iBroadcast = 1;
		int iLen;

		iLen = sizeof(iBroadcast);
		iRcode = setsockopt(xSock, SOL_SOCKET, SO_BROADCAST, (char *)&iBroadcast, iLen);
		Asura_Assert(iRcode != SOCKET_ERROR, "setsockopt error");
	
		// send broadcast message
		SendEnumHostsBroadcast();
	}
}


void Asura_GameSpy_GT2::StopSessionManager( void )
{
	for (u_int u = 0; u < s_uNumberOfActiveSessions; ++u)
	{
		delete s_ppxSessionsStorage[u];
	}
	s_uNumberOfActiveSessions = 0;
	delete[] s_ppxSessionsStorage;
	s_ppxSessionsStorage = NULL;
}

//void Asura_GameSpy_GT2::ConnectToSession( u_int uSessionIndex )
//{
//	Asura_Assert(uSessionIndex < s_uNumberOfActiveSessions, "invalid session");
//	Asura_GameSpy_GT2_SessionInfo* pxSessionInfo = s_ppxSessionsStorage[uSessionIndex];
//	u_int uIP;
//	u_short usPort;
//	pxSessionInfo->GetAddress(&uIP, &usPort);
//	char ucFullAddress[32];
//	gt2AddressToString( uIP, usPort, ucFullAddress );
//	ConnectToHost( ucFullAddress );
//
//	// set our copy of the session data to be this session's
//	SetSessionData(reinterpret_cast<const Asura_Network_SessionData*>(pxSessionInfo->GetSessionData()), pxSessionInfo->GetSessionDataSize());
//
//	// can get rid of the session data now.
//	for (u_int u = 0; u < s_uNumberOfActiveSessions; ++u)
//	{
//		delete s_ppxSessionsStorage[u];
//	}
//	s_uNumberOfActiveSessions = 0;
//	delete[] s_ppxSessionsStorage;
//	s_ppxSessionsStorage = NULL;
//}


void Asura_GameSpy_GT2::SendMessageToServer( const Asura_Message* pxMessage )
{
	if (s_pxMachineInfo->IsValid())
	{
		SendAsuraMessage(s_pxMachineInfo->GetConnectionHandle(), pxMessage);
	}
}


u_int Asura_GameSpy_GT2::GetLatency( void )
{
	if (s_pxMachineInfo->IsValid())
	{
		return s_pxMachineInfo->GetPing();
	}
	return 999;
}


void Asura_GameSpy_GT2::UpdateLANSessions( void )
{
	float fCurrentTime = Asura_Timers::GetLocalUnscaledTime();

	if (Platform_CanUseLANBroadcast())
	{
		if ( (fCurrentTime - s_fLastSessionRequestBroadcastTime) > dbg_fGameSpyGT2EnumHostsRefreshTime)
		{
			SendEnumHostsBroadcast();
		}
	}
	
	for (u_int u = 0; u < s_uNumberOfActiveSessions; u++)
	{
		Asura_GameSpy_GT2_SessionInfo* pxGT2Session = s_ppxSessionsStorage[u];
		Asura_Assert(pxGT2Session, "no session data");

		float fTimeSinceLastUpdate = fCurrentTime - pxGT2Session->GetLastUpdateTime();

		if (fTimeSinceLastUpdate > dbg_fGameSpyGT2EnumHostsExpiryTime) 
		{
			delete pxGT2Session;
			// swap with last, and set the loop to process this session
			s_ppxSessionsStorage[u] = s_ppxSessionsStorage[--s_uNumberOfActiveSessions];
		}
		else
		{
			Asura_Network_Session_System::SetSession(	u, 
														pxGT2Session->GetHostAddr(),
														pxGT2Session->GetHostPort(),
														pxGT2Session->GetNumPlayers(),
														pxGT2Session->GetMaxPlayers(),
														pxGT2Session->GetSessionName(),
														pxGT2Session->GetPing(),
														pxGT2Session->GetSessionData(),
														pxGT2Session->GetSessionDataSize() );
		}
	}

	// reset the remaining sessions in asura session system
	for (u_int u = s_uNumberOfActiveSessions; u < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS; ++u)
	{
		Asura_Network_Session_System::ClearSession(u);
	}

	// confirm the number of active sessions
	Asura_Network_Session_System::SetNumberOfSessions(s_uNumberOfActiveSessions);

}


// client and server code
void Asura_GameSpy_GT2::SetPassword( const char* szPassword )
{
	Asura_Assert(szPassword, "no password");
	if (szPassword)
	{
		s_uPasswordHash = Asura_GetHashID(szPassword);
	}
}


void Asura_GameSpy_GT2::GetNetworkBandwidthUsage(u_int& uSend, u_int& uReceive)
{
	uSend = s_uSendBPS;
	uReceive = s_uReceiveBPS;
}


void Asura_GameSpy_GT2::GetPeakNetworkBandwidthUsage(u_int& uSend, u_int& uReceive)
{
	uSend = s_uPeakSendBPS;
	uReceive = s_uPeakReceiveBPS;
}


// server utilities
void Asura_GameSpy_GT2::ProcessMachineConnection( GT2Connection pxConnection, u_short usNumPlayers, u_int uPing )
{
	// store machine info
	Asura_Handle iMachineHandle = Asura_Network::AddMachineToGame(usNumPlayers);

	Asura_Assert(iMachineHandle != ASURA_INVALID_HANDLE, "Too many machines! Should have been stopped before now!");
	Asura_Assert(static_cast<u_int>(iMachineHandle) < s_uMaxNumberOfMachines, "too many machines");
			
	s_pxMachineInfo[iMachineHandle].SetConnectionHandle(pxConnection);
	s_pxMachineInfo[iMachineHandle].InitPing(uPing);
    
	// need to send client guids to these players.
	for (u_int u = 0; u < usNumPlayers; ++u)
	{
		// get client guid, this function will add the guid to the Asura_Network machine info
		Asura_Guid uClientGuid = Asura_Network::GetNewClientGuid(iMachineHandle);
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_VALIDCLIENTGUID,sizeof(Asura_MsgData_ValidClientGuid));

		Asura_MsgData_ValidClientGuid* pxData = reinterpret_cast
			<Asura_MsgData_ValidClientGuid*>(pxMessage->GetData());

		pxData->Guid = uClientGuid;

		pxMessage->To = ASURA_GUID_CLIENT;
		pxMessage->UseGuaranteedDelivery();

		SendAsuraMessage(pxConnection, pxMessage);
		Asura_Network::DestroyMessage(pxMessage);
	}

//	// send the session data
//	Asura_Message* pxMsg = Asura_Network::CreateMessage(ASURA_MSG_NETWORKSESSIONDATA, s_uSessionDataSize, 0, s_pucSessionData);
//	pxMsg->To = ASURA_GUID_CLIENT;
//	pxMsg->UseGuaranteedDelivery();
//	SendAsuraMessage(pxConnection, pxMsg);
//	Asura_Network::DestroyMessage(pxMsg);
}


bool Asura_GameSpy_GT2::ProcessEnumHostsRequest( u_int uIP, u_short usPort, u_char* pucData, u_int uDataLength )
{
	Asura_GameSpy_GT2_LANBroadcastServerRequest* pxBroadcastRequestMsg = reinterpret_cast
		<Asura_GameSpy_GT2_LANBroadcastServerRequest*>(pucData);

	if (pxBroadcastRequestMsg->m_uApplicationGuid != Project_GetApplicationGuid())
	{
		return false;
	}

	if (pxBroadcastRequestMsg->m_usGameVersion != Asura_GameSpy::GetGameVersion())
	{
		return false;
	}

	SendEnumHostsResponse( uIP, usPort );

	return true;
}


void Asura_GameSpy_GT2::SendEnumHostsResponse(u_int uIP, u_short usPort)
{
	Asura_Assert(s_pxSocket, "no socket");
	if (!s_pxSocket) return;

	// (MH - 01.07.2004) ensure the session data is fresh
	Project_RefreshSessionData();

	u_int uSizeOfReplyData = sizeof(Asura_GameSpy_GT2_LANBroadcastServerReplyHeader) + s_uSessionDataSize;
	u_char* const pucReplyData = new u_char[uSizeOfReplyData];

	Asura_GameSpy_GT2_LANBroadcastServerReplyHeader* pxHeader = reinterpret_cast
		<Asura_GameSpy_GT2_LANBroadcastServerReplyHeader*>(pucReplyData);
	pxHeader->m_usMagicID = ASURA_GAMESPY_GT2_HOSTREPLY_MAGICID;
	pxHeader->m_usGameVersion = Asura_GameSpy::GetGameVersion();
	pxHeader->m_uApplicationGuid = Project_GetApplicationGuid();
	pxHeader->m_usDataSize = static_cast<u_short>(s_uSessionDataSize);
	pxHeader->m_ucNumPlayers = Asura_Network::GetNumberOfPlayers();
	pxHeader->m_ucMaxPlayers = Asura_Network::GetMaxNumberOfPlayers();
	Asura_PrintfToBfr(pxHeader->m_wszSessionName, L"%s", uASURA_NETWORK_MAX_SESSION_NAME + 1, true, Asura_Network::GetServerName());

	u_char* pucSessionData = reinterpret_cast<u_char*>(pxHeader + 1);
	memcpy(pucSessionData, s_pucSessionData, s_uSessionDataSize);

	GT2Bool iResult = gti2SocketSend(s_pxSocket, uIP, usPort, pucReplyData, uSizeOfReplyData);
	Asura_Assert(iResult, "failed to send broadcast message");

	delete[] pucReplyData;
}


// server callbacks
void Asura_GameSpy_GT2::ServerCallback_Connection( GT2Socket pxSocket,
												   GT2Connection pxConnection, 
												   unsigned int uIP, 
												   unsigned short usPort, 
												   int iLatency, 
												   GT2Byte* pucMessage, 
												   int iMessageLen )
{
	// We need to call either gt2Accept or gt2Reject in response to this, but 
	// we don't have to do it immediately if, say, we want to authenticate a 
	// CD key with another computer, or see if the user is banned.
	
	// check number of client machines
	if( Asura_Network::GetNumberOfMachines() >= Asura_Network::GetMaxNumberOfMachines() )
	{
		Asura_GameSpy_GT2_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uReason = ASURA_NDR_SERVERFULL;
		gt2Reject(pxConnection, reinterpret_cast<const u_char*>(&xRejectionReason), sizeof(xRejectionReason));
		Asura_OutputToDebugger_Retail(("connection from %s rejected: server full.\n", gt2AddressToString(uIP, usPort, NULL)));
		return;
	}

	// check data received for correct password etc.
	// first check validity of message data struct
	if (iMessageLen != sizeof(Asura_GameSpy_GT2_ConnectData))
	{
		Asura_GameSpy_GT2_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uReason = ASURA_NDR_VERSION;
		gt2Reject(pxConnection, reinterpret_cast<const u_char*>(&xRejectionReason), sizeof(xRejectionReason));
		Asura_OutputToDebugger_Retail(("connection from %s rejected: data struct size differs.\n", gt2AddressToString(uIP, usPort, NULL)));
		return;
	}

	const Asura_GameSpy_GT2_ConnectData* pxConnectData = reinterpret_cast
		<const Asura_GameSpy_GT2_ConnectData*>(pucMessage);
	
	// check game version
	if (pxConnectData->m_usGameVersion != Asura_GameSpy::GetGameVersion())
	{
		Asura_GameSpy_GT2_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uReason = ASURA_NDR_VERSION;
		gt2Reject(pxConnection, reinterpret_cast<const u_char*>(&xRejectionReason), sizeof(xRejectionReason));
		Asura_OutputToDebugger_Retail(("connection from %s rejected: game version differs.\n", gt2AddressToString(uIP, usPort, NULL)));
		return;
	}

	// check number of players
	u_int uNumPlayersIfAccepted = Asura_Network::GetNumberOfPlayers() + pxConnectData->m_usNumPlayers;
	if (uNumPlayersIfAccepted > Asura_Network::GetMaxNumberOfPlayers())
	{
		Asura_GameSpy_GT2_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uReason = ASURA_NDR_SERVERFULL;
		gt2Reject(pxConnection, reinterpret_cast<const u_char*>(&xRejectionReason), sizeof(xRejectionReason));
		Asura_OutputToDebugger_Retail(("connection from %s rejected: too many players.\n", gt2AddressToString(uIP, usPort, NULL)));
		return;
	}

	// check password
	if ( (GetPasswordHash() != 0) && (pxConnectData->m_uPasswordHash != GetPasswordHash()) )
	{
		Asura_GameSpy_GT2_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uReason = ASURA_NDR_PASSWORD;
		gt2Reject(pxConnection, reinterpret_cast<const u_char*>(&xRejectionReason), sizeof(xRejectionReason));
		Asura_OutputToDebugger_Retail(("connection from %s rejected: %s password.\n", gt2AddressToString(uIP, usPort, NULL), pxConnectData->m_uPasswordHash ? "incorrect" : "missing"));
		return;
	}


	GT2ConnectionCallbacks xConnectionCallbacks;
	xConnectionCallbacks.closed = ServerCallback_ClientDisconnected;
	xConnectionCallbacks.connected = 0;
	xConnectionCallbacks.received = MessageReceivedCallback;
	xConnectionCallbacks.ping = PingCallback;

	bool bOK = gt2Accept(pxConnection, &xConnectionCallbacks) != 0;
	if (bOK)
	{
		// connection OK
		ProcessMachineConnection(pxConnection, pxConnectData->m_usNumPlayers, static_cast<u_int>(iLatency));
		Asura_OutputToDebugger_Retail(("connection from %s accepted.\n", gt2AddressToString(uIP, usPort, NULL)));
	}
	else
	{
		// connection timed out, possibly query of CD key took too long?
		Asura_OutputToDebugger_Retail(("connection from %s failed: gt2Accept failed.\n", gt2AddressToString(uIP, usPort, NULL)));
	}
}


void Asura_GameSpy_GT2::ServerCallback_ClientDisconnected( GT2Connection pxConnection,
														   GT2CloseReason eReason )
{
	if (eReason == GT2NotEnoughMemory)
	{
		Asura_OutputToDebugger_Retail(("client kicked due to guaranteed message overflow.\n"));
	}

	Asura_GameSpy_GT2_MachineInfo* pxMachineInfo = GetMachineInfo(pxConnection);

	// use pointer arithmetic to find the machine index.
	u_int uMachineIndex = (pxMachineInfo - s_pxMachineInfo);

	// calling this function to remove the machine will cause the players also
	// to be removed from the system, sending out disconnected messages etc.
	Asura_Network::RemoveMachineFromGame(uMachineIndex);
		
	pxMachineInfo->ClearConnection();
}

//Callback prototype for your progress function
void Asura_GameSpy_GT2::NatNegProgress_ServerCallback(NegotiateState eState, void *pUserdata)
{
#if _DEBUG
	const char* szState = "unknown";
	switch (eState)
	{
	case ns_initsent:
		szState = "ns_initsent";
		break;
	case ns_initack:
		szState = "ns_initack";
		break;
	case ns_connectping:
		szState = "ns_connectping";
		break;
	case ns_finished:
		szState = "ns_finished";
		break;
	case ns_canceled:
		szState = "ns_canceled";
		break;
	}
	Asura_OutputToDebugger(("NatNegProgress_ServerCallback: state = %s.\n", szState));
	//Asura_Logfile(("sblog.txt", "NatNegProgress_ServerCallback: state = %s.\n", szState));
#endif
}

//Callback prototype for your negotiation completed function
void Asura_GameSpy_GT2::NatNegComplete_ServerCallback(NegotiateResult eResult, SOCKET xGameSocket, struct sockaddr_in* pxRemoteAddr, void *pUserdata)
{
#if _DEBUG
	const char* szResult = "unknown";
	switch(eResult) 
	{
	case nr_success:
		szResult = "nr_success";
		break;
	case nr_deadbeatpartner:
		szResult = "nr_deadbeatpartner";
		break;
	case nr_inittimeout:
		szResult = "nr_inittimeout";
		break;
	case nr_unknownerror:
		szResult = "nr_unknownerror";
		break;
	}
	Asura_OutputToDebugger(("NatNegComplete_ServerCallback for %s: result = %s.\n", gt2AddressToString(pxRemoteAddr->sin_addr.s_addr, ntohs(pxRemoteAddr->sin_port), NULL), szResult));
	//Asura_Logfile(("sblog.txt", "NatNegComplete_ServerCallback: result = %s.\n", szResult));
#endif
}


// client functions
void Asura_GameSpy_GT2::ProcessEnumHostsResponse( u_int uIP, u_short usPort, const u_char* pucData, int iDataSize )
{
	const Asura_GameSpy_GT2_LANBroadcastServerReplyHeader* pxHeader = reinterpret_cast
		<const Asura_GameSpy_GT2_LANBroadcastServerReplyHeader*>(pucData);
	const u_char* pucSessionData = reinterpret_cast<const u_char*>(pxHeader + 1);
	Asura_Assert(static_cast<u_int>(iDataSize) == sizeof(Asura_GameSpy_GT2_LANBroadcastServerReplyHeader) + pxHeader->m_usDataSize, "");

	// check project and version
	if (pxHeader->m_uApplicationGuid != Project_GetApplicationGuid()) return;
	if (pxHeader->m_usGameVersion != Asura_GameSpy::GetGameVersion()) return;

	// work out ping based on time since message was sent. Possible problem of
	// overlapping response and request...
	u_int uPing = Asura_Maths::FloatToInt(1000.0f * (Asura_Timers::GetLocalUnscaledTime() - s_fLastSessionRequestBroadcastTime));

	for (u_int u = 0; u < s_uNumberOfActiveSessions; ++u)
	{
		if (s_ppxSessionsStorage[u]->IsSameHost(uIP, usPort))
		{
			s_ppxSessionsStorage[u]->Refresh(uPing, pucSessionData, pxHeader->m_usDataSize, pxHeader->m_wszSessionName, pxHeader->m_ucNumPlayers, pxHeader->m_ucMaxPlayers);
			return;
		}
	}

	// this is a new server
	if (s_uNumberOfActiveSessions < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS)
	{
		s_ppxSessionsStorage[s_uNumberOfActiveSessions++] = new Asura_GameSpy_GT2_SessionInfo(uIP, 
			usPort, uPing, pucSessionData, pxHeader->m_usDataSize, pxHeader->m_wszSessionName, 
			pxHeader->m_ucNumPlayers, pxHeader->m_ucMaxPlayers);
	}
}


void Asura_GameSpy_GT2::ConnectToHost( const char* szFullAddress )
{
	Asura_GameSpy_GT2_ConnectData xUserConnectData;
	xUserConnectData.m_usGameVersion = Asura_GameSpy::GetGameVersion();
	xUserConnectData.m_usNumPlayers = 1;	// TODO - get number of local players from Asura_Network
	xUserConnectData.m_uPasswordHash = s_uPasswordHash;

	GT2ConnectionCallbacks xConnectionCallbacks;
	xConnectionCallbacks.closed = ClientCallback_Disconnected;
	xConnectionCallbacks.connected = ClientCallback_Connected;
	xConnectionCallbacks.received = MessageReceivedCallback;
	xConnectionCallbacks.ping = PingCallback;
	
	GT2Result eResult = gt2Connect( s_pxSocket, 
									NULL,			// ptr to connection handle - only used if blocking.
									szFullAddress, 
									reinterpret_cast<u_char*>(&xUserConnectData), 
									sizeof(xUserConnectData), 
									0,				// timeout in milliseconds, 0 = infinite
									&xConnectionCallbacks,	
									false );		// blocking or non-blocking
	Asura_Assertf(eResult == GT2Success, ("gt2Connect failed, reason = %d.\n", eResult));
	if (eResult != GT2Success)
	{
		Asura_Network::SetWaitingForConnection(false);
		Asura_Network::SetConnectionOkay(false);
	}
}
		
		
void Asura_GameSpy_GT2::SendEnumHostsBroadcast( void )
{
	Asura_Assert(s_pxSocket, "no socket");
	if (!s_pxSocket) return;

	u_int uBroadcastAddr = INADDR_BROADCAST;	// 255.255.255.255
	u_short usPort = static_cast<u_short>(GetPort());
	Asura_GameSpy_GT2_LANBroadcastServerRequest xBroadcastMsg;
	xBroadcastMsg.m_usMagicID = ASURA_GAMESPY_GT2_ENUMHOSTS_MAGICID;
	xBroadcastMsg.m_usGameVersion = Asura_GameSpy::GetGameVersion();
	xBroadcastMsg.m_uApplicationGuid = Project_GetApplicationGuid();
	GT2Bool iResult = gti2SocketSend(s_pxSocket, uBroadcastAddr, usPort, reinterpret_cast<const u_char*>(&xBroadcastMsg), sizeof(xBroadcastMsg));
	Asura_Assert(iResult, "failed to send broadcast message");

	s_fLastSessionRequestBroadcastTime = Asura_Timers::GetLocalUnscaledTime();
}


// client callbacks
void Asura_GameSpy_GT2::ClientCallback_Connected( GT2Connection pxConnection,
												  GT2Result eResult,
												  GT2Byte* pucMessage,
												  int iMessageLen )
{
	Asura_Network::SetWaitingForConnection(false);

	switch(eResult)
	{
		case GT2Success:
		{
			Asura_Network::SetConnectionOkay(true);
			s_pxMachineInfo->SetConnectionHandle(pxConnection);
			break;
		}
		case GT2Rejected:
		{
			if (iMessageLen == sizeof(Asura_GameSpy_GT2_ConnectionRejectedData))
			{
				const Asura_GameSpy_GT2_ConnectionRejectedData* pxResultCode = reinterpret_cast<const Asura_GameSpy_GT2_ConnectionRejectedData*>(pucMessage);
				
				ASURA_NETWORK_DISCONNECT_REASON eAsuraReason = static_cast<enum ASURA_NETWORK_DISCONNECT_REASON>(pxResultCode->m_uReason);
		
				Asura_Network::SetReasonForDisconnection(eAsuraReason);
			}
			else
			{
				Asura_Network::SetReasonForDisconnection(ASURA_NDR_VERSION);
			}
			break;
		}
		case GT2OutOfMemory:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
			SetNetworkError(ASURA_GT2_NETERROR_OUTOFMEMORY);
			break;
		}
		case GT2NetworkError:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
			SetNetworkError(ASURA_GT2_NETERROR_NETWORKERROR);
			break;
		}
		case GT2AddressError:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
			SetNetworkError(ASURA_GT2_NETERROR_ADDRESSERROR);
			break;
		}
		case GT2DuplicateAddress:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
			SetNetworkError(ASURA_GT2_NETERROR_DUPLICATEADDRESS);
			break;
		}
		case GT2TimedOut:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_TIMEOUT);
			break;
		}
		case GT2NegotiationError:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
			SetNetworkError(ASURA_GT2_NETERROR_NEGOTIATIONERROR);
			break;
		}
		default:
		{
			Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
			SetNetworkError(ASURA_GT2_NETERROR_UNKNOWN);
			break;
		}
	}
}


void Asura_GameSpy_GT2::ClientCallback_Disconnected( GT2Connection pxConnection,
													 GT2CloseReason eReason )
{
	//Asura_Logfile(("sblog.txt", "ClientCallback_Disconnected connection = 0x%p, reason = %d.\n", pxConnection, eReason));
//	if (pxConnection)
//	{
//		char szAddress[32];
//		gt2AddressToString(pxConnection->ip, pxConnection->port, szAddress);
//		//Asura_Logfile(("sblog.txt", "connection address = %s.\n", szAddress));
//	}

	if (s_pxMachineInfo->IsValid())
	{
		Asura_Assert(s_pxMachineInfo->GetConnectionHandle() == pxConnection, "unexpected connection ptr");
		s_pxMachineInfo->ClearConnection();
	}

	// if we thought we were connected, set the disconnected reason
	if (Asura_Network::IsConnectionOkay())
	{
		switch (eReason)
		{
			case GT2LocalClose:
			{
				// The connection was closed with gt2CloseConnection. Since that
				// was called on the client, no need to set the reason as it is already known.
				break;
			}

			case GT2RemoteClose:
			{
				// The connection was closed remotely. This may be due to the server quitting,
				// or it could be the server kicking this client due to running out of room for
				// guaranteed messages.
				Asura_Network::SetReasonForDisconnection(ASURA_NDR_SERVERQUIT);
				break;
			}

			case GT2CommunicationError:
			{
				// An invalid message was received (it was either unexpected or wrongly formatted).
				Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
				SetNetworkError(ASURA_GT2_NETERROR_COMMUNICATIONERROR);
				break;
			}

			case GT2SocketError:
			{        
				// An error with the socket forced the connection to close.
				Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
				SetNetworkError(ASURA_GT2_NETERROR_SOCKETERROR);
				break;
			}

			case GT2NotEnoughMemory:
			{
				// There wasn't enough memory to store an incoming or outgoing message.
				Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
				SetNetworkError(ASURA_GT2_NETERROR_TOOMANYDROPPEDPACKETS);
				break;
			}
			
			default:
			{
				Asura_Network::SetReasonForDisconnection(ASURA_NDR_UNKNOWN);
				break;
			}
		}
	}

	Asura_Network::SetConnectionOkay(false);
}

	
// general callbacks
void Asura_GameSpy_GT2::SocketErrorCallback( GT2Socket pxSocket, int iErrorCode )
{
	Asura_Assert(false, "fatal error on network socket");
	Asura_Assert(pxSocket == s_pxSocket, "unexpected socket ptr");
	s_pxSocket = NULL;
	Asura_Network::SetWaitingForConnection(false);
	Asura_Network::SetConnectionOkay(false);

	Asura_Network::SetReasonForDisconnection(ASURA_NDR_NETWORKING_ERROR);
	SetNetworkError(ASURA_GT2_NETERROR_SOCKETERROR);

	if (pxSocket)
	{
		Asura_OutputToDebugger_Retail(("socket error code = %d.\n", iErrorCode));
	}
}


void Asura_GameSpy_GT2::PingCallback( GT2Connection pxConnection, int iLatency )
{
	Asura_GameSpy_GT2_MachineInfo* pxMachineInfo = GetMachineInfo(pxConnection);
	pxMachineInfo->SetPing(static_cast<u_int>(iLatency));
}


void Asura_GameSpy_GT2::MessageReceivedCallback( GT2Connection pxConnection,
												 GT2Byte * pucData,
												 int iDataLen, 
												 GT2Bool bReliable )
{
	// (MH - 27.10.2003) zero-length messages are not normally sent, but my 
	// GT2 hack to enable out-of-order but guaranteed messages sends them.
	if (!iDataLen)
	{
		return;
	}

    u_short usMsgID = *reinterpret_cast<const u_short*>(pucData);
    if ( (usMsgID == ASURA_GAMESPY_GT2_ENUMHOSTS_MAGICID) || 
         (usMsgID == ASURA_GAMESPY_GT2_HOSTREPLY_MAGICID) )
    {
        // if this is an old enum hosts message ignore it, since we have already 
        // established a connection. We could pass it on to be dealt with by the 
        // same code that is called by the unrecognised message callback, but I 
        // don't see any advantage.
        return;
    }

//	const Asura_MessageHeader* pxHeader = reinterpret_cast<const Asura_MessageHeader*>(pucData);
//
//	Asura_Assert(pxHeader->m_uDataSize + sizeof(Asura_MessageHeader) == static_cast<u_int>(iDataLen), "data sizes do not match");
//	
//	// only add messages to queue if application is active, or if delivery was guaranteed
//	bool bWasGuaranteed = bReliable != 0;
//	if( ShouldAddMessageToQueue(pxHeader->m_usMsgID) || bWasGuaranteed)
//	{
//		Asura_Network::AddMessageToQueue(pxHeader, pucData + sizeof(Asura_MessageHeader), bWasGuaranteed);
//	}
	
	Asura_MessageHeader xMsgHeader;
	const void* pData = Asura_MessagePacking::UnpackBufferToMessageHeader(pucData, xMsgHeader);
	// only add messages to queue if application is active, or if delivery was guaranteed
	bool bWasGuaranteed = bReliable != 0;
	if( ShouldAddMessageToQueue(xMsgHeader.m_usMsgID) || bWasGuaranteed)
	{
		Asura_Network::AddMessageToQueue(&xMsgHeader, pData, bWasGuaranteed);
	}
}


GT2Bool Asura_GameSpy_GT2::UnrecognizedMessageCallback(	GT2Socket pxSocket,
														unsigned int uIP,
														unsigned short usPort,
														GT2Byte* pucMessage,
														int iLen )
{
	u_short usMagicID = *reinterpret_cast<u_short*>(pucMessage);
	// see if this is a broadcast request or a response to one.
	if (usMagicID == ASURA_GAMESPY_GT2_ENUMHOSTS_MAGICID)
	{
		if (!s_bServerLanReportingOn)
		{
			return GT2False;
		}
		// process accordingly
		if ( ProcessEnumHostsRequest(uIP, usPort, pucMessage, static_cast<u_int>(iLen)) )
		{
			return GT2True;
		}
	}
	else if (usMagicID == ASURA_GAMESPY_GT2_HOSTREPLY_MAGICID)
	{
		// only process reply if we're ready to receive the data
        if (s_ppxSessionsStorage)
        {
            ProcessEnumHostsResponse(uIP, usPort, pucMessage, iLen);
        }
		return GT2True;
	}
#ifdef ASURA_SUPPORT_GAMESPY
	else if ( (pucMessage[0] == QR_MAGIC_1) && (pucMessage[1] == QR_MAGIC_2) )
	{
		if (Asura_GameSpy_Peer::IsInitialised())
		{
			Asura_GameSpy_Peer::ParsePeerQuery(uIP, usPort, reinterpret_cast<char*>(pucMessage), iLen);
			return GT2True;
		}
		else if (Asura_GameSpy_QR2::IsInitialised())
		{
			// send to QR2 SDK
			sockaddr_in xSockAddress;
			xSockAddress.sin_family = AF_INET;
			xSockAddress.sin_port = htons(usPort);
			xSockAddress.sin_addr.s_addr = uIP;
			qr2_parse_query(NULL, reinterpret_cast<char*>(pucMessage), iLen, reinterpret_cast<sockaddr*>(&xSockAddress));
			return GT2True;
		}
	}
	else if ( (pucMessage[0] == NN_MAGIC_0) && (pucMessage[1] == NN_MAGIC_1) && 
				(pucMessage[2] == NN_MAGIC_2) && (pucMessage[3] == NN_MAGIC_3) && 
				(pucMessage[4] == NN_MAGIC_4) && (pucMessage[5] == NN_MAGIC_5) )
	{
		// send to NatNeg SDK
		sockaddr_in xSockAddress;
		xSockAddress.sin_family = AF_INET;
		xSockAddress.sin_port = htons(usPort);
		xSockAddress.sin_addr.s_addr = uIP;
		NNProcessData(reinterpret_cast<char*>(pucMessage), iLen, &xSockAddress);
		return GT2True;
	}
#endif // #ifdef ASURA_SUPPORT_GAMESPY
	return GT2False;
}


void Asura_GameSpy_GT2::SendDumpCallback( GT2Socket pxSocket,
										  GT2Connection pxConnection,
										  unsigned int uIp,
										  unsigned short usPort,
										  GT2Bool iReset,
										  const GT2Byte * pucMessage,
										  int iLen )
{
	Asura_Assert(iLen > 0 && iLen < 99999, "unexpected length of data sent");
	if (pxConnection && gt2GetConnectionData(pxConnection))
	{
		// (MH - 19.12.2003) I did some testing comparing the totals that come through 
		// this function with the totals reported by the library stats for the network
		// interface. Provided you take extra data that goes through the interface 
		// into account (e.g. other GameSpy SDKs, broadcast packets and replies etc.)
		// you can get the totals to match if you add an extra 14 bytes for the 
		// ethernet header. This header should not be sent over the local->internet
		// part of the connection so I don't think it's fair to include it in
		// the bandwidth stats. Therefore I'm just adding the IP+UDP header size (28).
		u_int uUDPHeaderSize = 28;
		GetMachineInfo(pxConnection)->AddSentBytes(uUDPHeaderSize + static_cast<u_int>(iLen));
	}

	if (dbg_bPrintAllNetworkTraffic)
	{
		Asura_OutputToDebugger(("To %s: %3d bytes: ", gt2AddressToString(uIp, usPort, NULL), iLen));
		int iBytesRemaining = iLen;
		const GT2Byte* pucData = pucMessage;
		u_int u4Bytes = 0;
		while (iBytesRemaining--)
		{
			Asura_OutputToDebugger(("%02x", *pucData++));
			if (++u4Bytes == 4)
			{
				u4Bytes = 0;
				Asura_OutputToDebugger((" "));
			}
		}
		Asura_OutputToDebugger(("\n"));
	}

}

void Asura_GameSpy_GT2::ReceiveDumpCallback( GT2Socket pxSocket,
											 GT2Connection pxConnection,
											 unsigned int uIp,
											 unsigned short usPort,
											 GT2Bool iReset,
											 const GT2Byte * pucMessage,
											 int iLen )
{
	if (pxConnection && gt2GetConnectionData(pxConnection))
	{
		u_int uUDPHeaderSize = 28;
		GetMachineInfo(pxConnection)->AddReceivedBytes(uUDPHeaderSize + static_cast<u_int>(iLen));
	}

	if (dbg_bPrintAllNetworkTraffic)
	{
		Asura_OutputToDebugger(("From %s: %3d bytes: ", gt2AddressToString(uIp, usPort, NULL), iLen));
		int iBytesRemaining = iLen;
		const GT2Byte* pucData = pucMessage;
		u_int u4Bytes = 0;
		while (iBytesRemaining--)
		{
			Asura_OutputToDebugger(("%02x", *pucData++));
			if (++u4Bytes == 4)
			{
				u4Bytes = 0;
				Asura_OutputToDebugger((" "));
			}
		}
		Asura_OutputToDebugger(("\n"));
	}
}


//Callback prototype for your progress function
void Asura_GameSpy_GT2::NatNegProgress_ClientCallback(NegotiateState eState, void *pUserdata)
{
	Asura_Assert(s_eClientConnectionState == CCS_WaitingForNatNeg, "invalid state in NatNegCompleteCallback");

#if _DEBUG
	const char* szState = "unknown";
	switch (eState)
	{
	case ns_initsent:
		szState = "ns_initsent";
		break;
	case ns_initack:
		szState = "ns_initack";
		break;
	case ns_connectping:
		szState = "ns_connectping";
		break;
	case ns_finished:
		szState = "ns_finished";
		break;
	case ns_canceled:
		szState = "ns_canceled";
		break;
	}
	Asura_OutputToDebugger(("NatNegProgress_ClientCallback: state = %s.\n", szState));
	//Asura_Logfile(("sblog.txt", "NatNegProgress_ClientCallback: state = %s.\n", szState));
#endif
}

//Callback prototype for your negotiation completed function
void Asura_GameSpy_GT2::NatNegComplete_ClientCallback(NegotiateResult eResult, SOCKET xGameSocket, struct sockaddr_in* pxRemoteAddr, void *pUserdata)
{
	Asura_Assert(s_eClientConnectionState == CCS_WaitingForNatNeg, "invalid state in NatNegCompleteCallback");

#if _DEBUG
	const char* szResult = "unknown";
	switch(eResult) 
	{
	case nr_success:
		szResult = "nr_success";
		break;
	case nr_deadbeatpartner:
		szResult = "nr_deadbeatpartner";
		break;
	case nr_inittimeout:
		szResult = "nr_inittimeout";
		break;
	case nr_unknownerror:
		szResult = "nr_unknownerror";
		break;
	}
	Asura_OutputToDebugger(("NatNegComplete_ClientCallback: result = %s.\n", szResult));
	//Asura_Logfile(("sblog.txt", "NatNegComplete_ClientCallback: result = %s.\n", szResult));
#endif

	if (eResult == nr_success)
	{
		s_eClientConnectionState = CCS_ReadyToConnect;
		u_int uIP = pxRemoteAddr->sin_addr.s_addr;
		u_short usPort = htons(pxRemoteAddr->sin_port);
		// set up the specified addresses for the connection
		gt2AddressToString(uIP, usPort, s_acServerAddress);
		SetPort(usPort);
	}
	else
	{
		s_eClientConnectionState = CCS_NotConnected;
	}
}


// general utility functions
Asura_GameSpy_GT2_MachineInfo* Asura_GameSpy_GT2::GetMachineInfo( GT2Connection pxConnection )
{
	Asura_GameSpy_GT2_MachineInfo* pxData = reinterpret_cast
		<Asura_GameSpy_GT2_MachineInfo*>(gt2GetConnectionData(pxConnection));
	Asura_Assert(pxData, "no machine data for connection");
	Asura_Assert(pxData->GetConnectionHandle() == pxConnection, "machine/connection mismatch");
	return pxData;
}


Asura_GameSpy_GT2_MachineInfo* Asura_GameSpy_GT2::GetMachineInfoFromIndex( u_int uMachine )
{
	Asura_Assert(uMachine < s_uMaxNumberOfMachines, "invalid machine handle");
	return &s_pxMachineInfo[uMachine];
}


void Asura_GameSpy_GT2::SendAsuraMessage( GT2Connection pxConnection, const Asura_Message* pxMessage )
{
	u_int uTotalSize = Asura_MessagePacking::PackMessageToBuffer( g_aucBuffer, ASURA_GAMESPY_GT2_BUFFERSIZE, *pxMessage );
	Asura_Assert(uTotalSize, "GT2 buffer not big enough");
	if (!uTotalSize) return;

	gt2Send(pxConnection, g_aucBuffer, uTotalSize, pxMessage->IsUsingGuaranteedDelivery());

//	Asura_MessageHeader* pxHeader = (Asura_MessageHeader*)g_aucBuffer;
//	
//	*pxHeader = *pxMessage;
//
//	unsigned int uHeaderSize = sizeof(Asura_MessageHeader);
//	unsigned int uTotalSize = uHeaderSize + pxHeader->m_uDataSize;
//
//	Asura_Assert(uTotalSize < ASURA_GAMESPY_GT2_BUFFERSIZE, "Packet buffer size too small");
//
//	memcpy(&g_aucBuffer[uHeaderSize], pxMessage->GetData(), pxHeader->m_uDataSize);
//
//	gt2Send(pxConnection, g_aucBuffer, uTotalSize, pxMessage->IsUsingGuaranteedDelivery());
}


bool Asura_GameSpy_GT2::ShouldAddMessageToQueue(u_int uMsgID)
{
	//if ready for messages , then add all messages we get
	if(Asura_Network::IsReadyForMessages()) return true;

	//otherwise just add a few special messages
	switch(uMsgID)
	{
		case ASURA_MSG_VALIDCLIENTGUID :
			return true;

		default :
			break;
	}
	return false;
}


void Asura_GameSpy_GT2::UpdateBandwidthStats( void )
{
	s_fTimeSinceLastBandwidthCalculation += Asura_Timers::GetFrameTime();
	if (s_fTimeSinceLastBandwidthCalculation > dbg_fTimeBetweenBandwidthCalculations)
	{
		u_int uSentCount = 0;
		u_int uReceivedCount = 0;

		for (u_int u = 0; u < s_uMaxNumberOfMachines; ++u)
		{
			Asura_GameSpy_GT2_MachineInfo& xMachineInfo = s_pxMachineInfo[u];
			if (xMachineInfo.IsValid())
			{
				uSentCount += xMachineInfo.GetBytesSent();
				uReceivedCount += xMachineInfo.GetBytesReceived();
			}
		}

		float fSentThisPeriod = static_cast<float>(uSentCount);
		float fReceivedThisPeriod = static_cast<float>(uReceivedCount);

		s_uSendBPS = Asura_Maths::FloatToInt(fSentThisPeriod / s_fTimeSinceLastBandwidthCalculation);
		s_uReceiveBPS = Asura_Maths::FloatToInt(fReceivedThisPeriod / s_fTimeSinceLastBandwidthCalculation);

		if (s_uSendBPS > s_uPeakSendBPS)
		{
			s_uPeakSendBPS = s_uSendBPS;
		}
		if (s_uReceiveBPS > s_uPeakReceiveBPS)
		{
			s_uPeakReceiveBPS = s_uReceiveBPS;
		}
		
		s_fTimeSinceLastBandwidthCalculation = 0.0f;
	}
}


Asura_GameSpy_GT2_SessionInfo::Asura_GameSpy_GT2_SessionInfo( u_int uHostAddr, 
															 u_short usHostPort, 
															 u_int uPing, 
															 const u_char* pcSessionData, 
															 u_int uSessionDataSize, 
															 const Asura_UnicodeChar* pszSessionName,
															 u_char ucNumPlayers,
															 u_char ucMaxPlayers)
:m_uHostAddr(uHostAddr)
,m_usHostPort(usHostPort)
,m_fLastUpdateTime(Asura_Timers::GetLocalUnscaledTime())
,m_uPing(uPing)
,m_uSessionDataSize(uSessionDataSize)
,m_ucNumPlayers(ucNumPlayers)
,m_ucMaxPlayers(ucMaxPlayers)
{
	m_pucSessionData = new u_char[uSessionDataSize];
	memcpy(m_pucSessionData, pcSessionData, uSessionDataSize);

	Asura_PrintfToBfr(m_wszSessionName, L"%s", uASURA_NETWORK_MAX_SESSION_NAME + 1, true, pszSessionName);
}

Asura_GameSpy_GT2_SessionInfo::~Asura_GameSpy_GT2_SessionInfo()
{
	delete[] m_pucSessionData;
}

void Asura_GameSpy_GT2_SessionInfo::Refresh(u_int uPing, 
											const u_char* pcSessionData, 
											u_int uSessionDataSize, 
											const Asura_UnicodeChar* pszSessionName,
											u_char ucNumPlayers,
											u_char ucMaxPlayers)
{
	Asura_Assert(uSessionDataSize == m_uSessionDataSize, "session data size changed!");
	if (uSessionDataSize != m_uSessionDataSize)
	{
		return;
	}

	m_fLastUpdateTime = Asura_Timers::GetLocalUnscaledTime();
	m_uPing = uPing;

	memcpy(m_pucSessionData, pcSessionData, uSessionDataSize);

	Asura_PrintfToBfr(m_wszSessionName, L"%s", uASURA_NETWORK_MAX_SESSION_NAME + 1, true, pszSessionName);

	m_ucNumPlayers = ucNumPlayers;
	m_ucMaxPlayers = ucMaxPlayers;
}

#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
