//////////////////////////////////////////////////////////////////////////////
//
// Asura_GameSpy_Peer.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     Asura wrapper for the Peer2 SDK
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_GAMESPY

#include "Network/GameSpy/Asura_GameSpy_Peer.h"

#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/GameSpy/Asura_GameSpy.h"
#include "Network/Asura_NetTransport_System.h"
#include "Network/Asura_NetSocket.h"
#include "Network/GameSpy/Asura_GameSpy_QR2.h"	// for custom keys
#include "Network/GameSpy/Asura_GameSpy_ServerBrowsing.h"	// to init server listing
#include "Network/Asura_LobbyChat.h"
#include "Network/Asura_Network.h"		// for ASURA_NETWORK_MAXHOSTNAMECHARS

#include "GameSpy/gp/gp.h"				// for nick and password lengths
#include "GameSpy/peer/peer.h"

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uASURA_GAMESPY_PEER_MAXGROUPROOMS = 16;
static const u_int uASURA_GAMESPY_PEER_MAXSERVERS = 256;

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
PEER Asura_GameSpy_Peer::s_pxPeer = NULL;
Asura_GameSpy_Peer::ConnectionState Asura_GameSpy_Peer::s_eConnectionState = Asura_GameSpy_Peer::PCN_None;
u_int Asura_GameSpy_Peer::s_uNicknameModifier = 1;
char Asura_GameSpy_Peer::s_acNickNameExtension[uASURA_GAMESPY_PEER_NICKNAME_EXT_MAXLEN];
RoomType Asura_GameSpy_Peer::s_eRoomType = NumRooms;	// impossible value
Asura_GameSpy_Peer_GroupRoom* Asura_GameSpy_Peer::s_pxGroupsRoomArray = NULL;
u_int Asura_GameSpy_Peer::s_uNumGroupRooms = 0;
u_int Asura_GameSpy_Peer::s_uGroupRoomIndex;
SBServer Asura_GameSpy_Peer::s_pxStagingServer;
bool Asura_GameSpy_Peer::s_bErrorAddingServer;
PEERJoinResult Asura_GameSpy_Peer::s_eLastRoomJoinResult;
char Asura_GameSpy_Peer::s_szAttemptedNick[uASURA_CHAT_MAX_USERNAME_LEN];

// L O C A L   F U N C T I O N S /////////////////////////////////////////////

static const char* TranslatePeerNick( const char* szNick, const char* szExtension )
{
	// (MH - 24.01.2005) strip unique nickname identifier. amazingly, there is no function
	// to do this from a const char* without first copying the nick.
	static char acTranslatedNick[uASURA_CHAT_MAX_USERNAME_LEN + 1];
	Asura_PrintfToBfr( acTranslatedNick, "%s", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, szNick );
	const char* szTranslatedNick = peerTranslateNick(acTranslatedNick, szExtension);
	if (!szTranslatedNick)
	{
		szTranslatedNick = szNick;
	}

	return szTranslatedNick;
}

static void PrepareBufferForNickFix( const Asura_UnicodeChar* wszInputNick, char* szOutputNick )
{
    // (MP - 09.03.2006) Previously we were converting nicknames to UTF8
    // before passing to peerFixNick(). This really makes no sense since that
    // function isn't expecting it internally. Instead, let's just limit the
    // characters we pass in to the non-extended (7-bit) ASCII range, since
    // peerFixNick()'s allowed list is just a subset of that.

    Asura_UnicodeChar wNext;
    do
    {
        wNext = *wszInputNick++;
        *szOutputNick++ = ( wNext < 128 ? static_cast<char>(wNext) : '_' );
    }
    while( wNext );
}

// S T A T I C   C L A S S   F U N C T I O N S ///////////////////////////////

bool Asura_GameSpy_Peer::Initialise( bool (*pfnAddUser)(const Asura_UnicodeChar*, bool*, int*) /* = 0 */ )
{
	// init auxiliary systems
	Asura_LobbyChat::Initialise(pfnAddUser);
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_GameSpy_ServerBrowsing::Initialise(uASURA_GAMESPY_PEER_MAXSERVERS, true);
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY

	Asura_Assert(!s_pxPeer, "Peer already Initialised!");
	Asura_Assert(s_eConnectionState == PCN_None, "peer connection wasn't closed?");
	
	PEERCallbacks xCallbacks;
	memset(&xCallbacks, 0, sizeof(PEERCallbacks));
	xCallbacks.disconnected = DisconnectedCallback;
	xCallbacks.roomMessage = RoomMessageCallback;
	xCallbacks.roomUTM = RoomUTMCallback;
	xCallbacks.roomNameChanged = RoomNameChangedCallback;
	xCallbacks.roomModeChanged = RoomModeChangedCallback;
	xCallbacks.playerMessage = PlayerMessageCallback;
	xCallbacks.playerUTM = PlayerUTMCallback;
	xCallbacks.readyChanged = ReadyChangedCallback;
	xCallbacks.gameStarted = GameStartedCallback;
	xCallbacks.playerJoined = PlayerJoinedCallback;
	xCallbacks.playerLeft = PlayerLeftCallback;
	xCallbacks.kicked = KickedCallback;
	xCallbacks.newPlayerList = NewPlayerListCallback;
	xCallbacks.playerChangedNick = PlayerChangedNickCallback;
	xCallbacks.playerInfo = PlayerInfoCallback;
	xCallbacks.playerFlagsChanged = PlayerFlagsChangedCallback;
	xCallbacks.ping = PingCallback;
	xCallbacks.crossPing = CrossPingCallback;
	xCallbacks.globalKeyChanged = GlobalKeyChangedCallback;
	xCallbacks.roomKeyChanged = RoomKeyChangedCallback;
	xCallbacks.qrServerKey = QRServerKeyCallback;
	xCallbacks.qrPlayerKey = QRPlayerKeyCallback;
	xCallbacks.qrTeamKey = QRTeamKeyCallback;
	xCallbacks.qrKeyList = QRKeyListCallback;
	xCallbacks.qrCount = QRCountCallback;
	xCallbacks.qrAddError = QRAddErrorCallback;
	xCallbacks.qrNatNegotiateCallback = QRNatNegotiateCallback;
	xCallbacks.qrPublicAddressCallback = QRPublicAddressCallback;
	xCallbacks.param = NULL;
	
	s_pxPeer = peerInitialize(&xCallbacks);
	Asura_Assert(s_pxPeer, "peerInitialise failed");

	bool bOK = (s_pxPeer != NULL);
	if (bOK)
	{
		int iMaxUpdates = 20;	// docs say 20-30 for broadband
		PEERBool eNatNegotiate = PEERTrue;
		// note that for GameSpy arcade compatability, all rooms must be ping rooms
		// and staging rooms only should be crossping rooms.
		PEERBool aePingRooms[NumRooms] = 
		{
			PEERTrue, // TitleRoom,  // The main room for a game.
			PEERTrue, // GroupRoom,  // A room which is, in general, for a particular type of gameplay (team, dm, etc.).
			PEERTrue, // StagingRoom,  // A room where players meet before starting a game.
		};
		PEERBool aeCrosspingRooms[NumRooms] = 
		{
			PEERFalse, // TitleRoom,  // The main room for a game.
			PEERFalse, // GroupRoom,  // A room which is, in general, for a particular type of gameplay (team, dm, etc.).
			PEERTrue, // StagingRoom,  // A room where players meet before starting a game.
		};
		PEERBool eOK = peerSetTitle(s_pxPeer, Asura_GameSpy::GetGameName(), Asura_GameSpy::GetGameKey(), Asura_GameSpy::GetGameName(), Asura_GameSpy::GetGameKey(), Asura_GameSpy::GetGameVersion(), iMaxUpdates, eNatNegotiate, aePingRooms, aeCrosspingRooms);
		Asura_Assert(eOK == PEERTrue, "peerSetTitle failed");	// should only happen if out of memory.
		bOK = (eOK == PEERTrue);
	}

	// register all custom keys
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_GameSpy_QR2::RegisterCustomKeys();
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	
	// restart the nickname modifier
	s_uNicknameModifier = 1;
	s_pxGroupsRoomArray = new Asura_GameSpy_Peer_GroupRoom[uASURA_GAMESPY_PEER_MAXGROUPROOMS];
	s_uNumGroupRooms = 0;

	s_pxStagingServer = NULL;

	s_eLastRoomJoinResult = PEERJoinSuccess;	// default to success
	s_szAttemptedNick[0] = 0;

	
	return bOK;
}

bool Asura_GameSpy_Peer::IsInitialised( void )
{
    return( NULL != s_pxPeer );
}

Asura_GameSpy_Peer::ConnectionState Asura_GameSpy_Peer::GetConnectionState()
{
	return s_eConnectionState;
}

void Asura_GameSpy_Peer::Update( void )
{
	if (s_pxPeer)
	{
		if(!Asura_Network::IsNetworkInitialised())
		{
			Asura_NetTransport_System::UpdateSocket();
		}

		peerThink(s_pxPeer);

		// update connection here
		if (s_eConnectionState == PCN_ConnectedToServer)
		{
			// get list of groups rooms
			PEERBool eBlocking = PEERFalse;
			char acFilterList[128];
			Asura_PrintfToBfr(acFilterList, "\\%s", 128, true, Project_GetGroupRoomFieldsString());
			peerListGroupRooms( s_pxPeer, acFilterList, ListGroupRoomsCallback, NULL, eBlocking );
			s_eConnectionState = PCN_GettingListOfGroupRooms;
		}
	}
}

void Asura_GameSpy_Peer::Shutdown( void )
{
	if (s_pxPeer)
	{
		peerShutdown(s_pxPeer);
		s_pxPeer = NULL;
	}

	// SNH - 18/10/2004 11:17:34 - if we've requested a socket from ANTS, it won't be removed
	//	when ANTS is shutdown (to allow us to rejoin a staging room when the game ends). 
	//	 Therefore we need to manually close the socket now.
	if(Asura_NetTransport_System::NeedToCloseSocketManually())
	{
		Asura_NetTransport_System::CloseSocket();
	}
	
	delete[] s_pxGroupsRoomArray;
	s_pxGroupsRoomArray = NULL;
	s_uNumGroupRooms = 0;

	s_pxStagingServer = NULL;

	// shutdown serverbrowsing in case we opened it
	Asura_GameSpy_ServerBrowsing::Shutdown();
}

void Asura_GameSpy_Peer::ConnectToChatServer(const Asura_UnicodeChar* wszPlayerName)
{
	if (s_pxPeer)
	{
		// first create a UTF8 version of the player name, truncating to uASURA_CHAT_MAX_USERNAME_LEN (31) chars plus NULL.
		Asura_UnicodeChar wszTruncatedNickname[uASURA_CHAT_MAX_USERNAME_LEN + 1];
		char szNickname[uASURA_CHAT_MAX_USERNAME_LEN + 1];
		char szNewNickname[uASURA_CHAT_MAX_USERNAME_LEN + 2];
		Asura_PrintfToBfr(wszTruncatedNickname, L"%s", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, wszPlayerName);
        PrepareBufferForNickFix( wszTruncatedNickname, szNickname );

		// fix up any illegal characters
		peerFixNick(reinterpret_cast<char*>(szNewNickname), reinterpret_cast<char*>(szNickname));
		
		int iGameSpyProfileID = 0;	// can be ignored if we're not logging on using gamespy profiles
		PEERBool eBlocking = PEERFalse;
		peerConnect(s_pxPeer, szNewNickname, iGameSpyProfileID, NickErrorCallback, ConnectCallback, NULL, eBlocking);

		// save nick
		Asura_PrintfToBfr( s_szAttemptedNick, "%s", uASURA_CHAT_MAX_USERNAME_LEN, true, szNewNickname );

		s_eConnectionState = PCN_ConnectingToServer;
	}
}

// CJM 13/07/06
// This version connects to GS with a unique nickname and password. The GS docs indicate that this should only be used with a custom GS namespace
//  as the unique nick can expire.
void Asura_GameSpy_Peer::ConnectWithGPLogin(const int iNameSpaceID, 
											const char* szNickNameExtension,
											const Asura_UnicodeChar* wszUniqueNickName, 
											const Asura_UnicodeChar* wszPassword)
{
	Asura_Assert(s_pxPeer, "no peer object");
	if (s_pxPeer)
	{
		// first create a UTF8 version of the player name, truncating to uASURA_CHAT_MAX_USERNAME_LEN (31) chars plus NULL.
		Asura_UnicodeChar wszTruncatedNickname[GP_UNIQUENICK_LEN + 1];
		char szNickname[GP_UNIQUENICK_LEN + 1];

		Asura_PrintfToBfr(wszTruncatedNickname, L"%s", GP_UNIQUENICK_LEN + 1, true, wszUniqueNickName);
		PrepareBufferForNickFix( wszTruncatedNickname, szNickname );

		// Deliberately don't fix up the nick here.

		//convert password
		char szPassword[GP_PASSWORD_LEN + 1];
		UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszPassword), szPassword);

		//copy extension
		Asura_Assert(strlen(szNickNameExtension) < uASURA_GAMESPY_PEER_NICKNAME_EXT_MAXLEN, "nickname extension too long");
		Asura_PrintfToBfr(s_acNickNameExtension, "%s", uASURA_GAMESPY_PEER_NICKNAME_EXT_MAXLEN, true, szNickNameExtension);
		
		PEERBool eBlocking = PEERFalse;
		peerConnectLogin(s_pxPeer, iNameSpaceID, NULL, NULL, szNickname, szPassword, NickErrorCallback, ConnectCallback, NULL, eBlocking);
		
		// save nick
		Asura_PrintfToBfr( s_szAttemptedNick, "%s", uASURA_CHAT_MAX_USERNAME_LEN, true, szNickname );

		s_eConnectionState = PCN_ConnectingToServer;
	}
}

// CJM 13/07/06
// This version connects to GS with a nickname, email and password.
void Asura_GameSpy_Peer::ConnectWithGPLogin(const int iNameSpaceID, 
											const char* szNickNameExtension,
											const Asura_UnicodeChar* wszNickName, 
											const Asura_UnicodeChar* wszEmail, 
											const Asura_UnicodeChar* wszPassword)
{
	Asura_Assert(s_pxPeer, "no peer object");
	if (s_pxPeer)
	{
		// first create a UTF8 version of the player name, truncating to uASURA_CHAT_MAX_USERNAME_LEN (31) chars plus NULL.
		Asura_UnicodeChar wszTruncatedNickname[GP_NICK_LEN + 1];
		char szNickname[GP_NICK_LEN + 1];

		Asura_PrintfToBfr(wszTruncatedNickname, L"%s", GP_NICK_LEN + 1, true, wszNickName);
		PrepareBufferForNickFix( wszTruncatedNickname, szNickname );

		// Deliberately don't fix up the nick here.

		//convert password
		char szPassword[GP_PASSWORD_LEN + 1];
		UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszPassword), szPassword);

		//convert email
		char szEmail[GP_EMAIL_LEN + 1];
		UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszEmail), szEmail);

		//copy extension
		Asura_Assert(strlen(szNickNameExtension) < uASURA_GAMESPY_PEER_NICKNAME_EXT_MAXLEN, "nickname extension too long");
		Asura_PrintfToBfr(s_acNickNameExtension, "%s", uASURA_GAMESPY_PEER_NICKNAME_EXT_MAXLEN, true, szNickNameExtension);
		
		PEERBool eBlocking = PEERFalse;
		peerConnectLogin(s_pxPeer, iNameSpaceID, szEmail, szNickname, NULL, szPassword, NickErrorCallback, ConnectCallback, NULL, eBlocking);
		
		// save nick
		Asura_PrintfToBfr( s_szAttemptedNick, "%s", uASURA_CHAT_MAX_USERNAME_LEN, true, szNickname );

		s_eConnectionState = PCN_ConnectingToServer;
	}
}

void Asura_GameSpy_Peer::DisconnectFromChatServer( void )
{
	if (s_pxPeer)
	{
		peerDisconnect(s_pxPeer);
	}
	s_eConnectionState = PCN_None;
	s_eRoomType = NumRooms;

	s_szAttemptedNick[0] = 0;

	// shutdown auxiliary systems
	Asura_LobbyChat::Shutdown();
}


void Asura_GameSpy_Peer::SetQuietMode(bool bQuietMode)
{
	if (s_pxPeer)
	{
		peerSetQuietMode(s_pxPeer, (bQuietMode?PEERTrue:PEERFalse));
	}
}


bool Asura_GameSpy_Peer::IsConnected()
{
	bool connected = false;

	if (s_pxPeer)
	{
		connected = (PEERTrue == peerIsConnected(s_pxPeer));
	}
//	else
//	{
//		return connected;
//	}

	return connected;
}


bool Asura_GameSpy_Peer::SendMessage(const Asura_UnicodeChar* wszMessage)
{
	if (s_pxPeer && (s_eRoomType != NumRooms))
	{
		// TODO - use unicode version instead of converting to ASCII
		// SNH - 30/06/2004 12:00:17 - use UTF8 rather than plain ascii.
		// Comments in UCS2ToUTF8String say 3 characters may be needed for each unicode char.
		char acMessage[3 * uASURA_CHAT_MAX_MESSAGE_SIZE];
		//Asura_PrintfToBfr(acMessage, "%S", uASURA_CHAT_MAX_MESSAGE_SIZE, true, wszMessage);
		
		UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszMessage), acMessage);
		
		peerMessageRoom(s_pxPeer, s_eRoomType, acMessage, NormalMessage);
		return true;
	}
	return false;
}

void Asura_GameSpy_Peer::SetReadyState(bool bReady)
{
	if(s_pxPeer && s_eRoomType == StagingRoom)
	{
		if(bReady)
		{
			peerSetReady(s_pxPeer, PEERTrue);
		}
		else
		{
			peerSetReady(s_pxPeer, PEERFalse);
		}
	}
}

bool Asura_GameSpy_Peer::GetReadyState(const Asura_UnicodeChar* wszPlayerName)
{
	if(s_pxPeer && s_eRoomType == StagingRoom)
	{
		// first create a UTF8 version of the player name, truncating to uASURA_CHAT_MAX_USERNAME_LEN (31) chars plus NULL.
		Asura_UnicodeChar wszTruncatedNickname[uASURA_CHAT_MAX_USERNAME_LEN + 1];
		char szNickname[uASURA_CHAT_MAX_USERNAME_LEN + 1];
		char szNewNickname[uASURA_CHAT_MAX_USERNAME_LEN + 2];
		Asura_PrintfToBfr(wszTruncatedNickname, L"%s", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, wszPlayerName);
		PrepareBufferForNickFix( wszTruncatedNickname, szNickname );

		// fix up any illegal characters
		peerFixNick(reinterpret_cast<char*>(szNewNickname), reinterpret_cast<char*>(szNickname));

		PEERBool eReady = PEERFalse;
		peerGetReady(s_pxPeer, szNickname, &eReady);

		if(eReady == PEERTrue)
		{
			return true;
		}
	}

	return false;
}

bool Asura_GameSpy_Peer::GetNick(Asura_UnicodeChar* wszNameBuffer, u_int uBufferSize)
{
	if(s_pxPeer)
	{
		// write UTF8 version of the nickname into unicode buffer
		Asura_UnicodeChar wszNickName[uASURA_CHAT_MAX_USERNAME_LEN + 1];
		UTF8ToUCS2String(peerGetNick(s_pxPeer), reinterpret_cast<u_short*>(wszNickName));
	
		// now put this into the user buffer
		Asura_PrintfToBfr(wszNameBuffer, L"%ls", Asura_Maths::Min(uBufferSize, uASURA_CHAT_MAX_USERNAME_LEN + 1), true, wszNickName);
		return true;
	}

	return false;
}

u_int Asura_GameSpy_Peer::GetNumPlayersInGroupRoom( u_int uGroupRoomIndex )
{
	if (uGroupRoomIndex < s_uNumGroupRooms)
	{
		return s_pxGroupsRoomArray[uGroupRoomIndex].m_uNumPlayersInRoom;
	}
	Asura_Assert(false, "invalid group room index");
	return 0;
}

u_int Asura_GameSpy_Peer::GetMaxPlayersInGroupRoom( u_int uGroupRoomIndex )
{
	if (uGroupRoomIndex < s_uNumGroupRooms)
	{
		return s_pxGroupsRoomArray[uGroupRoomIndex].m_uMaxPlayersInRoom;
	}
	Asura_Assert(false, "invalid group room index");
	return 0;
}

u_int Asura_GameSpy_Peer::GetNumGamesInGroupRoom( u_int uGroupRoomIndex )
{
	if(uGroupRoomIndex < s_uNumGroupRooms)
	{
		return s_pxGroupsRoomArray[uGroupRoomIndex].m_uNumGamesInRoom;
	}
	Asura_Assert(false, "invalid group room index");
	return 0;
}

u_int Asura_GameSpy_Peer::GetNumPlayingInGroupRoom( u_int uGroupRoomIndex )
{
	if(uGroupRoomIndex < s_uNumGroupRooms)
	{
		return s_pxGroupsRoomArray[uGroupRoomIndex].m_uNumPlayingInRoom;
	}
	Asura_Assert(false, "invalid group room index");
	return 0;
}

u_int Asura_GameSpy_Peer::GetFilterFlags( u_int uGroupRoomIndex )
{
	if(uGroupRoomIndex < s_uNumGroupRooms)
	{
		return s_pxGroupsRoomArray[uGroupRoomIndex].m_uFilterFlags;
	}
	Asura_Assert(false, "invalid group room index");
	return 0xffffffff;
}

Asura_UnicodeChar* Asura_GameSpy_Peer::GetGroupRoomName( u_int uGroupRoomIndex )
{
	if (uGroupRoomIndex < s_uNumGroupRooms)
	{
		return s_pxGroupsRoomArray[uGroupRoomIndex].m_awcRoomName;
	}
	Asura_Assert(false, "invalid group room index");
	return L"";
}

void Asura_GameSpy_Peer::JoinGroupRoom( u_int uGroupRoomIndex )
{
	Asura_Assert(s_pxPeer, "not initialised");
	Asura_Assert(s_eConnectionState == PCN_WaitingToJoinGroupRoom, "unexpected state");
	if (s_pxPeer && (s_eConnectionState == PCN_WaitingToJoinGroupRoom))
	{
		if (uGroupRoomIndex < s_uNumGroupRooms)
		{
			PEERBool eBlocking = PEERFalse;
			int iGroupID = s_pxGroupsRoomArray[uGroupRoomIndex].m_iGroupID;
			s_eConnectionState = PCN_JoiningGroupRoom;
			peerJoinGroupRoom(s_pxPeer, iGroupID, JoinRoomCallback, NULL, eBlocking);

			s_uGroupRoomIndex = uGroupRoomIndex;
		}
		else
		{
			Asura_Assert(false, "invalid group room index");
		}
	}
}

const char* Asura_GameSpy_Peer::GetGroupRoomCustomParam( u_int uGroupRoomIndex, const char* szParam)
{
	if (s_pxGroupsRoomArray[uGroupRoomIndex].m_pxServer)
	{
		return SBServerGetStringValue(s_pxGroupsRoomArray[uGroupRoomIndex].m_pxServer, szParam, "");
	}
	return "";
}

PEERJoinResult Asura_GameSpy_Peer::GetLastRoomJoinResult()
{
	Asura_Assert(s_pxPeer, "not initialised");
	return s_eLastRoomJoinResult;
}

//////////////////////////////////////////////////////////////////////////
//
// Interface for group and staging room operations
//
//////////////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::CreateGame( const Asura_UnicodeChar* wszRoomName, 
									 int iMaxPlayers,
									 const Asura_UnicodeChar* wszPassword /* = NULL */ )
{
	Asura_Assert(s_pxPeer, "not initialised");
	Asura_Assert(s_eConnectionState == PCN_InGroupRoom, "unexpected state");
	if (s_pxPeer && (s_eConnectionState == PCN_InGroupRoom))
	{
		// reset add server error flag
		s_bErrorAddingServer = false;

		// TODO - use Unicode versions (note use of capital %S to convert)
		// SNH - 30/06/2004 12:00:17 - use UTF8 rather than plain ascii.
		// Comments in UCS2ToUTF8String say 3 characters may be needed for each unicode char.
		char szRoomName[3 * ASURA_NETWORK_MAXHOSTNAMECHARS + 1];
		char szPassword[3 * PEER_PASSWORD_LEN];
	//	Asura_PrintfToBfr(szRoomName, "%S", ASURA_NETWORK_MAXHOSTNAMECHARS + 1, true, wszRoomName);
	//	Asura_PrintfToBfr(szPassword, "%S", PEER_PASSWORD_LEN, true, wszPassword);
		UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszRoomName), szRoomName);
		UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszPassword), szPassword);

		PEERBool eBlocking = PEERFalse;

		// get socket and port from ANTS
		SOCKET xSocket = 0;
		Asura_NetSocket* pxAsuraSocket = Asura_NetTransport_System::GetSocket();
		if(pxAsuraSocket)
		{
			xSocket = pxAsuraSocket->GetSOCKET();
		}
		else
		{
			Asura_Assert(false, "Can't get socket from ANTS");
		}
		u_short usPort = static_cast<u_short>(Asura_NetTransport_System::GetPort());

		peerCreateStagingRoomWithSocket(s_pxPeer, szRoomName, iMaxPlayers, szPassword, xSocket, usPort, JoinRoomCallback, NULL, eBlocking);
//		peerCreateStagingRoom(s_pxPeer, szRoomName, iMaxPlayers, szPassword, JoinRoomCallback, NULL, eBlocking);

		s_eConnectionState = PCN_CreatingStagingRoom;

		Asura_GameSpy_ServerBrowsing::Reset();
		
		// set the groupID reporting in qr2
		int iGroupID = peerGetGroupID(s_pxPeer);
		Asura_GameSpy_QR2::SetGroupID(iGroupID);
	}
}


bool Asura_GameSpy_Peer::StartGame( bool bKeepReporting /* = false */)
{
	Asura_Assert(s_pxPeer, "not initialised");
	Asura_Assert(s_eConnectionState == PCN_HostingStagingRoom, "unexpected state");
	if (s_pxPeer && (s_eConnectionState == PCN_HostingStagingRoom))
	{
		// for gamespy arcade compatability acMessage should be of the form "<ip>[:<port>]"
		// if port is not specified, default port for the game is assumed.
		char acMessage[32];
		u_int uIP = peerGetPublicIP(s_pxPeer);
		Asura_PrintfToBfr(acMessage, "%d.%d.%d.%d", 32, true, ((uIP & 0x000000ff) >> 0), ((uIP & 0x0000ff00) >> 8), ((uIP & 0x00ff0000) >> 16), ((uIP & 0xff000000) >> 24));

		// we call this with PEER_STOP_REPORTING as we will get the QR2 SDK to take over reporting.
		// SNH - 14/10/2004 17:19:40 - added an option for this, since Rogue uses peer in-game too.
		//	The default is to stop reporting.
		int iReportingOptions = PEER_STOP_REPORTING;
		if(bKeepReporting)
		{
			// get peer to continue automatically reporting on game info and players.
			//	Custom / Team / time+frag limit / scores still need reporting.
			iReportingOptions = PEER_KEEP_REPORTING | PEER_REPORT_INFO | PEER_REPORT_PLAYERS;
		}
		peerStartGame(s_pxPeer, acMessage, iReportingOptions);
		return true;
	}
	return false;
}

bool Asura_GameSpy_Peer::StopGame( void )
{
	Asura_Assert(s_pxPeer, "not initialised");
	if(s_pxPeer)
	{
		peerStopGame(s_pxPeer);

		// when returning from game, need to reset this so we don't immediately
		//	try to rejoin. Should still be in the same staging room as before.
		if(s_eConnectionState == PCN_ClientGameStarting)
		{
			s_eConnectionState = PCN_ClientInStagingRoom;
		}

		return true;
	}

	return false;
}

void Asura_GameSpy_Peer::Client_StartPlaying()
{
	Asura_Assert(s_pxPeer, "not initialised");
	if(s_pxPeer)
	{
		peerStartPlaying(s_pxPeer);

		s_eConnectionState = PCN_ClientGameStarting;
		s_eRoomType = StagingRoom;
	}
}

#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
void Asura_GameSpy_Peer::BeginListingServers( void )
{
	RefreshServerList();
}

bool Asura_GameSpy_Peer::StopListingServers( void )
{
	Asura_Assert(s_pxPeer, "not initialised");
	if(s_pxPeer)
	{
		peerStopListingGames(s_pxPeer);
		return true;
	}

	return false;
}

void Asura_GameSpy_Peer::Refresh( void )
{
	Asura_GameSpy_ServerBrowsing::Reset();
	RefreshServerList();
}

bool Asura_GameSpy_Peer::JoinStagingRoom( u_int uServerIndex, const Asura_UnicodeChar* wszPassword /* = NULL */ )
{
	Asura_Assert(s_pxPeer, "not initialised");
	Asura_Assert(s_eConnectionState == PCN_InGroupRoom, "unexpected state");

	if (s_pxPeer && (s_eConnectionState == PCN_InGroupRoom))
	{
		// get the server object
		SBServer pxServer = Asura_GameSpy_ServerBrowsing::GetServer(uServerIndex);
		Asura_Assert(pxServer, "invalid server index");

		// check this is a staging room, not already running the game
		Asura_Assert(IsInStaging(uServerIndex), "can't join staging room, game already running");

		if (pxServer && IsInStaging(uServerIndex))
		{
			// finished with all other servers from server browser now.
			Asura_GameSpy_ServerBrowsing::Reset();
			Asura_GameSpy_ServerBrowsing::AddServer(pxServer);
			
			// TODO - use Unicode versions (note use of capital %S to convert)
			// SNH - 30/06/2004 12:00:17 - use UTF8 rather than plain ascii.
			// Comments in UCS2ToUTF8String say 3 characters may be needed for each unicode char.
			char szPassword[3 * PEER_PASSWORD_LEN];
//			Asura_PrintfToBfr(szPassword, "%S", PEER_PASSWORD_LEN, true, wszPassword);
			UCS2ToUTF8String(reinterpret_cast<const u_short*>(wszPassword), szPassword);

			// clear the chat and leave this room
			Asura_LobbyChat::Reset();
			peerLeaveRoom( s_pxPeer, GroupRoom, NULL );
			s_eRoomType = NumRooms;

			// remember the server object
			s_pxStagingServer = pxServer;
			
			peerJoinStagingRoom( s_pxPeer, pxServer, szPassword, JoinRoomCallback, NULL, PEERFalse );

			s_eConnectionState = PCN_JoiningStagingRoom;
			return true;
		}
	}
	return false;
}


bool Asura_GameSpy_Peer::IsInStaging( u_int uServerIndex )
{
	SBServer pxServer = Asura_GameSpy_ServerBrowsing::GetServer(uServerIndex);
	Asura_Assert(pxServer, "invalid server index");
	if (pxServer)
	{
		const char* szGameMode = SBServerGetStringValue(pxServer, "gamemode", "");
		return strcmp(szGameMode, "openstaging") == 0;
	}
	return false;
}


void Asura_GameSpy_Peer::RefreshServerList( void )
{
	// NB hostname and gamemode are set by the SDK so no need to add them
	u_char aucFields[] = 
	{
		MAPNAME_KEY,
		GAMETYPE_KEY,
		NUMPLAYERS_KEY,
		MAXPLAYERS_KEY,
		ASURA_MAPID_KEY,
		ASURA_GAMETYPEID_KEY,
		PASSWORD_KEY,
		GAMEVER_KEY,
	};
	const char* szFilter = Asura_GameSpy_ServerBrowsing::GetFilter();
	peerStartListingGames(s_pxPeer, aucFields, sizeof(aucFields)/sizeof(aucFields[0]), szFilter, ListingGamesCallback, NULL);
}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY


bool Asura_GameSpy_Peer::LeaveGroupRoom( void )
{
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
		Asura_Assert(s_eRoomType == GroupRoom, "not in group room");
		if (s_eRoomType == GroupRoom)
		{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
			// shutdown server list
			Asura_GameSpy_ServerBrowsing::Shutdown();
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY

			// clear the chat and leave this room
			Asura_LobbyChat::Reset();
			peerLeaveRoom( s_pxPeer, GroupRoom, NULL );
			s_eRoomType = NumRooms;
			s_eConnectionState = PCN_WaitingToJoinGroupRoom;
			return true;
		}
	}
	return false;
}


bool Asura_GameSpy_Peer::LeaveStagingRoom( void )
{
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
 		Asura_Assert(s_eRoomType == StagingRoom, "not in staging room");
		if (s_eRoomType == StagingRoom)
		{
			// clear the chat and leave this room
			Asura_LobbyChat::Reset();
			peerLeaveRoom( s_pxPeer, StagingRoom, NULL );
			s_eRoomType = NumRooms;
			s_pxStagingServer = NULL;
			RejoinGroupRoom();
			return true;
		}
	}
	return false;
}


void Asura_GameSpy_Peer::RejoinGroupRoom( void )
{
	Asura_Assert(s_pxPeer, "not initialised");
	Asura_Assert(s_eRoomType == NumRooms, "already in a room");

	if (s_pxPeer && (s_eRoomType == NumRooms))
	{
		if (s_uGroupRoomIndex < s_uNumGroupRooms)
		{
			PEERBool eBlocking = PEERFalse;
			int iGroupID = s_pxGroupsRoomArray[s_uGroupRoomIndex].m_iGroupID;
			peerJoinGroupRoom(s_pxPeer, iGroupID, JoinRoomCallback, NULL, eBlocking);
			s_eConnectionState = PCN_JoiningGroupRoom;
		}
		else
		{
			Asura_Assert(false, "invalid group room index");
		}
	}
}


u_int Asura_GameSpy_Peer::GetMyPublicIPAddress( void )
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
		return peerGetPublicIP(s_pxPeer);
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	return 0;
}


bool Asura_GameSpy_Peer::SendNatNegotiateCookie( u_int uIP, u_short usPort, int iCookie )
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
		peerSendNatNegotiateCookie(s_pxPeer, uIP, usPort, iCookie);
		return true;
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	return false;
}

void Asura_GameSpy_Peer::ReportKeyValues()
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(s_pxPeer, "not initialised");
	if(s_pxPeer)
	{
		Asura_Assert(s_eConnectionState == PCN_HostingStagingRoom, "Unexpected state");
		peerStateChanged(s_pxPeer);
		Asura_OutputToDebugger_Retail(("Sending peer state change\n"));
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

void Asura_GameSpy_Peer::SendRoomUTM(const char* pcCommand, const char* pcParameters)
{
	Asura_Assert(s_pxPeer, "not initialised");
	if(s_pxPeer)
	{
		peerUTMRoom(s_pxPeer, s_eRoomType, pcCommand, pcParameters, PEERFalse );
	}
}

//////////////////////////////////////////////////////////////////////////
//
// Peer SDK Callbacks
//
//////////////////////////////////////////////////////////////////////////

// This gets called when the connection attempt finishes.
/////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::ConnectCallback
(
	PEER pxPeer,  // The peer object.
	PEERBool eSuccess,  // PEERTrue if success, PEERFalse if failure.
	int iFailureReason,  // If failure, the reason for it (PEER_DISCONNECTED, etc.)
	void* pParam  // User-data.
)
{
	if (eSuccess == PEERFalse)
	{
		//CHAT_DISCONNECTED = 0
		//CHAT_NICK_ERROR = 1
		//CHAT_LOGIN_FAILED = 2
		Asura_OutputToDebugger_Retail(("Failed to connect to chat server(%d)\n", iFailureReason));
		s_eConnectionState = PCN_ConnectionFailed;
	}
	else
	{
		Asura_OutputToDebugger_Retail(("Connected to chat server\n"));
		s_eConnectionState = PCN_ConnectedToServer;
	}
}

// This gets called if there is an error with
//   the nickname while connecting.
// Call peerRetryWithNick() to try another nick.  If it
//   is called with a NULL nick, then the connect will be
//   stopped, and peerConnectCallback will be called with
//   failure.
// The suggested nicks are only provided if type
//   is PEER_INVALID_UNIQUENICK.
/////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::NickErrorCallback
(
	PEER pxPeer,  // The peer object.
	int iType,  // The type of nick error (PEER_IN_USE, PEER_INVALID, etc.)
	const gsi_char* szNick,  // The bad nick.
	int iNumSuggestedNicks,  // The number of suggested nicks.
	const gsi_char** aszSuggestedNicks,  // The array of nicks.
	void* pParam  // User-data.
)
{
	switch (iType)
	{
		case CHAT_IN_USE:
		case CHAT_INVALID:
		{
			// add a number to the start of the name, the number is cycled 
			// so if this failed, next time we will try another number.

			// SNH: changed to always apply the number to the original name.
			char szNewNick[uASURA_CHAT_MAX_USERNAME_LEN + 1];
			Asura_PrintfToBfr(szNewNick, "[%u]%s", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, s_uNicknameModifier++, s_szAttemptedNick);

			char szFixedNickname[uASURA_CHAT_MAX_USERNAME_LEN + 2];
			
			// fix up any illegal characters
			peerFixNick(szFixedNickname, szNewNick);

			Asura_OutputToDebugger_Retail(("Chat Nickname error with %s; retrying with %s.\n", szNick, szFixedNickname));
			
			peerRetryWithNick(pxPeer, szFixedNickname);
			break;
		}
		
		case CHAT_UNIQUENICK_EXPIRED:
		case CHAT_NO_UNIQUENICK:
		case CHAT_INVALID_UNIQUENICK:
		default:
		{
			Asura_Assert(false, "unexpected nick error");
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// callbacks for listing group rooms
//////////////////////////////////////////////////////////////////////////

// Gets called once for each group room when listing group rooms.
// After this has been called for each group room, it will be
// called one more time with groupID==0 and name==NULL.
/////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::ListGroupRoomsCallback
(
	PEER pxPeer,  // The peer object.
	PEERBool eSuccess,  // PEERTrue if success, PEERFalse if failure.
	int iGroupID,  // A unique ID for this group.
	SBServer pxServerBrowser,  // The server object for this group room.
	const gsi_char* szName,  // The group room's name.
	int iNumWaiting,  // The number of players in the room.
	int iMaxWaiting,  // The maximum number of players allowed in the room.
	int iNumGames,  // The number of games either staging or running in the group.
	int iNumPlaying,  // The total number of players in games in the group.
	void* pParam  // User-data.
)
{
	Asura_Assert(s_eConnectionState == PCN_GettingListOfGroupRooms, "unexpected callback");
	if (iGroupID == 0)
	{
		Asura_Assert(!szName, "unexpected group room name");
		s_eConnectionState = PCN_WaitingToJoinGroupRoom;
	}
	else
	{
		Asura_Assert(s_uNumGroupRooms < uASURA_GAMESPY_PEER_MAXGROUPROOMS, "too many group rooms");
		if (s_uNumGroupRooms < uASURA_GAMESPY_PEER_MAXGROUPROOMS)
		{
			u_int uFilter = static_cast<u_int>(SBServerGetIntValue(pxServerBrowser, Project_GetGroupRoomFilterString(), -1));
			if (uFilter & Project_GetGroupRoomFilterMask())
			{
				u_int uGroupRoomNumber = s_uNumGroupRooms++;
				
				Asura_GameSpy_Peer_GroupRoom* pxRoom = &s_pxGroupsRoomArray[uGroupRoomNumber];
				pxRoom->m_iGroupID = iGroupID;
				pxRoom->m_uNumPlayersInRoom = iNumWaiting;
				pxRoom->m_uMaxPlayersInRoom = iMaxWaiting;
				pxRoom->m_uNumGamesInRoom = iNumGames;
				pxRoom->m_uNumPlayingInRoom = iNumPlaying;
				pxRoom->m_uFilterFlags = uFilter;
				pxRoom->m_pxServer = pxServerBrowser;
				// note use of capital %S to convert from ASCII to Unicode
				Asura_PrintfToBfr(pxRoom->m_awcRoomName, L"%S", uASURA_GAMESPY_PEER_GROUPROOMNAME_MAXLEN, true, szName);
			}
		}
	}
}



//////////////////////////////////////////////////////////////////////////
// callbacks for joining room
//////////////////////////////////////////////////////////////////////////

// This gets called when an attempt to join or create a room has finished.
//////////////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::JoinRoomCallback
(
	PEER pxPeer,  // The peer object.
	PEERBool eSuccess,  // PEERTrue if success, PEERFalse if failure.
	PEERJoinResult eResult,  // The result of the attempt.
	RoomType eRoomType,  // The type of room joined/created.
	void* pParam  // User-data.
)
{
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
		s_eLastRoomJoinResult = eResult;

		switch (s_eConnectionState)
		{
			case PCN_JoiningGroupRoom:
			{
				if (eSuccess == PEERTrue)
				{
					Asura_OutputToDebugger_Retail(("joined room, getting user and server list.\n"));
					s_eConnectionState = PCN_GettingGroupRoomUserList;
					peerEnumPlayers(s_pxPeer, eRoomType, EnumPlayersCallback, NULL);

					Asura_Assert(eRoomType == GroupRoom, "unexpected room type");
					s_eRoomType = eRoomType;

#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
					BeginListingServers();
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY

				}
				else
				{
					Asura_OutputToDebugger_Retail(("failed to join room.\n"));
					s_eConnectionState = PCN_JoinGroupRoomFailed;
				}
			}
			break;

#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
			case PCN_CreatingStagingRoom:
			{
				// clear the chat and leave this room
				Asura_LobbyChat::Reset();
				peerLeaveRoom( s_pxPeer, GroupRoom, NULL );
				s_eRoomType = NumRooms;
		
				if (eSuccess == PEERTrue)
				{
					Asura_OutputToDebugger_Retail(("created staging room.\n"));
					s_eConnectionState = PCN_HostingStagingRoom;

					// call enum players so we get listed in the chat
					peerEnumPlayers(s_pxPeer, eRoomType, EnumPlayersCallback, NULL);

					// check we're moving from group room to staging room
					Asura_Assert(eRoomType == StagingRoom, "unexpected room types");
					s_eRoomType = eRoomType;
				}
				else
				{
					Asura_OutputToDebugger_Retail(("failed to create room.\n"));
					s_eConnectionState = PCN_CreateStagingRoomFailed;
				}
			}
			break;

			case PCN_JoiningStagingRoom:
			{
				if (eSuccess == PEERTrue)
				{
					Asura_OutputToDebugger_Retail(("joined staging room, getting user list.\n"));
					s_eConnectionState = PCN_GettingStagingRoomUserList;
					peerEnumPlayers(s_pxPeer, eRoomType, EnumPlayersCallback, NULL);

					Asura_Assert(eRoomType == StagingRoom, "unexpected room type");
					s_eRoomType = eRoomType;
				}
				else
				{
					Asura_OutputToDebugger_Retail(("failed to join room.\n"));
					s_eConnectionState = PCN_JoinStagingRoomFailed;
					s_pxStagingServer = NULL;
				}
			}
			break;
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY

			default:
			{
				Asura_Assert(false, "unexpected state");
				break;
			}
		}
	}
}


//////////////////////////////////////////////////////////////////////////
// callbacks for listing players in a room
//////////////////////////////////////////////////////////////////////////

// Called for each player in a room being enumerated, and once
//   when finished, with index==-1 and nick==NULL.
//////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::EnumPlayersCallback
(
	PEER pxPeer,  // The peer object.
	PEERBool eSuccess,  // PEERTrue if success, PEERFalse if failure.
	RoomType eRoomType,  // The room whose players are being enumerated.
	int iIndex,  // The index of the player (0 to (N - 1)).  -1 when finished.
	const gsi_char* szNick,  // The nick of the player.
	int iFlags,  // This player's flags (see #define's above).  PANTS|03.12.01
	void* pParam  // User-data.
)
{
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
		if (eSuccess == PEERTrue)
		{
			if (iIndex == -1)
			{
				Asura_OutputToDebugger_Retail(("finished getting user list.\n"));
				switch (s_eConnectionState)
				{
					case PCN_GettingGroupRoomUserList:
					{
						s_eConnectionState = PCN_InGroupRoom;
						break;
					}

#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
					case PCN_GettingStagingRoomUserList:
					{
						s_eConnectionState = PCN_ClientInStagingRoom;
						break;
					}

					case PCN_HostingStagingRoom:
					{
						break;
					}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
					default:
					{
						Asura_Assert(false, "unexpected state");
						break;
					}
				}
			}
			else
			{
				Asura_LobbyChat::AddUser(TranslatePeerNick(szNick, s_acNickNameExtension));
			}
		}
		else
		{
			Asura_OutputToDebugger_Retail(("failed to join room.\n"));
			s_eConnectionState = PCN_ListPlayersFailed;
		}
	}
}


//////////////////////////////////////////////////////////////////////////
// callbacks needed for listing of games
//////////////////////////////////////////////////////////////////////////

// Called with info on games being listed.
// Used to maintain a list of running games and staging rooms.
// The server object is a unique way of identifying each game.
// It can also be used with the calls in the "SBServer Object Functions" section
//   of sb_serverbrowsing.h to find out more info about the server.
// If staging==PEERTrue, the game hasn't started yet, it's still in the staging room
//   use peerJoinStagingRoom() to join the staging room, or if staging==peerfalse
//   use the server object to get the game's IP and port to join with.
// During the _intial_ listing of games, progress is the percentage (0-100) of the
//   games that have been added.  Once the initial listing is completed,
//   progress will always be 100.
// PANTS|09.11.00 - The "password" key will be set to 1 for games that are
// passworded.  This can be checked with ServerGetIntValue(server, "password", 0).
/////////////////////////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::ListingGamesCallback
(
	PEER pxPeer,  // The peer object.
	PEERBool eSuccess,  // PEERTrue if success, PEERFalse if failure.
	const gsi_char* szName,  // The name of the game being listed.
	SBServer pxServer,  // The server object for this game.
	PEERBool eStaging,  // If PEERTrue, this is a staging room and not a running game.
	int iMsg,  // The type of message this is.
		// PEER_CLEAR:
		//   Clear the list.  This has the same effect as if this was called
		//   with PEER_REMOVE for every server listed.
		// PEER_ADD:
		//   This is a new server. Add it to the list.
		// PEER_UPDATE:
		//   This server is already on the list, and its been updated.
		// PEER_REMOVE:
		//   Remove this server from the list.  The server object for this server
		//   should not be used again after this callback returns.
	int iProgress,  // The percent of servers that have been added.
	void* pParam  // User-data.
)
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	if(Asura_GameSpy_ServerBrowsing::IsInitialised())
	{
		Asura_OutputToDebugger_Retail(("game listing callback:"));

		switch (iMsg)
		{
			case PEER_ADD:
			{
				Asura_OutputToDebugger_Retail((" add\n"));
				Asura_GameSpy_ServerBrowsing::AddServer(pxServer);
			}
			break;
			
			case PEER_UPDATE:
			{
				Asura_OutputToDebugger_Retail((" update\n"));
				Asura_GameSpy_ServerBrowsing::AddServer(pxServer);
			}
			break;

			case PEER_REMOVE:
			{
				Asura_OutputToDebugger_Retail((" remove\n"));
				Asura_GameSpy_ServerBrowsing::RemoveServer(pxServer);
			}
			break;

			case PEER_CLEAR:
			{
				Asura_OutputToDebugger_Retail((" clear\n"));
				Asura_GameSpy_ServerBrowsing::Reset();
			}
			break;

			case PEER_COMPLETE:
			{
				Asura_OutputToDebugger_Retail((" complete\n"));
				// initial listing of servers is complete.
			}
			break;

			default:
			{
				Asura_Assert(false, "unhandled peer message");
			}
			break;
		}
	}
	else
	{
		// ignore
		Asura_OutputToDebugger_Retail(("ignoring game listing callback %d as server browsing not initialised.\n"));
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY

}


//////////////////////////////////////////////////////////////////////////
// callbacks needed to create PEER object
//////////////////////////////////////////////////////////////////////////

// Called when the connection to the server gets disconnected.
//////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::DisconnectedCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szReason,  // The reason for the disconnection.
	void* pParam  // User-data.
) 
{
	Asura_OutputToDebugger_Retail(("Peer chat Disconnected: %s\n", szReason));
	s_eRoomType = NumRooms;
}

// Called when a message is sent to a room the local player is in.
//////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::RoomMessageCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the message was in.
	const gsi_char* szNick,  // The nick of the player who sent the message.
	const gsi_char* szMessage,  // The text of the message.
	MessageType eMessageType,  // The type of message.
	void* pParam  // User-data.
) 
{
	// only add the message if we're in the room it came from, so we don't
	// see group room chat in the staging room
	Asura_Assert(eRoomType == s_eRoomType, "wrong room type");
	if (eRoomType == s_eRoomType)
	{
		Asura_UnicodeChar awcUser[uASURA_CHAT_MAX_USERNAME_LEN + 1];
		Asura_UnicodeChar awcMessage[uASURA_CHAT_MAX_MESSAGE_SIZE];

		// SNH - 30/06/2004 12:00:17 - use UTF8 rather than plain ascii.
		//Asura_PrintfToBfr(awcUser, L"%S", uASURA_CHAT_MAX_USERNAME_LEN + 1, true, szNick);
		UTF8ToUCS2String(TranslatePeerNick(szNick, s_acNickNameExtension), reinterpret_cast<u_short*>(awcUser));
		UTF8ToUCS2String(szMessage, reinterpret_cast<u_short*>(awcMessage));
		Asura_LobbyChat::AddMessage(awcUser, awcMessage);
	}
}

// Called when a UTM is sent to a room the local player is in.
//////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::RoomUTMCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the UTM was in.
	const gsi_char* szNick,  // The nick of the player who sent the UTM.
	const gsi_char* szCommand, // The UTM command for this message.
	const gsi_char* szParameters,  // Any parameters for this UTM.
	PEERBool eAuthenticated,  // True if this has been authenticated by the server.
	void* pParam  // User-data.
) 
{
	// UTMs are used to pass arbitrary information around between players.
	Project_InterpretUTM(szCommand, szParameters);
}

// Called when the name of a room the player is in changes.
// The new name can be checked with peerGetRoomName.
///////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::RoomNameChangedCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the name changed in.
	void* pParam  // User-data
)
{}

// Called when a room's mode changes.
///////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::RoomModeChangedCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the name changed in.
	CHATChannelMode * mode,  // The current mode for this room.
	void* pParam  // User-data
) 
{}

// Called when a private message is received from another player.
/////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::PlayerMessageCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szNick,  // The nick of the player who sent the message.
	const gsi_char* szMessage,  // The text of the message.
	MessageType eMessageType,  // The type of message.
	void* pParam  // User-data
) 
{
	Asura_LobbyChat::AddMessage(TranslatePeerNick(szNick, s_acNickNameExtension), szMessage, true);
}

// Called when a private UTM is received from another player.
/////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::PlayerUTMCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szNick,  // The nick of the player who sent the UTM.
	const gsi_char* szCommand, // The UTM command for this message.
	const gsi_char* szParameters,  // Any parameters for this UTM.
	PEERBool eAuthenticated,  // True if this has been authenticated by the server.
	void* pParam  // User-data
) 
{
	// UTMs are used to pass arbitrary information around between players. We don't use them.
}

// Called when a player's ready state changes,
// from a call to peerSetReady().
//////////////////////////////////////////////
void Asura_GameSpy_Peer::ReadyChangedCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szNick,  // The nick of the player who's ready state changed.
	PEERBool eReady,  // The player's new ready state.
	void* pParam  // User-data.
) 
{}

// Called when the host of a staging room launches the game,
// with a call to peerStartGame().
// The public and private IPs and ports of the server can
// be obtained from the server object.
////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::GameStartedCallback
(
	PEER pxPeer,  // The peer object.
	SBServer xServer,  // A server object representing this host.
	const gsi_char* szMessage,  // A message that was passed into peerStartGame().
	void* pParam  // User-data.
) 
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	Asura_Assert(s_eConnectionState == PCN_ClientInStagingRoom, "unexpected state for game start callback");

	Asura_OutputToDebugger_Retail(("game has started.\n"));
	s_eConnectionState = PCN_ClientGameStarting;
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

// A player joined a room.
//////////////////////////
void Asura_GameSpy_Peer::PlayerJoinedCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the player joined.
	const gsi_char* szNick,  // The nick of the player that joined.
	void* pParam  // User-data.
) 
{
	Asura_Assert(eRoomType == s_eRoomType, "wrong room type");
	// only add if this is the room we're currently in.
	if (eRoomType == s_eRoomType)
	{
		Asura_LobbyChat::AddUser(TranslatePeerNick(szNick, s_acNickNameExtension));
	}
}

// A player left a room.
////////////////////////
void Asura_GameSpy_Peer::PlayerLeftCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the player left.
	const gsi_char* szNick,  // The nick of the player that left.
	const gsi_char* szReason,  // The reason the player left.
	void* pParam  // User-data.
) 
{
	// only remove if this is the room we're currently in.
	Asura_Assert(eRoomType == s_eRoomType, "wrong room type");
	if (eRoomType == s_eRoomType)
	{
		Asura_LobbyChat::RemoveUser(TranslatePeerNick(szNick, s_acNickNameExtension));
	}
	
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	if (eRoomType == StagingRoom)
	{
		if (peerIsPlayerHost(pxPeer, szNick, StagingRoom))
		{
			// host has quit staging room without starting the game
			s_eConnectionState = PCN_StagingRoomHostQuit;
		}
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

// The local player was kicked from a room.
///////////////////////////////////////////
void Asura_GameSpy_Peer::KickedCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room that the player was kicked from.
	const gsi_char* szNick,  // The nick of the player that did the kicking.
	const gsi_char* szReason,  // An optional reason for the kick.
	void* pParam  // User-data.
) 
{
}

// The entire player list for this room has been updated.
/////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::NewPlayerListCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of room.
	void* pParam  // User-data
) 
{
	Asura_OutputToDebugger_Retail(("PeerChat user list updated\n"));
}

// A player in one of the rooms changed his nick.
/////////////////////////////////////////////////
void Asura_GameSpy_Peer::PlayerChangedNickCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The type of the room the nick changed was in.
	const gsi_char* szOldNick,  // The player's old nick.
	const gsi_char* szNewNick,  // The player's new nick.
	void* pParam  // User-data.
)
{
	Asura_OutputToDebugger_Retail(("PeerChat user changed nick\n"));
}

// The IP and ProfileID for this player has just been received.
// PANTS|01.08.01
// This gets called for all players (who are using peer) in a room
// shortly after joining.  It will be called with nick==NULL after
// getting info for all the players.
//////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::PlayerInfoCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The room the info was gotten in.
	const gsi_char* szNick,  // The nick of the player the info is for.
	u_int uIP,  // The player's IP.
	int iProfileID,  // The player's profile ID.
	void* pParam  // User-data.
) 
{}

// This gets called when a player's flags have changed.
///////////////////////////////////////////////////////
void Asura_GameSpy_Peer::PlayerFlagsChangedCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The room the flags were changed in.
	const gsi_char* szNick,  // The player whose flags have changed.
	int iOldFlags,  // The player's old flags.
	int iNewFlags,  // The player's new flags.
	void* pParam  // User-data
) 
{}

// An updated ping for a player, who may be in any room(s).
///////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::PingCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szNick,  // The other player's nick.
	int iPing,  // The ping.
	void* pParam  // User-data.
)
{}

// An updated cross-ping between two players in the staging room.
/////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::CrossPingCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szNick1,  // The first player's nick.
	const gsi_char* szNick2,  // The second player's nick.
	int iCrossPing,  // The cross-ping.
	void* pParam  // User-data.
)
{}

// This is called for watch keys when a room is joined, for
// watch keys when another player joins, and for any newly 
// set watch keys.
///////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::GlobalKeyChangedCallback
(
	PEER pxPeer,  // The peer object.
	const gsi_char* szNick,  // The player whose key changed.
	const gsi_char* szKey,  // The key.
	const gsi_char* szValue,  // The value.
	void* pParam  // User-data.
)
{}

// This is called for watch keys when a room is joined, for
// watch keys when another player joins, for any newly 
// set watch keys, and when a broadcast key is changed.
///////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::RoomKeyChangedCallback
(
	PEER pxPeer,  // The peer object.
	RoomType eRoomType,  // The room the player is in.
	const gsi_char* szNick,  // The player whose key changed.
	const gsi_char* szKey,  // The key.
	const gsi_char* szValue,  // The value.
	void* pParam  // User-data.
)
{}

// Called to report QR server keys.
// Use qr2_buffer_add or qr2_buffer_add_int to
// add this key's information to the buffer.
//////////////////////////////////////////////
void Asura_GameSpy_Peer::QRServerKeyCallback
(
	PEER pxPeer,  // The peer object.
	int iKey,  // The key for which to report information.
	qr2_buffer_t pxBuffer,  // Fill in the information using this buffer.
	void* pParam  // User-data.
) 
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	//Asura_GameSpy_QR2::ServerKeyCallback(iKey, pxBuffer, NULL);
	char acTemp[256];

	switch (iKey)
	{
		case GAMEVER_KEY:
		{
			qr2_buffer_add_int(pxBuffer, Asura_GameSpy::GetGameVersion());
			break;
		}

		case MAPNAME_KEY:
		{
			const UCS2String wszMapName = reinterpret_cast<const u_short*>(Project_GetMapName());
			UCS2ToUTF8String(wszMapName, acTemp);
			qr2_buffer_add(pxBuffer, acTemp);
			break;
		}
		case GAMETYPE_KEY:
		{
			const UCS2String wszGameTypeString = reinterpret_cast<const u_short*>(Project_GetGameTypeString());
			UCS2ToUTF8String(wszGameTypeString, acTemp);
			qr2_buffer_add(pxBuffer, acTemp);
			break;
		}

		// SNH: only need to handle this while in-game. Possible values are in Asura_GameSpy_QR2.cpp
		case GAMEMODE_KEY:
		{
			const UCS2String wszGameModeString = reinterpret_cast<const u_short*>(Project_GetGameModeString());
			UCS2ToUTF8String(wszGameModeString, acTemp);
			qr2_buffer_add(pxBuffer, acTemp);
			break;
		}

		// asura's custom keys
		case ASURA_MAPID_KEY:
		{
			qr2_buffer_add_int(pxBuffer, Project_GetMapID());
			break;
		}
		case ASURA_GAMETYPEID_KEY:
		{
			qr2_buffer_add_int(pxBuffer, Project_GetGameTypeID());
			break;
		}

		default:
		{
			Asura_Assert(false, "unexpected key value in QRServerKeyCallback");
			qr2_buffer_add(pxBuffer, "");
			break;
		}
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

// Called to report QR player keys.
// Use qr2_buffer_add or qr2_buffer_add_int to
// add this key's information to the buffer.
//////////////////////////////////////////////
void Asura_GameSpy_Peer::QRPlayerKeyCallback
(
	PEER pxPeer,  // The peer object.
	int iKey,  // The key for which to report information.
	int iIndex,  // The index of the player for which to report info.
	qr2_buffer_t pxBuffer,  // Fill in the information using this buffer.
	void* pParam  // User-data.
) 
{	
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	//Asura_GameSpy_QR2::PlayerKeyCallback(iKey, iIndex, pxBuffer, NULL);
	Asura_OutputToDebugger(("QRPlayerKeyCallback\n"));
}

// Called to report QR team keys.
// Use qr2_buffer_add or qr2_buffer_add_int to
// add this key's information to the buffer.
//////////////////////////////////////////////
void Asura_GameSpy_Peer::QRTeamKeyCallback
(
	PEER pxPeer,  // The peer object.
	int iKey,  // The key for which to report information.
	int iIndex,  // The index of the team for which to report info.
	qr2_buffer_t pxBuffer,  // Fill in the information using this buffer.
	void* pParam  // User-data.
) 
{
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	//Asura_GameSpy_QR2::TeamKeyCallback(iKey, iIndex, pxBuffer, NULL);
	Asura_OutputToDebugger(("QRTeamKeyCallback\n"));
}

// Called to get a list of keys to be reported.
// Use qr2_keybuffer_add() to add keys.
// The following keys do not need to be added, as peer already adds them:
// GAMENAME_KEY, HOSTNAME_KEY, NUMPLAYERS_KEY, MAXPLAYERS_KEY,
// GAMEMODE_KEY, PASSWORD_KEY, GROUPID_KEY, PLAYER__KEY, PING__KEY.
/////////////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::QRKeyListCallback
(
	PEER pxPeer,  // The peer object.
	qr2_key_type eType,  // The type of keys being asked for (key_server, key_player, or key_team).
	qr2_keybuffer_t pxKeyBuffer,  // Fill in the keys using this buffer.
	void* pParam  // User-data.
)
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	//Asura_GameSpy_QR2::KeyListCallback(eType, pxKeyBuffer, NULL);
	//Asura_OutputToDebugger(("QRKeyListCallback\n"));
	// add map name, game type, map id and game type ID
	//need to add all the keys we support
	switch (eType)
	{
	case key_server:
		qr2_keybuffer_add(pxKeyBuffer, GAMEVER_KEY);
		qr2_keybuffer_add(pxKeyBuffer, MAPNAME_KEY);
		qr2_keybuffer_add(pxKeyBuffer, GAMETYPE_KEY);
		// asura custom keys
		qr2_keybuffer_add(pxKeyBuffer, ASURA_MAPID_KEY);
		qr2_keybuffer_add(pxKeyBuffer, ASURA_GAMETYPEID_KEY);
		break;
	default:
		break;
	}
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

// Called to get a count of the number of players or teams.
///////////////////////////////////////////////////////////
int Asura_GameSpy_Peer::QRCountCallback
(
	PEER pxPeer,  // The peer object.
	qr2_key_type eType,  // The type of count to return (key_player or key_team).
	void* pParam  // User-data.
) 
{
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	//Asura_OutputToDebugger(("QRCountCallback\n"));
	return 0;
	//return Asura_GameSpy_QR2::CountCallback(eType, NULL);
}

// Called when there is an error reporting the server.
//////////////////////////////////////////////////////
void Asura_GameSpy_Peer::QRAddErrorCallback
(
	PEER pxPeer,  // The peer object.
	qr2_error_t eError,  // The type of error.
	gsi_char* szErrorString,  // A text string containing the error.
	void* pParam  // User-data.
) 
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	Asura_OutputToDebugger_Retail(("QRAddErrorCallback: %s\n", szErrorString));

	s_bErrorAddingServer = (eError != e_qrnoerror);
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

// Called when hosting a server and a nat-negotiate cookie is received.
///////////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::QRNatNegotiateCallback
(
	PEER pxPeer,  // The peer object.
	int iCookie,  // A cookie sent from a potential client.
	void* pParam  // User-data.
) 
{
#ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	Asura_OutputToDebugger_Retail(("Got natneg cookie: %d\n", iCookie));

	Asura_NetTransport_System::ServerBeginNatNeg(iCookie);
#endif // #ifndef ASURA_GAMESPY_PEER_CHAT_ONLY
}

// Called when hosting a server with the server's public reporting address.
///////////////////////////////////////////////////////////////////////////
void Asura_GameSpy_Peer::QRPublicAddressCallback
(
	PEER pxPeer,  // The peer object.
	u_int uIP,  // The public reporting IP
	u_short usPort,  // The public reporting port
	void* pParam  // User-data.
)
{
	Asura_Assert(pxPeer == s_pxPeer, "unknown Peer object");
	// this gives the address and port as reported by the NAT to GameSpy's servers
//	char acAddress[32];
//	gt2AddressToString( uIP, usPort, acAddress );
//	Asura_OutputToDebugger_Retail(("public address = %s\n", acAddress));
	Asura_OutputToDebugger_Retail(("public address = %d.%d.%d.%d:%d\n", ((uIP & 0x000000ff) >> 0), ((uIP & 0x0000ff00) >> 8), ((uIP & 0x00ff0000) >> 16), ((uIP & 0xff000000) >> 24), usPort));
}


void Asura_GameSpy_Peer::ParsePeerQuery(u_int uIP, u_short usPort, char* pcQueryData, u_int uDataLength)
{
	Asura_Assert(s_pxPeer, "not initialised");
	if (s_pxPeer)
	{
		sockaddr_in xSockAddress;
		xSockAddress.sin_family = AF_INET;
		xSockAddress.sin_port = htons(usPort);
		xSockAddress.sin_addr.s_addr = uIP;

		peerParseQuery(s_pxPeer, pcQueryData, static_cast<int>(uDataLength), reinterpret_cast<sockaddr*>(&xSockAddress));
	}
}

#endif // #ifdef ASURA_SUPPORT_GAMESPY
