/**
*
* \file Asura_GameSpy_StatLogging.cpp
*
* Created by Richard May - 27/11/2003 11:19:07
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/


/************/
/* Includes.*/
/************/

// main asura header
#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_GAMESPY

// local header
#include "Network/GameSpy/Asura_GameSpy_StatLogging.h"

// asura
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_LogFile.h"
#include "Network/GameSpy/Asura_GameSpy.h"
#include "Network/GameSpy/Asura_GameSpy_GP.h"
#include "Network/Asura_Network_StatLogging.h"

// middleware
#include "GameSpy/gstats/gstats.h"


#ifdef ASURA_USE_NETWORK_STATLOGGING

/*********/
/* Data. */
/*********/

struct GuidAndPlayerID
{
	Asura_Guid uGuid;
	u_int uGSPlayerID;	
	bool  bIsGuid;
};

struct TeamAndTeamID
{	
	int   iTeam;
	u_int uGSTeamID;
};

char Asura_GameSpy_StatLogging::s_szSnapshotBuffer[];
GuidAndPlayerID* Asura_GameSpy_StatLogging::s_pxPlayerIDStore = NULL;
TeamAndTeamID*	 Asura_GameSpy_StatLogging::s_pxTeamIDStore   = NULL;

u_int Asura_GameSpy_StatLogging::s_uNextTeamID   = 0;
u_int Asura_GameSpy_StatLogging::s_uNextPlayerID = 0;
u_int Asura_GameSpy_StatLogging::s_uPlayerIDStoreSize = 0;
u_int Asura_GameSpy_StatLogging::s_uTeamIDStoreSize   = 0;

/**************/
/* Functions. */
/**************/

/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::Initialise
 *
 * Author        : Richard May
 *
 * Description   : PC specific initialise function - connects to
 *				   the gamespy stat logging server
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::Initialise ( void )
{
	// make sure this line is removed in retail builds. And the one in the frontend menu too.
	if (Asura_GameSpy::ShouldUseDevelopmentBackend())
	{
		strcpy( StatsServerHostname, "sdkdev.gamespy.com" );
	}
	
	// setup project gamename, etc. 
	// note that these may have already been set, but setting them again shouldn't hurt
	strcpy( gcd_gamename, Asura_GameSpy::GetGameName() );
	strcpy( gcd_secret_key, Asura_GameSpy::GetGameKey() );

	// connect to gamespy server
	int iResult = InitStatsConnection(0);
	if( iResult == GE_NOERROR )
	{
		Asura_OutputToDebugger(("Connected to gamespy tracking server"));
	}
	else
	{
		// churn out some useful debugging info
		switch(iResult)
		{
		case GE_NODNS     : Asura_Assert(false,"Connection failed: Unable to resolve stats server DNS"); break;
		case GE_NOSOCKET  : Asura_Assert(false,"Connection failed: Unable to create data socket"); break;
		case GE_NOCONNECT : Asura_Assert(false,"Connection failed: Unable to connect to stats server"); break;
		case GE_DATAERROR :	Asura_Assert(false,"Connection failed: Unable to receive challenge from stats server, or bad challenge"); break;
		default : Asura_Assert(false,"Connection failed: Unknown error"); break;
		}		
	}

	// initialise player and team id lists
	InitPlayerIDStorage();
	InitTeamIDStorage();
}
/* end of function Initialise */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::Shutdown
 *
 * Author        : Richard May
 *
 * Description   : PC specific shutdown - closes the connection
 *				   to the gamespy stat logging server
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::Shutdown ( void )
{
	delete [] s_pxPlayerIDStore;
	s_pxPlayerIDStore = NULL;

	delete [] s_pxTeamIDStore;
	s_pxTeamIDStore = NULL;

	CloseStatsConnection();
}
/* end of function Shutdown */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::StartSession
 *
 * Author        : Richard May
 *
 * Description   : 
 *
 *************************************************************/
bool Asura_GameSpy_StatLogging::StartSession ( void )
{
	NewGame(0); // '0' is to disable using buckets - we use our own storage

	// re-initialise player and team id lists
	InitPlayerIDStorage();
	InitTeamIDStorage();

	// add our hostname
	Asura_Network_StatLogging::RegisterVariable( "hostname", "SniperHost" );	

	return true;
}
/* end of function StartSession */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::EndSession
 *
 * Author        : Richard May
 *
 * Description   : 
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::EndSession ( void )
{
	// gamespy requires a final snapshot, regardless of the current 
	// snapshot mode so if we haven't already, send one now
	if( !(Asura_Network_StatLogging::GetSnapshotModeFlags() & ASURA_STATLOG_SNAPSHOTFLAG_ONEND) )
	{
		SendSnapshot(true);
	}

	FreeGame(NULL);
}
/* end of function EndSession */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::SendSnapshot
 *
 * Author        : Richard May
 *
 * Description   : Generates the gamespy-specific snapshot format
 *				   and sends it to the stat logging server
 *
 *************************************************************/
bool Asura_GameSpy_StatLogging::SendSnapshot ( bool bGameEnd )
{
	const Asura_StatLogging_Stat* pxVariableData = Asura_Network_StatLogging::GetStatData();

	// generate the snapshot from our stored data
	s_szSnapshotBuffer[0] = 0;
	for( u_int u=0; u<ASURA_STATLOG_MAXSTATS; u++ )
	{
		// ignore 'blank' stats
		if( pxVariableData[u].m_eType == STAT_NONE ) continue;
		
		char szKey[ASURA_STATLOG_MAXKEYSIZE];
		char szKeyValue[128];
		if( (pxVariableData[u].m_uOwner != ASURA_GUID_UNREGISTERED) ||
			(pxVariableData[u].m_iTeamID != ASURA_STATLOG_NOTEAM) )
		{
			// convert asura-style keys to gamespy-style keys
			GenerateGameSpyKey( szKey, &pxVariableData[u] );			
		}
		else
		{
			strcpy( szKey, pxVariableData[u].m_szName );
		}

		switch(pxVariableData[u].m_eType)
		{
		// basic types
		case STAT_BOOL      : Asura_PrintfToBfr( szKeyValue, "%d", 128, true, pxVariableData[u].m_bValue );   break;
		case STAT_INT	    : Asura_PrintfToBfr( szKeyValue, "%d", 128, true, pxVariableData[u].m_iValue );   break;
		case STAT_UINT	    : Asura_PrintfToBfr( szKeyValue, "%d", 128, true, pxVariableData[u].m_uValue );   break;
		case STAT_FLOAT	    : Asura_PrintfToBfr( szKeyValue, "%f", 128, true, Asura_Maths::FtoD(pxVariableData[u].m_fValue) );   break;
		// ptr types
		case STAT_PTR_BOOL  : Asura_PrintfToBfr( szKeyValue, "%d", 128, true, *pxVariableData[u].m_pbValue ); break;
		case STAT_PTR_INT   : Asura_PrintfToBfr( szKeyValue, "%d", 128, true, *pxVariableData[u].m_piValue ); break;
		case STAT_PTR_UINT  : Asura_PrintfToBfr( szKeyValue, "%d", 128, true, *pxVariableData[u].m_puValue ); break;
		case STAT_PTR_FLOAT : Asura_PrintfToBfr( szKeyValue, "%f", 128, true, Asura_Maths::FtoD(*pxVariableData[u].m_pfValue) ); break;
		// string
		case STAT_STRING    : Asura_PrintfToBfr( szKeyValue, "%s", 128, true, pxVariableData[u].m_szValue ); break;
		default : Asura_Assertf( false, ("Unknown variable type! (%d)",pxVariableData[u].m_eType) ); continue;
		}

		Asura_PrintfToBfr( s_szSnapshotBuffer, "%s\\%s\\%s", UINT_MAX, true, s_szSnapshotBuffer, szKey, szKeyValue );	
	}

	// add player profileID (pid) (GP currently only supports one connection so one player)
	for( u_int u=0; u<s_uPlayerIDStoreSize; ++u )
	{
		char szAuthKey[ASURA_STATLOG_MAXKEYSIZE];	

		if( ((s_pxPlayerIDStore[u].uGuid == ASURA_GUID_UNREGISTERED) && (s_pxPlayerIDStore[u].bIsGuid)) ||
			((s_pxPlayerIDStore[u].uGuid == UINT_MAX) && !(s_pxPlayerIDStore[u].bIsGuid)) )
		{
			break;
		}

		Asura_PrintfToBfr( szAuthKey, "pid_%d", ASURA_STATLOG_MAXKEYSIZE, true, s_pxPlayerIDStore[u].uGSPlayerID );		
		Asura_PrintfToBfr( s_szSnapshotBuffer, "%s\\%s\\%d", UINT_MAX, true, s_szSnapshotBuffer, szAuthKey, Asura_GameSpy_GP::GetCurrentProfileID() );	
	}

	// add player authentification
	for( u_int u=0; u<s_uPlayerIDStoreSize; ++u )
	{
		char szAuthKey[ASURA_STATLOG_MAXKEYSIZE];		

		if( ((s_pxPlayerIDStore[u].uGuid == ASURA_GUID_UNREGISTERED) && (s_pxPlayerIDStore[u].bIsGuid)) ||
			((s_pxPlayerIDStore[u].uGuid == UINT_MAX) && !(s_pxPlayerIDStore[u].bIsGuid)) )
		{
			break;
		}
	
		char szAuth[33];
		GetPlayerAuthentification(szAuth);

		Asura_PrintfToBfr( szAuthKey, "auth_%d", ASURA_STATLOG_MAXKEYSIZE, true, s_pxPlayerIDStore[u].uGSPlayerID );		
		Asura_PrintfToBfr( s_szSnapshotBuffer, "%s\\%s\\%d", UINT_MAX, true, s_szSnapshotBuffer, szAuthKey, szAuth );	
	}

	// send
	int iType = (bGameEnd == true) ? SNAP_FINAL : SNAP_UPDATE;
	int iResult = SendGameSnapShot( NULL, s_szSnapshotBuffer, iType );

	if( iResult != GE_NOERROR ) 
	{
		switch(iResult)
		{
		case GE_DATAERROR : Asura_Assert( false, "Snapshot send failed: No valid game (NewGame must've failed)" ); break;
		case GE_NOCONNECT : Asura_Assert( false, "Snapshot send failed: No connection to server" ); break;
		default : Asura_Assertf( false, ("Snapshot send failed: Unknown error (%d)",iResult) ); break;
		}

		return false;
	}

	return true;
}
/* end of function SendSnapshot */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::GenerateGameSpyKey
 *
 * Author        : Richard May
 *
 * Description   : Converts an asura-style variable key into a 
 *				   valid gamespy key
 *
 * Parameters    : szKey - gamespy-style output
 *				   pxStat  - pointer to the stat to convert
 *
 * Return value  : void
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::GenerateGameSpyKey ( char* szKey, const Asura_StatLogging_Stat* pxStat )
{
	if(!pxStat) return;

	int   iID = 0;
	u_int uHeaderSize = 0;	
	char  szTemp[ASURA_STATLOG_MAXKEYSIZE]; // working buffer

	// strip off our own asura-style header from the key
	if( pxStat->m_uOwner != ASURA_GUID_UNREGISTERED )
	{		
		Asura_PrintfToBfr( szTemp, "p%d_", ASURA_STATLOG_MAXKEYSIZE, true, pxStat->m_uOwner );
		uHeaderSize = strlen(szTemp);

		if( pxStat->m_bIsGuid )
		{
			iID = GetPlayerIDFromGuid(pxStat->m_uOwner);
		}	
		else
		{
			iID = GetPlayerIDFromID( (int)pxStat->m_uOwner );
		}
	}
	else if( pxStat->m_iTeamID != ASURA_STATLOG_NOTEAM )
	{	
		Asura_PrintfToBfr( szTemp, "t%d_", ASURA_STATLOG_MAXKEYSIZE, true, pxStat->m_iTeamID );
		uHeaderSize = strlen(szTemp);
		iID = GetTeamIDFromTeam(pxStat->m_iTeamID);
	}

	u_int v=0;
	for( u_int u=uHeaderSize; u<(ASURA_STATLOG_MAXKEYSIZE-uHeaderSize); u++ )
	{
		szTemp[v] = pxStat->m_szName[u];		
		v++;
	}
	szTemp[v] = 0;
		
	// add the gamespy footer (_N for players, _tN for teams)
	if( (pxStat->m_uOwner != ASURA_GUID_UNREGISTERED) || (!pxStat->m_bIsGuid) )
	{
		Asura_PrintfToBfr( szKey, "%s_%d", ASURA_STATLOG_MAXKEYSIZE, true, szTemp, iID );
	}
	else if( pxStat->m_iTeamID != ASURA_STATLOG_NOTEAM )
	{
		Asura_PrintfToBfr( szKey, "%s_t%d", ASURA_STATLOG_MAXKEYSIZE, true, szTemp, iID );
	}
}
/* end of function GenerateGameSpyKey */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::GetPlayerNumFromGuid
 *
 * Author        : Richard May
 *
 * Description   : Converts from a guid into a gamespy player ID
 *				   Gamespy player IDs must be contiguous in numbering,
 *				   unlike guids, so they can't be generated directly
 *
 *************************************************************/
u_int Asura_GameSpy_StatLogging::GetPlayerIDFromGuid ( Asura_Guid uPlayer )
{
	// search the player guid storage
	u_int u=0;
	for( ; u<s_uPlayerIDStoreSize; ++u )
	{
		if( s_pxPlayerIDStore[u].uGuid == uPlayer )
		{
			return s_pxPlayerIDStore[u].uGSPlayerID;
		}
		else if( ((s_pxPlayerIDStore[u].uGuid == ASURA_GUID_UNREGISTERED) && (s_pxPlayerIDStore[u].bIsGuid)) ||
				 ((s_pxPlayerIDStore[u].uGuid == UINT_MAX) && (s_pxPlayerIDStore[u].bIsGuid)) )
		{
			break;
		}
	}

	// if we've reached here, guid isn't in  the storage so allocate new id.
	if( u>=s_uPlayerIDStoreSize )
	{
		// allocate more space
		GuidAndPlayerID* pxTempStore = s_pxPlayerIDStore;
		u_int uOldSize = s_uPlayerIDStoreSize;
		
		s_uPlayerIDStoreSize *= 2;
		s_pxPlayerIDStore = new GuidAndPlayerID[s_uPlayerIDStoreSize];

		for( u_int v=0; v<s_uPlayerIDStoreSize; ++v )
		{
			if( v < uOldSize )
			{
				s_pxPlayerIDStore[v] = pxTempStore[v];
			}
			else
			{
				s_pxPlayerIDStore[v].uGuid = ASURA_GUID_UNREGISTERED;
				s_pxPlayerIDStore[v].uGSPlayerID = UINT_MAX;
				s_pxPlayerIDStore[v].bIsGuid = true;
			}
		}

		delete [] pxTempStore;
	}

	s_pxPlayerIDStore[u].uGuid = uPlayer;
	s_pxPlayerIDStore[u].uGSPlayerID = s_uNextPlayerID++; // increment after assignment
	s_pxPlayerIDStore[u].bIsGuid = true;

	return s_pxPlayerIDStore[u].uGSPlayerID;
}
/* end of function GetPlayerNumFromGuid */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::GetPlayerIDFromID
 *
 * Author        : Richard May
 *
 * Description   : Adds a plain pre-set player ID (for use in menus and the like)
 *				   while maintaining a concurrent order (which gamespy requires)
 *
 *************************************************************/
u_int Asura_GameSpy_StatLogging::GetPlayerIDFromID ( int iPlayerID )
{
	// search the player guid storage
	u_int u=0;
	for( ; u<s_uPlayerIDStoreSize; ++u )
	{
		if( !s_pxPlayerIDStore[u].bIsGuid && (s_pxPlayerIDStore[u].uGuid == (u_int)iPlayerID) )
		{
			return s_pxPlayerIDStore[u].uGSPlayerID;
		}
		else if( ((s_pxPlayerIDStore[u].uGuid == ASURA_GUID_UNREGISTERED) && (s_pxPlayerIDStore[u].bIsGuid)) ||
				 ((s_pxPlayerIDStore[u].uGuid == UINT_MAX) && (s_pxPlayerIDStore[u].bIsGuid)) )
		{
			break;
		}
	}

	// if we've reached here, id isn't in  the storage so allocate new.
	if( u>=s_uPlayerIDStoreSize )
	{
		// allocate more space
		GuidAndPlayerID* pxTempStore = s_pxPlayerIDStore;
		u_int uOldSize = s_uPlayerIDStoreSize;
		
		s_uPlayerIDStoreSize *= 2;
		s_pxPlayerIDStore = new GuidAndPlayerID[s_uPlayerIDStoreSize];

		for( u_int v=0; v<s_uPlayerIDStoreSize; ++v )
		{
			if( v < uOldSize )
			{
				s_pxPlayerIDStore[v] = pxTempStore[v];
			}
			else
			{
				s_pxPlayerIDStore[v].uGuid   = ASURA_GUID_UNREGISTERED;
				s_pxPlayerIDStore[v].bIsGuid = true;
				s_pxPlayerIDStore[v].uGSPlayerID = UINT_MAX;
			}
		}

		delete [] pxTempStore;
	}

	s_pxPlayerIDStore[u].uGuid = iPlayerID;
	s_pxPlayerIDStore[u].uGSPlayerID = s_uNextPlayerID++; // increment after assignment
	s_pxPlayerIDStore[u].bIsGuid = false;

	return s_pxPlayerIDStore[u].uGSPlayerID;
}
/* end of function AddPlayerID */



/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::GetTeamIDFromTeam
 *
 * Author        : Richard May
 *
 * Description   : Converts from a project-specific team into a gamespy 
 *				   Team ID. Gamespy Team IDs must be contiguous in numbering,
 *				   so they can't be generated directly in any guaranteed fashion
 *
 *************************************************************/
u_int Asura_GameSpy_StatLogging::GetTeamIDFromTeam ( int iTeam )
{
	// search the Team guid storage
	u_int u=0;
	for( ; u<s_uTeamIDStoreSize; ++u )
	{
		if( s_pxTeamIDStore[u].iTeam == iTeam )
		{
			return s_pxTeamIDStore[u].uGSTeamID;
		}
		else if( s_pxTeamIDStore[u].iTeam == ASURA_STATLOG_NOTEAM )
		{
			break;
		}
	}

	// if we've reached here, guid isn't in  the storage so allocate new id.
	if( u>=s_uTeamIDStoreSize )
	{
		// allocate more space
		TeamAndTeamID* pxTempStore = s_pxTeamIDStore;
		u_int uOldSize = s_uTeamIDStoreSize;
		
		s_uTeamIDStoreSize *= 2;
		s_pxTeamIDStore = new TeamAndTeamID[s_uTeamIDStoreSize];

		for( u_int v=0; v<s_uTeamIDStoreSize; ++v )
		{
			if( v < uOldSize )
			{
				s_pxTeamIDStore[v] = pxTempStore[v];
			}
			else
			{
				s_pxTeamIDStore[u].iTeam = ASURA_STATLOG_NOTEAM;
			}
		}

		delete [] pxTempStore;
	}

	s_pxTeamIDStore[u].iTeam = iTeam;
	s_pxTeamIDStore[u].uGSTeamID = s_uNextTeamID++; // increment after assignment

	return s_pxTeamIDStore[u].uGSTeamID;
}
/* end of function GetTeamIDFromGuid */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::InitPlayerIDStorage
 *
 * Author        : Richard May
 *
 * Description   : Sets up the player ID storage
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::InitPlayerIDStorage ( void )
{		
	delete [] s_pxPlayerIDStore; // handles null case internally (or should, according to ANSI C standards)

	s_uNextPlayerID		 = 0;
	s_uPlayerIDStoreSize = 8;

	s_pxPlayerIDStore = new GuidAndPlayerID[s_uPlayerIDStoreSize];

	for( u_int u=0; u<s_uPlayerIDStoreSize; ++u )
	{
		s_pxPlayerIDStore[u].uGuid = ASURA_GUID_UNREGISTERED;
		s_pxPlayerIDStore[u].uGSPlayerID = UINT_MAX;
		s_pxPlayerIDStore[u].bIsGuid = true;
	}
}
/* end of function InitPlayerIDStorage */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::InitTeamIDStorage
 *
 * Author        : Richard May
 *
 * Description   : Sets up the Team ID storage
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::InitTeamIDStorage ( void )
{		
	delete [] s_pxTeamIDStore; // handles null case internally (or should, according to ANSI C standards)

	s_uNextTeamID	   = 0;
	s_uTeamIDStoreSize = 8;

	s_pxTeamIDStore = new TeamAndTeamID[s_uTeamIDStoreSize];

	for( u_int u=0; u<s_uTeamIDStoreSize; ++u )
	{
		s_pxTeamIDStore[u].iTeam = ASURA_STATLOG_NOTEAM;
	}
}
/* end of function InitTeamIDStorage */


/*************************************************************
 *
 * Name          : Asura_GameSpy_StatLogging::GetPlayerAuthentification
 *
 * Author        : Richard May
 *
 * Description   : Gets a gamespy authentifcation code for the player
 *
 *************************************************************/
void Asura_GameSpy_StatLogging::GetPlayerAuthentification ( char* szAuth )
{
	// as the GP only supports one connection, we'll just access that for 
	// now. If this were a proper client/server setup, we'd need to handle
	// this better

	GenerateAuth( GetChallenge(NULL), const_cast<char*>(Asura_GameSpy_GP::GetCurrentProfilePassword()), szAuth );
}
/* end of function GetPlayerAuthentification */



#endif // ASURA_USE_NETWORK_STATLOGGING
#endif //#ifdef ASURA_SUPPORT_GAMESPY
/****************/
/* End of file. */
/****************/

