//////////////////////////////////////////////////////////////////////////////
//
// Asura_Leaderboards.h
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     game- and platform- independent interface for leaderboards
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_LEADERBOARDS_H___
#define ASURA_LEADERBOARDS_H___ 1

#ifdef ASURA_SUPPORT_LEADERBOARDS

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uASURA_LEADERBOARD_INVALID_NUM_ENTRIES = 0xffffffff;
static const u_int uASURA_LEADERBOARD_MAX_ROWS_IN_MEMORY = 100;
static const u_int uASURA_LEADERBOARD_MAX_COLUMNS = 7;

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Leaderboard;
class Asura_Leaderboard_Row;

// E N U M S /////////////////////////////////////////////////////////////////
enum ASURA_LB_DATAFORMAT
{
	ASURA_LB_DATAFORMAT_UINT,
	ASURA_LB_DATAFORMAT_FLOAT,
};

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////
class Asura_Leaderboard_System
{
public:

	static void AddConsoleCmdsAndVars();

	// init/update/shutdown functions that double up as console commands
	static bool Initialise( bool bHelp = false );
	static bool Update( bool bHelp = false );
	static bool Shutdown( bool bHelp = false );

	static bool ConsoleCmd_GetNumEntries( bool bHelp );
	static bool ConsoleCmd_IsDataPending( bool bHelp );
	static bool ConsoleCmd_ResetErrorFlag( bool bHelp );
	static bool SetCurrentLeaderboard( u_int uIndex, bool bInterrogate, bool bHelp );
	static bool ConsoleCmd_ToggleFriendsOnly( bool bHelp );
	static bool ConsoleCmd_PageUp( bool bHelp );
	static bool ConsoleCmd_PageDown( bool bHelp );

	// console cmds take input from 0 to (numrows-1)
	// projects can access these commands directly so they can be added with meaningful project-specific names.
	static bool ConsoleCmd_GetRank( u_int uIndex, bool bInterrogate, bool bHelp );
	static bool ConsoleCmd_GetName( u_int uIndex, bool bInterrogate, bool bHelp );
	// Get col command is templated for col0, col1,... colN
	template<typename T> static bool ConsoleCmd_GetCol( u_int uIndex, bool bInterrogate, bool bHelp );
		   
	// project-supplied functions - public access
	static u_int Project_GetNumLeaderboards();
	static u_int Project_GetNumLeaderboardColumns(u_int uLeaderboard );
	static u_int Project_GetRowsPerPage();
	static ASURA_LB_DATAFORMAT Project_GetDataFormat(u_int uLeaderboard, u_int uColumn);
	// Project_GetDataFormatString should return a printf-style specifier for column e.g. "%.02f", "%.f"
	static const char* Project_GetDataFormatString(u_int uLeaderboard, u_int uColumn);	

private:

	static void Platform_Initialise();
	static void Platform_Shutdown();

	static void PrintColumnToConsole( u_int uRow, u_int uColumn );

	// we maintain multiple leaderboards simultaneously.
	static u_int s_uCurrentLeaderboard;
	static Asura_Leaderboard** s_ppxLeaderboards;
	static Asura_Leaderboard** s_ppxFriendsLeaderboards;
	static Asura_Leaderboard* s_pxLeaderboard;
	static bool s_bError;
	static bool s_bFriendsOnly;
};

class Asura_Leaderboard
{
public:

	Asura_Leaderboard( u_int uLeaderboardIndex, bool bFriendsOnly );
	~Asura_Leaderboard();

	void Update();
	bool HasErrorOccurred() const { return m_bError; }
	void ResetErrorFlag() { m_bError = false; }
	bool IsCurrentPagePending() const { return !m_bDataValid; }	// returns true if current page isn't available yet
	u_int GetNumAvailableRowsInPage() const;

	// functions to change focus position
	u_int GetTotalEntries() const { return m_uTotalEntries; } // returns uASURA_LEADER_BOARD_INVALID_NUM_ENTRIES if not known.
	u_int GetViewPosition() const { return m_uFirstRow + m_uViewOffset; }
	
	void SetViewPosition( u_int uPos );
	void PageUp();
	void PageDown();

	// TODO:
	//void FindLoggedInUser();

	// stats data access
	const Asura_UnicodeChar* GetName(u_int uRowOffset) const;
	u_int GetRank(u_int uRowOffset) const;
	float GetData(u_int uRowOffset, u_int uColumn) const;	// offset is from current view position
	
private:

	void SetData(u_int uRow, u_int uColumn, float fData);
	// NY - 24 April 2009 - versions to set name for different source encodings - refered to by actual encoding for clarity, since two encodings are stored as c string
	// ASCII - normal c string, UCS2 - normal wide string, UTF8 - unicode in a c string, variable length encoding
	void SetASCIINameAndRank(u_int uRow, const char* szName, u_int uRank);
	void SetUCS2NameAndRank(u_int uRow, const Asura_UnicodeChar* wszName, u_int uRank);
	void SetUTF8NameAndRank(u_int uRow, const char* szName, u_int uRank);
	void SortData();

	void Platform_Initialise();
	void Platform_Shutdown();
	bool Platform_Update();		// should return true when a pending operation completes, may set m_bError.
	bool Platform_FetchData(); // requests data to fill current view, returns true if request is successful

	const bool m_bFriendsOnly;
	const u_int m_uLeaderboardIndex;
	u_int m_uNumColumns;

	Asura_Leaderboard_Row* m_pxData;

	u_int m_uNumRows;		// number of rows we have in memory
	u_int m_uFirstRow;		// index of first row we have (or have requested) in memory
	u_int m_uViewOffset;	// offset from first row in memory to first displayed row
	u_int m_uTotalEntries;	// total number of rows stored on the server

	bool m_bDataValid;
	bool m_bError;
	bool m_bFetchInProgress;
	bool m_bFetchQueued;

	void* m_pPlatformData;
};


#endif // #ifdef ASURA_SUPPORT_LEADERBOARDS

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ASURA_LEADERBOARDS_H___
