//////////////////////////////////////////////////////////////////////////////
//
// Asura_Merit.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     merit system to rank player skill in online games
//     See Wiki documentation at 
//     http://asurawiki.rebellion.co.uk/twiki/bin/view/Programming/MeritSystem
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_MERIT_SYSTEM

// This file's header.
#include "Network/Online/Asura_Merit.h"

// Asura includes
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_NetTransport_System.h"

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uASURA_MERIT_SYSTEM_INVALID_TEAM = 0xffffffff;
static const int iASURA_MERIT_MAX_POINTS_PER_ROUND = 150;

// I N T E R N A L   C L A S S E S ///////////////////////////////////////////
class Asura_Merit_Player
{
public:
	Asura_Merit_Player()
	{
		Set(ASURA_GUID_SERVER, uASURA_MERIT_SYSTEM_INVALID_TEAM, 0, 0, 0, false);

		m_iExpectedScore = 0;
		m_iRankingScore = 0;
		m_pExtraData = NULL;
	}

	~Asura_Merit_Player()
	{
		Asura_Deallocate(m_pExtraData);
	}

	void Set(Asura_Guid uGuid, u_int uTeam, int iScore, u_int uMeritLevel, int iMeritPoints, bool bDisconnected)
	{
		m_uGuid = uGuid;
		m_uTeam = uTeam;
		m_iScore = iScore;
		m_uMeritLevel = uMeritLevel;
		m_iMeritLevelChange = 0;
		m_iMeritPoints = iMeritPoints;
		m_iMeritPointsChange = 0;
		m_bDisconnected = bDisconnected;
	}

	void AllocateExtraData(size_t uSize)
	{
		Asura_Assert(!m_pExtraData, "memory leak");
		m_pExtraData = Asura_Allocate(uSize);
	}

	const void* GetExtraData() const
	{
		return m_pExtraData;
	}

	void SetExtraData(const void* pSource, size_t uSize) const
	{
		if (pSource)
		{
			Asura_Assert(m_pExtraData, "extra data not allocated");
			Asura_Mem::Copy(m_pExtraData, pSource, uSize);
		}
	}

	Asura_Guid m_uGuid;
	u_int m_uTeam;
	int m_iScore;
	u_int m_uMeritLevel;
	int m_iMeritLevelChange;
	int m_iMeritPoints;
	int m_iMeritPointsChange;
	int m_iExpectedScore;
	int m_iRankingScore;
	bool m_bDisconnected;

private:
	void* m_pExtraData;
};

class Asura_Merit_Team
{
public:
	Asura_Merit_Team()
	{
		m_uNumPlayers = 0;
		m_iTotalScore = 0;
		m_iTotalMeritPoints = 0;
	}

	void Set(u_int uNumPlayers, int iTotalScore, int iTotalMeritPoints)
	{
		m_uNumPlayers = uNumPlayers;
		m_iTotalScore = iTotalScore;
		m_iTotalMeritPoints = iTotalMeritPoints;
	}

	u_int m_uNumPlayers;
	int m_iTotalScore;
	int m_iTotalMeritPoints;
};

// G L O B A L S /////////////////////////////////////////////////////////////
static const int g_aiMaxMeritPoints[uASURA_MERIT_NUM_LEVELS] = {
	 150,  300,  450,  600,  750,
	1000, 1250, 1500, 1750, 2000,
	2500, 3000, 3500, 4000, 4500,
	5000, 6000, 7000, 8000, 10000 
};

static const int g_aiLossReduction[uASURA_MERIT_NUM_LEVELS] = {
	  0,   0,   0,  10,  20,
	 40,  80, 100, 100, 100,
	100, 100, 100, 100, 100, 
	100, 100, 100, 100, 100
};

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Merit_Calculator* Asura_Merit_System::s_pxCalculator;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////

// Asura_Merit ///////////////////////////////////////////////////////////////
void Asura_Merit::SetMeritLevel( u_int uMeritLevel )
{
	Asura_Assert(uMeritLevel < uASURA_MERIT_NUM_LEVELS, "invalid merit level");
	ASURA_STATIC_ASSERT(uASURA_MERIT_NUM_LEVELS <= UCHAR_MAX);
	m_ucMeritLevel = static_cast<u_char>(Asura_Maths::ClampToRange( uMeritLevel, 0u, uASURA_MERIT_NUM_LEVELS - 1 ));
}

void Asura_Merit::SetMeritPoints( int iMeritPoints )
{ 
	Asura_Assert(iMeritPoints >= 0 && iMeritPoints <= iASURA_MERIT_MAX_MERIT_POINTS, "invalid merit points");
	ASURA_STATIC_ASSERT(iASURA_MERIT_MAX_MERIT_POINTS <= USHRT_MAX);
	m_usMeritPoints = static_cast<u_short>(Asura_Maths::ClampToRange( iMeritPoints, 0, iASURA_MERIT_MAX_MERIT_POINTS ));
}

void Asura_Merit::ChangeNetworkByteOrder()
{
	ASURA_NTS_CHANGE_BYTE_ORDER(m_usMeritPoints);
	ASURA_NTS_CHANGE_BYTE_ORDER(m_ucMeritLevel);
}

// Asura_Merit_System ///////////////////////////////////////////////////////
void Asura_Merit_System::Initialise(u_int uMaxPlayers, u_int uMaxTeams, size_t uExtraDataSize)
{
	Asura_Assert(!s_pxCalculator, "memory leak");
	s_pxCalculator = new Asura_Merit_Calculator(uMaxPlayers, uMaxTeams, uExtraDataSize);
}

void Asura_Merit_System::ComputeStats()
{
	Asura_Assert(s_pxCalculator, "merit system not initialised correctly");
	if (!s_pxCalculator)
	{
		return;
	}

	// adjust scores for the winning team
	if (s_pxCalculator->IsTeamGame())
	{
		u_int uWinningTeam = 0;
		int iBonus = 0;
		if (Project_GetWinningTeam(uWinningTeam, iBonus))
		{
			Asura_Assert(uWinningTeam < s_pxCalculator->GetNumTeams(), "invalid winning team");
		}
		s_pxCalculator->FinaliseScores(uWinningTeam, iBonus);
	}
	s_pxCalculator->Process();
}

bool Asura_Merit_System::HasComputedStats()
{
	return s_pxCalculator && s_pxCalculator->HasDoneProcessing();
}

void Asura_Merit_System::Shutdown()
{
	delete s_pxCalculator;
	s_pxCalculator = NULL;
}

void Asura_Merit_System::StorePlayerStats( Asura_Guid uGuid, const Asura_Merit& xCurrentMerit, 
										  bool bDisconnectedEarly, const void* pExtraData )
{
	if (s_pxCalculator)
	{
		const u_int uPlayerIndex = s_pxCalculator->GetNumPlayers();
		u_int uTeam = 0;
		int iScore = 0;
		bool bIsTeamGame = s_pxCalculator->IsTeamGame();
		if (Project_GetPlayerData( uGuid, bDisconnectedEarly, uTeam, iScore ))
		{
			if (!bIsTeamGame)
			{
				uTeam = uPlayerIndex;
			}
			s_pxCalculator->SetPlayerData(uPlayerIndex, uGuid, uTeam, iScore, xCurrentMerit.GetMeritLevel(), 
											xCurrentMerit.GetMeritPoints(), bDisconnectedEarly, pExtraData);
		}
		else
		{
			Asura_Assert(false, "failed to get project data for merit system");
		}
	}
}

bool Asura_Merit_System::GetPlayerMerit(Asura_Guid uGuid, u_int& uMeritLevel, int& iMeritPoints, 
						   int& iMeritLevelChange, int& iMeritPointsChange)
{
	Asura_Assert(s_pxCalculator && s_pxCalculator->HasDoneProcessing(), "merit system not initialised correctly");
	if (s_pxCalculator && s_pxCalculator->HasDoneProcessing())
	{
		u_int uPlayerIndex = 0;
		if (s_pxCalculator->FindPlayerIndexFromGuid(uGuid, uPlayerIndex))
		{
			s_pxCalculator->GetPlayerMerit(uPlayerIndex, uMeritLevel, iMeritPoints);
			s_pxCalculator->GetPlayerMeritChange(uPlayerIndex, iMeritLevelChange, iMeritPointsChange);
			return true;
		}
	}
	Asura_Assert(false, "merit system failed to get player merit");
	iMeritLevelChange = 0;
	iMeritPointsChange = 0;
	return false;
}

const void* Asura_Merit_System::GetPlayerExtraData(Asura_Guid uGuid)
{
	Asura_Assert(s_pxCalculator, "merit system not initialised correctly");
	if (s_pxCalculator )
	{
		u_int uPlayerIndex = 0;
		if (s_pxCalculator->FindPlayerIndexFromGuid(uGuid, uPlayerIndex))
		{
			return s_pxCalculator->GetPlayerExtraData(uPlayerIndex);
		}
	}
	Asura_Assert(false, "merit system failed to find player");
	return NULL;
}


// Asura_Merit_Calculator ///////////////////////////////////////////////////
Asura_Merit_Calculator::Asura_Merit_Calculator(u_int uNumPlayers, 
											   u_int uNumTeams,
											   size_t uExtraDataSize)
: m_uMaxPlayers(uNumPlayers)
, m_uNumTeams(uNumTeams)
, m_uNumPlayers(0)
, m_iTotalMeritPoints(0)
, m_iTotalScore(0)
, m_iLargestGain(0)
, m_iTotalGain(0)
, m_iTotalLoss(0)
, m_bDoneProcessing(false)
, m_uExtraDataSize(uExtraDataSize)
{
	m_pxPlayers = new Asura_Merit_Player[uNumPlayers];
	for (u_int u = 0; u < uNumPlayers; ++u)
	{
		m_pxPlayers[u].AllocateExtraData(uExtraDataSize);
	}
	Asura_Assert(uNumTeams > 1, "project code should not allow ranked matches with only 1 team");
	Asura_Assert(uNumTeams <= uNumPlayers, "invalid number of teams specified");
	m_pxTeams = new Asura_Merit_Team[uNumTeams];
}

Asura_Merit_Calculator::~Asura_Merit_Calculator()
{
	delete[] m_pxTeams;
	delete[] m_pxPlayers;
}

void Asura_Merit_Calculator::SetPlayerData(u_int uPlayerIndex, Asura_Guid uGuid, u_int uTeam, 
										   int iScore, u_int uMeritLevel, int iMeritPoints, 
										   bool bDisconnectedEarly, const void* pExtraData)
{
	Asura_Assert(uPlayerIndex == m_uNumPlayers, "expecting players to be added in order");
	Asura_Assert(m_uNumPlayers < m_uMaxPlayers, "added more players than set in constructor");
	if (m_uNumPlayers >= m_uMaxPlayers)
	{
		return;
	}

	Asura_Assert(uTeam < m_uNumTeams, "invalid team for player");

	const u_int uIndex = m_uNumPlayers++;
	m_pxPlayers[uIndex].Set(uGuid, uTeam, iScore, uMeritLevel, iMeritPoints, bDisconnectedEarly);
	if (pExtraData)
	{
		m_pxPlayers[uIndex].SetExtraData(pExtraData, m_uExtraDataSize);
	}
}

void Asura_Merit_Calculator::CalculateTeamScores()
{
	u_int uPlayersProcessed = 0;
	// loop over teams
	for (u_int uTeam = 0; uTeam < m_uNumTeams; ++uTeam)
	{
		u_int uNumPlayersOnTeam = 0;
		int iTeamScore = 0;
		int iTeamMeritPoints = 0;
		for (u_int uPlayer = 0; uPlayer < m_uNumPlayers; ++uPlayer)
		{
			if (m_pxPlayers[uPlayer].m_uTeam == uTeam)
			{
				++uNumPlayersOnTeam;
				iTeamScore += m_pxPlayers[uPlayer].m_iScore;
				iTeamMeritPoints += m_pxPlayers[uPlayer].m_iMeritPoints;
			}
		}
		m_pxTeams[uTeam].Set(uNumPlayersOnTeam, iTeamScore, iTeamMeritPoints);

		m_iTotalMeritPoints += iTeamMeritPoints;
		m_iTotalScore += iTeamScore;
		uPlayersProcessed += uNumPlayersOnTeam;
	}

	Asura_Assert(uPlayersProcessed == m_uNumPlayers, "didn't process all players");
}

void Asura_Merit_Calculator::Process()
{
	// can't make these static asserts because of the array accesses
	Asura_Assert(iASURA_MERIT_MAX_MERIT_POINTS == g_aiMaxMeritPoints[uASURA_MERIT_NUM_LEVELS-1], "shouldn't change");
	Asura_Assert(iASURA_MERIT_MAX_POINTS_PER_ROUND == g_aiMaxMeritPoints[0], "shouldn't change");

	Asura_Assert(!m_bDoneProcessing, "already done the processing for this calculator");
	Asura_Assert(m_uNumPlayers == m_uMaxPlayers, "didn't get data for all players");

	// compute scores for each team.
	CalculateTeamScores();

	// set ranking score for each player to be (actualscore - expectedscore) 
	// where expected score = (merit/(merit + other team merit)) * (score + other team score)
	// expected score is just averaged if total merit is less than a fixed amount.
	// also find largest gain and loss in score
	for (u_int uPlayer = 0; uPlayer < m_uNumPlayers; ++uPlayer)
	{
		const u_int uTeam = m_pxPlayers[uPlayer].m_uTeam;
		Asura_Assert(uTeam < m_uNumTeams, "invalid team");
		const int iTeamMeritPoints = m_pxTeams[uTeam].m_iTotalMeritPoints;
		const int iTeamScore = m_pxTeams[uTeam].m_iTotalScore;
		const u_int uTeamSize = m_pxTeams[uTeam].m_uNumPlayers;

		// if this is a non-team game then we set the other team stats manually.
		int iTotalMeritPointsOfPlayersOnOtherTeams = m_iTotalMeritPoints - iTeamMeritPoints;
		int iTotalScoreOfPlayersOnOtherTeams = m_iTotalScore - iTeamScore;
		int iNumPlayersOnOtherTeams = m_uNumPlayers - uTeamSize;

		// update rankings
		{
			const int iPlayerMerit = m_pxPlayers[uPlayer].m_iMeritPoints;
			const int iTotalMerit = iPlayerMerit + iTotalMeritPointsOfPlayersOnOtherTeams;
			const int iTotalPlayers = 1 + iNumPlayersOnOtherTeams;
			const int iTotalScore = m_pxPlayers[uPlayer].m_iScore + iTotalScoreOfPlayersOnOtherTeams;
			const int iExpectedScore = (iTotalMerit > (iASURA_MERIT_MAX_POINTS_PER_ROUND * iTotalPlayers)) ? 
				(iPlayerMerit * iTotalScore / iTotalMerit) : 
				(iTotalScore / iTotalPlayers);

			m_pxPlayers[uPlayer].m_iExpectedScore = iExpectedScore;

			const int iRankingScore = m_pxPlayers[uPlayer].m_iScore - iExpectedScore;
			m_pxPlayers[uPlayer].m_iRankingScore = iRankingScore;
			if (iRankingScore > 0)
			{
				if (iRankingScore > m_iLargestGain)
				{
					m_iLargestGain = iRankingScore;
				}
				m_iTotalGain += iRankingScore;
			}
			else
			{
				const int iRankingScoreLoss = -iRankingScore;
				m_iTotalLoss += iRankingScoreLoss;
			}
		}
	}

	// now work out the merit changes from the ranking scores and largest gain and apply them
	for (u_int uPlayer = 0; uPlayer < m_uNumPlayers; ++uPlayer)
	{
		int iMeritChange;
		const int iScore = m_pxPlayers[uPlayer].m_iRankingScore;

		if ( (m_iLargestGain > 0) && (m_iTotalLoss > 0) )
		{
			if (iScore > 0)
			{
				// largest positive score gets iASURA_MERIT_MAX_POINTS_PER_ROUND (150), other positive scores get
				// a fraction of that.
				iMeritChange = iASURA_MERIT_MAX_POINTS_PER_ROUND * iScore / m_iLargestGain;
			}
			else
			{
				// negative scores are scaled to make the game zero-sum.
				// after loss reduction is applied, the game is no longer zero-sum.
				// Also, max loss is clamped to -150; this also prevents the game being zero-sum.
				iMeritChange = iASURA_MERIT_MAX_POINTS_PER_ROUND * iScore * m_iTotalGain / (m_iLargestGain * m_iTotalLoss);
				const u_int uPlayerLevel = m_pxPlayers[uPlayer].m_uMeritLevel;
				const int iLossReductionPercent = g_aiLossReduction[uPlayerLevel];
				iMeritChange = (iMeritChange * iLossReductionPercent) / 100;
				iMeritChange = Asura_Maths::Max(iMeritChange, -iASURA_MERIT_MAX_POINTS_PER_ROUND);
			}
		}
		else
		{
			iMeritChange = 0;
		}

		const u_int uOldLevel = m_pxPlayers[uPlayer].m_uMeritLevel;
		const int iOldMerit = m_pxPlayers[uPlayer].m_iMeritPoints;
		int iNewMerit = iOldMerit + iMeritChange;

		// clamp merit at zero and prevent it going above 10000
		iNewMerit = Asura_Maths::ClampToRange(iNewMerit, 0, iASURA_MERIT_MAX_MERIT_POINTS);

		// convert merit to level
		u_int uNewLevel = GetLevelFromMerit(iNewMerit, uOldLevel);

		// clamp merit at the bottom of the level for the first 5 ranks so you can't get demoted.
		if (uNewLevel < 5 && iNewMerit < GetMinMeritPointsForLevel(uNewLevel))
		{
			iNewMerit = GetMinMeritPointsForLevel(uNewLevel);
		}

#ifdef ASURA_LOGFILES_ENABLED
		// record the change to the debug log
		Asura_OutputToDebugger(("%u: team %u, expected %i, actual %i, new merit level %u, new merit points %i (change = %i)\n",
			uPlayer, m_pxPlayers[uPlayer].m_uTeam, m_pxPlayers[uPlayer].m_iExpectedScore,
			m_pxPlayers[uPlayer].m_iScore, uNewLevel, iNewMerit, iNewMerit - iOldMerit));
#endif

		// and store in the player data
		m_pxPlayers[uPlayer].m_uMeritLevel = uNewLevel;
		m_pxPlayers[uPlayer].m_iMeritLevelChange = static_cast<int>(uNewLevel) - static_cast<int>(uOldLevel);

		m_pxPlayers[uPlayer].m_iMeritPoints = iNewMerit;
		m_pxPlayers[uPlayer].m_iMeritPointsChange = iNewMerit - iOldMerit;
	}

	m_bDoneProcessing = true;
}


bool Asura_Merit_Calculator::FindPlayerIndexFromGuid( Asura_Guid uGuid, u_int& uPlayerIndex ) const
{
	for (u_int uPlayer = 0; uPlayer < m_uNumPlayers; ++uPlayer)
	{
		if (m_pxPlayers[uPlayer].m_uGuid == uGuid)
		{
			uPlayerIndex = uPlayer;
			return true;
		}
	}

	uPlayerIndex = 0;
	return false;
}

void Asura_Merit_Calculator::GetPlayerMerit(u_int uPlayerIndex, u_int& uMeritLevel, int& iMeritPoints) const
{
	Asura_Assert(m_bDoneProcessing, "GetPlayerMerit() called before processing done");
	Asura_Assert(uPlayerIndex < m_uNumPlayers, "invalid player index");

	if (m_bDoneProcessing && uPlayerIndex < m_uNumPlayers)
	{
		Asura_Assert(m_pxPlayers, "no player array");
		uMeritLevel = m_pxPlayers[uPlayerIndex].m_uMeritLevel;
		iMeritPoints = m_pxPlayers[uPlayerIndex].m_iMeritPoints;
	}
	else
	{
		uMeritLevel = 0;
		iMeritPoints = 0;
	}
}

void Asura_Merit_Calculator::GetPlayerMeritChange(u_int uPlayerIndex, int& iMeritLevelChange, int& iMeritPointsChange) const
{
	Asura_Assert(m_bDoneProcessing, "GetPlayerMerit() called before processing done");
	Asura_Assert(uPlayerIndex < m_uNumPlayers, "invalid player index");

	if (m_bDoneProcessing && uPlayerIndex < m_uNumPlayers)
	{
		Asura_Assert(m_pxPlayers, "no player array");
		iMeritLevelChange = m_pxPlayers[uPlayerIndex].m_iMeritLevelChange;
		iMeritPointsChange = m_pxPlayers[uPlayerIndex].m_iMeritPointsChange;
	}
	else
	{
		iMeritLevelChange = 0;
		iMeritPointsChange = 0;
	}
}

const void* Asura_Merit_Calculator::GetPlayerExtraData(u_int uPlayerIndex) const
{
	Asura_Assert(uPlayerIndex < m_uNumPlayers, "invalid player index");
	if (uPlayerIndex < m_uNumPlayers)
	{
		return m_pxPlayers[uPlayerIndex].GetExtraData();
	}
	return NULL;
}

bool Asura_Merit_Calculator::IsPlayerDisconnected(u_int uPlayerIndex) const
{
	Asura_Assert(uPlayerIndex < m_uNumPlayers, "invalid player index");
	if (uPlayerIndex < m_uNumPlayers)
	{
		return m_pxPlayers[uPlayerIndex].m_bDisconnected;
	}
	return false;
}

Asura_Guid Asura_Merit_Calculator::GetPlayerGuid(u_int uPlayerIndex) const
{
	Asura_Assert(uPlayerIndex < m_uNumPlayers, "invalid player index");
	if (uPlayerIndex < m_uNumPlayers)
	{
		return m_pxPlayers[uPlayerIndex].m_uGuid;
	}
	return false;
}

int Asura_Merit_Calculator::GetMinMeritPointsForLevel(u_int uMeritLevel)
{
	if (uMeritLevel == 0)
	{
		return 0;
	}
	return g_aiMaxMeritPoints[uMeritLevel - 1];
}

u_int Asura_Merit_Calculator::GetLevelFromMerit(int iMeritPoints, u_int uCurMeritLevel)
{
	for (u_int uLevel = 0; uLevel < uASURA_MERIT_NUM_LEVELS; ++uLevel)
	{
		if (iMeritPoints < g_aiMaxMeritPoints[uLevel])
		{
			// is this level lower than current?
			if (uLevel < uCurMeritLevel)
			{
				// you can't drop below the first 5 levels 
				if (uCurMeritLevel < 5)
				{
					return uCurMeritLevel;
				}
				// need to drop below halfway to previous
				int iHalfDiff;
				if (uLevel > 0)
				{
					iHalfDiff = (g_aiMaxMeritPoints[uLevel+1] - g_aiMaxMeritPoints[uLevel])/2;
				}
				else
				{
					iHalfDiff = (g_aiMaxMeritPoints[uLevel] / 2);
				}
				if ( iMeritPoints >= (g_aiMaxMeritPoints[uLevel] - iHalfDiff) )
				{
					uLevel++;
				}
			}
			return uLevel;
		}
	}

	return uASURA_MERIT_NUM_LEVELS - 1;
}

void Asura_Merit_Calculator::FinaliseScores( u_int uWinningTeam, int iBonus )
{
	if (!iBonus)
	{
		return;
	}
	Asura_Assert(IsTeamGame(), "should be no winning team bonus");
	Asura_Assert(uWinningTeam < m_uNumTeams, "invalid team");

	for (u_int uPlayer = 0; uPlayer < m_uNumPlayers; ++uPlayer)
	{
		const u_int uTeam = m_pxPlayers[uPlayer].m_uTeam;
		if (uTeam == uWinningTeam)
		{
			m_pxPlayers->m_iScore += iBonus;
		}
	}
}

// commented-out code to test merit system - could be turned into a unit test or just deleted eventually
//void TestMeritSystem()
//{
//	const u_int uNUMTESTS = 6;
//	const u_int uMAXPLAYERS = 8;
//	int aiPlayers[uNUMTESTS]					=	{1, 2, 2, 4, 8, 8};
//	int aiTeams[uNUMTESTS]						=	{1, 1, 2, 2, 8, 3};
//	int aiPlayerTeam[uNUMTESTS][uMAXPLAYERS]	=  {{0, -1, -1, -1, -1, -1, -1, -1}, 
//	{0,  0, -1, -1, -1, -1, -1, -1}, 
//	{0,  1, -1, -1, -1, -1, -1, -1}, 
//	{0,  1,  1,  0, -1, -1, -1, -1}, 
//	{0,  1,  2,  3,  4,  5,  6,  7}, 
//	{0,  1,  2,  0,  1,  2,  0,  1}  };
//	for (u_int u = 0; u < uNUMTESTS; ++u)
//	{
//		const u_int uNumPlayers = (u_int)aiPlayers[u];
//		const u_int uTeams = (u_int)aiTeams[u];
//		Asura_Merit_Calculator xCalculator(uNumPlayers, uTeams);
//		Asura_Guid uGuid = 1000;
//		for (u_int uPlayer = 0; uPlayer < uNumPlayers; ++uPlayer)
//		{
//			u_int uTeam = (u_int)aiPlayerTeam[u][uPlayer];
//			int iScore = Asura_Random::GetInt(-4, 50);
//			int iMeritPoints = Asura_Random::GetInt(0, iASURA_MERIT_MAX_MERIT_POINTS);
//			u_int uMeritLevel = Asura_Merit_Calculator::GetLevelFromMerit(iMeritPoints, 0);
//			xCalculator.SetPlayerData(uPlayer, uGuid++, uTeam, iScore, uMeritLevel, iMeritPoints);
//		}
//
//		xCalculator.Process();
//
//		for (u_int uPlayer = 0; uPlayer < uNumPlayers; ++uPlayer)
//		{
//			int iNewMerit;
//			u_int uNewLevel;
//			xCalculator.GetPlayerMerit(uPlayer, uNewLevel, iNewMerit);
//			Asura_OutputToDebugger(("%u: %u %i\n", uPlayer, uNewLevel, iNewMerit));
//		}
//	}
//}

#endif // #ifdef ASURA_USE_MERIT_SYSTEM
