//////////////////////////////////////////////////////////////////////////////
//
// Asura_Merit.h
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     merit system to rank player skill in online games
//     See Wiki documentation at 
//     http://asurawiki.rebellion.co.uk/twiki/bin/view/Programming/MeritSystem
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_MERIT_H___
#define ASURA_MERIT_H___ 1

#ifdef ASURA_USE_MERIT_SYSTEM

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Network/Online/Asura_Merit_Types.h"

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Merit_Player;
class Asura_Merit_Team;
class Asura_Merit_Calculator;

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////
class Asura_Merit_System
{
public:

	static void Initialise(u_int uMaxPlayers, u_int uMaxTeams, size_t uExtraDataSize);
	static void StorePlayerStats( Asura_Guid uGuid, const Asura_Merit& xCurrentMerit, 
									bool bDisconnectedEarly, const void* pExtraData );
	static void ComputeStats();
	static bool HasComputedStats();
	static void Shutdown();

	// GetPlayerMerit() allows you to retrieve the new values and the change in merit scores 
	// for any player after ComputeStats() has been called and before Shutdown() has been called. 
	// It returns false if the player could not be found with the input guid and sets the merit level, 
	// points and changes to zero in that case.
	static bool GetPlayerMerit(Asura_Guid uGuid, u_int& uMeritLevel, int& iMeritPoints, 
								int& iMeritLevelChange, int& iMeritPointsChange);

	static const void* GetPlayerExtraData(Asura_Guid uGuid);

	// by getting the calculator you can iterate over all players
	static const Asura_Merit_Calculator* GetCalculator() { return s_pxCalculator; }

private:
	static bool Project_GetPlayerData( Asura_Guid uGuid, bool bDisconnectedEarly, u_int& uTeam, int& iScore );
	static bool Project_GetWinningTeam( u_int& uTeam, int& iBonus );

	static Asura_Merit_Calculator* s_pxCalculator;
};

class Asura_Merit_Calculator
{
public:

	// for non-team-based games specify uNumPlayers for uNumTeams
	Asura_Merit_Calculator(u_int uNumPlayers, u_int uNumTeams, size_t uExtraDataSize);
	~Asura_Merit_Calculator();

	// for non-team-based games specify uPlayerIndex for uTeam
	void SetPlayerData(u_int uPlayerIndex, Asura_Guid uGuid, u_int uTeam, int iScore, u_int uMeritLevel, 
						int iMeritPoints, bool bDisconnectedEarly, const void* pExtraData);
	void Process();
	bool FindPlayerIndexFromGuid( Asura_Guid uGuid, u_int& uPlayerIndex ) const;
	void GetPlayerMerit(u_int uPlayerIndex, u_int& uMeritLevel, int& iMeritPoints) const;
	void GetPlayerMeritChange(u_int uPlayerIndex, int& iMeritLevelChange, int& iMeritPointsChange) const;
	const void* GetPlayerExtraData(u_int uPlayerIndex) const;
	bool IsPlayerDisconnected(u_int uPlayerIndex) const;
	Asura_Guid GetPlayerGuid(u_int uPlayerIndex) const;
	u_int GetNumPlayers() const { return m_uNumPlayers; }
	u_int GetNumTeams() const { return m_uNumTeams; }
	bool IsTeamGame() const { return m_uMaxPlayers != m_uNumTeams; }
	bool HasDoneProcessing() const { return m_bDoneProcessing; }
	void FinaliseScores( u_int uWinningTeam, int iBonus );

	// static functions
	static u_int GetLevelFromMerit(int iMeritPoints, u_int uCurMeritLevel);
	static int GetMinMeritPointsForLevel(u_int uMeritLevel);

	// remember to update the Wiki documentation with any changes that affect 
	// how projects use the system.

private:

	void CalculateTeamScores();

	Asura_Merit_Player* m_pxPlayers;
	Asura_Merit_Team* m_pxTeams;
	const u_int m_uMaxPlayers;
	const u_int m_uNumTeams;
	u_int m_uNumPlayers;
	int m_iTotalMeritPoints;
	int m_iTotalScore;
	int m_iLargestGain;
	int m_iTotalGain;
	int m_iTotalLoss;
	bool m_bDoneProcessing; 
	size_t m_uExtraDataSize;
};

#endif // #ifdef ASURA_USE_MERIT_SYSTEM

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ASURA_MERIT_H___
