//
// Asura_Physics.cpp
//
// Created by KJFL - 2004-03-17 11:22:30
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics.h"
// The physics configuration.
#include "ProjectSettings/Asura_PhysicsConfiguration.h"

//#define ASURA_USE_PHYSICS_VARIABLE_TIME_STEP 1
//#if defined(ASURA_USE_PHYSICS_VARIABLE_TIME_STEP) 
//#error ASURA_USE_PHYSICS_VARIABLE_TIME_STEP is defined, but this is currently incompatible with the current system
//#endif

#include "Core/Asura_Timers.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_DebugText.h"
#include "Dynamics/Asura_Collision_Object.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif

#include "Physics/Asura_Physics_Broadphase.h"
#include "Physics/Asura_Physics_CCD.h"
#include "Physics/Asura_Physics_Contact.h"
#include "Physics/Asura_Physics_CollisionFX_Report.h"
#include "Physics/Asura_Physics_Environment.h"
#include "Physics/Asura_Physics_EnvironmentCache.h" // for debugging counts
#include "Physics/Asura_Physics_Island.h"
#include "Physics/Asura_Physics_Object.h"
#include "Physics/Asura_Physics_Pairing.h"
#include "Physics/Asura_Physics_Ragdoll.h"
#include "Physics/Asura_Physics_RestFrame.h"
#include "Physics/Asura_Physics_RigidBody.h"
#include "Physics/Asura_Physics_RigidBodyDefinition.h"

#include "Physics/ForceFields/Asura_Physics_ForceField.h"
#include "Physics/Vehicles/Asura_Physics_Vehicle.h"

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
#include "Physics/Asura_Physics_Solver.h"
#endif

static const float fASURA_PHYSICS_TIMESTEP = 1.0f / 50.0f; // 50 hz
static const u_int uASURA_PHYSICS_RELAXATION_ITERATIONS = 5;


static float g_fTimeLeft = 0;

static u_int g_NumberOfIslands = 0;



Asura_Physics_Island*			Asura_Physics_Island::s_pxFirstIsland;
Asura_Physics_MemoryPool		Asura_Physics_Island::s_xMemory( 128*1024 ); // put aside 128k for the islands;



float	Asura_Physics_System::s_fPhysicsTime;
float	Asura_Physics_System::s_fPreviousPhysicsTime;
float	Asura_Physics_System::s_fTimeCorrectionScale;
u_int	Asura_Physics_System::s_uPhysicsUpdateCount;
bool	Asura_Physics_System::s_bFirstUpdateThisFrame;
bool	Asura_Physics_System::s_bStaticObjectsChanged;
bool	Asura_Physics_System::s_bMultithread = true;

#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
bool	Asura_Physics_System::s_bDisplayDebugInfo = false;
bool	Asura_Physics_System::s_bDisplayDebugEnvCaches = false;
bool	Asura_Physics_System::s_bDisplayDebugBroadphaseBBs = false;
bool	Asura_Physics_System::s_bDisplayDebugCollisionFX = false;
u_int	Asura_Physics_System::s_uNumActiveActors;	// not including vehicle actors
u_int	Asura_Physics_System::s_uNumActiveVehicleActors;
#endif

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
bool g_bFreezePhysics = false;
static bool g_bFreezePhysicsThisFrame = false;
#endif

Asura_AVector3	Asura_Physics_System::s_xGravityDirection;

void Asura_Physics_System::Initialise( void )
{
	s_xGravityDirection = Asura_AVector3( 0.0f, 1.0f, 0.0f );

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
	Asura_Physics_Solver::Initialise();
#endif

	Asura_Physics_Environment::Initialise();
	Asura_Physics_Contact::Initialise();
	Asura_Physics_RigidBodyInstance::Initialise();
	Asura_Physics_Pairing::Initialise();
	Asura_Physics_RestFrame::Initialise();
#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL
	Asura_Physics_Ragdoll_System::Initialise();
#endif
#ifdef ASURA_USE_COLLISIONFX
	Asura_Physics_CollisionFX_Report_System::Initialise();
#endif


#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
    AddConsoleCommands();
#endif

	g_fTimeLeft = 0;
	// set previous physics time to something non-zero but sensible
	// (AW - 04/08/05) Set s_fPhysicsTime too.
	s_fPhysicsTime = fASURA_PHYSICS_TIMESTEP;
	s_fPreviousPhysicsTime = fASURA_PHYSICS_TIMESTEP;
	s_fTimeCorrectionScale = 1.0f;

	s_uPhysicsUpdateCount = 0;

	s_bStaticObjectsChanged = true; 
}


void Asura_Physics_System::Reset( void )
{
	Asura_Physics_Environment::Reset();

	Asura_Physics_RigidBodyInstance::Reset();

	Asura_Physics_Pairing::Reset();

#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL
	Asura_Physics_Ragdoll_System::Reset();
#endif

	Asura_Physics_Contact::Reset();

	Asura_Physics_Vehicle::GetVehicleList().Clear();

	// NY - 16 December 2005 - reset physics object factory
	Asura_Physics_Object::Reset();

	Asura_Physics_RestFrame::Reset();
#ifdef ASURA_USE_COLLISIONFX
	Asura_Physics_CollisionFX_Report_System::Reset();
#endif

	g_fTimeLeft = 0;
	
	// set previous physics time to something non-zero but sensible
	// (AW - 04/08/05) Set s_fPhysicsTime too.
	s_fPhysicsTime = fASURA_PHYSICS_TIMESTEP;
	s_fPreviousPhysicsTime = fASURA_PHYSICS_TIMESTEP;
	s_fTimeCorrectionScale = 1.0f;

	s_uPhysicsUpdateCount = 0;

	s_bStaticObjectsChanged = true;

	Asura_Physics_Broadphase_System::Reset();

	Asura_Physics_ForceField::ResetAll();

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
	Asura_Physics_Solver::Reset();
#endif


}

void Asura_Physics_System::Shutdown( void )
{
	Asura_Physics_Environment::Shutdown();

	Asura_Physics_RigidBodyInstance::Shutdown();

	Asura_Physics_Pairing::Shutdown();

#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL
	Asura_Physics_Ragdoll_System::Shutdown();
#endif

	Asura_Physics_Contact::Shutdown();

	Asura_Physics_RestFrame::Shutdown();
#ifdef ASURA_USE_COLLISIONFX
	Asura_Physics_CollisionFX_Report_System::Shutdown();
#endif
	Asura_Physics_ForceField::ResetAll();

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
	Asura_Physics_Solver::Shutdown();
#endif

}

float Asura_Physics_System::GetGravityScalar( void )
{
	return fASURA_PHYSICS_GRAVITY;
}

void Asura_Physics_System::Update( void )
{
	const float fDeltaTime = Asura_Timers::GetGameTimeSlice();

	if (fDeltaTime < 0.00001f) return;

	ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__DYNAMICS, "Physics");

	g_fTimeLeft += fDeltaTime;

	s_fPreviousPhysicsTime = s_fPhysicsTime;


#if defined ASURA_USE_PHYSICS_VARIABLE_TIME_STEP
	s_fPhysicsTime = fDeltaTime;
	s_fTimeCorrectionScale = s_fPhysicsTime / s_fPreviousPhysicsTime;
#else
	s_fPhysicsTime = fASURA_PHYSICS_TIMESTEP;
	s_fTimeCorrectionScale = 1.0f;
#endif

	UpdateThread();

	ASURA_CPU_PROFILE__STOP(ASURA_CPU_PROFILE__DYNAMICS);
}


float Asura_Physics_System::GetTimeLeft( void )
{
	return g_fTimeLeft;
}

// KJFL - 8 September 2008 - not required any more, but leaving the code
// here for the lock-free queue!
#if 0 
static Asura_Physics_Island* g_pxIsland = 0;

static void* GetNextIsland( void )
{
	// KJFL - 12 March 2008 - lock-free queue access, much faster than using a Mutex
	Asura_Physics_Island* pxCurrent;

	do 
	{
		pxCurrent = g_pxIsland;

		if ( !pxCurrent ) return 0;

		Asura_Physics_Island* pxNext = pxCurrent->GetNextIsland();

		if ( Asura_Atomic::CompareAndSwapPointer( g_pxIsland, pxCurrent, pxNext ) == pxCurrent )
		{
			break;
		}
	} 
	while( true );

	return static_cast<void*>(pxCurrent);
}

#endif

void Asura_Physics_System::UpdateThread(void)
{
	Asura_Physics_EnvironmentCache::s_uNumberOfCacheUpdates = 0;

	float fDeltaTime = s_fPhysicsTime;

	
	while (g_fTimeLeft > s_fPhysicsTime*4)
	{
		g_fTimeLeft -= s_fPhysicsTime;
	}

	const Asura_AVector3 xForceEffectDueToGravity( s_xGravityDirection * (fASURA_PHYSICS_GRAVITY * fDeltaTime*fDeltaTime) );
#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL
	extern Asura_AVector3 evilg_xRagdollForceEffect;
	evilg_xRagdollForceEffect = xForceEffectDueToGravity;
#endif

	s_bFirstUpdateThisFrame = true;

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
    g_bFreezePhysicsThisFrame = g_bFreezePhysics;
#endif

	// After this any collision report pointers will be invalid, so we'll need
	// to call ClearCollisionReports() on any active instance before new reports
	// are added.
	if (g_fTimeLeft >= s_fPhysicsTime)
	{
		#ifdef ASURA_USE_COLLISIONFX
		Asura_Physics_CollisionFX_Report_System::ClearAllReports();
		#endif
		Asura_Physics_RigidBodyInstance::ResetAllCollisionReports();

		// Whip through the dynamic objects and clear their collision reports - external objects can be done as part of the broadphase loop below.
		// Because of the flag for Actors to make collision reports in their movement phases, we need to make sure dynamic objects have clear collision
		// reports *before* the first Actor moves.
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			pxInstance->ClearCollisionReports();
        }

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
        if( g_bFreezePhysicsThisFrame )
        {
		    for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		    {
			    if( !pxInstance->IsIgnore() && pxInstance->IsDynamic() )
			    {
                    pxInstance->StoreToBackup();
                }
            }
        }
#endif
	}

#if !defined ASURA_USE_PHYSICS_VARIABLE_TIME_STEP
	while (g_fTimeLeft >= s_fPhysicsTime)
#endif
	{
		g_fTimeLeft -= s_fPhysicsTime;

		// The physics update count is a simple integer that is
		// incremented every time the Physics does an update pass.
		// Note that this may not be the same as the frame number
		// count, and so if we want to use 'frame stamps' in the 
		// Physics code (e.g. to lazy-evaluate things) we need
		// to use this update count instead.
		++s_uPhysicsUpdateCount;
#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
  		Asura_Physics_Solver::Begin();
#endif
#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL
		Asura_Physics_Ragdoll_System::Update();
#endif

		// Scan through vehicle list and update them
		{
			Asura_Hashed_List_It<Asura_Physics_Vehicle> xVehicleIterator( &Asura_Physics_Vehicle::GetVehicleList() );

			while (!xVehicleIterator.Done())
			{
				Asura_Physics_Vehicle* pxVehicle = xVehicleIterator.GetCurrent();
				
				pxVehicle->Update();

				xVehicleIterator.Next();
			}
		}

		Asura_Physics_RestFrame::UpdateAll();

		Asura_Physics_ForceField::UpdateAll();


		Asura_Physics_Broadphase_System::ResetDynamicBroadphase();

		Asura_Physics_Island::ResetAllIslands();

		UpdateBroadphaseIfStaticObjectsHaveChanged();

		// and then add externally controlled objects to the dynamic object list - again not making pairs
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetExternallyControlledList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			pxInstance->ClearObstructionFlags();
			pxInstance->ClearCollisionReports();

			if (pxInstance->IsMeshDefinitionInvalid()) continue;

			if (pxInstance->IsIgnore()) continue;
			pxInstance->CheckSleepStatus();

			pxInstance->SetIsland( 0 );

			// We only need to add an external control object to the hash tree if we later want dynamic objects
			// to form pairings with it - so don't do it if the object is flagged as "m_bIgnoreDynamicObjects".
			if ( !(pxInstance->m_bIgnoreDynamicObjects) )
			{
				Asura_Physics_Broadphase_System::AddObjectToDynamicBroadphase( pxInstance, false, true );
			}
		}

		// move objects
	
#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
		s_uNumActiveActors = 0;
		s_uNumActiveVehicleActors = 0;
#endif

		// dynamic objects
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
		
			if (pxInstance->IsMeshDefinitionInvalid()) continue;

			if (pxInstance->IsIgnore()) continue;

			pxInstance->SetIsland( 0 );

			if ( pxInstance->IsActor() )
			{
				if ( pxInstance->GetActorMoveAsVehicle() )
				{
#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
					++s_uNumActiveVehicleActors;
#endif

					if ( pxInstance->m_bFeelsGravity )
					{
						pxInstance->m_xJumpVelocity += pxInstance->GetGravityDirection() * (fASURA_PHYSICS_GRAVITY * fDeltaTime * pxInstance->GetGravityScale());
					}
					else
					{
						pxInstance->m_xJumpVelocity.Zero();
					}

					pxInstance->MoveActor();//Vehicle();
				}
				else // normal actor
				{
#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
					++s_uNumActiveActors;
#endif

					if ( pxInstance->m_bFeelsGravity )
					{
						pxInstance->m_xJumpVelocity += pxInstance->GetGravityDirection() * (fASURA_PHYSICS_GRAVITY * fDeltaTime * pxInstance->GetGravityScale());
					}

					pxInstance->MoveActor();
				}

			}
			else
			{
				pxInstance->m_bHasAwakePartner = false;

				if (!pxInstance->IsAsleep() )
				{

					if (pxInstance->m_bFeelsGravity)
					{
						if ( pxInstance->GetUseOwnGravityDirection() )
						{
							const Asura_AVector3 xLocalGravityForce( pxInstance->GetOwnGravityDirection() * (fASURA_PHYSICS_GRAVITY * fDeltaTime*fDeltaTime) );
							pxInstance->m_xForceEffect += xLocalGravityForce * pxInstance->GetGravityScale();
						}
						else
						{
							pxInstance->m_xForceEffect += xForceEffectDueToGravity * pxInstance->GetGravityScale();
						}
					}
					pxInstance->MoveBase();

					pxInstance->m_xForceEffect.Zero();
				}
			}

		}
	
	
		// Now look for dynamic objects which are flagged as ignore other dynamic objects
		// (We're inefficiently going to skim through the list of dynamic objects twice,
		// since we don't have a separate external control spatial tree. Refactor here 
		// at some point if we come up on profiles.)
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			if ( pxInstance->IsIgnore() || (!pxInstance->m_bIgnoreDynamicObjects) ) continue;
	
			#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
			pxInstance->MergeIslandsUsingConstraints();
			#endif

			// check against the static object spatial hash (not adding this object, obviously)
			Asura_Physics_Broadphase_System::AddObjectToStaticBroadphase( pxInstance, true, false );

			// We only need to check the object against the hash tree if we want to collide with external control objects,
			// and as we're ignoring other dynamic objects we don't add the object itself to the tree.
			if ( !(pxInstance->m_bIgnoreExternalControlObjects) )
			{
				Asura_Physics_Broadphase_System::AddObjectToDynamicBroadphase( pxInstance, true, false );
			}

			// if the object managed to get this far without an island we'll have to give it one all to itself
			if ( !pxInstance->GetIsland() && !pxInstance->IsAsleep() )
			{
				Asura_Physics_Island* pxIsland = Asura_Physics_Island::Create();
				pxIsland->AddObject( pxInstance );
			}

		}
		// now look for dynamic objects not flagged as ignore other dynamic objects
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			if ( pxInstance->IsIgnore() || pxInstance->m_bIgnoreDynamicObjects ) continue;

			#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
			pxInstance->MergeIslandsUsingConstraints();
			#endif

			// check against the static object spatial hash (not adding this object, obviously)
			Asura_Physics_Broadphase_System::AddObjectToStaticBroadphase( pxInstance, true, false );

			// check against the dynamic spatial hash, and add this object for future objects to collide with
			Asura_Physics_Broadphase_System::AddObjectToDynamicBroadphase( pxInstance, true, true );

			// if the object managed to get this far without an island we'll have to give it one all to itself
			if ( !pxInstance->GetIsland() && !pxInstance->IsAsleep() )
			{
				Asura_Physics_Island* pxIsland = Asura_Physics_Island::Create();
				pxIsland->AddObject( pxInstance );
			}
		}
	
		// KJFL - 15 November 2005 - 
		// Scan through and remove all broken pairings - ones that weren't in any islands.
		// I'm not particularly happy with this approach at the moment, since it means we
		// have to touch lots of memory and potentially thrash our data cache. Before we
		// has islands we'd just process through all pairings, removing the broken ones
		// as we went. Now I do a scan through them all to check the update stamps.
        Asura_Physics_Pairing::RemoveBrokenParings();


		//Asura_DebugText::AddTextBlock( 10, 8, 0xffffffff, 0, "Pairings : %d", Asura_Physics_Pairing::GetNumberOfActivePairings() );

#if 0	// KJFL - 22 March 2007 - this debug code is now out of date;
		// I'll leave it here for now and somehow move it to the spatial hash class if it's needed again


		Asura_DebugText::AddTextBlock( 10, 8, 0xffffffff, 0, "Spatial hash size : %0.1f m", 1.0f / g_xStaticObjects_SpatialHash.GetBestGuessSpatialHashScale() );

		Asura_DebugText::AddTextBlock( 10, 4, 0xffffffff, 0, "Spatial hash entries" );

		char acBuffer[256] = "";

		for ( u_int u=0; u<g_xStaticObjects_SpatialHash.GetNumberOfHashNodes(); ++u )
		{
			char acNumber[10];
			sprintf( acNumber, "%d ", g_xStaticObjects_SpatialHash.GetNumberOfEntries( u ));
			strcat( acBuffer, acNumber );			
		}

		Asura_DebugText::AddTextBlock( 10, 5, 0xffffffff, 0, acBuffer );
#endif

#if _DEBUG //def ASURA_ASSERT
		// Right, we're about to resolve object collisions - let's check to see that all the collision reports 
		// have been successfully wiped
		if (IsFirstUpdateThisFrame())
		{
			// dynamic objects
			for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
			{
				// Actors can now get collision reports with their early object collision phase, so can't check this here anymore
				if (pxInstance->IsActor()) continue;
			
				Asura_Assert( pxInstance->GetObjectCollisionReportList() == 0, "Code bug: dynamic object collision report not cleared" );
			}

			// static objects
			for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetStaticList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
			{
				Asura_Assert( pxInstance->GetObjectCollisionReportList() == 0, "Code bug: static object collision report not cleared" );
			}
		}
#endif

		g_NumberOfIslands = 0;



		{
			for( Asura_Physics_Island* pxIsland = Asura_Physics_Island::GetFirstIsland(); pxIsland!=0; pxIsland = pxIsland->GetNextIsland() )
			{
#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
				Asura_Physics_Solver::SolveIsland( pxIsland );
#else
				UpdateIsland( pxIsland );

#endif
				g_NumberOfIslands++;
			}
		}

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
  		Asura_Physics_Solver::End();

		// Scan for sleep status - put here as a separate pass to match old behaviour prior to the new solver (i.e. do it after the islands work).
		// Obviously I'd prefer to have this elsewhere to avoid all these cache misses.
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			if ( (!pxInstance->IsIgnore()) && (pxInstance->IsDynamic()) && (!pxInstance->IsActor()) )
			{
				pxInstance->CheckSleepStatus();
			}
		}
#endif

		s_bFirstUpdateThisFrame = false;
	}

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
    if( g_bFreezePhysicsThisFrame && !s_bFirstUpdateThisFrame )
    {
	    for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
	    {
		    if( !pxInstance->IsIgnore() && pxInstance->IsDynamic() )
		    {
                pxInstance->RestoreFromBackup();
            }
	    }
    }
#endif
}



void Asura_Physics_System::UpdateIsland( const Asura_Physics_Island* pxIsland )
{
	for (u_int u=0; u<uASURA_PHYSICS_RELAXATION_ITERATIONS; u++)
	{
		for( Asura_Physics_Island_ObjectNode* pxNode = pxIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
		{
			Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxObject;

			Asura_Assert( !(pxInstance->IsIgnore()), "Shouldn't be ignorable objects here" );
		
			if (pxInstance->IsAsleep()) continue;

			pxInstance->m_xContactClusterList.Reset();
		}


		bool bNoContacts = true;

		if ( pxIsland->m_pxPairingNode )
		{
			for( Asura_Physics_Island_PairingNode* pxNode = pxIsland->m_pxPairingNode; pxNode!=0; pxNode = pxNode->m_pxNextPairingNode )
			{
				Asura_Physics_Pairing* pxPairing = pxNode->m_pxPairing;

				// pairing ptr can be null if it's destroyed in current iterations
				if ( pxPairing )
				{
					bool bShouldDestroy = pxPairing->Update();

					if ( bShouldDestroy )
					{
						Asura_Physics_Pairing::Destroy( pxPairing );

						
						pxNode->m_pxPairing = 0;
					}
					else
					{
						// assume contacts if we have a pairing
						bNoContacts = false;
					}
				}
			}
		}


		for( Asura_Physics_Island_ObjectNode* pxNode = pxIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
		{
			Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxObject;

			if ( !pxInstance->IsDynamic() || pxInstance->IsAsleep() ) continue;
			
#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
			// Resolve constraints
			pxInstance->ResolveConstraints();
#endif

			// Collide with environment
			pxInstance->TestForNewContacts();
			
			if ( bNoContacts )
			{
				if ( pxInstance->HasContacts() ) bNoContacts = false;
			}
		}

		if (bNoContacts)
		{
			break;
		}
	}
	
	for( Asura_Physics_Island_ObjectNode* pxNode = pxIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
	{
		Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxObject;

		if ( pxInstance->IsDynamic() && !pxInstance->IsActor() )
		{
			pxInstance->CheckSleepStatus();
		}
	}

}

#ifdef ASURA_TOOLS
void Asura_Physics_System::UpdateForWaypointing( void )
{
    const float fDeltaTime = fASURA_PHYSICS_TIMESTEP;

    g_fTimeLeft += fDeltaTime;

    s_fPreviousPhysicsTime = s_fPhysicsTime;
    s_fPhysicsTime         = fDeltaTime;
    s_fTimeCorrectionScale = 1.0f;

    UpdateThread();
}
#endif

float Asura_Physics_System::GetTimeStep( void )
{
    return fASURA_PHYSICS_TIMESTEP;
}

void Asura_Physics_System::DebuggingRender( void )
{
	if ( !GetDisplayDebugInfo() ) return; 

//	Asura_DebugText::AddTextBlock(1,15,0xffffffff,0, "Ragdoll Update: %0.0f", fRagdollTime * 1000000.0f);


	//Asura_DebugText::AddTextBlock(1,5,0xffffffff,0, "Env Cache Fills: %d", Asura_Physics_EnvironmentCache::s_uNumberOfCacheUpdates);
	//Asura_DebugText::AddTextBlock(1,6,0xffffffff,0, "Pairings: %d", Asura_Physics_Pairing::GetNumberOfActivePairings() );
	//Asura_DebugText::AddTextBlock(1,7,0xffffffff,0, "Spatial Hash mem %3.1fk", (float)g_xSpatialHash.GetMemUsed()/1024.0f);
/*	
		
	Asura_DebugText::AddTextBlock(1,15,0xffffffff,0, "Spatial Hash: %0.0f", fSpatialHashTime * 1000000.0f);
	Asura_DebugText::AddTextBlock(1,16,0xffffffff,0, "Links: %0.0f", fLinksTime * 1000000.0f);
	Asura_DebugText::AddTextBlock(1,17,0xffffffff,0, "Object-Object: %0.0f", fPairsTime * 1000000.0f);
	Asura_DebugText::AddTextBlock(1,18,0xffffffff,0, "Environment: %0.0f", fEnvironmentTime * 1000000.0f);
	Asura_DebugText::AddTextBlock(1,20,0xffffffff,0, "Total: %0.1f ms", fTotalTime * 1000.0f);
*/
#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL
	Asura_Physics_Ragdoll_System::DebugRender();
#endif

	//static bool bDrawText = false;

	//if (bDrawText)
	//{
	//	Asura_DebugText::AddTextBlock(1,1,0xffffff00,0, "Press H to toggle this text");
	//	Asura_DebugText::AddTextBlock(1,3,0xffffffff,0, "Press R to reset demo");
	//	Asura_DebugText::AddTextBlock(1,4,0xffffffff,0, "Press M to drop a box");
	//	Asura_DebugText::AddTextBlock(1,5,0xffffffff,0, "Press N to drop 10 boxes at once");
	//	Asura_DebugText::AddTextBlock(1,7,0xffffffff,0, "Press F1 to toggle wireframe");
	//	Asura_DebugText::AddTextBlock(1,8,0xffffffff,0, "Press F2, F3, F4 or F5 to recall set camera positions");
	//	Asura_DebugText::AddTextBlock(1,9,0xffffffff,0, "Press Y to show sleep status");

	//	Asura_DebugText::AddTextBlock(1,11,0xffffffff,0, "Press O to pick all objects up");
	//	Asura_DebugText::AddTextBlock(1,12,0xffffffff,0, "Press I,J,K or L to move objects around");
	//	Asura_DebugText::AddTextBlock(1,13,0xffffffff,0, "Press U to make time pass very slowly");

	//	Asura_DebugText::AddTextBlock(1,15,0xffffffff,0, "Press W/A/S/D to move; press Q to move forwards quickly");
	//	Asura_DebugText::AddTextBlock(1,16,0xffffffff,0, "Hold middle-mouse button and move mouse to rotate camera view");
	//}

    u_int uNumInstances;
	u_int uNumActiveInstances;
    u_int uNumAwakeInstances;

	// Render dynamic objects
    uNumInstances = 0;
	uNumActiveInstances = 0;
    uNumAwakeInstances = 0;
    u_int uNumActors = 0;
	u_int uNumActiveActors = 0;
    u_int uNumAwakeActors = 0;
	for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
	{
		if( pxInstance->IsMeshDefinitionInvalid() ) continue;

		pxInstance->DebuggingRender();

        ++uNumInstances;
		if( !pxInstance->IsIgnore() )
		{
			++uNumActiveInstances;
			if( !pxInstance->IsAsleep() )
			{
				++uNumAwakeInstances;
			}
		}

        if( pxInstance->IsActor() && !pxInstance->GetActorMoveAsVehicle() )
        {
            ++uNumActors;
			if( !pxInstance->IsIgnore() )
			{
				++uNumActiveActors;
				if( !pxInstance->IsAsleep() )
				{
					++uNumAwakeActors;
				}
			}
        }
	}
    Asura_DebugTextPrint(( 1, 1, 0xFFFFFFFF, 0, "Dynamic:  %3u/%3u (%3u)", uNumAwakeInstances, uNumActiveInstances, uNumInstances ));
    Asura_DebugTextPrint(( 1, 2, 0xFFFFFFFF, 0, "Actor:    %3u/%3u (%3u)", uNumAwakeActors,    uNumActiveActors,    uNumActors    ));

	// Render static objects
	uNumInstances = 0;
	uNumActiveInstances = 0;
	uNumAwakeInstances = 0;
	for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetStaticList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
	{
		if( pxInstance->IsMeshDefinitionInvalid() ) continue;

		pxInstance->DebuggingRender();

        ++uNumInstances;
		if( !pxInstance->IsIgnore() )
		{
			++uNumActiveInstances;
		}
	}
    Asura_DebugTextPrint(( 1, 3, 0xFFFFFFFF, 0, "Static:   %3u/%3u (%3u)", uNumAwakeInstances, uNumActiveInstances, uNumInstances ));

	// Render externally controlled objects
	uNumInstances = 0;
	uNumActiveInstances = 0;
	uNumAwakeInstances = 0;
	for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetExternallyControlledList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
	{
		if( pxInstance->IsMeshDefinitionInvalid() ) continue;

		pxInstance->DebuggingRender();

        ++uNumInstances;
		if( !pxInstance->IsIgnore() )
		{
			++uNumActiveInstances;
			if( !pxInstance->IsAsleep() )
			{
				++uNumAwakeInstances;
			}
		}
	}
    Asura_DebugTextPrint(( 1, 4, 0xFFFFFFFF, 0, "External: %3u/%3u (%3u)", uNumAwakeInstances, uNumActiveInstances, uNumInstances ));

	// Render vehicles
	Asura_Hashed_List_It<Asura_Physics_Vehicle> xVehicleIterator( &Asura_Physics_Vehicle::GetVehicleList() );
	while( !xVehicleIterator.Done() )
	{
		Asura_Physics_Vehicle* pxVehicle = xVehicleIterator.GetCurrent();
		
		pxVehicle->DebuggingRender();

		xVehicleIterator.Next();
	}
	
	Asura_Physics_ForceField::DebugRenderAll();
#if defined(ASURA_PHYSICS_ENABLE_DEBUG_INFO) && defined(ASURA_USE_COLLISIONFX)
	if (s_bDisplayDebugCollisionFX)
	{
		Asura_Physics_CollisionFX_Report_System::DebugRender();
	}
#endif
}

void Asura_Physics_System::SetGravityDirection( const Asura_AVector3& xDirection )
{
	s_xGravityDirection = xDirection;
	s_xGravityDirection.Normalise();
}

const Asura_AVector3& Asura_Physics_System::GetGravityDirection( void )
{
	return s_xGravityDirection;
}

void Asura_Physics_System::UpdateBroadphaseIfStaticObjectsHaveChanged( void )
{
	if ( s_bStaticObjectsChanged )
	{
		s_bStaticObjectsChanged = false;

		Asura_Physics_Broadphase_System::ResetStaticBroadphase();

		// add static objects etc first, not making any pairings (pairs of static objects makes no sense)
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetStaticList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			// since we're going through the static objects, let's check they don't have any collision reports,
			// which would indicate a bug
			Asura_Assert( !(pxInstance->GetObjectCollisionReportList()), "Code bug: orphaned object collision report" );

			if (pxInstance->IsMeshDefinitionInvalid()) continue;

			// check for ignored any objects, since it's better to reuse spatial hash than recalc it just because an object
			// has become ignored/not ignored.
			if (pxInstance->IsIgnore()) continue;

			pxInstance->SetIsland( 0 );

			Asura_Physics_Broadphase_System::AddObjectToStaticBroadphase( pxInstance, false, true );
		}
	}
}

bool Asura_Physics_System::DoesOBBIntersectStaticObjects( const Asura_OBB& xOBB, Asura_Handle iHandleOfObjectToIgnore )
{
	return Asura_Physics_Broadphase_System::DoesOBBIntersectStaticObjects( xOBB, iHandleOfObjectToIgnore );
}

bool Asura_Physics_System::DoesOBBIntersectDynamicObjects( const Asura_OBB& xOBB, Asura_Handle iHandleOfObjectToIgnore )
{
	return Asura_Physics_Broadphase_System::DoesOBBIntersectDynamicObjects( xOBB, iHandleOfObjectToIgnore );
}

bool Asura_Physics_System::CanPlaceObjectAt( const Asura_Vector_3& xPosition, const class Asura_Collision_Object& xObject,
	const u_short usFlagsOfPolysToIgnore, const Asura_Handle iHandleOfObjectToIgnore, const u_short usFlagsOfPolysToTestFor,
	bool bConsiderStaticObjects /*= false*/, bool bConsiderDynamicObjects /*= false*/ )
{

	
	if ( bConsiderDynamicObjects || bConsiderStaticObjects )
	{
		Asura_OBB xOBB( xObject.GetBoundingBox(), Asura_AVector3(xPosition) );

		if (bConsiderDynamicObjects)
		{
			if (Asura_Physics_Broadphase_System::DoesOBBIntersectDynamicObjects( xOBB, iHandleOfObjectToIgnore )) return false;
		}
		if (bConsiderStaticObjects)
		{
			if (Asura_Physics_Broadphase_System::DoesOBBIntersectStaticObjects( xOBB, iHandleOfObjectToIgnore )) return false;
		}
	}

	// very minimalist and suboptimal check...
	Asura_Physics_EnvironmentCache xCache;

	Asura_Bounding_Box xBB = xObject.GetBoundingBox();

	xBB.Displace( xPosition );

	xCache.SetFlagsOfPolysToIgnore( usFlagsOfPolysToIgnore );
	xCache.SetFacesOnly( true );
	xCache.Update( xBB );


	// AJG - 07/07/06 - Alternate version of the loop so the flags to test for are included.
	//					Not exactly optimal, but Gun needs a working version of this quickly and doing this through the cache is a much more complex process. Hopefully someone will improve this later.
	//					Note that with this, any poly with no flags will never be tested against.
	if (usFlagsOfPolysToTestFor != 0xFFFF)
	{
		for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
		{
			if ((Asura_Physics_Environment::GetFaceFlags(pxFace->m_uFaceID) & usFlagsOfPolysToTestFor) == 0) { continue; }
			if ( Asura_Physics_Environment::DoesFaceIntersectAABB( pxFace->m_uFaceID, xBB ) ) return false;
		}
	}
	else // If the flags to test for are 0xFFFF, behave as originally, so 0-flag polys are included.
	{
		for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
		{
			if ( Asura_Physics_Environment::DoesFaceIntersectAABB( pxFace->m_uFaceID, xBB ) ) return false;
		}
	}

	return true;
}


// AJG - 10/07/06 - Orientated version, needed for Gun's horses (which use orientated physics objects).
//					First version - probably not optimal, but quick version needed.
bool Asura_Physics_System::CanPlaceObjectAt( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation,
		const class Asura_Collision_Object& xObject, const u_short usFlagsOfPolysToIgnore,
		const Asura_Handle iHandleOfObjectToIgnore, const u_short usFlagsOfPolysToTestFor,
		bool bConsiderStaticObjects /*= false*/, bool bConsiderDynamicObjects /*= false*/ )
{

	if ( bConsiderDynamicObjects || bConsiderStaticObjects )
	{
		Asura_AMatrix3x3 xOrient;
		
		xOrient.FromMat3( xOrientation );

		Asura_OBB xOBB( xObject.GetBoundingBox(), Asura_AVector3(xPosition), xOrient );

		if (bConsiderDynamicObjects)
		{
			if (DoesOBBIntersectDynamicObjects( xOBB, iHandleOfObjectToIgnore )) return false;
		}
		if (bConsiderStaticObjects)
		{
			if (DoesOBBIntersectStaticObjects( xOBB, iHandleOfObjectToIgnore )) return false;
		}
	}

	// very minimalist and suboptimal check...
	Asura_Physics_EnvironmentCache xCache;

	// Calculate the smallest enclosing AABB to fill the cache.
	Asura_Vector_3 xMin(Asura_Maths::LargeFloat, Asura_Maths::LargeFloat, Asura_Maths::LargeFloat);
	Asura_Vector_3 xMax(-Asura_Maths::LargeFloat, -Asura_Maths::LargeFloat, -Asura_Maths::LargeFloat);

	Asura_Bounding_Box	xBB;

	xBB.FromOBB(xObject.GetBoundingBox(), xOrientation);

	xBB.Displace( xPosition );

	xCache.SetFlagsOfPolysToIgnore( usFlagsOfPolysToIgnore );
	xCache.SetFacesOnly( true );
	xCache.Update( xBB );


	// AJG - 07/07/06 - Alternate version of the loop so the flags to test for are included.
	//					Not exactly optimal, but Gun needs a working version of this quickly and doing this through the cache is a much more complex process. Hopefully someone will improve this later.
	//					Note that with this, any poly with no flags will never be tested against.
	if (usFlagsOfPolysToTestFor != 0xFFFF)
	{
		for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
		{
			if ((Asura_Physics_Environment::GetFaceFlags(pxFace->m_uFaceID) & usFlagsOfPolysToTestFor) == 0) { continue; }
			if ( Asura_Physics_Environment::DoesFaceIntersectOBB( pxFace->m_uFaceID, xObject, xPosition, xOrientation ) ) return false;
		}
	}
	else // If the flags to test for are 0xFFFF, behave as originally, so 0-flag polys are included.
	{
		for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
		{
			if ( Asura_Physics_Environment::DoesFaceIntersectOBB( pxFace->m_uFaceID, xObject, xPosition, xOrientation ) ) return false;
		}
	}


	return true;
}

// AJG - 06/08/07 - Sphere stack test similar to above. At this time, only tests against the environment.
// Heavily based on Asura_Physics_RigidBodyInstance::TestForNewContacts_SphereStack, and so there may be more we can do to improve this and tidy it up to allow sharing of code.
bool Asura_Physics_System::CanPlaceSphereStackAt(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, u_short usFlagsOfPolysToIgnore/* = 0*/, float fAllowedPenetration/* = 0.0f*/)
{
	// Create a sphere stack at the correct location.
	Asura_OBB xOBB(xBB, Asura_AVector3(xPosition));

	Asura_Physics_SphereGroup xSphereStack(xOBB, true);

	const float fSphereRadius = xSphereStack.GetSphereRadius();
	const float fSphereRadiusSquared = fSphereRadius * fSphereRadius;

	const bool bIsSphereStack = xSphereStack.GetNumberOfSpheres() > 1;

	Asura_AVector3 xPrimaryAxis;
	xPrimaryAxis.Zero();
	float fPrimaryAxisLength = 0.0f;

    if ( bIsSphereStack )
	{
		float fX = xOBB.GetHalfExtentX();
		float fY = xOBB.GetHalfExtentY();
		float fZ = xOBB.GetHalfExtentZ();

		if ( fX > fY )
		{
			if ( fX > fZ )
			{
				// it's X
				xPrimaryAxis.SetX( 1.0f );
				fPrimaryAxisLength = fX;
			}
			else // it's Z
			{
				xPrimaryAxis.SetZ( 1.0f );
				fPrimaryAxisLength = fZ;
			}
		}
		else if ( fY > fZ )
		{
			// it's Y
			xPrimaryAxis.SetY( 1.0f );
			fPrimaryAxisLength = fY;
		}
		else // it's Z
		{
			xPrimaryAxis.SetZ( 1.0f );
			fPrimaryAxisLength = fZ;
		}
	}

	// Create an env cache of surrounding polys.
	Asura_Bounding_Box xDisplacedBox;
	xOBB.CalculateEnclosingAABB(xDisplacedBox);

	Asura_Physics_EnvironmentCache xEnvironmentCache;
	xEnvironmentCache.SetFlagsOfPolysToIgnore(usFlagsOfPolysToIgnore);
	xEnvironmentCache.Update(xDisplacedBox, 0.1f);

	const Asura_AVector3 xCentreOfMass = xOBB.GetCentre();

	// test for intersection of the spheres with each poly in the environment cache
	for ( Asura_Physics_EnvironmentCache_Face* pxFace = xEnvironmentCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
	{
		const Asura_AVector3& xPV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xPV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xPV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		Asura_AVector3 xNormal = ((xPV1 - xPV0)^(xPV2 - xPV0));
		xNormal.Normalise();

		
		for ( u_int v=0; v<xSphereStack.GetNumberOfSpheres(); ++v )
		{
			bool bFacingTowards;
			float fHitDistance;

			if( Asura_Physics_System::RayIntersectsTri( xSphereStack.GetSphereCentres()[v], -xNormal, xPV0, xPV1, xPV2, bFacingTowards, fHitDistance ) )
			{
				// Because we're not trying to resolve, we don't care whether the face we intersect with is facing in the correct direction or not.
				// As such, just see if the hit distance is within the object (minus the allowed penetration).
				const float fPenetration = fSphereRadius - Asura_Maths::Abs(fHitDistance);
				if ( fPenetration > fAllowedPenetration )
				{
					return false;
				}
			}
		}

		// if we're a sphere-stack do an extra test along the primary axis (catches cases where individual spheres are on the other side of polys)
		if (bIsSphereStack)
		{
			bool bFacingTowards;
			float fHitDistance;

			if( Asura_Physics_System::RayIntersectsTri( xCentreOfMass, xPrimaryAxis, xPV0, xPV1, xPV2, bFacingTowards, fHitDistance ) )
			{
				// Because we're not trying to resolve, we don't care whether the face we intersect with is facing in the correct direction or not.
				// As such, just see if the hit distance is within the object (minus the allowed penetration).
				const float fPenetration = fPrimaryAxisLength - Asura_Maths::Abs(fHitDistance);
				if ( fPenetration > fAllowedPenetration )
				{
					return false;
				}
			}
			
		}
	}

	for ( Asura_Physics_EnvironmentCache_Edge* pxEdge = xEnvironmentCache.GetEdgeList(); pxEdge; pxEdge = pxEdge->m_pxNext )
	{
		const Asura_AVector3& xEdgeV0 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
		const Asura_AVector3& xEdgeV1 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );


		for ( u_int v=0; v<xSphereStack.GetNumberOfSpheres(); ++v )
		{
			Asura_AVector3 xPointOnEdge;
			Asura_AVector3 xSphereCentre = xSphereStack.GetSphereCentres()[v];

            Asura_Maths::GetNearestPointOnLineSegment( xEdgeV0, xEdgeV1, xSphereCentre, xPointOnEdge );

			Asura_AVector3 xDisplacement = (xSphereCentre - xPointOnEdge);

			float fDistanceSquared = xDisplacement.MagnitudeSquared();

			if ( fDistanceSquared > 0.0f && fDistanceSquared < fSphereRadiusSquared )
			{
				const float fDistance = xDisplacement.NormaliseReturningMagnitude();

				const float fPenetration = fSphereRadius - fDistance;

				if ( fPenetration > fAllowedPenetration )
				{
					return false;
				}
			}
		}
	}

	return true;
}


bool Asura_Physics_System::RayIntersectsTri(const Asura_AVector3& alpha, const Asura_AVector3& beta, const Asura_AVector3& v0, const Asura_AVector3& v1, const Asura_AVector3& v2, bool& facingToward, float& distance)
{
	const float epsilon = 1.0e-5f;
	Asura_AVector3 e1 = v1-v0;
	Asura_AVector3 e2 = v2-v0;

	Asura_AVector3 p = beta^e2;
	//Asura_Vector_3 n = e1^e2;
	float a = e1 * p;

	if (a>-epsilon && a<epsilon) return false;
	// if a is negative, it's facing away from the ray
	
	a = 1.0f/a; //because a can be negative, must take care of expression such as u > a && u/a > 1

	Asura_AVector3 s = alpha-v0;
	float u = (s*p)*a;

	if (u<0.0f || u>1.0f) return false;

	Asura_AVector3 q = s^e1;
	float v = (beta*q)*a;

	if (v<0.0f || u+v>1.0f) return false;

	float lambda = (e2*q)*a;

	if (a<0.0f) facingToward = false;
	else facingToward = true;

	distance = lambda;
	return true;
}


bool Asura_Physics_System::SmallestVectorBetweenTwoLines(const Asura_AVector3 &a1, const Asura_AVector3 &a2, const Asura_AVector3 &b1, const Asura_AVector3 &b2, float& out_along_a, float& out_along_b)
{
	const Asura_AVector3 &B0 = a1;
	const Asura_AVector3 &B1 = b1;
	
	const Asura_AVector3  M0 = a2 - a1;
	const Asura_AVector3  M1 = b2 - b1;
	
	float a =  M0 * M0;
	float b = -M0 * M1;
	float c =  M1 * M1;
	float d =  M0 * (B0 - B1);
	float e = -M1 * (B0 - B1);
	
	//
	// What's s and t?
	//
	
	float det = a*c - b*b;
	
	if (det <= 0.00001F)
	{
		return false;
	}
	
	float over_det = 1.0F / det;
	
	out_along_a = (b*e - c*d) * over_det;
	out_along_b = (b*d - a*e) * over_det;

	if (out_along_a <=0.0f || out_along_a >= 1.0f)
	{
		return false;
	}
	if (out_along_b <=0.0f || out_along_b >= 1.0f)
	{
		return false;
	}


	return true;
}


bool Asura_Physics_System::SweptSphereVsEdge(  const float fSphereRadius, const Asura_AVector3& xSphereStart, const Asura_AVector3& xSphereDir, const Asura_AVector3& xLineStart, const Asura_AVector3& xLineEnd, Asura_AVector3& xCollisionPoint, float& fAllowedSphereMovement )
{
	bool bHitSomething = false;

	fAllowedSphereMovement = 1.0f;

	// check against the line segment
	if ( SweptSphereVsLineSegment( fSphereRadius, xSphereStart, xSphereDir, xLineStart, xLineEnd, xCollisionPoint, fAllowedSphereMovement ) )
	{
		bHitSomething = true;
	}

	// and against the vertices themselves

	if ( SweptSphereVsPoint( fSphereRadius, xSphereStart, xSphereDir, xLineStart, xCollisionPoint, fAllowedSphereMovement ) )
	{
		bHitSomething = true;
	}

	if ( SweptSphereVsPoint( fSphereRadius, xSphereStart, xSphereDir, xLineEnd, xCollisionPoint, fAllowedSphereMovement ) )
	{
		bHitSomething = true;
	}


	return bHitSomething;
}

bool Asura_Physics_System::SweptSphereVsPoint( const float fSphereRadius, const Asura_AVector3& xSphereStart, const Asura_AVector3& xSphereDir, const Asura_AVector3& xPoint, Asura_AVector3& xCollisionPoint, float& fAllowedSphereMovement )
{
	const Asura_AVector3 xAlpha = xSphereStart - xPoint;

	// do the maths and we make make a quadratic...

	const float fA = xSphereDir * xSphereDir;
	const float fB = 2.0f * ( xAlpha * xSphereDir );
	const float fC = (xAlpha * xAlpha) - fSphereRadius*fSphereRadius;

	if ( Asura_Maths::Abs( fA ) < 0.000001f) return false;

	float fDeterminant = fB*fB - 4.0f * fA * fC;

	if (fDeterminant<0.0f) return false;

	float fSqrtDeterminant = Asura_Maths::Sqrt( fDeterminant );

	float fDenom = 1.0f / ( fA * 2.0f );

	// hmm, technically we should only ever be interested in the first intersection
	// let's try that and see if it's robust. (No problems so far.)
	float fLambda =( -fB - fSqrtDeterminant ) * fDenom;

	if ( fLambda>=0.0f && fLambda<=fAllowedSphereMovement )
	{
		fAllowedSphereMovement = fLambda;
		xCollisionPoint = xPoint;
		return true;
	}

	return false;

}


bool Asura_Physics_System::SweptSphereVsLineSegment(  const float fSphereRadius, const Asura_AVector3& xSphereStart, const Asura_AVector3& xSphereDir, const Asura_AVector3& xLineStart, const Asura_AVector3& xLineEnd, Asura_AVector3& xCollisionPoint, float& fAllowedSphereMovement )
{
	const Asura_AVector3 xAlpha = xSphereStart - xLineStart;
	const Asura_AVector3 xLineDir = xLineEnd - xLineStart;

	// Solution comes in the form of a quadratic

	const float fLineDirSqd = (xLineDir*xLineDir);
	const float fLineDirDotSphereDir = (xSphereDir*xLineDir);
	const float fAlphaDotLineDir = xAlpha*xLineDir;

	const float fRadiusSquared = fSphereRadius * fSphereRadius;

	float fA = (xSphereDir*xSphereDir)*fLineDirSqd - fLineDirDotSphereDir*fLineDirDotSphereDir;
	float fB = 2.0f * ( (xAlpha*xSphereDir)*fLineDirSqd - (fAlphaDotLineDir)*fLineDirDotSphereDir );
	float fC = ( xAlpha*xAlpha - fRadiusSquared ) * fLineDirSqd - (fAlphaDotLineDir*fAlphaDotLineDir);

	if ( Asura_Maths::Abs( fA ) < 0.000001f) return false;

	float fDeterminant = fB*fB - 4.0f * fA * fC;
	if ( fDeterminant<0.0f ) return false;

	float fDenom = 1.0f / ( fA * 2.0f );

	float fSqrtDet = Asura_Maths::Sqrt( fDeterminant );

	// try the lowest solution first (the "-" root)
	float fLambda = ( -fB - fSqrtDet ) * fDenom;

	// For this to be a valid solution, we must have the point both on the valid trajectory of the
	// sphere, and on the line segment, i.e.
	//
	//	0<= fLambda <=1 
	//  0<=   fMu   <=1		where fMu is the fraction along the line segment from xLineStart
	//

	if ( (fLambda >= 0.0f && fLambda <= 1.0f) )
	{
		float fMu = ((xAlpha + xSphereDir * fLambda) * xLineDir) / fLineDirSqd;

		if ( fMu >= 0.0f && fMu <= 1.0f)
		{
			// valid solution
			xCollisionPoint = xLineStart + xLineDir * fMu;

			// UNLESS we're just touching, in which case we should see if the sphere 
			// is trying to move away from the point of contact
			if ( fLambda == 0.0f )
			{
				const Asura_AVector3 xCollisionToCentre = xSphereStart - xCollisionPoint;

				if ( xCollisionToCentre * xSphereDir > 0.0f )
				{
					return false;
				}
			}

			fAllowedSphereMovement = fLambda;

			return true;
		}
	}

	// If the above solution isn't in a valid, we need to check the "+" root and do the same tests.
	fLambda = ( -fB + fSqrtDet ) * fDenom;

	if ( (fLambda >= 0.0f && fLambda <= 1.0f) )
	{
		float fMu = ((xAlpha + xSphereDir * fLambda) * xLineDir) / fLineDirSqd;

		if ( fMu >= 0.0f && fMu <= 1.0f)
		{
			// valid solution
			xCollisionPoint = xLineStart + xLineDir * fMu;

			// UNLESS we're just touching, in which case we should see if the sphere 
			// is trying to move away from the point of contact
			if ( fLambda == 0.0f )
			{
				const Asura_AVector3 xCollisionToCentre = xSphereStart - xCollisionPoint;

				if ( xCollisionToCentre * xSphereDir > 0.0f )
				{
					return false;
				}
			}

			fAllowedSphereMovement = fLambda;

			return true;
		}
	}

	// both solutions invalid
	return false;
}

float Asura_Physics_System::SquaredDistanceBetweenLineSegments( const Asura_AVector3& xSeg0Start, const Asura_AVector3& xSeg0End, const Asura_AVector3& xSeg1Start, const Asura_AVector3& xSeg1End )
{
	// KJFL - 15 August 2006 - see section on linear components in Geometric Tools for Computer Graphics, circa p. 417 for a discussion
	// on this, but beware the pseudocode which has quite a few errors. 

	const Asura_AVector3 xDir0 = xSeg0End - xSeg0Start;
	const Asura_AVector3 xDir1 = xSeg1End - xSeg1Start;
	const Asura_AVector3 xDisp = xSeg0Start - xSeg1Start;

	const float fA = xDir0 * xDir0;
	const float fB = xDir0 * xDir1;
	const float fC = xDir1 * xDir1;
	const float fD = xDir0 * xDisp;
	const float fE = xDir1 * xDisp;

	const float fDet = fA * fC - fB * fB;

	float fSNum, fTNum, fSDenom, fTDenom;

	fSDenom = fDet;
	fTDenom = fDet;

	if ( fDet < 0.00001f ) // (very near) parallel
	{
		fSNum = 0.0f;
		fSDenom = fA;

		fTNum = fE;
		fTDenom = fC;
	}
	else
	{
		fSNum = fB * fE - fC * fD;
		fTNum = fA * fE - fB * fD;
	}

	// check S
	if ( fSNum < 0.0f )
	{
		fSNum = 0.0f;
		fTNum = fE;
		fTDenom = fC;
	}
	else if ( fSNum > fSDenom )
	{
		fSNum = fSDenom;
		fTNum = fE + fB;
		fTDenom = fC;
	}

	// check T
	if ( fTNum < 0.0f )
	{
		fTNum = 0.0f;
		float fNewSNum = - fD;

		if ( fNewSNum < 0.0f )
		{
			fSNum = 0.0f;
		}
		else if ( fNewSNum > fA )
		{
			fSNum = fSDenom;
		}
		else
		{
			fSNum = fNewSNum;
			fSDenom = fA;
		}
	}
	else if ( fTNum > fTDenom )
	{
		fTNum = fTDenom;

		float fNewSNum = -fD + fB;

		if ( fNewSNum < 0.0f )
		{
			fSNum = 0.0f;
		}
		else if ( fNewSNum > fA )
		{
			fSNum = fSDenom;
		}
		else	
		{
			fSNum = fNewSNum;
			fSDenom = fA;
		}
	}

	const float fS = fSNum / fSDenom;
	const float fT = fTNum / fTDenom;

	const Asura_AVector3 xV = xDisp + ( xDir0 * fS ) - ( xDir1 * fT );

	return xV.MagnitudeSquared();
}


bool Asura_Physics_System::MoveParticle( Asura_Vector_3& xPosition, Asura_Vector_3& xVelocity, float fElasticity, u_short usFlagsOfPolysToIgnore/*=AsuraPolygonFlag_NoCollisions*/ )
{
    bool bCollisionOccurred = false;


	Asura_Vector_3 xDisplacement = xVelocity * Asura_Timers::GetGameTimeSlice();

	Asura_Bounding_Box xBB;

	{
		const Asura_Vector_3 xEnd = xPosition + xDisplacement;

		xBB.Reset();
		xBB.Set( xPosition );
		xBB.Set( xEnd );
	}

	Asura_Physics_EnvironmentCache xCache;
	xCache.SetFacesOnly( true );
	xCache.SetFlagsOfPolysToIgnore( usFlagsOfPolysToIgnore );
	xCache.Update( xBB );

    for (u_int uMoves = 0; uMoves<2; ++uMoves)
    {
		float fLeastDistance = 1.0f;
		//Needed to fix the PS3 release warning
		Asura_Vector_3 xFaceNormal( 0.0f, 0.0f, 1.0f );

		for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xCache.GetFaceList(); pxFace != 0; pxFace = pxFace->m_pxNext )
		{
            const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
			const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
			const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

	        bool bFacingTowards;

            float fLambda;
			
			if (RayIntersectsTri( Asura_AVector3(xPosition), Asura_AVector3(xDisplacement), xV0, xV1, xV2, bFacingTowards, fLambda ))
			{
	            if ( fLambda>=0 && fLambda<fLeastDistance && bFacingTowards )
		        {
			        fLeastDistance = fLambda;
				    xFaceNormal = ((xV1 - xV0)^(xV2 - xV0));
		
				}
			}

        }
		
		if (fLeastDistance < 1.0f)
		{
            // collision - elasticity 
            {
				xFaceNormal.Normalise();

                float fMagOfPerpImp = (xFaceNormal * xVelocity) * (1.0f + fElasticity);
                xVelocity -= xFaceNormal*fMagOfPerpImp;

				float fL = (xFaceNormal*xDisplacement) * (1.0f + fElasticity);
				xDisplacement -= xFaceNormal*fL;
            }

            fLeastDistance -=0.01f;
            bCollisionOccurred = true;
        }
        
        if (fLeastDistance>0.0f) xPosition += xDisplacement*fLeastDistance;

        if (fLeastDistance>0.9f) return bCollisionOccurred;

   

		{
			const Asura_Vector_3 xEnd = xPosition + xDisplacement;

			xBB.Reset();
			xBB.Set( xPosition );
			xBB.Set( xEnd );
	
			xCache.Update ( xBB );
    
		}
	}

    return bCollisionOccurred;
}

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 

static bool ConsoleCmd_Physics_WakeAllObjects( bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( L"Physics.WakeAllObjects" );
        Asura_CommandConsole::Printf( L"    Wakes all existing physics objects." );
    }
    else
    {
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
	    {
            pxInstance->WakeUp();
	    }
    }

    return true;
}

void Asura_Physics_System::AddConsoleCommands( void )
{
    Asura_CommandConsole::AddCmd( L"Physics.WakeAllObjects", ConsoleCmd_Physics_WakeAllObjects, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	//Asura_CommandConsole::AddVar( L"Physics.Multithread", &(s_bMultithread), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
	Asura_CommandConsole::AddVar( L"Physics.Render", &(s_bDisplayDebugInfo), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Physics.RenderEnvCaches", &(s_bDisplayDebugEnvCaches), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Physics.RenderBroadphaseBBs", &(s_bDisplayDebugBroadphaseBBs), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#ifdef ASURA_USE_COLLISIONFX
	Asura_CommandConsole::AddVar( L"Physics.RenderCollisionFX", &(s_bDisplayDebugCollisionFX), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif
#endif
#if 0
	static const Asura_UnicodeChar* const wszPHYSICS_NUMBEROFISLANDS = L"Physics.NumberOfIslands";

	Asura_CommandConsole::AddVar( wszPHYSICS_NUMBEROFISLANDS, &g_uNumberOfIslands, 0, INT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::MakeVarReadOnly( wszPHYSICS_NUMBEROFISLANDS );
#endif
}

#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE
