//
// Asura_Physics_CCD.cpp
//
// Created by KJFL - 17 November 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_CCD.h"

#include "Core/Memory/Asura_Mem.h"
#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_Environment.h"
#include "Physics/Asura_Physics_RigidBody.h"
#include "Physics/Asura_Physics_RigidBodyDefinition.h"

#include "Physics/Asura_Physics_Ragdoll.h"
#include "Physics/Asura_Physics_Ragdoll_Description.h"

// TODO - store in one place?
static const float fAXIALLYALIGNED_DIST_THRESHOLD = 0.005f;

static const float fCLIPPING_EPSILON = 1e-8f;


static u_int ClipPolyToPlane( Asura_AVector3* pxInVertices, u_int uNumberOfVertices, Asura_AVector3* pxOutVertices, Asura_AVector3& xPlaneNormal, float fPlaneConstant )
{
	// KJFL - 10 July 2006 - I've taken this fn direct from the clipper I wrote for the decal system.
	// If it works out ok I'll refactor it to an Asura_Maths fn or somesuch - it may prove too costly to
	// use in this manner.

	u_int uNewVertexCount = 0;

	float fDot[9];

	{
		bool bAllNegative = true;
		for (u_int u=0; u<uNumberOfVertices; u++)
		{
			fDot[u] = pxInVertices[u]*xPlaneNormal + fPlaneConstant;
			if (fDot[u]>=0.0f)
			{
				bAllNegative = false;
			}
		}

		/// Kevin Lea - 14/11/2002 10:49:43 - early exit if all vertices would be clipped
		if (bAllNegative) return 0;
	}

	for (u_int u=0; u<uNumberOfVertices; u++)
	{
		u_int v = (u!=0) ? u-1 : uNumberOfVertices-1; // previous vertex

		if (fDot[u]<0.0f)
		{
			if (fDot[v]>=0.0f)
			{
				// Current vertex is on negative side of plane,
				// but previous vertex is on positive side.
				Asura_AVector3 xEdge = pxInVertices[v] - pxInVertices[u];

				// KJFL - 7 August 2008 - in theory the dot product of xEdge and xPlaneNormal must be >0,
				// since we have one vertex behind the plane and one in front, i.e. the line between the 
				// vertices can't be at right angles to the plane normal. In practice due to floating point
				// error once in a blue moon the dot product can come up zero (seen on PSP), hence the raising
				// to a small epsilon with the Max fn.
				float f = Asura_Maths::Max( xEdge * xPlaneNormal, fCLIPPING_EPSILON);
				float fLambda = fDot[v] / f;

				// Asura_Assertf( fLambda >= -0.0001f && fLambda <= 1.0001f, ("Out of bounds: %f",Asura_Maths::FtoD(fLambda)));
				if (fLambda < 0.0f) fLambda = 0.0f;
				if (fLambda > 1.0f) fLambda = 1.0f;

				pxOutVertices[uNewVertexCount] = pxInVertices[v] - xEdge * fLambda;
				uNewVertexCount++;	
			}
		}
		else
		{
			if (fDot[v]<0.0f)
			{
				// Current vertex is on positive side of plane,
				// but previous vertex is on negative side.
				const Asura_AVector3 xEdge = pxInVertices[u] - pxInVertices[v];

				// KJFL - 7 August 2008 - in theory the dot product of xEdge and xPlaneNormal must be >0,
				// since we have one vertex behind the plane and one in front, i.e. the line between the 
				// vertices can't be at right angles to the plane normal. In practice due to floating point
				// error once in a blue moon the dot product can come up zero (seen on PSP), hence the raising
				// to a small epsilon with the Max fn.
				float f = Asura_Maths::Max( xEdge * xPlaneNormal, fCLIPPING_EPSILON);
				float fLambda = fDot[u] / f;

				// Asura_Assertf( fLambda >= -0.0001f && fLambda <= 1.0001f, ("Out of bounds: %f",Asura_Maths::FtoD(fLambda)));
				if (fLambda < 0.0f) fLambda = 0.0f;
				if (fLambda > 1.0f) fLambda = 1.0f;

				pxOutVertices[uNewVertexCount] = pxInVertices[u] - xEdge * fLambda;
				uNewVertexCount++;	
			}
			
			/// Kevin Lea - 14/11/2002 11:22:04 - include current vertex
			pxOutVertices[uNewVertexCount] = pxInVertices[u];
			uNewVertexCount++;
		}
	}


	return uNewVertexCount;
}


static u_int ClipEdgeToPlane( Asura_AVector3* pxInVertices, Asura_AVector3* pxOutVertices, Asura_AVector3& xPlaneNormal, float fPlaneConstant )
{
	// modified version of ClipPolyToPlane to just deal with just edges (2 vert line segments)

	float fDot0 = pxInVertices[0]*xPlaneNormal + fPlaneConstant;
	float fDot1 = pxInVertices[1]*xPlaneNormal + fPlaneConstant;

	bool bVert0Behind = (fDot0 < 0.0f);
	bool bVert1Behind = (fDot1 < 0.0f);

	if ( bVert0Behind && bVert1Behind )
	{
		// both behind plane - edge completely culled
		return 0;
	}
	else if ( bVert0Behind )
	{
		// vertex 0 is on negative side of plane, but previous vertex 1 is on positive side.

		Asura_AVector3 xEdge = pxInVertices[1] - pxInVertices[0];
		
		// KJFL - 7 August 2008 - in theory the dot product of xEdge and xPlaneNormal must be >0,
		// since we have one vertex behind the plane and one in front, i.e. the line between the 
		// vertices can't be at right angles to the plane normal. In practice due to floating point
		// error once in a blue moon the dot product can come up zero (seen on PSP), hence the raising
		// to a small epsilon with the Max fn.
		float f = Asura_Maths::Max( xEdge * xPlaneNormal, fCLIPPING_EPSILON);
		float fLambda = fDot1 / f;

		// Asura_Assertf( fLambda >= -0.0001f && fLambda <= 1.0001f, ("Out of bounds: %f",Asura_Maths::FtoD(fLambda)));
		if (fLambda < 0.0f) fLambda = 0.0f;
		if (fLambda > 1.0f) fLambda = 1.0f;

		pxOutVertices[0] = pxInVertices[1] - xEdge * fLambda;

		pxOutVertices[1] = pxInVertices[1];

		return 2;

	}
	else if (bVert1Behind )
	{
		Asura_AVector3 xEdge = pxInVertices[0] - pxInVertices[1];
		
		// KJFL - 7 August 2008 - in theory the dot product of xEdge and xPlaneNormal must be >0,
		// since we have one vertex behind the plane and one in front, i.e. the line between the 
		// vertices can't be at right angles to the plane normal. In practice due to floating point
		// error once in a blue moon the dot product can come up zero (seen on PSP), hence the raising
		// to a small epsilon with the Max fn.
		float f = Asura_Maths::Max( xEdge * xPlaneNormal, fCLIPPING_EPSILON);
		float fLambda = fDot0 / f;

		// Asura_Assertf( fLambda >= -0.0001f && fLambda <= 1.0001f, ("Out of bounds: %f",Asura_Maths::FtoD(fLambda)));
		if (fLambda < 0.0f) fLambda = 0.0f;
		if (fLambda > 1.0f) fLambda = 1.0f;

		pxOutVertices[0] = pxInVertices[0];

		pxOutVertices[1] = pxInVertices[0] - xEdge * fLambda;

		return 2;
	}
	else
	{
		pxOutVertices[0] = pxInVertices[0];
		pxOutVertices[1] = pxInVertices[1];
		return 2;
	}
}


static u_int ClipPolyToPoly( const Asura_AVector3* pxClipperVertices, const u_int uNumberOfClipperVertices, const Asura_AVector3* pxClippeeVertices, u_int uNumberOfClippeeVertices, Asura_AVector3* pxClippedVertices )
{
	// need at least 3 verts in the clipper poly to make a normal (e.g. we need a tri at least, not an edge)
	if ( uNumberOfClipperVertices < 3 )
	{
		Asura_Assert( false, "Too few verts in clipper poly" );
		return 0;
	}

	// and the clippee needs to be a poly
	if ( uNumberOfClippeeVertices < 3 )
	{
		Asura_Assert( false, "Too few verts in clippee poly" );
		return 0;
	}

	Asura_AVector3 xClipperNormal = (pxClipperVertices[1] - pxClipperVertices[0])^(pxClipperVertices[2] - pxClipperVertices[0]);
	float fMagnitude = xClipperNormal.NormaliseReturningMagnitude();

	if (fMagnitude == 0.0f)
	{
		Asura_Assert( false, "Clipper poly is degenerate" );
		return 0;
	}

	u_int uNumberOfClippedVertices = uNumberOfClippeeVertices;

	// need assert on max clipped verts
	Asura_Mem::Copy( pxClippedVertices, pxClippeeVertices, sizeof( Asura_AVector3 ) * uNumberOfClippeeVertices );

	Asura_AVector3 axBuffer[10];


	u_int uStart = uNumberOfClipperVertices - 1;

	Asura_AVector3 xCentre = pxClipperVertices[0] + pxClipperVertices[1] + pxClipperVertices[2] + pxClipperVertices[3];

	xCentre *=0.25f;

	for ( u_int uEnd = 0; uEnd < uNumberOfClipperVertices; uStart = uEnd, ++uEnd )
	{
		Asura_AVector3 xEdgeStart = pxClipperVertices[ uStart ];
		Asura_AVector3 xEdgeEnd = pxClipperVertices[ uEnd ];

		Asura_AVector3 xPlaneNormal = ( xEdgeEnd - xEdgeStart ) ^ xClipperNormal;


		// probably unused variable when assertions off; may not need to normalise anyway
#if defined(ASURA_ASSERT)
		float fPlaneNormalMagnitude =
#endif	// defined(ASURA_ASSERT)
			xPlaneNormal.NormaliseReturningMagnitude();

		Asura_Assert( fPlaneNormalMagnitude != 0, "edge parallel to normal!" );

		float fPlaneConstant = -xEdgeStart * xPlaneNormal;


		if ( xCentre*xPlaneNormal + fPlaneConstant < 0.0f )
		{
			xPlaneNormal = -xPlaneNormal;
			fPlaneConstant = -fPlaneConstant;
		}
		if ( xCentre*xPlaneNormal + fPlaneConstant < 0.0f )
		{
			return 0;
		}

		uNumberOfClippedVertices = ClipPolyToPlane( pxClippedVertices, uNumberOfClippedVertices, axBuffer, xPlaneNormal, fPlaneConstant );

		if (uNumberOfClippedVertices > 8)
		{
			return 0;
		}

		// need assert on max clipped verts
		Asura_Mem::Copy( pxClippedVertices, axBuffer, sizeof( Asura_AVector3 ) * uNumberOfClippedVertices );

		if ( uNumberOfClippedVertices == 0 ) return 0;
	}

	return uNumberOfClippedVertices;
}

static u_int ClipEdgeToPoly( const Asura_AVector3* pxClipperVertices, const u_int uNumberOfClipperVertices, const Asura_AVector3* pxClippeeVertices, Asura_AVector3* pxClippedVertices )
{
	u_int uNumberOfClippeeVertices = 2;

	// need at least 3 verts in the clipper poly to make a normal (e.g. we need a tri at least, not an edge)
	if ( uNumberOfClipperVertices < 3 )
	{
		Asura_Assert( false, "Too few verts in clipper poly" );
		return 0;
	}

	Asura_AVector3 xClipperNormal = (pxClipperVertices[1] - pxClipperVertices[0])^(pxClipperVertices[2] - pxClipperVertices[0]);
	xClipperNormal.Normalise();

	u_int uNumberOfClippedVertices = uNumberOfClippeeVertices;

	// need assert on max clipped verts
	Asura_Mem::Copy( pxClippedVertices, pxClippeeVertices, sizeof( Asura_AVector3 ) * uNumberOfClippeeVertices );

	Asura_AVector3 axBuffer[2];


	u_int uStart = uNumberOfClipperVertices - 1;

	Asura_AVector3 xCentre = pxClipperVertices[0] + pxClipperVertices[1] + pxClipperVertices[2] + pxClipperVertices[3];

	xCentre *=0.25f;

	for ( u_int uEnd = 0; uEnd < uNumberOfClipperVertices; uStart = uEnd, ++uEnd )
	{
		Asura_AVector3 xEdgeStart = pxClipperVertices[ uStart ];
		Asura_AVector3 xEdgeEnd = pxClipperVertices[ uEnd ];

		Asura_AVector3 xPlaneNormal = ( xEdgeEnd - xEdgeStart ) ^ xClipperNormal;


		// probably unused variable when assertions off; may not need to normalise anyway
#if defined(ASURA_ASSERT)
		float fMagnitude =
#endif	// defined(ASURA_ASSERT)
			xPlaneNormal.NormaliseReturningMagnitude();

		Asura_Assert( fMagnitude != 0, "edge parallel to normal!" );

		float fPlaneConstant = -xEdgeStart * xPlaneNormal;


		if ( xCentre*xPlaneNormal + fPlaneConstant < 0.0f )
		{
			xPlaneNormal = -xPlaneNormal;
			fPlaneConstant = -fPlaneConstant;
		}
		if ( xCentre*xPlaneNormal + fPlaneConstant < 0.0f )
		{
			return 0;
		}

		uNumberOfClippedVertices = ClipEdgeToPlane( pxClippedVertices, axBuffer, xPlaneNormal, fPlaneConstant );

		if ( uNumberOfClippedVertices == 0 )
		{
			return 0;
		}

		Asura_Mem::Copy( pxClippedVertices, axBuffer, sizeof( Asura_AVector3 ) * uNumberOfClippedVertices );

	}

	return uNumberOfClippedVertices;
}


static void ProjectVertexOntoEdge( const Asura_AVector3& xVertex, const Asura_AVector3* pxEdge, Asura_AVector3& xProjectedVertex )
{
	const Asura_AVector3 xAlongEdge = pxEdge[1] - pxEdge[0];
	const Asura_AVector3 xEdgeToVertex = xVertex - pxEdge[0];

	const float fLengthOfEdgeSquared = xAlongEdge.MagnitudeSquared();

	if ( fLengthOfEdgeSquared > 0.0f )
	{
		// find the distance along the edge:
		float fLambda = ( xAlongEdge * xEdgeToVertex ) / fLengthOfEdgeSquared;

		fLambda = Asura_Maths::ClampToRange( fLambda, 0.0f, 1.0f );

		xProjectedVertex = pxEdge[0] + xAlongEdge * fLambda;
	}
	else
	{
		xProjectedVertex = pxEdge[0];
	}
}

static void ProjectEdgeOntoEdge( const Asura_AVector3* pxEdgeA, const Asura_AVector3* pxEdgeB, Asura_AVector3& xProjectedVertexA, Asura_AVector3& xProjectedVertexB )
{
	const Asura_AVector3 xAlongEdgeA = pxEdgeA[1] - pxEdgeA[0];
	const Asura_AVector3 xAlongEdgeB = pxEdgeB[1] - pxEdgeB[0];

	const Asura_AVector3 xBetweenEdges = xAlongEdgeA ^ xAlongEdgeB;

	const Asura_AVector3 xM = xBetweenEdges ^ xAlongEdgeB;

	float fLambda = ( (pxEdgeB[0] - pxEdgeA[0]) * xM ) / ( xAlongEdgeA * xM );

	fLambda = Asura_Maths::ClampToRange( fLambda, 0.0f, 1.0f );

	xProjectedVertexA = pxEdgeA[0] + xAlongEdgeA * fLambda;

	ProjectVertexOntoEdge( xProjectedVertexA, pxEdgeB, xProjectedVertexB );
}

static void ProjectVertexOntoFace( const Asura_AVector3& xVertex, const Asura_AVector3* pxFace, Asura_AVector3& xProjectedVertex )
{
	Asura_AVector3 xFaceNormal = (pxFace[1] - pxFace[0])^(pxFace[2] - pxFace[0]);

	// hmm - check maths, can I drop the normalise?
	xFaceNormal.Normalise();

	float fDistanceFromFace = (xVertex - pxFace[0]) * xFaceNormal;

	xProjectedVertex = xVertex - xFaceNormal * fDistanceFromFace;
}

static u_int ProjectEdgeOntoFace( const Asura_AVector3* pxEdge, const Asura_AVector3* pxFace, Asura_AVector3* pxProjectedVerticesEdge, Asura_AVector3* pxProjectedVerticesFace, u_int uNumberOfVertsInFace )
{
	// take the edge and clip it with the planes made by the edges of the face and the face normal

	u_int uNumberOfPoints = ClipEdgeToPoly( pxFace, uNumberOfVertsInFace, pxEdge, pxProjectedVerticesEdge );

	if (uNumberOfPoints < 2)
	{
		//		Asura_Assert( false, "Edge completely clipped?" );
		return 0;
	}
	Asura_Assert( uNumberOfPoints == 2, "Code Bug: Not an Edge any more" );


	for ( u_int u=0; u<uNumberOfPoints; ++u )
	{
		ProjectVertexOntoFace( pxProjectedVerticesEdge[u], pxFace, pxProjectedVerticesFace[u] );
	}

	return uNumberOfPoints;
}
static u_int ProjectFaceOntoFace( const Asura_AVector3* pxFaceA, const Asura_AVector3* pxFaceB, Asura_AVector3* pxProjectedVerticesFaceA, Asura_AVector3* pxProjectedVerticesFaceB, u_int uNumberOfVertsInFaceA, u_int uNumberOfVertsInFaceB )
{
	u_int uNumberOfPoints = ClipPolyToPoly( pxFaceA, uNumberOfVertsInFaceA, pxFaceB, uNumberOfVertsInFaceB, pxProjectedVerticesFaceB );

	if (uNumberOfPoints <= 2)
	{
		//		Asura_Assert( false, "Face completely clipped?" );
		return 0;
	}

	for ( u_int u=0; u<uNumberOfPoints; ++u )
	{
		ProjectVertexOntoFace( pxProjectedVerticesFaceB[u], pxFaceA, pxProjectedVerticesFaceA[u] );
	}


	return uNumberOfPoints;
}


u_int Asura_Physics_CCD::FindContactPointsFromSupportFeatures( Asura_AVector3* pxContactPointsForA, Asura_AVector3* pxContactPointsForB, const Asura_Physics_RigidBodyInstance* pxInstanceA, const Asura_Physics_RigidBodyInstance* pxInstanceB, const Asura_AVector3& xAxis )
{

	// Find the support features for both objects. This can be a face (4 verts), and edge (2 verts) or just a single vertex
	Asura_AVector3 axSupportPointsForA[4];
	Asura_AVector3 axSupportPointsForB[4];

	u_int uNumberOfSupportPointsForA = pxInstanceA->GetSupportFeature( axSupportPointsForA, xAxis );
	u_int uNumberOfSupportPointsForB = pxInstanceB->GetSupportFeature( axSupportPointsForB, -xAxis );

	// We have 9 possible combinations. Might be able to minimise tests, but easiest if we do it as switches r
	switch( uNumberOfSupportPointsForA )
	{
		default:
		{
			Asura_Assert( false, "Unhandled number of support points" );
			return 0;
		}
		case 1: // vertex
		{
			// all these contacts will just be one vertex, and for A it's the support point
			pxContactPointsForA[0] = axSupportPointsForA[0];

			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-vertex case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					return 1;
				}
				case 2: // vertex-edge case
				{
					ProjectVertexOntoEdge( axSupportPointsForA[0], axSupportPointsForB, pxContactPointsForB[0] );
					return 1;
				}
				case 4: // vertex-face case
				{
					ProjectVertexOntoFace( axSupportPointsForA[0], axSupportPointsForB, pxContactPointsForB[0] );
					return 1;
				}
			}
		}
		case 2: // edge
		{
			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-edge case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					ProjectVertexOntoEdge( axSupportPointsForB[0], axSupportPointsForA, pxContactPointsForA[0] );
					return 1;
				}
				case 2: // edge-edge case
				{
					ProjectEdgeOntoEdge( axSupportPointsForA, axSupportPointsForB, pxContactPointsForA[0], pxContactPointsForB[0] );
					return 1;
				}
				case 4: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForA, axSupportPointsForB, pxContactPointsForA, pxContactPointsForB, 4 );

				}
			}
		}
		case 4:
		{
			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-face case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					ProjectVertexOntoFace( axSupportPointsForB[0], axSupportPointsForA, pxContactPointsForA[0] );
					return 1;
				}
				case 2: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForB, axSupportPointsForA, pxContactPointsForB, pxContactPointsForA, 4 );

				}
				case 4: // face-face case
				{
					return ProjectFaceOntoFace( axSupportPointsForA, axSupportPointsForB, pxContactPointsForA, pxContactPointsForB, 4, 4 );
				}
			}
		}
	}


}

bool Asura_Physics_CCD::FindPenetration_OBBVsOBB( const Asura_OBB& xObjectA, const Asura_OBB& xObjectB, float& fPenetration, Asura_AVector3& xAxis )
{
	const Asura_AMatrix3x3& xObjectAOrient = xObjectA.GetOrientation();
	const Asura_AMatrix3x3& xObjectBOrient = xObjectB.GetOrientation();

	// The potential separating axes are formed from the edges of the
	// OBBs, and the vectors perpendicular to the edge pairings. This
	// makes 15 axes to test

	const unsigned int uNumberOfSeparatingAxes = 15;

	// KJFL - 20 July 2006 - TODO - could save some time by calculating these as and when needed?
	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xObjectAOrient.GetColumn0(),
		xObjectAOrient.GetColumn1(),
		xObjectAOrient.GetColumn2(),

		xObjectBOrient.GetColumn0(),
		xObjectBOrient.GetColumn1(),
		xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn2(),
	};

	u_int uBestAxis = uNumberOfSeparatingAxes;
	float fBestMinSeparation = Asura_Maths::LargeFloat;

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		float fMinA = 0.0f;
		float fMaxA = 0.0f;
		float fMinB = 0.0f;
		float fMaxB = 0.0f;

		if ( axAxes[u].MagnitudeSquared() < 0.001f ) continue;

		// PHYSICSTODO: can we delay normalisation? Would need extra divides
		// to rescale separation distances though, so may not be a win
		axAxes[u].Normalise();

		xObjectA.GetExtentsAlongAxis( axAxes[u], fMinA, fMaxA );
		xObjectB.GetExtentsAlongAxis( axAxes[u], fMinB, fMaxB );

		float fMinSeparation = Asura_Maths::Min( fMaxA - fMinB, fMaxB - fMinA );

		if ( fMinSeparation < 0.0f ) return false;

		if ( fMinSeparation < fBestMinSeparation )
		{
			uBestAxis = u;
			fBestMinSeparation = fMinSeparation;
		}
	}

	if ( uBestAxis == uNumberOfSeparatingAxes )
	{
		return false;
    }

	fPenetration = fBestMinSeparation;
	xAxis = axAxes[uBestAxis];
	
	// Arbitrarily make the axis point towards box A.
	const Asura_AVector3 xAToB = xObjectB.GetCentre() - xObjectA.GetCentre();
	if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

	// looks like they overlap
	return true;
}

bool Asura_Physics_CCD::FindPenetration_SphereVsTri( const Asura_Physics_Tri& xTri, const Asura_AVector3& xSphereCentre, const float fRadius, float& fPenetration, Asura_AVector3& xPenetrationAxis, Asura_AVector3& xContactPoint )
{
	const Asura_AVector3* xTriVerts = xTri.GetVertexArray();

	Asura_AVector3 xNearestPoint;

	FindClosestPointOnTriangle( xSphereCentre, xTriVerts[0], xTriVerts[1], xTriVerts[2], xNearestPoint );

	xPenetrationAxis = xSphereCentre - xNearestPoint;

	if (xPenetrationAxis.MagnitudeSquared() > fRadius * fRadius) return false;

	Asura_AVector3 xNormal;
	xTri.GetUnnormalisedNormal( xNormal );

	// TODO - cope with the situation where the sphere centre is behind the poly,
	// or ignore it like this?
	if ( xPenetrationAxis * xNormal < 0.0f) return false;

	xContactPoint = xNearestPoint;
	
	float fSeparation = xPenetrationAxis.NormaliseReturningMagnitude();

	fPenetration = fRadius - fSeparation;

	return true;
}

bool Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsOBB( const Asura_OBB& xObjectA, const Asura_OBB& xObjectB, const Asura_AVector3& xRelativeVelocity, float& fTimeToImpact, Asura_AVector3& xAxisUsed, bool bIgnoreInitialOverlap /*=false*/ )
{
	const Asura_AMatrix3x3& xObjectAOrient = xObjectA.GetOrientation();
	const Asura_AMatrix3x3& xObjectBOrient = xObjectB.GetOrientation();

	// The potential separating axes are formed from the edges of the
	// OBBs, and the vectors perpendicular to the edge pairings. This
	// makes 15 axes to test

	const unsigned int uNumberOfSeparatingAxes = 15;

	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn0(),
		xObjectAOrient.GetColumn1(),
		xObjectAOrient.GetColumn2(),

		xObjectBOrient.GetColumn0(),
		xObjectBOrient.GetColumn1(),
		xObjectBOrient.GetColumn2()
	};

	float fOverlapMinTime = -0.0001f;
	float fOverlapMaxTime = 1.0f;
	u_int uBestAxis = uNumberOfSeparatingAxes;

	const Asura_AVector3 xAToB = xObjectB.GetCentre() - xObjectA.GetCentre();

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		Asura_AVector3& xAxis = axAxes[u];

		// KJFL - 27 June 2006 - ah, it's possible some of the axes are invalid
		// because they're zero vectors, simply because the edges of the two 
		// objects might be parallel (cross products above would give zero vectors).

		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

	//	if ( xRelativeVelocity * xAxis >= 0.0f ) continue;

		xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects

		float fCentreA = 0.0f;
		float fExtentsA = 0.0f;

		xObjectA.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreA, fExtentsA );

		float fCentreB = 0.0f;
		float fExtentsB = 0.0f;

		xObjectB.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreB, fExtentsB );

		float fRelativeVelocityAlongAxis = ( xRelativeVelocity ) * xAxis ;

		float fMinkowskiSumMin = fCentreA - fCentreB - fExtentsA - fExtentsB;
		float fMinkowskiSumMax = fCentreA - fCentreB + fExtentsA + fExtentsB;

		float fRayMin = Asura_Maths::Min( 0.0f, fRelativeVelocityAlongAxis );
		float fRayMax = Asura_Maths::Max( 0.0f, fRelativeVelocityAlongAxis );

		// test for a no overlap case
		if ( 0.0f < fMinkowskiSumMin + fRayMin || 0.0f > fMinkowskiSumMax + fRayMax )
		{
			// these objects definitely don't overlap
			return false;
		}

		if ( Asura_Maths::Abs( fRelativeVelocityAlongAxis ) > 0.0000001f )
		{
			float fTime1 = ( -fMinkowskiSumMax ) / ( fRelativeVelocityAlongAxis );
			float fTime2 = ( -fMinkowskiSumMin ) / ( fRelativeVelocityAlongAxis );

			float fMinTime = Asura_Maths::Min( fTime1, fTime2 ) ;
			float fMaxTime = Asura_Maths::Max( fTime1, fTime2 );

			if ( fMinTime > fOverlapMinTime )
			{
				fOverlapMinTime = fMinTime;
				uBestAxis = u;
			}
			
			fOverlapMaxTime = Asura_Maths::Min( fOverlapMaxTime, fMaxTime );

			if ( fOverlapMaxTime < fOverlapMinTime ) return false;
		}
	}

	if ( uBestAxis >= uNumberOfSeparatingAxes  )
	{
		if ( bIgnoreInitialOverlap )
		{
			return false;
		}
		else
		{
			// we're stuck in an object - cancel movement
			fTimeToImpact = 0.0f;

			// make a normal counter to the relative velocity
			xAxisUsed = -xRelativeVelocity;
			xAxisUsed.Normalise();
			return true;
		}
	}

	if ( fOverlapMinTime < fTimeToImpact )
	{
		fTimeToImpact = fOverlapMinTime;
		xAxisUsed = axAxes[uBestAxis];
		return true;
	}

	return false;
}

float Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsTri( const Asura_OBB& xOBB, const Asura_AVector3& xVelocity, const Asura_Physics_Tri& xTri, Asura_AVector3& xAxisUsed, u_int& uFirstPlane, bool& bHitEdge, bool bCheckVelocityVsNormal )
{
	bHitEdge = false;

	// value to return to indicate time to intersection does happen or is outside the 0.0f->1.0f range
	const float fNoIntersectionTime = 2.0f;

	// KJFL - 27 September 2006 - 
	// 
	// Let's see... there are 13 potential axes of separation we need to test:
	//
	// 1 face normal from the tri
	// 3 unique face normals from the OBB (the other three faces are just sign flips)
	// 9 unique axes given by the cross products of combinations of the edges of both objects

	const unsigned int uNumberOfSeparatingAxes = 13;

	Asura_AVector3 xTriNormal, xTriEdge0, xTriEdge1, xTriEdge2;

	xTri.GetUnnormalisedNormal( xTriNormal );

	if ( bCheckVelocityVsNormal )
	{
		if ( xTriNormal * xVelocity >= 0.0f ) return fNoIntersectionTime;
	}


	xTri.GetEdge0( xTriEdge0 );
	xTri.GetEdge1( xTriEdge1 );
	xTri.GetEdge2( xTriEdge2 );

	const Asura_AMatrix3x3& xObjectOrient = xOBB.GetOrientation();


	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xObjectOrient.GetColumn0() ^ xTriEdge0,
			xObjectOrient.GetColumn0() ^ xTriEdge1,
			xObjectOrient.GetColumn0() ^ xTriEdge2,

			xObjectOrient.GetColumn1() ^ xTriEdge0,
			xObjectOrient.GetColumn1() ^ xTriEdge1,
			xObjectOrient.GetColumn1() ^ xTriEdge2,

			xObjectOrient.GetColumn2() ^ xTriEdge0,
			xObjectOrient.GetColumn2() ^ xTriEdge1,
			xObjectOrient.GetColumn2() ^ xTriEdge2,

			xObjectOrient.GetColumn0(),
			xObjectOrient.GetColumn1(),
			xObjectOrient.GetColumn2(),
			xTriNormal,
	};


	float fOverlapMinTime = -0.0001f;
	float fOverlapMaxTime = 1.0f;
	u_int uBestAxis = uNumberOfSeparatingAxes;

	Asura_AVector3 xRelativeVelocity = xVelocity;

	const Asura_AVector3 xAToB = xTri.GetCentre() - xOBB.GetCentre();
	//	u_int uOverlap = 0;

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		u_int uAxis = (uFirstPlane + u);
		if (uAxis >= uNumberOfSeparatingAxes) uAxis -=uNumberOfSeparatingAxes;

		Asura_AVector3& xAxis = axAxes[uAxis];


		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

		// KJFL - in theory you can do this loop without normalising and save some 
		// cycles, but in practice floating point accuracy means you get different 
		// results to other algorithms.
		xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects

		float fCentreA = 0.0f;
		float fExtentsA = 0.0f;

		xOBB.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreA, fExtentsA );

		float fCentreB = 0.0f;
		float fExtentsB = 0.0f;

		xTri.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreB, fExtentsB );

		float fRelativeVelocityAlongAxis = ( xRelativeVelocity ) * xAxis ;

		float fMinkowskiSumMin = fCentreA - fCentreB - fExtentsA - fExtentsB;
		float fMinkowskiSumMax = fCentreA - fCentreB + fExtentsA + fExtentsB;

		float fRayMin = Asura_Maths::Min( 0.0f, fRelativeVelocityAlongAxis );
		float fRayMax = Asura_Maths::Max( 0.0f, fRelativeVelocityAlongAxis );

		// test for a no overlap case
		if ( 0.0f < fMinkowskiSumMin + fRayMin || 0.0f > fMinkowskiSumMax + fRayMax )
		{
			// these objects definitely don't overlap
			uFirstPlane = uAxis;
			return fNoIntersectionTime;
		}


		if ( Asura_Maths::Abs( fRelativeVelocityAlongAxis ) > 0.0000001f )
		{
			float fTime1 = ( -fMinkowskiSumMax ) / ( fRelativeVelocityAlongAxis );
			float fTime2 = ( -fMinkowskiSumMin ) / ( fRelativeVelocityAlongAxis );

			float fMinTime = Asura_Maths::Min( fTime1, fTime2 ) ;
			float fMaxTime = Asura_Maths::Max( fTime1, fTime2 );

		
			if ( fMinTime > fOverlapMinTime )
			{
				fOverlapMinTime = fMinTime;
				uBestAxis = uAxis;
			}

			fOverlapMaxTime = Asura_Maths::Min( fOverlapMaxTime, fMaxTime );
		}

		if (fOverlapMaxTime < fOverlapMinTime || fOverlapMaxTime <= 0.0f )
		{
			// these objects definitely don't overlap
			uFirstPlane = uAxis;
			return fNoIntersectionTime;
		}

	}

	if ( uBestAxis >= uNumberOfSeparatingAxes  )
	{
		// we're stuck in something
		//Asura_Assert( uBestAxis < uNumberOfSeparatingAxes, "Code bug: Invalid axis" );

		return -1.0f;
	}

	// We've found the axis of separation
	xAxisUsed = axAxes[uBestAxis];

	xAxisUsed.Normalise();

	if (fOverlapMinTime >= 1.0f)
	{
		// there isn't a collision this frame, but there may be later
		return fNoIntersectionTime;
	}
	else if ( fOverlapMinTime >= 0.0f )
	{
		// is this an edge case?

		bHitEdge = ( (axAxes[uNumberOfSeparatingAxes-1] * xAxisUsed ) < 0.999f );

		return fOverlapMinTime;
	}

	//Asura_Assertf( false, ("fOverlapMinTime %f", fOverlapMinTime ));

	// tiny negative value, let's ignore
	return 0.0f;
}

#if 0
float Asura_Physics_CCD::FindTimeOfIntersection_SweptSpheresVsTri( const Asura_Physics_SphereGroup& xSpheres, const Asura_AVector3& xVelocity, const float fTimeToImpact, const Asura_Physics_Tri& xTri, Asura_AVector3& xAxisUsed, Asura_AVector3& xContactPoint )
{
	float fMaxAllowed = fTimeToImpact;
    {
		Asura_AVector3 xNormal;

		xTri.GetUnnormalisedNormal( xNormal );

		const float fSphereRadius = xSpheres.GetSphereRadius();
		const Asura_AVector3 xOffset = xNormal * fSphereRadius;

		const Asura_AVector3* pxTriVerts = xTri.GetVertexArray();

		const Asura_AVector3 xV0 = pxTriVerts[0] + xOffset;
		const Asura_AVector3 xV1 = pxTriVerts[1] + xOffset;
		const Asura_AVector3 xV2 = pxTriVerts[2] + xOffset;


		for ( u_int v=0; v<xSpheres.GetNumberOfSpheres(); ++v )
		{
			bool bFacingTowards;
			float fHitDistance;

			const Asura_AVector3& xSphereStart = xSpheres.GetSphereCentres()[v];

			// first check against the face
			if( Asura_Physics_System::RayIntersectsTri( xSphereStart, xVelocity, xV0, xV1, xV2, bFacingTowards, fHitDistance ) )
			{
				if ( fHitDistance >= 0.0f && bFacingTowards )
				{
					if ( fHitDistance <= fMaxAllowed )
					{
						fMaxAllowed = fHitDistance;
						xAxisUsed = xNormal;

					}
				}	
				else if ( fHitDistance > -fSphereRadius )
				{
					// we're already overlapping with a polygon I'm afraid
					return fHitDistance;
				}
			}

			const Asura_AVector3& xSphereDir = xVelocity;

			bool bHitSomething = false;

			// then the three edges
			bHitSomething |= Asura_Physics_System::SweptSphereVsLineSegment( fSphereRadius, xSphereStart, xSphereDir, pxTriVerts[0], pxTriVerts[1], xContactPoint, fMaxAllowed );
			bHitSomething |= Asura_Physics_System::SweptSphereVsLineSegment( fSphereRadius, xSphereStart, xSphereDir, pxTriVerts[1], pxTriVerts[2], xContactPoint, fMaxAllowed );
			bHitSomething |= Asura_Physics_System::SweptSphereVsLineSegment( fSphereRadius, xSphereStart, xSphereDir, pxTriVerts[2], pxTriVerts[0], xContactPoint, fMaxAllowed );


			// and against the vertices themselves
			bHitSomething |= Asura_Physics_System::SweptSphereVsPoint( fSphereRadius, xSphereStart, xSphereDir, pxTriVerts[0], xContactPoint, fMaxAllowed );
			bHitSomething |= Asura_Physics_System::SweptSphereVsPoint( fSphereRadius, xSphereStart, xSphereDir, pxTriVerts[1], xContactPoint, fMaxAllowed );
			bHitSomething |= Asura_Physics_System::SweptSphereVsPoint( fSphereRadius, xSphereStart, xSphereDir, pxTriVerts[2], xContactPoint, fMaxAllowed );

			if (bHitSomething)
			{
				xAxisUsed = xSphereStart + xVelocity*fMaxAllowed - xContactPoint;
				xAxisUsed.Normalise();
			}
		}
	}

	return 0.0f;
}
#endif

bool Asura_Physics_CCD::DoesOBBIntersectTri( const Asura_OBB& xOBB, const Asura_Physics_Tri& xTri, float& fPenetration, Asura_AVector3& xPenetrationAxis )
{
	const Asura_AMatrix3x3& xObjectOrient = xOBB.GetOrientation();

	// KJFL - 27 September 2006 - 
	// 
	// Let's see... there are 13 potential axes of separation we need to test:
	//
	// 1 face normal from the tri
	// 3 unique face normals from the OBB (the other three faces are just sign flips)
	// 9 unique axes given by the cross products of combinations of the edges of both objects

	const unsigned int uNumberOfSeparatingAxes = 13;

	Asura_AVector3 xTriNormal, xTriEdge0, xTriEdge1, xTriEdge2;

	xTri.GetUnnormalisedNormal( xTriNormal );
	xTri.GetEdge0( xTriEdge0 );
	xTri.GetEdge1( xTriEdge1 );
	xTri.GetEdge2( xTriEdge2 );

	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
			xObjectOrient.GetColumn0() ^ xTriEdge0,
			xObjectOrient.GetColumn0() ^ xTriEdge1,
			xObjectOrient.GetColumn0() ^ xTriEdge2,

			xObjectOrient.GetColumn1() ^ xTriEdge0,
			xObjectOrient.GetColumn1() ^ xTriEdge1,
			xObjectOrient.GetColumn1() ^ xTriEdge2,

			xObjectOrient.GetColumn2() ^ xTriEdge0,
			xObjectOrient.GetColumn2() ^ xTriEdge1,
			xObjectOrient.GetColumn2() ^ xTriEdge2,

			xObjectOrient.GetColumn0(),
			xObjectOrient.GetColumn1(),
			xObjectOrient.GetColumn2(),
			
			xTriNormal,
	};

	float afSeparationDistance[uNumberOfSeparatingAxes];
	u_int uBestAxis = uNumberOfSeparatingAxes;

	const Asura_AVector3 xAToB = xTri.GetCentre() - xOBB.GetCentre();

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		afSeparationDistance[u] = Asura_Maths::LargeFloat / 4.0f;

		Asura_AVector3& xAxis = axAxes[u];

		// KJFL - 27 June 2006 - ah, it's possible some of the axes are invalid
		// because they're zero vectors, simply because the edges of the two 
		// objects might be parallel (cross products above would give zero vectors).

		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

		xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects

		float fCentreA = 0.0f;
		float fExtentsA = 0.0f;

		xOBB.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreA, fExtentsA );

		float fCentreB = 0.0f;
		float fExtentsB = 0.0f;

		xTri.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreB, fExtentsB );

		const float fMinkowskiSumMin = fCentreA - fCentreB - fExtentsA - fExtentsB;
		const float fMinkowskiSumMax = fCentreA - fCentreB + fExtentsA + fExtentsB;

		const float fRayMin = 0.0f;
		const float fRayMax = 0.0f;

		// test for a no overlap case
		if ( fRayMax < fMinkowskiSumMin || fRayMin > fMinkowskiSumMax )
		{
			// these objects definitely don't overlap
			return false;
		}

		afSeparationDistance[u] = Asura_Maths::Min( fMinkowskiSumMax, -fMinkowskiSumMin );

	}


	float fMinimumDistance = Asura_Maths::LargeFloat;

	{
		uBestAxis = uNumberOfSeparatingAxes;

		for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
		{
			// skip invalid axes (e.g. check for zero vectors)
			if ( axAxes[u].MagnitudeSquared() < 0.001f ) continue;

			const float f = afSeparationDistance[u];

			if ( f >= 0.0f && f < fMinimumDistance )
			{
				uBestAxis = u;
				fMinimumDistance = f;
			}
		}
	}

	if ( uBestAxis >= uNumberOfSeparatingAxes  )
	{
		// this is unexpected... let's assert and return a verdict of no collision
		Asura_Assert( uBestAxis < uNumberOfSeparatingAxes, "Invalid axis" );

		return false;
	}

	fPenetration = fMinimumDistance;
	xPenetrationAxis = axAxes[uBestAxis];

	// if the suggested penetration axis is facing away from the tri normal, ignore this poly
	// (this happens when you're actually on the wrong side of it, for example)
	if ( xPenetrationAxis * xTriNormal < 0.0f ) return false;

	// overlap found
	return true;
}

bool Asura_Physics_CCD::DoesOBBIntersectTri( const Asura_OBB& xOBB, const Asura_Physics_Tri& xTri )
{
	const Asura_AMatrix3x3& xObjectOrient = xOBB.GetOrientation();

	// ART - 03-09-2007
	// This is a cut-and-paste of the above code, we don't actually care about axes, so 
	// there are potentially better optimisations which could be used here.

	const unsigned int uNumberOfSeparatingAxes = 13;

	Asura_AVector3 xTriNormal, xTriEdge0, xTriEdge1, xTriEdge2;

	xTri.GetUnnormalisedNormal( xTriNormal );
	xTri.GetEdge0( xTriEdge0 );
	xTri.GetEdge1( xTriEdge1 );
	xTri.GetEdge2( xTriEdge2 );

	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xObjectOrient.GetColumn0() ^ xTriEdge0,
		xObjectOrient.GetColumn0() ^ xTriEdge1,
		xObjectOrient.GetColumn0() ^ xTriEdge2,

		xObjectOrient.GetColumn1() ^ xTriEdge0,
		xObjectOrient.GetColumn1() ^ xTriEdge1,
		xObjectOrient.GetColumn1() ^ xTriEdge2,

		xObjectOrient.GetColumn2() ^ xTriEdge0,
		xObjectOrient.GetColumn2() ^ xTriEdge1,
		xObjectOrient.GetColumn2() ^ xTriEdge2,

		xObjectOrient.GetColumn0(),
		xObjectOrient.GetColumn1(),
		xObjectOrient.GetColumn2(),

		xTriNormal,
	};

	const Asura_AVector3 xAToB = xTri.GetCentre() - xOBB.GetCentre();

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		Asura_AVector3& xAxis = axAxes[u];

		// KJFL - 27 June 2006 - ah, it's possible some of the axes are invalid
		// because they're zero vectors, simply because the edges of the two 
		// objects might be parallel (cross products above would give zero vectors).

		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

		xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects

		float fCentreA = 0.0f;
		float fExtentsA = 0.0f;

		xOBB.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreA, fExtentsA );

		float fCentreB = 0.0f;
		float fExtentsB = 0.0f;

		xTri.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreB, fExtentsB );

		const float fMinkowskiSumMin = fCentreA - fCentreB - fExtentsA - fExtentsB;
		const float fMinkowskiSumMax = fCentreA - fCentreB + fExtentsA + fExtentsB;

		const float fRayMin = 0.0f;
		const float fRayMax = 0.0f;

		// test for a no overlap case
		if ( fRayMax < fMinkowskiSumMin || fRayMin > fMinkowskiSumMax )
		{
			// these objects definitely don't overlap
			return false;
		}
	}

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		// skip invalid axes (e.g. check for zero vectors)
		if ( axAxes[u].MagnitudeSquared() >= 0.001f )
		{
			return(true);
		}
	}

	// No overlap found
	return(false);
}

// penetration axis points towards OBB
bool Asura_Physics_CCD::DoesOBBIntersectSphere( const Asura_OBB& xOBB, const Asura_AVector3& xSphereCentre, const float fRadius, float& fPenetration, Asura_AVector3& xPenetrationAxis )
{
	// find point on OBB nearest to sphere, and if the distance to that point from the sphere centre is less than the sphere radius we have an overlap

	// transform sphere into OBB's space
	Asura_AVector3 xSphereCentreInOBBSpace = xSphereCentre - xOBB.GetCentre();
	
	xSphereCentreInOBBSpace = xOBB.GetOrientation().ApplyTranspose( xSphereCentreInOBBSpace );


	const float fSX = xSphereCentreInOBBSpace.GetX();
	const float fSY = xSphereCentreInOBBSpace.GetY();
	const float fSZ = xSphereCentreInOBBSpace.GetZ();

	const float fPX = Asura_Maths::ClampToRange( fSX, -xOBB.GetHalfExtentX(), +xOBB.GetHalfExtentX() );
	const float fPY = Asura_Maths::ClampToRange( fSY, -xOBB.GetHalfExtentY(), +xOBB.GetHalfExtentY() );
	const float fPZ = Asura_Maths::ClampToRange( fSZ, -xOBB.GetHalfExtentZ(), +xOBB.GetHalfExtentZ() );

	Asura_AVector3 xDisp( fPX - fSX, fPY - fSY, fPZ - fSZ );
	
	float fDistanceSquared = xDisp.MagnitudeSquared();

	// if distance is greater than the radius, there's no intersection
	if (fDistanceSquared > fRadius * fRadius ) return false;

	if ( fDistanceSquared > 0.0f ) // penetrating, but centre of sphere not inside OBB
	{
		xPenetrationAxis = xDisp;

		fPenetration = fRadius - xPenetrationAxis.NormaliseReturningMagnitude();

		xPenetrationAxis = xPenetrationAxis * xOBB.GetOrientation();

		return true;
	}
	else // sphere centre inside or on surface of OBB 
	{
		// hmm, let's try finding the face nearest the sphere centre, and push it out that way.

		// distance to faces
		float fX = xOBB.GetHalfExtentX() - Asura_Maths::Abs(fSX);
		float fY = xOBB.GetHalfExtentY() - Asura_Maths::Abs(fSY);
		float fZ = xOBB.GetHalfExtentZ() - Asura_Maths::Abs(fSZ);

		if (fX<fY)
		{
			if (fX<fZ)
			{
				// X
				fPenetration = fX;
				xPenetrationAxis = Asura_AVector3( -Asura_Maths::Sign(fSX), 0.0f, 0.0f);
			}
			else 
			{
				// Z
				fPenetration = fZ;
				xPenetrationAxis = Asura_AVector3( 0.0f, 0.0f, -Asura_Maths::Sign(fSZ) );
			}
		}
		else 
		{
			if ( fY < fZ)
			{
				// Y
				fPenetration = fY;
				xPenetrationAxis = Asura_AVector3( 0.0f, -Asura_Maths::Sign(fSY), 0.0f );
			}
			else
			{
				// Z
				fPenetration = fZ;
				xPenetrationAxis = Asura_AVector3( 0.0f, 0.0f, -Asura_Maths::Sign(fSZ) );
			}
		}

		fPenetration += fRadius;
		xPenetrationAxis = xPenetrationAxis * xOBB.GetOrientation();


		return true;
	}
}



u_int Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( Asura_AVector3* pxContactPointsForPrimitive, const Asura_Physics_RigidBodyInstance* pxInstance, const Asura_Physics_Tri& xTri, const Asura_AVector3& xAxis )
{

	Asura_AVector3 pxContactPointsForB[10];

	// Find the support features for both objects. This can be a face (4 verts), and edge (2 verts) or just a single vertex
	Asura_AVector3 axSupportPointsForA[4];
	Asura_AVector3 axSupportPointsForB[4];

	u_int uNumberOfSupportPointsForA = pxInstance->GetSupportFeature( axSupportPointsForA, xAxis );
	u_int uNumberOfSupportPointsForB = xTri.GetSupportFeature( axSupportPointsForB, -xAxis );

	// We have 9 possible combinations. Might be able to minimise tests, but easiest if we do it as switches
	switch( uNumberOfSupportPointsForA )
	{
		default:
		{
			Asura_Assert( false, "Unhandled number of support points" );
			return 0;
		}
		case 1: // vertex
		{
			// all these contacts will just be one vertex, and for A it's the support point
			pxContactPointsForPrimitive[0] = axSupportPointsForA[0];
			return 1;
		}
		case 2: // edge
		{
			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-edge case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					ProjectVertexOntoEdge( axSupportPointsForB[0], axSupportPointsForA, pxContactPointsForPrimitive[0] );
					return 1;
				}
				case 2: // edge-edge case
				{
					ProjectEdgeOntoEdge( axSupportPointsForA, axSupportPointsForB, pxContactPointsForPrimitive[0], pxContactPointsForB[0] );
					return 1;
				}
				case 3: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForA, axSupportPointsForB, pxContactPointsForPrimitive, pxContactPointsForB, 3 );

				}
			}
		}
		case 4:
		{
			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-face case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					ProjectVertexOntoFace( axSupportPointsForB[0], axSupportPointsForA, pxContactPointsForPrimitive[0] );
					return 1;
				}
				case 2: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForB, axSupportPointsForA, pxContactPointsForB, pxContactPointsForPrimitive, 4 );

				}
				case 3: // face-face case
				{
					u_int uVerts =  ProjectFaceOntoFace( axSupportPointsForA, axSupportPointsForB, pxContactPointsForPrimitive, pxContactPointsForB, 4, 3 );

					Asura_AVector3 xFaceNormal = (axSupportPointsForA[1] - axSupportPointsForA[0])^(axSupportPointsForA[2] - axSupportPointsForA[0]);
					xFaceNormal.Normalise();

					for (u_int uV=0; uV<uVerts; ++uV)
					{
						#if defined ASURA_ASSERT
							float f = ( pxContactPointsForPrimitive[uV] - axSupportPointsForA[0] )* xFaceNormal;
						#endif

						Asura_Assert( Asura_Maths::Abs(f) < 0.01f, "Code bug: projection error" );
					}

					return uVerts;
				}
			}
		}
	}
}

u_int Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( Asura_AVector3* pxContactPointsForPrimitive, Asura_AVector3* pxContactPointsForTri, const Asura_Physics_RigidBodyInstance* pxInstance, const Asura_Physics_Tri& xTri, const Asura_AVector3& xAxis )
{

	// Find the support features for both objects. This can be a face (4 verts), and edge (2 verts) or just a single vertex
	Asura_AVector3 axSupportPointsForPrimitive[4];
	Asura_AVector3 axSupportPointsForTri[4];

	u_int uNumberOfSupportPointsForPrimitive = pxInstance->GetSupportFeature( axSupportPointsForPrimitive, xAxis );
	u_int uNumberOfSupportPointsForTri = xTri.GetSupportFeature( axSupportPointsForTri, -xAxis );

	// We have 9 possible combinations. Might be able to minimise tests, but easiest if we do it as switches r
	switch( uNumberOfSupportPointsForPrimitive )
	{
		default:
		{
			Asura_Assert( false, "Unhandled number of support points" );
			return 0;
		}
		case 1: // vertex
		{
			// all these contacts will just be one vertex, and for A it's the support point
			pxContactPointsForPrimitive[0] = axSupportPointsForPrimitive[0];

			switch( uNumberOfSupportPointsForTri )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-vertex case
				{
					pxContactPointsForTri[0] = axSupportPointsForTri[0];
					return 1;
				}
				case 2: // vertex-edge case
				{
					ProjectVertexOntoEdge( axSupportPointsForPrimitive[0], axSupportPointsForTri, pxContactPointsForTri[0] );
					return 1;
				}
				case 3: // vertex-face case
				{
					ProjectVertexOntoFace( axSupportPointsForPrimitive[0], axSupportPointsForTri, pxContactPointsForTri[0] );
					return 1;
				}
			}
		}
		case 2: // edge
		{
			switch( uNumberOfSupportPointsForTri )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-edge case
				{
					pxContactPointsForTri[0] = axSupportPointsForTri[0];
					ProjectVertexOntoEdge( axSupportPointsForTri[0], axSupportPointsForPrimitive, pxContactPointsForPrimitive[0] );
					return 1;
				}
				case 2: // edge-edge case
				{
					ProjectEdgeOntoEdge( axSupportPointsForPrimitive, axSupportPointsForTri, pxContactPointsForPrimitive[0], pxContactPointsForTri[0] );
					return 1;
				}
				case 3: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForPrimitive, axSupportPointsForTri, pxContactPointsForPrimitive, pxContactPointsForTri, 3 );

				}
			}
		}
		case 4:
		{
			switch( uNumberOfSupportPointsForTri )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-face case
				{
					pxContactPointsForTri[0] = axSupportPointsForTri[0];
					ProjectVertexOntoFace( axSupportPointsForTri[0], axSupportPointsForPrimitive, pxContactPointsForPrimitive[0] );
					return 1;
				}
				case 2: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForTri, axSupportPointsForPrimitive, pxContactPointsForTri, pxContactPointsForPrimitive, 4 );

				}
				case 3: // face-face case
				{
					return ProjectFaceOntoFace( axSupportPointsForPrimitive, axSupportPointsForTri, pxContactPointsForPrimitive, pxContactPointsForTri, 4, 3 );
				}
			}
		}
	}


}

u_int Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( Asura_AVector3* pxContactPointsForPrimitive, const Asura_Physics_MPR_Body* pxPrimitive, const Asura_Physics_Tri& xTri, const Asura_AVector3& xAxis )
{

	Asura_AVector3 pxContactPointsForB[10];

	// Find the support features for both objects. This can be a face (4 verts), and edge (2 verts) or just a single vertex
	Asura_AVector3 axSupportPointsForA[4];
	Asura_AVector3 axSupportPointsForB[4];

	// TODO - currently MPR just returns a support point rather than a feature, though for spheres and capsules it would be a point anyway
	u_int uNumberOfSupportPointsForA = 1;
	axSupportPointsForA[0] = pxPrimitive->GetSupportPoint( xAxis );
	u_int uNumberOfSupportPointsForB = xTri.GetSupportFeature( axSupportPointsForB, -xAxis );
	

	// We have 9 possible combinations. Might be able to minimise tests, but easiest if we do it as switches
	switch( uNumberOfSupportPointsForA )
	{
		default:
		{
			Asura_Assert( false, "Unhandled number of support points" );
			return 0;
		}
		case 1: // vertex
		{
			// all these contacts will just be one vertex, and for A it's the support point
			pxContactPointsForPrimitive[0] = axSupportPointsForA[0];
			return 1;
		}
		case 2: // edge
		{
			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-edge case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					ProjectVertexOntoEdge( axSupportPointsForB[0], axSupportPointsForA, pxContactPointsForPrimitive[0] );
					return 1;
				}
				case 2: // edge-edge case
				{
					ProjectEdgeOntoEdge( axSupportPointsForA, axSupportPointsForB, pxContactPointsForPrimitive[0], pxContactPointsForB[0] );
					return 1;
				}
				case 3: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForA, axSupportPointsForB, pxContactPointsForPrimitive, pxContactPointsForB, 3 );

				}
			}
		}
		case 4:
		{
			switch( uNumberOfSupportPointsForB )
			{
				default:
				{
					Asura_Assert( false, "Unhandled number of support points" );
					return 0;
				}
				case 1: // vertex-face case
				{
					pxContactPointsForB[0] = axSupportPointsForB[0];
					ProjectVertexOntoFace( axSupportPointsForB[0], axSupportPointsForA, pxContactPointsForPrimitive[0] );
					return 1;
				}
				case 2: // edge-face case
				{
					return ProjectEdgeOntoFace( axSupportPointsForB, axSupportPointsForA, pxContactPointsForB, pxContactPointsForPrimitive, 4 );

				}
				case 3: // face-face case
				{
					u_int uVerts =  ProjectFaceOntoFace( axSupportPointsForA, axSupportPointsForB, pxContactPointsForPrimitive, pxContactPointsForB, 4, 3 );

					Asura_AVector3 xFaceNormal = (axSupportPointsForA[1] - axSupportPointsForA[0])^(axSupportPointsForA[2] - axSupportPointsForA[0]);
					xFaceNormal.Normalise();

					for (u_int uV=0; uV<uVerts; ++uV)
					{
						#if defined ASURA_ASSERT
							float f = ( pxContactPointsForPrimitive[uV] - axSupportPointsForA[0] )* xFaceNormal;
						#endif

						Asura_Assert( Asura_Maths::Abs(f) < 0.01f, "Code bug: projection error" );
					}

					return uVerts;
				}
			}
		}
	}
}

Asura_Physics_Tri::Asura_Physics_Tri( const u_int uFaceID )
{
	m_xV[0] = Asura_Physics_Environment::GetFaceVertex( uFaceID, 0 );
	m_xV[1] = Asura_Physics_Environment::GetFaceVertex( uFaceID, 1 );
	m_xV[2] = Asura_Physics_Environment::GetFaceVertex( uFaceID, 2 );

}

u_int Asura_Physics_Tri::GetSupportFeature( Asura_AVector3* pxSupportPoints, const Asura_AVector3& xAxis ) const
{
	// we can either have a supporting face, edge or vertex
	
	// first check for the face case:
	Asura_AVector3 xNormal;
	GetNormal( xNormal );

	float fDot = Asura_Maths::Abs( xNormal * xAxis );
	const float fFaceDotThreshold = (0.9999f);

	if ( fDot > fFaceDotThreshold )
	{
		// hmm - may have to flip order based on sign of dot product
		pxSupportPoints[0] = m_xV[0];
		pxSupportPoints[1] = m_xV[1];
		pxSupportPoints[2] = m_xV[2];

		return 3;
	}

	// ok, what about any supporting edges - this happens when xAxis is perpendicular to an edge, put not parallel to one
	// (because that would be the face case).
	const float fEdgeDotThreshold = 2.6e-3f;
	
	u_int uVertB = 2;

	for ( u_int uVertA = 0 ; uVertA < 3; uVertB = uVertA++ )
	{
		Asura_AVector3 xEdge = m_xV[uVertA] - m_xV[uVertB];

		if ( Asura_Maths::Abs( xEdge * xAxis) < fEdgeDotThreshold )
		{
			pxSupportPoints[0] = m_xV[uVertB];
			pxSupportPoints[1] = m_xV[uVertA];

			return 2;
		}
	}

	// ok, if we've reach here its a supporting vert 
	u_int uVertex = 0;
	float fMinVertexDot = m_xV[0] * xAxis;
	for (u_int u=1; u<3; u++)	
	{
		float fVertexDot = m_xV[u] * xAxis;
		if ( fVertexDot < fMinVertexDot )
		{
			fMinVertexDot = fVertexDot;
			uVertex = u;
		}
	}
	pxSupportPoints[0] = m_xV[uVertex];
	return 1;
}

Asura_AVector3 Asura_Physics_Tri::GetSupportPoint( const Asura_AVector3& xAxis ) const
{
	u_int uVertex = 0;
	float fMinDot = m_xV[0] * xAxis;

	for (u_int u=1; u<3; u++)	
	{
		float fDot = m_xV[u] * xAxis;
		if ( fDot < fMinDot )
		{
			fMinDot = fDot;
			uVertex = u;
		}
	}
	return m_xV[uVertex];
}


Asura_Physics_SphereGroup::Asura_Physics_SphereGroup( const Asura_Physics_RigidBodyInstance& xInstance )
{
	const u_int	uMIN_NUMBER_OF_SPHERES_IN_STACK = 2;

	// grab volume properties for the instance
	Asura_AVector3 xCentre = xInstance.GetCollisionMeshCentre();

	if ( xInstance.GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERESTACK )
	{
		Asura_AVector3 xAxis;

		xInstance.GetDefinition()->GetSphereStackInfo( xAxis, m_fRadius );
		float fHalfHeight = xAxis.NormaliseReturningMagnitude();

		xAxis = xAxis * xInstance.GetOrientation(); // hmm - orientation not up-to-date enough?
	
		// work out how many spheres we need to cover this volume	
		m_uNumberOfSpheres = 1 + Asura_Maths::FloatToInt_Chop( fHalfHeight / m_fRadius );

		// clamp it to a sensible range
		m_uNumberOfSpheres = Asura_Maths::ClampToRange( m_uNumberOfSpheres, uMIN_NUMBER_OF_SPHERES_IN_STACK, uASURA_MAX_NUMBER_OF_SPHERES_IN_SPHEREGROUP );

		// find the position of the bottom sphere (which will be a radius away from the volume bottom)
		Asura_AVector3 xBottomOffset = xAxis * (fHalfHeight-m_fRadius);
		Asura_AVector3 xSphereCentre = xCentre + xBottomOffset;

		// find the offset between spheres
		float fSphereOffset = (fHalfHeight - m_fRadius) * 2.0f / static_cast<float>(m_uNumberOfSpheres - 1);
		Asura_AVector3 xSphereOffset = xAxis * fSphereOffset;

		// set the positions for each sphere
		for ( u_int u=0; u<m_uNumberOfSpheres; ++u )
		{
			m_axSphereCentres[u] = xSphereCentre;
			xSphereCentre -= xSphereOffset;
		}
	}
	else if ( xInstance.GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE )
	{
		m_fRadius = xInstance.GetDefinition()->GetMaximumRadius();
		m_uNumberOfSpheres = 1;
		m_axSphereCentres[0] = xCentre;
	}
	else
	{
		m_fRadius = xInstance.GetDefinition()->GetMinimumRadius()*0.5f;
		m_uNumberOfSpheres = 1;
		m_axSphereCentres[0] = xCentre;
	}
}

Asura_Physics_SphereGroup::Asura_Physics_SphereGroup( const Asura_OBB& xOBB, bool bSphereStack )
{
	const u_int	uMIN_NUMBER_OF_SPHERES_IN_STACK = 2;

	// grab volume properties for the instance
	Asura_AVector3 xCentre = xOBB.GetCentre();

	if ( bSphereStack )
	{
		Asura_AVector3 xAxis;

		xOBB.GetPrimaryAxisAndItsRadius( xAxis, m_fRadius );
		float fHalfHeight = xAxis.NormaliseReturningMagnitude();

		// work out how many spheres we need to cover this volume	
		m_uNumberOfSpheres = 1 + Asura_Maths::FloatToInt_Chop( fHalfHeight / m_fRadius );

		// clamp it to a sensible range
		m_uNumberOfSpheres = Asura_Maths::ClampToRange( m_uNumberOfSpheres, uMIN_NUMBER_OF_SPHERES_IN_STACK, uASURA_MAX_NUMBER_OF_SPHERES_IN_SPHEREGROUP );

		// find the position of the bottom sphere (which will be a radius away from the volume bottom)
		Asura_AVector3 xBottomOffset = xAxis * (fHalfHeight-m_fRadius);
		Asura_AVector3 xSphereCentre = xCentre + xBottomOffset;

		// find the offset between spheres
		float fSphereOffset = (fHalfHeight - m_fRadius) * 2.0f / static_cast<float>(m_uNumberOfSpheres - 1);
		Asura_AVector3 xSphereOffset = xAxis * fSphereOffset;

		// set the positions for each sphere
		for ( u_int u=0; u<m_uNumberOfSpheres; ++u )
		{
			m_axSphereCentres[u] = xSphereCentre;
			xSphereCentre -= xSphereOffset;

			//Asura_DebugLineRender::AddSphere( Asura_Vector_3(m_axSphereCentres[u]), m_fRadius, 0xff00ffff );

		}
	}
	else // just a sphere
	{
		m_fRadius = xOBB.GetMaximumHalfExtent();
		m_uNumberOfSpheres = 1;
		m_axSphereCentres[0] = xCentre;
	}
}

u_int Asura_Physics_SphereGroup::GetNumberOfSpheres( void ) const 
{
	return m_uNumberOfSpheres;
}

const Asura_AVector3* Asura_Physics_SphereGroup::GetSphereCentres( void ) const 
{
	return m_axSphereCentres;
}

float Asura_Physics_SphereGroup::GetSphereRadius( void ) const 
{
	return m_fRadius;
}

void Asura_Physics_SphereGroup::Displace( const Asura_AVector3& xDisplacement )
{
	for ( u_int u=0; u<m_uNumberOfSpheres; ++u )
	{
		m_axSphereCentres[u] += xDisplacement;
	}
}

float Asura_Physics_SphereGroup::SweptSpheresVsEnvironment( const Asura_AVector3& xDirection, const float fDistance, const Asura_Physics_EnvironmentCache& xEnvCache, Asura_AVector3& xLastCollisionNormal, u_int& uLastCollisionFaceID, bool& bHitEdge )
{
	bHitEdge = false;

	float fMaxAllowed = fDistance;
	for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xEnvCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
	{
		const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		Asura_AVector3 xPV0 = xV0;
		Asura_AVector3 xPV1 = xV1;
		Asura_AVector3 xPV2 = xV2;

		Asura_AVector3 xNormal = ((xPV1 - xPV0)^(xPV2 - xPV0));
		xNormal.Normalise();

		float fDot = ( -xNormal * xDirection );

		xPV0 = xPV0 + xNormal * GetSphereRadius();
		xPV1 = xPV1 + xNormal * GetSphereRadius();
		xPV2 = xPV2 + xNormal * GetSphereRadius();

		if (fDot <= 0.0f) continue;

		for ( u_int v=0; v<GetNumberOfSpheres(); ++v )
		{
			bool bFacingTowards;
			float fHitDistance;

			if( Asura_Physics_System::RayIntersectsTri( GetSphereCentres()[v], xDirection, xPV0, xPV1, xPV2, bFacingTowards, fHitDistance ) )
			{
				if ( fHitDistance >= 0.0f && bFacingTowards )
				{
					fHitDistance = Asura_Maths::Max( 0.0f, fHitDistance );

					if ( fHitDistance <= fMaxAllowed )
					{
						fMaxAllowed = fHitDistance;

						uLastCollisionFaceID = pxFace->m_uFaceID;

						xLastCollisionNormal = xNormal;
					}
				}	
				else if ( fHitDistance > -GetSphereRadius())
				{
					// we're already overlapping with a polygon I'm afraid
					// return a zero allowed distance
					fMaxAllowed = 0.0f;
					uLastCollisionFaceID = pxFace->m_uFaceID;
					xLastCollisionNormal = xNormal;
					
					return fMaxAllowed;
				}
			}
		}
	}

	for ( const Asura_Physics_EnvironmentCache_Edge* pxEdge = xEnvCache.GetEdgeList(); pxEdge; pxEdge = pxEdge->m_pxNext )
	{
		const Asura_AVector3& xEdgeV0 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
		const Asura_AVector3& xEdgeV1 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );

		//Asura_Physics_System::SmallestVectorBetweenTwoLines( xEdgeV0, xEdgeV1,)
		for ( u_int v=0; v<GetNumberOfSpheres(); ++v )
		{
			float fAllowedSphereMovement;
			Asura_AVector3 xCollisionPoint;

			Asura_AVector3 xSphereDirection = xDirection * fMaxAllowed;

			if ( Asura_Physics_System::SweptSphereVsEdge( GetSphereRadius(), GetSphereCentres()[v], xSphereDirection, xEdgeV0, xEdgeV1, xCollisionPoint, fAllowedSphereMovement ) )
			{
				float fSuggestedMaxAllowed = fAllowedSphereMovement * fMaxAllowed;

				if (fSuggestedMaxAllowed < fMaxAllowed )
				{
					fMaxAllowed = fSuggestedMaxAllowed;
					xLastCollisionNormal = ( GetSphereCentres()[v] + xSphereDirection * fMaxAllowed - xCollisionPoint );
					xLastCollisionNormal.Normalise();
					uLastCollisionFaceID = 0xffffffff;
					bHitEdge = true;
				}
			}
		}

	}

	if (fMaxAllowed != fDistance)
	{
		fMaxAllowed = Asura_Maths::Max( 0.0f, fMaxAllowed - fAXIALLYALIGNED_DIST_THRESHOLD);
	}

	return fMaxAllowed;
}
float Asura_Physics_SphereGroup::SweptSpheresVsEnvironment( const Asura_AVector3& xDirection, const float fDistance, const Asura_Physics_EnvironmentCache& xEnvCache,
								Asura_Physics_CollisionData* pxCollisionData, u_int& uNumberOfCollisions, u_int uMaxNumberOfCollisionsToRecord   )
{
	
	float fMaxAllowed = fDistance;
	for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xEnvCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
	{
		const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		Asura_AVector3 xPV0 = xV0;
		Asura_AVector3 xPV1 = xV1;
		Asura_AVector3 xPV2 = xV2;

		Asura_AVector3 xNormal = ((xPV1 - xPV0)^(xPV2 - xPV0));
		xNormal.Normalise();

		float fDot = ( -xNormal * xDirection );

		xPV0 = xPV0 + xNormal * GetSphereRadius();
		xPV1 = xPV1 + xNormal * GetSphereRadius();
		xPV2 = xPV2 + xNormal * GetSphereRadius();

		if (fDot <= 0.0f) continue;

		for ( u_int v=0; v<GetNumberOfSpheres(); ++v )
		{
			bool bFacingTowards;
			float fHitDistance;

			if( Asura_Physics_System::RayIntersectsTri( GetSphereCentres()[v], xDirection, xPV0, xPV1, xPV2, bFacingTowards, fHitDistance ) )
			{
				if ( fHitDistance >= 0.0f && bFacingTowards )
				{
					fHitDistance = Asura_Maths::Max( 0.0f, fHitDistance );

					if ( fHitDistance <= fMaxAllowed )
					{
						fMaxAllowed = fHitDistance;

						pxCollisionData[0].m_bIsEdge = false;
						pxCollisionData[0].m_uPolyFlags = Asura_Physics_Environment::GetFaceFlags( pxFace->m_uFaceID );
						pxCollisionData[0].m_uPolyID = pxFace->m_uFaceID;
						pxCollisionData[0].m_xNormal = xNormal;
						uNumberOfCollisions = 1;
					}
				}	
				else if ( fHitDistance > -GetSphereRadius())
				{
					// we're already overlapping with a polygon I'm afraid
					// return a zero allowed distance
					fMaxAllowed = 0.0f;

					pxCollisionData[0].m_bIsEdge = false;
					pxCollisionData[0].m_uPolyFlags = Asura_Physics_Environment::GetFaceFlags( pxFace->m_uFaceID );
					pxCollisionData[0].m_uPolyID = pxFace->m_uFaceID;
					pxCollisionData[0].m_xNormal = xNormal;
					uNumberOfCollisions = 1;

					return fMaxAllowed;
				}
			}
		}
	}

	for ( const Asura_Physics_EnvironmentCache_Edge* pxEdge = xEnvCache.GetEdgeList(); pxEdge; pxEdge = pxEdge->m_pxNext )
	{
		const Asura_AVector3& xEdgeV0 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
		const Asura_AVector3& xEdgeV1 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );

		//Asura_Physics_System::SmallestVectorBetweenTwoLines( xEdgeV0, xEdgeV1,)
		for ( u_int v=0; v<GetNumberOfSpheres(); ++v )
		{
			float fAllowedSphereMovement;
			Asura_AVector3 xCollisionPoint;

			Asura_AVector3 xSphereDirection = xDirection * fMaxAllowed;

			if ( Asura_Physics_System::SweptSphereVsEdge( GetSphereRadius(), GetSphereCentres()[v], xSphereDirection, xEdgeV0, xEdgeV1, xCollisionPoint, fAllowedSphereMovement ) )
			{
				float fSuggestedMaxAllowed = fAllowedSphereMovement * fMaxAllowed;

				if (fSuggestedMaxAllowed < fMaxAllowed )
				{
					fMaxAllowed = fSuggestedMaxAllowed;
					Asura_AVector3 xNormal = ( GetSphereCentres()[v] + xSphereDirection * fMaxAllowed - xCollisionPoint );
					xNormal.Normalise();

					uNumberOfCollisions = 0;

					for ( const Asura_Physics_EnvironmentCache_Face* pxFace = xEnvCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
					{
						u_int uMatchingVertex = 0;
						for ( u_int uVertex=0; uVertex<3; ++uVertex )
						{
							const Asura_AVector3& xFaceVertex = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, uVertex );
							
							if ( xEdgeV0.ApproximatelyEquals( xFaceVertex ) )
							{
								++uMatchingVertex;
							}
							else if ( xEdgeV1.ApproximatelyEquals( xFaceVertex ) )
							{
								++uMatchingVertex;
							}
						}

						if (uMatchingVertex==2)
						{
							if ( uNumberOfCollisions < uMaxNumberOfCollisionsToRecord )
							{
								pxCollisionData[uNumberOfCollisions].m_bIsEdge = true;
								pxCollisionData[uNumberOfCollisions].m_uPolyFlags = Asura_Physics_Environment::GetFaceFlags( pxFace->m_uFaceID );
								pxCollisionData[uNumberOfCollisions].m_uPolyID = pxFace->m_uFaceID;
								pxCollisionData[uNumberOfCollisions].m_xNormal = xNormal;	
							
								++uNumberOfCollisions;
							}
						}
					}

					if ( !uNumberOfCollisions )
					{
						Asura_Assert( false, "Code bug: couldn't match edge to any tris" );

						// If we didn't find any matching tris, we'll have to make do with just an edge with no flags or PolyID
						pxCollisionData[0].m_bIsEdge = true;
						pxCollisionData[0].m_uPolyFlags = 0;
						pxCollisionData[0].m_uPolyID = 0xffffffff;
						pxCollisionData[0].m_xNormal = xNormal;
						uNumberOfCollisions = 1;
					}

				}
			}
		}

	}

	if (fMaxAllowed != fDistance)
	{
		fMaxAllowed = Asura_Maths::Max( 0.0f, fMaxAllowed - fAXIALLYALIGNED_DIST_THRESHOLD);
	}

	return fMaxAllowed;
}


void Asura_Physics_CCD::FindClosestPointOnTriangle( const Asura_AVector3& xP, const Asura_AVector3& xA, const Asura_AVector3& xB, const Asura_AVector3& xC, Asura_AVector3& xClosestPoint )
{
	// Find the nearest point on the triangle, using Voronoi regions
	// adapted from pg. 141 of "Real Time Collision Detection"

	// Check if P in vertex region outside A
	const Asura_AVector3 xAB = xB - xA;
	const Asura_AVector3 xAC = xC - xA;
	const Asura_AVector3 xAP = xP - xA;

	const float fD1 = xAB * xAP;
	const float fD2 = xAC * xAP;

	if ( fD1 <= 0.0f && fD2 <= 0.0f ) 
	{
		// Closest point is indeed A - barycentric coordinates (1,0,0)
		xClosestPoint = xA;
		return;
	}

	// Check if P in vertex region outside B
	const Asura_AVector3 xBP = xP - xB;
	const float fD3 = xAB * xBP;
	const float fD4 = xAC * xBP;
	
	if (fD3 >= 0.0f && fD4 <= fD3 ) 
	{
		// Closest point is B - barycentric coordinates (0,1,0)
		xClosestPoint = xB;
		return;
	}

	// Check if P in edge region of AB, if so return projection of P onto AB
	const float fVC = fD1*fD4 - fD3*fD2;

	if ( fVC <= 0.0f && fD1 >= 0.0f && fD3 <= 0.0f )
	{
		// Closest point is barycentric coordinates (1-fV,fV,0)
		float fV = fD1 / (fD1 - fD3);
		xClosestPoint = xA + xAB * fV;
		return;
	}

	// Check if P in vertex region outside C
	const Asura_AVector3 xCP = xP - xC;
	const float fD5 = xAB * xCP;
	const float fD6 = xAC * xCP;

	if ( fD6 >= 0.0f && fD5 <= fD6 ) 
	{
		// Closest point is C - barycentric coordinates (0,0,1)
		xClosestPoint = xC;
		return;
	}

	// Check if P in edge region of AC, if so return projection of P onto AC
	const float fVB = fD5*fD2 - fD1*fD6;

	if ( fVB <= 0.0f && fD2 >= 0.0f && fD6 <= 0.0f )
	{
		// Closest point is barycentric coordinates (1-fW,0,fW)
		float fW = fD2 / (fD2 - fD6);
		xClosestPoint = xA + xAC * fW;
		return;
	}

	// Check if P in edge region of BC, if so return projection of P onto BC
	const float fVA = fD3*fD6 - fD5*fD4;

	if ( fVA <= 0.0f && (fD4 - fD3) >= 0.0f && (fD5 - fD6) >= 0.0f )
	{
		// Closest point is barycentric coordinates (0,1-fW,fW)
		float fW = (fD4 - fD3) / ((fD4 - fD3) + (fD5 - fD6));
		xClosestPoint = xB + (xC-xB) * fW;
		return;		
	}

	// P inside face region. Compute closest point through its barycentric coordinates (u,v,w)
	const float fDenom = 1.0f / ( fVA + fVB + fVC );

	//          fU = fVA * fDenom = 1.0f - fV - fW
	const float fV = fVB * fDenom;
	const float fW = fVC * fDenom;
	
    xClosestPoint = (xA) + (xAB * fV) + (xAC * fW); // equivalent to (xA * fU) + (xB * fV) + (xC * fW)
	return;
}


Asura_Physics_MPR_Body::Asura_Physics_MPR_Body( const Asura_Physics_RigidBodyInstance* pxBody )
{
	m_pxBody = pxBody;
	m_pxTri = 0;
	m_pxRagdollObject = 0;

	if ( m_pxBody )
	{
		m_xEnd = m_pxBody->GetCentreOfMass();
		m_xStart = m_pxBody->GetPreviousCentreOfMass();
	}
	else
	{
		m_xEnd.Zero();
		m_xStart.Zero();
	}
}

Asura_Physics_MPR_Body::Asura_Physics_MPR_Body( const Asura_Physics_RigidBodyInstance* pxBody, const Asura_AVector3& xStartPosition, const Asura_AVector3& xEndPosition )
{
	m_pxBody = pxBody;
	m_pxTri = 0;
	m_pxRagdollObject = 0;

	m_xStart = xStartPosition;
	m_xEnd = xEndPosition;
}

Asura_Physics_MPR_Body::Asura_Physics_MPR_Body( const Asura_Physics_Tri* pxTri )
{
	m_pxBody = 0;
	m_pxTri = pxTri;
	m_pxRagdollObject = 0;
	
	if ( pxTri )
	{
		m_xEnd = m_pxTri->GetCentre();
		m_xStart = m_pxTri->GetCentre();
	}
	else
	{
		m_xEnd.Zero();
		m_xStart.Zero();
	}
}

Asura_Physics_MPR_Body::Asura_Physics_MPR_Body( const Asura_Physics_Ragdoll_CollisionObject* pxObject )
{
	m_pxBody = 0;
	m_pxTri = 0;
	m_pxRagdollObject = pxObject;
	
	if ( pxObject )
	{
		m_xEnd = pxObject->GetPosition();
		m_xStart = pxObject->GetPreviousPosition();
	}
	else
	{
		m_xEnd.Zero();
		m_xStart.Zero();
	}
}

// KJFL - 17 July 2008 - Minkowski Portal Refinement (MPR) test
// Following the example given in Game Programming Gems 7
//
// Running in parallel with the existing OBB vs OBB collision code
// seems to work well, though I needed to tweak some floating point
// epsilons and normalise a few vectors.

// Can probably get rid of this once inner loop optimised - for
// now it aids code clarity below.
inline void Swap( Asura_AVector3& xV1, Asura_AVector3& xV2 )
{
	Asura_AVector3 xTemp = xV1;
	xV1 = xV2;
	xV2 = xTemp;
}
Asura_AVector3 Asura_Physics_MPR_Body::GetSupportPoint( const Asura_AVector3& xAxis, bool bPreviousPosition ) const
{
	if ( m_pxBody )
	{
		// KJFL - 17 July 2008 - while we're testing the MPR routine, I'm just using OBBs. 
		// The support point is just the OBB vertex most in the direction of the axis.
		//
		// Note this is is in world space, hence the orientation manipulation - long
		// hand you'd rotate the axis into object space, find the vertex, and then 
		// transform the vertex into world space. I've just simplified the maths
		// down to some dot products, some of which could be precalculated for speed.
		//
		// Later we'll want dedicated routines for different shapes etc.
		//
		const Asura_AVector3 xCentre = bPreviousPosition ? GetPreviousCentre() : GetCentre();

		if ( m_pxBody->GetCollisionProxy()==Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE )
		{
			Asura_AVector3 xPoint = xCentre	- xAxis * m_pxBody->GetDefinition()->GetMaximumRadius();
			return xPoint;
		}
		else if (m_pxBody->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERESTACK )
		{
			Asura_AVector3 xCapsuleAxis;
			float fRadius;
			m_pxBody->GetDefinition()->GetSphereStackInfo( xCapsuleAxis, fRadius );
		
			float fHalfHeight = xCapsuleAxis.NormaliseReturningMagnitude();

			xCapsuleAxis = xCapsuleAxis * m_pxBody->GetOrientation(); 

			Asura_AVector3 xOffset = xCapsuleAxis * (fHalfHeight-fRadius);

			Asura_AVector3 xPoint = xCentre - xAxis * fRadius;
	
			if ( xOffset * xAxis < 0.0f )
			{
				xPoint += xOffset;
			}
			else
			{
				xPoint -= xOffset;
			}

			return xPoint;
		}
		else if (m_pxBody->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_POLYGONAL )
		{
			// KJFL - 28 August 2008 - get support mapping for a convex polygonal object
			Asura_AVector3 xPoint;

			const Asura_CollisionMesh* pxMesh = m_pxBody->GetDefinition()->GetCollisionMesh();
			if (pxMesh)
			{
				const u_int uNumberOfVertices = pxMesh->GetNumberOfVertices();

				float fMinDot = Asura_Maths::LargeFloat;
				xPoint.Zero();

				// To reduce the per-vertex cost as much as possible, we'll do it in model space.
				const Asura_AVector3 xModelSpaceAxis = m_pxBody->GetOrientation().ApplyTranspose( xAxis );

				for ( u_int u=0; u<uNumberOfVertices; ++u )
				{
					Asura_AVector3 xPos( pxMesh->GetVertex(u) );

					float fDot = xPos * xModelSpaceAxis;

					if ( fDot < fMinDot )
					{
						xPoint = xPos;
						fMinDot = fDot;
					}
				}

				// KJFL - 28 August 2008 - NOTE: must remember that these points are offset by the
				// render mesh position, not the center of mass.
				xPoint = ( xPoint * m_pxBody->GetOrientation()) + m_pxBody->GetRenderingPosition() ;
			}
			else
			{
				// AJG - 17/09/08 - I don't really know what a sensible default if we don't have a collision mesh should be, but from reading the above the rendering position leaps out as the most obvious.
				xPoint = m_pxBody->GetRenderingPosition();
			}

			if (bPreviousPosition)
			{
				xPoint += (m_pxBody->GetPreviousCentreOfMass() - m_pxBody->GetCentreOfMass());
			}

			return xPoint;
		}
		else
		{
			const float fX = -xAxis * m_pxBody->GetOrientation().GetColumn0(); 
			const float fY = -xAxis * m_pxBody->GetOrientation().GetColumn1(); 
			const float fZ = -xAxis * m_pxBody->GetOrientation().GetColumn2(); 

			// lots of stuff here we could precalc / lazy eval
			const Asura_AVector3 xXOffset = m_pxBody->GetOrientation().GetColumn0() * m_pxBody->GetDefinition()->GetOBBHalfExtentsX();
			const Asura_AVector3 xYOffset = m_pxBody->GetOrientation().GetColumn1() * m_pxBody->GetDefinition()->GetOBBHalfExtentsY();
			const Asura_AVector3 xZOffset = m_pxBody->GetOrientation().GetColumn2() * m_pxBody->GetDefinition()->GetOBBHalfExtentsZ();

			// And the vertex in world space is...
			// KJFL - 17 July 2008 - using the sign op like this in theory means xAxis doesn't need to be normalised.
			// It might cause float epsilon errors elsewhere though if I were to leave it unnormalised.
			Asura_AVector3 xPoint = xCentre +  xXOffset * Asura_Maths::Sign(fX) + xYOffset * Asura_Maths::Sign(fY) + xZOffset * Asura_Maths::Sign(fZ);

			return xPoint;
		}
	}
	else if ( m_pxTri )
	{
		return m_pxTri->GetSupportPoint( xAxis );
	}
	else if ( m_pxRagdollObject )
	{
		return m_pxRagdollObject->GetSupportPoint( xAxis );
	}
		
	return Asura_Maths::ZeroAVector3;
	
}

Asura_AVector3 Asura_Physics_MPR_Body::GetSweptSupportPoint( const Asura_AVector3& xAxis, const Asura_AVector3& xSweep ) const
{
	Asura_AVector3 xSupport = GetSupportPoint( xAxis );

	if ( xSweep * xAxis < 0.0f )
	{
		xSupport += xSweep;
	}
	
	return xSupport;
}

Asura_AVector3 Asura_Physics_MPR_Body::GetCentre( void ) const
{
	return m_xEnd;
}

Asura_AVector3 Asura_Physics_MPR_Body::GetPreviousCentre( void ) const
{
	return m_xStart;
}
bool Asura_Physics_CCD::MPR_Test( const Asura_Physics_MPR_Body& xBody1, const Asura_Physics_MPR_Body& xBody2, Asura_AVector3& xAxis, float& fSeparation, bool bWarmStart )
{
	// This is the distance at which we give up searching.
	// This means we may miss an intersection smaller than this (in m).
	const float fTerminationDistance = 1e-6f;
	
	// define xV0 = center of Minkowski sum
	Asura_AVector3 xV0;

	if (bWarmStart)
	{
		xV0 = -xAxis;
	}
	else
	{
		xV0 = xBody2.GetCentre() - xBody1.GetCentre();

		// Avoid case where centers overlap -- any direction is fine in this case
		if (xV0 == Asura_Maths::ZeroAVector3 ) xV0 = Asura_AVector3(0.01f, 0.0f, 0.0f);
	}

	Asura_AVector3 xNormal = -xV0;
	xNormal.Normalise();

	// find xV1 = support in direction of origin
	Asura_AVector3 xV1 = xBody2.GetSupportPoint( -xNormal ) - xBody1.GetSupportPoint( xNormal );
		
	
	if ( xV1 * xNormal <= 0.0f)
	{
		xAxis = xNormal;
		return false;
	}

	xNormal = xV1 ^ xV0;
	
	// KJFL - 31 July 2008 - special case when xV0 & xV1 are parallel.
	// If xNormal is zero length, this means xV0, xV1 & the origin are colinear.
	// If xV0 is on one side of the origin by definition, and we've already
	// checked xV1 is on the other side of the origin, that must mean the origin
	// is on a line between the two. This means there is an intersection.
	if ( xNormal.NormaliseReturningMagnitude() == 0.0f )
	{
		xNormal = xV1 - xV0;
		xNormal.Normalise();
		xAxis = xNormal;

		fSeparation = -(xV1*xNormal);

		return true;
	}
	
	// find xV2 - support perpendicular to xV1,xV0
	Asura_AVector3 xV2 = xBody2.GetSupportPoint( -xNormal ) - xBody1.GetSupportPoint( xNormal );
		
	{
		float fDot = xV2 * xNormal;
		if (fDot <= 0.0f)
		{
			xAxis = xNormal;
			return false;
		}
	}

	// Determine whether origin is on + or - side of plane (xV1,xV0,xV2)
	xNormal = (xV1 - xV0) ^ (xV2 - xV0);
	xNormal.Normalise();

	// TODO - can xNormal be zero here?

	// If the origin is on the -ve side of the plane, reverse the direction of the plane
	if ( xNormal * xV0 > 0.0f )
	{
		Swap(xV1, xV2);
		xNormal = -xNormal;
	}

	// Phase One: Identify a portal

	u_int uPhaseOneIterations = 0;
	while (1)
	{
		++uPhaseOneIterations;
		// Obtain the support point in a direction perpendicular to the existing plane
		// Note: This point is guaranteed to lie off the plane
	
		Asura_AVector3 xV3 = xBody2.GetSupportPoint( -xNormal ) - xBody1.GetSupportPoint( xNormal );
		
		if (xV3*xNormal <= 0.0f)
		{
			xAxis = xNormal;
			return false;
		}

		// If origin is outside (xV1,xV0,xV3), then eliminate xV2 and loop
		if ((xV1 ^ xV3) * xV0 < 0.0f)
		{
			xV2 = xV3;
			xNormal = (xV1 - xV0) ^ (xV3 - xV0);
			xNormal.Normalise();
			continue;
		}

		// If origin is outside (xV3,xV0,xV2), then eliminate xV1 and loop
		if ((xV3 ^ xV2) * xV0 < 0.0f)
		{
			xV1 = xV3;
			xNormal = (xV3 - xV0) ^ (xV2 - xV0);
			xNormal.Normalise();
			continue;
		}

		bool bOriginInside = false;

		// Phase Two: Refine the portal
		u_int uPhaseTwoIterations=0;

		// We are now inside of a wedge...
		while (1)
		{
			++uPhaseTwoIterations;

			// Compute normal of the wedge face
			xNormal = (xV2 - xV1) ^ (xV3 - xV1);

			// Can this happen???  Can it be handled more cleanly?
			if ( xNormal.NormaliseReturningMagnitude() == 0.0f )
			{
				Asura_Assert( false, "Potential code bug: notify KJFL" );
				return false;
			}

			// Compute distance from origin to wedge face
			if (!bOriginInside)
			{
				float fDist = xNormal * xV1;

				// If the origin is inside the wedge, we have an intersection
				if (fDist >= 0.0f)
				{
					bOriginInside = true;
				}
			}

			// Find the support point in the direction of the wedge face
			Asura_AVector3 xV4 = xBody2.GetSupportPoint( -xNormal ) - xBody1.GetSupportPoint( xNormal );
		
			float fDelta = (xV4 - xV3) * xNormal;
			fSeparation = -(xV4 * xNormal);

			

			// If the boundary is thin enough or the origin is outside the support plane for the newly discovered vertex, then we can terminate
			if ( fDelta <= fTerminationDistance || fSeparation >= 0.0f || uPhaseTwoIterations>=20 )
			{
				xAxis = xNormal;
				return bOriginInside;
			}

			// KJFL - 21 July 2008 - rearrange and optimised with vector triple product
			// a * (b ^ c) = c * ( a ^ b )

			Asura_AVector3 xCross = xV0 ^ xV4;

			// Compute the tetrahedron dividing face (xV4,xV0,xV1)
			//float fD1 = (xV4 ^ xV1) * xV0;

			// Compute the tetrahedron dividing face (xV4,xV0,xV2)
			//float fD2 = (xV4 ^ xV2) * xV0;

			// Compute the tetrahedron dividing face (xV4,xV0,xV3)
			//float fD3 = (xV4 ^ xV3) * xV0;

			// KJFL - 21 July 2008 - TODO - can we make this branch free?
			// e.g. do we have Select... methods optimised on platforms?

			if ( xCross * xV1 < 0.0f )
			{
				if ( xCross * xV2 < 0.0f )
				{
					// Inside fD1 & inside fD2 ==> eliminate xV1
					xV1 = xV4;
				}
				else
				{
					// Inside fD1 & outside fD2 ==> eliminate xV3
					xV3 = xV4;
				}
			}
			else
			{
				if ( xCross * xV3 < 0.0f )
				{
					// Outside fD1 & inside fD3 ==> eliminate xV2
					xV2 = xV4;
				}
				else
				{
					// Outside fD1 & outside fD3 ==> eliminate xV1
					xV1 = xV4;
				}
			}
		}
	}
}

bool Asura_Physics_CCD::Swept_MPR_Test( const Asura_Physics_MPR_Body& xBody1, const Asura_Physics_MPR_Body& xBody2, Asura_AVector3& xAxis, float& fRewindTime, float& fPenetration  )
{
	// Keep one object "stationary", and sweep the other object by the relative motion
	Asura_AVector3 xVel1 = xBody1.GetCentre() - xBody1.GetPreviousCentre();
	Asura_AVector3 xVel2 = xBody2.GetCentre() - xBody2.GetPreviousCentre();
	Asura_AVector3 xSweepBody2 = (xVel1 - xVel2);

	// Make a conservative estimate of the first and last time of intersection,
	// by looking at the overlap along the axis of relative velocity
	Asura_AVector3 xNormal1 = xSweepBody2;

	float fEstimatedSpeed = xNormal1.NormaliseReturningMagnitude();

	float fMax = 1.0f;
	float fMin = 0.0f;

	if (fEstimatedSpeed>0.0f)
	{
		xNormal1.Normalise();
		Asura_AVector3 xX1 = xBody2.GetSupportPoint( -xNormal1, true ) - xBody1.GetSupportPoint( xNormal1, true );
		Asura_AVector3 xX2 = xBody2.GetSupportPoint( xNormal1, true ) - xBody1.GetSupportPoint( -xNormal1, true );

		fMax = xX1*xNormal1 / xSweepBody2.Magnitude();
		fMin = xX2*xNormal1 / xSweepBody2.Magnitude();
	
		// If no overlap just reject - otherwise we'll keep 
		// these min and max values to clamp any later result.
		if ( fMax < fMin || fMin >= 1.0f )
		{
			xAxis = xNormal1;
			return false;
		}	
	}

	// This is the distance at which we give up searching.
	// This means we may miss an intersection smaller than this (in m).
	const float fTerminationDistance = 1e-6f;

	// define xV0 = center of Minkowski sum
	Asura_AVector3 xV0;

	{
		xV0 = xBody2.GetCentre() - xBody1.GetCentre();

		// Avoid case where centers overlap -- any direction is fine in this case
		if (xV0 == Asura_Maths::ZeroAVector3 ) xV0 = Asura_AVector3(0.01f, 0.0f, 0.0f);
	}

	Asura_AVector3 xNormal = -xV0;
	xNormal.Normalise();

	// find xV1 = support in direction of origin
	Asura_AVector3 xV1 = xBody2.GetSweptSupportPoint( -xNormal, xSweepBody2 ) - xBody1.GetSupportPoint( xNormal );

	if ( xV1 * xNormal <= 0.0f )
	{
		xAxis = xNormal;
		return false;
	}

	xNormal = xV1 ^ xV0;

	// KJFL - 31 July 2008 - special case when xV0 & xV1 are parallel.
	// If xNormal is zero length, this means xV0, xV1 & the origin are colinear.
	// If xV0 is on one side of the origin by definition, and we've already
	// checked xV1 is on the other side of the origin, that must mean the origin
	// is on a line between the two. This means there is an intersection.
	if ( xNormal.NormaliseReturningMagnitude() == 0.0f )
	{
		xNormal = xV1 - xV0;
		xNormal.Normalise();
		xAxis = xNormal;
	
		float fSpeed = Asura_Maths::Max( - (xSweepBody2 * xNormal), 1e-8f);

		fPenetration = (xV1*xNormal);

		fRewindTime = fPenetration/fSpeed;
				
		// TODO - should be able to reduce the number of Min & Maxes ?
		fRewindTime = Asura_Maths::Max( fRewindTime, 1.0f-fMax );
		fRewindTime = Asura_Maths::Min( fRewindTime, 1.0f-fMin );

		fRewindTime = Asura_Maths::Max( fRewindTime, 0.0f );
		fRewindTime = Asura_Maths::Min( fRewindTime, 1.0f );

		return true;
	}

	// find xV2 - support perpendicular to xV1,xV0
	Asura_AVector3 xV2 = xBody2.GetSweptSupportPoint( -xNormal, xSweepBody2 ) - xBody1.GetSupportPoint( xNormal );

	{
		float fDot = xV2 * xNormal;
		if (fDot <= 0.0f)
		{
			xAxis = xNormal;
			return false;
		}
	}

	// Determine whether origin is on + or - side of plane (xV1,xV0,xV2)
	xNormal = (xV1 - xV0) ^ (xV2 - xV0);
	xNormal.Normalise();

	// TODO - can xNormal be zero here?

	// If the origin is on the -ve side of the plane, reverse the direction of the plane
	if ( xNormal * xV0 > 0.0f )
	{
		Swap(xV1, xV2);
		xNormal = -xNormal;
	}

	// Phase One: Identify a portal

	u_int uPhaseOneIterations = 0;
	while (1)
	{
		++uPhaseOneIterations;
		// Obtain the support point in a direction perpendicular to the existing plane
		// Note: This point is guaranteed to lie off the plane

		Asura_AVector3 xV3 = xBody2.GetSweptSupportPoint( -xNormal, xSweepBody2 ) - xBody1.GetSupportPoint( xNormal );

		if (xV3*xNormal <= 0.0f)
		{
			xAxis = xNormal;
			return false;
		}

		// If origin is outside (xV1,xV0,xV3), then eliminate xV2 and loop
		if ((xV1 ^ xV3) * xV0 < 0.0f)
		{
			xV2 = xV3;
			xNormal = (xV1 - xV0) ^ (xV3 - xV0);
			xNormal.Normalise();
			continue;
		}

		// If origin is outside (xV3,xV0,xV2), then eliminate xV1 and loop
		if ((xV3 ^ xV2) * xV0 < 0.0f)
		{
			xV1 = xV3;
			xNormal = (xV3 - xV0) ^ (xV2 - xV0);
			xNormal.Normalise();
			continue;
		}

		bool bOriginInside = false;

		// Phase Two: Refine the portal
		u_int uPhaseTwoIterations=0;

		// We are now inside of a wedge...
		while (1)
		{
			++uPhaseTwoIterations;

			// Compute normal of the wedge face
			xNormal = (xV2 - xV1) ^ (xV3 - xV1);

			// Can this happen???  Can it be handled more cleanly?
			if ( xNormal.NormaliseReturningMagnitude() == 0.0f )
			{
				Asura_Assert( false, "Potential code bug: notify KJFL" );
				return false;
			}

			// Compute distance from origin to wedge face
			if (!bOriginInside)
			{
				float fDist = xNormal * xV1;

				// If the origin is inside the wedge, we have an intersection
 				if (fDist >= 0.0f)
				{
					bOriginInside = true;
				}
			}

			// Find the support point in the direction of the wedge face
			Asura_AVector3 xV4 = xBody2.GetSweptSupportPoint( -xNormal, xSweepBody2 ) - xBody1.GetSupportPoint( xNormal );

			float fDelta = (xV4 - xV3) * xNormal;
			fRewindTime = -(xV4 * xNormal);



			// If the boundary is thin enough or the origin is outside the support plane for the newly discovered vertex, then we can terminate
			if ( fDelta <= fTerminationDistance || fRewindTime >= 0.0f || uPhaseTwoIterations>=20 )
			{
				xAxis = xNormal;

				float fSpeed = Asura_Maths::Max( - (xSweepBody2 * xNormal), 1e-8f);

				fPenetration = -fRewindTime;
				fRewindTime = fPenetration/fSpeed;

				// TODO - should be able to reduce the number of Min & Maxes ?
				fRewindTime = Asura_Maths::Max( fRewindTime, 1.0f-fMax );
				fRewindTime = Asura_Maths::Min( fRewindTime, 1.0f-fMin );

				fRewindTime = Asura_Maths::Max( fRewindTime, 0.0f );
				fRewindTime = Asura_Maths::Min( fRewindTime, 1.0f );

				return bOriginInside;
			}

			// KJFL - 21 July 2008 - rearrange and optimised with vector triple product
			// a * (b ^ c) = c * ( a ^ b )

			Asura_AVector3 xCross = xV0 ^ xV4;

			// Compute the tetrahedron dividing face (xV4,xV0,xV1)
			//float fD1 = (xV4 ^ xV1) * xV0;

			// Compute the tetrahedron dividing face (xV4,xV0,xV2)
			//float fD2 = (xV4 ^ xV2) * xV0;

			// Compute the tetrahedron dividing face (xV4,xV0,xV3)
			//float fD3 = (xV4 ^ xV3) * xV0;

			// KJFL - 21 July 2008 - TODO - can we make this branch free?
			// e.g. do we have Select... methods optimised on platforms?

			if ( xCross * xV1 < 0.0f )
			{
				if ( xCross * xV2 < 0.0f )
				{
					// Inside fD1 & inside fD2 ==> eliminate xV1
					xV1 = xV4;
				}
				else
				{
					// Inside fD1 & outside fD2 ==> eliminate xV3
					xV3 = xV4;
				}
			}
			else
			{
				if ( xCross * xV3 < 0.0f )
				{
					// Outside fD1 & inside fD3 ==> eliminate xV2
					xV2 = xV4;
				}
				else
				{
					// Outside fD1 & outside fD3 ==> eliminate xV1
					xV1 = xV4;
				}
			}
		}
	}
}

