//
// Asura_Physics_Environment.h
//
// Created by KJFL - 2004-03-24 16:43:47
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_PHYSICS_ENVIRONMENT_H__
#define __ASURA_PHYSICS_ENVIRONMENT_H__

#include "Core/Asura_HashTree.h"
#include "Core/Asura_LBAABB_Tree.h"

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	#include "Core/Asura_MaterialResponse_System.h"
#endif

#include "Core/Asura_Maths.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Rendering/Asura_Module.h"


// Forward declarations
class Asura_Chunk_Stream;
class Asura_Chunk_PhysicsEnvironment;


#ifdef TOOLS_INCLUDE_PHYSICS_MODULE_AND_POLY_LOOKUP_DATA
#ifndef ASURA_TOOLS
#error TOOLS_INCLUDE_PHYSICS_MODULE_AND_POLY_LOOKUP_DATA should only be defined in tools builds.
#endif
#endif

// MGH 7/4/2006: When streaming is enabled the indices for polygons, vertices and edges
// in the physics cache have a zone ID (6 bits), a sector ID (10 bits) and an index (16 bits).
// When streaming is disabled they have a sector ID (16 bits) and an index (16 bits).
#if defined(ASURA_USE_STREAMING)

// Number of bits reserved for zone index. (Projects can define this in Asura_EngineConfiguration.h if necessary).
#ifndef ASURA_ZONE_INDEX_NUM_BITS
#define ASURA_ZONE_INDEX_NUM_BITS (6)
#endif

// Maximum number of zones that can be simultaneously loaded.
const u_int uASURA_MAX_NUM_ZONES_LOADED = (1<<ASURA_ZONE_INDEX_NUM_BITS);
// Number of bits to shift to get to the zone index.
const u_int uASURA_ZONE_INDEX_SHIFT = (32-ASURA_ZONE_INDEX_NUM_BITS);
// Mask for the zone index bits.
const u_int uASURA_ZONE_INDEX_MASK  = ((uASURA_MAX_NUM_ZONES_LOADED-1)<<uASURA_ZONE_INDEX_SHIFT);
// Extract the sector from the index that contains a zone, sector and primitive ID.
inline u_int Asura_GetSectorFromIndex(const u_int uZoneSectorID) { return ((uZoneSectorID & (~uASURA_ZONE_INDEX_MASK))>>16); }

#else

// Extract the sector from the index that contains a sector and primitive ID.
inline u_int Asura_GetSectorFromIndex(const u_int uSectorID) { return (uSectorID>>16); }

#endif //defined(ASURA_USE_STREAMING)




class Asura_Physics_Environment_Face
{
public:
	inline u_int GetVertexIndex( u_int uVertex ) const
	{
		return static_cast<u_int>(m_usVertexID[uVertex]);
	}

	u_short	m_usVertexID[4]; // Fourth vertex will be 0xffff if NULL.

#if defined(ASURA_TOOLS)
	void WriteToChunkStream (Asura_Chunk_Stream& xStream) const;
#endif
};

class Asura_Physics_Environment_FaceFlags
{
public:
	union
	{
		struct
		{
			// NOTE: These bits are written out, adding any new bits means
			// ~~~~~ that you will need to add those bits to the bitfield
			//       preparation code in "WriteToChunkStream".
			u_char	m_bEdge0IsConvex : 1,
					m_bEdge1IsConvex : 1,
					m_bEdge2IsConvex : 1,
					m_bEdge3IsConvex : 1,

					m_bVertex0IsConvex : 1,
					m_bVertex1IsConvex : 1,
					m_bVertex2IsConvex : 1,
					m_bVertex3IsConvex : 1;
		};

		u_char m_ucFlags;
	};

#if defined(ASURA_TOOLS)
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
#endif
};


class Asura_Physics_Environment_Vertex
{
public:
	inline const Asura_AVector3& GetPosition( void ) const
	{
		return m_xPosition;
	}

	Asura_AVector3 m_xPosition;

#if defined(ASURA_TOOLS)
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
#endif
};

class Asura_Physics_Environment_Sector
{
public: 
	u_int				m_uFaceIndexOffset;
	u_int				m_uVertexIndexOffset;

	u_int				m_uNumberOfFaces;
	Asura_LBAABB_Tree	m_xFaceTree;

	u_int				m_uPolygonFlagOffset;
	u_int				m_uMaterialIDOffset;

	union
	{
		struct
		{
			// NOTE: These bits are written out, adding any new bits means
			// ~~~~~ that you will need to add those bits to the bitfield
			//       preparation code in "WriteToChunkStream".
			u_short		m_bAllFacesUseSamePolygonFlag	:1,
						m_bAllFacesUseSameMaterialID	:1,
														:0;
		};

		u_short	m_usFlags;
	};

#if defined(ASURA_TOOLS)
	void WriteToChunkStream( Asura_Chunk_Stream& xStream ) const;
#endif
};


class Asura_Physics_Zone
{
public:
	Asura_Physics_Zone(void);
	~Asura_Physics_Zone();

	bool Process( Asura_Handle hFile, const Asura_Chunk_PhysicsEnvironment& xChunk );
	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
#ifdef ASURA_TOOLS
	void WriteToChunkStream(Asura_Chunk_Stream& xStream,const int* piMaterialIDToIndexConversion) const;
	void Tools_Create( bool (*pfnShouldIncludeModule)(int iModuleIndex)=0 );
	void Invalidate() {m_bInvalid = true;}
	void ConvertMaterialIndecesToIDs(const int* piMaterialIndexToIDConversion);
#endif


	void Reset( void );

	// Query the environment polys to see what a ray hits first 
	bool FindPolyInLOS( const class Asura_LOS_InputData* pxInputData, class Asura_LOS_OutputData* pxOutputData );

	// Fill an environment cache with the environment poly face, edges and vertices that are inside the cache bounding volume
	void FillCache( class Asura_Physics_EnvironmentCache* pxCache );

	inline const Asura_AVector3& GetVertex( u_int uVertexID ) const
	{
		const u_int uSector = Asura_GetSectorFromIndex(uVertexID);
		const u_int u = (uVertexID)&0xffff;

		return m_pxVertices[m_pxSectors[uSector].m_uVertexIndexOffset+u].m_xPosition;
	}

	inline const Asura_AVector3& GetFaceVertex( u_int uFaceID, u_int uVertex ) const
	{
		const u_int uSector = Asura_GetSectorFromIndex(uFaceID);
		const u_int u = ((uFaceID)&0xffff)>>1;

		bool bSecondTriOfQuad = (uFaceID & 0x1);

		const Asura_Physics_Environment_Face& xFace = m_pxFaces[m_pxSectors[uSector].m_uFaceIndexOffset+u];

		if (bSecondTriOfQuad)
		{
			if (uVertex == 2)
			{
				uVertex = 0;
			}
			else
			{
				uVertex += 2;
			}
		}

		const u_int v = xFace.GetVertexIndex(uVertex);

		return m_pxVertices[m_pxSectors[uSector].m_uVertexIndexOffset+v].m_xPosition;
	}

	inline u_short GetFaceFlags( u_int uFaceID ) const
	{
		const u_int uSector = Asura_GetSectorFromIndex(uFaceID);
		const u_int u = ((uFaceID)&0xffff)>>1;

		const Asura_Physics_Environment_Sector& xSector =  m_pxSectors[uSector];

		return GetFlagsForPolygonInSector( xSector, u );
	}

#ifdef ASURA_TOOLS
    inline u_int Tools_GetFaceFlags( u_int uFaceID ) const
    {
        const u_int uSector = Asura_GetSectorFromIndex(uFaceID);
        const u_int u = ((uFaceID)&0xffff)>>1;

        const Asura_Physics_Environment_Sector& xSector =  m_pxSectors[uSector];

        return Tools_GetFlagsForPolygonInSector( xSector, u );
    }

#ifdef TOOLS_INCLUDE_PHYSICS_MODULE_AND_POLY_LOOKUP_DATA
	inline void Tools_CalcModuleAndPolyIndex( const u_int uFaceID, u_int& uModuleIndex, u_int& uPolyIndex ) const
	{
		const u_int uSector    = Asura_GetSectorFromIndex( uFaceID );
		const u_int uFaceIndex = ((uFaceID)&0xffff)>>1;
		uModuleIndex = m_puSectorLookupArray ? m_puSectorLookupArray[ uSector ] : 0;
		uPolyIndex   = m_ppuPolyLookupArrays ? m_ppuPolyLookupArrays[ uModuleIndex ][ uFaceIndex ] : 0;
	}
#endif
#endif // ASURA_TOOLS

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	inline Asura_Hash_ID GetFaceMaterialResponse(u_int uFaceID) const
	{
		const u_int uSector = Asura_GetSectorFromIndex(uFaceID);
        const u_int u = ((uFaceID)&0xffff)>>1;

        const Asura_Physics_Environment_Sector& xSector = m_pxSectors[uSector];
		return GetMaterialResponseForPolygonInSector(xSector, u);
	}
#endif //ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	bool DoesFaceIntersectAABB( const u_int uFaceID, const Asura_Bounding_Box& xTestBB ) const;
	bool DoesFaceIntersectOBB( const u_int uFaceID, const Asura_Collision_Object& xTestObject, const Asura_Vector_3 xObjectPosition, const Asura_Matrix_3x3 xObjectOrientation ) const;

	inline u_int GetNumberOfSectors( void ) const
	{
		return m_uNumberOfSectors;
	}

	inline u_int GetNumberOfVertices( void ) const
	{
		return m_uTotalNumberOfVertices;
	}

	inline u_int GetNumberOfFaces( void ) const
	{
		return m_uTotalNumberOfFaces;
	}

	inline u_int GetNumberOfPolyFlags( void ) const
	{
		return m_uTotalNumberOfPolygonFlags;
	}

	inline u_int GetNumberOfMaterialIDs( void ) const
	{
		return m_uTotalNumberOfMaterialIDs;
	}

	inline u_int GetZoneIndex() const { return m_uZoneIndex; }

private:

#ifdef ASURA_TOOLS
	void CreateFaceTrees( void );
#endif

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	inline Asura_Hash_ID GetMaterialResponseForPolygonInSector( const Asura_Physics_Environment_Sector& xSector, const u_int uPolygon ) const
	{
		if (xSector.m_bAllFacesUseSameMaterialID)
		{
			return m_puMaterialResponseHashIDs[ xSector.m_uMaterialIDOffset ];
		}
		else
		{
			return m_puMaterialResponseHashIDs[ xSector.m_uMaterialIDOffset + uPolygon ];
		}
	}

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	inline short GetMaterialForPolygonInSector( const Asura_Physics_Environment_Sector& xSector, const u_int uPolygon ) const
	{
		if (xSector.m_bAllFacesUseSameMaterialID)
		{
			return m_psMaterialIDs[ xSector.m_uMaterialIDOffset ];
		}
		else
		{
			return m_psMaterialIDs[ xSector.m_uMaterialIDOffset + uPolygon ];
		}
	}

	inline u_short GetFlagsForPolygonInSector( const Asura_Physics_Environment_Sector& xSector, const u_int uPolygon ) const
	{
		u_short usFlags;
		if (xSector.m_bAllFacesUseSamePolygonFlag)
		{
			usFlags = m_pusPolygonFlags[ xSector.m_uPolygonFlagOffset ];
		}
		else
		{
			usFlags = m_pusPolygonFlags[ xSector.m_uPolygonFlagOffset + uPolygon ];
		}

		//Check for any material flags that need to be transferred to the polygon flags
		//(For polygons with texture anims)
		//RAR 05/06/2006

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		const Asura_Hash_ID uMaterialResponseHashID = GetMaterialResponseForPolygonInSector( xSector, uPolygon );
		if( uMaterialResponseHashID != ASURA_HASH_ID_UNSET )
		{
			const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialResponseHashID );
			if( pxMaterial )
			{
				if( pxMaterial->GetDontRender() ) usFlags |= AsuraPolygonFlag_DontRender;
				if( pxMaterial->GetNoCollisions() ) usFlags |= AsuraPolygonFlag_NoCollisions;
				if( pxMaterial->GetIgnoreLOS() ) usFlags |= AsuraPolygonFlag_IgnoreLOS;
			}
		}

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		short sMaterial = GetMaterialForPolygonInSector(xSector,uPolygon);
		u_int uMatFlags = Asura_Material_System::GetMaterialFlags(sMaterial);
		if(uMatFlags)
		{
			if(uMatFlags & AsuraMaterialFlag_DontRender) usFlags |= AsuraPolygonFlag_DontRender;
			if(uMatFlags & AsuraMaterialFlag_NoCollisions) usFlags |= AsuraPolygonFlag_NoCollisions;
			if(uMatFlags & AsuraMaterialFlag_IgnoreLOS) usFlags |= AsuraPolygonFlag_IgnoreLOS;
		}
		return usFlags;

	}

#ifdef ASURA_TOOLS
    inline u_int Tools_GetFlagsForPolygonInSector( const Asura_Physics_Environment_Sector& xSector, const u_int uPolygon ) const
    {
        u_int uFlags;
        if (xSector.m_bAllFacesUseSamePolygonFlag)
        {
            uFlags = m_puToolsPolygonFlags[ xSector.m_uPolygonFlagOffset ];
        }
        else
        {
            uFlags = m_puToolsPolygonFlags[ xSector.m_uPolygonFlagOffset + uPolygon ];
        }

        //Check for any material flags that need to be transferred to the polygon flags
        //(For polygons with texture anims)
        //RAR 05/06/2006

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		const Asura_Hash_ID uMaterialResponseHashID = GetMaterialResponseForPolygonInSector( xSector, uPolygon );
		if( uMaterialResponseHashID != ASURA_HASH_ID_UNSET )
		{
			const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialResponseHashID );
			if( pxMaterial )
			{
				if( pxMaterial->GetDontRender() ) uFlags |= AsuraPolygonFlag_DontRender;
				if( pxMaterial->GetNoCollisions() ) uFlags |= AsuraPolygonFlag_NoCollisions;
				if( pxMaterial->GetIgnoreLOS() ) uFlags |= AsuraPolygonFlag_IgnoreLOS;
			}
		}

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		short sMaterial = GetMaterialForPolygonInSector(xSector,uPolygon);
		u_int uMatFlags = Asura_Material_System::GetMaterialFlags(sMaterial);
		if(uMatFlags)
		{
			if(uMatFlags & AsuraMaterialFlag_DontRender) uFlags |= AsuraPolygonFlag_DontRender;
			if(uMatFlags & AsuraMaterialFlag_NoCollisions) uFlags |= AsuraPolygonFlag_NoCollisions;
			if(uMatFlags & AsuraMaterialFlag_IgnoreLOS) uFlags |= AsuraPolygonFlag_IgnoreLOS;
		}
        return uFlags;
    }
#endif

	// callback functions used by FindPolyInLOS();
	static void FindPoly_SectorCallback( const u_int uIndex );
	static void FindPoly_PolyCallback( const u_int uIndex );

	// callback functions used by FillCache
	static void FillCache_SectorCallback( const u_int uIndex );
	static void FaceSearchCallback( const u_int uIndex );

	// Private variables...
	Asura_Physics_Environment_Sector*		m_pxSectors;
	u_int									m_uNumberOfSectors;
	Asura_LBAABB_Tree						m_xSectorsTree;

	u_int									m_uTotalNumberOfFaces;
	Asura_Physics_Environment_Face*			m_pxFaces;
	Asura_Physics_Environment_FaceFlags*	m_pxFaceFlags;

	u_int									m_uTotalNumberOfVertices;
	Asura_Physics_Environment_Vertex*		m_pxVertices;

	u_int									m_uTotalNumberOfPolygonFlags;
    u_short*								m_pusPolygonFlags; 
#ifdef ASURA_TOOLS
    u_int*								    m_puToolsPolygonFlags; 
#endif

	u_int									m_uTotalNumberOfMaterialIDs;

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	Asura_Hash_ID*							m_puMaterialResponseHashIDs;
#endif
	short*									m_psMaterialIDs;

	bool									m_bTreesUseOneHugeNodeArray;
	Asura_LBAABB_Tree_Double_Node*			m_pxOneHugeNodeArray;

	u_int									m_uZoneIndex;

#ifdef ASURA_TOOLS
	bool									m_bInvalid;
#endif

#ifdef TOOLS_INCLUDE_PHYSICS_MODULE_AND_POLY_LOOKUP_DATA
	u_int*									m_puSectorLookupArray;
	u_int**									m_ppuPolyLookupArrays;
#endif

	static u_int							s_uCurrentAABBSector;
	static Asura_Physics_EnvironmentCache*	s_pxCurrentCache;
	static Asura_Physics_Zone*				s_pxCurrentZone;
};


class Asura_Physics_Environment
{
public:

#ifdef ASURA_TOOLS
	static	void Tools_Create(bool (*pfnShouldIncludeModule)(int iModuleIndex)=0) { GetCurrentZone().Tools_Create(pfnShouldIncludeModule); }
	static  void Invalidate() {GetCurrentZone().Invalidate();}
	static  void ConvertMaterialIndecesToIDs(const int* piMaterialIndexToIDConversion) { GetCurrentZone().ConvertMaterialIndecesToIDs(piMaterialIndexToIDConversion); }
#endif

	static	void Initialise(void);
	static	void Reset(void);
	static	void Shutdown(void);

	// Query the environment polys to see what a ray hits first 
	static	bool FindPolyInLOS(const Asura_LOS_InputData* pxInputData, Asura_LOS_OutputData* pxOutputData);

	// Fill an environment cache with the environment poly face, edges and vertices that are inside the cache bounding volume
	static	void FillCache(Asura_Physics_EnvironmentCache* pxCache);

	static	bool DoesFaceIntersectAABB(const Asura_AVector3& xV0, const Asura_AVector3& xV1, const Asura_AVector3& xV2, const Asura_Bounding_Box& xTestBB);

	static	bool DoesEdgeIntersectAABB(const Asura_AVector3& xEdge0, const Asura_AVector3& xEdge1, const Asura_Bounding_Box& xTestBB);

	// Accessor functions...
	static	u_int GetNumberOfSectors(void) { return(GetCurrentZone().GetNumberOfSectors()); }
	static	u_int GetNumberOfVertices(void) { return(GetCurrentZone().GetNumberOfVertices()); }
	static	u_int GetNumberOfFaces(void) { return(GetCurrentZone().GetNumberOfFaces()); }
	static	u_int GetNumberOfPolyFlags(void) { return(GetCurrentZone().GetNumberOfPolyFlags()); }
	static	u_int GetNumberOfMaterialIDs(void) { return(GetCurrentZone().GetNumberOfMaterialIDs()); }


#if defined(ASURA_USE_STREAMING)
	// Zone functions...
	static	Asura_Physics_Zone* CreateZone(const Asura_Hash_ID uHash);
	static	bool DestroyZone(const Asura_Hash_ID uHash);

	static	void CreateAndSetZone(const Asura_Hash_ID uHash) { s_pxCurrentZone = CreateZone(uHash); }
	static	void SetCurrentZone(Asura_Physics_Zone* pxZone) { s_pxCurrentZone = pxZone; }
	static	void SetCurrentZone(const Asura_Hash_ID uHash) { s_pxCurrentZone = s_xZoneTree.GetDataByHash(uHash); }
	static	Asura_Physics_Zone& GetCurrentZone(void) { return(*s_pxCurrentZone); }
	static  Asura_Physics_Zone* GetZone(const Asura_Hash_ID uHash) { return s_xZoneTree.GetDataByHash(uHash); }

	static bool FillCache_ZoneCallback(Asura_Hash_ID uHash, Asura_Physics_Zone* pxZone, void* pxCache);
	static bool FindPolyInLOS_ZoneCallback(Asura_Hash_ID uHash, Asura_Physics_Zone* pxZone, void* pbHitPolygon);
	
	// Parameter: a FaceID, VertexID or EdgeID. Returns true if the corresponding zone is loaded.
	static bool IsValid(u_int uZoneSectorID) { return GetZoneFromIndex(uZoneSectorID) != 0; }
	static Asura_Physics_Zone* GetZoneFromIndex(u_int uZoneSectorID);

	// Accessor functions
	static u_short GetFaceFlags(u_int uFaceID)
	{
		const Asura_Physics_Zone* const pxZone = GetZoneFromIndex(uFaceID);
		return pxZone ? pxZone->GetFaceFlags(uFaceID) : 0;
	}
#ifdef ASURA_TOOLS
    static u_int Tools_GetFaceFlags(u_int uFaceID)
    {
        const Asura_Physics_Zone* const pxZone = GetZoneFromIndex(uFaceID);
        return pxZone ? pxZone->Tools_GetFaceFlags(uFaceID) : 0;
    }
#endif
	static const Asura_AVector3& GetFaceVertex(u_int uFaceID, u_int uVertex)
	{
		const Asura_Physics_Zone* const pxZone = GetZoneFromIndex(uFaceID);
		return pxZone ? pxZone->GetFaceVertex(uFaceID, uVertex) : Asura_Maths::ZeroAVector3;
	}
	static const Asura_AVector3& GetVertex(u_int uVertexID)
	{
		const Asura_Physics_Zone* const pxZone = GetZoneFromIndex(uVertexID);
		return pxZone ? pxZone->GetVertex(uVertexID) : Asura_Maths::ZeroAVector3;
	}

	static bool DoesFaceIntersectAABB(const u_int uFaceID, const Asura_Bounding_Box& xTestBB)
	{ 
		const Asura_Physics_Zone* const pxZone = GetZoneFromIndex(uFaceID);
		return pxZone ? pxZone->DoesFaceIntersectAABB(uFaceID, xTestBB) : false;
	}
	static bool DoesFaceIntersectOBB( const u_int uFaceID, const Asura_Collision_Object& xTestObject, const Asura_Vector_3 xObjectPosition, const Asura_Matrix_3x3 xObjectOrientation )
	{ 
		const Asura_Physics_Zone* const pxZone = GetZoneFromIndex(uFaceID);
		return pxZone ? pxZone->DoesFaceIntersectOBB(uFaceID, xTestObject, xObjectPosition, xObjectOrientation) : false;
	}
#else
	// Zone functions...
	static Asura_Physics_Zone& GetCurrentZone(void) { return(s_xZone); }
	// Only a single zone, so it is always loaded.
	static bool IsValid(u_int uZoneSectorID) { return true; }

	// Accessor functions
	static u_short GetFaceFlags(u_int uFaceID)									{ return s_xZone.GetFaceFlags(uFaceID); }
#ifdef ASURA_TOOLS
    static u_int Tools_GetFaceFlags(u_int uFaceID)                              { return s_xZone.Tools_GetFaceFlags(uFaceID); }
#ifdef TOOLS_INCLUDE_PHYSICS_MODULE_AND_POLY_LOOKUP_DATA
	static void Tools_CalcModuleAndPolyIndex( const u_int uFaceID, u_int& uModuleIndex, u_int& uPolyIndex ) { s_xZone.Tools_CalcModuleAndPolyIndex( uFaceID, uModuleIndex, uPolyIndex ); }
#endif
#endif // ASURA_TOOLS
	static const Asura_AVector3& GetFaceVertex(u_int uFaceID, u_int uVertex)	{ return s_xZone.GetFaceVertex(uFaceID, uVertex); }
	static const Asura_AVector3& GetVertex(u_int uVertexID)						{ return s_xZone.GetVertex(uVertexID); }

	static bool DoesFaceIntersectAABB( const u_int uFaceID, const Asura_Bounding_Box& xTestBB )
	{ 
		return s_xZone.DoesFaceIntersectAABB( uFaceID, xTestBB ); 
	}
	static bool DoesFaceIntersectOBB( const u_int uFaceID, const Asura_Collision_Object& xTestObject, const Asura_Vector_3 xObjectPosition, const Asura_Matrix_3x3 xObjectOrientation ) 
	{ 
		return s_xZone.DoesFaceIntersectOBB( uFaceID, xTestObject, xObjectPosition, xObjectOrientation ); 
	}

#endif	// defined(ASURA_USE_STREAMING)


private:
#if defined(ASURA_USE_STREAMING)
	// Private variables...
	static	Asura_Physics_Zone*							s_pxCurrentZone;		// Currently active handler.
	static	Asura_TemplateHashTree<Asura_Physics_Zone>	s_xZoneTree;			// Tree of all of the handlers available.
#else
	// Private variables...
	static	Asura_Physics_Zone							s_xZone;				// Single static-instance of a zone.
#endif	// defined(ASURA_USE_STREAMING)
};



#endif//__ASURA_PHYSICS_ENVIRONMENT_H__
