//
// Asura_Physics_Island.h
//
// Created by KJFL - 02 November 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_ISLAND_H__
#define __ASURA_PHYSICS_ISLAND_H__

class Asura_Physics_Island;
class Asura_Physics_Island_ObjectNode;
class Asura_Physics_Island_PairingNode;
class Asura_Physics_Pairing;
class Asura_Physics_RigidBodyInstance;

#include "Debugging/Asura_Assert.h"
// Need this include now because using a accessor from the class
#include "Physics/Asura_Physics_RigidBody.h"

// KJFL - this may become a generic class. Waiting to see what design needs the Island code / parallelism
// adds to the Physics system.
//
// Basically it acts as an 'allocate-only' memory heap that gets completely wiped before use, so there's
// no overhead for tracking allocations or handling deallocs.
class Asura_Physics_MemoryPool
{
	public:

		Asura_Physics_MemoryPool( u_int uSize )
		{
			m_pucBuffer = reinterpret_cast<u_char*>( Asura_AllocateAlign( uSize, ASURA_ALIGN_16_BYTE ) );
			m_pucEndOfBuffer = m_pucBuffer + uSize;

			Reset();
		}

		void Reset( void )
		{
			m_pucFreeMem = m_pucBuffer;
		}

		~Asura_Physics_MemoryPool( void )
		{
			Asura_DeallocateAlign( m_pucBuffer );
		}

		u_char* AllocateBytes( u_int uBytes )
		{
			// TODO handle running out of memory!

			u_char* pucMemory = m_pucFreeMem;

			m_pucFreeMem += uBytes;

			if (m_pucFreeMem <= m_pucEndOfBuffer)
			{
				return pucMemory;
			}
			Asura_Assert(false, "out of memory for physics islands");
			return NULL;
		}

	private:

		u_char*	m_pucFreeMem;
		u_char*	m_pucBuffer;
		u_char* m_pucEndOfBuffer;
};

class Asura_Physics_Island_ObjectNode
{
	public:
		Asura_Physics_RigidBodyInstance*	m_pxObject;
		Asura_Physics_Island_ObjectNode*	m_pxNextObjectNode;
};


class Asura_Physics_Island_PairingNode
{
	public:
		Asura_Physics_Pairing*				m_pxPairing;
		Asura_Physics_Island_PairingNode*	m_pxNextPairingNode;
};




class Asura_Physics_Island
{
	public:

		static void ResetAllIslands( void )
		{
			s_xMemory.Reset();
			s_pxFirstIsland = 0;
		}

		static Asura_Physics_Island* GetFirstIsland( void )
		{
			return s_pxFirstIsland;
		}


        static Asura_Physics_Island* Create( void )
		{
			Asura_Physics_Island* pxIsland = reinterpret_cast<Asura_Physics_Island*>( s_xMemory.AllocateBytes( sizeof( Asura_Physics_Island ) ) );

			if ( pxIsland  )
			{
				// blank out object and pairing entries
				pxIsland->m_pxObjectNode = 0;
				pxIsland->m_pxPairingNode = 0;

				// add island to head of list
				pxIsland->m_pxNextIsland = s_pxFirstIsland;
				s_pxFirstIsland = pxIsland;
			}

			return pxIsland;
					
		}

		static void AssignSameIsland( Asura_Physics_RigidBodyInstance* pxInstanceA, Asura_Physics_RigidBodyInstance* pxInstanceB, Asura_Physics_Pairing* pxPairing = 0 )
		{
			// KJFL - 31 January 2006 - Okay, we have three basic scenarios to deal with:
			//
			// 1. Neither object is in a island yet. In this case we need to make an island,
			// add the pairing, and add any dynamic objects.
			//
			// 2. Both of the objects has an island. This has two sub-categories:
			//    2a. The islands are the same - i.e. we've done this object before!
			//    2b. We need to merge the islands, and add the new pairing.
			//
			// 3. One of the objects has an island. 
			//    3a. If both objects are dynamic then we need to add the pairing and the new
			//        object to the existing island. 
			//    3b. Otherwise we have a dynamic object + static object pairing, and the dynamic 
			//        object already has an island. We may have handled this pairing already.
			//

			// Case 1: neither object has an island
			if ( (!pxInstanceA->GetIsland()) && (!pxInstanceB->GetIsland()) )
			{
				Asura_Physics_Island* pxIsland = Asura_Physics_Island::Create();

				// a new island, so we'll always have to add the pairing
				if ( pxPairing ) pxIsland->AddPairing( pxPairing );

				// and we also need to add any dynamic objects to the island too
				if ( pxInstanceA->IsDynamic() )
				{
					pxIsland->AddObject( pxInstanceA );
				}

				if ( pxInstanceB->IsDynamic() )
				{
					pxIsland->AddObject( pxInstanceB );
				}
			}
			// Case 2: both objects have an island
			else if ( pxInstanceA->GetIsland() && pxInstanceB->GetIsland() )
			{
				// 2a if the islands are the same, we've either handled them before
				// OR they're in the same island because of an island-merge. This
				// means we have to check the island to make sure.
				if ( pxInstanceA->GetIsland() == pxInstanceB->GetIsland() )
				{
					if ( pxPairing )
					{
						if ( !pxInstanceA->GetIsland()->IsPairingPresent( pxPairing ) )
						{
							pxInstanceA->GetIsland()->AddPairing( pxPairing );
						}
					}
				}
				else // 2b, we need to merge the islands
				{
					// merge island B into A... 
					pxInstanceA->GetIsland()->MergeIsland( pxInstanceB->GetIsland() );

					// and then add this new pairing to island A
					pxInstanceA->GetIsland()->AddPairing( pxPairing );

					// well, actually both instances should have the same island now - let's assert that
					Asura_Assert( pxInstanceA->GetIsland() == pxInstanceB->GetIsland(), "This should be the same island" );
				}
			}
			else // Case 3: One of the objects has an island
			{
				if ( pxInstanceA->GetIsland()  )
				{
					// Okay, A has an island already. It should be a dynamic object.
					Asura_Physics_Island* pxIsland = pxInstanceA->GetIsland();

					Asura_Assert( pxInstanceA->IsDynamic(), "Only dynamic objects should have an island" );
					Asura_Assert( !pxInstanceB->GetIsland(), "B shouldn't have an island" );

					// If B is dynamic (and logically at this stage without an island), then we need to add B to A's island
					if ( pxInstanceB->IsDynamic() )
					{
						pxIsland->AddObject( pxInstanceB );
						
						if ( pxPairing ) pxIsland->AddPairing( pxPairing );
					}
					else
					{
						// Otherwise we have a dynamic / non-dynamic match, where the dynamic object already has an island.
						// We *may* have seen this before, or it could be a new pairing.
						if ( pxPairing ) pxIsland->AddPairingIfNotAlreadyPresent( pxPairing );
					}
				}
				else // if ( pxInstanceB->m_pxIsland  )
				{
					// Okay, B has an island already. It should be a dynamic object.
					Asura_Physics_Island* pxIsland = pxInstanceB->GetIsland();

					Asura_Assert( pxInstanceB->IsDynamic(), "Only dynamic objects should have an island" );
					Asura_Assert( !pxInstanceA->GetIsland(), "A shouldn't have an island" );

					// If A is dynamic (and logically at this stage without an island), then we need to add A to B's island
					if ( pxInstanceA->IsDynamic() )
					{
						pxIsland->AddObject( pxInstanceA );
						if ( pxPairing ) pxIsland->AddPairing( pxPairing );
					}
					else
					{
						// Otherwise we have a dynamic / non-dynamic match, where the dynamic object already has an island.
						// We *may* have seen this before, or it could be a new pairing.
						if ( pxPairing ) pxIsland->AddPairingIfNotAlreadyPresent( pxPairing );
					}
				}

			}
		}

		void MergeIsland( Asura_Physics_Island* pxOtherIsland )
		{
			// KJFL - 05 July 2006 - 
			// We want to take all the pairings and objects from pxOtherIsland and move
			// them to "this" island. Then leave pxOtherIsland empty.
			//
			// TODO:
			// In theory we should destroy the other island rather than leaving it blank, 
			// e.g. put it on a use-next free queue or something.

			// Go through the other island's objects and tell them that they now belong to us.
			for( Asura_Physics_Island_ObjectNode* pxNode = pxOtherIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
			{
				pxNode->m_pxObject->SetIsland( this );
			}

			// Add all the objects from pxOtherIsland to "this" island by 
			// tacking their objectnode linked list onto the end of ours.
			{
				if ( m_pxObjectNode ) // do we actually have any nodes?
				{
					Asura_Physics_Island_ObjectNode* pxNode = m_pxObjectNode;

					// skip to the last one
					while ( pxNode->m_pxNextObjectNode != 0 )
					{
						pxNode = pxNode->m_pxNextObjectNode;
					}

					// chain the other islands object nodes to this node
					pxNode->m_pxNextObjectNode = pxOtherIsland->m_pxObjectNode;
				}
				else // we have no nodes; just set our root pointer to the other island's
				{
					m_pxObjectNode = pxOtherIsland->m_pxObjectNode;
				}
			
				// and blank the other islands memory of them
				pxOtherIsland->m_pxObjectNode = 0;
			}
			
			// and do the same for the pairings
			{
				if ( m_pxPairingNode ) // do we actually have any nodes?
				{
					Asura_Physics_Island_PairingNode* pxNode = m_pxPairingNode;

					// skip to the last one
					while ( pxNode->m_pxNextPairingNode != 0 )
					{
						pxNode = pxNode->m_pxNextPairingNode;
					}

					// chain the other islands pairing nodes to this node
					pxNode->m_pxNextPairingNode = pxOtherIsland->m_pxPairingNode;
				}
				else // we have no nodes; just set our root pointer to the other island's
				{
					m_pxPairingNode = pxOtherIsland->m_pxPairingNode;
				}
				
				// and blank the other islands memory of them
				pxOtherIsland->m_pxPairingNode = 0;
			}
		}

		void AddObject( Asura_Physics_RigidBodyInstance* pxObject )
		{
			// add the object to this island's list of objects
			Asura_Physics_Island_ObjectNode* pxNode = reinterpret_cast<Asura_Physics_Island_ObjectNode*>( s_xMemory.AllocateBytes( sizeof( Asura_Physics_Island_ObjectNode ) ) );

			if ( pxNode )
			{
				pxNode->m_pxNextObjectNode = this->m_pxObjectNode;
				pxNode->m_pxObject = pxObject;

				this->m_pxObjectNode = pxNode;
			}

			// make the object know it's in this island
			pxObject->SetIsland( this );
		}
		
		bool IsPairingPresent( const Asura_Physics_Pairing* pxPairing ) const
		{
			Asura_Physics_Island_PairingNode* pxNode = m_pxPairingNode;

			while ( pxNode )
			{
				if ( pxNode->m_pxPairing == pxPairing ) return true;
				pxNode = pxNode->m_pxNextPairingNode;
			}

			return false;
		}

		void AddPairing( Asura_Physics_Pairing* pxPairing )
		{
			Asura_Physics_Island_PairingNode* pxNode = reinterpret_cast<Asura_Physics_Island_PairingNode*>( s_xMemory.AllocateBytes( sizeof( Asura_Physics_Island_PairingNode ) ) );

			if ( pxNode )
			{
				pxNode->m_pxNextPairingNode = this->m_pxPairingNode;
				pxNode->m_pxPairing = pxPairing;

				this->m_pxPairingNode = pxNode;
			}
		}

		void AddPairingIfNotAlreadyPresent( Asura_Physics_Pairing* pxPairing )
		{
			// KJFL - 02 February 2006 - this is the only code here that needs to read back from the island 
			// with a loop - and as such I don't like it! I can get rid of it if I allow non-dynamic objects
			// to remember what island they are in - but we'd most definitely need to cope with merging islands 
			// in that case, and we'd get things in islands which aren't necessary. E.g. is dynamic objects A & B
			// both hit static object C, but A & B aren't colliding, then we can actually have two calculation 
			// islands AC and BC - we don't need ABC. More islands is good because we can parallelise the tasks
			// better.
			for( Asura_Physics_Island_PairingNode* pxNode = m_pxPairingNode; pxNode!=0; pxNode = pxNode->m_pxNextPairingNode )
			{
				if ( pxPairing == pxNode->m_pxPairing )
				{
					// already exists, let's exit
					return;
				}
			}

			// must be a new one - let's add it
			AddPairing( pxPairing );
		}

		Asura_Physics_Island* GetNextIsland( void )
		{
			return m_pxNextIsland;
		}


		Asura_Physics_Island_ObjectNode*	m_pxObjectNode;
		Asura_Physics_Island_PairingNode*	m_pxPairingNode;

	private:

		Asura_Physics_Island();
		~Asura_Physics_Island();

		Asura_Physics_Island*				m_pxNextIsland;

		static Asura_Physics_Island*		s_pxFirstIsland;
		static Asura_Physics_MemoryPool		s_xMemory;
};

#endif//__ASURA_PHYSICS_ISLAND_H__
