//
// Asura_Physics_Pairing.cpp
//
// Created by KJFL - 16 August 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_Pairing.h"

#include "Core/Asura_Timers.h"

#include "Physics/Asura_Physics_CCD.h"
#include "Physics/Asura_Physics_CollisionFX_Report.h"
#include "Physics/Asura_Physics_CollisionMesh.h"
#include "Physics/Asura_Physics_Island.h"
#include "Physics/Asura_Physics_RigidBody.h"
#include "Physics/Asura_Physics_RigidBodyDefinition.h"

#include "Physics/Asura_Physics.h" // only for tri test

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
#include "Physics/Asura_Physics_Solver.h"
#endif

static const float fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION = 0.001f; // 1mm

u_int													Asura_Physics_Pairing::s_uNumberOfActivePairings;
Asura_MempoolHashListTemplate<Asura_Physics_Pairing>	Asura_Physics_Pairing::s_xPairingsList(16);

// Returns a non-null pointer if the pairing is a valid one - 
// which means that it either existed already, or has just been created.
Asura_Physics_Pairing* Asura_Physics_Pairing::GetOrCreatePairing( Asura_Physics_RigidBodyInstance* pxInstanceA, Asura_Physics_RigidBodyInstance* pxInstanceB, bool bAssignIslands /*= true*/ )
{
	// check for simple ignore flags, and whether one object has been told to ignore the other explicitly
	if ( pxInstanceA->IsIgnore() ) return 0;
	if ( pxInstanceA->m_iHandleToIgnore == pxInstanceB->GetHandle() ) return 0;
	
	if ( pxInstanceB->IsIgnore() ) return 0;
	if ( pxInstanceB->m_iHandleToIgnore == pxInstanceA->GetHandle() ) return 0;

	if ( pxInstanceA->CollisionGroupsAllowCollision( pxInstanceB ) == false ) return 0;

	// do a closer BB test - the spatial hash is by nature overly inclusive
	if(! (pxInstanceA->m_xBoundingBox.Intersects(pxInstanceB->m_xBoundingBox)) )
	{
		return 0;
	}

	Asura_Hash_ID uHash = (pxInstanceA->GetHandle()) + (pxInstanceB->GetHandle()<<16);

	// The AddNode call below will return a ptr to a new node (setting bNewNodeCreated to true)
	// or a ptr to an existing node (setting bNewNodeCreated to false). It will only return
	// a null ptr if the system is out of memory.
	bool bNewNodeCreated = false;
	Asura_Physics_Pairing* pxPairing = s_xPairingsList.AddNode( uHash, bNewNodeCreated );

	// Are we out of memory?
	if ( !pxPairing ) return false;

	pxPairing->m_uLastUpdateStamp = Asura_Physics_System::GetPhysicsUpdateCount();

	if ( bAssignIslands )
	{
		Asura_Physics_Island::AssignSameIsland( pxInstanceA, pxInstanceB, pxPairing );
	}

	// if a new pairing was created, we need to initialise its values
	if ( bNewNodeCreated )
	{
		++s_uNumberOfActivePairings;

		pxPairing->m_pxInstanceA = pxInstanceA;

		pxPairing->m_pxInstanceB = pxInstanceB;

		pxPairing->m_uFXReportStamp = 0xffffffff;
	}
	else
	{
		// this should be a valid pairing with the pointers already set - let's just check
		Asura_Assert( pxPairing->m_pxInstanceA == pxInstanceA, "instance mismatch" );
		Asura_Assert( pxPairing->m_pxInstanceB == pxInstanceB, "instance mismatch" );
	}

	// Recalc the mass coefficient, even if the pairing already exists - this is to cope with people 
	// changing the mass of objects mid-collision
	const float fTotalInvMass = pxInstanceA->GetInvMass() + pxInstanceB->GetInvMass();

	if ( fTotalInvMass > 0.0f )
	{
		pxPairing->m_fMassCoefficient = 1.0f / fTotalInvMass;
	}
	else
	{
		// we can't resolve collision between two immovable objects!
		Asura_Assert( fTotalInvMass > 0.0f, "Two objects with infinite mass!" );
		
		pxPairing->m_fMassCoefficient = 0.0f;
	}

	return pxPairing;
}



void Asura_Physics_Pairing::RemoveBrokenParings(void)
{
	s_xPairingsList.RemoveBrokenParings();
}

// return true if pairing should be destroyed
bool Asura_Physics_Pairing::Update(void)
{
	// Try and find if the object in this pairing are separate.

	if ( m_pxInstanceA->IsAsleep() && m_pxInstanceB->IsAsleep() ) return false;

	// If one or both of the objects is suddenly flagged as ignore, destroy the pairing.
	// Static objects which are 'destroyed' may just set the physics object to be ignored,
	// or actors moving without physics (e.g. doing an animation of a vault) may similarly
	// want any existing pairings (and their contact points) to be destroyed.
	if ( m_pxInstanceA->IsIgnore() || m_pxInstanceB->IsIgnore() ) return true;


	if ( ! (m_pxInstanceA->m_xBoundingBox.Intersects(m_pxInstanceB->m_xBoundingBox)) )
	{
		// not intersecting; destroy
		return true;
	}

	/*
	// if we're not handling friction then we should just destroy all contacts
	{
		Asura_Physics_Pairing_Contact* pxContact = m_pxContactsList;

		pxContact = m_pxContactsList;
		while (pxContact)
		{
			Asura_Physics_Pairing_Contact* pxNext = pxContact->m_pxNext;

			pxContact->SetBrokenStatus( true );

			pxContact = pxNext;
		}

		DestroyBrokenContacts();
	}
	*/

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
	if (1)
	{
		if ( m_pxInstanceA->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_POLYGONAL
			 && m_pxInstanceA->GetDefinition()->GetCollisionMesh() )
		{
			return !FindContactPoints_MPR_Concave( m_pxInstanceA, m_pxInstanceB );
		}
		else if ( m_pxInstanceB->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_POLYGONAL 
				  && m_pxInstanceB->GetDefinition()->GetCollisionMesh() )
		{
			return !FindContactPoints_MPR_Concave( m_pxInstanceB, m_pxInstanceA );
		}
		else if ( m_pxInstanceA->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_OBB 
				&& m_pxInstanceB->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_OBB 
				&& !m_pxInstanceA->IsActor()
				&& !m_pxInstanceB->IsActor() )
		{
			return !FindContactPoints_SweptOBBs();
		}
		else return !FindContactPoints_MPR();
		
	}
#endif

#if 01
	if ( m_pxInstanceA->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_POLYGONAL
		 && m_pxInstanceA->GetDefinition()->GetCollisionMesh() )
	{
#if 01
		if ( m_pxInstanceB->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE )
		{
			return !FindContactPoints_SweptSphereVsSweptPolygons( m_pxInstanceB, m_pxInstanceA );
		}
		else
#endif
		{
			return !FindContactPoints_SweptOBBVsSweptPolygons( m_pxInstanceB, m_pxInstanceA );
		}
	}
	else if ( m_pxInstanceB->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_POLYGONAL 
			  && m_pxInstanceB->GetDefinition()->GetCollisionMesh() )
	{
#if 01
		if ( m_pxInstanceA->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE )
		{
			return !FindContactPoints_SweptSphereVsSweptPolygons( m_pxInstanceA, m_pxInstanceB );
		}
		else
#endif
		{
			return !FindContactPoints_SweptOBBVsSweptPolygons( m_pxInstanceA, m_pxInstanceB );
		}
	}
#if 0
	else if (!FindContactPoints_SweptOBBs())
	{
		return true;
	}

	//DestroyBrokenContacts();


	if (m_pxInstanceB->IsWideAwake())
	{
		m_pxInstanceA->m_bHasAwakePartner = true;
	}
	if (m_pxInstanceA->IsWideAwake())
	{
		m_pxInstanceB->m_bHasAwakePartner = true;
	}

	return false;
#else
	else
	{
		if ( m_pxInstanceA->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE )
		{
			return !FindContactPoints_SweptOBBVsSweptSphere( m_pxInstanceB, m_pxInstanceA );
		}
		else if ( m_pxInstanceB->GetCollisionProxy() == Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE )
		{
			return !FindContactPoints_SweptOBBVsSweptSphere( m_pxInstanceA, m_pxInstanceB );
		}

		return !FindContactPoints_SweptOBBs();
	}
#endif
#endif
}


void Asura_Physics_Pairing::OnDeletion( bool bWakeObjects /* = false */ )
{
	--s_uNumberOfActivePairings;

	if (bWakeObjects)
	{
		m_pxInstanceA->WakeUp();
		m_pxInstanceB->WakeUp();
	}

	//Asura_Physics_Pairing_Contact* pxContact = m_pxContactsList;

	//while (pxContact)
	//{
	//	Asura_Physics_Pairing_Contact* pxNext = pxContact->m_pxNext;

	//	m_pxInstanceA->RemoveContactVertex( &(pxContact->m_axContactVerticesForA ) );
	//	
	//	m_pxInstanceB->RemoveContactVertex( &(pxContact->m_axContactVerticesForB ) );
	//	
	//	Asura_Physics_Pairing_Contact::Destroy( pxContact );
	//	
	//	pxContact = pxNext;
	//}
}

bool Asura_Physics_Pairing::UsesInstance( const Asura_Physics_RigidBodyInstance* pxInstance ) const 
{
	return (pxInstance == m_pxInstanceA || pxInstance == m_pxInstanceB);
}

class PairingVelocityConstraint
{
public:

	Asura_AVector3 m_xRelA;
	Asura_AVector3 m_xRelB;

	float m_fRequiredNormalSpeed;
	float m_fInvNormalDenom;
};


void Asura_Physics_Pairing::CalculateCollisionResponses( u_int uNumberOfContacts, const Asura_AVector3* pxContactsOnA, const Asura_AVector3* pxContactsOnB, const Asura_AVector3& xAxis  )
{
#if 1
	if ( m_pxInstanceB->IsActor() && m_pxInstanceA->IsDynamic() && !m_pxInstanceA->IsActor() && m_pxInstanceA->GetActorsDontPushMe() )
	{
		CalculateCollisionResponses( m_pxInstanceA, m_pxInstanceB, uNumberOfContacts, pxContactsOnA, xAxis );
		return;
	} 
	else if ( m_pxInstanceA->IsActor() && m_pxInstanceB->IsDynamic() && !m_pxInstanceB->IsActor() && m_pxInstanceB->GetActorsDontPushMe() )
	{
		CalculateCollisionResponses( m_pxInstanceB, m_pxInstanceA, uNumberOfContacts, pxContactsOnB, -xAxis );
		return;
	}
#endif
	PairingVelocityConstraint axContact[10];

	Asura_AVector3 xCentreA;
	Asura_AVector3 xCentreB;

	xCentreA.Zero();
	xCentreB.Zero();
	
	float fNumVerts=0.0f;

	const Asura_AMatrix3x3 xInvInertiaTensorA = m_pxInstanceA->GetInvInertiaTensorDiagonal();
	const Asura_AMatrix3x3 xInvInertiaTensorB = m_pxInstanceB->GetInvInertiaTensorDiagonal();

	// TODO - how to combine elasticity? (sqrt of product?)
	float fElasticity = Asura_Maths::Sqrt( m_pxInstanceA->GetElasticity() * m_pxInstanceB->GetElasticity() );
	float fFriction = Asura_Maths::Sqrt( m_pxInstanceA->GetFriction() * m_pxInstanceB->GetFriction() );

	fFriction = Asura_Maths::Min( 1.0f, fFriction );
	float fFrictionNormalImpulse = 0;

	for (u_int u=0; u<uNumberOfContacts; ++u)
	{
		PairingVelocityConstraint& xContact = axContact[u];

		xContact.m_xRelA = pxContactsOnA[u] - m_pxInstanceA->m_xBodyState.m_xCentreOfMass;
		xContact.m_xRelB = pxContactsOnB[u] - m_pxInstanceB->m_xBodyState.m_xCentreOfMass;

		{
			Asura_AVector3 xDenomA = xContact.m_xRelA^xAxis;
			Asura_AVector3 xDenomB = xContact.m_xRelB^xAxis;

			xDenomA = xDenomA*xInvInertiaTensorA;
			xDenomB = xDenomB*xInvInertiaTensorB;

			Asura_AVector3 xDenom = (xDenomA ^ xContact.m_xRelA) + (xDenomB ^ xContact.m_xRelB);

			xContact.m_fInvNormalDenom = m_pxInstanceA->GetInvMass() + m_pxInstanceB->GetInvMass() + ( xDenom * xAxis );

			if ( xContact.m_fInvNormalDenom != 0.0f )
			{
				xContact.m_fInvNormalDenom = 1.0f / xContact.m_fInvNormalDenom;
			}
		}


//		Asura_AVector3 xRelVelocity = m_pxInstanceA->CalculateVelocityOfPointOnBody( pxContactsOnA[u] ) - m_pxInstanceB->CalculateVelocityOfPointOnBody( pxContactsOnB[u] );

		Asura_AVector3 xRelVelocity;
		{
			Asura_AVector3 xVelB;
			
			m_pxInstanceA->CalculateVelocityOfRelativePointOnBody( xContact.m_xRelA, xRelVelocity  );
			m_pxInstanceB->CalculateVelocityOfRelativePointOnBody( xContact.m_xRelB, xVelB  );

			xRelVelocity -= xVelB;
		}

		float fNormalVelocity = xRelVelocity * xAxis;

		{

			if (fNormalVelocity < 0.0f )
			{
				float fLocalElasticity = fElasticity;
			//	if (fNormalVelocity > -0.01f) fLocalElasticity = 0.0f;

				xContact.m_fRequiredNormalSpeed = -fNormalVelocity * fLocalElasticity;

				fFrictionNormalImpulse += -fNormalVelocity *(1.0f + fLocalElasticity ) * xContact.m_fInvNormalDenom;

				xCentreA += pxContactsOnA[u];
				xCentreB += pxContactsOnB[u];
				fNumVerts+=1.0f;
			}
			else
			{
				xContact.m_fRequiredNormalSpeed = fNormalVelocity;
			}
		}
	}

	if (fNumVerts == 0.0f) return;

	#ifdef ASURA_USE_COLLISIONFX
	if (ShouldAddCollisionFXReport())
	{
		Asura_AVector3 xCentre = xCentreA + xCentreB;
		xCentre /= fNumVerts*2.0f;

		AddCollisionFXReport( xCentre, xAxis );
	}
	#endif//#ifdef ASURA_USE_COLLISIONFX


	const u_int uPAIRING_ITERATIONS = 10;
	for (u_int uIterations = 0; uIterations<uPAIRING_ITERATIONS; ++uIterations)
	{
		bool bSatisfied = true;
		for (u_int u=0; u<uNumberOfContacts; ++u)
		{
			PairingVelocityConstraint& xContact = axContact[u];

			if (xContact.m_fInvNormalDenom == 0.0f) continue;

//			Asura_AVector3 xRelVelocity = m_pxInstanceA->CalculateVelocityOfPointOnBody( pxContactsOnA[u] ) - m_pxInstanceB->CalculateVelocityOfPointOnBody( pxContactsOnB[u] );

			Asura_AVector3 xRelVelocity;
			{
				Asura_AVector3 xVelB;

				m_pxInstanceA->CalculateVelocityOfRelativePointOnBody( xContact.m_xRelA, xRelVelocity  );
				m_pxInstanceB->CalculateVelocityOfRelativePointOnBody( xContact.m_xRelB, xVelB  );

				xRelVelocity -= xVelB;
			}
			float fNormalVelocity = xRelVelocity * xAxis;

			{
				float fNumerator = xContact.m_fRequiredNormalSpeed - fNormalVelocity;

		//		if ( Asura_Maths::Abs( fNumerator ) > 0.00001f )
				{
					bSatisfied = false;

					float fNormalImpulse = fNumerator * xContact.m_fInvNormalDenom;

					Asura_AVector3 xImpulse = xAxis * fNormalImpulse;

					m_pxInstanceA->ApplyImpulseAtRelativePosition( xContact.m_xRelA, xImpulse );
					
					xImpulse = -xImpulse;

					m_pxInstanceB->ApplyImpulseAtRelativePosition( xContact.m_xRelB, xImpulse );
				}
			}
		}

		if (bSatisfied) break;
	}
#if 1
#if 1
	if (m_pxInstanceA->m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceA->m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_pxInstanceA->m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceA->m_xBodyState.m_xLinearVelocity.Zero();
	}

	if (m_pxInstanceB->m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceB->m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_pxInstanceB->m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceB->m_xBodyState.m_xLinearVelocity.Zero();
	}
#endif
	

	if (fNumVerts!=0.0f )
	{
		xCentreA /= fNumVerts;
		xCentreB /= fNumVerts;

		const Asura_AVector3 xRelA = xCentreA - m_pxInstanceA->m_xBodyState.m_xCentreOfMass;
		const Asura_AVector3 xRelB = xCentreB - m_pxInstanceB->m_xBodyState.m_xCentreOfMass;

		Asura_AVector3 xRelVelocity;
		{
			Asura_AVector3 xVelB;

			m_pxInstanceA->CalculateVelocityOfRelativePointOnBody( xRelA, xRelVelocity  );
			m_pxInstanceB->CalculateVelocityOfRelativePointOnBody( xRelB, xVelB  );

			xRelVelocity -= xVelB;
		}


		float fNormalVelocity = xRelVelocity * xAxis;

		Asura_AVector3 xTangent = (xRelVelocity - xAxis * ( fNormalVelocity ));
		float fTangentialSpeed = xTangent.NormaliseReturningMagnitude();

		//if (fTangentialSpeed > 0.0001f)
		{
			
			Asura_AVector3 xDenomA = xRelA^xTangent;
			Asura_AVector3 xDenomB = xRelB^xTangent;

			xDenomA = xDenomA * xInvInertiaTensorA;
			xDenomB = xDenomB * xInvInertiaTensorB;

			Asura_AVector3 xDenom = (xDenomA ^ xRelA) + (xDenomB ^ xRelB);

			float fDenominator  = m_pxInstanceA->GetInvMass() + m_pxInstanceB->GetInvMass() + xDenom * xTangent;

			float fFrictionImpulse = fTangentialSpeed / fDenominator;

			fFrictionImpulse = Asura_Maths::Min( fFrictionImpulse, Asura_Maths::Abs(fFrictionNormalImpulse) * fFriction );

			Asura_AVector3 xImpulse = - xTangent * fFrictionImpulse;

			m_pxInstanceA->ApplyImpulseAtRelativePosition( xRelA, xImpulse );

			xImpulse = -xImpulse;

			m_pxInstanceB->ApplyImpulseAtRelativePosition( xRelB, xImpulse );
			
			m_pxInstanceA->m_xBodyState.m_xAngularVelocity = m_pxInstanceA->m_xBodyState.m_xAngularVelocity - xAxis * (m_pxInstanceA->m_xBodyState.m_xAngularVelocity*xAxis*fFriction*0.05f);
			m_pxInstanceB->m_xBodyState.m_xAngularVelocity = m_pxInstanceB->m_xBodyState.m_xAngularVelocity - xAxis * (m_pxInstanceB->m_xBodyState.m_xAngularVelocity*xAxis*fFriction*0.05f);
		}
		//	SetVelocitiesOfBody( m_xBodyState.m_xLinearVelocity, m_xBodyState.m_xAngularVelocity );
	}

/*
	if (m_pxInstanceA->m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceA->m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_pxInstanceA->m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceA->m_xBodyState.m_xLinearVelocity.Zero();
	}

	if (m_pxInstanceB->m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceB->m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_pxInstanceB->m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_pxInstanceB->m_xBodyState.m_xLinearVelocity.Zero();
	}
*/
#endif

}

void Asura_Physics_Pairing::CalculateCollisionResponses( Asura_Physics_RigidBodyInstance* m_pxDynamicInstance, Asura_Physics_RigidBodyInstance* m_pxStaticInstance, u_int uNumberOfContacts, const Asura_AVector3* pxContactsOnA, const Asura_AVector3& xAxis  )
{
	PairingVelocityConstraint axContact[10];

	Asura_AVector3 xCentreA;
	Asura_AVector3 xCentreB;

	xCentreA.Zero();
	xCentreB.Zero();

	float fNumVerts=0.0f;

	const Asura_AMatrix3x3 xInvInertiaTensorA = m_pxDynamicInstance->GetInvInertiaTensorDiagonal();
	
	// TODO - how to combine elasticity? (sqrt of product?)
	float fElasticity = Asura_Maths::Sqrt( m_pxDynamicInstance->GetElasticity() * m_pxStaticInstance->GetElasticity() );
	float fFriction = Asura_Maths::Sqrt( m_pxDynamicInstance->GetFriction() * m_pxStaticInstance->GetFriction() );

	fFriction = Asura_Maths::Min( 1.0f, fFriction );
	float fFrictionNormalImpulse = 0;

	for (u_int u=0; u<uNumberOfContacts; ++u)
	{
		PairingVelocityConstraint& xContact = axContact[u];

		xContact.m_xRelA = pxContactsOnA[u] - m_pxDynamicInstance->m_xBodyState.m_xCentreOfMass;
	
		{
			Asura_AVector3 xDenomA = xContact.m_xRelA^xAxis;

			xDenomA = xDenomA*xInvInertiaTensorA;
	
			Asura_AVector3 xDenom = (xDenomA ^ xContact.m_xRelA);

			xContact.m_fInvNormalDenom = m_pxDynamicInstance->GetInvMass() + ( xDenom * xAxis );

			if ( xContact.m_fInvNormalDenom != 0.0f )
			{
				xContact.m_fInvNormalDenom = 1.0f / xContact.m_fInvNormalDenom;
			}
		}


		Asura_AVector3 xRelVelocity;
	
		m_pxDynamicInstance->CalculateVelocityOfRelativePointOnBody( xContact.m_xRelA, xRelVelocity  );


		float fNormalVelocity = xRelVelocity * xAxis;

		{

			if (fNormalVelocity < 0.0f )
			{
				float fLocalElasticity = fElasticity;
				//	if (fNormalVelocity > -0.01f) fLocalElasticity = 0.0f;

				xContact.m_fRequiredNormalSpeed = -fNormalVelocity * fLocalElasticity;

				fFrictionNormalImpulse += -fNormalVelocity *(1.0f + fLocalElasticity ) * xContact.m_fInvNormalDenom;

				xCentreA += pxContactsOnA[u];
				fNumVerts+=1.0f;
			}
			else
			{
				xContact.m_fRequiredNormalSpeed = fNormalVelocity;
			}
		}
	}

	if (fNumVerts == 0.0f) return;

	#ifdef ASURA_USE_COLLISIONFX
	if (ShouldAddCollisionFXReport())
	{
		Asura_AVector3 xCentre = xCentreA;
		xCentre /= fNumVerts;

		AddCollisionFXReport( xCentre, xAxis );
	}
	#endif//#ifdef ASURA_USE_COLLISIONFX

	const u_int uPAIRING_ITERATIONS = 10;
	for (u_int uIterations = 0; uIterations<uPAIRING_ITERATIONS; ++uIterations)
	{
		bool bSatisfied = true;
		for (u_int u=0; u<uNumberOfContacts; ++u)
		{
			PairingVelocityConstraint& xContact = axContact[u];

			if (xContact.m_fInvNormalDenom == 0.0f) continue;

			//			Asura_AVector3 xRelVelocity = m_pxDynamicInstance->CalculateVelocityOfPointOnBody( pxContactsOnA[u] ) - m_pxStaticInstance->CalculateVelocityOfPointOnBody( pxContactsOnB[u] );

			Asura_AVector3 xRelVelocity;
			m_pxDynamicInstance->CalculateVelocityOfRelativePointOnBody( xContact.m_xRelA, xRelVelocity  );

			float fNormalVelocity = xRelVelocity * xAxis;

			{
				float fNumerator = xContact.m_fRequiredNormalSpeed - fNormalVelocity;

				//		if ( Asura_Maths::Abs( fNumerator ) > 0.00001f )
				{
					bSatisfied = false;

					float fNormalImpulse = fNumerator * xContact.m_fInvNormalDenom;

					Asura_AVector3 xImpulse = xAxis * fNormalImpulse;

					m_pxDynamicInstance->ApplyImpulseAtRelativePosition( xContact.m_xRelA, xImpulse );

				
				}
			}
		}

		if (bSatisfied) break;
	}
#if 1
#if 1
	if (m_pxDynamicInstance->m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_pxDynamicInstance->m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_pxDynamicInstance->m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_pxDynamicInstance->m_xBodyState.m_xLinearVelocity.Zero();
	}

	
#endif


	if (fNumVerts!=0.0f )
	{
		xCentreA /= fNumVerts;
	
		const Asura_AVector3 xRelA = xCentreA - m_pxDynamicInstance->m_xBodyState.m_xCentreOfMass;
	
		Asura_AVector3 xRelVelocity;

		m_pxDynamicInstance->CalculateVelocityOfRelativePointOnBody( xRelA, xRelVelocity  );

		float fNormalVelocity = xRelVelocity * xAxis;

		Asura_AVector3 xTangent = (xRelVelocity - xAxis * ( fNormalVelocity ));
		float fTangentialSpeed = xTangent.NormaliseReturningMagnitude();

		//if (fTangentialSpeed > 0.0001f)
		{

			Asura_AVector3 xDenomA = xRelA^xTangent;
	
			xDenomA = xDenomA * xInvInertiaTensorA;
	
			Asura_AVector3 xDenom = (xDenomA ^ xRelA);

			float fDenominator  = m_pxDynamicInstance->GetInvMass() + xDenom * xTangent;

			float fFrictionImpulse = fTangentialSpeed / fDenominator;

			fFrictionImpulse = Asura_Maths::Min( fFrictionImpulse, Asura_Maths::Abs(fFrictionNormalImpulse) * fFriction );

			Asura_AVector3 xImpulse = - xTangent * fFrictionImpulse;

			m_pxDynamicInstance->ApplyImpulseAtRelativePosition( xRelA, xImpulse );

		
			m_pxDynamicInstance->m_xBodyState.m_xAngularVelocity = m_pxDynamicInstance->m_xBodyState.m_xAngularVelocity - xAxis * (m_pxDynamicInstance->m_xBodyState.m_xAngularVelocity*xAxis*fFriction*0.05f);
		}
	}


#endif

}
bool Asura_Physics_Pairing::FindContactPoints_SweptOBBs2( void )
{
	/*
	const Asura_AMatrix3x3& xObjectAOrient = m_pxInstanceA->GetOrientation();
	const Asura_AMatrix3x3& xObjectBOrient = m_pxInstanceB->GetOrientation();

	const Asura_AVector3 xObjectAHalfExtents = m_pxInstanceA->GetHalfExtents();

	const Asura_AVector3 xObjectBHalfExtents = m_pxInstanceB->GetHalfExtents();

	const Asura_AVector3& xObjectAStartPosition = m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass;
	const Asura_AVector3& xObjectBStartPosition = m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass;

	Asura_OBB xObjectAOBB( xObjectAHalfExtents, xObjectAStartPosition, xObjectAOrient );

	Asura_OBB xObjectBOBB( xObjectBHalfExtents, xObjectBStartPosition, xObjectBOrient );

	bool bPenetration = xObjectAOBB.Intersects( xObjectBOBB );

	Asura_AVector3 xAxis;

	if (bPenetration)
	{
#if 01
		xObjectAOBB.SetCentre( m_pxInstanceA->GetCentreOfMass() );
		xObjectBOBB.SetCentre( m_pxInstanceB->GetCentreOfMass() );
#endif
		float fPenetration;
		bool bAgree = Asura_Physics_CCD::FindPenetration_OBBVsOBB( xObjectAOBB,xObjectBOBB, fPenetration, xAxis );

		Asura_Assert( bAgree, "Didn't find penetration" );
		// Ok, here's a scary bit. First we need to find the support features for both objects, and then we
		// need to project the support features and clip them to find the overlapping contact feature,
		// which gives us our contact points. Worst case is two faces whose common overlap is an octagon.

		if (!bAgree ) return false;
#if 0		
		if ( m_pxInstanceA->IsDynamic())
		{
			m_pxInstanceA->m_xBodyState.m_xCentreOfMass = m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass;
		}
		if ( m_pxInstanceB->IsDynamic() )
		{
			m_pxInstanceB->m_xBodyState.m_xCentreOfMass = m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass;
		}
#endif
		Asura_AVector3 axContactPointsForA[10];
		Asura_AVector3 axContactPointsForB[10];
		u_int uNumberOfContactPoints;


		uNumberOfContactPoints = Asura_Physics_CCD::FindContactPointsFromSupportFeatures( axContactPointsForA, axContactPointsForB, m_pxInstanceA, m_pxInstanceB, xAxis );

		if ( uNumberOfContactPoints )
		{
			CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForA, axContactPointsForB, xAxis );
		}

		NotifyObjectsOfCollision( xAxis );
#if 01
		if ( m_pxInstanceA->IsDynamic() )
		{
			Asura_AVector3 xImpulse = xAxis * fPenetration * 0.5f;
			m_pxInstanceA->m_xBodyState.m_xCentreOfMass += + xImpulse;
		}
		if ( m_pxInstanceB->IsDynamic()  )
		{
			Asura_AVector3 xImpulse = -xAxis * fPenetration * 0.5f;
			m_pxInstanceB->m_xBodyState.m_xCentreOfMass += + xImpulse;
		}
#else
		if ( m_pxInstanceA->IsDynamic() )
		{
			Asura_AVector3 xImpulse = xAxis * fPenetration * 0.5f;
			m_pxInstanceA->m_xBodyState.m_xCentreOfMass = m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass + xImpulse;
		}
		if ( m_pxInstanceB->IsDynamic()  )
		{
			Asura_AVector3 xImpulse = -xAxis * fPenetration * 0.5f;
			m_pxInstanceB->m_xBodyState.m_xCentreOfMass = m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass + xImpulse;
		}
#endif

	}
	else
	{
		float fTimeToImpact = 1.0f;

		Asura_AVector3 xVelocityA = m_pxInstanceA->m_xBodyState.m_xCentreOfMass - m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass;

		Asura_AVector3 xVelocityB = m_pxInstanceB->m_xBodyState.m_xCentreOfMass - m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass;

		Asura_AVector3 xRelativeVelocity = xVelocityA - xVelocityB;

		bool bImpact = Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsOBB( xObjectAOBB, xObjectBOBB, xRelativeVelocity, fTimeToImpact, xAxis);

		if (!bImpact) return false;

		Asura_AVector3 xImpulseA;
		Asura_AVector3 xImpulseB;

		xImpulseA.Zero();
		xImpulseB.Zero();


		// Ok, here's a scary bit. First we need to find the support features for both objects, and then we
		// need to project the support features and clip them to find the overlapping contact feature,
		// which gives us our contact points. Worst case is two faces whose common overlap is an octagon.

		Asura_AVector3 axContactPointsForA[10];
		Asura_AVector3 axContactPointsForB[10];
		u_int uNumberOfContactPoints;


		uNumberOfContactPoints = Asura_Physics_CCD::FindContactPointsFromSupportFeatures( axContactPointsForA, axContactPointsForB, m_pxInstanceA, m_pxInstanceB, xAxis );

		if ( uNumberOfContactPoints )
		{
			CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForA, axContactPointsForB, xAxis );
		}

		//fTimeToImpact = Asura_Maths::Max( 0.0f, fTimeToImpact - 0.1f );
		if ( m_pxInstanceA->IsDynamic() && !m_pxInstanceA->IsActor())
		{
			m_pxInstanceA->m_xBodyState.m_xCentreOfMass = m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass + xVelocityA*fTimeToImpact;
		}
		if ( m_pxInstanceB->IsDynamic() && !m_pxInstanceB->IsActor() )
		{
			m_pxInstanceB->m_xBodyState.m_xCentreOfMass = m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass + xVelocityB*fTimeToImpact;
		}

		NotifyObjectsOfCollision( xAxis );

	}
*/
	return true;
}

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER

static bool IntersectionTest( const Asura_Physics_RigidBodyInstance* pxBody1, const Asura_Physics_RigidBodyInstance* pxBody2, Asura_AVector3& xAxis, float& fSeparation )
{
	Asura_Physics_MPR_Body xBody1( pxBody1 );
	Asura_Physics_MPR_Body xBody2( pxBody2 );
	return Asura_Physics_CCD::MPR_Test( xBody1, xBody2, xAxis, fSeparation );
}

static bool SweptIntersectionTest(  Asura_Physics_RigidBodyInstance* pxBody1,  Asura_Physics_RigidBodyInstance* pxBody2, Asura_AVector3& xAxis, float& fRewindTime, float& fPenetration )
{
	Asura_Physics_MPR_Body xBody1( pxBody1 );
	Asura_Physics_MPR_Body xBody2( pxBody2 );
	bool bIntersect = Asura_Physics_CCD::Swept_MPR_Test( xBody1, xBody2, xAxis, fRewindTime, fPenetration );

#if 1 

	return bIntersect;

#else // Some other tests used to check the results of the MPR test.

	if(0)
	{
		float fTimeToImpact = 0;

		const Asura_AMatrix3x3& xObjectAOrient = pxBody1->GetOrientation();
		const Asura_AMatrix3x3& xObjectBOrient = pxBody2->GetOrientation();

		const Asura_AVector3 xObjectAHalfExtents = pxBody1->GetHalfExtents();
		const Asura_AVector3 xObjectBHalfExtents = pxBody2->GetHalfExtents();

		const Asura_AVector3& xObjectAStartPosition = pxBody1->GetPreviousCentreOfMass();
		const Asura_AVector3& xObjectBStartPosition = pxBody2->GetPreviousCentreOfMass();

		Asura_OBB xObjectA( xObjectAHalfExtents, xObjectAStartPosition, xObjectAOrient );
		Asura_OBB xObjectB( xObjectBHalfExtents, xObjectBStartPosition, xObjectBOrient );

		Asura_AVector3 xVelocityA = pxBody1->GetCentreOfMass() - pxBody1->GetPreviousCentreOfMass();
		Asura_AVector3 xVelocityB = pxBody2->GetCentreOfMass() - pxBody2->GetPreviousCentreOfMass();
		Asura_AVector3 xRelativeVelocity = xVelocityA - xVelocityB;

			
		bool bIntersectionFromOBBOBB = Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsOBB( xObjectA, xObjectB, xRelativeVelocity, fTimeToImpact, xAxis, false );

		Asura_Assert( bIntersectionFromOBBOBB == bIntersect, "Algorithms disagree");

		fRewindTime = 1.0f - fTimeToImpact;
		
		return bIntersectionFromOBBOBB;
	}

	if (1)
	{

		Asura_AVector3 xVelocityA = pxBody1->GetCentreOfMass() - pxBody1->GetPreviousCentreOfMass();
		Asura_AVector3 xVelocityB = pxBody2->GetCentreOfMass() - pxBody2->GetPreviousCentreOfMass();
		Asura_AVector3 xRelativeVelocity = xVelocityB - xVelocityA;

		if (xRelativeVelocity.Magnitude()>0.001f)
		{
			float fLambda = 0.0f;
			bool bHit = false;
			{
				Asura_AVector3 xPosition1 = pxBody1->GetCentreOfMass();
				Asura_AVector3 xPosition2 = pxBody2->GetCentreOfMass();

				for ( ; fLambda<= 1.0f; fLambda += 0.001f )
				{
					pxBody1->GetCentreOfMass() = (xPosition1 - pxBody1->GetPreviousCentreOfMass()) * fLambda + pxBody1->GetPreviousCentreOfMass();
					pxBody2->GetCentreOfMass() = (xPosition2 - pxBody2->GetPreviousCentreOfMass()) * fLambda + pxBody2->GetPreviousCentreOfMass();

					Asura_AVector3 xLocalAxis;
					float fSep;
					if ( Asura_Physics_CCD::MPR_Test( pxBody1, pxBody2, xLocalAxis, fSep ) )
					{
						xAxis = xLocalAxis;
						bHit = true;
						break;
					}
				}

				pxBody1->GetCentreOfMass() = xPosition1;
				pxBody2->GetCentreOfMass() = xPosition2;
			}

			Asura_Assert(bHit == bIntersect, "Algorithms disagree" );

			float fRewindTime2 = Asura_Maths::Min( 1.0f, 1.001f - fLambda );
/*			
			if ( bIntersect )
			{
				Asura_Assertf(Asura_Maths::Abs(fRewindTime2 - fRewindTime) <0.002f, ("Time out: CA: %f, MPR: %f", fRewindTime2, fRewindTime ));
				bIntersect = Asura_Physics_CCD::Swept_MPR_Test( pxBody1, pxBody2, xAxis, fRewindTime, fPenetration );
			}
*/
			fRewindTime = fRewindTime2;
				
			return bHit;
		}

		return false;
		
	}
#endif
}
bool Asura_Physics_Pairing::FindContactPoints_MPR_Concave( Asura_Physics_RigidBodyInstance* pxPolygonInstance, Asura_Physics_RigidBodyInstance* pxPrimitiveInstance )
{
	Asura_AVector3 xVelocityA = pxPolygonInstance->m_xBodyState.m_xCentreOfMass - pxPolygonInstance->m_xBodyState.m_xPreviousCentreOfMass;
	Asura_AVector3 xVelocityB = pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass - pxPrimitiveInstance->m_xBodyState.m_xPreviousCentreOfMass;
	Asura_AVector3 xRelativeVelocity = xVelocityA - xVelocityB;

	float fMinimumSafeDistance = Asura_Maths::Min( pxPolygonInstance->GetDefinition()->GetMinimumRadius(), pxPrimitiveInstance->GetDefinition()->GetMinimumRadius() ) *0.9f;

	bool bDoSweptTest = (xRelativeVelocity.Magnitude() > fMinimumSafeDistance);

	bDoSweptTest &=false;

	Asura_Physics_Tri xImpactTri;

	Asura_AVector3 xAxis;
	//float fTime = 0.0f;
	float fPenetration = 0.0f;

	bool bHaveIntersection = false;

	Asura_Physics_MPR_Body xBody2( pxPrimitiveInstance );	
	

	if ( bDoSweptTest )
	{
		float fRewindTime = 0.0f;
		
		Asura_Physics_CollisionMesh_Iterator xTriIterator( pxPolygonInstance->GetDefinition()->GetCollisionMesh() );
		Asura_Physics_Tri xTri;
		for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
		{
			// TODO - rather than calculating the OBB and Tri in world space, perhaps we can use the Tri's local space?
			// we'd have to transform something (i.e. OBB or Tri) so perhaps it wouldn't save much
			xTri.Transform( pxPolygonInstance->m_xBodyState.m_xOrientation, pxPolygonInstance->m_xInstancePosition );
/*
			Asura_AVector3 xTriNormal;
			xTri.GetUnnormalisedNormal( xTriNormal );
			if ( xRelativeVelocity * xTriNormal > 0.0f )
			{
				continue;
			}
*/
			Asura_Physics_MPR_Body xBody1( &xTri );

			float fThisRewindTime;
			Asura_AVector3 xThisAxis;
			bool bHit = Asura_Physics_CCD::Swept_MPR_Test( xBody1, xBody2, xThisAxis, fThisRewindTime, fPenetration );

			// if this isn't a penetration-solving case, look for the time of impact
			if ( bHit )
			{
				if ( fThisRewindTime > fRewindTime )
				{
					// record the new candidate for the first time of impact
					fRewindTime = fThisRewindTime;
					xAxis = xThisAxis;
					bHaveIntersection = true;
	
					// TODO - need to report tri for contact forces?
					// ideally you'd want all tris of the same contact time, which is more tricky!
					xImpactTri = xTri;

					/*
					if (fRewindTime >= 0.99999f)
					{
						bDoSweptTest = false;
						break;
					}
					*/
				}
			}
		}
	
		if (bHaveIntersection)
		{
		//	if (xRelativeVelocity * xAxis < 0.0f)
			{
				if ( pxPolygonInstance->IsDynamic() && !pxPolygonInstance->IsActor())
				{
					pxPolygonInstance->m_xBodyState.m_xCentreOfMass += - xVelocityA * fRewindTime;
				}
				if ( pxPrimitiveInstance->IsDynamic() && !pxPrimitiveInstance->IsActor())
				{
					pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass += - xVelocityB * fRewindTime;
				}
				fPenetration = 0.001f;
			}
			{
				Asura_AVector3 axContactPointsForPrimitive[10];
				Asura_AVector3 axContactPointsForTri[10];
				u_int uNumberOfContactPoints;

				uNumberOfContactPoints = Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( axContactPointsForPrimitive, axContactPointsForTri, pxPrimitiveInstance, xImpactTri, -xAxis );

				for ( u_int u=0; u<uNumberOfContactPoints; ++u)
				{
					Asura_AVector3 xRelA = axContactPointsForTri[u] - pxPolygonInstance->m_xBodyState.m_xCentreOfMass;
					Asura_AVector3 xRelB = axContactPointsForPrimitive[u] - pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass;
			
					Asura_Physics_Solver::AddPairingContact( pxPolygonInstance, xRelA, pxPrimitiveInstance, xRelB,  xAxis, fPenetration );
				}

				#ifdef ASURA_USE_COLLISIONFX
				if (uNumberOfContactPoints && ShouldAddCollisionFXReport())
				{
					// Find the average of all the contact points on both bodies,
					// and report that as the CollisionFX position.
					Asura_AVector3 xCentre;
					xCentre.Zero();

					for ( u_int u=0; u<uNumberOfContactPoints; ++u)
					{
						xCentre += axContactPointsForTri[u];
						xCentre += axContactPointsForPrimitive[u];
					}

					xCentre /= (static_cast<float>(uNumberOfContactPoints*2));

					AddCollisionFXReport( xCentre, xAxis );
				}
				#endif//#ifdef ASURA_USE_COLLISIONFX
			}
		}
	}
	
	//if (!bDoSweptTest)
	{
		pxPolygonInstance->CalculateRenderingPosition();
		fPenetration = 0.0f;
	
		Asura_OBB xOBB( pxPrimitiveInstance->GetHalfExtents(), pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass, pxPrimitiveInstance->GetOrientation() );
	
		

		Asura_Physics_CollisionMesh_Iterator xTriIterator( pxPolygonInstance->GetDefinition()->GetCollisionMesh() );
		Asura_Physics_Tri xTri;
		for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
		{
			// TODO - rather than calculating the OBB and Tri in world space, perhaps we can use the Tri's local space?
			// we'd have to transform something (i.e. OBB or Tri) so perhaps it wouldn't save much
			xTri.Transform( pxPolygonInstance->m_xBodyState.m_xOrientation, pxPolygonInstance->m_xInstancePosition );
			
			//Asura_Physics_MPR_Body xBody1( &xTri );

			Asura_AVector3 xThisAxis;
			float fThisPenetration;
		//	bool bHit = Asura_Physics_CCD::MPR_Test( xBody1, xBody2, xThisAxis, fThisPenetration );
			//fThisPenetration = -fThisPenetration;

			// KJFL - 19 November 2008 - TODO we need a generic PrimitiveVsTri method we can call. There's an
			// issue about what we can precalculate (don't want to make an OBB each time etc). The MPR test
			// is in theory the way to go, but it needs to be more close fitting.
			bool bHit;
			switch( pxPrimitiveInstance->GetCollisionProxy() )
			{
				case Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERE:
				{
					Asura_AVector3 xSphereCentre( pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass );
					Asura_AVector3 xContactPoint;
					float fRadius = pxPrimitiveInstance->GetDefinition()->GetMaximumRadius();

					bHit = Asura_Physics_CCD::FindPenetration_SphereVsTri( xTri, xSphereCentre, fRadius, fThisPenetration, xThisAxis, xContactPoint ); 
					xThisAxis = -xThisAxis;
					break;
				}

				case Asura_Physics_RigidBodyInstance::COLLISION_PROXY_SPHERESTACK:
				{
					Asura_Physics_MPR_Body xBody1( &xTri );
					bHit = Asura_Physics_CCD::MPR_Test( xBody1, xBody2, xThisAxis, fThisPenetration );
					fThisPenetration = -fThisPenetration;
					break;
				}

				default:
				case Asura_Physics_RigidBodyInstance::COLLISION_PROXY_OBB:
				{
					bHit = Asura_Physics_CCD::DoesOBBIntersectTri( xOBB, xTri, fThisPenetration, xThisAxis );
					xThisAxis = -xThisAxis;
					break;
				}
			}
		
			
			if ( bHit )
			{
				Asura_AVector3 xTriNormal;
				xTri.GetUnnormalisedNormal(xTriNormal);

				// if the suggested penetration axis is facing away from the tri normal, ignore this poly
				// (this happens when you're actually on the wrong side of it, for example)
				if ( xAxis * xTriNormal > 0.0f )
				{
					//bHit = false;
				}
			}
			



			// if this isn't a penetration-solving case, look for the time of impact
			if ( bHit )
			{
				if ( fThisPenetration > fPenetration )
				{
					// record the new candidate for the first time of impact
					fPenetration = fThisPenetration;
					xAxis = xThisAxis;
					bHaveIntersection = true;
					xImpactTri = xTri;
				}

			
				Asura_AVector3 axContactPointsForPrimitive[10];
				Asura_AVector3 axContactPointsForTri[10];
				u_int uNumberOfContactPoints;

				uNumberOfContactPoints = Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( axContactPointsForPrimitive, axContactPointsForTri, pxPrimitiveInstance, xTri, -xThisAxis );
	
				for ( u_int u=0; u<uNumberOfContactPoints; ++u)
				{
					Asura_AVector3 xRelA = axContactPointsForTri[u] - pxPolygonInstance->m_xBodyState.m_xCentreOfMass;
					Asura_AVector3 xRelB = axContactPointsForPrimitive[u] - pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass;
			
					Asura_Physics_Solver::AddPairingContact( pxPolygonInstance, xRelA, pxPrimitiveInstance, xRelB,  xThisAxis, fThisPenetration );
				}

				#ifdef ASURA_USE_COLLISIONFX
				if (uNumberOfContactPoints && ShouldAddCollisionFXReport())
				{
					// Find the average of all the contact points on both bodies,
					// and report that as the CollisionFX position.
					Asura_AVector3 xCentre;
					xCentre.Zero();

					for ( u_int u=0; u<uNumberOfContactPoints; ++u)
					{
						xCentre += axContactPointsForTri[u];
						xCentre += axContactPointsForPrimitive[u];
					}

					xCentre /= (static_cast<float>(uNumberOfContactPoints*2));

					AddCollisionFXReport( xCentre, xAxis );
				}
				#endif//#ifdef ASURA_USE_COLLISIONFX

			}
		}

		if (!bHaveIntersection) return false;

		{
			const float fScaledDistance = ( fPenetration + fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION ) * m_fMassCoefficient;
			const float fDistanceA = Asura_Maths::Min( fScaledDistance * pxPolygonInstance->GetInvMass(), pxPolygonInstance->GetDefinition()->GetMinimumRadius() * 0.5f );
			const float fDistanceB = Asura_Maths::Min( fScaledDistance * pxPrimitiveInstance->GetInvMass(), pxPrimitiveInstance->GetDefinition()->GetMinimumRadius() * 0.5f );

			if ( pxPolygonInstance->IsActor())
			{
				pxPolygonInstance->m_xBodyState.m_xCentreOfMass += xAxis * fDistanceA;
				pxPolygonInstance->m_xInstancePosition += xAxis * fDistanceA;
			}
			if ( pxPrimitiveInstance->IsActor())
			{
				pxPrimitiveInstance->m_xBodyState.m_xCentreOfMass += -xAxis * fDistanceB;
			}

		}
	}


	if ( pxPolygonInstance == m_pxInstanceA )
	{
		NotifyObjectsOfCollision( xAxis );
	}
	else
	{
		NotifyObjectsOfCollision( -xAxis );
	}
	return true;
}
bool Asura_Physics_Pairing::FindContactPoints_MPR( void )
{
	Asura_AVector3 xVelocityA = m_pxInstanceA->m_xBodyState.m_xCentreOfMass - m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass;
	Asura_AVector3 xVelocityB = m_pxInstanceB->m_xBodyState.m_xCentreOfMass - m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass;
	Asura_AVector3 xRelativeVelocity = xVelocityB - xVelocityA;

	float fMinimumSafeDistance = Asura_Maths::Min( m_pxInstanceA->GetDefinition()->GetMinimumRadius(), m_pxInstanceB->GetDefinition()->GetMinimumRadius() ) *0.9f;

	bool bDoSweptTest = (xRelativeVelocity.Magnitude() > fMinimumSafeDistance);


	Asura_AVector3 xAxis;
	float fTime = 0.0f;
	float fPenetration = 0.0f;

	bool bSweptIntersect = false;

	if ( bDoSweptTest )
	{
		bSweptIntersect = SweptIntersectionTest( m_pxInstanceA, m_pxInstanceB, xAxis, fTime, fPenetration );
	
		// objects don't intersect in 4D - we're done here
		if (!bSweptIntersect) return false;

		// if the objects aren't trying to move apart, roll back time to the point of contact
		if ( xRelativeVelocity*(xAxis) > 0.0f )
		{ 
			if ( m_pxInstanceA->IsDynamic() && !m_pxInstanceA->IsActor())
			{
				m_pxInstanceA->m_xBodyState.m_xCentreOfMass += - xVelocityA * fTime;
			}
			if ( m_pxInstanceB->IsDynamic() && !m_pxInstanceB->IsActor())
			{
				m_pxInstanceB->m_xBodyState.m_xCentreOfMass += - xVelocityB * fTime;
			}
		}
		// objects are now "just touching"
		fPenetration = 0.0f;
	}

	{
		bool bIntersect = IntersectionTest( m_pxInstanceA, m_pxInstanceB, xAxis, fPenetration );
		
		if (bIntersect)
		{
			// If one or both of the objects are actors, push the actor{s) out to solve the penetration directly
			// (as using velocity based methods won't work for actors).
			// Otherwise this penetration value will be passed onto the solver routine below.
			fPenetration = -fPenetration;
			const float fScaledDistance = ( fPenetration + fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION ) * m_fMassCoefficient;
			
			if ( m_pxInstanceA->IsActor())
			{
				const float fDistanceA = Asura_Maths::Min( fScaledDistance * m_pxInstanceA->GetInvMass(), m_pxInstanceA->GetDefinition()->GetMinimumRadius() * 0.9f );
				m_pxInstanceA->m_xBodyState.m_xCentreOfMass += xAxis * fDistanceA;
			}
			if ( m_pxInstanceB->IsActor())
			{
				const float fDistanceB = Asura_Maths::Min( fScaledDistance * m_pxInstanceB->GetInvMass(), m_pxInstanceB->GetDefinition()->GetMinimumRadius() * 0.9f );
				m_pxInstanceB->m_xBodyState.m_xCentreOfMass += -xAxis * fDistanceB;
			}
		}

		// if neither intersection test fired, we're done here.
		if (!bIntersect && !bSweptIntersect) return false;
	}


	Asura_AVector3 axContactPointsForA[10];
	Asura_AVector3 axContactPointsForB[10];
	u_int uNumberOfContactPoints;

	uNumberOfContactPoints = Asura_Physics_CCD::FindContactPointsFromSupportFeatures( axContactPointsForA, axContactPointsForB, m_pxInstanceA, m_pxInstanceB, xAxis );

	if ( !uNumberOfContactPoints )
	{
		// KJFL - 14 November 2008 - bit of an experiment. If the clipping code finds no contacts, rather than just returning,
		// just take the first support point of each object. We could move this into FindContactPointsFromSupportFeatures(),
		// but it's a special case that shouldn't happen often.
		m_pxInstanceA->GetSupportFeature( axContactPointsForA, xAxis );
		m_pxInstanceB->GetSupportFeature( axContactPointsForB, -xAxis );
		uNumberOfContactPoints = 1;
	}

	for ( u_int u=0; u<uNumberOfContactPoints; ++u)
	{
		Asura_AVector3 xRelA = axContactPointsForA[u] - m_pxInstanceA->m_xBodyState.m_xCentreOfMass;
		Asura_AVector3 xRelB = axContactPointsForB[u] - m_pxInstanceB->m_xBodyState.m_xCentreOfMass;

		Asura_Physics_Solver::AddPairingContact( m_pxInstanceA, xRelA, m_pxInstanceB, xRelB, xAxis, fPenetration );
	}

	#ifdef ASURA_USE_COLLISIONFX
	if (ShouldAddCollisionFXReport())
	{
		// Find the average of all the contact points on both bodies,
		// and report that as the CollisionFX position.
		Asura_AVector3 xCentre;
		xCentre.Zero();

		for ( u_int u=0; u<uNumberOfContactPoints; ++u)
		{
			xCentre += axContactPointsForA[u];
			xCentre += axContactPointsForB[u];
		}

		xCentre /= (static_cast<float>(uNumberOfContactPoints*2));

		AddCollisionFXReport( xCentre, xAxis );
	}
	#endif//#ifdef ASURA_USE_COLLISIONFX

	NotifyObjectsOfCollision( xAxis );

	return true;
}
#endif

bool Asura_Physics_Pairing::FindContactPoints_SweptOBBs( void )
{
	const Asura_AMatrix3x3& xObjectAOrient = m_pxInstanceA->GetOrientation();
	const Asura_AMatrix3x3& xObjectBOrient = m_pxInstanceB->GetOrientation();

	// The potential separating axes are formed from the edges of the
	// OBBs, and the vectors perpendicular to the edge pairings. This
	// makes 15 axes to test

	// KJFL - 10 July 2006 - 
	// Once this is working nicely it'll be worth profiling and seeing
	// how much of a hit these initial tests are - we could avoid doing
	// the cross-products to find the axes until we need them, and even
	// have an extra member in the pairing which holds the 'last' 
	// separating axis, so you can do a quick test-reject cycle for 
	// stationary objects close-by but actually separated.

	const unsigned int uNumberOfSeparatingAxes = 15;

	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn0() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn1() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn0(),
		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn1(),
		xObjectAOrient.GetColumn2() ^ xObjectBOrient.GetColumn2(),

		xObjectAOrient.GetColumn0(),
		xObjectAOrient.GetColumn1(),
		xObjectAOrient.GetColumn2(),

		xObjectBOrient.GetColumn0(),
		xObjectBOrient.GetColumn1(),
		xObjectBOrient.GetColumn2()
	};

	float afSeparationDistance[uNumberOfSeparatingAxes];

	float fOverlapMinTime = 0.0f;
	float fOverlapMaxTime = 1.0f;
	u_int uBestAxis = uNumberOfSeparatingAxes;

	Asura_AVector3 xVelocityA = m_pxInstanceA->m_xBodyState.m_xCentreOfMass - m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass;

	Asura_AVector3 xVelocityB = m_pxInstanceB->m_xBodyState.m_xCentreOfMass - m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass;

	Asura_AVector3 xRelativeVelocity = xVelocityB - xVelocityA;

	// KJFL - 15 January 2007 - can't just use the positions here, since you can get objects which have their "position" on their faces etc.
	// This should really be something like "previous CoM to previous CoM" something TODO when I rejig CoM and Position.
	const Asura_AVector3 xAToB = (m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass)- (m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass);


	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		afSeparationDistance[u] = Asura_Maths::LargeFloat / 4.0f;

		Asura_AVector3& xAxis = axAxes[u];

		// KJFL - 27 June 2006 - ah, it's possible some of the axes are invalid
		// because they're zero vectors, simply because the edges of the two 
		// objects might be parallel (cross products above would give zero vectors).
	
		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;
		

		xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects

		float fCentreA = 0.0f;
		float fExtentsA = 0.0f;

		m_pxInstanceA->GetCentreAndHalfExtentAlongAxis( xAxis, fCentreA, fExtentsA );

		float fCentreB = 0.0f;
		float fExtentsB = 0.0f;

		m_pxInstanceB->GetCentreAndHalfExtentAlongAxis( xAxis, fCentreB, fExtentsB );

        float fRelativeVelocityAlongAxis = ( xRelativeVelocity ) * xAxis ;

		float fMinkowskiSumMin = fCentreA - fCentreB - fExtentsA - fExtentsB;
		float fMinkowskiSumMax = fCentreA - fCentreB + fExtentsA + fExtentsB;

		float fRayMin = Asura_Maths::Min( 0.0f, -fRelativeVelocityAlongAxis );
		float fRayMax = Asura_Maths::Max( 0.0f, -fRelativeVelocityAlongAxis );

		// test for a no overlap case
		if ( fRayMax < fMinkowskiSumMin || fRayMin > fMinkowskiSumMax )
		{
			// these objects definitely don't overlap
			return false;
		}
	
		
		float fTime1 = 0.0f;
		float fTime2 = 1.0f;

		if ( Asura_Maths::Abs( fRelativeVelocityAlongAxis ) > 0.00001f )
		{
			fTime1 = ( fMinkowskiSumMax + fRelativeVelocityAlongAxis ) / ( fRelativeVelocityAlongAxis );
			fTime2 = ( fMinkowskiSumMin + fRelativeVelocityAlongAxis ) / ( fRelativeVelocityAlongAxis );
		}


		float fMinTime = Asura_Maths::Min( fTime1, fTime2 );
		float fMaxTime = Asura_Maths::Max( fTime1, fTime2 );
	
		if ( fMinTime > fOverlapMinTime )
		{
			fOverlapMinTime = fMinTime;
			uBestAxis = u;
		}
		fOverlapMaxTime = Asura_Maths::Min( fOverlapMaxTime, fMaxTime );


		if (fOverlapMaxTime < fOverlapMinTime ) return false;

		afSeparationDistance[u] = Asura_Maths::Min( fMinkowskiSumMax, -fMinkowskiSumMin );
	}

	
	float fMinimumDistance = Asura_Maths::LargeFloat;

	if (fOverlapMinTime <= 0.0f)
	{
		uBestAxis = uNumberOfSeparatingAxes;

		for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
		{
			// skip invalid axes (e.g. check for zero vectors)
			if ( axAxes[u].MagnitudeSquared() < 0.001f ) continue;

			const float f = afSeparationDistance[u];

			if ( f > 0.0f && f < fMinimumDistance )
			{
				uBestAxis = u;
				fMinimumDistance = f;
			}
			else if ( f <= 0.0f)
			{
				return false;
			}
		}
	}
	//else return false;

	if ( uBestAxis >= uNumberOfSeparatingAxes  )
	{
		// this is unexpected... let's assert and return a verdict of no collision
		Asura_Assert( uBestAxis < uNumberOfSeparatingAxes, "Invalid axis" );

		return false;
	}

	// We've found the axis of separation
	Asura_AVector3 xAxis = axAxes[uBestAxis];

	// Calculate the impulses to be given to each object - which depends on their relative masses
	Asura_AVector3 xImpulseA;
	Asura_AVector3 xImpulseB;

	xImpulseA.Zero();
	xImpulseB.Zero();
#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
	float fPenetration = 0.0f;
#endif

	if (fOverlapMinTime >= 1.0f)
	{
		// there isn't a collision this frame, but there may be later
		return false;
	}
	else if ( fOverlapMinTime > 0.0f )
	{
		// turn back time to point of impact
		float fRewindTime = 1.0f - fOverlapMinTime;
	
		if ( m_pxInstanceA->IsDynamic() && !m_pxInstanceA->IsActor())
		{
			xImpulseA = - xVelocityA * fRewindTime;
		}
		if ( m_pxInstanceB->IsDynamic() && !m_pxInstanceB->IsActor())
		{
			xImpulseB = - xVelocityB * fRewindTime;
		}
	}
	else
	{	
		// Limit the amount we can be moved to less than our min half-extent. This means we won't
		// have our centre pushed through the floor etc.
#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
		fPenetration = fMinimumDistance + fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION;
#else
		const float fScaledDistance = ( fMinimumDistance + fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION ) * m_fMassCoefficient;
		const float fDistanceA = Asura_Maths::Min( fScaledDistance * m_pxInstanceA->GetInvMass(), m_pxInstanceA->GetDefinition()->GetMinimumRadius() * 0.9f );
		const float fDistanceB = Asura_Maths::Min( fScaledDistance * m_pxInstanceB->GetInvMass(), m_pxInstanceB->GetDefinition()->GetMinimumRadius() * 0.9f );

		xImpulseA = xAxis * fDistanceA;
		xImpulseB = -xAxis * fDistanceB;
#endif
	}

	// Ok, here's a scary bit. First we need to find the support features for both objects, and then we
	// need to project the support features and clip them to find the overlapping contact feature,
	// which gives us our contact points. Worst case is two faces whose common overlap is an octagon.

	Asura_AVector3 axContactPointsForA[10];
	Asura_AVector3 axContactPointsForB[10];
	u_int uNumberOfContactPoints;


	uNumberOfContactPoints = Asura_Physics_CCD::FindContactPointsFromSupportFeatures( axContactPointsForA, axContactPointsForB, m_pxInstanceA, m_pxInstanceB, xAxis );

	if ( !uNumberOfContactPoints )
	{
		return false;
	}

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
	for ( u_int u=0; u<uNumberOfContactPoints; ++u)
	{
		Asura_AVector3 xRelA = axContactPointsForA[u] - m_pxInstanceA->m_xBodyState.m_xCentreOfMass;
		Asura_AVector3 xRelB = axContactPointsForB[u] - m_pxInstanceB->m_xBodyState.m_xCentreOfMass;

		Asura_Physics_Solver::AddPairingContact( m_pxInstanceA, xRelA, m_pxInstanceB, xRelB, xAxis, fPenetration );
	}
	if (fPenetration==0.0f)
	{


		if ( m_pxInstanceA->IsDynamic() )
		{
			//m_pxInstanceA->PositionProjection( xImpulseA, axContactPointsForA, uNumberOfContactPoints );

			m_pxInstanceA->m_xBodyState.m_xCentreOfMass += xImpulseA;
		}
		if ( m_pxInstanceB->IsDynamic() )
		{
			//m_pxInstanceB->PositionProjection( xImpulseB, axContactPointsForB, uNumberOfContactPoints );

			m_pxInstanceB->m_xBodyState.m_xCentreOfMass += xImpulseB;
		}
	}
#else
	CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForA, axContactPointsForB, xAxis );

	if ( m_pxInstanceA->IsDynamic() )
	{
		m_pxInstanceA->PositionProjection( xImpulseA, axContactPointsForA, uNumberOfContactPoints );

		//m_pxInstanceA->m_xBodyState.m_xCentreOfMass += xImpulseA;
	}
	if ( m_pxInstanceB->IsDynamic() )
	{
		m_pxInstanceB->PositionProjection( xImpulseB, axContactPointsForB, uNumberOfContactPoints );

		//m_pxInstanceB->m_xBodyState.m_xCentreOfMass += xImpulseB;
	}
#endif

	NotifyObjectsOfCollision( xAxis );

	// overlap found
	return true;
}

bool Asura_Physics_Pairing::FindContactPoints_SweptOBBVsSweptPolygons( Asura_Physics_RigidBodyInstance* pxOBBInstance, Asura_Physics_RigidBodyInstance* pxPolygonInstance )
{
	// We're going to first assume we need to do a continuous collision detection to find the first time of contact,
	// but if that turns out to be zero or negative, change over to an already-penetrating case.
	bool bCollisionOccurred = false;
	bool bPenetrationCase = false;

	Asura_OBB xOBB( pxOBBInstance->GetHalfExtents(), pxOBBInstance->m_xBodyState.m_xCentreOfMass, pxOBBInstance->GetOrientation() );
	
	// we need the rendering position of the poly instance so we can transform the Tris to world space
	pxPolygonInstance->CalculateRenderingPosition();

	// data to keep track of first impact
	float fTimeToImpact = 1.0f;
	Asura_AVector3 xAxisOfImpact;
	xAxisOfImpact.Zero();
	Asura_Physics_Tri xImpactTri;

	// the closing velocity of the two objects
	const Asura_AVector3 xRelativeVelocity = (pxOBBInstance->m_xBodyState.m_xCentreOfMass - pxOBBInstance->m_xBodyState.m_xPreviousCentreOfMass)
											-(pxPolygonInstance->m_xBodyState.m_xCentreOfMass - pxPolygonInstance->m_xBodyState.m_xPreviousCentreOfMass);

	// offset the OBB to account for the Tri's movement (i.e. as if Tri was static, and the OBB has all the motion)
	xOBB.SetCentre( pxOBBInstance->m_xBodyState.m_xPreviousCentreOfMass +  (pxPolygonInstance->m_xBodyState.m_xCentreOfMass - pxPolygonInstance->m_xBodyState.m_xPreviousCentreOfMass));

	// If nothing is moving, skip any swept tests (though algorithmically they're pretty identical).
	if ( xRelativeVelocity.MagnitudeSquared() == 0.0f )
	{
		bPenetrationCase = true;
	}

	Asura_Physics_CollisionMesh_Iterator xTriIterator( pxPolygonInstance->GetDefinition()->GetCollisionMesh() );
	Asura_Physics_Tri xTri;
	//u_int uTris = 0;

    for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
	{
		// TODO - rather than calculating the OBB and Tri in world space, perhaps we can use the Tri's local space?
		// we'd have to transform something (i.e. OBB or Tri) so perhaps it wouldn't save much
		xTri.Transform( pxPolygonInstance->m_xBodyState.m_xOrientation, pxPolygonInstance->m_xInstancePosition );
		//uTris++;

		// if this isn't a penetration-solving case, look for the time of impact
		if ( !bPenetrationCase )
		{
			u_int uFirstPlane = 0;
			Asura_AVector3 xAxisUsed;
			bool bHitEdge;

			float fHitTime = Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsTri( xOBB, xRelativeVelocity, xTri, xAxisUsed, uFirstPlane, bHitEdge, false );

			if ( fHitTime <= 0.0f ) 
			{
				// move the OBB to it's "current" position, and we'll find the penetration 
				xOBB.SetCentre( pxOBBInstance->m_xBodyState.m_xCentreOfMass );
				bPenetrationCase = true;
			}
			else if ( fHitTime < fTimeToImpact )
			{
				// record the new candidate for the first time of impact
				fTimeToImpact = fHitTime;
				xAxisOfImpact = xAxisUsed;
				xImpactTri = xTri;
			}
		}

		// if it's a penetration-solving case, find the contact points and penetration and penetration axis for this tri
		if ( bPenetrationCase )
		{
			Asura_AVector3 xNormal;
			float fPenetration;

			if ( Asura_Physics_CCD::DoesOBBIntersectTri( xOBB,xTri,fPenetration,xNormal ) )
			{
				bCollisionOccurred = true;			

				Asura_AVector3 axContactPointsForOBB[10];
				Asura_AVector3 axContactPointsForTri[10];
				u_int uNumberOfContactPoints = Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( axContactPointsForOBB, axContactPointsForTri, pxOBBInstance, xTri, xNormal );

				const Asura_AVector3 xImpulse = xNormal * (fPenetration + fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION ) * m_fMassCoefficient;

				const Asura_AVector3 xImpulseOBB = xImpulse * pxOBBInstance->GetInvMass();
				const Asura_AVector3 xImpulsePoly = - xImpulse * pxPolygonInstance->GetInvMass();

				if (m_pxInstanceA == pxOBBInstance )
				{
					NotifyObjectsOfCollision( xNormal );
					CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForOBB, axContactPointsForTri, xNormal );
				}
				else
				{
					NotifyObjectsOfCollision( -xNormal );
					CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForTri, axContactPointsForOBB, -xNormal );
				}

				if ( pxOBBInstance->IsDynamic() )
				{
					pxOBBInstance->m_xBodyState.m_xCentreOfMass += xImpulseOBB;
					xOBB.SetCentre( pxOBBInstance->m_xBodyState.m_xCentreOfMass );
				}
				if ( pxPolygonInstance->IsDynamic() )
				{
					pxPolygonInstance->m_xBodyState.m_xCentreOfMass += xImpulsePoly;

					// Currently need to recalc the rendering position so we can go from Tri->World space.
					// However, since this is a simple translation we might as well do it by handle rather
					// than call CalculateRenderingPosition().
					pxPolygonInstance->m_xInstancePosition += xImpulsePoly;
				}
			}
		}
	}

	// if this was a "time-to-impact" situation, deal with the collision response
	if ( !bPenetrationCase && fTimeToImpact<1.0f )
	{
		bCollisionOccurred = true;

		// turn back time to point of impact
		float fRewindTime = 1.0f - fTimeToImpact;

		xAxisOfImpact.Normalise();

		Asura_AVector3 axContactPointsForOBB[10];
		Asura_AVector3 axContactPointsForTri[10];
		u_int uNumberOfContactPoints = Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( axContactPointsForOBB, axContactPointsForTri, pxOBBInstance, xImpactTri, xAxisOfImpact );

		// make sure you get your minus signs the right way round!
		if (m_pxInstanceA == pxOBBInstance )
		{
			NotifyObjectsOfCollision( xAxisOfImpact );
			CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForOBB, axContactPointsForTri, xAxisOfImpact );
		}
		else
		{
			NotifyObjectsOfCollision( -xAxisOfImpact );
			CalculateCollisionResponses( uNumberOfContactPoints, axContactPointsForTri, axContactPointsForOBB, -xAxisOfImpact );
		}
	
		if ( m_pxInstanceA->IsDynamic() && !m_pxInstanceA->IsActor())
		{
			const Asura_AVector3 xVelocityA = m_pxInstanceA->m_xBodyState.m_xCentreOfMass - m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass;

			// move the object back to first time of collision
			m_pxInstanceA->m_xBodyState.m_xCentreOfMass -= xVelocityA*fRewindTime;

			// and then let it move on with its new velocity
			m_pxInstanceA->m_xBodyState.m_xCentreOfMass += m_pxInstanceA->m_xBodyState.m_xLinearVelocity * fRewindTime;
		}
		if ( m_pxInstanceB->IsDynamic() && !m_pxInstanceB->IsActor() )
		{
			const Asura_AVector3 xVelocityB = m_pxInstanceB->m_xBodyState.m_xCentreOfMass - m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass;

			// move the object back to first time of collision
			m_pxInstanceB->m_xBodyState.m_xCentreOfMass -= xVelocityB*fRewindTime;
				
			// and then let it move on with its new velocity
			m_pxInstanceB->m_xBodyState.m_xCentreOfMass += m_pxInstanceB->m_xBodyState.m_xLinearVelocity * fRewindTime;
		}

		
	}


	return bCollisionOccurred;
}
bool Asura_Physics_Pairing::FindContactPoints_SweptOBBVsSweptSphere( Asura_Physics_RigidBodyInstance* pxOBBInstance, Asura_Physics_RigidBodyInstance* pxSphereInstance )
{
	bool bCollisionOccurred = false;

	float fPenetration;
	Asura_AVector3 xAxis;

	Asura_OBB xOBB( pxOBBInstance->GetHalfExtents(), pxOBBInstance->GetCentreOfMass(), pxOBBInstance->GetOrientation() );

	float fSphereRadius = pxSphereInstance->GetDefinition()->GetMaximumRadius();
	Asura_AVector3 xSphereCentre = pxSphereInstance->m_xBodyState.m_xPreviousCentreOfMass;

	Asura_AVector3 xDisplacement = pxSphereInstance->GetCentreOfMass() - pxSphereInstance->m_xBodyState.m_xPreviousCentreOfMass;
	float fDistance = xDisplacement.NormaliseReturningMagnitude();

	u_int uIterations = Asura_Maths::FloatToInt( fDistance / fSphereRadius ) + 1;

    fDistance /= (float)uIterations;

	if ( fDistance > 0.0f )
	{
		xDisplacement *= fDistance;
	}
	else
	{
		xDisplacement.Zero();
	}

	bCollisionOccurred = Asura_Physics_CCD::DoesOBBIntersectSphere( xOBB, xSphereCentre, fSphereRadius, fPenetration, xAxis );
	
	// Unless it's the final collision test, ignore collisions when the sphere is trying to move away from the OBB.

	// ignore the collision if the sphere is trying to head away from the problem
	if ( bCollisionOccurred && (xAxis * xDisplacement < 0.0f) ) bCollisionOccurred = false;

	while ( !bCollisionOccurred && uIterations-- )
	{
		xSphereCentre += xDisplacement;
		bCollisionOccurred = Asura_Physics_CCD::DoesOBBIntersectSphere( xOBB, xSphereCentre, fSphereRadius, fPenetration, xAxis );
	
		// ignore the collision if the sphere is trying to head away from the problem
		if ( uIterations && bCollisionOccurred && (xAxis * xDisplacement < 0.0f) ) bCollisionOccurred = false;
	}

	pxSphereInstance->m_xBodyState.m_xCentreOfMass = xSphereCentre;


	if (bCollisionOccurred)
	{
#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
		{
			Asura_AVector3 xRelA = xAxis*(fSphereRadius);
			Asura_AVector3 xRelB = pxSphereInstance->m_xBodyState.m_xCentreOfMass + xAxis*(fSphereRadius-fPenetration) - pxOBBInstance->m_xBodyState.m_xCentreOfMass;

			Asura_Physics_Solver::AddPairingContact( pxSphereInstance, xRelA, pxOBBInstance, xRelB, -xAxis, fPenetration );
		}
		// make sure you get your minus signs the right way round!
		if (m_pxInstanceA == pxOBBInstance )
		{
			NotifyObjectsOfCollision( xAxis );
		}
		else
		{
			NotifyObjectsOfCollision( -xAxis );
		}
#else
		// Do min penetration epsilon after splitting the penetrations between objects - this means very heavy objects still get an epsilon, rather than the mass ratios completely eroding it.
		fPenetration *= m_fMassCoefficient;

		if ( pxOBBInstance->IsDynamic() )// && !pxOBBInstance->IsActor())
		{
			pxOBBInstance->m_xBodyState.m_xCentreOfMass += xAxis * Asura_Maths::Max( fPenetration * pxOBBInstance->GetInvMass(), fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION );
		}
		if ( pxSphereInstance->IsDynamic() )//&& !pxSphereInstance->IsActor() )
		{
			pxSphereInstance->m_xBodyState.m_xCentreOfMass -= xAxis * Asura_Maths::Max( fPenetration * pxSphereInstance->GetInvMass(), fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION );
		}

		Asura_AVector3 xContact = pxSphereInstance->m_xBodyState.m_xCentreOfMass + xAxis * fSphereRadius;

		// make sure you get your minus signs the right way round!
		if (m_pxInstanceA == pxOBBInstance )
		{
			NotifyObjectsOfCollision( xAxis );
			CalculateCollisionResponses( 1, &xContact, &xContact, xAxis );
		}
		else
		{
			NotifyObjectsOfCollision( -xAxis );
			CalculateCollisionResponses( 1, &xContact, &xContact, -xAxis );
		}
#endif
	}

	return bCollisionOccurred;
}

bool Asura_Physics_Pairing::FindContactPoints_SweptSphereVsSweptPolygons( Asura_Physics_RigidBodyInstance* pxSphereInstance, Asura_Physics_RigidBodyInstance* pxPolygonInstance )
{
	bool bCollisionOccurred = false;

	float fPenetration;
	Asura_AVector3 xAxis;
	Asura_AVector3 xContactPoint;

	float fSphereRadius = pxSphereInstance->GetDefinition()->GetMaximumRadius();
	Asura_AVector3 xSphereCentre = pxSphereInstance->m_xBodyState.m_xPreviousCentreOfMass;

	Asura_AVector3 xDisplacement = pxSphereInstance->GetCentreOfMass() - pxSphereInstance->m_xBodyState.m_xPreviousCentreOfMass;
	float fDistance = xDisplacement.NormaliseReturningMagnitude();

	u_int uIterations = Asura_Maths::FloatToInt( fDistance / fSphereRadius ) + 1;

	fDistance /= (float)uIterations;

	if ( fDistance > 0.0f )
	{
		xDisplacement *= fDistance;
	}
	else
	{
		xDisplacement.Zero();
	}

	Asura_Physics_CollisionMesh_Iterator xTriIterator( pxPolygonInstance->GetDefinition()->GetCollisionMesh() );
	Asura_Physics_Tri xTri;

	for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
	{
		// TODO - rather than calculating the OBB and Tri in world space, perhaps we can use the Tri's local space?
		// we'd have to transform something (i.e. OBB or Tri) so perhaps it wouldn't save much
		xTri.Transform( pxPolygonInstance->m_xBodyState.m_xOrientation, pxPolygonInstance->m_xInstancePosition );
	
		bool bCollisionWithThisTri = Asura_Physics_CCD::FindPenetration_SphereVsTri( xTri, xSphereCentre, fSphereRadius, fPenetration, xAxis, xContactPoint );

		if ( bCollisionWithThisTri && (xAxis * xDisplacement < 0.0f) )
		{
			bCollisionOccurred = true;
	
			// Do min penetration epsilon after splitting the penetrations between objects - this means very heavy objects still get an epsilon, rather than the mass ratios completely eroding it.
			fPenetration *= m_fMassCoefficient;

			// KFJL - 18 May - currently polygonal mesh is only supported on external control or static - so don't try to move it
			/*
			if ( pxTriInstance->IsDynamic() )// && !pxOBBInstance->IsActor())
			{
			pxTriInstance->m_xBodyState.m_xCentreOfMass += xAxis * Asura_Maths::Max( fPenetration * pxOBBInstance->GetInvMass(), fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION );
			}
			*/
			if ( pxSphereInstance->IsDynamic() )//&& !pxSphereInstance->IsActor() )
			{
				xSphereCentre += xAxis * Asura_Maths::Max( fPenetration * pxSphereInstance->GetInvMass(), fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION );
				pxSphereInstance->m_xBodyState.m_xCentreOfMass = xSphereCentre;
			}

			Asura_AVector3 xContact = xSphereCentre - xAxis * fSphereRadius;

			// make sure you get your minus signs the right way round!
			if (m_pxInstanceA == pxSphereInstance )
			{
				NotifyObjectsOfCollision( xAxis );
				CalculateCollisionResponses( 1, &xContact, &xContact, xAxis );
			}
			else
			{
				NotifyObjectsOfCollision( -xAxis );
				CalculateCollisionResponses( 1, &xContact, &xContact, -xAxis );
			}
		}

	}
	
	// Unless it's the final collision test, ignore collisions when the sphere is trying to move away from the OBB.

	// ignore the collision if the sphere is trying to head away from the problem
	while ( !bCollisionOccurred && uIterations-- )
	{
		xSphereCentre += xDisplacement;
		for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
		{
			// TODO - rather than calculating the OBB and Tri in world space, perhaps we can use the Tri's local space?
			xTri.Transform( pxPolygonInstance->m_xBodyState.m_xOrientation, pxPolygonInstance->m_xInstancePosition );

			bool bCollisionWithThisTri = Asura_Physics_CCD::FindPenetration_SphereVsTri( xTri, xSphereCentre, fSphereRadius, fPenetration, xAxis, xContactPoint );

			if ( bCollisionWithThisTri && (uIterations==0 || (xAxis * xDisplacement < 0.0f)))
			{
				bCollisionOccurred = true;

				// Do min penetration epsilon after splitting the penetrations between objects - this means very heavy objects still get an epsilon, rather than the mass ratios completely eroding it.
				fPenetration *= m_fMassCoefficient;

				// KFJL - 18 May - currently polygonal mesh is only supported on external control or static - so don't try to move it
/*
				if ( pxTriInstance->IsDynamic() )// && !pxOBBInstance->IsActor())
				{
					pxTriInstance->m_xBodyState.m_xCentreOfMass += xAxis * Asura_Maths::Max( fPenetration * pxOBBInstance->GetInvMass(), fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION );
				}
*/
				if ( pxSphereInstance->IsDynamic() )//&& !pxSphereInstance->IsActor() )
				{
					xSphereCentre += xAxis * Asura_Maths::Max( fPenetration * pxSphereInstance->GetInvMass(), fASURA_PHYSICS_PAIRING_EXTRA_SEPARATION );
					pxSphereInstance->m_xBodyState.m_xCentreOfMass = xSphereCentre;
				}

				Asura_AVector3 xContact = xSphereCentre - xAxis * fSphereRadius;

				// make sure you get your minus signs the right way round!
				if (m_pxInstanceA == pxSphereInstance )
				{
					NotifyObjectsOfCollision( xAxis );
					CalculateCollisionResponses( 1, &xContact, &xContact, xAxis );
				}
				else
				{
					NotifyObjectsOfCollision( -xAxis );
					CalculateCollisionResponses( 1, &xContact, &xContact, -xAxis );
				}
			}
		}
	}

	pxSphereInstance->m_xBodyState.m_xCentreOfMass = xSphereCentre;


	

	return bCollisionOccurred;
}
#ifdef ASURA_USE_COLLISIONFX
bool Asura_Physics_Pairing::ShouldAddCollisionFXReport( void )
{
	// MGH 12/10/2007: We only want a single report per frame, not one per physics update pass.
	// check frame stamp to see if we've done a report already
//	if (m_uFXReportStamp == Asura_Physics_System::GetPhysicsUpdateCount()) return false;
	if (m_uFXReportStamp == Asura_Timers::GetFrameNumber()) return false;

	// check instances flags
	return ( m_pxInstanceA->GetReportObjectCollisionFX() || m_pxInstanceB->GetReportObjectCollisionFX() );
}

void Asura_Physics_Pairing::AddCollisionFXReport( const Asura_AVector3& xPosition, const Asura_AVector3& xCollisionAxis )
{
	// MGH 12/10/2007: We only want a single report per frame, not one per physics update pass.
//	m_uFXReportStamp = Asura_Physics_System::GetPhysicsUpdateCount();
	m_uFXReportStamp = Asura_Timers::GetFrameNumber();

	Asura_AVector3 xRel = m_pxInstanceA->m_xBodyState.m_xLinearVelocity - m_pxInstanceB->m_xBodyState.m_xLinearVelocity;

	// MGH 10/10/2007: This seems to give the same result as the above calculation (which is quicker).
//	Asura_AVector3 xRel = (m_pxInstanceA->m_xBodyState.m_xCentreOfMass - m_pxInstanceA->m_xBodyState.m_xPreviousCentreOfMass) -
//						  (m_pxInstanceB->m_xBodyState.m_xCentreOfMass - m_pxInstanceB->m_xBodyState.m_xPreviousCentreOfMass);

	// MGH 10/10/2007: We want the actual velocity, so divide by the time.
	xRel /= Asura_Physics_System::GetTimeSlice();

	Asura_Physics_CollisionFX_Report_System::AddReport( xPosition, xRel, m_pxInstanceA->GetAssociatedEntityGuid(), m_pxInstanceB->GetAssociatedEntityGuid(), xCollisionAxis );
}
#endif //#ifdef ASURA_USE_COLLISIONFX

void Asura_Physics_Pairing::NotifyObjectsOfCollision( const Asura_AVector3& xAxis )
{
	if (m_pxInstanceA->IsAsleep()) m_pxInstanceA->WakeUp();
	if (m_pxInstanceB->IsAsleep()) m_pxInstanceB->WakeUp();

	if (m_pxInstanceB->IsWideAwake())
	{
		m_pxInstanceA->m_bHasAwakePartner = true;
	}
	if (m_pxInstanceA->IsWideAwake())
	{
		m_pxInstanceB->m_bHasAwakePartner = true;
	}

	m_pxInstanceA->AddObjectCollisionReport( m_pxInstanceB );
	m_pxInstanceB->AddObjectCollisionReport( m_pxInstanceA );

	// KJFL - remember xAxis points towards A, and away from B
	float fGravityADotAxis = m_pxInstanceA->GetGravityDirection() * xAxis;
	float fGravityBDotAxis = m_pxInstanceB->GetGravityDirection() * xAxis;

	static const float fVERTICAL_BOUNDARY = 0.70710678f; // = cosine(45degrees)

	// Judging by the collision axis and the gravity direction, is A supported?
	if ( fGravityADotAxis < -fVERTICAL_BOUNDARY )
	{
		// If A is dynamic, mark the appropriate flags for being supported
		if ( m_pxInstanceA->IsDynamic() )
		{
			m_pxInstanceA->m_bIsInContactWithFloor = true;
			m_pxInstanceA->m_bIsSupported = true;
			m_pxInstanceA->m_xJumpVelocity.Zero();
		}
		// If B has it's own restframe (and A doesn't) then we should link A as 'on' B
		if ( m_pxInstanceB->HasOwnRestFrame() && (!m_pxInstanceA->HasOwnRestFrame()) )
		{
			m_pxInstanceA->LinkToRestFrame( m_pxInstanceB->m_pxRestFrame );
		}
	}

	// Judging by the collision axis and the gravity direction, is B supported?
	if ( fGravityBDotAxis > fVERTICAL_BOUNDARY )
	{
		// If B is dynamic, mark the appropriate flags for being supported
		if ( m_pxInstanceB->IsDynamic() )
		{
			m_pxInstanceB->m_bIsInContactWithFloor = true;
			m_pxInstanceB->m_bIsSupported = true;
			m_pxInstanceB->m_xJumpVelocity.Zero();
		}
		// If A has it's own restframe (and B doesn't) then we should link B as 'on' A
		if ( m_pxInstanceA->HasOwnRestFrame() && (!m_pxInstanceB->HasOwnRestFrame()) )
		{
			m_pxInstanceB->LinkToRestFrame( m_pxInstanceA->m_pxRestFrame );
		}
	}
}

/*
void Asura_Physics_Pairing::DestroyBrokenContacts( void )
{

	{

		Asura_Physics_Pairing_Contact** ppxNextContactPointer = &m_pxContactsList;

		while (*ppxNextContactPointer) // is there a 'next' Contact?
		{
			// let's get it
			Asura_Physics_Pairing_Contact* pxContact = *ppxNextContactPointer;

			// do we need to remove this contact?
			if ( pxContact->IsBroken() )
			{
				// remove it from the list then.
				// make the previous 'next' point to this Contact's 'next'
				*ppxNextContactPointer = pxContact->m_pxNext;

				m_pxInstanceA->RemoveContactVertex( &(pxContact->m_axContactVerticesForA ) );

				m_pxInstanceB->RemoveContactVertex( &(pxContact->m_axContactVerticesForB ) );

				Asura_Physics_Pairing_Contact::Destroy( pxContact );
			}
			else
			{

				// move onto next ContactVertex in list, please
				ppxNextContactPointer = &(pxContact->m_pxNext);
			}
		}

	}
}

Asura_Physics_Pairing_Contact* Asura_Physics_Pairing::AddContact(void)
{
	if (m_pxInstanceA->IsAsleep()) m_pxInstanceA->WakeUp();
	if (m_pxInstanceB->IsAsleep()) m_pxInstanceB->WakeUp();


	Asura_Physics_Pairing_Contact* pxContact = new Asura_Physics_Pairing_Contact;
	
	if (pxContact)
	{
		pxContact->m_pxNext = m_pxContactsList;
		
		m_pxContactsList = pxContact;
	}
	return pxContact;
}
*/
bool Asura_Physics_Pairing::IsBroken( void ) const
{
	// KJFL - 15 November 2005 - Currently we're going to assume a pairing is broken
	// if it hasn't been flagged as up-to-date by being added to an island. If we 
	// find objects nearly touching tend to thrash in and out of a pairing we could
	// change this to say if a pairing has gone several updates without being in an 
	// island, then destroy it.
	return ( m_uLastUpdateStamp != Asura_Physics_System::GetPhysicsUpdateCount() );
}



