//
// Asura_Physics_Ragdoll.h
//
// Created by KJFL - 12 October 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_RAGDOLL_H__
#define __ASURA_PHYSICS_RAGDOLL_H__

#ifndef ASURA_PHYSICS_DISABLE_RAGDOLL

#include "Core/Asura_Maths.h"
#include "Core/Asura_MempoolFactory_Template.h"
#include "Core/Asura_HashTree.h"
#include "Network/Asura_Guid.h"
#include "Physics/Asura_Physics_BodyState.h"
#include "Physics/Asura_Physics_EnvironmentCache.h"
#include "Physics/Asura_Physics_Contact.h"
#include "Physics/Asura_Physics_ContactCluster.h"
#include "Physics/Asura_Physics_Ragdoll_Description.h"


class Asura_Animation;
class Asura_Physics_Ragdoll;
typedef Asura_MempoolFactory_Template<Asura_Physics_Ragdoll> Asura_Physics_Ragdoll_Factory;
class Asura_Physics_Ragdoll_CollisionObject;
class Asura_Physics_Ragdoll_Desc_Joint;
class Asura_Physics_Solver_ConeTwistConstraint;
class Asura_Physics_Solver_Body;



#ifdef ASURA_USE_COLLISIONFX
// KJFL - 3 March 2009
// This class acts as a buffer which only stores the fastest environment collision found
// so far, and then a report is made about just that one contact.
class Asura_Physics_Ragdoll_CFX_Reporter
{
	public:	

		void Reset( void )
		{
			m_fVelocityDotNormal = 0.0f;
		}

		void RegisterContact( const Asura_AVector3& xUnscaledVelocity, const Asura_AVector3& xPosition, const Asura_AVector3& xNormal, u_int uEnvironmentFaceID )
		{
			// KJFL - originally tried using the squared magnitude of the velocity,
			// but we're really after a measure of the "biggest impact" so we'll
			// find the speed perpendicularly into the polygon.
			float fVelocityDotNormal = -xUnscaledVelocity * xNormal;

			if ( fVelocityDotNormal > m_fVelocityDotNormal )
			{
				m_xUnscaledVelocity = xUnscaledVelocity;
				m_xPosition = xPosition;
				m_xNormal = xNormal;
				m_uEnvironmentFaceID = uEnvironmentFaceID;
				m_fVelocityDotNormal = fVelocityDotNormal;
			}
		}

		void CreateReports( Asura_Guid uEntityGuid ) const;

	private:

		Asura_AVector3	m_xUnscaledVelocity;
		Asura_AVector3	m_xPosition;
	
		Asura_AVector3	m_xNormal;
		u_int			m_uEnvironmentFaceID;

		float			m_fVelocityDotNormal;
};

#endif // ASURA_USE_COLLISIONFX

class Asura_Physics_Ragdoll
{
	public:

		void Initialise( Asura_Animation* pxAnim, const Asura_Physics_Ragdoll_Description* pxDesc, const Asura_Vector_3& xAnimPos, const Asura_Matrix_3x3& xAnimOri );
		void Shutdown( void );

		void Update( void );

		// TODO - be good to add version number parameter to ReadFromChunkStream
		void ReadFromChunkStream( Asura_Chunk_Stream& xStream ) {}
		void WriteToChunkStream( Asura_Chunk_Stream& xStream) const {}

		#ifdef ASURA_USE_COLLISIONFX
		void SetReportEnvironmentCollisionFX( bool bReport ) { m_bReportEnvironmentCollisionFX = bReport; }
		bool GetReportEnvironmentCollisionFX( void ) const { return m_bReportEnvironmentCollisionFX; }
		#endif

		void SetAssociatedEntityGuid( Asura_Guid uGuid ) { m_uEntityGuid = uGuid; }
		Asura_Guid GetAssociatedEntityGuid( void ) const { return m_uEntityGuid; }

		// this now returns the bounding box covering the animation nodes
		const Asura_Bounding_Box& GetBoundingBox( void ) const;
		
		void GetRenderingPosition( Asura_Vector_3& xPosition ) const;
		void GetRenderingOrientation( Asura_Matrix_3x3& xOrientation ) const;

		// Apply an impulse to the ragdoll
		void ApplyImpulseToAll( const Asura_AVector3& xImpulse );
		void ApplyImpulseToNearbyNodes( const Asura_AVector3& xPosition, const Asura_AVector3& xImpulse );


		// Ideally deprecated Asura_Vector_3 versions
		void ApplyImpulseToAll( const Asura_Vector_3& xImpulse ) { ApplyImpulseToAll( Asura_AVector3( xImpulse ) ); }
		void ApplyImpulseToNearbyNodes( const Asura_Vector_3& xPosition, const Asura_Vector_3& xImpulse ) { ApplyImpulseToNearbyNodes( Asura_AVector3(xPosition), Asura_AVector3(xImpulse) ); }


		// Get the position and/or orientation in World Space of a marker point on the ragdoll.
		// TODO - not implemented yet	
		//bool GetMarkerLocation( int iIndex, Asura_Vector_3* pxPosition, Asura_Matrix_3x3* pxOrientation) const
		//{ return false; }
	
		// Query sleep status. Sleeping ragdolls take no physics time.
		inline bool IsAsleep( void ) const
		{
			return m_bAsleep;
		}

		// Is the ragdoll moving? Useful as the ragdoll's position may remain constant between 
		// entity updates due to the frequency of physics updates.
		inline bool IsMoving( void ) const
		{
			return (m_fTimeMotionless == 0.0f);
		}

		// Tell a ragdoll to wake up and become active. Note, this has an effect even if the ragdoll
		// is already awake, since it resets the 'time spent motionless' counter to zero.
		inline void WakeUp( void )
		{
			m_bAsleep = false;
			m_fTimeMotionless = 0.0f;
			m_fTimeInContact = 0.0f;
		}

		// Tell the ragdoll to deactivate. It will only wake up if something interacts with it.
		void Sleep( void );

		// update the anim data prior to rendering (doesn't need to be called if the ragdoll isn't moving)
		void UpdateAnimation( void );

		u_short GetFlagsOfPolysToIgnore( void ) const
		{
			return m_usFlagsOfPolysToIgnore;
		}

		void SetFlagsOfPolysToIgnore( const u_short usFlags );

		void	SetDampingScale	(const float fScale)	{ m_fDampingScale = fScale; }
		float	GetDampingScale	(void) const			{ return(m_fDampingScale); }

		void	DebugRender( void );

		Asura_Animation* GetAnimation( void	) const		{ return m_pxAnim; }

#if defined(ASURA_USE_STREAMING)
		// MGH 10/4/2006: Zone has been unloaded. Clean up any remaining references to it
		// in the environment cache.
		void RemoveZone(u_int uZoneIndex);
#endif //defined(ASURA_USE_STREAMING)


		// KJFL - 28 January 2008 - TODO
		// Function AvP added to the old ragdoll system to force the position of a particular node - 
		// this interface will most likely have to change, and currently it does nothing. 
		// Note: the function returns a bool, but the calling code in AvP never seems to use it.
		bool SetBonePosition(const u_int uBoneID, const Asura_Vector_3& xPosition) { return false; }

	#ifdef ASURA_USE_COLLISIONFX

		void ResetCFX( void )
		{
			m_xCFXReporter.Reset();
		}
		void ReportCFX( void )
		{
			if ( GetReportEnvironmentCollisionFX() )
			{
				m_xCFXReporter.CreateReports( GetAssociatedEntityGuid() );
			}
		}
		void RegisterCFX( const Asura_AVector3& xUnscaledVelocity, const Asura_AVector3& xPosition, const Asura_AVector3& xNormal, u_int uEnvironmentFaceID )
		{
			m_xCFXReporter.RegisterContact( xUnscaledVelocity, xPosition, xNormal, uEnvironmentFaceID );		
		} 
	#endif


	private:

		friend class Asura_Physics_Solver;

		// Update the bounding box from the animation data
		void UpdateBoundingBox( void );

		// Zero the momentum of the entire ragdoll, i.e. all its constituent parts
		void ZeroMomentum( void );
			
		void SetupConeTwist( Asura_Physics_Solver_ConeTwistConstraint& xConstraint, const Asura_Physics_Ragdoll_Desc_Joint& xJoint, Asura_Physics_Solver_Body* pxBodies );

		void CheckSleepiness( void );

		friend class Asura_Physics_Ragdoll_System;

		Asura_Animation*						m_pxAnim;

		const Asura_Physics_Ragdoll_Description* m_pxRagdollDesc;

		Asura_Bounding_Box						m_xBoundingBox;

		Asura_Physics_Ragdoll_CollisionObject*	m_pxCollisionObjects;

		Asura_Matrix_3x3*						m_pxNodeOrientations;
		Asura_Vector_3*							m_pxNodePositions;

		u_int									m_bAsleep :1;
		u_int									m_bReportEnvironmentCollisionFX :1;

		Asura_Guid								m_uEntityGuid;

		float									m_fTimeMotionless;
		float									m_fTimeInContact;
		
		float									m_fDampingScale;

		#ifdef ASURA_USE_COLLISIONFX
		Asura_Physics_Ragdoll_CFX_Reporter		m_xCFXReporter;
		#endif

		u_short									m_usFlagsOfPolysToIgnore;

};

class Asura_Physics_Ragdoll_System
{
	public:
		
		// Initialise, Reset and Shutdown functions - to be called by Asura_Physics only.
		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

		static void Update( void );
		static void	DebugRender( void );

		// const accessor used in-game, to get ragdoll description from a hash id
		static const Asura_Physics_Ragdoll_Description*	GetRagdollDescription( const Asura_Hash_ID uHashID );

		// TODO - I think this should be Tools_ only, but we'll need to modify calling fns and check all projects
		static Asura_Physics_Ragdoll_Description*	GetRagdollDescription(const Asura_String& xName);

		// Create a ragdoll with the given description. returns ASURA_INVALID_HANDLE if creation was not successful.
		static Asura_Handle CreateRagdoll( const Asura_Physics_Ragdoll_Description& xRagdollDesc, class Asura_Animation* pxAnim, const Asura_Vector_3& xAnimPos, const Asura_Matrix_3x3& xAnimOri, bool bAnimateInitialPositions = false );

		// Destroy ragdoll via its handle
		static void DestroyRagdoll( const Asura_Handle iHandle );

		// TODO - possibly move this to the class itself?
		static Asura_Physics_Ragdoll_Description* CreateNewRagdollDescription( const Asura_String& xName );

		// get an access pointer for a ragdoll from its handle
		static Asura_Physics_Ragdoll* GetRagdollPointer( const Asura_Handle iHandle )
		{
			return s_xFactory.GetInstancePointer( iHandle );
		}
	
		static u_int GetNumberOfActiveRagdolls( void );

#if defined(ASURA_USE_STREAMING)
		// TODO - much match functionality, and reset environment caches etc
		// MGH 10/4/2006: Zone has been unloaded. Clean up any remaining references to it
		// in the environment cache.
		static void RemoveZone(u_int uZoneIndex) {}
#endif //defined(ASURA_USE_STREAMING)

#ifdef ASURA_TOOLS
		// TODO - rename Tools_ and fix up any calling code
		static int GetNumRagdollDescriptions(  u_int uBoneCountToMatch = 0 );
		static void GetRagdollDescriptionNames( const char** pszNames,  u_int uBoneCountToMatch = 0 );
		//static Asura_Physics_Ragdoll_Description*	CreateNewRagdollDescription(const Asura_String& xName, const u_int uNumberOfAnimationNodes);
		//static Asura_Physics_Ragdoll_Description*	CreateNewRagdollDescriptionFrom(const Asura_String& xName, const Asura_Physics_Ragdoll_Description& pxFrom);
#endif

		static float GetTimeSlice( void ) { return 1.0f/200.0f; }	

private:

		static Asura_Physics_Ragdoll_Factory	s_xFactory;

		typedef Asura_TemplateHashTree<Asura_Physics_Ragdoll_Description> RagdollDescriptionTree;

		static RagdollDescriptionTree s_xRagdollDescriptionTree;// Tree of ragdoll descriptions, accessed via hash-ID.
};

// An instance of a collision object, with all the data that changes in-game
// This may become a generic "rigid body" base class for all systems
class Asura_Physics_Ragdoll_CollisionObject
{
	public:

		Asura_Physics_BodyState* GetBodyState( void )
		{
			return &(m_xBodyState);
		}


		const Asura_AMatrix3x3& GetOrientation( void ) const { return m_xBodyState.m_xOrientation; }

		const Asura_AVector3& GetPosition( void ) const { return m_xBodyState.m_xCentreOfMass; }
		const Asura_AVector3& GetPreviousPosition( void ) const { return m_xBodyState.m_xPreviousCentreOfMass; }

		const Asura_AVector3& GetLinearVelocity( void ) const { return m_xBodyState.m_xLinearVelocity; }
		const Asura_AVector3& GetAngularVelocity( void ) const { return m_xBodyState.m_xAngularVelocity; }

		const Asura_AMatrix3x3& GetWorldInvInertiaTensor( void ) const { return m_xBodyState.m_xWorldInvInertiaTensor; }

		void Integrate( void );

		void ZeroMomentum( void );

		void CalculateWorldInvInertiaTensor( void );

	
		inline void ApplyImpulseAtRelativePosition( const Asura_AVector3& xRel, const Asura_AVector3& xImpulse )
		{
			m_xBodyState.m_xLinearVelocity += xImpulse * GetInvMass();
			m_xBodyState.m_xAngularVelocity += ( xRel ^ xImpulse )* m_xBodyState.m_xWorldInvInertiaTensor;
		}

		inline void ApplyImpulse( const Asura_AVector3& xImpulse )
		{
			m_xBodyState.m_xLinearVelocity += xImpulse * GetInvMass();
		}
	
		inline void CalculateVelocityOfRelativePointOnBody( const Asura_AVector3& xRelPoint, Asura_AVector3& xVelocity) const
		{
			xVelocity = ( m_xBodyState.m_xLinearVelocity + ( m_xBodyState.m_xAngularVelocity ^ (xRelPoint) ));
		}

		inline float GetInvMass( void ) const { return m_pxDesc->GetInvMass(); }

		void FindEnvironmentContacts( Asura_Physics_Ragdoll* pxParentRagdoll );

		void FindObjectContacts( Asura_Physics_Ragdoll* pxParentRagdoll );

		Asura_AVector3 GetSupportPoint( const Asura_AVector3& xAxis ) const;

		Asura_Physics_BodyState					m_xBodyState;
	
		Asura_Physics_ContactCluster_List		m_xContactClusterList;
		Asura_Physics_EnvironmentCache			m_xEnvironmentCache;

		Asura_Bounding_Box						m_xBroadphaseAABB;

		const Asura_Physics_Ragdoll_Desc_CollisionObject* m_pxDesc;
};

#endif // #ifndef ASURA_PHYSICS_DISABLE_RAGDOLL

#endif//__ASURA_PHYSICS_RAGDOLL_H__
