//
// Asura_Physics_Ragdoll_Description.h
//
// Created by KJFL - 14 October 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_RAGDOLL_DESCRIPTION_H__
#define __ASURA_PHYSICS_RAGDOLL_DESCRIPTION_H__

#include "Core/Asura_List.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"

class Asura_Physics_Ragdoll_Desc_CollisionObject;
class Asura_Physics_Ragdoll_Desc_HierarchyNode;
class Asura_Physics_Ragdoll_Desc_Joint;

class Asura_Physics_Ragdoll_Description
{
	public:
		enum { CURRENT_CHUNK_VERSION = 15 };

		// A safe value for an "invalid index" identifier, shared amongst all the description
		// classes (e.g. collision object index, joint object index, hierarchy node index).
		// Avoids magic numbers in code and aids readability.
		enum { uINVALID_INDEX = (0x7fffffff) };

		Asura_Physics_Ragdoll_Description();
		
		~Asura_Physics_Ragdoll_Description();

		// const accessors - data doesn't need to be modified in-game
		u_int GetNumberOfHierarchyNodes() const { return m_uNumberOfHierarchyNodes; }
		u_int GetNumberOfCollisionObjects() const { return m_uNumberOfCollisionObjects; }
		u_int GetNumberOfJoints() const { return m_uNumberOfJoints; }

		const Asura_Physics_Ragdoll_Desc_HierarchyNode& GetHierarchyNode( u_int uIndex ) const;
		const Asura_Physics_Ragdoll_Desc_CollisionObject& GetCollisionObject( u_int uIndex ) const;
		const Asura_Physics_Ragdoll_Desc_Joint& GetJoint( u_int uIndex ) const;

		void ReadFromChunkStream( Asura_Chunk_Stream& xStream, int iVersion );
		void WriteToChunkStream( Asura_Chunk_Stream& xStream) const;

		const Asura_String& GetName( void ) const { return m_xName; }
		void SetName( const Asura_String& xName ) { m_xName = xName; }


		// Tools specific methods, to initialise or change data on the fly
		#if defined (ASURA_TOOLS)

			void Tools_Initialise( u_int uNumberOfHierarchyNodes );

			// non-const accessors for Tools write-access
			Asura_Physics_Ragdoll_Desc_HierarchyNode* Tools_GetHierarchyNode( u_int uIndex );
			Asura_Physics_Ragdoll_Desc_CollisionObject* Tools_GetCollisionObject( u_int uIndex ) ;
			Asura_Physics_Ragdoll_Desc_Joint* Tools_GetJoint( u_int uIndex );
		
			// Methods for adding new elements to the ragdoll description on the fly
			Asura_Physics_Ragdoll_Desc_CollisionObject* Tools_AddCollisionObject( void );
			Asura_Physics_Ragdoll_Desc_Joint* Tools_AddJoint( void );

			// Methods for removing elements from the ragdoll description on the fly.
			void Tools_RemoveCollisionObject( Asura_Physics_Ragdoll_Desc_CollisionObject* pxObjectToRemove );
			void Tools_RemoveJoint( Asura_Physics_Ragdoll_Desc_Joint* pxJointToRemove );
		#endif

	private:

		// The hierarchy nodes, and how they are linked to the collision objects
		u_int										m_uNumberOfHierarchyNodes;
		Asura_Physics_Ragdoll_Desc_HierarchyNode*	m_pxHierarchyNodes;
				
		// The collision objects, and how they are linked to the hierarchy nodes
		u_int										m_uNumberOfCollisionObjects;
		Asura_Physics_Ragdoll_Desc_CollisionObject* m_pxCollisionObjects;

		// The constraints, and how they are linked to the collision objects
		u_int										m_uNumberOfJoints;
		Asura_Physics_Ragdoll_Desc_Joint*			m_pxJoints;

		Asura_String								m_xName;



};

class Asura_Physics_Ragdoll_Desc_HierarchyNode
{
	public:
	
		Asura_Physics_Ragdoll_Desc_HierarchyNode()
		{
			m_uCollisionObject = Asura_Physics_Ragdoll_Description::uINVALID_INDEX;
			m_uOverrideCollisionObject = Asura_Physics_Ragdoll_Description::uINVALID_INDEX;
		}

		void ReadFromChunkStream( Asura_Chunk_Stream& xStream, int iVersion );
		void WriteToChunkStream( Asura_Chunk_Stream& xStream) const;

		// const accessors - data doesn't need to be modified in-game
		const Asura_AMatrix3x3& GetRelativeOrientation( void ) const { return m_xRelativeOrientation; }
		const Asura_AVector3& GetRelativePosition( void ) const { return m_xRelativePosition; }
		u_int GetCollisionObject( void ) const { return m_uCollisionObject; }

		// Tools specific methods, to initialise or change data on the fly
		#if defined( ASURA_TOOLS )
			void Tools_Initialise( const Asura_AMatrix3x3& xRelativeOrientation, const Asura_AVector3& xRelativePosition, u_int uCollisionObject )
			{
				m_xRelativeOrientation = xRelativeOrientation;
				m_xRelativePosition = xRelativePosition;
				m_uCollisionObject = uCollisionObject;
			}	

			void Tools_SetOverrideCollisionObject( u_int uIndex ) { m_uOverrideCollisionObject = uIndex; }
			u_int Tools_GetOverrideCollisionObject( void ) const { return m_uOverrideCollisionObject; }
		#endif // ASURA_TOOLS

	private:

		// Position relative to given collision object
		Asura_AMatrix3x3	m_xRelativeOrientation; 
		Asura_AVector3		m_xRelativePosition;			
		u_int				m_uCollisionObject;
		u_int				m_uOverrideCollisionObject;
};

class Asura_Physics_Ragdoll_Desc_CollisionObject
{
	public:

		enum COLLISION_PROXY
		{
			COLLISION_SPHERE=0,
			COLLISION_BOX=1,
			COLLISION_CAPSULE=2,
		};

		Asura_Physics_Ragdoll_Desc_CollisionObject(){}

		Asura_Physics_Ragdoll_Desc_CollisionObject( const Asura_AMatrix3x3& xRelativeOrientation, const Asura_AVector3& xRelativePosition,	
													const Asura_AVector3& xCollisionExtents, u_int uHierarchyNode,
													enum COLLISION_PROXY eCollisionProxy )
		: m_xRelativeOrientation( xRelativeOrientation )
		, m_xRelativePosition( xRelativePosition )
		, m_xCollisionExtents( xCollisionExtents )
		, m_uHierarchyNode( uHierarchyNode )
		, m_eCollisionProxy( eCollisionProxy )
		, m_fInvMass( 1.0f )
		{
			CalculateInvInertiaTensorDiagonal();
		}

		void ReadFromChunkStream( Asura_Chunk_Stream& xStream, int iVersion );
		void WriteToChunkStream( Asura_Chunk_Stream& xStream) const;

		// const accessors - data doesn't need to be modified in-game
		const Asura_AMatrix3x3& GetRelativeOrientation( void ) const { return m_xRelativeOrientation; }
		const Asura_AVector3& GetRelativePosition( void ) const { return m_xRelativePosition; }
		
		const Asura_AVector3& GetCollisionExtents( void ) const { return m_xCollisionExtents; }

		u_int GetHierarchyNode( void ) const { return m_uHierarchyNode; }

		enum COLLISION_PROXY GetCollisionProxyType( void ) const { return m_eCollisionProxy; }

		const Asura_AVector3& GetInvInertiaTensor( void ) const { return m_xInvInertiaTensorDiagonal; }
			
		float GetInvMass( void ) const { return m_fInvMass; }

		void CalculateInvInertiaTensorDiagonal( void );

		// Tools specific methods, to initialise or change data on the fly
		#if defined (ASURA_TOOLS)
			void Tools_SetCollisionExtents( const Asura_AVector3& xExtents )
			{
				m_xCollisionExtents = xExtents;
				CalculateInvInertiaTensorDiagonal();
			}

			void Tools_SetMass( float fMass )
			{
				m_fInvMass = 1.0f/fMass;
				CalculateInvInertiaTensorDiagonal();
			}
		#endif

	private:

		// Position relative to given hierarchy node
		Asura_AMatrix3x3		m_xRelativeOrientation;
		Asura_AVector3			m_xRelativePosition;

		// Collision extents along axes in object space
		Asura_AVector3			m_xCollisionExtents;

		Asura_AVector3			m_xInvInertiaTensorDiagonal;

		// "Parent" hierarchy node, which determines initial position/orientation
		u_int					m_uHierarchyNode;

		// Collision shape used
		enum COLLISION_PROXY	m_eCollisionProxy;

		// mass in kg
		float					m_fInvMass;

};

class Asura_Physics_Ragdoll_Desc_Joint
{
	public:
		
		Asura_Physics_Ragdoll_Desc_Joint(){}

		Asura_Physics_Ragdoll_Desc_Joint( const Asura_AVector3& xJointPosRelativeToA, const Asura_AVector3& xJointPosRelativeToB, u_int uObjectA, u_int uObjectB, u_int uHierarchyNode )
		: m_xJointPosRelativeToA( xJointPosRelativeToA )
		, m_xJointPosRelativeToB( xJointPosRelativeToB )
		, m_uObjectA( uObjectA )
		, m_uObjectB( uObjectB )
		, m_uHierarchyNode( uHierarchyNode )
		{
			m_xOriginalJointOrientation.Identity();
			m_xJointOrientation.Identity();

			m_xJointSpaceRelativeToA.Identity();
			m_xJointSpaceRelativeToB.Identity();

			m_bUseAngularConstraint = true;
			m_fTwistLimit = 0.0f;
			m_fSwingLimitX = 0.0f;
			m_fSwingLimitY = 0.0f;
		}

		void ReadFromChunkStream( Asura_Chunk_Stream& xStream, int iVersion );
		void WriteToChunkStream( Asura_Chunk_Stream& xStream) const;

		bool IsValid( void ) const 
		{
			return ( m_uObjectA!=Asura_Physics_Ragdoll_Description::uINVALID_INDEX && m_uObjectB!=Asura_Physics_Ragdoll_Description::uINVALID_INDEX );
		}

		// const accessors - data doesn't need to be modified in-game
		u_int GetObjectA( void ) const { return m_uObjectA; }
		u_int GetObjectB( void ) const { return m_uObjectB; }

		// Positions inside that object's space
		const Asura_AVector3& GetPositionRelativeToA( void ) const { return m_xJointPosRelativeToA; }
		const Asura_AVector3& GetPositionRelativeToB( void ) const { return m_xJointPosRelativeToB; }

		// Object->Joint transform
		const Asura_AMatrix3x3& GetJointSpaceRelativeToA( void ) const { return m_xJointSpaceRelativeToA; }
		const Asura_AMatrix3x3& GetJointSpaceRelativeToB( void ) const { return m_xJointSpaceRelativeToB; }

		u_int GetHierarchyNode( void ) const { return m_uHierarchyNode; }


		float GetSwingLimitX( void ) const { return m_fSwingLimitX; }
		float GetSwingLimitY( void ) const { return m_fSwingLimitY; }
		float GetTwistLimit( void ) const { return m_fTwistLimit; }

		bool UseAngularConstraint( void ) const { return m_bUseAngularConstraint; }
		
		// TODO - Tools_ only?
		void SetObjectA( u_int uIndex ) { m_uObjectA = uIndex; }
		void SetObjectB( u_int uIndex ) { m_uObjectB = uIndex; }
		void SetPositionRelativeToA( const Asura_AVector3& xRel ) { m_xJointPosRelativeToA = xRel; }
		void SetPositionRelativeToB( const Asura_AVector3& xRel ) { m_xJointPosRelativeToB = xRel; }
		void SetJointSpaceRelativeToA( const Asura_AMatrix3x3& xMat ) { m_xJointSpaceRelativeToA = xMat; }
		void SetJointSpaceRelativeToB( const Asura_AMatrix3x3& xMat ) { m_xJointSpaceRelativeToB = xMat; }

		void Tools_SetTwistLimit( float fLimit ) { m_fTwistLimit = fLimit; }
		void Tools_SetSwingLimitX( float fLimit ) { m_fSwingLimitX = fLimit; }
		void Tools_SetSwingLimitY( float fLimit ) { m_fSwingLimitY = fLimit; }

		const Asura_AMatrix3x3& GetJointOrientation( void ) const { return m_xJointOrientation; }
		void SetJointOrientation( const Asura_AMatrix3x3& xOri ) { m_xJointOrientation = xOri; }
	
		const Asura_AMatrix3x3& GetOriginalJointOrientation( void ) const { return m_xOriginalJointOrientation; }
		void SetOriginalJointOrientation( const Asura_AMatrix3x3& xOri ) { m_xOriginalJointOrientation = xOri; }

	private:

		// Joint orientations - these are only needed by the Editor really,
		// but it's tricky not to include it here.
		Asura_AMatrix3x3 m_xJointOrientation;
		Asura_AMatrix3x3 m_xOriginalJointOrientation;

		// Joint spacial orientation relative to both collision objects it connects
		Asura_AMatrix3x3 m_xJointSpaceRelativeToA;
		Asura_AMatrix3x3 m_xJointSpaceRelativeToB;

		// Joint position relative to both collision objects it connects
		Asura_AVector3 m_xJointPosRelativeToA;
		Asura_AVector3 m_xJointPosRelativeToB;

		// Collision object indices
		u_int m_uObjectA;
		u_int m_uObjectB;

		// "Parent" hierarchy node on which the joint sits - 
		// possibly won't be need in-game, but needed in ragdoll construction 
		u_int m_uHierarchyNode; 	

		// Angular constraint limits
		float m_fTwistLimit;
		float m_fSwingLimitX;
		float m_fSwingLimitY;

		bool m_bUseAngularConstraint;
};

#endif//__ASURA_PHYSICS_RAGDOLL_DESCRIPTION_H__

