//
// Asura_Physics_RigidBody.cpp
//
// Created by KJFL - 2004-03-17 11:23:03
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_RigidBody.h"

#include "Physics/Asura_Physics_Broadphase.h"
#include "Physics/Asura_Physics_CCD.h"
#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
#include "Physics/Asura_Physics_Island.h"
#endif
#include "Physics/Asura_Physics_RigidBodyDefinition.h"
#include "Physics/Asura_Physics_CollisionMesh.h"
#ifdef ASURA_USE_COLLISIONFX
#include "Physics/Asura_Physics_CollisionFX_Report.h"
#endif
#include "Physics/Asura_Physics_CollisionReport.h"
#include "Physics/Asura_Physics_Environment.h"
#include "Physics/Asura_Physics_EnvironmentCache.h"

#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_Pairing.h"
#include "Physics/Asura_Physics_RestFrame.h"

#include "Rendering/Asura_Render.h"

#include "Mesh/Asura_Mesh_PolygonFlags.h" 

#include "Debugging/Asura_Logfile.h"

#ifdef ASURA_ASSERT
#include "Entity/Asura_Entity.h"
#include "Network/Asura_Network.h"
#endif

// debugging rendering values
DEBUGVAR bool dbg_bDrawEdges = true;

// The maximum number of times in one update you resolve your motion against normals of obstructions
static const u_int uMAX_NUMBER_OF_ITERATIONS_FOR_ACTOR_MOVEMENT = 10;

static const float fTWIST_FRICTION_SCALE = 0.10f;

static const float fMOTIONLESS_SPEED_THRESHOLD = 0.005f;
static const float fMOTIONLESS_ANGULAR_SPEED_THRESHOLD = 0.05f;

static const float fMOTIONLESS_TIME_THRESHOLD = 1.0f;
static const float fAXIALLYALIGNED_DIST_THRESHOLD = 0.005f;
static const float fAXIALLYALIGNED_MOVEMENT_THRESHOLD = 0.00001f;

static const float fVERTICAL_BOUNDARY = 0.70710678f; // = cosine(45degrees)
static const float fEDGE_AS_FLOOR_THRESHOLD = 0.9848f; // = cosine(10degrees)

static const float fDEFAULT_AIR_DRAG = 0.999f;
static const float fDEFAULT_ELASTICITY = 0.5f;
static const float fDEFAULT_FRICTION = 1.0f;
static const float fDEFAULT_ANGULAR_DRAG = 0.99f;

static const float fSELF_RIGHTING_CONSTRAINT = 0.1950903f; // sine(  22.5 / 2 )

static const float fASURA_PHYSICS_ACTOR_MOVEMENT_DEADZONE = 0.0004f; //( 2 cm at 50 physics updates per sec)

static const float fACTOR_SLIDE_OFF_MOVEMENT_OVERRIDE_SPEED = 0.025f;
static const float fACTOR_SLIDE_OFF_JUMP_SPEED_LIMIT = 10.0f;

DEBUGVAR bool dbg_bOnlyUseLinearProjection = false;

DEBUGVAR float dbg_ActorVehicle_GroundingSpeed = 0.005f;
DEBUGVAR float dbg_ActorVehicle_GroundTestDistanceDown = 0.1f;
DEBUGVAR float dbg_ActorVehicle_GroundTestDistanceUp = dbg_ActorVehicle_GroundTestDistanceDown + 0.001f;
DEBUGVAR float dbg_ActorVehicle_GroundAlignmentDeadZone = 0.0001f;

static void Render4VectorLine( const Asura_AVector3& xV0, const Asura_AVector3& xV1, u_int uColour )
{
	Asura_Render::RenderLine( xV0, xV1, uColour );
}

static void RenderEllipsoid( const Asura_AVector3& xPosition,  float fRadiusX,  float fRadiusY, const u_int uARGB, const Asura_AMatrix3x3* pxOrientation = 0 )
{
	const int		iNumSegments	= 16;
	const int		iNumSlices		= 8;
	Asura_AVector3*	pxVertices		= new Asura_AVector3 [iNumSegments * (iNumSlices + 1)];
	Asura_AVector3*	pxVertex		= pxVertices;

	// Calculate the vertices one slice at a time...
	for (int iSlice = 0; iSlice <= iNumSlices; iSlice++)
	{
		const Asura_Angle	xSliceAngle		= (Asura_Maths::Deg180 * static_cast<float>(iSlice)) / iNumSlices;
		const float			fSliceRadius	= fRadiusX * Asura_Maths::Sine(xSliceAngle);
		const float fY = (fRadiusY * Asura_Maths::Cosine(xSliceAngle));

		// ...and one segment at a time...
		for (int iSegment = 0; iSegment < iNumSegments; iSegment++)
		{
			const Asura_Angle	xAngle	= (Asura_Maths::Deg360 * static_cast<float>(iSegment)) / iNumSegments;
			const float fX = (fSliceRadius * Asura_Maths::Sine(xAngle));
			const float fZ = (fSliceRadius * Asura_Maths::Cosine(xAngle));
	
			*pxVertex = Asura_AVector3( fX, fY, fZ );

			if (pxOrientation)
			{
				*pxVertex = (*pxVertex)*(*pxOrientation);
			}
			
			*pxVertex += xPosition;

			// Next vertex...
			pxVertex++;
		}
	}

	// Render the slices...
	for (int iSlice = 1; iSlice < iNumSlices; iSlice++)
	{
		// Go through each segement...
		for (int iSegment = 0; iSegment < iNumSegments; iSegment++)
		{
			const int	iStartVertex	= (iSlice * iNumSegments) + iSegment;
			const int	iEndVertex		= (iSlice * iNumSegments) + ((iSegment + 1) % iNumSegments);

			// Render this line...
			Render4VectorLine(pxVertices [iStartVertex], pxVertices [iEndVertex], uARGB);
		}
	}

	// Render the segments...
	for (int iSegment = 0; iSegment < iNumSegments; iSegment++)
	{
		// Go through each slice...
		for (int iSlice = 0; iSlice < iNumSlices; iSlice++)
		{
			const int	iStartVertex	= (iSlice * iNumSegments) + iSegment;
			const int	iEndVertex		= ((iSlice + 1) * iNumSegments) + iSegment;

			// Render this line...
			Render4VectorLine(pxVertices [iStartVertex], pxVertices [iEndVertex], uARGB);
		}
	}

	// Finally, destroy the vertices...
	delete []	(pxVertices);
}

#if 01
static void RenderContactPoint( const Asura_AVector3& xCentre, const Asura_AVector3& xNormal )
{
	const Asura_AVector3 xOffsetX(0.025f, 0.0f, 0.0f);
	const Asura_AVector3 xOffsetY(0.0f, 0.025f, 0.0f);
	const Asura_AVector3 xOffsetZ(0.0f, 0.0f, 0.025f);
	{
		Asura_AVector3 xV0 = xCentre;
		Asura_AVector3 xV1 = xCentre;
		xV0 -= xOffsetX;
		xV1 += xOffsetX;
		Render4VectorLine( xV0, xV1, 0xffffff00);
	}
	{
		Asura_AVector3 xV0 = xCentre;
		Asura_AVector3 xV1 = xCentre;
		xV0 -= xOffsetY;
		xV1 += xOffsetY;
		Render4VectorLine( xV0, xV1, 0xffffff00);
	}
	{
		Asura_AVector3 xV0 = xCentre;
		Asura_AVector3 xV1 = xCentre;
		xV0 -= xOffsetZ;
		xV1 += xOffsetZ;
		Render4VectorLine( xV0, xV1, 0xffffff00);
	}

	{
		Asura_AVector3 xV0 = xCentre;
		Asura_AVector3 xV1 = xCentre + xNormal * 0.1f;
		Render4VectorLine( xV0, xV1, 0xffffff00);
	}
}
#endif

const u_int uINSTANCES_PER_POOL = 8;
Asura_TrackingHandlePool<Asura_Physics_RigidBodyInstance>	Asura_Physics_RigidBodyInstance::s_xInstancePool( uINSTANCES_PER_POOL );

// collision report mempools
Asura_ResizingMempool g_xObjectCollisionReportsPool( 16, sizeof(Asura_Physics_ObjectCollisionReport), __FILE__, __LINE__ );
Asura_ResizingMempool g_xEnvironmentCollisionReportsPool( 16, sizeof(Asura_Physics_EnvironmentCollisionReport), __FILE__, __LINE__ );

Asura_Physics_RigidBodyInstance*	Asura_Physics_RigidBodyInstance::s_pxDynamicList = 0;
Asura_Physics_RigidBodyInstance*	Asura_Physics_RigidBodyInstance::s_pxExternallyControlledList = 0;
Asura_Physics_RigidBodyInstance*	Asura_Physics_RigidBodyInstance::s_pxStaticList = 0;


#if 0 // work in progress test stuff
/*

static inline bool TestAxis( Asura_OBB& xOBB, Asura_Physics_Tri& xTri, Asura_AVector3& xAxis )
{
	// Use ray vs Minkowski sum to test for space-time intersection of objects

	float fCentreA = 0.0f;
	float fExtentsA = 0.0f;

	xOBB.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreA, fExtentsA );

	float fCentreB = 0.0f;
	float fExtentsB = 0.0f;

	xTri.GetCentreAndHalfExtentAlongAxis( xAxis, fCentreB, fExtentsB );

	const float fMinkowskiSumMin = fCentreA - fCentreB - fExtentsA - fExtentsB;
	const float fMinkowskiSumMax = fCentreA - fCentreB + fExtentsA + fExtentsB;

	const float fRayMin = 0.0f;
	const float fRayMax = 0.0f;

	// test for a no overlap case
	if ( fRayMax < fMinkowskiSumMin || fRayMin > fMinkowskiSumMax )
	{
		// these objects definitely don't overlap
		return false;
	}
	return true;
}

bool DoesOBBIntersectTri_Opt( Asura_OBB& xOBB, Asura_Physics_Tri& xTri )
{
	const Asura_AMatrix3x3& xObjectOrient = xOBB.GetOrientation();

	// KJFL - 27 September 2006 - 
	// 
	// Let's see... there are 13 potential axes of separation we need to test:
	//
	// 1 face normal from the tri
	// 3 unique face normals from the OBB (the other three faces are just sign flips)
	// 9 unique axes given by the cross products of combinations of the edges of both objects

	const unsigned int uNumberOfSeparatingAxes = 13;

	Asura_AVector3 xTriNormal, xTriEdge0, xTriEdge1, xTriEdge2;

	xTri.GetNormal( xTriNormal );
	xTri.GetEdge0( xTriEdge0 );
	xTri.GetEdge1( xTriEdge1 );
	xTri.GetEdge2( xTriEdge2 );

	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xObjectOrient.GetColumn0() ^ xTriEdge0,
		xObjectOrient.GetColumn0() ^ xTriEdge1,
		xObjectOrient.GetColumn0() ^ xTriEdge2,

		xObjectOrient.GetColumn1() ^ xTriEdge0,
		xObjectOrient.GetColumn1() ^ xTriEdge1,
		xObjectOrient.GetColumn1() ^ xTriEdge2,

		xObjectOrient.GetColumn2() ^ xTriEdge0,
		xObjectOrient.GetColumn2() ^ xTriEdge1,
		xObjectOrient.GetColumn2() ^ xTriEdge2,
		xObjectOrient.GetColumn0(),
		xObjectOrient.GetColumn1(),
		xObjectOrient.GetColumn2(),
	
		xTriNormal,
	};

	//float afSeparationDistance[uNumberOfSeparatingAxes];
	//u_int uBestAxis = uNumberOfSeparatingAxes;

	//const Asura_AVector3 xAToB = xTri.GetCentre() - xOBB.GetCentre();

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
//		afSeparationDistance[u] = Asura_Maths::LargeFloat / 4.0f;

		Asura_AVector3& xAxis = axAxes[u];

		// KJFL - 27 June 2006 - ah, it's possible some of the axes are invalid
		// because they're zero vectors, simply because the edges of the two 
		// objects might be parallel (cross products above would give zero vectors).

		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		//if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

	//	xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects
		if (!TestAxis( xOBB, xTri, xAxis )) return false;

		//afSeparationDistance[u] = Asura_Maths::Min( fMinkowskiSumMax, -fMinkowskiSumMin );

	}

#if 0
	float fMinimumDistance = Asura_Maths::LargeFloat;

	{
		uBestAxis = uNumberOfSeparatingAxes;

		for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
		{
			// skip invalid axes (e.g. check for zero vectors)
			if ( axAxes[u].MagnitudeSquared() < 0.001f ) continue;

			const float f = afSeparationDistance[u];

			if ( f > 0.0f && f < fMinimumDistance )
			{
				uBestAxis = u;
				fMinimumDistance = f;
			}
		}
	}

	if ( uBestAxis >= uNumberOfSeparatingAxes  )
	{
		// this is unexpected... let's assert and return a verdict of no collision
		Asura_Assert( uBestAxis < uNumberOfSeparatingAxes, "Invalid axis" );

		return false;
	}
#endif

	// overlap found
	return true;
}

bool DoesAABBIntersectTri_Opt( Asura_OBB& xOBB, Asura_Physics_Tri& xTri )
{
	//const Asura_AMatrix3x3& xObjectOrient = xOBB.GetOrientation();

	// KJFL - 27 September 2006 - 
	// 
	// Let's see... there are 13 potential axes of separation we need to test:
	//
	// 1 face normal from the tri
	// 3 unique face normals from the OBB (the other three faces are just sign flips)
	// 9 unique axes given by the cross products of combinations of the edges of both objects

	const unsigned int uNumberOfSeparatingAxes = 13;

	Asura_AVector3 xTriNormal, xTriEdge0, xTriEdge1, xTriEdge2;

	xTri.GetNormal( xTriNormal );
	xTri.GetEdge0( xTriEdge0 );
	xTri.GetEdge1( xTriEdge1 );
	xTri.GetEdge2( xTriEdge2 );

	Asura_AVector3 xXAxis(1.0f,0.0f,0.0f);
	Asura_AVector3 xYAxis(0.0f,1.0f,0.0f);
	Asura_AVector3 xZAxis(0.0f,0.0f,1.0f);

	Asura_AVector3 axAxes[uNumberOfSeparatingAxes] = 
	{
		xXAxis ^ xTriEdge0,
		xXAxis ^ xTriEdge1,
		xXAxis ^ xTriEdge2,

		xYAxis ^ xTriEdge0,
		xYAxis ^ xTriEdge1,
		xYAxis ^ xTriEdge2,

		xZAxis ^ xTriEdge0,
		xZAxis ^ xTriEdge1,
		xZAxis ^ xTriEdge2,
		
		xXAxis,
		xYAxis,
		xZAxis,

		xTriNormal,
	};

	//float afSeparationDistance[uNumberOfSeparatingAxes];
	//u_int uBestAxis = uNumberOfSeparatingAxes;

	//const Asura_AVector3 xAToB = xTri.GetCentre() - xOBB.GetCentre();

	for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
	{
		//		afSeparationDistance[u] = Asura_Maths::LargeFloat / 4.0f;

		Asura_AVector3& xAxis = axAxes[u];

		// KJFL - 27 June 2006 - ah, it's possible some of the axes are invalid
		// because they're zero vectors, simply because the edges of the two 
		// objects might be parallel (cross products above would give zero vectors).

		// skip invalid axes (e.g. check for zero vectors)
		if ( xAxis.MagnitudeSquared() < 0.001f ) continue;

		// Arbitrarily make the axis point towards box A.
		//if ( xAxis*xAToB > 0.0f ) xAxis = -xAxis;

		//	xAxis.Normalise();

		// Use ray vs Minkowski sum to test for space-time intersection of objects
		if (!TestAxis( xOBB, xTri, xAxis )) return false;

		//afSeparationDistance[u] = Asura_Maths::Min( fMinkowskiSumMax, -fMinkowskiSumMin );

	}

#if 0
	float fMinimumDistance = Asura_Maths::LargeFloat;

	{
		uBestAxis = uNumberOfSeparatingAxes;

		for ( u_int u = 0; u<uNumberOfSeparatingAxes; ++u )
		{
			// skip invalid axes (e.g. check for zero vectors)
			if ( axAxes[u].MagnitudeSquared() < 0.001f ) continue;

			const float f = afSeparationDistance[u];

			if ( f > 0.0f && f < fMinimumDistance )
			{
				uBestAxis = u;
				fMinimumDistance = f;
			}
		}
	}

	if ( uBestAxis >= uNumberOfSeparatingAxes  )
	{
		// this is unexpected... let's assert and return a verdict of no collision
		Asura_Assert( uBestAxis < uNumberOfSeparatingAxes, "Invalid axis" );

		return false;
	}
#endif

	// overlap found
	return true;
}

bool DoesAABBIntersectTri( const Asura_Bounding_Box& xAABB, u_int uTri )
{
	Asura_OBB xOBB( xAABB );
	Asura_Physics_Tri xTri(uTri); 

	return DoesOBBIntersectTri( xOBB, xTri);
}
bool DoesAABBIntersectTri_Opt( const Asura_Bounding_Box& xAABB, u_int uTri )
{
	//Asura_OBB xOBB( xAABB );
	//Asura_Physics_Tri xTri(uTri); 

	return false;
//	return DoesAABBIntersectTri_Opt( xOBB, xTri);
}
*/
#endif // work in progress test stuff

void Asura_Physics_RigidBodyInstance::Initialise( void )
{
	s_pxDynamicList = 0;
	s_pxExternallyControlledList = 0;
	s_pxStaticList = 0;
}

void Asura_Physics_RigidBodyInstance::Reset( void )
{
	Asura_Assert( s_pxDynamicList == 0, "Code bug: list error" );
	Asura_Assert( s_pxExternallyControlledList == 0, "Code bug: list error" );
	Asura_Assert( s_pxStaticList == 0, "Code bug: list error" );

	s_xInstancePool.Reset();

	ResetAllCollisionReports();
}


void Asura_Physics_RigidBodyInstance::Shutdown( void )
{
    Reset();
}

void Asura_Physics_RigidBodyInstance::ResetAllCollisionReports( void)
{
	g_xObjectCollisionReportsPool.ForceReset();
	g_xEnvironmentCollisionReportsPool.ForceReset();
}


void Asura_Physics_RigidBodyInstance::InitialiseInstanceAfterCreation( Asura_Handle iHandle )
{
	m_iHandle = iHandle;
	m_uEntityGuid = ASURA_GUID_UNREGISTERED;

	m_fInvMass = 1.0f;

	m_xBodyState.m_xWorldInvInertiaTensor.Zero();
	m_xInvInertiaTensorDiagonal.Zero();
	m_xBodyState.m_xAngularVelocity.Zero();
	m_xBodyState.m_xLinearVelocity.Zero();
	
	m_xBodyState.m_xCentreOfMass.Zero();
	m_xBodyState.m_xPreviousCentreOfMass.Zero();


	m_xInstancePosition.Zero();
	
	m_xBoundingBox.Reset();

	m_xBodyState.m_xOrientation.Identity();

	m_pxObjectVertexPosition = 0;
	m_pxObjectVertexPrevPosition = 0;

	m_pxDefinition = 0; // no definition to begin with - they have to set the collision mesh

	m_xEnvironmentCache.Reset();

	m_fTimeMotionless = 0.0f;
	m_bAsleep = true;
	m_bExternallyControlled = false;

	m_bHasHitSomething = false;
	m_bHasHitAnObject = false;
	m_bIsSupported = false;

	m_bIgnore = false;
	m_iHandleToIgnore = ASURA_INVALID_HANDLE;

	m_xForceEffect.Zero();
	m_xJumpVelocity.Zero();
	m_fGravityScale = 1.0f;
	m_xOwnGravityDirection = Asura_AVector3(0.0f, 1.0f, 0.0f);
	m_fElasticity = fDEFAULT_ELASTICITY;
	m_fFriction = fDEFAULT_FRICTION;
	m_fDrag = fDEFAULT_AIR_DRAG;
	m_fAngularDrag = fDEFAULT_ANGULAR_DRAG;

	m_bAxiallyAligned = false;
	m_bStatic = false;
	m_bActor = false;
	m_bActor_MoveAsVehicle = false;
	m_bActor_DontPushActors = false;
	m_bOwnsRestFrame = false;

	m_eCollisionProxy = COLLISION_PROXY_OBB;

	m_bLockYAxisToBeVertical = false;
	m_bSelfRighting = false;
	m_bIgnoreDynamicObjects = false;
	m_bIgnoreExternalControlObjects = false;

	m_bFeelsGravity = true;
	m_bUseOwnGravityDirection = false;
	m_bActor_SuperSlidy = true;
	m_bActor_TreatSteepSlopesAsVertical = false;
	m_bActorsSlideOffMe = false;
	m_bActorsDontPushMe = false;

	m_bReportObjectCollisions = false;
	m_bReportEnvironmentCollisions = false;
	m_bReportObjectCollisionFX = false;
	m_bReportEnvironmentCollisionFX  = false;
	m_bReportedEnvCollisionFXThisFrame = false;

	m_usPolyFlagsToReturnForActorCollision = 0;
	m_usNotWalkableSurfaceFlags = 0;
	m_bAllSurfacesWalkable = false;

	ClearCollisionReports();

#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
	Asura_Assert( m_xConstraintList.GetList() == 0, "Code bug: constraint list not empty on object creation" );
#endif

	m_pxRestFrame = 0;

	m_uCollisionGroupsBelongedTo = 0xffffffff;
	m_uCollisionGroupsToCollideWith = 0xffffffff;

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
    m_pxBackup = NULL;
#endif
}

void Asura_Physics_RigidBodyInstance::SetInvMass( float fInvMass )
{
	Asura_Assertf(fInvMass > 0.0f, ("Setting rigid body instance (belonging to entity %d) mass to infinite. There should no longer be any reason to do this.", m_uEntityGuid));
	m_fInvMass = fInvMass;
	CalculateInvInertiaTensorDiagonal();
}

void Asura_Physics_RigidBodyInstance::CalculateInvInertiaTensorDiagonal( void )
{
	if (!GetDefinition())
	{
		m_xInvInertiaTensorDiagonal.Zero();
		CalculateWorldInvInertiaTensor();
		return;
	}

    const float fInvMass = GetInvMass();

#if 0 // use the inertia tensor of the verlet tetrahedron
	{
		float fOneOverNumberOfVertices = 1.0f / static_cast<float>(GetNumberOfBaseVertices());

		Asura_AVector3 xCentre;
		
		xCentre.Zero();

		if (fInvMass)
		{
			for ( u_int i=0; i<GetNumberOfBaseVertices(); ++i )
			{
				xCentre +=GetDefinition()-> m_pxBaseVertices[i].m_xPosition;
			}
			xCentre *= fOneOverNumberOfVertices;

			float fInertiaXX = 0.0000001f;
			float fInertiaYY = 0.0000001f;
			float fInertiaZZ = 0.0000001f;
	
			for ( u_int i=0; i<GetNumberOfBaseVertices(); ++i )
			{
				Asura_AVector3 xRel = GetDefinition()->m_pxBaseVertices[i].m_xPosition - xCentre;

				fInertiaXX += xRel.GetY()*xRel.GetY() + xRel.GetZ()*xRel.GetZ();
				fInertiaYY += xRel.GetX()*xRel.GetX() + xRel.GetZ()*xRel.GetZ();
				fInertiaZZ += xRel.GetX()*xRel.GetX() + xRel.GetY()*xRel.GetY();
			}

			m_xInvInertiaTensorDiagonal.SetX( 1.0f / fInertiaXX );
			m_xInvInertiaTensorDiagonal.SetY( 1.0f / fInertiaYY );
			m_xInvInertiaTensorDiagonal.SetZ( 1.0f / fInertiaZZ );
		}
	}
#else 

	// We're going to get infinities in our inverse interial values if we allow tiny->zero extents.
	// Cap things to 1cm radius.
	const float fMIN_EXTENT = 0.05f; 

	switch ( m_eCollisionProxy )
	{
		case COLLISION_PROXY_SPHERE:
		{
			// Completely symmetric so comes down to calculating a scalar value:
			// I = (2/5)(Mr*r)
			// 1/I = (5/2)*(1/M)*(1/(r*r))

			// inelegant way of getting radius atm!
			Asura_Physics_SphereGroup xSphereStack( *this );
			const float fSphereRadius = Asura_Maths::Max( xSphereStack.GetSphereRadius(), fMIN_EXTENT );
			const float fSphereRadiusSquared = fSphereRadius * fSphereRadius;

			const float fInvInertia = 2.5f * fInvMass / fSphereRadiusSquared;
			
			m_xInvInertiaTensorDiagonal = Asura_AVector3( fInvInertia, fInvInertia, fInvInertia );
	
			break;
		}
		default: // default to the OBB, e.g. for polygonal meshes or sphere stacks
		case COLLISION_PROXY_OBB:
		{
			const float fMin = fMIN_EXTENT;

			const float fX = Asura_Maths::Max( GetDefinition()->GetOBBHalfExtentsX(), fMin );
			const float fY = Asura_Maths::Max( GetDefinition()->GetOBBHalfExtentsY(), fMin );
			const float fZ = Asura_Maths::Max( GetDefinition()->GetOBBHalfExtentsZ(), fMin );
		
			const float fXSquared = fX*fX;
			const float fYSquared = fY*fY;
			const float fZSquared = fZ*fZ;

			// usually quoted with a 12.0f constant, but using width, height, depth - we're using a half-extents 
			// so there is a factor of (2.0f*2.0f) to divide out, which gives us a 3.0f.
			float fMassConst = 3.0f * fInvMass;

			float fInvIX = fMassConst / ( fYSquared + fZSquared );
			float fInvIY = fMassConst / ( fXSquared + fZSquared );
			float fInvIZ = fMassConst / ( fXSquared + fYSquared );

			m_xInvInertiaTensorDiagonal = Asura_AVector3( fInvIX, fInvIY, fInvIZ );

			break;
		}

	}
#endif
	CalculateWorldInvInertiaTensor();
}

void Asura_Physics_RigidBodyInstance::CalculateWorldInvInertiaTensor( void )
{
	// If you're not going to rotate when something hits you, let's just cut to the chase 
	// and zero the matrix.
	if ( ( IsActor() && !GetActorMoveAsVehicle() ) || IsExternallyControlled() || IsStatic() )
	{
		m_xBodyState.m_xWorldInvInertiaTensor.Zero();
	}
	else
	{
		Asura_AMatrix3x3 xM( m_xInvInertiaTensorDiagonal.GetX(), 0.0f, 0.0f,
							0.0f,	m_xInvInertiaTensorDiagonal.GetY(), 0.0f,
							0.0f, 0.0f, m_xInvInertiaTensorDiagonal.GetZ() );

		Asura_AMatrix3x3 xO = m_xBodyState.m_xOrientation;
		xO.Transpose();

		m_xBodyState.m_xWorldInvInertiaTensor = (xO*xM)*m_xBodyState.m_xOrientation;
	}
}

const Asura_AMatrix3x3& Asura_Physics_RigidBodyInstance::GetInvInertiaTensorDiagonal( void ) const
{
	return m_xBodyState.m_xWorldInvInertiaTensor;
}
	

Asura_Handle Asura_Physics_RigidBodyInstance::CreateDynamicInstance( void )
{
	Asura_Handle iHandle = s_xInstancePool.Allocate();

	if (iHandle == ASURA_INVALID_HANDLE)
	{
		return iHandle;
	}

	Asura_Physics_RigidBodyInstance* pxInstance = s_xInstancePool.GetPointer(iHandle);
	
	if (!pxInstance)
	{
		Asura_Assert( pxInstance, "Out of memory?" );
		return ASURA_INVALID_HANDLE;
	}

	pxInstance->AddToList( &s_pxDynamicList );

	pxInstance->InitialiseInstanceAfterCreation( iHandle );

	return iHandle;
}

Asura_Handle Asura_Physics_RigidBodyInstance::CreateStaticInstance( void )
{
	Asura_Handle iHandle = s_xInstancePool.Allocate();

	if (iHandle == ASURA_INVALID_HANDLE)
	{
		return iHandle;
	}

	Asura_Physics_RigidBodyInstance* pxInstance = s_xInstancePool.GetPointer(iHandle);
	
	if (!pxInstance)
	{
		Asura_Assert( pxInstance, "Out of memory?" );
		return ASURA_INVALID_HANDLE;
	}

	pxInstance->AddToList( &s_pxStaticList );

	pxInstance->InitialiseInstanceAfterCreation( iHandle );

	pxInstance->m_bStatic = true;

	// tell the physics system that there's a new static object - it will want to recalc spatial hash
	Asura_Physics_System::NotifyOfChangeToStaticObjects();

	return iHandle;
}

Asura_Handle Asura_Physics_RigidBodyInstance::CreateExternallyControlledInstance( void )
{
	Asura_Handle iHandle = s_xInstancePool.Allocate();

	if (iHandle == ASURA_INVALID_HANDLE)
	{
		return iHandle;
	}

	Asura_Physics_RigidBodyInstance* pxInstance = s_xInstancePool.GetPointer(iHandle);
	
	if (!pxInstance)
	{
		Asura_Assert( pxInstance, "Out of memory?" );
		return ASURA_INVALID_HANDLE;
	}

	pxInstance->AddToList( &s_pxExternallyControlledList );

	pxInstance->InitialiseInstanceAfterCreation( iHandle );

	pxInstance->m_bExternallyControlled = true;

	return iHandle;
}


Asura_Handle Asura_Physics_RigidBodyInstance::CreateVehicle( const Asura_Bounding_Box& xBB )
{
	Asura_Assert( false, "Code bug: this method is deprecated" );
	return ASURA_INVALID_HANDLE;
}

void Asura_Physics_RigidBodyInstance::Destroy( const Asura_Handle iHandle )
{
	Asura_Physics_RigidBodyInstance* pxInstance = GetInstancePointer( iHandle );

	if (!pxInstance) return;

	pxInstance->RemoveFromCurrentList();

	// check for any rest frames

	if ( pxInstance->m_pxRestFrame )
	{
		if ( pxInstance->HasOwnRestFrame() )
		{
			Asura_Physics_RestFrame::Destroy( pxInstance->m_pxRestFrame );
		}
		else
		{
			pxInstance->UnlinkFromRestFrame();
		}
	}

	// destroy associated instance data (e.g. pairings, contact points, etc)
	{
		// tell any pairs that we were in that we no longer exist
		Asura_Physics_Pairing::RemovePairingsUsingInstance( pxInstance );

		if ( pxInstance->m_pxDefinition )
		{
			// definitions may have a separate life and not be created/destroyed in the same way,
			// i.e. act as premade templates
			delete pxInstance->m_pxDefinition;
			pxInstance->m_pxDefinition = 0;
		}


		// delete things that aren't in mempools but probably should be...
		delete [] pxInstance->m_pxObjectVertexPosition;
		delete [] pxInstance->m_pxObjectVertexPrevPosition;

		pxInstance->m_pxObjectVertexPosition = 0;
		pxInstance->m_pxObjectVertexPrevPosition = 0;
	}

	if ( pxInstance->IsStatic() )
	{
		// tell the physics system that there's been a change to static objects - it will want to recalc spatial hash
		Asura_Physics_System::NotifyOfChangeToStaticObjects();
	}

#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
	// destroy references to any constraints
	pxInstance->m_xConstraintList.Reset();
#endif

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
    pxInstance->DeallocateBackup();
#endif

	// destroy the instance
	return s_xInstancePool.Deallocate(iHandle);

}

void Asura_Physics_RigidBodyInstance::FlushCachedData( void )
{
	m_xContactClusterList.Reset();

	m_xEnvironmentCache.Flush();

	// we've got to forget any collision reports, since if we're out of the update loops
	// by being asleep or static etc, this pointer will get out-of-date.
	ClearCollisionReports();
}

void Asura_Physics_RigidBodyInstance::FlushCachedDataForAllInstances()
{
	Asura_TrackingHandlePool_Iterator<Asura_Physics_RigidBodyInstance> xIt = s_xInstancePool.GetIterator();
	for(xIt.Begin(); !xIt.End(); xIt.Next())
	{
		xIt.GetCurrentPointer()->FlushCachedData();
	}
}


Asura_AVector3 Asura_Physics_RigidBodyInstance::GetHalfExtents( void ) const
{
	if (!GetDefinition()) return Asura_AVector3( 0.0f, 0.0f, 0.0f );

	return Asura_AVector3( GetDefinition()->GetOBBHalfExtentsX(), GetDefinition()->GetOBBHalfExtentsY(), GetDefinition()->GetOBBHalfExtentsZ() );
}

u_int Asura_Physics_RigidBodyInstance::GetSupportFeature( Asura_AVector3* pxSupportPoints, const Asura_AVector3& xAxis ) const
{
	switch ( GetCollisionProxy() )
	{
		case COLLISION_PROXY_SPHERE:
		{
			pxSupportPoints[0] = GetCollisionMeshCentre() - xAxis * GetDefinition()->GetMaximumRadius();
			return 1;
		}

		case COLLISION_PROXY_SPHERESTACK:
		{
			pxSupportPoints[0] = GetCollisionMeshCentre() - xAxis * GetDefinition()->GetMaximumRadius();

			Asura_AVector3 xCapsuleAxis;
			float fRadius;
			GetDefinition()->GetSphereStackInfo( xCapsuleAxis, fRadius );
		
			float fHalfHeight = xCapsuleAxis.NormaliseReturningMagnitude();

			xCapsuleAxis = xCapsuleAxis * GetOrientation(); 

			Asura_AVector3 xOffset = xCapsuleAxis * (fHalfHeight-fRadius);

			pxSupportPoints[0] = GetCollisionMeshCentre() - xAxis * fRadius;
	
			if ( xOffset * xAxis < 0.0f )
			{
				pxSupportPoints[0] += xOffset;
			}
			else
			{
				pxSupportPoints[0] -= xOffset;
			}

			return 1;
		}

		default:
		case COLLISION_PROXY_OBB:
		{
			const float fX = -xAxis * GetOrientation().GetColumn0(); 
			const float fY = -xAxis * GetOrientation().GetColumn1(); 
			const float fZ = -xAxis * GetOrientation().GetColumn2(); 

	const Asura_AVector3 xCentre = GetCollisionMeshCentre();

			const Asura_AVector3 xXOffset = GetOrientation().GetColumn0() * GetDefinition()->GetOBBHalfExtentsX();
			const Asura_AVector3 xYOffset = GetOrientation().GetColumn1() * GetDefinition()->GetOBBHalfExtentsY();
			const Asura_AVector3 xZOffset = GetOrientation().GetColumn2() * GetDefinition()->GetOBBHalfExtentsZ();

	// first check to see if we have supporting faces - this happens when xAxis is very nearly parallel with an edge.
	const float fFaceDotThreshold = (0.9999f);

	if ( Asura_Maths::Abs(fX) > fFaceDotThreshold ) // support feature is an X face
	{
		Asura_AVector3 xCentreOfFace = xCentre + xXOffset * Asura_Maths::Sign( fX );

		pxSupportPoints[0] = xCentreOfFace + xYOffset + xZOffset;
		pxSupportPoints[1] = xCentreOfFace + xYOffset - xZOffset;
		pxSupportPoints[2] = xCentreOfFace - xYOffset - xZOffset;
		pxSupportPoints[3] = xCentreOfFace - xYOffset + xZOffset;

		return 4;
	}
	else if ( Asura_Maths::Abs(fY) > fFaceDotThreshold ) // support feature is an Y face
	{
		Asura_AVector3 xCentreOfFace = xCentre + xYOffset * Asura_Maths::Sign( fY );

		pxSupportPoints[0] = xCentreOfFace + xXOffset + xZOffset;
		pxSupportPoints[1] = xCentreOfFace + xXOffset - xZOffset;
		pxSupportPoints[2] = xCentreOfFace - xXOffset - xZOffset;
		pxSupportPoints[3] = xCentreOfFace - xXOffset + xZOffset;

		return 4;
	}
	else if ( Asura_Maths::Abs(fZ) > fFaceDotThreshold ) // support feature is an Z face
	{
		Asura_AVector3 xCentreOfFace = xCentre + xZOffset * Asura_Maths::Sign( fZ );

		pxSupportPoints[0] = xCentreOfFace + xYOffset + xXOffset;
		pxSupportPoints[1] = xCentreOfFace + xYOffset - xXOffset;
		pxSupportPoints[2] = xCentreOfFace - xYOffset - xXOffset;
		pxSupportPoints[3] = xCentreOfFace - xYOffset + xXOffset;

		return 4;
	}

	// ok, what about any supporting edges - this happens when xAxis is perpendicular to an edge, put not parallel to one
	// (because that would be the face case).
	const float fEdgeDotThreshold = 2.6e-3f;

	if ( Asura_Maths::Abs(fX) < fEdgeDotThreshold ) // edge crosses X axis
	{
		Asura_AVector3 xCentreOfEdge = xCentre + xYOffset * Asura_Maths::Sign(fY) + xZOffset * Asura_Maths::Sign(fZ);

		pxSupportPoints[0] = xCentreOfEdge + xXOffset;
		pxSupportPoints[1] = xCentreOfEdge - xXOffset;

		return 2;
	}
	else if ( Asura_Maths::Abs(fY) < fEdgeDotThreshold ) // edge crosses Y axis
	{
		Asura_AVector3 xCentreOfEdge = xCentre + xXOffset * Asura_Maths::Sign(fX) + xZOffset * Asura_Maths::Sign(fZ);

		pxSupportPoints[0] = xCentreOfEdge + xYOffset;
		pxSupportPoints[1] = xCentreOfEdge - xYOffset;

		return 2;
	}
	else if ( Asura_Maths::Abs(fZ) < fEdgeDotThreshold ) // edge crosses Z axis
	{
		Asura_AVector3 xCentreOfEdge = xCentre + xYOffset * Asura_Maths::Sign(fY) + xXOffset * Asura_Maths::Sign(fX);

		pxSupportPoints[0] = xCentreOfEdge + xZOffset;
		pxSupportPoints[1] = xCentreOfEdge - xZOffset;

		return 2;
	}

	// then it must be a vert...
	pxSupportPoints[0] = xCentre +  xXOffset * Asura_Maths::Sign(fX) + xYOffset * Asura_Maths::Sign(fY) + xZOffset * Asura_Maths::Sign(fZ);

	return 1;
}
	}
}


void Asura_Physics_RigidBodyInstance::InitialiseFromDefinition( void )
{
	if ( IsStatic() )
	{
		// tell the physics system that there's been a change to static objects - it will want to recalc spatial hash
		Asura_Physics_System::NotifyOfChangeToStaticObjects();
	}

	// if we have any contacts, they'll need to be eliminated
	m_xContactClusterList.Reset();

	// the centre of mass will need to be calculated afresh
	m_xBodyState.m_xCentreOfMass = m_xInstancePosition + m_pxDefinition->GetOBBCentreOffset() * m_xBodyState.m_xOrientation;
	m_xBodyState.m_xPreviousCentreOfMass = m_xBodyState.m_xCentreOfMass;


	delete [] m_pxObjectVertexPosition;
	delete [] m_pxObjectVertexPrevPosition;

	const u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
	m_pxObjectVertexPosition = new Asura_AVector3[uNumberOfObjectVertices];
	m_pxObjectVertexPrevPosition = new Asura_AVector3[uNumberOfObjectVertices];

	for (u_int u=0; u<uNumberOfObjectVertices; u++)
	{
		m_pxObjectVertexPosition [u]		 = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation;
		m_pxObjectVertexPosition [u]		+= m_xInstancePosition;
		m_pxObjectVertexPrevPosition [u]	 = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation;
		m_pxObjectVertexPrevPosition [u]	+= m_xInstancePosition;
	}

	m_xEnvironmentCache.Flush();

	CalculateInvInertiaTensorDiagonal();

	SetCollisionTestAABB();
}
//#include "Core/Asura_Timers.h"


void Asura_Physics_RigidBodyInstance::GetVelocitiesOfBody( Asura_AVector3& xLinearVelocity, Asura_AVector3& xAngularVelocity ) const
{
	xLinearVelocity = m_xBodyState.m_xLinearVelocity;
	xAngularVelocity = m_xBodyState.m_xAngularVelocity;
}

Asura_AVector3 Asura_Physics_RigidBodyInstance::CalculateVelocityOfPointOnBody( const Asura_AVector3& xPoint ) const
{
	return Asura_AVector3( m_xBodyState.m_xLinearVelocity + ( m_xBodyState.m_xAngularVelocity ^ (xPoint - m_xBodyState.m_xCentreOfMass) ));
}

void Asura_Physics_RigidBodyInstance::SetVelocitiesOfBody( const Asura_AVector3& xLinearVelocity, const Asura_AVector3& xAngularVelocity )
{
 	m_xBodyState.m_xLinearVelocity = xLinearVelocity;
	m_xBodyState.m_xAngularVelocity = xAngularVelocity;
}

void Asura_Physics_RigidBodyInstance::SetAngularVelocityOfBody( const Asura_AVector3& xAngularVelocity )
{
	m_xBodyState.m_xAngularVelocity = xAngularVelocity;
}

void Asura_Physics_RigidBodyInstance::SetLinearVelocityOfBody( const Asura_AVector3& xLinearVelocity )
{
	m_xBodyState.m_xLinearVelocity = xLinearVelocity;
}


void Asura_Physics_RigidBodyInstance::ZeroMomentum( void )
{
	m_xBodyState.m_xLinearVelocity.Zero();
	m_xBodyState.m_xAngularVelocity.Zero();

	m_xJumpVelocity.Zero();
}

void Asura_Physics_RigidBodyInstance::ZeroAngularMomentum( void )
{
	Asura_AVector3 xZeroVector;
	xZeroVector.Zero();
	SetAngularVelocityOfBody( xZeroVector );
}

void Asura_Physics_RigidBodyInstance::ZeroLinearMomentum( void )
{
	Asura_AVector3 xZeroVector;
	xZeroVector.Zero();
	SetLinearVelocityOfBody( xZeroVector );
}

float Asura_Physics_RigidBodyInstance::GetGravityScale( void ) const
{
	return m_fGravityScale;
}
void Asura_Physics_RigidBodyInstance::SetGravityScale( float fGravityScale )
{
	m_fGravityScale = fGravityScale;
}

const Asura_AVector3& Asura_Physics_RigidBodyInstance::GetGravityDirection( void ) const
{
	if ( m_bUseOwnGravityDirection )
	{
		return m_xOwnGravityDirection;
	}
	else
	{
		return Asura_Physics_System::GetGravityDirection();
	}
}

const Asura_AVector3& Asura_Physics_RigidBodyInstance::GetOwnGravityDirection( void ) const
{
	return m_xOwnGravityDirection;
}

void Asura_Physics_RigidBodyInstance::SetOwnGravityDirection( const Asura_AVector3& xDirection )
{
	m_xOwnGravityDirection = xDirection;

	m_xOwnGravityDirection.Normalise();
}


void Asura_Physics_RigidBodyInstance::MoveBase( void )
{
	// save off the previous positions of the object's collision mesh verts	
	const u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
	for (u_int u=0; u<uNumberOfObjectVertices; u++)
	{
		m_pxObjectVertexPrevPosition[u] = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation + m_xInstancePosition;
	}

	ClearObstructionFlags();

	if (IsExternallyControlled())
	{
		
	}
	else
	{
		MoveWithRestFrame();

		m_xBodyState.m_xLinearVelocity = m_xBodyState.m_xLinearVelocity*GetDrag() + GetForceEffect();
		m_xBodyState.m_xAngularVelocity = m_xBodyState.m_xAngularVelocity*GetAngularDrag();

		// KJFL - 15 July 2008 - using the last update's contacts like this was something I added during the Simpsons project, 
		// where we had a dynamic object that often had a large velocity slightly into the ground (Homerball). The swept test
		// would result in the Homerball sticking, rather than moving along the ground. Profiling shows this piece of code
		// is actually quite costly, and it doesn't really gel with the new solver, so I'm going to remove it for now.
		// If a project comes up with a similar need I may reinstate a similar concept with an object flag so not everything
		// has to do it.
#ifdef ASURA_PHYSICS_USE_OLD_SOLVER
		// use the last update's contacts to resolve the velocity vector before we try to move
		for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
		{
			const Asura_AVector3&	xNormal				= pxCluster->GetContactNormal();
			const float				fNormalDotGravity	= xNormal * GetGravityDirection();

			// if it looks floor-like and our velocity is towards it, set the floor flag
			if ( ( fNormalDotGravity < -fVERTICAL_BOUNDARY ) && ( xNormal * m_xBodyState.m_xLinearVelocity < 0.0f ) )
			{
				m_bIsInContactWithFloor = true;
				m_bIsSupported = true;
			}

			CalculateCollisionResponses( pxCluster, true );
		}
#endif

		// KJFL - 18 July 2006 - if object is moving fast, do an extra swept sphere test and stop the object at the collision.
		// This is to stop objects from going too fast at an environment boundary for the contact system to cope.
		float fAllowedDisplacementScale = 1.0f;
		{
			// find the average displacement of the object:
			Asura_AVector3 xDisplacement = m_xBodyState.m_xLinearVelocity;

			// find how far the object wants to travel
			float fDistanceToTravel = xDisplacement.NormaliseReturningMagnitude();

			float fSphereRadius = GetDefinition()->GetMinimumRadius();

			// consider using the sphere test if the distance to travel is more than than smallest half extent of the object
			if (fDistanceToTravel > fSphereRadius )
			{
				m_xBoundingBox.Reset();
				m_xBoundingBox.Set( Asura_Vector_3(m_xBodyState.m_xCentreOfMass) );
				m_xBoundingBox.Set( Asura_Vector_3(m_xBodyState.m_xCentreOfMass + m_xBodyState.m_xLinearVelocity ) );
				m_xBoundingBox.Grow( m_pxDefinition->GetMaximumRadius() );

				// give a 10cm extension when fill cache to minimise number of searches
				m_xEnvironmentCache.Update( m_xBoundingBox, 0.1f );
		
				Asura_Physics_SphereGroup xSphereGroup( *this );
			
				Asura_AVector3 xLastCollisionNormal;
				u_int uLastCollisionFaceID;
				bool bHitEdge;

				float fAllowedDistance = xSphereGroup.SweptSpheresVsEnvironment( xDisplacement, fDistanceToTravel, m_xEnvironmentCache, xLastCollisionNormal, uLastCollisionFaceID, bHitEdge );
				
				// allow a small overlap of this minimum radius, and then the contact resolution stage can pick that up
				fAllowedDistance = Asura_Maths::Min( fAllowedDistance + 0.01f, fDistanceToTravel);

				if ( fAllowedDistance < fDistanceToTravel )
				{
					fAllowedDisplacementScale = fAllowedDistance / fDistanceToTravel;
					m_bHasHitSomething = true;
				}
			}
		}

		m_xBodyState.m_xPreviousCentreOfMass = m_xBodyState.m_xCentreOfMass;
		
		
		m_xBodyState.m_xCentreOfMass += m_xBodyState.m_xLinearVelocity * fAllowedDisplacementScale;

		Asura_AMatrix3x3 xDelta;
		xDelta.Tilde( m_xBodyState.m_xAngularVelocity );		
		xDelta = m_xBodyState.m_xOrientation*xDelta;
		m_xBodyState.m_xOrientation.Add( xDelta );
		m_xBodyState.m_xOrientation.Orthogonalise();

		HandleOrientationConstraints();

		CalculateWorldInvInertiaTensor();
	}

	SetCollisionTestAABB();
}

void Asura_Physics_RigidBodyInstance::Integrate( const Asura_AVector3& xLinearVelocity, const Asura_AVector3& xAngularVelocity )
{
	m_xBodyState.m_xCentreOfMass += xLinearVelocity;

	Asura_AMatrix3x3 xDelta;
	xDelta.Tilde( xAngularVelocity );		
	xDelta = m_xBodyState.m_xOrientation*xDelta;
	m_xBodyState.m_xOrientation.Add( xDelta );
	m_xBodyState.m_xOrientation.Orthogonalise();

	HandleOrientationConstraints();

	CalculateWorldInvInertiaTensor();
}


void Asura_Physics_RigidBodyInstance::MoveActor( void )
{
	ClearObstructionFlags();

	if (!GetActorMoveAsVehicle())
	{
		m_xBodyState.m_xAngularVelocity.Zero();
		HandleOrientationConstraints();
		CalculateWorldInvInertiaTensor();
	}
	
	// if we're not axially aligned (and we're an OBB), our orientation may have changed and stuck us through something
	if ( !m_bAxiallyAligned && m_eCollisionProxy == COLLISION_PROXY_OBB )
	{
		const u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
		for (u_int u=0; u<uNumberOfObjectVertices; u++)
		{
			m_pxObjectVertexPrevPosition[u] = m_pxObjectVertexPosition[u];
		}

		m_xContactClusterList.Reset();
		TestForNewContacts();

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER
		RelaxPositionConstraints();
#endif
	}
	
	Asura_AVector3 xDisp = GetForceEffect();

	xDisp += m_xJumpVelocity * Asura_Physics_System::GetTimeSlice();

	// save off the previous positions of the object's collision mesh verts	
	const u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
	for (u_int u=0; u<uNumberOfObjectVertices; u++)
	{
		m_pxObjectVertexPrevPosition[u] = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation + m_xInstancePosition;
	}

	m_xBodyState.m_xPreviousCentreOfMass = m_xBodyState.m_xCentreOfMass;
	
	MoveWithRestFrame();


	//TODO - need to somehow set the actors "velocity" so we have momentum when hitting other objects
	m_xBodyState.m_xLinearVelocity = xDisp;


	// normalise the direction of travel and calculate how far we should be going in that direction
	float fDistanceLeftToTravel = xDisp.NormaliseReturningMagnitude();
	
	// if the distance we're flagged to move is tiny, ignore it
	if (fDistanceLeftToTravel < fAXIALLYALIGNED_MOVEMENT_THRESHOLD) return;


	bool bSlidingOffActor = false;
	const bool bFalling = ( m_xBodyState.m_xLinearVelocity * GetGravityDirection() > 0.003f );

	Asura_Physics_Broadphase_Query xBroadphaseQuery;

	u_int uIterations = 0;

	Asura_Bounding_Box xActorAABB;

	{
		float fWorldSpaceExtentX = Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsX() * m_xBodyState.m_xOrientation.GetColumn0().GetX() )
								 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsY() * m_xBodyState.m_xOrientation.GetColumn1().GetX() )
								 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsZ() * m_xBodyState.m_xOrientation.GetColumn2().GetX() );

		float fWorldSpaceExtentY = Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsX() * m_xBodyState.m_xOrientation.GetColumn0().GetY() )
								 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsY() * m_xBodyState.m_xOrientation.GetColumn1().GetY() )
								 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsZ() * m_xBodyState.m_xOrientation.GetColumn2().GetY() );

		float fWorldSpaceExtentZ = Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsX() * m_xBodyState.m_xOrientation.GetColumn0().GetZ() )
								 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsY() * m_xBodyState.m_xOrientation.GetColumn1().GetZ() )
								 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsZ() * m_xBodyState.m_xOrientation.GetColumn2().GetZ() );

	
		// KJFL - 24 September 2008 - we do have vector min and max methods, but since
		// the bounding box members aren't vectorised we'll have little gain in using 
		// them. Something to revisit if/when we have an AABB based on Asura_AVector.
		xActorAABB.MinX =  - fWorldSpaceExtentX;
		xActorAABB.MaxX =  + fWorldSpaceExtentX;
							
		xActorAABB.MinY =  - fWorldSpaceExtentY;
		xActorAABB.MaxY =  + fWorldSpaceExtentY;
							
		xActorAABB.MinZ =  - fWorldSpaceExtentZ;
		xActorAABB.MaxZ =  + fWorldSpaceExtentZ;
	}
	

	while ( fDistanceLeftToTravel > 0.0f )
	{
		// KJFL - 27 September 2006 - keep track of how many times we resolve our motion against obstructions
		// and at some point give up, e.g. we're at the bottom of a V like pit, and non of the polygons are
		// flagged as a walkable surface, so the actor is well and truly stuck.
		if ( ++uIterations > uMAX_NUMBER_OF_ITERATIONS_FOR_ACTOR_MOVEMENT )
		{
			// KJFL - 27 September 2006 - In this case zero the 'falling' velocity, 
			// otherwise it'll keep growing making it:
			//
			// a) even more difficult for the actor to move
			// b) the EnvironmentCache will keep growing even though we're not moving
			//
			// We *may* want to set a flag to indicate this state of affairs to the project code via a 'stuck' flag
			// e.g. allow a player to jump if on a 'floor' poly or 'stuck'.
			m_xJumpVelocity.Zero();
			m_bIsSupported = true;
			break;
		}
		
		const u_int uMAX_NUMBER_OF_COLLISIONS_TO_RECORD = 8;
		
		Asura_Physics_CollisionData axCollision[uMAX_NUMBER_OF_COLLISIONS_TO_RECORD];
		u_int uNumberOfCollisions = 0;
		Asura_Physics_RigidBodyInstance* pxObstacleHit = 0;

		float fDistanceAllowed = 0.0f;

		
		switch ( m_eCollisionProxy )
		{
			default:
			case COLLISION_PROXY_POLYGONAL:
			case COLLISION_PROXY_OBB:
			{
				// find the bounding box covering the old and new positions
				m_xBoundingBox.Reset();

				{
	
					m_xBoundingBox = xActorAABB;
					m_xBoundingBox.Displace( Asura_Vector_3( m_xBodyState.m_xCentreOfMass ) );
					
					Asura_AVector3 xSweep = xDisp*fDistanceLeftToTravel;

					if (xSweep.GetX() > 0.0f)
					{
						m_xBoundingBox.MaxX += xSweep.GetX();
					}
					else
					{
						m_xBoundingBox.MinX += xSweep.GetX();
					}
					
					if (xSweep.GetY() > 0.0f)
					{
						m_xBoundingBox.MaxY += xSweep.GetY();
					}
					else
					{
						m_xBoundingBox.MinY += xSweep.GetY();
					}
					
					if (xSweep.GetZ() > 0.0f)
					{
						m_xBoundingBox.MaxZ += xSweep.GetZ();
					}
					else
					{
						m_xBoundingBox.MinZ += xSweep.GetZ();
				}
				}

				// test against environment and see what's in our way
				m_xEnvironmentCache.Update( m_xBoundingBox, 0.1f );

                float fTimeToImpact = 1.0f;
 				
				if ( m_xEnvironmentCache.GetFaceList() )
				{
					Asura_AVector3 xCentre = m_xBodyState.m_xCentreOfMass;
					
					Asura_AVector3 xVelocity = xDisp * fDistanceLeftToTravel;

					Asura_OBB xOBB( GetHalfExtents(), xCentre, GetOrientation() );

					for ( Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvironmentCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
					{
						Asura_Physics_Tri xTri( pxFace->m_uFaceID);
						
						Asura_AVector3 xNormal;
						bool bHitEdge;

						float fHitTime = Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsTri( xOBB, xVelocity, xTri, xNormal, pxFace->m_uLastSeparationPlane, bHitEdge );

						if (fHitTime < -0.1f)
						{
							// We're stuck. 
							
							// Kill the falling velocity, and flag that we've hit something - don't set the floor flags though
							m_xJumpVelocity.Zero();
							m_bIsSupported = true;
							UnlinkFromRestFrame();
							m_bHasHitSomething = true;

							// KJFL - 22 April 2007 - Due to the wonders of floating point maths, it's possible what we see as
							// an intersection here isn't seen as such by the code to remove penetrations vs the environment due
							// to the different algorithms involved - so if this is a small penetration, we'll try and fix it
							// here and now - otherwise we'll leave it for the "real" check-environment-penetration phase. 
							Asura_AVector3 xSeparationAxis;
							float fPenetration;

							if ( Asura_Physics_CCD::DoesOBBIntersectTri( xOBB, xTri, fPenetration, xSeparationAxis ) )
							{
								// if it's less than a few mm, we'll try to remove the penetration directly
								const float fSAFE_PENETRATION_TO_REMOVE = 0.003f;

								// KJFL - 26 April 2007 - in the world of floating point a + b = a if b is a lot smaller 
								// than a. If we going to try to solve a penetration, make at least a minimal distance
								// to move the object. Otherwise the addition may have no effect.
								const float fMIN_PENETRATION = 0.001f;

								if ( fPenetration < fSAFE_PENETRATION_TO_REMOVE )
								{
									m_xBodyState.m_xCentreOfMass += xSeparationAxis * Asura_Maths::Max( fMIN_PENETRATION, fPenetration );
								}
							}

							SetCollisionTestAABB();
							m_xBodyState.m_xLinearVelocity = m_xBodyState.m_xCentreOfMass - m_xBodyState.m_xPreviousCentreOfMass;

							return;
						}

						const float fTIME_EPSILON = 0.0001f;
						if ( uNumberOfCollisions && ( Asura_Maths::Abs(fHitTime - fTimeToImpact) < fTIME_EPSILON ) )
						{
							if ( uNumberOfCollisions < uMAX_NUMBER_OF_COLLISIONS_TO_RECORD )
							{
								axCollision[uNumberOfCollisions].m_bIsEdge = bHitEdge;
								axCollision[uNumberOfCollisions].m_uPolyFlags = Asura_Physics_Environment::GetFaceFlags( pxFace->m_uFaceID );
								axCollision[uNumberOfCollisions].m_uPolyID = pxFace->m_uFaceID;
								axCollision[uNumberOfCollisions].m_xNormal = xNormal;
								fTimeToImpact = Asura_Maths::Min( fHitTime, fTimeToImpact );

								++uNumberOfCollisions;
							}
						}
						else if ( fHitTime<fTimeToImpact )
						{
							fTimeToImpact = fHitTime;
							axCollision[0].m_bIsEdge = bHitEdge;
							axCollision[0].m_uPolyFlags = Asura_Physics_Environment::GetFaceFlags( pxFace->m_uFaceID );
							axCollision[0].m_uPolyID = pxFace->m_uFaceID;
							axCollision[0].m_xNormal = xNormal;
							uNumberOfCollisions = 1;
						}
						
					}
				}

				if (fTimeToImpact < 1.0f)
				{
					// we've hit something... find how far we're allowed to travel, and subtract a 
					// teeny distance to allow for float inaccuracies.
					fDistanceAllowed = Asura_Maths::Max((fTimeToImpact * fDistanceLeftToTravel) - fAXIALLYALIGNED_DIST_THRESHOLD, 0.0f);
				}
				else
				{
					fDistanceAllowed = fDistanceLeftToTravel;
				}

				break;
			}
			case COLLISION_PROXY_SPHERE:
			case COLLISION_PROXY_SPHERESTACK:
			{
				// find the bounding box covering the old and new positions
				m_xBoundingBox.Reset();

				m_xBoundingBox.Set( Asura_Vector_3( m_xBodyState.m_xCentreOfMass ) );
				m_xBoundingBox.Set( Asura_Vector_3( m_xBodyState.m_xCentreOfMass + xDisp*fDistanceLeftToTravel ) );
				m_xBoundingBox.Grow( m_pxDefinition->GetMaximumRadius() );
	
				// test against environment and see what's in our way
				m_xEnvironmentCache.Update( m_xBoundingBox, 0.1f );

				Asura_Physics_SphereGroup xSphereStack( *this );

				fDistanceAllowed = xSphereStack.SweptSpheresVsEnvironment( xDisp, fDistanceLeftToTravel, m_xEnvironmentCache, axCollision, uNumberOfCollisions, uMAX_NUMBER_OF_COLLISIONS_TO_RECORD );
			}	
		}
		if (fDistanceAllowed > 0.0f)
		{
			// check what we should be testing against
			const bool bCollideWithExternals = !m_bIgnoreExternalControlObjects;
			const bool bCollideWithActors = (!m_bIgnoreDynamicObjects) && GetActorDontPushActors();
			const bool bCollideWithDynamics = bCollideWithActors;

			xBroadphaseQuery.MakeQuery_StandardActorMovement( m_xBoundingBox, bCollideWithExternals, bCollideWithActors, bCollideWithDynamics );

			Asura_AVector3 xRelativeVelocity = xDisp * fDistanceAllowed;

			const Asura_OBB xInstanceOBB( GetHalfExtents(), m_xBodyState.m_xCentreOfMass, GetOrientation() );

			Asura_Physics_MPR_Body xBody1( this, m_xBodyState.m_xCentreOfMass, m_xBodyState.m_xCentreOfMass+xRelativeVelocity );

			float fTimeToImpact = 1.0f;
		
			for ( const Asura_Physics_Broadphase_Entry* pxEntry = xBroadphaseQuery.GetFirstEntry(); pxEntry; pxEntry = pxEntry->GetNextEntry() )
			{
				Asura_Physics_RigidBodyInstance* pxObstacle = pxEntry->GetInstance();

				if ( CollisionGroupsAllowCollision( pxObstacle ) == false ) continue;

				if ( pxObstacle==this || pxObstacle->GetHandle() == GetHandleToIgnore() ) continue;

				if ( pxObstacle->GetIgnoreDynamicObjects() ) continue;

				
				switch ( pxObstacle->GetCollisionProxy() )
				{
					default:
					case COLLISION_PROXY_OBB:
					{
						Asura_AVector3 xNormal;

						// KJFL - 6 November 2008 - if an obstacle isn't static, ignore it if we're already overlapping it. This produces
						// smoother movement against things that have moved into us (e.g. rocking platforms, lifts, etc)
						// KJFL - 11 March 2009 - this has now gone full circle, and I'm not ignoring initial overlaps at all.
						// The problem is that fast moving characters could run through external control objects (doors, vehicles). The 
						// rocking platforms from the Simpsons were a very special case, and every other game we've made needs to treat
						// overlaps in a penalising way. If this causes issues in the future we could make this behaviour be determined
						// by a per actor flag.
						bool bIgnoreInitialOverlap = false;//!(pxObstacle->IsStatic());

						bool bIntersect;
		
						// KJFL - 1 December 2008 - Any Actor not an OBB uses the MPR routine, otherwise we'll use the specialist OBB vs OBB 
						// we've always used previously. I want to move everything to the MPR version eventually, though there may be issues
						// to resolve before we can do that.
						if ( GetCollisionProxy() != COLLISION_PROXY_OBB )
						{
							Asura_Physics_MPR_Body xBody2( pxObstacle );
	
							float fRewindTime;
							float fPenetration;
							bIntersect = Asura_Physics_CCD::Swept_MPR_Test( xBody1, xBody2, xNormal, fRewindTime, fPenetration );
	
							// TODO - messy bit of logic - could wrap into swept routine.
							if (bIntersect)
							{
								// If the relative velocity points in the same direction as the impact normal,
								// this is probably a t=0 collision we're trying to move away from - ignore it.
								if ( xRelativeVelocity * xNormal > 0.0f )
								{
									bIntersect = false;
								}
								else if (fRewindTime==1.0f)
								{
									if ( bIgnoreInitialOverlap )
									{
										bIntersect = false;
									}
									else // do another non-swept test to make sure there really is an initial overlap
									{
										// KJFL - 20 April 2008 - Do the exact same test the pairing code will do to push
										// things apart - we're seeing issues sticking because the swept test says one thing,
										// and the non-swept test says another.
										Asura_Physics_MPR_Body xUnmovedBody(this);
										bIntersect = Asura_Physics_CCD::MPR_Test( xUnmovedBody, xBody2, xNormal, fPenetration );
									
										if (bIntersect) fTimeToImpact = 0.0f;
									}
								}
								else 
								{
									float fTime = 1.0f - fRewindTime;

									if (fTime<fTimeToImpact)
									{
										fTimeToImpact = fTime;
									}
									else
									{
										bIntersect = false;
									}
								}
							}
						}
						else
						{
							const Asura_OBB xObstacleOBB( pxObstacle->GetHalfExtents(), pxObstacle->m_xBodyState.m_xCentreOfMass, pxObstacle->GetOrientation() );
		 
							bIntersect = Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsOBB( xInstanceOBB, xObstacleOBB, xRelativeVelocity, fTimeToImpact, xNormal, bIgnoreInitialOverlap );
						}

						if (bIntersect)
						{
							xNormal.Normalise();
							axCollision[0].m_bIsEdge = false;
							axCollision[0].m_uPolyFlags = pxObstacle->GetPolyFlagsToReturnForActorCollision();
							axCollision[0].m_uPolyID = 0xffffffff;
							axCollision[0].m_xNormal = xNormal;
							uNumberOfCollisions = 1;
	
							pxObstacleHit = pxObstacle;
						}
						break;
					}
					case COLLISION_PROXY_POLYGONAL:
					{
						if ( !pxObstacle->GetDefinition() )
						{
							break;
						}
						
						const Asura_CollisionMesh* pxMesh = pxObstacle->GetDefinition()->GetCollisionMesh();

						if ( !(pxMesh) )
						{
							Asura_Assertf( false, ("Polygonal Proxy object has no collision mesh - associated Guid is %d", pxObstacle->GetAssociatedEntityGuid()) );
							break;
						}

						Asura_Physics_CollisionMesh_Iterator xTriIterator( pxMesh );
						Asura_Physics_Tri xTri;
						for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
						{
							// TODO - rather than calculating the OBB and Tri in world space, perhaps we can use the Tri's local space?
							// we'd have to transform something (i.e. OBB or Tri) so perhaps it wouldn't save much
							xTri.Transform( pxObstacle->m_xBodyState.m_xOrientation, pxObstacle->m_xInstancePosition );

							u_int uFirstPlane = 0;
							Asura_AVector3 xAxisUsed;
							bool bHitEdge = false;

							float fHitTime;
							
							// Use the old trusty OBB vs Tri is possible - otherwise anything else is a MPR test
							if ( GetCollisionProxy() == COLLISION_PROXY_OBB )
							{
								fHitTime = Asura_Physics_CCD::FindTimeOfIntersection_SweptOBBVsTri( xInstanceOBB, xRelativeVelocity, xTri, xAxisUsed, uFirstPlane, bHitEdge );
							}
							else
							{
								Asura_Physics_MPR_Body xBody2( &xTri );
	
								float fRewindTime;
								float fPenetration;
								bool bIntersect = Asura_Physics_CCD::Swept_MPR_Test( xBody1, xBody2, xAxisUsed, fRewindTime, fPenetration );

								// If there's an intersection, and the relative velocity is definitely heading towards the collision normal,
								// flag a collision - otherwise ignore it. (The relative velocity bits stops sticking at glancing angles.)
								if ( bIntersect && ( xRelativeVelocity * xAxisUsed < -0.0001f ) )
								{
									fHitTime = 1.0f - fRewindTime;
								}
								else
								{
									// no collision - just do a time > 1.0f
									fHitTime = 2.0f;
								}
							}
							
							if ( fHitTime < 0.0f ) 
							{
								// KJFL - 29 August 2007 - commented this line out so we instead ignore this polygon. Stops sticky cases.
								// KJFL - 11 March 2009 - now blocking on intersection with polys. This stops actors getting through polygonal
								// objects. This may create sticky issues, but it solves "going through" issues.
								fTimeToImpact = -1.0f;
								break;
							}
							// if you've got an earlier impact, OR if you've found a poly with the same time of impact which is "more in the way"
							else if ( ( fHitTime < fTimeToImpact ) || ( fHitTime == fTimeToImpact && xAxisUsed*xDisp < axCollision[0].m_xNormal*xDisp) )
							{
								// record the new candidate for the first time of impact
								fTimeToImpact = fHitTime;
							
								axCollision[0].m_bIsEdge = bHitEdge;
								axCollision[0].m_uPolyFlags = pxObstacle->GetPolyFlagsToReturnForActorCollision() | xTriIterator.GetFlagsOfCurrentTri();
								axCollision[0].m_uPolyID = 0xffffffff;
								axCollision[0].m_xNormal = xAxisUsed;
								uNumberOfCollisions = 1;

								pxObstacleHit = pxObstacle;


							}
						}
						break;
					}
				}
			}
			if (fTimeToImpact<0.0f)
			{
				fDistanceAllowed = 0.0f;
			}
			else if (fTimeToImpact<1.0f)
			{
				fDistanceAllowed = Asura_Maths::Max((fTimeToImpact * fDistanceAllowed) - fAXIALLYALIGNED_DIST_THRESHOLD, 0.0f);
				
				Asura_Assert( pxObstacleHit, "Code bug: null pointer" );

				if (pxObstacleHit)
				{
					if (pxObstacleHit->IsAsleep()) pxObstacleHit->WakeUp();
					AddObjectCollisionReport( pxObstacleHit );
					pxObstacleHit->AddObjectCollisionReport( this );


					if ( pxObstacleHit->HasOwnRestFrame() )
					{
						const float	fNormalDotGravity = axCollision[0].m_xNormal * GetGravityDirection();

						if ( ! HasOwnRestFrame() && ( fNormalDotGravity < -fVERTICAL_BOUNDARY ) )
						{
							LinkToRestFrame( pxObstacleHit->m_pxRestFrame );
						}
					}
				}
			}
		}

		// if we can move at all, let's do so and update the necessary positional data
		if (fDistanceAllowed > 0.0f)
		{
			const Asura_AVector3 xAllowedDisplacement = xDisp * fDistanceAllowed;

			m_xBodyState.m_xCentreOfMass += xAllowedDisplacement;

			// subtract how far we've already come from the distance we were expecting to travel.
			fDistanceLeftToTravel -= fDistanceAllowed;
		}



		// if the distance left to travel is tiny then just don't bother.
		if (fDistanceLeftToTravel < 0.0001f) break;

		// well, we've definitely hit something...
		m_bHasHitSomething = true;

		// but do we know what?
		if (uNumberOfCollisions == 0) break;

		u_int uBestCollision = 0;

		for ( u_int u=0; u<uNumberOfCollisions; ++u )
		{
			// add any environment polys to the reports
			if ( axCollision[u].m_uPolyID != 0xffffffff )
			{
				// TODO: currently using the centre of mass for the actor as the collision FX report position
				AddEnvironmentCollisionReport( axCollision[u].m_uPolyID, axCollision[u].m_xNormal, m_xBodyState.m_xCentreOfMass );
				UnlinkFromRestFrame();
			}

			// always take a face over an edge
			if ( (!axCollision[u].m_bIsEdge) && (axCollision[uBestCollision].m_bIsEdge) )
			{
				uBestCollision = u;
			}
			// otherwise find the collision normal that most blocks our path
			else if ( (axCollision[uBestCollision].m_xNormal*xDisp > axCollision[u].m_xNormal*xDisp) && (axCollision[u].m_bIsEdge == axCollision[uBestCollision].m_bIsEdge))
			{
				uBestCollision = u;
			}

		}

		// most blocking normal
		Asura_AVector3 xLastCollisionNormal = axCollision[uBestCollision].m_xNormal;

		const Asura_AVector3& xGravityDirection = GetGravityDirection();
		float fNormalDotGravity = xLastCollisionNormal * xGravityDirection;
	
		if ( pxObstacleHit )
		{
			// if we've hit a normal that looks floor-like but the object doesn't want to be stood on,
			// slide the Actor away sideways in the plane of the collision.
			if ( pxObstacleHit->GetActorsSlideOffMe() && fNormalDotGravity < -fVERTICAL_BOUNDARY )
			{
				// AJG - 20/08/07 - See if someone's trying to force us away, and let them have a go if they are.
				const Asura_AVector3 xForceEffectParallelToGravity( xGravityDirection * (xGravityDirection * m_xForceEffect) );
				const Asura_AVector3 xForceEffectPerpToGravity( m_xForceEffect - xForceEffectParallelToGravity );

				if (xForceEffectPerpToGravity.MagnitudeSquared() >= (fACTOR_SLIDE_OFF_MOVEMENT_OVERRIDE_SPEED * fACTOR_SLIDE_OFF_MOVEMENT_OVERRIDE_SPEED))
				{
					xDisp = xForceEffectPerpToGravity;
					xDisp.Normalise();
					fNormalDotGravity = 0.0f;
					bSlidingOffActor = true;
				}
				else
				{
					Asura_AVector3 xDirectionToSlideOff = GetCentreOfMass() - pxObstacleHit->GetCentreOfMass();

					xDirectionToSlideOff = xDirectionToSlideOff - xLastCollisionNormal*(xDirectionToSlideOff*xLastCollisionNormal);

					// KJFL - 29 May 2007 - technically if you're absolutely exactly aligned, we won't have 
					// a direction to slide off; may need to revisit this if that's a problem.
					if ( xDirectionToSlideOff.NormaliseReturningMagnitude() > 0.0f )
					{
						xDisp = xDirectionToSlideOff;
						fNormalDotGravity = 0.0f;
						bSlidingOffActor = true;
					}
				}

				// If we're sliding off, clamp our jump velocity so we don't end up with a huge sideways motion after doing this
				if (bSlidingOffActor)
				{
					Asura_AVector3 xJumpVelocityParallelToGravity( xGravityDirection * (xGravityDirection * m_xJumpVelocity) );
					const Asura_AVector3 xJumpVelocityPerpToGravity( m_xJumpVelocity - xJumpVelocityParallelToGravity );
					if (xJumpVelocityParallelToGravity.MagnitudeSquared() > (fACTOR_SLIDE_OFF_JUMP_SPEED_LIMIT * fACTOR_SLIDE_OFF_JUMP_SPEED_LIMIT))
					{
						xJumpVelocityParallelToGravity.Normalise();
						xJumpVelocityParallelToGravity *= fACTOR_SLIDE_OFF_JUMP_SPEED_LIMIT;
						m_xJumpVelocity = xJumpVelocityPerpToGravity + xJumpVelocityParallelToGravity;
					}
				}
			}
		}


		bool bFloorFlagSet = GetAllSurfacesWalkable();
	
		// if we've hit an edge, we need to AND all the poly flags that meet at the edge 
		// to get a composite flags we can use for testing for walkable surfaces etc
		if ( axCollision[uBestCollision].m_bIsEdge )
		{
			if ( (m_eCollisionProxy == COLLISION_PROXY_OBB) && (uNumberOfCollisions == 1) )
			{
				// if we've got a solitary edge and we're an OBB - don't set any
				// floor flags as there isn't enough to go on.
						//bFloorFlagSet = (axCollision[uBestCollision].m_uPolyFlags & AsuraPolygonFlag_Physics_WalkableSurface) != 0;
			}
			else
			{
				u_int uFlags = axCollision[uBestCollision].m_uPolyFlags;

				// all edges must agree on the flags.
				for ( u_int u=0; u<uNumberOfCollisions; ++u )
				{
					// Not sure if it's slower doing this extra check. ;)
					// if ( u == uBestCollision ) continue;

					if ( axCollision[u].m_bIsEdge )
					{
						uFlags &= axCollision[u].m_uPolyFlags;
					}
				}
				if ( bFloorFlagSet )
				{
					bFloorFlagSet = (uFlags & GetNotWalkableSurfaceFlags() ) == 0;
		
				}
				else
				{
				bFloorFlagSet = (uFlags & AsuraPolygonFlag_Physics_WalkableSurface) != 0;
			}
		}
		}
		else
		{
			if ( bFloorFlagSet )
			{
				bFloorFlagSet = (axCollision[uBestCollision].m_uPolyFlags & GetNotWalkableSurfaceFlags() ) == 0;

			}
			else
			{
			bFloorFlagSet = (axCollision[uBestCollision].m_uPolyFlags & AsuraPolygonFlag_Physics_WalkableSurface) != 0;
		}
		}

		// KJFL - 20 February 2006 - 
		// If we've hit a walkable surface, i.e. "floor", then we need to zero the jump velocity (actors don't 
		// bounce after a jump) and set the IsInContactWithFloor flag to true.
		// Currently also allowing normals 45 degrees off straight up - because existing art may not have floor flagged correctly.
	
		const float fAngleThreshold = (axCollision[uBestCollision].m_bIsEdge) ? -fEDGE_AS_FLOOR_THRESHOLD : -fVERTICAL_BOUNDARY;
		
		if ( bFloorFlagSet || (fNormalDotGravity < fAngleThreshold) )
		{
			m_bIsInContactWithFloor = true;
			m_bIsSupported = true;
			m_xJumpVelocity.Zero();
		}


		// If we're in contact with the floor, and we have no lateral movement (i.e. the character 
		// would like to stand still), don't do another iteration (which would make the character slide
		// down a slope, etc.)
		if ( m_bIsInContactWithFloor )
		{
			const Asura_AVector3 xForceEffectParallelToGravity( xGravityDirection * (xGravityDirection * m_xForceEffect) );
			const Asura_AVector3 xForceEffectPerpToGravity( m_xForceEffect - xForceEffectParallelToGravity );
			
			if ( xForceEffectPerpToGravity.MagnitudeSquared() < (2.0f * fASURA_PHYSICS_ACTOR_MOVEMENT_DEADZONE * fASURA_PHYSICS_ACTOR_MOVEMENT_DEADZONE) )
			{
				break;
			}
		}

		
		// if the actor has hit a vertical surface, which is flagged as walkable, go upwards, e.g. treat it as stairs
		if ( bFloorFlagSet && fNormalDotGravity < fVERTICAL_BOUNDARY && fNormalDotGravity > -fVERTICAL_BOUNDARY)
		{
			const Asura_AVector3 xDispParallelToGravity( xGravityDirection * (xGravityDirection * xDisp) );
			const Asura_AVector3 xDispPerpToGravity(xDisp - xDispParallelToGravity);
			const Asura_AVector3 xAdditionalDisp(xGravityDirection * -30.0f); // a strong "upwards" movement

			xDisp = xDispPerpToGravity + xAdditionalDisp;
			xDisp.Normalise();
		}

		//Asura_DebugText::AddTextBlock( 5, 1+uIterations, 0xffffffff, 0, "%f,%f,%f", xLastCollisionNormal.GetX(),xLastCollisionNormal.GetY(), xLastCollisionNormal.GetZ() );
		// resolve the direction of travel against this latest collision
		const float fDot = -xDisp * xLastCollisionNormal;

		// if the direction of travel is blocked by the collision vector...
		if (fDot > 0.0f)
		{
			// give the resolution a tiny boost to avoid sticking against this surface again next frame
			// (we need to minimise this value as much as possible really)
			Asura_AVector3 xDirectionAdjust = xLastCollisionNormal * fDot * 1.01f;
         	
			// resolve jump velocity against collision surface
			m_xJumpVelocity -= xLastCollisionNormal * (m_xJumpVelocity * xLastCollisionNormal);

			if ( bFalling && !m_bIsSupported )
			{
				float fFallComponent = m_xJumpVelocity * GetGravityDirection();

				// If component of jump velocity in direction of gravity is small, mark the object as supported.
				if ( fFallComponent >= 0.0f && fFallComponent < 0.001f ) 
				{
					m_bIsSupported = true;
					m_xJumpVelocity.Zero();
				}
			}

			xDisp += xDirectionAdjust;

			// If flagged to do so, treating steep, "non floor" polys as vertical walls if we're going up them, 
			// but leaving them as is if we're sliding/running down them. 
			if (GetActorShouldTreatSteepSlopesAsVertical())
			{
				// if we're stood on something which is an upwards slope which isn't flagged as floor, 
				// and when we resolve the direction we're heading upwards, we've done something wrong.
				const bool bUpwardsSlope = (fNormalDotGravity < 0.0f && fNormalDotGravity > -fVERTICAL_BOUNDARY) && (!bFloorFlagSet);

				if ( bUpwardsSlope && ( xDisp * xGravityDirection < 0.0f ) )
				{
					// throw away that last correction
					xDisp -= xDirectionAdjust;

					// pretend that last normal was due to a vertical wall, by removing the component of the normal
					// parallel to gravity
					xLastCollisionNormal -= xGravityDirection * (fNormalDotGravity);
					const float fMagnitude = xLastCollisionNormal.NormaliseReturningMagnitude();

					// if that didn't make a nearly zero normal, let's collide against that
					if (fMagnitude > 0.01f)
					{
						xDisp -= xLastCollisionNormal * (xDisp * xLastCollisionNormal * 1.01f);
					}
					else // otherwise it's time to give up
					{
						break;
					}
				}
			}
		}

		// renormalise the direction of travel
		const float fMagnitude = xDisp.NormaliseReturningMagnitude();

		// however, if the resolved direction of travel vector was very small
		// prior to normalisation, this indicates the object is heading pretty
		// much straight towards an obstacle - so we should just stop.
		if (fMagnitude < 0.00001f)
		{
			break;
		}

		if (!GetActorIsSuperSlidy() && !bFloorFlagSet)
		{
			fDistanceLeftToTravel *= fMagnitude;
		}

	}

	//TODO - need to somehow set the actors "velocity" so we have momentum when hitting other objects
	m_xBodyState.m_xLinearVelocity = m_xBodyState.m_xCentreOfMass - m_xBodyState.m_xPreviousCentreOfMass;

	// if we're sliding off an Actor then explicitly set the floor and supported flags to be false
    if ( bSlidingOffActor )
	{
		m_bIsSupported = false;
		m_bIsInContactWithFloor = false;
	}
	else if ( m_bHasHitSomething && bFalling && !m_bIsSupported ) // do a last check to see if we should set the support flag
	{
		// If component of velocity in direction of gravity is small, mark the object as supported.
		float fFallComponent = m_xBodyState.m_xLinearVelocity * GetGravityDirection();
		if ( fFallComponent >= 0.0f && fFallComponent < 0.001f )
		{
			m_bIsSupported = true;
			m_xJumpVelocity.Zero();
		}
	}


	if (GetActorMoveAsVehicle())
	{
		const float				fMaxExtent			= Asura_Maths::Max( GetDefinition()->GetOBBHalfExtentsX(), GetDefinition()->GetOBBHalfExtentsZ() );
		const Asura_AVector3&	xGravityDirection	= GetGravityDirection();

		m_xBodyState.m_xLinearVelocity = xGravityDirection * dbg_ActorVehicle_GroundingSpeed; // * fMinExtent;
	//	m_xBodyState.m_xAngularVelocity.Zero();
		m_xBodyState.m_xCentreOfMass += xGravityDirection * dbg_ActorVehicle_GroundTestDistanceDown;
		
		SetCollisionTestAABB( true );

		
// 		const u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
		for (u_int u=0; u<uNumberOfObjectVertices; u++)
		{
			m_pxObjectVertexPosition[u] = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation + m_xInstancePosition;
		}
		
		TestForNewContacts_ObjectVertices();
		TestForNewContacts_EnvironmentVertices();
		TestForNewContacts_Edges();
#if 0
		for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
		{
			CalculateCollisionResponses( pxCluster, false );
		}
#else
		CalculateAllCollisionResponses( false );
#endif

		m_xContactClusterList.Reset();

		m_xBodyState.m_xCentreOfMass -= xGravityDirection * dbg_ActorVehicle_GroundTestDistanceUp;
		SetCollisionTestAABB();

		m_xBodyState.m_xAngularVelocity	= m_xBodyState.m_xAngularVelocity - (xGravityDirection * (m_xBodyState.m_xAngularVelocity * xGravityDirection));
		
		Asura_AVector3 xAngDelta( m_xBodyState.m_xAngularVelocity );

		float fAngSpeed = xAngDelta.NormaliseReturningMagnitude();

		fAngSpeed -= dbg_ActorVehicle_GroundAlignmentDeadZone*fMaxExtent;
		if ( fAngSpeed > 0.0f )
		{
			const float fRotationScale = fAngSpeed * Asura_Maths::Max( 1.0f, fMaxExtent*fMaxExtent ) ;//* Asura_Physics_System::GetTimeSlice() * 50.0f;

			xAngDelta *= fRotationScale;

			Asura_AMatrix3x3 xDelta;
			xDelta.Tilde( xAngDelta );		
			xDelta = m_xBodyState.m_xOrientation*xDelta;
			m_xBodyState.m_xOrientation.Add( xDelta );
			m_xBodyState.m_xOrientation.Orthogonalise();
		}
		m_xBodyState.m_xAngularVelocity *= 0.60f;

		HandleOrientationConstraints();
	
		CalculateWorldInvInertiaTensor();
	}
	else
	{
		SetCollisionTestAABB();
	}
	
}

void Asura_Physics_RigidBodyInstance::SetOrientation( const Asura_AMatrix3x3& xNewOrientation )
{
	if ( IsStatic() )
	{
		// tell the physics system that there's been a change to static objects - it will want to recalc spatial hash
		Asura_Physics_System::NotifyOfChangeToStaticObjects();
	}
	else if (IsExternallyControlled()) 
	{
		WakeUp();
	}

	// instance position stays where it was, but centre of mass may change
	CalculateRenderingPosition();
	m_xBodyState.m_xOrientation = xNewOrientation;

	CalculateWorldInvInertiaTensor();

	// if we don't have a valid mesh definition yet, just store the position and orientation
	if ( !m_pxDefinition )
	{
		return;
	}

	Asura_AVector3 xOldOffset = m_xBodyState.m_xPreviousCentreOfMass - m_xBodyState.m_xCentreOfMass;

	m_xBodyState.m_xCentreOfMass = m_xInstancePosition + m_pxDefinition->GetOBBCentreOffset() * m_xBodyState.m_xOrientation;
	m_xBodyState.m_xPreviousCentreOfMass = xOldOffset + m_xBodyState.m_xCentreOfMass;

	{
		SetCollisionTestAABB();
	}
}

void Asura_Physics_RigidBodyInstance::SetPosition( const Asura_AVector3& xNewPosition )
{
	if ( IsStatic() )
	{
		// tell the physics system that there's been a change to static objects - it will want to recalc spatial hash
		Asura_Physics_System::NotifyOfChangeToStaticObjects();
	}

	// if you're not externally controlled, flush any cached data
	if ( !IsExternallyControlled() )
	{
		// Flush cached data, since they won't be valid if we teleport to somewhere else.
		FlushCachedData();

		// and destroy any pairings we might be in
		Asura_Physics_Pairing::RemovePairingsUsingInstance( this );
	}
	else
	{
		// KJFL - 14 December 2006 - we only WakeUp external control objects because things like dynamic objects
		// get teleported around by constructors/loading etc and we don't want to wake them up!
		WakeUp();
	}

	m_xInstancePosition = xNewPosition;

	// if we don't have a valid mesh definition yet, just store the position and exit
	if ( !m_pxDefinition )
	{
		return;
	}

	m_xBodyState.m_xPreviousCentreOfMass = m_xBodyState.m_xCentreOfMass;
	m_xBodyState.m_xCentreOfMass = m_xInstancePosition + m_pxDefinition->GetOBBCentreOffset() * m_xBodyState.m_xOrientation;


#ifdef ASURA_ASSERT
	{
		const Asura_AVector3 xDisplacement = m_xBodyState.m_xCentreOfMass - m_xBodyState.m_xPreviousCentreOfMass;
		const Asura_Entity* pxEntity = Asura_Entity_Server::FindEntity(GetAssociatedEntityGuid());
		Asura_Assertf( xDisplacement.MagnitudeSquared() < ASURA_PHYSICS_MAX_EXPECTED_OBJECT_SIZE * ASURA_PHYSICS_MAX_EXPECTED_OBJECT_SIZE, 
			("Code Bug: Object was just moved across a very large distance via SetPosition(). \nEntity guid is %d."
			"Classification %s\n Old centre of mass = %.02f,%.02f,%.02f\nNew centre of mass = %.02f,%.02f,%.02f", 
			GetAssociatedEntityGuid(), pxEntity ? Asura_Network::GetEntityClassificationDesc(pxEntity->GetClassification()) : "unknown",
			Asura_Maths::FtoD(m_xBodyState.m_xPreviousCentreOfMass.GetX()), 
			Asura_Maths::FtoD(m_xBodyState.m_xPreviousCentreOfMass.GetY()), 
			Asura_Maths::FtoD(m_xBodyState.m_xPreviousCentreOfMass.GetZ()), 
			Asura_Maths::FtoD(m_xBodyState.m_xCentreOfMass.GetX()), 
			Asura_Maths::FtoD(m_xBodyState.m_xCentreOfMass.GetY()), 
			Asura_Maths::FtoD(m_xBodyState.m_xCentreOfMass.GetZ()) ));
	}
#endif

	{
		SetCollisionTestAABB();
	}
}


void Asura_Physics_RigidBodyInstance::TeleportTo( const Asura_AVector3& xNewPosition )
{
	if (m_xInstancePosition == xNewPosition)
	{
		return;
	}

	if ( IsStatic() )
	{
		// tell the physics system that there's been a change to static objects - it will want to recalc spatial hash
		Asura_Physics_System::NotifyOfChangeToStaticObjects();
	}

	// if you're not externally controlled, flush any cached data
	if ( !IsExternallyControlled() )
	{
		// Flush cached data, since they won't be valid if we teleport to somewhere else.
		FlushCachedData();

		// and destroy any pairings we might be in
		Asura_Physics_Pairing::RemovePairingsUsingInstance( this );
	}
	else
	{
		// KJFL - 14 December 2006 - we only WakeUp external control objects because things like dynamic objects
		// get teleported around by constructors/loading etc and we don't want to wake them up!
		WakeUp();
	}

	m_xInstancePosition = xNewPosition;

	// if we don't have a valid mesh definition yet, just store the position and exit
	if ( !m_pxDefinition )
	{
		return;
	}

	m_xBodyState.m_xCentreOfMass = m_xInstancePosition + m_pxDefinition->GetOBBCentreOffset() * m_xBodyState.m_xOrientation;
	m_xBodyState.m_xPreviousCentreOfMass = m_xBodyState.m_xCentreOfMass;

	{
		SetCollisionTestAABB();
		}
}


bool Asura_Physics_RigidBodyInstance::HasContacts( void ) const
{
	return false;
}



void Asura_Physics_RigidBodyInstance::TestForNewContacts(void) 
{
#if 0
	SetCollisionTestAABB( !Asura_Physics_System::IsMultithreaded() );
#else
	SetCollisionTestAABB( true );
#endif

	switch ( m_eCollisionProxy )
	{
		default:
		case COLLISION_PROXY_POLYGONAL:
		case COLLISION_PROXY_OBB:
		{
#if 0
			TestForNewContacts_OBB();
#else

			if (!GetDefinition()) return;
			u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
			for (u_int u=0; u<uNumberOfObjectVertices; u++)
			{
				m_pxObjectVertexPosition[u] = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation + m_xInstancePosition;
			}

			TestForNewContacts_ObjectVertices();
			TestForNewContacts_EnvironmentVertices();
			TestForNewContacts_Edges();
#endif
			break;
		}
		case COLLISION_PROXY_SPHERE:
		case COLLISION_PROXY_SPHERESTACK:
		{
		
			TestForNewContacts_SphereStack();
			break;
		}
	}

#ifdef ASURA_PHYSICS_USE_OLD_SOLVER

	// if we have any contacts, resolve velocities and then positions
	if ( m_xContactClusterList.GetFirstCluster() )
	{ 
		if (GetActorMoveAsVehicle())
		{
			CalculateAllCollisionResponses( false );
		}
		else
		{
			for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
			{
				CalculateCollisionResponses( pxCluster, false );
			}
		}

		RelaxPositionConstraints();
	}
#endif
}

void Asura_Physics_RigidBodyInstance::TestForNewContacts_OBB( void )
{
	
	if ( m_xEnvironmentCache.GetFaceList() )
	{
		Asura_AVector3 xCentre = m_xBodyState.m_xCentreOfMass;
		Asura_AVector3 xHalfExtents = GetHalfExtents();

		Asura_OBB xOBB( xHalfExtents, xCentre, GetOrientation() );

		for ( Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvironmentCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
		{
			Asura_Physics_Tri xTri( pxFace->m_uFaceID);

			Asura_AVector3 xNormal;
			float fPenetration;

			if ( Asura_Physics_CCD::DoesOBBIntersectTri( xOBB,xTri,fPenetration,xNormal) )
			{
				Asura_AVector3 axContactPoints[10];
				u_int uNumberOfContactPoints;

				uNumberOfContactPoints = Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( axContactPoints, this, xTri, xNormal );

				for ( u_int u=0; u<uNumberOfContactPoints; ++u )
				{

					Asura_AVector3 xRel = axContactPoints[u] - m_xBodyState.m_xCentreOfMass;

					Asura_Assert ( xRel.Magnitude() < 3.0f, "Code bug: force point far from object");

					if (xRel.Magnitude()>3.0f)
					{
						uNumberOfContactPoints = Asura_Physics_CCD::FindContactPoints_PrimitiveVsTri( axContactPoints, this, xTri, xNormal );

					}

					AddContactPosition( axContactPoints[u], xNormal, fPenetration );
				}

				if ( uNumberOfContactPoints )
				{
					// TODO - make average of contact points?
					AddEnvironmentCollisionReport( pxFace->m_uFaceID, xNormal, axContactPoints[0] );
				}

		
			}
		}
	}
}


void Asura_Physics_RigidBodyInstance::TestForNewContacts_SphereStack( void )
{
	Asura_Physics_SphereGroup xSphereStack( *this );

	const float fSphereRadius = xSphereStack.GetSphereRadius();
	const float fSphereRadiusSquared = fSphereRadius * fSphereRadius;

	const bool bIsSphereStack = xSphereStack.GetNumberOfSpheres() > 1;

	Asura_AVector3 xPrimaryAxis;
	xPrimaryAxis.Zero();
	float fPrimaryAxisLength = 0.0f;

    if ( bIsSphereStack )
	{
		float fX = GetHalfExtents().GetX();
		float fY = GetHalfExtents().GetY();
		float fZ = GetHalfExtents().GetZ();

		if ( fX > fY )
		{
			if ( fX > fZ )
			{
				// it's X
				xPrimaryAxis.SetX( 1.0f );
				fPrimaryAxisLength = fX;
			}
			else // it's Z
			{
				xPrimaryAxis.SetZ( 1.0f );
				fPrimaryAxisLength = fZ;
			}
		}
		else if ( fY > fZ )
		{
			// it's Y
			xPrimaryAxis.SetY( 1.0f );
			fPrimaryAxisLength = fY;
		}
		else // it's Z
		{
			xPrimaryAxis.SetZ( 1.0f );
			fPrimaryAxisLength = fZ;
		}

		xPrimaryAxis = xPrimaryAxis * GetOrientation();
	}


	// test for intersection of the spheres with each poly in the environment cache
	for ( Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvironmentCache.GetFaceList(); pxFace; pxFace = pxFace->m_pxNext )
	{
		const Asura_AVector3& xPV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xPV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xPV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		Asura_AVector3 xNormal = ((xPV1 - xPV0)^(xPV2 - xPV0));
		xNormal.Normalise();

		
		for ( u_int v=0; v<xSphereStack.GetNumberOfSpheres(); ++v )
		{
			bool bFacingTowards;
			float fHitDistance;

			if( Asura_Physics_System::RayIntersectsTri( xSphereStack.GetSphereCentres()[v], -xNormal, xPV0, xPV1, xPV2, bFacingTowards, fHitDistance ) )
			{
				if ( fHitDistance >= 0.0f && bFacingTowards )
				{
					const float fPenetration = fSphereRadius - fHitDistance;

					if ( fPenetration > 0.0f  )
					{
						Asura_AVector3 xCollisionPoint = xSphereStack.GetSphereCentres()[v] - xNormal * fHitDistance;
						
						AddEnvironmentCollisionReport( pxFace->m_uFaceID, xNormal, xCollisionPoint );

						AddContactPosition( xCollisionPoint, xNormal, fPenetration );

					}
				}	
			}
		}

		// if we're a sphere-stack do an extra test along the primary axis (catches cases where individual spheres are on the other side of polys)
		if (bIsSphereStack)
		{
			bool bFacingTowards;
			float fHitDistance;

			if( Asura_Physics_System::RayIntersectsTri( m_xBodyState.m_xCentreOfMass, xPrimaryAxis, xPV0, xPV1, xPV2, bFacingTowards, fHitDistance ) )
			{
				if ( fHitDistance >= 0.0f && bFacingTowards )
				{
					const float fPenetration = fPrimaryAxisLength - fHitDistance;

					if ( fPenetration > 0.0f  )
					{
						Asura_AVector3 xCollisionPoint = m_xBodyState.m_xCentreOfMass + xPrimaryAxis * fHitDistance;

						AddEnvironmentCollisionReport( pxFace->m_uFaceID, -xPrimaryAxis, xCollisionPoint );

						AddContactPosition( xCollisionPoint, -xPrimaryAxis, fPenetration );
					}
				}	
				else if ( fHitDistance < 0.0f && !bFacingTowards )
				{
					const float fPenetration = fPrimaryAxisLength + fHitDistance;

					if ( fPenetration > 0.0f  )
					{
						Asura_AVector3 xCollisionPoint = m_xBodyState.m_xCentreOfMass + xPrimaryAxis * fHitDistance;

						AddEnvironmentCollisionReport( pxFace->m_uFaceID, xPrimaryAxis, xCollisionPoint );

						AddContactPosition( xCollisionPoint, xPrimaryAxis, fPenetration );
					}
				}	
			}
			
		}
	}

	for ( Asura_Physics_EnvironmentCache_Edge* pxEdge = m_xEnvironmentCache.GetEdgeList(); pxEdge; pxEdge = pxEdge->m_pxNext )
	{
		const Asura_AVector3& xEdgeV0 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
		const Asura_AVector3& xEdgeV1 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );


		for ( u_int v=0; v<xSphereStack.GetNumberOfSpheres(); ++v )
		{
			Asura_AVector3 xPointOnEdge;
			Asura_AVector3 xSphereCentre = xSphereStack.GetSphereCentres()[v];

            Asura_Maths::GetNearestPointOnLineSegment( xEdgeV0, xEdgeV1, xSphereCentre, xPointOnEdge );

			Asura_AVector3 xDisplacement = (xSphereCentre - xPointOnEdge);

			float fDistanceSquared = xDisplacement.MagnitudeSquared();

			if ( fDistanceSquared > 0.0f && fDistanceSquared < fSphereRadiusSquared )
			{
				const float fDistance = xDisplacement.NormaliseReturningMagnitude();

				const float fPenetration = fSphereRadius - fDistance;

				AddContactPosition( xPointOnEdge, xDisplacement, fPenetration );
			}
		}

	}

}

void Asura_Physics_RigidBodyInstance::TestForNewContacts_ObjectVertices( void )
{
	const Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvironmentCache.GetFaceList();

	if (!pxFace) return;

	while (pxFace)
	{
		float fDistance;
		bool bFacingTowards;


		const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		if ( !Asura_Physics_Environment::DoesFaceIntersectAABB( xV0, xV1, xV2, m_xBoundingBox ) )
		{
			pxFace = pxFace->m_pxNext;
			continue;
		}

		Asura_AVector3 xNormal = ((xV1 - xV0)^(xV2 - xV0));
		xNormal.Normalise();


		for (u_int u=0; u<m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices(); u++)
		{
			const Asura_AVector3 xVertexNormal = m_pxDefinition->m_pxObjectMesh->GetVertexNormal(u) * m_xBodyState.m_xOrientation;

			if (xVertexNormal*xNormal < -0.1f)
			{
				{
					const Asura_AVector3& xEnd = m_pxObjectVertexPosition[u];

					Asura_AVector3 xStart = m_pxObjectVertexPrevPosition[u];

					Asura_AVector3 xDisp = xEnd - xStart;
					xStart -= xDisp;
					const float fLength = xDisp.NormaliseReturningMagnitude() * 2.0f;	
					if (fLength != 0.0f)
					{
						// if vertex points away from direction of motion, ignore it
						//if (xVertexNormal * xDisp <= 0.0f) continue;

						float fPenetration = - (xEnd - xV0) * xNormal;

						if (fPenetration>=0.0f)
						{

							bool bIntersect = Asura_Physics_System::RayIntersectsTri(xStart, xDisp, xV0, xV1, xV2, bFacingTowards, fDistance);

							if (bIntersect && bFacingTowards && fDistance >= 0.0f && fDistance <= fLength)
							{
								Asura_AVector3 xContactPoint = xEnd;

								AddEnvironmentCollisionReport( pxFace->m_uFaceID, xNormal, xContactPoint );
								AddContactPosition( xContactPoint, xNormal, fPenetration );
			

								continue;
							}
						}
					}
				}

				// test for embedded objects
#if 01
				{
					const Asura_AVector3& xEnd = m_pxObjectVertexPosition[u];

					float fPenetration = - (xEnd - xV0) * xNormal;

					if (fPenetration>=0.0f)
					{
						const Asura_AVector3& xStart = GetCollisionMeshCentre();//xEnd + xNormal;

						Asura_AVector3 xDisp = xEnd - xStart;
						const float fLength = xDisp.NormaliseReturningMagnitude();	
						if (fLength == 0.0f)
						{
							continue;
						}

						bool bIntersect = Asura_Physics_System::RayIntersectsTri(xStart, xDisp, xV0, xV1, xV2, bFacingTowards, fDistance);

						if (bIntersect && bFacingTowards && fDistance >= 0.0f && fDistance <= fLength)
						{
							Asura_AVector3 xContactPoint = xEnd;
								
							AddEnvironmentCollisionReport( pxFace->m_uFaceID, xNormal, xContactPoint );
							AddContactPosition( xContactPoint, xNormal, fPenetration );

							continue;
						}
					}
				}
#endif
			}
		}

		pxFace = pxFace->m_pxNext;
	}

}

void Asura_Physics_RigidBodyInstance::TestForNewContacts_Edges( void )
{
#if 01
	// get object's collision mesh centre - we may want to cache this per object but it's another 16 bytes
	const Asura_AVector3 xObjectCentre = GetCollisionMeshCentre();

	// check for edges
	const Asura_Physics_RigidBody_ObjectMesh* pxObjectMesh = m_pxDefinition->m_pxObjectMesh;

	for (u_int u=0; u<pxObjectMesh->GetNumberOfEdges(); u++)
	{
		const Asura_Physics_RigidBody_ObjectEdge& xEdge = pxObjectMesh->GetEdge(u);
	
		const Asura_AVector3& xEdge0 = m_pxObjectVertexPosition[xEdge.GetFirstVertexIndex()];
		const Asura_AVector3& xEdge1 = m_pxObjectVertexPosition[xEdge.GetSecondVertexIndex()];

		Asura_Physics_EnvironmentCache_Edge* pxEdge = m_xEnvironmentCache.GetEdgeList();

		while (pxEdge)
		{
			const Asura_AVector3& xStart = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
			const Asura_AVector3& xEnd = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );

			if (!Asura_Physics_Environment::DoesEdgeIntersectAABB( xStart, xEnd, m_xBoundingBox ))
			{
				pxEdge = pxEdge->m_pxNext;
				continue;
			}


			Asura_AVector3 xDisp = xEnd - xStart;
			float fLength = xDisp.NormaliseReturningMagnitude();
			Asura_Assert(fLength != 0.0f, "invalid vector passed to Normalise");

			float fDistance;
			bool bFacingTowards;

			bool bIntersect =	Asura_Physics_System::RayIntersectsTri
								(
									xStart,
									xDisp,
									xEdge0,
									xEdge1,
									xObjectCentre,
									bFacingTowards,
									fDistance
								);
/*
			if (!bIntersect)
			{
				bIntersect =	Asura_Physics_System::RayIntersectsTri
								(
									xStart,
									xDisp,
									xEdge1,
									xPrevEdge1,
									xPrevEdge0,
									bFacingTowards,
									fDistance
								);
			}

			if (!bIntersect)
			{
				bIntersect =	Asura_Physics_System::RayIntersectsTri
								(
									xStart,
									xDisp,
									xEdge0,
									xEdge1,
									xPrevEdge0,
									bFacingTowards,
									fDistance
								);
			}
*/
			if (bIntersect && fDistance>=0.0f && fDistance<=fLength)
			{
				float fAlongObjectEdge, fAlongEnvEdge;

				if (Asura_Physics_System::SmallestVectorBetweenTwoLines(xEdge0, xEdge1, xStart, xEnd, fAlongObjectEdge, fAlongEnvEdge))
				{
					Asura_Assertf( fAlongObjectEdge>=0.0f && fAlongObjectEdge<=1.0f, ("object edge: %d", Asura_Maths::FtoD(fAlongObjectEdge)));
					Asura_Assertf( fAlongEnvEdge>=0.0f && fAlongEnvEdge<=1.0f, ("env edge: %d", Asura_Maths::FtoD(fAlongEnvEdge)));

					// the position where the two edges meet (naturally this is on the non-moving environment edge)
					Asura_AVector3 xEnvironmentPoint = xStart + (xEnd-xStart) * fAlongEnvEdge;

					// given the point on the object edge which will be the point of contact, where would
					// that point be if the collision hadn't taken place?
					Asura_AVector3 xContactPoint =  xEdge0 + (xEdge1-xEdge0) * fAlongObjectEdge;

					Asura_AVector3 xNormal = (xEnd-xStart) ^ (xEdge1 - xEdge0);
					xNormal.Normalise();

					const Asura_AVector3 xEdgeNormal = xEdge.GetNormal() * m_xBodyState.m_xOrientation;

					if (xNormal * xEdgeNormal > 0.0f)
					{
						xNormal = -xNormal;
					}

					const float fPenetration = (xEnvironmentPoint - xContactPoint) * xNormal;

					//Asura_Assert( fPenetration>=0.0f,"-ve penetration");
					if (fPenetration > 0.0f)
					{
						AddContactPosition( xContactPoint, xNormal, fPenetration );
					}

				}
			}



			pxEdge = pxEdge->m_pxNext;
			
		}

	}
#endif
}

void Asura_Physics_RigidBodyInstance::TestForNewContacts_EnvironmentVertices( void )
{
	// check for environment vertices inside the body's mesh

	Asura_Physics_EnvironmentCache_Vertex* pxVertex = m_xEnvironmentCache.GetVertexList();
	const Asura_Physics_RigidBody_ObjectMesh* pxMesh = m_pxDefinition->m_pxObjectMesh;

	while (pxVertex)
	{
		const Asura_AVector3& xVertexPosition = Asura_Physics_Environment::GetVertex(pxVertex->m_uVertexID);

		if (!m_xBoundingBox.Contains( *(const Asura_Vector_3*)(&xVertexPosition) ) )
		{
			pxVertex = pxVertex->m_pxNext;
			continue;
		}
		// test vertex against every face in mesh. 

		// convex hulls only?

		// assume penetration until proven wrong
		// optimisation: use previously rejected face first?
		bool bPenetration = true;

		float fLeastPenetration = Asura_Maths::LargeFloat;
		u_int uLeastPenetratedFace = 0;

		for (u_int u=0; u<pxMesh->GetNumberOfFaces(); u++)
		{
			Asura_AVector3 xFaceNormal;
			GetInstanceFaceNormal( xFaceNormal, u );

			const Asura_AVector3& xPointOnFacePosition = GetPositionOnInstanceFace(u);

			float fPenetration = (xPointOnFacePosition - xVertexPosition) * xFaceNormal;

			if (fPenetration < 0.0f) 
			{
				// no penetration with this face; vertex is outside mesh
				bPenetration = false;
				break;
			}
			else
			{
				if (fLeastPenetration > fPenetration)
				{
					fLeastPenetration = fPenetration;
					uLeastPenetratedFace = u;
				}
			}
		}

		if (bPenetration)
		{

			Asura_AVector3 xNormal;

			GetInstanceFaceNormal( xNormal, uLeastPenetratedFace );

			Asura_AVector3 xContactPoint = xVertexPosition;
			AddContactPosition( xContactPoint, -xNormal, fLeastPenetration );

		}

		pxVertex = pxVertex->m_pxNext;
	}

}


#if 0 // shouldn't be needed any more, but keeping around out of paranoia
/*
float Asura_Physics_RigidBodyInstance::TestForNewContacts_LeastDistance_ObjectVertices( const Asura_AVector3& xDisplacement, const float fDistance )
{
	float fLeastDistance = fDistance;
	bool bHitSomething = false;

	const Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvironmentCache.GetFaceList();

	while (pxFace)
	{


		const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		if ( !Asura_Physics_Environment::DoesFaceIntersectAABB( xV0, xV1, xV2, m_xBoundingBox ) )
		{
			pxFace = pxFace->m_pxNext;
			continue;
		}


		Asura_AVector3 xFaceNormal = ((xV1 - xV0)^(xV2 - xV0));
		xFaceNormal.Normalise();
	
		// is vertex heading towards face?
		if (xDisplacement*xFaceNormal < 0.0f)
		{			
			for (u_int u=0; u<m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices(); u++)
			{

				const Asura_AVector3& xStart = m_pxObjectVertexPosition[u] - xDisplacement * fAXIALLYALIGNED_DIST_THRESHOLD;
				const Asura_AVector3 xEnd = m_pxObjectVertexPosition[u] + xDisplacement * (fLeastDistance + fAXIALLYALIGNED_DIST_THRESHOLD); 

				// is the end behind the plane of the face?
				float fPenetration = - (xEnd - xV0) * xFaceNormal;
				if (fPenetration>=0.0f)
				{
					// well, is the line intersecting the face then, and where?
					float fDistanceAllowed;
					bool bFacingTowards;
					bool bIntersect = Asura_Physics_System::RayIntersectsTri(xStart, xDisplacement, xV0, xV1, xV2, bFacingTowards, fDistanceAllowed);

					// probably don't have to check bFacingTowards... indeed we could assert on it! (or make an optimised fn which doesn't
					// use that parameter)
				
					if (bIntersect && bFacingTowards && fDistanceAllowed >= 0.0f)// && fDistanceAllowed <= fDistance)
					{
						fDistanceAllowed -= fAXIALLYALIGNED_DIST_THRESHOLD;


						// distance to intersection is fDistance
						if (fLeastDistance > fDistanceAllowed)
						{
							fLeastDistance = fDistanceAllowed ;
							xLastCollisionNormal = xFaceNormal;
							uLastCollisionFaceID = pxFace->m_uFaceID;
							bHitSomething = true;
						}
					}
				}
			}
		}

		pxFace = pxFace->m_pxNext;
	}

	// If we've hit something, exaggerate how close the obstacle is. This is so the body
	// doesn't move to exactly touch it, but stops a little before it.
	if ( bHitSomething )
	{
		fLeastDistance -= fAXIALLYALIGNED_DIST_THRESHOLD;
	}
	return fLeastDistance;
}

float Asura_Physics_RigidBodyInstance::TestForNewContacts_LeastDistance_Edges( const Asura_AVector3& xDisplacement, const float fDistance )
{
	if (!m_xEnvironmentCache.GetEdgeList()) return fDistance;

	bool bHitSomething = false;
	float fLeastDistance = fDistance;

	Asura_AVector3 xObjectCentreDoubled;

	xObjectCentreDoubled.Zero();

	for (u_int u=0; u<m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices(); u++)
	{
		xObjectCentreDoubled += m_pxObjectVertexPosition[u];
	}

	xObjectCentreDoubled /= 0.5f * (float)m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();

	// check for edges
	const Asura_Physics_RigidBody_ObjectMesh* pxObjectMesh = m_pxDefinition->m_pxObjectMesh;

	for (u_int u=0; u<pxObjectMesh->GetNumberOfEdges(); u++)
	{
		const Asura_Physics_RigidBody_ObjectEdge& xEdge = pxObjectMesh->GetEdge(u);
	
		const Asura_AVector3& xEdge0 = m_pxObjectVertexPosition[xEdge.GetFirstVertexIndex()];
		const Asura_AVector3& xEdge1 = m_pxObjectVertexPosition[xEdge.GetSecondVertexIndex()];

		const Asura_AVector3 xEdgeNormal = (xEdge0+xEdge1) - xObjectCentreDoubled;
		
		if (xEdgeNormal * xDisplacement < 0.0f) continue;
		
		const Asura_AVector3& xEndEdge0 = xDisplacement*fLeastDistance + xEdge0;
		const Asura_AVector3& xEndEdge1 = xDisplacement*fLeastDistance + xEdge1;


		Asura_AVector3 xEdgeVector = xEdge1 - xEdge0;
		xEdgeVector.Normalise();

		Asura_Physics_EnvironmentCache_Edge* pxEdge = m_xEnvironmentCache.GetEdgeList();


		// KJFL - imagine I can make this faster if the while loop is on the outside
		while (pxEdge)
		{


			const Asura_AVector3& xStart = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
			const Asura_AVector3& xEnd = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );

			if (!Asura_Physics_Environment::DoesEdgeIntersectAABB( xStart, xEnd, m_xBoundingBox ))
			{
				pxEdge = pxEdge->m_pxNext;
				continue;
			}

		
			// test to see if the line segment formed by the environment edge intersects the parallelogram
			// formed by sweeping the object edge through it's proposed motion
			
			Asura_AVector3 xDisp = xEnd - xStart;
			float fLength = xDisp.NormaliseReturningMagnitude();
			Asura_Assert(fLength != 0.0f, "invalid vector");
			
			float fIntersectionDistance;
			bool bFacingTowards;

			// consider the parallelogram as two triangles, and test separately

			bool bIntersect =	Asura_Physics_System::RayIntersectsTri
								(
									xStart,
									xDisp,
									xEdge0,
									xEdge1,
									xEndEdge0,
									bFacingTowards,
									fIntersectionDistance
								);

			// if the line segment (environment edge) didn't go through this tri, try the other tri

			if (!(bIntersect && fIntersectionDistance>=0.0f && fIntersectionDistance<=fLength))
			{
				bIntersect =	Asura_Physics_System::RayIntersectsTri
								(
									xStart,
									xDisp,
									xEndEdge0,
									xEndEdge1,
									xEdge1,
									bFacingTowards,
									fIntersectionDistance
								);
			}

			if (bIntersect && fIntersectionDistance>=0.0f && fIntersectionDistance<=fLength)
			{
				// we have an intersection at this position:
				const Asura_AVector3 xIntersection = xStart + xDisp*fIntersectionDistance;
				
				// KJFL - 23 September 2004 - 
				//
				// the question is, how far does the object edge have to move before it reaches this point?
				//
				// equation of line segment (vectors capitals, scalars lower case)
				//
				// R = A + l * B
				//
				// if line segment moves through space in direction D, a point on this plane is
				//
				// R = A + l * B + m * D
				//
				// since we know R in this case is the intersection point...
				//
				// m * D = R - A - l * B
				//
				// cross product both sides with B, as B^B = 0...
				//
				// m * (D ^ B) = (R - A) ^ B
				//
				// Therefore if D is a normalised vector, the distance m the edge travels before it hits the
				// other edge is
				//
				// m = |( (R - A) ^ B )| / |(D ^ B)|
				//
				// obviously there are zero terms to check for there, which indicate m=0 and/or the direction of travel
				// is parallel to the object edge. The latter is something we can probably early reject on (it'd be a
				// vertex hit instead, I imagine).

				Asura_AVector3 xDivisor = xDisplacement ^ xEdgeVector;

				float fDivisor = xDivisor.Magnitude();

				if (fDivisor != 0.0f)
				{
					Asura_AVector3 xNumerator = ((xIntersection - xEdge0) ^ xEdgeVector);
					float fNumerator = xNumerator.Magnitude();

					float fDistanceAllowed =  fNumerator / fDivisor;

					if (fDistanceAllowed < fLeastDistance)
					{
						fLeastDistance = fDistanceAllowed ;

						// make a collision 'normal' by calculating a vector perpendicular to both edges
						xLastCollisionNormal = xEdgeVector ^ xDisp;
						// KOR - 30/01/2006 - cross-product doesn't guarantee a normalised vector.
						xLastCollisionNormal.Normalise();
						// Also do an extra check in case the cross product gave us complete rubbish
						Asura_Assert(0.99f <= xLastCollisionNormal.Magnitude() && xLastCollisionNormal.Magnitude() <= 1.01f, "xLastCollisionNormal not normalised");
						
						// let's choose the normal as pointing away from direction of motion
						if (xDisplacement * xLastCollisionNormal > 0.0f)
						{
							xLastCollisionNormal = -xLastCollisionNormal;
						}
						bHitSomething = true;
					}
				}
			}



			pxEdge = pxEdge->m_pxNext;
			
		}

	}
	
	// If we've hit something, exaggerate how close the obstacle is. This is so the body
	// doesn't move to exactly touch it, but stops a little before it.
	if ( bHitSomething )
	{
		fLeastDistance -= fAXIALLYALIGNED_DIST_THRESHOLD;
	}
	return fLeastDistance;
}

float Asura_Physics_RigidBodyInstance::TestForNewContacts_LeastDistance_EnvironmentVertices( const Asura_AVector3& xDisplacement, const float fDistance )
{
	if (!m_xEnvironmentCache.GetVertexList()) return fDistance;

	bool bHitSomething = false;

	float fLeastDistance = fDistance;
	const Asura_AVector3 xReflectedDisplacement = -xDisplacement;

	const Asura_Physics_RigidBody_ObjectMesh* pxMesh = m_pxDefinition->m_pxObjectMesh;

	for (u_int u=0; u<pxMesh->GetNumberOfFaces(); u++)
	{
		const Asura_AVector3& xV0 = m_pxObjectVertexPosition[pxMesh->m_pusFaceVertexIndices[0+u*3]];
		const Asura_AVector3& xV1 = m_pxObjectVertexPosition[pxMesh->m_pusFaceVertexIndices[1+u*3]];
		const Asura_AVector3& xV2 = m_pxObjectVertexPosition[pxMesh->m_pusFaceVertexIndices[2+u*3]];

		Asura_AVector3 xFaceNormal = ((xV1 - xV0)^(xV2 - xV0));
		xFaceNormal.Normalise();

		// if face normal is counter the direction of movement, then we can ignore this face 
		if (xDisplacement*xFaceNormal < 0.0f) continue;

		Asura_Physics_EnvironmentCache_Vertex* pxVertex = m_xEnvironmentCache.GetVertexList();
	
		while (pxVertex)
		{
			// consider the face to be stationary and the vertex to move towards the face (reflect displacement vector)

			const Asura_AVector3& xPos = Asura_Physics_Environment::GetVertex(pxVertex->m_uVertexID); 

			if (!m_xBoundingBox.Contains( *(const Asura_Vector_3*)(&xPos) ) )
			{
				pxVertex = pxVertex->m_pxNext;
				continue;
			}

			const Asura_AVector3& xStart = xPos - xReflectedDisplacement * fAXIALLYALIGNED_DIST_THRESHOLD;
			const Asura_AVector3& xEnd = xPos + xReflectedDisplacement * (fLeastDistance + fAXIALLYALIGNED_DIST_THRESHOLD);
		

			// is the end behind the plane of the face?
			float fPenetration = - (xEnd - xV0) * xFaceNormal;
			if (fPenetration>=0.0f)
			{
				// well, is the line intersecting the face then, and where?
				float fDistanceAllowed;
				bool bFacingTowards;
				bool bIntersect = Asura_Physics_System::RayIntersectsTri(xStart, xReflectedDisplacement, xV0, xV1, xV2, bFacingTowards, fDistanceAllowed);

				// probably don't have to check bFacingTowards... indeed we could assert on it! (or make an optimised fn which doesn't
				// use that parameter)

				if (bIntersect && bFacingTowards && fDistanceAllowed >= 0.0f)
				{
					fDistanceAllowed -= fAXIALLYALIGNED_DIST_THRESHOLD;

					// distance to intersection is fDistance
					if (fLeastDistance > fDistanceAllowed)
					{
						fLeastDistance = fDistanceAllowed;
						xLastCollisionNormal = -xFaceNormal;
						bHitSomething = true;
					}
				}
			}

			pxVertex = pxVertex->m_pxNext;
		}
	}

	// If we've hit something, exaggerate how close the obstacle is. This is so the body
	// doesn't move to exactly touch it, but stops a little before it.
	if ( bHitSomething )
	{
		fLeastDistance -= fAXIALLYALIGNED_DIST_THRESHOLD;
	}
	return fLeastDistance;

}

*/
#endif


class VelocityContact
{
public:

	Asura_AVector3 m_xRel;
	Asura_AVector3 m_xNormal;

	float m_fRequiredNormalSpeed;

	float m_fInvNormalDenom;

};


void Asura_Physics_RigidBodyInstance::CalculateCollisionResponses( const Asura_Physics_ContactCluster* pxCluster, bool bInelastic )
{
	if ( IsActor() && !GetActorMoveAsVehicle() ) return;
	
	Asura_AVector3 xCentreOfMass;

	// TODO - wouldn't need to do this if could set CoM offset
#if 01
	if (IsActor() && GetActorMoveAsVehicle())
	{
		// fake CoM to be at pivot of Actor
		CalculateRenderingPosition();
		xCentreOfMass = m_xInstancePosition;
	}
	else
#endif
	{
		xCentreOfMass = m_xBodyState.m_xCentreOfMass;
	}

	float fFriction = Asura_Maths::Min(m_fFriction,1.0f);

	const Asura_AVector3& xNormal = pxCluster->GetContactNormal();

	const u_int uMAX_NUMBER_OF_CONTACTS = uASURA_CONTACTCLUSTER_MAXCONTACTS;
	const u_int uNumberOfContacts = Asura_Maths::Min( pxCluster->GetNumberOfContacts(), uMAX_NUMBER_OF_CONTACTS );

	VelocityContact axContact[uMAX_NUMBER_OF_CONTACTS];

	Asura_AVector3 xCentre;

	xCentre.Zero();
	float fNumVerts=0.0f;
	float fFrictionNormalImpulse = 0.0f;

	const Asura_AMatrix3x3 xInvInertiaTensor = GetInvInertiaTensorDiagonal();

	for (u_int u=0; u<uNumberOfContacts; ++u)
	{
		VelocityContact& xContact = axContact[u];

		xContact.m_xRel = pxCluster->GetContactPosition(u) - xCentreOfMass;

		{
			Asura_AVector3 xDenom = axContact[u].m_xRel^xNormal;

			xDenom = xDenom*xInvInertiaTensor;

			xDenom = xDenom ^ axContact[u].m_xRel;

			xContact.m_fInvNormalDenom = GetInvMass() + xDenom * xNormal;

			if (xContact.m_fInvNormalDenom != 0.0f )
			{
				xContact.m_fInvNormalDenom = 1.0f / xContact.m_fInvNormalDenom;
			}
		}
		

		Asura_AVector3 xVelocity;
		
		CalculateVelocityOfRelativePointOnBody( xContact.m_xRel, xVelocity );

		float fNormalVelocity = xVelocity * xNormal;

		{

			//float fSeparation = pxCluster->GetContactPosition(u)*xNormal - pxCluster->GetPlanePenetrationConstant();
			if (fNormalVelocity < 0.0f )//&& fSeparation < 0.000001f)
			{
				float fElasticity = bInelastic ? 0.0f : GetElasticity();

				if (fNormalVelocity > -0.01f) fElasticity = 0.0f;

				
				xContact.m_fRequiredNormalSpeed = -fNormalVelocity * fElasticity;

				fFrictionNormalImpulse += -fNormalVelocity *(1.0f + fElasticity ) * xContact.m_fInvNormalDenom;

				xCentre += pxCluster->GetContactPosition(u);
				fNumVerts+=1.0f;
			}
			else
			{
				xContact.m_fRequiredNormalSpeed = fNormalVelocity;
			}
		}

	}

	
	u_int uNumberOfIterations = (uNumberOfContacts==1) ? 1 : 10;

	for (u_int uIterations = 0; uIterations<uNumberOfIterations; ++uIterations)
	{
		for (u_int u=0; u<uNumberOfContacts; ++u)
		{
			VelocityContact& xContact = axContact[u];
		

			Asura_AVector3 xVelocity;
			CalculateVelocityOfRelativePointOnBody( xContact.m_xRel, xVelocity );

			float fNormalVelocity = xVelocity * xNormal;

			{
				if (xContact.m_fInvNormalDenom > 0.0f) 
				{
					float fNumerator = xContact.m_fRequiredNormalSpeed - fNormalVelocity;

					float fNormalImpulse = fNumerator * xContact.m_fInvNormalDenom;

					const Asura_AVector3 xImpulse = xNormal * fNormalImpulse;

					ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );
				}
			}
		}
	}
#if 01

		if (m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
		{
			m_xBodyState.m_xAngularVelocity.Zero();
		}

		if (fNumVerts!=0.0f )
		{
			xCentre /= fNumVerts;
			fFrictionNormalImpulse /= fNumVerts;

			Asura_AVector3 xVelocity = CalculateVelocityOfPointOnBody( xCentre );
			Asura_AVector3 xRel = xCentre - xCentreOfMass;
			float fNormalVelocity = xVelocity * xNormal;

			Asura_AVector3 xTangent = (xVelocity - xNormal * ( fNormalVelocity ));
			float fTangentialSpeed = xTangent.NormaliseReturningMagnitude();

			if (fTangentialSpeed > 0.0001f)
			{
				Asura_AVector3 xDenom = xRel^xTangent;

				xDenom = xDenom*xInvInertiaTensor;

				xDenom = xDenom ^ xRel;

				float fDenominator = GetInvMass() + xDenom * xTangent;

				//if (fDenominator <= 0.0f) return;

				float fFrictionImpulse = fTangentialSpeed / fDenominator;

				fFrictionImpulse = Asura_Maths::Min( fFrictionImpulse, Asura_Maths::Abs(fFrictionNormalImpulse) * GetFriction() );

				const Asura_AVector3 xImpulse = - xTangent * fFrictionImpulse;

				ApplyImpulseAtRelativePosition( xRel, xImpulse );

				m_xBodyState.m_xAngularVelocity = m_xBodyState.m_xAngularVelocity - xNormal * (m_xBodyState.m_xAngularVelocity*xNormal*fFriction*fTWIST_FRICTION_SCALE);
			}
			//	SetVelocitiesOfBody( m_xBodyState.m_xLinearVelocity, m_xBodyState.m_xAngularVelocity );
		}

	if (m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_xBodyState.m_xLinearVelocity.Zero();
	}
	
#endif

}

void Asura_Physics_RigidBodyInstance::CalculateAllCollisionResponses( bool bInelastic )
{
	if ( IsActor() && !GetActorMoveAsVehicle() ) return;

	Asura_AVector3 xCentreOfMass;

	// TODO - wouldn't need to do this if could set CoM offset
#if 01
	if (IsActor() && GetActorMoveAsVehicle())
	{
		// fake CoM to be at pivot of Actor
		CalculateRenderingPosition();
		xCentreOfMass = m_xInstancePosition;
	}
	else
#endif
	{
		xCentreOfMass = m_xBodyState.m_xCentreOfMass;
	}

	const u_int uMAX_NUMBER_OF_CONTACTS = 50;

	VelocityContact axContact[ uMAX_NUMBER_OF_CONTACTS ];
	u_int uContactsSoFar = 0;

	for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
	{
		u_int uNumberOfContacts = pxCluster->GetNumberOfContacts();

		const Asura_AVector3& xNormal = pxCluster->GetContactNormal();
		
		const Asura_AMatrix3x3 xInvInertiaTensor = GetInvInertiaTensorDiagonal();

		for (u_int u=0; u<uNumberOfContacts; ++u)
		{
			// skip the rest if we'd run out of space for these contacts 
			if ( u + uContactsSoFar >= uMAX_NUMBER_OF_CONTACTS )
			{
				uContactsSoFar = uMAX_NUMBER_OF_CONTACTS;
				break;
			}

			VelocityContact& xContact = axContact[u + uContactsSoFar];

			xContact.m_xNormal = xNormal;

			xContact.m_xRel = pxCluster->GetContactPosition(u) - xCentreOfMass;

			{
				Asura_AVector3 xDenom = xContact.m_xRel^xNormal;

				xDenom = xDenom*xInvInertiaTensor;

				xDenom = xDenom ^ xContact.m_xRel;

				xContact.m_fInvNormalDenom = GetInvMass() + xDenom * xNormal;

				if (xContact.m_fInvNormalDenom != 0.0f )
				{
					xContact.m_fInvNormalDenom = 1.0f / xContact.m_fInvNormalDenom;
				}
			}


			Asura_AVector3 xVelocity;

			CalculateVelocityOfRelativePointOnBody( xContact.m_xRel, xVelocity );

			float fNormalVelocity = xVelocity * xNormal;

			{

				//float fSeparation = pxCluster->GetContactPosition(u)*xNormal - pxCluster->GetPlanePenetrationConstant();
				if (fNormalVelocity < 0.0f )//&& fSeparation < 0.000001f)
				{
					float fElasticity = bInelastic ? 0.0f : GetElasticity();

					if (fNormalVelocity > -0.01f) fElasticity = 0.0f;


					xContact.m_fRequiredNormalSpeed = -fNormalVelocity * fElasticity;

				}
				else
				{
					xContact.m_fRequiredNormalSpeed = fNormalVelocity;
				}
			}
		}
		// skip the rest if we'd run out of space for these contacts 
		if ( uContactsSoFar >= uMAX_NUMBER_OF_CONTACTS ) break;
		
		uContactsSoFar += pxCluster->GetNumberOfContacts();
	}

	u_int uTotalNumberOfContacts = uContactsSoFar;

	u_int uNumberOfIterations = (uTotalNumberOfContacts==1) ? 1 : 10;

	for (u_int uIterations = 0; uIterations<uNumberOfIterations; ++uIterations)
	{
		for (u_int u=0; u<uTotalNumberOfContacts; ++u)
		{
			VelocityContact& xContact = axContact[u];


			Asura_AVector3 xVelocity;
			CalculateVelocityOfRelativePointOnBody( xContact.m_xRel, xVelocity );

			float fNormalVelocity = xVelocity * xContact.m_xNormal;

			{
				if (xContact.m_fInvNormalDenom > 0.0f) 
				{
					float fNumerator = xContact.m_fRequiredNormalSpeed - fNormalVelocity;

					float fNormalImpulse = fNumerator * xContact.m_fInvNormalDenom;

					const Asura_AVector3 xImpulse = xContact.m_xNormal * fNormalImpulse;

					ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );
				}
			}
		}
		
	}

	if (m_xBodyState.m_xAngularVelocity.Magnitude() < 0.0001f)
	{
		m_xBodyState.m_xAngularVelocity.Zero();
	}
	if (m_xBodyState.m_xLinearVelocity.Magnitude() < 0.0001f)
	{
		m_xBodyState.m_xLinearVelocity.Zero();
	}
}


void Asura_Physics_RigidBodyInstance::PositionProjection( const Asura_AVector3& xDisplacement, const Asura_AVector3 axContacts[], u_int uNumberOfContacts )
{
	bool bUseLinearProjection = IsActor() || dbg_bOnlyUseLinearProjection;

	if (bUseLinearProjection)
	{
		m_xBodyState.m_xCentreOfMass += xDisplacement;
	}
	else
	{
		Asura_AVector3 xPoint;
		xPoint.Zero();

		for (u_int u=0; u<uNumberOfContacts; ++u)
		{
			xPoint += axContacts[u];
		}
		xPoint /= (float)uNumberOfContacts;

		Asura_AVector3 xRel = xPoint - m_xBodyState.m_xCentreOfMass;

		// KJFL - 6 September 2007 - we want to share the movement between linear and angular components.
		// Since this is part of a relaxation step it does not need to remove a whole penetration in one go,
		// just improve matters. Scale factors determined empirically - TODO: there may be a more robust mathematical
		// way of doing this (as long as it's not too costly).
		// Currently using a small scale factor for the ang contribution since that produced good results...
		// we may want to cap the amount of angular as well / instead. 
		Asura_AVector3 xLinImpulse = xDisplacement * 0.5f;
		// order of calculations to reduce inaccuracy (i.e. inv inertia * mass in brackets).
		Asura_AVector3 xAngImpulse = ( xRel ^ (xDisplacement * 0.1f) ) * ( GetInvInertiaTensorDiagonal() * (1.0f/GetInvMass()) );

		m_xBodyState.m_xCentreOfMass += xLinImpulse;

		Asura_AMatrix3x3 xDelta;
		xDelta.Tilde( xAngImpulse );		
		xDelta = m_xBodyState.m_xOrientation*xDelta;
		m_xBodyState.m_xOrientation.Add( xDelta );
		m_xBodyState.m_xOrientation.Orthogonalise();

		HandleOrientationConstraints();

		CalculateWorldInvInertiaTensor();
	}
}


void Asura_Physics_RigidBodyInstance::RelaxPositionConstraints( void )
{
#if 1
	// Ok, do yet more iterations until penetration removed
	for( u_int uIterations=1; uIterations<10; ++uIterations)
	{
		bool bNoPenetrations = true;

		float fMaxPenetration = 0.0f;


		for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
		{
			const Asura_AVector3& xNormal = pxCluster->GetContactNormal();

			float fPenetration = pxCluster->GetPlanePenetrationConstant() - m_xBodyState.m_xCentreOfMass*xNormal;

			if (fPenetration > 0.0f)
			{
				Asura_AVector3 xDisplacement = (xNormal*(fPenetration + 0.0002f))*0.5f;
				
				PositionProjection( xDisplacement, &pxCluster->GetContactPosition(0), pxCluster->GetNumberOfContacts() );
				
				bNoPenetrations = false;
				fMaxPenetration = Asura_Maths::Max( fPenetration, fMaxPenetration );

			}
		}

		if ( bNoPenetrations ) return;

		// if you're not an actor and the penetration will be relatively small 
		// (we should have removed half of it already), let's stop.
		if ( !IsActor() && fMaxPenetration < 0.05f ) return;
	}
#else
	// Do a first pass of the contacts - this might be good enough for most rigid bodies,
	// where a small amount of "slop" penetration actually reduces jittering
	{
		bool bNoPenetrations = true;

		float fMaxPenetration = 0.0f;
	
		for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
		{
			const Asura_AVector3& xNormal = pxCluster->GetContactNormal();

			float fPenetration = pxCluster->GetPlanePenetrationConstant() - m_xBodyState.m_xCentreOfMass*xNormal;

			if (fPenetration > 0.0f)
			{
				Asura_AVector3 xDisplacement = (xNormal*(fPenetration + 0.0002f))*0.5f;
				m_xBodyState.m_xCentreOfMass += xDisplacement;
			
				bNoPenetrations = false;
				fMaxPenetration = Asura_Maths::Max( fPenetration, fMaxPenetration );
			}
		}
	
		// if there weren't any penetrations, let's not waste any more time here
		if ( bNoPenetrations ) return;

		// if you're not an actor and the penetration will be relatively small 
		// (we should have removed half of it already), let's stop.
		if ( !IsActor() && fMaxPenetration < 0.05f ) return;
	}

	// Ok, do yet more iterations until penetration removed
	for( u_int uIterations=1; uIterations<10; ++uIterations)
	{
		bool bNoPenetrations = true;

		for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
		{
			const Asura_AVector3& xNormal = pxCluster->GetContactNormal();

			float fPenetration = pxCluster->GetPlanePenetrationConstant() - m_xBodyState.m_xCentreOfMass*xNormal;

			if (fPenetration > 0.0f)
			{
				Asura_AVector3 xDisplacement = (xNormal*(fPenetration + 0.0002f))*0.5f;
				m_xBodyState.m_xCentreOfMass += xDisplacement;
				bNoPenetrations = false;
			}
		}

		if ( bNoPenetrations ) return;
	}
#endif
}

void Asura_Physics_RigidBodyInstance::AddContactPosition( const Asura_AVector3& xPosition, const Asura_AVector3& xNormal, float fPenetration )
{
	RegisterCollision( xNormal );

	m_xContactClusterList.AddContactPosition( m_xBodyState.m_xCentreOfMass, xPosition, xNormal, fPenetration );
}



void Asura_Physics_RigidBodyInstance::RegisterCollision( const Asura_AVector3& xNormal )
{
	const float	fNormalDotGravity	= xNormal * GetGravityDirection();

	// check if we've hit something floor-like
	if ( fNormalDotGravity < -fVERTICAL_BOUNDARY )
	{
		m_bIsInContactWithFloor = true;
		m_bIsSupported = true;
	}

	// flag this object as having hit something
	m_bHasHitSomething = true;
}

void Asura_Physics_RigidBodyInstance::ClearObstructionFlags( void )
{
	if (Asura_Physics_System::IsFirstUpdateThisFrame())
	{
		m_bHasHitSomething = false;
		m_bHasHitAnObject = false;
		m_bIsSupported = false;

		m_bWasInContactWithFloor = m_bIsInContactWithFloor;
		m_bIsInContactWithFloor = false;
	}
}

void Asura_Physics_RigidBodyInstance::HandleOrientationConstraints( void )
{
	if (m_bAxiallyAligned)
	{
		m_xBodyState.m_xOrientation.Identity();
	}	
	else if (m_bLockYAxisToBeVertical)
	{
		// "down" vector always along world Y axis:
		Asura_AVector3 xDown( 0.0f, 1.0f, 0.0f );

		// "forwards" vector goes from the bottom pt to the most +ve z vertex in local space,
		// but with any y axis component zero-ed out (effectively projecting it onto the XZ plane)
		Asura_AVector3 xForwards = m_xBodyState.m_xOrientation.GetColumn2();
		xForwards.SetY(0.0f);
		float fMagnitude = xForwards.NormaliseReturningMagnitude();

		Asura_AVector3 xRight;

		if ( fMagnitude > 0.0f )
		{
			// and now the "right" vector must be perpendicular to both of these: x = y^z
			xRight = xDown ^ xForwards;
		}
		else // "forward" vector was actually parallel to y-axis - have to use "right" vector instead
		{
			xRight = m_xBodyState.m_xOrientation.GetColumn0();
			xRight.SetY(0.0f);
			xRight.Normalise();

			// make "forward" perpendicular to "right" and "down"
			xForwards = xRight ^ xDown;
		}

		// hmm renormalise again? recalc up vector? need to check for orthogonality
		// fill out matrix
		m_xBodyState.m_xOrientation.SetColumn0(xRight);
		m_xBodyState.m_xOrientation.SetColumn1(xDown);
		m_xBodyState.m_xOrientation.SetColumn2(xForwards);
	}
	else if (m_bSelfRighting)
	{
		Asura_AVector3 xUpwards( 0.0f, 1.0f, 0.0f );

		Asura_AVector3 xInstanceUp = m_xBodyState.m_xOrientation.GetColumn1();

		Asura_Quat xQuat;

		xQuat.RotationFromAToB( xUpwards, xInstanceUp );

		Asura_AMatrix3x3 xToUpright;

		xQuat.ToMatrix( xToUpright );
		xToUpright.Transpose();

		// TODO - is this less maths to go with a "axis and angle" calc?
		xQuat.x = Asura_Maths::ClampToRange( xQuat.x, -fSELF_RIGHTING_CONSTRAINT, +fSELF_RIGHTING_CONSTRAINT );
		xQuat.y = Asura_Maths::ClampToRange( xQuat.y, -fSELF_RIGHTING_CONSTRAINT, +fSELF_RIGHTING_CONSTRAINT );
		xQuat.z = Asura_Maths::ClampToRange( xQuat.z, -fSELF_RIGHTING_CONSTRAINT, +fSELF_RIGHTING_CONSTRAINT );
		xQuat.w = Asura_Maths::Sqrt( 1.0f - (xQuat.x*xQuat.x + xQuat.y*xQuat.y + xQuat.z*xQuat.z) );

		Asura_AMatrix3x3 xToConstraint;
		xQuat.ToMatrix( xToConstraint );

		m_xBodyState.m_xOrientation = (m_xBodyState.m_xOrientation * xToUpright) * xToConstraint;
	}
}


void Asura_Physics_RigidBodyInstance::SetCollisionTestAABB( bool bUpdateEnvCache )
{
	// find the bounding box covering the old and new positions
	m_xBoundingBox.Reset();

	if (!GetDefinition()) return;

	switch ( m_eCollisionProxy )
	{
		case COLLISION_PROXY_SPHERE:
		case COLLISION_PROXY_SPHERESTACK:
		{
			// make a bounding box which contains the currently positioned sphere and the sphere
			// at its previous position, and then grow the bb by the sphere radius
			m_xBoundingBox.Set( Asura_Vector_3(m_xBodyState.m_xCentreOfMass) );
			m_xBoundingBox.Set( Asura_Vector_3(m_xBodyState.m_xPreviousCentreOfMass) );
			m_xBoundingBox.Grow( m_pxDefinition->GetMaximumRadius() );
	
			break;
		}
		default:
		{
			CalculateRenderingPosition();

			{
				// KJFL - 24 September 2008 - calculate the overall AABB from the OBB swept
				// between old and new positions.
				
				// Hmm, can't quite see how I can vectorise this nicely since I need to do those
				// abs on each bit. Possibly we need some sort of AbsDot method... who knows, some
				// hardware may have native vector instructions for this.
			
				float fWorldSpaceExtentX = Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsX() * m_xBodyState.m_xOrientation.GetColumn0().GetX() )
										 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsY() * m_xBodyState.m_xOrientation.GetColumn1().GetX() )
										 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsZ() * m_xBodyState.m_xOrientation.GetColumn2().GetX() );

				float fWorldSpaceExtentY = Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsX() * m_xBodyState.m_xOrientation.GetColumn0().GetY() )
										 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsY() * m_xBodyState.m_xOrientation.GetColumn1().GetY() )
										 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsZ() * m_xBodyState.m_xOrientation.GetColumn2().GetY() );

				float fWorldSpaceExtentZ = Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsX() * m_xBodyState.m_xOrientation.GetColumn0().GetZ() )
										 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsY() * m_xBodyState.m_xOrientation.GetColumn1().GetZ() )
										 + Asura_Maths::Abs( m_pxDefinition->GetOBBHalfExtentsZ() * m_xBodyState.m_xOrientation.GetColumn2().GetZ() );

			
				// KJFL - 24 September 2008 - we do have vector min and max methods, but since
				// the bounding box members aren't vectorised we'll have little gain in using 
				// them. Something to revisit if/when we have an AABB based on Asura_AVector.
				m_xBoundingBox.MinX = Asura_Maths::Min( m_xBodyState.m_xCentreOfMass.GetX(), m_xBodyState.m_xPreviousCentreOfMass.GetX() ) - fWorldSpaceExtentX;
				m_xBoundingBox.MaxX = Asura_Maths::Max( m_xBodyState.m_xCentreOfMass.GetX(), m_xBodyState.m_xPreviousCentreOfMass.GetX() ) + fWorldSpaceExtentX;

				m_xBoundingBox.MinY = Asura_Maths::Min( m_xBodyState.m_xCentreOfMass.GetY(), m_xBodyState.m_xPreviousCentreOfMass.GetY() ) - fWorldSpaceExtentY;
				m_xBoundingBox.MaxY = Asura_Maths::Max( m_xBodyState.m_xCentreOfMass.GetY(), m_xBodyState.m_xPreviousCentreOfMass.GetY() ) + fWorldSpaceExtentY;
			
				m_xBoundingBox.MinZ = Asura_Maths::Min( m_xBodyState.m_xCentreOfMass.GetZ(), m_xBodyState.m_xPreviousCentreOfMass.GetZ() ) - fWorldSpaceExtentZ;
				m_xBoundingBox.MaxZ = Asura_Maths::Max( m_xBodyState.m_xCentreOfMass.GetZ(), m_xBodyState.m_xPreviousCentreOfMass.GetZ() ) + fWorldSpaceExtentZ;
				}

			break;
		}
	}

	if (bUpdateEnvCache)
	{
		// give a 10cm extension when fill cache to minimise number of searches
		m_xEnvironmentCache.Update( m_xBoundingBox, 0.1f );
	}
}

void Asura_Physics_RigidBodyInstance::CalculateRenderingPosition( void ) const
{
	if (!m_pxDefinition) return;

	m_xInstancePosition = m_xBodyState.m_xCentreOfMass - m_pxDefinition->GetOBBCentreOffset() * m_xBodyState.m_xOrientation;
}

bool Asura_Physics_RigidBodyInstance::IsWideAwake( void ) const
{
	// static objects are always asleep
	if ( IsStatic() ) return false;

	if (m_pxRestFrame) return true;
	
	return m_bHasAwakePartner || (m_fTimeMotionless < fMOTIONLESS_TIME_THRESHOLD);
}

void Asura_Physics_RigidBodyInstance::CheckSleepStatus( void )
{
	if (!IsAsleep())
	{	
		Asura_AVector3 xDisplacement = m_xBodyState.m_xCentreOfMass - m_xBodyState.m_xPreviousCentreOfMass;		
		
		if (xDisplacement.Magnitude() < fMOTIONLESS_SPEED_THRESHOLD && m_xBodyState.m_xAngularVelocity.Magnitude() < fMOTIONLESS_ANGULAR_SPEED_THRESHOLD )
		{
			m_fTimeMotionless += Asura_Physics_System::GetTimeSlice();
		}
		else
		{
			m_fTimeMotionless = 0.0f;
		}

		if (!IsWideAwake())
		{
			// put object to sleep, and kill its momentum (so it won't jitter when it wakes up)
			ZeroMomentum();

			// flush any cached data to save memory
			FlushCachedData();

			m_bAsleep = true;
		}
	}
}


const Asura_AVector3 Asura_Physics_RigidBodyInstance::GetPositionOnInstanceFace( u_int uFace, u_int uVertex )
{
	return m_pxObjectVertexPosition[ m_pxDefinition->m_pxObjectMesh->GetVertexIndexOnFace( uFace, uVertex ) ];
}
void Asura_Physics_RigidBodyInstance::GetInstanceFaceNormal( Asura_AVector3& xNormal, const u_int uFace )
{
	xNormal = m_pxDefinition->m_pxObjectMesh->GetFaceNormal(uFace) * m_xBodyState.m_xOrientation;
}

void Asura_Physics_RigidBodyInstance::GetExtentsAlongAxis( const Asura_AVector3& xAxis, float& fMin, float& fMax ) const
{
	// project the OBB onto the axis given by xAxis, and find the min and max extents along that axis.
	const float fHalfExtent = Asura_Maths::Abs( xAxis * GetOrientation().GetColumn0() ) * GetDefinition()->GetOBBHalfExtentsX()
		+ Asura_Maths::Abs( xAxis * GetOrientation().GetColumn1() ) * GetDefinition()->GetOBBHalfExtentsY()
		+ Asura_Maths::Abs( xAxis * GetOrientation().GetColumn2() ) * GetDefinition()->GetOBBHalfExtentsZ();

	const float fProjectedCentre = GetCollisionMeshCentre() * xAxis;

	fMin = fProjectedCentre - fHalfExtent;
	fMax = fProjectedCentre + fHalfExtent;
}

void Asura_Physics_RigidBodyInstance::GetCentreAndHalfExtentAlongAxis( const Asura_AVector3& xAxis, float& fCentre, float& fHalfExtent ) const
{
	// project the OBB onto the axis given by xAxis, and find the new extents and centre position along that axis.
	fHalfExtent = Asura_Maths::Abs( xAxis * GetOrientation().GetColumn0() ) * GetDefinition()->GetOBBHalfExtentsX()
				+ Asura_Maths::Abs( xAxis * GetOrientation().GetColumn1() ) * GetDefinition()->GetOBBHalfExtentsY()
				+ Asura_Maths::Abs( xAxis * GetOrientation().GetColumn2() ) * GetDefinition()->GetOBBHalfExtentsZ();

	fCentre = GetCollisionMeshCentre() * xAxis;
}

void Asura_Physics_RigidBodyInstance::SetIgnore( bool bIgnore )
{
	m_bIgnore = bIgnore;

	// If we're now ignored, destroy all our cached data.
	// Otherwise when we become un-ignored we'll get dragged back 
	// to our old contacts etc.
	if ( bIgnore )
	{
		FlushCachedData();
	}

	// If we're a static object, we need to notify the main system so it can recalc the static object spatial hashes
	if ( IsStatic() )
	{
		Asura_Physics_System::NotifyOfChangeToStaticObjects();
	}
}

void Asura_Physics_RigidBodyInstance::SetUseAsRestFrame( bool bUseAsRestFrame )
{
	if ( bUseAsRestFrame )
	{
		if ( m_bOwnsRestFrame )
		{
			// we've already got one, thanks
			// nothing to do here
		}
		else // we don't own one already
		{
			// unlink from any frame we're currently part of
			UnlinkFromRestFrame();

			// and make our own
			CreateOwnRestFrame();
		}
	}
	else // if we have our own rest frame, destroy it
	{
		if ( m_bOwnsRestFrame )
		{
			DestroyOwnRestFrame();
		}
	}
}
void Asura_Physics_RigidBodyInstance::LinkToRestFrame( Asura_Physics_RestFrame* pxRestFrame )
{
	if (m_pxRestFrame == pxRestFrame ) return;

	UnlinkFromRestFrame();

	pxRestFrame->Link();
	m_pxRestFrame = pxRestFrame;
}

void Asura_Physics_RigidBodyInstance::UnlinkFromRestFrame( void )
{
	if ( m_pxRestFrame && !HasOwnRestFrame())
	{
		m_pxRestFrame->Unlink();
		m_pxRestFrame = 0;
	}
}

void Asura_Physics_RigidBodyInstance::MoveWithRestFrame( void )
{
	if ( !m_pxRestFrame || HasOwnRestFrame() )
	{
		return;
	}

	m_pxRestFrame->TransformPosition( m_xBodyState.m_xCentreOfMass );
	m_pxRestFrame->TransformOrientation( m_xBodyState.m_xOrientation );

	UnlinkFromRestFrame();
}

void Asura_Physics_RigidBodyInstance::CreateOwnRestFrame( void )
{
	m_pxRestFrame = Asura_Physics_RestFrame::Create( this );
	m_bOwnsRestFrame = true;
}

void Asura_Physics_RigidBodyInstance::DestroyOwnRestFrame( void )
{
	Asura_Physics_RestFrame::Destroy( m_pxRestFrame );
	m_pxRestFrame = 0;
	m_bOwnsRestFrame = false;
}

bool Asura_Physics_RigidBodyInstance::HasRestFrameFrom(const Asura_Physics_RigidBodyInstance* pxOtherInstance) const
{
	Asura_Assert(pxOtherInstance, "Asura_Physics_RigidBodyInstance::HasRestFrameFrom - other instance is null");
	if (pxOtherInstance == this && m_bOwnsRestFrame)
	{
		return true;
	}
	if (!m_bOwnsRestFrame && m_pxRestFrame)
	{
		if (pxOtherInstance && pxOtherInstance->m_bOwnsRestFrame && pxOtherInstance->m_pxRestFrame == m_pxRestFrame)
		{
			return true;
		}
	}
	return false;
}

const Asura_Physics_RestFrame* Asura_Physics_RigidBodyInstance::GetOwnedRestFrame() const
{
	return m_bOwnsRestFrame ? m_pxRestFrame : NULL;
}

bool Asura_Physics_RigidBodyInstance::HasMatchingRestFrame(const Asura_Physics_RestFrame* pxRestFrame) const
{
	Asura_Assert(pxRestFrame, "Asura_Physics_RigidBodyInstance::HasMatchingRestFrame - rest frame is null");
	// Reasonable to say the a NULL rest frame never matches.
	return pxRestFrame && (pxRestFrame == m_pxRestFrame);
}


// clear this object's collision reports
void Asura_Physics_RigidBodyInstance::ClearCollisionReports( void )
{
	// no need to release memory, just clear the list pointer
	m_pxObjectCollisionReports = 0;
	m_pxEnvironmentCollisionReports = 0;

	m_bReportedEnvCollisionFXThisFrame = false;
}

void Asura_Physics_RigidBodyInstance::AddObjectCollisionReport( const Asura_Physics_RigidBodyInstance* xCollidee )
{
	// set the basic collision flags, and then see if we want to store a report
	m_bHasHitSomething = true;
	m_bHasHitAnObject = true;

	if (!GetReportObjectCollisions()) return;

	// KJFL - 24 January - check to see if report with this object already exists? Not sure
	// what exact behaviour people might want... for now don't allow duplicates. Later
	// if extra info (e.g. positions, normals) is required, we may allow multiple reports
	{
		const Asura_Physics_ObjectCollisionReport* pxReport = m_pxObjectCollisionReports;

		while (pxReport)
		{
			// quit if we've seen this before
			if (pxReport->GetPhysicsHandle() == xCollidee->GetHandle())	return;
			
			pxReport = pxReport->GetNextReport();
		}
	}


	// allocate a report
	Asura_Physics_ObjectCollisionReport* pxReport = static_cast<Asura_Physics_ObjectCollisionReport*>(g_xObjectCollisionReportsPool.Allocate());

	if (!pxReport)
	{
		Asura_Assert( pxReport, "Code bug: out of memory?" );
		return;
	}

	// set the data held in the report
	pxReport->m_uEntityGuid = xCollidee->GetAssociatedEntityGuid();
	pxReport->m_iPhysicsHandle = xCollidee->GetHandle();

	// put the report at the head of the list
	pxReport->m_pxNextReport = m_pxObjectCollisionReports;
	m_pxObjectCollisionReports = pxReport;
}

void Asura_Physics_RigidBodyInstance::AddEnvironmentCollisionReport( u_int uFaceID, const Asura_AVector3& xNormal, const Asura_AVector3& xPosition )
{
	// set the basic collision flags, and then see if we want to store a report
	m_bHasHitSomething = true;

	#ifdef ASURA_USE_COLLISIONFX
	if ( GetReportEnvironmentCollisionFX() && !m_bReportedEnvCollisionFXThisFrame )
	{
		m_bReportedEnvCollisionFXThisFrame = true;

		// MGH 10/10/2007: We want the actual velocity, not m_xBodyState.m_xLinearVelocity whatever that is.
		Asura_AVector3 xVel = m_xBodyState.m_xCentreOfMass - m_xBodyState.m_xPreviousCentreOfMass;
		xVel /= Asura_Physics_System::GetTimeSlice();

		Asura_Physics_CollisionFX_Report_System::AddReport( xPosition, xVel, xNormal, GetAssociatedEntityGuid(), uFaceID );
	}
	#endif

	if (!GetReportEnvironmentCollisions()) return;

	// KJFL - 8 February - check to see if report with this poly already exists? Not sure
	// what exact behaviour people might want... for now don't allow duplicates. Later
	// if extra info (e.g. exact positions) is required, we may allow multiple reports
	{
		const Asura_Physics_EnvironmentCollisionReport* pxReport = m_pxEnvironmentCollisionReports;

		while (pxReport)
		{
			// quit if we've seen this before
			if (pxReport->GetFaceID() == uFaceID) return;

			pxReport = pxReport->GetNextReport();
		}
	}


	// allocate a report
	Asura_Physics_EnvironmentCollisionReport* pxReport = static_cast<Asura_Physics_EnvironmentCollisionReport*>(g_xEnvironmentCollisionReportsPool.Allocate());

	if (!pxReport)
	{
		Asura_Assert( pxReport, "Code bug: out of memory?" );
		return;
	}

	// set the data held in the report
	pxReport->m_uFaceID = uFaceID;
	pxReport->m_xNormal = xNormal;

	// put the report at the head of the list
	pxReport->m_pxNextReport = m_pxEnvironmentCollisionReports;
	m_pxEnvironmentCollisionReports = pxReport;
}

#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
void Asura_Physics_RigidBodyInstance::AddConstraint( Asura_Physics_Constraint* pxConstraint )
{
	m_xConstraintList.Add( pxConstraint );
}

void Asura_Physics_RigidBodyInstance::RemoveConstraint( Asura_Physics_Constraint* pxConstraint )
{
	m_xConstraintList.Remove( pxConstraint );
}

void Asura_Physics_RigidBodyInstance::ResolveConstraints( void )
{
	Asura_Physics_Constraint_Node* pxNode = m_xConstraintList.GetList();

	// scan through nodes and resolve the constraints
	while ( pxNode )
	{
		Asura_Physics_Constraint* pxConstraint = pxNode->GetConstraint();

		if (pxConstraint)
		{
			// only do the constraints if you're the primary body
			// avoids doing constraints twice etc
			if (pxConstraint->GetPrimaryBody() == this )
			{
				pxConstraint->ResolveConstraint();
			}
		}

		pxNode = pxNode->GetNextNode();
	}
}


void Asura_Physics_RigidBodyInstance::MergeIslandsUsingConstraints( void )
{
	Asura_Physics_Constraint_Node* pxNode = m_xConstraintList.GetList();

	// scan through nodes and resolve the constraints
	while ( pxNode )
	{
		Asura_Physics_Constraint* pxConstraint = pxNode->GetConstraint();

		if (pxConstraint)
		{
			Asura_Physics_RigidBodyInstance* pxPrimary = pxConstraint->GetPrimaryBody();
			Asura_Physics_RigidBodyInstance* pxSecondary = pxConstraint->GetSecondaryBody();

			// if there are two bodies involved, check the islands
			if ( pxPrimary && pxSecondary )
			{
				if ( !(pxPrimary->IsAsleep() && pxSecondary->IsAsleep()) )
				{
					if ( pxPrimary->IsAsleep() ) pxPrimary->WakeUp();
					if ( pxSecondary->IsAsleep() ) pxSecondary->WakeUp();

					Asura_Physics_Island::AssignSameIsland( pxPrimary, pxSecondary );
				}
			}
		}

		pxNode = pxNode->GetNextNode();
	}
}

#endif

static void RenderBoundingBox( const Asura_Bounding_Box& xBB, const u_int uColour = 0xff00ff00 )
{
	Asura_Vector_3 axBoxVertices[8];
	
	// { -1.0f, -1.0f, -1.0f, 1.0f }
	// { -1.0f, -1.0f, +1.0f, 1.0f }
	// { -1.0f, +1.0f, -1.0f, 1.0f }
	// { -1.0f, +1.0f, +1.0f, 1.0f }
	// { +1.0f, -1.0f, -1.0f, 1.0f }
	// { +1.0f, -1.0f, +1.0f, 1.0f }
	// { +1.0f, +1.0f, -1.0f, 1.0f }
	// { +1.0f, +1.0f, +1.0f, 1.0f }

	axBoxVertices[0] = Asura_Vector_3(xBB.MinX, xBB.MinY, xBB.MinZ);
	axBoxVertices[1] = Asura_Vector_3(xBB.MinX, xBB.MinY, xBB.MaxZ);
	axBoxVertices[2] = Asura_Vector_3(xBB.MinX, xBB.MaxY, xBB.MinZ);
	axBoxVertices[3] = Asura_Vector_3(xBB.MinX, xBB.MaxY, xBB.MaxZ);
	axBoxVertices[4] = Asura_Vector_3(xBB.MaxX, xBB.MinY, xBB.MinZ);
	axBoxVertices[5] = Asura_Vector_3(xBB.MaxX, xBB.MinY, xBB.MaxZ);
	axBoxVertices[6] = Asura_Vector_3(xBB.MaxX, xBB.MaxY, xBB.MinZ);
	axBoxVertices[7] = Asura_Vector_3(xBB.MaxX, xBB.MaxY, xBB.MaxZ);
	


	Asura_Render::RenderLine( axBoxVertices[0], axBoxVertices[1], uColour );
	Asura_Render::RenderLine( axBoxVertices[1], axBoxVertices[3], uColour );
	Asura_Render::RenderLine( axBoxVertices[2], axBoxVertices[3], uColour );
	Asura_Render::RenderLine( axBoxVertices[0], axBoxVertices[2], uColour );
	
	Asura_Render::RenderLine( axBoxVertices[0], axBoxVertices[4], uColour );
	Asura_Render::RenderLine( axBoxVertices[4], axBoxVertices[6], uColour );
	Asura_Render::RenderLine( axBoxVertices[2], axBoxVertices[6], uColour );
	Asura_Render::RenderLine( axBoxVertices[6], axBoxVertices[7], uColour );

	Asura_Render::RenderLine( axBoxVertices[4], axBoxVertices[5], uColour );
	Asura_Render::RenderLine( axBoxVertices[5], axBoxVertices[7], uColour );
	Asura_Render::RenderLine( axBoxVertices[1], axBoxVertices[5], uColour );
	Asura_Render::RenderLine( axBoxVertices[3], axBoxVertices[7], uColour );

}

DEBUGVAR bool  dbg_bPhysicsCullDebugRender    = true;
DEBUGVAR float dbg_fPhysicsMaxDebugRenderDist = 50.0f;



void Asura_Physics_RigidBodyInstance::DebuggingRender( void )
{
//	extern void RenderOutline( const Asura_Vector_3* pxPoints, int iNumPoints, u_int uColour, bool bWireframe );

	//SetCollisionTestAABB();
    Asura_Camera::SetCameraIndex( 0 );

    if( dbg_bPhysicsCullDebugRender )
    {
        // Check it's in the frustum.
        if( !Asura_Camera::IsAABBInFrustum(m_xBoundingBox) )
        {
            return;
        }

        // Cull for distance.
        if( dbg_fPhysicsMaxDebugRenderDist > 0.0f )
        {
            const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();
            if( m_xBoundingBox.GetSqdMinDistFromPoint(*pxCamPos) > dbg_fPhysicsMaxDebugRenderDist * dbg_fPhysicsMaxDebugRenderDist )
            {
                return;
            }
        }

        // TODO: Could check visibility here as well.
    }

	const Asura_Physics_RigidBody_ObjectMesh& xMesh = *(m_pxDefinition->m_pxObjectMesh);
	
	
	if( Asura_Physics_System::GetDisplayBroadPhaseBBs() )
	{
		RenderBoundingBox( m_xBoundingBox, 0xffffffff );
	}

	if ( Asura_Physics_System::GetDisplayDebugEnvCaches() )
	{
		RenderBoundingBox( m_xEnvironmentCache.GetCacheBoundingBox(), 0xffff8921 );
	}


#if 0
	if (!Asura_Render::GetWireframe())
	{
		for (u_int u=0; u<xMesh.GetNumberOfFaces(); u++)
		{


			Asura_Vector_3 xPoints[3];

			xPoints[0].x = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[0+u*3]].m_fX;
			xPoints[0].y = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[0+u*3]].m_fY;
			xPoints[0].z = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[0+u*3]].m_fZ;

			xPoints[1].x = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[1+u*3]].m_fX;
			xPoints[1].y = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[1+u*3]].m_fY;
			xPoints[1].z = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[1+u*3]].m_fZ;

			xPoints[2].x = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[2+u*3]].m_fX;
			xPoints[2].y = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[2+u*3]].m_fY;
			xPoints[2].z = m_pxObjectVertexPosition[xMesh.m_pusFaceVertexIndices[2+u*3]].m_fZ;


			Asura_Vector_3 xNormal = (xPoints[1] - xPoints[0]) ^ ( xPoints[2] - xPoints[1]);
			xNormal.Normalise();
			
			u_int uBrightness = Asura_Maths::FloatToInt( 100.0f * xNormal.y * xNormal.y + 127.0f );

			u_int uDiffuse = 0xff000000 + uBrightness + (uBrightness<<8) + (uBrightness<<16);
			
//			RenderOutline( xPoints, 3, uDiffuse, false );
		}
	}
#endif
	if (dbg_bDrawEdges || Asura_Render::GetWireframe())
	{
		u_int uDiffuse = 0xffff0000;

		//if (m_bIsInContactWithFloor) uDiffuse = 0xffffffff;
		
		if (IsIgnore())
		{	
			uDiffuse = 0xff00007f;
		}
		else if (IsStatic())
		{
			uDiffuse = 0xff00ff00;
		}
		else if (IsAsleep())
		{
			uDiffuse = 0xffff00ff;
		}
		else if (IsExternallyControlled())
		{
			uDiffuse = 0xff00ffff;
		}
		else if (IsActor())
		{
			uDiffuse = 0xffffff00;
		}

		//if (GetObjectCollisionsReportList()) uDiffuse = 0xffffffff;

		switch ( m_eCollisionProxy )
		{
			default:
			case COLLISION_PROXY_OBB:
			{
				Asura_Vector_3 axBoxVertices[8];
				
				// { -1.0f, -1.0f, -1.0f, 1.0f }
				// { -1.0f, -1.0f, +1.0f, 1.0f }
				// { -1.0f, +1.0f, -1.0f, 1.0f }
				// { -1.0f, +1.0f, +1.0f, 1.0f }
				// { +1.0f, -1.0f, -1.0f, 1.0f }
				// { +1.0f, -1.0f, +1.0f, 1.0f }
				// { +1.0f, +1.0f, -1.0f, 1.0f }
				// { +1.0f, +1.0f, +1.0f, 1.0f }

				axBoxVertices[0] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MinX, m_xEnvironmentCache.GetCacheBoundingBox().MinY, m_xEnvironmentCache.GetCacheBoundingBox().MinZ);
				axBoxVertices[1] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MinX, m_xEnvironmentCache.GetCacheBoundingBox().MinY, m_xEnvironmentCache.GetCacheBoundingBox().MaxZ);
				axBoxVertices[2] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MinX, m_xEnvironmentCache.GetCacheBoundingBox().MaxY, m_xEnvironmentCache.GetCacheBoundingBox().MinZ);
				axBoxVertices[3] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MinX, m_xEnvironmentCache.GetCacheBoundingBox().MaxY, m_xEnvironmentCache.GetCacheBoundingBox().MaxZ);
				axBoxVertices[4] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MaxX, m_xEnvironmentCache.GetCacheBoundingBox().MinY, m_xEnvironmentCache.GetCacheBoundingBox().MinZ);
				axBoxVertices[5] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MaxX, m_xEnvironmentCache.GetCacheBoundingBox().MinY, m_xEnvironmentCache.GetCacheBoundingBox().MaxZ);
				axBoxVertices[6] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MaxX, m_xEnvironmentCache.GetCacheBoundingBox().MaxY, m_xEnvironmentCache.GetCacheBoundingBox().MinZ);
				axBoxVertices[7] = Asura_Vector_3(m_xEnvironmentCache.GetCacheBoundingBox().MaxX, m_xEnvironmentCache.GetCacheBoundingBox().MaxY, m_xEnvironmentCache.GetCacheBoundingBox().MaxZ);
					
				if (m_pxDefinition)
				{
					const u_int uNumberOfObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
					for (u_int u=0; u<uNumberOfObjectVertices; u++)
					{
						m_pxObjectVertexPosition[u] = m_pxDefinition->m_pxObjectMesh->GetVertexPosition(u) * m_xBodyState.m_xOrientation + m_xInstancePosition;
					}
				}

				for (u_int u=0; u<xMesh.GetNumberOfEdges(); u++)
				{
					const Asura_Physics_RigidBody_ObjectEdge& xEdge = xMesh.GetEdge(u);
					u_int uIndex0 = xEdge.GetFirstVertexIndex();
					u_int uIndex1 = xEdge.GetSecondVertexIndex();

					Asura_AVector3 xV0 = m_pxObjectVertexPosition[uIndex0];
					Asura_AVector3 xV1 = m_pxObjectVertexPosition[uIndex1];

			
			
					Render4VectorLine( xV0, xV1, uDiffuse);
				}
				break;
			}
			case COLLISION_PROXY_SPHERE:
			{
				Asura_Physics_SphereGroup xStack( *this );

				float fRadius = xStack.GetSphereRadius();

				RenderEllipsoid( xStack.GetSphereCentres()[0], fRadius, fRadius, uDiffuse, &(GetOrientation()) );
			
				break;
			}
			case COLLISION_PROXY_SPHERESTACK:
			{
				Asura_Physics_SphereGroup xStack( *this );

				float fRadius = xStack.GetSphereRadius();

				for (u_int u=0; u<xStack.GetNumberOfSpheres(); ++u)
				{
					RenderEllipsoid( xStack.GetSphereCentres()[u], fRadius, fRadius, uDiffuse );
				}
				break;
			}
			case COLLISION_PROXY_POLYGONAL:
			{
				if (GetDefinition()->HasCollisionMesh())
				{
					Asura_Physics_CollisionMesh_Iterator xTriIterator( GetDefinition()->GetCollisionMesh() );

					Asura_Physics_Tri xTri;

					for (xTriIterator.AccessFirstTri(); xTriIterator.GetCurrentTri( xTri ); xTriIterator.AccessNextTri() )
					{
						xTri.Transform( m_xBodyState.m_xOrientation, m_xInstancePosition );

						const Asura_AVector3* pxVertices = xTri.GetVertexArray();

						Render4VectorLine( pxVertices[0],pxVertices[1], uDiffuse );
						Render4VectorLine( pxVertices[1],pxVertices[2], uDiffuse );
						Render4VectorLine( pxVertices[2],pxVertices[0], uDiffuse );
					}

				}
			}
		}
	}

	// render contacts
	for ( const Asura_Physics_ContactCluster* pxCluster = m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
	{
		const Asura_AVector3& xNormal = pxCluster->GetContactNormal();
	
		for ( u_int u=0; u<pxCluster->GetNumberOfContacts(); ++u )
		{
			RenderContactPoint( pxCluster->GetContactPosition(u), xNormal );
		}
	}

#if 0
	const Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvironmentCache.GetFaceList();

	while (pxFace)
	{
		const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		Asura_Vector_3 xPoints[3];

		xPoints[0].x = xV0.m_fX;
		xPoints[0].y = xV0.m_fY;
		xPoints[0].z = xV0.m_fZ;

		xPoints[1].x = xV1.m_fX;
		xPoints[1].y = xV1.m_fY;
		xPoints[1].z = xV1.m_fZ;

		xPoints[2].x = xV2.m_fX;
		xPoints[2].y = xV2.m_fY;
		xPoints[2].z = xV2.m_fZ;


	//	RenderOutline( xPoints, 3, 0xff00ffff, false );

		if ( Asura_Physics_Environment::GetFaceFlags( pxFace->m_uFaceID ) & AsuraPolygonFlag_Physics_WalkableSurface )
		{
		
			Asura_Render::RenderLine(xPoints[0],xPoints[1],0xffffff00);
			Asura_Render::RenderLine(xPoints[1],xPoints[2],0xffffff00);
			Asura_Render::RenderLine(xPoints[2],xPoints[0],0xffffff00);
		}

		pxFace = pxFace->m_pxNext;
	}
#endif


#if 001
	const Asura_Physics_EnvironmentCache_Edge* pxEdge = m_xEnvironmentCache.GetEdgeList();

	while (pxEdge)
	{
		const Asura_AVector3& xEdgeV0 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex0ID );
		const Asura_AVector3& xEdgeV1 = Asura_Physics_Environment::GetVertex( pxEdge->m_uEdgeVertex1ID );

		{
			Asura_Render::RenderLine(xEdgeV0,xEdgeV1,0xffff00ff);
		}

	
		pxEdge = pxEdge->m_pxNext;
	}





#if 0
	for (u_int u=0; u<xMesh.GetNumberOfVertices(); u++)
	{
		Asura_AVector3 xVertex = xMesh.GetVertexPosition(u);

		xVertex = xVertex * m_xBodyState.m_xOrientation + m_xInstancePosition;

		Asura_AVector3 xNormal = xMesh.GetVertexNormal(u)*0.25f;

		xNormal = xNormal * m_xBodyState.m_xOrientation + xVertex;

	

		Render4VectorLine( xVertex, xNormal, 0xff00ff00);
	}
#endif
//	for (u_int u=0; u<xMesh.GetNumberOfFaces(); u++)
//	{
//		Asura_AVector3 xVertex1 = xMesh.G
//		Asura_AVector3 xNormal = xVertex + xMesh.GetVertexNormal(u)*0.25f;
//
//		Render4VectorLine( xVertex, xNormal, 0xff00ff00);
//	}

#endif
}

#if defined(ASURA_USE_STREAMING)
// MGH 11/4/2006: Zone has been unloaded. Clean up any remaining references to it
// in the environment caches.
void Asura_Physics_RigidBodyInstance::RemoveZone(u_int uZoneIndex)
{
	// Clean up dynamic, external and static objects.
	// Clean up dynamic, external and static objects.
	RemoveZoneFromPool( uZoneIndex, s_pxDynamicList );
	RemoveZoneFromPool( uZoneIndex, s_pxExternallyControlledList );
	RemoveZoneFromPool( uZoneIndex, s_pxStaticList );
}

// MGH 11/4/2006: Remove the given zone from the pool of rigid bodies.
void Asura_Physics_RigidBodyInstance::RemoveZoneFromPool( u_int uZoneIndex, Asura_Physics_RigidBodyInstance* pxPoolHead )
{
	for( Asura_Physics_RigidBodyInstance* pxInstance = pxPoolHead; pxInstance; pxInstance = pxInstance->GetNextInstance() )
	{
		pxInstance->RemoveZoneFromInstance(uZoneIndex);
	}
}

// MGH 11/4/2006: Zone has been unloaded. Clean up any remaining references to it
// in the environment cache and contact list.
void Asura_Physics_RigidBodyInstance::RemoveZoneFromInstance(u_int uZoneIndex)
{
	// KJFL - 9 May 2007 - this isn't needed currently, but we may want this method if we 
	// start using persistent contacts again
}

#endif // defined(ASURA_USE_STREAMING)

// These methods are here to avoid an extra cpp just for some one line methods (or an
// unnecessary include in Asura_Physics_CollisionReport).
u_short Asura_Physics_EnvironmentCollisionReport::GetPolyFlags( void ) const
{
	return Asura_Physics_Environment::GetFaceFlags( m_uFaceID );
}

#ifdef ASURA_TOOLS
u_int Asura_Physics_EnvironmentCollisionReport::Tools_GetPolyFlags( void ) const
{
    return Asura_Physics_Environment::Tools_GetFaceFlags( m_uFaceID );
}

#ifdef TOOLS_INCLUDE_PHYSICS_MODULE_AND_POLY_LOOKUP_DATA
void Asura_Physics_EnvironmentCollisionReport::Tools_CalcModuleAndPolyIndex( u_int& uModuleIndex, u_int& uPolyIndex ) const
{
	Asura_Physics_Environment::Tools_CalcModuleAndPolyIndex( m_uFaceID, uModuleIndex, uPolyIndex );
}
#endif
#endif


#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE

struct Asura_Physics_RigidBodyInstance::Backup
{
    Asura_Physics_BodyState xBodyState;
    Asura_AVector3			xOwnGravityDirection;
    Asura_AVector3			xForceEffect;
    Asura_AVector3			xInvInertiaTensorDiagonal;
    Asura_AVector3			xJumpVelocity;
    Asura_AVector3			xInstancePosition;
    u_int                   uAllBitFlags;
	float					fTimeMotionless;
	Asura_AVector3*			pxObjectVertexPosition;
	Asura_AVector3*			pxObjectVertexPrevPosition;
};


void Asura_Physics_RigidBodyInstance::DeallocateBackup( void )
{
    if( m_pxBackup )
    {
        delete[] m_pxBackup->pxObjectVertexPrevPosition;
        delete[] m_pxBackup->pxObjectVertexPosition;
        delete m_pxBackup;
        m_pxBackup = NULL;
    }
}


void Asura_Physics_RigidBodyInstance::StoreToBackup( void )
{
    const u_int uNumObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();

    if( !m_pxBackup )
    {
        m_pxBackup = new Asura_Physics_RigidBodyInstance::Backup;
        m_pxBackup->pxObjectVertexPosition = new Asura_AVector3[uNumObjectVertices];
        m_pxBackup->pxObjectVertexPrevPosition = new Asura_AVector3[uNumObjectVertices];
    }

    m_pxBackup->xBodyState                           = m_xBodyState;
    m_pxBackup->xOwnGravityDirection                 = m_xOwnGravityDirection;
    m_pxBackup->xForceEffect                         = m_xForceEffect;
    m_pxBackup->xInvInertiaTensorDiagonal            = m_xInvInertiaTensorDiagonal;
    m_pxBackup->xJumpVelocity                        = m_xJumpVelocity;
    m_pxBackup->xInstancePosition                    = m_xInstancePosition;
    m_pxBackup->uAllBitFlags                         = m_uAllBitFlags;
    m_pxBackup->fTimeMotionless                      = m_fTimeMotionless;

    for( u_int u = 0; u < uNumObjectVertices; ++u )
    {
        m_pxBackup->pxObjectVertexPosition[u] = m_pxObjectVertexPosition[u];
        m_pxBackup->pxObjectVertexPrevPosition[u] = m_pxObjectVertexPrevPosition[u];
    }
}


void Asura_Physics_RigidBodyInstance::RestoreFromBackup( void )
{
    if( !m_pxBackup )
    {
        return;
    }

    m_xBodyState                           = m_pxBackup->xBodyState;
    m_xOwnGravityDirection                 = m_pxBackup->xOwnGravityDirection;
    m_xForceEffect                         = m_pxBackup->xForceEffect;
    m_xInvInertiaTensorDiagonal            = m_pxBackup->xInvInertiaTensorDiagonal;
    m_xJumpVelocity                        = m_pxBackup->xJumpVelocity;
    m_xInstancePosition                    = m_pxBackup->xInstancePosition;
    m_uAllBitFlags                         = m_pxBackup->uAllBitFlags;
    m_fTimeMotionless                      = m_pxBackup->fTimeMotionless;

    const u_int uNumObjectVertices = m_pxDefinition->m_pxObjectMesh->GetNumberOfVertices();
    for( u_int u = 0; u < uNumObjectVertices; ++u )
    {
        m_pxObjectVertexPosition[u] = m_pxBackup->pxObjectVertexPosition[u];
        m_pxObjectVertexPrevPosition[u] = m_pxBackup->pxObjectVertexPrevPosition[u];
    }

}

#endif // #ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
