//
// Asura_Colour.h
//
// Created by KOR - 23/02/2004 16:21:42
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_COLOUR_H__
#define __ASURA_COLOUR_H__

// unfortunately need this include for FloatToInt...
#include "Core/Asura_Maths.h"

//!
//! KJFL - 2004-03-12 15:11:56 - colour utility class. 
//!
//! The intention is to keep it lean, i.e. no more data members. Hopefully it can be used throughout Asura 
//! whenever a colour representation is need (e.g. particles, lighting, etc).
//!
//! The current four float representation should be most friendly to memory & hardware; traditionally
//! there's been a lot of u_int 32bit colours used in Asura, but this is mainly because of pre-vertex shader
//! PC hardware. Nowadays most GPUs want colours in floats, so having 32bit compacted colours in
//! independent code can often be a false economy.
//!
class Asura_RGBA
{
	public:
		// Constructors
		inline Asura_RGBA() {}
		inline Asura_RGBA(const Asura_RGBA& xOther) { m_fR = xOther.m_fR; m_fG = xOther.m_fG; m_fB = xOther.m_fB; m_fA = xOther.m_fA; }
		inline Asura_RGBA(float fR, float fG, float fB, float fA): m_fR(fR), m_fG(fG), m_fB(fB), m_fA(fA) {}
		inline Asura_RGBA( u_int uRGB, u_int uAlpha ) { FromUIntARGB( uRGB | (uAlpha << 24) ); }
		inline Asura_RGBA(u_int uARGB) { FromUIntARGB(uARGB); }

		inline Asura_RGBA& operator = (const Asura_RGBA& xOther) 
		{ 
			m_fR = xOther.m_fR; m_fG = xOther.m_fG; m_fB = xOther.m_fB; m_fA = xOther.m_fA; 
			return *this;
		}
		inline Asura_RGBA operator + ( const Asura_RGBA& c) const
		{
			return Asura_RGBA(m_fR + c.m_fR, m_fG + c.m_fG, m_fB + c.m_fB, m_fA + c.m_fA);
		}
		inline Asura_RGBA operator - ( const Asura_RGBA& c) const
		{
			return Asura_RGBA(m_fR - c.m_fR, m_fG - c.m_fG, m_fB - c.m_fB, m_fA - c.m_fA);
		}
	    inline Asura_RGBA operator * ( float fScale ) const
		{
			return Asura_RGBA(m_fR * fScale, m_fG * fScale, m_fB * fScale, m_fA * fScale);
		}
        inline Asura_RGBA operator += ( const Asura_RGBA& xOther )
        {
            m_fR += xOther.m_fR; m_fG += xOther.m_fG; m_fB += xOther.m_fB; m_fA += xOther.m_fA;
            return *this;
        }
        inline Asura_RGBA operator -= ( const Asura_RGBA& xOther )
        {
            m_fR -= xOther.m_fR; m_fG -= xOther.m_fG; m_fB -= xOther.m_fB; m_fA -= xOther.m_fA;
            return *this;
        }
        inline Asura_RGBA operator *= ( float fScale )
        {
            m_fR *= fScale; m_fG *= fScale; m_fB *= fScale; m_fA *= fScale;
            return *this;
        }
		inline bool operator == ( const Asura_RGBA& c) const
		{
			return (m_fR == c.m_fR) && (m_fG == c.m_fG) && (m_fB == c.m_fB) && (m_fA == c.m_fA);
		}
		inline bool operator != ( const Asura_RGBA& c) const
		{
			return (m_fR != c.m_fR) || (m_fG != c.m_fG) || (m_fB != c.m_fB) || (m_fA != c.m_fA);
		}

		inline void Zero( void )
		{
			m_fR = m_fG = m_fB = m_fA = 0.0f;
		}

		inline void Set( float fR, float fG, float fB, float fA )
		{
			m_fR = fR, m_fG = fG, m_fB = fB, m_fA = fA;
		}

		inline void Clamp( void )
		{
			m_fR = Asura_Maths::ClampToRange(m_fR, 0.f, 1.f);
			m_fG = Asura_Maths::ClampToRange(m_fG, 0.f, 1.f);
			m_fB = Asura_Maths::ClampToRange(m_fB, 0.f, 1.f);
			m_fA = Asura_Maths::ClampToRange(m_fA, 0.f, 1.f);
		}

        inline float GetLuminance()
        {
            return ( 0.30f * m_fR ) + ( 0.59f * m_fG ) + ( 0.11f * m_fB );
        }

        // Linear interpolation between this colour and xOther, giving xOther
        // when fValue is 1.
        Asura_RGBA Lerp( const Asura_RGBA& xOther, float fValue ) const
        {
            return Asura_RGBA( m_fR * (1.0f - fValue) + xOther.m_fR * fValue,
                               m_fG * (1.0f - fValue) + xOther.m_fG * fValue,
                               m_fB * (1.0f - fValue) + xOther.m_fB * fValue,
                               m_fA * (1.0f - fValue) + xOther.m_fA * fValue );
        }

		// Linear interpolation between two colours A & B, giving A when fValue is 0 and B 
		// when fValue is 1. 
		void Lerp( const Asura_RGBA& xA, const Asura_RGBA& xB, float fValue )
        {
			m_fR = xA.m_fR + (xB.m_fR - xA.m_fR) * fValue;
            m_fG = xA.m_fG + (xB.m_fG - xA.m_fG) * fValue;
            m_fB = xA.m_fB + (xB.m_fB - xA.m_fB) * fValue;
		    m_fA = xA.m_fA + (xB.m_fA - xA.m_fA) * fValue;
        }


		//! KJFL - 2004-03-12 - Make a 32bit representation, with A in the highest byte and B in the lowest.
		//! Note there is no attempt to cache this result or use lazy evaluation, since this class is intended
		//! to be a lean representation of colour. Unless there are memory issues, in independent code it's
		//! generally best to store colours in float form.
		//! KOR - 2004-07-19 - added optional clamp parameter (default = false to retain old behaviour)
		//! If the floating point values are outside the range [0, 1], it will clamp them before conversion
		//! This is slightly slower, but safer if you are unsure that your values are in range
		inline u_int ToUIntARGB(bool bClamp = false) const
		{
			u_int uColour;
			if(bClamp)
			{
				uColour  = Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fA, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fR, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fG, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fB, 0.f, 1.f)*255.0f);
			}
			else
			{
				uColour  = Asura_Maths::FloatToInt(m_fA*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fR*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fG*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fB*255.0f);
			}
			return uColour;
		}

        inline Asura_RGBA& FromUIntARGB( u_int uARGB )
        {
            m_fB = static_cast<int>(uARGB & 0xFF) / 255.0f; // Prevent software u_int->float on PS2.
            uARGB >>=8;
            m_fG = static_cast<int>(uARGB & 0xFF) / 255.0f;
            uARGB >>=8;
            m_fR = static_cast<int>(uARGB & 0xFF) / 255.0f;
            uARGB >>=8;
            m_fA = static_cast<int>(uARGB & 0xFF) / 255.0f;
            return *this;
        }

		inline Asura_RGBA& FromUIntRGBA( u_int uRGBA )
		{
			m_fA = static_cast<int>(uRGBA & 0xFF) / 255.0f; // Prevent software u_int->float on PS2.
            uRGBA >>=8;
            m_fB = static_cast<int>(uRGBA & 0xFF) / 255.0f;
            uRGBA >>=8;
            m_fG = static_cast<int>(uRGBA & 0xFF) / 255.0f;
            uRGBA >>=8;
            m_fR = static_cast<int>(uRGBA & 0xFF) / 255.0f;
            return *this;
		}

		inline u_int ToUIntABGR(bool bClamp = false) const
		{
			u_int uColour;

			if(bClamp)
			{
				uColour  = Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fA, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fB, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fG, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fR, 0.f, 1.f)*255.0f);
			}
			else
			{
				uColour  = Asura_Maths::FloatToInt(m_fA*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fB*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fG*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fR*255.0f);
			}

			return uColour;
		}

		inline u_int ToUIntRGBA(bool bClamp = false) const
		{
			u_int uColour;

			if (bClamp)
			{
				uColour  = Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fR, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fG, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fB, 0.f, 1.f)*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(Asura_Maths::ClampToRange(m_fA, 0.f, 1.f)*255.0f);
			}
			else
			{
				uColour  = Asura_Maths::FloatToInt(m_fR*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fG*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fB*255.0f);	uColour <<=8;
				uColour += Asura_Maths::FloatToInt(m_fA*255.0f);
			}

			return uColour;
		}
        
		// Members
		// MH 6.12.05 -- got rid of union with m_afChannels[4] to avoid the compiler generating extra loads and stores.
		float m_fR;
		float m_fG;
		float m_fB;
		float m_fA;
};

#endif//__ASURA_COLOUR_H__
