
//
// Asura_DeferredShading.cpp
//
// Created by DMR 13/02/07
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Handles deferred lighting passes
//


#include "Core/Asura.h"

#include "Rendering/Asura_DeferredShading.h"

#include "Debugging/Asura_CPUProfile.h"
#include "Entity/DynamicLight/Asura_ClientEntity_DynamicLight.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Render.h"
#include "Projectors/Asura_Projector_System.h"
#include "Shadows/Asura_CSM.h"
#include "Shadows/Asura_Shadow_System.h"

#ifdef ASURA_USE_DEFERRED_SHADING
	#error The project has ASURA_USE_DEFERRED_SHADING defined. This is now obsolete. \
		   See programming announcement from AAM on 6/11/2007.
#endif

#ifdef ASURA_TARGET_SM3

bool Asura_DeferredShading::s_bShowingDeferredBuffers = false;
bool Asura_DeferredShading::s_bShowingPhotonOutlines  = false;
bool Asura_DeferredShading::s_bShowProjectorFrustums  = false;
bool Asura_DeferredShading::s_bUseZBufferOptimization = true;
bool Asura_DeferredShading::s_bCubesForPhotons = false;
bool Asura_DeferredShading::s_bUseStencilBuffer = true;
bool Asura_DeferredShading::s_bShowStencils			= false;

static bool g_bRenderDarkeningShadowsAfterNoShadowLights = true;

void Asura_DeferredShading::Initialise()
{
	Platform_Initialise(); 

	AddConsoleCommands();
}


void Asura_DeferredShading::AddConsoleCommands()
{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
	Asura_CommandConsole::AddVar( L"Render.Deferred.ShowBuffers", &(s_bShowingDeferredBuffers) , Asura_CommandConsole::REMOVE_AT_SHUTDOWN );	
	Asura_CommandConsole::AddVar( L"Render.Deferred.ShowPhotons", &(s_bShowingPhotonOutlines), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.Deferred.UseStencilBuffer", &(s_bUseStencilBuffer), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.Deferred.ShowStencils", &s_bShowStencils, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	Asura_CommandConsole::AddVar( L"Render.Deferred.ShowProjectors", &(s_bShowProjectorFrustums), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );	

	Asura_CommandConsole::AddVar( L"Render.Deferred.CubesForPhotons", &(s_bCubesForPhotons), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );	

		Asura_CommandConsole::AddVar( L"Render.Deferred.RenderDarkeningShadowsAfterNoShadowLights", &(g_bRenderDarkeningShadowsAfterNoShadowLights), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );	
#endif
}


void Asura_DeferredShading::Shutdown()
{
	Platform_Shutdown();
}

void Asura_DeferredShading::Reset()
{
	Platform_Reset();
}


void Asura_DeferredShading::RenderPhotons(bool bWithDarkeningShadows, bool bWithRealShadows, bool bWithNoShadows)
{
	if (Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_PHOTONS))
	{
		Platform_RenderPhotons(bWithDarkeningShadows, bWithRealShadows, bWithNoShadows);
	}		
}


void Asura_DeferredShading::Render()
{
	Platform_BeginRender();
	{
		#ifdef ASURA_USE_CSM
			Asura_CSM::ApplyToScene();
		#endif

		ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__DEFERRED_LIGHTS, "Deferred Lights");

		// First, iterate through the dynamic light entities and update their positions. This has to be done after the entity update, 
		// to avoid lights ending up a frame behind entities they're attached to.
		//
		// (This used to be done in the entity update, but the update order was indeterminate)
		for( Asura_Hashed_List_It<Asura_ClientEntity_DynamicLight> xIt(Asura_ClientEntity_DynamicLight::GetList()); !xIt.Done(); xIt.Next() )
		{
			xIt.GetCurrent()->UpdatePositionAfterEntityUpdate();
		}



		ASURA_CPU_PROFILE__STOP(ASURA_CPU_PROFILE__DEFERRED_LIGHTS);


		ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__CREATE_SHADOWS, "Create Object Shadows");

		// Create the shadow instances
		Asura_Shadow_System::RenderInstances();

		ASURA_CPU_PROFILE__STOP(ASURA_CPU_PROFILE__CREATE_SHADOWS);

		ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__DEFERRED_LIGHTS, "Deferred Lights");

		// Render any lights casting darkening-pass shadows
		RenderPhotons(true, false, false);
		Asura_Projector_System::Render(true, false, g_bRenderDarkeningShadowsAfterNoShadowLights );

		// Render the darkening-pass shadows
		Asura_Shadow_System::RenderOnToScene();

		// Now render any lights with their own shadows
		RenderPhotons(false, true, true);
		Asura_Projector_System::Render(false, true, !g_bRenderDarkeningShadowsAfterNoShadowLights );

		// Dealloc the shadow instances
		Asura_Shadow_System::FinishedRenderingShadows();

		ASURA_CPU_PROFILE__STOP(ASURA_CPU_PROFILE__DEFERRED_LIGHTS);

		// Fog is rendered here now - before translucent stuff
		Asura_Fog::Render();
	}
	Platform_EndRender();
}


void Asura_DeferredShading::RenderDebugOutput()
{
	Platform_RenderDebugOutput();
}




// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Asura_Vector_3 Asura_DeferredShading::GetPointOnSphere
// Gets a point on a radius=1.0f sphere, given 2 angles (in the 0-1 range)
// ------------------------------------------------------------------------------------------------------------------------------------------------------

Asura_Vector_3 Asura_DeferredShading::GetPointOnSphere(float fElev,float fRad)
{
	// Inputs are both in the 0-1 range
	Asura_Vector_3 xPos;

	float fElevationAngle = fElev * Asura_Maths::Pi;
	xPos.y =  Asura_Maths::Cosine(fElevationAngle);
	xPos.x =  Asura_Maths::Sine(fElevationAngle);
	xPos.z =  0.0f;

	// Make sure top/bottom points are clamped, to avoid possibility of holes
	if (xPos.y<=-0.99f)
	{
		xPos.x =  0.0f;
		xPos.y = -1.0f;		
	}
	else
	{	
		if (xPos.y>=0.99f)
		{
			xPos.x = 0.0f;
			xPos.y = 1.0f;		
		}
		else
		{
			// Rotate it
			Asura_Matrix_3x3 xMtx;
			float fAngle = fRad * 2.0f * Asura_Maths::Pi;			
			Asura_Euler_3 xAngles = Asura_Euler_3(0, fAngle, 0);		
			xAngles.ToMatrix(xMtx);
			xPos = xPos * xMtx;			
		}
	}	

	return xPos;
}



// ------------------------------------------------------------------------------------------------------------------------------------------------------
// Asura_DeferredShading::CreateSphere
// This fills a buffer of Asura_Vector_3 with verts for a sphere, as a single tristrip, with the given number of segments
// ------------------------------------------------------------------------------------------------------------------------------------------------------

void Asura_DeferredShading::CalculateSphereData(const int iSphereRadialSegments, const int iSphereVerticalSegments,
												int& iNumVerts)
{
	iNumVerts	= ((iSphereRadialSegments * 4) + 2) * iSphereVerticalSegments;
}
 
void Asura_DeferredShading::CreateSphere(Asura_Vector_3 *pxVerts, const int iNumVerts,
										 const int iSphereRadialSegments, const int iSphereVerticalSegments,
										 const float fSphereRadiusScale)
{
#if defined(ASURA_ASSERT)
	int	iExpectedNumVerts;

	CalculateSphereData(iSphereRadialSegments, iSphereVerticalSegments, iExpectedNumVerts);
	Asura_Assert(pxVerts && iNumVerts >= iExpectedNumVerts, "Vertex buffer null or too small - this shouldn't happen");
#endif	//defined(ASURA_ASSERT)

	int iVertCount= 0;
	for(int j = 0; j< iSphereVerticalSegments; j++)
	{
		float fTop    = ((float)(j))/((float)(iSphereVerticalSegments));
		float fBottom = ((float)(j+1))/((float)(iSphereVerticalSegments));

		for(int i = 0; i< iSphereRadialSegments; i++)
		{
			int i2=i+1;
			if (i2==iSphereRadialSegments) i2=0;

			float fLeft    = ((float)(i))/((float)(iSphereRadialSegments));
			float fRight   = ((float)(i2))/((float)(iSphereRadialSegments));					

			if (i==0)
			{
				// Add extra vert to link strips with a degen tri
				pxVerts[iVertCount++] = GetPointOnSphere(fTop,fLeft) * fSphereRadiusScale;
			}

			pxVerts[iVertCount++] = GetPointOnSphere(fTop,fLeft)	 * fSphereRadiusScale;
			pxVerts[iVertCount++] = GetPointOnSphere(fBottom,fLeft)  * fSphereRadiusScale;
			pxVerts[iVertCount++] = GetPointOnSphere(fTop,fRight)	 * fSphereRadiusScale;
			pxVerts[iVertCount++] = GetPointOnSphere(fBottom,fRight) * fSphereRadiusScale;

			if (i==(iSphereRadialSegments-1))
			{
				// Add extra vert to link strips with a degen tri
				pxVerts[iVertCount++] = GetPointOnSphere(fBottom,fRight) * fSphereRadiusScale;
			}
		}
	}

#if defined(ASURA_ASSERT)
	Asura_Assert(iExpectedNumVerts == iVertCount, "Problem creating sphere VB - vertex count mismatch");
#endif	//defined(ASURA_ASSERT)
}




////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DeferredShading::CreateSphereSegment
// Author       : Andy Tate
// Date         : 8-May-2009
////////////////////////////////////////////////////////////////////////////////
// What we have here is a function that writes out a list of vertices and indices
// to be able to render either a segment of a sphere OR a sphere with the top
// sliced off (if the limit-angle is greater than 180 degrees). The sphere itself
// is split into vertical "slices" with a fixes number of vertices per-slice and
// two extra vertices - one for the bottom of the sphere and the second for either
// the point of the cone (if the limit-angle is less-than 180 degrees), or the top
// of the slice (if the limit-angle is greater-than 180 degrees).
//
// There vertices aren't co-ordinates, merely 3 data values to be interpreted by
// the vertex-shader as follows:
//
// - x = the angle (in radians) of the vertex on the current "slice"
// - y = the parametric scale value applied to the angle-limit (used to calculate
//       the vertical position of the current "slice" of the sphere),
// - z = 0.0f for the top/centre of the sphere, 1.0f in all other cases.
////////////////////////////////////////////////////////////////////////////////

void Asura_DeferredShading::CalculateIndexedSphereData(const int iSphereRadialSegments, const int iSphereVerticalSegments,
													   int& iNumVerts,
													   int& iNumIndices)
{
	iNumVerts	= (iSphereRadialSegments * (iSphereVerticalSegments - 1)) + 2;
	iNumIndices	= iSphereRadialSegments * (((iSphereVerticalSegments - 1) * 2) + 4);
}

void Asura_DeferredShading::CreateIndexedSphere(Asura_Vector_2* pxVerts, int iNumVerts,
												u_short* pusIndices, int iNumIndices,
												int iSphereRadialSegments, int iSphereVerticalSegments)
{
#if defined(ASURA_ASSERT)
	int	iExpectedNumVerts;
	int	iExpectedNumIndices;

	CalculateIndexedSphereData(iSphereRadialSegments, iSphereVerticalSegments, iExpectedNumVerts, iExpectedNumIndices);
	Asura_Assert(pxVerts && (iNumVerts >= iExpectedNumVerts) && (iNumIndices >= iExpectedNumIndices), "Vertex buffer null or too small - this shouldn't happen");
#endif	//defined(ASURA_ASSERT)

	int	iVertCount	= 0;
	int	iIndexCount	= 0;

	// Create the vertex-buffer...
	pxVerts[iVertCount].x	= 0.0f;
	pxVerts[iVertCount++].y	= Asura_Maths::Pi;

	pxVerts[iVertCount].x	= 0.0f;
	pxVerts[iVertCount++].y	= 0.0f;

	// Create vertices by sphere vertical-segment, the per vertex around this segment...
	for(int j = 1; j < iSphereVerticalSegments; j++)
	{
		const float	fVerticalAngle	= (j * Asura_Maths::Pi) / iSphereVerticalSegments;

		for(int i = 0; i < iSphereRadialSegments; i++)
		{
			const float	fRadialAngle	= (i * 2.0f * Asura_Maths::Pi) / iSphereRadialSegments;

			pxVerts[iVertCount].x	= fRadialAngle;
			pxVerts[iVertCount++].y	= fVerticalAngle;
		}
	}

	// Create the index-buffer...
	for(int i = 0; i < iSphereRadialSegments; i++)
	{
		const int	iNext	= (i >= (iSphereRadialSegments - 1)) ? 0 : (i + 1);

		// Create the base point of the triangle...
		pusIndices[iIndexCount++]	 = 1;
		pusIndices[iIndexCount++]	 = 1;

		for(int j = 0; j < (iSphereVerticalSegments - 1); j++)
		{
			// Create the triangles going up the side...
			pusIndices[iIndexCount++]	 = static_cast<u_short>(2 + (j * iSphereRadialSegments) + i);
			pusIndices[iIndexCount++]	= static_cast<u_short>(2 + (j * iSphereRadialSegments) + iNext);
		}

		pusIndices[iIndexCount++]	= 0;
		pusIndices[iIndexCount++]	= 0;
	}

#if defined(ASURA_ASSERT)
	Asura_Assert(iExpectedNumVerts == iVertCount, "Problem creating sphere VB - vertex count mismatch");
	Asura_Assert(iExpectedNumIndices == iIndexCount, "Problem creating sphere IB - index count mismatch");
#endif	//defined(ASURA_ASSERT)
}

/// End of function 'Asura_DeferredShading::CreateSphereSegment'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DeferredShading::CreateSphereSegment
// Author       : Andy Tate
// Date         : 8-May-2009
////////////////////////////////////////////////////////////////////////////////
// What we have here is a function that writes out a list of vertices and indices
// to be able to render either a segment of a sphere OR a sphere with the top
// sliced off (if the limit-angle is greater than 180 degrees). The sphere itself
// is split into vertical "slices" with a fixes number of vertices per-slice and
// two extra vertices - one for the bottom of the sphere and the second for either
// the point of the cone (if the limit-angle is less-than 180 degrees), or the top
// of the slice (if the limit-angle is greater-than 180 degrees).
//
// There vertices aren't co-ordinates, merely 3 data values to be interpreted by
// the vertex-shader as follows:
//
// - x = the angle (in radians) of the vertex on the current "slice"
// - y = the parametric scale value applied to the angle-limit (used to calculate
//       the vertical position of the current "slice" of the sphere),
// - z = 0.0f for the top/centre of the sphere, 1.0f in all other cases.
////////////////////////////////////////////////////////////////////////////////

void Asura_DeferredShading::CalculateIndexedSphereSegmentData(const int iSphereRadialSegments, const int iSphereVerticalSegments,
															  int& iNumVerts,
															  int& iNumIndices)
{
	iNumVerts	= (iSphereRadialSegments * iSphereVerticalSegments) + 2;
	iNumIndices	= iSphereRadialSegments * ((iSphereVerticalSegments * 2) + 4);
}

void Asura_DeferredShading::CreateIndexedSphereSegment(Asura_Vector_3* pxVerts, int iNumVerts,
													   u_short* pusIndices, int iNumIndices,
													   int iSphereRadialSegments,
													   int iSphereVerticalSegments)
{
#if defined(ASURA_ASSERT)
	int	iExpectedNumVerts;
	int	iExpectedNumIndices;

	CalculateIndexedSphereSegmentData(iSphereRadialSegments, iSphereVerticalSegments, iExpectedNumVerts, iExpectedNumIndices);
	Asura_Assert(pxVerts && (iNumVerts >= iExpectedNumVerts) && (iNumIndices >= iExpectedNumIndices), "Vertex buffer null or too small - this shouldn't happen");
#endif	//defined(ASURA_ASSERT)

	int	iVertCount	= 0;
	int	iIndexCount	= 0;

	// Create the vertex-buffer...
	pxVerts[iVertCount].x	= 0.0f;
	pxVerts[iVertCount].y	= 1.0f;
	pxVerts[iVertCount++].z	= 0.0f;

	pxVerts[iVertCount].x	= 0.0f;
	pxVerts[iVertCount].y	= 0.0f;
	pxVerts[iVertCount++].z	= 1.0f;

	// Create vertices by sphere vertical-segment, the per vertex around this segment...
	for(int j = 0; j < iSphereVerticalSegments; j++)
	{
		const float	fScale	= static_cast<float>(j + 1) / iSphereVerticalSegments;

		for(int i = 0; i < iSphereRadialSegments; i++)
		{
			const float	fRadialAngle	= (i * Asura_Maths::Pi * 2.0f) / iSphereRadialSegments;

			pxVerts[iVertCount].x	= fRadialAngle;
			pxVerts[iVertCount].y	= fScale;
			pxVerts[iVertCount++].z	= 1.0f;
		}
	}

	// Create the index-buffer...
	for(int i = 0; i < iSphereRadialSegments; i++)
	{
		const int	iNext	= (i >= (iSphereRadialSegments - 1)) ? 0 : (i + 1);

		// Create the base point of the triangle...
		pusIndices[iIndexCount++]	 = 1;
		pusIndices[iIndexCount++]	 = 1;

		for(int j = 0; j < iSphereVerticalSegments; j++)
		{
			// Create the triangles going up the side...
			pusIndices[iIndexCount++]	 = static_cast<u_short>(2 + (j * iSphereRadialSegments) + i);
			pusIndices[iIndexCount++]	 = static_cast<u_short>(2 + (j * iSphereRadialSegments) + iNext);
		}

		pusIndices[iIndexCount++]	= 0;
		pusIndices[iIndexCount++]	= 0;
	}

#if defined(ASURA_ASSERT)
	Asura_Assert(iExpectedNumVerts == iVertCount, "Problem creating sphere-segment VB - vertex count mismatch");
	Asura_Assert(iExpectedNumIndices == iIndexCount, "Problem creating sphere-segment IB - index count mismatch");
#endif	//defined(ASURA_ASSERT)
}

/// End of function 'Asura_DeferredShading::CreateSphereSegment'. ///


#endif
