//
// Asura_Fog.cpp
//
// Created by KJFL - 2004-03-12 12:50:10
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Fog.h"

// Engine independent.
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Colour.h"



////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

bool				 Asura_Fog::s_bFogIsEnabled = false;
bool				 Asura_Fog::s_bHideFog = false;
Asura_Fog_Parameters Asura_Fog::s_xFogParameters;
Asura_Fog_Parameters Asura_Fog::s_xInitialFogParameters;



void Asura_Fog::Initialise()
{

}

void Asura_Fog::Reset()
{
	s_xInitialFogParameters.DestroyGraphFogGraph();
	s_xFogParameters.DestroyGraphFogGraph();
}

void Asura_Fog::Shutdown()
{
	Reset();
}

//! Turns fog on and off
void Asura_Fog::EnableFog( bool b )
{
	s_bFogIsEnabled = b;

    Platform_UpdateForNewParameters();
}


//! Set fog colour; components should be in the range 0.0f to 1.0f (except if using HDR.)
void Asura_Fog::SetFogColour( float fRed, float fGreen, float fBlue )
{
	s_xFogParameters.SetFogColour( fRed, fGreen, fBlue );
	
	Platform_UpdateForNewParameters();
}

//! Set fog near plane, in metres
void Asura_Fog::SetFogNearPlane( float fNearPlane )
{	
	s_xFogParameters.SetFogNearPlane( fNearPlane );

	Platform_UpdateForNewParameters();
}

//! Set fog far plane, in metres
void Asura_Fog::SetFogFarPlane( float fFarPlane )
{	
	s_xFogParameters.SetFogFarPlane( fFarPlane );

	Platform_UpdateForNewParameters();
}

void Asura_Fog::SetFogValueAtFarPlane( float fFogValueAtFarPlane )
{
	s_xFogParameters.SetFogValueAtFarPlane( fFogValueAtFarPlane );

    Platform_UpdateForNewParameters();
}

void Asura_Fog::SetSkyboxFogBlendFactor( float fBlendFactor )
{
	s_xFogParameters.SetSkyboxFogBlendFactor( fBlendFactor );

	Platform_UpdateForNewParameters();
}

// -------- Setting initial values. --------

void Asura_Fog::SetInitialFogColour( const Asura_RGBA& xColour )
{	
	s_xInitialFogParameters.SetFogColour( xColour );

	SetFogColour( xColour );
}

void Asura_Fog::SetInitialFogColour( float fRed, float fGreen, float fBlue)
{	
	s_xInitialFogParameters.SetFogColour( fRed, fGreen, fBlue);

	SetFogColour( fRed, fGreen, fBlue);
}

void Asura_Fog::SetInitialFogNearPlane(float fNearPlane)
{
	s_xInitialFogParameters.SetFogNearPlane( fNearPlane );

	SetFogNearPlane( fNearPlane );
}

void Asura_Fog::SetInitialFogFarPlane( float fFarPlane )
{	
	s_xInitialFogParameters.SetFogFarPlane( fFarPlane );

	SetFogFarPlane( fFarPlane );
}

void Asura_Fog::SetInitialFogValueAtFarPlane( float fFogValueAtFarPlane )
{	
	s_xInitialFogParameters.SetFogValueAtFarPlane( fFogValueAtFarPlane );
	
	SetFogValueAtFarPlane( fFogValueAtFarPlane );
}

void Asura_Fog::SetInitialSkyboxFogBlendFactor( float fBlendFactor )
{	
	s_xInitialFogParameters.SetSkyboxFogBlendFactor( fBlendFactor );

	SetSkyboxFogBlendFactor( fBlendFactor );
}

void Asura_Fog::SetInitialGraphFogPoint( const Asura_Vector_2 &xGraphPoint )
{
	s_xInitialFogParameters.SetGraphFogPoint( xGraphPoint );

	SetGraphFogPoint( xGraphPoint );
}


void Asura_Fog::BlendParameters( Asura_Fog_Parameters& xFogParametersToBlend, float fStrength )
{ 
	if( fStrength==0.0f )
	{
		//there is no need to blend anything.
		return;
	}
	if( fStrength==1.0f )
	{
		// We don't need to call interpolation function, when fStrength is at maximum. It is enought to simply overwrite the values.
		s_xFogParameters = xFogParametersToBlend;
	}
	else
	{
		s_xFogParameters.BlendParameters( xFogParametersToBlend, fStrength );	
	}

	Platform_UpdateForNewParameters();
}

void Asura_Fog::EnableGraphFog( bool bEnable)
{
	s_xFogParameters.EnableGraphFog( bEnable );
	
	Platform_UpdateForNewParameters();
}

void Asura_Fog::SetGraphFogPoint( const Asura_Vector_2 &xGraphPoint )
{
	s_xFogParameters.SetGraphFogPoint( xGraphPoint );

	Platform_UpdateForNewParameters();
}

float Asura_Fog::GetFogDensityAtDistance(float fDistance)
{
	if( !s_bFogIsEnabled )	{	return 0.0f;	}

	return s_xFogParameters.GetFogDensityAtDistance( fDistance );
}

/*
void Asura_Fog::CalculateLerpConstant( void )
{
s_xInitialFogParameters.CalculateLerpConstant();
s_xFogParameters.CalculateLerpConstant();

Platform_UpdateForNewParameters();
}
*/

   
#ifdef ASURA_TOOLS
void Asura_Fog::SetToDefaultsIfNecessary( void )
{
    // If it's been setup, the value at the far plane should be non-zero.
    // (Zero is invalid since that's equivalent to just turning the fog off.)
    if( 0.0f != GetFogValueAtFarPlane() ) { return; }

    // For a while we weren't setting up defaults correctly. Let's fix cases
    // where the fog has already been setup but the incorrect value-at-far-
    // plane default hasn't been overridden.
    if( s_bFogIsEnabled && 0.0f == GetFogValueAtFarPlane() )
    {
        SetInitialFogValueAtFarPlane( 1.0f );

        // Don't want to fiddle with any other settings since they should be correct.
        return;
    }

    SetInitialFogColour(1.0f,1.0f,1.0f);
    SetInitialFogNearPlane( 25.0f );
    SetInitialFogFarPlane( 250.0f );
    SetInitialFogValueAtFarPlane( 1.0f );
    SetInitialSkyboxFogBlendFactor( 1.0f );
	EnableGraphFog(false);
	SetGraphFogPoint(Asura_Vector_2(0.5f, 0.5f));
}
#endif





//---------------------------------------------------------------------------- new param class


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Version	Author	Comment
// ~~~~~~~	~~~~~~	~~~~~~~
// 1		EP		Created.
////////////////////////////////////////////////////////////////////////////////
const u_int	Asura_Fog_Parameters::s_uChunkVersion	= 1;



Asura_Fog_Parameters& Asura_Fog_Parameters::operator = (const Asura_Fog_Parameters& xFog)
{
	if( this!=&xFog )
	{
		m_xFogColour					= xFog.m_xFogColour;
		m_uFogColourRGB					= xFog.m_uFogColourRGB;
		m_fFogNearPlane					= xFog.m_fFogNearPlane;
		m_fFogFarPlane					= xFog.m_fFogFarPlane;
		m_fFogValueAtFarPlane			= xFog.m_fFogValueAtFarPlane;
		m_fOneOverFogPlaneSeparation	= xFog.m_fOneOverFogPlaneSeparation;
		m_fSkyboxFogBlendFactor			= xFog.m_fSkyboxFogBlendFactor;

		// Extra bits for graph fog.
		m_bGraphFogEnabled				= xFog.m_bGraphFogEnabled;
		m_xGraphFogExtraPoint			= xFog.m_xGraphFogExtraPoint;

		if( m_pxGraphFogGraph )
		{
			delete m_pxGraphFogGraph;
			m_pxGraphFogGraph = NULL;
		}

		// is there something to copy?
		if( xFog.m_pxGraphFogGraph )
		{
			// Create new instance of the GraphFogGraph
			m_pxGraphFogGraph = new Asura_Graph<float>(2, 0.0f, GetFogValueAtFarPlane() );

			// Invoke operator =
			*m_pxGraphFogGraph			= *xFog.m_pxGraphFogGraph;
		}

	}

	return *this;
}

float Asura_Fog_Parameters::GetFogDensityAtDistance(float fDistance)
{
	if( fDistance<m_fFogNearPlane )		{	return 0.0f;	}
	if( fDistance>m_fFogFarPlane )		{	return m_fFogValueAtFarPlane;	}

	if(m_bGraphFogEnabled )
	{
		if( !m_pxGraphFogGraph )		{	return 0.0f;	}

		return m_pxGraphFogGraph->GetValueAtTime((fDistance - m_fFogNearPlane)*m_fOneOverFogPlaneSeparation);
	}
	else
	{
		return ((fDistance - m_fFogNearPlane)*m_fOneOverFogPlaneSeparation)*m_fFogValueAtFarPlane;
	}
}


void Asura_Fog_Parameters::SetFogColour( const Asura_RGBA& xColour )
{
	SetFogColour( xColour.m_fR, xColour.m_fG, xColour.m_fB );
}

//! Set fog colour; components should be in the range 0.0f to 1.0f (except if using HDR.)
void Asura_Fog_Parameters::SetFogColour( float fRed, float fGreen, float fBlue )
{
#ifndef ASURA_USE_HDR
	Asura_Assert( fRed>=0 && fRed<= 1.0f, "Invalid colour value" );
	Asura_Assert( fGreen>=0 && fGreen<= 1.0f, "Invalid colour value" );
	Asura_Assert( fBlue>=0 && fBlue<= 1.0f, "Invalid colour value" );
#endif

	m_xFogColour.Set( fRed, fGreen, fBlue, 0.0f );

	m_uFogColourRGB = m_xFogColour.ToUIntARGB( true );
}

//! Set fog near plane, in meters
void Asura_Fog_Parameters::SetFogNearPlane( float fNearPlane )
{
	m_fFogNearPlane = fNearPlane;
	CalculateLerpConstant();
}

//! Set fog far plane, in meteres
void Asura_Fog_Parameters::SetFogFarPlane( float fFarPlane )
{
	m_fFogFarPlane = fFarPlane;
	CalculateLerpConstant();
}

void Asura_Fog_Parameters::SetFogValueAtFarPlane( float fFogValueAtFarPlane )
{
	m_fFogValueAtFarPlane = fFogValueAtFarPlane;
}

void Asura_Fog_Parameters::SetSkyboxFogBlendFactor( float fBlendFactor )
{
	if (fBlendFactor>1.0f)
	{
		Asura_Assert( fBlendFactor<=1.0f, "Invalid blend factor" );
		fBlendFactor = 1.0f;
	}
	else if (fBlendFactor<0.0f)
	{
		Asura_Assert( fBlendFactor>=1.0f, "Invalid blend factor" );
		fBlendFactor = 0.0f;
	}
	m_fSkyboxFogBlendFactor = fBlendFactor;
}

void Asura_Fog_Parameters::CalculateLerpConstant( void )
{
	float fDistance = m_fFogFarPlane - m_fFogNearPlane;

	// this allows for far plane being nearer the near plane (which is actually possible on PC/Xbox hardware)

	if (fDistance != 0.0f)
	{
		m_fOneOverFogPlaneSeparation = 1.0f / fDistance;
	}
	else // we could return a very large float, but that would just make it foggy everywhere, rather than after the near plane
	{
		m_fOneOverFogPlaneSeparation = 1.0f;
	}
}


void Asura_Fog_Parameters::EnableGraphFog( bool bEnable)
{
	m_bGraphFogEnabled = bEnable;
	if(m_bGraphFogEnabled != (m_pxGraphFogGraph != NULL))
	{
		DestroyGraphFogGraph();
		
		if(m_bGraphFogEnabled)
		{
			m_pxGraphFogGraph = new Asura_Graph<float>(2, 0.0f, GetFogValueAtFarPlane() );
			m_pxGraphFogGraph->AddPoint(m_xGraphFogExtraPoint.y, m_xGraphFogExtraPoint.x);
		}
	}
}

void Asura_Fog_Parameters::SetGraphFogPoint( const Asura_Vector_2 &xGraphPoint )
{
	m_xGraphFogExtraPoint = xGraphPoint;
	if(m_pxGraphFogGraph)
	{
		delete m_pxGraphFogGraph;
		m_pxGraphFogGraph = new Asura_Graph<float>(2, 0.0f, GetFogValueAtFarPlane() );
		m_pxGraphFogGraph->AddPoint(m_xGraphFogExtraPoint.y, m_xGraphFogExtraPoint.x);
	}
}

void Asura_Fog_Parameters::DestroyGraphFogGraph()
{
	if( m_pxGraphFogGraph )
	{
		delete m_pxGraphFogGraph;
		m_pxGraphFogGraph = NULL;
	}
}

Asura_Fog_Parameters::Asura_Fog_Parameters() :
  m_xFogColour( 1.0f, 1.0f, 1.0f, 0.0f )
, m_fFogNearPlane( 25.0f)
, m_fFogFarPlane( 250.0f )
, m_fFogValueAtFarPlane( 1.0f )
, m_fSkyboxFogBlendFactor( 1.0f )
, m_bGraphFogEnabled( false )
, m_xGraphFogExtraPoint( 0.5f, 0.5f )
, m_pxGraphFogGraph( NULL )
{
	m_uFogColourRGB = m_xFogColour.ToUIntARGB( true );
	CalculateLerpConstant();
}

Asura_Fog_Parameters::~Asura_Fog_Parameters()
{
	DestroyGraphFogGraph();
}

void Asura_Fog_Parameters::BlendParameters( Asura_Fog_Parameters& xFogParametersToBlend, float fStrength )
{
	m_xFogColour.Lerp( m_xFogColour, xFogParametersToBlend.m_xFogColour, fStrength );
	m_uFogColourRGB = m_xFogColour.ToUIntARGB( true );

	m_fFogNearPlane =				Asura_Maths::Lerp( m_fFogNearPlane,				xFogParametersToBlend.m_fFogNearPlane,				fStrength );
	m_fFogFarPlane =				Asura_Maths::Lerp( m_fFogFarPlane,				xFogParametersToBlend.m_fFogFarPlane,				fStrength );
	m_fFogValueAtFarPlane =			Asura_Maths::Lerp( m_fFogValueAtFarPlane,		xFogParametersToBlend.m_fFogValueAtFarPlane,		fStrength );
	m_fOneOverFogPlaneSeparation =	Asura_Maths::Lerp( m_fOneOverFogPlaneSeparation,xFogParametersToBlend.m_fOneOverFogPlaneSeparation, fStrength );
	m_fSkyboxFogBlendFactor =		Asura_Maths::Lerp( m_fSkyboxFogBlendFactor,		xFogParametersToBlend.m_fSkyboxFogBlendFactor,		fStrength );

	CalculateLerpConstant();
}


bool Asura_Fog_Parameters::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	//check our version
	u_int uVersion;
	xStream >> uVersion;
	if( uVersion>s_uChunkVersion )
	{
		Asura_Assertf( false,("Unknown chunk version %d for Asura_Fog_Parameters",uVersion));
		return false;
	}

	xStream >> m_xFogColour;
	xStream >> m_fFogNearPlane;
	xStream >> m_fFogFarPlane;
	xStream >> m_fFogValueAtFarPlane;
	xStream >> m_fSkyboxFogBlendFactor;

	// Recalculate this value again.
	m_uFogColourRGB = m_xFogColour.ToUIntARGB( true );

	xStream >> m_fOneOverFogPlaneSeparation;
	xStream >> m_uFogColourRGB;

	// Extra bits for graph fog.
	xStream >> m_bGraphFogEnabled;
	xStream >> m_xGraphFogExtraPoint;

	// Check if we need to stream graph fog.
	bool bReadGraphFog;
	xStream >> bReadGraphFog;

	if( bReadGraphFog )
	{
		DestroyGraphFogGraph();
		m_pxGraphFogGraph = new Asura_Graph< float >( xStream );
	}
	else
	{
		m_pxGraphFogGraph = NULL;
	}

	return true;
}

void Asura_Fog_Parameters::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	//save our version
	xStream << s_uChunkVersion;

	xStream << m_xFogColour;
	xStream << m_fFogNearPlane;
	xStream << m_fFogFarPlane;
	xStream << m_fFogValueAtFarPlane;
	xStream << m_fSkyboxFogBlendFactor;

	// m_uFogColourRGB Do not store this value. Instead refresh this value when we will read the data back from disk. 

	xStream << m_fOneOverFogPlaneSeparation;
	xStream << m_uFogColourRGB;

	// Extra bits for graph fog.
	xStream << m_bGraphFogEnabled;
	xStream << m_xGraphFogExtraPoint;

	bool bReadGraphFog = (m_bGraphFogEnabled && m_pxGraphFogGraph);	// Only store graph when it is enabled but also when it actualy exists.
	xStream << bReadGraphFog;

	if( bReadGraphFog )
	{
		bool bSamplesOnly = true;
#ifdef ASURA_TOOLS
		bSamplesOnly = Tools_Level_Export::IsExport();
#endif
		m_pxGraphFogGraph->WriteToChunkStream( xStream, bSamplesOnly );
	}
}

void Asura_Fog::SetFog( const Asura_Fog_Parameters& xFogParameters )
{
	s_xFogParameters = xFogParameters;

	Platform_UpdateForNewParameters();
}

void Asura_Fog::SetInitialFog( const Asura_Fog_Parameters& xFogParameters )
{
	s_xInitialFogParameters = xFogParameters;
}
