/**
*
* \file Asura_MovieToTexture_System.cpp
*
* Created by Kevin Lea - 19/08/2002 16:26:38
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "Rendering/Asura_MovieToTexture_System.h"

#include "Core/Asura_TextureManagement.h"
#include "Rendering/Asura_Platform_MovieToTextureStream.h"

u_int									Asura_MovieToTexture_System::s_uMaxNumberOfMovieStreams;
Asura_Tracking_Handle_List*				Asura_MovieToTexture_System::s_pxHandleList;
Asura_Platform_MovieToTextureStream**	Asura_MovieToTexture_System::s_ppxStreamStorage;

void Asura_MovieToTexture_System::Initialise(const u_int uMaxNumberOfMovieStreams, const bool bStartAllFMVTextures)
{
	Asura_Assert(!s_pxHandleList, "Asura_MovieToTexture_System already initialised");
	if (s_pxHandleList) return;

	s_uMaxNumberOfMovieStreams = uMaxNumberOfMovieStreams;

	s_pxHandleList = new Asura_Tracking_Handle_List(s_uMaxNumberOfMovieStreams);

	s_ppxStreamStorage = new Asura_Platform_MovieToTextureStream*[s_uMaxNumberOfMovieStreams];
    for( u_int u = 0; u < s_uMaxNumberOfMovieStreams; ++u )
    {
        s_ppxStreamStorage[u] = NULL;
    }

    Platform_Initialise();

	if (bStartAllFMVTextures)
	{
		/// Kevin Lea - 26/07/2002 15:40:02 - scan through texture flags and find movie textures
		for (int i=0; i<ASURA_MAX_NO_OF_TEXTURES; i++)
		{
			if(Asura_Texture_Management::IsTextureLoaded(i) && Asura_Texture_Management::IsTextureFMV(i))
			{
				Asura_String xMovieName("FMV");
				Asura_String xPath;
				Asura_String xFile;
				Asura_Texture_Management::GetTextureName(i)->SplitFilename(&xPath, &xFile, NULL);

				// (MH - 25.05.2004) if movie texture is in graphics/FMV we're already looking for it in the FMV folder.
				xPath.MakeLowercase();
				if (!xPath.EqualsIgnoringExtension("\\fmv"))
				{
					xMovieName += xPath;
				}
				xMovieName += "\\";
				xMovieName += xFile;
				xMovieName += ".bik";

				CreateStream(i, xMovieName.GetString());
			}
		}
	}
}

void Asura_MovieToTexture_System::Reset(void)
{
	Asura_Assert(s_pxHandleList, "Asura_MovieToTexture_System not initialised");
	if (!s_pxHandleList) return;

	s_pxHandleList->Traverse(DestroyStream);
    Platform_Reset();
}

void Asura_MovieToTexture_System::Shutdown(void)
{
	Asura_Assert(s_pxHandleList, "Asura_MovieToTexture_System not initialised");
	if (!s_pxHandleList) return;

	s_pxHandleList->Traverse(DestroyStream);

    Platform_Shutdown();

	delete s_pxHandleList;
	s_pxHandleList = NULL;
	delete [] s_ppxStreamStorage;
	s_ppxStreamStorage = NULL;
}

void Asura_MovieToTexture_System::Update(void)
{
	Platform_Update();
}

Asura_Handle Asura_MovieToTexture_System::CreateStream(const int iTextureIndex, const char* pxMovieName, bool bLooping, bool bHoldAtEnd)
{
	Asura_Assert(s_pxHandleList, "Asura_MovieToTexture_System not initialised");
	if (!s_pxHandleList) return ASURA_INVALID_HANDLE;

	Asura_Handle iHandle;

	if(s_pxHandleList->GetNewHandle(iHandle))
	{
        bool bOK = false;
		Asura_Platform_MovieToTextureStream* pxNewStream = new Asura_Platform_MovieToTextureStream( iTextureIndex, pxMovieName, &bOK, bLooping, bHoldAtEnd );
        
        if( pxNewStream && bOK )
        {
            s_ppxStreamStorage[iHandle] = pxNewStream;
        }
        else
        {
            if( pxNewStream )
            {
                delete pxNewStream;
            }

            s_pxHandleList->ReleaseHandle( iHandle );
            iHandle = ASURA_INVALID_HANDLE;
        }
	}

	return iHandle;
}


////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_MovieToTexture_System
//  NAME     :  IsReadyToBeDestroyed()
//  CREATED  :  Adam A. Brown (17.2.2005)
//  DESC     :  Returns true if the movie stream should be destroyed.
////////////////////////////////////////////////////////////////////////////////
bool Asura_MovieToTexture_System::IsReadyToBeDestroyed(Asura_Handle iMovieHandle)
{
	Asura_Assert(s_pxHandleList, "Asura_MovieToTexture_System not initialised");
	if (!s_pxHandleList) return false;

	/// Kevin Lea - 26/07/2002 12:44:56 - check handle to see if it's okay
	if(iMovieHandle == ASURA_INVALID_HANDLE || s_ppxStreamStorage[iMovieHandle] == NULL)
	{
		Asura_Assert(false, "Invalid handle");
		return false;
	}

	return s_ppxStreamStorage[iMovieHandle]->IsReadyToBeDestroyed();
}


void Asura_MovieToTexture_System::DestroyStream(Asura_Handle iMovieHandle)
{
	Asura_Assert(s_pxHandleList, "Asura_MovieToTexture_System not initialised");
	if (!s_pxHandleList) return;

	/// Kevin Lea - 26/07/2002 12:44:56 - check handle to see if it's okay
	Asura_Assert(iMovieHandle!=ASURA_INVALID_HANDLE,"Invalid handle");
	Asura_Assert(s_ppxStreamStorage[iMovieHandle]!=NULL,"Invalid handle");

	/// Kevin Lea - 26/07/2002 12:45:23 - delete the movie and release handle
	delete s_ppxStreamStorage[iMovieHandle];
	s_ppxStreamStorage[iMovieHandle] = NULL;
	s_pxHandleList->ReleaseHandle(iMovieHandle);
}

Asura_Platform_MovieToTextureStream* Asura_MovieToTexture_System::GetStreamByTexID( int iTextureID )
{
	Asura_Assert(s_pxHandleList, "Asura_MovieToTexture_System not initialised");
	if (!s_pxHandleList) return NULL;

	Asura_Tracking_Handle_List::Iterator xIterator = s_pxHandleList->GetIterator();
    for(xIterator.Begin(); !xIterator.End(); xIterator.Next())
	{
        Asura_Platform_MovieToTextureStream* pxNextStream = s_ppxStreamStorage[xIterator.GetCurrentHandle()];
        if( pxNextStream->GetTextureIndex() == iTextureID )
        {
            return pxNextStream;
        }
    }

    return NULL;
}
