//
// Asura_Water.h
//
// Created by Rob Dibley - 16/11/2004 16:06
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifndef __ASURA_WATER_H__
#define __ASURA_WATER_H__

#include "Core/Asura_Maths.h"

#ifdef ASURA_TOOLS
#include "Core/Asura_String.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#endif

// defines for flags
#define AsuraWaterFlag_Transparent		0x1
#define AsuraWaterFlag_HorizonReflect	0x2
#define AsuraWaterFlag_SunReflect		0x4
#define AsuraWaterFlag_BlendHorizon		0x8	// normally both reflections are ADDED, this flag makes horizon be blended

#define AsuraWaterFlag_Reflective		( AsuraWaterFlag_HorizonReflect | AsuraWaterFlag_SunReflect )

// RJF - 21 February 2005 - Some of this code is really horribly structured... todo for the next refactoring, see about fixing it a bit (e.g. changing
//							changing Asura_Water_Packet so that each platform derives from it instead of it containing a platform structure about which
//							we have no guarantees for interface but make calls on anyway...)
//							We might also want to look into per-packet water controls so that we can have different regions of water with different
//							characteristics (e.g. large ocean + small bay + water tank in lab).

class Asura_Platform_Water_Packet;
template<class T> class Asura_TrackingHandlePool;
class Asura_Water_Packet;

class Asura_Water_System
{
	public:
		struct Water_Controls
		{
			float m_fWaterCycleScale1;
			float m_fWaterCycleScale2;
			float m_fWaterCycleScale3;	// in case we decide we want a third wave

			float m_fXScale1;
			float m_fXScale2;
			float m_fXScale3;

			float m_fZScale1;
			float m_fZScale2;
			float m_fZScale3;

			float m_fYScale1;
			float m_fYScale2;
			float m_fYScale3;

			float m_fFalloff1;
			float m_fFalloff2;
			float m_fFalloff3;

			float m_fThreshold;

			float m_fLightOffset;
			float m_fLightScale;

			float m_fAlphaLevel;
			float m_fSunBrightness;

			bool  m_bUseLevelSun;
			float m_afSunDir[3];

			const char* m_aszTextureNames[3];	// one for each permitted pass
			int   m_aiTextureID[3];		// used to store IDs, but may be used to set them if no names used

			float m_afTexOffset[2];		// Texture offset for the sun reflection texture
			float m_afTexScale[2];		// Texture scale for the sun reflection texture

			union
			{
				struct
				{
					bool m_bTransparency		: 1;
					bool m_bHorizonReflection	: 1;
					bool m_bSunReflection		: 1;
					bool m_bBlendHorizon		: 1;
				};
				u_int m_uFlags;
			};

			float m_fHorizonVScale;		// controls V texture co-ordinate scale for horizon reflection
			float m_fHorizonVOffset;	// ditto for offset
			float m_fHorizonYScale;		// controls horizon effect normal scaling for flatter waves
			float m_fHorizonBrightness;	// controls brightness or contribution of horizon reflection

			float m_fSubdivisionYScale;
			float m_fSunlightFalloffNear;
			float m_fSunlightFalloffFar;

            float m_afUVSpeeds[2];

			float m_fSunYScale;
			float m_fHorizonFresnelScale;

			float m_fLightRescale;		// compensation factor to prevent large waves or large lighting values from over-saturating the output
		};

		//The global water control for the level.. at the moment , this only has extra
		//data while in tools. RAR 10/01/2005
		struct Env_Water_Controls : Water_Controls
		{
			#ifdef ASURA_TOOLS
			Asura_String m_axToolsTextureNames[3];
			Asura_String m_xTemplateName;
			#endif
		};

		// this structure is to be used to simplify sharing the areas of water between platforms
		// the actual mesh generation code will have to be platform specific.
		struct Water_Quad
		{
			Water_Controls* pxControl;
			float fXPos,  fZPos;
			float fXSize, fZSize;
			float fHeight;
			float fRed, fGreen, fBlue;
			float fURepeats, fVRepeats;
		};

		// platform independent code
		static void Initialise();//initialise using the env water controller that will be setup from load
		static void Initialise( Water_Controls* pxControls );
		static void Shutdown();
		static void Reset();
		static void Update( float fFrameTime );
		static void Render();

		// platform specific code
		static void Platform_Initialise(void);
		static void Platform_BeginRender(void);
		static void Platform_Render(Asura_Platform_Water_Packet *pPacket);
		static void Platform_EndRender(void);
        static void Platform_Reset( void );
        static void Platform_Shutdown(void);

		static Asura_Water_Packet* AllocateWaterPacket();

#ifdef ASURA_TOOLS
		static void BuildWaterPackets(Asura_SaveBlock& xSaveBlock, bool bShouldSubdivide);
#endif

		static float GetWaterCycleTime1( void );
		static float GetWaterCycleTime2( void );
		static float GetWaterCycleTime3( void );

		static void	ComputeFracSpace(Asura_Vector_3 *pOutput, const struct Water_Controls *pControls, const Asura_Vector_3 *pInput)
		{
			Asura_Vector_3 xScales(pControls->m_fXScale1,	pControls->m_fXScale2,	pControls->m_fXScale3);
			Asura_Vector_3 zScales(pControls->m_fZScale1,	pControls->m_fZScale2,	pControls->m_fZScale3);
			Asura_Vector_3 tScales(s_fWaterCycleTime1,		s_fWaterCycleTime2,		s_fWaterCycleTime3);	

			*pOutput = xScales * pInput->x;
			*pOutput += zScales * pInput->z;
			*pOutput += tScales;

			pOutput->x -= Asura_Maths::FloatToInt_Chop( pOutput->x );
			pOutput->y -= Asura_Maths::FloatToInt_Chop( pOutput->y );
			pOutput->z -= Asura_Maths::FloatToInt_Chop( pOutput->z );
		}

		static void	ComputePolynomial(Asura_Vector_3 *pOutput, const Asura_Vector_3 *pInput)
		{
			Asura_Vector_3 square = pInput->Vector_Multiply(*pInput);
			Asura_Vector_3 result = *pInput * 0.5f;
			result -= square * 1.5f;
			result += pInput->Vector_Multiply(square);

			*pOutput = result;
		}

		static float	ComputePolynomial(float fInput)
		{
			return (fInput * fInput * fInput) - 1.5f * (fInput * fInput) + 0.5f * (fInput);
		}

		static float	ComputeDerivedPolynomial(float fInput)
		{
			return 3 * (fInput * fInput) - 3.0f * (fInput) + 0.5f;
		}

		static float	CalculateHeightOffset(const struct Water_Controls *pControls, const Asura_Vector_3 *pInput)
		{
			Asura_Vector_3 fHeights;
			Asura_Vector_3 yScales(pControls->m_fYScale1,	pControls->m_fYScale2,	pControls->m_fYScale3);

			ComputeFracSpace(&fHeights, pControls, pInput);
			ComputePolynomial(&fHeights, &fHeights);
			fHeights = fHeights.Vector_Multiply(yScales);

			return(fHeights.x + fHeights.y + fHeights.z);
		}

        static const float* GetBasePassUVOffset( void )
        {
            return s_afBasePassUVOffset;
        }
// (MP - 25.01.2005) Currently only exposed through the U/V speed stuff in
// the water controls.
//		static void SetBasePassUVOffset( const float* pfData )
//		{
//			s_afBasePassUVOffset[0] = *pfData;
//			s_afBasePassUVOffset[1] = *(pfData+1);
//		}

        // One to use for rendering, etc.
        static const Water_Controls* GetCurController( void ) { return s_pxController; }

        // For storing the environment settings.
        // ... Currently this will pretty much always be the same as the
        // above ends up pointing to.
        static Env_Water_Controls& GetEnvController( void ) { return s_xEnvController; }

		static void SetDefaults(Env_Water_Controls* pxControls){Project_SetDefaults(pxControls);}

		static bool IsAnyWaterPresent() { return s_uNumberOfWaterPackets > 0; }

		static bool IsAnyWaterVisible();

		static float GetVisibleWaterYPosition();
	private:

        static void Project_SetDefaults( Env_Water_Controls* pxControls );
		static void RenderPacket( Asura_Water_Packet* pxPacket );

		static Water_Controls* s_pxController;

		static Env_Water_Controls s_xEnvController;
		static float             s_afBasePassUVOffset[2];	// allows external code to control movement of water base texture directly

		static float s_fWaterCycleTime1;
		static float s_fWaterCycleTime2;
		static float s_fWaterCycleTime3;

		static Asura_TrackingHandlePool<class Asura_Water_Packet>* s_pxWaterPacketList;
		static u_int s_uNumberOfWaterPackets;
		
};

class Asura_Water_Packet
{
	public :
		Asura_Water_Packet();
		~Asura_Water_Packet();

		int m_iModuleIndex;
		float m_fYPosition;
		Asura_Platform_Water_Packet* m_pxPlatformPacket;
};

#endif
