/////////////////////////////////////////////////////////////////////////////////////////
//
//	FILE:			Asura_Shapes.h
//
//	AUTHOR:			Richard M. Smith (5 August 2002)
//
/////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_SHAPES_H__
#define __ASURA_SHAPES_H__


//////////////
// INCLUDES //
//////////////

#include "Animation/Asura_AnimationAttachment.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_MaterialFlags.h"
#include "Light/Asura_Lighting_Solution.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Network/Asura_Guid.h"
#include "Rendering/Asura_Module_Defines.h" // for #define NUM_MODULES_TO_STORE_FOR_CLIPPING 4

#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
#include "Shadows/Asura_Shadow_System.h"
#endif

#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_ModuleList.h"
#endif	// defined(ASURA_USE_STREAMING)

#if defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
#include "Visibility/Asura_Visibility_System.h"
#endif

/////////////
// DEFINES //
/////////////
//#define	ASURA_HASH_ID_UNSET	( 0 )	// (RMS) Reserving zero as an "unset" value for Asura_Hash_ID

#define	ASURA_SHAPE_FLAG_HAS_COLLISION_DATA	( 1 << 0 )
#define	ASURA_SHAPE_FLAG_HAS_RENDER_DATA	( 1 << 1 )
#define	ASURA_SHAPE_FLAG_HAS_FRAGMENT_DATA	( 1 << 2 )
#define ASURA_SHAPE_FLAG_HAS_DYNAMICS_COLLISIONMESH (1 << 3)

// Maximum number of zones for module visibility clipping of shape instances.
#define ASURA_MAX_ZONES_FOR_SI_CLIPPING 2

/////////////
// CLASSES //
/////////////
class Asura_Chunk_Stream;
class Asura_CollisionMesh;
class Asura_Collision_Box;
class Asura_Shape_Instance;
class Asura_Shape_RenderData;
class Asura_Animation;
#ifdef ASURA_TOOLS
class Asura_Mesh;
template<typename T> class List;
#endif

/////////////
// FLAGS   //
/////////////

const int ASURA_SHAPE_INSTANCE_AMBIENT_ONLY         = 0x00000001;
const int ASURA_SHAPE_INSTANCE_NO_SHADOW            = 0x00000002; // This shape instance doesn't require a shadow.
const int ASURA_SHAPE_INSTANCE_FLAT_LIGHTING		= 0x00000010; // Use a single light value for the whole object.
const int ASURA_SHAPE_INSTANCE_USES_DEFAULT_AMBIENT = 0x00000020; // Use the default entity ambient lighting
const int ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING   = 0x00000040; // Ignore module linking when rendering this entity.
const int ASURA_SHAPE_INSTANCE_FULLBRIGHT			= 0x00000080; // Forces shape to be drawn with just white ambient light. If this is set , the other ambient flags/settings are ignored.
const int ASURA_SHAPE_INSTANCE_JUST_LIGHT_ONCE      = 0x00000100; // Shape will be lit when first rendered, using only static lighting, and it will never be updated.
const int ASURA_SHAPE_INSTANCE_USE_ANIM_MOVEMENT    = 0x00000200; // Compass movement from animation will be added onto the shape instance's position/orient for rendering and LOS tests. RAR 28/06/2006
const int ASURA_SHAPE_INSTANCE_BONES_IN_WORLDSPACE  = 0x00000400; // Bones required are all in worldspace.
const int ASURA_SHAPE_INSTANCE_NO_FRUSTUM_TEST      = 0x00000800; // Don't cull based on frustum tests.

enum ASURA_SHAPE_DATA_FLAGS
{
	ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB = 1,
	ASURA_SHAPE_DATA_FLAGS_HAS_NO_COLLISIONS = 2,
	ASURA_SHAPE_DATA_FLAGS_NEVER_RENDERED = 4,
};

///////////////////////
// Asura_Shape class //
///////////////////////
class Asura_Shape
{
friend class Asura_Shape_Handler;
friend class Asura_Shape_Instance;
friend class Asura_Shadow_Instance_Shape;
template<class Asura_Shape> friend class Asura_TemplateHashTree_Node;

public:
// Get functions
	inline Asura_Hash_ID GetShapeHashID( void ) const { return m_uHashID; }
	inline Asura_Hash_ID GetFragmentHashID( void ) const { return m_uFragmentHashID; }
    inline int           GetMaterialIdx( void ) const { return Platform_GetMaterialIdx(); }
    inline bool          IsOpaque( void ) const { return Platform_IsOpaque(); }

	inline const Asura_Collision_Object* GetDynamicsCollisionData(void) const
	{
		if(m_pxDynamicsCollisionMesh)
		{
			return m_pxDynamicsCollisionMesh;
		}
		if (m_uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB)
		{
			return m_pxDynamicsCollisionData;
		}
		return m_pxLOSCollisionData;
	}

    // Distinct from the above since that will always return something while
    // this specifically returns details of the separate specific dynamics
    // collision mesh, which may not exist.
    const Asura_CollisionMesh* GetDynamicsCollisionMesh( void ) const
    {
        return m_pxDynamicsCollisionMesh;
    }
    
	inline const Asura_CollisionMesh* GetLOSCollisionData(void) const
	{
		return m_pxLOSCollisionData;
	}

	inline bool HasNoCollisions() const
	{
		return (m_uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_NO_COLLISIONS) != 0;
	}
	inline bool IsNeverRendered() const
	{
		return (m_uFlags & ASURA_SHAPE_DATA_FLAGS_NEVER_RENDERED) != 0;
	}
	inline unsigned GetFlags() const
	{
		return m_uFlags;
	}
	inline Asura_Shape_RenderData* GetRenderData() const
	{
		return m_pxRenderData;
	}

// Loading functions
	void LoadShape();
#ifdef ASURA_TOOLS
    bool CreateShape(const char* szShapeFileName, const char* szMeshName );
	void Initialise(Asura_Mesh* pxMesh);
#endif

// Saving functions
#ifdef ASURA_TOOLS
	int GetChunkBufferSize( void );
	void FillInChunkBuffer( char*& pcChunkBuffer, const List<Asura_Mesh*>& xMeshList );
#endif

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

#ifdef ASURA_TOOLS
	//saves the chunks needed for the Asura_Shape (for exported files).
	void SaveShapeChunks( class Asura_SaveBlock* pxSaveBlock, const List<Asura_Mesh*>& xMeshList );
#endif

#ifdef ASURA_TOOLS
// Rebuilding the Tools_Shape_Table (PC Only)
	void GetShapeFilename( char* szString, unsigned int uMaxLen );
	void GetShapeName( char* szString, unsigned int uMaxLen );
#endif //ASURA_TOOLS


	void SetShapeData(unsigned uFlags, Asura_Bounding_Box *pxBB);

	int GetTextureIndex() const;

protected:
	// Constructors and destructors
	Asura_Shape( void );
	Asura_Shape( Asura_Hash_ID uHashID );
	Asura_Shape( Asura_Hash_ID uHashID, Asura_Chunk_Stream& xStream );
	virtual ~Asura_Shape();

private:
	// Prevent use of copy constructor/operator since we've got allocated data...
	Asura_Shape(const Asura_Shape& xOther);

	Asura_Shape& operator= (const Asura_Shape& xOther);

	// Project and platform specific functions...
	void Project_Render( const Asura_Shape_Instance* pxShapeInst, u_int uAmbientARGB, u_int uCamVisFlags ) const;
	void Platform_Render( const Asura_Shape_Instance* pxShapeInst, u_int uAmbientARGB, u_int uCamVisFlags ) const;

	void Platform_RenderShadow( const Asura_Shape_Instance* pxShapeInst, const class Asura_Shadow& xShadow, u_int uCamVisFlags ) const;

    int Platform_GetMaterialIdx( void ) const;
    bool Platform_IsOpaque( void ) const;

	// Private data...
	unsigned					m_uFlags;
	Asura_Hash_ID				m_uHashID;
	Asura_Hash_ID				m_uFragmentHashID;
	Asura_CollisionMesh*	    m_pxLOSCollisionData;
	Asura_CollisionMesh*	    m_pxDynamicsCollisionMesh;//Low polygon collision mesh used for dynamics.(Used in preference , if it exists)
	Asura_Collision_Box*		m_pxDynamicsCollisionData;
	Asura_Shape_RenderData*		m_pxRenderData;
};
// End class Asura_Shape





// (MH - 30.08.2002) this is just Asura_CollisionMesh
///////////////////////////////////////
//// Asura_Shape_CollisionData class //
///////////////////////////////////////
//class Asura_Shape_CollisionData
//{
//public:
//
//// Constructors and destrcutors
//	Asura_Shape_CollisionData::Asura_Shape_CollisionData() {};
//	Asura_Shape_CollisionData::Asura_Shape_CollisionData( char*& pcChunkBuffer )
//	{
//	}
//	virtual Asura_Shape_CollisionData::~Asura_Shape_CollisionData() {};
//
//// Saving
//	int GetChunkBufferSize( void ){ return 0; };
//	void FillInChunkBuffer( char*& pcChunkBuffer )
//	{
//	}
//
//};
//// End class Asura_Shape_CollisionData





//////////////////////////////////
// Asura_Shape_RenderData class //
//////////////////////////////////
class Asura_Shape_RenderData
{

public:

    enum { NAME_BUFFER_SIZE = 260 };

// Constructors and destrcutors
	Asura_Shape_RenderData() {};
	virtual ~Asura_Shape_RenderData() {};

//// Render
//	virtual void Render( const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri, 
//						 const unsigned int uAmbientRGBA, unsigned int uFlags = 0,
//						 const unsigned uCamVisFlags = 0xffffffff) = 0;

// Saving
	virtual int GetChunkBufferSize( void ){ return 0; };
	virtual void FillInChunkBuffer( char*& pcChunkBuffer ){};

// Loading
	virtual void Load( Asura_Hash_ID uShapeHash ){};

#ifdef ASURA_TOOLS
// Rebuilding the Tools_Shape_Table (PC Only)
	virtual void GetShapeFilename( char* szString, unsigned int uMaxLen ){};
	virtual void GetShapeName( char* szString, unsigned int uMaxLen ){};
	virtual class Asura_Mesh* GetMesh( void ){ return NULL; };
#endif //ASURA_TOOLS

	virtual void GetFragmentHashFromMesh(Asura_Hash_ID& xOutputHash) const {};

//	virtual int GetTextureIndex() const {return ASURA_TEXTURE_NONE;}
//	virtual short GetMaterialIndex() const {return ASURA_MATERIAL_NONE;}

};
// End class Asura_Shape_RenderData





////////////////////////////////
// Asura_Shape_Instance class //
////////////////////////////////
class Asura_Shape_Instance
{
	friend class Asura_Shape;

public:

// Constructors and destrcutors
	Asura_Shape_Instance();
	Asura_Shape_Instance( const Asura_Hash_ID uShapeHashID );
	virtual ~Asura_Shape_Instance();

// Set functions
	void SetShape( const Asura_Hash_ID uShapeHashID );

	void SetAnimationPointer( Asura_Animation* pxAnimation );
    void SetSkinAndAnimID( const Asura_Hash_ID uSkinID, const Asura_Hash_ID uAnimID, const int iAnimFlags,float fInitialAnimTime = 0.0f );
    void ChangeAnimation(const Asura_Hash_ID uAnimID, const int iAnimFlags, float fTweenDuration);
	void ChangeAnimation(const Asura_Hash_ID uAnimID, const int iAnimFlags, float fTweenDuration, float fStartPos, float fEndPos=1.0f );

	// range of ambient light for each component is 0x00 to 0xFF.
	// a value of 0x80 counts as maximum normal brightness.
	void Platform_SetAmbientRGB( const unsigned char ucR, const unsigned char ucG, const unsigned char ucB, const unsigned char ucA = 0xff);
	void Platform_SetAmbientARGB( const unsigned int uARGB );
	void Platform_SetDefaultAmbientRGB( void );

	void SetFlags( const unsigned int uFlags ) { m_uFlags = uFlags; }
    void AddFlags( const unsigned int uFlags ) { m_uFlags |= uFlags; }
    void RemoveFlags( const unsigned int uFlags ) { m_uFlags &= ~uFlags; }
	void SetNewPosAndOri(const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri);
	bool IsShape() const;
	void SetIgnoreForPlayer( Asura_Guid uIgnore ) { m_uIgnoreForThisPlayer = uIgnore; };
	void SetRenderOnlyForPlayer( Asura_Guid uPlayer ) { m_uRenderOnlyForThisPlayer = uPlayer; };

// Get functions
	inline Asura_Hash_ID GetShapeHashID( void ) const
	{
		return m_uShapeHashID;
	}

	//Gets the hash id for the shape if its a shape, or the skin if its a skin	
	Asura_Hash_ID GetShapeOrSkinHashID() const;

	// MGH 27/5/03: Added non-const version so that we can add attachments to the hierarchy.
	Asura_Animation* GetAnimation() { return m_pxAnimation; } 
	
	const Asura_Animation* GetAnimation() const {return m_pxAnimation;}
	const Asura_Shape* GetShape() const { return m_pxTheShape; }

	inline const Asura_Vector_3& GetPosition() const {return m_xPos;}
	inline const Asura_Matrix_3x3& GetOrientation() const {return m_xOri;}

    // The BB returned is in worldspace.
    inline const Asura_Bounding_Box& GetBB( void ) const { return m_xRotatedDisplacedBB; }
    inline Asura_Bounding_Box& GetBB( void ) { return m_xRotatedDisplacedBB; }

    inline u_int GetFlags( void ) const { return m_uFlags; }

    unsigned int GetAmbientARGB() const { return m_uAmbientARGB; }

	bool IsSphereInFrustum();
#if !defined ( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	bool TestModuleVisibility();
	void CalculateModuleList();
	void InvalidateModuleList();
#endif
	void ResetLastKnownModuleID();
	
    // Render
	void Render(unsigned uCamVisFlags = 0);
	void RenderShadow(const u_int uCamVisFlags = 0);

#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
	static void AddShadowCallback_Shape( const Asura_ShadowInstance_CreationParams& xParams,  const void *pxUserData );
	static void AddShadowCallback_Anim(  const Asura_ShadowInstance_CreationParams& xParams,  const void *pxUserData );
#endif

    // Line of sight test, on instance entity with given position and orientation.
    // If it is a hit (closer than other hits so far) then returns true and updates
	// Asura_LOS_OutputData.
    // (MH - 02.09.2002) now making the caller take care of checking and setting GUIDs.

	// AJG - 18/04/08 - Up until now, if an animation had an invalid tween store (i.e. had never been updated) we'd just test against a bounding sphere, which may have caused some oddities but seemingly no major issues.
	//					However, on AVP we're now seeing some very strange IK behaviour which has been traced to the IK LOS hitting a particularly large object that needs to be initially active, but that doesn't get
	//					rendered immediately. To solve this, we can update the animation if we hit that situation, giving it a valid tween store. However, doing that requires making the los test function non-const.
	bool LOSTest(const Asura_Vector_3& xPosition, 
				 const Asura_Matrix_3x3& xOrient,
				 const Asura_LOS_InputData* pxInput,
				 Asura_LOS_OutputData* pxOutput);

	int GetTextureIndex() const;

	// Functions to get and set the shape instance's alpha value.
	// When Alpha = 0.0 the shape instance is completly transparent.		
    // When Alpha = 1.0 the shape instance is completly opaque (then modified by texture)
	void SetAlpha( const float fAlpha )			{	m_fAlpha = fAlpha;	}
	float GetAlpha( void ) const				{	return m_fAlpha;	}

	void SetAnimTimestep( const float fAnimTimestep ) { m_fAnimTimestep = fAnimTimestep; }
	float GetAnimTimestep( void ) const				  { return m_fAnimTimestep; }


	//Gets the world position/orientation for a marker point , providing this shape instance
	//is a skin/hierarchy that has a matching marker point.
	bool GetMarkerLocationFromHash( Asura_Hash_ID uMarkerHash, Asura_Vector_3* pxOutputPosition, Asura_Matrix_3x3* pxOutputOrient, bool bWorldSpace=true, int iAttachID=Asura_AnimationAttachment::INVALID_ID ) const;

	//calculate/update the local photon list in Asura_Photon_System
 	void UpdateLighting() const;
	
	// if we have a static lighting solution, this invalidates it
	void InvalidateLighting();

	//Informs the shape instance if its being used as an animation attachment. This
	//allows it to use the parent animation's lighting solution.
	void SetParentAnimation(Asura_Animation* pxParentAnimation);
	const Asura_Animation* GetParentAnimation(void) const	{ return (m_pxParentAnimation); }
	Asura_Animation*	   GetParentAnimation(void) 		{ return (m_pxParentAnimation); }

 	//For setting texture animation controller... only supported
	//by shapes that have Asura_Animation
	void SetTextureAnimSequence(u_int uSequence);
	void SetTextureAnimSpeed(float fSpeed);

	//Checks for never rendered flag on the shape used by the shape instance.
	bool ShouldNeverBeRendered() const {return m_pxTheShape && m_pxTheShape->IsNeverRendered();}

	#if defined(ASURA_USE_STREAMING)
	// Return last known containing zone, based on module visibility tests.
	Asura_Hash_ID GetContainingZoneFromModuleTest() const;
	#endif //defined(ASURA_USE_STREAMING)

private:

	Asura_Hash_ID		m_uShapeHashID;
	const Asura_Shape*	m_pxTheShape;
    Asura_Animation*    m_pxAnimation;  // (MH - 11.10.2002) either this or TheShape will be NULL.

	Asura_Animation*	m_pxParentAnimation;//parent animation if this shape instance is being used as an attachment

	// VISTODO - m_iLastKnownModule shouldn't be needed with Asura_Visibility_System
	int					m_iLastKnownModule;

#if defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)
	Asura_Visibility_Query m_xVisibilityQuery;
	Asura_Visibility_Query m_xShadowVisibilityQuery;
#else // !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

#if defined(ASURA_USE_STREAMING)
	u_int				m_uFrameNumberCalculatedModules;
	Asura_Hash_ID		m_uLastKnownZone;
	Asura_Streaming_ModuleList<NUM_MODULES_TO_STORE_FOR_CLIPPING, ASURA_MAX_ZONES_FOR_SI_CLIPPING> m_xModuleList;
#else
	unsigned short		m_ausModuleList[NUM_MODULES_TO_STORE_FOR_CLIPPING];
#endif

#endif // !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)


	Asura_Bounding_Box	m_xRotatedDisplacedBB;
	Asura_Vector_3		m_xPos;
	Asura_Matrix_3x3	m_xOri;
	unsigned int		m_uAmbientARGB;
    float               m_fAlpha;
	float				m_fAnimTimestep;
	unsigned int		m_uFlags;
	
	/// Richard May - 11/11/2002 17:02:47 - allows shapes to be rendered for all but a given player
	/// to allow non-hud objects to be rendered and ignored on that player's screen
	Asura_Guid m_uIgnoreForThisPlayer; 
	Asura_Guid m_uRenderOnlyForThisPlayer;

	mutable Asura_Lighting_Solution m_xLightingSolution;

private:
    // (NJD - 29.04.2003) - explicitly deny access to the copy constructor
    Asura_Shape_Instance( const Asura_Shape_Instance& rhs );
};
// End class Asura_Shape_Instance





#endif //__ASURA_SHAPES_H__

////////////////////////////////////// End of file //////////////////////////////////////
