// Asura_Audio_Voice.cpp
// Thomas Young - 17.09.08
// Copyright (C) 2008 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_AUDIO_SYSTEM

#include "Core/Asura_LOS.h"
#include "Core/Asura_LOS_Penetration.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Sound/Asura_Audio_Voice.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_ReverbVolume_System.h"
#include "Volume/Asura_VolumeGroup.h"

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

static const float fASURA_AUDIO_VOICE_MIN_SPACIALISATION_DISTANCE	= 2.0f;
static const float fASURA_AUDIO_VOICE_ONE_OVER_TWO_PI				= 0.159154f;
static const float fASURA_INVALID_OCCLUSION_VALUE					= -1.0f;
static const float fASURA_INVALID_OBSTRUCTION_VALUE					= 2.0f;

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

u_int	Asura_Audio_Voice::s_uActiveOcclusionCounter	= 0;
u_int	Asura_Audio_Voice::s_uTicksSpentOnOcclusion		= 0;

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////

class Asura_Sound_Penetration : public Asura_Penetration
{
public:

	Asura_Sound_Penetration()
	{
		m_iMaxEnvLoS = 0;
		m_iMaxEntityLoS = 0;
	}

	//We don't care about the stopping power, it isn't relevant
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	float GetStoppingPower(Asura_Hash_ID uMaterialResponseHashID) const
	{
		return 0.0f;
	}
	#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	virtual float GetStoppingPower(short sMaterialID) const
	{
		return 0.0f;
	}
	#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	float GetStoppingPower( const Asura_LOS_OutputData& xOutput, unsigned short usClassification) const
	{
		return 0.0f;
	}

	int GetMaxEnvironmentLoSHits () const
	{
		return m_iMaxEnvLoS;
	}
	int GetMaxEntityLoSHits () const
	{
		return m_iMaxEntityLoS;
	}

	void SetMaxEnvironmentLoSHits (const int iHits)
	{
		m_iMaxEnvLoS = iHits;
	}
	void SetMaxEntityLoSHits (const int iHits)
	{
		m_iMaxEntityLoS = iHits;
	}

	void SetDefaults ()
	{
		m_iMaxEnvLoS	= 4;
		m_iMaxEntityLoS = 4;
	}

private:

	int m_iMaxEnvLoS;
	int m_iMaxEntityLoS;
};

static Asura_Sound_Penetration g_xPenetrationProps;

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

Asura_Audio_Voice::Asura_Audio_Voice ()
{
	m_fVolume						= 1.0f;
	m_fPitch						= 1.0f;
	m_bPitchUsesTimescale			= true;
	m_bReleased						= false;

	m_xPosition.Zero ();

	m_pxClipVolume					= 0;
	m_bInvertClipVolume				= false;
	m_bMutedByClipVolume			= false;

	m_eFalloffType					= FALLOFF_TYPE_NOT_SET;	
	m_bHasPosition					= false;

	m_bInnerRadiusNotSpacial		= false;

	m_eSpeakerMappingType			= ASURA_ASMT_DIRECT;
			
	Asura_Mem::Set( m_fSpeakerAmplitudes,		0, Asura_Audio_System::ASURA_NUM_SPEAKERS * sizeof(float) );
	Asura_Mem::Set( m_fSavedSpeakerAmplitudes,	0, Asura_Audio_System::ASURA_NUM_SPEAKERS * sizeof(float) );

	m_bUsingOcclusion				= false;
	m_bHasInitialisedOcclusion		= false;
	m_fOcclusionFactor				= fASURA_INVALID_OCCLUSION_VALUE;
	m_fObstructionFactor			= fASURA_INVALID_OCCLUSION_VALUE;
	m_fDesiredOcclusionFactor		= fASURA_INVALID_OCCLUSION_VALUE;
	m_fDesiredObstructionFactor		= fASURA_INVALID_OCCLUSION_VALUE;
	m_uOcclusionUpdateCounter		= 0;

	m_bUsingDistanceFiltering		= false;

	m_bUsingReverb					= false;
	m_bUsingMasterDSP				= false;

	#ifdef ASURA_AUDIO_STORE_FILENAMES
	m_szFilename = NULL;
	#endif
}


void Asura_Audio_Voice::PreUpdate ()
{
	s_uTicksSpentOnOcclusion = 0;
}


void Asura_Audio_Voice::Update ( const bool bFirstUpdate )
{
	//Update the clip volume

	m_bMutedByClipVolume = false;
	if( m_pxClipVolume )
	{
		bool bInside = false;
		for (u_int uListenerIdx = 0; !bInside && uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
		{
			if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
			{
				continue;
			}

			Asura_Vector_3 xListenerPosition;
			Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );

			bInside |= m_pxClipVolume->Contains( xListenerPosition );
		}

		if( bInside == m_bInvertClipVolume )
		{
			m_bMutedByClipVolume = true;
		}
	}

	//Update the speaker mappings

	switch( m_eSpeakerMappingType )
	{
		case ASURA_ASMT_DIRECT:
		{
			float fVolume = 1.0f;
			if( m_bHasPosition )
			{
				const float fRelativeSoundDistance = GetRelativeSoundDistance();
				fVolume = 1.0f - fRelativeSoundDistance;
			}
			SetAllSpeakerAmplitudes( 0.0f );
			Asura_Assert(		Asura_Audio_System::IsSpeakerActive( Asura_Audio_System::ASURA_SPEAKER_FRONT_LEFT ) 
							&&	Asura_Audio_System::IsSpeakerActive( Asura_Audio_System::ASURA_SPEAKER_FRONT_RIGHT )
							, "Expected FL and FR speakers to be present" );
			m_fSpeakerAmplitudes[ Asura_Audio_System::ASURA_SPEAKER_FRONT_LEFT	] = fVolume;
			m_fSpeakerAmplitudes[ Asura_Audio_System::ASURA_SPEAKER_FRONT_RIGHT ] = fVolume;
		}
		break;

		case ASURA_ASMT_SPACIAL:
		{
			float fVolume = 1.0f;

			Asura_Assert( m_bHasPosition, "Spacial sounds should have a position");

			if( m_bHasPosition )
			{
				const float fBleed					= 0.2f;				
				const float fRelativeSoundDistance	= GetRelativeSoundDistance();

				fVolume = 1.0f - fRelativeSoundDistance;
				
				float fMinVol = 0.0f;
				
				if ( m_bInnerRadiusNotSpacial && ( fRelativeSoundDistance <= fBleed ) )
				{
					//If we are approaching (or in) the inner radius, we need to tend all speaker volumes towards 1
					fMinVol = ( 1.0f - ( fRelativeSoundDistance * ( 1.0f / fBleed ) ) );
				}

				UpdateSpeakerAmplitudes( fVolume, m_xPosition, fMinVol );
			}
		}
		break;

		case ASURA_ASMT_FIXED_SPACIAL:
		{
			if( m_bHasPosition )
			{
				const float fRelativeSoundDistance = GetRelativeSoundDistance();
				for( int iSpeaker = 0; iSpeaker < Asura_Audio_System::ASURA_NUM_SPEAKERS; ++iSpeaker )
				{
					m_fSpeakerAmplitudes[ iSpeaker ] = m_fSavedSpeakerAmplitudes[ iSpeaker ] * ( 1.0f - fRelativeSoundDistance );
				}
			}
		}
		break;

		case ASURA_ASMT_NOT_SET:
		default:
		{
			// Keep PS3 happy
			Asura_Assert( false, "Speaker mapping type not set" );
		}
		break;
	}

	// Update the occlusion factor if relevant (always do this on the first update)
	if( Asura_Audio_System::IsOcclusionUpdateEnabled() || bFirstUpdate )
	{
		if( ShouldUpdateOcclusion() )
		{
			Asura_TickCount xUpdateTimer = Asura_Timers::GetCurTicks();

			UpdateOcclusionData();

			++m_uOcclusionUpdateCounter;	// Flag this voice as having occlusion updated

			s_uTicksSpentOnOcclusion += static_cast<u_int>( Asura_Timers::GetElapsedTicks( xUpdateTimer ) );
		}

		// We should tend to our desired occlusion/obstruction values regardless of if they were updated this frame
		if( IsUsingOcclusion() )
		{
			if( !m_bHasInitialisedOcclusion )
			{
				// Make sure the occlusion/obstruction values start correctly
				m_fOcclusionFactor   = m_fDesiredOcclusionFactor;
				m_fObstructionFactor = m_fDesiredObstructionFactor;

				m_bHasInitialisedOcclusion = true;
			}
			else
			{
				const float fChange = Asura_Timers::GetFrameTime();

				if( Asura_Maths::Abs( m_fOcclusionFactor - m_fDesiredOcclusionFactor ) < fChange )
				{
					m_fOcclusionFactor = m_fDesiredOcclusionFactor;
				}
				else
				{
					if( m_fDesiredOcclusionFactor > m_fOcclusionFactor )	{	m_fOcclusionFactor += fChange;	}
					else													{	m_fOcclusionFactor -= fChange;	}
				}

				if( Asura_Maths::Abs( m_fObstructionFactor - m_fDesiredObstructionFactor ) < fChange )
				{
					m_fObstructionFactor = m_fDesiredObstructionFactor;
				}
				else
				{
					if( m_fDesiredObstructionFactor > m_fObstructionFactor )	{	m_fObstructionFactor += fChange;	}
					else														{	m_fObstructionFactor -= fChange;	}
				}
			}
		}
	}
}

void Asura_Audio_Voice::PostUpdate ()
{
	if( s_uTicksSpentOnOcclusion == 0 )
	{
		// Didnt update anything, increment the counter. The nature of this update mechanism means that
		// we miss one frame of updates (up to every other frame), but thats fine for something like occlusion
		++ s_uActiveOcclusionCounter;
	}
}


float Asura_Audio_Voice::GetRelativeSoundDistance () const
{
	// This function returns the relative sound distance from inner (1.0) to outer (0.0) radius

	Asura_Assert( HasPosition(), "Calling GetRelativeSoundDistance on a sound with no position" );

	float fReturnVal = 0.0f;
	
	float fNearestDistSq = Asura_Maths::LargeFloat;
	Asura_Vector_3 xNearestListenerPos;
	xNearestListenerPos.Zero();
	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xListenerPosition;
		Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );

		const float fDistSq = ( xListenerPosition - m_xPosition ).MagnitudeSquared();
		if ( fDistSq < fNearestDistSq )
		{
			xNearestListenerPos = xListenerPosition;
			fNearestDistSq = fDistSq;
		}
	}

	switch( m_eFalloffType )
	{
		case FALLOFF_TYPE_CUBOID:
		{
			Asura_Vector_3 xRelativePos( &m_xPosition.x );
			xRelativePos -= xNearestListenerPos;
			xRelativePos = m_xCuboidFalloff.m_xOrient.RotateVector( xRelativePos);

			fReturnVal = Asura_Audio_System::CalcVolFalloff( m_xCuboidFalloff.m_xCuboidInner, m_xCuboidFalloff.m_xCuboidOuter, xRelativePos );
		}
		break;

		case FALLOFF_TYPE_SPHERICAL:
		{
			Asura_Vector_3 xListenerToSound = Asura_Vector_3( xNearestListenerPos - m_xPosition );

			const float fDistance = xListenerToSound.Magnitude();

			if( fDistance > m_xSphericalFalloff.m_fOuterRadius )
			{
				//Outside the outer range
				fReturnVal = 0.0f;
			}
			else if( fDistance <= m_xSphericalFalloff.m_fInnerRadius )
			{
				//Inside the inner range
				fReturnVal = 1.0f;
			}
			else
			{
				//Somewhere between inner and outer
				fReturnVal = CalcVolFalloff( m_xSphericalFalloff.m_fInnerRadius, m_xSphericalFalloff.m_fOuterRadius, fDistance );
			}
		}
		break;

		case FALLOFF_TYPE_NOT_SET:
		{
			// Assume we are exactly on the sound
			fReturnVal = 1.0f;
		}
		break;

		default:					break;
	}
	
	return ( 1.0f - fReturnVal );
};


float Asura_Audio_Voice::GetReverbSendLevel() const
{
	if( HasPosition() )
	{
		return ( 1.0f - GetRelativeSoundDistance() );
	}
	return 1.0f;
}


float Asura_Audio_Voice::CalcVolFalloff(	const Asura_Vector_3& xInnerRange,
											const Asura_Vector_3& xOuterRange, 
											const Asura_Vector_3& xPos ) const
{
	float afFalloff[3];

	for( u_int u = 0; u < 3; ++u )
	{
		const float fInner = xInnerRange.GetElement(u);
		const float fOuter = xOuterRange.GetElement(u);

		const float fAbsDist = Asura_Maths::Abs( xPos.GetElement(u) );
		if( fAbsDist > fOuter ) return 0.0f;

		afFalloff[u] = CalcVolFalloff( fInner, fOuter, fAbsDist );
	}

	return Asura_Maths::Min( Asura_Maths::Min( afFalloff[0], afFalloff[1] ), afFalloff[2] );
}


float Asura_Audio_Voice::CalcVolFalloff(	const float fInnerRadius, 
											const float fOuterRadius,
											const float fDistance ) const
{
	const float fRange = fOuterRadius - fInnerRadius;
	const float fProportion = (fOuterRadius - fDistance) / fRange;
	return Asura_Maths::ClampToRange( fProportion, 0.0f, 1.0f );
}


void Asura_Audio_Voice::SetAllSpeakerAmplitudes ( const float fAmplitude )
{
	//Set all speakers to this amplitude
	for( u_int uSpeaker = 0; uSpeaker < static_cast<u_int>( Asura_Audio_System::ASURA_NUM_SPEAKERS ); ++uSpeaker )
	{
		if( Asura_Audio_System::IsSpeakerActive(static_cast<Asura_Audio_System::Speakers>( uSpeaker ) ) )
		{
			m_fSpeakerAmplitudes[ uSpeaker ] = fAmplitude;
		}
	}
}


void Asura_Audio_Voice::SetAllSpeakerMinAmplitudes ( const float fMinAmplitude )
{
	//Set all speakers to at least this amplitude
	for( u_int uSpeaker = 0; uSpeaker < static_cast<u_int>( Asura_Audio_System::ASURA_NUM_SPEAKERS ); ++uSpeaker )
	{
		if( Asura_Audio_System::IsSpeakerActive(static_cast<Asura_Audio_System::Speakers>( uSpeaker ) ) )
		{
			m_fSpeakerAmplitudes[ uSpeaker ] = Asura_Maths::Max( fMinAmplitude, m_fSpeakerAmplitudes[ uSpeaker ] );
		}
	}
}

void Asura_Audio_Voice::UpdateSpeakerAmplitudes ( const float fBaseAmplitude, const Asura_Vector_3& xSoundPosition, const float fMinVol )
{
	Asura_Vector_3 xListenerPos;
	xListenerPos.Zero();
	Asura_Matrix_3x3 xListenerOri;
	xListenerOri.Identity();

	float fNearestDistSq = Asura_Maths::LargeFloat;
	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xCurrentListenerPosition;
		Asura_Matrix_3x3 xCurrentListenerOri;
		Asura_Audio_System::GetListenerPosition( &xCurrentListenerPosition, &xCurrentListenerOri, uListenerIdx );

		const float fDistSq = ( xCurrentListenerPosition - m_xPosition ).MagnitudeSquared();
		if ( fDistSq < fNearestDistSq )
		{
			xListenerPos = xCurrentListenerPosition;
			xListenerOri = xCurrentListenerOri;
			fNearestDistSq = fDistSq;
		}
	}

	//Get the angle from the listener to the sound
	Asura_Vector_3 xListenerToSound = Asura_Vector_3( xSoundPosition - xListenerPos );
	const float fDistanceToSound2D	= Asura_Vector_2( xListenerToSound.x, xListenerToSound.z ).Magnitude();
	xListenerToSound.Normalise ();
	
	float fAngle = Asura_Maths::ArcTangent(xListenerToSound.x, xListenerToSound.z);

	//Mofidy this angle with the direction the Listener is facing
	fAngle -= xListenerOri.Heading();
	
	//Make a vector of this angle (this step is a bit redundant if we were to store the speaker positions as angles)
	Asura_Vector_2 xDirection2D = Asura_Vector_2( Asura_Maths::Sine( fAngle ), Asura_Maths::Cosine( fAngle ) );
	xDirection2D.Normalise();
	
	//The bleed defines how far around the spacial sound field a sound will travel (i.e. how many speakers will render it)
	// TODO: I want to add special cases here for stereo since it sort of b0rks the spacialisation algorithm
	const float fBleed = Asura_Maths::Min( Asura_Audio_System::GetMaxAngleBetweenSpeakers() * fASURA_AUDIO_VOICE_ONE_OVER_TWO_PI * 0.75f, 1.0f );
	
	float fInternalMinVol = 0.0f;

	//If we are very close to the sound, increase the amplitude of all speakers to avoid sudden snaps from one speaker to another
	if( fDistanceToSound2D < fASURA_AUDIO_VOICE_MIN_SPACIALISATION_DISTANCE )
	{
		fInternalMinVol = ( fASURA_AUDIO_VOICE_MIN_SPACIALISATION_DISTANCE - fDistanceToSound2D ) / fASURA_AUDIO_VOICE_MIN_SPACIALISATION_DISTANCE;
	}

	float fTotal = 0.0f;
	
	for( u_int uSpeaker = 0; uSpeaker < static_cast<u_int>( Asura_Audio_System::ASURA_NUM_SPEAKERS ); ++uSpeaker )
	{
		Asura_Audio_System::Speakers eSpeaker = static_cast<Asura_Audio_System::Speakers>( uSpeaker );
		if( Asura_Audio_System::IsSpeakerActive( eSpeaker ) )
		{
			const float fDot = Asura_Audio_System::GetSpeakerPosition( eSpeaker ) * xDirection2D;
			const float fAngleBetween = Asura_Maths::ArcCosine( fDot ) * fASURA_AUDIO_VOICE_ONE_OVER_TWO_PI;
			
			if( fAngleBetween > fBleed )
			{
				//Beyond the audible angle
				m_fSpeakerAmplitudes[ uSpeaker ] = 0.0f;
			}
			else if ( fAngleBetween == 0.0f )
			{
				//Avoid division by zero
				m_fSpeakerAmplitudes[ uSpeaker ] = 1.0f;
			}
			else
			{
				m_fSpeakerAmplitudes[ uSpeaker ] = 1.0f - fAngleBetween / fBleed;
			}

			m_fSpeakerAmplitudes[ uSpeaker ] = Asura_Maths::Max( m_fSpeakerAmplitudes[ uSpeaker ], fInternalMinVol );
			m_fSpeakerAmplitudes[ uSpeaker ] = Asura_Maths::Max( m_fSpeakerAmplitudes[ uSpeaker ], fMinVol );
			m_fSpeakerAmplitudes[ uSpeaker ] = Asura_Maths::Min( m_fSpeakerAmplitudes[ uSpeaker ], 1.0f );

			fTotal += m_fSpeakerAmplitudes[ uSpeaker ];
		}
	}

	// Normalise and apply base volume
	fTotal = 1.0f / fTotal;

	// (AAM - 14.05.2009) Removing this temporarily in AVP E3 branch, till we can work out the
	// differences between direct and spatialised sounds
	const float fMultiplier = fBaseAmplitude /** fTotal*/;

	for( u_int uSpeaker = 0; uSpeaker < static_cast<u_int>( Asura_Audio_System::ASURA_NUM_SPEAKERS ); ++uSpeaker )
	{
		Asura_Audio_System::Speakers eSpeaker = static_cast<Asura_Audio_System::Speakers>( uSpeaker );
		if( Asura_Audio_System::IsSpeakerActive( eSpeaker ) )
		{
			m_fSpeakerAmplitudes[ uSpeaker ] *= fMultiplier;
		}
	}
}


void Asura_Audio_Voice::SetVolume ( const float fVolume )
{
	m_fVolume = fVolume;
}


float Asura_Audio_Voice::GetVolume () const
{
	if( m_bMutedByClipVolume )
	{
		return 0.0f;
	}
	return m_fVolume;
}


void Asura_Audio_Voice::SetPitch ( const float fPitch )
{
	m_fPitch = fPitch;
}


void Asura_Audio_Voice::SetPitchUsesTimescale ( const bool bUseTimescale )
{
	m_bPitchUsesTimescale = bUseTimescale;
}


float Asura_Audio_Voice::GetPitch () const
{
	float fFinalPitch = m_fPitch;

	if( m_bPitchUsesTimescale )
	{
		fFinalPitch *= Asura_Timers::GetGameTimeScale();
	}

	//TODO: Multiply pitch by doppler factor here

	return fFinalPitch;
}


void Asura_Audio_Voice::SetPos ( const Asura_Vector_3& xPos )
{
	Asura_Assert( !xPos.AnyComponentIsNaN(), "Asura Audio voice got a garbage position - PLEASE BREAK HERE AND CHECK YOUR CALL STACK TO SEE WHY" );
	if( !xPos.AnyComponentIsNaN() )
	{
		m_xPosition = xPos;
	}
}


void Asura_Audio_Voice::SetRadii ( const float fInner, const float fOuter )
{
	m_eFalloffType = FALLOFF_TYPE_SPHERICAL;

	m_xSphericalFalloff.m_fInnerRadius = fInner;
	m_xSphericalFalloff.m_fOuterRadius = fOuter;
}


void Asura_Audio_Voice::SetCuboid ( const Asura_Vector_3& xInner, const Asura_Vector_3& xOuter, const Asura_Quat& xOrient )
{
	m_eFalloffType = FALLOFF_TYPE_CUBOID;

	m_xCuboidFalloff.m_xCuboidInner		= xInner;
	m_xCuboidFalloff.m_xCuboidOuter		= xOuter;
	m_xCuboidFalloff.m_xOrient			= xOrient;
	m_xCuboidFalloff.m_xOrient.w		= -xOrient.w;
}


void Asura_Audio_Voice::SetInvertClipVolume ( const bool bInvert )
{
	m_bInvertClipVolume = bInvert;
}


void Asura_Audio_Voice::SetClipVolume ( const Asura_VolumeGroup* const pxVolume )
{
	m_pxClipVolume = pxVolume;
}


float Asura_Audio_Voice::GetSpeakerAmplitude ( Asura_Audio_System::Speakers eSpeaker ) const
{
	if( Asura_Audio_System::IsSpeakerActive( eSpeaker ) )
	{
		return m_fSpeakerAmplitudes[ eSpeaker ];
	}
	return 0.0f;
}

void Asura_Audio_Voice::SetFixedSoundFieldPos( const Asura_Vector_2& xPosition )
{
	Asura_Vector_3 xListenerPos;
	xListenerPos.Zero();
	Asura_Matrix_3x3 xListenerOri;
	xListenerOri.Identity();

	float fNearestDistSq = Asura_Maths::LargeFloat;
	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xCurrentListenerPosition;
		Asura_Matrix_3x3 xCurrentListenerOri;
		Asura_Audio_System::GetListenerPosition( &xCurrentListenerPosition, &xCurrentListenerOri, uListenerIdx );

		const float fDistSq = ( xCurrentListenerPosition - m_xPosition ).MagnitudeSquared();
		if ( fDistSq < fNearestDistSq )
		{
			xListenerPos = xCurrentListenerPosition;
			xListenerOri = xCurrentListenerOri;
			fNearestDistSq = fDistSq;
		}
	}

	Asura_Vector_3 xRelativePos;

	xRelativePos.Zero();
	xRelativePos.x = xPosition.x;
	xRelativePos.z = xPosition.y;
	xRelativePos.Normalise();
		
	xRelativePos = xRelativePos * xListenerOri;							// Rotate the position to be relative to the Listener
	xRelativePos *= fASURA_AUDIO_VOICE_MIN_SPACIALISATION_DISTANCE;		// Multiply by the min spacialisation distance so that the extremeties are fully spacialised
	xRelativePos += xListenerPos;										// Offset by the Listener position

	UpdateSpeakerAmplitudes( 1.0f, xRelativePos );						// Use our standard spacialisation algorithm to get the speaker volumes

	// Save these speaker amplitudes rather than re-calculating them every time
	Asura_Mem::Copy( m_fSavedSpeakerAmplitudes, m_fSpeakerAmplitudes, sizeof( float ) * static_cast<u_int>( Asura_Audio_System::ASURA_NUM_SPEAKERS ) );
}


float Asura_Audio_Voice::GetDistanceFilterFactor () const
{
	if( !m_bHasPosition )
	{
		Asura_Assertf( false , ("Trying to apply distance filter to a voice with no position '%s'.", GetFilename()) );
		return 0.0f;
	}
	
	// return the normalised position between our min and max filter ranges

	float fNearestDistSq = Asura_Maths::LargeFloat;
	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xCurrentListenerPosition;
		Asura_Matrix_3x3 xCurrentListenerOri;
		Asura_Audio_System::GetListenerPosition( &xCurrentListenerPosition, &xCurrentListenerOri, uListenerIdx );

		const float fDistSq = ( xCurrentListenerPosition - m_xPosition ).MagnitudeSquared();
		if ( fDistSq < fNearestDistSq )
		{
			fNearestDistSq = fDistSq;
		}
	}

	const float fDistanceToListener = Asura_Maths::Sqrt( fNearestDistSq );

	const float fMinRange = Asura_Audio_System::GetDistanceFilterMinRange();
	const float fMaxRange = Asura_Audio_System::GetDistanceFilterMaxRange();

	if( fDistanceToListener <= fMinRange )
	{
		return 0.0f;
	}
	else if( fDistanceToListener > fMaxRange )
	{
		return 1.0f;
	}
	
	return ( ( fDistanceToListener - fMinRange ) / ( fMaxRange - fMinRange ) );
}


bool Asura_Audio_Voice::ShouldUpdateOcclusion ()
{
	if( !IsUsingOcclusion() )
	{
		return false;
	}

	if( !m_bHasInitialisedOcclusion )
	{
		// Always update occlusion if we havent done so at least once
		return true;
	}

	if( s_uActiveOcclusionCounter == m_uOcclusionUpdateCounter )
	{
		return false;
	}

	if( !m_bHasPosition )
	{
		// IsUsingOcclusion should return false for non-spacial sounds
		Asura_Assertf( false , ("Sounds with no position should not be flagged to use occlusion '%s'.", GetFilename()) );
		return false;
	}

	if( s_uTicksSpentOnOcclusion >= Asura_Timers::SecondsToTicks( Asura_Audio_System::GetOcclusionCPUTimePerFrame() ) )
	{
		// Spent enough time this frame updating occlusion
		return false;
	}

	#ifdef ASURA_TOOLS
	//Dont enable occlusion for most tools modes, can cause big slowdowns
	if(Tools_System::IsActive())
	{
		const int iMode = Tools_System::GetCurrentMode();
		switch (iMode)
		{
			// These modes are okay
			case TOOLS_MODE_OBJECT:
			case TOOLS_MODE_CUTSCENE:
			case TOOLS_MODE_STANDBY:
			case TOOLS_MODE_SOUND:		break;

			default:
			{
				return false;
			}
		}
	}
	#endif

	return true;
}


void Asura_Audio_Voice::UpdateOcclusionData ()
{
	//update the occlusion/obstruction data for this sound
	//this involves doing LoS tests and then calling project code to determine the occlusion factor of the hit item
	
	//Set the default penetration properties
	g_xPenetrationProps.SetDefaults();

	float fLeastObstruction = fASURA_INVALID_OBSTRUCTION_VALUE;
	Asura_Vector_3 xListenerPosition = m_xPosition;

	GetOcclusionForLeastOccludedListener( fLeastObstruction, xListenerPosition );

	if(fLeastObstruction > 1.0f)
	{
		//Didnt assign fLeastOcclusion, default to no occlusion
		fLeastObstruction = 0.0f;
	}

	bool bOcclude = false;	// Occlude or obstruct?

	//Now check if we are in the same reverb zone as the listener ( occlusion factor = obstruction value if in the same zone )
	if( ( fLeastObstruction != 0.0f ) )
	{
		#ifdef ASURA_USE_REVERBVOLUME_SYSTEM
		Asura_ReverbVolume_Handler xHandler = Asura_ReverbVolume_System::GetCurrentHandler();

		const Asura_ReverbVolume* pxCamVolume = xHandler.GetReverbVolumeFromPosition( xListenerPosition );
		if(pxCamVolume)
		{
			const Asura_ReverbVolume* pxSoundVolume = xHandler.GetReverbVolumeFromPosition( m_xPosition );
			if(pxCamVolume != pxSoundVolume)
			{
				// We are in a different reverb zone, so occlude
				bOcclude = true;
			}

		}
		#endif
	}

	if( bOcclude )
	{
		m_fDesiredObstructionFactor = 0.0f;
		m_fDesiredOcclusionFactor	= fLeastObstruction;
	}
	else
	{
		m_fDesiredObstructionFactor = fLeastObstruction;
		m_fDesiredOcclusionFactor	= 0.0f;
	}
	
	// If the obstruction/occlusion factors are uninitialised, set them now
	if(		( m_fObstructionFactor	== fASURA_INVALID_OCCLUSION_VALUE )
		||	( m_fOcclusionFactor	== fASURA_INVALID_OCCLUSION_VALUE ) )
	{
		m_fObstructionFactor	= m_fDesiredObstructionFactor;
		m_fOcclusionFactor		= m_fDesiredOcclusionFactor;
	}
}

void Asura_Audio_Voice::GetOcclusionForLeastOccludedListener( float& fLeastObstruction, Asura_Vector_3& xListenerPosition )
{
	//loop over all all cameras and take the least occluded as our occlusion factor
	for ( u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx )
	{
		float fCurrentListenerObstruction = fASURA_INVALID_OBSTRUCTION_VALUE;
		if ( !Asura_Audio_System::IsListenerActive( uListenerIdx ) )
		{
			continue;
		}
		//Fire a ray from the sound centre to the sound and check the occlusion factor of everything it intersects
		fCurrentListenerObstruction = GetOcclusionForListener( uListenerIdx );

		//Check that we set the camera occlusion first
		if(fCurrentListenerObstruction <= 1.0f)
		{
			if(fCurrentListenerObstruction < fLeastObstruction)
			{
				fLeastObstruction = fCurrentListenerObstruction;
				Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );
			}
		}
	}
}

float Asura_Audio_Voice::GetOcclusionForListener( const u_int uListenerIdx )
{
	float fLeastObstruction = fASURA_INVALID_OBSTRUCTION_VALUE;

	Asura_Audio_OcclusionFactor_Callback pfnEnvFactorCallback = Asura_Audio_System::GetEnvironmentOcclusionFactorCallback();
	Asura_Audio_OcclusionFactor_Callback pfnEntFactorCallback = Asura_Audio_System::GetEntityOcclusionFactorCallback();

	Asura_Vector_3 xListenerPosition;
	Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );

	Asura_Vector_3 xDirection = m_xPosition - xListenerPosition;
	const float fMagnitude = xDirection.Normalise();

	const Asura_LOS_InputData xInput( xListenerPosition, xDirection, Asura_Audio_System::GetGUIDToIgnoreForOcclusion( uListenerIdx ), fMagnitude );

	Asura_LOS_Penetration::TestAgainstEverything(&xInput,g_xPenetrationProps);

	for(int iHit = 0; iHit < Asura_LOS_Penetration::GetNumberOfHits(); ++iHit)
	{
		const Asura_LOS_OutputData* pxOutput = Asura_LOS_Penetration::GetHit(iHit);
		if( pxOutput && ( pxOutput->m_bHitEnvironment || pxOutput->m_uHitEntity ) )
		{
			float fHitObstruction = 2.0f;
			if( pxOutput->m_bHitEnvironment )
			{
				if ( pfnEnvFactorCallback)
				{
					fHitObstruction = pfnEnvFactorCallback( pxOutput );
					Asura_Assert((fHitObstruction >= 0.0f) && (fHitObstruction <= 1.0f), "Invalid occlusiton factor returned from project 'get occlusion' function");
				}
			}
			else if ( pfnEntFactorCallback )
			{
				fHitObstruction = pfnEntFactorCallback( pxOutput );
				Asura_Assert((fHitObstruction >= 0.0f) && (fHitObstruction <= 1.0f), "Invalid occlusiton factor returned from project 'get occlusion' function");
			}
			if(fHitObstruction < fASURA_INVALID_OBSTRUCTION_VALUE) //assuming we have set it to something reasonable
			{
				if(fLeastObstruction == fASURA_INVALID_OBSTRUCTION_VALUE)
				{
					//Hasnt been set yet
					fLeastObstruction = 1.0f - fHitObstruction; //store the inverse so we can just multiply them up
				}
				else
				{
					//Has been set, so attenuate
					fLeastObstruction *= (1.0f - fHitObstruction);
				}
			}
		}
	}
	//flip it back again
	if(fLeastObstruction <= 1.0f)
	{
		fLeastObstruction = 1.0f - fLeastObstruction;
	}

	return fLeastObstruction;
}

void Asura_Audio_Voice::SetUsingOcclusion( bool bOcclude )
{
	if( IsPlaying() )
	{
		Asura_Assert( false, "Cannot set a voice 'use occlusion' property after it has started playing" );
		return;
	}
	m_bUsingOcclusion = bOcclude;
}


void Asura_Audio_Voice::SetUsingDistanceFiltering( bool bFilter )
{
	if( IsPlaying() )
	{
		Asura_Assert( false, "Cannot set a voice 'use occlusion' property after it has started playing" );
		return;
	}
	m_bUsingDistanceFiltering = bFilter;
}


void Asura_Audio_Voice::SetUsingReverb( bool bReverb )
{ 
	if( IsPlaying() )
	{
		Asura_Assert( false, "Cannot set a voice 'use reverb' property after it has started playing" );
		return;
	}
	m_bUsingReverb = bReverb; 
}


void Asura_Audio_Voice::SetUsingMasterDSP( bool bDSP )
{
	if( IsPlaying() )
	{
		Asura_Assert( false, "Cannot set a voice 'use master DSP' propert after it has started playing" );
		return;
	}
	m_bUsingMasterDSP = bDSP;
}

#endif	//ASURA_USE_AUDIO_SYSTEM

//eof
