//////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_Voice.h
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#if defined ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#if !defined __ASURA_AUDIO_VOICE_H__
#define __ASURA_AUDIO_VOICE_H__

#if defined ASURA_USE_AUDIO_SYSTEM

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Asura Platform-independent
#include "Core/Asura_Maths.h"
#include "Sound/Asura_Audio_System.h"

/******************************************************************************/
/* Classes                                                                    */
/******************************************************************************/

class Asura_Quat;
class Asura_VolumeGroup;

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_SpatialType
//
// Spatial types for voice processing
//
////////////////////////////////////////////////////////////////////////////////

enum Asura_Audio_SpeakerMappingType
{
	ASURA_ASMT_NOT_SET,

	ASURA_ASMT_DIRECT,			// Each channel in the audio maps to a speaker (except mono which is upmixed)
	ASURA_ASMT_SPACIAL,			// Calculate the speaker volumes from position and orientation
	ASURA_ASMT_FIXED_SPACIAL,	// Sound always appears to come from one place in the sound field
};

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_Voice
//
// Represents a playback voice. Platforms will derive their own versions
//
////////////////////////////////////////////////////////////////////////////////

class Asura_Audio_Voice
{
	friend class Asura_Audio_System;

public:

	// Platform-independent functions

	Asura_Audio_Voice();
	virtual ~Asura_Audio_Voice() {}

	#if defined ASURA_AUDIO_STORE_FILENAMES
		const char* GetFilename() const { return m_szFilename; }
	#else
		const char* GetFilename() const { return "[Unknown]"; }
	#endif

	void SetSpeakerMappingType( const Asura_Audio_SpeakerMappingType eMapping )	{ m_eSpeakerMappingType	= eMapping;			}
	void SetHasPosition( const bool bHasPosition )								{ m_bHasPosition		= bHasPosition;		}
	bool HasPosition()	const													{ return m_bHasPosition;					}
	
	void SetInnerRadiusDoesNotSpacialise( const bool bNoSpacial = true )		{ m_bInnerRadiusNotSpacial	= bNoSpacial;	}

	void SetVolume( const float fVolume );
	
	void SetPitch( const float fPitch );
	void SetPitchUsesTimescale( const bool bUseTimescale );
	
	void SetPos( const Asura_Vector_3& xPos );
	void SetFixedSoundFieldPos( const Asura_Vector_2& xPosition );
	
	void SetRadii( const float fInner, const float fOuter );
	void SetCuboid( const Asura_Vector_3& xInner, const Asura_Vector_3& xOuter, const Asura_Quat& xOrient );

	void SetInvertClipVolume( const bool bInvert );
	void SetClipVolume( const Asura_VolumeGroup* const pxVolume );

	void Release()			{ m_bReleased = true;	}
	bool IsReleased() const	{ return m_bReleased;	}

	void SetUsingOcclusion( bool bOcclude = true );
	bool IsUsingOcclusion() const { return m_bUsingOcclusion; }

	void SetUsingDistanceFiltering( bool bFilter = true );
	bool IsUsingDistanceFiltering() const { return m_bUsingDistanceFiltering; }

	void SetUsingReverb( bool bReverb = true );
	bool IsUsingReverb() const { return m_bUsingReverb; }

	void SetUsingMasterDSP( bool bDSP = true );
	bool IsUsingMasterDSP() const { return m_bUsingMasterDSP; }

	// Virtual functions

	virtual void Update( const bool bFirstUpdate = false );

	virtual void SetMixerChannel( const Asura_Hash_ID uChannel ) {};

	// Pure virtual functions

	virtual void EnqueueWave( const Asura_Hash_ID uWave ) = 0;

	virtual void Start() = 0;
	virtual bool IsPlaying() const = 0;

	virtual void SeekAtPlaybackTime( const float fSeekPosition, const float fPlaybackTime ) = 0;
	virtual bool SeekIsPending() = 0;

	virtual float GetPlayPosition() const = 0;

	virtual bool SetSubmixOutputs( const Asura_Handle* auSubmixVoices, const u_int uNumVoices ) = 0;
	virtual void ClearSubmixOutputs () = 0;

	virtual bool IsPreparing() const { return false; }

protected:

	float	GetVolume() const;
	float	GetPitch() const;
	float	GetSpeakerAmplitude( Asura_Audio_System::Speakers eSpeaker ) const;
	float	GetDistanceFilterFactor() const;
	float	GetObstructionFactor()		const { return m_fObstructionFactor;	}
	float	GetOcclusionFactor()		const { return m_fOcclusionFactor;		}
	float	GetRelativeSoundDistance()	const;
	float	GetReverbSendLevel()		const;
	
	#if defined ASURA_AUDIO_STORE_FILENAMES
		const char* m_szFilename;
	#endif

private:

	// ------------------- Functions ------------------------

	void	SetAllSpeakerAmplitudes		( const float fAmplitude );
	void	UpdateSpeakerAmplitudes		( const float fBaseAmplitude, const Asura_Vector_3& xSoundPosition, const float fMinVol = 0.0f );
	void	SetAllSpeakerMinAmplitudes	( const float fMinAmplitude );

	float	CalcVolFalloff	(	const Asura_Vector_3& xInnerRange,
								const Asura_Vector_3& xOuterRange, 
								const Asura_Vector_3& xPos ) const;

	float	CalcVolFalloff	(	const float fInnerRadius, 
								const float fOuterRadius,
								const float fDistance ) const;

	static void PreUpdate	();
	static void PostUpdate	();

	bool	ShouldUpdateOcclusion	( );
	void	UpdateOcclusionData		( );
	void	GetOcclusionForLeastOccludedListener( float& fLeastObstruction, Asura_Vector_3& xListenerPosition );
	float	GetOcclusionForListener( const u_int uListenerIdx );
	// ------------------ Enumerations ------------------------

	enum FalloffType
	{
		FALLOFF_TYPE_NOT_SET,
		FALLOFF_TYPE_SPHERICAL,
		FALLOFF_TYPE_CUBOID
	};

	// ------------------- Structures ------------------------

	struct Falloff_Spherical
	{
		float					m_fInnerRadius;
		float					m_fOuterRadius;

	} m_xSphericalFalloff;

	struct Falloff_Cuboid
	{
		Asura_Vector_3			m_xCuboidInner;
		Asura_Vector_3			m_xCuboidOuter;
		Asura_Quat				m_xOrient;

	} m_xCuboidFalloff;

	// -------------------- Variables ------------------------
	
	float							m_fVolume;
	float							m_fPitch;
	bool							m_bPitchUsesTimescale;
	bool							m_bReleased;
	
	Asura_Vector_3					m_xPosition;
	
	const Asura_VolumeGroup*		m_pxClipVolume;
	bool							m_bInvertClipVolume;
	bool							m_bMutedByClipVolume;
	FalloffType						m_eFalloffType;
	bool							m_bHasPosition;
	bool							m_bInnerRadiusNotSpacial;

	Asura_Audio_SpeakerMappingType	m_eSpeakerMappingType;
	float							m_fSpeakerAmplitudes		[ Asura_Audio_System::ASURA_NUM_SPEAKERS ];
	float							m_fSavedSpeakerAmplitudes	[ Asura_Audio_System::ASURA_NUM_SPEAKERS ];

	bool							m_bUsingOcclusion;
	bool							m_bHasInitialisedOcclusion;
	float							m_fOcclusionFactor;
	float							m_fObstructionFactor;
	float							m_fDesiredOcclusionFactor;
	float							m_fDesiredObstructionFactor;
	u_int							m_uOcclusionUpdateCounter;

	bool							m_bUsingDistanceFiltering;
	
	bool							m_bUsingReverb;
	bool							m_bUsingMasterDSP;

	static u_int					s_uActiveOcclusionCounter;
	static u_int					s_uTicksSpentOnOcclusion;
};

#endif // defined ASURA_USE_AUDIO_SYSTEM

#endif // !defined __ASURA_AUDIO_VOICE_H__

//eof
