//////////////////////////////////////////////////////////////////////////////
//
// Asura_Dialogue.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     internal header for dialogue system
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_DIALOGUE_H___
#define ___ASURA_DIALOGUE_H___ 1

#ifdef ASURA_USE_DIALOGUE_SYSTEM

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura_Hash.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_HeapArray.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_StaticMessageBlock.h"
#include "Network/Asura_Guid.h"

#include "Core/Asura_PlatformBasics.h"

#ifdef ASURA_USE_ARTICULATION_SYSTEM
#include "LipSync/Asura_Lipsync_Manager.h"
#endif

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Dialogue_CharacterMapping;
class Asura_Dialogue_CharacterMappingList;
class Asura_Dialogue_Event;
class Asura_Dialogue_Instance;
class Asura_Dialogue_Line;
class Asura_Dialogue_Line_Template;
class Asura_Dialogue_Message;
class Asura_Sound_Mixer_Channel;
class Asura_Sound_Handler;
class Asura_Dialogue_StaticMessageBlock;
class Asura_Dialogue_PlaybackObject;
class Asura_Stream;
template <class T> class Asura_ResourceSet;

// D E F I N E S /////////////////////////////////////////////////////////////
#ifndef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
#if defined(ASURA_TOOLS) || defined(_DEBUG)
#define ASURA_DIALOGUE_INCLUDE_TOOLS_DATA	1	// text names, sound categories etc.
#endif
#endif

// E N U M S /////////////////////////////////////////////////////////////////
enum ASURA_DIALOGUE_PLAYBACK_TYPE
{
	ASURA_DPT_ALLSEQUENTIAL,	// one at a time triggering next one when finished
	ASURA_DPT_ONEBYONE,			// current one in list, stopping and advancing in list when finished 
	ASURA_DPT_RANDOM,			// pick at random
	ASURA_DPT_CDPLAYERRANDOM,	// like a CD random function (plays all in random order)
};

//playable class types for use in the playable ID class
enum ASURA_DIALOGUE_PLAYABLE_TYPE
{
	ASURA_DIALOGUE_PLAYABLE_LINE,
	ASURA_DIALOGUE_PLAYABLE_EVENT,
};

// C L A S S E S /////////////////////////////////////////////////////////////


class Asura_Dialogue_Storage
{
	friend class Asura_Chunk_Dialogue_InGame;
	friend class Asura_Dialogue_Event;
	friend class Asura_Dialogue_Line;
	friend class Asura_Dialogue_Line_Template;
	friend class Asura_Dialogue_Handler;

public:

	static void Initialise();
	static void Reset();
	static void Shutdown();

	//resets the state of the dialogue events (last played item)
	static void ResetInGameState();

	static const Asura_Dialogue_Event* FindEvent(Asura_Hash_ID uHashID);
	static const Asura_Dialogue_Line* FindLine(Asura_Hash_ID uHashID);
	static const Asura_Dialogue_Line_Template* FindLineTemplate(Asura_Hash_ID uHashID);

	//Allocates a new event/line/line template and adds it to the current resource bank.
	//If it already existed , then a reference will be added to the current resource bank
	//and the existing event/line/line template will be returned	
	static Asura_Dialogue_Event* CreateEvent(const char* szName, bool& bAlreadyExists);
	static Asura_Dialogue_Line* CreateLine(const char* szName, bool& bAlreadyExists);
	static Asura_Dialogue_Line_Template* CreateLineTemplate(const char* szName, bool& bAlreadyExists);

	
#ifdef ASURA_TOOLS
	//Tools accessors for the lines and events.
	//Note that new lines/events aren't added to the end of the 'indexed' list , so
	//it isn't safe to store an index value for any length of time.
	//RAR 04/07/2006
	static u_int GetNumberOfEvents();
	static Asura_Dialogue_Event* GetEventByHash(Asura_Hash_ID uHashID);
	static Asura_Dialogue_Event* GetEventByIndex(u_int uIndex);

	static u_int GetNumberOfLines();
	static Asura_Dialogue_Line* GetLineByHash(Asura_Hash_ID uHashID);
	static Asura_Dialogue_Line* GetLineByIndex(u_int uIndex);

	static u_int GetNumberOfLineTemplates();
	static Asura_Dialogue_Line_Template* GetLineTemplateByHash(Asura_Hash_ID uHashID);
	static Asura_Dialogue_Line_Template* GetLineTemplateByIndex(u_int uIndex);


	static bool RenameEvent(const char* szOldName,const char* szNewName);
	static bool RenameLineTemplate(const char* szOldName,const char* szNewName);

	static void DeleteEvent(Asura_Dialogue_Event* pxEvent);
	static void DeleteLineTemplate(Asura_Dialogue_Line_Template* pxTemplate);

	//Clears out all the dialogue lines to make way for them being reloaded from the excel/unicode files. RAR 06/11/2006
	static void ResetLines();
#endif




private:


	static void ResetInGameStateCallback(Asura_Hash_ID uHash,Asura_Dialogue_Event* pxEvent);


	static Asura_ResourceSet<Asura_Dialogue_Event> s_xEventContainer;
	static Asura_ResourceSet<Asura_Dialogue_Line> s_xLineContainer;
	static Asura_ResourceSet<Asura_Dialogue_Line_Template> s_xLineTemplateContainer;
	

};

class Asura_Dialogue_Playable
{
	friend class Asura_Chunk_Dialogue_Event;

public:
	Asura_Dialogue_Playable(const char* szName);
	virtual ~Asura_Dialogue_Playable() {}

	virtual bool IsLine() const = 0;
	bool IsEvent() const { return !IsLine(); }

#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA	// tools-only data (also in debug)
	const char* GetName() const { return m_xName.GetString(); }
#else
	const char* GetName() const { return "<Unknown>"; }
#endif
	Asura_Hash_ID GetHashID() const { return m_uHashID; }


protected:
	// protected data
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_String	m_xName;	// must be unique, converted to hashID
#endif
	Asura_Hash_ID	m_uHashID;

private:
	// non-copyable
	Asura_Dialogue_Playable( const Asura_Dialogue_Playable& );
	Asura_Dialogue_Playable& operator=( const Asura_Dialogue_Playable& );
};

//playable ID is used to identify the lines/events during loading and saving. 
class Asura_Dialogue_PlayableID
{
public :
	ASURA_DIALOGUE_PLAYABLE_TYPE m_eType;
	Asura_Hash_ID m_uHashID; 

	inline bool operator == (const Asura_Dialogue_PlayableID& xID) const
	{
		return m_eType == xID.m_eType && m_uHashID == xID.m_uHashID;
	}	
	inline bool operator != (const Asura_Dialogue_PlayableID& xID) const
	{
		return !(*this == xID);
	}
};

//For the tools playable list to avoid having a second array for each of ID, weight and delay
class Asura_Dialogue_PlayableIDListItem
{
public :
	Asura_Dialogue_PlayableID m_xID;
	float m_fDelay;
	u_char m_ucWeight;
	Asura_Hash_ID m_uStartMessageID;//ID of message to send at start of playable
	Asura_Hash_ID m_uEndMessageID;//ID of message to send at end of playable
	Asura_String m_xName;//name is a fallback to provide more meaningful messages in tools if the event/line can't be found from the ID

	inline bool operator == (const Asura_Dialogue_PlayableIDListItem& xItem) const
	{
		return m_xID == xItem.m_xID && 
				m_fDelay == xItem.m_fDelay && 
				m_ucWeight == xItem.m_ucWeight &&
				m_uEndMessageID == xItem.m_uEndMessageID &&
				m_uStartMessageID == xItem.m_uStartMessageID;
	}	
	inline bool operator != (const Asura_Dialogue_PlayableIDListItem& xItem) const
	{
		return !(*this == xItem);
	}

	void SetDefaults()
	{
		m_xID.m_eType = ASURA_DIALOGUE_PLAYABLE_LINE;
		m_xID.m_uHashID = 0;
		m_fDelay = 0.0f;
		m_ucWeight = 1;
		m_uStartMessageID = 0;
		m_uEndMessageID = 0;
		m_xName.MakeNull();
	}

};

class Asura_Dialogue_Event : public Asura_Dialogue_Playable
{
	friend class Asura_Dialogue_Storage;//for renaming events and for checking for permanent sounds
	friend class Asura_Chunk_Dialogue_Event;
	friend class Asura_Chunk_Dialogue_InGame;
	
private :
	Asura_Dialogue_Event(const char* szName);//allocated by Asura_Dialogue_Event::CreateEvent()
public:
	~Asura_Dialogue_Event();

	virtual bool IsLine() const { return false; }

	// TODO: Function to get list of HashID of all Characters used by dialogue lines?
	u_int GetNumberOfReferencedLines() const;

	// GetPlayableIndex() returns an index of an entry from the list as appropriate and sets m_uLastPlayedItem.
	// Returns -1 if none are available.
	int GetPlayableIndex() const;

	// this overloaded version of GetPlayable() is simply an accessor and 
	// allows sequential playback.
	const Asura_Dialogue_Playable* GetPlayable(u_int uIndex) const;
	u_int GetNumPlayables() const { return m_xPlayableList.Size(); }

	//triggers messages from the specified message block
	void TriggerMessages(Asura_Hash_ID uMessageID,Asura_Guid uSpeakerGuid,u_short usMessageBlockFromInstance,Asura_Hash_ID uStaticMessageHandler) const;

	void ResetInGameState() 
	{
		m_uLastPlayedItem = (u_int)-1;
		m_fLastPlayTime = 0.0f;
	}

	//Returns false if the event fails its die roll for determining whether to play.
	bool CheckPlayProbability() const {return (m_ucPlayProbability==100) ? true : (Asura_Random::GetInt(1,100) <= m_ucPlayProbability);}

	// accessors
	ASURA_DIALOGUE_PLAYBACK_TYPE GetPlaybackType() const { return m_ePlaybackType; }
	float GetDelay(u_int uIndex) const { return m_xDelayList[uIndex]; }
	u_char GetWeight(u_int uIndex) const { return m_xWeightsList[uIndex]; }
	Asura_Hash_ID GetStartMessageID(u_int uIndex) const {return m_xStartMessageList[uIndex];}
	Asura_Hash_ID GetEndMessageID(u_int uIndex) const {return m_xEndMessageList[uIndex];}
	float GetReplayDelayTime() const { return m_fReplayDelayTime; }
	u_char GetPlayProbability() const { return m_ucPlayProbability; }
	char GetPriority() const { return m_cPriority; }
	bool GetIsGlobal() const { return m_bGlobalEvent; }
	const Asura_Dialogue_Message* GetMessageFromID(Asura_Hash_ID uHashID) const;
	Asura_Dialogue_Message* GetMessageFromID(Asura_Hash_ID uHashID);
	Asura_Hash_ID GetInitialMessageID() const {return m_uInitialMessage;}
	Asura_Hash_ID GetFinalMessageID() const {return m_uFinalMessage;}

	u_int GetMessageStoreSize() const {return m_xMessageStore.Size();}
	Asura_Dialogue_Message* GetMessageFromStore(u_int uIndex) { return m_xMessageStore[uIndex];}
	const Asura_Dialogue_Message* GetMessageFromStore(u_int uIndex) const { return m_xMessageStore[uIndex];}
	const Asura_Dialogue_Message* GetMessageFromStoreByHash(const Asura_Hash_ID uHash) const;

	void UpdateLastPlayedTime () const { m_fLastPlayTime = Asura_Timers::GetLocalScaledTime(); }
	bool CheckReplayDelayTime () const { return ( (Asura_Timers::GetLocalScaledTime() - m_fLastPlayTime) > m_fReplayDelayTime); }

#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	const char* GetFilename() const {return m_xFilename.GetString();}
#endif 

#ifdef ASURA_TOOLS
	void SetPlaybackType( ASURA_DIALOGUE_PLAYBACK_TYPE eType );
	void SetPlayableList( const Asura_Dialogue_Playable* const* apxPlayables, u_int uNumPlayables );
	void SetDelay( u_int uItem, float fDelay );
	void SetWeight( u_int uItem, u_char ucWeight );
	void SetReplayDelayTime( float fReplayDelayTime ) { m_fReplayDelayTime = fReplayDelayTime; }
	void SetPlayProbability( u_char ucProbability ) { m_ucPlayProbability = ucProbability; }
	void SetPriority( char cPriority ) { m_cPriority = cPriority; }
	void SetIsGlobal( bool bGlobal ) { m_bGlobalEvent = bGlobal; }
	void SetFilename(const char* szFilename) {m_xFilename = szFilename; }
	void SetInitialMessageID(Asura_Hash_ID uID)  {m_uInitialMessage = uID;}
	void SetFinalMessageID(Asura_Hash_ID uID)  {m_uFinalMessage = uID;}

	void SetMessageStore(const Asura_Dialogue_Message* const * ppxMessages,u_int uNumMessages);
	// TODO: messsages accessors

	u_int GetNumPlayableID() const { return m_xPlayableIDList.Size(); }
	const Asura_Dialogue_PlayableIDListItem&  GetPlayableID(u_int uIndex) const;
	Asura_Dialogue_PlayableIDListItem* GetPlayableIDPointer(u_int uIndex);
	
	//Updates list of playables , returns true if there was any change
	bool SetPlayableIDList( const Asura_Dialogue_PlayableIDListItem* pxPlayables, u_int uNumPlayables);
	//Removes playable if the event uses it. Returns true if anything was removed.
	bool RemovePlayableID(const Asura_Dialogue_PlayableID& xID);
	//Replaces/renames playable if the event uses it. Returns true if any were replaced
	bool ReplacePlayableID(const Asura_Dialogue_PlayableID& xOldID,const Asura_Dialogue_PlayableID& xNewID);

	//updates the list of playable pointers from the tools list of playable IDs
	void UpdatePlayableListFromIDList();

	void RemoveUnusedMessageBlocks ();

#endif // ASURA_TOOLS

private:
	
	// properties
	ASURA_DIALOGUE_PLAYBACK_TYPE					m_ePlaybackType;
	Asura_HeapArray<const Asura_Dialogue_Playable*>	m_xPlayableList;
	Asura_HeapArray<float>							m_xDelayList;	// for "all" playback type
	Asura_HeapArray<u_char> 						m_xWeightsList;	// for "random" playback type (0-100)
	Asura_HeapArray<Asura_Hash_ID>					m_xStartMessageList;// sent at start of line's playback
	Asura_HeapArray<Asura_Hash_ID>					m_xEndMessageList;// sent at end of line's playback
	Asura_Hash_ID									m_uInitialMessage;	// sent before first item playback 
	Asura_Hash_ID									m_uFinalMessage;	// sent at end of dialogue event 

	Asura_HeapArray<Asura_Dialogue_Message*>		m_xMessageStore;//Store of all message blocks used by this event.
	
	float	m_fReplayDelayTime; // minimum amount of time that must have passed before the dialogue event will accept any more play requests. 
	u_char	m_ucPlayProbability;	// the probability that this event will play at all when triggered. (0-100, defaults to 100).
	char	m_cPriority; // 0 = min priority, 200 = max priority, default 100
	bool	m_bGlobalEvent;

	mutable float	m_fLastPlayTime;	//When was this event last played

	// runtime playback data shared between all event references
	mutable u_int						m_uLastPlayedItem;	// initialised to m_uNumPlayables. Used for sequential and one-by-one playback
	mutable class Asura_PrimeSearch*	m_pxPrimeSearch;	// for ASURA_DPT_CDPLAYERRANDOM

	// tools data
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_HeapArray<Asura_Dialogue_PlayableIDListItem> m_xPlayableIDList;
	Asura_String m_xFilename;	// where it is saved to (Asura_String) 
#endif

};

//Used to store each language's text while in tools
struct Asura_Dialogue_Text
{
	int m_iLanguage;//matches enum in Asura_LocalisedText
	Asura_UnicodeChar* m_wszText;
};

class Asura_Dialogue_Line : public Asura_Dialogue_Playable
{
	friend class Asura_Chunk_Dialogue_Line;
	friend class Asura_Dialogue_Storage;

private :
	Asura_Dialogue_Line(const char* szName);//Allocated by Asura_Dialogue_Storage::CreateLine()
public:
	~Asura_Dialogue_Line();

	virtual bool IsLine() const { return true; }
	const Asura_Dialogue_Line_Template* GetTemplate() const { return m_pxTemplate; }
	Asura_Hash_ID GetCharacter() const { return m_xCharacterID; }
	const Asura_UnicodeChar* GetText() const { return m_wszText; }
	Asura_Hash_ID GetSoundHash() const { return m_uSoundHash; }
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	Asura_Hash_ID GetFaceAnimationHash(void) const	{ return(m_uFaceAnimationHash); }
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	u_int GetAnimation() const {return m_uAnimation;}
	float GetSoundLength() const {return m_fSoundLength;}

#ifndef ASURA_USE_AUDIO_SYSTEM
	//Gets the sound ID using the hash value , searching in the appropriate handler for streaming builds.
	int GetSoundID() const;
#endif

	void SetTemplate( const Asura_Dialogue_Line_Template* pxTemplate ) { m_pxTemplate = pxTemplate; }

#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	const char* GetCharacterName() const {return m_xCharacterName.GetString();}
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	bool		HasFaceType(void) const		{ return(!m_xFaceTypeName.IsBlank()); }
	const char*	GetFaceTypeName(void) const	{ return(m_xFaceTypeName.GetString()); }

	bool		HasBaseFacePose(void) const			{ return(!m_xBaseFacePoseName.IsBlank()); }
	const char*	GetBaseFacePoseName(void ) const	{ return(m_xBaseFacePoseName.GetString()); }
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	const char* GetTemplateName() const {return m_xTemplateName.GetString();}
#endif

#ifdef ASURA_USE_ARTICULATION_SYSTEM
	bool HasArticulation() const { return m_pxArticulation == NULL? false : true; }
	Asura_LipSync_Articulation* GetArticulation() const { return m_pxArticulation; }
	int StartArticulation(const Asura_Handle iSoundHandle, const Asura_Hash_ID uSpeakerGuid) const { return m_pxArticulation->Start( iSoundHandle, uSpeakerGuid ); }
	bool StopArticulation( const int iIndex ) const { return m_pxArticulation->Stop( iIndex ); }
	void SetArticulation( Asura_LipSync_Articulation* const pxArt ) {	m_pxArticulation = pxArt; }
#endif

#ifdef ASURA_TOOLS
	void SetCharacter(const char* szCharacter);
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	void SetFaceType(const char* szFaceType)		{ m_xFaceTypeName = szFaceType; }
	void SetBaseFacePose(const char* szFacePose)	{ m_xBaseFacePoseName = szFacePose; }
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	void SetText(const Asura_Dialogue_Text* pxText,u_int uNumLanguages);
	void SetTemplate(const char* szTemplateName);
//	void SetSoundHash(Asura_Hash_ID uHash) {m_uSoundHash = uHash;}
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	void SetFaceAnimationHash(const Asura_Hash_ID uHash)	{ m_uFaceAnimationHash = uHash; }
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	void SetAnimation(u_int uAnimation) {m_uAnimation = uAnimation;}
	void SetSoundLength(float fSoundLength) {m_fSoundLength = fSoundLength;}
	void SetSoundFilename(const char* szFilename);
	const char* GetSoundFilename() const {return m_xSoundFilename.GetString();}//note that there will most likely be a filename even if the wav file doesn't exist

	void SetSourceDialogueFilename( const char* szFilename );
	const char* const GetSourceDialogueFilename() const { return m_xSourceDialogueFilename.GetString(); }

	//FindTemplate() updates the template pointer from the current template name
	void FindTemplate();

	//Gets the text for the specified language , or 0 if not available
	const Asura_UnicodeChar* GetTextForLanguage(int iLanguage) const;
#endif



private:
	// properties
	Asura_Hash_ID m_xCharacterID;
	const Asura_UnicodeChar* m_wszText;	// current language only
	const Asura_Dialogue_Line_Template* m_pxTemplate;
	Asura_Hash_ID m_uSoundHash;
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	Asura_Hash_ID	m_uFaceAnimationHash;	// The hash-ID of the face-animation used for this line of dialogue.
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	u_int m_uAnimation; //Determines what animation should be played on the speaking character. Exactly what this value represents is project dependent , but a value of zero is used for not playing an animation.
	float m_fSoundLength;//Length of original wav file in seconds. Only used for retail builds where that information may not be otherwise available.


#ifdef ASURA_USE_ARTICULATION_SYSTEM
	Asura_LipSync_Articulation* m_pxArticulation;
#endif


#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_String m_xCharacterName;
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	Asura_String m_xFaceTypeName;
	Asura_String m_xBaseFacePoseName;
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	Asura_String m_xTemplateName;
	Asura_HeapArray<Asura_Dialogue_Text> m_xLocalisedText;	// matches an array in Tools_Dialogue_System containing language names.
#endif

#ifdef ASURA_TOOLS
	Asura_String m_xSoundFilename;//The wav filename that will be used (if it exists)
	mutable bool m_bNeedToGenerateSoundID;//Indicates whether an attempt has been made to generate a sound ID for this yet.
	Asura_String m_xSourceDialogueFilename;	// which .txt file this line was loaded from - to help with output messages
#endif

};

class Asura_Dialogue_Line_Template
{
	friend class Asura_Chunk_Dialogue_Line_Template;
	friend class Asura_Dialogue_Storage;

private :
	Asura_Dialogue_Line_Template(const char* szName);//Allocated by Asura_Dialogue_Storage::CreateLineTemplate()
public:
	~Asura_Dialogue_Line_Template();

	Asura_Hash_ID m_uHashID;
	float m_fGain;
	float m_fInnerRange;
	float m_fOuterRange;
	float m_fMinGain;
	bool m_bVoiceover;
	bool m_bNoWavForPSP;
	bool m_bNoSubtitle;
	Asura_Guid m_uMixerPreset;
	Asura_Hash_ID m_uDSPChain;

	const Asura_Sound_Mixer_Channel* GetMixerChannel() const;
	#ifdef ASURA_TOOLS
	void SetMixerChannel(const Asura_Sound_Mixer_Channel* pxMixer) {m_pxMixingChannel = pxMixer;}
	#endif
private :
	mutable const Asura_Sound_Mixer_Channel* m_pxMixingChannel;	// use "Master" if NULL
	mutable Asura_Hash_ID m_uMixerHash;

public :

	bool m_bStreamed;

#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_String m_xName;
	u_short m_usSampleRateForPS2;//If zero then the global setting is used instead
	u_short m_usSampleRateForPSP;//ditto
	Asura_String m_xMixingChannelName;
	u_short m_usCompressionQualityPS3;
	u_short m_usCompressionQuality360;
#endif
 };


// Asura_Dialogue_Instance
// This used to be known as a Tannoy. NB this is different from the Sound Event System where an 
// Asura_Sound_Event_Instance is a playback instance. Asura_Dialogue_Instance is more like an Asura_Phonon.
class Asura_Dialogue_Instance
{
public:
	const Asura_Dialogue_Event* m_pxEvent;
	const Asura_Dialogue_CharacterMappingList* m_pxCharacterMappings;	// one per character in event.

#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_String m_xEventName;
#endif 
};

// Asura_Dialogue_CharacterMapping - maps a character to a specific instance via a GUID
class Asura_Dialogue_CharacterMapping
{
public:
	Asura_Hash_ID m_uCharacterHashID;
	Asura_Guid m_uCharacterGuid;
};

class Asura_Dialogue_CharacterMappingList
{
public:
	Asura_HeapArray<Asura_Dialogue_CharacterMapping> m_xMappings;
};

//stores single message block for use by dialogue event , along with an ID/name
class Asura_Dialogue_Message
{
public :
	Asura_Hash_ID m_uHashID;
	Asura_StaticMessageBlock m_xMessageBlock;

	#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_String m_xName;
	#endif
};

#endif // #ifdef ASURA_USE_DIALOGUE_SYSTEM

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_DIALOGUE_H___
