//////////////////////////////////////////////////////////////////////////////
//
// Asura_Dialogue_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     provides the playback interface and manages access to event storage
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_DIALOGUE_SYSTEM

// This file's header.
#include "Sound/Asura_Dialogue_System.h"

#include <wchar.h>

// Asura includes
#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_DebugOverlay.h"
#include "Entity/Sound/Asura_ServerEntity_DialogueController.h"
#include "Entity/Triggers/Asura_ServerEntity_DialogueTrigger.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Message.h"
#include "Rendering/Asura_Font_System.h"
#include "Sound/Asura_Dialogue.h"
#include "Sound/Asura_Dialogue_Playback.h"

// Tools includes
#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "Sound/Tools_Dialogue_System.h"
#endif

// G L O B A L S /////////////////////////////////////////////////////////////
#if defined(ASURA_TOOLS)
static const Asura_Dialogue_Event* g_pxDialogueEventDeleted;
#endif

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Dialogue_PlaybackObject* Asura_Dialogue_System::s_pxPlaybackObjectList;
Asura_Tracking_Handle_List* Asura_Dialogue_System::s_pxPlaybackObjectHandleList;
const Asura_Dialogue_Line* Asura_Dialogue_System::s_pxCurrentDialogueLine = NULL;
Asura_Guid Asura_Dialogue_System::s_uSpeakerGuid = ASURA_GUID_UNREGISTERED;
Asura_UnicodeChar Asura_Dialogue_System::s_wszSubtitleLine[ASURA_DIALOGUE_MAX_SUBTITLE_LENGTH] = (L"");
float Asura_Dialogue_System::s_fSubtitleDuration;
float Asura_Dialogue_System::s_fSubtitleTimer;
int Asura_Dialogue_System::s_iSubtitleLength;
bool Asura_Dialogue_System::s_bSplitSubtitle;
bool Asura_Dialogue_System::s_bShowSubtitles = true;

void (*Asura_Dialogue_System::s_pfnProjectRenderSubtitles)(const Asura_UnicodeChar* wszText) = 0;
void (*Asura_Dialogue_System::s_pfnProjectTriggerAnimation)(Asura_Guid uSpeaker,u_int uAnimation) = 0;
void (*Asura_Dialogue_System::s_pfnProjectPostSetCurrentDialogueLine)(const Asura_Dialogue_Line* pxLine, Asura_Guid uSpeakerGuid) = 0;
void (*Asura_Dialogue_System::s_pfnProjectPostRemoveCurrentDialogueLine)(const Asura_Dialogue_Line* pxLine) = 0;
bool (*Asura_Dialogue_System::s_pfnProjectPlayEvent)(const Asura_Dialogue_Event *pxEvent, const Asura_Dialogue_CharacterMappingList* pxMappings, Asura_Guid uSpeakerGuid) = 0;
Asura_Handle (*Asura_Dialogue_System::s_pfnProjectVoiceCreateEvent)(const Asura_Dialogue_Line& xLine, Asura_Guid uSpeakerGuid, float fVolumeScale) = 0;


Asura_Dialogue_StaticMessage_QueueEntry Asura_Dialogue_System::s_axStaticMessageQueue[ASURA_DIALOGUE_STATIC_MESSAGE_QUEUE_SIZE];
u_int Asura_Dialogue_System::s_uStaticMessageQueue_FirstEntry = 0;
u_int Asura_Dialogue_System::s_uStaticMessageQueue_NumEntries = 0;



// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

// Asura_Dialogue_System
void Asura_Dialogue_System::Initialise()
{
	Asura_Assert(!s_pxPlaybackObjectList, "already initialised");
	Asura_Assert(!s_pxPlaybackObjectHandleList, "already initialised");
	s_pxPlaybackObjectList = new Asura_Dialogue_PlaybackObject[ASURA_DIALOGUE_MAX_ACTIVE_EVENTS];
	s_pxPlaybackObjectHandleList = new Asura_Tracking_Handle_List(ASURA_DIALOGUE_MAX_ACTIVE_EVENTS);
	s_wszSubtitleLine[0] = 0;

#if defined(ASURA_TOOLS)
	//(In streaming builds the intialisation needs to wait until the handlers are set up)
	Tools_Dialogue_System::Initialise();
#endif

#if defined( ASURA_USE_DEBUGOVERLAY )
	Asura_DebugOverlay::RegisterMode( "Dialogue", ASURA_DO_DIALOGUE, DoDebugOverlayListSounds, NULL );
#endif
}

void Asura_Dialogue_System::Reset()
{
	// stop all active instances with no external handle.
	s_pxPlaybackObjectHandleList->Traverse(StopAllWithNoExternalHandleCallback);
	Asura_Assert(!s_pxPlaybackObjectHandleList->GetNumberOfActiveHandles(), "external dialogue event handles remain at Reset()");

	Asura_Dialogue_Storage::Reset();

	Asura_Assert(!s_uStaticMessageQueue_NumEntries,"Proboably shouldn't be any messages queued up when resetting dialogue system");
	s_uStaticMessageQueue_NumEntries = 0;
	s_wszSubtitleLine[0] = 0;
}

void Asura_Dialogue_System::Shutdown()
{
	// stop all active instances with no external handle.
	s_pxPlaybackObjectHandleList->Traverse(StopAllWithNoExternalHandleCallback);
	Asura_Assert(!s_pxPlaybackObjectHandleList->GetNumberOfActiveHandles(), "external dialogue event handles remain at Reset()");

	delete[] s_pxPlaybackObjectList;
	s_pxPlaybackObjectList = NULL;
	delete s_pxPlaybackObjectHandleList;
	s_pxPlaybackObjectHandleList = NULL;

#ifdef ASURA_TOOLS
	Tools_Dialogue_System::Shutdown();
#endif

	Asura_Dialogue_Storage::Shutdown();

	Asura_Assert(!s_uStaticMessageQueue_NumEntries,"Proboably shouldn't be any messages queued up when shutting down dialogue system");
	s_uStaticMessageQueue_NumEntries = 0;
	s_wszSubtitleLine[0] = 0;
}

// Update() will traverse the active event instance list checking playback status etc.
void Asura_Dialogue_System::Update()
{
	s_pxPlaybackObjectHandleList->Traverse( UpdatePlaybackObject );
	//trigger any messages queued up during the update
	TriggerQueuedMessages();
}

// Checks to see if the provided guid is currently playing dialogue
bool Asura_Dialogue_System::IsGuidCurrentlyPlayingDialogue( Asura_Guid uGuid )
{
	if ( GetNumActiveHandles() == 0 )
	{
		return false;
	}

	for ( u_int u = 0; u < GetNumActiveHandles(); ++u )
	{
		if ( s_pxPlaybackObjectList[u].GetSpeakerGuid() == uGuid && s_pxPlaybackObjectList[u].IsPlaying() )
		{
			return true;
		}
	}

	return false;
}

void Asura_Dialogue_System::StopByGuid( Asura_Guid uGuid )
{
	for ( int i = 0; i < (int)GetNumActiveHandles(); ++i )
	{
		if ( s_pxPlaybackObjectList[i].GetSpeakerGuid() == uGuid )
		{
			Stop( i );
		}
	}
}


const Asura_Dialogue_Event* Asura_Dialogue_System::FindEvent(Asura_Hash_ID uHashID)
{
	const Asura_Dialogue_Event* pxEvent = Asura_Dialogue_Storage::FindEvent(uHashID);
	Asura_Assertf( pxEvent, ("Dialogue event not found - Hash 0x%08X", uHashID));
	return pxEvent;
}

// PlayEvent()
// if pointer to Asura_Handle is passed in, the external handle is returned and must be 
// released when the calling code is finised with it (by calling Stop() or FinishedWithHandle()
// The version with a GUID parameter is meant for code-driven one-shot AI speech and the other
// version has the mappings structure to cope with more complex conversations.
void Asura_Dialogue_System::PlayEvent( const Asura_Dialogue_Event* pxEvent, 
									  Asura_Guid uGuid, 
									  Asura_Handle* piHandleOut )
{

	PlayEvent( pxEvent, 0, uGuid, NULL, piHandleOut,ASURA_STATICMESSAGE_EMPTYBLOCK,ASURA_GUID_UNREGISTERED, 0 );
}

void Asura_Dialogue_System::PlayEvent( const Asura_Dialogue_Event* pxEvent, 
									  const Asura_Dialogue_CharacterMappingList* pxMappings,
									  Asura_Handle* piHandleOut )
{
	PlayEvent(pxEvent,pxMappings,ASURA_GUID_UNREGISTERED, NULL,piHandleOut,ASURA_STATICMESSAGE_EMPTYBLOCK,ASURA_GUID_UNREGISTERED, 0);
}

void Asura_Dialogue_System::PlayEvent( const Asura_Dialogue_Event* pxEvent, 
									  const Asura_Vector_3& xPosition, 
									  Asura_Guid uGuid,
									  Asura_Handle* piHandleOut )
{

	PlayEvent( pxEvent, 0, uGuid, &xPosition,piHandleOut,ASURA_STATICMESSAGE_EMPTYBLOCK,ASURA_GUID_UNREGISTERED, 0 );
}


void Asura_Dialogue_System::PlayEvent( const Asura_Dialogue_Event* pxEvent, 
									  const Asura_Dialogue_CharacterMappingList* pxMappings,
									  Asura_Guid uSpeakerGuid,
									  const Asura_Vector_3* pxPosition,
									  Asura_Handle* piHandleOut,
									  unsigned short usMessageBlock,
									  Asura_Guid uInstanceGuid,
									  Asura_Hash_ID uInstanceHandlerID )
{
	// parameter checking
	if (piHandleOut)
	{
		Asura_Assert(*piHandleOut == ASURA_INVALID_HANDLE, "handle is already set, possible leak");
	}

	Asura_Assert(pxEvent, "no event specified");
	if (!pxEvent || !pxEvent->CheckReplayDelayTime() || !pxEvent->CheckPlayProbability())
	{
		if (piHandleOut)
		{
			*piHandleOut = ASURA_INVALID_HANDLE;
		}
		SendNotifyFinishedMessageIfRequired(uInstanceGuid, 0);	
		return;
	}

	if(!pxEvent->GetNumPlayables())
	{
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA	// tools-only data (also in debug)
		Asura_Assertf( false, ( "No entries in event '%s'", pxEvent->GetName() ) );
#else // #ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA	// tools-only data (also in debug)
		Asura_Assert( false, "No entries in event" );
#endif // #ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA	// tools-only data (also in debug)

		if (piHandleOut)
		{
			*piHandleOut = ASURA_INVALID_HANDLE;
		}
		SendNotifyFinishedMessageIfRequired(uInstanceGuid, 0);	
		return;
	}

	if(s_pfnProjectPlayEvent)
	{
		if(!s_pfnProjectPlayEvent(pxEvent, pxMappings, uSpeakerGuid))
		{
			*piHandleOut = ASURA_INVALID_HANDLE;
			SendNotifyFinishedMessageIfRequired(uInstanceGuid, 0);	
			return;
		}
	}

	// allocate an event instance
	Asura_Handle iHandle;
	bool bOK = s_pxPlaybackObjectHandleList->GetNewHandle(iHandle);
	if (piHandleOut)
	{
		*piHandleOut = iHandle;
	}
	if (!bOK)
	{
		Asura_Assert(false, "ran out of dialogue event instances to allocate");
		SendNotifyFinishedMessageIfRequired(uInstanceGuid, 0);
		return;
	}

	// deal with playback
	// initialise the playback object and play the Dialogue
	s_pxPlaybackObjectList[iHandle].Play(pxEvent, pxMappings, uSpeakerGuid, pxPosition, usMessageBlock, uInstanceGuid, uInstanceHandlerID);
	s_pxPlaybackObjectList[iHandle].SetHasExternalHandle(piHandleOut != NULL);

	pxEvent->UpdateLastPlayedTime();

	//Trigger any messages from starting dialogue events
	TriggerQueuedMessages();
}

void Asura_Dialogue_System::PlayLine( const Asura_Dialogue_Line* pxLine, 
									  Asura_Guid uSpeakerGuid,
									  Asura_Handle* piHandleOut,
									  Asura_Guid uInstance /*= ASURA_GUID_UNREGISTERED*/,
									  int iNotifyFinishedMessageData /*=0*/ )
{
	// parameter checking
	if (piHandleOut)
	{
		Asura_Assert(*piHandleOut == ASURA_INVALID_HANDLE, "handle is already set, possible leak");
	}

	Asura_Assert(pxLine, "no line specified");
	if (!pxLine)
	{
		if (piHandleOut)
		{
			*piHandleOut = ASURA_INVALID_HANDLE;
		}
		SendNotifyFinishedMessageIfRequired(uInstance, iNotifyFinishedMessageData);
		return;
	}

	// allocate an event instance
	Asura_Handle iHandle;
	bool bOK = s_pxPlaybackObjectHandleList->GetNewHandle(iHandle);
	if (piHandleOut)
	{
		*piHandleOut = iHandle;
	}
	if (!bOK)
	{
		Asura_Assert(false, "ran out of dialogue event instances to allocate");
		SendNotifyFinishedMessageIfRequired(uInstance, iNotifyFinishedMessageData);
		return;
	}

	// deal with playback
	// initialise the playback object and play the Dialogue
	s_pxPlaybackObjectList[iHandle].Play(pxLine, uSpeakerGuid, uInstance, iNotifyFinishedMessageData);
	s_pxPlaybackObjectList[iHandle].SetHasExternalHandle(piHandleOut != NULL);

	//Trigger any messages from starting dialogue lines
	//(actually , I don't think this is really necessary , but no real harm in doing so to be sure)
	TriggerQueuedMessages();
}



// Stop() destroys the dialogue playback object and releases the handle, so it uses the reference
// to set the handle to be invalid automatically.
void Asura_Dialogue_System::Stop( Asura_Handle& iHandleRef )
{
	// allow invalid handle with no asserts for easier cleaning up etc.
	if (iHandleRef == ASURA_INVALID_HANDLE)
    {
        return;
    }

	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandleRef), "unexpected handle");
	s_pxPlaybackObjectList[iHandleRef].Stop();
	s_pxPlaybackObjectList[iHandleRef].SetHasExternalHandle(false);
	s_pxPlaybackObjectHandleList->ReleaseHandle(iHandleRef);
	iHandleRef = ASURA_INVALID_HANDLE;
}

// FinishedWithHandle() allows the dialogue system code to free up the event handle when the 
// event has finished playing. The reference is used to set the handle to be invalid automatically.
void Asura_Dialogue_System::FinishedWithHandle( Asura_Handle& iHandleRef )
{
	// allow invalid handle with no asserts for easier cleaning up etc.
	if (iHandleRef == ASURA_INVALID_HANDLE)
    {
        return;
    }

	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandleRef), "unexpected handle");
	s_pxPlaybackObjectList[iHandleRef].SetHasExternalHandle(false);
}

// IsPlaying() returns whether the given event instance is playing
bool Asura_Dialogue_System::IsPlaying( Asura_Handle iHandle )
{
	// allow invalid handle with no asserts for easier cleaning up etc.
	if (iHandle == ASURA_INVALID_HANDLE)
    {
        return false;
    }

	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	return s_pxPlaybackObjectList[iHandle].IsPlaying();
}

// GetNumberPlaying() returns the number of active dialogue handles - PJC 11/04/07
u_int Asura_Dialogue_System::GetNumActiveHandles()
{
	return (s_pxPlaybackObjectHandleList) ? static_cast<u_int> (s_pxPlaybackObjectHandleList->GetNumberOfActiveHandles()) : 0;
}

// JAT 08/02/07 - set & get VolumeScale
void Asura_Dialogue_System::SetVolumeScale( Asura_Handle iHandle, const float fVolumeScale )
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	s_pxPlaybackObjectList[iHandle].SetVolumeScale( fVolumeScale );
}

float Asura_Dialogue_System::GetVolumeScale( Asura_Handle iHandle )
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	return s_pxPlaybackObjectList[iHandle].GetVolumeScale();
}

void Asura_Dialogue_System::UpdatePosition( Asura_Handle iHandle, const Asura_Vector_3& xPos)
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	s_pxPlaybackObjectList[iHandle].SetPositionOverride( xPos);
}

void Asura_Dialogue_System::UpdatePlaybackObject(Asura_Handle iHandle)
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Dialogue_PlaybackObject& xPlaybackObject = s_pxPlaybackObjectList[iHandle];
	xPlaybackObject.Update();
	if (!xPlaybackObject.IsPlaying() && xPlaybackObject.NoExternalHandle())
	{
		xPlaybackObject.Stop();
		s_pxPlaybackObjectHandleList->ReleaseHandle(iHandle);
	}
}

void Asura_Dialogue_System::StopAllWithNoExternalHandle()
{     
	s_pxPlaybackObjectHandleList->Traverse(StopAllWithNoExternalHandleCallback);
}

void Asura_Dialogue_System::StopAllWithNoExternalHandleCallback(Asura_Handle iHandle)
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Dialogue_PlaybackObject& xPlaybackObject = s_pxPlaybackObjectList[iHandle];
	if (xPlaybackObject.NoExternalHandle())
	{
		Stop(iHandle);
	}
}

void Asura_Dialogue_System::StopAll()
{
	s_pxPlaybackObjectHandleList->Traverse(StopAllCallback);

	// Remove the current subtitle if there is one.
	s_fSubtitleDuration = 0.0f;
	s_fSubtitleTimer = 0.0f;
	s_wszSubtitleLine[0] = '\0';
}

void Asura_Dialogue_System::StopAllCallback(Asura_Handle iHandle)
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Dialogue_PlaybackObject& xPlaybackObject = s_pxPlaybackObjectList[iHandle];
	xPlaybackObject.Stop();
}


void Asura_Dialogue_System::SkipLine()
{
	s_pxPlaybackObjectHandleList->Traverse(SkipLineCallback);
}

void Asura_Dialogue_System::SkipLineCallback(Asura_Handle iHandle)
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Dialogue_PlaybackObject& xPlaybackObject = s_pxPlaybackObjectList[iHandle];
	xPlaybackObject.SkipLine();
}


#if defined(ASURA_TOOLS)
void Asura_Dialogue_System::DialogueEventDeleted(const Asura_Dialogue_Event* pxEvent)
{
	g_pxDialogueEventDeleted = pxEvent;
	s_pxPlaybackObjectHandleList->Traverse(DialogueEventDeletedCallback);
	g_pxDialogueEventDeleted = NULL;
}

void Asura_Dialogue_System::DialogueEventDeletedCallback(Asura_Handle iHandle)
{
	Asura_Assert(s_pxPlaybackObjectHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Dialogue_PlaybackObject& xPlaybackObject = s_pxPlaybackObjectList[iHandle];
	xPlaybackObject.DialogueEventDeleted(g_pxDialogueEventDeleted);
}
#endif

void Asura_Dialogue_System::HandleMessage(const Asura_Message* pxMessage)
{
	switch(pxMessage->GetMsgID())
	{
		case ASURA_MSG_DIALOGUE_TRIGGER :
		{
			const Asura_DialogueTrigger_MessageHeader* pxData = reinterpret_cast
				<const Asura_DialogueTrigger_MessageHeader*>(pxMessage->GetData());

			const Asura_Dialogue_CharacterMapping* pxMapping = reinterpret_cast<const Asura_Dialogue_CharacterMapping*>(pxData + 1);

			//find the event to be triggered
#if defined(ASURA_USE_STREAMING)
			//Need to get the event from the correct handler
			Asura_Hash_ID uHandlerID = pxData->m_uHandlerID;
			Asura_Dialogue_Handler* pxHandler = Asura_Dialogue_Storage::GetHandler(pxData->m_uHandlerID);
			if(!pxHandler)
			{
				Asura_Assert(0,"Failed to get streaming dialogue handler for dialogue trigger");
				return;
			}
			const Asura_Dialogue_Event* pxEvent = pxHandler->FindEvent(pxData->m_uEventID);
#else
			Asura_Hash_ID uHandlerID = 0;
			const Asura_Dialogue_Event* pxEvent = Asura_Dialogue_Storage::FindEvent(pxData->m_uEventID);
#endif	// defined(ASURA_USE_STREAMING)
			if(!pxEvent)
			{
				return;
			}

			//fill in a temporary character mapping to pass into the play function
			Asura_Dialogue_CharacterMappingList xList;
			xList.m_xMappings.SetSize(pxData->m_uNumMappings);
			for(u_int i=0;i<pxData->m_uNumMappings;i++)
			{
				xList.m_xMappings[i] = pxMapping[i];
			}

			//play the event
			PlayEvent(pxEvent,&xList,ASURA_GUID_UNREGISTERED, NULL,0,pxData->m_usMessageBlock,pxMessage->GetFromGuid(),uHandlerID);
			break;
		}

		case ASURA_MSG_DIALOGUE_PLAYLINE :
		case ASURA_MSG_DIALOGUE_PLAYLINEONENTITY :
		{
			const Asura_Message_Dialogue_PlayLine* pxData = reinterpret_cast
				<const Asura_Message_Dialogue_PlayLine*>(pxMessage->GetData());

			//find the line to be triggered
			const Asura_Dialogue_Line* pxLine = Asura_Dialogue_Storage::FindLine(pxData->m_uLineID);
			
			if(!pxLine)
			{
				return;
			}
			//If using the OnEntity variation , then the line is played using the sender's guid
			Asura_Guid uGuid = (pxMessage->GetMsgID() == ASURA_MSG_DIALOGUE_PLAYLINEONENTITY) ? pxMessage->GetFromGuid() : ASURA_GUID_UNREGISTERED;

			//play the line
			PlayLine(pxLine,uGuid,NULL);
			break;
			
		}

		case ASURA_MSG_DIALOGUE_PLAYLINE_CONTROLLER :
		{
			const Asura_Msg_Dialogue_PlayLine_Controller* pxData = reinterpret_cast
				<const Asura_Msg_Dialogue_PlayLine_Controller*>(pxMessage->GetData());

			//find the line to be triggered
			const Asura_Dialogue_Line* pxLine = Asura_Dialogue_Storage::FindLine(pxData->m_uLineID);

			if(!pxLine)
			{
				return;
			}
			//play the line
			PlayLine( pxLine, pxData->m_uSpeaker, NULL, pxMessage->From, pxData->m_iNotifyFinishedMessageData );
		}
		break;

		case ASURA_MSG_DIALOGUE_STOPLINE_CONTROLLER:
		{
			const Asura_Msg_Dialogue_StopLine_Controller* pxData = reinterpret_cast<const Asura_Msg_Dialogue_StopLine_Controller*>( pxMessage->GetData() );

			Asura_Tracking_Handle_List_Iterator xIter( s_pxPlaybackObjectHandleList );
			for(; !xIter.End(); xIter.Next() )
			{
				const Asura_Dialogue_Line* pxLine = s_pxPlaybackObjectList[ xIter.GetCurrentHandle() ].GetLine();
				if( pxLine && pxLine->GetHashID() == pxData->m_uLineID )
				{
					Asura_Handle xHandle = xIter.GetCurrentHandle();
					Stop( xHandle );
					break;
				}
			}
		}
		break;

		case ASURA_MSG_DIALOGUE_CANCELSUBTITLE :
			CancelSubtitle();
			break;

		default :
			Asura_Assert(0,"Unexpected message sent to Asura_Dialogue_System::HandleMessage()");
			break;

	}
}

void Asura_Dialogue_System::SetCurrentDialogueLine(const Asura_Dialogue_Line* pxLine,Asura_Guid uSpeakerGuid, float fDuration)
{
	s_pxCurrentDialogueLine = pxLine;
	s_uSpeakerGuid = uSpeakerGuid;
	s_fSubtitleDuration = fDuration;
	s_fSubtitleTimer = 0.0f;
	s_iSubtitleLength = 0;
	s_bSplitSubtitle = false;

	if(s_fSubtitleDuration < ASURA_DIALOGUE_MIN_SUBTITLE_DURATION)
	{	s_fSubtitleDuration = ASURA_DIALOGUE_MIN_SUBTITLE_DURATION;}

	if(s_pxCurrentDialogueLine)
	{
		const Asura_UnicodeChar* wszText = s_pxCurrentDialogueLine->GetText();
		if(wszText)
		{
			s_iSubtitleLength = wcslen(wszText);
			//if the subtitle contains any newlines , it will be split when displayed
			s_bSplitSubtitle = wcschr(wszText,L'\n')!=0;

			wcsncpy(s_wszSubtitleLine, wszText, ASURA_DIALOGUE_MAX_SUBTITLE_LENGTH-1);
		}
	}

	if(s_pfnProjectPostSetCurrentDialogueLine)
	{
		// perform project specific processing when current dialogue line is set
		s_pfnProjectPostSetCurrentDialogueLine(pxLine, uSpeakerGuid);
	}
}

void Asura_Dialogue_System::RemoveCurrentDialogueLine(const Asura_Dialogue_Line* pxLine)
{
	if(pxLine == s_pxCurrentDialogueLine)
	{
		s_pxCurrentDialogueLine = NULL;
		s_uSpeakerGuid = ASURA_GUID_UNREGISTERED;

		if(s_pfnProjectPostRemoveCurrentDialogueLine)
		{
			// perform project specific processing when current dialogue line is removed
			s_pfnProjectPostRemoveCurrentDialogueLine(pxLine);
		}
	}
}

const Asura_UnicodeChar* Asura_Dialogue_System::GetSubtitleText()
{
	return s_pxCurrentDialogueLine ? s_pxCurrentDialogueLine->GetText() : 0;
}

Asura_Hash_ID Asura_Dialogue_System::GetSubtitleCharacterID()
{
	return s_pxCurrentDialogueLine ? s_pxCurrentDialogueLine->GetCharacter() : 0;
}


void Asura_Dialogue_System::RenderSubtitles()
{
	#ifdef ASURA_TOOLS
	if(Tools_System::IsActive())
	{
		//reset the store of the current subtitle in tools
		Tools_Dialogue_System::SetSubtitle(L"");
	}
	#endif

	
//	if(!s_pxCurrentDialogueLine || !s_iSubtitleLength)
//	{	return;}
//	const Asura_UnicodeChar* wszText = s_pxCurrentDialogueLine->GetText();
//	if(!wszText) return;

	const Asura_UnicodeChar* wszText = s_wszSubtitleLine;
	if(wcslen(wszText) == 0)
	{	return;}

	s_fSubtitleTimer += Asura_Timers::GetGameTimeSlice();
	if(s_fSubtitleTimer >= s_fSubtitleDuration)
	{
		s_wszSubtitleLine[0] = 0;
		return;
	}
	
	if(s_bSplitSubtitle)
	{
		//get the character that has been reached
		int iCurrentPos = Asura_Maths::FloatToInt((s_fSubtitleTimer*s_iSubtitleLength)/s_fSubtitleDuration);
		if(iCurrentPos >= s_iSubtitleLength) return;

		//search back for a newline before this position
		const Asura_UnicodeChar* pwcStart = &wszText[iCurrentPos];
		while(iCurrentPos > 0)
		{
			if(*pwcStart == L'\n') break;
			iCurrentPos--;
			pwcStart--;
		}

		//skip forwards to the start of a line
		while(*pwcStart == L'\n') pwcStart++;

		//copy until the end of the line into a temporary buffer
		const int iBufferSize = 256;
		Asura_UnicodeChar awcBuffer[iBufferSize];

		const Asura_UnicodeChar* pwcSource = pwcStart;
		Asura_UnicodeChar* pwcDest = awcBuffer;
		int iPos = 0;

		while(*pwcSource && *pwcSource != L'\n')
		{
			if(iPos >= iBufferSize - 1)
			{
				//not going to have enough space for the entire string
				Asura_Assert(0,"Buffer not large enough to split dialogue string.\nEither need to increase iBufferSize , or split the line into more segments.");
				break;
			}
			*pwcDest++ = *pwcSource++;
			iPos++;
		}
		*pwcDest = 0;

		//render this string
		#ifdef ASURA_TOOLS
		if(Tools_System::IsActive())
		{	Tools_Dialogue_System::SetSubtitle(awcBuffer);}
		else
		#endif
		if(s_pfnProjectRenderSubtitles)
		{	s_pfnProjectRenderSubtitles(awcBuffer);}
		else
		{	RenderSubtitles(awcBuffer);}
	}
	else
	{	
		#ifdef ASURA_TOOLS
		if(Tools_System::IsActive())
			Tools_Dialogue_System::SetSubtitle(wszText);
		else
		#endif
		if(s_pfnProjectRenderSubtitles)
		{	s_pfnProjectRenderSubtitles(wszText);}
		else
		{	RenderSubtitles(wszText);}
	}
}

void Asura_Dialogue_System::RenderSubtitles(const Asura_UnicodeChar* wszText)
{
    Asura_Bounding_Rectangle xRect;
	//put dialogue subtitles at the top
	xRect.MinX = 0.06f;
	xRect.MaxX = 0.94f;
	xRect.MinY = 0.0f;
	xRect.MaxY = 0.125f;


	Asura_Font_System::SetFont(0);

    u_int uColour = 0xFFFFFFFF;

	Asura_Font_System::Print( xRect, 0.0f, 0.0f, 0.75f, 0.9f, uColour,
							 ASURA_TEXTFORMAT_RELATIVEPOSITION|ASURA_TEXTFORMAT_CENTREJUSTIFIED|ASURA_TEXTFORMAT_VERTICALCENTREJUSTIFIED|ASURA_TEXTFORMAT_WORDWRAP,
							 wszText);
}

void Asura_Dialogue_System::TriggerAnimation(Asura_Guid uSpeakerGuid,u_int uAnimation)
{
	if(s_pfnProjectTriggerAnimation)
	{
		s_pfnProjectTriggerAnimation(uSpeakerGuid,uAnimation);
	}
	else if(uAnimation)
	{
		Asura_Assert(0,"Attempting to trigger an animation from dialogue , but no project callback has been set yet.");
	}
}

void Asura_Dialogue_System::QueueStaticMessage(const Asura_Dialogue_Event* pxEvent,Asura_Hash_ID uMessageID, Asura_Guid uSpeakerGuid, const Asura_Dialogue_PlaybackObject* pxPlaybackObject)
{
	if(!uMessageID) return;
	Asura_Assert(pxEvent,"");
	Asura_Assert(pxPlaybackObject,"");

	if(s_uStaticMessageQueue_NumEntries >= ASURA_DIALOGUE_STATIC_MESSAGE_QUEUE_SIZE)
	{
		Asura_Assert(0,"Too many static messages queued up in dialogue system");
		return;
	}

	u_int uSlot = (s_uStaticMessageQueue_FirstEntry+s_uStaticMessageQueue_NumEntries) % ASURA_DIALOGUE_STATIC_MESSAGE_QUEUE_SIZE;
	s_uStaticMessageQueue_NumEntries++;

	s_axStaticMessageQueue[uSlot].m_pxEvent = pxEvent;
	s_axStaticMessageQueue[uSlot].m_uMessageID = uMessageID;
	s_axStaticMessageQueue[uSlot].m_uSpeakerGuid = uSpeakerGuid;
	s_axStaticMessageQueue[uSlot].m_usMessageBlockFromInstance = pxPlaybackObject->GetMessageBlockFromInstance();
	s_axStaticMessageQueue[uSlot].m_uStaticMessageHandler = 0;
}

void Asura_Dialogue_System::TriggerQueuedMessages()
{
	//This function may get called again , or more static messages queued up as a result of messages sent,
	//so need to make sure that the state is updated before each message is sent.
	//RAR 11/07/2007

	while(s_uStaticMessageQueue_NumEntries)
	{
		Asura_Dialogue_StaticMessage_QueueEntry* pxEntry = &s_axStaticMessageQueue[s_uStaticMessageQueue_FirstEntry];

		s_uStaticMessageQueue_FirstEntry++;
		if(s_uStaticMessageQueue_FirstEntry==ASURA_DIALOGUE_STATIC_MESSAGE_QUEUE_SIZE) s_uStaticMessageQueue_FirstEntry = 0;
		s_uStaticMessageQueue_NumEntries--;

		pxEntry->m_pxEvent->TriggerMessages(pxEntry->m_uMessageID,
											pxEntry->m_uSpeakerGuid,
											pxEntry->m_usMessageBlockFromInstance,
											pxEntry->m_uStaticMessageHandler);
	}
}


#if defined( ASURA_USE_DEBUGOVERLAY )
// helper for sorting events
static int CompareDialogueEventInstances( const void* pEventInstance1, const void* pEventInstance2 )
{
	const Asura_Dialogue_PlaybackObject* const * ppxInstance1 = reinterpret_cast<const Asura_Dialogue_PlaybackObject* const *>(pEventInstance1);
	const Asura_Dialogue_PlaybackObject* const * ppxInstance2 = reinterpret_cast<const Asura_Dialogue_PlaybackObject* const *>(pEventInstance2);
	const Asura_Dialogue_Playable* pxEvent1 = (*ppxInstance1)->GetEventOrLine();
	const Asura_Dialogue_Playable* pxEvent2 = (*ppxInstance2)->GetEventOrLine();
	const Asura_Hash_ID uHash1 = pxEvent1 ? pxEvent1->GetHashID() : ASURA_HASH_ID_UNSET;
	const Asura_Hash_ID uHash2 = pxEvent2 ? pxEvent2->GetHashID() : ASURA_HASH_ID_UNSET;
	if (uHash1 > uHash2)
	{
		return 1;
	}
	else if (uHash1 < uHash2)
	{
		return -1;
	}
	return 0;
}

void Asura_Dialogue_System::DoDebugOverlayListSounds( void* const pxIgnore )
{
	(void)pxIgnore;

	// print num active
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DIALOGUE, "%u/%u active dialogue events", 
		s_pxPlaybackObjectHandleList->GetNumberOfActiveHandles(), ASURA_DIALOGUE_MAX_ACTIVE_EVENTS );

	// build sorted array of active event instance pointers
	const Asura_Dialogue_PlaybackObject* apxSortedEvents[ASURA_DIALOGUE_MAX_ACTIVE_EVENTS];
	u_int uNumEvents = 0;

	for (u_int uCurEvent = 0; uCurEvent < ASURA_DIALOGUE_MAX_ACTIVE_EVENTS; ++uCurEvent)
	{
		const Asura_Dialogue_PlaybackObject& xEventInstance = s_pxPlaybackObjectList[uCurEvent];
		if (xEventInstance.IsPlaying() || !xEventInstance.NoExternalHandle())
		{
			apxSortedEvents[uNumEvents++] = &xEventInstance;
		}
	}
	Asura_SortAndSearch::Quicksort(apxSortedEvents, uNumEvents, CompareDialogueEventInstances);

	Asura_Hash_ID uPrevEventHash = ASURA_HASH_ID_UNSET;
	for (u_int uCurEvent = 0; uCurEvent < uNumEvents; ++uCurEvent)
	{
		const Asura_Dialogue_PlaybackObject& xEventInstance = *apxSortedEvents[uCurEvent];
		const Asura_Dialogue_Playable* pxEventOrLine = xEventInstance.GetEventOrLine();

		if ( pxEventOrLine )
		{
			const float fVolume = xEventInstance.GetVolumeScale();

			// check if we've already printed info about this event.
			Asura_Hash_ID uHashID = pxEventOrLine->GetHashID();
			if (uPrevEventHash == uHashID)
			{
				continue;
			}
			uPrevEventHash = uHashID;

			// peek at other instances of this event ahead of us.
			u_int uNumInstances = 1;
			for (u_int uTemp = uCurEvent + 1; uTemp < uNumEvents; ++uTemp)
			{
				const Asura_Dialogue_Playable* pxNextEventOrLine = (apxSortedEvents[uTemp]->GetEventOrLine());
				if (!pxNextEventOrLine || (pxNextEventOrLine->GetHashID() != uHashID))
				{
					break;
				}
				++uNumInstances;
			}

			const Asura_Char* szName = "";
#if defined( ASURA_DIALOGUE_INCLUDE_TOOLS_DATA )
			szName = pxEventOrLine->GetName();
#else
			char acNameBuffer[11];
			Asura_PrintfToBfr(acNameBuffer, "0x%x", 11, true, pxEventOrLine->GetHashID());
			szName = acNameBuffer;
#endif
			Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DIALOGUE, 
				"%u of %s (vol %.02f)", 
				uNumInstances, szName, Asura_Maths::FtoD(fVolume)
			);
		}
		else
		{
			Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DIALOGUE, "inactive event");
		}
	}
}
#endif // #if defined( ASURA_USE_DEBUGOVERLAY )


void Asura_Dialogue_System::SendNotifyFinishedMessageIfRequired(Asura_Guid uGuid, int iNotifyFinishedMessageData)
{
	//Let the dialogue trigger/controller know that the dialogue has played as much as it
	//is going to.
	//There might be something to be said for distinguishing between failed events and ones that
	//have finished properly , but for now the same behaviour is used either way.
	//RAR 14/04/2009
	if(uGuid != ASURA_GUID_UNREGISTERED && Asura_Network::IsServer())
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_NOTIFYFINISHED, sizeof(Asura_Message_Dialogue_NotifyFinished));
		pxMessage->To =	uGuid;
		pxMessage->From = ASURA_GUID_UNREGISTERED;
		
		Asura_Message_Dialogue_NotifyFinished* pxData = reinterpret_cast<Asura_Message_Dialogue_NotifyFinished*>( pxMessage->GetData() );
		pxData->m_iMessageData = iNotifyFinishedMessageData;

		Asura_Network::Send(pxMessage);
	}
}

#endif // ASURA_USE_DIALOGUE_SYSTEM
