/**
*
* \file Asura_NullSound.h
*
* Copyright (C) 2007 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_OPENAL_H__
#define __ASURA_OPENAL_H__

#include "Core/Asura.h"
#if defined( ASURA_USE_OPENAL ) && ( ASURA_USE_OPENAL == 1 ) && !defined(ASURA_USE_AUDIO_SYSTEM)
#include "OpenAL/OpenAL1-1SDK/include/al.h"
#include "OpenAL/OpenAL1-1SDK/include/alc.h"
#include "OpenAL/OpenAL1-1SDK/include/efx.h"
#include "Sound/Asura_Sound.h"

/* ------------------------------------------------------------------------- */

typedef class Asura_OpenAL_ActiveSound_API			Asura_PC_ActiveSound_API;
typedef class Asura_OpenAL_Sound_API				Asura_PC_Sound_API;
typedef struct Asura_OpenAL_ReverbParameters		Asura_PC_Sound_ReverbParams;

/* ------------------------------------------------------------------------- */

struct Asura_OpenAL_ReverbParameters
{
	ALuint		uEffectSlotPrevious;
	ALuint		uEffectSlotCurrent;
};

/* ------------------------------------------------------------------------- */

enum ASURA_OPENAL_FILTER_ID
{
	ASURA_OPENAL_FILTER_ID_DISTANCE = 0,	//distance filter
	ASURA_OPENAL_FILTER_ID_OCCLUSION,		//occlusion filter
	ASURA_OPENAL_FILTER_ID_WET,				//wet path filter, combination of distance + wet volume

	ASURA_OPENAL_FILTER_ID_NUM_ITEMS
};

/* ------------------------------------------------------------------------- */

// This class must match the implicit interface of Asura_PC_ActiveSound_API
// (defined by the calls made on it). NOTE an alternative would be to define
// an explicit interface class, but while logically clearer to understand,
// it introduces an additional vtable that isn't necessary.
class Asura_OpenAL_ActiveSound_API
{
public:

	Asura_OpenAL_ActiveSound_API();
	~Asura_OpenAL_ActiveSound_API();

	bool Initialise( bool bUseHardware3dPan, u_int uSize, const void* pvFormat, const u_char* pucData );
	void Update();
	void Destroy();

	bool IsValid() const;
	bool IsPlaying() const;
	bool IsPaused() const;

	// Basic behaviour
	bool Play( bool bIsLooping );
	bool Pause();
	bool Unpause( bool bIsLooping );
	bool Stop();

	// More advanced control
	bool SetVolume( float fVolume );
	bool SetSpatial( bool bIsSpatial );
	bool SetPosition( const Asura_Vector_3& xPosition );
	bool SetRadii( float fInner, float fOuter );
	bool SetPan( float fPan );
	bool SetForcedPanAndFade( float fPan, float fFade );
	bool SetPitch( float fPitch );
	bool SetCone( const Asura_Vector_3& xDirection, float fInnerAngle, float fOuterAngle, float fVolumeOutsideCone );
	bool SetUseReverb( bool bUseReverb );
	bool SetUseDistanceFilter( bool bUseDistancefilter );
	void SetReverbParams( const Asura_OpenAL_ReverbParameters& xReverbParams );
	bool SetOcclusionLPF( const float fOcclusion );
	bool SetWetDryVolume (const float fWetVolume, const float fDryVolume);
	void SetDistanceFilterGain (const float fGain);
	
	#if _DEBUG
		void CheckStatus() const;
	#endif

private:

	// Platform-specific implementations of OpenAL behaviour
	static bool Platform_FillBuffer( ALuint m_uBufferID, const void* pvFormat, const u_char* pucData, u_int uSize );

	ALuint			m_uSourceID;
	ALuint			m_uBufferID;
	ALuint			m_auFilterID[ASURA_OPENAL_FILTER_ID_NUM_ITEMS];

	float			m_fInnerRadius;
	float			m_fOuterRadius;
	Asura_Vector_3	m_xPosition;
	Asura_Vector_3	m_xPreviousPosition;
	bool			m_bIsSpatial;
	bool			m_bUseReverb;
	float			m_fDryVolume;
	float			m_fWetVolume;
	float			m_fOcclusionFilterGain;
	float			m_fOcclusion;
	bool			m_bNeedUpdateReverbParams;
	float			m_fDistanceFilterGain;
	float			m_fLastDistanceFilterGain;

	// Disallow copying and assignment
	Asura_OpenAL_ActiveSound_API( Asura_OpenAL_ActiveSound_API& other );
	const Asura_OpenAL_ActiveSound_API& operator=( Asura_OpenAL_ActiveSound_API& other );
};

/* ------------------------------------------------------------------------- */

// Same notes as for Asura_OpenAL_ActiveSound_API
class Asura_OpenAL_Sound_API
{
public:

	Asura_OpenAL_Sound_API();
	~Asura_OpenAL_Sound_API();

	bool Initialise();
	void Update();
	void Destroy();
	bool IsValid() const;

	static Asura_OpenAL_Sound_API& Platform_GetAPI();

	void SetListener( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation );
	void ForceListenerUpdate();

	void SetReverb( Asura_Sound_System::ReverbType eType, float fReverbAmount );

	bool IsEFXAvailable() const						{ return m_bEFXAvailable; }

	const Asura_Vector_3& GetListenerPos() const	{ return m_xListenerPos; }
	int	GetNumAuxSendsPerSource() const				{ return m_iAuxSendsPerSource; }
	const Asura_OpenAL_ReverbParameters& GetReverbParams() const { return m_xReverbParameters; }

	#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
		void ForceReverb( int iReverbID );
		int GetForcedReverb() const					{ return ( m_bManualReverbControl ? static_cast<int>( m_eForcedReverb ) : -1 ); }
	#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

private:

	// Platform-specific implementations of OpenAL behaviour
	// JAT 18/10/07 - let the h/w platform specify this value
	int	Platform_GetNumAuxSendsPerSource( void );
	static void Platform_Initialise();

	bool SwitchToReverb( Asura_Sound_System::ReverbType eReverbID, float fReverbAmount );
	bool Asura_OpenAL_Sound_API::SetReverbProperties( int index, Asura_Sound_System::ReverbType eReverbID, float fReverbAmount );

	ALCdevice*			m_pxDevice;
	ALCcontext*			m_pxContext;
	Asura_Vector_3		m_xListenerPos;
	Asura_Vector_3		m_xPreviousListenerPos;

	// Effects extensions. NOTE we ask for ASURA_OPENAL_REQUEST_AUX_SENDS_NUM auxiliary sends,
	// but get only m_iAuxSendsPerSource (which might be less if the hardware supports less).
	bool				m_bEFXAvailable;
	int					m_iAuxSendsPerSource;

	// An effect slot could theoretically have more than one effect; but
	// for us, at the moment, it's 1:1 and all effects are reverbs
	static const int	ASURA_OPENAL_REQUEST_AUX_SENDS_NUM = 4;

	ALuint				m_uEffects[ASURA_OPENAL_REQUEST_AUX_SENDS_NUM];
	ALuint				m_uEffectSlots[ASURA_OPENAL_REQUEST_AUX_SENDS_NUM];

	static const float	ASURA_OPENAL_REVERB_CROSSFADE_TIME;
	float				m_fReverbCrossfadeTime;
	float				m_fCurrentReverbAmount;
	Asura_Sound_System::ReverbType m_eCurrentReverb;
	int					m_iEffectSlotIndexPrevious;
	int					m_iEffectSlotIndexCurrent;
	Asura_OpenAL_ReverbParameters m_xReverbParameters;

	#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
		// Debugging and development
		bool			m_bManualReverbControl;
		Asura_Sound_System::ReverbType m_eForcedReverb;
		Asura_Sound_System::ReverbType m_eOldReverb;
		float			m_fOldReverbAmount;
	#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

	//Reverb
	float				m_fReverbCrossfadeProportion;
	
	// Disallow copying and assignment
	Asura_OpenAL_Sound_API( Asura_OpenAL_Sound_API& other );
	const Asura_OpenAL_Sound_API& operator=( Asura_OpenAL_Sound_API& other );

	#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
		friend bool ConsoleCmd_AudioDiagnostics( bool bHelp );
	#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM
};

/* ------------------------------------------------------------------------- */

extern LPALGENEFFECTS					alGenEffects;
extern LPALDELETEEFFECTS				alDeleteEffects;
extern LPALISEFFECT						alIsEffect;
extern LPALEFFECTI						alEffecti;
extern LPALEFFECTIV						alEffectiv;
extern LPALEFFECTF						alEffectf;
extern LPALEFFECTFV						alEffectfv;
extern LPALGETEFFECTI					alGetEffecti;
extern LPALGETEFFECTIV					alGetEffectiv;
extern LPALGETEFFECTF					alGetEffectf;
extern LPALGETEFFECTFV					alGetEffectfv;

extern LPALGENFILTERS					alGenFilters;
extern LPALDELETEFILTERS				alDeleteFilters;
extern LPALISFILTER						alIsFilter;
extern LPALFILTERI						alFilteri;
extern LPALFILTERIV						alFilteriv;
extern LPALFILTERF						alFilterf;
extern LPALFILTERFV						alFilterfv;
extern LPALGETFILTERI					alGetFilteri;
extern LPALGETFILTERIV					alGetFilteriv;
extern LPALGETFILTERF					alGetFilterf;
extern LPALGETFILTERFV					alGetFilterfv;

extern LPALGENAUXILIARYEFFECTSLOTS		alGenAuxiliaryEffectSlots;
extern LPALDELETEAUXILIARYEFFECTSLOTS	alDeleteAuxiliaryEffectSlots;
extern LPALISAUXILIARYEFFECTSLOT		alIsAuxiliaryEffectSlot;
extern LPALAUXILIARYEFFECTSLOTI			alAuxiliaryEffectSloti;
extern LPALAUXILIARYEFFECTSLOTIV		alAuxiliaryEffectSlotiv;
extern LPALAUXILIARYEFFECTSLOTF			alAuxiliaryEffectSlotf;
extern LPALAUXILIARYEFFECTSLOTFV		alAuxiliaryEffectSlotfv;
extern LPALGETAUXILIARYEFFECTSLOTI		alGetAuxiliaryEffectSloti;
extern LPALGETAUXILIARYEFFECTSLOTIV		alGetAuxiliaryEffectSlotiv;
extern LPALGETAUXILIARYEFFECTSLOTF		alGetAuxiliaryEffectSlotf;
extern LPALGETAUXILIARYEFFECTSLOTFV		alGetAuxiliaryEffectSlotfv;
	
/* ------------------------------------------------------------------------- */

#endif	//defined( ASURA_USE_OPENAL ) && ( ASURA_USE_OPENAL == 1 ) && !defined(ASURA_USE_AUDIO_SYSTEM)

#endif // !defined __ASURA_OPENAL_H__
