//
// Asura_Phonon.cpp
//
// Created by Kevin Lea - 12/06/2001 11:12:40
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Sound/Asura_Phonon.h"

#include "Core/Asura_Stream.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Sound.h"
#include "Sound/Asura_Sound_Event.h"
#include "Sound/Asura_Sound_Event_System.h"
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_TOOLS)
#include "Interface/Asura_CommandConsole.h"
#endif

#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#include "Lighting/Tools_LightSets.h"
#include "Mesh/Tools_PlatformMesh_System.h"
#endif

#include "Volume/Asura_VolumeGroup.h"


#if defined(ASURA_USE_MULTIPLE_HANDLERS)
Asura_Phonon_Handler*							Asura_Phonon_System::s_pxCurrentHandler	= NULL;		// Currently active handler.
Asura_TemplateHashTree<Asura_Phonon_Handler>	Asura_Phonon_System::s_xHandlerTree;				// Tree of all of the handlers available.
#else
Asura_Phonon_Handler							Asura_Phonon_System::s_xHandler;					// Single static-instance of a zone.
#endif

u_int	Asura_Phonon_System::s_uNextCommonGuid		= ASURA_PHONONGUID_COMMONFILE_START;
u_int	Asura_Phonon_System::s_uNextMissionGuid		= ASURA_PHONONGUID_MISSIONFILE_START;
float	Asura_Phonon_System::s_fVolumeScale		= 1.0f;
bool	Asura_Phonon_System::s_bDisablePhonons	= false;
u_int	Asura_Phonon_System::s_uNumPlayingPhonons = 0;
bool	Asura_Phonon_System::s_bPausePhononUpdates = false;

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_TOOLS)

#define CONSOLE_CMDNAME_SILENCEPHONONS L"SilencePhonons"

static bool ConsoleCmd_SilencePhonons( bool bSilence, bool bInterrogate, bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_SILENCEPHONONS );
        Asura_CommandConsole::Printf( "    Silences all current phonons." );       
    }
    else
    {
        if( !bInterrogate )
        {
            if ( bSilence )
            {
                Asura_Phonon_System::DisablePhonons();
            }
            else
            {
                Asura_Phonon_System::EnablePhonons();
            }
        }

        Asura_CommandConsole::Printf( "Phonons %s.", (Asura_Phonon_System::ArePhononsDisabled() ? "off" : "on") );
    }

	return true;
}

#endif // #if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_TOOLS)
	

static bool IsSphereInRange( const Asura_Vector_3& xPosition, float fRadius )
{
	// This function is actually duplicated in Asura_Sound_Event_System because I cant
	// really think of a nice place to put it, I'm reluctant to move it to the Audio system
	// because it isnt explicity related to low level audio.

	// Check this sphere is in range of the listener

	#ifndef ASURA_USE_AUDIO_SYSTEM

	return Asura_Sound_System::IsSphereInRange( xPosition, fRadius );
	
	#else // defined ASURA_USE_AUDIO_SYSTEM

	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xListenerPosition;
		Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );

		// Is the camera inside this sphere
		if( Asura_Vector_3( xPosition - xListenerPosition ).Magnitude() < fRadius )
		{
			return true;
		}
	}
	return false;

	#endif	// ASURA_USE_AUDIO_SYSTEM
}


static bool IsCuboidVolumeInRange(const Asura_Vector_3& xPosition, const Asura_Vector_3&  xExtents, const Asura_Quat* pxOrient)
{
	#ifndef ASURA_USE_AUDIO_SYSTEM

	return Asura_Sound_System::IsCuboidVolumeInRange( xPosition, xExtents, pxOrient );
	
	#else
	
	if(pxOrient)
	{
		Asura_Quat xInverseOrient(*pxOrient);
		xInverseOrient.Invert();
		for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
		{
			if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
			{
				continue;
			}

			Asura_Vector_3 xListenerPosition;
			Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );

			Asura_Vector_3 xRelativePosition = xPosition - xListenerPosition;

			xRelativePosition = xInverseOrient.RotateVector(xRelativePosition);

			if (Asura_Maths::Abs(xRelativePosition.x) <= xExtents.x &&
				Asura_Maths::Abs(xRelativePosition.y) <= xExtents.y &&
				Asura_Maths::Abs(xRelativePosition.z) <= xExtents.z)
			{
				return true;
			}
		}
	}
	else
	{
		for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
		{
			if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
			{
				continue;
			}

			Asura_Vector_3 xListenerPosition;
			Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );

			Asura_Vector_3 xRelativePosition = xPosition - xListenerPosition;

			if (Asura_Maths::Abs(xRelativePosition.x) <= xExtents.x &&
				Asura_Maths::Abs(xRelativePosition.y) <= xExtents.y &&
				Asura_Maths::Abs(xRelativePosition.z) <= xExtents.z)
			{
				return true;
			}
		}
	}

	return false;

	#endif
}


static bool IsBBInRange( const Asura_Bounding_Box& xBB )
{
	#ifndef ASURA_USE_AUDIO_SYSTEM

	return Asura_Sound_System::IsBBInRange( xBB );

	#else // defined ASURA_USE_AUDIO_SYSTEM

	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xListenerPosition;
		Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );
		
		if ( xBB.Contains( xListenerPosition ) )
		{
			return true;
		}
	}
	return false;

	#endif	// ASURA_USE_AUDIO_SYSTEM
}

////////////////////// Asura_Phonon_Handler ////////////////////////////////////////

void Asura_Phonon_Handler::UpdatePhonon(Asura_Hash_ID uHash, Asura_Phonon* pxPhonon)
{
	// (MH - 29.05.2002) in tools, phonons do not get a sound immediately
#ifdef ASURA_TOOLS
	if (!pxPhonon->GetEvent())
	{
		return;
	}
#endif // #ifdef ASURA_TOOLS

	bool bDisabled = Asura_Phonon_System::ArePhononsDisabled();
	#ifdef ASURA_TOOLS
	//Disable phonons that aren't in the 'light set' being used. RAR 14/06/2006
	bDisabled |= !pxPhonon->IsMemberOfCurrentLightSet();
	bDisabled |= !pxPhonon->IsEnabledForPlatform(Tools_PlatformMesh_System::GetPlatform(false));
	#endif
	if(bDisabled)
	{
		pxPhonon->DestroySound();
		return;
	}

	bool bPhononInRange;
	if(pxPhonon->m_bUseCuboidFalloff)
	{
		const Asura_Quat* pxOrient = pxPhonon->m_bUseOrient ? &pxPhonon->m_xOrient : 0;
		bPhononInRange = IsCuboidVolumeInRange( pxPhonon->m_xPosition, pxPhonon->m_xOuterCuboidRadius, pxOrient);
	}
	else
	{
		bPhononInRange = IsSphereInRange( pxPhonon->m_xPosition, pxPhonon->m_fLargestRadius );
	}

	if (!pxPhonon->HasValidSoundHandle())
	{
		if (!pxPhonon->m_bIsActive)
		{
			return;
		}
		if ( !bPhononInRange )
		{
			return;
		}
		pxPhonon->CreateSound();
		pxPhonon->m_bHasChanged = pxPhonon->HasValidSoundHandle();
	}
	else if(!pxPhonon->m_bIsActive)
	{
		//Phonon is no longer active... so stop the sound. RAR 11/04/03
		pxPhonon->DestroySound();
		return;
	}
	else if ( !bPhononInRange )
	{
		// (NJD - 11.07.2003) - phonon is no longer in range, so stop the sound
		pxPhonon->DestroySound();
		return;
	}
	if (!pxPhonon->HasValidSoundHandle())
	{
		Asura_Assert(false, "Failed to create sound for phonon.");
		return;
	}

	if (pxPhonon->m_bHasChanged)
	{
		if (pxPhonon->CurrentlyPlaying())
		{
			Asura_Sound_Event_System::UpdatePhononInstanceParameters( pxPhonon->m_iSoundEventHandle, pxPhonon );
		}
		pxPhonon->m_bHasChanged = false;
	}

	if (pxPhonon->m_bIsActive)
	{
		if (!pxPhonon->CurrentlyPlaying())
		{
			// see if we should retrigger
			if (!pxPhonon->m_pxEvent)                                 
			{
				char szTmp[128];
				Asura_PrintfToBfr(szTmp, "Phonon with GUID=%d has a missing sound event",128,true, pxPhonon->GetGuid());
				Asura_Assert(false, szTmp);
			}
			else
			{         
				if (pxPhonon->m_pxEvent->IsEndlessPlayback())
				{
					pxPhonon->m_bShouldTrigger = true;
				}
				else if (pxPhonon->m_bRetriggerable)
				{
					if (pxPhonon->m_bUseRetriggerRadius)
					{
						pxPhonon->m_bShouldTrigger = IsSphereInRange( pxPhonon->m_xPosition, pxPhonon->m_fRetriggerRadius );
					}
					else if (pxPhonon->m_bUseRetriggerBoundingBox)
					{
						pxPhonon->m_bShouldTrigger = IsBBInRange( pxPhonon->m_xRetriggerBoundingBox );
					}
					else
					{
						// we've already established that we're in range
						Asura_Assert( IsSphereInRange( pxPhonon->m_xPosition, pxPhonon->m_fLargestRadius ), "");
						
						pxPhonon->m_bShouldTrigger = true;
					}
				}

				if (pxPhonon->m_bShouldTrigger)
				{
					Asura_Sound_Event_System::TriggerPhononInstanceEventPlay( pxPhonon->m_iSoundEventHandle, pxPhonon );
					pxPhonon->m_bShouldTrigger = false;
				}
			}
		}
	}
	else
	{
		pxPhonon->DestroySound();
	}

	if (pxPhonon->IsPlaying())
	{
		++Asura_Phonon_System::s_uNumPlayingPhonons;
	}
}

void Asura_Phonon_Handler::SilencePhonon(Asura_Hash_ID uHash, Asura_Phonon* pxPhonon)
{
	pxPhonon->DestroySound();
}

Asura_Phonon* Asura_Phonon_Handler::AllocatePhonon(bool bMissionFile)
{
	// guarantee the next guid isn't used.
	Asura_Guid& uNextGuid = bMissionFile ? Asura_Phonon_System::s_uNextMissionGuid : Asura_Phonon_System::s_uNextCommonGuid;

	while(GetPhonon(uNextGuid) != NULL)
	{
		uNextGuid++;
	}

	// allocate a phonon using this guid
	return AllocatePhonon(uNextGuid++);
}

Asura_Phonon* Asura_Phonon_Handler::AllocatePhonon(Asura_Guid uGuid)
{
	if(uGuid == ASURA_GUID_UNREGISTERED)
	{
		Asura_Assert(uGuid != ASURA_GUID_UNREGISTERED, "Guid should have been assigned already");
		return AllocatePhonon(false);
	}

	Asura_Phonon*	pxPhonon;
	
	// (MH - 15.11.2002) try to make sure that if phonons got deleted, the empty
	// guids won't get reallocated to newly-created ones.
	if(uGuid >= ASURA_PHONONGUID_MISSIONFILE_START)
	{
		if (uGuid >= Asura_Phonon_System::s_uNextMissionGuid)
		{
			Asura_Phonon_System::s_uNextMissionGuid = uGuid + 1;
		}
	}
	else
	{
		if (uGuid >= Asura_Phonon_System::s_uNextCommonGuid)
		{
			Asura_Phonon_System::s_uNextCommonGuid = uGuid + 1;
		}
	}
	

	// Create the phonon...
	pxPhonon = m_xPhononTree.AddNode(uGuid);
	Asura_Assert( pxPhonon, "m_xPhononTree.AddNode() failed, probably duplicate phonon guid");
	if (!pxPhonon)
	{
		return NULL;
	}

	// Initialise the phonon...
	pxPhonon->Initialise();
	pxPhonon->m_uGuid = uGuid;
	pxPhonon->m_bShouldTrigger = true;
	pxPhonon->m_iSoundEventHandle = ASURA_INVALID_HANDLE;
	pxPhonon->m_bIsActive = false;
#ifdef ASURA_TOOLS
	pxPhonon->m_uTemplateGuid = ASURA_GUID_UNREGISTERED;
	pxPhonon->m_uLightSets = 0xffffffff;
	pxPhonon->m_uPlatforms = 0xffffffff;
#endif

	// Return the phonon...
	return(pxPhonon);
}

void Asura_Phonon_Handler::DeallocatePhonon(Asura_Phonon *pxPhonon)
{
	pxPhonon->DestroySound();
	m_xPhononTree.DeleteNodeByHash(pxPhonon->m_uGuid);
}


#if defined(ASURA_TOOLS)
void Asura_Phonon_Handler::SoundEventDeleted(const Asura_Sound_Event* pxEvent)
{
	m_xPhononTree.Traverse(SoundEventDeletedCallback,(void*)pxEvent);
}

bool Asura_Phonon_Handler::SoundEventDeletedCallback(Asura_Hash_ID uHash,Asura_Phonon* pxPhonon,void* pUserData)
{
	if((void*)pxPhonon->m_pxEvent == pUserData)
	{
		pxPhonon->m_pxEvent = 0;
		pxPhonon->m_uEventHashID = 0;
	}
	return true;
}

struct Tools_RenameEventData
{
	Asura_Hash_ID uOldHash;
	Asura_Hash_ID uNewHash;
};

void Asura_Phonon_Handler::SoundEventRenamed(Asura_Hash_ID uOldHash,Asura_Hash_ID uNewHash)
{
	Tools_RenameEventData xData;
	xData.uOldHash = uOldHash;
	xData.uNewHash = uNewHash;
	m_xPhononTree.Traverse(SoundEventRenamedCallback,(void*)&xData);
}

bool Asura_Phonon_Handler::SoundEventRenamedCallback(Asura_Hash_ID uHash,Asura_Phonon* pxPhonon,void* pUserData)
{
	Tools_RenameEventData* pxData = (Tools_RenameEventData*) pUserData;

	if(pxPhonon->m_uEventHashID == pxData->uOldHash)
	{
		pxPhonon->m_uEventHashID = pxData->uNewHash;
	}
	return true;
}

#endif //defined(ASURA_TOOLS)


void Asura_Phonon_System::Initialise(void)
{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_TOOLS)
    Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_SILENCEPHONONS, ConsoleCmd_SilencePhonons );
#endif
}

void Asura_Phonon_System::Update (void)
{
	#ifndef ASURA_USE_AUDIO_SYSTEM
	Asura_Assert( false, "Should not be calling Asura_Phonon_System::Update, Asura_Sound::Update will perform the same actions" );
	return;
	#else

	if( !AreUpdatesPaused() )
	{
		ResetNumPlayingPhonons();
		#if defined(ASURA_USE_MULTIPLE_HANDLERS)
		// Many phonon handlers - update each one in-turn...
		//s_xHandlerTree.Traverse( TraverseUpdatePhonons );
		// AUDIOTODO: This, see the old sound system (it seems horribly complicated)
		Asura_Assert( false, "Phonon system does not support multiple handlers yet");
		#else
		// Only one phonon handler - update it directly...
		GetCurrentHandler().Update();
		#endif
	}
	#endif
}

void Asura_Phonon_System::Shutdown(void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	Asura_Phonon_Handler&	xHandler	= GetCurrentHandler();

	// Reset the current instance...
	xHandler.Reset();
#endif

	// Reset the new GUID...
	s_uNextCommonGuid	= ASURA_PHONONGUID_COMMONFILE_START;
	s_uNextMissionGuid	= ASURA_PHONONGUID_MISSIONFILE_START;
}

void Asura_Phonon_System::Reset(void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	GetCurrentHandler().Reset();
#else
	// Shutdown the system, then re-initialise...
	Shutdown();
	Initialise();
#endif
}


#if defined(ASURA_USE_MULTIPLE_HANDLERS)

void Asura_Phonon_System::ReparentHandler(Asura_Phonon_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash)
{
	Asura_Phonon_Handler*	pxParentHandler	= s_xHandlerTree.GetDataByHash(uNewParentHash);

	if(pxParentHandler != NULL)
	{
		xSourceHandler.m_xPhononTree.SetParent(&(pxParentHandler->m_xPhononTree));
	}
}

void Asura_Phonon_System::ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash)
{
	Asura_Phonon_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uSourceHash);

	if(pxHandler != NULL)
	{
		ReparentHandler(*pxHandler, uNewParentHash);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Phonon_System::CreateHandler
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Attempts to create a phonon-system handler with the specified
//				  hash-ID.
//
// Parameters   : The hash-ID of the handler to create.
//
// Return value : A pointer to the handler created.
////////////////////////////////////////////////////////////////////////////////

Asura_Phonon_Handler* Asura_Phonon_System::CreateHandler (const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)
{
	Asura_Phonon_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHash);

		if(pxHandler != NULL)
		{
			ReparentHandler(*pxHandler, uParentHandlerHash);
		}
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an phonon handler with this hash-ID.");
	}

	// Return the handler...
	return(pxHandler);
}

/// End of function 'Asura_Phonon_System::CreateHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Phonon_System::DestroyHandler
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Attempts to destroy the phonon-handler with the specified
//				  unique hash-ID.
//
// Parameters   : The hash-ID of the handler to destroy.
//
// Return value : "true" if destroyed, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_Phonon_System::DestroyHandler (const Asura_Hash_ID uHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

/// End of function 'Asura_Phonon_System::DestroyHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Phonon_System::WriteHandlerLog
// Author       : Paul Cardy
// Date         : 20-12-07
//
// Description  : Writes a log of the contents of the supplied handler
//
// Parameters   : The hash-ID of the handler to log and the filename to log to
//
// Return value : None
////////////////////////////////////////////////////////////////////////////////

void Asura_Phonon_System::WriteHandlerReport(u_int uHandler, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 64;
	char szLogBuffer[uLOGBUFFERSIZE];

	Asura_Phonon_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHandler);
	if (pxHandler)
	{
		if (pxHandler->m_xPhononTree.GetNumberOfNodes())
		{
			Asura_PrintfToBfr(szLogBuffer, "Phonon:\t\t\t\t%d Phonon(s)\n", uLOGBUFFERSIZE, true, pxHandler->m_xPhononTree.GetNumberOfNodes());
			xLog << szLogBuffer;
		}
		else
		{
			xLog << "Phonon:\t\t\t\tNone\n";
		}
	}
	else
	{
		xLog << "Phonon:\t\t\t\tNo Handler\n";
	}
}

/// End of function 'Asura_Phonon_System::WriteHandlerLog'. ///


void Asura_Phonon_System::SilenceHandler (Asura_Hash_ID uHash, Asura_Phonon_Handler* pxHandler)
{
	pxHandler->Silence();
}

#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)

void Asura_Phonon_System::Silence (void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	s_xHandlerTree.Traverse(SilenceHandler);
#else
	s_xHandler.Silence();
#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)
}


void Asura_Phonon::Initialise()
{
	m_pxClipVolume = new Asura_VolumeGroup;
}

void Asura_Phonon::Shutdown()
{
	delete m_pxClipVolume;
	m_pxClipVolume = 0;
}

// functions for sending messages
void Asura_Phonon::Activate()
{
    m_bIsActive = true;
}

void Asura_Phonon::Deactivate()
{
    m_bIsActive = false;
}

void Asura_Phonon::Trigger()
{
    Activate();
	m_bShouldTrigger = true;
}

// functions for internal use only
void Asura_Phonon::Restart(void)
{
	m_bShouldTrigger = true;
	DestroySound();
}


void Asura_Phonon::SetPosition(const Asura_Vector_3& xPos)
{
	m_xPosition = xPos;
	if (m_iSoundEventHandle != ASURA_INVALID_HANDLE)
    {
        Asura_Sound_Event_System::UpdatePosition(m_iSoundEventHandle, m_xPosition);
    }
}

bool Asura_Phonon::IsPlaying() const
{
    if (m_iSoundEventHandle == ASURA_INVALID_HANDLE)
    {
        return false;
    }
	return Asura_Sound_Event_System::IsPlaying(m_iSoundEventHandle);
}

void Asura_Phonon::CalculateLargestRadius()
{
	if ( m_bUseCuboidFalloff )
	{
	    // (NJD - 11.07.2003) - approximate the largest radius by using the largest axis
	    // if this is not accurate enough we'll have to calculate the largest diagonal
		// (PSJE - 09.09.03) - It's not... Calculate the diagonal
		m_fLargestRadius = m_xOuterCuboidRadius.Magnitude();
	}
	else
	{
	    m_fLargestRadius = m_fOuterRadius;
	}
}

bool Asura_Phonon::CurrentlyPlaying() const
{
	return Asura_Sound_Event_System::IsPlaying(m_iSoundEventHandle);
}

bool Asura_Phonon::HasValidSoundHandle() const
{
	return m_iSoundEventHandle != ASURA_INVALID_HANDLE;
}

void Asura_Phonon::CreateSound()
{
	m_iSoundEventHandle = Asura_Sound_Event_System::CreatePhononInstance(this);
}

void Asura_Phonon::DestroySound()
{
	Asura_Sound_Event_System::Stop(m_iSoundEventHandle);
}

#ifdef ASURA_TOOLS
void Asura_Phonon::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	m_xPosition = pxInstance->TransformPosition(m_xPosition);
	m_xInitialPosition = pxInstance->TransformPosition(m_xInitialPosition);
	m_xOrient = pxInstance->TransformQuat(m_xOrient);
	m_pxClipVolume->ApplyChangesForTemplateInstance(pxInstance);
}

bool Asura_Phonon::IsMemberOfCurrentLightSet() const
{
	return Tools_LightSets::IsMemberOfCurrentLightSets(m_uLightSets);
}
#endif //ASURA_TOOLS

const Asura_Phonon& Asura_Phonon::operator = (const Asura_Phonon& xPhonon)
{
	if(this != &xPhonon)
	{
    	m_uGuid = xPhonon.m_uGuid;
		m_pxEvent = xPhonon.m_pxEvent;

		m_xPosition = xPhonon.m_xPosition;
		m_fInnerRadius = xPhonon.m_fInnerRadius;
		m_fOuterRadius = xPhonon.m_fOuterRadius;
		m_xInnerCuboidRadius = xPhonon.m_xInnerCuboidRadius;
		m_xOuterCuboidRadius = xPhonon.m_xOuterCuboidRadius;
		m_fLargestRadius = xPhonon.m_fLargestRadius;
		m_xOrient = xPhonon.m_xOrient;

		Asura_Assert(m_pxClipVolume && xPhonon.m_pxClipVolume,"Source or destination phonon not initialised correctly");
		m_pxClipVolume->Copy(xPhonon.m_pxClipVolume);


		m_bUseCuboidFalloff = xPhonon.m_bUseCuboidFalloff;
		m_bStartsActive = xPhonon.m_bStartsActive;
		m_bUseClipVolume = xPhonon.m_bUseClipVolume;
		m_bInvertClipVolume = xPhonon.m_bInvertClipVolume;
		m_bUseOrient = xPhonon.m_bUseOrient;
		m_bRetriggerable = xPhonon.m_bRetriggerable;
			
		m_bUseRetriggerRadius = xPhonon.m_bUseRetriggerRadius;	   
		m_bUseRetriggerBoundingBox = xPhonon.m_bUseRetriggerBoundingBox;
		m_fRetriggerRadius = xPhonon.m_fRetriggerRadius;		   
		m_xRetriggerBoundingBox = xPhonon.m_xRetriggerBoundingBox;   

		m_iSoundEventHandle = xPhonon.m_iSoundEventHandle;

		m_bShouldTrigger = xPhonon.m_bShouldTrigger;

		m_bHasChanged = xPhonon.m_bHasChanged;
		m_bIsActive = xPhonon.m_bIsActive;

		#ifdef ASURA_TOOLS
		m_uTemplateGuid = xPhonon.m_uTemplateGuid;
		m_xName = xPhonon.m_xName;
		m_uLightSets = xPhonon.m_uLightSets;
		m_uPlatforms = xPhonon.m_uPlatforms;

		m_uEventHashID = xPhonon.m_uEventHashID;

		m_xInitialPosition = xPhonon.m_xInitialPosition;
		#endif
	
	}
	return *this;
}


void Asura_Phonon::Set(const Asura_Phonon& xNewSettings)
{
		// always stop sound (Restart) if using events (for now)
		Restart();

        // (MH - 15.11.2002) don't set the guid - this is for copying properties.
		m_pxEvent = xNewSettings.m_pxEvent;
	
		m_xPosition = xNewSettings.m_xPosition;
        m_fInnerRadius = xNewSettings.m_fInnerRadius;
		m_fOuterRadius = xNewSettings.m_fOuterRadius;
        m_xInnerCuboidRadius = xNewSettings.m_xInnerCuboidRadius;
		m_xOuterCuboidRadius = xNewSettings.m_xOuterCuboidRadius;
		m_xOrient = xNewSettings.m_xOrient;

		m_pxClipVolume->Copy(xNewSettings.m_pxClipVolume);

		m_bUseCuboidFalloff = xNewSettings.m_bUseCuboidFalloff;
	    m_bStartsActive = xNewSettings.m_bStartsActive;
		m_bUseClipVolume = xNewSettings.m_bUseClipVolume;
		m_bInvertClipVolume = xNewSettings.m_bInvertClipVolume;
		m_bUseOrient = xNewSettings.m_bUseOrient;
		m_bRetriggerable = xNewSettings.m_bRetriggerable;

		m_bUseRetriggerRadius = xNewSettings.m_bUseRetriggerRadius;
		m_bUseRetriggerBoundingBox = xNewSettings.m_bUseRetriggerBoundingBox;
		m_fRetriggerRadius = xNewSettings.m_fRetriggerRadius;
		m_xRetriggerBoundingBox = xNewSettings.m_xRetriggerBoundingBox;


#ifdef ASURA_TOOLS
		m_xName = xNewSettings.m_xName;
		m_uLightSets = xNewSettings.m_uLightSets;
		m_uPlatforms = xNewSettings.m_uPlatforms;
#endif

#if defined(ASURA_TOOLS)
		m_uEventHashID = xNewSettings.m_uEventHashID;
#endif

#ifdef ASURA_TOOLS
		m_xInitialPosition = xNewSettings.m_xInitialPosition;
#endif

		CalculateLargestRadius();
		        
		m_bHasChanged = true;
}



