//////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

#include "Sound/Asura_XAudio2.h"

// Asura Platform-independent
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Mempool.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Resources/Asura_ResourceSet.h"
#include "Core/Threads/Asura_ProducerConsumerQueue.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_XAudio2_Buffer.h"
#include "Sound/Asura_XAudio2_Stream.h"
#include "Sound/Asura_Audio_DSP.h"
#include "Sound/XAPO_DSP/Asura_XAPO_DSP_AddNoise.h"
#include "Sound/XAPO_DSP/Asura_XAPO_DSP_SimpleFilter.h"
#include "Sound/XAPO_DSP/Asura_XAPO_DSP_Flanger.h"
#include "Sound/XAPO_DSP/Asura_XAPO_DSP_Chorus.h"
#include "Volume/Asura_VolumeGroup.h"

#ifdef ASURA_TOOLS
#include "Sound/XAPO_DSP/Asura_XAPO_DSP_Monitor.h"
#endif

// Misc
#include <xapofx.h>
#include <xaudio2.h>
#include <xaudio2fx.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// This array maps from XAudio2 speaker indices to Asura_Audio_System speaker IDs. The order is
// defined by audiodefs.h

const Asura_Audio_System::Speakers aeINDEPENDENT_SPEAKERS[] =
{
	Asura_Audio_System::ASURA_SPEAKER_FRONT_LEFT,         // SPEAKER_FRONT_LEFT
	Asura_Audio_System::ASURA_SPEAKER_FRONT_RIGHT,        // SPEAKER_FRONT_RIGHT
	Asura_Audio_System::ASURA_SPEAKER_FRONT_CENTRE,       // SPEAKER_FRONT_CENTER
	Asura_Audio_System::ASURA_SPEAKER_LOW_FREQ,           // SPEAKER_LOW_FREQUENCY
	Asura_Audio_System::ASURA_SPEAKER_REAR_LEFT,          // SPEAKER_BACK_LEFT
	Asura_Audio_System::ASURA_SPEAKER_REAR_RIGHT,         // SPEAKER_BACK_RIGHT
	Asura_Audio_System::ASURA_SPEAKER_FRONT_CENTRE_LEFT,  // SPEAKER_FRONT_LEFT_OF_CENTER
	Asura_Audio_System::ASURA_SPEAKER_FRONT_CENTRE_RIGHT, // SPEAKER_FRONT_RIGHT_OF_CENTER
	Asura_Audio_System::ASURA_SPEAKER_REAR_CENTRE,        // SPEAKER_BACK_CENTER
    Asura_Audio_System::ASURA_SPEAKER_SIDE_LEFT,          // SPEAKER_SIDE_LEFT
	Asura_Audio_System::ASURA_SPEAKER_SIDE_RIGHT,         // SPEAKER_SIDE_RIGHT
    
	// These have no equivalent in the independent list yet
	Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_CENTER
    Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_FRONT_LEFT
    Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_FRONT_CENTER
    Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_FRONT_RIGHT
    Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_BACK_LEFT
    Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_BACK_CENTER
    Asura_Audio_System::ASURA_NUM_SPEAKERS,               // SPEAKER_TOP_BACK_RIGHT
};

const size_t uINDEP_SPEAKER_ARRAY_SIZE =
	sizeof( aeINDEPENDENT_SPEAKERS ) / sizeof( aeINDEPENDENT_SPEAKERS[0] );

const u_int uMAX_CONCURRENT_VOICE_PREPERATIONS = 6;

#ifdef ASURA_TOOLS
static const u_int uMONITOR_BUFFER_SIZE = ( sizeof( float ) * 480 * 8 );	// DSP's give 480 floats per channel
#endif

/******************************************************************************/
/* Asura_XAudio2 Static Class Data                                            */
/******************************************************************************/

IXAudio2* Asura_XAudio2::s_pxEngine = NULL;
IXAudio2MasteringVoice* Asura_XAudio2::s_pxMasteringVoice = NULL;

IXAudio2SubmixVoice* Asura_XAudio2::s_pxReverbVoiceWithDSP = NULL;
IXAudio2SubmixVoice* Asura_XAudio2::s_pxReverbVoiceNoDSP = NULL;

IUnknown* Asura_XAudio2::s_apxReverbAPOs[2] = {0};
XAUDIO2FX_REVERB_PARAMETERS Asura_XAudio2::s_xReverbParams;

IXAudio2SubmixVoice* Asura_XAudio2::s_apxDSPVoices[2] = {0};
u_int Asura_XAudio2::s_uCurrDSPVoice = 0;
float Asura_XAudio2::s_fDSPFadeTimer = 0.0f;
float Asura_XAudio2::s_fDSPFadeDuration = 0.0f;

u_int Asura_XAudio2::s_uDeviceNumChannels = 0;
u_int Asura_XAudio2::s_uDeviceChannelMask = 0;
u_int Asura_XAudio2::s_uDeviceSampleRate = 0;

Asura_ResourceSet< Asura_XAudio2_Buffer > Asura_XAudio2::s_xBuffers( "XAudio2Buffer" );

Asura_TrackingHandlePool< Asura_XAudio2_Voice > Asura_XAudio2::s_xVoices(16);
Asura_TrackingHandlePool< Asura_XAudio2_Stream > Asura_XAudio2::s_xStreamPool(4);

#ifdef ASURA_TOOLS
char*			Asura_XAudio2::s_pMonitorBuffer;
volatile bool	Asura_XAudio2::s_bMonitorBufferReady;
u_int			Asura_XAudio2::s_uMonitorBufferChannels;
u_int			Asura_XAudio2::s_uMonitorBufferFrames;
Asura_XAPO_DSP_Monitor_Params Asura_XAudio2::s_xMonitorParams;
#endif

/******************************************************************************/
/* Data Types                                                                 */
/******************************************************************************/

struct VoicePrepData
{
	Asura_XAudio2_Voice*			pxVoice;
	Asura_XAudio2_Stream*			pxStream;
	const Asura_Audio_StreamInfo*	pxStreamInfo;
	float							fTimeOffset;
	bool							bMusic;
	bool							bFilter;
	Asura_Hash_ID					xDSPChain;
};


/******************************************************************************/
/* Globals                                                                    */
/******************************************************************************/

static Asura_ProducerConsumerQueue	g_xVoicePrepQueue( uMAX_CONCURRENT_VOICE_PREPERATIONS, 1, ASURA_MULTITHREAD_STACK_SIZE_SMALL );
static u_int						g_uNextVoicePrepData = 0;
static VoicePrepData				g_axVoicePrepData[ uMAX_CONCURRENT_VOICE_PREPERATIONS ];

/******************************************************************************/
/* Asura_XAudio2 Static Class Functions                                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateBuffer()
//
// Creates an empty XAudio2 buffer and adds it to the resource set
//
////////////////////////////////////////////////////////////////////////////////


Asura_XAudio2_Buffer* Asura_XAudio2::CreateBuffer( const char* const szFilename, bool& bAlreadyExisted )
{
	const Asura_Hash_ID uNameHash = Asura_GetHashID( szFilename );
	//see if a buffer already exists for this filename
	Asura_XAudio2_Buffer* pxBuffer = s_xBuffers.GetResource(uNameHash);
	if(pxBuffer)
	{
		bAlreadyExisted = true;
		//add a reference for the current resource bank
		s_xBuffers.AddResource(uNameHash, *pxBuffer, false);
	}
	else
	{
		bAlreadyExisted = false;
		//create a new buffer and add it
		pxBuffer = new Asura_XAudio2_Buffer;
		Asura_Assert( pxBuffer, "Failed to allocate memory for audio buffer." );

		if( pxBuffer )
		{
			#if defined ASURA_AUDIO_STORE_FILENAMES
			pxBuffer->m_szFilename = Asura_StringUtil::Duplicate( szFilename );
			#endif
			s_xBuffers.AddResource(uNameHash, *pxBuffer, true);
		}
	}

	return pxBuffer;
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::GetBuffer()
//
// Retrieves a sound buffer from storage in an audio handler
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Buffer* Asura_XAudio2::GetBuffer( const Asura_Hash_ID uHash )
{
	Asura_XAudio2_Buffer* const pxBuffer = s_xBuffers.GetResource( uHash );

	// (AAM - 10.11.2008) Used to assert here, but it causes problems because the dialogue system
	// doesn't necessarily have wave files associated with its lines
	//Asura_Assertf( pxBuffer, ("Failed to find sound buffer with hash '0x%X'.", uHash) );

	return pxBuffer;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::GetVoice()
//
// Retrieves the audio voice with a given handle
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Voice* Asura_XAudio2::GetVoice( const Asura_Handle iHandle )
{
	if( iHandle == ASURA_INVALID_HANDLE ) return NULL;

	Asura_XAudio2_Voice* const pxVoice = s_xVoices.GetPointer( iHandle );
	Asura_Assertf( pxVoice, ("Failed to get audio voice with handle %d.", iHandle) );

	return pxVoice;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::Initialise()
//
// Initialises XAudio2
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::Initialise()
{
	Asura_Assert( !s_pxEngine && !s_pxMasteringVoice, "XAudio2 already initialised." );

	UINT32 uFlags = 0;
	#if defined _DEBUG
		uFlags |= XAUDIO2_DEBUG_ENGINE;
	#endif

	s_pxEngine = NULL;
	HRESULT xResult = XAudio2Create( &s_pxEngine, uFlags );

	if( FAILED( xResult ) || !s_pxEngine )
	{
		Asura_Assert( false, "Failed to initialise XAudio2." );
		return;
	}

	s_pxMasteringVoice = NULL;
	xResult = s_pxEngine->CreateMasteringVoice( &s_pxMasteringVoice );

	if( FAILED( xResult ) || !s_pxMasteringVoice )
	{
		Asura_Assert( false, "Failed to create XAudio2 mastering voice." );
		SAFE_RELEASE( s_pxEngine );
		return;
	}

	// Get the default device details

	XAUDIO2_DEVICE_DETAILS xDetails;
	s_pxEngine->GetDeviceDetails( 0, &xDetails );

	s_uDeviceNumChannels = xDetails.OutputFormat.Format.nChannels;
	s_uDeviceChannelMask = xDetails.OutputFormat.dwChannelMask;
	s_uDeviceSampleRate = xDetails.OutputFormat.Format.nSamplesPerSec;

	Asura_OutputToDebugger_Retail(( "\nXAudio2 device ID: %S\n", xDetails.DeviceID ));
	Asura_OutputToDebugger_Retail(( "XAudio2 device name: %S\n", xDetails.DisplayName ));
	Asura_OutputToDebugger_Retail(( "XAudio2 sample rate: %u Hz\n", s_uDeviceSampleRate ));

	Asura_OutputToDebugger_Retail(( "XAudio2 num channels: %u (mask 0x%X)\n\n",
		s_uDeviceNumChannels, s_uDeviceChannelMask ));


	#ifdef ASURA_TOOLS
	// Attach our monitor DSP to the mastering voice
	s_pMonitorBuffer = static_cast<char*>( Asura_Allocate( uMONITOR_BUFFER_SIZE ) );
	if( s_pMonitorBuffer )
	{
		Asura_Mem::Set( s_pMonitorBuffer, 0, uMONITOR_BUFFER_SIZE );

		s_uMonitorBufferChannels = 0;
		s_uMonitorBufferFrames = 0;
		s_bMonitorBufferReady = false;

		XAUDIO2_EFFECT_DESCRIPTOR* pxDescriptors = new XAUDIO2_EFFECT_DESCRIPTOR[ 1 ];

		s_xMonitorParams.m_pbBufferWritten		= &s_bMonitorBufferReady;
		s_xMonitorParams.m_pfBuffer				= reinterpret_cast<float*>( s_pMonitorBuffer );
		s_xMonitorParams.m_puChannels			= &s_uMonitorBufferChannels;
		s_xMonitorParams.m_puSamples			= &s_uMonitorBufferFrames;
		s_xMonitorParams.m_uBufferSize			= uMONITOR_BUFFER_SIZE;

		pxDescriptors[0].pEffect = static_cast<IXAPO*>( new Asura_XAPO_DSP_Monitor );
		
		pxDescriptors[0].InitialState	= TRUE;
		pxDescriptors[0].OutputChannels = s_uDeviceNumChannels;

		XAUDIO2_EFFECT_CHAIN xChain;
		xChain.EffectCount = 1;
		xChain.pEffectDescriptors = pxDescriptors;

		s_pxMasteringVoice->SetEffectChain( &xChain );
		s_pxMasteringVoice->SetEffectParameters( 0, &s_xMonitorParams, sizeof( Asura_XAPO_DSP_Monitor_Params ), XAUDIO2_COMMIT_NOW );

		SAFE_RELEASE( pxDescriptors[0].pEffect );
		delete []pxDescriptors;
	}
	#endif

	// Let Asura_Audio_System know which speakers are active

	for( u_int u = 0; u < 32; ++u )
	{
		if( s_uDeviceChannelMask & (1 << u) )
		{
			const Asura_Audio_System::Speakers eSpeaker = GetIndependentSpeaker(u);

			if( eSpeaker != Asura_Audio_System::ASURA_NUM_SPEAKERS )
			{
				Asura_Audio_System::SetSpeakerActive( eSpeaker );
			}
			else
			{
				Asura_Assertf( false,
					("Speaker %u is not supported by the Asura audio system.", u) );
			}
		}
	}

	// Create submixes voices for master DSP effects. Can't just set the effects on the mastering
	// voices, because some sounds want to bypass the effect. Also need to crossfade new & old DSPs

	for( u_int u = 0; u < 2; ++u )
	{
		xResult = s_pxEngine->CreateSubmixVoice( &s_apxDSPVoices[u], s_uDeviceNumChannels,
			s_uDeviceSampleRate, 0, 1 );

		Asura_Assert( SUCCEEDED( xResult ), "Failed to create master DSP submix voice." );
	}

	// Create submix voice for reverb, which outputs to mastering voice. Create another one that
	// outputs to the DSP voices, so we can choose which based on sound event flags

	IXAudio2SubmixVoice* apxReverbVoices[2] = {0};

	for( u_int u = 0; u < 2; ++u )
	{
		UINT32 uReverbFlags = 0;
		#if defined _DEBUG
			uReverbFlags |= XAUDIO2FX_DEBUG;
		#endif

		xResult = XAudio2CreateReverb( &s_apxReverbAPOs[u], uReverbFlags );

		if( SUCCEEDED( xResult ) && s_apxReverbAPOs[u] )
		{
			XAUDIO2_EFFECT_DESCRIPTOR xDescriptor;

			xDescriptor.InitialState = TRUE;
			xDescriptor.OutputChannels = 1;
			xDescriptor.pEffect = s_apxReverbAPOs[u];

			XAUDIO2_EFFECT_CHAIN xChain;

			xChain.EffectCount = 1;
			xChain.pEffectDescriptors = &xDescriptor;

			xResult = s_pxEngine->CreateSubmixVoice( &apxReverbVoices[u], 1, s_uDeviceSampleRate, 0,
				0, NULL, &xChain );

			Asura_Assert( SUCCEEDED( xResult ) && apxReverbVoices[u],
				"Failed to create XAudio2 reverb submix voice." );
		}
		else
		{
			Asura_Assert( false, "Failed to create XAudio2 reverb APO." );
		}
	}

	s_pxReverbVoiceWithDSP = apxReverbVoices[0];
	s_pxReverbVoiceNoDSP   = apxReverbVoices[1];

	if( s_pxReverbVoiceWithDSP )
	{
		// The reverb voice that forwards on for master DSP outputs to both DSP voices. The volumes
		// of the DSP voices are updated based on the fade timer, with vol1 = 1-vol2 at all times

		#if defined MARCH_2009_SDK
			XAUDIO2_SEND_DESCRIPTOR axSendDescriptors[2] =
				{ { 0, s_apxDSPVoices[0] }, { 0, s_apxDSPVoices[1] } };

			const XAUDIO2_VOICE_SENDS xSends = { 2, axSendDescriptors };
		#else
			IXAudio2Voice* apxSends[2] = { s_apxDSPVoices[0], s_apxDSPVoices[1] };
			const XAUDIO2_VOICE_SENDS xSends = { 2, apxSends };
		#endif

		xResult = s_pxReverbVoiceWithDSP->SetOutputVoices( &xSends );
		Asura_Assert( SUCCEEDED( xResult ), "Failed to set reverb voice outputs to master DSP." );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::Reset()
//
// Resets XAudio2
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::Reset()
{
	#ifdef ASURA_ASSERT
	Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Voice > xIter( &s_xVoices );
	xIter.Begin();
	Asura_XAudio2_Voice* const pxVoice = xIter.GetCurrentPointer();
	if( pxVoice )
	{
		Asura_Assert( false, "Asura_Audio_Voice is still present at reset" );
	}
	#endif
	
	s_xStreamPool.Reset();
	s_xVoices.Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::Shutdown()
//
// Shuts XAudio2 down
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::Shutdown()
{
	// Wait for all streaming sounds to be finished
	Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Stream > xIter( &s_xStreamPool );
	for( xIter.Begin(); !xIter.End(); xIter.Next() )
	{
		Asura_XAudio2_Stream* const pxStream = xIter.GetCurrentPointer();

		while( !pxStream->CanBeDestroyed() );

		pxStream->CloseFile();
		s_xStreamPool.Deallocate( xIter.GetCurrentHandle() );
	}

	#ifdef ASURA_ASSERT
	Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Voice > xVoiceIter( &s_xVoices );
	xVoiceIter.Begin();
	Asura_XAudio2_Voice* const pxVoice = xVoiceIter.GetCurrentPointer();
	if( pxVoice )
	{
		Asura_Assert( false, "Asura_Audio_Voice is still present at shutdown" );
	}
	#endif

	s_xVoices.Reset();
	s_xBuffers.Reset();

	if( s_pxReverbVoiceWithDSP )
	{
		s_pxReverbVoiceWithDSP->DestroyVoice();
		s_pxReverbVoiceWithDSP = NULL;
	}

	if( s_pxReverbVoiceNoDSP )
	{
		s_pxReverbVoiceNoDSP->DestroyVoice();
		s_pxReverbVoiceNoDSP = NULL;
	}

	for( u_int u = 0; u < 2; ++u )
	{
		if( s_apxDSPVoices[u] )
		{
			s_apxDSPVoices[u]->DestroyVoice();
			s_apxDSPVoices[u] = NULL;
		}
	}

	if( s_pxMasteringVoice )
	{
		s_pxMasteringVoice->DestroyVoice();
		s_pxMasteringVoice = NULL;
	}

	#ifdef ASURA_TOOLS
	Asura_Deallocate( s_pMonitorBuffer );
	#endif

	SAFE_RELEASE( s_pxEngine );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::Update()
//
// Once per frame update for XAudio2
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::Update()
{
	// Update streams
	{
		Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Stream > xIter( &s_xStreamPool );

		for( xIter.Begin(); !xIter.End(); xIter.Next() )
		{
			Asura_XAudio2_Stream* const pxStream = xIter.GetCurrentPointer();

			if( pxStream->CanBeDestroyed() )
			{
				pxStream->CloseFile();
				s_xStreamPool.Deallocate( xIter.GetCurrentHandle() );
			}
		}
	}

	// Update voices
	{
		Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Voice > xIter( &s_xVoices );

		for( xIter.Begin(); !xIter.End(); xIter.Next() )
		{
			Asura_XAudio2_Voice* const pxVoice = xIter.GetCurrentPointer();
			if( pxVoice )			//extra check to prevent crash
			{
				pxVoice->Update();

				if (pxVoice->IsPendingDestruction() && !pxVoice->IsPreparing())
				{
					DestroyVoice(pxVoice->m_iPendingDestructionHandle);
					pxVoice->SetPendingDestruction(ASURA_INVALID_HANDLE);
				}
			}
		}
	}

	// Update reverb

	for( u_int u = 0; u < 2; ++u )
	{
		// Update both reverb voices. One routes to the DSP voices, the other direct to the
		// mastering output

		IXAudio2SubmixVoice* const apxReverbVoices[2] =
			{ s_pxReverbVoiceWithDSP, s_pxReverbVoiceNoDSP };

		if( apxReverbVoices[u] )
		{
			Asura_Audio_Reverb xReverb;
			float fAmount;

			Asura_Audio_System::GetCurrentReverb( xReverb, fAmount );

			XAUDIO2FX_REVERB_I3DL2_PARAMETERS xI3DL2;

			xI3DL2.WetDryMix         = 100.0f;
			xI3DL2.Room              = xReverb.m_iRoom;
			xI3DL2.RoomHF            = xReverb.m_iRoomHF;
			xI3DL2.RoomRolloffFactor = xReverb.m_fRoomRolloffFactor;
			xI3DL2.DecayTime         = xReverb.m_fDecayTime;
			xI3DL2.DecayHFRatio      = xReverb.m_fDecayHFRatio;
			xI3DL2.Reflections       = xReverb.m_iReflections;
			xI3DL2.ReflectionsDelay  = xReverb.m_fReflectionsDelay;
			xI3DL2.Reverb            = xReverb.m_iReverb;
			xI3DL2.ReverbDelay       = xReverb.m_fReverbDelay;
			xI3DL2.Diffusion         = xReverb.m_fDiffusion;
			xI3DL2.Density           = xReverb.m_fDensity;
			xI3DL2.HFReference       = xReverb.m_fHFReference;

			ReverbConvertI3DL2ToNative( &xI3DL2, &s_xReverbParams );

			HRESULT xResult;
			const UINT32 uSize = sizeof( s_xReverbParams );

			xResult = apxReverbVoices[u]->SetEffectParameters( 0, &s_xReverbParams, uSize );
			Asura_Assert( SUCCEEDED( xResult ), "Failed to set XAudio2 reverb parameters." );

			xResult = apxReverbVoices[u]->SetVolume( fAmount );
			Asura_Assert( SUCCEEDED( xResult ), "Failed to set reverb voice volume." );
		}
	}

	// Update DSP crossfade
	{
		const float fOldDSPVol =
			(s_fDSPFadeDuration > 0.0f ? s_fDSPFadeTimer / s_fDSPFadeDuration : 0.0f);

		const float fNewDSPVol = 1.0f - fOldDSPVol;

		IXAudio2SubmixVoice* const pxOldDSPVoice = s_apxDSPVoices[ 1 - s_uCurrDSPVoice ];
		IXAudio2SubmixVoice* const pxNewDSPVoice = s_apxDSPVoices[ s_uCurrDSPVoice ];

		HRESULT xResult;

		if( pxOldDSPVoice )
		{
			xResult = pxOldDSPVoice->SetVolume( fOldDSPVol );
			Asura_Assert( SUCCEEDED( xResult ), "Failed to set outgoing DSP volume." );
		}

		if( pxNewDSPVoice )
		{
			xResult = pxNewDSPVoice->SetVolume( fNewDSPVol );
			Asura_Assert( SUCCEEDED( xResult ), "Failed to set incoming DSP volume." );
		}

		s_fDSPFadeTimer -= Asura_Timers::GetGameTimeSlice();
		s_fDSPFadeTimer = Asura_Maths::Max( 0.0f, s_fDSPFadeTimer );
	}
}

#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::IsWaveLoaded()
//
// Queries whether a given wave is already loaded. Only available in Tools
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2::IsWaveLoaded( const char* const szFilename )
{
	const Asura_Hash_ID uHash = Asura_GetHashID( szFilename );
	return s_xBuffers.GetResource( uHash ) != NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::ForceWaveToBeLoaded()
//
// Forces a delayed wave to be loaded now. Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::ForceWaveToBeLoaded( const Asura_Hash_ID uWave )
{
	Asura_XAudio2_Buffer* const pxBuffer = GetBuffer( uWave );
	if( pxBuffer )
	{
		pxBuffer->LoadIfDelayed();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::ReloadWave()
//
// Reloads the wave with the given hash ID. Only available in Tools
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::ReloadWave( const Asura_Hash_ID uWave, const bool bForceReloadNow )
{
	Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Voice > xIter( &s_xVoices );

	for( xIter.Begin(); !xIter.End(); xIter.Next() )
	{
		Asura_XAudio2_Voice* const pxVoice = xIter.GetCurrentPointer();
		if( Asura_GetHashID( pxVoice->GetFilename() ) == uWave )
		{
			pxVoice->Stop();
		}
	}

	Asura_XAudio2_Buffer* const pxBuffer = GetBuffer( uWave );
	if( !pxBuffer ) return;

	pxBuffer->Unload();

	if( bForceReloadNow )
	{
		pxBuffer->LoadIfDelayed();
	}
}

void Asura_XAudio2::SetWaveLooping( const Asura_Hash_ID uWave, const bool bLoop )
{
	// Adjust the looping property of the buffer itself
	{
		Asura_XAudio2_Buffer* const pxBuffer = Asura_XAudio2::GetBuffer( uWave );

		Asura_Assertf( pxBuffer, ("Failed to find XAudio2 buffer 0x%X.", uWave) );
		if( !pxBuffer ) return;

		pxBuffer->SetLooping( bLoop );
	}

	// If any voices were playing this buffer, they need to be updated for the new looping property.
	// This won't change automatically because the buffer's already been submitted

	Asura_TrackingHandlePool_Iterator< Asura_XAudio2_Voice > xIter( &s_xVoices );

	for( xIter.Begin(); !xIter.End(); xIter.Next() )
	{
		Asura_XAudio2_Voice* const pxVoice = xIter.GetCurrentPointer();
		if( Asura_GetHashID( pxVoice->GetFilename() ) == uWave && pxVoice->IsPlaying() )
		{
			if( bLoop )
			{
				// Enqueue the wave again, this time looping
				pxVoice->EnqueueWave( uWave );
			}
			else if( pxVoice->m_pxSourceVoice )
			{
				// Stop looping after the current one is done
				pxVoice->m_pxSourceVoice->ExitLoop();
			}
		}
	}
}

#endif // defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::GetWaveFilename()
//
// Retrieves the filename of the given wave resource
//
////////////////////////////////////////////////////////////////////////////////

#if defined ASURA_AUDIO_STORE_FILENAMES

const char* Asura_XAudio2::GetWaveFilename( const Asura_Hash_ID uWave )
{
	const Asura_XAudio2_Buffer* const pxBuffer = s_xBuffers.GetResource( uWave );
	Asura_Assertf( pxBuffer, ("Failed to find wave with hash ID 0x%X.", uWave) );

	return pxBuffer ? pxBuffer->GetFilename() : NULL;
}

#endif // defined ASURA_AUDIO_STORE_FILENAMES

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateVoiceInternal()
//
// Creates and initialises an Asura_XAudio2_Voice, asserting on failure
//
////////////////////////////////////////////////////////////////////////////////

// Disable new macro so we can use placement form
#include "Core/Asura_MemoryManagement_DebugOff.h"

Asura_XAudio2_Voice* Asura_XAudio2::CreateVoiceInternal( Asura_Handle& iHandle )
{
	// Allocate an Asura_XAudio2_Voice wrapper to manage the voice

	iHandle = s_xVoices.Allocate();
	if( iHandle == ASURA_INVALID_HANDLE )
	{
		Asura_Assert( false, "Unable to allocate new audio voice." );
		return NULL;
	}

	Asura_XAudio2_Voice* pxVoice = s_xVoices.GetPointer( iHandle );

	// Use placement new to construct the object. Tracking handle pools don't do this on
	// Allocate() (though they do call constructors at pool creation time, which might result in
	// double construction)
	pxVoice = new (pxVoice) Asura_XAudio2_Voice;
	
	return pxVoice;
}

// Re-enable new macro
#include "Core/Asura_MemoryManagement_DebugOn.h"

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateXAudioSource()
//
// Wrapper for creating an IXAudio2SourceVoice, asserting on failure
//
////////////////////////////////////////////////////////////////////////////////

IXAudio2SourceVoice* Asura_XAudio2::CreateXAudioSource( const WAVEFORMATEX& xFmt,
	const bool bFilter, const bool bMusic, IXAudio2VoiceCallback* const pxCallback )
{
	Asura_Assert( s_pxEngine, "XAudio2 engine not initialised." );
	if( !s_pxEngine ) return NULL;

	// XAudio2 uses a memory pool internally, allowing us to create and destroy voices on the fly

	u_int uFlags = 0;
	uFlags |= (bFilter ? XAUDIO2_VOICE_USEFILTER : 0);
	uFlags |= (bMusic ? XAUDIO2_VOICE_MUSIC : 0);

	IXAudio2SourceVoice* pxSourceVoice = NULL;
	s_pxEngine->CreateSourceVoice( &pxSourceVoice, &xFmt, uFlags, XAUDIO2_DEFAULT_FREQ_RATIO,
		pxCallback );

	Asura_Assert( pxSourceVoice, "Failed to create XAudio2 source voice." );

	return pxSourceVoice;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateXAudioSubmix()
//
// Wrapper for creating an IXAudio2SubmixVoice, asserting on failure
//
////////////////////////////////////////////////////////////////////////////////

IXAudio2SubmixVoice* Asura_XAudio2::CreateXAudioSubmix( const bool bFilter, const bool bMusic )
{
	Asura_Assert( s_pxEngine, "XAudio2 engine not initialised." );
	if( !s_pxEngine ) return NULL;

	// XAudio2 uses a memory pool internally, allowing us to create and destroy voices on the fly

	u_int uFlags = 0;
	uFlags |= (bFilter ? XAUDIO2_VOICE_USEFILTER : 0);
	uFlags |= (bMusic ? XAUDIO2_VOICE_MUSIC : 0);

	IXAudio2SubmixVoice* pxSubmixVoice = NULL;
	s_pxEngine->CreateSubmixVoice( &pxSubmixVoice, 1, s_uDeviceSampleRate, uFlags );

	Asura_Assert( pxSubmixVoice, "Failed to create XAudio2 submix voice." );

	return pxSubmixVoice;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateVoice()
//
// Creates a playback voice, initialising it with a given audio buffer
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_XAudio2::CreateVoice( const Asura_Hash_ID uWave,
	const bool bFilter, const bool bMusic, const Asura_Hash_ID xDSPChain )
{
	Asura_XAudio2_Buffer* const pxMutableBuffer = GetBuffer( uWave );
	if( !pxMutableBuffer ) return ASURA_INVALID_HANDLE;

	#if defined ASURA_TOOLS
		if( !pxMutableBuffer->LoadIfDelayed() ) return ASURA_INVALID_HANDLE;
	#endif

	const Asura_XAudio2_Buffer* const pxBuffer = pxMutableBuffer;
	const WAVEFORMATEX& xWaveFormat = pxBuffer->m_xFmt.m_xWFX.Format;

	// Check for duff wave formats. XAudio2 doesn't gracefully handle this, it just crashes
	if( xWaveFormat.nChannels == 0 || xWaveFormat.nSamplesPerSec == 0 )
	{
		Asura_Assertf( false, ("'%s' has bad wave format.", pxBuffer->GetFilename()) );
		return ASURA_INVALID_HANDLE;
	}

	Asura_Handle iVoiceHandle;
	Asura_XAudio2_Voice* const pxVoice = CreateVoiceInternal( iVoiceHandle );

	if( !pxVoice ) return ASURA_INVALID_HANDLE;

	#if defined ASURA_AUDIO_STORE_FILENAMES
		// Store this for ease of debugging
		pxVoice->m_szFilename = Asura_Audio_System::GetWaveFilename( uWave );
	#endif

	IXAudio2SourceVoice* const pxSourceVoice =
		CreateXAudioSource( xWaveFormat, bFilter, bMusic, NULL );

	if( !pxSourceVoice )
	{
		s_xVoices.Deallocate( iVoiceHandle );
		return ASURA_INVALID_HANDLE;
	}

	#if defined ASURA_ASSERT
		const HRESULT xResult =
	#endif
		pxSourceVoice->SubmitSourceBuffer( pxBuffer->GetParams(), pxBuffer->GetWMA() );

	Asura_Assert( SUCCEEDED( xResult ), "Failed to submit XAudio2 source buffer to voice." );

	pxVoice->m_pxSourceVoice = pxSourceVoice;

#ifdef ASURA_USE_XAUDIO_OCCLUSION
	IXAudio2SubmixVoice* pxSubmix = Asura_XAudio2::CreateXAudioSubmix( true, false );
	Asura_Assert( pxSubmix, "Failed to create occlusion submix voice." );

	pxVoice->m_pxDirectPathSubmixVoice = pxSubmix;
#endif

	// Create the DSP chain for this voice
	CreateDSPChain( xDSPChain, pxSourceVoice, xWaveFormat.nChannels );

	return iVoiceHandle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateVoice()
//
// Creates a source voice. Instead of taking the format from a given buffer,
// it's directly passed in. No audio data is queued on the voice.
//
// Callbacks are set up on this voice, to allow the user to manage the buffers
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_XAudio2::CreateVoice( const WAVEFORMATEX& xFmt,
	const bool bFilter, const bool bMusic, const Asura_Hash_ID xDSPChain )
{
	Asura_Handle iVoiceHandle;
	Asura_XAudio2_Voice* const pxVoice = CreateVoiceInternal( iVoiceHandle );

	if( !pxVoice ) return ASURA_INVALID_HANDLE;

	IXAudio2SourceVoice* const pxSourceVoice = CreateXAudioSource( xFmt, bFilter, bMusic, pxVoice );
	if( !pxSourceVoice )
	{
		s_xVoices.Deallocate( iVoiceHandle );
		return ASURA_INVALID_HANDLE;
	}

	pxVoice->m_pxSourceVoice = pxSourceVoice;

#ifdef ASURA_USE_XAUDIO_OCCLUSION
	IXAudio2SubmixVoice* pxSubmix = Asura_XAudio2::CreateXAudioSubmix( true, false );
	Asura_Assert( pxSubmix, "Failed to create occlusion submix voice." );

	pxVoice->m_pxDirectPathSubmixVoice = pxSubmix;
#endif

	// Create the DSP chain for this voice
	CreateDSPChain( xDSPChain, pxSourceVoice, xFmt.nChannels );

	return iVoiceHandle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateStreamingVoice()
//
// Creates a playback voice, initialised to play the given stream
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_XAudio2::CreateStreamingVoice( const Asura_Hash_ID uStreamHash,
	const bool bFilter, const bool bMusic, const float fTimeOffset, const Asura_Hash_ID xDSPChain )
{
	// Get info for the streaming wave
	const Asura_Audio_StreamInfo* const pxStreamInfo = Asura_Audio_System::GetStreamInfo( uStreamHash );

	if( !pxStreamInfo ) 
	{
		return ASURA_INVALID_HANDLE;
	}

	// Allocate a new streaming object
	const Asura_Handle iStreamHandle = s_xStreamPool.Allocate();
	if( iStreamHandle == ASURA_INVALID_HANDLE )
	{
		#if defined ASURA_ASSERT
			#if defined ASURA_AUDIO_STORE_FILENAMES
				const char* const szFriendlyName = pxStreamInfo->m_szFilename;
			#else
				const char* const szFriendlyName = "<Unknown>, define ASURA_AUDIO_STORE_FILENAMES";
			#endif
		#endif
		Asura_Assertf( false, ("Failed to allocate new audio stream for '%s'.", szFriendlyName) );
		return ASURA_INVALID_HANDLE;
	}

	Asura_XAudio2_Stream* const pxStream = s_xStreamPool.GetPointer( iStreamHandle );

	// We don't use placement new on these, because their constructors need to allocate platform-
	// specific memory. Don't want to do that each time here. Must ensure we clear out the flags
	// from whichever voice used this stream last though, especially m_bReleased
	pxStream->Reset();

	// Create a new voice
	Asura_Handle iVoiceHandle;
	Asura_XAudio2_Voice* const pxVoice = CreateVoiceInternal( iVoiceHandle );

	if( !pxVoice )
	{
		s_xStreamPool.Deallocate( iStreamHandle );

		return ASURA_INVALID_HANDLE;
	}
	pxVoice->m_iStreamHandle = iStreamHandle;

	#if defined ASURA_AUDIO_STORE_FILENAMES
	// Store this for ease of debugging
	pxVoice->m_szFilename = pxStreamInfo->m_szFilename;
	#endif

	// Thread out the initialisation of this streaming voice
	VoicePrepData* pxPrepData = &g_axVoicePrepData[ g_uNextVoicePrepData ];
	g_uNextVoicePrepData = ( g_uNextVoicePrepData + 1 ) % uMAX_CONCURRENT_VOICE_PREPERATIONS;

	pxPrepData->pxVoice			= pxVoice;
	pxPrepData->pxStream		= pxStream;
	pxPrepData->pxStreamInfo	= pxStreamInfo;
	pxPrepData->fTimeOffset		= fTimeOffset;
	pxPrepData->bMusic			= bMusic;
	pxPrepData->bFilter			= bFilter;
	pxPrepData->xDSPChain		= xDSPChain;
	
	pxVoice->SetPreparing( true );

	g_xVoicePrepQueue.AddToQueue( ThreadFunc_PrepareStreamingVoice, pxPrepData, sizeof( VoicePrepData ) );

	return iVoiceHandle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::ThreadFunc_PrepareStreamingVoice()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::ThreadFunc_PrepareStreamingVoice( void* pData, const u_int uDataSize )
{
	VoicePrepData* pxPrepData = static_cast<VoicePrepData*>( pData );
	
	#if defined ASURA_ASSERT
		#if defined ASURA_AUDIO_STORE_FILENAMES
			const char* const szFriendlyName = pxPrepData->pxStreamInfo->m_szFilename;
		#else
			const char* const szFriendlyName = "<Unknown>, define ASURA_AUDIO_STORE_FILENAMES";
		#endif
	#endif

	// Open and parse the wave file
	if( !pxPrepData->pxStream->OpenFile( pxPrepData->pxStreamInfo->m_szPackage, pxPrepData->pxStreamInfo->m_uLocation ) )
	{
		Asura_Assertf( false, ("Failed to open audio stream '%s'.", szFriendlyName) ); 
		pxPrepData->pxVoice->SetPreparing( false );
		return;	
	}

	const WAVEFORMATEX& xFmt = pxPrepData->pxStream->GetFormat().m_xWFX.Format;
	IXAudio2SourceVoice* const pxSourceVoice = CreateXAudioSource( xFmt, pxPrepData->bFilter, pxPrepData->bMusic, pxPrepData->pxVoice );

	if( !pxSourceVoice )
	{
		Asura_Assert( false, "Failed to create xaudio source voice" );
		pxPrepData->pxStream->CloseFile();
		pxPrepData->pxVoice->SetPreparing( false );
		return;
	}

	pxPrepData->pxVoice->m_pxSourceVoice = pxSourceVoice;
	pxPrepData->pxVoice->SetStartOffset( pxPrepData->fTimeOffset );

	pxPrepData->pxStream->SetLooping( pxPrepData->pxStreamInfo->m_bLooping ); // Do this before binding, for first reads!
	pxPrepData->pxStream->Bind( pxSourceVoice, pxPrepData->fTimeOffset );

	#ifdef ASURA_USE_XAUDIO_OCCLUSION
	IXAudio2SubmixVoice* pxSubmix = CreateXAudioSubmix( true, false );
	Asura_Assert( pxSubmix, "Failed to create occlusion submix voice." );

	pxPrepData->pxVoice->m_pxDirectPathSubmixVoice = pxSubmix;
	#endif

	// TODO: are there threading issues here? erk
	CreateDSPChain( pxPrepData->xDSPChain, pxSourceVoice, xFmt.nChannels );

	pxPrepData->pxVoice->SetPreparing( false );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateSubmixVoice()
//
// Creates a submix voice. You can route regular voices to these with
// Asura_Audio_Voice::SetSubmixOutputs()
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_XAudio2::CreateSubmixVoice( const bool bFilter, const bool bMusic, const Asura_Hash_ID xDSPChain )
{
	Asura_Handle iVoiceHandle;

	Asura_XAudio2_Voice* const pxVoice = CreateVoiceInternal( iVoiceHandle );
	if( !pxVoice ) return ASURA_INVALID_HANDLE;

	IXAudio2SubmixVoice* const pxSubmixVoice = CreateXAudioSubmix( bFilter, bMusic );
	if( !pxSubmixVoice )
	{
		s_xVoices.Deallocate( iVoiceHandle );
		return ASURA_INVALID_HANDLE;
	}

	pxVoice->m_pxSubmixVoice = pxSubmixVoice;

	CreateDSPChain( xDSPChain, pxSubmixVoice, 1 );	// Force mono DSP? Not sure how to handle that

	return iVoiceHandle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateWrapperVoice()
//
// Creates an Asura voice to wrap an existing IXAudio2SourceVoice. Used for FMV
// playback, for instance, where the 360 XMV player creates a voice for audio
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_XAudio2::CreateWrapperVoice( IXAudio2SourceVoice* const pxSourceVoice )
{
	Asura_Handle iVoiceHandle;

	Asura_XAudio2_Voice* const pxVoice = CreateVoiceInternal( iVoiceHandle );
	if( !pxVoice ) return ASURA_INVALID_HANDLE;

	pxVoice->m_pxSourceVoice = pxSourceVoice;
	pxVoice->m_bExternalVoice = true;

	return iVoiceHandle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::DestroyVoice()
//
// Destroys a playback voice
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::DestroyVoice( const Asura_Handle iVoiceHandle )
{
	Asura_XAudio2_Voice* const pxVoice = GetVoice( iVoiceHandle );
	if( !pxVoice ) return;

	Asura_Mutex* pxMutex = NULL;

	if( pxVoice->IsPreparing() )
	{
		pxVoice->SetPendingDestruction(iVoiceHandle);
		return;
	}

	const Asura_Handle iStreamHandle = pxVoice->m_iStreamHandle;

	if( iStreamHandle != ASURA_INVALID_HANDLE )
	{
		Asura_XAudio2_Stream* const pxStream = GetStream( iStreamHandle );
		if( pxStream )
		{
			// Lock the mutex. We don't want the streaming thread to submit more data right now
			pxMutex = &pxStream->GetMutex();
			pxMutex->Lock();

			pxStream->Release();
		}

		pxVoice->m_iStreamHandle = ASURA_INVALID_HANDLE;
	}

	IXAudio2SourceVoice* const pxSourceVoice = pxVoice->m_pxSourceVoice;
	if( pxSourceVoice )
	{
		if( !pxVoice->m_bExternalVoice )
		{
			pxSourceVoice->DestroyVoice();
		}

		pxVoice->m_pxSourceVoice = NULL;
	}

	IXAudio2SubmixVoice* const pxSubmixVoice = pxVoice->m_pxSubmixVoice;
	if( pxSubmixVoice )
	{
		if( !pxVoice->m_bExternalVoice )
		{
			pxSubmixVoice->DestroyVoice();
		}

		pxVoice->m_pxSubmixVoice = NULL;
	}

	#ifdef ASURA_USE_XAUDIO_OCCLUSION
	if ( pxVoice->m_pxDirectPathSubmixVoice )
	{
		pxVoice->m_pxDirectPathSubmixVoice->DestroyVoice();
		pxVoice->m_pxDirectPathSubmixVoice = NULL;
	}
	#endif

	// Would normally call the destructor here as above, but that causes something weird to happen
	// internally in the tracking handle pool. Don't need to do anything 'cause we cleaned up anyway

	s_xVoices.Deallocate( iVoiceHandle );

	if( pxMutex )
	{
		// Streaming thread is now free to update, but it will bail out due to having been released
		pxMutex->Unlock();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateAPO()
//
// Creates an XAudio2 audio processing object of the given type
//
////////////////////////////////////////////////////////////////////////////////

IXAPO* Asura_XAudio2::CreateAPO( const Asura_Audio_DSPType eType )
{
	switch( eType )
	{
		case ASURA_DSP_ADD_NOISE:		return new Asura_XAPO_DSP_AddNoise;
		case ASURA_DSP_SIMPLE_FILTER:	return new Asura_XAPO_DSP_SimpleFilter;
		case ASURA_DSP_FLANGER:			return new Asura_XAPO_DSP_Flanger;
		case ASURA_DSP_CHORUS:			return new Asura_XAPO_DSP_Chorus;
	}
 
	Asura_Assert( false, "Unknown type");
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::CreateDSPChain()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::CreateDSPChain( const Asura_Hash_ID uChainHash, IXAudio2Voice* pxVoice, u_int uChannels )
{
	Asura_Assert( pxVoice, "No voice passed to Asura_XAudio2::CreateDSPChain" );
	if( !pxVoice ) return;

	if( uChainHash == ASURA_HASH_ID_UNSET )
	{
		pxVoice->SetEffectChain( NULL );
		return;
	}

	// First create a list of effect descriptors

	Asura_Audio_DSPChain* pxChain = Asura_Audio_System::GetDSPChainByHash( uChainHash );

	Asura_Assertf( pxChain, ( "Failed to find DSP chain with hash %u", uChainHash ) );

	if( !pxChain ) return;
	
	if( pxChain->m_uNumEffects == 0 )
	{
		pxVoice->SetEffectChain( NULL );
		return;
	}

	XAUDIO2_EFFECT_DESCRIPTOR* pxDescriptors =
		new XAUDIO2_EFFECT_DESCRIPTOR[ pxChain->m_uNumEffects ];

	for( u_int u = 0; u < pxChain->m_uNumEffects; ++u )
	{
		pxDescriptors[u].pEffect = CreateAPO( pxChain->m_ppxEffects[u]->GetType() );

		pxDescriptors[u].InitialState = TRUE;
		pxDescriptors[u].OutputChannels = uChannels;
	}

	XAUDIO2_EFFECT_CHAIN xChain;
	xChain.EffectCount = pxChain->m_uNumEffects;
	xChain.pEffectDescriptors = pxDescriptors;

	pxVoice->SetEffectChain( &xChain );

	for( u_int u = 0; u < pxChain->m_uNumEffects; ++u )
	{
		// Don't need these interfaces around any more, effects are reference counted
		SAFE_RELEASE( pxDescriptors[u].pEffect );
	}

	delete[] pxDescriptors;

	// Finally set the parameters of all the effects. In XAudio2 this is done through the voice

	for( u_int u = 0; u < pxChain->m_uNumEffects; ++u )
	{
		const Asura_DSP_Params* const pxEffect = pxChain->m_ppxEffects[u];
		if( !pxEffect ) continue;

		pxVoice->SetEffectParameters( u, pxEffect->GetParams(), pxEffect->GetParamSize() );
		pxVoice->EnableEffect( u, XAUDIO2_COMMIT_NOW );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::SetMasterDSP()
//
// Sets the currently active master DSP
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::SetMasterDSP( const Asura_Hash_ID uDSPChain, const float fFadeTime )
{
	s_uCurrDSPVoice = 1 - s_uCurrDSPVoice;
	CreateDSPChain( uDSPChain, s_apxDSPVoices[ s_uCurrDSPVoice ], GetDeviceNumChannels() );

	s_fDSPFadeTimer = s_fDSPFadeDuration = fFadeTime;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::GetIndependentSpeaker()
//
// Maps from an XAudio2 speaker index to an Asura_Audio_System speaker ID
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_System::Speakers Asura_XAudio2::GetIndependentSpeaker( const u_int uXAudio2Speaker )
{
	if( uXAudio2Speaker < uINDEP_SPEAKER_ARRAY_SIZE )
	{
		return aeINDEPENDENT_SPEAKERS[ uXAudio2Speaker ];
	}
	else
	{
		Asura_Assertf( false, ("Speaker index %u out of range.", uXAudio2Speaker) );
		return Asura_Audio_System::ASURA_NUM_SPEAKERS;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::GetStream()
//
// Retrieves the XAudio2 stream referenced by the given handle
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Stream* Asura_XAudio2::GetStream( const Asura_Handle iHandle )
{
	return iHandle == ASURA_INVALID_HANDLE ? NULL : s_xStreamPool.GetPointer( iHandle );
}

#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::GetMonitorBuffer()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::GetMonitorBuffer ( char* &pBuffer, u_int &uChannels, u_int &uFrames )
{
	if( !s_bMonitorBufferReady )
	{
		pBuffer		= 0;
		uChannels	= 0;
		uFrames		= 0;
		return;
	}
	pBuffer		= s_pMonitorBuffer;
	uChannels	= s_uMonitorBufferChannels;
	uFrames		= s_uMonitorBufferFrames;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2::ReadyForMoreMonitorData()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2::ReadyForMoreMonitorData	( )
{
	s_bMonitorBufferReady = false;
}

#endif

/******************************************************************************/
/* Asura_XAudio2::EffectParams Class Functions                                */
/******************************************************************************/

/*Asura_XAudio2::EffectParams::EffectParams()
	: m_uSize(0)
{
}

void Asura_XAudio2::EffectParams::Initialise( const Asura_Audio_DSPEffectParams& xEssence )
{
	switch( xEssence.m_eType )
	{
		case ASURA_DSP_REVERB:
		{
			const Asura_Audio_DSPEffect_Reverb& xReverbEssence =
				static_cast< const Asura_Audio_DSPEffect_Reverb& >( xEssence );

			XAUDIO2FX_REVERB_I3DL2_PARAMETERS xI3DL2;

			xI3DL2.WetDryMix         = 100.0f;
			xI3DL2.Room              = xReverbEssence.m_iRoom;
			xI3DL2.RoomHF            = xReverbEssence.m_iRoomHF;
			xI3DL2.RoomRolloffFactor = xReverbEssence.m_fRoomRolloffFactor;
			xI3DL2.DecayTime         = xReverbEssence.m_fDecayTime;
			xI3DL2.DecayHFRatio      = xReverbEssence.m_fDecayHFRatio;
			xI3DL2.Reflections       = xReverbEssence.m_iReflections;
			xI3DL2.ReflectionsDelay  = xReverbEssence.m_fReflectionsDelay;
			xI3DL2.Reverb            = xReverbEssence.m_iReverb;
			xI3DL2.Diffusion         = xReverbEssence.m_fDiffusion;
			xI3DL2.Density           = xReverbEssence.m_fDensity;
			xI3DL2.HFReference       = xReverbEssence.m_fHFReference;

			ReverbConvertI3DL2ToNative( &xI3DL2, &m_xReverb );

			m_uSize = sizeof( XAUDIO2FX_REVERB_PARAMETERS );

			break;
		}

		case ASURA_DSP_LIGHT_REVERB:
		{
			const Asura_Audio_DSPEffect_LightReverb& xLightReverbEssence =
				static_cast< const Asura_Audio_DSPEffect_LightReverb& >( xEssence );

			m_xLightReverb.Diffusion = xLightReverbEssence.m_fDiffusion;
			m_xLightReverb.RoomSize = xLightReverbEssence.m_fRoomSize;

			m_uSize = sizeof( FXREVERB_PARAMETERS );

			break;
		}

		case ASURA_DSP_EQUALIZER:
		{
			const Asura_Audio_DSPEffect_Equalizer& xEqualizerEssence =
				static_cast< const Asura_Audio_DSPEffect_Equalizer& >( xEssence );

			m_xEqualizer.FrequencyCenter0 = xEqualizerEssence.m_afFrequencyCentres[0];
			m_xEqualizer.FrequencyCenter1 = xEqualizerEssence.m_afFrequencyCentres[1];
			m_xEqualizer.FrequencyCenter2 = xEqualizerEssence.m_afFrequencyCentres[2];
			m_xEqualizer.FrequencyCenter3 = xEqualizerEssence.m_afFrequencyCentres[3];

			m_xEqualizer.Gain0 = xEqualizerEssence.m_afGains[0];
			m_xEqualizer.Gain1 = xEqualizerEssence.m_afGains[1];
			m_xEqualizer.Gain2 = xEqualizerEssence.m_afGains[2];
			m_xEqualizer.Gain3 = xEqualizerEssence.m_afGains[3];

			m_xEqualizer.Bandwidth0 = xEqualizerEssence.m_afBandwidths[0];
			m_xEqualizer.Bandwidth1 = xEqualizerEssence.m_afBandwidths[1];
			m_xEqualizer.Bandwidth2 = xEqualizerEssence.m_afBandwidths[2];
			m_xEqualizer.Bandwidth3 = xEqualizerEssence.m_afBandwidths[3];

			m_uSize = sizeof( FXEQ_PARAMETERS );

			break;
		}

		case ASURA_DSP_MASTERING_LIMITER:
		{
			const Asura_Audio_DSPEffect_MasteringLimiter& xMasteringLimiterEssence =
				static_cast< const Asura_Audio_DSPEffect_MasteringLimiter& >( xEssence );

			m_xMasteringLimiter.Loudness = xMasteringLimiterEssence.m_uLoudness;
			m_xMasteringLimiter.Release = xMasteringLimiterEssence.m_uRelease;

			m_uSize = sizeof( FXMASTERINGLIMITER_PARAMETERS );

			break;
		}

		case ASURA_DSP_ECHO:
		{
			const Asura_Audio_DSPEffect_Echo& xEchoEssence =
				static_cast< const Asura_Audio_DSPEffect_Echo& >( xEssence );

			m_xEcho.WetDryMix = xEchoEssence.m_fWetDryMix;
			m_xEcho.Feedback = xEchoEssence.m_fFeedback;
			m_xEcho.Delay = xEchoEssence.m_fDelay;

			m_uSize = sizeof( FXECHO_PARAMETERS );

			break;
		}

		default:
		{
			Asura_Assertf( false, ("Unexpected DSP type (%d).", xEssence.m_eType) );
			break;
		}
	}
}

void Asura_XAudio2::EffectParams::Apply( IXAudio2Voice* const pxVoice, const u_int uIndex ) const
{
	Asura_Assert( pxVoice, "Invalid voice." );

	if( pxVoice && m_uSize > 0 )
	{
		#if defined ASURA_ASSERT
			const HRESULT xResult =
		#endif
			pxVoice->SetEffectParameters( uIndex, &m_xLightReverb, m_uSize );

		Asura_Assert( SUCCEEDED( xResult ), "Failed to set DSP effect parameters." );
	}
}
*/
#endif // defined ASURA_USE_AUDIO_SYSTEM
