//////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

#include "Sound/Asura_XAudio2_Buffer.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Endian.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Sound/Asura_XAudio2.h"

// Tools
#if defined ASURA_TOOLS
	#include "Sound/Tools_Sound_System.h"
#endif

// Misc. This is only needed for the version check on XAUDIO2_BUFFER::pAudioData, and can be
// removed once we're all above the November 2008 SDKs
#if defined _XBOX
	#include <xdk.h>
#else
	#include "Core/Asura_MemoryManagement_DebugOff.h"
	#include <d3dx9.h>
	#include "Core/Asura_MemoryManagement_DebugOn.h"
#endif

/******************************************************************************/
/* Macros                                                                     */
/******************************************************************************/

#define UNPACK_FOURCC(x)                       \
{                                              \
	static_cast< char >( ((x) >>  0) & 0xFF ), \
	static_cast< char >( ((x) >>  8) & 0xFF ), \
	static_cast< char >( ((x) >> 16) & 0xFF ), \
	static_cast< char >( ((x) >> 24) & 0xFF )  \
}

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const u_int u4CC_DATA = MAKEFOURCC( 'd', 'a', 't', 'a' );
const u_int u4CC_DPDS = MAKEFOURCC( 'd', 'p', 'd', 's' );
const u_int u4CC_FMT  = MAKEFOURCC( 'f', 'm', 't', ' ' );
const u_int u4CC_RIFF = MAKEFOURCC( 'R', 'I', 'F', 'F' );
const u_int u4CC_SEEK = MAKEFOURCC( 's', 'e', 'e', 'k' );
const u_int u4CC_SMPL = MAKEFOURCC( 's', 'm', 'p', 'l' );
const u_int u4CC_WAVE = MAKEFOURCC( 'W', 'A', 'V', 'E' );
const u_int u4CC_WSMP = MAKEFOURCC( 'w', 's', 'm', 'p' );
const u_int u4CC_XMA2 = MAKEFOURCC( 'X', 'M', 'A', '2' );
const u_int u4CC_XWMA = MAKEFOURCC( 'X', 'W', 'M', 'A' );

/******************************************************************************/
/* Asura_XAudio2_Buffer Class Functions                                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::Asura_XAudio2_Buffer()
//
// Constructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Buffer::Asura_XAudio2_Buffer()
	: m_pAudioData( NULL )
{
	Asura_Mem::Set( &m_xParams, 0, sizeof( m_xParams ) );
	Asura_Mem::Set( &m_xWMA, 0, sizeof( m_xWMA ) );
	Asura_Mem::Set( &m_xFmt, 0, sizeof( m_xFmt ) );

	#if defined ASURA_AUDIO_STORE_FILENAMES
		m_szFilename = NULL;
	#endif

	#if defined ASURA_TOOLS
		m_bAttemptedLoad = false;
	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::~Asura_XAudio2_Buffer()
//
// Destructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Buffer::~Asura_XAudio2_Buffer()
{
	if( m_xFmt.m_xWFX.Format.wFormatTag != 0 )
	{
		Asura_XAudio2::Platform_Deallocate( m_xFmt.m_xWFX.Format.wFormatTag, m_pAudioData );
	}
	else
	{
		Asura_Assert( !m_pAudioData, "Audio data shouldn't be allocated without a format tag." );
	}

	#if defined ASURA_AUDIO_STORE_FILENAMES
		delete[] m_szFilename;
	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::LoadWave()
//
// Processes a chunk stream containing a wave, and stores it in this buffer
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_Buffer::LoadWave( Asura_Chunk_Stream& xStream, const u_int uFileSize,
	const u_int uFlags )
{
	#if defined ASURA_TOOLS
		if( Tools_Sound_System::IsDontLoadWavs() )
		{
			xStream.Skip( uFileSize );
			CreateBlankWave();

			return true;
		}
	#endif

	const u_int uStreamStartPos = xStream.GetPosition();

	Asura_XAudio2_WaveReader xWaveReader( xStream, m_xFmt, m_xParams.AudioBytes, &m_pAudioData );
	if( !xWaveReader.ReadWave() ) return false;

	#if defined NOVEMBER_2008_SDK
		m_xParams.pAudioData = static_cast< BYTE* >( m_pAudioData );
	#else
		m_xParams.pAudioData = static_cast< const BYTE* >( m_pAudioData );
	#endif

	// AUDIOTODO: allow buffer queues
	m_xParams.Flags = XAUDIO2_END_OF_STREAM;

	if( uFlags & Asura_Chunk_ResourceFile::SF_LOOPING )
	{
		SetLooping( true );
	}

	// Don't do anything if looping flag is false, as that might overwrite earlier set up for
	// delayed waves

	const u_int uBytesRead = xStream.GetPosition() - uStreamStartPos;
	if( uBytesRead < uFileSize )
	{
		// For some reason, the contents of resource file chunks are always padded out to multiples
		// of four bytes. Make up for the potential padding here
		xStream.Skip( uFileSize - uBytesRead );
	}

	#if defined ASURA_TOOLS
		m_bAttemptedLoad = true;
	#endif

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::SetLooping()
//
// Specifies whether an XAudio2 buffer should loop until stopped or play once
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Buffer::SetLooping( const bool bLooping )
{
	m_xParams.LoopCount = (bLooping ? XAUDIO2_LOOP_INFINITE : 0);
}

#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::LoadIfDelayed()
//
// This version of LoadWave() reads the audio buffer from a wave file. Only
// available in Tools builds. Optionally, info about a chunk stream can be
// passed in, and used instead of the recorded disk filename. This is used when
// loading from ToolsGmSnd.asr a file which has already been registered through
// e.g. a sound event chunk load
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_Buffer::LoadIfDelayed( Asura_Chunk_Stream* pxStream /*= NULL*/,
	const u_int uFileSize /*= 0*/ )
{
	// Delayed loading is Tools-only, and requires ASURA_AUDIO_STORE_FILENAMES
	if( m_bAttemptedLoad )
	{
		if( pxStream )
		{
			// This happens if a sound was forced loaded by being played through Tools, then quit
			// to game caused ToolsGmSnd.asr to be loaded and we've got another set of data for
			// the same wave file. Just skip it
			pxStream->Skip( uFileSize );
		}

		return true;
	}

	if( Tools_Sound_System::IsDontLoadWavs() )
	{
		// Do this now to avoid even opening the file
		CreateBlankWave();
		return true;
	}

	if( pxStream )
	{
		return LoadWave( *pxStream, uFileSize, 0 );
	}
	else
	{
		char szNetworkFilename[ MAX_PATH ];
		Asura_FileAccess::PlatformGetNetworkFilename( m_szFilename, szNetworkFilename, MAX_PATH );

		const Asura_Handle iFileHandle = Asura_FileAccess::Open( szNetworkFilename );
		if( iFileHandle == ASURA_INVALID_HANDLE )
		{
			// (AAM - 10.11.2008) Used to assert here, but it causes problems because the dialogue
			// system doesn't necessarily have wave files associated with its lines
			//Asura_Assertf( false, ("Failed to load audio file '%s'.", m_szFilename) );
			return false;
		}

		const u_int uFileSize = Asura_FileAccess::GetFileSize( iFileHandle );
		Asura_Chunk_File_Stream xStream( iFileHandle );

		const bool bSuccess = LoadWave( xStream, uFileSize, 0 );

		Asura_FileAccess::Close( iFileHandle );
		return bSuccess;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::Unload()
//
// Unloads the contents of an XAudio2 wave buffer. Only available in Tools
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Buffer::Unload()
{
	Asura_Deallocate( m_pAudioData );

	// Preserve the filename after the blat
	const char* const szFilename = GetFilename();

	Asura_Mem::Set( this, 0, sizeof( Asura_XAudio2_Buffer ) );
	m_szFilename = szFilename;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::GetDuration()
//
// Get the length, in seconds, of the audio data stored in the buffer
//
////////////////////////////////////////////////////////////////////////////////

float Asura_XAudio2_Buffer::GetDuration() const
{
	if( !m_pAudioData )
	{
		return 0.0f;
	}

	if( m_xFmt.m_xWFX.Format.wFormatTag != WAVE_FORMAT_PCM )
	{
		Asura_Assert( false, "Expected PCM wav format" );
		return 0.0f;
	}

	return static_cast<float>( m_xParams.AudioBytes / ( m_xFmt.m_xWFX.Format.nBlockAlign ) ) / static_cast<float>( m_xFmt.m_xWFX.Format.nSamplesPerSec );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::CreateBlankWave()
//
// Fills an XAudio2 buffer with a short amount of silence, for "don't load wavs"
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Buffer::CreateBlankWave()
{
	// Create a short 8-bit mono PCM wave filled with silence. Old sound system used 150 Hz, but
	// XAudio2 can' go that low

	const u_int uSize = 1024;

	m_pAudioData = Asura_Allocate( uSize );
	if( m_pAudioData )
	{
		// 127 is zero volume for 8-bit samples
		Asura_Mem::Set( m_pAudioData, 0x7F, uSize );
	}

	m_xFmt.m_xWFX.Format.nAvgBytesPerSec = uSize;
	m_xFmt.m_xWFX.Format.nChannels = 1;
	m_xFmt.m_xWFX.Format.nBlockAlign = 1;
	m_xFmt.m_xWFX.Format.nSamplesPerSec = uSize;
	m_xFmt.m_xWFX.Format.wBitsPerSample = 8;
	m_xFmt.m_xWFX.Format.wFormatTag = WAVE_FORMAT_PCM;
	m_xFmt.m_xWFX.Format.cbSize = 0;

	#if defined NOVEMBER_2008_SDK
		m_xParams.pAudioData = static_cast< BYTE* >( m_pAudioData );
	#else
		m_xParams.pAudioData = static_cast< const BYTE* >( m_pAudioData );
	#endif

	m_xParams.AudioBytes = uSize;
	m_xParams.Flags = XAUDIO2_END_OF_STREAM;

	m_bAttemptedLoad = true;
}

#endif // defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Buffer::GetWMA()
//
// Retrieves xWMA params for an audio buffer, or NULL if it's not that format
//
////////////////////////////////////////////////////////////////////////////////

const XAUDIO2_BUFFER_WMA* Asura_XAudio2_Buffer::GetWMA() const
{
	if( m_xFmt.m_xWFX.Format.wFormatTag == WAVE_FORMAT_WMAUDIO2
	 || m_xFmt.m_xWFX.Format.wFormatTag == WAVE_FORMAT_WMAUDIO3 )
	{
		return &m_xWMA;
	}
	else return NULL;
}

/******************************************************************************/
/* Asura_XAudio2_WaveReader Class Functions                                   */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_WaveReader::Asura_XAudio2_WaveReader()
//
// Constructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_WaveReader::Asura_XAudio2_WaveReader( Asura_Chunk_Stream& xStream,
	Asura_XAudio2_WaveFormat& xFmt, u_int& uTotalDataSize, void** ppAudioData /*= NULL*/ )

	: m_xStream( xStream )
	, m_xFmt( xFmt )
	, m_uTotalDataSize( uTotalDataSize )
	, m_ppAudioData( ppAudioData )
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_WaveReader::ReadWave()
//
// Reads a wave file. Outputs the wave format and total data size. If the
// provided buffer is empty, allocates enough for the entire file and stores
// the pointer where it was told. Returns false iff error occurred
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_WaveReader::ReadWave()
{
	const u_int uStreamStartPos = m_xStream.GetPosition();

	u_int uFileSize;

	// Read RIFF header
	{
		u_int uChunkType;
		u_int uChunkSize;

		m_xStream >> uChunkType >> uChunkSize;

		#if defined ASURA_PLATFORM_IS_BIG_ENDIAN
			uChunkType = Asura_Endian::SwapBytes( uChunkType );
			uChunkSize = Asura_Endian::SwapBytes( uChunkSize );
		#endif

		Asura_Assert( uChunkType == u4CC_RIFF, "Not a RIFF file." );
		if( uChunkType != u4CC_RIFF ) return false;

		// File size is RIFF chunk size + the above chunk header
		uFileSize = uChunkSize + 8;
	}

	// RIFF chunk contents are the subtype FOURCC then subtype-dependent data

	u_int uSubType;
	m_xStream >> uSubType;

	#if defined ASURA_PLATFORM_IS_BIG_ENDIAN
		uSubType = Asura_Endian::SwapBytes( uSubType );
	#endif

	Asura_Assert( uSubType == u4CC_WAVE || uSubType == u4CC_XWMA, "Invalid audio chunk." );
	if( uSubType != u4CC_WAVE && uSubType != u4CC_XWMA ) return false;

	while( m_xStream.GetPosition() < uStreamStartPos + uFileSize )
	{
		// Parsing returns false if error occurs, or we've finished reading a streaming stub
		const PCReturnCode eStatus = ParseChunk();

		if( eStatus == ASURA_PCRC_FINISHED ) return true;
		if( eStatus == ASURA_PCRC_ERROR ) return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_WaveReader::ParseChunk()
//
// Parses and processes a wave chunk from a stream. Return value indicates
// whether another chunk should be read, the file is finished, or error occurred
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_WaveReader::PCReturnCode Asura_XAudio2_WaveReader::ParseChunk()
{
	u_int uChunkType, uChunkSize;
	m_xStream >> uChunkType >> uChunkSize;

	#if defined ASURA_PLATFORM_IS_BIG_ENDIAN
		uChunkType = Asura_Endian::SwapBytes( uChunkType );
		uChunkSize = Asura_Endian::SwapBytes( uChunkSize );
	#endif

	switch( uChunkType )
	{
		case u4CC_DATA:
		{
			m_uTotalDataSize = uChunkSize;

			if( m_ppAudioData )
			{
				*m_ppAudioData =
					Asura_XAudio2::Platform_Allocate( m_xFmt.m_xWFX.Format.wFormatTag, uChunkSize );

				Asura_Assert( *m_ppAudioData, "Failed to allocate memory for sound." );
				if( !*m_ppAudioData ) return ASURA_PCRC_ERROR;

				// Don't endian-swap this
				const u_int uBytesRead = m_xStream.Read( m_uTotalDataSize, *m_ppAudioData );

				return uBytesRead == m_uTotalDataSize ? ASURA_PCRC_CONTINUE : ASURA_PCRC_ERROR;
			}
			else
			{
				// When streaming into provided buffers, don't continue with the rest of the file
				return ASURA_PCRC_FINISHED;
			}
		}

		// AUDIOTODO: no need to handle xWMA files yet
		//case u4CC_DPDS:
		//{
		//	XAUDIO2_BUFFER_WMA& xWMA = m_xWMA;

		//	m_xStream >> xWMA.PacketCount;

		//	#if defined ASURA_PLATFORM_IS_BIG_ENDIAN
		//		xWMA.PacketCount = Asura_Endian::SwapBytes( xWMA.PacketCount );
		//	#endif

		//	// This needs to be specificed in DWORDs
		//	Asura_Assert( xWMA.PacketCount % 4 == 0, "Packet count not a multiple of 4." );
		//	xWMA.PacketCount /= 4;

		//	// AUDIOTODO: need to endian-swap
		//	const u_int uBytesRead =
		//		m_xStream.Read( uChunkSize, &xWMA.pDecodedPacketCumulativeBytes );

		//	return uBytesRead == uChunkSize;
		//}

		case u4CC_FMT:
		{
			if( uChunkSize > sizeof( Asura_XAudio2_WaveFormat ) )
			{
				Asura_Assert( false, "Audio file contains invalid format chunk." );
				return ASURA_PCRC_ERROR;
			}

			const u_int uBytesRead = m_xStream.Read( uChunkSize, &m_xFmt );
			if( uBytesRead != uChunkSize ) return ASURA_PCRC_ERROR;

			#if defined ASURA_PLATFORM_IS_BIG_ENDIAN

				m_xFmt.m_xWFX.Format.wFormatTag =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.wFormatTag );
				m_xFmt.m_xWFX.Format.nChannels =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.nChannels );
				m_xFmt.m_xWFX.Format.nSamplesPerSec =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.nSamplesPerSec );
				m_xFmt.m_xWFX.Format.nAvgBytesPerSec =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.nAvgBytesPerSec );
				m_xFmt.m_xWFX.Format.nBlockAlign =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.nBlockAlign );
				m_xFmt.m_xWFX.Format.wBitsPerSample =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.wBitsPerSample );
				m_xFmt.m_xWFX.Format.cbSize =
					Asura_Endian::SwapBytes( m_xFmt.m_xWFX.Format.cbSize );

			#endif

			switch( m_xFmt.m_xWFX.Format.wFormatTag )
			{
				case WAVE_FORMAT_ADPCM:
				{
					ADPCMWAVEFORMAT& xFormat = m_xFmt.m_xADPCM;

					#if defined ASURA_PLATFORM_IS_BIG_ENDIAN

						xFormat.wSamplesPerBlock =
							Asura_Endian::SwapBytes( xFormat.wSamplesPerBlock );
						xFormat.wNumCoef =
							Asura_Endian::SwapBytes( xFormat.wNumCoef );

						for( u_int u = 0; u < xFormat.wNumCoef; ++u )
						{
							xFormat.aCoef[u] = Asura_Endian::SwapBytes( xFormat.aCoef[u] );
						}

					#endif

					const bool bCorrectSize = (uChunkSize == sizeof( xFormat ));
					Asura_Assert( bCorrectSize, "Audio file contains invalid ADPCM format chunk." );

					return bCorrectSize ? ASURA_PCRC_CONTINUE : ASURA_PCRC_ERROR;
				}

				case WAVE_FORMAT_XMA2:
				{
					XMA2WAVEFORMATEX& xFormat = m_xFmt.m_xXMA2;

					#if defined ASURA_PLATFORM_IS_BIG_ENDIAN

						xFormat.NumStreams     = Asura_Endian::SwapBytes( xFormat.NumStreams );
						xFormat.ChannelMask    = Asura_Endian::SwapBytes( xFormat.ChannelMask );
						xFormat.SamplesEncoded = Asura_Endian::SwapBytes( xFormat.SamplesEncoded );
						xFormat.BytesPerBlock  = Asura_Endian::SwapBytes( xFormat.BytesPerBlock );
						xFormat.PlayBegin      = Asura_Endian::SwapBytes( xFormat.PlayBegin );
						xFormat.PlayLength     = Asura_Endian::SwapBytes( xFormat.PlayLength );
						xFormat.LoopBegin      = Asura_Endian::SwapBytes( xFormat.LoopBegin );
						xFormat.LoopLength     = Asura_Endian::SwapBytes( xFormat.LoopLength );
						xFormat.LoopCount      = Asura_Endian::SwapBytes( xFormat.LoopCount );
						xFormat.EncoderVersion = Asura_Endian::SwapBytes( xFormat.EncoderVersion );
						xFormat.BlockCount     = Asura_Endian::SwapBytes( xFormat.BlockCount );

					#endif

					const bool bCorrectSize = (uChunkSize == sizeof( xFormat ));
					Asura_Assert( bCorrectSize, "Audio file contains invalid XMA2 format chunk." );

					return bCorrectSize ? ASURA_PCRC_CONTINUE : ASURA_PCRC_ERROR;
				}

				case WAVE_FORMAT_PCM:
				{
					// The chunk is actually just a WAVEFORMATEX - swapping all handled above
					return ASURA_PCRC_CONTINUE;
				}

				case WAVE_FORMAT_WMAUDIO2:
				case WAVE_FORMAT_WMAUDIO3:
				{
					#if defined ASURA_PLATFORM_IS_BIG_ENDIAN

						WAVEFORMATEXTENSIBLE& xFmt = m_xFmt.m_xWFX;

						// Note that xWMA files with <=2 channels don't have these extra fields,
						// but that's OK. We'll just be swapping meaningless bytes

						xFmt.Samples.wReserved = Asura_Endian::SwapBytes( xFmt.Samples.wReserved );
						xFmt.dwChannelMask = Asura_Endian::SwapBytes( xFmt.dwChannelMask );

						xFmt.SubFormat.Data1 = Asura_Endian::SwapBytes( xFmt.SubFormat.Data1 );
						xFmt.SubFormat.Data2 = Asura_Endian::SwapBytes( xFmt.SubFormat.Data2 );
						xFmt.SubFormat.Data3 = Asura_Endian::SwapBytes( xFmt.SubFormat.Data3 );
						// Fourth component is an array of bytes, which don't need to be swapped

					#endif

					// Have to make do without a size check here, since it could vary
					return ASURA_PCRC_CONTINUE;
				}

				// May want to support these formats in the future

				case WAVE_FORMAT_IEEE_FLOAT:
				case WAVE_FORMAT_EXTENSIBLE:
				default:
				{
					Asura_Assert( false, "Unrecognised audio format chunk." );
					return ASURA_PCRC_ERROR;
				}
			}
		}

		default:
		{
			bool bSkipped = (m_xStream.Skip( uChunkSize ) == uChunkSize);

			// Don't assert here, there are all kinds of different chunks pumped out by the various
			// sound programs and it should be safe to skip them all. Except for XMA2 chunks - these
			// are found only in old XMA2 files, we want to make sure they've all been converted

			if( uChunkType == u4CC_XMA2 )
			{
				Asura_Assert( false, "Found old XMA2 chunk in wave file." );
				bSkipped = false;
			}

			return bSkipped ? ASURA_PCRC_CONTINUE : ASURA_PCRC_ERROR;
		}
	}
}

#endif // defined ASURA_USE_AUDIO_SYSTEM
