//////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

// Asura Platform-independent
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_XAudio2.h"
#include "Sound/Asura_XAudio2_Buffer.h"
#include "Sound/Asura_XAudio2_Stream.h"
#include "Sound/Asura_XAudio2_Voice.h"

/******************************************************************************/
/* Asura_XAudio2_Voice Class Functions                                        */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::Asura_XAudio2_Voice()
//
// Constructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Voice::Asura_XAudio2_Voice()
	: m_pxSourceVoice( NULL )
	, m_pxSubmixVoice( NULL )

#ifdef ASURA_USE_XAUDIO_OCCLUSION
	, m_pxDirectPathSubmixVoice (NULL)
#endif

	, m_pfnBufferEndCallback( NULL )
	, m_iStreamHandle( ASURA_INVALID_HANDLE )
	, m_iPendingDestructionHandle( ASURA_INVALID_HANDLE )
	, m_iSamplesPlayedOffset( 0 )
	, m_bPlaying( false )
	, m_bSeekPending( false )
	, m_bPausedByPitch( false )
	, m_bPausedExternally( false )
	, m_bSubmixOutputs( false )
	, m_bExternalVoice( false )
	, m_bPlayAfterPrepared( false )
	, m_bPreparing( false )
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::~Asura_XAudio2_Voice()
//
// Destructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_XAudio2_Voice::~Asura_XAudio2_Voice()
{
	if( IsPreparing() )
	{
		Asura_Assert( false, "voice is preparing, cannot destroy yet" );
		return;
	}

	if( m_pxSourceVoice && !m_bExternalVoice )
	{
		m_pxSourceVoice->DestroyVoice();
	}

	if( m_pxSubmixVoice && !m_bExternalVoice )
	{
		m_pxSubmixVoice->DestroyVoice();
	}

#ifdef ASURA_USE_XAUDIO_OCCLUSION
	if (m_pxDirectPathSubmixVoice)
	{
		m_pxDirectPathSubmixVoice->DestroyVoice();
		m_pxDirectPathSubmixVoice = NULL;
	}
#endif

	Asura_XAudio2_Stream* const pxStream = Asura_XAudio2::GetStream( m_iStreamHandle );
	if( pxStream )
	{
		pxStream->Release();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::EnqueueWave()
//
// Adds an audio buffer to a voice's queue
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::EnqueueWave( const Asura_Hash_ID uWave )
{
	const Asura_XAudio2_Buffer* const pxBuffer = Asura_XAudio2::GetBuffer( uWave );
	if( !pxBuffer ) return;

	Asura_Assert( m_pxSourceVoice, "Submitting buffer to invalid audio voice." );
	if( !m_pxSourceVoice ) return;

	#if defined ASURA_ASSERT
		const HRESULT xResult =
	#endif
		m_pxSourceVoice->SubmitSourceBuffer( pxBuffer->GetParams(), pxBuffer->GetWMA() );

	Asura_Assert( SUCCEEDED( xResult ), "Failed to submit XAudio2 source buffer." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::EnqueueData()
//
// Enqueues arbitrary data on a voice. Specify whether this is the last packet
// in the audio stream, to prevent starvation warnings. You can also give a
// 'context' pointer that will be passed to callbacks when the buffer finishes
// being processed. Can't currently cope with xWMA format data.
//
// Ensure the data is not destroyed or overwritten before it's been played
// (check the context in the end-of-buffer callback)
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::EnqueueData( void* const pData, const u_int uDataSize,
	const bool bFinalPacket, void* const pContext /*= NULL*/ )
{
	if( IsPreparing() )
	{
		Asura_Assert( false, "voice is preparing, cannot enqueue data" );
		return;
	}

	Asura_Assert( m_pxSourceVoice, "Submitting data to invalid audio voice." );
	if( !m_pxSourceVoice ) return;

	XAUDIO2_BUFFER xPacket = {0};

	xPacket.Flags = (bFinalPacket ? XAUDIO2_END_OF_STREAM : 0);
	xPacket.AudioBytes = uDataSize;

	#if defined NOVEMBER_2008_SDK
		xPacket.pAudioData = static_cast< BYTE* >( pData );
	#else
		xPacket.pAudioData = static_cast< const BYTE* >( pData );
	#endif

	xPacket.pContext = pContext;

	#if defined ASURA_ASSERT
		const HRESULT xResult =
	#endif
		m_pxSourceVoice->SubmitSourceBuffer( &xPacket );

	Asura_Assert( SUCCEEDED( xResult ), "Failed to submit data to XAudio2 voice." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetBufferEndCallback()
//
// Sets a callback to be executed when each queued buffer ends playing
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SetBufferEndCallback( BufferCallback* const pfnCallback )
{
	m_pfnBufferEndCallback = pfnCallback;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::Start()
//
// Starts a voice playing
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::Start()
{
	if( IsPreparing() )
	{
		if( !m_bPausedByPitch && !m_bPausedExternally )
		{
			m_bPlayAfterPrepared = true;
		}
	}
	else
	{
		Asura_Assert( m_pxSourceVoice, "Trying to start invalid audio voice." );
		if( !m_pxSourceVoice ) 
		{
			return;
		}

		SetupSubbusRouting();

		// Initialise spatialisation to avoid a pop
		Update( true );

		if( !m_bPausedByPitch && !m_bPausedExternally )
		{
			const HRESULT xResult = m_pxSourceVoice->Start(0);

			if( !SUCCEEDED( xResult ) )
			{
				Asura_Assert( false, "Failed to start XAudio2 source voice." );
				return;
			}
		}
	}

	m_bPlaying = true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::Stop()
//
// Stops a voice
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::Stop()
{
	if( IsPreparing() )
	{
		if( m_bPlayAfterPrepared )
		{
			m_bPlayAfterPrepared = false;
		}
		return;
	}
	
	Asura_Assert( m_pxSourceVoice, "Trying to start invalid audio voice." );
	if( !m_pxSourceVoice ) return;

	const HRESULT xResult = m_pxSourceVoice->Stop(0);

	if( SUCCEEDED( xResult ) )
	{
		m_bPlaying = false;
		m_iSamplesPlayedOffset = 0;
	}
	else
	{
		Asura_Assert( false, "Failed to stop XAudio2 source voice." );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::IsPlaying()
//
// Queries whether an XAudio2 voice is currently playing
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_Voice::IsPlaying() const
{
	if( IsPreparing() )
	{
		return m_bPlaying;
	}

	Asura_Assert( m_pxSourceVoice, "Trying to query invalid audio voice." );
	if( !m_pxSourceVoice ) return false;

	XAUDIO2_VOICE_STATE xState;
	m_pxSourceVoice->GetState( &xState );

	return ( m_bPlaying && xState.BuffersQueued > 0 );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetupSubbusRouting()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SetupSubbusRouting()
{
	// Each time a voice is started, set up its routing to submix voices. This means we can't
	// change this stuff on the fly, but that's fine. Don't do this if already set up submix outputs
	if( !m_bSubmixOutputs )
	{
		// Output to either the reverb voice that goes on for master DSP, or the one that doesn't
		IXAudio2Voice* const pxReverbVoice = IsUsingMasterDSP()
			? Asura_XAudio2::GetReverbVoiceWithDSP()
			: Asura_XAudio2::GetReverbVoiceNoDSP();

		// API changes here as of March 2009 SDK, to allow per-send filtering. This can go once all
		// projects have upgraded

		#if defined MARCH_2009_SDK

			XAUDIO2_SEND_DESCRIPTOR apxOutputs[] =
			#ifdef ASURA_USE_XAUDIO_OCCLUSION
			{ { 0, m_pxDirectPathSubmixVoice }, { 0, pxReverbVoice } };
			#else
			{ { 0, Asura_XAudio2::GetMasteringVoice() }, { 0, pxReverbVoice } };
			#endif

		#else

			IXAudio2Voice* apxOutputs[] =
			#ifdef ASURA_USE_XAUDIO_OCCLUSION
			{ m_pxDirectPathSubmixVoice, pxReverbVoice };
			#else
			{ Asura_XAudio2::GetMasteringVoice(), pxReverbVoice };
			#endif

		#endif

		XAUDIO2_VOICE_SENDS xSends;

		#if defined MARCH_2009_SDK
			xSends.SendCount = (IsUsingReverb() && pxReverbVoice ? 2 : 1);
			xSends.pSends = apxOutputs;
		#else
			xSends.OutputCount = (IsUsingReverb() && pxReverbVoice ? 2 : 1);
			xSends.pOutputVoices = apxOutputs;
		#endif

		IXAudio2Voice* pxVoice;
		if (m_pxSourceVoice)
		{
			pxVoice = m_pxSourceVoice;
		}
		else
		{
			pxVoice = m_pxSubmixVoice;
		}
		Asura_Assert( pxVoice, "No voice?" );
		if( pxVoice )
		{
			#if defined ASURA_ASSERT
			HRESULT xResult =
			#endif
			pxVoice->SetOutputVoices( &xSends );
			Asura_Assert( SUCCEEDED( xResult ), "Failed to set outputs for XAudio2 source/submix voice." );
		}

		#ifdef ASURA_USE_XAUDIO_OCCLUSION
		// Connect the occlusion submix to the mastering voice

		#if defined MARCH_2009_SDK

			if( IsUsingMasterDSP() )
			{
				// Output to both DSP voices. Their volumes are updated based on the DSP crossfade
				// timer

				for( u_int u = 0; u < 2; ++u )
				{
					apxOutputs[u].Flags = 0;
					apxOutputs[u].pOutputVoice = Asura_XAudio2::GetDSPVoice(u);
				}

				xSends.SendCount = 2;
			}
			else
			{
				// In this case, output straight to the mastering voice

				apxOutputs[0].Flags = 0;
				apxOutputs[0].pOutputVoice = Asura_XAudio2::GetMasteringVoice();
				xSends.SendCount = 1;
			}

			xSends.pSends = apxOutputs;

		#else // !defined MARCH_2009_SDK

			if( IsUsingMasterDSP() )
			{
				for( u_int u = 0; u < 2; ++u )
				{
					apxOutputs[u] = Asura_XAudio2::GetDSPVoice(u);
				}

				xSends.OutputCount = 2;
			}
			else
			{
				apxOutputs[0] = Asura_XAudio2::GetMasteringVoice();
				xSends.OutputCount = 1;
			}

			xSends.pOutputVoices = apxOutputs;

		#endif // !defined MARCH_2009_SDK

		#if defined ASURA_ASSERT
		HRESULT xResult =
		#endif
		m_pxDirectPathSubmixVoice->SetOutputVoices( &xSends );
		Asura_Assert( SUCCEEDED( xResult ), "Failed to set outputs for XAudio2 occlusion submix voice." );

		//these values are acoustically equivalent to the filter being fully bypassed. Ie, it's switched off to start with
		XAUDIO2_FILTER_PARAMETERS xParams;
		xParams.Type		= LowPassFilter;
		xParams.Frequency	= 1.0f;
		xParams.OneOverQ	= 1.0f;

		if( m_pxDirectPathSubmixVoice )
		{
			#ifdef ASURA_ASSERT
			HRESULT iResult =
			#endif
			m_pxDirectPathSubmixVoice->SetFilterParameters( &xParams );
			Asura_Assert( SUCCEEDED( iResult ), "Failed to set filer parameters" );
		}
		#endif	// ASURA_USE_XAUDIO_OCCLUSION
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetMixerChannel()
//
// Sets the mixer channel through which an audio voice should be routed
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SetMixerChannel( const Asura_Hash_ID uChannelHash )
{
	// For now, do nothing. We're not going to have different DSP effects on each mixer channel
	// any more. This can be removed once the dust has settled

	//// Route from this voice into the mixer channel's submix voice
	//IXAudio2Voice* const pxSubmixVoice = Asura_XAudio2::GetMixerChannel( uChannelHash );
	//IXAudio2Voice* const pxMasteringVoice = Asura_XAudio2::GetMasteringVoice();

	//IXAudio2Voice* apxOutputVoices[1] = { pxSubmixVoice ? pxSubmixVoice : pxMasteringVoice };

	//XAUDIO2_VOICE_SENDS xSends;
	//xSends.OutputCount = 1;
	//xSends.pOutputVoices = apxOutputVoices;

	//#if defined ASURA_ASSERT
	//	const HRESULT xResult =
	//#endif
	//	m_pxSourceVoice->SetOutputVoices( &xSends );

	//Asura_Assert( SUCCEEDED( xResult ), "Failed to set XAudio2 voice send list." );

	//// Need to store the output voice for subsequent calls to SetOutputMatrix(), because passing
	//// NULL to that only works if you've not explicitly set output mappings
	//m_pxDestVoice = apxOutputVoices[0];
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SeekAtPlaybackTime()
//
// Sets up a seek at a specified time in the future. Only valid for streaming
// voices. The times are measured in seconds from the start of the file
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SeekAtPlaybackTime( const float fSeekPosition, const float fPlaybackTime )
{
	Asura_Assert( !IsPreparing(), "Cannot seek at playback time, preparing" );

	Asura_Assert( m_iStreamHandle != ASURA_INVALID_HANDLE,
		"SeekAtPlaybackTime() called for non-streaming voice." );

	Asura_XAudio2_Stream* const pxStream = Asura_XAudio2::GetStream( m_iStreamHandle );

	if( pxStream )
	{
		if( pxStream->SeekAtPlaybackTime( fSeekPosition, fPlaybackTime ) )
		{
			m_bSeekPending = true;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SeekIsPending()
//
// Queries whether a seek is pending on the voice. Only for streaming sounds
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_Voice::SeekIsPending()
{
	Asura_Assert( m_iStreamHandle != ASURA_INVALID_HANDLE,
		"SeekIsPending() called for non-streaming voice." );

	return m_bSeekPending;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::Update()
//
// Once per frame update for an XAudio2 voice
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::Update( const bool bFirstUpdate )
{
	if( IsPreparing() )
	{
		// Nothing to do
		return;
	}

	Asura_Audio_Voice::Update( bFirstUpdate );
	IXAudio2Voice* pxVoice;
	if (m_pxSourceVoice)
	{
		pxVoice = m_pxSourceVoice;
	}
	else
	{
		pxVoice = m_pxSubmixVoice;
	}

	if( !pxVoice ) 
	{
		// This can happen if we had an error creating a voice
		return;
	}

	{
		// Deal with starting the voice when we are waiting for the stream to be prepared
		if( m_bPlayAfterPrepared )
		{
			m_bPlayAfterPrepared = false;

			SetupSubbusRouting();

			// Initialise spatialisation to avoid a pop
			Update( true );

			#if defined ASURA_ASSERT
				const HRESULT xResult =
			#endif
				m_pxSourceVoice->Start(0);

			Asura_Assert( SUCCEEDED( xResult ), "Failed to start source voice" );
		}
	}


	// Set overall volume
	{
		float fVol = GetVolume();

		Asura_Assertf( -XAUDIO2_MAX_VOLUME_LEVEL <= fVol && fVol <= XAUDIO2_MAX_VOLUME_LEVEL,
			("Voice volume (%f) out of range.", fVol) );

		fVol =
			Asura_Maths::ClampToRange( fVol, -XAUDIO2_MAX_VOLUME_LEVEL, XAUDIO2_MAX_VOLUME_LEVEL );

		#if defined ASURA_ASSERT
			const HRESULT xResult =
		#endif
			pxVoice->SetVolume( fVol );

		Asura_Assert( SUCCEEDED( xResult ), "Failed to set XAudio2 voice volume." );
	}

	XAUDIO2_VOICE_DETAILS xVoiceDetails;
	pxVoice->GetVoiceDetails(&xVoiceDetails);

	// Set pitch, including game timescale and doppler factors.
	// We can only do this if we have pitch control on the source voice

	if (m_pxSourceVoice)
	{
		if (!(xVoiceDetails.CreationFlags & XAUDIO2_VOICE_NOPITCH))
		{
			float fPitch = GetPitch();

			if( fPitch < XAUDIO2_MIN_FREQ_RATIO )
			{
				if( !IsPausedByPitch() )
				{
					SetPausedByPitch();
				}
			}
			else
			{
				Asura_Assertf( fPitch <= XAUDIO2_MAX_FREQ_RATIO, ("Voice pitch (%f) out of range.", fPitch) );

				if( IsPausedByPitch() )
				{
					SetPausedByPitch( false );
				}
				
				fPitch = Asura_Maths::ClampToRange( fPitch, XAUDIO2_MIN_FREQ_RATIO, XAUDIO2_MAX_FREQ_RATIO );

				#if defined ASURA_ASSERT
					const HRESULT xResult =
				#endif
					m_pxSourceVoice->SetFrequencyRatio( fPitch );

				Asura_Assert( SUCCEEDED( xResult ), "Failed to set XAudio2 voice pitch." );
			}
		}
	}

	UpdateFilters ();

	// Can only do this for mono voices. Multichannel sounds should just route through with
	// their original mappings
	// (Don't do this if we have submix destinations)
	if( (xVoiceDetails.InputChannels == 1) && !m_bSubmixOutputs )
	{
		// Set output matrix coefficients based on spatialisation
		{
			const u_int uChannelMask = Asura_XAudio2::GetDeviceChannelMask();
			u_int uXAudio2Speaker = 0;

			static float afMatrix[32];

			for( u_int u = 0; u < 32; ++u )
			{
				if( uChannelMask & (1 << u) )
				{
					const Asura_Audio_System::Speakers eSpeaker =
						Asura_XAudio2::GetIndependentSpeaker(u);

					afMatrix[ uXAudio2Speaker++ ] = GetSpeakerAmplitude( eSpeaker );
				}
			}

			IXAudio2Voice* pxThisVoice = pxVoice;
#ifdef ASURA_USE_XAUDIO_OCCLUSION
			// if we've got a submix on the direct path, then do the spacialisation there
			if (m_pxDirectPathSubmixVoice)
			{
				pxThisVoice = m_pxDirectPathSubmixVoice;
			}
#endif

			HRESULT xResult;
			const u_int uDeviceNumChannels = Asura_XAudio2::GetDeviceNumChannels();

			if( IsUsingMasterDSP() )
			{
				if( IsPlaying() || bFirstUpdate )
				{
					// Sigh. XAudio2 crashes if you set the output matrix to a voice which hasn't
					// yet been registered as an output, instead of just returning an error. This
					// happens if a voice is being updated that hasn't started playing yet, since
					// that's when we set up the routing. So we can't call this till that's done

					for( u_int u = 0; u < 2; ++u )
					{
						xResult = pxThisVoice->SetOutputMatrix( Asura_XAudio2::GetDSPVoice(u), 1,
							uDeviceNumChannels, afMatrix );

						Asura_Assert( SUCCEEDED( xResult ),
							"Failed to set voice DSP output matrix." );
					}
				}
			}
			else
			{
				xResult = pxThisVoice->SetOutputMatrix( Asura_XAudio2::GetMasteringVoice(), 1,
					uDeviceNumChannels, afMatrix );

				Asura_Assert( SUCCEEDED( xResult ), "Failed to set voice master output matrix." );
			}
		}

		IXAudio2SubmixVoice* const pxReverbVoice = IsUsingMasterDSP()
			? Asura_XAudio2::GetReverbVoiceWithDSP()
			: Asura_XAudio2::GetReverbVoiceNoDSP();

		// If using reverb, set send level to the reverb submix voice
		if( IsUsingReverb() && pxReverbVoice && (m_bPlaying || bFirstUpdate) )
		{
			// Ensure we're actually outputting to the reverb voice, which will only be set up
			// once the voice has started. Can use m_bPlaying for this

			const float fReverbLevel = GetReverbSendLevel();

			#if defined ASURA_ASSERT
				const HRESULT xResult =
			#endif
				pxVoice->SetOutputMatrix( pxReverbVoice, 1, 1, &fReverbLevel );

			Asura_Assert( SUCCEEDED( xResult ), "Failed to set voice reverb output matrix." );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::GetPlayPosition()
//
// Queries the current play cursor position for a voice. Only supported on
// streaming sounds for the moment
//
////////////////////////////////////////////////////////////////////////////////

float Asura_XAudio2_Voice::GetPlayPosition() const
{
	if( IsPreparing() )
	{
		return 0.0f;
	}

	Asura_Assert( m_pxSourceVoice, "Calling GetPlayPosition() for invalid audio voice." );
	if( !m_pxSourceVoice ) return 0.0f;

	Asura_Assert( m_iStreamHandle != ASURA_INVALID_HANDLE,
		"GetPlayPosition() not supported for non-streaming voices yet." );

	Asura_XAudio2_Stream* const pxStream = Asura_XAudio2::GetStream( m_iStreamHandle );

	if( !pxStream )
	{
		return 0.0f;
	}

	XAUDIO2_VOICE_STATE xState;
	m_pxSourceVoice->GetState( &xState );

	UINT64 uSamplesPlayed = xState.SamplesPlayed + m_iSamplesPlayedOffset;

	// NOTE: I'm not 100% sure if the samples specified here are the number of decoded samples or the 
	// number of samples output by xaudio, I have assumed the former
	const float fPlayPos = static_cast<float>( uSamplesPlayed ) / static_cast<float>( pxStream->GetSampleRate() );

	return fPlayPos;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::UpdateFilters()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::UpdateFilters()
{
	// Update the filter on the voice (used for distance + occlusion filtering )
	if( m_pxSourceVoice && ( IsUsingDistanceFiltering() || IsUsingOcclusion() ) )
	{
		XAUDIO2_VOICE_DETAILS xVoiceDetails;
		m_pxSourceVoice->GetVoiceDetails(&xVoiceDetails);

		if( !( xVoiceDetails.CreationFlags & XAUDIO2_VOICE_USEFILTER ) )
		{
			Asura_Assert (xVoiceDetails.CreationFlags & XAUDIO2_VOICE_USEFILTER, "Filter effects not allowed on this voice");
			return;
		}
		
		float fVoiceFilterFactor = 0.0f;
		
		if( IsUsingDistanceFiltering() )
		{
			fVoiceFilterFactor += GetDistanceFilterFactor();
		}

		#ifdef ASURA_USE_XAUDIO_OCCLUSION
		if ( IsUsingOcclusion() )
		{
			fVoiceFilterFactor += GetOcclusionFactor();
		}
		#endif

		// To my ear, this most closely approximates a linear attenuation
		fVoiceFilterFactor = Asura_Maths::Power( fVoiceFilterFactor, 0.3f );	//Maybe replace this with a lookup table
		
		XAUDIO2_FILTER_PARAMETERS xVoiceFilterParams;
		xVoiceFilterParams.Type			= LowPassFilter;
		xVoiceFilterParams.Frequency	= ( 1.0f - fVoiceFilterFactor) * 0.95f + 0.05f;
		xVoiceFilterParams.OneOverQ		= 1.414213f;	//Apparently 1/root(2) approximates a butterworth filter

		xVoiceFilterParams.Frequency = Asura_Maths::ClampToRange( xVoiceFilterParams.Frequency, 0.0f, XAUDIO2_MAX_FILTER_FREQUENCY );

		#ifdef ASURA_ASSERT
		HRESULT iResult = m_pxSourceVoice->SetFilterParameters( &xVoiceFilterParams );
		Asura_Assert( SUCCEEDED( iResult ), "Failed to set filer parameters" );
		#else
		m_pxSourceVoice->SetFilterParameters( &xVoiceFilterParams );
		#endif
	}

	// Update the direct path submix filter (used for obstruction)
	#ifdef ASURA_USE_XAUDIO_OCCLUSION
	if( m_pxDirectPathSubmixVoice && IsUsingOcclusion() )
	{
		float fSubmixFilterFactor = GetObstructionFactor();

		// To my ear, this most closely approximates a linear attenuation
		fSubmixFilterFactor = Asura_Maths::Power( fSubmixFilterFactor, 0.3f );	//Maybe replace this with a lookup table

		XAUDIO2_FILTER_PARAMETERS xSubmixFilterParams;
		xSubmixFilterParams.Type		= LowPassFilter;
		xSubmixFilterParams.Frequency	= ( 1.0f - fSubmixFilterFactor) * 0.95f + 0.05f;
		xSubmixFilterParams.OneOverQ	= 1.414213f;	//Apparently 1/root(2) approximates a butterworth filter

		xSubmixFilterParams.Frequency = Asura_Maths::ClampToRange( xSubmixFilterParams.Frequency, 0.0f, XAUDIO2_MAX_FILTER_FREQUENCY );
		
		#ifdef ASURA_ASSERT
		HRESULT iResult =
		#endif
		m_pxDirectPathSubmixVoice->SetFilterParameters( &xSubmixFilterParams );
		Asura_Assert( SUCCEEDED( iResult ), "Failed to set filer parameters" );
	}
	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::OnBufferEnd()
//
// Called by XAudio2 when a buffer has finished on a voice. Enqueues a read
// on the stream, if this is a streaming voice
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::OnBufferEnd( void* pBufferContext )
{
	Asura_XAudio2_Stream* const pxStream = Asura_XAudio2::GetStream( m_iStreamHandle );
	if( pxStream )
	{
		pxStream->EnqueueRead();
	}

	if( m_pfnBufferEndCallback )
	{
		m_pfnBufferEndCallback( pBufferContext );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::OnBufferStart()
//
// Called by XAudio2 when a buffer has started on a voice. We need to update
// our play position offset here
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::OnBufferStart( void* pBufferContext )
{
	if( m_iStreamHandle != ASURA_INVALID_HANDLE )
	{
		// The buffer context contains the offset in seconds of this buffer start time from the previous buffer end time
		const int iSamplesOffset = reinterpret_cast<int>( pBufferContext );
		m_iSamplesPlayedOffset += iSamplesOffset;

		if( m_bSeekPending && ( iSamplesOffset != 0 ) )
		{
			// We are playing a buffer which seeked, so the seek is no longer pending
			m_bSeekPending = false;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetStartOffset()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SetStartOffset ( const float fOffsetTime)
{
	Asura_Assert( m_iStreamHandle != ASURA_INVALID_HANDLE, "Expected stream to be created at this point" );
	
	Asura_XAudio2_Stream* const pxStream = Asura_XAudio2::GetStream( m_iStreamHandle );

	if( pxStream )
	{
		m_iSamplesPlayedOffset = static_cast<int>( fOffsetTime * pxStream->GetSampleRate() );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetSubmixOutputs()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_Voice::SetSubmixOutputs( const Asura_Handle* piSubmixVoices,
	const u_int uNumVoices )
{
	Asura_Assert( m_pxSourceVoice, "Calling SetSubmixOutputs() for invalid voice." );
	if( !m_pxSourceVoice ) return false;

	// SetOutputVoices() mustn't be called with 0 entries
	Asura_Assert( uNumVoices > 0, "Invalid number of submix voices." );
	if( uNumVoices == 0 ) return false;

	// Gather the submix pointers into an array

	#if defined MARCH_2009_SDK
		XAUDIO2_SEND_DESCRIPTOR* const pxDescriptors = new XAUDIO2_SEND_DESCRIPTOR[ uNumVoices ];
		if( !pxDescriptors ) return false;
	#else
		IXAudio2Voice** const ppxVoices = new IXAudio2Voice*[ uNumVoices ];
		if( !ppxVoices ) return false;
	#endif

	for( u_int u = 0; u < uNumVoices; ++u )
	{
		Asura_XAudio2_Voice* const pxAsuraVoice = Asura_XAudio2::GetVoice( piSubmixVoices[u] );

		// SetOutputVoices() mustn't be called with NULL entries in the list
		if( !pxAsuraVoice || !pxAsuraVoice->m_pxSubmixVoice )
		{
			Asura_Assertf( false, ("Failed to get submix voice %i.", piSubmixVoices[u]) );

			#if defined MARCH_2009_SDK
				delete[] pxDescriptors;
			#else
				delete[] ppxVoices;
			#endif

			return false;
		}

		#if defined MARCH_2009_SDK
			pxDescriptors[u].Flags = 0;
			pxDescriptors[u].pOutputVoice = pxAsuraVoice->m_pxSubmixVoice;
		#else
			ppxVoices[u] = pxAsuraVoice->m_pxSubmixVoice;
		#endif
	}

	// Set the send list

	#if defined MARCH_2009_SDK
		const XAUDIO2_VOICE_SENDS xSends = { uNumVoices, pxDescriptors };
	#else
		const XAUDIO2_VOICE_SENDS xSends = { uNumVoices, ppxVoices };
	#endif

	const HRESULT xResult = m_pxSourceVoice->SetOutputVoices( &xSends );

	Asura_Assertf( SUCCEEDED( xResult ), ("Failed to set submix outputs for voice playing '%s'.",
		GetFilename()) );

	#if defined MARCH_2009_SDK
		delete[] pxDescriptors;
	#else
		delete[] ppxVoices;
	#endif

	if( SUCCEEDED( xResult ) )
	{
		// Prevent reassigning the outputs when we start playing
		m_bSubmixOutputs = true;
		return true;
	}
	else
	{
		return false;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::ClearSubmixOutputs()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::ClearSubmixOutputs ()
{
	// Clear the flag, so that when we're started our outputs will be reassigned to defaults
	m_bSubmixOutputs = false;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetPausedByPitch()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SetPausedByPitch( bool bPaused )
{
	// Prevent warning about signed/unsigned mismatch now we're using a bitfield
	const bool bPausedByPitch = m_bPausedByPitch;
	if( bPausedByPitch == bPaused ) return;

	const bool bWasPaused = m_bPausedExternally || m_bPausedByPitch;

	m_bPausedByPitch = bPaused;

	const bool bCurrentlyPaused = m_bPausedExternally || m_bPausedByPitch;

	if( m_pxSourceVoice )
	{
		if( !bWasPaused && bCurrentlyPaused && IsPlaying() )
		{
			m_pxSourceVoice->Stop(0);
		}
		else if( bWasPaused && !bCurrentlyPaused )
		{
			m_pxSourceVoice->Start(0);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::SetPausedExternally()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_XAudio2_Voice::SetPausedExternally( bool bPaused )
{
	// Prevent warning about signed/unsigned mismatch now we're using a bitfield
	const bool bPausedExternally = m_bPausedExternally;
	if( bPausedExternally == bPaused ) return;

	const bool bWasPaused = m_bPausedExternally || m_bPausedByPitch;

	m_bPausedExternally = bPaused;

	const bool bCurrentlyPaused = m_bPausedExternally || m_bPausedByPitch;

	if( m_pxSourceVoice )
	{
		if( !bWasPaused && bCurrentlyPaused && IsPlaying() )
		{
			m_pxSourceVoice->Stop(0);
		}
		else if( bWasPaused && !bCurrentlyPaused )
		{
			m_pxSourceVoice->Start(0);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_XAudio2_Voice::IsPreparing()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_XAudio2_Voice::IsPreparing() const
{
	return m_bPreparing;
}

#endif // defined ASURA_USE_AUDIO_SYSTEM

//eof
