// Asura_XAPO_DSP_SimpleFilter.cpp
// Thomas Young - 31.03.09
// Copyright (C) 2009 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_AUDIO_SYSTEM

#include "Sound/XAPO_DSP/Asura_XAPO_DSP_SimpleFilter.h"
#include "Sound/Asura_Audio_DSP.h"

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

Asura_XAPO_DSP_SimpleFilter::Asura_XAPO_DSP_SimpleFilter( ) : Asura_XAPOBase< Asura_XAPO_DSP_SimpleFilter, Asura_DSP_Params_SimpleFilter::Params >( )
{
	m_pfTempBuffer = 0;
	m_uExpectedChans = 0;
	m_uExpectedFrames = 0;
}


Asura_XAPO_DSP_SimpleFilter::~Asura_XAPO_DSP_SimpleFilter()
{
	if( m_pfTempBuffer )
	{
		Asura_DeallocateAlign( m_pfTempBuffer );
	}
}


void Asura_XAPO_DSP_SimpleFilter::DoProcess ( float* pfInputBuffer, u_int uFrames, u_int uChannels, const Asura_DSP_Params_SimpleFilter::Params& xParams )
{
	if( !m_pfTempBuffer )
	{
		// We need to create a temp buffer, so we make some assumptions about the buffer format based on this first call
		m_uExpectedChans = uChannels;
		m_uExpectedFrames = uFrames;

		// Alloc size is the size of the buffer xaudio gives us ( sizeof float * num Frames * num channels ) plus the extra we need
		// to store out of the previous buffer for the convo ( sizeof float * num channels * num coefficients )
		const u_int uAlocSize = ( uChannels * 2 * uFrames + Asura_DSP_Params_SimpleFilter::uFILTER_NUM_COEFFICIENTS ) * sizeof( float );

		m_pfTempBuffer = reinterpret_cast<float*>( Asura_AllocateAlign( uAlocSize, ASURA_ALIGN_128_BIT ) );

		// Zero it since we will be using it immediately
		Asura_Mem::Set( m_pfTempBuffer, 0, uAlocSize );
	}
	
	if( ( m_uExpectedChans != uChannels ) || ( m_uExpectedFrames != uFrames ) )
	{
		return;
	}

	// Copy the end of the previous buffer to the start of our temp buffer, and copy the new data into the temp buffer after it
	// So that we have || end of last buffer | new buffer ||
	const u_int uPadSize = uChannels * Asura_DSP_Params_SimpleFilter::uFILTER_NUM_COEFFICIENTS;
	const u_int uPadLoc = uChannels * uFrames;

	Asura_Mem::Move( &m_pfTempBuffer[0], &m_pfTempBuffer[ uPadLoc ], uPadSize * sizeof( float ) );				// Copy bit of old buffer
	Asura_Mem::Copy( &m_pfTempBuffer[ uPadSize ], pfInputBuffer, uChannels * uFrames * sizeof( float ) );
	
	Asura_Mem::Set( pfInputBuffer, 0, uChannels * uFrames * sizeof( float ) );

	// Convolve

	//		|1|2|3|			- Original Signal	( pfTempBuffer )
	//	|4|5|6|	->			- Coefficients		( xParams.m_afCoefficients )
	//		|x|y|z|			- Result			( pfResult )

	float* pfResult = pfInputBuffer;	// Just to make the code less confusing
	
	for( int iResultPos = 0; iResultPos < static_cast<int>( uFrames ); ++iResultPos )
	{
		const int iResultFramePos = iResultPos * static_cast<int>( uChannels );

		// loop over coefficients and multiply-add to the result
		for( int iCoeff = 0; iCoeff < Asura_DSP_Params_SimpleFilter::uFILTER_NUM_COEFFICIENTS; ++iCoeff )
		{
			const int iSourceFramePos = ( iResultPos + iCoeff ) * static_cast<int>( uChannels );

			// for all channels in this frame
			for( int iChan = 0; iChan < static_cast<int>( uChannels ); ++iChan )
			{
				const int iResultSamplePos = iResultFramePos + iChan;
				const int iSourceSamplePos = iSourceFramePos + iChan;
			
				pfResult[ iResultSamplePos ] += m_pfTempBuffer[ iSourceSamplePos ] * xParams.m_afCoefficients[ iCoeff ];
			}
		}

		// Apply the gain adjustment afterwards
		for( int iChan = 0; iChan < static_cast<int>( uChannels ); ++iChan )
		{
			const int iResultSamplePos = iResultFramePos + iChan;
			
			pfResult[ iResultSamplePos ] *= xParams.m_fGain;

			// Clip if necessary
			if( pfResult[ iResultSamplePos ] > 1.0f )
			{
				pfResult[ iResultSamplePos ] = 1.0f;
			}
			else if( pfResult[ iResultSamplePos ] < -1.0f )
			{
				pfResult[ iResultSamplePos ] = -1.0f;
			}
		}
	}
}


#endif

//eof
