//////////////////////////////////////////////////////////////////////////////
//
// Asura_3dDecal_System.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     3d objects (e.g. small craters) placed as decals
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// JAT 15/03/06 to get memset for Asura_Shapes.h
#include <string.h>

// This file's header.
#include "SpecialFX/Asura_3dDecal_System.h"

// Asura includes
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Dynamics/Asura_Dynamics.h"
#include "Rendering/Asura_SortedRender.h"
#include "Shapes/Asura_Shapes.h"
#include "Shapes/Asura_Shape_System.h"

// internal defines
#ifndef ASURA_MAX_NO_OF_3D_DECALS
#define ASURA_MAX_NO_OF_3D_DECALS 64
#endif

DEBUGVAR float dbg_fMinDecalCameraDist = 10.0f;
DEBUGVAR float dbg_fMaxDecalCameraDist = 100.0f;
DEBUGVAR float dbg_fMinDecalMovement = 0.005f;
DEBUGVAR float dbg_fMaxDecalMovement = 0.15f;

// internal classes
class Asura_3dDecal
{

	public:
	
		Asura_3dDecal()
		{
			// init the shape instances - SetStatic allocates memory so I want to get that out of the way
			// RIP SetStatic()... but the flags still need to be set I guess. RAR 12/05/2005
			Asura_Shape_Instance* apxShapeInstances[3] = { &m_xShapeInstanceMain, &m_xShapeInstanceFace, &m_xShapeInstanceBlend };
			for (u_int u = 0; u < 3; ++u)
			{
				Asura_Shape_Instance& xShapeInstance = *apxShapeInstances[u];
				xShapeInstance.AddFlags(ASURA_SHAPE_INSTANCE_NO_SHADOW);
			}
		}

		void Init(const Asura_3dDecal_Template* pxTemplate, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri)
		{
			Asura_Hash_ID auShapes[3] = { pxTemplate->m_uShapeHashID, pxTemplate->m_uTopFaceHashID, pxTemplate->m_uBlendHashID[0] };
			Asura_Shape_Instance* apxShapeInstances[3] = { &m_xShapeInstanceMain, &m_xShapeInstanceFace, &m_xShapeInstanceBlend };
			for (u_int u = 0; u < 3; ++u)
			{
				Asura_Shape_Instance& xShapeInstance = *apxShapeInstances[u];

				xShapeInstance.SetShape(auShapes[u]);
				Asura_Assert(xShapeInstance.GetShape(), "shape for 3d decal not found");

				// force lighting update
				xShapeInstance.InvalidateLighting();

				xShapeInstance.SetNewPosAndOri(xPos, xOri);

				xShapeInstance.ResetLastKnownModuleID();
			}

			m_pxDecalTemplate = pxTemplate;
			m_fTimer = 0.0f;
			m_xBasePos = xPos;
			m_xBaseOri = xOri;

			const Asura_Vector_3* pxCameraPos = 0;
			if (Asura_Camera::GetNumberOfCameras() == 1)
			{
				pxCameraPos = Asura_Camera::GetPosition();
			}
			Update(pxCameraPos);
		}

		void Reset()
		{
			Asura_Shape_Instance* apxShapeInstances[3] = { &m_xShapeInstanceMain, &m_xShapeInstanceFace, &m_xShapeInstanceBlend };
			for (u_int u = 0; u < 3; ++u)
			{
				Asura_Shape_Instance& xShapeInstance = *apxShapeInstances[u];

				xShapeInstance.SetShape(ASURA_HASH_ID_UNSET);
			}
		}

		// SetPos() only intended for small offsets to fix sorting issues
		void SetPos(const Asura_Vector_3& xOffset)
		{
			Asura_Vector_3 xNewPos = m_xBasePos + xOffset;
			Asura_Shape_Instance* apxShapeInstances[2] = { &m_xShapeInstanceMain, &m_xShapeInstanceFace };
			for (u_int u = 0; u < 2; ++u)
			{
				Asura_Shape_Instance& xShapeInstance = *apxShapeInstances[u];
				xShapeInstance.SetNewPosAndOri( xNewPos, xShapeInstance.GetOrientation() );
			}
			// top layer is moved a bit more.
			m_xShapeInstanceBlend.SetNewPosAndOri( xNewPos + xOffset, m_xShapeInstanceBlend.GetOrientation() );
		}

		// SetOri() intended for changing scale of face and main shapes (for grow effect, used on acid holes)
		void SetOri(const Asura_Matrix_3x3& xOri)
		{
			for (u_int u = 0; u < 2; ++u)
			{
				Asura_Shape_Instance& xShapeInstance = u ? m_xShapeInstanceFace : m_xShapeInstanceMain;

				xShapeInstance.SetNewPosAndOri(xShapeInstance.GetPosition(), xOri);
			}
		}

		void Update(const Asura_Vector_3* pxCameraPos)
		{
			const Asura_3dDecal_Template* pxTemplate = m_pxDecalTemplate;
			
			// deal with special update modes
			if (pxTemplate->m_uFlags & ASURA_3DDECAL_GROW)
			{
				float fGrowTime = pxTemplate->m_fTimer;
				if (m_fTimer <= fGrowTime)
				{
					m_fTimer += Asura_Timers::GetGameTimeSlice();
					if (m_fTimer >= fGrowTime)
					{
						SetOri(m_xBaseOri);
					}
					else
					{
						float fScale = m_fTimer / fGrowTime;
						Asura_Matrix_3x3 xOri = m_xBaseOri * fScale;
						SetOri(xOri);
					}
				}

			}

			if (pxTemplate->m_uFlags & ASURA_3DDECAL_ANIMATEBLEND)
			{
				float fAnimTime = pxTemplate->m_fTimer;
				if (m_fTimer < fAnimTime)
				{
					float fFrameTime = fAnimTime / static_cast<float>(uASURA_3DDECAL_MAX_BLEND_SHAPES);
					u_int uFrame = Asura_Maths::FloatToInt(m_fTimer / fFrameTime);
					if (uFrame >= uASURA_3DDECAL_MAX_BLEND_SHAPES)
					{
						uFrame = uASURA_3DDECAL_MAX_BLEND_SHAPES - 1;
					}
					m_xShapeInstanceBlend.SetShape(pxTemplate->m_uBlendHashID[uFrame]);

					m_fTimer += Asura_Timers::GetGameTimeSlice();
				}
			}

			// update pos based on camera pos?
			if (pxCameraPos)
			{
				Asura_Vector_3 xDecalToCamera = *pxCameraPos - m_xBasePos;
				float fDistToCamera = xDecalToCamera.Magnitude();
				float fDistToMove;
				if (fDistToCamera < dbg_fMinDecalCameraDist)
				{
					fDistToMove = dbg_fMinDecalMovement;
				}
				else if (fDistToCamera < dbg_fMaxDecalCameraDist)
				{
					fDistToMove = dbg_fMinDecalMovement + 
						(dbg_fMaxDecalMovement * (fDistToCamera - dbg_fMinDecalCameraDist) / (dbg_fMaxDecalCameraDist - dbg_fMinDecalCameraDist));
				}
				else 
				{
					fDistToMove = dbg_fMaxDecalMovement;
				}
				xDecalToCamera *= fDistToMove / fDistToCamera;
				SetPos( xDecalToCamera );
			}
		}

		// member data
		Asura_Shape_Instance m_xShapeInstanceMain;
		Asura_Shape_Instance m_xShapeInstanceFace;
		Asura_Shape_Instance m_xShapeInstanceBlend;
		const Asura_3dDecal_Template* m_pxDecalTemplate;
		float m_fTimer;
		Asura_Vector_3 m_xBasePos;
		Asura_Matrix_3x3 m_xBaseOri;
};


// internal globals
static u_int g_u3dDecalCounter = 0;
static Asura_3dDecal* g_px3dDecalArray = NULL;

// debugvars
DEBUGVAR float dbg_f3dDecalOffset = 0.005f;	// placed this far from surface to avoid Z-fighting
DEBUGVAR bool dbg_ab3dDecalPasses[3] = {true, true, true};
DEBUGVAR float dbg_f3dDecalNormalThreshold = 0.1f;

void Asura_3dDecal_System::Initialise(void)
{
	Asura_Assert(!g_px3dDecalArray, "already initialised");

	g_u3dDecalCounter = 0;
	g_px3dDecalArray = new Asura_3dDecal[ASURA_MAX_NO_OF_3D_DECALS];
}

void Asura_3dDecal_System::Reset(void)
{
	u_int uNumDecals = g_u3dDecalCounter;
	if (uNumDecals > ASURA_MAX_NO_OF_3D_DECALS)
	{
		uNumDecals = ASURA_MAX_NO_OF_3D_DECALS;
	}

	for (u_int u = 0; u < uNumDecals; ++u)
	{
		g_px3dDecalArray[u].Reset();
	}
	
	g_u3dDecalCounter = 0;
}

void Asura_3dDecal_System::Shutdown(void)
{
	delete[] g_px3dDecalArray;
	g_px3dDecalArray = NULL;
	g_u3dDecalCounter = 0;
}

void Asura_3dDecal_System::Update(void)
{
	u_int uNumDecals = g_u3dDecalCounter;
	if (uNumDecals > ASURA_MAX_NO_OF_3D_DECALS)
	{
		uNumDecals = ASURA_MAX_NO_OF_3D_DECALS;
	}

	// pass the camera pos if we're not in split screen. You may want to disable 3d decals in split screen.
	const Asura_Vector_3* pxCameraPos = 0;
	if (Asura_Camera::GetNumberOfCameras() == 1)
	{
		pxCameraPos = Asura_Camera::GetPosition();
	}
	for (u_int u = 0; u < uNumDecals; ++u)
	{
		g_px3dDecalArray[u].Update(pxCameraPos);
	}
}

void Asura_3dDecal_System::EarlyRender(void)
{
	u_int uNumDecals = g_u3dDecalCounter;
	if (uNumDecals > ASURA_MAX_NO_OF_3D_DECALS)
	{
		uNumDecals = ASURA_MAX_NO_OF_3D_DECALS;
	}
	
	// render main shapes
	if (dbg_ab3dDecalPasses[0])
	{
		Asura_SortedRender::Start();

		for (u_int u = 0; u < uNumDecals; ++u)
		{
			g_px3dDecalArray[u].m_xShapeInstanceMain.Render();
		}

		Asura_SortedRender::End();
	}

	// render top face polys
	if (dbg_ab3dDecalPasses[1])
	{
		Asura_SortedRender::Start();

		for (u_int u = 0; u < uNumDecals; ++u)
		{
			g_px3dDecalArray[u].m_xShapeInstanceFace.Render();
		}

		Asura_SortedRender::End();
	}
}

void Asura_3dDecal_System::LateRender(void)
{
	u_int uNumDecals = g_u3dDecalCounter;
	if (uNumDecals > ASURA_MAX_NO_OF_3D_DECALS)
	{
		uNumDecals = ASURA_MAX_NO_OF_3D_DECALS;
	}
	
	// render blend shapes
	if (dbg_ab3dDecalPasses[2])
	{
		Asura_SortedRender::Start();

		for (u_int u = 0; u < uNumDecals; ++u)
		{
			g_px3dDecalArray[u].m_xShapeInstanceBlend.Render();
		}

		Asura_SortedRender::End();
	}
}

bool Asura_3dDecal_System::AddDecal(const Asura_Vector_3& xPosition, 
									const Asura_Vector_3& xSurfaceNormal, 
									u_int uTemplateID, 
									float fRotation /* = 0.0f */,
									Asura_Vector_3 xScale /* = Asura_Vector_3(1.0f, 1.0f, 1.0f)*/)
{
	const Asura_3dDecal_Template* pxTemplate = Project_GetTemplate(uTemplateID);
	Asura_Assertf(pxTemplate, ("no decal template with ID %u", uTemplateID));
	if (!pxTemplate) return false;

	// move pos a bit along normal
	// now done in update().
	Asura_Vector_3 xBasePos(xPosition);
//	Asura_Vector_3 xOffset(xSurfaceNormal * dbg_f3dDecalOffset);
//	xBasePos += xOffset;
	
	// create orientation from surface normal
	Asura_Matrix_3x3 xOrientation;
	Asura_Maths::OrientationFromDirection(xOrientation, xSurfaceNormal);

	// rotate around normal
	if (fRotation!=0.0f)
	{
		fRotation /= 2.0f;
		float fSine = Asura_Maths::Sine(fRotation);
		float fCosine = Asura_Maths::Cosine(fRotation);
		Asura_Quat quat(xSurfaceNormal.x * fSine, xSurfaceNormal.y * fSine, xSurfaceNormal.z * fSine, fCosine);
		Asura_Matrix_3x3 xRotation;
		quat.ToMatrix(xRotation);
		xOrientation = xRotation * xOrientation;
	}

	// add scale into local->world matrix. Since culling will still be done using original bounding
	// radius, and this is quite hard to change, only allow decals to be scaled down.
	// (MH - 31.05.2004) disabled this assert - it's ok for smallish shapes to be made a bit larger...
	float fScaleX = xScale.x;
	float fScaleY = xScale.y;
	float fScaleZ = xScale.z;
	if (fScaleX != 1.0f)
	{
//		Asura_Assert(fScaleX > 0.0f && fScaleX < 1.0f, "can't scale 3d decals to be larger than original shape");
		xOrientation[0][0] *= fScaleX;
		xOrientation[1][0] *= fScaleX;
		xOrientation[2][0] *= fScaleX;
	}
	if (fScaleY != 1.0f)
	{
//		Asura_Assert(fScaleY > 0.0f && fScaleY < 1.0f, "can't scale 3d decals to be larger than original shape");
		xOrientation[0][1] *= fScaleY;
		xOrientation[1][1] *= fScaleY;
		xOrientation[2][1] *= fScaleY;
	}
	if (fScaleZ != 1.0f)
	{
//		Asura_Assert(fScaleZ > 0.0f && fScaleZ < 1.0f, "can't scale 3d decals to be larger than original shape");
		xOrientation[0][2] *= fScaleZ;
		xOrientation[1][2] *= fScaleZ;
		xOrientation[2][2] *= fScaleZ;
	}

	// get the decal shape so we can access the bounding box
	const Asura_Shape* pxShape = Asura_Shape_System::GetShape(pxTemplate->m_uShapeHashID);
	Asura_Assert(pxShape, "no shape for 3d decal");
	if (!pxShape) return false;

    // create worldspace bounding box
	Asura_Assert(pxShape->GetLOSCollisionData(), "");
	Asura_Bounding_Box xRotatedDisplacedBB;
	xRotatedDisplacedBB.FromOBB( pxShape->GetLOSCollisionData()->GetBoundingBox(), xOrientation );
	xRotatedDisplacedBB.Displace(xBasePos);

    // find all visible polys that intersect (not just collision polys, we're checking for visual artifacts)
	u_int uNumberOfTris = Asura_Dynamics_System::FindVisiblePolysInBoundingBox(xRotatedDisplacedBB);

	// if we don't intersect any visible polygons, that's probably because we were created
	// on collision mesh and there are no visible polygons in range. This decal will most
	// likely float above the visible geometry, so don't create it.
	if (!uNumberOfTris)
	{
		return false;
	}

	const Asura_Collision_Poly* pxTriangle = Asura_Dynamics_System::GetCollisionPolysArray();

 	for (u_int u=0; u<uNumberOfTris; u++, pxTriangle++)
	{
		if (!pxTriangle->Normal.ApproximatelyEquals(xSurfaceNormal, dbg_f3dDecalNormalThreshold))
		{
			// collided with a poly which isn't facing the same way as the one we hit.
			// The 3d decal will probably look bad so don't add it
			return false;
		}
	}
	
	// check for intersections with other 3d decals. Because the Z value of the base
	// of the hole will be the same for all decals of the same type you get bad z-fighting.
	// Also if they intersect you get draw order issues with the Z mask of one decal
	// and the polygons of an other.
	{
		u_int uNumDecals = g_u3dDecalCounter;
		if (uNumDecals > ASURA_MAX_NO_OF_3D_DECALS)
		{
			uNumDecals = ASURA_MAX_NO_OF_3D_DECALS;
		}
		
		for (u_int u = 0; u < uNumDecals; ++u)
		{
			const Asura_Bounding_Box& xRotatedDisplacedBBToTest = g_px3dDecalArray[u].m_xShapeInstanceMain.GetBB();
			if (xRotatedDisplacedBBToTest.Intersects(xRotatedDisplacedBB))
			{
				return false;
			}

		}
	}

	u_int uDecalIndex = (g_u3dDecalCounter++) % ASURA_MAX_NO_OF_3D_DECALS;
	g_px3dDecalArray[uDecalIndex].Init(pxTemplate, xBasePos, xOrientation);
	// place the top decal a bit higher up
	//g_px3dDecalArray[uDecalIndex].m_xShapeInstanceBlend.SetNewPosAndOri(xBasePos + xOffset, xOrientation);
	return true;
}

