//////////////////////////////////////////////////////////////////////////////
//
// Asura_Beam_System.cpp
//
// Copyright (c) 2004-2005 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "SpecialFX/Asura_Beam_System.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Network/Asura_Message.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int uFIRST_VALID_BEAM_ID = 1;

DEBUGVAR bool dbg_bRenderBeams = true;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static float LerpForLifetime( float fInitialValue, float fEndValue, float fLifetime )
{
    return( fEndValue + fLifetime * (fInitialValue - fEndValue) );
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Beam_System.
//

Asura_Beam  Asura_Beam_System::s_axBeams[ASURA_MAX_NO_OF_BEAMS];
Asura_Beam* Asura_Beam_System::s_apxBeamPtrs[ASURA_MAX_NO_OF_BEAMS];
u_int       Asura_Beam_System::s_uNumActiveBeams;
u_int       Asura_Beam_System::s_uNextBeamID;
float       Asura_Beam_System::s_fEndEffectAngleTolerance = 0.0002f;
int			Asura_Beam_System::s_iTiledTextureTextureID;

void Asura_Beam_System::Initialise(void)
{
    if( Asura_Texture_Management::LoadAsuraFileUsingResetGroup(ASURA_BEAM_SYSTEM_TILED_TEXTURE_FILENAME, ASURA_TEXTURE_LOADED_PERMANENT) )
    {
        // This assumes that the beam texture is new at this point. But it
        // should be.
        s_iTiledTextureTextureID = Asura_Texture_Management::GetLastAddedTexID();
    }
    else
	{
		Asura_Assertf( false, ("Failed to load beams tiled texture file. '%s'", ASURA_BEAM_SYSTEM_TILED_TEXTURE_FILENAME ));
	    s_iTiledTextureTextureID = ASURA_TEXTURE_NONE;
	}

    Reset();

	Platform_Initialise();

    // This should call back to SetupTemplates() with array(s) of the project-
    // specific templates.
    Project_Initialise();
}

void Asura_Beam_System::Shutdown(void)
{
	Project_Shutdown();

	Platform_Shutdown();

    for ( u_int u = 0; u < ASURA_MAX_NO_OF_BEAMS; ++u )
    {
        s_axBeams[u].m_uID = uINVALID_BEAM_ID;
        s_apxBeamPtrs[u] = &s_axBeams[u];
    }

	s_uNumActiveBeams = 0;
	s_uNextBeamID     = uFIRST_VALID_BEAM_ID;
}

void Asura_Beam_System::Reset(void)
{
	Project_Reset();

    for ( u_int u = 0; u < ASURA_MAX_NO_OF_BEAMS; ++u )
    {
        s_axBeams[u].m_uID = uINVALID_BEAM_ID;
        s_apxBeamPtrs[u] = &s_axBeams[u];
    }

	s_uNumActiveBeams = 0;
	s_uNextBeamID     = uFIRST_VALID_BEAM_ID;
}

void Asura_Beam_System::SetupTemplates( Asura_Beam_Template* pxTemplates, u_int uCount )
{
    for( u_int u = 0; u < uCount; ++u )
    {
        Asura_Assert( pxTemplates[u].iTileHandle == ASURA_INVALID_HANDLE, "" );

        pxTemplates[u].iTileHandle = Asura_TiledTextureManagement::GetHandleOfTile( pxTemplates[u].uTileHashID );

	    Asura_Assertf( pxTemplates[u].iTileHandle != ASURA_INVALID_HANDLE,
                       ("Failed to find tile info for tile with hash ID 0x%08X. No beams using this template can be created.", pxTemplates[u].uTileHashID) );
    }
}

void Asura_Beam_System::ResetTemplates( Asura_Beam_Template* pxTemplates, u_int uCount )
{
	for( u_int u = 0; u < uCount; ++u )
	{
		pxTemplates[u].iTileHandle	= ASURA_INVALID_HANDLE;
	}
}

void Asura_Beam_System::DeallocateBeam(u_int uBeamID)
{
	Asura_Assert( uBeamID < s_uNextBeamID, "Invalid beam ID." );
    if( uINVALID_BEAM_ID == uBeamID ) { return; }

    int iIdx = GetBeamArrayPtrIdxByID( uBeamID );
    if( -1 != iIdx ) { DeallocateBeamByPtrArrayIdx( iIdx ); }

#if 0 // _DEBUG
    for( u_int u = 0; u < ASURA_MAX_NO_OF_BEAMS; ++u )
    {
        bool bFound = false;
        for( u_int v = 0; v < ASURA_MAX_NO_OF_BEAMS; ++v )
        {
            if( &s_axBeams[u] == s_apxBeamPtrs[v] ) { bFound = true; break; }
        }
        Asura_Assert( bFound, "" );
    }
#endif
}

void Asura_Beam_System::Render( void )
{
    if( !s_uNumActiveBeams ) { return; }
    if( !dbg_bRenderBeams  ) { return; }

	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_BEAMS))
	{
		return;
	}

    Platform_RenderStart();
	Platform_Render();
	Platform_RenderEnd();
}

void Asura_Beam_System::HandleMessage( const Asura_Message* pxMessage )
{
    switch( pxMessage->GetMsgID() )
    {   
        case ASURA_MSG_CREATE_BEAM:
        {
            const Asura_Msg_Create_Beam* pxData = static_cast<const Asura_Msg_Create_Beam*>( pxMessage->GetData() );
            Asura_Beam* pxBeam = Asura_Beam_System::CreateBeam( pxData->m_xInitialTailPos, pxData->m_xInitialHeadPos, pxData->m_xTemplate );
            if( pxBeam && pxData->m_fDecayRate > 0.0f )
            {
                pxBeam->m_fDecayRate = pxData->m_fDecayRate;
            }
            break;
        }
    }
}

void Asura_Beam_System::GlobalCreate( const Asura_Vector_3& xInitialTailPos,
                                      const Asura_Vector_3& xInitialHeadPos,
                                      const Asura_Beam_Template& xTemplate,
                                      Asura_Guid uExceptForPlayerGuid,
                                      float fDecayRate )
{
    Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_CREATE_BEAM, sizeof(Asura_Msg_Create_Beam) );

    pxMsg->SetToGuid( ASURA_GUID_OTHER_PLAYERS_CLIENTMACHINES );
    pxMsg->SetFromGuid( uExceptForPlayerGuid );
    
    // Setup data.
    {
        Asura_Msg_Create_Beam* pxData = static_cast<Asura_Msg_Create_Beam*>( pxMsg->GetData() );

        pxData->m_xInitialTailPos = xInitialTailPos;
        pxData->m_xInitialHeadPos = xInitialHeadPos;
        pxData->m_xTemplate       = xTemplate;
        pxData->m_fDecayRate      = fDecayRate;
    }

    Asura_Network::Send( pxMsg );
}

Asura_Beam* Asura_Beam_System::CreateBeam( const Asura_Vector_3& xInitialTailPos,
                                           const Asura_Vector_3& xInitialHeadPos,
                                           const Asura_Beam_Template& xTemplate ) 
{    
    if( s_uNumActiveBeams >= ASURA_MAX_NO_OF_BEAMS )
    {
        Asura_Assert( false, "Too many beams! Consider increasing your project's/platform's ASURA_MAX_NO_OF_BEAMS value." );
        return NULL;
    }

    // Don't allow beams to created if the tile they're trying to reference
    // is missing.
    if( xTemplate.iTileHandle == ASURA_INVALID_HANDLE ) { return NULL; }

	if(xInitialHeadPos.ApproximatelyEquals(xInitialTailPos))
	{
		Asura_Assert(0, "Tried to create a zero length beam.");
		return NULL;
	}

	Asura_Beam* const pxBeam = s_apxBeamPtrs[s_uNumActiveBeams++];

	pxBeam->SetPositions( xInitialTailPos, xInitialHeadPos );

    pxBeam->m_fTailRadiusScale = 1.0f;
    pxBeam->m_fHeadRadiusScale = 1.0f;
    pxBeam->m_fSpeedScale      = 1.0f;
    pxBeam->m_fTailAlphaScale  = 1.0f;
    pxBeam->m_fHeadAlphaScale  = 1.0f;
    
	pxBeam->m_fLifeTime          = 1.0f;
	pxBeam->m_fDecayRate         = xTemplate.fDecayRate;
	pxBeam->m_fTimesliceOverride = -1.0f;

#ifdef ASURA_BEAMS_ALL_MODES
    pxBeam->m_eTranslucencyMode = ASURA_TRANSLUCENCY_GLOWING;
#endif

    pxBeam->m_pxTemplate = &xTemplate;
    pxBeam->m_uID = s_uNextBeamID++;

#ifdef ASURA_BEAMS_SUPPORT_CAMERA_SELECTION
	pxBeam->m_uOnlyRenderIfCameraIgnoresThisGuid = ASURA_GUID_UNREGISTERED;
	pxBeam->m_uDontRenderIfCameraIgnoresThisGuid = ASURA_GUID_UNREGISTERED;
#endif

	pxBeam->m_bInvisible = false;
	pxBeam->m_bDontUpdateUntilRenderedOnce = false;
		
	return pxBeam;
}

void Asura_Beam_System::Update(void)
{
	const float fGameTimeslice = Asura_Timers::GetGameTimeSlice();

	// update all active Beams
	for( u_int u = 0; u < s_uNumActiveBeams; )
	{
        Asura_Beam* const pxBeam = s_apxBeamPtrs[u];
		if( pxBeam->m_bDontUpdateUntilRenderedOnce )
		{
			// just skip this beam
			++u;
		}
		else
		{		
			float fTimeslice = fGameTimeslice;

			if( pxBeam->m_fTimesliceOverride >= 0.0f )
			{
				//overide the time for this frame. RAR 20/03/03
				fTimeslice = pxBeam->m_fTimesliceOverride;
				pxBeam->m_fTimesliceOverride = -1.0f;
			}
			
			pxBeam->m_fLifeTime -= (pxBeam->m_fDecayRate * fTimeslice);
			
			// remove any timed-out Beams
			if( pxBeam->m_fLifeTime <= 0.0f )
			{
				DeallocateBeamByPtrArrayIdx( u );
			}
			else
			{
				const Asura_Beam_Template* pxTemplate = pxBeam->m_pxTemplate;
				const float fSpeedScale = pxBeam->m_fSpeedScale;

				// Update position and length.
				{
					float fTailMove = fTimeslice * pxTemplate->fTailSpeed * fSpeedScale;
					float fHeadMove = fTimeslice * pxTemplate->fHeadSpeed  * fSpeedScale;
	            
					pxBeam->m_xTailPosition += pxBeam->m_xDirection * fTailMove;
					pxBeam->m_fLength       += fHeadMove - fTailMove;
				}

				++u;
			}
		}
	}	
}

void Asura_Beam_System::SetTimesliceForBeam(u_int uBeamID,float fTimeslice)
{
	Asura_Assert(uBeamID<s_uNextBeamID,"Unexpected beam ID");
	if(!uBeamID) return;

    Asura_Beam* pxBeam = GetBeamByID( uBeamID );
    if( pxBeam )
    {
		//set the time for the beam next frame. If a time has already been set , add it on.
		//RAR 20/03/03
		if(pxBeam->m_fTimesliceOverride<0.0f) pxBeam->m_fTimesliceOverride = 0.0f;
		pxBeam->m_fTimesliceOverride += fTimeslice;			
	}
}

int Asura_Beam_System::GetBeamArrayPtrIdxByID( u_int uID )
{
	if( uINVALID_BEAM_ID == uID ) return -1;

	for( u_int u = 0; u < s_uNumActiveBeams; ++u )
	{
        if( s_apxBeamPtrs[u]->m_uID == uID ) { return u; }
    }

    return -1;
}

Asura_Beam* Asura_Beam_System::GetBeamByID( u_int uID )
{
    int iIdx = GetBeamArrayPtrIdxByID( uID );
    return( -1 == iIdx ? NULL : s_apxBeamPtrs[iIdx] );
}

void Asura_Beam_System::DeallocateBeamByPtrArrayIdx( u_int uPtrArrayIdx )
{
    --s_uNumActiveBeams;

    Asura_Beam* const pxDeleting = s_apxBeamPtrs[uPtrArrayIdx];
    Asura_Beam* const pxLast     = s_apxBeamPtrs[s_uNumActiveBeams];

    s_apxBeamPtrs[uPtrArrayIdx]      = pxLast;
    s_apxBeamPtrs[s_uNumActiveBeams] = pxDeleting;
}

//
// Asura_Beam.
//

void Asura_Beam::CalcCurColours( Asura_RGBA* pxTail, Asura_RGBA* pxHead ) const
{
    const float fLifetime = m_fLifeTime;

    // Tail.
    {
        const float* pfInitialTail = m_pxTemplate->xInitialTailColour.m_afChannels;
        const float* pfEndTail     = m_pxTemplate->xEndTailColour.m_afChannels;

        for( u_int u = 0; u < 4; ++u )
        {
			// MH 6.12.05 -- got rid of union in Asura_Colour, therefore need to use Asura_Beam_Colour to access by channel.
            reinterpret_cast<Asura_Beam_Colour*>(pxTail)->m_afChannels[u] = LerpForLifetime( pfInitialTail[u], pfEndTail[u], fLifetime );
        }

        pxTail->m_fA *= m_fTailAlphaScale;
    }

    // Head.
    {
        const float* pfInitialHead = m_pxTemplate->xInitialHeadColour.m_afChannels;
        const float* pfEndHead     = m_pxTemplate->xEndHeadColour.m_afChannels;

        for( u_int u = 0; u < 4; ++u )
        {
			// MH 6.12.05 -- got rid of union in Asura_Colour, therefore need to use Asura_Beam_Colour to access by channel.
            reinterpret_cast<Asura_Beam_Colour*>(pxHead)->m_afChannels[u] = LerpForLifetime( pfInitialHead[u], pfEndHead[u], fLifetime );
        }

        pxHead->m_fA *= m_fHeadAlphaScale;
    }
}

Asura_Vector_3 Asura_Beam::CalcCurHeadPos( void ) const
{
    return( m_xTailPosition + m_xDirection * m_fLength );
}

void Asura_Beam::CalcCurRadii( float* pfTail, float* pfHead ) const
{
    float fCurRadius = LerpForLifetime( m_pxTemplate->fInitialRadius, m_pxTemplate->fEndRadius, m_fLifeTime );

    *pfTail = fCurRadius * m_fTailRadiusScale;
    *pfHead = fCurRadius * m_fHeadRadiusScale;
}

void Asura_Beam::SetPositions( const Asura_Vector_3& xTailPos, const Asura_Vector_3& xDir, float fLength )
{
	m_xTailPosition = xTailPos;
	m_xDirection    = xDir;
    m_fLength       = fLength;
}

void Asura_Beam::SetPositions( const Asura_Vector_3& xTailPos, const Asura_Vector_3& xHeadPos )
{
	m_xTailPosition  = xTailPos;
	m_xDirection	 = xHeadPos - xTailPos;
	m_fLength		 = m_xDirection.Normalise();
	// To avoid this, the caller needs to check that the tail and head position are not the same.
	// There's not much we can do about it in here.
	Asura_Assert(m_fLength > 0.f, "Zero length beam");
}


/// End of file. ///
