///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_CFX_Effect.cpp
//
// Created by Martin G.Harris (10/10/2007)
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "SpecialFX/Asura_CFX_Effect.h"

// Only compile this code if the symbol is defined.
#ifdef ASURA_USE_COLLISIONFX

#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"

#include "Physics/Asura_Physics_CollisionFX_Report.h"

#include "Sound/Asura_Sound.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "Sound/Asura_Sound_Event_System.h"

#include "SpecialFX/Asura_CollisionFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Entity_Effect.h"
#include "SpecialFX/PFX/Asura_PFX_Enums.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"

#ifdef ASURA_USE_DECAL_MANAGER
#include "SpecialFX/Asura_Decal_Manager.h"
#endif //ASURA_USE_DECAL_MANAGER


///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Effect
///////////////////////////////////////////////////////////////////////////////

Asura_CFX_Effect::Asura_CFX_Effect(u_int uFlags)
: m_pxEntry(0)
, m_fDestroyDelay(fASURA_CFX_DESTROY_DELAY)
, m_uFlags(0)
, m_fMinRepeatDelay(fASURA_CFX_MIN_REPEAT_DELAY)
, m_fMaxRepeatDelay(fASURA_CFX_MAX_REPEAT_DELAY)
, m_fNextRepeatTime(0.0f)
, m_uLastFaceID(0)
, m_xLastPosition(0.0f, 0.0f, 0.0f)
{
	// Make sure that the state flags are cleared.
	SetFlags(uFlags);
}

//////////////
// SetFlags //
//////////////
// Set the flags, but preserve the state flags.
void Asura_CFX_Effect::SetFlags(u_int uFlags)
{
	const u_int uStateFlags = (m_uFlags & ASURA_CFX_FLAG_STATE_MASK);
	m_uFlags = (uFlags & ~ASURA_CFX_FLAG_STATE_MASK);
	m_uFlags |= uStateFlags;
}

/////////////////////
// SetStartedState //
/////////////////////
void Asura_CFX_Effect::SetStartedState(bool bStarted)
{
	// Set or clear the started state flag.
	if (bStarted)
	{
		m_uFlags |= ASURA_CFX_FLAG_STATE_STARTED;
	}
	else
	{
		m_uFlags &= ~ASURA_CFX_FLAG_STATE_STARTED;
	}
}

///////////////////
// SetLastFaceID //
///////////////////
void Asura_CFX_Effect::SetLastFaceID(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	if (xCollInfo.IsEnvironmentCollision())
	{
		m_uLastFaceID = xCollInfo.GetEnvironmentFaceID();
	}
}

///////////////////////
// SetNextRepeatTime //
///////////////////////
// Set the time at which the next repeat is allowed.
void Asura_CFX_Effect::SetNextRepeatTime()
{
	if (m_uFlags & ASURA_CFX_FLAG_RESTART_AFTER_DELAY)
	{
		m_fNextRepeatTime = Asura_Timers::GetLocalScaledTime();

		if (m_fMinRepeatDelay == m_fMaxRepeatDelay)
		{
			m_fNextRepeatTime += m_fMinRepeatDelay;
		}
		else
		{
			// Select delay between minimum and maximum using a uniform random distribution.
			m_fNextRepeatTime += Asura_Random::GetFloat(m_fMinRepeatDelay, m_fMaxRepeatDelay);
		}
	}
	else
	{
		// Not using the repeat delay.
		m_fNextRepeatTime = 0.0f;
	}
}

///////////////////////////
// HasRepeatDelayExpired //
///////////////////////////
bool Asura_CFX_Effect::HasRepeatDelayExpired() const
{
	if (GetShouldRestartAfterDelay())
	{
		return (Asura_Timers::GetLocalScaledTime() >= m_fNextRepeatTime);
	}

	// No need to wait for the delay.
	return true;
}

///////////////////
// ShouldRestart //
///////////////////
// Should we restart the effect?
bool Asura_CFX_Effect::ShouldRestart(bool bPlaying) const
{
	if (GetShouldRestartIfFinished())
	{
		// Need to wait for effect to finish.
		if (!bPlaying)
		{
			return HasRepeatDelayExpired();
		}
	}
	else if (GetShouldRestartAfterDelay())
	{
		// Don't need to wait for effect to finish.
		return HasRepeatDelayExpired();
	}

	return false;
}

///////////
// Start //
///////////
void Asura_CFX_Effect::Start(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	SetStartedState(true);
	SetLastFaceID(xCollInfo);
	SetNextRepeatTime();
	m_xLastPosition = xCollInfo.GetPosition();
}

//////////
// Stop //
//////////
// Collision entry has been removed. Stop PFX
void Asura_CFX_Effect::Stop(bool bImmediate)
{
	SetStartedState(false);
}

/////////////////////
// SetLastPosition //
/////////////////////
// Update the last position at which a collision occurred (only need to do this when
// the effect starts or whilst the effect is playing).
void Asura_CFX_Effect::SetLastPosition(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	m_xLastPosition = xCollInfo.GetPosition();
}

/////////////////////
// CheckForRestart //
/////////////////////
// Check whether we should restart the effect.
// Return true, if we have restarted the effect.
bool Asura_CFX_Effect::CheckForRestart(const Asura_Physics_CollisionFX_Report& xCollInfo, bool bPlaying)
{
	if (ShouldRestart(bPlaying))
	{
		// Update the parameters, before restarting.
		UpdateParameters(xCollInfo);

		// Stop existing effect if necessary and start a new one.
		Restart(xCollInfo);
		return true;
	}

	// Sliding collisions tend to use looping effects that may need to be
	// restarted if the environmental surface type has changed.
	// However we don't want to change the effect every couple of frames
	// (e.g. if we are stradling two surface types), so apply the repeat delay
	// if it has been set.
	if (m_uFlags & ASURA_CFX_FLAG_RESTART_SLIDE_ON_TO_NEW_SURFACE)
	{
		if (GetSlidingState() && HasRepeatDelayExpired())
		{
			// Only restart if the parameters have changed significantly.
			const bool bRestart = UpdateParameters(xCollInfo);
			if (bRestart)
			{
				// Stop existing effect if necessary and start a new one.
				Restart(xCollInfo);
				return true;
			}
		}
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_PFX_Handle
///////////////////////////////////////////////////////////////////////////////
Asura_CFX_PFX_Handle::Asura_CFX_PFX_Handle(Asura_Hash_ID uPFXHash, u_int uFlags)
	: m_uPFXHash(uPFXHash)
	, m_uPFXFlags(uFlags)
	, m_uPFXEffectID(ASURA_PFX_INVALID_EFFECT_ID)
	, m_xLastNormal(0.0f, 0.0f, 0.0f)
{
}

// Start effect.
void Asura_CFX_PFX_Handle::Start(Asura_CFX_Effect& xParent, const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	if (m_uPFXHash != ASURA_HASH_ID_UNSET)
	{
		const Asura_Vector_3 xPosition(xCollInfo.GetPosition());
		const Asura_Vector_3 xNormal(xCollInfo.GetNormal());
		Start(xParent, xPosition, xNormal);
	}
}


void Asura_CFX_PFX_Handle::Start(Asura_CFX_Effect& xParent, const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal)
{
	if (m_uPFXHash != ASURA_HASH_ID_UNSET)
	{
		Asura_Assert(m_uPFXEffectID == ASURA_PFX_INVALID_EFFECT_ID, "PFX effect ID should be invalid.");
		m_xLastNormal = xNormal;

		m_uPFXEffectID = Asura_PFX_Entity_Effect::Create(xPosition, m_xLastNormal, m_uPFXHash, 0, static_cast<u_short>(m_uPFXFlags));
		Asura_Assertf(m_uPFXEffectID != ASURA_PFX_INVALID_EFFECT_ID, ("Failed to start PFX (0x%08x)", m_uPFXHash));
		if (m_uPFXEffectID != ASURA_PFX_INVALID_EFFECT_ID)
		{
			xParent.SetCreatedPFXState();
			xParent.SetUpdatedPositionState();
			xParent.SetUpdatedNormalState();
		}
	}
}

// Update PFX position and normal.
void Asura_CFX_PFX_Handle::Update(Asura_CFX_Effect& xParent, const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	if (xParent.GetUpdatePosition() && m_uPFXEffectID != ASURA_PFX_INVALID_EFFECT_ID)
	{
		const Asura_Vector_3 xPosition(xCollInfo.GetPosition());

		if (xParent.GetUpdateDirection())
		{
			m_xLastNormal = xCollInfo.GetNormal();
			Asura_PFX_System::SetEntityEffectPosition(m_uPFXEffectID, xPosition, m_xLastNormal);
			xParent.SetUpdatedNormalState();
		}
		else
		{
			Asura_PFX_System::SetEntityEffectPosition(m_uPFXEffectID, xPosition);
		}
		xParent.SetUpdatedPositionState();
	}
}

void Asura_CFX_PFX_Handle::Update(Asura_CFX_Effect& xParent)
{
	if (xParent.GetUpdatePosition() && m_uPFXEffectID != ASURA_PFX_INVALID_EFFECT_ID)
	{
		if (xParent.GetUpdateDirection())
		{
			Asura_PFX_System::SetEntityEffectPosition(m_uPFXEffectID, xParent.GetLastPosition(), m_xLastNormal);
			xParent.SetUpdatedNormalState();
		}
		else
		{
			Asura_PFX_System::SetEntityEffectPosition(m_uPFXEffectID, xParent.GetLastPosition());
		}
		xParent.SetUpdatedPositionState();
	}
}

// Stop effect.
void Asura_CFX_PFX_Handle::Stop(bool bImmediate)
{
	if (m_uPFXEffectID != ASURA_PFX_INVALID_EFFECT_ID)
	{
		Asura_PFX_System::StopEntityEffect(m_uPFXEffectID, bImmediate);
		m_uPFXEffectID = ASURA_PFX_INVALID_EFFECT_ID;
	}
}

// Is effect playing?
bool Asura_CFX_PFX_Handle::IsPlaying() const
{
	return  m_uPFXEffectID != ASURA_PFX_INVALID_EFFECT_ID &&
			Asura_PFX_System::GetEntityEffect(m_uPFXEffectID) != 0;
}


///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_SE_Handle
///////////////////////////////////////////////////////////////////////////////
Asura_CFX_SE_Handle::Asura_CFX_SE_Handle(Asura_Hash_ID uSEHash, float fVolumeScale)
	: m_uSEHash(uSEHash)
	, m_iSoundHandle(ASURA_INVALID_HANDLE)
	, m_fVolumeScale(fVolumeScale)
{
}

// Start effect.
void Asura_CFX_SE_Handle::Start(Asura_CFX_Effect& xParent, const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	if (m_uSEHash != ASURA_HASH_ID_UNSET)
	{
		const Asura_Vector_3 xPosition(xCollInfo.GetPosition());
		Start(xParent, xPosition);
	}
}

void Asura_CFX_SE_Handle::Start (Asura_CFX_Effect& xParent, const Asura_Vector_3& xPosition)
{
	if (m_uSEHash != ASURA_HASH_ID_UNSET)
	{
		Asura_Assert(m_iSoundHandle == ASURA_INVALID_HANDLE, "Sound event handle should be invalid.");

		const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::FindEvent( m_uSEHash );
		Asura_Assertf(pxEvent, ("Failed to find sound event (0x%08x)", m_uSEHash));
		if (pxEvent)
		{
			Asura_Sound_Event_System::PlayEvent( pxEvent, &m_iSoundHandle, &xPosition, m_fVolumeScale );
			xParent.SetCreatedSEState();
			xParent.SetUpdatedPositionState();
		}
	}
}

// Update position.
void Asura_CFX_SE_Handle::Update(Asura_CFX_Effect& xParent, const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	const Asura_Vector_3 xPosition(xCollInfo.GetPosition());
	Update(xParent, xPosition);
}

void Asura_CFX_SE_Handle::Update(Asura_CFX_Effect& xParent, const Asura_Vector_3& xPosition)
{
	if (xParent.GetUpdatePosition() && IsPlaying())
	{
		Asura_Sound_Event_System::UpdatePosition(m_iSoundHandle, xPosition);
		xParent.SetUpdatedPositionState();
	}
}

// Stop effect.
void Asura_CFX_SE_Handle::Stop(bool bImmediate)
{
	if (m_iSoundHandle != ASURA_INVALID_HANDLE)
	{
		float fFadeOut = 0.0f;
		if (!bImmediate && Asura_Sound_Event_System::IsPlaying(m_iSoundHandle))
		{
			fFadeOut = Asura_CollisionFX_System::GetParameters().m_fSoundEventFadeOutTime;
		}

		Asura_Sound_Event_System::Stop(m_iSoundHandle, fFadeOut);
	}
}

// Is effect playing?
bool Asura_CFX_SE_Handle::IsPlaying() const
{
	return  (m_iSoundHandle != ASURA_INVALID_HANDLE) &&
			Asura_Sound_Event_System::IsPlaying(m_iSoundHandle);
}


// Set the volume scale and optionally update the playing sound event.
void Asura_CFX_SE_Handle::SetVolumeScale(Asura_CFX_Effect& xParent, float fScale, bool bUpdateSE)
{
	const float fOldVolume = m_fVolumeScale;
	Asura_Assert(fScale>=0.0f, "Invalid volume scale");
	if (fScale>=0.0f)
	{
		m_fVolumeScale = fScale;
	}

	if (bUpdateSE && IsPlaying())
	{
		Asura_Sound_Event_System::SetVolumeScale(m_iSoundHandle, m_fVolumeScale);
		if (fOldVolume != m_fVolumeScale)
		{
			xParent.SetUpdatedVolumeState();
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Decal_Handle
///////////////////////////////////////////////////////////////////////////////
// Wrapper for decal.
#ifdef ASURA_USE_DECAL_MANAGER
Asura_CFX_Decal_Handle::Asura_CFX_Decal_Handle(Asura_Hash_ID uDecalSetHash,
											   int iDecalIndex,
											   float fScale /*= 1.0f*/,
											   float fOrientation /*= 0.0f*/)
: m_uDecalSetHash(uDecalSetHash)
, m_iDecalIndex(iDecalIndex)
, m_fScale(fScale)
, m_fOrientation(fOrientation)
{
}

const float fMAX_DECAL_NORMAL_VELOCITY_COSINE = 0.9962f; // 5 degrees

void Asura_CFX_Decal_Handle::Start (const Asura_CFX_Effect& xParent, const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	if (m_uDecalSetHash != ASURA_HASH_ID_UNSET)
	{
		const bool bLocal = xParent.GetLocalEffect();

		const Asura_Vector_3 xPosition(xCollInfo.GetPosition());
		const Asura_Vector_3 xNormal(xCollInfo.GetNormal());

		if (xParent.GetAlignDecalTopWithVelocity())
		{
			// If velocity is almost parallel to the surface normal then don't create the decal.
			Asura_Vector_3 xVel(xCollInfo.GetRelVelocity());
			xVel.Normalise();
			if (Asura_Maths::Abs(xVel*xNormal) < fMAX_DECAL_NORMAL_VELOCITY_COSINE)
			{
				// Align the decal top direction with the velocity direction.
				Asura_Decal_Manager::AddDecal(xPosition, xNormal, xVel, m_uDecalSetHash, m_iDecalIndex, m_fScale, m_fOrientation);

				if (!bLocal)
				{
					Asura_CollisionFX_System::SendClientCreateDecalMessage(*this, xPosition, xNormal, &xVel);
				}
			}
		}
		else
		{
			Asura_Decal_Manager::AddDecal(xPosition, xNormal, m_uDecalSetHash, m_iDecalIndex, m_fScale, m_fOrientation);

			if (!bLocal)
			{
				Asura_CollisionFX_System::SendClientCreateDecalMessage(*this, xPosition, xNormal, 0);
			}
		}
	}
}
#endif  //ASURA_USE_DECAL_MANAGER



///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Effect_PFX
///////////////////////////////////////////////////////////////////////////////
Asura_CFX_Effect_PFX::Asura_CFX_Effect_PFX(Asura_Hash_ID uPFXHash,
										   u_int uFlags)
	: Asura_CFX_Effect(uFlags)
	, m_xPFXHandle(uPFXHash, ASURA_PFX_EFFECT_CAMERA_FLAGS)
{
}

///////////
// Start //
///////////
// First collision has happened or restarting effect.
void Asura_CFX_Effect_PFX::Start(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	m_xPFXHandle.Start(*this, xCollInfo);
	Asura_CFX_Effect::Start(xCollInfo);
}

////////////
// Update //
////////////
// Subsequent collision has happened.
void Asura_CFX_Effect_PFX::Update(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	const bool bPlaying = m_xPFXHandle.IsPlaying();

	if (CheckForRestart(xCollInfo, bPlaying))
	{
		return;
	}

	if (bPlaying)
	{
		Asura_CFX_Effect::SetLastPosition(xCollInfo);
		m_xPFXHandle.Update(*this, xCollInfo);
	}
}

//////////
// Stop //
//////////
// Collision entry has been removed. Stop PFX
void Asura_CFX_Effect_PFX::Stop(bool bImmediate)
{
	m_xPFXHandle.Stop(bImmediate);
	Asura_CFX_Effect::Stop(bImmediate);
}


///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Effect_Sound_Event
///////////////////////////////////////////////////////////////////////////////

Asura_CFX_Effect_Sound_Event::Asura_CFX_Effect_Sound_Event(Asura_Hash_ID uSEHash,
														   u_int uFlags,
														   float fVolumeScale /*= 1.0f*/)
	: Asura_CFX_Effect(uFlags)
	, m_xSEHandle(uSEHash, fVolumeScale)
{
}

///////////
// Start //
///////////
// First collision has happened or restarting effect.
void Asura_CFX_Effect_Sound_Event::Start(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	m_xSEHandle.Start(*this, xCollInfo);
	Asura_CFX_Effect::Start(xCollInfo);
}

////////////
// Update //
////////////
// Subsequent collision has happened.
void Asura_CFX_Effect_Sound_Event::Update(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	const bool bPlaying = m_xSEHandle.IsPlaying();

	if (CheckForRestart(xCollInfo, bPlaying))
	{
		return;
	}
	
	if (bPlaying)
	{
		Asura_CFX_Effect::SetLastPosition(xCollInfo);
		m_xSEHandle.Update(*this, xCollInfo);
	}
}

//////////
// Stop //
//////////
// Collision entry has been removed. Stop sound event.
void Asura_CFX_Effect_Sound_Event::Stop(bool bImmediate)
{
	m_xSEHandle.Stop(bImmediate);
	Asura_CFX_Effect::Stop(bImmediate);
}



///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Effect_PFX_Sound_Event
///////////////////////////////////////////////////////////////////////////////
Asura_CFX_Effect_PFX_Sound_Event::Asura_CFX_Effect_PFX_Sound_Event(Asura_Hash_ID uPFXHash,
																   Asura_Hash_ID uSEHash,
																   u_int uFlags,
																   float fVolumeScale /*= 1.0f*/)
	: Asura_CFX_Effect(uFlags)
	, m_xPFXHandle(uPFXHash, ASURA_PFX_EFFECT_CAMERA_FLAGS)
	, m_xSEHandle(uSEHash, fVolumeScale)
{
}

///////////
// Start //
///////////
// First collision has happened or restarting effect.
void Asura_CFX_Effect_PFX_Sound_Event::Start(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	m_xPFXHandle.Start(*this, xCollInfo);
	m_xSEHandle.Start(*this, xCollInfo);
	Asura_CFX_Effect::Start(xCollInfo);
}

////////////
// Update //
////////////
// Subsequent collision has happened.
void Asura_CFX_Effect_PFX_Sound_Event::Update(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	const bool bPlayingPFX = m_xPFXHandle.IsPlaying();
	const bool bPlayingSE  = m_xSEHandle.IsPlaying();

	const bool bPlaying = (bPlayingPFX || bPlayingSE);

	if (CheckForRestart(xCollInfo, bPlaying))
	{
		return;
	}

	if (bPlayingPFX)
	{
		m_xPFXHandle.Update(*this, xCollInfo);
	}
	if (bPlayingSE)
	{
		m_xSEHandle.Update(*this, xCollInfo);
	}
	if (bPlaying)
	{
		Asura_CFX_Effect::SetLastPosition(xCollInfo);
	}
}

//////////
// Stop //
//////////
// Collision entry has been removed. Stop effects.
void Asura_CFX_Effect_PFX_Sound_Event::Stop(bool bImmediate)
{
	m_xPFXHandle.Stop(bImmediate);
	m_xSEHandle.Stop(bImmediate);
	Asura_CFX_Effect::Stop(bImmediate);
}


///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Effect_PFX_SE_Decal
///////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_USE_DECAL_MANAGER
Asura_CFX_Effect_PFX_SE_Decal::Asura_CFX_Effect_PFX_SE_Decal(Asura_Hash_ID uPFXHash,
								  Asura_Hash_ID uSEHash,
								  Asura_Hash_ID uDecalSetHash,
								  int iDecalIndex,
								  u_int uFlags /*= ASURA_CFX_FLAG_DEFAULT*/,
								  float fVolumeScale /*= 1.0f*/)

	: Asura_CFX_Effect_PFX_Sound_Event(uPFXHash, uSEHash, uFlags, fVolumeScale)
	, m_xDecalHandle(uDecalSetHash, iDecalIndex)
{
}

void Asura_CFX_Effect_PFX_SE_Decal::Start(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	// MGH 19/11/2007: For now only do decals for environmental collisions. We may want to
	// do decals on static objects later, but this is more complicated.
	if (xCollInfo.IsEnvironmentCollision())
	{
		m_xDecalHandle.Start(*this, xCollInfo);
	}
	Asura_CFX_Effect_PFX_Sound_Event::Start(xCollInfo);
}

void Asura_CFX_Effect_PFX_SE_Decal::Update(const Asura_Physics_CollisionFX_Report& xCollInfo)
{
	Asura_CFX_Effect_PFX_Sound_Event::Update(xCollInfo);
}

void Asura_CFX_Effect_PFX_SE_Decal::Stop(bool bImmediate)
{
	Asura_CFX_Effect_PFX_Sound_Event::Stop(bImmediate);
}
#endif //ASURA_USE_DECAL_MANAGER

///////////////////////////////////////////////////////////////////////////////
// Asura_CFX_Effect_PFX_SE_RemoteClient
///////////////////////////////////////////////////////////////////////////////
void Asura_CFX_Effect_PFX_SE_RemoteClient::StartClient(const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal)
{
	m_xPFXHandle.Start(*this, xPosition, xNormal);
	m_xSEHandle.Start(*this, xPosition);

	SetStartedState(true);
	SetNextRepeatTime();
	m_xLastPosition = xPosition;
}

// Update some or all of the parameters.
void Asura_CFX_Effect_PFX_SE_RemoteClient::UpdateClient(const Asura_Vector_3* pxPosition,
														const Asura_Vector_3* pxNormal,
														const float* pfVolumeScale)
{
	if (pxPosition)
	{
		m_xLastPosition = *pxPosition;
		m_xSEHandle.Update(*this, *pxPosition);
	}
	if (pfVolumeScale && *pfVolumeScale != m_xSEHandle.GetVolumeScale())
	{
		m_xSEHandle.SetVolumeScale(*this, *pfVolumeScale, true);
	}

	if (pxNormal)
	{
		m_xPFXHandle.m_xLastNormal = *pxNormal;
	}

	if (pxPosition || (pxNormal && GetUpdateDirection()))
	{
		m_xPFXHandle.Update(*this);
	}
}

// Stop a remote client effect.
void Asura_CFX_Effect_PFX_SE_RemoteClient::StopClient()
{
	Stop(false);
}

#endif //ASURA_USE_COLLISIONFX

