/**
*
* \file Asura_Corona_System.h
*
* Created by Kevin Lea - 27/01/2003 09:22:14
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/
#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_CORONA_SYSTEM_H__
#define __ASURA_CORONA_SYSTEM_H__

#include "Core/Asura.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Mempool.h"

#include "Rendering/Asura_Module_Defines.h"

#if defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
#include "Visibility/Asura_Visibility_System.h"
#endif

#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_ModuleList.h"
#endif	// defined(ASURA_USE_STREAMING)


#define ASURA_CORONA_FADE_DIST_BEGIN (60.0f)
#define ASURA_CORONA_FADE_DIST_END   (75.0f)


template<typename T> class Asura_TrackingHandlePool;

class Asura_Corona
{
	public:
		//get the distance at which the corona starts fading out when close to the camera
		float GetNearFadeDistance() const {return m_bDoNotFadeOutNear ? 0.0f : m_fRadius;}

		#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
		bool IsInVisibleModule();
		bool TestModuleVisibility();
		void CalculateModuleList();
		void InvalidateModuleList();
		#endif
		
		Asura_Vector_3	m_xPosition;
		Asura_Vector_3	m_xDirection;

		union
		{
			/// Kevin Lea - 27/01/2003 12:26:55 - union of bitfields with simple u_int
			/// enables guaranteed setting of all flags with one line of code i.e. m_uFlags = 0
			u_int		m_uFlags;
			
			struct 
			{
				u_int	m_bIsHidden :1;

				u_int	m_bOnlyDrawIfPositionSetPerFrame :1;
				u_int	m_bPositionUpdatedThisFrame	:1;
				
				u_int	m_bUseDirectionEffect :1;
				
				u_int	m_bScaleWithDistance :1;
				u_int	m_bDoNotFadeOutWithDistance :1;
				u_int	m_bDoNotFadeOutNear : 1;
				u_int	m_bUseModuleLinking:1;
				
			};
		};

		float			m_fRadius;

		int				m_iTextureIndex;

		float			m_fRed;
		float			m_fGreen;
		float			m_fBlue;

		float			m_fVolumetricDepthScale;

		//For coronas that use module linking
#if defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
		Asura_Visibility_Query m_xVisibilityQuery;
#else
		#if defined(ASURA_USE_STREAMING)
		u_int				m_uFrameNumberCalculatedModules;
		Asura_Hash_ID		m_uLastKnownZone;
		Asura_Streaming_ModuleList<NUM_MODULES_TO_STORE_FOR_CLIPPING, 2> m_xModuleList;
		#else
		unsigned short		m_ausModuleList[NUM_MODULES_TO_STORE_FOR_CLIPPING];
		#endif
		int					m_iLastKnownModule;
#endif

		
};

class Asura_Corona_Handler
{
public:
	Asura_Corona_Handler();
	~Asura_Corona_Handler();

	void Initialise(void);
	void Reset(void);
	void Shutdown(void);

	void Update(void) {} // This was never previously defined in the corona system, since coronas don't need updating. However, so we're not removing things in these changes this is left in.
	void Render(void);

	Asura_Handle CreateCorona(void);
	Asura_Handle CopyCorona( Asura_Handle iSrcHandle );
	// (MH - 15.04.2004) iHandle is now passed by reference and automatically set to ASURA_INVALID_HANDLE
	void DestroyCorona(Asura_Handle& iHandle);

	void SetPosition(const Asura_Handle iHandle, const Asura_Vector_3& xPosition)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
		if(pxCorona->m_bUseModuleLinking)
		{
			if(pxCorona->m_xPosition != xPosition)
			{
				pxCorona->CalculateModuleList();
			}
		}
		#endif

		pxCorona->m_xPosition = xPosition;
		pxCorona->m_bPositionUpdatedThisFrame = true;
	}

	bool GetPosition(const Asura_Handle iHandle, Asura_Vector_3& xPosition)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return false;

		xPosition = pxCorona->m_xPosition;
		return true;
	}

	void SetDirection(const Asura_Handle iHandle, const Asura_Vector_3& xDirection)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_xDirection = xDirection;
	}
	
	void SetRadius(const Asura_Handle iHandle, float fRadius)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_fRadius = fRadius;
	}
	
	bool GetRadius(const Asura_Handle iHandle, float& fRadius)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return false;

		fRadius = pxCorona->m_fRadius;
		return true;
	}

	void SetTextureIndex(const Asura_Handle iHandle, int iTextureIndex)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_iTextureIndex = iTextureIndex;
	}
	
	
	void SetColour(const Asura_Handle iHandle, const float fRed, const float fGreen, const float fBlue)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_fRed = fRed;
		pxCorona->m_fGreen = fGreen;
		pxCorona->m_fBlue = fBlue;
	}

	void SetFlag_OnlyDrawIfPositionSetPerFrame(const Asura_Handle iHandle, bool bStatus)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_bOnlyDrawIfPositionSetPerFrame = bStatus;
	}

	void SetFlag_UseDirectionEffect(const Asura_Handle iHandle, bool bStatus)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_bUseDirectionEffect = bStatus;
	}

	void SetFlag_Hidden(const Asura_Handle iHandle, const bool bStatus)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_bIsHidden = bStatus;
	}

	void SetFlag_DontFadeWithDistance(const Asura_Handle iHandle, const bool bStatus)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_bDoNotFadeOutWithDistance = bStatus;
	}

	void SetFlag_DontFadeNear(const Asura_Handle iHandle, const bool bStatus)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_bDoNotFadeOutNear = bStatus;
	}

	void SetFlag_UseModuleLinking(const Asura_Handle iHandle, const bool bStatus)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
		if(bStatus && !pxCorona->m_bUseModuleLinking) pxCorona->CalculateModuleList();
#endif

		pxCorona->m_bUseModuleLinking = bStatus;
	}

	void SetVolumetricDepthScale(const Asura_Handle iHandle, float fVolumetricDepthScale)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);
		if (!pxCorona) return;

		pxCorona->m_fVolumetricDepthScale = fVolumetricDepthScale;
	}

private:

	Asura_TrackingHandlePool<Asura_Corona>* m_pxCoronaPool;
};


class Asura_Corona_System
{
public:
	static void Initialise(void);
	static void Reset(void);
	static void Shutdown(void);

	static void Update();
	static void Render();

	static Asura_Handle CreateCorona(void)
	{
		return GetCurrentHandler().CreateCorona();
	}

	static Asura_Handle CopyCorona( Asura_Handle iSrcHandle )
	{
		return GetCurrentHandler().CopyCorona(iSrcHandle);
	}

	// (MH - 15.04.2004) iHandle is now passed by reference and automatically set to ASURA_INVALID_HANDLE
	static void DestroyCorona(Asura_Handle& iHandle)
	{
		GetCurrentHandler().DestroyCorona(iHandle);
	}

	static void SetPosition(const Asura_Handle iHandle, const Asura_Vector_3& xPosition)
	{
		GetCurrentHandler().SetPosition(iHandle, xPosition);
	}

	static bool GetPosition(const Asura_Handle iHandle, Asura_Vector_3& xPosition)
	{
		return GetCurrentHandler().GetPosition(iHandle, xPosition);
	}

	static void SetDirection(const Asura_Handle iHandle, const Asura_Vector_3& xDirection)
	{
		GetCurrentHandler().SetDirection(iHandle, xDirection);
	}
	
	static void SetRadius(const Asura_Handle iHandle, float fRadius)
	{
		GetCurrentHandler().SetRadius(iHandle, fRadius);
	}

	static bool GetRadius(const Asura_Handle iHandle, float& fRadius)
	{
		return GetCurrentHandler().GetRadius(iHandle, fRadius);
	}

	static void SetTextureIndex(const Asura_Handle iHandle, int iTextureIndex)
	{
		GetCurrentHandler().SetTextureIndex(iHandle, iTextureIndex);
	}

	static void SetColour(const Asura_Handle iHandle, const float fRed, const float fGreen, const float fBlue)
	{
		GetCurrentHandler().SetColour(iHandle, fRed, fGreen, fBlue);
	}

	static void SetFlag_OnlyDrawIfPositionSetPerFrame(const Asura_Handle iHandle, bool bStatus)
	{
		GetCurrentHandler().SetFlag_OnlyDrawIfPositionSetPerFrame(iHandle, bStatus);
	}

	static void SetFlag_UseDirectionEffect(const Asura_Handle iHandle, bool bStatus)
	{
		GetCurrentHandler().SetFlag_UseDirectionEffect(iHandle, bStatus);
	}

	static void SetFlag_Hidden(const Asura_Handle iHandle, const bool bStatus)
	{
		GetCurrentHandler().SetFlag_Hidden(iHandle, bStatus);
	}

	static void SetFlag_DontFadeWithDistance(const Asura_Handle iHandle, const bool bStatus)
	{
		GetCurrentHandler().SetFlag_DontFadeWithDistance(iHandle, bStatus);
	}

	static void SetFlag_DontFadeNear(const Asura_Handle iHandle, const bool bStatus)
	{
		GetCurrentHandler().SetFlag_DontFadeNear(iHandle, bStatus);
	}

	static void SetFlag_UseModuleLinking(const Asura_Handle iHandle, const bool bStatus)
	{
		GetCurrentHandler().SetFlag_UseModuleLinking(iHandle, bStatus);
	}

	static void SetVolumetricDepthScale(const Asura_Handle iHandle, float fVolumetricDepthScale)
	{
		GetCurrentHandler().SetVolumetricDepthScale(iHandle, fVolumetricDepthScale);
	}

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// Handler functions...
	static	Asura_Corona_Handler*	CreateHandler(const Asura_Hash_ID uHash);
	static	bool					DestroyHandler(const Asura_Hash_ID uHash);

	static	void					SetCurrentHandler(Asura_Corona_Handler* pxHandler) { s_pxCurrentHandler = pxHandler; }
	static	void					SetCurrentHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static	Asura_Corona_Handler&	GetCurrentHandler(void) { return *s_pxCurrentHandler; }
	static	Asura_Corona_Handler*	GetHandler(const Asura_Hash_ID uHash) { return s_xHandlerTree.GetDataByHash(uHash); }

	static void						CreateAndSetHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = CreateHandler(uHash); }
#else
	// Handler functions...
	static	Asura_Corona_Handler&	GetCurrentHandler(void) { return s_xHandler; }
#endif

#ifdef ASURA_TOOLS
	//Disables rendering of coronas
	static void DisableCoronas(bool bDisable) {s_bDisabled = bDisable;} 
#endif

	static float GetDefaultVolumetricDepthScale(void);

private:

	static void RenderCorona( Asura_Corona* pxCorona );

	static void Platform_Initialise();
	static void Platform_Shutdown();

	static void Platform_RenderStart(void);
	static void	Platform_RenderCorona( const Asura_Corona* pxCorona, u_int uCamVisFlags );
	static void Platform_RenderEnd(void);

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	static void Reset_Callback(Asura_Hash_ID uHash, Asura_Corona_Handler* pxData);
	static void Render_Callback(Asura_Hash_ID uHash, Asura_Corona_Handler* pxData)	{ if (pxData) pxData->Render(); }
	static void Update_Callback(Asura_Hash_ID uHash, Asura_Corona_Handler* pxData)	{ if (pxData) pxData->Update(); }

	static Asura_Corona_Handler*						s_pxCurrentHandler;		// Currently active handler.
	static Asura_TemplateHashTree<Asura_Corona_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.
#else
	static Asura_Corona_Handler s_xHandler;				// Single static-instance of a zone.
#endif

#ifdef ASURA_TOOLS
	static bool s_bDisabled;
#endif

	friend class Asura_Corona_Handler; // So the handler can access RenderCorona, removing the need for platform changes to the corona system.

};

#endif//__ASURA_CORONA_SYSTEM_H__
