
/**
*
* \file Asura_Decal_Manager.cpp
*
* Created by Dave Reed (08/12/06)
*
* Manages decal template definitions when the decal editor is being used
*
* Copyright (C) 2006 Rebellion, All Rights Reserved.
*
**/

#include "Core/Asura.h"

#ifdef ASURA_USE_DECAL_MANAGER

#ifdef ASURA_TARGET_SM3
#include "Core/Asura_MaterialFlags.h"
#endif
#include "Core/Asura_TextureManagement.h" 
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "SpecialFX/Asura_Decal_System.h"
#include "SpecialFX/Asura_Decal_Template.h"
#include "SpecialFX/Asura_Decal_Manager.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"

#ifdef ASURA_TOOLS
#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#include "Materials/Tools_MaterialResponse_System.h"
#endif
#endif

// Decal templates for Asura_Decal_System
Asura_Decal_Template		*Asura_Decal_Manager::s_pxDecalTemplates   = NULL;

#if defined(ASURA_TARGET_SM3)
Asura_Decal_Material		*Asura_Decal_Manager::s_pxDecalMaterials   = NULL;
#endif

int							 Asura_Decal_Manager::s_iNumDecalTemplates = 0;
int							 Asura_Decal_Manager::s_iNumDecalMaterials = 0;

// Decal Manager's decal store

Asura_DecalManager_Material *Asura_Decal_Manager::s_pxMaterials		   = NULL;
Asura_DecalManager_DecalSet *Asura_Decal_Manager::s_pxSets			   = NULL;	
int							 Asura_Decal_Manager::s_iNumMaterials	   = 0;
int						     Asura_Decal_Manager::s_iNumSets		   = 0;

void Asura_Decal_Manager::Initialise()
{
#if defined(ASURA_TOOLS)
	// Load the surface properties. In retail builds, the decals are exported with each level
	Asura_ChunkLoading_System::LoadAsuraFile("misc\\decals." ASURA_STANDARD_FILE_EXTENSION);
#endif
}


void Asura_Decal_Manager::Shutdown()
{
#if defined(ASURA_TARGET_SM3)
	if (s_pxDecalMaterials)
	{
		delete [] s_pxDecalMaterials;
		s_pxDecalMaterials = NULL;
		s_iNumDecalMaterials = 0;
	}
#endif

	if (s_pxDecalTemplates)
	{
		delete [] s_pxDecalTemplates;
		s_pxDecalTemplates = NULL;
		s_iNumDecalTemplates = 0;
	}

	if (s_pxSets)
	{
		delete [] s_pxSets;
		s_pxSets = NULL;	
		s_iNumSets = 0;
	}

	if (s_pxMaterials)
	{			
		delete [] s_pxMaterials;
		s_pxMaterials = NULL;
		s_iNumMaterials = 0;
	}
}



void Asura_Decal_Manager::Reset()
{
#ifndef ASURA_DECAL_MANAGER_USE_COMMON_ASR
	Shutdown();
	Initialise();
#endif // #ifndef ASURA_DECAL_MANAGER_USE_COMMON_ASR	
}



// This rebuilds the Asura lists of decal templates and material
// from the contents of the decal manager
void Asura_Decal_Manager::RebuildTemplates(bool bResetDecalSystem)
{
#if defined(ASURA_TARGET_SM3)
	if (s_pxDecalMaterials)
	{
		delete [] s_pxDecalMaterials;
		s_pxDecalMaterials = NULL;
		s_iNumDecalMaterials = 0;
	}
#endif //defined(ASURA_TARGET_SM3)

	if (s_pxDecalTemplates)
	{
		delete [] s_pxDecalTemplates;
		s_pxDecalTemplates = NULL;
		s_iNumDecalTemplates = 0;
	}

	// First create the material templates
	// We create an extra material, to ensure there's always at least one, at index 0
	s_iNumDecalMaterials = GetNumDecalMaterials()+1;

#if defined(ASURA_TARGET_SM3)
	s_pxDecalMaterials = new Asura_Decal_Material[s_iNumDecalMaterials];
	// set 0th entry to some valid defaults
	Asura_Mem::Set(&s_pxDecalMaterials[0], 0, sizeof(Asura_Decal_Material)*s_iNumDecalMaterials);
	s_pxDecalMaterials[0].m_iMaxDecals = 1;
	s_pxDecalMaterials[0].m_fFadeInTime = 0.001f;
	s_pxDecalMaterials[0].m_fFadeOutTime = 0.001f;
#endif //defined(ASURA_TARGET_SM3)

	for(int i=0; i<GetNumDecalMaterials(); i++)
	{
		Asura_DecalManager_Material  *pxSrcMtl = GetDecalMaterialByIndex(i);
#if defined(ASURA_TARGET_SM3)	
		Asura_Decal_Material		 *pxMtl = &s_pxDecalMaterials[i+1];
	
		pxMtl->m_fDecalStayTime      = pxSrcMtl->m_fLifeTime;
		pxMtl->m_fFadeInTime         = pxSrcMtl->m_fFadeInLength;
		pxMtl->m_fFadeOutTime		 = pxSrcMtl->m_fFadeOutLength;
		pxMtl->m_iMaxDecals			 = pxSrcMtl->m_iMaxDecals;

		pxMtl->m_fUScrollRange		 = pxSrcMtl->m_fUScrollRange;
		pxMtl->m_fVScrollRange		 = pxSrcMtl->m_fVScrollRange;
		pxMtl->m_fUScrollSpeed		 = pxSrcMtl->m_fUScrollSpeed;
		pxMtl->m_fVScrollSpeed		 = pxSrcMtl->m_fVScrollSpeed;
			
		if (!pxSrcMtl->m_bEnableFades)
		{		
			pxMtl->m_fDecalStayTime = -1.0f;
		}

		// This is a workaround for a bug in the decal renderer - could do with
		// a better fix in the renderer code, really
		if (pxMtl->m_fFadeInTime<=0.0f) pxMtl->m_fFadeInTime = 0.001f;
		if (pxMtl->m_fFadeOutTime<=0.0f) pxMtl->m_fFadeOutTime = 0.001f;
#endif //defined(ASURA_TARGET_SM3)

#ifdef ASURA_TOOLS
	#if !defined(ASURA_TARGET_SM3)
		// In tools builds, we need to load the textures
		Asura_Texture_Management::AddTexture(pxSrcMtl->m_xTextureName);
	#endif //defined(ASURA_TARGET_SM3)
		Asura_Texture_Management::LoadTextures();

#endif

#if defined(ASURA_TARGET_SM3)
		pxMtl->m_uMaterialResponseHash = pxSrcMtl->m_uMaterialResponseHash;
#else		
		pxSrcMtl->m_iTextureIndex	= Asura_Texture_Management::GetIndexFromHashID(Asura_GetHashID_IgnoreExtension(pxSrcMtl->m_xTextureName.GetString()));
#endif //defined(ASURA_TARGET_SM3)

	}

	// Now create the decal templates
	// Again, index 0 is always created, as a default
	int iCount = 0;
	for(int i=0;i<GetNumDecalSets();i++)
	{
		iCount += GetDecalSetByIndex(i)->m_iNumDecals;
	}

	s_iNumDecalTemplates = iCount+1;
	s_pxDecalTemplates = new Asura_Decal_Template [s_iNumDecalTemplates];
	// set the 0th template to all zeroes
	Asura_Mem::Set(&s_pxDecalTemplates[0], 0, sizeof(Asura_Decal_Template) * s_iNumDecalTemplates);
	// and some valid defaults
#ifdef ASURA_TARGET_SM3
	s_pxDecalTemplates[0].m_iDecalMaterialIndex = ASURA_MATERIAL_NONE;
	s_pxDecalTemplates[0].m_uFlags = ASURA_DECAL_LIGHTONCREATION;
#else
	s_pxDecalTemplates[0].m_iTextureIndex = ASURA_TEXTURE_NONE;
#endif

	iCount = 0;
	for(int i=0; i<GetNumDecalSets(); i++)
	{
		Asura_DecalManager_DecalSet *pxSet = GetDecalSetByIndex(i);

		pxSet->m_iFirstAsuraDecalTemplate = iCount+1;
#if defined(ASURA_TARGET_SM3)
		pxSet->m_iAsuraDecalMaterial = GetMaterialIndex(pxSet->m_uMtlNameHash)+1;
#else
		pxSet->m_iAsuraDecalTexture = GetTextureIndex(pxSet->m_uMtlNameHash);
#endif

		for(int j=0;j<pxSet->m_iNumDecals; j++)
		{
			Asura_Decal_Template *pxTmp = &s_pxDecalTemplates[iCount+1];
			Asura_DecalManager_Decal *pxSrcTmp = &pxSet->m_pxDecals[j];

			pxTmp->m_fWidth  = pxSrcTmp->m_fWidth;
			pxTmp->m_fHeight = pxSrcTmp->m_fHeight;
			pxTmp->m_fDepth  = 0.5f*(pxTmp->m_fWidth+pxTmp->m_fHeight);

			// Decals were coming out y-flipped, so I'm flipping the UVs here.
			pxTmp->m_fU      = pxSrcTmp->m_fU;
			pxTmp->m_fV      = pxSrcTmp->m_fV+pxSrcTmp->m_fDeltaV;
			pxTmp->m_fDeltaU = pxSrcTmp->m_fDeltaU;
			pxTmp->m_fDeltaV = -pxSrcTmp->m_fDeltaV;			

			if (pxSrcTmp->m_iFlags & ASURA_DECALMANAGER_FLAG_UFLIP)
			{
				pxTmp->m_fU += pxSrcTmp->m_fDeltaU;
				pxTmp->m_fDeltaU =- pxTmp->m_fDeltaU;
			}

			if (pxSrcTmp->m_iFlags & ASURA_DECALMANAGER_FLAG_VFLIP)
			{
				pxTmp->m_fV += pxSrcTmp->m_fDeltaV;
				pxTmp->m_fDeltaV =- pxTmp->m_fDeltaV;
			}

			pxTmp->m_uFlags  = 0;
#if defined(ASURA_TARGET_SM3)
			pxTmp->m_iDecalMaterialIndex	= pxSet->m_iAsuraDecalMaterial;
			pxTmp->m_uFlags = ASURA_DECAL_LIGHTONCREATION;
#else
			pxTmp->m_iTextureIndex			= pxSet->m_iAsuraDecalTexture;
#endif	

			iCount++;
		}
	}

	 if (bResetDecalSystem)
	 {	 
		// Reset the decal system, as there may be existing decals referencing
		// now-invalid materials (in tools builds)
		Asura_Decal_System::Reset();
	 }
}




void Asura_Decal_Manager::PostLoadInitialise(void)
{
	for(int i=0; i<GetNumDecalMaterials(); i++)
	{
		Asura_DecalManager_Material  *pxSrcMtl = GetDecalMaterialByIndex(i);
#if defined(ASURA_TARGET_SM3)	
		Asura_Decal_Material		 *pxMtl = &s_pxDecalMaterials[i+1];

		pxMtl->m_uMaterialResponseHash = pxSrcMtl->m_uMaterialResponseHash;
#else
		pxSrcMtl->m_iTextureIndex	= Asura_Texture_Management::GetIndexFromHashID(Asura_GetHashID_IgnoreExtension(pxSrcMtl->m_xTextureName.GetString()));
#endif //defined(ASURA_TARGET_SM3)
	}
}


Asura_DecalManager_DecalSet *Asura_Decal_Manager::GetDecalSet(uint32 uNameHash)
{
	for(int i=0;i<s_iNumSets;i++)
	{
		if (s_pxSets[i].m_uSetNameHash==uNameHash)
		{
			return &s_pxSets[i];
		}
	}
	return NULL;
}


Asura_DecalManager_Material *Asura_Decal_Manager::GetMaterial(uint32 uNameHash)
{	
	for(int i=0;i<s_iNumMaterials;i++)
	{
		if (s_pxMaterials[i].m_uMtlNameHash==uNameHash)
		{
			return &s_pxMaterials[i];
		}
	}
	return NULL;
}

#if defined(ASURA_TARGET_SM3)

int Asura_Decal_Manager::GetMaterialIndex(uint32 uNameHash)
{	
	for(int i=0;i<s_iNumMaterials;i++)
	{
		if (s_pxMaterials[i].m_uMtlNameHash==uNameHash)
		{
			return i;
		}
	}
	return -1;
}
#else
int Asura_Decal_Manager::GetTextureIndex(uint32 uNameHash)
{	
	for(int i=0;i<s_iNumMaterials;i++)
	{
		if (s_pxMaterials[i].m_uMtlNameHash==uNameHash)
		{
			return s_pxMaterials[i].m_iTextureIndex;
		}
	}
	return -1;
}
#endif //defined(ASURA_TARGET_SM3)

int Asura_Decal_Manager::GetDecalSetIndex(uint32 uNameHash)
{	
	for(int i=0;i<s_iNumSets;i++)
	{
		if (s_pxSets[i].m_uSetNameHash==uNameHash)
		{
			return i;
		}
	}
	return -1;
}


int	Asura_Decal_Manager::GetRandomDecalFromSet(uint32 uNameHash)
{
	Asura_DecalManager_DecalSet *pxSet = GetDecalSet(uNameHash);

	if (pxSet)
	{
		return pxSet->m_iFirstAsuraDecalTemplate + Asura_Random::GetInt(0,pxSet->m_iNumDecals-1);
	}
	else
	{
		return 0;
	}
}



#ifdef ASURA_TOOLS

// ---------------------------------------------------------------------------------------------
// Tools-only DecalManager functions, for dynamically adding/removing sets/materials/decals
// ---------------------------------------------------------------------------------------------


Asura_DecalManager_Material *Asura_Decal_Manager::AddDecalMaterial(Asura_String xName)
{
	for(int i=0; i<s_iNumMaterials; i++)
	{
		if (s_pxMaterials[i].m_uMtlNameHash == Asura_GetHashID(xName.GetString()))
		{
			// Name already exists, can't create this one
			return NULL;
		}
	}

	Asura_DecalManager_Material *pxNewMtls = new Asura_DecalManager_Material[s_iNumMaterials+1];

	if (s_pxMaterials)
	{
		for(int i=0; i<s_iNumMaterials; i++)
		{
			pxNewMtls[i] = s_pxMaterials[i];
		}
		delete [] s_pxMaterials;
	}

	s_iNumMaterials++;
	s_pxMaterials = pxNewMtls;	
	s_pxMaterials[s_iNumMaterials-1].m_xMtlName = xName;
	s_pxMaterials[s_iNumMaterials-1].m_uMtlNameHash = Asura_GetHashID(xName.GetString());

	return &s_pxMaterials[s_iNumMaterials-1];
}

void Asura_Decal_Manager::RemoveDecalMaterial(int iIndex)
{
	// Note - we don't shrink the array, easier to just leave the unused elements on the end
	for(int i=iIndex; i<s_iNumMaterials-1;i++)
	{
		s_pxMaterials[i]=s_pxMaterials[i+1];		
	}
	s_iNumMaterials--;
}

Asura_DecalManager_DecalSet *Asura_Decal_Manager::AddDecalSet(Asura_String xName)
{
	for(int i=0; i<s_iNumSets; i++)
	{
		if (s_pxSets[i].m_uSetNameHash == Asura_GetHashID(xName.GetString()))
		{
			// Name already exists, can't create this one
			return NULL;
		}
	}

	Asura_DecalManager_DecalSet *pxNewSets = new Asura_DecalManager_DecalSet[s_iNumSets+1];

	if (s_pxSets)
	{
		for(int i=0;i<s_iNumSets;i++)
		{
			pxNewSets[i] = s_pxSets[i];
		}
		delete [] s_pxSets;
	}

	s_iNumSets++;
	s_pxSets = pxNewSets;	
	s_pxSets[s_iNumSets-1].m_xSetName = xName;
	s_pxSets[s_iNumSets-1].m_uSetNameHash = Asura_GetHashID(xName.GetString());

	return &s_pxSets[s_iNumSets-1];
}


void Asura_Decal_Manager::RemoveDecalSet(int iIndex)
{
	// Note - we don't shrink the array, easier to just leave the unused elements on the end
	for(int i=iIndex; i<s_iNumMaterials-1;i++)
	{
		s_pxSets[i]=s_pxSets[i+1];		
	}
	s_iNumSets--;
}



void Asura_DecalManager_DecalSet::ChangeNumDecalTypes(int iCount)
{
	Asura_DecalManager_Decal *pxNewDecals = NULL;

	if (iCount>0)
	{	
		pxNewDecals = new Asura_DecalManager_Decal[iCount];

		int iMin = iCount;
		if (m_iNumDecals<iCount)
		{
			iMin = m_iNumDecals;
		}

		for(int i=0;i<iMin;i++)
		{
			pxNewDecals[i] = m_pxDecals[i];
		}
	}

	if (m_pxDecals)
	{
		delete [] m_pxDecals;
	}	

	m_iNumDecals = iCount;
	m_pxDecals = pxNewDecals;
}

#endif // #ifdef ASURA_TOOLS



// Project functions should call these. If you want to keep hard-coded/code-generated templates,
// use a different range of IDs for them and remap them within the project functions?

const Asura_Decal_Template* Asura_Decal_Manager::GetDecalTemplate(u_int uDecalTemplateID)
{
	Asura_Assert(uDecalTemplateID < (uint32)s_iNumDecalTemplates, "Invalid decal ID");

	if(uDecalTemplateID < ((uint32)s_iNumDecalTemplates))
	{		
		return &s_pxDecalTemplates[uDecalTemplateID]; 		
	}
	else
	{		
		return &s_pxDecalTemplates[0];
	}
}

#if defined(ASURA_TARGET_SM3)
const Asura_Decal_Material* Asura_Decal_Manager::GetDecalMaterial(u_int uDecalMaterialID)
{
	Asura_Assert(uDecalMaterialID < (uint32)s_iNumDecalMaterials, "Invalid decal material ID");

	if(uDecalMaterialID < ((uint32)s_iNumDecalMaterials))
	{		
		return &s_pxDecalMaterials[uDecalMaterialID]; 
	}
	else
	{
		return &s_pxDecalMaterials[0];
	}
}
#endif

void Asura_Decal_Manager::AddDecal(const Asura_Vector_3 &xPos, const Asura_Vector_3 &xNormal, uint32 uDecalSetHash, int iDecalIndex, float fScale, float fOrientation)
{
	Asura_DecalManager_DecalSet *pxSet = GetDecalSet(uDecalSetHash);				
	
	if (pxSet)
	{	
		if (pxSet->m_iNumDecals>0)
		{		
			int iTemplateIndex = pxSet->m_iFirstAsuraDecalTemplate;

			if (iDecalIndex<0 || iDecalIndex>=pxSet->m_iNumDecals)
			{
				iDecalIndex = Asura_Random::GetInt(0,pxSet->m_iNumDecals-1);
			}

			if (pxSet->m_bRandomOrientation)
			{
				fOrientation = Asura_Random::GetFloat(0,6.284f);
			}

			if (iTemplateIndex+iDecalIndex >=0 && iTemplateIndex+iDecalIndex < s_iNumDecalTemplates)
			{			
				Asura_Decal_System::AddDecal(xPos,xNormal,iTemplateIndex+iDecalIndex,fScale,fOrientation);
			}
			else
			{
				Asura_Assert(false, "Out-of-range decal!");
			}
		}
	}
}

// Add decal, but also specify the decal top direction.
void Asura_Decal_Manager::AddDecal(const Asura_Vector_3& xPos,
								   const Asura_Vector_3& xNormal,
								   const Asura_Vector_3& xTopDir,
								   uint32 uDecalSetHash,
								   int iDecalIndex /*= -1*/,
								   float fScale /*= 1.0f*/,
								   float fOrientation /*= 0.0f*/)
{
	const Asura_DecalManager_DecalSet* const pxSet = GetDecalSet(uDecalSetHash);
	
	if (pxSet && pxSet->m_iNumDecals>0)
	{	
		const int iTemplateIndex = pxSet->m_iFirstAsuraDecalTemplate;

		if (iDecalIndex < 0 || iDecalIndex >= pxSet->m_iNumDecals)
		{
			iDecalIndex = Asura_Random::GetInt(0, pxSet->m_iNumDecals-1);
		}

		if (pxSet->m_bRandomOrientation)
		{
			fOrientation = Asura_Random::GetFloat(0, Asura_Maths::Deg360);
		}

		if (iTemplateIndex+iDecalIndex >=0 && iTemplateIndex+iDecalIndex < s_iNumDecalTemplates)
		{			
			Asura_Decal_System::AddDecal(xPos, xNormal, xTopDir, iTemplateIndex+iDecalIndex, fScale, fOrientation);
		}
		else
		{
			Asura_Assert(false, "Out-of-range decal!");
		}
	}
}



#ifdef ASURA_TOOLS
int Asura_Decal_Manager::GetTextureUsage( uint32 uTextureNameHash )
{
	return 0;
}
#endif


#endif // ASURA_USE_DECAL_MANAGER

