//
// Asura_PerlinNoise_System.cpp
//
// Created by KOR - 16/12/2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "SpecialFX/Asura_PerlinNoise_System.h"

// Static data
int				Asura_PerlinNoise_System::s_aiPermutation[ASURA_PERLINNOISE_REPEAT_DISTANCE * 2 + 2];
Asura_Vector_3	Asura_PerlinNoise_System::s_axGradients[ASURA_PERLINNOISE_REPEAT_DISTANCE * 2 + 2];
bool			Asura_PerlinNoise_System::s_bNoiseInitialised = false;

void Asura_PerlinNoise_System::Initialise()
{
	Asura_Assert(!s_bNoiseInitialised, "Noise system already initialised");
	if (s_bNoiseInitialised) return;
	s_bNoiseInitialised = true;

	int i, j, k;
	// Generate random coefficients
	for (i = 0 ; i < ASURA_PERLINNOISE_REPEAT_DISTANCE ; i++)
	{
		s_aiPermutation[i] = i;
		s_axGradients[i].RandomDirection();
	}
	// Generate permutation table
	while (--i)
	{
		k = s_aiPermutation[i];
		s_aiPermutation[i] = s_aiPermutation[j = Asura_Random::GetIntU() % ASURA_PERLINNOISE_REPEAT_DISTANCE];
		s_aiPermutation[j] = k;
	}
	// Extend tables
	for (i = 0 ; i < ASURA_PERLINNOISE_REPEAT_DISTANCE + 2 ; i++)
	{
		s_aiPermutation[ASURA_PERLINNOISE_REPEAT_DISTANCE + i] = s_aiPermutation[i];
		s_axGradients[ASURA_PERLINNOISE_REPEAT_DISTANCE + i] = s_axGradients[i];
	}
}

// Cubic approximation to weight function for drop-off. (__fValue between 0 and 1)
// The clever thing about this is that s_curve(t - 1) = (1 - s_curve(t)),
// so we can sum two wavelets using the lerp function
#define s_curve(__fValue) ( __fValue * __fValue * (3.f - 2.f * __fValue) )

// Ronseal
#define lerp(__fScale, __fValue1, __fValue2) ( __fValue1 + __fScale * (__fValue2 - __fValue1) )

// Setup macro does the following
// Offsets value by a large positive amount to prevent negative value weirdnesses
// Sets iB0 to low grid coordinate
// Sets iB1 to high grid coordinate
// Sets fR0 to fractional part of value
// Sets fR1 to fR0 - 1 (i.e. offset from high grid coordinate
#define setup(i,iB0,iB1,fR0,fR1)\
{\
	float fTemp = afVec[i] + ASURA_PERLINNOISE_OFFSET;\
	int iTemp = Asura_Maths::FloatToInt_Chop(fTemp);\
	iB0 = iTemp & ASURA_PERLINNOISE_BITMASK;\
	iB1 = iB0+1;\
	fR0 = fTemp - iTemp;\
	fR1 = fR0 - 1.f;\
}

float Asura_PerlinNoise_System::GetNoise(float fValue, float* pfGradient /* = NULL */)
{
	Asura_Assert(s_bNoiseInitialised, "Noise system not initialised");
	// High and low grid references
	int iBX0, iBX1;
	// Fractional offsets from grid references
	float fRX0, fRX1;
	// Linear wavelet values
	float fU, fV;
	float afVec[1] = { fValue };

	setup(0, iBX0, iBX1, fRX0, fRX1);

	// Compute influence from wavelets
	// Linear function
	fU = fRX0 * s_axGradients[ s_aiPermutation[ iBX0 ] ].x;
	fV = fRX1 * s_axGradients[ s_aiPermutation[ iBX1 ] ].x;

	if(pfGradient)
	{
		*pfGradient = s_axGradients[ s_aiPermutation[ iBX0 ] ].x * (1 - 9 * fRX0 * fRX0 + 8 * fRX0 * fRX0 * fRX0)
					- s_axGradients[ s_aiPermutation[ iBX1 ] ].x * (6 * fRX0 - 15 * fRX0 * fRX0 + 8 * fRX0 * fRX0 * fRX0);
	}
	// Lerp using weight function.
	return lerp(s_curve(fRX0), fU, fV);
}

float Asura_PerlinNoise_System::GetNoise2D(const float* afVec, Asura_Vector_2* pxGradient /* = NULL */)
{
	Asura_Assert(s_bNoiseInitialised, "Noise system not initialised");
	int iBX0, iBX1, iBY0, iBY1, iB00, iB10, iB01, iB11;
	float fRX0, fRX1, fRY0, fRY1, fSX, fSY, fA, fB, fBX, fU, fV;
	float fAX = 0.0f;
	Asura_Vector_3 *pxQ00, *pxQ01, *pxQ10, *pxQ11;
	int i, j;

	// Set up x axis
	setup(0, iBX0,iBX1, fRX0,fRX1);
	// Set up y axis
	setup(1, iBY0,iBY1, fRY0,fRY1);

	i = s_aiPermutation[ iBX0 ];
	j = s_aiPermutation[ iBX1 ];

	iB00 = s_aiPermutation[ i + iBY0 ];
	iB10 = s_aiPermutation[ j + iBY0 ];
	iB01 = s_aiPermutation[ i + iBY1 ];
	iB11 = s_aiPermutation[ j + iBY1 ];

	fSX = s_curve(fRX0);
	fSY = s_curve(fRY0);

	pxQ00 = s_axGradients + iB00; fU = fRX0 * pxQ00->x + fRY0 * pxQ00->y;//DotProdWithQ2(fRX0,fRY0);
	pxQ10 = s_axGradients + iB10; fV = fRX1 * pxQ10->x + fRY0 * pxQ10->y;//DotProdWithQ2(fRX1,fRY0);
	fA = lerp(fSX, fU, fV);
	if(pxGradient)	fAX = (6.f * fRX0 * fRX1) * (fU - fV) + lerp(fSX, pxQ00->x, pxQ10->x);

	pxQ01 = s_axGradients + iB01; fU = fRX0 * pxQ01->x + fRY1 * pxQ01->y;//DotProdWithQ2(fRX0,fRY1);
	pxQ11 = s_axGradients + iB11; fV = fRX1 * pxQ11->x + fRY1 * pxQ11->y;//DotProdWithQ2(fRX1,fRY1);
	fB = lerp(fSX, fU, fV);
	if(pxGradient)
	{
		fBX = (6.f * fRX0 * fRX1) * (fU - fV) + lerp(fSX, pxQ01->x, pxQ11->x);

		pxGradient->x = lerp(fSY, fAX, fBX);
		pxGradient->y = (6.f * fRY0 * fRY1) * (fA - fB) + (1 - fSY) * (lerp(fSX, pxQ00->y, pxQ10->y)) + fSY * (lerp(fSX, pxQ01->y, pxQ11->y));
	}

	return lerp(fSY, fA, fB);;
}

float Asura_PerlinNoise_System::GetNoise3D(const float* afVec, Asura_Vector_2* pxGradient /* = NULL */)
{
	Asura_Assert(s_bNoiseInitialised, "Noise system not initialised");
	int iBX0, iBX1, iBY0, iBY1, iBZ0, iBZ1, iB00, iB10, iB01, iB11;
	float fRX0, fRX1, fRY0, fRY1, fRZ0, fRZ1, fSX, fSY, fSZ, fA, fB, fC, fD, fU, fV;
//	Asura_Vector_3* pxQ;
	int i, j;
	float fAX = 0.0f;
	float fBX = 0.0f;
	float fCX = 0.0f;
	float fCY = 0.0f;
	float fDX, fDY;
	Asura_Vector_3 *pxQ00, *pxQ01, *pxQ10, *pxQ11;

	// Set up x axis
	setup(0, iBX0,iBX1, fRX0,fRX1);
	// Set up y axis
	setup(1, iBY0,iBY1, fRY0,fRY1);
	// Set up z axis
	setup(2, iBZ0,iBZ1, fRZ0,fRZ1);

	i = s_aiPermutation[ iBX0 ];
	j = s_aiPermutation[ iBX1 ];

	iB00 = s_aiPermutation[ i + iBY0 ];
	iB10 = s_aiPermutation[ j + iBY0 ];
	iB01 = s_aiPermutation[ i + iBY1 ];
	iB11 = s_aiPermutation[ j + iBY1 ];

	fSX = s_curve(fRX0);
	fSY = s_curve(fRY0);
	fSZ = s_curve(fRZ0);

#define DotProdWithQ3(rx,ry,rz) ( rx * pxQ->x + ry * pxQ->y + rz * pxQ->z )
	// Calculate values at z0
	// Calculate value between x0 and x1 at y0
	pxQ00 = s_axGradients + iB00 + iBZ0; fU = fRX0 * pxQ00->x + fRY0 * pxQ00->y + fRZ0 * pxQ00->z;//DotProdWithQ3(fRX0,fRY0,fRZ0);
	pxQ10 = s_axGradients + iB10 + iBZ0; fV = fRX1 * pxQ10->x + fRY0 * pxQ10->y + fRZ0 * pxQ10->z;//DotProdWithQ3(fRX1,fRY0,fRZ0);
	fA = lerp(fSX, fU, fV);
	if(pxGradient) fAX = (6 * fRX0 * fRX1) * (fU - fV) + lerp(fSX, pxQ00->x, pxQ10->x);
	// Calculate value between x0 and x1 at y1
	pxQ01 = s_axGradients + iB01 + iBZ0; fU = fRX0 * pxQ01->x + fRY1 * pxQ01->y + fRZ0 * pxQ01->z;//DotProdWithQ3(fRX0,fRY1,fRZ0);
	pxQ11 = s_axGradients + iB11 + iBZ0; fV = fRX1 * pxQ11->x + fRY1 * pxQ11->y + fRZ0 * pxQ11->z;//DotProdWithQ3(fRX1,fRY1,fRZ0);
	fB = lerp(fSX, fU, fV);
	if(pxGradient) fBX = (6 * fRX0 * fRX1) * (fU - fV) + lerp(fSX, pxQ01->x, pxQ11->x);
	// Get overall value for z0
	fC = lerp(fSY, fA, fB);
	if(pxGradient) fCX = lerp(fSY, fAX, fBX);
	if(pxGradient) fCY = (6.f * fRY0 * fRY1) * (fA - fB) + (1 - fSY) * (lerp(fSX, pxQ00->y, pxQ10->y)) + fSY * (lerp(fSX, pxQ01->y, pxQ11->y));

	// Calculate value between x0 and x1 at y0
	pxQ00 = s_axGradients + iB00 + iBZ1; fU = fRX0 * pxQ00->x + fRY0 * pxQ00->y + fRZ1 * pxQ00->z;//DotProdWithQ3(fRX0,fRY0,fRZ1);
	pxQ10 = s_axGradients + iB10 + iBZ1; fV = fRX1 * pxQ10->x + fRY0 * pxQ10->y + fRZ1 * pxQ10->z;//DotProdWithQ3(fRX1,fRY0,fRZ1);
	fA = lerp(fSX, fU, fV);
	if(pxGradient) fAX = (6 * fRX0 * fRX1) * (fU - fV) + lerp(fSX, pxQ00->x, pxQ10->x);
	// Calculate value between x0 and x1 at y1
	pxQ01 = s_axGradients + iB01 + iBZ1; fU = fRX0 * pxQ01->x + fRY1 * pxQ01->y + fRZ1 * pxQ01->z;//DotProdWithQ3(fRX0,fRY1,fRZ1);
	pxQ11 = s_axGradients + iB11 + iBZ1; fV = fRX1 * pxQ11->x + fRY1 * pxQ11->y + fRZ1 * pxQ11->z;//DotProdWithQ3(fRX1,fRY1,fRZ1);
	fB = lerp(fSX, fU, fV);
	if(pxGradient) fBX = (6 * fRX0 * fRX1) * (fU - fV) + lerp(fSX, pxQ01->x, pxQ11->x);
	// Get overall value for z1
	fD = lerp(fSY, fA, fB);
	if(pxGradient)
	{
		fDX = lerp(fSY, fAX, fBX);
		fDY = (6.f * fRY0 * fRY1) * (fA - fB) + (1 - fSY) * (lerp(fSX, pxQ00->y, pxQ10->y)) + fSY * (lerp(fSX, pxQ01->y, pxQ11->y));

		pxGradient->x = lerp(fSZ, fCX, fDX);
		pxGradient->y = lerp(fSZ, fCY, fDY);
	}
	// Get value between z0 and z1
	return lerp(fSZ, fC, fD);
}



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PerlinNoise_System::CalculateNoiseVector
// Author       : Andy Tate
// Date         : 8-Dec-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PerlinNoise_System::CalculateNoiseVector (const Asura_Vector_3& xPosition,
													 Asura_Vector_3& xVector)
{
	const float afVec0[] = { xPosition[1], xPosition[2] };
	xVector[0]	= GetNoise2D(afVec0);
	const float afVec1[] = { xPosition[2], xPosition[0] };
	xVector[1]	= GetNoise2D(afVec1);
	const float afVec2[] = { xPosition[0], xPosition[1] };
	xVector[2]	= GetNoise2D(afVec2);
}

/// End of function 'Asura_PerlinNoise_System::CalculateNoiseVector'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PerlinNoise_System::CalculateNoiseVector
// Author       : Andy Tate
// Date         : 8-Dec-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PerlinNoise_System::CalculateNoiseVector (const Asura_AVector3& xPosition,
													 Asura_AVector3& xVector)
{
	const float afVec0[] = { xPosition.GetY(), xPosition.GetZ() };
	xVector.SetX(GetNoise2D(afVec0));
	const float afVec1[] = { xPosition.GetZ(), xPosition.GetX() };
	xVector.SetY(GetNoise2D(afVec1));
	const float afVec2[] = { xPosition.GetX(), xPosition.GetY() };
	xVector.SetZ(GetNoise2D(afVec2));
}

/// End of function 'Asura_PerlinNoise_System::CalculateNoiseVector'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PerlinNoise_System::CalculateNoiseVector
// Author       : Andy Tate
// Date         : 8-Dec-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PerlinNoise_System::CalculateNoiseVector (const Asura_Vector_3& xPosition,
													 const float fTime,
													 Asura_Vector_3& xVector)
{
	const float afVec0[] = { xPosition[1], xPosition[2], fTime };
	xVector[0]	= GetNoise3D(afVec0);
	const float afVec1[] = { xPosition[2], xPosition[0], fTime };
	xVector[1]	= GetNoise3D(afVec1);
	const float afVec2[] = { xPosition[0], xPosition[1], fTime };
	xVector[2]	= GetNoise3D(afVec2);
}

/// End of function 'Asura_PerlinNoise_System::CalculateNoiseVector'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PerlinNoise_System::CalculateNoiseVector
// Author       : Andy Tate
// Date         : 8-Dec-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PerlinNoise_System::CalculateNoiseVector (const Asura_AVector3& xPosition,
													 const float fTime,
													 Asura_AVector3& xVector)
{
	const float afVec0[] = { xPosition.GetY(), xPosition.GetZ(), fTime };
	xVector.SetX(GetNoise3D(afVec0));
	const float afVec1[] = { xPosition.GetZ(), xPosition.GetX(), fTime };
	xVector.SetY(GetNoise3D(afVec1));
	const float afVec2[] = { xPosition.GetX(), xPosition.GetY(), fTime };
	xVector.SetZ(GetNoise3D(afVec2));
}

/// End of function 'Asura_PerlinNoise_System::CalculateNoiseVector'. ///


/// End of file. ///
