//
// Asura_ShadowDecal_System.cpp
//
// Created by KJFL - 15 March 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "SpecialFX/Asura_ShadowDecal_System.h"
#include "SpecialFX/Asura_ShadowDecal_System_Internal.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Colour.h"


#include "Physics/Asura_Physics_Environment.h"
#include "Physics/Asura_Physics_EnvironmentCache.h"

#ifndef ASURA_SHADOW_DECAL_MAX_RENDER_DIST_SQRD
	#define ASURA_SHADOW_DECAL_MAX_RENDER_DIST_SQRD (100.0f) // 10 metres by define
#endif

const u_int uNUMBER_OF_SHADOW_CACHES = 16;
const float fENV_CACHE_BOUNDARY = 0.1f;

Asura_ShadowDecal_Cache* Asura_ShadowDecal_System::s_pxShadowCache;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_ShadowDecal_System::Initialise(void)
{
	Platform_Initialise();

	s_pxShadowCache = new Asura_ShadowDecal_Cache[ uNUMBER_OF_SHADOW_CACHES ];
}

void Asura_ShadowDecal_System::Reset(void)
{
	Platform_Reset();

	if(s_pxShadowCache)
	{
		for(u_int u = 0; u < uNUMBER_OF_SHADOW_CACHES; ++u)
		{
			s_pxShadowCache[u].Reset();
		}
	}
}

void Asura_ShadowDecal_System::Shutdown(void)
{
	Platform_Shutdown();

	delete [] s_pxShadowCache;
	s_pxShadowCache = 0;
}

void Asura_ShadowDecal_System::AddShadow( const Asura_Bounding_Box& xBB, const class Asura_Animation* pxAnim )
{
	u_int uBestCacheForRecycling = 0;

	for( u_int u=0; u<uNUMBER_OF_SHADOW_CACHES; ++u )
	{
		Asura_ShadowDecal_Cache& xCache = s_pxShadowCache[u];

		// if we already have a cache, update it and then we're done
		if ( xCache.IsCacheForAnimation( pxAnim ) )
		{
			xCache.Update( xBB, pxAnim );
			return;
		}
		// otherwise if this isn't our cache, see if it's a valid one for recycling,
		// just in case we don't actually have a valid cache.
		else
		{
			if ( xCache.GetFrameStamp() < s_pxShadowCache[uBestCacheForRecycling].GetFrameStamp() )
			{
				uBestCacheForRecycling = u;
			}
		}
	}

	// right, looks like we need to recycle a cache then.
	Asura_ShadowDecal_Cache& xCache = s_pxShadowCache[uBestCacheForRecycling];

	xCache.Reset();
	xCache.Update( xBB, pxAnim );
}

void Asura_ShadowDecal_Cache::Update( const Asura_Bounding_Box& xBB, const Asura_Animation* pxAnim )
{
	m_pxAnimation = pxAnim;

	m_uFrameStamp = Asura_Timers::GetFrameNumber();

    
	{
		// KJFL - 24 March 2005 - test to see if we'll notice shadows popping out after 10m
		// WARNING: metric ignores splitscreen
		Asura_Camera::SetCameraIndex();
		const Asura_Vector_3 xDisp = xBB.CalculateCentre() - *Asura_Camera::GetPosition();
		if (xDisp.MagnitudeSquared() > ASURA_SHADOW_DECAL_MAX_RENDER_DIST_SQRD) return;
	}
	

	const float fHeight	= (xBB.MaxZ-xBB.MinZ);
	const float fWidth	= (xBB.MaxX-xBB.MinX);
	const float fDepth	= 1.2f;

	// If height or width is zero, return otherwise we'll get a divide by zero later
	if( (fHeight == 0.0f) || (fWidth == 0.0f) ) return; 

	Asura_Vector_3 xPosition = xBB.CalculateCentre();
	xPosition.y = xBB.MaxY;

	Asura_Bounding_Box xBox;
	xBox.Reset();
	
	{
		const Asura_Vector_3 xUOffset(-fWidth, 0.0f, 0.0f);
		const Asura_Vector_3 xVOffset( 0.0f, 0.0f, fHeight);
		const Asura_Vector_3 xNOffset( 0.0f,-fDepth, 0.0f);
		const Asura_Vector_3 xUpOffset( 0.0f, -0.2f, 0.0f);

		Asura_Vector_3 xVertex1 = -xUOffset -xVOffset +xUpOffset; 
		Asura_Vector_3 xVertex2 =  xUOffset -xVOffset +xUpOffset; 
		Asura_Vector_3 xVertex3 = -xUOffset -xVOffset -xNOffset; 
		Asura_Vector_3 xVertex4 =  xUOffset -xVOffset -xNOffset; 
		Asura_Vector_3 xVertex5 = -xUOffset +xVOffset +xUpOffset; 
		Asura_Vector_3 xVertex6 =  xUOffset +xVOffset +xUpOffset; 
		Asura_Vector_3 xVertex7 = -xUOffset +xVOffset -xNOffset; 
		Asura_Vector_3 xVertex8 =  xUOffset +xVOffset -xNOffset; 

		xBox.Set(xVertex1);
		xBox.Set(xVertex2);
		xBox.Set(xVertex3);
		xBox.Set(xVertex4);
		xBox.Set(xVertex5);
		xBox.Set(xVertex6);
		xBox.Set(xVertex7);
		xBox.Set(xVertex8);

		xBox.Displace(xPosition);
	}


	const Asura_Vector_3 xUOffset(-1.0f, 0.0f, 0.0f);
	const Asura_Vector_3 xVOffset( 0.0f, 0.0f, 1.0f);
	const Asura_Vector_3 xNOffset( 0.0f,-1.0f, 0.0f);

	// make an environment cache of the polys
	// if we can see it, add a decal to it
	m_xEnvCache.SetFlagsOfPolysToIgnore( AsuraPolygonFlag_DontRender );
	m_xEnvCache.SetFacesOnly( true );

	m_xEnvCache.Update( xBox, fENV_CACHE_BOUNDARY );

	const Asura_Physics_EnvironmentCache_Face* pxFace = m_xEnvCache.GetFaceList();

	for (; pxFace!=NULL; pxFace = pxFace->m_pxNext )
	{
		const Asura_AVector3& xV0 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 0 );
		const Asura_AVector3& xV1 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 1 );
		const Asura_AVector3& xV2 = Asura_Physics_Environment::GetFaceVertex( pxFace->m_uFaceID, 2 );

		Asura_Vector_3 xNormal;
		
		xNormal = ((xV1 - xV0)^(xV2 - xV0));
	
		xNormal.Normalise();


		const float fSurfaceAngle = -xNormal.y;

		if (fSurfaceAngle<=0.02f) continue;



		Asura_ShadowDecal_Vertex xVertexBufferA[9];
		Asura_ShadowDecal_Vertex xVertexBufferB[9];

		u_int uNumberOfVertices = 3;

		xVertexBufferA[0].m_xPosition = xV0;
		xVertexBufferA[1].m_xPosition = xV1;
		xVertexBufferA[2].m_xPosition = xV2;

		
		//	Platform_AddStrip(xVertexBufferA,uNumberOfVertices);

		/// Kevin Lea - 13/11/2002 17:47:42 - left plane
		float fD = xPosition*xUOffset;

		uNumberOfVertices = ClipPolyToPlane( xVertexBufferA, uNumberOfVertices, xVertexBufferB, xUOffset, fWidth - fD);
		if (uNumberOfVertices==0) continue;

		/// Kevin Lea - 13/11/2002 17:50:20 - right plane
		Asura_Vector_3 xPlaneNormal = -xUOffset;

		uNumberOfVertices = ClipPolyToPlane( xVertexBufferB, uNumberOfVertices, xVertexBufferA, xPlaneNormal, fWidth + fD);
		if (uNumberOfVertices==0) continue;

		/// Kevin Lea - 13/11/2002 17:51:47 - top plane
		fD = xPosition*xVOffset;

		uNumberOfVertices = ClipPolyToPlane( xVertexBufferA, uNumberOfVertices, xVertexBufferB, xVOffset, fHeight - fD);
		if (uNumberOfVertices==0) continue;

		/// Kevin Lea - 13/11/2002 17:53:12 - bottom plane
		xPlaneNormal = -xVOffset;

		uNumberOfVertices = ClipPolyToPlane( xVertexBufferB, uNumberOfVertices, xVertexBufferA, xPlaneNormal, fHeight + fD);
		if (uNumberOfVertices==0) continue;

		/// Kevin Lea - 13/11/2002 17:54:55 - front plane
		fD = xPosition*xNOffset;

		uNumberOfVertices = ClipPolyToPlane( xVertexBufferA, uNumberOfVertices, xVertexBufferB, xNOffset, fDepth - fD);
		if (uNumberOfVertices==0) continue;

		/// Kevin Lea - 13/11/2002 17:56:09 - back plane
		xPlaneNormal = -xNOffset;

		uNumberOfVertices = ClipPolyToPlane( xVertexBufferB, uNumberOfVertices, xVertexBufferA, xPlaneNormal, fDepth + fD);
		if (uNumberOfVertices==0) continue;

        // Set the colour (lighting if required)
        //Asura_RGBA xDiffuseRGBA;
        //{
        //    xDiffuseRGBA.m_fR = 1.0f;
        //    xDiffuseRGBA.m_fG = 1.0f;
        //    xDiffuseRGBA.m_fB = 1.0f;
        //}
        //xDiffuseRGBA.m_fA = 1.0f;

		for (u_int v=0; v<uNumberOfVertices; v++)
		{
			//xVertexBufferA[v].m_fAlpha = 1.0f - Asura_Maths::Abs((xSurfaceNormal * (xPosition-xVertexBufferA[v].m_xPosition))/fDepth);

			Asura_Vector_3 xDisp = xVertexBufferA[v].m_xPosition - xPosition;

			xVertexBufferA[v].m_fU = (0.5f * (xDisp*xUOffset) / fWidth  + 0.5f);
			xVertexBufferA[v].m_fV = (0.5f * (xDisp*xVOffset) / fHeight + 0.5f);
		}
		/// Kevin Lea - 13/11/2002 17:57:41 - okay, now output stip to platform dependent layer

		// reorder fan into strip
		{
			u_int u = 0;
			u_int v = 0;
			u_int w = uNumberOfVertices-1;

			xVertexBufferB[u++] = xVertexBufferA[v++];

			while(true)
			{
				xVertexBufferB[u++] = xVertexBufferA[v++];
				if (u==uNumberOfVertices) break;

				xVertexBufferB[u++] = xVertexBufferA[w--];
				if (u==uNumberOfVertices) break;
			}
		}

		Asura_ShadowDecal_System::Platform_AddStrip( xVertexBufferB, uNumberOfVertices );
	}

}

u_int Asura_ShadowDecal_Cache::ClipPolyToPlane(Asura_ShadowDecal_Vertex* pxInVertices, u_int uNumberOfVertices, Asura_ShadowDecal_Vertex* pxOutVertices, const Asura_Vector_3& xPlaneNormal, const float fPlaneConstant)
{
	u_int uNewVertexCount = 0;

	float fDot[9];

	{
		bool bAllNegative = true;
		for (u_int u=0; u<uNumberOfVertices; u++)
		{
			fDot[u] = pxInVertices[u].m_xPosition*xPlaneNormal + fPlaneConstant;
			if (fDot[u]>=0.0f)
			{
				bAllNegative = false;
			}
		}

		/// Kevin Lea - 14/11/2002 10:49:43 - early exit if all vertices would be clipped
		if (bAllNegative) return 0;
	}

	for (u_int u=0; u<uNumberOfVertices; u++)
	{
		u_int v = (u!=0) ? u-1 : uNumberOfVertices-1; // previous vertex
		
		if (fDot[u]<0.0f)
		{
			if (fDot[v]>=0.0f)
			{
				// Current vertex is on negative side of plane,
				// but previous vertex is on positive side.
				Asura_Vector_3 xEdge = pxInVertices[v].m_xPosition - pxInVertices[u].m_xPosition;
				float f = xEdge * xPlaneNormal;
				
				Asura_Assert( f!=0.0f, "edge parallel to plane");

				float fLambda = fDot[v] / f;

                Asura_Assertf( fLambda >= -0.0001f && fLambda <= 1.0001f, ("Out of bounds: %f",Asura_Maths::FtoD(fLambda)));
				if (fLambda < 0.0f) fLambda = 0.0f;
				if (fLambda > 1.0f) fLambda = 1.0f;

				pxOutVertices[uNewVertexCount].m_xPosition = pxInVertices[v].m_xPosition - xEdge * fLambda;
				uNewVertexCount++;	
			}
		}
		else
		{
			if (fDot[v]<0.0f)
			{
				// Current vertex is on positive side of plane,
				// but previous vertex is on negative side.
				const Asura_Vector_3 xEdge = pxInVertices[u].m_xPosition - pxInVertices[v].m_xPosition;
				float f = xEdge * xPlaneNormal;
				
				Asura_Assert( f!=0.0f, "edge parallel to plane");

				float fLambda = fDot[u] / f;

                Asura_Assertf( fLambda >= -0.0001f && fLambda <= 1.0001f, ("Out of bounds: %f",Asura_Maths::FtoD(fLambda)));
				if (fLambda < 0.0f) fLambda = 0.0f;
				if (fLambda > 1.0f) fLambda = 1.0f;

				pxOutVertices[uNewVertexCount].m_xPosition = pxInVertices[u].m_xPosition - xEdge * fLambda;
				uNewVertexCount++;	
			}

			/// Kevin Lea - 14/11/2002 11:22:04 - include current vertex
			pxOutVertices[uNewVertexCount].m_xPosition = pxInVertices[u].m_xPosition;
			uNewVertexCount++;
		}
	}


	return uNewVertexCount;
}

void Asura_ShadowDecal_System::Render(void)
{
	Platform_Render();
}

