//
// Asura_PFX_Source.cpp
//
// Created by KOR - 15/07/2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"


// This file's header.
#include "SpecialFX/PFX/Asura_PFX_Source.h"

// Asura headers
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#ifdef ASURA_PFX_DONT_REORDER_PARTICLES
#include "Core/Memory/Asura_Mem.h"
#endif
#include "Physics/Asura_Physics.h"
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
#include "Jobs/Asura_Job_PFX.h"
#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
#include "Light/Asura_GI_Solution.h"
#include "Light/Asura_Lighting_Solution.h"
#include "Sound/Asura_Sound_Event_System.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "SpecialFX/PFX/Asura_PFX_Classes.h"
#include "SpecialFX/PFX/Asura_PFX_Effect.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"
#include "Debugging/Asura_Debug_Timers.h"

#if _DEBUG
u_int g_uParticlesAllocatedInSources = 0;
u_int g_uParticlesUsedInSources      = 0;
#endif

static const float g_fGravity = 10.0f;

/////////////////////////////////////////////////////////////////////////////////////////////////


class Asura_PFX_Lighting
{
public :
	Asura_PFX_Lighting()
	: m_fCachedScale(-1.0f)
	, m_xLightingValues(1.0f, 1.0f, 1.0f, 1.0f)
	, m_xLightingSolution(true,false)
	{
	}

	void UpdateLighting(const Asura_PFX_Particle_Template* pxTemplate,const Asura_Vector_3& xPos);

	const Asura_RGBA GetLighting() const {return m_xLightingValues;}

private :
	float						m_fCachedScale;
	Asura_RGBA					m_xLightingValues;
	Asura_Lighting_Solution		m_xLightingSolution;
};

void Asura_PFX_Lighting::UpdateLighting(const Asura_PFX_Particle_Template* pxParticleTemplate,const Asura_Vector_3& xPos)
{
	const float	fScale	= pxParticleTemplate->m_fLightingScale;
	bool		bLightingChanged;

	//Update the lighting position , unless the template has the prelight option and we have already 
	//done the calculations once
	if(!(pxParticleTemplate->m_bPreLighting && m_xLightingSolution.IsStaticValid()))
	{
		m_xLightingSolution.SetPosition(xPos);
	} 

	if(m_xLightingSolution.IsUpToDate(true, false))
	{
		bLightingChanged	= (m_fCachedScale != fScale);
	}
	else
	{
		bLightingChanged = true;
	}

	if(bLightingChanged)
	{
//		ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_PFX_Lighting::UpdateLighting - lighting changed" );
		const Asura_SphericalHarmonics_Order2_RGB& xSH = Asura_GI_Solution::s_xSHForCurrentRender;

		m_xLightingSolution.RecalculateLighting(true, false);

		m_xLightingValues.m_fR	= Asura_Maths::Abs( xSH.m_afRedCoefficients[0]   )
								+ Asura_Maths::Abs( xSH.m_afRedCoefficients[1]   )
								+ Asura_Maths::Abs( xSH.m_afRedCoefficients[2]   )
								+ Asura_Maths::Abs( xSH.m_afRedCoefficients[3]   );
		m_xLightingValues.m_fG	= Asura_Maths::Abs( xSH.m_afGreenCoefficients[0] )
								+ Asura_Maths::Abs( xSH.m_afGreenCoefficients[1] )
								+ Asura_Maths::Abs( xSH.m_afGreenCoefficients[2] )
								+ Asura_Maths::Abs( xSH.m_afGreenCoefficients[3] );
		m_xLightingValues.m_fB	= Asura_Maths::Abs( xSH.m_afBlueCoefficients[0]  )
								+ Asura_Maths::Abs( xSH.m_afBlueCoefficients[1]  )
								+ Asura_Maths::Abs( xSH.m_afBlueCoefficients[2]  )
								+ Asura_Maths::Abs( xSH.m_afBlueCoefficients[3]  );
		m_xLightingValues.m_fA = 1.0f;

		if(fScale < 1.f)
		{
			m_xLightingValues.m_fR	= Asura_Maths::Lerp(1.0f, m_xLightingValues.m_fR, fScale);
			m_xLightingValues.m_fG	= Asura_Maths::Lerp(1.0f, m_xLightingValues.m_fG, fScale);
			m_xLightingValues.m_fB	= Asura_Maths::Lerp(1.0f, m_xLightingValues.m_fB, fScale);
		}

		m_fCachedScale	= fScale;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////

Asura_PFX_Source::Asura_PFX_Source(Asura_PFX_Effect* pxEffect, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xDirection, Asura_Hash_ID uTemplateNameHash, const Asura_PFX_Volume& xVolume)
: m_pxEffect(pxEffect)
, m_xPosition(xPosition)
, m_xDirection(xDirection)
, m_xVolume(xVolume)
, m_pxParticles(NULL)
, m_uMaxParticles(0)
, m_uNumParticles(0)
, m_uTemplateNameHash(uTemplateNameHash)
, m_fParticleReleaseCounter(0.f)
, m_fAge(0.f)
, m_fRangeTimer(0.f)
, m_pxLighting(NULL)
, m_uFlags(0)
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
, m_pPlatformRenderData(NULL)
#endif
#ifdef ASURA_TOOLS
, m_pxRNG(NULL)
#endif
{
	Asura_Assert(pxEffect, "No parent effect given for source");
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(uTemplateNameHash);
	Asura_Assert(pxTemplate, "No template for source");
	if(!pxTemplate)
	{
		//pxEffect->SetActive(false);
		return;
	}
	m_fRangeTimer = pxTemplate->m_fRangeDelayTime;
#if _DEBUG
    g_uParticlesAllocatedInSources += pxTemplate->m_uMaxParticles;
#endif
	Asura_PFX_Particle_Template *pxParticleTemplate = Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash);

#if defined ASURA_TOOLS || defined _DEBUG
	Asura_Assertf(pxParticleTemplate, ("No particle template for source %s", pxTemplate->m_xName.GetString()));
#else
	Asura_Assert(pxParticleTemplate, "No particle template for source");
#endif

	if(pxParticleTemplate)
	{
		// Lighting may get turned on in tools
#ifndef ASURA_TOOLS
		if(pxParticleTemplate->m_bUseLighting)
#endif
		{
			m_pxLighting = new Asura_PFX_Lighting;
		}
	}
	m_bInRange = true;
	m_bHasDoneBurst = false;
	m_fReleaseRate = 1.0f;
}

Asura_PFX_Source::~Asura_PFX_Source()
{
	if(m_pxLighting) delete m_pxLighting;
	RemoveAllParticles();
#if _DEBUG
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
    if(pxTemplate)
	{
		g_uParticlesAllocatedInSources -= pxTemplate->m_uMaxParticles;
	}
#endif
	delete [] m_pxParticles;
	m_uMaxParticles = 0;
}

void Asura_PFX_Source::Update(bool bLinkedIn, bool bInFrustum, const Asura_Spline<Asura_Vector_3>* pxSpline /* = NULL */)
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
	Asura_PFX_Particle_Template *pxParticleTemplate = pxTemplate ? Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash) : NULL;

	if(!pxTemplate || !pxParticleTemplate ||
		pxTemplate->m_fDuration == 0.0f) return;

	// Age source
	float fTimeSlice = Asura_Timers::GetGameTimeSlice();
	if(pxTemplate->m_bUseTimeScale)
	{
		fTimeSlice *= pxTemplate->m_fTimeScale;
	}
	m_fAge += fTimeSlice / pxTemplate->m_fDuration;
	if(m_fAge > 1.f)
	{
		if(/* Loop */ 0)
		{
			m_fAge = 0.f;
			m_bHasDoneBurst = false; // reset so we can burst again as soon as we're visible
		}
		else
		{
			//return;
		}
	}

	// Move source
	if(pxTemplate->m_bUseSourceDynamics)
	{
		// Move source
		Asura_Physics_System::MoveParticle(m_xPosition, m_xVelocity, pxTemplate->m_fSourceBounce);
		//m_xPosition += m_xVelocity * fTimeSlice;
		m_xVelocity.y += g_fGravity * fTimeSlice;
	}

	u_int uInitialNumParticles = m_uNumParticles;

	// Update in range - used to be done after move particles, since I moved the lighting up,
	// I also moved this up (MH - 08.09.2005) .
	if(pxTemplate->m_bUseRange)
	{
		bool bInRange = false;
		for(int iCamera = 0; iCamera < Asura_Camera::GetNumberOfCameras(); ++iCamera)
		{
			Asura_Camera::SetCameraIndex(iCamera);
			if(Asura_Camera::IsCameraActive())
			{
				const Asura_Vector_3* pxCameraPos = Asura_Camera::GetPosition();
				const float fDistance = (*pxCameraPos - m_xPosition).Magnitude();
				if(fDistance > pxTemplate->m_fRangeMin && fDistance < pxTemplate->m_fRangeMax)
				{
					bInRange = true;
					break;
				}
			}
		}
		if(bInRange)
		{
			m_fRangeTimer += Asura_Timers::GetGameTimeSlice();
		}
		else
		{
			m_fRangeTimer -= Asura_Timers::GetGameTimeSlice();
		}
		if(m_bInRange)
		{
			if(m_fRangeTimer <= 0.f)
			{
				m_bInRange = false;
			}
		}
		else
		{
			if(m_fRangeTimer > pxTemplate->m_fRangeDelayTime)
			{
				m_bInRange = true;
			}
		}
		m_fRangeTimer = Asura_Maths::ClampToRange(m_fRangeTimer, 0.f, pxTemplate->m_fRangeDelayTime);
	}
#ifdef ASURA_TOOLS
	// Could have changed template using range - make sure we're generating if toggled off.
	else
	{
		m_bInRange = true;
	}
#endif

	// Don't remove particles if we're just out of view.
	const bool bRemoveParticles = !(bLinkedIn && !bInFrustum);
	// remove any particles which have expired
	if(m_uNumParticles && bRemoveParticles)
	{
		if(pxParticleTemplate->m_bRandomLifetime)
		{
			for(u_int uParticle = 0; uParticle < m_uNumParticles; )
			{
				// Age particle
				float fLifeTime = pxParticleTemplate->GetLifetimeFromSeed(m_pxParticles[uParticle].m_fSeed);
				m_pxParticles[uParticle].m_fAge += fLifeTime > 0.0f ? (fTimeSlice / fLifeTime) : 0.0f;
				// Delete old particles
				if(m_pxParticles[uParticle].m_fAge > 1.f)
				{
					if(pxParticleTemplate->m_bLooped)
					{
						m_pxParticles[uParticle].m_fAge = 0.f;
						++uParticle;
					}
					else
					{
						RemoveParticle(uParticle);
					}
				}
				else
				{
					++uParticle;
				}
			}
		}
		else
		{
			float fLifeTime = pxParticleTemplate->m_fMinLifetime;
			const float fAgeIncrement = fLifeTime > 0.0f ? (fTimeSlice / fLifeTime) : 0.0f;
			for(u_int uParticle = 0; uParticle < m_uNumParticles; )
			{
				// Age particle
				m_pxParticles[uParticle].m_fAge += fAgeIncrement;
				// Delete old particles
				if(m_pxParticles[uParticle].m_fAge > 1.f)
				{
					if(pxParticleTemplate->m_bLooped && (m_pxEffect->GetState() != ASURA_PFX_EFFECT_STOPPING))
					{
						m_pxParticles[uParticle].m_fAge = 0.f;
						++uParticle;
					}
					else
					{
						RemoveParticle(uParticle);
					}
				}
				else
				{
					++uParticle;
				}
			}
		}
	}

	// Create particles (used to be done after UpdateParticles). To maintain old behaviour,
	// remember the number of particles before creating and only update those. This assumes
	// that new particles are created at the end of the list.
	u_int uNumParticlesToUpdate = m_uNumParticles;
	//Only bother creating new particles if the system is visible
	//RAR 16/09/04
	if(m_fAge < 1.f && bInFrustum  && pxParticleTemplate && m_bInRange)
	{
		if(pxTemplate->m_bBurst)
		{
			// Create all particles once
			if(!m_bHasDoneBurst)
			{
				m_bHasDoneBurst = true;
				for(u_int uParticle = m_uNumParticles; uParticle < pxTemplate->m_uMaxParticles; ++uParticle)
				{
					CreateParticle(fTimeSlice);
				}
			}
		}
		else
		{
			const float fReleaseRateScale = pxTemplate->m_bScaleReleaseRateWithVolume ? m_xVolume.GetVolumeSize() : 1.f;
			const float fReleaseRate = fReleaseRateScale * pxTemplate->m_xReleaseRate.GetValueAtTime_Discrete(m_fAge) * m_pxEffect->GetReleaseRateScale();
			if(fReleaseRate > 0.f)
			{
				m_fParticleReleaseCounter += fReleaseRate*fTimeSlice*m_fReleaseRate;
				// Create new particle if necessary
				while(m_fParticleReleaseCounter >= 1.0f)
				{
					m_fParticleReleaseCounter -= 1.0f;
					const float fTimeBetweenParticles = 1.f/fReleaseRate;
					//get frame time for this particle , based on when it was created in frame
					const float fTimer = m_fParticleReleaseCounter*fTimeBetweenParticles;
					CreateParticle(fTimer);
				}
			}
		}
	}

	// update lighting (used to be done after particles had moved. I moved it here so that the 
	// UpdateParticles thread can prepare the render buffers if necessary. (MH - 8.9.05)
	if(pxParticleTemplate->m_bUseLighting && (m_uNumParticles > 0) && bInFrustum)
	{
		Asura_Assert(m_pxLighting, "No lighting solution");
		// Update lighting info from the GI solution
		if(m_pxLighting)
		{
			m_pxLighting->UpdateLighting(pxParticleTemplate,m_xPosition);
		}
	}

	//Only bother updating particle positions if the system is visible
	//RAR 16/09/04
	if(bInFrustum && m_uNumParticles)
	{
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
		// reset this
		m_pPlatformRenderData = NULL;
  		// don't do skybox effects as job queue as they need to be ready earlier.
  		bool bIsSkyboxEffect = (ASURA_PFX_RENDER_SKYBOX == m_pxEffect->GetRenderType());
  		// create job and add to job queue
		if (Asura_PFX_System::GetEnableMultithreading() && !bIsSkyboxEffect)
  		{
  			Asura_Vector_3 xCollisionPosition, xCollisionDirection, xGravityPosition;
  			bool bUseCollisionPlane = false;
  			bool bUsePointGravity = false;
  			if(pxTemplate->m_bUseParticleDynamics)
  			{
  				bUseCollisionPlane = m_pxEffect->GetCollisionPlane(&xCollisionPosition, &xCollisionDirection);
  			}
  			if(pxTemplate->m_bPointGravity)
  			{
  				bUsePointGravity = m_pxEffect->GetGravityPosition(&xGravityPosition);
  			}

			UpdateRenderData();

			if( NULL != m_pPlatformRenderData)
			{
				Asura_Job_PFX_Update xJob;

	//			bool bPaused = Asura_Timers::GetGameTimeScale() == 0.0f;

				xJob.InitialiseData(	m_pxParticles, m_uNumParticles, uNumParticlesToUpdate, pxTemplate,
  										bUseCollisionPlane, xCollisionPosition, xCollisionDirection, 
  										bUsePointGravity, xGravityPosition, 
  										m_pxEffect->GetSizeScale(), m_pPlatformRenderData,
										m_pxLighting ? m_pxLighting->GetLighting() : Asura_RGBA(1.f,1.f,1.f,1.f), *m_pxEffect->GetOrientation(), 
										m_pxEffect->GetCameraFlags(true), pxSpline );  			

				xJob.AddToQueue();
			}
			
  		}
  		else
#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
		{
			// do it the old single-threaded way.
			m_pxEffect->UpdateParticles( m_pxParticles, uNumParticlesToUpdate, pxTemplate, pxParticleTemplate, pxSpline );
		}
	}
	
	// Create particles was here, moved it to before update particles to test.

	if(m_pxParticles && !m_uNumParticles && !uInitialNumParticles)
	{
		//Particle system has had zero particles for an entire frame , so
		//delete the particle array to save space.
		delete [] m_pxParticles;
		m_pxParticles = 0;
	}
}

void Asura_PFX_Source::Render(bool bDontQueue /* = false */)
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
	Asura_PFX_Particle_Template *pxParticleTemplate = pxTemplate ? Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash) : NULL;

	if(!(pxTemplate && pxParticleTemplate)) return;
	if(!m_uNumParticles) return;
	if (IsHeatHazeSource() &&
		Asura_PFX_System::AddHeatHazeSource(this))
	{
		// (AW - 26.10.2004) - Successfully added to the heat haze list. Will render later.
		return;
	}

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
		void* pPlatformRenderData = m_pPlatformRenderData;
#else
		void* pPlatformRenderData = NULL;
#endif

	const Asura_RGBA	xColour	= (m_pxLighting == NULL)
								? Asura_RGBA(1.0f, 1.0f, 1.0f, 1.0f)
								: m_pxLighting->GetLighting();

	// RJF - 19 October 2004 - Queued render commands
	if(bDontQueue || pxParticleTemplate->m_eTranslucency == ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE)
	{
		// Don't bother queuing particles if it's an ordered render or they're heat haze particles
		m_pxEffect->RenderParticles(m_pxParticles, m_uNumParticles, pxParticleTemplate, xColour, pPlatformRenderData);
	}
	else
	{
		Asura_PFX_RenderQueue&	xRenderQueue	= Asura_PFX_System::GetRenderQueue();

		xRenderQueue.AddParticles(*pxParticleTemplate, m_pxParticles, m_uNumParticles, xColour, *m_pxEffect, pPlatformRenderData);
	}

	
}

void Asura_PFX_Source::Stop(bool bImmediate)
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);

	//Check for sources that always stop immediately. RAR 20/02/2006
	if(pxTemplate && pxTemplate->m_bStopInstantly)
	{
		bImmediate = true;
	}
	
	if(bImmediate)
	{
		RemoveAllParticles();
	}
}

void Asura_PFX_Source::CreateParticle(float fReleaseTimer)
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
	Asura_PFX_Particle_Template *pxParticleTemplate = pxTemplate ? Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash) : NULL;

	// Don't create particles in an inactive effect
	if(m_pxEffect->GetState() != ASURA_PFX_EFFECT_RUNNING) return;

	const float fProbability = pxTemplate->m_xReleaseRand.GetValueAtTime_Discrete(m_fAge);
	if(GetRNG()->GetFloat() > fProbability) return;

	//Not entirely sure its helpful to assert on creating too many particles here , since random release
	//rate will cause this too happen unless MaxParticles is set much higher than really needed.
	//RAR 26/11/04
	//Asura_Assert(m_uNumParticles < m_pxTemplate->m_uMaxParticles, "Attempted to create too many particles");
	const float fReleaseRateScale = pxTemplate->m_bScaleReleaseRateWithVolume ? m_xVolume.GetVolumeSize() : 1.f;
	const u_int uMaxParticles = Asura_Maths::FloatToInt_Near(pxTemplate->m_uMaxParticles * fReleaseRateScale);
	if(m_uNumParticles >= uMaxParticles) return;


	float fTimeSlice = Asura_Timers::GetGameTimeSlice();
	if(pxTemplate->m_bUseTimeScale)
	{
		fTimeSlice *= pxTemplate->m_fTimeScale;
	}
	const Asura_Vector_3 xEffectVelocity = (fTimeSlice <= 0.f) ? Asura_Vector_3(0.f, 0.f, 0.f) : m_pxEffect->GetDelta() / fTimeSlice;

	// KOR - 08/06/2005 - spawn particles from particles of another source.
	const Asura_PFX_Particle* pxPrevParticle = NULL;
	if(pxTemplate->m_iSpawnSourceIndex != -1)
	{
		const Asura_PFX_Source* pxSource = m_pxEffect->GetSource(pxTemplate->m_iSpawnSourceIndex); // should get source number from source template if spawn from particles turned on
		if(pxSource)
		{
			pxPrevParticle = pxSource->GetRandomParticle();
			if(!pxPrevParticle) return;
		}
	}
	const float fSourceAge = Asura_Maths::ClampToRange(m_fAge - fReleaseTimer/pxTemplate->m_fDuration, 0.f, 1.f);
//	if(Asura_PFX_System::CanCreateParticle())
	{
		// Allocate particle
		if(!m_pxParticles)
		{
			//allocate particle array
			m_pxParticles = new Asura_PFX_Particle[uMaxParticles];
			m_uMaxParticles = uMaxParticles;
		}
		if(m_uNumParticles >= m_uMaxParticles) return;

		// don't change the way new particles are added at the end. The way particles are
		// updated assumes that any particles added to the list in a frame are at the end.
		Asura_PFX_Particle* pxParticle = &m_pxParticles[m_uNumParticles];
		++m_uNumParticles;
#if _DEBUG
        ++g_uParticlesUsedInSources;
#endif
		// KOR - 08/06/2005 - spawn particles from particles of another source.
		if(pxPrevParticle)
		{
			pxParticle->m_xPosition = pxPrevParticle->m_xPosition - pxPrevParticle->m_xVelocity * fReleaseTimer;
		}
		else
		{
			pxParticle->m_xPosition = m_xPosition;
		}
		// Compensate for source movement
		if(pxTemplate->m_bUseSourceDynamics)
		{
			pxParticle->m_xPosition -= m_xVelocity * fReleaseTimer;
		}
		// Compensate for entity effect movement
		if(!pxTemplate->m_bMovePclesWithSource)
		{
			pxParticle->m_xPosition -= xEffectVelocity * fReleaseTimer;
		}
		
		//Offset position by a point chosen from the volume
		//RAR 13/09/04
		if(m_xVolume.m_bUseOrientation)
		{
			pxParticle->m_xPosition += m_xVolume.GeneratePointInVolume(GetRNG())*m_xDirection;
		}
		else
		{
			pxParticle->m_xPosition += m_xVolume.GeneratePointInVolume(GetRNG());
		}
		if(pxParticleTemplate->m_uSoundEvent)
		{
			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::FindEvent(pxParticleTemplate->m_uSoundEvent);
			if(pxEvent)
			{
				Asura_Assert(!pxEvent->IsEndlessPlayback(), "Particle template is using a looping sound event.");
				Asura_Sound_Event_System::PlayEvent(pxEvent, NULL, &pxParticle->m_xPosition);
			}
		}
		const float fSpeed = pxTemplate->m_xSpeed.GetValueAtTime_Discrete(fSourceAge);
		const Asura_Vector_3 xTemplateAngle = pxTemplate->m_xAngle.GetValueAtTime_Discrete(fSourceAge);
		const float fSpeedSpread = pxTemplate->m_xSpeedSpread.GetValueAtTime_Discrete(fSourceAge);
		const Asura_Vector_3 xAngleSpread = pxTemplate->m_xAngleSpread.GetValueAtTime_Discrete(fSourceAge);
		// TODO - optimise: this is a rather complicated way of going about it at the moment - there must be a simpler version.
		Asura_Vector_3 xVelocity( 0.f, 0.f,	fSpeed * (1.0f - fSpeedSpread * GetRNG()->GetFloat()));

		// (AAM - 22.02.2005) Use code speed scaling factor
		if( pxTemplate->m_bUseSpeedScale )
		{
			xVelocity *= m_pxEffect->GetSpeedScale();
		}

		const Asura_Euler_3 xEuler( (xTemplateAngle.x * Asura_Maths::Pi) + xAngleSpread.x * GetRNG()->GetFloat(-1.f * Asura_Maths::Pi, Asura_Maths::Pi),
									(xTemplateAngle.y * Asura_Maths::Pi) + xAngleSpread.y * GetRNG()->GetFloat(-1.f * Asura_Maths::Pi, Asura_Maths::Pi), 
									0.f);
		const Asura_Quat xPitchYawQuat(xEuler);
		// Roll must be applied after pitch and yaw, otherwise it does nothing to the z-axis
		const Asura_Euler_3 xRollEuler( 0.f, 0.f, (xTemplateAngle.z * Asura_Maths::Pi) + xAngleSpread.z * GetRNG()->GetFloat(-1.f * Asura_Maths::Pi, Asura_Maths::Pi));
		const Asura_Quat xRollQuat(xRollEuler);
		const Asura_Quat xQuat = xRollQuat * xPitchYawQuat;
		xVelocity = xQuat.RotateVector(xVelocity);
		pxParticle->m_xVelocity = xVelocity * m_xDirection;
		if(pxTemplate->m_bInheritVelocity)
		{
			pxParticle->m_xVelocity += xEffectVelocity * pxTemplate->m_fVelocityInheritance;
		}

		if(!pxParticleTemplate) return;

		// Random seed
		pxParticle->m_fSeed = GetRNG()->GetFloat();

		// Age it within the frame
		const float fLifetime = pxParticleTemplate->GetLifetimeFromSeed(pxParticle->m_fSeed);
		pxParticle->m_fAge = fLifetime > 0.0f
						   ? Asura_Maths::Min(1.f, fReleaseTimer / fLifetime)
						   : 0.0f;

		// Move particle along path according to time within frame it should have been alive
		// (unless looped, in which case it should never move)
		if(!pxParticleTemplate->m_bLooped)
		{
			pxParticle->m_xPosition += pxParticle->m_xVelocity * fReleaseTimer;
		}
		if(pxTemplate->m_bDirectionalGravity)
		{
			pxParticle->m_xPosition += pxTemplate->m_xGravityVector * (0.5f * fReleaseTimer * fReleaseTimer);
		}
	}
}

void Asura_PFX_Source::RemoveParticle(u_int uParticle)
{
	--m_uNumParticles;
#if _DEBUG
    --g_uParticlesUsedInSources;
#endif
#ifdef ASURA_PFX_DONT_REORDER_PARTICLES
	// This will move the whole remaining section of the array down.
	Asura_Mem::Move(&m_pxParticles[uParticle], &m_pxParticles[uParticle + 1], sizeof(Asura_PFX_Particle) * (m_uNumParticles - uParticle));
#else
	// This simply drops the last particle into the slot emptied by the current one.
	m_pxParticles[uParticle] = m_pxParticles[m_uNumParticles];
#endif
}

void Asura_PFX_Source::RemoveAllParticles()
{
	while(m_uNumParticles)
	{
		RemoveParticle(0);
	}
}

bool Asura_PFX_Source::IsHeatHazeSource() const
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
	Asura_PFX_Particle_Template *pxParticleTemplate = pxTemplate ? Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash) : NULL;
	return( pxParticleTemplate && pxParticleTemplate->m_eTranslucency == ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE );
}

void Asura_PFX_Source::ExpandBoundingBox(Asura_Bounding_Box& xOutputBB) const
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
	Asura_PFX_Particle_Template *pxParticleTemplate = pxTemplate ? Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash) : NULL;
    if( !pxTemplate || !pxParticleTemplate ) { return; }
	
	//Get the bounding box of the spawn volume
	Asura_Bounding_Box xSourceBB;

	switch(m_xVolume.m_eType)
	{
		case ASURA_PFX_VOLUME_POINT :
			xSourceBB.Zero();
			break;
		case ASURA_PFX_VOLUME_CUBOID :
			xSourceBB = m_xVolume.m_xCuboidDesc.m_xBB;
			break;
		case ASURA_PFX_VOLUME_SPHERE :
			xSourceBB.MinX = -m_xVolume.m_xSphereDesc.m_fRadius;
			xSourceBB.MinY = -m_xVolume.m_xSphereDesc.m_fRadius;
			xSourceBB.MinZ = -m_xVolume.m_xSphereDesc.m_fRadius;
			xSourceBB.MaxX = m_xVolume.m_xSphereDesc.m_fRadius;
			xSourceBB.MaxY = m_xVolume.m_xSphereDesc.m_fRadius;
			xSourceBB.MaxZ = m_xVolume.m_xSphereDesc.m_fRadius;
			break;

		case ASURA_PFX_VOLUME_SPHEROID :
			xSourceBB.MinX = -m_xVolume.m_xSpheroidDesc.m_fExtentX;
			xSourceBB.MinY = -m_xVolume.m_xSpheroidDesc.m_fExtentY;
			xSourceBB.MinZ = -m_xVolume.m_xSpheroidDesc.m_fExtentZ;
			xSourceBB.MaxX = m_xVolume.m_xSpheroidDesc.m_fExtentX;
			xSourceBB.MaxY = m_xVolume.m_xSpheroidDesc.m_fExtentY;
			xSourceBB.MaxZ = m_xVolume.m_xSpheroidDesc.m_fExtentZ;
			break;

		default :
			Asura_Assert(0,"Unknown volume type");
			xSourceBB.Zero();
			break;

	}
	//apply the source's position and orientation(if required)
	if(m_xVolume.m_bUseOrientation)
	{
		if(m_xVolume.m_eType == ASURA_PFX_VOLUME_SPHEROID || m_xVolume.m_eType == ASURA_PFX_VOLUME_CUBOID)
		xSourceBB.FromOBB(xSourceBB,m_xDirection);
	}
	xSourceBB.Displace(m_xPosition);


	//Expand the source BB to include the maximum distance it could travel in its lifetime
	Asura_Bounding_Box xTravelBB;
	//get the highest particle speed
	float fMaxSpeed = 0.0f;
	GetGraphMax(pxTemplate->m_xSpeed, fMaxSpeed);

	float fMaxSpread = 0.0f;
	GetGraphMax(pxTemplate->m_xAngleSpread, fMaxSpread, fMaxSpread);
	float fMaxAngle = 0.0f;
	GetGraphMax(pxTemplate->m_xAngle, fMaxAngle, fMaxAngle);

	fMaxSpread = Asura_Maths::Min(1.f, fMaxAngle + fMaxSpread);

	const float fMaxPcleLifetime = pxParticleTemplate->GetMaxLifetime();
	float fDistance = fMaxSpeed * fMaxPcleLifetime;
	// Forwards
	xTravelBB.MaxZ = fDistance;
	// Backwards, if spread is large enough
	if(fMaxSpread > 0.5f)
	{
		xTravelBB.MinZ = Asura_Maths::Cosine(Asura_Maths::Pi * fMaxSpread) * fDistance;
		// Now clamp at 0.5, so that we get maximum spread at the sides
		fMaxSpread = 0.5f;
	}
	else
	{
		xTravelBB.MinZ = 0.f;
	}
	const float fSideDistance = fDistance * Asura_Maths::Sine(Asura_Maths::Pi * fMaxSpread);
	xTravelBB.MinX = -fSideDistance;
	xTravelBB.MinY = -fSideDistance;
	xTravelBB.MaxX = fSideDistance;
	xTravelBB.MaxY = fSideDistance;

	// Now rotate travel bb into worldspace
	xTravelBB.FromOBB(xTravelBB,m_xDirection);

	if(!pxTemplate->m_bUseSourceBBForVisibility && !pxParticleTemplate->m_bLooped)
	{
		// Expand source volume by travel BB
		xSourceBB.MaxX += xTravelBB.MaxX;
		xSourceBB.MinX += xTravelBB.MinX;
		xSourceBB.MaxY += xTravelBB.MaxY;
		xSourceBB.MinY += xTravelBB.MinY;
		xSourceBB.MaxZ += xTravelBB.MaxZ;
		xSourceBB.MinZ += xTravelBB.MinZ;

		//expand to include effects of gravity on the particle
		if(pxTemplate->m_bDirectionalGravity || pxTemplate->m_bPointGravity)
		{
			float fMinGravity = 0.f;
			float fMaxGravity = 0.f;
			GetGraphMinAndMax(pxTemplate->m_xGravityStrength, fMinGravity, fMaxGravity);

			if(pxTemplate->m_bDirectionalGravity)
			{
				if(fMinGravity < 0.f)
				{
					Asura_Vector_3 xTravel = pxTemplate->m_xGravityVector * fMinGravity * (0.5f * fMaxPcleLifetime * fMaxPcleLifetime);
					ApplyGravity(xSourceBB, xTravel);
				}
				if(fMaxGravity > 0.f)
				{
					Asura_Vector_3 xTravel = pxTemplate->m_xGravityVector * fMaxGravity * (0.5f * fMaxPcleLifetime * fMaxPcleLifetime);
					ApplyGravity(xSourceBB, xTravel);
				}
			}
			else
			{
				// Point gravity. Could be in any direction so do every direction
				const float fTravel = Asura_Maths::Max(fMaxGravity, -fMinGravity) * (0.5f * fMaxPcleLifetime * fMaxPcleLifetime);
				xSourceBB.MaxX += fTravel;
				xSourceBB.MinX -= fTravel;
				xSourceBB.MaxY += fTravel;
				xSourceBB.MinY -= fTravel;
				xSourceBB.MaxZ += fTravel;
				xSourceBB.MinZ -= fTravel;
			}
		}
	}

	if(pxTemplate->m_bUseSourceDynamics)
	{
		//expand to include movement of the source
		float fTravel =  m_xVelocity.Magnitude()*pxTemplate->m_fDuration;

		xSourceBB.MinX -= fTravel;
		xSourceBB.MaxX += fTravel;
		xSourceBB.MinY -= fTravel;
		xSourceBB.MaxY += fTravel;
		xSourceBB.MinZ -= fTravel;
		xSourceBB.MaxZ += fTravel;

		//and gravity on the source
		fDistance = 0.5f * g_fGravity * pxTemplate->m_fDuration * pxTemplate->m_fDuration;
		xSourceBB.MaxY += fDistance;
	}

	// Expand by maximum size of particle
	// We still want to do this even if m_bUseSourceBBForVisibility is true.
	float fMaxSize = 0.0f;
	GetGraphMax( pxParticleTemplate->m_xSize, fMaxSize );
	xSourceBB.MinX -= fMaxSize;
	xSourceBB.MaxX += fMaxSize;
	xSourceBB.MinY -= fMaxSize;
	xSourceBB.MaxY += fMaxSize;
	xSourceBB.MinZ -= fMaxSize;
	xSourceBB.MaxZ += fMaxSize;

	//expand the output BB to include the source's BB
	xOutputBB.Set(xSourceBB);
}

void Asura_PFX_Source::ApplyGravity(Asura_Bounding_Box& xBB, const Asura_Vector_3& xTravel) const
{
	if(xTravel.x > 0.0f) xBB.MaxX += xTravel.x;
	else xBB.MinX += xTravel.x;
	if(xTravel.y > 0.0f) xBB.MaxY += xTravel.y;
	else xBB.MinY += xTravel.y;
	if(xTravel.z > 0.0f) xBB.MaxZ += xTravel.z;
	else xBB.MinZ += xTravel.z;
}

void Asura_PFX_Source::GetGraphMax(const Asura_Graph<float>& xGraph, float& fMax) const
{
	for(u_int u = xGraph.GetNumDiscreteSamples();u--;)
	{
		fMax = Asura_Maths::Max(xGraph.GetDiscreteSample(u), fMax);
	}
}

void Asura_PFX_Source::GetGraphMinAndMax(const Asura_Graph<float>& xGraph, float& fMin, float& fMax) const
{
	for(u_int u = xGraph.GetNumDiscreteSamples();u--;)
	{
		fMin = Asura_Maths::Min(xGraph.GetDiscreteSample(u), fMin);
		fMax = Asura_Maths::Max(xGraph.GetDiscreteSample(u), fMax);
	}
}

void Asura_PFX_Source::GetGraphMax(const Asura_Graph<Asura_Vector_3>& xGraph, float& fMaxX, float& fMaxY) const
{
	for(u_int u = xGraph.GetNumDiscreteSamples();u--;)
	{
		fMaxX = Asura_Maths::Max(xGraph.GetDiscreteSample(u).x, fMaxX);
		fMaxY = Asura_Maths::Max(xGraph.GetDiscreteSample(u).y, fMaxY);
	}
}

void Asura_PFX_Source::Generate(float fStart, float fDuration)
{
	Asura_PFX_Source_Template *pxTemplate = Asura_PFX_System::GetSourceTempByHash(m_uTemplateNameHash);
	Asura_PFX_Particle_Template *pxParticleTemplate = pxTemplate ? Asura_PFX_System::GetParticleTempByHash(pxTemplate->m_uParticleTemplateHash) : NULL;

	const float fReleaseRate = pxTemplate->m_xReleaseRate.GetValueAtTime_Discrete(m_fAge) * m_pxEffect->GetReleaseRateScale();
	if(fReleaseRate <= 0.f) { return; }
	const float fTimeBetweenParticles = 1.f/fReleaseRate;
	
	// No point generating for longer than particle lifetime
	fDuration = Asura_Maths::Min(pxParticleTemplate->GetMaxLifetime(), fDuration);

	m_fParticleReleaseCounter = fReleaseRate * fDuration;

	// Create new particle if necessary
	while(m_fParticleReleaseCounter>=1.0f)
	{
		m_fParticleReleaseCounter -= 1.0f;
		//get frame time for this particle , based on when it was created in frame
		const float fTimer = m_fParticleReleaseCounter * fTimeBetweenParticles;
		CreateParticle(fTimer + fStart);
	}
}

const Asura_PFX_Particle* Asura_PFX_Source::GetRandomParticle() const
{
	if(m_uNumParticles)
	{
		return &m_pxParticles[GetRNG()->GetInt(0, m_uNumParticles - 1)];
	}
	else
	{
		return NULL;
	}
}



#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
void Asura_PFX_Source::UpdateRenderData()
{
	// if we're going to be rendered get a vertex buffer otherwise set to NULL
	if (m_pxEffect->GetCameraFlags(true) != 0)
	{
		m_pPlatformRenderData = Asura_PFX_System::Platform_GetRenderData(m_uNumParticles);
	}
	else
	{
		m_pPlatformRenderData = NULL;
	}
}
#endif



/// End of file. ///
