///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Streaming_ModuleList.h
// Author : Martin G. Harris
// Date   : 25-May-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
// List of modules possibly from a number of different zones.
// Used for visibility testing in the PFX system and so forth.
//
///////////////////////////////////////////////////////////////////////////////

// RJT NOTE: Required for GCC builds as the file list generation doesn't account for VC project exclusion from build
#ifdef ASURA_USE_STREAMING

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_STREAMING_MODULE_LIST_H__
#define __ASURA_STREAMING_MODULE_LIST_H__

// Includes
#include "Core/Asura_Hash.h"

#ifdef ASURA_TOOLS
#include "Misc/Tools_Render.h"
#endif //ASURA_TOOLS

// Forward declarations
class Asura_Module_Handler;
class Asura_Bounding_Box;
class Asura_Streaming_ModuleList_Base;

/////////////////////////////////////////
// Asura_Streaming_ModuleList_UserData //
/////////////////////////////////////////
// User data for callback and function calls.
class Asura_Streaming_ModuleList_UserData
{
public:
	Asura_Streaming_ModuleList_UserData(const Asura_Bounding_Box& xBox, Asura_Hash_ID* puZoneList,
										u_short* pusModuleList, u_char* pucNumModulesForZone,
										Asura_Streaming_ModuleList_Base* pxThis)
		: m_puZoneList(puZoneList)
		, m_pusModuleList(pusModuleList)
		, m_pucNumModulesForZone(pucNumModulesForZone)
		, m_pxThis(pxThis)
		, m_xBox(xBox)
	{}
	
	Asura_Hash_ID* const	m_puZoneList;
	u_short* const			m_pusModuleList;
	u_char*	const			m_pucNumModulesForZone;

	Asura_Streaming_ModuleList_Base* const m_pxThis;

	const Asura_Bounding_Box& m_xBox;
};


/////////////////////////////////////
// Asura_Streaming_ModuleList_Base //
/////////////////////////////////////
// Base class, which exists mainly so that I don't have to stick everything in the header file.
class Asura_Streaming_ModuleList_Base
{
protected:
	Asura_Streaming_ModuleList_Base(u_char ucSizeModules, u_char ucSizeZones)
		: m_aucMaxNumModules(ucSizeModules), m_aucMaxNumZones(ucSizeZones)
	{
		Reset();
	}

	void Reset()
	{
		m_aucNumModules = 0;
		m_aucNumZones   = 0;
	}

	// Is there enough space to add a new module/zone to the list?
	bool CanAddModule() const { return m_aucNumModules<m_aucMaxNumModules; }
	bool CanAddZone()   const { return m_aucNumZones<m_aucMaxNumZones; }

	u_int GetNumModules() const { return m_aucNumModules; }
	u_int GetNumZones()   const { return m_aucNumZones; }

	// Fill the list with all modules that intersect the bounding box.
	// Return true if we generated a valid list.
	bool Fill(Asura_Streaming_ModuleList_UserData& xUserData);

	// Is the module list visible to the active camera?
	bool IsVisibleToActiveCamera(const Asura_Hash_ID* const puZoneList,
								 const u_short* const pusModuleList,
								 const u_char* const pucNumModulesForZone) const;

	// Callback for module handler tree traversal.
	static bool Fill_Callback(Asura_Hash_ID uHash, Asura_Module_Handler* pxHandler, void* pUserData);

private:
	const u_char  m_aucMaxNumModules;	// Max number of modules.
	const u_char  m_aucMaxNumZones;		// Max number of zones.
	      u_char  m_aucNumModules;		// Number of modules in list.
	      u_char  m_aucNumZones;		// Number of zones in list.
};

////////////////////////////////
// Asura_Streaming_ModuleList //
////////////////////////////////
// List of modules with their associated zones.
// ucNUM_MODULES = maximum number of different modules in the list (1..254). Maximum of MAX_VARIABLE_LENGTH_CLIPPING_COUNT per zone.
// ucNUM_ZONES   = maximum number of different zones in the list (1..254).
template<u_char ucNUM_MODULES, u_char ucNUM_ZONES>
class Asura_Streaming_ModuleList : public Asura_Streaming_ModuleList_Base
{
public:
	Asura_Streaming_ModuleList()
		: Asura_Streaming_ModuleList_Base(ucNUM_MODULES, ucNUM_ZONES)
		, m_bIsValid(false)
	{}

	// Find all the modules that intersect the specified bounding box.
	// Return true if the call generated a valid list.
	// Return false if the list is invalid (eg overflowed the available space).
	bool CalculateModuleList(const Asura_Bounding_Box& xBox)
	{
		Asura_Streaming_ModuleList_UserData xUserData(xBox, m_auZoneList, m_ausModuleList, m_aucNumModulesForZone, this);
		m_bIsValid = Fill(xUserData);
		return m_bIsValid;
	}

	// Is module list visible by active camera?
	bool IsVisible() const
	{
		if (!m_bIsValid) return false;
		return IsVisibleToActiveCamera(m_auZoneList, m_ausModuleList, m_aucNumModulesForZone);
	}

	// Mark the list as invalid.
	void MakeInvalid()
	{
		m_bIsValid = false;
		Reset();
	}

	// Is the list valid? (ie we have calculated the module list and it didn't overflow)
	bool IsValid() const { return m_bIsValid; }

	// Is the module list valid, but empty? (Could happen if an object is in an unloaded zone or outside all the zones).
	bool IsEmpty() const { return m_bIsValid && (GetNumModules() == 0); }

	// Return the hash of one of the containing zones, or ASURA_HASH_UNSET if the module list is invalid or empty.
	Asura_Hash_ID GetContainingZoneHash() const
	{
		if (m_bIsValid && GetNumZones()>0)
		{
			return m_auZoneList[0];
		}
		return ASURA_HASH_ID_UNSET;
	}

#ifdef ASURA_TOOLS
	// NOTE: tools only loads a single zone.
	void RenderVisibility() const
	{
		if (m_bIsValid && GetNumZones() && GetNumModules())
		{
			for (u_int u=0; u<m_aucNumModulesForZone[0]; ++u)
			{
				Tools_Render::RenderModuleOutline(m_ausModuleList[u], 0xff00ffff, true);
			}
		}
	}
#endif //ASURA_TOOLS

private:
	Asura_Hash_ID m_auZoneList[ucNUM_ZONES];		// The hashes of the different zones.
	u_short m_ausModuleList[ucNUM_MODULES];			// The indicies of the different modules.
	u_char  m_aucNumModulesForZone[ucNUM_ZONES];	// For each zone in turn, how many modules are there in the list.
	bool    m_bIsValid;								// Is the module list valid (ie didn't overflow)?
};

#endif //__ASURA_STREAMING_MODULE_LIST_H__

#endif	// ASURA_USE_STREAMING
