
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Streaming_Jobs_FileAccess.cpp
// Author : Andy Tate
// Date   : 4-May-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

// RJT NOTE: Required for GCC builds as the file list generation doesn't account for VC project exclusion from build
#ifdef ASURA_USE_STREAMING

#include "Core/Asura.h"

// This file's header...
#include "Streaming/Jobs/Asura_Streaming_Jobs_FileAccess.h"

// Asura independent header-files...
#include "ChunkSystem/Asura_BackgroundChunkProcessing_System.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_ProcessChunkTimingList.h"
#include "FileAccess/Asura_FileAccess.h"
#include "FileAccess/Asura_FileCache.h"

// Asura platform header-files...

// Game independent header-files...

// Game platform header-files...

// Tools header-files...


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Streaming_Job_OpenAsuraFile::Update
// Author       : Andy Tate
// Date         : 4-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Streaming_Job_OpenAsuraFile::Update (float& fTimeAllowed)
{
	const Asura_Handle	iFileHandle	= Asura_FileAccess::Open(m_szFilename);

	// Call the parent update function...
	PARENT::Update(fTimeAllowed);

	// Did we succeed in opening the file?
	if(iFileHandle == ASURA_INVALID_HANDLE)
	{
		m_pxFileStream	= NULL;
		return(true);
	}
	else
	{
		m_pxFileStream	= new Asura_Chunk_File_Stream (iFileHandle);
		return(true);
	}
}

/// End of function 'Asura_Streaming_Job_OpenAsuraFile::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Streaming_Job_OpenStreamingCachedAsuraFile::Update
// Author       : Andy Tate
// Date         : 4-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Streaming_Job_OpenStreamingCachedAsuraFile::Update (float& fTimeAllowed)
{
	const Asura_FileCache*	pxFileCache	= Asura_FileAccess::GetFileCacheForStreaming();

	// Call the parent update function...
	PARENT::Update(fTimeAllowed);

	// Are we still opening the file?
	if(pxFileCache->IsOpening())
	{
		// Yes - try again next frame...
		fTimeAllowed	= 0.0f;
		return(false);
	}
	else
	{
		const Asura_Handle	iFileHandle	= Asura_FileAccess::OpenForStreaming(m_szFilename);

		// Did we succeed in opening the file?
		if(iFileHandle == ASURA_INVALID_HANDLE)
		{
			if(Asura_FileAccess::FileExists(m_szFilename))
			{
				// File exists, therefore we must assume that we couldn't open the file for
				// some reason (such as the disc not being present). We need to retry the
				// open again at a later date - ART 04/09/06.
				return(false);
			}
			else
			{
				m_pxFileStream	= NULL;
				return(true);
			}
		}
		else
		{
			m_pxFileStream	= new Asura_Chunk_File_Stream (iFileHandle);
			return(true);
		}
	}
}

/// End of function 'Asura_Streaming_Job_OpenStreamingCachedAsuraFile::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Streaming_Job_ReadAsuraFileHeader::Update
// Author       : Andy Tate
// Date         : 4-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Streaming_Job_ReadAsuraFileHeader::Update (float& fTimeAllowed)
{
	// Call the parent update function...
	PARENT::Update(fTimeAllowed);

	// Do we have a stream?
	if(m_pxFileStream == NULL)
	{
		// No - we cannot do any more...
		return(true);
	}
	else
	{
		const Asura_Handle	iFileHandle	= m_pxFileStream->GetFileHandle();

		// Yes - are we able to read it?
		if(Asura_FileAccess::CanBeRead(iFileHandle))
		{
			if(Asura_ChunkLoading_System::ProcessAsuraFileHeader(*m_pxFileStream))
			{
				return(true);
			}
			else
			{
				Asura_Assert(false, "File being read is not a valid Asura file. ");
				return(true);
			}
		}
		else
		{
			// No - try again next frame...
			fTimeAllowed	= 0.0f;
			return(false);
		}
	}
}

/// End of function 'Asura_Streaming_Job_ReadAsuraFileHeader::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Streaming_Job_ReadAsuraFileChunkData::Update
// Author       : Andy Tate
// Date         : 4-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Streaming_Job_ReadAsuraFileChunkData::Update (float& fTimeAllowed)
{
	// Call the parent update function...
	PARENT::Update(fTimeAllowed);

	// Do we have a stream?
	if(m_pxFileStream == NULL)
	{
		// No - we cannot do any more...
		return(true);
	}
	else
	{
		const Asura_Handle	iFileHandle	= m_pxFileStream->GetFileHandle();

#if defined(ASURA_USE_BACKGROUND_CHUNK_PROCESSING)
		// Yes - is this the first frame?
		if(m_uNumUpdates == 1)
		{
			Asura_BackgroundChunkProcessing_System::BeginProcessing(*m_pxFileStream);
		}
#endif	// defined(ASURA_USE_BACKGROUND_CHUNK_PROCESSING)

		// Are we able to read it?
		if(Asura_FileAccess::CanBeRead(iFileHandle))
		{
			bool	bMoreToProcess;

#if defined(ASURA_USE_BACKGROUND_CHUNK_PROCESSING)
			// Wait for the chunk to be processed (or partially processed)...
			bMoreToProcess	= Asura_BackgroundChunkProcessing_System::WaitForProcessing(fTimeAllowed);
#else
#if defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)
			// Start timing this chunk...
			Asura_ProcessChunkTimingList::BeginTiming();
#endif	// defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)
			{
				// Process the chunk itself...
				bMoreToProcess	= Asura_ChunkLoading_System::ProcessChunk(*m_pxFileStream);
			}
#if defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)
			// Stop timing this chunk...
			Asura_ProcessChunkTimingList::EndTiming(Asura_ChunkLoading_System::GetCurrentChunkID());
#endif	// defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)
#endif	// defined(ASURA_USE_BACKGROUND_CHUNK_PROCESSING)

			// Yes - decide if we have any more work to do...
			return(!bMoreToProcess);
		}
		else
		{
			// No - try again next frame...
			fTimeAllowed	= 0.0f;
			return(false);
		}
	}
}

/// End of function 'Asura_Streaming_Job_ReadAsuraFileChunkData::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Streaming_Job_CloseAsuraFile::Update
// Author       : Andy Tate
// Date         : 4-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Streaming_Job_CloseAsuraFile::Update (float& fTimeAllowed)
{
	// Call the parent update function...
	PARENT::Update(fTimeAllowed);

	// Do we have a file to close?
	if(m_pxFileStream != NULL)
	{
		const Asura_Handle	iFileHandle	= m_pxFileStream->GetFileHandle();

		// Yes - close the file...
		Asura_ChunkLoading_System::EndProcessAsuraFile(*m_pxFileStream);
		Asura_FileAccess::Close(iFileHandle);

		// ...then prevent us from doing anything else with it...
		delete(m_pxFileStream);
		m_pxFileStream	= NULL;
	}

	// We always complete after the first update...
	return(true);
}

/// End of function 'Asura_Streaming_Job_CloseAsuraFile::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Streaming_Job_CloseStreamingAsuraFile::Update
// Author       : Andy Tate
// Date         : 4-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_Streaming_Job_CloseStreamingAsuraFile::Update (float& fTimeAllowed)
{
	const Asura_FileCache*	pxFileCache	= Asura_FileAccess::GetFileCacheForStreaming();

	// Call the parent update function...
	PARENT::Update(fTimeAllowed);

	if(pxFileCache->IsClosing())
	{
		// Yes - try again next frame...
		fTimeAllowed	= 0.0f;
		return(false);
	}
	else
	{
		// Do we have a file to close?
		if(pxFileCache->IsOpen())
		{
			const Asura_Handle	iFileHandle	= m_pxFileStream->GetFileHandle();

			// Yes - close the file...
			Asura_ChunkLoading_System::EndProcessAsuraFile(*m_pxFileStream);
			Asura_FileAccess::Close(iFileHandle);
			return(false);
		}
		else
		{
			// ...then prevent us from doing anything else with it...
			delete(m_pxFileStream);
			m_pxFileStream	= NULL;

			// We always complete after the first update...
			return(true);
		}
	}
}

/// End of function 'Asura_Streaming_Job_CloseStreamingAsuraFile::Update'. ///

#endif	// ASURA_USE_STREAMING

/// End of file. ///