
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Teleport_System.cpp
// Author : Andy Tate
// Date   : 27-Feb-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Teleport/Asura_Teleport_System.h"

// Asura independent header-files...
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Intangible/Asura_ServerEntity_Teleporter.h"
#include "Entity/PhysicalObject/Asura_ClientEntity_PhysicalObject.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "Entity/Spawners/Asura_ServerEntity_TeleportDestination.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_NetMsg.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING)
#include "Teleport/Asura_Teleport_Messages.h"

#if !defined(ASURA_DONT_USE_TELEPORTERS)

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

Asura_Teleport_System::State			Asura_Teleport_System::s_eCurrentState;
Asura_Teleport_System::State			Asura_Teleport_System::s_eNewState;
u_int									Asura_Teleport_System::s_uCurrentFadeColour;
float									Asura_Teleport_System::s_fTimeInState;

float									Asura_Teleport_System::s_fFadeInDuration;
float									Asura_Teleport_System::s_fFadeOutDuration;
u_int									Asura_Teleport_System::s_uTargetFadeColour;

Asura_TeleportedPlayerData				Asura_Teleport_System::s_axTeleportedPlayerData[ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS];

ProjectTeleportServerEntityFunction*	Asura_Teleport_System::s_pfnProjectTeleportServerEntity = NULL;

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::ChangeState
// Author       : Andy Tate
// Date         : 1-Mar-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Teleport_System::ChangeState(const State eNewState)
{
	if(eNewState == s_eCurrentState)
	{
		return(false);
	}
	else
	{
		s_eNewState	= eNewState;
		return(true);
	}
}

/// End of function 'Asura_Teleport_System::ChangeState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::TeleportClientEntity
// Author       : Andy Tate
// Date         : 28-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::TeleportClientEntity(const Asura_Guid uGuidToTeleport,
												 const Asura_Vector_3& xPosition,
												 const Asura_Quat& xOrientation)
{
	Asura_ClientEntity_PhysicalObject*	pxClient	= Asura_ClientEntity_PhysicalObject::FindEntity(uGuidToTeleport);

	if(pxClient != NULL)
	{
		Asura_Matrix_3x3	xOrientationMatrix;

		xOrientation.ToMatrix(xOrientationMatrix);
		pxClient->TeleportTo(xPosition, xOrientationMatrix);
	}
}

/// End of function 'Asura_Teleport_System::TeleportClientEntity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::TeleportServerEntity
// Author       : Andy Tate
// Date         : 28-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::TeleportServerEntity(const Asura_Guid uGuidToTeleport,
												 const Asura_Vector_3& xPosition,
												 const Asura_Quat& xOrientation)
{
    // See if project wants to handle it first.
	if(s_pfnProjectTeleportServerEntity)
	{
		if( s_pfnProjectTeleportServerEntity(uGuidToTeleport, xPosition, xOrientation) )
        {
            return;
        }
	}

    Asura_ServerEntity_PhysicalObject*	pxServer	= Asura_ServerEntity_PhysicalObject::FindEntity(uGuidToTeleport);
	if(pxServer == NULL)
	{
		Asura_Assert(false, "Cannot find server-entity to teleport.");
	}
	else
	{
			Asura_Matrix_3x3	xOrientationMatrix;

		xOrientation.ToMatrix(xOrientationMatrix);
		pxServer->TeleportTo(xPosition, xOrientationMatrix, true);
	}
}

/// End of function 'Asura_Teleport_System::TeleportServerEntity'. ///


const Asura_TeleportedPlayerData* Asura_Teleport_System::GetPlayerTeleportData(const Asura_Guid uGuidTeleported)
{
	for(u_int u = 0; u < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS; u++)
	{
		if (s_axTeleportedPlayerData[u].m_uPlayerGuid == uGuidTeleported )
		{
			return (&s_axTeleportedPlayerData[u]);
		}
	}

	return NULL;
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::BeginPlayerTeleport
// Author       : Andy Tate
// Date         : 28-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::BeginPlayerTeleport(const Asura_Message_BeginPlayerTeleport* pxData)
{
	if(s_eCurrentState == UNUSED)
	{
		s_uTargetFadeColour	= pxData->GetFadeColour();

		if(s_uTargetFadeColour == 0x00000000)
		{
			s_fFadeInDuration	= 0.0f;
			s_fFadeOutDuration	= 0.0f;
		}
		else
		{
			s_fFadeInDuration	= pxData->GetFadeInDuration();
			s_fFadeOutDuration	= pxData->GetFadeOutDuration();
		}

		for(u_int u = 0; u < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS; u++)
		{
			pxData->GetDestinationData(u, s_axTeleportedPlayerData[u].m_xPosition, s_axTeleportedPlayerData[u].m_xOrientation);
		}

		if(s_fFadeInDuration == 0.0f)
		{
#if defined(ASURA_USE_STREAMING)
			ChangeState(WAITING_FOR_STREAMING);
#else
			ChangeState(TELEPORTING);
#endif	// defined(ASURA_USE_STREAMING)
		}
		else
		{
			ChangeState(FADE_OUT);
		}
	}
	else
	{
		Asura_Assert(false, "Unexpected state to teleport from.");
	}
}

/// End of function 'Asura_Teleport_System::BeginPlayerTeleport'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::TeleportPlayers
// Author       : Andy Tate
// Date         : 28-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::TeleportPlayers(void)
{
	// Client and server need to be handled differently...
	if(Asura_Network::IsServer())
	{
		for(u_int u = 0; u < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS; u++)
		{
			const Asura_Guid		uGuidToTeleport	= s_axTeleportedPlayerData[u].m_uPlayerGuid;
			const Asura_Vector_3&	xPosition		= s_axTeleportedPlayerData[u].m_xPosition;
			const Asura_Quat&		xOrientation	= s_axTeleportedPlayerData[u].m_xOrientation;

			TeleportServerEntity(uGuidToTeleport, xPosition, xOrientation);
		}
	}
	else
	{
		for(u_int u = 0; u < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS; u++)
		{
			const Asura_Guid		uGuidToTeleport	= s_axTeleportedPlayerData[u].m_uPlayerGuid;
			const Asura_Vector_3&	xPosition		= s_axTeleportedPlayerData[u].m_xPosition;
			const Asura_Quat&		xOrientation	= s_axTeleportedPlayerData[u].m_xOrientation;

			TeleportClientEntity(uGuidToTeleport, xPosition, xOrientation);
		}
	}

	// Teleport is done, time to fade-in...
	ChangeState((s_fFadeInDuration > 0.0f) ? FADE_IN : UNUSED);
}

/// End of function 'Asura_Teleport_System::TeleportPlayers'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::Initialise
// Author       : Andy Tate
// Date         : 27-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::Initialise(void)
{
	s_eCurrentState			= UNUSED;
	s_eNewState				= MAX_STATES;
	s_uCurrentFadeColour	= 0xFF000000;
	s_fTimeInState			= 0.0f;
	
	for(u_int u = 0; u < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS; u++)
	{
		s_axTeleportedPlayerData[u].m_uPlayerGuid	= ASURA_GUID_UNREGISTERED;
	}
}

/// End of function 'Asura_Teleport_System::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::Reset
// Author       : Andy Tate
// Date         : 27-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::Reset(void)
{
	Shutdown();
	Initialise();
}

/// End of function 'Asura_Teleport_System::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::Shutdown
// Author       : Andy Tate
// Date         : 27-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::Shutdown(void)
{
}

/// End of function 'Asura_Teleport_System::Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::RegisterPlayerGuid
// Author       : Andy Tate
// Date         : 1-Mar-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::RegisterPlayerGuid(const u_int uIndex,
											   const Asura_Guid uGuid)
{
	Asura_Assert(uIndex < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS, "Index out of range, increase ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS?");

	if(uIndex < ASURA_TELEPORTER_ENTITY_MAX_DESTINATIONS) // CW - Sanity check!
	{
		s_axTeleportedPlayerData[uIndex].m_uPlayerGuid	= uGuid;
	}
}

/// End of function 'Asura_Teleport_System::RegisterPlayerGuid'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::TeleportEntity
// Author       : Andy Tate
// Date         : 28-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::TeleportEntity(const Asura_Guid uGuidToTeleport,
										   const Asura_Guid uDestinationGuid)
{
	const Asura_ServerEntity_TeleportDestination*	pxDestinationEntity	= Asura_ServerEntity_TeleportDestination::FindEntity(uDestinationGuid);

	if(pxDestinationEntity == NULL)
	{
		Asura_Assert(false, "Cannot find destination to teleport to.");
	}
	else
	{
		const Asura_Vector_3&	xPosition	= pxDestinationEntity->GetPosition();

		Asura_Quat				xOrientation(pxDestinationEntity->GetOrientation());

		TeleportServerEntity(uGuidToTeleport, xPosition, xOrientation);
	}
}

/// End of function 'Asura_Teleport_System::TeleportEntity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::Update
// Author       : Andy Tate
// Date         : 27-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::Update(void)
{
	Asura_RGBA	xRGBA;
	float		fAlpha;

	// Change state and update timers...
	if(s_eNewState == MAX_STATES)
	{
		s_fTimeInState	+= Asura_Timers::GetGameTimeSlice();
	}
	else
	{
		s_eCurrentState	= s_eNewState;
		s_eNewState		= MAX_STATES;
		s_fTimeInState	= 0.0f;
	}

	// Store the target colour...
	xRGBA.FromUIntARGB(s_uTargetFadeColour);

	// Handle state updating...
	switch(s_eCurrentState)
	{
		case UNUSED:
		{
			fAlpha	= 0.0f;
			break;
		}

		case FADE_OUT:
		{
			if(s_fTimeInState < s_fFadeInDuration)
			{
				fAlpha	= (s_fTimeInState / s_fFadeInDuration);
			}
			else
			{
#if defined(ASURA_USE_STREAMING)
				ChangeState(WAITING_FOR_STREAMING);
#else
				ChangeState(TELEPORTING);
#endif	// defined(ASURA_USE_STREAMING)
				fAlpha	= 1.0f;
			}
			break;
		}

#if defined(ASURA_USE_STREAMING)
		case WAITING_FOR_STREAMING:
		{
			if(Asura_Streaming_System::GetJobQueue().Size() == 0)
			{
				ChangeState(TELEPORTING);
			}
			fAlpha	= 1.0f;
			break;
		}
#endif	// defined(ASURA_USE_STREAMING)

		case TELEPORTING:
		{
			TeleportPlayers();

			ChangeState((s_fFadeInDuration == 0.0f) ? UNUSED : FADE_IN);
			fAlpha	= 1.0f;
			break;
		}

		case FADE_IN:
		{
			if(s_fTimeInState < s_fFadeInDuration)
			{
				fAlpha	= (1.0f - (s_fTimeInState / s_fFadeInDuration));
			}
			else
			{
				ChangeState(UNUSED);
				fAlpha	= 0.0f;
			}
			break;
		}

		default:
		{
			Asura_Assert(false, "Asura_Teleport_System::Update(): Unexpected state.");
			return;
		}
	}		// End of switch.

	// Update the current fade-colour...
	xRGBA.m_fA				= fAlpha;
	s_uCurrentFadeColour	= xRGBA.ToUIntARGB();
}

/// End of function 'Asura_Teleport_System::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::Render
// Author       : Andy Tate
// Date         : 27-Feb-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_Teleport_System::Render(void)
{
	switch(s_eCurrentState)
	{
		case UNUSED:
		{
			break;
		}

		case FADE_OUT:
#if defined(ASURA_USE_STREAMING)
		case WAITING_FOR_STREAMING:
#endif	// defined(ASURA_USE_STREAMING)
		case TELEPORTING:
		case FADE_IN:
		{
			const float	fWidth	= static_cast<float>(Asura_Viewport::GetWidth());
			const float	fHeight	= static_cast<float>(Asura_Viewport::GetHeight());

			Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
			Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
			Asura_Render::Overlay.Colour			= s_uCurrentFadeColour;

			Asura_Render::Overlay.SetRectXY(0.0f, 0.0f, fWidth, fHeight);
			Asura_Render::Overlay.Render();
			break;
		}

		default:
		{
			Asura_Assert(false, "Asura_Teleport_System::Render(): Unexpected state.");
			break;
		}
	}		// End of switch.
}

/// End of function 'Asura_Teleport_System::Render'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Teleport_System::HandleClientMessages
// Author       : Andy Tate
// Date         : 27-Feb-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_Teleport_System::HandleClientMessages(const Asura_Message* pxMessage)
{
	switch(pxMessage->GetMsgID())
	{
		case ASURA_MSGID_BEGIN_PLAYER_TELEPORT:
		{
			const Asura_Message_BeginPlayerTeleport*	pxData	= static_cast<const Asura_Message_BeginPlayerTeleport*>(pxMessage->GetData());

			if(pxData != NULL)
			{
				BeginPlayerTeleport(pxData);
			}
			return(true);
		}

		default:
		{
			return(false);
		}
	}		// End of switch.
}

/// End of function 'Asura_Teleport_System::HandleClientMessages'. ///

#endif	// !defined(ASURA_DONT_USE_TELEPORTERS)

// NetMsg Registration
#ifdef ASURA_NETMSG_PACK_ALL_DATA
void Asura_Message_Teleport::RegisterFields(Asura_NetMsg* pxNetMsg) const
{
	Asura_Assert(pxNetMsg->GetMsgID() == ASURA_MSGID_TELEPORT, "unexpected message ID");
	Asura_Assertf(pxNetMsg->GetEntityClassification() == AsuraEntityClass_NotSet, 
		("unexpected entity classification (%s)", Asura_Network::GetEntityClassificationDesc(pxNetMsg->GetEntityClassification())));

	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_vector3, m_xPosition );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_quat, m_xOrientation );
}
#endif


/// End of file. ///
