
//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     Unit testing code, based on CppUnitLite
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_UNITTEST

#include "UnitTest/Asura_UnitTest.h"

#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Interface/Asura_CommandConsole.h"
#include "Debugging/Asura_Logfile.h"

#include <stdio.h>
#include <string.h> 

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_UnitTest_Test;

// C L A S S E S /////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest_Failure
//
//////////////////////////////////////////////////////////////////////////////
Asura_UnitTest_Failure::Asura_UnitTest_Failure (const Asura_String& xTestName, 
												const Asura_String& xFileName, 
												int	  			    iLineNumber,
												const Asura_String& xCondition)
: m_xMessage (xCondition), 
  m_xTestName (xTestName), 
  m_xFileName (xFileName), 
  m_iLineNumber (iLineNumber)
{
}


Asura_UnitTest_Failure::Asura_UnitTest_Failure (const Asura_String&	xTestName, 
												const Asura_String&	xFileName, 
												int					iLineNumber,
												const Asura_String&	xExpected,
												const Asura_String&	xActual)
: m_xTestName (xTestName), 
  m_xFileName (xFileName), 
  m_iLineNumber (iLineNumber)
{
	char* szPart1 = "expected ";
	char* szPart3 = " but was: ";

	char* acBuffer = new char [strlen (szPart1) 
					+ xExpected.GetLength() 
					+ strlen (szPart3)
					+ xActual.GetLength()
					+ 1];

	sprintf(acBuffer, "%s%s%s%s", 
		szPart1, 
		xExpected.GetString(), 
		szPart3, 
		xActual.GetString());

	m_xMessage = Asura_String(acBuffer);

	delete[] acBuffer;
}


//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest_TestResult
//
//////////////////////////////////////////////////////////////////////////////

Asura_UnitTest_TestResult::Asura_UnitTest_TestResult()
	: m_iFailureCount (0)
{
}


void Asura_UnitTest_TestResult::TestsStarted() 
{
#ifdef ASURA_UNITTEST_BATCHMODE
#else
	Asura_OutputToDebugger_Retail(("Started running unit tests.\n"));
#endif
}


void Asura_UnitTest_TestResult::AddFailure(const Asura_UnitTest_Failure& xFailure) 
{
#ifdef ASURA_UNITTEST_BATCHMODE
	Asura_CommandConsole::Printf("%s(%i) : unit test \"%s\" failed : %s", xFailure.m_xFileName.GetString(), xFailure.m_iLineNumber, xFailure.m_xTestName.GetString(), xFailure.m_xMessage.GetString());
#else
	Asura_OutputToDebugger_Retail(( "%s%s%s%s%s%s%ld%s%s\n",
		"Failure in test ",
		xFailure.m_xTestName.GetString(),
		": \"",
		xFailure.m_xMessage.GetString(),
		"\" " ,
		"line ",
		xFailure.m_iLineNumber,
		" in ",
		xFailure.m_xFileName.GetString()));
#endif	
	m_iFailureCount++;
}


void Asura_UnitTest_TestResult::TestsEnded(u_int uNumTestsRun)
{
#ifdef ASURA_UNITTEST_BATCHMODE
	Asura_CommandConsole::Printf("Unit testing complete - %i failure(s), %i test(s) run", m_iFailureCount, uNumTestsRun);
#else
	Asura_OutputToDebugger_Retail(("Finished running %u unit tests.\n", uNumTestsRun));
	if (m_iFailureCount > 0)
	{
        for( u_int u = 8; u--; ) { Asura_OutputToDebugger_Retail(( "********************************\n" )); }
		Asura_OutputToDebugger_Retail(("There were %d unit test failures\n", m_iFailureCount));
        for( u_int u = 8; u--; ) { Asura_OutputToDebugger_Retail(( "********************************\n" )); }
	}
	else
	{
		Asura_OutputToDebugger_Retail(("There were no test failures\n"));
	}
#endif
}


//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest_TestRegistry 
// 
// a singleton collection of all the tests to run in a system.  
// 
///////////////////////////////////////////////////////////////////////////////

class Asura_UnitTest_TestRegistry
{
public:
	static void AddTest(Asura_UnitTest_Test* pxTest);
	static void RunAllTests(Asura_UnitTest_TestResult& xResult);

private:

	static Asura_UnitTest_TestRegistry&	GetInstance();
	void					Add(Asura_UnitTest_Test* pxTest);
	void					Run(Asura_UnitTest_TestResult& xResult);

	
	Asura_UnitTest_Test* m_pxTests;

};

void Asura_UnitTest_TestRegistry::AddTest(Asura_UnitTest_Test* pxTest) 
{
	GetInstance().Add(pxTest);
}


void Asura_UnitTest_TestRegistry::RunAllTests (Asura_UnitTest_TestResult& xResult) 
{
	GetInstance().Run(xResult);
}


Asura_UnitTest_TestRegistry& Asura_UnitTest_TestRegistry::GetInstance() 
{
	static Asura_UnitTest_TestRegistry ls_Registry;
	return ls_Registry;
}


void Asura_UnitTest_TestRegistry::Add (Asura_UnitTest_Test* pxTest) 
{
	if (m_pxTests == 0) 
	{
		m_pxTests = pxTest;
		return;
	}
	
	pxTest->SetNext(m_pxTests);
	m_pxTests = pxTest;
}


void Asura_UnitTest_TestRegistry::Run (Asura_UnitTest_TestResult& xResult) 
{
	u_int uNumTests = 0;
	xResult.TestsStarted();

	for (Asura_UnitTest_Test *pxTest = m_pxTests; pxTest != 0; pxTest = pxTest->GetNext())
	{
		//Asura_OutputToDebugger_Retail(("Running test %s...\n", pxTest->GetName()));
		pxTest->Run(xResult);
		++uNumTests;
	}

	xResult.TestsEnded(uNumTests);
}



//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest_Test
// 
///////////////////////////////////////////////////////////////////////////////

Asura_UnitTest_Test::Asura_UnitTest_Test (const Asura_String& xTestName) 
	: m_xName (xTestName) 
{
	Asura_UnitTest_TestRegistry::AddTest(this);
}


Asura_UnitTest_Test *Asura_UnitTest_Test::GetNext() const
{
	return m_pxNext;
}


void Asura_UnitTest_Test::SetNext(Asura_UnitTest_Test* pxTest)
{	
	m_pxNext = pxTest;
}


// global function to run all tests
void Asura_UnitTest_RunAllTests()
{
	Asura_UnitTest_TestResult xTestResult;
    Asura_UnitTest_TestRegistry::RunAllTests(xTestResult);
}

// RJF 26-09-06: overloaded version that makes the results available to the caller
void Asura_UnitTest_RunAllTests(Asura_UnitTest_TestResult& xResult)
{
	Asura_UnitTest_TestRegistry::RunAllTests(xResult);
}

// Asura_String utility functions - should go in Asura_String.h
Asura_String StringFrom(bool bValue)
{
	char acBuffer[6];
	Asura_PrintfToBfr(acBuffer, "%s", 6, true, bValue ? "true" : "false");
	return Asura_String(acBuffer);
}

Asura_String StringFrom(const char *szValue)
{
	return Asura_String(szValue);
}

Asura_String StringFrom (int iValue)
{
	char acBuffer[20];
	Asura_PrintfToBfr(acBuffer, "%d", 20, true, iValue);
	return Asura_String(acBuffer);
}

Asura_String StringFrom (u_int uValue)
{
	char acBuffer[20];
	Asura_PrintfToBfr(acBuffer, "%u", 20, true, uValue);
	return Asura_String(acBuffer);
}

Asura_String StringFrom (float fValue)
{
	union
	{
		float	m_fValue;
		u_int	m_uValue;
	} xTemp;

	xTemp.m_fValue	= fValue;

	char acBuffer[40];
	Asura_PrintfToBfr(acBuffer, "%f [0x%08x]",
        40, true, Asura_Maths::FtoD(fValue),
        xTemp.m_uValue);
	return Asura_String(acBuffer);
}

Asura_String StringFrom (const Asura_String& xValue)
{
	return Asura_String(xValue);
}

Asura_String StringFrom (const Asura_Vector_3& xValue)
{
	char acBuffer[60];
	Asura_PrintfToBfr(acBuffer, "{%f,%f,%f}", 60, true, Asura_Maths::FtoD(xValue.x), Asura_Maths::FtoD(xValue.y), Asura_Maths::FtoD(xValue.z));
	return Asura_String(acBuffer);
}

Asura_String StringFrom(const Asura_IntVector_2& xValue)
{
	char acBuffer[40];
	Asura_PrintfToBfr(acBuffer, "{%d,%d}", 20, true, xValue.iX, xValue.iY);
	return Asura_String(acBuffer);
}

#endif // #ifdef ASURA_SUPPORT_UNITTEST
