//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest.h
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     Unit testing code, based on CppUnitLite
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_UNITTEST_H___
#define ___ASURA_UNITTEST_H___ 1

#ifdef ASURA_SUPPORT_UNITTEST

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura_String.h"

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Vector_3;
class Asura_IntVector_2;
class Asura_UnitTest_TestResult;

// M A C R O S ///////////////////////////////////////////////////////////////

#ifndef ASURA_UNITTEST_FAILURE_ACTION
#define ASURA_UNITTEST_FAILURE_ACTION {}	// can be defined e.g. to {asm("break 1");asm("nop");}
#endif

// ASURA_UNITTEST - declares a class derived from Asura_UnitTest_Test with a 
// name based on the testName and testGroup macro parameters which overrides 
// the Run() method. It then creates an instance of the class. Finally it 
// begins the definition of the Run() method for you. You must complete this
// definition (i.e. write the bit in curly braces) by writing some tests using 
// the other test macros in this file. The test is automatically added to the 
// test registry by the Asura_UnitTest_Test constructor. 
// Example usage (must be in file scope):
// 		ASURA_UNITTEST(Addition, IntVector2)
// 		{
// 		    Asura_IntVector_2 xV1(1,2);
// 		    Asura_IntVector_2 xV2(3,5);
// 		    Asura_IntVector_2 xV3(4,7);
// 		    ASURA_UNITTEST_CHECK_EQUAL( xV3, xV1 + xV2 );
// 		}
#define ASURA_UNITTEST(testName, testGroup)\
	class testGroup##testName##Test : public Asura_UnitTest_Test \
	{ \
		public: \
		testGroup##testName##Test () : Asura_UnitTest_Test (#testGroup": "#testName) {} \
			void Run (Asura_UnitTest_TestResult& xResult); \
	} \
    testGroup##testName##Instance; \
	void testGroup##testName##Test::Run (Asura_UnitTest_TestResult& xResult) 



#define ASURA_UNITTEST_CHECK(condition)\
{ \
	if (!(condition)) \
	{ \
		xResult.AddFailure(Asura_UnitTest_Failure(m_xName, __FILE__,__LINE__, #condition)); \
		ASURA_UNITTEST_FAILURE_ACTION; \
	} \
}



#define ASURA_UNITTEST_CHECK_EQUAL(expected,actual)\
{ \
	if ((expected) != (actual)) \
	{ \
		xResult.AddFailure(Asura_UnitTest_Failure(m_xName, __FILE__, __LINE__, StringFrom(expected), StringFrom(actual))); \
		ASURA_UNITTEST_FAILURE_ACTION; \
	} \
}


#define ASURA_UNITTEST_FLOATS_EQUAL(fExpected, fActual, fThreshold)\
{ \
	float fExpectedTemp = fExpected; \
	float fActualTemp = fActual; \
	if (fabsf((fExpectedTemp)-(fActualTemp)) > (fThreshold)) \
	{ \
		xResult.AddFailure (Asura_UnitTest_Failure (m_xName, __FILE__, __LINE__, \
			StringFrom((fExpectedTemp)), StringFrom((fActualTemp)))); \
		ASURA_UNITTEST_FAILURE_ACTION; \
	} \
}


#define ASURA_UNITTEST_FLOATARRAY_EQUAL(afExpected, afActual, uArraySize, fThreshold)\
{ \
	for (u_int u = 0; u < uArraySize; ++u) \
	{ \
		if (fabsf((afExpected[u])-(afActual[u])) > (fThreshold)) \
		{ \
			Asura_String xExpected = "{ "; \
			Asura_String xActual = "{ "; \
			for (u_int u2 = 0; u2 < uArraySize; ++u2) \
			{ \
				Asura_String xPostfix = (u2 == (uArraySize - 1)) ? " }" : ", "; \
				xExpected += StringFrom((afExpected[u2])); \
				xExpected += xPostfix; \
				xActual += StringFrom((afActual[u2])); \
				xActual += xPostfix; \
			} \
			xResult.AddFailure (Asura_UnitTest_Failure (m_xName, __FILE__, __LINE__, \
				xExpected, xActual)); \
			ASURA_UNITTEST_FAILURE_ACTION; \
			break; \
		} \
	} \
}


#define ASURA_UNITTEST_VECTOR3_EQUAL(xExpected, xActual, fEpsilon)\
{ \
	if (!xExpected.ApproximatelyEquals(xActual, fEpsilon)) \
	{ \
		xResult.AddFailure (Asura_UnitTest_Failure (m_xName, __FILE__, __LINE__, \
			StringFrom((xExpected)), StringFrom((xActual)))); \
		ASURA_UNITTEST_FAILURE_ACTION; \
	} \
}



#define ASURA_UNITTEST_FAIL(text) \
{ \
	xResult.AddFailure(Asura_UnitTest_Failure(m_xName, __FILE__, __LINE__,(text))); \
	ASURA_UNITTEST_FAILURE_ACTION; \
	return; \
}

// Tests an array for approximate uniformity. Ensure all entries in array are within
// 5-sigma of the mean number of entries per element.
#define ASURA_UNITTEST_ARRAY_UNIFORM_POPULATION(afArray, iNumberElements ) \
{ \
	float fMean = 0; \
	for (int i=0;i < iNumberElements; ++i) \
	{ \
		fMean +=static_cast<float>(afArray[i]); \
	} \
	fMean = fMean/static_cast<float>(iNumberElements); \
	float fStandardDeviation = Asura_Maths::Sqrt(fMean); \
	for (int i = 0; i < iNumberElements; ++i) \
	{ \
		if ( (fMean-afArray[i])/fStandardDeviation > 5.0f ) \
		{ \
			xResult.AddFailure (Asura_UnitTest_Failure (m_xName, __FILE__, __LINE__, \
			"app. equal array element pop.", \
			" not the case here: one or more array elements out of bounds")); \
			ASURA_UNITTEST_FAILURE_ACTION; \
			break; \
		} \
	} \
}

// F U N C T I O N   D E C L A R A T I O N S /////////////////////////////////
extern void Asura_UnitTest_RunAllTests();
// RJF 26-09-06: overloaded version that makes the results available to the caller
extern void Asura_UnitTest_RunAllTests(Asura_UnitTest_TestResult& xResult);

Asura_String StringFrom(bool bValue);
Asura_String StringFrom(const char *szValue);
Asura_String StringFrom(int iValue);
Asura_String StringFrom(u_int uValue);
Asura_String StringFrom(float fValue);
Asura_String StringFrom(const Asura_String& xValue);
Asura_String StringFrom(const Asura_Vector_3& xValue);
Asura_String StringFrom(const Asura_IntVector_2& xValue);



// C L A S S E S /////////////////////////////////////////////////////////////
class Asura_UnitTest_Test
{
public:
	Asura_UnitTest_Test(const Asura_String& xTestName);
	virtual ~Asura_UnitTest_Test() {}

	virtual void	Run (Asura_UnitTest_TestResult& xResult) = 0;


	void					SetNext(Asura_UnitTest_Test* pxTest);
	Asura_UnitTest_Test*	GetNext() const;

	const char*				GetName() const { return m_xName.GetString(); }

protected:

	// not sure it's worth having real functions when the macros can do this for us.
	// Obviously the functions are more type-safe but they're less flexible as you 
	// need an overloaded version for each type. Templates would be a good 
	// alternative, maybe we'll get round to that someday.
	//bool Check(int iExpected, int iActual, Asura_UnitTest_TestResult& xResult, const Asura_String& xFileName, int iLineNumber);
	//bool Check(const Asura_String& xExpected, const Asura_String& xActual, Asura_UnitTest_TestResult& xResult, const Asura_String& xFileName, int iLineNumber);

	Asura_String			m_xName;
	Asura_UnitTest_Test*	m_pxNext;

};

class Asura_UnitTest_Failure
{

public:
	Asura_UnitTest_Failure (const Asura_String&	xTestName, 
							const Asura_String&	xFileName, 
							int	  				iLineNumber,
							const Asura_String&	xCondition);

	Asura_UnitTest_Failure (const Asura_String&	xTestName, 
							const Asura_String&	xFileName, 
							int					iLineNumber,
							const Asura_String&	xExpected,
							const Asura_String&	xActual);

	Asura_String		m_xMessage;
	Asura_String		m_xTestName;
	Asura_String		m_xFileName;
	int					m_iLineNumber;
};

class Asura_UnitTest_TestResult
{
public:
	Asura_UnitTest_TestResult(); 
	void TestsStarted ();
	void AddFailure (const Asura_UnitTest_Failure& xFailure);
	void TestsEnded (u_int uNumTestsRun);

	int GetFailureCount() { return m_iFailureCount; }
private:
	int m_iFailureCount;
};

// KOR - 08/12/2006 - workaround for broken Wii compiler.
// Codewarrior uses rather overzealous dead-stripping, and gets rid of
// globals that aren't referenced, even if they do useful stuff in their
// constructors, like our unit tests. This pragma forces it to include them.
// We should be able to remove it when they fix the compiler.
#ifdef RVL
#pragma force_active on
#endif

#endif // #ifdef ASURA_SUPPORT_UNITTEST

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_UNITTEST_H___
