//////////////////////////////////////////////////////////////////////////////
//
// Asura_UnitTest_Maths.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     Unit tests for Asura_Maths
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_UNITTEST

#include "UnitTest/Asura_UnitTest.h"

#include "Core/Asura_Maths.h"

static const float fSMALLEPSILON = 1e-6f;	// on PSP we were OK with this at 1e-10 but on PC needs to be 1e-6
static const float fEPSILON = 1e-5f;
static const float fEPSILON_3DP = 1e-3f;

static const float fSinZero = 0.0f;
static const float fSin45 = 0.707106f;
static const float fSin90 = 1.0f;
static const float fSin135 = 0.707106f;
static const float fSin180 = 0.0f;
static const float fSin225 = -0.707106f;
static const float fSin360 = 0.0f;
static const float fSinMinus90 = -1.0f;

static const float fCosZero = 1.0f;
static const float fCos45 = 0.707106f;
static const float fCos90 = 0.0f;
static const float fCos135 = -0.707106f;
static const float fCos180 = -1.0f;
static const float fCos225 = -0.707106f;
static const float fCos360 = 1.0f;
static const float fCosMinus90 = -0.0f;

static const float fTanZero = 0.0f;
static const float fTan45 = 1.0f;
static const float fTanMinus45 = -1.0f;
static const float fTan10 = 0.1763269807f;
static const float fTan20 = 0.3639702343f;

ASURA_UNITTEST(DegToRad, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::Pi, 3.1415926536f, fSMALLEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( 0.0f, Asura_Maths::DegToRad(0.0f), 0.0f  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::Pi/2.0f, Asura_Maths::DegToRad(90.0f), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::Pi, Asura_Maths::DegToRad(180.0f), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::Pi * 2.0f, Asura_Maths::DegToRad(360.0f), fEPSILON  );
}

ASURA_UNITTEST(RadToDeg, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( 0.0f, Asura_Maths::RadToDeg(0.0f), 0.0f  );
	ASURA_UNITTEST_FLOATS_EQUAL( (90.0f), Asura_Maths::RadToDeg(Asura_Maths::Pi/2.0f), fEPSILON  );

    // (MP - 02.09.2008) The next 2 tests are particularly inaccurate on
    // PS2 debug for some reason. However, I don't think these tests are
    // that great anyway; in release the compiler is almost certainly
    // spotting how many components of the expression cancel each other out
    // and making it much simpler (and more accurate).
	ASURA_UNITTEST_FLOATS_EQUAL( (180.0f), Asura_Maths::RadToDeg(Asura_Maths::Pi), 2.0f * fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( (360.0f), Asura_Maths::RadToDeg(Asura_Maths::Pi * 2.0f), 4.0f * fEPSILON  );
}

ASURA_UNITTEST(Sine, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( fSinZero, Asura_Maths::Sine(0.0f), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin45, Asura_Maths::Sine(Asura_Maths::DegToRad(45.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin90, Asura_Maths::Sine(Asura_Maths::DegToRad(90.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin135, Asura_Maths::Sine(Asura_Maths::DegToRad(135.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin180, Asura_Maths::Sine(Asura_Maths::DegToRad(180.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin225, Asura_Maths::Sine(Asura_Maths::DegToRad(225.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin360, Asura_Maths::Sine(Asura_Maths::DegToRad(360.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinMinus90, Asura_Maths::Sine(Asura_Maths::DegToRad(-90.0f)), fEPSILON  );
	// KJS Added 28/11/05 (randomize x 1->2Pi) 
	// Check sine is anti-symmetric & maps onto cosine
	for (u_int u = 0; u < 100; ++u) 
	{ 
		
		float fRandomFloat = Asura_Random::GetFloat(0.0f, Asura_Maths::Pi*2.0f);
		ASURA_UNITTEST_FLOATS_EQUAL( 
			Asura_Maths::Sine(fRandomFloat), -1.0f*Asura_Maths::Sine(-1.0f*fRandomFloat), fEPSILON );
		ASURA_UNITTEST_FLOATS_EQUAL( 
			Asura_Maths::Sine(fRandomFloat), Asura_Maths::Cosine(fRandomFloat-Asura_Maths::Pi*0.5f), fEPSILON );

	}
}

ASURA_UNITTEST(ArcSine, Trig)
{
	// assume (and check) ArcSine returns results in [-90, 90]
	ASURA_UNITTEST_FLOATS_EQUAL( 0.0f, Asura_Maths::ArcSine( fSinZero ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(45.0f), Asura_Maths::ArcSine( fSin45 ), fEPSILON_3DP  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(90.0f), Asura_Maths::ArcSine( fSin90 ), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(45.0f), Asura_Maths::ArcSine( fSin135 ), fEPSILON_3DP  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(0.0f), Asura_Maths::ArcSine( fSin180 ), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(-45.0f), Asura_Maths::ArcSine( fSin225 ), fEPSILON_3DP  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(0.0f), Asura_Maths::ArcSine( fSin360 ), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(-90.0f), Asura_Maths::ArcSine( fSinMinus90 ), fEPSILON  );
}

ASURA_UNITTEST(Cos, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( fCosZero, Asura_Maths::Cosine(0.0f), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos45, Asura_Maths::Cosine(Asura_Maths::DegToRad(45.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos90, Asura_Maths::Cosine(Asura_Maths::DegToRad(90.0f)), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos135, Asura_Maths::Cosine(Asura_Maths::DegToRad(135.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos180, Asura_Maths::Cosine(Asura_Maths::DegToRad(180.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos225, Asura_Maths::Cosine(Asura_Maths::DegToRad(225.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos360, Asura_Maths::Cosine(Asura_Maths::DegToRad(360.0f)), fEPSILON  );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosMinus90, Asura_Maths::Cosine(Asura_Maths::DegToRad(-90.0f)), fEPSILON  );
	// KJS Added 28/11/05 (randomize x 1->2Pi) 
	// Check cosine is symmetric & maps onto sine
	for (u_int u = 0; u < 100; ++u) 
	{ 
		float fRandomFloat = Asura_Random::GetFloat(0.0f, Asura_Maths::Pi*2.0f);
		ASURA_UNITTEST_FLOATS_EQUAL( 
			Asura_Maths::Cosine(fRandomFloat), Asura_Maths::Cosine(-1.0f*fRandomFloat), fEPSILON );
		ASURA_UNITTEST_FLOATS_EQUAL( 
			Asura_Maths::Cosine(fRandomFloat), Asura_Maths::Sine(fRandomFloat+Asura_Maths::Pi*0.5f), fEPSILON );
	}
}

ASURA_UNITTEST(ArcCos, Trig)
{
	// assume (and check) ArcCos returns results in [0, 180]
	ASURA_UNITTEST_FLOATS_EQUAL( 0.0f, Asura_Maths::ArcCosine( fCosZero ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(45.0f), Asura_Maths::ArcCosine( fCos45 ), fEPSILON_3DP );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(90.0f), Asura_Maths::ArcCosine( fCos90 ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(135.0f), Asura_Maths::ArcCosine( fCos135 ), fEPSILON_3DP );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(180.0f), Asura_Maths::ArcCosine( fCos180 ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(135.0f), Asura_Maths::ArcCosine( fCos225 ), fEPSILON_3DP );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(0.0f), Asura_Maths::ArcCosine( fCos360 ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(90.0f), Asura_Maths::ArcCosine( fCosMinus90 ), fEPSILON );
}

ASURA_UNITTEST(Tan, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( fTanZero, Asura_Maths::Tangent(Asura_Maths::DegToRad(0.0f)), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fTan45, Asura_Maths::Tangent(Asura_Maths::DegToRad(45.0f)), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fTanMinus45, Asura_Maths::Tangent(Asura_Maths::DegToRad(-45.0f)), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fTan10, Asura_Maths::Tangent(Asura_Maths::DegToRad(10.0f)), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fTan20, Asura_Maths::Tangent(Asura_Maths::DegToRad(20.0f)), fEPSILON );
	// KJS Added 28/11/05 (randomize x 1->2Pi) 
	// Check Tangent is anti-symmetric
	for (u_int u = 0; u < 100; ++u)
	{ 
		float fRandomFloat = Asura_Random::GetFloat(0.0f, Asura_Maths::Pi*2.0f);
		ASURA_UNITTEST_FLOATS_EQUAL( 
			Asura_Maths::Tangent(fRandomFloat), -1.0f*Asura_Maths::Tangent(-1.0f*fRandomFloat), fEPSILON );
	}
}

ASURA_UNITTEST(ArcTan, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(0.0f), Asura_Maths::ArcTangent( fTanZero, 1.0f ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(45.0f), Asura_Maths::ArcTangent( fTan45, 1.0f ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(-45.0f), Asura_Maths::ArcTangent( fTanMinus45, 1.0f ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(10.0f), Asura_Maths::ArcTangent( fTan10, 1.0f ), fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( Asura_Maths::DegToRad(20.0f), Asura_Maths::ArcTangent( fTan20, 1.0f ), fEPSILON );
}

ASURA_UNITTEST(SinCos, Trig)
{
	float fSin, fCos;
	
	Asura_Maths::SinCos( 0.0f, fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinZero, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosZero, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(45.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin45, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos45, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(90.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin90, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos90, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(135.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin135, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos135, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(180.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin180, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos180, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(225.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin225, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos225, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(360.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin360, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos360, fCos, fEPSILON );

	Asura_Maths::SinCos( Asura_Maths::DegToRad(-90.0f), fSin, fCos );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinMinus90, fSin, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosMinus90, fCos, fEPSILON );
}


ASURA_UNITTEST(SinCos2, Trig)
{
	float fSin0, fCos0;
	float fSin1, fCos1;
	
	Asura_Maths::SinCos2( 0.0f, Asura_Maths::DegToRad(45.0f), fSin0, fCos0, fSin1, fCos1 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinZero, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosZero, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin45, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos45, fCos1, fEPSILON);

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(45.0f), 0.0f, fSin1, fCos1, fSin0, fCos0 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinZero, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosZero, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin45, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos45, fCos1, fEPSILON );

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(90.0f), Asura_Maths::DegToRad(135.0f), fSin0, fCos0, fSin1, fCos1 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin90, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos90, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin135, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos135, fCos1, fEPSILON );

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(135.0f), Asura_Maths::DegToRad(90.0f), fSin1, fCos1, fSin0, fCos0 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin90, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos90, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin135, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos135, fCos1, fEPSILON );

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(180.0f), Asura_Maths::DegToRad(225.0f), fSin0, fCos0, fSin1, fCos1 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin180, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos180, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin225, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos225, fCos1, fEPSILON );

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(225.0f), Asura_Maths::DegToRad(180.0f), fSin1, fCos1, fSin0, fCos0 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin180, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos180, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin225, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos225, fCos1, fEPSILON );

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(360.0f), Asura_Maths::DegToRad(-90.0f), fSin0, fCos0, fSin1, fCos1 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin360, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos360, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinMinus90, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosMinus90, fCos1, fEPSILON );

	Asura_Maths::SinCos2( Asura_Maths::DegToRad(-90.0f), Asura_Maths::DegToRad(360.0f), fSin1, fCos1, fSin0, fCos0 );
	ASURA_UNITTEST_FLOATS_EQUAL( fSin360, fSin0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCos360, fCos0, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fSinMinus90, fSin1, fEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( fCosMinus90, fCos1, fEPSILON );
}

ASURA_UNITTEST(SquareRoot, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( 0.0f, Asura_Maths::Sqrt(0.0f), 0.0f );
	ASURA_UNITTEST_FLOATS_EQUAL( 1.0f, Asura_Maths::Sqrt(1.0f), 0.0f );
	ASURA_UNITTEST_FLOATS_EQUAL( 2.0f, Asura_Maths::Sqrt(4.0f), fSMALLEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( 3.0f, Asura_Maths::Sqrt(9.0f), fSMALLEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( 154.365f, Asura_Maths::Sqrt(154.365f * 154.365f), fEPSILON );

	for (u_int u = 0; u < 100; ++u)
	{
		float fRandomFloat = Asura_Random::GetFloat(0.0f, 999.9f);
		ASURA_UNITTEST_FLOATS_EQUAL( fRandomFloat, Asura_Maths::Sqrt(fRandomFloat * fRandomFloat), fEPSILON_3DP );
	}
}

ASURA_UNITTEST(InvSquareRoot, Trig)
{
	ASURA_UNITTEST_FLOATS_EQUAL( 1.0f, Asura_Maths::InvSqrt(1.0f), 0.0f );
	ASURA_UNITTEST_FLOATS_EQUAL( 0.5f, Asura_Maths::InvSqrt(4.0f), fSMALLEPSILON );
	ASURA_UNITTEST_FLOATS_EQUAL( 0.333333333f, Asura_Maths::InvSqrt(9.0f), fEPSILON );

	for (u_int u = 0; u < 100; ++u)
	{
		float fRandomFloat = Asura_Random::GetFloat(0.1f, 999.9f);
		ASURA_UNITTEST_FLOATS_EQUAL( 1.0f / Asura_Maths::Sqrt(fRandomFloat), Asura_Maths::InvSqrt(fRandomFloat), fEPSILON_3DP );
	}
}

#endif // #ifdef ASURA_SUPPORT_UNITTEST

// E N D   O F   F I L E /////////////////////////////////////////////////////
