//
// Asura_Visibility_Node.cpp
//
// Created by KJFL - 18 December 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"


#if defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

// This file's header.
#include "Visibility/Asura_Visibility_Node.h"


#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Memory/Asura_Mem.h"
#include "Visibility/Asura_Visibility_System.h"


void Asura_Visibility_Bitfield::Save( Asura_Chunk_Stream& xStream ) const
{
	// If we are an all zero bitfield then the data ptr is zero,
	// and the compressed data size is actually the number of zeros.

	if ( m_pucCompressedData )
	{
		xStream << m_uSizeOfCompressedData;

		for ( u_int u=0; u<m_uSizeOfCompressedData; ++u )
		{
			xStream << m_pucCompressedData[u];
		}
	}
	else
	{
		const u_int uZero = 0;

		// use zero to indicate an all zero bitfield, followed by the
		// number of zeros
		xStream << uZero;
		xStream << m_uSizeOfCompressedData;
	}
}

void Asura_Visibility_Bitfield::Load( Asura_Chunk_Stream& xStream )
{
	Reset();

	xStream >> m_uSizeOfCompressedData;

	if ( m_uSizeOfCompressedData )
	{
		m_pucCompressedData = new u_char[ m_uSizeOfCompressedData ];

		for ( u_int u=0; u<m_uSizeOfCompressedData; ++u )
		{
			xStream >> m_pucCompressedData[u];
		}
	}
	else // all zero bitfield, next number is how many zeros
	{
		xStream >> m_uSizeOfCompressedData;
	}
}

void Asura_Visibility_Bitfield::Compress( const bool* const pbSource, u_int uNumberOfBytes )
{
	Reset();

	const u_int uOriginalSize = uNumberOfBytes;

	if ( !uOriginalSize ) return;

	if ( !pbSource )
	{
		// an all zero bitfield
		m_uSizeOfCompressedData = uNumberOfBytes;
		return;
	}

	bool bCurrentData = pbSource[0];
	u_char ucCurrentRun = 0;
	++m_uSizeOfCompressedData;

	for ( u_int u=1; u<uOriginalSize; ++u )
	{
		if ( bCurrentData == pbSource[u] )
		{
			++ucCurrentRun;

			if ( ucCurrentRun == 128 )
			{
				++m_uSizeOfCompressedData;
				ucCurrentRun = 0;
			}	
		}
		else 
		{
			++m_uSizeOfCompressedData;
			bCurrentData = pbSource[u];
			ucCurrentRun = 0;
		}
	}

	if ( !m_uSizeOfCompressedData ) return;

	m_pucCompressedData = new u_char [ m_uSizeOfCompressedData ];

	bCurrentData = pbSource[0];
	ucCurrentRun = 0;

	u_int uCompressionIndex=0;

	for ( u_int u=1; u<uOriginalSize; ++u )
	{
		Asura_Assert( uCompressionIndex <= m_uSizeOfCompressedData, "Code bug: invalid index" );

		if ( bCurrentData == pbSource[u] )
		{
			++ucCurrentRun;

			if ( ucCurrentRun == 128 )
			{
				u_char ucCompressedByte = 127;

				if (bCurrentData) ucCompressedByte |= 128;

				m_pucCompressedData[uCompressionIndex++] = ucCompressedByte;

				ucCurrentRun = 0;
			}	
		}
		else 
		{
			u_char ucCompressedByte = ucCurrentRun;

			if (bCurrentData) ucCompressedByte |= 128;

			m_pucCompressedData[uCompressionIndex++] = ucCompressedByte;

			bCurrentData = pbSource[u];
			ucCurrentRun = 0;
		}
	}

	{
		u_char ucCompressedByte = ucCurrentRun;

		if (bCurrentData) ucCompressedByte |= 128;

		m_pucCompressedData[uCompressionIndex] = ucCompressedByte;
	}
}

u_int Asura_Visibility_Bitfield::Decompress( u_char* pucTarget ) const
{
	if ( !m_pucCompressedData )
	{
		// must be all zeros
		Asura_Mem::Set( pucTarget, 0, sizeof( u_char ) * m_uSizeOfCompressedData );
		return m_uSizeOfCompressedData;
	}

	u_int uDecompressedBytes = 0;

	for ( u_int u=0; u<m_uSizeOfCompressedData; ++u )
	{
		u_char ucCompressedData = m_pucCompressedData[u];
		u_char ucByteToWrite = ucCompressedData >> 7;
		u_int uNumberOfBytes = (ucCompressedData & 127) + 1;

		for ( u_int v=0; v<uNumberOfBytes; ++v )
		{
			pucTarget[uDecompressedBytes] = ucByteToWrite;
			uDecompressedBytes++;
		}
	}

	return uDecompressedBytes;
}

u_int Asura_Visibility_Bitfield::Decompress( bool* pbTarget ) const
{
	if ( !m_pucCompressedData )
	{
		Asura_Assert( m_uSizeOfCompressedData, "Code bug: bitfield truly empty" );
		// must be all zeros
		Asura_Mem::Set( pbTarget, 0, sizeof( bool ) * m_uSizeOfCompressedData );
		return m_uSizeOfCompressedData;
	}

	u_int uDecompressedBytes = 0;

	for ( u_int u=0; u<m_uSizeOfCompressedData; ++u )
	{
		u_char ucCompressedData = m_pucCompressedData[u];
		bool bToWrite = (ucCompressedData >> 7)!=0;
		u_int uNumberOfBytes = (ucCompressedData & 127) + 1;

		for ( u_int v=0; v<uNumberOfBytes; ++v )
		{
			pbTarget[uDecompressedBytes] = bToWrite;
			uDecompressedBytes++;
		}
	}

	return uDecompressedBytes;
}


bool Asura_Visibility_Bitfield::Query( u_int uIndex ) const
{
	// nice and easy if there's no data, since that indicates
	// an all zero bitfield.
	if ( !m_pucCompressedData ) return false;

	u_int uDecompressedBytes = 0;

	for ( u_int u=0; u<m_uSizeOfCompressedData; ++u )
	{
		u_char ucCompressedData = m_pucCompressedData[u];
		u_int uNumberOfBytes = (ucCompressedData & 127) + 1;

		if ( uNumberOfBytes + uDecompressedBytes > uIndex )
		{
			u_char ucByteToWrite = ucCompressedData >> 7;
			return ucByteToWrite!=0;
		}

		uDecompressedBytes += uNumberOfBytes;
	}

	return false;
}

bool Asura_Visibility_Bitfield::Set( u_int uIndex ) 
{
	// TODO - cope with empty values
	u_int uDecompressedSize = 0;

	if ( m_pucCompressedData )
	{
		// TODO - cunning in place method without decompress

		for ( u_int u=0; u<m_uSizeOfCompressedData; ++u )
		{
			u_char ucCompressedData = m_pucCompressedData[u];
			u_int uNumberOfBytes = (ucCompressedData & 127) + 1;

			if ( uDecompressedSize <= uIndex && uNumberOfBytes + uDecompressedSize > uIndex )
			{
				u_char ucValue = ucCompressedData >> 7;

				// already set
				if (ucValue) return false;
			}

			uDecompressedSize += uNumberOfBytes;
		}
	}
	else
	{
		uDecompressedSize = m_uSizeOfCompressedData;
	}

	if ( uIndex >= uDecompressedSize )
	{
		Asura_Assert( false, "Code bug: index out of range" );
		return false;
	}

	bool* pbTempBuffer = new bool[ uDecompressedSize ];

    Decompress( pbTempBuffer );

    Asura_Assert( !pbTempBuffer[uIndex], "Should have already eliminated redundant set above" );
	pbTempBuffer[uIndex] = true;

	Compress( pbTempBuffer, uDecompressedSize );

	delete [] pbTempBuffer;

	return true;

}

// Given an array, fill in the given number of features with our visibility info.
bool Asura_Visibility_Leaf::SetVisibleFeaturesArray( u_char* pucArray, u_char ucBitPattern, u_int uNumberOfFeatures ) const 
{
	if ( !m_pucFeatureVisibilities ) return false;

	Asura_Mem::Copy( pucArray, m_pucFeatureVisibilities, uNumberOfFeatures );

	return true;
}

#ifdef ASURA_TOOLS // Data can only be refined in Tools builds 

bool Asura_Visibility_Leaf::Tools_RefineFeatureVisibility( u_int uFeatureID0, u_int uFeatureID1 )
{
	if ( !m_pucFeatureVisibilities ) return false;

	Asura_Assert( uFeatureID0 < Asura_Visibility_System::GetNumberOfVisibilityFeatures(), "Code bug: index out of range" );
	Asura_Assert( uFeatureID1 < Asura_Visibility_System::GetNumberOfVisibilityFeatures(), "Code bug: index out of range" );

	bool bChanged = (m_pucFeatureVisibilities[ uFeatureID0 ]==0) || (m_pucFeatureVisibilities[ uFeatureID1 ]==0);

	m_pucFeatureVisibilities[ uFeatureID0 ] = 1;
	m_pucFeatureVisibilities[ uFeatureID1 ] = 1;

	return bChanged;
}

bool Asura_Visibility_Leaf::Tools_RefineLeafVisibility( u_int uLeafIndex )
{
	// returns true if changed
	return m_xLeafVisibilities.Set( uLeafIndex );
}

void Asura_Visibility_Leaf::Tools_DeleteModule( u_int uModuleIndex )
{
	// Replace visibility of module index with last module's, and move the
	// skybox into what was the last module's slot.
	u_int uSkybox = Asura_Visibility_System::GetNumberOfVisibilityFeatures()-1;
	u_int uLastModule = uSkybox - 1;

	m_pucFeatureVisibilities[ uModuleIndex ] = m_pucFeatureVisibilities[ uLastModule ];
	m_pucFeatureVisibilities[ uLastModule ] = m_pucFeatureVisibilities[ uSkybox ];
}

void Asura_Visibility_Leaf::Tools_RemoveLinks( void )
{
	// set everything to zero
	m_xLeafVisibilities.Compress( 0, Asura_Visibility_System::GetNumberOfVisibilityLeaves() );

	Asura_Mem::Set( m_pucFeatureVisibilities, 0, Asura_Visibility_System::GetNumberOfVisibilityFeatures() );
}
	

#endif


void Asura_Visibility_Leaf::Compress( const bool* const pbSource, u_int uNumberOfBytes )
{
	m_xLeafVisibilities.Compress( pbSource, uNumberOfBytes );
}

void Asura_Visibility_Leaf::Decompress( bool* pbTarget )
{
#if defined(ASURA_ASSERT)
	const u_int uDecompressedBytes =
#endif	// defined(ASURA_ASSERT)
	m_xLeafVisibilities.Decompress( pbTarget );

	Asura_Assert( uDecompressedBytes == Asura_Visibility_System::GetNumberOfVisibilityLeaves(), "Code bug: decompression failure" );
}

bool Asura_Visibility_Leaf::IsLeafVisible( u_int uLeafIndex ) const
{
	return m_xLeafVisibilities.Query( uLeafIndex );
}

bool Asura_Visibility_Leaf::IsFeatureVisible( u_int uFeatureIndex ) const
{
	return m_pucFeatureVisibilities[ uFeatureIndex ] != 0;
}



#endif //#if defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

