//
// Asura_Visibility_System.h
//
// Created by KJFL - 06 May 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_VISIBILITY_SYSTEM_H__
#define __ASURA_VISIBILITY_SYSTEM_H__

#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM

// shame about this camera include - currently used to get ASURA_MAX_NO_OF_CAMERAS
#include "Camera/Asura_Camera.h"

const u_int uASURA_VISIBILITY_INVALID_NODE_INDEX = 0xFFFFFFFF;

// KJFL - 13 February 2008 - Asura_Visibility_Query
//
// A small helper class to encapsulate caching the "last known node" of the visibility system
// (no Reset is required as the visibility system will automatically cope with invalid inputs).
// Moved to the top of the file so it's easier to see.
//
// DO NOT add extra external parameters (modules, zones etc) to this class. It is designed
// to be a complete black-box taking only position/bounding volume parameters and returning 
// the visibility status. Any data required to cache previous results/values etc should
// be held internally in the class and completely hidden e.g. no get/set methods, so we can
// potentially change the mechanism in the future with no side-effects.
//
class Asura_Visibility_Query
{
	public:

		Asura_Visibility_Query() : m_uLastKnownNode(0) {}

		// Is the AABB visible from the current camera? This will automatically
		// do a frustum test if required.
		inline bool IsAABBVisibleFromCurrentCamera( const Asura_Bounding_Box& xAABB ) const;

		// Is the AABB visible from the current camera position, ignoring the camera direction?
		// i.e. if you could see in all directions from the current camera position.
		inline bool IsAABBVisibleFromCurrentCameraIgnoringDirection( const Asura_Bounding_Box& xAABB ) const;

		inline bool IsPositionVisibleFromCurrentCamera( const Asura_Vector_3& xPosition ) const;
		inline bool IsSphereVisibleFromCurrentCamera( const Asura_Vector_3& xPosition, const float fRadius ) const;

	
	private:

		mutable u_int m_uLastKnownNode;
};

class Asura_Static_Visibility_Query
{
	public:

		Asura_Static_Visibility_Query() : m_uLastKnownNode(uASURA_VISIBILITY_INVALID_NODE_INDEX) {}

		// Is the AABB visible from the current camera? This will automatically
		// do a frustum test if required.
		inline bool IsAABBVisibleFromCurrentCamera( const Asura_Bounding_Box& xAABB ) const;

		// Is the AABB visible from the current camera position, ignoring the camera direction?
		// i.e. if you could see in all directions from the current camera position.
		inline bool IsAABBVisibleFromCurrentCameraIgnoringDirection( const Asura_Bounding_Box& xAABB ) const;

		inline bool IsSphereVisibleFromCurrentCamera( const Asura_Vector_3& xPosition, const float fRadius ) const;

	private:

		mutable u_int m_uLastKnownNode;
};


class Asura_Visibility_Node;
class Asura_Visibility_Leaf;
class Asura_Visibility_NodeFlags;

class Asura_Visibility_System
{
	public:

		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

		static void Update( void );

		// See top of file for the Asura_Visibility_Query class designed as a helper for visibility queries.

		// Is AABB visible from the current camera?
		static bool IsAABBVisibleFromCurrentCamera( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode );
		
		// Is AABB visible from the cameras whose bits are set in uCameraFlags? Defaults to all cameras,
		// effectively asking if the AABB is visible from any camera.
		static bool IsAABBVisible( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode, u_char ucCameraFlags = 0xff, bool bUpdateNode = true );

		// Is AABB visible from the cameras whose bits are set in uCameraFlags, ignoring direction/frusta tests? 
		// Defaults to all cameras, effectively asking if the AABB is visible from any camera position, ignoring frustum culling.
		static bool IsAABBVisibleIgnoringCameraDirection( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode, u_char ucCameraFlags = 0xff, bool bUpdateNode = true );


		static bool IsModuleLinkedInFromCurrentCamera( u_int uModuleIndex );

		static bool IsSkyboxVisibleFromCurrentCamera( void );

		static void FindNodeContainingPoint( const Asura_Vector_3& xPoint, u_int& uLastKnownNode );
		
		// returns true if AABB is wholly contained by root viewcell 
		static bool FindNodeContainingAABB( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode );
		static bool FindNodeContainingSphere( const Asura_Vector_3& xCentre, float fRadius, u_int& uLastKnownNode );

		// This method provides a bridge to the old-style module-system setting of a RenderList array.
		static void GetVisibleFeaturesFromCamera( u_char* pucVisibleFeaturesArray, u_int uNumberOfFeatures );

		static u_int GetNumberOfVisibilityFeatures( void ) { return s_uNumberOfVisibilityFeatures; }
		static u_int GetNumberOfVisibilityLeaves( void ) { return s_uNumberOfLeaves; }

		#ifdef ASURA_TOOLS // Data can only be changed like this in Tools builds 
		static void Tools_DeleteModule( u_int uModuleIndex );
		#endif

		static void ToggleRenderEverything( void ) { s_bRenderEverything = !s_bRenderEverything; }

		static bool ShouldRenderInXRayMode( void ) { return s_bXRayMode; }

	private:
		friend class Tools_Visibility_System;
		friend class Asura_Chunk_Visibility;

		static void CreateNodeVisibilityFlags( void );
		
		static void UpdateCameraVisibilities( void );
		static void UpdateViewIndependentVisibilties( void );

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
		static void AddConsoleCommands( void );
#endif

		static u_int					s_uNumberOfNodes;
		static Asura_Visibility_Node*	s_pxVisibilityNodes;	

		// KJFL - 18 February 2008 - Extra bitfields per node for hierarchical frustum tests
		// - uses a single bit per camera frustum
		// - better memory access if a separate array, and can be more easily zeroed etc
		// - dynamic so no need to load/save
		static Asura_Visibility_NodeFlags*	s_pxNodeFlags;

		static u_int					s_uNumberOfLeaves;
		static Asura_Visibility_Leaf*	s_pxVisibilityLeaves;

		static u_int					s_uNumberOfVisibilityFeatures;
		static u_char*					s_pucVisibleFeatures;

		static u_int					s_auLastKnownCameraNode[ASURA_MAX_NO_OF_CAMERAS];
		static Asura_Visibility_Leaf*	s_apxCameraLeaf[ASURA_MAX_NO_OF_CAMERAS];

		static bool						s_bForceVisRecalcForCurrentCameraNode;
		static bool						s_bRenderEverything; 
		static bool						s_bFreezeCameraNode;
		static bool						s_bXRayMode;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////
// inline methods from Asura_Visibilty_Query - need to be after Asura_Visibility_System is declared. //
///////////////////////////////////////////////////////////////////////////////////////////////////////

inline bool Asura_Visibility_Query::IsAABBVisibleFromCurrentCamera( const Asura_Bounding_Box& xAABB ) const
{
	return Asura_Visibility_System::IsAABBVisibleFromCurrentCamera( xAABB, m_uLastKnownNode );
}

inline bool Asura_Visibility_Query::IsAABBVisibleFromCurrentCameraIgnoringDirection( const Asura_Bounding_Box& xAABB ) const
{
	return Asura_Visibility_System::IsAABBVisibleIgnoringCameraDirection( xAABB, m_uLastKnownNode, Asura_Camera::GetCameraIndexBitFlag() );
}

inline bool Asura_Visibility_Query::IsPositionVisibleFromCurrentCamera( const Asura_Vector_3& xPosition ) const
{
	Asura_Bounding_Box	xAABB;

	// ART - 17/01/08 - We could do with a point function in "Asura_Visibility_System",
	// assuming that this is actually a valid thing to be doing...
	xAABB.SetToPoint(xPosition);
	return Asura_Visibility_System::IsAABBVisibleFromCurrentCamera( xAABB, m_uLastKnownNode );
}

inline bool Asura_Visibility_Query::IsSphereVisibleFromCurrentCamera( const Asura_Vector_3& xPosition, const float fRadius ) const
{
	Asura_Bounding_Box	xAABB;

	xAABB.SetToPoint(xPosition);
	xAABB.Grow(fRadius);
	return Asura_Visibility_System::IsAABBVisibleFromCurrentCamera( xAABB, m_uLastKnownNode );
}

///////////////////////////////////
// Asura_Static_Visibility_Query //
///////////////////////////////////
inline bool Asura_Static_Visibility_Query::IsAABBVisibleFromCurrentCamera( const Asura_Bounding_Box& xAABB ) const
{
#ifdef ASURA_TOOLS
	bool bUpdateNode = true;
#else
	bool bUpdateNode = false;
#endif
	return Asura_Visibility_System::IsAABBVisible( xAABB, m_uLastKnownNode, Asura_Camera::GetCameraIndexBitFlag(), bUpdateNode );

}

inline bool Asura_Static_Visibility_Query::IsAABBVisibleFromCurrentCameraIgnoringDirection( const Asura_Bounding_Box& xAABB ) const
{
#ifdef ASURA_TOOLS
	bool bUpdateNode = true;
#else
	bool bUpdateNode = false;
#endif
	return Asura_Visibility_System::IsAABBVisibleIgnoringCameraDirection( xAABB, m_uLastKnownNode, Asura_Camera::GetCameraIndexBitFlag(), bUpdateNode );
}


inline bool Asura_Static_Visibility_Query::IsSphereVisibleFromCurrentCamera( const Asura_Vector_3& xPosition, const float fRadius ) const
{
#ifdef ASURA_TOOLS
	bool bUpdateNode = true;
#else
	bool bUpdateNode = false;
#endif
	Asura_Bounding_Box	xAABB;
	xAABB.SetToPoint(xPosition);
	xAABB.Grow(fRadius);

	return Asura_Visibility_System::IsAABBVisible( xAABB, m_uLastKnownNode, Asura_Camera::GetCameraIndexBitFlag(), bUpdateNode );
}


#endif // #ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM

#if defined(ASURA_USE_VISIBILITY_SYSTEM)

class Asura_Visibility_Node;
class Asura_Bounding_Box;
class Asura_Vector_3;

class Asura_Visibility_System
{
	public:

		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );


		static void Update( void );

		static bool IsEnvironmentStripVisible( u_int uStripID );
		static const Asura_Bounding_Box& GetStripBB( u_int uStripID );

		static u_int GetNumberOfVisibleStrips() { return s_uNumberOfVisibleStrips; }
		static u_int GetVisibleStrip( const u_int uVisibleStripIndex );

#ifdef ASURA_TOOLS
		static void Tools_Create( const u_int uNumberOfNodes, Asura_Bounding_Box* pxNodeBBs, const Asura_Vector_3* pxNodeCentres );
#endif

	private: friend class Asura_Chunk_Visibility;

		static void SortVisibleStripsOnCameraDistance( void );
		static void SortVisibleStripsOnIndexOrder( void );

		static int NodeSortOnCameraDistance( const void *pNode1, const void *pNode2 );
		static int NodeSortOnIndexOrder( const void *pNode1, const void *pNode2 );

		static void Platform_RenderStripZOnly( u_int uStripID );
		//static void Platform_RenderZPassOfVisibleStrips( void );

		static void Platform_IssueOcclusionQuery( u_int uNodeIndex, const Asura_Bounding_Box& xBB );

		static void Platform_IssueOcclusionQuery( u_int uNodeIndex, u_int uStripID );

		static u_int Platform_GetNumberOfOutstandingOcclusionQueries( void );

		static bool Platform_GetOcclusionResults( u_int& uNodeIndex, bool bForceFlush  );


		static u_int					s_uNumberOfNodes;
		static Asura_Visibility_Node*	s_pxNodes;

		static u_int 					s_uLastFrameStamp;
		static u_int 					s_uFrameStamp;

		static u_int					s_uNumberOfStrips;
		static u_int					s_uNumberOfVisibleStrips;
		static u_int*					s_puVisibleStrips;
};
#endif //defined(ASURA_USE_VISIBILITY_SYSTEM)

#endif//__ASURA_VISIBILITY_SYSTEM_H__

