//////////////////////////////////////////////////////////////////////
// Asura_VolumeGroup												//
// SJL 6/11/07														//
// Copyright (C) 2007 Rebellion, All Rights Reserved.			    //
//																	//
// A container class for a collection of Asura_Volume objects    	//
//////////////////////////////////////////////////////////////////////

// Main asura file
#include "Core/Asura.h"

// This file's header.
#include "Volume/Asura_VolumeGroup.h"

// Asura headers
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_LBAABB_Tree.h"

// Volume types
#include "Volume/Asura_Volume.h"
#include "Volume/Asura_Volume_Cuboid.h"
#include "Volume/Asura_Volume_Sphere.h"
#include "Volume/Asura_Volume_Cylinder.h"

// Static class members
const u_int Asura_VolumeGroup::s_uCHUNKVERSION = 1;
Asura_VolumeGroup_Test_Parameters Asura_VolumeGroup::s_xTestParams;

Asura_VolumeGroup::Asura_VolumeGroup()
: m_uNumberOfVolumes(0)
, m_ppxVolumes(NULL)
, m_pxLBAABBTree(NULL)
{
}

Asura_VolumeGroup::~Asura_VolumeGroup()
{
	Clear();
}

void Asura_VolumeGroup::Clear()
{
   if (m_ppxVolumes)
   {
		for (u_int i=0 ; i< m_uNumberOfVolumes ; ++i)
		{
			if (m_ppxVolumes[i]) delete m_ppxVolumes[i];
		}
		Asura_Deallocate(m_ppxVolumes);
		m_ppxVolumes = NULL;
   }
   m_uNumberOfVolumes = 0;
   RemoveLBAABBTree();
}

void Asura_VolumeGroup::DisplaceVolumes(const Asura_Vector_3& xOffset)
{
	for (u_int i=0; i<m_uNumberOfVolumes ; ++i)
	{
		if (m_ppxVolumes[i]) m_ppxVolumes[i]->Displace(xOffset);
	}
}
void Asura_VolumeGroup::RotateVolumesAbout(const Asura_Vector_3& xCentre, const Asura_Matrix_3x3& xOri)
{
	for (u_int i=0; i<m_uNumberOfVolumes ; ++i)
	{
		if (m_ppxVolumes[i]) m_ppxVolumes[i]->RotateAbout(xCentre, xOri);
	}
}


#if !defined(ASURA_NO_DEBUG_OUTPUT)
void Asura_VolumeGroup::AddToLineSet(const u_int uARGB,
									 Asura_DebugLineRender_LineSet& xLineSet) const
{
	for(u_int u = 0; u < m_uNumberOfVolumes; u++)
	{
		Asura_Volume*	pxVolume	= m_ppxVolumes[u];

		if(pxVolume != NULL)
		{
			pxVolume->AddToLineSet(uARGB, xLineSet);
		}
	}
}
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)

#ifdef ASURA_TOOLS
void Asura_VolumeGroup::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	for (u_int i=0; i<m_uNumberOfVolumes ; ++i)
	{
		if (m_ppxVolumes[i]) m_ppxVolumes[i]->ApplyChangesForTemplateInstance(pxInstance);
	}
}
#endif //ASURA_TOOLS

bool Asura_VolumeGroup::Contains(const Asura_Vector_3& xPosition) const
{
	if(m_pxLBAABBTree)
	{
		s_xTestParams.m_pxGroup = this;
		s_xTestParams.m_pxPosition = &xPosition;
		s_xTestParams.m_bResult = false;
		s_xTestParams.m_pfnCallback = NULL;
		m_pxLBAABBTree->TestTreeWithPoint(xPosition, LBAABBContainsCallback);
		// Not necessary, but best to be safe.
		s_xTestParams.m_pxGroup = NULL;
		return s_xTestParams.m_bResult;
	}
	for (u_int i=0; i<m_uNumberOfVolumes ; ++i)
	{
		if (m_ppxVolumes[i] && m_ppxVolumes[i]->Contains(xPosition)) return true;
	}
	return false;
}

bool Asura_VolumeGroup::Contains(const Asura_Vector_3& xPosition, Asura_VolumeGroup_Contains_Callback pfnCallback) const
{
	Asura_Assert(pfnCallback, "Must provide a callback");
	if(m_pxLBAABBTree)
	{
		s_xTestParams.m_pxGroup = this;
		s_xTestParams.m_pxPosition = &xPosition;
		s_xTestParams.m_bResult = false;
		s_xTestParams.m_pfnCallback = pfnCallback;
		m_pxLBAABBTree->TestTreeWithPoint(xPosition, LBAABBContainsCallback);
		// Not necessary, but best to be safe.
		s_xTestParams.m_pxGroup = NULL;
		return s_xTestParams.m_bResult;
	}
	bool bContains = false;
	for (u_int u = 0; u < m_uNumberOfVolumes ; ++u)
	{
		if (m_ppxVolumes[u] && m_ppxVolumes[u]->Contains(xPosition))
		{
			// If we have a callback, we need to evaluate every volume.
			pfnCallback(u);
			bContains = true;
		}
	}
	return bContains;
}

bool Asura_VolumeGroup::Overlaps(const Asura_Bounding_Box& xBB) const
{
	if(m_pxLBAABBTree)
	{
		s_xTestParams.m_pxGroup = this;
		s_xTestParams.m_pxBB = &xBB;
		s_xTestParams.m_bResult = false;
		m_pxLBAABBTree->TestTreeWithBB(xBB, LBAABBOverlapsCallback);
		// Not necessary, but best to be safe.
		s_xTestParams.m_pxGroup = NULL;
		return s_xTestParams.m_bResult;
	}
	for (u_int i=0; i<m_uNumberOfVolumes ; ++i)
	{
		if (m_ppxVolumes[i] && m_ppxVolumes[i]->Overlaps(xBB)) return true;
	}
	return false;
}


Asura_Quat* Asura_VolumeGroup::GetVolumeOrient(u_int uIndex)
{
	Asura_Assert(uIndex < m_uNumberOfVolumes, "GetVolumeOrient, index out of range");
	Asura_Assert(m_ppxVolumes[uIndex], "GetVolumeOrient, no orient to get");
	if (m_ppxVolumes[uIndex] && uIndex < m_uNumberOfVolumes) 
	{
		return m_ppxVolumes[uIndex]->GetVolumeOrient();
	} 
	else
	{
		return NULL;
	}
}

void Asura_VolumeGroup::SetCapacity(u_int uNumberOfVolumes)
{
	// This first assert is unnecessary, as if you copy a volume trigger in tools without ever editing the volume group of the original, it won't have any volumes yet.
//	Asura_Assert(uNumberOfVolumes, "Asura_VolumeGroup::SetCapacity, argument should be greater than 0");
	Asura_Assert(!m_uNumberOfVolumes, "Asura_VolumeGroup::SetCapactiy can only be called on an empty group");
	Asura_Assert(!m_ppxVolumes, "Asura_VolumeGroup::SetCapacity can only be called on an empty group");
	if ( (!uNumberOfVolumes) || m_uNumberOfVolumes || m_ppxVolumes) return;

	// allocate array
	m_ppxVolumes = (Asura_Volume**)(Asura_Allocate(uNumberOfVolumes* sizeof(Asura_Volume*)));
	// set end marker
	m_uNumberOfVolumes = uNumberOfVolumes;
	// zero pointers
	u_int i;
	for (i=0; i<m_uNumberOfVolumes ; i++)
	{
		m_ppxVolumes[i] = NULL;
	}
}

void Asura_VolumeGroup::AddVolume(Asura_Volume* pxVolume)
{
	Asura_Assert(pxVolume, "Warning: adding a NULL volume");
	// See if theres already space for a volume in the array
	for (u_int i=0 ; i< m_uNumberOfVolumes ; i++)
	{
		// found an empty entry
		if (!m_ppxVolumes[i])
		{
			m_ppxVolumes[i] = pxVolume;
			return;
		}
	}

	// We will need to grow the array by 1
	// Calculate new array size and allocate storage
	u_int uNewSize = m_uNumberOfVolumes + 1;
	Asura_Volume ** ppxNewVolumes = (Asura_Volume**)(Asura_Allocate(uNewSize * sizeof(Asura_Volume*)));

	// Copy pointers into new array
	for (u_int i=0 ; i< m_uNumberOfVolumes ; ++i)
	{
		ppxNewVolumes[i] = m_ppxVolumes[i];
	}

	// Free old array
	if (m_ppxVolumes) Asura_Deallocate(m_ppxVolumes);

	// and swap new and old array pointers
	m_ppxVolumes = ppxNewVolumes;

	// and add the new element / increase count
	m_ppxVolumes[m_uNumberOfVolumes] = pxVolume;
	++m_uNumberOfVolumes;
}

void Asura_VolumeGroup::RemoveAndDestroyVolume(Asura_Volume* pxVolume)
{
	Asura_Assert(m_uNumberOfVolumes, "Trying to remove a volume from a volume group, but none currently exist!");
	Asura_Assert(pxVolume, "Trying to remove a NULL volume, thats kind of pointless");
	if (!pxVolume) return;

	// Find the volume passed in.
	for (u_int i=0 ; i<m_uNumberOfVolumes ; ++i)
	{
		if ( m_ppxVolumes[i] == pxVolume)
		{
			RemoveAndDestroyVolume(i);
			return;
		}
	}
	Asura_Assert(0, "RemoveAndDestroyVolue: Volume not found");
}

void Asura_VolumeGroup::RemoveAndDestroyVolume(u_int uIndex)
{
	Asura_Assert(uIndex < m_uNumberOfVolumes, "");

	// Delete volume
	delete m_ppxVolumes[uIndex];
	m_ppxVolumes[uIndex] = NULL;

	// Calculate new array size and allocate storage
	u_int uNewSize = m_uNumberOfVolumes - 1;
	Asura_Volume ** ppxNewVolumes = (Asura_Volume**)(Asura_Allocate(uNewSize * sizeof(Asura_Volume*)));

	// Copy elements before 'uIndex'
	for (u_int i=0 ; i<uIndex ; ++i)
	{
		ppxNewVolumes[i] = m_ppxVolumes[i];
	}
	// Copy elements after 'i'
	for (u_int i=uIndex ; i<uNewSize; ++i)
	{
		ppxNewVolumes[i] = m_ppxVolumes[i+1];
	}

	// Free old array
	Asura_Deallocate(m_ppxVolumes);

	// and swap new and old array pointers
	m_ppxVolumes = ppxNewVolumes;

	// and decrease the count
	--m_uNumberOfVolumes;
}

void Asura_VolumeGroup::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	Clear();

	// Read and check version number
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert(uVersion <= s_uCHUNKVERSION, "Unknown chunk version loading Asura_VolumeGroup");
	if (uVersion > s_uCHUNKVERSION )
	{
		return;
	}
	u_int uChunkFlags = 0;
	if(uVersion >= 1)
	{
		xStream >> uChunkFlags;
	}

	// Read VolumeGroup data
	u_int uNumberOfVolumes;
	xStream >> uNumberOfVolumes;

	// Allocate space for these volumes in the array
	SetCapacity(uNumberOfVolumes);

	// Read Volumes
	for (u_int i=0; i<uNumberOfVolumes ; i++)
	{
		int iClassification;
		xStream >> iClassification;

		Asura_Assert( (iClassification >= 0) && (iClassification < ASURA_VOLUME_CLASSIFICATION_NUMBEROFCLASSES), "Asura_VolumeGroup::Read from Chunk Stream, unknown volume type");
		if ( (iClassification < 0) || (iClassification >= ASURA_VOLUME_CLASSIFICATION_NUMBEROFCLASSES) ) return;

		m_ppxVolumes[i] = NULL;

		switch (iClassification)
		{
			case ASURA_VOLUME_CLASSIFICATION_CUBOID:
				m_ppxVolumes[i] = new Asura_Volume_Cuboid();
				break;
			case ASURA_VOLUME_CLASSIFICATION_SPHERE:
				m_ppxVolumes[i] = new Asura_Volume_Sphere();
				break;
			case ASURA_VOLUME_CLASSIFICATION_CYLINDER:
				m_ppxVolumes[i] = new Asura_Volume_Cylinder();
				break;
			case ASURA_VOLUME_CLASSIFICATION_NULL: // drop through
			default:
				{
				Asura_Assert(0, "Bad volume classifiction, did you set it in your constructor?");
				}
				break;
		}

		if (m_ppxVolumes[i]) m_ppxVolumes[i]->ReadFromChunkStream(xStream);
	}

	if(uChunkFlags & s_uCHUNK_FLAG_HAS_LBAABB)
	{
		m_pxLBAABBTree = new Asura_LBAABB_Tree();
		m_pxLBAABBTree->ReadTreeAndNodesFromChunkStream(xStream);
	}
}

void Asura_VolumeGroup::WriteToChunkStream (Asura_Chunk_Stream& xStream, bool bAddLBAABBTree /* = false */, Asura_VolumeGroup_Reorder_Callback pfnReorderCallback /* = NULL */) const
{
#ifdef ASURA_TOOLS
	if(bAddLBAABBTree)
	{
		// Can't add the tree to ourselves since we're const. Instead, create a copy.
		Asura_VolumeGroup xCopy;
		xCopy.Copy(this);
		xCopy.GenerateLBAABBTree(pfnReorderCallback);
		xCopy.WriteToChunkStream(xStream);
		return;
	}
#else
	Asura_Assert(!bAddLBAABBTree, "Cannot add LBAABB tree to volume group outside tools");
#endif
	// Write version number
	xStream << s_uCHUNKVERSION;
	const u_int uChunkFlags = m_pxLBAABBTree ? s_uCHUNK_FLAG_HAS_LBAABB : 0;
	xStream << uChunkFlags;
	// Write VolumeGroup data
	xStream << m_uNumberOfVolumes;	
	// Write Volumes
	for (u_int i=0 ; i<m_uNumberOfVolumes ; i++)
	{
		if (m_ppxVolumes[i])
		{
			m_ppxVolumes[i]->WriteToChunkStream(xStream);
		} 
		else
		{
			xStream << ASURA_VOLUME_CLASSIFICATION_NULL;
		}
	}
	if(m_pxLBAABBTree)
	{
		m_pxLBAABBTree->WriteTreeAndNodesToChunkStream(xStream);
	}
}

void Asura_VolumeGroup::ReadOrientedVolumeArrayFromChunkStream(Asura_Chunk_Stream& xStream, u_int uNumberOfVolumes)
{

	// Allocate space for these volumes in the array
	SetCapacity(uNumberOfVolumes);

	// Read Volumes
	for (u_int i=0; i<uNumberOfVolumes ; i++)
	{
		Asura_Bounding_Box xBox;
		Asura_Quat xOrient;
		xStream >> xBox;
		xStream >> xOrient;

		m_ppxVolumes[i] = new Asura_Volume_Cuboid(xBox, xOrient);
	}

}


void Asura_VolumeGroup::Copy(const Asura_VolumeGroup* pxFrom)
{
	// Remove old array.
	if (m_ppxVolumes)
	{
		for (u_int i=0 ; i< m_uNumberOfVolumes ; ++i)
		{
			if (m_ppxVolumes[i]) delete m_ppxVolumes[i];
		}
		Asura_Deallocate(m_ppxVolumes);
	}
	m_ppxVolumes = 0;
	m_uNumberOfVolumes = 0;
	// Set it up again
	SetCapacity(pxFrom->GetNumberOfVolumes());

	// Copy the volumes over
	for (u_int i=0; i< pxFrom->GetNumberOfVolumes() ; ++i)
	{
		// Construct and copy volume
		if (pxFrom->m_ppxVolumes[i]) 
		{
			m_ppxVolumes[i] = pxFrom->m_ppxVolumes[i]->Duplicate();
			// Assert the classification is correct, if not this may imply that
			// the duplicate or copy functions have not been created properly.
			Asura_Assert(m_ppxVolumes[i]->GetClassification() == pxFrom->m_ppxVolumes[i]->GetClassification(), "Copying volumes, classification mis-match, check code");
		}
	}


}

void Asura_VolumeGroup::LBAABBContainsCallback(const u_int uIndex)
{
	// If we already have a positive result, there's no point testing further volumes.
	if(s_xTestParams.m_bResult && !s_xTestParams.m_pfnCallback) { return; }
	Asura_Assert(s_xTestParams.m_pxGroup && s_xTestParams.m_pxPosition, "");
	const Asura_Volume* pxVolume = s_xTestParams.m_pxGroup->GetVolume(uIndex);
	if(pxVolume->Contains(*s_xTestParams.m_pxPosition))
	{
		if(s_xTestParams.m_pfnCallback)
		{
			s_xTestParams.m_pfnCallback(uIndex);
		}
		s_xTestParams.m_bResult = true;
	}
}

void Asura_VolumeGroup::LBAABBOverlapsCallback(const u_int uIndex)
{
	// If we already have a positive result, there's no point testing further volumes.
	if(s_xTestParams.m_bResult) { return; }
	Asura_Assert(s_xTestParams.m_pxGroup && s_xTestParams.m_pxBB, "");
	const Asura_Volume* pxVolume = s_xTestParams.m_pxGroup->GetVolume(uIndex);
	s_xTestParams.m_bResult = pxVolume->Overlaps(*s_xTestParams.m_pxBB);
}

#ifdef ASURA_TOOLS
void Asura_VolumeGroup::GenerateLBAABBTree(Asura_VolumeGroup_Reorder_Callback pfnReorderCallback /* = NULL */)
{
	if(!m_uNumberOfVolumes)
	{
		RemoveLBAABBTree();
		return;
	}
	if(m_pxLBAABBTree)
	{
		m_pxLBAABBTree->DeleteNodes();
	}
	else
	{
		m_pxLBAABBTree = new Asura_LBAABB_Tree();
	}
	// Generate array of BBs and centres.
	Asura_Bounding_Box* pxBBs = new Asura_Bounding_Box[m_uNumberOfVolumes];
	Asura_Vector_3* pxCentres = new Asura_Vector_3[m_uNumberOfVolumes];
	Asura_Assert(pxBBs && pxCentres, "Out of memory?");

	for(u_int uVolume = 0; uVolume < m_uNumberOfVolumes; ++uVolume)
	{
		m_ppxVolumes[uVolume]->GetAABB(pxBBs[uVolume]);
		pxCentres[uVolume] = pxBBs[uVolume].CalculateCentre();
	}
	const u_int* puTranslationArray = Tools_LBAABB_Tree::CreateRunTimeTree(*m_pxLBAABBTree, m_uNumberOfVolumes, pxBBs, pxCentres);
	// Re-order volumes using the translation array
	Asura_Volume** ppxNewVolumeArray = (Asura_Volume**)(Asura_Allocate(m_uNumberOfVolumes * sizeof(Asura_Volume*)));
	for(u_int uVolume = 0; uVolume < m_uNumberOfVolumes; ++uVolume)
	{
		ppxNewVolumeArray[uVolume] = m_ppxVolumes[puTranslationArray[uVolume]];
	}
	Asura_Deallocate(m_ppxVolumes);
	m_ppxVolumes = ppxNewVolumeArray;
	if(pfnReorderCallback)
	{
		pfnReorderCallback(puTranslationArray);
	}
	delete [] pxBBs;
	delete [] pxCentres;
	delete [] puTranslationArray;
}
#endif
void Asura_VolumeGroup::RemoveLBAABBTree()
{
	if(m_pxLBAABBTree)
	{
		m_pxLBAABBTree->DeleteNodes();
	}
	delete m_pxLBAABBTree;
	m_pxLBAABBTree = NULL;
}


void Asura_VolumeGroup::GetEncompassingAABB( Asura_Bounding_Box& xEncompassingBB ) const
{
    xEncompassingBB.Reset();

    Asura_Bounding_Box xTempBB;
	for( u_int u = 0; u < m_uNumberOfVolumes; ++u )
	{
        xTempBB.Reset();
        m_ppxVolumes[u]->GetAABB( xTempBB );
        xEncompassingBB.Set( xTempBB );
    }
}
