// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_PC_SUPPORT_STEAMWORKS

#include "Debugging/Asura_Logfile.h"						// Asura
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_NetTransport_System.h"
#include "Network/Steamworks/Asura_PC_NetSocketSteamworks.h"
#include "Network/Steamworks/Asura_PC_Steamworks.h"
#include "Network/Steamworks/Asura_PC_SteamworksClient.h"
#include "Network/Steamworks/Asura_PC_SteamworksServer.h"
#include "Core/Asura_Memheap.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_Timers.h"
//#include "Sound/Asura_XAudio2.h"

#include "steam/steam_api.h"								// Steam

// D E F I N E S /////////////////////////////////////////////////////////////

#define ASURA_PC_STEAM_MAX_SERVER_SEARCH_RESULTS		1024
#define ASURA_PC_STEAM_SERVER_SEARCH_TIMEOUT			10.0f
#define ASURA_PC_STEAM_SERVER_SEARCH_UPDATE_INTERVAL	0.25f

// G L O B A L   V A R I A B L E S ///////////////////////////////////////////

static	u_int	g_uFrameNumber = 0;

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////

// ____________________ Asura_PC_Steamworks_ServerList _________________________

class Asura_PC_Steamworks_ServerList : public ISteamMatchmakingServerListResponse
{
	// This class handles server browsing functionality
	public:

		void ServerResponded( int iServer );
		void ServerFailedToRespond( int iServer );

		void RefreshComplete( EMatchMakingServerResponse response ); 
};

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

bool								Asura_PC_Steamworks::s_bSteamworksEnabled		= false;

bool								Asura_PC_Steamworks::s_bIsInitialised			= false;

Asura_PC_SteamServer_Data*			Asura_PC_Steamworks::s_pxServerList				= 0;
u_int								Asura_PC_Steamworks::s_uNumServers				= 0;

bool								Asura_PC_Steamworks::s_bSearching				= false;
float								Asura_PC_Steamworks::s_fSearchStartTime			= 0.0f;
ASURA_PC_STEAMWORKS_SERVER_TYPE		Asura_PC_Steamworks::s_eSearchServerType		= ASURA_PC_STEAMWORKS_SERVER_LAN;

Asura_PC_Steamworks_ServerList		Asura_PC_Steamworks::s_xServerListResponse;

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

extern "C" void __cdecl Asura_Steam_DebugTextHook( int nSeverity, const char *szDebugText )
{
	Asura_OutputToDebugger( ( "[Steam] %s ( severity %d )", szDebugText ) );
}

// _____________________ Asura_PC_Steamworks ___________________________

Asura_PC_Steamworks::Asura_PC_Steamworks ()
{
}

Asura_PC_Steamworks::~Asura_PC_Steamworks ()
{
}


void Asura_PC_Steamworks::Initialise ()
{
	if ( s_bIsInitialised )
	{
		Asura_Assert( false, "Asura_PC_Steamworks is already initialised" );
		return;
	}

	s_bIsInitialised = SteamAPI_Init();

	// TY 05.04.09 - Allow steam init to fail silently for now, so that people without steam arent bugged by asserts
	//Asura_Assert( s_bIsInitialised, "Failed to initialise Steamworks (is steam running?)" );

	if ( s_bIsInitialised )
	{
		SteamClient()->SetWarningMessageHook( &Asura_Steam_DebugTextHook );

		// Initialise the client and server classes
		Asura_PC_SteamworksClient::Initialise();
		Asura_PC_SteamworksServer::Initialise();
	}
}


void Asura_PC_Steamworks::Update ()
{
	if( g_uFrameNumber == Asura_Timers::GetFrameNumber() )
	{
		// Already updated this frame
		return;
	}
	
	g_uFrameNumber = Asura_Timers::GetFrameNumber();
	
	SteamAPI_RunCallbacks();

	if( s_bSearching )
	{
		// Handle search timing out
		if( ( Asura_Timers::GetLocalUnscaledTime() - s_fSearchStartTime ) > ASURA_PC_STEAM_SERVER_SEARCH_TIMEOUT )
		{
			s_bSearching = false;
		}
	}

	// Update the client and server classes
	Asura_PC_SteamworksClient::Update();
	Asura_PC_SteamworksServer::Update();
}


void Asura_PC_Steamworks::Shutdown ()
{
	if( s_bIsInitialised )
	{
		Asura_Assert( !Asura_PC_SteamworksServer::IsServerRunning(), "A steam server is still running, should have been shutdown by now" );
		Asura_Assert( !Asura_PC_SteamworksClient::IsConnectedToServer(), "Our steam client is still conencted to a server, should have been disconnected by now" );

		// Shutdown the client and server classes
		Asura_PC_SteamworksClient::Shutdown();
		Asura_PC_SteamworksServer::Shutdown();

		s_bSteamworksEnabled = false;

		SteamAPI_Shutdown ();
	}
}


void Asura_PC_Steamworks::BeginServerSearch ( const ASURA_PC_STEAMWORKS_SERVER_TYPE eServerType, const Asura_PC_Steamworks_ServerFilter* pxFilters, const u_int uNumFilters )
{
	if( s_bSearching )
	{
		Asura_Assert( false, "A search is already being conducted, please wait for it to finish before starting a new search" );
		return;
	}

	s_fSearchStartTime = Asura_Timers::GetLocalUnscaledTime();

	switch( s_eSearchServerType )
	{
		case ASURA_PC_STEAMWORKS_SERVER_INTERNET:
		{
			// Only internet searches allow us to specify filters
			const u_int uMaxFilters = 16;

			Asura_Assert( uNumFilters < uMaxFilters, "server searching can only take up to 16 filters" );

			MatchMakingKeyValuePair_t pxSteamFilters[16];

			for( u_int uFilter = 0; uFilter < Asura_Maths::Min( uMaxFilters, uNumFilters ); ++uFilter )
			{
				Asura_PrintfToBfr( pxSteamFilters[ uFilter ].m_szKey, "%s", ASURA_PC_STEAMWORKS_FILTER_MAX_CHARS, true, pxFilters[ uFilter ].m_szKey );
				Asura_PrintfToBfr( pxSteamFilters[ uFilter ].m_szValue, "%s", ASURA_PC_STEAMWORKS_FILTER_MAX_CHARS, true, pxFilters[ uFilter ].m_szValue );
			}

			s_eSearchServerType = eServerType;

			MatchMakingKeyValuePair_t* _pxSteamFilters = &pxSteamFilters[0];	// need a pointer to pointer

			SteamMatchmakingServers()->RequestInternetServerList( SteamUtils()->GetAppID(), &_pxSteamFilters , uNumFilters, &s_xServerListResponse );
		}
		break;

		case ASURA_PC_STEAMWORKS_SERVER_LAN:
		{
			SteamMatchmakingServers()->RequestLANServerList( SteamUtils()->GetAppID(), &s_xServerListResponse );
		}
		break;
	}

	s_pxServerList = new Asura_PC_SteamServer_Data[ ASURA_PC_STEAM_MAX_SERVER_SEARCH_RESULTS ];
	s_uNumServers = 0;

	s_bSearching = true;
}


bool Asura_PC_Steamworks::IsServerSearchFinished			( )
{
	return !s_bSearching;
}


const Asura_PC_SteamServer_Data* Asura_PC_Steamworks::GetServerSearchResult ( const u_int uResult )
{
	if( s_bSearching )
	{
		Asura_Assert( false, "A search is being conducted, please wait for it to finish" );
		return 0;
	}
	if( uResult >= s_uNumServers )
	{
		// Dont assert here, caller might be doing something like while( GetSearchResult() )
		return 0;
	}

	return &s_pxServerList[ uResult ];
}


u_int Asura_PC_Steamworks::GetNumServerSearchResults ( )
{
	if( s_bSearching )
	{
		Asura_Assert( false, "A search is being conducted, please wait for it to finish" );
		return 0;
	}
	return s_uNumServers;
}


void Asura_PC_Steamworks::ClearServerSearch ( )
{
	// Clean up our search data
	if( s_pxServerList )
	{
		delete []s_pxServerList;
		s_pxServerList = 0;
	}
	s_uNumServers = 0;
	s_bSearching = false;
}

void Asura_PC_Steamworks::EnableSteamworks ()
{
	if(		Asura_Network::IsNetworkInitialised() 
		||	Asura_NetTransport_System::IsInitialised() )
	{
		Asura_Assert( false, "Some Asura networking is initialised, you cannot hot-swap networking protocols" );
		return;
	}

	if( !Asura_PC_Steamworks::IsInitialised() )
	{
		Asura_PC_Steamworks::Initialise();
	}
	
	if( Asura_PC_Steamworks::IsInitialised() )
	{
		s_bSteamworksEnabled = true;
	}
}

void Asura_PC_Steamworks::DisableSteamworks ()
{
	if(		Asura_Network::IsNetworkInitialised()
		||	Asura_NetTransport_System::IsInitialised() )
	{
		Asura_Assert( false, "Some Asura networking is initialised, you cannot hot-swap networking protocols" );
		return;
	}
	if(		Asura_PC_SteamworksClient::IsConnectedToServer() 
		||	Asura_PC_SteamworksServer::IsServerRunning() )
	{
		Asura_Assert( false, "Steamworks client/server is active, you must terminate client connections and servers before disabling steamworks" );
		return;
	}

	if( Asura_PC_Steamworks::IsInitialised() )
	{
		Asura_PC_Steamworks::Shutdown();
	}

	s_bSteamworksEnabled = false;
}


void Asura_PC_Steamworks::ShowSteamOverlay ( const ASURA_PC_STEAMWORKS_OVERLAY eOverlay )
{
	if( !s_bSteamworksEnabled )
	{
		Asura_Assert( false, "Steamworks not enabled" );
		return;
	}

	if( SteamFriends() )
	{
		switch( eOverlay )
		{
			case ASURA_PC_STEAMWORKS_OVERLAY_FRIENDS:		SteamFriends()->ActivateGameOverlay( "Friends"   );		break;
			case ASURA_PC_STEAMWORKS_OVERLAY_COMMUNITY:		SteamFriends()->ActivateGameOverlay( "Community" );		break;
			case ASURA_PC_STEAMWORKS_OVERLAY_PLAYERS:		SteamFriends()->ActivateGameOverlay( "Players"   );		break;
			case ASURA_PC_STEAMWORKS_OVERLAY_SETTINGS:		SteamFriends()->ActivateGameOverlay( "Settings"  );		break;
		}
	}
}


void Asura_PC_Steamworks::RegisterConsoleCmds ( )
{
	#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Enable"			, Asura_PC_Steamworks::ConsoleCmd_EnableSteamworks		, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Disable"		, Asura_PC_Steamworks::ConsoleCmd_DisableSteamworks		, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Status"			, Asura_PC_Steamworks::ConsoleCmd_SteamworksStatus		, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Voice.Enable"	, Asura_PC_Steamworks::ConsoleCmd_StartVoiceRecording	, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Voice.Disable"	, Asura_PC_Steamworks::ConsoleCmd_StopVoiceRecording	, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif
}

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

bool Asura_PC_Steamworks::ConsoleCmd_EnableSteamworks( bool bHelp )
{
	if (!bHelp)
	{
		Asura_PC_Steamworks::EnableSteamworks();
	}
	return true;
}


bool Asura_PC_Steamworks::ConsoleCmd_DisableSteamworks( bool bHelp )
{
	if (!bHelp)
	{
		Asura_PC_Steamworks::DisableSteamworks();
	}
	return true;
}


bool Asura_PC_Steamworks::ConsoleCmd_SteamworksStatus( bool bHelp )
{
	if (!bHelp)
	{
		if( Asura_PC_Steamworks::IsSteamworksEnabled() )
		{
			Asura_CommandConsole::Printf( L"Steamworks is enabled\n" );
			
			if( Asura_PC_SteamworksClient::IsConnectedToServer() )
			{
				Asura_CommandConsole::Printf( L"Steamworks client is connected to a server\n" );
			}
			else
			{
				Asura_CommandConsole::Printf( L"Steamworks client is not connected to a server\n" );
			}
			
			if( Asura_PC_SteamworksServer::IsServerRunning() )
			{
				Asura_CommandConsole::Printf( L"Steamworks server is running\n" );
			}
			else
			{
				Asura_CommandConsole::Printf( L"Steamworks server is not running\n" );
			}
		}
		else
		{
			Asura_CommandConsole::Printf( L"Steamworks is not enabled\n" );
		}
	}
	return true;
}

bool Asura_PC_Steamworks::ConsoleCmd_StartVoiceRecording ( bool bHelp )
{
	if (!bHelp)
	{
		if( Asura_PC_Steamworks::IsSteamworksEnabled() )
		{
			Asura_PC_SteamworksClient::StartVoiceRecording();
		}
	}
	return true;
}

bool Asura_PC_Steamworks::ConsoleCmd_StopVoiceRecording ( bool bHelp )
{
	if (!bHelp)
	{
		if( Asura_PC_Steamworks::IsSteamworksEnabled() )
		{
			Asura_PC_SteamworksClient::StopVoiceRecording();
		}
	}
	return true;
}

#endif


//_____________________________ Asura_PC_Steamworks_ServerList __________________________________


void Asura_PC_Steamworks_ServerList::ServerResponded( int iServer )
{
	if( Asura_PC_Steamworks::IsServerSearchFinished() )
	{
		return;
	}

	if( Asura_PC_Steamworks::s_uNumServers == ASURA_PC_STEAM_MAX_SERVER_SEARCH_RESULTS )
	{
		// We cant store any more server details
		return;
	}

	gameserveritem_t *pxServer = 0;

	switch( Asura_PC_Steamworks::s_eSearchServerType )
	{
		case ASURA_PC_STEAMWORKS_SERVER_LAN:		pxServer = SteamMatchmakingServers()->GetServerDetails( eLANServer		, iServer );	break;
		case ASURA_PC_STEAMWORKS_SERVER_INTERNET:	pxServer = SteamMatchmakingServers()->GetServerDetails( eInternetServer	, iServer );	break;
	}

	if ( pxServer )
	{
		// Filter out servers that don't match our appid here (might get these in LAN calls since we can't put more filters on it)
		if ( pxServer->m_nAppID == static_cast<int>( SteamUtils()->GetAppID() ) )
		{
			Asura_PC_SteamServer_Data& xEntry = Asura_PC_Steamworks::s_pxServerList[ Asura_PC_Steamworks::s_uNumServers ];

			xEntry.m_uAddress	= pxServer->m_NetAdr.GetIP();
			xEntry.m_usPort		= pxServer->m_NetAdr.GetConnectionPort();

			xEntry.m_ucNumPlayers = static_cast<unsigned char>( pxServer->m_nPlayers );
			xEntry.m_ucMaxPlayers = static_cast<unsigned char>( pxServer->m_nMaxPlayers );

			++Asura_PC_Steamworks::s_uNumServers;
		}
	}
}


void Asura_PC_Steamworks_ServerList::ServerFailedToRespond( int iServer )
{
	// not really a very useful callback
}


void Asura_PC_Steamworks_ServerList::RefreshComplete( EMatchMakingServerResponse response )
{
	if( Asura_PC_Steamworks::IsServerSearchFinished() )
	{
		return;
	}

	if( Asura_PC_Steamworks::s_bSearching )
	{
		Asura_PC_Steamworks::s_bSearching = false;
	}
}

#endif	// ASURA_PC_SUPPORT_STEAMWORKS

//eof
