//////////////////////////////////////////////////////////////////////////////
//
// Asura_Bink.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Bink/Asura_Bink.h"

// Engine platform-specific.

// Asura_Direct3D.h needs to be included before Bink.h
// Bink defines f32 which clashes with DirectX includes
#include "Rendering/Asura_Direct3D.h"
#include "BinkPC/bink.h"

#include "Camera/Asura_Camera.h"
#include "Core/Asura_Windows.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#include "Sound/Asura_PC_Sound.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

#if defined( _WIN64)
	#undef BinkSoundUseDirectSound
	#define BinkSoundUseDirectSound( _A)

	#undef BinkSoundUseWaveOut
	#define BinkSoundUseWaveOut()
	
	#undef BinkGoto
	#define BinkGoto( _A, _B, _C)
	
	#undef BinkWait
	#define BinkWait( _A) (false)
	
	#undef BinkDoFrame
	#define BinkDoFrame( _A)
	
	#undef BinkNextFrame
	#define BinkNextFrame( _A)
	
	#undef BinkOpen
	#define BinkOpen( _A, _B) (NULL)
	
	#undef BinkClose
	#define BinkClose( _A)
	
	#undef BinkPause
	#define BinkPause( _A, _B)
	
	#undef BinkSetVolume
	#define BinkSetVolume( _A, _B, _C)
	
	#undef BinkShouldSkip
	#define BinkShouldSkip( _A) (false)
	
	#undef BinkSetSoundTrack
	#define BinkSetSoundTrack( _A, _B)
	
	#undef BinkCopyToBuffer
	#define BinkCopyToBuffer( _1, _2, _3, _4, _5, _6, _7)
#endif

// Play all FMVs at 75% volume to match the rest of the sound system.
// Volume 0-1 goes 0-32768 under Bink
#if defined( ASURA_USE_AUDIO_SYSTEM)
	const S32 iBINK_VOLUME = 32768;
#else
	const S32 iBINK_VOLUME = static_cast< S32 >( ASURA_SOUND_GLOBAL_VOLUME_SCALE * 32768.0f );
#endif

// G L O B A L S /////////////////////////////////////////////////////////////

static HBINK                  g_pxBink         = NULL;
static Direct3D_Texture*      g_pxTexture      = NULL;
static Direct3D_VertexBuffer* g_pxVertexBuffer = NULL;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void CleanUp( void );

static float GetPlayedTimeInSeconds( void )
{
    float fFPS = static_cast<float>(g_pxBink->FrameRate) / static_cast<float>(g_pxBink->FrameRateDiv);
    return( static_cast<float>(g_pxBink->FrameNum) / fFPS );
}

static bool RenderCurFrame( Asura_PC_Bink::RENDER_SUBTITLE_CB_FN pfnRenderSubtitleCB );

static bool SetupRendering( void );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_PC_Bink.
//

Asura_PC_Bink::PLAYRESULT Asura_PC_Bink::Play( const char* szFilename,
                                               float fTimeBeforeSkipAllowed,
                                               RENDER_SUBTITLE_CB_FN pfnRenderSubtitleCB,
                                               INPUT_CHECK_CB_FN pfnInputCheckCB,
                                               u_int uAudioStreamIdx )
{
    // Tell Bink to play the correct audio stream.
    BinkSetSoundTrack( 1, &uAudioStreamIdx );

	BinkSoundUseDirectSound(0);

    bool bError = false;

    // Open the media file. We need to specify the BINKSNDTRACK flag so that
    // it will pay attention to the sound track we've requested above.
    g_pxBink = BinkOpen( szFilename, BINKSNDTRACK );
    if( g_pxBink )
    {
        // Setup our buffers, etc.
        if( !SetupRendering() )
        {
            bError = true;
        }

		// Play all FMVs at 75% volume
		BinkSetVolume( g_pxBink, uAudioStreamIdx, iBINK_VOLUME );
    }
    else
    {
        bError = true;
    }

    bool bSkipped = false;

    // Play the movie.
    while( !bError )
    {
        Asura_Windows::CheckForWindowsMessages();

        // Deal with input.
        {
            Asura_UserInput::Update();

            // Provided callback function will return true if the user
            // has requested the remainder movie to be skipped.
            bool bWantsSkip = pfnInputCheckCB && pfnInputCheckCB();

            // However, we'll ignore that if we're not past the minimum
            // length.
            if( bWantsSkip && GetPlayedTimeInSeconds() >= fTimeBeforeSkipAllowed )
            {
                break;
            }
        }

        // Ready for next frame yet?
        if( BinkWait(g_pxBink) ) { continue; }

        BinkDoFrame( g_pxBink );

        // Check if the video is lagging behind - if so, we need to skip
        // one or more frames in order to catch up.
        if( !BinkShouldSkip(g_pxBink) )
        {
            if( !RenderCurFrame(pfnRenderSubtitleCB) )
            {
                bError = true;
                break;
            }
        }

        // Finished?
        if( g_pxBink->FrameNum == g_pxBink->Frames )
        {
            break;
        }

        BinkNextFrame( g_pxBink );
    }

    CleanUp();

    if( bError )
    {
        return PLAYRESULT_ERROR;
    }
    else if( bSkipped )
    {
        return PLAYRESULT_SKIPPED;
    }
    else
    {
        return PLAYRESULT_FINISHED;
    }
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void CleanUp( void )
{
    if( g_pxVertexBuffer )
    {
        g_pxVertexBuffer->Release();
        g_pxVertexBuffer = NULL;
    }

    if( g_pxTexture )
    {
        g_pxTexture->Release();
        g_pxTexture = NULL;
    }

    if( g_pxBink )
    {
        BinkClose( g_pxBink );
        g_pxBink = NULL;
    }
}

static bool RenderCurFrame( Asura_PC_Bink::RENDER_SUBTITLE_CB_FN pfnRenderSubtitleCB )
{
    // Copy current frame into our temporary texture.
    {
        D3DLOCKED_RECT xLockedRect;
        if( FAILED(g_pxTexture->LockRect(0, &xLockedRect, NULL, 0)) )
        {
            return false;
        }

        // (AAM - 14.02.2006) We need to use the BINKCOPYALL flag here because BinkCopyToBuffer()
		// returns immediately if the video is lagging behind. Without using the flag Bink only
		// copies pixels that it knows have changed since last frame, which corrupts the image
		// if frames had been skipped. In Rogue this resulted in bits of the Rebellion logo
		// being left on-screen after it crumbled away.
        BinkCopyToBuffer( g_pxBink, xLockedRect.pBits, xLockedRect.Pitch, g_pxBink->Height, 0, 0, BINKSURFACE32 | BINKCOPYALL );

        if( FAILED(g_pxTexture->UnlockRect(0)) )
        {
            return false;
        }
    }

    Asura_Render::BeginFrame();

    // Clear back buffer as well to deal with any bits the movie doesn't
    // cover.
    Asura_Direct3D::D3DDevice->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(0,0,0,0), 1.0f, 0 );

    Asura_Render::CurrentStates.RequestAlphaTest( false );
    Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
    Asura_Render::CurrentStates.RequestTextureOpMode( ASURA_TEXTUREOP_OPAQUE );
    Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
    Asura_Render::CurrentStates.RequestZBufferEnabled( false );
    Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

	DWORD ulOldAddressU, ulOldAddressV;
	Asura_Direct3D::D3DDevice->GetSamplerState( 0, D3DSAMP_ADDRESSU, &ulOldAddressU );
	Asura_Direct3D::D3DDevice->GetSamplerState( 0, D3DSAMP_ADDRESSV, &ulOldAddressV );

	Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
	Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );

    Asura_Direct3D::D3DDevice->SetTexture( 0, g_pxTexture );

	Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader( D3DFVF_QUADVERTEX );
    Asura_Direct3D::D3DDevice->SetStreamSource( 0, g_pxVertexBuffer, 0, sizeof(D3DQuadVertex) );
    Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );

	Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_ADDRESSU, ulOldAddressU );
	Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_ADDRESSV, ulOldAddressV );

    // Ensure any texture required for further rendering gets uploaded, since
    // we've bypassed the renderstate stuff above for that.
    Asura_Render::CurrentStates.RequestTextureIndex( 0, ASURA_TEXTURE_INVALID );

    pfnRenderSubtitleCB( GetPlayedTimeInSeconds() );

    Asura_Render::EndFrame();

    return true;
}

static bool SetupRendering( void )
{
    HRESULT iResult;

    // Create temporary texture for copying each frame into.
    iResult = Asura_Direct3D::D3DDevice->CreateTexture( g_pxBink->Width, g_pxBink->Height, 1,
                                                        0, D3DFMT_X8R8G8B8, D3DPOOL_MANAGED,
                                                        &g_pxTexture, NULL );
    if( FAILED(iResult) )
    {
        return false;
    }

    // And also a  vertex buffer with some fixed mesh for rendering with that
    // texture.
    iResult = Asura_Direct3D::D3DDevice->CreateVertexBuffer( 4 * sizeof(D3DQuadVertex),
                                                             D3DUSAGE_WRITEONLY, D3DFVF_QUADVERTEX, D3DPOOL_MANAGED,
                                                             &g_pxVertexBuffer, NULL );
    if( FAILED(iResult) )
    {
        return false;
    }

    // Fill the vertex buffer.
    {
        D3DQuadVertex* pxVertices;
        
        iResult = g_pxVertexBuffer->Lock( 0, 0, reinterpret_cast<void**>(&pxVertices), 0 );
        if( FAILED(iResult) )
        {
            return false;
        }

        // Calculate and write vertex data.
        {
            // Set common values.
            for( u_int u = 0; u < 4; ++u )
            {
                pxVertices[u].Z       = 0.0f;
                pxVertices[u].RHW     = 1000.0f;
                pxVertices[u].Diffuse = 0xFFFFFFFFu;
            }

            float fViewportWidth  = static_cast<float>( Asura_Viewport::GetWidth()  );
            float fViewportHeight = static_cast<float>( Asura_Viewport::GetHeight() );

            float fInsetX;
            float fInsetY;

            // Scale the movie to account for different screen / movie aspect
            // ratios. May want some options for this but for now assume 
            // that we want to display the movie at the correct aspect ratio.
            {
                float fMovieAspectRatio  = static_cast<float>(g_pxBink->Width) / static_cast<float>(g_pxBink->Height);
                float fScreenAspectRatio = ( Asura_Camera::IsWidescreen() ? 16.0f / 9.0f : 4.0f / 3.0f );

                float fQuadWidth, fQuadHeight;

                if( fMovieAspectRatio > fScreenAspectRatio )
                {
                    fQuadWidth  = fViewportWidth;
                    fQuadHeight = fViewportHeight * (fScreenAspectRatio / fMovieAspectRatio);
                }
                else
                {
                    fQuadWidth  = fViewportWidth * (fScreenAspectRatio / fMovieAspectRatio);
                    fQuadHeight = fViewportHeight;
                }

                fInsetX = (fViewportWidth - fQuadWidth) * 0.5f;
                fInsetY = (fViewportHeight - fQuadHeight) * 0.5f;
            }

            // Bottom left.
            pxVertices[0].X = fInsetX;
            pxVertices[0].Y = fViewportHeight - fInsetY;
            pxVertices[0].U = 0.0f;
            pxVertices[0].V = 1.0f;

            // Top left.
            pxVertices[1].X = fInsetX;
            pxVertices[1].Y = fInsetY;
            pxVertices[1].U = 0.0f;
            pxVertices[1].V = 0.0f;

            // Top right.
            pxVertices[2].X = fViewportWidth - fInsetX;
            pxVertices[2].Y = fInsetY;
            pxVertices[2].U = 1.0f;
            pxVertices[2].V = 0.0f;

            // Bottom right.
            pxVertices[3].X = fViewportWidth - fInsetX;
            pxVertices[3].Y = fViewportHeight - fInsetY;
            pxVertices[3].U = 1.0f;
            pxVertices[3].V = 1.0f;
        }

        iResult = g_pxVertexBuffer->Unlock();
        if( FAILED(iResult) )
        {
            return false;
        }
    }

    return true;
}
