//*******************************************************************
//
// File: Asura_Direct3D_Carpet_Game_Render_MakeData.cpp
//
// Created: 26 Jan 2006
//
// Description: 
//			These routines create the optimised data version of the Carpet Data
//			1) make an RTree
//			2) Clip the RTree into more coares sections
//			3) Make batches of data for the coarse RTree data
//			4) Create Simple RTree data
//
//*******************************************************************

// system libraries
#include "Core/Asura_MemoryManagement_DebugOff.h"
#pragma warning( push )
#pragma warning( disable : 4702 ) // unreachable code
#include <vector>
#include <list>
#pragma warning( pop )
#include "Core/Asura_MemoryManagement_DebugOn.h"

#include "Core/Asura.h"

#if defined( ASURA_USE_CARPET) && defined( ASURA_TOOLS)

// main header
#include "Asura_Direct3D_Carpet_Game_Render.h"

// other libraries
#include "Carpet/Asura_Carpet_MeshManager.h"
#include "Carpet/Asura_Carpet_StyleManager.h"

// internal dependencies
#include "Debugging/Asura_Debug_Timers.h"
#include "Lighting/Tools_PhotonMapping_System.h"
#include "Lighting/Tools_EnvironmentalLightTransport.h"

#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Module.h"

// Tools includes
#include "Carpet/Tools_Carpet_Mesh.h"

//*******************************************************************
// Constants
//*******************************************************************
#define MANGLED_SB_TEXTURE_COUNT 0x1000
#define MANGLED_SB_STYLE_COUNT 0x1000
#define MANGLED_SB_BILLBOARD_COUNT 0x1000

#define MANGLED_SB_TEXTURE_OFFSET 1
#define MANGLED_SB_STYLE_OFFSET  (MANGLED_SB_TEXTURE_COUNT * MANGLED_SB_TEXTURE_OFFSET)
#define MANGLED_SB_BILLBOARD_OFFSET (MANGLED_SB_STYLE_COUNT * MANGLED_SB_STYLE_OFFSET)

//*******************************************************************
// Local Classes
//*******************************************************************

class Asura_RTree_RawRecord
{
public:
	Asura_RTree_RawRecord();
	void Asura_RTree_RawRecord::Set(const Asura_Vector_3 &xMin, const Asura_Vector_3 &xMax, int32 iIndex);
	Asura_Vector_3 GetMin() const;
	Asura_Vector_3 GetMax() const;
	int32 GetIndex() const;
	int32 GetParentIndex() const;
	int32 GetSiblingIndex() const;
	int32 GetFirstChildIndex() const;
	~Asura_RTree_RawRecord();
	
	void SetIndex( const int32 iIndex);
	void SetFirstChild( const int32 iChild);

private:

	Asura_Vector_3 m_xMin;
	Asura_Vector_3 m_xMax;
	int32 m_iIndex;
	int32 m_iParentIndex;
	int32 m_iSiblingIndex;
	int32 m_iFirstChildIndex;

	friend class Asura_RTree_Builder;
};

class Asura_RTree_Builder
{
public:

	Asura_RTree_Builder();
	~Asura_RTree_Builder();
	
	void Initialise(const uint32 i_type, const int32 i_maxSiblings);

	uint32 GetType() const;
	int32 GetRootIndex() const;
	int32 GetNumRecords() const;

	const Asura_RTree_RawRecord* GetRecord(const int32 i_index) const;
	Asura_RTree_RawRecord* GetRecord(const int32 i_index);
	void SplitRecords( int32 parentIndex);

	float CalcBBVolume( const Asura_Vector_3 &bbMin, const Asura_Vector_3 &bbMax);
	
	void AddRecord( const int32 startIndex, const int32 newIndex);
	void InsertAnonObject(const Asura_Vector_3 &aMin, const Asura_Vector_3 &aMax, int32 aIndex);
	//void InsertTriangle(CoreRawDetectionTriangle &tri, int32 triIndex);

	void InsertTriangle( const Asura_Vector_3* pxVectors, const int32 uiTriIndex);
	
private:
	uint32 m_type;
	std::vector< Asura_RTree_RawRecord> m_records;
	int32 m_rootIndex;
	int32 m_maxSiblings;

};

class Asura_CacheTriangle : public Asura_Carpet_Triangle
{
public:
	float m_fDensity;
	uint32 m_uInstanceCount;
	Asura_Carpet_Style* m_pxStyle;
	uint32 m_uTexture;
};


//*******************************************************************
// Function Definitions
//*******************************************************************

static Asura_Vector_3 GetMaximize( const Asura_Vector_3& xA, const Asura_Vector_3& xB)
{
	Asura_Vector_3 xR(
		Asura_Maths::Max( xA[0], xB[0]),
		Asura_Maths::Max( xA[1], xB[1]),
		Asura_Maths::Max( xA[2], xB[2])
	);
	
	return( xR);
}

static Asura_Vector_3 GetMinimize( const Asura_Vector_3& xA, const Asura_Vector_3& xB)
{
	Asura_Vector_3 xR(
		Asura_Maths::Min( xA[0], xB[0]),
		Asura_Maths::Min( xA[1], xB[1]),
		Asura_Maths::Min( xA[2], xB[2])
	);
	
	return( xR);
}

Asura_RTree_RawRecord::Asura_RTree_RawRecord()
{
	m_xMin = Asura_Vector_3(0, 0, 0);
	m_xMax = Asura_Vector_3(0, 0, 0);
	m_iIndex = -1;
	m_iParentIndex = -1;
	m_iSiblingIndex = -1;
	m_iFirstChildIndex = -1;
}

void Asura_RTree_RawRecord::Set(const Asura_Vector_3 &xMin, const Asura_Vector_3 &xMax, int32 iIndex)
{
	m_xMin = xMin;
	m_xMax = xMax;
	m_iIndex = iIndex;
}

Asura_Vector_3 Asura_RTree_RawRecord::GetMin() const
{
	return m_xMin;
}

Asura_Vector_3 Asura_RTree_RawRecord::GetMax() const
{
	return m_xMax;
}

int32 Asura_RTree_RawRecord::GetIndex() const
{
	return m_iIndex;
}

int32 Asura_RTree_RawRecord::GetParentIndex() const
{
	return m_iParentIndex;
}

int32 Asura_RTree_RawRecord::GetSiblingIndex() const
{
	return m_iSiblingIndex;
}

int32 Asura_RTree_RawRecord::GetFirstChildIndex() const
{
	return m_iFirstChildIndex;
}

void Asura_RTree_RawRecord::SetIndex( const int32 iIndex)
{
	this->m_iIndex = iIndex;
}

void Asura_RTree_RawRecord::SetFirstChild( const int32 iChild)
{
	this->m_iFirstChildIndex = iChild;
}

Asura_RTree_RawRecord::~Asura_RTree_RawRecord()
{
	// do nothing
}

Asura_RTree_Builder::Asura_RTree_Builder()
{
	m_type = 0;
	m_maxSiblings = 0;
	m_rootIndex = -1;
}

void Asura_RTree_Builder::Initialise(const uint32 i_type, const int32 i_maxSiblings)
{
	m_type = i_type;
	m_records.clear();
	m_maxSiblings = i_maxSiblings;
	m_rootIndex = -1;
}

uint32 Asura_RTree_Builder::GetType() const
{
	return m_type;
}

int32 Asura_RTree_Builder::GetRootIndex() const
{
	return m_rootIndex;
}

int32 Asura_RTree_Builder::GetNumRecords() const
{
	return m_records.size();
}

const Asura_RTree_RawRecord* Asura_RTree_Builder::GetRecord(const int32 i_index) const
{
	return &m_records[i_index];
}

Asura_RTree_RawRecord* Asura_RTree_Builder::GetRecord(const int32 i_index)
{
	return &m_records[i_index];
}

void Asura_RTree_Builder::SplitRecords( int32 parentIndex)
{
	if (parentIndex == -1)
	{
		// use rootIndex as start point
		parentIndex = m_records.size();
		
		//m_records.Insert(parentIndex, 1);
		{
			std::vector< Asura_RTree_RawRecord>::iterator xIteratorParent = this->m_records.begin() + parentIndex;
			this->m_records.insert( xIteratorParent, Asura_RTree_RawRecord());
		}

		m_records[parentIndex].m_iFirstChildIndex = m_rootIndex;
		m_rootIndex = parentIndex;
	}

	// insert spouse after parent, then split children up
	if (1)
	{
		int32 spouseIndex = m_records.size();
		//m_records.Insert(spouseIndex, 1);
		{
			std::vector< Asura_RTree_RawRecord>::iterator xIteratorSpouse = this->m_records.begin() + spouseIndex;
			this->m_records.insert( xIteratorSpouse, Asura_RTree_RawRecord());
		}

		m_records[spouseIndex].m_iParentIndex = m_records[parentIndex].m_iParentIndex;
		m_records[spouseIndex].m_iSiblingIndex = m_records[parentIndex].m_iSiblingIndex;
		m_records[parentIndex].m_iSiblingIndex = spouseIndex;

		int32 childIndex = m_records[parentIndex].m_iFirstChildIndex;
		int32 childCount = 0;

		int32 halfMax = m_maxSiblings >> 1;

		while (childIndex != -1)
		{
			int32 nextIndex = m_records[childIndex].m_iSiblingIndex;

			if (childCount == 0)
			{
				m_records[parentIndex].m_xMin = m_records[childIndex].m_xMin;
				m_records[parentIndex].m_xMax = m_records[childIndex].m_xMax;
			}
			else if (childCount <= halfMax)
			{
				//m_records[parentIndex].m_xMin.Minimize(m_records[childIndex].m_xMin);
				m_records[parentIndex].m_xMin = GetMinimize( m_records[parentIndex].m_xMin, m_records[childIndex].m_xMin);
				//m_records[parentIndex].m_xMax.Maximize(m_records[childIndex].m_xMax);
				m_records[parentIndex].m_xMax = GetMaximize( m_records[parentIndex].m_xMax, m_records[childIndex].m_xMax);
			}
			else if (childCount == (halfMax + 1))
			{
				m_records[spouseIndex].m_xMin = m_records[childIndex].m_xMin;
				m_records[spouseIndex].m_xMax = m_records[childIndex].m_xMax;
			}
			else if (childCount >= (halfMax + 1))
			{
				//m_records[spouseIndex].m_xMin.Minimize(m_records[childIndex].m_xMin);
				m_records[spouseIndex].m_xMin = GetMinimize( m_records[spouseIndex].m_xMin, m_records[childIndex].m_xMin);
				//m_records[spouseIndex].m_xMax.Maximize(m_records[childIndex].m_xMax);
				m_records[spouseIndex].m_xMax = GetMaximize( m_records[spouseIndex].m_xMax, m_records[childIndex].m_xMax);
			}

			if (childCount == halfMax)
			{
				m_records[childIndex].m_iSiblingIndex = -1;
			}

			if (childCount == (halfMax + 1))
			{
				m_records[spouseIndex].m_iFirstChildIndex = childIndex;
			}

			if (childCount <= halfMax)
			{
				m_records[childIndex].m_iParentIndex = parentIndex;
			}
			else if (childCount >= (halfMax + 1))
			{
				m_records[childIndex].m_iParentIndex = spouseIndex;
			}

			childIndex = nextIndex;
			childCount++;
		}
	}

	// now check if spouse has made too many siblings
	if (1)
	{
		int32 grandParentIndex = m_records[parentIndex].m_iParentIndex;

		int32 sibIndex = -1;
		int32 sibCount = 0;

		if (grandParentIndex == -1)
		{
			sibIndex = m_rootIndex;
		}
		else
		{
			sibIndex = m_records[grandParentIndex].m_iFirstChildIndex;
		}

		while (sibIndex != -1)
		{
			sibIndex = m_records[sibIndex].m_iSiblingIndex;
			sibCount++;
		}

		if (sibCount > m_maxSiblings)
		{
			SplitRecords(grandParentIndex);
		}
	}
}

float Asura_RTree_Builder::CalcBBVolume( const Asura_Vector_3 &bbMin, const Asura_Vector_3 &bbMax)
{
	Asura_Vector_3 diff = bbMax - bbMin;
	return float(diff.GetX() * diff.GetY() * diff.GetZ());
}

void Asura_RTree_Builder::AddRecord( const int32 startIndex, const int32 newIndex)
{
	if (m_records[startIndex].m_iFirstChildIndex == -1)
	{
		// is leaf
		int32 sibIndex = startIndex;
		int32 sibCount = 1;

		while (m_records[sibIndex].m_iSiblingIndex != -1)
		{
			sibIndex = m_records[sibIndex].m_iSiblingIndex;
			sibCount++;
		}

		m_records[sibIndex].m_iSiblingIndex = newIndex;

		m_records[newIndex].m_iParentIndex = m_records[sibIndex].m_iParentIndex;

		if (sibCount > m_maxSiblings)
		{
			SplitRecords(m_records[sibIndex].m_iParentIndex);
		}
	}
	else
	{
		// find best sib and descend tree
		int32 sibIndex = startIndex;

		float minVolumeDelta = 0.0f;
		int32 minIndex = -1;

		while (sibIndex != -1)
		{
			//Asura_Vector_3 tMin = m_records[sibIndex].m_xMin.GetMinimize(m_records[newIndex].m_xMin);
			//Asura_Vector_3 tMax = m_records[sibIndex].m_xMax.GetMaximize(m_records[newIndex].m_xMax);
			Asura_Vector_3 tMin = GetMinimize( m_records[sibIndex].m_xMin, m_records[newIndex].m_xMin);
			Asura_Vector_3 tMax = GetMaximize( m_records[sibIndex].m_xMax, m_records[newIndex].m_xMax);

			float tVolumeDelta = CalcBBVolume(tMin, tMax) - CalcBBVolume(m_records[sibIndex].m_xMin, m_records[sibIndex].m_xMax);

			if (minIndex == -1 || tVolumeDelta < minVolumeDelta)
			{
				minVolumeDelta = tVolumeDelta;
				minIndex = sibIndex;
			}

			sibIndex = m_records[sibIndex].m_iSiblingIndex;
		}

		if (minIndex != -1)
		{
			// modify BB
			//m_records[minIndex].m_xMin = m_records[minIndex].m_xMin.GetMinimize(m_records[newIndex].m_xMin);
			//m_records[minIndex].m_xMax = m_records[minIndex].m_xMax.GetMaximize(m_records[newIndex].m_xMax);
			m_records[minIndex].m_xMin = GetMinimize( m_records[minIndex].m_xMin, m_records[newIndex].m_xMin);
			m_records[minIndex].m_xMax = GetMaximize( m_records[minIndex].m_xMax, m_records[newIndex].m_xMax);

			// descend
			AddRecord(m_records[minIndex].m_iFirstChildIndex, newIndex);
		}
	}
}

void Asura_RTree_Builder::InsertAnonObject(const Asura_Vector_3 &aMin, const Asura_Vector_3 &aMax, int32 aIndex)
{
	// allocate new record
	int32 recIndex = m_records.size();
	//m_records.Insert(recIndex, 1);
	{
		std::vector< Asura_RTree_RawRecord>::iterator xIteratorRec = this->m_records.begin() + recIndex;
		this->m_records.insert( xIteratorRec, Asura_RTree_RawRecord());
	}

	m_records[recIndex].Set(aMin, aMax, aIndex);

	if (m_rootIndex == -1)
	{
		m_rootIndex = recIndex;
	}
	else
	{
		AddRecord(m_rootIndex, recIndex);
	}
}

#if 0
void Asura_RTree_Builder::InsertTriangle(CoreRawDetectionTriangle &tri, int32 triIndex)
{
	// allocate new record
	int32 recIndex = m_records.GetSize();
	m_records.Insert(recIndex, 1);

	// calc AABB
	Asura_Vector_3 tMin = tri.GetVertex(0);
	tMin = tMin.GetMinimize(tri.GetVertex(1));
	tMin = tMin.GetMinimize(tri.GetVertex(2));

	Asura_Vector_3 tMax = tri.GetVertex(0);
	tMax = tMax.GetMaximize(tri.GetVertex(1));
	tMax = tMax.GetMaximize(tri.GetVertex(2));

	m_records[recIndex].Set(tMin, tMax, triIndex);

	if (m_rootIndex == -1)
	{
		m_rootIndex = recIndex;
	}
	else
	{
		AddRecord(m_rootIndex, recIndex);
	}
}
#endif

void Asura_RTree_Builder::InsertTriangle( const Asura_Vector_3* pxVectors, const int32 uiTriIndex)
{
	// allocate new record
	int32 iRecIndex = m_records.size();
	//m_records.Insert(recIndex, 1);
	{
		std::vector< Asura_RTree_RawRecord>::iterator xIteratorRec = this->m_records.begin() + iRecIndex;
		this->m_records.insert( xIteratorRec, Asura_RTree_RawRecord());
	}

	// calc AABB
	Asura_Vector_3 xMin = pxVectors[0];
	xMin = GetMinimize( xMin, pxVectors[1]);
	xMin = GetMinimize( xMin, pxVectors[2]);

	Asura_Vector_3 xMax = pxVectors[0];
	xMax = GetMaximize( xMax, pxVectors[1]);
	xMax = GetMaximize( xMax, pxVectors[2]);

	m_records[ iRecIndex].Set( xMin, xMax, uiTriIndex);

	if (m_rootIndex == -1)
	{
		m_rootIndex = iRecIndex;
	}
	else
	{
		AddRecord(m_rootIndex, iRecIndex);
	}
}

Asura_RTree_Builder::~Asura_RTree_Builder()
{
	// nothing
}

static uint32 MangleTSB( const uint32 uiTexture, const uint32 uiStyle, const uint32 uiBillboard)
{
	Asura_Assertf( uiTexture < MANGLED_SB_TEXTURE_COUNT, ("uiTexture(%d) is larger than allowed(%d)", uiTexture, MANGLED_SB_TEXTURE_COUNT));
	Asura_Assertf( uiStyle < MANGLED_SB_STYLE_COUNT, ("uiStyle(%d) is larger than allowed(%d)", uiStyle, MANGLED_SB_STYLE_COUNT));
	Asura_Assertf( uiBillboard < MANGLED_SB_BILLBOARD_OFFSET, ("uiBillboard(%d) is larger than allowed(%d)", uiBillboard, MANGLED_SB_BILLBOARD_COUNT));
	
	const uint32 uiMangled = (uiTexture * MANGLED_SB_TEXTURE_OFFSET) + (uiStyle * MANGLED_SB_STYLE_OFFSET) + (uiBillboard * MANGLED_SB_BILLBOARD_OFFSET);
	
	return( uiMangled);
}

static void UnMangleTSB( const uint32 uiMangled, uint32 &uiTexture, uint32 &uiStyle, uint32 &uiBillboard)
{
	uiTexture = (uiMangled / MANGLED_SB_TEXTURE_OFFSET) % MANGLED_SB_TEXTURE_COUNT;
	uiStyle = (uiMangled / MANGLED_SB_STYLE_OFFSET) % MANGLED_SB_STYLE_COUNT;
	uiBillboard = (uiMangled / MANGLED_SB_BILLBOARD_OFFSET) % MANGLED_SB_BILLBOARD_OFFSET;
}

/*
static uint8 FloatToUint8( const float fValue)
{
	uint32 uiValue = uint32( 255.0f * fValue);
	Asura_Assert( uiValue < 0x100, "Float value not in the range of 0 to 1");
	
	uint8 ucValue =uint8(  uiValue);
	return( ucValue);
}
*/

bool Asura_Direct3D_Carpet_Game_Render::MakeDataSet_Meshes( const Asura_LinkList< Asura_Carpet_Triangle>& pxListCarpetTriangles)
{
	const uint32 uiInvalidIndex = uint32( -1);
	
	// DLH - 19 January 2006 14:28:12 - Create a list of unique textures
	Asura_LinkList< uint32> xListTextures;
	Asura_LinkList< uint32> xListMangledTS;
	Asura_LinkList< uint32> xListMangledTSB;
	Asura_LinkList< Asura_CacheTriangle> xListCachedTriangles;
	
	// Init the mesh matching
	Tools_Carpet_Mesh::BeginFindScenePoly();
	

	{
		for( Asura_ListElt< Asura_Carpet_Triangle>* pxIteratorTriangle = pxListCarpetTriangles.Front(); pxIteratorTriangle; pxIteratorTriangle = pxIteratorTriangle->Next())
		{		
			const Asura_Carpet_Triangle& xTriangle = pxIteratorTriangle->Data();
			const uint32 uiStyle = xTriangle.iStyle;
			Asura_Carpet_Style* pxStyle = Asura_Carpet_StyleManager::GetStyle( uiStyle);
			
			//Asura_Assertf( NULL != pxStyle, ("Style %d does not return a valid pointer", uiStyle));
			if(NULL == pxStyle)
			{
				continue;
			}
			
			// DLH - 19 January 2006 15:21:00 - Make List of Unique texture
			uint32 uiTexture = pxStyle->iTextureID;
			if( NULL == xListTextures.Find( uiTexture))
			{
				xListTextures.Push_Front( uiTexture);
			}
			
			// DLH - 19 January 2006 15:21:10 - Make List of Unique Texture/Styles
			uint32 uiMangledTS = MangleTSB( uiTexture, uiStyle, 0);
			if( NULL == xListMangledTS.Find( uiMangledTS))
			{
				xListMangledTS.Push_Front( uiMangledTS);
			}
			
			// DLH - 19 January 2006 15:38:03 - Calc Cache Triangle
			Asura_ListElt< Asura_CacheTriangle>* xIteratorCachedTriangle = xListCachedTriangles.Push_Back( Asura_CacheTriangle());
			Asura_CacheTriangle& xCacheTriangle = xIteratorCachedTriangle->Data();
			
			*static_cast<Asura_Carpet_Triangle*>( &xCacheTriangle) = xTriangle;
			xCacheTriangle.m_fDensity = pxStyle->fDensity * (xTriangle.xVerts[0].fDensity + xTriangle.xVerts[1].fDensity + xTriangle.xVerts[2].fDensity) * 0.3333f;
			xCacheTriangle.m_uInstanceCount = uint32(xTriangle.fArea * xCacheTriangle.m_fDensity);
			xCacheTriangle.m_pxStyle = pxStyle;
			xCacheTriangle.m_uTexture = uiTexture;
			
			// DLH - 19 January 2006 15:21:27 - Make List of Unique Texture/Style/Billboards
			for( uint32 uiInstanceIndex = 0; uiInstanceIndex < xCacheTriangle.m_uInstanceCount; uiInstanceIndex += 1)
			{
				uint32 uiBillBoardIndex = pxStyle->aiSequence[ uiInstanceIndex % ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH];
				
				uint32 uiMangledTSB = MangleTSB( uiTexture, uiStyle, uiBillBoardIndex);
				if( NULL == xListMangledTSB.Find( uiMangledTSB))
				{
					xListMangledTSB.Push_Front( uiMangledTSB);
				}
			}
		}
	}
	
	Asura_Assertf( xListCachedTriangles.Size() == pxListCarpetTriangles.Size(), ("Cached Triangle List is a different size Cache %d != Original %d", xListCachedTriangles.Size(), pxListCarpetTriangles.Size()));

	// DLH - 19 January 2006 14:28:42 - Create Mesh list from the textures
	const uint32 uiTextureCount = xListTextures.Size();
//	const uint32 uiMangledTSCount = xListMangledTS.Size();
//	const uint32 uiMangledTSBCount = xListMangledTSB.Size();
	
	const uint32 uiCGRMeshCount = uiTextureCount;
	
	s_pxMeshes = new Asura_Direct3D_CGR_Mesh[ uiCGRMeshCount];
	s_uMeshesCount = uiCGRMeshCount;
	
	{
		Asura_Direct3D_CGR_Mesh* pxCGRMesh = s_pxMeshes;
		for( Asura_ListElt< uint32>* pxIteratorTexture = xListTextures.Front(); pxIteratorTexture; pxIteratorTexture = pxIteratorTexture->Next(), pxCGRMesh += 1)
		{		
			const uint32 uiTexture = pxIteratorTexture->Data();
			
			pxCGRMesh->m_iTexture = uiTexture;
			
			// DLH - 19 January 2006 16:04:23 - Count and allocated the styles
			Asura_LinkList< uint32> xListMangledSB;
			Asura_LinkList< uint32> xListStyle;
			for( Asura_ListElt< uint32> *pxIteratorMTSB = xListMangledTSB.Front(); pxIteratorMTSB; pxIteratorMTSB = pxIteratorMTSB->Next())
			{
				const uint32 uiMangledTSB = pxIteratorMTSB->Data();
				uint32 uiMangledTexture;
				uint32 uiMangledStyle;
				uint32 uiMangledBillboard;
				
				UnMangleTSB( uiMangledTSB, uiMangledTexture, uiMangledStyle, uiMangledBillboard);
				if( uiTexture != uiMangledTexture)
				{
					continue;
				}
				
				if( NULL == xListMangledSB.Find( uiMangledTSB))
				{
					xListMangledSB.Push_Front( uiMangledTSB);
				}
				if( NULL == xListStyle.Find( uiMangledStyle))
				{
					xListStyle.Push_Front( uiMangledStyle);
				}
			}
			
			// DLH - 19 January 2006 17:32:27 - This is a list that matches the styles length
			// DLH - 19 January 2006 17:32:51 - Each entry is a list of pointers to cached triangles!
			Asura_LinkList< Asura_LinkList< Asura_CacheTriangle*> > xListListStyle_pxCacheTriangles;
			
			pxCGRMesh->m_uPolygonCount = 0;
			pxCGRMesh->m_uStyleCount = xListMangledSB.Size();
			pxCGRMesh->m_pxStyles = new Asura_Direct3D_CGR_Style[ pxCGRMesh->m_uStyleCount];
			
			pxCGRMesh->m_uPolygonStyleCount = xListStyle.Size();
			pxCGRMesh->m_pxPolygonStyles = new Asura_Direct3D_CGR_PolygonStyle[ pxCGRMesh->m_uPolygonStyleCount];
			
			// DLH - 20 January 2006 17:01:33 - Converts Style into CGRPolygonStyle index
			std::vector< uint32> xArrayStyleToPolygonStyle;
			xArrayStyleToPolygonStyle.resize( Asura_Carpet_StyleManager::GetNumStyles(), uiInvalidIndex);
			
			// DLH - 20 January 2006 17:01:54 - Converts CGRPolygonStyle and Billboard to CGRStyle
			std::vector< std::vector< uint32> > xArrayPolygonStyleArrayBillboardToCGRStyle;
			xArrayPolygonStyleArrayBillboardToCGRStyle.resize( pxCGRMesh->m_uPolygonStyleCount);
			
			
			// DLH - 20 January 2006 16:53:42 - Create The PolygonStyle
			{
				uint32 uiPolygonStyle = 0;
				Asura_Direct3D_CGR_PolygonStyle* pxPolygonStyle = pxCGRMesh->m_pxPolygonStyles;
				for( Asura_ListElt< uint32>* pxIteratorStyle = xListStyle.Front(); pxIteratorStyle; pxIteratorStyle = pxIteratorStyle->Next(), pxPolygonStyle += 1, uiPolygonStyle += 1)
				{
					const uint32 uiStyle = pxIteratorStyle->Data();
					
					pxPolygonStyle->m_uStyle = uiStyle;
					xArrayStyleToPolygonStyle[ uiStyle] = uiPolygonStyle;
					
					std::vector< uint32>& xArrayBillboardToCGRStyle = xArrayPolygonStyleArrayBillboardToCGRStyle[ uiPolygonStyle];
					xArrayBillboardToCGRStyle.resize( ASURA_CARPET_BILLBOARDS_PER_STYLE, uiInvalidIndex);
				}
			}
			
			// DLH - 19 January 2006 16:09:55 - Fill in the CGR Style Data
			{
				Asura_Direct3D_CGR_Style* pxCGRStyle = pxCGRMesh->m_pxStyles;
				uint32 uiCGRStyle = 0;
				for( Asura_ListElt< uint32>* pxIteratorMSB = xListMangledSB.Front(); pxIteratorMSB; pxIteratorMSB = pxIteratorMSB->Next(), pxCGRStyle += 1, uiCGRStyle += 1)
				{
					const uint32 uiMangledSB = pxIteratorMSB->Data();
					uint32 uiMangledTexture;
					uint32 uiMangledStyle;
					uint32 uiMangledBillboard;
					
					UnMangleTSB( uiMangledSB, uiMangledTexture, uiMangledStyle, uiMangledBillboard);
					
					Asura_Assertf( uiTexture == uiMangledTexture, ("Texture does not match %d != %d", uiTexture, uiMangledTexture));
					
					Asura_Carpet_Style* pxStyle = Asura_Carpet_StyleManager::GetStyle( uiMangledStyle);
					Asura_Assertf( NULL != pxStyle, ("Style %d does not return a valid pointer", uiMangledStyle));
					Asura_Assert( uiMangledBillboard < ASURA_CARPET_BILLBOARDS_PER_STYLE, "Billboard Style index too high");
					Asura_Carpet_Billboard *pxBillboard = &pxStyle->axBillboards[uiMangledBillboard];

					pxCGRStyle->m_fWidth = pxBillboard->fWidth;
					pxCGRStyle->m_fHeight = pxBillboard->fHeight;
					pxCGRStyle->m_afU[ 0] = pxBillboard->fU[ 0];
					pxCGRStyle->m_afU[ 1] = pxBillboard->fU[ 1];
					pxCGRStyle->m_afV[ 0] = pxBillboard->fV[ 0];
					pxCGRStyle->m_afV[ 1] = pxBillboard->fV[ 1];
					pxCGRStyle->m_fScaleMin = pxBillboard->fMinScale;

					Asura_Assert( pxStyle->fLodFadeNear <= pxStyle->fLodFadeFar, "Near Clip is greater than Far clip!");
					pxCGRStyle->m_fLodFadeNear = pxStyle->fLodFadeNear;
					pxCGRStyle->m_fLodFadeFar = pxStyle->fLodFadeFar;
					pxCGRStyle->m_fLodFadeDelta = pxStyle->fLodFadeFar - pxCGRStyle->m_fLodFadeNear;
					pxCGRStyle->m_fLodFadeDelta = Asura_Maths::Max( 1e-3f, pxCGRStyle->m_fLodFadeDelta);
					pxCGRStyle->m_fMovementScale = pxStyle->fMovementScale;
					
					// DLH - 20 January 2006 17:15:00 - Save the remapping Data for the style
					uint32 uiCGRPolygonStyle = xArrayStyleToPolygonStyle[ uiMangledStyle];
					Asura_Assert( uiCGRPolygonStyle < pxCGRMesh->m_uPolygonStyleCount, "Style remapping");
					
					std::vector< uint32>& xArrayBillboardToCGRStyle = xArrayPolygonStyleArrayBillboardToCGRStyle[ uiCGRPolygonStyle];
					xArrayBillboardToCGRStyle[ uiMangledBillboard] = uiCGRStyle;
				}
			}
					
			// DLH - 20 January 2006 17:16:39 - Save off the remapped CGRStyle into the PolygonStyle
			{
				pxCGRMesh->m_fLodFadeFar = 0;
				
				uint32 uiPolygonStyle = 0;
				Asura_Direct3D_CGR_PolygonStyle* pxPolygonStyle = pxCGRMesh->m_pxPolygonStyles;
				for( Asura_ListElt< uint32>* pxIteratorStyle = xListStyle.Front(); pxIteratorStyle; pxIteratorStyle = pxIteratorStyle->Next(), pxPolygonStyle += 1, uiPolygonStyle += 1)
				{
					const uint32 uiStyle = pxIteratorStyle->Data();
					Asura_Carpet_Style* pxStyle = Asura_Carpet_StyleManager::GetStyle( uiStyle);
					
					std::vector< uint32>& xArrayBillboardToCGRStyle = xArrayPolygonStyleArrayBillboardToCGRStyle[ uiPolygonStyle];
					
					for( uint32 uiSequenceIndex = 0; uiSequenceIndex < ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH; uiSequenceIndex += 1)
					{
						const uint32 uiBillboardIndex = pxStyle->aiSequence[ uiSequenceIndex];
						Asura_Assert( uiBillboardIndex < ASURA_CARPET_BILLBOARDS_PER_STYLE, "Billboard Index is too high");
						Asura_Assert( uiBillboardIndex < xArrayBillboardToCGRStyle.size(), "Billboard Index is too high");
						const uint32 uiCGRStyleIndex = xArrayBillboardToCGRStyle[ uiBillboardIndex];
						Asura_Assert( uiInvalidIndex == uiCGRStyleIndex || uiCGRStyleIndex < pxCGRMesh->m_uStyleCount, "Style Index is too high");
						
						if( uiInvalidIndex != uiCGRStyleIndex)
						{
							const Asura_Direct3D_CGR_Style& xCGRStyle = pxCGRMesh->m_pxStyles[ uiCGRStyleIndex];
							
							pxPolygonStyle->m_fLodFadeFar = Asura_Maths::Max( pxPolygonStyle->m_fLodFadeFar, xCGRStyle.m_fLodFadeFar);
						}
						
						pxPolygonStyle->m_aucStyleLookup[ uiSequenceIndex] = uiCGRStyleIndex;
					}
					
					pxCGRMesh->m_fLodFadeFar = Asura_Maths::Max( pxCGRMesh->m_fLodFadeFar, pxPolygonStyle->m_fLodFadeFar);
				}
			}
			
			// DLH - 19 January 2006 17:19:22 - Count the Triangles for the peresent Style
			// DLH - Create a matching cached triangle list 
			Asura_LinkList< Asura_CacheTriangle*> xListTextureCacheTriangles;
			
			for( Asura_ListElt< Asura_CacheTriangle>* pxIteratorCachedTexture = xListCachedTriangles.Front(); pxIteratorCachedTexture; pxIteratorCachedTexture = pxIteratorCachedTexture->Next())
			{
				Asura_CacheTriangle& xCachedTriangle = pxIteratorCachedTexture->Data();
				
				if( uiTexture != xCachedTriangle.m_uTexture)
				{
					continue;
				}
				
				xListTextureCacheTriangles.Push_Back( &xCachedTriangle);
			}
			
			// DLH - 19 January 2006 17:34:05 - Allocate the polygon list
			pxCGRMesh->m_uPolygonCount = xListTextureCacheTriangles.Size();
			pxCGRMesh->m_pxPolygons = new Asura_Direct3D_CGR_Polygon[ pxCGRMesh->m_uPolygonCount];
			pxCGRMesh->m_uVertexCount = 3 * pxCGRMesh->m_uPolygonCount;
			pxCGRMesh->m_pxVertices = new Asura_Direct3D_CGR_Vertex[ pxCGRMesh->m_uVertexCount];
			
			// DLH - 20 January 2006 17:24:47 - Need To Add a sort routine here to order the polygons in Style
//			SortPolygonsIntoStyleOrder();
			
			pxCGRMesh->m_xBoundingBox.SetToPoint( Asura_Vector_3( 0, 0, 0));
			// DLH - 19 January 2006 17:35:07 - Copy the Cached polygons into the mesh in style order
			if( 0 != xListTextureCacheTriangles.Size())
			{
//				Asura_Direct3D_CGR_Style* pxCGRStyle = pxCGRMesh->m_pxStyles;
				Asura_Direct3D_CGR_Polygon* pxCGRPolygon = pxCGRMesh->m_pxPolygons;
				Asura_Direct3D_CGR_Vertex* pxCGRVertex = pxCGRMesh->m_pxVertices;
				
				bool bIsBBInitialised = false;
				
				for( Asura_ListElt< Asura_CacheTriangle*>* pxIteratorPolygon = xListTextureCacheTriangles.Front(); pxIteratorPolygon; pxIteratorPolygon = pxIteratorPolygon->Next(), pxCGRPolygon += 1, pxCGRVertex += 3)
				{
					const Asura_CacheTriangle* pxCacheTriangle = pxIteratorPolygon->Data();
					
					Asura_Assert( pxCGRPolygon < (pxCGRMesh->m_pxPolygons + pxCGRMesh->m_uPolygonCount), "Invalid pointer");
					Asura_Assert( pxCGRVertex < (pxCGRMesh->m_pxVertices + pxCGRMesh->m_uVertexCount), "Invalid pointer");
					
					pxCGRPolygon->m_fDenisty = pxCacheTriangle->m_fDensity;
					pxCGRPolygon->m_uInstanceCount = pxCacheTriangle->m_uInstanceCount;
					pxCGRPolygon->m_uRandomSeed = pxCacheTriangle->uSeed;
					
					pxCGRPolygon->m_xCentre = pxCacheTriangle->xCentre;
					pxCGRPolygon->m_fRadius = pxCacheTriangle->fRadius;

					const uint32 uiVertexBaseIndex = pxCGRVertex - pxCGRMesh->m_pxVertices;
					Asura_Assert( (pxCGRMesh->m_pxVertices + uiVertexBaseIndex) == pxCGRVertex, "Vertex Pointer out of Sync");
					
					pxCGRPolygon->m_uVertexIndex[ 0] = uiVertexBaseIndex + 0;
					pxCGRPolygon->m_uVertexIndex[ 1] = uiVertexBaseIndex + 1;
					pxCGRPolygon->m_uVertexIndex[ 2] = uiVertexBaseIndex + 2;
					
					for( uint32 uiCornerIndex = 0; uiCornerIndex < 3; uiCornerIndex += 1)
					{
						const Asura_Carpet_Vertex* pxVerts = &pxCacheTriangle->xVerts[ uiCornerIndex];
						Asura_Direct3D_CGR_Vertex* pxCGRVertexCorner = &pxCGRVertex[ uiCornerIndex];
						
						if( false == bIsBBInitialised)
						{
							pxCGRMesh->m_xBoundingBox.SetToPoint( pxVerts->xPos);
							bIsBBInitialised = true;
						}
						else						
						{
							pxCGRMesh->m_xBoundingBox.Set( pxVerts->xPos);
						}
						
						pxCGRVertexCorner->m_xPosition = pxVerts->xPos;
						
						{
							int				iModule = -1;
							int				iPoly = -1;
							int				iVtx = -1;
							Asura_Vector_3  xNormal;
							u_int			uPaintedColour;

							if(false == Tools_Carpet_Mesh::FindScenePoly( pxCacheTriangle, uiCornerIndex, &iModule, &iPoly, &iVtx, &xNormal, &uPaintedColour))
							{	
								iModule = -1;
								iPoly = -1;
								iVtx = -1;
							}
							
							pxCGRVertexCorner->m_xNormal = xNormal;
							pxCGRVertexCorner->m_uModule = iModule;
							pxCGRVertexCorner->m_uPoly = iPoly;
							pxCGRVertexCorner->m_uVtx = iVtx;
						}

#if defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
						pxCGRVertexCorner->m_xColourRGB[0] = pxVerts->afColour[0];
						pxCGRVertexCorner->m_xColourRGB[1] = pxVerts->afColour[1];
						pxCGRVertexCorner->m_xColourRGB[2] = pxVerts->afColour[2];
#else
						pxCGRVertexCorner->m_xColourRGB[0] = (0xFF & (pxVerts->uColour >> 0)) / float( 0xFF);
						pxCGRVertexCorner->m_xColourRGB[1] = (0xFF & (pxVerts->uColour >> 8)) / float( 0xFF);
						pxCGRVertexCorner->m_xColourRGB[2] = (0xFF & (pxVerts->uColour >> 16)) / float( 0xFF);
#endif

						pxCGRVertexCorner->m_fDensity = pxVerts->fDensity;
					}
					
					uint32 uiStyle = pxCacheTriangle->iStyle;
					Asura_Assert( uiStyle < uint32( Asura_Carpet_StyleManager::GetNumStyles()), "Style Index is too high");
					uint32 uiCGRPolygonStyle = xArrayStyleToPolygonStyle[ uiStyle];
					Asura_Assert( uiCGRPolygonStyle < pxCGRMesh->m_uPolygonStyleCount, "PolygonStyle index is too high");
					if (uiCGRPolygonStyle >= pxCGRMesh->m_uPolygonStyleCount)
					{
						// MGH 25/1/2008: Better to default the style to zero than have out-of-range array accesses.
						uiCGRPolygonStyle = 0;
					}

					pxCGRPolygon->m_uPolygonStyle = uiCGRPolygonStyle;
					
					const Asura_Direct3D_CGR_PolygonStyle& xPolygonStyle = pxCGRMesh->m_pxPolygonStyles[ uiCGRPolygonStyle];
					pxCGRPolygon->m_fLodFadeFar = xPolygonStyle.m_fLodFadeFar;
				}
				
				Asura_Assert( (pxCGRMesh->m_pxPolygons + pxCGRMesh->m_uPolygonCount) == pxCGRPolygon, "Processing has messed up.");
				Asura_Assert( (pxCGRMesh->m_pxVertices + pxCGRMesh->m_uVertexCount) == pxCGRVertex, "Processing has messed up.");
				
			}
	
		MakeDataSet_MeshesRTree( pxCGRMesh);
		MakeDataSet_CalcMinimumVertexBufferSize( pxCGRMesh);
		}
	}

	Tools_Carpet_Mesh::EndFindScenePoly();
	
	return( true);
}

bool Asura_Direct3D_Carpet_Game_Render::MakeDataSet_MeshesRTree( Asura_Direct3D_CGR_Mesh* pxMesh)
{
	// DLH - 26 January 2006 15:59:26 - Make an RTree that contains all the triangles
	Asura_RTree_Builder xRTreeBuilder;
	
	xRTreeBuilder.Initialise( 0, 4);
	
	const uint32 uiPolygonCount = pxMesh->m_uPolygonCount;
	for( uint32 uiPolygonIndex = 0; uiPolygonIndex < uiPolygonCount; uiPolygonIndex += 1)
	{
		const Asura_Direct3D_CGR_Polygon& xPolygon = pxMesh->m_pxPolygons[ uiPolygonIndex];
		const uint32* pxVertexIndices = xPolygon.m_uVertexIndex;
		const Asura_Direct3D_CGR_Vertex* pxVertices = pxMesh->m_pxVertices;
		const Asura_Vector_3 axCorners[ 3] = 
		{
			pxVertices[ pxVertexIndices[ 0]].m_xPosition,
			pxVertices[ pxVertexIndices[ 1]].m_xPosition,
			pxVertices[ pxVertexIndices[ 2]].m_xPosition,
		};
		
#if 0
		// DLH - 26 January 2006 15:23:47 - adds the style widths to the BB
		{
			Asura_Vector_3 xSizeMax( 0, 0, 0);
			const Asura_Direct3D_CGR_PolygonStyle& xPolygonStyle = pxMesh->m_pxPolygonStyles[ xPolygon.m_uPolygonStyle];
			for( uint32 uiStyleListIndex = 0; uiStyleListIndex < ASURA_DIRECT3D_CGR_POLYGONSTYLE_LOOKUP_LENGTH; uiStyleListIndex += 1)
			{
				const uint32 uiStyleIndex = xPolygonStyle.m_aucStyleLookup[ uiStyleListIndex];
				if( Asura_Direct3D_CGR_PolygonStyle::uiInvalidStyleIndex == uiStyleIndex)
				{
					continue;
				}
				
				Asura_Direct3D_CGR_Style& xStyle = pxMesh->m_pxStyles[ uiStyleIndex];
				
				const float fSize = Asura_Maths::Max( xStyle.m_fHeight, xStyle.m_fWidth);
				Asura_Vector_3 xSize( fSize, fSize, fSize);
				
				xSizeMax = GetMaximize( xSizeMax, xSize);
			}

			axCorners[ 0] += 			
		}	
#endif	
		
		
		// DLH - 26 January 2006 15:46:31 - Calc the large RTree
		xRTreeBuilder.InsertTriangle( axCorners, uiPolygonIndex);
	}
	
	// DLH - 31 January 2006 14:41:33 - Truncate Data Tree
	// DLH - 31 January 2006 14:40:51 - Make Data Batches. Convert RTree from Polygons to Batches
	MakeDataSet_MeshesBatches( *pxMesh, xRTreeBuilder);
	
	// DLH - 31 January 2006 14:41:47 - Copy Data Tree
	MakeDataSet_RTreeBuilderToCGRRTree( xRTreeBuilder, *pxMesh);
	
	return( true);	
}

static uint32 CollapseRTreeNode_CountPolygons( const Asura_RTree_Builder &xRTreeBuilder, const Asura_RTree_RawRecord* pxRTreeFirstNode)
{
	uint32 uiPolygonCount = 0;
	
	// DLH - 31 January 2006 15:31:27 - Process First node
	uiPolygonCount += (-1 != pxRTreeFirstNode->GetIndex()) ? 1 : 0;
	
	std::list< uint32 > xListRTreeNodeToProcess;
	
	{
		const uint32 uiChildIndex = pxRTreeFirstNode->GetFirstChildIndex();
		xListRTreeNodeToProcess.push_back( uiChildIndex);
	}
	
	// DLH - 31 January 2006 15:31:39 - Process Children
	while( 0 < xListRTreeNodeToProcess.size())
	{
		const uint32 uiNodeIndex = xListRTreeNodeToProcess.front();
		xListRTreeNodeToProcess.pop_front();
		
		if( -1 == uiNodeIndex)
		{
			continue;
		}
		
		const Asura_RTree_RawRecord* pxRTreeNode = xRTreeBuilder.GetRecord( uiNodeIndex);
		
		{
			const uint32 uiNodeChildIndex = pxRTreeNode->GetFirstChildIndex();
			xListRTreeNodeToProcess.push_back( uiNodeChildIndex);
		}
		
		{
			const uint32 uiNodeSiblingIndex = pxRTreeNode->GetSiblingIndex();
			xListRTreeNodeToProcess.push_back( uiNodeSiblingIndex);
		}
		
		
		uiPolygonCount += (-1 != pxRTreeNode->GetIndex()) ? 1 : 0;
	}
	
	return( uiPolygonCount);
}

static bool CollapseRTreeNode_BuildBatch( const Asura_RTree_Builder &xRTreeBuilder, const Asura_RTree_RawRecord* pxRTreeFirstNode, std::list< uint32>& xListuiPolygonIndices)
{
	// DLH - 31 January 2006 15:31:27 - Process First node
	{
		const int32 iPolygonIndex = pxRTreeFirstNode->GetIndex();
		if( -1 != iPolygonIndex)
		{
			xListuiPolygonIndices.push_back( iPolygonIndex);
		}
	}
	
	std::list< uint32 > xListRTreeNodeToProcess;
	
	{
		const uint32 uiChildIndex = pxRTreeFirstNode->GetFirstChildIndex();
		xListRTreeNodeToProcess.push_back( uiChildIndex);
	}
	
	// DLH - 31 January 2006 15:31:39 - Process Children
	while( 0 < xListRTreeNodeToProcess.size())
	{
		const uint32 uiNodeIndex = xListRTreeNodeToProcess.front();
		xListRTreeNodeToProcess.pop_front();
		
		if( -1 == uiNodeIndex)
		{
			continue;
		}
		
		const Asura_RTree_RawRecord* pxRTreeNode = xRTreeBuilder.GetRecord( uiNodeIndex);
		
		{
			const uint32 uiNodeChildIndex = pxRTreeNode->GetFirstChildIndex();
			xListRTreeNodeToProcess.push_back( uiNodeChildIndex);
		}
		
		{
			const uint32 uiNodeSiblingIndex = pxRTreeNode->GetSiblingIndex();
			xListRTreeNodeToProcess.push_back( uiNodeSiblingIndex);
		}
		
		
		const int32 iPolygonIndex = pxRTreeNode->GetIndex();
		if( -1 != iPolygonIndex)
		{
			xListuiPolygonIndices.push_back( iPolygonIndex);
		}
	}
	
	return( true);
}

//-------------------------------------------------------------------
// Function:    CollapseRTreeNodeToCount
//
// Description: Calculates how many polygons to collapse this RTree brach down to for a batch
//
// Parameters:  xRTreeBuilder   - RTree DBase
//              pxRTreeNode     - RTree Start Node
//
// Returns:     The number of polygons to batch together
//				If the branch is too big then 0 will returned
//				This means any RTree branch must process the children until a non-zero value is found 
//-------------------------------------------------------------------
static uint32 CollapseRTreeNodeToCount( const Asura_RTree_Builder &xRTreeBuilder, const Asura_RTree_RawRecord* pxRTreeNode)
{
	// DLH - 31 January 2006 15:07:09 - Leaf node so batch it
#if 0
	if( -1 == pxRTreeNode->GetFirstChildIndex())
	{
		return( 0);
	}
	
	if( -1 != pxRTreeNode->GetIndex())
	{
		return( 0);
	}
#endif
	
	const uint32 uiPolygonCount = CollapseRTreeNode_CountPolygons( xRTreeBuilder, pxRTreeNode);
	if( 0 < uiPolygonCount && uiPolygonCount <= 30)
//	if( 0 < uiPolygonCount && uiPolygonCount <= 1)
	{
		return( uiPolygonCount);
	}
	
	return( 0);
}

bool Asura_Direct3D_Carpet_Game_Render::MakeDataSet_MeshesBatches( Asura_Direct3D_CGR_Mesh& xMesh, Asura_RTree_Builder &xRTreeBuilder)
{
	std::list< std::list< uint32> > xListBatches_xListuiPolygonIndices;
	
	std::list< uint32 > xListRTreeNodeToProcess;
	
	xListRTreeNodeToProcess.push_back( xRTreeBuilder.GetRootIndex());
	
	while( 0 < xListRTreeNodeToProcess.size())
	{
		// DLH - 31 January 2006 14:52:31 - Process one leaf node
		const uint32 uiRTreeNodeIndex = xListRTreeNodeToProcess.front();
		xListRTreeNodeToProcess.pop_front();
		
		if( -1 == uiRTreeNodeIndex)
		{
			continue;
		}
		
		Asura_RTree_RawRecord* pxRTreeNode = xRTreeBuilder.GetRecord( uiRTreeNodeIndex);
		
		{
			// DLH - 31 January 2006 15:08:27 - Add this nodes next sibling to the process stack
			const int32 iSiblingIndex = pxRTreeNode->GetSiblingIndex();
			xListRTreeNodeToProcess.push_back( iSiblingIndex);
		}
		
		const uint32 uiPolygonCount = CollapseRTreeNodeToCount( xRTreeBuilder, pxRTreeNode);
		
		if( 0 < uiPolygonCount)
		{
			// DLH - 31 January 2006 15:12:08 - Add a new batch to the list
			const uint32 uiBatchIndex = xListBatches_xListuiPolygonIndices.size();
			
			{
				std::list< uint32> xListEmpty;
				xListBatches_xListuiPolygonIndices.push_back( xListEmpty);
			}
			std::list< uint32>& xListuiPolygonIndices = xListBatches_xListuiPolygonIndices.back();
			
			// DLH - 31 January 2006 15:12:22 - Fill batch with data
			CollapseRTreeNode_BuildBatch( xRTreeBuilder, pxRTreeNode, xListuiPolygonIndices);
			
			Asura_Assert( xListuiPolygonIndices.size() == uiPolygonCount, "The wrong number of polygons have been inserted into the list");
			
			// DLH - 31 January 2006 15:46:14 - Change Index to Batch number
			pxRTreeNode->SetIndex( uiBatchIndex);
			// DLH - 31 January 2006 15:44:20 - Terminate Branch
			pxRTreeNode->SetFirstChild( -1);
			
			continue;
		}

		{
			// DLH - 31 January 2006 15:12:38 - Add Child to process list
			const int32 iChildIndex = pxRTreeNode->GetFirstChildIndex();
			xListRTreeNodeToProcess.push_back( iChildIndex);
		}		
	}
	
	// DLH - 31 January 2006 15:58:42 - Create and fill the batches
	// DLH - 31 January 2006 16:41:57 - Create a reordered list of polygons
	
	
	{
		xMesh.m_uPolygonBatchCount = xListBatches_xListuiPolygonIndices.size();
		xMesh.m_pxPolygonBatches = new Asura_Direct3D_CGR_PolygonBatch[ xMesh.m_uPolygonBatchCount];
		
		Asura_Direct3D_CGR_Polygon* pxNewCGRPolygons = new Asura_Direct3D_CGR_Polygon[ xMesh.m_uPolygonCount];
		
		Asura_Direct3D_CGR_PolygonBatch* pxCGRPolygonBatch = xMesh.m_pxPolygonBatches;
		Asura_Direct3D_CGR_Polygon* pxNewCGRPolygon = pxNewCGRPolygons;
		for( std::list< std::list< uint32> >::iterator xItBatch = xListBatches_xListuiPolygonIndices.begin(); xItBatch != xListBatches_xListuiPolygonIndices.end(); xItBatch++, pxCGRPolygonBatch += 1)
		{
			std::list< uint32>& xListuiPolygonIndices = *xItBatch;
			
			Asura_Assert( xMesh.m_pxPolygonBatches <= pxCGRPolygonBatch, "Bad CGRBatch Pointer");
			Asura_Assert( pxCGRPolygonBatch < (xMesh.m_pxPolygonBatches + xMesh.m_uPolygonBatchCount), "Bad CGRBatch Pointer");
			
			const uint32 uiBatchPolygonCount = xListuiPolygonIndices.size();
			
			// DLH - 31 January 2006 16:47:43 - Setup Batch Data 
			pxCGRPolygonBatch->m_uDataOffset = pxNewCGRPolygon - pxNewCGRPolygons;
			Asura_Assert( 0 <= pxCGRPolygonBatch->m_uDataOffset, "Invalid Data Offset");
			Asura_Assert( pxCGRPolygonBatch->m_uDataOffset < xMesh.m_uPolygonCount, "Invalid Data Offset");
			pxCGRPolygonBatch->m_uDataSize = uiBatchPolygonCount;
			Asura_Assert( 0 < pxCGRPolygonBatch->m_uDataSize, "Should not have any empty Batchs");
			Asura_Assert( (pxCGRPolygonBatch->m_uDataOffset + pxCGRPolygonBatch->m_uDataSize) <= xMesh.m_uPolygonCount , "Invalid Data Size");

			// DLH - 31 January 2006 16:47:51 - Copy Batch's polygons into a linear order on the new polygon list
			for( std::list< uint32>::iterator xItBatchPolygon = xListuiPolygonIndices.begin(); xItBatchPolygon != xListuiPolygonIndices.end(); xItBatchPolygon++, pxNewCGRPolygon += 1)
			{
				Asura_Assert( pxNewCGRPolygon < (pxNewCGRPolygons + xMesh.m_uPolygonCount), "Polygon pointer out of range");
				
				const uint32 uiPolygonIndex = *xItBatchPolygon;
				Asura_Assert( uiPolygonIndex < xMesh.m_uPolygonCount, "Polygon Index out of range");
				
				Asura_Direct3D_CGR_Polygon xOldCGRPolygon = xMesh.m_pxPolygons[ uiPolygonIndex];
				
				*pxNewCGRPolygon = xOldCGRPolygon;
				
			}
		}
		
		Asura_Assert( (pxNewCGRPolygons + xMesh.m_uPolygonCount) == pxNewCGRPolygon, "Polygon pointer is not in the correct place");
		
		// DLH - 31 January 2006 16:54:44 - Replace old polygon list with new reordered one
		delete[] xMesh.m_pxPolygons;
		xMesh.m_pxPolygons = pxNewCGRPolygons;
	}
	
	return( true);	
}

static uint32 RTreeNode_CountPolygons( const Asura_RTree_Builder &xRTreeBuilder, std::vector< int32>& xArrayRTreeBuilderToCGRIndex)
{
	uint32 uiNodeCount = 0;
	
	std::list< uint32 > xListRTreeNodeToProcess;
	xListRTreeNodeToProcess.push_back( xRTreeBuilder.GetRootIndex());
	
	while( 0 < xListRTreeNodeToProcess.size())
	{
		const uint32 uiNodeIndex = xListRTreeNodeToProcess.front();
		xListRTreeNodeToProcess.pop_front();
		
		if( -1 == uiNodeIndex)
		{
			continue;
		}
		
		xArrayRTreeBuilderToCGRIndex[ uiNodeIndex] = uiNodeCount;
		
		const Asura_RTree_RawRecord* pxRTreeNode = xRTreeBuilder.GetRecord( uiNodeIndex);
		
		{
			const uint32 uiNodeChildIndex = pxRTreeNode->GetFirstChildIndex();
			xListRTreeNodeToProcess.push_back( uiNodeChildIndex);
		}
		
		{
			const uint32 uiNodeSiblingIndex = pxRTreeNode->GetSiblingIndex();
			xListRTreeNodeToProcess.push_back( uiNodeSiblingIndex);
		}
		
		uiNodeCount += 1;
	}
	
	return( uiNodeCount);
}

static Asura_Direct3D_CGR_RTreeNode* GetPointerFromIndex( const Asura_Direct3D_CGR_Mesh& xMesh, uint32 uiIndex)
{
	if( uint32( -1) == uiIndex)
	{
		return( NULL);
	}
	
	Asura_Assert( uiIndex < xMesh.m_uRTreeCount, "RTree Node Index is out of Range");
	
	return( &xMesh.m_pxRTree[ uiIndex]);
}

bool Asura_Direct3D_Carpet_Game_Render::MakeDataSet_RTreeBuilderToCGRRTree( Asura_RTree_Builder &xRTreeBuilder, Asura_Direct3D_CGR_Mesh& xMesh)
{
	std::vector< int32> xArrayRTreeBuilderToCGRIndex;
	// DLH - 31 January 2006 16:07:54 - Create the index conversion array with every value set to -1
	xArrayRTreeBuilderToCGRIndex.insert( xArrayRTreeBuilderToCGRIndex.begin(), xRTreeBuilder.GetNumRecords(), -1);
	
	const uint32 uiRTreeRecordCount = RTreeNode_CountPolygons( xRTreeBuilder, xArrayRTreeBuilderToCGRIndex);
	xMesh.m_uRTreeCount = uiRTreeRecordCount;
	xMesh.m_pxRTree = new Asura_Direct3D_CGR_RTreeNode[ xMesh.m_uRTreeCount];
	
	
	std::list< uint32 > xListRTreeNodeToProcess;
	xListRTreeNodeToProcess.push_back( xRTreeBuilder.GetRootIndex());
	
	uint32 uiCGRNodeCount = 0;
	
	while( 0 < xListRTreeNodeToProcess.size())
	{
		const uint32 uiNodeIndex = xListRTreeNodeToProcess.front();
		xListRTreeNodeToProcess.pop_front();
		
		if( -1 == uiNodeIndex)
		{
			continue;
		}
		
		const Asura_RTree_RawRecord* pxRTreeNode = xRTreeBuilder.GetRecord( uiNodeIndex);
		
		{
			const uint32 uiNodeChildIndex = pxRTreeNode->GetFirstChildIndex();
			xListRTreeNodeToProcess.push_back( uiNodeChildIndex);
		}
		
		{
			const uint32 uiNodeSiblingIndex = pxRTreeNode->GetSiblingIndex();
			xListRTreeNodeToProcess.push_back( uiNodeSiblingIndex);
		}
		
		{
			const uint32 uiCGRNodeIndex = uiCGRNodeCount;
			Asura_Assert( uiCGRNodeIndex < xMesh.m_uRTreeCount, "RTree miscount");
			Asura_Assert( (int)uiCGRNodeIndex == xArrayRTreeBuilderToCGRIndex[ uiNodeIndex], "Index translation table is not correct");
			
			Asura_Direct3D_CGR_RTreeNode& xCGRRTreeNode = xMesh.m_pxRTree[ uiCGRNodeIndex];
			
			int32 iRTreeChildIndex = pxRTreeNode->GetFirstChildIndex();
			int32 iRTreeSiblingIndex = pxRTreeNode->GetSiblingIndex();
			
			iRTreeChildIndex = -1 != iRTreeChildIndex ? xArrayRTreeBuilderToCGRIndex[ iRTreeChildIndex] : -1;
			iRTreeSiblingIndex = -1 != iRTreeSiblingIndex ? xArrayRTreeBuilderToCGRIndex[ iRTreeSiblingIndex] : -1;
			
			xCGRRTreeNode.m_pxChild = GetPointerFromIndex( xMesh, iRTreeChildIndex);
			xCGRRTreeNode.m_pxSibling = GetPointerFromIndex( xMesh, iRTreeSiblingIndex);
			xCGRRTreeNode.m_uData = pxRTreeNode->GetIndex();
			Asura_Assert( -1 == xCGRRTreeNode.m_uData || xCGRRTreeNode.m_uData < xMesh.m_uPolygonBatchCount, "Batch index is out of range");
			xCGRRTreeNode.m_xBoundingBox.SetToPoint( pxRTreeNode->GetMin());
			xCGRRTreeNode.m_xBoundingBox.Set( pxRTreeNode->GetMax());
		}
		
		uiCGRNodeCount += 1;
	}
	
	Asura_Assert( uiCGRNodeCount == xMesh.m_uRTreeCount, "RTree node miscount");
	return( true);
}

//-------------------------------------------------------------------
// Function:    Asura_Direct3D_Carpet_Game_Render::MakeDataSet_CalcMinimumVertexBufferSize
//
// Description: Search the Data for the minimum size vertex buffer needed to show the all the visible carpet from one location
//
// Parameters:  pxMesh          - 
//
// Returns:     
//-------------------------------------------------------------------
bool Asura_Direct3D_Carpet_Game_Render::MakeDataSet_CalcMinimumVertexBufferSize( Asura_Direct3D_CGR_Mesh* pxMesh)
{
	uint32 uMaxItemCount = 0;
	
	const uint32 uRTreeNodeCount = pxMesh->m_uRTreeCount;
	for( uint32 uRTreeNodeOriginalIndex = 0; uRTreeNodeOriginalIndex < uRTreeNodeCount; uRTreeNodeOriginalIndex += 1)
	{
		const Asura_Direct3D_CGR_RTreeNode& xRTreeNodeOriginal = pxMesh->m_pxRTree[ uRTreeNodeOriginalIndex];
		const uint32 uBatchOriginalIndex = xRTreeNodeOriginal.m_uData;
		
		if( -1 == uBatchOriginalIndex)
		{
			continue;
		}
		
		Asura_Bounding_Box xBoundingBoxOriginalBatch = xRTreeNodeOriginal.m_xBoundingBox;
		
		xBoundingBoxOriginalBatch.Grow( pxMesh->m_fLodFadeFar);
		
		uint32 uPresentItemCount = 0;
		
		for( uint32 uRTreeNodeSecondIndex = 0; uRTreeNodeSecondIndex < uRTreeNodeCount; uRTreeNodeSecondIndex += 1)
		{
			const Asura_Direct3D_CGR_RTreeNode& xRTreeNodeSecond = pxMesh->m_pxRTree[ uRTreeNodeSecondIndex];
			const uint32 uBatchSecondIndex = xRTreeNodeSecond.m_uData;
			
			if( -1 == uBatchSecondIndex)
			{
				continue;
			}
			
			const Asura_Direct3D_CGR_PolygonBatch& xSecondBatch = pxMesh->m_pxPolygonBatches[ uBatchSecondIndex];
			const Asura_Bounding_Box& xBoundingBoxSecondBatch = xRTreeNodeSecond.m_xBoundingBox;
			
			const bool bTouches = xBoundingBoxSecondBatch.Intersects( xBoundingBoxOriginalBatch);
			if( false == bTouches)
			{
				continue;
			}
			
			uint32 uIndexCount = 0;
			uint32 uVertexCount = 0;
			
			Asura_Direct3D_Carpet_Game_Render::BatchGetRequiredBufferSize( *pxMesh, xSecondBatch, uIndexCount, uVertexCount);

			uPresentItemCount += uVertexCount;
		}
		
		uMaxItemCount = Asura_Maths::Max( uMaxItemCount, uPresentItemCount);
	}
	
	pxMesh->m_uBatchVertexBufferSize = uMaxItemCount;
	return( true);
}

bool Asura_Direct3D_Carpet_Game_Render::MakeDataSet( const Asura_LinkList< Asura_Carpet_Triangle>& pxListCarpetTriangles)
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Direct3D_Carpet_Game_Render::MakeDataSet");
	
	ClearDataSet();
	
	MakeDataSet_Meshes( pxListCarpetTriangles);
	
	return( true);
}

bool Asura_Direct3D_Carpet_Game_Render::RelightEverything()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Direct3D_Carpet_Game_Render::RelightEverything");
	
	ClearDirect3DData_AllBuffers();
	
	const u_int uInvalidIndex = static_cast< uint32>( -1);
	
	for( uint32 uiMeshIndex = 0; uiMeshIndex < s_uMeshesCount; uiMeshIndex += 1)
	{
		Asura_Direct3D_CGR_Mesh& xCGRMesh = s_pxMeshes[ uiMeshIndex];
		
		const uint32 uVertexCount = xCGRMesh.m_uVertexCount;
		for( uint32 uVertexIndex = 0; uVertexIndex < uVertexCount; uVertexIndex += 1)
		{
			Asura_Direct3D_CGR_Vertex& xVertex = xCGRMesh.m_pxVertices[ uVertexIndex];
			
			#if 0
			if( uInvalidIndex == xVertex.m_uModule && uInvalidIndex == xVertex.m_uPoly && uInvalidIndex == xVertex.m_uVtx)
			{
				int				iModule = uInvalidIndex;
				int				iPoly = uInvalidIndex;
				int				iVtx = uInvalidIndex;
				Asura_Vector_3  xNormal;

				//if( true == Tools_Carpet_Mesh::FindScenePoly( pxCacheTriangle, uiCornerIndex, &iModule, &iPoly, &iVtx, &xNormal))
				{
				}
			}
			#endif

			if( uInvalidIndex != xVertex.m_uModule && uInvalidIndex != xVertex.m_uPoly && uInvalidIndex != xVertex.m_uVtx)
			{
				const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule( xVertex.m_uModule );

#ifndef ASURA_TARGET_SM3
				uint32 uColour = 0xFFFFFFFF;

				const u_int uPaintedColour = pxMesh->m_pxPaintedColourDataPtr[xVertex.m_uPoly].Colour[xVertex.m_uVtx];
				uColour = Tools_EnvironmentalLightTransport::GetDiffuseForVertex( xVertex.m_xPosition, xVertex.m_xNormal, xVertex.m_uPoly, xVertex.m_uVtx, xVertex.m_uModule, uPaintedColour );

				xVertex.m_xColourRGB[0] = (0xFF & (uColour >> 0)) / float( 0xFF);
				xVertex.m_xColourRGB[1] = (0xFF & (uColour >> 8)) / float( 0xFF);
				xVertex.m_xColourRGB[2] = (0xFF & (uColour >> 16)) / float( 0xFF);
#else				
				Asura_SphericalHarmonics_Order2_RGB xDiffuseSH;
				Asura_RGBA xDiffuse;
				
				int iVertID = pxMesh->PolygonDataPtr[xVertex.m_uPoly].VertexID[xVertex.m_uVtx];

				Tools_EnvironmentalLightTransport::GetDiffuseForVertex( xDiffuseSH, iVertID, xVertex.m_uModule );

                Tools_EnvironmentalLightTransport::UpdateLightingFromVertexPainting( xDiffuseSH, pxMesh, xVertex.m_uPoly, xVertex.m_uVtx );

				Asura_RGBA xColour;
				xDiffuseSH.EvaluateInDirection( xVertex.m_xNormal, xColour );

				xVertex.m_xColourRGB[0] = xColour.m_fR;
				xVertex.m_xColourRGB[1] = xColour.m_fG;
				xVertex.m_xColourRGB[2] = xColour.m_fB;

#endif

				
			}
		}
	}
		
	return( true);
}

bool Asura_Direct3D_Carpet_Game_Render::DeleteModule( const u_int uModuleIndex)
{
	return( true);
}

//*******************************************************************
//
// (c) 2006 Rebellion (Derby)
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************

#endif //defined( ASURA_USE_CARPET) && defined( ASURA_TOOLS)
