//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformTimers.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     PC implementation of Asura_Timers platform-specific functionality
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// independent includes
#include "Core/Asura_Timers.h"

// PC includes
#include "Core/Asura_PlatformBasics.h"
#include "Debugging/Asura_Assert.h"

#ifdef ASURA_USE_STREAMING
#include "Core/Asura_Multithread.h"
#endif //ASURA_USE_STREAMING


// DLH - 06 June 2007 11:22:50 - Storage and Initialise class for the PC's timer data
class Asura_PC_Timers_Data
{
public:
	Asura_PC_Timers_Data();
};

// G L O B A L S /////////////////////////////////////////////////////////////

static LARGE_INTEGER g_xFrequency;
static LARGE_INTEGER g_xLastTickCount;
static float g_fFrequency = 0.0f;
static HANDLE g_pTimer = NULL;
// DLH - 06 June 2007 11:23:12 - PC Timer Data. This will be initilaised in the global scope. Which means Asura_Debug_Timers can be used anywhere within the main scope with out having to check for the timer system to be initiliased. 
static Asura_PC_Timers_Data g_xPCTimerData;

Asura_PC_Timers_Data::Asura_PC_Timers_Data()
{
#if defined( ASURA_ASSERT)
	BOOL bQPFError = 
#endif
	QueryPerformanceFrequency(&g_xFrequency);
	Asura_Assert( 0 != bQPFError, "Error Initialising Query Performance Counter");

	g_fFrequency = static_cast<float>(g_xFrequency.QuadPart);
	Asura_Assert( 0.0f != g_fFrequency, "Error Initialising Asura_Timers, frequency is 0");

	g_pTimer = CreateWaitableTimer(NULL, TRUE, NULL);//"WaitableTimer");
	Asura_Assertf(g_pTimer, ("CreateWaitableTimer failed (%d)\n", GetLastError()));
}

void Asura_Timers::Platform_Initialise()
{
	// increase the granularity of timers from 10 to 1 ms
	timeBeginPeriod( 1 );
}


void Asura_Timers::Platform_Reset()
{
	QueryPerformanceCounter(&g_xLastTickCount);
}


void Asura_Timers::Platform_Shutdown()
{
	// must match each call to timeBeginPeriod with one of these for the same value.
	timeEndPeriod( 1 );
}


void Asura_Timers::Platform_Update( void )
{
    const float fEffectiveMinFrameTime = GetEffectiveMinFrameTime();
	if( fEffectiveMinFrameTime > 0.0f )
	{
        const LONGLONG iMinFrameTicks = static_cast<LONGLONG>( g_fFrequency * fEffectiveMinFrameTime );
		const LONGLONG iTargetTicks = g_xLastTickCount.QuadPart + iMinFrameTicks;

		// use Win32 sync objects to clamp framerate. This means we don't tie up
		// the whole CPU.
		LARGE_INTEGER xTime;
		if ( QueryPerformanceCounter(&xTime) && xTime.QuadPart < iTargetTicks )
		{
			LARGE_INTEGER liDueTime;
			liDueTime.QuadPart = -(static_cast<LONGLONG>((iTargetTicks - xTime.QuadPart) * (10000000.0f / g_fFrequency)));

            // Timer granularity is at least 1ms, even with our call to
            // timeBeginPeriod() at initialisation time, and so in the worst
            // case we could end  up not waking up from the timer wait until
            // 1ms or so after we needed to. We need to avoid waiting too
            // long (and ending up with a lower than intended average
            // framerate) and so we should therefore round off 1ms of the
            // wait time to compensate. Note that if we end up undershooting
            // the necessary wait time then the independent busy-wait code
            // will deal with the rest (and that ends up calling through to
            // a theoretically more accurate counter, ie. the QPC stuff).
            liDueTime.QuadPart += 10000;

            // Skip waiting at all if the initial amount of time was smaller
            // than the expected granularity.
            if( liDueTime.QuadPart < 0 )
            {
			    if (SetWaitableTimer(g_pTimer, &liDueTime, 0, NULL, NULL, 0))
			    {
				    // Wait for the timer .
				    if (WaitForSingleObject(g_pTimer, INFINITE) != WAIT_OBJECT_0)
				    {
					    Asura_Assertf(false, ("WaitForSingleObject failed (%d)\n", GetLastError()));
				    }
			    }
			    else
			    {
				    Asura_Assert(false, "SetWaitableTimer() failed");
			    }
            }
		}
	}
}


u_int Asura_Timers::Platform_GetTicksThisFrame()
{
	LARGE_INTEGER xNewTickCount;
	
	QueryPerformanceCounter(&xNewTickCount);

	u_int uTicksThisFrame = static_cast<u_int>(xNewTickCount.QuadPart - g_xLastTickCount.QuadPart);
	g_xLastTickCount = xNewTickCount;
	return uTicksThisFrame;
}


u_int Asura_Timers::Platform_GetTicksInFrame()
{
	LARGE_INTEGER xNewTickCount;
	QueryPerformanceCounter( &xNewTickCount );
    return static_cast< u_int >( xNewTickCount.QuadPart - g_xLastTickCount.QuadPart );
}


float Asura_Timers::Platform_GetTickFrequency()
{
	Asura_Assert( 0.0f != g_fFrequency, "Asura_Timers not setup yet");
	return g_fFrequency;
}


float Asura_Timers::Platform_GetDefaultFrameRate()
{
	return 60.0f;
}


Asura_TickCount Asura_Timers::Platform_GetCurTicks()
{
	LARGE_INTEGER xNewTickCount;
	QueryPerformanceCounter(&xNewTickCount);
	return xNewTickCount.QuadPart;
}

Asura_TickCount Asura_Timers::Platform_GetElapsedTicks(Asura_TickCount xStartTicks)
{
	LARGE_INTEGER xNewTickCount;
	QueryPerformanceCounter(&xNewTickCount);
	return xNewTickCount.QuadPart - xStartTicks;
}

float Asura_Timers::Platform_TicksToSeconds(Asura_TickCount xTicks)
{
	Asura_Assert( 0.0f != g_fFrequency, "Asura_Timers not setup yet");
	return static_cast<float>(xTicks) / g_fFrequency;
}

Asura_TickCount Asura_Timers::Platform_SecondsToTicks(float fSeconds)
{
	Asura_Assert( 0.0f != g_fFrequency, "Asura_Timers not setup yet");
	return static_cast<Asura_TickCount>(fSeconds * g_fFrequency);
}
