/**
*
* \file Asura_Logfile.cpp
*
* Created by Kevin Lea - 30/10/2001 09:30:55
*
* Copyright (C) 2001 Rebellion, All Rights Reserved.
*
* \brief Platform specific logfile functionality
*
**/

#include "Core/Asura.h"

// This file's header.
#include "Debugging/Asura_Logfile.h"

// Stdlib independent.
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_NameTree.h"
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_String.h"

#ifdef ASURA_TOOLS
// Tools.
#include "Tools_System.h"
#endif

#if defined( _DEBUG)
	// DLH - 10 April 2006 11:15:16 - Add this to allow the output of large items like HLSL Vertex shader when they do not compile
	static const int OUTPUT_BUFFER_SIZE = 40*1024;
#else
	static const int OUTPUT_BUFFER_SIZE = 8*1024;
#endif

static const Asura_Char* File;
static unsigned int Line;
static Asura_NameTree OpenedLogfiles;
static char* g_szLogfilesDir;
static char g_acLogfileDirStorage[MAX_PATH];

static bool Asura_LogfileClear(const Asura_Char* filename);
static bool TryAndCreateDirForLogfile( const Asura_Char* szFilename );

// to override logfiles dir, call this global function (using Asura_OverrideLogfilesDir()). 
// It will optionally backup the dir if it exists.
bool __Asura_OverrideLogfilesDir( const char* szDir, bool bArchiveExisting )
{
	// (MH - 21.02.2005) archive old dir
	if (bArchiveExisting)
	{
		struct stat xStat;
		int iResult = stat( szDir, &xStat );
		// exists?
		if (iResult == 0)
		{
			const char* szLogfileBackupDir = "logfile_backups";
			CreateDirectory( szLogfileBackupDir, NULL );

			const char* modifiedDate = ctime( &xStat.st_mtime );
			char acNewName[MAX_PATH];
			strcpy( acNewName, szLogfileBackupDir );
			strcat( acNewName, "\\" );
			strcat( acNewName, szDir );

			// append date
			strcat( acNewName, "_");
			strcat( acNewName, modifiedDate );
			// remove trailing '\n'
			acNewName[strlen(acNewName) - 1] = 0;

			// remove spaces
			while ( char* pcChar = strchr( acNewName, ' ' ) )
			{
				*pcChar = '_';
			}
			// remove :
			while ( char* pcChar = strchr( acNewName, ':' ) )
			{
				*pcChar = '-';
			}
			
			// rename logfiles dir to version with timestamp.
			MoveFile( szDir, acNewName );
		}
	}
		
	// set g_szLogfilesDir to point to version with trailing slash
	g_szLogfilesDir = g_acLogfileDirStorage;
	Asura_PrintfToBfr( g_acLogfileDirStorage, "%s\\", MAX_PATH, true, szDir );
	return true;
}

// return string with "logfiles" substituted with g_szLogfilesDir
// temporarily made global scope to allow external code to access the correct logfile dir.
const char* Asura_PC_FixupLogfileName(const char* szFilename, char* pcBuffer, u_int uBufferSize)
{
	if (!g_szLogfilesDir)
	{
		return szFilename;
	}

	const char* szDefaultDir = "logfiles\\";
	int iDefaultLen = int( strlen(szDefaultDir));
	if (_strnicmp( szFilename, szDefaultDir, iDefaultLen ) == 0)
	{
		Asura_PrintfToBfr( pcBuffer, "%s%s", uBufferSize, true, g_szLogfilesDir, szFilename + iDefaultLen );
		return pcBuffer;
	}
	// filename didn't contain "logfiles"
	return szFilename;
}

/// Kevin Lea - 30/10/2001 16:56:42 - wipe a file and timestamp it.
// (MP - 25.04.2002) Return true iff successful.
static bool Asura_LogfileClear(const Asura_Char* filename)
{
	char acBufferForLogFilename[MAX_PATH];
	filename = Asura_PC_FixupLogfileName(filename, acBufferForLogFilename, MAX_PATH);

    FILE* file = fopen( filename, "w" );

    // Did that fail because the directory doesn't exist?
    if( !file && errno == ENOENT )
    {
        if( TryAndCreateDirForLogfile(filename) )
        {
            file = fopen( filename, "w" );
        }
    }

	if( file )
	{
		char tbuffer[100];
		char tbuffer2[100];
		{
			SYSTEMTIME time_of_day;
			GetLocalTime( &time_of_day );

			GetDateFormat( GetThreadLocale(), DATE_LONGDATE, &time_of_day, NULL, tbuffer, 99 );
			GetTimeFormat( GetThreadLocale(), 0, &time_of_day, NULL, tbuffer2, 99 );
		}
		fprintf( file, "Asura Logfile \"%s\" - opened %s %s\n\n", filename, tbuffer, tbuffer2 );

		fclose( file );

        return true;
	}

    return false;
}

/// Kevin Lea - 30/10/2001 16:57:00 - helper function to set debugging data
void __Asura_LogfileDebugData(const Asura_Char* file, unsigned int line)
{
	File = file;
	Line = line;
}

/// Kevin Lea - 30/10/2001 16:57:16 - output info to logfile
void __Asura_LogfileOutput(const Asura_Char* logfilename, const Asura_Char* string, ...)
{
	Asura_Assert(logfilename, "Log filename is NULL");
	if (!logfilename) return;

	char acBufferForLogFilename[MAX_PATH];
	logfilename = Asura_PC_FixupLogfileName(logfilename, acBufferForLogFilename, MAX_PATH);

	/// Kevin Lea - 30/10/2001 16:41:33 - check to see if we've already opened a file called this.
	{
		int dummy;

		if (OpenedLogfiles.FindNodeIndexByName( logfilename, true, dummy ) == false)
		{
			/// Kevin Lea - 30/10/2001 16:48:46 - this file hasn't been seen before, so let's wipe it
			if( !Asura_LogfileClear(logfilename) )
            {
                // (MP - 25.04.2002) If we fail to clear the file, we won't
                // be able to write out to it below either so give up now.
                return;
            }
		}
	}

	char buffer[OUTPUT_BUFFER_SIZE];

	/// Kevin Lea - 30/10/2001 16:55:50 - process the format string and any other parameters
	/// into a buffer, just like printf does.
    // (MP - 25.04.2002) Truncate it if necessary.
	{
		va_list ap;
		va_start( ap, string );
		int iNumChars = _vsnprintf( buffer, OUTPUT_BUFFER_SIZE, string, ap );
		va_end( ap );

        if( -1 == iNumChars )
        {
            buffer[OUTPUT_BUFFER_SIZE - 1] = '\0';
        }
	}

	/// Kevin Lea - 30/10/2001 16:49:21 - open file for 'a'ppending
	FILE* fileHandle = fopen( logfilename, "a" );

	/// Kevin Lea - 30/10/2001 16:54:09 - if we could open the file, print out the contents
	/// of 'buffer'. Depending on the debug level, also output date/time, and the line no
	/// and file of the caller.
	if (fileHandle)
	{
		char tbuffer[100];
		char tbuffer2[100];
		{
			SYSTEMTIME time_of_day;
			GetLocalTime(&time_of_day);
	
			GetDateFormat( GetThreadLocale(), DATE_LONGDATE, &time_of_day, NULL, tbuffer, 99 );
			GetTimeFormat( GetThreadLocale(), 0, &time_of_day, NULL, tbuffer2, 99 );
		}	

		#ifdef NDEBUG
		fprintf( fileHandle, "%s\t(%s %s)\n\n", buffer, tbuffer, tbuffer2 );
		#else
		fprintf( fileHandle, "%s, Line %d:\n%s\t(%s %s)\n\n", File, Line, buffer, tbuffer, tbuffer2 );
		#endif
		fclose( fileHandle );
	}
}

// This version of the log clear function should not make use of any globals
// with constructors or any functions that use some. In particular, this
// function can't assert or use the OpenedLogfiles tree to check if the file's
// been used before or not.
void __Asura_SimpleLogfileClear(const Asura_Char* file)
{
    // (MP - 25.04.2002) At the moment, the standard clear function is ok.
    Asura_LogfileClear( file );
}

// This version of the logging function should not make use of any globals
// with constructors or any functions that use some. In particular, this
// function can't assert or use the OpenedLogfiles tree to check if the file's
// been used before or not.
void __Asura_SimpleLogfileOutput(const Asura_Char* logfilename, const Asura_Char* string, ...)
{
	Asura_Assert(logfilename, "Log filename is NULL");
	if (!logfilename) return;

	char acBufferForLogFilename[MAX_PATH];
	logfilename = Asura_PC_FixupLogfileName(logfilename, acBufferForLogFilename, MAX_PATH);

	char buffer[OUTPUT_BUFFER_SIZE];

	/// Kevin Lea - 30/10/2001 16:55:50 - process the format string and any other parameters
	/// into a buffer, just like printf does.
    // (MP - 25.04.2002) Truncate it if necessary.
	{
		va_list ap;
		va_start( ap, string );
		int iNumChars = _vsnprintf( buffer, OUTPUT_BUFFER_SIZE, string, ap );
		va_end( ap );

        if( -1 == iNumChars )
        {
            buffer[OUTPUT_BUFFER_SIZE - 1] = '\0';
        }
	}

	/// Kevin Lea - 30/10/2001 16:49:21 - open file for 'a'ppending
	FILE* fileHandle = fopen( logfilename, "a" );

    // Did that fail because the directory doesn't exist?
    if( !fileHandle && errno == ENOENT )
    {
        // Try and create the directory then.
        if( TryAndCreateDirForLogfile(logfilename) )
        {
            fileHandle = fopen( logfilename, "a" );
        }
    }

	/// Kevin Lea - 30/10/2001 16:54:09 - if we could open the file, print out the contents
	/// of 'buffer'.
	if( fileHandle )
	{
        fputs( buffer, fileHandle ); 
		fclose( fileHandle );
	}
}

#if defined( ASURA_TOOLS)
static const char* g_pszXMLHeader = "<?xml version=\"1.0\"?><asura>";
static const char* g_pszXMLFooter = "</asura>";
		
// This version of the log clear function should not make use of any globals
// with constructors or any functions that use some. In particular, this
// function can't assert or use the OpenedLogfiles tree to check if the file's
// been used before or not.
void __Asura_SimpleLogfileClearXML(const Asura_Char* filename)
{
	__Asura_SimpleLogfileOpenXML( filename);
	__Asura_SimpleLogfileCloseXML( filename);
}

void __Asura_SimpleLogfileOpenXML( const char* filename )
{
    FILE* fileHandle = fopen( filename, "w" );

    // Did that fail because the directory doesn't exist?
    if( !fileHandle && errno == ENOENT )
    {
        if( TryAndCreateDirForLogfile(filename) )
        {
            fileHandle = fopen( filename, "w" );
        }
    }

	if( fileHandle)
	{
		fputs( g_pszXMLHeader, fileHandle); 
		
		fclose( fileHandle);
	}
}

void __Asura_SimpleLogfileCloseXML( const char* szLogFilename )
{
	__Asura_SimpleLogfileOutput( szLogFilename, g_pszXMLFooter);
}

// This version of the logging function should not make use of any globals
// with constructors or any functions that use some. In particular, this
// function can't assert or use the OpenedLogfiles tree to check if the file's
// been used before or not.
// It is a very simple add XML node routine. It assumes that the document has a root node of "asura" and that the last X chracters are the closing of that node "</asura>"
// If this is not the case it will go wrong
void __Asura_SimpleLogfileOutputXML(const Asura_Char* logfilename, const Asura_Char* string, ...)
{
	Asura_Assert(logfilename, "Log filename is NULL");
	if (!logfilename) return;

	char buffer[OUTPUT_BUFFER_SIZE];

	/// Kevin Lea - 30/10/2001 16:55:50 - process the format string and any other parameters
	/// into a buffer, just like printf does.
    // (MP - 25.04.2002) Truncate it if necessary.
	{
		va_list ap;
		va_start( ap, string );
		int iNumChars = _vsnprintf( buffer, OUTPUT_BUFFER_SIZE, string, ap );
		va_end( ap );

        if( -1 == iNumChars )
        {
            buffer[OUTPUT_BUFFER_SIZE - 1] = '\0';
        }
	}

	// DLH - 02 April 2007 13:28:51 - Open the file to allow seeking and insertion
	FILE* fileHandle = fopen( logfilename, "r+" );
	bool bEmpty = false;
    // Did that fail because the directory doesn't exist?
    if( !fileHandle)
    {
		if( errno == ENOENT )
		{
			// Try and create the directory then.
			if( TryAndCreateDirForLogfile(logfilename) )
			{
			}
		}
		
		fileHandle = fopen( logfilename, "w+" );
		bEmpty = true;
	}
    
	/// Kevin Lea - 30/10/2001 16:54:09 - if we could open the file, print out the contents
	/// of 'buffer'.
	
	if( fileHandle )
	{
		{
			fseek( fileHandle, 0, SEEK_END);
			const long uFileSize = ftell( fileHandle);
			if( true == bEmpty || 0 == uFileSize)
			{
				// DLH - 02 April 2007 13:29:38 - Insert Header
				fputs( g_pszXMLHeader, fileHandle); 
			}
			else
			{
				// DLH - 02 April 2007 13:41:23 - Seek to the end of the contents of the root node
				const long uFooterSize = strlen( g_pszXMLFooter);
				fseek( fileHandle, uFileSize - uFooterSize, SEEK_SET);
			}
		}

		fputs( buffer, fileHandle); 
		fputs( g_pszXMLFooter, fileHandle); 
		
		fclose( fileHandle );
	}
}

void __Asura_SimpleLogfileConvertToXML(const Asura_Char* pszInput, Asura_Char* pszOutput, const u_int uOutputSize)
{
	const Asura_Char* pszOutputCharacterMax = pszOutput + uOutputSize - 1;
	Asura_Char* pszOutputCharacter = pszOutput;
	for( const Asura_Char* pszInputCharacter = pszInput; 0 != *pszInputCharacter; )
	{
		const Asura_Char* pszReplacement = NULL;
		
		if( '&' == *pszInputCharacter)
		{
			pszReplacement = "&amp;";
		}
		if( '\'' == *pszInputCharacter)
		{
			pszReplacement = "&apos;";
		}
		if( '\"' == *pszInputCharacter)
		{
			pszReplacement = "&quot;";
		}
		else if( '<' == *pszInputCharacter)
		{
			pszReplacement = "&lt;";
		}
		else if( '>' == *pszInputCharacter)
		{
			pszReplacement = "&gt;";
		}
		
		if( NULL != pszReplacement)
		{
			u_int uReplacementSize = strlen( pszReplacement);
			if( (pszOutputCharacterMax - uReplacementSize) <= pszOutputCharacter)
			{
				break;
			}
			
			strcpy( pszOutputCharacter, pszReplacement);
			pszOutputCharacter += uReplacementSize;
			pszInputCharacter += 1;
		}
		else
		{
			if( pszOutputCharacterMax <= pszOutputCharacter)
			{
				break;
			}
			
			pszOutputCharacter[0] = pszInputCharacter[0];
			pszOutputCharacter += 1;
			pszInputCharacter += 1;
		}
	}
	
	pszOutputCharacter[0] = '\0';
}
#endif

static bool TryAndCreateDirForLogfile( const Asura_Char* __szFilename )
{
    const Asura_Char* szSlash = strrchr( __szFilename, '\\' );
    if( szSlash )
    {
        Asura_Char szDir[MAX_PATH];

        int iNumChars = Asura_Maths::Min( MAX_PATH - 1, int( szSlash - __szFilename) );
        strncpy( szDir, __szFilename, iNumChars );
        szDir[iNumChars] = '\0';

        BOOL bOK = CreateDirectory( szDir, NULL );
        if( bOK )
        {
            return true;
        }
    }

    return false;
}

void __Asura_OutputToDebugger( const Asura_Char* __szFormat, ... )
{
	char buffer[OUTPUT_BUFFER_SIZE];

	/// Kevin Lea - 30/10/2001 16:55:50 - process the format string and any other parameters
	/// into a buffer, just like printf does.
    // (MP - 25.04.2002) Truncate it if necessary.
	int iNumChars = 0;
	{
		va_list ap;
		va_start( ap, __szFormat );
		iNumChars = _vsnprintf( buffer, OUTPUT_BUFFER_SIZE - 1, __szFormat, ap );
		va_end( ap );

        if( -1 == iNumChars )
        {
			iNumChars = OUTPUT_BUFFER_SIZE;
            buffer[OUTPUT_BUFFER_SIZE - 1] = '\0';
        }
	}

    // Output.
    OutputDebugString( buffer );

#if defined(ASURA_LOGFILES_ENABLED)
	// also put debugger output in logfile
	{
		const char* szDebuggerOutputLog = "logfiles\\debugger_output.txt";
		static bool ls_bFirstTime = true;
		if (ls_bFirstTime)
		{
			// clear the file the first time.
			ls_bFirstTime = false;
			Asura_LogfileClear(szDebuggerOutputLog);
		}
		__Asura_SimpleLogfileOutput_PreFormatted(szDebuggerOutputLog, buffer, iNumChars);
	}
#endif // #if defined(ASURA_LOGFILES_ENABLED)
}

#if defined(ASURA_LOGFILES_ENABLED)

bool __Asura_Logfile_CanLogYet( void )
{
#ifdef ASURA_TOOLS
	#ifndef _FONTMAKER
		return Tools_System::s_bIsCurrentDirectorySet;
	#else
		return true;
	#endif
#else
	return true;
#endif
}


void __Asura_SimpleLogfileOutput_PreFormatted( const char* szLogFilename, const char* szStr, u_int uLen )
{
	char acBufferForLogFilename[MAX_PATH];
	szLogFilename = Asura_PC_FixupLogfileName(szLogFilename, acBufferForLogFilename, MAX_PATH);

	/// Kevin Lea - 30/10/2001 16:49:21 - open file for 'a'ppending
	FILE* fileHandle = fopen( szLogFilename, "a" );

    // Did that fail because the directory doesn't exist?
    if( !fileHandle && errno == ENOENT )
    {
        // Try and create the directory then.
        if( TryAndCreateDirForLogfile(szLogFilename) )
        {
            fileHandle = fopen( szLogFilename, "a" );
        }
    }

	/// Kevin Lea - 30/10/2001 16:54:09 - if we could open the file, print out the contents
	/// of 'buffer'.
	if( fileHandle )
	{
        fputs( szStr, fileHandle ); 
		fclose( fileHandle );
	}
}

#endif // ASURA_LOGFILES_ENABLED