//
// Asura_PlatformFileAccess.cpp
//
// Created by Kevin Lea - 07/09/2001 16:21:41
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <windows.h>
#include <process.h>

// Engine independent.
#include "Core/Asura_List.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"

#ifdef ASURA_TOOLS
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Misc/Tools_IniFile.h"
#endif

#if defined( ASURA_PC_ALLOW_LOAD_FROM_SEARCHPATHS)
#include "Core/Asura_LinkList.h"
#endif

// Kevin Lea - 07/09/2001 16:22:31 - implementation note
//
// valid fileHandles are in the range 0 to (ASURA_MAX_NO_OF_OPEN_FILES-1) inclusive
// 

// Kevin Lea - 25/09/2001 10:32:29 - okay, I've changed the loader to read in the entire file into memory and then
// just memcpy the data as required, rather than using lots of separate freads. This gave a loading time for a test 
// environment of just 1.4s opposed to the original 4.2s (3 times faster).
//
// A further improvements might be to map the file as virtual memory, and remove the need to new/delete a large chunk of memory.
//
// (MP - 11.04.2003) Done.

struct OpenFileInfo
{
	union
	{
		struct
		{
			u_int bAsynchronous : 1;
			u_int bFinishedLoading : 1;
			u_int bErrorLoading : 1;
			u_int bAbort : 1;

		};
		u_int uAsyncFlags;
	};

    HANDLE hFile;    // Remainder of state is invalid iff this is INVALID_HANDLE_VALUE.
	HANDLE hFileMapping; //file mapping handle , not used by async loading
    void*  pvData;
    u_int  uPos;     // Bytes.
    u_int  uSize;    //   "
	u_int  uFileMappingStart; 
	u_int  uFileMappingEnd; 
	#ifdef ASURA_TOOLS
	bool   bClosedForLZX;
	#endif
};



static OpenFileInfo g_axOpenFiles[ASURA_MAX_NO_OF_OPEN_FILES];

//Maximum size of file mapping to create at a time
#ifdef ASURAIMPORT_EXPORTS
//Trying a small file mapping size with the max plugins , since memory is a bit tight on larger levels.
//RAR 08/08/2008
static const u_int g_uMAX_FILE_MAPPING_SIZE = 1024*1024;
#else
static const u_int g_uMAX_FILE_MAPPING_SIZE = 32*1024*1024;
#endif

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
static const char* FILE_SRC_LOGFILE = "logfiles\\file_sources.txt";

//path to shadow directories
static char g_szNetworkCfgFile[MAX_PATH] = "NetworkPath.cfg";
static char g_szNetworkDirectory[MAX_PATH] = "";

bool evilg_bUsingLocalFileNotOnNetwork;
#endif

static char g_szAsuraFilesDir[MAX_PATH] = "";
static int g_iAsuraFilesDirLength;
static bool g_bAsuraFilesDirSetExternally = false;

bool Asura_FileAccess::s_bNetworkDirectoryInitialised = false;

static void AsyncronousFileThread(void* pData);

#ifndef INVALID_FILE_ATTRIBUTES
#define  INVALID_FILE_ATTRIBUTES 0xffffffff
#endif

void Asura_FileAccess::Platform_Initialise( void )
{
    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        g_axOpenFiles[u].hFile = INVALID_HANDLE_VALUE;
    }
}

void Asura_FileAccess::Platform_Shutdown( void )
{
    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        Asura_Assertf( INVALID_HANDLE_VALUE == g_axOpenFiles[u].hFile, ("File with handle %u wasn't closed.", u) );
    }
}

bool Asura_FileAccess::Platform_DiscExists(void)
{
	return(true);
}

bool Asura_FileAccess::Platform_FileExists( const char* szFilename )
{
	const HANDLE	hFile	= ::CreateFile(szFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	if(INVALID_HANDLE_VALUE == hFile)
	{
		return(false);
	}
	else
	{
		::CloseHandle(hFile);
		return(true);
	}
}

u_int Asura_FileAccess::Platform_GetFileSize (const char* szFilename)
{
	const HANDLE	hFile	= ::CreateFile(szFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	if(INVALID_HANDLE_VALUE == hFile)
	{
		Asura_Assert(false, "Cannot open file to get the size of.");
		return(0);
	}
	else
	{
		const DWORD	uFileSize	= ::GetFileSize(hFile, NULL);

		::CloseHandle(hFile);
		return(uFileSize);
	}
}

u_int Asura_FileAccess::Platform_GetFileSize( const Asura_Handle iFileHandle )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid handle (%d) passed to Asura_FileAccess::Platform_GetFileSize().", iFileHandle) );
		return 0;
	}

	return ::GetFileSize( g_axOpenFiles[ iFileHandle ].hFile, NULL );
}

bool Asura_FileAccess::Platform_Open( Asura_Handle iHandle, const char* szFilename )
{
	HANDLE hFile = CreateFile( szFilename,
                               GENERIC_READ,
                               FILE_SHARE_READ,
                               NULL,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL );

    if( INVALID_HANDLE_VALUE == hFile ) { return false; }

	DWORD uFileSize = ::GetFileSize( hFile, NULL );
    if( INVALID_FILE_SIZE == uFileSize )
    {
        CloseHandle( hFile );
        return false;
    }

    HANDLE hFileMapping = CreateFileMapping( hFile, NULL, PAGE_READONLY, 0, 0, NULL );
    if( INVALID_HANDLE_VALUE == hFileMapping )
    {
        Asura_Assert(0,"CreateFileMapping() failed");
		CloseHandle( hFile );
        return false;
    }

    u_int uMappingSize = Asura_Maths::Min(static_cast<u_int>(uFileSize),g_uMAX_FILE_MAPPING_SIZE);
    void* pvFileData = MapViewOfFile( hFileMapping, FILE_MAP_READ, 0, 0, uMappingSize );


    if( !pvFileData )
    {
		// (MH - 26.08.2008) allow silent failure for zero-length files
        Asura_Assert(uMappingSize == 0,"MapViewOfFile() failed");
		CloseHandle( hFileMapping);
		CloseHandle( hFile );
        return false;
    }

    g_axOpenFiles[iHandle].uAsyncFlags  = 0;
    g_axOpenFiles[iHandle].hFile  = hFile;
    g_axOpenFiles[iHandle].hFileMapping  = hFileMapping;
    g_axOpenFiles[iHandle].pvData = pvFileData;
    g_axOpenFiles[iHandle].uPos   = 0;
    g_axOpenFiles[iHandle].uFileMappingStart = 0;
    g_axOpenFiles[iHandle].uFileMappingEnd = uMappingSize;
    g_axOpenFiles[iHandle].uSize  = uFileSize;

    return true;
}


bool Asura_FileAccess::Platform_LoadAsynchronous(Asura_Handle iHandle,const char* szFilename)
{
	HANDLE hFile = CreateFile( szFilename,
                               GENERIC_READ,
                               FILE_SHARE_READ,
                               NULL,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL );

    if( INVALID_HANDLE_VALUE == hFile ) { return false; }

	DWORD uFileSize = ::GetFileSize( hFile, NULL );
    if( INVALID_FILE_SIZE == uFileSize )
    {
        CloseHandle( hFile );
        return false;
    }

	//allocate buffer to read file into
	void* pvFileData = Asura_Allocate(uFileSize);
	if(!pvFileData)
	{
        CloseHandle( hFile );
        return false;
	}

    g_axOpenFiles[iHandle].uAsyncFlags  = 0;
    g_axOpenFiles[iHandle].bAsynchronous  = true;
    g_axOpenFiles[iHandle].hFile  = hFile;
    g_axOpenFiles[iHandle].pvData = pvFileData;
    g_axOpenFiles[iHandle].uPos   = 0;
    g_axOpenFiles[iHandle].uSize  = uFileSize;

	//start the loading process
	_beginthread(AsyncronousFileThread,0,&g_axOpenFiles[iHandle]);

    return true;
	
}


static void AsyncronousFileThread(void* pData)
{
	OpenFileInfo* pxOpenFile = (OpenFileInfo*) pData;

	char* pcBuffer = (char*) pxOpenFile->pvData;
	u_int uBytesLeft = pxOpenFile->uSize;

	while(!pxOpenFile->bAbort && uBytesLeft)
	{
		//read a small amount at a time , so we can respond to requests to abort
		
		u_int uBytesToRead = uBytesLeft;
		if(uBytesToRead > 4096) uBytesToRead = 4096;

		unsigned long uBytesRead = 0;
		ReadFile(pxOpenFile->hFile,pcBuffer,uBytesToRead,&uBytesRead,0);

		if(uBytesRead != uBytesToRead)
		{
			pxOpenFile->bErrorLoading = true;
			break;
		}
		else
		{
			uBytesLeft -= uBytesToRead;
			pcBuffer += uBytesToRead;
		}
	}

	pxOpenFile->bAbort = false;
	pxOpenFile->bFinishedLoading = !uBytesLeft;
}


enum ASURA_ASYNCLOADING_STATUS Asura_FileAccess::Platform_GetAsynchronousLoadingStatus(Asura_Handle iFileHandle)
{
    OpenFileInfo& xOpenFile = g_axOpenFiles[iFileHandle];
    Asura_Assert( INVALID_HANDLE_VALUE != xOpenFile.hFile, "No file open with specified handle." );
	Asura_Assert( xOpenFile.bAsynchronous,"This file isn't being loadad asyncronously");

	if(xOpenFile.bErrorLoading) return ASURA_ASYNCLOADING_STATUS_ERROR;
	if(xOpenFile.bFinishedLoading) return ASURA_ASYNCLOADING_STATUS_FINISHEDLOADING;

	return ASURA_ASYNCLOADING_STATUS_LOADING;
}


bool Asura_FileAccess::Platform_Close( Asura_Handle iFileHandle )
{
    OpenFileInfo& xOpenFile = g_axOpenFiles[iFileHandle];

	#ifdef ASURA_TOOLS
	if(xOpenFile.bClosedForLZX)
	{
		//The file handle has already been closed , but this second call to close is expected. RAR 08/01/2008
		Asura_Assert( INVALID_HANDLE_VALUE == xOpenFile.hFile, "File should already have been closed." );
		xOpenFile.bClosedForLZX = false;
		return false;
	}
	#endif

    Asura_Assert( INVALID_HANDLE_VALUE != xOpenFile.hFile, "No file open with specified handle." );

    if(xOpenFile.bAsynchronous)
	{
		xOpenFile.bAbort = true;
		while(!xOpenFile.bAbort) Sleep(10);

		Asura_Deallocate(xOpenFile.pvData);
    	int iClosedOK   = CloseHandle( xOpenFile.hFile );

		xOpenFile.pvData = NULL;
	    xOpenFile.hFile = INVALID_HANDLE_VALUE;

		return( iClosedOK!=0 );
	}
	else
	{
    	int iUnmappedOK = UnmapViewOfFile( xOpenFile.pvData );
    	int iClosedOK   = CloseHandle( xOpenFile.hFileMapping );
    	iClosedOK   &= CloseHandle( xOpenFile.hFile );

		xOpenFile.pvData = NULL;
	    xOpenFile.hFile = INVALID_HANDLE_VALUE;
	    xOpenFile.hFileMapping = INVALID_HANDLE_VALUE;

		return( iUnmappedOK && iClosedOK );
	}
}

#ifdef ASURA_TOOLS
void Asura_FileAccess::Platform_CloseFileForLZXDecompression( Asura_Handle iFileHandle )
{
    OpenFileInfo& xOpenFile = g_axOpenFiles[iFileHandle];
    Asura_Assert( INVALID_HANDLE_VALUE != xOpenFile.hFile, "No file open with specified handle." );

	Platform_Close(iFileHandle);
	//Flag that it was closed for this reason , so we know not to assert when closing for real the second time round
	xOpenFile.bClosedForLZX = true;
}
#endif //ASURA_TOOLS

bool Asura_FileAccess::Platform_SeekFile(Asura_Handle iFileHandle, u_int uPosition)
{
	if( uPosition < g_axOpenFiles[iFileHandle].uSize )
	{
		g_axOpenFiles[iFileHandle].uPos = uPosition;
		return true;
	}
	
	return false;
}

bool Asura_FileAccess::Platform_ReadData( Asura_Handle iFileHandle, void* pvDest, u_int uNumBytes )
{
    OpenFileInfo& xOpenFile = g_axOpenFiles[iFileHandle];
    Asura_Assert( INVALID_HANDLE_VALUE != xOpenFile.hFile, "No file open with specified handle." );

	if(xOpenFile.bAsynchronous)
	{
		//make sure its safe to read the buffer
		if(!xOpenFile.bFinishedLoading)
		{
			Asura_Assert(0,"Asyncronous loading hasn't finished yet");
			return false;
		}
	}

	//pvData may be NULL if there has been an error reading the file already. RAR 17/02/2006
	if(!xOpenFile.pvData) return false;


	if( xOpenFile.uPos + uNumBytes <= xOpenFile.uSize )
	{
		if(!xOpenFile.bAsynchronous)
		{
			u_char* pucDest = reinterpret_cast<u_char*>(pvDest);
			while(uNumBytes)
			{
				//move file mapping if start position isn't covered
				if(xOpenFile.uPos < xOpenFile.uFileMappingStart || xOpenFile.uPos >= xOpenFile.uFileMappingEnd)
				{
					#if defined ASURA_ASSERT
						int iUnmappedOkay =
					#endif
						UnmapViewOfFile( xOpenFile.pvData );

					Asura_Assert(iUnmappedOkay,"Error unmapping file");

					//File mapping start has to be on certain boundaries to work. Sticking to multiples of our
					//file mapping size ought to do the trick. RAR 17/02/2006
					xOpenFile.uFileMappingStart = (xOpenFile.uPos/g_uMAX_FILE_MAPPING_SIZE) * g_uMAX_FILE_MAPPING_SIZE;
					u_int uMappingSize = Asura_Maths::Min(xOpenFile.uSize-xOpenFile.uFileMappingStart,g_uMAX_FILE_MAPPING_SIZE);

					xOpenFile.uFileMappingEnd = xOpenFile.uFileMappingStart + uMappingSize;

				    xOpenFile.pvData = MapViewOfFile( xOpenFile.hFileMapping, FILE_MAP_READ, 0, xOpenFile.uFileMappingStart, uMappingSize );

					if(!xOpenFile.pvData)
					{
						//Failed to move file mapping , which is probably not so good.
						Asura_Assertf(0,("Error mapping view of file. Further attempts to read this file will fail.\nFile %s\nFile mapping start %u",GetFilename(iFileHandle),xOpenFile.uFileMappingStart));
						return false;
					}
				}
				
				//read as much as can be done , without going beyond end of mapping
				u_int uBytesToRead = Asura_Maths::Min(uNumBytes,xOpenFile.uFileMappingEnd - xOpenFile.uPos);
				memcpy(pucDest,reinterpret_cast<const u_char*>(xOpenFile.pvData) + (xOpenFile.uPos - xOpenFile.uFileMappingStart),uBytesToRead);

				xOpenFile.uPos += uBytesToRead;
				pucDest += uBytesToRead;
				uNumBytes -= uBytesToRead;
			}
			return true;

		}
		else
		{
			memcpy( pvDest, reinterpret_cast<const u_char*>(xOpenFile.pvData) + xOpenFile.uPos, uNumBytes );
			xOpenFile.uPos += uNumBytes;
			return true;
		}
	}
	else
	{
        // NJD 12/04/2005 - I've changed this to a console output warning, as it is valid for this
        // to happen when calling Asura_FileAccess::ReadLine
        Asura_OutputToDebugger( ( "WARNING: Reading beyond end of file %s. Position %u; file size %u; attempting to read %u.\n", GetFilename(iFileHandle), xOpenFile.uPos, xOpenFile.uSize, uNumBytes ) );
		return false;
	}
}

bool Asura_FileAccess::Platform_SkipUncompressedData( Asura_Handle iFileHandle, u_int uNumBytes )
{
	g_axOpenFiles[iFileHandle].uPos += uNumBytes;
	// JAT 10/09/07 - test if the new position is beyond the end of the file to catch bad things
	const u_int uPos = g_axOpenFiles[iFileHandle].uPos;
	const u_int uFileSize = g_axOpenFiles[iFileHandle].uSize;
	if ( uPos > uFileSize )
	{
		Asura_Assertf( false, ( "Can't skip past end of file uPos 0x%X bytes uFileSize 0x%X\n", uPos, uFileSize ) );
		return false;
	}
	return true;
}

static void DetermineNetworkAndCurrentDirectory()
{
	if( !g_bAsuraFilesDirSetExternally )
    {
	    GetCurrentDirectory( MAX_PATH, g_szAsuraFilesDir );
	    g_iAsuraFilesDirLength = int( strlen( g_szAsuraFilesDir ));
    }
	
	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	//read the network directory from NetwordPath.cfg
	HANDLE pFile = CreateFile( g_szNetworkCfgFile, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS, 0 );
	if(pFile == INVALID_HANDLE_VALUE) 
    {
        Asura_Assertf( false, ("Failed to find \"%s\". Local copies will be used for all files.", g_szNetworkCfgFile) );
        return;
    }

	//read the contents of the file into a buffer
	int iFileSize = ::GetFileSize(pFile,0);
 	unsigned long uBytesRead;
	char* pcBuffer = new char[iFileSize+1];
	pcBuffer[iFileSize] = 0;
	ReadFile(pFile,pcBuffer,iFileSize,&uBytesRead,0);
	CloseHandle(pFile);

	if((int)uBytesRead == iFileSize)
	{
		//scan for the start of the path
		char* pcPos = pcBuffer;
		while(*pcPos == ' ' || *pcPos == '\t' || *pcPos == '\r' || *pcPos == '\n') pcPos++;

		//copy the rest of the string
		strncpy(g_szNetworkDirectory,pcPos,MAX_PATH);

		//search for the last non-space character on the line to put in a terminator
		pcPos = g_szNetworkDirectory;
		char* pcLastNonSpace = pcPos;

		while(*pcPos && *pcPos != '\t' && *pcPos != '\r' && *pcPos != '\n')
		{
			if(*pcPos != ' ') pcLastNonSpace = pcPos;
			pcPos++;
		}
		*(pcLastNonSpace+1) = 0;
	}

	delete [] pcBuffer;

    
	// (MH - 24.06.2004) check we can open the shared directory, if not then use local copies
	{
        // Check it exists and is indeed a directory.
	    DWORD uAttributes = GetFileAttributes( g_szNetworkDirectory );
	    if( (INVALID_FILE_ATTRIBUTES == uAttributes) || !(uAttributes & FILE_ATTRIBUTE_DIRECTORY) )
	    {
	        Asura_Assertf( false, ("Failed to open dir \"%s\". Local copies will be used for all files.", g_szNetworkDirectory) );
			*g_szNetworkDirectory = 0;
            return;
        }
	}
	
	#endif //ASURA_PC_ALLOW_LOAD_FROM_NETWORK
}

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK

void Asura_FileAccess::PlatformGetNetworkDir( const char* szDir, char* szOutputBuffer, u_int uOutputBufferSize,bool bWarnIfOnlyLocal )
{
	Asura_Assert( szDir, "" );
	Asura_Assert( strlen(szDir) < uOutputBufferSize, "" );

	if(!s_bNetworkDirectoryInitialised)
	{
		DetermineNetworkAndCurrentDirectory();
		s_bNetworkDirectoryInitialised = true;
	}
	
    // If we haven't got a network directory or if the specified filename
    // already has the absolute network path specified then we'll just copy
    // to the output buffer.
	if( *g_szNetworkDirectory && strnicmp(szDir, g_szNetworkDirectory, strlen(g_szNetworkDirectory)) )
	{
	    const char* szRelativeDir = szDir;
	    //If the path appears to be a full path to a subdirectory of the executable directory ,
	    //cut off the beginning part to convert to a relative dir
	    if(!strnicmp(szDir,g_szAsuraFilesDir,g_iAsuraFilesDirLength))
	    {
		    szRelativeDir += g_iAsuraFilesDirLength;
		    if(*szRelativeDir == '\\' || *szRelativeDir == '/') szRelativeDir++;
	    }

	    //get the network dir
	    Asura_PrintfToBfr(szOutputBuffer,"%s\\%s",uOutputBufferSize,true,g_szNetworkDirectory,szRelativeDir);

        // Check it exists and is indeed a directory.
	    DWORD uAttributes = GetFileAttributes( szOutputBuffer );
	    if( INVALID_FILE_ATTRIBUTES != uAttributes && (uAttributes & FILE_ATTRIBUTE_DIRECTORY) )
	    {
            return;
        }

        // This assert is just misleading if the local directory doesn't
        // exist either so keep quiet in that case. Any errors resulting from
        // the local dir not existing should be handled externally.
		uAttributes = GetFileAttributes(szDir);
        if( INVALID_FILE_ATTRIBUTES != uAttributes && (uAttributes & FILE_ATTRIBUTE_DIRECTORY) && bWarnIfOnlyLocal)
        {
	        Asura_Assertf( false, ("Using local dir \"%s\" which does not exist on the network.", szDir) );
        }
    }

    // Copy over the input path.
	strcpy( szOutputBuffer, szDir );
}

void Asura_FileAccess::PlatformGetNetworkFilename( const char* szFilename, char* szOutputBuffer,u_int uOutputBufferSize )
{
	Asura_Assert( szFilename, "" );
	Asura_Assert( strlen(szFilename) < uOutputBufferSize, "" );
	// (MH - 20.08.2004) assert about files starting with a backslash as they behave 
	// differently according to whether NetworkPath.cfg exists. If it does they get 
	// Z:\\Project_Art\\Asura_Files prepended, if not the filename gets used as-is and
	// that means an initial backslash will go to the root of the current drive.
	Asura_Assert( (szFilename[0] != '\\') || (szFilename[1] == '\\'), "filenames should not start with \\");

    evilg_bUsingLocalFileNotOnNetwork = false;

    const char* szNetworkPath = PlatformGetNetworkPath();

    // If we haven't got a network path, give up and just copy across the
    // input filename.
	if( !*szNetworkPath )
	{
		strcpy( szOutputBuffer, szFilename );
		return;
	}

    const char* szRelativeFilename = PlatformRemoveNetworkOrExecutableDir( szFilename );

    bool bLocalCopyExists = true;
    bool bNetworkCopyExists = false;
    bool bUseNetworkCopy = false;
    FILETIME xNetworkFileTime = {0};

    // Get timestamp for the network copy, if one exists.
    {
	    Asura_PrintfToBfr( szOutputBuffer, "%s\\%s", uOutputBufferSize, true, szNetworkPath, szRelativeFilename );

		WIN32_FILE_ATTRIBUTE_DATA xNetFileAttribs = { 0};
        BOOL bAttributesStatus = GetFileAttributesEx( szOutputBuffer, GetFileExInfoStandard, &xNetFileAttribs);
        if( 0 != bAttributesStatus)
        {
			if( INVALID_FILE_ATTRIBUTES != xNetFileAttribs.dwFileAttributes && !(xNetFileAttribs.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
			{
				xNetworkFileTime = xNetFileAttribs.ftLastWriteTime;
				bNetworkCopyExists = true;
			}
        }
    }

    // Look for a local copy.
    {
	    Asura_PrintfToBfr( szOutputBuffer, "%s\\%s", uOutputBufferSize, true, g_szAsuraFilesDir, szRelativeFilename );

		WIN32_FILE_ATTRIBUTE_DATA xLocalFileAttribs = { 0};
		BOOL bAttributesStatus = GetFileAttributesEx( szOutputBuffer, GetFileExInfoStandard, &xLocalFileAttribs);
		DWORD uAttributes = (0 != bAttributesStatus) ? xLocalFileAttribs.dwFileAttributes : INVALID_FILE_ATTRIBUTES;
        // If that failed, try using the input filename as-is, to account for
        // opening stuff outside the main Asura files directory. Ignore input
        // filenames that are actually network paths.
        if( INVALID_FILE_ATTRIBUTES == uAttributes && strnicmp(szFilename, szNetworkPath, strlen(szNetworkPath)) )
        {
            strcpy( szOutputBuffer, szFilename );
			bAttributesStatus = GetFileAttributesEx( szOutputBuffer, GetFileExInfoStandard, &xLocalFileAttribs);
			uAttributes = (0 != bAttributesStatus) ? xLocalFileAttribs.dwFileAttributes : INVALID_FILE_ATTRIBUTES;
        }


        // Local copy available?
        if( INVALID_FILE_ATTRIBUTES != uAttributes && !(uAttributes & FILE_ATTRIBUTE_DIRECTORY) )
        {
            // Writeable local copies override network copies.
            if( uAttributes & FILE_ATTRIBUTE_READONLY )
            {
                if( bNetworkCopyExists )
                {
                    // Get timestamp for the local file.
                    FILETIME xLocalFileTime = xLocalFileAttribs.ftLastWriteTime;
					// If the network copy is newer, go with that.
					#ifdef ASURA_TOOLS
						if( Tools_FileSystem::ApproxCompareFileTime(&xLocalFileTime, &xNetworkFileTime) < 0 )
					#else
						if( CompareFileTime(&xLocalFileTime, &xNetworkFileTime) < 0 )
					#endif
					{
						bUseNetworkCopy = true;
					}
                }
                else
                {
		            Asura_Assertf( false, ("Loading local file \"%s\" which does not exist on the network.", szOutputBuffer) );
                }
            }

            evilg_bUsingLocalFileNotOnNetwork = !bNetworkCopyExists;
        }
        else
        {
            bUseNetworkCopy = bNetworkCopyExists;
            bLocalCopyExists = false;
        }
    }

    if( bUseNetworkCopy )
    {
        // Unfortunately we need to construct the full path again at this
        // point, since we're restricted to working in a single buffer.
	    Asura_PrintfToBfr( szOutputBuffer, "%s\\%s", uOutputBufferSize, true, szNetworkPath, szRelativeFilename );
    }

    // Log source used for the file. Don't bother if the file doesn't
    // actually exist, since that'll just be misleading.
    if( bLocalCopyExists || bNetworkCopyExists )
    {
        static bool bFirst = true;
        if( bFirst )
        {
            Asura_SimpleLogfileClear( FILE_SRC_LOGFILE );           
            bFirst = false;
        }

        Asura_SimpleLogfileOutput(( FILE_SRC_LOGFILE, "%s - %s\n", (bUseNetworkCopy ? "Network" : "Local  "), szRelativeFilename ));
    }
}

void Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy( const char* szFilename, char* szOutputBuffer,u_int uOutputBufferSize )
{
	PlatformGetNetworkFilename(szFilename,szOutputBuffer,uOutputBufferSize);

	#if defined(ASURA_TOOLS) && defined(ASURA_MAIN)

	static bool s_bFirstTime = true;
	static bool s_bGetExistingFiles = false;
	static u_int s_uGetFilesMin = 0;
	static u_int s_uGetFilesMax = 0;
	//Get user settings if this is the first time
	if(s_bFirstTime)
	{
		s_bFirstTime = false;
		//This needs to be done before the rest of the options file is used in Tools_OptionsFile , so
		//just read in the settings we care about here.
		Tools_IniFile xIni("ToolsIniData\\AsuraTools.ini",false);
		xIni.GetBoolInCurrentSection("Always Get Existing Files",s_bGetExistingFiles);

		float fTemp;
		if(xIni.GetFloatInCurrentSection("Get Files (Min)",fTemp))
		{
			s_uGetFilesMin = Asura_Maths::FloatToInt(1024*1024*fTemp);
		}
		if(xIni.GetFloatInCurrentSection("Get Files (Max)",fTemp))
		{
			if(fTemp>1024.0f) fTemp = 1024.0f;//clamp at 1 gigabyte for now...
			s_uGetFilesMax = Asura_Maths::FloatToInt(1024*1024*fTemp);
		}
	}
	
	if(!s_bGetExistingFiles && s_uGetFilesMin>=s_uGetFilesMax)
	{
		//Never want to fetch files from the network in this case , so forget about further tests
		//Well , you could get a file if Min and Max was the same , but it is more
		//convenient to skip in that case ,so both can be left at zero by default.
		return;
	}


	//file taken from network?
    const char* szNetworkPath = PlatformGetNetworkPath();
    if( !*szNetworkPath ) return;
    if(strnicmp( szOutputBuffer, szNetworkPath, strlen(szNetworkPath)))
	{
		//not taken from network ,so nothing more to do
		return;
	}


	//get size of network copy (and confirm existence...)
	WIN32_FILE_ATTRIBUTE_DATA xNetFileAttribs = { 0};
	BOOL bNetFileAttribsOkay = GetFileAttributesEx(szOutputBuffer, GetFileExInfoStandard, &xNetFileAttribs);
	if(!bNetFileAttribsOkay) return;
	u_int uFileSize = xNetFileAttribs.nFileSizeLow;


	//determine local filename
	const char* szLocalFilename = PlatformRemoveNetworkOrExecutableDir(szOutputBuffer);

	//if 'local' and 'network' files are the same , then skip this
	//(probably happen if loading file outside of Asura_Files for some reason)
	if(szLocalFilename == szOutputBuffer) return;

	//if local file is writeable , then leave alone
	//(typically the local file should have been chosen by PlatformGetNetworkFilename , but
	//it might not if the original filename given was expicitly a network path).
	WIN32_FILE_ATTRIBUTE_DATA xLocalFileAttribs = { 0};
	BOOL bLocalFileAttribsOkay = GetFileAttributesEx(szLocalFilename, GetFileExInfoStandard, &xLocalFileAttribs);

	if( bLocalFileAttribsOkay && !(xLocalFileAttribs.dwFileAttributes & FILE_ATTRIBUTE_READONLY)) return;

	if(!bLocalFileAttribsOkay || !s_bGetExistingFiles)
	{
		//Either a new file , or not set to get existing files by default.
		//So check if the file is within the size range.
		if(uFileSize<s_uGetFilesMin) return;
		if(uFileSize>s_uGetFilesMax) return;
	}




	//check filetimes , only interested in copying the network file if it is newer
	//(again , this is only likely to be an issue if th szFilename was a network path)
	if(bLocalFileAttribsOkay && bNetFileAttribsOkay)
	{
		if( Tools_FileSystem::ApproxCompareFileTime(&xLocalFileAttribs.ftLastWriteTime, &xNetFileAttribs.ftLastWriteTime) >= 0 )
		{
			return;//local file is more recent
		}
	}

	//Okay , we do want to copy the files.
	//Make sure the local file can be overwritten
	if(bLocalFileAttribsOkay)
	{
		SetFileAttributes(szLocalFilename,xLocalFileAttribs.dwFileAttributes &~ FILE_ATTRIBUTE_READONLY);
	}

	Tools_FileSystem::EnsureDirectoryExists(szLocalFilename);
	
	if(!CopyFile(szOutputBuffer,szLocalFilename,false))
	{
		
		Asura_Assertf(0,("Failed to copy %s to %s",szOutputBuffer,szLocalFilename));
		//copy failed for some reason , restore the readonly attribute
		if(bLocalFileAttribsOkay)
		{
			SetFileAttributes(szLocalFilename,xLocalFileAttribs.dwFileAttributes);
		}
		return;
	}

	Asura_SimpleLogfileOutput(( FILE_SRC_LOGFILE, "Updated Local - %s\n", szLocalFilename ));

	DWORD uAttributes = GetFileAttributes(szLocalFilename);
	if(INVALID_FILE_ATTRIBUTES != uAttributes)
	{
		//files in alienbrain shadows are hidden , this is probably undesirable for the files being got
		SetFileAttributes(szLocalFilename,uAttributes &~ FILE_ATTRIBUTE_HIDDEN);

		//finally update the output filename , so it uses the local file
		//(can't use strcpy since the buffers overlap)
		char* pcTo = szOutputBuffer;
		const char* pcFrom = szLocalFilename;
		while(*pcFrom) *pcTo++ = *pcFrom++;
		*pcTo++ = 0;
	}


	#endif //ASURA_TOOLS && ASURA_MAIN
}


const char* Asura_FileAccess::PlatformGetNetworkPath()
{
	if(!s_bNetworkDirectoryInitialised)
	{
		DetermineNetworkAndCurrentDirectory();
		s_bNetworkDirectoryInitialised = true;
	}
    return g_szNetworkDirectory;
}

void Asura_FileAccess::PlatformSetNetworkCfgFile( const char* szFilename )
{
    strncpy( g_szNetworkCfgFile, szFilename, MAX_PATH );
    g_szNetworkCfgFile[MAX_PATH - 1] = '\0';

    s_bNetworkDirectoryInitialised = false;
}

bool Asura_FileAccess::PlatformUsingNetworkFile( const char* szFilename )
{
    const char* szNetworkPath = PlatformGetNetworkPath();
    if( !*szNetworkPath ) { return false; }

    char szBuffer[MAX_PATH + 1];
    PlatformGetNetworkFilename( szFilename, szBuffer, MAX_PATH + 1 );

    return !strnicmp( szBuffer, szNetworkPath, strlen(szNetworkPath) );
}

#else // not ASURA_PC_ALLOW_LOAD_FROM_NETWORK

void Asura_FileAccess::PlatformGetNetworkDir( const char* szDir, char* szOutputBuffer, u_int uOutputBufferSize,bool bWarnIfOnlyLocal )
{
	Asura_Assert( szDir, "" );
	Asura_Assert( strlen(szDir) < uOutputBufferSize, "" );
	strcpy( szOutputBuffer, szDir );
}

void Asura_FileAccess::PlatformGetNetworkFilename( const char* szFilename, char* szOutputBuffer,u_int uOutputBufferSize )
{
	//network loading not enabled.. just return the original filename
	Asura_Assert(szFilename,"");
	Asura_Assert(strlen(szFilename)<uOutputBufferSize,"");
	strcpy(szOutputBuffer,szFilename);
}

void Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy( const char* szFilename, char* szOutputBuffer,u_int uOutputBufferSize )
{
	PlatformGetNetworkFilename(szFilename,szOutputBuffer,uOutputBufferSize);
}

bool Asura_FileAccess::PlatformUsingNetworkFile( const char* szFilename )
{
	return false;
}

#endif // else not ASURA_PC_ALLOW_LOAD_FROM_NETWORK

void Asura_FileAccess::PlatformSetAsuraFilesDir( const char* szPath )
{
    strncpy( g_szAsuraFilesDir, szPath, MAX_PATH );
    g_szAsuraFilesDir[MAX_PATH - 1] = '\0';

    g_iAsuraFilesDirLength = int( strlen( g_szAsuraFilesDir ));

    g_bAsuraFilesDirSetExternally  = true;
    s_bNetworkDirectoryInitialised = false;
}

const char* Asura_FileAccess::PlatformGetAsuraFilesDir()
{
	return g_szAsuraFilesDir;
}

const char* Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir( const char* szPath )
{
	if(!s_bNetworkDirectoryInitialised)
	{
		DetermineNetworkAndCurrentDirectory();
		s_bNetworkDirectoryInitialised = true;
	}

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
    // Check against network path.
    {
        const char* szNetworkPath = PlatformGetNetworkPath();
        if( *szNetworkPath && !strnicmp(szPath, szNetworkPath, strlen(szNetworkPath)) )
        {
            const char* szRet = szPath + strlen(szNetworkPath);
            return( *szRet == '\\' ? szRet + 1 : szRet );
        }
    }
#endif //ASURA_PC_ALLOW_LOAD_FROM_NETWORK

    // Check against executable dir.
	// replace any forward slashes
	Asura_String xTemp(szPath);
	xTemp.Replace("/", "\\");
	Asura_Assert( !strstr(g_szAsuraFilesDir, "/" ), "not expecting '/' in executable path");
    if( !strnicmp(xTemp.GetString(), g_szAsuraFilesDir, g_iAsuraFilesDirLength) )
    {
        const char* szRet = szPath + g_iAsuraFilesDirLength;
        return( *szRet == '\\' ? szRet + 1 : szRet );
    }

    // Must have been relative in the first place then, or maybe an absolute
    // path to somewhere else entirely.
    return szPath;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_FileAccess::Platform_OpenFromMemory().
// PARAMETERS   iFileHandle - effectively, index into open file array
//              pMem        - pointer to buffer containing file contents
//              uSize       - size of buffer
// RETURNS      false, as we don't support this method on this platform.
// DESCRIPTION  sets up file reading from memory that already contains a file.
// AUTHOR       Mike Healey - 17.08.2005.
//
//////////////////////////////////////////////////////////////////////////////
bool Asura_FileAccess::Platform_OpenFromMemory( Asura_Handle iFileHandle, const void* pMem, u_int uSize )
{
	return false;
}
// End function Asura_FileAccess::Platform_OpenFromMemory().


#if defined( ASURA_PC_ALLOW_LOAD_FROM_SEARCHPATHS)

class Asura_FileAccess_SearchPath
{
public:
	Asura_FileAccess_SearchPath();
	~Asura_FileAccess_SearchPath();
	
	void SetPath( const char* pszPath);
	const char* GetPath( void) const;
	
private:
//	Asura_FileAccess_SearchPath( const Asura_FileAccess_SearchPath& xRHS);
//	Asura_FileAccess_SearchPath& operator=( const Asura_FileAccess_SearchPath& xRHS);
	
	char m_szPath[ MAX_PATH];
};

Asura_FileAccess_SearchPath::Asura_FileAccess_SearchPath()
{
}

Asura_FileAccess_SearchPath::~Asura_FileAccess_SearchPath()
{
}

void Asura_FileAccess_SearchPath::SetPath( const char* pszPath)
{
	Asura_PrintfToBfr( this->m_szPath, "%s", sizeof( this->m_szPath), true, pszPath);
}

const char* Asura_FileAccess_SearchPath::GetPath( void) const
{
	return( this->m_szPath);
}

static Asura_LinkList< Asura_FileAccess_SearchPath> g_xList_xSearchPaths;

void Asura_FileAccess::PlatformSearchPathsClear( void)
{
	g_xList_xSearchPaths.Clear();
}

void Asura_FileAccess::PlatformSearchPathsAddSearchPath( const char* pszPath)
{
	Asura_FileAccess_SearchPath xSearchPath;
	
	xSearchPath.SetPath( pszPath);
	
	g_xList_xSearchPaths.Push_Back( xSearchPath);
}

void Asura_FileAccess::PlatformSearchPathsGetExistingFilename( const char* pszFilename, char* szOutputBuffer, u_int uOutputBufferSize )
{
	for( Asura_ListElt< Asura_FileAccess_SearchPath>* pxIterator = g_xList_xSearchPaths.Front(); NULL != pxIterator; pxIterator = pxIterator->Next())
	{
		const Asura_FileAccess_SearchPath& xSearchPath = pxIterator->Data();

		const u_int uFullFilenameSize = MAX_PATH;
		char szFullFilename[ uFullFilenameSize + 1];
		
		Asura_PrintfToBfr( szFullFilename, "%s%s", uFullFilenameSize, true, xSearchPath.GetPath(), pszFilename);
		
		bool bFileExists = Asura_FileAccess::FileExists( szFullFilename);
		if( false == bFileExists)
		{
			continue;
		}
		
		Asura_PrintfToBfr( szOutputBuffer, "%s", uOutputBufferSize, true, szFullFilename);
		return;
	}
	
	Asura_Assertf( false, ("Cannot find file '%s' in any of the search paths", pszFilename));
	Asura_PrintfToBfr( szOutputBuffer, "%s", uOutputBufferSize, true, pszFilename);
}

#endif//ASURA_PC_ALLOW_LOAD_FROM_SEARCHPATHS

static void FindDirectories( const char* szSearchString, Asura_List<Asura_String>& xOutputList )
{
	Asura_Assert( szSearchString && *szSearchString, "" );

	Asura_List_ReverseIt<Asura_String> xIt( &xOutputList );
	//xIt.end();

	char szBuffer[MAX_PATH];
	WIN32_FIND_DATA xFindData;

	const int iInitialNumEntries = xOutputList.GetNumEntries();
	for( int iDir = 0; iDir < 1 + (xOutputList.GetNumEntries() - iInitialNumEntries); ++iDir )
	{
		char* szFilenameStart;

		//copy the search string into a buffer for buidling the filenames , 
		//and work out where the filename part should be copied
		{
			if( iDir )
			{
				if(iDir==1 && iInitialNumEntries==0)
				{
					//If there were no entries in the list to start with , then the iterator will currently be
					//on the sentinel , so we need to restart to get to the first of the newly added entries.
					//If not , then the iterator will be on the last of the original entries , so using next()
					//will get us to the first newly added entry.RAR 17/12/03
					xIt.Restart();
				}
				else
				{
					xIt.Next();
				}

				const Asura_String& xNext = *xIt.GetCurrent();
				strcpy( szBuffer, xNext.GetString() );
				strcpy( &szBuffer[xNext.GetLength()], "\\*.*" );

			}
			else
			{
				strcpy( szBuffer, szSearchString );
			}

			char* szSlash = strrchr( szBuffer, '\\' );
			if( !szSlash )
			{
				szSlash = strrchr( szBuffer, '/' );
			}
			szFilenameStart = ( szSlash ? szSlash + 1 : szBuffer );
		}

		HANDLE hFindFile = ::FindFirstFile( szBuffer, &xFindData );
		if( INVALID_HANDLE_VALUE != hFindFile )
		{
			do
			{	
				if( xFindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
				{

					// Ignore the "." and ".." directories.
					if( '.' == xFindData.cFileName[0] )
					{
						if( !xFindData.cFileName[1] || ('.' == xFindData.cFileName[1] && !xFindData.cFileName[2]) )
						{
							continue;
						}
					}

					strcpy( szFilenameStart, xFindData.cFileName );
					xOutputList.AddToList( new Asura_String(szBuffer) );
				}
			}
			while( ::FindNextFile(hFindFile, &xFindData) );

			::FindClose( hFindFile );
		}
	}
}

void Asura_FileAccess::Platform_FindFiles(const char* szDirectory, const char* szExtension, void (*pfnCallback)(const char* szFilename), bool bRecursive /* = true */)
{
	// Check the strings have the right formatting. Could insert the slash or dot manually here, but
	// that means more Asura_String work which we might as well avoid
	{
		#if defined ASURA_ASSERT
			bool bSlash = Asura_StringUtil::EndsWith(szDirectory, "\\", false);
			bSlash = bSlash || Asura_StringUtil::EndsWith(szDirectory, "/", false);

			const bool bDot = (szExtension[0] == '.');
		#endif

		Asura_Assertf(bSlash, ("\"%s\" must end with a slash.", szDirectory));
		Asura_Assertf(bDot, ("\"%s\" must start with a dot.", szExtension));
	}

	Asura_String xStrWildcard(szDirectory);
	xStrWildcard += "*";
	xStrWildcard += szExtension;
	const char* szFullWildCard = xStrWildcard.GetString();

	Asura_List<Asura_String> xDirList;

	const char* szWildcard;

	{
		const char* szSlash = strrchr( szFullWildCard, '\\' );
		if( !szSlash )
		{
			szSlash = strrchr( szFullWildCard, '/' );
		}

		char szTempBuffer[MAX_PATH];

		if( szSlash && szSlash > szFullWildCard )
		{
			const int iLen = szSlash - szFullWildCard;
			strncpy( szTempBuffer, szFullWildCard, iLen );
			szTempBuffer[iLen] = '\0';
			xDirList.AddToList( new Asura_String(szTempBuffer) );
		}
		else
		{
			xDirList.AddToList( new Asura_String(szFullWildCard) );
		}

		if( bRecursive )
		{
			const Asura_String& xRoot = *xDirList.GetHeadObject();
			strcpy( szTempBuffer, xRoot.GetString() );
			strcpy( &szTempBuffer[xRoot.GetLength()], "\\*.*" );
			FindDirectories( szTempBuffer, xDirList );
		}

		if( szSlash && *(szSlash + 1) )
		{
			szWildcard = szSlash + 1;
		}
		else
		{
			szWildcard = szFullWildCard;
		}
	}

	//    const int iLengthOfWildcard = strlen( szWildcard );

	char szBuffer[MAX_PATH];

	for( Asura_List_It<Asura_String> xIt(&xDirList); !xIt.Done(); xIt.Next() )
	{
		char* szFilenameStart;

		//copy the search string into a buffer for buidling the filenames , 
		//and work out where the filename part should be copied
		{
			const Asura_String& xNext = *xIt.GetCurrent();
			const int iLen = xNext.GetLength();

			strcpy( szBuffer, xNext.GetString() );
			szBuffer[iLen] = '\\';
			strcpy( &szBuffer[iLen + 1], szWildcard );

			szFilenameStart = &szBuffer[iLen + 1];
		}

		WIN32_FIND_DATA xFindData;
		HANDLE hFindFile = ::FindFirstFile( szBuffer, &xFindData );
		if( INVALID_HANDLE_VALUE != hFindFile )
		{
			do
			{	
				if( !(xFindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
				{
					// (MP - 17.03.2006) Make sure this really does match
					// the wildcard. The file search stuff in Windows does
					// some slightly odd stuff in specific cases. For example,
					// *.asr will match *.asrBE, though bizarrely
					// *.a will not match *.asr. I think it's related to
					// 3 letter extensions somewhow.
					if( Asura_StringUtil::CompareAgainstWildcardString(xFindData.cFileName, szWildcard, true) )
					{
						strcpy( szFilenameStart, xFindData.cFileName );
						pfnCallback(szBuffer);
					}
				}
			}
			while( ::FindNextFile(hFindFile, &xFindData) );

			::FindClose( hFindFile );
		}
	}
	//delete strings from xDirList
	for( Asura_List_It<Asura_String> xIt(&xDirList); !xIt.Done(); xIt.Next() )
	{
		delete (xIt.GetCurrent());
	}
}
/// End of file. ///
