//*******************************************************************
// File: Asura_PC_Job_Manager.cpp
//
// Created: 25 June 2007 DLH
//
// Description: 
//
//*******************************************************************

// system libraries
#include "Core/Asura.h"

#if defined( ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)

// main header
#include "Jobs/Asura_Job_Manager.h"

// other libraries
#include "Core/Asura_Atomic.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_LogFile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Jobs/Asura_Job.h"
#include "Jobs/Asura_Job_TaskQueue.h"

// internal dependencies

//*******************************************************************
// Constants
//*******************************************************************

// DLH - Turn on defines for ThreadProfiling applications
#if 0 && defined( ASURA_SUPPORT_EXTERNAL_THREAD_PROFILING)
	// DLH - Turn on support for VTune ThreadProfiling API
	#define ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER
#endif

#if 0 || defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
	#include <libittnotify.h>
	#pragma comment( lib, "libittnotify.lib")
#endif

//*******************************************************************
// Local Classes
//*******************************************************************
class Asura_PC_Job_Manager_Data;

// DLH - Description of each job
class Asura_PC_Job_Manager_Job_Data
{
public:
	Asura_PC_Job_Manager_Job_Data();
	Asura_PC_Job_Manager_Job_Data( Asura_Job_Queue* pxJobQueue, Asura_Job_TaskQueue* pxTaskQueue);
	
	void JobFinished( void) const;
	void ExecuteJob( void) const;
	
public:

	Asura_Job_Queue*					m_pxJobQueue;
	Asura_Job_TaskQueue*				m_pxTaskQueue;
};

// DLH - One per thread
class Asura_PC_Job_Manager_Thread_Data
{
public:
	Asura_PC_Job_Manager_Thread_Data();
	~Asura_PC_Job_Manager_Thread_Data();
	
public:
	HANDLE	m_xThreadHandle;
	DWORD	m_dwThreadID;
	
	Asura_PC_Job_Manager_Data* m_pxThreadManager;
};

// DLH - Main Working class
class Asura_PC_Job_Manager_Data
{
public:
	Asura_PC_Job_Manager_Data();
	~Asura_PC_Job_Manager_Data();
	
	void Initialise();
	void Shutdown();
	
	void AddJobCountToQueue( const u_int uJobsCount);
	void RemoveJobCountFromQueue( const u_int uJobsCount);
	
	//void AddJob( Asura_Job_Manager_Job_Function pfnJobFuncation, void* pJobData);
	void AddJob( Asura_Job_Queue* pxJobQueue, Asura_Job_TaskQueue* pxJobTaskQueue);
	
	u_int GetNumUnfinishedJobs() const;

	void WaitForAllJobsToFinish();
	
	u_int GetNumWorkerThreads();
	
private:
	static long WINAPI WorkerThread( void* pThreadData);
	
	void AddJob( const Asura_PC_Job_Manager_Job_Data& xJobData);

public:

private:
	Asura_Job_TaskQueue m_xJobQueue;
	

	CRITICAL_SECTION					m_xJobsCriticalSection;
	volatile	u_int					m_uNumPendingJobs;
	HANDLE								m_xNewJobEvent;
	HANDLE								m_xExitThreadEvent;
	
	u_int								m_uNumWorkerThreads;
	HANDLE*								m_paxWorkerThreads_Handles;
	Asura_PC_Job_Manager_Thread_Data*	m_paxWorkerThreads_Data;
	
	bool								m_bAllowTheading;
	
	u_int								m_uQueueTasks;
	u_int								m_uFreeTasks;
	
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
	__itt_event m_xVTTPAddTask;
	__itt_event m_xVTTPExecuteTask;
	__itt_event m_xVTTPExecuteQueue;
	__itt_event m_xVTTPWaitForSpace;
	__itt_event m_xVTTPWaitForFinish;
#endif
};

//*******************************************************************
// Local Datas
//*******************************************************************

// DLH - Instance of main working class
static Asura_PC_Job_Manager_Data s_xJobManagerData;

//*******************************************************************
// Function Definitions
//*******************************************************************

//--------------------------------------
// Asura_PC_Job_Manager_Job_Data
//--------------------------------------

Asura_PC_Job_Manager_Job_Data::Asura_PC_Job_Manager_Job_Data( Asura_Job_Queue* pxJobQueue, Asura_Job_TaskQueue* pxTaskQueue) :
	m_pxJobQueue( pxJobQueue),
	m_pxTaskQueue( pxTaskQueue)
{
	
}

Asura_PC_Job_Manager_Job_Data::Asura_PC_Job_Manager_Job_Data() :
	m_pxJobQueue( NULL),
	m_pxTaskQueue( NULL)
{
	
}

void Asura_PC_Job_Manager_Job_Data::ExecuteJob( void) const
{
	Asura_Job_Queue* pxJobQueue = this->m_pxJobQueue;
	if( NULL != pxJobQueue)
	{
		Asura_Job_TaskQueue* pxTaskQueue = this->m_pxTaskQueue;
		
		Asura_Job_TaskQueue::Task* pxTask;
		bool bTryPopStatus = pxTaskQueue->PopTask( pxTask);
		
		if( true == bTryPopStatus)
		{
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			char szVTuneTitle[64];
			Asura_PrintfToBfr( szVTuneTitle, "PCJobQueueExecuteTask '%s'", sizeof( szVTuneTitle), true, pxTaskQueue->GetName());

			__itt_event xVTTPExecuteTask = __itt_event_create( szVTuneTitle, strlen( szVTuneTitle) );
			__itt_event_start( xVTTPExecuteTask);
#endif
	
			pxJobQueue->ExecuteJob( pxTask->GetData());

			pxJobQueue->JobFinished();
			
			
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_end( xVTTPExecuteTask);
#endif
			
			pxTaskQueue->ReturnTask( pxTask);
		}
		return;
	}
}

void Asura_PC_Job_Manager_Job_Data::JobFinished( void) const
{
	Asura_Job_Queue* pxJobQueue = this->m_pxJobQueue;
	if( NULL != pxJobQueue)
	{
		pxJobQueue->JobFinished();
	}
}

//--------------------------------------
// Asura_PC_Job_Manager_Thread_Data
//--------------------------------------

Asura_PC_Job_Manager_Thread_Data::Asura_PC_Job_Manager_Thread_Data() :
	m_xThreadHandle( NULL),
	m_dwThreadID( 0),
	m_pxThreadManager( NULL)
{
}

Asura_PC_Job_Manager_Thread_Data::~Asura_PC_Job_Manager_Thread_Data()
{
}

//--------------------------------------
// Asura_PC_Job_Manager_Data
//--------------------------------------

Asura_PC_Job_Manager_Data::Asura_PC_Job_Manager_Data() :
	m_xNewJobEvent( NULL),
	m_xExitThreadEvent( NULL),
	m_paxWorkerThreads_Handles( NULL),
	m_paxWorkerThreads_Data( NULL),
	m_bAllowTheading( true),
	m_xJobQueue( 100, sizeof( Asura_PC_Job_Manager_Job_Data))
{

}

Asura_PC_Job_Manager_Data::~Asura_PC_Job_Manager_Data()
{
	Asura_Assert( NULL == m_paxWorkerThreads_Data, "Worker Threads Data not deallocated");
	Asura_Assert( NULL == m_paxWorkerThreads_Handles, "Worker Threads Handles not deallocated");
}

void Asura_PC_Job_Manager_Data::Initialise()
{
	this->m_uQueueTasks = 0;
	this->m_uFreeTasks = 0;

	// Init pending job count
	InitializeCriticalSectionAndSpinCount( &this->m_xJobsCriticalSection, 1000);
	this->m_uNumPendingJobs = 0;

	// Create the 'new job' event
	this->m_xNewJobEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
	this->m_xExitThreadEvent = CreateEvent( NULL, TRUE, FALSE, NULL);

	// Create worker threads
	// How many?  Will depend on platform. On PC, check how many cores we've got. On 360,
	// it'll be fixed.		
	SYSTEM_INFO xSystemInfo;
	GetSystemInfo( &xSystemInfo);
	this->m_uNumWorkerThreads = xSystemInfo.dwNumberOfProcessors-1;
	
	// DLH - 25 June 2007 15:34:34 - Create Thread Data
	this->m_paxWorkerThreads_Data = new Asura_PC_Job_Manager_Thread_Data[ this->m_uNumWorkerThreads];
	this->m_paxWorkerThreads_Handles = new HANDLE[ this->m_uNumWorkerThreads];
	
	Asura_OutputToDebugger(("PC Job Manager creating %d worker thread(s) [SystemInfo.dwNumberOfProcessors = %d]\n\n", this->m_uNumWorkerThreads, xSystemInfo.dwNumberOfProcessors));
	
	SetThreadIdealProcessor( GetCurrentThread(),0);
	for( u_int uThreadIndex = 0; uThreadIndex < this->m_uNumWorkerThreads; uThreadIndex++)
	{		
		Asura_PC_Job_Manager_Thread_Data& xThreadData = this->m_paxWorkerThreads_Data[ uThreadIndex];
		
		xThreadData.m_pxThreadManager = this;
		
		HANDLE& xThreadHandle = this->m_paxWorkerThreads_Handles[ uThreadIndex];
		
		xThreadHandle = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)WorkerThread, &xThreadData, 0, &xThreadData.m_dwThreadID);
		//xThreadHandle = xThreadHandle = _beginthread( NULL, 0, (LPTHREAD_START_ROUTINE)WorkerThread, (void*)&xThreadData, 0, &xThreadData.m_dwThreadID);
		if( NULL == xThreadHandle)
		{
			continue;
		}
		
		xThreadData.m_xThreadHandle = xThreadHandle;
		
		SetThreadIdealProcessor( xThreadData.m_xThreadHandle, uThreadIndex+1);
	}
	
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
	{
		char szVTuneTitleAddTask[] = "PCJobQueueAddTask";
		this->m_xVTTPAddTask = __itt_event_create( szVTuneTitleAddTask, strlen( szVTuneTitleAddTask) );
		char szVTuneTitleExecuteTask[] = "PCJobQueueExecuteTask";
		this->m_xVTTPExecuteTask = __itt_event_create( szVTuneTitleExecuteTask, strlen( szVTuneTitleExecuteTask) );
		char szVTuneTitleExecuteQueue[] = "PCJobQueueExecuteQueue";
		this->m_xVTTPExecuteQueue = __itt_event_create( szVTuneTitleExecuteQueue, strlen( szVTuneTitleExecuteQueue) );
		char szVTuneTitleWaitForSpace[] = "PCJobQueueWaitForSpace";
		this->m_xVTTPWaitForSpace = __itt_event_create( szVTuneTitleWaitForSpace, strlen( szVTuneTitleWaitForSpace) );
		char szVTuneTitleWaitForFinish[] = "PCJobQueueWaitForFinish";
		this->m_xVTTPWaitForFinish = __itt_event_create( szVTuneTitleWaitForFinish, strlen( szVTuneTitleWaitForFinish) );
	}
#endif

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	Asura_CommandConsole::AddVar( L"CPU.MultiThreading.JobQueue.Threaded", &(this->m_bAllowTheading), Asura_CommandConsole::REMOVE_AT_SHUTDOWN);	
	Asura_CommandConsole::AddVar( L"CPU.MultiThreading.JobQueue.QueueTasks", &(this->m_uQueueTasks), 0, 99999, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);	
	Asura_CommandConsole::AddVar( L"CPU.MultiThreading.JobQueue.FreeTasks", &(this->m_uFreeTasks), 0, 99999, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);	
#endif
}

void Asura_PC_Job_Manager_Data::Shutdown()
{
	{
		// DLH - 25 June 2007 16:45:22 - Tell all the threads to quit
		SetEvent( this->m_xExitThreadEvent);
		
		// DLH - 25 June 2007 17:24:03 - Wait for all the threads to die
		u_int uNumWorkerThreadsTotal = this->m_uNumWorkerThreads;
		HANDLE* pxWorkerThreadsHandles = this->m_paxWorkerThreads_Handles;
		while( 0 < uNumWorkerThreadsTotal)
		{
			static const u_int uNumWorkerThreadsPerWaitMax = MAXIMUM_WAIT_OBJECTS;
			u_int uNumWorkerThreadsBatch = Asura_Maths::Min( uNumWorkerThreadsTotal, uNumWorkerThreadsPerWaitMax);
			
			const DWORD uWaitStatus = WaitForMultipleObjectsEx( uNumWorkerThreadsBatch, pxWorkerThreadsHandles, TRUE, 1000, FALSE);
			if( WAIT_TIMEOUT == uWaitStatus)
			{
				u_int uFirstThreadIndex = static_cast< u_int>( pxWorkerThreadsHandles - this->m_paxWorkerThreads_Handles);
				uFirstThreadIndex = uFirstThreadIndex;
				Asura_OutputToDebugger(("PC Job Manager waiting for threads %d to %d the finish.\n", uFirstThreadIndex, uFirstThreadIndex + uNumWorkerThreadsBatch));
				continue;
			}
			
			pxWorkerThreadsHandles += uNumWorkerThreadsBatch;
			uNumWorkerThreadsTotal -= uNumWorkerThreadsBatch;
		}
	}

	// Delete the event
	CloseHandle( this->m_xNewJobEvent);
	CloseHandle( this->m_xExitThreadEvent);

	// Delete the thread data
	delete[] this->m_paxWorkerThreads_Data;
	this->m_paxWorkerThreads_Data = NULL;
	
	delete[] this->m_paxWorkerThreads_Handles;
	this->m_paxWorkerThreads_Handles = NULL;

	DeleteCriticalSection(&this->m_xJobsCriticalSection);		
}

// Worker thread function - very simple - Wait for signal, then  
// attempt to pop a job. If succeeded, then do the job.
// --------------------------------------------------------------------
long WINAPI Asura_PC_Job_Manager_Data::WorkerThread( void* pThreadData)
{
	Asura_Assert( NULL != pThreadData, "Thread Data not setup");
	
	Asura_PC_Job_Manager_Thread_Data* pxThreaData = static_cast< Asura_PC_Job_Manager_Thread_Data*>( pThreadData);
	Asura_Assert( NULL != pxThreaData, "Thread Data not setup");
	
	Asura_PC_Job_Manager_Data* pxThreadManager = pxThreaData->m_pxThreadManager;
	Asura_Assert( NULL != pxThreadManager, "Thread Manager not setup");
	
	bool bExit = false;
	
	Asura_Job_TaskQueue& xJobQueue = pxThreadManager->m_xJobQueue;

	ASURA_CPU_PROFILE__START_THREAD( "Job");
	
	ASURA_CPU_PROFILE__START_MULTITHREADED_LABEL( ASURA_CPU_PROFILE__PRESENT_FRAME, "Wait for Job" );
	
	HANDLE axWaitEventHandles[ 2] = 
	{
		pxThreadManager->m_xNewJobEvent,
		pxThreadManager->m_xExitThreadEvent
	};
	static const u_int uWaitEventHandlesCount = sizeof( axWaitEventHandles) / sizeof( axWaitEventHandles[0]);
	
	do
	{
		// Sleep until a new job comes in.
		DWORD uWaitForObjects = WaitForMultipleObjects( uWaitEventHandlesCount, axWaitEventHandles, false, INFINITE);
		
		if( (WAIT_OBJECT_0 + 0) != uWaitForObjects)
		{
			// DLH - 25 June 2007 16:25:32 - Exit Event Triggered
			break;
		}
		
//		ASURA_CPU_PROFILE__STOP_MULTITHREADED( ASURA_CPU_PROFILE__PRESENT_FRAME);
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
		__itt_event_start( pxThreadManager->m_xVTTPExecuteQueue);
#endif

		Asura_Job_TaskQueue::Task* pxTask;
		
		while( true == xJobQueue.PopTask( pxTask))
		{	
			Asura_Assert( NULL != pxTask, "This should not happen");
			void* pJobData = pxTask->GetData();
			Asura_PC_Job_Manager_Job_Data xJobData = *(static_cast< Asura_PC_Job_Manager_Job_Data*>( pJobData));
			
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_start( pxThreadManager->m_xVTTPExecuteTask);
#endif

			ASURA_CPU_PROFILE__STOP_MULTITHREADED( ASURA_CPU_PROFILE__PRESENT_FRAME);
	
			try
			{
				xJobData.ExecuteJob();
			}
			catch (...)
			{
			}
			
			ASURA_CPU_PROFILE__START_MULTITHREADED_LABEL( ASURA_CPU_PROFILE__PRESENT_FRAME, "Wait for Job" );
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_end( pxThreadManager->m_xVTTPExecuteTask);
#endif
			
			xJobQueue.ReturnTask( pxTask);
			
		}
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
		__itt_event_end( pxThreadManager->m_xVTTPExecuteQueue);
#endif
			
//		ASURA_CPU_PROFILE__START_MULTITHREADED_LABEL( ASURA_CPU_PROFILE__PRESENT_FRAME, "Wait for Job" );
	}
	while (false == bExit);

	ASURA_CPU_PROFILE__STOP_MULTITHREADED( ASURA_CPU_PROFILE__PRESENT_FRAME);

	ASURA_CPU_PROFILE__STOP_THREAD();
	
	return( 0);
}

void Asura_PC_Job_Manager_Data::WaitForAllJobsToFinish()
{
	Asura_Job_TaskQueue& xJobQueue = this->m_xJobQueue;

	Asura_Job_TaskQueue::Task* pxTask;
	
	while( true == xJobQueue.PopTask( pxTask))
	{	
		Asura_Assert( NULL != pxTask, "This should not happen");
		void* pJobData = pxTask->GetData();
		Asura_PC_Job_Manager_Job_Data xJobData = *(static_cast< Asura_PC_Job_Manager_Job_Data*>( pJobData));
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
		__itt_event_start( this->m_xVTTPExecuteTask);
#endif

		try
		{
			xJobData.ExecuteJob();
		}
		catch (...)
		{
		}
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
		__itt_event_end( this->m_xVTTPExecuteTask);
#endif
		
		xJobQueue.ReturnTask( pxTask);
	}
	
	{
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
		__itt_event_start( this->m_xVTTPWaitForFinish);
#endif

		while( false == xJobQueue.IsIdle())
		{
			// DLH - 28 June 2007 10:54:40 - Let this processor do something else but come back ASAP
			SwitchToThread();
		}
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
		__itt_event_end( this->m_xVTTPWaitForFinish);
#endif
	}
	
	this->m_uFreeTasks = xJobQueue.GetNumberOfFreeTasks();
	this->m_uQueueTasks = xJobQueue.GetNumberOfQueuedTasks();
	
	Asura_Assert( xJobQueue.IsValid(), "Job List Not Valid");
}

u_int Asura_PC_Job_Manager_Data::GetNumUnfinishedJobs() const
{
	return( 0);
}

void Asura_PC_Job_Manager_Data::AddJobCountToQueue( const u_int uJobsCount)
{
}

void Asura_PC_Job_Manager_Data::RemoveJobCountFromQueue( const u_int uJobsCount)
{
}

void Asura_PC_Job_Manager_Data::AddJob( const Asura_PC_Job_Manager_Job_Data& xJobData)
{
	// DLH - Threading might have been turned off to observer performance
	// DLH - Or it could be a single processor machine!!!!!! Oh, please minor deity, NOOOOOOOOOOoooo.....    
	if( false == this->m_bAllowTheading || 0 == this->m_uNumWorkerThreads)
	{
		xJobData.ExecuteJob();
		return;
	}
	
	{
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_start( this->m_xVTTPAddTask);
#endif
		// Increment the number of pending jobs
		Asura_Atomic::Increment( this->m_uNumPendingJobs);

		const void* pJobData = static_cast< const void*>( &xJobData);
		while( false == this->m_xJobQueue.Push( pJobData))
		{
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_start( this->m_xVTTPWaitForSpace);
#endif

			// DLH - 28 June 2007 10:54:40 - Let this processor do something else but come back ASAP
			BOOL bSwitchStatus = SwitchToThread();
			if( 0 == bSwitchStatus)
			{
				//Sleep( 0);
			}
			
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_end( this->m_xVTTPWaitForSpace);
#endif
		}
		
		// Signal the worker threads that there's something in the queue
		SetEvent( this->m_xNewJobEvent);
		
#if defined( ASURA_PC_JOB_MANAGER_SUPPORT_VTUNE_THREADPROFILER)
			__itt_event_end( this->m_xVTTPAddTask);
#endif
	}
}

void Asura_PC_Job_Manager_Data::AddJob( Asura_Job_Queue* pxJobQueue, Asura_Job_TaskQueue* pxTaskQueue)
{
	Asura_PC_Job_Manager_Job_Data xJobData( pxJobQueue, pxTaskQueue);
	
	this->AddJob( xJobData);
}

u_int Asura_PC_Job_Manager_Data::GetNumWorkerThreads()
{
	return( this->m_uNumWorkerThreads);
}

//--------------------------------------
// Asura_PC_Job_Manager
//--------------------------------------

void Asura_Job_Manager::Initialise()
{
	s_xJobManagerData.Initialise();
}

void Asura_Job_Manager::Shutdown()
{
	s_xJobManagerData.Shutdown();
}

void Asura_Job_Manager::AddJobCountToQueue( const u_int uJobsCount)
{
	s_xJobManagerData.AddJobCountToQueue( uJobsCount);
}

void Asura_Job_Manager::RemoveJobCountFromQueue( const u_int uJobsCount)
{
	s_xJobManagerData.RemoveJobCountFromQueue( uJobsCount);
}

void Asura_Job_Manager::AddJob( Asura_Job_Queue* pxJobQueue, Asura_Job_TaskQueue* pxJobTaskQueue)
{
	s_xJobManagerData.AddJob( pxJobQueue, pxJobTaskQueue);
}

void Asura_Job_Manager::WaitForAllJobsToFinish()
{
	s_xJobManagerData.WaitForAllJobsToFinish();
}

//*******************************************************************
//
// (c) 2006 Rebellion
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************
#endif //defined( ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
