//
// Asura_PC_GCDKey.cpp
//
// Created by Richard Rice - 09/09/2003
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Network/Asura_PC_GCDKey.h"

//asura independent 
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Network/Asura_Network.h"

//pc independent
#include "GameSpy/gcdkey/gcdkeys.h"
#include "Network/Asura_PC_DirectPlay.h"
#include "Network/Asura_PC_Message.h"



bool Asura_PC_GCDKey::s_bInitialised = false;
int Asura_PC_GCDKey::s_iMaxPlayers = 0;
Asura_GCD_PlayerInfo* Asura_PC_GCDKey::s_pxPlayerInfo = NULL;

bool Asura_PC_GCDKey::s_bWaitingForAuthorisation = false;
bool Asura_PC_GCDKey::s_bAuthorised = false;
Asura_String Asura_PC_GCDKey::s_xErrorMessage;


CRITICAL_SECTION g_csGCDKey;

class GCDCriticalSectionInit
{
public:
	GCDCriticalSectionInit(){InitializeCriticalSection(&g_csGCDKey);}
	~GCDCriticalSectionInit(){DeleteCriticalSection(&g_csGCDKey);}
}; 
static GCDCriticalSectionInit g_xCriticalSectionInit;

#define GCDKEY_LOCK()                   EnterCriticalSection( &g_csGCDKey ) 
#define GCDKEY_UNLOCK()                 LeaveCriticalSection( &g_csGCDKey )



struct Asura_GCD_PlayerInfo
{
	unsigned int  	m_bValid:1,
					m_bWaiting:1,
					m_bMessageSent:1,
					m_bAuthorised:1;

	int m_iIndex;
	Asura_Guid m_uGuid;

	float m_fTimer;
	char m_szChallenge[8];
	Asura_String m_xErrorMessage;
};



void Asura_PC_GCDKey::Initialise()
{
	#ifndef ASURA_PC_DISABLE_GCD_CHECK
	if(!s_bInitialised)
	{
		#ifdef QR2CDKEY_INTEGRATION
		gcd_init_qr2(NULL,Project_GetProductID());
		#else
		gcd_init(Project_GetProductID());
		#endif
		s_bInitialised = true;

		//initialise player info
		//hmm , max number of machines not necessarily set yet. Just going for a fixed number for now.
		s_iMaxPlayers = 32;//Asura_Network::GetMaxNumberOfMachines();
		s_pxPlayerInfo = new Asura_GCD_PlayerInfo[s_iMaxPlayers];
		for(int i=0;i<s_iMaxPlayers;i++)
		{
			s_pxPlayerInfo[i].m_bValid = false;
		}
	}
	#endif
}

void Asura_PC_GCDKey::Shutdown()
{
	if(s_bInitialised)
	{
		GCDKEY_LOCK();//need to take critical section to make sure there are no calls to RemovePlayer in process
		//check that all players were removed first
		for(int i=0;i<s_iMaxPlayers;i++)
		{
			Asura_Assert(!s_pxPlayerInfo[i].m_bValid,"Player should have been removed by now.");
		}

		delete [] s_pxPlayerInfo;
		s_pxPlayerInfo = 0;
		
		gcd_shutdown();
		s_bInitialised = false;
		GCDKEY_UNLOCK();
	}
}

//This function will be called by the main thread on the server
void Asura_PC_GCDKey::Update()
{
	if(!s_bInitialised) return;

	GCDKEY_LOCK(); //take critical section to make sure no players are aded/removed while we do this
	for(int i=0;i<s_iMaxPlayers;i++)
	{
		if(s_pxPlayerInfo[i].m_bValid)
		{
			if(s_pxPlayerInfo[i].m_bWaiting)
			{
				s_pxPlayerInfo[i].m_fTimer += Asura_Timers::GetFrameTime();
				if(s_pxPlayerInfo[i].m_fTimer > 60.0f)
				{
					//client is taking too long replying to authorisation challenge.
					//so assume they aren't authorised
					s_pxPlayerInfo[i].m_bWaiting = false;
					s_pxPlayerInfo[i].m_bAuthorised = false;
					//Fill in a string for this case , so we don't simply have a null error message
					//Should possibly be localised , but since this is only used as part of the gamespy setup , and
					//none of that is localised , it probably won't make much odds.
					s_pxPlayerInfo[i].m_xErrorMessage = "Disconnected from server";

					
				}
			}
			else
			{
				if(!s_pxPlayerInfo[i].m_bMessageSent)
				{
					//finished waiting on authorisation , but haven't sent messag to client yet.
					//do so now
					SendAuthorisationMessage(i,s_pxPlayerInfo[i].m_bAuthorised,s_pxPlayerInfo[i].m_xErrorMessage.GetString());
					s_pxPlayerInfo[i].m_bMessageSent = true;
					if(!s_pxPlayerInfo[i].m_bAuthorised)
					{
						//if the player wasn't authorised , start a timer till we disconnect them
						//(not doing it immediately , in case it stops the failed authorisation message
						//from getting through);
						s_pxPlayerInfo[i].m_fTimer = 0;
					}
				}
				else if(!s_pxPlayerInfo[i].m_bAuthorised)
				{
					s_pxPlayerInfo[i].m_fTimer += Asura_Timers::GetFrameTime();
					if(s_pxPlayerInfo[i].m_fTimer > 1.0f)
					{
						if(i)//can only kick off clients in this way. if this is the server , then we'll need to rely on checks elsewhere
						{
							//time to get rid of this player
							Asura_PC_DirectPlay::DisconnectPlayer(i);
						}
					}
				}
			}

		}
	}

	gcd_think();
	GCDKEY_UNLOCK();

}

//This function will be called by the directplay server thread
void Asura_PC_GCDKey::AddPlayer(int iIndex)
{
	GCDKEY_LOCK();
	if(!s_bInitialised)
	{
		//If GCD system isn't in use , then just assume client is authorised
		SendAuthorisationMessage(iIndex,true,0);
		GCDKEY_UNLOCK();
		return;
	}
	Asura_Assert(iIndex>=0 && iIndex<s_iMaxPlayers,"");

	
	if(s_pxPlayerInfo[iIndex].m_bValid)
	{
		Asura_Assert(!s_pxPlayerInfo[iIndex].m_bValid,"This player has already been added");
		GCDKEY_UNLOCK();
		return;
	}
	

	Asura_GCD_PlayerInfo* pxPlayer = &s_pxPlayerInfo[iIndex];
	pxPlayer->m_bValid = true;
	pxPlayer->m_bAuthorised = false;
	pxPlayer->m_bWaiting = true;
	pxPlayer->m_bMessageSent = false;
	pxPlayer->m_iIndex = iIndex;
	pxPlayer->m_uGuid = Asura_PC_DirectPlay::GetPlayerInfo(iIndex)->ClientGuid;
	pxPlayer->m_fTimer = 0;
	pxPlayer->m_xErrorMessage.MakeNull();

	//generate challenge string
	for(int i=0;i<7;i++)
	{
		pxPlayer->m_szChallenge[i] = (char)('a' + (Asura_Random::GetIntU() % 26));
	}
	pxPlayer->m_szChallenge[7] = 0;

	//and send off the challenge message to the client
    Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_PC_MSG_GCDCHALLENGE,sizeof(Asura_Msg_GCD_Challenge));
	pxMessage->UseGuaranteedDelivery();

	Asura_Msg_GCD_Challenge* pxData = (Asura_Msg_GCD_Challenge*) pxMessage->GetData();
	strcpy(pxData->m_szChallenge,pxPlayer->m_szChallenge);
	pxData->m_uClientGuid = pxPlayer->m_uGuid;

    pxMessage->To = iIndex ? ASURA_GUID_PLAYERS_CLIENTMACHINE : ASURA_GUID_CLIENT;
	pxMessage->From = pxPlayer->m_uGuid;

	Asura_Network::Send(pxMessage);
	
	GCDKEY_UNLOCK();


}

//This function will be called by the directplay server thread
void Asura_PC_GCDKey::RemovePlayer(int iIndex)
{
	GCDKEY_LOCK();
	if(!s_bInitialised)
	{
		GCDKEY_UNLOCK();
		return;//If GCD system isn't in use , just ignore this
	}
	Asura_Assert(iIndex>=0 && iIndex<s_iMaxPlayers,"");

	if(!s_pxPlayerInfo[iIndex].m_bValid)
	{
		Asura_Assert(s_pxPlayerInfo[iIndex].m_bValid,"This player has already been removed");
		GCDKEY_UNLOCK();
		return;
	}
	gcd_disconnect_user(Project_GetProductID(),Asura_PC_DirectPlay::GetPlayerInfo(iIndex)->NetworkID);
	s_pxPlayerInfo[iIndex].m_bValid = false;

	GCDKEY_UNLOCK();
}


void Asura_PC_GCDKey::HandleGCDChallenge(Asura_Guid uClientGuid,const char* pszChallenge)
{
	char szKey[50];
	if(!Project_GetCDKey(szKey,50))
	{
		return;
	}
	
	char szChallenge[8];
	strcpy(szChallenge,pszChallenge);

	//create message to send back to server
    Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_PC_MSG_GCDRESPONSE,sizeof(Asura_Msg_GCD_Response));
	pxMessage->UseGuaranteedDelivery();

	Asura_Msg_GCD_Response* pxData = (Asura_Msg_GCD_Response*) pxMessage->GetData();
	//fill in the response string
	gcd_compute_response(szKey,szChallenge,pxData->m_szResponse);

    pxMessage->To = ASURA_GUID_SERVER;
	pxMessage->From = uClientGuid;

	Asura_Network::Send(pxMessage);
}

//This function will be called by the main thread on the server
void Asura_PC_GCDKey::HandleGCDResponse(Asura_Guid uClientGuid,const char* pacResponse)
{
	Asura_Assert(s_bInitialised,"");

	char acResponse[RESPONSE_SIZE];
	memcpy(acResponse,pacResponse,RESPONSE_SIZE);

	GCDKEY_LOCK();
	//find the player info from the guid
	int iIndex;
	for(iIndex=0;iIndex<s_iMaxPlayers;iIndex++)
	{
		if(!s_pxPlayerInfo[iIndex].m_bValid) continue;
		if(s_pxPlayerInfo[iIndex].m_uGuid == uClientGuid)
		{
			break;
		}
	}
	if(iIndex == s_iMaxPlayers)
	{
		Asura_Assert(0,"Failed to find player info");
		return;
	}

	Asura_GCD_PlayerInfo* pxPlayer = &s_pxPlayerInfo[iIndex];

	//start the authentication process
	gcd_authenticate_user(	Project_GetProductID(), 
							Asura_PC_DirectPlay::GetPlayerInfo(iIndex)->NetworkID, 
							Asura_PC_DirectPlay::GetPlayerIPAddress(iIndex), 
							pxPlayer->m_szChallenge, 
							acResponse,
							GCDAuthenticationCallback, 
							pxPlayer);
	GCDKEY_UNLOCK();
	
}

//This is ultimately called by Asura_PCD_GCDKey::Update in main thread ,so critical section should already be taken
void Asura_PC_GCDKey::GCDAuthenticationCallback(int iProductID, int iLocalID, int iAuthenticated, char *pszErrmsg, void *pUserData)
{
	Asura_GCD_PlayerInfo* pxPlayer = (Asura_GCD_PlayerInfo*)pUserData;
	
	Asura_Assert(iProductID == Project_GetProductID(),"");
	Asura_Assert(iLocalID == (int)Asura_PC_DirectPlay::GetPlayerInfo(pxPlayer->m_iIndex)->NetworkID,"Local ID doesn't match");
		
	pxPlayer->m_bWaiting = false;
	if(iAuthenticated)
	{
		//mark player as authorised... rejoice!
		pxPlayer->m_bAuthorised = true;
	}
	else
	{
		//bad player , need to kick them from the game
		//(somehow)
		pxPlayer->m_xErrorMessage = pszErrmsg;
		pxPlayer->m_bAuthorised = false;
	}
	
}

//This can be called by either thread.... but hopefully shouldn't be using anything from the GCD code that needs locking
void Asura_PC_GCDKey::SendAuthorisationMessage(int iIndex,bool bAuthorised,const char* pszErrorMessage)
{
	if(bAuthorised) pszErrorMessage = "";
    if(!pszErrorMessage) pszErrorMessage = "";

    Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_PC_MSG_GCDAUTHORISATION,sizeof(Asura_Msg_GCD_Authorisation)+strlen(pszErrorMessage)+1);
	pxMessage->UseGuaranteedDelivery();

	Asura_Msg_GCD_Authorisation* pxData = (Asura_Msg_GCD_Authorisation*) pxMessage->GetData();
	pxData->m_ucAuthorised = bAuthorised;

	char* pszDestString = (char*)(pxData+1);
	strcpy(pszDestString,pszErrorMessage);


    pxMessage->To = iIndex ? ASURA_GUID_PLAYERS_CLIENTMACHINE : ASURA_GUID_CLIENT;
	pxMessage->From = Asura_PC_DirectPlay::GetPlayerInfo(iIndex)->ClientGuid;

	Asura_Network::Send(pxMessage);
	
}
