// Asura_PC_Steamworks.cpp
// Thomas Young - 13.01.09
// Copyright (C) 2008 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_PC_SUPPORT_STEAMWORKS

#include "Debugging/Asura_Logfile.h"						// Asura
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_NetTransport_System.h"
#include "Network/Steamworks/Asura_PC_NetSocketSteamworks.h"
#include "Network/Steamworks/Asura_PC_Steamworks.h"
#include "Network/Steamworks/Asura_PC_SteamworksClient.h"
#include "Network/Steamworks/Asura_PC_SteamworksServer.h"
#include "Network/Steamworks/Asura_PC_SteamworksLobby.h"
#include "Network/Steamworks/Asura_PC_SteamworksStats.h"
#include "Core/Asura_Memheap.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_Timers.h"
#include "Sound/Asura_XAudio2.h"

#ifdef ASURA_ENABLE_QOS_SYSTEM
#include "Network/Asura_Qos_System.h"
#endif //ASURA_ENABLE_QOS_SYSTEM

#include "steam/steam_api.h"								// Steam

// D E F I N E S /////////////////////////////////////////////////////////////

#define ASURA_PC_STEAM_MAX_GAME_SEARCH_RESULTS		1024
#define ASURA_PC_STEAM_GAME_SEARCH_TIMEOUT			10.0f
#define ASURA_PC_STEAM_GAME_SEARCH_UPDATE_INTERVAL	0.25f

// G L O B A L   V A R I A B L E S ///////////////////////////////////////////

static u_int g_uFrameNumber = 0;

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////

// ____________________ Asura_PC_Steamworks_ServerList _________________________

class Asura_PC_Steamworks_ServerList : public ISteamMatchmakingServerListResponse
{
	// This class handles server browsing functionality
	public:

		void ServerResponded( int iServer );
		void ServerFailedToRespond( int iServer );

		void RefreshComplete( EMatchMakingServerResponse response ); 
};

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

bool											Asura_PC_Steamworks::s_bIsInitialised			= false;

bool											Asura_PC_Steamworks::s_bSearching				= false;
float											Asura_PC_Steamworks::s_fSearchStartTime			= 0.0f;
ASURA_PC_STEAMWORKS_SEARCH_TYPE					Asura_PC_Steamworks::s_eSearchType				= ASURA_PC_STEAMWORKS_SEARCH_TYPE_NONE;
u_int											Asura_PC_Steamworks::s_uNumSearchResults		= 0;

Asura_PC_SteamServer_Data*						Asura_PC_Steamworks::s_pxServerList				= 0;
Asura_PC_Steamworks_ServerList					Asura_PC_Steamworks::s_xServerListResponse;
ASURA_PC_STEAMWORKS_SERVER_TYPE					Asura_PC_Steamworks::s_eSearchServerType		= ASURA_PC_STEAMWORKS_SERVER_LAN;

Asura_PC_SteamworksLobby_Data*					Asura_PC_Steamworks::s_pxLobbyList				= 0;
Asura_List<Asura_String>						Asura_PC_Steamworks::s_xLobbyMetaData;

Asura_PC_Steam_Callback< LobbyMatchList_t >		Asura_PC_Steamworks::s_xCallback_LobbyList		( Asura_PC_Steamworks::Callback_LobbyList );
Asura_PC_Steam_Callback< LobbyDataUpdate_t >	Asura_PC_Steamworks::s_xCallback_LobbyData		( Asura_PC_Steamworks::Callback_LobbyData );

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

extern "C" void __cdecl Asura_Steam_DebugTextHook( int nSeverity, const char *szDebugText )
{
	Asura_OutputToDebugger( ( "[Steam] %s ( severity %d )", szDebugText ) );
}

// _____________________ Asura_PC_Steamworks ___________________________

Asura_PC_Steamworks::Asura_PC_Steamworks ()
{
}

Asura_PC_Steamworks::~Asura_PC_Steamworks ()
{
}


void Asura_PC_Steamworks::Initialise ()
{
	if ( s_bIsInitialised )
	{
		Asura_Assert( false, "Asura_PC_Steamworks is already initialised" );
		return;
	}

	if( Asura_Network::IsNetworkInitialised() || Asura_NetTransport_System::IsInitialised() )
	{
		Asura_Assert( false, "Some networking has already been initialised, you must initialise steam before this" );
		return;
	}

	s_bIsInitialised = SteamAPI_Init();

	// TY 05.04.09 - Allow steam init to fail silently for now, so that people without steam arent bugged by asserts
	//Asura_Assert( s_bIsInitialised, "Failed to initialise Steamworks (is steam running?)" );

	if ( s_bIsInitialised )
	{
		s_xCallback_LobbyData.Register	( false );

		SteamClient()->SetWarningMessageHook( &Asura_Steam_DebugTextHook );

		// Initialise the client and server classes
		Asura_PC_SteamworksClient::Initialise();
		Asura_PC_SteamworksServer::Initialise();
		Asura_PC_SteamworksLobby_System::Initialise();

		Asura_PC_SteamworksStats::Initialise();
	}
}


void Asura_PC_Steamworks::Update ()
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( g_uFrameNumber == Asura_Timers::GetFrameNumber() )
	{
		// Already updated this frame
		return;
	}
	
	g_uFrameNumber = Asura_Timers::GetFrameNumber();
	
	SteamAPI_RunCallbacks();

	if( s_bSearching )
	{
		// Handle search timing out
		if( ( Asura_Timers::GetFrameTime() - s_fSearchStartTime ) > ASURA_PC_STEAM_GAME_SEARCH_TIMEOUT )
		{
			ClearSearch();
		}
	}

	// Update the client and server classes
	Asura_PC_SteamworksClient::Update();
	Asura_PC_SteamworksServer::Update();
	Asura_PC_SteamworksLobby_System::Update();

	Asura_PC_SteamworksStats::Update();
}


void Asura_PC_Steamworks::Shutdown ()
{
	if( !Asura_PC_Steamworks::IsInitialised() )
	{
		// dont assert, its annoying for the caller to need to check steam is enabled in this case
		return;
	}

	Asura_Assert( !Asura_PC_SteamworksServer::IsServerRunning(), "A steam server is still running, should have been shutdown by now" );
	Asura_Assert( !Asura_PC_SteamworksClient::IsConnectedToServer(), "Our steam client is still conencted to a server, should have been disconnected by now" );

	ClearSearch();

	// Shutdown the client and server classes
	Asura_PC_SteamworksClient::Shutdown();
	Asura_PC_SteamworksServer::Shutdown();
	Asura_PC_SteamworksLobby_System::Shutdown();
	Asura_PC_SteamworksStats::Shutdown();

	s_bIsInitialised = false;

	#ifdef ASURA_ENABLE_QOS_SYSTEM
	Asura_Qos_System::Shutdown();
	#endif

	SteamAPI_Shutdown ();
}


void Asura_PC_Steamworks::BeginServerSearch ( const ASURA_PC_STEAMWORKS_SERVER_TYPE eServerType, const Asura_PC_Steamworks_KeyValuePair* pxFilters, const u_int uNumFilters )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( s_bSearching )
	{
		Asura_Assert( false, "A search is already being conducted, please wait for it to finish before starting a new search" );
		return;
	}

	Asura_Assert( s_eSearchType == ASURA_PC_STEAMWORKS_SEARCH_TYPE_NONE, "Search type not reset, did you forget to call 'ClearSearch()'?");

	if( s_pxServerList )
	{
		Asura_Assert( false, "server list is not null?" );
		ClearSearch();
	}

	if( !SteamMatchmakingServers() || !SteamUtils() )
	{
		Asura_Assert( false, "Missing steam API objects, cannot begin server search" );
		return;
	}

	s_fSearchStartTime = Asura_Timers::GetLocalUnscaledTime();

	switch( s_eSearchServerType )
	{
		case ASURA_PC_STEAMWORKS_SERVER_INTERNET:
		{
			// Only internet searches allow us to specify filters
			const u_int uMaxFilters = 16;

			Asura_Assert( uNumFilters < uMaxFilters, "server searching can only take up to 16 filters" );

			MatchMakingKeyValuePair_t pxSteamFilters[16];

			for( u_int uFilter = 0; uFilter < Asura_Maths::Min( uMaxFilters, uNumFilters ); ++uFilter )
			{
				Asura_PrintfToBfr( pxSteamFilters[ uFilter ].m_szKey, "%s", ASURA_PC_STEAMWORKS_FILTER_MAX_CHARS, true, pxFilters[ uFilter ].m_szKey );
				Asura_PrintfToBfr( pxSteamFilters[ uFilter ].m_szValue, "%s", ASURA_PC_STEAMWORKS_FILTER_MAX_CHARS, true, pxFilters[ uFilter ].m_szValue );
			}

			s_eSearchServerType = eServerType;

			MatchMakingKeyValuePair_t* _pxSteamFilters = &pxSteamFilters[0];	// need a pointer to pointer

			SteamMatchmakingServers()->RequestInternetServerList( SteamUtils()->GetAppID(), &_pxSteamFilters , uNumFilters, &s_xServerListResponse );
		}
		break;

		case ASURA_PC_STEAMWORKS_SERVER_LAN:
		{
			SteamMatchmakingServers()->RequestLANServerList( SteamUtils()->GetAppID(), &s_xServerListResponse );
		}
		break;
	}

	s_pxServerList = new Asura_PC_SteamServer_Data[ ASURA_PC_STEAM_MAX_GAME_SEARCH_RESULTS ];
	s_uNumSearchResults = 0;

	s_bSearching = true;
	s_eSearchType = ASURA_PC_STEAMWORKS_SEARCH_TYPE_SERVER;
}


bool Asura_PC_Steamworks::IsSearchFinished ( )
{
	Asura_Assert( s_eSearchType != ASURA_PC_STEAMWORKS_SEARCH_TYPE_NONE, "No search is being conducted" );
	
	return !s_bSearching;
}


const Asura_PC_SteamServer_Data* Asura_PC_Steamworks::GetServerSearchResult ( const u_int uResult )
{
	if( s_bSearching )
	{
		Asura_Assert( false, "A search is being conducted, please wait for it to finish" );
		return 0;
	}
	if( s_eSearchType != ASURA_PC_STEAMWORKS_SEARCH_TYPE_SERVER )
	{
		Asura_Assert( false, "No server search has been conducted" );
		return 0;
	}
	if( uResult >= s_uNumSearchResults )
	{
		// Dont assert here, caller might be doing something like while( GetSearchResult() )
		return 0;
	}

	return &s_pxServerList[ uResult ];
}


u_int Asura_PC_Steamworks::GetNumSearchResults ( )
{
	if( s_bSearching )
	{
		Asura_Assert( false, "A search is being conducted, please wait for it to finish" );
		return 0;
	}
	return s_uNumSearchResults;
}


void Asura_PC_Steamworks::ClearSearch ( )
{
	// Clean up our search data
	if( s_pxServerList )
	{
		delete []s_pxServerList;
		s_pxServerList = 0;
	}

	if( s_pxLobbyList )
	{
		delete []s_pxLobbyList;
		s_pxLobbyList = 0;
	}

	while( !s_xLobbyMetaData.IsEmpty() )
	{
		delete s_xLobbyMetaData.PopHeadObject();
	}

	s_uNumSearchResults = 0;
	s_bSearching = false;
	s_eSearchType = ASURA_PC_STEAMWORKS_SEARCH_TYPE_NONE;
}


void Asura_PC_Steamworks::BeginLobbySearch ( const Asura_PC_Steamworks_LobbyFilter* pxFilters, const u_int uNumFilters, const Asura_List<Asura_String>* pxMetaData )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( s_bSearching )
	{
		Asura_Assert( false, "A search is already being conducted, please wait for it to finish before starting a new search" );
		return;
	}

	Asura_Assert( s_eSearchType == ASURA_PC_STEAMWORKS_SEARCH_TYPE_NONE, "Search type not reset, did you forget to call 'ClearSearch()'?");

	if( s_pxLobbyList )
	{
		Asura_Assert( false, "Lobby list is not null?" );
		ClearSearch();
	}
	
	if( !SteamMatchmaking() )
	{
		Asura_Assert( false, "Missing steam API objects, cannot begin server search" );
		return;
	}

	s_pxLobbyList = new Asura_PC_SteamworksLobby_Data[ ASURA_PC_STEAM_MAX_GAME_SEARCH_RESULTS ];
	s_uNumSearchResults = 0;

	// Copy the list of what metadata we should get from each lobby
	// this is a pain because we are dealing with lists of strings everywhere

	while( !s_xLobbyMetaData.IsEmpty() )
	{
		delete s_xLobbyMetaData.PopHeadObject();
	}
	
	if( pxMetaData )
	{
		Asura_List_ConstIt<Asura_String> xIter( *pxMetaData );

		for( ; !xIter.Done(); xIter.Next() )
		{
			Asura_String* pxNewString = new Asura_String( *xIter.GetCurrent() );
			s_xLobbyMetaData.AddToList( pxNewString );
		}
	}
	
	// Add all the filters

	for( u_int uFilter = 0; uFilter < uNumFilters; ++uFilter )
	{
		int iComparisonType = k_ELobbyComparisonEqual;
		
		switch( pxFilters[ uFilter ].m_eOperator )
		{
			case ASURA_PC_STEAMWORKS_FILTER_OP_EQUAL:		iComparisonType = k_ELobbyComparisonEqual;					break;
			case ASURA_PC_STEAMWORKS_FILTER_OP_LESS_EQUAL:	iComparisonType = k_ELobbyComparisonEqualToOrLessThan;		break;
			case ASURA_PC_STEAMWORKS_FILTER_OP_GREAT_EQUAL:	iComparisonType	= k_ELobbyComparisonEqualToOrGreaterThan;	break;

			default:	Asura_Assert( false, "Unknown filter operator" );
		}

		SteamMatchmaking()->AddRequestLobbyListNumericalFilter( pxFilters[ uFilter ].m_szKey, pxFilters[ uFilter ].m_iValue, iComparisonType );
	}

	s_bSearching = true;
	s_eSearchType = ASURA_PC_STEAMWORKS_SEARCH_TYPE_LOBBY;
	
	// Request our lobby list, the result will be in the callback Callback_LobbyList
	SteamAPICall_t xResult = SteamMatchmaking()->RequestLobbyList();

	s_xCallback_LobbyList.Set( xResult );

}


const Asura_PC_SteamworksLobby_Data* Asura_PC_Steamworks::GetLobbySearchResult ( const u_int uResult )
{
	if( s_bSearching )
	{
		Asura_Assert( false, "A search is being conducted, please wait for it to finish" );
		return 0;
	}
	if( s_eSearchType != ASURA_PC_STEAMWORKS_SEARCH_TYPE_LOBBY )
	{
		Asura_Assert( false, "No lobby search has been conducted" );
		return 0;
	}
	if( uResult >= s_uNumSearchResults )
	{
		// Dont assert here, caller might be doing something like while( GetLobbySearchResult() )
		return 0;
	}

	return &s_pxLobbyList[ uResult ];
}


void Asura_PC_Steamworks::Callback_LobbyList ( const LobbyMatchList_t* const pxParam )
{
	if( !s_pxLobbyList )
	{
		return;
	}

	s_uNumSearchResults = static_cast<u_int>( Asura_Maths::Min( static_cast<u_int>( pxParam->m_nLobbiesMatching ), (u_int)ASURA_PC_STEAM_MAX_GAME_SEARCH_RESULTS ) );

	for( int iLobby = 0; iLobby < static_cast<int>( s_uNumSearchResults ); ++iLobby )
	{
		s_pxLobbyList[ iLobby ].m_xSteamID = SteamMatchmaking()->GetLobbyByIndex( iLobby );

		// Request data about this lobby, it will be recieved in the Callback_LobbyData
		SteamMatchmaking()->RequestLobbyData( s_pxLobbyList[ iLobby ].m_xSteamID );
	}
}


void Asura_PC_Steamworks::Callback_LobbyData( const LobbyDataUpdate_t* const pxParam )
{
	// TODO: We can get this callback outside of searching, do something with it in those cases
	if( s_bSearching )
	{
		bool bFound = false;

		// Find this lobby in our list
		for( u_int uLobby = 0; uLobby < s_uNumSearchResults; ++uLobby )
		{
			if( s_pxLobbyList[ uLobby ].m_xSteamID == pxParam->m_ulSteamIDLobby )
			{
				s_pxLobbyList[ uLobby ].m_bMetaDataReceived = true;

				Asura_List_It<Asura_String> xIter( s_xLobbyMetaData );

				for( ; !xIter.Done(); xIter.Next() )
				{
					Asura_String* pxKey = xIter.GetCurrent();

					Asura_PC_Steamworks_KeyValuePair* pxMetaDataEntry = new Asura_PC_Steamworks_KeyValuePair;
					if( pxMetaDataEntry )
					{
						// Copy the key string

						Asura_PrintfToBfr( pxMetaDataEntry->m_szKey, "%s", ASURA_PC_STEAMWORKS_FILTER_MAX_CHARS, true, pxKey->GetString() );

						// Copy the value string

						const char *pcData = SteamMatchmaking()->GetLobbyData( s_pxLobbyList[ uLobby ].m_xSteamID, pxMetaDataEntry->m_szKey );

						Asura_PrintfToBfr( pxMetaDataEntry->m_szValue, "%s", ASURA_PC_STEAMWORKS_FILTER_MAX_CHARS, true, pcData );

						s_pxLobbyList[ uLobby ].m_xMetaData.AddToList( pxMetaDataEntry );
					}
				}
				bFound = true;

				break;	// we found the lobby and got its data so break
			}
		}

		// Check if we have found the meta data for every lobby

		if( bFound )
		{
			for( u_int uLobby = 0; uLobby < s_uNumSearchResults; ++uLobby )
			{
				if( !s_pxLobbyList[ uLobby ].m_bMetaDataReceived )
				{
					return;
				}
			}
		}

		// If we got here then we must have got all the metadata

		s_bSearching = false;
	}
	
}


void Asura_PC_Steamworks::ShowSteamOverlay ( const ASURA_PC_STEAMWORKS_OVERLAY eOverlay )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( SteamFriends() )
	{
		switch( eOverlay )
		{
			case ASURA_PC_STEAMWORKS_OVERLAY_FRIENDS:		SteamFriends()->ActivateGameOverlay( "Friends"   );		break;
			case ASURA_PC_STEAMWORKS_OVERLAY_COMMUNITY:		SteamFriends()->ActivateGameOverlay( "Community" );		break;
			case ASURA_PC_STEAMWORKS_OVERLAY_PLAYERS:		SteamFriends()->ActivateGameOverlay( "Players"   );		break;
			case ASURA_PC_STEAMWORKS_OVERLAY_SETTINGS:		SteamFriends()->ActivateGameOverlay( "Settings"  );		break;
		}
	}
}


void Asura_PC_Steamworks::RegisterConsoleCmds ( )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Status"			, Asura_PC_Steamworks::ConsoleCmd_SteamworksStatus		, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Voice.Enable"	, Asura_PC_Steamworks::ConsoleCmd_StartVoiceRecording	, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Asura.Steam.Voice.Disable"	, Asura_PC_Steamworks::ConsoleCmd_StopVoiceRecording	, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif
}

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

bool Asura_PC_Steamworks::ConsoleCmd_SteamworksStatus( bool bHelp )
{
	if (!bHelp)
	{
		// TODO:	Need to fill this console command in
		//			It should probably list any active sockets, joined/hosted lobbies and/or servers
		
		if( IsInitialised() )
		{
			Asura_CommandConsole::Printf( L"Steamworks is initialised\n" );
		}
	}
	return true;
}

bool Asura_PC_Steamworks::ConsoleCmd_StartVoiceRecording ( bool bHelp )
{
	if (!bHelp)
	{
		if( IsInitialised() )
		{
			Asura_PC_SteamworksClient::StartVoiceRecording();
		}
		else
		{
			Asura_CommandConsole::Printf( L"Steamworks is not initialised\n" );
		}
	}
	return true;
}

bool Asura_PC_Steamworks::ConsoleCmd_StopVoiceRecording ( bool bHelp )
{
	if (!bHelp)
	{
		if( IsInitialised() )
		{
			Asura_PC_SteamworksClient::StopVoiceRecording();
		}
		else
		{
			Asura_CommandConsole::Printf( L"Steamworks is not initialised\n" );
		}
	}
	return true;
}

#endif


//_____________________________ Asura_PC_Steamworks_ServerList __________________________________


void Asura_PC_Steamworks_ServerList::ServerResponded( int iServer )
{
	if( Asura_PC_Steamworks::IsSearchFinished() )
	{
		return;
	}

	if( Asura_PC_Steamworks::s_uNumSearchResults == ASURA_PC_STEAM_MAX_GAME_SEARCH_RESULTS )
	{
		// We cant store any more server details
		return;
	}

	gameserveritem_t *pxServer = 0;

	switch( Asura_PC_Steamworks::s_eSearchServerType )
	{
		case ASURA_PC_STEAMWORKS_SERVER_LAN:		pxServer = SteamMatchmakingServers()->GetServerDetails( eLANServer		, iServer );	break;
		case ASURA_PC_STEAMWORKS_SERVER_INTERNET:	pxServer = SteamMatchmakingServers()->GetServerDetails( eInternetServer	, iServer );	break;
	}

	if ( pxServer )
	{
		// Filter out servers that don't match our appid here (might get these in LAN calls since we can't put more filters on it)
		if ( pxServer->m_nAppID == static_cast<int>( SteamUtils()->GetAppID() ) )
		{
			Asura_PC_SteamServer_Data& xEntry = Asura_PC_Steamworks::s_pxServerList[ Asura_PC_Steamworks::s_uNumSearchResults ];

			xEntry.m_uAddress	= pxServer->m_NetAdr.GetIP();
			xEntry.m_usPort		= pxServer->m_NetAdr.GetConnectionPort();

			xEntry.m_ucNumPlayers = static_cast<unsigned char>( pxServer->m_nPlayers );
			xEntry.m_ucMaxPlayers = static_cast<unsigned char>( pxServer->m_nMaxPlayers );

			++Asura_PC_Steamworks::s_uNumSearchResults;
		}
	}
}


void Asura_PC_Steamworks_ServerList::ServerFailedToRespond( int iServer )
{
	// not really a very useful callback
}


void Asura_PC_Steamworks_ServerList::RefreshComplete( EMatchMakingServerResponse response )
{
	if( Asura_PC_Steamworks::IsSearchFinished() )
	{
		return;
	}

	if( Asura_PC_Steamworks::s_bSearching )
	{
		Asura_PC_Steamworks::s_bSearching = false;
	}
}

#endif	// ASURA_PC_SUPPORT_STEAMWORKS

//eof
