// Asura_PC_SteamworksClient.cpp
// Thomas Young - 04.02.09
// Copyright (C) 2009 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_PC_SUPPORT_STEAMWORKS

#include "Core/Asura_List.h"								// Asura
#include "Core/Asura_Timers.h"
#include "Core/Asura_Memheap.h"
#include "Network/Asura_NetSocket.h"
#include "Network/Asura_Platform_NetSocket.h"
#include "Network/Asura_ReliableSocket.h"
#include "Network/Steamworks/Asura_PC_Steamworks.h"
#include "Network/Steamworks/Asura_PC_SteamworksClient.h"
#include "Network/Steamworks/Asura_PC_SteamworksMessages.h"
#include "Network/Steamworks/Asura_PC_NetSocketSteamworks.h"
#include "Network/Steamworks/Asura_PC_SteamworksServer.h"
#include "Sound/Asura_XAudio2.h"
#include "Sound/Asura_XAudio2_Voice.h"

// E N U M E R A T I O N S ///////////////////////////////////////////////////

enum ASURA_STEAM_CLIENT_MESSAGE
{
	// These should ideally not conflict with ASURA_STEAM_RELIABLESOCKET_MESSAGE

	ASURA_STEAM_CLIENT_MESSAGE_VOICE = 20,
};

// D E F I N E S /////////////////////////////////////////////////////////////

#define ASURA_PC_STEAMWORKS_CLIENT_VOICE_REC_UPDATE_INTERVAL	0.25f				// 1/4 Second
#define ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_REC_DATA			1024
#define ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_PLAY_DATA			( 1024 * 64 )
#define ASURA_PC_STEAMWORKS_CLIENT_DECOMPRESS_BUFFER_SIZE		( 1024 * 32 )
#define ASURA_PC_STEAMWORKS_CLIENT_VOICE_VPORT					5

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int uASURA_PC_CLIENT_VOICE_MSG_BUFFER_SIZE = 2048;

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////

class Asura_SC_VoicePeerSocket
{
	// This class represents another client we need to send voice data to

	public:

		enum STATE
		{
			STATE_CONNECTING,
			STATE_ERROR,
			STATE_CONNECTED
		};

		static const int iCONNECT_TIMEOUT = 10;

		CSteamID			m_xSteamID;
		SNetSocket_t		m_xSocket;
		STATE				m_eState;

		Asura_SC_VoicePeerSocket ( const CSteamID xSteamID )
		{
			m_xSocket = 0;
			m_xSteamID = xSteamID;

			m_xSocket = SteamNetworking()->CreateP2PConnectionSocket( m_xSteamID, ASURA_PC_STEAMWORKS_CLIENT_VOICE_VPORT, iCONNECT_TIMEOUT, true );

			m_eState = m_xSocket ? STATE_CONNECTING : STATE_ERROR;
		}

		~Asura_SC_VoicePeerSocket ()
		{
			SteamNetworking()->DestroySocket( m_xSocket, false );
		}
};


class Asura_SC_PeerVoice
{
	// This class represents the audio voice for another client's speech

	public:
	
		Asura_Handle	m_xVoiceHandle;
		CSteamID		m_xSteamID;

		Asura_SC_PeerVoice ()
		{
			m_xVoiceHandle	= ASURA_INVALID_HANDLE;
			m_xSteamID		= CSteamID().ConvertToUint64();
		}

		void DestroyAudioVoice ()
		{
			if( m_xVoiceHandle != ASURA_INVALID_HANDLE )
			{
				Asura_XAudio2::DestroyVoice( m_xVoiceHandle );
			}
			m_xVoiceHandle = ASURA_INVALID_HANDLE;
		}

		~Asura_SC_PeerVoice ()
		{
			DestroyAudioVoice();
		}
};


class Asura_PC_Steamworks_AuthTicket
{
	public:
	
		Asura_PC_Steamworks_AuthTicket()
		{
			m_xTicket = k_HAuthTicketInvalid;
			m_pcAuthData = 0;
			m_uAuthDataSize = 0;
		}

		~Asura_PC_Steamworks_AuthTicket()
		{
			ClearAuthData();
		}

		void ClearAuthData ()
		{
			if( m_pcAuthData )
			{
				Asura_Deallocate( m_pcAuthData );
				m_pcAuthData = 0;
			}
			m_uAuthDataSize = 0;
		}

		CSteamID		m_xSteamID;		// steam ID of the client this ticket is for
		HAuthTicket		m_xTicket;
		char*			m_pcAuthData;
		u_int			m_uAuthDataSize;
};


// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

Asura_UnicodeChar										Asura_PC_SteamworksClient::s_awcSteamIDText		[ASURA_PC_STEAM_MAX_STEAMID_LENGTH];
char*													Asura_PC_SteamworksClient::s_pcAuthBlob			= 0;
u_int													Asura_PC_SteamworksClient::s_uAuthBlobSize		= 0;

bool													Asura_PC_SteamworksClient::s_bConnectedToServer	= false;
Asura_NetAddress										Asura_PC_SteamworksClient::s_xServerAddress;
unsigned short											Asura_PC_SteamworksClient::s_usServerPort		= 0;

bool													Asura_PC_SteamworksClient::s_bVoiceRecActive		= false;
bool													Asura_PC_SteamworksClient::s_bMicrophoneAttached	= false;
char*													Asura_PC_SteamworksClient::s_pcVoiceRecData			= 0;
u_int													Asura_PC_SteamworksClient::s_uVoiceRecDataSize		= 0;
float													Asura_PC_SteamworksClient::s_fLastVoiceRecUpdate	= 0.0f;

Asura_List<Asura_SC_VoicePeerSocket>	Asura_PC_SteamworksClient::s_xVoicePeerSocketList;
SNetListenSocket_t										Asura_PC_SteamworksClient::s_xVoiceListenSocket		= 0;
bool													Asura_PC_SteamworksClient::s_bAllowVoicePlayback	= true;
Asura_List<Asura_SC_PeerVoice>			Asura_PC_SteamworksClient::s_xPeerVoices;
Asura_MemHeap*											Asura_PC_SteamworksClient::s_pxVoicePlaybackMemHeap;
char*													Asura_PC_SteamworksClient::s_pcVoicePlaybackMemory;

Asura_List<Asura_PC_Steamworks_AuthTicket>				Asura_PC_SteamworksClient::s_xPeerAuthTicketList;

void*													Asura_PC_SteamworksClient::s_pCallbackUserData_Accept = 0;
void*													Asura_PC_SteamworksClient::s_pCallbackUserData_Reject = 0;

Asura_PC_SteamClient_Callback_ClientAccept				Asura_PC_SteamworksClient::s_pfnAsuraCallback_ClientAccept = 0;
Asura_PC_SteamClient_Callback_ClientReject				Asura_PC_SteamworksClient::s_pfnAsuraCallback_ClientReject = 0;

Asura_PC_Steam_Callback< ValidateAuthTicketResponse_t >	Asura_PC_SteamworksClient::s_xCallback_TicketResponse ( Asura_PC_SteamworksClient::Callback_TicketResponse );

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

// _____________________ Asura_PC_SteamworksClient ___________________________

Asura_PC_SteamworksClient::Asura_PC_SteamworksClient ()
{
}

Asura_PC_SteamworksClient::~Asura_PC_SteamworksClient ()
{
}


void Asura_PC_SteamworksClient::Initialise ()
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( SteamFriends() )
	{
		const char* szPersona = SteamFriends()->GetPersonaName();

		// Save the steam ID of the user
		MultiByteToWideChar( CP_UTF8, 0, szPersona, -1, s_awcSteamIDText, ASURA_PC_STEAM_MAX_STEAMID_LENGTH );
	}

	s_xCallback_TicketResponse.Register( false );

	s_pcVoiceRecData			= static_cast<char*>( Asura_Allocate( ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_REC_DATA ) );
	s_pcVoicePlaybackMemory		= static_cast<char*>( Asura_Allocate( ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_PLAY_DATA ) );
	s_pxVoicePlaybackMemHeap	= new Asura_MemHeap( s_pcVoicePlaybackMemory, ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_PLAY_DATA );

	s_pxVoicePlaybackMemHeap->EnableThreadSafety();

	if( SteamNetworking() )
	{
		s_xVoiceListenSocket = SteamNetworking()->CreateListenSocket( ASURA_PC_STEAMWORKS_CLIENT_VOICE_VPORT, 0, 0, true );
	}
	
}


void Asura_PC_SteamworksClient::Update ()
{
	// Check for new voice messages
	if( s_xVoiceListenSocket )
	{
		char			pBuffer[uASURA_PC_CLIENT_VOICE_MSG_BUFFER_SIZE];
		uint32			uRecvSize;
		SNetSocket_t	xSourceSocket;

		while( SteamNetworking()->RetrieveData( s_xVoiceListenSocket, pBuffer, uASURA_PC_CLIENT_VOICE_MSG_BUFFER_SIZE, &uRecvSize, &xSourceSocket ) )
		{
			// Check if this is a valid message
			Asura_Steam_Msg_Header* pxHdr = reinterpret_cast< Asura_Steam_Msg_Header* >( &pBuffer[0] );

			CSteamID	xRemoteSteamID;
			int			iIgnore1;
			uint32		uIgnore2;
			uint16		usIgnore3;
			
			if(		pxHdr->IsValid() 
				&&	( pxHdr->GetType() == ASURA_STEAM_CLIENT_MESSAGE_VOICE )
				&&	SteamNetworking()->GetSocketInfo( xSourceSocket, &xRemoteSteamID, &iIgnore1, &uIgnore2, &usIgnore3 ) )
			{
				// This is a voice message (as we expect)
				HandleMessage_VoiceData( pxHdr, uRecvSize, xRemoteSteamID );
			}
		}
	}
	
	// Do we have any voice chat to send?
	if( SteamUser() )
	{
		if( ( Asura_Timers::GetLocalUnscaledTime() - s_fLastVoiceRecUpdate ) > ASURA_PC_STEAMWORKS_CLIENT_VOICE_REC_UPDATE_INTERVAL )
		{
			s_fLastVoiceRecUpdate = Asura_Timers::GetLocalUnscaledTime();
			
			#ifdef ASURA_ASSERT
			EVoiceResult xResult = SteamUser()->GetCompressedVoice( s_pcVoiceRecData, ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_REC_DATA, &s_uVoiceRecDataSize );
			switch( xResult )
			{
				case k_EVoiceResultNotInitialized:	Asura_Assert( false, "Steam: Voice not initialised" );			break;
				case k_EVoiceResultBufferTooSmall:	Asura_Assert( false, "Steam: voice data buffer too small" );	break;
				case k_EVoiceResultDataCorrupted:	Asura_Assert( false, "Steam: Voice data corrupt" );				break;
			}
			#else
			SteamUser()->GetCompressedVoice( s_pcVoiceRecData, ASURA_PC_STEAMWORKS_CLIENT_MAX_VOICE_REC_DATA, &s_uVoiceRecDataSize );
			#endif

			// If we have some voice data to send then we must have a microphone attached.
			if (s_uVoiceRecDataSize > 0)
			{
				s_bMicrophoneAttached = true;
			}

			// Make a message from this data (early out if we have no peers to send the data to)
			if( ( s_uVoiceRecDataSize > 0 ) && !s_xVoicePeerSocketList.IsEmpty() )
			{
				unsigned short uMsgSize = static_cast<unsigned short>( sizeof( Asura_Steam_Msg_Header ) + sizeof( ASC_Msg_Voice ) + s_uVoiceRecDataSize );
				char* pcMsgData = static_cast<char*>( Asura_Allocate( uMsgSize ) );

				Asura_Steam_Msg_Header* pxMsgHdr = reinterpret_cast< Asura_Steam_Msg_Header* >( pcMsgData );

				pxMsgHdr->SetType( ASURA_STEAM_CLIENT_MESSAGE_VOICE );

				ASC_Msg_Voice* pxMsgBody = reinterpret_cast< ASC_Msg_Voice* >( pxMsgHdr->GetMessageBody() );
				
				Asura_Mem::Copy( pxMsgBody->GetDataAddress(), s_pcVoiceRecData, s_uVoiceRecDataSize );
				pxMsgBody->usDataSize = static_cast< unsigned short >( s_uVoiceRecDataSize );
				pxMsgBody->usChecksum = Asura_NetTransport_System::ComputeChecksum( pxMsgBody->GetDataAddress(), pxMsgBody->usDataSize );

				// loop over every peer and send the data to them
				Asura_List_It< Asura_SC_VoicePeerSocket > xIter( s_xVoicePeerSocketList );
				
				for( ; !xIter.Done(); xIter.Next() )
				{
					Asura_SC_VoicePeerSocket* pxPeer = xIter.GetCurrent();
					
					// Send the voice to the client... if the client is also the server use our primary socket
					if( pxPeer->m_eState == Asura_SC_VoicePeerSocket::STATE_CONNECTED )
					{	
						if( !SteamNetworking()->SendDataOnSocket( pxPeer->m_xSocket, pcMsgData, uMsgSize, false ) )
						{
							pxPeer->m_eState = Asura_SC_VoicePeerSocket::STATE_ERROR;
						}
					}
				}
				Asura_Deallocate( pcMsgData );
			}
		}
	}

	// Check if we need to remove any peers playing voice data

	Asura_List_ReverseIt< Asura_SC_PeerVoice > xVoiceIter( s_xPeerVoices );
	
	for( ; !xVoiceIter.Done(); xVoiceIter.Next() )
	{
		Asura_SC_PeerVoice* pxPeerVoice = xVoiceIter.GetCurrent();
		
		// If our voice playback is finished, destroy the peer entry in our list
		Asura_XAudio2_Voice* pxVoice = Asura_XAudio2::GetVoice( pxPeerVoice->m_xVoiceHandle );
		if( pxVoice && !pxVoice->IsPlaying() )
		{
			delete pxPeerVoice;
			xVoiceIter.DeleteCurrent();
		}
	}

	// Check the status of our sockets used for sending voice

	Asura_List_ReverseIt< Asura_SC_VoicePeerSocket > xPeerIter( s_xVoicePeerSocketList );

	for( ; !xPeerIter.Done(); xPeerIter.Next() )
	{
		Asura_SC_VoicePeerSocket* pxVoiceSocket = xPeerIter.GetCurrent();

		if( !pxVoiceSocket )
		{
			continue;
		}

		if( pxVoiceSocket->m_eState == Asura_SC_VoicePeerSocket::STATE_CONNECTING )
		{
			// Check if the socket is fully connected yet
			CSteamID	xIgnore1;
			uint32		uIgnore2;
			uint16		usIgnore3;
			int			iState;
			
			if( SteamNetworking()->GetSocketInfo( pxVoiceSocket->m_xSocket, &xIgnore1, &iState, &uIgnore2, &usIgnore3 ) )
			{
				if( iState == k_ESNetSocketStateConnected )
				{
					// The socket is now connected
					pxVoiceSocket->m_eState = Asura_SC_VoicePeerSocket::STATE_CONNECTED;
				}
				else if( iState >= k_ESNetSocketStateDisconnecting )
				{
					// Socket is disconnecting, possibly a legitimate disconnect, we trash it either way
					pxVoiceSocket->m_eState = Asura_SC_VoicePeerSocket::STATE_ERROR;
				}
			}
			else
			{
				// Failed to get socket info, trash it
				pxVoiceSocket->m_eState = Asura_SC_VoicePeerSocket::STATE_ERROR;
			}
		}

		if( pxVoiceSocket->m_eState == Asura_SC_VoicePeerSocket::STATE_ERROR )
		{
			// Destroy this socket
			delete pxVoiceSocket;
			xPeerIter.DeleteCurrent();
		}
	}
}


void Asura_PC_SteamworksClient::Shutdown ()
{
	ReleaseAuthBlob();
	
	while( s_xPeerVoices.GetNumEntries() > 0 )
	{
		delete s_xPeerVoices.PopHeadObject();
	}

	while( s_xVoicePeerSocketList.GetNumEntries() > 0 )
	{
		delete s_xVoicePeerSocketList.PopHeadObject();
	}

	while ( s_xPeerAuthTicketList.GetNumEntries() > 0 )
	{
		delete s_xPeerAuthTicketList.PopHeadObject();
	}

	Asura_Deallocate( s_pcVoiceRecData );
	s_pcVoiceRecData = 0;
	s_uVoiceRecDataSize = 0;

	delete s_pxVoicePlaybackMemHeap;
	s_pxVoicePlaybackMemHeap = 0;

	Asura_Deallocate( s_pcVoicePlaybackMemory );
	s_pcVoicePlaybackMemory = 0;
	
	if( SteamNetworking() && s_xVoiceListenSocket )
	{
		SteamNetworking()->DestroyListenSocket( s_xVoiceListenSocket, false );
		s_xVoiceListenSocket = 0;
	}
}


CSteamID Asura_PC_SteamworksClient::GetSteamID ()
{
	if( SteamUser() )
	{
		return SteamUser()->GetSteamID();
	}
	return CSteamID();
}


uint64 Asura_PC_SteamworksClient::GetSteamID_Int64 ()
{
	if( SteamUser() )
	{
		return SteamUser()->GetSteamID().ConvertToUint64();
	}
	return 0;
}


const Asura_UnicodeChar* Asura_PC_SteamworksClient::GetSteamIDText	()
{
	return s_awcSteamIDText;
}


bool Asura_PC_SteamworksClient::InitiateServerConnection ( const CSteamID xServerSteamID, const ASURA_PC_STEAMWORKS_AUTH_LEVEL eAuthLevel, const Asura_NetAddress xAddress, const unsigned short usPort )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( !SteamUser() )
	{
		Asura_Assert( false, "No steam user in steamworks" );
		return false;
	}

	ReleaseAuthBlob();

	// TODO: Sort out the voice comms
	/*if( !Asura_PC_SteamworksServer::IsServerRunning() )
	{
		// This is as good a time as any to create our listen socket for voice
		Asura_Assert( !s_pxListenSocket, "Listen socket should not be created here" );
		s_pxListenSocket = new Asura_PC_NetSocketSteamworks( ASURA_PC_STEAMWORKS_CLIENT_MAX_PEERS, ASURA_PC_STEAMSOCKET_TYPE_SERVER );
		
		unsigned short usVoicePort = ASURA_PC_STEAMWORKS_VOICE_PORT;
		s_pxListenSocket->Bind( Asura_NetAddress(0), usVoicePort );
	}*/

	s_pcAuthBlob = new char[ ASURA_PC_STEAM_MAX_AUTH_BLOB_SIZE ];

	// Now call InitiateGameConnection, this will give us our auth blob
	if( eAuthLevel > ASURA_PC_STEAMWORKS_NO_AUTHENTICATE )
	{
		const bool bVACSecure = ( eAuthLevel == ASURA_PC_STEAMWORKS_AUTHENTICATE_AND_VAC );

		s_uAuthBlobSize = SteamUser()->InitiateGameConnection( s_pcAuthBlob, ASURA_PC_STEAM_MAX_AUTH_BLOB_SIZE, xServerSteamID, xAddress.m_uAddress, usPort, bVACSecure );
	}
	else
	{
		// Dont authenticate, but still call InitiateGameConnection to communicate presence data
		s_uAuthBlobSize = SteamUser()->InitiateGameConnection( 0, 0, k_steamIDNonSteamGS, xAddress.m_uAddress, usPort, false );
	}

	s_bConnectedToServer = true;

	s_xServerAddress = xAddress;
	s_xServerAddress.m_xSteamID = xServerSteamID;
	s_usServerPort = usPort;

	return true;
}


bool Asura_PC_SteamworksClient::AcquireAuthBlob ( void* &pBlob, u_int &uBlobSize )
{
	if( !s_pcAuthBlob )
	{
		return false;
	}

	pBlob = s_pcAuthBlob;
	uBlobSize = s_uAuthBlobSize;

	return true;
}


void Asura_PC_SteamworksClient::ReleaseAuthBlob ( )
{
	if( s_pcAuthBlob )
	{
		delete []s_pcAuthBlob;
		s_pcAuthBlob = 0;
		s_uAuthBlobSize = 0;
	}
}


void Asura_PC_SteamworksClient::TerminateServerConnection ( )
{
	if( !IsConnectedToServer() )
	{
		return;
	}

	SteamUser()->TerminateGameConnection( s_xServerAddress.m_uAddress, s_usServerPort );

	s_bConnectedToServer = false;
	s_xServerAddress = Asura_NetAddress();
	s_usServerPort = 0;

	// Clear our client list
	while( s_xVoicePeerSocketList.GetNumEntries() )
	{
		delete s_xVoicePeerSocketList.PopHeadObject();
	}

	// Clear our peer voices
	while( s_xPeerVoices.GetNumEntries() )
	{
		delete s_xPeerVoices.PopHeadObject();
	}

	ReleaseAuthBlob();
}


void Asura_PC_SteamworksClient::StartVoiceRecording ( )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	// Prevent a conflict with steam voice chat
	if( SteamFriends() )
	{
		SteamFriends()->SetInGameVoiceSpeaking( GetSteamID(), true );
	}
	if( SteamUser() )
	{
		SteamUser()->StartVoiceRecording();
	}

	s_bVoiceRecActive = true;
}


void Asura_PC_SteamworksClient::StopVoiceRecording ( )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( SteamFriends() )
	{
		SteamFriends()->SetInGameVoiceSpeaking( GetSteamID(), false );
	}
	if( SteamUser() )
	{
		SteamUser()->StopVoiceRecording();
	}
	
	s_bVoiceRecActive = false;
}


void Asura_PC_SteamworksClient::HandleMessage_VoiceData ( const void* pData, const u_int uDataSize, const CSteamID xSender )
{
	const char* pcData = static_cast< const char* >( pData );
	const Asura_Steam_Msg_Header* pxMsgHdr = reinterpret_cast< const Asura_Steam_Msg_Header* >( pcData );

	if( !pxMsgHdr->IsValid() || ( pxMsgHdr->GetType() != ASURA_STEAM_CLIENT_MESSAGE_VOICE ) )
	{
		Asura_Assert( false, "Invalid steam message" );
		return;
	}

	const ASC_Msg_Voice* pxMsgBody = reinterpret_cast< const ASC_Msg_Voice* >( pxMsgHdr->GetMessageBody() );

	if( pxMsgBody->usChecksum != Asura_NetTransport_System::ComputeChecksum( pxMsgBody->GetDataAddress(), static_cast<int>( pxMsgBody->usDataSize ) ) )
	{
		Asura_Assert( false, "Checksum check failed on voice message" );
		return;
	}

	// See if we need to create a new voice for this peer
	Asura_SC_PeerVoice* pxPeerVoice = 0;

	Asura_List_It< Asura_SC_PeerVoice > xIter( s_xPeerVoices );

	for( ; !xIter.Done(); xIter.Next() )
	{
		pxPeerVoice = xIter.GetCurrent();
		if( pxPeerVoice && ( pxPeerVoice->m_xSteamID == xSender ) )
		{
			break;
		}
		pxPeerVoice = 0;
	}

	if( !pxPeerVoice )
	{
		pxPeerVoice = new Asura_SC_PeerVoice();

		if( pxPeerVoice )
		{
			s_xPeerVoices.AddToList( pxPeerVoice );

			pxPeerVoice->m_xSteamID = xSender;

			WAVEFORMATEX xFormat;
			xFormat.wFormatTag			= WAVE_FORMAT_PCM;
			xFormat.nChannels			= 1;
			xFormat.nSamplesPerSec		= 11025;
			xFormat.nAvgBytesPerSec		= 22050;	// 2 * 11025
			xFormat.nBlockAlign			= 2;
			xFormat.wBitsPerSample		= 16;
			xFormat.cbSize				= 0;

			// create the audio voice
			pxPeerVoice->m_xVoiceHandle = Asura_XAudio2::CreateVoice( xFormat, false, false );

			Asura_XAudio2_Voice* pxVoice = Asura_XAudio2::GetVoice( pxPeerVoice->m_xVoiceHandle );

			if( pxVoice )
			{
				pxVoice->SetBufferEndCallback( AudioVoiceBufferCallback );
			}
		}
	}
	
	if( !pxPeerVoice )
	{
		Asura_Assert( false, "Could not create peer voice structure" );
		return;
	}

	// Decompress the received data
	Asura_Assert( s_pxVoicePlaybackMemHeap, "No voice heap" );

	// This is a bit big to go on the stack really
	char* pcDecompressBuffer[ ASURA_PC_STEAMWORKS_CLIENT_DECOMPRESS_BUFFER_SIZE ];
	uint32 iBytesWritten = 0;

	EVoiceResult xResult = SteamUser()->DecompressVoice( const_cast<char*>( pxMsgBody->GetDataAddress() ), pxMsgBody->usDataSize, pcDecompressBuffer, ASURA_PC_STEAMWORKS_CLIENT_DECOMPRESS_BUFFER_SIZE, &iBytesWritten );

	Asura_Assertf( xResult == k_EVoiceResultOK, ( "Voice decompress failed (%d)", xResult ) );

	if( xResult == k_EVoiceResultOK )
	{
		// Put this data on our memheap and pass it as a new buffer to the voice
		char* pcNewVoiceData = static_cast<char*>( s_pxVoicePlaybackMemHeap->Alloc( iBytesWritten ) );
		if( pcNewVoiceData )
		{
			Asura_Mem::Copy( pcNewVoiceData, pcDecompressBuffer, iBytesWritten );

			Asura_XAudio2_Voice* pxVoice = Asura_XAudio2::GetVoice( pxPeerVoice->m_xVoiceHandle );

			if( pxVoice )
			{
				pxVoice->EnqueueData( pcNewVoiceData, static_cast<u_int>( iBytesWritten ), false, pcNewVoiceData );
				if( !pxVoice->IsPlaying() )
				{
					pxVoice->Start();
				}
			}
		}
	}
}


void Asura_PC_SteamworksClient::AudioVoiceBufferCallback ( void* pContext )
{
	// The context will contain the address of the buffer so we can free it
	Asura_Assert( s_pxVoicePlaybackMemHeap, "No voice heap" );

	s_pxVoicePlaybackMemHeap->Free( pContext );
}


bool Asura_PC_SteamworksClient::IsPeerVoicePlaying ( const CSteamID xSteamID )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	Asura_List_It< Asura_SC_PeerVoice > xIter( s_xPeerVoices );

	for( ; !xIter.Done(); xIter.Next() )
	{
		Asura_SC_PeerVoice* pxPeerVoice = xIter.GetCurrent();
		if( pxPeerVoice->m_xSteamID == xSteamID )
		{
			return true;
		}
	}
	
	return false;
}

bool Asura_PC_SteamworksClient::IsMicrophoneAttached()
{
	return s_bMicrophoneAttached;
}

bool Asura_PC_SteamworksClient::WriteToRemoteStorage ( const void* pData, const u_int uDataSize, const char* szFilename )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( !SteamRemoteStorage() )
	{
		Asura_Assert( false, "No remote storage instance in steamworks");
		return false;
	}

	return SteamRemoteStorage()->FileWrite( szFilename, pData, static_cast<int32>( uDataSize ) );
}


bool Asura_PC_SteamworksClient::ReadFromRemoteStorage ( void* pData, const u_int uDataSize, const char* szFilename )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( !SteamRemoteStorage() )
	{
		Asura_Assert( false, "No remote storage instance in steamworks");
		return false;
	}

	int32 iToRead = static_cast<int32>( uDataSize );

	return ( SteamRemoteStorage()->FileRead( szFilename, pData, iToRead ) == iToRead );
}


bool Asura_PC_SteamworksClient::RemoteStorageFileExists ( const char* szFilename )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( !SteamRemoteStorage() )
	{
		Asura_Assert( false, "No remote storage instance in steamworks");
		return false;
	}

	return SteamRemoteStorage()->FileExists( szFilename );
}


bool Asura_PC_SteamworksClient::IsFriend ( const CSteamID xSteamID )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( SteamFriends() )
	{
		return ( SteamFriends()->GetFriendRelationship( xSteamID ) == k_EFriendRelationshipFriend );
	}
	return false;
}


Asura_NetAddress Asura_PC_SteamworksClient::GetServerAddress ( )
{
	if( IsConnectedToServer() )
	{
		return s_xServerAddress;
	}
	Asura_Assert( false, "Steamworks client is not connected to a server, cannot return server address");

	return Asura_NetAddress();
}


unsigned short Asura_PC_SteamworksClient::GetServerPort( )
{
	if( IsConnectedToServer() )
	{
		return s_usServerPort;
	}
	Asura_Assert( false, "Steamworks client is not connected to a server, cannot return server address");

	return 0;
}


bool Asura_PC_SteamworksClient::InitiateP2PConnection ( const CSteamID xSteamID )
{
	// We are connecting to another machine

	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( !SteamUser() )
	{
		Asura_Assert( false, "No steam user in steamworks" );
		return false;
	}

	if( FindAuthTicket( xSteamID ) )
	{
		Asura_Assert( false, "Trying to establish a connection to a client we already have a ticket for" );
		return false;
	}

	Asura_PC_Steamworks_AuthTicket* pxTicket = new Asura_PC_Steamworks_AuthTicket();

	if( pxTicket )
	{
		pxTicket->m_xSteamID	= xSteamID;
		pxTicket->m_pcAuthData	= static_cast<char*>( Asura_Allocate( ASURA_PC_STEAM_MAX_AUTH_TICKET_SIZE ) );
		pxTicket->m_xTicket		= SteamUser()->GetAuthSessionTicket( pxTicket->m_pcAuthData, ASURA_PC_STEAM_MAX_AUTH_TICKET_SIZE, &pxTicket->m_uAuthDataSize );

		if( pxTicket->m_xTicket != k_HAuthTicketInvalid )
		{
			s_xPeerAuthTicketList.AddToList( pxTicket );

			return true;
		}
		else
		{
			delete pxTicket;
		}
	}

	return false;
}


void Asura_PC_SteamworksClient::TerminateP2PConnection ( const CSteamID xSteamID )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	// Find this connection in our list
	Asura_PC_Steamworks_AuthTicket* pxTicket = FindAuthTicket( xSteamID );

	if( !pxTicket )
	{
		//Asura_Assert( false, "Failed to find auth ticket" );
		return;
	}

	SteamUser()->CancelAuthTicket( pxTicket->m_xTicket );

	s_xPeerAuthTicketList.RemoveFromList( pxTicket );
	delete pxTicket;
}


bool Asura_PC_SteamworksClient::HandleUserConnection ( const void* pAuthData, const u_int uAuthDataSize, CSteamID xSteamID )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	if( !SteamUser() )
	{
		Asura_Assert( false, "No steam user in steamworks" );
		return false;
	}

	int iAuthDataSize = static_cast< int >( uAuthDataSize );

	// Authorise a connecting user (Callback_TicketResponse will be called)
	
	EBeginAuthSessionResult xResult = SteamUser()->BeginAuthSession( pAuthData, iAuthDataSize, xSteamID );

	return ( xResult == k_EBeginAuthSessionResultOK );
}


void Asura_PC_SteamworksClient::HandleUserDisconnection ( const CSteamID xSteamID )
{
	ASURA_PC_STEAM_RETURN_VOID_IF_UNINITIALISED

	if( !SteamUser() )
	{
		Asura_Assert( false, "No steam user in steamworks" );
		return;
	}

	SteamUser()->EndAuthSession( xSteamID );
}


Asura_PC_Steamworks_AuthTicket* Asura_PC_SteamworksClient::FindAuthTicket ( const CSteamID xSteamID )
{
	ASURA_PC_STEAM_RETURN_ZERO_IF_UNINITIALISED

	Asura_List_It< Asura_PC_Steamworks_AuthTicket > xIter( s_xPeerAuthTicketList );

	for( ; !xIter.Done(); xIter.Next() )
	{
		Asura_PC_Steamworks_AuthTicket* pxTicket = xIter.GetCurrent();

		if( pxTicket && ( pxTicket->m_xSteamID == xSteamID ) )
		{
			return pxTicket;
		}
	}

	return 0;
}


void Asura_PC_SteamworksClient::Callback_TicketResponse ( const ValidateAuthTicketResponse_t* const pxParam )
{
	if( pxParam->m_eAuthSessionResponse == k_EAuthSessionResponseOK )
	{
		RunCallback_AcceptClient( pxParam->m_SteamID );
	}
	else
	{
		RunCallback_RejectClient( pxParam->m_SteamID, static_cast<u_int>( pxParam->m_eAuthSessionResponse ) );
	}
}


void Asura_PC_SteamworksClient::RunCallback_AcceptClient ( const CSteamID xSteamID )
{
	if( s_pfnAsuraCallback_ClientAccept )
	{
		s_pfnAsuraCallback_ClientAccept( xSteamID, s_pCallbackUserData_Accept );
	}
}


void Asura_PC_SteamworksClient::RunCallback_RejectClient ( const CSteamID xSteamID, const u_int uReason )
{
	if( s_pfnAsuraCallback_ClientReject )
	{
		s_pfnAsuraCallback_ClientReject( xSteamID, uReason, s_pCallbackUserData_Reject );
	}
}


void Asura_PC_SteamworksClient::RegisterCallback_ClientAccept ( Asura_PC_SteamClient_Callback_ClientAccept pfnCallback, void* pUserData )
{
	s_pfnAsuraCallback_ClientAccept = pfnCallback;
	s_pCallbackUserData_Accept = pUserData;
}


void Asura_PC_SteamworksClient::RegisterCallback_ClientReject ( Asura_PC_SteamClient_Callback_ClientReject pfnCallback, void* pUserData )
{
	s_pfnAsuraCallback_ClientReject = pfnCallback;
	s_pCallbackUserData_Reject = pUserData;
}


bool Asura_PC_SteamworksClient::GetAuthTicketForClient ( const CSteamID xSteamID, char* &pcAuthData, u_int &uAuthDataSize )
{
	ASURA_PC_STEAM_RETURN_FALSE_IF_UNINITIALISED

	Asura_PC_Steamworks_AuthTicket* pxTicket = FindAuthTicket ( xSteamID );

	if( pxTicket )
	{
		pcAuthData = pxTicket->m_pcAuthData;
		uAuthDataSize = pxTicket->m_uAuthDataSize;

		return true;
	}

	return false;
}


void Asura_PC_SteamworksClient::ClearAuthTicketForClient( const CSteamID xSteamID )
{
	Asura_PC_Steamworks_AuthTicket* pxTicket = FindAuthTicket ( xSteamID );

	if( pxTicket )
	{
		pxTicket->ClearAuthData ();
	}
}


void Asura_PC_SteamworksClient::SetVoiceSendList ( const CSteamID* pxPeerList, const u_int uListSize )
{
	CSteamID axNewPeerList[ 64 ];

	// Make a copy of the list so we can modify it

	Asura_Mem::Copy( &axNewPeerList[0], pxPeerList, Asura_Maths::Min( uListSize, (u_int)64 ) * sizeof( CSteamID ) );

	Asura_List_ReverseIt< Asura_SC_VoicePeerSocket > xIter( s_xVoicePeerSocketList );

	for( ; !xIter.Done(); xIter.Next() )
	{
		bool bFound = false;

		Asura_SC_VoicePeerSocket* pxVoicePeer = xIter.GetCurrent();

		for( u_int uListItem = 0; uListItem < uListSize; ++uListItem )
		{
			// Check if this peer is in our list
			if( axNewPeerList[ uListItem ].IsValid() && (pxVoicePeer->m_xSteamID == axNewPeerList[ uListItem ] ) )
			{
				axNewPeerList[ uListItem ] = CSteamID();
				bFound = true;
				break;
			}
		}

		if( !bFound )
		{
			// We need to delete this from our list
			delete pxVoicePeer;
			xIter.DeleteCurrent();
		}
	}

	// We have performed removal where necessary, now check for additions

	for( u_int uListItem = 0; uListItem < uListSize; ++uListItem )
	{
		if( axNewPeerList[ uListItem ].IsValid() )
		{
			// If valid, this peer needs adding
			Asura_SC_VoicePeerSocket* pxNewVoicePeer = new Asura_SC_VoicePeerSocket( axNewPeerList[ uListItem ] );
			
			s_xVoicePeerSocketList.AddToList( pxNewVoicePeer );
		}
	}
}


#endif	// ASURA_PC_SUPPORT_STEAMWORKS

//eof
