
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_SCEUtility_SaveListHandler.cpp
// Author : Andy Tate
// Date   : 5-Oct-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "PSPEmul/SCEUtility/Asura_PSP_SCEUtility_SaveListHandler.h"

// Asura independent header-files...
#include "Core/Asura_Checksum.h"
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Font_System.h"

// Asura platform header-files...
#include "FileAccess/Asura_DebugFileStream.h"

// Game independent header-files...

// Game platform header-files...


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const Asura_IntVector_2	g_xSaveListNewDataPosition (240, 168);


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_SaveListHandler::OnSelect
// Author       : Andy Tate
// Date         : 5-Oct-2005
//
// Description  : Handles the saving of the specified slot upon selection.
//
// Parameters   : A reference to the slot data to use.
//
// Return value : "true" if successful, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SCEUtility_SaveListHandler::OnSelect (Asura_PSP_SCEUtility_FileData& xFileData)
{
	const u_int	uTotalSize	= m_uBufferSize + Asura_Checksum::CHECKSUM_LENGTH;
	u_char*		pucTemp		= new u_char [uTotalSize];

	bool		bSuccess;

	// Update the file-data...
	xFileData.UpdateHeaderData	(m_wszSaveTitle, m_wszSaveDescription, uTotalSize);

	// Checksum the data...
	memset								(pucTemp, 0x00, uTotalSize);
	Asura_Checksum::EncodeFromBuffer	(static_cast<u_char const*>(m_pvData), m_uDataSize, pucTemp);
	memcpy								(&pucTemp [Asura_Checksum::CHECKSUM_LENGTH], m_pvData, m_uDataSize);

	// Write the data to this file...
	bSuccess	= xFileData.WriteData( uTotalSize, pucTemp );

	Asura_Assert( bSuccess, "Cannot write data to file.");

	// Delete the temporary storage before returning...
	delete[]( pucTemp );
	return( bSuccess );
}

/// End of function 'Asura_PSP_SCEUtility_SaveListHandler::OnSelect'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_SaveListHandler::RenderEmptySlot
// Author       : Andy Tate
// Date         : 5-Oct-2005
//
// Description  : Renders the data for this empty slot.
//
// Parameters   : A reference to the current slot data and the screen-scale to
//				  use when calculating positions.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_SCEUtility_SaveListHandler::RenderEmptySlot (const float fScreenScale,
															const Asura_PSP_SCEUtility_FileData& xFileData) const
{
	Asura_Vector_2	xPosition;

	// Call the base-function...
	Asura_PSP_SCEUtility_FileListHandler::RenderEmptySlot	(fScreenScale, xFileData);

	// Calculate the current slot-data positions...
	CalculateScreenPosition	(fScreenScale, g_xSaveListNewDataPosition, xPosition);

	// Display the current slot's title...
	Asura_Font_System::PrintLongString
	(
		m_wszNewData,
		NULL,
		xPosition [0],
		xPosition [1],
		1.0f,
		0.5f * fScreenScale,
		0xFF808080,
		ASURA_TEXTFORMAT_LEFTJUSTIFIED
	);
}

/// End of function 'Asura_PSP_SCEUtility_SaveListHandler::RenderEmptySlot'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_SaveListHandler::Asura_PSP_SCEUtility_SaveListHandler
// Author       : Andy Tate
// Date         : 5-Oct-2005
//
// Description  : Constructor function for the save-list handler.
//
// Parameters   : 
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_PSP_SCEUtility_SaveListHandler::Asura_PSP_SCEUtility_SaveListHandler (const int iMaxSaveSlots,
																			const Asura_UnicodeChar* wszSaveTitle,
																			const Asura_UnicodeChar* wszSaveDescription,
																			const Asura_UnicodeChar* wszNewData,
																			const void* pvData,
																			const u_int uDataSize,
																			const u_int uBufferSize) :
	Asura_PSP_SCEUtility_FileListHandler(L"Select Slot To Save Data", iMaxSaveSlots, true),
	m_pvData(pvData),
	m_uDataSize(uDataSize),
	m_uBufferSize(uBufferSize)
{
	// Allocate the necessary strings...
	m_wszSaveTitle			= new Asura_UnicodeChar [wcslen(wszSaveTitle) + 1];
	m_wszSaveDescription	= new Asura_UnicodeChar [wcslen(wszSaveDescription) + 1];
	m_wszNewData			= new Asura_UnicodeChar [wcslen(wszNewData) + 1];

	// ...and copy the data...
	wcscpy	(m_wszSaveTitle,		wszSaveTitle);
	wcscpy	(m_wszSaveDescription,	wszSaveDescription);
	wcscpy	(m_wszNewData,			wszNewData);
}

/// End of function 'Asura_PSP_SCEUtility_SaveListHandler::Asura_PSP_SCEUtility_SaveListHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_SaveListHandler::~Asura_PSP_SCEUtility_SaveListHandler
// Author       : Andy Tate
// Date         : 5-Oct-2005
//
// Description  : Constructor function for the save-list handler.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_PSP_SCEUtility_SaveListHandler::~Asura_PSP_SCEUtility_SaveListHandler ()
{
	// Delete the necessary strings...
	delete []	(m_wszSaveTitle);
	delete []	(m_wszSaveDescription);
	delete []	(m_wszNewData);
}

/// End of function 'Asura_PSP_SCEUtility_SaveListHandler::~Asura_PSP_SCEUtility_SaveListHandler'. ///


/// End of file. ///
