// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_PC_WindowsRegistry.h													//
//																						//
//	Description: This supplies access to the window registry							//
//																						//
//	Created: Steve Keen & Mark Parry - 25/02/08											//
//																						//
//	Copyright (C) 2008 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"

#ifdef ASURA_PC_SUPPORT_WINDOWS_REGISTRY

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef _ASURA_PC_WINDOWSREGISTRY_H_
#define _ASURA_PC_WINDOWSREGISTRY_H_

enum ASURA_WINDOWS_REGISTRY_VALUE_TYPE
{
	ASURA_WINDOWS_REGISTRY_VALUE_TYPE_UNKNOWN	= -2,
	ASURA_WINDOWS_REGISTRY_VALUE_TYPE_ERROR		= -1,

	ASURA_WINDOWS_REGISTRY_VALUE_TYPE_NUMBER		= 0,
	ASURA_WINDOWS_REGISTRY_VALUE_TYPE_STRING,
	ASURA_WINDOWS_REGISTRY_VALUE_TYPE_BINARY,
	ASURA_WINDOWS_REGISTRY_VALUE_TYPE_UNSUPPORTED,

};

//return false to stop enumerating
typedef bool (*ASURA_ENUMERATE_REGISTRY_KEYS_CALLBACK)(
	const int index,							//entry index
	const char* const pszKeyName,				//enumerated key's name
	const char* const pszKeyClass,				//enumerated key's class
	const char* const pszParentKey,				//the key that is being enumerated
	const char* const pszParentKeyTidied,		//the key that is being enumerated (the tidied name)
	void* const pUserData);						//user data

//return false to stop enumerating
typedef bool (*ASURA_ENUMERATE_REGISTRY_VALUES_CALLBACK)(
	const int iIndex,										//entry index
	const char* const pszValueName,							//enumerated value's name
	void* const pValueBuffer,								//the value (if a buffer was passed to the enumerate function)
	const int iValueSize,									//the size of the value
	const bool bValueTooLargeForBuffer,						//was the value too large for the passed buffer?
	const ASURA_WINDOWS_REGISTRY_VALUE_TYPE eValueType,		//the processed value type
	u_int uValueTypeUnprocessed,							//the unprocessed value type
	const char* const pszParentKey,							//the key that is being enumerated
	const char* const pszParentKeyTidied,					//the key that is being enumerated (the tidied name)
	void* const pUserData);									//user data

//----------------------------------------------------------------------------------------------

class Asura_PC_WindowsRegistry
{
public:
	enum
	{
		eTidySubKeySize = 256,
		eBufferSizeCheckIfChanged = 8192 
	};
	
	Asura_PC_WindowsRegistry();
	~Asura_PC_WindowsRegistry();

//----------------------------------------------------------------------------------------------
//the default master key
private:
	static HKEY m_hDefaultMasterKeyHandle;
public:
	static HKEY GetDefaultMasterKey( void);
	static void SetDefaultMasterKey( HKEY hDefaultMasterKeyHandle);
	static void SetDefaultMasterKey_ClassesRoot( void);
	static void SetDefaultMasterKey_CurrentUser( void);
	static void SetDefaultMasterKey_LocalMachine( void);
	static void SetDefaultMasterKey_Users( void);

	static HKEY GetMasterKeyClassesRoot( void);
	static HKEY GetMasterKeyCurrentUser( void);
	static HKEY GetMasterKeyLocalMachine( void);
	static HKEY GetMasterKeyUsers( void);

//----------------------------------------------------------------------------------------------
private:
	//tidy the subKey (remove '/' off the end if necessary, some versions of windows don't like it!)
	static bool TidySubKey( char* const pszOutputSubKey, const char* const pszSubKey, HKEY& pMasterKeyHandle);
	static bool OverrideKeyHandle( const char* &pszSubKey, HKEY& hMasterKeyHandle, const char* const pszOverrideName, HKEY hOverrideMasterKeyHandle);
	static bool OverrideKeyHandle( const char* &pszSubKey, HKEY& hMasterKeyHandle);

	//static bool OverrideKeyHandle( const char* &pszSubKey, void* &pMasterKeyHandle, const char* const pszOverrideName, void* const pOverrideMasterKeyHandle);
	//static bool OverrideKeyHandle( const char* &pszSubKey, void* &pMasterKeyHandle);

//----------------------------------------------------------------------------------------------
//open / create / close / Delete a subkey
private:
	HKEY m_hSubKeyHandle;
	bool m_bSubKeyOpen;
	bool m_bCreatedSubKeyAlreadyExisted;

public:
	bool CloseSubKey(void);
	bool OpenSubKey(const char* const pszSubKey, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool CreateSubKey(const char* const pszSubKey, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool DeleteSubKey(const char* const pszSubKey, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool DoesSubKeyExists(const char* const pszSubKey, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);

//----------------------------------------------------------------------------------------------
//get value type and size, get value, set value and Delete value on an already open subkey
private:
	u_int m_uValueTypeUnprocessed;
	int	  m_iValueSize;
	ASURA_WINDOWS_REGISTRY_VALUE_TYPE m_eValueTypeProcessed;

	bool GetValueTypeAndSizeFromOpen(const char* const pszValueName);
	bool CheckProcessedValueTypeAndSizeFromOpen( const char* const pszValueName,
												const ASURA_WINDOWS_REGISTRY_VALUE_TYPE eProcessedValueType,
												const int iValueSize,
												const bool bReportSizeDifference = true);

	bool GetValueFromOpen(const char* const pszValueName, void* const pDataBuffer, const int iDataSize, int* const piOutputSize);
	bool SetValueFromOpen(const char* const pszValueName, const void* pDataBuffer, const int iDataSize, const u_int uValueType);
	bool DeleteValueFromOpen(const char* const pszValueName);

public:
	static ASURA_WINDOWS_REGISTRY_VALUE_TYPE GetProcessedValueType(const u_int uValueTypeUnprocessed);
	static const char* GetProcessedValueTypeString(const ASURA_WINDOWS_REGISTRY_VALUE_TYPE eValueType);
	const char* GetProcessedValueTypeString(void)	const		{ return(GetProcessedValueTypeString(m_eValueTypeProcessed)); }

	//read from an open subkey
	bool ReadNumberFromOpen(const char* const pszValueName, u_int* const puOutputNumber);
	bool ReadStringFromOpen(const char* const pszValueName, char* const pszStringBuffer, const int iStringBufferSize, int* const piOutputStringSize = NULL);
	bool ReadBinaryFromOpen(const char* const pszValueName, void* const pbuffer, const int iBufferSize, int* const piOutputSize = NULL);

	//write to an open subkey
	bool WriteNumberFromOpen(const char* const pszValueName, const unsigned int number);
	bool WriteStringFromOpen(const char* const pszValueName, const char* const stringBuffer, int stringSize = -1);
	bool WriteBinaryFromOpen(const char* const pszValueName, const void* const buffer, const int bufferSize);
	bool WriteNumberIfChangedFromOpen(const char* const pszValueName, const u_int uNumber);
	bool WriteStringIfChangedFromOpen(const char* const pszValueName, const char* const pszStringBuffer, int iStringSize = -1);
	bool WriteBinaryIfChangedFromOpen(const char* const pszValueName, const void* const pBuffer, const int iBufferSize);

//----------------------------------------------------------------------------------------------
// ok, the calls that are intended for general use!
//----------------------------------------------------------------------------------------------

public:
	//read
	bool ReadNumber(const char* const pszSubKey, const char* const pszValueName, unsigned int* piOutputNumber, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool ReadString(const char* const pszSubKey, const char* const pszValueName, char* const pszStringBuffer, const int iStringBufferSize, int* const piOutputStringSize = NULL, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool ReadBinary(const char* const pszSubKey, const char* const pszValueName, void* const pBuffer, const int iBufferSize, int* const piOutputSize = NULL, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);

	//write
	bool WriteNumber(const char* const pszSubKey, const char* const pszValueName, const u_int uNumber, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool WriteString(const char* const pszSubKey, const char* const pszValueName, const char* const pszStringBuffer, const int iStringSize = -1, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool WriteBinary(const char* const pszSubKey, const char* const pszValueName, const void* const pBuffer, const int iBufferSize, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool WriteNumberIfChanged(const char* const pszSubKey, const char* const pszValueName, u_int uNumber, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool WriteStringIfChanged(const char* const pszSubKey, const char* const pszValueName, const char* const pszStringBuffer, const int iStringSize = -1, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool WriteBinaryIfChanged(const char* const pszSubKey, const char* const pszValueName, const void* const pBuffer, const int iBufferSize, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);

	bool DeleteValue(const char* const pszSubKey, const char* const pszValueName, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	bool DoesValueExist(const char* const pszSubKey, const char* const pszValueName, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);

	static int EnumerateKeys(const char* const pszSubKey, ASURA_ENUMERATE_REGISTRY_KEYS_CALLBACK callBack, void* const pUserData, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
	static int EnumerateValues(const char* const pszSubKey, ASURA_ENUMERATE_REGISTRY_VALUES_CALLBACK callBack, void* const pUserData, void* const pValueBuffer = NULL, const int iVvalueBufferSize = 0, HKEY hMasterKeyHandle = m_hDefaultMasterKeyHandle);
};

#endif // _REGISTRY_H_

#endif //ASURA_PC_SUPPORT_WINDOWS_REGISTRY

//eof
