//
// Asura_Direct3D_EffectParameterBlock.h
//
// Created by RJF - 11 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_DIRECT3D_EFFECTPARAMETERBLOCK_H__
#define __ASURA_DIRECT3D_EFFECTPARAMETERBLOCK_H__

#include "Core/Asura_String.h"
#include "Rendering/Asura_Direct3D.h"
#include "Core/Asura_List.h"

class Asura_Direct3D_Effect;
class Asura_RGBA;

//////////////////////////////////////////////////////////////////////////
// Asura_Direct3D_EffectParameterBlock
// -----------------------------------
// Asura_Direct3D_EffectParameterBlock is an object that stores a group of effect/shader parameters
// in a platform-efficient format, allowing parameters to be set from client code with a single call
// instead of a series of Effect->SetParameter() calls.
//
// Designed by KJFL and RJF, original implementation by RJF. Any questions, ask.
//
//////////////////////////////////////////////////////////////////////////
//
// Clients:
// --------
// It's pretty simple to use. You can do the following things to an Asura_Direct3D_EffectParameterBlock:
//
// * Create/delete it:		Just use the standard new/delete operators. Nothing special.
//
// * Add parameters to it:  There are two kinds of parameters, 'static' parameters and 'dynamic' parameters.
//
//                          Static parameters are parameters that never change; when you add the parameter 
//                          intially, you give it a value, and this value is stored in the block. You cannot
//                          change the value once the parameter has been added.
//
//                          Dynamic parameters don't take direct values, but take pointers to them instead; 
//                          when the block is used to set shader parameters, the pointer is dereferenced and
//                          the current value used to set the shader parameter. This way you can 'bind' a 
//							shader parameter to a (persistant) variable.
//
//                          Note as well that parameters are generally considered as arrays; setting a single 
//							value is equivalent to setting an array with a size of 1. Shortcut functions are 
//							in place to make this a little simpler for static parameters.
//
// * Bind it:				This binds the parameter block to a particular effect, allowing it to be optimized.
//							New parameters cannot be added once the block is bound. Don't try binding the block
//							to an effect which is currently active (i.e. between BeginEffect/FlushEffects), it 
//							won't work. In general, you should always bind.
//
// * Apply it:				This sends the parameter values to the graphics hardware, setting them on the shader. 
//							If the block hasn't been bound, then you can specify which effect it should set them
//							on. (If the block has been bound, then you can only apply it to the effect you bound
//							it to, by passing NULL or nothing).
//
//////////////////////////////////////////////////////////////////////////
//
// Maintainers:
// ------------
// EffectParameterBlock has been written in as generic a manner as possible whilst still maintaing type-safety.
// The block itself doesn't need to know anything about the types of the parameters it is wrangling - the types
// only need to be unwrapped when the parameter is actually being set on the effect, so that we know which Set 
// function to call. Thus we deal with three internal support classes (or groups of support classes):
//
// * Parameter:					Abstract base type for a parameter. Contains things common to any kind of 
//								parameter, like the name and handle. Also contains a pure virtual function
//								Apply() that is partially implemented by...
//
// * TypedParameter<>:			A templated type, derived from Parameter, that actually stores parameter values.
//								Manages the temporary storage needed by static parameters (the storage in which 
//								copies of values are stored at add-parameter time). It implements the Apply 
//								function, but just forwards the call to...
//
// * ParameterSettingPolicy<>:  A group of templated policy classes that define exactly how a particular type of
//								parameter is set on the effect. Each just contains a static Apply() function. The
//								base template's implementation simply asserts; the class is then specialized to 
//								create a policy for each supported type.
//
// Last but not least, the Asura_Direct3D_EffectParameterBlock class itself provides templated AddStaticParameter and
// AddDynamicParameter functions that create the required TypedParameter<> objects and adds them to the list.
//
// When parameters are initially added, they store their names (as strings). Binding the block to an effect cashes in
// these names for handles, which are a better plan when setting parameters. The names can then be discarded; the exception
// to this is in Tools builds, where effects can be reloaded at runtime. In that situation we keep the names so that we can
// retrieve fresh handles after the effect the block is bound to gets reloaded.
//
// Binding also converts static parameters into a D3DXEffect Parameter Block, which is (in theory) more optimal than making
// a bunch of SetParameter calls. When this conversion is done, the static parameters are completely removed from the block
// to save memory. (Again, the exception is in Tools builds - the parameters are kept in case the block needs to be rebuilt
// after an effect reload).
//
// Applying the block does two things: It applies the D3DXEffect parameter block if one exists, and then it looks through
// the remaining (dynamic) parameters, applying each one of them individually.
//
// If you want to add support for a new parameter type, all you should need to do is add a new ParameterSettingPolicy<>
// specialization at the end of this file. 
//
//////////////////////////////////////////////////////////////////////////
//
// Cor, that was a lot of text, wasn't it? C'mon, grab your safari hat and let's head out...

class Asura_Direct3D_EffectParameterBlock
{
public:

	// Constructor/destructor: as you can see, nothing special.
	Asura_Direct3D_EffectParameterBlock();
	~Asura_Direct3D_EffectParameterBlock();

protected:

	// The base parameter type. Contains data and functionality common to all parameters.
	struct Parameter
	{
		Parameter(const Asura_Char* szParameterName, bool bIsStatic);
		virtual ~Parameter() { } // required for derived type-specific classes to be able to destruct correctly

		const Asura_Char* m_szParameterName;
		D3DXHANDLE m_xParameterHandle;
		bool m_bIsStatic;
		D3DXHANDLE GetHandle(Direct3D_Effect* pxEffect);

		// Implemented by the derived, type-specific classes:
		virtual void Apply(Direct3D_Effect* pxEffect) = 0;
	};

	// Base, unspecialised ParameterSettingPolicy. Only used if someone tries throwing us a parameter type that isn't supported.
	template<class ParameterType>
	struct ParameterSettingPolicy
	{
		static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, ParameterType* pxData, u_int uNumValues)
		{
			// If you hit this assert, check the call stack to see what class you're in - the template parameter should give away the problem.
			Asura_Assert(false, "Parameter setting policy is not defined for this parameter type.");
		}
	};

	// The derived, type-specific parameter class. Handles value data (external pointer for dynamic params, internally 
	// allocated array for static) and makes use of the policy classes for implementing Apply().
	template<class ParameterType>
	struct TypedParameter : Parameter
	{
		ParameterType* m_pxValues;
		u_int m_uNumValues;

		TypedParameter(const Asura_Char* szParameterName, 
			bool bIsStatic,
			ParameterType* pxValues,
			u_int uNumValues)
			: Asura_Direct3D_EffectParameterBlock::Parameter(szParameterName, bIsStatic)
		{
			if(m_bIsStatic)
			{
				// Allocate storage to keep a copy of the passed-in values
				m_pxValues = new ParameterType[uNumValues];
				memcpy(m_pxValues, pxValues, sizeof(ParameterType) * uNumValues);
			}
			else
			{
				// store the external pointer
				m_pxValues = pxValues;
			}
			m_uNumValues = uNumValues;
		}

		void Apply(Direct3D_Effect* pxEffect)
		{
			D3DXHANDLE xHandle = GetHandle(pxEffect);
			Asura_Assertf(xHandle, ("Trying to apply shader parameter '%s', but it's unbound.", m_szParameterName));
			if(!xHandle) return;
			ParameterSettingPolicy<ParameterType>::Apply(pxEffect, xHandle, m_pxValues, m_uNumValues);
		}

		~TypedParameter()
		{
			if(m_bIsStatic)
			{
				// Delete the storage we allocated
				delete[] m_pxValues;
			}
		}
	};

public:

	// Syntactic sugar for single-value parameters. Just forwards them on as single-value arrays.
	// Note that the data is pass-by-value here, while it's pass-by-pointer in t'other one.
	template<typename ParameterType>
	void AddStaticParameter(const Asura_Char* szParameterName, ParameterType xValue)
	{
		AddStaticParameter(szParameterName, &xValue, 1);
	}
	
	// Add parameters. Why not just let people call AddParameter directly? Because this will make 
	// people write clearer-to-read code, and make them think more about which function they should be calling.
	template<typename ParameterType>
	void AddStaticParameter(const Asura_Char* szParameterName, ParameterType* pxValues, u_int uNumValues)		{ AddParameter(szParameterName, pxValues, uNumValues, true); }

	template<typename ParameterType>
	void AddDynamicParameter(const Asura_Char* szParameterName, ParameterType* pxValues, u_int uNumValues = 1)	{ AddParameter(szParameterName, pxValues, uNumValues, false); }

	void Bind(Asura_Direct3D_Effect* pxEffectToBindTo);
	void Apply(Asura_Direct3D_Effect* pxEffectToApplyTo = NULL);

	bool IsBound() const { return m_pxEffect != NULL; }

protected:

	template<typename ParameterType>
		void AddParameter(const Asura_Char* szParameterName, ParameterType* pxValues, u_int uNumValues, bool bIsStatic)
	{
		Asura_Assert(!IsBound(), "Adding a new parameter to a block that is already bound");
		if(IsBound()) return;
		m_xParameters.AddToList(new TypedParameter<ParameterType>(szParameterName, bIsStatic, pxValues, uNumValues));
	}

	Asura_List<Parameter> m_xParameters;
	Asura_Direct3D_Effect* m_pxEffect;
	D3DXHANDLE m_xStaticParameterBlock; // The D3DX Parameter Block handle.

#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
	static Asura_List<Asura_Direct3D_EffectParameterBlock> s_xBlockList;
	public:
	static void Tools_RegenerateAll();
#endif
};

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// ParameterSettingPolicy Specializations
// --------------------------------------
// Each one of these contains the code for submitting one particular type of parameter to the effect.
// It's assumed that an array of size 1 is to be treated as a single element.

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<float>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, float* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetFloat(xHandle, *pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			xResult = pxEffect->SetFloatArray(xHandle, pxData, uNumValues);
			ASURA_D3D_OK( xResult);
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<int>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, int* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetInt(xHandle, *pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			xResult = pxEffect->SetIntArray(xHandle, pxData, uNumValues);
			ASURA_D3D_OK( xResult);
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<bool>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, bool* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetBool(xHandle, *pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			if(sizeof(bool) != sizeof(BOOL))
			{
				// This is extremely unfortunate, and usually true, but hopefully we don't work with bool arrays too often
				// The BOOL datatype accepted by SetBoolArray is an int type - 4 bytes wide - while bool is usually only 1 byte wide
				// So we can't pass our array of bool directly
				// Our options are to allocate an array of BOOL and copy things across, or to set the bools individually
				// I suspect that setting them individually is faster and less hassle so that is what we do
				for(u_int uValue = 0; uValue < uNumValues; ++uValue)
				{
					D3DXHANDLE xElementHandle = pxEffect->GetParameterElement(xHandle, uValue);
					xResult = pxEffect->SetBool(xElementHandle, pxData[uValue]);
					ASURA_D3D_OK( xResult);
				}
			}
			else
			{
				// The strides match! Set it directly. (What system are we on, exactly?)
				xResult = pxEffect->SetBoolArray(xHandle, reinterpret_cast<BOOL*>(pxData), uNumValues);
				ASURA_D3D_OK( xResult);
			}
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<D3DXMATRIX>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, D3DXMATRIX* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetMatrix(xHandle, pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			xResult = pxEffect->SetMatrixArray(xHandle, pxData, uNumValues);
			ASURA_D3D_OK( xResult);
		}
	}
};


template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<D3DXVECTOR2>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, D3DXVECTOR2* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetValue(xHandle, pxData, sizeof( D3DXVECTOR2));
			ASURA_D3D_OK( xResult);
		}
		else
		{
			Asura_Assert( false, "Setting Arrays of D3DXVECTOR2 is not allowed");
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<D3DXVECTOR4>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, D3DXVECTOR4* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetVector(xHandle, pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			xResult = pxEffect->SetVectorArray(xHandle, pxData, uNumValues);
			ASURA_D3D_OK( xResult);
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<Direct3D_BaseTexture*>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, Direct3D_BaseTexture** pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetTexture(xHandle, *pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			Asura_Assert(false, "Effect parameter blocks do not support texture arrays");
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<Direct3D_Texture*>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, Direct3D_Texture** pxData, u_int uNumValues)
	{
		HRESULT xResult;
		if(uNumValues == 1)
		{
			xResult = pxEffect->SetTexture(xHandle, *pxData);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			Asura_Assert(false, "Effect parameter blocks do not support texture arrays");
		}
	}
};

template<>
struct Asura_Direct3D_EffectParameterBlock::ParameterSettingPolicy<Asura_RGBA>
{
	static void Apply(Direct3D_Effect* pxEffect, D3DXHANDLE xHandle, Asura_RGBA* pxData, u_int uNumValues)
	{
		HRESULT xResult;
		xResult = pxEffect->SetFloatArray(xHandle, (float*)pxData, 4 * uNumValues);
		ASURA_D3D_OK( xResult);
	}
};

#endif//__ASURA_DIRECT3D_EFFECTPARAMETERBLOCK_H__