//*******************************************************************
//
// File: Asura_Direct3D_IndexBuffer_Dynamic.h
//
// Created: 24 Oct 2006
//
// Description: 
//
//*******************************************************************

#ifndef _ASURA_DIRECT3D_INDEXBUFFER_DYNAMIC_H_
#define _ASURA_DIRECT3D_INDEXBUFFER_DYNAMIC_H_

// other libraries
#include "Rendering/Asura_Direct3D.h"

// internal dependencies

// forward declarations

//*******************************************************************
// Constants
//*******************************************************************


//*******************************************************************
// Class Definitions
//*******************************************************************

//-------------------------------------------------------------------
// Class:       Asura_Direct3D_IndexBuffer_Dynamic
//
// Description: Dynamic Index Buffer manager class.
//				You can ask it for a block of Index Buffer and it will give you a unique memory block to fill in.
//				When finished you tell it how much you have used
//				
//				Do not Lock1/Unlock1, Lock2/Unlock2 Render( Lock1 Data) Render( Lock2 Data) 
//				Only Lock1/Unlock1 Render() Lock2/Unlock2 Render()
//				If the Lock1 fills the VB then Lock2 will lock at the beginning of a discarded VB 
//
//				Once the VB has been Unlocked do not try to use any data pointers returned from the Lock
//
//				Do NOT Read from the memory returned by this class
//				Do NOT access memory outside of the range that was asked for
//-------------------------------------------------------------------
class Asura_Direct3D_IndexBuffer_Dynamic
{
public:
	Asura_Direct3D_IndexBuffer_Dynamic();
	~Asura_Direct3D_IndexBuffer_Dynamic();
	
	void Clear();
	
	bool ClearDirect3DData();
	bool CreateDirect3DData();
	
	bool SetFormat( const D3DFORMAT eIndexFormat);
	D3DFORMAT GetFormat( void) const { return( this->m_eIndexFormat);}
	
	bool SetSize( const u_int uSizeBytes);
	
	Direct3D_IndexBuffer* GetIndexBuffer() const;
	
	bool Lock( const u_int uSizeBytesRequired, u_int& uLockedIndexBufferOffsetBytes, void*& pLockedMemory);
	void Unlock( const u_int uSizeBytesUsed);
	
	bool IsLocked() const { return( NULL != this->m_pIndexBufferLockedStart);}
	
#if defined( ASURA_ASSERT)
	void* Assert_LockedMemoryStart() const { return( this->m_pIndexBufferLockedStart);}
	void* Assert_LockedMemoryEnd() const { return( this->m_pIndexBufferLockedEnd);}
#endif
	
private:
	void Init();
		
private:
	
	D3DFORMAT m_eIndexFormat;
	u_int m_uIndexBufferSizeBytes;
	Direct3D_IndexBuffer* m_pxD3DIndexBuffer;
	
	u_int m_uIndexBufferUsedSizeBytes;
	u_int m_uIndexBufferLockedSizeBytes;
	void* m_pIndexBufferLockedStart;
#if defined( ASURA_ASSERT)
	void* m_pIndexBufferLockedEnd;
#endif
};

//-------------------------------------------------------------------
// Class:       Asura_Direct3D_IndexBuffer_Dynamic_Type
//
// Description: This allows a typed interface to a non-types Index Buffer
//-------------------------------------------------------------------
template< class Tindex>
class Asura_Direct3D_IndexBuffer_Dynamic_Type
{
public:
	Asura_Direct3D_IndexBuffer_Dynamic_Type( Asura_Direct3D_IndexBuffer_Dynamic &xIndexBuffer);
	
	bool SetSize( const u_int uIndexCount);
	
	bool Lock( const u_int uIndexCountRequired, u_int& uLockedIndexBufferOffsetBytes, Tindex*& ptLockedMemory);
	void Unlock( const u_int uIndexCountUsed);
	
private:
	Asura_Direct3D_IndexBuffer_Dynamic& m_xIndexBuffer;
};

//*******************************************************************
// Inline Function Definitions
//*******************************************************************

inline Direct3D_IndexBuffer* Asura_Direct3D_IndexBuffer_Dynamic::GetIndexBuffer() const
{
	Asura_Assert( NULL != this->m_pxD3DIndexBuffer, "No index buffer has been allocated yet.");
	Asura_Assert( NULL == this->m_pIndexBufferLockedStart, "The Index Buffer is locked. Unlock() before calling this method.");
	return( this->m_pxD3DIndexBuffer);
}

template< class Tindex> 
Asura_Direct3D_IndexBuffer_Dynamic_Type< Tindex>::Asura_Direct3D_IndexBuffer_Dynamic_Type( Asura_Direct3D_IndexBuffer_Dynamic &xIndexBuffer) :
	m_xIndexBuffer( xIndexBuffer)
{
	Asura_Assert( 2 == sizeof( Tindex) || 4 == sizeof( Tindex), "Unsupported Data Type");
}

template< class Tindex> 
inline bool Asura_Direct3D_IndexBuffer_Dynamic_Type< Tindex>::SetSize( const u_int uIndexCount)
{
	return( this->m_xIndexBuffer.SetSize( uIndexCount * sizeof( Tindex)));
}

template< class Tindex> 
inline bool Asura_Direct3D_IndexBuffer_Dynamic_Type< Tindex>::Lock( const u_int uIndexCountRequired, u_int& uLockedIndexBufferOffsetBytes, Tindex*& ptLockedMemory)
{
	void* pLockedMemory = NULL;
	const bool bStatus = this->m_xIndexBuffer.Lock( uIndexCountRequired * sizeof( Tindex), uLockedIndexBufferOffsetBytes, pLockedMemory);
	ptLockedMemory = static_cast< Tindex*>( pLockedMemory);
	return( bStatus);
}

template< class Tindex> 
inline void Asura_Direct3D_IndexBuffer_Dynamic_Type< Tindex>::Unlock( const u_int uIndexCountUsed)
{
	this->m_xIndexBuffer.Unlock( uIndexCountUsed * sizeof( Tindex));
}

//*******************************************************************
//
// (c) 2006 Rebellion Derby
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************

#endif // _ASURA_DIRECT3D_VERTEXBUFFER_DYNAMIC_H_
