//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE        :  Asura_Direct3D_ShadowBuffer.h
//
//  CREATED     :  Richard J. Fine  (15.11.2004)
//
//  Copyright (C) 2004 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_DIRECT3D_SHADOWBUFFER_H__
#define __ASURA_DIRECT3D_SHADOWBUFFER_H__

#include "Rendering/Asura_Direct3D.h"

class Asura_Animation;
class Asura_Direct3D_ShadowBuffer;
class Asura_Direct3D_ShadowBufferSystem;

class Asura_Direct3D_ShadowBufferStrategy
{
public:
	Asura_Direct3D_ShadowBufferStrategy(Asura_Direct3D_ShadowBuffer* pxParentBuffer) : m_pxParent(pxParentBuffer) {}
	virtual ~Asura_Direct3D_ShadowBufferStrategy() {}

	virtual void CreateResources()			= 0;
	virtual void ReleaseResources()			= 0;
	virtual void Clean()					= 0;
	virtual void BeginCollectingShadows()	= 0;
	virtual void EndCollectingShadows()		= 0;
	virtual void BeginRenderingShadows()	= 0;
	virtual void EndRenderingShadows()		= 0;

	virtual float GetZScale() { return 1.0f; }

protected:
	Asura_Direct3D_ShadowBuffer* m_pxParent;
};

class Asura_Direct3D_ShadowBuffer
{
	friend class Asura_Direct3D_ShadowBufferSystem;

public:

	// (SB - 16/11/06) For some rendering techniques it's necessary to omit things while shadows are being collected
	static bool AreInstancesCollectingShadows() { return s_iNumInstancesCollectingShadows > 0; };

protected:

	// (SB - 16/11/06) For some rendering techniques it's necessary to omit things while shadows are being collected
	static int s_iNumInstancesCollectingShadows;

	Asura_Direct3D_ShadowBufferStrategy* m_pxStrategy;
	void SelectBufferStrategy();

	// Strategy-independent code

	enum BufferStatus
	{
		BUFFER_UNUSED,
		BUFFER_CLEAN,
		BUFFER_DIRTY
	};
	BufferStatus m_eBufferStatus;

	D3DXMATRIX m_xViewMatrix;
	D3DXMATRIX m_xProjectionMatrix;

	D3DXMATRIX m_xShadowViewportMatrix;
	D3DXMATRIX m_xShadowMatrix;

	D3DXMATRIX m_xOldViewMatrix;
	D3DXMATRIX m_xOldProjectionMatrix;

	Asura_Vector_3 m_xLightPosition;

	Asura_Bounding_Box m_xWorldSpaceShadowBox;
	bool m_bShadowTransformIsDirty;
	void RecalcShadowTransform();

	enum { AFFECTED_MODULES_LIST_SIZE = 16 };
	u_short m_ausModuleIDs[AFFECTED_MODULES_LIST_SIZE];
	u_int m_uModulesFound;

	const Asura_Animation* m_pxAssociatedAnim;

	void MakeClean();

	static void FindAffectedModulesCallback(const u_int uModuleIndex);
	static Asura_Direct3D_ShadowBuffer* s_pxCurrentShadowBuffer;

#ifdef ASURA_ASSERT
	bool m_bIsCurrentlyRenderingToBuffer;
#endif

	u_int m_uBufferWidth, m_uBufferHeight;

	float m_fBaseLevel;

	Direct3D_Viewport m_xViewport;
	Direct3D_Viewport m_xOldViewport;

	void RenderShadowsOnEnvironment(float fAlpha);

public:
	enum { DEFAULT_WIDTH = 256 };
	enum { DEFAULT_HEIGHT = 256 };

	// Shadow buffers may not be available on some hardware, so call this before trying to create any
	static bool IsSupported();

	Asura_Direct3D_ShadowBuffer(u_int uBufferWidth = DEFAULT_WIDTH, u_int uBufferHeight = DEFAULT_HEIGHT);
	~Asura_Direct3D_ShadowBuffer();

	void BeginCollectingShadows();
	void EndCollectingShadows();

	inline bool IsUnused() const { return m_eBufferStatus == BUFFER_UNUSED; }
	inline bool IsDirty() const { return m_eBufferStatus == BUFFER_DIRTY; }

	void SetLightPosition(const Asura_Vector_3& xLightPosition);
	void SetPerspectiveEnclosingSphere(const Asura_Vector_3& xSphereCenter, float fSphereRadius, float fDistance);
	inline const Asura_Vector_3& GetLightPosition() { return m_xLightPosition; }

	// Returns a matrix that transforms coordinates from worldspace into shadowbuffer space
	const D3DXMATRIX& GetShadowSpaceTransform();

	// Returns an AABB that encloses the shadowed volume
	const Asura_Bounding_Box &GetWorldSpaceShadowBox();

	// Uses the world space shadow box to find modules that may need shadow drawing on them
	void FindAffectedModules();

	inline int GetBufferWidth() { return m_uBufferWidth; }
	inline int GetBufferHeight() { return m_uBufferHeight; }

#ifdef _DEBUG
	void DrawShadowBox();
#endif

	inline const Asura_Animation* GetAssociatedAnim() const { return m_pxAssociatedAnim; }

};

class Asura_Direct3D_ShadowBufferSystem
{
public:
	static void Initialise();
	static void PreReset();
	static void PostReset();
	static void Shutdown();

	static Asura_Direct3D_ShadowBuffer* GetShadowBufferForAnim(const Asura_Animation* pxAnim, const Asura_Vector_3& xBoundingSphereCentre, float fSphereRadius);

	static void BeginNewFrame();
	static void FinishedCollectingShadows();

protected:
	enum { NUMBER_OF_SHADOW_BUFFERS = 16 };
	static Asura_Direct3D_ShadowBuffer* s_apxShadowBuffers[NUMBER_OF_SHADOW_BUFFERS];
};

#endif//__ASURA_DIRECT3D_SHADOWBUFFER_H__
