//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE        :  Asura_Direct3D_ShadowBuffer_DepthTexture.cpp
//
//  CREATED     :  Richard J. Fine  (16.11.2004)
//
//  Copyright (C) 2004 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////


#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Direct3D_ShadowBuffer_DepthTexture.h"

#include "Rendering/Asura_Direct3D_Caps.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Render.h"

const float cfZBias = -0.01f;
const float cfZSlopeScale = -0.005f;

Asura_Direct3D_ShadowBufferStrategy_DepthTexture::Asura_Direct3D_ShadowBufferStrategy_DepthTexture(Asura_Direct3D_ShadowBuffer* pxParent)
	: Asura_Direct3D_ShadowBufferStrategy(pxParent),
	  m_pxDepthTexture(0),
	  m_pxDepthSurface(0),
	  m_pxDummyColourRT(0),
	  m_pxDummyColourRTTexture(0)
{

}

Asura_Direct3D_ShadowBufferStrategy_DepthTexture::~Asura_Direct3D_ShadowBufferStrategy_DepthTexture()
{
	ReleaseResources();
}

bool Asura_Direct3D_ShadowBufferStrategy_DepthTexture::IsSupported()
{
	if(!Asura_Direct3D_Caps::HasDepthTextures()) return false;

	// check for depth bias too
	Direct3D_Caps caps;
	Asura_Direct3D::D3DDevice->GetDeviceCaps(&caps);  
	if(!(caps.RasterCaps & D3DPRASTERCAPS_DEPTHBIAS)) return false;
	if(!(caps.RasterCaps & D3DPRASTERCAPS_SLOPESCALEDEPTHBIAS)) return false;

	return true;

}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::CreateResources()
{
	ReleaseResources();

	D3DFORMAT depthTextureFormat = (Asura_Direct3D_Caps::GetDepthTextureBitDepth() == 16) ? D3DFMT_D16 : D3DFMT_D24S8;

	Asura_Direct3D::D3DDevice->CreateTexture(m_pxParent->GetBufferWidth(), 
											 m_pxParent->GetBufferHeight(), 
											 1, 
											 D3DUSAGE_DEPTHSTENCIL, 
											 depthTextureFormat, 
											 D3DPOOL_DEFAULT, 
											 &m_pxDepthTexture,
											 NULL);

	Asura_Assert(m_pxDepthTexture, "Could not create depth texture");

	m_pxDepthTexture->GetSurfaceLevel(0, &m_pxDepthSurface);
	Asura_Assert(m_pxDepthSurface, "Could not get depth surface");

	// Create a dummy colour render target to match
	Asura_Direct3D::D3DDevice->CreateTexture(m_pxParent->GetBufferWidth(),
											 m_pxParent->GetBufferHeight(),
											 1,
											 D3DUSAGE_RENDERTARGET,
											 D3DFMT_R5G6B5,
											 D3DPOOL_DEFAULT,
											 &m_pxDummyColourRTTexture,
											 NULL);
	m_pxDummyColourRTTexture->GetSurfaceLevel(0, &m_pxDummyColourRT);

	Asura_Assert(m_pxDummyColourRT, "Could not create dummy colour RT");
}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::ReleaseResources()
{
	if(m_pxDummyColourRT)
	{
		m_pxDummyColourRT->Release();
		m_pxDummyColourRT = NULL;
	}
	if(m_pxDummyColourRTTexture)
	{
		m_pxDummyColourRTTexture->Release();
		m_pxDummyColourRTTexture = NULL;
	}
	if(m_pxDepthSurface)
	{
		m_pxDepthSurface->Release();
		m_pxDepthSurface = NULL;
	}
	if(m_pxDepthTexture)
	{
		m_pxDepthTexture->Release();
		m_pxDepthTexture = NULL;
	}
}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::Clean()
{
	if(!m_pxDummyColourRT)CreateResources();

	Direct3D_Surface* pxOldColourBuffer = NULL;
	Direct3D_Surface* pxOldDepthBuffer	= NULL;

	Asura_Direct3D::D3DDevice->GetRenderTarget(0, &pxOldColourBuffer);
	Asura_Direct3D::D3DDevice->GetDepthStencilSurface(&pxOldDepthBuffer);

	Asura_Direct3D::D3DDevice->SetRenderTarget(0, m_pxDummyColourRT);
	Asura_Direct3D::D3DDevice->SetDepthStencilSurface(m_pxDepthSurface);

	Asura_Direct3D::D3DDevice->Clear(0, NULL, D3DCLEAR_ZBUFFER, 0xffffffff, 1.f, 0);

	Asura_Direct3D::D3DDevice->SetRenderTarget(0, pxOldColourBuffer);
	Asura_Direct3D::D3DDevice->SetDepthStencilSurface(pxOldDepthBuffer);

	if(pxOldColourBuffer)pxOldColourBuffer->Release();
	if(pxOldDepthBuffer)pxOldDepthBuffer->Release();
}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::BeginCollectingShadows()
{
	// Save out the current render target and depth stencil surface
	Asura_Direct3D::D3DDevice->GetRenderTarget(0, &m_pxOldRT);
	Asura_Direct3D::D3DDevice->GetDepthStencilSurface(&m_pxOldDS);
	Asura_Assert(m_pxOldRT && m_pxOldDS, "Could not get old RT/DS");

	// Set up our dummy render target and shadow buffer
	Asura_Direct3D::D3DDevice->SetRenderTarget(0, m_pxDummyColourRT);
	Asura_Direct3D::D3DDevice->SetDepthStencilSurface(m_pxDepthSurface);

	// Enable Z-buffering
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);

	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_JUSTSETZ);

	// Set up some zbias
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_DEPTHBIAS, *((DWORD*)&cfZBias));
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, *((DWORD*)&cfZSlopeScale));

}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::EndCollectingShadows()
{
	// Turn off zbias
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_DEPTHBIAS, 0);
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, 0);

	// Restore render target
	Asura_Direct3D::D3DDevice->SetRenderTarget(0, m_pxOldRT);
	Asura_Direct3D::D3DDevice->SetDepthStencilSurface(m_pxOldDS);

	if(m_pxOldRT)m_pxOldRT->Release(); m_pxOldRT = NULL;
	if(m_pxOldDS)m_pxOldDS->Release(); m_pxOldDS = NULL;
}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::BeginRenderingShadows()
{
	Asura_Direct3D::D3DDevice->GetTexture(2, &m_pxOldTexture);
	Asura_Direct3D::D3DDevice->SetTexture(2, m_pxDepthTexture);
	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_BORDER);
	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_BORDER);
	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_BORDERCOLOR, 0xffffffff);
	Asura_Direct3D::D3DDevice->SetTextureStageState(2, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_PROJECTED);

	Asura_Direct3D_Shaders::RequestPixelShader(ASURA_D3D_PS_SHADOWMAPPING_DEPTHTEXTURE);

	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_INVCOLOUR);

	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
}

void Asura_Direct3D_ShadowBufferStrategy_DepthTexture::EndRenderingShadows()
{
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);

	Asura_Direct3D_Shaders::RequestPixelShader(ASURA_D3D_PS_NONE);

	Asura_Direct3D::D3DDevice->SetTextureStageState(2, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);

	Asura_Direct3D::D3DDevice->SetTexture(2, m_pxOldTexture);
	if(m_pxOldTexture)
	{
		m_pxOldTexture->Release();
		m_pxOldTexture = NULL;
	}
}

float Asura_Direct3D_ShadowBufferStrategy_DepthTexture::GetZScale()
{
	return (float)((2 << (Asura_Direct3D_Caps::GetDepthTextureBitDepth())) - 1);
}
