//////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer.cpp
//
// Copyright (c) 2009 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined(ASURA_TARGET_SM3)

// Our header
#include "Rendering/Asura_Direct3D_SkinRenderer.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Resources/Asura_ResourceSet.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_CommandConsole.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#include "Shadows/Asura_Shadow.h"

// Asura PC
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Fog.h"
#include "Rendering/Asura_Direct3D_MaterialResponse.h"

// Asura Tools
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile_PC_Skin.h"
#include "MeshProcessing/Tools_HierarchyProcessor_MultiMaterial_NormalMapped.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor_PC.h"
#include "PC/Tools_PC_Export.h"
#endif

// Shaders
#include "Skin_VS.h"
#include "Skin_Shadow_VS.h"

/******************************************************************************/
/* Asura_Direct3D_SkinRenderer Static Class Data                               */
/******************************************************************************/

// Must match below vertex declaration, stream 0 only
const u_int Asura_Direct3D_SkinRenderer::uVERTEX_SIZE = 60;

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Remember to update Asura_Direct3D_SkinRenderer::uVERTEX_SIZE if this changes
// (and also remember to update vertex shader input type accordingly)
const Direct3D_VertexElement axVD[] =
{
	// Mesh vertices themselves in first stream
	{ 0,  0, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,     0 },
	{ 0, 12, D3DDECLTYPE_UBYTE4,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDINDICES, 0 },
	{ 0, 16, D3DDECLTYPE_SHORT4N,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT,  0 },
	{ 0, 24, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,      0 },
	{ 0, 32, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL,     0 },
	{ 0, 40, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,       0 },
	{ 0, 48, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     0 },
	{ 0, 56, D3DDECLTYPE_FLOAT1,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     1 },
	D3DDECL_END()
};

ASURA_STATIC_ASSERT( Asura_Direct3D_SkinRenderer::uVERTEX_SIZE == 60 );

const char szSOLID_SHADER[] =
	"float4 main( uniform float4 xColour : register(c0) ) : COLOR { return xColour; }";

const char szSHADOW_SHADER[] =
	"float4 main( const float4 fDepth : TEXCOORD ) : COLOR { return fDepth; }";

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Direct3D_VertexShader*		g_pxMainVS = NULL;
static Direct3D_VertexShader*		g_pxShadowVS = NULL;
static Direct3D_PixelShader*		g_pxSolidPS = NULL;
static Direct3D_PixelShader*		g_pxShadowPS = NULL;
static Direct3D_VertexDeclaration*	g_pxMainVSDeclaration = NULL;

static Asura_ResourceSet< Asura_Direct3D_Skin > g_xSkins( "Direct3DSkin" );

static u_int g_uNumRendered = 0;

// This can be removed when the old renderers are deleted
#if defined ASURA_USE_NEW_SKIN_RENDERERS
	DVSVAR bool dvs_bEnabled = true;
#else
	DVSVAR bool dvs_bEnabled = false;
#endif

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// IsStripVisible()
//
// Helper function for checking whether strips are visible in this render pass
//
////////////////////////////////////////////////////////////////////////////////

static bool IsStripVisible( const Asura_Animation& xAnim, const int iElement,
	const Asura_MaterialResponse* const pxMaterial, const bool bDepthPass, const bool bShadow )
{

	// Check skin element isn't hidden
#if defined ASURA_SUPPORT_SKIN_ELEMENTS
	if( !xAnim.GetSkinElementVisibility( iElement ) ) return false;
#endif

	if( pxMaterial )
	{
		// Check if material is hidden
		if( bShadow && pxMaterial->GetDontCastShadow() ) return false;
		if( !bShadow && pxMaterial->GetDontRender() ) return false;

		// Check we're in the right transparency pass (not applicable for shadows)
		if( !bShadow && (pxMaterial->IsTransparent() != Asura_SortedRender::IsTransparentPass()) )
		{
			return false;
		}

		// Check we're in the right depth pass
		if( bDepthPass && !pxMaterial->GetZPrePass() ) return false;
	}

	return true;
}


////////////////////////////////////////////////////////////////////////////////
//
// SetMatrices()
//
// Uploads the world, world-view-proj, camera pos etc. to GPU registers
//
////////////////////////////////////////////////////////////////////////////////

static void SetMatrices( Asura_AMatrix4x4_InParam xWorldMat, const Asura_Shadow* const pxShadow )
{
	D3DXMATRIX xWorldViewProj;
	xWorldMat.StoreColsUnaligned( &xWorldViewProj._11 );

	Asura_FixedArray< D3DXVECTOR4, 9 > axConstants;

	if( pxShadow )
	{
		// Shadow frustum first
		const float fNear = pxShadow->GetFrustum().GetNearPlane();
		const float fFar = pxShadow->GetFrustum().GetFarPlane();
		axConstants[0].x = fNear;
		axConstants[0].y = 1.0f / (fFar - fNear);

		// Light position
		const Asura_Vector_3& xLightPos = pxShadow->GetLightPosition();
		D3DXMATRIX xShadowMat;
		D3DXMatrixTranslation( &xShadowMat, -xLightPos.x, -xLightPos.y, -xLightPos.z );
		xWorldViewProj *= xShadowMat;

		// Light Orientation
		Asura_Matrix_3x3 xLightOri;
		pxShadow->GetLightOrientation( xLightOri );
		Asura_Mem::Set( &xShadowMat, 0, sizeof( D3DXMATRIX ) );

		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xShadowMat.m[u][v] = xLightOri.Element[u][v];

				if( v == 1 )
				{
					// Account for y-axis being inverted between D3D and Asura
					xShadowMat.m[u][v] = -xShadowMat.m[u][v];
				}
			}
		}

		xShadowMat._44 = 1.0f;
		xWorldViewProj *= xShadowMat;

		// Projection
		const float fVertFOV =
			Asura_Maths::DegToRad( pxShadow->GetFrustum().GetVerticalFOVInDegrees() );

		D3DXMatrixPerspectiveFovLH( &xShadowMat, fVertFOV, 1.0f, fNear, fFar );
		xWorldViewProj *= xShadowMat;

		Asura_Mem::Copy( &axConstants[4], &xWorldViewProj, sizeof( D3DXMATRIX ) );
	}
	else
	{
		// World matrix first
		Asura_Mem::Copy( &axConstants[0], &xWorldViewProj, sizeof( D3DXMATRIX ) );

		// Then world-view-proj
		xWorldViewProj *= Asura_Direct3D::GetViewProjectionMatrix();
		Asura_Mem::Copy( &axConstants[4], &xWorldViewProj, sizeof( D3DXMATRIX ) );

		// Camera position in world space
		axConstants[8].x = Asura_Camera::GetPosition()->x;
		axConstants[8].y = Asura_Camera::GetPosition()->y;
		axConstants[8].z = Asura_Camera::GetPosition()->z;
		axConstants[8].w = 1.0f;
	}

	Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( 2, axConstants[0], 9 );
}


////////////////////////////////////////////////////////////////////////////////
//
// SetFog()
//
// Sets up fogging for transparent strips
//
////////////////////////////////////////////////////////////////////////////////

static void SetFog()
{
	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;

	const BOOL bFogEnabled = Asura_Fog::IsFogEnabled();

	const float afFogParams[4] =
		{ Asura_Fog::GetFogNearPlane(), Asura_Fog::GetOneOverFogPlaneSeparation(), 0.0f, 0.0f };

	pxDevice->SetPixelShaderConstantB( 2, &bFogEnabled, 1 );
	pxDevice->SetPixelShaderConstantF( 6, afFogParams, 1 );
	pxDevice->SetTexture( 1, Asura_Direct3D_Fog::GetTexture() );

	// Actual fog colour is set per strip, since additive strips need to always fog to black
}


////////////////////////////////////////////////////////////////////////////////
//
// CopyBones()
//
// Copies bone matrices for a skin into the vertex shader constants
//
////////////////////////////////////////////////////////////////////////////////

static void CopyBones( const Asura_AMatrix4x4* const pxBoneMats, u_int uNumBones,
	const Asura_Animation& xAnim, const Asura_Direct3D_Skin::BoneMap* const pxBoneMap = NULL )
{
	SCOPEDPIXEVENT(( 0xFF884499, L"Set bone matrices" ));

	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;
	const u_int uMaxBones = ASURA_MAX_NO_OF_BONES_PER_MATERIAL;

	float afBoneMatricesBuffer[ uMaxBones * 12 ];
	Asura_Mem::Set( afBoneMatricesBuffer, 0, sizeof( afBoneMatricesBuffer ) );

	Asura_Assertf( uNumBones <= uMaxBones, ("%s has too many bones (%u > %u).",
		(pxBoneMap ? "Bone map" : "Skin"), uNumBones, uMaxBones) );

	uNumBones = Asura_Maths::Min( uMaxBones, uNumBones );

	for( u_int u = 0; u < uNumBones; ++u )
	{
		const u_int uBone = (pxBoneMap ? pxBoneMap->m_puEntries[u] : u);

		pxBoneMats[ uBone ].GetRow0().StoreUnaligned( &afBoneMatricesBuffer[ u*12+0 ] );
		pxBoneMats[ uBone ].GetRow1().StoreUnaligned( &afBoneMatricesBuffer[ u*12+4 ] );
		pxBoneMats[ uBone ].GetRow2().StoreUnaligned( &afBoneMatricesBuffer[ u*12+8 ] );

		// We pack per-bone damage values into the matrix to save space, as we can
		// reconstruct the blatted row in the vertex shader (orthogonal rows). Invert
		// here since 1 means fully damaged, but in texturing terms that's material 0
		afBoneMatricesBuffer[ u*12+10 ] = 1.0f - xAnim.GetBoneDamageMapValue( uBone );
	}

	pxDevice->SetVertexShaderConstantF( 11, afBoneMatricesBuffer, uNumBones * 3 );
}


////////////////////////////////////////////////////////////////////////////////
//
// SetSH()
//
// Helper function for setting SH-related shader constants
//
////////////////////////////////////////////////////////////////////////////////

static void SetSH( const Asura_MaterialResponse* const pxMaterial )
{
	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;

	Asura_FixedArray< D3DXVECTOR4, 6 > axConstants;
	Asura_SphericalHarmonics_Order2_RGB xSH;

	if( pxMaterial )
	{
		// Set SH lighting by considering static and dynamic components separately

		xSH = Asura_GI_Solution::s_xStaticSHForCurrentRender;

		Asura_RGBA xColour = pxMaterial->GetAmbientColour();

		const float fPrelightingColourise = pxMaterial->GetPrelightingColourise();
		xSH.Colourise( xColour.m_fR, xColour.m_fG, xColour.m_fB, fPrelightingColourise );

		const float fAmbientContribution = pxMaterial->GetAmbientContribution();
		xSH.Scale( fAmbientContribution );

		if( pxMaterial->IsTransparent() )
		{	
			xSH += Asura_GI_Solution::s_xDynamicSHForCurrentRender;
		}
	}
	else
	{
		// send through basic SH
		xSH = Asura_GI_Solution::s_xSHForCurrentRender;
	}

	for( u_int u = 0; u < 3; ++u )
	{
		// Need to swizzle because the constant term is stored first in the struct
		
		axConstants[u].x = xSH[ u*4 + 1 ];
		axConstants[u].y = xSH[ u*4 + 2 ];
		axConstants[u].z = xSH[ u*4 + 3 ];
		axConstants[u].w = xSH[ u*4 + 0 ];
		
	}

	// If rendering transparent strips, need to know the brightest light direction & colour
	if( pxMaterial && pxMaterial->IsTransparent() )
	{
		Asura_Vector_3 xLightDirection;
		xSH.GetBrightestDirection( xLightDirection );

		if( xLightDirection.ApproximatelyEquals( Asura_Maths::ZeroVector3 ) )
		{
			// If there are no lights, we'll get a zero direction. Not much we can do here,
			// might as well consider the light as coming from above
			xLightDirection.y = -1.0f;
		}

		axConstants[3].x = xLightDirection.x;
		axConstants[3].y = xLightDirection.y;
		axConstants[3].z = xLightDirection.z;
		axConstants[3].w = 0.0f;

		// Evaluate the colour in that direction
		Asura_RGBA xLightColour;
		xSH.EvaluateInDirection( xLightDirection, xLightColour );

		// Use the same method to calculate light brightness as the photon shaders. Should
		// probably use standard luminance calculation instead
		const float fLightBrightness =
			xLightColour.m_fR / 3.0f + xLightColour.m_fG / 3.0f + xLightColour.m_fB / 3.0f;

		axConstants[4].x = xLightColour.m_fR;
		axConstants[4].y = xLightColour.m_fG;
		axConstants[4].z = xLightColour.m_fB;
		axConstants[4].w = fLightBrightness;

		// Fog colour is set per strip, since additive materials need to always fog to black
		if( pxMaterial->GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_ADDITIVE )
		{
			axConstants[5].x = axConstants[5].y = axConstants[5].z = axConstants[5].w = 0.0f;
		}
		else
		{
			axConstants[5].x = Asura_Fog::GetFogColour().m_fR;
			axConstants[5].y = Asura_Fog::GetFogColour().m_fG;
			axConstants[5].z = Asura_Fog::GetFogColour().m_fB;
			axConstants[5].w = Asura_Fog::GetFogColour().m_fA;
		}
	}

	pxDevice->SetPixelShaderConstantF( 128, axConstants[0], axConstants.GetSize() );
}


#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// RenderWireframe()
//
// Renders a strip's wireframe. Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

static void RenderWireframe( const Asura_Direct3D_Skin& xSkin, const u_int uNumIndices,
	const u_int uStartIndex, const u_int uARGB )
{
	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;

	Asura_Render::CurrentStates.RequestWireFrameMode( true );
	pxDevice->SetPixelShader( g_pxSolidPS );

	const Asura_RGBA xColour( uARGB );
	pxDevice->SetPixelShaderConstantF( 0, &xColour.m_fR, 1 );

	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );

	// Apply a little depth bias so the lines don't break up across the skin
	float fDepthBias = -0.0001f;
	DWORD uValue = *reinterpret_cast< const DWORD* >( &fDepthBias );
	pxDevice->SetRenderState( D3DRS_DEPTHBIAS, uValue );

	if( xSkin.IsUsingTriLists() )
	{
		pxDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, xSkin.GetNumVertices(),
			uStartIndex, uNumIndices / 3 );
	}
	else
	{
		pxDevice->DrawIndexedPrimitive( D3DPT_TRIANGLESTRIP, 0, 0, xSkin.GetNumVertices(),
			uStartIndex, uNumIndices - 2 );
	}

	// Reset depth bias
	fDepthBias = 0.0f;
	uValue = *reinterpret_cast< const DWORD* >( &fDepthBias );
	pxDevice->SetRenderState( D3DRS_DEPTHBIAS, uValue );
}

#endif // defined ASURA_TOOLS


////////////////////////////////////////////////////////////////////////////////
//
// RenderStrip()
//
// Helper function for rendering a single strip from a skin
//
////////////////////////////////////////////////////////////////////////////////

static void RenderStrip( const Asura_Direct3D_Skin& xSkin, const Asura_AMatrix4x4* const pxBoneMats,
	const u_int uStrip, const Asura_Animation& xAnim, const Asura_Hierarchy_Skin& xHierarchySkin,
	const u_int uStartIndex, const bool bDepthPass, const bool bShadow )
{
	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;

	const Asura_Direct3D_SkinStrip& xStrip = xSkin.GetStrip( uStrip );

	const Asura_MaterialResponse* const pxMaterial1 =
		Asura_MaterialResponse_System::GetMaterialByHashID( xStrip.m_auMaterials[0] );

	// Skip strip if appropriate
	if( !IsStripVisible( xAnim, xStrip.m_iElementID, pxMaterial1, bDepthPass, bShadow ) ) return;

	SCOPEDPIXEVENT(( 0, L"Strip %u", uStrip ));

	// Number of bone weights per vertex
	{
		const int aiLoops[4] = { xStrip.m_uNumWeights, 0, 1, 0 };
		pxDevice->SetVertexShaderConstantI( 0, aiLoops, 1 );
	}

	// Set bone matrices
	if( xSkin.GetNumBoneMaps() > 0 )
	{
		const Asura_Direct3D_Skin::BoneMap& xBoneMap = xSkin.GetBoneMap( xStrip.m_uBoneMap );
		CopyBones( pxBoneMats, xBoneMap.m_uSize, xAnim, &xBoneMap );
	}

	if( !bDepthPass && !bShadow )
	{
		const Asura_MaterialResponse* const pxMaterial2 =
			Asura_MaterialResponse_System::GetMaterialByHashID( xStrip.m_auMaterials[1] );

		const Asura_Hash_ID uCombinedHash = Asura_MaterialResponse_System::CombineShaderHashes(
			pxMaterial1, pxMaterial2, true );

		// it seems that this will set pixel shader
		Asura_Direct3D_MaterialResponse::SetMaterialConstants( uCombinedHash, pxMaterial1,
			pxMaterial2, &xAnim );

		SetSH( pxMaterial1 );
		Asura_Direct3D_MaterialResponse::SetBlendStates( pxMaterial1 );

		Asura_Render::CurrentStates.RequestAlphaTest( true );

		#if defined ASURA_TOOLS
			Asura_Render::CurrentStates.RequestWireFrameMode( Asura_Render::GetWireframe() );
		#endif

		if( Asura_MaterialResponse_System::GetMaterialProjectContextCallback() )
		{
			const u_int uContext1 = pxMaterial1
				? pxMaterial1->GetProjectContext()
				: ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID;

			const u_int uContext2 = pxMaterial2
				? pxMaterial2->GetProjectContext()
				: ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID;

			Asura_MaterialResponse_System::GetMaterialProjectContextCallback()( uContext1,
				uContext2 );
		}
	}
	else
	{
		Asura_Render::CurrentStates.RequestAlphaTest( false );
	}

	// TODO: Looks like we'll be moving to trilists only in the near future. Can remove this then

	if( xSkin.IsUsingTriLists() )
	{
		pxDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, xSkin.GetNumVertices(),
			uStartIndex, xStrip.m_uNumIndices/3 );
	}
	else
	{
		pxDevice->DrawIndexedPrimitive( D3DPT_TRIANGLESTRIP, 0, 0, xSkin.GetNumVertices(),
			uStartIndex, xStrip.m_uNumIndices-2 );
	}

	#if defined ASURA_TOOLS

		if( !bDepthPass && xAnim.GetOutlineColour() && xHierarchySkin.IsSkinnedModel() )
		{
			// Render again using wireframe effect, but not if hierarchy (see comment in
			// Asura_Hierarchy_System::Platform_RenderHierarchy())
			RenderWireframe( xSkin, xStrip.m_uNumIndices, uStartIndex, xAnim.GetOutlineColour() );
		}

	#endif
}


/******************************************************************************/
/* Asura_Direct3D_SkinRenderer Static Class Functions                          */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::Initialise()
//
// Initialises the PC skin renderer. Called at app startup
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_SkinRenderer::Initialise()
{
	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;
	if( !pxDevice ) return;

	HRESULT xRes;

	// Create vertex shaders

	const DWORD* puFunction = reinterpret_cast< const DWORD* >( g_auSkin_VS );
	pxDevice->CreateVertexShader( puFunction, &g_pxMainVS );

	puFunction = reinterpret_cast< const DWORD* >( g_auSkin_Shadow_VS );
	pxDevice->CreateVertexShader( puFunction, &g_pxShadowVS );

	Asura_Assert( g_pxMainVS && g_pxShadowVS, "Failed to create skin vertex shader." );

	// Create vertex declaration

	xRes = pxDevice->CreateVertexDeclaration( axVD, &g_pxMainVSDeclaration );
	Asura_Assert( SUCCEEDED( xRes ), "Failed to create skin vertex declaration." );

	// Create pixel shader for solid-colour rendering. Used for wireframe, and also depth-only
	// passes (since you can't set a NULL pixel shader with VS 3.0)

	ID3DXBuffer* pxBuffer = NULL;
	xRes = D3DXCompileShader( szSOLID_SHADER, sizeof( szSOLID_SHADER ), NULL, NULL, "main",
		"ps_3_0", 0, &pxBuffer, NULL, NULL );

	Asura_Assert( SUCCEEDED( xRes ), "Failed to compile skin solid pixel shader." );

	if( SUCCEEDED( xRes ) )
	{
		puFunction = static_cast< const DWORD* >( pxBuffer->GetBufferPointer() );
		xRes = pxDevice->CreatePixelShader( puFunction, &g_pxSolidPS );
		Asura_Assert( SUCCEEDED( xRes ), "Failed to create skin solid pixel shader." );
		SAFE_RELEASE( pxBuffer );
	}

	// Set up toggle DVS var. This can be removed when the old renderers are deleted

	#if defined ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_CommandConsole::AddVar( L"Render.UseNewSkins", &dvs_bEnabled,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::Reset()
//
// Resets the PC skin renderer. Typically called between levels
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_SkinRenderer::Reset()
{
	// Nothing to do here. Skins will be destroyed when the project clears the resource bank
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::Shutdown()
//
// Shuts down the PC skin renderer.
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_SkinRenderer::Shutdown()
{
	SAFE_RELEASE( g_pxMainVS );
	SAFE_RELEASE( g_pxShadowVS );
	SAFE_RELEASE( g_pxSolidPS );
	SAFE_RELEASE( g_pxShadowPS );
	SAFE_RELEASE( g_pxMainVSDeclaration );
	g_xSkins.Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::IsEnabled()
//
// Queries whether the new skin renderer or old ones should be used. This can be
// removed when the old renderers are deleted
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Direct3D_SkinRenderer::IsEnabled()
{
	return dvs_bEnabled;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::ResetCount()
//
// Resets the rendered skin count
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_SkinRenderer::ResetCount()
{
	g_uNumRendered = 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::CreateSkin()
//
// Allocates a new PC render skin and adds it to the current resource bank
// Passes back a bool stating whether a skin with that name already existed
//
////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Skin* Asura_Direct3D_SkinRenderer::CreateSkin( const Asura_Hash_ID uHash, bool& bAlreadyExisted )
{
	bAlreadyExisted = false;

	if( g_xSkins.GetResource( uHash ) )
	{
		bAlreadyExisted = true;
		return NULL;
	}

	Asura_Direct3D_Skin* const pxSkin = new Asura_Direct3D_Skin;
	if( !pxSkin ) return NULL;

	if( !g_xSkins.AddResource( uHash, *pxSkin ) )
	{
		delete pxSkin;
		return NULL;
	}

	return pxSkin;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::GetSkin()
//
// Returns the skin for the given Asura_Hash_ID
//
////////////////////////////////////////////////////////////////////////////////

const Asura_Direct3D_Skin* Asura_Direct3D_SkinRenderer::GetSkin( const Asura_Hash_ID uHash )
{
	return g_xSkins.GetResource( uHash );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::GetVertexDeclaration()
//
// Retrieves the skin renderer vertex declaration for external use
//
////////////////////////////////////////////////////////////////////////////////

Direct3D_VertexDeclaration* Asura_Direct3D_SkinRenderer::GetVertexDeclaration()
{
	return g_pxMainVSDeclaration;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::RenderSkin()
//
// Renders a skin on PC
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_SkinRenderer::RenderSkin( const Asura_Animation& xAnimation,
	const Asura_Hierarchy_Skin& xHierarchySkin, Asura_AMatrix4x4_InParam xWorldMat,
	const Asura_AMatrix4x4* const pxBoneMats, const Asura_Shadow* const pxShadow /*= NULL*/ )
{
	const Asura_Direct3D_Skin* pxSkin = g_xSkins.GetResource( xHierarchySkin.GetHashID() );

	#if defined ASURA_TOOLS
		if( !pxSkin )
		{
			pxSkin = Tools_InitialiseSkinFromRawData( &xHierarchySkin );
		}
	#endif

	Asura_Assertf( pxSkin, ("Failed to find mesh for skin '%s'.", xHierarchySkin.GetName()) );
	if( !pxSkin ) return;

	SCOPEDPIXEVENT(( 0, L"Skin %u - '%S'", ++g_uNumRendered, xHierarchySkin.GetName() ));

	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;
	if( !pxDevice ) return;

	// Set up
	{
		// Blending and z-write states set up per strip, others here
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestStencilEnabled( false );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );

		pxDevice->SetVertexDeclaration( g_pxMainVSDeclaration );

		pxDevice->SetVertexShader( pxShadow ? g_pxShadowVS : g_pxMainVS );

		if( pxShadow )
		{
			pxDevice->SetPixelShader( g_pxShadowPS );
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		}

		pxDevice->SetStreamSource( 0, pxSkin->GetVB(), 0, Asura_Direct3D_SkinRenderer::uVERTEX_SIZE );
		pxDevice->SetIndices( pxSkin->GetIB() );

		SetMatrices( xWorldMat, pxShadow );

		// Bone matrices
		if( pxSkin->GetNumBoneMaps() == 0 )
		{
			CopyBones( pxBoneMats, xHierarchySkin.NumberOfBones, xAnimation );
		}

		if( Asura_SortedRender::IsTransparentPass() )
		{
			SetFog();
		}
	}

	const u_int uNumStrips = pxSkin->GetNumStrips();
	const bool bShadow = (pxShadow != NULL );
	const bool bDoDepthPass = Asura_SortedRender::IsTransparentPass() && !bShadow;

	for( u_int u = (bDoDepthPass ? 0 : 1); u < 2; ++u )
	{
		const bool bIsDepthPass = (u == 0);

		if( bDoDepthPass )
		{
			// Don't bother with the PIX event if we're only doing one pass
			D3DPERFCALL( BeginEvent( 0, (bIsDepthPass ? L"Depth Pass" : L"Colour Pass") ) );

			if( bIsDepthPass )
			{
				pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE, 0 );
				pxDevice->SetPixelShader( g_pxSolidPS );
				Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
			}
		}

		u_int uStartIndex = 0;

		for( u_int v = 0; v < uNumStrips; ++v )
		{
			RenderStrip( *pxSkin, pxBoneMats, v, xAnimation, xHierarchySkin, uStartIndex,
				bIsDepthPass, bShadow );

			uStartIndex += pxSkin->GetStrip(v).m_uNumIndices;
		}

		if( bDoDepthPass )
		{
			D3DPERFCALL( EndEvent() );

			pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALL  );
		}
	}

	// Cleanup
	{
		if( pxShadow )
		{
			Asura_Render::AddToTriPassesRendered( pxSkin->GetNumTriangles() );
		}
		else
		{
			Asura_Render::AddToTrianglesRendered( pxSkin->GetNumTriangles() );
		}

		// Reset wireframe render state, in case we were doing selected objects
		Asura_Render::CurrentStates.RequestWireFrameMode( Asura_Render::GetWireframe() );

		// Many renderers don't request this, so turn it back on
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

		const float fDepthBias = 0.0f;
		const DWORD uDepthBias = *reinterpret_cast< const DWORD* >( &fDepthBias );
		pxDevice->SetRenderState( D3DRS_DEPTHBIAS, uDepthBias );

		// "Blend into albedo" materials might have modified these
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE,  D3DCOLORWRITEENABLE_ALL );
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL );
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL );
	}
}


#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_SkinRenderer::Tools_InitialiseSkinFromRawData()
//
// In Tools builds, constructs D3D skin render data from Asura_Hierarchy_Skin.
// Similar to what happens at export time
//
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Skin* Asura_Direct3D_SkinRenderer::Tools_InitialiseSkinFromRawData(
	const Asura_Hierarchy_Skin* const pxConstSkin )
{
	//////////////////////////////////////////////////////////////////////////
	// Get a non-const copy of the skin to operate on
	// Need to get the top-most skin for this , since the LOD's aren't accessible via GetSkinByHash
	int iLODLevel = 0;
	const Asura_Hierarchy_Skin* pxConstParentSkin = pxConstSkin;
	while(pxConstParentSkin->m_pxNextHigherSkin)
	{
		pxConstParentSkin = pxConstParentSkin->m_pxNextHigherSkin;
		iLODLevel++;
	}
	Asura_Hierarchy_Skin* pxSkin = Asura_Animation_System::GetSkinByHash(pxConstParentSkin->m_uNameHashID);
	if(!pxSkin) 
		return NULL;
	//skip back down to the correct LOD
	for(int i=0;i<iLODLevel;i++)
	{
		if(!pxSkin->m_pxNextLowerSkin) return NULL;
		pxSkin = pxSkin->m_pxNextLowerSkin;
	}
	//////////////////////////////////////////////////////////////////////////

	Tools_SmoothSkin_Processor xProcessor;
	xProcessor.SetIncludeNormalMapData( true ); // This is unnecessary for a few models, but it's going to get pretty rare.

	Tools_SmoothSkin_PC_Processor xPCProcessor(xProcessor);

	if( pxSkin->IsSkinnedModel() )
	{
		if( !xProcessor.Initialise( pxSkin ) ) return NULL;
		xProcessor.CreateBoneRemapTables(ASURA_MAX_NO_OF_BONES_PER_MATERIAL);
		if( !xPCProcessor.InitialiseV2( false ) ) return false;
	}
	else
	{
		if( !Tools_HierarchyProcessor_MultiMaterial_NormalMapped::Initialise( pxSkin ) )	return false;
		if( !xPCProcessor.InitialiseV2FromHierarchyProcessor() )							return false;
	}

	// use dummy stream to get the memory size required.
	Asura_Chunk_Dummy_Stream xDummyStream;

	xPCProcessor.OutputToChunkV2( xDummyStream );

	u_int uSize = xDummyStream.GetPosition();
	void* pBuffer = Asura_Allocate(uSize);
	if (!pBuffer) return NULL;

	// write chunk data to memory buffer
	Asura_Chunk_Memory_Stream xMemoryStream( uSize, pBuffer );
	if( ! xPCProcessor.OutputToChunkV2( xMemoryStream ) )
	{
		Asura_Deallocate(pBuffer);
		return NULL;
	}

	if( !pxSkin->IsSkinnedModel() )
	{
		Tools_HierarchyProcessor_MultiMaterial_NormalMapped::Shutdown();
	}

	//////////////////////////////////////////////////////////////////////////
	// restart stream
	xMemoryStream.SetPosition(0);
	//
	Asura_Chunk_ResourceFile_PC_Skin xChunk;
	xMemoryStream.Read( sizeof(xChunk), &xChunk );
	//const bool bWarnIfMeshExists = false;// if creating skin data for the loaded mesh in tools we want to skip this warning.
	xMemoryStream.SkipString();	// skip resource name (mesh name is after).

	bool bAlreadyExisted = false;
	Asura_Direct3D_Skin* pxResult = xChunk.CreateAndRead( pxConstSkin->GetHashID(), xMemoryStream, bAlreadyExisted );
	if( bAlreadyExisted )
	{
#if !defined ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA
		Asura_Assertf( !bAlreadyExisted, ("Skin being loaded more than once.") );
#endif
		return NULL;
	}

	Asura_Deallocate(pBuffer);

	return pxResult;
}

#endif // defined ASURA_TOOLS


/******************************************************************************/
/* Asura_Direct3D_Skin Class Functions                                         */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Skin::Asura_Direct3D_Skin()
//
// Constructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Skin::Asura_Direct3D_Skin()
: m_pxStrips( NULL )
, m_uNumStrips(0)
, m_uNumTriangles(0)
, m_uNumVertices(0)
, m_pxIB( NULL )
, m_pxVB( NULL )
, m_pxBoneMaps( NULL )
, m_uNumBoneMaps( 0 )
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Skin::~Asura_Direct3D_Skin()
//
// Destructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Skin::~Asura_Direct3D_Skin()
{
	delete[] m_pxStrips;
	delete[] m_pxBoneMaps;

	SAFE_RELEASE( m_pxIB );
	SAFE_RELEASE( m_pxVB );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Skin::SetNumBoneMaps()
//
// Reallocates the bone maps for a D3D skin
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Skin::SetNumBoneMaps( const u_int uNumBoneMaps )
{
	delete[] m_pxBoneMaps;

	m_uNumBoneMaps = uNumBoneMaps;
	if (uNumBoneMaps)
	{
		m_pxBoneMaps = new BoneMap[uNumBoneMaps];
		Asura_Assert(m_uNumBoneMaps, "Unable to allocate smooth skin Bonemaps");
	}
	else
	{
		m_pxBoneMaps = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Skin::IsAnyMaterialTransparent()
//
// Scans all strips to see if any of them are using a transparent material
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Direct3D_Skin::IsAnyMaterialTransparent() const
{
	for( u_int u = 0; u < m_uNumStrips; ++u )
	{
		const Asura_MaterialResponse* const pxMaterial =
			Asura_MaterialResponse_System::GetMaterialByHashID( m_pxStrips[u].m_auMaterials[0] );

		if( pxMaterial && pxMaterial->IsTransparent() ) return true;
	}

	return false;
}

#endif // #if defined(ASURA_TARGET_SM3)

