//
// Asura_PlatformRenderMesh.cpp
//
// Created by Kevin Lea - 05/10/2001 16:39:06
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_TOOLS

// This file's header
#include "Rendering/Asura_PlatformRenderMesh.h"

// Asura headers
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Light/Asura_Light.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Bicubic_Patch.h"
#include "Rendering/Asura_Direct3D_Caps.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_PatchSubdivision.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Surface.h"
#include "Rendering/Asura_PlatformRenderEnvironment.h"

#ifdef ASURA_SUPPORT_WEATHER_SYSTEM
#include "SpecialFX/Asura_Weather_System.h"
#endif

#ifdef ASURA_TOOLS
#include "Misc/Tools_Render_Options.h"
#endif

class Asura_Binary_Tree_Node_Vector
{
	public:
 
		VertexFormat V;
		int Index;
		class Asura_Binary_Tree_Node_Vector *Lower;
		class Asura_Binary_Tree_Node_Vector *Higher;

		Asura_Binary_Tree_Node_Vector(VertexFormat* v, int index)
		{
			Lower = 0;
			Higher = 0;	

			V = *v;
			Index = index;
		}
		~Asura_Binary_Tree_Node_Vector()
		{
		}
		
	    inline bool operator == ( const Asura_Binary_Tree_Node_Vector& v) const
		{
		    if (V == v.V) return true;
			else return false;
		}

	    inline bool operator > ( const Asura_Binary_Tree_Node_Vector& v) const
		{
		    if (V > v.V) return true;
			else return false;
		}
	    inline bool operator < ( const Asura_Binary_Tree_Node_Vector& v) const
		{
		    if (V < v.V) return true;
			else return false;
		}


			
	    inline bool operator == ( const VertexFormat *n) const
		{
		    if (V == *n) return true;
			else return false;
		}

	    inline bool operator > ( const VertexFormat *n) const
		{
		    if (V > *n) return true;
			else return false;
		}
	    inline bool operator < ( const VertexFormat *n) const
		{
		    if (V < *n) return true;
			else return false;
		}
};

class Asura_Binary_Tree_Vector
{
	public:
		
		Asura_Binary_Tree_Node_Vector *Root;
		int NumberOfNodes;

		Asura_Binary_Tree_Vector()
		{
			NumberOfNodes = 1;
			VertexFormat zero;
			zero.X = 0;
			zero.Y = 0;
			zero.Z = 0;
			Root = new Asura_Binary_Tree_Node_Vector(&zero,0);
		}
		~Asura_Binary_Tree_Vector()
		{
			FreeAll();
		}

		// KJL 08/03/00 17:08:16 - returns true if the node existed already
		bool FindNodeIndexByName(VertexFormat *searchName, bool addToTreeIfRequired, int& index)
		{
			Asura_Binary_Tree_Node_Vector *nodePtr = Root;
			
			while(!(*nodePtr==searchName))
			{
				while(*nodePtr<searchName)
				{
					if (nodePtr->Higher)
					{
						nodePtr = nodePtr->Higher;
						continue;
					}
					else if (addToTreeIfRequired)
					{
						nodePtr->Higher = new Asura_Binary_Tree_Node_Vector(searchName,NumberOfNodes++);
						
						index = nodePtr->Higher->Index;
					}
					return false;
				}
				while(*nodePtr>searchName)
				{
					if (nodePtr->Lower)
					{
						nodePtr = nodePtr->Lower;
						continue;
					}
					else if (addToTreeIfRequired)
					{
						nodePtr->Lower = new Asura_Binary_Tree_Node_Vector(searchName,NumberOfNodes++);
					
						index =  nodePtr->Lower->Index;
					}
					return false;
				}

			}
			index = nodePtr->Index;
			return true;
		}

		void FreeAll(void)
		{
			FreeAllRecursive(Root);
		}

		void FreeAllRecursive(Asura_Binary_Tree_Node_Vector *nodePtr)
		{
			if (nodePtr->Lower)
			{
				FreeAllRecursive(nodePtr->Lower);
			}
			if (nodePtr->Higher)
			{
				FreeAllRecursive(nodePtr->Higher);
			}

			//if (nodePtr!=Root)
			{
				delete nodePtr;
			}
		}

		void Traverse(Asura_Binary_Tree_Node_Vector *nodePtr, void (*function)(VertexFormat& v, int index))
		{
			if (nodePtr->Lower)
			{
				Traverse(nodePtr->Lower,function);
			}
			if (nodePtr->Higher)
			{
				Traverse(nodePtr->Higher,function);
			}

			if (nodePtr!=Root)
			{
				function(nodePtr->V,nodePtr->Index);
			}
		}


};

static const int MAX_SMOOTHED_VERTICES = 65536;

static unsigned int s_uPrelightingFrameStamp = 0;//counter to stagger recalculation of prelighting
static unsigned int s_uPrelightingDoneThisFrame = 0;
static const unsigned int s_uMaxPrelightingPerFrame = 400;
static bool s_bRecalculatePrelightingOnSubdividedMeshes = false;

Asura_Direct3D_Mesh* RenderMeshStorage[ASURA_MAX_NO_OF_MESHES];
int SetupLightsForMesh(int offset, int step);

void Asura_Direct3D_Mesh::Initialise(class Asura_Mesh* mesh)
{
	Asura_Assert(!IsValid,"RenderMesh already initialised");

	VertexFormat* vertexData = new VertexFormat[mesh->NumberOfTriangles*3];

	int freeVertex = 0;
	if(IndexList) delete [] IndexList;
	IndexList = new unsigned short[mesh->NumberOfTriangles*3];

	int* piMaterialList = new int[mesh->NumberOfTriangles];
	int* polysInBatch = new int[mesh->NumberOfTriangles];

	if (!mesh->AreNormalsValid()) mesh->CreateNormals();

	int freeTexture = 0;
	{
		for (int pID=0; pID<mesh->NumberOfPolygons; pID++)
		{
			#ifdef ASURA_TOOLS
			if(Tools_Render_Options::IsHidden(mesh,pID)) continue;
			#endif

			
			int iMatIndex;
			if (mesh->TextureDataValid)
			{
				iMatIndex = mesh->TextureDataPtr[pID].m_iMaterialID;
			}
			else
			{		
				iMatIndex = ASURA_MATERIAL_NONE;
			}
			
			int v = freeTexture;
			bool newTexture = true;

			while(v--)
			{
				if (piMaterialList[v] == iMatIndex)
				{
					newTexture = false;
					break;
				}
			}

			if (newTexture)
			{
				piMaterialList[freeTexture] = iMatIndex;
				polysInBatch[freeTexture] = 1;
				if (mesh->PolygonDataPtr[pID].IsQuad()) polysInBatch[freeTexture]++;
				freeTexture++;
			}
			else
			{
				polysInBatch[v]++;
				if (mesh->PolygonDataPtr[pID].IsQuad()) polysInBatch[v]++;
			}
		}
	}

	PacketList = new Asura_Polygon_Packet[freeTexture];
	NumberOfPackets = freeTexture;
	TotalNumberOfIndices = 0;
	int i = 0;

	Asura_Binary_Tree_Vector tree;

	for (int t=0; t<freeTexture; t++)
	{
		PacketList[t].IndexStart = i;
		PacketList[t].NumberOfIndices = 0;
		PacketList[t].m_iMaterialID = piMaterialList[t];

		for (int pID=0; pID<mesh->NumberOfPolygons; pID++)
		{
			int iMatIndex;
			if (mesh->TextureDataValid)
			{
				iMatIndex = mesh->TextureDataPtr[pID].m_iMaterialID;
			}
			else
			{
				iMatIndex = ASURA_MATERIAL_NONE;
			}

			if (iMatIndex != piMaterialList[t]) continue;

			#ifdef ASURA_TOOLS
			if(Tools_Render_Options::IsHidden(mesh,pID)) continue;
			#endif

		
			for (int offset = 0; offset<3; offset+=2)
			{
				if (offset && !mesh->PolygonDataPtr[pID].IsQuad())
				{
					continue;
				}
				for (int e=0; e<3; e++)
				{
					int a = (e+offset)%4;
					
				

					VertexFormat vertex;
					
					int vID = mesh->PolygonDataPtr[pID].VertexID[a];

					Asura_Vector_3 p = mesh->VertexDataPtr[vID].Position;

					vertex.X = p.x;
					vertex.Y = p.y;
					vertex.Z = p.z;
					
					Asura_Vector_3 n = mesh->NormalsDataPtr[pID].VertexNormal[a];

					vertex.m_NX = n.x;
					vertex.m_NY = n.y;
					vertex.m_NZ = n.z;

					if (mesh->TextureDataValid)
					{
						vertex.U0 = mesh->TextureDataPtr[pID].U[0][a];
						vertex.V0 = mesh->TextureDataPtr[pID].V[0][a];
				
						vertex.U1 = vertex.U0*32.0f;
						vertex.V1 = vertex.V0*32.0f;
						if (mesh->PrelightingDataValid)
						{
							vertex.Colour = mesh->PrelightingDataPtr[pID].Colour[a];
						}
						else
						{
							vertex.Colour = 0xffffffff;
			
						}
					}
					else
					{
						vertex.U0 = 0;
						vertex.V0 = 0;
					
						vertex.Colour = 0xffffffff;
					}
				
			

					int v;
					bool newVertex = true;

#if 01
					if(tree.FindNodeIndexByName(&vertex,true,v))
					{
						newVertex = false;
					}
					v-=1;
#endif
					if (newVertex)
					{
						IndexList[i+e] = (unsigned short)freeVertex;
						vertexData[freeVertex++] = vertex;
					}
					else
					{
						IndexList[i+e] = (unsigned short)v;
					}
				}
				i+=3;
				PacketList[t].NumberOfIndices+=3;
				TotalNumberOfIndices +=3;
			}
		}
	}

	
	NumberOfUniqueVertices = freeVertex;

//	MakeVertexBuffer(mesh->Index,NumberOfUniqueVertices,vertexData,IndexList,TotalNumberOfIndices);

	if(VertexData) delete [] VertexData;
	VertexData = new VertexFormat[NumberOfUniqueVertices];
	for (int index=0; index<NumberOfUniqueVertices; index++)
	{
		VertexData[index] = vertexData[index];
	}

	MakeVertexBuffer(vertexData,IndexList);

	delete [] vertexData;
	delete [] piMaterialList;
	delete [] polysInBatch;

	IsValid = true;
	m_bVerticesOffsetByModulePosition = false;
}



void Asura_Direct3D_Mesh::MakeVertexBuffer(VertexFormat* vertices, unsigned short* indexList)
{
	Asura_Assert(NumberOfUniqueVertices<MAX_SMOOTHED_VERTICES,"Too many vertices in an object!");
	Asura_Assert(TotalNumberOfIndices<MAX_SMOOTHED_VERTICES*3,"Too many polys in an object!");
	if(!NumberOfUniqueVertices) return;

	Asura_Direct3D::D3DDevice->CreateVertexBuffer
	(
		NumberOfUniqueVertices*sizeof(D3DGenericVertex),
		0, // We will need to read these verts when copying to the final renderer
		D3DFVF_GENERICVERTEX,
		D3DPOOL_MANAGED,
		&VertexBuffer,
		0
	);

	Asura_Direct3D::D3DDevice->CreateIndexBuffer
	(
		TotalNumberOfIndices*sizeof(unsigned short),
		D3DUSAGE_WRITEONLY,
		D3DFMT_INDEX16,
		D3DPOOL_MANAGED,
		&IndexBuffer,
		0
	);

	D3DGenericVertex* vertexdata;
	if (SUCCEEDED(VertexBuffer->Lock(0,sizeof(D3DGenericVertex)*NumberOfUniqueVertices,(void**)&vertexdata,0)))
	{
		for (int i=0; i<NumberOfUniqueVertices; i++)
		{
			vertexdata[i].X = vertices[i].X;
			vertexdata[i].Y = vertices[i].Y;
			vertexdata[i].Z = vertices[i].Z;
			vertexdata[i].Normal.x = vertices[i].m_NX;
			vertexdata[i].Normal.y = vertices[i].m_NY;
			vertexdata[i].Normal.z = vertices[i].m_NZ;


			vertexdata[i].U = vertices[i].U0;
			vertexdata[i].V = vertices[i].V0;
			vertexdata[i].U2 = vertices[i].U1;
			vertexdata[i].V2 = vertices[i].V1;

			//	vertices[i].ColourComponents.R /= 2;
			//	vertices[i].ColourComponents.G /= 2;
			//	vertices[i].ColourComponents.B /= 2;


			vertexdata[i].Diffuse = vertices[i].Colour;
		}
		VertexBuffer->Unlock();
	} else {
		Asura_Assert(0, "Failed to lock VB");
	}

	unsigned short* indexdata;
	if (SUCCEEDED(IndexBuffer->Lock(0,sizeof(unsigned short)*TotalNumberOfIndices,(void**)&indexdata,0)))
	{
		for(int i=0; i<TotalNumberOfIndices; i++)
		{
			indexdata[i] = indexList[i];
		}
		IndexBuffer->Unlock();
	} else {
		Asura_Assert(0, "Failed to lock IB");
	}

	m_iVertexPrelightingProgress = 0;	
	
}

#ifdef ASURA_TOOLS
void Asura_Direct3D_Mesh::CalculatePrelighting()
{
	Asura_Assert(m_bPrelightingBeingCalculated,"");
	Asura_Assert(m_pxMesh && m_piPolygonIndeces,"");
	Asura_Assert(m_pxMesh->AreNormalsValid(),"");
	D3DGenericVertex* vertexdata;
	//lock the vertex buffer , and calculate the prelighting for as many vertices as we can
	if (SUCCEEDED(VertexBuffer->Lock(0,sizeof(D3DGenericVertex)*NumberOfUniqueVertices,(void**)&vertexdata,0)))
	{
		for (;m_iVertexPrelightingProgress<NumberOfUniqueVertices && s_uPrelightingDoneThisFrame<s_uMaxPrelightingPerFrame; m_iVertexPrelightingProgress++,s_uPrelightingDoneThisFrame++)
		{
			Asura_Vector_3 xPos,xNormal;


			xPos.x = vertexdata[m_iVertexPrelightingProgress].X;
			xPos.y = vertexdata[m_iVertexPrelightingProgress].Y;
			xPos.z = vertexdata[m_iVertexPrelightingProgress].Z;
			xNormal.x = vertexdata[m_iVertexPrelightingProgress].Normal.x;
			xNormal.y = vertexdata[m_iVertexPrelightingProgress].Normal.y;
			xNormal.z = vertexdata[m_iVertexPrelightingProgress].Normal.z;

			Asura_Colour_Union xColour;

			int iPolyIndex = m_piPolygonIndeces[m_iVertexPrelightingProgress];
			int iPolygonFlags = 0;
			if(m_pxMesh->PolygonFlagDataValid)
			{
				iPolygonFlags = m_pxMesh->PolygonFlagDataPtr[iPolyIndex].Flags;
			}

			//for the LOS test need to project the vertex onto the original polygon , so it doesn't get blocked
			//by the unsubdivided mesh.
			//const Asura_Vector_3& xFaceNormal = m_pxMesh->NormalsDataPtr[iPolyIndex].FaceNormal;
			//const Asura_Vector_3& xOriginalVertex = m_xPosition+m_pxMesh->VertexDataPtr[m_pxMesh->PolygonDataPtr[iPolyIndex].VertexID[0]].Position;

			//Asura_Vector_3 xOffset = xOriginalVertex - xPos;
			//float fDistance = xOffset*xFaceNormal + 0.05f;
			//Asura_Vector_3 xPointOnPoly = xPos + xFaceNormal*fDistance;

			const Asura_Mesh_Polygon* pxPoly = &m_pxMesh->PolygonDataPtr[iPolyIndex];
			const Asura_Vector_3& xFaceNormal = m_pxMesh->NormalsDataPtr[iPolyIndex].FaceNormal;
			float fDistance = -Asura_Maths::LargeFloat;
			for(int i=0;i<pxPoly->NumberOfVertices();i+=3)
			{
				const Asura_Vector_3& xOriginalVertex = m_xPosition+m_pxMesh->VertexDataPtr[m_pxMesh->PolygonDataPtr[iPolyIndex].VertexID[i]].Position;

				Asura_Vector_3 xOffset = xOriginalVertex - xPos;
				float fTempDistance = xOffset*xFaceNormal + 0.002f;
				if(fTempDistance>fDistance) fDistance = fTempDistance;

			}
			Asura_Vector_3 xPointOnPoly = xPos + xFaceNormal*fDistance;

			Asura_Lighting_System::CalculatePreLightValue(xPos,xPointOnPoly,xNormal,xColour.Colour,iPolygonFlags);

	///		xColour.ColourComponents.R /= 2;
	//		xColour.ColourComponents.G /= 2;
	//		xColour.ColourComponents.B /= 2;


			vertexdata[m_iVertexPrelightingProgress].Diffuse = 0xff000000|xColour.Colour;
		}
		VertexBuffer->Unlock();
	} else {
		Asura_Assert(0, "Failed to lock VB");
	}


}
#endif //ASURA_TOOLS


void Asura_Direct3D_Mesh::Shutdown(void)
{
	if(IsValid)
	{
		if (IndexBuffer) IndexBuffer->Release();
		if (VertexBuffer) VertexBuffer->Release();
		IsValid = false;

		if (VertexData) delete [] VertexData;
		if (PacketList) delete [] PacketList;
		if (IndexList) delete [] IndexList;

		if(m_piPolygonIndeces) delete [] m_piPolygonIndeces;

		VertexData = 0;
		PacketList = 0;
		IndexList = 0;
		IndexBuffer = 0;
		VertexBuffer = 0; 
		m_piPolygonIndeces = 0;
		NumberOfPackets = 0;
	}
}

void Asura_Direct3D_Mesh::ReleaseBuffers()
{
	Shutdown();
}

void Asura_Direct3D_Mesh::UpdateLighting(Asura_Mesh* mesh, Asura_Vector_3* position, Asura_Matrix_3x3* orientation)
{
return;

}

// Ugh.
#ifdef ASURA_TOOLS
#include "Misc/Tools_Render_Options.h"
#endif

void Asura_Direct3D_Mesh::RenderHighlight( const Asura_Vector_3* pxPos, const Asura_Matrix_3x3* pxOri, u_int uColour, bool bWireframe, bool bOffset )
{
	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestTextureOpMode( ASURA_TEXTUREOP_SINGLE_PASS );
	
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
	Asura_Render::CurrentStates.RequestTextureIndex( 0, ASURA_TEXTURE_NONE );
    Asura_Render::CurrentStates.RequestWireFrameMode( bWireframe );

    Asura_Direct3D_Shaders::RequestVertexShader( bOffset ? ASURA_D3D_VS_MESHHIGHLIGHT_WITHOFFSET : ASURA_D3D_VS_MESHHIGHLIGHT );

#ifdef ASURA_TOOLS
    if( Tools_Render_Options::NoBackFaceCull )
    {
        Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
    }
    else
#endif
    {
        Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);
    }

	Asura_Vector_3 xPos(0,0,0);
	if(m_bVerticesOffsetByModulePosition)
	{
		//module offset has already been applied... ignore the position passed. RAR 10/04/03
		pxPos = &xPos;
	}

    // Setup object->screen matrix and also object space cam pos and offset.
    {
    	D3DXMATRIX xObjectToWorldMat;
    	D3DXMATRIX xWorldToViewMat;
    	D3DXMATRIX xViewToScreenMat;
        D3DXMATRIX xTempMat;

        // Setup obj->world matrix.
		Asura_Direct3D::SetupD3DMatrix(xObjectToWorldMat, pxOri ? *pxOri : Asura_Maths::IdentityMatrix3x3, *pxPos, true);

        // Fetch world->view and view->screen matrices from DirectX.
	    Asura_Direct3D::D3DDevice->GetTransform( D3DTS_VIEW, &xWorldToViewMat );
	    Asura_Direct3D::D3DDevice->GetTransform( D3DTS_PROJECTION, &xViewToScreenMat );

        // Concatenate.
        D3DXMatrixMultiply( &xTempMat, &xObjectToWorldMat, &xWorldToViewMat );
        D3DXMatrixMultiplyTranspose( &xTempMat, &xTempMat, &xViewToScreenMat );

        // Copy to vertex shader's constant registers.
        Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( 0, (float*)&xTempMat, 4 );

        if( bOffset )
        {
            // Setup offset parameters in constant registers.
            {
                const float MIN_OFFSET   = 0.05f;

#if ASURA_PC_OFFSET_MESH_HIGHLIGHT_ACCORDING_TO_CAM_DIST
                const float MAX_OFFSET   = 5.0f;
                const float MIN_CAM_DIST = 10.0f;
                const float MAX_CAM_DIST = 200.0f;

                // Just to make table tidier below...
                const float OFFSET_RANGE       = MAX_OFFSET - MIN_OFFSET;
                const float CAM_DIST_RANGE     = MAX_CAM_DIST - MIN_CAM_DIST;
                const float O_O_CAM_DIST_RANGE = 1.0f / CAM_DIST_RANGE;

                const u_int FIRST_CONST_REGISTER_TO_SET = 6;
                const u_int NUM_CONST_REGISTERS_TO_SET  = 6;

                const float afData[NUM_CONST_REGISTERS_TO_SET][4] =
                {
                    { MIN_OFFSET,         MIN_OFFSET,         MIN_OFFSET,         0.0f }, // 6
                    { OFFSET_RANGE,       OFFSET_RANGE,       OFFSET_RANGE,       0.0f }, // 7
                    { MIN_CAM_DIST,       MIN_CAM_DIST,       MIN_CAM_DIST,       0.0f }, // 8
                    { CAM_DIST_RANGE,     CAM_DIST_RANGE,     CAM_DIST_RANGE,     0.0f }, // 9
                    { O_O_CAM_DIST_RANGE, O_O_CAM_DIST_RANGE, O_O_CAM_DIST_RANGE, 0.0f }, // 10
                    { 0.0f,               0.0f,               0.0f,               0.0f }, // 11
                };

                Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( FIRST_CONST_REGISTER_TO_SET, (float*)afData, NUM_CONST_REGISTERS_TO_SET );
#else
                const float afOffset[4] = { MIN_OFFSET, MIN_OFFSET, MIN_OFFSET, 0.0f };

                Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( 6, afOffset, 1 );
#endif
            }

            // Setup object space camera position in constant register.
            {
                const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();

                float afTransformedCamPos[4] =
                {
                    pxCamPos->x, pxCamPos->y, pxCamPos->z, 1.0f
                };
                
                // *ahem*
                Asura_Vector_3& xTransformedCamPos = reinterpret_cast<Asura_Vector_3&>( afTransformedCamPos[0] );

                // Transform camera position from world space to object space.
                if( pxOri )
                {
                    Asura_Matrix_3x3 xInverseRot( *pxOri );
                    xInverseRot.Transpose();
                    xTransformedCamPos = (xTransformedCamPos - *pxPos) * xInverseRot;
                }
                else
                {
                    xTransformedCamPos -= *pxPos;
                }

                Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( 5, afTransformedCamPos, 1 );
            }
        }
    }

    // Setup colour.
    {
        const float fOneOver255 = 1.0f / 255.0f;

        float afColours[4];
        for( u_int u = 4; u--; )
        {
            const u_int uShift = (u << 3);
			const u_int uIndex = (u==3) ? 3 : 2-u;//Rearranged so that uColour is read as ARGB. RAR 15/05/03
            afColours[uIndex] = static_cast<float>((uColour & (0xff << uShift)) >> uShift) * fOneOver255;
        }

        Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( 4, afColours, 1 );
    }

	Asura_Direct3D::D3DDevice->SetStreamSource( 0, this->VertexBuffer,0, sizeof(D3DGenericVertex) );
	Asura_Direct3D::D3DDevice->SetIndices( this->IndexBuffer );

    // Send each packet.
    Asura_Assert( this->NumberOfPackets >= 0, "" );
	for( int iPacket = this->NumberOfPackets; iPacket--; )
	{
        const Asura_Polygon_Packet& xPacket = this->PacketList[iPacket];
		Asura_Direct3D::D3DDevice->DrawIndexedPrimitive
		(
			D3DPT_TRIANGLELIST,
			0,
			0,
            this->NumberOfUniqueVertices,
            xPacket.IndexStart,
            xPacket.NumberOfIndices/3
		);
    }

//    // Reset obj->world matrix.
//	D3DXMatrixTranslation( &xObjectToWorldMat, 0.0f, 0.0f, 0.0f );
//	Asura_Direct3D::D3DDevice->SetTransform( D3DTS_WORLD, &xObjectToWorldMat );

    // Some renderers are less thorough with their renderstate setup so reset
    // a couple of bits.
	Asura_Render::CurrentStates.RequestWireFrameMode( Asura_Render::GetWireframe() );
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW);
}

void Asura_Direct3D_Mesh::Render( const Asura_Vector_3* position, const Asura_Matrix_3x3* orientation )
{
#if defined( ASURA_TARGET_SM3)
	HRESULT xResult;
	
	// DLH - 14 February 2007 12:33:53 - Store and Clear all the RenderTargets after the first one
	const u_int uSurfaceCountMax = 8;
	Direct3D_Surface* apxSurface[ uSurfaceCountMax] = {NULL}; 
	u_int uSurfaceCount = Asura_Direct3D_Caps::GetMaxNumberOfRenderTargets();
	
	Asura_Assertf( uSurfaceCount < uSurfaceCountMax, ("%d Possible Render Targets useable. Only %d coped with.", uSurfaceCount, uSurfaceCountMax));
	// DLH - 14 February 2007 12:31:18 - Stops array over runs
	uSurfaceCount = Asura_Maths::Min( uSurfaceCount, uSurfaceCountMax);
	
	for ( u_int uSurfaceIndex = 1; uSurfaceIndex < uSurfaceCount; uSurfaceIndex += 1)
	{
		xResult = Asura_Direct3D::D3DDevice->GetRenderTarget( uSurfaceIndex, &(apxSurface[ uSurfaceIndex]));
//		ASURA_D3D_OK( xResult);
		xResult = Asura_Direct3D::D3DDevice->SetRenderTarget( uSurfaceIndex, NULL);
		ASURA_D3D_OK( xResult);
	}
#endif

	Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);
	Asura_Render::CurrentStates.RequestAlphaTest(false);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);

	
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);


#if 0
	Asura_Matrix_3x3 m;
	Asura_Camera_Desc *cameraPtr = Asura_Camera_System::GetCameraDescription();
	m = cameraPtr->Orientation;			
	m.Transpose();
#endif
	HRESULT h;
	
	D3DXMATRIX matWorld;
	D3DXMatrixTranslation( &matWorld, position->x,position->y,position->z );
	if(orientation)
	{
		matWorld._11 = orientation->e11;
		matWorld._12 = orientation->e21;
		matWorld._13 = orientation->e31;
		matWorld._21 = orientation->e12;
		matWorld._22 = orientation->e22;
		matWorld._23 = orientation->e32;
		matWorld._31 = orientation->e13;
		matWorld._32 = orientation->e23;
		matWorld._33 = orientation->e33;
	}
	Asura_Direct3D::D3DDevice->SetTransform( D3DTS_WORLD,      &matWorld );

	h=Asura_Direct3D::D3DDevice->SetStreamSource(0,VertexBuffer,0,sizeof(D3DGenericVertex));
	h=Asura_Direct3D::D3DDevice->SetIndices(IndexBuffer);

	for(int i=0; i<NumberOfPackets; i++)
	{
		// KJL - 22-May-01 15:45:09 - warning textureindex may be -1
		//if (!Asura_Texture_Management::TextureIsTranslucent(PacketList[i].TextureIndex))
		{
			Asura_Render::CurrentStates.RequestTextureIndex(0,Asura_Material_System::GetTextureID(PacketList[i].m_iMaterialID));
		
			// KJL - Monday January 22 2001 15:56:22
			// 
			// Note: Optimisations can be made here wrt start and end vertex to transform 
			h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,PacketList[i].IndexStart,PacketList[i].NumberOfIndices/3);
		}
//		else
		{
		#if 0
			int s = PacketList[i].IndexStart;
			for(int p=0; p<PacketList[i].NumberOfIndices; p+=3)
			{
				Asura_Vector_3 pos;
				
				pos.Zero();
				for(int v=0; v<3; v++)
				{
					pos.x += VertexData[IndexList[s+p+v]].X;
					pos.y += VertexData[IndexList[s+p+v]].Y;
					pos.z += VertexData[IndexList[s+p+v]].Z;
				}
				
				pos = (pos-cameraPtr->Position)*m;
				
				Asura_Translucent_Polygon_System::Add(s+p,PacketList[i].TextureIndex,vbIndex,pos.z);
			}
		#endif
		}
	}

	D3DXMatrixTranslation( &matWorld, 0,0,0 );
	Asura_Direct3D::D3DDevice->SetTransform( D3DTS_WORLD,      &matWorld );
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW);
#if defined( ASURA_TARGET_SM3)
	// DLH - 14 February 2007 12:33:53 - Restore all the RenderTargets after the first one
	for ( u_int uSurfaceIndex = 1; uSurfaceIndex < uSurfaceCount; uSurfaceIndex += 1)
	{
		Direct3D_Surface* pxSurface = apxSurface[ uSurfaceIndex];
		
		xResult = Asura_Direct3D::D3DDevice->SetRenderTarget( uSurfaceIndex, pxSurface);
		ASURA_D3D_OK( xResult);
		if( NULL != pxSurface)
		{
			pxSurface->Release();
		}
	}

#endif
}

#include "Light/Asura_Photon.h"

void Asura_Direct3D_Mesh::SimpleRender(void)
{
	const bool bWireframe = Asura_Render::GetWireframe();

	Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);

	//Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);

	Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
	Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
	Asura_Direct3D::D3DDevice->SetSamplerState( 1, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
	Asura_Direct3D::D3DDevice->SetSamplerState( 1, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );

	HRESULT h;
		
	h=Asura_Direct3D::D3DDevice->SetStreamSource(0,VertexBuffer,0,sizeof(D3DGenericVertex));
	h=Asura_Direct3D::D3DDevice->SetIndices(IndexBuffer);
		
	
	/// Kevin Lea - 29/01/2002 11:45:24 - first render ambient light only
    if( bWireframe )
    {
    	Asura_Render::CurrentStates.RequestTextureOpMode( ASURA_TEXTUREOP_WHITE );
    }
    else
    {
	    Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_EMISSIVEMATERIALSOURCE, D3DMCS_COLOR1);
    	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
    }
	Asura_Render::CurrentStates.RequestTextureIndex(0,ASURA_TEXTURE_NONE);

	bool bMeshUsesTransparency = false;

    if( !bWireframe )
    {
	    for(int i=0; i<NumberOfPackets; i++)
	    {
		    if (Asura_Material_System::IsMaterialTransparent(PacketList[i].m_iMaterialID))
		    {
			    bMeshUsesTransparency = true;
			    break;
		    }
	    }
    }

	if (!bMeshUsesTransparency)
	{
		Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
		h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,0,TotalNumberOfIndices/3);
		Asura_Render::AddToTrianglesRendered(TotalNumberOfIndices/3);
	}
	else
	{
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE );
//		float fBias = -2.0f;
//		Asura_Direct3D::D3DDevice->SetTextureStageState(0, D3DTSS_MIPMAPLODBIAS, *((LPDWORD) (&fBias)));
	
		for(int i=0; i<NumberOfPackets; i++)
		{
			//Asura_Render::CurrentStates.RequestTextureIndex(0,PacketList[i].TextureIndex);
				
			if (Asura_Material_System::IsMaterialTransparent(PacketList[i].m_iMaterialID))
			{
				Asura_Render::CurrentStates.RequestAlphaTest(true);
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_JUSTALPHA);
				Asura_Render::CurrentStates.RequestTextureIndex(0,Asura_Material_System::GetTextureID(PacketList[i].m_iMaterialID));
				h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,PacketList[i].IndexStart,PacketList[i].NumberOfIndices/3);
			}
			else
			{
				Asura_Render::CurrentStates.RequestAlphaTest(false);
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
				Asura_Render::CurrentStates.RequestTextureIndex(0,ASURA_TEXTURE_NONE);
				h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,PacketList[i].IndexStart,PacketList[i].NumberOfIndices/3);
			}
		
		}
		Asura_Render::AddToTrianglesRendered(TotalNumberOfIndices/3);
	}
	
    // That's your lot in wireframe.
    if( bWireframe ) { return; }

	Asura_Render::CurrentStates.RequestAlphaTest(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_EQUALS);

	static int AttenuationMap = 0;
	static int DetailMap = 0;
	static int RainSplashTexture = 0;

	/// Kevin Lea - 29/01/2002 11:53:15 - render light passes
	int noOfPhotons = Asura_Photon_System::NumberOfLocalPhotons;
	
	if (noOfPhotons)
	{
	
		if (!AttenuationMap)
		{
			AttenuationMap = Asura_Texture_Management::AddTextureWithResetGroup("\\graphicnovel\\expattenuation.dds", ASURA_TEXTURE_LOADED_PERMANENT );
			Asura_Texture_Management::LoadTextures();
		}
		Asura_Render::CurrentStates.RequestTextureIndex(0,AttenuationMap);
		Asura_Render::CurrentStates.RequestTextureIndex(1,AttenuationMap);
 
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_NONE );

		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_NONE );

		Asura_Render::CurrentStates.RequestDynamicLights(true);
	
		Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
	
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_EMISSIVEMATERIALSOURCE, D3DMCS_MATERIAL);

		Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_LIGHTPASS);

		Asura_Direct3D::D3DDevice->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEPOSITION | 0 );
		Asura_Direct3D::D3DDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);// | D3DTTFF_PROJECTED );//D3DTTFF_COUNT2 );

		Asura_Direct3D::D3DDevice->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEPOSITION | 1 );
		Asura_Direct3D::D3DDevice->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);// | D3DTTFF_PROJECTED );//D3DTTFF_COUNT2 );

		for (int n=0; n<noOfPhotons; n++)
		{
			Asura_Local_Photon* photon = &Asura_Photon_System::LocalPhotons[n];
			const Asura_Vector_3& xPosition = photon->GetPosition();

			{
				Direct3D_Light light;
				ZeroMemory( &light, sizeof(Direct3D_Light) );
				light.Type = D3DLIGHT_POINT;
				light.Diffuse.r  = photon->GetRedComponent();
				light.Diffuse.g  = photon->GetGreenComponent();
				light.Diffuse.b  = photon->GetBlueComponent();

				
				light.Position.x = xPosition.x;
				light.Position.y = xPosition.y;
				light.Position.z = xPosition.z;
			
				light.Range      = photon->GetRange();
			
				light.Attenuation0 = 1.0f;//0.01f;
				light.Attenuation1 = 0.0f;
				light.Attenuation2 = 0; 

				Asura_Direct3D::D3DDevice->SetLight( 0, &light );
				Asura_Direct3D::D3DDevice->LightEnable( 0, TRUE);
			}
	
			Direct3D_Material mat;

			// Set the RGBA for diffuse reflection. 
			mat.Diffuse.r = 0.5f; 
			mat.Diffuse.g = 0.5f; 
			mat.Diffuse.b = 0.5f; 
			mat.Diffuse.a = 0.5f; 

			// Set the RGBA for ambient reflection. 
			 mat.Ambient.r = 0.0f; 
			 mat.Ambient.g = 0.0f; 
			 mat.Ambient.b = 0.0f; 
			 mat.Ambient.a = 0.0f; 

			// Set the color and sharpness of specular highlights. 
			mat.Specular.r = 0.0f; 
			mat.Specular.g = 0.0f; 
			mat.Specular.b = 0.0f; 
			mat.Specular.a = 0.0f; 
			mat.Power = 0.0f; 

			mat.Emissive.r = 0.5f*photon->GetRedComponent();
			mat.Emissive.g = 0.5f*photon->GetGreenComponent();
			mat.Emissive.b = 0.5f*photon->GetBlueComponent();
			mat.Emissive.a = 0.5f;

			Asura_Direct3D::D3DDevice->SetMaterial(&mat);

			D3DXMATRIX matView;
			Asura_Direct3D::D3DDevice->GetTransform( D3DTS_VIEW, &matView );
			D3DXMatrixInverse( &matView,  NULL, &matView );


			{
				D3DXMATRIX matrix;
				D3DXMatrixIdentity(&matrix); 
				matrix._11 = 1.0f/(2.0f*photon->GetRange());
				matrix._41 = 0.5f - matrix._11*xPosition.x;
				matrix._22 = 0.0f;
				matrix._32 = matrix._11;
				matrix._42 = 0.5f - matrix._11*xPosition.z;
				matrix._33 = matrix._44 = 0.0f;

				D3DXMatrixMultiply(&matrix,&matView,&matrix);

				D3DXMATRIX matrix2;
				D3DXMatrixIdentity(&matrix2); 
				matrix2._11 = 0.0;
				matrix2._21 = 1.0f/(2.0f*photon->GetRange());
				matrix2._41 = 0.5f - matrix2._21*xPosition.y;
				matrix2._22 = 0.0f;
				matrix2._42 = 0.5f;
				matrix2._33 = matrix2._44 = 0.0f;

				D3DXMatrixMultiply(&matrix2,&matView,&matrix2);

				Asura_Direct3D::D3DDevice->SetTransform(D3DTS_TEXTURE0, &matrix);
				Asura_Direct3D::D3DDevice->SetTransform(D3DTS_TEXTURE1, &matrix2);
			}

			{
				h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,0,TotalNumberOfIndices/3);
				Asura_Render::AddToTrianglesRendered(TotalNumberOfIndices/3);
			}

		}
		Asura_Direct3D::D3DDevice->SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
		Asura_Direct3D::D3DDevice->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU);
		Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
		Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU|1);

		
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );

		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
		Asura_Render::CurrentStates.RequestDynamicLights(false);
	}
	else
	{
		Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR );
	//	float fBias = 0.0f;
	//	Asura_Direct3D::D3DDevice->SetTextureStageState(0, D3DTSS_MIPMAPLODBIAS, *((LPDWORD) (&fBias)));


	}


	/// Kevin Lea - 29/01/2002 11:48:54 - render in textures
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_PLAINTEXTURE);
	if (!DetailMap)
	{
		DetailMap = Asura_Texture_Management::AddTextureWithResetGroup("\\graphicnovel\\detail.dds", ASURA_TEXTURE_LOADED_PERMANENT );
		RainSplashTexture = Asura_Texture_Management::AddTextureWithResetGroup("\\specialfx\\splash.bmp", ASURA_TEXTURE_LOADED_PERMANENT );
		Asura_Texture_Management::LoadTextures();

		
		//D3DXCreateCubeTextureFromFile(Asura_Direct3D::D3DDevice,"graphics\\specialfx\\envmap2.dds",&cubemap);
	}

      //                             D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR | 0); 
	//Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_NORMALIZENORMALS , TRUE);

//	void SetupVertexFog(DWORD Color, DWORD Mode, BOOL UseRange, FLOAT Density);
//	SetupVertexFog(0xff202040,D3DFOG_LINEAR,false,0.1f);	
	Asura_Render::CurrentStates.RequestTextureIndex(1,DetailMap);
//	Asura_Direct3D::D3DDevice->SetTexture(1,cubemap);

	if (01)
	{
		for(int i=0; i<NumberOfPackets; i++)
		{
			Asura_Render::CurrentStates.RequestTextureIndex(0,Asura_Material_System::GetTextureID(PacketList[i].m_iMaterialID));

			if (PacketList[i].m_iMaterialID == ASURA_MATERIAL_NONE) continue;
		
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_DOUBLECOLOUR);
		
//			#ifdef ASURA_SUPPORT_WEATHER_SYSTEM
//			if (Asura_Weather_System::IsRaining() && Asura_Material_System::IsMaterialRainSplashes(PacketList[i].m_iMaterialID))
//			{				
//				Asura_Render::CurrentStates.RequestTextureIndex(1,RainSplashTexture);
//		
//				D3DXMATRIX mat;
//				D3DXMatrixIdentity(&mat);
//				mat._11 = 1.0f/16.0f; mat._12 = 0.0f; mat._13 = 0.0f; mat._14 = 0.0f; 
//				mat._21 = 0.0f; mat._22 = 1.0f/16.0f; mat._23 = 0.0f; mat._24 = 0.0f; 
//				mat._31 = (Asura_Timers::GetGameTimeScale() == 0.0f) ? 0.0f : Asura_Random::GetFloat()*5.0f;	
//				mat._32 = (Asura_Timers::GetGameTimeScale() == 0.0f) ? 0.0f : Asura_Random::GetFloat()*5.0f;
//				mat._33 = 1.0f; mat._34 = 0.0f; 
//				mat._41 = 0.0f; mat._42 = 0.0f; mat._43 = 0.0f; mat._44 = 1.0f; 
//	
//				/// Kevin Lea - 03/12/2002 17:05:52 - these floats should really be once per frame
//
//				Asura_Direct3D::D3DDevice->SetTransform( D3DTS_TEXTURE1, &mat );
//				Asura_Direct3D::D3DDevice->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2 );
//				Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU|1);
//				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_PLAINTEXTURE);
//			}
//			else
//			#endif
			if (Asura_Material_System::IsMaterialUsingDetailMap(PacketList[i].m_iMaterialID))
			{
				D3DXMATRIX mat;
				D3DXMatrixIdentity(&mat);
				mat._11 = 32.0f;
				mat._22 = 32.0f;
				/// Kevin Lea - 03/12/2002 17:05:52 - these floats should really be once per frame
				Asura_Direct3D::D3DDevice->SetTransform( D3DTS_TEXTURE1, &mat );
				Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2 );
				Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU|0);
				Asura_Render::CurrentStates.RequestTextureIndex(1,DetailMap);
				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_PLAINTEXTURE);
			}
			else
			{
				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_OPAQUE);
			}
			
			// turn on fog if we have to...
			if (Asura_Fog::IsFogRenderable())
			{
				Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);
			}
	
			h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,PacketList[i].IndexStart,PacketList[i].NumberOfIndices/3);
				
			// turn off fog if we have to...
			if (Asura_Fog::IsFogRenderable())
			{
				Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, FALSE);
			}

			
			Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
		

			if (Asura_Material_System::IsMaterialSelfIlluminating(PacketList[i].m_iMaterialID))
			{
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_OPAQUE);
				h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,PacketList[i].IndexStart,PacketList[i].NumberOfIndices/3);
			}
#ifndef ASURA_TARGET_SM3
			if (Asura_Material_System::IsMaterialReflective(PacketList[i].m_iMaterialID))
			{
				/// Kevin Lea - 04/06/2002 14:44:44 - the following can all be calculated once per frame, per camera.
				D3DXMATRIX matView;
				Asura_Direct3D::D3DDevice->GetTransform( D3DTS_VIEW, &matView );
				matView._14 = matView._24 = matView._34 = 0.0f;
				matView._41 = matView._42 = matView._43 = 0.0f;
			//	matView._44 = 1.0f;
				D3DXMatrixInverse( &matView,  NULL, &matView );

				D3DXMATRIX mat;
				D3DXMatrixIdentity(&mat);
				mat._11 = 0.5f; mat._12 = 0.0f; mat._13 = 0.0f; mat._14 = 0.0f; 
				mat._21 = 0.0f; mat._22 = -0.5f; mat._23 = 0.0f; mat._24 = 0.0f; 
				mat._31 = 0.0f; mat._32 = 0.0f; mat._33 = 1.0f; mat._34 = 0.0f; 
				mat._41 = 0.5f; mat._42 = 0.5f; mat._43 = 0.0f; mat._44 = 1.0f; 

				D3DXMatrixMultiply(&mat,&matView,&mat);

				
				Asura_Direct3D::D3DDevice->SetTransform( D3DTS_TEXTURE1, &mat );
				Asura_Direct3D::D3DDevice->SetTextureStageState( 1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2 );
				Asura_Direct3D::D3DDevice->SetTextureStageState( 1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR|1 );


				Asura_Render::CurrentStates.RequestTextureIndex( 1, Asura_Render::GetReflectanceMap());

				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_COLOUR);
				Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_REFLECTIVE);
				h=Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,0,0,NumberOfUniqueVertices,PacketList[i].IndexStart,PacketList[i].NumberOfIndices/3);

				Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
				Asura_Direct3D::D3DDevice->SetTextureStageState(1, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU|1);
			}
#endif //#ifndef ASURA_TARGET_SM3
		}
		Asura_Render::AddToTrianglesRendered(TotalNumberOfIndices/3);
	}
//	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, FALSE);

	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_EMISSIVEMATERIALSOURCE, D3DMCS_COLOR1);	
//	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

	// );
	//
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);


}

void Asura_Mesh_System::PlatformInitialiseRenderMeshes(void)
{
//return;
	Asura_Assert(Asura_Mesh_System::GetNumberOfMeshes()<ASURA_MAX_NO_OF_MESHES,"uhoh");

	for(int i=0; i<ASURA_MAX_NO_OF_MESHES; i++)
	{
		RenderMeshStorage[i] = new Asura_Direct3D_Mesh;
	}
}

void Asura_Mesh_System::PlatformShutdownRenderMeshes(void)
{
//return;
	Asura_Direct3D_EnvRenderer::ClearRenderMesh();

	for (int i=0; i<ASURA_MAX_NO_OF_MESHES; i++)
	{
		if (RenderMeshStorage[i]) delete RenderMeshStorage[i];
		RenderMeshStorage[i] = 0;
	}
}

void Asura_Mesh_System::PlatformResetRenderMeshes(void)
{
	PlatformShutdownRenderMeshes();
	PlatformInitialiseRenderMeshes();
	extern bool ForceRedrawAll;
	ForceRedrawAll = true;

	for (int i=0; i<Asura_Mesh_System::GetNumberOfMeshes(); i++)
	{
		Asura_Mesh* mesh = Asura_Mesh_System::MeshStorage[i];

		if (mesh)
		{
			Asura_Surface_Mesh* surfaceMesh = Asura_Surface_Mesh_System::GetSurfaceMesh(mesh->Index);
			if (surfaceMesh) surfaceMesh->Invalidate();

			mesh->PrelightingDataValid = false;
		}
	}
}

#ifdef ASURA_TOOLS
void Asura_Mesh_System::PlatformResetRenderMeshesOnly()
{
	for (int i=0; i<Asura_Mesh_System::GetNumberOfMeshes(); i++)
	{
		if(RenderMeshStorage[i])
		{
			RenderMeshStorage[i]->ReleaseBuffers();
		}
	}
}

u_int Asura_Mesh_System::Platform_GetRenderMeshTriCount( int iIndex )
{
    return RenderMeshStorage[iIndex]->GetNumTris();
}
#endif //ASURA_TOOLS

void Asura_Mesh_System::PlatformRenderMeshHighlight( Asura_Mesh* pxMesh,
                                                     const Asura_Vector_3* pxPos,
                                                     const Asura_Matrix_3x3* pxOri,
                                                     u_int uColour,
                                                     bool bWireframe,
                                                     bool bOffset )
{
	Asura_Direct3D_Mesh* pxRenderMesh = RenderMeshStorage[pxMesh->Index];

	if( pxRenderMesh->NotInitialised() )
	{
		pxRenderMesh->Initialise( pxMesh );
	}
	
//	pxRenderMesh->UpdateLighting( pxMesh, pxPos, pxOri );

	pxRenderMesh->RenderHighlight( pxPos, pxOri, uColour, bWireframe, bOffset );
}

void Asura_Mesh_System::PlatformRenderMesh( Asura_Mesh* mesh, const Asura_Vector_3* position, const Asura_Matrix_3x3* orientation)
{

	Asura_Direct3D_Mesh* renderMesh = RenderMeshStorage[mesh->Index];

	if (renderMesh->NotInitialised())
	{
		renderMesh->Initialise(mesh);
	}
	
//	renderMesh->UpdateLighting(mesh,position,orientation);

	renderMesh->Render(position,orientation);
}

void Asura_Mesh_System::PlatformResetRenderMesh(Asura_Mesh* mesh)
{
	Asura_Direct3D_Mesh*& renderMesh = RenderMeshStorage[mesh->Index];

	if (renderMesh && !renderMesh->NotInitialised())
	{
		delete renderMesh;
		renderMesh = new Asura_Direct3D_Mesh;

	}
#if 01
	mesh->InvalidateNormals();
#endif

	Asura_Surface_Mesh* surfaceMesh = Asura_Surface_Mesh_System::GetSurfaceMesh(mesh->Index);
	if (surfaceMesh) surfaceMesh->Invalidate();
	mesh->PrelightingDataValid = false;
}

#ifdef ASURA_TOOLS
void Asura_Mesh_System::PlatformResetRenderMeshOnly(const Asura_Mesh* mesh)
{
	//For texturing mode...
	//Avoid resetting the normals and lighting to reduce the amount
	//of unneeded recalculation. RAR 17/03/03

	Asura_Direct3D_Mesh*& renderMesh = RenderMeshStorage[mesh->Index];

	if (renderMesh && !renderMesh->NotInitialised())
	{
		delete renderMesh;
		renderMesh = new Asura_Direct3D_Mesh;

	}
}

void Asura_Mesh_System::PlatformResetMeshLightingOnly(Asura_Mesh* mesh)
{
	//(PSJE) 23-05-2003
	//Avoid resetting the render mesh and normals of reduce the amountof unneeded recalculation.
	Asura_Surface_Mesh* surfaceMesh = Asura_Surface_Mesh_System::GetSurfaceMesh(mesh->Index);
	if (surfaceMesh) 
	{
		surfaceMesh->Invalidate();
	}
	mesh->PrelightingDataValid = false;
}
#endif // ASURA_TOOLS

void Asura_Mesh_System::PlatformDeleteMesh( int iIndex )
{
	Asura_Direct3D_Mesh* pxMeshToDelete = RenderMeshStorage[ iIndex ];
	Asura_Direct3D_Mesh* pxLastMesh = RenderMeshStorage[ Asura_Mesh_System::GetNumberOfMeshes() - 1 ];

	delete pxMeshToDelete;

	RenderMeshStorage[ iIndex ] = pxLastMesh;

	RenderMeshStorage[ Asura_Mesh_System::GetNumberOfMeshes() - 1 ] = new Asura_Direct3D_Mesh;
}

//#include "Rendering/Asura_PlatformRenderMesh.h"
//#include "Rendering/Asura_Surface.h"
//#include "Mesh/Asura_Mesh.h"
//#include "Rendering/Asura_Render.h" // for defn of ASURA_TEXTURE_NONE
//#include "MeshProcessing/Tools_SubdividedMesh.h"

//extern Asura_Direct3D_Mesh* RenderMeshStorage[];
extern VertexFormat* VertexArray;
extern unsigned short* IndexArray;
extern Asura_Bicubic_Patch patch;

#if defined(ASURA_TOOLS)
// (SB - 11/12/06) Hook to generate vertex colour from  flags
Asura_Surface_Mesh_System::GeneratePatchVertexColourHook* Asura_Surface_Mesh_System::s_pfnGeneratePatchVertexColourHook;
#endif

void Asura_Surface_Mesh_System::MakeVertexBufferAndIndexList(Asura_Mesh& mesh, const Asura_Vector_3& position)
{
	Asura_Direct3D_Mesh* d3dMesh = RenderMeshStorage[mesh.Index];
	if (!VertexArray)
	{
		VertexArray = new VertexFormat[MAX_SMOOTHED_VERTICES];
		IndexArray = new unsigned short [MAX_SMOOTHED_VERTICES*3];
	}

	//Check to see if its this meshes turn for having prelighting recalcuated. RAR
	#ifdef ASURA_TOOLS
	if(s_bRecalculatePrelightingOnSubdividedMeshes)
	{
		if(s_uPrelightingFrameStamp != Asura_Timers::GetFrameNumber())
		{
			//reset the  count of vertices done this frame
			s_uPrelightingFrameStamp = Asura_Timers::GetFrameNumber();
			s_uPrelightingDoneThisFrame = 0;
		}
		//check if any more preighting should be done this frame
		if(s_uPrelightingDoneThisFrame<s_uMaxPrelightingPerFrame)
		{
			if(d3dMesh->IsValid)
			{
				if(!d3dMesh->m_bPrelightingBeingCalculated)
				{
					//mesh wasn't made with writable vertex buffer , so redo it
					d3dMesh->ReleaseBuffers();
				}
				else if(d3dMesh->m_iVertexPrelightingProgress<d3dMesh->NumberOfUniqueVertices)
				{
					//do some prelighting for this mesh	
					d3dMesh->CalculatePrelighting();
				}
			}
		}
	}
	#endif //ASURA_TOOLS

	if (!d3dMesh->IsValid)
	{
		
		Asura_PatchSubdivision::FindPointsNearModule(&mesh,position);
		
		Asura_Surface_Mesh* surfaceMesh = &s_axSurfaceMeshStorage[mesh.Index];
		patch.SetMaxDetailLevel(Asura_Render::GetPatchesSubdivLevel());
		
		//CalculatePrelighting(mesh,position);

		int totalVertices = 0;
		int totalIndices = 0;

		//Asura_Polygon_Packet axConstructionPacketList[ASURA_MAX_MATERIALS+1];
		
		int numberOfPackets=0;

		//constructionPacketList[numberOfPackets].NumberOfIndices=0;
		//constructionPacketList[numberOfPackets].m_iMaterialID = surfaceMesh->Surfaces[0].m_iMaterialID;
		//constructionPacketList[numberOfPackets].IndexStart = 0;

		//determine how many materials are used
		int aiMaterialsUsed[ASURA_MAX_NO_OF_MATERIALS+1];
		for (int s=0; s<surfaceMesh->NumberOfSurfaces; s++)
		{
			int iPolygonFlags = 0;
			if(mesh.PolygonFlagDataValid)
			{
				iPolygonFlags = mesh.PolygonFlagDataPtr[s].Flags;
			}

			if (iPolygonFlags & AsuraPolygonFlag_DontRender)
			{
				continue;
			}

			Asura_Surface* surfaceDataPtr = &surfaceMesh->Surfaces[s];
			int iMaterialID = surfaceDataPtr->m_iMaterialID;

			int iPacket = 0;
			for(;iPacket<numberOfPackets;iPacket++)
			{
				if(iMaterialID == aiMaterialsUsed[iPacket]) break;
			}
			if(iPacket == numberOfPackets)
			{
				//found a new material
				aiMaterialsUsed[numberOfPackets++] = iMaterialID;
			}
		}

		Asura_Polygon_Packet* pxPacketList = new Asura_Polygon_Packet[numberOfPackets];
		for(int iPacket = 0;iPacket<numberOfPackets;iPacket++)
		{
			pxPacketList[iPacket].NumberOfIndices = 0;
			pxPacketList[iPacket].m_iMaterialID = aiMaterialsUsed[iPacket];
			pxPacketList[iPacket].IndexStart = 0;
		}

		int* piTempPolygonIndeces = new int[MAX_SMOOTHED_VERTICES];


		for(int iPacket = 0;iPacket<numberOfPackets;iPacket++)
		{
			Asura_Polygon_Packet* pxPacket = &pxPacketList[iPacket];
			if(iPacket>0)
			{
				pxPacket->IndexStart = pxPacketList[iPacket-1].NumberOfIndices + pxPacketList[iPacket-1].IndexStart;
			}
			//check for all polygons using this material
			
			for (int s=0; s<surfaceMesh->NumberOfSurfaces; s++)
			{
				int iPolygonFlags = 0;
				if(mesh.PolygonFlagDataValid)
				{
					iPolygonFlags = mesh.PolygonFlagDataPtr[s].Flags;
				}

				if (iPolygonFlags & AsuraPolygonFlag_DontRender)
				{
					continue;
				}

				Asura_Surface* surfaceDataPtr = &surfaceMesh->Surfaces[s];

				if(surfaceDataPtr->m_iMaterialID != pxPacket->m_iMaterialID)
				{
					continue;
				}

		
				patch.Render(surfaceDataPtr,false);

				u_int uIterationsRequired = (iPolygonFlags & AsuraPolygonFlag_DoubleSided) ? 1 : 0;
				
				for (u_int u = 0; u <= uIterationsRequired; u++) 
				{
					if (totalVertices + patch.nextAvailableVertex >= MAX_SMOOTHED_VERTICES)
					{
						Asura_Assertf( totalVertices + patch.nextAvailableVertex < MAX_SMOOTHED_VERTICES, ("Too many smoothed vertices in mesh %s. Try reducing curves or splitting it.", mesh.Name.GetString()));
						// attempt to recover
						break;
					}

					if (totalIndices + patch.noOfIndices >= MAX_SMOOTHED_VERTICES * 3)
					{
						Asura_Assertf( totalIndices + patch.noOfIndices < MAX_SMOOTHED_VERTICES * 3, ("Too many smoothed polygons in mesh %s. Try reducing curves or splitting it.", mesh.Name.GetString()));
						// attempt to recover
						break;
					}

					for (int v=0; v<patch.nextAvailableVertex; v++)
					{
						VertexArray[totalVertices+v].X = patch.Vertices[v].du0dv0.x;
						VertexArray[totalVertices+v].Y = patch.Vertices[v].du0dv0.y;
						VertexArray[totalVertices+v].Z = patch.Vertices[v].du0dv0.z;
						
						VertexArray[totalVertices+v].m_NX = patch.Vertices[v].Normal.x;
						VertexArray[totalVertices+v].m_NY = patch.Vertices[v].Normal.y;
						VertexArray[totalVertices+v].m_NZ = patch.Vertices[v].Normal.z;

						if (u)
						{
							VertexArray[totalVertices+v].m_NX = -VertexArray[totalVertices+v].m_NX;
							VertexArray[totalVertices+v].m_NY = -VertexArray[totalVertices+v].m_NY;
							VertexArray[totalVertices+v].m_NZ = -VertexArray[totalVertices+v].m_NZ;
						}

						VertexArray[totalVertices+v].U0 = patch.Vertices[v].u;
						VertexArray[totalVertices+v].V0 = patch.Vertices[v].v;

						if(Asura_Material_System::IsMaterialDualTexture(patch.m_iMaterialID))
						{
							VertexArray[totalVertices+v].U1 = patch.Vertices[v].u2;
							VertexArray[totalVertices+v].V1 = patch.Vertices[v].v2;
						}
						else
						{
							VertexArray[totalVertices+v].U1 = patch.Vertices[v].u;
							VertexArray[totalVertices+v].V1 = patch.Vertices[v].v;
						}

#if defined(ASURA_TOOLS)
						// (SB - 11/12/06) Project hook to set colour based on flags etc
						if (s_pfnGeneratePatchVertexColourHook)
						{
							VertexArray[totalVertices + v].Colour = s_pfnGeneratePatchVertexColourHook(mesh, patch, s, v);
						}
						else
#endif
						{
							VertexArray[totalVertices+v].Colour = patch.Vertices[v].Diffuse;
						}

						piTempPolygonIndeces[totalVertices+v] = s;

					}

					for(int i=0; i<patch.noOfIndices; i++)
					{
						if (u)
						{
							IndexArray[totalIndices+i] = (unsigned short)(patch.indexOrder[patch.noOfIndices-1-i]+totalVertices);
						}
						else
						{
							IndexArray[totalIndices+i] = (unsigned short)(patch.indexOrder[i]+totalVertices);
						}
					}

					pxPacket->NumberOfIndices += patch.noOfIndices;

					totalIndices+=patch.noOfIndices;
					totalVertices+=patch.nextAvailableVertex;
				}
			}
		}

		d3dMesh->PacketList = pxPacketList;
		d3dMesh->NumberOfPackets = numberOfPackets;


		//copy the polygon indeces for the vertices into the d3dMesh. RAR 27/05/03
		d3dMesh->m_pxMesh = &mesh;
		d3dMesh->m_xPosition = position;
		d3dMesh->m_piPolygonIndeces = new int[totalVertices];
		for(int i=0;i<totalVertices;i++)
		{
			d3dMesh->m_piPolygonIndeces[i] = piTempPolygonIndeces[i];
		}
		delete [] piTempPolygonIndeces;


		d3dMesh->NumberOfUniqueVertices = totalVertices;
		d3dMesh->TotalNumberOfIndices = totalIndices;
		d3dMesh->m_bPrelightingBeingCalculated = s_bRecalculatePrelightingOnSubdividedMeshes;

		if (totalVertices!=0)
		{
			d3dMesh->MakeVertexBuffer(VertexArray,IndexArray);
		}
		d3dMesh->IsValid = true;
		d3dMesh->m_bVerticesOffsetByModulePosition = true;
	}
}

void Asura_Surface_Mesh_System::RenderFromVertexBuffer(Asura_Mesh& xMesh)
{
	Asura_Direct3D_Mesh* pxD3DMesh = RenderMeshStorage[xMesh.Index];
	Asura_Assert(pxD3DMesh && pxD3DMesh->IsValid, "D3D mesh not created for this mesh yet");
	if (pxD3DMesh && pxD3DMesh->NumberOfUniqueVertices)
	{
		pxD3DMesh->SimpleRender();
	}
}

void RenderMesh(Asura_Mesh* pxMesh, const Asura_Vector_3& xPosition,int iModuleIndex)
{
	Asura_Bounding_Box xBB;
	xBB = pxMesh->BoundingBox;
	xBB.Displace(xPosition);

//	return;

	if(Asura_Lighting_System::MakeLightListForObject(xBB))
	{
		Asura_Mesh_System::PlatformResetRenderMesh(pxMesh);	
	}

	if(!pxMesh->PrelightingDataValid)
	{
		Asura_Lighting_System::CalculatePrelighting(*pxMesh,xPosition,iModuleIndex);
	}

	const Asura_Vector_3 xBBCentre = xBB.CalculateCentre();
	Asura_Photon_System::MakePhotonListForBoundingBox( xBB, xBBCentre, false, true );

	Asura_Render::SetPatchesOn(true);
//	Asura_Render::PatchesSubdivLevel = 3;
//	Asura_Render::StaticSubdivision = true;

	Asura_Surface_Mesh_System::RenderAsSurfaces(*pxMesh,xPosition);
}

void SetRecalculatePrelightingForSubdividedMeshes(bool bRecalc)
{
	s_bRecalculatePrelightingOnSubdividedMeshes = bRecalc;
}

#endif // #ifdef ASURA_TOOLS
