/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float4x4 g_xProjectionMatrix;
const float4x4 g_xInvViewMatrix;
const float4x4 g_xLightOrientation;

const float4 g_xTwoOverScreen;			// dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float4 g_xCameraTanHalfFOV;		// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.

// Directional light parameters
const float4 g_xLightDirection;

// Omni light parameters
const float4 g_xFalloff;
const float4 g_xLightPosScale;			// World-space light pos in xyz, 1/range in w
const float4 g_xLightColour;			// Colour in rgb, intensity in a
const float4 g_xMeshScaleAndPos;		// Position/scale for the light volume

const bool   g_bUseProjCube;
const bool	 g_bUseDualProjCube;
const bool   g_bUseAngularRange;

const float  g_fProjCubeLerpDist;

#define DEFINE_SAMPLER( TextureName, Register )											\
	const texture g_x##TextureName;														\
	const sampler2D g_x##TextureName##Sampler : register( Register ) = sampler_state	\
	{																					\
		Texture = <g_x##TextureName>;													\
		AddressU = CLAMP;																\
		AddressV = CLAMP;																\
		MinFilter = POINT;																\
		MagFilter = POINT;																\
		MipFilter = POINT;																\
	};

DEFINE_SAMPLER( NormalDepthBuffer, s6 );
DEFINE_SAMPLER( AlbedoBuffer, s7 );

const texture g_xProjCubeMap;
const samplerCUBE g_xProjCubeMapSampler = sampler_state
{
	Texture = <g_xProjCubeMap>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = POINT;
	sRGBTexture = <g_bUseGammaCorrection>;
};

const texture g_xProjCubeMapBlurred;
const samplerCUBE g_xProjCubeMapBlurredSampler = sampler_state
{
	Texture = <g_xProjCubeMapBlurred>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = POINT;
	sRGBTexture = <g_bUseGammaCorrection>;
};


#undef DEFINE_SAMPLER

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct VS_Output
{
	float4 xPosition  : POSITION;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// The vertex shader recreates the light frustum in world space, then projects
// it as an object in the standard camera view

VS_Output VShader( const float4 xInPosition : POSITION )
{
	VS_Output xOut;

	float4 xPos = xInPosition;

	xPos.xyz *= g_xMeshScaleAndPos.www;
	xPos.xyz += g_xMeshScaleAndPos.xyz;
	
	xOut.xPosition = mul( xPos, g_xViewProjectionMatrix );	

	return xOut;
}

// -----------------------------------------------------------------------------
// Pixel Shader (ClearDestAlpha)
// Just output 0
// -----------------------------------------------------------------------------

float4 PShader_ClearDestAlpha( void ) : COLOR
{
	return float4(0,0,0,0);
}



// -----------------------------------------------------------------------------
// Pixel Shader (Photon)
// Material response lighting from a point light
// -----------------------------------------------------------------------------

float4 PShader_Photon( const VS_Output xIn, float2 xScreenPos : VPOS ) : COLOR
{
	// Fetch the world-space normal and screen depth
	const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * (0.5 * (g_xTwoOverScreen.xy));
	const float4 xBufferNormal = tex2D(g_xNormalDepthBufferSampler, xNormalizedCoords);
	const half4 xBufferAlbedo = tex2D(g_xAlbedoBufferSampler, xNormalizedCoords);
	
	half3 xNormal;
	float fScreenDepth;
	half3 xAlbedoColour;
	half  hSpecularStrength;
	float fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.x * fScreenDepth *  (xScreenPos.x * g_xTwoOverScreen.x - 1.0),
													g_xCameraTanHalfFOV.y * fScreenDepth * -(xScreenPos.y * g_xTwoOverScreen.y - 1.0),
													fScreenDepth,
													1.0 );

	float4 xOutput;

	float4 xWorldPosition	= mul( xCameraViewSpacePosition, g_xInvViewMatrix );
	float3 xLightDelta		= g_xLightPosScale.xyz - xWorldPosition.xyz;
	half3  xLightStrength	= xLightDelta.xyz * g_xLightPosScale.w;
	half   fLightDistance	= sqrt( dot( xLightStrength.xyz, xLightStrength.xyz ) );
	float3 xCameraDir		= normalize(g_xCameraPosition.xyz - xWorldPosition.xyz);

	// Early-out if not in range of the light
	if( 0.5 - abs( fLightDistance - 0.5 ) >= 0 )
	{
		// Handle the falloff
		half  fLightAttenuation = min(1,pow(abs((1.0f - fLightDistance) * g_xFalloff.x), g_xFalloff.y));

		half3 xLightAttenuatedColour = g_xLightColour.rgb * fLightAttenuation;

		xOutput.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
			half4( xNormal, 1 ), xCameraDir, normalize( xLightDelta ), xLightAttenuatedColour,
			xLightAttenuatedColour, dot( xLightAttenuatedColour, float3( 0.333, 0.333, 0.333 ) ) );

		xOutput.a = 1;
	}
	else
	{
		xOutput = float4( 0, 0, 0, 0 );
	}

	return xOutput;
}


// -----------------------------------------------------------------------------
// Pixel Shader (Limited Photon)
// Material response lighting from a point light - with angular constraints
// -----------------------------------------------------------------------------

float4 PShader_LimitedPhoton( const VS_Output xIn, float2 xScreenPos : VPOS ) : COLOR
{
	// Fetch the world-space normal and screen depth
	const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * (0.5 * (g_xTwoOverScreen.xy));
	const float4 xBufferNormal = tex2D(g_xNormalDepthBufferSampler, xNormalizedCoords);
	const half4 xBufferAlbedo = tex2D(g_xAlbedoBufferSampler, xNormalizedCoords);
	
	half3 xNormal;
	float fScreenDepth;
	half3 xAlbedoColour;
	half  hSpecularStrength;
	float fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.x * fScreenDepth *  (xScreenPos.x * g_xTwoOverScreen.x - 1.0),
													g_xCameraTanHalfFOV.y * fScreenDepth * -(xScreenPos.y * g_xTwoOverScreen.y - 1.0),
													fScreenDepth,
													1.0 );

	float4 xOutput;

	float4 xWorldPosition	= mul( xCameraViewSpacePosition, g_xInvViewMatrix );
	float3 xLightDelta		= g_xLightPosScale.xyz - xWorldPosition.xyz;
	half3  xLightStrength	= xLightDelta.xyz * g_xLightPosScale.w;
	half   fLightDistance	= sqrt( dot( xLightStrength.xyz, xLightStrength.xyz ) );
	float3 xCameraDir		= normalize(g_xCameraPosition.xyz - xWorldPosition.xyz);

	// Early-out if not in range of the light
	if( 0.5 - abs( fLightDistance - 0.5 ) >= 0 )
	{
		// Handle the falloff
		half  fLightAttenuation = min(1,pow(abs((1.0f - fLightDistance) * g_xFalloff.x), g_xFalloff.y));
		
				
		float3 xOfs = normalize(g_xLightPosScale.xyz-xWorldPosition);
		
		if (g_bUseAngularRange)
		{
			// Handle the angular falloff		
			float  fDot = dot(xOfs, g_xLightDirection);		
			fLightAttenuation *= saturate((fDot + g_xFalloff.z) * g_xFalloff.w);
		}
		
		half3 xLightAttenuatedColour = g_xLightColour.rgb * fLightAttenuation;


		if (g_bUseProjCube)
		{
			float3 xCol;
			
			// Cubemap projector
			float3 xDir = mul(xOfs, g_xLightOrientation);
			xCol = texCUBE( g_xProjCubeMapSampler, xDir);		
			
			if (g_bUseDualProjCube)
			{
				float fLerp = min(fLightDistance/g_fProjCubeLerpDist,1.0f);
				xCol = lerp(xCol, texCUBE( g_xProjCubeMapBlurredSampler, xDir), fLerp);
			}
			
			xLightAttenuatedColour *= xCol;
		}

		xOutput.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
			half4( xNormal, 1 ), xCameraDir, normalize( xLightDelta ), xLightAttenuatedColour,
			xLightAttenuatedColour, dot( xLightAttenuatedColour, float3( 0.333, 0.333, 0.333 ) ) );

		xOutput.a = 1;
	}
	else
	{
		xOutput = float4( 0, 0, 0, 0 );
	}

	return xOutput;
}


// -----------------------------------------------------------------------------
// Pixel Shader (Infinite)
// Material response lighting from a parallel light
// -----------------------------------------------------------------------------

float4 PShader_Infinite( const VS_Output xIn, float2 xScreenPos : VPOS ) : COLOR
{
	// Fetch the world-space normal and screen depth
	const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * (0.5 * (g_xTwoOverScreen.xy));
	const float4 xBufferNormal = tex2D(g_xNormalDepthBufferSampler, xNormalizedCoords);
	const half4 xBufferAlbedo = tex2D(g_xAlbedoBufferSampler, xNormalizedCoords);
	
	half3 xNormal;
	float fScreenDepth;
	half3 xAlbedoColour;
	half  hSpecularStrength;
	float fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial );

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.x * fScreenDepth * (xScreenPos.x * g_xTwoOverScreen.x  - 1.0),
													g_xCameraTanHalfFOV.y  * fScreenDepth * -(xScreenPos.y * g_xTwoOverScreen.y - 1.0),
													fScreenDepth,
	 	  											1.0);
		   														 
	float4 xOutput;

	float4 xWorldPosition   = mul(  xCameraViewSpacePosition, g_xInvViewMatrix);			
	float3 xLightDelta		= g_xLightPosScale.xyz - xWorldPosition.xyz;
	half3  xLightStrength	= xLightDelta.xyz * (g_xLightPosScale.w);	
	half   fLightDistance	= sqrt( dot( xLightStrength.xyz, xLightStrength.xyz ))	;					   
	float3 xCameraDir		= normalize(g_xCameraPosition.xyz - xWorldPosition.xyz);
	
	clip( 0.5 - abs( fLightDistance - 0.5));
	
	// Handle the falloff
	half  fLightAttenuation = min(1,pow(abs((1.0f - fLightDistance) * g_xFalloff.x), g_xFalloff.y));			
	
	half3 xLightAttenuatedColour = g_xLightColour * fLightAttenuation;
	xOutput.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
		half4( xNormal, 1), xCameraDir, g_xLightDirection, xLightAttenuatedColour,
		xLightAttenuatedColour, dot(xLightAttenuatedColour, float3(0.333,0.333,0.333)) );

	xOutput.a = 1;		 	

	return xOutput;
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	// This is now implemented as several 'passes', instead of one ubershader with loads of static
	// branching. We only ever render 1-2 of these passes per photon

	// Pass Type 0 - Clear dest. alpha (clear shadow mask)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_ClearDestAlpha
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_ClearDestAlpha();
	}
	
	// Pass Type 1 - Render as a photon (point light)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_Photon
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_Photon();
	}

	// Pass Type 2 - Render as a photon (point light)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_LimitedPhoton
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton();
	}

	// Pass Type 3 - Render as an infinite light
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_Infinite
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_Infinite();
	}
}
