/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

// First pass, creating the maps

float4x4 g_xShadowViewProj;

sampler2D g_xTransparencyMask : register(s0) = sampler_state
{
	AddressU = WRAP;
	AddressV = WRAP;
};

float2 g_xTexAnim;

// Second & third passes, Gaussian blur

// (1/tex width, 0.5/tex width)
float2 g_xRcpBlurTexSize;

// Horz pass:
// {-1.5/width, -0.5/width, 0.5/width, 1.5/width, 2.5/width}
// Vert pass:
// {-1/width, 1/width, 1/width, 2/width} (only four needed)
float g_afBlurOffsets[5];

// VPOS isn't viewport-relative, but render target-relative
float g_fCascadeOffset;

texture g_xBlurTexture;
sampler2D g_xBlurSampler = sampler_state
{
	Texture = <g_xBlurTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};

// Fourth pass, applying to scene

float4x4 g_axShadowViewProj[4];
float4x4 g_xInvCamViewProj;

// (1/width, 1/height, 0.5/width, 0.5/height)
float4 g_xScreenDims;

float2 g_xTwoOverScreenDims;
float2 g_xCameraTanHalfFOV;

float4 g_xSplits;

float g_fRcpMaxMaps;

float g_fPower;
float g_afStrengths[4];

texture g_xDepthTexture;
sampler2D g_xDepthSampler = sampler_state
{
	Texture = <g_xDepthTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};

texture g_xShadowMapTexture;
sampler2D g_xShadowMap = sampler_state
{
	Texture = <g_xShadowMapTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// Create shadow map, no transparency

float4 VShader_CreateShadowMap( const float4 xPos : POSITION, out float fDepth : TEXCOORD )
	: POSITION
{
	const float4 xResult = mul( xPos, g_xShadowViewProj );

	fDepth = xResult.z;

	return xResult;
}

float4 PShader_CreateShadowMap( const float fDepth : TEXCOORD ) : COLOR
{
	return float4( fDepth, fDepth * fDepth, 0.0f, 1.0f );
}

// Create shadow map, 1-bit transparent strips. We need to alpha test against the texture

float4 VShader_CreateShadowMap_Trans( const float4 xPos : POSITION, out float fDepth : TEXCOORD,
	const float2 xUV : TEXCOORD, out float2 xOutUV : TEXCOORD1 ) : POSITION
{
	const float4 xResult = mul( xPos, g_xShadowViewProj );

	fDepth = xResult.z;

	xOutUV = xUV + g_xTexAnim;

	return xResult;
}

float4 PShader_CreateShadowMap_Trans( const float fDepth : TEXCOORD, const float2 xUV : TEXCOORD1 )
	: COLOR
{
	float4 xResult;

	xResult.r = fDepth;
	xResult.g = fDepth * fDepth;
	xResult.b = 0.0f;
	xResult.a = tex2D( g_xTransparencyMask, xUV ).a;

	return xResult;
}

// Blur horizontally then vertically

float4 VShader_Blur( const float4 xPos : POSITION ) : POSITION
{
	return xPos;
}

float4 PShader_BlurHorz( const float2 xScreenPos : VPOS ) : COLOR
{
	float2 xUV;
	xUV.y = xScreenPos.y * g_xRcpBlurTexSize.x + g_xRcpBlurTexSize.y;

	// Point-sample the five taps, since we can't apply filtering before computing the squares

	float4 xSamples0123;
	float  fSample4;

	xUV.x = xScreenPos.x * g_xRcpBlurTexSize.x + g_afBlurOffsets[0];
	xSamples0123.x = tex2D( g_xBlurSampler, xUV );

	xUV.x = xScreenPos.x * g_xRcpBlurTexSize.x + g_afBlurOffsets[1];
	xSamples0123.y = tex2D( g_xBlurSampler, xUV );

	xUV.x = xScreenPos.x * g_xRcpBlurTexSize.x + g_afBlurOffsets[2];
	xSamples0123.z = tex2D( g_xBlurSampler, xUV );

	xUV.x = xScreenPos.x * g_xRcpBlurTexSize.x + g_afBlurOffsets[3];
	xSamples0123.w = tex2D( g_xBlurSampler, xUV );

	xUV.x = xScreenPos.x * g_xRcpBlurTexSize.x + g_afBlurOffsets[4];
	fSample4 = tex2D( g_xBlurSampler, xUV );

	const float4 xKernel = { 1.f/16.f, 4.f/16.f, 6.f/16.f, 4.f/16.f };

	const float fBlurredDepth = dot( xSamples0123, xKernel ) + fSample4 * xKernel.x;

	// Square the samples
	const float4 xSquaredSamples0123 = xSamples0123 * xSamples0123;
	const float fSquaredSample4 = fSample4 * fSample4;

	// Find the "average" of the squares = E(x^2)
	const float fBlurredSquaredDepth =
		dot( xSquaredSamples0123, xKernel ) + fSquaredSample4 * xKernel.x;

	return float4( fBlurredDepth, fBlurredSquaredDepth, 0.0f, 0.0f );
}

float4 PShader_BlurVert( const float2 xScreenPos : VPOS ) : COLOR
{
	float2 xUV;
	xUV.x = xScreenPos.x * g_xRcpBlurTexSize.x + g_xRcpBlurTexSize.y;

	// VPOS is relative to RT, not viewport. Subtract off our viewport location
	const float fBaseY = (xScreenPos.y - g_fCascadeOffset) * g_xRcpBlurTexSize.x;

	// This shader works a little differently to the horizontal pass. Because we've got the squares
	// already, we can use linear filtering to reduce the number of samples needed to four

	float4 xSamples1, xSamples2;

	xUV.y = fBaseY + g_afBlurOffsets[0];
	xSamples1.xy = tex2D( g_xBlurSampler, xUV );

	xUV.y = fBaseY + g_afBlurOffsets[1];
	xSamples1.zw = tex2D( g_xBlurSampler, xUV );

	xUV.y = fBaseY + g_afBlurOffsets[2];
	xSamples2.xy = tex2D( g_xBlurSampler, xUV );

	xUV.y = fBaseY + g_afBlurOffsets[3];
	xSamples2.zw = tex2D( g_xBlurSampler, xUV );

	// NB kernel slightly different from horz pass, but maths works out the same due to filtering
	const float4 xKernel = { 2.f/16.f, 6.f/16.f, 6.f/16.f, 2.f/16.f };

	const float fBlurredDepth = dot( float4( xSamples1.xz, xSamples2.xz ), xKernel );
	const float fBlurredSquaredDepth = dot( float4( xSamples1.yw, xSamples2.yw ), xKernel );

	return float4( fBlurredDepth, fBlurredSquaredDepth, 0.0f, 0.0f );
}

// Apply shadows to scene

float4 VShader_ApplyShadows( const float4 xPos : POSITION ) : POSITION
{
	return xPos;
}

float Chebychev( const float t, const float m, const float s2 )
{
	const float d = t - m;
	return s2 / (s2 + d*d);
}

float4 PShader_ApplyShadows( const float2 xUV : VPOS ) : COLOR
{
	// Reconstruct camera view space position
	const float2 xScreenUV = xUV * g_xScreenDims.xy + g_xScreenDims.zw;
	const float fViewSpaceDepth = abs( tex2D( g_xDepthSampler, xScreenUV ).b );

	float4 xViewSpacePos =
	{
		g_xCameraTanHalfFOV * fViewSpaceDepth * (xUV * g_xTwoOverScreenDims - 1.0f),
		fViewSpaceDepth,
		1.0f
	};

	// In view space, y goes up (our view matrices correct for Asura's y axis)
	xViewSpacePos.y = -xViewSpacePos.y;

	// Actually this is the inv view matrix only now
	const float4 xWorldSpacePos = mul( xViewSpacePos, g_xInvCamViewProj );

	// Calculate which segment of the cascade to use

	const float4 xDepths = fViewSpaceDepth.xxxx;
	const float4 xResults = (xDepths > g_xSplits);
	const float4 xOne = { 1.0f, 1.0f, 1.0f, 1.0f };
	const float fMapToUse = dot( xResults, xOne );

	// Reproject into shadow map clip space
	float4 xShadowCoords = mul( xWorldSpacePos, g_axShadowViewProj[ fMapToUse ] );

	// Scale from [-1,1] into [0,1] and invert y for texture lookup
	// TODO: incorporate this into the above matrix
	xShadowCoords.xy = xShadowCoords.xy * 0.5f + 0.5f;
	xShadowCoords.y = 1.0f - xShadowCoords.y;

	clip( xShadowCoords.xyz );
	clip( 1.0f - xShadowCoords.xyz );

	// Translate to the correct part of the packed texture
	float2 xShadowUV = xShadowCoords.xy;
	xShadowUV.y = (xShadowUV.y + fMapToUse) * g_fRcpMaxMaps;

	// Sample depth from shadow's POV and work out if we're in shadow

	const float2 xSample = tex2D( g_xShadowMap, xShadowUV );

	const float fMean = xSample.r;
	const float fVariance = saturate( xSample.g - fMean * fMean + 0.0001f );

	const float fChebychev = Chebychev( xShadowCoords.z, fMean, fVariance );

	const float fResult = xShadowCoords.z > fMean
						? 1.0f - pow( abs( fChebychev ), g_fPower )
						: 0.0f;

	return float4( 0.0f, 0.0f, 0.0f, saturate( fResult * g_afStrengths[ fMapToUse ] ) );
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass CreateShadowMap
	{
		VertexShader = compile vs_3_0 VShader_CreateShadowMap();
		PixelShader = compile ps_3_0 PShader_CreateShadowMap();
	}

	pass CreateShadowMap_Trans
	{
		VertexShader = compile vs_3_0 VShader_CreateShadowMap_Trans();
		PixelShader = compile ps_3_0 PShader_CreateShadowMap_Trans();
	}

	pass BlurHorz
	{
		VertexShader = compile vs_3_0 VShader_Blur();
		PixelShader = compile ps_3_0 PShader_BlurHorz();
	}

	pass BlurVert
	{
		VertexShader = compile vs_3_0 VShader_Blur();
		PixelShader = compile ps_3_0 PShader_BlurVert();
	}

	pass ApplyShadows
	{
		VertexShader = compile vs_3_0 VShader_ApplyShadows();
		PixelShader = compile ps_3_0 PShader_ApplyShadows();
	}
}