#include "Shared.fxh"
#include "Deferred_Shared.fxh"

//------------------------------------
struct VS_Input
{
	float4	m_xPosition		: POSITION0;
	float4	m_xData1		: COLOR0;
	float4	m_xData2		: COLOR1;
};

struct VS_Output
{
	float4	m_xPosition		: POSITION0;	
	float2	m_xUV			: TEXCOORD0;
	float4	m_xColour		: TEXCOORD1;
	float4	m_xNormalDistance: TEXCOORD2;
	float4	m_xPos			: TEXCOORD3;
};

// Expected constants:
//  c00-c03: ViewProj matrix
//	c04: Billboard Right X, Y, Z, 0.5
//	c05: Billboard Up X, Y, Z, 1.0
//  c06: Camera Position
//  c07: 255, 512, ASURA_DIRECT3D_CARPET_GAME_RENDER_VS_CONST_STYLE_START, 0
//	c08 09 10 11:
//		Array[4] Corner Offsets X, Y, U, V
//  
//  c12:
//		Array[16] Wave Table XYZ
//Linear Fog
//	c28: Fog Plane
//  c29: Fog Values
//Graph Fog
//	c28: Fog Plane (Scaled for Spline Enteries)
//	c29: Fog *, *, Dist Clamp Min, Dist Clamp Max
//	c29: 
//		Array[2] 
//			Dist Scale, Dist Offset, *, *
//	c31: 
//		Array[2] 
//			Spline Coeffs XYZW
//Dynamic Lights
//	c33: Dynamic LightX transform
//	c34: Dynamic LightY transform
//	c35: Dynamic LightZ transform
//
struct CarpetData
{
	float4x4 m_xViewProj;
	
	float3 m_xBillboardRight;
	float3 m_xBillboardUp;
	
	float4 m_xCameraPosition;
	float4 m_xConstants0;
	
	float4 m_axCorners[4];
	
	float4 m_axWave[16];
	
//	float4 m_xFog0;
//	float4 m_xFog1;
//	float4 m_xFog2;
	
//	float4 m_FogSplineCoeffs[2];
	
//	float4 m_xDynamicLightTransform[3];
};

//Billboard Styles
//  c36: 
//		Array[Dynamic] Billboard Styles
//			Size Width, Height
//			UV	TL, TR, BL, BR
struct CarpetStyle
{
	float4 m_xSize;
	float4 m_xUV;
};

// Per-style movement scale


//------------------------------------
CarpetData  g_xCarpetData : register( c0);
float4		g_f4MovementScale : register( c35);
CarpetStyle g_xCarpetStyles[100] : register( c36);


sampler g_sAlbedoTexture : register( s0) =
{
	sampler_state
	{
    	AddressU	= WRAP;
    	AddressV	= WRAP;
    	AddressW	= CLAMP;
    	MIPFILTER	= LINEAR;
    	MINFILTER	= LINEAR;
    	MAGFILTER	= LINEAR;
		sRGBTexture	= <g_bUseGammaCorrection>;
	}

};


//------------------------------------
VS_Output VS_Carpet(VS_Input xIn)
{
	VS_Output xOut;
	
	// Get the Corner
	int uCornerIndex = 0x100 * xIn.m_xData2.z;
	float4 xCorner = g_xCarpetData.m_axCorners[ uCornerIndex];
	
	// Get the Style
	int uStyleIndex = 0x100 * xIn.m_xData2.x;
	CarpetStyle xStyle = g_xCarpetStyles[ uStyleIndex];
	
	// Get the Wave
	int uWaveIndex = 0x100 * xIn.m_xData2.y;
	float4 xWave = g_xCarpetData.m_axWave[ uWaveIndex];
	
	// Setup the Fade
	float fFadeDistance = dot( xIn.m_xPosition.xyzw, g_xCarpetData.m_xCameraPosition.xyzw);
	float fFade = (fFadeDistance * xStyle.m_xSize.w) + xStyle.m_xSize.z;
	fFade = min( fFade, 1.0f);
	
	// Setup the Instance Scales
	xCorner.xy *= xIn.m_xData2.w;
	// Setup the Style Scales
	xCorner.xy *= xStyle.m_xSize.xy;
	// Setup the fade scales. Made the billboard degenerate if it is >= 100% faded
	xCorner.xy *= (0.0 < fFade) ? 1 : 0;
	
	// Setup the Position
	float3 xOffsetRight = g_xCarpetData.m_xBillboardRight * xCorner.x;
	float3 xOffsetUp = g_xCarpetData.m_xBillboardUp * xCorner.y;
	float3 xOffsetWave = xCorner.y * xWave.xyz;
	
	float4 xWorldPosition = xIn.m_xPosition;
	xWorldPosition.xyz += xOffsetRight;
	xWorldPosition.xyz += xOffsetUp;
	xWorldPosition.xyz += xOffsetWave * g_f4MovementScale.x;
	
	xOut.m_xPos = xWorldPosition;
	xOut.m_xPosition = mul( xWorldPosition, g_xCarpetData.m_xViewProj);
	
	// Setup the Textures
	xOut.m_xUV = (xStyle.m_xUV.zw * xCorner.zw) + xStyle.m_xUV.xy;

	// Setup Colour
	xOut.m_xColour.rgb = 2.0 * xIn.m_xData1.xyz;
	xOut.m_xColour.a = (-0.5 * fFade) + 1.0;
	
	// Setup Camera Information
	//xOut.m_xNormalDistance.xyz = normalize( g_xCarpetData.m_xCameraPosition.xyz - xWorldPosition);
	xOut.m_xNormalDistance.xyz = normalize( cross( g_xCarpetData.m_xBillboardRight.xyz, g_xCarpetData.m_xBillboardUp.xyz));
	xOut.m_xNormalDistance.w = xOut.m_xPosition.w;

	return xOut;
}

//------------------------------------
DeferredOutputPixel PS_Carpet(VS_Output xIn)
{
	half4 xColour;
	half4 xTexture = tex2D( g_sAlbedoTexture, xIn.m_xUV);
	
	xColour.rgb = xTexture * xIn.m_xColour.rgb;
	xColour.a = xTexture.a - xIn.m_xColour.a;	

	float3 xNormal = normalize( xIn.m_xNormalDistance.xyz);
	DeferredOutputPixel xOut = Deferred_FillOutputPixel(
										xColour,
										normalize(float3( 0.0, -1.0, 0.0 ) +  normalize(g_xCameraPosition.xyz - xIn.m_xPos)),
										xIn.m_xNormalDistance.w,
										xTexture,
										0.0,
										0.0
										);										
	return( xOut);
}



//-----------------------------------
technique Carpet
{
    pass p0
    {
		VertexShader	= compile vs_3_0 VS_Carpet();
		PixelShader		= compile ps_3_0 PS_Carpet();
    }
}
 