
/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

const float  g_fTwoOverScreenWidth  = 0;
const float  g_fTwoOverScreenHeight	= 0;
const float  g_fTimer = 0.0f;

const float4 g_xUVScroll;
const float4 g_xUVScroll2;
const float4 g_xWaterFogColour;
const float4 g_xWaterReflectionColour;
const float4 g_xWaterRefractionColour;
const float  g_fWaterFogScale;
const float  g_fVolumetricDepthScale;
const float  g_fRippleScale;
const float  g_fDistortionScale;

const float4 g_xWavesX;
const float4 g_xWavesZ;

const texture g_xDepthTexture;
const sampler g_xDepthTextureSampler = sampler_state
{
	Texture = <g_xDepthTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = POINT;
	MagFilter = POINT;
};

const texture g_xBackBufferTexture;
const sampler g_xBackBufferTextureSampler = sampler_state
{
	Texture = <g_xBackBufferTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = POINT;
	MagFilter = POINT;
};


const texture g_xReflectionTexture;
const sampler g_xReflectionTextureSampler = sampler_state
{
	Texture = <g_xReflectionTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
};


/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 m_xPosition : POSITION;
	float3 m_xNormal   : NORMAL;
	float4 m_xTangent  : TANGENT;
	float2 m_axUV[2]   : TEXCOORD0;

	float4 m_xRedSH    : TEXCOORD2;
	float4 m_xGreenSH  : TEXCOORD3;
	float4 m_xBlueSH   : TEXCOORD4;
};

struct OutputVertex
{
	float4 m_xPosition       : POSITION;
	float4 m_axUVs           : TEXCOORD0; // UV pairs in (x,y) & (z,w)
	float4 m_xNormalDistance : TEXCOORD1;
	float3 m_xTangent		 : TEXCOORD2;
	float3 m_xBinormal		 : TEXCOORD3;
	float3 m_xWorldPos		 : TEXCOORD4;
	float3 m_xEyeDistance	 : TEXCOORD5;

	float4 m_xRedSH          : TEXCOORD6;
	float4 m_xGreenSH        : TEXCOORD7;
	float4 m_xBlueSH         : TEXCOORD8;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	OutputVertex xOut;

	float4 xPos = xIn.m_xPosition;
	xPos.y += sin(g_xWavesX.z + xIn.m_xPosition.x * g_xWavesX.y)*g_xWavesX.x;
	xPos.y += sin(0.7f * (g_xWavesX.z + xIn.m_xPosition.x * g_xWavesX.y))*g_xWavesX.x;

	xPos.y += sin(g_xWavesZ.z + xIn.m_xPosition.z * g_xWavesZ.y)*g_xWavesZ.x;
	xPos.y += sin(0.7f * (g_xWavesZ.z + xIn.m_xPosition.z * g_xWavesZ.y))*g_xWavesZ.x;

	xOut.m_xPosition = mul( xPos, g_xViewProjectionMatrix );

	// Pack texcoords to save an interpolator
	xOut.m_axUVs.xy = xIn.m_axUV[0];
	xOut.m_axUVs.zw = xIn.m_axUV[1];

	// Bundle these together to save an interpolator
	xOut.m_xNormalDistance.xyz = xIn.m_xNormal;
	xOut.m_xNormalDistance.w   = xOut.m_xPosition.w;

	xOut.m_xTangent  = xIn.m_xTangent.xyz;
	xOut.m_xBinormal = normalize(cross(xIn.m_xTangent, xIn.m_xNormal)) * xIn.m_xTangent.w;
	xOut.m_xWorldPos = xIn.m_xPosition.xyz;

	xOut.m_xRedSH   = xIn.m_xRedSH;
	xOut.m_xGreenSH = xIn.m_xGreenSH;
	xOut.m_xBlueSH  = xIn.m_xBlueSH;

	// Calculate the eye direction in tangent-space...
	float3		xEyeDistance	= g_xCameraPosition - xIn.m_xPosition;
	xOut.m_xEyeDistance	= RotateIntoTangentSpace (xEyeDistance, xIn.m_xTangent, xOut.m_xBinormal, xIn.m_xNormal);

	return xOut;
}

float4 PShader( const OutputVertex xIn, const float2 xScreenPos : VPOS, float fFace: VFACE ) : COLOR
{
	const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * float2(g_fTwoOverScreenWidth, g_fTwoOverScreenHeight) / 2;
	float4 xDepth = tex2D(g_xDepthTextureSampler, xNormalizedCoords);
	float fDepth = xDepth.z;
	if (fDepth<0) fDepth=-fDepth;

	float4 xBackBuffer = tex2D(g_xBackBufferTextureSampler, xNormalizedCoords);
	float fHeight = xBackBuffer.a;
	float2 xOfs = normalize(xIn.m_xEyeDistance) * 0.075f * fHeight;

	float4 xNormal  = tex2D( g_xNormalMapSampler, g_xUVScroll.z * (xIn.m_axUVs.xy + g_xUVScroll.xy + xOfs)).agaa;
	xNormal.xyz -= 0.5f;

	if (g_bDualNormalMap)
	{
		float4 xNormal2 = tex2D( g_xNormalMap2Sampler, g_xUVScroll2.z * (xIn.m_axUVs.xy + g_xUVScroll2.xy + xOfs) ).agaa;
		xNormal = xNormal+xNormal2;
		xNormal.xyz -= 0.5f;
	}

	// Our normal maps are output upside-down. TODO: do this in a preprocessing step
	xNormal.xy *= 2.0f;
	xNormal.y = -xNormal.y;
	xNormal.z = sqrt( 1.0h - dot( xNormal.xy, xNormal.xy ) );
	xNormal.w = 1.0h;
	xNormal.xyz = normalize(xNormal.xyz);

	float3 xNormalWorld;
	xNormalWorld.xyz  = xIn.m_xTangent.xyz * xNormal.x;
	xNormalWorld.xyz += xIn.m_xBinormal.xyz * xNormal.y;
	xNormalWorld.xyz += xIn.m_xNormalDistance.xyz * xNormal.z;

	float3 xNormalWorldUnflattened = xNormalWorld;

	// 'Flatten' the normals based on the ripplescale value. If we've got dynamic ripples, lerp out the flattening
	// -----------------------------------------------------------------------------------------------------------
	xNormalWorld.xyz = normalize(lerp(xIn.m_xNormalDistance.xyz, xNormalWorld.xyz, lerp(g_fRippleScale,1,fHeight) ));

	float3 xViewDir			 = normalize(g_xCameraPosition.xyz - xIn.m_xWorldPos.xyz);
	float3 xReflectionVector = (2*xNormalWorld.xyz)-xViewDir;
	float4 xReflection		 = tex2D( g_xReflectionTextureSampler, (xReflectionVector.xz * 0.5f) + 0.5f);
	float4 xColour;

	// Do the volumetric alpha calculation
	// -----------------------------------------------------------------------------------------------------------
	float fSoftValue   = (fDepth-xIn.m_xNormalDistance.w);
	float fDistortion  = min(abs(fSoftValue*0.01f),1.0f) * g_fDistortionScale;
	fSoftValue         = saturate(fSoftValue*g_fVolumetricDepthScale);
	xColour.a          = fSoftValue;

	if (g_bUseTransparencyMap)
	{
		xColour.a *= tex2D(g_xTransparencyMapSampler, xIn.m_axUVs.xy).a;
	}

	// Generate the offset/distorted UVs for the fake refraction (distorted version of the image below)
	// -----------------------------------------------------------------------------------------------------------
	float2 xRefractedCoords = xNormalizedCoords+ ((fDistortion/(1.0+xIn.m_xNormalDistance.w))*xNormalWorldUnflattened.xz);

	// Re-read the depth from the refracted pos
	// -----------------------------------------------------------------------------------------------------------
	float fOldDepth = fDepth;
	xDepth = tex2D(g_xDepthTextureSampler, xRefractedCoords);
	fDepth = xDepth.z;
	if (fDepth<0) fDepth=-fDepth;

	// If we're sampling an object closer than the water, reject this sample, and use the non-wibbled sample
	// -----------------------------------------------------------------------------------------------------------
	if (fDepth < xIn.m_xNormalDistance.w)
	{
		xColour.rgb = tex2D(g_xBackBufferTextureSampler, xNormalizedCoords);
		fDepth      = fOldDepth;
	}
	else
	{
		xColour.rgb = tex2D(g_xBackBufferTextureSampler, xRefractedCoords);
	}

	// Lerp between reflection and refraction based on viewing angle
	// -----------------------------------------------------------------------------------------------------------
	if (fFace<0)
	{
		// Above water

		// Fog the refracted colour
		// -----------------------------------------------------------------------------------------------------------
		float fFogAmount = min((fDepth-xIn.m_xNormalDistance.w)*g_fWaterFogScale,1.0f);
		xColour.rgb      = lerp(xColour.rgb, g_xWaterFogColour.rgb, fFogAmount);

		// Fresnel-ish lerp between reflection and refraction, based on view angle
		// -----------------------------------------------------------------------------------------------------------
		// A few magic numbers... values that 'looked about right'. Might be better off as params in the material, but
		// water has tons of parameters already

		float fLerp  = 0.3f + 0.7f * pow(1.0f-dot(xViewDir, xIn.m_xNormalDistance.xyz),3);	
		xColour.rgb = 2.0f * lerp (xColour.rgb * g_xWaterRefractionColour, xReflection.rgb * g_xWaterReflectionColour, fLerp);
	}
	else
	{
		// Underwater
		xColour.rgb = xColour.rgb;
	}

	// Shade dynamic ripples
	xColour = xColour * (1.0+fHeight*0.2);

	// Multiply by prelighting, so areas under piers etc. are darker
	{
		const float4 xNormalWorld4 = { xNormalWorld, 1.0f };

		float3 xSHLookup;
		xSHLookup.r = mul( xNormalWorld4, xIn.m_xRedSH );
		xSHLookup.g = mul( xNormalWorld4, xIn.m_xGreenSH );
		xSHLookup.b = mul( xNormalWorld4, xIn.m_xBlueSH );

		xColour.rgb *= xSHLookup;
	}

	xColour = ApplyFog(xColour, xIn.m_xNormalDistance.w);

	return xColour;
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader();
	}
}
