/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#define DEFINE_SAMPLER( Name, Filter )           \
	texture g_x##Name##Texture;                  \
	sampler2D g_x##Name##Sampler = sampler_state \
	{                                            \
		Texture = <g_x##Name##Texture>;          \
		AddressU = CLAMP;                        \
		AddressV = CLAMP;                        \
		MinFilter = Filter;                      \
		MagFilter = Filter;                      \
		MipFilter = NONE;                        \
	};

// Copy of the scene to be downsized
DEFINE_SAMPLER( Downsize, POINT );

// Single texel holding the current adapted luminance
DEFINE_SAMPLER( CurrentLuminance, POINT );

// Single texel holding the scene's luminance, our target value over time
DEFINE_SAMPLER( SceneLuminance, POINT );

// Current HDR scene to apply tone mapping on
DEFINE_SAMPLER( InputScene, POINT );

#undef DEFINE_SAMPLER

float2 g_xTexCoordScale;
float2 g_xViewportScale;

float g_fAdaptationFactor;  // Adaption rate * time elapsed this frame
float g_fMaxLuminance;      // Maximum luminance to adapt to
float g_fMinLuminance;      // Minimum luminance to adapt to
float g_fKeyValue;          // Exposure setting for the scene
float g_fWhitePointSquared; // World luminances beyond this point are whited out

// (white pt / (white pt - black pt), -black pt * white pt / (white pt - black pt)
float2 g_xScaleParams;

// Allow the user to specify a fixed luminance
bool g_bDynamicExposure = true;
float g_fFixedLuminance;

float g_fHistogramCopyScale;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

// Vertex structure used in second pass
struct PosUVVert
{
	float4 m_xPos : POSITION;
	float4 m_xUV  : TEXCOORD;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Luminance = 0.27r + 0.67g + 0.06b
static float3 xLUMINANCE_CONSTANTS = float3( 0.27f, 0.67f, 0.06f );

// Tex coord offsets for 4x4 downsampling
static float2 axDOWNSAMPLE_4X4_OFFSETS[16] =
{
 	float2( 0.5f, 0.5f ), float2( 0.5f, 1.5f ), float2( 0.5f, 2.5f ), float2( 0.5f, 3.5f ),
 	float2( 1.5f, 0.5f ), float2( 1.5f, 1.5f ), float2( 1.5f, 2.5f ), float2( 1.5f, 3.5f ),
 	float2( 2.5f, 0.5f ), float2( 2.5f, 1.5f ), float2( 2.5f, 2.5f ), float2( 2.5f, 3.5f ),
 	float2( 3.5f, 0.5f ), float2( 3.5f, 1.5f ), float2( 3.5f, 2.5f ), float2( 3.5f, 3.5f ),

// 	float2( -1.5f, -1.5f ), float2( -0.5f, -1.5f ), float2( 0.5f, -1.5f ), float2( 1.5f, -1.5f ),
// 	float2( -1.5f, -0.5f ), float2( -0.5f, -0.5f ), float2( 0.5f, -0.5f ), float2( 1.5f, -0.5f ),
// 	float2( -1.5f, +0.5f ), float2( -0.5f, +0.5f ), float2( 0.5f, +0.5f ), float2( 1.5f, +0.5f ),
// 	float2( -1.5f, +1.5f ), float2( -0.5f, +1.5f ), float2( 0.5f, +1.5f ), float2( 1.5f, +1.5f ),
};

// Tex coord offsets for 3x3 downsampling
const float2 axDOWNSAMPLE_3X3_OFFSETS[9] =
{
 	float2( 0.5f, 0.5f ), float2( 0.5f, 1.5f ), float2( 0.5f, 2.5f ), 
 	float2( 1.5f, 0.5f ), float2( 1.5f, 1.5f ), float2( 1.5f, 2.5f ), 
 	float2( 2.5f, 0.5f ), float2( 2.5f, 1.5f ), float2( 2.5f, 2.5f ), 

// 	float2( -1.0f, -1.0f ), float2( 0.0f, -1.0f ), float2( 1.0f, -1.0f ),
// 	float2( -1.0f, +0.0f ), float2( 0.0f, +0.0f ), float2( 1.0f, +0.0f ),
// 	float2( -1.0f, +1.0f ), float2( 0.0f, +1.0f ), float2( 1.0f, +1.0f ),
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

float4 VShader( const float4 xIn : POSITION ) : POSITION
{
	return xIn;
}

float2 CalcUVFromVPOS(const float2 xPos)
{
	return (xPos) * g_xViewportScale;
}

float4 Average16_Point( const float2 xUV )
{
	// Can't use filtering on float textures. Manually average each block
	float4 xSum = float4(0,0,0,0);

	for( int i = 0; i < 16; ++i )
	{
		float2 xModifiedUV = xUV + axDOWNSAMPLE_4X4_OFFSETS[i] * g_xTexCoordScale;
		xSum += tex2D( g_xDownsizeSampler, xModifiedUV );
	}

	return xSum / 16.0f;
}


float Average9Log_Point( const float2 xUV )
{
	// Can't use filtering on float textures. Manually average each block
	float fSum = 0.0;
	// Add a tiny amount to avoid log(0) on pure black pixels
	const float fDELTA = 0.0001f;

	for( int i = 0; i < 9; ++i )
	{
		float2 xModifiedUV = xUV + axDOWNSAMPLE_3X3_OFFSETS[i] * g_xTexCoordScale;
		float fLuminance = dot( tex2D( g_xDownsizeSampler, xModifiedUV ).rgb, xLUMINANCE_CONSTANTS );


		fSum += log( abs(fLuminance) + fDELTA );
	}

	return fSum / 9.0f;
}



float4 PShader_Downsample( float2 xScreenPos : VPOS ) : COLOR
{
	return Average16_Point( CalcUVFromVPOS(xScreenPos));
}

PosUVVert VShader_CalcLogLuminance( const PosUVVert xIn )
{
	return xIn;
}

float4 PShader_CalcLogLuminance( const float2 xUV : TEXCOORD ) : COLOR
{
	// We should be squishing the scene into a smaller RT at this point (for 720p, 320x180 -> 64x64)

	// Find the average natural log
	return Average9Log_Point(xUV);
}

float4 PShader_CalcFinalLuminance() : COLOR
{
	// We should now be rendering into a 1x1 RT, from a 4x4 source. Take one final average and then
	// undo the log

	const float2 xUV = float2( 0.0f, 0.0f );

	// NB: exp() HLSL intrinsic operates on every component, but assembly instruction is scalar.
	// Leaving off the .r here dramatically increases the register usage and number of ALU ops
	const float fAvg = Average16_Point( xUV ).r;

	return exp( fAvg );
}

float4 PShader_AdaptLuminance(uniform bool bIgnoreLerp) : COLOR
{
	const float2 xCurrentLumUV = float2( 0.5f, 0.5f );
	const float fCurrentLuminance = tex2D( g_xCurrentLuminanceSampler, xCurrentLumUV ).r;

	// Sample the top-left texel of the 16x16 working texture
	const float2 xSceneLumUV = float2( 0.5f / 16.0f, 0.5f / 16.0f );
	const float fSceneLuminance = tex2D( g_xSceneLuminanceSampler, xSceneLumUV ).r;


	const float fAdaptedLuminance = bIgnoreLerp ? fSceneLuminance : lerp( fCurrentLuminance, fSceneLuminance, g_fAdaptationFactor );
	return clamp( fAdaptedLuminance, g_fMinLuminance, g_fMaxLuminance );
}

float4 PShader_ApplyToneMapping( float2 xScreenPos : VPOS ) : COLOR
{
	float4 xInputSample = tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 0.5f, 0.5f )) );

	// Scale to account for "black point"
	xInputSample.rgb = max( xInputSample.rgb * g_xScaleParams.x + g_xScaleParams.y, 0.0f );

	const float2 xUV = float2( 0.5f, 0.5f );

	const float fAdaptedLuminance = g_bDynamicExposure
								  ? tex2D( g_xCurrentLuminanceSampler, xUV ).r
								  : g_fFixedLuminance;

	const float fHDRLuminance = dot( xInputSample.rgb, xLUMINANCE_CONSTANTS );

	// Apply a slight delta to prevent divide-by-zero
	const float fDELTA = 0.0001f;
	const float fScaledLuminance = g_fKeyValue / (fAdaptedLuminance + fDELTA) * fHDRLuminance;

	const float fNumerator = fScaledLuminance * (1.0f + fScaledLuminance / g_fWhitePointSquared);
	const float fDenominator = 1.0f + fScaledLuminance;

	const float fLDRLuminance = fNumerator / fDenominator;

	// Rescale the colours by the ratio of new to old luminance. Slight delta to avoid divide by 0
	xInputSample.rgb = xInputSample.rgb * fLDRLuminance / max( fHDRLuminance, fDELTA );

	return xInputSample;
}
float4 PShader_Debug( float2 xScreenPos : VPOS ) : COLOR
{
	return tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 0.5f, 0.5f )));
}

float4 PShader_Histogram_copy( float2 xScreenPos : VPOS ) : COLOR
{
	float4 xColour = tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 0.5f, 0.5f )));
	xColour += tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 1.0f, 0.5f )));
	xColour += tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 0.5f, 1.0f )));
	xColour += tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 1.0f, 1.0f )));
	xColour *= 0.25;
	xColour.a = dot(xColour.rgb,xLUMINANCE_CONSTANTS.rgb);
	
	return saturate(xColour * g_fHistogramCopyScale);
}

float4 PShader_Histogram_ToneMap_copy( float2 xScreenPos : VPOS ) : COLOR
{
	float4 xInputSample = tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 0.5f, 0.5f )));
	xInputSample += tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 1.0f, 0.5f )));
	xInputSample += tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 0.5f, 1.0f )));
	xInputSample += tex2D( g_xInputSceneSampler, CalcUVFromVPOS( xScreenPos + float2( 1.0f, 1.1f )));
	xInputSample *= 0.25;

	// Scale to account for "black point"
	xInputSample.rgb = max( xInputSample.rgb * g_xScaleParams.x + g_xScaleParams.y, 0.0f );

	const float2 xUV = float2( 0.5f, 0.5f );

	const float fAdaptedLuminance = g_bDynamicExposure
								  ? tex2D( g_xCurrentLuminanceSampler, xUV ).r
								  : g_fFixedLuminance;

	const float fHDRLuminance = dot( xInputSample.rgb, xLUMINANCE_CONSTANTS );

	// Apply a slight delta to prevent divide-by-zero
	const float fDELTA = 0.0001f;
	const float fScaledLuminance = g_fKeyValue / (fAdaptedLuminance + fDELTA) * fHDRLuminance;

	const float fNumerator = fScaledLuminance * (1.0f + fScaledLuminance / g_fWhitePointSquared);
	const float fDenominator = 1.0f + fScaledLuminance;

	const float fLDRLuminance = fNumerator / fDenominator;

	// Rescale the colours by the ratio of new to old luminance. Slight delta to avoid divide by 0
	xInputSample.rgb = xInputSample.rgb * fLDRLuminance / max( fHDRLuminance, fDELTA );

	xInputSample.a = dot(xInputSample.rgb,xLUMINANCE_CONSTANTS.rgb);
	
	return saturate(xInputSample * g_fHistogramCopyScale);
}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass // 0
	{
		// First pass downsamples scene to quarter size
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Downsample();
	}

	pass  // 1
	{
		// Second pass calculates log luminance averages, writing into a 64x64
		// R32F target
		VertexShader = compile vs_3_0 VShader_CalcLogLuminance();
		PixelShader = compile ps_3_0 PShader_CalcLogLuminance();
	}

	pass  // 2
	{
		// Then need to downsample that R32F source down to 4x4. This pass is
		// executed twice (64x64 -> 16x16 -> 4x4)
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Downsample();
	}

	pass  // 3
	{
		// Final pass downsamples from 4x4 to 1x1 and undoes the log
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_CalcFinalLuminance();
	}

	pass  // 4
	{
		// This pass adapts the eye luminance towards the target
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_AdaptLuminance(false);
	}
	pass  // 5
	{
		// This pass adapts the eye luminance towards the target by just setting it to a new value (used post reset)
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_AdaptLuminance(true);
	}

	pass  // 6
	{
		// Finally we apply tone mapping using the current light adaptation
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_ApplyToneMapping();
	}
	pass  // 7
	{
		// Debug shader
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Debug();
	}
	pass  // 8
	{
		// Used for producing a histogram in tools
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Histogram_copy();
	}
	pass  // 9
	{
		// Used for producing a histogram in tools
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Histogram_ToneMap_copy();
	}
}