/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"
#include "Fog.fxh"


const texture g_xTexture;
const sampler g_xTextureSampler = sampler_state
{
	Texture = <g_xTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	sRGBTexture = <g_bUseGammaCorrection>;
};

const texture g_xDepthTexture;
const sampler g_xDepthTextureSampler = sampler_state
{
	Texture = <g_xDepthTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = POINT;		// bilinear doesn't work on FP16 textures? getting weird glitches
	MagFilter = POINT;
};

const bool   g_bUseLighting;
const bool   g_bCrossfadeTextureFrames		= true;
const bool   g_bZReadEnable					= false;
const bool   g_bVolumetric					= true;

const float  g_fTwoOverScreenWidth		    = 0;
const float  g_fTwoOverScreenHeight			= 0;

const float  g_fFirstInstance				= 0;
const float  g_fVolumetricDepthScale		= 0.0f;

const float4 g_xXDelta						= float4(1,0,0,0);
const float4 g_xYDelta						= float4(0,0,1,0);
const float4 g_xLightColour					= float4(1,1,1,1);
const float4 g_xPosOffsetAndSizeScale	    = float4(0,0,0,1);
const float4 g_xNearFarFade;



/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	// Vertex Data (Stream 0)
	float  fVertexIndex			 : TEXCOORD4;

	// Instance Data (Stream 1)
	float4 xInstancePosition	 : POSITION0;
	float4 xDiffuse				 : COLOR0;
	float4 xUVSet1				 : TEXCOORD0;	
	float4 xUVSet2				 : TEXCOORD1;	
	float2 xSize				 : TEXCOORD2;	
	float2 xParams				 : TEXCOORD3;

};

struct OutputVertex
{
	float4 Position				 : POSITION;
	float4 Diffuse				 : COLOR;
	float2 UV1					 : TEXCOORD0;
	float2 UV2					 : TEXCOORD1;
	float4 UVDepth				 : TEXCOORD2;		// Unused when using the non-soft PS
	float3 xParams				 : TEXCOORD4;		// .x = fDepth, .y = fSize, .z = fCrossFadeBlendWeight
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/




OutputVertex VShader( const InputVertex xIn )
{
	// The only per-vert input is a vertex index for the particle, used to 
	// generate a vertex using some little lookup tables:
		
	const float4 axVerts[4]=	   {	{ -1.0, -1.0  ,0.0,  1.0},
										{  1.0, -1.0  ,0.0,  1.0},
										{  1.0,  1.0  ,0.0,  1.0},
										{ -1.0,  1.0  ,0.0,  1.0}};

	// UV generation is a bit more tricky, as we've got to 4 possible texture orientations (90 degree rotations)
	// Still cheap enough to do in the shader - it's just a lookup and dot product
	
	const float4 axUSelector[16] = {    {1,0,0,0}, {0,0,1,0}, {0,0,1,0}, {1,0,0,0},
										{1,0,0,0}, {0,0,1,0}, {0,0,1,0}, {1,0,0,0},
										{0,0,1,0}, {1,0,0,0}, {1,0,0,0}, {0,0,1,0},
										{0,0,1,0}, {1,0,0,0}, {1,0,0,0}, {0,0,1,0}};
										
	const float4 axVSelector[16] = {    {0,1,0,0}, {0,1,0,0}, {0,0,0,1}, {0,0,0,1},
										{0,0,0,1}, {0,0,0,1}, {0,1,0,0}, {0,1,0,0},
										{0,0,0,1}, {0,0,0,1}, {0,1,0,0}, {0,1,0,0},
										{0,1,0,0}, {0,1,0,0}, {0,0,0,1}, {0,0,0,1}};								 
				 
	
	// This is all we get from the vertex data stream - (in the order 0,1,3,3,2,1 - as we're using tri lists
	int   iVertexIndex = xIn.fVertexIndex;	
		
	float4 xInstancePosition = xIn.xInstancePosition;
	float4 xDiffuse		 = xIn.xDiffuse;
	float2 xParams		 = xIn.xParams;
	float2 xSize		 = xIn.xSize;
	float4 xUVSet1		 = xIn.xUVSet1;	
	float4 xUVSet2		 = xIn.xUVSet2;

	
	float fRotation = xInstancePosition.w;	
	
	// Handle global size scale
	xSize = xSize * g_xPosOffsetAndSizeScale.w;
	
	OutputVertex xOut;
	
	// Colour and Misc params
	// ---------------------------------------------
	xOut.xParams.z = xParams.y;										// CrossFadeBlendWeight
	xOut.xParams.y = xSize.x*2.0f * g_fVolumetricDepthScale;		// Size
	xOut.Diffuse = xDiffuse;
	xOut.Diffuse.rgb *= g_xLightColour.rgb;


	// Handle the near/far fade (and culling when
	// alpha=0 by setting the size to 0)
	// ---------------------------------------------
		
	float4 xPos;
	xPos.xyz = xInstancePosition.xyz + g_xPosOffsetAndSizeScale.xyz;
	xPos.w   = 1.0f;
				
	float4 x2 = mul( xPos , g_xViewProjectionMatrix);				// <- OPTIMIZE THIS...
	float fDist = x2.w;
		
	float fNearFade = min(1,max(0,(fDist - g_xNearFarFade.y) / (g_xNearFarFade.x - g_xNearFarFade.y)));
	float fFarFade  = min(1,max(0,(fDist - g_xNearFarFade.w) / (g_xNearFarFade.z - g_xNearFarFade.w)));
	xOut.Diffuse.a *= min(fNearFade,fFarFade);
	
	if (xOut.Diffuse.a<=0)
	{
		// Cull it if alpha is 0, by degenerating it to a point. 
		xSize.xy=float2(0,0);
	}
		
	// Get the base vertex			
	// ---------------------------------------------
	float4 xTmp = axVerts[iVertexIndex];
	
	// Scale it	
	// ---------------------------------------------
	xTmp.xy = xTmp.xy * xSize.xy;
	
	// Rotate it
	// ---------------------------------------------
	float2 xRot0,xRot1,xRotated;
	sincos(-fRotation, xRot0.y, xRot0.x);        
	xRot1.x= xRot0.y;							
	xRot1.y=-xRot0.x;	
	xRotated.x = dot(xRot0, xTmp.xy);
	xRotated.y =-dot(xRot1, xTmp.xy);
	
	// Align it
	// ---------------------------------------------
	xTmp.xyz = xRotated.x*g_xXDelta.xyz + xRotated.y*g_xYDelta.xyz;	

	// Position it
	// ---------------------------------------------
	xTmp.xyz = xTmp.xyz + xInstancePosition.xyz + g_xPosOffsetAndSizeScale.xyz;
		
	// Xform to screen
	// ---------------------------------------------
	xOut.Position = mul( xTmp , g_xViewProjectionMatrix );
	
	// Do the UVs
	// ---------------------------------------------
	int  iOrient = xParams.x;	
	int  iTexOrientIndex = iOrient*4 + iVertexIndex;
	xOut.UV1.x = dot(xUVSet1 , axUSelector[iTexOrientIndex]);
	xOut.UV1.y = dot(xUVSet1 , axVSelector[iTexOrientIndex]);
	xOut.UV2.x = dot(xUVSet2 , axUSelector[iTexOrientIndex]);
	xOut.UV2.y = dot(xUVSet2 , axVSelector[iTexOrientIndex]);
	
	// For soft particles - UVs to read from the depth texture
	// ---------------------------------------------
	xOut.UVDepth.x = xOut.Position.x / xOut.Position.w;
	xOut.UVDepth.y = -xOut.Position.y / xOut.Position.w;
	xOut.UVDepth.z = 0;
	xOut.UVDepth.w = 1;
	
	xOut.xParams.x = xOut.Position.w;
		
	return xOut;
}




float4 PShader_Volumetric( const OutputVertex xIn, const float2 xScreenPos : VPOS ) : COLOR
{
	const float4 xTexture[2] = { tex2D( g_xTextureSampler, xIn.UV1 ), tex2D( g_xTextureSampler, xIn.UV2 ) };
	float4 xColour;    
	
	if( g_bCrossfadeTextureFrames )
	{
		xColour = lerp( xTexture[1], xTexture[0], xIn.xParams.z ) * xIn.Diffuse;
	}
	else
	{
		xColour = xTexture[0] * xIn.Diffuse;
	}	
	
	if( g_bUseLighting )
	{
		// Now supports crossfading with lighting. Dunno why it didn't before?
		xColour.rgb *= 2.0f;		
	}
	
	if (g_bVolumetric)
	{
		// Do the soft particle stuff
		const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * float2(g_fTwoOverScreenWidth, g_fTwoOverScreenHeight) / 2;
		const float4 xDepth = tex2D(g_xDepthTextureSampler, xNormalizedCoords);
		
		float fDepth = xDepth.z;
		if (fDepth<0) fDepth=-fDepth;
		
	    float fSoftValue = min(abs(fDepth-xIn.xParams.x)/xIn.xParams.y,1.0f);
		xColour.a *= fSoftValue;	
	}
	
	if (g_bFogEnabled)
	{
		return ApplyFog(xColour, xIn.xParams.x);
	}
	else
	{
		return xColour;
	}	
}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique Volumetric
{
	pass
	{
		ZEnable = <g_bZReadEnable>;
		ZFunc = LESSEQUAL;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Volumetric();
	}
}