/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Ensure we strip out all the effect stuff
#define NOT_EFFECT

#include "MaterialResponse_Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

float4x4 g_xWorld             : register(c2);
float4x4 g_xWorldViewProj     : register(c6);
float3   g_xCamPos            : register(c10);
row_major float3x4 g_axBoneMatrices[80] : register(c11);

int g_iNumBoneWeightings  : register(i0);

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 m_xPosition        : POSITION;
	float4 m_xBlendIndices    : BLENDINDICES;
	float4 m_xBlendWeights    : BLENDWEIGHT;

	float3 m_xTangent         : TANGENT;
	float3 m_xBinormal        : BINORMAL;
	float3 m_xNormal          : NORMAL;

	float4 m_xUVPair          : TEXCOORD0;
	float  m_fDualTextureLerp : TEXCOORD1;
};

struct OutputVertex
{
	float4 m_xPosition                : POSITION;

	float3 m_xTangent                 : TANGENT;
	float3 m_xBinormal                : BINORMAL;
	float3 m_xNormal                  : NORMAL;

	float4 m_xUVPair                  : TEXCOORD0;

	// World-space camera dir in xyz, view-space depth in w
	float4 m_xWorldSpaceCamDir_Dist   : TEXCOORD1;

	// Tangent-space camera dir in xyz, dual-texture lerp factor in w
	float4 m_xTangentSpaceCamDir_Lerp : TEXCOORD2;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

float3x4 FetchBone( const int fIndex, out float fDamage )
{
	float3x4 xMatrix = g_axBoneMatrices[ fIndex ];

	// Per-bone texturing blend factors ("damage values") are packed into one of the matrix
	// elements. We can reconstruct the row (assuming they're orthogonal) with a cross product
	fDamage = xMatrix._33;
	xMatrix[2].xyz = cross( xMatrix[0].xyz, xMatrix[1].xyz );

	return xMatrix;
}

OutputVertex main( const InputVertex xIn )
{
	// Accumulate the blended skinning matrix and damage value

	float3x4 xBoneMat =
	{
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
	};

	float fDamageValue = 0.0f;

	for( int i = 0; i < g_iNumBoneWeightings; ++i )
	{
		const float fWeight = xIn.m_xBlendWeights[i];

		[branch]
		if( fWeight > 0.0f )
		{
			float fDamage;
			const float3x4 xMat = FetchBone( xIn.m_xBlendIndices[i], fDamage );

			xBoneMat += xMat * fWeight;
			fDamageValue += fDamage * fWeight;
		}
	}

	// Apply skinning to the input geometry
	const float4 xPos = float4( mul( xBoneMat, xIn.m_xPosition ).xyz, 1.0 );
	const float3 xTangent = mul( xBoneMat, xIn.m_xTangent );
	const float3 xBinormal = mul( xBoneMat, xIn.m_xBinormal );
	const float3 xNormal = mul( xBoneMat, xIn.m_xNormal );

	OutputVertex xOut;

	xOut.m_xPosition = mul( g_xWorldViewProj, xPos );

	// Transform TBN into world space
	xOut.m_xTangent.xyz = mul( g_xWorld, xTangent );
	xOut.m_xBinormal.xyz = mul( g_xWorld, xBinormal );
	xOut.m_xNormal.xyz = mul( g_xWorld, xNormal );

	xOut.m_xUVPair = MaterialResponse_AnimateUVs( xIn.m_xUVPair );
	xOut.m_xTangentSpaceCamDir_Lerp.w = xIn.m_fDualTextureLerp * fDamageValue;

	// World space direction to camera and depth
	const float3 xWorldSpacePos = mul( g_xWorld, xPos );
	const float3 xWorldSpaceCameraDir = g_xCamPos - xWorldSpacePos;

	xOut.m_xWorldSpaceCamDir_Dist.xyz = xWorldSpaceCameraDir;
	xOut.m_xWorldSpaceCamDir_Dist.w = xOut.m_xPosition.w;

	// Tangent space camera dir, needed for parallax mapping
	const float3x3 xTangentToWorldMatrix = { xOut.m_xTangent, xOut.m_xBinormal, xOut.m_xNormal };

	// Note implicit transpose on the matrix here, by swapping operand order
	xOut.m_xTangentSpaceCamDir_Lerp.xyz = mul( xTangentToWorldMatrix, xWorldSpaceCameraDir );

	return xOut;
}