//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_CSM.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_CSM

#include "Shadows/Asura_CSM.h"

// Asura Platform-independent
#include "Animation/Asura_TextureAnimation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// Asura PC
#include "Core/Asura_PlatformBasics.h"
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Effects_System.h"
#include "Rendering/Asura_Direct3D_EnvironmentRenderer.h"
#include "Rendering/Asura_Direct3D_RenderTargetManager.h"

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const u_int uEFFECT_HASH = 0x000181FD; // "CSM"

const u_int uPASS_CREATE       = 0;
const u_int uPASS_CREATE_TRANS = 1;
const u_int uPASS_BLUR_HORZ    = 2;
const u_int uPASS_BLUR_VERT    = 3;
const u_int uPASS_APPLY        = 4;

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

// Our pixel shader only supports up to 5 maps at the moment, to speed up determining which map to
// use for each pixel. This could be extended if necessary, at the cost of some ALU ops

#if ASURA_CSM_MAX_MAPS > 5
	#error PC CSM only works with up to 5 maps
#endif

static D3DXMATRIX g_axCSMMatrices[ ASURA_CSM_MAX_MAPS ];

DVSVAR float g_fSlopeScaleDepthBias = 0.0f;
DVSVAR float g_fDepthBias = 0.0f;
DVSVAR float g_fPower = 1.0f;
DVSVAR bool  g_bCullStrips = true;
DVSVAR bool  g_bStaggerUpdates = true;

// This is what we render into
static Direct3D_Texture* g_pxRenderTexture = NULL;
static Direct3D_Surface* g_pxRenderSurface = NULL;

// Scratch surface for blurring
static Direct3D_Texture* g_pxBlurTexture = NULL;
static Direct3D_Surface* g_pxBlurSurface = NULL;

// Then copy into here, containing all cascade levels packed vertically
static Direct3D_Texture* g_pxCSMTexture = NULL;
static Direct3D_Surface* g_pxCSMSurface = NULL;

// Use a depth buffer while rendering too, to enable Hi-Z etc.
static Direct3D_Surface* g_pxCSMDS = NULL;

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

static bool UpdateThisFrame( const u_int uMap )
{
	if( !g_bStaggerUpdates ) return true;

	const bool bEvenFrame = ((Asura_Timers::GetFrameNumber() & 1) == 0);

	// This assumes 4 maps, not a problem at the moment
	return bEvenFrame ? (uMap == 0 || uMap == 3) : (uMap == 1 || uMap == 2);
}

////////////////////////////////////////////////////////////////////////////////
//
// CreateViewMatrix()
//
// Creates a D3DXMATRIX suitable for use as a view matrix, given translation and
// rotation components in Asura form. The rotation is transposed in the
// function, but the translation is assumed to already be negated
//
////////////////////////////////////////////////////////////////////////////////

static void CreateViewMatrix( D3DXMATRIX& xOut, const Asura_Vector_3& xAsuraTrans,
	const Asura_Matrix_3x3& xAsuraRot )
{
	D3DXMatrixTranslation( &xOut, xAsuraTrans.x, xAsuraTrans.y, xAsuraTrans.z );

	D3DXMATRIX xRot;
	D3DXMatrixIdentity( &xRot );

	// Fudging like this compensates for Asura's y-axis being upside-down. Also note there is an
	// implicit transpose here, as the naming is the other way round too

	xRot._11 =  xAsuraRot.e11;
	xRot._12 = -xAsuraRot.e12;
	xRot._13 =  xAsuraRot.e13;
	xRot._21 =  xAsuraRot.e21;
	xRot._22 = -xAsuraRot.e22;
	xRot._23 =  xAsuraRot.e23;
	xRot._31 =  xAsuraRot.e31;
	xRot._32 = -xAsuraRot.e32;
	xRot._33 =  xAsuraRot.e33;

	xOut *= xRot;
}

////////////////////////////////////////////////////////////////////////////////
//
// CreateResources()
//
// Creates D3DPOOL_DEFAULT resources for CSM system. Called on system
// initialisation and after device is regained
//
////////////////////////////////////////////////////////////////////////////////

static bool CreateResources( void* pData = NULL )
{
	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;
	if( !pxDevice ) return false;

	// Create initial render target
	{
		pxDevice->CreateTexture( ASURA_CSM_SIZE, ASURA_CSM_SIZE, 1, D3DUSAGE_RENDERTARGET,
			D3DFMT_R32F, D3DPOOL_DEFAULT, &g_pxRenderTexture, NULL );

		if( g_pxRenderTexture )
		{
			g_pxRenderTexture->GetSurfaceLevel( 0, &g_pxRenderSurface );
		}

		Asura_Assert( g_pxRenderTexture && g_pxRenderSurface, "Failed to create CSM RT." );
	}

	// Create depth stencil surface. Could avoid this and use a min blend op, but probably a good
	// idea to get the benefit of hierarchical Z etc.
	{
		pxDevice->CreateDepthStencilSurface( ASURA_CSM_SIZE, ASURA_CSM_SIZE, D3DFMT_D24X8,
			D3DMULTISAMPLE_NONE, 0, TRUE, &g_pxCSMDS, NULL );

		Asura_Assert( g_pxCSMDS, "Failed to create cascaded shadow map depth-stencil surface." );
	}

	// Create scratch texture for squaring & blurring
	{
		pxDevice->CreateTexture( ASURA_CSM_SIZE, ASURA_CSM_SIZE, 1, D3DUSAGE_RENDERTARGET,
			D3DFMT_G32R32F, D3DPOOL_DEFAULT, &g_pxBlurTexture, NULL );

		if( g_pxBlurTexture )
		{
			g_pxBlurTexture->GetSurfaceLevel( 0, &g_pxBlurSurface );
		}

		Asura_Assert( g_pxBlurTexture && g_pxBlurSurface,
			"Failed to create cascaded shadow map render target." );
	}

	// Create texture to store the shadow maps
	{
		// All cascade maps packed vertically
		pxDevice->CreateTexture( ASURA_CSM_SIZE, ASURA_CSM_SIZE * ASURA_CSM_MAX_MAPS, 1,
			D3DUSAGE_RENDERTARGET, D3DFMT_G32R32F, D3DPOOL_DEFAULT, &g_pxCSMTexture, NULL );

		if( g_pxCSMTexture )
		{
			g_pxCSMTexture->GetSurfaceLevel( 0, &g_pxCSMSurface );
		}

		Asura_Assert( g_pxCSMTexture && g_pxCSMSurface, "Failed to create CSM texture." );
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// DestroyResources()
//
// Releases D3DPOOL_DEFAULT resources for CSM. Called on shutdown & device loss
//
////////////////////////////////////////////////////////////////////////////////

static bool DestroyResources( void* pData = NULL )
{
	SAFE_RELEASE( g_pxRenderSurface );
	SAFE_RELEASE( g_pxRenderTexture );

	SAFE_RELEASE( g_pxCSMDS );

	SAFE_RELEASE( g_pxBlurSurface );
	SAFE_RELEASE( g_pxBlurTexture );

	SAFE_RELEASE( g_pxCSMSurface );
	SAFE_RELEASE( g_pxCSMTexture );

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// RenderGeometry()
//
// Renders the environment geometry using the shadow creation passes
//
////////////////////////////////////////////////////////////////////////////////

static void RenderGeometry( Asura_Direct3D_Effect& xEffect, const Asura_Vector_3& xBSCentre,
	const float fBSRadius, const float fExtraFrustumDepth )
{
	SCOPEDPIXEVENT(( 0, L"Environment geometry" ));

	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;
	if( !pxDevice ) return;

	pxDevice->SetVertexDeclaration( Asura_Direct3D_EnvironmentRenderer::GetVD() );

	// Construct an OBB around the CSM split bounding sphere. Could get tighter by just surrounding
	// the view frustum split section, but don't have OBB functions for that right now
	Asura_OBB xShadowOBB;

	// Enable toggling for perf testing. Once satisfied this bool can go
	if( g_bCullStrips )
	{
		Asura_Matrix_3x3 xCSMOrient;
		Asura_Maths::OrientationFromDirection( xCSMOrient, Asura_CSM::GetShadowDir() );

		xShadowOBB.SetOrientation( Asura_AMatrix3x3( xCSMOrient ) );

		// Offset centre to account for "extra frustum depth"
		Asura_AVector3 xCentre( xBSCentre );
		xCentre -= Asura_AVector3( Asura_CSM::GetShadowDir() ) * (fExtraFrustumDepth * 0.5f);
		xShadowOBB.SetCentre( xCentre );

		// Same for OBB depth
		const float fHalfExtentZ = fBSRadius + fExtraFrustumDepth * 0.5f;
		const Asura_AVector3 xHalfExtents( fBSRadius, fBSRadius, fHalfExtentZ );
		xShadowOBB.SetHalfExtents( xHalfExtents );
	}

	const D3DPRIMITIVETYPE ePrimType = Asura_Direct3D_EnvironmentRenderer::GetPrimType();
	const u_int uNumStrips = Asura_Direct3D_EnvironmentRenderer::GetNumStrips();

	int iCurrPass = -1;

	for( u_int u = 0; u < uNumStrips; ++u )
	{
		const Asura_Direct3D_EnvironmentRenderer_Strip& xStrip =
			Asura_Direct3D_EnvironmentRenderer::GetStrip(u);

		// Enable toggling for perf testing. Once satisfied this bool can go
		if( g_bCullStrips )
		{
			// Only render strips within the map's bounding OBB
			if( !Asura_OBB( xStrip.m_xBB ).Intersects( xShadowOBB ) ) continue;
		}

		const Asura_MaterialResponse* const pxMaterial =
			Asura_MaterialResponse_System::GetMaterialByHashID(
			xStrip.m_auMaterialResponseHashID[0] );

		// Check for strips we should skip

		if( pxMaterial->GetDontCastShadow()
		 || pxMaterial->IsTransparent()
		 || pxMaterial->GetSpecialFXMode() != Asura_MaterialResponse::SPECIALFX_MODE_NONE )
		{
			continue;
		}

		SCOPEDPIXEVENT(( 0, L"Strip %u", u ));

		// Bind the vertices

		Direct3D_VertexBuffer* const pxVB =
			Asura_Direct3D_EnvironmentRenderer::GetVB( xStrip.m_uVertexBuffer );

		pxDevice->SetStreamSource( 0, pxVB, 0,
			sizeof( Asura_Direct3D_EnvironmentRenderer_Vertex ) );

		// Bind the indices

		Direct3D_IndexBuffer* const pxIB =
			Asura_Direct3D_EnvironmentRenderer::GetIB( xStrip.m_uIndexBuffer );

		pxDevice->SetIndices( pxIB );

		// Set the transparency mask from the material

		const int iTransparencyMask = pxMaterial
			? pxMaterial->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP )
			: ASURA_TEXTURE_NONE;

		Asura_Render::CurrentStates.RequestTextureIndex( 0, iTransparencyMask );

		// Select the correct pass

		const int iNewPass =
			(iTransparencyMask == ASURA_TEXTURE_NONE ? uPASS_CREATE : uPASS_CREATE_TRANS);

		if( iNewPass == uPASS_CREATE_TRANS )
		{
			// Texture animation if using transparent pass

			Asura_TextureAnimationOutput xTexAnim;
			Asura_TextureAnimation_System::GetAnimationState( pxMaterial->GetTextureAnim(), NULL,
				&xTexAnim );

			const D3DXVECTOR4 xVec( xTexAnim.m_fUOffset, xTexAnim.m_fVOffset, 0.0f, 0.0f );
			xEffect.SetParameterByHash( "g_xTexAnim", 0xABF3D418, xVec );
		}

		if( iNewPass != iCurrPass )
		{
			if( iNewPass != -1 )
			{
				xEffect.EndPass();
			}

			xEffect.BeginPass( iNewPass );
			iCurrPass = iNewPass;
		}
		else if( iNewPass == uPASS_CREATE_TRANS )
		{
			// BeginPass() will commit for us. If not switching, do it manually, only for
			// transparent strips
			xEffect.GetD3DEffect()->CommitChanges();
		}

		// Finally we can draw

		pxDevice->DrawIndexedPrimitive( ePrimType, 0, xStrip.m_uStartVertex,
			xStrip.m_uNumberOfVertices, xStrip.m_uStartIndex, xStrip.m_uNumberOfPrimitives );
	}

	if( iCurrPass != -1 )
	{
		xEffect.EndPass();
	}
}

/******************************************************************************/
/* Asura_CSM Static Class Function                                            */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Platform_Initialise()
//
// Initialises the CSM system on PC
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Platform_Initialise()
{
	Asura_Direct3D::AddPreResetCallback( DestroyResources, NULL );
	Asura_Direct3D::AddPostResetCallback( CreateResources, NULL );

	CreateResources();

	#if defined ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_CommandConsole::AddVar( L"Render.CSM.SlopeScaleDepthBias", &g_fSlopeScaleDepthBias,
			-FLT_MAX, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.DepthBias", &g_fDepthBias,
			-FLT_MAX, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.Power", &g_fPower, 0.1f, 100.0f,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.EnableCull", &g_bCullStrips,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.StaggerUpdates", &g_bStaggerUpdates,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Platform_Reset()
//
// Resets the CSM system on PC
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Platform_Reset()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Platform_Shutdown()
//
// Shuts the CSM system down on PC
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Platform_Shutdown()
{
	DestroyResources();

	Asura_Direct3D::RemovePostResetCallback( CreateResources, NULL );
	Asura_Direct3D::RemovePreResetCallback( DestroyResources, NULL );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Platform_CreateShadowMaps()
//
// Constructs cascaded shadow maps on PC, based on camera 0 pos & ori
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Platform_CreateShadowMaps()
{
	SCOPEDPIXEVENT(( 0, L"Create cascaded shadow maps" ));

	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;

	Asura_Direct3D_Effect* const pxEffect =
		Asura_Direct3D_Effects_System::GetEffect( uEFFECT_HASH );

	Asura_Assert( pxDevice && pxEffect, "Missing device or effect for CSM." );
	if( !pxDevice || !pxEffect ) return;

	Direct3D_Surface* pxOldRT = NULL;
	Direct3D_Surface* pxOldDepthStencil = NULL;

	// Set up
	{
		SCOPEDPIXEVENT(( 0, L"Set up" ));

		// Ensure the shadow state is valid
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );

		Asura_Render::CurrentStates.RequestDitheringMode( false );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestStencilEnabled( false );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

		pxDevice->GetRenderTarget( 0, &pxOldRT );
		pxDevice->GetDepthStencilSurface( &pxOldDepthStencil );

		const D3DXVECTOR4 xSplits( &s_afSplits[1] );
		pxEffect->SetParameterByHash( "g_xSplits", 0xCA9E5F39, xSplits );

		pxEffect->Begin();

		const DWORD uSSDB = *reinterpret_cast< const DWORD* >( &g_fSlopeScaleDepthBias );
		pxDevice->SetRenderState( D3DRS_SLOPESCALEDEPTHBIAS, uSSDB );

		const DWORD uDB = *reinterpret_cast< const DWORD* >( &g_fDepthBias );
		pxDevice->SetRenderState( D3DRS_DEPTHBIAS, uDB );
	}

	// Create each cascaded shadow texture. This step writes depth, square and blurs, then packs
	// into main texture

	for( u_int u = 0; u < s_uNumMaps; ++u )
	{
		if( !UpdateThisFrame(u) ) continue;

		SCOPEDPIXEVENT(( 0, L"Map %u", u ));

		// Write depths
		{
			SCOPEDPIXEVENT(( 0, L"Write depths" ));

			pxDevice->SetRenderTarget( 0, g_pxRenderSurface );
			pxDevice->SetDepthStencilSurface( g_pxCSMDS );

			const u_int uFlags = D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER;
			pxDevice->Clear( 0, NULL, uFlags, D3DCOLOR_XRGB( 0xFF, 0, 0 ), 1.0f, 0 );

			CreateViewMatrix( g_axCSMMatrices[u], s_axShadowViewMatrixTrans[u],
				s_axShadowViewMatrixOrients[u] );

			D3DXMATRIX xProj;
			const float fDiameter = s_afBSRadii[u] * 2.0f;

			// Push the near plane back out to catch shadow casters that aren't in frustum
			D3DXMatrixOrthoLH( &xProj, fDiameter, fDiameter, -s_fExtraFrustumDepth, fDiameter );

			g_axCSMMatrices[u] *= xProj;

			// Find out where the origin will project to in the shadow map
			D3DXVECTOR4 xOrigin( 0.0f, 0.0f, 0.0f, 1.0f );
			D3DXVec4Transform( &xOrigin, &xOrigin, &g_axCSMMatrices[u] );

			float fOriginX = (xOrigin.x * 0.5f) * ASURA_CSM_SIZE;
			float fOriginY = (xOrigin.y * 0.5f) * ASURA_CSM_SIZE;

			float fPixelDiffX = floorf( fOriginX ) - fOriginX;
			float fPixelDiffY = floorf( fOriginY ) - fOriginY;

			float fRoundingX = (fPixelDiffX / ASURA_CSM_SIZE) * 2.0f;
			float fRoundingY = (fPixelDiffY / ASURA_CSM_SIZE) * 2.0f;

			D3DXMATRIX xRounding;
			D3DXMatrixTranslation( &xRounding, fRoundingX, fRoundingY, 0.0f );

			g_axCSMMatrices[u] *= xRounding;

			pxEffect->SetParameterByHash( "g_xShadowViewProj", 0xD6D0D342, g_axCSMMatrices[u] );

			Asura_Render::CurrentStates.RequestAlphaTest( true );
			Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
			Asura_Render::CurrentStates.RequestZBufferEnabled( true );

			RenderGeometry( *pxEffect, s_axBSCentres[u], s_afBSRadii[u], s_fExtraFrustumDepth );
		}

		// Square and horz blur
		{
			SCOPEDPIXEVENT(( 0, L"Square and horizontal blur" ));

			pxDevice->SetRenderTarget( 0, g_pxBlurSurface );
			pxDevice->SetDepthStencilSurface( NULL );

			pxDevice->SetStreamSource( 0, Asura_Direct3D::GetFullScreenVB(), 0, sizeof( short )*2 );
			pxDevice->SetVertexDeclaration( Asura_Direct3D::GetFullScreenVD() );

			Asura_Render::CurrentStates.RequestAlphaTest( false );
			Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
			Asura_Render::CurrentStates.RequestZBufferEnabled( false );

			pxEffect->SetParameterTextureByHash( "g_xBlurTexture", 0xBDB3C034, g_pxRenderTexture );

			// Blur texcoord offsets, including standard half-texel offset
			const float afOffsets[5] =
			{
				-1.5f / ASURA_CSM_SIZE,
				-0.5f / ASURA_CSM_SIZE,
				+0.5f / ASURA_CSM_SIZE,
				+1.5f / ASURA_CSM_SIZE,
				+2.5f / ASURA_CSM_SIZE,
			};

			const u_int uSize = sizeof( afOffsets );

			pxEffect->SetParameterByHash( "g_afBlurOffsets", 0x20CCF8DC, afOffsets, uSize );

			D3DXVECTOR4 xRcpTexSize( 1.0f / ASURA_CSM_SIZE, 0.5f / ASURA_CSM_SIZE, 0.0f, 0.0f );
			pxEffect->SetParameterByHash( "g_xRcpBlurTexSize", 0x370D2F42, xRcpTexSize );

			pxEffect->BeginPass( uPASS_BLUR_HORZ );
			pxDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
			pxEffect->EndPass();
		}

		// Vert blur and pack into main texture
		{
			SCOPEDPIXEVENT(( 0, L"Vertical blur and pack" ));

			pxDevice->SetRenderTarget( 0, g_pxCSMSurface );

			// Set the viewport to the section of the texture we're rendering to
			const Direct3D_Viewport xViewport =
				{ 0, u * ASURA_CSM_SIZE, ASURA_CSM_SIZE, ASURA_CSM_SIZE, 0.0f, 1.0f };

			pxDevice->SetViewport( &xViewport );

			pxEffect->SetParameterTextureByHash( "g_xBlurTexture", 0xBDB3C034, g_pxBlurTexture );

			// Blur texcoord offsets, including standard half-texel offset
			const float afOffsets[5] =
			{
				-1.0f / ASURA_CSM_SIZE,
				+1.0f / ASURA_CSM_SIZE,
				+1.0f / ASURA_CSM_SIZE,
				+2.0f / ASURA_CSM_SIZE,
				0.0f // Not used
			};

			const u_int uSize = sizeof( afOffsets );

			pxEffect->SetParameterByHash( "g_afBlurOffsets", 0x20CCF8DC, afOffsets, uSize );

			const float fCascadeOffset = static_cast< float >( u * ASURA_CSM_SIZE );
			pxEffect->SetParameterByHash( "g_fCascadeOffset", 0xFE899C79, fCascadeOffset );

			pxEffect->BeginPass( uPASS_BLUR_VERT );
			pxDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
			pxEffect->EndPass();
		}
	}

	// Clean up
	{
		SCOPEDPIXEVENT(( 0, L"Clean up" ));

		pxEffect->SetParameterTextureByHash( "g_xBlurTexture", 0xBDB3C034, NULL );

		pxDevice->SetRenderState( D3DRS_SLOPESCALEDEPTHBIAS, 0 );
		pxDevice->SetRenderState( D3DRS_DEPTHBIAS, 0 );

		pxEffect->End();

		pxDevice->SetRenderTarget( 0, pxOldRT );
		pxDevice->SetDepthStencilSurface( pxOldDepthStencil );

		SAFE_RELEASE( pxOldRT );
		SAFE_RELEASE( pxOldDepthStencil );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Platform_ApplyToScene()
//
// Applies CSM data to the scene on PC, as a darkening pass
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Platform_ApplyToScene()
{
	SCOPEDPIXEVENT(( 0, L"Apply Cascaded Shadow Maps" ));

	Direct3D_Device* const pxDevice = Asura_Direct3D::D3DDevice;

	Asura_Direct3D_Effect* const pxEffect =
		Asura_Direct3D_Effects_System::GetEffect( uEFFECT_HASH );

	if( !pxDevice || !pxEffect ) return;

	// Set up
	{
		SCOPEDPIXEVENT(( 0, L"Set up" ));

		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
		Asura_Render::CurrentStates.RequestDitheringMode( false );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestStencilEnabled( false );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL);
		Asura_Render::CurrentStates.RequestZBufferEnabled( false );

		pxDevice->SetVertexDeclaration( Asura_Direct3D::GetFullScreenVD() );
		pxDevice->SetStreamSource( 0, Asura_Direct3D::GetFullScreenVB(), 0, sizeof( short ) * 2 );

		{
			Asura_Direct3D_RenderTarget* const pxRT = Asura_Direct3D::GetDeferredRenderingTarget();

			Direct3D_Texture* const pxNormalDepth =
				pxRT->GetTexture( ASURA_DIRECT3D_DEFFEDRENDERTARGET_NORMAL_INDEX );

			pxEffect->SetParameterTextureByHash( "g_xDepthTexture", 0x73BB9798, pxNormalDepth );
		}

		pxEffect->SetParameterTextureByHash( "g_xShadowMapTexture", 0xF2BC8FFF, g_pxCSMTexture );

		{
			const D3DXMATRIX& xCamViewProj = Asura_Direct3D::GetViewMatrix();

			D3DXMATRIX xInvCamViewProj;
			D3DXMatrixInverse( &xInvCamViewProj, NULL, &xCamViewProj );

			pxEffect->SetParameterByHash( "g_xInvCamViewProj", 0x81367260, xInvCamViewProj );
		}

		{
			const D3DXVECTOR4 xTwoOverScreenDims
			(
				2.0f / static_cast< float >( Asura_Viewport::GetWidth() ),
				2.0f / static_cast< float >( Asura_Viewport::GetHeight() ),
				0.0f, 0.0f
			);

			pxEffect->SetParameterByHash( "g_xTwoOverScreenDims", 0x24474DD7, xTwoOverScreenDims );
		}

		{
			const D3DXVECTOR4 xTanHalfFOV
			(
				1.0f / Asura_Camera::GetOneOverTanHalfFovX(),
				1.0f / Asura_Camera::GetOneOverTanHalfFovY(),
				0.0f, 0.0f
			);

			pxEffect->SetParameterByHash( "g_xCameraTanHalfFOV", 0x90E6BC7E, xTanHalfFOV );
		}

		{
			const D3DXVECTOR4 xScreenDims
			(
				1.0f / Asura_Viewport::GetWidth(),
				1.0f / Asura_Viewport::GetHeight(),
				0.5f / Asura_Viewport::GetWidth(),
				0.5f / Asura_Viewport::GetHeight()
			);

			pxEffect->SetParameterByHash( "g_xScreenDims", 0x83FE7DF7, xScreenDims );
		}

		{
			const float fRcpMaxMaps = 1.0f / ASURA_CSM_MAX_MAPS;
			pxEffect->SetParameterByHash( "g_fRcpMaxMaps", 0xE8CD7DCA, fRcpMaxMaps );
		}

		{
			const D3DXVECTOR4 xSplits( &s_afSplits[1] );
			pxEffect->SetParameterByHash( "g_xSplits", 0xCA9E5F39, xSplits );
		}

		pxEffect->GetD3DEffect()->SetMatrixArray( "g_axShadowViewProj", g_axCSMMatrices,
			ASURA_CSM_MAX_MAPS );

		pxEffect->SetParameterByHash( "g_fPower", 0x2175E337, g_fPower );

		pxEffect->SetParameterByHash( "g_afStrengths", 0x7F7381F5, &s_afStrengths[0],
			sizeof( float ) * s_afStrengths.GetSize() );

		pxEffect->Begin();
		pxEffect->BeginPass( uPASS_APPLY );
	}

	// Apply to scene
	{
		pxDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
	}

	// Clean up
	{
		SCOPEDPIXEVENT(( 0, L"Clean up" ));

		pxEffect->EndPass();
		pxEffect->End();

		pxEffect->SetParameterTextureByHash( "g_xDepthTexture", 0x73BB9798, NULL );
		pxEffect->SetParameterTextureByHash( "g_xShadowMapTexture", 0xF2BC8FFF, NULL );
	}
}

#endif // defined ASURA_USE_CSM
