//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Audio.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

#include "Sound/Asura_Audio_System.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_DSP.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_Sound_Mixer.h"
#include "Sound/Asura_XAudio2.h"
#include "Sound/Asura_XAudio2_Buffer.h"

// Tools
#if defined ASURA_TOOLS
	#include "Sound/Tools_Audio_DSP.h"
	#include "Sound/Tools_Sound_System.h"
#endif

// Misc
#include <objbase.h>

/******************************************************************************/
/* Asura_Audio_System Static Class Functions                                  */
/******************************************************************************/

void Asura_Audio_System::Platform_Initialise()
{
	// Must initialise COM before XAudio2. COINIT_MULTITHREADED causes a crash on exit though...
	CoInitializeEx( NULL, COINIT_APARTMENTTHREADED );
	Asura_XAudio2::Initialise();

	#ifdef ASURA_TOOLS
		LoadAsuraFile( Asura_Chunk_DSP::GetFilename() );
		LoadAsuraFile( szTOOLS_AUDIO_REVERB_FILENAME );
	#endif
}

void Asura_Audio_System::Platform_Reset()
{
	Asura_XAudio2::Reset();
}

void Asura_Audio_System::Platform_Shutdown()
{
	Asura_XAudio2::Shutdown();
	CoUninitialize();
}

void Asura_Audio_System::Platform_Update()
{
	Asura_XAudio2::Update();
}

void Asura_Audio_System::Platform_CreateDSPChain( const Asura_Hash_ID uHash, const Asura_Audio_DSPChain& xChain )
{
	//Asura_XAudio2::CreateDSPChain( uHash, xChain );
}

Asura_Handle Asura_Audio_System::Platform_CreateVoice( const Asura_Hash_ID uWave,
	const bool bFilter, const bool bMusic, const Asura_Hash_ID uDSPChain /* = ASURA_HASH_ID_UNSET*/ )
{
	return Asura_XAudio2::CreateVoice( uWave, bFilter, bMusic, uDSPChain );
}

Asura_Handle Asura_Audio_System::Platform_CreateStreamingVoice( const Asura_Hash_ID uStreamHash, const bool bFilter,
	const bool bMusic, const float fTimeOffset, const Asura_Hash_ID uDSPChain /* = ASURA_HASH_ID_UNSET*/ )
{
	return Asura_XAudio2::CreateStreamingVoice( uStreamHash, bFilter, bMusic, fTimeOffset, uDSPChain );
}

Asura_Handle Asura_Audio_System::Platform_CreateSubmixVoice( const bool bFilter, const bool bMusic,
	const Asura_Hash_ID uDSPChain /* = ASURA_HASH_ID_UNSET*/ )
{
	return Asura_XAudio2::CreateSubmixVoice( bFilter, bMusic, uDSPChain );
}

void Asura_Audio_System::Platform_DestroyVoice( const Asura_Handle iVoiceHandle )
{
	Asura_XAudio2::DestroyVoice( iVoiceHandle );
}

Asura_Audio_Voice* Asura_Audio_System::Platform_GetVoice( const Asura_Handle iVoiceHandle )
{
	return Asura_XAudio2::GetVoice( iVoiceHandle );
}

void Asura_Audio_System::Platform_SetMasterDSP( const Asura_Hash_ID uDSPChain, const float fFadeTime )
{
	Asura_XAudio2::SetMasterDSP( uDSPChain, fFadeTime );
}

bool Asura_Audio_System::Platform_LoadWave( Asura_Chunk_Stream& xStream, const u_int uFileSize,
	const u_int uSubType, const u_int uFlags, const char* const szFilename )
{
	#if defined ASURA_TOOLS

		// In Tools builds, we might have already created a reference to this sound file by parsing
		// say a sound event chunk. We might now be loading ToolsGmSnd.asr, which contains the
		// actual data

		if( Asura_XAudio2::IsWaveLoaded( szFilename ) )
		{
			Asura_XAudio2_Buffer* const pxBuffer =
				Asura_XAudio2::GetBuffer( Asura_GetHashID( szFilename ) );

			return pxBuffer->LoadIfDelayed( &xStream, uFileSize );
		}

	#endif
	bool bAlreadyExists;
	Asura_XAudio2_Buffer* const pxBuffer = Asura_XAudio2::CreateBuffer( szFilename, bAlreadyExists );
	if( pxBuffer && !bAlreadyExists)
	{
		if( !pxBuffer->LoadWave( xStream, uFileSize, uFlags ) )
		{
			Asura_Assertf( false, ("Failed to load wave file '%s'.", szFilename) );
		}
		return true;
	}
	else
	{
		return false;
	}
}

#if defined ASURA_TOOLS

bool Asura_Audio_System::Platform_LoadWave( const char* const szFilename, const bool bLooping )
{
	if( Asura_XAudio2::IsWaveLoaded( szFilename ) )
	{
		// Do nothing if the file is already loaded
		return true;
	}

	// In Tools, use delayed loading. We'll only actually open the file when it's first played
	bool bAlreadyExists;
	Asura_XAudio2_Buffer* const pxBuffer = Asura_XAudio2::CreateBuffer( szFilename, bAlreadyExists );

	if( pxBuffer && !bAlreadyExists)
	{
		pxBuffer->SetLooping( bLooping );
		return true;
	}
	else
	{
		return false;
	}
}

void Asura_Audio_System::Platform_ForceWaveToBeLoaded( const Asura_Hash_ID uWave )
{
	Asura_XAudio2::ForceWaveToBeLoaded( uWave );
}

void Asura_Audio_System::Platform_ReloadWave( const Asura_Hash_ID uWave,
	const bool bForceReloadNow )
{
	Asura_XAudio2::ReloadWave( uWave, bForceReloadNow );
}

void Asura_Audio_System::Platform_SetWaveLooping( const Asura_Hash_ID uWave, const bool bLoop )
{
	Asura_XAudio2::SetWaveLooping( uWave, bLoop );
}

#endif // defined ASURA_TOOLS

#if defined ASURA_AUDIO_STORE_FILENAMES

const char* Asura_Audio_System::Platform_GetWaveFilename( const Asura_Hash_ID uWave )
{
	return Asura_XAudio2::GetWaveFilename( uWave );
}

#endif // defined ASURA_AUDIO_STORE_FILENAMES

/******************************************************************************/
/* Asura_XAudio2 Static Class Functions                                       */
/******************************************************************************/

void* Asura_XAudio2::Platform_Allocate( const WORD usFormatTag, const u_int uSize )
{
	// No need to do anything special on PC
	return Asura_Allocate( uSize );
}

void Asura_XAudio2::Platform_Deallocate( const WORD usFormatTag, void* const pBuffer )
{
	Asura_Deallocate( pBuffer );
}

#endif // defined ASURA_USE_AUDIO_SYSTEM
