/**
*
* \file Asura_Platform_Decal_SM3.cpp
*
* Created by Dave Reed 08/12/2006
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* This is the Shader Model 3.0 version of the decal renderer - upgraded to support 
* normal+specular maps, and the static decals system
*
**/

#include "Core/Asura.h"

#ifdef ASURA_TARGET_SM3

#include "Core/Asura_Timers.h"
#include "Core/Asura_HalfFloat.h"
#include "Core/Asura_TextureManagement.h"
#include "Light/Asura_GI_Solution.h"
#include "SpecialFX/Asura_Decal_System.h"
#include "SpecialFX/Asura_Decal_Manager.h"
#include "SpecialFX/Asura_Decal_Template.h"
#include "SpecialFX/Asura_Platform_Decal_SM3.h"
#include "Camera/Asura_Camera.h"
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Direct3D_Effects_System.h"
#include "Rendering/Asura_Direct3D_MaterialResponse.h"
#include "Rendering/Asura_Colour.h"
#include "Visibility/Asura_Visibility_System.h"

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
 #ifdef ASURA_TOOLS
  #include "SpecialFX/Asura_Decal_Manager.h"
  #include "Decals/Tools_StaticDecal_System.h"
 #endif
#endif

#ifdef ASURA_TOOLS
#include "Misc/Tools_Render_Options.h"
#endif


// ------------------------------------------------------------------------------------
// Vertex format for normal-mapped decals
// ------------------------------------------------------------------------------------

const Direct3D_VertexElement g_axDecalVertexDeclaration[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,     0 },
	{ 0, 12, D3DDECLTYPE_FLOAT4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     0 },
	{ 0, 28, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,		  0 },
	{ 0, 40, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,	  0 },
	{ 0, 52, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL,	  0 },	
	{ 0, 64, D3DDECLTYPE_FLOAT4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  1 },	 // SH Red
	{ 0, 80, D3DDECLTYPE_FLOAT4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  2 },	 // SH Green
	{ 0, 96, D3DDECLTYPE_FLOAT4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  3 },	 // SH Blue
	D3DDECL_END()
};

// ====================================================================================
// Asura_Platform_Decal_System_SM3
// ====================================================================================

DEBUGVAR float	dbg_fPlatformDecalSystemSM3DepthBias	= -0.00001f;

// Asura_Decal_System - Platform functions redirected to Asura_Platform_Decal_System_SM3
void Asura_Decal_System::Platform_Initialise( void )			{Asura_Platform_Decal_System_SM3::Initialise();			}
void Asura_Decal_System::Platform_Shutdown( void )				{Asura_Platform_Decal_System_SM3::Shutdown();			}
void Asura_Decal_System::Platform_Reset( void )					{Asura_Platform_Decal_System_SM3::Reset();				}

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
void Asura_Decal_System::Platform_RenderStaticDecals( void )    {Asura_Platform_Decal_System_SM3::RenderStaticDecals(); }
#endif

// The declaration objects
Direct3D_VertexDeclaration*				Asura_Platform_Decal_System_SM3::s_pxDecalVertexDecl		= NULL;
Direct3D_VertexDeclaration*				Asura_Platform_Decal_System_SM3::s_pxDecalVertexSimpleDecl	= NULL;

// The shader
Asura_Direct3D_Effect*					Asura_Platform_Decal_System_SM3::s_pxDecalEffect			= NULL;

// Its parameters
Asura_Direct3D_Decal_ShaderParameters	Asura_Platform_Decal_System_SM3::s_xShaderParams;

int										Asura_Platform_Decal_System_SM3::s_iNumStaticDecalBuffers	= 0;
Asura_Direct3D_StaticDecalBuffer*		Asura_Platform_Decal_System_SM3::s_pxStaticDecalBuffers		= NULL;



// ------------------------------------------------------------------------------------
// Platform_Initialise() - Create vertex decls
// ------------------------------------------------------------------------------------
void Asura_Platform_Decal_System_SM3::Initialise()
{
	if( NULL == Asura_Direct3D::D3DDevice)
	{
		return;
	}

	Asura_Direct3D::D3DDevice->CreateVertexDeclaration(g_axDecalVertexDeclaration, &s_pxDecalVertexDecl);
	Asura_Assert(g_axDecalVertexDeclaration, "Decal renderer couldn't create vertex decl");	
	CaptureEffect();	
}


// ------------------------------------------------------------------------------------
// Platform_Reset()
// ------------------------------------------------------------------------------------
void Asura_Platform_Decal_System_SM3::Reset()
{
	// MGH 14/1/2008: Fixed memory fragmentation for PC retail build.
#if !defined(ASURA_TOOLS)
	if (s_pxStaticDecalBuffers)
	{
		delete [] s_pxStaticDecalBuffers;
		s_pxStaticDecalBuffers = NULL;
		s_iNumStaticDecalBuffers = 0;
	}
#endif //!defined(ASURA_TOOLS)
}


// ------------------------------------------------------------------------------------
// Platform_Shutdown() - Destroy resources
// ------------------------------------------------------------------------------------
void Asura_Platform_Decal_System_SM3::Shutdown()
{
	if (NULL != s_pxStaticDecalBuffers)
	{
		delete [] s_pxStaticDecalBuffers;
		s_pxStaticDecalBuffers = NULL;
	}

	if( NULL != s_pxDecalVertexDecl)
	{
		s_pxDecalVertexDecl->Release();
		s_pxDecalVertexDecl = NULL;
	}

	if( NULL != s_pxDecalVertexSimpleDecl)
	{
		s_pxDecalVertexSimpleDecl->Release();
		s_pxDecalVertexSimpleDecl = NULL;
	}
}

// -----------------------------------------------------------------------------------
// AllocateStaticDecalBuffers() - Called from the resource chunk loader
// -----------------------------------------------------------------------------------

void Asura_Platform_Decal_System_SM3::AllocateStaticDecalBuffers(int iNumBuffers)
{
	if (s_pxStaticDecalBuffers)
	{
		delete [] s_pxStaticDecalBuffers;
		s_pxStaticDecalBuffers = NULL;
		s_iNumStaticDecalBuffers = 0;
	}

	s_pxStaticDecalBuffers = new Asura_Direct3D_StaticDecalBuffer[iNumBuffers];
	s_iNumStaticDecalBuffers = iNumBuffers;
}

// -----------------------------------------------------------------------------------
// GetStaticDecalBuffer()
// -----------------------------------------------------------------------------------

Asura_Direct3D_StaticDecalBuffer* Asura_Platform_Decal_System_SM3::GetStaticDecalBuffer(int iIndex)
{
	Asura_Assert(iIndex>=0 && iIndex<s_iNumStaticDecalBuffers, "Invalid staticdecal buffer index!");
	return &s_pxStaticDecalBuffers[iIndex];
}



// ------------------------------------------------------------------------------------
// Shader loading + parameter getting
// ------------------------------------------------------------------------------------

void Asura_Platform_Decal_System_SM3::CaptureEffect(void)
{
	s_pxDecalEffect = Asura_Direct3D_Effects_System::GetEffect("decals");

	if (s_pxDecalEffect)
	{	
#if ASURA_LOAD_EFFECT_FILES == 1
		s_pxDecalEffect->SetOnLoadCallback(OnEffectLoaded);
		s_pxDecalEffect->SetOnUnloadCallback(OnEffectUnloaded);
#endif
		s_xShaderParams.Get(s_pxDecalEffect);
	}
}

#if ASURA_LOAD_EFFECT_FILES == 1
void Asura_Platform_Decal_System_SM3::OnEffectLoaded(Asura_Direct3D_Effect* pxEffect)
{
	CaptureEffect();
}

void Asura_Platform_Decal_System_SM3::OnEffectUnloaded(Asura_Direct3D_Effect* pxEffect)
{
	CaptureEffect();
}
#endif



// ------------------------------------------------------------------------------------
// Asura_Platform_Decal_System_SM3::PrepareRenderStates()
// ------------------------------------------------------------------------------------

void Asura_Platform_Decal_System_SM3::PrepareRenderStates()
{
	Asura_Render::CurrentStates.RequestAlphaTest(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);	
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_DEPTHBIAS, *((DWORD *)&dbg_fPlatformDecalSystemSM3DepthBias) );

	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SEPARATEALPHABLENDENABLE, TRUE );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALPHA);
}


// ------------------------------------------------------------------------------------
// Asura_Platform_Decal_System_SM3::RestoreRenderStates()
// ------------------------------------------------------------------------------------

void Asura_Platform_Decal_System_SM3::RestoreRenderStates()
{
	const float fZBias = 0.0f;
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_DEPTHBIAS, *((DWORD *)&fZBias) );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);	
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);

	for( u_int u = 0; u < 16; ++u )
	{
		Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_SRGBTEXTURE, FALSE );
	}

	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SEPARATEALPHABLENDENABLE, FALSE );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SRCBLENDALPHA, D3DBLEND_SRCALPHA );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DESTBLENDALPHA, D3DBLEND_INVSRCALPHA );

	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, 0xF);
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, 0xF);
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, 0xF);

	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);


}



// ------------------------------------------------------------------------------------
// Asura_Xbox360_Decal_System::SetupMaterial()
// PC only ever uses iPass = 0 at the moment - 360 version uses 2 passes to handle
// blending into the specular/normal buffers, which we can't do currently on PC
// due to the encoding of the normals
// ------------------------------------------------------------------------------------

void  Asura_Platform_Decal_System_SM3::SetupMaterial(int iPass, const Asura_Decal_Material *pxMtl, float *pfFadeValues)
{
	Asura_Direct3D_Effect *pxEffect = Asura_Platform_Decal_System_SM3::GetEffect();
	Asura_Direct3D_Decal_ShaderParameters *pxParams = Asura_Platform_Decal_System_SM3::GetEffectParameters();		
	Direct3D_Effect *pxD3DEffect = pxEffect->GetD3DEffect();

	const Asura_MaterialResponse* pxMaterial
		= Asura_MaterialResponse_System::GetMaterialByHashID( pxMtl->m_uMaterialResponseHash );

	const Asura_MaterialResponse::BLEND_MODE eBlendMode = pxMaterial
		? pxMaterial->GetBlendMode()
		: Asura_MaterialResponse::BLEND_MODE_OPAQUE;

	if (iPass == 0 && pxMaterial)
	{
		Asura_RGBA xPrelightingColour = pxMaterial->GetAmbientColour();
		xPrelightingColour *= pxMaterial->GetAmbientContribution();

		D3DXVECTOR4 xPrelightingColourVector = D3DXVECTOR4( xPrelightingColour.m_fR, xPrelightingColour.m_fG, xPrelightingColour.m_fB, 0.0f );

		pxD3DEffect->SetVector(pxParams->m_xPrelightingColourFilter, &xPrelightingColourVector);


		Asura_Direct3D_MaterialResponse::SetEffectParameters( pxMaterial, NULL,	NULL, *pxEffect );

		Asura_Direct3D_MaterialResponse::SetSHParameters( pxMaterial, *pxEffect );
		Asura_Direct3D_MaterialResponse::SetFogParameters( false, pxMaterial, *pxEffect );


	}

	if (iPass == 0)
	{

		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
		if (eBlendMode == Asura_MaterialResponse::BLEND_MODE_ADDITIVE)
		{
			Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
		} 
		else
		{
			Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
		}


		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SRCBLENDALPHA, D3DBLEND_ZERO);
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DESTBLENDALPHA, D3DBLEND_INVSRCALPHA);

		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE);
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE);

		if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback())
		{
			Asura_MaterialResponse_System::GetMaterialProjectContextCallback()(
				pxMaterial ? pxMaterial->GetProjectContext() : ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID, 
				ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
		}

		pxD3DEffect->SetVector(pxParams->m_xFadeValues,(D3DXVECTOR4 *)pfFadeValues);

		// UV scrolling
		float afUVScroll[4];

		if (pxMtl->m_fUScrollSpeed>0.0f || pxMtl->m_fVScrollSpeed>0.0f)
		{	
			float fTimer = Asura_Timers::GetLocalScaledTime();		
			afUVScroll[0] = Asura_Maths::Sine(fTimer * pxMtl->m_fUScrollSpeed) * pxMtl->m_fUScrollRange;
			afUVScroll[1] = Asura_Maths::Cosine(fTimer * pxMtl->m_fVScrollSpeed) * pxMtl->m_fVScrollRange;
		}
		else
		{
			afUVScroll[0] = afUVScroll[1] = 0;
		}

		pxD3DEffect->SetVector(pxParams->m_xUVScroll, (D3DXVECTOR4 *)afUVScroll);

	}
	else
	{
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SRCBLENDALPHA, D3DBLEND_ONE);
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DESTBLENDALPHA, D3DBLEND_ONE);

		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, 0);
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, 0);

	}

	if (pxMaterial->GetTexture(Asura_MaterialResponse::NORMAL_MAP) != -1 && iPass == 0)
	{
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_ALPHA);
	} 
	else 
	{
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALPHA);
	}


}




// ====================================================================================
// Asura_Direct3D_DecalBuffer 
// ====================================================================================

// ------------------------------------------------------------------------------------
// Constructor and Destructor
// ------------------------------------------------------------------------------------

Asura_Direct3D_DecalBuffer::Asura_Direct3D_DecalBuffer(const Asura_Decal_Template* pxDecalTemplate)
{
	const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(pxDecalTemplate->m_iDecalMaterialIndex);
	
	m_iDecalMaterialIndex = pxDecalTemplate->m_iDecalMaterialIndex;
	m_iMaxDecals		  = pxMtl->m_iMaxDecals;		
	m_apxDecals			  = new Decal[m_iMaxDecals];
	m_iNumVerts			  = m_iMaxDecals*8;
	m_apxVerts			  = new D3D_DecalVertex[m_iNumVerts];		 // Worst-case is 8 verts in a decal
	m_iNumIndices		  = m_iMaxDecals*18;
	m_apxIndices		  = new uint16[m_iNumIndices];				 // Worst-case is 18 indices for a decal  (if all 8 verts are used)
	m_iFirstDecal		  = 0;
	m_iNumDecals		  = 0;
	m_iNextVertex		  = 0;
	m_iNumDecalTris		  = 0;									 
	m_pxNextDecalBuffer   = NULL;		
}


Asura_Direct3D_DecalBuffer::~Asura_Direct3D_DecalBuffer()
{
	if (m_pxNextDecalBuffer)
	{	
		delete m_pxNextDecalBuffer;
	}

	delete [] m_apxDecals;
	delete [] m_apxVerts;
	delete [] m_apxIndices;
}

	

// ------------------------------------------------------------------------------------
// AddStrip() - Adds a strip to the decalbuffer. Note - a decal may consist of
//              several strips
// ------------------------------------------------------------------------------------

void Asura_Direct3D_DecalBuffer::AddStrip(const Asura_Decal_Vertex* pxStrip, const u_int uNumberOfVertices, const Asura_Decal_Template* pxDecalTemplate, const Asura_RGBA& xDiffuseRGBA)
{
	if (m_iNumDecals>=m_iMaxDecals)
	{
		// No room - flush the first one
		m_iNumDecalTris -= (m_apxDecals[m_iFirstDecal].iNumIndices/3);
		m_iFirstDecal++;
		if (m_iFirstDecal>=m_iMaxDecals)
		{
			m_iFirstDecal=0;
		}
		m_iNumDecals--;				
	}		

	// Is there space in the index buffer?
	int iFirstIndex = 0;
	if (m_iNumDecals>0)
	{		
		int iLastDecal = (m_iFirstDecal+m_iNumDecals-1);
		if (iLastDecal>=m_iMaxDecals)
		{
			iLastDecal-=m_iMaxDecals;
		}

		iFirstIndex = m_apxDecals[iLastDecal].iFirstIndex+
					  m_apxDecals[iLastDecal].iNumIndices;

		// De we need to shift the index buffer down
		if (iFirstIndex + ((int)((uNumberOfVertices-2)*3)) >= m_iNumIndices)
		{					
			int iShiftDown = m_apxDecals[m_iFirstDecal].iFirstIndex;			
			int iSrc	   = iShiftDown;
			int iDst	   = 0;
			int iDecal	   = m_iFirstDecal;

			for(int i=0;i<m_iNumDecalTris*3;i++)
			{				
				m_apxIndices[iDst++]=m_apxIndices[iSrc++];
			}

			for(int i=0; i<m_iNumDecals; i++)
			{										
				m_apxDecals[iDecal].iFirstIndex-=iShiftDown;										
				iDecal++;
				if (iDecal>=m_iMaxDecals)
				{
					iDecal-=m_iMaxDecals;
				}
			}	

			iFirstIndex = m_apxDecals[iLastDecal].iFirstIndex+
						  m_apxDecals[iLastDecal].iNumIndices;
		}
	}
	
	// Fill in the decal struct
	int iDecal = m_iFirstDecal+m_iNumDecals;			
	if (iDecal>=m_iMaxDecals)
	{
		iDecal-=m_iMaxDecals;
	}
	m_apxDecals[iDecal].iFirstIndex		= iFirstIndex;
	m_apxDecals[iDecal].iNumIndices		= (uNumberOfVertices-2)*3;
	m_apxDecals[iDecal].fCreationTime	= Asura_Timers::GetLocalScaledTime();
	m_iNumDecals++;
	

	// Fill in the verts+indices, include degenerates to join the strips		
	if ((m_iNextVertex+(int)uNumberOfVertices)>m_iNumVerts)
	{
		m_iNextVertex = 0;
	}

	// Compute the tangent/binormal
	// We take a single normal for the decal now, to avoid a discontinuity when the
	// decal is 'bent' over a polygon edge.
	Asura_Vector_3 xNormal,xTangent,xBinormal;
	xNormal = Asura_Decal_System::GetCurrentDecalNormal();
	xNormal.Normalise();
	{		
		// There's room for some optimization here...
		Asura_Vector_3 axPositions[3];
		Asura_Vector_2 axUVs[3];
		for(int i=0;i<3;i++)
		{			
			axPositions[i] = pxStrip[i].m_xPosition;
			axUVs[i].x	   = pxStrip[i].m_fU;
			axUVs[i].y     = 1.0f - pxStrip[i].m_fV;
		}
		Asura_Decal_System::CalculateTangentAndBinormal(axPositions, axUVs, xTangent, xBinormal);

		// The calculation above is incomplete, as it doesn't take the normal into account... so finish the job:
		xBinormal.Normalise();
		xTangent  = xBinormal ^  xNormal;
		xTangent.Normalise();
		xBinormal = xNormal ^ xTangent;
		xBinormal.Normalise();
	}


	// Converting from a strip to an indexed list for now - avoids messing with degens, and
	// may actually be faster?
	for(u_int i=2; i<uNumberOfVertices; i++)
	{	
		m_apxIndices[iFirstIndex++] = (uint16)(m_iNextVertex+i-2);			
		m_apxIndices[iFirstIndex++] = (uint16)(m_iNextVertex+i-1);			
		m_apxIndices[iFirstIndex++] = (uint16)(m_iNextVertex+i);			
	}

	// Fill in the vertices
	for(u_int i=0; i<uNumberOfVertices; i++)
	{						
		m_apxVerts[m_iNextVertex].fX = pxStrip[i].m_xPosition.x;
		m_apxVerts[m_iNextVertex].fY = pxStrip[i].m_xPosition.y;
		m_apxVerts[m_iNextVertex].fZ = pxStrip[i].m_xPosition.z;
		m_apxVerts[m_iNextVertex].fU = pxStrip[i].m_fU;
		m_apxVerts[m_iNextVertex].fV = pxStrip[i].m_fV;
		
		// Tangent frame
		m_apxVerts[m_iNextVertex].afNormal[0] = xNormal.x;
		m_apxVerts[m_iNextVertex].afNormal[1] = xNormal.y;
		m_apxVerts[m_iNextVertex].afNormal[2] = xNormal.z;
		m_apxVerts[m_iNextVertex].afTangent[0] = xTangent.x;
		m_apxVerts[m_iNextVertex].afTangent[1] = xTangent.y;
		m_apxVerts[m_iNextVertex].afTangent[2] = xTangent.z;
		m_apxVerts[m_iNextVertex].afBinormal[0] = xBinormal.x;
		m_apxVerts[m_iNextVertex].afBinormal[1] = xBinormal.y;
		m_apxVerts[m_iNextVertex].afBinormal[2] = xBinormal.z;

		// Lighting
		m_apxVerts[m_iNextVertex].fAlpha = xDiffuseRGBA.m_fA;

		m_apxVerts[m_iNextVertex].afSHRed[0] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHRed[1] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHRed[2] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHRed[3] = xDiffuseRGBA.m_fR;

		m_apxVerts[m_iNextVertex].afSHGreen[0] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHGreen[1] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHGreen[2] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHGreen[3] = xDiffuseRGBA.m_fG;

		m_apxVerts[m_iNextVertex].afSHBlue[0] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHBlue[1] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHBlue[2] = 0.0f;
		m_apxVerts[m_iNextVertex].afSHBlue[3] = xDiffuseRGBA.m_fB;

		
		m_apxVerts[m_iNextVertex].fCreationTime = m_apxDecals[iDecal].fCreationTime;
		m_iNextVertex++;						
		if (m_iNextVertex>=m_iNumVerts) m_iNextVertex = 0;
	}				

	m_iNumDecalTris+= (uNumberOfVertices-2);
}



// ------------------------------------------------------------------------------------
// Render() - Render the contents of this buffer
// ------------------------------------------------------------------------------------

void Asura_Direct3D_DecalBuffer::Render(void)
{
	if (Asura_Platform_Decal_System_SM3::GetEffect()==NULL)
	{
		Asura_Platform_Decal_System_SM3::CaptureEffect();	

		if (Asura_Platform_Decal_System_SM3::GetEffect()==NULL)
		{
			return;
		};
	}

	// Set up the fade values		
	const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(m_iDecalMaterialIndex);				

	float fCurrentTime = Asura_Timers::GetLocalScaledTime();
	float afFadeValues[4] = {fCurrentTime, pxMtl->m_fFadeInTime,pxMtl->m_fFadeOutTime, pxMtl->m_fDecalStayTime};
	if (pxMtl->m_fDecalStayTime<0)
	{
		// Not 'forever', but should be long enough?
		// (AAM - 17.01.2007) If this value is too large, PC cards
		// don't render properly. Some register overflow perhaps?
		afFadeValues[3] = 10000.0f;
	}

	// Remove faded-out decals here - so it doesn't keep rendering them invisibly.
	while (m_iNumDecals>0 && (fCurrentTime > m_apxDecals[m_iFirstDecal].fCreationTime + (afFadeValues[3]+pxMtl->m_fFadeOutTime)))
	{
		// Decal is faded out, remove it
		m_iNumDecalTris -= (m_apxDecals[m_iFirstDecal].iNumIndices/3);
		m_iFirstDecal++;
		if (m_iFirstDecal>=m_iMaxDecals)
		{
			m_iFirstDecal=0;
		}
		m_iNumDecals--;				
	}			

	Asura_Direct3D_Effect *pxDecalEffect = Asura_Platform_Decal_System_SM3::GetEffect();

	Asura_Direct3D_Effects_System::BeginEffect(*pxDecalEffect);
	Asura_Assert(pxDecalEffect->IsD3DEffectValid(), "Decal effect invalid");
	if (pxDecalEffect->IsD3DEffectValid() && m_iNumDecals>0)
	{
		
		// Render them
		for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
		{
			Asura_Camera::SetCameraIndex(i);

			if (Asura_Camera::IsCameraActive())
			{
				Asura_Render::ApplyCameraSettings();
											
				Asura_Direct3D::D3DDevice->SetVertexDeclaration(Asura_Platform_Decal_System_SM3::GetVertexDecl());

				for (int i = 0; i < 2; ++i)
				{
					Asura_Platform_Decal_System_SM3::SetupMaterial(i, pxMtl, afFadeValues);

					pxDecalEffect->GetD3DEffect()->BeginPass(i);
					pxDecalEffect->GetD3DEffect()->CommitChanges();

					int iFirstIndex = m_apxDecals[m_iFirstDecal].iFirstIndex;

					Asura_Direct3D::D3DDevice->DrawIndexedPrimitiveUP
						(
						D3DPT_TRIANGLELIST, 0, m_iNumVerts,
						m_iNumDecalTris, 
						(void*)&m_apxIndices[iFirstIndex], D3DFMT_INDEX16,
						(void*)m_apxVerts, sizeof(D3D_DecalVertex)
						);

					pxDecalEffect->GetD3DEffect()->EndPass();
				}
			}
		}
	}

	// Flush any effects we may still be using
	Asura_Direct3D_Effects_System::FlushEffects();

	if (m_pxNextDecalBuffer)
	{
		m_pxNextDecalBuffer->Render();
	}
}




// ------------------------------------------------------------------------------------
// Platform_AddStrip - Called from the cross-platform Asura_Decal_System to add
// a tristrip
// ------------------------------------------------------------------------------------
void Asura_Decal_Handler::Platform_AddStrip(const Asura_Decal_Vertex* pxStrip,
										    u_int uNumVertices,
										    const Asura_RGBA& xDiffuseRGBA,
										    const Asura_Decal_Template* pxDecalTemplate )
{
	Asura_Direct3D_DecalBuffer* pxDecalBuffer = static_cast<Asura_Direct3D_DecalBuffer*>(m_pxFirstDecalBuffer);
		
	// Do we have a decal buffer for the current decal material?
	while (pxDecalBuffer)
	{
		if (pxDecalBuffer->m_iDecalMaterialIndex == pxDecalTemplate->m_iDecalMaterialIndex)
		{			
			pxDecalBuffer->AddStrip(pxStrip, uNumVertices, pxDecalTemplate, xDiffuseRGBA);
			return;
		}
		pxDecalBuffer = pxDecalBuffer->m_pxNextDecalBuffer;
	}

	// Need to create a new buffer
	pxDecalBuffer = new Asura_Direct3D_DecalBuffer(pxDecalTemplate);
	pxDecalBuffer->m_pxNextDecalBuffer = static_cast<Asura_Direct3D_DecalBuffer*>(m_pxFirstDecalBuffer);
	m_pxFirstDecalBuffer = pxDecalBuffer;
	pxDecalBuffer->AddStrip(pxStrip,uNumVertices,pxDecalTemplate, xDiffuseRGBA);
}



// ------------------------------------------------------------------------------------
// Platform_Render() - Render all decal buffers in this handler
// ------------------------------------------------------------------------------------
void Asura_Decal_Handler::Platform_Render(void)
{
	if (m_pxFirstDecalBuffer == 0) return;

	SCOPEDPIXEVENT((0xff00ff00, L"Decal Handler"));

	Asura_Platform_Decal_System_SM3::PrepareRenderStates();

	static_cast<Asura_Direct3D_DecalBuffer*>(m_pxFirstDecalBuffer)->Render();

	Asura_Platform_Decal_System_SM3::RestoreRenderStates();
}





#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
// ------------------------------------------------------------------------------------
// Platform_RenderStaticDecals() - Render any static decals.
//
// 2 different paths here, one for tools mode, and one for retail mode. Tools mode
// uses the tools data and DrawIndexedPrimitiveUP. Retail mode uses vertex buffers,
// and 2 different vertex formats - using a smaller one for unlit decals
// ------------------------------------------------------------------------------------
void Asura_Platform_Decal_System_SM3::RenderStaticDecals()
{
	float afFadeValues[4] = {5,1,1,10};

	SCOPEDPIXEVENT((0xff00ff00, L"Static Decals"));

	if (GetEffect()==NULL)
	{
		CaptureEffect();	

		if (GetEffect()==NULL)
		{
			return;
		};
	}

	// Invalidate the texture cache values as the environment which just rendered has screwed them.
	Asura_Render::CurrentStates.RequestTextureIndex(0, ASURA_TEXTURE_INVALID);
	Asura_Render::CurrentStates.RequestTextureIndex(1, ASURA_TEXTURE_INVALID);
	Asura_Render::CurrentStates.RequestTextureIndex(2, ASURA_TEXTURE_INVALID);
	Asura_Render::CurrentStates.RequestTextureIndex(3, ASURA_TEXTURE_INVALID);

	PrepareRenderStates();

	Asura_Direct3D_Effect *pxDecalEffect = GetEffect();

#ifdef ASURA_TOOLS
	// ============================================================================================================
	// Tools Version - Render from the tools version of the static decal buffers
	// ============================================================================================================
	
	for(int i=0; i<Tools_StaticDecal_System::GetNumDecalBuffers();i++)
	{
		const Tools_StaticDecal_Buffer *pxBuffer = Tools_StaticDecal_System::GetDecalBufferByIndex(i);			
		int iIdx = Asura_Decal_Manager::GetMaterialIndex(pxBuffer->m_uDecalMaterial);				
		const Asura_DecalManager_Material *pxMgrMtl = Asura_Decal_Manager::GetDecalMaterialByIndex(iIdx);
		const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(iIdx+1);
		
		// Any polys? 
		if (pxBuffer->m_iNumPolys>0 && pxMtl)
		{	
			// Hmmm... this may cause problems in split screen. 
            if( (pxBuffer->m_iModuleIndex < 0) || Asura_Visibility_System::IsModuleLinkedInFromCurrentCamera(pxBuffer->m_iModuleIndex) )
			{		
				bool abAABBInFrustum[ASURA_MAX_NO_OF_CAMERAS];
				int  iNumCamerasVisibleFrom = 0;
				
				for(int j=0; j<Asura_Camera::GetNumberOfCameras(); j++)
				{
					Asura_Camera::SetCameraIndex(j);
					abAABBInFrustum[j] = Asura_Camera::IsAABBInFrustum(pxBuffer->m_xAABB);
					if (abAABBInFrustum[j])
					{					
						iNumCamerasVisibleFrom++;
					}
				}

				if (iNumCamerasVisibleFrom>0)
				{					
					if (pxMgrMtl->m_bStaticDecalBackFaceCull)
					{
						Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW); 
					}
					else
					{				
						Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE); 
					}
												

					Asura_Direct3D_Effects_System::BeginEffect(*pxDecalEffect);
					Asura_Assert(pxDecalEffect->IsD3DEffectValid(), "Decal effect invalid");

					if (pxDecalEffect->IsD3DEffectValid())
					{
						// Render them
						for(int j=0; j<Asura_Camera::GetNumberOfCameras(); j++)
						{
							Asura_Camera::SetCameraIndex(j);
							
							if (Asura_Camera::IsCameraActive())
							{
								if (abAABBInFrustum[j])
								{							
									Asura_Render::ApplyCameraSettings();

									for (int i = 0; i < 2; ++i)
									{
										Asura_Platform_Decal_System_SM3::SetupMaterial(i, pxMtl, afFadeValues);

										Asura_Direct3D::D3DDevice->SetVertexDeclaration(Asura_Platform_Decal_System_SM3::GetVertexDecl());
										pxDecalEffect->GetD3DEffect()->BeginPass(i);
										pxDecalEffect->GetD3DEffect()->CommitChanges();

										Asura_Direct3D::D3DDevice->DrawIndexedPrimitiveUP
											(
												D3DPT_TRIANGLELIST, 0, pxBuffer->m_iNumVerts,
																	   pxBuffer->m_iNumPolys, 
																		(void*)pxBuffer->m_piIndices, D3DFMT_INDEX32,
																		(void*)pxBuffer->m_pxVerts, sizeof(D3D_DecalVertex)
											);

										pxDecalEffect->GetD3DEffect()->EndPass();
									}
									if (Tools_Render_Options::GetWireframeMode()==TOOLS_WIREFRAME_ON || Tools_Render_Options::GetWireframeMode()==TOOLS_WIREFRAME_MIXED)
									{
										Asura_Platform_Decal_System_SM3::SetupMaterial(0, pxMtl, afFadeValues);

										pxDecalEffect->GetD3DEffect()->BeginPass(i);
										pxDecalEffect->GetD3DEffect()->CommitChanges();

										Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_WHITE);
										Asura_Render::CurrentStates.RequestWireFrameMode( true);

										Asura_Direct3D::D3DDevice->DrawIndexedPrimitiveUP
											(
											D3DPT_TRIANGLELIST, 0, pxBuffer->m_iNumVerts,
											pxBuffer->m_iNumPolys, 
											(void*)pxBuffer->m_piIndices, D3DFMT_INDEX32,
											(void*)pxBuffer->m_pxVerts, sizeof(D3D_DecalVertex)
											);

										pxDecalEffect->GetD3DEffect()->EndPass();
									}

								}
							}
						}
					}

					// Flush any effects we may still be using
					Asura_Direct3D_Effects_System::FlushEffects();	
				}
			}

		}
	}
#else
	// ============================================================================================================
	// Non-tools version: Render the exported platform data
	// ============================================================================================================

	for(int i=0; i<s_iNumStaticDecalBuffers; i++)
	{
		Asura_Direct3D_StaticDecalBuffer *pxBuffer = &s_pxStaticDecalBuffers[i];

		// Any polys? 
		if (pxBuffer->m_iNumVertices>0)
		{	

			if (pxBuffer->m_iDecalMaterialIndex<0)
			{			
				pxBuffer->m_iDecalMaterialIndex = Asura_Decal_Manager::GetMaterialIndex(pxBuffer->m_uDecalMaterialHash);						
			}

			if (pxBuffer->m_iDecalMaterialIndex>=0)
			{		
				const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(pxBuffer->m_iDecalMaterialIndex+1);
				const Asura_DecalManager_Material *pxMgrMtl = Asura_Decal_Manager::GetDecalMaterialByIndex(pxBuffer->m_iDecalMaterialIndex);

				// Hmmm... this may cause problems in split screen. 
                if( (pxBuffer->m_iModuleIndex < 0) || Asura_Visibility_System::IsModuleLinkedInFromCurrentCamera(pxBuffer->m_iModuleIndex) )
				{
					bool abAABBInFrustum[ASURA_MAX_NO_OF_CAMERAS];
					int  iNumCamerasVisibleFrom = 0;

					for(int j=0; j<Asura_Camera::GetNumberOfCameras(); j++)
					{
						Asura_Camera::SetCameraIndex(j);
						abAABBInFrustum[j] = Asura_Camera::IsAABBInFrustum(pxBuffer->m_xAABB);

						if (abAABBInFrustum[j])
						{					
							iNumCamerasVisibleFrom++;
						}
					}

					if (iNumCamerasVisibleFrom>0)
					{
						if (pxMgrMtl->m_bStaticDecalBackFaceCull)
						{
							Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW); 
						}
						else
						{				
							Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE); 
						}

						int iVertSize;

						Asura_Direct3D::D3DDevice->SetVertexDeclaration(GetVertexDecl());						
						iVertSize = sizeof(D3D_DecalVertex);					

						Asura_Direct3D::D3DDevice->SetStreamSource(0, pxBuffer->m_pxVertexBuffer, 0, iVertSize);
						Asura_Direct3D::D3DDevice->SetIndices(pxBuffer->m_pxIndexBuffer);

						Asura_Direct3D_Effects_System::BeginEffect(*pxDecalEffect);
						Asura_Assert(pxDecalEffect->IsD3DEffectValid(), "Decal effect invalid");

						if (pxDecalEffect->IsD3DEffectValid())
						{

							// Render them
							for(int j=0; j<Asura_Camera::GetNumberOfCameras(); j++)
							{
								Asura_Camera::SetCameraIndex(j);

								if (Asura_Camera::IsCameraActive() && abAABBInFrustum[j])
								{
									Asura_Render::ApplyCameraSettings();

									for (int i = 0; i < 2; ++i)
									{
										Asura_Platform_Decal_System_SM3::SetupMaterial(i, pxMtl, afFadeValues);

										pxDecalEffect->GetD3DEffect()->BeginPass(i);
										pxDecalEffect->GetD3DEffect()->CommitChanges();

										Asura_Direct3D::D3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,0, pxBuffer->m_iNumVertices, 0, pxBuffer->m_iNumIndices / 3);

										pxDecalEffect->GetD3DEffect()->EndPass();
									}
								}
							}
							
							// Flush any effects we may still be using
							Asura_Direct3D_Effects_System::FlushEffects();						
						}
					}
				}
			}
		}
	}


#endif // ASURA_TOOLS


	RestoreRenderStates();
}

#endif // ASURA_USE_STATIC_DECAL_SYSTEM



#endif // #ifdef ASURA_TARGET_SM3