// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Animation/Asura_PS2_Hierarchy.h"

// Engine independent.
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Stream.h"
#include "Core/Asura_TextureManagement.h"
#if _DEBUG
#include "Debugging/Asura_Logfile.h"
#include "Mesh/Asura_Env_System.h"
#endif

// Engine platform-specific.
#include "Animation/Asura_PS2_SkinTypes.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

static const float THRESHOLD_SCALE_FOR_1_VIEWPORT       = 0.6f;
static const float THRESHOLD_SCALE_FOR_2_VIEWPORTS      = 0.45f;
static const float THRESHOLD_SCALE_FOR_3_OR_4_VIEWPORTS = 0.3f;

// G L O B A L S /////////////////////////////////////////////////////////////

static bool g_bOnOffTransparencyConsideredOpaque = false;

// S T A T I C S /////////////////////////////////////////////////////////////

#if defined(ASURA_USE_MULTIPLE_HANDLERS) || defined(ASURA_USE_STREAMING)
Asura_TemplateHashTree<Asura_PS2_Skin_Handler>	Asura_PS2_SkinSystem::s_xHandlerTree;
Asura_PS2_Skin_Handler*							Asura_PS2_SkinSystem::s_pxCurrentHandler	= NULL;
#else
Asura_PS2_Skin_Handler							Asura_PS2_SkinSystem::s_xHandler;
#endif

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static inline const void* GetSkinLOD( const Asura_PS2_SkinEntry* pxSkinEntry, u_int uLOD, bool bIsSmooth )
{
    Asura_Assert( pxSkinEntry, "Invalid skin" );
    
    Asura_Assert( pxSkinEntry->m_usNumLODs, "No LODs loaded for this skin." );
    if( uLOD < pxSkinEntry->m_usNumLODs )
    {
        Asura_Assert( pxSkinEntry->m_bIsSmooth == bIsSmooth, "" );
        return pxSkinEntry->m_apvLODs[uLOD];
    }
    else
    {
        return NULL;
    }
}

static inline bool IsMaterialOpaque( u_int uMaterialFlags )
{
	if(g_bOnOffTransparencyConsideredOpaque)
	{
		// DJP 20/04/07 - allow on/off transparency to be treated as opaque
		return !(uMaterialFlags & AsuraMaterialFlag_PS2AdditiveAlpha);
	}

    return !(uMaterialFlags & (AsuraMaterialFlag_Transparent | AsuraMaterialFlag_PS2AdditiveAlpha));
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Asura_PS2_SkinSystem
//////////////////////////////////////////////////////////////////////////

void Asura_PS2_SkinSystem::Initialise( void )
{
	// Nothing to do here
}

void Asura_PS2_SkinSystem::Shutdown( void )
{
#if defined(ASURA_USE_STREAMING) || defined(ASURA_USE_MULTIPLE_HANDLERS)
	s_xHandlerTree.Reset();
	s_pxCurrentHandler	= NULL;
#else
	GetCurrentHandler().Reset();
#endif
}

void Asura_PS2_SkinSystem::Reset( void )
{
#if _DEBUG
    Asura_OutputToDebugger(( "Asura_PS2_SkinSystem::Reset(): %u skins at reset\n", CountSkins() ));
#endif
    
#if defined(ASURA_USE_STREAMING)
	s_xHandlerTree.Reset();
	s_pxCurrentHandler	= NULL;
#else
	GetCurrentHandler().Reset();
#endif
}

int Asura_PS2_SkinSystem::CountSkins()
{
#if defined(ASURA_USE_STREMING) || defined(ASURA_USE_MULTIPLE_HANDLERS)
	int iObjectCount = 0;

	Asura_TemplateHashTree_Iterator<Asura_PS2_Skin_Handler> xIt(s_xHandlerTree);
	for ( ; !xIt.Done(); xIt.Next() )
	{
		Asura_PS2_Skin_Handler* pxHandler = xIt.GetCurrent();
		iObjectCount += pxHandler->CountSkins();
	}

	return iObjectCount;
#else
	return GetCurrentHandler().CountSkins();
#endif
}

void Asura_PS2_SkinSystem::SetNumViewports( u_int uNumViewports )
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS) || defined(ASURA_USE_STREAMING)
	Asura_TemplateHashTree_Iterator<Asura_PS2_Skin_Handler> xIt(s_xHandlerTree);
	for ( ; !xIt.Done(); xIt.Next() )
	{
		Asura_PS2_Skin_Handler* pxHandler = xIt.GetCurrent();
		Asura_Assert(pxHandler,"");

		pxHandler->SetNumViewports( uNumViewports );
	}
#else
	s_xHandler.SetNumViewports( uNumViewports );
#endif
}

#if defined(ASURA_USE_MULTIPLE_HANDLERS) || defined(ASURA_USE_STREAMING)

void Asura_PS2_SkinSystem::ReparentHandler(Asura_PS2_Skin_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash)
{
	Asura_PS2_Skin_Handler*	pxParentHandler	= s_xHandlerTree.GetDataByHash(uNewParentHash);

	if(pxParentHandler != NULL)
	{
		xSourceHandler.m_xSkinStorage.SetParent(&(pxParentHandler->m_xSkinStorage));
	}
}

void Asura_PS2_SkinSystem::ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash)
{
	Asura_PS2_Skin_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uSourceHash);

	if(pxHandler != NULL)
	{
		ReparentHandler(*pxHandler, uNewParentHash);
	}
}

Asura_PS2_Skin_Handler* Asura_PS2_SkinSystem::CreateHandler (const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)
{
	Asura_PS2_Skin_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this zone?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHash);

		if (pxHandler != NULL)
		{
			ReparentHandler(*pxHandler, uParentHandlerHash);
		}
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an handler with this hash-ID.");
	}

	// Return the zone...
	return(pxHandler);
}


bool Asura_PS2_SkinSystem::DestroyHandler (const Asura_Hash_ID uHash)
{
	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

void Asura_PS2_SkinSystem::WriteHandlerReport(u_int uHandler, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 96;
	char szLogBuffer[uLOGBUFFERSIZE];

	Asura_PS2_Skin_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHandler);
	if (pxHandler)
	{
		if (!pxHandler->m_xSkinStorage.GetNumberOfNodes())
		{
			xLog << "PS2 Skin:\t\t\tNone\n";
		}
		else
		{
			Asura_TemplateHashTree_Iterator<Asura_PS2_SkinEntry> xIt(pxHandler->m_xSkinStorage);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_PS2_SkinEntry* pxSkinEntry = xIt.GetCurrent();
#ifdef _DEBUG
				Asura_PrintfToBfr( szLogBuffer, "PS2 Skin:\t(0x%08x)\t%s\n", uLOGBUFFERSIZE, true, pxSkinEntry->m_uNameHashID, pxSkinEntry->m_szName);
#else
				Asura_PrintfToBfr( szLogBuffer, "PS2 Skin:\t(0x%08x)\n", uLOGBUFFERSIZE, true, pxSkinEntry->m_uNameHashID );
#endif
				xLog << szLogBuffer;
			}		
		}
	}
	else
	{
		xLog << "PS2 Skin:\t\t\tNo Handler\n";
	}
}

#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS) || defined(ASURA_USE_STREAMING)

//////////////////////////////////////////////////////////////////////////
// Asura_PS2_SkinSystem_Handler
//////////////////////////////////////////////////////////////////////////

void Asura_PS2_Skin_Handler::Reset( void )
{
	m_xSkinStorage.Reset();
	m_fAppliedThresholdMultiplier = 0.0f;
}

int Asura_PS2_Skin_Handler::CountSkins()
{
	return m_xSkinStorage.GetNumberOfNodes();
}

bool Asura_PS2_Skin_Handler::AddHierSkinLOD( const char* szSkinName, void* pvHierSkinData, u_int uDataSizeInBytes )
{
    return AddSkinLODInternal( szSkinName, pvHierSkinData, uDataSizeInBytes, false );
}

bool Asura_PS2_Skin_Handler::AddSmoothSkinLOD( const char* szSkinName, void* pvSmoothSkinData, u_int uDataSizeInBytes )
{
    return AddSkinLODInternal( szSkinName, pvSmoothSkinData, uDataSizeInBytes, true );
}

const Asura_PS2_HierSkin_SkinHdr* Asura_PS2_Skin_Handler::GetHierSkinHdr( const Asura_PS2_SkinEntry* pxSkinEntry, const Asura_Vector_3& xPos, u_int uCamFlags, float* pfDistFromNearestCamSqd )
{
    u_int uLOD = CalculateLOD( pxSkinEntry, xPos, uCamFlags, pfDistFromNearestCamSqd );
    return static_cast<const Asura_PS2_HierSkin_SkinHdr*>( GetSkinLOD(pxSkinEntry, uLOD, false) );
}

const Asura_PS2_SmoothSkin_SkinHdr* Asura_PS2_Skin_Handler::GetSmoothSkinHdr( const Asura_PS2_SkinEntry* pxSkinEntry, const Asura_Vector_3& xPos, u_int uCamFlags, float* pfDistFromNearestCamSqd )
{
    u_int uLOD = CalculateLOD( pxSkinEntry, xPos, uCamFlags, pfDistFromNearestCamSqd );
    return static_cast<const Asura_PS2_SmoothSkin_SkinHdr*>( GetSkinLOD(pxSkinEntry, uLOD, true) );
}

Asura_PS2_SkinEntry* Asura_PS2_Skin_Handler::AddSkin( const char* szSkinName )
{
	Asura_Assert(m_fAppliedThresholdMultiplier == 0.0f,"Skins should be added before SetNumViewpoints() is called");

    Asura_Hash_ID uNameHashID = Asura_GetHashID( szSkinName );
 
	bool bDidExist = true;
	Asura_PS2_SkinEntry* pxSkinEntry = m_xSkinStorage.AddOrFindNode(uNameHashID,&bDidExist,true);
    if( pxSkinEntry && !bDidExist )
    {
        Asura_Assert( ASURA_HASH_ID_UNSET == pxSkinEntry->m_uNameHashID, "Hash clash?" );

#if _DEBUG
		if (pxSkinEntry->m_szName) delete[] pxSkinEntry->m_szName;
        pxSkinEntry->m_szName      = Asura_StringUtil::Duplicate( szSkinName );
#endif
        pxSkinEntry->m_uNameHashID = uNameHashID;

		while (pxSkinEntry->m_usNumLODs)
		{
			pxSkinEntry->m_usNumLODs--;
			Asura_DeallocateAlign( pxSkinEntry->m_apvLODs[pxSkinEntry->m_usNumLODs] );
		}
	}
	else
	{
#ifdef ASURA_ASSERT
		Asura_PS2_SkinEntry* pxLoadedSkinEntry = FindEntryForHashID( uNameHashID, false );
#ifdef _DEBUG
		Asura_Assertf(false,("Skin %s already loaded, ignoring.",pxSkinEntry->m_szName));
#else
		Asura_Assertf(false,("Skin %0x%x already loaded, ignoring.",pxSkinEntry->m_uNameHashID));
#endif
#endif
	}
    
    return pxSkinEntry;
}

void Asura_PS2_Skin_Handler::SetNumViewports( u_int uNumViewports )
{
	if (m_xSkinStorage.GetNumberOfNodes() > 0)
	{
		float fScaleSqd;
		switch( uNumViewports )
		{
			default:
				Asura_Assert( false, "Invalid number of viewports, initalising as 1." );
				// Fall through to a valid case just so fScale gets initialised.

			case 1:
				fScaleSqd = THRESHOLD_SCALE_FOR_1_VIEWPORT * THRESHOLD_SCALE_FOR_1_VIEWPORT;
				break;

			case 2:
				fScaleSqd = THRESHOLD_SCALE_FOR_2_VIEWPORTS * THRESHOLD_SCALE_FOR_2_VIEWPORTS;
				break;
		        
			case 3:
			case 4:
				fScaleSqd = THRESHOLD_SCALE_FOR_3_OR_4_VIEWPORTS * THRESHOLD_SCALE_FOR_3_OR_4_VIEWPORTS;
				break;
		}

		if (fScaleSqd != m_fAppliedThresholdMultiplier)
		{
			bool bRemoveOld = (m_fAppliedThresholdMultiplier > 0.0f);

			Asura_TemplateHashTree_Iterator<Asura_PS2_SkinEntry> xIt(m_xSkinStorage);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_PS2_SkinEntry* pxSkinEntry = xIt.GetCurrent();
				Asura_Assert(pxSkinEntry,"");

				for( u_int u = pxSkinEntry->m_usNumLODs; u--; ) 
				{ 
					if (bRemoveOld) pxSkinEntry->m_afThresholdsSqd[u] /= m_fAppliedThresholdMultiplier;
					pxSkinEntry->m_afThresholdsSqd[u] *= fScaleSqd;
#if _DEBUG
					pxSkinEntry->m_bHasBeenScaled = true;
#endif
				}	
			}

			m_fAppliedThresholdMultiplier = fScaleSqd;
		}
	}
}

int Asura_PS2_Skin_Handler::GetTextureID( Asura_Hash_ID uNameHashID )
{
	Asura_PS2_SkinEntry* pxSkinEntry = FindEntryForHashID( uNameHashID, false );
	return( pxSkinEntry ) ? pxSkinEntry->m_iTexID : ASURA_TEXTURE_NONE;
}

Asura_PS2_SkinEntry* Asura_PS2_Skin_Handler::FindEntryForHashID( Asura_Hash_ID uNameHashID, bool bAddIfNotFound )
{
	Asura_PS2_SkinEntry* pxSkinEntry = NULL;
	if (bAddIfNotFound)
	{
		pxSkinEntry = m_xSkinStorage.AddOrFindNode( uNameHashID );
		Asura_Assertf( pxSkinEntry, ("Couldn't find or add skin 0x%x", uNameHashID) );
	}
	else
	{
		pxSkinEntry = m_xSkinStorage.GetDataByHash( uNameHashID );
		Asura_Assertf( pxSkinEntry, ("Skin with hash ID 0x%08X not loaded.", uNameHashID) );
	}

    return pxSkinEntry;
}

bool Asura_PS2_Skin_Handler::IsOpaque( Asura_Hash_ID uNameHashID )
{
    const Asura_PS2_SkinEntry* pxSkinEntry = FindEntryForHashID( uNameHashID );
    return ( pxSkinEntry ) ? pxSkinEntry->m_bAllMaterialAreOpaque : true;
}

void Asura_PS2_SkinSystem::SetOnOffTransparencyConsideredOpaque(bool bSet)
{
	g_bOnOffTransparencyConsideredOpaque = bSet;
}

bool Asura_PS2_Skin_Handler::AddSkinLODInternal( const char* szSkinName, void* pvSkinData, u_int uDataSizeInBytes, bool bIsSmooth )
{
    Asura_Assert( pvSkinData, "No skin data." );

	bool bOK = false;
    Asura_Hash_ID uNameHashID = Asura_GetHashID( szSkinName );

	Asura_PS2_SkinEntry* pxSkinEntry = FindEntryForHashID( uNameHashID );
    Asura_Assert( pxSkinEntry, "LOD thresholds should come before skin data and the skin should get added there." );

	if (pxSkinEntry)
	{
		Asura_Assert( uNameHashID == pxSkinEntry->m_uNameHashID, "" );
#if _DEBUG
		Asura_Assert( !stricmp(szSkinName, pxSkinEntry->m_szName), "Hash clash?" );
#endif

		Asura_Assert( pxSkinEntry->m_usNumLODs < ASURA_PS2_SKINSYSTEM_MAX_LODS_PER_SKIN, "Too many LODs for this skin." );
		bool bFirstLODAdded = !pxSkinEntry->m_usNumLODs;

		if( bFirstLODAdded )
		{
			pxSkinEntry->m_bIsSmooth = bIsSmooth;
		}
		else
		{
			Asura_Assert( bIsSmooth == pxSkinEntry->m_bIsSmooth, "" );
		}

		bOK = ( bIsSmooth
						 ? PreprocessSmoothSkin(szSkinName, pvSkinData, uDataSizeInBytes, pxSkinEntry, bFirstLODAdded)
						 : PreprocessHierSkin(szSkinName, pvSkinData, uDataSizeInBytes, pxSkinEntry, bFirstLODAdded) );

		// The caller is expected to deallocate the data on failure so we'd
		// better lose our ptr to it in that case. However, we should increment
		// the LOD count regardless since otherwise thresholds and LOD ptrs will
		// get out of sync.

		// Need to check the thresholds array to see if we already have all our LODs (and therefore this is a duplicate load)
		if ((pxSkinEntry->m_usNumLODs == 0)||((pxSkinEntry->m_usNumLODs < ASURA_PS2_SKINSYSTEM_MAX_LODS_PER_SKIN) && (pxSkinEntry->m_afThresholdsSqd[pxSkinEntry->m_usNumLODs] > 0.0f)))
		{
			pxSkinEntry->m_apvLODs[pxSkinEntry->m_usNumLODs++] = ( bOK ? pvSkinData : NULL );
		}
		else
		{
#ifdef _DEBUG
			Asura_Assertf(false,("Ignoring duplicate LOD for skin %s",pxSkinEntry->m_szName));
#else
			Asura_Assertf(false,("Ignoring duplicate LOD for skin 0x%08x",pxSkinEntry->m_uNameHashID));
#endif
			bOK = false;
		}
	}

    return bOK;
}

u_int Asura_PS2_Skin_Handler::CalculateLOD( const Asura_PS2_SkinEntry* pxSkinEntry, const Asura_Vector_3& xPos, u_int uCamVisFlags, float* pfDistFromNearestCamSqd )
{
    Asura_Assert(pxSkinEntry,"");

	// Verify that LOD thresholds have been scaled appropriately.
#ifdef _DEBUG
	Asura_Assert( pxSkinEntry->m_bHasBeenScaled, "Skin LOD thresholds haven't been scaled. Are you calling SetNumViewports()?" );
#endif

    float fClosestDistSqd = Asura_Maths::LargeFloat;
    
    // Find closest cam.
    const int iNumCams = Asura_Camera::GetNumberOfCameras();
    for( int iCam = 0; iCam < iNumCams; ++iCam )
    {
        // If the hierarchy is visible to this camera, determine the LOD
        // required for it. Start at the highest LOD required so far.
        if( uCamVisFlags & (1u << iCam) )
        {
    		Asura_Camera::SetCameraIndex( iCam );
    
            const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();
            Asura_Vector_3 xDelta( *pxCamPos - xPos );
            float fCamLODMetric = Asura_Camera::GetLODMetric( iCam );
            float fDistSqd = (xDelta * xDelta) * fCamLODMetric * fCamLODMetric;

            fClosestDistSqd = Asura_Maths::Min( fClosestDistSqd, fDistSqd );
        }
    }

    if( pfDistFromNearestCamSqd ) { *pfDistFromNearestCamSqd = fClosestDistSqd; }
    
    // Find appropriate LOD.
    u_int uGreatestLOD = ( pxSkinEntry->m_usNumLODs ? pxSkinEntry->m_usNumLODs - 1 : 0 );
    for( ; uGreatestLOD; --uGreatestLOD )
    {
        if( fClosestDistSqd > pxSkinEntry->m_afThresholdsSqd[uGreatestLOD] )
        {
            break;
        }
    }

    return uGreatestLOD;
}

//static void DeallocateAllSkinEntries( void )
//{	
//    for( u_int u = 0; u < ASURA_MAX_NO_OF_HIERARCHY_SKINS; ++u )
//    {
//        SkinEntry& xSkinEntry = g_axSkinStorage[u];
//        if( ASURA_HASH_ID_UNSET != xSkinEntry.uNameHashID )
//        {
//            while( xSkinEntry.usNumLODs-- )
//            {
//                Asura_DeallocateAlign( xSkinEntry.apvLODs[xSkinEntry.usNumLODs] );
//            }
//#if _DEBUG
//            delete[] xSkinEntry.szName;
//#endif
//            xSkinEntry.uNameHashID = ASURA_HASH_ID_UNSET;
//        }
//    }
//}

bool Asura_PS2_Skin_Handler::PreprocessHierSkin( const char* szSkinName, void* pvHierSkinData, u_int uNumBytes, Asura_PS2_SkinEntry* pxSkinEntry, bool bIsFirstLOD )
{
    Asura_PS2_HierSkin_SkinHdr* pxSkinHdr = static_cast<Asura_PS2_HierSkin_SkinHdr*>( pvHierSkinData );

    // Fix up texture ID. This should always be 0 in the stored data since we
    // only support a single texture per skin.
    Asura_Assert( 0 == pxSkinHdr->sTextureID, "" );
    pxSkinHdr->sTextureID = Asura_Texture_Management::ConvertIndex( 0 );

    // Fix up pointers and addresses in skin header.
    {
        u_char* pucSecMatAddr = static_cast<u_char*>(pvHierSkinData) + reinterpret_cast<const u_int&>(pxSkinHdr->pxSectionMatHdrStart);
        pxSkinHdr->pxSectionMatHdrStart = reinterpret_cast<Asura_PS2_HierSkin_SectionMatHdr*>( pucSecMatAddr );

        pxSkinHdr->uStripDataStart = reinterpret_cast<const u_int&>(pvHierSkinData) + (pxSkinHdr->uStripDataStart << 4);
    }

    // Fix up material IDs.
    {
        const u_short* pusSectionMaterialCounts = reinterpret_cast<const u_short*>( pxSkinHdr + 1 );
	    Asura_PS2_HierSkin_SectionMatHdr* pxSectionMatHdr = pxSkinHdr->pxSectionMatHdrStart;
        for( u_int uSectionsLeft = pxSkinHdr->usSectionCount; uSectionsLeft; --uSectionsLeft )
        {
            for( u_int uMaterialsLeft = *pusSectionMaterialCounts++; uMaterialsLeft; --uMaterialsLeft )
            {
                pxSectionMatHdr->sMaterialID = Asura_Material_System::ConvertIndex( pxSectionMatHdr->sMaterialID );
                ++pxSectionMatHdr;
            }
        }
    }

    if( bIsFirstLOD )
    {
        pxSkinEntry->m_iTexID = pxSkinHdr->sTextureID;
        pxSkinEntry->m_bAllMaterialAreOpaque = IsMaterialOpaque( pxSkinHdr->uAllMaterialFlags );
    }
    else
    {
        // All LODs are assumed to use the same texture.
        Asura_Assert( pxSkinHdr->sTextureID == pxSkinEntry->m_iTexID, "" );
    }

    return true;
}

bool Asura_PS2_Skin_Handler::PreprocessSmoothSkin( const char* szSkinName, void* pvSmoothSkinData, u_int uNumBytes, Asura_PS2_SkinEntry* pxSkinEntry, bool bIsFirstLOD )
{
	// Grab ptr to skin hdr.
    Asura_PS2_SmoothSkin_SkinHdr* pxHdr = static_cast<Asura_PS2_SmoothSkin_SkinHdr*>( pvSmoothSkinData );

    const u_int uNumMaterials = pxHdr->usNumMaterials;
    Asura_Assert( uNumMaterials > 0, "" );

    // Convert material IDs.
    int* piMaterialIDs = reinterpret_cast<int*>( pxHdr + 1 );
    for( u_int u = 0; u < uNumMaterials; ++u )
    {
        piMaterialIDs[u] = Asura_Material_System::ConvertIndex( piMaterialIDs[u] );
    }

    // Setup the texture ID. We'll just take it from the first material. They
    // should all be using the same one anyway.
    const int iTexID = Asura_Material_System::GetTextureID( piMaterialIDs[0] );
    for( u_int u = 1; u < uNumMaterials; ++u )
    {
        Asura_Assert( iTexID == Asura_Material_System::GetTextureID(piMaterialIDs[u]), "" );
    }

    if( bIsFirstLOD )
    {
        pxSkinEntry->m_iTexID = iTexID;
        
        pxSkinEntry->m_bAllMaterialAreOpaque = true;
        for( u_int u = 0; u < uNumMaterials; ++u )
        {
            if( !IsMaterialOpaque(Asura_Material_System::GetMaterialFlags(piMaterialIDs[u])) )
            {
                pxSkinEntry->m_bAllMaterialAreOpaque = false;
                break;
            }
        }
    }
    else
    {
        // All LODs are assumed to use the same texture.
        Asura_Assert( iTexID == pxSkinEntry->m_iTexID, "" );
    }

    return true;
}