//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_DMA.h
//
// By: Morgan Parry.
//
// Copyright (c) 2001 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_DMA_H___
#define ___ASURA_PS2_DMA_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

#include <eeregs.h>

#include "Core/Asura_PlatformBasics.h"
#include "Debugging/Asura_Assert.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// Assertions will prevent functions below getting inlined so only do them
// in debug.
#if _DEBUG
#define Asura_PS2_DMA_Assert Asura_Assert
#else
#define Asura_PS2_DMA_Assert(_arg0,_arg1)
#endif

#define ASURA_PS2_DMA_SPR_FLAG (0x80000000)
#define ASURA_PS2_DMA_MEM_MASK (0x0FFFFFFF)

// MH copied from DMA tutor
#define	PY_DMA_REFE	(0)
#define	PY_DMA_CNT	(1 << 28)
#define	PY_DMA_NEXT	(2 << 28)
#define	PY_DMA_REF	(3 << 28)
#define	PY_DMA_REFS	(4 << 28)
#define	PY_DMA_CALL	(5 << 28)
#define	PY_DMA_RET	(6 << 28)
#define	PY_DMA_END	(7 << 28)

// see EE User's manual page 75.
#define CMD_S_32  (0)   //0000
#define CMD_S_16  (1)   //0001
#define CMD_S_8   (2)   //0010
#define CMD_V2_32 (4)   //0100
#define CMD_V2_16 (5)   //0101
#define CMD_V2_8  (6)   //0110
#define CMD_V3_32 (8)   //1000
#define CMD_V3_16 (9)   //1001
#define CMD_V3_8  (10)  //1010
#define CMD_V4_32 (12)  //1100
#define CMD_V4_16 (13)  //1101
#define CMD_V4_8  (14)  //1110
#define CMD_V4_5  (15)  //1111

#define ASURA_PS2_FROM_SPR_CHANNEL_MASK  (1 << Asura_PS2_DMA::CHANNEL_FROMSPR)
#define ASURA_PS2_TO_SPR_CHANNEL_MASK    (1 << Asura_PS2_DMA::CHANNEL_TOSPR)
#define ASURA_PS2_BOTH_SPR_CHANNELS_MASK (ASURA_PS2_FROM_SPR_CHANNEL_MASK | ASURA_PS2_TO_SPR_CHANNEL_MASK)

// (MP - 17.04.2002) This macro is for waiting for completion of a DMA
// transfer that has been instructed to report its status to cop0. If you
// want to be able to profile this, simply create an aptly function
// containing this macro and watch for that in the profiler. The uChannelBits
// argument should be a bitwise ORing of (1 << channel_num), eg (1 << 8) for
// fromSPR while the bDisable... argument should be true iff you want the 
// macro to turn cop0 reporting back off again for the relevant channel(s)
// after its done waiting.
// (MP - 05.06.2002) This macro seems huge but in release it should just be
// the 8 cycle bc0f loop unless the disable-cop0-reporting option is used or
// you're waiting for both a SPR channel and a non-SPR channel. In other
// cases, the rest should get optimised out, assuming the arguments can be
// determined at compile time.
#ifdef _MSC_VER
    // Just a dummy for MSVC build; our asm replacement macro in PS2_in_VC.h
    // can't be used inside a multi-line macro.
    #define ASURA_PS2_WAIT_FOR_DMA_COMPLETION( uChannelBits, bDisableCop0ReportingAfterwards ) { }
#else
    #ifndef __GNUC__
        #error I removed NOPs in the loop below because GCC adds them anyway. Other compilers may need them put back in.
    #endif

    #define ASURA_PS2_WAIT_FOR_DMA_COMPLETION( uChannelBits, bDisableCop0ReportingAfterwards ) \
    { \
        /* Verify that cop0 reporting was setup correctly for the channel(s) \
           we're interested in. Note that if the relevant D_PCR bit isn't \
           set, the bf0? loop below will simply terminate immediately. */ \
        Asura_PS2_DMA_Assert( *D_PCR & (uChannelBits), "" ); \
        \
        /* Wait for transfer to complete using bc0f if any non-SPR channels \
           are involved. Note that this isn't necessary if we're only waiting \
           for SPR channels because checking Dx_CHCR doesn't tie up the bus \
           for them - as long as cycle stealing is off, which it should be. */ \
        const u_int uNonSPRChannelBits = (uChannelBits) & ~ASURA_PS2_BOTH_SPR_CHANNELS_MASK; \
        if( uNonSPRChannelBits ) \
        { \
            /* The two 0: lines below are local labels. The 0f and 0b \
               arguments to the bc0? instructions mean local label 0 \
               forwards and local label 0 backwards respectively. */ \
            /* (MP - 05.07.2002) GCC adds nops after the bc0? instructions \
               anyway so there's no need to put them in ourselves. */ \
	        asm __volatile__( " 0:          \n"    \
	                          "     bc0t 0f \n"    \
	                          "     bc0t 0f \n"    \
	                          "     bc0t 0f \n"    \
	                          "     bc0f 0b \n"    \
                              " 0:          \n" ); \
            \
            /* Verify that the transfer(s) we were waiting for have \
               finished. */ \
            Asura_PS2_DMA_Assert( (*D_STAT & uNonSPRChannelBits) == uNonSPRChannelBits, "" ); \
        } \
        \
        /* If we're waiting for any SPR channels, verify that the transfer \
           has actually finished. Sony have confirmed it's possible for the \
           bc0f loop to fail sometimes with SPR transfers. See the "Warning \
           about BC0F when DMAing to SPR" thread in the sce.dev.toptips \
           newsgroup. */ \
        if( (uChannelBits) & ASURA_PS2_BOTH_SPR_CHANNELS_MASK ) \
        { \
            if( (uChannelBits) & ASURA_PS2_FROM_SPR_CHANNEL_MASK ) \
            { \
                while( *D8_CHCR & D_CHCR_STR_M ) { } \
            } \
            if( (uChannelBits) & ASURA_PS2_TO_SPR_CHANNEL_MASK ) \
            { \
                while( *D9_CHCR & D_CHCR_STR_M ) { } \
            } \
        } \
        \
        /* Disable cop0 reporting for relevant channel(s) now we're done? */ \
        if( bDisableCop0ReportingAfterwards ) \
        { \
            *D_PCR &= ~(uChannelBits); \
            asm __volatile__( "sync.l" ); \
        } \
    }
#endif

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_PS2_DmaTag
{

    public:

        // T Y P E S /////////////////////////////////////////////////////////////

        // Tag ID. In the comments, the transfer specifications are inclusive.
        // The letters are: b = QW of this DMA tag, ie. base QW; n = QW count
        // specified in the tag's uQWC field; a = the address specified in the
        // uAddr field. See also section 3.6 of the EE User's Manual.
        enum ID
        {
            ID_REFE,        // Transfers (a) -> (a+n-1), and then ends transfer, setting Dx_CHCR.STR to 0.
            ID_CNT,         // Transfers (b+1) -> (b+n). Next tag at (b+n+1).
            ID_NEXT,        // Transfers (b+1) -> (b+n). Next tag at (a).
            ID_REF,         // Transfers (a) -> (a+n-1). Next tag at (b+1).
            ID_REFS,        // Transfers (a) -> (a+n-1), while controlling stalls. Next tag at (b+1). Only effective on VIF1, GIF1 and SIF1 channels.
            ID_CALL,        // Transfers (b+1) -> (b+n). See User's Manual for next tag behaviour. Only effective on VIF1, GIF1 and SIF1 channels.
            ID_RET,         // Transfers (b+1) -> (b+n). See User's Manual for next tag behaviour. Only effective on VIF1, GIF1 and SIF1 channels.
            ID_END,         // Transfers (b+1) -> (b+n), and then ends transfers, setting Dx_CHCR.STR to 0.
        };

        // Priority control enable.
        enum PCE
        {
            PCE_NOCHANGE,   // Nothing performed - D_PCR.PCE unchanged.
            PCE_RESERVED,   // Do not use.
            PCE_DISABLE,    // Priority setting disabled, ie. D_PCR.PCE = 0.
            PCE_ENABLE,     // Priority setting enabled, ie. D_PCR.PCE = 1.
        };

        // C O N S T A N T S /////////////////////////////////////////////////////

        enum
        {
            SHIFT_QWC  = 0,
            SHIFT_PCE  = 26,
            SHIFT_ID   = 28,
            SHIFT_IRQ  = 31,
            SHIFT_ADDR = 32,
            SHIFT_SPR  = 63,
        };
    
        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////////

        Asura_PS2_DmaTag( void ) { }
        explicit Asura_PS2_DmaTag( u_long128 uValue ) { m_u128 = uValue; }
    
        // O P E R A T O R S /////////////////////////////////////////////////////

        Asura_PS2_DmaTag& operator=( const Asura_PS2_DmaTag& xOther )
        {
            m_u128 = xOther.m_u128;
            return *this;
        }

        // F U N C T I O N S /////////////////////////////////////////////////////

        inline void Set( u_int uQWC, PCE ePCE, ID eID, bool bIRQ, u_int uAddr, bool bSPR )
        {
            // Check the scratchpad flag is set correctly.
            Asura_PS2_DMA_Assert( bSPR == ((uAddr & ASURA_PS2_SPR_MEM_BASE) == ASURA_PS2_SPR_MEM_BASE), "Scratchpad flag is incorrect." );
        
            // The DMAC doesn't seem to take too kindly to having the uncached
            // or uncached unaccelerated flags set in the address for transfers
            // from main memory so let's check they've been masked off.
            if( !bSPR )
            {
                Asura_PS2_DMA_Assert( !(uAddr & (ASURA_PS2_UNCACHED_MEM_FLAG | ASURA_PS2_UNCACHED_ACCELERATED_MEM_FLAG)), "You need to mask off the uncached / uncached accelerated flags." );
            }

            m_uTagPart = (static_cast<u_long>(uQWC)  << SHIFT_QWC )
                       | (static_cast<u_long>(ePCE)  << SHIFT_PCE )
                       | (static_cast<u_long>(eID)   << SHIFT_ID  )
                       | (static_cast<u_long>(bIRQ)  << SHIFT_IRQ )
                       | (static_cast<u_long>(uAddr) << SHIFT_ADDR)
                       | (static_cast<u_long>(bSPR)  << SHIFT_SPR );
        }

        inline void SetSpareWords( u_int uWord2, u_int uWord3 )
        {
            m_uSpareDoubleWord = static_cast<u_long>(uWord2) | (static_cast<u_long>(uWord3) << 32u);
        }

        inline void Set( u_int uQWC, PCE ePCE, ID eID, bool bIRQ, u_int uAddr, bool bSPR, u_int uWord2, u_int uWord3 )
        {
            Set( uQWC, ePCE, eID, bIRQ, uAddr, bSPR );
            SetSpareWords( uWord2, uWord3 );
        }

        inline void SetSpareWordsToNOPs( void )
        {
            m_uSpareDoubleWord = 0uL;
        }

        inline void SetWithNOPs( u_int uQWC, PCE ePCE, ID eID, bool bIRQ, u_int uAddr, bool bSPR )
        {
            Set( uQWC, ePCE, eID, bIRQ, uAddr, bSPR ); 
            SetSpareWordsToNOPs();
        }

        // D A T A ///////////////////////////////////////////////////////////////

        union
        {
            struct
            {
                u_int m_uQWC    : 16;
                u_int _padding1 : 10;
                PCE   m_ePCE    : 2;
                ID    m_eID     : 3;
                u_int m_bIRQ    : 1;
                u_int m_uAddr   : 31;
                u_int m_bSPR    : 1;

                // Additional words that can be used for VIF codes, etc., if tag
                // transfer is enabled.
                u_int m_uWord2;
                u_int m_uWord3;
            };

            struct
            {
                u_long m_uTagPart;
                u_long m_uSpareDoubleWord;
            };

            u_long128 m_u128;

            u_long m_auDoubleWords[2];

            u_int m_auWords[4];
        };

};
// End class Asura_PS2_DmaTag.

// General DMA class providing functionality for simple transfers. Please 
// bear the usual DMA transfer restrictions/recommendations in mind:
//
//     (i) All transfer addresses (src/dest) must be QW-aligned.
//    (ii) Large transfers benefit from addresses being on 128-byte (ie. 8 QW)
//         boundaries.
//   (iii) Concurrent memory accesses that cause cache misses will slow
//         transfers.
class Asura_PS2_DMA
{

    public:

        // T Y P E S /////////////////////////////////////////////////////////
        
        // To save including libdma.h just for the SCE_DMA_xxx #defines...
        enum CHANNEL
        {
            CHANNEL_VIF0,
            CHANNEL_VIF1,
            CHANNEL_GIF,
            CHANNEL_FROMIPU,
            CHANNEL_TOIPU,
            CHANNEL_SIF0,
            CHANNEL_SIF1,
            CHANNEL_SIF2,
            CHANNEL_FROMSPR,
            CHANNEL_TOSPR,
        };

        // C L A S S   M E T H O D S /////////////////////////////////////////

        // The uChannelBits argument should be a bitwise ORing of
        // (1 << CHANNEL_xxx) for relevant channels, where CHANNEL_xxx comes
        // from the enum above.
        static inline void ReportCompletionOnCop0( unsigned uChannelBits )
        {
            // Request completion report on cop0.
            *D_PCR = uChannelBits;

            // Clear transfer completion bit in D_STAT. Note that this 
            // assignment only changes the relevant bits, not the whole
            // register, because of that way that D_STAT works. (See the EE
            // User's Manual for more details.)
            *D_STAT = uChannelBits;
        }
        
        static inline volatile u_int* GetPtrToCHCR( CHANNEL eChannel )
        {
            // The compiler should be able to optimise all of this out, as
            // the channel should be known at compile time. Note that the
            // channel addresses are irregular so we can't just add an offset
            // based on the channel number.
            switch( eChannel )
            {
                default:
                    Asura_PS2_DMA_Assert( false, "Invalid DMA channel." );

                case CHANNEL_VIF0:    return D0_CHCR;
                case CHANNEL_VIF1:    return D1_CHCR;
                case CHANNEL_GIF:     return D2_CHCR;
                case CHANNEL_FROMIPU: return D3_CHCR;
                case CHANNEL_TOIPU:   return D4_CHCR;
                case CHANNEL_SIF0:    return D5_CHCR;
                case CHANNEL_SIF1:    return D6_CHCR;
                case CHANNEL_SIF2:    return D7_CHCR;
                case CHANNEL_FROMSPR: return D8_CHCR;
                case CHANNEL_TOSPR:   return D9_CHCR;
            }
        }

        static inline volatile u_int* GetPtrToMADR( CHANNEL eChannel )
        {
            // See comment in GetPtrToCHCR().
            switch( eChannel )
            {
                default:
                    Asura_PS2_DMA_Assert( false, "Invalid DMA channel." );

                case CHANNEL_VIF0:    return D0_MADR;
                case CHANNEL_VIF1:    return D1_MADR;
                case CHANNEL_GIF:     return D2_MADR;
                case CHANNEL_FROMIPU: return D3_MADR;
                case CHANNEL_TOIPU:   return D4_MADR;
                case CHANNEL_SIF0:    return D5_MADR;
                case CHANNEL_SIF1:    return D6_MADR;
                case CHANNEL_SIF2:    return D7_MADR;
                case CHANNEL_FROMSPR: return D8_MADR;
                case CHANNEL_TOSPR:   return D9_MADR;
            }
        }

        static inline volatile u_int* GetPtrToQWC( CHANNEL eChannel )
        {
            // See comment in GetPtrToCHCR().
            switch( eChannel )
            {
                default:
                    Asura_PS2_DMA_Assert( false, "Invalid DMA channel." );

                case CHANNEL_VIF0:    return D0_QWC;
                case CHANNEL_VIF1:    return D1_QWC;
                case CHANNEL_GIF:     return D2_QWC;
                case CHANNEL_FROMIPU: return D3_QWC;
                case CHANNEL_TOIPU:   return D4_QWC;
                case CHANNEL_SIF0:    return D5_QWC;
                case CHANNEL_SIF1:    return D6_QWC;
                case CHANNEL_SIF2:    return D7_QWC;
                case CHANNEL_FROMSPR: return D8_QWC;
                case CHANNEL_TOSPR:   return D9_QWC;
            }           
        }

        static inline volatile u_int* GetPtrToTADR( CHANNEL eChannel )
        {
            // Note that some channels do not have a TADR register. See also
            // comment in GetPtrToCHCR().
            switch( eChannel )
            {
                case CHANNEL_FROMIPU:
                case CHANNEL_SIF0:
                case CHANNEL_SIF2:
                case CHANNEL_FROMSPR:
                default:
                    Asura_PS2_DMA_Assert( false, "Invalid DMA channel." );

                case CHANNEL_VIF0:  return D0_TADR;
                case CHANNEL_VIF1:  return D1_TADR;
                case CHANNEL_GIF:   return D2_TADR;
                case CHANNEL_TOIPU: return D4_TADR;
                case CHANNEL_SIF1:  return D6_TADR;
                case CHANNEL_TOSPR: return D9_TADR;
            }           
        }

        static inline void SetCHCR( CHANNEL eChannel, u_int uValue )
        {
            *GetPtrToCHCR(eChannel) = uValue;
        }

        static inline void SetMADR( CHANNEL eChannel, u_int uAddr, bool bFromSPR )
        {
            Asura_PS2_DMA_Assert( !(uAddr & 0x0000000Fu), "Memory address must be QW-aligned." );
            Asura_PS2_DMA_Assert( bFromSPR || !(uAddr & 0x30000000u), "Uncached / uncached-accelerated bits should be masked off." );

            // Set SPR flag if necessary.
            if( bFromSPR )
            {
                uAddr |= 0x80000000;
            }

            *GetPtrToMADR(eChannel) = uAddr;
        }

        static inline void SetSADR( CHANNEL eChannel, u_int uAddr )
        {
            Asura_PS2_DMA_Assert( !(uAddr & 0x0fffc00f), "Invalid SPR addr." );

            // Most channels don't have the SADR register.
            Asura_PS2_DMA_Assert( CHANNEL_FROMSPR == eChannel || CHANNEL_TOSPR == eChannel, "Specified channel doesn't have SADR register." );

            *(CHANNEL_FROMSPR == eChannel ? D8_SADR : D9_SADR) = uAddr;
        }

        static inline void SetTADR( CHANNEL eChannel, u_int uAddr )
        {
            Asura_PS2_DMA_Assert( !(uAddr & 0x0000000f), "Tag address must be QW-aligned." );
            *GetPtrToTADR(eChannel) = uAddr;
        }

        static inline void SetQWC( CHANNEL eChannel, u_int uCount )
        {
            Asura_PS2_DMA_Assert( !(uCount & 0xffff0000), "Count must be <= 65535 as QWC register only provides 16 bits." );
            *GetPtrToQWC(eChannel) = uCount;
        }   

        // Handles standard chain-mode transfers from memory. Makes some 
        // assumptions about the transfer; in particular, it assumes the call
        // tag doesn't push an address and that the DMA tag(s) should be sent.
        static inline void ChainTransferFromMem( CHANNEL eChannel, const void* pvSrcAddr, bool bFromSPR, bool bReportOnCop0, bool bStall )
        {
            // Check the caller hasn't requested a channel where transfers
            // from memory are impossible.
            Asura_PS2_DMA_Assert( CHANNEL_FROMIPU != eChannel && CHANNEL_SIF0 != eChannel && CHANNEL_FROMSPR != eChannel, "Can't transfer from memory on specified channel." );

            // Verify that the specified channel is not currently in use. If
            // it is, things are about to break horribly.
            Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(eChannel) & D_CHCR_STR_M), "Specified DMA channel is already transferring data." );

            // Reset transferred QW count.
            SetQWC( eChannel, 0 );

            // Set tag addr.
            {
                u_int uSrcAddr = reinterpret_cast<u_int>( pvSrcAddr );
                Asura_PS2_DMA_Assert( ((uSrcAddr & ASURA_PS2_SPR_MEM_BASE) != 0) == bFromSPR, "Specified SPR flag was incorrect." );
                if( bFromSPR )
                {
                    uSrcAddr |= ASURA_PS2_DMA_SPR_FLAG;
                }

                SetTADR( eChannel, uSrcAddr );
            }

            // Request transfer completion notification on cop0 if necessary.
            if( bReportOnCop0 ) { ReportCompletionOnCop0(1 << eChannel); }

            // (MP - 05.07.2002) This appears to be necessary to guarantee
            // that all registers - in particular the relevant bits in D_PCR
            // and D_STAT that setup completion notification on cop0 - are
            // written before Dx_CHCR.
            asm __volatile__( "sync.l" );

            // Setup channel control register.
            SetCHCR( eChannel, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
                               (1 << D_CHCR_MOD_O) |  // MOD: chain mode.
                               (0 << D_CHCR_ASP_O) |  // ASP: assume call tag doesn't push any addresses.
                               (1 << D_CHCR_TTE_O) |  // TTE: transfer tags.
                               (1 << D_CHCR_TIE_O) |  // TIE: allow tags to dictate interrupt setting.
                               (1 << D_CHCR_STR_O) ); // STR: start DMA.

            // Ensure this is all written back to registers before we start 
            // waiting for completion notification on cop0. Note that both of
            // these synchronisation instructions are required, as noted in
            // DMA FAQ "faq4.txt".
            asm __volatile__( "sync.p" );
            asm __volatile__( "sync.l" );

            // Wait for transfer to complete before returning?
            if( bStall )
            {
                Asura_PS2_DMA_Assert( bReportOnCop0, "Can't stall if you didn't request transfer completion notification." );
                ASURA_PS2_WAIT_FOR_DMA_COMPLETION( 1 << eChannel, false ); //true );
            }
        }

        static inline void NormalTransferFromMem( CHANNEL eChannel, const void* pvSrcAddr, u_int uNumQWs, bool bFromSPR, bool bReportOnCop0, bool bStall )
        {
            // Check the caller hasn't requested a channel where transfers
            // from memory are impossible.
            Asura_PS2_DMA_Assert( CHANNEL_FROMIPU != eChannel && CHANNEL_SIF0 != eChannel && CHANNEL_FROMSPR != eChannel, "Can't transfer from memory on specified channel." );

            // Verify that the specified channel is not currently in use. If
            // it is, things are about to break horribly.
            Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(eChannel) & D_CHCR_STR_M), "Specified DMA channel is already transferring data." );

            // Verify they're not trying to transfer 0 QWs. This is illegal
            // with normal mode transfers. See the Dn_QWC documentation in
            // the EE User's Manual.
            Asura_PS2_DMA_Assert( uNumQWs, "Bad QWC." );

            // Set src addr.
            SetMADR( eChannel, reinterpret_cast<u_int>(pvSrcAddr), bFromSPR );

            // Set transfer size.
            SetQWC( eChannel, uNumQWs );
            
            // Request transfer completion notification on cop0 if necessary.
            if( bReportOnCop0 ) { ReportCompletionOnCop0(1 << eChannel); }

            // (MP - 05.07.2002) This appears to be necessary to guarantee
            // that all registers - in particular the relevant bits in D_PCR
            // and D_STAT that setup completion notification on cop0 - are
            // written before Dx_CHCR.
            asm __volatile__( "sync.l" );

            // Start transfer.
            SetCHCR( eChannel, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
                               (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                               (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                               (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                               (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                               (1 << D_CHCR_STR_O) ); // STR: start DMA.

            // (MP - 22.04.2002) As stated in one of the DMA faqs (the file
            // is called faq4.txt), both sync.p and sync.l instructions are
            // necessary before entering the corresponding bc0x loop.
            asm __volatile__( "sync.p" );
            asm __volatile__( "sync.l" );

            // Wait for transfer to complete before returning?
            if( bStall )
            {
                Asura_PS2_DMA_Assert( bReportOnCop0, "Can't stall if you didn't request transfer completion notification." );
                ASURA_PS2_WAIT_FOR_DMA_COMPLETION( 1 << eChannel, false ); //true );
            }
        }
        
        static inline void TransferFromSPR( u_int uSrcAddr, void* pvDest, u_int uNumQWs, bool bReportOnCop0, bool bStall )
        {
            // Verify that the fromSPR channel is not currently in use. If it
            // is, things are about to break horribly.
            Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(CHANNEL_FROMSPR) & D_CHCR_STR_M), "The fromSPR DMA channel is already transferring data." );

            // Validate and set src SPR addr.
            SetSADR( CHANNEL_FROMSPR, uSrcAddr );

            // Validate and set dest mem addr.
            SetMADR( CHANNEL_FROMSPR, reinterpret_cast<u_int>(pvDest), false );

            // Set transfer size.
            Asura_PS2_DMA_Assert( ((uSrcAddr & 0x0fffffff) >> 4) + uNumQWs <= 1024, "Specified region to transfer exceeds SPR bounds." );
            SetQWC( CHANNEL_FROMSPR, uNumQWs );

            // Request transfer completion notification on cop0 if necessary.
            if( bReportOnCop0 ) { ReportCompletionOnCop0(1 << CHANNEL_FROMSPR); }

            // (MP - 05.07.2002) This appears to be necessary to guarantee
            // that all registers - in particular the relevant bits in D_PCR
            // and D_STAT that setup completion notification on cop0 - are
            // written before Dx_CHCR.
            asm __volatile__( "sync.l" );

            // Start transfer.
            SetCHCR( CHANNEL_FROMSPR, (0 << D_CHCR_DIR_O) |  // DIR: to memory.
                                      (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                                      (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                                      (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                                      (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                                      (1 << D_CHCR_STR_O) ); // STR: start DMA.

            // (MP - 22.04.2002) As stated in one of the DMA faqs (the file
            // is called faq4.txt), both sync.p and sync.l instructions are
            // necessary before entering the corresponding bc0x loop.
            asm __volatile__( "sync.p" );
            asm __volatile__( "sync.l" );

            // Wait for transfer to complete before returning?
            if( bStall )
            {
                Asura_PS2_DMA_Assert( bReportOnCop0, "Can't stall if you didn't request transfer completion notification." );
                ASURA_PS2_WAIT_FOR_DMA_COMPLETION( 1 << CHANNEL_FROMSPR, false ); //true );
            }
        }

        static inline void TransferFromSPR( const void* pvSrc, void* pvDest, u_int uNumQWs, bool bReportOnCop0, bool bStall )
        {
            const u_int uSrcAddr = reinterpret_cast<u_int>( pvSrc );
            Asura_PS2_DMA_Assert( uSrcAddr & ASURA_PS2_SPR_MEM_BASE, "Invalid SPR addr." );
            TransferFromSPR( uSrcAddr, pvDest, uNumQWs, bReportOnCop0, bStall );
        }

        static inline void TransferToSPR( const void* pvSrc, u_int uDestAddr, u_int uNumQWs, bool bReportOnCop0, bool bStall )
        {
            // Verify that the specified region in SPR has been allocated 
            // properly.
            Asura_PS2_DMA_Assert( Asura_PS2_DebugSPR::IsAllocated((uDestAddr & 0x0fffffff) >> 4, uNumQWs), "Specified region of SPR wasn't allocated with Asura_PS2_SPR::Alloc()." );

            // Verify that the toSPR channel is not currently in use. If it
            // is, things are about to break horribly.
            Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(CHANNEL_TOSPR) & D_CHCR_STR_M), "The toSPR DMA channel is already transferring data." );

            // Set src mem addr.
            SetMADR( CHANNEL_TOSPR, reinterpret_cast<u_int>(pvSrc), false );

            // Set dest SPR addr.
            SetSADR( CHANNEL_TOSPR, uDestAddr );

            // Set transfer size.
            Asura_PS2_DMA_Assert( ((uDestAddr & 0x0fffffff) >> 4) + uNumQWs <= 1024, "Transfer will overflow SPR." );
            SetQWC( CHANNEL_TOSPR, uNumQWs );

            // Request transfer completion notification on cop0 if necessary.
            if( bReportOnCop0 ) { ReportCompletionOnCop0(1 << CHANNEL_TOSPR); }

            // (MP - 05.07.2002) This appears to be necessary to guarantee
            // that all registers - in particular the relevant bits in D_PCR
            // and D_STAT that setup completion notification on cop0 - are
            // written before Dx_CHCR.
            asm __volatile__( "sync.l" );

            // Start transfer.
            SetCHCR( CHANNEL_TOSPR, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
                                    (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                                    (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                                    (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                                    (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                                    (1 << D_CHCR_STR_O) ); // STR: start DMA.

            // (MP - 22.04.2002) As stated in one of the DMA faqs (the file
            // is called faq4.txt), both sync.p and sync.l instructions are
            // necessary before entering the corresponding bc0x loop.
            asm __volatile__( "sync.p" );
            asm __volatile__( "sync.l" );

            // Wait for transfer to complete before returning?
            if( bStall )
            {
                Asura_PS2_DMA_Assert( bReportOnCop0, "Can't stall if you didn't request transfer completion notification." );
                ASURA_PS2_WAIT_FOR_DMA_COMPLETION( 1 << CHANNEL_TOSPR, false ); //true );
            }
        }

        static inline void TransferToSPR( const void* pvSrc, void* pvDest, u_int uNumQWs, bool bReportOnCop0, bool bStall )
        {
            const u_int uDestAddr = reinterpret_cast<u_int>( pvDest );
            Asura_PS2_DMA_Assert( uDestAddr & ASURA_PS2_SPR_MEM_BASE, "Invalid SPR addr." );
            TransferToSPR( pvSrc, uDestAddr, uNumQWs, bReportOnCop0, bStall );
        }


		static inline void TransferToThenFromSPR( u_int uToSrcAddr,	u_int uToDestAddr,	 u_int uToNumQWs, 
												 u_int uFromSrcAddr,	u_int uFromDestAddr, u_int uFromNumQWs )
		{																				
            // Verify that the specified regions in SPR have been allocated 
            // properly.
            Asura_PS2_DMA_Assert( Asura_PS2_DebugSPR::IsAllocated((uToDestAddr & 0x0fffffff) >> 4, uToNumQWs), "Specified region of SPR wasn't allocated with Asura_PS2_SPR::Alloc()." );
            Asura_PS2_DMA_Assert( Asura_PS2_DebugSPR::IsAllocated((uFromSrcAddr & 0x0fffffff) >> 4, uFromNumQWs), "Specified region of SPR wasn't allocated with Asura_PS2_SPR::Alloc()." );
            // Verify that the toSPR and fromSPR channels are not currently in use. If it is, things are about to break horribly.
            Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(CHANNEL_TOSPR) & D_CHCR_STR_M), "The toSPR DMA channel is already transferring data." );
			Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(CHANNEL_FROMSPR) & D_CHCR_STR_M), "The fromSPR DMA channel is already transferring data." );
            // Set src mem addr.
            SetMADR( CHANNEL_TOSPR,	  uToSrcAddr,    false );
			SetMADR( CHANNEL_FROMSPR, uFromDestAddr, false );
            // Set dest SPR addr.
            SetSADR( CHANNEL_TOSPR,   uToDestAddr  );
			SetSADR( CHANNEL_FROMSPR, uFromSrcAddr );
            // Set transfer size.
            Asura_PS2_DMA_Assert( ((uToDestAddr & 0x0fffffff) >> 4) + uToNumQWs <= 1024, "Transfer will overflow SPR." );
			Asura_PS2_DMA_Assert( ((uFromSrcAddr & 0x0fffffff) >> 4) + uFromNumQWs <= 1024, "Transfer will overflow SPR." );
            SetQWC( CHANNEL_TOSPR,	 uToNumQWs );
			SetQWC( CHANNEL_FROMSPR, uFromNumQWs );
            // (MP - 05.07.2002) This appears to be necessary to guarantee
            // that all registers - in particular the relevant bits in D_PCR
            // and D_STAT that setup completion notification on cop0 - are
            // written before Dx_CHCR.
			asm __volatile__( "sync.p" );
            asm __volatile__( "sync.l" );
			// Both DMA's are read to go... The following to start dma writes will sit in the WBB until the sync.l
			// at which point they get flushed together and avoid the second write stalling on the first dma
            SetCHCR( CHANNEL_TOSPR, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
                                    (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                                    (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                                    (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                                    (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                                    (1 << D_CHCR_STR_O) ); // STR: start DMA.
            SetCHCR( CHANNEL_FROMSPR, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
									  (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                                      (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                                      (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                                      (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                                      (1 << D_CHCR_STR_O) ); // STR: start DMA.
			asm __volatile__( "sync.p" );												
			asm __volatile__( "sync.l" );
		}

		static inline void TransferFromThenToSPR( u_int uFromSrcAddr,	u_int uFromDestAddr, u_int uFromNumQWs,
												  u_int uToSrcAddr,		u_int uToDestAddr,	 u_int uToNumQWs )
		{																				
            // Verify that the specified regions in SPR have been allocated 
            // properly.
            Asura_PS2_DMA_Assert( Asura_PS2_DebugSPR::IsAllocated((uToDestAddr & 0x0fffffff) >> 4, uToNumQWs), "Specified region of SPR wasn't allocated with Asura_PS2_SPR::Alloc()." );
            Asura_PS2_DMA_Assert( Asura_PS2_DebugSPR::IsAllocated((uFromSrcAddr & 0x0fffffff) >> 4, uFromNumQWs), "Specified region of SPR wasn't allocated with Asura_PS2_SPR::Alloc()." );
            // Verify that the toSPR and fromSPR channels are not currently in use. If it is, things are about to break horribly.
            Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(CHANNEL_TOSPR) & D_CHCR_STR_M), "The toSPR DMA channel is already transferring data." );
			Asura_PS2_DMA_Assert( !(*GetPtrToCHCR(CHANNEL_FROMSPR) & D_CHCR_STR_M), "The fromSPR DMA channel is already transferring data." );
            // Set src mem addr.
            SetMADR( CHANNEL_TOSPR,	  uToSrcAddr,    false );
			SetMADR( CHANNEL_FROMSPR, uFromDestAddr, false );
            // Set dest SPR addr.
            SetSADR( CHANNEL_TOSPR,   uToDestAddr  );
			SetSADR( CHANNEL_FROMSPR, uFromSrcAddr );
            // Set transfer size.
            Asura_PS2_DMA_Assert( ((uToDestAddr & 0x0fffffff) >> 4) + uToNumQWs <= 1024, "Transfer will overflow SPR." );
			Asura_PS2_DMA_Assert( ((uFromSrcAddr & 0x0fffffff) >> 4) + uFromNumQWs <= 1024, "Transfer will overflow SPR." );
            SetQWC( CHANNEL_TOSPR,	 uToNumQWs );
			SetQWC( CHANNEL_FROMSPR, uFromNumQWs );
            // (MP - 05.07.2002) This appears to be necessary to guarantee
            // that all registers - in particular the relevant bits in D_PCR
            // and D_STAT that setup completion notification on cop0 - are
            // written before Dx_CHCR.
			asm __volatile__( "sync.p" );
            asm __volatile__( "sync.l" );
			// Both DMA's are read to go... The following to start dma writes will sit in the WBB until the sync.l
			// at which point they get flushed together and avoid the second write stalling on the first dma
            SetCHCR( CHANNEL_FROMSPR, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
									  (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                                      (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                                      (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                                      (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                                      (1 << D_CHCR_STR_O) ); // STR: start DMA.
            SetCHCR( CHANNEL_TOSPR, (1 << D_CHCR_DIR_O) |  // DIR: from memory.
                                    (0 << D_CHCR_MOD_O) |  // MOD: normal mode.
                                    (0 << D_CHCR_ASP_O) |  // ASP: irrelevant in normal mode.
                                    (0 << D_CHCR_TTE_O) |  // TTE: irrelevant in normal mode.
                                    (0 << D_CHCR_TIE_O) |  // TIE: irrelevant in normal mode.
                                    (1 << D_CHCR_STR_O) ); // STR: start DMA.
			asm __volatile__( "sync.p" );												
			asm __volatile__( "sync.l" );
		}  
};
// End class Asura_PS2_DMA.

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

template< typename T >
static inline u_long128* Asura_PS2_DMAFromMem( T* __p )
{
    Asura_PS2_DMA_Assert( (reinterpret_cast<u_int>(__p) & ASURA_PS2_SPR_MEM_BASE) != ASURA_PS2_SPR_MEM_BASE, "Should be using Asura_PS2_DMAFromSPR()!" );
    return reinterpret_cast<u_long128*>( reinterpret_cast<u_int>(__p) & ASURA_PS2_DMA_MEM_MASK );
}

template< typename T >
static inline u_long128* Asura_PS2_DMAFromSPR( T* __p )
{
    Asura_PS2_DMA_Assert( (reinterpret_cast<u_int>(__p) & ASURA_PS2_SPR_MEM_BASE) == ASURA_PS2_SPR_MEM_BASE, "Should be using Asura_PS2_DMAFromMem()!" );
    return reinterpret_cast<u_long128*>( reinterpret_cast<u_int>(__p) | ASURA_PS2_DMA_SPR_FLAG );
}

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_DMA_H___
