//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_OS.cpp
//
// Copyright (c) 2001, 2002, 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_PS2_OS.h"

// Standard independent.
#include <ctype.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>

// Standard platform-specific.
#include <eekernel.h>
#include <libver.h>
#include <libcdvd.h>	// for poweroff
#include <libsn.h>		// DLL support
#if SCE_LIBRARY_VERSION	>=0x3020  
#include <sceerrno.h>	// for error numbers
#endif // #if SCE_LIBRARY_VERSION	>=0x3020  
#include <sifdev.h>		// for error numbers

// Engine independent.
#include "Core/Asura_Handle.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	#if (defined ASURA_SUPPORT_GAMESPY) 
		#if !(defined ASURA_GAMESPY_USEANTS)
		#include "Network/Asura_GameSpy_GT2.h"
		#else
		#include "Network/Asura_NetTransport_System.h"
		#endif
	#endif
#endif

// Engine platform-specific.
#include "Core/Asura_PS2_IRX.h"
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
#include "Network/Asura_PS2_eznetcnf.h"
#endif
#include "Sound/Asura_PS2_Multistream.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#define MAX_RETRIES     (256) // (MH - 03.12.2002) was 100. Seeking from one 
                              // end of the CD to the other seems to take longer 
                              // than 100 retries though

#define CD_SECTOR_SIZE  (2048)

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void DelayCallback( int iID, u_short usCount, void* pvArg );

extern "C" void attySetMode( int iMode );

// G L O B A L S /////////////////////////////////////////////////////////////

// (MH - 10.05.2002) fileroot buffer for AtWinMon compatability
#ifndef ASURA_PS2_CDEMUL
char ASURA_PS2_FILEROOT[256];
#endif

static char g_szBufferForFilePaths[ASURA_PS2_FULL_PATH_BFR_SIZE] = "";

static const char *g_szIRXModuleLoadResultDescs[] = 
{
    "Resident",
    "NON RESIDENT",
    "Removable Resident"
};

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#ifndef ASURA_PS2_SONY_SUBMISSION
bool Asura_PS2_OS::s_bIsTOOL        = false;
bool Asura_PS2_OS::s_bIsTestKit     = false;
bool Asura_PS2_OS::s_bIsNTSCTestKit = false;
#endif

#ifndef ASURA_PS2_CDEMUL
bool Asura_PS2_OS::s_bIsRunningUnderAtwinmon = false;
#endif

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
bool Asura_PS2_OS::s_bIsNetworkAdaptorConnected = false;
#endif

int Asura_PS2_OS::s_iMainThreadID;

int Asura_PS2_OS::s_iUsedMem = 0;
int Asura_PS2_OS::s_iMaxUsedMem = 0;
int Asura_PS2_OS::s_iFreeMem = 0;
int Asura_PS2_OS::s_iMinFreeMem = 128*1024*1024;

#ifdef ASURA_PS2_SUPPORT_DLLS
Asura_PS2_OS::LoadedDllInfo Asura_PS2_OS::s_axLoadedDllInfo[ASURA_PS2_MAX_LOADED_DLLS];
u_int Asura_PS2_OS::s_uNumLoadedDlls = 0;
#endif

// JGP Project specific command line
#ifdef ASURA_USE_PROJECT_COMMAND_LINE
extern const bool Project_ProcessCmdLineToken( const char* const szToken );
#endif

void Asura_PS2_OS::Initialise( void )
{
    // NRM 12/03/03 - change priority of main apps thread to allow middleware to perform correctly
	s_iMainThreadID = GetThreadId();
    ChangeThreadPriority(s_iMainThreadID, ASURA_PS2_MAIN_THREAD_PRIORITY);

#ifndef ASURA_PS2_SONY_SUBMISSION
    CheckForTOOLOrTestKit();

    {
        char szBfr[128] = "";

        if( IsTOOL() )
        {
            strcat( szBfr, "TOOL detected." );
        }
        else
        {
            if( IsRunningUnderAtwinmon() )
            {
                strcat( szBfr, "Running under atwinmon." );
            }

            if( IsTestKit() )
            {
                if( *szBfr ) { strcat( szBfr, " " ); }

                strcat( szBfr, (IsNTSCTestKit() ? "NTSC" : "PAL") );
                strcat( szBfr, " testkit detected." );
            }
        }

        if( *szBfr )
        {
            Asura_OutputToDebugger_Retail(( "%s\n", szBfr ));
        }
    }
#endif

#ifdef ASURA_PS2_SUPPORT_DLLS
	int iRet = snInitDllSystem( NULL );
	Asura_Assert(iRet == SN_DLL_SUCCESS, "DLL system init failed");
#endif
}

void Asura_PS2_OS::Shutdown( void )
{
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	Asura_PS2_EZNetCnf::ShutdownNetworkConnection();
#endif
	// unload all DLLs here? I'll assume that any system which loaded one
	// will be shutdown and that will unload the DLL.
#ifdef ASURA_PS2_SUPPORT_DLLS
	Asura_Assert(s_uNumLoadedDlls == 0, "some DLLs were not unloaded");
#endif
}

void Asura_PS2_OS::ParseCommandLine( int iNumArgs, char** pszArgs )
{
#ifndef ASURA_PS2_CDEMUL

    s_bIsRunningUnderAtwinmon = false;

	if( iNumArgs && !strnicmp(pszArgs[0], "atfile", 6) )
	{
        s_bIsRunningUnderAtwinmon = true;

        const char* szStart = pszArgs[0];
		const char* szEnd   = szStart;

		while( *szEnd != ','  ) { ++szEnd; }
		while( *szEnd != '.'  ) { ++szEnd; }
		while( *szEnd != '\\' ) { --szEnd; }
		++szEnd;

        strncpy( ASURA_PS2_FILEROOT, szStart, szEnd - szStart );
        ASURA_PS2_FILEROOT[szEnd - szStart] = '\0';
	}
    else
    {
		strcpy( ASURA_PS2_FILEROOT, ASURA_PS2_PCFILEROOT );
    }

#if !_DEBUG
    if( !s_bIsRunningUnderAtwinmon )
    {
        attySetMode(0);
    }
#endif

    Asura_OutputToDebugger_Retail(( "fileroot = %s\n", ASURA_PS2_FILEROOT ));

// JGP Project specific command line
#ifdef ASURA_USE_PROJECT_COMMAND_LINE
	for ( int i = 1; i < iNumArgs; i++ )
	{
		Project_ProcessCmdLineToken( pszArgs[ i ] );
	}
#endif

#endif // ndef ASURA_PS2_CDEMUL
}

#ifndef ASURA_PS2_SONY_SUBMISSION
void Asura_PS2_OS::CheckForTOOLOrTestKit( void )
{
    int iFile = sceOpen( "rom0:ROMVER", SCE_RDONLY );
    
    // Shouldn't fail...
    Asura_Assert( iFile >= 0, "Failed to access ROM version information." );

    // This should be a 16 byte file. Read it in and then close it.
    char szRomVer[16];
    sceRead( iFile, szRomVer, 16 );
    sceClose( iFile );

    // This should apparently begin with "0100TD" on all TOOLs. See "How do
    // I detect if the game is running on a dev kit or a real system?" in
    // sce.dev.devkit.
    if( !strncmp(szRomVer, "0100TD", 6) ) { s_bIsTOOL = true; }

#if _DEBUG
    // We must be running on a TOOL in debug builds so verify that the
    // assumption above is correct.
    Asura_Assert( !strncmp(szRomVer, "0100TD", 6), "This seems to be a TOOL with a different ROM version prefix. Please tell Morgan." );
#endif

    // If we're not on a TOOL then the 6th character should be a D iff it's
    // a test kit (otherwise it should be a C.)
    if( !s_bIsTOOL )
    {
        Asura_Assert( 'C' == szRomVer[5] || 'D' == szRomVer[5], "" );
        if( 'D' == szRomVer[5] )
        {
            s_bIsTestKit = true;

            // NTSC?
            if( 'A' == szRomVer[4] )
            {
                s_bIsNTSCTestKit = true;
            }
        }
    }
}
#endif

u_char* Asura_PS2_OS::LoadFileAligned( const char* szFilename, ASURA_ALIGN_ID eAlignment, u_int* puFileSizeInBytes )
{
    Asura_Handle iHandle = Asura_FileAccess::Open( szFilename );
    if( ASURA_INVALID_HANDLE == iHandle ) { return NULL; }

    // Get file size.
    u_int uRealSize = Asura_PS2_Multistream_FastLoad::GetFileSize( iHandle );

    // Round up to alignment.
    u_int uBfrSize = ((uRealSize + (eAlignment - 1)) / eAlignment) * eAlignment;

    // Allocate buffer.
    u_char* const pucBfr = reinterpret_cast<u_char*>( Asura_AllocateAlign(uBfrSize, eAlignment) );
    if( !pucBfr ) { Asura_FileAccess::Close( iHandle ); return NULL; }

    // Read in the data.
    u_char* pucOut = pucBfr;
    while( true )
    {
        const void* pvNext;
        int iNumBytes = Asura_PS2_Multistream_FastLoad::GetData( iHandle, &pvNext );

        if( Asura_PS2_Multistream_FastLoad::REACHED_EOF == iNumBytes ) { break; }

        memcpy( pucOut, pvNext, iNumBytes );      
        pucOut += iNumBytes;
    }

    Asura_FileAccess::Close( iHandle );

    // Read correct amount?
    if( static_cast<u_int>(pucOut - pucBfr) != uRealSize )
    {
        Asura_DeallocateAlign( pucBfr );
        return NULL;
    }
	
	// (MH - 07.12.2004) this has been missing since June 2003 :)
	if (puFileSizeInBytes)
	{
		*puFileSizeInBytes = uRealSize;
	}

	return pucBfr;
}

void Asura_PS2_OS::BackupFullPath( char* szBackupBfr )
{
    strcpy( szBackupBfr, g_szBufferForFilePaths );
}

void Asura_PS2_OS::RestoreFullPath( char* szBackupBfr )
{
    strcpy( g_szBufferForFilePaths, szBackupBfr );
}

const char* Asura_PS2_OS::GetFullPath( const char* szInputFilename, bool bPrependDeviceName, const char* szReplaceExt, bool bIsDirectory )
{
    // (MP - 17.03.2003) NB: Any asserts that fire inside this method will
    // result in execution re-entering it (when we construct the full path to
    // the assert logfile), thereby destroying the contents of the global
    // g_szBufferForFilePaths buffer and possibly producing a crash,
    // depending on exactly where the assert is.
    Asura_Assert( !strchr(szInputFilename, '/'), "Filenames containing '/' will break CD builds; use '\\' instead." );

#ifdef ASURA_PS2_CDEMUL

    char* szOut;
	if(bPrependDeviceName)
	{
#ifdef ASURA_PS2_DEMODISK
        Asura_PrintfToBfr( g_szBufferForFilePaths, "%s\\%s", ASURA_PS2_FULL_PATH_BFR_SIZE, true, ASURA_PS2_CD_DEVICENAME, ASURA_PS2_DEMODISK_ROOT );
#else
		strcpy( g_szBufferForFilePaths, ASURA_PS2_CD_DEVICENAME );
#endif
	}
	else
	{
#ifdef ASURA_PS2_DEMODISK
        Asura_PrintfToBfr( g_szBufferForFilePaths, "\\%s", ASURA_PS2_FULL_PATH_BFR_SIZE, true, ASURA_PS2_DEMODISK_ROOT );
#else
		g_szBufferForFilePaths[0] = '\0';
#endif
	}
	szOut = &g_szBufferForFilePaths[strlen(g_szBufferForFilePaths)];

    // Add leading backslash?
    if( *szInputFilename != '\\' )
    {
        *szOut++ = '\\';
    }

    // Copy filename, ensuring all characters are uppercase.
    while( (*szOut = toupper(*szInputFilename++)) != '\0' ) { ++szOut; }

    // Append ";1".
    Asura_Assert( ';' != *(szOut - 2) || '1' != *(szOut - 1), "Filename should not already contain \";1\"." );
    if( !bIsDirectory )
    {
        *szOut++ = ';';
        *szOut++ = '1';
        *szOut++ = '\0';
    }

#else // not defined ASURA_PS2_CDEMUL

    Asura_Assert( !strstr(szInputFilename, "host0:"), "Filename should be relative to root, not absolute." );
    strcpy( g_szBufferForFilePaths, ASURA_PS2_FILEROOT );
    strcat( g_szBufferForFilePaths, szInputFilename );

#endif // else not defined  ASURA_PS2_CDEMUL

    // Replace extension?
    if( szReplaceExt )
    {
        char* const szNulTerminator = g_szBufferForFilePaths + strlen(g_szBufferForFilePaths);

        char* szExt = NULL;
        for( char* szSearch = szNulTerminator - 1; szSearch > g_szBufferForFilePaths; --szSearch )
        {
            // If we encounter a slash then we've reached the directory
            // containing the file and so definitely won't find an extension.
            // We want to stop searching since directory names could contain
            // a . character as well and we don't want to go replacing
            // anything there.
            if( '\\' == *szSearch ) { break; }

            if( '.' == *szSearch ) { szExt = szSearch + 1; break; }
        }

        if( szExt )
        {
#ifdef ASURA_PS2_CDEMUL
            // Ignore the last two characters in this comparison since they
            // will be ";1" and not part of the extension.
            Asura_Assertf( (szNulTerminator - 2) - szExt == static_cast<int>(strlen(szReplaceExt)), ("Replacement extension \"%s\" differs in length from extension found in filename \"%s\".", szReplaceExt, szInputFilename) );
#else
            Asura_Assertf( szNulTerminator - szExt == static_cast<int>(strlen(szReplaceExt)), ("Replacement extension \"%s\" differs in length from extension \"%s\" found in filename \"%s\".", szReplaceExt, szExt, szInputFilename) );
#endif

            const char* szSrc = szReplaceExt;
            while( *szSrc )
            {
                *szExt++ = toupper( *szSrc++ );
            }
        }
        else
        {
            Asura_Assertf( false, ("Replacement extension specified but extension not found in filename \"%s\".", szInputFilename) );
        }
    }

    // (MP - 16.09.2002) This is potentially dangerous but the alternative 
    // would be to truncate the output and I'm not sure that's any better
    // really, ie. we'd fail to open the file instead of crashing but it's
    // still unlikely we'd be able to continue.)
    Asura_Assert( strlen(g_szBufferForFilePaths) + 1 <= ASURA_PS2_FULL_PATH_BFR_SIZE, "Internal buffer wasn't big enough." );

    return g_szBufferForFilePaths;
}

#ifdef ASURA_PS2_CDEMUL
const char* Asura_PS2_OS::GetFullPathOnPC( const char* szInputFilename )
{
    strcpy( g_szBufferForFilePaths, ASURA_PS2_PCFILEROOT );
    strcat( g_szBufferForFilePaths, szInputFilename );
    return g_szBufferForFilePaths;
}
#endif // #ifdef ASURA_PS2_CDEMUL

static u_int g_uLastMemoryUpdateTimestamp = 0;

void Asura_PS2_OS::UpdateMemoryUsageStats()
{
    // (MP - 25.04.2003) Calling mallinfo() is very slow when using the DL
    // memory manager and with the number of allocations we tend to have
    // outstanding in Dredd. To get around this, let's just update the memory
    // statistics every couple of seconds.
    if( !g_uLastMemoryUpdateTimestamp || Asura_Timers::GetFrameNumber() > g_uLastMemoryUpdateTimestamp + 50 )
    {
        struct mallinfo xMallInfo = mallinfo();
        s_iFreeMem = xMallInfo.fordblks;
        s_iUsedMem = xMallInfo.uordblks;

        g_uLastMemoryUpdateTimestamp = Asura_Timers::GetFrameNumber();
    }

    if (s_iUsedMem > s_iMaxUsedMem)
    {
        s_iMaxUsedMem = s_iUsedMem;
    }
    
    if (s_iFreeMem < s_iMinFreeMem)
    {
        s_iMinFreeMem = s_iFreeMem;
    }
}

void Asura_PS2_OS::GetMemStats(int& iUsedMem, int& iMaxUsedMem, int& iFreeMem, int& iMinFreeMem)
{
    UpdateMemoryUsageStats();
    iUsedMem = s_iUsedMem;
    iMaxUsedMem = s_iMaxUsedMem;
    iFreeMem = s_iFreeMem;
    iMinFreeMem = s_iMinFreeMem;
}

void Asura_PS2_OS::DelayThread( int milliseconds )
{
	if (milliseconds <=0 )
	{
		return;
	}

	// Very approximately 16 scanlines per ms
	SetAlarm(16*milliseconds,DelayCallback,(void *)GetThreadId());
	SleepThread();
}

// (MP - 16.09.2002) Currently unused.
//// MH 17.10.01 -- code to print out memory sizes - may be moved to memory manager.
//
//// this typedef prevents the globals being put in small data section. 
//// the startup code (crt.s) assumes they are in the normal section
//typedef unsigned int __attribute__ ((mode (TI))) heap_size_type;
//
//extern "C" heap_size_type _stack_size;
//extern "C" heap_size_type _stack;
//extern "C" heap_size_type _heap_size;
//extern "C" heap_size_type _end;
//
//void Asura_PS2_PrintMemoryInfo( void )
//{
//    //int ProgEnd =  (int)&_end;
//    int StackStart = (int)&_stack;
//    int StackSize = (int)&_stack_size;
//    int HeapSize = (int)&_heap_size;
//
//    printf("stack size  = 0x%08x (%d)\n", StackSize, StackSize);
//    
//    if (StackStart == -1)
//    {
//        printf("stack location is default.\n");
//    }
//    else
//    {
//        printf("stack       = 0x%08x (%d)\n", StackStart, StackStart);
//    }
//    
//    if (HeapSize == -1)
//    {
//        printf("heap size is maximum.\n");
//    }
//    else
//    {
//        printf("heap size   = 0x%08x (%d)\n", HeapSize, HeapSize);
//    }
//}

void Asura_PS2_OS::DumpMemoryInfo( void )
{
    struct mallinfo amal = mallinfo();
    int iMemAllocated = GetMemAlloced();
    int iMemAllocatedMb = iMemAllocated / 1024 / 1024;
    int iMemAllocatedKb = (iMemAllocated - (iMemAllocatedMb * 1024 * 1024)) / 1024;
    int iMemAllocatedBytes = (iMemAllocated - ((iMemAllocatedMb * 1024 * 1024) + (iMemAllocatedKb * 1024)));

    int iMemFree = GetMemFree();
    int iMemFreeMb = iMemFree / 1024 / 1024;
    int iMemFreeKb = (iMemFree - (iMemFreeMb * 1024 * 1024)) / 1024;
    int iMemFreeBytes = (iMemFree - ((iMemFreeMb * 1024 * 1024) + (iMemFreeKb * 1024)));

    int iFragmentation = amal.ordblks;
    int iFastBinsFree = amal.fsmblks;
    int iFastBinsUsed = amal.smblks;

    Asura_OutputToDebugger_Retail(("%d Mb %d Kb %d bytes allocated. %d Mb %d Kb %d bytes free. \n"
        "Fragmentation %d. FastBins free blocks = %d, FastBins use = %d bytes.\n",
        iMemAllocatedMb, iMemAllocatedKb, iMemAllocatedBytes,
        iMemFreeMb, iMemFreeKb, iMemFreeBytes,
        iFragmentation, iFastBinsFree, iFastBinsUsed));
}

int Asura_PS2_OS::GetMemAlloced( void )
{
    struct mallinfo amal = mallinfo();
    return ( amal.uordblks );
}

int Asura_PS2_OS::GetMemFree( void )
{
    struct mallinfo amal = mallinfo();
    return ( amal.fordblks );
}


bool Asura_PS2_OS::IsMainThread( void )
{
	return (s_iMainThreadID == GetThreadId());
}


int Asura_PS2_OS::LoadIRXModule(const char *szFilename, int argc, const char *argv, bool bForceHost0 /* = false */, const u_int uMaxRetries /* = 0 */, int* piResult /* = NULL */)
{
    int iErr;
	int iResult;

#ifdef ASURA_PS2_CDEMUL
	const char* szFullpath = bForceHost0 ? GetFullPathOnPC (szFilename) : GetFullPath(szFilename, true);
#else
	const char* szFullpath = GetFullPath(szFilename, true);
#endif	

	u_int uRetry = uMaxRetries + 1;

    do
	{
		iErr = sceSifLoadStartModule(szFullpath, argc, argv, &iResult);
		if (uMaxRetries)
		{
			--uRetry;
		}
    }
	while (uRetry && (iErr < 0));

    if (iErr >= 0) 
	{
		printf("Loaded %s, %s\n", szFullpath, g_szIRXModuleLoadResultDescs[iResult & 3]);
	}
	else
	{
		printf("Failed to load %s, after %u attempts; iErr = %d, reason = %d\n", szFullpath, uMaxRetries, iErr, iResult);
	}

	if (piResult)
	{
		*piResult = iResult;
	}
	return iErr;	
}

/*
 *  returns:
 *     0 when module successfully unloads,
 *     1 when module could not be unloaded, or
 *    -1 if module ID parameter is invalid.
 */
int Asura_PS2_OS::UnloadIRXModule(const int id)
{
    int err = -1, result;
    if ( id >= 0 ) {
        if (0 > (err = sceSifStopModule(id, 0, NULL, &result))) {
            printf("Unloading> module id %d did not stop (%d)\n", id, err);
        } else {
            if (0 > (err = sceSifUnloadModule(id))) {
                printf("Unloading> module id %d did not unload (%d)\n", id, err);
            } else {
                err = 0;
            }
        }
    }
    return err;
}

/*
 *  returns:
 *     0 when module is not found (so we can abuse this function),
 *     0 when module successfully unloads, or
 *     1 when module could not be unloaded.
 */
int Asura_PS2_OS::UnloadIRXModuleByName(const char *name)
{
    int id;
    id = sceSifSearchModuleByName(name);
    if (id >= 0) {
        return UnloadIRXModule(id);
    }
    return 0;
}

void Asura_PS2_OS::Reset( void )
{
    g_uLastMemoryUpdateTimestamp = 0;
}


#ifdef ASURA_PS2_SUPPORT_DLLS
// internal (private) function to get DLL info by ID
int Asura_PS2_OS::GetDllIndex( Asura_Hash_ID uID )
{
	for (u_int u = 0; u < s_uNumLoadedDlls; ++u)
	{
		if (s_axLoadedDllInfo[u].m_uID == uID)
		{
			return static_cast<int>(u);
		}
	}
	return -1;
}

// call AllocateDllMem to get memory to load a Dll to.
u_long128* Asura_PS2_OS::AllocateDllMem( u_int uSize )
{
	if (s_uNumLoadedDlls >= ASURA_PS2_MAX_LOADED_DLLS)
	{
		Asura_Assert(false, "can't load any more Dlls");
		return NULL;
	}
	return reinterpret_cast<u_long128*>(Asura_AllocateAlign( uSize, ASURA_ALIGN_128_BYTE ));
}

// call with the stripped name of the dll (eg "ps2_dnas")
bool Asura_PS2_OS::IsDllLoaded( const char* szName )
{
	Asura_Hash_ID uHashID = Asura_GetHashID(szName);
	
	if(GetDllIndex(uHashID) != -1)
	{
		return true;
	}

	return false;
}

// call InitDll when you have loaded a DLL into that memory to initialise it.
void Asura_PS2_OS::InitDll(const char* szName, u_long128* puDllAddr)
{
	Asura_String xStrippedFilename(szName);
	// remove path and ext
	xStrippedFilename.SplitFilename(NULL, &xStrippedFilename, NULL);
	szName = xStrippedFilename.GetString();
		
	Asura_Hash_ID uHashID = Asura_GetHashID(szName);
	Asura_Assert(puDllAddr, "NULL address for DLL");
	Asura_Assert(GetDllIndex(uHashID) == -1, "DLL already loaded");
	Asura_Assert(s_uNumLoadedDlls < ASURA_PS2_MAX_LOADED_DLLS, "too many DLLs");

	Asura_OutputToDebugger(("1. num loaded DLLs = %d.\n", s_uNumLoadedDlls));
	LoadedDllInfo* pxDllInfo = &s_axLoadedDllInfo[s_uNumLoadedDlls++];
	Asura_OutputToDebugger(("2. num loaded DLLs = %d.\n", s_uNumLoadedDlls));

#ifndef __SNC__
	int iRet = snDllLoaded(puDllAddr, NULL);
	// DLL loader has problems patching if breakpoints are set - if stepping
	// through a temporary breakpoint gets set here so make sure it's a nop
	// as per SNs FAQ. 
	asm("nop");
	Asura_OutputToDebugger(("3. num loaded DLLs = %d.\n", s_uNumLoadedDlls));
	Asura_Assert(iRet == SN_DLL_SUCCESS, "DLL init failed");
#endif // #ifndef __SNC__

	pxDllInfo->m_uID = uHashID;
	pxDllInfo->m_puAddress = puDllAddr;
	Asura_OutputToDebugger(("4. num loaded DLLs = %d.\n", s_uNumLoadedDlls));
}

// call ShutdownAndUnloadDll to free the memory the DLL was allocated.
void Asura_PS2_OS::ShutdownAndUnloadDll(const char* szName)
{
	Asura_Hash_ID uHashID = Asura_GetHashID(szName);
	Asura_Assert(s_uNumLoadedDlls, "no DLLs loaded");
	int iDllIndex = GetDllIndex(uHashID);
	Asura_Assert(iDllIndex != -1, "DLL with that ID not loaded");
	if (iDllIndex == -1)
	{
		return;
	}

	u_long128* puDllAddr = s_axLoadedDllInfo[iDllIndex].m_puAddress;
	Asura_Assert(puDllAddr, "DLL with that ID not loaded");

	int iRet = snDllUnload(puDllAddr);
	Asura_Assert(iRet == SN_DLL_SUCCESS, "DLL shutdown failed");

	// free the memory
	Asura_DeallocateAlign(puDllAddr);

	// remove from list of loaded DLLs (swap with last one and decrease num loaded)
	s_axLoadedDllInfo[iDllIndex] = s_axLoadedDllInfo[--s_uNumLoadedDlls];
}
#endif // #ifdef ASURA_PS2_SUPPORT_DLLS

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

static const int POWEROFF_THREAD_STACKSZ = (512 * 16);

static u_char aucPoweroffThreadStack[POWEROFF_THREAD_STACKSZ] __attribute__ ((aligned(16)));
static int g_iSemaphoreID;
static int g_iPoweroffThreadID;

void Asura_PS2_OS::PowerOffThread(void *arg) 
{
	int sid = (int)arg;
	int stat;
	while(1) 
	{
		WaitSema(sid);
		scePrintf("power off request has come.\n");

		// stop the main thread now - this prevents anything else happening after we shut things down here
		SuspendThread( s_iMainThreadID );

		// this is the list of extra tasks from the Network SDK FAQ
		//	Required: Gracefully shut down the network interface, releasing DHCP, PPP, and PPPoE sessions
		//	Optional: Shut down open socket connections
		//	Optional: Fade video display to black
		//	Optional: Fade SPU2 master volume to 0

		// shutdown open socket connections (optional)
		// This will speed up the time before any other player in the game are
		// aware we have disconnected.
#if (defined ASURA_SUPPORT_GAMESPY)
	#if !(defined ASURA_GAMESPY_USEANTS)
		Asura_GameSpy_GT2::Shutdown();
	#else
		Asura_NetTransport_System::Shutdown();
	#endif
#endif

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
		// shutdown network interface (required)
		Asura_PS2_EZNetCnf::ShutdownNetworkConnection();
#endif //ASURA_SUPPORT_EXTERNAL_NETWORK

		// dev9 power off, need to power off PS2
		while(sceDevctl("dev9x:", DDIOC_OFF, NULL, 0, NULL, 0) < 0);
		// PS2 power off
		while(!sceCdPowerOff(&stat) || stat);

		scePrintf("power off complete.\n");
    }
}

void Asura_PS2_OS::PowerOffHandler(void *arg) 
{
    int sid = (int)arg;
    iSignalSema(sid);
}

void Asura_PS2_OS::PreparePowerOff(void) 
{
    struct ThreadParam tparam;
    struct SemaParam   sparam;

    sparam.initCount = 0;
    sparam.maxCount  = 1;
    sparam.option    = 0;
    g_iSemaphoreID = CreateSema(&sparam);

    tparam.stackSize = POWEROFF_THREAD_STACKSZ;
    tparam.gpReg = &_gp;
    tparam.entry = PowerOffThread;
    tparam.stack = (void *)aucPoweroffThreadStack;
    tparam.initPriority = 1;
    
	g_iPoweroffThreadID = CreateThread(&tparam);
    StartThread(g_iPoweroffThreadID, (void *)g_iSemaphoreID);

    sceCdPOffCallback(PowerOffHandler, (void *)g_iSemaphoreID);
}

void Asura_PS2_OS::UndoPowerOff(void)
{
    // sanity check
    if (sceSifSearchModuleByName("dev9") >= 0) {
        // remove poweroff handler, thread, and semaphore
        sceCdPOffCallback( NULL, NULL );
        TerminateThread( g_iPoweroffThreadID );
        DeleteThread( g_iPoweroffThreadID );
        DeleteSema( g_iSemaphoreID );
        // close any and all open files
        sceDevctl("pfs:", PDIOC_CLOSEALL, NULL, 0, NULL, 0);
        // turn off the expansion bus
        while (sceDevctl("dev9x:", DDIOC_OFF, NULL, 0, NULL, 0) < 0) {};
    }
}
#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

// (PSJE-14.01.2003)	Function to sleep the current thread for n milliseconds.
// Added so other threads (ie multistream) get a chance to complete disk access ect, whilst we are sitting in a sceCdSync loop
// Code copied directly from a sony newsgroups cddvd message thread (just search for DelayThread)
static void DelayCallback( int id, u_short count,void *arg)
{
	iWakeupThread((int)arg);
	asm __volatile__("sync.l; ei");
}

#ifndef ASURA_PS2_SONY_SUBMISSION
void Asura_PS2_OS::DecodeSceeErrNo( const int iErrNo )
{
#if SCE_LIBRARY_VERSION	>=0x3020  
	switch ( iErrNo )
	{
		case SCE_STDIO_EBADF:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EBADF\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENODEV:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENODEV\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EINVAL:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EINVAL\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EMFILE:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EMFILE\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EIO:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EIO\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENOMEM:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENOMEM\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENOTDIR:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENOTDIR\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENXIO:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENXIO\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EACCES:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EACCES\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENOENT:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENOENT\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EBUSY:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EBUSY\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENOSPC:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENOSPC\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EFBIG:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EFBIG\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENAMETOOLONG:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENAMETOOLONG\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ELOOP:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ELOOP\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EROFS:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EROFS\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EISDIR:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EISDIR\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EEXIST:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EEXIST\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_ENOTEMPTY:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_ENOTEMPTY\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EVERSION:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EVERSION\n", iErrNo ) );
			break;
		}
		case SCE_STDIO_EDEVICE_BROKEN:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X SCE_STDIO_EDEVICE_BROKEN\n", iErrNo ) );
			break;
		}
		default:
		{
			Asura_OutputToDebugger_Retail( ( "0x%08X UNKNOWN\n", iErrNo ) );
			break;
		}
	}
#else // #if SCE_LIBRARY_VERSION	>=0x3020  
	// Hopefully ignore warning
	(void)iErrNo;
#endif // #if SCE_LIBRARY_VERSION	>=0x3020  
}
#endif // #ifndef ASURA_PS2_SONY_SUBHMISSION

///////////////////////////////////// END OF FILE //////////////////////////////////////
