//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_SystemInfo.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     System information for PS2
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_PS2_SystemInfo.h"

// Stdlib independent.
#include <limits.h>
#include <time.h>

// Engine independent.
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"

// Stdlib platform-specific.
#include <libcdvd.h>
#include <libscf.h>

// C O N S T A N T S /////////////////////////////////////////////////////////

// Max acceptable age of the cached date and time, in milliseconds.
static const u_int uMAX_CACHE_AGE = 30 * 60 * 1000;

// G L O B A L S /////////////////////////////////////////////////////////////

static tm g_xCachedDateAndTime;
static u_int g_uCacheTimestamp; // Milliseconds since start.

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static int BCDToInt(u_char ucBCD)
{
    // the members in the xClock struct are in binary coded decimal!
    return (ucBCD & 0x0F) + ((ucBCD >> 4) * 10);
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_PS2_SystemInfo::ConsoleInfo Asura_PS2_SystemInfo::s_xConsoleInfo;

void Asura_PS2_SystemInfo::Initialise(void)
{
    // Get all the info and store it in the struct
    s_xConsoleInfo.i24Hour         = sceScfGetTimeNotation();
    s_xConsoleInfo.iAspectRatio    = sceScfGetAspect();
    s_xConsoleInfo.iDateFormat     = sceScfGetDateNotation();
    s_xConsoleInfo.iDaylightSaving = sceScfGetSummerTime();
    s_xConsoleInfo.iLanguage       = sceScfGetLanguage();
    s_xConsoleInfo.iSpdif          = sceScfGetSpdif();
    s_xConsoleInfo.iTimeZone       = sceScfGetTimeZone();
}

void Asura_PS2_SystemInfo::GetSystemDateAndTime( tm* pxDateAndTime )
{
    u_int uCurMS = Asura_Timers::GetMSecsSinceStart_Continuous();
    u_int uDelta = uCurMS - g_uCacheTimestamp;

    // Since calling sceCdReadClock() can be so insanely expensive - it's
    // always costly but can occasionally block for hundreds of ms under
    // heavy IOP or RPC load - we'll avoid calling it as much as possible.
    if( !g_uCacheTimestamp || uDelta > uMAX_CACHE_AGE )
    {
        sceCdCLOCK xClock;
        sceCdReadClock( &xClock );
        // (MP - 28.07.2005) If xClock.stat is non-zero then this has also
        // failed. I don't know whether or not this is reflected in the
        // return value. For example, this happens if the PS2's internal
        // battery is dead. In this case the default time (00/03/04 05:06:07)
        // is returned. According to the library documentation, it is
        // acceptable to use this in the failure case, so I guess it's ok not
        // to bother doing anything here. Not sure what we'd do anyway.

        // Convert to local time.
        sceScfGetLocalTimefromRTC( &xClock );

        // Populate cached tm.
        g_xCachedDateAndTime.tm_mday = BCDToInt( xClock.day    );
        g_xCachedDateAndTime.tm_mon  = BCDToInt( xClock.month  );
        g_xCachedDateAndTime.tm_year = BCDToInt( xClock.year   );
        g_xCachedDateAndTime.tm_hour = BCDToInt( xClock.hour   );
        g_xCachedDateAndTime.tm_min  = BCDToInt( xClock.minute );
        g_xCachedDateAndTime.tm_sec  = BCDToInt( xClock.second );

        // Update timestamp.
        g_uCacheTimestamp = uCurMS;

        // Reset delta so we don't go offsetting below.
        uDelta = 0;
    }

    // Copy cached data to specified output.
    *pxDateAndTime = g_xCachedDateAndTime;

    // Apply delta.
    {
        // Convert delta to seconds.
        uDelta /= 1000;

        // Assume that delta will be at most a number of minutes.
        Asura_Assert( uDelta < 60 * 60, "" );
        
        u_int uMinuteDelta = uDelta / 60;
        u_int uSecondDelta = uDelta - uMinuteDelta * 60;

        pxDateAndTime->tm_min += uMinuteDelta;
        pxDateAndTime->tm_sec += uSecondDelta;
    }
}
