//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_VU0.cpp
//
// By: Mike Healey, based on VU1 equivalent
//
// Copyright (c) 2001 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////


#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_PS2_VU0.h"

// Standard platform-specific.
#include <devvif0.h>                        // For sceDevVif0Reset()
#include <devvu0.h>                         // For sceDevVu0Reset()
#include <eestruct.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static u_int* g_apuVU0MicroPrograms[Asura_PS2_VU0::VU0_PROG_NUMPROGS] =
{
	&VU0_MathsLib,
//  &VU0_Skinning,
};

Asura_PS2_VU0::VU0_MICROPROGRAM Asura_PS2_VU0::s_eCurrentProg = VU0_PROG_NONE;

static Asura_PS2_QW* g_pxNonBlockingLoadBfr = NULL;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////


// C L A S S E S /////////////////////////////////////////////////////////////

Asura_PS2_QW* Asura_PS2_VU0::s_pxQuickLoadCmdBfr = NULL;

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU0::Initialise().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Sets up VU0 stuff.
// AUTHOR       Mike Healey 28.11.01
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU0::Initialise( void )
{
    // Reset.
	sceDevVif0Reset();
	sceDevVu0Reset();

    // setup VU debugging
    sceDevVu0PutDBit(1);

    // Reset VU renderer state cache.
    s_eCurrentProg = VU0_PROG_NUMPROGS;

    s_pxQuickLoadCmdBfr    = NULL;
    g_pxNonBlockingLoadBfr = NULL;

	// (MH - 15.07.2003) load in the maths library
	LoadMicroprogram(VU0_PROG_MATHS_LIB);
}
// End function Asura_PS2_VU0::Initialise().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION 	Asura_PS2_VU0::LoadMicroprogram
// PARAMETERS	Which program to upload.
// RETURNS		None.
// DESCRIPTION	Uploads the specified microcode to VU0.
// AUTHOR		Mike Healey 28.11.01
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU0::LoadMicroprogram( VU0_MICROPROGRAM eMicroprog, bool bBlocking )
{
	Asura_Assert( VU0_PROG_MIN <= eMicroprog && eMicroprog < VU0_PROG_NUMPROGS, "Invalid microprogram!" );

	Asura_Assert( eMicroprog == VU0_PROG_MATHS_LIB, "Currently we assume maths lib is loaded on to VU0 at all times" );

    if( eMicroprog != s_eCurrentProg )
    {
        Asura_PS2_DMA::ChainTransferFromMem( Asura_PS2_DMA::CHANNEL_VIF0,
                                             g_apuVU0MicroPrograms[eMicroprog],
                                             false,
                                             true,
                                             bBlocking );

        s_eCurrentProg = eMicroprog;
    }
}
// End function Asura_PS2_VU0::LoadMicroprogram().


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU0::Shutdown().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  
// AUTHOR       Mike Healey 28.11.01
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU0::Shutdown( void )
{
    Asura_Assert( !s_pxQuickLoadCmdBfr, "Did you forget to call ShutdownQuickDataLoading()?" );
    Asura_Assert( !g_pxNonBlockingLoadBfr, "Did you forget to call ShutdownNonBlockingDataLoading()?" );
}
// End function Asura_PS2_VU0::Shutdown().

void Asura_PS2_VU0::InitNonBlockingDataLoading( void )
{
    Asura_Assert( !g_pxNonBlockingLoadBfr, "Did you forget to call ShutdownNonBlockingDataLoading()?" );
    g_pxNonBlockingLoadBfr = Asura_PS2_SPR_Alloc( 0, 2 );
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU0::LoadDataNonBlocking().
// PARAMETERS   pointer to 128-bit aligned data, 
//              num qwords to upload,
//              dest address in VU mem, which should have the top bit set for 
//              double buffered offset if required.
// RETURNS      None.
// DESCRIPTION  
// AUTHOR       Mike Healey 28.11.01
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU0::LoadDataNonBlocking(const u_long128* data_ptr, u_int num_quadwords, u_int vu_mem_dest, u_int kick_vu)
{
    Asura_Assert( g_pxNonBlockingLoadBfr, "Did you forget to call InitNonBlockingDataLoading()?" );
    u_int* pPacket = reinterpret_cast<u_int*>( g_pxNonBlockingLoadBfr );

    // (MP - 14.03.2003) For optimal transfer rates - because we're using a
    // REF tag and have TTE on - the data should be one QW ahead of a 8 QW
    // boundary.
    Asura_Assert( (reinterpret_cast<u_int>(data_ptr) & 127) == 16, "Data isn't correctly aligned." );
    
    *(pPacket++) = PY_DMA_REF | num_quadwords;
    *(pPacket++) = reinterpret_cast<u_int>(data_ptr);
    *(pPacket++) = SCE_VIF0_SET_STCYCL(4, 4, 0);    // CL == WL, i.e. write 1 qword every cycle. 
    *(pPacket++) = SCE_VIF0_SET_UNPACK(vu_mem_dest, num_quadwords, CMD_V4_32, 0); // 0 = address in VU mem

	// start VU microprogram and end dma transfer
	*(pPacket++) = PY_DMA_END | 0;
	*(pPacket++) = 0;
    *(pPacket++) = kick_vu;
	*(pPacket++) = SCE_VIF0_SET_NOP(0);

    // quad word count (no initial qword sent)
    DPUT_D0_QWC(0);

    // source address
    DPUT_D0_TADR( reinterpret_cast<u_int>(Asura_PS2_DMAFromSPR(g_pxNonBlockingLoadBfr)) );

    // for polling of completion
    DPUT_D_PCR(D_PCR_CPC0_M);
    DPUT_D_STAT(D_STAT_CIS0_M);

    asm __volatile__("sync.l");

    // start actual kick
    DPUT_D0_CHCR(1  // from memory
        | (1 << 2)  // chain mode
        | (1 << 6)  // transfer tag on
        | (1 << 8));    // dma start

    asm __volatile__("sync.l");
}

void Asura_PS2_VU0::ShutdownNonBlockingDataLoading( void )
{
    Asura_Assert( g_pxNonBlockingLoadBfr, "" );
    Asura_PS2_SPR_Dealloc( g_pxNonBlockingLoadBfr );
    g_pxNonBlockingLoadBfr = NULL;
}
