//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformBasics.cpp
//
// Created by RMS.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_PlatformBasics.h"

// Stdlib independent.
#include <wchar.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"

// T Y P E S /////////////////////////////////////////////////////////////////

#if _DEBUG
struct SPRReservation
{
    Asura_PS2_QW* pxMem;                // Ptr that was returned by Alloc().
    u_short       uQWOffset;            // Beginning of allocation in SPR.
    u_short       uNumQWs;              // Size of allocation.
    const char*   szFile;               // The file where the allocation was made.
    int           iLine;                // The line where it was made.
};

struct SPRReservationListNode
{
    SPRReservation          xData;      // Allocation info.
    SPRReservationListNode* pxNext;     // Next entry in list. NULL => this is the end of the list.
};
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

#if _DEBUG
static const u_int MAX_RESERVATION_LIST_NODES = 8;
#endif

// G L O B A L S /////////////////////////////////////////////////////////////

#if _DEBUG
static SPRReservationListNode* g_pxReservationListHead = NULL;
static SPRReservationListNode* g_pxNodePool = NULL;
static SPRReservationListNode* g_apxFreeListNodes[MAX_RESERVATION_LIST_NODES];
static u_int g_uNumFreeListNodes;
DEBUGVAR bool dbg_bFillAllocatedSPR = false;
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#if _DEBUG

void Asura_PS2_DebugSPR::Initialise( void )
{
    {
        g_pxNodePool = new SPRReservationListNode[MAX_RESERVATION_LIST_NODES];
        Asura_Assert( g_pxNodePool, "Insufficient memory to allocate SPR reservation list nodes." );

        for( u_int u = 0; u < MAX_RESERVATION_LIST_NODES; ++u )
        {
            g_apxFreeListNodes[u] = &g_pxNodePool[u];
        }

        g_uNumFreeListNodes = MAX_RESERVATION_LIST_NODES;
    }

    g_pxReservationListHead = NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugSPR::Alloc().
// PARAMETERS   uQWOffset - Offset to beginning of memory area required.
//              uNumQWs   - Size of memory area required.
// RETURNS      Ptr to beginning of allocated SPR area. Will not return NULL.
// DESCRIPTION  Verifies that the specified area of SPR is available and then
//              returns a ptr to it.
// AUTHOR       Morgan Parry - 08.05.2002.
//
//////////////////////////////////////////////////////////////////////////////
Asura_PS2_QW* Asura_PS2_DebugSPR::Alloc( u_int uQWOffset, u_int uNumQWs, const char* szFile, int iLine )
{
    // Verify the requested section of memory is valid.
    Asura_Assert( uNumQWs, "Stupid Alloc() request." );
    Asura_Assert( uNumQWs <= 1024, "Attempted to Alloc() invalid number of SPR QWs." );
    Asura_Assert( uQWOffset + uNumQWs <= 1024, "Attempted to Alloc() area of memory that exceeds SPR bounds." );

    // Calculate the pointer to be returned.
    Asura_PS2_QW* const pxMemRegion = reinterpret_cast<Asura_PS2_QW*>(ASURA_PS2_SPR_MEM_BASE) + uQWOffset;

    // Verify the specified section of memory is available.
    for( SPRReservationListNode* pxCur = g_pxReservationListHead; pxCur; pxCur = pxCur->pxNext )
    {
        // Calculate next free QW after this reservation.
        u_int uNextFreeQW = pxCur->xData.uQWOffset + pxCur->xData.uNumQWs;

        // Verify this reserved area and the requested area don't overlap.
        if( uQWOffset < uNextFreeQW && uQWOffset + uNumQWs > pxCur->xData.uQWOffset )
        {
            Asura_Assert( false, "Attempted to Alloc() area of SPR that's already in use." );

            // (MP - 07.05.2002) Could return NULL here but I think it makes
            // more sense to behave as the release build would now that we've
            // given an error message and return the ptr anyway.
            break;
        }
    }

    // Check there's a free node.
    if( !g_uNumFreeListNodes )
    {
        Asura_Assert( false, "Not enough reservation list nodes available; increase MAX_RESERVATION_LIST_NODES." );

        // (MP - 19.08.2002) Again, we could return NULL here but let's mimic
        // the release build, except now accepting we won't be able to record
        // the allocation.
        return pxMemRegion;
    }

    // Get node.
    SPRReservationListNode* pxNode = g_apxFreeListNodes[--g_uNumFreeListNodes];

    // Fill it out.
    pxNode->xData.pxMem     = pxMemRegion;
    pxNode->xData.uQWOffset = uQWOffset;
    pxNode->xData.uNumQWs   = uNumQWs;
    pxNode->xData.szFile    = szFile;
    pxNode->xData.iLine     = iLine;

    // Add to the front of the list.
    pxNode->pxNext = g_pxReservationListHead;
    g_pxReservationListHead = pxNode;

    if( dbg_bFillAllocatedSPR )
    {
        u_long128* puOut = reinterpret_cast<u_long128*>( pxMemRegion );
        const u_long128 uFill = ASURA_PS2_INIT_ULONG128( 0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDLL );
        for( u_int u = uNumQWs; u--; )
        {
            asm( "sq %0, 0(%1)" : : "r" (uFill), "r" (puOut++) );
        }
    }

    // And return ptr.
    return pxMemRegion;
}
// End function Asura_PS2_DebugSPR::Alloc().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugSPR::Dealloc().
// PARAMETERS   pxMem - Area of SPR to be released.
// RETURNS      None.
// DESCRIPTION  Verifies that the specified area of SPR was allocated using
//              Alloc() and then releases it.
// AUTHOR       Morgan Parry - 08.05.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_DebugSPR::Dealloc( Asura_PS2_QW* pxMem )
{
    // Find this section of memory is the reservation list.
    SPRReservationListNode* pxPrev = NULL;
    SPRReservationListNode* pxCur  = g_pxReservationListHead;
    while( pxCur )
    {
        if( pxCur->xData.pxMem == pxMem )
        {
            // Remove from list.
            if( pxPrev )
            {
                pxPrev->pxNext = pxCur->pxNext;
            }
            else
            {
                g_pxReservationListHead = pxCur->pxNext;
            }
            
            // Return node to free list.
            Asura_Assert( g_uNumFreeListNodes < MAX_RESERVATION_LIST_NODES, "" );
            g_apxFreeListNodes[g_uNumFreeListNodes++] = pxCur;

            // And we're done.
            return;
        }

        pxPrev = pxCur;
        pxCur  = pxCur->pxNext;
    }

    // Oh dear.
    // (MP - 19.08.2002) NB: This assert will fire erroneously if the value
    // of MAX_RESERVATION_LIST_NODES is too low.
    Asura_Assert( false, "Dealloc()'d SPR memory that wasn't Alloc()'d." );
}
// End function Asura_PS2_DebugSPR::Dealloc().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugSPR::IsAllocated().
// PARAMETERS   uQWOffset - Offset to beginning of memory area to be checked.
//              uNumQWs   - Size of memory area to be checked.
// RETURNS      True if the specified memory region is (part of) an area
//              reserved with Alloc().
// DESCRIPTION  See RETURNS. Note that this function does not cope with the
//              case where the region of memory to be checked covers more
//              than one reserved area; it will only return true if a single
//              Alloc() call reserved (at least) the specified area of memory.
// AUTHOR       Morgan Parry - 08.05.2002.
//
//////////////////////////////////////////////////////////////////////////////
bool Asura_PS2_DebugSPR::IsAllocated( u_int uQWOffset, u_int uNumQWs )
{
    // Verify the requested section of memory is valid.
    Asura_Assert( uNumQWs, "Stupid IsAllocated() call." );
    Asura_Assert( uNumQWs <= 1024, "Call to IsAllocated() with invalid number of SPR QWs." );
    Asura_Assert( uQWOffset + uNumQWs <= 1024, "Call to IsAllocated() with area of memory that exceeds SPR bounds." );

    for( SPRReservationListNode* pxCur = g_pxReservationListHead; pxCur; pxCur = pxCur->pxNext )
    {
        // Calculate next QW after this reservation.
        u_short uNextQW = pxCur->xData.uQWOffset + pxCur->xData.uNumQWs;

        // Does this reserved area contain the area we're checking?
        if( uQWOffset >= pxCur->xData.uQWOffset && uQWOffset + uNumQWs <= uNextQW )
        {
            return true;
        }
    }

    return false;
}
// End function Asura_PS2_DebugSPR::IsAllocated().

void Asura_PS2_DebugSPR::Shutdown( void )
{
    if( g_pxNodePool )
    {
        delete [] g_pxNodePool;
        g_pxNodePool = NULL;
    }
}

void Asura_PS2_DebugSPR::VerifyAllDeallocated( bool bForceAnyOutstandingToBeDeallocated )
{
    Asura_Assert( !g_pxReservationListHead, "SPR reservation list isn't empty. Did you forget to call Dealloc() at some point?" );

    if( bForceAnyOutstandingToBeDeallocated )
    {
        for( SPRReservationListNode* pxCur = g_pxReservationListHead; pxCur; pxCur = pxCur->pxNext )
        {
            g_apxFreeListNodes[g_uNumFreeListNodes++] = pxCur;
        }
        Asura_Assert( MAX_RESERVATION_LIST_NODES == g_uNumFreeListNodes, "" );

        g_pxReservationListHead = NULL;
    }
}

#endif // #if _DEBUG


extern "C"
{


int iswalnum( Asura_UnicodeChar wChar )
{
    return( iswalpha(wChar) || iswdigit(wChar) );
}


int wcsicmp( const Asura_UnicodeChar* wszA, const Asura_UnicodeChar* wszB )
{
    Asura_UnicodeChar wA, wB;

    do
    {
        wA = towlower( *wszA++ );
        wB = towlower( *wszB++ );
    }
    while( wA && wA == wB );

    return( wA - wB );
}


int wcsnicmp( const Asura_UnicodeChar* wszA, const Asura_UnicodeChar* wszB, size_t uMax )
{
    if( !uMax ) { return 0; }

    Asura_UnicodeChar wA, wB;

    do
    {
        wA = towlower( *wszA++ );
        wB = towlower( *wszB++ );
    }
    while( --uMax && wA && wA == wB );

    return( wA - wB );
}


Asura_UnicodeChar* wcsrchr( const Asura_UnicodeChar* wszHaystack, Asura_UnicodeChar wNeedle )
{
    // The dubious casting in here seems to be required by the ANSI spec for
    // the function signature...

    Asura_UnicodeChar* const wszStart = const_cast<Asura_UnicodeChar*>( wszHaystack );

    // Go to end of string.
    while( *wszHaystack++ ) { }

    // And search backwards.
    for( ; wszHaystack >= wszStart; --wszHaystack )
    {
        if( *wszHaystack == wNeedle )
        {
            return const_cast<Asura_UnicodeChar*>( wszHaystack );
        }
    }

    return NULL;
}


};
// End extern "C".
