//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_DebugInfo.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     code to draw debugging info to the screen
//
// History:
//
//     (MH  - 10.10.2002) - Created using code originally in Asura_PS2_GS.cpp.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Debugging/Asura_PS2_DebugInfo.h"

#ifndef ASURA_PS2_SONY_SUBMISSION

// Stdlib platform-specific.
#include <devfont.h>

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_DebugText.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_RenderStates.h"
#include "Rendering/Asura_Viewport.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Core/Asura_PS2_OS.h"
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_GS.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#define CONSOLE_OFFX    ( ( ( 4096 - Asura_PS2_GS::GetScreenWidth()  ) / 2 ) << 4 )
#define CONSOLE_OFFY    ( ( ( 4096 - Asura_PS2_GS::GetScreenHeight() ) / 2 ) << 4 )

// G L O B A L S /////////////////////////////////////////////////////////////

//// (MP - 11.10.2002) Set this to a value >= 0 to get information about mesh
//// in specified camera's LOS.
//DEBUGVAR int dbg_iCamForMeshDataInLOS = -1;

static bool g_bEnableDbgInfoDisplay = true;

static bool g_abDisplayToggles[Asura_PS2_DebugInfo::DISPLAYTOGGLE_NUM_TOGGLES];

static int g_iConsoleID;

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugInfo::Initialise
// PARAMETERS   None
// RETURNS      None
// DESCRIPTION  Prepare the console for output to the screen
// AUTHOR       KJL
//
//////////////////////////////////////////////////////////////////////////////
const u_int iConsoleOffsetX = 32;
const u_int iConsoleOffsetY = 48;
void Asura_PS2_DebugInfo::Initialise(void)
{
    sceDevConsInit();
    // (MH - 17.04.2002) note that x and y are in GS 12.4 coords, w/h are in console characters.
    g_iConsoleID = sceDevConsOpen( CONSOLE_OFFX + (iConsoleOffsetX << 4),  CONSOLE_OFFY + (iConsoleOffsetY << 4), 56, 48 );
    sceDevConsClear( g_iConsoleID );

    // Set default toggle values.
    {
        bool bNormal  = Asura_PS2_OS::IsTOOL() || Asura_PS2_OS::IsRunningUnderAtwinmon();
        bool bVerbose = Asura_PS2_OS::IsTOOL();
#ifndef ASURA_PS2_CDEMUL
		// JAT 14/09/06 - if not on a CD build then turn normal on if on a test kit
		if ( Asura_PS2_OS::IsTestKit() )
		{
			bNormal = true;
			// If not on an atwinmon build then put verbose on as well
			if ( !Asura_PS2_OS::IsRunningUnderAtwinmon() )
			{
				bVerbose = true;
			}
		}
#endif // #ifndef ASURA_PS2_CDEMUL

        g_abDisplayToggles[DISPLAYTOGGLE_CAM_POS]                   = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_CAM_DIR_FWD]               = bVerbose;
        g_abDisplayToggles[DISPLAYTOGGLE_CAM_DIR_OTHER]             = false;
        g_abDisplayToggles[DISPLAYTOGGLE_CAM_MODULE]                = false;
        g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_FPS_INDICATOR]       = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_MEM]                 = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_RENDER]              = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_MEM_STATS]                 = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_NETWORK]                   = false;
        g_abDisplayToggles[DISPLAYTOGGLE_RENDER_FPS]                = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_RENDER_TRIS_AND_TEX]       = bNormal;
        g_abDisplayToggles[DISPLAYTOGGLE_RENDER_TRI_PASS_BREAKDOWN] = bVerbose;
        g_abDisplayToggles[DISPLAYTOGGLE_RENDER_MODULE_NOCLIP]      = false;
    }
#if defined(ASURA_USE_DEBUG_TEXT) && defined(ASURA_PS2_USE_DEBUGTEXT_OVERRIDE)
	Asura_DebugText::SetPlatformOverride(RenderDebugText);
#endif
}
// End function Asura_PS2_DebugInfo::Initialise().


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION     Asura_PS2_DebugInfo::Render
// PARAMETERS   None
// RETURNS      None
// DESCRIPTION  Output info to the console
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_PS2_DebugInfo::Render(void)
{
    if( !Asura_PS2_OS::IsTOOL() && !Asura_PS2_OS::IsTestKit() ) { return; }

    if( !g_bEnableDbgInfoDisplay ) { return; }

    Asura_PS2_GS::AllowFullScreenRendering();

    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1 );
    Asura_PlatformRenderStates::ApplyRenderStates();

    sceDevConsLocate( g_iConsoleID, 0, 0 );// (MH - 17.04.2002) strangely, setting y to non-0 is *really* slow.
    sceDevConsAttribute( g_iConsoleID, 7 );
    
    bool bPrintedAnythingToConsole = false;

    #define PRINT_TO_CONSOLE( args ) \
    { \
        sceDevConsPrintf args ;\
        bPrintedAnythingToConsole = true; \
    }
    
    // Memory.
    if( g_abDisplayToggles[DISPLAYTOGGLE_MEM_STATS] || g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_MEM] )
    {
        int iFreeMem;
        int iUsedMem;
        int iMaxUsedMem;
        int iMinFreeMem;
        Asura_PS2_OS::GetMemStats(iUsedMem, iMaxUsedMem, iFreeMem, iMinFreeMem);

        // I think Kb is enough detail...
        iFreeMem >>= 10;
        iUsedMem >>= 10;
        iMaxUsedMem >>= 10;
        iMinFreeMem >>= 10;

        if( g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_MEM] )
        {
            DrawMemoryBar( iUsedMem, iMaxUsedMem, iFreeMem + iUsedMem );
        }

        if( g_abDisplayToggles[DISPLAYTOGGLE_MEM_STATS] )
        {
//          PRINT_TO_CONSOLE(( g_iConsoleID, "free %i min %i\n", iFreeMem, iMinFreeMem ));
//          PRINT_TO_CONSOLE(( g_iConsoleID, "used %i max %i\n", iUsedMem, iMaxUsedMem ));

            // (MP - 06.02.2005) I don't think anyone really ever looks at
            // much except this. We could restore some of the above on a
            // separate DISPLAYTOGGLE, if necessary, but this should be the
            // default at least.
            PRINT_TO_CONSOLE(( g_iConsoleID, "free %i\n", iFreeMem ));
        }
    }

    // FPS.
    if( g_abDisplayToggles[DISPLAYTOGGLE_RENDER_FPS] )
    {
        char szOutputBuffer[40];
		Asura_PrintfToBfr( szOutputBuffer, "fps %5.1f %5.1f\n", 40, true,
						   Asura_Maths::FtoD( Asura_Timers::GetSmoothedFrameRate() ), 
						   Asura_Maths::FtoD( Asura_Timers::GetFrameRate()) );
        PRINT_TO_CONSOLE( ( g_iConsoleID, szOutputBuffer ) );
    }
    if( g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_FPS_INDICATOR] )
    {
        DrawFrameRateBar();
    }

    // Camera position.
    if( g_abDisplayToggles[DISPLAYTOGGLE_CAM_POS] )
    {
        const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();
        char szOutputBuffer[40];
		Asura_PrintfToBfr( szOutputBuffer, "cam @ %.1f %.1f %.1f\n", 40, true, Asura_Maths::FtoD(pxCamPos->x), Asura_Maths::FtoD(pxCamPos->y), Asura_Maths::FtoD(pxCamPos->z) );
        PRINT_TO_CONSOLE(( g_iConsoleID, szOutputBuffer ));
    }

    // Camera orientation.
    if( g_abDisplayToggles[DISPLAYTOGGLE_CAM_DIR_FWD] )
    {
        const Asura_Matrix_3x3* pxCamOri = Asura_Camera::GetOrientation();
        char szOutputBuffer[40];
		Asura_PrintfToBfr( szOutputBuffer, "cam F %.1f %.1f %.1f\n", 40, true, Asura_Maths::FtoD(pxCamOri->e13), Asura_Maths::FtoD(pxCamOri->e23), Asura_Maths::FtoD(pxCamOri->e33) );
        PRINT_TO_CONSOLE(( g_iConsoleID, szOutputBuffer ));
    }
    if( g_abDisplayToggles[DISPLAYTOGGLE_CAM_DIR_OTHER] )
    {
        const Asura_Matrix_3x3* pxCamOri = Asura_Camera::GetOrientation();
        char szOutputBuffer[40];
		Asura_PrintfToBfr( szOutputBuffer, "cam R %.1f %.1f %.1f\n", 40, true, Asura_Maths::FtoD(pxCamOri->e11), Asura_Maths::FtoD(pxCamOri->e21), Asura_Maths::FtoD(pxCamOri->e31) );
        PRINT_TO_CONSOLE(( g_iConsoleID, szOutputBuffer ));
		Asura_PrintfToBfr( szOutputBuffer, "cam U %.1f %.1f %.1f\n", 40, true, Asura_Maths::FtoD(pxCamOri->e12), Asura_Maths::FtoD(pxCamOri->e22), Asura_Maths::FtoD(pxCamOri->e32) );
        PRINT_TO_CONSOLE(( g_iConsoleID, szOutputBuffer ));
    }

    // Module containing camera.
    {
        const Asura_Vector_3& xCamPos = *Asura_Camera::GetPosition();

        static int iCamModuleIdx = -1;
        iCamModuleIdx = Asura_Module_System::GetModuleIndexFromPosition( xCamPos, iCamModuleIdx );

        // If we're not in any module, always display an error message.
        // Otherwise only display currently module if that option is enabled.
        if( -1 == iCamModuleIdx )
        {
            // Skip error message in menus, etc.
            if( Asura_Module_System::GetNumberOfModules() )
            {
                // Change colour to red so it's obvious.
                sceDevConsAttribute( g_iConsoleID, 2 );

                PRINT_TO_CONSOLE(( g_iConsoleID, "CAM NOT IN ANY MODULE\n" ));

                // Reset colour.
                sceDevConsAttribute( g_iConsoleID, 7 );
            }
        }
        else if( g_abDisplayToggles[DISPLAYTOGGLE_CAM_MODULE] )
        {
            const u_int uBFR_SIZE = 128;
            char szList[uBFR_SIZE] = "";
            u_int uLen = 0;
        
            for( int iModule = 0; uLen + 1 < uBFR_SIZE && iModule < Asura_Module_System::GetNumberOfModules(); ++iModule )
            {
                const Asura_Module* pxModule = Asura_Module_System::GetModuleData( iModule );
                if( pxModule->m_pxCollisionMesh->GetBoundingBox().Contains(xCamPos - Asura_Module_System::GetPositionOfModule(iModule)) )
                {
                    uLen += Asura_PrintfToBfr( &szList[uLen], "%s%i", uBFR_SIZE - uLen, true, (uLen ? "," : ""), iModule );
                }
            }

            PRINT_TO_CONSOLE(( g_iConsoleID, "modules %s\n", szList ));
        }
    }
    
    // Tri and texture counts.
    if( g_abDisplayToggles[DISPLAYTOGGLE_RENDER_TRIS_AND_TEX] )
    {
#if ASURA_PS2_TEXTUREMANAGER__PROFILING
        PRINT_TO_CONSOLE(( g_iConsoleID,
                           "tex %u / %u  %uKb\n",
                           Asura_PS2_TextureManagement::Profile_GetNumUploads(),
                           Asura_PS2_TextureManagement::Profile_GetNumUploads() + Asura_PS2_TextureManagement::Profile_GetNumCacheHits(),
                           (Asura_PS2_TextureManagement::Profile_GetNumUploadedBytes() >> 10) ));
#endif

        PRINT_TO_CONSOLE(( g_iConsoleID, "tris %i / %i\n", Asura_Render::GetNumberOfTrianglesRendered(), Asura_Render::GetNumTriPassesRendered() ));
    }

#if ASURA_PS2_COLLECT_RENDER_STATS
    // Tri count breakdown.
    if( g_abDisplayToggles[DISPLAYTOGGLE_RENDER_TRI_PASS_BREAKDOWN] )
    {
        const u_int uEnv    = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_ENV         );
        const u_int uHier   = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_HIER        );
        const u_int uSmooth = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_SMOOTH_SKIN );
        const u_int uObj    = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_OBJ         );
        const u_int uWater  = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_WATER       );
        const u_int uShadow = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_SHADOW      );
		const u_int uCarpet = Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_CARPET		);
		const u_int uBots	= Asura_PS2_Render::GetTriPassCount( Asura_PS2_Render::TRICOUNT_BOTANICALS	);
        const u_int uTotal      = Asura_Render::GetNumTriPassesRendered();
        const u_int uKnownTotal = uEnv + uHier + uSmooth + uObj + uWater + uShadow + uCarpet + uBots;

        Asura_Assert( uTotal >= uKnownTotal, "" );

        PRINT_TO_CONSOLE( ( g_iConsoleID,
                            "e %u hs %u ss %u o %u w %u sh %u ? %u\n",
                            uEnv,
                            uHier,
                            uSmooth,
                            uObj,
                            uWater,
						    uShadow,
                            uTotal - uKnownTotal ) );
#if defined( ASURA_USE_CARPET ) || defined( ASURA_USE_BOTANICALS )
        PRINT_TO_CONSOLE( ( g_iConsoleID,
                            "b %u c %d\n",
						    uBots, uCarpet ) );
#endif // #if defined ASURA_USE_CARPET || defined ASURA_USE_BOTANICALS
    }

    // Percentage of modules going down no-clip path.
    if( g_abDisplayToggles[DISPLAYTOGGLE_RENDER_MODULE_NOCLIP] )
    {
        char szOutputBuffer[40];
        float fRatio = 100.0f * Asura_PS2_Render::GetNoClipModulesRatio();
		Asura_PrintfToBfr( szOutputBuffer, "noclip modules %.1f%%\n", 40, true, Asura_Maths::FtoD(fRatio) );
        PRINT_TO_CONSOLE(( g_iConsoleID, szOutputBuffer ) );

//      extern int g_NumModulesRendered;
//      extern int g_NumModulesClipped;
//      extern int g_NumModulesNotClipped;
//      Asura_Assert( g_NumModulesClipped + g_NumModulesNotClipped == g_NumModulesRendered, "" );
//      const float fHundredOverNumModulesRendered = 100.0f / g_NumModulesRendered;
//      PRINT_TO_CONSOLE(( g_iConsoleID,
//                         "Module clipping: %.1f%% clipped, %.1f%% not clipped\n",
//                         Asura_Maths::FtoD(fHundredOverNumModulesRendered * g_NumModulesClipped),
//                         Asura_Maths::FtoD(fHundredOverNumModulesRendered * g_NumModulesNotClipped) ));
    }
#endif // #if ASURA_PS2_COLLECT_RENDER_STATS

//  if( g_abDisplayToggles[DISPLAYTOGGLE_CAM_DIRS_AND_MODULE] )
//  {
//      if( dbg_iCamForMeshDataInLOS >= 0 && dbg_iCamForMeshDataInLOS < Asura_Camera::GetNumberOfCameras() )
//      {
//          Asura_Camera::SetCameraIndex( dbg_iCamForMeshDataInLOS );
//
//          // Get fwd vector.
//          const Asura_Matrix_3x3* pxOri = Asura_Camera::GetOrientation();
//          const Asura_Vector_3 xFwdVec( pxOri->e13,
//                                        pxOri->e23,
//                                        pxOri->e33 );
//
//          int iModuleIdx;
//          bool bHit = Asura_Module_System::FindPolyInLOS( *Asura_Camera::GetPosition(), xFwdVec, iModuleIdx );
//          if( bHit )
//          {
//              PRINT_TO_CONSOLE(( g_iConsoleID, "module in cam %i's LOS: %i\n", dbg_iCamForMeshDataInLOS, iModuleIdx ));
//          }
//      }
//  }
    
//#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
//#if ASURA_PS2_SCE_RT_BANDWIDTH_TEXT
//  if( g_abDisplayToggles[DISPLAYTOGGLE_NETWORK] )
//  {
//      PRINT_TO_CONSOLE(( g_iConsoleID, Asura_PS2_SCE_RT::GetBandwidthString() ));
//  }
//#endif
//#endif

#if _DEBUG
    Asura_PS2_DebugSPR::VerifyAllDeallocated( false );
#endif

    if( bPrintedAnythingToConsole )
    {
        sceDevConsDrawS( g_iConsoleID );
        sceDevConsClear( g_iConsoleID );
    }

    if( g_abDisplayToggles[DISPLAYTOGGLE_GRAPH_RENDER] )
    {
        ASURA_PS2_PROFILE__RENDER
    }

    #undef PRINT_TO_CONSOLE
}
// End Function 'Asura_PS2_DebugInfo::Render'

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugInfo::Shutdown().
// PARAMETERS   None
// RETURNS      None
// DESCRIPTION  None
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_DebugInfo::Shutdown( void )
{
}
// End function Asura_PS2_DebugInfo::ShutdownConsole().

int Asura_PS2_DebugInfo::GetConsoleID( void )
{
    return g_iConsoleID;
}

bool Asura_PS2_DebugInfo::GetDisplayToggle( DISPLAYTOGGLE eToggle )
{
    Asura_Assert( eToggle < DISPLAYTOGGLE_NUM_TOGGLES, "" );
    return g_abDisplayToggles[eToggle];
}

void Asura_PS2_DebugInfo::SetDisplayToggle( DISPLAYTOGGLE eToggle, bool bOn )
{
    Asura_Assert( eToggle < DISPLAYTOGGLE_NUM_TOGGLES, "" );
    g_abDisplayToggles[eToggle] = bOn;
}

void Asura_PS2_DebugInfo::EnableDisplay( bool bEnable )
{
    g_bEnableDbgInfoDisplay = bEnable;
}

bool Asura_PS2_DebugInfo::IsDisplayEnabled( void )
{
    return g_bEnableDbgInfoDisplay;
}

void Asura_PS2_DebugInfo::RenderBBForSphere( const Asura_Vector_3& xCentre, float fRadius )
{
    const Asura_Bounding_Box xBB = { { { -fRadius, fRadius, -fRadius, fRadius, -fRadius, fRadius } } };

    Asura_Render::RenderBB( xBB, xCentre, Asura_Maths::IdentityMatrix3x3, 0xFFFFFFFFu );
}
        
        
//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugInfo::DrawFrameRateBar().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Draws bar to indicate current frame rate in a graphical manner
// AUTHOR       Mike Healey - 19.08.2002.
//
//////////////////////////////////////////////////////////////////////////////
DEBUGVAR u_int dbg_uFrameRateColourOK  = 0xff40ff40;   // ARGB 0-ff
DEBUGVAR u_int dbg_uFrameRateColourLow = 0xffff4040;   // ARGB 0-ff
DEBUGVAR float dbg_fFrameRateBarX = 0.06f;
DEBUGVAR float dbg_fFrameRateBarY = 0.05f;
DEBUGVAR float dbg_fFrameRateBarH = 0.008f;
DEBUGVAR float dbg_fFudgeFactor = 0.02f;

// replacement for using Asura_Overlay_Quad to render debug info after EndFrame.
// Adapted from profiler rendering code.
void Asura_PS2_DebugInfo::RenderDebugQuad( float fX, float fY, float fW, float fH, u_int uARGB )
{
	// Render the profile bar
    Asura_PS2_QW* const pxBase = Asura_PS2_SPR_Alloc( 0, 1024 );	// We don't know how much we'll need yet
	Asura_PS2_QW* pxOut = pxBase;

	// Create a DMA tag - fill in actual dma tag at the end when we know how much we wrote
//	pxOut->u32s[0] = PY_DMA_END | uNLoop+1;
	pxOut->u32s[1] = 0;
	pxOut->u32s[2] = SCE_VIF1_SET_NOP(0);
//	pxOut->u32s[3] = SCE_VIF1_SET_DIRECT( uNLoop+1, 0 );
	pxOut++;

	// Leave space for the gif tag - again filled in when we know howmuch data we wrote
	pxOut->u64s[1] = SCE_GIF_PACKED_AD;
	pxOut++;

    // Setup Test for Z always passes and Alpha always fails and only updates the framebuffer (to disable z writes)
    pxOut->u64s[0] = SCE_GS_SET_TEST_1(	1,					// ATE   = ON
										0,					// ATST  = ALL FAIL
										0,					// AREF  = N/A
										1,					// AFAIL = Frame buffer only (no z writes)
										0,					// DATE  = OFF
										0,					// DATM  = N/A
										1,					// ZTE   = ON
										SCE_GS_ZALWAYS );	// ZTST  = ALWAYS
    pxOut->u64s[1] = SCE_GS_TEST_1;
    pxOut++;

	// Set blend equation
	pxOut->u64s[0] = SCE_GS_SET_ALPHA( 0, 1, 0, 1, 0 );
	pxOut->u64s[1] = SCE_GS_ALPHA_1;
	pxOut++;

	// RGBA
    u_int uR, uG, uB, uA;
    uB = uARGB & 0xffu;
    uARGB >>= 8u;
    uG = uARGB & 0xffu;
    uARGB >>= 8u;
    uR = uARGB & 0xffu;
    uARGB >>= 8u;
    uA = uARGB & 0xffu;
    // [0,255] -> [0,128]
    ++uR;
    ++uG;
    ++uB;
    ++uA;
    uR >>= 1;
    uG >>= 1;
    uB >>= 1;
    uA >>= 1;
	pxOut->u64s[0] = SCE_GS_SET_RGBAQ(uR, uG, uB, uA, 0x3f800000);
	pxOut->u64s[1] = SCE_GS_RGBAQ;
	pxOut++;

	// convert floats to GS coords
	int iOffsetX = ASURA_PS2_SCREEN_CENTRE_X - (Asura_PS2_GS::GetScreenWidth() >> 1);
    int iOffsetY = ASURA_PS2_SCREEN_CENTRE_Y - (Asura_PS2_GS::GetScreenHeight()>> 1);
	int iX = (static_cast<int>(fX) + iOffsetX) << 4;
	int iY = (static_cast<int>(fY) + iOffsetY) << 4;
	int iW = static_cast<int>(fW) << 4;
	int iH = static_cast<int>(fH) << 4;

	// XYZ top-left
    pxOut->u64s[0] = SCE_GS_SET_XYZ( iX, iY, 0 );
    pxOut->u64s[1] = SCE_GS_XYZ2;
    pxOut++;

	// XYZ bottom-right
    pxOut->u64s[0] = SCE_GS_SET_XYZ( iX + iW, iY + iH, 0 );
	pxOut->u64s[1] = SCE_GS_XYZ2;
    pxOut++;

	// Restore pixel test to : disable alpha test and enable depth test
    pxOut->u64s[0] = SCE_GS_SET_TEST_1(	0,					// ATE   = ON
										0,					// ATST  = ALL FAIL
										0,					// AREF  = N/A
										0,					// AFAIL = Frame buffer only (no z writes)
										0,					// DATE  = OFF
										0,					// DATM  = N/A
										1,					// ZTE   = ON
										SCE_GS_ZGEQUAL );	// ZTST  = ALWAYS
    pxOut->u64s[1] = SCE_GS_TEST_1;
    ++pxOut;

    // Calculate how big the dma packet is and update the dma tag
	const u_int uDmaPacketSize = (pxOut - pxBase);
	pxBase[0].u32s[0] = PY_DMA_END | (uDmaPacketSize-1);
	pxBase[0].u32s[3] = SCE_VIF1_SET_DIRECT( (uDmaPacketSize-1), 0 );

	// And the gif tag
	pxBase[1].u64s[0] = SCE_GIF_SET_TAG( uDmaPacketSize-2,		// NLOOP - Sneaked camera count in here !!!!!!
										 1,						// EOP (End Of Packet) - Added by microcode
										 1,						// PRE (PRIM field Enable)
										 SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE,		// PRIM Type
														  1,				// IIP (Flat/Gouraud Shading)
														  0,				// TME (Texture Mapping Enabled)
														  0,				// FGE (Fogging Enabled)
														  1,				// ABE (Alpha Blending Enabled)
														  0,				// AA1 (1 Pass Anti-Aliasing Enabled)
														  0,				// FST (Method of specifying Texture Coords)
														  0,				// CTXT (Context)
														  0 ),				// FIX (Fragment Value Control)
										 SCE_GIF_PACKED,			// FLG (Data format, 0 = PACKED, 1 = REGLIST, 2 = IMAGE)
										 1 );						// NREG (Number of registers)

	// Make sure nothing is drawing at the moment
    Asura_PS2_BetterGsSyncPath( 0, 0 );

	// Fire of the packet
	DPUT_D1_QWC(0);
	// source address
	DPUT_D1_TADR((u_int)pxBase | ASURA_PS2_DMA_SPR_FLAG);   // when transferring from the scratchpad, set the top bit to 1.
	// for polling of completion
	DPUT_D_PCR(D_PCR_CPC1_M);
	DPUT_D_STAT(D_STAT_CIS1_M);

	asm __volatile__("sync.l");

	// start actual kick
	DPUT_D1_CHCR(1  // from memory
		| (1 << 2)  // chain mode
		| (1 << 6)  // transfer tag on
		| (1 << 8));    // dma start

	asm __volatile__("sync.l");
	asm __volatile__("sync.p");

	// Wait for completion.
	asm __volatile__( " _AsuraPS2Profiling_Render_WaitForDMA:         " );
	asm __volatile__( "     bc0f _AsuraPS2Profiling_Render_WaitForDMA " );
	asm __volatile__( "     nop                               " );

	Asura_PS2_SPR_Dealloc( pxBase );
} // end function Asura_PS2_DebugInfo::RenderDebugQuad()

void Asura_PS2_DebugInfo::DrawFrameRateBar()
{
    float fScreenWidth = static_cast<float>( Asura_Viewport::GetWidth() );
    float fScreenHeight = static_cast<float>( Asura_Viewport::GetHeight() );

    const float fMaxFramerate = ( Asura_Timers::GetMaxFrameRate() > 0.0f ? Asura_Timers::GetMaxFrameRate() : 0.5f * Asura_PS2_GS::GetRefreshRate() );
    const float fTargetFramerate = ( Asura_PS2_GS::GetTargetFramerate() > 0.0f ? Asura_PS2_GS::GetTargetFramerate() : 0.5f * Asura_PS2_GS::GetRefreshRate() );

    float fFrameRate = Asura_Timers::GetSmoothedFrameRate();
    float fMaxBarWidth = fScreenWidth * (1.0f - (2.0f * dbg_fFrameRateBarX));
    float fBarWidth = fMaxBarWidth * Asura_Maths::Min(1.0f, fFrameRate / fMaxFramerate);
    float fBarHeight = fScreenHeight * dbg_fFrameRateBarH;

    float fBarX = fScreenWidth * dbg_fFrameRateBarX;
    float fBarY = fScreenHeight * dbg_fFrameRateBarY;

    u_int uColour = ( fFrameRate < fTargetFramerate - dbg_fFudgeFactor ? dbg_uFrameRateColourLow : dbg_uFrameRateColourOK );
	RenderDebugQuad( fBarX, fBarY, fBarWidth, fBarHeight, uColour );
}
// End function Asura_PS2_DebugInfo::DrawFrameRateBar().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_DebugInfo::DrawMemoryBar().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Draws bar to indicate memory usage rate in a graphical manner
// AUTHOR       Mike Healey - 6.1.03
//
//////////////////////////////////////////////////////////////////////////////
DEBUGVAR u_int dbg_uMemoryUsageColourCurrent = 0xff40ff40;   // ARGB 0-ff
DEBUGVAR u_int dbg_uMemoryUsageColourMax = 0xff208020;   // ARGB 0-ff
DEBUGVAR u_int dbg_uMemoryUsageColourLow = 0xff802020;   // ARGB 0-ff
DEBUGVAR u_int dbg_uMemoryUsageColourCurrentLow = 0xffff4040;   // ARGB 0-ff
DEBUGVAR float dbg_fMemoryUsageBarX = 0.06f;
DEBUGVAR float dbg_fMemoryUsageBarY = 0.062f;
DEBUGVAR float dbg_fMemoryUsageBarH = 0.008f;
DEBUGVAR float dbg_fMemSafetyFactor = 128.0f * 1024.0f;

void Asura_PS2_DebugInfo::DrawMemoryBar(int iUsedMem, int iMaxUsedMem, int iTotalMem)
{
    float fScreenWidth = static_cast<float>( Asura_Viewport::GetWidth() );
    float fScreenHeight = static_cast<float>( Asura_Viewport::GetHeight() );

    float fMemoryUsed = static_cast<float>(iUsedMem);
    float fMemoryUsedMax = static_cast<float>(iMaxUsedMem);
    float fTotalMem = static_cast<float>(iTotalMem);
    float fMaxBarWidth = fScreenWidth * (1.0f - (2.0f * dbg_fMemoryUsageBarX));
    float fBarWidth = fMaxBarWidth * fMemoryUsed / fTotalMem;
    float fBarWidth2 = fMaxBarWidth * fMemoryUsedMax / fTotalMem;
    float fBarHeight = fScreenHeight * dbg_fMemoryUsageBarH;

    float fBarX = fScreenWidth * dbg_fMemoryUsageBarX;
    float fBarY = fScreenHeight * dbg_fMemoryUsageBarY;

    u_int uARGB;
	if ( fTotalMem - fMemoryUsedMax > dbg_fMemSafetyFactor )
    {
        uARGB = dbg_uMemoryUsageColourCurrent;
    }
    else
    {
        uARGB = dbg_uMemoryUsageColourCurrentLow;
    }
    RenderDebugQuad(fBarX, fBarY, fBarWidth, fBarHeight, uARGB);

    if ( fTotalMem - fMemoryUsedMax > dbg_fMemSafetyFactor )
    {
        uARGB = dbg_uMemoryUsageColourMax;
    }
    else
    {
        uARGB = dbg_uMemoryUsageColourLow;
    }
    RenderDebugQuad(fBarWidth, fBarY, fBarWidth2 - fBarWidth, fBarHeight, uARGB);
}
// End function Asura_PS2_DebugInfo::DrawMemoryBar().

void Asura_PS2_DebugInfo::ShowMessage( const char* szText )
{
	// check sceDevConsPrintf isn't going to start looking for things to format
	for (int i = 0; i < static_cast<int>(strlen(szText)) - 1; ++i)
	{
		if (szText[i] == '%')
		{
			if (szText[++i] != '%')
			{
				Asura_Assert( false, "message must not contain format specifiers" );
				return;
			}
		}
	}

    Asura_PS2_GS::AllowFullScreenRendering();

    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1 );
    Asura_PlatformRenderStates::ApplyRenderStates();

	sceDevConsLocate( g_iConsoleID, 0, 0 );// (MH - 17.04.2002) strangely, setting y to non-0 is *really* slow.
    sceDevConsAttribute( g_iConsoleID, 7 );

	sceDevConsClear( g_iConsoleID );

	sceDevConsPrintf( g_iConsoleID, szText );

#if _DEBUG
    Asura_PS2_DebugSPR::VerifyAllDeallocated( false );
#endif
	sceDevConsDrawS( g_iConsoleID );
	sceDevConsClear( g_iConsoleID );
}

#if defined(ASURA_USE_DEBUG_TEXT) && defined(ASURA_PS2_USE_DEBUGTEXT_OVERRIDE)
static u_int g_uDebugTextOffset = 0;
void Asura_PS2_DebugInfo::RenderDebugText(const Asura_DebugTextBlock* pxBlock)
{
	const u_int uColour = pxBlock->uColour;
	sceDevConsAttribute(g_iConsoleID, 0);
	sceDevConsSetColor(g_iConsoleID, 0, (uColour & 0xff0000) >> 16, (uColour & 0xff00) >> 8, (uColour & 0xff));
	sceDevConsLocate(g_iConsoleID, pxBlock->uX, 2* pxBlock->uY + g_uDebugTextOffset);
	sceDevConsPrintf(g_iConsoleID, pxBlock->szText);
	sceDevConsDrawS(g_iConsoleID);
	sceDevConsClear(g_iConsoleID);
	sceDevConsSetColor(g_iConsoleID, 0, 0, 0, 0);
}
#endif

#endif//#ifndef ASURA_PS2_SONY_SUBMISSION

