/**
*
* \file Asura_PlatformLogfile.cpp
*
* Created by Kevin Lea - 30/10/2001 17:30:53
* Modified by Morgan Parry - 31.10.2001 - Ported to PS2.
*
* Copyright (C) 2001 Rebellion, All Rights Reserved.
*
* \brief PS2 version of Logfile file access
*
**/

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Standard independent.
#include <errno.h>                          // For ENOENT #define.
#include <stdarg.h>                         // For va_list, etc.
#include <string.h>                         // For sprintf().

// Standard platform-specific.
#include <libcdvd.h>                        // For sceCdCLOCK struct and btoi() and sceCdReadClock().
#include <libscf.h>                         // For sceScfGetLocalTimefromRTC().
#include <sifdev.h>                         // For sceClose(), sceOpen() and sceWrite().
#include <eekernel.h>						// For QueryIntrContext()

#if defined( ASURA_USE_PROVIEW_TTY )
/* Call this to send TTY data out.          */
extern "C" int snputs(const char* szStr);
#endif // #if defined( ASURA_USE_PROVIEW_TTY )

// Engine independent.
#include "Core/Asura_String.h"
#include "Core/Asura_Hash.h"
#include "Debugging/Asura_Logfile.h"

// Engine platform-specific.
#include "Core/Asura_PS2_OS.h"
#include "Sound/Asura_PS2_Multistream.h"

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_LOGFILES_ENABLED

// M A C R O S ///////////////////////////////////////////////////////////////

// Currently passing the SCE_APPEND flag to sceOpen() isn't supported by the
// SN target manager. Turn this on when it is.
#define SCE_APPEND_IS_WORKING (0)

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int LOGFILE_INFO_BUFFER_SIZE   = 256;
static const u_int LOGFILE_OUTPUT_BUFFER_SIZE = 2048;

static const char* DEBUGGER_OUTPUT_LOGFILE = "logfiles\\debugger_output.txt";

// G L O B A L S /////////////////////////////////////////////////////////////

static char g_szInfoBuffer[LOGFILE_INFO_BUFFER_SIZE];
static char g_szOutputBuffer[LOGFILE_OUTPUT_BUFFER_SIZE];
static const Asura_Char* g_szFile;
static unsigned g_uLine;
// (MH - 17.12.2002) evil global
//static Asura_NameTree g_xOpenedLogfiles;
static const int ASURA_PS2_MAX_OPENED_LOGFILES = 16;
static Asura_Hash_ID g_aiOpenedLogfiles[ASURA_PS2_MAX_OPENED_LOGFILES];
static int g_iNumOpenedLogfiles = 0;

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR bool dbg_bLogDebuggerOutputToFile = true;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void Asura_LogfileClear( const Asura_Char* szLogfileName, const Asura_Char* szLogfileFullPath );

static u_int Asura_LogfileGetDateAndTime( Asura_Char* szBuffer );

static inline bool CanLogInCurrentSetup( void )
{
    return( Asura_PS2_OS::IsTOOL() || Asura_PS2_OS::IsRunningUnderAtwinmon() );
}

// F U N C T I O N S /////////////////////////////////////////////////////////
bool Asura_PS2_LogfileHasBeenOpened(const char* szFilename, bool bAdd)
{
    Asura_Hash_ID iHashID = Asura_GetHashID(szFilename);
    for (int i = 0; i < g_iNumOpenedLogfiles; ++i)
    {
        if (iHashID == g_aiOpenedLogfiles[i])
        {
            return true;
        }
    }
    if (bAdd)
    {
        Asura_Assert(g_iNumOpenedLogfiles < ASURA_PS2_MAX_OPENED_LOGFILES, "opened too many logfiles, increase ASURA_PS2_MAX_OPENED_LOGFILES");
        if (g_iNumOpenedLogfiles < ASURA_PS2_MAX_OPENED_LOGFILES)
        {
            g_aiOpenedLogfiles[g_iNumOpenedLogfiles++] = iHashID;
        }
    }
    return false;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_LogfileClear().
// PARAMETERS   Logfile name and full path.
// RETURNS      None.
// DESCRIPTION  Empties file and timestamps it.
// AUTHOR       Morgan Parry - 31.10.2001.
//
//////////////////////////////////////////////////////////////////////////////
static void Asura_LogfileClear( const Asura_Char* szLogfileName, const Asura_Char* szLogfileFullPath )
{
	//if( Asura_PS2_Multistream_FastLoad::AreAnyFilesOpen() ) { return; }
	Asura_PS2_Multistream_FastLoad::PauseFastload();

    // Try and open specified logfile.
    int iFileHandle = sceOpen( szLogfileFullPath, SCE_WRONLY | SCE_CREAT | SCE_TRUNC );
    if( iFileHandle >= 0 )
    {
        u_int uOutputLen = Asura_PrintfToBfr( g_szInfoBuffer, "Asura logfile \"%s\" - opened at ", LOGFILE_INFO_BUFFER_SIZE, false, szLogfileName );

        // Append timestamp.
        if( uOutputLen < LOGFILE_INFO_BUFFER_SIZE - 23 )
        {
            uOutputLen += Asura_LogfileGetDateAndTime( &g_szInfoBuffer[uOutputLen] );
        }

        // Append ".\n\n".
        if( uOutputLen < LOGFILE_INFO_BUFFER_SIZE - 3 )
        {
            g_szInfoBuffer[uOutputLen++] = '.';
            g_szInfoBuffer[uOutputLen++] = '\r';
            g_szInfoBuffer[uOutputLen++] = '\n';
            g_szInfoBuffer[uOutputLen++] = '\r';
            g_szInfoBuffer[uOutputLen++] = '\n';
        }

        // Output.
        sceWrite( iFileHandle, g_szInfoBuffer, uOutputLen );
        
        // Close file.
        sceClose( iFileHandle );
    }
    else
    {
        if( iFileHandle == -ENOENT || iFileHandle == -ENODEV )
        {
            printf( "Failed to open logfile \"%s\" for clearing - invalid path - check fileserver root.\n", szLogfileFullPath );
        }
        else
        {
            printf( "Failed to open logfile \"%s\" for clearing - error %i.\n", szLogfileFullPath, iFileHandle );
        }
    }
	
	Asura_PS2_Multistream_FastLoad::ResumeFastload();
}
// End function Asura_LogfileClear().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     __Asura_LogfileDebugData().
// PARAMETERS   Name of calling file and line within that file.
// RETURNS      None.
// DESCRIPTION  Helper function to set the name of the calling file and the
//              line within that file that the call occurs.
// AUTHOR       Kevin Lea - 30.10.2001.
//
//////////////////////////////////////////////////////////////////////////////
void __Asura_LogfileDebugData( const Asura_Char* __szFile, unsigned __uLine )
{
	g_szFile = __szFile;
	g_uLine = __uLine;
}
// End function __Asura_LogfileDebugData().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_LogfileGetDateAndTime().
// PARAMETERS   Pointer to string buffer.
// RETURNS      Num characters output, excluding nul terminator.
// DESCRIPTION  Writes date and time into buffer in the format
//              "<time> on <date>" - unless it fails to read the clock, in
//              which case an error string is written instead. Buffer must be
//              at least 23 characters in size.
// AUTHOR       Morgan Parry - 01.11.2001.
//
//////////////////////////////////////////////////////////////////////////////
static u_int Asura_LogfileGetDateAndTime( Asura_Char* szBuffer )
{    
    sceCdCLOCK xClock;
    int iClockReadRetVal = sceCdReadClock( &xClock );
    if( !iClockReadRetVal || xClock.stat )
    {
        // Write error message - 22 chars including nul terminator.
        strcpy( szBuffer, "*error reading clock*" );
        Asura_Assert( strlen(szBuffer) == 21, "" );
        return 21;
    }
    else
    {
        // Convert to local time.
        sceScfGetLocalTimefromRTC( &xClock );

        // Write time - 23 chars including nul terminator.  
        int iNumChars = sprintf( szBuffer,
                                 "%02u:%02u:%02u on %02u.%02u.20%02u",
                                 btoi(xClock.hour),
                                 btoi(xClock.minute),
                                 btoi(xClock.second),
                                 btoi(xClock.day),
                                 btoi(xClock.month),
                                 btoi(xClock.year) );

        Asura_Assert( 22 == iNumChars, "" );
        return iNumChars;
    }
}
// End function Asura_LogfileGetDateAndTime().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     __Asura_LogfileOutput().
// PARAMETERS   Name of logfile, format string for output and arguments to
//              format string - ie. printf style.
// RETURNS      None.
// DESCRIPTION  Creates a string according to the specified format and
//              outputs it to the specified logfile.
// AUTHOR       Kevin Lea - 30.10.2001.
// AUTHOR       Morgan Parry - 31.10.2001 - Ported to PS2.
//
//////////////////////////////////////////////////////////////////////////////
void __Asura_LogfileOutput( const Asura_Char* __szLogfile, const Asura_Char* __szFormat, ... )
{
    // (MP - 04.01.2002) Before doing anything with the file, check we can
    // do any logging. If we're not running on a TOOL or on a test PS2 with
    // ProView, we can't log.
    if( !CanLogInCurrentSetup() ) { return; }

	//if( Asura_PS2_Multistream_FastLoad::AreAnyFilesOpen() ) { return; }
	Asura_PS2_Multistream_FastLoad::PauseFastload();

    char szFullPathBackup[ASURA_PS2_FULL_PATH_BFR_SIZE];
    Asura_PS2_OS::BackupFullPath( szFullPathBackup );

    // Construct full logfile path.
#ifdef ASURA_PS2_CDEMUL
    const char* szFullPath = Asura_PS2_OS::GetFullPathOnPC( __szLogfile );
#else
    const char* szFullPath = Asura_PS2_OS::GetFullPath( __szLogfile );
#endif

	// Kevin Lea - 30/10/2001 16:41:33 - check to see if we've already opened
    // a file called this.
	{
		// (MH - 17.12.2002) no need for name tree
        //int iIgnored;
		//if( !g_xOpenedLogfiles.FindNodeIndexByName(__szLogfile, true, iIgnored) )
        if (!Asura_PS2_LogfileHasBeenOpened(__szLogfile, true))
		{
			/// Kevin Lea - 30/10/2001 16:48:46 - this file hasn't been seen before, so let's wipe it
			Asura_LogfileClear( __szLogfile, szFullPath );
		}
	}

    u_int uLogStrLen;

    {
		va_list xArgs;
		va_start( xArgs, __szFormat );
        uLogStrLen = Asura_PrintfToBfr( g_szOutputBuffer, __szFormat, xArgs, LOGFILE_OUTPUT_BUFFER_SIZE, false );
		va_end( xArgs );
    }

    // Add a newline.
    if( LOGFILE_OUTPUT_BUFFER_SIZE - 1 == uLogStrLen )
    {
        // We've hit the buffer limit so overwrite the last character.
        g_szOutputBuffer[uLogStrLen - 1] = '\n';
    }
    else
    {
        g_szOutputBuffer[uLogStrLen++] = '\n';
    }

    // Try and open specified logfile.
#if SCE_APPEND_IS_WORKING
    int iFileHandle = sceOpen( szFullPath, SCE_WRONLY | SCE_APPEND );
#else
    int iFileHandle = sceOpen( szFullPath, SCE_WRONLY );
#endif
    if( iFileHandle >= 0 )
    {
#if !SCE_APPEND_IS_WORKING
        // Move to end of file.
        int iSeekRetVal = sceLseek( iFileHandle, 0, SCE_SEEK_END );
        if( iSeekRetVal >= 0 )
#endif
        {
			u_int uLogInfoLen;
        
#if _DEBUG
			uLogInfoLen = Asura_PrintfToBfr( g_szInfoBuffer, "Logged from %s (%u) at ", LOGFILE_INFO_BUFFER_SIZE, false, g_szFile, g_uLine );
#else
			uLogInfoLen = sprintf( g_szInfoBuffer, "Logged at " );
#endif

			// Append timestamp.
			if( uLogInfoLen < LOGFILE_INFO_BUFFER_SIZE - 23 )
			{
				uLogInfoLen += Asura_LogfileGetDateAndTime( &g_szInfoBuffer[uLogInfoLen] );
			}

			// Append ":\n".
			if( uLogInfoLen < LOGFILE_INFO_BUFFER_SIZE - 2 )
			{
				g_szInfoBuffer[uLogInfoLen++] = ':';
				g_szInfoBuffer[uLogInfoLen++] = '\n';
			}

			// Output.
			sceWrite( iFileHandle, g_szInfoBuffer, uLogInfoLen );
			sceWrite( iFileHandle, g_szOutputBuffer, uLogStrLen );

			// Close file.
		}

        sceClose( iFileHandle );
    }
    else
    {
        if( iFileHandle == -ENOENT || iFileHandle == -ENODEV )
        {
            printf( "ERROR: Failed to open logfile \"%s\" to append log msg - invalid path - check fileserver root.\n", szFullPath );
        }
        else
        {
            printf( "ERROR: Failed to open logfile \"%s\" to append log msg - error value %i.\n", szFullPath, iFileHandle );
        }
    }

    Asura_PS2_OS::RestoreFullPath( szFullPathBackup );

	Asura_PS2_Multistream_FastLoad::ResumeFastload();
}
// End function __Asura_LogfileOutput().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     __Asura_SimpleLogfileBegin().
// PARAMETERS   Name of logfile
// RETURNS      None.
// DESCRIPTION  open specified logfile and writes the time. It is designed to log
//              errors on shutdown, so don't use globals with constructors
//              or functions which do, like Asura_Assert.
// AUTHOR       Mike Healey 18.3.02, based on Morgan's logfile stuff
//
//////////////////////////////////////////////////////////////////////////////
void __Asura_SimpleLogfileClear( const Asura_Char* szLogfile )
{
    if( !CanLogInCurrentSetup() ) { return; }

    char szFullPathBackup[ASURA_PS2_FULL_PATH_BFR_SIZE];
    Asura_PS2_OS::BackupFullPath( szFullPathBackup );

#ifdef ASURA_PS2_CDEMUL
    Asura_LogfileClear( szLogfile, Asura_PS2_OS::GetFullPathOnPC(szLogfile) );
#else
    Asura_LogfileClear( szLogfile, Asura_PS2_OS::GetFullPath(szLogfile) );
#endif

    Asura_PS2_OS::RestoreFullPath( szFullPathBackup );
}


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     __Asura_SimpleLogfileOutput().
// PARAMETERS   Name of logfile, format string for output and arguments to
//              format string - ie. printf style.
// RETURNS      None.
// DESCRIPTION  Creates a string according to the specified format and
//              outputs it to the specified logfile. It is designed to log
//              errors on shutdown, so don't use globals with constructors
//              or functions which do, like Asura_Assert.
// AUTHOR       Mike Healey 18.3.02, based on Morgan's logfile stuff
//
//////////////////////////////////////////////////////////////////////////////
void __Asura_SimpleLogfileOutput( const Asura_Char* szLogFilename, const Asura_Char* szFormat, ... )
{
    if( !CanLogInCurrentSetup() ) { return; }

    u_int uLen;

    {
	    va_list xArgs;
	    va_start( xArgs, szFormat );
        uLen = Asura_PrintfToBfr( g_szOutputBuffer, szFormat, xArgs, LOGFILE_OUTPUT_BUFFER_SIZE, false );
	    va_end( xArgs );
    }

    __Asura_SimpleLogfileOutput_PreFormatted( szLogFilename, g_szOutputBuffer, uLen );
}

void __Asura_SimpleLogfileOutput_PreFormatted( const char* szLogFilename, const char* szStr, u_int uLen )
{
    if( !CanLogInCurrentSetup() ) { return; }

    // Prevent infinite recursion in the case where we can't open a logfile,
    // since otherwise we'll try and report that fact in the debugger and
    // then end up back in here again.
    static bool bRecursed = false;
    if( bRecursed ) { return; }
    bRecursed = true;

	//if( Asura_PS2_Multistream_FastLoad::AreAnyFilesOpen() ) { return; }
	Asura_PS2_Multistream_FastLoad::PauseFastload();

    char szFullPathBackup[ASURA_PS2_FULL_PATH_BFR_SIZE];
    Asura_PS2_OS::BackupFullPath( szFullPathBackup );

#ifdef ASURA_PS2_CDEMUL
    const char* szFullPath = Asura_PS2_OS::GetFullPathOnPC( szLogFilename );
#else
    const char* szFullPath = Asura_PS2_OS::GetFullPath( szLogFilename );
#endif

#if SCE_APPEND_IS_WORKING
    int iFileHandle = sceOpen( szFullPath, SCE_WRONLY | SCE_APPEND );
#else
    int iFileHandle = sceOpen( szFullPath, SCE_WRONLY );
#endif
    if( iFileHandle >= 0 )
    {
#if !SCE_APPEND_IS_WORKING
        // Move to end of file.
        int iSeekRetVal = sceLseek( iFileHandle, 0, SCE_SEEK_END );
        if( iSeekRetVal >= 0 )
#endif
        {
			// Output the 'string' in upto each /n marker
			const char* pcOutput = szStr;
			u_int uOutputLen = 0;
			for ( u_int uI = 0; uI < uLen; uI++ )
			{
				if ( pcOutput[uOutputLen] != '\n' )
				{
					uOutputLen++;
				}
				else
				{
					// Catch case of /n/n
					if ( uOutputLen > 0 )
					{
						sceWrite( iFileHandle, pcOutput, uOutputLen );
					}
					// Output PC end of line marker
					const char acPCeol[2] = { 0x0D, 0x0A };
            		sceWrite( iFileHandle, acPCeol, 2 );

					pcOutput += ( uOutputLen + 1 );
					uOutputLen = 0;
				}
			}
			if ( uOutputLen > 0 )
			{
				// Catch any straggles
				sceWrite( iFileHandle, pcOutput, uOutputLen );
				// Output PC end of line marker
				const char acPCeol[2] = { 0x0D, 0x0A };
       			sceWrite( iFileHandle, acPCeol, 2 );
			}
        }
        sceClose( iFileHandle );
    }
    else
    {
        if( iFileHandle == -ENOENT || iFileHandle == -ENODEV )
        {
            Asura_OutputToDebugger_Retail(( "ERROR: Failed to open logfile \"%s\" to append log msg - invalid path - check fileserver root.\n", szFullPath ));
        }
        else
        {
            Asura_OutputToDebugger_Retail(( "ERROR: Failed to open logfile \"%s\" to append log msg - error value %i.\n", szFullPath, iFileHandle ));
        }
    }

    Asura_PS2_OS::RestoreFullPath( szFullPathBackup );

    bRecursed = false;

	Asura_PS2_Multistream_FastLoad::ResumeFastload();
}

bool __Asura_Logfile_CanLogYet( void )
{
#ifdef ASURA_PS2_CDEMUL
    return Asura_PS2_OS::IsTOOL();
#else
    return( CanLogInCurrentSetup() && *ASURA_PS2_FILEROOT );
#endif
}


bool __Asura_OverrideLogfilesDir( const char* szDir, bool bArchiveExisting )
{
	return false;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     __Asura_OutputToDebugger().
// PARAMETERS   szFormat - Format string.
//              ...      - Arguments for format string.
// RETURNS      None.
// DESCRIPTION  Prepares a string according to the specified format and 
//              outputs it to the debugger.
// AUTHOR       Morgan Parry - 25.04.2002.
//
//////////////////////////////////////////////////////////////////////////////
void __Asura_OutputToDebugger( const Asura_Char* szFormat, ... )
{
    u_int uLen;

	// Kevin Lea - 30/10/2001 16:55:50 - process the format string and any
    // other parameters into a buffer, just like printf does.
    {
		va_list xArgs;
		va_start( xArgs, szFormat );
        uLen = Asura_PrintfToBfr( g_szOutputBuffer, szFormat, xArgs, LOGFILE_OUTPUT_BUFFER_SIZE, true );
		va_end( xArgs );
	}


    // Echo to TTY.
	if (QueryIntrContext())
	{
		scePrintf( "%s", g_szOutputBuffer );
	}
	else
	{
		fputs( g_szOutputBuffer, stdout );
	}
#if defined( ASURA_USE_PROVIEW_TTY )
	snputs( g_szOutputBuffer );
#endif // #if ASURA_USE_PROVIEW

    // (MP - 05.01.2004) Useful, I think, to log all of this stuff as well.
    // However, we need to make sure it's safe first, since we may output to
    // the debugger before anything has been initialised.
	// (PJC 15.02.2008) Support for asserts in interrupts
    if( dbg_bLogDebuggerOutputToFile && __Asura_Logfile_CanLogYet() && (QueryIntrContext() == 0) )
    {
        static bool bFirst = true;
        if( bFirst )
        {
            __Asura_SimpleLogfileClear( DEBUGGER_OUTPUT_LOGFILE );
            bFirst = false;
        }

        __Asura_SimpleLogfileOutput_PreFormatted( DEBUGGER_OUTPUT_LOGFILE, g_szOutputBuffer, uLen );
    }
}
// End function __Asura_OutputToDebugger().

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifdef ASURA_LOGFILES_ENABLED

