
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_ScreenShots.cpp (PS2 version)
// Author : Andy Tate
// Date   : 10-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_SCREENSHOT_SYSTEM

// This file's header...
#include "Debugging/Asura_ScreenShot_System.h"

// Asura header-files...
#include "Core/Asura_Stream.h"
#include "Core/Asura_String.h"
#include "Core/Asura_UserInput.h"
#include "Rendering/Asura_Viewport.h"

#include "Core/Asura_PS2_OS.h"
#include "Debugging/Asura_PS2_DebugInfo.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_VU1.h"
#include "Rendering/Asura_TIM2.h"

// Game header-files...


////////////////////////////////////////////////////////////////////////////////
// Type declarations...
////////////////////////////////////////////////////////////////////////////////

// structure for information to take snap shots of arbitrary areas of VRAM
struct VRAM_SHOT_DATA
{
    bool  TakeNow;
    u_int tbp;
    u_int tbw;
    u_int psm;
    short x;
    short y;
    short w;
    short h;
};

// PS2 pixels formats
struct PS2_16BITPIXEL
{
    unsigned short R : 5;
    unsigned short G : 5;
    unsigned short B : 5;
    unsigned short A : 1;
};

struct PS2_32BITPIXEL
{
    u_int R : 8;
    u_int G : 8;
    u_int B : 8;
    u_int A : 8;
};


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// S T A T I C V A R S /////////////////////////////////////////////////////////
static bool	g_bOldVsync	= false;

// D E B U G V A R S ///////////////////////////////////////////////////////////
DEBUGVAR bool	dbg_bScreenShotsAsTM2s	= false;

#if _DEBUG
static VRAM_SHOT_DATA dbg_VRAMShot;
#endif

DEBUGVAR float dbg_fMsgX = 0.01f;
DEBUGVAR float dbg_fMsgY = 0.085f;
DEBUGVAR float dbg_fMsgW = 0.98f;
DEBUGVAR float dbg_fMsgH = 0.035f;

////////////////////////////////////////////////////////////////////////////////
// static functions
////////////////////////////////////////////////////////////////////////////////
static void Screenshots_DrawMessageBackground()
{
    float fScreenWidth = static_cast<float>( Asura_Viewport::GetWidth() );
    float fScreenHeight = static_cast<float>( Asura_Viewport::GetHeight() );
	// background for white text
	Asura_PS2_DebugInfo::RenderDebugQuad(dbg_fMsgX * fScreenWidth, 
										 dbg_fMsgY * fScreenHeight, 
										 dbg_fMsgW * fScreenWidth, 
										 dbg_fMsgH * fScreenHeight, 
										 0xcc000000);
}

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_RenderProgressBar(const Asura_Vector_2& xPosition, 
														 const Asura_Vector_2& xSize, 
														 const u_int uARGB)
{
	Screenshots_DrawMessageBackground();

	char acBuffer[256];
	u_int uMag = GetMagnification();
	u_int uWidth = Asura_PS2_GS::GetScreenWidth();
	u_int uHeight = Asura_PS2_GS::GetScreenHeight();
	Asura_PrintfToBfr( acBuffer, "Taking hi-res shot (%ux%u, %uX AA)", 256, true, 
		uWidth * uMag, uHeight * uMag, GetAntiAliasing() );
	Asura_PS2_DebugInfo::ShowMessage( acBuffer );

	Asura_PS2_DebugInfo::RenderDebugQuad(xPosition.x, xPosition.y, xSize.x, xSize.y, uARGB);
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_GetScreenDimensions
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Gets the specified screen dimensions.
//
// Parameters   : Pointers to the position, size and bit-depth to get, if these
//				  are NULL then don't bother fetching the data.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_GetScreenDimensions (Asura_IntVector_2* pxPosition,
															Asura_IntVector_2* pxSize,
															u_int* puBitDepth)
{
	// Do we need to get the position?

	if (NULL != pxPosition)
	{
		// Yes - do so...

		pxPosition->iX	= 0;
		pxPosition->iY	= 0;
	}


	// Do we need to get the size?

	if (NULL != pxSize)
	{
		// Yes - do so...

		pxSize->iX	= Asura_PS2_GS::GetScreenWidth();
		pxSize->iY	= Asura_PS2_GS::GetScreenHeight();
	}


	// Do we need to get the bit-depth?

	if (NULL != puBitDepth)
	{
		// (MH - 23.05.2005) to get the FBP of the draw buffer, we need to call GetDispBfrFBP().
		// This is because we are rendering directly, not using deferred rendering and the function
		// accounts for the 1-frame delay. (If using 32-bit back and 24-bit front GetDispBfrFBP()
		// and GetDrawBfrFBP() always return the same anyway).
		u_int	uFBP	= Asura_PS2_GS::GetDispBfrFBP();
		u_int	tbp		= uFBP << 5;
		u_int	tbw		= (Asura_PS2_GS::GetScreenWidth() + 63) >> 6;
		u_int	psm		= Asura_PS2_GS::GetDrawBfrPSM();


		// Decide upon the bit-depth...

		switch (psm)
		{
			case SCE_GS_PSMCT16:
			case SCE_GS_PSMCT16S:
			{
				*puBitDepth	= 16;
				break;
			}

			case SCE_GS_PSMCT32:
			{
				*puBitDepth	= 32;
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected screen bit-depth found.");
				break;
			}
		}		// End of switch.
	}
}

/// End of function 'Asura_ScreenShot_System::Platform_GetScreenDimensions'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_WriteScreenAs24Bit
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Writes the specified portion of the screen to the specified
//				  stream as a 24-bit BMP.
//
// Parameters   : A reference to the stream to write to, along with the position
//				  size and format of the screen.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_WriteScreenAs24Bit (Asura_Stream& xStream,
														   const Asura_IntVector_2& xPosition,
														   const Asura_IntVector_2& xSize,
														   const u_int uBitDepth)
{
	// (MH - 30.06.2005) note that the bit depth parameter is ignored as we take
	// screenshots from the 32-bit draw buffer. The 16-bit case below is commented-out
	// (DJP - 11.04.2005) put 16-bit support back in for the unusual case where we're
	// using 16-bit double buffers (as in The Simpsons).
	bool bIsScreen32Bit = (uBitDepth == 32) ? true : false;

	// (MH - 23.05.2005) to get the FBP of the draw buffer in double-buffered mode
	// we need to call GetDispBfrFBP(). This is because we are rendering directly, 
	// not using deferred rendering and the function accounts for the 1-frame delay. 
	// (If using 32-bit back and 16-bit front GetDispBfrFBP() and GetDrawBfrFBP() 
	// always return the same value so we must call GetDrawBfrFBP()).
	bool bSingleBufferedFrameBuffers = (16 == Asura_PS2_GS::GetDispBfrBitDepth() 
										&& 32 == Asura_PS2_GS::GetDrawBfrBitDepth());

	u_int uFBP = bSingleBufferedFrameBuffers ? 
					Asura_PS2_GS::GetDrawBfrFBP() : 
					Asura_PS2_GS::GetDispBfrFBP();
	u_int tbp = uFBP << 5;
	u_int tbw = (Asura_PS2_GS::GetScreenWidth() + 63) >> 6;
	u_int psm = Asura_PS2_GS::GetDrawBfrPSM();
	short x = xPosition.iX;
	short y = xPosition.iY;
	short w = xSize.iX;
	short h = xSize.iY;

	// Ensure all paths are idle
	sceGsSyncPath( 0, 0 );

	// Store the DMA channel 1 settings since these are changed by sceGsExecStoreImage().
	u_int	store_D1_CHCR = *D1_CHCR;

	// Write image data (upside-down)
	u_int uBytesPerLine = w * 3;
	u_char* buf2 = reinterpret_cast<u_char*>(Asura_AllocateAlign(4096, ASURA_ALIGN_16_BYTE));
	if (buf2)
	{
		if (bIsScreen32Bit)
		{
			for( int i = h - 1; i >= 0; --i )
			{
				PS2_32BITPIXEL buf[1024];			// Reserve 4KB raster buffer
				
				sceGsStoreImage si;

				sceGsSetDefStoreImage( &si, tbp, tbw, psm, x, y + i, w, 1);
				FlushCache( WRITEBACK_DCACHE );				// Flush D cache

				sceGsExecStoreImage( &si, (u_long128*)buf );	// Start transfer to VRAM
			
				FlushCache( WRITEBACK_DCACHE );				// Flush D cache

				u_char* puc24Bit = buf2;
				for (int j = 0; j < w; ++j)
				{
					PS2_32BITPIXEL xPixel = buf[j];

					*puc24Bit++ = u_char(xPixel.B);
					*puc24Bit++ = u_char(xPixel.G);
					*puc24Bit++ = u_char(xPixel.R);
				}

				FlushCache( WRITEBACK_DCACHE );				// Flush D cache

				xStream.WriteData	(uBytesPerLine, buf2);						// Write data for one raster line
			}
		}
		else
		{
			// 16-bit
			for( int i = h - 1; i >= 0; --i )
			{
				PS2_16BITPIXEL buf[2048];			// Reserve 4KB raster buffer
				
				sceGsStoreImage si;

				sceGsSetDefStoreImage( &si, tbp, tbw, psm, x, y + i, w, 1);
				FlushCache( WRITEBACK_DCACHE );				// Flush D cache

				sceGsExecStoreImage( &si, (u_long128*)buf );	// Start transfer to VRAM
			
				FlushCache( WRITEBACK_DCACHE );				// Flush D cache

				u_char* puc24Bit = buf2;
				for (int j = 0; j < w; ++j)
				{
					PS2_16BITPIXEL xPixel = buf[j];

					*puc24Bit++ = u_char((xPixel.B * 255) / 31);
					*puc24Bit++ = u_char((xPixel.G * 255) / 31);
					*puc24Bit++ = u_char((xPixel.R * 255) / 31);
				}

				FlushCache( WRITEBACK_DCACHE );				// Flush D cache

				xStream.WriteData	(uBytesPerLine, buf2);						// Write data for one raster line
			}
			
		}

		Asura_DeallocateAlign( buf2 );
		buf2 = 0;
	}

	// Restore the DMA channel 1 settings.
	*D1_CHCR = store_D1_CHCR;
}

/// End of function 'Asura_ScreenShot_System::Platform_WriteScreenAs24Bit'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_AddScreenToFPBuffer
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Adds the specified portion of the screen to the floating-point
//				  RGB buffer.
//
// Parameters   : 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_AddScreenToFPBuffer (const Asura_IntVector_2& xPosition,
															const Asura_IntVector_2& xSize,
															const u_int uBitDepth,
															float* pfFPBuffer)
{
	if (32 != uBitDepth)
	{
		Asura_Assert(false, "bit depth not supported for FP buffer output");
		return;
	}

	// (MH - 23.05.2005) to get the FBP of the draw buffer in double-buffered mode
	// we need to call GetDispBfrFBP(). This is because we are rendering directly, 
	// not using deferred rendering and the function accounts for the 1-frame delay. 
	// (If using 32-bit back and 16-bit front GetDispBfrFBP() and GetDrawBfrFBP() 
	// always return the same value so we must call GetDrawBfrFBP()).
	bool bSingleBufferedFrameBuffers = (16 == Asura_PS2_GS::GetDispBfrBitDepth() 
										&& 32 == Asura_PS2_GS::GetDrawBfrBitDepth());

	u_int uFBP = bSingleBufferedFrameBuffers ? 
					Asura_PS2_GS::GetDrawBfrFBP() : 
					Asura_PS2_GS::GetDispBfrFBP();
	u_int tbp = uFBP << 5;
	u_int tbw = (Asura_PS2_GS::GetScreenWidth() + 63) >> 6;
	u_int psm = Asura_PS2_GS::GetDrawBfrPSM();
	short x = xPosition.iX;
	short y = xPosition.iY;
	short w = xSize.iX;
	short h = xSize.iY;

	// Ensure all paths are idle
	sceGsSyncPath( 0, 0 );

	// Store the DMA channel 1 settings since these are changed by sceGsExecStoreImage().
	u_int	store_D1_CHCR = *D1_CHCR;

	//void* pRasterBuf = Asura_AllocateAlign( 4096, ASURA_ALIGN_128_BYTE );
	//void* pStoreImage = Asura_AllocateAlign( sizeof(sceGsStoreImage), ASURA_ALIGN_64_BYTE );

	//if (pRasterBuf && pStoreImage)
	{
		//PS2_32BITPIXEL* pxBuf = reinterpret_cast<PS2_32BITPIXEL*>(Asura_PS2_UncachedMem(pRasterBuf));
		//sceGsStoreImage* pxStoreImage = reinterpret_cast<sceGsStoreImage*>(Asura_PS2_UncachedMem(pStoreImage));
		sceGsStoreImage xStoreImage;
		sceGsStoreImage* pxStoreImage = &xStoreImage;
		//PS2_32BITPIXEL pxBuf[1024];
		PS2_32BITPIXEL* pxBuf = reinterpret_cast<PS2_32BITPIXEL*>(Asura_PS2_SPR_Alloc( 0, 256 ));

		// call sceGsSetDefStoreImage with 0s for size to set up the unchanging bits.
		sceGsSetDefStoreImage( pxStoreImage, tbp, tbw, psm, 0, 0, 0, 0);

		for( int i = h - 1; i >= 0; --i )
		{
			// just set the bits of the sceGsStoreImage struct that changed
			*(u_long *)&pxStoreImage->trxpos = SCE_GS_SET_TRXPOS(x, y + i, 0, 0, 0);
			*(u_long *)&pxStoreImage->trxreg = SCE_GS_SET_TRXREG(w, 1);

			FlushCache( WRITEBACK_DCACHE );				// Flush D cache

			sceGsExecStoreImage( pxStoreImage, (u_long128*)pxBuf );	// Start transfer to VRAM
		
			//FlushCache( WRITEBACK_DCACHE );				// Flush D cache

			for (int j = 0; j < w; ++j)
			{
				PS2_32BITPIXEL xPixel = pxBuf[j];

				// add into buffer
				float* pfInput = pfFPBuffer;
				float fB = *pfInput++;
				float fG = *pfInput++;
				float fR = *pfInput++;
				*pfFPBuffer++ = fB + float(int(xPixel.B));
				*pfFPBuffer++ = fG + float(int(xPixel.G));
				*pfFPBuffer++ = fR + float(int(xPixel.R));
			}
		}

		Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(pxBuf) );
	}
	//Asura_DeallocateAlign(pRasterBuf);
	//Asura_DeallocateAlign(pStoreImage);
	
	// Restore the DMA channel 1 settings.
	*D1_CHCR = store_D1_CHCR;
}

/// End of function 'Asura_ScreenShot_System::Platform_AddScreenToFPBuffer'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_BeginHiResRender
// Author       : Andy Tate
// Date         : 13-Jun-2005
//
// Description  : Handles the platform specific start of the high-resolution render.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_BeginHiResRender (void)
{
	// Ensure all paths are idle...

	sceGsSyncPath	(0, 0);


	// Disable the VSync...

	g_bOldVsync	= Asura_PS2_GS::GetEnableVSync ();

	Asura_PS2_GS::SetEnableVSync	(false);

	// set the keys for accept + cancel
	s_iAcceptKey = ASURA_KEYID_ENTER;
	s_aiCancelKey[0] = ASURA_KEYID_BACKSPACE;
	s_aiCancelKey[1] = ASURA_KEYID_ESCAPE;
}

/// End of function 'Asura_ScreenShot_System::Platform_BeginHiResRender'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_EndHiResRender
// Author       : Andy Tate
// Date         : 13-Jun-2005
//
// Description  : Handles the platform specific end of the high-resolution render.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_EndHiResRender (void)
{
	// Enable the VSync...

	Asura_PS2_GS::SetEnableVSync	(g_bOldVsync);
}

/// End of function 'Asura_ScreenShot_System::Platform_EndHiResRender'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_TakeScreenShot
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Handles the taking of screen-shots.
//
// Parameters   : The name of the file to be saved.
//
// Return value : The result of the screen shot function.
////////////////////////////////////////////////////////////////////////////////

int Asura_ScreenShot_System::Platform_TakeScreenShot (const char* szFilename)
{
	// What sort of screen-shot do we want?

	if (dbg_bScreenShotsAsTM2s)
	{
        const char*	szFullFilename	= Asura_PS2_OS::GetFullPath (szFilename);

		const int	iScreenWidth	= Asura_PS2_GS::GetScreenWidth ();
		const int	iScreenHeight	= Asura_PS2_GS::GetScreenHeight ();

		const u_int	uFBP			= Asura_PS2_GS::GetDispBfrFBP ();
		const u_int	tbp				= (uFBP << 5);
		const u_int	tbw				= ((Asura_PS2_GS::GetScreenWidth() + 63) >> 6);
		const u_int	psm				= Asura_PS2_GS::GetDrawBfrPSM ();


		// TIM2 - take the screen-shot...

		return	(Tim2TakeSnapshot(szFullFilename, tbp, tbw, psm, 0, 0, iScreenWidth, iScreenHeight));
	}

	else
	{
		Asura_IntVector_2	xScreenPosition;
		Asura_IntVector_2	xScreenSize;
		u_int				uScreenBitDepth;


		// Must be a BMP...

		Platform_GetScreenDimensions	(&xScreenPosition, &xScreenSize, &uScreenBitDepth);
		return							(BMPTakeScreenShot(szFilename,  xScreenPosition, xScreenSize, uScreenBitDepth));
	}
}

/// End of function 'Asura_ScreenShot_System::Platform_TakeScreenShot'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_NotTakingScreenShot
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Handles the case of being in screen-shot mode but not actually
//				  taking a screen-shot.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ScreenShot_System::Platform_NotTakingScreenShot (void)
{
#if _DEBUG
    if (dbg_VRAMShot.TakeNow)
    {
		char szFilename[256];
		const char* szFullFilename;

		static int	nIndex2 = 0;
        dbg_VRAMShot.TakeNow = false;
        sprintf( szFilename, "SShots\\VRAM%03d.TM2", nIndex2 );
        szFullFilename = Asura_PS2_OS::GetFullPath( szFilename );
		printf("Taking VRAM Shot: %s...\n", szFullFilename );
		int ok = Tim2TakeSnapshot( szFullFilename, dbg_VRAMShot.tbp, dbg_VRAMShot.tbw, dbg_VRAMShot.psm,
            dbg_VRAMShot.x, dbg_VRAMShot.y, dbg_VRAMShot.w, dbg_VRAMShot.h);
		if (ok)
        {
            printf("Done.\n");
		    nIndex2++;
        }
        else
        {
            printf("Failed.\n");
        }
    }
#endif // _DEBUG
}

/// End of function 'Asura_ScreenShot_System::Platform_NotTakingScreenShot'. ///

static float* pfDEBUG_MEMORY_112MEG = reinterpret_cast<float*>(0x07000000);// 112 MB - use last 16 MB of the 128 MB memory.

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_AllocateFloatBuffer
// Author       : Mike Healey
// Date         : 28-Jun-2005
//
// Description  : returns float buffer big enough for screenshots.
//
// Return value : ptr to float buffer.
////////////////////////////////////////////////////////////////////////////////
float* Asura_ScreenShot_System::Platform_AllocateFloatBuffer(u_int uNumFloats)
{
	// always use debug memory on a TOOL to avoid crash when new[] fails.
	if (!Asura_PS2_OS::IsTOOL())
	{
		return new float[uNumFloats];
	}
	else
	{
		return pfDEBUG_MEMORY_112MEG;
	}
}
/// End of function 'Asura_ScreenShot_System::Platform_AllocateFloatBuffer'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_FreeFloatBuffer
// Author       : Mike Healey
// Date         : 28-Jun-2005
//
// Description  : frees float buffer allocated earlier (if necessary)
////////////////////////////////////////////////////////////////////////////////
void Asura_ScreenShot_System::Platform_FreeFloatBuffer(float* pfBuffer)
{
	if (pfBuffer != pfDEBUG_MEMORY_112MEG)
	{
		delete[] pfBuffer;
	}
}
/// End of function 'Asura_ScreenShot_System::Platform_FreeFloatBuffer'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ScreenShot_System::Platform_RenderPrompt
// Author       : Mike Healey
// Date         : 28-Jun-2005
//
// Description  : render prompt for yes/no functionality. 
////////////////////////////////////////////////////////////////////////////////
void Asura_ScreenShot_System::Platform_RenderPrompt(void)
{
	Screenshots_DrawMessageBackground();
	Asura_PS2_DebugInfo::ShowMessage( "[ENTER] take this screen - [ESCAPE]/[BKSPACE] - cancel" );
}
/// End of function 'Asura_ScreenShot_System::Platform_RenderPrompt'. ///

#endif // #ifdef ASURA_USE_SCREENSHOT_SYSTEM

/// End of file. ///
