//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Mpeg.cpp
//
// Created by Morgan Parry.
//
// Mostly copyright (c) 2002 Sony Computer Entertainment Europe.
// Portions copyright (c) 2002 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Standard independent.
#include <stdio.h>
#include <string.h> // (MP - 13.09.2002) This is just for memcpy(). Don't seem to have memory.h on PS2.

// Standard PS2.
#include <eestruct.h>
#include <libmpeg.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"
#include "Core/Asura_Handle.h"
#include "COre/Asura_TextureFlags.h"
#include "Core/Asura_Timers.h"

// Engine PS2.
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_PS2_DMA.h"
#include "FMV/Asura_PS2_Mpeg.h"
#include "Rendering/Asura_Platform_MovieToTextureStream.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_TIM2.h"
#include "Sound/Asura_PS2_Multistream.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#if ASURA_PS2_USE_NEW_FMV_STUFF

// Four quadwords header, six quads per macroblock, one quadword end tag,
// rounded up to a 64 byte boundary.
#define PATH3_CHAIN_SIZE_ROWS(mbw, mbh) (((((5 + (6 * (mbw) * (mbh))) * 16) + 63) / 64) * 64)

// Four quadwords header, six quads per column, one quadword end tag, rounded
// up to a 64 byte boundary.
#define PATH3_CHAIN_SIZE_COLUMNS(mbw) (((((5 + (6 * (mbw) * (1))) * 16) + 63) / 64) * 64)

// T Y P E S /////////////////////////////////////////////////////////////////

struct ADSHeader
{
	char id[4];		// 'S''S''h''d'
	int size;		// 24
	int type;		// 0: 16bit big endian
					// 1: 16bit little endian
					// 16: SPU2-ADPCM (VAG) 
	int rate;		// sampling rate (Hz)
	int ch;			// number of channels
	int interSize;	// interleave size
	int loopStart;	// loop start block address
	int loopEnd;	// loop end block sddress
};
// End struct ADSHeader.

struct ADSBody
{
	char id[4];		// 'S''S''b''d'
	int size;		// size of audio data
};
// End struct ADSBody.

struct MpegStream
{
    bool bInUse;
	sceMpeg mp;

	u_char*pPssBuffer;
	uint32 uPssOffset;
	uint32 uPssEnd;

	int ipuNumFrames;
	int framesDecoded;
	float fFrameTime;
	float fNextFrameDueAt;

	u_int width, height;

    void (*pfnFileInitCallback)( int );
    int (*pfnFileLoadCallback)( u_char*, int, int );
    int iFileCallbackParam;
    void (*pfnIdleCallback)( void );

	u_int dbw;			// Texture width / 64		for Texture uploading and TEX0 setting
	u_int tw;			// 2^tw = texture width		for Tex0 setting
	u_int th;			// 2^th = texture height	for Tex0 setting
	sceIpuDmaEnv dmaEnv;

	sceIpuRGB32* rgb32[2];

	u_char *workArea;

    Asura_PS2_QW* apxDmaChains[2];

	// ipuBuff - data that will be sent to the IPU, either a .m2v or .ipu.

	// Ring buffer for demultiplexed m2v or ipu data (uncached access)
	//
	//  Regions: times are relative to DMA and circular wrt BuffStart, BuffSize
	//
	//  ipuBuffPresent -> ipuBuffFuture   = Area currently being processed
	//  ipuBuffFuture  -> ipuBuffPast     = Demuxed data ready for processing
	//  ipuBuffPast    -> ipuBuffPresent  = Processed data
	//
	u_char *ipuBuff;
	u_int ipuBuffSize;
	u_char *ipuBuffPast;
	u_char *ipuBuffPresent;
	u_char *ipuBuffFuture;

	// Ring buffer for audio data (audioBuff)
	//
	// Note that audioBuffFullness can wrap
	//
	//  |<----------------------- audioBuffSize ----------------------->|
	//  |<--- audioBuffPos --->|<--- audioBuffFullness --->|
	//
 	//  +----------------------+---------------------------+------------+
	//  |      old data        |         new data          |  old data  |
	//  +----------------------+---------------------------+------------+
	//
	//  ^ audioBuff

	u_char *audioBuff;
	u_int audioBuffSize;
	int audioBuffPos;
	int audioBuffFullness;

	ADSHeader sshd;
	ADSBody ssbd;

	int headerCount;

	ASURA_PS2_MPEG_MULTIPLEXTYPE mplex;
	ASURA_PS2_MPEG_VIDEOTYPE video;
	ASURA_PS2_MPEG_AUDIOTYPE audio;
	bool bDoubleBuffer;
};
// End struct MpegStream.

// C O N S T A N T S /////////////////////////////////////////////////////////

#define IPU_BUFF_SIZE   (256 * 1024)
#define ADS_HEADER_SIZE (40)
#define AUDIO_BUFF_SIZE (49152)
#define DMACHUNKSIZE    (8192)  // bytes

// Maximum number of macroblocks that fit in a single DMA transfer
#define MAX_DMA_MBCOUNT ((1024*1024 - 16)/sizeof(sceIpuRGB32))

// G L O B A L S /////////////////////////////////////////////////////////////

static MpegStream g_axStreams[ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS];

static MpegStream* g_pxCurStream = NULL;

// (MP - 17.09.2002) TODO: Could get rid of this static buffer by allocating
// a temporary one on the heap in _createCore() instead.
static u_char tempWorkArea[SCE_MPEG_BUFFER_SIZE(0, 0)] __attribute__ ((aligned(64)));

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static bool _createCore( bool bCreate,
                         MpegStream* pxStream,
                         ASURA_PS2_MPEG_MULTIPLEXTYPE eMultiplexType,
                         ASURA_PS2_MPEG_VIDEOTYPE eVideoType,
                         ASURA_PS2_MPEG_AUDIOTYPE eAudioType,
                         int audioChannel,
                         void (*pfnFileInitCallback)( int ),
                         int (*pfnFileLoadCallback)( u_char*, int, int ),
                         int iFileCallbackParam,
                         void (*pfnIdleCallback)( void ),
                         bool bDoubleBuffer,
                         int iDBW );

static void _mywaitipuidle( bool bReturnOnOutputBlocked );
static int _videoCallback( sceMpeg *mp, sceMpegCbDataStr *cbstr, void *data );
static int _audioCallback( sceMpeg *mp, sceMpegCbDataStr *cbstr, void *data );
static int _nodataCallback( sceMpeg *mp, sceMpegCbData *cbstr, void *data );
static int _backgroundCallback( sceMpeg *mp, sceMpegCbData *cbstr, void *data );
static int _errorCallback( sceMpeg *mp, sceMpegCbData *cbstr, void *data );
static int _fillBuff( bool bBlocking );

static void SetLoadImageDest( Asura_PS2_QW* pxBuffer, u_int dbp, u_int dbw );
static void SetLoadImageTagsRows( Asura_PS2_QW* pxBuffer, void* image, u_int mbw, u_int mbh );
static void SetLoadImageTagsColumns( Asura_PS2_QW* pxBuffer, void* image, u_int mbw, u_int mbh );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_PS2_MpegSystem.
//

void Asura_PS2_MpegSystem::Initialise( void )
{
    sceIpuInit();
    sceMpegInit();

    for( u_int u = 0; u < ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS; ++u )
    {
        g_axStreams[u].bInUse = false;
    }

    Asura_PS2_StreamingMpegSystem::Initialise();
}

void Asura_PS2_MpegSystem::Shutdown( void )
{
    Asura_PS2_StreamingMpegSystem::Shutdown();

    for( u_int u = 0; u < ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS; ++u )
    {
        if( g_axStreams[u].bInUse )
        {
            Destroy( u );
        }
    }
}

int Asura_PS2_MpegSystem::Create( ASURA_PS2_MPEG_MULTIPLEXTYPE eMultiplexType,
                                  ASURA_PS2_MPEG_VIDEOTYPE eVideoType,
                                  ASURA_PS2_MPEG_AUDIOTYPE eAudioType,
                                  int audioChannel,
                                  void (*pfnFileInitCallback)( int ),
                                  int (*pfnFileLoadCallback)( u_char*, int, int ),
                                  int iFileCallbackParam,
                                  void (*pfnIdleCallback)( void ),
                                  bool bDoubleBuffer )
{
    int iHandle = -1;

    // Find free slot.
    {
        for( u_int u = 0; u < ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS; ++u )
        {
            if( !g_axStreams[u].bInUse )
            {
                iHandle = u;
                break;
            }
        }

        if( iHandle < 0 )
        {
            return -1;
        }
    }

    bool bCreatedOK = _createCore( true,
                                   &g_axStreams[iHandle],
                                   eMultiplexType,
                                   eVideoType,
                                   eAudioType,
                                   audioChannel,
                                   pfnFileInitCallback,
                                   pfnFileLoadCallback,
                                   iFileCallbackParam,
                                   pfnIdleCallback,
                                   bDoubleBuffer,
                                   -1 );

    if( bCreatedOK )
    {
        g_axStreams[iHandle].bInUse = true;
        return iHandle;
    }
    else
    {
        return -1;
    }
}

void Asura_PS2_MpegSystem::Destroy( int iHandle )
{
    Asura_Assert( 0 <= iHandle && iHandle < static_cast<int>(ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS), "Invalid handle." );

    MpegStream& xStream = g_axStreams[iHandle];
    if( xStream.bInUse )
    {
        if( &xStream == g_pxCurStream )
        {
            // (MP - 17.09.2002) TODO?
        }

        xStream.bInUse = false;
    }
}

void Asura_PS2_MpegSystem::PrepareNextFrame( void )
{
    if( !g_pxCurStream )
    {
        Asura_Assert( false, "" );
        return;
    }

	// If the next frame isn't dues yet, don't do anything
	g_pxCurStream->fNextFrameDueAt -= Asura_Timers::GetGameTimeSlice();

	if ( g_pxCurStream->fNextFrameDueAt <= 0.0f )
	{	// Need another frame - set timer for next frame
		g_pxCurStream->fNextFrameDueAt += g_pxCurStream->fFrameTime;
	}
	else
	{
		// Not time for a new frame yet...
		return;
	}

/*	// This code pauses the video if the data is unlikely to be read...
	// It assumes that 2 * DMACHUNKSIZE is enough data to decode one frame of video,
	// the real amount of data required will depend on the video size and the compression rate
	// If we don't pausing the video when data is not available, instead the game will 
	// pause if we run out of data mid way through decoding the frame
	// Check if there is enough data to decode this frame
	int availableData = g_pxCurStream->ipuBuffPast - g_pxCurStream->ipuBuffFuture;   // data ready for processing
	if (availableData<0)
	{
		availableData += g_pxCurStream->ipuBuffSize; // ring buffer
	}

	if ( availableData<(DMACHUNKSIZE<<1) )
	{
		_fillBuff( false );

		availableData = g_pxCurStream->ipuBuffPast - g_pxCurStream->ipuBuffFuture;   // data ready for processing
		if (availableData<0)
		{
			availableData += g_pxCurStream->ipuBuffSize; // ring buffer
		}

		if ( availableData<(DMACHUNKSIZE<<1) )
		{
			// No enough data skip it for a frame
			printf("SKIPPING DUE TO INSUFFICENT DATA\n");
			return;
		}
	}
*/
	
	const u_int uBuffer = ( (g_pxCurStream->framesDecoded & 1) && g_pxCurStream->bDoubleBuffer );

    sceIpuRestartDMA(&g_pxCurStream->dmaEnv);

    if( g_pxCurStream->video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD || g_pxCurStream->video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_COLUMNS )
    {
        if (g_pxCurStream->framesDecoded == 0)
        {
            sceIpuReset();
            sceIpuSync(0, 0);
            
            sceIpuBCLR(0);
            sceIpuSync(0, 0);
        }

        if (g_pxCurStream->framesDecoded >= (g_pxCurStream->ipuNumFrames - 1))
        {
            *D4_QWC  = 0;
            *D4_CHCR = 0;
            return;
        }
    }
	
    if(g_pxCurStream->video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU)
    {
        if(sceMpegIsEnd(&g_pxCurStream->mp))
        {
            return;
        }
        
        if( sceMpegGetPicture(&g_pxCurStream->mp, g_pxCurStream->rgb32[uBuffer], g_pxCurStream->width * g_pxCurStream->height) < 0 )
        {
            return;
        }
    }
    else  // g_pxCurStream->video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD or ..._COLUMNS
    {        
        u_int flags;
        u_int mbcount = (g_pxCurStream->width / 16) * (g_pxCurStream->height / 16);
        
        sceIpuRGB32 *cscptr = g_pxCurStream->rgb32[uBuffer];
        
        // Get flags
        sceIpuFDEC(0);
        _mywaitipuidle( false );
        
        flags = sceIpuGetFVdecResult();
        flags = (flags >> 24) & 0xff;
        
        sceIpuFDEC(8);  // proceed 7 bit
        _mywaitipuidle( false );
        
        DPUT_IPU_CTRL((flags & ~0x00000004) << 16);
        
        sceIpuIDEC(
                   0,            // ofm: RGB32
                   0,            // dte: no dither
                   0,            // sgn: no offset
                   (flags >> 2) & 0x1,        // dtd: no dt decode
                   1,            // qsc: dummy
                   0            // fb
                   );
                           
        while(mbcount)
        {
            u_int mb = ( mbcount < MAX_DMA_MBCOUNT ? mbcount : MAX_DMA_MBCOUNT );
            
            *D3_MADR = reinterpret_cast<u_int>( cscptr );
            *D3_QWC  = mb * sizeof(sceIpuRGB32) / 16;
            *D3_CHCR = (1 << D_CHCR_STR_O) | (0 << D_CHCR_DIR_O);
            
            _mywaitipuidle( true );
            
            mbcount -= mb;
            cscptr += mb;
        }
        
        // skip start code
        sceIpuFDEC(32);
        _mywaitipuidle( false );
    }
    
    sceIpuStopDMA(&g_pxCurStream->dmaEnv);
    
    ++g_pxCurStream->framesDecoded;
}

u_int Asura_PS2_MpegSystem::GetVideo( u_int uDestBfrBase )
{
    Asura_Assert( g_pxCurStream, "" );

    if( !g_pxCurStream->framesDecoded )
    {
        Asura_Assert( false, "No frames have been decoded." );
        return 0;
    }

    // We want to return the last frame decoded.
    u_int uBuffer = ( ((g_pxCurStream->framesDecoded - 1) & 1) && g_pxCurStream->bDoubleBuffer );
        
    Asura_PS2_QW* const pxDmaChain = g_pxCurStream->apxDmaChains[uBuffer];
    if( !pxDmaChain )
    {
        return 0;
    }

    // Fix up DMA chain with GS destination buffer base address.
    SetLoadImageDest( pxDmaChain, uDestBfrBase, g_pxCurStream->dbw );

    return reinterpret_cast<u_int>(pxDmaChain);
}

void Asura_PS2_MpegSystem::GetTex0( u_long *pulTex0, u_int uTBP )
{
	*pulTex0 = SCE_GS_SET_TEX0( uTBP,					// Texture VRAM address
								g_pxCurStream->dbw,		// Texture buffer width/64
								SCE_GS_PSMCT24,			// Texture format (rgb no alpha)
								g_pxCurStream->tw,		// Texture Width (2^tw)
								g_pxCurStream->th,		// Texture Height (2^th)
								0,						// Alpha component ?  None
								0,						// Texture function = Modulate
								0,						// Clut VRAM address N/A
								0,						// Clut Format N/A
								0,						// Clut storage mode N/A
								0,						// Clut entry offset N/A
								0);						// Clut buffer load control N/A
}

bool Asura_PS2_MpegSystem::IsFinished( void )
{
    Asura_Assert( g_pxCurStream, "" );

    switch(g_pxCurStream->video)
    {
        case ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD:
        case ASURA_PS2_MPEG_VIDEOTYPE_IPU_COLUMNS:
            return (g_pxCurStream->framesDecoded >= (g_pxCurStream->ipuNumFrames - 1));
        
        case ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU:
            return (sceMpegIsEnd(&g_pxCurStream->mp));
    }

    Asura_Assert( false, "Stream has invalid video type." );
    return false;
}

void Asura_PS2_MpegSystem::Rewind( void )
{
    Asura_Assert( g_pxCurStream, "" );

	Asura_PS2_Multistream_FastLoad::Seek( (reinterpret_cast<const Asura_Platform_MovieToTextureStream*>( g_pxCurStream->iFileCallbackParam )->GetFileHandle()), 0 );

    _createCore( false,
                 g_pxCurStream,
                 g_pxCurStream->mplex,
                 g_pxCurStream->video,
                 g_pxCurStream->audio,
                 0,
                 g_pxCurStream->pfnFileInitCallback,
                 g_pxCurStream->pfnFileLoadCallback,
                 g_pxCurStream->iFileCallbackParam,
                 g_pxCurStream->pfnIdleCallback,
                 g_pxCurStream->bDoubleBuffer,
                 g_pxCurStream->dbw );
}

void Asura_PS2_MpegSystem::GetAudio( u_int (*receiver)( u_char *, u_int, u_int, u_int, int ), int param )
{
    Asura_Assert( g_pxCurStream, "" );

    u_int bytesUsed = receiver( g_pxCurStream->audioBuff,
                                g_pxCurStream->audioBuffSize,
                                g_pxCurStream->audioBuffPos,
                                g_pxCurStream->audioBuffFullness,
                                param );
    
    g_pxCurStream->audioBuffFullness -= bytesUsed;
    g_pxCurStream->audioBuffPos = (g_pxCurStream->audioBuffPos + bytesUsed) % g_pxCurStream->audioBuffSize;    
}

void Asura_PS2_MpegSystem::Recharge( void )
{
    Asura_Assert( g_pxCurStream, "" );
    _fillBuff( false );
}

bool Asura_PS2_MpegSystem::Bind( int iHandle )
{
    Asura_Assert( 0 <= iHandle && iHandle < static_cast<int>(ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS), "Invalid handle." );

    if( g_axStreams[iHandle].bInUse )
    {
        g_pxCurStream = &g_axStreams[iHandle];
        return true;
    }
    else
    {
        return false;
    }
}

//
// Asura_PS2_StreamingMpegSystem.
//

// Index within array is the handle returned by Asura_PS2_MpegSystem::Create().
int g_aiStreamingMpegTextureIDs[ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS];
u_int g_uNumStreamingMpegs;

int Asura_PS2_StreamingMpegSystem::Create( int iTextureIdx )
{
    // (MP - 19.09.2002) TODO
    return -1;
}

void Asura_PS2_StreamingMpegSystem::Destroy( int iHandle )
{
    Asura_Assert( 0 <= iHandle && iHandle < static_cast<int>(ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS), "Invalid handle." );
    Asura_Assert( ASURA_TEXTURE_NONE != g_aiStreamingMpegTextureIDs[iHandle], "Invalid handle." );
    Asura_PS2_MpegSystem::Destroy( iHandle );
    g_aiStreamingMpegTextureIDs[iHandle] = ASURA_TEXTURE_NONE;
}

int Asura_PS2_StreamingMpegSystem::GetTextureIdxForHandle( int iHandle )
{
    Asura_Assert( 0 <= iHandle && iHandle < static_cast<int>(ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS), "Invalid handle." );
    Asura_Assert( ASURA_TEXTURE_NONE != g_aiStreamingMpegTextureIDs[iHandle], "Invalid handle." );
    return g_aiStreamingMpegTextureIDs[iHandle];
}

void Asura_PS2_StreamingMpegSystem::Initialise( void )
{
    for( u_int u = 0; u < ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS; ++u )
    {
        g_aiStreamingMpegTextureIDs[u] = ASURA_TEXTURE_NONE;
    }
    g_uNumStreamingMpegs = 0;
}

void Asura_PS2_StreamingMpegSystem::Shutdown( void )
{
    for( u_int u = 0; u < ASURA_PS2_MPEG_MAX_CONCURRENT_STREAMS; ++u )
    {
        if( ASURA_TEXTURE_NONE != g_aiStreamingMpegTextureIDs[u] )
        {
            Asura_PS2_MpegSystem::Destroy( u );
            g_aiStreamingMpegTextureIDs[u] = ASURA_TEXTURE_NONE;
        }
    }
    g_uNumStreamingMpegs = 0;
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static bool _createCore( bool bCreate,
                         MpegStream* str,
                         ASURA_PS2_MPEG_MULTIPLEXTYPE mplex,
                         ASURA_PS2_MPEG_VIDEOTYPE video,
                         ASURA_PS2_MPEG_AUDIOTYPE audio,
                         int audioChannel,
                         void (*pfnFileInitCallback)( int ),
                         int (*pfnFileLoadCallback)( u_char*, int, int ),
                         int iFileCallbackParam,
                         void (*pfnIdleCallback)( void ),
                         bool bDoubleBuffer,
                         int iDBW )
{
	return false;
	uint32 uMemoryAllcated = 0;

    int n;

    g_pxCurStream = str;

    *D4_QWC  = 0;
    *D4_CHCR = 0;
    
    if( bCreate )
    {
        // Initialise buffer for unmultiplexed video data (.m2v or .ipu)        
        str->ipuBuff = reinterpret_cast<u_char*>( Asura_AllocateAlign(IPU_BUFF_SIZE, ASURA_ALIGN_64_BYTE) );
		if ( !str->ipuBuff )
		{
			return false;
		}

		uMemoryAllcated += IPU_BUFF_SIZE;
    }

    str->pPssBuffer = NULL;
	str->uPssEnd = 0;
	str->uPssOffset = 0;

    str->ipuBuffSize	= IPU_BUFF_SIZE;    
    str->ipuBuffPast    = str->ipuBuff + 64;
    str->ipuBuffPresent = str->ipuBuff;
    str->ipuBuffFuture  = str->ipuBuff + 64;
    
    // Initialise buffer for audio data
    str->audio = audio;
    str->headerCount = 0;

    if (audio != ASURA_PS2_MPEG_AUDIOTYPE_NONE)
    {
        if( bCreate )
        {
            str->audioBuffSize = AUDIO_BUFF_SIZE;
    
            str->audioBuff = reinterpret_cast<u_char*>( Asura_AllocateAlign(str->audioBuffSize, ASURA_ALIGN_64_BYTE) );

            if( !str->audioBuff )
            {
				// Allcoation Failed - free up everything we allocated up to this point
				Asura_Assert( 0, "Failed to allocate memory for mpeg decoder" );
				if (str->ipuBuff) { Asura_DeallocateAlign( str->ipuBuff );	str->ipuBuff = NULL; }
                return false;
            }

			uMemoryAllcated += str->audioBuffSize;
        }

        str->audioBuffPos = 0;
        str->audioBuffFullness = 0;
    }

    if( bCreate )
    {
        str->pfnFileInitCallback = pfnFileInitCallback;
        str->pfnFileLoadCallback = pfnFileLoadCallback;
        str->iFileCallbackParam  = iFileCallbackParam;    
        str->pfnIdleCallback     = pfnIdleCallback;
        str->video               = video;
        str->mplex               = mplex;
        str->bDoubleBuffer       = bDoubleBuffer;
    }
    
    if( pfnFileInitCallback )
    {
        pfnFileInitCallback( iFileCallbackParam );
    }
    
    str->framesDecoded = 0;
	str->fNextFrameDueAt = 0.0f;
    
    // Get video dimensions
    if (mplex == ASURA_PS2_MPEG_MULTIPLEXTYPE_PSS && bCreate)
    {
        // Just for the initial demux. We'll create a real one later.
        sceMpegCreate(&str->mp, tempWorkArea, sizeof(tempWorkArea));
        
        // Set the stream callbacks just so we can get the start of the video
        if (video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU)
        {
            sceMpegAddStrCallback(&str->mp, sceMpegStrM2V, 0, (sceMpegCallback) _videoCallback, 0);
        }
        else // video == SCEE_MPEG_VIDEO_IPU_STANDARD or ..._IPU_COLUMNS
        {
            sceMpegAddStrCallback(&str->mp, sceMpegStrIPU, 0, (sceMpegCallback) _videoCallback, 0);
        }
    }
    
    if( !_fillBuff(true) )
    {
        Asura_Assert( false, "Couldn't read stream header." );
    }
    
    if(video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD || video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_COLUMNS)
    {
        if( bCreate )
        {
			// Get movie specs
            str->width  = (u_int) *(u_short *) (str->ipuBuffFuture + 8);
            str->height = (u_int) *(u_short *) (str->ipuBuffFuture + 10);
            str->ipuNumFrames = (int) *(u_int *) (str->ipuBuffFuture + 12);
			// Find the framerate - Not sure if this is correct for IPU streams...
			uint32 uFrameRateCode = (str->ipuBuffFuture[7] & 0x0f);

			switch (uFrameRateCode)
			{
				case 1:
					str->fFrameTime = 1001.0f / 24000.0f;
					break;
				case 2:
					str->fFrameTime = 1.0f / 24.0f;
					break;
				case 3:
					str->fFrameTime = 1.0f / 25.0f;
					break;
				case 4:
					str->fFrameTime = 1001.0f / 30000.0f;
					break;
				case 5:
					str->fFrameTime = 1.0f / 30.0f;
					break;
				case 6:
					str->fFrameTime = 1.0f / 50.0f;
					break;
				case 7:
					str->fFrameTime = 1001.0f / 60000.0f;
					break;
				case 8:
					str->fFrameTime = 1.0f / 60.0f;
					break;
				default:
					Asura_Assert( 0, "Movie has Illegal frame rate code" );
					str->fFrameTime = 0.0f;
					break;
			}
        }
       
        // Advance the demuxed stream past the header, in a roundabout way        
        str->ipuBuffFuture += 16; // header is 16 bytes
    }
    // Or get m2v dimensions...
    else  // video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU 
    {
        if( bCreate )
        {
            str->width  = (u_int) (str->ipuBuffFuture[4] << 4) | (str->ipuBuffFuture[5] >> 4);
            str->height = (u_int) ((str->ipuBuffFuture[5] & 0xf) << 8) | str->ipuBuffFuture[6];
			// Find the framerate
			uint32 uFrameRateCode = (str->ipuBuffFuture[7] & 0x0f);

			switch (uFrameRateCode)
			{
				case 1:
					str->fFrameTime = 1001.0f / 24000.0f;
					break;
				case 2:
					str->fFrameTime = 1.0f / 24.0f;
					break;
				case 3:
					str->fFrameTime = 1.0f / 25.0f;
					break;
				case 4:
					str->fFrameTime = 1001.0f / 30000.0f;
					break;
				case 5:
					str->fFrameTime = 1.0f / 30.0f;
					break;
				case 6:
					str->fFrameTime = 1.0f / 50.0f;
					break;
				case 7:
					str->fFrameTime = 1001.0f / 60000.0f;
					break;
				case 8:
					str->fFrameTime = 1.0f / 60.0f;
					break;
				default:
					Asura_Assert( 0, "Movie has Illegal frame rate code" );
					str->fFrameTime = 0.0f;
					break;
			}
        }
    }

    // (MP - 20.09.2002) Calculate DBW now, unless it was specified.
    if( iDBW < 0 )
    {
        str->dbw = Tim2CalcBufWidth( SCE_GS_PSMCT32, str->width );
    }
    else
    {
        Asura_Assert( 1 <= iDBW && iDBW <= 32, "Invalid DBW." );
        str->dbw = iDBW;
    }

	// calculate tw and th
	str->tw = 0;
	while ( ((u_int)(1<<str->tw)) < str->width )	str->tw++;
	str->th = 0;
	while ( ((u_int)(1<<str->th)) < str->height )	str->th++;

    
    // Set up the sceMpeg structure; only not needed if we're playing a raw
    // ipu stream, in which case it won't actually harm us.
    
    // Recreate the sceMpeg structure to replace the temporary one we created
    // just for demuxing which had a zero-sized work area attached.
    if( video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU || mplex == ASURA_PS2_MPEG_MULTIPLEXTYPE_PSS )
    {
        unsigned int workAreaSize;
        
        // Set up work area
        if( video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD || video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_COLUMNS )
        {
            workAreaSize = SCE_MPEG_BUFFER_SIZE(0,0);
        }
        else  // video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU
        {
            workAreaSize = SCE_MPEG_BUFFER_SIZE(str->width, str->height);
        }
        
        if( bCreate )
        {
            str->workArea = reinterpret_cast<u_char*>( Asura_AllocateAlign(workAreaSize, ASURA_ALIGN_64_BYTE) );
            if( !str->workArea )
			{
				// Allcoation Failed - free up everything we allocated up to this point
				Asura_Assert( 0, "Failed to allocate memory for mpeg decoder" );
	            if (str->audioBuff) { Asura_DeallocateAlign( str->audioBuff );	str->audioBuff = NULL; }
				if (str->ipuBuff)	{ Asura_DeallocateAlign( str->ipuBuff );	str->ipuBuff = NULL; }
                return false;
            }
			uMemoryAllcated += workAreaSize;
        }
        
        sceMpegCreate(&str->mp, str->workArea, (int) workAreaSize);
        
        // Set up callbacks (all of them, for real this time)
        if(video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU)
        {
            sceMpegAddCallback(&str->mp, sceMpegCbNodata, _nodataCallback, (void *) str);
            
            if (str->pfnIdleCallback)
            {
                // Should reintroduce this at some point
                sceMpegAddCallback( &str->mp, sceMpegCbBackground, _backgroundCallback, str );
            }
        }
        
        sceMpegAddCallback(&str->mp, sceMpegCbError, _errorCallback, NULL);
        
        
        if (mplex == ASURA_PS2_MPEG_MULTIPLEXTYPE_PSS)
        {
            // Video callbacks            
            if(video == ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU)
            {
                sceMpegAddStrCallback(&str->mp, sceMpegStrM2V, 0, (sceMpegCallback) _videoCallback, 0);
            }
            
            else // video == SCEE_MPEG_VIDEO_IPU_STANDARD or .._IPU_COLUMNS
            {
                sceMpegAddStrCallback(&str->mp, sceMpegStrIPU, 0, (sceMpegCallback) _videoCallback, 0);
            }
            
            // Audio callbacks
            if(audio == ASURA_PS2_MPEG_AUDIOTYPE_PCM)
            {
                sceMpegAddStrCallback(&str->mp, sceMpegStrPCM, audioChannel, (sceMpegCallback) _audioCallback, 0);
            }

            else if(audio == ASURA_PS2_MPEG_AUDIOTYPE_ADPCM)
            {
                sceMpegAddStrCallback(&str->mp, sceMpegStrADPCM, audioChannel, (sceMpegCallback) _audioCallback, 0);
            }

            else if(audio == ASURA_PS2_MPEG_AUDIOTYPE_SPDIF)
            {
                sceMpegAddStrCallback(&str->mp, sceMpegStrDATA, audioChannel, (sceMpegCallback) _audioCallback, 0);
            }
        }
    }
    
    for (n = 0; bCreate && n < (bDoubleBuffer ? 2 : 1); n++)
    {
        {
            const u_int uSize = (str->width / 16) * (str->height / 16) * sizeof(sceIpuRGB32);
            str->rgb32[n] = reinterpret_cast<sceIpuRGB32*>( Asura_AllocateAlign(uSize, ASURA_ALIGN_64_BYTE) );
            if( !str->rgb32[n] )
            {
				// Allcoation Failed - free up everything we allocated up to this point
				Asura_Assert( 0, "Failed to allocate memory for mpeg decoder" );

				for (int nd = 0; nd<n; nd++)
				{
					if (str->rgb32[nd]		 )	{ Asura_DeallocateAlign(str->rgb32[nd]       ); str->rgb32[nd]		  = NULL; }
					if (str->apxDmaChains[nd])	{ Asura_DeallocateAlign(str->apxDmaChains[nd]); str->apxDmaChains[nd] = NULL; }
				}

				sceMpegDelete( &str->mp );
				if (str->workArea)  { Asura_DeallocateAlign(str->workArea ); str->workArea  = NULL; }
	            if (str->audioBuff) { Asura_DeallocateAlign(str->audioBuff); str->audioBuff = NULL; }
				if (str->ipuBuff)	{ Asura_DeallocateAlign(str->ipuBuff  ); str->ipuBuff   = NULL; }
                return false;
            }
			uMemoryAllcated += uSize;
        }
        
        // Set up DMA transfer tags
        if(str->video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD)
        {
			const u_int uSize = PATH3_CHAIN_SIZE_ROWS( str->width / 16, str->height / 16 );
			str->apxDmaChains[n] = reinterpret_cast<Asura_PS2_QW*>( Asura_AllocateAlign(uSize, ASURA_ALIGN_64_BYTE) );
			if (!str->apxDmaChains[n])
			{
				// Allcoation Failed - free up everything we allocated up to this point
				Asura_Assert( 0, "Failed to allocate memory for mpeg decoder" );

				for (int nd = 0; nd<n; nd++)
				{
					if (str->rgb32[nd]		 )	{ Asura_DeallocateAlign(str->rgb32[nd]       ); str->rgb32[nd]		  = NULL; }
					if (str->apxDmaChains[nd])	{ Asura_DeallocateAlign(str->apxDmaChains[nd]); str->apxDmaChains[nd] = NULL; }
				}

				if (str->rgb32[n] )	{ Asura_DeallocateAlign(str->rgb32[n] ); str->rgb32[n]	= NULL; }
				sceMpegDelete( &str->mp );
				if (str->workArea ) { Asura_DeallocateAlign(str->workArea ); str->workArea  = NULL; }
	            if (str->audioBuff) { Asura_DeallocateAlign(str->audioBuff); str->audioBuff = NULL; }
				if (str->ipuBuff  )	{ Asura_DeallocateAlign(str->ipuBuff  ); str->ipuBuff   = NULL; }
                return false;
            }

			uMemoryAllcated += uSize;

            SetLoadImageTagsRows( str->apxDmaChains[n], str->rgb32[n], str->width / 16, str->height / 16 );
        }
        else
        {
			const u_int uSize = PATH3_CHAIN_SIZE_COLUMNS( str->width / 16 );
			str->apxDmaChains[n] = reinterpret_cast<Asura_PS2_QW*>( Asura_AllocateAlign(uSize, ASURA_ALIGN_64_BYTE) );

			if (!str->apxDmaChains[n])
			{
				// Allcoation Failed - free up everything we allocated up to this point
				Asura_Assert( 0, "Failed to allocate memory for mpeg decoder" );

				for (int nd = 0; nd<n; nd++)
				{
					if (str->rgb32[nd]		 )	{ Asura_DeallocateAlign(str->rgb32[nd]       ); str->rgb32[nd]		  = NULL; }
					if (str->apxDmaChains[nd])	{ Asura_DeallocateAlign(str->apxDmaChains[nd]); str->apxDmaChains[nd] = NULL; }
				}

				if (str->rgb32[n] )	{ Asura_DeallocateAlign(str->rgb32[n] ); str->rgb32[n]	= NULL; }
				sceMpegDelete( &str->mp );
				if (str->workArea ) { Asura_DeallocateAlign(str->workArea ); str->workArea  = NULL; }
	            if (str->audioBuff) { Asura_DeallocateAlign(str->audioBuff); str->audioBuff = NULL; }
				if (str->ipuBuff  )	{ Asura_DeallocateAlign(str->ipuBuff  ); str->ipuBuff   = NULL; }
                return false;
            }

			uMemoryAllcated += uSize;
           
            SetLoadImageTagsColumns( str->apxDmaChains[n], str->rgb32[n], str->width / 16, str->height / 16 );
        }
    }
        
    sceIpuStopDMA(&str->dmaEnv);

	printf( "Allocated %d bytes\n", uMemoryAllcated );
    return true;
}

static void _mywaitipuidle( bool bReturnOnOutputBlocked )
{
    while( true )
    {
        if( *IPU_CTRL & IPU_CTRL_ECD_M )
        {
            printf( "IPU error code detected.\n" );
            return;
        }
        
        if( !sceIpuIsBusy() )
        {
            return;
        }
                
        if( !(*D4_CHCR & D_CHCR_STR_M) )
        {
            // DMA stopped
            
            if(sceIpuIsBusy())
            {
                _nodataCallback( NULL, NULL, g_pxCurStream );
            }
        }

        if( bReturnOnOutputBlocked && !(*D3_CHCR & D_CHCR_STR_M) )
        {
            return;
        }
    }
}

static int _videoCallback( sceMpeg*, sceMpegCbDataStr* cbstr, void* )
{
    int availSpace;
    int spill;

    // We can arrive at this function either from libmpeg's demuxer,
    // or our own passing raw data straight to it...
    if( (g_pxCurStream->video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_STANDARD || g_pxCurStream->video == ASURA_PS2_MPEG_VIDEOTYPE_IPU_COLUMNS)
		&& g_pxCurStream->mplex == ASURA_PS2_MPEG_MULTIPLEXTYPE_PSS )
    {
        // Ignore the sub-stream ID
        cbstr->data += 4;
        cbstr->len -= 4;
    }

    availSpace = g_pxCurStream->ipuBuffPresent - g_pxCurStream->ipuBuffPast;    // processed data
    if (availSpace < 0) 
	{
		availSpace += (int) g_pxCurStream->ipuBuffSize;  // ring buffer
	}
    
    if ((availSpace == 0) && (g_pxCurStream->ipuBuffPresent == g_pxCurStream->ipuBuffFuture))
	{
        availSpace = (int) g_pxCurStream->ipuBuffSize;    // special case - buffer is empty, not full
	}
    
    if( availSpace < static_cast<int>(cbstr->len) )
    {
        return 0;
    }
    
    spill = ((int) g_pxCurStream->ipuBuffPast - (int) g_pxCurStream->ipuBuff) + (int) cbstr->len - (int) g_pxCurStream->ipuBuffSize;
    
    
    if (spill > 0)
    {
        memcpy((void *) (((int) g_pxCurStream->ipuBuffPast & 0x0fffffff) | 0x20000000), cbstr->data, cbstr->len - spill);
        memcpy((void *) (((int) g_pxCurStream->ipuBuff & 0x0fffffff) | 0x20000000), cbstr->data + (cbstr->len - spill), (u_int) spill);
        g_pxCurStream->ipuBuffPast = g_pxCurStream->ipuBuff + spill;
    }
    
    else
    {
        memcpy((void *) (((int) g_pxCurStream->ipuBuffPast & 0x0fffffff) | 0x20000000), cbstr->data, cbstr->len);
        g_pxCurStream->ipuBuffPast += cbstr->len;
    }
    
    return 1;
}


static int _audioCallback( sceMpeg*, sceMpegCbDataStr* cbstr, void* )
{
    int xferStart;
    int spill;

    cbstr->len  -= 4;  // first four bytes are sub-stream ID
    cbstr->data += 4;

    if( (g_pxCurStream->audio != ASURA_PS2_MPEG_AUDIOTYPE_SPDIF) && (g_pxCurStream->headerCount < ADS_HEADER_SIZE) )
    {
        // Note, for these copies the sshd will spill over to the ssbd.
    
        if( static_cast<int>(cbstr->len) > ADS_HEADER_SIZE - g_pxCurStream->headerCount )
        {
            memcpy( (char *)&g_pxCurStream->sshd + g_pxCurStream->headerCount, cbstr->data, ADS_HEADER_SIZE - (u_int)g_pxCurStream->headerCount );

            cbstr->data += ADS_HEADER_SIZE - g_pxCurStream->headerCount;
            cbstr->len -= ADS_HEADER_SIZE - (u_int) g_pxCurStream->headerCount;
            g_pxCurStream->headerCount = ADS_HEADER_SIZE;
        }

        else
        {
            memcpy( (char *)&g_pxCurStream->sshd + g_pxCurStream->headerCount, cbstr->data, cbstr->len );
        
            g_pxCurStream->headerCount += (int) cbstr->len;
            cbstr->len = 0;
        }
    }
    
    if( g_pxCurStream->audioBuffFullness + cbstr->len > g_pxCurStream->audioBuffSize )
    {
        return 0;
    }

    xferStart = (g_pxCurStream->audioBuffPos + g_pxCurStream->audioBuffFullness) % (int)g_pxCurStream->audioBuffSize;

    spill = xferStart + (int)cbstr->len - (int)g_pxCurStream->audioBuffSize;

    if(spill > 0)
    {
        memcpy( g_pxCurStream->audioBuff + xferStart, cbstr->data, cbstr->len - spill );
        memcpy( g_pxCurStream->audioBuff, cbstr->data + (cbstr->len - spill), (u_int)spill );
    }
    else
    {
        memcpy( g_pxCurStream->audioBuff + xferStart, cbstr->data, cbstr->len );
    }

    g_pxCurStream->audioBuffFullness += (int)cbstr->len;
    return 1;
}

static int _nodataCallback( sceMpeg*, sceMpegCbData*, void* pvData )
{
    MpegStream* pxStream = reinterpret_cast<MpegStream*>( pvData );
    if( !pxStream )
    {
        // (MP - 17.09.2002) Return value is irrelevant apparently.
        return 1;
    }

    u_int dmaSize;
    int availData;
    
    pxStream->ipuBuffPresent = pxStream->ipuBuffFuture;
    
    availData = pxStream->ipuBuffPast - pxStream->ipuBuffFuture;    // data ready for processing
    if (availData < 0)
    {
        availData += (int) pxStream->ipuBuffSize;  // ring buffer
    }
    
    while(availData < DMACHUNKSIZE)
    {
		int bytesUsed = _fillBuff(true);
		if( !bytesUsed )
        {            
            // More data cannot be obtained from the demultiplexer so this
            // must be the end of the stream. Round up the data to ensure
            // there's no remainder from the quadword-multiple transfer,
            // otherwise the decoder won't receive the MPEG end sequence code.
            printf("End of stream presumed\n");
            availData += 15;
			break;
        }

       availData = pxStream->ipuBuffPast - pxStream->ipuBuffFuture;   // data ready for processing
        if (availData < 0)
        {
            availData += pxStream->ipuBuffSize; // ring buffer
        }
    }

    dmaSize = (u_int) availData & ~0xf;
    if (dmaSize > DMACHUNKSIZE) dmaSize = DMACHUNKSIZE;
    
    if ((pxStream->ipuBuffPresent + dmaSize) > (pxStream->ipuBuff + pxStream->ipuBuffSize))
    {
        dmaSize = (u_int) (pxStream->ipuBuff + pxStream->ipuBuffSize - pxStream->ipuBuffPresent);
    }

    *D4_QWC  = dmaSize / 16;
    *D4_MADR = (u_int) pxStream->ipuBuffPresent;
    *D4_CHCR = 1 | 1 << 8;  // start, normal, from memory
    
    pxStream->ipuBuffFuture = pxStream->ipuBuffPresent + dmaSize;
    
    if (pxStream->ipuBuffFuture >= (pxStream->ipuBuff + pxStream->ipuBuffSize))
    {
        // ring buffer
        pxStream->ipuBuffFuture -= pxStream->ipuBuffSize;  
    }
    
    return 1;
}

static int _backgroundCallback( sceMpeg*, sceMpegCbData*, void* pvData )
{
    const MpegStream* pxStream = reinterpret_cast<const MpegStream*>( pvData );
    if( !pxStream || !pxStream->pfnIdleCallback )
    {
		// Load a bit more data if we aren't doing anything
		_fillBuff(false);
		// No stream ?  Or no idle call back for stream..
        // (MP - 17.09.2002) Return value is irrelevant apparently.
        return 1;
    }

    u_int startTime, elapsedTime, availTime;
    
    asm __volatile__ ("mfc0 %0, $9" : "=r" (startTime));
    
    // IPU takes 768 bus cycles @ 300MHz to perform CSC on one macroblock
    availTime = (pxStream->width/16) * (pxStream->height/16) * 768;
    do
    {
		pxStream->pfnIdleCallback();
     
        asm __volatile__ ("mfc0 %0, $9" : "=r" (elapsedTime));
        elapsedTime -= startTime;
    }
    while( elapsedTime < availTime );
    
    return 1;
}

static int _errorCallback( sceMpeg*, sceMpegCbData* cbstr, void* )
{
    printf( "ERROR: Problem while decoding MPEG: %s", (reinterpret_cast<const sceMpegCbDataError*>(cbstr))->errMessage );
    return 1;
}

// Fill the demultiplex buffers from at most ASURA_PS2_MPEG_READCHUNKSIZE
// worth of multiplexed data. Returns number of bytes put into buffers.
static int _fillBuff( bool bBlocking )
{
    u_char* startPos;
    int procBytes;
    
    if( g_pxCurStream->mplex == ASURA_PS2_MPEG_MULTIPLEXTYPE_NONE)
    {
        // Streaming from an .m2v or .ipu

        sceMpegCbDataStr cbstr;

        if (g_pxCurStream->uPssOffset<g_pxCurStream->uPssEnd)
        {
            startPos = g_pxCurStream->pPssBuffer + g_pxCurStream->uPssOffset;
        }
        else
        {
            do
            {
			    // Get stream Handle
				const Asura_Platform_MovieToTextureStream* pxStream = reinterpret_cast<const Asura_Platform_MovieToTextureStream*>( g_pxCurStream->iFileCallbackParam );
				Asura_Handle xFileHandle = pxStream->GetFileHandle();

				void *pBuffer;
				int iLoadSize = Asura_PS2_Multistream_FastLoad::GetData( xFileHandle, (const void **)&pBuffer );;
	
				// Check for end of file...
				/*if ( iLoadSize == Asura_PS2_Multistream_FastLoad::REACHED_EOF )
				{
					// We've reach the end of the movie...
					u_int *puDest = (u_int*)pucDest;
					puDest[ 0] = 0x000001B7;	// Movie end code
					puDest[ 1] = 0;			// Padding for IPU FIFO
					puDest[ 2] = 0;			// Padding for IPU FIFO
					puDest[ 3] = 0;			// Padding for IPU FIFO
					puDest[ 4] = 0;			// Padding for IPU FIFO
					puDest[ 5] = 0;			// Padding for IPU FIFO
					puDest[ 6] = 0;			// Padding for IPU FIFO
					puDest[ 7] = 0;			// Padding for IPU FIFO
					puDest[ 8] = 0;			// Padding for IPU FIFO
					puDest[ 9] = 0;			// Padding for IPU FIFO
					puDest[10] = 0;			// Padding for IPU FIFO
					puDest[11] = 0;			// Padding for IPU FIFO
					puDest[12] = 0;			// Padding for IPU FIFO
					puDest[13] = 0;			// Padding for IPU FIFO
					puDest[14] = 0;			// Padding for IPU FIFO
					puDest[15] = 0;			// Padding for IPU FIFO
					return (64);
				}*/

				if ( iLoadSize )
				{
					g_pxCurStream->pPssBuffer = (u_char*)pBuffer;
					g_pxCurStream->uPssOffset = 0;
					g_pxCurStream->uPssEnd = iLoadSize;
				}
                
                if( (g_pxCurStream->uPssOffset>=g_pxCurStream->uPssEnd) && !bBlocking )
                {
                    return 0;
                }    
            }
            while( g_pxCurStream->uPssOffset>=g_pxCurStream->uPssEnd );
            
            startPos = g_pxCurStream->pPssBuffer + g_pxCurStream->uPssOffset;
        }
        
        cbstr.len = g_pxCurStream->uPssEnd - g_pxCurStream->uPssOffset;
        cbstr.data = startPos;

        procBytes = ( _videoCallback(&g_pxCurStream->mp, &cbstr, 0) ? (int)cbstr.len : 0 );
        
        g_pxCurStream->uPssOffset += procBytes;
        
        return procBytes;
    }
    else  // Stream from a multiplex
    {
        if (g_pxCurStream->uPssOffset<g_pxCurStream->uPssEnd)
        {
            startPos = g_pxCurStream->pPssBuffer + g_pxCurStream->uPssOffset;
        }
        else
        {
            do
            {
			    // Get stream Handle
				const Asura_Platform_MovieToTextureStream* pxStream = reinterpret_cast<const Asura_Platform_MovieToTextureStream*>( g_pxCurStream->iFileCallbackParam );
				Asura_Handle xFileHandle = pxStream->GetFileHandle();

				void *pBuffer;
				int iLoadSize = Asura_PS2_Multistream_FastLoad::GetData( xFileHandle, (const void **)&pBuffer );
	
				// Check for end of file...
				if ( iLoadSize == Asura_PS2_Multistream_FastLoad::REACHED_EOF )
				{
					asm("break");	// Does this actually happen ?
					// We've reach the end of the movie...
					pBuffer = (void*)g_pxCurStream->pPssBuffer;

					if ( pBuffer )
					{
						u_int *puDest = (u_int*)g_pxCurStream->pPssBuffer;

						int iRemoveThisCrap = ASURA_PS2_MPEG_READCHUNKSIZE; 
						while (iRemoveThisCrap > 0)
						{
							puDest[ 0] = 0x000001B7;	// Movie end code
							puDest[ 1] = 0;			// Padding for IPU FIFO
							puDest[ 2] = 0;			// Padding for IPU FIFO
							puDest[ 3] = 0;			// Padding for IPU FIFO
							puDest[ 4] = 0;			// Padding for IPU FIFO
							puDest[ 5] = 0;			// Padding for IPU FIFO
							puDest[ 6] = 0;			// Padding for IPU FIFO
							puDest[ 7] = 0;			// Padding for IPU FIFO
							puDest[ 8] = 0;			// Padding for IPU FIFO
							puDest[ 9] = 0;			// Padding for IPU FIFO
							puDest[10] = 0;			// Padding for IPU FIFO
							puDest[11] = 0;			// Padding for IPU FIFO
							puDest[12] = 0;			// Padding for IPU FIFO
							puDest[13] = 0;			// Padding for IPU FIFO
							puDest[14] = 0;			// Padding for IPU FIFO
							puDest[15] = 0;			// Padding for IPU FIFO
							puDest += 16;
							iRemoveThisCrap -= 64;
						}
					}
					else
					{
						Asura_PS2_Multistream_FastLoad::Seek( xFileHandle, 0 );
					}
				}

				if ( iLoadSize )
				{
					g_pxCurStream->pPssBuffer = (u_char*)pBuffer;
					g_pxCurStream->uPssOffset = 0;
					g_pxCurStream->uPssEnd = iLoadSize;
				}
                
                if( (g_pxCurStream->uPssOffset >= g_pxCurStream->uPssEnd) && !bBlocking )
                {
                    return 0;
                }    
            }
            while( g_pxCurStream->uPssOffset >= g_pxCurStream->uPssEnd );
            
            startPos = g_pxCurStream->pPssBuffer + g_pxCurStream->uPssOffset;
        }
        
		uint32 dataAvailable = g_pxCurStream->uPssEnd - g_pxCurStream->uPssOffset;
		if ( dataAvailable > ASURA_PS2_MPEG_READCHUNKSIZE )
		{
			dataAvailable = ASURA_PS2_MPEG_READCHUNKSIZE;
		}

        procBytes = sceMpegDemuxPss(&g_pxCurStream->mp, (u_char *) startPos, dataAvailable);
        g_pxCurStream->uPssOffset += procBytes;

        return procBytes;
    }
}

// Path 3 tags for decoded video transfer:
//
// The tags created here for uploading a decoded image are intended to be
// 'call'ed from a channel 2 DMA chain.
//
// Videos of type SCEEMPEG_VIDEO_IPU_STANDARD are transferred one macroblock
// at a time, in row order. All other video types can have their data sent in
// units of one-macroblock-wide columns.
//
//  ------ (Header) -----------------------------------------------
//  t[0]   DMA cnt
//  t[1]     GIF tag A&D
//  t[2]       BITBLTBUF
//  t[3]       TRXREG
//  ------ (Transfer repeat unit) ----------------------------------
//  t[4]   DMA cnt
//  t[5]     GIF tag A&D
//  t[6]       TRXPOS
//  t[7]       TRXDIF
//  t[8]     GIF tag image, loop = 1 macroblock or 1 column
//  t[9]   DMA ref, qwc = 1 macroblock or 1 column
//  ------ (above repeated for each macroblock, or for each strip of image) 
//  ------ (End) --------------------------------------------------
//  ....   DMA ret


static void SetLoadImageDest( Asura_PS2_QW* pxBuffer, u_int dbp, u_int dbw )
{
    // (MP - 17.09.2002) NB: Assuming buffer is not in cache. This should be
    // the case because it's a multiple of 64 bytes in size, allocated on a
    // 64 byte boundary and we only access it here.
	pxBuffer = Asura_PS2_UncachedMem( pxBuffer );
	pxBuffer[2].u64s[0] = SCE_GS_SET_BITBLTBUF( 0, 0, 0, dbp, dbw, 0 );
	pxBuffer[2].u64s[1] = SCE_GS_BITBLTBUF;
}

static void SetLoadImageTagsRows( Asura_PS2_QW* pxBuffer, void* pvImageData, u_int mbw, u_int mbh )
{
    // (MP - 17.09.2002) NB: Assuming buffer is not in cache. This should be
    // the case because it's a multiple of 64 bytes in size, allocated on a
    // 64 byte boundary and we only access it here.
	pxBuffer = Asura_PS2_UncachedMem( pxBuffer );

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 3,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_CNT,
                                                            false,
                                                            0,
                                                            false,
                                                            0,
                                                            0 );

    (reinterpret_cast<Asura_PS2_GifTag*>(pxBuffer++))->Set( 2,
                                                            false,
                                                            false,
                                                            0,
                                                            SCE_GIF_PACKED,
                                                            1,
                                                            SCE_GIF_PACKED_AD );

	// This QW will contain a SCE_GS_BITBLTBUF; that's written prior to use
    // by a call to SetLoadImageDest().
#if _DEBUG
    pxBuffer->u128 = ASURA_PS2_INIT_ULONG128( 0xcdcdcdcdcdcdcdcdcdcdcdcdcdcdcdcdLL );
#endif
    ++pxBuffer;

	pxBuffer->u64s[0] = SCE_GS_SET_TRXREG( 16, 16 );
	pxBuffer->u64s[1] = SCE_GS_TRXREG;
    ++pxBuffer;

	for( u_int my = 0; my < mbh; ++my )
	{
		for( u_int mx = 0; mx < mbw; ++mx )
		{
			bool bEOP = ( mx == mbw - 1 && my == mbh - 1 );
		
            (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 4,
                                                                    Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                    Asura_PS2_DmaTag::ID_CNT,
                                                                    false,
                                                                    0,
                                                                    false,
                                                                    0,
                                                                    0 );

            (reinterpret_cast<Asura_PS2_GifTag*>(pxBuffer++))->Set( 2,
                                                                    false,
                                                                    false,
                                                                    0,
                                                                    SCE_GIF_PACKED,
                                                                    1,
                                                                    SCE_GIF_PACKED_AD );
            
			pxBuffer->u64s[0] = SCE_GS_SET_TRXPOS( 0, 0, mx*16, my*16, 0 );
			pxBuffer->u64s[1] = SCE_GS_TRXPOS;
            ++pxBuffer;

            pxBuffer->u64s[0] = SCE_GS_SET_TRXDIR( 0 );
			pxBuffer->u64s[1] = SCE_GS_TRXDIR;
            ++pxBuffer;

            (reinterpret_cast<Asura_PS2_GifTag*>(pxBuffer++))->Set( 64,
                                                                    bEOP,
                                                                    false,
                                                                    0,
                                                                    SCE_GIF_IMAGE,
                                                                    0 );          

            (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 64,
                                                                    Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                    Asura_PS2_DmaTag::ID_REF,
                                                                    false,
                                                                    reinterpret_cast<u_int>(pvImageData),
                                                                    false,
                                                                    0,
                                                                    0 );

			pvImageData = reinterpret_cast<u_long*>(pvImageData) + ((16 * 16 * 4) / 8);
		}
	}

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 0,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_RET,
                                                            false,
                                                            0,
                                                            false,
                                                            0,
                                                            0 );
}

static void SetLoadImageTagsColumns( Asura_PS2_QW* pxBuffer, void* pvImageData, u_int mbw, u_int mbh )
{
	u_int mx;

    // (MP - 17.09.2002) NB: Assuming buffer is not in cache. This should be
    // the case because it's a multiple of 64 bytes in size, allocated on a
    // 64 byte boundary and we only access it here.
	pxBuffer = Asura_PS2_UncachedMem( pxBuffer );

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 3,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_CNT,
                                                            false,
                                                            0,
                                                            false,
                                                            0,
                                                            0 );

    (reinterpret_cast<Asura_PS2_GifTag*>(pxBuffer++))->Set( 2,
                                                            false,
                                                            false,
                                                            0,
                                                            SCE_GIF_PACKED,
                                                            1,
                                                            SCE_GIF_PACKED_AD );

	// This QW will contain a SCE_GS_BITBLTBUF; that's written prior to use
    // by a call to SetLoadImageDest().
#if _DEBUG
    pxBuffer->u128 = ASURA_PS2_INIT_ULONG128( 0xcdcdcdcdcdcdcdcdcdcdcdcdcdcdcdcdLL );
#endif
    ++pxBuffer;

	pxBuffer->u64s[0] = SCE_GS_SET_TRXREG( 16, mbh * 16 );
	pxBuffer->u64s[1] = SCE_GS_TRXREG;
    ++pxBuffer;

	for( u_int mx = 0; mx < mbw; ++mx )
	{
		bool bEOP = ( mx == mbw - 1 );

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 4,
                                                                Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                Asura_PS2_DmaTag::ID_CNT,
                                                                false,
                                                                0,
                                                                false,
                                                                0,
                                                                0 );

        (reinterpret_cast<Asura_PS2_GifTag*>(pxBuffer++))->Set( 2,
                                                                false,
                                                                false,
                                                                0,
                                                                SCE_GIF_PACKED,
                                                                1,
                                                                SCE_GIF_PACKED_AD );
		
		pxBuffer->u64s[0] = SCE_GS_SET_TRXPOS( 0, 0, mx*16, 0, 0 );
		pxBuffer->u64s[1] = SCE_GS_TRXPOS;
        ++pxBuffer;

        pxBuffer->u64s[0] = SCE_GS_SET_TRXDIR( 0 );
		pxBuffer->u64s[1] = SCE_GS_TRXDIR;
        ++pxBuffer;

        (reinterpret_cast<Asura_PS2_GifTag*>(pxBuffer++))->Set( 64 * mbh,
                                                                bEOP,
                                                                false,
                                                                0,
                                                                SCE_GIF_IMAGE,
                                                                0 );          

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 64 * mbh,
                                                                Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                Asura_PS2_DmaTag::ID_REF,
                                                                false,
                                                                reinterpret_cast<u_int>(pvImageData),
                                                                false,
                                                                0,
                                                                0 );

		pvImageData = reinterpret_cast<u_long*>(pvImageData) + ((16 * 16 * 4 * mbh) / 8);
	}
	
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxBuffer++))->Set( 0,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_RET,
                                                            false,
                                                            0,
                                                            false,
                                                            0,
                                                            0 );
}

#endif // ASURA_PS2_USE_NEW_FMV_STUFF
