//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_fmv.cpp
//
// Copyright (c) 2004-2005 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "FMV/Asura_PS2_fmv.h"

// Stdlib platform-specific.
#include <libmpeg.h>

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Logfile.h"
#include "Sound/Asura_Sound.h"

// Engine platform-specific.
#include "Core/Asura_PS2_OS.h"
#include "Sound/Asura_PS2_Multistream.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

#ifdef ASURA_PS2_SUPPORT_DLLS
static const char* const szMPEG_DLL_NAME = "ps2_mpeg";
static const char* const szMPEG_DLL_FILE = "Misc\\mpeg_dll.asr";
#endif

// G L O B A L S /////////////////////////////////////////////////////////////

DEBUGVAR bool dbg_bEnableFMVs = true;

// (MP - 17.07.2003) Force a minimum time so user can't skip accidentally.
DEBUGVAR float dbg_fMinTimeBeforeSkipAllowed = 0.25f;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

u_int                                Asura_PS2_FMV::s_uFramesBeforeSkipAllowed;
Asura_PS2_FMV::RENDER_SUBTITLE_CB_FN Asura_PS2_FMV::s_pfnRenderSubtitleCB;
Asura_PS2_VideoMode                  Asura_PS2_FMV::s_eVideoModeToUse;
u_int                                Asura_PS2_FMV::s_uAudioStreamIdx;
Asura_PS2_FMV::MOVIE_SIZE_CB_FN      Asura_PS2_FMV::s_pfnMovieSizeCB;

Asura_PS2_FMV::PLAYRESULT Asura_PS2_FMV::Play( const char* szFilename,
                                               float fTimeBeforeSkipAllowed,
                                               RENDER_SUBTITLE_CB_FN pfnRenderSubtitleCB,
                                               Asura_PS2_VideoMode eForceVideoMode,
                                               u_int uAudioStreamIdx,
                                               MOVIE_SIZE_CB_FN pfnMovieSizeCB )
{
	if( !dbg_bEnableFMVs ) { return PLAYRESULT_DIDNOTSTART; }

    s_pfnRenderSubtitleCB = pfnRenderSubtitleCB;
    s_eVideoModeToUse     = ( ASURA_PS2_VIDEOMODE_USEDEFAULT == eForceVideoMode ? Asura_PS2_GS::GetVideoMode() : eForceVideoMode );
    s_uAudioStreamIdx     = uAudioStreamIdx;
    s_pfnMovieSizeCB      = pfnMovieSizeCB;

	// (MH - 18.11.2004) moved this hack to allow skip on non-CD builds out of DLL.
#ifdef ASURA_PS2_CDEMUL
    fTimeBeforeSkipAllowed = Asura_Maths::Max( fTimeBeforeSkipAllowed, dbg_fMinTimeBeforeSkipAllowed );
#else
	fTimeBeforeSkipAllowed = 0.0f;
#endif
	
	const float fFPS = ( ASURA_PS2_VIDEOMODE_PAL == s_eVideoModeToUse ? 25.0f : 29.97f );

	// AW - 09/07/03 - Take care to avoid overflow - they might use a very large value to ban skipping.
	if( fTimeBeforeSkipAllowed > 50000.0f )
	{
		s_uFramesBeforeSkipAllowed = UINT_MAX;
	}
	else
	{
		s_uFramesBeforeSkipAllowed = static_cast<u_int>( Asura_Maths::FloatToInt(fTimeBeforeSkipAllowed * fFPS) );
	}

	// HE - 20/08/2007 - make sure no sounds of any kind are playing
	Asura_Sound_System::PauseAllSystems( true );

	Asura_PS2_Multistream::DisableDSPProcessing( true );

#ifdef ASURA_PS2_SUPPORT_DLLS
	// don't fail if libmpeg.a is statically linked.
	int (*pfnMpegInit)(void) = sceMpegInit;
	if( !pfnMpegInit && !IsDLLLoaded() )
	{
		Asura_OutputToDebugger_Retail(( "FMV DLLs not loaded, cannot play movie.\n" ));
		return PLAYRESULT_DIDNOTSTART;
	}
#endif

    PLAYRESULT eResult = PlayInternal( szFilename );

    Asura_PS2_Multistream::EnableDSPProcessing();
	Asura_Sound_System::UnpauseAllSystems();

    return eResult;
}

#ifdef ASURA_PS2_SUPPORT_DLLS

bool Asura_PS2_FMV::IsDLLLoaded( void )
{
    return Asura_PS2_OS::IsDllLoaded( szMPEG_DLL_NAME );
}

bool Asura_PS2_FMV::LoadDLL( void )
{
    if( !Asura_PS2_OS::IsDllLoaded(szMPEG_DLL_NAME) )
    {
        return Asura_ChunkLoading_System::LoadAsuraFile( szMPEG_DLL_FILE );
    }
    return true;
}

void Asura_PS2_FMV::UnloadDLL( void )
{
    if( Asura_PS2_OS::IsDllLoaded(szMPEG_DLL_NAME) )
    {
        Asura_PS2_OS::ShutdownAndUnloadDll( szMPEG_DLL_NAME );
    }
}

#endif // #ifdef ASURA_PS2_SUPPORT_DLLS



#if 0

// (MH - 08.11.2004) if you need to debug the FMV stuff without it being in a DLL, add 
// libmpeg.a libipu.a to the link list and #include the following files.
// (MP - 17.10.2005) Note that VisualC won't bother recompiling this file
// if any of these included files change. Yay.
#include "FMV/Asura_PS2_audiodec.cpp"
#include "FMV/Asura_PS2_fmvmain.cpp"
#include "FMV/Asura_PS2_ldimage.cpp"
#include "FMV/Asura_PS2_playpss.cpp"
#include "FMV/Asura_PS2_strfile.cpp"

#endif
