//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_usbmouse.cpp
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Derived class implementation of USB mouse support
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Input/Asura_PS2_usbmouse.h"

#include "Debugging/Asura_Assert.h"
#include "Core/Asura_UserInput.h"
#include "Core/Asura_Timers.h"

#include <string.h>     // for memset
#include <stdio.h>

#if ASURA_PS2_SUPPORT_MOUSE

Asura_PS2_USBMouse::Asura_PS2_USBMouse(int iPortNumber)
				   :Asura_PS2_USBDevice(iPortNumber)
{
	m_xDeviceType = USB_DT_MOUSE;
}

Asura_PS2_USBMouse::~Asura_PS2_USBMouse(void)
{
}

void Asura_PS2_USBMouse::Update(void)
{
    switch(m_eCurrentPhase)
    {
        case USB_PHASE_NONE:
        break;

        case USB_PHASE_INITIALISING:
        {
            // Reset variables
            memset(&m_xMouseDeviceDesc, 0, sizeof(m_xMouseDeviceDesc));
            memset(&m_xMouseData, 0, sizeof(m_xMouseData));
            m_iEnumNeeded  = LGKBM_HINT_NOENUMNEEDED;
            m_iMouseDevice = LGKBM_INVALID_DEVICE;
            m_iLocation = -1;
            m_eCurrentPhase = USB_PHASE_ENUMERATING;
        }
        break;

        case USB_PHASE_ENUMERATING:
        {
            // Enumerate only one mouse as we only support one
            int iRC = lgMouseEnumerate(0, &m_xMouseDeviceDesc);
            if(iRC == LGKBM_SUCCESS)
            {
                m_eCurrentPhase = USB_PHASE_ENUMERATED;
            }
            else
            {
                m_eCurrentPhase = USB_PHASE_ERROR;
            }
        }
        break;

        case USB_PHASE_ENUMERATED:
        {
            // Try to open the mouse
            if(lgMouseOpen(0, &m_iMouseDevice) == LGKBM_SUCCESS)
            {
                m_eCurrentPhase = USB_PHASE_GET_LOCATIONS;
            }
            else
            {
                m_eCurrentPhase = USB_PHASE_ERROR;
            }
        }
        break;

        case USB_PHASE_GET_LOCATIONS:
		{
            m_iLocation = static_cast<int>(m_xMouseDeviceDesc.device_location[0]);
// (MH - 15.01.2004) this method of forcing device numbers to match locations
// appears flawed when you consider that we should support hubs. The new PSX
// model of PS2 has only a single USB port. Therefore we allow any location
// and skip this test.
//            // Check whether it's plugged into the right port - we only support one mouse remember!
//            if(m_iLocation != m_iDeviceNumber)
//            {
//                m_eCurrentPhase = USB_PHASE_ERROR;
//            }
//            else
//            {
//                m_eCurrentPhase = USB_PHASE_READY;
//            }
			m_eCurrentPhase = USB_PHASE_READY;
		}
        break;

        case USB_PHASE_BUSY:
            // Doesn't do anything - here to stop compiler complaining
        break;

        case USB_PHASE_ERROR:
            // Error - so close the device and set the state to poll for new devices
            lgMouseClose(m_iMouseDevice);
            m_iMouseDevice  = LGKBM_INVALID_DEVICE;
            m_eCurrentPhase = USB_PHASE_NO_DEVICES;
        break;

        case USB_PHASE_NO_DEVICES:
            // Check for an enumhint and enumerate if needed
            if(m_iEnumNeeded == LGKBM_HINT_ENUMNEEDED)
            {
                m_eCurrentPhase = USB_PHASE_INITIALISING;
            }
            lgMouseEnumHint(&m_iEnumNeeded);
        break;

        case USB_PHASE_CALIBRATE:
            // Doesn't do anything - here to stop compiler complaining
        break;

        case USB_PHASE_READY:
            // Check previous frames return of hint needed
            if(m_iEnumNeeded == LGKBM_HINT_ENUMNEEDED)
            {
                m_eCurrentPhase = USB_PHASE_ERROR;
            }
            // Read data
            else if(lgMouseRead(m_iMouseDevice, &m_xMouseData) != LGKBM_SUCCESS)
            {
                m_eCurrentPhase = USB_PHASE_ERROR;
            }
            // Check for plug / unplug hint
            else
            {
                lgMouseEnumHint(&m_iEnumNeeded);
            }
        break;

        case USB_PHASE_DEVICE_SPECIFIC:
            // Doesn't do anything - here to stop compiler complaining
        break;
    }
}

void Asura_PS2_USBMouse::Translate(u_char *piAsuraInputBufferOffset, float *piAsuraAxis_BufferOffset)
{  
    // Check that device is ready and data has been read
    if(m_eCurrentPhase != USB_PHASE_READY)
    {
        return;
    }
    // Check for packets and translate them into mouse movement
    for(int i=0; i<m_xMouseData.num_pkts; ++i)
    {
        // Translate the buttons first
        lgMousePacket* pxCurrentPacket = &(m_xMouseData.mouse_pkts[i]);
	    *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEBUTTON1) = 255 * lgIsMouseButtonPressed(pxCurrentPacket->button, 1);
	    *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEBUTTON2) = 255 * lgIsMouseButtonPressed(pxCurrentPacket->button, 2);
	    *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEBUTTON3) = 255 * lgIsMouseButtonPressed(pxCurrentPacket->button, 3);
	    *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEBUTTON4) = 255 * lgIsMouseButtonPressed(pxCurrentPacket->button, 4);
	    *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEBUTTON5) = 255 * lgIsMouseButtonPressed(pxCurrentPacket->button, 5);

        // Mouse wheel
        if(pxCurrentPacket->Wheel > 0)
        {
            *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEWHEELUP) = 255;
        }
        else if(pxCurrentPacket->Wheel < 0)
        {
            *(piAsuraInputBufferOffset + ASURA_KEYID_MOUSEWHEELDOWN) = 255;
        }
    }

    // Do the axis - from the summary packet is fine
    // This is pretty much copied from what the PC does
    float fHorizontalVelocity = m_xMouseData.summary_pkt.X * Asura_Timers::GetFrameRate();
    float fVerticalVelocity   = m_xMouseData.summary_pkt.Y * Asura_Timers::GetFrameRate();
    piAsuraAxis_BufferOffset[ASURA_AXIS_MOUSE_V] = fVerticalVelocity/512.0f;
	piAsuraAxis_BufferOffset[ASURA_AXIS_MOUSE_H] = fHorizontalVelocity/512.0f;
}

bool Asura_PS2_USBMouse::Available(void)
{
    return m_eCurrentPhase == USB_PHASE_READY;
}

void Asura_PS2_USBMouse::Initialise(void)
{
    // Set initial phase only - everything else handled by state machine above
    m_eCurrentPhase = USB_PHASE_NO_DEVICES;
}

#endif // #if ASURA_PS2_SUPPORT_MOUSE
