//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_MemoryCard.cpp
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Memory card save / load functions
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "MemoryCard/Asura_PS2_MemoryCard.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libcdvd.h>
#include <libscf.h>

#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "MemoryCard/Asura_PS2_MemoryCard_SubSystem.h"
#include "MemoryCard/Asura_PS2_MemoryCard_Error.h"
#include "MemoryCard/SCE_PS2_ShiftJis.h"
#include "Core/Asura_Checksum.h"

// library includes
#include <time.h>

#ifndef ASURA_PS2_DEMODISK
/*************************** ICON CLASS FUNCTIONS *********************************************************/

int		Asura_PS2_MemoryCard_Icon::s_iIconBufferSize[3];
char* 	Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[3];
sceMcIconSys   Asura_PS2_MemoryCard_Icon::s_xIconSysFile;

void Asura_PS2_MemoryCard_Icon::Initialise(void)
{
	s_pcIconBuffer[0] = 
	s_pcIconBuffer[1] = 
	s_pcIconBuffer[2] = NULL;

	s_xIconSysFile.Head[0] = 'P';	// NB see Table 4-6 in "PS2_EE-Overview-Device"
	s_xIconSysFile.Head[1] = 'S';
	s_xIconSysFile.Head[2] = '2';
	s_xIconSysFile.Head[3] = 'D';

	// Set the reserved areas of the icon files to zero
	s_xIconSysFile.Reserv1 = 0x0000;
	s_xIconSysFile.Reserv2 = 0x00000000;
	s_xIconSysFile.OffsLF = 32;
	unsigned int* puiReserve3 = reinterpret_cast<unsigned int*>(s_xIconSysFile.Reserve3);
	for(int i = 0; i < 128; ++i)
	{
		puiReserve3[i] = 0x00000000;
	}

	s_xIconSysFile.FnameView[0] = 0;
	s_xIconSysFile.FnameCopy[0] = 0;
	s_xIconSysFile.FnameDel[0]  = 0;
}

void Asura_PS2_MemoryCard_Icon::Shutdown(void)
{
	for(int i=0; i<3; ++i)
	{
		if(s_pcIconBuffer[i])
		{
			Asura_Deallocate(s_pcIconBuffer[i]);
			Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[i] = NULL;
		}
	}
}

void Asura_PS2_MemoryCard_Icon::SetFileName(const Asura_PS2_MemoryCard_IconType iIconType, 
											const char* pcIconFileName, int iFileSize)
{
	// Allocate the memory and load the icon files
	int iFileID;
	Asura_Assert(strlen(pcIconFileName) < 64, "Invalid file name for icon : length >= 64");
	switch(iIconType)
	{
		case ASURA_PS2_MEMORYCARD_ICON_FILE:
		{
			if(s_pcIconBuffer[0])
			{
				Asura_Deallocate(s_pcIconBuffer[0]);
			}

			strcpy(reinterpret_cast<char*>(s_xIconSysFile.FnameView), pcIconFileName);

			s_iIconBufferSize[0] = iFileSize;
			s_pcIconBuffer[0] = reinterpret_cast<char*>(Asura_Allocate(s_iIconBufferSize[0]));
			iFileID = Asura_FileAccess::Open(pcIconFileName);
            Asura_Assertf(iFileID != ASURA_INVALID_HANDLE, ("Failed to open icon file %s.", pcIconFileName));
            if (iFileID != ASURA_INVALID_HANDLE)
            {
			    Asura_FileAccess::ReadData(iFileID, s_pcIconBuffer[0], s_iIconBufferSize[0]);
			    Asura_FileAccess::Close(iFileID);
            }
		}
		break;
		case ASURA_PS2_MEMORYCARD_ICON_DELETE:
		{
			if(s_pcIconBuffer[1])
			{
				Asura_Deallocate(s_pcIconBuffer[1]);
			}

			strcpy(reinterpret_cast<char*>(s_xIconSysFile.FnameCopy), pcIconFileName);

			s_iIconBufferSize[1] = iFileSize;
			iFileID = Asura_FileAccess::Open(pcIconFileName);
            Asura_Assertf(iFileID != ASURA_INVALID_HANDLE, ("Failed to open icon file %s.", pcIconFileName));
            if (iFileID != ASURA_INVALID_HANDLE)
            {
			    s_pcIconBuffer[1] = reinterpret_cast<char*>(Asura_Allocate(s_iIconBufferSize[1]));
			    Asura_FileAccess::ReadData(iFileID, s_pcIconBuffer[1], s_iIconBufferSize[1]);
			    Asura_FileAccess::Close(iFileID);
            }
		}
		break;
		case ASURA_PS2_MEMORYCARD_ICON_COPY:
		{
			if(s_pcIconBuffer[2])
			{
				Asura_Deallocate(s_pcIconBuffer[2]);
			}

			strcpy(reinterpret_cast<char*>(s_xIconSysFile.FnameDel), pcIconFileName);

			s_iIconBufferSize[2] = iFileSize;
			iFileID = Asura_FileAccess::Open(pcIconFileName);
            Asura_Assertf(iFileID != ASURA_INVALID_HANDLE, ("Failed to open icon file %s.", pcIconFileName));
            if (iFileID != ASURA_INVALID_HANDLE)
            {
			    s_pcIconBuffer[2] = reinterpret_cast<char*>(Asura_Allocate(s_iIconBufferSize[2]));
			    Asura_FileAccess::ReadData(iFileID, s_pcIconBuffer[2], s_iIconBufferSize[2]);
			    Asura_FileAccess::Close(iFileID);
            }
		}
		break;
	}
}

void Asura_PS2_MemoryCard_Icon::SetBackgroundColour(const Asura_PS2_MemoryCard_Icon_BackgroundPosition iBackPos, 
			   										const unsigned int usColour)
{
	SetBackgroundColour(iBackPos,
						(usColour & 0x00FF0000) >> 16,
						(usColour & 0x0000FF00) >> 8,
						(usColour & 0x000000FF));
}

void Asura_PS2_MemoryCard_Icon::SetBackgroundColour(const Asura_PS2_MemoryCard_Icon_BackgroundPosition iBackPos, 
													const unsigned char ucR, 
													const unsigned char ucG, 
													const unsigned char ucB)
{
	Asura_Assert(iBackPos < NUMBER_POSITIONS, "iBackPos >= NUMBER_POSITIONS");
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.BgColor[iBackPos][0] = ucR;
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.BgColor[iBackPos][1] = ucG;
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.BgColor[iBackPos][2] = ucB;
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.BgColor[iBackPos][3] = 0;
}

void Asura_PS2_MemoryCard_Icon::SetBackgroundTransparency(const unsigned char ucTrans)
{
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.TransRate = 0x000000FF & ucTrans;
}

void Asura_PS2_MemoryCard_Icon::SetLightSource(const int iLightSourceNumber, 
											   const Asura_Vector_3& xLightSourcePosition,
											   const Asura_Vector_3& xLightSourceColour)
{
	Asura_Assert(iLightSourceNumber < 3, "Only three light sources allowed for memory card icon");

	// Direction
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightDir[iLightSourceNumber][0] = xLightSourcePosition[0];
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightDir[iLightSourceNumber][1] = xLightSourcePosition[1];
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightDir[iLightSourceNumber][2] = xLightSourcePosition[2];
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightDir[iLightSourceNumber][3] = 0.0f;

	// Colour
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightColor[iLightSourceNumber][0] = xLightSourceColour[0];
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightColor[iLightSourceNumber][1] = xLightSourceColour[1];
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightColor[iLightSourceNumber][2] = xLightSourceColour[2];
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.LightColor[iLightSourceNumber][3] = 0.f;
}

void Asura_PS2_MemoryCard_Icon::SetAmbient(const float fR, 
										   const float fG, 
										   const float fB)
{
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.Ambient[0] = fR;
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.Ambient[1] = fG;
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.Ambient[2] = fB;
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.Ambient[3] = 0.0f;
}

/*************************** SYSTEM FUNCTIONS *********************************************************/

char*	     Asura_PS2_MemoryCard_System::s_pcFileTitleOutput;
sceMcIconSys Asura_PS2_MemoryCard_System::s_xTempIconSysHeader;
int			 Asura_PS2_MemoryCard_System::s_iPendingOperation = 0;

enum		 
{
	GET_FILE_TITLE = 1,
};

char  Asura_PS2_MemoryCard_System::s_acSaveName[21];
char  Asura_PS2_MemoryCard_System::s_acFileName[7];
char  Asura_PS2_MemoryCard_System::s_acProductCode[13];
int	  Asura_PS2_MemoryCard_System::s_iSaveNumber;

void Asura_PS2_MemoryCard_System::Initialise()
{
	Asura_PS2_MemoryCard_SubSystem::Initialise();
	Asura_PS2_MemoryCard_Icon::Initialise();
	s_acFileName[0]	= '\0';
	s_acSaveName[0] = '\0';
	s_acProductCode[0] = '\0';
	s_iSaveNumber	= -1;
}

void Asura_PS2_MemoryCard_System::Shutdown()
{
	Asura_PS2_MemoryCard_Icon::Shutdown();
	Asura_PS2_MemoryCard_SubSystem::Shutdown();
}

void Asura_PS2_MemoryCard_System::Update()
{
	Asura_PS2_MemoryCard_SubSystem::Update();

	if(Asura_PS2_MemoryCard_SubSystem::s_xState[0] == ASURA_PS2_MEMORY_CARD_READY)
	{
		switch(s_iPendingOperation)
		{
			case 0:
			break;
			case GET_FILE_TITLE:
			{
				Sjis2AsciiString(s_xTempIconSysHeader.TitleName, s_pcFileTitleOutput);
				s_iPendingOperation = 0;
			}
			break;
			default:
			{
			}
		}
	}
}

void Asura_PS2_MemoryCard_System::Rescan(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
}

bool Asura_PS2_MemoryCard_System::Sync(const int iSyncMode)
{
	switch(iSyncMode)
	{
		case 0:		// wait
		{
			while(!(Asura_PS2_MemoryCard_SubSystem::s_iCurrentNumberOfCommands == 0 &&
					Asura_PS2_MemoryCard_SubSystem::s_xState[0] == ASURA_PS2_MEMORY_CARD_READY))
			{
				Update();
			}
			return true;
		}

		case 1:		// do not wait
		{
			return (Asura_PS2_MemoryCard_SubSystem::s_iCurrentNumberOfCommands == 0 &&
					Asura_PS2_MemoryCard_SubSystem::s_xState[0] == ASURA_PS2_MEMORY_CARD_READY);
		}

		default:
		{
		}
	}
	return false;
}

bool Asura_PS2_MemoryCard_System::Available(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	return Asura_PS2_MemoryCard_SubSystem::s_xMemoryCardInfo[iMemoryCardPort][iMemoryCardSlot].iMemoryCardType == sceMcTypePS2;
}

bool Asura_PS2_MemoryCard_System::Formatted(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	return Asura_PS2_MemoryCard_SubSystem::s_xMemoryCardInfo[iMemoryCardPort][iMemoryCardSlot].iFormatted == 1;
}

int Asura_PS2_MemoryCard_System::GetFreeSpace(const int iMemoryCardPort, const int iMemoryCardSlot)
{
    // (NRM - 24.03.2003) convert to bytes
	return 1024 * Asura_PS2_MemoryCard_SubSystem::s_xMemoryCardInfo[iMemoryCardPort][iMemoryCardSlot].iFreeKiloBytes;
}

int Asura_PS2_MemoryCard_System::GetNumberOfEntries(void)
{
    return Asura_PS2_MemoryCard_SubSystem::s_iNumberOfEntries;
}

int Asura_PS2_MemoryCard_System::GetProjectFileSize(void)
{
	int iFileSize = 0;
	for(int i=0; i < Asura_PS2_MemoryCard_SubSystem::s_iNumberOfEntries; ++i)
	{
		// Don't look at the attribs of "." and ".."
		if(Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].EntryName[0] != '.' &&
			Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].EntryName[1] != '.')
		{
			iFileSize += Asura_Maths::FloatToInt_Chop((Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].FileSizeByte + 1023.f) / 1024.f) * 1024;
		}
	}
	// Adjustment according to tech note
	if(iFileSize > 0)		// but only if files were actually found
	{
        // (NRM - 02.07.2003) -1 as it's +1 from the technote but -2 for "." & ".."
        iFileSize += Asura_Maths::FloatToInt_Chop(((Asura_PS2_MemoryCard_SubSystem::s_iNumberOfEntries - 1.f) / 2.f) + 2.f) * 1024;
	}
	return iFileSize;
}

void Asura_PS2_MemoryCard_System::GetDirList(const int iMemoryCardPort, 
										     const int iMemoryCardSlot,
											 char* pcOutputBuffer,
											 const bool bProjectOnlyEntries/*= false*/)
{
	// Add a command to check the card - will still work if swapped
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoGetDir;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	char* pcFileFilter = reinterpret_cast<char*>(&(xCommand.iParameters[2]));
	if(bProjectOnlyEntries)
	{
		strncpy(pcFileFilter, s_acSaveName, 11);
		pcFileFilter[12] = 0;
		strcat(pcFileFilter, "*");
	}
	else
	{
		pcFileFilter[0] = '*';
		pcFileFilter[1] = '\0';
	}
	xCommand.iParameters[31] = reinterpret_cast<int>(pcOutputBuffer);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

bool Asura_PS2_MemoryCard_System::AttributesOfProjectFilesCorrect(void)
{
	// Correct attribute for file is 0x8497
	bool bCorrect = true;
	for(int i=0; i < Asura_PS2_MemoryCard_SubSystem::s_iNumberOfEntries; ++i)
	{
		// Don't look at the attribs of "." and ".."
		if(Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].EntryName[0] != '.' &&
			Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].EntryName[1] != '.')
		{
            // 0x8497 is the correct attributes for a file written succesfully
			bCorrect &= (Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].AttrFile == 0x8497);
		}
	}
	return bCorrect;
}

void Asura_PS2_MemoryCard_System::ChangeDir(const int iMemoryCardPort, 
										    const int iMemoryCardSlot,
											const char* pcDirectory)
{
	// Add a command to check the card - will still work if swapped
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoChDir;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	// NRM - as max length is 20 characters the 30 ints is plenty room to store the directory path
	strcpy(reinterpret_cast<char*>(&(xCommand.iParameters[2])), pcDirectory);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_System::ChangeToProductDir(const int iMemoryCardPort, const int iMemoryCardSlot)
{
    MakeDirectoryName();
	char acDirectory[22];
	acDirectory[0] = '/';
	acDirectory[1] = '\0';
	strcat(acDirectory, s_acSaveName);
	ChangeDir(iMemoryCardPort, iMemoryCardSlot, acDirectory);
}

void Asura_PS2_MemoryCard_System::Rename(const int iMemoryCardPort, const int iMemoryCardSlot,
										 const char* pcOriginalName, const char* pcNewName)
{
	// Add a command to check the card - will still work if swapped
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoRename;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	strcpy(reinterpret_cast<char*>(&(xCommand.iParameters[2])), pcOriginalName);
	strcpy(reinterpret_cast<char*>(&(xCommand.iParameters[16])), pcNewName);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_System::Format(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	// Add a command to check the card - will still work if swapped
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoFormat;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

#ifdef _DEBUG		// Don't want to encourage use of this in commercial release
void Asura_PS2_MemoryCard_System::Unformat(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	// Add a command to check the card - will still work if swapped
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoUnformat;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}
#endif


void Asura_PS2_MemoryCard_System::SetProductCode(const char* pcProductCode)
{
	Asura_Assert(strlen(pcProductCode) == 10, "Invalid string length for product code");

	switch(pcProductCode[0] + pcProductCode[1] + pcProductCode[2] + pcProductCode[3])
	{
		// Add the key id based on product code and copy string across
		case ('S' + 'L' + 'E' + 'S'):			// Europe, Middle East and Australasia
		{
			s_acProductCode[0] = 'B';
			s_acProductCode[1] = 'E';
			strcpy(&(s_acProductCode[2]), pcProductCode);
		}
		break;
		case ('S' + 'L' + 'P' + 'S'):			// Japan and Far East
		{
			s_acProductCode[0] = 'B';
			s_acProductCode[1] = 'I';
			strcpy(&(s_acProductCode[2]), pcProductCode);
		}
		break;
		case ('S' + 'L' + 'U' + 'S'):			// USA, Canada and South America
		{
			s_acProductCode[0] = 'B';
			s_acProductCode[1] = 'A';
			strcpy(&(s_acProductCode[2]), pcProductCode);
		}
		break;
		case ('S' + 'L' + 'K' + 'A'):			// Korea
		{
			s_acProductCode[0] = 'B';
			s_acProductCode[1] = 'K';
			strcpy(&(s_acProductCode[2]), pcProductCode);
		}
		break;
		default:
			Asura_Assert(0, "Unknown region in product code");
	}
    // (NRM - 17.07.2003) ensure that the last char is actually set to null terminate - paranoia after Dredd pre-sub
    s_acProductCode[12] = '\0';
}

void Asura_PS2_MemoryCard_System::SetSaveName(const char* pcSaveName)
{
	// copy the first six characters to the filename
	strncpy(s_acFileName, pcSaveName, 6);
}

void Asura_PS2_MemoryCard_System::SetBrowserName(const char* pcDirectoryName, u_short usLineBreakPos)
{
	Asura_Assert(strlen(pcDirectoryName) < 32, "Browser display can only show 32 characters");
	Asura_Assert(usLineBreakPos <= 16, "Browser will only display 16 characters per line");
	Asura_Assert(static_cast<int>(strlen(pcDirectoryName)) - static_cast<int>(usLineBreakPos) <= 16, "Browser will only display 16 characters per line");

    // convert the file name to SHIFT-JIS and copy it to s_xIconSysFile
	// NRM - slightly evil casting to use the Sony Shift-JIS stuff
	AsciiString2Sjis(reinterpret_cast<const unsigned char*>(pcDirectoryName), 
					 reinterpret_cast<unsigned short*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.TitleName));
	// (AW - 08.03.2005) - Because they're wide characters, we must multiply the line break pos by 2.
	Asura_PS2_MemoryCard_Icon::s_xIconSysFile.OffsLF = (usLineBreakPos << 1);
}

void Asura_PS2_MemoryCard_System::SetSaveNumber(const int iSaveNumber)
{
	s_iSaveNumber = iSaveNumber;
}

void Asura_PS2_MemoryCard_System::MakeDirectoryName(void)
{
	// Build the file name string : 1. pProduct code, 2. 6 digit savename, 3. save number
	char acSaveNumberString[2];
	acSaveNumberString[0] = (s_iSaveNumber / 10) + 48;
	acSaveNumberString[1] = (s_iSaveNumber % 10) + 48;
	strcpy(s_acSaveName, s_acProductCode);
	strcat(s_acSaveName, s_acFileName);
	int iNumberPosition = strlen(s_acSaveName);
	s_acSaveName[iNumberPosition]   = acSaveNumberString[0];
	s_acSaveName[iNumberPosition+1] = acSaveNumberString[1];
	s_acSaveName[iNumberPosition+2] = '\0';
}

void Asura_PS2_MemoryCard_System::GetPrimaryName(char* szPrimaryNameOutput)
{
    MakeDirectoryName();
    strcpy(szPrimaryNameOutput, s_acSaveName);
}


void Asura_PS2_MemoryCard_System::WritePrimary(const int iMemoryCardPort, const int iMemoryCardSlot, 
											   const void* pcDataBuffer,
											   const int   iDataSize,
                                               bool bWriteChecksum /*= false*/)
{
    MakeDirectoryName();
	WriteSecondary(iMemoryCardPort, iMemoryCardSlot, pcDataBuffer, s_acSaveName, iDataSize, bWriteChecksum);
}

void Asura_PS2_MemoryCard_System::WriteSecondary(const int iMemoryCardPort, const int iMemoryCardSlot, 
												 const void* pcDataBuffer, const char* pcFileName,
												 const int iDataSize,
                                                 bool bWriteChecksum /*= false*/)
{
	// New write consists of 5 commands:
	// 1. change to the root directory (/) from anywhere (memcard only supports directories in the root)
	// 2. open the file with write properties
	// 3. write,
	// 4. flush the file buffer
	// 5. close the file

	// Write secondary assumes that the primary directory has been created
	// change to that directory - even if in root
	Asura_PS2_MemoryCard_Command xCommand;
	char	acDirectory[25] = "/";
	strcat(acDirectory, s_acSaveName);
	ChangeDir(iMemoryCardPort, iMemoryCardSlot, acDirectory);

	int iWriteMode = SCE_WRONLY | SCE_CREAT;

	// Add the open command
	xCommand.iCommand = sceMcFuncNoOpen;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	xCommand.iParameters[2] = iWriteMode;
	strcpy(reinterpret_cast<char*>(&(xCommand.iParameters[3])), pcFileName);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

	// Add the write command
	xCommand.iCommand = sceMcFuncNoWrite;
	xCommand.iParameters[0] = iDataSize;
	xCommand.iParameters[1] = reinterpret_cast<int>(pcDataBuffer);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

	// Add the flush command
	xCommand.iCommand = sceMcFuncNoFlush;
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

    // write the checksum if requested
    if(bWriteChecksum)
    {
        static u_char aucChecksum[Asura_Checksum::CHECKSUM_LENGTH];
        Asura_Checksum::EncodeFromBuffer(reinterpret_cast<const u_char*>(pcDataBuffer), iDataSize, aucChecksum);

        // Seek to end
//        xCommand.iCommand = sceMcFuncNoSeek;
//        xCommand.iParameters[0] = iDataSize + 1;
//        xCommand.iParameters[1] = 0;
//    	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

        xCommand.iCommand = sceMcFuncNoWrite;
        xCommand.iParameters[0] = Asura_Checksum::CHECKSUM_LENGTH;
        xCommand.iParameters[1] = reinterpret_cast<int>(aucChecksum);
	    Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

	    xCommand.iCommand = sceMcFuncNoFlush;
    	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
    }

	// Add the close command
	xCommand.iCommand = sceMcFuncNoClose;
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_System::WriteDirectory(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	// Create the save directory
    MakeDirectoryName();
	// Change to the root
	Asura_PS2_MemoryCard_Command xCommand;
	ChangeDir(iMemoryCardPort, iMemoryCardSlot, "/");

	// Add the create command
	xCommand.iCommand = sceMcFuncNoMkdir;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	strcpy(reinterpret_cast<char*>(&(xCommand.iParameters[2])), s_acSaveName);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

	// Change to the directory of the product number
	char	acDirectory[25] = "/";
	strcat(acDirectory, s_acSaveName);
	ChangeDir(iMemoryCardPort, iMemoryCardSlot, acDirectory);
}

void Asura_PS2_MemoryCard_System::WriteIcon(const int iMemoryCardPort, const int iMemoryCardSlot, bool bMakeDir /*=true*/)
{
	// CW - 25/04/2008 - The TRC says that if writing fails the save should be displayed as corrupt data.
	// The best way to do this is to write out icon.sys last, as without it data is concidered corrupt.
	// I have reordered the write commands here so that icon.sys is last, also, I have added the option of not
	// creating the directory at this point so that project data can be written between the directory and icons being written.

	if(bMakeDir)
	{
		WriteDirectory(iMemoryCardPort, iMemoryCardSlot);
	}

	// Check that delete and copy have had their filenames set
	if(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameCopy[0] == '\0')
	{
		strcpy(reinterpret_cast<char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameCopy),
			   reinterpret_cast<const char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameView));
	}

	if(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameDel[0] == '\0')
	{
		strcpy(reinterpret_cast<char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameDel),
			   reinterpret_cast<const char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameView));
	}

	// Write the icon files
	if(Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[0])
	{
		WriteSecondary(iMemoryCardPort, iMemoryCardSlot, 
					   Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[0], 
					   reinterpret_cast<char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameView), 
					   Asura_PS2_MemoryCard_Icon::s_iIconBufferSize[0]);
	}

	if(Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[1])
	{
		WriteSecondary(iMemoryCardPort, iMemoryCardSlot, 
					   Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[1], 
					   reinterpret_cast<char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameCopy), 
					   Asura_PS2_MemoryCard_Icon::s_iIconBufferSize[1]);
	}

	if(Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[2])
	{
		WriteSecondary(iMemoryCardPort, iMemoryCardSlot, 
					   Asura_PS2_MemoryCard_Icon::s_pcIconBuffer[2], 
					   reinterpret_cast<char*>(Asura_PS2_MemoryCard_Icon::s_xIconSysFile.FnameDel),
					   Asura_PS2_MemoryCard_Icon::s_iIconBufferSize[2]);
	}

	// Write the icon.sys file
	WriteSecondary(iMemoryCardPort, iMemoryCardSlot, 
		&Asura_PS2_MemoryCard_Icon::s_xIconSysFile, 
		"icon.sys", sizeof(sceMcIconSys));
}

// Read in the data
void Asura_PS2_MemoryCard_System::ReadPrimary(const int iMemoryCardPort, const int iMemoryCardSlot, 
							                  void* pcDataBuffer,  const int iDataSize,
                                              u_char* pucChecksum /* = NULL*/)
{
    MakeDirectoryName();
	Read(iMemoryCardPort, iMemoryCardSlot, s_acSaveName, pcDataBuffer, iDataSize, pucChecksum);
}

void Asura_PS2_MemoryCard_System::Read(const int iMemoryCardPort, const int iMemoryCardSlot, 
									   const char* pcFileName, void* pcDataBuffer,
									   const int iDataSize,
                                       u_char* pucChecksum /* = NULL*/)
{
	// Read consists of 3 commands: open, read, close
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;

	// Add the open command
	xCommand.iCommand = sceMcFuncNoOpen;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	xCommand.iParameters[2] = SCE_RDONLY;		// Read only
	// NRM - as max length is 20 characters the maximum of 30 ints here is plenty room 
	// to store the directory path
	strcpy(reinterpret_cast<char*>(&(xCommand.iParameters[3])), pcFileName);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);	

	// Add the read command
	xCommand.iCommand = sceMcFuncNoRead;
	xCommand.iParameters[0] = iDataSize;
	// Store the address of the output (pcDataBuffer)
	xCommand.iParameters[1] = reinterpret_cast<int>(pcDataBuffer);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);

    // (NRM - 06.08.2003) read in the checksum if data is set
    if(pucChecksum)
    {
        // (NRM - 06.08.2003) seek to the position of the checksum at the end of the file
//        xCommand.iCommand = sceMcFuncNoSeek;
//        xCommand.iParameters[0] = iDataSize + 1;
//        xCommand.iParameters[1] = 0;
//    	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
        
        // read the checksum
        xCommand.iCommand       = sceMcFuncNoRead;
        xCommand.iParameters[0] = Asura_Checksum::CHECKSUM_LENGTH;
        xCommand.iParameters[1] = reinterpret_cast<int>(pucChecksum);
    	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
    }

	// Add the close command
	xCommand.iCommand = sceMcFuncNoClose;
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_System::GetFileTitle(const int iMemoryCardPort, const int iMemoryCardSlot,
											   const char* pcFileName, char* pcFileTitle)
{
	// Get the actual real file title from a file name - use to display the save name to the user
	// This is stored in the file icon.sys in the directory pointed at by pcFileName
	s_pcFileTitleOutput = pcFileTitle;
	ChangeDir(iMemoryCardPort, iMemoryCardSlot, pcFileName);
	Read(iMemoryCardPort, iMemoryCardSlot, "icon.sys", &s_xTempIconSysHeader, sizeof(sceMcIconSys));
	s_iPendingOperation = GET_FILE_TITLE;
}

void Asura_PS2_MemoryCard_System::GetTimeStamp(const char* pcFileName, tm* pxTime)
{
	// Get the date and put it in the pcDate string
	bool found = false;
	int i = 0;
	while(i < MAX_SAVE_GAMES && !found)
	{
		if(strcmp(pcFileName, 
				  reinterpret_cast<const char*>(Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i].EntryName)) == 0)
		{
			found = true;
		}
		else
		{
			++i;
		}
	}

	if (found)
	{
        // (NRM - 22.08.2003) this is all such a convoluted faff - thanks Sony
		sceMcStDateTime& xModifyDate = Asura_PS2_MemoryCard_SubSystem::s_xTableDir[i]._Modify;
        sceCdCLOCK       xModifyLocalDate;

        xModifyLocalDate.second = itob(xModifyDate.Sec);
        xModifyLocalDate.minute = itob(xModifyDate.Min);
        xModifyLocalDate.hour   = itob(xModifyDate.Hour);
        xModifyLocalDate.day    = itob(xModifyDate.Day);
        xModifyLocalDate.month  = itob(xModifyDate.Month);
        xModifyLocalDate.year   = itob(xModifyDate.Year % 100);

        sceScfGetLocalTimefromRTC(&xModifyLocalDate);

        // Output in the correct date format
        pxTime->tm_year = btoi(xModifyLocalDate.year);
        pxTime->tm_mon  = btoi(xModifyLocalDate.month);
        pxTime->tm_mday = btoi(xModifyLocalDate.day);
        pxTime->tm_hour = btoi(xModifyLocalDate.hour);
        pxTime->tm_min  = btoi(xModifyLocalDate.minute);
        pxTime->tm_sec  = btoi(xModifyLocalDate.second);
	}
	else
	{
		pxTime->tm_year = 0;
        pxTime->tm_mon  = 0;
        pxTime->tm_mday = 0;
        pxTime->tm_hour = 0;
        pxTime->tm_min  = 0;
        pxTime->tm_sec  = 0;
	}
}

void Asura_PS2_MemoryCard_System::Delete(const int iMemoryCardPort, const int iMemoryCardSlot, 
										 const char* pcFileName)
{
	Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(iMemoryCardPort, iMemoryCardSlot);
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoDelete;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	strcpy(reinterpret_cast<char*>(&xCommand.iParameters[2]), pcFileName);
	Asura_PS2_MemoryCard_SubSystem::AddCommand(xCommand);
}

int Asura_PS2_MemoryCard_System::GetCardType(const int iMemoryCardPort, const int iMemoryCardSlot)
{
	return Asura_PS2_MemoryCard_SubSystem::s_xMemoryCardInfo[iMemoryCardPort][iMemoryCardSlot].iMemoryCardType;
}

#endif
