//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_CarpetMeshDefines.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//  Notes:
//
//		- The carpet mesh code is based on the GSM VCL renderer
//
//      - The CarpetMesh VU code is written mainly for use with MSCALs. The only
//        place where an MSCNT is ok for rendering a strip. Note that ITOP
//        needs to be set to the vertex count before the MSCNT and that
//        executing the dummy NOP program will be necessary if anything
//        other than the BEGIN_STRIP program has been run immediately
//        beforehand.
//
//      - For point lighting, the number of lights is assumed to be in the
//        range [0, 7]. The light count is stored in the w component of the
//        2nd QW of each light buffer; we don't use the alpha component of
//        the light colours so this is spare.
//
//      - Dynamic light components should be floating point in [0.0, 255.0]
//        (again, when UNPACKed.)
//
//      - Camera flags should be set so that for camera c, bit (7 - 2c) is
//        high iff the mesh is visible to it and bit (6 - 2c) is high iff
//        clipping is required. This slightly odd arrangement is for
//        convenience in the microcode - it's easy to shift up but basically
//        impossible to shift down.
//
// VU memory usage:
//
//      Buffer sizes, etc.:
//
//          GIF tags                    = 4               \__ constant input
//          Constant cam data buffer    = 4 * 5           /   
//          
//          Renderstate                 = 2 * 5           \
//          Per-mesh camera data buffer = 2 * 4 * 7       |__ input that changes sporadically, though at least once per mesh
//          Lighting data buffers       = 2 * ( 2 * 7)    |
//          Carpet-style data           = 2 * (16+8)      /
//          
//          Input triangle buffers      = 2 * 7 * Ntri    >--- input triangle data
//          Working quad buffers        = 3 * Nout        >--- generate quad data
//
//          Clip input buffer           = 3 * Nout        >--- intermediate data
//
//          Geom output buffers         = 2 * (1 + 3Nout) \___ output data
//          RS output buffer            = 8               /
//
//          VI reg stack                = 7               >--- internal usage
//
//      Max vertices:
//
//          4 + 4*5 + 2*5 + 2*4*7 + 2*14 + 2*24 + + 14*Ntri + 2*3Nout + 2*(1+3Nout) + 8 + 7 <= 1024
//
//          4 + 20 + 10 + 56 + 28 + 48 + 14*Ntri + 2*3Nout + 2*(1+3Nout) + 15 <= 1024
//
//          4 + 20 + 10 + 56 + 28 + 48 + 15 + 2 + 14*Ntri + 2*3Nout + 2*(3Nout) <= 1024
//
//              =>
//          183 + 14Ntri + 12Nout <= 1024
//              =>  .
//			14Ntri + 12Nout <= 841
//          
//			Choices
// Note: Nout needs to be multiple 4
//			Ntri | Nout | Nquads  | NoutputVerts (assuming each triangle makes max number of quads))
//			---------------------------------------------------------------------------------------
//            1  | 68   |  17     | 68
//            5  | 64   |  16     | 320
//           10  | 56   |  14     | 560
//           20  | 44   |  11     | 880
//           25  | 40   |  10     | 1000
//           30  | 32   |   8     | 960
//           35  | 28   |   7     | 980
//           40  | 20   |   5     | 800
//
//          For reference the GSM renderer has NoutputVerts = 44
//			So match that to get similar GS performance and not throw up odd DMA stalls and timings
//			e.g. suddenly affect PATH3 timings!
//
//			Ntri = 20, Nout = 44
//
//			So memory used = 183 + 14 * 20 + 12 * 44 
//						   = 183 + 280 + 528
//						   = 999
//
//      Memory layout:
//      
//             0   /--------------------------\
//                 |         GIF tags         |    4 QWs - geometry setup and draw, shadow setup and draw
//             4   |--------------------------|
//                 |    constant cam data     |   20 QWs - (SCISSOR, XYOFFSET, GBC->S scale, GBC->S trans, GBC->SC scale) * 4
//            24   |--------------------------|
//                 |        renderstate       |   10 QWs - (base (ALPHA, TEX1, TEX0, MIPTBP1, TEST) ) * 2
//            34   |--------------------------|
//                 |     per-mesh cam data    |   56 QWs - (O->GBC mat, viewWidthVec, viewHeightVec, viewFwdVec ) * 4 * 2
//            90   |--------------------------|
//                 |        light data        |   28 QWs - (point/dir light data) * 2
//           118   |--------------------------|
//                 |        style data        |   48 QWs - (16 QW's + 8 QW's ) * 2
//           166   |--------------------------|
//                 | input triangle buffers   |  280 QWs - (XYZ * 3 (RGBA+DENSITY) * 3, INFO ) * Ntri * 2
//           446   |--------------------------|
//                 |   working quad buffer    |  132 QWs - (XYZ, ST + ADC, RGBA ) * Nout 
//           578   |--------------------------|
//                 |     clip input buffer    |  132 QWs - (XYZW, ST + should-clip flag, RGBA) * Nout
//           710   |--------------------------|
//                 |      output buffers      |  266 QWs - (draw GIF tag, (ST, RGBAQ, XYZ) * Nout) * 2
//           976   |--------------------------|
//                 |     RS output buffer     |    8 QWs - (setup GIF tag, up to 7 registers)
//           984   |--------------------------|
//                 |        module pos        |    1 QW  - (XYZ)
//           985   |--------------------------|
//                 |         <unused>         |   15 QWs
//          1000   |--------------------------|
//                 |       VI reg stack       |    7 QWs
//          1007   |--------------------------|
//                 |         <unused>         |   16 QWs
//          1023   \--------------------------/
//
//      Memory layout of the output buffers during clipping:
//
//                 :                          :
//             K   |--------------------------|
//                 |   clipped output buffer  | 1+3M QWs
//                 |--------------------------|
//                 |    clipper workspace 1   |   36 QWs
//                 |--------------------------|
//                 |    clipper workspace 2   |   36 QWs
//                 |--------------------------|
//                 :                          :
//
//          where M is the maximum number of vertices that will fit.
//
//              1+3M + 36 + 36 <= 132
//                  =>
//              73 + 3M <= 132
//                  =>
//              M <= 19
//
//          => set M=18.
//
//          (NB: M must be at least 9 since just a single input tri can
//          generate that many output vertices.)
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_CARPET_MESH_DEFINES_H___
#define ___ASURA_PS2_CARPET_MESH_DEFINES_H___ 1

// M A C R O S ///////////////////////////////////////////////////////////////

// Limits.
#define ASURA_PS2_CARPET_MESH__MAX_VERTS_PER_STRIP              (46)
#define ASURA_PS2_CARPET_MESH__MAX_QUADS_PER_STRIP              (11)
#define ASURA_PS2_CARPET_MESH__MAX_TRIS_PER_BATCH               (20)
#define ASURA_PS2_CARPET_MESH__MAX_POINT_LIGHTS                 (7)
#define ASURA_PS2_CARPET_MESH__MAX_CAMERAS                      (4)

// ITOP data: mesh flags.
#define ASURA_PS2_CARPET_MESH__FLAG__POINT_LIT                  (0x080)

// ITOP data: cam flags.
#define ASURA_PS2_CARPET_MESH__FLAG__VIS_FOR_MORE_THAN_ONE_CAM  (0x100)
#define ASURA_PS2_CARPET_MESH__MASK__CAM_VIS_FLAGS              (0x0AA)
#define ASURA_PS2_CARPET_MESH__MASK__CAM_CLIP_FLAGS             (0x055)
#define ASURA_PS2_CARPET_MESH__MASK__CAM_VIS_AND_CLIP_FLAGS     (ASURA_PS2_CARPET_MESH__MASK__CAM_VIS_FLAGS | ASURA_PS2_CARPET_MESH__MASK__CAM_CLIP_FLAGS)

// ITOP data: per-strip, inserted into the packet data at export time.
#define ASURA_PS2_CARPET_MESH__STRIP_MASK__VERTEX_QTY           (0x07F)
// (Space for flags available at 0x080 and 0x100 but not currently required.)

// These match the VU1 clipping bits.
#define ASURA_PS2_CARPET_MESH__CLIP_RIGHT                       (0x001)
#define ASURA_PS2_CARPET_MESH__CLIP_LEFT                        (0x002)
#define ASURA_PS2_CARPET_MESH__CLIP_BOTTOM                      (0x004)
#define ASURA_PS2_CARPET_MESH__CLIP_TOP                         (0x008)
#define ASURA_PS2_CARPET_MESH__CLIP_FAR                         (0x010)
#define ASURA_PS2_CARPET_MESH__CLIP_NEAR                        (0x020)
#define ASURA_PS2_CARPET_MESH__CLIP_ALL                         (0x03F)

// Sizes.
#define ASURA_PS2_CARPET_MESH__SIZE__PER_POINT_LIGHT            (2)

// Data area addresses.
#define ASURA_PS2_CARPET_MESH__ADDR__GIF_TAGS                   (0)
#define ASURA_PS2_CARPET_MESH__ADDR__CONST_CAM_DATA             (4)
#define ASURA_PS2_CARPET_MESH__ADDR__RENDERSTATE_BFR_0          (24)
#define ASURA_PS2_CARPET_MESH__ADDR__RENDERSTATE_BFR_1          (29)
#define ASURA_PS2_CARPET_MESH__ADDR__PER_OBJ_CAM_DATA_BFR_0     (34)
#define ASURA_PS2_CARPET_MESH__ADDR__PER_OBJ_CAM_DATA_BFR_1     (66)
#define ASURA_PS2_CARPET_MESH__ADDR__LIGHT_DATA_BFR_0           (90)
#define ASURA_PS2_CARPET_MESH__ADDR__LIGHT_DATA_BFR_1           (104)
#define ASURA_PS2_CARPET_MESH__ADDR__STYLE_DATA_BFR_0           (118)
#define ASURA_PS2_CARPET_MESH__ADDR__STYLE_DATA_BFR_1           (142)
#define ASURA_PS2_CARPET_MESH__ADDR__INPUT_BFR_0                (166)
#define ASURA_PS2_CARPET_MESH__ADDR__INPUT_BFR_1                (306)
#define ASURA_PS2_CARPET_MESH__ADDR__WORK_QUAD_BFR              (446)
#define ASURA_PS2_CARPET_MESH__ADDR__ST_OVERWRITE_BFR           (578)
#define ASURA_PS2_CARPET_MESH__ADDR__CLIP_INPUT_BFR             (578)
#define ASURA_PS2_CARPET_MESH__ADDR__GEOM_OUTPUT_BFR_0          (710)
#define ASURA_PS2_CARPET_MESH__ADDR__GEOM_OUTPUT_BFR_1          (843) 
#define ASURA_PS2_CARPET_MESH__ADDR__RS_OUTPUT_BFR              (976)
#define ASURA_PS2_CARPET_MESH__ADDR__MODULE_POS_BFR             (984)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK               (994)

// Data area sizes.
#define ASURA_PS2_CARPET_MESH__SIZE__GIF_TAGS                   (4)
#define ASURA_PS2_CARPET_MESH__SIZE__CONST_CAM_DATA             (20)
#define ASURA_PS2_CARPET_MESH__SIZE__RENDERSTATE_BFR            (5)
#define ASURA_PS2_CARPET_MESH__SIZE__PER_OBJ_CAM_DATA_BFR       (28)
#define ASURA_PS2_CARPET_MESH__SIZE__PER_STYLE_DATA_BFR         (24)
#define ASURA_PS2_CARPET_MESH__SIZE__LIGHT_DATA_BFR             (7)
#define ASURA_PS2_CARPET_MESH__SIZE__INPUT_BFR                  (140)
#define ASURA_PS2_CARPET_MESH__SIZE__CLIP_INPUT_BFR             (132)
#define ASURA_PS2_CARPET_MESH__SIZE__OUTPUT_BFR                 (133)
#define ASURA_PS2_CARPET_MESH__SIZE__RS_OUTPUT_BFR              (8)
#define ASURA_PS2_CARPET_MESH__SIZE__MODULE_POS_BFR             (1)
#define ASURA_PS2_CARPET_MESH__SIZE__VI_REG_STACK               (7)
#define ASURA_PS2_CARPET_MESH_SIZE__PER_CAM_DATA_BFR			(7)

// GIF tag buffer details.
#define ASURA_PS2_CARPET_MESH__OFFSET__GIF_TAGS_GEOM_SETUP      (0)
#define ASURA_PS2_CARPET_MESH__OFFSET__GIF_TAGS_GEOM_DRAW       (1)

// Constant cam data buffer details.
#define ASURA_PS2_CARPET_MESH__OFFSET__CONST_CAM_SCISSOR        (0)
#define ASURA_PS2_CARPET_MESH__OFFSET__CONST_CAM_XYOFFSET       (1)
#define ASURA_PS2_CARPET_MESH__OFFSET__CONST_CAM_GBC2S_SCALE    (2)
#define ASURA_PS2_CARPET_MESH__OFFSET__CONST_CAM_GBC2S_TRANS    (3)
#define ASURA_PS2_CARPET_MESH__OFFSET__CONST_CAM_GBC2SC_SCALE   (4)
#define ASURA_PS2_CARPET_MESH__OFFSET__CONST_CAM_NEXT_CAM       (5)

// GIF tag and renderstate input buffer details.
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_IN_BASE_ALPHA         (0)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_IN_BASE_TEX1          (1)	
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_IN_BASE_MIPTBP1       (2)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_IN_BASE_TEX0          (3)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_IN_BASE_TEST          (4)

// Per-object cam data buffer details.
#define ASURA_PS2_CARPET_MESH__OFFSET__CAM_PER_OBJ_O2GBC_MAT		(0)
#define ASURA_PS2_CARPET_MESH__OFFSET__CAM_PER_OBJ_VIEW_WIDTH_VEC	(4)
#define ASURA_PS2_CARPET_MESH__OFFSET__CAM_PER_OBJ_VIEW_HEIGHT_VEC	(5)
#define ASURA_PS2_CARPET_MESH__OFFSET__CAM_PER_OBJ_VIEW_FWD_VEC     (6)
#define ASURA_PS2_CARPET_MESH__OFFSET__CAM_PER_OBJ_NEXT_CAM			ASURA_PS2_CARPET_MESH_SIZE__PER_CAM_DATA_BFR

// Light data buffer details.
#define ASURA_PS2_CARPET_MESH__OFFSET__QW_CONTAINING_QTY        (1)

// Clipper workspace buffer details.
#define ASURA_PS2_CARPET_MESH__SIZE__CLIP_WORKSPACE             (36)  // Up to 9 vertices with 4 QWs per vertex.
#define ASURA_PS2_CARPET_MESH__SIZE__CLIPPED_STS                (18)  // M = max number clip verts
#define ASURA_PS2_CARPET_MESH__OFFSET__CLIP_WORKSPACE           (55)  // 1 + 3M
#define ASURA_PS2_CARPET_MESH__OFFSET__CLIPPED_STS              (127) // 1+3M + 36 + 36

// Geometry output buffer details for normal geometry.
#define ASURA_PS2_CARPET_MESH__OFFSET__GEOM_OUT_DRAW_GIF_TAG    (0)
#define ASURA_PS2_CARPET_MESH__OFFSET__GEOM_OUT_FIRST_VERTEX    (1)

// RS output buffer details for first (base) pass.
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_SETUP_GIF_TAG   (0)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_SCISSOR         (1)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_XYOFFSET        (2)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_ALPHA           (3)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_TEX1            (4)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_MIPTBP1         (5)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_TEX0            (6)
#define ASURA_PS2_CARPET_MESH__OFFSET__RS_OUT_0_TEST            (7)
#define ASURA_PS2_CARPET_MESH__RS_OUT_0_NLOOP                   (7)

// VI register stack aliases.
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_0                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 0)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_1                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 1)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_2                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 2)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_3                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 3)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_4                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 4)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_5                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 5)
#define ASURA_PS2_CARPET_MESH__ADDR__VI_STACK_6                 (ASURA_PS2_CARPET_MESH__ADDR__VI_REG_STACK + 6)

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_CARPET_MESHDEFINES_H___
