//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_GlassDefines.h
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     Contains common settings between Asura_PlatformRenderGlass.cpp and
//     VU1_GlassRender.vcl.
//
// Memory layout:
//
//     (MP - 14.07.2003) NB: Currently we only send one quad at a time so
//     this just budgets for tris (which we actually batch up) and assumes
//     quad stuff will just fit.
//
//     Calculation of max tris per buffer, T:
//
//         Input buffer          = 1 + (2 + 3 * 2) * T
//         Output buffer         = 4 + 3 * 3 * T
//         Clipper input buffer  = 3 * 3 * T
//         IO total              = 2 * (1 + (2 + 3 * 2) * T) + 2 * (4 + 3 * 3 * T)
//                               = 2 + 16 * T + 8 + 18 * T
//                               = 10 + 43 * T
//
//         Cam setup GIF tag     = 1
//         Camera data           = 4 * 17
//         Clipped strip GIF tag = 1
//
//         Space for IO buffers  = 1024 - (1 + 4 * 17 + 1)
//                               = 1024 - 70
//                               = 954
//
//         => T = (94 - 10) / 43
//              = 21
//
//     Overall layout:
//
//        Key: P is poly count for batch.
//             I is per poly vertex count, obtained from ITOP register.
//             V is the total number of vertices in the batch.
//
//        0    /--------------------------\ 
//             |   camera setup GIF tag   |    1 QW  - GIF tag [ NLOOP 2, EOP 0, NREG 1, REG(S): A+D ]
//        1    |--------------------------|
//             |        camera data       |   68 QWs - (W->S, W->C, C->S, SCISSOR, XYOFFSET, pos, tex mat 0, tex mat 1) * 4
//       69    |--------------------------|
//             |      input buffer 0      |  169 QWs - GIF tag, (normal, RGBA, (XYZ, STQ) * I) * P
//      238    |--------------------------|
//             |      input buffer 1      |  169 QWs - GIF tag, (normal, RGBA, (XYZ, STQ) * I) * P
//      407    |--------------------------|
//             |      output buffer 0     |  193 QWs - GIF tag, SCISSOR, XYOFFSET, GIF tag, (STQ, RGBAQ, XYZ) * I * P
//      600    |--------------------------|
//             |      output buffer 1     |  193 QWs - GIF tag, SCISSOR, XYOFFSET, GIF tag, (STQ, RGBAQ, XYZ) * I * P
//      793    |--------------------------|
//             |    clipper input buffer  |  189 QWs - (clip XYZ, ST + clip flags + input ADC, RGBA) * V
//      982    |--------------------------|
//             |   clipped strip GIF tag  |    1 QW  - GIF tag [ NLOOP <from microcode>, EOP 1, NREG 3, REG(S): ST, RGBAQ, XYZ ]
//      983    |--------------------------|
//             |          unused          |   41 QWs
//     1024    \--------------------------/
//
//     Clipper layout:
//
//         During clipping, we use the current output buffer as follows:
//
//             :                          :
//             |--------------------------|
//             |   clipped output buffer  |  4 + 3M QWs
//             |--------------------------|
//             |    clipper workspace 0   |      27 QWs
//             |--------------------------|
//             |    clipper workspace 1   |      27 QWs
//             |--------------------------|
//             :                          :
//
//         This means the maximum number of vertices per clipped output
//         buffer is:
//
//             M = (193 - (4 + 27 + 27)) / 3 = 135 / 3 = 45
//
//         Hence:
//
//             :                          :
//        O    |--------------------------|
//             |   clipped output buffer  |  139 QWs - GIF tag, SCISSOR, XYOFFSET, GIF tag, (STQ, RGBA, XYZ) * M
//  O + 139    |--------------------------|
//             |    clipper workspace 0   |   27 QWs - (XYZ, STQ, RGBA) * 9
//  O + 166    |--------------------------|
//             |    clipper workspace 1   |   27 QWs - (XYZ, STQ, RGBA) * 9
//  O + 193    |--------------------------|
//             :                          :
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_GLASSDEFINES_H___
#define ___ASURA_PS2_GLASSDEFINES_H___

// M A C R O S ///////////////////////////////////////////////////////////////

#define ASURA_PS2_GLASS_MAX_TRIS_PER_BATCH             (  21)
#define ASURA_PS2_GLASS_MAX_QUADS_PER_BATCH            (  15)
#define ASURA_PS2_GLASS_MAX_VERTICES_IN_CLIPPER_OUTPUT (  45)

#define ASURA_PS2_GLASS_NUM_QWS_PER_CAM                (  17)
#define ASURA_PS2_GLASS_NUM_QWS_LIGHT_DATA             (   7)
#define ASURA_PS2_GLASS_NUM_QWS_PER_TRI                (   8)
#define ASURA_PS2_GLASS_NUM_QWS_PER_QUAD               (  10)

#define ASURA_PS2_GLASS_MASK_VERTICES_PER_POLY         (0x0F)

#define ASURA_PS2_GLASS_FLAG_NO_CLIPPING_REQD          (0x10)

#define ASURA_PS2_GLASS_ADDR_CAM_SETUP_GIF_TAG         (   0)
#define ASURA_PS2_GLASS_ADDR_CAM_DATA                  (   1)
#define ASURA_PS2_GLASS_ADDR_INPUT_BUFFER_0            (  69)
#define ASURA_PS2_GLASS_ADDR_INPUT_BUFFER_1            ( 238)
#define ASURA_PS2_GLASS_ADDR_OUTPUT_BUFFER_0           ( 407)
#define ASURA_PS2_GLASS_ADDR_OUTPUT_BUFFER_1           ( 600)
#define ASURA_PS2_GLASS_ADDR_CLIPPER_INPUT_BUFFER      ( 793)
#define ASURA_PS2_GLASS_ADDR_CLIPPED_STRIP_GIF_TAG     ( 982)

#define ASURA_PS2_GLASS_OFFSET_CLIPPER_OUTPUT_BUFFER   (   0)
#define ASURA_PS2_GLASS_OFFSET_CLIPPER_WORKSPACE_0     ( 139)
#define ASURA_PS2_GLASS_OFFSET_CLIPPER_WORKSPACE_1     ( 166)

#define ASURA_PS2_GLASS_SIZE_CAM_DATA                  (4 * ASURA_PS2_GLASS_NUM_QWS_PER_CAM)
#define ASURA_PS2_GLASS_SIZE_INPUT_BUFFER              (ASURA_PS2_GLASS_ADDR_INPUT_BUFFER_1 - ASURA_PS2_GLASS_ADDR_INPUT_BUFFER_0)
#define ASURA_PS2_GLASS_SIZE_OUTPUT_BUFFER             (ASURA_PS2_GLASS_ADDR_OUTPUT_BUFFER_1 - ASURA_PS2_GLASS_ADDR_OUTPUT_BUFFER_0)
#define ASURA_PS2_GLASS_SIZE_CLIPPER_WORKSPACE         (ASURA_PS2_GLASS_OFFSET_CLIPPER_WORKSPACE_1 - ASURA_PS2_GLASS_OFFSET_CLIPPER_WORKSPACE_0)

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_GLASSDEFINES_H___
