//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_RenderEnvironment.cpp
//
// Copyright (c) 2001-2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <libgraph.h>

// Engine independent.
#include "Animation/Asura_TextureAnimation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Light/Asura_Photon.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"

// Engine platform-specific.
#include "Debugging/Asura_Assert.h"
#include "Mesh/Asura_PS2_Env_Store.h"
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_GSMDefines.h"
#include "Rendering/Asura_PS2_GSMRenderer.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"
#include "Rendering/Asura_PS2_VU1.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static const u_int MAX_LIGHTS_PER_MODULE = 6;

#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
static int g_iRainsplashTexID = ASURA_TEXTURE_NONE;
static const char* RAINSPLASH_TEXTURE_NAME = "\\specialfx\\splash.tm2";
#endif

static u_int g_uNumVisibleModules;
static u_int g_uNumVisibleMaterials;
#if defined(ASURA_PS2_SUPPORT_DYNAMIC_LOADING) && _DEBUG
static u_int g_uNumUnloadedVisModules;
#endif

// For both of these, 0 => none of that type of material.
static u_int g_uReflectiveTexBfr;
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
static u_int g_uRainsplashTexBfr;
#endif

static u_int* g_puVisAndClipFlags;
static const Asura_PS2_Env_RuntimeData_MaterialHdr** g_ppxMaterialDrawList;

DEBUGVAR bool  dbg_bOnlyDrawSpecifiedModules   = false;
DEBUGVAR u_int dbg_auListOfModulesToRender[]   = { 0xFFFFFFFFu, 0xFFFFFFFFu, 0xFFFFFFFFu, 0xFFFFFFFFu, 0xFFFFFFFFu, 0xFFFFFFFFu };
DEBUGVAR bool  dbg_bOnlyDrawSpecifiedMaterials = false;
DEBUGVAR int   dbg_aiListOfMaterialsToRender[] = { -1, -1, -1, -1, -1, -1 };
DEBUGVAR bool  dbg_bSkipDualTextureMaterials   = false;

bool dbg_bSkipWater = false;

// C L A S S E S /////////////////////////////////////////////////////////////

class ModuleLightingCache
{

    public:

        // T Y P E S /////////////////////////////////////////////////////////

        typedef Asura_PS2_GSMRenderer::Light Light;

        // C L A S S   M E T H O D S /////////////////////////////////////////

        static void Initialise( void );
        static void ResetForNewFrame( void );
        static void Shutdown( void );

        static void GetLightsForModule( u_int uModuleIdx, const Light** ppxLights, u_int* puNumLights );

    private:

        // C L A S S   S T A T E /////////////////////////////////////////////

        static Light* s_axLightBfr;
        static u_int  s_uLightBfrCapacity;
        static u_int  s_uNumLightsInBfr;

        // Information about each module's entry in the data buffer. If the
        // relevant light count is -1, then we haven't yet calculated the
        // light data for the module this frame.
        static int* s_aiNumLightsPerModule;

};
// End class ModuleLightingCache.

#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
class EnvWaterControl
{

    public:

        static void Update( float fFrameTime );

        static void GetParams( u_int uMaterialFlags, float* pfTransS, float* pfTransT, float* pfC0, float* pfC1 );
        static void GetWaterParams( float* pfTransS, float* pfTransT, float* pfC0, float* pfC1 );
        static void GetWaterfallParams( float* pfTransS, float* pfTransT, float* pfC0, float* pfC1 );

    private:

        static float s_fWaterUVDeltaU;
        static float s_fWaterUVDeltaV;
        static float s_fWaterUPhase;
        static float s_fWaterVPhase;

        static float s_fWaterfallUVDeltaU;
        static float s_fWaterfallUVDeltaV;

};
// End class EnvWaterControl.
#endif

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void SetupRenderstateForFlaggedMaterial( int iMaterialID, u_int uMaterialFlags );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// ModuleLightingCache.
//

ModuleLightingCache::Light* ModuleLightingCache::s_axLightBfr           = NULL;
u_int                       ModuleLightingCache::s_uLightBfrCapacity    = 0;
u_int                       ModuleLightingCache::s_uNumLightsInBfr      = 0;
int*                        ModuleLightingCache::s_aiNumLightsPerModule = NULL;

void ModuleLightingCache::Initialise( void )
{
    Asura_Assert( !s_axLightBfr && !s_aiNumLightsPerModule, "" );

    const u_int uNumModules = Asura_Module_System::GetNumberOfModules();
    const u_int uMaxLights  = uNumModules * MAX_LIGHTS_PER_MODULE;

    s_axLightBfr        = new Light[uMaxLights];
    s_uLightBfrCapacity = uMaxLights;
    s_uNumLightsInBfr   = 0;

    // NB: Stick per-module count array at QW boundary to allow memset() to
    // fill it quickly. (See ResetForNewFrame().)
    s_aiNumLightsPerModule = reinterpret_cast<int*>( Asura_AllocateAlign(uNumModules * sizeof(int), ASURA_ALIGN_16_BYTE) );

    Asura_Assert( s_axLightBfr && s_aiNumLightsPerModule, "Out of memory." );

    // Could force a reset here but that should be done at the beginning of
    // the first frame anyway.
}

void ModuleLightingCache::GetLightsForModule( u_int uModuleIdx, const Light** ppxLights, u_int* puNumLights )
{
    Light* const pxFirstLightForModule = &s_axLightBfr[uModuleIdx * MAX_LIGHTS_PER_MODULE];
    *ppxLights = pxFirstLightForModule;

    u_int uNumLights;

    // Have we already calculated the light data for this module this frame?
    const int iCachedNumLights = s_aiNumLightsPerModule[uModuleIdx];
    if( -1 != iCachedNumLights )
    {
        uNumLights = iCachedNumLights;
    }
    else
    {
        if( Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DYN_LIGHT_ENV) )
        {
            const Asura_CollisionMesh* pxCMesh = Asura_Module_System::GetCollisionMeshForModule( uModuleIdx );

            Asura_Bounding_Box xBB = pxCMesh->GetBoundingBox();
            xBB.Displace( Asura_Module_System::GetPositionOfModule(uModuleIdx) );

            const Asura_Vector_3& xModuleCentre = Asura_Module_System::GetCentreOfModule(uModuleIdx);
            Asura_Photon_System::MakePhotonListForBoundingBox( xBB, xModuleCentre, false, true, MAX_LIGHTS_PER_MODULE );

            uNumLights = Asura_Photon_System::NumberOfLocalPhotons;
        }
        else
        {
            uNumLights = 0;
        }

        Light* pxLightOut = pxFirstLightForModule;
        for( u_int u = 0; u < uNumLights; ++u )
        {
		    const Asura_Local_Photon& xLocalPhoton = Asura_Photon_System::LocalPhotons[u];

            const Asura_Vector_3& xPhotonPosition = xLocalPhoton.GetPosition();
            pxLightOut->afLightPos[0] = xPhotonPosition.x;
            pxLightOut->afLightPos[1] = xPhotonPosition.y;
            pxLightOut->afLightPos[2] = xPhotonPosition.z;

            pxLightOut->fLightRange = xLocalPhoton.GetRange();

            pxLightOut->afLightRGB[0] = xLocalPhoton.GetRedComponent();
            pxLightOut->afLightRGB[1] = xLocalPhoton.GetGreenComponent();
            pxLightOut->afLightRGB[2] = xLocalPhoton.GetBlueComponent();

            ++pxLightOut;
        }

        s_aiNumLightsPerModule[uModuleIdx] = uNumLights;
    }

    *puNumLights = uNumLights;
}

void ModuleLightingCache::ResetForNewFrame( void )
{
    // Set all data lengths to UINT_MAX so that they're all recalculated when
    // required. Note that we don't need to reset ptrs because the length is
    // used to determine whether or not a given ptr is valid.
    memset( &s_aiNumLightsPerModule[0], -1, Asura_Module_System::GetNumberOfModules() * sizeof(int) );
}

void ModuleLightingCache::Shutdown( void )
{
    delete[] s_axLightBfr;
    s_axLightBfr = NULL;

    Asura_DeallocateAlign( s_aiNumLightsPerModule );
    s_aiNumLightsPerModule = NULL;
}

#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS

//
// EnvWaterControl.
//

float EnvWaterControl::s_fWaterUVDeltaU = 0.0f;
float EnvWaterControl::s_fWaterUVDeltaV = 0.0f;
float EnvWaterControl::s_fWaterUPhase = 0.0f;
float EnvWaterControl::s_fWaterVPhase = 0.0f;
float EnvWaterControl::s_fWaterfallUVDeltaU = 0.0f;
float EnvWaterControl::s_fWaterfallUVDeltaV = 0.0f;

 bool dbg_bEnableWaterEffect        = true;
 bool dbg_bEnableWaterEffectUpdates = true;

 float dbg_fWaterScrollUSpeed = 0.7f;
 float dbg_fWaterScrollVSpeed = 0.0f;
 float dbg_fWaterfallScrollUSpeed = 0.3f;
 float dbg_fWaterfallScrollVSpeed = 0.0f;

 float dbg_fWaterAmplitudeUSpeed = 4.8f;
 float dbg_fWaterAmplitudeVSpeed = 12.0f;

 float dbg_fMinWaterAmplitudeU = 0.05f;
 float dbg_fMinWaterAmplitudeV = 0.05f;
 float dbg_fWaterAmplitudeRangeU = 0.1f;
 float dbg_fWaterAmplitudeRangeV = 0.1f;

void EnvWaterControl::Update(const float fFrameTime)
{
    if( !dbg_bEnableWaterEffectUpdates ) { return; }

    s_fWaterUVDeltaU += fFrameTime * dbg_fWaterScrollUSpeed;
    while (s_fWaterUVDeltaU < 0.0f)
    {
        s_fWaterUVDeltaU += 1.0f;
    }
    while (s_fWaterUVDeltaU >= 1.0f)
    {
        s_fWaterUVDeltaU -= 1.0f;
    }

    s_fWaterUVDeltaV += fFrameTime * dbg_fWaterScrollVSpeed;
    while (s_fWaterUVDeltaV < 0.0f)
    {
        s_fWaterUVDeltaV += 1.0f;
    }
    while (s_fWaterUVDeltaV >= 1.0f)
    {
        s_fWaterUVDeltaV -= 1.0f;
    }

    s_fWaterUPhase += fFrameTime * dbg_fWaterAmplitudeUSpeed;
    while (s_fWaterUPhase < 0.0f)
    {
        s_fWaterUPhase += Asura_Maths::Deg360;
    }
    while (s_fWaterUPhase >= Asura_Maths::Deg360)
    {
        s_fWaterUPhase -= Asura_Maths::Deg360;
    }

    s_fWaterVPhase += fFrameTime * dbg_fWaterAmplitudeVSpeed;
    while (s_fWaterVPhase < 0.0f)
    {
        s_fWaterVPhase += Asura_Maths::Deg360;
    }
    while (s_fWaterVPhase >= Asura_Maths::Deg360)
    {
        s_fWaterVPhase -= Asura_Maths::Deg360;
    }

    s_fWaterfallUVDeltaU += fFrameTime * dbg_fWaterfallScrollUSpeed;
    while (s_fWaterfallUVDeltaU < 0.0f)
    {
        s_fWaterfallUVDeltaU += 1.0f;
    }
    while (s_fWaterfallUVDeltaU >= 1.0f)
    {
        s_fWaterfallUVDeltaU -= 1.0f;
    }

    s_fWaterfallUVDeltaV += fFrameTime * dbg_fWaterfallScrollVSpeed;
    while (s_fWaterfallUVDeltaV < 0.0f)
    {
        s_fWaterfallUVDeltaV += 1.0f;
    }
    while (s_fWaterfallUVDeltaV >= 1.0f)
    {
        s_fWaterfallUVDeltaV -= 1.0f;
    }
}

void EnvWaterControl::GetParams( u_int uMaterialFlags, float* pfTransS, float* pfTransT, float* pfC0, float* pfC1 )
{
    if( uMaterialFlags & AsuraMaterialFlag_FlowingWater )
    {
        GetWaterParams( pfTransS, pfTransT, pfC0, pfC1 );
    }
    else
    {
        Asura_Assert( uMaterialFlags & AsuraMaterialFlag_Waterfall, "" );
        GetWaterfallParams( pfTransS, pfTransT, pfC0, pfC1 );
    }
}

void EnvWaterControl::GetWaterParams( float* pfTransS, float* pfTransT, float* pfC0, float* pfC1 )
{
    *pfTransS = s_fWaterUVDeltaU;
    *pfTransT = s_fWaterUVDeltaV;
    *pfC0     = dbg_fMinWaterAmplitudeU + dbg_fWaterAmplitudeRangeU * Asura_Maths::Sine(s_fWaterUPhase);
    *pfC1     = dbg_fMinWaterAmplitudeV + dbg_fWaterAmplitudeRangeV * Asura_Maths::Sine(s_fWaterVPhase);
}

void EnvWaterControl::GetWaterfallParams( float* pfTransS, float* pfTransT, float* pfC0, float* pfC1 )
{
    *pfTransS = s_fWaterfallUVDeltaU;
    *pfTransT = s_fWaterfallUVDeltaV;
    *pfC0     = 0.0f;
    *pfC1     = 0.0f;
}

#endif // #ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS

//
// Asura_Module_System.
//

void Asura_Module_Handler::Platform_Initialise(void)
{
    ModuleLightingCache::Initialise();
}

void Asura_Module_Handler::Platform_Shutdown(void)
{
    ModuleLightingCache::Shutdown();
}

// copied from other renderer
//DEBUGVAR unsigned short dbg_MipMapK2 = 0xffce;// -50
//DEBUGVAR unsigned short dbg_MipMapL2 = 0;
DEBUGVAR u_int dbg_uMMAG = SCE_GS_LINEAR;
#ifdef ASURA_PS2_ENV_USE_TRILINEAR
DEBUGVAR u_int dbg_uMMIN = SCE_GS_LINEAR_MIPMAP_LINEAR;// most expensive mode!
#else
DEBUGVAR u_int dbg_uMMIN = SCE_GS_LINEAR_MIPMAP_NEAREST;
#endif

unsigned short g_uMipMapK;

DEBUGVAR float dbg_fMinFOV = 3.5f * Asura_Maths::DegToRadConstant;
DEBUGVAR float dbg_fMipMapKAtMinFOV = 3991.0f;
DEBUGVAR float dbg_fMaxFOV = 120.0f * Asura_Maths::DegToRadConstant;
DEBUGVAR float dbg_fFOVCutoff = 75.0f * Asura_Maths::DegToRadConstant;
DEBUGVAR float dbg_fMipMapKAtFOVCutoff = 4046.0f;

static const Asura_PS2_Env_RuntimeData_MaterialHdr** g_ppxEnvDrawList = NULL;

static inline u_short GetMipMapK()
{
    const float fCurFOV = Asura_Camera::GetHorizontalFieldOfView();
    Asura_Assert( dbg_fMinFOV <= fCurFOV && fCurFOV <= dbg_fMaxFOV, "FOV outside allowed range." );

    if( fCurFOV > dbg_fFOVCutoff )
    {
        return static_cast<u_short>( Asura_Maths::FloatToInt(dbg_fMipMapKAtFOVCutoff) );
    }
    else
    {
        const float fOneOverFOVRange = 1.0f / (dbg_fFOVCutoff - dbg_fMinFOV);
        const float fLerp = (fCurFOV - dbg_fMinFOV) * fOneOverFOVRange;
        const float fMipMapK = (1.0f - fLerp) * dbg_fMipMapKAtMinFOV + fLerp * dbg_fMipMapKAtFOVCutoff;
        return static_cast<u_short>( Asura_Maths::FloatToInt(fMipMapK) );
    }
}

void Asura_Module_Handler::Platform_BeginRender( void )
{
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DRAW_ENV) )
    {
        g_uNumVisibleMaterials = 0;
        return;
    }

    ModuleLightingCache::ResetForNewFrame();

#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    // (MP - 26.08.2004) TODO: Move this into the level data, as with the
    // reflective texture.
    if( Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_RAINSPLASH) && ASURA_TEXTURE_NONE == g_iRainsplashTexID )
    {
        bool bWasNew;
        g_iRainsplashTexID = Asura_Texture_Management::AddTexture( RAINSPLASH_TEXTURE_NAME, &bWasNew );
        Asura_Assert( !bWasNew, "Rainsplash texture should be included in core gfx.asr if required." );
    }
#endif

#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    if( dbg_bEnableWaterEffect ) { EnvWaterControl::Update( Asura_Timers::GetGameTimeSlice() ); }
#endif
    
    // Setup mipmap K for this frame.
#ifdef ASURA_PS2_USE_LOWEST_MIPMAP_K
    // (AW - 03.09.2003) - This sets the mipmap level for the most zoomed in camera.
    // Not ideal, because if one camera zooms in, then all others will see the change.
    g_uMipMapK = USHRT_MAX;
    for (int i=0;i<Asura_Camera::GetNumberOfCameras();i++)
    {
        Asura_Camera::SetCameraIndex(i);
        g_uMipMapK = Asura_Maths::Min(g_uMipMapK, GetMipMapK());
    }    
#else
    // (MH - 01.05.2002) note that this is setting one mipmap setting for all camera based on the last
    // camera used. There is a chance that this is not the desired behaviour, e.g. if each camera can
    // have a different FOV.
    g_uMipMapK = GetMipMapK();
#endif

    const u_int uNumCams      = Asura_Camera::GetNumberOfCameras();
    const u_int uNumModules   = Asura_Module_System::GetNumberOfModules();
    const u_int uNumMaterials = Asura_PS2_Env_Store::GetNumUniqueMaterials();

    const u_int uQWsReqdForDualTexturePalette = (ASURA_PS2_GSM__MAX_DUAL_TEXTURE_PALETTE_SIZE + 3) >> 2;
    const u_int uQWsForDrawLists              = 1024 - (ASURA_PS2_GSM_SPR_QWS_RESERVED + uQWsReqdForDualTexturePalette);

    // Temporary buffer for dual texture palette. This slightly reduces the
    // maximum amount of stuff we can render - see the calculation just below
    // - but that's pretty huge anyway.
    int* piDualTexturePalette = (Asura_PS2_SPR_Alloc(ASURA_PS2_GSM_SPR_QWS_RESERVED, uQWsReqdForDualTexturePalette))->i32s;

    // Longer lasting buffer for draw list stuff. First section will be vis/
    // clip flags; second section will be visible materials.
    {
        Asura_PS2_QW* pxSPR = Asura_PS2_SPR_Alloc( ASURA_PS2_GSM_SPR_QWS_RESERVED + uQWsReqdForDualTexturePalette, uQWsForDrawLists );

        const u_int uQWsForVisAndClipFlags  = (uNumModules   + 3) >> 2;
        const u_int uQWsForMaterialDrawList = (uNumMaterials + 3) >> 2;

        // Check it'll all fit.
        Asura_Assert( uQWsForVisAndClipFlags + uQWsForMaterialDrawList <= uQWsForDrawLists, "Oh dear - too many materials and/or modules." );

        g_puVisAndClipFlags   = &pxSPR->u32s[0];
        g_ppxMaterialDrawList = reinterpret_cast<const Asura_PS2_Env_RuntimeData_MaterialHdr**>( pxSPR + uQWsForVisAndClipFlags );
    }

    u_int uNumModulesRendered = 0;
#if ASURA_PS2_COLLECT_RENDER_STATS
    u_int uNumNoClipModules   = 0;
#endif
    u_int uNumVisMaterials    = 0;
    bool  bAnyReflective      = false;
    bool  bAnyRainsplash      = false;

    for( u_int uModule = 0; uModule < uNumModules; ++uModule )
    {
        u_int uVisFlags = Asura_Module_System::GetModuleVisFlags( uModule );

        if( dbg_bOnlyDrawSpecifiedModules )
        {
            bool bFound = false;

            const u_int uLen = sizeof(dbg_auListOfModulesToRender) / sizeof(dbg_auListOfModulesToRender[0]);
            for( u_int u = 0; u < uLen; ++u )
            {
                if( dbg_auListOfModulesToRender[u] == uModule ) { bFound = true; break; }
            }

            uVisFlags = ( bFound ? (1 << uNumCams) - 1 : 0 );
        }

        if( uVisFlags )
        {
            ++uNumModulesRendered;

            const Asura_Vector_3& xBSCentre = Asura_Module_System::GetCentreOfModule( uModule );
            const float fBSRadius = Asura_Module_System::GetRadiusOfModule( uModule );

            u_int uClipFlags = 0;

            u_int uClipBit = 1 << 6;
            for( u_int uCam = 0; uCam < uNumCams; ++uCam )
            {
                if( uVisFlags & (1 << uCam) )
                {
                    Asura_Camera::SetCameraIndex( uCam );

                    if( !Asura_PS2_VU1::IsSphereEntirelyWithinGuardbandFrustum(xBSCentre, fBSRadius) )
                    {
                        uClipFlags |= uClipBit;
                    }
                }

                uClipBit >>= 2;
            }

#if ASURA_PS2_COLLECT_RENDER_STATS
            if( !uClipFlags ) { ++uNumNoClipModules; }
#endif

            g_puVisAndClipFlags[uModule] = uVisFlags | (uClipFlags << ASURA_MAX_NO_OF_CAMERAS);
        }
        else
        {
            g_puVisAndClipFlags[uModule] = 0;
        }
    }

    u_int uDualTexturePaletteSize = 0;
    for( Asura_PS2_Env_Store::ConstIt xIt; !xIt.IsFinished(); xIt.Next() )
    {
        const Asura_PS2_Env_RuntimeData_MaterialHdr& xCurMaterialHdr = xIt.GetCurMaterialHdr();
        const Asura_PS2_Env_RuntimeData_ModuleHdr*   pxCurModuleHdr  = xIt.GetModuleHdrsForCurMaterial();
    
        if( dbg_bOnlyDrawSpecifiedMaterials )
        {
            bool bFound = false;
            const u_int uLen = sizeof(dbg_aiListOfMaterialsToRender) / sizeof(dbg_aiListOfMaterialsToRender[0]);
            for( u_int u = 0; u < uLen; ++u )
            {
                if( dbg_aiListOfMaterialsToRender[u] == xCurMaterialHdr.iMaterialID ) { bFound = true; break; }
            }
            if( !bFound ) { continue; }
        }

        // Search for a module using this material that's visible.
        int iNumLeft = xCurMaterialHdr.uNumModules;
        for( ; iNumLeft--; ++pxCurModuleHdr )
        {
            if( pxCurModuleHdr->uChainCALLAddr && g_puVisAndClipFlags[pxCurModuleHdr->iModuleIdx] ) { break; }
        }
        // Skip material if we didn't find one.
        if( -1 == iNumLeft ) { continue; }

        const Asura_Material* const pxMaterial = Asura_Material_System::GetMaterial( xCurMaterialHdr.iMaterialID );

        // We don't bother dealing with ASURA_MATERIAL_NONE, missing
        // materials, etc., instead assuming that the export process will
        // ensure we have valid data.
        Asura_Assert( pxMaterial, "" );

		// JAT 100506 Ignore materials that are marked as don't render
		// Get the flags or if it has an animation, so need to check if it is altering the flags
		u_int uMaterialFlags = pxMaterial->m_uMaterialFlags;
		if ( pxMaterial->m_uAnimHash ) 
		{
			uMaterialFlags = Asura_TextureAnimation_System::GetMaterialFlags( pxMaterial->m_uAnimHash, 0, uMaterialFlags );
		}
		if ( uMaterialFlags & AsuraMaterialFlag_DontRender )
		{
			continue;
		}
		// END - JAT 100506 Ignore materials that are marked as don't render

        // Add to material draw list.
        g_ppxMaterialDrawList[uNumVisMaterials++] = &xCurMaterialHdr;

        // Check for interesting material flags.
        if( uMaterialFlags & AsuraMaterialFlag_Reflective   ) { bAnyReflective = true; }
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
        if( uMaterialFlags & AsuraMaterialFlag_RainSplashes ) { bAnyRainsplash = true; }
#endif

        // Deal with dual texturing case.
        {
            const int iSecondTexID = pxMaterial->m_iTextureID[1];
            if( ASURA_TEXTURE_NONE != iSecondTexID )
            {
                // Obviously we're going to end up with loads of duplicates
                // here for different materials using the same second texture
                // so we need to filter those out.
                bool bDuplicate = false;
                for( u_int u = uDualTexturePaletteSize; u; )
                {
                    --u;
                    if( iSecondTexID == piDualTexturePalette[u] )
                    {
                        bDuplicate = true;
                        break;
                    }
                }

                // Jam it on the end of the list if it's new.
                if( !bDuplicate ) { piDualTexturePalette[uDualTexturePaletteSize++] = iSecondTexID; }
            }
        }
    }

    Asura_Assert( uNumModulesRendered <= uNumModules, "" );
    Asura_Assert( uNumVisMaterials <= uNumMaterials, "" );
    Asura_Assert( uDualTexturePaletteSize <= ASURA_PS2_GSM__MAX_DUAL_TEXTURE_PALETTE_SIZE, "" );

    g_uNumVisibleModules   = uNumModulesRendered;
    g_uNumVisibleMaterials = uNumVisMaterials;

#if ASURA_PS2_COLLECT_RENDER_STATS
    {
        Asura_Assert( uNumNoClipModules <= uNumModulesRendered, "" );

        // Note that this statistic is only particularly interesting for the
        // single camera case.
        float fRatio = static_cast<float>(static_cast<int>(uNumNoClipModules)) / static_cast<float>(static_cast<int>(uNumModulesRendered));
        Asura_PS2_Render::SetNoClipModulesRatio( fRatio );
    }
#endif

    if( !uNumVisMaterials )
    {
        Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(g_puVisAndClipFlags ) );
        Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(piDualTexturePalette) );
        return;
    }

    // Check rendering options and disregard material flags again as
    // necessary.
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_REFLECTION)     ) { bAnyReflective          = false; }
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_RAINSPLASH)     ) { bAnyRainsplash          = false; }
#endif
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_DUAL_TEXTURING) ) { uDualTexturePaletteSize = 0;     }

    u_int uNumTexBfrs;
    if( uDualTexturePaletteSize )
    {
        // We assume that no more than three out of reflective, rainsplash or
        // dual texturing are simultaneously required.
        Asura_Assert( !(bAnyReflective && bAnyRainsplash), "" );

        uNumTexBfrs = ( bAnyReflective || bAnyRainsplash ? 4 : 3 );
    }
    else
    {
        uNumTexBfrs = ( bAnyReflective && bAnyRainsplash ? 4 : 3 );
    }

    ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__ENVIRONMENT_MARKER );

    Asura_PS2_GSMRenderer::Begin();

    // Could use the '2 large' layout when we only require 2 tex bfrs but
    // there's really little point since we can fit our largest supported
    // texture size (512 8-bit with mipmaps) into the '2+1' layout.
    ASURA_PS2_TEXBFRLAYOUT eLayout = ( 4 == uNumTexBfrs ? ASURA_PS2_TEXBFRLAYOUT_4_SMALL : ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL );
    Asura_PS2_TextureManagement::SetBufferLayout( eLayout, true, 0, 1 );

    // The buffers are ordered as:
    //     [base dbl bfr], reflective, rainsplash, dual texture palette
    // Obviously at least one of the latter three will be omitted.
    g_uReflectiveTexBfr = ( bAnyReflective ? 2 : 0 );
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    g_uRainsplashTexBfr = ( bAnyRainsplash ? (bAnyReflective ? 3 : 2) : 0 );
#endif
    if( uDualTexturePaletteSize )
    {
        Asura_PS2_GSMRenderer::SetDualTexturePalette( piDualTexturePalette, uDualTexturePaletteSize, uNumTexBfrs - 1, dbg_uMMAG, dbg_uMMIN );
    }

    // Release temporary palette buffer again.
    Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(piDualTexturePalette) );
}

void Asura_Module_Handler::Platform_Render( void )
{
#if defined(ASURA_PS2_SUPPORT_DYNAMIC_LOADING) && _DEBUG
    g_uNumUnloadedVisModules  = 0;
#endif

	if( !g_uNumVisibleMaterials ) { return; }

    // When any of these material flags are set, we have to do some extra
    // work sorting out renderstate.
    const u_int uMaterialFlagsToCheck = AsuraMaterialFlag_Reflective
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
                                      | AsuraMaterialFlag_FlowingWater
                                      | AsuraMaterialFlag_Waterfall
                                      | AsuraMaterialFlag_RainSplashes
#endif
                                      | AsuraMaterialFlag_TextureAnim;

    const Asura_PS2_Env_RuntimeData_MaterialHdr** ppxNextMat = g_ppxMaterialDrawList;
    for( u_int u = g_uNumVisibleMaterials; u; --u )
    {
        const Asura_PS2_Env_RuntimeData_MaterialHdr* const pxCurMaterialHdr = *ppxNextMat++;

        const int   iMaterialID = pxCurMaterialHdr->iMaterialID;
        const u_int uNumModules = pxCurMaterialHdr->uNumModules;

        const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial( iMaterialID );

        // We don't bother dealing with ASURA_MATERIAL_NONE, missing
        // materials, etc., instead assuming that the export process will
        // ensure we have valid data.
        Asura_Assert( pxMaterial, "" );

        if( dbg_bSkipDualTextureMaterials && ASURA_TEXTURE_NONE != pxMaterial->m_iTextureID[1] ) { continue; }

        Asura_PS2_GSMRenderer::SetTexture( pxMaterial->m_iTextureID[0], dbg_uMMAG, dbg_uMMIN, Asura_PS2_TextureManagement::BFR_AUTO );           

        const u_int uMaterialFlags = pxMaterial->m_uMaterialFlags;
        if( uMaterialFlags & uMaterialFlagsToCheck )
        {
            SetupRenderstateForFlaggedMaterial( iMaterialID, uMaterialFlags );
        }
        else
        {
            Asura_PS2_GSMRenderer::SetRenderstate( uMaterialFlags );
        }

        // Deal with dual texturing case.
        {
            const int iSecondTexID = pxMaterial->m_iTextureID[1];
            if( ASURA_TEXTURE_NONE != iSecondTexID )
            {
                // NB: Don't need to bother checking rendering option again.
                // This function call will just do nothing if it's off since
                // we won't have called SetDualTexturePalette() in the render
                // setup above.
                Asura_PS2_GSMRenderer::SelectFromDualTexturePalette( iSecondTexID );
            }
        }

        u_int uTris = 0;

        const Asura_PS2_Env_RuntimeData_ModuleHdr* pxCurModule = Asura_PS2_Env_Store::GetModuleHdrsForMaterial( *pxCurMaterialHdr );
        for( u_int uNumModulesLeft = uNumModules; uNumModulesLeft--; ++pxCurModule )
        {
            const u_int uChainCALLAddr = pxCurModule->uChainCALLAddr;
            if( !uChainCALLAddr ) { continue; }

            const int iModuleIdx = pxCurModule->iModuleIdx;

            // Setup vis and clip flags.
            {
                const u_int uVisAndClipFlags = g_puVisAndClipFlags[iModuleIdx];
                if( !uVisAndClipFlags ) { continue; }

                const u_int uVisFlags  = uVisAndClipFlags & ((1 << ASURA_MAX_NO_OF_CAMERAS) - 1);
                const u_int uClipFlags = uVisAndClipFlags >> ASURA_MAX_NO_OF_CAMERAS;
                Asura_PS2_GSMRenderer::SetCamFlags( uVisFlags, uClipFlags );
            }

            // Lighting.
            const ModuleLightingCache::Light* pxLights;
            u_int uNumLights;
            ModuleLightingCache::GetLightsForModule( iModuleIdx, &pxLights, &uNumLights );
            Asura_PS2_GSMRenderer::SetPointLightingParametersForPrelitMesh( pxLights, uNumLights );

            Asura_PS2_GSMRenderer::RenderStripChain( uChainCALLAddr );

            uTris += pxCurModule->usNumTris * Asura_PS2_GSMRenderer::GetNumCamsVisFor();
        }

        Asura_PS2_Render::UpdateTriStats( uTris, Asura_PS2_GSMRenderer::GetNumPassesPerTri(), Asura_PS2_Render::TRICOUNT_ENV );
    }
}

void Asura_Module_Handler::Platform_EndRender( void )
{
	if( g_uNumVisibleMaterials ) 
	{ 
        Asura_PS2_GSMRenderer::End();

#if defined(ASURA_PS2_SUPPORT_DYNAMIC_LOADING) && _DEBUG
        if( g_uNumUnloadedVisModules )
        {
            // (MP - 14.02.2003) NB: This count may be less than the true number
            // of modules that ought to be loaded because they will only be
            // counted if the texture(s) they use have some other modules that
            // are both visible and loaded.
            Asura_OutputToDebugger_Retail(( "WARNING: >= %3u modules visible but not loaded. (Frame %u.)\n", g_uNumUnloadedVisModules, Asura_Timers::GetFrameNumber() ));
        }
#endif

        ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__ENVIRONMENT_MARKER );

        Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(g_puVisAndClipFlags) );
    }
}

static void SetupRenderstateForFlaggedMaterial( int iMaterialID, u_int uMaterialFlags )
{
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    bool bRainsplash = false;
#endif

    if( uMaterialFlags & AsuraMaterialFlag_Reflective )
    {
        const u_int uBfr = g_uReflectiveTexBfr;
        if( uBfr )
        {
            Asura_PS2_GSMRenderer::SetTexture( Asura_Render::GetReflectanceMap(), dbg_uMMAG, dbg_uMMIN, uBfr, false );
        }
    }
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    else if( uMaterialFlags & AsuraMaterialFlag_RainSplashes )
    {
        const u_int uBfr = g_uRainsplashTexBfr;
        if( uBfr )
        {
            Asura_PS2_GSMRenderer::SetTexture( g_iRainsplashTexID, dbg_uMMAG, dbg_uMMIN, uBfr, false );
            bRainsplash = true;
        }
    }
#endif

    Asura_PS2_GSMRenderer::SetRenderstate( uMaterialFlags );
    
    if( uMaterialFlags & AsuraMaterialFlag_TextureAnim )
    {
        Asura_TextureAnimationOutput xTexAnimState;
        Asura_Material_System::GetTextureAnimationState( iMaterialID, NULL, &xTexAnimState );
        Asura_PS2_GSMRenderer::SetupPass0STManipulations( true, 1.0f, 1.0f, xTexAnimState.m_fUOffset, xTexAnimState.m_fVOffset );
    }
#ifdef ASURA_PS2_SUPPORT_OLD_WATER_MATERIALS
    else if( bRainsplash )
    {
        const float fMul = ( Asura_Timers::GetGameTimeScale() > 0.0f ? 0.5f / static_cast<float>(INT_MAX) : 0.0f );
        Asura_PS2_GSMRenderer::SetupPass2STManipulations( true,
														  1.75f, 1.75f, fMul * Asura_Random::GetIntU(), fMul * Asura_Random::GetIntU(),
                                                          SCE_GS_ALPHA_CD, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_ZERO, 0 );
    }
    else if( uMaterialFlags & (AsuraMaterialFlag_FlowingWater | AsuraMaterialFlag_Waterfall) )
    {
        if( dbg_bEnableWaterEffect )
        {
            float fTransS, fTransT;
            float fC0, fC1;

            EnvWaterControl::GetParams( uMaterialFlags, &fTransS, &fTransT, &fC0, &fC1 );

            Asura_PS2_GSMRenderer::SetupPass0STManipulations( true, 1.0f, 1.0f, fTransS, fTransT );

            Asura_PS2_GSMRenderer::SetupYManipulations( true,
                                                        2.0f * Asura_Maths::Pi, 0.0f, 0.0f, 8.0f * Asura_Maths::Pi,
                                                        fC0, fC1, 1.0f, 0.0f );
        }
    }
#endif
}
