//
// Asura_PS2_ShadowBuffer.h
//
// Created by RCD - 23 December 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

// based initially on Richard Fines XBox version

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PS2_SHADOWBUFFER_H__
#define __ASURA_PS2_SHADOWBUFFER_H__

#include "Animation/Asura_Animation.h"
#include "Rendering/Asura_PS2_Shadows.h"
#include "Mesh/Asura_PS2_Env_Store.h"

#ifndef ASURA_PS2_SHADOW_MAX_BUFFERS
	#define ASURA_PS2_SHADOW_MAX_BUFFERS (16)
#endif

#ifndef ASURA_PS2_SHADOW_MAX_PROJECTED_BUFFERS
	#define ASURA_PS2_SHADOW_MAX_PROJECTED_BUFFERS (8)
#endif

#if ASURA_PS2_SHADOW_MAX_PROJECTED_BUFFERS > ASURA_PS2_SHADOW_MAX_BUFFERS
    #error ASURA_PS2_SHADOW_MAX_PROJECTED_BUFFERS must be no greater than ASURA_PS2_SHADOW_MAX_BUFFERS
#endif

class Asura_PS2_ShadowBufferSystem;

/******************************************************************************************
*
* SHADOW BUFFER OBJECT																	  
* --------------------																	  
*
* Shadow buffer objects encapsulate shadow buffers, exposing methods for both compositing
* shadows and for rendering them.
*
* Create a shadow buffer object with new. The width and height passed to the constructor
* determine the resolution of the buffer - they do not affect the area of the scene recorded
* by the buffer, which is set using SetShadowBox. 
* The buffers are actually embedded in the Z Buffer, in the top 8 bits, (ie the alpha channel)
* so don't try anything fancy like colours!
*
* Use SetLightPosition to change the 'light POV' from which shadows are calculated. Rendering
* shadows from multiple light positions into the same buffer will not work.
*
* Unlike the XBox the PS2 doesn't have depth textures, and so we don't need to know the "box"
* in light space which will be captured.
*
* To collect shadows for rendering, call BeginRenderingIntoBuffer and render your occluders.
* BeginRenderingIntoBuffer sets up render targets and states for you; 
* When you've rendered all your occluders, call EndRenderingIntoBuffer. It will set everything 
* back to normal.  Note that you can render into a buffer multiple times, and not necessarily 
* consecutively, but you must call BeginRenderingIntoBuffer each time you need to render into
* another buffer.  Most cases for this are already covered in the relevant renderers 
* (Skin, Hierarchy etc)
*
* To render shadows onto the scene, call BeginRenderingFromBuffer. 
*
* Because the PS2 works rather differently to the XBox, we don't attempt to restore the renderstate
* after rendering the shadows - everything else will just have to watch out !
*
* The shadow rendering will be done using the same renderer which normally renders the environment
* (so don't start thinking you can cast shadows onto skins!) with an extra flag in place to tell it
* to use a texture projection matrix.  Because the projection matrix is so complex, we can't get away
* with just using a cut-down version of it (eg just knowing position and radius) and so we will send
* the whole thing in - which has the huge advantage that the same code could actually be used to project
* any texture of any type onto the environment.
*
* The is no current facility for a special large shadow buffer, but it would be very easy to re-organise
* to support this - where possible I've made sure everything is done based on the actual size of the 
* shadow buffer in question, so that different sized buffers won't break anything.
* 
*
*****************************************************************************************/
class Asura_PS2_ShadowBuffer
{
	friend class Asura_PS2_ShadowBufferSystem;
	friend class Asura_PS2_Shadows;
	
protected:
	u_int m_uIndex;	// save time searching later !
	u_int m_uBufferWidth;
	u_int m_uBufferHeight;

	u_int m_uBufferBasePointer;
	
	sceVu0FMATRIX m_xWorldToShadowMatrix, m_xShadowProjectionMatrix;

	bool m_bIsCurrentlyRenderingToBuffer;
	bool m_bForceFallbackRender;

//	sceVu0FMATRIX m_xShadowViewportTransform;
//	sceVu0FMATRIX m_xShadowTransform;

	bool m_bShadowTransformIsDirty;
	void RecalcShadowTransform();

	Asura_Bounding_Box m_xWorldSpaceShadowBox;

	Asura_Vector_3 m_xLightPosition;
	Asura_Vector_3 m_xLightDirection;

	float m_fBaseLevel;
	float m_fDistanceLimit;	// how far the shadow will cast at most FROM THE LIGHT SOURCE
	float m_fBaseWidth;		// half width of shadow pyramid (if perspective) or of shadow box (if ortho)

	float m_fLightLevel;	// allows for brightness of world to effect light

	float m_fBufferScale;

	// Array for storing number of affected modules
	// We'll only track the first AFFECTED_MODULES_LIST_SIZE modules
	enum { AFFECTED_MODULES_LIST_SIZE = 16 };
	u_short m_ausModuleIDs[AFFECTED_MODULES_LIST_SIZE];
	u_int m_uModulesFound;

	const Asura_Animation* m_pxAssociatedAnim;

    u_int m_uCamVisFlagsForCaster;

	void MakeClean();
	void AllocateVRAM(void);

	enum BufferStatus
	{ 
		BUFFER_UNUSED, 
		BUFFER_ALLOCATED,
		BUFFER_CLEAN, 
		BUFFER_DIRTY 
	};
	BufferStatus m_eBufferStatus;
	
public:
	enum { DEFAULT_WIDTH = 128 };
	enum { DEFAULT_HEIGHT = 128 };

	Asura_PS2_ShadowBuffer(u_int uBufferWidth = DEFAULT_WIDTH, u_int uBufferHeight = DEFAULT_HEIGHT);
	~Asura_PS2_ShadowBuffer();

	void BeginRenderingIntoBuffer();
	void EndRenderingIntoBuffer();
	void ReleaseShadowBuffer();

    // Clear buffer and setup context 2 ready for rendering to it. Note
    // that the buffer is assumed to be 32-bit.
	void ClearAndSetupRenderToBfr( u_int uFLUSH );

	// As above but only setup, not clear - needed because we now share buffers, so we have to be able to switch
	void JustSetupRenderToBfr( u_int uFLUSH );

	inline bool IsUnused() const { return m_eBufferStatus == BUFFER_UNUSED; }
	inline bool IsDirty() const { return m_eBufferStatus == BUFFER_DIRTY; }
	inline bool IsForcedFallback() const { return m_bForceFallbackRender; }
	inline void ForceFallback() { m_bForceFallbackRender = true; }

	void SetPositionAndSize(const Asura_Vector_3& xBSCentre, float fBSRadius);
	void SetLightPosition(const Asura_Vector_3& xLightPosition);
    void SetLightLevel( float fLightLevel ) { m_fLightLevel = fLightLevel; }
    float GetLightLevel( void ) const { return m_fLightLevel; }
	void SetShadowBufferQuadHeight(float fQuadHeight);

	void SetPerspectiveEnclosingSphere(const Asura_Vector_3& xSphereCenter, float fSphereRadius, float fDistance);
	void CalcShadowProjMat( float fDistance, const Asura_Vector_3& xLightPos, float fNearLimit, float fFarLimit);

    // Calculates the shadow projection matrix, writing it to the
    // specified location. Can pass NULL for o->w mat if not required.
	// This is just the concatenation of the objects matrix with the shadows ready made projection matrix
    void Proj_GetShadowProjMat( float* pfO2W, float* pfOut );

	const sceVu0FMATRIX *GetShadowSpaceTransform();
	const Asura_Bounding_Box &GetWorldSpaceShadowBox();

	inline const Asura_Vector_3& GetLightPosition( void ) const { return m_xLightPosition; }

	void Proj_GetTEX1andTEX0( u_long* puTEX1, u_long* puTEX0 ) const;

	void FindAffectedModules();

	static void FindAffectedModulesCallback(const u_int uModuleIndex);
	static Asura_PS2_ShadowBuffer* s_pxCurrentShadowBuffer;
	static u_int s_uShadowBufferNextPointer;
	static u_int s_uShadowBufferLimit;
#if _DEBUG
	static bool dbg_bShadowUseAAForRenderToBfr;
#endif

#ifdef _DEBUG
	// Debugging function - shows the volume captured by the shadow buffer
	void DrawShadowBox();
#endif

	// shadow fall-off values for specific object
	float m_fDistanceToLight;
	float m_fObjectRadius;

	// shared fall-off controls across all objects
	static float s_fFalloffEnd;		// distance beyond object at which no shadows cast
	static float s_fFalloffLength;	// distance at which to start fading shadows out
};

class Asura_PS2_ShadowBufferSystem
{
public:
	static void Initialise();
	static void Shutdown();

	static Asura_PS2_ShadowBuffer* FindUnusedShadowBuffer(void);
	static void Asura_PS2_ShadowBufferSystem::PreAllocate(const Asura_Animation* pxAnim, u_int uBufferWidth, float scale);
	static Asura_PS2_ShadowBuffer* GetUnusedShadowBuffer( const Asura_Vector_3& xBSCentre,
                                                          float fBSRadius,
                                                          const Asura_Animation* pxAnim,
                                                          u_int uCamVisFlagsForCaster );
	static Asura_PS2_ShadowBuffer* GetShadowBufferForAnim( const Asura_Animation* pxAnim,
                                                           const Asura_Vector_3& xBSCentre,
                                                           float fBSRadius,
                                                           u_int uCamVisFlagsForCaster );
	static Asura_PS2_ShadowBuffer* GetTemporaryShadowBuffer( const Asura_Vector_3& xBSCentre,
                                                             float fBSRadius,
                                                             u_int uCamVisFlagsForCaster );
	static void Traverse_ModuleData( void );
	static void Destroy_ModuleData( void );

	static void BeginNewFrame();
	static void EndFrame();
	static void FinishedCollectingShadows();
	static void ShowUsedShadowBuffers(bool bAlpha);
	static void AddModuleReference(u_short usModuleIndex, u_int uShadowBufferIndex);

	static inline float GetShadowBufferRadius( u_int uShadowBufferIndex)
	{
		return s_apxShadowBuffers[uShadowBufferIndex]->m_fObjectRadius;
	}

	static inline float GetShadowBufferDistanceToLight( u_int uShadowBufferIndex)
	{
		return s_apxShadowBuffers[uShadowBufferIndex]->m_fDistanceToLight;
	}

	static inline float GetShadowBufferLightLevel( u_int uShadowBufferIndex)
	{
		return s_apxShadowBuffers[uShadowBufferIndex]->GetLightLevel();
	}

	struct ModuleShadowReference
	{
		int iModuleIndex;	// necessary to allow possibility of storing -1 for unused entries
		u_short usShadowBufferFlags;
	};

	static void GetShadowBufferPosition(Asura_PS2_QW *pxOut, u_int uShadowBufferIndex)
	{
		Asura_PS2_ShadowBuffer *pBuffer = s_apxShadowBuffers[uShadowBufferIndex];
		pxOut->f32s[0] = pBuffer->m_xLightPosition.x;
		pxOut->f32s[1] = pBuffer->m_fBaseLevel;
		pxOut->f32s[2] = pBuffer->m_xLightPosition.z;
		pxOut->f32s[3] = pBuffer->m_fObjectRadius;
	};

	static ModuleShadowReference *GetModuleReference( u_short usModuleIndex);
	
	enum
    {
        MAXIMUM_NUMBER_OF_MODULES = ASURA_PS2_SHADOW_MAX_BUFFERS * Asura_PS2_ShadowBuffer::AFFECTED_MODULES_LIST_SIZE,
    };

	static bool s_bBufferHasChanged;

protected:

	enum
	{
		MAX_SHADOWS_IN_ONE_PASS = 5
	};
	static Asura_PS2_ShadowBuffer* s_apxShadowBuffers[ASURA_PS2_SHADOW_MAX_BUFFERS];

	static bool s_bStillCollectingShadowsThisFrame;

	static ModuleShadowReference s_axModuleControls[MAXIMUM_NUMBER_OF_MODULES];

	static const Asura_PS2_Env_RuntimeData_ModuleHdr*** s_paxModuleIndices;
	static int *s_paiModuleChunkCounts;
	static int s_moduleCount;
	static Asura_PS2_ShadowBuffer* s_pxLastUsedShadowBuffer;

};

#if !_DEBUG
	#define dbg_bShadowUseAAForRenderToBfr true
#else
	#define dbg_bShadowUseAAForRenderToBfr Asura_PS2_ShadowBuffer::dbg_bShadowUseAAForRenderToBfr
#endif

#endif//__ASURA_PS2_SHADOWBUFFER_H__