//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_VU1.cpp
//
// By: Morgan Parry.
//
// Copyright (c) 2001 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS2_VU1.h"

// Standard platform-specific.
#include <devvif1.h>
#include <devvu1.h>

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_PlatformBasics.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Render.h"	// for mirror mode

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"

// T Y P E S /////////////////////////////////////////////////////////////////

struct MeshFrustumTestData
{
    float fTanHalfFovX;
    float fOneOverCosHalfFovX;
    float fOneOverCosHalfFovY;
};

// G L O B A L S /////////////////////////////////////////////////////////////

extern u_int VU1_Micro_ParticleRenderer	__attribute__((section(".vudata")));
#ifdef ASURA_USE_SKYSPHERE
extern u_int VU1_FractalSky				__attribute__((section(".vudata")));
#endif
extern u_int VU1_GSM                    __attribute__((section(".vudata")));
extern u_int VU1_Micro_CoronaRenderer	__attribute__((section(".vudata")));
//extern u_int VU1_GlassRender			__attribute__((section(".vudata")));
extern u_int VU1_DecalRenderer			__attribute__((section(".vudata")));
extern u_int VU1_Micro_Beam				__attribute__((section(".vudata")));
extern u_int VU1_SMOOTHSKIN             __attribute__((section(".vudata")));
extern u_int VU1_Micro_BlueToAlpha      __attribute__((section(".vudata")));
#ifdef ASURA_USE_WATER
extern u_int VU1_Water                  __attribute__((section(".vudata")));
#endif
#ifdef ASURA_USE_CARPET
extern u_int VU1_CarpetMesh             __attribute__((section(".vudata")));
#endif
#ifdef ASURA_USE_BOTANICALS
extern u_int VU1_Botanicals             __attribute__((section(".vudata")));
#endif

static const u_int* g_apuVU1MicroPrograms[Asura_PS2_VU1::RENDERER_NUMRENDERERS] =
{
    &VU1_Micro_ParticleRenderer, // RENDERER_PARTICLES
#ifdef ASURA_USE_SKYSPHERE
    &VU1_FractalSky,             // RENDERER_FRACTALSKY
#endif
    &VU1_GSM,                    // RENDERER_GSM
    &VU1_Micro_CoronaRenderer,   // RENDERER_CORONAS
//  &VU1_GlassRender,            // RENDERER_GLASS
    &VU1_DecalRenderer,          // RENDERER_DECAL
    &VU1_Micro_Beam,             // RENDERER_BEAM,
    &VU1_SMOOTHSKIN,             // RENDERER_SMOOTHSKIN
    &VU1_Micro_BlueToAlpha,      // RENDERER_BLUE_TO_ALPHA
#ifdef ASURA_USE_WATER
    &VU1_Water,                  // RENDERER_WATER
#endif
#ifdef ASURA_USE_CARPET
    &VU1_CarpetMesh,             // RENDERER_CARPET
#endif
#ifdef ASURA_USE_BOTANICALS
    &VU1_Botanicals,             // RENDERER_BOTANICALS
#endif
};

// Record of the address from which the last renderer was uploaded. We assume
// these don't change and can therefore use this value for caching purposes.
static u_int g_uCurrentRendererAddr;

// In theory, we should be able to clip precisely to the screen bounds. To be
// paranoid, let's allow a small margin of error to account for numerical
// inaccuracies.
DEBUGVAR float dbg_fScreenClipEpsilon = 1.0f;

DEBUGVAR float dbg_fGuardbandClip = 512.0f;

// Output z value range. This is the range of values that the unit cube of
// clipspace gets mapped into. Generally you'd want this to be full range, to
// use all of the available z buffer bits and improve accuracy, but
// unfortunately the PS2 fog method means we have to fudge the max value.
// (MP - 21.04.2005) The palettised PS2 fog also requires that the minimum
// value used for rendered geometry is high enough to force the middle 8 bits
// of the z-buffer (assuming a total of 24 bits) are never all 0. This 0 value
// is reserved for the skybox and we don't want any geometry to end up using
// that palette entry, since the skybox's fog value may be very different from
// the value used for the far plane. To do this, we need to bias output z
// values by 256. However, since we cheat when doing the FTOI conversion on
// VU1 and end up shifting the z values by 4 bits, we only need to offset by
// 16 in our transform.
DEBUGVAR float dbg_fMinOutputZ = 16.0f; //1.0f;
static float g_fMaxOutputZ;

sceVu0FMATRIX Asura_PS2_VU1::s_xWorldToScreenMat[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xWorldToViewMat[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xWorldToClipMatScreen[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xWorldToClipMatGuardband[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xViewToClipMatScreen[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xViewToClipMatGuardband[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xViewToScreenMat[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FMATRIX Asura_PS2_VU1::s_xClipToScreenMat[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FVECTOR Asura_PS2_VU1::s_xGC2SScale[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FVECTOR Asura_PS2_VU1::s_xGC2STrans[ASURA_MAX_NO_OF_CAMERAS];
sceVu0FVECTOR Asura_PS2_VU1::s_xGC2SCScale[ASURA_MAX_NO_OF_CAMERAS];

static MeshFrustumTestData g_axMeshFrustumTestData[ASURA_MAX_NO_OF_CAMERAS];

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void SetupMeshFrustumTest( int iCamID, float fScreenZ );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU1::BeginFrame().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Sets up matrices.
// AUTHOR       Morgan Parry - 07.11.2001.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU1::BeginFrame()
{
    // MH hack: this is done here in case Asura_PS2_GS::SetViewport hasn't 
    // been called this frame.
	// (MH - 16.04.2002) now this is the only place it gets done - matrices are
	// calculated for each camera and stored.
	SetupMatrices();
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU1::SetupMatrices().
// PARAMETERS   uCamera: Bitfield specifying which camera(s) to setup matrices for.
// RETURNS      None.
// DESCRIPTION  Sets up matrices.
// AUTHOR       Morgan Parry - 07.11.2001.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU1::SetupMatrices( u_int uCameras /*= UINT_MAX*/ )
{
	const int iOldCamID = Asura_Camera::GetCameraIndex();

#if _DEBUG
    // We can't support per-camera near and far z clip on PS2, due to full
    // screen fog and probably other reasons, so let's check that's not 
    // being used.
    Asura_Camera::SetCameraIndex( 0 );
    const float fCam0ClipNearZ = Asura_Camera::GetNearClip();
    const float fCam0ClipFarZ = Asura_Camera::GetFarClip();
#endif

	for( int iCamID = 0; iCamID < Asura_Camera::GetNumberOfCameras(); ++iCamID )
	{
		if ((uCameras & (1 << iCamID)) == 0) { continue; }

		Asura_Camera::SetCameraIndex( iCamID );
        if( !Asura_Camera::IsCameraActive() ) { continue; }

#if _DEBUG
        Asura_Assert( Asura_Camera::GetNearClip() == fCam0ClipNearZ, "Per-camera near or far clip z not supported on PS2" );
        Asura_Assert( Asura_Camera::GetFarClip() == fCam0ClipFarZ, "Per-camera near or far clip z not supported on PS2" );
#endif

        {
            // This is just the way the PS2 library and sample code deals
            // with FOV. It achieves exactly the same thing, since the
            // position of the projection plane is dictated by the size of
            // the viewport and the desired FOV.
		    const float fScreenZ = static_cast<float>(Asura_PS2_GS::GetViewportWidth(iCamID) >> 1) * Asura_Camera::GetOneOverTanHalfFovX();

		    SetupViewClipAndViewScreenMatrices( iCamID, fScreenZ );
            SetupMeshFrustumTest( iCamID, fScreenZ );
        }

		// Setup world->view matrix.
        SetupPS2MatFromAsuraData( &s_xWorldToViewMat[iCamID][0][0], *Asura_Camera::GetOrientation(), *Asura_Camera::GetPosition() );
		sceVu0InversMatrix( s_xWorldToViewMat[iCamID], s_xWorldToViewMat[iCamID] );

		sceVu0MulMatrix( s_xWorldToClipMatScreen[iCamID],    s_xViewToClipMatScreen[iCamID],    s_xWorldToViewMat[iCamID] );
		sceVu0MulMatrix( s_xWorldToClipMatGuardband[iCamID], s_xViewToClipMatGuardband[iCamID], s_xWorldToViewMat[iCamID] );
		sceVu0MulMatrix( s_xWorldToScreenMat[iCamID],        s_xViewToScreenMat[iCamID],        s_xWorldToViewMat[iCamID] );
	}
    // End loop through cameras.

	Asura_Camera::SetCameraIndex( iOldCamID );

    // (MP - 26.06.2002) This is actually necessary. If any rendering process
    // uses the matrices soon after this function has completed, the cache
    // may not have been written back to main memory yet. For example, I was
    // bitten by this bug when working on the skybox renderer (which is the
    // first thing drawn after BeginFrame() in the main game loop.)
    ( QueryIntrContext() ? iFlushCache(WRITEBACK_DCACHE) : FlushCache(WRITEBACK_DCACHE) );
}
// End function Asura_PS2_VU1::SetupMatrices().

bool Asura_PS2_VU1::IsSphereEntirelyWithinGuardbandFrustum( const Asura_Vector_3& xCentre, float fRadius )
{
    sceVu0FVECTOR xVSCentre = { xCentre.x, xCentre.y, xCentre.z, 1.0f };
    sceVu0ApplyMatrix( xVSCentre, Asura_PS2_VU1::GetWorldToViewMat(), xVSCentre );

//	if (viewSpacePos.z+radius>=0)// return true;
//	{
//		float r = radius * Camera->m_fOneOverCosHalfFovX;
//		float z = viewSpacePos.z * Camera->m_fTanHalfFovX;
//	
//
//		if ((viewSpacePos.x-z)<=r && (-viewSpacePos.x-z)<=r)
//		{
//			z *= Camera->m_fYOverXAspectRatio;
//		
//			r = radius * Camera->m_fOneOverCosHalfFovY;
//
//			if ((viewSpacePos.y-z)<=r && (-viewSpacePos.y-z)<=r)
//			{
//				return true;
//			}
//		}
//	}
    
    const float fNearClipZ = Asura_Camera::GetNearClip();
    const float fFarClipZ  = Asura_Camera::GetFarClip();

    if( xVSCentre[2] - fRadius < fNearClipZ || xVSCentre[2] + fRadius >= fFarClipZ ) { return false; }

    const MeshFrustumTestData& xTestData = g_axMeshFrustumTestData[Asura_Camera::GetCameraIndex()];

    float fR = fRadius * xTestData.fOneOverCosHalfFovX;
    float fZ = xVSCentre[2] * xTestData.fTanHalfFovX;

    if( fZ - xVSCentre[0] < fR || fZ + xVSCentre[0] < fR ) { return false; }

    fR = fRadius * xTestData.fOneOverCosHalfFovY;
    fZ *= Asura_Camera::GetYOverXAspectRatio();

    if( fZ - xVSCentre[1] < fR || fZ + xVSCentre[1] < fR ) { return false; }

    return true;
}

void Asura_PS2_VU1::EndFrame( void )
{
    // Reset VU renderer state cache.
    g_uCurrentRendererAddr = 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU1::Initialise().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Sets up VU1 stuff.
// AUTHOR       Morgan Parry - 07.11.2001.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU1::Initialise( void )
{
	// JAT 11/07/06 check the version number of the VU1 code library to make sure
	// compiled and running with the latest version
#ifndef ASURA_NO_DEBUG_OUTPUT
    {
        u_int uRequiredVersion, uActualVersion;
        Project_GetVULibVersions( &uRequiredVersion, &uActualVersion );

        if( uActualVersion != uRequiredVersion )
        {	    
            while( true )
            {
	            Asura_Assert( false, "BAD VU LIB VERSION" );
                Asura_OutputToDebugger_Retail(( "BAD VU LIB VERSION\n" ));
            }
        }

        Asura_OutputToDebugger(( "VU lib version: 0x%04X\n", uActualVersion ));
    }
#endif

    // Reset.
	sceDevVif1Reset();
	sceDevVu1Reset();

    // setup VU debugging
    sceDevVu1PutDBit(1);

    SetZMaxToDefault();

    // Reset VU renderer state cache.
    g_uCurrentRendererAddr = 0;
}
// End function Asura_PS2_VU1::Initialise().

bool Asura_PS2_VU1::LoadRenderer( RENDERER eRenderer )
{
	Asura_Assert( eRenderer < RENDERER_NUMRENDERERS, "" );
    return LoadCustomRenderer( g_apuVU1MicroPrograms[eRenderer] );
}

bool Asura_PS2_VU1::LoadCustomRenderer( const void* pvRenderer )
{
    const u_int uAddr = reinterpret_cast<u_int>( pvRenderer );

    // Early out if already uploaded.
    if( uAddr == g_uCurrentRendererAddr ) { return false; }
    
#if _DEBUG
    // Verify the size of the upload. The chain should just be a large RET
    // tag with a whole bunch of MPG VIF codes. 8 full MPGs will fill VU1's
    // instruction memory. This is 16Kb of data, plus 4 QWs of overhead for
    // the MPGs and some NOPs, and therefore the RET tag's QWC should be
    // <= 1028. Note that we read in the tag using an uncached ptr, since
    // we want to avoid dragging any of this stuff into the cache.
    const Asura_PS2_DmaTag xRETTag = *Asura_PS2_UncachedMem(reinterpret_cast<const Asura_PS2_DmaTag*>(uAddr));
    Asura_Assert( xRETTag.m_eID == Asura_PS2_DmaTag::ID_RET, "" );
    Asura_Assert( xRETTag.m_uQWC <= 1028, "Oh dear, VU1 is about to explode." );
#endif

	Asura_PS2_QW* pxOut = Asura_PS2_DeferredRendering::GetVifBuffer( 1 );
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut))->SetWithNOPs( 0,
                                                               Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                               Asura_PS2_DmaTag::ID_CALL,
                                                               false,
                                                               uAddr,
                                                               false );

    g_uCurrentRendererAddr = uAddr;
	return true;
}

bool Asura_PS2_VU1::IsRendererLoaded( RENDERER eRenderer )
{
	Asura_Assert( eRenderer < RENDERER_NUMRENDERERS, "" );
    return IsCustomRendererLoaded( g_apuVU1MicroPrograms[eRenderer] );
}

bool Asura_PS2_VU1::IsCustomRendererLoaded( const void* pvRenderer )
{
    return( reinterpret_cast<u_int>(pvRenderer) == g_uCurrentRendererAddr );
}

void Asura_PS2_VU1::InvalidateCachedRenderer()
{
	g_uCurrentRendererAddr = 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PS2_VU1::Shutdown().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  
// AUTHOR       Morgan Parry - 07.11.2001.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU1::Shutdown( void )
{
}
// End function Asura_PS2_VU1::Shutdown().

DEBUGVAR float dbg_fProjMatrixViewportScale = 1.0f;

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     SetupViewClipAndViewScreenMatrices().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Creates view->clip and view->screen matrices.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_PS2_VU1::SetupViewClipAndViewScreenMatrices( int iCamID, float fScreenZ )
{
    // NB: The CalculateZMaxReqd() and TransformZBfrValueToViewspaceZ()
    // methods down below will need updating if our projection matrix changes.

    const float fNearClipZ = Asura_Camera::GetNearClip();
    const float fFarClipZ  = Asura_Camera::GetFarClip();

	float fAspectX = 1.0f;
	float fAspectY = Asura_Camera::GetPhysicalScreenAspectRatio() * Asura_PS2_GS::GetScreenHeight() / Asura_PS2_GS::GetScreenWidth();
    float fAspectZ = fFarClipZ * fNearClipZ * (-dbg_fMinOutputZ + g_fMaxOutputZ) / (-fNearClipZ + fFarClipZ);

    if( Asura_Render::ShouldRenderMirrored() ) { fAspectX = -fAspectX; }

	const Asura_Vector_2& xOffset = Asura_Render::GetSubPixelOffset();
    float fCentreX = ASURA_PS2_SCREEN_CENTRE_X + xOffset.x;
    float fCentreY = ASURA_PS2_SCREEN_CENTRE_Y + xOffset.y;
    float fCentreZ = (-g_fMaxOutputZ * fNearClipZ + dbg_fMinOutputZ * fFarClipZ) / (-fNearClipZ + fFarClipZ);

    float fScreenClipX = ((Asura_PS2_GS::GetViewportWidth(iCamID)  >> 1) + dbg_fScreenClipEpsilon) / fAspectX;
    float fScreenClipY = ((Asura_PS2_GS::GetViewportHeight(iCamID) >> 1) + dbg_fScreenClipEpsilon) / fAspectY;

    const Asura_ViewingArea* pxViewingArea = Asura_Camera::GetViewingArea();
    float fGBClipX = (dbg_fGuardbandClip * pxViewingArea->m_fWidth ) / fAspectX;
    float fGBClipY = (dbg_fGuardbandClip * pxViewingArea->m_fHeight) / fAspectY;

    if( dbg_fProjMatrixViewportScale != 1.0f )
    {
        fScreenClipX *= dbg_fProjMatrixViewportScale;
        fGBClipX     *= dbg_fProjMatrixViewportScale;
        fScreenClipY *= dbg_fProjMatrixViewportScale;
        fGBClipY     *= dbg_fProjMatrixViewportScale;
    }

    //                      | ax  0  0 cx | | scrz   0  0 0 |   | ax . scrz      0       cx    0 |
    // s_xViewToScreenMat = |  0 ay  0 cy | |   0  scrz 0 0 | = |     0      ay . scrz   cy    0 |
    //                      |  0  0 az cz | |   0    0  0 1 |   |     0          0       cz   az |
    //                      |  0  0  0	1 | |   0    0  1 0 |   |     0          0        1    0 |

	sceVu0UnitMatrix( s_xViewToScreenMat[iCamID] ); 
	s_xViewToScreenMat[iCamID][0][0] = fAspectX * fScreenZ;
    s_xViewToScreenMat[iCamID][1][1] = fAspectY * fScreenZ;
	s_xViewToScreenMat[iCamID][2][0] = fCentreX;
	s_xViewToScreenMat[iCamID][2][1] = fCentreY  ;
	s_xViewToScreenMat[iCamID][2][2] = fCentreZ;
	s_xViewToScreenMat[iCamID][2][3] = 1.0f;
    s_xViewToScreenMat[iCamID][3][2] = fAspectZ;
    s_xViewToScreenMat[iCamID][3][3] = 0.0f;

    //                    | 2n/2w    0	       0           0      |
    // s_xViewToClipMat = |   0	   2n/2h       0           0      |
    //                    |   0      0    (f+n)/(f-n) -2f*n/(f-n) |
    //                    |   0	   	 0         1           0      |
    //
    // (This is the equivalent of the D3DTS_PROJECTION matrix in Direct3D or
    // GL_PROJECTION_MATRIX in OpenGL. In fact, it's derived from the
    // standard perspective matrix setup by the latter API, with some changes
    // of sign, support for FOV and provision for the near clip plane being
    // distinct from the projection plane.)

	sceVu0UnitMatrix( s_xViewToClipMatScreen[iCamID] );
	s_xViewToClipMatScreen[iCamID][0][0] = fScreenZ / fScreenClipX;
	s_xViewToClipMatScreen[iCamID][1][1] = fScreenZ / fScreenClipY;
	s_xViewToClipMatScreen[iCamID][2][2] = (fFarClipZ + fNearClipZ) / (fFarClipZ - fNearClipZ);
	s_xViewToClipMatScreen[iCamID][3][2] = -2.0f * fFarClipZ * fNearClipZ / (fFarClipZ - fNearClipZ);
	s_xViewToClipMatScreen[iCamID][2][3] = 1.0f;
	s_xViewToClipMatScreen[iCamID][3][3] = 0.0f;

	sceVu0UnitMatrix( s_xViewToClipMatGuardband[iCamID] );
	s_xViewToClipMatGuardband[iCamID][0][0] = fScreenZ / fGBClipX;
	s_xViewToClipMatGuardband[iCamID][1][1] = fScreenZ / fGBClipY;
	s_xViewToClipMatGuardband[iCamID][2][2] = (fFarClipZ + fNearClipZ) / (fFarClipZ - fNearClipZ);
	s_xViewToClipMatGuardband[iCamID][3][2] = -2.0f * fFarClipZ * fNearClipZ / (fFarClipZ - fNearClipZ);
	s_xViewToClipMatGuardband[iCamID][2][3] = 1.0f;
	s_xViewToClipMatGuardband[iCamID][3][3] = 0.0f;

	//                      | ax*w    0          0              cx      |
	// s_xClipToScreenMat = |  0     ay*h        0              cy      |
	//                      |  0      0    (-zmax+zmin)/2 (zmin+zmax)/2 |
	//                      |  0      0          0               1      |
    //
    // (Here we're just after:
    //
    //      s_xClipToScreenMat * s_xViewToClipMat == s_xViewToScreenMat
    //
    //  Of course, floating point inaccuracies mean we probably won't get
    //  exactly that.)

	sceVu0UnitMatrix( s_xClipToScreenMat[iCamID] );
	s_xClipToScreenMat[iCamID][0][0] = fAspectX * fScreenClipX;
	s_xClipToScreenMat[iCamID][1][1] = fAspectY * fScreenClipY;
	s_xClipToScreenMat[iCamID][2][2] = (-g_fMaxOutputZ + dbg_fMinOutputZ) / 2.0f;
	s_xClipToScreenMat[iCamID][3][0] = fCentreX;	
    s_xClipToScreenMat[iCamID][3][1] = fCentreY;	
	s_xClipToScreenMat[iCamID][3][2] = (g_fMaxOutputZ + dbg_fMinOutputZ) / 2.0f;
	s_xClipToScreenMat[iCamID][3][3] = 1.0f;

    // Rather than using separate W->C, W->S and C->S matrices, we can just
    // scale and translate by these vectors.
    {
        s_xGC2SScale[iCamID][0] = fAspectX * fGBClipX;
        s_xGC2SScale[iCamID][1] = fAspectY * fGBClipY;
        s_xGC2SScale[iCamID][2] = (-g_fMaxOutputZ + dbg_fMinOutputZ) / 2.0f;
        s_xGC2SScale[iCamID][3] = 1.0f;

        // GC->S translation is same as the one for SC->S so just copy from
        // the SC->S matrix above.
        for( u_int u = 0; u < 3; ++u ) { s_xGC2STrans[iCamID][u] = s_xClipToScreenMat[iCamID][3][u]; }
        s_xGC2STrans[iCamID][3] = 0.0f;
    }
    
    // Finally, to save uploading both world->guardbandclip and
    // world->screenclip matrices, we can just scale between them.
    s_xGC2SCScale[iCamID][0] = fGBClipX / fScreenClipX;
    s_xGC2SCScale[iCamID][1] = fGBClipY / fScreenClipY;
    s_xGC2SCScale[iCamID][2] = 1.0f;
    s_xGC2SCScale[iCamID][3] = 1.0f;
}
// End function SetupViewClipAndViewScreenMatrices().

void Asura_PS2_VU1::SetupPS2MatFromAsuraData( float* pfOut, const Asura_Matrix_3x3& xOri, const Asura_Vector_3& xPos )
{
    for( u_int u = 0; u < 3; ++u )
    {
        for( u_int v = 0; v < 3; ++v ) { *pfOut++ = xOri.Element[v][u]; }
        *pfOut++ = 0.0f;
    }
    
    for( unsigned u = 0; u < 3; ++u ) { *pfOut++ = xPos.GetElement(u); }
    *pfOut++ = 1.0f;
}

bool Asura_PS2_VU1::IsVtxInsideGuardband( u_int uCam, const Asura_Vector_3& xVtx )
{
    sceVu0FVECTOR xClipVtx = { xVtx.x, xVtx.y, xVtx.z, 1.0f };
    sceVu0ApplyMatrix( xClipVtx, GetWorldToClipMatGuardband(uCam), xClipVtx );

    const float fAbsW = Asura_Maths::Abs( xClipVtx[3] );
    for( u_int u = 0; u < 3; ++u )
    {
        if( xClipVtx[u] > fAbsW || xClipVtx[u] < -fAbsW ) { return false; }
    }

    return true;
}

void Asura_PS2_VU1::TransformVtxToScreenspace( u_int uCam, const Asura_Vector_3& xVtxIn, Asura_PS2_QW* pxQWOut, float* pfOOW )
{
    Asura_Vector_3 xTmp;
    TransformVtxToScreenspace( uCam, xVtxIn, &xTmp, pfOOW );

    pxQWOut->u32s[0] = Asura_Maths::FloatToInt_Near( xTmp.x * 16.0f );
    pxQWOut->u32s[1] = Asura_Maths::FloatToInt_Near( xTmp.y * 16.0f );
    pxQWOut->u32s[2] = Asura_Maths::FloatToInt_Near( xTmp.z * 16.0f ); // Scale to be consistent with VU code.
    pxQWOut->u32s[3] = 0;
}

void Asura_PS2_VU1::TransformVtxToScreenspace( u_int uCam, const Asura_Vector_3& xVtxIn, Asura_Vector_3* pxVtxOut, float* pfOOW )
{
    sceVu0FVECTOR xVtxIn4 = { xVtxIn.x, xVtxIn.y, xVtxIn.z, 1.0f };

    sceVu0FVECTOR xVtxOut4;
    sceVu0ApplyMatrix( xVtxOut4, GetWorldToScreenMat(uCam), xVtxIn4 );

    float fOOW = 1.0f / xVtxOut4[3];
    for( u_int u = 0; u < 3; ++u ) { (*pxVtxOut)[u] = xVtxOut4[u] * fOOW; }

    if( pfOOW ) { *pfOOW = fOOW; }
}

float Asura_PS2_VU1::GetZMax( void )
{
    return g_fMaxOutputZ;
}

void Asura_PS2_VU1::SetZMax( float fZMax )
{
    g_fMaxOutputZ = fZMax;
}

void Asura_PS2_VU1::SetZMaxToDefault( void )
{
    float fZMax;

    // We want to use the full available range, ie. (2^b)-1, by default.
    // However, due to the FTOI4 we do on VU1, instead of an FTOI0, we need
    // to divide the max value by 4.
    switch( Asura_PS2_GS::GetZBfrBitDepth() )
    {
        default: Asura_Assert( false, "" ); // ... and fall through to a valid case.

        case 16: fZMax = 16383.0f;      break;
        case 24: fZMax = 4194303.0f;    break;
        case 32: fZMax = 1073741760.0f; break; // Can't use 1073741823 due to floating point inaccuracy.
    }

    SetZMax( fZMax );
}

float Asura_PS2_VU1::CalculateZMaxReqd( float fViewspaceZ, float fZBfrValue )
{
    // Given the transform parameters other than z max and fixed viewspace z
    // and z buffer values, the z max required is
    //
    //           b * (-n + f)             f * n * min
    //           ------------ - min * f + -----------
    //                16                       v
    //
    //     max = ------------------------------------
    //
    //                       f * n
    //                       ----- - n
    //                         v
    //
    //     where b = z bfr value, n = z near, f = z far,
    //           min = z min, v = viewspace z
    //
    // This equation is derived from the projection matrix setup in the
    // SetupViewClipAndViewScreenMatrices() method and the perspective divide
    // done on VU1, which sets
    //
    //                   az
    //     b = 16 ( cz + -- )
    //                    v
    //
    //     where az = aspect z, cz = centre z
    //
    // Note that the factor of 16 comes from us doing a FTOI4 instead of a
    // FTOI0 on the output z value on VU1.
    //
    // Obviously this will all break horribly if the projection matrix
    // changes.

    if( !fViewspaceZ ) { return 0.0f; }
    
    // NB: Current camera index doesn't matter much for these lookups since
    // we assume that all of them have the same clip bounds. See
    // SetupMatrices().
    const float fNearClipZ = Asura_Camera::GetNearClip();
    const float fFarClipZ  = Asura_Camera::GetFarClip();

    float fFNOverV = (fFarClipZ * fNearClipZ) / fViewspaceZ;

    float f0 = (fZBfrValue * (-fNearClipZ + fFarClipZ)) / 16.0f;
    float f1 = dbg_fMinOutputZ * fFarClipZ;
    float f2 = fFNOverV * dbg_fMinOutputZ;
    float f3 = fFNOverV - fNearClipZ;

    return( (f0 - f1 + f2) / f3 );
}

float Asura_PS2_VU1::TransformZBfrValueToViewspaceZ( float fZBfrValue )
{
    // As noted in CalculateZMaxReqd() above,
    //
    //                   az
    //     b = 16 ( cz + -- )
    //                    v
    //
    // and so obviously
    //
    //            az
    //
    //     v = ---------
    //
    //           b
    //          -- - cz
    //          16

    // NB: Current camera index doesn't matter much for these lookups since
    // we assume that all of them have the same clip bounds. See
    // SetupMatrices().
    const float fNearClipZ = Asura_Camera::GetNearClip();
    const float fFarClipZ  = Asura_Camera::GetFarClip();

    float fAZ = fFarClipZ * fNearClipZ * (-dbg_fMinOutputZ + g_fMaxOutputZ) / (-fNearClipZ + fFarClipZ);
    float fCZ = (-g_fMaxOutputZ * fNearClipZ + dbg_fMinOutputZ * fFarClipZ) / (-fNearClipZ + fFarClipZ);

    return( fAZ / ((fZBfrValue / 16.0f) - fCZ) );
}

float Asura_PS2_VU1::TransformViewspaceZToZBfrValue( float fViewspaceZ )
{
    // As noted in CalculateZMaxReqd() above,
    //
    //                   az
    //     b = 16 ( cz + -- )
    //                    v
    //

    // NB: Current camera index doesn't matter much for these lookups since
    // we assume that all of them have the same clip bounds. See
    // SetupMatrices().
    const float fNearClipZ = Asura_Camera::GetNearClip();
    const float fFarClipZ  = Asura_Camera::GetFarClip();

    float fAspectZ = fFarClipZ * fNearClipZ * (-dbg_fMinOutputZ + g_fMaxOutputZ) / (-fNearClipZ + fFarClipZ);
    float fCentreZ = (-g_fMaxOutputZ * fNearClipZ + dbg_fMinOutputZ * fFarClipZ) / (-fNearClipZ + fFarClipZ);

	return 16.0f * (fCentreZ + (fAspectZ / fViewspaceZ)); 
}

void Asura_PS2_VU1::FillMem( u_long128 uFillValue )
{
    const u_int uNumQWs = 4;

    Asura_PS2_QW* const pxVIFBfr = Asura_PS2_DeferredRendering::GetVifBuffer( uNumQWs );
    Asura_PS2_QW* pxOut = pxVIFBfr;

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->SetWithNOPs( uNumQWs - 1, Asura_PS2_DmaTag::PCE_NOCHANGE, Asura_PS2_DmaTag::ID_CNT, false, 0, false );

    // Filling write, 4 QWs at a time, each QW actually being a copy of the
    // STROW register. This needs to be done in 4 steps, since UNPACKs are
    // limited to 256 QWs.

    pxOut->u32s[0] = SCE_VIF1_SET_STCYCL( 4, 0, false );

    pxOut->u32s[1] = SCE_VIF1_SET_STMASK( false );
    pxOut->u32s[2] = SCE_VIF1_SET_STMASK_REG( 1, 1, 1, 1,
                                        1, 1, 1, 1,
                                        1, 1, 1, 1,
                                        1, 1, 1, 1 );
    
    pxOut->u32s[3] = SCE_VIF1_SET_STROW( false );
    ++pxOut;
    pxOut->u128 = uFillValue;
    ++pxOut;

    for( u_int u = 0; u < 4; ++u )
    {
        // Recall that NUM=0 is interpreted as NUM=256.
        pxOut->u32s[u] = SCE_VIF1_SET_UNPACK( u << 8, 0, Asura_PS2_VU1::VIFPK_V4_32, false );
    }
    ++pxOut;

    Asura_Assert( static_cast<u_int>(pxOut - pxVIFBfr) == uNumQWs, "" );
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void SetupMeshFrustumTest( int iCamID, float fScreenZ )
{
    g_axMeshFrustumTestData[iCamID].fTanHalfFovX = dbg_fGuardbandClip / fScreenZ;

    float fHalfFovX = Asura_Maths::ArcTangent( dbg_fGuardbandClip, fScreenZ );
    g_axMeshFrustumTestData[iCamID].fOneOverCosHalfFovX = 1.0f / Asura_Maths::Cosine( fHalfFovX );

    float fHalfFovY = Asura_Maths::ArcTangent( g_axMeshFrustumTestData[iCamID].fTanHalfFovX, Asura_Camera::GetXOverYAspectRatio() );
    g_axMeshFrustumTestData[iCamID].fOneOverCosHalfFovY = 1.0f / Asura_Maths::Cosine( fHalfFovY );
}

