//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_WaterRenderer.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_WATER

// This file's header.
#include "Rendering/Asura_PS2_WaterRenderer.h"

// Stdlib platform-specific.
#include <eestruct.h>

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "SpecialFX/Asura_Weather_System.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_VU1.h"
#include "Rendering/Asura_PS2_WaterDefines.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// The following is rather ugly but we do have a lot of state and it's quite
// a lot faster to keep it in SPR. Note that some relatively strong
// assumptions are made in the memory layout about buffer sizes, etc. This
// also needs to be kept in sync with the constants in Asura_PS2_GSMDefines.h.

// General.
#define g_uMeshFlags         (*reinterpret_cast<u_int*         >(0x70000000))
#define g_uPrevMeshFlags     (*reinterpret_cast<u_int*         >(0x70000004)) // Mesh flags from previous WaterSendPendingData().
#define g_uPendingFlags      (*reinterpret_cast<u_int*         >(0x70000008))
#define g_uPendingVtxCount   (*reinterpret_cast<u_int*         >(0x7000000C))
// Camera.
#define g_uCamFlagsForVU     (*reinterpret_cast<u_int*         >(0x70000010))
#define g_bO2WIsIdentity     (*reinterpret_cast<bool*          >(0x70000014))
#define g_uCurVUCamBfr       (*reinterpret_cast<u_int*         >(0x70000018))
// Texture.
#define g_bTexUploaded       (*reinterpret_cast<bool*          >(0x7000001C))
#define g_aiCurTex           ( reinterpret_cast<int*           >(0x70000020))	// array of texture IDs - up to 3, for multiple passes
// Sun dir
#define g_pxSunDir           ( reinterpret_cast<Asura_Vector_3*>(0x70000030))
// Renderstate
#define g_uCurVURSBfr        (*reinterpret_cast<u_int*         >(0x7000003C))
#define g_uNumPassesPerTri   (*reinterpret_cast<u_int*         >(0x70000040))
#define g_pxRSBfr            ( reinterpret_cast<Asura_PS2_QW*  >(0x70000050)) // Up to 12 QWs of renderstate.
#define g_pxYParams          ( reinterpret_cast<Asura_PS2_QW*  >(0x70000110)) // currently 8 QW reserved
// Matrices.
#define g_pxO2W              ( reinterpret_cast<Asura_PS2_QW*  >(0x70000190)) // 4x4 matrix.
#define g_pxMatTempBfr       ( reinterpret_cast<Asura_PS2_QW*  >(0x700001D0)) //     "
#define g_pxTexMatBfr        ( reinterpret_cast<Asura_PS2_QW*  >(0x70000210)) //     "
// Total.
#define TOTAL_SPR_QWS_REQD                                      (0x0000025 )

// Compile-time checks against stuff in the header.
#if TOTAL_SPR_QWS_REQD > ASURA_PS2_WATER_SPR_QWS_RESERVED
    #error Out of sync with header.
#endif

// See notes in Asura_PS2_GSMDefines.h for what's going on here.
#define MAKE_CAM_VIS_BIT(uCamIdx) (1 << (7 - ((uCamIdx) << 1)))
#define MAKE_CAM_CLIP_BIT_FROM_VIS_BIT(uVisBit) ((uVisBit) >> 1)
#define ALL_CAM_VIS_BITS (0xAA)

#define MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT(puExtern, uNumMSCALs) (++(uNumMSCALs), ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(puExtern))

DEBUGVAR  Asura_Vector_3 dbg_xNormalOffset;
DEBUGVAR  Asura_Vector_3 dbg_xNormalScale;

// T Y P E S /////////////////////////////////////////////////////////////////

enum PENDING
{
    PENDING_UPLOAD_TEX_SYNC         =  1 <<  0,
    PENDING_UPLOAD_RENDERSTATE      =  1 <<  1,
    PENDING_UPLOAD_CAM              =  1 <<  2,
    PENDING_UPLOAD_MESH_FLAGS       =  1 <<  3,
    PENDING_UPLOAD_CAM_FLAGS        =  1 <<  4,
    PENDING_UPLOAD_Y_PARAMS         =  1 <<  5,
};

// G L O B A L S /////////////////////////////////////////////////////////////

// Debugging controls.
DEBUGVAR bool dbg_bWATERFillSPRInBegin     = false;
DEBUGVAR bool dbg_bWATERFillVU1MemInBegin  = false;
DEBUGVAR bool dbg_bWATERCullOnly           = false;
DEBUGVAR bool dbg_bRemoveRenderKicks       = false;	// NOTE : needs a define enabled in water VCL code to actually do anything !
#if _DEBUG
static bool g_bBreakOnNextStrip = false;
static bool g_bSkipNextStrip    = false;
#endif

// Optimisation controls.
DEBUGVAR bool dbg_bWATEROptimiseOutRedundantTexRegUploads = true;

// used to allow use of water renderer without using visibility flags - once you start using vis flags, then
// it will pay attention to them, but if you don't it assumes everything is always visible.
static bool g_bCamFlagsSet = false;	

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static Asura_PS2_QW* CalculateAndWriteCamData( Asura_PS2_QW* pxOut, u_int uCamIdx, bool bO2WIsIdentity, bool bReflective );
static void WaterSendPendingData( bool bWillMSCNTForNextStrip );

static inline void FillQWWithNOPs( Asura_PS2_QW* pxOut )
{
    pxOut->u128 = 0;
}

static inline void UpdatePassesPerTri( u_int uMeshFlags )
{
    g_uNumPassesPerTri = 1 + (uMeshFlags & ASURA_PS2_WATER__MASK__PASS_1 ? 1 : 0) + (uMeshFlags & ASURA_PS2_WATER__MASK__PASS_2 ? 1 : 0);
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

static bool dbg_water_useTexture = true; 
static bool dbg_water_doAlpha = true;

void Asura_PS2_WaterRenderer::Begin( void )
{

    // If renderer was already there, we assume some setup stuff has
    // already been done.
    const bool bNeedFullSetup = Asura_PS2_VU1::LoadRenderer( Asura_PS2_VU1::RENDERER_WATER );

    // Reserve some of the beginning of SPR.
    (void) Asura_PS2_SPR_Alloc( 0, TOTAL_SPR_QWS_REQD ); // The cast fixes warning in release.

    const Asura_Water_System::Water_Controls* pxControls = Asura_Water_System::GetCurController();

    if( bNeedFullSetup )
    {
        if( dbg_bWATERFillSPRInBegin )
        {
            const u_long128 uFillValue = ASURA_PS2_INIT_ULONG128( 0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDuLL );
            u_long128* puOut = reinterpret_cast<u_long128*>( 0x70000000 );
            for( u_int u = ASURA_PS2_WATER_SPR_QWS_RESERVED; u; --u )
            {
                *puOut++ = uFillValue;
            }
        }
        if( dbg_bWATERFillVU1MemInBegin )
        {
            Asura_PS2_VU1::FillMem( ASURA_PS2_INIT_ULONG128(0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDuLL) );
        }
        
		g_uPendingFlags = 0;
	
#if _DEBUG
		{
	        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( 3 );
			
			Asura_PS2_QW* pxOut = pxBase;

			(reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 2,
																 Asura_PS2_DmaTag::PCE_NOCHANGE,
																 Asura_PS2_DmaTag::ID_CNT,
																 false,
																 0,
																 false,
																 SCE_VIF1_SET_STCYCL(1, 1, false),
																 SCE_VIF1_SET_UNPACK(ASURA_PS2_WATER__ADDR__DEBUG_PARAMS, 2, Asura_PS2_VU1::VIFPK_V4_32, false) );

			(pxOut)->f32s[0] = dbg_xNormalOffset.GetElement(0);
			(pxOut)->f32s[1] = dbg_xNormalOffset.GetElement(1);
			(pxOut)->f32s[2] = dbg_xNormalOffset.GetElement(2);
			pxOut++;

			(pxOut)->f32s[0] = dbg_xNormalScale.GetElement(0);
			(pxOut)->f32s[1] = dbg_xNormalScale.GetElement(1);
			(pxOut)->f32s[2] = dbg_xNormalScale.GetElement(2);
			pxOut++;
		}
#endif

		Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL, false );

// (MP - 06.02.2005) Ok, this is a bad idea, seeing as we've yet to setup a
// bunch of default state below (mesh flags, etc.)
//		SetupYManipulations( Asura_Water_System::GetWaterCycleTime1(), 
//							 Asura_Water_System::GetWaterCycleTime2(), 
//							 Asura_Water_System::GetWaterCycleTime3(), 
//							 pxControls );
//
//		for (int iPass = 0; iPass < 3; iPass++)
//		{
//			int iMask = 1<<iPass;
//			if( (iPass == 0) || (pxControls->m_uFlags & iMask) )
//			{
//				if( pxControls->m_aiTextureID[iPass] != ASURA_TEXTURE_NONE )
//				{
//					Asura_PS2_WaterRenderer::SetTexture( pxControls->m_aiTextureID[iPass], SCE_GS_LINEAR, SCE_GS_LINEAR, iPass, iPass );
//				}
//			}
//		}
//		WaterSendPendingData( false );

        
        const u_int uNumCams = Asura_Camera::GetNumberOfCameras();

        const u_int uNumQWsForCamUNPACK    = uNumCams * 5;
        const u_int uNumQWsForGIFTagUNPACK = ASURA_PS2_WATER__SIZE__GIF_TAGS;

		const u_int uNumQWsForParameters = 2;

        const u_int uNumQWsForDMA = uNumQWsForCamUNPACK
                                  + 1 + uNumQWsForGIFTagUNPACK
//								  + uNumQWsForParameters
                                  + 1 + Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS
                                  + 1;

        const u_int uTotalQWs = 1 + uNumQWsForDMA;

        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumQWsForDMA,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_CNT,
                                                             false,
                                                             0,
                                                             false,
                                                             SCE_VIF1_SET_STCYCL(1, 1, false),
                                                             SCE_VIF1_SET_UNPACK(ASURA_PS2_WATER__ADDR__CONST_CAM_DATA, uNumQWsForCamUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false) );

        // Const cam data.
        for( u_int uCam = 0; uCam < uNumCams; ++uCam )
        {
            pxOut = Asura_PS2_GS::GetViewportSetupGifCmds( uCam, Asura_PS2_GS::CONTEXT_1, pxOut );

            (pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrScaleVec(uCam) );
            (pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrTransVec(uCam) );
            (pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrClipScaleVec(uCam) );

            if( dbg_bWATERCullOnly )
            {
                // Make x,y,z as large as possible and w as small as possible
                // so (practically) all clip tests fail and the clipper
                // thinks all the geometry's entirely outside the frustum.
                --pxOut;
                pxOut->f32s[0] = pxOut->f32s[1] = pxOut->f32s[2] = Asura_Maths::LargeFloat;
                pxOut->f32s[3] = 0.0f;
                ++pxOut;
            }
        }

        // Setup BASE and OFFSET and start UNPACK of GIF tags.
        pxOut->u32s[0] = SCE_VIF1_SET_BASE( ASURA_PS2_WATER__ADDR__INPUT_BFR_0, false );
        pxOut->u32s[1] = SCE_VIF1_SET_OFFSET( ASURA_PS2_WATER__SIZE__INPUT_BFR, false );
        pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_WATER__ADDR__GIF_TAGS, uNumQWsForGIFTagUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false );
        ++pxOut;

        // Geometry setup GIF tag. The default NLOOP should be for the base
        // pass; we'll fix it up on VU1 for anything else.
        {
            Asura_PS2_GifTag* pxGeomSetupGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
            pxGeomSetupGifTag->SetExceptRegs( ASURA_PS2_WATER__RS_OUT_0_NLOOP, true, false, 0, SCE_GIF_PACKED, 1 );
            pxGeomSetupGifTag->uAllRegs = SCE_GIF_PACKED_AD;
        }
        
        // Geometry draw GIF tag. Note that NLOOP is setup in code.
        {
            Asura_PS2_GifTag* pxGeomDrawGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );

            u_int uPrimType = SCE_GS_PRIM_TRI;
            bool  bTME      = true;

#ifndef ASURA_PS2_SONY_SUBMISSION
            if( Asura_Render::GetWireframe() )
            {
                uPrimType = SCE_GS_PRIM_LINESTRIP;
                bTME      = false;
            }
#endif
												// prim, gouraud, tex,                   fog,   alpha,         AntiA, stq, ctxt, fix
            const u_int uPRIM = SCE_GS_SET_PRIM( uPrimType,    1, dbg_water_useTexture, false, dbg_water_doAlpha, false, 0,   0,    0 );
            pxGeomDrawGifTag->SetExceptRegs( 0, true, true, uPRIM, SCE_GIF_PACKED, 3 );

            pxGeomDrawGifTag->uAllRegs = (SCE_GS_ST << 0) | (SCE_GS_RGBAQ << 4) | (SCE_GS_XYZ2 << 8);
        }

        // Run INIT program and start UNPACK of shadow constant data.
        pxOut->u32s[0] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_WATER_INIT );
        pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );

        // DIRECT for default renderstate setup.
        pxOut->u32s[3] = SCE_VIF1_SET_DIRECT( Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS, false );
        ++pxOut;

        Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1, pxOut, true, false );
        pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

        // Calculate and upload sun dir.
        {
            const Asura_Water_System::Water_Controls* pxControls = Asura_Water_System::GetCurController();

			if( pxControls->m_bUseLevelSun )
			{
				*g_pxSunDir = Asura_Weather_System::GetSunPos();
                g_pxSunDir->Normalise();
			}
			else
			{
				*g_pxSunDir = reinterpret_cast<const Asura_Vector_3&>( pxControls->m_afSunDir[0] );
                // We expect that to be normalised anyway.
			}

            // Upload a modified negated sun dir without the y component for
            // our dot product.
            Asura_Vector_2 xModifiedSunDir( -g_pxSunDir->x, -g_pxSunDir->z );
            xModifiedSunDir.Normalise();
            
            pxOut->u32s[0] = SCE_VIF1_SET_UNPACK( ASURA_PS2_WATER__ADDR__NEG_SUN_DIR, 1, Asura_PS2_VU1::VIFPK_V3_32, false );
            pxOut->f32s[1] = xModifiedSunDir.x;
            pxOut->f32s[2] = 0.0f;
            pxOut->f32s[3] = xModifiedSunDir.y;
            ++pxOut;
        }

        Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );

        // Reset buffer indices. We assume these are still in sync if
        // renderer was already present on VU; the only time they'll change
        // is when we run one of the VU1_WATER_SWITCH... subroutines.
        g_uCurVURSBfr  = 0;
        g_uCurVUCamBfr = 0;

        // Fill in constant elements of renderstate bfr.
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_ALPHA      )->u64s[1] = SCE_GS_ALPHA_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEX1       )->u64s[1] = SCE_GS_TEX1_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEX0       )->u64s[1] = SCE_GS_TEX0_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_MIPTBP1    )->u64s[1] = SCE_GS_MIPTBP1_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEST       )->u64s[1] = SCE_GS_TEST_1;

        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_HORIZON_TEX1    )->u64s[1] = SCE_GS_TEX1_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_HORIZON_TEX0    )->u64s[1] = SCE_GS_TEX0_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_HORIZON_MIPTBP1 )->u64s[1] = SCE_GS_MIPTBP1_1;
		
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_ALPHA       )->u64s[1] = SCE_GS_ALPHA_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_TEX1        )->u64s[1] = SCE_GS_TEX1_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_TEX0        )->u64s[1] = SCE_GS_TEX0_1;
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_MIPTBP1     )->u64s[1] = SCE_GS_MIPTBP1_1;

        // Setup some defaults.
        g_uMeshFlags       = 0;
        g_uPrevMeshFlags   = 0;
        g_uPendingFlags    = PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM | PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS;
        g_uPendingVtxCount = 0;

		if (!g_bCamFlagsSet)
		{
			int iCams = Asura_Camera::GetNumberOfCameras();
			switch (iCams)
			{
			case 1:
				g_uCamFlagsForVU   = 0xC0;
				break;
			case 2:
				g_uCamFlagsForVU   = 0xF0;
				break;
			case 3:
				g_uCamFlagsForVU   = 0xFC;
				break;
			case 4:
				g_uCamFlagsForVU   = 0xFF;
				break;
			default:
				g_uCamFlagsForVU   = 0xC0;
				break;
			}
			if (iCams > 1)
			{
				g_uCamFlagsForVU += 0x100;
			}
		}
		else
		{
			g_uCamFlagsForVU   = 0;		// this implies that setting the cam flags must be done AFTER beginning the water renderer
		}

        g_bO2WIsIdentity   = true;
        g_uNumPassesPerTri = 1;

        // Setup default renderstate.
        g_aiCurTex[0] = ASURA_TEXTURE_NONE;
        g_aiCurTex[1] = ASURA_TEXTURE_NONE;
        g_aiCurTex[2] = ASURA_TEXTURE_NONE;
    }
    else // not bNeedFullSetup
    {
        // (MP - 06.02.2005) I'm pretty sure we never reach this case since
        // all the water gets rendered in one batch. Let's assert to be sure,
        // as it almost certainly wouldn't work if we ever get here.
        Asura_Assert( false, "Multiple water renders?" );
    }

	SetupYManipulations( Asura_Water_System::GetWaterCycleTime1(), 
						 Asura_Water_System::GetWaterCycleTime2(), 
						 Asura_Water_System::GetWaterCycleTime3(), 
						 pxControls );

	for (int iPass = 0; iPass < 3; iPass++)
	{
		int iMask = 1<<iPass;
		if( (iPass == 0) || (pxControls->m_uFlags & iMask) )
		{
			if( pxControls->m_aiTextureID[iPass] != ASURA_TEXTURE_NONE )
			{
				Asura_PS2_WaterRenderer::SetTexture( pxControls->m_aiTextureID[iPass], SCE_GS_LINEAR, SCE_GS_LINEAR, iPass, iPass );
			}
		}
	}

    SetRenderstate( pxControls->m_uFlags );

    // (MP - 06.02.2005) I don't think we really want to be doing this. Most
    // of this stuff never changes anyway so we might as well explicitly
    // upload the constant setup stuff (bits based on the water controls,
    // renderstate, etc.) and then leave it at that. We're doing loads of
    // unnecessary work in WaterSendPendingData() every time we send some
    // strips later on.
	WaterSendPendingData( false );
}

bool Asura_PS2_WaterRenderer::IsMaterialReflective( u_int uWaterFlags )
{
    return( uWaterFlags & AsuraWaterFlag_Reflective );
}


void Asura_PS2_WaterRenderer::SetRenderstate( u_int uWaterFlags )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewMeshFlags    = uOldMeshFlags;
    u_int uNewPendingFlags = g_uPendingFlags;

    // Mask off mesh flags for RS set up externally and after this method.
//    uNewMeshFlags &= ~ASURA_PS2_WATER__MASK__ANY_VTX_MANIPULATIONS;

    {
        // Mask off the mesh flags that we may or may not turn on below.
        uNewMeshFlags &= ~(ASURA_PS2_WATER__FLAG__SUN_REFLECTIVE | ASURA_PS2_WATER__FLAG__HORIZON_REFLECT);

        u_long uTEST = SCE_GS_SET_TEST( false, 0, 0, 0, false, 0, true, SCE_GS_ZGEQUAL );

        // Pass 0: Cs * As + Cd * (1 - As) - NOTE: non-transparency is now handled by disabling the texture alpha and setting water alpha to 128 (ie solid)
        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
            SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );

        (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEST)->u64s[0] = uTEST;

        if( IsMaterialReflective(uWaterFlags) )
        {
			if (uWaterFlags & AsuraWaterFlag_SunReflect)
			{
			
				uNewMeshFlags |= ASURA_PS2_WATER__FLAG__SUN_REFLECTIVE;

				// If reflective flag wasn't previously on, setup tex matrix now.
				if( !(uOldMeshFlags & ASURA_PS2_WATER__FLAG__SUN_REFLECTIVE) )
				{
				// in order to provide a proper texture matrix we are going to use the look-at code, which will make a 
				// matrix such that (0,0,1) will point at the sun when it is applied.
				// we then use the transpose of this (i.e. cheap inverse) together with appropriate scaling and offset to
				// make the texture matrix we need

                    const Asura_Water_System::Water_Controls* pxControls = Asura_Water_System::GetCurController();

					Asura_Matrix_3x3 xTexMat;
                    Asura_Maths::OrientationFromDirection( xTexMat, *g_pxSunDir );

					// transpose, offset and scale all in one go !
					float fScaleX  = pxControls->m_afTexScale[0];	// massively scaled up, results in small window getting whole texture
					float fScaleY  = pxControls->m_afTexScale[1];
					float fOffsetX = pxControls->m_afTexOffset[0];	
					float fOffsetY = pxControls->m_afTexOffset[1];	
					
					sceVu0UnitMatrix( &g_pxTexMatBfr->Vu0FVec );	// initialise to empty (I)
					(g_pxTexMatBfr + 0)->f32s[0] = xTexMat.e11 * fScaleX;
					(g_pxTexMatBfr + 0)->f32s[1] = xTexMat.e12 * fScaleY;
					(g_pxTexMatBfr + 0)->f32s[2] = xTexMat.e13 * fScaleX;

					(g_pxTexMatBfr + 1)->f32s[0] = xTexMat.e21 * fScaleX;
					(g_pxTexMatBfr + 1)->f32s[1] = xTexMat.e22 * fScaleY;
					(g_pxTexMatBfr + 1)->f32s[2] = xTexMat.e23 * fScaleX;

					(g_pxTexMatBfr + 2)->f32s[0] = xTexMat.e31 * fScaleX;
					(g_pxTexMatBfr + 2)->f32s[1] = xTexMat.e32 * fScaleY;
					(g_pxTexMatBfr + 2)->f32s[2] = xTexMat.e33 * fScaleX;

					(g_pxTexMatBfr + 3)->f32s[0] =  fOffsetX;
					(g_pxTexMatBfr + 3)->f32s[1] =  fOffsetY;

					uNewPendingFlags |= PENDING_UPLOAD_CAM;
				}
			}

			if (uWaterFlags & AsuraWaterFlag_HorizonReflect)
			{
				uNewMeshFlags |= ASURA_PS2_WATER__FLAG__HORIZON_REFLECT;

				// If reflective flag wasn't previously on, setup tex matrix now.
				if( !(uOldMeshFlags & ASURA_PS2_WATER__FLAG__HORIZON_REFLECT) )
				{
					// nothing to do, because all the work for this particular case is done
					// in the camera upload code
				}
			}

			// NOTE : one alpha blend mode is shared between both reflective layers, as they are both additive
			// For water the alpha already used is likely to be variable and you DON'T want it to effect this layer
            // Pass 2: Cd += Cs - plain old additive
            (g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_ALPHA)->u64s[0] =
                SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0x80 );

			// pass through request to blend the horizon reflection instead of add it
			if (uWaterFlags & AsuraWaterFlag_BlendHorizon)
			{
				uNewMeshFlags |= ASURA_PS2_WATER__FLAG__SWAP_BLEND_HORIZON;
			}
        }

        uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
    }

	if (dbg_bRemoveRenderKicks)
	{
		uNewMeshFlags |= ASURA_PS2_WATER__FLAG__SKIP_KICKS;
	}

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}



void Asura_PS2_WaterRenderer::SetTexture( int iTexID, int iMMAG, int iMMIN, u_int uBfr, int iPass )
{
    const u_int uCurPendingFlags = g_uPendingFlags;
    u_int uNewPendingFlags = uCurPendingFlags;

    bool bUploaded = false;

    {
        bool bUpload = true;

        // If there's another texture sync pending, say from another pass's
        // texture upload, then sync with that first.
        if( uCurPendingFlags & PENDING_UPLOAD_TEX_SYNC )
        {
            // We want to avoid doing this sync if possible since it means a
            // FLUSH much earlier, probably forcing a stall while the
            // previous strip finishes rendering. A safe case for doing this
            // is if this next texture we're going to upload is going to some
            // fixed buffer and is already resident there. Note that if we
            // skip the sync, we have to skip the upload below as well (since
            // otherwise internal texture management settings may not be
            // right when we come to do the sync later.) This means that it's
            // not safe to skip the sync if auto buffer switching since we
            // can't skip the logic in UploadTexture() that remembers which
            // buffer we used last (and therefore doesn't go overwriting the
            // texture in it while we're still using it.)
            // (MP - 09.01.2004) TODO: This restriction may become a
            // significant problem and bottleneck if/when we support more
            // general multipassing.
            if( Asura_PS2_TextureManagement::BFR_AUTO != uBfr )
            {
                if( Asura_PS2_TextureManagement::BFR_NONE != Asura_PS2_TextureManagement::HasTextureBeenUploaded(iTexID, uBfr) )
                {
                    bUpload = false;
                }
            }

            if( bUpload )
            {
                Asura_PS2_TextureManagement::SetupTextureEnvironment( g_bTexUploaded, ASURA_PS2_TEXREGFLAG_TEXFLUSH );
            }
        }

        if( bUpload )
        {
            bUploaded = Asura_PS2_TextureManagement::UploadTexture( iTexID, uBfr );

            g_bTexUploaded = bUploaded;
            uNewPendingFlags |= PENDING_UPLOAD_TEX_SYNC;
        }
    }

    // Refresh tex setup stuff if necessary. Note that a check on whether
    // or not the setup has already been done for the specified texture ID
    // breaks in two cases: (i) if they're requesting the same texture but
    // with different filtering settings; (ii) if texture caching isn't
    // working, so that we've uploaded the texture again despite it being
    // resident. I don't think (i) is likely to happen, mainly because we
    // use bilinear everywhere on meshes, and we can workaround (ii) by
    // forcing a refresh if we found the texture needed to be uploaded above.

	switch (iPass)
	{
	case 0:
		if( bUploaded || iTexID != g_aiCurTex[iPass] )
		{
			Asura_PS2_TextureManagement::GetMIPTBP1Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_MIPTBP1)->u64s[0], iTexID, uBfr );
	        Asura_PS2_TextureManagement::GetTEX1Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEX1)->u64s[0], iTexID, iMMAG, iMMIN );
			Asura_PS2_TextureManagement::GetTEX0Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEX0)->u64s[0], iTexID, uBfr );
	        uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
	        g_aiCurTex[iPass] = iTexID;

			if( !(Asura_Water_System::GetCurController()->m_uFlags & AsuraWaterFlag_Transparent) )
			{
				reinterpret_cast<sceGsTex0 *>(&(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_BASE_TEX0)->u64s[0])->TCC = 0;	// disable texture alpha
			}
		}
		break;

	case 1:
		if( bUploaded || iTexID != g_aiCurTex[iPass] )
		{
			Asura_PS2_TextureManagement::GetMIPTBP1Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_HORIZON_MIPTBP1)->u64s[0], iTexID, uBfr );
	        Asura_PS2_TextureManagement::GetTEX1Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_HORIZON_TEX1)->u64s[0], iTexID, iMMAG, iMMIN );
			Asura_PS2_TextureManagement::GetTEX0Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_HORIZON_TEX0)->u64s[0], iTexID, uBfr );
	        uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
	        g_aiCurTex[iPass] = iTexID;
		}
		break;

	case 2:
		if( bUploaded || iTexID != g_aiCurTex[iPass] )
		{
			Asura_PS2_TextureManagement::GetMIPTBP1Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_MIPTBP1)->u64s[0], iTexID, uBfr );
	        Asura_PS2_TextureManagement::GetTEX1Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_TEX1)->u64s[0], iTexID, iMMAG, iMMIN );
			Asura_PS2_TextureManagement::GetTEX0Setting( &(g_pxRSBfr + ASURA_PS2_WATER__OFFSET__RS_IN_SUN_TEX0)->u64s[0], iTexID, uBfr );
	        uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
	        g_aiCurTex[iPass] = iTexID;
		}
		break;

	default:
		Asura_Assert( false, "Water Texture pass number too large" );
	}

    g_uPendingFlags = uNewPendingFlags;
}



void Asura_PS2_WaterRenderer::SetupYManipulations( float fT1, float fT2, float fT3, const Asura_Water_System::Water_Controls* pC )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewPendingFlags = g_uPendingFlags;
    u_int uNewMeshFlags    = uOldMeshFlags;

	// compute constants for sunlight reflection falloff
	float fK1, fK2;

	fK1 = (pC->m_fSunlightFalloffNear * pC->m_fSunlightFalloffFar) / (pC->m_fSunlightFalloffNear - pC->m_fSunlightFalloffFar);
	fK2 = fK1 / pC->m_fSunlightFalloffNear;

    (g_pxYParams + 0)->f32s[0] = pC->m_fXScale1;
    (g_pxYParams + 0)->f32s[1] = pC->m_fXScale2;
    (g_pxYParams + 0)->f32s[2] = pC->m_fXScale3;
    (g_pxYParams + 0)->f32s[3] = pC->m_fSunBrightness;

    (g_pxYParams + 1)->f32s[0] = pC->m_fZScale1;
    (g_pxYParams + 1)->f32s[1] = pC->m_fZScale2;
    (g_pxYParams + 1)->f32s[2] = pC->m_fZScale3;	// Z direction
    (g_pxYParams + 1)->f32s[3] = pC->m_fThreshold;	// Threshold for subdivision

	(g_pxYParams + 2)->f32s[0] = fT1;
	(g_pxYParams + 2)->f32s[1] = fT2;
	(g_pxYParams + 2)->f32s[2] = fT3;
	(g_pxYParams + 2)->f32s[3] = pC->m_fAlphaLevel;

	(g_pxYParams + 3)->f32s[0] = pC->m_fYScale1;
	(g_pxYParams + 3)->f32s[1] = pC->m_fYScale2;
	(g_pxYParams + 3)->f32s[2] = pC->m_fYScale3;
	(g_pxYParams + 3)->f32s[3] = -pC->m_fLightScale * pC->m_fLightRescale;

	(g_pxYParams + 4)->f32s[0] = 1.0f / pC->m_fFalloff1;
	(g_pxYParams + 4)->f32s[1] = 1.0f / pC->m_fFalloff2;
	(g_pxYParams + 4)->f32s[2] = 1.0f / pC->m_fFalloff3;
	(g_pxYParams + 4)->f32s[3] = pC->m_fLightOffset * pC->m_fLightRescale;	// lighting fakery controls

	(g_pxYParams + 5)->f32s[0] = pC->m_fHorizonBrightness;
	(g_pxYParams + 5)->f32s[1] = pC->m_fHorizonYScale;
	(g_pxYParams + 5)->f32s[2] = pC->m_fHorizonVOffset;
	(g_pxYParams + 5)->f32s[3] = pC->m_fHorizonVScale;

	(g_pxYParams + 6)->f32s[0] = 0.0f;		// unused
	(g_pxYParams + 6)->f32s[1] = pC->m_fSubdivisionYScale;
	(g_pxYParams + 6)->f32s[2] = fK1;		// see calcs at top of function
	(g_pxYParams + 6)->f32s[3] = fK2;

    const float* pfUVOffsets = Asura_Water_System::GetBasePassUVOffset();
	(g_pxYParams + 7)->f32s[0] = pfUVOffsets[0];
	(g_pxYParams + 7)->f32s[1] = pfUVOffsets[1];
	(g_pxYParams + 7)->f32s[2] = pC->m_fSunYScale;
	(g_pxYParams + 7)->f32s[3] = pC->m_fHorizonFresnelScale * 128.0f;	// note : scaled up because on PS2 this goes in the alpha channel

    uNewPendingFlags |= PENDING_UPLOAD_Y_PARAMS;

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_WaterRenderer::End( void )
{
    // Deal with case of having to render mesh with no tris. Unfortunately
    // there are cases where this occurs, eg. with invisible objects being
    // used for collisions only.
    if( g_uPendingFlags )
    {
        WaterSendPendingData( false );
    }

    // FLUSH to prevent subsequent rendering trashing stuff in VU memory that
    // we're still in the middle of using. This gets optimised out wherever
    // possible; see comment below.
    Asura_PS2_QW* pxQWForTag = Asura_PS2_DeferredRendering::GetVifBuffer( 1 );
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxQWForTag))->Set( 0,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_CNT,
                                                            false,
                                                            0,
                                                            false,
                                                            SCE_VIF1_SET_NOP(false),
                                                            SCE_VIF1_SET_FLUSH(false) );

    // Return region of SPR that we reserved in Begin().
    Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(0x70000000) );
}

void Asura_PS2_WaterRenderer::Initialise( void )
{
    Reset(); 
}

void Asura_PS2_WaterRenderer::RenderMesh( u_int uREFAddr, u_int uNumQWsForREF, u_int uNumVertices )
{
    Asura_PS2_DmaTag* pxTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
    pxTag->Set( uNumQWsForREF,
                Asura_PS2_DmaTag::PCE_NOCHANGE,
                Asura_PS2_DmaTag::ID_REF,
                false,
                uREFAddr,
                false );

    pxTag->SetSpareWordsToNOPs();

    WaterSendPendingData( false );

    g_uPendingVtxCount = uNumVertices;
}

void Asura_PS2_WaterRenderer::Reset( void )
{
    
}

void Asura_PS2_WaterRenderer::Shutdown( void )
{
    Reset();
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

float g_waterYscale = 2.0f;
float g_waterXscale = 1.0f;

float g_waterYoffset = 0.0f;

static Asura_PS2_QW* CalculateAndWriteCamData( Asura_PS2_QW* pxOut, u_int uCamIdx, bool bO2WIsIdentity, bool bReflective )
{
    // O->GBC matrix.
    {
        sceVu0FMATRIX& xW2GBC = Asura_PS2_VU1::GetWorldToClipMatGuardband( uCamIdx );
        if( bO2WIsIdentity )
        {
            Asura_PS2_VU1::CopyPS2Mat( pxOut, xW2GBC );
        }
        else
        {
            sceVu0MulMatrix( &pxOut->Vu0FVec, xW2GBC, &g_pxO2W->Vu0FVec );
        }
        pxOut += 4;
    }


	Asura_Camera::SetCameraIndex(uCamIdx);	// use the right damn camera you fool !

    // O->V and texture matrices.
    if( bReflective )
    {

		// Write out camera position for computing camera relative position in reflection calcs
		
		const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();
		* (reinterpret_cast<Asura_Vector_3 *>(pxOut)) = * pxCamPos;

		pxOut+= 4;	// skip space reserved for O2V matrix (not required if working in world space)

		Asura_PS2_VU1::CopyPS2Mat( pxOut, g_pxTexMatBfr);
		pxOut+= 4;

		// setup horizontal rotation, scale and offset for horizon reflection
		Asura_Angle fHeading;
		fHeading = Asura_Maths::ArcTangent( Asura_Camera::GetOrientation()->e13, 
											Asura_Camera::GetOrientation()->e33);

		float fSin, fCos;
		float xOffset;
		fSin = Asura_Maths::Sine(fHeading);
		fCos = Asura_Maths::Cosine(fHeading);
		xOffset = fHeading / (2.0f * Asura_Maths::Pi);

		pxOut->f32s[0] = fSin;
		pxOut->f32s[2] = fCos;
		pxOut->f32s[3] = xOffset;

		// send a zoom scale factor so that cameras which are zoomed in 
		// will set the threshold to be bigger !
		float fVary = Asura_Camera::GetLODMetric();	// will be 1.0f for normal camera, and gets smaller as you zoom in
		pxOut->f32s[1] = 1.0f / fVary;

		pxOut++;


    }

    return pxOut;
}


static void WaterSendPendingData( bool bWillMSCNTForNextStrip )
{
    const u_int uMeshFlags = g_uMeshFlags; //& 0x1ff; // mask out anything else which could cause a DMA crash !!
                                           // (MP - 06.02.2005) What's the point? Correct code should never write anything like that.
    Asura_Assert( !(uMeshFlags & ~0x1FFu), "" );

    u_int uPendingFlags = g_uPendingFlags;

    // Turn on need-to-send-mesh-flags flag if required.
    if( uMeshFlags != g_uPrevMeshFlags )
    {
        uPendingFlags |= PENDING_UPLOAD_MESH_FLAGS;
        g_uPrevMeshFlags = uMeshFlags;
    }

    // Early out if nothing's changed since the previous strip.
    if( !uPendingFlags ) { return; }

    // Anything to UNPACK? Do this first, so we can leave it as long as
    // possible before FLUSHing.
    if( uPendingFlags & (PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM) )
    {
        // Estimate how many QWs we'll take up with it all. We'll err on the
        // side of overestimation and then just hand back any QWs we don't
        // use.
        u_int uEstimatedQWC = 0;
        if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE  ) { uEstimatedQWC += 1 + ASURA_PS2_WATER__SIZE__RENDERSTATE_BFR;           }
        if( uPendingFlags & PENDING_UPLOAD_CAM          ) { uEstimatedQWC += 1 + ASURA_PS2_WATER__SIZE__PER_OBJ_CAM_DATA_BFR;      }
        if( uPendingFlags & PENDING_UPLOAD_Y_PARAMS     ) { uEstimatedQWC += 2 + ASURA_PS2_WATER__SIZE__Y_SCALE_AND_TRANS_PARAMS;  }

        const u_int uTotalQWs = 1 + uEstimatedQWC;
        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        Asura_PS2_DmaTag* const pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut );
        pxCNTTag->Set( 0, // Will be fixed up below as necessary.
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        pxCNTTag->m_uWord2 = SCE_VIF1_SET_STCYCL( 1, 1, false );

        bool bNeedNOPs = false;

        if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE )
        {
            const u_int uBfr  = g_uCurVURSBfr ^ 1;
            const u_int uAddr = ( uBfr ? ASURA_PS2_WATER__ADDR__RENDERSTATE_BFR_1 : ASURA_PS2_WATER__ADDR__RENDERSTATE_BFR_0 );

            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
            }
            pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, ASURA_PS2_WATER__SIZE__RENDERSTATE_BFR, Asura_PS2_VU1::VIFPK_V4_32, false );
            ++pxOut;

            const u_long128* puSrc = &g_pxRSBfr->u128;
            for( u_int u = ASURA_PS2_WATER__SIZE__RENDERSTATE_BFR; u; --u )
            {
                (pxOut++)->u128 = *puSrc++;
            }

            bNeedNOPs     = true;
            g_uCurVURSBfr = uBfr;
        }

        if( uPendingFlags & PENDING_UPLOAD_CAM )
        {
            const u_int uBfr         = g_uCurVUCamBfr ^ 1;
            const bool  bReflective  = true;	//( uMeshFlags & ASURA_PS2_WATER__FLAG__SUN_REFLECTIVE );// need this on always because we need view space co-ordinates
            const u_int uSize        = ( bReflective ? 13 : 4 );
            const u_int uNumCams     = Asura_Camera::GetNumberOfCameras();

            u_int uAddr = ( uBfr ? ASURA_PS2_WATER__ADDR__PER_OBJ_CAM_DATA_BFR_1 : ASURA_PS2_WATER__ADDR__PER_OBJ_CAM_DATA_BFR_0 );

            const bool bO2WIsIdentity = g_bO2WIsIdentity;
            for( u_int uCam = 0; uCam < uNumCams; uAddr += ASURA_PS2_WATER__OFFSET__CAM_PER_OBJ_NEXT_CAM, ++uCam )
            {
//                if( !(uCamVisFlags & (1 << uCam)) ) { continue; }

                if( bNeedNOPs )
                {
                    FillQWWithNOPs( pxOut );
                }
                bNeedNOPs = true;

                (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, uSize, Asura_PS2_VU1::VIFPK_V4_32, false );

                pxOut = CalculateAndWriteCamData( pxOut, uCam, bO2WIsIdentity, bReflective );
            }

            g_uCurVUCamBfr = uBfr;
        }

        if( uPendingFlags & (PENDING_UPLOAD_Y_PARAMS) )
        {
            // Note that we don't bother double buffering these parameters
            // because we don't expect them to change very often. We've
            // UNPACKed all the strip data and other stuff above already
            // anyway so a FLUSHE here isn't much of an issue.

            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
                pxOut->u32s[2] = SCE_VIF1_SET_FLUSHE( false );
            }
            else
            {
                (pxOut++)->u32s[3] = SCE_VIF1_SET_FLUSHE( false );
                FillQWWithNOPs( pxOut );
            }

            if( uPendingFlags & PENDING_UPLOAD_Y_PARAMS )
            {
                (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_WATER__ADDR__Y_SCALE_AND_TRANS_PARAMS,
                                                          ASURA_PS2_WATER__SIZE__Y_SCALE_AND_TRANS_PARAMS,
                                                          Asura_PS2_VU1::VIFPK_V4_32,
                                                          false );

                (pxOut++)->u128 = (g_pxYParams + 0)->u128;
                (pxOut++)->u128 = (g_pxYParams + 1)->u128;
                (pxOut++)->u128 = (g_pxYParams + 2)->u128;	// extended parameters
                (pxOut++)->u128 = (g_pxYParams + 3)->u128;	// more extended parameters
                (pxOut++)->u128 = (g_pxYParams + 4)->u128;	// some more extended parameters
                (pxOut++)->u128 = (g_pxYParams + 5)->u128;	// and even more extended parameters
                (pxOut++)->u128 = (g_pxYParams + 6)->u128;	// and yet more extended parameters
                (pxOut++)->u128 = (g_pxYParams + 7)->u128;	// and still more extended parameters
            }

            bNeedNOPs = true;
        }

        if( !bNeedNOPs ) { (pxOut++)->u32s[3] = SCE_VIF1_SET_NOP( false ); }

        // Verify that we didn't underestimate.
        Asura_Assert( pxOut <= pxBase + uTotalQWs, "" );

        // Fix up tag.
        u_int uActualQWC = pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCNTTag + 1);
        pxCNTTag->m_uQWC = uActualQWC;

        // Return remainder of buffer if necessary.
        u_int uDiff = (pxBase + uTotalQWs) - pxOut;
        if( uDiff ) { Asura_PS2_DeferredRendering::ReturnVifBuffer( pxOut, uDiff ); }
    }

    if( uPendingFlags & PENDING_UPLOAD_TEX_SYNC )
    {
        Asura_PS2_TextureManagement::SetupTextureEnvironment( g_bTexUploaded, ASURA_PS2_TEXREGFLAG_TEXFLUSH );
    }

    // Add any necessary MSCALs.
    {
        u_int uEstimatedQWC = 0;
        // MSCALs to update ptrs after UNPACKs above.
        if( uPendingFlags & (PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM) ) { ++uEstimatedQWC; }
        // ITOPs and MSCALs for mesh/cam flags.
        if( uPendingFlags & (PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS) ) { ++uEstimatedQWC; }

        const u_int uTotalQWs = 1 + uEstimatedQWC;
        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        Asura_PS2_DmaTag* const pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut );
        pxCNTTag->Set( 0, // Will be fixed up below as necessary.
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        u_int uNumMSCALs = 0;
        u_int uWordOut   = 2;

        if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE )
        {
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_WATER_SWITCH_RENDERSTATE_BFR, uNumMSCALs );
        }   
        if( uPendingFlags & PENDING_UPLOAD_CAM )
        {
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_WATER_SWITCH_CAM_DATA_BFR, uNumMSCALs );
        }
        if( uPendingFlags & PENDING_UPLOAD_MESH_FLAGS )
        {
            pxOut->u32s[uWordOut++] = SCE_VIF1_SET_ITOP( uMeshFlags, false );
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_WATER_LOAD_MESH_FLAGS, uNumMSCALs );

            UpdatePassesPerTri( uMeshFlags );
        }
        if( uPendingFlags & PENDING_UPLOAD_CAM_FLAGS )
        {
            pxOut->u32s[uWordOut++] = SCE_VIF1_SET_ITOP( g_uCamFlagsForVU, false );
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_WATER_LOAD_CAM_FLAGS, uNumMSCALs );
        }

        // We need to run the NOP program up to two times:
        //     (i) If we're transferring a whole chain of strips and so will
        //         be MSCNTing to render the next strip, we need a NOP to get
        //         the VU1 PC to point to the right place.
        //    (ii) If we've written an odd number of MSCALs above, including
        //         (i), we need a dummy one to get the TOP register back to
        //         the correct value, ie. pointing to the strip data that the
        //         caller UNPACKed.
        if( bWillMSCNTForNextStrip ) { pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_WATER_NOP, uNumMSCALs ); }
        if( uNumMSCALs & 1         ) { pxOut->u32s[uWordOut++] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_WATER_NOP ); }

        // Pad to end of QW.
        if( uWordOut & 3 )
        {
            for( u_int u = 4 - (uWordOut & 3); u--; )
            {
                pxOut->u32s[uWordOut++] = SCE_VIF1_SET_NOP( false );
            }
        }

        // Fix up output ptr.
        pxOut += (uWordOut >> 2);

        // Verify that we didn't underestimate.
        Asura_Assert( pxOut <= pxBase + uTotalQWs, "" );

        // Fix up tag.
        u_int uActualQWC = pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCNTTag + 1);
        pxCNTTag->m_uQWC = uActualQWC;

        // Return remainder of buffer if necessary.
        u_int uDiff = (pxBase + uTotalQWs) - pxOut;
        if( uDiff ) { Asura_PS2_DeferredRendering::ReturnVifBuffer( pxOut, uDiff ); }
    }

    // Clear pending flags.
    g_uPendingFlags = 0;
}

u_int Asura_PS2_WaterRenderer::GetNumPassesPerTri( void )
{
    return g_uNumPassesPerTri;
}

#endif // #ifdef ASURA_USE_WATER
