//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_MovieToTextureStream.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     Class for streaming in MPEG data. Provides all functionality necessary
//     for uploading each frame's texture data.
//
// History:
//
//     (MH  - 22.08.2002) - Created. Placeholder version for linkage only.
//     (MP  - 20.09.2002) - Initial PS2 version.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Platform_MovieToTextureStream.h"

// Independent
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_MovieToTexture_System.h"

// PS2
#include "Core/Asura_PS2_DMA.h"
#include "Core/Asura_PS2_OS.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_GS.h"

#if defined(ASURA_PS2_SUPPORT_BINK) && defined(ASURA_SUPPORT_FMV_TEXTURES)

#include <sifdev.h>

static const u_int NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD = 4;
static const u_int NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION  = 1 + NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD + 1;
static const u_int NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION  = 1 + NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION + 1;

static bool g_bUpdateFMVTextures = true;

Asura_Platform_MovieToTextureStream::Asura_Platform_MovieToTextureStream(int iTextureIndex, const char* pxMovieName, bool* pbOK)
: m_pcTextureBuffer(NULL)
{
	m_apxDMAChains[0] = NULL;
	m_apxDMAChains[1] = NULL;
	m_uLastUsedFrame  = 0;

	m_iTextureIndex = iTextureIndex;	

	m_pxBinkHandle = BinkOpen( Asura_PS2_OS::GetFullPath(pxMovieName, true), BINKPRELOADALL );

	if (!m_pxBinkHandle) 
	{
		Asura_OutputToDebugger_Retail(("failed to open FMV texture %s\n", pxMovieName));
		*pbOK = false;
		return;
	}

	
	u_int uBytesRequired = m_pxBinkHandle->Width * m_pxBinkHandle->Height * 4;

	Asura_OutputToDebugger(("opened FMV texture %s. w = %u, h = %u, texture buffer size = %u bytes\n", 
		pxMovieName, m_pxBinkHandle->Width, m_pxBinkHandle->Height, uBytesRequired));
	
	m_pcTextureBuffer = reinterpret_cast<char*>(Asura_AllocateAlign(uBytesRequired, ASURA_ALIGN_128_BYTE));
	
	m_ucTBW = m_pxBinkHandle->Width / 64;
	m_ucTW = Asura_Maths::GetCeilLog2(m_pxBinkHandle->Width);
	m_ucTH = Asura_Maths::GetCeilLog2(m_pxBinkHandle->Height);

	if (m_pcTextureBuffer)
	{
		*pbOK = true;
	}
	else
	{
		BinkClose(m_pxBinkHandle);
		m_pxBinkHandle = NULL;
		*pbOK = false;
	}
}

Asura_Platform_MovieToTextureStream::~Asura_Platform_MovieToTextureStream()
{
	if (m_pxBinkHandle)
	{
		BinkClose(m_pxBinkHandle);
	}

	if (m_pcTextureBuffer)
	{
		Asura_DeallocateAlign(m_pcTextureBuffer);
	}

	for (u_int u = 0; u < 2; ++u)
	{
		if (m_apxDMAChains[u])
		{
			Asura_DeallocateAlign(m_apxDMAChains[u]);
		}
	}
}

void Asura_Platform_MovieToTextureStream::Update(void)
{
	if (!m_pxBinkHandle) return;

	// only decode the next frame if we're using this texture to render something
	u_int uFramesSinceUsed = Asura_Timers::GetFrameNumber() - m_uLastUsedFrame;
	if (uFramesSinceUsed <= 1)
	{
		if (!BinkWait(m_pxBinkHandle))
		{
			BinkDoFrame(m_pxBinkHandle);

			BinkCopyToBuffer( m_pxBinkHandle, Asura_PS2_UncachedMem(m_pcTextureBuffer), m_pxBinkHandle->Width * 4, m_pxBinkHandle->Height, 0, 0, BINKSURFACE32 );

			BinkNextFrame(m_pxBinkHandle);
		}
	}
}

u_int Asura_Platform_MovieToTextureStream::GetDmaChainCallAddr( u_int uBfr ) const
{
	Asura_Assert(uBfr < 2, "invalid textuer buffer for FMV texture");

	if (uBfr < 2)
	{
		if (!m_apxDMAChains[uBfr])
		{
			u_int uQWsForDMAChain = GetNumQWsForDMAChain(m_pxBinkHandle->Width, m_pxBinkHandle->Height);
			// Allocate on a 8 QW (ie. 128 byte) boundary to speed up transfers and to avoid cache issues.
			m_apxDMAChains[uBfr] = reinterpret_cast<Asura_PS2_QW*>(Asura_AllocateAlign(uQWsForDMAChain << 4, ASURA_ALIGN_128_BYTE));

			WriteDMAChain(Asura_PS2_UncachedAcceleratedMem(m_apxDMAChains[uBfr]), uBfr);
		}

		m_uLastUsedFrame = Asura_Timers::GetFrameNumber();
		return reinterpret_cast<u_int>(m_apxDMAChains[uBfr]);
	}
	else
	{
		return 0;
	}
}

void Asura_Platform_MovieToTextureStream::GetTex0( u_long *pulTex0, u_int uTBP ) const
{
	*pulTex0 = SCE_GS_SET_TEX0( uTBP,					// Texture VRAM address
								m_ucTBW,				// Texture buffer width/64
								SCE_GS_PSMCT32,			// Texture format (rgba)
								m_ucTW,					// Texture Width (2^tw)
								m_ucTH,					// Texture Height (2^th)
								0,						// Alpha component ?  None
								0,						// Texture function = Modulate
								0,						// Clut VRAM address N/A
								0,						// Clut Format N/A
								0,						// Clut storage mode N/A
								0,						// Clut entry offset N/A
								0);						// Clut buffer load control N/A
}


u_int Asura_Platform_MovieToTextureStream::GetNumQWsForDMAChain( u_int uWidth, u_int uHeight )
{
	// this function is based on the same-named one in Asura_PS2_Texture
	u_int uNumQWs = 0;
	
	// we are uploading PSM32
	u_int uQWsPerLine  = uWidth >> 2;
    u_int uLinesPerSec = (uQWsPerLine ? Asura_PS2_GifTag::MAX_NLOOP / uQWsPerLine : uHeight);
    u_int uNumSections = (uHeight + (uLinesPerSec - 1)) / uLinesPerSec;
    uNumQWs += NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION * uNumSections;

    // Finally, we'll require a RET tag.
    ++uNumQWs;

    // Align to a 64 byte boundary to prevent cache issues.
    uNumQWs = ((uNumQWs + 3) >> 2) << 2;

    return uNumQWs;
}


u_int Asura_Platform_MovieToTextureStream::GetDataSize() const
{
	return m_pxBinkHandle->Width * m_pxBinkHandle->Height * 4;
}


bool Asura_Platform_MovieToTextureStream::WriteDMAChain( Asura_PS2_QW* pxOut, u_int uBuffer ) const
{
	// this function is based on Asura_PS2_Texture::WriteTextureUploadDmaData
	Asura_PS2_QW* const pxOutputStart  = pxOut;

    // Get a pointer to the texture image data.
	const u_long128* puImgData = reinterpret_cast<const u_long128*>(m_pcTextureBuffer);

    // Properties of the texture that are constant across all mipmaps.
    const u_int uUploadPSM  = SCE_GS_PSMCT32;
    const u_int uNumMipmaps = 1;
		  u_int uGsDestAddr	= Asura_PS2_TextureManagement::GetBfrTBP( uBuffer );
		  u_int uUploadW    = m_pxBinkHandle->Width;
		  u_int uUploadH    = m_pxBinkHandle->Height;
	
#if 0//def ASURA_ASSERT
	// Where is the end of this buffer ?
	const u_int uBufferEnd64 = uGsDestAddr + g_aauTexBfrSizesIn64wUnits[g_eCurTexBfrLayout][uBuffer];

	// Check this texture isn't too big for the buffer it's going to
	Asura_Assert( uGsDestAddr + m_ausTBPOffsets[0], "Texture too big for buffer." );
#endif // ASURA_ASSERT

    // For keeping track of when the EOP is required yet...
    u_int uPacketsRemaining = uNumMipmaps;

     // Transfer image data for full size texture.
    if( !WriteUploadDmaDataForOneLevel(m_pxBinkHandle->Width, m_pxBinkHandle->Height, &pxOut, 0, uGsDestAddr, puImgData, !--uPacketsRemaining) )
    {
        return false;
    }
    
 	// Write RET tag.
	(reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 0,
														 Asura_PS2_DmaTag::PCE_NOCHANGE,
														 Asura_PS2_DmaTag::ID_RET,
														 false,
														 0,
														 false,
														 0,
														 0 );

    // Align QWs written to a 64 byte boundary, like GetNumQWsForDMAChain does to prevent cache issues.
    Asura_Assert( GetNumQWsForDMAChain(uUploadW, uUploadH) == ((static_cast<u_int>(pxOut - pxOutputStart) + 3) >> 2) << 2, "" );

    return true;
}

bool Asura_Platform_MovieToTextureStream::WriteUploadDmaDataForOneLevel( u_int uWidth, u_int uHeight, 
	Asura_PS2_QW** ppxOut, u_int uMipLevel, u_int uTBP, const u_long128* puImgData, bool bNeedEOP )
{
    Asura_PS2_QW* pxOut = *ppxOut;

    u_int uUploadPSM = SCE_GS_PSMCT32;
    u_int uUploadW   = uWidth;
    u_int uUploadH   = uHeight;

	u_int uTBW = (uUploadW + 63) >> 6;

    // How many QWs per line of the texture?
    const u_int uQWsPerLine = uUploadW >> 2;

    // How many lines of the texture will we do in each packet?
    const u_int uLinesPerPacket = ( uQWsPerLine ? Asura_PS2_GifTag::MAX_NLOOP / uQWsPerLine : uUploadH );

	u_int uLinesRemaining = uUploadH;
    u_int uDestOffset     = 0;

	do
	{
        const u_int uNumLinesInThisPacket = Asura_Maths::Min( uLinesPerPacket, uLinesRemaining );
        const u_int uNumImageDataQWs = uNumLinesInThisPacket * uQWsPerLine;

        Asura_PS2_QW* const pxSectionStart = pxOut;

		// Write DMA tag.
		// PSJE - Added a VIF_DIRECT to the end of the dma tag.. this allows the texture to go:
		//		via path2 - The DIRECT code is used by vif1 
		//		via path3 - When dmaing to GIF we don't send the dma tag (TTE=0)
        Asura_PS2_DmaTag* const pxCntDmaTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut++ );
        pxCntDmaTag->Set( NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION,
                          Asura_PS2_DmaTag::PCE_NOCHANGE,
                          Asura_PS2_DmaTag::ID_CNT,
                          false,
                          0,
                          false,
                          0,
                          SCE_VIF1_SET_DIRECT( NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION, false ) );

        // GIF tag for register settings.
        Asura_PS2_GifTag* const pxRegGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
        pxRegGifTag->Set( NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD,
                          false,
                          false,
                          0,
                          SCE_GIF_PACKED,
                          1,
                          SCE_GIF_PACKED_AD );

	    // Set transmission between buffers.
		pxOut->u64s[0] = SCE_GS_SET_BITBLTBUF( 0, 0, 0, uTBP, uTBW, uUploadPSM );
		pxOut->u64s[1] = SCE_GS_BITBLTBUF;
        ++pxOut;

	    // Set transmission area between buffers (source (x, y), dest (x, y)
        // and direction.)
		pxOut->u64s[0] = SCE_GS_SET_TRXPOS( 0, 0, 0, uDestOffset, 0 );
		pxOut->u64s[1] = SCE_GS_TRXPOS;
        ++pxOut;

	    // Set size of transmission area.
		pxOut->u64s[0] = SCE_GS_SET_TRXREG( uUploadW, uNumLinesInThisPacket );
		pxOut->u64s[1] = SCE_GS_TRXREG;
        ++pxOut;

	    // Set transmission direction to HOST->LOCAL.
		pxOut->u64s[0] = SCE_GS_SET_TRXDIR( 0 );
		pxOut->u64s[1] = SCE_GS_TRXDIR;
        ++pxOut;

        Asura_Assert( static_cast<u_int>(pxOut - reinterpret_cast<Asura_PS2_QW*>(pxRegGifTag)) - 1 == NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD, "" );

        // GIF tag for IMAGE data.
		// Check if we need an EOP yet... (only put EOPs when bNeedEOP and it's the last chunk of a big upload)
		const bool bEOPThisTime = ( bNeedEOP && uLinesRemaining == uNumLinesInThisPacket );
#ifdef ASURA_ASSERT
		if  (uNumImageDataQWs == 0)
		{
			Asura_Assert( 0, "This texture is too small, and will crash the engine in a non Assert build" );
			Asura_Assert( 0, "Absaloute Minimum texture widths are : 32bit - 4, 16bit - 8, 8bit - 32, 4bit - 64" );

			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( 1,
				                                                 bEOPThisTime,
					                                             false,
						                                         0,
							                                     SCE_GIF_IMAGE,
								                                 0 );
		}
		else
#endif
		{
			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNumImageDataQWs,
				                                                 bEOPThisTime,
					                                             false,
						                                         0,
							                                     SCE_GIF_IMAGE,
								                                 0 );
		}

        Asura_Assert( static_cast<u_int>(pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCntDmaTag)) - 1 == NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION, "" );

    	// Transfer IMAGE data.
		// PSJE - Added a VIF_DIRECT to the end of the dma tag.. this allows the texture to go:
		//		via path2 - The DIRECT code is used by vif1 
		//		via path3 - When dmaing to GIF we don't send the dma tag (TTE=0)
        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumImageDataQWs,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_REF,
                                                             false,
                                                             reinterpret_cast<u_int>(puImgData),
                                                             false,
                                                             0,
                                                             SCE_VIF1_SET_DIRECT( uNumImageDataQWs, false ) );

        Asura_Assert( static_cast<u_int>(pxOut - pxSectionStart) == NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION, "" );

		uLinesRemaining -= uNumLinesInThisPacket;
		uDestOffset     += uNumLinesInThisPacket;
		puImgData       += uNumImageDataQWs;
	}
    while( uLinesRemaining );

    // Update output ptr.
    *ppxOut = pxOut;

    return true;
}


/**** Functions for the PS2 Bink Version of the MovieToTexture_System ****/

static void* g_pIOPAddr;
void Asura_MovieToTexture_System::Platform_Initialise( void )
{
	// allocate some IOP memory for Bink
	g_pIOPAddr = sceSifAllocIopHeap( RAD_IOPMemoryAmount( RAD_IOP_NEED_CORE1 ) );

	// start up the RAD reading and sound IOP module
	if ( RAD_IOPStartUp( 1, -1, g_pIOPAddr, RAD_IOP_NEED_CORE1 | RAD_IOP_NO_INIT_LIBSD ) )
	{
		// use core1's pcm channel
		RAD_IOPHardwareVolumes( 1, 0x3fff, 0x3fff, 0x3fff, 0x3fff );
#ifndef ASURA_PS2_IN_VC
		BinkSoundUseRAD_IOP( 1 );
#endif
	}
	else
	{
		Asura_Assert(false, "failed to start bink");
	}

}

void Asura_MovieToTexture_System::Platform_Shutdown( void )
{
	RAD_IOPShutDown();

	// free the memory we allocated previously
	sceSifFreeIopHeap( g_pIOPAddr );
}

void Asura_MovieToTexture_System::Platform_Reset( void )
{
}

void Asura_MovieToTexture_System::Platform_Update( void )
{
	/// Kevin Lea - 25/09/2002 14:55:10 - independent-esque code: this code be moved to the Update 
	/// method if other platforms are okay with it.
	if (!g_bUpdateFMVTextures) return;

	Asura_Tracking_Handle_List::Iterator xIterator = s_pxHandleList->GetIterator();
    for(xIterator.Begin(); !xIterator.End(); xIterator.Next())
	{
		Asura_Handle iHandle = xIterator.GetCurrentHandle();

		s_ppxStreamStorage[iHandle]->Update();
	}
}

#else // not #if defined(ASURA_PS2_SUPPORT_BINK) && defined(ASURA_SUPPORT_FMV_TEXTURES)


// Stdlib platform-specific.
#include <sifrpc.h>

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_MovieToTexture_System.h"

// Engine PS2.
#include "Core/Asura_PS2_OS.h"
#include "FMV/Asura_PS2_Mpeg.h"
#include "Sound/Asura_PS2_Multistream.h"

// middleware
#include "Sony/multistream/ee/sound.h"

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

// Call back functions
static void InitMovieLoading( int iCallbackParam );
static int LoadMovieData( u_char* pucDest, int iNumBytesToRead, int iCallbackParam );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_MovieToTexture_System.
//

void Asura_MovieToTexture_System::Platform_Initialise( void )
{
}

void Asura_MovieToTexture_System::Platform_Reset( void )
{
    // Shouldn't be anything to do here; the independent code should take
    // care of destroying all active streams. That will also release all used
    // file IDs again (see Asura_Platform_MovieToTextureStream destructor.)
}

void Asura_MovieToTexture_System::Platform_Shutdown( void )
{
}

void Asura_MovieToTexture_System::Platform_Update( void )
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
    for( u_int u = s_uMaxNumberOfMovieStreams; u--; )
    {
        if( s_ppxStreamStorage[u] )
        {
            s_ppxStreamStorage[u]->Update();
        }
    }
#endif
}

//
// Asura_Platform_MovieToTextureStream.
//

Asura_Platform_MovieToTextureStream::Asura_Platform_MovieToTextureStream( int iTextureIdx, const char* pxMovieName, bool* pbOK, bool bLooping, bool bHoldAtEnd )
    : m_xFileHandle( ASURA_INVALID_HANDLE )
    , m_iMpegHandle( -1 )
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
    // Let's be pessimistic and assume we're going to fail until we discover
    // otherwise.
    *pbOK = false;

    if( iTextureIdx < 0 )
    {
        return;
    }
    m_iTextureIdx = iTextureIdx;

    // Construct full relative path.
	Asura_String xFileName;
	Asura_String xFullPath( pxMovieName );
	xFullPath.SplitFilename( 0, &xFileName, 0 );

    char szFullRelPath[64];
#if 1
	const ASURA_PS2_MPEG_MULTIPLEXTYPE eMultiplexing = ASURA_PS2_MPEG_MULTIPLEXTYPE_PSS;
    Asura_PrintfToBfr( szFullRelPath, "fmvs\\%s.pss", 64, true, xFileName.GetString() );
#else
	const ASURA_PS2_MPEG_MULTIPLEXTYPE eMultiplexing = ASURA_PS2_MPEG_MULTIPLEXTYPE_NONE;
    Asura_PrintfToBfr( szFullRelPath, "fmvs\\%s.m2v", 64, true, xFileName.GetString() );
#endif

    // Open the file
	m_xFileHandle = Asura_FileAccess::Open( szFullRelPath );
//	m_xFileHandle = 2;
//	if ( !Asura_PS2_Multistream_FastLoad::OpenFile( m_xFileHandle, szFullRelPath ) )
//	{
//		m_xFileHandle= ASURA_INVALID_HANDLE;
//	}

	if( m_xFileHandle == ASURA_INVALID_HANDLE )
    {
		printf( "ERROR: Couldn't find FMV file \"%s\" for texture stream.\n", szFullRelPath );
		Asura_Assert( false, "" );
		return;
	}

	// Get the file length, so we know when too loop
	m_uFileLength = 0x03200000;	// - 50Mb

    // Create MPEG.
    // (MP - 20.09.2002) TODO: Proper mplex/video/audio parameters...
    // (MP - 20.09.2002) TODO: Idle callback?
    // (MP - 20.09.2002) TODO: Double buffer?
    m_iMpegHandle = Asura_PS2_MpegSystem::Create( eMultiplexing,
                                                  ASURA_PS2_MPEG_VIDEOTYPE_M2V_CSCIPU,
                                                  ASURA_PS2_MPEG_AUDIOTYPE_ADPCM,
                                                  0,
                                                  InitMovieLoading,
                                                  LoadMovieData,
                                                  reinterpret_cast<int>(this),
                                                  NULL,
                                                  false );

    // Check we created that okay.
    if( m_iMpegHandle < 0 )
    {
        Asura_FileAccess::Close( m_xFileHandle );
		m_xFileHandle = ASURA_INVALID_HANDLE;
        return;
    }

    // Success.
    *pbOK = true;
#endif
}

Asura_Platform_MovieToTextureStream::~Asura_Platform_MovieToTextureStream( void )
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
    if( m_iMpegHandle >= 0 )
    {
        Asura_PS2_MpegSystem::Destroy( m_iMpegHandle );
    }

    if( m_xFileHandle != ASURA_INVALID_HANDLE )
    {
        Asura_FileAccess::Close( m_xFileHandle );
		m_xFileHandle = ASURA_INVALID_HANDLE;
    }
#endif
}

u_int Asura_Platform_MovieToTextureStream::GetDmaChainCallAddr( u_int uDestBfrBase ) const
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
	// Bind the texture?
    Asura_PS2_MpegSystem::Bind( m_iMpegHandle );
	// Get the dma chain
    return Asura_PS2_MpegSystem::GetVideo( uDestBfrBase );
#else
    return 0;
#endif
}

void Asura_Platform_MovieToTextureStream::GetTex0( u_long *pulTex0, u_int uTBP ) const
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
	// Bind the texture .. Select this movie stream into g_pxStream ?
    Asura_PS2_MpegSystem::Bind( m_iMpegHandle );
    Asura_PS2_MpegSystem::GetTex0( pulTex0, uTBP );
#endif
}


static u_int AudioReceiver( u_char *pcBuffer, u_int uBufferSize, u_int uReadOffset, u_int uDataAvailable, int param )
{
	// (ToDo) Get data from audio buffer and play it... 
	
	// Return the amount of data we used
	return uDataAvailable;
}

void Asura_Platform_MovieToTextureStream::Update( void )
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
    // (MP - 01.10.2002) TODO: Need to regulate the playback speed.
    Asura_PS2_MpegSystem::Bind( m_iMpegHandle );

	// If the video has finished
	if ( Asura_PS2_MpegSystem::IsFinished() )
	{	// Restart it
		Asura_PS2_MpegSystem::Rewind();
	}

	Asura_PS2_MpegSystem::PrepareNextFrame();

 
	Asura_PS2_MpegSystem::GetAudio( AudioReceiver, 0 );
#endif
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     InitMovieLoading().
// PARAMETERS   iCallbackParam  - The address of the relevant instance of Asura_Platform_MovieToTextureStream.
// RETURNS      nothing
// DESCRIPTION  
// AUTHOR       MP - 20.09.2002.
//
//////////////////////////////////////////////////////////////////////////////
static bool	 bRequestMade	  = 0;
static u_int uTotalLoadedSize = 0;
static bool bMovieStarted	  = false;
static void InitMovieLoading( int iCallbackParam )
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
	uTotalLoadedSize	= 0;
	bRequestMade		= false;
	bMovieStarted		= false;
#endif
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     LoadMovieData().
// PARAMETERS   pucDest         - Destination address.
//              iNumBytesToRead - Duh.
//              iCallbackParam  - The address of the relevant instance of
//                                Asura_Platform_MovieToTextureStream.
// RETURNS      Number of bytes read.
// DESCRIPTION  
// AUTHOR       MP - 20.09.2002.
//
//////////////////////////////////////////////////////////////////////////////
static int LoadMovieData( u_char* pucDest, int iNumBytesToRead, int iCallbackParam )
{
#if ASURA_PS2_USE_NEW_FMV_STUFF
    // Get ptr.
    const Asura_Platform_MovieToTextureStream* pxStream = reinterpret_cast<const Asura_Platform_MovieToTextureStream*>( iCallbackParam );
    
    // Get stream ID.
    Asura_Handle xFileHandle = pxStream->GetFileHandle();

	const u_int uDestAddr = reinterpret_cast<u_int>( pucDest );
	
	void *pBuffer;
	int iLoadSize = Asura_PS2_Multistream_FastLoad::GetData( xFileHandle, (const void **)&pBuffer );;
	
	// Check for end of file...
	if ( iLoadSize == Asura_PS2_Multistream_FastLoad::REACHED_EOF )
	{
		// We've reach the end of the movie...
		u_int *puDest = (u_int*)pucDest;
		puDest[ 0] = 0x000001B7;	// Movie end code
		puDest[ 1] = 0;			// Padding for IPU FIFO
		puDest[ 2] = 0;			// Padding for IPU FIFO
		puDest[ 3] = 0;			// Padding for IPU FIFO
		puDest[ 4] = 0;			// Padding for IPU FIFO
		puDest[ 5] = 0;			// Padding for IPU FIFO
		puDest[ 6] = 0;			// Padding for IPU FIFO
		puDest[ 7] = 0;			// Padding for IPU FIFO
		puDest[ 8] = 0;			// Padding for IPU FIFO
		puDest[ 9] = 0;			// Padding for IPU FIFO
		puDest[10] = 0;			// Padding for IPU FIFO
		puDest[11] = 0;			// Padding for IPU FIFO
		puDest[12] = 0;			// Padding for IPU FIFO
		puDest[13] = 0;			// Padding for IPU FIFO
		puDest[14] = 0;			// Padding for IPU FIFO
		puDest[15] = 0;			// Padding for IPU FIFO
		return (64);
	}
	else
	{
		if ( iLoadSize )
		{
			memcpy( pucDest, pBuffer, Asura_Maths::Min( iLoadSize, iNumBytesToRead) );
			return Asura_Maths::Min( iLoadSize, iNumBytesToRead);
		}
	}

	return iLoadSize;
#else
    return 0;
#endif
}
// End function LoadMovieData().

#endif // end #else // not #if defined(ASURA_PS2_SUPPORT_BINK) && defined(ASURA_SUPPORT_FMV_TEXTURES)
