//////////////////////////////////////////////////////////////////////////////
//
// VU1_CommonMacros.i
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// A collection of Gasp macros that are used in various renderers.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "vcl_sml.i"

// M A C R O S ///////////////////////////////////////////////////////////////

// Macros to unpack normals and/or ADCs uploaded using V4_5. These macros
// all work in-place.
.macro UnpackNormalADC fNrm, fUnpackScale
    ITOF4 \fNrm, \fNrm
    SUBAw ACC,   VF00,  VF00
    MADD  \fNrm, \fNrm, \fUnpackScale
.endm
.macro UnpackNormal fNrm, fUnpackScale
    ITOF4.xyz \fNrm, \fNrm
    SUBAw.xyz ACC,   VF00,  VF00
    MADD.xyz  \fNrm, \fNrm, \fUnpackScale
.endm
.macro UnpackADC fOut, fIn, fUnpackScale
    ITOF4.w \fOut, \fIn
    MUL.w   \fOut, \fOut, \fUnpackScale
.endm

// Macro to make unpack scale QW for use by macros above.
// Explanation of 0x45808002 magic number used:
// This is 4112.0009765625. In theory we want exactly 4112 since then 4112 *
// 8 (8 being the value in w if ADC is on for the vertex) will give 32896, ie.
// 0x47008000. However, there is the 1-bit multiplication error, meaning 4112
// * 8 actually gives 0x47007FFF. 4112 is 0x45808000 so in theory 0x45808001
// would do the trick; however, there's another error involved in getting the
// value into the register in the first place since we're multiplying by 1 to
// get it in there. So, finally, 0x45808002 is required.
.macro MakeNormalUnpackScale fUnpackScale
    LOI 0.125
    ADDi.xyz \fUnpackScale, VF00, I
    LOI 0x45808002 // See below.
    MULi.w \fUnpackScale, VF00, I
.endm

// Switches pCur between two equal sized buffers. Assumes pCur is already set
// to first_bfr or second_bfr and then switches it to the other. Also
// requires that first_bfr and second_bfr are integer constants and that
// second_bfr >= first_bfr. Uses two temp registers called _pFirst and
// _pOffset.
.macro SwitchBfr_Constants pCur, first_bfr, second_bfr
    // Setup ptr to first bfr in a register.
    IADDIU _pFirst, VI00, \first_bfr
    // Calculate current offset from second bfr.
    ISUBIU _pOffset, \pCur, \second_bfr
    // _pOffset is now either 0 if pCur was pSecond or -size if pCur was
    // pFirst. Hence (_pFirst - _pOffset) will give other buffer.
    ISUB   \pCur, _pFirst, _pOffset
.endm
// Same except the two buffer ptrs are specified in registers instead of as
// constants.
.macro SwitchBfr_Regs pCur, pFirstBfr, pSecondBfr
    ISUB _pOffset, \pCur, \pSecondBfr
    ISUB \pCur, \pFirstBfr, _pOffset
.endm

// Linear interpolation along the vector A->B according to the value R, which
// should be in the range [0,1]. Note that this macro trashes ACC.
.macro Lerp fOut, fA, fB, fR, componentMask=xyzw
    ADDAx.\componentMask  ACC,   \fA, VF00 // = A
    MADDAx.\componentMask ACC,   \fB, \fR  // = A + (B * R)
    MSUBx.\componentMask  \fOut, \fA, \fR  // = A + (B * R) - (A * R) = (1 - R) * A + R * B
.endm

// Very simple macro to fix up light colours to make it easier to include
// mesh colour alpha in CalcDirLighting below.
.macro FixUpDirLightColours afLightColours
    // Set w component of third light colour to 0.
    MINIx.w \afLightColours[2], VF00, VF00[x]
.endm
// Macro to calculate directional lighting for a vertex. Does not clamp
// output light values or setup output alpha since it's generally easier for
// caller to do so. Note that this macro trashes the contents of ACC. It also
// uses a temporary register called _fLdotN. Assumes that the light colours
// have previously been fixed up by FixUpDirLightColours.
.macro CalcDirLighting fOutRGBA, fNrm, fPrelightRGBA, fAmbientRGBA, afLightDirs, afLightColours
    // Calculate (L.N).
    MULAx.xyz  ACC,     \afLightDirs[0], \fNrm[x]
    MADDAy.xyz ACC,     \afLightDirs[1], \fNrm[y]
    MADDz.xyz  _fLdotN, \afLightDirs[2], \fNrm[z]
    // Clamp to non-negative range.
    MAXx.xyz _fLdotN, _fLdotN, VF00[x]
    // Accumulate lights. Note that we know the w component of third light
    // colour will be 0 - see FixUpDirLightColours above - so we'll just get
    // alpha from (prelight+ambient) here.
    ADDA       ACC,       \fPrelightRGBA,     \fAmbientRGBA
    MADDAx.xyz ACC,       \afLightColours[0], _fLdotN[x]
    MADDAy.xyz ACC,       \afLightColours[1], _fLdotN[y]
    MADDz      \fOutRGBA, \afLightColours[2], _fLdotN[z]
.endm
// As above except without prelighting option.
.macro CalcDirLighting_NoPrelighting fOutRGBA, fNrm, fAmbientRGBA, afLightDirs, afLightColours
    // Calculate (L.N).
    MULAx.xyz  ACC,     \afLightDirs[0], \fNrm[x]
    MADDAy.xyz ACC,     \afLightDirs[1], \fNrm[y]
    MADDz.xyz  _fLdotN, \afLightDirs[2], \fNrm[z]
    // Clamp to non-negative range.
    MAXx.xyz _fLdotN, _fLdotN, VF00[x]
    // Accumulate lights. Note that we know the w component of third light
    // colour will be 0 - see FixUpDirLightColours above - so we'll just get
    // alpha from (prelight+ambient) here.
    ADDAx      ACC,       \fAmbientRGBA,      VF00[x]
    MADDAx.xyz ACC,       \afLightColours[0], _fLdotN[x]
    MADDAy.xyz ACC,       \afLightColours[1], _fLdotN[y]
    MADDz      \fOutRGBA, \afLightColours[2], _fLdotN[z]
.endm

// Lights the vertex using a 2nd order SH matrix. The matrix is assumed to be
// in the form
//
//      r0 g0 b0 A
//      r1 g1 b1 0
//      r2 g2 b2 ?
//      r3 g3 b3 ?
//
// where r_out = r1 * nx + r2 * ny + r3 * nz + r0 (and so on for the other
// components.) The input alpha (A above) is just copied to the output QW.
//
// As with CalcDirLighting above, the caller is expected to deal with
// clamping output values to <=255, since it's easier and more efficient to
// ensure the I register (or whatever) is correctly setup there. However, the
// macro deals with clamping to >=0. (The SH calculation can produce negative
// values so the additional clamp is necessary here.)
.macro LightVtxWithSH fOutRGBA, fNrm, afSHMatrix
    MULAx      ACC,       \afSHMatrix[1], \fNrm[x]
    MADDAy.xyz ACC,       \afSHMatrix[2], \fNrm[y]
    MADDAz.xyz ACC,       \afSHMatrix[3], \fNrm[z]
    MADDw      \fOutRGBA, \afSHMatrix[0], VF00[w]
    // Clamp to >=0.
    MAXx.xyz \fOutRGBA, \fOutRGBA, VF00[x]
.endm

// Swaps the two specified integer registers. Uses a temporary register
// called _iTmp.
.macro Swap iA, iB
    IADDIU _iTmp, \iA, 0
    IADDIU \iA, \iB, 0
    IADDIU \iB, _iTmp, 0
.endm

// Calculates the inverse square root of the x component of the input
// register. Only the x component of the output register is modified.
// Accurate to about +/- 4%.
.macro ApproxInvSqrt fOut, fIn
    // This is Matthew Jones's specular power trick, also applicable for
    // calculating inv sqrt since it allows arbitrary powers. See the
    // "Calculating power functions on the VU" thread in sce.dev.prog.vu
    // (started on 16/06/2003) for details. There's a slight tweak from
    // another thread ("Analysis of Fast Specular code" from 26/07/2001) to
    // improve accuracy. The original method calculates a*(x^k), ie. in this
    // case 1*(x^-0.5). This is accurate within about 8.86% but the error is
    // always positive. If we multiply by a factor of about 0.965 instead of
    // 1, this accuracy changes to +/- 4%, which is rather more tolerable.
    // The only change required to do this is to the magic number.
    ITOF0.x \fOut, \fIn
    LOI 1597442621.0 // Magic number. This is 1064766013 - 1065353216 * -0.5, ie. (float)FloatAsInteger(a) - (float)FloatAsInteger(1.0) * k
    ADDAi.x ACC, VF00, I
    LOI -0.5 // Power, ie. k.
    MADDi.x \fOut, \fOut, I
    FTOI0.x \fOut, \fOut
.endm
// More general version to allow up to four inv sqrts to be done in parallel.
// However, this requires an additional register to be setup in advance.
.macro ApproxInvSqrtParallel4Setup componentMask=xyzw
    MINIx.\componentMask _aisp4_0000, VF00, VF00
.endm
.macro ApproxInvSqrtParallel4 fOut, fIn, componentMask=xyzw
    ITOF0.\componentMask \fOut, \fIn
    LOI 1597442621.0
    ADDAi.\componentMask ACC, _aisp4_0000, I
    LOI -0.5
    MADDi.\componentMask \fOut, \fOut, I
    FTOI0.\componentMask \fOut, \fOut
.endm
// Even more general version for arbitary powers. You're best off passing the
// magic number in hex since it tends to be very large and ee-gasp likes to
// replace such values with random numbers.
.macro ApproxPowParallel4Setup specMagicNum, specPower, componentMask=xyzw
    // Setup specular power in x component of one register.
    LOI \specPower
    MAXi.x _app4_power, VF00, I
    // We need the magic number in all components (as specified by the mask)
    // of another register.
    LOI \specMagicNum
    MINIx.\componentMask _app4_0000, VF00, VF00
    ADDi.\componentMask _app4_magic, _app4_0000, I
.endm
.macro ApproxPowParallel4 fOut, fIn, componentMask=xyzw
    ITOF0.\componentMask \fOut, \fIn
    ADDAx.\componentMask ACC, _app4_magic, VF00[x]
    MADDx.\componentMask \fOut, \fOut, _app4_power[x]
    FTOI0.\componentMask \fOut, \fOut
.endm

// Approximately normalises the the specified vector. Does not modify the w
// component of the output register. Trashes the contents of ACC in the
// process and uses a temporary register called _fTmp.
.macro ApproxNormaliseXYZ fOut, fIn
    // Calculate squared magnitude of input vector.
    VectorDotProductACC _fTmp, \fIn, \fIn
    // Calculate 1/magnitude.
    ApproxInvSqrt _fTmp, _fTmp
    // Output normalised vector.
    MULx.xyz \fOut, \fIn, _fTmp
.endm
// Same as above except making less use of the accumulator. This may be
// quicker or slower, depending on the context.
.macro ApproxNormaliseXYZ_LessACC fOut, fIn
    VectorDotProduct _fTmp, \fIn, \fIn
    ApproxInvSqrt _fTmp, _fTmp
    MULx.xyz \fOut, \fIn, _fTmp
.endm

