;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;    
;//;    FILE            VU1_DecalRenderer.vcl
;//;    
;//;    DESCRIPTION     VU1 microcode for rendering decal strips
;//;
;//;    AUTHOR          Pete Ellacott 12.12.02
;//;    
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//; 
;//;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;
;//;    Calculation for buffer sizes :
;//;    ------------------------------
;//;
;//;    Camera data = 4*10 = 40(0x28)
;//;    Camera setup list = 3
;//;    Space Remainging = 1024 - 43 = 981(0x3d5)
;//;
;//;    Output buffer sizes = 2*((NVerts*OutSize)+1)
;//;    Input buffer sizes = 2*(NVerts*InSize)
;//;    :. 12*NVerts = 979
;//;    NVerts = 81
;//;
;//;    Output Buffer Size = 1+NVerts*3 = 244(0xF4)
;//;    Input Buffer Size = NVerts*3 = 243(0xF3)
;//;
;//;    Constants uploaded at startup:
;//;    ------------------------------
;//;    0x000   Camera 1 Screeen Matrix
;//;    0x004   Camera 1 Clipping Matrix
;//;    0x008   Camera 1 Scissor Value
;//;    0x009   Camera 1 XYOffset Value
;//;    0x00a   Camera 2 Screeen Matrix
;//;    0x00e   Camera 2 Clipping Matrix
;//;    0x012   Camera 2 Scissor Value
;//;    0x013   Camera 2 XYOffset Value
;//;    0x014   Camera 3 Screeen Matrix
;//;    0x018   Camera 3 Clipping Matrix
;//;    0x01c   Camera 3 Scissor Value
;//;    0x01d   Camera 3 XYOffset Value
;//;    0x01e   Camera 4 Screeen Matrix
;//;    0x022   Camera 4 Clipping Matrix
;//;    0x026   Camera 4 Scissor Value
;//;    0x027   Camera 4 XYOffset Value
;//;
;//;    0x028   Camera Setup - Giftag
;//;    0x029   Drawing - Giftag
;//;    0x02a
;//;
;//;    Buffer layout during rendering :
;//;    --------------------------------
;//;    0x000
;//;            Per Camera Data (screen space/clip space/Scissor/Offset) * (4)
;//;    0x028
;//;            Output Buffer 0 : 
;//;    0x11c
;//;            Output Buffer 1 : 
;//;    0x210
;//;            Input Buffer  0 : 
;//;    0x303
;//;            Input Buffer  1 : 
;//;    0x3f6
;//;            Unused :
;//;    0x3fd
;//;            Camera Setup List : GifTag, Scissor, XYOffset
;//;    0x400
;//;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;// I N C L U D E S   &   D E F I N E S ////////////////////////////////////////
#include "vcl_sml.i"

#define CAMERA_SETUP_BUFFER         (1024 - 3)
#define ASURA_PS2_DECAL_PERCAMERADATA_SIZE      (8)
#define OUTPUT_BUFFER_START         (40)
#define OUTPUT_BUFFER_SIZE          (244)

;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.syntax new

.init_vf_all
.init_vi_all

    --enter
    --endenter

START:
    ;// How many cameras have we got
    XITOP   iCameraQty

    ;// Setup the output buffer pointers
    IADDIU  pThisOutputBuffer, VI00, OUTPUT_BUFFER_START
    IADDIU  pNextOutputBuffer,  pThisOutputBuffer,  OUTPUT_BUFFER_SIZE

;// Preload the constants
    // Load the W->GBC matrix (camera 0)
    MatrixLoad  mO2GBC, 0, VI00
    // Load the GBC->S scale and translation vectors (camera 0)
    LQ fGBC2SScale, 4(VI00)
    LQ fGBC2STrans, 5(VI00)
    ;// Load Scissor and xyoffset values  (camera 1)
    LQ  fSCISSOR,   6(VI00)
    LQ  fXYOFFSET,  7(VI00)
    ;// Load the Camera Setup GIF tag
    LQ  fCamera_SetupGIF, (4*ASURA_PS2_DECAL_PERCAMERADATA_SIZE)(VI00)
    ;// Load the Drawig GIF Tabg
    LQ.yzw  fDrawing_GIF, (4*ASURA_PS2_DECAL_PERCAMERADATA_SIZE + 1)(VI00)
    ;// Create the first cameras setup drawlist in the right place
    SQ  fCamera_SetupGIF,   (CAMERA_SETUP_BUFFER+0)(VI00)
    SQ  fSCISSOR,           (CAMERA_SETUP_BUFFER+1)(VI00)
    SQ  fXYOFFSET,          (CAMERA_SETUP_BUFFER+2)(VI00)
;// Create usefull i0x8000 register - for setting EOP
    IADDIU  i0x8000,    VI00,       0x7fff
    IADDIU  i0x8000,    i0x8000,    1

    --cont

DECAL_BATCH_SETUP:
;// Setup constants in a register for colour scaling and clamping
    LOI 128
    ADDi.x vColourScaleAndClampConstants, VF00, I
    LOI 255
    ADDi.y vColourScaleAndClampConstants, VF00, I
;// Get the input buffer address
    XTOP    pInputBuffer
;// Get the vertex count from ITOP
    XITOP   iVertexQty
;// Render the strips for each camera
    IADD    iCameraCount,   VI00,   iCameraQty
    IADDIU  pCameraData,    VI00,   0
    // Calculate end of input ptr.
    IADD pInEnd, pInputBuffer, iVertexQty
    IADD pInEnd, pInEnd, iVertexQty
    IADD pInEnd, pInEnd, iVertexQty
PER_CAMERA_RENDERING_LOOP:
    --LoopCS    1,0
        ;// Render the batch of verts for this camera
        IADDIU  pIn,            pInputBuffer,       0
        IADDIU  pOut,           pThisOutputBuffer,  1
        IADD    iVertexCount,   VI00,   iVertexQty
        ;// Clear the clipping flags
        FCSET   0
        // Fix input Q at 1
        MAXw.z fSTQIn, VF00, VF00

VERTEX_LOOP:
        --LoopCS    3,0
            ;// Load the vertex, ST and colour
            LQ.xyz  fVertexIn,   0(pIn)
            ILW.w   iDegenerate, 0(pIn)
            LQ.xy   fSTQIn,      1(pIn)
            LQ      fRGBA,       2(pIn)
            // Convert input ST.
            ITOF12.xy fSTQIn, fSTQIn
            // Transform to GB clip space.
            MatrixMultiplyVertexW1 fGBCPos, mO2GBC, fVertexIn
            // Perform clipping judgement.
            CLIPw.xyz fGBCPos, fGBCPos
            // Scale and translate to screenspace.
            MULA  ACC,        fGBCPos,     fGBC2SScale
            MADDw fVertexOut, fGBC2STrans, fGBCPos[w]
            ;// Do the Perspective divide
            DIV Q,      VF00[w],    fVertexOut[w]
            ;// Apply the perspective divide to vertex and stq
            MULq.xyz    fVertexOut, fVertexOut, Q
            MULq.xyz    fSTQOut,    fSTQIn,     Q
            ;// Convert XYZ for the GIF packet
            FTOI4.xyz   fVertexOut, fVertexOut
            ;// Output STQ, XYZ
            SQ.xyz fSTQOut,    0(pOut) 
            SQ     fRGBA,      1(pOut)
            SQ.xyz fVertexOut, 2(pOut)
            ;// Clip the vertex
            FCAND       VI01,           0x3ffff 
            IADDIU      iADCFlag,       VI01,   0x7fff
            ;// Or in the vertex degeneracy flag we loaded earlier and write it out
            IOR         iADCFlag, iADCFlag, iDegenerate
            ISW.w       iADCFlag, 2(pOut)

            ;// Decrement loop counters and move input and output pointers along
            IADDIU  pIn,    pIn,    3
            IADDIU  pOut,   pOut,   3

            IBNE    pIn,   pInEnd,   VERTEX_LOOP

        ;// Finished / Write out he GIF flag and the vertex count
        IOR     iVertsInList,   iVertexQty, i0x8000
        ISW.x   iVertsInList,   0(pThisOutputBuffer)
        SQ.yzw  fDrawing_GIF,   0(pThisOutputBuffer)
        
        ;// Kick the camera setup list
        IADDIU  pKickMe,    VI00,   CAMERA_SETUP_BUFFER
        XGKICK  pKickMe

        ;// Swap output buffers
        IADDIU  iTemp,              pThisOutputBuffer,  0
        IADDIU  pThisOutputBuffer,  pNextOutputBuffer,  0
        IADDIU  pNextOutputBuffer,  iTemp,              0

        // Update camera count.
        IADDI   iCameraCount,   iCameraCount,   -1

        ;// Send this strip to the GS
        XGKICK  pNextOutputBuffer

        ;// Are there anymore cameras to come ?
        IBEQ    iCameraCount,   VI00,   FINISHED_RENDERING_CAMERAS

        ;// Load the next cameras settings
        IADDIU  pCameraData, pCameraData, ASURA_PS2_DECAL_PERCAMERADATA_SIZE
        MatrixLoad mO2GBC, 0, pCameraData
        LQ fGBC2SScale, 4(pCameraData)
        LQ fGBC2STrans, 5(pCameraData)
        LQ fSCISSOR, 6(pCameraData)
        LQ fXYOFFSET, 7(pCameraData)
        ;// Update the camera setup draw list
        SQ  fSCISSOR,       (CAMERA_SETUP_BUFFER+1)(VI00)
        SQ  fXYOFFSET,      (CAMERA_SETUP_BUFFER+2)(VI00)

        B   PER_CAMERA_RENDERING_LOOP

FINISHED_RENDERING_CAMERAS:
    ;// If we only had one camera, skip the reloading of the first camera
    IADDI   iCameraQtyMinusOne,   iCameraQty, -1
    IBEQ    iCameraQtyMinusOne,   VI00,   ONLY_ONE_CAMERA

    ;// Reload the first cameras data
    MatrixLoad mO2GBC, 0, VI00
    LQ fGBC2SScale, 4(VI00)
    LQ fGBC2STrans, 5(VI00)
    LQ fSCISSOR, 6(VI00)
    LQ fXYOFFSET, 7(VI00)
    ;// Update the camera setup draw list
    SQ  fSCISSOR,       (CAMERA_SETUP_BUFFER+1)(VI00)
    SQ  fXYOFFSET,      (CAMERA_SETUP_BUFFER+2)(VI00)

ONLY_ONE_CAMERA:

    --cont
    
    B    DECAL_BATCH_SETUP

    --exit
    --endexit

.end

