//////////////////////////////////////////////////////////////////////////////
//
// VU1_Skybox.vcl
//
// Memory layout:
//
//     See Asura_PS2_SkyboxDefines.h.
//
//////////////////////////////////////////////////////////////////////////////

// V C L   S E T U P /////////////////////////////////////////////////////////

.init_vf_all
.init_vi_all

.syntax new

.global VU1_SKYBOX_SETUP
.global VU1_SKYBOX_RENDER_BATCH

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "vcl_sml.i"
#include "Asura_PS2_SkyboxDefines.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// These correspond to the bits produced by the CLIP instruction.
#define CLIP_FLAG_RIGHT  (0x01)
#define CLIP_FLAG_LEFT   (0x02)
#define CLIP_FLAG_BOTTOM (0x04)
#define CLIP_FLAG_TOP    (0x08)
#define CLIP_FLAG_FAR    (0x10)
#define CLIP_FLAG_NEAR   (0x20)

// See thread "Texturing problem when clipping" (12.12.2002) in
// sce.dev.prog.gs for details on this stuff.
#define NORMALISE_CLIPPED_STS (0)

.macro GetOtherOutputBuffer pCur, pOther
    ISUBIU pDelta, \pCur, ASURA_PS2_SKYBOX_ADDR_OUTPUT_BFR_1
    // pDelta now contains either 0 if cur bfr is 1 or else -size if it
    // was 0 so we just want to subtract it from the address of bfr 0.
    IADDIU pBfr0, VI00, ASURA_PS2_SKYBOX_ADDR_OUTPUT_BFR_0
    ISUB \pOther, pBfr0, pDelta
.endm

.macro SwitchOutputBuffer p
    GetOtherOutputBuffer \p, \p
.endm

// C O D E ///////////////////////////////////////////////////////////////////

--enter
--endenter

VU1_SKYBOX_SETUP:

    // Prepare a couple of global ptrs.
    IADDIU pRotMat,    VI00, ASURA_PS2_SKYBOX_ADDR_ROT_MATS
    IADDIU pOutputBfr, VI00, ASURA_PS2_SKYBOX_ADDR_OUTPUT_BFR_0
    // Grab camera count from ITOP register.
    XITOP iNumCams
    // Setup constant RGBA in a register. Unfortunately it's necessary to
    // output this per vertex to get perspective correct texture mapping.
    // RGB is specified by artists and so uploaded.
    XTOP   pIn
    LQ.xyz fRGBA, 0(pIn)
    // We fix alpha at 128 though.
    LOI    128.0
    MAXi.w fRGBA, VF00, I
    // Convert ready for GS.
    FTOI0 fRGBA, fRGBA

--exit
    out_vi iNumCams(VI13)
    out_vi pRotMat(VI14)
    out_vi pOutputBfr(VI15)
    out_vf fRGBA(VF31)
--endexit

--enter
    in_vi iNumCams(VI13)
    in_vi pRotMat(VI14)
    in_vi pOutputBfr(VI15)
    in_vf fRGBA(VF31)
--endenter

VU1_SKYBOX_RENDER_BATCH:
    
    // Read in vis flags and num-rot-mat-ptr-updates-reqd count.
    XITOP iITOP
    // Separate.
    IADDIU iMask,             VI00,  ASURA_PS2_SKYBOX_MASK_MAT_PTR_UPDATES
    IAND   iNumMatPtrUpdates, iITOP, iMask
    IADDIU iMask,             VI00,  ASURA_PS2_SKYBOX_MASK_VIS_FLAGS
    IAND   iVisFlags,         iITOP, iMask
    // Setup state for cam loop.
    IADD   iCamLpCtr, VI00, iNumCams
    IADDIU pCamData,  VI00, ASURA_PS2_SKYBOX_ADDR_CAM_DATA
    IADDIU iVisBit,   VI00, ASURA_PS2_SKYBOX_FIRST_VIS_FLAG
    // Keep VCL quiet about the CLIPa register.
    FCSET 0

    VU1_SKYBOX_ROT_MAT_UPDATE_LOOP:
    
        IBEQ iNumMatPtrUpdates, VI00, VU1_SKYBOX_DONE_ROT_MAT_UPDATES
        IADDIU pRotMat, pRotMat, 3
        ISUBIU iNumMatPtrUpdates, iNumMatPtrUpdates, 1
        B VU1_SKYBOX_ROT_MAT_UPDATE_LOOP

VU1_SKYBOX_DONE_ROT_MAT_UPDATES:

    VU1_SKYBOX_CAMERA_LOOP:

        --LoopCS 1, 0

        // Skip if this cam can't see the current face.
        IAND iVis, iVisFlags, iVisBit
        IBEQ iVis, VI00,      VU1_SKYBOX_NEXT_CAM
        // Load W->GC matrix and GC->S scale and translation vectors.
        MatrixLoad mInW2GC,    0,pCamData
        LQ         fGC2SScale, 4(pCamData)
        LQ         fGC2STrans, 5(pCamData)
        // Prepare world transform.
        LQ mWT[0], 0(pRotMat)
        LQ mWT[1], 1(pRotMat)
        LQ mWT[2], 2(pRotMat)
        // Translation row is just cam pos.
        LQ mWT[3], 7(pCamData)
        // Concatenate with W->GC transform.
        MatrixMultiply mW2GC, mWT, mInW2GC
        // Copy GIF tags and GS reg setup to output bfr.
        LQ fGIFTag0,  ASURA_PS2_SKYBOX_ADDR_GIF_TAGS(VI00)
        LQ fSCISSOR,  8(pCamData)
        LQ fXYOFFSET, 9(pCamData)
        LQ fGIFTag1,  (ASURA_PS2_SKYBOX_ADDR_GIF_TAGS+1)(VI00)
        SQ fGIFTag0,  0(pOutputBfr)
        SQ fSCISSOR,  1(pOutputBfr)
        SQ fXYOFFSET, 2(pOutputBfr)
        SQ fGIFTag1,  3(pOutputBfr)
        // Prepare input and output ptrs.
        XTOP   pIn
        IADDIU pOut, pOutputBfr, 4
        // Prepare loop ctr. The vertex count is hiding in the first vertex's
        // position's w component.
        ILW.w iVtxLpCtr, 1(pIn)// (MP - 20.08.2003) 
        // GIF tag provided just has a zero NLOOP so fix that up.
        ILW.x iNLOOPandEOP, 3(pOutputBfr)
        IOR   iNLOOPandEOP, iNLOOPandEOP, iVtxLpCtr
        ISW.x iNLOOPandEOP, 3(pOutputBfr)
        // Keep a record of whether or not we need to go clipping any polys.
        IADD iAnyClipped, VI00, VI00

        VU1_SKYBOX_VERTEX_LOOP:

            --LoopCS  3, 0
            --LoopAbs 16

            // Load.
            LQ.xyz fSTQ, 0(pIn)
            ILW.w  iADC, 0(pIn)
            LQ.xyz fXYZ, 1(pIn)
            // Convert ST from 12:4 fixed point.
            ITOF4.xy fSTQ, fSTQ
            // Transform to (guardband) clipspace and perform clip test.
            MatrixMultiplyVertexW1 fGClipXYZ, mW2GC, fXYZ
            CLIPw.xyz fGClipXYZ, fGClipXYZ
            // Scale and translate to screenspace.
            MULA  ACC,     fGClipXYZ,  fGC2SScale
            MADDw fScrXYZ, fGC2STrans, fGClipXYZ[w]
            // Get clip result and OR on degeneracy flag.
            FCAND VI01, 0x3FFFF
            IOR iAnyClipped, iAnyClipped, VI01
            IADDIU iClipADC, VI01, 0x7FFF
            IOR iADC, iClipADC, iADC
            // Perspective divide.
            DIV Q, VF00[w], fScrXYZ[w]
            MULq.xyz fScrXYZ, fScrXYZ, Q
            MULq.xyz fSTQ,    fSTQ,    Q
            // Convert XYZ to GS fixed point format.
            FTOI4.xyz fScrXYZ, fScrXYZ
            // Output.
            SQ.xyz fSTQ,    0(pOut)
            SQ     fRGBA,   1(pOut)
            SQ.xyz fScrXYZ, 2(pOut)
            ISW.w  iADC,    2(pOut)
            // Update loop state.
            IADDIU pIn,       pIn,       ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_IN
            IADDIU pOut,      pOut,      ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT
            ISUBIU iVtxLpCtr, iVtxLpCtr, 1
            // Are we done yet?
            IBNE iVtxLpCtr, VI00, VU1_SKYBOX_VERTEX_LOOP

        // Send to GS and swap output bfrs.
        XGKICK pOutputBfr
        SwitchOutputBuffer pOutputBfr
        // Any vertices clipped?
        IBEQ iAnyClipped, VI00, VU1_SKYBOX_NEXT_CAM
        // Ugh, yes. Off we go then.
        B VU1_SKYBOX_CLIP

    VU1_SKYBOX_NEXT_CAM:

        // Shift vis bit left.
        IADD iVisBit, iVisBit, iVisBit
        // Update loop state.
        IADDIU pCamData,  pCamData,  10
        ISUBIU iCamLpCtr, iCamLpCtr, 1
        // Any more cameras?
        IBNE   iCamLpCtr, VI00,      VU1_SKYBOX_CAMERA_LOOP

VU1_FINISHED_BATCH:

--exit
    out_vi iNumCams(VI13)
    out_vi pRotMat(VI14)
    out_vi pOutputBfr(VI15)
    out_vf fRGBA(VF31)
--endexit

VU1_SKYBOX_CLIP_FINISHED:

    // Retrieve int registers from stack.
    MTIR pRotMat,   fStack0[x]
    MTIR iVisBit,   fStack0[y]
    MTIR iCamLpCtr, fStack0[z]
    // Check if there's anything in the output buffer that needs rendering.
    // (We could just kick the packet anyway, but we don't want to swap
    // output bfr and therefore synchronise if it's not strictly necessary.)
    ILW.x  iVtxCount, 3(pOutputBfr)
    IADDIU iMask, VI00, 0x7FFF
    IAND   iVtxCount, iVtxCount, iMask
    IBEQ   iVtxCount, VI00, VU1_SKYBOX_NEXT_CAM
    // Ok, we have something left to render.
    XGKICK pOutputBfr
    SwitchOutputBuffer pOutputBfr
    B VU1_SKYBOX_NEXT_CAM

VU1_SKYBOX_CLIP:
    
    // We've already setup this camera and we don't have any extra
    // renderstate setup rubbish so we just need a single GIF tag to start
    // off the output buffer.
    LQ.yzw fGIFTag, (ASURA_PS2_SKYBOX_ADDR_GIF_TAGS+1)(VI00)
    SQ.yzw fGIFTag, 0(pOutputBfr)
    // Overwrite NLOOP. We'll increment it as the clipper outputs more
    // tristrips.
    IADDIU iEOPAndZeroNLOOP, VI00,             0x7FFF
    IADDIU iEOPAndZeroNLOOP, iEOPAndZeroNLOOP, 1
    ISW.x  iEOPAndZeroNLOOP, 0(pOutputBfr)
    // Setup an output ptr.
    IADDIU pOut, pOutputBfr, 1
    // Pack some integer registers into a float one to relieve register
    // pressure.
    MFIR.x fStack0, pRotMat
    MFIR.y fStack0, iVisBit
    MFIR.z fStack0, iCamLpCtr
    MFIR.w fStack0, pOut
    // Setup input ptr for clipper.
    XTOP pIn
    // We also need to read ADC bits from the main render loop's output.
    // Note that this is slightly risky since in theory we could fill our
    // output buffer in the clipper and then start overwriting this data.
    // However, we would need to generate a massive number of clipped tris
    // for this to happen - something like five or six vertices for *every*
    // input tri and I don't think that can happen in any real situation.
    GetOtherOutputBuffer pOutputBfr, pMainRenderOutput
    // Skip over GIF tags and GS setup QWs.
    IADDIU pMainRenderOutput, pMainRenderOutput, 4
    // Instead of a loop ctr, we'll use an end-of-input ptr and compare to
    // that.
    ILW.w iNumVerts, 1(pIn)
    IADD pInEnd, pIn, iNumVerts
    IADD pInEnd, pInEnd, iNumVerts
    // Setup a couple of constants we'll need below.
    IADDIU iClipMask, VI00, 0x3F
    IADDIU iNonADC,   VI00, 0x7FFF
    // We'll also be needing the guardbandclip->screenclip scale vector.
    LQ fGC2SCScale, 6(pCamData)
    // Get screenclipspace pos for vtx 0.
    LQ.xyz fXYZ, 1(pIn)
    MatrixMultiplyVertexW1 fGClipXYZ, mW2GC, fXYZ
    MUL fSClipXYZ, fGClipXYZ, fGC2SCScale
    // Generate clip results for it.
    CLIPw.xyz fSClipXYZ, fSClipXYZ
    FCGET     iClipFlag2
    IAND      iClipFlag2, iClipFlag2, iClipMask
    // Get screenclipspace pos for vtx 0.
    LQ.xyz fXYZ, 3(pIn)
    MatrixMultiplyVertexW1 fGClipXYZ, mW2GC, fXYZ
    MUL fSClipXYZ, fGClipXYZ, fGC2SCScale
    // Generate clip results for it.
    CLIPw.xyz fSClipXYZ, fSClipXYZ
    FCGET     iClipFlag1
    IAND      iClipFlag1, iClipFlag1, iClipMask
    // Skip ahead to third vtx.
    IADDIU pIn,               pIn,               (2 * ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_IN)
    IADDIU pMainRenderOutput, pMainRenderOutput, (2 * ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT)
    // Calculate half of first clipping test.
    IAND iClipAll21, iClipFlag2, iClipFlag1

FIND_CLIP_TRIANGLE_LOOP:

        --LoopCS 1, 0

        // Have we finished yet ?
        IBEQ pIn, pInEnd, VU1_SKYBOX_CLIP_FINISHED
        // Calculate screenclipspace pos for cur vtx.
        LQ.xyz fXYZ, 1(pIn)
        MatrixMultiplyVertexW1 fGClipXYZ, mW2GC, fXYZ
        MUL fSClipXYZ, fGClipXYZ, fGC2SCScale
        // And get clipping results.
        CLIPw.xyz fSClipXYZ, fSClipXYZ
        FCGET     iClipFlag0
        IAND      iClipFlag0, iClipFlag0, iClipMask
        // Calculate second half of clipping test.
        IAND iClipAll10, iClipFlag1, iClipFlag0
        IAND iClipAll,   iClipAll21, iClipAll10
        // Move clipping flags/result along.
        IADD iClipFlag1, VI00, iClipFlag0
        IADD iClipAll21, VI00, iClipAll10
        // Update loop state.
        IADDIU pIn,               pIn,               ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_IN
        IADDIU pMainRenderOutput, pMainRenderOutput, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT
        // Ignore vertices that we accepted in the main render loop above.
        // Otherwise, we'll think we have to clip polys here that we've
        // already rendered, since we're checking against smaller bounds.
        ILW.w iADC, -1(pMainRenderOutput)
        IBEQ  iADC, iNonADC, FIND_CLIP_TRIANGLE_LOOP // Will be 0x8000 if rejected by main render loop or else 0x7FFF.
        // Ignore degenerate vertices.
        ILW.w iADC, -2(pIn)
        IBNE  iADC, iNonADC, FIND_CLIP_TRIANGLE_LOOP // Will be 0x8000 if ADC on or else 0x7FFF.
        // Does this tri need clipping?
        FCAND VI01,     0x3ffff
        IBEQ  VI01,     VI00, FIND_CLIP_TRIANGLE_LOOP // Tri is completely inside clip bounds.
        IBNE  iClipAll, VI00, FIND_CLIP_TRIANGLE_LOOP // Tri is completely outside at least one clip plane.

    // Ok, this triangle needs clipping.
    // Pack some integer registers into a float one to relieve register
    // pressure.
    MFIR.x fStack1, iVisFlags
    MFIR.y fStack1, pInEnd
    MFIR.z fStack1, iNonADC
    MFIR.w fStack1, iClipFlag1
    MFIR.x fStack2, iClipAll21
    MFIR.y fStack2, pMainRenderOutput
    
--barrier

CLIP_TRIANGLE:
    
    // Setup the clipping variables
    IADDIU pClipSrc,  VI00, ASURA_PS2_SKYBOX_ADDR_CLIP_WORKSP_0
    IADDIU pClipDst,  VI00, ASURA_PS2_SKYBOX_ADDR_CLIP_WORKSP_1
    IADDIU pClipSwap, VI00, (ASURA_PS2_SKYBOX_ADDR_CLIP_WORKSP_0 + ASURA_PS2_SKYBOX_ADDR_CLIP_WORKSP_1)
    IADDIU iSrcPtQty, VI00, 3
    // Copy src vertices into first clip workspace. Do third vtx first
    // because we've already got the guardbandclipspace pos for that.
    SQ       fGClipXYZ,  4(pClipSrc)
    LQ.xyz   fSTQ,      -2(pIn)
    ITOF4.xy fSTQ,       fSTQ
    SQ.xyz   fSTQ,       5(pClipSrc)
    // Generate guardbandclipspace pos for vtx 0.
    LQ.xyz fXYZ, -5(pIn)
    MatrixMultiplyVertexW1 fGClipXYZ, mW2GC, fXYZ
    // And copy that and STQ to clip workspace.
    SQ       fGClipXYZ,  0(pClipSrc)
    LQ.xyz   fSTQ,      -6(pIn)
    ITOF4.xy fSTQ,       fSTQ
    SQ.xyz   fSTQ,       1(pClipSrc)
    // Generate guardbandclipspace pos for vtx 1.
    LQ.xyz fXYZ, -3(pIn)
    MatrixMultiplyVertexW1 fGClipXYZ, mW2GC, fXYZ
    // And copy that and STQ to clip workspace.
    SQ       fGClipXYZ,  2(pClipSrc)
    LQ.xyz   fSTQ,      -4(pIn)
    ITOF4.xy fSTQ,       fSTQ
    SQ.xyz   fSTQ,       3(pClipSrc)

    //----------------------------------
    // Clip to the near plane (Z -ve)
    //----------------------------------
    IADDIU    iCurrentClipMask, VI00, CLIP_FLAG_NEAR        // Setup the current clip mask
    SUBw.xw    fPlaneSign,    VF00, VF00                            // Clear destination vertex count and set plane sign to -ve
//    IADDIU    pInterpolate, VI00, INTERPOLATE_Z                // No VCL JUMP TABLE SUPPORT :o(
    IADDIU    pInterpolate, VI00, 0                            // No VCL JUMP TABLE SUPPORT :o(
    BAL        pRtA, CLIP_POINTS                                // Clip the points
    IADDIU pClipSrc, pClipDst,  0
    ISUB   pClipDst, pClipSwap, pClipSrc
    FTOI0.w    fPlaneSign, fPlaneSign                            //    and counters for next clip
    MTIR    iSrcPtQty, fPlaneSign[w]
    IBEQ    iSrcPtQty, VI00,    CLIP_FINISHED_THIS_TRIANGLE    // All the points have been clipped away...

    //----------------------------------
    // Clip to the far plane (Z +ve)
    //----------------------------------
CLIP_FAR_PLANE:
    IADDIU    iCurrentClipMask, VI00, CLIP_FLAG_FAR        // Setup the current clip mask
    ADDw.x    fPlaneSign,    VF00, VF00                            // Set plane sig to +ve
    SUB.w    fPlaneSign, VF00, VF00                            // Clear vertex count
//    IADDIU    pInterpolate, VI00, INTERPOLATE_Z                // No VCL JUMP TABLE SUPPORT :o(
    IADDIU    pInterpolate, VI00, 0                            // No VCL JUMP TABLE SUPPORT :o(
    BAL        pRtA, CLIP_POINTS                                // Clip the points
    IADDIU pClipSrc, pClipDst,  0
    ISUB   pClipDst, pClipSwap, pClipSrc
    FTOI0.w    fPlaneSign, fPlaneSign                            //    and counters for next clip
    MTIR    iSrcPtQty, fPlaneSign[w]
    IBEQ    iSrcPtQty, VI00,    CLIP_FINISHED_THIS_TRIANGLE    // All the points have been clipped away...

    //----------------------------------
    // Clip to the top plane (Y -ve)
    //----------------------------------
CLIP_TOP_PLANE:
    IADDIU    iCurrentClipMask, VI00, CLIP_FLAG_TOP        // Setup the current clip mask
    SUBw.xw    fPlaneSign,    VF00, VF00                            // Clear destination vertex count and set plane sign to -ve
//    IADDIU    pInterpolate, VI00, INTERPOLATE_Y                // No VCL JUMP TABLE SUPPORT :o(
    IADDIU    pInterpolate, VI00, 1                            // No VCL JUMP TABLE SUPPORT :o(
    BAL        pRtA, CLIP_POINTS                                // Clip the points
    IADDIU pClipSrc, pClipDst,  0
    ISUB   pClipDst, pClipSwap, pClipSrc
    FTOI0.w    fPlaneSign, fPlaneSign                            //    and counters for next clip
    MTIR    iSrcPtQty, fPlaneSign[w]
    IBEQ    iSrcPtQty, VI00,    CLIP_FINISHED_THIS_TRIANGLE    // All the points have been clipped away...

    //----------------------------------
    // Clip to the bottom plane (Y +ve)
    //----------------------------------
CLIP_BOTTOM_PLANE:
    IADDIU    iCurrentClipMask, VI00, CLIP_FLAG_BOTTOM    // Setup the current clip mask
    ADDw.x    fPlaneSign,    VF00, VF00                            // Set plane sign to +ve
    SUB.w    fPlaneSign, VF00, VF00                            // Clear destination vertex counter
//    IADDIU    pInterpolate, VI00, INTERPOLATE_Y                // No VCL JUMP TABLE SUPPORT :o(
    IADDIU    pInterpolate, VI00, 1                            // No VCL JUMP TABLE SUPPORT :o(
    BAL        pRtA, CLIP_POINTS                                // Clip the points
    IADDIU pClipSrc, pClipDst,  0
    ISUB   pClipDst, pClipSwap, pClipSrc
    FTOI0.w    fPlaneSign, fPlaneSign                            //    and counters for next clip
    MTIR    iSrcPtQty, fPlaneSign[w]
    IBEQ    iSrcPtQty, VI00,    CLIP_FINISHED_THIS_TRIANGLE    // All the points have been clipped away...

    //----------------------------------
    // Clip to the left plane (X -ve)
    //----------------------------------
CLIP_LEFT_PLANE:
    IADDIU    iCurrentClipMask, VI00, CLIP_FLAG_LEFT        // Setup the current clip mask
    SUBw.xw    fPlaneSign,    VF00, VF00                            // Set plane sign to -ve and clear destination vertex count
//    IADDIU    pInterpolate, VI00, INTERPOLATE_X                // No VCL JUMP TABLE SUPPORT :o(
    IADDIU    pInterpolate, VI00, 2                            // No VCL JUMP TABLE SUPPORT :o(
    BAL        pRtA, CLIP_POINTS                                // Clip the points
    IADDIU pClipSrc, pClipDst,  0
    ISUB   pClipDst, pClipSwap, pClipSrc
    FTOI0.w    fPlaneSign, fPlaneSign                            //    and counters for next clip
    MTIR    iSrcPtQty, fPlaneSign[w]
    IBEQ    iSrcPtQty, VI00,    CLIP_FINISHED_THIS_TRIANGLE    // All the points have been clipped away...

    //----------------------------------
    // Clip to the right plane (X +ve)
    //----------------------------------
CLIP_RIGHT_PLANE:
    IADDIU    iCurrentClipMask, VI00, CLIP_FLAG_RIGHT    // Setup the current clip mask
    ADDw.x    fPlaneSign,    VF00, VF00                            // Set plane sign to +ve
    SUB.w    fPlaneSign, VF00, VF00                            // Clear destination vertex count
//    IADDIU    pInterpolate, VI00, INTERPOLATE_X                // No VCL JUMP TABLE SUPPORT :o(
    IADDIU    pInterpolate, VI00, 2                            // No VCL JUMP TABLE SUPPORT :o(
    BAL        pRtA, CLIP_POINTS                                // Clip the points
    IADDIU    pClipSrc, pClipDst, 0                            // Setup the source pointer,
    FTOI0.w    fPlaneSign, fPlaneSign                            //    and counters
    MTIR    iSrcPtQty, fPlaneSign[w]                
    IBEQ    iSrcPtQty, VI00,    CLIP_FINISHED_THIS_TRIANGLE    // All the points have been clipped away...

    //-----------------------------------------------
    // Finished clipping - Project in to screen space
    //-----------------------------------------------

CLIP_FINISHED_PLANES:

    // Retrieve int registers from stack.
    MTIR pOut, fStack0[w]
    // Check there is enough space in the output buffer for this many extra
    // vertices.
    ISUB   iTemp, pOut,  pOutputBfr                           // Space used in buffer already.
    IADD   iTemp, iTemp, iSrcPtQty
    IADD   iTemp, iTemp, iSrcPtQty
    IADD   iTemp, iTemp, iSrcPtQty                            // Plus space required for this strip. Assuming ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT is 3 here :-/
    IADDIU iTemp, iTemp, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT // Plus space for overrun (see comment below.)
    ISUBIU iTemp, iTemp, ASURA_PS2_SKYBOX_SIZE_OUTPUT_BFR     // Now <= 0 if there is space for this strip
    IBLEZ  iTemp, CLIP_ENOUGH_SPACE_IN_OUTPUT
    // Not enough space for this strip. Send what we have and swap output
    // buffers.
    XGKICK pOutputBfr
    SwitchOutputBuffer pOutputBfr
    // Copy GIF tag into output buffer.
    LQ.yzw fGIFTag, (ASURA_PS2_SKYBOX_ADDR_GIF_TAGS+1)(VI00)
    SQ.yzw fGIFTag, 0(pOutputBfr)
    // Overwrite NLOOP. We'll increment it as the clipper outputs more
    // tristrips.
    IADDIU iEOPAndZeroNLOOP, VI00,             0x7FFF
    IADDIU iEOPAndZeroNLOOP, iEOPAndZeroNLOOP, 1
    ISW.x  iEOPAndZeroNLOOP, 0(pOutputBfr)
    // Reset output ptr.
    IADDIU pOut, pOutputBfr, 1
    
CLIP_ENOUGH_SPACE_IN_OUTPUT:
    // Increment NLOOP count in GIF tag.
    ILW.x iNLOOP, 0(pOutputBfr)
    IADD  iNLOOP, iNLOOP, iSrcPtQty
    ISW.x iNLOOP, 0(pOutputBfr)
    // Setup the end-of-input ptr.
    IADD   pClipSrcE, pClipSrc,  iSrcPtQty                             // Assuming ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP is 2 here :-/
    IADD   pClipSrcE, pClipSrcE, iSrcPtQty
    ISUBIU pClipSrcE, pClipSrcE, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP // Last vertex in fan.
    // Set degeneracy flag for first two output verts. We'll add this value
    // to itself each time; that'll give us 0xC000, 0x8000, 0x0000 (and
    // obviously 0s thereafter), which means the ADC bit is on for precisely
    // the first two vertices.
    IADDIU    iADC,    VI00,    0x6000
#if NORMALISE_CLIPPED_STS
    // Calculate ST normalising offset.
    LOI 0.5
    ADD.xy   fNormST, fMinSTs, fMaxSTs
    MULi.xy  fNormST, fNormST, I
    FTOI0.xy fNormST, fNormST
    ITOF0.xy fNormST, fNormST
#endif

CLIP_OUTPUT_STRIP_LOOP:

        // (MP - 21.08.2003) We'll always write a multiple of two vertices
        // here to save branching in the loop. This then allows VCL to
        // optimise it for us. Of course, this means we'll overrun by up to
        // one vertex so we have to account for this when checking for free
        // space above. The GIF tag's NLOOP won't count this extra vertex but
        // we'll have to fix up the output ptr after the loop.

        --LoopCS  1, 0
        --LoopAbs 22

        // Load vtx from the front.
        LQ     fXYZ, 0(pClipSrc)
        LQ.xyz fSTQ, 1(pClipSrc)
#if NORMALISE_CLIPPED_STS
        // Normalise ST.
        SUB.xy fSTQ, fSTQ, fNormST
#endif
        // Transform to screenspace.
        MULA  ACC,     fXYZ,       fGC2SScale
        MADDw fScrXYZ, fGC2STrans, fXYZ[w]
        // Perspective divide.
        DIV Q, VF00[w], fScrXYZ[w]
        MULq.xyz fScrXYZ, fScrXYZ, Q
        MULq.xyz fSTQ,    fSTQ,    Q
        // Convert XYZ to GS fixed point format.
        FTOI4.xyz fScrXYZ, fScrXYZ
        // Output.
        SQ.xyz  fSTQ,     0(pOut)
        SQ      fRGBA,    1(pOut)
        SQ.xyz  fScrXYZ,  2(pOut)
        // Set the ADC bit if necessary.
        IADD  iADC, iADC, iADC
        ISW.w iADC, 2(pOut)
        // Update loop state.
        IADDIU pClipSrc, pClipSrc, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
        IADDIU pOut,     pOut,     ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT
        // Load vtx from the back.
        LQ     fXYZ, 0(pClipSrcE)
        LQ.xyz fSTQ, 1(pClipSrcE)
#if NORMALISE_CLIPPED_STS
        // Normalise ST.
        SUB.xy fSTQ, fSTQ, fNormST
#endif
        // Transform to screenspace.
        MULA  ACC,     fXYZ,       fGC2SScale
        MADDw fScrXYZ, fGC2STrans, fXYZ[w]
        // Perspective divide.
        DIV Q, VF00[w], fScrXYZ[w]
        MULq.xyz fScrXYZ, fScrXYZ, Q
        MULq.xyz fSTQ,    fSTQ,    Q
        // Convert XYZ to GS fixed point format.
        FTOI4.xyz fScrXYZ, fScrXYZ
        // Output.
        SQ.xyz  fSTQ,     0(pOut)
        SQ      fRGBA,    1(pOut)
        SQ.xyz  fScrXYZ,  2(pOut)
        // Set the ADC bit if necessary.
        IADD  iADC, iADC, iADC
        ISW.w iADC, 2(pOut)
        // Update loop state.
        ISUBIU pClipSrcE, pClipSrcE, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
        IADDIU pOut,      pOut,      ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT
        // Check if we've finished yet.
        ISUB  iTemp, pClipSrcE, pClipSrc
        IBGEZ iTemp, CLIP_OUTPUT_STRIP_LOOP

CLIP_FINISHED_OUTPUT_THIS_TRIANGLE:

    // Did we have an odd number of vertices?
    IADDIU iMask, VI00,      0x1
    IAND   iOdd,  iSrcPtQty, iMask
    IBEQ   iOdd,  VI00,      CLIP_SKIP_OUTPUT_PTR_FIX_UP
    // Yes. We need to correct the output ptr.
    ISUBIU pOut, pOut, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT

CLIP_SKIP_OUTPUT_PTR_FIX_UP:

    // Pack some integer registers into a float one to relieve register
    // pressure.
    MFIR.w fStack0, pOut

CLIP_FINISHED_THIS_TRIANGLE:

    // Retrieve int registers from stack.
    MTIR iVisFlags,         fStack1[x]
    MTIR pInEnd,            fStack1[y]
    MTIR iNonADC,           fStack1[z]
    MTIR iClipFlag1,        fStack1[w]
    MTIR iClipAll21,        fStack2[x]
    MTIR pMainRenderOutput, fStack2[y]
    // Carry out previous 2 clips again so the clip find triangle loop has
    // the correct clip history.
    LQ.xyz fXYZ, -3(pIn)
    MatrixMultiplyVertexW1 fGCXYZ, mW2GC, fXYZ
    MUL fSCXYZ, fGCXYZ, fGC2SCScale
    CLIPw.xyz fSCXYZ, fSCXYZ
    LQ.xyz fXYZ, -1(pIn)
    MatrixMultiplyVertexW1 fGCXYZ, mW2GC, fXYZ
    MUL fSCXYZ, fGCXYZ, fGC2SCScale
    CLIPw.xyz fSCXYZ, fSCXYZ
    // Restore clip mask, to save VCL preserving the register value for us.
    IADDIU iClipMask, VI00, 0x3F
    // And look for more tris to clip.
    B FIND_CLIP_TRIANGLE_LOOP

--barrier

CLIP_POINTS:

    // (MP - 19.08.2003) We work in guardbandclipspace in general. However,
    // our clipping logic works in the screenclipspace, so we don't end up
    // sending large bits of polys that are outside the screen bounds. Hence
    // the following code keeps tracks of the current and next positions in
    // both spaces.

    // Keep track of the start of the input buffer so we can add the first
    // vertex to the end of the list at the end.
    IADDIU pClipSrcStart, pClipSrc, 0
    // Create an output pointer.
    IADDIU pClipOut, pClipDst, 0
    // Load the first vtx from the buffer.
    LQ     fNxtGCXYZ, 0(pClipSrc)
    LQ.xyz fNxtSTQ,   1(pClipSrc)
    // Update input ptr.
    IADDIU pClipSrc, pClipSrc, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
#if NORMALISE_CLIPPED_STS
    // Track min and max STs so we can normalise them post-clipping.
    ADD.xy fMinSTs, fNxtSTQ, VF00
    ADD.xy fMaxSTs, fNxtSTQ, VF00
#endif
    // Generate clip result.
    MUL       fNxtSCXYZ, fNxtGCXYZ, fGC2SCScale
    CLIPw.xyz fNxtSCXYZ, fNxtSCXYZ
    // Check if it's in or out.
    FCGET VI01
    IAND  iTemp, VI01, iCurrentClipMask
    IBEQ  iTemp, VI00, CLIP_CURRENT_IN

CLIP_CURRENT_OUT:

    // Current point is out. Any more points?
    ISUBIU iSrcPtQty, iSrcPtQty, 1
    IBEQ   iSrcPtQty, VI00, CLIP_POINTS_END
    // Move next point to current point.
    MOVE     fCurGCXYZ, fNxtGCXYZ
    MOVE     fCurSCXYZ, fNxtSCXYZ
    MOVE.xyz fCurSTQ,   fNxtSTQ
    // Load the next point.
    LQ     fNxtGCXYZ, 0(pClipSrc)
    LQ.xyz fNxtSTQ,   1(pClipSrc)
    // Update input ptr.
    IADDIU pClipSrc, pClipSrc, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
#if NORMALISE_CLIPPED_STS
    // Track min and max STs so we can normalise them post-clipping.
    ADD.xy fMinSTs, fNxtSTQ, VF00
    ADD.xy fMaxSTs, fNxtSTQ, VF00
#endif
    // Generate clip result.
    MUL       fNxtSCXYZ, fNxtGCXYZ, fGC2SCScale
    CLIPw.xyz fNxtSCXYZ, fNxtSCXYZ
    // Check if it's in or out.
    FCGET VI01
    IAND  iTemp, VI01, iCurrentClipMask
    IBEQ  iTemp, VI00, CLIP_OUT_IN

CLIP_OUT_OUT:

    // Carry on to next point.
    B CLIP_CURRENT_OUT

CLIP_OUT_IN:

    // Adding 1 to destination point count
    ADD.w fPlaneSign, fPlaneSign, VF00
    // Write INTERPOLATED - Setup retrun address to jump to the next point
//    IADDIU    pRtA2, VI00, CLIP_CURRENT_IN
//    JR        pInterpolate
    // Hack until VCL support JumpTables or I re write it in VSM
    BAL pRtA2, INTERPOLATE
    B   CLIP_CURRENT_IN

CLIP_CURRENT_IN:

    // Current point is in. Any more points?
    ISUBIU iSrcPtQty, iSrcPtQty, 1
    IBEQ   iSrcPtQty, VI00, CLIP_POINTS_END
    // Move next point to current point
    MOVE     fCurGCXYZ, fNxtGCXYZ
    MOVE     fCurSCXYZ, fNxtSCXYZ
    MOVE.xyz fCurSTQ,   fNxtSTQ
    // Load the next point
    LQ     fNxtGCXYZ, 0(pClipSrc)
    LQ.xyz fNxtSTQ,   1(pClipSrc)
    // Update input ptr.
    IADDIU pClipSrc, pClipSrc, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
#if NORMALISE_CLIPPED_STS
    // Track min and max STs so we can normalise them post-clipping.
    ADD.xy fMinSTs, fNxtSTQ, VF00
    ADD.xy fMaxSTs, fNxtSTQ, VF00
#endif
    // Generate clip result.
    MUL       fNxtSCXYZ, fNxtGCXYZ, fGC2SCScale
    CLIPw.xyz fNxtSCXYZ, fNxtSCXYZ
    // Check if it's in or out.
    FCGET VI01
    IAND  iTemp, VI01, iCurrentClipMask
    IBNE  iTemp, VI00, CLIP_IN_OUT

CLIP_IN_IN:

    // Adding 1 to destination point count
    ADD.w fPlaneSign, fPlaneSign, VF00
    // Write current.
    SQ     fCurGCXYZ, 0(pClipOut)
    SQ.xyz fCurSTQ,   1(pClipOut)
    // Update output ptr.
    IADDIU pClipOut, pClipOut, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
    // Carry on to next point.
    B CLIP_CURRENT_IN

CLIP_IN_OUT:

    // Adding 2 to destination point count
    ADD.w fPlaneSign, fPlaneSign, VF00
    ADD.w fPlaneSign, fPlaneSign, VF00
    // Write current.
    SQ     fCurGCXYZ, 0(pClipOut)
    SQ.xyz fCurSTQ,   1(pClipOut)
    // Update output ptr.
    IADDIU pClipOut, pClipOut, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
    // Write INTERPOLATED - Setup retrun address to jump to the next point
//    IADDIU    pRtA2, VI00, CLIP_CURRENT_OUT
//    JR        pInterpolate
    // Hack until VCL support JumpTables or I re write it in VSM
    BAL pRtA2, INTERPOLATE
    B   CLIP_CURRENT_OUT

CLIP_POINTS_END:

    // Check if we've done the final edge yet, ie. last vtx -> first.
    IBNE pClipSrcStart, VI00, CLIP_POINT_LAST_SEGMENT
    JR   pRtA

CLIP_POINT_LAST_SEGMENT:
    
    // Get the start of the input buffer and clear the input buffer pointer
    // so we know we've finished.
    IADDIU pClipSrc,      pClipSrcStart, 0
    IADDIU pClipSrcStart, VI00,          0
    // One final segment to do.
    IADDIU iSrcPtQty, iSrcPtQty, 2
    // Check if the last clip was in or out
    FCGET VI01
    IAND  iTemp, VI01, iCurrentClipMask
    IBEQ  iTemp, VI00, CLIP_CURRENT_IN
    B CLIP_CURRENT_OUT

--barrier

    //         |-------L-------|
    //        C ----------|---- N
    //                    P
    //         |----A-----l--B-|
    //                    a
    //                    n
    //                    e
    //
    //    we need to find the ratio R = |A/L|

    // (MP - 19.08.2003) Find R based on screenclipspace positions but
    // actually interpolate the guardbandclipspace ones.

    // Hack until VCL support JumpTables or I re write it in VSM
INTERPOLATE:
    IBEQ   pInterpolate, VI00, INTERPOLATE_Z
    ISUBIU iInt, pInterpolate, 1
    IBEQ   iInt, VI00, INTERPOLATE_Y
    // Interpolate the current line segment against the clip plane
INTERPOLATE_X:
    // Get distance from clip plane to current (taking into account -/+ clip plane)        (-A in above diagram)
    ADDAx.x ACC, VF00, fCurSCXYZ              // ACC.x = cur.x
    MSUBw.x fC2Plane, fPlaneSign, fCurSCXYZ   // fC2Plane.x = cur.x - fPlaneSign * cur.w
    // Get distance from next to clip plane (taking into account -/+ clip plane)            (B in above diagram)
    ADDAx.x ACC, VF00, fNxtSCXYZ              // ACC.x = next.x
    MSUBw.x fN2Plane, fPlaneSign, fNxtSCXYZ   // fN2Plane.x = next.x - fPlaneSign * next.w
    B INTERPOLATE_COMMON
INTERPOLATE_Y:
    // Get distance from clip plane to current (taking into account -/+ clip plane)        (-A in above diagram)
    ADDAy.x ACC, VF00, fCurSCXYZ              // ACC.x = cur.y
    MSUBw.x fC2Plane, fPlaneSign, fCurSCXYZ   // fC2Plane.x = cur.y - fPlaneSign * cur.w
    // Get distance from next to clip plane (taking into account -/+ clip plane)            (B in above diagram)
    ADDAy.x ACC, VF00, fNxtSCXYZ              // ACC.x = next.y
    MSUBw.x fN2Plane, fPlaneSign, fNxtSCXYZ   // fN2Plane.x = next.y - fPlaneSign * next.w
    B INTERPOLATE_COMMON
INTERPOLATE_Z:
    // Get distance from clip plane to current (taking into account -/+ clip plane)        (-A in above diagram)
    ADDAz.x ACC, VF00, fCurSCXYZ              // ACC.x = cur.z
    MSUBw.x fC2Plane, fPlaneSign, fCurSCXYZ   // fC2Plane.x = cur.z - fPlaneSign * cur.w
    // Get distance from next to clip plane (taking into account -/+ clip plane)            (B in above diagram)
    ADDAz.x ACC, VF00, fNxtSCXYZ              // ACC.x = next.z
    MSUBw.x fN2Plane, fPlaneSign, fNxtSCXYZ   // fN2Plane.x = next.z - fPlaneSign * next.w
//    B INTERPOLATE_COMMON
INTERPOLATE_COMMON:
    // Calculate L = B + A = B - -A
    SUB.x fDiff, fN2Plane, fC2Plane
    // Calculate R = |A/L|
    DIV Q, fC2Plane[x], fDiff[x]
    ADDq.x fR, VF00, Q
    ABS.x  fR, fR
    // Create a new vertex from the old verts and the ratio (New = Current + (Next-Current)*R)
    // XYZ
    ADDAx  ACC,     fCurGCXYZ, VF00      // = Current
    MADDAx ACC,     fNxtGCXYZ, fR        // = Current + (Next*R)
    MSUBx  fNewGCXYZ, fCurGCXYZ, fR      // = Current + (Next*R) - (Current*R)
    SQ     fNewGCXYZ, 0(pClipOut)
    // ST
    ADDAx.xy  ACC,     fCurSTQ, VF00      // = Current
    MADDAx.xy ACC,     fNxtSTQ, fR        // = Current + (Next*R)
    MSUBx.xy  fNewST,  fCurSTQ, fR        // = Current + (Next*R) - (Current*R)
    SQ.xy     fNewST,  1(pClipOut)
    SQ.z      fCurSTQ, 1(pClipOut)
    // Update output ptr.
    IADDIU pClipOut, pClipOut, ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP
    // Done.
    JR pRtA2

.end
