//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_ShellMCUI.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description: Asura Interface to the Sony Memory Card User Interface 
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// this file's header
#include "ShellMCUI/Asura_PS2_ShellMCUI.h"

// Engine independent
#include "Core/Asura_Handle.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_LocalisedText.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"

// project includes
#include "ASURA_PS2_MCUI_Text_Enum.h"	// need to have path to this file set in project settings.

// Engine platform dependent
#include "Sound/Asura_PS2_Multistream.h"

// middleware
#include "Sony/PS2Shell/shellMCUI.h"

// G L O B A L S /////////////////////////////////////////////////////////////

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
bool	Asura_PS2_ShellMCUI::s_bInitialised = false;

char	Asura_PS2_ShellMCUI::s_acKeyCode[2];
char	Asura_PS2_ShellMCUI::s_acProductCode[11];
char	Asura_PS2_ShellMCUI::s_acFileName[7];		
char*	Asura_PS2_ShellMCUI::s_pcIconBuffer = NULL; 

sceMcIconSys	Asura_PS2_ShellMCUI::s_xIconSysFile;

ShellMCUIOutputs	Asura_PS2_ShellMCUI::s_xOutputParms;
ShellMCUIInputs 	Asura_PS2_ShellMCUI::s_xInputParms;

EShellMCUIScreen 	Asura_PS2_ShellMCUI::s_eScreen		= SHELLMCUI_SCREEN_NULL;
EShellMCUIScreen 	Asura_PS2_ShellMCUI::s_ePrevScreen	= SHELLMCUI_SCREEN_NULL;

bool	Asura_PS2_ShellMCUI::s_bCardChanged		= false;
bool	Asura_PS2_ShellMCUI::s_bCardWasInserted = false;




bool Asura_PS2_ShellMCUI::Initialise( void )
{
	ShellMCUIInitParms	xInitParms;

	if (s_bInitialised)
	{
		Asura_Assert(false, "Trying to initialise Asura_PS2_ShellMCUI twice");
		return false;
	}

	const char* szProductCode = Asura_PS2_ShellMCUI::Project_GetProductCode();
	
	// Ok, lets check the validity of our product code and set the correct region
	if (strlen(szProductCode) != 10)
	{
		Asura_Assert(false, "Invalid string length for product code");
		return false;
	}

	switch(szProductCode[0] + szProductCode[1] + szProductCode[2] + szProductCode[3])
	{
		// Add the key id based on product code and copy string across
		case ('S' + 'L' + 'E' + 'S'):			// Europe, Middle East and Australasia
		{
			s_acKeyCode[0] = 'B';
			s_acKeyCode[1] = 'E';
			strcpy(s_acProductCode, szProductCode);
		}
		break;
		case ('S' + 'L' + 'P' + 'S'):			// Japan and Far East
		{
			s_acKeyCode[0] = 'B';
			s_acKeyCode[1] = 'I';
			strcpy(s_acProductCode, szProductCode);
		}
		break;
		case ('S' + 'L' + 'U' + 'S'):			// USA, Canada and South America
		{
			s_acKeyCode[0] = 'B';
			s_acKeyCode[1] = 'A';
			strcpy(s_acProductCode, szProductCode);
		}
		break;
		default:
			Asura_Assert(false, "Unknown region in product code");
			return false;
	}
    
	// (NRM - 17.07.2003) ensure that the last char is actually set to null terminate - paranoia after Dredd pre-sub
    s_acProductCode[10] = '\0';

	const int iMaxNumSavesPerCard = Asura_PS2_ShellMCUI::Project_GetMaxNumSavesPerCard();

	if (iMaxNumSavesPerCard <= 0)
	{
		Asura_Assert(false, "Max number of saves per card must be > 0");
		return false;
	}

	//Initialise shellMCUI:
	xInitParms.separateDirPerSave	= true;
	xInitParms.maxNumSavesPerCard 	= iMaxNumSavesPerCard;
	xInitParms.keyCode				= s_acKeyCode;
	xInitParms.productModelNo		= s_acProductCode;
	
	shellMCUIInit(&xInitParms);

	//Initialise xInputParms:
	s_xInputParms.saveParms.pIconSysData		= &s_xIconSysFile;
	s_xInputParms.userInput.userMainMenuOption	= SHELLMCUI_MAIN_MENU_OPTION_NULL;
	s_xInputParms.portNo						= 0;
	s_xInputParms.slotNo						= 0;

	s_bInitialised = true;

	return true;
}

void Asura_PS2_ShellMCUI::Shutdown()
{
	if (!s_bInitialised)
	{
		Asura_Assert(false, " Shell MCUI had not been initialised");
		return;
	}

	// If we have a memory card icon loaded, free it
	if (s_pcIconBuffer)
	{
		Asura_Deallocate(s_pcIconBuffer);
	}

	s_bInitialised = false;
}



bool Asura_PS2_ShellMCUI::SetMemoryCardIcon( const char* szIconName )
{
	if (!s_bInitialised)
	{
		Asura_Assert(false, "MCUI Shell has not yet been initialised\n");
		return false;
	}
	
	// Get the size of the icon
	Asura_Handle iFileID = Asura_FileAccess::Open(szIconName);
	u_int uFileSize;
	
	if (iFileID == ASURA_INVALID_HANDLE)
	{
		Asura_Assert(false, "Cannot open memory card icon file - memory card save will be non-TRC compliant\n");
		return false;
	}
	else
	{
		uFileSize = Asura_PS2_Multistream_FastLoad::GetFileSize(iFileID);
	
		if (s_pcIconBuffer)
		{
			Asura_Deallocate(s_pcIconBuffer);
		}

		s_pcIconBuffer = reinterpret_cast<char*>(Asura_Allocate(uFileSize));

		if (!s_pcIconBuffer)
		{
			Asura_Assert(false, "Unable to allocate memory for icon file memory\n");
			return false;
		}

		Asura_FileAccess::ReadData(iFileID, s_pcIconBuffer, uFileSize);
	    Asura_FileAccess::Close(iFileID);
	}

	// Set the icon files
	for (u_int i = 0; i < SHELLMEMCARD_ICON_NUM_ICONS; i++)
	{
		s_xInputParms.saveParms.pIconFileData[i]	= s_pcIconBuffer;
		s_xInputParms.saveParms.iconFileSize[i]	= uFileSize;
//		strcpy(s_xInputParms.saveParms.iconFileNames[i], szIconName);
		strcpy(s_xInputParms.saveParms.iconFileNames[i], "Icon1.ico");
	}		

	return true;
}

bool Asura_PS2_ShellMCUI::IsMainMenuLoadOptionAvailable( void )
{
	if (s_xOutputParms.mainMenuOptionsValid[SHELLMCUI_MAIN_MENU_OPTION_LOAD])
	{
		return true;
	}

	return false;
}


bool Asura_PS2_ShellMCUI::IsMainMenuSaveOptionAvailable( void )
{
	if (s_xOutputParms.mainMenuOptionsValid[SHELLMCUI_MAIN_MENU_OPTION_SAVE])
	{
		return true;
	}

	return false;
}


const int Asura_PS2_ShellMCUI::GetNumberOfSavesFound( void ) 
{
	if (s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LIST_SAVES_SUCCESSFUL])
	{
		if (s_xOutputParms.numSavesFound > Project_GetMaxNumSavesPerCard() )
		{
			return Project_GetMaxNumSavesPerCard();
		}
		
		return s_xOutputParms.numSavesFound;
	}

	return -1;
}


const int Asura_PS2_ShellMCUI::GetFreeSpaceOnCard( void ) 
{
	return s_xOutputParms.freeSpaceToDisplay;
}


const char* Asura_PS2_ShellMCUI::GetSaveNameForIndex( const int iIndex ) 
{
	if (iIndex < s_xOutputParms.numSavesFound)
	{
		char *pSaveName = s_xOutputParms.saveNames[iIndex];
		
		return (pSaveName + 12);	// Key code is 2 chars and product code is 10 chars
	}

	return NULL;
}


bool Asura_PS2_ShellMCUI::IsLoading( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LOADING];
}


bool Asura_PS2_ShellMCUI::LoadSuccessful( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LOAD_SUCCESSFUL];
}

bool Asura_PS2_ShellMCUI::LoadFailed( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LOAD_FAILED];
}


const char* Asura_PS2_ShellMCUI::GetLoadedDataPtrForIndex( const int iIndex ) 
{
	if ((iIndex < 0) || (iIndex >= ASURA_PS2_SHELLMCUI_MAX_NUM_DATA_FILES))
	{
		Asura_Assert(false, "Invalid index passed into GetLoadedDataPtrForIndex\n");
		return NULL;
	}

	if (s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LOAD_SUCCESSFUL])
	{
		Asura_Assert(s_xInputParms.pDataFileData[iIndex] != NULL, "Pointer to loaded data is NULL\n");
		return s_xInputParms.pDataFileData[iIndex];
	}

	return NULL;
}


void Asura_PS2_ShellMCUI::FreeLoadedDataPtrForIndex( const int iIndex ) 
{
	if ((iIndex < 0) || (iIndex >= ASURA_PS2_SHELLMCUI_MAX_NUM_DATA_FILES))
	{
		Asura_Assert(false, "Invalid index passed into GetLoadedDataPtrForIndex\n");
		return;
	}

	if (s_xInputParms.pDataFileData[iIndex] == NULL)
	{
		Asura_Assert(false, "Pointer to loaded data has already been free'd or held no data\n");
	}
	else
	{
		// Free up the data loading pointer for the next load
        Asura_Deallocate(s_xInputParms.pDataFileData[iIndex]);
		s_xInputParms.pDataFileData[iIndex] = NULL;
	}
}



const int Asura_PS2_ShellMCUI::GetLoadedDataSizeForIndex( const int iIndex ) 
{
	if ((iIndex < 0) || (iIndex >= ASURA_PS2_SHELLMCUI_MAX_NUM_DATA_FILES))
	{
		Asura_Assert(false, "Invalid index passed into GetLoadedDataSizeForIndex\n");
		return 0;
	}

	return s_xInputParms.dataFileSize[iIndex];
}

/////////////////////////////////////////////////////////////
//
// ResetAllDataPtrs(void)
//
// This should only be called after doing a save, since the 
// pointers have not been allocated, but need to be set to 
// NULL for the next data save or load. Use carefully.
//
/////////////////////////////////////////////////////////////

void Asura_PS2_ShellMCUI::ResetAllDataPtrs( void ) 
{
	for(int i=0; i<ASURA_PS2_SHELLMCUI_MAX_NUM_DATA_FILES; i++)
	{
		s_xInputParms.pDataFileData[i] = NULL;
	}
}


bool Asura_PS2_ShellMCUI::IsSaving( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_SAVING];
}


bool Asura_PS2_ShellMCUI::SaveSuccessful( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_SAVE_SUCCESSFUL];
}


bool Asura_PS2_ShellMCUI::SaveFailed( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_SAVE_FAILED];
}


bool Asura_PS2_ShellMCUI::IsFormatting( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_FORMATTING];
}


bool Asura_PS2_ShellMCUI::FormatSuccessful( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_FORMAT_SUCCESSFUL];
}


bool Asura_PS2_ShellMCUI::FormatFailed( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_FORMAT_FAILED];
}


bool Asura_PS2_ShellMCUI::IsCardDetecting( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_DETECTING];
}

bool Asura_PS2_ShellMCUI::IsOverwriteRequested( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_REQUEST_USER_OVERWRITE];
}

bool Asura_PS2_ShellMCUI::IsFormatRequested( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_REQUEST_USER_FORMAT];
}


bool Asura_PS2_ShellMCUI::IsCardPresent( void )
{
	const bool bCardInserted	= s_xOutputParms.statusFlags[SHELLMCUI_STATUS_PS2_CARD_INSERTED];
	const bool bCardUnformatted = s_xOutputParms.statusFlags[SHELLMCUI_STATUS_UNFORMATTED_CARD];

	if (bCardInserted || bCardUnformatted)
	{
		return true;
	}

	return false;
}


bool Asura_PS2_ShellMCUI::IsCardUnformatted( void )
{
	return s_xOutputParms.statusFlags[SHELLMCUI_STATUS_UNFORMATTED_CARD];
}

////////////////////////////////////////////////////////////////////////////////
// Autosaving titles need to know if a previously inserted card was changed. 
////////////////////////////////////////////////////////////////////////////////

void Asura_PS2_ShellMCUI::CheckForCardChanged( void )
{

	if (s_bCardWasInserted && (s_xOutputParms.cardStatus.cardChanged == 1))
	{
		s_bCardChanged = true;
	}

	s_bCardWasInserted  = s_xOutputParms.statusFlags[SHELLMCUI_STATUS_PS2_CARD_INSERTED];

}



void Asura_PS2_ShellMCUI::Update( void )
{

	if (!s_bInitialised)
	{
		Asura_Assert(false, "MCUI Shell has not yet been initialised\n");
		return;
	}

	s_eScreen = shellMCUIUpdate(&s_xOutputParms, &s_xInputParms);

	// Autosaving titles need to know if a card has been removed and the existing or a new
	// card reinserted since this function was last called.
	CheckForCardChanged();
	
	switch (s_eScreen)
	{
		case SHELLMCUI_SCREEN_MAIN:
		{
			UpdateMainScreen();	
		}			
		break;

		case SHELLMCUI_SCREEN_LIST_SAVES:
		{
			UpdateListSavesScreen();			
		}			
		break;
	
		case SHELLMCUI_SCREEN_LOAD:
		{
			UpdateLoadScreen();	
		}			
		break;

		case SHELLMCUI_SCREEN_SAVE:
		{
			UpdateSaveScreen();
		}			
		break;	

		default:
			Asura_Assert(false, "Invalid screen returned in Asura_PS2_ShellMCUI::Update\n");
	}				
	
	// Handle user inputs:
	UpdateUserInput();
	
//	MenuDisplayTitle(_pLanguageText->TEXT_SCREEN_TITLE[screen]);
//	MenuDisplayStatusMessages(&_outputParms);
//	MenuAppUpdateProgressBar(&_outputParms);

}



void Asura_PS2_ShellMCUI::UpdateMainScreen( void )
{
	//Display main menu options (load or save):
	if (IsMainMenuLoadOptionAvailable() && Project_LoadScreenRequested() )
	{
		s_xInputParms.userInput.userMainMenuOption = SHELLMCUI_MAIN_MENU_OPTION_LOAD;
	}
	else if (IsMainMenuSaveOptionAvailable() && Project_SaveScreenRequested() )
	{
		s_xInputParms.userInput.userMainMenuOption = SHELLMCUI_MAIN_MENU_OPTION_SAVE;
	}

}



void Asura_PS2_ShellMCUI::UpdateListSavesScreen( void )
{
	//Pass in list save parms:
	if (s_xOutputParms.requestData[SHELLMCUI_REQUEST_DATA_LIST_SAVE_PARMS])
	{
		strcpy(s_xInputParms.dirNameExtension, "");
		s_xInputParms.dataValid = true;		
	}

	//Display list of saves:
	//It is up to the app to determine which save dir corresponds to which 
	//displayed save:
	if (s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LIST_SAVES_SUCCESSFUL])
	{			
		// Quick paranoid check to see if number of saves on card exceeds whats allowed
		if (s_xOutputParms.numSavesFound > Project_GetMaxNumSavesPerCard() )
		{
			Asura_Assert(false, "Found more saves than allowed per card\n");
		}
		
		if (Project_ContinueToLoading() || Project_ContinueToSaving() )
		{
			s_xInputParms.userInput.userContinue = true;
		}
	}	
}



void Asura_PS2_ShellMCUI::UpdateLoadScreen( void )
{
	
	// Request load file parms
	if (s_xOutputParms.requestData[SHELLMCUI_REQUEST_DATA_LOAD_PARMS])
	{
		Asura_PS2_ShellMCUI_LoadParams xLoadParams;

		Project_GetLoadParams(xLoadParams);
		
		if ( SetLoadParams(xLoadParams) )
		{
			s_xInputParms.dataValid = true;
		}
	}

	//*********************************************************************	    				
	// For testing only:	
	//*********************************************************************
	if(s_xOutputParms.statusFlags[SHELLMCUI_STATUS_LOAD_SUCCESSFUL])
	{
	
	}
}




void Asura_PS2_ShellMCUI::UpdateSaveScreen( void )
{

	//If requested, pass in save file parms:
	if (s_xOutputParms.requestData[SHELLMCUI_REQUEST_DATA_SAVE_PARMS])
	{
		Asura_PS2_ShellMCUI_SaveParams xSaveParams;

		Project_GetSaveParams(xSaveParams);
		
		if ( SetSaveParams(xSaveParams) )
		{
			s_xInputParms.dataValid = true;
		}
	}

	//If insufficient space, display space reqd and space available.
	//(Ref "Suggested memory card (8MB) (for PlayStation2) messages" doc, 
	// section 3.0.2):
	if (s_xOutputParms.statusFlags[SHELLMCUI_STATUS_INSUFFICIENT_SPACE])
	{	
	
	
	} 
}


bool Asura_PS2_ShellMCUI::SetLoadParams( const Asura_PS2_ShellMCUI_LoadParams &xLoadParams )
{
	
	s_xInputParms.numDataFiles = xLoadParams.m_iNumDataFiles;

	strcpy(s_xInputParms.dirNameExtension, xLoadParams.m_szDirNameExtension);

	for (int i = 0; i < s_xInputParms.numDataFiles; i++)
	{
		Asura_Assert(s_xInputParms.pDataFileData[i] == NULL, "ShellMCUI pDataFileData ptrs are not NULL");
		strcpy(s_xInputParms.dataFileNames[i], xLoadParams.m_szDataFileNames[i]);
	}

	return true;
}



bool Asura_PS2_ShellMCUI::SetSaveParams( const Asura_PS2_ShellMCUI_SaveParams &xSaveParams )
{

	//Set dirNameExtension:
	strcpy(s_xInputParms.dirNameExtension, xSaveParams.m_szDirNameExtension);
	
	//Set user data file parms:
	s_xInputParms.saveParms.pUserFileData = xSaveParams.m_pcUserFileData;
	s_xInputParms.saveParms.userFileSize  = xSaveParams.m_iUserFileSize;
	
	//Set data file parms:
	s_xInputParms.numDataFiles = xSaveParams.m_iNumDataFiles;

	for (int i = 0; i < s_xInputParms.numDataFiles; i++)
	{	
		strcpy(s_xInputParms.dataFileNames[i], xSaveParams.m_szDataFileNames[i]); 
		
		s_xInputParms.dataFileSize[i]  = xSaveParams.m_iDataFileSizes[i];	
		s_xInputParms.pDataFileData[i] = xSaveParams.m_pcDataFileData[i];
	}	

	//Set icon sys data:
	SetIconSys(xSaveParams.m_szTitleName);

	return true;
}




void Asura_PS2_ShellMCUI::UpdateUserInput( void )
{
	// Get user required inputs:
	
	// Display continue option if requested:
	if (s_xOutputParms.requestUser[SHELLMCUI_REQUEST_USER_CONTINUE])
	{
		Project_RequestInputUserContinue();
		
		if (Project_GetInputUserContinue() )
		{
			s_xInputParms.userInput.userContinue = true;
		}					
	}

	// Cancel current menu if triangle pressed (where applicable):
	if (s_xOutputParms.requestUser[SHELLMCUI_REQUEST_USER_CANCEL])
	{
		Project_RequestInputUserCancel();

		if (Project_GetInputUserCancel() )
		{
			s_xInputParms.userInput.userCancel = true;
		}
	}

	// Display yes / no menu if requested:
	if (s_xOutputParms.requestUser[SHELLMCUI_REQUEST_USER_YESNO])
	{
		Project_RequestInputUserYesNo();

		if (Project_GetInputUserYes() )
		{
			s_xInputParms.userInput.userYesNo = SHELLMCUI_YESNO_YES;
		}
		else if (Project_GetInputUserNo() )
		{
			s_xInputParms.userInput.userYesNo = SHELLMCUI_YESNO_NO;
		}
	}	
}




void Asura_PS2_ShellMCUI::SetIconSys( const char *szTitleName )
{

	const int aiBgcolor[4][4] = 
	{	
		{ 0x80,    0,    0, 0 },
		{    0, 0x80,    0, 0 },
		{    0,    0, 0x80, 0 },
		{ 0x80, 0x80, 0x80, 0 },
	};
	const Asura_Vector_3 axLightdir[3]= 
	{
		Asura_Vector_3(0.5f, 0.5f, 0.5f),
		Asura_Vector_3(0.0f,-0.4f,-0.1f),
		Asura_Vector_3(-0.5f,-0.5f, 0.5f),
	};
	static Asura_Vector_3 axLightcol[3]= 
	{
		Asura_Vector_3(0.25f, 0.25f, 0.25f),
		Asura_Vector_3(0.25f, 0.25f, 0.25f),
		Asura_Vector_3(0.25f, 0.25f, 0.25f),
	};
	static float aiAmbient[4] = { 0.50, 0.50, 0.50, 0.00 };			


	shellMemCardResetIconSys(&s_xIconSysFile);

	// Background colour
	for (int i=0; i<4; ++i)
	{
		SetBackgroundColour(static_cast<eAsura_PS2_ShellMCUI_Icon_BackgroundPosition>(i),
							aiBgcolor[i][0],
							aiBgcolor[i][1],
							aiBgcolor[i][2]);
	}
	
	// Light position
	for (int i=0; i<3; ++i)
	{
		SetLightSource(i, axLightdir[i], axLightcol[i]);
	}
	
	SetAmbient(aiAmbient[0], aiAmbient[1], aiAmbient[2]);
	SetBackgroundTransparency(0x00);

	s_xIconSysFile.OffsLF = 24;			

	strcpy((char *) s_xIconSysFile.FnameView, s_xInputParms.saveParms.iconFileNames[0]);
	strcpy((char *) s_xIconSysFile.FnameCopy, s_xInputParms.saveParms.iconFileNames[1]);
	strcpy((char *) s_xIconSysFile.FnameDel,  s_xInputParms.saveParms.iconFileNames[2]);

	//Convert ascii strings to shift-JIS strings as required by
	//sceMcIconSys.TitleName:
	shellCoreAsciiToJis(s_xIconSysFile.TitleName, szTitleName);

}


void Asura_PS2_ShellMCUI::SetBackgroundColour(const eAsura_PS2_ShellMCUI_Icon_BackgroundPosition eBackPos, 
													const unsigned char ucR, 
													const unsigned char ucG, 
													const unsigned char ucB)
{
	Asura_Assert(eBackPos < NUMBER_POSITIONS, "iBackPos >= NUMBER_POSITIONS");

	s_xIconSysFile.BgColor[eBackPos][0] = ucR;
	s_xIconSysFile.BgColor[eBackPos][1] = ucG;
	s_xIconSysFile.BgColor[eBackPos][2] = ucB;
	s_xIconSysFile.BgColor[eBackPos][3] = 0;
}


// Background transparency range is 0 (fully transparent) to 0x80 (fully opaque)
void Asura_PS2_ShellMCUI::SetBackgroundTransparency(const unsigned char ucTrans)
{
	s_xIconSysFile.TransRate = 0x000000FF & ucTrans;
}


void Asura_PS2_ShellMCUI::SetLightSource(const int iLightSourceNumber, 
										 const Asura_Vector_3& xLightSourcePosition,
										 const Asura_Vector_3& xLightSourceColour)
{
	Asura_Assert(iLightSourceNumber < 3, "Only three light sources allowed for memory card icon");

	// Direction
	s_xIconSysFile.LightDir[iLightSourceNumber][0] = xLightSourcePosition[0];
	s_xIconSysFile.LightDir[iLightSourceNumber][1] = xLightSourcePosition[1];
	s_xIconSysFile.LightDir[iLightSourceNumber][2] = xLightSourcePosition[2];
	s_xIconSysFile.LightDir[iLightSourceNumber][3] = 0.0f;

	// Colour
	s_xIconSysFile.LightColor[iLightSourceNumber][0] = xLightSourceColour[0];
	s_xIconSysFile.LightColor[iLightSourceNumber][1] = xLightSourceColour[1];
	s_xIconSysFile.LightColor[iLightSourceNumber][2] = xLightSourceColour[2];
	s_xIconSysFile.LightColor[iLightSourceNumber][3] = 0.f;
}


void Asura_PS2_ShellMCUI::SetAmbient(const float fR, 
									 const float fG, 
									 const float fB)
{
	s_xIconSysFile.Ambient[0] = fR;
	s_xIconSysFile.Ambient[1] = fG;
	s_xIconSysFile.Ambient[2] = fB;
	s_xIconSysFile.Ambient[3] = 0.0f;
}



const Asura_UnicodeChar* Asura_PS2_ShellMCUI::GetStatusMsg(eStatusMessageType eStatusMsgType)
{
	
	if ((eStatusMsgType < STATUS_MSG_TYPE_CARD) || (eStatusMsgType >= STATUS_MSG_MAX))
	{
		Asura_Assert(false, "Invalid status msg type requested");
		return L"";
	}

	const Asura_LocalisedText_Page* pxPage = Asura_LocalisedText::GetPage(Asura_GetHashID("ASURA_PS2_MCUI"));
	Asura_Assert(pxPage, "MCUI text page not loaded");

	if (pxPage)
	{
		for (u_int i = 0; i < SHELLMCUI_STATUS_NUM; i++)
		{
			if (s_xOutputParms.statusFlags[i] && (SHELLMCUI_STATUS_TYPES[i] == (EShellMCUIStatusType)(eStatusMsgType)) )
			{	
				return pxPage->GetString(ASURA_PS2_MCUI_TEXT_ID_SHELLMCUI_STATUS_DETECTING + i);
			}
		}
	}

	return L"";
}

