//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Multistream.cpp
//
// Created by Mike Healey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Class to provide an Asura interface to PS2 multistream library
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

//  This file's header.
#include "Sound/Asura_PS2_Multistream.h"

// Stdlib independent.
#include <stdio.h>

// Stdlib platform-specific.
#include <eekernel.h>
#ifdef ASURA_PS2_CDEMUL
#include <libcdvd.h>
#endif
#include <libver.h>	// for SCE_LIBRARY_VERSION

#include <sifdev.h>
#include <sifrpc.h>
#include <sdmacro.h>

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Sound/Asura_Sound.h"

// Engine platform-specific.
#include "Core/Asura_PS2_OS.h"
#include "Core/Asura_PS2_Timers.h"
#ifdef ASURA_PS2_CDEMUL
#include "FileAccess/Asura_PS2_TOCCache.h"
#endif

// middleware
#include "Sony/Multistream/ee/sound.h"

// D E F I N E S /////////////////////////////////////////////////////////////
#ifdef ASURA_PS2_DISABLE_REVERB	// DSP is not compatible with reverb
	#define ASURA_PS2_SUPPORT_SOUND_DSP 1
#endif 

#define SIZE_OF_VAGHEADER (48)

#define ASURA_PS2_MAX_MUSIC_STREAM_BUFFER_SIZE (24064) // must be multiple of 64. largest value set in fpacker for all streamed sounds.
#define ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE (2*ASURA_PS2_MAX_MUSIC_STREAM_BUFFER_SIZE)	// each of L and R in stereo music uses this much IOP RAM
#define ASURA_PS2_MS_SOUND_LOADING_STREAM_BUFFER_SIZE (32768)// must be multiple of 64. Used for loading data from CD->SPU

#define ASURA_PS2_MS_62_LOADING 1

#ifdef ASURA_PS2_MS_62_LOADING
#define ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE (8192)
#define ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE (40960)// must be multiple of 64. Used for non-music streams
#else // not #ifdef ASURA_PS2_MS_62_LOADING
#define ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE (20480)// must be multiple of 64. Used for non-music streams
#endif // else not #ifdef ASURA_PS2_MS_62_LOADING

#define INVALID_VOICE_NUM (-1)

// T Y P E S /////////////////////////////////////////////////////////////////

// MIH is the multistream interleaved header
// It consists of a SOUND_MIH_HEADER, followed by numfiles SOUND_MIH_DATA
struct SOUND_MIH_HEADER
{
	unsigned int	MIHSize;		// .MIH file size
	unsigned int	MIHPad;			// Unused data in .MIH file (.MIH is always a multiple of 64 bytes)
	unsigned int	NumberOfFiles;	// Number of files interleaved into a single .MIB file

	unsigned int	Pitch;			// Pitch (in hz) to play streams in .MIB file
	unsigned int	BufferSize;	    // Size of each interleaved segment
	unsigned int	NumSegmentsPerFile;
};

struct SOUND_MIH_DATA
{
    unsigned int ID;   // can be anything - set per track in fpacker.
};

struct StreamingSound
{
    int voice_num_L;
	int voice_num_R;
    u_int file_id;
	int tracksize;
	int base_pitch;
    int volume;
    u_int looping:1;
	u_int m_uPad:15;
	u_short m_usFrequency;	// 0-65536 Hz
	u_int m_uSizeInBytes;	// total size of file
};

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS

struct Asura_PS2_Multistream::PartiallyStreamedSound
{
    u_short usLoadedSoundIdx;
    u_short usBasePitch;
    u_int   uOffset;        // byte offset from start of the partially streamed sound file.
    u_int   uSize;          // Data size in bytes.
//    u_int   uBufferSize;       // Bytes.
};

struct Asura_PS2_Multistream::ActivePartiallyStreamedSound
{
    bool    bInUse;
	char    cVoiceNum;
    u_short usIdx;             // Index back into array of PartiallyStreamedSound instances.
    u_int   uFileID;

    // (MP - 27.02.2003) NB: No need to store the stream ID since it can be
    // calculated by adding the instance's index within the global array to
    // ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST.
};

#endif // #ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int INVALID_MS_FILE_ID = 0xFFFFFFFFu;

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_PS2_Multistream::ReplaceableSound
{
public:
    
    ReplaceableSound()
    {
        m_uFileID = INVALID_MS_FILE_ID;
        m_uFileOffset = 0;
        m_uSize = 0;
        m_uSPUAddress = 0;
        m_usBasePitch = 0;
        m_usPartiallyStreamedIndex = -1;
    }

    u_int m_uFileID;
    u_int m_uFileOffset;
    u_int m_uSize;
    u_int m_uSPUAddress;
    u_short m_usBasePitch;
    short m_usPartiallyStreamedIndex;   // -1 if not partially streamed
};

class MICFileHeaderManager
{

    public:

        static void Initialise( void );
        static void Shutdown( void );

        static bool GetHeaderAndFilesize( const char* szFullPath, const SOUND_MIH_HEADER*& pxHeader, u_int& uSize );
		static u_int GetFileSize( void );

    private:

#ifdef ASURA_PS2_CDEMUL
        struct MICFileData
        {
            SOUND_MIH_HEADER xHdr;
            Asura_Hash_ID uFilenameHash;
			u_int uFileSize;
        };

        static MICFileData* s_pxMICFileData;
        static u_int       s_uNumHdrs;
#endif

};

// G L O B A L S /////////////////////////////////////////////////////////////

#ifdef ASURA_ASSERT
static int g_iMaxLoadedSounds = 0;
#endif

static bool g_bResetPointSet = false;
#ifdef ASURA_ASSERT
static u_int g_uMaxFreeSoundRamAtReset = 0;
static u_int g_uFreeSoundRamAtReset = 0;
#endif

static StreamingSound* ActiveStreamingSounds = NULL;

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
static u_int g_uMaxPartiallyStreamedSounds = 0;
static u_int g_uNumPartiallyStreamedSounds = 0;
static Asura_PS2_Multistream::PartiallyStreamedSound* g_pxPartiallyStreamedSounds = NULL;
static Asura_PS2_Multistream::ActivePartiallyStreamedSound* g_pxActivePartiallyStreamedSounds = NULL;
#endif // #ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS

#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
static int g_aiEffectType[ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS];
static int g_aiFilterType[ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS];
static u_int g_auFilterFreq[ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS];
static bool g_abEffectMute[ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS];
#endif//#ifdef ASURA_PS2_SUPPORT_SOUND_DSP

// D E B U G V A R S /////////////////////////////////////////////////////////
#ifdef ASURA_PS2_SUPPORT_DTS
DEBUGVAR bool dbg_bDTSSwitchOffSpeakers = true;
#endif
DEBUGVAR float dbg_fMinTimeForMultistreamWarning = 0.25f;
DEBUGVAR bool dbg_bAllowStreamPreload = false;// (MH - 22.01.2003) preload appears to be broken.
DEBUGVAR bool dbg_bDisplayMultistreamStats = false;
DEBUGVAR bool dbg_bReportPSSLoads = false;

#if _DEBUG
DEBUGVAR bool dbg_bLogSPUUsage = false;
static const char* SPU_USAGE_LOGFILE = "logfiles\\spu_usage.txt";
#endif

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static inline bool CanSendMultistreamCmds( void )
{
    return !SOUND_CheckRPCStatus();
}

static void CloseMSFile( u_int uFileID );

static inline void SendQueuedMultistreamCmds( bool bGuarantee, int iType )
{
    // Wait for idle?
    if( bGuarantee )
    {
        while( !CanSendMultistreamCmds() ) { }
    }

    // And send.
    int iRet = SOUND_FlushIOPCommand( iType );

    if( bGuarantee )
    {
        // Shouldn't fail because we wait until it's possible for MS to make
        // an RPC call.
        Asura_Assert( iRet >= 0, "" );
    }
}

#ifdef ASURA_PS2_CDEMUL
static u_int SetupFileInfoForMS( const char* szFilename, bool bFilenameHasFullPath, u_int uOffset = 0, u_int* puFileSize = NULL, u_int* puLSN = NULL );
#else
static u_int SetupFileInfoForMS( const char* szFilename, bool bFilenameHasFullPath, u_int uOffset = 0, u_int* puFileSize = NULL );
#endif

static inline int GetMusicStreamSPUAddress(int iMusicStreamID)
{
#ifdef ASURA_PS2_MS_62_LOADING
	return SOUND_MemBaseAddress + iMusicStreamID * ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE;
#else
	return SOUND_MemBaseAddress + iMusicStreamID * ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE;
#endif
}

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
static inline int GetPartiallyStreamedSPUAddress(int iPartiallyStreamedIndex)
{
	// get spu address past last stereo stream, and add index of this pss
#ifdef ASURA_PS2_MS_62_LOADING
	return GetMusicStreamSPUAddress(ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS * 2) + iPartiallyStreamedIndex * ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE;
#else
	return GetMusicStreamSPUAddress(ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS * 2) + iPartiallyStreamedIndex * ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE;
#endif
}
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_PS2_Multistream.
//

int Asura_PS2_Multistream::s_iNumFramesSinceLastIOPUpdate = 0;
int Asura_PS2_Multistream::s_iFastLoadStackAddress = 0;
#ifdef ASURA_PS2_SUPPORT_DTS
bool Asura_PS2_Multistream::s_bDTSOn = false;
#endif
bool Asura_PS2_Multistream::s_bDSPOn = false;
u_int* Asura_PS2_Multistream::s_auSPUAddressOfLoadedSound = 0;
Asura_PS2_Multistream::ReplaceableSound* Asura_PS2_Multistream::s_axReplaceableSounds = 0;
u_int Asura_PS2_Multistream::s_uEQOnBitfield = 0xffffffff;
bool Asura_PS2_Multistream::s_bStreamingSoundsAreGlobal = false;


bool Asura_PS2_Multistream::Initialise(int iMaxLoadedSpotFX)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    if( !Asura_PS2_Multistream_FastLoad::Initialise() ) { return false; }
    
    return true;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

#if _DEBUG
    if( dbg_bLogSPUUsage )
    {
        Asura_SimpleLogfileClear( SPU_USAGE_LOGFILE );
    }
#endif

    // (MP - 07.05.2003) Before setting MS loose on the IOP, cache all the
    // .MIC file headers. This isn't necessary in non-CD builds but the
    // MICFileHeaderManager class takes care of that detail internally.
    MICFileHeaderManager::Initialise();

    if (SOUND_InitIOP () < 0)
    {
        Asura_Assert(0, "failed to init IOP for sound system");
        Asura_OutputToDebugger_Retail(("failed to init IOP for sound system"));
        return false;
    }

    ActiveStreamingSounds = new StreamingSound[ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS];

#ifdef ASURA_ASSERT
    g_iMaxLoadedSounds = iMaxLoadedSpotFX;
#endif
    s_auSPUAddressOfLoadedSound = new u_int[iMaxLoadedSpotFX];
    s_axReplaceableSounds = new ReplaceableSound[iMaxLoadedSpotFX];

	SOUND_InitSPU();
    SOUND_InitCD(SOUND_SET_CALLBACK);
    SOUND_InitCD( SOUND_CD_SPIN_STREAM );

	SOUND_InitStreamData(SND_LOAD_INTERNAL,0,0);	// Use MultiStream CD/Host loading routines
    //sg_SpotFX_FileID = ASURA_PS2_STREAM_ID_SFX_LOADING;// we reserve enough stream fileIDs, then 1 for spot fx usage
	SOUND_AllocateFileMemory(ASURA_PS2_STREAM_ID_COUNT);	// Could pass as the second param of SOUND_InitStreamData
	SOUND_AllocateSpotFXMemory(iMaxLoadedSpotFX);		// Cound pass as the third param of SOUND_InitStreamData
    
    // (MH - 05.03.2003) try the SPU malloc stuff
#ifdef ASURA_PS2_DISABLE_REVERB
	int iNumReverbBuffers = 0;
#else
	int iNumReverbBuffers = 2;
#endif
    int iNumMallocsRequired = iMaxLoadedSpotFX + 
							  1 +	// for fixed allocs
							  iNumReverbBuffers;

    int iMemAllocRet = SOUND_MemAllocInit(iNumMallocsRequired, SOUND_ALLOCATE_ALL_RAM);
    Asura_Assert(iMemAllocRet == 0, "MemAllocInit failed");

#ifndef ASURA_PS2_DISABLE_REVERB
#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
#error // can't have DSP and reverb together (possibly could work using separate core for each)
#endif
	int iFXBufferSize = SOUND_GetHardwareEffectSize(SOUND_FX_STUDIO_C);
	u_int uFXBuffer1 = SOUND_MemAllocateEffectBuffer(SOUND_FX_STUDIO_C);
	u_int uFXBuffer2 = SOUND_MemAllocateEffectBuffer(SOUND_FX_STUDIO_C);
	int iEffectsResult = SOUND_SetEffectBufferAddress(uFXBuffer1 | 0xffff, uFXBuffer2 | 0xffff);
	Asura_Assertf(iEffectsResult == 0, ("SOUND_SetEffectBufferAddress failed with code %d.", iEffectsResult));
#endif // #ifndef ASURA_PS2_DISABLE_REVERB

	// fixed spu buffers
#ifdef ASURA_PS2_MS_62_LOADING
	u_int uTotalFixedSPUBufferSize = ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE * 2 * ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS;	// music streams
#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
	uTotalFixedSPUBufferSize += ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE * ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS;
#endif
#else//#ifdef ASURA_PS2_MS_62_LOADING
	u_int uTotalFixedSPUBufferSize = ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE * 2 * ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS;	// music streams
#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
	uTotalFixedSPUBufferSize += ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE * ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS;
#endif
#endif//#else not #ifdef ASURA_PS2_MS_62_LOADING

	const int iBaseSPUAddress = AllocateSPURAM(uTotalFixedSPUBufferSize);
	Asura_Assert(iBaseSPUAddress == SOUND_MemBaseAddress, "must alloc fixed addresses first");

#if _DEBUG
    if( dbg_bLogSPUUsage )
    {
        Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6u bytes @ SPU address 0x%08X for streams, etc.\n", uTotalFixedSPUBufferSize, iBaseSPUAddress ));
    }
#endif

	SOUND_SetMaxStreamLimit( ASURA_PS2_STREAM_ID_COUNT );

    for (int iStreamID = 0; iStreamID < ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS; ++iStreamID)
    {
		int iSPUAddress = GetMusicStreamSPUAddress(iStreamID);
        SOUND_AllocateStreamBuffer( iStreamID, iSPUAddress, ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE * 2 );
		// (MH - 16.07.2003) gone back to MS5.2 loading which doesn't allow resizeable SPU buffers
#ifdef ASURA_PS2_MS_62_LOADING
		SOUND_ResizeSPUBuffer( iStreamID, ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE);
		SOUND_ResizeSPUBuffer( iStreamID + 2, ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE);
#endif
    }

#if _DEBUG
	Asura_OutputToDebugger(("stream 0 (0L) had 0x%x bytes of SPU at 0x%x.\n", ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE, GetMusicStreamSPUAddress(0)));
	Asura_OutputToDebugger(("stream 1 (1L) had 0x%x bytes of SPU at 0x%x.\n", ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE, GetMusicStreamSPUAddress(1)));
	Asura_OutputToDebugger(("stream 2 (0R) had 0x%x bytes of SPU at 0x%x.\n", ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE, GetMusicStreamSPUAddress(2)));
	Asura_OutputToDebugger(("stream 3 (1R) had 0x%x bytes of SPU at 0x%x.\n", ASURA_PS2_MS_MONO_MUSIC_STREAM_BUFFER_SIZE, GetMusicStreamSPUAddress(3)));
#endif

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    // We'll initialise the ActivePartiallyStreamingSound stuff here since we
    // know in advance counts and sizes required, etc. The remainder of the
    // PSS state will be initialised when we load the appropriate chunk and
    // discover how many different sounds we have, etc.

    // Allocate instance array.
    g_pxActivePartiallyStreamedSounds = new ActivePartiallyStreamedSound[ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS];

    // Reset in-use flag for all instances and also store an invalid file ID
    // for each in case we reset or shutdown this system without having
    // loaded the relevant chunk. We'll also allocate buffer space on the IOP
    // and reserve space in SPU memory at this point.
    for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
    {
        g_pxActivePartiallyStreamedSounds[u].bInUse  = false;
		g_pxActivePartiallyStreamedSounds[u].cVoiceNum = INVALID_VOICE_NUM;
        g_pxActivePartiallyStreamedSounds[u].uFileID = INVALID_MS_FILE_ID;

        int iSPUAddress = GetPartiallyStreamedSPUAddress(u);
        SOUND_AllocateStreamBuffer( ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + u, iSPUAddress, ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE);
		// (MH - 16.07.2003) gone back to MS5.2 loading which doesn't allow resizeable SPU buffers
#ifdef ASURA_PS2_MS_62_LOADING
		SOUND_ResizeSPUBuffer( ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + u, ASURA_PS2_MS_STREAM_SPU_BUFFER_SIZE);
#endif
    }

#if _DEBUG
	Asura_OutputToDebugger(("partially streamed channel 0 had 0x%x bytes of SPU at 0x%x.\n", ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE, GetPartiallyStreamedSPUAddress(0)));
	Asura_OutputToDebugger(("partially streamed channel 1 had 0x%x bytes of SPU at 0x%x.\n", ASURA_PS2_MS_SOUND_STREAM_BUFFER_SIZE, GetPartiallyStreamedSPUAddress(1)));
#endif

    // allocate IOP buffer for dynamic loading
    SOUND_AllocateStreamBuffer( ASURA_PS2_STREAM_ID_SFX_LOADING, SOUND_DATA_STREAM, ASURA_PS2_MS_SOUND_LOADING_STREAM_BUFFER_SIZE );
#endif // #ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS

    // (MH - 07.12.2002) hack to set the thread priority for fast load
    // NRM - 13/03/03 - ChangeThreadPriority now done in Asura_PS2_OS - see Asura_PS2_OS.h for the thread priority constants
//    int iMainThreadPriority = 10;
    int iStackSize = 4096*8;
    s_iFastLoadStackAddress = reinterpret_cast<int>(Asura_AllocateAlign(iStackSize, ASURA_ALIGN_64_BYTE));

//    ChangeThreadPriority(GetThreadId(),iMainThreadPriority);
    int iRetVal = SOUND_InitFastLoad(ASURA_PS2_MULTISTREAM_THREAD_PRIORITY, s_iFastLoadStackAddress, iStackSize);
    Asura_Assert(iRetVal == 0, "SOUND_InitFastLoad failed\n.");
    
	SOUND_FlushIOPCommand(0);
// You no longer need to do SOUND_GetStatus(streamnumber,0) if you call this..
	SOUND_SetFastLoadMode(SOUND_FASTLOAD_ON);
    // (MP - 24.08.2005) This defaults to on but there's no need for us to
    // do it, since we do it manually in FastLoadCallback(). We also don't
    // properly handle the stream channel being set to 999, which happens
    // when MS is busy flushing the D-cache.
    SOUND_SetFastLoadMode(SOUND_FASTLOAD_INVALID_DCACHE_OFF);
// If EE loading, wait for user to allow load after each data packet sent..
	SOUND_SetFastLoadMode(SOUND_FASTLOAD_STOP);

    SOUND_InitWaitUpdate();

    // Wait for the above SOUND_ commands to be processed before continuing.
	SOUND_FlushIOPCommand( 0 );

    if( !Asura_PS2_Multistream_FastLoad::Initialise() )
    {
        return false;
    }

#ifdef ASURA_PS2_SUPPORT_DTS
	int iDTSQuality = 100;	// 0-100
	int iDTSInitReturnValue = SOUND_Init_EE_DTS(ASURA_PS2_MULTISTREAM_DTS_THREAD_PRIORITY, SOUND_DTS_51, iDTSQuality);
	Asura_Assert(iDTSInitReturnValue == 0, "init DTS failed");
	SOUND_SetDTSMode(SOUND_DTS_INIT_THREAD, 0);
	UpdateIOPBlocking();
#endif

    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

void Asura_PS2_Multistream::Shutdown()
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    Asura_PS2_Multistream_FastLoad::Shutdown();
    
#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    ResetPartiallyStreamedSoundState();

    for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
    {
        SOUND_CloseStreamBuffer( ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + u );
    }

    delete[] g_pxActivePartiallyStreamedSounds;
    g_pxActivePartiallyStreamedSounds = NULL;
	
	delete[] g_pxPartiallyStreamedSounds;
	g_pxPartiallyStreamedSounds = NULL;
#endif

    delete[] ActiveStreamingSounds;

    delete[] s_auSPUAddressOfLoadedSound;
    delete[] s_axReplaceableSounds;
    SOUND_CloseEverything();

#ifdef ASURA_PS2_SUPPORT_DTS
	SOUND_SetDTSMode(SOUND_DTS_OFF, 1);// 1 shuts down thread as well callbacks etc.
#endif

    UpdateIOPBlocking();
#endif

    Asura_DeallocateAlign((void*)s_iFastLoadStackAddress);

    MICFileHeaderManager::Shutdown();
}

void Asura_PS2_Multistream::SetResetPoint()
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    //s_ResetPointInSpuRAM = SOUND_GetSPUAddress();

    g_bResetPointSet = true;

#ifdef ASURA_ASSERT
    g_uMaxFreeSoundRamAtReset = SOUND_MemFindLargest();
    g_uFreeSoundRamAtReset = SOUND_MemGetTotalFree();
#endif

#endif
}

void Asura_PS2_Multistream::Reset()
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    ResetPartiallyStreamedSoundState();
#endif

    //SOUND_SetSPUWriteAddress(s_ResetPointInSpuRAM);

    Asura_Assert(g_uMaxFreeSoundRamAtReset == SOUND_MemFindLargest(), "reset didn't return Multistream to desired state");
    Asura_Assert(g_uFreeSoundRamAtReset == SOUND_MemGetTotalFree(), "reset didn't return Multistream to desired state");
#endif
}

void Asura_PS2_Multistream::SetReverb(u_int uType, u_int uDepth, u_int uDelay /* = 0 */, u_int uFeedback /* = 0 */)
{
#ifdef ASURA_PS2_DISABLE_REVERB
	return;
#else
#ifdef ASURA_ASSERT
    if (uType == SOUND_FX_OFF)
    {
        Asura_Assert(uDepth == 0 && uDelay == 0 && uFeedback == 0, "");
    }
    else if (uType == SOUND_FX_DELAY)
    {
        Asura_Assert(uFeedback == 0, "");
    }
    else if (uType != SOUND_FX_ECHO)
    {
        Asura_Assert(uDelay == 0 && uFeedback == 0, "");
    }
#endif
    if (uType > SOUND_FX_PIPE)
    {
        Asura_Assert(false, "invalid reverb type");
        return;
    }
    if (uDepth > 0x7fff)
    {
        Asura_Assert(false, "invalid reverb depth");
        return;
    }
    if (uDelay > 127)
    {
        Asura_Assert(false, "invalid delay value");
        return;
    }
    if (uFeedback > 127)
    {
        Asura_Assert(false, "invalid feedback value");
        return;
    }       

    SOUND_EnableEffects(0, uType, uDepth, uDepth, uDelay, uFeedback);    // core 0
    SOUND_EnableEffects(1, uType, uDepth, uDepth, uDelay, uFeedback);    // core 1
#endif // #else not #ifdef ASURA_PS2_DISABLE_REVERB

}


// SetEQ() - ensure DSP is on and apply settings
// type from SOUND_FILTER_LOWPASS, SOUND_FILTER_HIPASS, SOUND_FILTER_BANDPASS
// freq from 0 to 512 ( not exactly sure about the scale, roughly 0Hz to 20kHz? )
void Asura_PS2_Multistream::SetEQ(int iEffectSlot, EQ_FILTER_TYPE eFilterType, u_int uFilterFreq)
{
	Asura_Assert(iEffectSlot >= 0 && iEffectSlot < ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS, "invalid effect slot");

#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	if (!s_bDSPOn)
	{
		EnableDSPProcessing();
	}

	int iRet;
	u_int uDSPCore = 0;
	if (g_aiEffectType[iEffectSlot] != SOUND_FLAG_DSP_FILTER)
	{
		iRet = SOUND_DSPSetEffect(iEffectSlot, uDSPCore, SOUND_FLAG_DSP_FILTER);   // Slot 0 - Filter
		Asura_Assert(iRet >= 0, "SOUND_DSPSetEffect failed");

		iRet = SOUND_SetDSP(iEffectSlot, SOUND_FILTER_RESONANCE, 0);
		Asura_Assert(iRet >= 0, "SOUND_SetDSP failed");

		g_aiEffectType[iEffectSlot] = SOUND_FLAG_DSP_FILTER;
	}

	if (g_aiFilterType[iEffectSlot] != eFilterType)
	{
		iRet = SOUND_SetDSP(iEffectSlot, SOUND_FILTER_TYPE,		eFilterType); // Set up initial values for filter
		Asura_Assert(iRet >= 0, "SOUND_SetDSP failed");

		g_aiFilterType[iEffectSlot] = eFilterType;
	}

	if (g_auFilterFreq[iEffectSlot] != uFilterFreq)
	{
		iRet = SOUND_SetDSP(iEffectSlot, SOUND_FILTER_FREQ,		uFilterFreq);
		Asura_Assert(iRet >= 0, "SOUND_SetDSP failed");

		g_auFilterFreq[iEffectSlot] = uFilterFreq;
	}

	SetEQMute(iEffectSlot, false);
#else
	Asura_Assert(false, "EQ not supported - need ASURA_PS2_SUPPORT_SOUND_DSP");
#endif
}


void Asura_PS2_Multistream::SetEQMute(int iEffectSlot, bool bMute)
{
#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	if (g_abEffectMute[iEffectSlot] != bMute)
	{
		int iRet = SOUND_SetDSP(iEffectSlot, SOUND_DSP_MUTE, bMute);
		Asura_Assert(iRet >= 0, "SOUND_SetDSP failed");

		g_abEffectMute[iEffectSlot] = bMute;
	}
#endif // #ifdef ASURA_PS2_SUPPORT_SOUND_DSP
}

void Asura_PS2_Multistream::SetVoiceUseEQ(int iVoiceNum, bool bUseEQ)
{
#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	Asura_Assert((iVoiceNum >= 0) && (iVoiceNum < 24), "invalid voice num for EQ");
	
	// if DSP is on we may need to alter the settings for this voice
	if (s_bDSPOn)
	{
		bool bCurrentEQ = (s_uEQOnBitfield & (1<<iVoiceNum)) != 0;
		if (bUseEQ != bCurrentEQ)
		{
			if (bUseEQ)
			{
				SOUND_SetChannelOutputOff(iVoiceNum);
				SOUND_SetChannelEffectOn(iVoiceNum);
			}
			else
			{
				SOUND_SetChannelOutputOn(iVoiceNum);
				SOUND_SetChannelEffectOff(iVoiceNum);
			}
		}
	}

	if (bUseEQ)
	{
		s_uEQOnBitfield |= (1<<iVoiceNum);
	}
	else
	{
		s_uEQOnBitfield &= ~(1<<iVoiceNum);
	}
#endif // #ifdef ASURA_PS2_SUPPORT_SOUND_DSP
}

// DisableDSPProcessing / EnableDSPProcessing - use to resolve conflicts for PCM channel usage.
// Shutting it down can take a short while, hence the blocking parameter and query function.
void Asura_PS2_Multistream::DisableDSPProcessing(bool bBlocking)
{
#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	if (s_bDSPOn)
	{
		u_int uDSPCore = 0;

		SOUND_AllChannelOutputsOn(uDSPCore);

		SOUND_DSPOnOff(uDSPCore, SOUND_DSP_OFF);
		SOUND_CloseDSP();
		if (bBlocking)
		{
			while (IsDSPProcessingEnabled())
			{
				UpdateIOPBlocking();
				Asura_PS2_OS::DelayThread(1);
			}
		}
		
		for (u_int u = 0; u < 2; ++u)
		{
			g_abEffectMute[u] = false;
		}

		s_bDSPOn = false;
	}
#endif
}

void Asura_PS2_Multistream::EnableDSPProcessing()
{
#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	if (!s_bDSPOn)
	{
		int iIgnored = 0;
		u_int uDSPCore = 0;
		SOUND_SetDSP(iIgnored, SOUND_INIT_DSP, 0);   // Initialise DSP using SPU channel0 (first parameter ignored!)
		SOUND_DSPOnOff(uDSPCore, SOUND_DSP_ON);    // Turn DSP processing ON for core0
		SOUND_SetParam(0|SD_P_MMIX, 0xfc0);   // Turn PCM reverb OFF for core0
		SOUND_EnableEffects(uDSPCore, 0, 0x3fff, 0x3fff, 0, 0); // Enable hardware effects, using NO revery type (hall/studio A..etc..)

		// Audio will play from WET area from core0
		SOUND_AllChannelOutputsOff(uDSPCore);    // Turn ALL of core1 output off..
		SOUND_AllChannelEffectsOn(uDSPCore);     // Turn ALL of core1 effects on..

		// go through voices to see if any have been set to EQ off.
		u_int uEQBitfield = s_uEQOnBitfield;
		if (uEQBitfield != 0xffffffff)
		{
			for (u_int u = 0; u < 24; ++u)
			{
				if (!(uEQBitfield & (1<<u)))
				{
					SOUND_SetChannelEffectOff(u);
					SOUND_SetChannelOutputOn(u);
				}
			}
		}

		SOUND_SetPCMVolume(0x7fff,0x7fff);

		// update the IOP with these commands
		SOUND_FlushIOPCommand(SIF_RPCM_NOWAIT);

		for (u_int u = 0; u < ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS; ++u)
		{
			g_aiEffectType[u] = -1;
			g_aiFilterType[u] = -1;
			g_auFilterFreq[u] = 0xffffffff;
			g_abEffectMute[u] = false;
		}

		s_bDSPOn = true;
	}
#endif
}

bool Asura_PS2_Multistream::IsDSPProcessingEnabled()
{
#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	return (SOUND_GetDSPInfo(NULL) != SOUND_DSP_OFF);
#else
	return false;
#endif
}


void Asura_PS2_Multistream::Update()
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // update IOP side
	int result = 0;

// Don't need to do this when using fastload.
//    int iStreamRequiringLoad = SOUND_GetLoadRequest();
//    if (iStreamRequiringLoad != -1)
//    {
//        SOUND_GetStatus(iStreamRequiringLoad,0);
//    }

#if _DEBUG
	if (dbg_bDisplayMultistreamStats)
	{
		dbg_bDisplayMultistreamStats = false;
		SOUND_DisplayStats();
	}
#endif

#ifdef ASURA_PS2_SUPPORT_DTS

	int iDTSStatus = SOUND_GetDTSStatus();
	if (IsDTSOn())
	{
		if (iDTSStatus != SOUND_DTS_ON)
		{
			Asura_OutputToDebugger(("Attempt to start DTS. cur status = %d.\n", iDTSStatus));
			SOUND_SetDTSMode(SOUND_DTS_START, 47);// spare SPU channel
			if (dbg_bDTSSwitchOffSpeakers)
			{
				SOUND_SetTVOutputOff();
			}
		}
	}
	else
	{
		if (iDTSStatus == SOUND_DTS_ON)
		{
			Asura_OutputToDebugger(("Attempt to stop DTS. cur status = %d.\n", iDTSStatus));
			SOUND_SetDTSMode(SOUND_DTS_OFF, 0);// 0 => don't shutdown IOP thread
			SOUND_SetTVOutputOn();
		}
	}
	
#endif

	result = SOUND_FlushIOPCommand(SIF_RPCM_NOWAIT);			// Transfer EE command list to IOP buffer

    if (result == -1)
    {
        s_iNumFramesSinceLastIOPUpdate++;
    }
    else
    {
        s_iNumFramesSinceLastIOPUpdate = 0;
#if _DEBUG
		static float fLastUpdateTime = Asura_Timers::GetLocalUnscaledTime();
		float fCurTime = Asura_Timers::GetLocalUnscaledTime();
		float fTimeSinceLastUpdate = fCurTime - fLastUpdateTime;
		if (fTimeSinceLastUpdate >= dbg_fMinTimeForMultistreamWarning)
		{
			Asura_OutputToDebugger(("Multistream failed with IOP update for %.02f seconds.\n", 
				Asura_Maths::FtoD(fTimeSinceLastUpdate)));
		}
		fLastUpdateTime = fCurTime;
#endif
    }
    

    static int iLastResult = 0;
	result = SOUND_HandleCDErrors();
#if _DEBUG
    if (result != iLastResult)
    {
        Asura_OutputToDebugger(("SOUND_HandleCDErrors returned %d.\n", result));
		iLastResult = result;
    }
#endif

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    // Check all active PSSs to see if they've finished yet. Only do so if we
    // called SOUND_FlushIOPCommand() successfully above since this will 
    // guarantee Multistream has processed requests made in the last frame.
    if( !s_iNumFramesSinceLastIOPUpdate )
    {
        for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
        {
            if( g_pxActivePartiallyStreamedSounds[u].bInUse )
            {
                if( !SOUND_GetStreamStatus(ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + u) )
                {
                    g_pxActivePartiallyStreamedSounds[u].bInUse = false;
					g_pxActivePartiallyStreamedSounds[u].cVoiceNum = INVALID_VOICE_NUM;
                }
            }
        }
    }
#endif
    
#endif//#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

void Asura_PS2_Multistream::UpdateIOPBlocking()
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    while( SOUND_CheckRPCStatus() ) { }

    int i = SOUND_FlushIOPCommand( 0 );

    // Shouldn't fail because we waited until MS was able to make an RPC call.
    Asura_Assert( i >= 0, "" );
#endif
}


bool Asura_PS2_Multistream::ReplaceSound(const int iExistingLoadedSoundID, const int iNewLoadedSoundID, u_short& usNewBasePitch)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // verify arguments
    Asura_Assert(iExistingLoadedSoundID >= 0 && iExistingLoadedSoundID < g_iMaxLoadedSounds, "");
    Asura_Assert(iNewLoadedSoundID >= 0 && iNewLoadedSoundID < g_iMaxLoadedSounds, "");
    
    ReplaceableSound& xExistingSound = s_axReplaceableSounds[iExistingLoadedSoundID];
    ReplaceableSound& xNewSound = s_axReplaceableSounds[iNewLoadedSoundID];

    // get address of existing sound in SPU RAM and verify
    u_int uSPUAddress = xExistingSound.m_uSPUAddress;
    Asura_Assert(uSPUAddress < SOUND_MemEndOfRAM, "out of range");
    Asura_Assert(xNewSound.m_uSPUAddress == 0, "new sound already loaded?");

    // check the loading stream is free if we will need it.
    if (xNewSound.m_uSize && (SOUND_GetStreamStatus(ASURA_PS2_STREAM_ID_SFX_LOADING) != STREAM_STATUS_OFF))
    {
        // this might happen quite often unless we use a queue.
        return false;
    }
    
    // Free SPU RAM, clear address and MS spot FX ID of existing sound.
    if (uSPUAddress)
	{
		int iRet = SOUND_MemFree(uSPUAddress);
		Asura_Assertf(iRet == 0, ("error freeing SPU address %x.", uSPUAddress));
		xExistingSound.m_uSPUAddress = 0;
		SOUND_FreeSpotFXID((u_short)iExistingLoadedSoundID);

#if _DEBUG
        if( dbg_bLogSPUUsage )
        {
            const char* szSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iExistingLoadedSoundID );
            Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Free  SPU address 0x%08X (ready for replacing sound \"%s\".)\n", uSPUAddress, szSoundName ));
        }
#endif
	}

	if (xNewSound.m_uSize == 0)
	{
		// new sound has no stub, it is fully-streamed.
		return true;
	}

    // allocate SPU RAM
    {
		int iSPUAddress = AllocateSPURAM(xNewSound.m_uSize);
		if (iSPUAddress < 0)
		{
#if _DEBUG
            if( dbg_bLogSPUUsage )
            {
                const char* szOldSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iExistingLoadedSoundID );
                const char* szNewSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iNewLoadedSoundID );
                Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6u bytes FAILED for replacement sound \"%s\" (replacing \"%s\".)\n", xNewSound.m_uSize, szNewSoundName, szOldSoundName ));
            }
#endif
			return false;
		}
		uSPUAddress = iSPUAddress;

#if _DEBUG
        if( dbg_bLogSPUUsage )
        {
            const char* szOldSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iExistingLoadedSoundID );
            const char* szNewSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iNewLoadedSoundID );
            Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6u bytes @ SPU address 0x%08X for replacement sound \"%s\" (replacing \"%s\".)\n", xNewSound.m_uSize, iSPUAddress, szNewSoundName, szOldSoundName ));
        }
#endif
	}
    xNewSound.m_uSPUAddress = uSPUAddress;
    
    // do the load
    u_int uFileID = xNewSound.m_uFileID;
    u_int uFileOffset = xNewSound.m_uFileOffset;
    u_int uSize = xNewSound.m_uSize;

    SOUND_SetFileOffsetAndPlaySize(uFileID, uFileOffset, uSize);
    SOUND_SetSPUWriteAddress(uSPUAddress);
    SOUND_PlayStream(uFileID, ASURA_PS2_STREAM_ID_SFX_LOADING, 0, 0, 0, 0, STREAM_LOAD_SPU, 0, 0);
    // (MH - 10.03.2003) give this next highest priority. I'm assuming that music streams will have 
    // big enough buffers to handle taking a bit longer. These bits should be fairly small.
    SOUND_SetStreamPriority(ASURA_PS2_STREAM_ID_SFX_LOADING, 1);
    SOUND_PatchSFX(iNewLoadedSoundID, uSPUAddress);

    // flush the queue, why not?
    SOUND_FlushIOPCommand(SIF_RPCM_NOWAIT);
    usNewBasePitch = xNewSound.m_usBasePitch;
    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

DEBUGVAR bool dbg_bNoReplaceableSounds = false;
DEBUGVAR u_int dbg_uMinPssStubSize = 8064;	// helps prevent fragmentation, this should match the default value in resource packer.
bool Asura_PS2_Multistream::LoadReplaceableSound(const int iLoadedSoundID, u_short& usNewBasePitch)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

	if (dbg_bNoReplaceableSounds) return false;

    // verify argument
    Asura_Assert(iLoadedSoundID >= 0 && iLoadedSoundID < g_iMaxLoadedSounds, "");

    ReplaceableSound& xNewSound = s_axReplaceableSounds[iLoadedSoundID];

    // check the sound has been setup, fileID will be valid in that case
    Asura_Assert(xNewSound.m_uFileID != INVALID_MS_FILE_ID, "LoadReplaceableSound called with sound that doesn't exist yet");

    // (MH - 12.05.2003) if size is 0, nothing to preload as the sound is fully-streamed.
	u_int uSize = xNewSound.m_uSize;
    if (uSize == 0)
    {
		usNewBasePitch = xNewSound.m_usBasePitch;
        return true;
    }

    // check the loading stream is free
    if (SOUND_GetStreamStatus(ASURA_PS2_STREAM_ID_SFX_LOADING) != STREAM_STATUS_OFF)
    {
        Asura_Assert(false, "loading stream is in use");
        return false;
    }
    
    // allocate SPU RAM
	if (uSize < dbg_uMinPssStubSize)
	{
		uSize = dbg_uMinPssStubSize;
	}
    int iSPUAddress = AllocateSPURAM(uSize);
    if (iSPUAddress < 0)
    {
#if _DEBUG
        if( dbg_bLogSPUUsage )
        {
            const char* szSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iLoadedSoundID );
            Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6u bytes FAILED for new replaceable sound \"%s\".\n", uSize, szSoundName ));
        }
#endif
        return false;
    }

#if _DEBUG
    if( dbg_bLogSPUUsage )
    {
        const char* szSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( iLoadedSoundID );
        Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6u bytes @ SPU address 0x%08X for new replaceable sound \"%s\".\n", uSize, iSPUAddress, szSoundName ));
    }
#endif

    xNewSound.m_uSPUAddress = iSPUAddress;

    // do the load
    u_int uFileID = xNewSound.m_uFileID;
    u_int uFileOffset = xNewSound.m_uFileOffset;

    SOUND_SetFileOffsetAndPlaySize(uFileID, uFileOffset, uSize);
    SOUND_SetSPUWriteAddress(iSPUAddress);
    SOUND_PlayStream(uFileID, ASURA_PS2_STREAM_ID_SFX_LOADING, 0, 0, 0, 0, STREAM_LOAD_SPU, 0, 0);
    SOUND_PatchSFX(iLoadedSoundID, iSPUAddress);
    
    SOUND_WaitForFileToLoad(ASURA_PS2_STREAM_ID_SFX_LOADING | SOUND_CHAN_LOAD_ONLY, SOUND_WAIT_HOT);// we already called SOUND_InitWaitUpdate()

    usNewBasePitch = xNewSound.m_usBasePitch;
	
    {
	    static int iNumStubsLoaded = 0;
        ++iNumStubsLoaded;
        if( dbg_bReportPSSLoads )
        {
	        Asura_OutputToDebugger(( "Loaded partially streamed stub (%i loaded.)\n", iNumStubsLoaded ));
        }
    }
    
    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

bool Asura_PS2_Multistream::LoadSoundFromMemory( const char* vag_in_memory, const int num_bytes, const int loadedSoundIndex )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    return false;
#else
//    if (s_bTempSPURAMAllocated)
//    {
//        Asura_Assert(false, "can't load sound while temp SPU RAM is allocated");
//        return false;
//    }
    
//    int spu_pointer_after_write = SOUND_GetSPUAddress() + num_bytes - SIZE_OF_VAGHEADER;
//    if (spu_pointer_after_write >= sg_EndOfSpuRAM)
//    {
//        Asura_Assert(0, "failed to load sound - out of SPU RAM");
//        return false;
//    }
    void * IOP_address = sceSifAllocIopHeap(num_bytes);
    if (!IOP_address)
    {
        Asura_Assertf(0, ("failed to allocate %d bytes of IOP memory for sound loading, max available = %d, total available = %d.\n", 
			num_bytes, sceSifQueryMaxFreeMemSize(), sceSifQueryTotalFreeMemSize()));
        return false;
    }

    int iSPUAddress = AllocateSPURAM(num_bytes - SIZE_OF_VAGHEADER);
    if (iSPUAddress < 0)
    {
#if _DEBUG
        if( dbg_bLogSPUUsage )
        {
            const char* szSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( loadedSoundIndex );
            Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6i bytes FAILED for new resident sound \"%s\".\n", num_bytes - SIZE_OF_VAGHEADER, szSoundName ));
        }
#endif
		sceSifFreeIopHeap(IOP_address);
        return false;
    }

#if _DEBUG
    if( dbg_bLogSPUUsage )
    {
        const char* szSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( loadedSoundIndex );
        Asura_SimpleLogfileOutput(( SPU_USAGE_LOGFILE, "Alloc %6i bytes @ SPU address 0x%08X for new resident sound \"%s\".\n", num_bytes - SIZE_OF_VAGHEADER, iSPUAddress, szSoundName ));
    }
#endif

#ifdef ASURA_PS2_SUPPORT_SOUND_DSP
	if (s_bDSPOn)
	{
		// transfers using SOUND_TransIOPToSPU take much longer if DSP is on.
		DisableDSPProcessing(true);
	}
#endif
    
	SOUND_TransferDataToIOP((u_int)(vag_in_memory+SIZE_OF_VAGHEADER), (u_int)IOP_address, num_bytes - SIZE_OF_VAGHEADER, SOUND_DMA_WAIT);
    SOUND_PatchSFX(loadedSoundIndex, iSPUAddress);
    SOUND_TransIOPToSPU((u_int)IOP_address, iSPUAddress, num_bytes - SIZE_OF_VAGHEADER);
    
	UpdateIOPBlocking();

    sceSifFreeIopHeap(IOP_address);

    // remember the address
    s_auSPUAddressOfLoadedSound[loadedSoundIndex] = iSPUAddress;
    
    return true;
#endif
}

bool Asura_PS2_Multistream::UnloadSound(const int loadedSoundIndex)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    return false;
#else
    SOUND_FreeSpotFXID((u_short)loadedSoundIndex);

#if _DEBUG
    if( dbg_bLogSPUUsage )
    {
        const char* szSoundName = Asura_Sound_System::GetFileNameFromSoundIndex( loadedSoundIndex );
        Asura_Logfile(( SPU_USAGE_LOGFILE, "Freed SPU address 0x%08X (was (resident?) sound \"%s\".)\n", s_auSPUAddressOfLoadedSound[loadedSoundIndex], szSoundName ));
    }
#endif

    int iRet = SOUND_MemFree(s_auSPUAddressOfLoadedSound[loadedSoundIndex]);
    Asura_Assertf(iRet == 0, ("error freeing SPU address %x.", s_auSPUAddressOfLoadedSound[loadedSoundIndex]));
    return (iRet == 0);
#endif
}

bool Asura_PS2_Multistream::IsVoiceFree(int voice_num)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    return false;
#else
    int actualIndex = SOUND_FindFreeSPUChannel(voice_num, voice_num);
    return (actualIndex == voice_num);
#endif
}


bool Asura_PS2_Multistream::IsSoundPlaying(int voice_num)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    return false;
#else
    return (SOUND_GetKeyStatus(voice_num) != SPU_KEY_STATUS_OFF);
#endif
}


#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
bool Asura_PS2_Multistream::IsPartiallyStreamedSoundPlaying(int iVoiceNum)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    return false;
#else
    bool bPlaying = (SOUND_GetKeyStatus(iVoiceNum) != SPU_KEY_STATUS_OFF);
	if (!bPlaying)
	{
		// (MH - 11.07.2003) flag this one as being free to use
		for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
		{
			if( (int)g_pxActivePartiallyStreamedSounds[u].cVoiceNum == iVoiceNum)
			{
				g_pxActivePartiallyStreamedSounds[u].bInUse = false;
				g_pxActivePartiallyStreamedSounds[u].cVoiceNum = INVALID_VOICE_NUM;
				break;
			}
		}
	}

	return bPlaying;
#endif
}
#endif // #ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS


void Asura_PS2_Multistream::PlaySound(int loadedSoundIndex, int voice_num, int volumeL, int volumeR, int pitch)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    int attack = 0;
    int release = 0;

    SOUND_PlaySFX(loadedSoundIndex, voice_num, volumeL, volumeR, pitch, attack, release);
#endif
}

void Asura_PS2_Multistream::PlayLoopingSound(int loadedSoundIndex, int voice_num, int volumeL, int volumeR, int pitch, u_int uLoopCount)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    int attack = 0;
    int release = 0;

    SOUND_PlaySFXLoop(loadedSoundIndex, voice_num, volumeL, volumeR, pitch, attack, release, uLoopCount);
#endif
}

void Asura_PS2_Multistream::StopSound(int voice_num)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    SOUND_StopSound(voice_num);
#endif
}

void Asura_PS2_Multistream::SetVolume(int voice_num, int volumeL, int volumeR)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	// (MH - 12.06.2003) allow negative volumes to be passed in, then mask
	// off the top bit which is used for other purposes e.g. fadeout effects
	// on the SPU2. This still results in the correct inverse phase for PL2.
    SOUND_SetChannelVolume(voice_num, volumeL & 0x7fff, volumeR & 0x7fff);
#endif
}

void Asura_PS2_Multistream::SetPitch(int voice_num, int pitch)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    SOUND_SetChannelPitch(voice_num, pitch);
#endif
}

bool Asura_PS2_Multistream::OpenStreamingSoundFile(int iStreamID, const char* szFilename)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // We assume there are no more than two streams and therefore no more
    // than two distinct stream IDs. We need to know these values in advance
    // so we know what other IDs to use with file streaming, etc. See the
    // ASURA_PS2_STREAM_ID enum in this file's header. Note that we should
    // only be asked to start streams' parents here; the child streams are
    // generated internally at offset IDs.
    Asura_Assert( iStreamID == ASURA_PS2_STREAM_ID_SOUND_STREAM_0_L || iStreamID == ASURA_PS2_STREAM_ID_SOUND_STREAM_1_L, "Invalid stream ID." );

    const char* szFullPath = Asura_PS2_OS::GetFullPath( szFilename, false, "MIC" );

    const SOUND_MIH_HEADER* pxMIHHdr = NULL;
	u_int uFileSize = 0;
	bool bOK = MICFileHeaderManager::GetHeaderAndFilesize( szFullPath, pxMIHHdr, uFileSize );
    if( !bOK ) { return false; }
	Asura_Assert(pxMIHHdr && uFileSize, "MICFileHeaderManager::GetHeaderAndFilesize return OK but didn't fill out required data.");
	if ( !(pxMIHHdr && uFileSize) ) { return false; }

  
    // (MH - 14.01.2003) hack - assume no more than 2 streams i.e. 4 voices
    int iVoiceNumL;
    if (iStreamID == 0)
    {
        iVoiceNumL = ASURA_PS2_MS_AVAILABLE_VOICES;	// first 2 voices after sounds
    }
    else
    {
        iVoiceNumL = ASURA_PS2_MS_AVAILABLE_VOICES + 2;// second 2 voices after sounds
    }

	ActiveStreamingSounds[iStreamID].voice_num_L = iVoiceNumL ;
	ActiveStreamingSounds[iStreamID].voice_num_R = (iVoiceNumL + 1) ;
    ActiveStreamingSounds[iStreamID].tracksize   = pxMIHHdr->BufferSize;
    ActiveStreamingSounds[iStreamID].volume      = 0x1fff;
    ActiveStreamingSounds[iStreamID].base_pitch  = SOUND_HzToSPU( pxMIHHdr->Pitch );
	ActiveStreamingSounds[iStreamID].m_usFrequency = pxMIHHdr->Pitch;
	ActiveStreamingSounds[iStreamID].m_uSizeInBytes = uFileSize;
	
// Resize the allocated stream buffer for this track
	int	iBufferSize = pxMIHHdr->BufferSize;

	Asura_Assert( iBufferSize <= ASURA_PS2_MAX_MUSIC_STREAM_BUFFER_SIZE, "Error - Buffer size is too big" );

//	SOUND_CloseStreamBuffer( ASURA_PS2_STREAM_ID_STREAM_LEFT_CHAN );
//	SOUND_AllocateStreamBuffer( ASURA_PS2_STREAM_ID_STREAM_LEFT_CHAN, ASURA_PS2_SPU_BASE_ADDR, iBufferSize ); 
    SOUND_ResizeStreamBuffer(iStreamID, iBufferSize * 2 * 2);

	// File ID, filename, offset into file where sample data starts.
	ActiveStreamingSounds[iStreamID].file_id = SetupFileInfoForMS( szFullPath, true, pxMIHHdr->MIHSize );

    // That can't fail since we've already opened the file above.
    Asura_Assert( INVALID_MS_FILE_ID != ActiveStreamingSounds[iStreamID].file_id, "" );

	Asura_OutputToDebugger(("stream ID %d, tracksize %d, pitch %d, %s.\n", iStreamID, iBufferSize, pxMIHHdr->Pitch, szFullPath));
    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

void Asura_PS2_Multistream:: PlayStreamingSound(int iStreamID, bool bLoop, bool bPreload, bool bMono)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    
#if 0 //def ASURA_PS2_MS_62_LOADING	// looping is more-or-less broken in 62 mode.
	int loop = STREAM_ONCE_OFF;
#else
	int loop = bLoop ? STREAM_LOOPING : STREAM_ONCE_OFF;
#endif
	int vol = ActiveStreamingSounds[iStreamID].volume;
	int pitch = ActiveStreamingSounds[iStreamID].base_pitch;
	int voice_L = ActiveStreamingSounds[iStreamID].voice_num_L;
	int voice_R = ActiveStreamingSounds[iStreamID].voice_num_R;
	u_int file_id = ActiveStreamingSounds[iStreamID].file_id;
	int tracksize = ActiveStreamingSounds[iStreamID].tracksize;

	Asura_Assert(file_id != INVALID_MS_FILE_ID, "invalid file id");

    // Note, as we are playing back two streams to create a stereo image, 
    //this stream only plays out of the LEFT speaker..
	SOUND_PlayStream(file_id, iStreamID, voice_L, vol, 0, pitch, loop, 0, 0);

    // Tell MultiStream that this is the "Parent", and to use track 0 (left channel) data.
	SOUND_SetStreamParent_Int(iStreamID, 0, tracksize);	// Parent Stream, track to play, track size

    // Make a child.
    // Note, this stream plays track 1 (right channel) data and only out of the RIGHT speaker..
	u_int uSPUAddr = GetMusicStreamSPUAddress(iStreamID + 2);
	SOUND_SetStreamChild_Int(iStreamID + 2, iStreamID, voice_R, 1, 0, vol, uSPUAddr );

    SOUND_StreamIgnoreVAGPitch(iStreamID);
    SOUND_StreamIgnoreVAGPitch(iStreamID + 2);

    if (bMono)
    {
        ChangeMonoOrStereo(iStreamID, bMono);
    }

    if (dbg_bAllowStreamPreload && bPreload)
    {
        SOUND_StreamDisableKeyOn(iStreamID);
        // (MH - 22.01.2003) SOUND_StreamDisableKeyOn(iStreamID + 2);
    }
    else
    {
        SOUND_StreamAllowKeyOn(iStreamID);
        // (MH - 22.01.2003) SOUND_StreamAllowKeyOn(iStreamID + 2);
    }
    
#endif
}

void Asura_PS2_Multistream:: StopStreamingSound(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
//    if (ActiveStreamingSounds[iStreamID].voice_num_L != INVALID_VOICE_NUM)
//	{
//		SOUND_StopStream(ActiveStreamingSounds[iStreamID].voice_num_L);
//        ActiveStreamingSounds[iStreamID].voice_num_L = INVALID_VOICE_NUM;
//    }
//    if (ActiveStreamingSounds[iStreamID].voice_num_R != INVALID_VOICE_NUM)
//    {
//        SOUND_StopStream(ActiveStreamingSounds[iStreamID].voice_num_R);
//		ActiveStreamingSounds[iStreamID].voice_num_R = INVALID_VOICE_NUM;
//	}
    SOUND_StopStream(iStreamID);
#endif
}

void Asura_PS2_Multistream:: PauseStreamingSound(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
//    SOUND_SetChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_L, 0);
//	SOUND_SetChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_R, 0);
	// (MH - 05.02.2004) workaround MS inconsistency where it doesn't mask the group off by doing &0xffff
    SOUND_InitMultiChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_L & 0xffff, 0);
    SOUND_AddMultiChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_R & 0xffff);
    SOUND_SetMultiChannelPitch();
#endif
}

void Asura_PS2_Multistream:: ResumeStreamingSound(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
//	SOUND_SetChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_L, ActiveStreamingSounds[iStreamID].base_pitch);
//	SOUND_SetChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_R, ActiveStreamingSounds[iStreamID].base_pitch);
    SOUND_InitMultiChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_L & 0xffff, 
        ActiveStreamingSounds[iStreamID].base_pitch);
    SOUND_AddMultiChannelPitch(ActiveStreamingSounds[iStreamID].voice_num_R & 0xffff);
    SOUND_SetMultiChannelPitch();
#endif
}

void Asura_PS2_Multistream:: CloseStreamingSoundFile(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    CloseMSFile( ActiveStreamingSounds[iStreamID].file_id );
	// that call also stops the streams

	ActiveStreamingSounds[iStreamID].voice_num_L = INVALID_VOICE_NUM;
	ActiveStreamingSounds[iStreamID].voice_num_R = INVALID_VOICE_NUM;
	ActiveStreamingSounds[iStreamID].file_id = INVALID_MS_FILE_ID;
#endif
}

bool Asura_PS2_Multistream::StreamingSoundIsPreloaded(int iStreamID)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    return false;
#else
    SOUND_STREAM_INFO xStreamInfo;
    SOUND_GetStreamInfo(iStreamID, &xStreamInfo);
    return (xStreamInfo.Active != 0);
#endif
}

void Asura_PS2_Multistream::StartPreloadedSound(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    SOUND_StreamAllowKeyOn(iStreamID);
#endif
}

DEBUGVAR int dbg_iFixedPointMonoScale = 65536 * 2 / 3;
void Asura_PS2_Multistream::ChangeMonoOrStereo(int iStreamID, bool bMono)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    int iLVolL, iLVolR, iRVolL, iRVolR;
    if (bMono)
    {
        iLVolL = iLVolR = iRVolL = iRVolR = (dbg_iFixedPointMonoScale * ActiveStreamingSounds[iStreamID].volume) >> 16;;
    }
    else
    {
        iLVolL = iRVolR = ActiveStreamingSounds[iStreamID].volume;
        iLVolR = iRVolL = 0;
    }

    SOUND_SetChannelVolume(ActiveStreamingSounds[iStreamID].voice_num_L, iLVolL, iLVolR);
    SOUND_SetChannelVolume(ActiveStreamingSounds[iStreamID].voice_num_R, iRVolL, iRVolR);
#endif
}


bool Asura_PS2_Multistream::IsStreamPlaying(int iStreamID)
{
	return (SOUND_GetStreamStatus(iStreamID) != STREAM_STATUS_OFF);
}


float Asura_PS2_Multistream::GetStreamPlayedSeconds(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    SOUND_STREAM_INFO xStreamInfo;
    SOUND_GetStreamInfo(iStreamID, &xStreamInfo);
	float fFrames = static_cast<float>(xStreamInfo.Time_Frames);
	fFrames += static_cast<float>(xStreamInfo.Time_Seconds) * 60.0f;
	fFrames += static_cast<float>(xStreamInfo.Time_Minutes) * 60.0f * 60.0f;
	fFrames += static_cast<float>(xStreamInfo.Time_Hours) * 60.0f * 60.0f * 60.0f;
	return fFrames / 60.0f;
#else
	return 0.0f;
#endif
}


void Asura_PS2_Multistream::SetStreamVolume(int iStreamID, int iVol)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    ActiveStreamingSounds[iStreamID].volume = iVol;
#endif
}


// WARNING: this play offset only gets updated every time a buffer is filled
// so it's not at all accurate. You may want to use GetStreamPlayedSeconds()
// instead.
u_int Asura_PS2_Multistream::GetStreamPlayOffset(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    SOUND_STREAM_INFO xStreamInfo;
    SOUND_GetStreamInfo(iStreamID, &xStreamInfo);
	return xStreamInfo.PlayOffset;
#else
	return 0;
#endif
}


u_int Asura_PS2_Multistream::GetStreamFilesize(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	return ActiveStreamingSounds[iStreamID].m_uSizeInBytes;
#else
	return 0;
#endif
}


u_int Asura_PS2_Multistream::GetStreamFrequency(int iStreamID)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	return ActiveStreamingSounds[iStreamID].m_usFrequency;
#else
	return 0;
#endif
}

int Asura_PS2_Multistream::AllocateSPURAM(u_int uBytes)
{
    int iSPUAddress = SOUND_MemAllocate(uBytes);
#ifdef ASURA_ASSERT
    if (iSPUAddress < 0)
    {
        const char* szReason;
        switch (iSPUAddress)
        {
        case -1:
            szReason = "Too many allocs";
            break;
        case -2:
            szReason = "Out of SPU RAM";
            break;
        default:
            szReason = "unknown";
            break;
        }
        Asura_Assertf(iSPUAddress >= 0, ("failed to alloc sound RAM, reason = %d (%s)", iSPUAddress, szReason));
    }
#endif
    return iSPUAddress;
}

u_int Asura_PS2_Multistream::AllocateTempSPURAM(u_int uBytes)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return 0;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    int iSPUAddress = AllocateSPURAM(uBytes);
    if (iSPUAddress < 0)
    {
        return 0;
    }
    return iSPUAddress;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

void Asura_PS2_Multistream::FreeTempSPURAM(u_int uAddress)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    int iRet = SOUND_MemFree(uAddress);
    Asura_Assertf(iRet == 0, ("error freeing SPU address %x.", uAddress));
#endif
}

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS

// This struct came from newsgroup message "re: VAG header format" 25.01.02
struct VAGHeader
{
    unsigned int  format;  /* always 'VAGp' for identifying*/
    unsigned int  ver;  /* format version (2) */
    unsigned int  ssa;  /* Source Start Address, always 0 (reserved for VAB format) */
    unsigned int  size;  /* Sound Data Size in byte */
    unsigned int  fs;   /* sampling frequency, 44100(>pt1000), 32000(>pt),22000(>pt0800)... */
    unsigned short  volL;  /* base volume for Left channel */
    unsigned short  volR;  /* base volume for Right channel */
    unsigned short  pitch;  /* base pitch (includes fs modulation)*/
    unsigned short  ADSR1;  /* base ADSR1 (see SPU manual) */
    unsigned short  ADSR2;  /* base ADSR2 (see SPU manual) */
    unsigned short  reserved; /* not in use */
    char    name[16];
};

//static unsigned short GetPitchFromVAGHeader(const VAGHeader * vag_header)
//{
//    int wrong_endian_sample_rate = vag_header->fs;
//    int sampling_rate = ((wrong_endian_sample_rate & 0xff000000) >> 24) | 
//        ((wrong_endian_sample_rate & 0x00ff0000) >> 8);
//    u_short pitch = (sampling_rate*4096)/48000;
//    return pitch;
//}
//
//static bool PS2SoundIsLooping( const char* pcVagData, int iDataSize)
//{
//    const char* pcLoopByte = pcVagData + iDataSize - 15;
//
//    // (MH - 16.01.2003) I think this is wrong
//    //if ( *pcLoopByte & 0x3 )
//    if (*pcLoopByte == 0x03)
//    {
//        return true;
//    }
//    else
//    {
//        return false;
//    }
//}
//
//static const int PSS_TEST_SOUND_INDEX = 60000; // (MP - 28.02.2003) Just an index I know isn't used in Dredd.
//static const int PSS_TEST_SOUND_VOICE = 40;    // (MP - 28.02.2003) And a voice that should be free.

//void SetupPartiallyStreamingSoundTest( void )
//{
//    static const char* SFX_TEST_FILE = "testsfx.vag";
//    static const char* STR_TEST_FILE = "teststr.vag";
//
//    Asura_PS2_Multistream::InitPartiallyStreamingSounds( 1, STR_TEST_FILE );
//
//    int   iFile;
//    u_int uFileSize;
//    char* pcBfr;
//
//    iFile = sceOpen( Asura_PS2_OS::GetFullPath(SFX_TEST_FILE), SCE_RDONLY );
//    uFileSize = sceLseek( iFile, 0, SCE_SEEK_END );
//    Asura_Assert( !(uFileSize & 15), "" );
//    sceLseek( iFile, 0, SCE_SEEK_SET );
//
//    pcBfr = reinterpret_cast<char*>( Asura_AllocateAlign(uFileSize, ASURA_ALIGN_128_BYTE) );
//    sceRead( iFile, pcBfr, uFileSize );
//    Asura_Assert( PS2SoundIsLooping(pcBfr, uFileSize), "The SFX sample must be looping!" );
//    Asura_PS2_Multistream::LoadSoundFromMemory( pcBfr + sizeof(VAGHeader), uFileSize - sizeof(VAGHeader), PSS_TEST_SOUND_INDEX );
//    Asura_DeallocateAlign( pcBfr );
//
//    sceClose( iFile );
//
//    iFile = sceOpen( Asura_PS2_OS::GetFullPath(STR_TEST_FILE), SCE_RDONLY );
//    uFileSize = sceLseek( iFile, 0, SCE_SEEK_END );
//    Asura_Assert( !(uFileSize & 15), "" );
//    sceLseek( iFile, 0, SCE_SEEK_SET );
//
//    VAGHeader xStrVAGHdr;
//    sceRead( iFile, &xStrVAGHdr, sizeof(xStrVAGHdr) );
//
//    sceClose( iFile );
//
//    // (MP - 03.03.2003) Obviously assuming SFX sample and stream use same
//    // pitch here...
//    g_pxPartiallyStreamedSounds[0].usBasePitch      = GetPitchFromVAGHeader( &xStrVAGHdr );
//    g_pxPartiallyStreamedSounds[0].usQWOffset       = static_cast<u_short>( (sizeof(xStrVAGHdr) >> 4) );
//    g_pxPartiallyStreamedSounds[0].usQWSize         = static_cast<u_short>( ((uFileSize - sizeof(xStrVAGHdr)) >> 4) );
//    g_pxPartiallyStreamedSounds[0].usLoadedSoundIdx = PSS_TEST_SOUND_INDEX;
//}
//
//void DoPartiallyStreamingSoundTest( void )
//{
//    Asura_PS2_Multistream::PlayPartiallyStreamingSound( 0, PSS_TEST_SOUND_VOICE, 0x3FFF, 0x3FFF, Asura_Sound_System::SoundGroup_SoundFX );
//}

void Asura_PS2_Multistream::InitPartiallyStreamingSounds( u_int uNumPSSs, const char* szStreamFilename )
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( !g_pxPartiallyStreamedSounds, "Already initialised." );

    // Only currently supporting per-level streaming sounds, not global stuff.
    // See comment in ResetPartiallyStreamedSoundState().
    Asura_Assert( g_bResetPointSet, "Unsupported!" );

    // Record count.
    g_uMaxPartiallyStreamedSounds = uNumPSSs;

    // Allocate instance array.
    g_pxPartiallyStreamedSounds = new PartiallyStreamedSound[uNumPSSs];

    // Setup file info. Each partially streamed sound needs a separate file
    // ID so we can set file offsets.
    for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
    {
        Asura_Assert( g_pxActivePartiallyStreamedSounds[u].uFileID == INVALID_MS_FILE_ID, "" );
        g_pxActivePartiallyStreamedSounds[u].uFileID = SetupFileInfoForMS( szStreamFilename, false, 0 );

        // Only complain once.
        if( !u )
        {
            Asura_Assertf( INVALID_MS_FILE_ID != g_pxActivePartiallyStreamedSounds[u].uFileID, ("File \"%s\" not found.", szStreamFilename) );
        }
    }

    // (MP - 27.02.2003) NB: The PartiallyStreamedSound instances' state will
    // be initialised by the chunk loading code.

#endif // #if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

bool Asura_PS2_Multistream::AddPartiallyStreamingSoundData( int iSoundID, int iFrequency, int iOffset1, int iSize1, int iOffset2, int iSize2, Asura_Hash_ID uFilenameHash )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // check arguments
    Asura_Assert(iSoundID > 0, "invalid sound ID");
    Asura_Assert(iFrequency > 1000 && iFrequency <= 48000, "invalid frequency");
    bool bPartiallyStreamed = iOffset2 != 0;
    
    if (bPartiallyStreamed && (g_uNumPartiallyStreamedSounds >= g_uMaxPartiallyStreamedSounds))
    {
        Asura_Assert(false, "no room to add partially streaming sounds.");
        return false;
    }

    int iLoadedSoundIndex = Asura_Sound_System::SoundLoadedByPlatformCode(iSoundID, uFilenameHash);
    if (iLoadedSoundIndex < 0)
    {
        return false;
    }


    u_short usBasePitch = SOUND_HzToSPU(iFrequency);

    // (MH - 12.05.2003) now valid to have 0 size for non-streamed bit
	if (iOffset1)
	{
		iOffset1 += SIZE_OF_VAGHEADER;
	}
	if (iSize1)
	{
		iSize1 -= SIZE_OF_VAGHEADER;
	}

    s_axReplaceableSounds[iLoadedSoundIndex].m_uFileID = g_pxActivePartiallyStreamedSounds[0].uFileID;
    s_axReplaceableSounds[iLoadedSoundIndex].m_uFileOffset = iOffset1;
    s_axReplaceableSounds[iLoadedSoundIndex].m_uSize = iSize1;
    s_axReplaceableSounds[iLoadedSoundIndex].m_usBasePitch = usBasePitch;
    
    if (bPartiallyStreamed)
    {
        s_axReplaceableSounds[iLoadedSoundIndex].m_usPartiallyStreamedIndex = g_uNumPartiallyStreamedSounds;

        PartiallyStreamedSound& xSound = g_pxPartiallyStreamedSounds[g_uNumPartiallyStreamedSounds++];
        xSound.usLoadedSoundIdx = iLoadedSoundIndex;
        xSound.uOffset = iOffset2;
        xSound.uSize = iSize2;
        xSound.usBasePitch = usBasePitch;
		
		// (MH - 14.05.2004) might as well call the sound properly loaded if it is fully-streamed
		if (!iSize1)
		{
			// soundID is dynamic.
			if (iSoundID == ASURA_SID_DYNAMIC)
			{
				Asura_Assert(uFilenameHash != ASURA_HASH_ID_UNSET, "no hashID for sound");
				iSoundID = Asura_Sound_System::GetSoundIDFromHash(uFilenameHash);
			}
			Asura_Sound_System::LoadReplaceableSound((Asura_Sound_ID)iSoundID);
		}
    }
    else
    {
        s_axReplaceableSounds[iLoadedSoundIndex].m_usPartiallyStreamedIndex = -1;
    }

    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

int Asura_PS2_Multistream::GetReplaceableSoundSize(const int iLoadedSoundIndex)
{
    return s_axReplaceableSounds[iLoadedSoundIndex].m_uSize;
}

bool Asura_PS2_Multistream::IsLoadedIndexReplaceableSound( int iLoadedIndex )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // verify argument
    Asura_Assert(iLoadedIndex >= 0 && iLoadedIndex < g_iMaxLoadedSounds, "");
    
    // loaded if it has an SPU address
    const ReplaceableSound& xExistingSound = s_axReplaceableSounds[iLoadedIndex];
    return (xExistingSound.m_uFileID != INVALID_MS_FILE_ID);

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

bool Asura_PS2_Multistream::IsLoadedIndexPartiallyStreamedSound(int iLoadedIndex)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // verify argument
    Asura_Assert(iLoadedIndex >= 0 && iLoadedIndex < g_iMaxLoadedSounds, "");
    
    // loaded if it has an SPU address
    const ReplaceableSound& xExistingSound = s_axReplaceableSounds[iLoadedIndex];
    return (xExistingSound.m_usPartiallyStreamedIndex != -1);

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

bool Asura_PS2_Multistream::IsLoadedIndexFullyStreamedSound(int iLoadedIndex)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // verify argument
    Asura_Assert(iLoadedIndex >= 0 && iLoadedIndex < g_iMaxLoadedSounds, "");
    
    // loaded if it has an SPU address
    const ReplaceableSound& xExistingSound = s_axReplaceableSounds[iLoadedIndex];
    return ( (xExistingSound.m_usPartiallyStreamedIndex != -1) && (s_axReplaceableSounds[iLoadedIndex].m_uSize == 0));

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}


bool Asura_PS2_Multistream::PlayPartiallyStreamingSound( int iLoadedIndex, int iVoiceNum, int iVolumeL, int iVolumeR)
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    return false;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    const ReplaceableSound& xReplaceableSound = s_axReplaceableSounds[iLoadedIndex];
	bool bFullyStreamed = xReplaceableSound.m_uSize == 0;

    int iIdx = xReplaceableSound.m_usPartiallyStreamedIndex;
    if (iIdx == -1)
    {
        Asura_Assert(false, "sound is not partially streaming");
        return false;
    }

    Asura_Assert( 0 <= iIdx && iIdx < static_cast<int>(g_uNumPartiallyStreamedSounds), "Invalid index." );
    const PartiallyStreamedSound& xSound = g_pxPartiallyStreamedSounds[iIdx];

    int iFreeEntry = -1;
    for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
    {
        if( !g_pxActivePartiallyStreamedSounds[u].bInUse )
        {
            iFreeEntry = u;
            break;
        }
    }

    if( -1 == iFreeEntry )
    {
        Asura_Assertf( false, ("Only %u partially streamed sounds may be active simultaneously.", ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS) );
        return false;
    }

    ActivePartiallyStreamedSound* pxActiveSound = &g_pxActivePartiallyStreamedSounds[iFreeEntry];

    pxActiveSound->bInUse = true;
    pxActiveSound->usIdx  = static_cast<u_short>(iIdx);
	pxActiveSound->cVoiceNum = iVoiceNum;
    // (MP - 27.02.2003) NB: File ID is setup at initialisation time.

    u_int uStreamID = ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + iFreeEntry;

//    // Resize the stream buffer for this particular sound.
//    SOUND_ResizeStreamBuffer( uStreamID, xSound.uBufferSize );

    // Play the resident SFX sample.
    if (!bFullyStreamed)
	{
		SOUND_PlaySFX( xSound.usLoadedSoundIdx, iVoiceNum, iVolumeL, iVolumeR, xSound.usBasePitch, 0, 0 );
	}

    // Tell MS that when this SFX finishes, it should switch to the stream.
    // This should give us a seamless transition.
	// (MH - 12.05.2003) added fully-streamed option.
    {
//        u_int uOffset = (static_cast<u_int>(xSound.usQWOffset) << 4);
//        u_int uSize   = (static_cast<u_int>(xSound.usQWSize  ) << 4);
        SOUND_SetFileOffsetAndPlaySize( pxActiveSound->uFileID, xSound.uOffset, xSound.uSize );

        if (bFullyStreamed)
		{
			SOUND_PlayStream(pxActiveSound->uFileID, uStreamID, iVoiceNum , 
				iVolumeL, iVolumeR, xSound.usBasePitch, STREAM_ONCE_OFF, 0, 0);
		}
		else
		{
			SOUND_PlayStreamFromSFX( pxActiveSound->uFileID, uStreamID, iVoiceNum , STREAM_ONCE_OFF );
		}
        // (MH - 10.03.2003) give this the maximum priority. If one of these streams doesn't complete loading
        // in time, the preloaded bit will loop continuously. These bits should be fairly small.
        SOUND_SetStreamPriority(uStreamID, 0);
    }

    SendQueuedMultistreamCmds( true, SIF_RPCM_NOWAIT );

    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}


#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
void Asura_PS2_Multistream::StopPartiallyStreamingSound(int iVoiceNum)
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    SOUND_StopSound(iVoiceNum);
    // (MH - 11.07.2003) flag this one as being free to use
	for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
    {
        if( (int)g_pxActivePartiallyStreamedSounds[u].cVoiceNum == iVoiceNum)
		{
			g_pxActivePartiallyStreamedSounds[u].bInUse = false;
			g_pxActivePartiallyStreamedSounds[u].cVoiceNum = INVALID_VOICE_NUM;
			return;
		}
	}
	Asura_Assert(false, "failed to find partially streamed sound in list");
#endif
}
#endif // #ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS


void Asura_PS2_Multistream::ResetPartiallyStreamedSoundState( void )
{
#if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    // (MP - 30.05.2007) I've restored the code here to deallocate everything
    // because this is necessary for the current scheme, on the Simpsons
    // anyway, of having per-level streamed dialogue, rather than a global
    // file that means we only go through InitPartiallyStreamingSounds() once.
    // This means that we're not supporting the latter currently; we assert
    // about this in InitPartiallyStreamingSounds().

	for( u_int u = 0; u < ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS; ++u )
	{
		SOUND_StopStream( ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + u );
		if (!AreStreamingSoundsGlobal())
		{

			// (AMM - 15 Aug 2008) 
			// dont pass invalid file handles to the IOP. Although it initially looks like
			// the IOP handles this it does issue errors. Over the course of time
			// this causes more severe errors to get reported and eventually a hang
			// a simple check prevents the errors from occuring. Note that I could of
			// put this into the SOUND_FreeFileID function in sound.c as this fix is in
			// two places, but as the middleware may get replaced I though it better here.
			//
			if (g_pxActivePartiallyStreamedSounds[u].uFileID != INVALID_MS_FILE_ID)
			{
				SOUND_FreeFileID( g_pxActivePartiallyStreamedSounds[u].uFileID );
			}

			g_pxActivePartiallyStreamedSounds[u].bInUse    = false;
			g_pxActivePartiallyStreamedSounds[u].cVoiceNum = INVALID_VOICE_NUM;
			g_pxActivePartiallyStreamedSounds[u].uFileID   = INVALID_MS_FILE_ID;
		}
	}

	SendQueuedMultistreamCmds( true, 0 );

	if (!AreStreamingSoundsGlobal())
	{
		delete[] g_pxPartiallyStreamedSounds;
		g_pxPartiallyStreamedSounds = NULL;

		g_uMaxPartiallyStreamedSounds = 0;
		g_uNumPartiallyStreamedSounds = 0;
	}

	for ( u_int u = 0; u < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS; ++u)
	{
		ReplaceableSound& xSound = s_axReplaceableSounds[u];
		if (xSound.m_uSPUAddress)
		{
			int iRet = SOUND_MemFree(xSound.m_uSPUAddress);
			Asura_Assertf(iRet == 0, ("error freeing SPU address %x.", xSound.m_uSPUAddress));
			xSound.m_uSPUAddress = 0;
			SOUND_FreeSpotFXID((u_short)u);
		}
	}

	SendQueuedMultistreamCmds( true, 0 );

#endif // #if !ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

#endif // #ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS

u_int Asura_PS2_Multistream::GetTotalFreeSPURAM( void )
{
    return SOUND_MemGetTotalFree();
}

//////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Multistream_FastLoad.
//
//////////////////////////////////////////////////////////////////////////

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
u_int Asura_PS2_Multistream_FastLoad::s_uNumOpenFiles;
u_int Asura_PS2_Multistream_FastLoad::s_uLastFastloadMode = 0xffffffff;

#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

struct FileInfo
{
    enum { BUFFER_SIZE = 16 * 2048 };

    char acData[BUFFER_SIZE];
    int  iHandle;
};

static FileInfo g_axFileInfo[ASURA_MAX_NO_OF_OPEN_FILES];

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

#define USE_CONTINUE_RET_VAL (0)
#define USE_3_BUFFERS        (0)

#if USE_3_BUFFERS
static const u_int FL_BUFFER_SIZE = 16 * 2048;  // This need to be a multiple of 16K if the fmv player is using this system
static const u_int FL_BUFFER_QTY  = 3;
#else
static const u_int FL_BUFFER_SIZE = 24 * 2048;  // This need to be a multiple of 16K if the fmv player is using this system
static const u_int FL_BUFFER_QTY  = 2;
#endif

struct LocalFastLoadInfo
{
    bool                bInUse;
	bool				bPreloadAll;      // if set, uses dynamically-allocated buffer pucPreloadBuffer.
	bool				bPreloadFinished; // whether entire file is loaded in yet.
    
    // File state.
    u_int               uFileID;          // MS file ID.
    u_int               uFileSize;
#ifdef ASURA_PS2_CDEMUL
    u_int               uLSN;             // Logical sector number, as retrieved from the TOC cache.
#endif

    // EE-side buffer state.
	int					iReadBuffer;					// The buffer the user is reading from
	int					iWriteBuffer;					// The buffer that multistream is writting to
    u_char*             apucFLBuffers[FL_BUFFER_QTY];	// The buffers
	u_int				uDataInBuffer[FL_BUFFER_QTY];	// The amount of data in the buffer (if any)

	// preload state
	u_char*             pucPreloadBufferBase;
	u_char*             pucPreloadBufferCur;

    // Stream state.
    ASURA_PS2_STREAM_ID eStreamID;
    bool                bFirstLoad;			// True iff the file has been opened or Seek() has been called without a subsequent GetData() call.
    /*volatile*/ bool   bLoadReqMade;	    // True iff MS is currently writing into apucFLBuffers[iWriteBuffer]. ( this seems to work ok if this is not volatile)
    bool                bStopLoading;       // True iff we've had a GetData() call with bCancelLoad set to true.
    u_int               uDataWritten;		// The total amount of data written by multistream - used to determin EOF for multistream loads
	u_int               uDataRead;			// The total amount of data read by the user - used to determin EOF to user
    u_int               uTotalDataSize;		// Used for determining EOF. This is equal to (uFileSize - <last offset seeked to>).
};


#ifdef ASURA_ASSERT
static bool g_bFLInitd = false;
#endif

static LocalFastLoadInfo g_axFLFileInfo[ASURA_MAX_NO_OF_OPEN_FILES];
volatile bool g_bFastloadCDError = false;
volatile bool g_bFastloadDoneCDErrorPrintf = false;
static u_int g_uLastGetDataIOPCheckCycles;

static void FL_InitInfo( LocalFastLoadInfo* pxFileInfo, ASURA_PS2_STREAM_ID eStreamID )
{
    pxFileInfo->bInUse				= false;
    pxFileInfo->bPreloadAll			= false;
    pxFileInfo->bPreloadFinished	= false;
    pxFileInfo->eStreamID			= eStreamID;
    pxFileInfo->iWriteBuffer		= -1;
	pxFileInfo->iReadBuffer			= -1;
    pxFileInfo->bFirstLoad			= false;
    pxFileInfo->bLoadReqMade		= false;
	pxFileInfo->uDataWritten		= 0;
	pxFileInfo->uDataRead			= 0;
    pxFileInfo->uTotalDataSize		= 0;

    // Allocate IOP-side file buffer.
	SOUND_AllocateStreamBuffer( pxFileInfo->eStreamID, SOUND_DATA_STREAM, FL_BUFFER_SIZE );
    SendQueuedMultistreamCmds( true, 0 );

#ifdef ASURA_ASSERT
	// Check that succeeded.
	SOUND_STREAM_INFO xStreamInfo;
    SOUND_GetStreamInfo( pxFileInfo->eStreamID, &xStreamInfo );
    Asura_Assert( xStreamInfo.IOPAddress, "Not enough space on IOP for file buffer." );
#endif

    // Allocate EE-side file buffers.
    Asura_Assert( !(FL_BUFFER_SIZE & 63), "Buffers need to be cache-aligned." );
    for ( uint32 u=0; u<FL_BUFFER_QTY; u++ )
    {
        if( !(pxFileInfo->apucFLBuffers[u] = reinterpret_cast<u_char*>(Asura_AllocateAlign(FL_BUFFER_SIZE, ASURA_ALIGN_64_BYTE))) )
        {
            // Not very much point in dealing properly with this. We'll be
            // allocating these buffers very early and we're completely
            // screwed if it fails.
            Asura_Assert( false, "Out of memory." );
        }
		pxFileInfo->uDataInBuffer[u] = 0;
    }
}

static int FastLoadCallback( void )
{
    // Find out which stream multistream wants to tell us about
	SOUND_FASTLOAD_INFO xMSFLInfo;
    int iRetVal = SOUND_GetFastLoadInfo( &xMSFLInfo );
	Asura_Assert( iRetVal == 1, "FastLoadCallback called but no data has arrived" );

	// Check for cd errors - if we have one, the rest of the info will be rubbish
	if ( xMSFLInfo.Status >= 2 )
	{
		// (MH - 18.11.2004) printf in the callback seems to mess up error recovery :(
		//Asura_OutputToDebugger(("FastLoadCallback Status = %d.\n", xMSFLInfo.Status));

		// It's a cd error !
		g_bFastloadCDError = true;
		return SOUND_FASTLOAD_STOP;
	}

	// Which file is this ?
	LocalFastLoadInfo *pxFLInfo = NULL;

	// We can just calculate it at the moment
	Asura_Assert( (xMSFLInfo.StreamChannel - ASURA_PS2_STREAM_ID_FILELOAD_FIRST)<ASURA_MAX_NO_OF_OPEN_FILES, "Recieved data from an unknown stream." );
	pxFLInfo = &g_axFLFileInfo[xMSFLInfo.StreamChannel - ASURA_PS2_STREAM_ID_FILELOAD_FIRST];
	
	// are we managing the preloading of the entire file?
	if (pxFLInfo->bPreloadAll)
	{
		Asura_Assert(!pxFLInfo->bPreloadFinished, "already finished loading this file");

		u_int uLoadedBytes = xMSFLInfo.LoadSize;
		pxFLInfo->pucPreloadBufferCur += uLoadedBytes;
		Asura_Assert(pxFLInfo->pucPreloadBufferCur <= (pxFLInfo->pucPreloadBufferBase + pxFLInfo->uFileSize), "loaded past end of buffer");
		if (pxFLInfo->pucPreloadBufferCur == (pxFLInfo->pucPreloadBufferBase + pxFLInfo->uFileSize))
		{
			// finished loading
			pxFLInfo->bPreloadFinished = true;
			pxFLInfo->bLoadReqMade = false;
			// invalidate the DCache for all the data we've loaded.
			InvalidDCache( pxFLInfo->pucPreloadBufferBase, pxFLInfo->pucPreloadBufferCur - 1 ); // NB: The range is inclusive.
			pxFLInfo->pucPreloadBufferCur = pxFLInfo->pucPreloadBufferBase;
			return SOUND_FASTLOAD_STOP;
		}
		// continue loading into buffer
		return ( reinterpret_cast<int>(pxFLInfo->pucPreloadBufferCur) | SOUND_FASTLOAD_CONTINUE);
	}

	Asura_Assert( pxFLInfo->uDataWritten < xMSFLInfo.TotalSize, "We seem to have more data than MS has written" );

	// Update the write buffers index and flags
	pxFLInfo->uDataInBuffer[pxFLInfo->iWriteBuffer] = xMSFLInfo.LoadSize;
	pxFLInfo->uDataWritten += xMSFLInfo.LoadSize;
	pxFLInfo->iWriteBuffer++;
	pxFLInfo->iWriteBuffer %= FL_BUFFER_QTY;
	pxFLInfo->bLoadReqMade = false;

    int iReturnValue;

	// Should we request more data for this buffer ?
#if USE_CONTINUE_RET_VAL
	if( pxFLInfo->iWriteBuffer != pxFLInfo->iReadBuffer && xMSFLInfo.TotalSize != pxFLInfo->uTotalDataSize && !pxFLInfo->bStopLoading )
	{
		// Ok, we've got space, some data left to load and no request to stop
        // so let's load some more.
        iReturnValue = ( reinterpret_cast<int>(pxFLInfo->apucFLBuffers[pxFLInfo->iWriteBuffer]) | SOUND_FASTLOAD_CONTINUE );
		pxFLInfo->bLoadReqMade = true;
	}
    else
#endif
    {
        iReturnValue = SOUND_FASTLOAD_STOP;
    }

    return iReturnValue;
}

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

void Asura_PS2_Multistream_FastLoad::CloseFile( Asura_Handle iHandle )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );
    Asura_Assert( -1 != g_axFileInfo[iHandle].iHandle, "No file open for this file." );
    
    sceClose( g_axFileInfo[iHandle].iHandle );
    g_axFileInfo[iHandle].iHandle = -1;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Initialise() failed or was not called." );
    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );

    LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
    Asura_Assert( xFLInfo.bInUse, "No file open for this handle." );

    // If there's an outstanding load, wait for it to complete.
    if( xFLInfo.bLoadReqMade )
    {
        const void* pvDummy;
        while( !GetData(iHandle, &pvDummy, true) ) { }
    }

    SOUND_StopStream( xFLInfo.eStreamID );

    CloseMSFile( xFLInfo.uFileID );
    
    SendQueuedMultistreamCmds( true, 0 );

    xFLInfo.bInUse = false;

	if ( xFLInfo.bPreloadAll )
	{
		// cleanup preload data
		xFLInfo.bPreloadAll = false;
		Asura_DeallocateAlign( xFLInfo.pucPreloadBufferBase );
		xFLInfo.pucPreloadBufferBase = NULL;
		xFLInfo.pucPreloadBufferCur = NULL;
	}

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

	--s_uNumOpenFiles;
}

int Asura_PS2_Multistream_FastLoad::GetData( Asura_Handle iHandle, const void** ppvBfr, bool bCancelLoad /*= false*/ )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );
    Asura_Assert( -1 != g_axFileInfo[iHandle].iHandle, "No file open for this file." );

    int iBytes = sceRead( g_axFileInfo[iHandle].iHandle, g_axFileInfo[iHandle].acData, FileInfo::BUFFER_SIZE );

    if( iBytes > 0 )
    {
        *ppvBfr = g_axFileInfo[iHandle].acData;
    }

    return Asura_Maths::Max( 0, iBytes );

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Initialise() failed or was not called." );
    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );

	// Which file are we dealing with this time
	LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
    Asura_Assert( xFLInfo.bInUse, "No file open for this handle." );

	if ( xFLInfo.bPreloadAll )
	{
		Asura_Assert(!bCancelLoad, "cancel load not supported for preloaded files yet");
		if (xFLInfo.bPreloadFinished)
		{
			// if we've "seeked" into the file these will have been modified accordingly.
			// try emulating the normal load by returning chunks of FL_BUFFER_SIZE
			int iBytesAvailable = static_cast<int>(Asura_Maths::Min(FL_BUFFER_SIZE, xFLInfo.uTotalDataSize));
			*ppvBfr = xFLInfo.pucPreloadBufferCur;
			xFLInfo.uTotalDataSize -= iBytesAvailable;
			xFLInfo.pucPreloadBufferCur += iBytesAvailable;
			return iBytesAvailable;
		}
		else
		{
			return 0;
		}
	}
	
	// Check for cd errors .. this seems a horrible place to do this, but we get stuck in infinite loops if we don't do.
	// (MH - 18.11.2004) I've had to change a couple of things to get this to work:-
	// 1) only do the FASTLOAD CD ERROR printf once for each error. (avoids tying up the IOP)
	// 2) call UpdateIOPBlocking() before SOUND_HandleCDErrors() to setup the globals that the latter uses.
	// 3) call that every so often as you don't get a fastload callback with the error in some situations unless it called.
	// 4) don't tie up the IOP by checking this more than once per second
	// 5) don't printf inside the callback.
	u_int uCurCycles = Asura_PS2_Timers::GetCycleCount();
	// call once per second
	if (Asura_PS2_Timers::GetElapsedCycles( g_uLastGetDataIOPCheckCycles, uCurCycles ) > u_int(ASURA_PS2_EE_CYCLES_PER_SEC))
	{
		g_uLastGetDataIOPCheckCycles = uCurCycles;
		if ( g_bFastloadCDError )
		{
			if (!g_bFastloadDoneCDErrorPrintf)
			{
				printf("FASTLOAD CD ERROR\n" );
				g_bFastloadDoneCDErrorPrintf = true;
			}
			Asura_PS2_Multistream::UpdateIOPBlocking();
			if ( SOUND_HandleCDErrors() == 0 )
			{
				// Ok, we've recovered from the error
				g_bFastloadCDError = false;
				g_bFastloadDoneCDErrorPrintf = false;
			}
			SOUND_FlushIOPCommand(0);	
		}
		else 
		{
			SOUND_FlushIOPCommand(SIF_RPCM_NOWAIT);
			SOUND_HandleCDErrors();
		}
	}

	// Have we reached the end of this file yet ?
	if ( xFLInfo.uDataRead == xFLInfo.uTotalDataSize )
	{
		Asura_Assert( xFLInfo.bLoadReqMade == false, "More data requested, but we are at the end of the file" );
		return REACHED_EOF;
	}

	// Are we waiting for the load to be cancelled ?
    xFLInfo.bStopLoading = bCancelLoad;
	if( bCancelLoad )
	{
		// We are.. has it stopped now ?
		if ( xFLInfo.bLoadReqMade )
		{	// Nope, it's still reading
			return 0;
		}
		else
		{	// Yup It's stopped
			return 1;
		}
	}

	// Is this the first request for data from this stream
	if( xFLInfo.bFirstLoad )
	{
		// It is, so start loading
		xFLInfo.bFirstLoad = false;
		xFLInfo.bLoadReqMade = true;	// Mark the load as requested before we request it, just in case it completes really quickly
		xFLInfo.iWriteBuffer = 0;
		const u_int uWriteAddr = reinterpret_cast<u_int>(xFLInfo.apucFLBuffers[xFLInfo.iWriteBuffer]);
		SOUND_LoadFile( STREAM_LOAD_EE, xFLInfo.eStreamID, xFLInfo.uFileID, uWriteAddr );
		SendQueuedMultistreamCmds( true, 0 );
	}

	// Ok, the user really does want data let's see if we have any
	int iNumBytes = 0;

	// Is the next buffer ready for the user
	if ( xFLInfo.uDataInBuffer[(xFLInfo.iReadBuffer+1) % FL_BUFFER_QTY] != 0 )
	{
		// It is... mark the last buffer (if any) as free for loading
		if ( xFLInfo.iReadBuffer != -1 )
		{
			// It's now unused
			xFLInfo.uDataInBuffer[xFLInfo.iReadBuffer] = 0;
		}
		// Update the buffer pointers
		xFLInfo.iReadBuffer++;
		xFLInfo.iReadBuffer %= FL_BUFFER_QTY;
        // (MP - 21.01.2003) NB: Returning a UCA ptr is actually *slower*
        // here because much of the data we're reading needs to be written
        // out again. *Any* store operation will invalidate the contents of
        // the UCAB, rendering it pretty much useless here.
		u_char* pucBfrToReturn = xFLInfo.apucFLBuffers[xFLInfo.iReadBuffer];
        *ppvBfr = pucBfrToReturn;
        // We need to invalidate the relevant cache lines before returning.
        // Our buffers are small enough that they'll stay in the cache a lot
        // of the time and the cache won't know that we've DMA'd a bunch of
        // new data into the relevant region of memory.
        InvalidDCache( pucBfrToReturn, pucBfrToReturn + (FL_BUFFER_SIZE - 1) ); // NB: The range is inclusive.
		// Tell them how much data they've got
		iNumBytes = xFLInfo.uDataInBuffer[xFLInfo.iReadBuffer];
		// And keep track of how much data we have given them in total
		xFLInfo.uDataRead += iNumBytes;
	}

	// Should we start another load for this file
	if ( (!xFLInfo.bLoadReqMade) &&							    // If this file isn't loading at the moment
		 (xFLInfo.iWriteBuffer != xFLInfo.iReadBuffer) &&		// And we have room for more data
		 (xFLInfo.uDataWritten != xFLInfo.uTotalDataSize) )		// And we haven't reached the end of the file yet
	{
		// Start data loading into the next buffer
		xFLInfo.bLoadReqMade = true;	// Mark the load as requested before we request it, just in case it completes really quickly
		const u_int uWriteAddr = reinterpret_cast<u_int>(xFLInfo.apucFLBuffers[xFLInfo.iWriteBuffer]);
		SOUND_ModifyEETransferAddress( xFLInfo.eStreamID, uWriteAddr );
		SOUND_SetFastLoadNext( xFLInfo.eStreamID );
		SendQueuedMultistreamCmds( true, SIF_RPCM_NOWAIT );
	}

    return iNumBytes;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

u_int Asura_PS2_Multistream_FastLoad::GetFilePos( Asura_Handle iHandle )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	return 0;
#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Initialise() failed or was not called." );
    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );

    const LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
    Asura_Assert( xFLInfo.bInUse, "No file open for this handle." );

	if (xFLInfo.bPreloadAll)
	{
		Asura_Assert(xFLInfo.bPreloadFinished, "preload not finished!");
		return xFLInfo.pucPreloadBufferCur - xFLInfo.pucPreloadBufferBase;
	}
	
    u_int uLastSeekOffset = xFLInfo.uFileSize - xFLInfo.uTotalDataSize;
    return uLastSeekOffset + xFLInfo.uDataRead;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

u_int Asura_PS2_Multistream_FastLoad::GetFileSize( Asura_Handle iHandle )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );
    Asura_Assert( -1 != g_axFileInfo[iHandle].iHandle, "No file open for this file." );

    int iOldPos = sceLseek( g_axFileInfo[iHandle].iHandle, 0, SCE_SEEK_CUR );
    if( iOldPos >= 0 )
    {
        int iFileSize = sceLseek( g_axFileInfo[iHandle].iHandle, SCE_SEEK_END, 0 );
        sceLseek( g_axFileInfo[iHandle].iHandle, SCE_SEEK_SET, iOldPos );
        return iFileSize;
    }

    return 0;

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Initialise() failed or was not called." );
    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );

    const LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
    Asura_Assert( xFLInfo.bInUse, "No file open for this handle." );

    return xFLInfo.uFileSize;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

bool Asura_PS2_Multistream_FastLoad::Initialise( void )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        g_axFileInfo[u].iHandle = -1;
    }

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    
    Asura_Assert( !g_bFLInitd, "Duplicate call to Initialise()?" );
	// Initialise the file structures
    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        FL_InitInfo( &g_axFLFileInfo[u], static_cast<ASURA_PS2_STREAM_ID>(ASURA_PS2_STREAM_ID_FILELOAD_FIRST + u) );
    }

	// Add the FastLoad callback
	SOUND_SetFastLoadCallback( FastLoadCallback );

#ifdef ASURA_ASSERT
    g_bFLInitd = true;
#endif

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
    
    return true;
}

bool Asura_PS2_Multistream_FastLoad::OpenFile( Asura_Handle iHandle, const char* szFilename )
{

#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );
    Asura_Assert( -1 == g_axFileInfo[iHandle].iHandle, "File already open for this handle." );

    int iSceHandle = sceOpen( Asura_PS2_OS::GetFullPath(szFilename, true), SCE_RDONLY );
    if( iSceHandle >= 0 )
    {
        g_axFileInfo[iHandle].iHandle = iSceHandle;
		++s_uNumOpenFiles;
        return true;
    }
    else
    {
        return false;
    }

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Initialise() failed or was not called." );
    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );

    LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
    Asura_Assert( !xFLInfo.bInUse, "File already open for this handle." );

#ifdef ASURA_PS2_CDEMUL
    u_int uFileID = SetupFileInfoForMS( szFilename, false, 0, &xFLInfo.uFileSize, &xFLInfo.uLSN );
#else
    u_int uFileID = SetupFileInfoForMS( szFilename, false, 0, &xFLInfo.uFileSize );
#endif
    if( INVALID_MS_FILE_ID == uFileID )
    {
        return false;
    }
	++s_uNumOpenFiles;

    xFLInfo.bInUse			= true;
	xFLInfo.bPreloadAll     = false;
	xFLInfo.uFileID			= uFileID;
    xFLInfo.bFirstLoad		= true;
    xFLInfo.bLoadReqMade	= false;
    xFLInfo.bStopLoading    = false;
	xFLInfo.uDataWritten	= 0;
	xFLInfo.uDataRead		= 0;
    xFLInfo.uTotalDataSize	= xFLInfo.uFileSize;
    xFLInfo.iWriteBuffer	= -1;
	xFLInfo.iReadBuffer		= -1;

	for ( uint32 u=0; u<FL_BUFFER_QTY; u++ )
	{
		xFLInfo.uDataInBuffer[u] = 0;
    }

    SendQueuedMultistreamCmds( true, 0 );

    return true;

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

bool Asura_PS2_Multistream_FastLoad::OpenFileForPreload( Asura_Handle iHandle, const char* szFilename )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	Asura_Assert(false, "not supported");
	return false;
#else
	bool bOK = OpenFile( iHandle, szFilename );
	if (bOK)
	{
		LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
		
		xFLInfo.bPreloadAll			= true;
		xFLInfo.bPreloadFinished	= false;
		xFLInfo.pucPreloadBufferBase = reinterpret_cast<u_char*>(Asura_AllocateAlign((xFLInfo.uFileSize + 63) & ~63, ASURA_ALIGN_64_BYTE));
		if (!xFLInfo.pucPreloadBufferBase)
		{
			// out of memory
			CloseFile( iHandle );
			return false;
		}
		xFLInfo.pucPreloadBufferCur = xFLInfo.pucPreloadBufferBase;

		const u_int uWriteAddr = reinterpret_cast<u_int>(xFLInfo.pucPreloadBufferBase);
		SOUND_LoadFile( STREAM_LOAD_EE, xFLInfo.eStreamID, xFLInfo.uFileID, uWriteAddr );
		xFLInfo.bLoadReqMade = true;
		SendQueuedMultistreamCmds( true, 0 );
	}
	return bOK;
#endif
}

bool Asura_PS2_Multistream_FastLoad::IsPreloadFinished( Asura_Handle iHandle )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	Asura_Assert(false, "not supported");
	return false;
#else
	LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
	Asura_Assert(xFLInfo.bPreloadAll, "not opened with preload");
	return xFLInfo.bPreloadFinished;
#endif
}

u_int Asura_PS2_Multistream_FastLoad::GetNumBytesPreloaded( Asura_Handle iHandle )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	Asura_Assert(false, "not supported");
	return 0;
#else
	LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
	Asura_Assert(xFLInfo.bPreloadAll, "not opened with preload");
	if (xFLInfo.bPreloadFinished)
	{
		return xFLInfo.uFileSize;
	}
	else
	{
		return xFLInfo.pucPreloadBufferCur - xFLInfo.pucPreloadBufferBase;
	}
#endif
}

void Asura_PS2_Multistream_FastLoad::Seek( Asura_Handle iHandle, u_int uByteOffset )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );
    Asura_Assert( -1 != g_axFileInfo[iHandle].iHandle, "No file open for this file." );

    sceLseek( g_axFileInfo[iHandle].iHandle, uByteOffset, SCE_SEEK_SET );

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Initialise() failed or was not called." );
    Asura_Assert( 0 <= iHandle && iHandle < ASURA_MAX_NO_OF_OPEN_FILES, "Invalid handle specified." );

    LocalFastLoadInfo& xFLInfo = g_axFLFileInfo[iHandle];
    Asura_Assert( xFLInfo.bInUse, "No file open for this handle." );

    // Also verify the offset is valid. Technically seeking to EOF could be
    // valid here but I'll disallow it for now to pick up logic errors, etc.
    Asura_Assert( uByteOffset < xFLInfo.uFileSize, "Invalid offset specified." );

	if (xFLInfo.bPreloadAll)
	{
		Asura_Assert(xFLInfo.bPreloadFinished, "preload not finished");
		// fix up data that will be returned by next call to GetData
		xFLInfo.uTotalDataSize = xFLInfo.uFileSize - uByteOffset;
		xFLInfo.pucPreloadBufferCur = xFLInfo.pucPreloadBufferBase + uByteOffset;
		return;
	}

    // If there's an outstanding load, wait for it to complete.
    if( xFLInfo.bLoadReqMade )
    {
        const void* pvDummy;
        while( !GetData(iHandle, &pvDummy, true) ) { }
    }

    u_int uRemainingData = xFLInfo.uFileSize - uByteOffset;

    SOUND_SetFileOffsetAndPlaySize( xFLInfo.uFileID, uByteOffset, uRemainingData );
    SendQueuedMultistreamCmds( true, 0 );

    xFLInfo.bFirstLoad		= true;
    xFLInfo.bLoadReqMade	= false;
    xFLInfo.bStopLoading    = false;
    xFLInfo.uDataWritten	= 0;
	xFLInfo.uDataRead		= 0;
    xFLInfo.uTotalDataSize	= uRemainingData;
    xFLInfo.iWriteBuffer	= -1;
	xFLInfo.iReadBuffer		= -1;
	for ( uint32 u=0; u<FL_BUFFER_QTY; u++ )
	{
		xFLInfo.uDataInBuffer[u] = 0;
    }

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

void Asura_PS2_Multistream_FastLoad::Shutdown( void )
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        Asura_Assert( -1 == g_axFileInfo[u].iHandle, "File wasn't closed." );
    }

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

    Asura_Assert( g_bFLInitd, "Shutdown() when Initialise() failed or was not called." );

    // All files should have been closed externally by now.
    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        Asura_Assert( !g_axFLFileInfo[u].bInUse, "File wasn't closed." );
    }

	// Remove the fastload callback
	SOUND_SetFastLoadCallback( NULL );

    for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
    {
        for( u_int v = 0; v < 2; ++v )
        {
            if( g_axFLFileInfo[u].apucFLBuffers[v] )
            {
                Asura_DeallocateAlign( g_axFLFileInfo[u].apucFLBuffers[v] );
                g_axFLFileInfo[u].apucFLBuffers[v] = NULL;
            }
        }
    }

#ifdef ASURA_ASSERT
    g_bFLInitd = false;
#endif

#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

// functions to allow sceOpen to be used without conflict
void Asura_PS2_Multistream_FastLoad::PauseFastload()
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	if (!AreAnyFilesOpen()) return;

	Asura_Assert(s_uLastFastloadMode == 0xffffffff, "fastload not running");

	SOUND_FASTLOAD_INFO xMSFLInfo;
	SOUND_GetFastLoadInfo( &xMSFLInfo );
	s_uLastFastloadMode = xMSFLInfo.EEStatus;

	LocalFastLoadInfo *pxFLInfo = NULL;
	Asura_Assert( (xMSFLInfo.StreamChannel - ASURA_PS2_STREAM_ID_FILELOAD_FIRST)<ASURA_MAX_NO_OF_OPEN_FILES, "fastload operating on an unknown stream." );
	pxFLInfo = &g_axFLFileInfo[xMSFLInfo.StreamChannel - ASURA_PS2_STREAM_ID_FILELOAD_FIRST];

	volatile bool bStillLoading = pxFLInfo->bLoadReqMade;
	while (bStillLoading)
	{
		Asura_PS2_OS::DelayThread(1);
		bStillLoading = pxFLInfo->bLoadReqMade;
	}

	SOUND_SetFastLoadMode(SOUND_FASTLOAD_OFF);
	SendQueuedMultistreamCmds( true, 0 );
#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}

void Asura_PS2_Multistream_FastLoad::ResumeFastload()
{
#if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM

#else // not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
	if (!AreAnyFilesOpen()) return;

	Asura_Assert(s_uLastFastloadMode != 0xffffffff, "fastload not paused");
	SOUND_SetFastLoadMode(s_uLastFastloadMode);
	SendQueuedMultistreamCmds( true, 0 );

	s_uLastFastloadMode = 0xffffffff;
#endif // else not #if ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM
}



//
// MICFileHeaderManager.
//

#ifdef ASURA_PS2_CDEMUL
MICFileHeaderManager::MICFileData* MICFileHeaderManager::s_pxMICFileData;
u_int                             MICFileHeaderManager::s_uNumHdrs;
#endif

bool MICFileHeaderManager::GetHeaderAndFilesize( const char* szFullPath, const SOUND_MIH_HEADER*& pxHeader, u_int& uSize )
{
#ifdef ASURA_PS2_CDEMUL
    // Skip "cdrom0:"...
    if( !strncmp(szFullPath, "cdrom0:", 7) ) { szFullPath += 7; }

    const Asura_Hash_ID uHash = Asura_GetHashID( szFullPath );
    for( u_int u = 0; u < s_uNumHdrs; ++u )
    {
        if( uHash == s_pxMICFileData[u].uFilenameHash )
        {
            pxHeader = &s_pxMICFileData[u].xHdr;
			uSize = s_pxMICFileData[u].uFileSize;
			return true;
        }
    }

#else // not #ifdef ASURA_PS2_CDEMUL

    // Assume caller won't require more than one hdr at a time. This is only
    // used by MS streaming sound stuff at the moment and that's the case
    // there.
    static SOUND_MIH_HEADER xHdr;

    int iHandle = sceOpen( szFullPath, SCE_RDONLY );
    if( iHandle >= 0 )
    {
        sceRead( iHandle, &xHdr, sizeof(SOUND_MIH_HEADER) );

        Asura_Assert( xHdr.NumberOfFiles == 2, "Streaming files should be stereo." );
        // could change this if we want to handle mono streams

		// get filesize while we have the file open
		sceLseek( iHandle, 0, SCE_SEEK_SET );
		uSize = sceLseek( iHandle, 0, SCE_SEEK_END );

        sceClose( iHandle );

        pxHeader = &xHdr;
		return true;
    }

#endif // else not #ifdef ASURA_PS2_CDEMUL

    Asura_Assertf( false, ("File \"%s\" not found.", szFullPath) );
    return false;
}

void MICFileHeaderManager::Initialise( void )
{
#ifdef ASURA_PS2_CDEMUL

    const Asura_PS2_TOCCache::Entry* pxEntries = Asura_PS2_TOCCache::GetEntryList();

    const u_int uNumEntries = Asura_PS2_TOCCache::GetNumEntries();

    const u_int uSearchExt = (static_cast<u_int>('M') <<  0)
                           | (static_cast<u_int>('I') <<  8)
                           | (static_cast<u_int>('C') << 16);

    s_uNumHdrs = 0;
    for( u_int uEntry = 0; uEntry < uNumEntries; ++uEntry )
    {
        if( uSearchExt == pxEntries[uEntry].m_uExtAsUint )
        {
            ++s_uNumHdrs;
        }
    }

    if( !s_uNumHdrs ) { return; }

    s_pxMICFileData = new MICFileData[s_uNumHdrs];

    u_char* pucTmpSectorBuffer = reinterpret_cast<u_char*>( Asura_AllocateAlign(2048, ASURA_ALIGN_64_BYTE) );

	sceCdRMode xReadMode;
	xReadMode.trycount    = 0;
	xReadMode.spindlctrl  = SCECdSpinNom;
	xReadMode.datapattern = SCECdSecS2048;       

    sceCdSync( 0 );

    // Assume hdr fits in a single sector.
    Asura_Assert( sizeof(SOUND_MIH_HEADER) <= 2048, "" );

    u_int uHdrOut = 0;
    for( u_int uEntry = 0; uEntry < uNumEntries; ++uEntry )
    {
        const Asura_PS2_TOCCache::Entry& xNextEntry = pxEntries[uEntry];
        if( uSearchExt == xNextEntry.m_uExtAsUint )
        {
            do
            {
                sceCdDiskReady( 0 );
                while( !sceCdRead(xNextEntry.m_uFirstSector, 1, pucTmpSectorBuffer, &xReadMode) ) { }
                sceCdSync( 0 );
            }
            while( SCECdErNO != sceCdGetError() );

            memcpy( &s_pxMICFileData[uHdrOut].xHdr, pucTmpSectorBuffer, sizeof(SOUND_MIH_HEADER) );

            Asura_Assertf( s_pxMICFileData[uHdrOut].xHdr.NumberOfFiles == 2, ("Streaming files should be stereo. File entry %d isn't.", uEntry) );
            // could change this if we want to handle mono streams

            s_pxMICFileData[uHdrOut].uFilenameHash = xNextEntry.m_uFilenameHash;
			s_pxMICFileData[uHdrOut].uFileSize = xNextEntry.m_uLengthInBytes;

            ++uHdrOut;
        }
    }

    Asura_Assert( uHdrOut == s_uNumHdrs, "" );

    Asura_DeallocateAlign( pucTmpSectorBuffer );
    
#endif // #ifdef ASURA_PS2_CDEMUL
}

void MICFileHeaderManager::Shutdown( void )
{
#ifdef ASURA_PS2_CDEMUL
    delete[] s_pxMICFileData;
#endif
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void CloseMSFile( u_int uFileID )
{
    if( Asura_PS2_OS::IsRunningUnderAtwinmon() )
    {
        SOUND_ATWinMonCloseFile( uFileID );
    }

	// (AMM - 15 Aug 2008) 
	// dont pass invalid file handles to the IOP. Although it initially looks like
	// the IOP handles this it does issue errors. Over the course of time
	// this causes more severe errors to get reported and eventually a hang
	// a simple check prevents the errors from occuring. Note that I could of
	// put this into the SOUND_FreeFileID function in sound.c as this fix is in
	// two places, but as the middleware may get replaced I though it better here.
	//
	if (uFileID != INVALID_MS_FILE_ID)
	{
	    SOUND_FreeFileID( uFileID );
	}
}

#ifdef ASURA_PS2_CDEMUL

static u_int SetupFileInfoForMS( const char* szFilename,
                                 bool        bFilenameHasFullPath,
                                 u_int       uOffset /*= 0*/,
                                 u_int*      puFileSize /*= NULL*/,
                                 u_int*      puLSN /*= NULL*/ )
{
    if( !bFilenameHasFullPath )
    {
        szFilename = Asura_PS2_OS::GetFullPath( szFilename );
    }

    sceCdlFILE xCDFileInfo;
    if( !Asura_PS2_TOCCache::CdSearchFile(&xCDFileInfo, szFilename) )
    {
// (MP - 03.08.2005) I'm going to remove this for now, because it kind of
// breaks Asura_FileAccess::FileExists() by complaining every time we check
// for a file. We maybe want to selectively disable output, depending on
// whether or not we're just checking, or perhaps add a Platform_FileExists().
//      Asura_OutputToDebugger_Retail(( "File \"%s\" not found.\n", szFilename ));
        return INVALID_MS_FILE_ID;
    }

    if( puFileSize ) { *puFileSize = xCDFileInfo.size; }
    if( puLSN      ) { *puLSN      = xCDFileInfo.lsn;  }

    return SOUND_SetFileInfo( SOUND_CREATE_FILE, xCDFileInfo.lsn, xCDFileInfo.size, ADPCM_DATA_ON_CD, uOffset );
}

#else // not def ASURA_PS2_CDEMUL

static u_int SetupFileInfoForMS( const char* szFilename,
                                 bool        bFilenameHasFullPath,
                                 u_int       uOffset /*= 0*/,
                                 u_int*      puFileSize /*= NULL*/ )
{
    if( !bFilenameHasFullPath )
    {
        szFilename = Asura_PS2_OS::GetFullPath( szFilename );
    }

    u_int uFileSize;

    // Check the file exists. Doesn't currently seem to be any way of
    // determining this via the MS interface.
    {
        int i = sceOpen( szFilename, SCE_RDONLY );
        if( i < 0 )
        {
// (MP - 03.08.2005) I'm going to remove this for now, because it kind of
// breaks Asura_FileAccess::FileExists() by complaining every time we check
// for a file. We maybe want to selectively disable output, depending on
// whether or not we're just checking, or perhaps add a Platform_FileExists().
//          Asura_OutputToDebugger_Retail(( "Unable to open file \"%s\" (file not found.)\n", szFilename ));
            return INVALID_MS_FILE_ID;
        }
        else
        {
            // Get file size while we've got the file open.
            uFileSize = sceLseek( i, 0, SCE_SEEK_END );

			// JAT 220206 - if sceLseek returns -ve number then an erorr occured
			int iErrNo = static_cast<int>( uFileSize );
			if ( iErrNo < 0 )
			{
#if SCE_LIBRARY_VERSION	>=0x3020  
				iErrNo = sceStdioConvertError( SCE_STDIO_FUNC_ANYTHING, uFileSize );
#endif
				Asura_OutputToDebugger_Retail(( "Error getting file size '%s' Error 0x%08X/0x%08X\n", szFilename, -uFileSize, iErrNo ));
#if SCE_LIBRARY_VERSION	>=0x3020  
				Asura_PS2_OS::DecodeSceeErrNo( iErrNo );
#endif
				return INVALID_MS_FILE_ID;
			}

            sceClose( i );
        }
    }

    if( puFileSize ) { *puFileSize = uFileSize; }

    u_int uFileID = SOUND_SetFileInfo( SOUND_CREATE_FILE, 0, uFileSize, ADPCM_DATA_ON_HOST, uOffset );
    SOUND_StoreHOST0FileName( uFileID, szFilename );
    if( Asura_PS2_OS::IsRunningUnderAtwinmon() )
    {
        SOUND_ATWinMonOpenFile( uFileID );
    }

    return uFileID;
}

#endif  // else not def ASURA_PS2_CDEMUL
