//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_FractalSkyDefines.h
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Notes:
//
//     - This file is intended to be shared between the EE and VU1 sides of
//       the PS2 runtime code and also tools export code.
// 
//     - ST generation parameter QW should consist of the following:
//
//           x = p2
//           y = -p1
//           z = 0.9
//           w = don't care
//
//       where p1 and p2 are the phase parameters from the PC version.
//
//     - ADC bits should be stored in the ST QW's w component. They are
//       assumed to be clean, ie. low 16 bits are 0x8000 iff the vertex is
//       degenerate and 0 iff it's not.
// 
// VU memory usage:
//
//     Buffer sizes, etc.:
//
//         Renderstate     = 3
//         Cam data buffer = 4 * 9
//         Input buffers   = 2 * 2N
//         Output buffers  = 2 * (5 + 2N)
//         Clip input bfr  = 2N
//         VI reg stack    = 2
// 
//     Max vertices:
//
//         3 + 4 * 9 + 2 * 2N + 2 * (5 + 2N) + 2N + 2 <= 1024
//             =>
//         51 + 10N <= 1024
//             =>
//         N <= 97.3
//
//         To allow plenty of space for future changes, set N=70. That's
//         more than enough anyway.
//
//     Memory layout:
//
//             0   /--------------------------\
//                 |        renderstate       |    3 QWs - (ALPHA) * 3
//             3   |--------------------------|
//                 |         cam data         |   36 QWs - (SCISSOR, XYOFFSET, W->GBC mat, GBC->S scale, GBC->S trans, GBC->SC scale) * 4
//            39   |--------------------------|
//                 |      input buffers       |  280 QWs - (XYZ, ST+ADC) * N * 2
//           319   |--------------------------|
//                 |      output buffers      |  290 QWs - (GIF tag, XYOFFSET, SCISSOR, ALPHA, GIF tag, (ST, XYZ+ADC) * N) * 2
//           609   |--------------------------|
//                 |     clip input buffer    |  140 QWs - (XYZW, ST + should-clip flag) * N
//           749   |--------------------------|
//                 |       VI reg stack       |    2 QWs
//           751   |--------------------------|
//                 |         <unused>         |  275 QWs
//          1024   \--------------------------/
//
//      Memory layout of the output buffers during clipping:
//
//                 :                          :
//             K   |--------------------------|
//                 |   clipped output buffer  | 5+2M QWs - (GIF tag, SCISSOR, XYOFFSET, ALPHA, GIF tag, (ST, XYZ+ADC) * M)
//                 |--------------------------|
//                 |    clipper workspace 1   |   36 QWs
//                 |--------------------------|
//                 |    clipper workspace 2   |   36 QWs
//                 |--------------------------|
//                 :                          :
//
//          where M is the maximum number of vertices that will fit.
//
//              5+2M + 36 + 36 <= 215
//                  =>
//              77 + 2M <= 145
//                  =>  .
//              M <= 34
//
//          So set M=34.
//
//          (NB: M must be at least 9 since just a single input tri can
//          generate that many output vertices.)
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_FRACTALSKYDEFINES_H___
#define ___ASURA_PS2_FRACTALSKYDEFINES_H___ 1

// M A C R O S ///////////////////////////////////////////////////////////////

// Version history:
//     (MP - 25.06.2002) - 1 - Original format.
//     (MP - 19.11.2003) - 2 - Changed UNPACK stuff for new renderer.
#define ASURA_PS2_FS__CHUNK_VERSION                   (2)

// Limits.
#define ASURA_PS2_FS__MAX_VERTS_PER_STRIP             (70)
#define ASURA_PS2_FS__MAX_CAMERAS                     (4)

// These match the VU1 clipping bits.
#define ASURA_PS2_FS__CLIP_RIGHT                      (0x001)
#define ASURA_PS2_FS__CLIP_LEFT                       (0x002)
#define ASURA_PS2_FS__CLIP_BOTTOM                     (0x004)
#define ASURA_PS2_FS__CLIP_TOP                        (0x008)
#define ASURA_PS2_FS__CLIP_FAR                        (0x010)
#define ASURA_PS2_FS__CLIP_NEAR                       (0x020)
#define ASURA_PS2_FS__CLIP_ALL                        (0x03F)

// Data area addresses.
#define ASURA_PS2_FS__ADDR__RENDERSTATE               (0)
#define ASURA_PS2_FS__ADDR__CAM_DATA                  (3)
#define ASURA_PS2_FS__ADDR__INPUT_BFR_0               (39)
#define ASURA_PS2_FS__ADDR__INPUT_BFR_1               (179)
#define ASURA_PS2_FS__ADDR__OUTPUT_BFR_0              (319)
#define ASURA_PS2_FS__ADDR__OUTPUT_BFR_1              (464)
#define ASURA_PS2_FS__ADDR__CLIP_INPUT_BFR            (609)
#define ASURA_PS2_FS__ADDR__VI_REG_STACK              (749)

// Data area sizes.
#define ASURA_PS2_FS__SIZE__RENDERSTATE               (3)
#define ASURA_PS2_FS__SIZE__CAM_DATA                  (36)
#define ASURA_PS2_FS__SIZE__INPUT_BFR                 (140)
#define ASURA_PS2_FS__SIZE__OUTPUT_BFR                (145)
#define ASURA_PS2_FS__SIZE__CLIP_INPUT_BFR            (140)
#define ASURA_PS2_FS__SIZE__VI_REG_STACK              (2)

// Renderstate buffer details.
#define ASURA_PS2_FS__OFFSET__RS_ALPHA_PASS_0         (0)
#define ASURA_PS2_FS__OFFSET__RS_ALPHA_PASS_1         (1)
#define ASURA_PS2_FS__OFFSET__RS_ALPHA_PASS_2         (2)

// Cam data buffer details.
#define ASURA_PS2_FS__OFFSET__CAM_SCISSOR             (0)
#define ASURA_PS2_FS__OFFSET__CAM_XYOFFSET            (1)
#define ASURA_PS2_FS__OFFSET__CAM_V2GBC_MAT           (2)
#define ASURA_PS2_FS__OFFSET__CAM_GBC2S_SCALE         (6)
#define ASURA_PS2_FS__OFFSET__CAM_GBC2S_TRANS         (7)
#define ASURA_PS2_FS__OFFSET__CAM_GBC2SC_SCALE        (8)
#define ASURA_PS2_FS__OFFSET__CAM_NEXT_CAM            (9)

// Output buffer details.
#define ASURA_PS2_FS__OFFSET__OUTPUT_SETUP_GIF_TAG    (0)
#define ASURA_PS2_FS__OFFSET__OUTPUT_SCISSOR          (1)
#define ASURA_PS2_FS__OFFSET__OUTPUT_XYOFFSET         (2)
#define ASURA_PS2_FS__OFFSET__OUTPUT_ALPHA            (3)
#define ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG     (4)
#define ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX     (5)

// Clipper workspace buffer details.
#define ASURA_PS2_FS__OFFSET__CLIP_OUTPUT_BFR         (0)
#define ASURA_PS2_FS__OFFSET__CLIP_WORKSPACE          (73)
#define ASURA_PS2_FS__SIZE__CLIP_OUTPUT_BFR           (73)
#define ASURA_PS2_FS__SIZE__CLIP_WORKSPACE            (36)

// VI register stack aliases.
#define ASURA_PS2_FS__ADDR__VI_STACK_0                (ASURA_PS2_FS__ADDR__VI_REG_STACK + 0)
#define ASURA_PS2_FS__ADDR__VI_STACK_1                (ASURA_PS2_FS__ADDR__VI_REG_STACK + 1)

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_FRACTALSKYDEFINES_H___
