//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Skybox.cpp
//
// Copyright (c) 2001-2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "SpecialFX/Asura_Platform_Skybox.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"        
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Render.h"         
#include "Rendering/Asura_RenderStates.h"   
#include "SpecialFX/Asura_Skybox.h"
#include "SpecialFX/Asura_Sphere.h"

// Engine platform-specific.
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_PS2_DMA.h"             
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_GSMDefines.h"
#include "Rendering/Asura_PS2_GSMRenderer.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"
#include "Rendering/Asura_PS2_VU1.h"
#ifdef ASURA_USE_SKYSPHERE
#include "SpecialFX/Asura_PS2_FractalSkyDefines.h"
#endif

// M A C R O S ///////////////////////////////////////////////////////////////

#define SKYBOX_ENABLE_SUBDIV_TEST_CODE (0)

// T Y P E S /////////////////////////////////////////////////////////////////

struct SkyboxCamTestData
{
    float fCachedMaxHalfFOV;
    float fCachedTestVal;
};

// C O N S T A N T S /////////////////////////////////////////////////////////

// This is arctan(sqrt(2)). This represents the angle from the centre of a
// face to a corner.
static const float SKYBOX_ANGLE_TO_CORNER = 0.95531661812450927816385710251576f;

static const u_int SKYBOX_NUM_FACES = 6;

// G L O B A L S /////////////////////////////////////////////////////////////

// Subroutines on the VU.
extern u_int VU1_SKYBOX_SETUP[]        __attribute__(( section(".vudata") ));
extern u_int VU1_SKYBOX_RENDER_BATCH[] __attribute__(( section(".vudata") ));
extern u_int VU1_FS_SETUP[]            __attribute__(( section(".vudata") ));

static int g_aiSkyboxTextureIDs[SKYBOX_NUM_FACES] = { ASURA_TEXTURE_NONE, ASURA_TEXTURE_NONE, ASURA_TEXTURE_NONE, ASURA_TEXTURE_NONE, ASURA_TEXTURE_NONE, ASURA_TEXTURE_NONE };
#ifdef ASURA_USE_SKYSPHERE
static int g_iFractalSkyTextureID = ASURA_TEXTURE_NONE;
static int g_iFractalSky2TextureID = ASURA_TEXTURE_NONE;
#endif

DEBUGVAR bool dbg_bSkyboxRenderAllFaces     = false;
DEBUGVAR int  dbg_iSkyboxForceSingleFaceVis = -1;

static float g_fSkyboxCosEulerY = Asura_Maths::LargeFloat;
static float g_fSkyboxSinEulerY = Asura_Maths::LargeFloat;
static Asura_Vector_3 g_axSkyboxFaceDirs[SKYBOX_NUM_FACES];

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

#ifdef ASURA_USE_SKYSPHERE
static void FractalSky_Deallocate( void );
static bool FractalSky_Render( void );
static void FractalSky_SetupMatrices( void );
#endif

static void Skybox_Init( void );
static void Skybox_GenerateFaces( void );
static void Skybox_GenerateFace(u_int uFaceType);
static void Skybox_Render( bool bNeedToBlendWithFractalSky );
static void Skybox_DeallocateFace( void );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_Skybox::Platform_DoRenderSetup( void )
{
    bool bAnyNew = false;
    bool bNew;

    for( unsigned uSide = 0; uSide < SKYBOX_NUM_FACES; ++uSide )
    {
        Asura_Assert( ASURA_TEXTURE_NONE == g_aiSkyboxTextureIDs[uSide], "Skybox textures have already been initialised." );
        if( !GetTexture(uSide).IsBlank() )
        {
            g_aiSkyboxTextureIDs[uSide] = Asura_Texture_Management::AddTexture( GetTexture(uSide), &bNew );
            bAnyNew |= bNew;
        }
    }

#ifdef ASURA_USE_SKYSPHERE
	// AJG - 19/02/2004 - I don't really like this, these could be in the array above if we extend it,
	// but since the original fractal texture was kept separate, for clarity I guess, I'll do the same for the second
    Asura_Assert( ASURA_TEXTURE_NONE == g_iFractalSkyTextureID, "Fractal sky texture has already been initialised." );
    if( !GetTexture(ASURA_SKYBOX_FRACTALSKY).IsBlank() )
    {
        g_iFractalSkyTextureID = Asura_Texture_Management::AddTexture( GetTexture(ASURA_SKYBOX_FRACTALSKY), &bNew );
        bAnyNew |= bNew;
    }
    Asura_Assert( ASURA_TEXTURE_NONE == g_iFractalSky2TextureID, "Second fractal sky texture has already been initialised." );
    if( !GetTexture(ASURA_SKYBOX_FRACTALSKY_2).IsBlank() )
    {
        g_iFractalSky2TextureID = Asura_Texture_Management::AddTexture( GetTexture(ASURA_SKYBOX_FRACTALSKY_2), &bNew );
        bAnyNew |= bNew;
    }
#endif

#ifndef ASURA_PS2_SONY_SUBMISSION
    if( bAnyNew )
    {
        Asura_Assert( false, "All skybox textures must be included in level archive." );
        Asura_Texture_Management::LoadTextures();
    }
#endif
    
    Skybox_Init();
}

void Asura_Skybox::Platform_Initialise( void )
{
    Skybox_GenerateFaces();
}

void Asura_Skybox::Platform_Render( void )
{
    bool bRenderedFractalSky = false;
    
#ifdef ASURA_USE_SKYSPHERE
    if( Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DRAW_FRACTAL_SKY) && s_bDrawClouds )
    {
		ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__FRACTALSKY_MARKER );
        bRenderedFractalSky = FractalSky_Render();
		ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__FRACTALSKY_MARKER );
    }
#endif
    
    if( Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DRAW_SKYBOX) && GetNumSkyboxTextures() )
    {
		ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__SKYBOX_MARKER );
        Skybox_Render( bRenderedFractalSky );
		ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__SKYBOX_MARKER );
    }
}

void Asura_Skybox::Platform_Reset( void )
{
    for( u_int u = 0; u < SKYBOX_NUM_FACES; ++u ) { g_aiSkyboxTextureIDs[u] = ASURA_TEXTURE_NONE; }

#ifdef ASURA_USE_SKYSPHERE
    g_iFractalSkyTextureID = ASURA_TEXTURE_NONE;
    g_iFractalSky2TextureID = ASURA_TEXTURE_NONE;

    FractalSky_Deallocate();
#endif
}

void Asura_Skybox::Platform_Shutdown( void )
{
    Platform_Reset();
    Skybox_DeallocateFace();
}

bool Asura_Skybox::Platform_ShouldModifyFogForSkyboxObjects( void )
{
    return false;
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

//
// Fractal sky renderer.
// 

#ifdef ASURA_USE_SKYSPHERE

struct FSStrip
{
    u_int   uAddrForREF;
    u_short usQWC;
    u_short usNumVertices;
};

static u_int    g_uFractalSkyNumStrips = 0;
static u_int    g_uFractalSkyNumTris   = 0;
static FSStrip* g_pxFractalSkyStrips   = NULL;

bool Asura_PS2_LoadStrippedSkysphere( Asura_Handle hFile, u_int uNumBytes )
{
    Asura_Assert( !g_uFractalSkyNumStrips, "Skysphere already loaded. It should be embedded in level files; don't load it manually." );

    // Expected data:
    //
    //     Given s strips...
    //
    //       /------------------------------------------------\
    //       | data                          | num bytes      |
    //       |-------------------------------|----------------|
    //       | num strips                    | 4              |
    //       | num tris                      | 4              |
    //       | strip stats (QWC + num verts) | 4 * num strips |
    //       | render data                   |                |
    //       \------------------------------------------------/

#ifdef ASURA_ASSERT
    u_int uNumBytesRead = 0;
    #define READ_DATA( dest_ptr, num_bytes ) \
        Asura_FileAccess::ReadData( hFile, (dest_ptr), (num_bytes) ); \
        uNumBytesRead += (num_bytes);
#else
    #define READ_DATA( dest_ptr, num_bytes ) \
        Asura_FileAccess::ReadData( hFile, (dest_ptr), (num_bytes) );
#endif

    READ_DATA( &g_uFractalSkyNumStrips, 4 );
    READ_DATA( &g_uFractalSkyNumTris,   4 );

    g_pxFractalSkyStrips = new FSStrip[g_uFractalSkyNumStrips];

    FSStrip* pxStrip = &g_pxFractalSkyStrips[0];
    for( u_int u = g_uFractalSkyNumStrips; u--; ++pxStrip )
    {
        READ_DATA( &pxStrip->usQWC,         2 );
        READ_DATA( &pxStrip->usNumVertices, 2 );
    }

    pxStrip = &g_pxFractalSkyStrips[0];
    for( u_int u = g_uFractalSkyNumStrips; u--; ++pxStrip )
    {
        void* pvBfr = Asura_AllocateAlign( pxStrip->usQWC << 4, ASURA_ALIGN_128_BYTE );
        READ_DATA( pvBfr, pxStrip->usQWC << 4 );
        pxStrip->uAddrForREF = reinterpret_cast<u_int>( pvBfr );
    }

    Asura_Assert( uNumBytesRead == uNumBytes, "" );

    return true;
}

static void FractalSky_Deallocate( void )
{
    if( g_pxFractalSkyStrips )
    {
        for( u_int u = g_uFractalSkyNumStrips; u--; )
        {
            Asura_DeallocateAlign( reinterpret_cast<void*>(g_pxFractalSkyStrips[u].uAddrForREF) );
        }
        delete[] g_pxFractalSkyStrips;
        g_pxFractalSkyStrips = NULL;
    }

    g_iFractalSkyTextureID = ASURA_TEXTURE_NONE;
    g_iFractalSky2TextureID = ASURA_TEXTURE_NONE;
    g_uFractalSkyNumStrips = 0;
}

static bool FractalSky_Render( void )
{
	int iTextureToUse = (Asura_Skybox::GetIsUsingSecondClouds() ? g_iFractalSky2TextureID : g_iFractalSkyTextureID);
    if( !g_uFractalSkyNumStrips || ASURA_TEXTURE_NONE == iTextureToUse ) { return false; }

    Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_LARGE );
    Asura_PS2_TextureManagement::SetupTextureEnvironment( Asura_PS2_TextureManagement::UploadTexture(iTextureToUse),
                                                          ASURA_PS2_TEXREGFLAG_ALL,
                                                          iTextureToUse,
                                                          Asura_PS2_TextureManagement::BFR_AUTO,
                                                          SCE_GS_LINEAR,
                                                          SCE_GS_LINEAR_MIPMAP_NEAREST );

    Asura_PS2_VU1::LoadRenderer( Asura_PS2_VU1::RENDERER_FRACTALSKY );
    
    // Count num active cams.
    u_int uNumActiveCams = 0;
    for( int iCam = 0; iCam < Asura_Camera::GetNumberOfCameras(); ++iCam )
    {
        Asura_Camera::SetCameraIndex( iCam );
        if( Asura_Camera::IsCameraActive() ) { ++uNumActiveCams; }
    }

    const u_int uSizeOfDIRECT               = Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS + 1;
    const u_int uSizeOfUNPACKForRS          = ASURA_PS2_FS__SIZE__RENDERSTATE;
    const u_int uSizeOfUNPACKForCams        = uNumActiveCams * ASURA_PS2_FS__OFFSET__CAM_NEXT_CAM;
    const u_int uSizeOfUNPACKForSetupProg   = 3;
    const u_int uTotalQWsForFirstCNT        = uSizeOfDIRECT + 1 + uSizeOfUNPACKForRS + 1 + uSizeOfUNPACKForCams + 1 + uSizeOfUNPACKForSetupProg;
    const u_int uQWsForFinalCNT             = 1;
    const u_int uTotalQWs                   = 1 + uTotalQWsForFirstCNT + g_uFractalSkyNumStrips + 1 + uQWsForFinalCNT;
    //                                       CNT                            REF per strip        CNT

    Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
    Asura_PS2_QW* pxOut = pxBase;

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uTotalQWsForFirstCNT,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         SCE_VIF1_SET_NOP(false),
                                                         SCE_VIF1_SET_DIRECT(uSizeOfDIRECT, false) );

    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1, pxOut, true, true );
    Asura_PlatformRenderStates::SetPixelTest( false, false );
    Asura_PlatformRenderStates::ForgetMyBuffer( pxOut );

    // Append RGBAQ.
    (pxOut + Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS)->u64s[0] = SCE_GS_SET_RGBAQ( 128, 128, 128, 128, 0x3F800000 );
    (pxOut + Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS)->u64s[1] = SCE_GS_RGBAQ;

    // Fix up NLOOP in GIF tag.
    ++((reinterpret_cast<Asura_PS2_GifTag*>(pxOut + Asura_PlatformRenderStates::QWID_TAG))->uNLOOP);

    pxOut += uSizeOfDIRECT;

    pxOut->u32s[0] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[2] = SCE_VIF1_SET_STCYCL( 1, 1, false );
    pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_FS__ADDR__RENDERSTATE, ASURA_PS2_FS__SIZE__RENDERSTATE, Asura_PS2_VU1::VIFPK_V4_32, false );
    ++pxOut;

    // Pass 0: Cd = Cs.
    pxOut->u64s[0] = SCE_GS_SET_ALPHA( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CS, 0 );
    pxOut->u64s[1] = SCE_GS_ALPHA_1;
    ++pxOut;

    // Pass 1: Cd += Cs.
    pxOut->u64s[0] = SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CD, 128 );
    pxOut->u64s[1] = SCE_GS_ALPHA_1;
    ++pxOut;

    // Pass 2: Cd = FIX * Cs + (1 - FIX) * Cd
    {
        float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalScaledTime() / 8.0f );
        fAlpha = 128.0f * fAlpha * fAlpha;
        pxOut->u64s[0] = SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CD, Asura_Maths::FloatToInt(fAlpha) );
        pxOut->u64s[1] = SCE_GS_ALPHA_1;
        ++pxOut;
    }

    // Sneak BASE and OFFSET setup into this next QW...
    pxOut->u32s[0] = SCE_VIF1_SET_BASE( ASURA_PS2_FS__ADDR__INPUT_BFR_0, false );
    pxOut->u32s[1] = SCE_VIF1_SET_OFFSET( ASURA_PS2_FS__SIZE__INPUT_BFR, false );

    // UNPACK for camera data.
    pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_FS__ADDR__CAM_DATA, uSizeOfUNPACKForCams, Asura_PS2_VU1::VIFPK_V4_32, false );
    ++pxOut;

    // Write cam data.
    {
        Asura_PS2_QW* pxSPR = Asura_PS2_SPR_Alloc( 0, 4 );

        for( int iCam = 0; iCam < Asura_Camera::GetNumberOfCameras(); ++iCam )
        {
            Asura_Camera::SetCameraIndex( iCam );
            if( !Asura_Camera::IsCameraActive() ) { continue; }

            // SCISSOR and XYOFFSET.
            pxOut = Asura_PS2_GS::GetViewportSetupGifCmds( iCam, Asura_PS2_GS::CONTEXT_1, pxOut );

            // V->GBC mat.
            {
                sceVu0UnitMatrix( &pxSPR->Vu0FVec );

                const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();
                (pxSPR + 3)->f32s[0] = pxCamPos->x;
                (pxSPR + 3)->f32s[1] = pxCamPos->y;
                (pxSPR + 3)->f32s[2] = pxCamPos->z;

                sceVu0MulMatrix( &pxOut->Vu0FVec, Asura_PS2_VU1::GetWorldToClipMatGuardband(), &pxSPR->Vu0FVec );
                pxOut += 4;
            }

            // GBC->S scale, trans.
            (pxOut++)->u128 = *(reinterpret_cast<const u_long128*>(Asura_PS2_VU1::GetGBClipToScrScaleVec(iCam)));
            (pxOut++)->u128 = *(reinterpret_cast<const u_long128*>(Asura_PS2_VU1::GetGBClipToScrTransVec(iCam)));

            // GBC->SC scale.
            (pxOut++)->u128 = *(reinterpret_cast<const u_long128*>(Asura_PS2_VU1::GetGBClipToScrClipScaleVec(iCam)));
        }

        Asura_PS2_SPR_Dealloc( pxSPR );
    }

    // UNPACK for data required by setup program.
    pxOut->u32s[0] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[3] = SCE_VIF1_SET_UNPACKR( 0, 3, Asura_PS2_VU1::VIFPK_V4_32, false );
    ++pxOut;

    // GIF tag for GS setup. EOP is off because draw GIF tag will follow
    // immediately.
    (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( (ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG - ASURA_PS2_FS__OFFSET__OUTPUT_SETUP_GIF_TAG) - 1,
                                                         false,
                                                         false,
                                                         0,
                                                         SCE_GIF_PACKED,
                                                         1,
                                                         SCE_GIF_PACKED_AD );

    // GIF tag for drawing.
    {
        const u_int uPRIM = SCE_GS_SET_PRIM( SCE_GS_PRIM_TRISTRIP, // PRIM
                                             0,                    // IIP
                                             true,                 // TME
                                             false,                // FGE
                                             true,                 // ABE
                                             false,                // AA1
                                             0,                    // FST
                                             0,                    // CTXT
                                             0 );                  // FIX

        (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( (ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG - ASURA_PS2_FS__OFFSET__OUTPUT_SETUP_GIF_TAG) - 1,
                                                             true,
                                                             true,
                                                             uPRIM,                 
                                                             SCE_GIF_PACKED,
                                                             2,
                                                             SCE_GS_ST,
                                                             SCE_GS_XYZ2 );
    }

    // ST generation parameters.
    {
        float fPhase1 = Asura_Timers::GetLocalScaledTime() / 64.0f;
        float fPhase2 = Asura_Timers::GetLocalScaledTime() / 128.0f;
            
        // Normalise.
        fPhase1 -= Asura_Maths::FloatToInt_Chop( fPhase1 );
        fPhase2 -= Asura_Maths::FloatToInt_Chop( fPhase2 );

        pxOut->f32s[0] = fPhase2;
        pxOut->f32s[1] = -fPhase1;
        pxOut->f32s[2] = 0.9f;
        ++pxOut;
    }

    // REF tags for strips. In each tag, we setup ITOP and run the program
    // for the previous strip uploaded. In the case of the first tag, we
    // run the setup program.
    const FSStrip* pxStrip = &g_pxFractalSkyStrips[0];
    for( u_int uStrip = 0; uStrip < g_uFractalSkyNumStrips; ++pxStrip, ++uStrip )
    {
        u_int uWord2, uWord3;
        if( uStrip )
        {
            uWord2 = SCE_VIF1_SET_ITOP( (pxStrip - 1)->usNumVertices, false );
            uWord3 = SCE_VIF1_SET_MSCNT( false );
        }
        else
        {
            uWord2 = SCE_VIF1_SET_ITOP( uNumActiveCams, false );
            uWord3 = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_FS_SETUP );
        }

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( pxStrip->usQWC,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_REF,
                                                             false,
                                                             pxStrip->uAddrForREF,
                                                             false,
                                                             uWord2,
                                                             uWord3 );
    }

    // Final CNT tag to render last strip and then FLUSH to ensure we're done
    // before next renderer, whatever that may be, gets started.
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWsForFinalCNT,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         SCE_VIF1_SET_NOP(false),
                                                         SCE_VIF1_SET_NOP(false) );

    pxOut->u32s[0] = SCE_VIF1_SET_ITOP( (pxStrip - 1)->usNumVertices, false );
    pxOut->u32s[1] = SCE_VIF1_SET_MSCNT( false );
    pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[3] = SCE_VIF1_SET_FLUSH( false );
    ++pxOut;

    Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );

    Asura_PS2_Render::UpdateTriStats( uNumActiveCams * g_uFractalSkyNumTris, 3 );

    return true;
}

#endif // #ifdef ASURA_USE_SKYSPHERE

//
// Skybox renderer.
//

#if SKYBOX_ENABLE_SUBDIV_TEST_CODE
    static u_int SKYBOX_FACE_NUM_SUBDIVS = 1;
#else
    #ifndef SKYBOX_FACE_NUM_SUBDIVS
        #define SKYBOX_FACE_NUM_SUBDIVS (1)
    #endif
#endif

#define SKYBOX_FACE_ROW_NUM_VERTS (((1 << SKYBOX_FACE_NUM_SUBDIVS) + 1) << 1)
#define SKYBOX_FACE_NUM_ROWS      (1 << SKYBOX_FACE_NUM_SUBDIVS)
#define SKYBOX_FACE_NUM_TRIS      (1 << ((SKYBOX_FACE_NUM_SUBDIVS << 1) + 1))

#define SKYBOX_BATCH_NUM_ST_UPK_WORDS(verts)  ( verts )
#define SKYBOX_BATCH_NUM_ST_WORDS(verts)      ( 1 + SKYBOX_BATCH_NUM_ST_UPK_WORDS(verts) )
#define SKYBOX_BATCH_NUM_ADC_UPK_WORDS(verts) ( (verts + 3) >> 2 )
#define SKYBOX_BATCH_NUM_ADC_WORDS(verts)     ( 1 + SKYBOX_BATCH_NUM_ADC_UPK_WORDS(verts) )
#define SKYBOX_BATCH_NUM_XYZ_UPK_WORDS(verts) ( verts * 3 )
#define SKYBOX_BATCH_NUM_XYZ_WORDS(verts)     ( 1 + SKYBOX_BATCH_NUM_XYZ_UPK_WORDS(verts) )
#define SKYBOX_BATCH_TOTAL_WORDS(verts)       (   1                                 \
                                                + SKYBOX_BATCH_NUM_ST_WORDS(verts)  \
                                                + SKYBOX_BATCH_NUM_ADC_WORDS(verts) \
                                                + SKYBOX_BATCH_NUM_XYZ_WORDS(verts) \
                                                + 2 )

#define SKYBOX_BATCH_MAX_ROWS   (ASURA_PS2_GSM__MAX_VERTS_PER_STRIP / SKYBOX_FACE_ROW_NUM_VERTS)
#define SKYBOX_BATCH_REQD_COUNT ((SKYBOX_FACE_NUM_ROWS + (SKYBOX_BATCH_MAX_ROWS - 1)) / SKYBOX_BATCH_MAX_ROWS)

enum SKYBOX_FACE_TYPE
{
	SKYBOX_FACE_DEFAULT = 0,
	SKYBOX_FACE_MIRRORED,
	SKYBOX_FACE_MIRRORED_FLIPPED,
	SKYBOX_NUM_FACE_TYPES,
};

static Asura_PS2_QW* g_apxSkyboxFaceData[SKYBOX_NUM_FACE_TYPES];
static u_int g_aauSkyboxBatchAddrs[SKYBOX_NUM_FACE_TYPES][SKYBOX_BATCH_REQD_COUNT];

static void Skybox_DeallocateFace( void )
{
	for(u_int u = 0; u < SKYBOX_NUM_FACE_TYPES; ++u)
	{
		Asura_DeallocateAlign( g_apxSkyboxFaceData[u] );
	}
}

static void Skybox_GenerateFaces()
{
	for(u_int u = 0; u < SKYBOX_NUM_FACE_TYPES; ++u)
	{
		Skybox_GenerateFace(u);
	}

}

// Generates one RET DMA chain per batch. This data is constant and only
// needs to be generated once at startup.
static void Skybox_GenerateFace(u_int uFaceType)
{
    // To simplify matters, we'll output a fixed number of rows per batch.
    // Well, except for maybe the last one.

    // Average out num rows per batch as best we can.
    const u_int uNumRowsPerBatch    = SKYBOX_FACE_NUM_ROWS / SKYBOX_BATCH_REQD_COUNT;
    const u_int uNumRowsInLastBatch = SKYBOX_FACE_NUM_ROWS - (uNumRowsPerBatch * (SKYBOX_BATCH_REQD_COUNT - 1));

    // Calculate the resultant number of vertices per batch.
    const u_int uNumVertsPerBatch    = uNumRowsPerBatch    * SKYBOX_FACE_ROW_NUM_VERTS;
    const u_int uNumVertsInLastBatch = uNumRowsInLastBatch * SKYBOX_FACE_ROW_NUM_VERTS;

    // Calculate the additional amount of data required for each one. Each
    // batch will be able to put two of its data words in the DMA tag.
    const u_int uNumDMAQWsPerBatch    = ((SKYBOX_BATCH_TOTAL_WORDS(uNumVertsPerBatch)    - 2) + 3) >> 2;
    const u_int uNumDMAQWsInLastBatch = ((SKYBOX_BATCH_TOTAL_WORDS(uNumVertsInLastBatch) - 2) + 3) >> 2;

    // So the total size including tags is...
    const u_int uTotalNumQWs = (SKYBOX_BATCH_REQD_COUNT - 1) * (1 + uNumDMAQWsPerBatch) // All batches except last.
                             + (1 + uNumDMAQWsInLastBatch);                             // And the last batch.

    // However, the allocation needs to be for a multiple of 64 bytes to
    // avoid cache issues.
    const u_int uNumQWsToAlloc = ((uTotalNumQWs + 3) >> 2) << 2;
    
    // Allocate it on a 128 byte boundary too since we're DMAing it.
    g_apxSkyboxFaceData[uFaceType] = reinterpret_cast<Asura_PS2_QW*>( Asura_AllocateAlign(uNumQWsToAlloc << 4, ASURA_ALIGN_128_BYTE) );

    // We used to make the skybox size quite large but there's no particular
    // reason to do so, I don't think, since we're not using it to clear the
    // z buffer and we're not writing to it either. This also prevents issues
    // with stuff getting culled against the far clip plane, as happened
    // initially when moving this code over to use the GSM renderer instead
    // of a custom one.
    const float SKYBOX_SIZE      = 1.0f;
    const float NUM_QUADS        = (1 << SKYBOX_FACE_NUM_SUBDIVS);
    const float HALF_SKYBOX_SIZE = 0.5f * SKYBOX_SIZE;
    const float O_O_NUM_QUADS    = 1.0f / NUM_QUADS;

    float fY = 0.0f;

    // Ok, output each batch. Per vertex we just need to write XYZ, ADC and
    // ST values. RGB and normals are irrelevant since we're faking up some
    // ambient lighting and alpha is irrelevant since the blend mode will
    // ignore it anyway.
    u_int* puOut = &g_apxSkyboxFaceData[uFaceType]->u32s[0];
    for( u_int uBatch = 0; uBatch < SKYBOX_BATCH_REQD_COUNT; ++uBatch )
    {
        u_int* const puBatchStart = puOut;

        const bool  bIsLast    = ( uBatch + 1 == SKYBOX_BATCH_REQD_COUNT );
        const u_int uNumVerts  = ( bIsLast ? uNumVertsInLastBatch : uNumVertsPerBatch );
        const u_int uNumDMAQWs = ( bIsLast ? uNumDMAQWsInLastBatch : uNumDMAQWsPerBatch );

        // Record batch addr.
        g_aauSkyboxBatchAddrs[uFaceType][uBatch] = reinterpret_cast<u_int>( Asura_PS2_DMAMem(puOut) );

        // DMA tag.
        {
            Asura_PS2_DmaTag* pxTag = reinterpret_cast<Asura_PS2_DmaTag*>( puOut );
            pxTag->Set( uNumDMAQWs,
                        Asura_PS2_DmaTag::PCE_NOCHANGE,
                        Asura_PS2_DmaTag::ID_RET,
                        false,
                        0,
                        false );

            puOut += 2;
        }

        *puOut++ = SCE_VIF1_SET_STCYCL( 1, 4, false );

        // STs.
        *puOut++ = SCE_VIF1_SET_UNPACKR( 1, uNumVerts, Asura_PS2_VU1::VIFPK_V2_16, false );
        short* psSTOut = reinterpret_cast<short*>( puOut );
        puOut += SKYBOX_BATCH_NUM_ST_UPK_WORDS( uNumVerts );

        // ADCs. Note that we'll be trashing the normal's XYZ here but, as
        // noted above, we won't be using it anyway.
        *puOut++ = SCE_VIF1_SET_UNPACKR( 3, uNumVerts, Asura_PS2_VU1::VIFPK_S_8, false );
        u_char* pucADCOut = reinterpret_cast<u_char*>( puOut );
        puOut += SKYBOX_BATCH_NUM_ADC_UPK_WORDS( uNumVerts );

        // XYZs.
        *puOut++ = SCE_VIF1_SET_UNPACKR( 0, uNumVerts, Asura_PS2_VU1::VIFPK_V3_32, false );
        float* pfXYZOut = reinterpret_cast<float*>( puOut );
        puOut += SKYBOX_BATCH_NUM_XYZ_UPK_WORDS( uNumVerts );

        // Fill in the gaps in the above.
        for( u_int uRowsLeft = (bIsLast ? uNumRowsInLastBatch : uNumRowsPerBatch); uRowsLeft--; fY += 1.0f )
        {
            const float fTopY = (fY + 0.0f) * SKYBOX_SIZE * O_O_NUM_QUADS - HALF_SKYBOX_SIZE;
            const float fBtmY = (fY + 1.0f) * SKYBOX_SIZE * O_O_NUM_QUADS - HALF_SKYBOX_SIZE;
			float fAdjustedY = fY;
			float fDirection = 1.f;
			if(uFaceType == SKYBOX_FACE_MIRRORED)
			{
				const bool bMirrored = (fY >= SKYBOX_FACE_NUM_ROWS / 2);
				fAdjustedY = bMirrored ? (SKYBOX_FACE_NUM_ROWS - fY) : fY;
				fDirection = (bMirrored ? -1.f : 1.f);
			}
			else if(uFaceType == SKYBOX_FACE_MIRRORED_FLIPPED)
			{
				const bool bMirrored = (fY < SKYBOX_FACE_NUM_ROWS / 2);
				fAdjustedY = bMirrored ? (SKYBOX_FACE_NUM_ROWS - fY) : fY;
				fDirection = (bMirrored ? -1.f : 1.f);
			}
            const float fTopT = (fAdjustedY + 0.0f) * 1.0f * O_O_NUM_QUADS;
			const float fBtmT = (fAdjustedY + fDirection) * 1.0f * O_O_NUM_QUADS;

            // Convert STs to 1:3:12 fixed point.
            const short sTopT = static_cast<short>( Asura_Maths::FloatToInt_Near(fTopT * 4096.0f) );
            const short sBtmT = static_cast<short>( Asura_Maths::FloatToInt_Near(fBtmT * 4096.0f) );

            // Special case for first two vertices of each row. They need to
            // have ADC bits set and also are the only time we need to output
            // the lefthand vertices, since we're starting off the strip. The
            // values required are 8 for ADC on and 0 for ADC off, since
            // the GSM renderer's decompressing of them is designed to
            // support values coming from the last component of a V4_5 UNPACK.
            for( u_int u = 0; u < 2; ++u )
            {
                *psSTOut++   = 0;
                *psSTOut++   = ( u ? sBtmT : sTopT );
                *pucADCOut++ = 8;
                *pfXYZOut++  = -HALF_SKYBOX_SIZE;
                *pfXYZOut++  = ( u ? fBtmY : fTopY );
                *pfXYZOut++  = HALF_SKYBOX_SIZE;
            }

            // And the rest.
            for( float fX = 0.0f; fX < NUM_QUADS; fX += 1.0f )
            {
                const float fRtX = (fX + 1.0f) * SKYBOX_SIZE * O_O_NUM_QUADS - HALF_SKYBOX_SIZE;
                const float fRtS = (fX + 1.0f) * 1.0f * O_O_NUM_QUADS;

                // Convert ST to 1:3:12 fixed point.
                const short sRtS = static_cast<short>( Asura_Maths::FloatToInt_Near(fRtS * 4096.0f) );

                for( u_int u = 0; u < 2; ++u )
                {
                    *psSTOut++   = sRtS;
                    *psSTOut++   = ( u ? sBtmT : sTopT );
                    *pucADCOut++ = 0;
                    *pfXYZOut++  = fRtX;
                    *pfXYZOut++  = ( u ? fBtmY : fTopY );
                    *pfXYZOut++  = HALF_SKYBOX_SIZE;
                }
            }
        }

        Asura_Assert( reinterpret_cast<u_int*>(pfXYZOut) == puOut, "" );
        
        // Set vertex count and start microprogram.
        *puOut++ = SCE_VIF1_SET_ITOP( uNumVerts, false );
        *puOut++ = SCE_VIF1_SET_MSCNT( false );

        // Pad to QW alignment ready for next DMA tag.
        u_int uRemainder = (puOut - puBatchStart) & 3;
        if( uRemainder )
        {
            for( u_int uPad = 4 - uRemainder; uPad; --uPad )
            {
                *puOut++ = SCE_VIF1_SET_NOP( false );
            }
        }

        Asura_Assert( static_cast<u_int>((puOut - puBatchStart) >> 2) == 1 + uNumDMAQWs, "" );
    }

    Asura_Assert( (static_cast<u_int>(puOut - &g_apxSkyboxFaceData[uFaceType]->u32s[0]) >> 2) == uTotalNumQWs, "" );
}

static void Skybox_Render( bool bNeedToBlendWithFractalSky )
{
#if SKYBOX_ENABLE_SUBDIV_TEST_CODE
    static u_int uCachedSubdivLevel = SKYBOX_FACE_NUM_SUBDIVS;
    if( SKYBOX_FACE_NUM_SUBDIVS != uCachedSubdivLevel )
    {
        Asura_PS2_DeferredRendering::WaitForDeferredRendering();
        Skybox_DeallocateFace();
        Skybox_GenerateFaces();
        uCachedSubdivLevel = SKYBOX_FACE_NUM_SUBDIVS;
    }
#endif

    const u_int uNumCams = Asura_Camera::GetNumberOfCameras();

    // Setup some values for the visibility tests.
    float afCamTestValues[ASURA_MAX_NO_OF_CAMERAS];
    for( u_int uCam = 0; uCam < uNumCams; ++uCam )
    {
        Asura_Camera::SetCameraIndex( uCam );
        const float fMaxHalfFOV = Asura_Maths::Max( 0.5f * Asura_Camera::GetHorizontalFieldOfView(), 0.5f * Asura_Camera::GetVerticalFieldOfView() );
        afCamTestValues[uCam] = Asura_Maths::Cosine( fMaxHalfFOV + SKYBOX_ANGLE_TO_CORNER );
    }

    // Some stuff for setting up the object->world transforms later.
    const float PI = Asura_Maths::Pi;
    const float HALF_PI = 0.5f * Asura_Maths::Pi;
    const float SKYBOX_FACE_ROT_EULERS[SKYBOX_NUM_FACES][2] =
    {
        { -HALF_PI,       PI }, // +y (recall our Euler code generates an yxz matrix, ie. z followed by x followed by y)
        {     0.0f,       PI }, // -z
        {     0.0f,     0.0f }, // +z
        {     0.0f,  HALF_PI }, // -x
        {     0.0f, -HALF_PI }, // +x
        {  HALF_PI,       PI }, // -y (see note for +y)
    };
    Asura_Matrix_3x3 xSkyToWorld(  g_fSkyboxCosEulerY, 0.0f, g_fSkyboxSinEulerY,
                                                 0.0f, 1.0f,               0.0f,
                                  -g_fSkyboxSinEulerY, 0.0f, g_fSkyboxCosEulerY  );
    Asura_Matrix_3x3 xFaceToWorld;

    Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_LARGE );

    Asura_PS2_GSMRenderer::Begin();

    // Enable texture clamping to fix seam issues.
    Asura_PS2_GSMRenderer::SetTextureClamp( true );

    Asura_PS2_GSMRenderer::SetRenderstate( 0 );
    Asura_PS2_GSMRenderer::ManuallyModifyRenderstate( Asura_PS2_GSMRenderer::CB_DisableZReadsAndWrites );

    // Fake lighting to avoid having to mess around with normals and vertex
    // colours. See also Skybox_GenerateFace().
    {
        Asura_SphericalHarmonics_Order2_RGB xFakeSH;

		xFakeSH.Zero();
		xFakeSH.m_afRedCoefficients[0]   = 0.5f * (Asura_Skybox::GetR() / 255.0f);
		xFakeSH.m_afGreenCoefficients[0] = 0.5f * (Asura_Skybox::GetG() / 255.0f);
		xFakeSH.m_afBlueCoefficients[0]  = 0.5f * (Asura_Skybox::GetB() / 255.0f);

        Asura_PS2_GSMRenderer::UseExistingSHForMesh( xFakeSH, 1.0f );
    }

    // Run through the faces, rendering for particular cameras as appropriate.
    u_int uTrisRendered = 0;
    for( u_int uFace = 0; uFace < SKYBOX_NUM_FACES; ++uFace )
    {
        const int iTexID = g_aiSkyboxTextureIDs[uFace];
        if( ASURA_TEXTURE_NONE == iTexID ) { continue; }

        u_int uVisFlags = 0;
        for( u_int uCam = 0; uCam < uNumCams; ++uCam )
        {
            Asura_Camera::SetCameraIndex( uCam );
            if( !Asura_Camera::IsCameraActive() ) { continue; }

            const Asura_Matrix_3x3* pxCamOri = Asura_Camera::GetOrientation();

            float fDot = pxCamOri->e13 * g_axSkyboxFaceDirs[uFace].x
                       + pxCamOri->e23 * g_axSkyboxFaceDirs[uFace].y
                       + pxCamOri->e33 * g_axSkyboxFaceDirs[uFace].z;

            if( !dbg_bSkyboxRenderAllFaces && -1 != dbg_iSkyboxForceSingleFaceVis )
            {
                fDot = ( static_cast<int>(uFace) == dbg_iSkyboxForceSingleFaceVis ? 1.0f : -1.0f );
            }

            if( dbg_bSkyboxRenderAllFaces || fDot >= afCamTestValues[uCam] )
            {
                uVisFlags |= 1 << uCam;
            }
        }

        if( !uVisFlags ) { continue; }

		SKYBOX_FACE_TYPE eFaceType = SKYBOX_FACE_DEFAULT;
		if(Asura_Skybox::GetMirrorBelowHorizon())
		{
			if(uFace != Asura_Skybox::ASURA_SKYBOX_DN && uFace != Asura_Skybox::ASURA_SKYBOX_UP)
			{
				if(uFace == Asura_Skybox::ASURA_SKYBOX_RT && Asura_Skybox::GetRightTextureIsLeftUpsideDown())
				{
					eFaceType = SKYBOX_FACE_MIRRORED_FLIPPED;
				}
				else if(uFace == Asura_Skybox::ASURA_SKYBOX_BK && Asura_Skybox::GetBackTextureIsFrontUpsideDown())
				{
					eFaceType = SKYBOX_FACE_MIRRORED_FLIPPED;
				}
				else
				{
					eFaceType = SKYBOX_FACE_MIRRORED;
				}
			}
		}

        // Set this face's texture.
        Asura_PS2_GSMRenderer::SetTexture( iTexID, SCE_GS_NEAREST, SCE_GS_NEAREST, Asura_PS2_TextureManagement::BFR_AUTO );

        // Prepare face->world rotation matrix. This is constant across all
        // cameras - only the translation changes.
        {
            Asura_Euler_3 xFaceToSkyEuler( SKYBOX_FACE_ROT_EULERS[uFace][0], SKYBOX_FACE_ROT_EULERS[uFace][1], 0.0f );
        
            Asura_Matrix_3x3 xFaceToSky;
            xFaceToSkyEuler.ToMatrix( xFaceToSky );

            // Modify middle element of the sky->world matrix depending on
            // whether or not we're doing any texture sharing stuff.
			// KOR - or if we're doing mirroring
            if( (Asura_Skybox::ASURA_SKYBOX_BK == uFace && Asura_Skybox::GetBackTextureIsFrontUpsideDown())
                ||
                (Asura_Skybox::ASURA_SKYBOX_RT == uFace && Asura_Skybox::GetRightTextureIsLeftUpsideDown()) )
            {
                xSkyToWorld.e22 = -1.0f;
            }
            else
            {
                xSkyToWorld.e22 = 1.0f;
            }

			if((Asura_Skybox::ASURA_SKYBOX_DN == uFace && Asura_Skybox::GetMirrorBelowHorizon()))
			{
				xFaceToSky.e32 = -1.f;
			}

            xFaceToWorld = xSkyToWorld * xFaceToSky;
        }

        // Run through the cameras again, rendering for each one that can
        // see this face. Obviously this could be a little more efficient -
        // the VU1 microprogram could internally support object instancing
        // or something - but it's the way we need to do things with the
        // existing GSM renderer. In terms of overall performance, it
        // shouldn't really make any difference, since the bottlenecks are
        // the texture uploads and rasterising not vertex work, plus we're
        // only uploading a very small amount of data per face given the
        // (currently) low level of subdivision employed.
        for( u_int uCam = 0; uCam < uNumCams; ++uCam )
        {
            const u_int uCamVisBit = 1 << uCam;
            if( !(uVisFlags & uCamVisBit) ) { continue; }

            Asura_Camera::SetCameraIndex( uCam );

            // Update object->world matrix.
            Asura_PS2_GSMRenderer::SetObjectToWorldMat( xFaceToWorld, *Asura_Camera::GetPosition() );

            // Technically we should also refresh the lighting data now since
            // it'll be in the wrong space. However, we're only using ambient
            // so it doesn't actually matter.

            // Update camera flags. For details of what's going on here, see
            // comment in Asura_PS2_GSMRenderer.h about SetCamFlags().
            Asura_PS2_GSMRenderer::SetCamFlags( uCamVisBit, 1 << (6 - (uCam << 1)) );
        
            // Send geometry chain(s) and kick off microprogram after each
            // batch.
            {
                Asura_PS2_GSMRenderer::RenderStripsWithManualSend();

                Asura_PS2_DmaTag* pxTags = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(SKYBOX_BATCH_REQD_COUNT) );
                for( u_int u = 0; u < SKYBOX_BATCH_REQD_COUNT; ++u )
                {
                    (pxTags++)->SetWithNOPs( 0,
                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                             Asura_PS2_DmaTag::ID_CALL,
                                             false,
                                             g_aauSkyboxBatchAddrs[eFaceType][u],
                                             false );
                }
            }

            uTrisRendered += SKYBOX_FACE_NUM_TRIS;
        }
    }

    Asura_PS2_Render::UpdateTriStats( uTrisRendered, 1 );

    Asura_PS2_GSMRenderer::SetTextureClamp( false );

    Asura_PS2_GSMRenderer::End();
}

static void Skybox_Init( void )
{
    // Precalculate cosine and sine of Y Euler and dirs to faces.
    {
        float fEulerY;
        Asura_Skybox::GetOrientationAroundYAxis( fEulerY );

        Asura_Maths::SinCos( fEulerY, g_fSkyboxSinEulerY, g_fSkyboxCosEulerY );

        g_axSkyboxFaceDirs[0].x = 0.0f;                 // +y (bottom)
        g_axSkyboxFaceDirs[0].y = 1.0f;
        g_axSkyboxFaceDirs[0].z = 0.0f;

        g_axSkyboxFaceDirs[1].x = -g_fSkyboxSinEulerY;  // -z (back)
        g_axSkyboxFaceDirs[1].y = 0.0f;
        g_axSkyboxFaceDirs[1].z = -g_fSkyboxCosEulerY;

        g_axSkyboxFaceDirs[2].x = g_fSkyboxSinEulerY;   // +z (front)
        g_axSkyboxFaceDirs[2].y = 0.0f;
        g_axSkyboxFaceDirs[2].z = g_fSkyboxCosEulerY;

        g_axSkyboxFaceDirs[3].x = g_fSkyboxCosEulerY;   // +x (right)
        g_axSkyboxFaceDirs[3].y = 0.0f;
        g_axSkyboxFaceDirs[3].z = -g_fSkyboxSinEulerY;

        g_axSkyboxFaceDirs[4].x = -g_fSkyboxCosEulerY;  // -x (left)
        g_axSkyboxFaceDirs[4].y = 0.0f;
        g_axSkyboxFaceDirs[4].z = g_fSkyboxSinEulerY;

        g_axSkyboxFaceDirs[5].x = 0.0f;                 // -y (top)
        g_axSkyboxFaceDirs[5].y = -1.0f;
        g_axSkyboxFaceDirs[5].z = 0.0f;
    }
}
