//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Multithread.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     PS3-specific code for multithreading code
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_Multithread.h"

// Asura independent
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"

// library includes
#include <sys/ppu_thread.h>
#include <sys/synchronization.h>
#include <sys/timer.h>

u_int Asura_Thread_System::s_auStackSizes[]  =	{ 16 * 1024		//ASURA_MULTITHREAD_STACK_SIZE_SMALL
												, 32 * 1024		//ASURA_MULTITHREAD_STACK_SIZE_MEDIUM
												, 64 * 1024 };	//ASURA_MULTITHREAD_STACK_SIZE_LARGE

union Asura_PS3_ThreadParams
{
	struct  
	{
		Asura_ThreadFunc m_pfnThreadFunc;
		const void* m_pUserData;
	};
	uint64_t m_xDataAs64;
};

static void Asura_PS3_ThreadStartup( uint64_t uParam )
{
	Asura_PS3_ThreadParams xParams;
	xParams.m_xDataAs64 = uParam;
	xParams.m_pfnThreadFunc(xParams.m_pUserData);
	sys_ppu_thread_exit(0);
}

u_int Asura_Thread_System::Platform_GetNumHardwareThreads()
{
    return 2;
}

bool Asura_Thread_System::Platform_CreateThread( const char* szThreadName, Asura_ThreadFunc pfnThreadFunc, const void* pUserData, const ASURA_MULTITHREAD_STACK_SIZE eStackSize )
{
	Asura_Assert(eStackSize != ASURA_MULTITHREAD_STACK_NUM_SIZES, "Invalid stack size enumeration passed to Platform_CreateThread");

	int iRet;
	Asura_PS3_ThreadParams xParams;
	xParams.m_pfnThreadFunc = pfnThreadFunc;
	xParams.m_pUserData = pUserData;
	ASURA_STATIC_ASSERT( sizeof(xParams) == sizeof(xParams.m_xDataAs64));

	sys_ppu_thread_t xThreadID;
	
	// get current thread priority and use that
	sys_ppu_thread_t xCurThreadID;
	iRet = sys_ppu_thread_get_id( &xCurThreadID );
	Asura_Assertf(iRet == CELL_OK, ("sys_ppu_thread_get_id failed (%d)", iRet));
	int iPriority;
	sys_ppu_thread_get_priority(xCurThreadID, &iPriority);
	Asura_Assertf(iRet == CELL_OK, ("sys_ppu_thread_get_priority failed (%d)", iRet));

	iRet = sys_ppu_thread_create(&xThreadID
								, Asura_PS3_ThreadStartup
								, xParams.m_xDataAs64
								, iPriority
								, s_auStackSizes[eStackSize]
								, 0
								, szThreadName);

	Asura_Assertf(iRet == CELL_OK, ("sys_ppu_thread_create failed (%d)", iRet));

	return iRet == CELL_OK;
}

void Asura_Thread_System::Platform_DelayCurrentThread(float fTimeInSeconds, bool bAllowCallbacks)
{
	sys_timer_usleep( Asura_Maths::FloatToInt_Chop(fTimeInSeconds * 1000000) /* In microseconds */ );
}

void Asura_Thread_System::Platform_YieldCurrentThread()
{
	sys_ppu_thread_yield();
}

Asura_ThreadID Asura_Thread_System::Platform_GetCurrentThreadID()
{ 
	sys_ppu_thread_t xCurThreadID;
#ifdef ASURA_ASSERT
	int iRet = 
#endif // #ifdef ASURA_ASSERT
		sys_ppu_thread_get_id( &xCurThreadID );
	Asura_Assertf(iRet == CELL_OK, ("sys_ppu_thread_get_id failed (%d)", iRet));
	return xCurThreadID; 
}

Asura_ThreadID Asura_Thread_System::Platform_GetInvalidThreadID()
{ 
	return (Asura_ThreadID)(-1);
}

void Asura_Mutex::Platform_Lock()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sys_lwmutex_lock( &m_xMutex, 0 );
	Asura_Assert( iRet == CELL_OK, "sys_lwmutex_lock failed");
}

void Asura_Mutex::Platform_Unlock()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sys_lwmutex_unlock( &m_xMutex );
	Asura_Assert( iRet == CELL_OK, "sys_lwmutex_unlock failed");
}

void Asura_SyncEvent::Platform_WaitForSignal()
{
	switch( sys_lwcond_wait( &m_xCond, 0 ) )//( 2^48 - 1)
	{
	case CELL_OK:
		break;
	case ETIMEDOUT:
		Asura_Assert( false, "Expiration of the specified period" );
		break;
	case ESRCH:
		Asura_Assert( false, "Invalid ID (The ID has not been issued by the kernel)" );
		break;
	case EPERM:
		Asura_Assert( false, "The caller thread is not the owner of the mutex" );
		break;
	case EABORT:
		Asura_Assert( false, "The condition variable was destroyed because the process terminated (Only the interrupt thread gets this return value)" );
		break;
	}
}

void Asura_SyncEvent::Platform_SetSignal()
{
	sys_lwcond_signal( &m_xCond );
}

void Asura_SyncEvent::Platform_ClearSignal()
{
}

//////////////////////////////////////////////////////////////////////////
// Semaphore
//////////////////////////////////////////////////////////////////////////
void Asura_Semaphore::Platform_Wait()
{
	int iResult = sys_semaphore_wait(m_xSemaphore, 0); // 0 => no timeout
	if (iResult < CELL_OK)
	{
		Asura_Assertf(false, ("sys_semaphore_wait failed (%d)\n", iResult));
	}
}

bool Asura_Semaphore::Platform_TryWait()
{
	int iResult = sys_semaphore_trywait(m_xSemaphore);
	if (iResult == CELL_OK)
	{
		return true;
	}
	else
	{
		Asura_Assertf(iResult == EBUSY, ("sys_semaphore_trywait failed (%d)\n", iResult));
	}
	return false;
}

bool Asura_Semaphore::Platform_Signal()
{
	int iResult = sys_semaphore_post(m_xSemaphore, 1);
	Asura_Assertf(iResult == CELL_OK, ("sys_semaphore_post failed (%d)\n", iResult));
	return (iResult == CELL_OK);
}

// EOF //////////////////////////////////////////////////////////////////////////
