
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_Timers.cpp
// Author : Andy Tate
// Date   : 21-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include <stdio.h>
#include <sys/sys_time.h>

// This file's header...
#include "Core/Asura_Timers.h"

// Asura header-files...
#include "Core/Asura_Multithread.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

//#define USE_TIMEBASE


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#if defined(USE_TIMEBASE)
// Timebase uses 64-bit unsigned integers...

typedef uint64_t TickCount;


// Timebase needs a stored frequency...

static float	g_fTimeBaseFrequency;
#else
// System-time has it's own type...

typedef system_time_t TickCount;


// System-time reports in micro-seconds...

static const float	g_fTimeBaseFrequency	= 1000000.0f;
#endif 	// defined(USE_TIMEBASE)


// The tick-count at the start of the current frame...

static TickCount	g_xTickCountAtStartOfFrame;


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : GetCurrentTickCount
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Gets the tick-count.
//
// Return value : The current system-time.
////////////////////////////////////////////////////////////////////////////////

static TickCount GetCurrentTickCount (void)
{
	TickCount	xTickCount;


#if defined(USE_TIMEBASE)
	// Get the timebase count...

	xTickCount	= sys_time_get_lpar_timebase_count ();
#else
	// Get the system-time and return it...

	xTickCount = sys_time_get_system_time();
#endif 	// defined(USE_TIMEBASE)


	// Return the tick-count...
	
	return	(xTickCount);
}

/// End of function 'GetCurrentTickCount'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_Initialise
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Initialise the PS3 timers.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Timers::Platform_Initialise (void)
{
#if defined(USE_TIMEBASE)
	TickCount	xFrequency;


	// Get the frequency...

	xFrequency	= sys_time_get_timebase_frequency	();


	// Store the floating-point version...

	g_fTimeBaseFrequency	= static_cast<float>(xFrequency);
#endif 	// defined(USE_TIMEBASE)
}

/// End of function 'Asura_Timers::Platform_Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_Reset
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Resets the PS3 timers.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Timers::Platform_Reset (void)
{
	// Store the current tick-count...

	g_xTickCountAtStartOfFrame	= GetCurrentTickCount ();
}

/// End of function 'Asura_Timers::Platform_Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_Shutdown
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Handles the shutdown of the timers.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Timers::Platform_Shutdown (void)
{
	// Nothing to do...
}

/// End of function 'Asura_Timers::Platform_Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_Update
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Handles the PS3 specific timer update.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Timers::Platform_Update (void)
{
	const float	fEffectiveMinFrameTime = GetEffectiveMinFrameTime();

	if(fEffectiveMinFrameTime > 0.0f)
	{
		const float	fTimeInFrame	= GetTimeInFrame();
		const float	fTimeToDelay	= (fEffectiveMinFrameTime - fTimeInFrame) - 0.001f;

		if(fTimeToDelay > 0.0f)
		{
			Asura_Thread_System::DelayCurrentThread(fTimeToDelay);
		}
	}
}

/// End of function 'Asura_Timers::Platform_Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_GetTicksThisFrame
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Gets the number of ticks this frame.
//
// Return value : The number of ticks this frame.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Timers::Platform_GetTicksThisFrame (void)
{
	TickCount	xCurrentTickCount	= GetCurrentTickCount ();
	TickCount	xTicksThisFrame		= xCurrentTickCount - g_xTickCountAtStartOfFrame;


	// Store the last tick-count...

	g_xTickCountAtStartOfFrame	= xCurrentTickCount;


	// Return the number of ticks this frame...

	return	(static_cast<u_int>(xTicksThisFrame));
}

/// End of function 'Asura_Timers::Platform_GetTicksThisFrame'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_GetTicksInFrame
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Returns the number of ticks since the start of the current frame.
//
// Return value : The number of ticks into this frame.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Timers::Platform_GetTicksInFrame (void)
{
	TickCount	xCurrentTickCount	= GetCurrentTickCount ();


	// Return the difference in tick-counts...

	return	(static_cast<u_int>(xCurrentTickCount - g_xTickCountAtStartOfFrame));
}

/// End of function 'Asura_Timers::Platform_GetTicksInFrame'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_GetTickFrequency
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Gets the hardware tick frequency.
//
// Return value : The number of ticks-per-second.
////////////////////////////////////////////////////////////////////////////////

float Asura_Timers::Platform_GetTickFrequency (void)
{
	// We've already stored off the frequency, return it...

	return	(g_fTimeBaseFrequency);
}

/// End of function 'Asura_Timers::Platform_GetTickFrequency'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Timers::Platform_GetDefaultFrameRate
// Author       : Andy Tate
// Date         : 21-Jun-2005
//
// Description  : Gets the default frame-rate for the PS3.
//
// Return value : The number of frames-per-second.
////////////////////////////////////////////////////////////////////////////////

float Asura_Timers::Platform_GetDefaultFrameRate (void)
{
	// Assume NTSC for now...

	return	(60.0f);
}


Asura_TickCount Asura_Timers::Platform_GetCurTicks()
{
	return GetCurrentTickCount();
}


Asura_TickCount Asura_Timers::Platform_GetElapsedTicks(Asura_TickCount xStartTicks)
{
	TickCount xCurrentTickCount = GetCurrentTickCount();
	return xCurrentTickCount - xStartTicks;
}


float Asura_Timers::Platform_TicksToSeconds(Asura_TickCount uTicks)
{
	return static_cast<float>(uTicks) / g_fTimeBaseFrequency;
}


Asura_TickCount Asura_Timers::Platform_SecondsToTicks( float fSeconds )
{
    return static_cast<Asura_TickCount>( fSeconds * g_fTimeBaseFrequency );
}
