//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_CommandBufferInspector.h
//
// Functions to get debug info about the "graphics command management"
// e.g. inspecting the command buffer
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

//
// Includes
//
#include "Core/Asura.h"

#if defined ASURA_PS3_USE_COMMAND_BUFFER_INSPECTOR

#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_PS3_CommandBufferInspector.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_SwapChain.h"

//
// Functions
//

void Asura_PS3_CommandBufferInspector::Initialise()
{
	Asura_CommandConsole::AddCmd(L"Debug.RSXCommands.Inspect", RunCommandConsoleCallback, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddCmd(L"Debug.RSXCommands.RunTests", RunTests, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
}


void Asura_PS3_CommandBufferInspector::Run( const u_int* puBufferStart, const u_int* puBufferEnd )
{
	Asura_OutputToDebugger_Retail(("\nAnalysing command buffer from %08x to %08x\n", puBufferStart, puBufferEnd));

	const u_int* puCmd = puBufferStart;
	bool bDump = false;

	while ( puCmd < puBufferEnd )
	{
		switch ( *puCmd )
		{
			// these codes were found using RunTests()
			
			case 0x00040064:
			{
				// looks like it's a wait message
				if ( puCmd[2] == 0x00040068 ) // cellGcmSetWaitLabel sets this too, anything else could be one of the other waits (if we want to track them use a switch)
				{
					u_int uLabel = puCmd[1]>>4;
					u_int uValue = puCmd[3];
					u_int* pAddress = cellGcmGetLabelAddress(uLabel);
					
					if ( uLabel == ASURA_GCM_LABEL_RSX_PROFILING )
					{
						// special case, the value this waits for gets update by RSX profiling code
						// hence this is actually quite a good measure of how far the frame has got.
						if ( uValue == *pAddress )
						{
							// the RSX has just passed this profiling point, it is somewhere between this and the next one.
							Asura_OutputToDebugger_Retail(("RSX HAS STOPPED AFTER THIS POINT! profiling label, waiting for 0x%08x - current value 0x%08x\n", uValue, *pAddress));
						}
						else if ( uValue == (*pAddress) + 1 )
						{
							// the RSX has not reached this profiling label, if it is stuck, it stuck before this.
							Asura_OutputToDebugger_Retail(("RSX HAS NOT PASSED THIS POINT! profiling label, waiting for 0x%08x - current value 0x%08x\n", uValue, *pAddress));
						}
						else
						{
// 							Asura_OutputToDebugger_Retail(("RSX profiling label %08x (currently %08x)\n", uValue, *pAddress));
						}
					}
					else
					{
						Asura_OutputToDebugger_Retail(("Wait label at 0x%08x: label %d (at 0x%08x) waiting for value 0x%08x - current value is 0x%08x\n",
														puCmd, uLabel, pAddress, uValue, *pAddress));
					}

					puCmd += 3; //in addition to the 1 it will do anyway
				}
				break;
			}
			
			default:
			{
				if ( bDump )
				{
					Asura_OutputToDebugger_Retail(("0x%08x\n", *puCmd));
				}
				break;
			}
		}

		puCmd++;
	}
	
	if ( puCmd < puBufferEnd ) Asura_OutputToDebugger_Retail(("Asura_PS3_CommandBufferInspector - Encountered invalid message header 0x%08x at 0x%08x\n", *puCmd, puCmd ));
}


bool Asura_PS3_CommandBufferInspector::RunCommandConsoleCallback( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf("Analyse this frame's graphics commands. Output to TTY.");
		return true;
	}
	
	//this might be shite, probably the wrong frame
	//Run( Asura_PS3_Graphics::GetGcmCurrentContext()->begin, Asura_PS3_Graphics::GetGcmCurrentContext()->current );
	Run( Asura_PS3_SwapChain::s_paxCommandBuffer[1-Asura_PS3_SwapChain::GetWhich()], Asura_PS3_SwapChain::s_paxCommandBuffer[1-Asura_PS3_SwapChain::GetWhich()] + (4 * 1024 * 1024 - 192 * 1024)/(2*4));
	
	return true;
}

bool Asura_PS3_CommandBufferInspector::RunTests( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf("Create some RSX commands and print out the resulting bits. Output to TTY.");
		return true;
	}
	
	u_int* pStart = Asura_PS3_Graphics::GetGcmCurrentContext()->current;
	
	for ( u_int uPass = 0; uPass < 2; uPass++ )
	{
		u_int uMask = uPass?0xffffffff:0;

		// add whatever commands we want to study, flip the bits of all parameters to see what they are
		cellGcmSetWaitLabelInline( Asura_PS3_Graphics::GetGcmCurrentContext(), 0xff&(uMask^64), uMask);
	}

	u_int* pEnd = Asura_PS3_Graphics::GetGcmCurrentContext()->current;
	
	// output the results
	u_int u2 = (pEnd - pStart)/2;
	for ( u_int u=0; u < u2; u++ )
	{
		Asura_OutputToDebugger_Retail(("%08x\t%08x\t param mask: %08x\n", pStart[u], pStart[u+u2], pStart[u] ^ pStart[u+u2]));
	}
	
	// try to fix the commandbuffer
	Asura_PS3_Graphics::GetGcmCurrentContext()->current = pStart;
	
	return true;
}


#endif //#if defined ASURA_PS3_USE_COMMAND_BUFFER_INSPECTOR
