//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_DebugInfo.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined ASURA_PS3_ALLOW_STATS
// This file's header.
#include "Debugging/Asura_PS3_DebugInfo.h"

// Stdlib independent.
#include <stdlib.h>

// Stdlib platform-specific.
#include <cell/dbgfont.h>
#include <cell/sysmodule.h>

// Engine independent.
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"

// Engine platform-specific.
#include "Rendering/Asura_PS3_ScratchRenderMemory.h"
#include "Rendering/Asura_PS3_TextureManagement.h"
#include "Rendering/Asura_PS3_VideoMemoryManager.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static bool		g_bInit = false;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////
bool	Asura_PS3_DebugInfo::s_bShowStats	= true;

//
// Asura_PS3_DebugInfo
//

void Asura_PS3_DebugInfo::Initialise( void )
{
#if defined( ASURA_SUPPORT_COMMAND_CONSOLE )
	Asura_CommandConsole::AddVar( L"Debug.RenderStats", &s_bShowStats, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif // #if defined( ASURA_SUPPORT_COMMAND_CONSOLE )
	g_bInit = true;
}

void Asura_PS3_DebugInfo::Render( void )
{
	// Don't display the stats if the DVS is up, they get in the way
#if defined ASURA_DEBUG_VARIABLE_SYSTEM
	const bool bDVSOpen = Asura_DebugVariableSystem::IsOpen();
#else
	const bool bDVSOpen = false;
#endif

	if( !g_bInit || !Asura_Render::IsRendering() ) return;

	if( AreStatsEnabled() && !bDVSOpen )
	{
		u_int uMemFreeColour = 0xFFFFFFFF;

		float fVram = static_cast< float >( Asura_PS3_VideoMemoryManager::GetFreeLocal() ) / 1048576.0f;

		if( fVram > 10.0f ) uMemFreeColour = 0xFFFFFF00; // Gold; plenty of memory free.
		if( fVram < 2.0f  ) uMemFreeColour = 0xFFFF0000; // Red; memory getting low.

		u_int	uX	= 15;
		u_int	uY	= 4;

		Asura_DebugText::AddTextBlock(uX, uY++, uMemFreeColour, 0, "Video Mem free: %0.1f MB", fVram);
		
		u_int uAvailableMemory = Asura_PS3_GetFreeMainMemory();

		float fMemFree = static_cast< float >( uAvailableMemory ) / 1048576.0f;

		if( fMemFree > 10.0f ) uMemFreeColour = 0xFFFFFF00; // Gold; plenty of memory free.
		if( fMemFree < 2.0f  ) uMemFreeColour = 0xFFFF0000; // Red; memory getting low.

		// Additionally, if we can, get hold of the system memory not yet allocated, as this can help spot a) when we can't allocate from anything but the heap and b) when the heap is getting heavily fragmented.
		sys_memory_info_t xMemInfo;
		if( CELL_OK == sys_memory_get_user_memory_size(&xMemInfo) )
		{
			u_int uSysMemory = xMemInfo.available_user_memory;
			float fSysFree = static_cast< float >( uSysMemory ) / 1048576.0f;
			Asura_DebugText::AddTextBlock(uX, uY++, uMemFreeColour, 0, "Mem Free:       %0.1f MB (Sys: %0.1f MB)", fMemFree, fSysFree);
		}
		else
		{
			Asura_DebugText::AddTextBlock(uX, uY++, uMemFreeColour, 0, "Mem Free:       %0.1f MB", fMemFree);
		}

		Asura_DebugText::AddTextBlock(uX, uY++, 0xFFFFFFFF,     0, "Total Polys:    %d", Asura_Render::GetNumTriPassesRendered());

		uX				= 60;
		uY				= 4;
		uMemFreeColour	= 0xFFFFFFFF;

        float fMem3, fMem4, fMem5;

//		fMem3 = Asura_PS3_ScratchRenderMemory::GetCurMainMemoryUsage() / 1024.0f;
//		fMem4 = Asura_PS3_ScratchRenderMemory::GetMaxMainMemoryUsage() / 1024.0f;
//		fMem5 = ASURA_PS3_MAIN_SCRATCH_MEMORY_SIZE / 1024.0f;
//		uFrameRateColour = 0xFFFFFF00;
//		if( fMem5 - fMem4 < 100.0f ) uFrameRateColour = 0xFFFFFFFF;
//		if( fMem5 - fMem4 < 25.0f ) uFrameRateColour = 0xFFFF0000;
//		Asura_Font_System::Print( fX4, fY2, fScale, uFrameRateColour, uFormat, L"Main: %0.1f / %0.1f / %0.1f KB", fMem3, fMem4, fMem5 );

		fMem3 = Asura_PS3_ScratchRenderMemory::GetCurMainVideoMemoryUsage() / 1024.0f;
		fMem4 = Asura_PS3_ScratchRenderMemory::GetMaxMainVideoMemoryUsage() / 1024.0f;
		fMem5 = ASURA_PS3_MAIN_VIDEO_SCRATCH_MEMORY_SIZE / 1024.0f;
		uMemFreeColour = 0xFFFFFF00;
		if( fMem5 - fMem4 < 512.0f ) uMemFreeColour = 0xFFFFFFFF;
		if( fMem5 - fMem4 < 128.0f ) uMemFreeColour = 0xFFFF0000;
		Asura_DebugText::AddTextBlock(uX, uY++, uMemFreeColour, 0, "Scratch Render Memory: %0.1f KB (Max: %0.1f KB)", fMem3, fMem4);
		Asura_DebugText::AddTextBlock(uX, uY++, uMemFreeColour, 0, "    (Total available:  %0.1f KB)", fMem5 );

		Asura_Render::FlushDisplay();
	}

	// Artists/designers are too often running master builds and not seeing the stats or asserts, so I'm forcing this to display for
	// the first 250 frames even if stats are turned off
	//
	// Note - the stats are only valid if the texture budget is disabled (a snapshot of the values when the load has finished)
	if (!Asura_PS3_TextureManagement::TextureBudgetEnabled())
	{
		static int s_iTexBudgetFrameCount = 250;		
		if (Asura_PS3_TextureManagement::GetTextureBudgetExceededAmount()>0)
		{
			if (!bDVSOpen && (AreStatsEnabled() || s_iTexBudgetFrameCount>0))
			{				
				float fLimit =   ((float)Asura_PS3_TextureManagement::GetTextureBudget())/(1024.0f*1024.0f);
				float fCurrent = fLimit + ((float)Asura_PS3_TextureManagement::GetTextureBudgetExceededAmount())/(1024.0f*1024.0f);

				Asura_DebugText::AddTextBlock(70, 8, 0xffff3333, 0, "Textures overbudget!: %0.02f MB/%0.02f MB", fCurrent, fLimit);							

				s_iTexBudgetFrameCount--;
			}
		}				
		else
		{
			s_iTexBudgetFrameCount = 250;
		}

		if (Asura_PS3_TextureManagement::GetTextureBudgetExceededAmount()<0)
		{
			if (!bDVSOpen && AreStatsEnabled())
			{				
				uint32 uColour = 0xff33ff33;

				// Change to yellow if we're down to the last 64Mbytes
				if ((-Asura_PS3_TextureManagement::GetTextureBudgetExceededAmount()/1024)<65536)
				{
					uColour = 0xffffff00;
				}					

				float fLimit =   ((float)Asura_PS3_TextureManagement::GetTextureBudget())/(1024.0f*1024.0f);
				float fCurrent = fLimit + ((float)Asura_PS3_TextureManagement::GetTextureBudgetExceededAmount())/(1024.0f*1024.0f);					

				Asura_DebugText::AddTextBlock(70, 8, uColour, 0, "Textures: %0.02f MB/%0.02f MB", fCurrent, fLimit);		
			}
		}
	}
	Asura_Render::FlushDisplay();
}

void Asura_PS3_DebugInfo::Shutdown( void )
{
}

#endif //#if defined ASURA_PS3_ALLOW_STATS
