//
//	Asura_PS3_ExceptionHandler.cpp
//
//	Created By Ben Weston - 2/4/08
//
//	Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_PS3_SUPPORT_EXCEPTION_HANDLER

#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Asura_PS3_ExceptionHandler.h"

#include <cell/sysmodule.h>
#include <string.h>
#include <sys/dbg.h> 
#include <sys/sys_time.h> 

#define EXCEPTION_DUMP_REGISTER_FILE (0)

bool Asura_PS3_ExceptionHandler::bRegistered	= false;

void Asura_PS3_ExceptionHandler::Initialise( void )
{
	int iStatus = cellSysmoduleLoadModule(CELL_SYSMODULE_LV2DBG);
	if (iStatus != CELL_OK)
	{
		Asura_Assert(false,"Error initialising exception handler");
		cellSysmoduleUnloadModule(CELL_SYSMODULE_LV2DBG);
		return;
	}

	//sys_dbg_set_stacksize_ppu_exception_handler( smallish ); default is 16KB

	iStatus = sys_dbg_initialize_ppu_exception_handler(0);
	if ( iStatus != CELL_OK )
	{
		if ( iStatus == (int)CELL_LV2DBG_ERROR_DEOPERATIONDENIED )
		{
			Asura_OutputToDebugger_Retail(("\n"));
			Asura_OutputToDebugger_Retail(("**********************************************************************************\n"));
			Asura_OutputToDebugger_Retail(("* Exception handler is disabled - please enable in your devkit's system settings *\n"));
			Asura_OutputToDebugger_Retail(("**********************************************************************************\n"));
			Asura_OutputToDebugger_Retail(("\n"));
			//BW it's on the debug menu in the main system settings, or you can override that in the target manager's target properties.
		}

		return;
	}

	sys_dbg_register_ppu_exception_handler( Callback, SYS_DBG_PPU_THREAD_STOP );

	bRegistered = true;
}

void Asura_PS3_ExceptionHandler::Shutdown( void )
{
	if ( bRegistered )
	{
		sys_dbg_unregister_ppu_exception_handler();

		sys_dbg_finalize_ppu_exception_handler();
	}
}

void Asura_PS3_ExceptionHandler::Callback( uint64_t uId, sys_ppu_thread_t xThread, uint64_t uDAR )
{
	static system_time_t s_xCoreDumpStartTime;

	if ( uId == SYS_DBG_EVENT_CORE_DUMPED )
	{
		system_time_t duration = sys_time_get_system_time()-s_xCoreDumpStartTime;
		Asura_OutputToDebugger_Retail(("Core dump complete: Please zip it up with the .self file and send it to a programmer!\n(time taken %.0fs)\n", static_cast<float>(duration) / 1000000.0f));
		return;
	}

	Asura_OutputToDebugger_Retail(("\n\n"));
	Asura_OutputToDebugger_Retail(("******************\n"));
	Asura_OutputToDebugger_Retail(("Exception detected\n"));
	Asura_OutputToDebugger_Retail(("******************\n"));

	sys_dbg_ppu_thread_context_t xContextInfo;
	sys_dbg_read_ppu_thread_context( xThread, &xContextInfo );

#if ( EXCEPTION_DUMP_REGISTER_FILE )
	//dump register file
	Asura_OutputToDebugger_Retail(("\nRegisters:\n"));

	for ( int i=0; i < 32; i++ )
	{
		Asura_OutputToDebugger_Retail(("VR[%2d]: %8.5f %8.5f %8.5f %8.5f ", i,
		//Asura_OutputToDebugger_Retail(("VR[%2d]: 0x%08X 0x%08X 0x%08X 0x%08X ", i,
			static_cast<float>(xContextInfo.vr[i].word[0]), static_cast<float>(xContextInfo.vr[i].word[1]),
			static_cast<float>(xContextInfo.vr[i].word[2]), static_cast<float>(xContextInfo.vr[i].word[3])));
		if ( !((i+1)%2) ) Asura_OutputToDebugger_Retail(("\n"));
	}

	Asura_OutputToDebugger_Retail(("\n"));

	for ( int i=0; i < 32; i++ )
	{
		Asura_OutputToDebugger_Retail(("FPR[%2d]: %14.6f ", i, static_cast<double>(xContextInfo.fpr[i])));
		if ( !((i+1)%4) ) Asura_OutputToDebugger_Retail(("\n"));
	}

	Asura_OutputToDebugger_Retail(("\n"));

	for ( int i=0; i < 32; i++ )
	{
		Asura_OutputToDebugger_Retail(("GPR[%2d]: 0x%016LX ", i, xContextInfo.gpr[i]));
		if ( !((i+1)%4) ) Asura_OutputToDebugger_Retail(("\n"));
	}

	Asura_OutputToDebugger_Retail(("\n"));

	Asura_OutputToDebugger_Retail(("VSCR: 0x%08X 0x%08X 0x%08X 0x%08X\n",
		xContextInfo.vscr.word[0], xContextInfo.vscr.word[1], xContextInfo.vscr.word[2], xContextInfo.vscr.word[3]));
	Asura_OutputToDebugger_Retail(("FPSCR: 0x%08X ", xContextInfo.fpscr));
	Asura_OutputToDebugger_Retail(("XER: 0x%016LX ", xContextInfo.xer));
	Asura_OutputToDebugger_Retail(("CTR: 0x%016LX ", xContextInfo.ctr));
	Asura_OutputToDebugger_Retail(("CR: 0x%08X\n", xContextInfo.cr));
#endif

	//dump the important stuff:
	Asura_OutputToDebugger_Retail(("\nPlease Copy & Paste the following into a bug report:\n\n"));

	Asura_OutputToDebugger_Retail(("PC: %#016LX ", xContextInfo.pc));
	Asura_OutputToDebugger_Retail(("LR: %#016LX\n", xContextInfo.lr));

	//type of error
	switch( uId )
	{
		default:
		{
			Asura_OutputToDebugger_Retail(("Exception caused by unknown error (code 0x%016LX), this should be impossible!\n", uId));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_TRAP:
		{
			Asura_OutputToDebugger_Retail(("Exception: Trap instruction\n"));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_PREV_INST:
		{
			Asura_OutputToDebugger_Retail(("Exception: Privileged instruction\n"));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_ILL_INST:
		{
			Asura_OutputToDebugger_Retail(("Exception: Illegal instruction\n"));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_TEXT_HTAB_MISS:
		{
			Asura_OutputToDebugger_Retail(("Exception: Instruction Storage Interrupt Occurred\n"));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_TEXT_SLB_MISS:
		{
			Asura_OutputToDebugger_Retail(("Exception: Instruction Segment Interrupt Occurred\n"));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_DATA_HTAB_MISS:
		{
			Asura_OutputToDebugger_Retail(("Exception: Data Storage Interrupt Occurred.\nData register: 0x%016LX\n", uDAR));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_DATA_SLB_MISS:
		{
			Asura_OutputToDebugger_Retail(("Exception: Data Segment Interrupt Occurred.\nData register: 0x%016LX\n", uDAR));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_FLOAT:
		{
			Asura_OutputToDebugger_Retail(("Floating point exception - Sony says this can't cause an exception!!\n"));
			//BW liblv2dbg-Reference_e.pdf -
			//"This interrupt does not occur in the Reference Tool, Debugging Station, and PLAYSTATION3 environment because
			//the floating point exception mode is set to ignore exception"
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_DABR_MATCH:
		{
			Asura_OutputToDebugger_Retail(("Exception: Data Storage Interrupt Occurred.\nData register: 0x%016LX\n", uDAR));
			break;
		}
		case SYS_DBG_PPU_EXCEPTION_ALIGNMENT:
		{
			Asura_OutputToDebugger_Retail(("Exception: Data Storage Interrupt Occurred.\nData register: 0x%016LX\n", uDAR));
			break;
		}
	}

	//is core dump enabled?
	//Sony doesn't seem to have a nice way to do this, so try to detect core dump thread
	bool bCoreDump = false;
	sys_ppu_thread_t axThreads[64];
	uint64_t uNumIDs = 64;
	uint64_t uNumAllIDs;
	sys_dbg_get_ppu_thread_ids( axThreads, &uNumIDs, &uNumAllIDs );

	for ( int i=0; i<(int)uNumIDs; i++ )
	{
		char name[28];
		sys_dbg_get_ppu_thread_name( axThreads[i], name );
		if ( strcmp( name, "__COREDUMP_HANDLER__" ) == 0 )
			bCoreDump = true;
	}

	if ( bCoreDump )
	{
		Asura_OutputToDebugger_Retail(("\nBeginning core dump, this could take a couple of minutes.\n"));
		Asura_OutputToDebugger_Retail(("Please do not reboot this system until the \"Core Dump Complete\" message...\n...\n"));
		s_xCoreDumpStartTime = sys_time_get_system_time();
	}
	else
	{
		Asura_OutputToDebugger_Retail(("\n*CANNOT DUMP CORE* Core dump is disabled, please enable Core Dump in the debug settings of your test kit.\n"));
	}
}

#endif //ASURA_PS3_SUPPORT_EXCEPTION_HANDLER

//eof
