//
// File      : Asura_PS3_FMV.cpp
// Author    : Philip Hall
// Created   : 18/10/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "FMV/Asura_PS3_FMV.h"

// Asura Headers
#include "Core/Asura_UserInput.h"

// Asura PS3 headers
#include "FMV/Asura_PS3_FMV_pamf.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_QuadOverlayRenderer.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_SwapChain.h"

#include <cell/codec.h>
#include <cell/codec/adec.h>
#include <cell/audio.h>
#include <cell/gcm.h>


const u_int     PS3_FMV_ES_NUM = 2;
const int       PS3_FMV_PPU_THREAD_PRIO = 600;
const size_t    PS3_FMV_PPU_STACK_SIZE = 16384;
const int       PS3_FMV_1VSYNC_ON_TIMESTAMP = (90000/(60 * 1000 / 1001));
const int       PS3_FMV_VIDEO_SYNC_DELTA = PS3_FMV_1VSYNC_ON_TIMESTAMP * 2;

const usecond_t PS3_FMV_UPDATE_POLL_WAIT = 1000;
const usecond_t PS3_FMV_UPDATE_DECODE_WAIT = 2000;

enum ASURA_PS3_FMV_TERMINATION
{
	ASURA_PS3_FMV_TERMINATION_KEY = 1,	//FMV terminated due to user keypress
	ASURA_PS3_FMV_TERMINATION_DATA,		//FMV terminated because the video ended
	ASURA_PS3_FMV_TERMINATION_SHUTDOWN	//FMV terminated because the system is shutting down
};

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::Asura_PS3_FMV
// Description : FMV playback mechanism
//
// Parameters  : 
//   const char* szFilename  - Name of movie file (copied and ext modified)
//   int iAudioStream        - Index of audio stream to play
//   u_int uFlags            - Playback flags
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_FMV::Asura_PS3_FMV(const char* szFilename, int iAudioStream, u_int uFlags, Asura_PS3_SubtitleCallback pfnCallback)
{
	// Remember the playback flags 
	m_uFlags = uFlags;

	// Regardless of the filename we were given, strip the extension and append the PS3 specific one
	strncpy( m_szFilename, szFilename, ASURA_PS3_MAX_PATH );
	char *szExt = strrchr(m_szFilename, '.');
	if ( szExt == NULL ) { szExt = m_szFilename + strlen(m_szFilename); }
	strncpy(szExt, ".pam", 5);

	// Init member variables
	m_iAudioStream = iAudioStream;
	m_bAudioStarted = false;
	m_pAudioInfo = NULL;
	m_pVideoInfo = NULL;
	m_paxVideoFrame[ 0 ] = NULL;
	m_paxVideoFrame[ 1 ] = NULL;
	m_iDMuxTime = 0;
	m_iVideoTimeout = 0;
	m_iDMuxLookahead = 0;
	m_uDecodeThreadID = 0;
	m_pfnSubtitleCallback = pfnCallback;
	m_pSubtitleArg = this;
}

Asura_PS3_FMV::~Asura_PS3_FMV()
{
	Shutdown();
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::Initialise
// Description : Initialise the FMV playback system
//
// Parameters  : none
//
// Returns     : bool - Returns true on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV::Initialise()
{
	// Load and extract stream information from the files
	Asura_PS3_FMV_pamf xPAMFInfo;
	if ( !xPAMFInfo.Open(m_szFilename) )
	{
		Asura_Assertf(false, ("[Asura_PS3_FMV::Initialise] - Failed to open file for PAMF information extraction:\nFile: '%s'\n", m_szFilename));
		return false;
	}

	// Initialise the AV components
	m_xStreamer.Initialise(&m_xControl, xPAMFInfo.GetFilename(), xPAMFInfo.GetStreamOffset(),	xPAMFInfo.GetStreamSize());
	m_xDMux.Initialise(&m_xControl, PS3_FMV_ES_NUM);
	m_xVDec.Initialise(&m_xControl);
	m_xADec.Initialise(&m_xControl);

	// PAMF open
	if ( !xPAMFInfo.GetVideoStreamInfo(&m_xVideoInfo) )
	{
		Asura_Assert(false, "[Asura_PS3_FMV::Initialise] - Couldn't get PAMF video stream info\n");
		return false;
	}
	
	if ( !xPAMFInfo.GetAudioStreamInfo(&m_xAudioInfo, m_iAudioStream) )
	{
		// PAZ - default to stream 0 if the specified stream does not exist
		if ( !xPAMFInfo.GetAudioStreamInfo(&m_xAudioInfo, 0) )
		{
			Asura_Assert(false, "[Asura_PS3_FMV::Initialise] - Couldn't get PAMF audio stream info\n");
			return false;
		}
	}

	// Check there are streams present
	if ( !(m_xVideoInfo.bPresent || m_xAudioInfo.bPresent) )
	{
		Asura_Assert(false, "[Asura_PS3_FMV::Initialise] - Video contains no valid video nor audio streams\n");
		return false;
	}

	// DMux open
	if ( !m_xDMux.Open() )
	{
		Asura_Assert(false, "[Asura_PS3_FMV::Initialise] - Failed to open DMUX\n");
		return false;
	}

	// Initialise audio decoder	
	if ( m_xAudioInfo.bPresent )
	{
		m_pAudioInfo = m_xDMux.EnableEs(m_xAudioInfo.uType, &m_xAudioInfo.esFilterId, &m_xAudioInfo.esSpecInfo);

		Asura_OutputToDebugger(("[Asura_PS3_FMV::Initialise] - Audio: %d channels of %s\n", m_xAudioInfo.uChannels, m_xAudioInfo.szType));
		if ( !m_xADec.Open(m_pAudioInfo, m_xAudioInfo.uChannels) )
		{
			Asura_Assert(false, "[Asura_PS3_FMV::Initialise] - Failed to open audio decoder\n");
			return false;
		}
	}
	
	// Initialise the video decoder
	if ( m_xVideoInfo.bPresent )
	{
		m_pVideoInfo = m_xDMux.EnableEs(m_xVideoInfo.uType, &m_xVideoInfo.esFilterId, &m_xVideoInfo.esSpecInfo);

		Asura_OutputToDebugger(("[Asura_PS3_FMV::Initialise] - Video: %dx%d %s, using %d SPUs\n", m_xVideoInfo.uWidth, m_xVideoInfo.uHeight, m_xVideoInfo.szType, m_xVideoInfo.vdecSpuNum));
		if ( !m_xVDec.Open(m_pVideoInfo, m_xVideoInfo.uWidth, m_xVideoInfo.uHeight, m_xVideoInfo.vdecSpuNum) )
		{
			Asura_Assert(false, "[Asura_PS3_FMV::Initialise] - Failed to open video decoder\n");
			return false;
		}
	}

	InitialiseVideo();

	if ( !m_xStreamer.Open() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::Open] - Streamer failed to open, can't continue\n"));
		return false;
	}

	return true;
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::Shutdown
// Description : Shutdown and clean up the FMV system post playback
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV::Shutdown()
{
	if ( m_paxVideoFrame[ 0 ] != NULL )
	{
		Asura_PS3_VideoMemoryManager::FreeMain( m_paxVideoFrame[ 0 ] );
		m_paxVideoFrame[ 0 ] = NULL;
	}

	if ( m_paxVideoFrame[ 1 ] != NULL )
	{
		Asura_PS3_VideoMemoryManager::FreeMain( m_paxVideoFrame[ 1 ] );
		m_paxVideoFrame[ 1 ] = NULL;
	}
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::InitialiseVideo
// Description : Initialise the video system for fullscreen FMV playback
//
// Parameters  : none
//
// Returns     : bool - Returns true on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV::InitialiseVideo()
{
//	Asura_Render::SetEnableScreenClear(true);
	float fAspect = static_cast<float>(m_xVideoInfo.uHeight) / m_xVideoInfo.uWidth;

	float fActualHeight = fAspect * Asura_PS3_Graphics::GetScreenWidth();
	float fBorder = (Asura_PS3_Graphics::GetScreenHeight() - fActualHeight ) / 2.0f;

	m_xVideoQuad.SetRectXY(0, fBorder, Asura_PS3_Graphics::GetScreenWidth()-1.0f, fActualHeight );
	m_xVideoQuad.SetRectUV( 0.0f, 0.0f, 1.0f, 1.0f );

	m_xVideoQuad.TranslucencyMode =	ASURA_TRANSLUCENCY_OFF;
	m_xVideoQuad.Colour = 0xFFFFFFF;
	m_xVideoQuad.TextureIndex = ASURA_TEXTURE_NONE;

	// Allocate video frame texture memory
	m_paxVideoFrame[ 0 ] = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_xVideoInfo.uWidth * m_xVideoInfo.uHeight * 4, 128);
	m_paxVideoFrame[ 1 ] = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_xVideoInfo.uWidth * m_xVideoInfo.uHeight * 4, 128);

	// Populate GCM texture information
	CellGcmTexture xTexture;
	xTexture.width = m_xVideoInfo.uWidth;
	xTexture.height = m_xVideoInfo.uHeight;
	xTexture.mipmap = 1;
	xTexture.format = CELL_GCM_TEXTURE_A8R8G8B8 | CELL_GCM_TEXTURE_LN | CELL_GCM_TEXTURE_NR;
	xTexture.pitch = xTexture.width * 4;
	xTexture.cubemap = CELL_GCM_FALSE;
	xTexture.location = CELL_GCM_LOCATION_MAIN;
	xTexture._padding = 0;
	xTexture.depth = 1;
	xTexture.dimension = CELL_GCM_TEXTURE_DIMENSION_2;
	xTexture.remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_A,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );

	// Create texture
	cellGcmAddressToOffset( m_paxVideoFrame[ 0 ], &xTexture.offset );;
	m_axVideoTexture[ 0 ] = Asura_PS3_Texture( &xTexture );
	m_axVideoTexture[ 0 ].SetTextureData( m_paxVideoFrame[ 0 ] );

	cellGcmAddressToOffset( m_paxVideoFrame[ 1 ], &xTexture.offset );;
	m_axVideoTexture[ 1 ] = Asura_PS3_Texture( &xTexture );
	m_axVideoTexture[ 1 ].SetTextureData( m_paxVideoFrame[ 1 ] );
	m_iCurrentTexture = 0;

	return true;
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::Begin
// Description : Begin video playback
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV::Begin()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV::Begin] - Beginning...\n"));
	m_xControl.Begin();

	m_bAudioStarted = false;
	m_iDMuxLookahead = 0;

	m_bSkipping = false;

	if ( m_xStreamer.Begin() )
	{
		// Decoding thread
		int iResult = sys_ppu_thread_create(&m_uDecodeThreadID, &Asura_PS3_FMV::DecodeThreadProxy, (uintptr_t)this, 
											PS3_FMV_PPU_THREAD_PRIO, PS3_FMV_PPU_STACK_SIZE, SYS_PPU_THREAD_CREATE_JOINABLE, "Asura_PS3_FMV::DecodeThread");
		if (iResult < CELL_OK)
		{
			Asura_Assert(false, "[Asura_PS3_FMV::Begin] - Can't create PPU thread\n");
		}
	}
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::End
// Description : End video playback (waits for either completion or skip)
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV::End()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV::End] - Ending...\n"));
	uint64_t uStat;
	int	iResult = sys_ppu_thread_join(m_uDecodeThreadID, &uStat);
	if (iResult < CELL_OK)
	{
		Asura_Assert(false, "[Asura_PS3_FMV_dmux::WaitComplete] - Unable to join thread\n");
	}

	// End streamer thread
	m_xStreamer.End();
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::Play
// Description : Play the movie file with the given parameters, if asynchronous
//               it's up to you to delete the returned object.
//
// Parameters  :
//   const char* szFilename   - Name of file to play back
//   int         iAudioStream - Audio stream index
//   u_int       uFlags       - Playback flags
//   Asura_PS3_SubtitleCallback pfnCallback - Subtitle callback
//
// Returns     : bool  - Returns false if playback failed
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV::Play(const char* szFilename, int iAudioStream, u_int uFlags, Asura_PS3_SubtitleCallback pfnCallback)
{
	Asura_Assert(!( ( uFlags & ASURA_PS3_FMV_LOOP) && !(uFlags & ASURA_PS3_FMV_ALLOWSKIP) ) ,"PS3 Video set to loop but unskippable");

	if ( Asura_PS3_Graphics::IsInShutdown() ) return true; //the video didn't actually fail

	u_int uTermination = 0;
	//TY 11.01.08 - Loop until we get user input if necessary; For now delete and recreate the FMV object since
	//the class doesnt work if we repeatedly call Begin -> Update -> End. However we will probably want to code
	//it slightly differently in future to avoid this.
	do 
	{
		Asura_PS3_FMV* pxFMV = new Asura_PS3_FMV(szFilename, iAudioStream, uFlags, pfnCallback);
		if ( !pxFMV->Initialise() )
		{
			delete pxFMV;
			return false;
		}
		
		pxFMV->Begin();
		uTermination = pxFMV->Update();
		pxFMV->End();
		delete pxFMV;
	} 
	while((uTermination == ASURA_PS3_FMV_TERMINATION_DATA) && (uFlags & ASURA_PS3_FMV_LOOP));	//Loop if necessary
	
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::Update
// Description : Update video playback
//
// Parameters  : none
//
// Returns     : the reason why it terminated (key or end of file)
///////////////////////////////////////////////////////////////////////////////
u_int Asura_PS3_FMV::Update()
{
	u_int uTermination = 0;

	// Set vsync on for playback (otherwise heavy tearing results)
	u_int uPresentInterval = Asura_PS3_SwapChain::GetPresentInterval();
	cellGcmSetFlipMode( CELL_GCM_DISPLAY_VSYNC );
	Asura_PS3_SwapChain::SetPresentInterval( 0 );
	Asura_PS3_SwapChain::SetIsInFMV( true );

	m_iVideoTimeout = 0;
	int iSkipDelay = 2;
	
	m_xStartTicks = Asura_Timers::GetCurTicks();
	
	// Force an input update just before we start	
	Asura_UserInput::Update();
	
	// Loop until aborted or end of video playback
	while ( !m_xControl.IsComplete() )
	{
		if ( Asura_PS3_Graphics::IsInShutdown() )
		{
			m_xControl.Complete();
			uTermination = ASURA_PS3_FMV_TERMINATION_SHUTDOWN;
		}

		if ( m_xControl.IsEndOfAudioData() && m_xControl.IsEndOfVideoData() && m_xVDec.Buffer().IsEmpty() && m_xADec.Buffer().IsEmpty() )
		{
			m_xControl.Complete();
			uTermination = ASURA_PS3_FMV_TERMINATION_DATA;
		}

		// Start audio playback if required
		if ( m_xAudioInfo.bPresent && !m_bAudioStarted )
		{
			if ( !m_xVDec.Buffer().IsEmpty() )
			{	
				m_bAudioStarted = m_xADec.StartAudio();
			}
		}

		// Keypress skip
		if(m_bSkipping)
		{
			Asura_UserInput::Update();

			if ( !Asura_UserInput::HasAnyKeyBeenPressed() )
			{
				m_xControl.Abort(CELL_OK);
				m_xStreamer.Abort();
				m_xADec.StopAudio();
				uTermination = ASURA_PS3_FMV_TERMINATION_KEY;
			}
		}
		else if ( ((m_uFlags & ASURA_PS3_FMV_ALLOWSKIP) != 0) )
		{
			Asura_UserInput::Update();

			if ( ( iSkipDelay <= 0 ) && Asura_UserInput::DebouncedHasAnyKeyBeenPressed() )
			{
				m_bSkipping = true;
			}
			else
			{
				iSkipDelay--;
			}
		}

		// Draw current video frame		
		RenderVideo();
		
		Asura_PS3_Graphics::CheckExitCallBack();
	}

	m_xADec.StopAudio();

	Asura_PS3_SwapChain::SetPresentInterval( uPresentInterval );
	Asura_PS3_SwapChain::SetIsInFMV( false );
	cellGcmSetFlipMode( CELL_GCM_DISPLAY_HSYNC );
	Asura_OutputToDebugger( ("[Asura_PS3_FMV::Update] - cellGcmSetFlipMode( CELL_GCM_DISPLAY_HSYNC )\n") );

	return uTermination;
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::RenderVideo
// Description : Render the video frames and framesync
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV::RenderVideo()
{
	// Sleep and return if there's no video stream
	if ( !m_xVideoInfo.bPresent )
	{
		sys_timer_usleep(PS3_FMV_UPDATE_POLL_WAIT);
		return;
	}

	bool bDisplay = false;
	float fDisplayTime = 0.0f;
	int iATime = m_xControl.GetATime().lower;

	Asura_PS3_VideoBlock* pBlock = m_xVDec.Buffer().Peek(0);
	if ( pBlock != NULL)
	{
		// Only sync if there's still audio to sync against
		if ( m_xAudioInfo.bPresent && !m_xControl.IsEndOfAudioData() && m_bAudioStarted )
		{
			// Calculate video skew
			int iDiff = iATime - static_cast<int>(pBlock->xTimestamp.lower);

			if ( (iDiff > -PS3_FMV_VIDEO_SYNC_DELTA) && (iDiff < PS3_FMV_VIDEO_SYNC_DELTA) )
			{
				Asura_Mem::Copy( m_paxVideoFrame[ m_iCurrentTexture ], pBlock->pFrame, pBlock->uSize );
				m_xVDec.Buffer().Drop();
				bDisplay = true;
				m_iVideoTimeout = 0;
			}
			// We're behind, drop frames
			else if ( iDiff >= PS3_FMV_VIDEO_SYNC_DELTA )
			{
				m_xVDec.Buffer().Drop();
				m_iVideoTimeout = 0;
			}
			// If we're ahead wait for a while...
			else if ( iDiff < -PS3_FMV_VIDEO_SYNC_DELTA )
			{
				m_iVideoTimeout++;
				
				// If we've been ahead for too long, start waiting	
				if ( m_iVideoTimeout > 2500 )
				{
					// If we're very far ahead, abort playback
					if ( m_iVideoTimeout > 5000 ) 
					{
						Asura_Assert(false, "[Asura_PS3_FMV::RenderVideo] - Aborted due to long-term audio sync error (the audio doesn't seem to be playing fast enough!)\n");
						m_xControl.Abort(RET_CODE_ERR_FATAL);
					}
					
					sys_timer_usleep(50);  // Sleep to allow the audio playback to catch up - allows for a max 125ms wait with a counter difference of 2500
				}
			}

			// Use measured time (should really use sync time, but it's miles out)
			fDisplayTime = Asura_Timers::TicksToSeconds( Asura_Timers::GetElapsedTicks(m_xStartTicks) );
		}
		else 
		{
			// Normal display if there's no audio (so manually synced at the moment)
			Asura_Mem::Copy( m_paxVideoFrame[ m_iCurrentTexture ], pBlock->pFrame, pBlock->uSize );
			m_xVDec.Buffer().Drop();
			sys_timer_usleep(1000000 / 30); // manual sync hack for no-audio
			bDisplay = true;

			// Use measured time
			fDisplayTime = Asura_Timers::TicksToSeconds( Asura_Timers::GetElapsedTicks(m_xStartTicks) );
		}
 	}

	// This hack allows us to display the blank frame when we've completed
	if ( m_xControl.IsComplete() )
	{
		// Blank the video buffer
		Asura_Mem::Set( m_paxVideoFrame[ m_iCurrentTexture ], 0, m_xVideoInfo.uWidth * m_xVideoInfo.uHeight * 4 ); 
		bDisplay = true;
	}

	if ( bDisplay )
	{
		// Simply draws the video texture - the texture is filled in from the vdec class.
		Asura_Render::BeginFrame();
		Asura_PS3_DeferredRendering::BindAsSingleRT();
		Asura_PS3_QuadOverlayRenderer::RenderEx( m_xVideoQuad, &m_axVideoTexture[ m_iCurrentTexture ], false, Asura_PS3_Graphics::GetScreenWidth(), Asura_PS3_Graphics::GetScreenHeight() );
		m_iCurrentTexture = ( m_iCurrentTexture + 1 ) % 2;

		// Make the subtitle callback for this frame
		if ( m_pfnSubtitleCallback != NULL )
		{		
			m_pfnSubtitleCallback(fDisplayTime, m_pSubtitleArg);
		}
		Asura_Render::EndFrame();                        // This will cause a vertical-retrace sync in submission mode
		Asura_Timers::HandlePrimaryTimer();
	}
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::DecodeThread
// Description : Demux->Video->Audio decoding thread
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV::DecodeThread()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV::Thread] - Started...\n"));

	m_iDMuxTime = 0;

	// Decode until completed
	while (true)
	{
		if ( m_xControl.IsComplete() ) { break; }

		// Do the demuxing		
		m_iDMuxTime--;
		if ( m_iDMuxTime <= 0 )
		{
			ThreadDMUX();
			m_iDMuxTime = 100; // We don't need to demux very often (decoders can request more if needed)
		}

		// Do the audio decoding (if there's an audio stream)
		if ( m_xAudioInfo.bPresent )
		{
			Asura_PS3_FMV_DecodeResult eResult = m_xADec.StartProcessing();
			if ( eResult == ASURA_PS3_FMV_DECODE_EMPTY )
			{
				m_iDMuxTime = 0;
				if ( m_xControl.IsEndOfDMux() && !m_xControl.IsEndOfAudioData() && m_xADec.Buffer().IsEmpty() )
				{
					m_xControl.EndAudioData();  // Finished audio (there shouldn't be any AUs left at this point)
				}
			}
		}

		// Do the video decoding
		if ( m_xVideoInfo.bPresent )
		{
			Asura_PS3_FMV_DecodeResult eResult = m_xVDec.StartProcessing();
			if ( eResult == ASURA_PS3_FMV_DECODE_EMPTY )
			{
				m_iDMuxTime = 0;
				if ( m_xControl.IsEndOfDMux() && !m_xControl.IsEndOfVideoData() && m_xVDec.Buffer().IsEmpty() )
				{
					m_xControl.EndVideoData();	// Finished video (there shouldn't be any AUs left at this point)
				} 
			}
		}

		sys_timer_usleep(PS3_FMV_UPDATE_DECODE_WAIT);
	}
	
	Asura_OutputToDebugger(("[Asura_PS3_FMV::Thread] - ...Finished\n"));
	sys_ppu_thread_exit(0);
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV::ThreadDMUX
// Description : Process demuxing
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV::ThreadDMUX()
{
	if ( !m_xDMux.IsBusy() && !m_xControl.IsEndOfDMux() )
	{
		// If the streamer's stopped and the buffer's empty then we're done
		if ( m_xControl.IsEndOfStream() && m_xStreamer.Buffer().IsEmpty() )
		{
			if ( m_xDMux.ProcessedCount() == 0 )
			{
				m_xControl.EndDMux();
			}
			return;
		}

		// If streamer has new blocks, feed them to the demuxer
		if ( m_xStreamer.Signal().GetValue() > 0 )
		{
			if ( m_xStreamer.Lock(0) )
			{
				Asura_PS3_StreamBlock* pBlock = m_xStreamer.Buffer().Peek(m_iDMuxLookahead);
				if ( pBlock->uSize > 0 )
				{
					if ( m_xDMux.StartProcessing( pBlock->acData, pBlock->uSize, pBlock->bStart ) )
					{
						m_xStreamer.Signal().Wait(1);
						m_iDMuxLookahead++;
					}
				}
				else
				{
					m_xStreamer.Buffer().Drop();   // Drop empty blocks
				}

				m_xStreamer.Unlock();
			}
		}
	}

	// Drop completed blocks from the stream buffers
	const int iDMUXProcessed = m_xDMux.ProcessedCount();
	if ( iDMUXProcessed > 0 )
	{
		if ( m_xStreamer.Lock(0) )
		{
			m_xStreamer.Buffer().Drop(iDMUXProcessed);
			m_xDMux.AdjustProcessedCount( -iDMUXProcessed );
			m_iDMuxLookahead -= iDMUXProcessed;
		}
		m_xStreamer.Unlock();
	}
}

//eof
