///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_FMV_dmux.cpp
// Author    : Philip Hall
// Created   : 18/10/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "FMV/Asura_PS3_FMV_dmux.h"

// Asura headers
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_LogFile.h"

#include <cell/sysmodule.h>


const uint32_t  DMUX_LIB_PPU_THREAD_PRIO = 450;
const size_t    DMUX_LIB_PPU_STACK_SIZE  = 16384;
const uint32_t  DMUX_LIB_SPU_THREAD_PRIO = 150;
const u_int     DMUX_LIB_SPU_NUM         = 1;


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::Asura_PS3_FMV_dmux
// Description : Initialise dmux wrapper 
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_FMV_dmux::Asura_PS3_FMV_dmux()
{
	m_pWorkMemory = NULL;
	m_uMemSize    = NULL;
	m_axESInfo    = NULL;
	m_handle      = NULL;
	m_bDmuxBusy   = false;
	m_iProcessedCount = 0;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::~Asura_PS3_FMV_dmux
// Description : Clean up and free allocated memory
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_FMV_dmux::~Asura_PS3_FMV_dmux()
{
	Close();
}

void Asura_PS3_FMV_dmux::Initialise(Asura_PS3_FMV_Control* pControl, uint32_t uESCount)
{
	m_pControl = pControl;
	m_xType.streamType = CELL_DMUX_STREAM_TYPE_PAMF;
	m_xType.reserved1 = 0;
	m_xType.reserved2 = 0;
	m_uESCount = uESCount;
	m_iProcessedCount = 0;
	m_iRunningFrameCount = 0;
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::Open
// Description : Open the dmuxer with the current proivded settings
//
// Parameters  : none
//
// Returns     : bool - Returns true on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV_dmux::Open()
{
	int iResult;
	CellDmuxCb cb = { &Asura_PS3_FMV_dmux::Callback, this };

	Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::Open] - Opening dmuxer\n"));

	// Load the dmux module
	iResult = cellSysmoduleLoadModule(CELL_SYSMODULE_DMUX_PAMF);
	if (iResult < CELL_OK)
	{
		Asura_Assert(false, "[Asura_PS3_FMV_dmux::Open] - Unable to load DMUX module\n");
		goto labelError;
	}

	CellDmuxAttr xAttr;
	iResult = cellDmuxQueryAttr(&m_xType, &xAttr);
	if (iResult < CELL_OK)
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::Open] - cellDmuxQueryAttr failed (error = %X)\n", iResult));
		goto labelError;
	}

	m_uMemSize = xAttr.memSize;
	m_pWorkMemory = reinterpret_cast<uint8_t*>( Asura_Allocate(m_uMemSize) );
	if (m_pWorkMemory == NULL)
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::Open] - Unable to allocate working memory\n"));
		goto labelError;
	}
	
	m_axESInfo = new Asura_PS3_FMV_ESInfo[m_uESCount];
	if ( m_axESInfo == NULL )
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::Open] - Unable to allocate mux ES memory\n"));
		goto labelError;
	}
	
	// Initialise dmux resource block
	CellDmuxResource res;
	res.memAddr				= m_pWorkMemory;
	res.memSize				= m_uMemSize;
	res.ppuThreadPriority	= DMUX_LIB_PPU_THREAD_PRIO;
	res.ppuThreadStackSize	= DMUX_LIB_PPU_STACK_SIZE;
	res.spuThreadPriority	= DMUX_LIB_SPU_THREAD_PRIO;
	res.numOfSpus			= DMUX_LIB_SPU_NUM;

	// Open the dmuxer
	iResult = cellDmuxOpen(&m_xType, &res, &cb, &m_handle);
	if (iResult < CELL_OK)
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::Open] - cellDmuxOpen failed (error = %X)\n", iResult));
		goto labelError;
	}
	
	return true;
		
// Exit after error with cleanup
labelError:
	Close();
	return false;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::Close
// Description : Close down and clean up after the dmuxer
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV_dmux::Close()
{
	if ( m_handle != NULL )
	{
		int iResult;
		iResult = cellDmuxClose(m_handle);
		if (iResult < CELL_OK)
		{
			Asura_Assertf(false, ("[Asura_PS3_FMV_dmux::Close] - cellDmuxClose failed (error = %X)", iResult));
		}
		m_handle = NULL;
	}

	if ( m_pWorkMemory != NULL )
	{
		Asura_Deallocate(m_pWorkMemory);
		m_pWorkMemory = NULL;
	}
	
	if ( m_axESInfo != NULL )
	{
		delete[] m_axESInfo;
		m_axESInfo = NULL;
	}

	m_xLock.Destroy();

	// Unload the dmux module
	cellSysmoduleUnloadModule(CELL_SYSMODULE_DMUX_PAMF);
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::EnableEs
// Description : Enable processing for the selected elementary stream (ES)
//
// Parameters  : 
//   uint8_t uStreamType               - Type of stream (audio/video/etc)
//   CellDmuxEsFilterId* pEsFilterId  - ES filter
//   UDmuxEsSpecInfo* pEsSpecInfo     - ES type-specific information
//
// Returns     : Asura_PS3_FMV_ESInfo* - ES processing information buffer
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_FMV_ESInfo* Asura_PS3_FMV_dmux::EnableEs(uint8_t uStreamType, CellDmuxEsFilterId* pEsFilterId, UDmuxEsSpecInfo* pEsSpecInfo)
{
	int iResult;
	Asura_PS3_FMV_ESInfo* pDmuxEsInfo = NULL;
	for (uint32_t index = 0; index < m_uESCount; ++index)
	{
		if (m_axESInfo[index].pWorkMemory == NULL)
		{
			pDmuxEsInfo = &m_axESInfo[index];
			break;
		}
	}
	
	if (pDmuxEsInfo == NULL)
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::EnableEs] - Can't allocate ES info"));
		return NULL;
	}
	
	pDmuxEsInfo->uStreamType = uStreamType;
	
	CellDmuxEsAttr	xAttr;
	iResult = cellDmuxQueryEsAttr(&m_xType, pEsFilterId, pEsSpecInfo, &xAttr);
	if (iResult  < CELL_OK)
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::EnableEs] - cellDmuxQueryEsAttr failed (error = %d)", iResult));
		return NULL;
	}
	
	// Initialise dmux ES storage
	if ( !pDmuxEsInfo->Create(xAttr.memSize) )
	{
		return NULL;
	}
	
	CellDmuxEsResource	res = { (void*)pDmuxEsInfo->pWorkMemory, xAttr.memSize };
	CellDmuxEsCb		cb  = { &Asura_PS3_FMV_dmux::ESCallback, this };
	iResult = cellDmuxEnableEs(m_handle, pEsFilterId, &res, &cb, pEsSpecInfo, &pDmuxEsInfo->xHandle);
	if (iResult < CELL_OK)
	{
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::EnableEs] - cellDmuxEnableEs failed (error = %X)\n", iResult));
		pDmuxEsInfo->Destroy();
		return NULL;
	}

	pDmuxEsInfo->xSpecInfo = *pEsSpecInfo;
	return pDmuxEsInfo;
}


void Asura_PS3_FMV_dmux::DisableEs(Asura_PS3_FMV_ESInfo* pDmuxEsInfo)
{
	int iResult = cellDmuxDisableEs( pDmuxEsInfo->xHandle );
	if ( iResult != CELL_OK )
	{
		// This is not particularly bad since we're closing it down 
		Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::DisableEs] - cellDmuxDisableES failed (error = %X)\n", iResult));
	}

	pDmuxEsInfo->Destroy();
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::StartProcessing
// Description : Start processing the given block (dmuxer must be idle)
//
// Parameters  : 
//   void* pData   - Data block to process
//   size_t uSize  - Size of data to process
//   bool bFirst   - Set true if this is the first of the sequence
//
// Returns     : bool - Returns true if processing already started 
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV_dmux::StartProcessing(void* pData, size_t uSize, bool bFirst)
{
	int iResult = cellDmuxSetStream(m_handle, pData, uSize, bFirst, reinterpret_cast<uint64_t>(m_axESInfo) );
	switch( iResult)
	{
		case CELL_OK:
		break;

		case CELL_DMUX_ERROR_ARG:
			Asura_Assert(false, "[Asura_PS3_FMV_dmux::StartProcessing] - Bad DMUX argument" );
			m_pControl->Abort(RET_CODE_ERR_FATAL);
		break;

		case CELL_DMUX_ERROR_FATAL:
			Asura_OutputToDebugger(("[Asura_PS3_FMV_dmux::StartProcessing] - DMUX Fatal error"));
			m_pControl->Abort(RET_CODE_ERR_FATAL);
		break;

		case CELL_DMUX_ERROR_BUSY:
			//Asura_OutputToDebugger(("DMUX Busy!\n"));
			if ( m_xLock.Lock(0) )
			{
				m_bDmuxBusy = true;
				m_xLock.Unlock();
			}
		break;
		
		default:
			Asura_Assertf(false, ("[Asura_PS3_FMV_dmux::StartProcessing] - Result unhandled %d\n", iResult) );
			m_pControl->Abort(RET_CODE_ERR_FATAL);
		break;
	}

	return !m_bDmuxBusy;
}


bool Asura_PS3_FMV_dmux::IsBusy()
{
	return m_bDmuxBusy;
}

int Asura_PS3_FMV_dmux::ProcessedCount()
{
	int iProcessed = 0;
	if ( m_xLock.Lock(0) )
	{
		iProcessed = m_iProcessedCount;
		m_xLock.Unlock();
	}
	return iProcessed;
}

void Asura_PS3_FMV_dmux::AdjustProcessedCount(int iCount)
{
	if ( m_xLock.Lock(0) )
	{
		m_iProcessedCount += iCount;
		m_xLock.Unlock();
	}
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::ESCallback
// Description : Elementary stream callback
//
// Parameters  : 
//   CellDmuxHandle handle            - 
//   CellDmuxEsHandle esHandle        - 
//   const CellDmuxEsMsg* pDmuxEsMsg  - 
//   void* cbArg                      - 
//
// Returns     : uint32_t - 
///////////////////////////////////////////////////////////////////////////////
uint32_t Asura_PS3_FMV_dmux::ESCallback(CellDmuxHandle handle, CellDmuxEsHandle esHandle, const CellDmuxEsMsg* pDmuxEsMsg, void* cbArg)
{
	(void)handle;
	(void)esHandle;
	Asura_PS3_FMV_dmux* pThis = reinterpret_cast<Asura_PS3_FMV_dmux*>(cbArg);
	
	switch (pDmuxEsMsg->msgType)
	{
		case CELL_DMUX_ES_MSG_TYPE_AU_FOUND:
		case CELL_DMUX_ES_MSG_TYPE_FLUSH_DONE:
		{
			Asura_PS3_FMV_ESInfo* pESInfo = pThis->m_axESInfo;
			for (u_int uLoop=0; uLoop < pThis->m_uESCount; pESInfo++, uLoop++)
			{
				if ( pESInfo->xHandle == esHandle )
				{
					if ( pESInfo->Lock() )
					{
						pESInfo->iDMUXFoundAUCount++;
						pESInfo->Unlock();
					}
					break;
				}
			}
		}
		break;

		default:
			Asura_Assertf(false, ("[Asura_PS3_FMV_adec::DMuxESCallback] - Callback type unknown %d\n", pDmuxEsMsg->msgType));
			pThis->m_pControl->Abort(RET_CODE_ERR_FATAL);
		break;
	}
	
	return 0;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_dmux::Callback
// Description : DMuxer data block callback
//
// Parameters  : 
//   CellDmuxHandle handle        - DMuxer handle
//   const CellDmuxMsg* pDmuxMsg  - Dmux message info
//   void* callbackArg            - User specific argument (this pointer)
//
// Returns     : uint32_t - Result
///////////////////////////////////////////////////////////////////////////////
uint32_t Asura_PS3_FMV_dmux::Callback(CellDmuxHandle handle, const CellDmuxMsg* pDmuxMsg, void* callbackArg)
{
	(void)handle;
	Asura_PS3_FMV_dmux* pThis = reinterpret_cast<Asura_PS3_FMV_dmux*>(callbackArg);

	switch (pDmuxMsg->msgType)
	{
		case CELL_DMUX_MSG_TYPE_DEMUX_DONE:
			//Asura_OutputToDebugger(("DMUX done block %d\n", m_iRunningFrameCount++));
			if ( pThis->m_xLock.Lock(0) )
			{
				pThis->m_iProcessedCount++;
				pThis->m_bDmuxBusy = false;
				pThis->m_xLock.Unlock();
			}
		break;
		
		case CELL_DMUX_MSG_TYPE_FATAL_ERR:
			pThis->m_pControl->Abort(RET_CODE_ERR_FATAL);
		break;
		
		default:
			Asura_OutputToDebugger(("Callback type unknown %u\n", pDmuxMsg->msgType));
			pThis->m_pControl->Abort(RET_CODE_ERR_FATAL);
		break;
	}

	return 0;
}





