
//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_VideoToTexture.cpp
// Author    : Jason Turner
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM

//
// Asura independent includes
//
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FMV/Asura_VideoToTexture.h"
#include "Rendering/Asura_Viewport.h"

//
// Asura platform includes
//
#include "FileAccess/Asura_PS3_FileSystem.h"

#ifdef	SPURS_VIDEO
	#include "PPU_SPU_Utils/Asura_PS3_SPURS.h"
#endif

#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_QuadOverlayRenderer.h"
#include "Rendering/Asura_PS3_TextureManagement.h"
#include "Rendering/Asura_PS3_VideoMemoryManager.h"
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Sound/Asura_PS3_Multistream_Audio.h"

//
// Shaders definition
//
#include "Rendering/ShadersDefinition/FP/YUV_to_RGB_FP.h"

//
// PS3 includes
//
#include <cell/mstream.h>
#include <cell/sail.h>
#include <sys/fs_external.h>
#include <sys/paths.h>
#include <sys/sys_time.h>
#include <sysutil/sysutil_sysparam.h>

//
// Local defines
//

// Enables YUV->RGB colour space conversion in the fragment shader
#define ENABLE_SHADER_CSC
//#define ENABLE_DEBUG_LOG_RENDER_TIME
//#define ENABLE_DEBUG_LOG_MEM_ALLOCS
//#define ENABLE_AUDIO_VIDEO_SYNC

#define PROTOCOL_CELL_FS		"x-cell-fs://"
#define MAX_URI_LENGTH (CELL_FS_MAX_FS_PATH_LENGTH + sizeof(PROTOCOL_CELL_FS))
#define INVALID_PTS				(0xFFFFFFFFFFFFFFFFULL)

#define CAST_PTR_TO_UINT64(x)	((uint64_t)(intptr_t)(x))
#define CAST_UINT64_TO_PTR(t, x) ((t)(intptr_t)(x))
#define ROUNDUP(x, a)			(((x)+((a)-1))&(~((a)-1)))

//	#define g_szSpursNamePrefix "libsailPlr-"	// max 13 chars
static const char g_szSpursNamePrefix[] = "LibSailPlr-";	// max 13 chars

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data
//
////////////////////////////////////////////////////////////////////////////////

class Asura_VideoToTexture_PS3Data
{
public:
	friend class Asura_VideoToTexture;

	Asura_VideoToTexture_PS3Data();
	~Asura_VideoToTexture_PS3Data();

private:
	static void OnPlayerEvent(void* pArg, CellSailEvent xEvent, uint64_t ulArg0, uint64_t ulArg1);
	static void* OnMemAlloc(Asura_VideoToTexture *pxSelf, size_t uBoundary, size_t uSize);
	static void OnMemFree(Asura_VideoToTexture *pxSelf, size_t uBoundary, void *pMemory);

	static int OnSoundAdapterMakeup(Asura_VideoToTexture_PS3Data *pxArg);
	static int OnSoundAdapterCleanup(Asura_VideoToTexture_PS3Data *pxArg);
	static int OnSoundAdapterFormatChanged(Asura_VideoToTexture_PS3Data *pxArg, CellSailAudioFormat *pxFormat, uint32_t uCount);

	static int OnGraphicsAdapterMakeup(Asura_VideoToTexture *pxArg);
	static int OnGraphicsAdapterCleanup(Asura_VideoToTexture *pxArg);
	static int OnGraphicsAdapterFormatChanged(Asura_VideoToTexture *pxArg, CellSailVideoFormat *pxFormat, uint32_t uCount);
	static int OnGraphicsAdapterAllocFrame(Asura_VideoToTexture_PS3Data *pxArg, size_t uSize, int iNum, uint8_t **ppucFrame);
	static int OnGraphicsAdapterFreeFrame(Asura_VideoToTexture_PS3Data *pxArg, int iNum, uint8_t **ppucFrame);
	
#ifndef ASURA_NO_DEBUG_OUTPUT
	static void OutputEvent(CellSailEvent xEvent, uint64_t ulArg0, uint64_t ulArg1);
#endif	// ASURA_NO_DEBUG_OUTPUT

	static bool AudioVoiceDataCallback( char* &pcData, u_int &uDataSize, void *pArg );
	bool OnAudioDataNeeded( char* &pcData, u_int &uDataSize);

	void CreateThreadEventQueue();
	int GetPlayerState()
	{
		return m_iPlayerState;
	}
	void Render();
	void Reset();
	bool InitialiseRenderTarget(const int iTextureRTIndex);
	void SwapInPlaceholderTexture();
	void SwapOutPlaceholderTexture();

	Asura_Handle CreateVoiceForFMV();
	bool PlayVoiceForFMV();

	static const u_int		s_uMaxSamplesPerPlayerRead = 2048;
	static const u_int		s_uTotalYUVPlanes = 3;

	// size of event queue to receive from CellSailPlayer 
	static const u_int		s_uEventQueueSize = 32;

	//
	// Spurs parameters
	//
//	static const u_int s_uSpursNumOfSpus = 4;	// this causes a hang
	static const u_int		s_uSpursNumOfSpus = 2;
	static const u_int		s_uSpursSpuPriority = 100;
	static const u_int		s_uSpursPpuPriority = 300;
	static const bool		s_bSpursExitIfNoWork = true;

	// Total internal video frames the player allocs - must be >=3, <=6.
	// (the lower the number, the less it can use for buffering and the more choppy the playback).
	static const u_int		s_uTotalPlayerVideoFrames = 4;

	char					m_szFilename[ASURA_PS3_MAX_PATH];
	char					m_szURI[MAX_URI_LENGTH];
	CellSailDescriptor*		m_pxDescriptor;
	CellSailMemAllocator	m_xAllocator;
	CellSailPlayer			m_xPlayer;
	CellSailSoundAdapter	m_xSoundAdapter;
	CellSailGraphicsAdapter	m_xGraphicsAdapter;
	CellSailAviMovie		m_xAviMovieDatabase;

	//
	// PS3 rendering
	//
	uint8_t*				m_pcVideoFrames[s_uTotalPlayerVideoFrames];
	uint8_t*				m_pcVideoTextureTargetData;
	Asura_PS3_Texture*		m_pxVideoTextureTarget;
	CellGcmTexture			m_xPlaceholderGcmTextureData;
	void*					m_pPlaceholderTextureData;
	bool					m_bRenderingPlaceholderTexture;
	Asura_PS3_Texture		m_axYUVTextures[s_uTotalYUVPlanes];
	Asura_PS3_FragmentShader* m_pxFragmentShader;
	Asura_PS3_YUV_to_RGB_FP_Handles* m_pxFragmentShaderHandles;
	Asura_Overlay_Quad		m_xVideoQuad;
	Asura_PS3_FrameBuffer*	m_pxRenderTarget;

#ifdef ENABLE_AUDIO_VIDEO_SYNC
	#define VIDEO_TAG_FEEDBACK_DELAY  (2)

	// Needs to be bigger than VIDEO_TAG_FEEDBACK_DELAY
	#define VIDEO_TAG_NUMBER   (4)

	//
	// Video tags
	//
	uint32_t  				maVideoTags[VIDEO_TAG_NUMBER];
	int       				mRenderTagIndex;
	int       				mUpdateTagIndex;

	//
	// Sound tags
	//
	// bigger than 4
	#define AUDIO_TAG_FEEDBACK_DELAY  (5)

	// bigger than AUDIO_TAG_FEEDBACK_DELAY
	#define AUDIO_TAG_NUMBER  (7)

	uint32_t				mPortNo;
	bool					maUsed[AUDIO_TAG_NUMBER];
	uint64_t				maMixerTag[AUDIO_TAG_NUMBER];
	uint32_t				maSyncTag[AUDIO_TAG_NUMBER];

#endif

	volatile bool			m_bPlayerInited;
	int						m_iPlayerState;

	// PS3 event queue for passing data between threads
	sys_event_queue_t		m_xPlayerEventQueue;
	sys_event_port_t		m_xPlayerEventPort;
	uint64_t				m_ulAudioTime;
	uint64_t				m_ulVideoTime;
	uint32_t            	m_uAudioSessionId;
	uint32_t            	m_uVideoSessionId;

	uint32_t				m_uVideoWidth;
	uint32_t				m_uVideoHeight;
	uint32_t				m_uVideoBytesPerColour;		// for ARGB this is 4
	uint32_t				m_uMSecPerFrame;
	uint64_t				m_ulLastPTS;				// Presentation Time Stamp from last video frame
	uint64_t				m_ulMovieDurationMsec;

	CellSailAudioFormat 	m_xAudioFormat;
	CellSailVideoFormat 	m_xVideoFormat;
	bool					m_bPlayingAudio;
	u_int					m_uLastRenderTime;

#ifndef	SPURS_VIDEO
	CellSpurs*				m_pxSpurs;// __attribute__((aligned(128)));
#endif
	sys_memory_container_t	m_xContainer;

	Asura_VideoToTexture*	m_pxParentObject;		// pointer to parent Asura_VideoToTexture object
};


////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::Asura_VideoToTexture_PS3Data()
//
////////////////////////////////////////////////////////////////////////////////

Asura_VideoToTexture_PS3Data::Asura_VideoToTexture_PS3Data()
{
	m_pxDescriptor = NULL;
	m_bPlayerInited = false;
	m_iPlayerState = 0;
	m_uVideoWidth = 0;
	m_uVideoHeight = 0;
	m_uVideoBytesPerColour = 0;
	m_uMSecPerFrame = 0;
	m_ulLastPTS = 0;
	m_ulMovieDurationMsec = 0;
	m_pcVideoTextureTargetData = NULL;
	m_pxVideoTextureTarget = NULL;
	m_pPlaceholderTextureData = NULL;
	m_pxParentObject = NULL;

#ifndef	SPURS_VIDEO
	m_pxSpurs = (CellSpurs *) Asura_AllocateAlign(sizeof(CellSpurs), ASURA_ALIGN_128_BYTE);
#endif

	m_pxFragmentShader = NULL;
	m_pxFragmentShaderHandles = NULL;
	m_pxRenderTarget = NULL;
	m_ulAudioTime = 0;
	m_ulVideoTime = 0;

	m_xPlayerEventQueue = 0;
	m_xPlayerEventPort = 0;
	m_bRenderingPlaceholderTexture = true;
	m_bPlayingAudio = false;
	m_uLastRenderTime = 0;
};

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::~Asura_VideoToTexture_PS3Data()
//
////////////////////////////////////////////////////////////////////////////////

Asura_VideoToTexture_PS3Data::~Asura_VideoToTexture_PS3Data()
{
	Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::Reset()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::Reset()
{
	int iRet;

	m_bPlayerInited = false;

	// wait for finish of call
	// - cellSailSoundAdapterGetFrame()
	//(- cellSailGraphicsAdapterGetFrame())
	sys_timer_usleep(1000*10); // 10msec

	// release all resource including descriptor
	
	cellSailPlayerFinalize(&m_xPlayer);
	cellSailSoundAdapterFinalize(&m_xSoundAdapter);
	cellSailGraphicsAdapterFinalize(&m_xGraphicsAdapter);

#ifndef	SPURS_VIDEO
	if (m_pxSpurs)
	{
		iRet = cellSpursFinalize(m_pxSpurs);
		Asura_OutputToDebuggerAssert(iRet >= CELL_OK, ("cellSpursFinalize() failed: 0x%X\n", iRet));
	}
#endif

	iRet = sys_memory_container_destroy(m_xContainer);
	Asura_OutputToDebuggerAssert(iRet >= CELL_OK, ("sys_memory_container_destroy() failed: 0x%X\n", iRet));

#ifndef	SPURS_VIDEO
	if (m_pxSpurs)
	{
		Asura_DeallocateAlign(m_pxSpurs);
		m_pxSpurs = NULL;
	}
#endif

	iRet = sys_event_port_disconnect(m_xPlayerEventPort);
	Asura_OutputToDebuggerAssert(iRet >= CELL_OK, ("sys_event_port_disconnect() failed: 0x%X\n", iRet));

	iRet = sys_event_port_destroy(m_xPlayerEventPort);
	Asura_OutputToDebuggerAssert(iRet >= CELL_OK, ("sys_event_port_destroy() failed: 0x%X\n", iRet));

	iRet = sys_event_queue_destroy(m_xPlayerEventQueue, SYS_EVENT_QUEUE_DESTROY_FORCE);
	Asura_OutputToDebuggerAssert(iRet >= CELL_OK, ("sys_event_queue_destroy() failed: 0x%X\n", iRet));

	if (!m_bRenderingPlaceholderTexture)
	{
		SwapInPlaceholderTexture();
	}

	// wait until renderer has finished with this?
	if (m_pcVideoTextureTargetData)
	{
		Asura_DeallocateAlign(m_pcVideoTextureTargetData);
		m_pcVideoTextureTargetData = NULL;
	}

	if (m_pxFragmentShader)
	{
		delete m_pxFragmentShader;
		m_pxFragmentShader = NULL;
	}

	if (m_pxFragmentShaderHandles)
	{
		delete m_pxFragmentShaderHandles;
		m_pxFragmentShaderHandles = NULL;
	}

	if (m_pxRenderTarget)
	{
		delete m_pxRenderTarget;
		m_pxRenderTarget = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Initialise()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Initialise() 
{
	// create and initialise PS3 data
	Asura_VideoToTexture_PS3Data* pxPs3Data = new Asura_VideoToTexture_PS3Data();
	Asura_Assert(pxPs3Data, "Failed to allocate Asura_VideoToTexture_PS3Data. Out of memory?");

	pxPs3Data->m_pxParentObject = this;

	int iRet;

	m_pPlatformData = pxPs3Data;

	// Create PS3 thread queue
	pxPs3Data->CreateThreadEventQueue();

#ifndef	SPURS_VIDEO
	//
	// Init spurs
	//
	{
		// memory container for spurs
		iRet = sys_memory_container_create(&pxPs3Data->m_xContainer, Asura_VideoToTexture_PS3Data::s_uSpursNumOfSpus * 512 * 1024);
		if (0 > iRet)
		{
			Asura_Assertf(iRet >= CELL_OK, ("sys_memory_container_create() failed: 0x%X\n", iRet));
			return false;
		}

		CellSpursAttribute xAttr;
		iRet = cellSpursAttributeInitialize(&xAttr,
			Asura_VideoToTexture_PS3Data::s_uSpursNumOfSpus,
			Asura_VideoToTexture_PS3Data::s_uSpursSpuPriority,
			Asura_VideoToTexture_PS3Data::s_uSpursPpuPriority,
			Asura_VideoToTexture_PS3Data::s_bSpursExitIfNoWork);
		if (0 > iRet)
		{
			Asura_Assertf(iRet >= CELL_OK, ("cellSpursAttributeInitialize() failed: 0x%X\n", iRet));
			return false;
		}

		iRet = cellSpursAttributeSetNamePrefix(&xAttr, g_szSpursNamePrefix, strlen(g_szSpursNamePrefix) - 1);
		if (0 > iRet)
		{
			Asura_Assertf(iRet >= CELL_OK, ("cellSpursAttributeSetNamePrefix() failed: 0x%X\n", iRet));
			return false;
		}

		iRet = cellSpursAttributeSetMemoryContainerForSpuThread(&xAttr, pxPs3Data->m_xContainer);
		if (0 > iRet)
		{
			Asura_Assertf(iRet >= CELL_OK, ("cellSpursAttributeSetMemoryContainerForSpuThread() failed: 0x%X\n", iRet));
			return false;
		}

		iRet = cellSpursInitializeWithAttribute(pxPs3Data->m_pxSpurs, &xAttr);
		if (0 > iRet)
		{
			Asura_Assertf(iRet >= CELL_OK, ("cellSpursInitializeWithAttribute() failed: 0x%X\n", iRet));
			return false;
		}
	}
#endif

	CellSailPlayerResource xPlayerResource;
	Asura_Mem::Set(&xPlayerResource, 0, sizeof(xPlayerResource));

#ifdef SPURS_VIDEO
	xPlayerResource.pSpurs = Asura_PS3_SPURS::GetSPURSInstance();
#else
	xPlayerResource.pSpurs = pxPs3Data->m_pxSpurs;
#endif


	//
	// Init Player
	//
	CellSailPlayerAttribute xAttr;
	Asura_Mem::Set(&xAttr, 0, sizeof(xAttr));
#ifdef ENABLE_AUDIO_VIDEO_SYNC
//	xAttr.playerPreset      = CELL_SAIL_PLAYER_PRESET_AV_SYNC_AUTO_DETECT;
	xAttr.playerPreset      = CELL_SAIL_PLAYER_PRESET_AV_SYNC_59_94HZ;
#else
	xAttr.playerPreset      = CELL_SAIL_PLAYER_PRESET_AS_IS;
#endif

	xAttr.maxAudioStreamNum = 1;
//	xAttr.maxAudioStreamNum = m_eFlags & ASURA_FMV_FLAG_NO_SOUND ? 0 : 1;
	xAttr.maxVideoStreamNum = 1;
	xAttr.maxUserStreamNum  = 0;
	xAttr.queueDepth        = 0;

	// mem allocator
	CellSailMemAllocatorFuncs xMemFuncs =
	{
		(CellSailMemAllocatorFuncAlloc) Asura_VideoToTexture_PS3Data::OnMemAlloc,
		(CellSailMemAllocatorFuncFree) Asura_VideoToTexture_PS3Data::OnMemFree,
	};

	iRet = cellSailMemAllocatorInitialize(&pxPs3Data->m_xAllocator, &xMemFuncs, this);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailMemAllocatorInitialize() failed: 0x%X\n", iRet));

	// player
	iRet = cellSailPlayerInitialize2(
		&pxPs3Data->m_xPlayer,
		&pxPs3Data->m_xAllocator,					// mem allocator
		Asura_VideoToTexture_PS3Data::OnPlayerEvent,// callback
		m_pPlatformData,							// callback arg
		&xAttr,
		&xPlayerResource);							// the SPURS instance

	Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerInitialize() failed: 0x%X\n", iRet));

	//
	// Init sound adapter
	//
	static const CellSailSoundAdapterFuncs xSoundFuncs =
	{
		(CellSailSoundAdapterFuncMakeup)       Asura_VideoToTexture_PS3Data::OnSoundAdapterMakeup,
		(CellSailSoundAdapterFuncCleanup)      Asura_VideoToTexture_PS3Data::OnSoundAdapterCleanup,
		(CellSailSoundAdapterFuncFormatChanged)Asura_VideoToTexture_PS3Data::OnSoundAdapterFormatChanged,
	};

	iRet = cellSailSoundAdapterInitialize(&pxPs3Data->m_xSoundAdapter, &xSoundFuncs, pxPs3Data);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailSoundAdapterInit() failed: 0x%X\n", iRet));

	iRet = cellSailPlayerSetSoundAdapter(&pxPs3Data->m_xPlayer, 0, &pxPs3Data->m_xSoundAdapter);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerSetSoundAdapter() failed: 0x%X\n", iRet));

	#ifdef ENABLE_AUDIO_VIDEO_SYNC
	pxPs3Data->mRenderTagIndex = 0;
	pxPs3Data->mUpdateTagIndex = -VIDEO_TAG_FEEDBACK_DELAY;
	Asura_Assert(pxPs3Data->mUpdateTagIndex % (VIDEO_TAG_FEEDBACK_DELAY+1) == -VIDEO_TAG_FEEDBACK_DELAY, "");
	#endif

	//
	// Init graphics adapter
	//
	static const CellSailGraphicsAdapterFuncs xGraphicsFuncs =
	{
		(CellSailGraphicsAdapterFuncMakeup)       Asura_VideoToTexture_PS3Data::OnGraphicsAdapterMakeup,
		(CellSailGraphicsAdapterFuncCleanup)      Asura_VideoToTexture_PS3Data::OnGraphicsAdapterCleanup,
		(CellSailGraphicsAdapterFuncFormatChanged)Asura_VideoToTexture_PS3Data::OnGraphicsAdapterFormatChanged,
		(CellSailGraphicsAdapterFuncAllocFrame)   Asura_VideoToTexture_PS3Data::OnGraphicsAdapterAllocFrame,
		(CellSailGraphicsAdapterFuncFreeFrame)    Asura_VideoToTexture_PS3Data::OnGraphicsAdapterFreeFrame,
	};
	iRet = cellSailGraphicsAdapterInitialize(&pxPs3Data->m_xGraphicsAdapter, &xGraphicsFuncs, pxPs3Data);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailGraphicsAdapterInit() failed: 0x%X\n", iRet));

	//
	// Set the output format you want
	//
	CellSailVideoFormat xVideoFormat;
	Asura_Mem::Set(&xVideoFormat, 0xFF, sizeof(xVideoFormat));

#ifdef ENABLE_SHADER_CSC
	xVideoFormat.coding = CELL_SAIL_VIDEO_CODING_YUV420_PLANAR;;
#else
	xVideoFormat.coding = CELL_SAIL_VIDEO_CODING_ARGB_INTERLEAVED;;
#endif

	iRet = cellSailGraphicsAdapterSetPreferredFormat(&pxPs3Data->m_xGraphicsAdapter, &xVideoFormat);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailGraphicsAdapterSetPreferredFormat() failed: 0x%X\n", iRet));

	iRet = cellSailPlayerSetGraphicsAdapter(&pxPs3Data->m_xPlayer, 0, &pxPs3Data->m_xGraphicsAdapter);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerSetGraphicsAdapter() failed: 0x%X\n", iRet));

	//
	// Boot the player 
	//
	{
		iRet = cellSailPlayerBoot(&pxPs3Data->m_xPlayer, CAST_PTR_TO_UINT64(m_pPlatformData));
		Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerBoot() failed: 0x%X\n", iRet));

		Asura_OutputToDebugger(("cellSailPlayerBoot() completed\n"));
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::InitialiseRenderTarget
//
// iTextureRTIndex = texture index of render target
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture_PS3Data::InitialiseRenderTarget(const int iTextureRTIndex)
{
#ifdef ENABLE_SHADER_CSC
	m_pxFragmentShaderHandles = new Asura_PS3_YUV_to_RGB_FP_Handles();
	Asura_Assert(m_pxFragmentShaderHandles, "Unable to allocate Asura_PS3_YUV_to_RGB_FP_Handles - out of memory?");

	m_pxFragmentShader = new Asura_PS3_FragmentShader( (unsigned char *)&_binary_PS3_YUV_to_RGB_FP_bin_start, (unsigned char *)&_binary_PS3_YUV_to_RGB_FP_bin_end, m_pxFragmentShaderHandles );
	Asura_Assert(m_pxFragmentShader, "Unable to allocate m_pxFragmentShader - out of memory?");

	//
	// Setup YUV textures
	//
	for (unsigned int i = 0; i < s_uTotalYUVPlanes; ++i)
	{
		CellGcmTexture xGCMTexture;

		xGCMTexture.format    = CELL_GCM_TEXTURE_B8 | CELL_GCM_TEXTURE_LN | CELL_GCM_TEXTURE_NR;
		xGCMTexture.mipmap    = 1;
		xGCMTexture.dimension = CELL_GCM_TEXTURE_DIMENSION_2;
		xGCMTexture.cubemap   = CELL_GCM_FALSE;
		xGCMTexture.remap =
			CELL_GCM_TEXTURE_REMAP_REMAP  << 14 | // b
			CELL_GCM_TEXTURE_REMAP_REMAP  << 12 | // g = b
			CELL_GCM_TEXTURE_REMAP_REMAP  << 10 | // r = b
			CELL_GCM_TEXTURE_REMAP_REMAP  <<  8 | // a = 0xFF
			CELL_GCM_TEXTURE_REMAP_FROM_R <<  6 |
			CELL_GCM_TEXTURE_REMAP_FROM_G <<  4 |
			CELL_GCM_TEXTURE_REMAP_FROM_B <<  2 |
			CELL_GCM_TEXTURE_REMAP_FROM_A;
		
		// Y plane is normal size, U/V are half width and height
		xGCMTexture.width  = (0 == i ? m_uVideoWidth : m_uVideoWidth / 2);
		xGCMTexture.height = (0 == i ? m_uVideoHeight : m_uVideoHeight / 2);
		xGCMTexture.depth  = 1;
		xGCMTexture.location = CELL_GCM_LOCATION_MAIN;

		m_axYUVTextures[i] = Asura_PS3_Texture( &xGCMTexture );
	}

	//
	// Setup quad renderer
	//
	// if we're doing full screen rendering...
	if (iTextureRTIndex == ASURA_TEXTURE_NONE)
	{
		m_xVideoQuad.SetRectXY(-2, -2, Asura_Viewport::GetWidth()+4, Asura_Viewport::GetHeight()+4);
	}
	else
	{
		m_xVideoQuad.SetRectXY(-2, -2, m_uVideoWidth+4, m_uVideoHeight+4);
	}

	m_xVideoQuad.SetRectUV( 0.0f, 0.0f, 1.0f, 1.0f );
	m_xVideoQuad.TranslucencyMode =	ASURA_TRANSLUCENCY_OFF;
	m_xVideoQuad.Colour = 0xFFFFFFFF;
	m_xVideoQuad.TextureIndex = ASURA_TEXTURE_NONE;

	// if we're not doing full screen rendering...
	if (iTextureRTIndex != ASURA_TEXTURE_NONE)
	{
		m_pxRenderTarget = new Asura_PS3_FrameBuffer( false, m_uVideoWidth, m_uVideoHeight, 1, ASURA_PS3_FRAMEBUFFER_RGBA, false, false, "VideoToTexture RT", -1);
		Asura_Assert(m_pxRenderTarget, "Unable to allocate Asura_VideoToTexture::Asura_PS3_FrameBuffer - out of memory?");

		Asura_VideoToTexture::s_uTotalVidMemUsed += m_uVideoWidth * m_uVideoHeight * m_uVideoBytesPerColour;
	}

#ifdef ENABLE_DEBUG_LOG_MEM_ALLOCS
	Asura_OutputToDebugger_Retail(("Asura_VideoToTexture_PS3Data:VidAlloc of %dB [Total:%dB, 0x%X]\n", m_uVideoWidth * m_uVideoHeight * m_uVideoBytesPerColour, Asura_VideoToTexture::s_uTotalVidMemUsed, m_pxRenderTarget));
#endif

#endif	// ENABLE_SHADER_CSC

	// if we're NOT doing full screen rendering...
	if (iTextureRTIndex != ASURA_TEXTURE_NONE)
	{
		m_pxVideoTextureTarget = Asura_PS3_TextureManagement::GetTexture(iTextureRTIndex);

		// save current texture data
		m_xPlaceholderGcmTextureData = *m_pxVideoTextureTarget->GetTexture();
		m_pPlaceholderTextureData = m_pxVideoTextureTarget->GetTextureData();

		// Allocate video frame texture memory
		m_pcVideoTextureTargetData = (uint8_t *) ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_uVideoWidth * m_uVideoHeight * m_uVideoBytesPerColour, ASURA_ALIGN_128_BYTE);
		Asura_VideoToTexture::s_uTotalMemUsed += m_uVideoWidth * m_uVideoHeight * m_uVideoBytesPerColour;

	#ifdef ENABLE_DEBUG_LOG_MEM_ALLOCS
		Asura_OutputToDebugger_Retail(("Asura_VideoToTexture_PS3Data:Alloc of %dB [Total:%dB, 0x%X]\n", m_uVideoWidth * m_uVideoHeight * m_uVideoBytesPerColour, Asura_VideoToTexture::s_uTotalMemUsed, m_pcVideoTextureTargetData));
	#endif
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::CreateThreadEventQueue()
//
// The event queue is used to pass data FROM the player callback thread TO the main thread
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::CreateThreadEventQueue()
{
	// message queue
	sys_event_queue_attribute_t xQueueAttr;

	sys_event_queue_attribute_initialize(xQueueAttr);
	int iRet = sys_event_queue_create(&m_xPlayerEventQueue, &xQueueAttr, SYS_EVENT_QUEUE_LOCAL, s_uEventQueueSize);
	Asura_Assertf(iRet >= CELL_OK, ("sys_event_queue_create() failed: 0x%X\n", iRet));

	iRet = sys_event_port_create(&m_xPlayerEventPort, SYS_EVENT_PORT_LOCAL, SYS_EVENT_PORT_NO_NAME);
	Asura_Assertf(iRet >= CELL_OK, ("sys_event_port_create() failed: 0x%X\n", iRet));

	iRet = sys_event_port_connect_local(m_xPlayerEventPort, m_xPlayerEventQueue);
	if (iRet < CELL_OK)
	{
		Asura_Assertf(iRet >= CELL_OK, ("sys_event_port_connect_local() failed: 0x%X\n", iRet));

		sys_event_port_destroy(m_xPlayerEventPort);
		(void) sys_event_queue_destroy(m_xPlayerEventQueue, SYS_EVENT_QUEUE_DESTROY_FORCE);
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::CreateVoiceForFMV()
//
// Creates an audio voice for us to throw PCM data at
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_VideoToTexture_PS3Data::CreateVoiceForFMV()
{
	Asura_Handle iVoiceHandle = ASURA_INVALID_HANDLE;

	int iRet = cellSailSoundAdapterGetFormat(&m_xSoundAdapter, &m_xAudioFormat);
	if (iRet >= 0)
	{
		Asura_Assertf(m_xAudioFormat.chNum <= 2, ("Only mono/stereo sounds supported. This audio stream has %d channels", m_xAudioFormat.chNum));
		Asura_Assert(m_xAudioFormat.coding == CELL_SAIL_AUDIO_CODING_LPCM_FLOAT32, "Audio stream not Linear PCM data");

		if (m_xAudioFormat.coding == CELL_SAIL_AUDIO_CODING_LPCM_FLOAT32)
		{
			if ((m_xAudioFormat.chNum > 0) && (m_xAudioFormat.chNum <= 2))
			{
				iVoiceHandle = Asura_MultiStream_Audio::CreateCustomStreamingVoice();
				Asura_MultiStream_Custom_Streaming_Voice* pxVoice = reinterpret_cast<Asura_MultiStream_Custom_Streaming_Voice*>( Asura_MultiStream_Audio::GetVoice(iVoiceHandle) );
				if( pxVoice )
				{
					pxVoice->Prepare(AudioVoiceDataCallback, CELL_MS_32BIT_FLOAT, m_xAudioFormat.fs, m_xAudioFormat.chNum, this);
				}
			}
		}
	}
	else
	{
		Asura_Assertf(false, ("Unable to get FMV audio stream info. Error [0x%X]", iRet));
	}

	return iVoiceHandle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::PlayVoiceForFMV()
//
// Starts playing our voice
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture_PS3Data::PlayVoiceForFMV()
{
	Asura_Handle iVoiceHandle = m_pxParentObject->GetSourceVoiceHandle();
	if (iVoiceHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Audio_Voice* pxVoice = Asura_Audio_System::GetVoice(iVoiceHandle);
		Asura_Assertf (pxVoice, ("Unable to get Voice Handle [%d]", iVoiceHandle));

		pxVoice->Start();
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::AudioVoiceDataCallback
//
// Called when the voice playback needs more data
// pcData = [in] data pointer to copy into
// uDataSize = [in/out] size available to copy into/size we have copied
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture_PS3Data::AudioVoiceDataCallback( char* &pcData, u_int &uDataSize, void* pArg )
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(pArg);
	Asura_Assert(pxPS3Data, "")	;
	return (pxPS3Data->OnAudioDataNeeded(pcData, uDataSize));
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnAudioDataNeeded
//
// Class member function of AudioVoiceDataCallback().
// Grabs the current audio data from the player and gives it to our voice channel.
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture_PS3Data::OnAudioDataNeeded( char*& pcData, u_int& uDataSize)
{
	if (!m_bPlayerInited)
	{
		return true;
	}

	int iRet;
	Asura_Assert(uDataSize, "No space in voice!");
	const u_int uSampleSize = m_xAudioFormat.chNum * 4;
	u_int uSampleTotal = uDataSize / uSampleSize;

	if (uSampleTotal > s_uMaxSamplesPerPlayerRead)
	{
		uSampleTotal = s_uMaxSamplesPerPlayerRead;
	}

	CellSailSoundFrameInfo xFrameInfo;

	iRet = cellSailSoundAdapterGetFrame(&m_xSoundAdapter, uSampleTotal, &xFrameInfo);
	// >= 256
//	if (CELL_SAIL_SYNC_IS_FRAME_AVAILABLE(iRet))
	if ((iRet >= 0) && (xFrameInfo.pBuffer))
	{
		Asura_Assert(xFrameInfo.pBuffer, "");
		if (xFrameInfo.sessionId != m_uAudioSessionId)
		{
			cellSailSoundAdapterGetFormat(&m_xSoundAdapter, &m_xAudioFormat);
			Asura_OutputToDebugger(("* new audio format: sampleNum = %d\n", m_xAudioFormat.sampleNum));
			Asura_OutputToDebugger(("* new audio format: chNum     = %d\n", m_xAudioFormat.chNum));
			m_uAudioSessionId = xFrameInfo.sessionId;
		}

		uint64_t time = 0;
		iRet = cellSailSoundAdapterPtsToTimePosition(&m_xSoundAdapter, xFrameInfo.pts, &time);
		if (0 > iRet)
		{
			m_ulAudioTime = INVALID_PTS;
		} else
		{
			m_ulAudioTime = time;
		}

		const u_int uCopySize = uSampleTotal * uSampleSize;
		Asura_Assert(pcData, "");
		Asura_Mem::Copy(pcData, xFrameInfo.pBuffer, uCopySize);
		uDataSize = uCopySize;
	}

	system_time_t thisTime = sys_time_get_system_time();
	cellSailSoundAdapterUpdateAvSync(&m_xSoundAdapter, xFrameInfo.tag, thisTime);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Shutdown
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture::Platform_Shutdown()
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	delete pxPS3Data;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Update
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture::Platform_Update()
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	Asura_Assert(pxPS3Data, "");

	int iRet;
	sys_event_t xEvents[2];
	int iEventsReceived;

	// receive events from callback thread
	iRet = sys_event_queue_tryreceive(pxPS3Data->m_xPlayerEventQueue, xEvents, 1, &iEventsReceived);
	if (0 > iRet)
	{
		Asura_OutputToDebuggerAssert(0 > iRet, ("sys_event_queue_tryreceive() failed: 0x%X\n", iRet));
		return;
	}

	//
	// Process player message
	//
	for (int iEventNo = 0; iEventNo < iEventsReceived; ++iEventNo)
	{
		CellSailEvent xEvent;
		uint64_t      ulArg0;
		uint64_t      ulArg1;
		xEvent.u64.value = xEvents[iEventNo].data1;
		ulArg0            = xEvents[iEventNo].data2;
		ulArg1            = xEvents[iEventNo].data3;

#ifndef ASURA_NO_DEBUG_OUTPUT
		Asura_VideoToTexture_PS3Data::OutputEvent(xEvent, ulArg0, ulArg1);
#endif	// ASURA_NO_DEBUG_OUTPUT

		switch (xEvent.u32x2.major)
		{
			case CELL_SAIL_EVENT_ERROR_OCCURED:
			{
				// Unable to open file
				if ((int) ulArg0 == ENOENT)
				{
					Asura_Assertf(false, ("Unable to open FMV \"%s\"\n", pxPS3Data->m_szFilename));
					m_ePlayerState = PLAYER_STATE_ERROR;
				}
				else
				{
					Asura_Assertf(false, ("CELL_SAIL_EVENT_ERROR_OCCURED ret=0x%08X\n", (int) ulArg0));
				}
				return;
			}

			case CELL_SAIL_EVENT_PLAYER_CALL_COMPLETED:
				switch (xEvent.u32x2.minor)
				{
					case CELL_SAIL_PLAYER_CALL_BOOT:
					{
						m_ePlayerState = PLAYER_STATE_INITIALISED;
						pxPS3Data->m_bPlayerInited = true;
					}
					break;

					case CELL_SAIL_PLAYER_CALL_OPEN_STREAM:
					{
						pxPS3Data->m_ulAudioTime = 0;
						pxPS3Data->m_ulVideoTime = 0;

						if (m_ePlayerState != PLAYER_STATE_ERROR)
						{
							m_ePlayerState = PLAYER_STATE_FILE_OPENED;
						}

						break;
					}
					case CELL_SAIL_PLAYER_CALL_CLOSE_STREAM:
					{
						pxPS3Data->m_ulAudioTime = 0;
						pxPS3Data->m_ulVideoTime = 0;
						break;
					}
					default:
					{
						break;
					}
				}
				break;

			case CELL_SAIL_EVENT_PLAYER_STATE_CHANGED:
			{
				pxPS3Data->m_iPlayerState = ulArg0;
				break;
			}

			case CELL_SAIL_EVENT_STREAM_OPENED:
			{
				CellSailDescriptor* pxDesc = CAST_UINT64_TO_PTR(CellSailDescriptor*, ulArg0);
				int                 iStreamType;
				iStreamType = cellSailDescriptorGetStreamType(pxDesc);
				pxPS3Data->m_pxDescriptor = pxDesc;

				iRet = cellSailDescriptorCreateDatabase(pxDesc, &pxPS3Data->m_xAviMovieDatabase, sizeof(CellSailAviMovie), 0);
				if (0 > iRet)
				{
					Asura_OutputToDebuggerAssert(0 > iRet, ("cellSailDescriptorDestroyDatabase() failed: 0x%X\n", iRet));
					return;
				}

				switch (iStreamType)
				{
					case CELL_SAIL_STREAM_AVI:
					{
						CellSailAviMovieInfo		xAviInfo;
						CellSailAviMainHeader		xAviMainHdr;
						CellSailAviExtendedHeader	xAviExtHdr;

						iRet = cellSailAviMovieGetMovieInfo(&pxPS3Data->m_xAviMovieDatabase, &xAviInfo);
						Asura_Assertf(iRet >= 0, ("cellSailAviMovieGetMovieInfo() failed: 0x%X\n", iRet));

						pxPS3Data->m_ulLastPTS = ((uint64_t) xAviInfo.length - 1) * (uint64_t) xAviInfo.scale * 1000000 / xAviInfo.rate;
						pxPS3Data->m_ulMovieDurationMsec = (uint64_t)xAviInfo.length * xAviInfo.scale * 1000 / xAviInfo.rate;

						iRet = cellSailAviMovieGetHeader(&pxPS3Data->m_xAviMovieDatabase, &xAviMainHdr, &xAviExtHdr);
						Asura_Assertf(iRet >= 0, ("cellSailAviMovieGetHeader() failed: 0x%X\n", iRet));

						pxPS3Data->m_uVideoWidth = xAviMainHdr.width;
						pxPS3Data->m_uVideoHeight = xAviMainHdr.height;
						pxPS3Data->m_uVideoBytesPerColour = 4;
						pxPS3Data->m_uMSecPerFrame = (xAviMainHdr.microSecPerFrame/1000)-12;

						m_xVideoInfo.m_uWidth = xAviMainHdr.width;
						m_xVideoInfo.m_uHeight = xAviMainHdr.height;
						m_xVideoInfo.m_fDuration = static_cast<float>(pxPS3Data->m_ulMovieDurationMsec) / 1000.f;

						CellSailAviMediaType	xAviMediaType;
						CellSailAviStream		xAviStream;

						Asura_Mem::Set(&xAviMediaType, 0, sizeof(xAviMediaType));
						xAviMediaType.fccType             = CELL_SAIL_AVI_STREAM_VIDEO;
						xAviMediaType.fccHandler         = CELL_SAIL_AVI_HANDLER_UNSPECIFIED;
						iRet = cellSailAviMovieGetStreamByTypeAndIndex(&pxPS3Data->m_xAviMovieDatabase, &xAviMediaType, 0, &xAviStream);
						if (0 > iRet) {
							/* Error handling */ 
						}

						iRet = cellSailAviStreamGetMediaType(&xAviStream, &xAviMediaType);
						if ((xAviMediaType.fccHandler != CELL_SAIL_AVI_HANDLER_divx)
							&& (xAviMediaType.fccHandler != CELL_SAIL_AVI_HANDLER_DIVX))
						{
							Asura_Assertf(false,
								("Unsupported video compression: \"%c%c%c%c\"\n",
								xAviMediaType.fccHandler >> 24,
								(xAviMediaType.fccHandler >> 16) & 0xFF,
								(xAviMediaType.fccHandler >> 8) & 0xFF,
								xAviMediaType.fccHandler & 0xFF));
						}

						pxPS3Data->InitialiseRenderTarget(GetTextureRTIndex());
						break;
					}
					default:
					{
						Asura_Assertf(false, ("Unsupported stream type: %d\n", iStreamType));
						pxPS3Data->m_ulLastPTS = 0;
						pxPS3Data->m_ulMovieDurationMsec = 0;
						pxPS3Data->m_uVideoWidth = 0;
						pxPS3Data->m_uVideoHeight = 0;
						pxPS3Data->m_uMSecPerFrame = 0;
						pxPS3Data->m_uVideoBytesPerColour = 0;
						break;
					}
				}

				if (!(m_eFlags & ASURA_FMV_FLAG_NO_SOUND))
				{
					m_iSourceVoiceHandle = pxPS3Data->CreateVoiceForFMV();
				}

				break;
			}

			// end of current source stream (end of file)
			case CELL_SAIL_EVENT_SOURCE_EOS:
			{
				break;
			}
			case CELL_SAIL_EVENT_STREAM_CLOSED:
			{
				CellSailDescriptor *pxDesc = (CellSailDescriptor*)(uintptr_t) ulArg0;
				iRet = cellSailDescriptorDestroyDatabase(pxDesc, &pxPS3Data->m_xAviMovieDatabase);
				Asura_Assertf(iRet >= CELL_OK, ("cellSailDescriptorDestroyDatabase() failed: 0x%X\n", iRet));

				iRet = cellSailPlayerDestroyDescriptor(&pxPS3Data->m_xPlayer, pxDesc);
				Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerDestroyDescriptor() failed: 0x%X\n", iRet));

				pxPS3Data->m_pxDescriptor = NULL;
				pxPS3Data->m_ulLastPTS = 0;
				pxPS3Data->m_ulMovieDurationMsec = 0;
				pxPS3Data->m_uVideoWidth = 0;
				pxPS3Data->m_uVideoHeight = 0;
				pxPS3Data->m_uVideoBytesPerColour = 0;
				pxPS3Data->m_uMSecPerFrame = 0;

				break;
			}

			case CELL_SAIL_EVENT_PAUSE_STATE_CHANGED:
			{
				break;
			}

			default:
			{
				break;
			}
		}
	}

	if (pxPS3Data->m_iPlayerState == CELL_SAIL_PLAYER_STATE_RUNNING)
	{
		if (!pxPS3Data->m_bPlayingAudio)
		{
			pxPS3Data->PlayVoiceForFMV();
			pxPS3Data->m_bPlayingAudio = true;
		}
	}

	if (m_ePlayerState == PLAYER_STATE_PLAYING)
	{
		if ((pxPS3Data->m_ulVideoTime) >= pxPS3Data->m_ulLastPTS)
		{
			// if we're not looping and we're at the end of the video... then stop
			if (!(m_eFlags & ASURA_FMV_FLAG_LOOP))
			{
				m_ePlayerState = PLAYER_STATE_COMPLETE;
				return;
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Render
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture::Platform_Render()
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	Asura_Assert(pxPS3Data, "");

	if (pxPS3Data->m_iPlayerState == CELL_SAIL_PLAYER_STATE_RUNNING)
	{
		pxPS3Data->Render();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::Render
//
// Grabs the current video frame from the player and copies it to our render target
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::Render()
{
	int iRet;
	if (m_bPlayerInited)
	{
		u_int uThisTime = Asura_Timers::GetMSecsSinceStart_Continuous();
		if ((uThisTime - m_uLastRenderTime) < m_uMSecPerFrame)
			return;

		m_uLastRenderTime = Asura_Timers::GetMSecsSinceStart_Continuous();

#ifdef ENABLE_DEBUG_LOG_RENDER_TIME
		u_int uRenderStartTime = Asura_Timers::GetMSecsSinceStart_Continuous();
#endif

		CellSailGraphicsFrameInfo xFrameInfo;
		iRet = cellSailGraphicsAdapterGetFrame(&m_xGraphicsAdapter, &xFrameInfo);

		// >= 256
//		if (CELL_SAIL_SYNC_IS_FRAME_AVAILABLE(iRet))
		if ((iRet >= 0) && (xFrameInfo.pBuffer))
		{
			uint64_t luTime = 0;

			// Make sure we're rendering the texture containing the video frames
			if (m_bRenderingPlaceholderTexture)
			{
				SwapOutPlaceholderTexture();
			}

			// luTime is in micro seconds?
			iRet = cellSailGraphicsAdapterPtsToTimePosition(&m_xGraphicsAdapter, xFrameInfo.pts, &luTime);
			if (0 > iRet)
			{
				m_ulVideoTime = INVALID_PTS;
			}
			else
			{
				m_ulVideoTime = luTime;
			}

//			Asura_OutputToDebugger_Retail(("m_ulVideoTime: %d/%d\n", (int) (m_ulVideoTime/1000), (int) m_ulMovieDurationMsec));

			if (xFrameInfo.sessionId != m_uVideoSessionId)
			{
				(void) cellSailGraphicsAdapterGetFormat(&m_xGraphicsAdapter, &m_xVideoFormat);
				Asura_OutputToDebugger(("* new video format: width  = %d\n", m_xVideoFormat.width));
				Asura_OutputToDebugger(("* new video format: height = %d\n", m_xVideoFormat.height));
				m_uVideoSessionId = xFrameInfo.sessionId;
			}

#ifdef ENABLE_SHADER_CSC

			uint8_t *pBuf = xFrameInfo.pBuffer;
			for (unsigned int i = 0; i < s_uTotalYUVPlanes; i++)
			{
				uint32_t uTextureOffset;

				cellGcmAddressToOffset(pBuf, &uTextureOffset);
				m_axYUVTextures[i].GetTexture()->offset = uTextureOffset;
				m_axYUVTextures[i].SetTextureData(pBuf);
				uint32_t frameSize = m_xVideoFormat.pitch * m_xVideoFormat.height;
				if (i == 0)
				{
					m_axYUVTextures[i].GetTexture()->pitch = m_xVideoFormat.pitch;
				}
				else
				{
					m_axYUVTextures[i].GetTexture()->pitch = m_xVideoFormat.pitch / 2;
					frameSize /= 4;
				}
				pBuf += frameSize;
			}

			// Set render target
			if (m_pxParentObject->GetTextureRTIndex() == ASURA_TEXTURE_NONE)
			{
				// render full screen
				Asura_PS3_DeferredRendering::BindAsSingleRT();
			}
			else
			{
				m_pxRenderTarget->BindAsSingleRT();
			}

//			cellGcmSetClearColor(Asura_PS3_Graphics::GetGcmCurrentContext(), 0);
//			cellGcmSetClearSurface( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_CLEAR_R | CELL_GCM_CLEAR_G | CELL_GCM_CLEAR_B | CELL_GCM_CLEAR_A );

			// set 3 samplers.
			// Last param means we're giving it normalized texture coords (UV's 0->1)
			m_pxFragmentShader->SetTexture( Y_Sampler, &m_axYUVTextures[0], true);
			m_pxFragmentShader->SetTexture( U_Sampler, &m_axYUVTextures[1], true);
			m_pxFragmentShader->SetTexture( V_Sampler, &m_axYUVTextures[2], true);

			if (m_pxParentObject->GetTextureRTIndex() == ASURA_TEXTURE_NONE)
			{
				// rendering full screen, so use viewport size
				Asura_PS3_QuadOverlayRenderer::RenderCurrentRenderTarget( m_xVideoQuad, m_pxFragmentShader, true );
			}
			else
			{
				// use our render target size
				Asura_PS3_QuadOverlayRenderer::RenderCurrentRenderTarget( m_xVideoQuad, m_pxFragmentShader, true, m_xVideoFormat.width, m_xVideoFormat.height );
			}

			// Render the subtitles over the frame
			if (m_pxParentObject->GetSubtitleCallback())
			{
				m_pxParentObject->GetSubtitleCallback()(m_pxParentObject->GetPlayedTimeInSeconds());
			}

			// This will be NULL for full-screen rendering
			if (m_pxVideoTextureTarget)
			{
				// Resolve into our final texture
				m_pxRenderTarget->ResolveToTexture(m_pxVideoTextureTarget);
			}

			// restore old render target after call to BindAsSingleRT
			Asura_PS3_DeferredRendering::Bind();

#else
			// copy the video frame to our texture
			Asura_Mem::Copy(m_pcVideoTextureTargetData, xFrameInfo.pBuffer, m_uVideoWidth * m_uVideoHeight * m_uVideoBytesPerColour);
#endif
		}

#ifdef ENABLE_DEBUG_LOG_RENDER_TIME
		u_int uRenderTime = Asura_Timers::GetMSecsSinceStart_Continuous() - uRenderStartTime;
		Asura_OutputToDebugger_Retail(("FMV Render time: %dms\n", uRenderTime));
#endif

#ifdef ENABLE_AUDIO_VIDEO_SYNC
		// frame sync
		maVideoTags[mRenderTagIndex] = xFrameInfo.tag;
		mRenderTagIndex = (mRenderTagIndex + 1) % VIDEO_TAG_NUMBER;
		uint64_t timebase = cellGcmGetLastFlipTime();
		if (mUpdateTagIndex >= 0)
		{
			uint32_t tag = maVideoTags[mUpdateTagIndex];
			cellSailGraphicsAdapterUpdateAvSync(&m_xGraphicsAdapter, tag, timebase);
		}
		mUpdateTagIndex = (mUpdateTagIndex + 1) % VIDEO_TAG_NUMBER;
#endif
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Open
//
// Starts opening FMV
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Open(const char* szFilename, const int iTextureRTIndex, u_int uSoundTrackIndex)
{
	Asura_Char szFullPath[ASURA_PS3_MAX_PATH];

	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	Asura_Assert(pxPS3Data, "");
	Asura_Assert(pxPS3Data->m_bPlayerInited, "");

	Asura_PrintfToBfr(pxPS3Data->m_szFilename, "FMV/%s.avi", ASURA_PS3_MAX_PATH, 0, szFilename);
	m_xFileName = pxPS3Data->m_szFilename;
	Asura_PS3_FileSystem::GenerateFullFilename( pxPS3Data->m_szFilename, szFullPath, ASURA_PS3_MAX_PATH );
	Asura_PrintfToBfr(pxPS3Data->m_szURI, "%s%s", MAX_URI_LENGTH, 0, PROTOCOL_CELL_FS, szFullPath);

	Asura_OutputToDebugger(("uri = \"%s\"\n", pxPS3Data->m_szURI));

	int iRet;
	Asura_Assert(pxPS3Data->m_pxDescriptor == NULL, "stream already open\n");

	iRet = cellSailPlayerGetDescriptorCount(&pxPS3Data->m_xPlayer);
	Asura_OutputToDebugger(("added descriptor num = %d\n", iRet));
	if (0 == iRet)
	{
		CellSailDescriptor *pxDesc;
		iRet = cellSailPlayerCreateDescriptor(
			&pxPS3Data->m_xPlayer,
			CELL_SAIL_STREAM_AVI,	// stream type
			NULL,					// media info
			pxPS3Data->m_szURI,		// the URI of the file (this string is NOT copied internally - it stores the pointer, therefore must be on heap)
			&pxDesc);

		Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerCreateDescriptor() failed: 0x%X\n", iRet));

		iRet = cellSailPlayerAddDescriptor(&pxPS3Data->m_xPlayer, pxDesc);
		Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerAddDescriptor() failed: 0x%X\n", iRet));
	}

	iRet = cellSailPlayerOpenStream(&pxPS3Data->m_xPlayer, NULL);
	Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerOpenStream() failed: 0x%X\n", iRet));
	m_ePlayerState = PLAYER_STATE_FILE_OPENING;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Play
//
// Starts playing FMV - must be opened already.
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Play()
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	int iRet;

	if (GetPlayerState() == PLAYER_STATE_PAUSED)
	{
		cellSailPlayerSetPaused(&pxPS3Data->m_xPlayer, false);
		m_ePlayerState = PLAYER_STATE_PLAYING;
	}
	else if (CELL_SAIL_PLAYER_STATE_OPENED == pxPS3Data->m_iPlayerState)
	{
		Asura_Assert(pxPS3Data->m_pxDescriptor, "");

		// set video to repeat at the end
		if (m_eFlags & ASURA_FMV_FLAG_LOOP)
		{
			iRet = cellSailPlayerSetRepeatMode(&pxPS3Data->m_xPlayer, CELL_SAIL_REPEAT_LOOP, NULL);
			Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerSetRepeatMode() failed: 0x%X\n", iRet));
		}

		CellSailStartCommand xCommand;
		Asura_Mem::Set(&xCommand, 0, sizeof(xCommand));
		xCommand.startType = CELL_SAIL_START_NORMAL;

		iRet = cellSailPlayerStart(&pxPS3Data->m_xPlayer, &xCommand, CAST_PTR_TO_UINT64(pxPS3Data));
		Asura_Assertf(iRet >= CELL_OK, ("cellSailPlayerStart() failed: 0x%X\n", iRet));
		m_ePlayerState = PLAYER_STATE_PLAYING;
	}
	else
	{
		Asura_Assert(false, "can't start/pause/resume now!!\n");
		m_ePlayerState = PLAYER_STATE_FILE_OPENED;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Pause()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Pause()
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	bool bPaused = false;

	if (GetPlayerState() == PLAYER_STATE_PLAYING)
	{
		cellSailPlayerSetPaused(&pxPS3Data->m_xPlayer, true);
		m_ePlayerState = PLAYER_STATE_PAUSED;
		bPaused = true;
	}

	return bPaused;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_SetPlaySpeed()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_SetPlaySpeed(float fSpeed)
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);
	int iRet;

	if ((GetPlayerState() == PLAYER_STATE_PLAYING)
		|| (GetPlayerState() == PLAYER_STATE_PAUSED))
	{
		//
		// Don't think DivX supports dynamically changing play speed
		//
		CellSailStartCommand xCommand;
		Asura_Mem::Set(&xCommand, 0, sizeof(xCommand));

		xCommand.startType = CELL_SAIL_START_TIME_SCALE;
		uint32_t speed = (uint32_t)(fSpeed * 256.f);
		xCommand.startArg = speed;
		xCommand.seekType = CELL_SAIL_SEEK_CURRENT_POSITION;
		xCommand.flags |= CELL_SAIL_START_FLAG_PAUSE_END;

		iRet = cellSailPlayerStart(&pxPS3Data->m_xPlayer, &xCommand, NULL);
		Asura_Assertf(iRet >= CELL_OK, ("SetPlaySpeed() failed: 0x%X\n", iRet));
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnPlayerEvent
//
// Main event callback for the player.
// Forwards events FROM the player callback thread TO the main thread
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::OnPlayerEvent(void* pArg, CellSailEvent xEvent, uint64_t ulArg0, uint64_t ulArg1)
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(pArg);
	if (!pxPS3Data)
	{
		Asura_OutputToDebugger(("Asura_VideoToTexture_PS3Data::OnPlayerEvent pArg == NULL"));
	}

	static const int uTryNum = 4;
	int iRet;

	for (int i = uTryNum; i > 0; --i)
	{
		// put this event + args in the queue, so the main thread can read it
		iRet = sys_event_port_send(pxPS3Data->m_xPlayerEventPort, xEvent.u64.value, ulArg0, ulArg1);
		if (0 == iRet)
		{
			return;
		}
		// sleep for 5 msec
		sys_timer_usleep(1000*5);
	}

	Asura_OutputToDebugger(("sys_event_port_send() failed: 0x%X\n", iRet));
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnMemAlloc
//
// Called when the player needs to alloc main memory.
////////////////////////////////////////////////////////////////////////////////

void* Asura_VideoToTexture_PS3Data::OnMemAlloc(Asura_VideoToTexture *pxSelf, size_t uBoundary, size_t uSize)
{
	(void) pxSelf;
	Asura_Assertf(uBoundary != 0, ("Asura_VideoToTexture_PS3Data::OnMemAlloc(), Invalid boundary: %d", uBoundary));

	Asura_VideoToTexture::s_uTotalMemUsed += uSize;
	void *pNewMem = Asura_AllocateAlign(uSize, (ASURA_ALIGN_ID) uBoundary);

#ifdef ENABLE_DEBUG_LOG_MEM_ALLOCS
	Asura_OutputToDebugger_Retail(("Asura_VideoToTexture_PS3Data:Alloc of %dB [Total:%dB, 0x%X]\n", uSize, Asura_VideoToTexture::s_uTotalMemUsed, pNewMem));
#endif

	return pNewMem;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnMemFree
//
// Called when the player needs to free main memory.
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::OnMemFree(Asura_VideoToTexture *pxSelf, size_t uBoundary, void *pMemory)
{
	(void) pxSelf;
	(void) uBoundary;

#ifdef ENABLE_DEBUG_LOG_MEM_ALLOCS
	Asura_OutputToDebugger_Retail(("Asura_VideoToTexture_PS3Data:Free 0x%X\n", pMemory));
#endif

	return Asura_DeallocateAlign(pMemory);
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnSoundAdapterMakeup
//
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnSoundAdapterMakeup(Asura_VideoToTexture_PS3Data *pxArg)
{
	(void) pxArg;
	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnSoundAdapterCleanup
//
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnSoundAdapterCleanup(Asura_VideoToTexture_PS3Data *pxArg)
{
	(void) pxArg;
	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnSoundAdapterFormatChanged
//
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnSoundAdapterFormatChanged(Asura_VideoToTexture_PS3Data *pxArg, CellSailAudioFormat *pxFormat, uint32_t uCount)
{
	(void) pxArg;
	(void) uCount;

#ifndef ASURA_NO_DEBUG_OUTPUT

	const char *p;
	switch (pxFormat->coding)
	{
		case CELL_SAIL_AUDIO_CODING_LPCM_FLOAT32:
		{
			p = "CELL_SAIL_AUDIO_CODING_LPCM_FLOAT32";
			break;
		}

		case CELL_SAIL_AUDIO_CODING_UNSPECIFIED:
		{
			p = "CELL_SAIL_AUDIO_CODING_UNSPECIFIED";
			break;
		}

		default:
		{
			p = "unknown";
			break;
		}
	}

	Asura_OutputToDebugger(("coding = %s (%d)\n", p, pxFormat->coding));
	Asura_OutputToDebugger(("chNum = %d\n", pxFormat->chNum));
	Asura_OutputToDebugger(("sampleNum = %d\n", pxFormat->sampleNum));

	switch (pxFormat->fs)
	{
		case CELL_SAIL_AUDIO_FS_8000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_8000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_11025HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_11025HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_12000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_12000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_16000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_16000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_22050HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_22050HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_24000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_24000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_32000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_32000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_44100HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_44100HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_48000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_48000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_64000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_64000HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_88200HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_88200HZ";
			break;
		}

		case CELL_SAIL_AUDIO_FS_96000HZ:
		{
			p = "CELL_SAIL_AUDIO_FS_96000HZ";
			break;
		}

		default:
		{
			p = "unknown";
			break;
		}
	}
	Asura_OutputToDebugger(("fs = %s (%d)\n", p, pxFormat->fs));

	switch (pxFormat->chLayout)
	{
		case CELL_SAIL_AUDIO_CH_LAYOUT_UNDEFINED:
			p = "CELL_SAIL_AUDIO_CH_LAYOUT_UNDEFINED";
		break;

		case CELL_SAIL_AUDIO_CH_LAYOUT_1CH:
			p = "CELL_SAIL_AUDIO_CH_LAYOUT_1CH";
		break;

		case CELL_SAIL_AUDIO_CH_LAYOUT_2CH_LR:
			p = "CELL_SAIL_AUDIO_CH_LAYOUT_2CH_LR";
		break;

		case CELL_SAIL_AUDIO_CH_LAYOUT_6CH_LCRlrE:
			p = "CELL_SAIL_AUDIO_CH_LAYOUT_6CH_LCRlrE";
		break;

		case CELL_SAIL_AUDIO_CH_LAYOUT_6CH_LCRxyE:
			p = "CELL_SAIL_AUDIO_CH_LAYOUT_6CH_LCRxyE";
		break;

		case CELL_SAIL_AUDIO_CH_LAYOUT_8CH_LRCElrxy:
			p = "CELL_SAIL_AUDIO_CH_LAYOUT_8CH_LRCElrxy";
		break;

		default:
			p = "unknown";
		break;
	}
	Asura_OutputToDebugger(("chLayout = %s (%d)\n", p, pxFormat->chLayout));

#endif	// ASURA_NO_DEBUG_OUTPUT

	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnGraphicsAdapterMakeup
//
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnGraphicsAdapterMakeup(Asura_VideoToTexture *pxArg)
{
	(void) pxArg;
	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnGraphicsAdapterCleanup
//
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnGraphicsAdapterCleanup(Asura_VideoToTexture *pxArg)
{
	(void) pxArg;
	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnGraphicsAdapterFormatChanged
//
// Outputs the new graphics format
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnGraphicsAdapterFormatChanged(Asura_VideoToTexture *pxArg, CellSailVideoFormat *pxFormat, uint32_t uCount)
{
	(void) pxArg;
	(void) uCount;

#ifndef ASURA_NO_DEBUG_OUTPUT

	const char *p;
	switch (pxFormat->coding)
	{
		case CELL_SAIL_VIDEO_CODING_ARGB_INTERLEAVED:
		{
			p = "CELL_SAIL_VIDEO_CODING_ARGB_INTERLEAVED";
			break;
		}
		case CELL_SAIL_VIDEO_CODING_RGBA_INTERLEAVED:
		{
			p = "CELL_SAIL_VIDEO_CODING_RGBA_INTERLEAVED";
			break;
		}

		case CELL_SAIL_VIDEO_CODING_UYVY422_INTERLEAVED:
		{
			p = "CELL_SAIL_VIDEO_CODING_UYVY422_INTERLEAVED";
			break;
		}

		case CELL_SAIL_VIDEO_CODING_YUV420_PLANAR:
		{
			p = "CELL_SAIL_VIDEO_CODING_YUV420_PLANAR";
			break;
		}

		case CELL_SAIL_VIDEO_CODING_UNSPECIFIED:
		{
			p = "CELL_SAIL_VIDEO_CODING_UNSPECIFIED";
			break;
		}

		default:
		{
			p = "unknown";
			break;
		}
	}
	Asura_OutputToDebugger(("coding = %s (%d)\n", p, pxFormat->coding));

	switch (pxFormat->scan)
	{
		case CELL_SAIL_VIDEO_SCAN_PROGRESSIVE:
		{
			p = "CELL_SAIL_VIDEO_SCAN_PROGRESSIVE";
			break;
		}

		case CELL_SAIL_VIDEO_SCAN_INTERLACE:
		{
			p = "CELL_SAIL_VIDEO_SCAN_INTERLACE";
			break;
		}
		case CELL_SAIL_VIDEO_SCAN_UNSPECIFIED:
		{
			p = "unknown";
			break;
		}
	}

	Asura_OutputToDebugger(("scan = %s (%d)\n", p, pxFormat->scan));
	Asura_OutputToDebugger(("bitsPerColor = %d\n", pxFormat->bitsPerColor));

	switch (pxFormat->frameRate)
	{
		case CELL_SAIL_VIDEO_FRAME_RATE_24000_1001HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_24000_1001HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_24HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_24HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_25HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_25HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_30000_1001HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_30000_1001HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_30HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_30HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_50HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_50HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_60000_1001HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_60000_1001HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_60HZ:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_60HZ";
			break;
		case CELL_SAIL_VIDEO_FRAME_RATE_UNSPECIFIED:
			p = "CELL_SAIL_VIDEO_FRAME_RATE_UNSPECIFIED";
			break;
		default:
			p = "unknown";
			break;
	}

	Asura_OutputToDebugger(("frameRate = %s (%d)\n", p, pxFormat->frameRate));
	Asura_OutputToDebugger(("width  = %d\n", pxFormat->width));
	Asura_OutputToDebugger(("height = %d\n", pxFormat->height));
	Asura_OutputToDebugger(("pitch  = %d\n", pxFormat->pitch));
	Asura_OutputToDebugger(("alpha  = %d\n", pxFormat->alpha));

	switch (pxFormat->colorMatrix)
	{
		case CELL_SAIL_VIDEO_COLOR_MATRIX_BT601:
			p = "CELL_SAIL_VIDEO_COLOR_MATRIX_BT601";
			break;
		case CELL_SAIL_VIDEO_COLOR_MATRIX_BT709:
			p = "CELL_SAIL_VIDEO_COLOR_MATRIX_BT709";
			break;
		case CELL_SAIL_VIDEO_COLOR_MATRIX_UNSPECIFIED:
			p = "CELL_SAIL_VIDEO_COLOR_MATRIX_UNSPECIFIED";
			break;
		default:
			Asura_OutputToDebugger(("colorMatrix = unknown(%d)\n", pxFormat->colorMatrix));
			break;
	}
	Asura_OutputToDebugger(("colorMatrix = %s (%d)\n", p, pxFormat->colorMatrix));

	switch (pxFormat->aspectRatio)
	{
		case CELL_SAIL_VIDEO_ASPECT_RATIO_1_1:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_1_1(1920x1080 1280x720)";
			break;
		case CELL_SAIL_VIDEO_ASPECT_RATIO_12_11:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_12_11(720x576 normal)";
			break;
		case CELL_SAIL_VIDEO_ASPECT_RATIO_10_11:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_10_11(720x480 normal)";
			break;
		case CELL_SAIL_VIDEO_ASPECT_RATIO_16_11:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_16_11(720x576 wide)";
			break;
		case CELL_SAIL_VIDEO_ASPECT_RATIO_40_33:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_40_33(720x480 wide)";
			break;
		case CELL_SAIL_VIDEO_ASPECT_RATIO_4_3:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_4_3(1440x1080)";
			break;
		case CELL_SAIL_VIDEO_ASPECT_RATIO_UNSPECIFIED:
			p = "CELL_SAIL_VIDEO_ASPECT_RATIO_UNSPECIFIED";
			break;
		default:
			p = "unknown";
			break;
	}
	Asura_OutputToDebugger(("aspectRatio = %s (%d)\n", p, pxFormat->aspectRatio));

#endif // ASURA_NO_DEBUG_OUTPUT

	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnGraphicsAdapterAllocFrame
//
// Called when the player needs to alloc graphics memory
// Returns : 0 = all requested frames were allocated.
//			 >0 = number of allocated frames.
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnGraphicsAdapterAllocFrame(Asura_VideoToTexture_PS3Data *pxArg, size_t uSize, int iNum, uint8_t **ppucFrame)
{
	uSize = ROUNDUP(uSize, 128);

	// the player's requesting iNum frames, but we're only going to give it s_uTotalPlayerVideoFrames
	for (unsigned int iFrmNum = 0; iFrmNum < s_uTotalPlayerVideoFrames; iFrmNum++)
	{
		ppucFrame[iFrmNum] = pxArg->m_pcVideoFrames[iFrmNum] = (uint8_t*) ASURA_PS3_VIDEOMEMORY_ALLOCMAIN(uSize, 128);

		Asura_VideoToTexture::s_uTotalMemUsed += uSize;

#ifdef ENABLE_DEBUG_LOG_MEM_ALLOCS
		Asura_OutputToDebugger_Retail(("Asura_VideoToTexture_PS3Data:VidAlloc of %dB [Total:%dB, 0x%X]\n", uSize, Asura_VideoToTexture::s_uTotalMemUsed, ppucFrame[iFrmNum]));
#endif
	}

	return s_uTotalPlayerVideoFrames;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OnGraphicsAdapterFreeFrame
//
// Called when the player needs to free graphics memory
////////////////////////////////////////////////////////////////////////////////

int Asura_VideoToTexture_PS3Data::OnGraphicsAdapterFreeFrame(Asura_VideoToTexture_PS3Data *pxArg, int iNum, uint8_t **ppucFrame)
{
	(void) pxArg;

	for (int iFrmNum = 0; iFrmNum < iNum; iFrmNum++)
	{
#ifdef ENABLE_DEBUG_LOG_MEM_ALLOCS
		Asura_OutputToDebugger_Retail(("Asura_VideoToTexture_PS3Data:VidFree 0x%X\n", ppucFrame[iFrmNum]));
#endif
		Asura_PS3_VideoMemoryManager::FreeMain(ppucFrame[iFrmNum]);
	}

	return CELL_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_GetPlayedTimeInSeconds
//
////////////////////////////////////////////////////////////////////////////////

float Asura_VideoToTexture::Platform_GetPlayedTimeInSeconds()
{
	Asura_VideoToTexture_PS3Data *pxPS3Data = reinterpret_cast<Asura_VideoToTexture_PS3Data*>(m_pPlatformData);

	return( static_cast<float> (pxPS3Data->m_ulVideoTime / 1000000));
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::SwapInPlaceholderTexture()
{
	if (m_pxParentObject)
	{
		//
		// Put the old texture data back
		//
		int iTex = m_pxParentObject->GetTextureRTIndex();
		if (iTex != ASURA_TEXTURE_NONE)
		{
			Asura_PS3_Texture* const pxTexture = Asura_PS3_TextureManagement::GetTexture(iTex);
			if (pxTexture)
			{
				*pxTexture->GetTexture() = m_xPlaceholderGcmTextureData;
				pxTexture->SetTextureData(m_pPlaceholderTextureData);
				m_bRenderingPlaceholderTexture = true;
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::SwapOutPlaceholderTexture()
{
	int iTex = m_pxParentObject->GetTextureRTIndex();
	if (iTex != ASURA_TEXTURE_NONE)
	{
		Asura_Assert(m_pxVideoTextureTarget, "Video texture NULL!");
		Asura_Assert(m_pcVideoTextureTargetData, "Video texture NULL!");

		// Populate GCM texture information
		CellGcmTexture xTexture;
		xTexture.width = m_uVideoWidth;
		xTexture.height = m_uVideoHeight;
		xTexture.mipmap = 1;
		xTexture.format = CELL_GCM_TEXTURE_A8R8G8B8 | CELL_GCM_TEXTURE_LN | CELL_GCM_TEXTURE_NR;
		xTexture.pitch = xTexture.width * 4;
		xTexture.cubemap = CELL_GCM_FALSE;
		xTexture.location = CELL_GCM_LOCATION_MAIN;
		xTexture._padding = 0;
		xTexture.depth = 1;
		xTexture.dimension = CELL_GCM_TEXTURE_DIMENSION_2;
		xTexture.remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_A,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );

		// Create texture
		cellGcmAddressToOffset( m_pcVideoTextureTargetData, &xTexture.offset );

		// this will disconnect us from the placeholder texture data and use the new buffer instead
		*m_pxVideoTextureTarget = Asura_PS3_Texture( &xTexture );
		m_pxVideoTextureTarget->SetTextureData(m_pcVideoTextureTargetData);

		m_bRenderingPlaceholderTexture = false;
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_PS3Data::OutputEvent
//
// Writes an event to the debug log
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_PS3Data::OutputEvent(CellSailEvent xEvent, uint64_t ulArg0, uint64_t ulArg1)
{
	if (xEvent.u32x2.major != 0)
	{
		Asura_OutputToDebugger(("event:%d:%d, arg0:%lld, arg1:%lld\n",
			xEvent.u32x2.major,
			xEvent.u32x2.minor,
			ulArg0,
			ulArg1));
	}

	switch (xEvent.u32x2.major)
	{
		case CELL_SAIL_EVENT_ERROR_OCCURED:
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_ERROR_OCCURED ret=0x%08X\n", (int) ulArg0));
			break;

		case CELL_SAIL_EVENT_PLAYER_CALL_COMPLETED:
		{
			const char* p = 0;

			switch (xEvent.u32x2.minor)
			{
				case CELL_SAIL_PLAYER_CALL_BOOT:
					p = "CELL_SAIL_PLAYER_CALL_BOOT";
					break;
				case CELL_SAIL_PLAYER_CALL_OPEN_STREAM:
					p = "CELL_SAIL_PLAYER_CALL_OPEN_STREAM";
					break;
				case CELL_SAIL_PLAYER_CALL_CLOSE_STREAM:
					p = "CELL_SAIL_PLAYER_CALL_CLOSE_STREAM";
					break;
				case CELL_SAIL_PLAYER_CALL_START:
					p = "CELL_SAIL_PLAYER_CALL_START";
					break;
				case CELL_SAIL_PLAYER_CALL_STOP:
					p = "CELL_SAIL_PLAYER_CALL_STOP";
					break;
				case CELL_SAIL_PLAYER_CALL_OPEN_ES_AUDIO:
					p = "CELL_SAIL_PLAYER_CALL_OPEN_ES_AUDIO";
					break;
				case CELL_SAIL_PLAYER_CALL_OPEN_ES_VIDEO:
					p = "CELL_SAIL_PLAYER_CALL_OPEN_ES_VIDEO";
					break;
				case CELL_SAIL_PLAYER_CALL_CLOSE_ES_AUDIO:
					p = "CELL_SAIL_PLAYER_CALL_CLOSE_ES_AUDIO";
					break;
				case CELL_SAIL_PLAYER_CALL_CLOSE_ES_VIDEO:
					p = "CELL_SAIL_PLAYER_CALL_CLOSE_ES_VIDEO";
					break;

				case CELL_SAIL_PLAYER_CALL_REOPEN_ES_AUDIO:
					p = "CELL_SAIL_PLAYER_CALL_REOPEN_ES_AUDIO";
					break;
				case CELL_SAIL_PLAYER_CALL_REOPEN_ES_VIDEO:
					p = "CELL_SAIL_PLAYER_CALL_REOPEN_ES_VIDEO";
					break;

				case CELL_SAIL_PLAYER_CALL_NEXT:
					p = "CELL_SAIL_PLAYER_CALL_NEXT";
					break;
				default:
					p  = "unknown";
					break;
			}

			Asura_OutputToDebugger(("CELL_SAIL_EVENT_PLAYER_CALL_COMPLETED ret:%lld\n", ulArg0));
			Asura_OutputToDebugger(("  %s (%u)\n", p, xEvent.u32x2.minor));
			break;
		}

		case CELL_SAIL_EVENT_PLAYER_STATE_CHANGED:
		{
			const char* p = 0;
			switch (ulArg0)
			{
				case CELL_SAIL_PLAYER_STATE_INITIALIZED:
					p = "CELL_SAIL_PLAYER_STATE_INITIALIZED";
					break;
				case CELL_SAIL_PLAYER_STATE_BOOT_TRANSITION:
					p = "CELL_SAIL_PLAYER_STATE_BOOT_TRANSITION";
					break;
				case CELL_SAIL_PLAYER_STATE_CLOSED:
					p = "CELL_SAIL_PLAYER_STATE_CLOSED";
					break;
				case CELL_SAIL_PLAYER_STATE_OPEN_TRANSITION:
					p = "CELL_SAIL_PLAYER_STATE_OPEN_TRANSITION";
					break;

				case CELL_SAIL_PLAYER_STATE_OPENED:
					p = "CELL_SAIL_PLAYER_STATE_OPENED";
					break;

				case CELL_SAIL_PLAYER_STATE_START_TRANSITION:
					p = "CELL_SAIL_PLAYER_STATE_START_TRANSITION";
					break;

				case CELL_SAIL_PLAYER_STATE_RUNNING:
					p = "CELL_SAIL_PLAYER_STATE_RUNNING";
					break;

				case CELL_SAIL_PLAYER_STATE_STOP_TRANSITION:
					p = "CELL_SAIL_PLAYER_STATE_STOP_TRANSITION";
					break;

				case CELL_SAIL_PLAYER_STATE_CLOSE_TRANSITION:
					p = ("CELL_SAIL_PLAYER_STATE_CLOSE_TRANSITION\n");
					break;

				case CELL_SAIL_PLAYER_STATE_LOST:
					p = "CELL_SAIL_PLAYER_STATE_LOST";
					break;

				default:
					p = "unknown";
			}
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_PLAYER_STATE_CHANGED\n"));
			Asura_OutputToDebugger(("  %s (%llu)\n", p, ulArg0));
		}
		break;

		case CELL_SAIL_EVENT_STREAM_OPENED:
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_STREAM_OPEN pDesc=0x%llX\n", ulArg0));
			break;

		case CELL_SAIL_EVENT_STREAM_CLOSED:
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_STREAM_CLOSED pDesc=0x%llX\n", ulArg0));
			break;

		case CELL_SAIL_EVENT_SESSION_STARTED:
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_SESSION_STARTED arg=%lld\n", ulArg0));
			break;

		case CELL_SAIL_EVENT_PAUSE_STATE_CHANGED:
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_PAUSE_STATE_CHANGED arg=%lld\n", ulArg0));
			break;

		case CELL_SAIL_EVENT_SOURCE_EOS:
			Asura_OutputToDebugger(("CELL_SAIL_EVENT_SOURCE_EOS arg=%lld\n", ulArg0));
			break;

		case CELL_SAIL_EVENT_ES_OPENED:
			switch (ulArg0)
			{
				case CELL_SAIL_ES_AUDIO:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_OPENED AUDIO index=%lld\n", ulArg1));
					break;
				case CELL_SAIL_ES_VIDEO:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_OPENED VIDEO index=%lld\n", ulArg1));
					break;
				case CELL_SAIL_ES_USER:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_OPENED USER index=%lld\n", ulArg1));
					break;
				default:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_OPENED unknown index=%lld\n", ulArg1));
					break;
			}
			break;

		case CELL_SAIL_EVENT_ES_CLOSED:
			switch (ulArg0)
			{
				case CELL_SAIL_ES_AUDIO:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_CLOSED AUDIO index=%lld\n", ulArg1));
					break;
				case CELL_SAIL_ES_VIDEO:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_CLOSED VIDEO index=%lld\n", ulArg1));
					break;
				case CELL_SAIL_ES_USER:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_CLOSED USER index=%lld\n", ulArg1));
					break;
				default:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_ES_CLOSED unknown index=%lld\n", ulArg1));
					break;
			}
			break;

		case CELL_SAIL_EVENT_MEDIA_STATE_CHANGED:
			switch (ulArg0)
			{
				case CELL_SAIL_MEDIA_STATE_FINE:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_MEDIA_STATE_CHANGED state=FINE desc=%p\n",
						CAST_UINT64_TO_PTR(CellSailDescriptor*, ulArg1)));
					break;
				case CELL_SAIL_MEDIA_STATE_BAD:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_MEDIA_STATE_CHANGED state=BAD desc=%p\n",
						CAST_UINT64_TO_PTR(CellSailDescriptor*, ulArg1)));
					break;
				case CELL_SAIL_MEDIA_STATE_LOST:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_MEDIA_STATE_CHANGED state=LOST desc=%p\n",
						CAST_UINT64_TO_PTR(CellSailDescriptor*, ulArg1)));
					break;
				default:
					Asura_OutputToDebugger(("CELL_SAIL_EVENT_MEDIA_STATE_CHANGED state=unknown desc=%p\n",
						CAST_UINT64_TO_PTR(CellSailDescriptor*, ulArg1)));
			}
			break;

		default:
			if (0 != xEvent.u64.value)
			{
				Asura_OutputToDebugger(("unknown event\n"));
			}
			break;
	}
}
#endif	// ASURA_NO_DEBUG_OUTPUT

#endif // ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM
