//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Job.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     PS3 implementations of Asura_Job code
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE   

// Stdlib platform-specific.
#include <cell/sync.h>
#include <spu_printf.h>
#include <string.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"
#include "Jobs/Asura_Job.h"


// Engine platform-specific.
#include "PPU_SPU_Utils/spu_thread_group_utils.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// This won't work terribly well without support for the data blocks.
#ifndef ASURA_PFX_TEMPLATES_SUPPORT_DATABLOCK
#error You must define ASURA_PFX_TEMPLATES_SUPPORT_DATABLOCK to use the PS3 PFX job code
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int uSPU_THREAD_PRIORITY = 200;

//////////////////////////////////////////////////////////////////////////////
// Asura_Job_Queue
//////////////////////////////////////////////////////////////////////////////

#ifdef	SPURS_PFX	//	See Asura_Engine_Configuration for appropriate project

	#include "SpecialFX/Asura_PS3_SPURS_PFX.h"

	#include "PS3/PPU_SPU_UTILS/Asura_PS3_SPU_HelperStructs.h"	
	#include "PS3/PPU_SPU_UTILS/Asura_PS3_SPURS.h"
	#include "PS3/PPU_SPU_UTILS/Asura_PS3_SPURS_TaskController.h"	
#else

	#include "Jobs/Asura_Job_PFX.h"	// hack

#endif

static const u_int uMAX_QUEUE_DATA_SIZE = 512;	// bytes

// new system, written for PS3 to use libsync library's Queue implementation.
class Asura_Job_Queue_PS3Data
{
public:
#ifdef SPURS_PFX
	CellSpursLFQueue* m_pxQueue;
#else
	CellSyncQueue* m_pxQueue;
#endif
	u_int m_uNumUnfinished;
	void* m_pxBuffer;
	u_int m_uElementDataSize; 
	u_int m_uAlignedElementDataSize;
};

static UtilSpuThreadGroupInfo g_xSPUThreadGroupInfo;

void Asura_Job_Queue::Platform_Initialise( u_int uDepth, u_int uDataSize )
{
#ifdef	SPURS_PFX
	u_int uElementSize = (uDataSize + 15) & ~15;	// common
	
	Asura_PS3_SPURS_PFX::Initialise( uDepth, uElementSize );

	Asura_Job_Queue_PS3Data* pxPs3Data = new Asura_Job_Queue_PS3Data;
	m_pxPlatformData = pxPs3Data;
	Asura_Assert(uElementSize <= uMAX_QUEUE_DATA_SIZE, "data element too big");
	pxPs3Data->m_pxBuffer = NULL;
	pxPs3Data->m_pxQueue = Asura_PS3_SPURS_PFX::GetTaskController()->GetQueue();
	pxPs3Data->m_uNumUnfinished = 0;
	pxPs3Data->m_uElementDataSize = uDataSize;
	pxPs3Data->m_uAlignedElementDataSize = uElementSize;

	Asura_PS3_SPURS_PFX::GetTaskParameters()->uNumUnfinishedAddr = reinterpret_cast<uint64_t>( &pxPs3Data->m_uNumUnfinished );
	Asura_PS3_SPURS_PFX::Begin();
#else
	// create and initialise PS3 data
	Asura_Job_Queue_PS3Data* pxPs3Data = new Asura_Job_Queue_PS3Data;
	m_pxPlatformData = pxPs3Data;
	u_int uElementSize = (uDataSize + 15) & ~15;// align to 16 bytes
	Asura_Assert(uElementSize <= uMAX_QUEUE_DATA_SIZE, "data element too big");
	pxPs3Data->m_pxBuffer = Asura_AllocateAlign(uDepth * uElementSize, ASURA_ALIGN_16_BYTE);
	pxPs3Data->m_pxQueue = reinterpret_cast<CellSyncQueue*>(Asura_AllocateAlign( sizeof(CellSyncQueue), ASURA_ALIGN_32_BYTE));
	pxPs3Data->m_uNumUnfinished = 0;
	pxPs3Data->m_uElementDataSize = uDataSize;
	pxPs3Data->m_uAlignedElementDataSize = uElementSize;

#ifdef ASURA_ASSERT
    int iRetCode;
#endif

	// create queue
#ifdef ASURA_ASSERT
	iRetCode = 
#endif
	cellSyncQueueInitialize( pxPs3Data->m_pxQueue, pxPs3Data->m_pxBuffer, uElementSize, uDepth );
	Asura_Assertf( iRetCode == CELL_OK, ("cellSyncQueueInitialize failed with code 0x%08X", iRetCode) );

	// create SPU worker threads
	extern u_int _binary_PFX_SPU_elf_start;
	const u_int* const puSPUProg = &_binary_PFX_SPU_elf_start;

	// code stolen from queue sample (sample_sync_queue)
	const u_int uNUM_SPU_THREADS = 1;
	UtilSpuThreadGroupArg  groupArg;

	groupArg.group_name = "PFX_Update";
	groupArg.priority   = uSPU_THREAD_PRIORITY;
	groupArg.nThreads   = uNUM_SPU_THREADS;

	for(u_int u = 0; u < uNUM_SPU_THREADS; ++u)
	{
		groupArg.thread[u].name   = "PFX_Update_SPU";	// for debugger
		groupArg.thread[u].option = SYS_SPU_THREAD_OPTION_NONE;
		groupArg.thread[u].elf    = reinterpret_cast<sys_addr_t>( puSPUProg );
		groupArg.thread[u].arg[0] = SYS_SPU_THREAD_ARGUMENT_LET_32( u );
		groupArg.thread[u].arg[1] = reinterpret_cast<uint64_t>( pxPs3Data->m_pxQueue );
		groupArg.thread[u].arg[2] = reinterpret_cast<uint64_t>( &pxPs3Data->m_uNumUnfinished );
		groupArg.thread[u].arg[3] = 0;
	}

#ifdef ASURA_ASSERT
	iRetCode = 
#endif
	utilInitializeSpuThreadGroupAll( &g_xSPUThreadGroupInfo, &groupArg );
	Asura_Assertf( iRetCode == CELL_OK, ("utilInitializeSpuThreadGroupAll() failed with code 0x%08X", iRetCode) );

#ifndef ASURA_NO_DEBUG_OUTPUT
	for( u_int u = 0; u < uNUM_SPU_THREADS; ++u ) 
	{
#ifdef ASURA_ASSERT
		iRetCode =
#endif
			spu_printf_attach_thread( g_xSPUThreadGroupInfo.thread_id[u] );
#ifdef ASURA_ASSERT
        if( iRetCode != CELL_OK )
        {
		    Asura_Assertf( false, ("spu_printf_attach_thread() failed for thread %u with code 0x%08X", u, iRetCode) );
        }
#endif
	}
#endif

#ifdef ASURA_ASSERT
	iRetCode = 
#endif
	utilStartSpuThreadGroup( &g_xSPUThreadGroupInfo );
	Asura_Assertf( iRetCode == CELL_OK, ("utilStartSpuThreadGroup() failed with code %i", iRetCode) );

#endif	//	SPURS_PFX

}


void Asura_Job_Queue::Platform_Shutdown( void )
{
#ifdef SPURS_PFX
	Asura_PS3_SPURS_PFX::Shutdown();
#else
	// destroy worker threads by posting invalid job
	Asura_Job_PFX_Update xJob;
	Push(&xJob);

    UtilSpuThreadGroupStatus xGroupStatus;
#ifdef ASURA_ASSERT
	int iRet =
#endif
    utilWaitSpuThreadGroup( &g_xSPUThreadGroupInfo, &xGroupStatus );
	Asura_Assertf( iRet == CELL_OK, ("utilWaitSpuThreadGroup() failed with code %i", iRet) );

#ifdef ASURA_ASSERT
    int iStatus = -1;
	switch( xGroupStatus.cause )
    {
	    case SYS_SPU_THREAD_GROUP_JOIN_GROUP_EXIT:
		    iStatus = xGroupStatus.status;
            break;

	    case SYS_SPU_THREAD_GROUP_JOIN_ALL_THREADS_EXIT:
		    iStatus = CELL_OK;
		    for( int i = 0; i < xGroupStatus.nThreads; i++ )
            {
			    if( xGroupStatus.threadStatus[i] != CELL_OK )
                {
				    iStatus = xGroupStatus.threadStatus[i];
                    break;
			    }
		    }
            break;

	    case SYS_SPU_THREAD_GROUP_JOIN_TERMINATED: 
		    iStatus = xGroupStatus.status;
            break;
	}
    Asura_Assertf( iStatus == CELL_OK, ("bad status: %i", iStatus) );
#endif

#ifdef ASURA_ASSERT
	iRet =
#endif
    utilFinalizeSpuThreadGroupAll( &g_xSPUThreadGroupInfo );
	Asura_Assertf( iRet == CELL_OK, ("utilFinalizeSpuThreadGroupAll() failed with code %i", iRet) );

	// free mem allocations
	Asura_Job_Queue_PS3Data* pxPs3Data = static_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData);
	Asura_DeallocateAlign(pxPs3Data->m_pxBuffer);
	Asura_DeallocateAlign(pxPs3Data->m_pxQueue);
	delete pxPs3Data;
	m_pxPlatformData = NULL;
#endif
}

u_int Asura_Job_Queue::Platform_GetNumEntries( void ) const
{
	int iRet = 0;
#ifdef	SPURS_PFX
	iRet = Asura_PS3_SPURS_PFX::GetTaskController()->GetNumEntriesInQueue();

#else
	CellSyncQueue* pxQueue = static_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
	iRet = cellSyncQueueSize(pxQueue);
#endif
	if (iRet < 0)
	{
		Asura_Assertf(false, ("error: cellSyncQueueSize returned %d\n", iRet));
		return 0;
	}
	return static_cast<u_int>(iRet);
}

// Platform_Push should increment the number of unfinished jobs
void Asura_Job_Queue::Platform_Push( void* pData )
{
	cellAtomicIncr32( &(reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_uNumUnfinished) );
#ifdef	SPURS_PFX
	CellSpursLFQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#else
	CellSyncQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#endif

#ifdef ASURA_ASSERT
	int iRet = 
#endif

#ifdef	SPURS_PFX
	Asura_PS3_SPURS_PFX::GetTaskController()->PushTask_Blocking( pData );
#else
		cellSyncQueuePush( pxQueue, pData );
#endif
	Asura_Assertf(iRet == CELL_OK, ("cellSyncQueuePush failed (%d)", iRet));
}

void Asura_Job_Queue::Platform_Pop( void* pBuffer )
{
	// input buffer may just be an object on the stack. It is not guaranteed to be
	// 128-byte aligned, or large enough to take the 128-byte aligned data size.

#ifdef SPURS_PFX
	CellSpursLFQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#else
	CellSyncQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#endif

	// create buffer on the stack and align to 128 bytes.
	char acBuffer[uMAX_QUEUE_DATA_SIZE + 127];
	uint64_t uBuffer = reinterpret_cast<uint64_t>(acBuffer);
	char* pcAlignedBuffer = reinterpret_cast<char*>((uBuffer + 127) & ~127);

#ifdef ASURA_ASSERT
	int iRet = 
#endif

#ifdef	SPURS_PFX
		Asura_PS3_SPURS_PFX::GetTaskController()->PopTask_Blocking( pcAlignedBuffer );
#else
		cellSyncQueuePop( pxQueue, pcAlignedBuffer );
#endif
	Asura_Assertf(iRet == CELL_OK, ("cellSyncQueuePop failed (%d)", iRet));

	// copy to (possibly smaller and unaligned) input buffer
	memcpy( pBuffer, pcAlignedBuffer, reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_uElementDataSize );
}

bool Asura_Job_Queue::Platform_TryPop( void* pBuffer )
{
	// input buffer may just be an object on the stack. It is not guaranteed to be
	// 128-byte aligned, or large enough to take the 128-byte aligned data size.

#ifdef SPURS_PFX
	CellSpursLFQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#else
	CellSyncQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#endif

	// create buffer on the stack and align to 128 bytes.
	char acBuffer[uMAX_QUEUE_DATA_SIZE + 127];
	uint64_t uBuffer = reinterpret_cast<uint64_t>(acBuffer);
	char* pcAlignedBuffer = reinterpret_cast<char*>((uBuffer + 127) & ~127);

	int iRet = 0;
#ifdef	SPURS_PFX
	iRet = Asura_PS3_SPURS_PFX::GetTaskController()->PopTask( pBuffer );
#else
	iRet = cellSyncQueueTryPop( pxQueue, pcAlignedBuffer );
#endif
	Asura_Assertf(iRet == CELL_OK || iRet == CELL_SYNC_ERROR_BUSY, ("cellSyncQueuePop failed (%d)", iRet));

	if (iRet == CELL_SYNC_ERROR_BUSY)
	{
		return false;
	}

	// copy to (possibly smaller and unaligned) input buffer
	memcpy( pBuffer, pcAlignedBuffer, reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_uElementDataSize );
	return true;
}

// After a job has been executed, the code that did it should call JobFinished()
// Platform_JobFinished should decrement the number of unfinished jobs
void Asura_Job_Queue::Platform_JobFinished( void ) const
{
	cellAtomicDecr32( &(reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_uNumUnfinished) );
}

// Number of unfinished jobs is maintained so code that requires all jobs to be 
// complete can wait until they are.
u_int Asura_Job_Queue::Platform_GetNumUnfinishedJobs( void ) const
{
	return (reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_uNumUnfinished);
}

void Asura_Job_Queue::Platform_Clear( void )
{
#ifdef SPURS_PFX
	CellSpursLFQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#else
	CellSyncQueue* pxQueue = reinterpret_cast<Asura_Job_Queue_PS3Data*>(m_pxPlatformData)->m_pxQueue;
#endif

#ifdef ASURA_ASSERT
	int iRet = 
#endif

#ifdef	SPURS_PFX
		Asura_PS3_SPURS_PFX::GetTaskController()->ClearQueue();
#else
		cellSyncQueueClear( pxQueue );
#endif

	Asura_Assertf(iRet == CELL_OK, ("cellSyncQueueClear failed (%d)", iRet));
}


void Asura_Job_Queue::Platform_SetThreadName( const char* szThreadName )
{
}


void Asura_Job_Queue::Platform_WaitForAllJobsToFinish( void )
{
}

#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
