//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_DLC_System.cpp
//
// Copyright (c) 2009 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT

// Own header
#include "Network/Asura_PS3_DLC_System.h"

// Asura independent
#include "Debugging/Asura_Assert.h"
#include "Network/Asura_DLC_System.h"

// Asura PS3
#include "FileAccess/Asura_PS3_FileSystem.h"
#include "Utility/Asura_PS3_GameContent.h"

// Library headers
#include <cell/cell_fs.h>
#include <cell/sysmodule.h>
#include <np.h>

// Default licensee key data. TODO - This should probably become a #error as a game approaches completion.
#ifndef ASURA_PS3_DLC_PROJECT_LICENSEE_KEY
	#define ASURA_PS3_DLC_PROJECT_LICENSEE_KEY { '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0', '\0' }
#endif //#ifndef ASURA_PS3_DLC_PROJECT_LICENSEE_KEY

// String to make the directory name unique.
static const char* szASURA_PS3_DLC_DIRECTORY = "DLC";
static const char* szDRM_ENCRYPTED_FILE_EXTENSION = ".edat";

// NP memory pool.
static const int iNP_POOL_SIZE = 128 * 1024;
static uint8_t g_aucNpPool[iNP_POOL_SIZE];

// Set up the licensee key.
const SceNpDrmKey Asura_PS3_DLC_System::s_xDRMKey = { ASURA_PS3_DLC_PROJECT_LICENSEE_KEY };

Asura_PS3_GameContent* Asura_PS3_DLC_System::s_pxGameContent = NULL;
char Asura_PS3_DLC_System::s_acRootDirectory[128];

void Asura_PS3_DLC_System::Initialise()
{
	int iRet = cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_NP);
	Asura_Assertf(iRet == CELL_OK, ("cellSysmoduleLoadModule failed to load NP module (0x%x)", iRet));

	if (iRet != CELL_OK) { return; }

	iRet = sceNpInit(iNP_POOL_SIZE, g_aucNpPool);
	Asura_Assertf(iRet == CELL_OK, ("sceNpInit failed (0x%x)", iRet));

	if (iRet != CELL_OK) { return; }

	s_acRootDirectory[0] = '\0';
	s_pxGameContent = Asura_PS3_GameContent::CreateGameDataContent(NULL); // Always use the default content directory, we'll be enforcing the DLC content is within a directory called "DLC".
	if (s_pxGameContent)
	{
		const char* szUsrDir = s_pxGameContent->GetUsrDirectory();
		if (szUsrDir && (*szUsrDir != '\0'))
		{
			Asura_PrintfToBfr(s_acRootDirectory, "%s/%s", 128, true, szUsrDir, szASURA_PS3_DLC_DIRECTORY);
			// Make sure this directory exists - it's entirely valid to have existing content in this location without having created the DLC folder.
			int iFileDesc = -1;
			CellFsErrno xError = cellFsOpendir(s_acRootDirectory, &iFileDesc);
			if (xError == CELL_FS_SUCCEEDED)
			{
				// Opened fine, close the handle again.
				cellFsClosedir(iFileDesc);
			}
			else
			{
				// Failed to open, should just mean the directory doesn't exist. Invalidate the path.
				s_acRootDirectory[0] = '\0';
			}
		}
	}
}

void Asura_PS3_DLC_System::Shutdown()
{
	Asura_PS3_GameContent::DestroyContent(s_pxGameContent);
	s_pxGameContent = NULL;
	s_acRootDirectory[0] = '\0';

#ifdef ASURA_ASSERT
	int iRet =
#endif //#ifdef ASURA_ASSERT
		sceNpTerm();

	Asura_Assertf(iRet == CELL_OK, ("sceNpTerm failed (0x%x), was it initialised or has it already been shut down?", iRet));

#ifdef ASURA_ASSERT
	iRet =
#endif //#ifdef ASURA_ASSERT
		cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_NP);

	Asura_Assertf(iRet == CELL_OK, ("cellSysmoduleUnloadModule failed to unload NP module (0x%x)", iRet));
}

void Asura_PS3_DLC_System::UpdateDownloadedFiles(Asura_TemplateHashTree<Asura_DLC_DownloadedFileData*>& xDLCFiles)
{
	// We need to search the folders, but we're doing this manually so as to avoid searching non-dlc files.
	if (s_acRootDirectory[0] == '\0') { return; }

	Asura_List<Asura_String> xFiles;

	Asura_PS3_FileSystem::FindFiles(xFiles, s_acRootDirectory, "", "*");

	char acFullPath[256];

	for (Asura_List_It<Asura_String> xIt(xFiles); !xIt.Done();)
	{
		Asura_String* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			const char* szName = pxCurrent->GetString();
			// For consistency, always remove a leading slash if it exists.
			if (szName[0] == '/')
			{
				szName = &(szName[1]);
			}
			Asura_PrintfToBfr(acFullPath, "%s/%s", 256, true, s_acRootDirectory, szName);

			int iRet = 0; // If the file isn't encrypted then it's always available.
			if (pxCurrent->EndsWith(szDRM_ENCRYPTED_FILE_EXTENSION, true))
			{
				iRet = sceNpDrmIsAvailable(GetDRMKey(), acFullPath);
			}

			if (iRet == 0) // Documentation doesn't specify a name for the success value of sceNpDrmIsAvailable.
			{
				// The PS3 can use the default file data for now. It also has no way to fill the restrictions flags at present.
				Asura_DLC_DownloadedFileData* pxData = new Asura_DLC_DownloadedFileData(0);
				if (pxData)
				{
					xDLCFiles.AddNode(Asura_GetHashID(szName), pxData);
				}
			}
			else
			{
				// Check that it's simply that we don't have the entitlement to this file, any other error should be warned about.
				Asura_Assertf(static_cast<u_int>(iRet) == SCE_NP_DRM_ERROR_NO_ENTITLEMENT, ("sceNpDrmIsAvailable failed with error code 0x%x on file %s", iRet, szName));
			}
		}
		// Clear the entry from the list and delete the string.
		xIt.DeleteCurrent();
		delete pxCurrent;
	}
}

bool Asura_PS3_DLC_System::IsNewContentAvailable()
{
	return false;
}

void Asura_PS3_DLC_System::OpenStore()
{
}

const char* Asura_PS3_DLC_System::GetRootDirectory()
{
	if (s_acRootDirectory[0] != '\0')
	{
		return s_acRootDirectory;
	}
	else
	{
		return NULL;
	}
}

bool Asura_PS3_DLC_System::IsEncryptedFile(const char* szFilename)
{
	// For now, just check the extension.
	// If we decide to treat encrypted data as the original filename with .edat appended we'll have to store the encrypted state in the hash tree (so it can be accessed by calling code without the .edat extension).
	return Asura_StringUtil::EndsWith(szFilename, szDRM_ENCRYPTED_FILE_EXTENSION, true);
}

#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
