//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_Network.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     wrapper for various network libraries
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Asura_PS3_Network.h"

// Asura includes
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// library includes
#include <cell/sysmodule.h>
#include <netex/libnetctl.h>
#include <netex/net.h>

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_PS3_Network::s_bInitialised;
bool Asura_PS3_Network::s_bConnecting;
bool Asura_PS3_Network::s_bConnectionError;
int Asura_PS3_Network::s_iEventHandlerID = -1;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
bool Asura_PS3_Network::Initialise()
{
	Asura_Assert(!s_bInitialised, "Asura_PS3_Network already initialised");
	if (!s_bInitialised)
	{
		int iRet = 0;

#ifndef ASURA_USE_GCM_REPLAY		
		iRet = cellSysmoduleLoadModule(CELL_SYSMODULE_NET);
		if (iRet < 0)
		{
			Asura_Assertf(false, ("cellSysmoduleLoadModule failed %d (0x%x)", iRet));
			return false;
		}

		sys_net_initialize_network();

		iRet = cellNetCtlInit();
#endif

		if (iRet < 0) 
		{
			sys_net_finalize_network();
			cellSysmoduleUnloadModule(CELL_SYSMODULE_NET);
			Asura_Assertf(false, ("cellNetCtlInit() failed: %d (0x%x)", iRet, iRet));
			return false;
		}

		iRet = cellNetCtlAddHandler(EventHandler, NULL, &s_iEventHandlerID);
		if(iRet < 0)
		{
			Asura_Assertf(false, ("cellNetCtlAddHandler() failed: %d (0x%x)", iRet, iRet));
		}

		// start the network connection dialog. 
		// This will do the dialog to help the player if the network is down
		CellNetCtlNetStartDialogParam xNetStartParams;
		xNetStartParams.size = sizeof(xNetStartParams);
#ifdef ASURA_SUPPORT_LOBBY
		xNetStartParams.type = CELL_NET_CTL_NETSTART_TYPE_NP;
#else
		xNetStartParams.type = CELL_NET_CTL_NETSTART_TYPE_NET; // if not using NP.
#endif
		xNetStartParams.cid = 0; // unused

		iRet = cellNetCtlNetStartDialogLoadAsync( &xNetStartParams );
		if(iRet < 0)
		{
			Asura_Assertf(false, ("cellNetCtlNetStartDialogLoadAsync() failed: %d (0x%x)", iRet, iRet));
		}

		s_bInitialised = true;
		s_bConnecting = true;
		s_bConnectionError = false;
	}
	return true;
}

void Asura_PS3_Network::Shutdown()
{
	Asura_Assert(s_bInitialised, "Asura_PS3_Network not initialised");
	if (s_bInitialised)
	{
		int iRet = cellNetCtlDelHandler(s_iEventHandlerID);
		if(iRet < 0)
		{
			Asura_Assertf(false, ("cellNetCtlDelHandler() failed: %d (0x%x)", iRet, iRet));
		}
		s_iEventHandlerID = -1;

#ifndef ASURA_USE_GCM_REPLAY		
		cellNetCtlTerm();

		sys_net_finalize_network();
		cellSysmoduleUnloadModule(CELL_SYSMODULE_NET);
#endif

		s_bInitialised = false;
		s_bConnecting = false;
	}
}


bool Asura_PS3_Network::IsConnected()
{
	Asura_Assert(s_bInitialised, "Asura_PS3_Network not initialised");
	if (!s_bInitialised)
	{
		return false;
	}

	int iState;
	int iRet = cellNetCtlGetState(&iState);
	if (iRet < 0) 
	{
		Asura_Assertf(false, ("cellNetCtlGetState() failed(%x)", iRet));
		return false;
	}
	return (iState == CELL_NET_CTL_STATE_IPObtained);
}

void Asura_PS3_Network::SysUtilNetCtlEvent( uint64_t uStatus )
{
	switch (uStatus)
	{
		case CELL_SYSUTIL_NET_CTL_NETSTART_LOADED:
		{
			break;
		}
		case CELL_SYSUTIL_NET_CTL_NETSTART_FINISHED:
		{
			CellNetCtlNetStartDialogResult xResult;
			xResult.size = sizeof(xResult);
			int iRet = cellNetCtlNetStartDialogUnloadAsync( &xResult );
			if(iRet < 0)
			{
				Asura_Assertf(false, ("cellNetCtlNetStartDialogLoadAsync() failed: %d (0x%x)", iRet, iRet));
			}
			ProcessSysUtilNetCtlResult( xResult.result );
			break;
		}
		case CELL_SYSUTIL_NET_CTL_NETSTART_UNLOADED:
		{
			break;
		}
		default:
		{
			Asura_Assert(false, "unhandled status");
			break;
		}
	}
}

void Asura_PS3_Network::EventHandler( int iPrevState, int iNewState, int iEvent, int iErrorCode, void* pArg )
{
	// printf for all events
	Asura_OutputToDebugger_Retail(( "Asura_PS3_Network::EventHandler( %d, %d, %d, %d )\n", iPrevState, iNewState, iEvent, iErrorCode ));

//	CELL_NET_CTL_STATE_Disconnected = 0
//	CELL_NET_CTL_STATE_Connecting = 1
//	CELL_NET_CTL_STATE_IPObtaining = 2
//	CELL_NET_CTL_STATE_IPObtained = 3

	// SDK docs indicate the following 2 cases should be treated as link lost.
	if (iPrevState == CELL_NET_CTL_STATE_IPObtained)
	{
		if ( (iEvent == CELL_NET_CTL_EVENT_ERROR && iNewState == CELL_NET_CTL_STATE_Disconnected) || 
			(iEvent == CELL_NET_CTL_EVENT_LINK_DISCONNECTED && iNewState == CELL_NET_CTL_STATE_Connecting) )
		{
			s_bConnectionError = true;
		}
	}

}


void Asura_PS3_Network::ProcessSysUtilNetCtlResult( int iResult )
{
	s_bConnecting = false;
	if (iResult)
	{
		// failure
		s_bConnectionError = true;
	}
	Asura_OutputToDebugger_Retail(( "ProcessSysUtilNetCtlResult = 0x%x\n", iResult ));
}

// E N D   O F   F I L E /////////////////////////////////////////////////////
