//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_NP.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     Asura wrapper for PSN Network Platform code
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Online/Asura_PS3_NP.h"

// Asura independent
#include "Core/Asura_HeapArray.h"
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Online/Asura_Lobby.h"
#include "Network/Online/Asura_Merit_Types.h"

// Asura PS3
#include "Network/Online/Asura_PS3_NP_Basic.h"
#include "Network/Online/Asura_PS3_NP_TUS.h"

// library headers
#include <cell/sysmodule.h>
#include <np.h>

#ifdef ASURA_SUPPORT_LOBBY
#include "Network/Online/Asura_Lobby.h"
#endif //#ifdef ASURA_SUPPORT_LOBBY

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
ASURA_PS3_NP_STATE Asura_PS3_NP::s_eState;
void* Asura_PS3_NP::s_pPool;
int Asura_PS3_NP::s_iConnectionStatus = SCE_NP_MANAGER_STATUS_OFFLINE;
SceNpId Asura_PS3_NP::s_xNpId;
SceNpMatching2ContextId Asura_PS3_NP::s_xCtxId;
SceNpMatching2ServerId* Asura_PS3_NP::s_pxServerIDs;
SceNpMatching2World Asura_PS3_NP::s_xWorld;
u_int Asura_PS3_NP::s_uNumServers;
u_int Asura_PS3_NP::s_uCurrentServer;

// C O N S T A N T S /////////////////////////////////////////////////////////
const u_int uNUM_MERIT_TUS_VARS = 2;
const ASURA_PS3_TUS_VAR_SLOT aeTUS_VAR_SLOTS[uNUM_MERIT_TUS_VARS] = 
	{ ASURA_PS3_TUS_VAR_SLOT_MERIT_LEVEL, ASURA_PS3_TUS_VAR_SLOT_MERIT_POINTS };

// G L O B A L S /////////////////////////////////////////////////////////////
static Asura_PS3_TUSTransactionManagerGetVars g_xMeritTUSVarManager(uNUM_MERIT_TUS_VARS);

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_PS3_NP::Initialise()
{
	// init static data
	s_eState = ASURA_PS3_NP_STATE_LOGGING_IN;
	s_iConnectionStatus = SCE_NP_MANAGER_STATUS_OFFLINE;
	s_uNumServers = 0;

	int iRet;
	
	iRet = cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_NP2);
	Asura_Assertf( iRet == CELL_OK, ("cellSysmoduleLoadModule failed to load NP2 module (0x%x)", iRet) );

	Asura_Assert(!s_pPool, "Asura_PS3_NP already initialised");
	s_pPool = Asura_Allocate(SCE_NP_MIN_POOL_SIZE);

	iRet = sceNp2Init(SCE_NP_MIN_POOL_SIZE, s_pPool);
	Asura_Assertf( iRet == CELL_OK, ("sceNp2Init failed (0x%x)", iRet) );

	iRet = sceNpManagerRegisterCallback(ConnectionStatusCallback, NULL);
	Asura_Assertf( iRet == CELL_OK, ("sceNpManagerRegisterCallback failed (0x%x)", iRet) );

	iRet = sceNpManagerGetStatus(&s_iConnectionStatus);
	Asura_Assertf( iRet == CELL_OK, ("sceNpManagerGetStatus failed (0x%x)", iRet) );

	if (iRet != CELL_OK)
	{
		Asura_OutputToDebugger_Retail(("Asura_PS3_NP::Initialise() failed\n"));
	}

	Asura_PS3_NP_Basic::Initialise();
}

bool Asura_PS3_NP::UpdateInitialisation()
{
	switch (s_eState)
	{
		case ASURA_PS3_NP_STATE_LOGGING_IN:
		{
			if (s_iConnectionStatus == SCE_NP_MANAGER_STATUS_ONLINE)
			{
				if (InitialiseMatching2())
				{
					s_eState = ASURA_PS3_NP_STATE_GETTING_SERVERINFO;
				}
				else
				{
					s_eState = ASURA_PS3_NP_STATE_ERROR;
				}
			}
			break;
		}
		
		case ASURA_PS3_NP_STATE_GETTING_SERVERINFO:
		{
			UpdateGettingServerInfo();
			break;
		}

		case ASURA_PS3_NP_STATE_GETTING_WORLD:
		{
			UpdateGettingWorld();
			break;
		}

		case ASURA_PS3_NP_STATE_GETTING_MERIT:
		{
			UpdateGettingMerit();
			break;
		}

		case ASURA_PS3_NP_STATE_GETTING_GAMECOMPLETION:
		{
			UpdateGettingGameCompletion();
			break;
		}

		default:
		{
			Asura_Assert(false, "unexpected state");
			break;
		}
	}

	// return true if we're finished. Finished means error of some sort (or managed to end up in offline state) or we're online.
	return s_eState == ASURA_PS3_NP_STATE_ONLINE || s_eState == ASURA_PS3_NP_STATE_OFFLINE || s_eState == ASURA_PS3_NP_STATE_ERROR;
}

void Asura_PS3_NP::Update()
{
	Asura_PS3_NP_Basic::Update();
}

void Asura_PS3_NP::Shutdown()
{
	if (s_pPool)
	{
		Asura_PS3_NP_Basic::Shutdown();

		ShutdownMatching2();

		sceNpManagerUnregisterCallback();
		sceNp2Term();
		Asura_Deallocate(s_pPool);
		s_pPool = NULL;
		cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_NP2);

		delete[] s_pxServerIDs;
		s_pxServerIDs = NULL;
	}

	s_iConnectionStatus = SCE_NP_MANAGER_STATUS_OFFLINE;
}

bool Asura_PS3_NP::InitialiseMatching2()
{
	int iRet = sceNpMatching2Init(0, 0);
	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2Init failed (0x%x)", iRet) );

	iRet = sceNpManagerGetNpId(&s_xNpId);
	Asura_Assertf( iRet == CELL_OK, ("sceNpManagerGetNpId failed (0x%x)", iRet) );

	int iOption = (SCE_NP_MATCHING2_CONTEXT_OPTION_USE_ONLINENAME |
		SCE_NP_MATCHING2_CONTEXT_OPTION_USE_AVATARURL);

	// Create a context
	iRet = sceNpMatching2CreateContext(
		&s_xNpId, &s_xProject_CommunicationId, &s_xProject_PassPhrase, &s_xCtxId, iOption);
	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2CreateContext failed (0x%x)", iRet) );

	iRet = sceNpMatching2RegisterContextCallback(s_xCtxId, ContextCallback, NULL);
	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2RegisterContextCallback failed (0x%x)", iRet) );

	// Start the context
	iRet = sceNpMatching2ContextStart(s_xCtxId);
	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2ContextStart failed (0x%x)", iRet) );

	// Get Server ID List
	iRet = sceNpMatching2GetServerIdListLocal(s_xCtxId, NULL, 0);
	Asura_Assertf( iRet > 0, ("sceNpMatching2GetServerIdListLocal failed (0x%x)", iRet) );
	// iRet is the number of servers
	if (iRet > 0) 
	{
		// allocate memory for list of servers and get their IDs
		Asura_OutputToDebugger_Retail(("sceNpMatching2GetServerIdListLocal reports %i servers\n", iRet));
		s_pxServerIDs = new SceNpMatching2ServerId[iRet];
		s_uNumServers = static_cast<u_int>(iRet);
		s_uCurrentServer = 0;
	
		iRet = sceNpMatching2GetServerIdListLocal(s_xCtxId, s_pxServerIDs, s_uNumServers);
		Asura_Assertf( iRet > 0, ("sceNpMatching2GetServerIdListLocal failed (0x%x)", iRet) );
	}

	return (iRet > 0);
}

void Asura_PS3_NP::ShutdownMatching2()
{
	// Stop the context
	int iRet = sceNpMatching2ContextStop(s_xCtxId);
	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2ContextStop failed (0x%x)", iRet) );

	// Destroy the context
	iRet = sceNpMatching2DestroyContext(s_xCtxId);
	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2DestroyContext failed (0x%x)", iRet) );

	s_xCtxId = 0;

	// Terminate the NP matching 2 utility
	sceNpMatching2Term();
}


void Asura_PS3_NP::ConnectionStatusCallback(int iEvent, int iResult, void* pArg )
{
	Asura_OutputToDebugger_Retail(("Asura_PS3_NP::ConnectionStatusCallback(), iEvent = %d, iResult = 0x%x\n", iEvent, iResult));
	s_iConnectionStatus = iEvent;

	if (iEvent == SCE_NP_MANAGER_STATUS_OFFLINE)
	{
		s_eState = ASURA_PS3_NP_STATE_ERROR;
	}
}

void Asura_PS3_NP::ContextCallback(
						 SceNpMatching2ContextId  xCtxId,
						 SceNpMatching2Event      usEvent,
						 SceNpMatching2EventCause ucEventCause,
						 int iErrorCode,
						 void *arg
						 )
{
	Asura_OutputToDebugger_Retail(("[ %s ]\n", __FUNCTION__));

	Asura_OutputToDebugger_Retail(("Asura_PS3_NP::ContextCallback, ctxId = %d, event = %d, cause = %d, error = %d\n", 
		xCtxId, usEvent, ucEventCause, iErrorCode));

	Asura_Assert(usEvent != SCE_NP_MATCHING2_CONTEXT_EVENT_StartOver, "NP Context no longer valid");
	if (usEvent == SCE_NP_MATCHING2_CONTEXT_EVENT_StartOver)
	{
		s_eState = ASURA_PS3_NP_STATE_ERROR;
	}
}


// This data is initialised by the main thread and after that should not be touched 
// by that thread. The callback thread will modify it and when m_bDone is set to true
// the main thread can use it again.
static struct Asura_PS3_GetServerInfo_CallbackData {
	SceNpMatching2ServerStatus m_xStatus;
	int m_iErrorCode;
	bool m_bDone;
	bool m_bInProgress;
	bool m_bError;
} g_xGetServerInfoCallbackData;

static struct Asura_PS3_GetWorlds_CallbackData {
	bool m_bDone;
	bool m_bError;
} g_xGetWorldsCallbackData;

static u_int g_uReqSeqNum;
static SceNpMatching2RequestId g_xAssignedReqId;

void Asura_PS3_NP::UpdateGettingServerInfo()
{
	int iRet = 0;
	SceNpMatching2RequestOptParam xOptParam;
	SceNpMatching2GetServerInfoRequest xReqParam;

	if (!g_xGetServerInfoCallbackData.m_bInProgress) 
	{
		Asura_Assert(s_pxServerIDs && s_uCurrentServer < s_uNumServers, "invalid state");
		if (!s_pxServerIDs || s_uCurrentServer >= s_uNumServers)
		{
			return;
		}

		Asura_Mem::Set(&g_xGetServerInfoCallbackData, 0, sizeof(g_xGetServerInfoCallbackData));
		g_xGetServerInfoCallbackData.m_bDone = false;
		g_xGetServerInfoCallbackData.m_bInProgress = true;

		Asura_Mem::Set(&xReqParam, 0, sizeof(xReqParam));
		xReqParam.serverId = s_pxServerIDs[s_uCurrentServer];

		Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
		xOptParam.cbFunc = GetServerInfoCallback;
		xOptParam.cbFuncArg = NULL;
		xOptParam.timeout = 0; // default
		xOptParam.appReqId = ++g_uReqSeqNum;

		iRet = sceNpMatching2GetServerInfo(
			s_xCtxId,
			&xReqParam,
			&xOptParam,
			&g_xAssignedReqId);
		Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2GetServerInfo failed (0x%x)", iRet) );
	}
	else if (g_xGetServerInfoCallbackData.m_bDone)
	{
		if (g_xGetServerInfoCallbackData.m_iErrorCode >= 0) 
		{
			if (g_xGetServerInfoCallbackData.m_xStatus == SCE_NP_MATCHING2_SERVER_STATUS_AVAILABLE) 
			{
				// now we have a server, find the world ID. Assuming just one world for Asura games for now.
				if (GetWorlds())
				{
					s_eState = ASURA_PS3_NP_STATE_GETTING_WORLD;
				}
				else
				{
					s_eState = ASURA_PS3_NP_STATE_ERROR;
				}
			}
			else 
			{
				s_uCurrentServer++;
				if (s_uCurrentServer >= s_uNumServers) 
				{
					s_uCurrentServer = 0;
					Asura_Assert(false, "no NP server available");

					s_eState = ASURA_PS3_NP_STATE_ERROR;
				}
				else
				{
					Asura_OutputToDebugger_Retail(("challenge next server\n"));
				}
			}
		}
		else
		{
			s_eState = ASURA_PS3_NP_STATE_ERROR;
		}
		g_xGetServerInfoCallbackData.m_bInProgress = false;
	}
}


// Callback function will be executed on a different thread to the main one.
void Asura_PS3_NP::GetServerInfoCallback(SceNpMatching2ContextId xCtxId, 
										 SceNpMatching2RequestId xReqId, 
										 SceNpMatching2Event usEvent,
										 SceNpMatching2EventKey xEventKey,
										 int iErrorCode, 
										 size_t uDataSize, 
										 void* pArg)
{		
	int iRet = 0;

	if (xReqId != g_xAssignedReqId)
	{
		Asura_Assert(false, "unexpected server info callback");
		return;
	}

	if (iErrorCode < 0) 
	{
		Asura_Assertf(false, ("server info callback error 0x%x", iErrorCode));
		g_xGetServerInfoCallbackData.m_iErrorCode = iErrorCode;
		g_xGetServerInfoCallbackData.m_bDone = true;
		g_xGetServerInfoCallbackData.m_bError = true;
		return;
	}

	SceNpMatching2GetServerInfoResponse xRespData;
	Asura_Assert(sizeof(xRespData) == uDataSize, "unexpected data size");
	Asura_Mem::Set(&xRespData, 0, sizeof(xRespData));
	iRet = sceNpMatching2GetEventData( xCtxId, xEventKey, &xRespData, sizeof(xRespData));

	Asura_Assertf( iRet >= 0, ("sceNpMatching2GetEventData failed (0x%x)", iRet) );
	if (iRet < 0) 
	{
		g_xGetServerInfoCallbackData.m_iErrorCode = iRet;
		g_xGetServerInfoCallbackData.m_bDone = true;
		g_xGetServerInfoCallbackData.m_bError = true;
		return;
	}

	g_xGetServerInfoCallbackData.m_xStatus = xRespData.server.status;
	g_xGetServerInfoCallbackData.m_bDone = true;

	return;
}

bool Asura_PS3_NP::GetWorlds()
{
	SceNpMatching2RequestOptParam xOptParam;
	SceNpMatching2GetWorldInfoListRequest xReqParam;

	Asura_Mem::Set(&g_xGetWorldsCallbackData, 0, sizeof(g_xGetWorldsCallbackData));

	Asura_Mem::Set(&xReqParam, 0, sizeof(xReqParam));
	Asura_Assert(s_pxServerIDs && s_uCurrentServer < s_uNumServers, "invalid state");
	xReqParam.serverId = s_pxServerIDs[s_uCurrentServer];

	Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
	xOptParam.cbFunc = GetWorldInfoCallback;
	xOptParam.cbFuncArg = NULL;
	xOptParam.timeout = 0; // default
	xOptParam.appReqId = ++g_uReqSeqNum;

	int iRet = sceNpMatching2GetWorldInfoList(
		s_xCtxId,
		&xReqParam,
		&xOptParam,
		&g_xAssignedReqId);

	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2GetWorldInfoList failed (0x%x)", iRet) );
	return iRet == CELL_OK;
}

void Asura_PS3_NP::UpdateGettingWorld()
{
	if (g_xGetWorldsCallbackData.m_bDone)
	{
		if (g_xGetWorldsCallbackData.m_bError)
		{
			s_eState = ASURA_PS3_NP_STATE_ERROR;
		}
		else
		{
			if (GetMerit())
			{
				s_eState = ASURA_PS3_NP_STATE_GETTING_MERIT;
			}
			else
			{
				s_eState = ASURA_PS3_NP_STATE_ERROR;
			}
		}
	}
}

bool Asura_PS3_NP::GetMerit()
{
	// start the process of getting the merit.
#ifdef ASURA_USE_MERIT_SYSTEM
	Asura_PS3_NP_TUS::Initialise();
	const int iRet = g_xMeritTUSVarManager.GetVarsAsync(s_xNpId, uNUM_MERIT_TUS_VARS, aeTUS_VAR_SLOTS);
	return iRet == CELL_OK;
#else
	return true;
#endif
}

bool Asura_PS3_NP::UpdateGettingMerit()
{
#ifdef ASURA_USE_MERIT_SYSTEM
	bool bDone = g_xMeritTUSVarManager.Update();
	if (bDone)
	{
		bool bError = false;
		if (g_xMeritTUSVarManager.GetOperationResult() != 
			static_cast<int>(SCE_NP_COMMUNITY_SERVER_ERROR_USER_STORAGE_DATA_NOT_FOUND))
		{
			Asura_Assertf(!g_xMeritTUSVarManager.HasErrorOccurred(), 
				("error getting merit from TUS: 0x%x", g_xMeritTUSVarManager.GetOperationResult()));
			bError = g_xMeritTUSVarManager.HasErrorOccurred();
		}
		
		Asura_PS3_NP_TUS::Shutdown();

		if (bError)
		{
			s_eState = ASURA_PS3_NP_STATE_ERROR;
		}
		else
		{
			Asura_Assert(aeTUS_VAR_SLOTS[0] == ASURA_PS3_TUS_VAR_SLOT_MERIT_LEVEL, "level/points mixup");
			Asura_Assert(aeTUS_VAR_SLOTS[1] == ASURA_PS3_TUS_VAR_SLOT_MERIT_POINTS, "level/points mixup");
			u_int uMeritLevel = static_cast<u_int>(g_xMeritTUSVarManager.GetVarValue(0));
			int iMeritPoints = static_cast<int>(g_xMeritTUSVarManager.GetVarValue(1));

			Asura_OutputToDebugger_Retail(("retrieved merit: level = %u, points = %d\n", 
				uMeritLevel, iMeritPoints));
			Asura_Assert(uMeritLevel < uASURA_MERIT_NUM_LEVELS, "invalid merit level");
			Asura_Assert(iMeritPoints < iASURA_MERIT_MAX_MERIT_POINTS, "invalid merit points");
			// clamp
			uMeritLevel = Asura_Maths::ClampToRange(uMeritLevel, 0u, uASURA_MERIT_NUM_LEVELS - 1);
			iMeritPoints = Asura_Maths::ClampToRange(iMeritPoints, 0, iASURA_MERIT_MAX_MERIT_POINTS);

			Asura_Lobby_System::SetLocalPlayerMerit(uMeritLevel, iMeritPoints);

			// Start getting game completion data. Ignore any errors. There is no need to break
			// the whole MP game just because we couldn't read this data.
			GetGameCompletion();
			s_eState = ASURA_PS3_NP_STATE_GETTING_GAMECOMPLETION;
		}
	}
	return bDone;
#else
	GetGameCompletion();
	s_eState = ASURA_PS3_NP_STATE_GETTING_GAMECOMPLETION;
	
	return true;
#endif
}

// Return false if an error occurred.
bool Asura_PS3_NP::GetGameCompletion()
{
#ifdef ASURA_SUPPORT_LOBBY
	Asura_Lobby_System::GetGameCompletionHistory().Platform_GetData_Start();
	return Asura_Lobby_System::GetGameCompletionHistory().IsGettingData();
#else
	return true;
#endif //#ifdef ASURA_SUPPORT_LOBBY
}

// Return false if we are waiting to get the data.
bool Asura_PS3_NP::UpdateGettingGameCompletion()
{
#ifdef ASURA_SUPPORT_LOBBY
	const bool bDone = Asura_Lobby_System::GetGameCompletionHistory().Platform_GetData_Update();
	if (bDone)
	{
		// Ignore any errors. If we failed to get the game completion data then we'll just
		// default everything to zero, and not write back the stats to the server.
		// There is no need to break the whole MP game just because we couldn't read this data.
		s_eState = ASURA_PS3_NP_STATE_ONLINE;
	}
	return bDone;
#else
	s_eState = ASURA_PS3_NP_STATE_ONLINE;
	return true;
#endif //#ifdef ASURA_SUPPORT_LOBBY
}

// Callback function will be executed on a different thread to the main one.
void Asura_PS3_NP::GetWorldInfoCallback(SceNpMatching2ContextId xCtxId,
								 SceNpMatching2RequestId xReqId, SceNpMatching2Event usEvent,
								 SceNpMatching2EventKey xEventKey, int iErrorCode, size_t uDataSize, void* pArg)
{

	if (xReqId != g_xAssignedReqId)
	{
		Asura_Assert(false, "unexpected server info callback");
		g_xGetWorldsCallbackData.m_bError = true;
		g_xGetWorldsCallbackData.m_bDone = true;
		return;
	}

	if (iErrorCode < 0) 
	{
		Asura_Assertf(false, ("server info callback error 0x%x", iErrorCode));
		g_xGetWorldsCallbackData.m_bError = true;
		g_xGetWorldsCallbackData.m_bDone = true;
		return;
	}

	// using HeapArray to create a temporary buffer that is automatically 
	// deleted when it goes out of scope.
	Asura_HeapArray<u_char> xBuffer;
	xBuffer.SetSize(uDataSize);

	SceNpMatching2GetWorldInfoListResponse* pxResponseData = reinterpret_cast
		<SceNpMatching2GetWorldInfoListResponse*>(&xBuffer[0]);
	if (!pxResponseData) 
	{
		g_xGetWorldsCallbackData.m_bError = true;
		g_xGetWorldsCallbackData.m_bDone = true;
		return;
	}

	Asura_Mem::Set(pxResponseData, 0, uDataSize);

	int iRet = sceNpMatching2GetEventData(xCtxId, xEventKey, pxResponseData, uDataSize);
	Asura_Assertf( iRet >= 0, ("sceNpMatching2GetEventData failed (0x%x)", iRet) );
	if (iRet < 0 || pxResponseData->worldNum < 1) 
	{
		Asura_Assert(pxResponseData->worldNum > 0, "no worlds");
		g_xGetWorldsCallbackData.m_bError = true;
		g_xGetWorldsCallbackData.m_bDone = true;
		return;
	}	
	else
	{
		Asura_Mem::Copy(&s_xWorld, &pxResponseData->world[0], sizeof(SceNpMatching2World));
	}	

	g_xGetWorldsCallbackData.m_bDone = true;
}
