//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_NP_RoomFinder.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     class to manage searching for PSN Rooms. This class is intended to be
//     completely independent of Asura_Lobby.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Asura_PS3_NP_RoomFinder.h"

// Asura independent headers
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"

// Asura PS3 headers
#include "Network/Online/Asura_PS3_NP.h"

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
u_int Asura_PS3_NP_RoomFinder::s_uRequestSequenceNumber;
Asura_PS3_SearchRoom_CallbackData Asura_PS3_NP_RoomFinder::s_xCallbackData;
Asura_PS3_NP_RoomFinder* Asura_PS3_NP_RoomFinder::s_pxThis;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
Asura_PS3_NP_RoomFinder::Asura_PS3_NP_RoomFinder( u_int uMaxResults )
{
	m_uMaxResults = uMaxResults;
	m_xAssignedReqID = 0;
	m_pxResponseData = NULL;

	Asura_Assert(!s_pxThis, "only one instance currently supported");
	s_pxThis = this;
}

Asura_PS3_NP_RoomFinder::~Asura_PS3_NP_RoomFinder()
{
	Asura_Deallocate(m_pxResponseData);

	Asura_Assert(s_pxThis == this, "only one instance currently supported");
	s_pxThis = NULL;
}

bool Asura_PS3_NP_RoomFinder::BeginSearch(const SceNpMatching2IntSearchFilter* pxSearchFilters, 
										  u_int uNumSearchFilters, 
										  u_int uNumAttributesToFetch)
{
	if (m_pxResponseData)
	{
		Asura_Assert(false, "only expected one search per instanced RoomFinder");
		Asura_Deallocate(m_pxResponseData);
		m_pxResponseData = NULL;
	}

	// init callback data
	s_xCallbackData.m_bError = false;
	s_xCallbackData.m_bDone = false;

	SceNpMatching2SearchRoomRequest xReqParam;
	SceNpMatching2RequestOptParam xOptParam;
	SceNpMatching2AttributeId axAttrId[SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_NUM];
	axAttrId[0] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_1_ID;
	axAttrId[1] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_2_ID;
	axAttrId[2] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_3_ID;
	axAttrId[3] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_4_ID;
	axAttrId[4] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_5_ID;
	axAttrId[5] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_6_ID;
	axAttrId[6] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_7_ID;
	axAttrId[7] = SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_8_ID;
	ASURA_STATIC_ASSERT(SCE_NP_MATCHING2_ROOM_SEARCHABLE_INT_ATTR_EXTERNAL_NUM == 8);
	
	//E Request parameters
	memset(&xReqParam, 0, sizeof(xReqParam));
	xReqParam.worldId = Asura_PS3_NP::GetWorldID();
	xReqParam.lobbyId = 0;
	xReqParam.rangeFilter.startIndex =
		SCE_NP_MATCHING2_RANGE_FILTER_START_INDEX_MIN;
	xReqParam.rangeFilter.max = m_uMaxResults;
	// set the flag filters to avoid getting rooms which are full or closed.
	xReqParam.flagFilter = (SCE_NP_MATCHING2_ROOM_FLAG_ATTR_CLOSED | SCE_NP_MATCHING2_ROOM_FLAG_ATTR_FULL);
	xReqParam.flagAttr = 0;
	xReqParam.intFilter = uNumSearchFilters ? const_cast<SceNpMatching2IntSearchFilter*>(pxSearchFilters) : NULL;	// library is not const-correct
	xReqParam.intFilterNum = uNumSearchFilters;
	xReqParam.binFilter = NULL;
	xReqParam.binFilterNum = 0;
	xReqParam.attrId = axAttrId;
	xReqParam.attrIdNum = uNumAttributesToFetch;
	//E Get only the NP ID of the room owner
	xReqParam.option = SCE_NP_MATCHING2_SEARCH_ROOM_OPTION_WITH_NPID | SCE_NP_MATCHING2_SEARCH_ROOM_OPTION_WITH_ONLINENAME;

	Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
	xOptParam.cbFunc = SearchRoomCallback;
	xOptParam.cbFuncArg = NULL;
	xOptParam.timeout = 0; // default
	xOptParam.appReqId = ++s_uRequestSequenceNumber;

	const int iRet = sceNpMatching2SearchRoom(Asura_PS3_NP::GetContextID(), 
		&xReqParam, &xOptParam, &m_xAssignedReqID);
	if (iRet < 0) 
	{
		//E Error handling
		return false;
	}
	return true;
}


void Asura_PS3_NP_RoomFinder::SearchRoomCallback(SceNpMatching2ContextId xCtxId, 
						SceNpMatching2RequestId xReqId, 
						SceNpMatching2Event usEvent,
						SceNpMatching2EventKey xEventKey,
						int iErrorCode, 
						size_t uDataSize, 
						void* pArg)
{
	Asura_Assert(s_pxThis, "no room finder object");
	if (!s_pxThis)
	{
		return;
	}

	if (xReqId != s_pxThis->m_xAssignedReqID)
	{
		Asura_Assert(false, "unexpected search room callback");
		s_xCallbackData.m_bError = true;
		s_xCallbackData.m_bDone = true;
		return;
	}

	if (iErrorCode < 0) 
	{
		Asura_Assertf(false, ("search room callback error 0x%x", iErrorCode));
		s_xCallbackData.m_bError = true;
		s_xCallbackData.m_bDone = true;
		return;
	}

	s_pxThis->m_pxResponseData = reinterpret_cast <SceNpMatching2SearchRoomResponse*>(Asura_Allocate(uDataSize));
	if (!s_pxThis->m_pxResponseData)
	{
		s_xCallbackData.m_bError = true;
		s_xCallbackData.m_bDone = true;
		return;
	}

	if (!s_pxThis->m_pxResponseData) 
	{
		s_xCallbackData.m_bError = true;
		s_xCallbackData.m_bDone = true;
		return;
	}

	Asura_Mem::Set(s_pxThis->m_pxResponseData, 0, uDataSize);

	int iRet = sceNpMatching2GetEventData(xCtxId, xEventKey, s_pxThis->m_pxResponseData, uDataSize);
	Asura_Assertf( iRet >= 0, ("sceNpMatching2GetEventData failed (0x%x)", iRet) );
	if (iRet < 0) 
	{
		Asura_Assert(false, "search room failed");
		s_xCallbackData.m_bError = true;
		s_xCallbackData.m_bDone = true;
		return;
	}	

	s_xCallbackData.m_bDone = true;
}


u_int Asura_PS3_NP_RoomFinder::GetNumSearchResults() const
{
	if (m_pxResponseData)
	{
		return m_pxResponseData->range.size;
	}
	return 0;
}


const SceNpMatching2RoomDataExternal& Asura_PS3_NP_RoomFinder::GetResult( u_int uIndex ) const
{
	Asura_Assert(s_xCallbackData.m_bDone && !s_xCallbackData.m_bError, "results not valid");
	Asura_Assert(m_pxResponseData, "no data");
	Asura_Assert(uIndex < m_pxResponseData->range.size, "invalid result index" );
	const SceNpMatching2RoomDataExternal* pxResult = m_pxResponseData->roomDataExternal;
	u_int uCurrentIndex = 0;
	while (uCurrentIndex < uIndex)
	{
		++uCurrentIndex;
		pxResult = pxResult->next;
	}
	// check the linked list is terminated as expected.
	Asura_Assert(uIndex != (m_pxResponseData->range.size - 1) || !pxResult->next, "linked list error");
	Asura_Assert(pxResult, "invalid room search result");
	return *pxResult;
}
