//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_SPUGroupController.h
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     A helper class for dealing with a SPU thread group where data is
//     passed from PPU to SPU via a lock-free queue and all the threads are
//     grabbing work from it as fast as they can.
//
//     Pseudo-code for the work per frame is:
//
//         Begin()
//         queue = GetQueue()
//         while( job )
//             cellSyncLFQueuePush( queue, job )
//         End()
//
//     The call to Begin() can set ELF 'command line' arguments as required.
//     If not specified these will default to:
//
//         arg1: thread num (32-bits)
//         arg2: lfqueue addr
//         arg3: 0
//         arg4: 0
//
//     For debugging / profiling purposes the number of threads used can be
//     dynamically altered via SetTempThreadLimit(). Do not modify this limit
//     while the threads are running. Passing 0 resets the limit.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_PS3_SPUGROUPCONTROLLER_H___
#define ASURA_PS3_SPUGROUPCONTROLLER_H___ 1

// C L A S S E S /////////////////////////////////////////////////////////////

struct CellSyncLFQueue;
struct UtilSpuThreadGroupInfo;

class Asura_PS3_SPUGroupController
{

public:

    enum
    {
        PRIORITY_HIGH = 64,  // Where work needs to be completed ASAP, probably PPU is waiting on it.
        PRIORITY_LOW  = 199, // Work to be done in the background while PPU gets on with other stuff and maybe other SPU groups jump in temporarily.

        MAX_THREADS = 5,     // As many SPUs as can be occupied simultaneously, taking into account necessarily resident stuff like the audio code.
    };

    Asura_PS3_SPUGroupController( const char* szGroupName,
                                  const void* pvELFAddr,
                                  u_int uThreadCount,
                                  u_int uGroupPriority,
                                  u_int uQueueEntrySize, u_int uMaxQueueEntries,
                                  const char* const * pszThreadNames = NULL,
                                  bool bSuspendGroup = false );
    ~Asura_PS3_SPUGroupController( void );

    void SetTerminatorJob( const void* pvTerminatorJob ) { m_pvTerminatorJob = pvTerminatorJob; }
    void SetTempThreadLimit( u_int uTempThreadLimit = 0 ) { m_uTempThreadLimit = uTempThreadLimit; }

    void Begin( uint64_t* puArg1 = NULL,
                uint64_t* puArg2 = NULL,
                uint64_t* puArg3 = NULL,
                uint64_t* puArg4 = NULL );
    CellSyncLFQueue* GetQueue( void ) { return m_pxQueue; }
	void PushTerminators( void );
    void End( void );
	
private:

    UtilSpuThreadGroupInfo* m_pxThreadGroupInfo;
    CellSyncLFQueue*        m_pxQueue;
    void*                   m_pvQueueBfr;
    const void*             m_pvTerminatorJob;
    u_int                   m_uTempThreadLimit;
    bool					m_bQueueTerminated;
    bool                    m_bHasEverBeenStarted;
    bool                    m_bSuspendGroup;

};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ASURA_PS3_SPUGROUPCONTROLLER_H___
