/* SCE CONFIDENTIAL
PLAYSTATION(R)3 Programmer Tool Runtime Library 210.001
* Copyright (C) 2007 Sony Computer Entertainment Inc.
* All Rights Reserved.
*/


#include "spu_thread_group_utils.h"
#include <string.h>
#include <stdlib.h>

int utilInitializeSpuThread(UtilSpuThreadInfo *spuThreadInfo, 	
							const char *name,
							sys_addr_t elf,
							const uint64_t arg[4],
							uint32_t option	)
{
	int ret;
	ret = sys_spu_image_import(&spuThreadInfo->thread_image, (const void*)elf, SYS_SPU_IMAGE_DIRECT);
	if (ret != CELL_OK) {
		CELL_SRE_FPRINTF(stderr, "sys_spu_image_import failed: %d\n", ret);
		return ret;
	}
	
	/* SPU thread attribute */
	spuThreadInfo->threadAttr.name   = name;
	spuThreadInfo->threadAttr.nsize  = strlen(name) + 1;
	spuThreadInfo->threadAttr.option =  option;
	
	spuThreadInfo->arg.arg1 = arg[0];
	spuThreadInfo->arg.arg2 = arg[1];
	spuThreadInfo->arg.arg3 = arg[2];
	spuThreadInfo->arg.arg4 = arg[3];
	return 0;	
}


int utilInitializeSpuThreadGroup(UtilSpuThreadGroupInfo *groupInfo, 
								 const char *group_name,
								 int priority,
								 int nThreads,
								 UtilSpuThreadInfo *spuThreadInfos)
{
	int ret;
	
	CELL_SRE_PRINTF("Creating an SPU thread group.\n");
	
	/* SPU thread group attribute */
	sys_spu_thread_group_attribute_t attrGroup;
	attrGroup.name  = group_name;
	attrGroup.nsize = strlen(attrGroup.name) + 1; 
	attrGroup.type  = SYS_SPU_THREAD_GROUP_TYPE_NORMAL;
	ret = sys_spu_thread_group_create(&groupInfo->group_id, 
									  nThreads, priority, &attrGroup);
	if (ret != CELL_OK) {
		CELL_SRE_FPRINTF(stderr, "sys_spu_thread_group_create failed: %d\n", ret);
		return ret;
	}
	
	groupInfo->threads = spuThreadInfos;

    int id;
	for ( id = 0; id < nThreads; id++) {
		CELL_SRE_PRINTF("Initializing SPU thread %d\n", id);
		
		ret = sys_spu_thread_initialize(&groupInfo->thread_id[id],
										groupInfo->group_id, 
										id, 
										&spuThreadInfos[id].thread_image, 
										&spuThreadInfos[id].threadAttr, 
										&spuThreadInfos[id].arg);
		if (ret != CELL_OK) {
			CELL_SRE_FPRINTF(stderr, "sys_spu_thread_initialize failed: %d\n", ret);
			return ret;
		}
	}

	CELL_SRE_PRINTF("All SPU threads have been successfully initialized.\n");
	
	groupInfo->nThreads = nThreads;

	return CELL_OK;
	

}


int utilStartSpuThreadGroup(const UtilSpuThreadGroupInfo *info)
{
	int ret;
	CELL_SRE_PRINTF("Starting the SPU thread group.\n");
	ret = sys_spu_thread_group_start(info->group_id);
	if (ret != CELL_OK) {
		CELL_SRE_FPRINTF(stderr, "sys_spu_thread_group_start failed: group id = %d, ret = %d\n", 
						 info->group_id, ret);
		return ret;
	}
	
	return CELL_OK;
	

}
int utilWaitSpuThreadGroup(const UtilSpuThreadGroupInfo *info,
						   UtilSpuThreadGroupStatus* status)
{
	int ret;
    int i;

	ret = sys_spu_thread_group_join(info->group_id, &status->cause, &status->status);
	if (ret != CELL_OK) {
		CELL_SRE_FPRINTF(stderr, "sys_spu_thread_group_join failed: %d\n", ret);
		return ret;
	}

	switch(status->cause) {
	case SYS_SPU_THREAD_GROUP_JOIN_GROUP_EXIT:
		CELL_SRE_PRINTF("The SPU thread group exited by sys_spu_thread_group_exit().\n"
						"The group's exit status = %d\n", status->status);
		break;
	case SYS_SPU_THREAD_GROUP_JOIN_ALL_THREADS_EXIT:
		status->nThreads = info->nThreads;
		CELL_SRE_PRINTF("All SPU threads exited by sys_spu_thread_exit().\n");
		for ( i = 0; i < info->nThreads; i++) {
			ret = sys_spu_thread_get_exit_status(info->thread_id[i], &status->threadStatus[i]);
			if (ret != CELL_OK) {
				fprintf(stderr, "sys_spu_thread_get_exit_status failed:%d\n", ret);
			}
			printf("SPU thread %d's exit status = %d\n", i, status->threadStatus[i]);
		}
		break;
	case SYS_SPU_THREAD_GROUP_JOIN_TERMINATED:
		CELL_SRE_PRINTF("The SPU thread group was terminated by sys_spu_thread_terminate().\n"
						"The group's exit status = %d\n", status->status);
		break;
	default:
		CELL_SRE_FPRINTF(stderr,
						 "Unknown exit cause: %d\n", status->cause);
		break;
	}
	return ret;
	

}

int utilFinalizeSpuThread(UtilSpuThreadInfo *info)
{
	int ret;
	ret = sys_spu_image_close(&info->thread_image);
	if (ret != CELL_OK) {
		fprintf(stderr, "sys_spu_image_close failed: %d\n", ret);
		return ret;
	}
	return CELL_OK;
	
}
int utilFinalizeSpuThreadGroup(UtilSpuThreadGroupInfo *info)
{
	int ret;
	/*
	 * Destroy the SPU thread group and clean up resources.
	 */
	ret = sys_spu_thread_group_destroy(info->group_id);
	if (ret != CELL_OK) {
		fprintf(stderr, "sys_spu_thread_group_destroy failed: %d\n", ret);
		return ret;
	}
	return CELL_OK;

}


int utilInitializeSpuThreadGroupAll(UtilSpuThreadGroupInfo *info, 
									UtilSpuThreadGroupArg  *arg)
{
	int ret;
    int i;

	UtilSpuThreadInfo *thread_infos = (UtilSpuThreadInfo*)malloc(sizeof(UtilSpuThreadInfo) * arg->nThreads);
	for( i=0; i<arg->nThreads; i++){
		ret = utilInitializeSpuThread(&thread_infos[i],
									  arg->thread[i].name,
									  arg->thread[i].elf,
									  arg->thread[i].arg,
									  arg->thread[i].option);
		if(ret != CELL_OK){
			fprintf(stderr, "utilInitializeSpuThread failed: %d\n", ret);
			return ret;
		}
	}
	
	ret = utilInitializeSpuThreadGroup(info, 
									   arg->group_name,
									   arg->priority,
									   arg->nThreads,
									   thread_infos);
	if(ret != CELL_OK){
		fprintf(stderr, "utilInitializeSpuThread failed: %d\n", ret);
		return ret;
	}
	return CELL_OK;
}


int utilFinalizeSpuThreadGroupAll(UtilSpuThreadGroupInfo *info)
{
	int ret;
    int i;

	for( i=0; i<info->nThreads; i++){	
		ret = utilFinalizeSpuThread(&info->threads[i]);
		if(ret != CELL_OK){
			fprintf(stderr, "utilFinalizeSpuThread failed: %d\n", ret);
			return ret;
		}
	}
	free(info->threads);
	ret = utilFinalizeSpuThreadGroup(info);
	if(ret != CELL_OK){
		fprintf(stderr, "utilFinalizeSpuThreadGroup failed: %d\n", ret);
		return ret;
	}
	return CELL_OK;
	
}
