//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_CharacterRenderer.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS3_CharacterRenderer.h"

// Engine independent.
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Animation/Asura_TextureAnimation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_SortedRender.h"

// Engine platform-specific.
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS3.h"
#include "Rendering/Asura_PS3_Fog.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_MaterialResponse_System.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"
#include "Rendering/Asura_PS3_TextureManagement.h"
#include "Rendering/Asura_PS3_SPU_Skinning.h"
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Rendering/Shaders/Asura_PS3_VertexShader.h"

//Shaders definition
#include "Rendering/ShadersDefinition/FP/ClearAlpha_FP.h"
#include "Rendering/ShadersDefinition/FP/SkinDebug_FP.h"
#include "Rendering/ShadersDefinition/FP/SmoothSkin_Forward_FP.h"
#include "Rendering/ShadersDefinition/FP/SmoothSkin_FP.h"
#include "Rendering/ShadersDefinition/VD/SmoothSkin_VD.h"
#include "Rendering/ShadersDefinition/VD/SmoothSkinShadow_VD.h"
#include "Rendering/ShadersDefinition/VD/SmoothSkin_SpuSkinning_VD.h"
#include "Rendering/ShadersDefinition/VD/SmoothSkin_StraightThrough_VD.h"
#include "Rendering/ShadersDefinition/VP/SmoothSkin_VP.h"
#include "Rendering/ShadersDefinition/VP/SmoothSkinShadow_VP.h"
#include "Rendering/ShadersDefinition/VP/SmoothSkin_SpuSkinning_VP.h"
#include "Rendering/ShadersDefinition/VP/SmoothSkin_StraightThrough_VP.h"

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR bool dbg_bTestDamageMaps = false;

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Asura_TemplateHashTree< Asura_PS3_CharacterRenderer_Skin > g_xCharacterTree;

static Asura_TemplateHashTree<Asura_PS3_CharacterRenderer_Skin*> g_xSkinHashTree;

static Asura_PS3_GCM_VertexShader*			g_pxVertexShader			= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxVertexShaderDM			= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxSmoothSkinSpu			= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxSmoothSkinSpuDM			= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxStraightThroughShadowVS	= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxInUseVS					= NULL;

static Asura_PS3_GCM_VertexShader*			g_pxShadowVertexShader		= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxZPassVertexShader		= NULL;
static Asura_PS3_GCM_VertexShader*			g_pxZPassVertexShaderDM		= NULL;
static Asura_PS3_FragmentShader*			g_pxFragmentShaderZ			= NULL;
Asura_PS3_MaterialResponse_Handles*			g_xMaterialResponseHandles;

static 	Asura_PS3_SPU_SkinningConstants		g_xShaderConstants;

static Asura_AMatrix4x4						g_axBones[ ASURA_MAX_NO_OF_BONES_PER_MODEL ]; //BW if SPUs were changed to read the 4x3 array we wouldn't need this
static float								g_axBoneData4x3[150*3*4];


static Asura_PS3_SkinDebug_FP_Handles		g_xSkinDebugHandles;
#define HASH_SKIN_DEBUG_SHADER	(0xf000baaa)

DVSVAR bool									g_bUseDebugShader			= false;
DVSVAR bool									g_bCharacterStripsVisible	= true;
DVSVAR bool									g_bDrawSmoothSkin			= true;
DVSVAR float								g_fDamageMapTest			= 0.0f; //if this is <= 0 the real damage map values will be used


/******************************************************************************/
/* Asura_PS3_CharacterRenderer_Skin Class Functions                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_CharacterRenderer_Skin::Asura_PS3_CharacterRenderer_Skin()
//
// Constructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_PS3_CharacterRenderer_Skin::Asura_PS3_CharacterRenderer_Skin()
	: m_pxStrips( NULL )
	, m_uNumStrips(0)
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_CharacterRenderer_Skin::~Asura_PS3_CharacterRenderer_Skin()
//
// Destructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_PS3_CharacterRenderer_Skin::~Asura_PS3_CharacterRenderer_Skin()
{
	m_xIndexBuffer.Shutdown();
	m_xVertexBuffer.Shutdown();

	delete[] m_pxStrips;
}


void Asura_PS3_CharacterRenderer::Initialise(void)
{
#if defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SUPPORT_COMMAND_CONSOLE )
	Asura_CommandConsole::AddVar( L"Render.Enable.SmoothSkin",	&g_bDrawSmoothSkin,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.DamageMapTest",	&g_fDamageMapTest, -1.0f, 1.0f,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.UseDebugShader",	&g_bUseDebugShader,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.CharacterStripsVisible",	&g_bCharacterStripsVisible,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
	g_pxVertexShader			= new Asura_PS3_VP_SmoothSkin();
	g_pxVertexShaderDM			= new Asura_PS3_VP_SmoothSkinDM();
	g_pxShadowVertexShader		= new Asura_PS3_VP_SmoothSkinShadow();
	g_pxZPassVertexShader		= new Asura_PS3_VP_SmoothSkinZPass();
	g_pxZPassVertexShaderDM		= new Asura_PS3_VP_SmoothSkinZPassDM();

	// SPU skinning shaders
	g_pxSmoothSkinSpu			= new Asura_PS3_VP_SmoothSkin_SpuSkinning();
	g_pxSmoothSkinSpuDM			= new Asura_PS3_VP_SmoothSkin_SpuSkinning_DM();
	g_pxStraightThroughShadowVS	= new Asura_PS3_VP_SmoothSkin_StraightThroughShadow();
	
	g_pxFragmentShaderZ			= new Asura_PS3_FragmentShader( (unsigned char *)&_binary_PS3_DeferredPhoton_P0_FP_bin_start, (unsigned char *)&_binary_PS3_DeferredPhoton_P0_FP_bin_end, NULL );
	Asura_PS3_GCM_ShadersLibrary::AddFP( ASURA_SHADER_HASH_CHARACTER, HASH_SKIN_DEBUG_SHADER, (unsigned char *)&_binary_PS3_SkinDebug_FP_bin_start, (unsigned char *)&_binary_PS3_SkinDebug_FP_bin_end - (unsigned char *)&_binary_PS3_SkinDebug_FP_bin_start, &g_xSkinDebugHandles, 1, true);

	g_pxVertexShader->InitVertexBufferDeclaration( g_axSmoothSkinVertexDeclaration, NUM_SMOOTHSKIN_VD_ELEMENTS );
	g_pxShadowVertexShader->InitVertexBufferDeclaration( g_axVertexDeclarationShadow, VDS_ELEMENTS );
	g_pxSmoothSkinSpu->InitVertexBufferDeclaration( g_axSpuSkinningVD_FromSpu, NUM_SPUSKINNING_FROMSPU_ELEMENTS );
	g_pxSmoothSkinSpu->InitVertexBufferDeclaration( g_axSpuSkinningVD_FromData, NUM_SPUSKINNING_FROMDATA_ELEMENTS );
	g_pxSmoothSkinSpuDM->InitVertexBufferDeclaration( g_axSpuSkinningVD_FromSpu, NUM_SPUSKINNING_FROMSPU_ELEMENTS );
	g_pxSmoothSkinSpuDM->InitVertexBufferDeclaration( g_axSpuSkinningVD_FromData, NUM_SPUSKINNING_FROMDATA_ELEMENTS );
	g_pxStraightThroughShadowVS->InitVertexBufferDeclaration( g_axStraightThroughVertexDeclarationShadow, VDS_STRAIGHTTHROUGH_ELEMENTS );

	//BW some sanity checks for the SPU skinning code
	Asura_Assertf( sizeof(Asura_PS3_SPU_Skinning_InputVertex) == ASURA_PS3_CHARACTER_VERTEX_SIZE, ("Vertex definition is wrong! SPU code depends on this!\n %d != %d", sizeof(Asura_PS3_SPU_Skinning_InputVertex), ASURA_PS3_CHARACTER_VERTEX_SIZE) );
	Asura_Assertf( sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly) == SPUSKINNING_FROMSPU_VERTEX_SIZE, ("SPU-skinning: output vertex size does not match declaration, this could be fatal!\n %d != %d", sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly), SPUSKINNING_FROMSPU_VERTEX_SIZE) );
	Asura_Assertf( sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly) == STRAIGHTTHROUGH_SHADOW_VERTEX_SIZE, ("SPU-skinning: output vertex size does not match declaration, this could be fatal!\n %d != %d", sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly), STRAIGHTTHROUGH_SHADOW_VERTEX_SIZE) );
}

int Asura_PS3_CharacterRenderer::GetVertexSize( void )
{
	return ASURA_PS3_CHARACTER_VERTEX_SIZE;
}


const Asura_AMatrix4x4* Asura_PS3_CharacterRenderer::GetBonesForRender( const Asura_AMatrix4x4* pxBones, int iNumberOfBones, const Asura_Animation& xAnimation, bool bUseDamageMaps, bool bSpuVS )
{
#ifdef ASURA_USE_DAMAGE_MAPS	
	if( bUseDamageMaps && !bSpuVS )
	{
		for( int i = 0; i < iNumberOfBones; ++i )
		{
			float fDamage;
			if ( g_fDamageMapTest > 0.0f )
			{
				fDamage = g_fDamageMapTest; //override the damage map for testing
			}
			else
			{
				fDamage = xAnimation.GetBoneDamageMapValue( i );
			}

			g_axBones[ i ] = pxBones[ i ];

			g_axBones[ i ].SetElement( 2, 2, fDamage );

			if (dbg_bTestDamageMaps)
			{
				Asura_OutputToDebugger(( "Damage: %f\n", xAnimation.GetBoneDamageMapValue( i ) ));
			}
		}

		return g_axBones;
	}
	else
#endif // #ifdef ASURA_USE_DAMAGE_MAPS	
	{
		return pxBones;
	}
}


// Make sure the data has been skinned this frame, and update the render caches accordingly.
void Asura_PS3_CharacterRenderer::UseSpuSkinning(	Asura_PS3_CharacterRenderer_Skin* pxCharacterStrips,
													const Asura_Animation& xAnimation,
													SPU_SKINNING_MODE eMode,
													Asura_AMatrix4x4_InParam xTransform,
													const Asura_AMatrix4x4* const pxBoneTransforms,
													u_int uNumBones,
													u_int uSkinIdx )
{
	Asura_Platform_Animation_RenderCache* pxRenderCache = xAnimation.GetRenderCache();
	Asura_PS3_SPU_SkinningConstants xShaderConstants;
	
	// need to set up all of these correctly so the SPU'd vertices can be re-used by character render
	xShaderConstants.m_xWorldViewProjectionMatrix = Asura_PS3_Graphics::GetWorldViewProjectionMatrix( &xTransform );
	xShaderConstants.m_xWorldMatrix = xTransform;

	Asura_AVector4	xCamPos(*Asura_Camera::GetPosition(), 1.0f);
	xShaderConstants.m_xCameraPosition = xCamPos;

	// update the cached bones (input to the SPU) if they're out dated
	if ( !pxRenderCache->IsBoneCacheCurrent() )
	{
		const Asura_AMatrix4x4 *pxBones = pxBoneTransforms;
		pxRenderCache->UpdateCachedBones( pxBones, uNumBones );
	}
	xShaderConstants.m_uBoneArrayAddr = reinterpret_cast<uint64_t>(pxRenderCache->GetCachedBones());
	xShaderConstants.m_uNumBones = uNumBones;
	
	u_int uNumVerts = pxCharacterStrips->GetVertexBuffer()->GetNumVerts();

	// make sure the cache is big enough for the highest LOD
	//BW otherwise there will be a lot of allocing, though that would give better memory usage if it weren't for the fragmentation.
	Asura_PS3_CharacterRenderer_Skin* pxTopLod = Asura_PS3_CharacterRenderer::GetCharacter( xAnimation.GetSkinHashID(uSkinIdx) );
	u_int uMaxVerts = pxTopLod->GetVertexBuffer()->GetNumVerts();
	u_int uMaxIdxs = pxTopLod->GetIndexBuffer()->GetIndexBufferSize()/2;

	Asura_Assert( uNumVerts <= uMaxVerts, "Sanity check failed, maybe GetSkinHashID() did not return the highest LOD" );

	pxRenderCache->ObtainVertexBuffer( ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS, uSkinIdx, uMaxVerts, sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly) );
	
	if ( eMode != SPU_SKINNING_POS_ONLY_NO_CULL )
	{
		pxRenderCache->ObtainIndexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED,uSkinIdx,uMaxIdxs); // make sure it's alloced
	}

	//
	// SPU vertex processing
	//

	if ( !pxRenderCache->IsVertexCacheCurrent( ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS, uSkinIdx )
		|| ( eMode != SPU_SKINNING_POS_ONLY_NO_CULL && !xAnimation.GetRenderCache()->IsIndexCacheCurrent( ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED, uSkinIdx ) ) )
	{
		bool bRet = Asura_PS3_SPU_Skinning::ProcessSkin( pxCharacterStrips, &xShaderConstants, xAnimation.GetRenderCache(), uSkinIdx, uNumVerts, eMode );

		//BW conceivably this could be handled, e.g. if we run out of memory, but that's probably fatal anyway.
		Asura_Assert( bRet, "RenderCharacter: ProcessSkin() != IsSpuable(), this is probably fatal" );
	}
	
	
/*	if ( eMode == SPU_SKINNING_POS_AND_CULL_AND_AA )
	{
		send AA edges to AA if not already;
	}*/
}


void Asura_PS3_CharacterRenderer::SetVPData( Asura_PS3_CharacterRenderer_Skin* pxCharacterStrips,
											const Asura_Animation& xAnimation,
											const Asura_Hierarchy_Skin& xSkin,
											u_int uSkinIdx,
											Asura_AMatrix4x4_InParam xTransform,
											Asura_AMatrix4x4_InParam xInverseTransform,
											const Asura_AMatrix4x4* const pxBones,
											bool bUseDamageMaps, bool bTransparentPass, bool bSpuVS )
{
	// Render States

	Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CW );
	Asura_PS3_RenderingStates::RequestZBufferEnabled( true );

	if( bTransparentPass )
	{
		Asura_PS3_DeferredRendering::BindAsSingleRT();
		Asura_PS3_RenderingStates::RequestAlphaTest( false );
	}
	else
	{
		Asura_PS3_DeferredRendering::Bind();
		Asura_PS3_RenderingStates::RequestAlphaTest( false );
	}
	
	// pick a shader for the current circumstances
	if ( bSpuVS )
	{
		if ( bUseDamageMaps )	g_pxInUseVS = g_pxSmoothSkinSpuDM;
		else					g_pxInUseVS = g_pxSmoothSkinSpu;
	}
	else
	{
		if ( bUseDamageMaps )	g_pxInUseVS = g_pxVertexShaderDM;
		else					g_pxInUseVS = g_pxVertexShader;
	}
	g_pxInUseVS->Bind();
	

	// Send constants to shader
	g_pxInUseVS->SetUniformParameter( SK_WorldMatrix,  xTransform.GetElements() );

	const Asura_AMatrix4x4& xWorldProjectionView = Asura_PS3_Graphics::GetWorldViewProjectionMatrix( &xTransform );
	g_pxInUseVS->SetUniformParameter( SK_WorldViewProj,  xWorldProjectionView.GetElements() );


	// Skin shader needs world-space camera position
	Asura_AVector4	xCamPos(*Asura_Camera::GetPosition(), 1.0f);
	g_pxInUseVS->SetUniformParameter( SK_CamPos, xCamPos.GetElements() );


	// Set bone matrices
	// Pack bone matrices as 4x3 for vertex shader
	const Asura_AMatrix4x4 *pxRenderBones = GetBonesForRender(pxBones, xSkin.NumberOfBones, xAnimation, bUseDamageMaps, false);

	Asura_Assertf( xSkin.NumberOfBones <= 150, ("PS3 only supports up to 220 bones, this skin has %d", xSkin.NumberOfBones) );
	for ( int i=0; i < xSkin.NumberOfBones; i++ )
	{
		// store transposed, so only need 3 float4's not 4 in shader
		pxRenderBones[i].StoreRowsUnaligned(&g_axBoneData4x3[i*3*4]); // this stores 4x4 to our 3x4 space, hopefully it won't break anything
	}

	//avoid the huge copy by just setting it directly
	cellGcmSetVertexProgramParameterBlockUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), g_pxInUseVS->GetRegisterIndex(SK_BoneTransforms), xSkin.NumberOfBones*3, g_axBoneData4x3 );
	
	//invalidate the cached vertex parameters, since we just set them directly
	g_pxInUseVS->InvalidateRegisterCache();


	// run the SPU skinning with the same settings
	if ( bSpuVS )
	{
		const Asura_AMatrix4x4 *pxSpuRenderBones = GetBonesForRender(pxBones, xSkin.NumberOfBones, xAnimation, bUseDamageMaps, true);
		Asura_PS3_CharacterRenderer::UseSpuSkinning( pxCharacterStrips, xAnimation, SPU_SKINNING_POS_AND_CULL,
														xTransform, pxSpuRenderBones, xSkin.NumberOfBones, uSkinIdx );
	}
}



#ifdef ASURA_USE_ZPREPASS
void Asura_PS3_CharacterRenderer::RenderCharacterZPassOnly(const Asura_Animation& xAnimation,
												  const Asura_Hierarchy_Skin& xSkin,
												  Asura_AMatrix4x4_InParam xTransform,
												  Asura_AMatrix4x4_InParam xInverseTransform,
												  const Asura_AMatrix4x4* const pxBones, const u_int uSkinIdx,
												  bool bUseDamageMaps)
{
	Asura_PS3_CharacterRenderer_Skin* const pxCharacterStrips = GetCharacter( xSkin.m_uNameHashID );
	Asura_Assert( pxCharacterStrips, "Character has no strips." );

	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__SMOOTHSKIN, "RenderCharacterZPassOnly" );
	const bool bSpuVS = Asura_PS3_SPU_Skinning::IsSpuable(pxCharacterStrips);
	const bool bTransparentPass = Asura_SortedRender::IsTransparentPass();

	// this is a little misleading, we get this function to set things up although we use a different shader.
	// this works because the z-pass shaders have the same constants as the other ones, but it would be better if that interdependency were more visible.
	SetVPData( pxCharacterStrips, xAnimation, xSkin, uSkinIdx, xTransform, xInverseTransform, pxBones, bUseDamageMaps, bTransparentPass, bSpuVS );

	if ( bSpuVS )
	{
		Asura_PS3_DeferredRendering::BindAsZPassOnly();
		Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( true );

		// use the verts processed by SPU
		Asura_PS3_GCM_VertexBuffer* pProcessedVerts = xAnimation.GetRenderCache()->ObtainVertexBuffer( ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS, uSkinIdx );
		pProcessedVerts->Bind( g_axStraightThroughVertexDeclarationShadow, VDS_STRAIGHTTHROUGH_ELEMENTS );
		
		// use indices processed by SPU
		Asura_PS3_GCM_IndexBuffer* pIndexBuffer = xAnimation.GetRenderCache()->ObtainIndexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED,uSkinIdx);
		
		g_pxFragmentShaderZ->Bind();

		for( u_int u = 0; u < pxCharacterStrips->m_uNumStrips; ++u )
		{
#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
			if (xAnimation.GetSkinElementVisibility(pxCharacterStrips->m_pxStrips[u].m_iElementID))
#endif
			{
				const Asura_Hash_ID uHash = pxCharacterStrips->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
				const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uHash );

				if( pxMaterial )
				{
					const bool bTransparentStrip = pxMaterial->IsTransparent();

					if( bTransparentStrip != bTransparentPass )
					{
						// don't draw opaque strips in transparent pass and vice versa
						continue;
					}
					
					if( bTransparentStrip )
					{
						// only draw z-prepass on a transparent things if the material's ZPrePass flag is set.
						// this flag gives a kind of transparency where occluded parts of the model aren't visible, i.e. like a superimposed image.
						if ( !pxMaterial->GetZPrePass() )
						{
							continue;
						}
					}
					else
					{
						if ( pxMaterial->GetDontRender() || pxMaterial->GetIsDecal() )
						{
							// only draw opaque materials that will be drawn in the main render
							// decals need to use their alpha-texture
							continue;
						}
					}
				}

				pIndexBuffer->RenderTriangles( pxCharacterStrips->m_pxStrips[ u ].m_uNumIndices, pxCharacterStrips->m_pxStrips[ u ].m_uStartIndex );
					
				Asura_PS3_Graphics::Flush();
			}
		}

		Asura_PS3_SPU_Skinning::NotifyBufferUsed();
	}
	else										
	{
		bool bFirstDrawCall = true;
		
		Asura_PS3_GCM_VertexShader* pxVertexShader;

#ifdef ASURA_USE_DAMAGE_MAPS	
		bool bUseDamageMaps = xAnimation.IsUsingDamageMap() || g_fDamageMapTest > 0.0f;
#else
		const bool bUseDamageMaps = false;
#endif
		if ( bUseDamageMaps )
		{
			pxVertexShader = g_pxZPassVertexShaderDM;
		}
		else
		{
			pxVertexShader = g_pxZPassVertexShader;
		}

		for( u_int u = 0; u < pxCharacterStrips->GetNumStrips() ; ++u )
		{
	#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
			if (xAnimation.GetSkinElementVisibility(pxCharacterStrips->m_pxStrips[u].m_iElementID))
	#endif
			{			
				const Asura_Hash_ID uMaterialHash1 = pxCharacterStrips->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
				const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

				if( pxMaterial1 )
				{
					const bool bTransparentStrip = pxMaterial1->IsTransparent();

					if( bTransparentStrip != bTransparentPass )
					{
						// don't draw opaque strips in transparent pass and vice versa
						continue;
					}
					
					if( bTransparentStrip )
					{
						// only draw z-prepass on a transparent things if the material's ZPrePass flag is set.
						// this flag gives a kind of transparency where occluded parts of the model aren't visible, i.e. like a superimposed image.
						if ( !pxMaterial1->GetZPrePass() )
						{
							continue;
						}
					}
					else
					{
						if ( pxMaterial1->GetDontRender() || pxMaterial1->GetIsDecal() )
						{
							// only draw opaque materials that will be drawn in the main render
							// decals need to use their alpha-texture
							continue;
						}
					}
				}

				if( bFirstDrawCall )
				{
					Asura_PS3_DeferredRendering::BindAsZPassOnly();
					Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_NONE );

					Asura_PS3_RenderingStates::RequestAlphaTest( false );
					Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CW );
					Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
					Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
					Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( true );
					Asura_PS3_RenderingStates::SetStencilTestEnabled( false );
					Asura_PS3_RenderingStates::SetStencilFunction( ASURA_PS3_STENCIL_NEVER, 0, 0 );

					pxVertexShader->Bind();

					const Asura_AMatrix4x4& xWorldProjectionView = Asura_PS3_Graphics::GetWorldViewProjectionMatrix( &xTransform );

					pxVertexShader->SetUniformParameter( SKZ_WorldViewProj, xWorldProjectionView.GetElements() );

					// set bones
					Asura_Assertf( xSkin.NumberOfBones <= 150, ("PS3 only supports up to 220 bones, this skin has %d", xSkin.NumberOfBones) );
					for ( int i=0; i < xSkin.NumberOfBones; i++ )
					{
						// store transposed, so only need 3 float4's not 4 in shader
						pxBones[i].StoreRowsUnaligned(&g_axBoneData4x3[i*3*4]); // this stores 4x4 to our 3x4 space, hopefully it won't break anything
					}

					//avoid the huge copy by just setting it directly
					cellGcmSetVertexProgramParameterBlockUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), pxVertexShader->GetRegisterIndex(SKZ_BoneTransforms), xSkin.NumberOfBones*3, g_axBoneData4x3 );
					
					//invalidate the cached vertex parameters, since we just set them directly
					pxVertexShader->InvalidateRegisterCache();
					
					pxCharacterStrips->m_xVertexBuffer.Bind( g_axVertexDeclarationShadow, VDS_ELEMENTS );
					g_pxFragmentShaderZ->Bind();
					bFirstDrawCall = false;
				}

				pxCharacterStrips->m_xIndexBuffer.RenderTriangles( pxCharacterStrips->m_pxStrips[ u ].m_uNumIndices, pxCharacterStrips->m_pxStrips[ u ].m_uStartIndex );
				Asura_PS3_Graphics::Flush();
			}
		}
	}
	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__SMOOTHSKIN );
}
#endif

void Asura_PS3_CharacterRenderer::RenderCharacter(const Asura_Animation& xAnimation,
													  const Asura_Hierarchy_Skin& xSkin,
													  Asura_AMatrix4x4_InParam xTransform,
													  Asura_AMatrix4x4_InParam xInverseTransform,
													  const Asura_AMatrix4x4* const pxBones,
													  u_int uSkinIdx)
{
	if( !g_bDrawSmoothSkin ) return;

	Asura_PS3_CharacterRenderer_Skin* const pxCharacterStrips = GetCharacter( xSkin.m_uNameHashID );
	Asura_Assert( pxCharacterStrips, "Character has no strips." );

	//	New render logic
	// Decide if we're going to use the vertex shader or SPU for rendering
	const	bool bSpuVS				= Asura_PS3_SPU_Skinning::IsSpuable(pxCharacterStrips);
	const	bool bTransparentPass	= Asura_SortedRender::IsTransparentPass();

	Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
	Asura_PS3_RenderingStates::SetStencilTestEnabled(false);

#ifdef ASURA_USE_DAMAGE_MAPS	
	bool bUseDamageMaps = xAnimation.IsUsingDamageMap() || g_fDamageMapTest > 0.0f;
#else
	const bool bUseDamageMaps = false;
#endif

#ifdef ASURA_USE_ZPREPASS
	RenderCharacterZPassOnly( xAnimation, xSkin, xTransform, xInverseTransform, pxBones, uSkinIdx, bUseDamageMaps );
	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
#else
	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( true );
#endif

	SetVPData( pxCharacterStrips, xAnimation, xSkin, uSkinIdx, xTransform, xInverseTransform, pxBones, bUseDamageMaps, bTransparentPass, bSpuVS );

	Asura_PS3_GCM_VertexBuffer*	pProcessedPositions = NULL;
	Asura_PS3_GCM_IndexBuffer*	pIndexBuffer		= NULL;

	if ( bSpuVS )
	{
		//use the verts processed by SPU
		pProcessedPositions = xAnimation.GetRenderCache()->ObtainVertexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS,uSkinIdx);
		pIndexBuffer = xAnimation.GetRenderCache()->ObtainIndexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED,uSkinIdx);

		//bind SPU-skinned things
		pProcessedPositions->Bind( g_axSpuSkinningVD_FromSpu, NUM_SPUSKINNING_FROMSPU_ELEMENTS, 1, true ); //true => clear other bindings
		
		//bind non-SPU-skinned things
		pxCharacterStrips->m_xVertexBuffer.Bind(g_axSpuSkinningVD_FromData, NUM_SPUSKINNING_FROMDATA_ELEMENTS, 1, false);

		Asura_Assert( pIndexBuffer, "RenderCharacter: pIndexBuffer is NULL" );
		Asura_Assert( pIndexBuffer->GetIndexBufferSize() >= pxCharacterStrips->m_xIndexBuffer.GetIndexBufferSize(), "Cached buffer num indices too small" );
	}
	else
	{
		//failed to setup SPU processing, do it using a vertex shader
		pxCharacterStrips->m_xVertexBuffer.Bind( g_axSmoothSkinVertexDeclaration, NUM_SMOOTHSKIN_VD_ELEMENTS );

		pIndexBuffer = &pxCharacterStrips->m_xIndexBuffer;
	}
	
	Asura_PS3_FP_List_Tree* fpTree = Asura_PS3_GCM_ShadersLibrary::GetResettableFPTree( ASURA_SHADER_HASH_CHARACTER );
	Asura_Assert(fpTree,"ASURA_SHADER_HASH_CHARACTER shader tree is NULL, no character shaders have been loaded!");

	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__SMOOTHSKIN, "RenderCharacter" );

	for( u_int u = 0; u < pxCharacterStrips->m_uNumStrips; ++u )
	{
		Asura_PS3_FragmentShaderCopy* pxShader = NULL;
#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		if (xAnimation.GetSkinElementVisibility(pxCharacterStrips->m_pxStrips[u].m_iElementID))
#endif
		{			
			const Asura_Hash_ID uMaterialHash1 = pxCharacterStrips->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
			const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

			if( pxMaterial1 )
			{
				const bool bTransparentStrip = pxMaterial1->IsTransparent();
				if( ( pxMaterial1->GetDontRender() ) || ( bTransparentStrip != bTransparentPass ) )
				{
					continue;
				}
			}
			
			const Asura_Hash_ID uMaterialHash2 = pxCharacterStrips->m_pxStrips[u].m_auMaterialResponseHashIDs[1];
			const Asura_MaterialResponse* const pxMaterial2 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash2 );
			const Asura_TextureAnimationController* pxTextureAnimationController = xAnimation.GetTextureAnimController();

			Asura_Hash_ID uHashID = pxCharacterStrips->m_pxStrips[ u ].m_uFPHashID;
			if( Asura_MaterialResponse_System::GetCurrentOverrideMapHash() != ASURA_HASH_ID_UNSET )
			{
				uHashID = Asura_PS3_MaterialResponse_System::CreateShaderHashID( uMaterialHash1, uMaterialHash2 );
			}
			if( bTransparentPass && Asura_PS3_Fog::GetFogParameters().m_bFogEnabled )
			{
				uHashID += ASURA_PS3_SHADER_FOG;
			}
			if ( bUseDamageMaps )
			{
				uHashID += ASURA_PS3_SHADER_DAMAGE_MAP;
			}

			pxShader = fpTree->GetShader( uHashID );

			if ( !pxShader && bUseDamageMaps ) pxShader = fpTree->GetShader( uHashID&~ASURA_PS3_SHADER_DAMAGE_MAP );

			if ( pxShader )
			{
				Asura_PS3_MaterialResponse_System::SetBlendStates( pxMaterial1 );

				Asura_PS3_MaterialResponse_System::SetEffectParameters( pxMaterial1, pxMaterial2, &xAnimation, pxShader );
				Asura_PS3_MaterialResponse_System::SetSHParameters( pxMaterial1, pxShader );
				Asura_PS3_MaterialResponse_System::SetProjectContext( pxMaterial1, pxMaterial2 );

				Asura_TextureAnimationOutput xTexAnim1, xTexAnim2;

				Asura_TextureAnimation_System::GetAnimationState( pxMaterial1->GetTextureAnim(), pxTextureAnimationController, &xTexAnim1 );
				Asura_TextureAnimation_System::GetAnimationState( pxMaterial2 ? pxMaterial2->GetTextureAnim() : ASURA_HASH_ID_UNSET, pxTextureAnimationController, &xTexAnim2 );

				const float afTexCoordOffset[4] = { xTexAnim1.m_fUOffset, xTexAnim1.m_fVOffset, xTexAnim2.m_fUOffset, xTexAnim2.m_fVOffset };
				g_pxInUseVS->SetUniformParameter( (u_int)SK_UV_Offsets, afTexCoordOffset );


				if ( g_bUseDebugShader == false )
				{
					pxShader->Bind();
				}
				else
				{
					//each time we Get this FP it will create a new instance that can have a different set of parameters
					Asura_PS3_FragmentShaderCopy* pxDebugShader = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_CHARACTER, HASH_SKIN_DEBUG_SHADER );

					int v = u; //= white => it was 6. but that ain't possible. *IT NEVER HITS THIS CODE WITH THAT VAL!!!*
					float afDebugColour[4] = { (float)(v&1), (float)((v/2)&1), (float)((v/4)&1), 1 };//{ batchnum&1, batchnum&2, batchnum&4, 1 }; I don't have the batch number at this point, so use strip number (it's mostly the same thing)
					//afDebugColour[0] = (float)(v&1); afDebugColour[1] = (float)((v/2)&1); afDebugColour[2] = (float)((v/4)&1);
					pxDebugShader->SetFloats( FP_DebugColour, 4, afDebugColour );
					float afMaterialProperties[ 4 ] = { pxMaterial1->GetMaterialIndex(), pxMaterial1->GetReflectivity(), pxMaterial1->GetParallaxScale(), -pxMaterial1->GetParallaxScale() * 0.5f };
					pxDebugShader->SetFloats( FP_DebugMaterialProperties, 4, afMaterialProperties );
					pxDebugShader->Bind();
				}

				if ( g_bCharacterStripsVisible ) //this just skips the draw call, still do all the setup (useful for SPU testing)
				{
					pIndexBuffer->RenderTriangles( pxCharacterStrips->m_pxStrips[ u ].m_uNumIndices, pxCharacterStrips->m_pxStrips[ u ].m_uStartIndex );
				}
			}
			else
			{
				Asura_Assert( false, "Error: No shader set for this strip" );
			}
		}
	}

	if ( bSpuVS )
	{
		Asura_PS3_SPU_Skinning::NotifyBufferUsed();
	}

	Asura_PS3_RenderingStates::SetStencilTestEnabled( false );

	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__SMOOTHSKIN );
}

void Asura_PS3_CharacterRenderer::RenderShadow(	const Asura_Animation& xAnimation,
												const Asura_Hierarchy_Skin& xSkin,
												const Asura_Shadow& xShadow,
												Asura_AMatrix4x4_InParam xTransform,
												const Asura_AMatrix4x4* const pxBones,
												u_int uSkinIdx )
{
	Asura_PS3_CharacterRenderer_Skin* const pxCharacterStrips = GetCharacter( xSkin.m_uNameHashID );

	if ( pxCharacterStrips )
	{
		// Decide if we're going to use the vertex shader or SPU for rendering
		bool bSpuVS = Asura_PS3_SPU_Skinning::IsSpuable(pxCharacterStrips);

		Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( true );
		Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
		Asura_PS3_RenderingStates::RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		Asura_PS3_RenderingStates::SetBlendEnableMrt( false, false, false );
		Asura_PS3_RenderingStates::SetAlphaBlending( false );
		Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALL );
		Asura_PS3_RenderingStates::SetUseDepthBias( false );
		Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CW );

		Asura_AMatrix4x4 xWorldProjectionView;
		Asura_AMatrix4x4 xObjectMatrix;

		xObjectMatrix = xTransform;

		Asura_Matrix_3x3 xLightOri;
		xShadow.GetLightOrientation(xLightOri);
		const Asura_Vector_3& xLightPos = xShadow.GetLightPosition();

		Asura_AMatrix4x4 xPerspective;
		Asura_AMatrix4x4 xLight( xLightOri, xLightPos );
		xLight.InvertOrthogonal();

		xPerspective.SetPerspectiveForLH( Asura_Maths::DegToRad( xShadow.GetFrustum().GetVerticalFOVInDegrees() ), 1.0f, xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane() );

		xWorldProjectionView = xObjectMatrix * xLight * xPerspective;

		if ( bSpuVS )
		{
			Asura_PS3_CharacterRenderer::UseSpuSkinning( pxCharacterStrips, xAnimation, SPU_SKINNING_POS_ONLY_NO_CULL,
															xTransform, pxBones, xSkin.NumberOfBones, uSkinIdx );

			g_pxStraightThroughShadowVS->Bind();
			g_pxStraightThroughShadowVS->SetUniformParameter( VS_TRANSFORM_FLOAT4_Transform, xWorldProjectionView.GetElements() );
		}
		else
		{
			g_pxShadowVertexShader->Bind();

			g_pxShadowVertexShader->SetUniformParameter( SKS_WorldViewProj, xWorldProjectionView.GetElements() );

			Asura_Assertf( xSkin.NumberOfBones <= 150, ("PS3 only supports up to 220 bones, this skin has %d", xSkin.NumberOfBones) );
			for ( int i=0; i < xSkin.NumberOfBones; i++ )
			{
				// store transposed, so only need 3 float4's not 4 in shader
				pxBones[i].StoreRowsUnaligned(&g_axBoneData4x3[i*3*4]); // this stores 4x4 to our 3x4 space, hopefully it won't break anything
			}

			//avoid the huge copy by just setting it directly
			cellGcmSetVertexProgramParameterBlockUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), g_pxShadowVertexShader->GetRegisterIndex(SKS_BoneTransforms), xSkin.NumberOfBones*3, g_axBoneData4x3 );
			
			//invalidate the cached vertex parameters, since we just set them directly
			g_pxShadowVertexShader->InvalidateRegisterCache();
		}


		Asura_PS3_GCM_VertexBuffer *pProcessedVerts = NULL;
		if ( bSpuVS )
		{
			//use the verts processed by SPU
			pProcessedVerts = xAnimation.GetRenderCache()->ObtainVertexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS,uSkinIdx);
			pProcessedVerts->Bind( g_axStraightThroughVertexDeclarationShadow, VDS_STRAIGHTTHROUGH_ELEMENTS );
		}
		else
		{
			//failed to setup SPU processing, do it using a vertex shader
 			pxCharacterStrips->m_xVertexBuffer.Bind( g_axVertexDeclarationShadow, VDS_ELEMENTS );
		}

		for( u_int u = 0; u < pxCharacterStrips->m_uNumStrips; ++u )
		{
	#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
			if (xAnimation.GetSkinElementVisibility(pxCharacterStrips->m_pxStrips[u].m_iElementID))
	#endif
			{
				const Asura_Hash_ID uHash = pxCharacterStrips->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
				const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uHash );

				if( pxMaterial && !pxMaterial->IsTransparent() && (pxMaterial->GetRenderType() != Asura_MaterialResponse::RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW) )
				{
					pxCharacterStrips->m_xIndexBuffer.RenderTriangles( pxCharacterStrips->m_pxStrips[ u ].m_uNumIndices, pxCharacterStrips->m_pxStrips[ u ].m_uStartIndex );
					
					Asura_PS3_Graphics::Flush();
				}
			}
		}

		if ( bSpuVS )
		{
			Asura_PS3_SPU_Skinning::NotifyBufferUsed();
		}
	}
}


Asura_PS3_CharacterRenderer_Skin* Asura_PS3_CharacterRenderer::GetCharacter( const Asura_Hash_ID uHash )
{
	Asura_PS3_CharacterRenderer_Skin* const pxSkin = g_xCharacterTree.GetDataByHash( uHash );
	Asura_Assert( pxSkin, "Failed to get character data." );

	return pxSkin;
}
////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_CharacterRenderer::IsCharacterOpaque()
//
// Queries whether a character is entirely opaque
//
// Parameters:		uHash - Hash ID of character name
// Return value:	true if all strips are opaque
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_PS3_CharacterRenderer::IsCharacterOpaque( const Asura_Hash_ID uHash )
{
	Asura_PS3_CharacterRenderer_Skin* const pxSkin = GetCharacter( uHash );

	if ( pxSkin )
	{
		for( u_int u = 0; u < pxSkin->m_uNumStrips; ++u )
		{
			const Asura_Hash_ID uMaterialResponseHashID = pxSkin->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
			const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialResponseHashID );

			// Decal-type materials count as opaque
			if( pxMaterial && pxMaterial->IsTransparent() )
			{
				return false;
			}
		}
	}
	else
	{
		return false;
	}

	return true;
}

void Asura_PS3_CharacterRenderer::Reset()
{
	g_xCharacterTree.Reset();
}

void Asura_PS3_CharacterRenderer::Shutdown(void)
{
    Reset();

	delete g_pxVertexShader;
	delete g_pxVertexShaderDM;
	delete g_pxShadowVertexShader;
	delete g_pxZPassVertexShader;
	delete g_pxZPassVertexShaderDM;
	delete g_pxSmoothSkinSpu;
	delete g_pxSmoothSkinSpuDM;
	delete g_pxStraightThroughShadowVS;
	delete g_pxFragmentShaderZ;

	g_pxVertexShader			= NULL;
	g_pxVertexShaderDM			= NULL;
	g_pxShadowVertexShader		= NULL;
	g_pxZPassVertexShader		= NULL;
	g_pxZPassVertexShaderDM		= NULL;
	g_pxSmoothSkinSpu			= NULL;
	g_pxSmoothSkinSpuDM			= NULL;
	g_pxStraightThroughShadowVS	= NULL;
	g_pxFragmentShaderZ			= NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_CharacterRenderer::CreateCharacter()
//
// Creates a new PS3 character. Called by the chunk loader
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_PS3_CharacterRenderer_Skin* Asura_PS3_CharacterRenderer::CreateCharacter( const Asura_Hash_ID uHash, const u_int uNumStrips )
{
	Asura_PS3_CharacterRenderer_Skin* const pxNewCharacter = g_xCharacterTree.AddNode( uHash );
	if( !pxNewCharacter )
	{
		return NULL;
	}

	pxNewCharacter->m_pxStrips = new Asura_PS3_CharacterRenderer_Strip[ uNumStrips ];
	if( !pxNewCharacter->m_pxStrips )
	{
		Asura_Assert( pxNewCharacter->m_pxStrips, "Unable to allocate character materials." );
		return NULL;
	}

	pxNewCharacter->m_uNumStrips = uNumStrips;

	return pxNewCharacter;
}
