//
// Asura_PS3_FrameBuffer.cpp
//
// Created by KJFL - 06 December 2005
// Modified Catalin Arsenescu - 27 March 2007
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS3_FrameBuffer.h"

#if defined(ASURA_USE_PROFILING_SYSTEM)
#include "Debugging/Asura_CPUProfile.h"
#endif	// defined(ASURA_USE_PROFILING_SYSTEM)

#include "Core/Memory/Asura_Mem.h"
#include "Rendering/Asura_Overlay_Quad.h"

#include "Rendering/Asura_PS3_TextureManagement.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_QuadOverlayRenderer.h"
#include "Rendering/Asura_PS3_Helper.h"
#include "Rendering/Asura_PS3_SwapChain.h"

#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_LogFile.h"

#include <sys/timer.h>

#define roundup(x,a)						(((x)+(a)-1)&(~((a)-1)))

#define TILED_REGION_BASE_BOUNDARY			0x10000
#define TILED_REGION_SIZE_BOUNDARY			0x10000

static u_int g_uTotalMemRequired = 0;
static u_int g_uTotalMemAllocated = 0;

// each tiled region must have a unique index.
// If we plan to alloc & free frame buffers a lot we should handle this as 15 availability slots rather than just one incrementing value.
static u_int g_usTile = 0;


void Asura_PS3_FrameBuffer::ResolveBackBufferNonBlock( Asura_PS3_Texture* pxResolveTexture, int iWidth, int iHeight )
{
	CellGcmTexture* xTexture	= pxResolveTexture->GetTexture();
	CellGcmTexture* xDest		= m_pxTextures[ m_ucCurrentSurface ]->GetTexture();

	CellGcmEnum eTransferMode = static_cast<CellGcmEnum>((xTexture->location<<1) | xDest->location); // see definition of CELL_GCM_TRANSFER_LOCAL_TO_MAIN etc

	u_int uBufferResolveHeight, uBufferResolveWidth;
	if( iWidth < 0 )
	{
		uBufferResolveHeight = xTexture->height;
		uBufferResolveWidth = xTexture->width;
	}
	else
	{
		uBufferResolveHeight = iHeight;
		uBufferResolveWidth = iWidth;
	}

	u_int uBufferResolveBytesPerPixel = 4;
	switch( m_eFormat )
	{
	default:
		{
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA16F:
		{
			uBufferResolveWidth *= 2;
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA32F:
		{
			uBufferResolveWidth *= 4;
			break;
		}
	}

	cellGcmSetTransferImageUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), eTransferMode, xTexture->offset, xTexture->pitch, 0, 0, xDest->offset, xDest->pitch, 0, 0, uBufferResolveWidth, uBufferResolveHeight, uBufferResolveBytesPerPixel );
	Asura_PS3_Graphics::Flush();
}

void Asura_PS3_FrameBuffer::ResolveBackBuffer( Asura_PS3_Texture* pxResolveTexture, int iWidth, int iHeight )
{
	CellGcmTexture* xTexture	= pxResolveTexture->GetTexture();
	CellGcmTexture* xDest		= m_pxTextures[ m_ucCurrentSurface ]->GetTexture();

	CellGcmEnum eTransferMode = static_cast<CellGcmEnum>((xTexture->location<<1) | xDest->location); // see definition of CELL_GCM_TRANSFER_LOCAL_TO_MAIN etc

	u_int uBufferResolveHeight, uBufferResolveWidth;
	if( iWidth < 0 )
	{
		uBufferResolveHeight = xTexture->height;
		uBufferResolveWidth = xTexture->width;
	}
	else
	{
		uBufferResolveHeight = iHeight;
		uBufferResolveWidth = iWidth;
	}

	u_int uBufferResolveBytesPerPixel = 4;
	switch( m_eFormat )
	{
	default:
		{
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA16F:
		{
			uBufferResolveWidth *= 2;
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA32F:
		{
			uBufferResolveWidth *= 4;
			break;
		}
	}

	cellGcmSetTransferImageUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), eTransferMode, xTexture->offset, xTexture->pitch, 0, 0, xDest->offset, xDest->pitch, 0, 0, uBufferResolveWidth, uBufferResolveHeight, uBufferResolveBytesPerPixel );
    Asura_PS3_Helper::WriteBlockingWaitForBackEndCmds( Asura_PS3_Graphics::GetGcmCurrentContext() );
	Asura_PS3_Graphics::Flush();
}

void Asura_PS3_FrameBuffer::ResolveToRenderTarget( Asura_PS3_FrameBuffer* pxFrameBuffer )
{
	Asura_PS3_Texture* xPS3Texture = pxFrameBuffer->GetRTAsTexture( ASURA_RENDER_TARGET_CURRENT_FRONT_BUFFER );
	ResolveToTexture( xPS3Texture );
}

void Asura_PS3_FrameBuffer::ResolveToCubemapTexture( Asura_PS3_Texture *pxResolveTexture, int iFace )
{
	if( iFace < 0 || iFace > 5 )
	{
		Asura_Assert( false, "Invalid cube map texture face." );
		return;
	}

	CellGcmTexture* xSourceTexture = m_pxTextures[ 0 ]->GetTexture();
	CellGcmTexture* xTexture = pxResolveTexture->GetTexture();

	static CellGcmTransferScale		xScale;
	static CellGcmTransferSurface	xDest;

	xScale.conversion	= CELL_GCM_TRANSFER_CONVERSION_DITHER;
	xScale.operation	= CELL_GCM_TRANSFER_OPERATION_SRCCOPY;

	xScale.format		= CELL_GCM_TRANSFER_SCALE_FORMAT_A8R8G8B8;

	xScale.clipX = 0;
	xScale.clipY = 0;
	xScale.clipW = xTexture->width;
	xScale.clipH = xTexture->height;

	xScale.outX = 0;
	xScale.outY = 0;
	xScale.outW = xTexture->width;
	xScale.outH = xTexture->height;

	xScale.ratioX = cellGcmGetFixedSint32( ( float )xSourceTexture->width  / ( float )xTexture->width );
	xScale.ratioY = cellGcmGetFixedSint32( ( float )xSourceTexture->height / ( float )xTexture->height );

	xScale.inX = cellGcmGetFixedUint16( 0.0f );
	xScale.inY = cellGcmGetFixedUint16( 0.0f );
	xScale.inW = xSourceTexture->width;
	xScale.inH = xSourceTexture->height;

	xScale.origin = CELL_GCM_TRANSFER_ORIGIN_CENTER;
	xScale.interp = CELL_GCM_TRANSFER_INTERPOLATOR_FOH;

	xScale.pitch = xSourceTexture->pitch;
	xScale.offset = xSourceTexture->offset;

	xDest.format = CELL_GCM_TRANSFER_SURFACE_FORMAT_A8R8G8B8;
	xDest.pitch = xTexture->pitch;
	cellGcmAddressToOffset( ( (char*)pxResolveTexture->GetTextureData() ) + ( pxResolveTexture->GetTextureFaceSize() * iFace ), &xDest.offset );

	if( xSourceTexture->location == CELL_GCM_LOCATION_LOCAL )
	{
		if( xTexture->location == CELL_GCM_LOCATION_LOCAL )
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_LOCAL_TO_LOCAL, CELL_GCM_TRANSFER_SURFACE );
		}
		else
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_LOCAL_TO_MAIN, CELL_GCM_TRANSFER_SURFACE );
		}
	}
	else
	{
		if( xTexture->location == CELL_GCM_LOCATION_LOCAL )
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_MAIN_TO_LOCAL, CELL_GCM_TRANSFER_SURFACE );
		}
		else
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_MAIN_TO_MAIN, CELL_GCM_TRANSFER_SURFACE );
		}
	}

	cellGcmSetTransferScaleSurfaceUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), &xScale, &xDest );

	Asura_PS3_Helper::WriteBlockingWaitForBackEndCmds( Asura_PS3_Graphics::GetGcmCurrentContext() );

	Asura_PS3_Graphics::Flush();
}

void Asura_PS3_FrameBuffer::ResolveToTexture( Asura_PS3_Texture* pxResolveTexture, bool bUseGammaCorrection )
{
	bool bUseAQuad = true;
	CellGcmTexture* xSourceTexture = m_pxTextures[ m_ucCurrentSurface ]->GetTexture();
	CellGcmTexture* xTexture = pxResolveTexture->GetTexture();

	uint8_t ucClearFormatIn = ~CELL_GCM_TEXTURE_UN;
	uint8_t ucFormatIn = xSourceTexture->format, ucFormatOut = xTexture->format;
	ucFormatIn &= ucClearFormatIn; ucFormatOut &= ucClearFormatIn;
	ucClearFormatIn = ~CELL_GCM_TEXTURE_LN;
	ucFormatIn &= ucClearFormatIn; ucFormatOut &= ucClearFormatIn;


	if( ucFormatIn == ucFormatOut && ( ucFormatIn == CELL_GCM_TEXTURE_A8R8G8B8 || ucFormatIn == CELL_GCM_TEXTURE_X32_FLOAT ) )
	{
		bUseAQuad = false;
	}

#ifdef ASURA_PS3_USE_GAMMA_CORRECTION
	if( bUseGammaCorrection )
	{
		CellGcmTexture* xTexture = pxResolveTexture->GetTexture();

		Asura_Overlay_Quad xQuad;

		xQuad.SetRectUV( 0.0f, 0.0f, 1.0f, 1.0f );
		xQuad.SetRectXY(0.0f, 0.0f, xTexture->width, xTexture->height );
		xQuad.TextureIndex = ASURA_TEXTURE_NONE;
		xQuad.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
		xQuad.Colour = 0xFFFFFFFF;

		bool bStencil = Asura_PS3_RenderingStates::GetStencilTestEnabled();
		Asura_PS3_RenderingStates::SetStencilTestEnabled(false);

		Asura_PS3_QuadOverlayRenderer::RenderGamma( xQuad, m_pxTextures[ m_ucCurrentSurface ], false, xTexture->width, xTexture->height );
		Asura_PS3_Graphics::Flush();

		Asura_PS3_RenderingStates::SetStencilTestEnabled(bStencil);

		return;
	}
#endif
	if( bUseAQuad )
	{
		Asura_Overlay_Quad xQuad;

		xQuad.SetRectUV( 0.0f, 0.0f, 1.0f, 1.0f );
		xQuad.SetRectXY(0.0f, 0.0f, xTexture->width, xTexture->height );
		xQuad.TextureIndex = ASURA_TEXTURE_NONE;
		xQuad.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
		xQuad.Colour = 0xFFFFFFFF;

		bool bStencil = Asura_PS3_RenderingStates::GetStencilTestEnabled();
		Asura_PS3_RenderingStates::SetStencilTestEnabled(false);

		Asura_PS3_QuadOverlayRenderer::RenderEx( xQuad, m_pxTextures[ m_ucCurrentSurface ], false, xTexture->width, xTexture->height );
		Asura_PS3_Graphics::Flush();

		Asura_PS3_RenderingStates::SetStencilTestEnabled(bStencil);

		return;
	}

	static CellGcmTransferScale		xScale;
	static CellGcmTransferSurface	xDest;
	int iMult = 1;
/*
	switch( ucFormatIn )
	{
	case CELL_GCM_TEXTURE_W16_Z16_Y16_X16_FLOAT:
		iMult = 2;
		break;
	case CELL_GCM_TEXTURE_W32_Z32_Y32_X32_FLOAT:
		iMult = 4;
		break;
	default:
		iMult = 1;
	}
*/	
	xScale.conversion	= CELL_GCM_TRANSFER_CONVERSION_DITHER;
	xScale.operation	= CELL_GCM_TRANSFER_OPERATION_SRCCOPY;

	xScale.format		= CELL_GCM_TRANSFER_SCALE_FORMAT_A8R8G8B8;

	xScale.clipX = 0;
	xScale.clipY = 0;
	xScale.clipW = xTexture->width * iMult;
	xScale.clipH = xTexture->height;

	xScale.outX = 0;
	xScale.outY = 0;
	xScale.outW = xTexture->width * iMult;
	xScale.outH = xTexture->height;

	xScale.ratioX = cellGcmGetFixedSint32( ( float )xSourceTexture->width * iMult  / ( float )xTexture->width * iMult );
	xScale.ratioY = cellGcmGetFixedSint32( ( float )xSourceTexture->height / ( float )xTexture->height );

	xScale.inX = cellGcmGetFixedUint16( 0.0f );
	xScale.inY = cellGcmGetFixedUint16( 0.0f );
	xScale.inW = xSourceTexture->width * iMult;
	xScale.inH = xSourceTexture->height;

	xScale.origin = CELL_GCM_TRANSFER_ORIGIN_CENTER;
	xScale.interp = CELL_GCM_TRANSFER_INTERPOLATOR_FOH;

	xScale.pitch = xSourceTexture->pitch;
	xScale.offset = xSourceTexture->offset;

	xDest.format = CELL_GCM_TRANSFER_SURFACE_FORMAT_A8R8G8B8;
	xDest.pitch = xTexture->pitch;
	xDest.offset = xTexture->offset;

	if( xSourceTexture->location == CELL_GCM_LOCATION_LOCAL )
	{
		if( xTexture->location == CELL_GCM_LOCATION_LOCAL )
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_LOCAL_TO_LOCAL, CELL_GCM_TRANSFER_SURFACE );
		}
		else
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_LOCAL_TO_MAIN, CELL_GCM_TRANSFER_SURFACE );
		}
	}
	else
	{
		if( xTexture->location == CELL_GCM_LOCATION_LOCAL )
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_MAIN_TO_LOCAL, CELL_GCM_TRANSFER_SURFACE );
		}
		else
		{
			cellGcmSetTransferScaleModeUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRANSFER_MAIN_TO_MAIN, CELL_GCM_TRANSFER_SURFACE );
		}
	}

	cellGcmSetTransferScaleSurfaceUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), &xScale, &xDest );

    Asura_PS3_Helper::WriteBlockingWaitForBackEndCmds( Asura_PS3_Graphics::GetGcmCurrentContext() );

	Asura_PS3_Graphics::Flush();
}

u_int	Asura_PS3_FrameBuffer::CalcColourPitch( u_int uWidth, u_int uBytesPerPixel )
{
	u_int uResult = 0;

	u_int uBufferWidth	= roundup( uWidth, 256 / uBytesPerPixel );
	uResult				= Asura_PS3_Helper::GetPitchSize( uBufferWidth * uBytesPerPixel );

	return uResult;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FrameBuffer::Asura_PS3_FrameBuffer
// Author       : Catalin Arsenescu (27-March-2007)
//
// Notes:
// if pxDepthBuffer is specified the size must match the Height and Width of the render target.
////////////////////////////////////////////////////////////////////////////////
Asura_PS3_FrameBuffer::Asura_PS3_FrameBuffer( bool bWaitForIdleState, u_int uWidth, u_int uHeight, unsigned char ucNumberOfRenderTargets, ASURA_PS3_FRAMEBUFFER_FORMAT eFormat, bool bHasDepthBuffer, bool bUseDoubleBuffer, char* szName, char cDisplayID, bool bTiled, void** ppxDepthBuffer, bool bUseZCullMemory, u_int uAllocWidth, u_int uAllocHeight, CellGcmEnum eLocation )
{
	unsigned char	ucBank = 0;	// rotate bank assignments for MRT, apparently it can help performance //BW I've tested a lot of different settings with no benefit.
	CellGcmTexture	xTextures[ 5 ];
	void*			paxTextureData[ 5 ];

	m_eFormat						= eFormat;
	m_ucCurrentSurface				= 0;
	m_ucNumberOfRenderTargets		= ucNumberOfRenderTargets;
	m_bDoubleBuffer					= bUseDoubleBuffer;
	m_cDisplayID					= cDisplayID;
	m_eLocation						= eLocation;

	// we want to swizzle shadow textures (z buffer with no colour buffer), but must not swizzle tiled memory.
	const bool bSwizzled = ( ucNumberOfRenderTargets == 0 && !bTiled );

	if( bWaitForIdleState )
	{
		Asura_PS3_Graphics::LockMutex();
		Asura_PS3_Helper::WaitForIdleState();
	}
	
	if( ucNumberOfRenderTargets > 4 )
	{
		Asura_Assert( false, "Asura_PS3_FrameBuffer::Asura_PS3_FrameBuffer -> Invalid number of render targets (0 - 3)" );
		return;
	}

	if( bUseDoubleBuffer && ucNumberOfRenderTargets != 1 )
	{
		Asura_Assert( false, "Asura_PS3_FrameBuffer::Asura_PS3_FrameBuffer -> Double buffers and MRT not implemented" );
	}
	
	u_int uBytesPerPixel;
	switch( eFormat )
	{
	default:
		{
			Asura_OutputToDebugger( ("Unknown format for frame buffer\n") );
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA16F:
		{
			uBytesPerPixel = 8;
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA:
		{
			uBytesPerPixel = 4;
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_RGBA32F:
		{
			uBytesPerPixel = 16;
			break;
		}
	case ASURA_PS3_FRAMEBUFFER_32F:
		{
			uBytesPerPixel = 4;
			break;
		}
	}

	// initialise textures, some of these values get updated later when we set up the buffers
	for( u_int uCounter = 0; uCounter < 5; ++uCounter )
	{
		m_pxTextures[uCounter] = NULL;
		xTextures[ uCounter ].format	= CELL_GCM_TEXTURE_LN | CELL_GCM_TEXTURE_UN;
		switch( eFormat )
		{
		case ASURA_PS3_FRAMEBUFFER_RGBA16F:
			{
				xTextures[ uCounter ].format |= CELL_GCM_TEXTURE_W16_Z16_Y16_X16_FLOAT;
				break;
			}
		case ASURA_PS3_FRAMEBUFFER_RGBA:
			{
				xTextures[ uCounter ].format |= CELL_GCM_TEXTURE_A8R8G8B8;
				break;
			}
		case ASURA_PS3_FRAMEBUFFER_RGBA32F:
			{
				xTextures[ uCounter ].format |= CELL_GCM_TEXTURE_W32_Z32_Y32_X32_FLOAT;
				break;
			}
		case ASURA_PS3_FRAMEBUFFER_32F:
			{
				xTextures[ uCounter ].format |= CELL_GCM_TEXTURE_X32_FLOAT;
//				xTextures[ uCounter ].format |= CELL_GCM_TEXTURE_Y16_X16_FLOAT;
				break;
			}
		}
		xTextures[ uCounter ].mipmap	= 1;
		xTextures[ uCounter ].dimension	= CELL_GCM_TEXTURE_DIMENSION_2;
		xTextures[ uCounter ].cubemap	= CELL_GCM_FALSE;
		xTextures[ uCounter ].remap		= CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_A,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );
		xTextures[ uCounter ].width		= uWidth;
		xTextures[ uCounter ].height	= uHeight;
		xTextures[ uCounter ].depth		= 1;
		xTextures[ uCounter ].location	= CELL_GCM_LOCATION_LOCAL;
		xTextures[ uCounter ]._padding	= 0;
		xTextures[ uCounter ].offset	= 0;
		xTextures[ uCounter ].pitch		= 0;
	}

	const u_char ucMaxRenderSurf	= bUseDoubleBuffer ? 2 : 1;
	const u_int uColorPitch			= CalcColourPitch( uWidth, uBytesPerPixel );

	u_int uDepthWidth, uDepthPitch;
	if( ucNumberOfRenderTargets == 0 )	//shadow buffers (depth-only) must access depth as depth
	{
		uDepthWidth		= uWidth;
		xTextures[ 4 ].format = CELL_GCM_TEXTURE_UN | CELL_GCM_TEXTURE_DEPTH24_D8;
		xTextures[ 4 ].format |= bSwizzled ? CELL_GCM_TEXTURE_SZ : CELL_GCM_TEXTURE_LN;
		uDepthPitch		= roundup( uDepthWidth * 4, 0x40 ) ; // 24Bits for Z and 8Bits for Stencil
	}
	else	//non-shadow buffers want depth so it can be read in shaders (if they ever read it), i.e. RGBA then read using Deferred_GetDepth() in MaterialResponse_Shared.ps3FXH
	{
		uDepthWidth		= roundup( roundup( uWidth, 256 / 4 ), 0x40 );
		xTextures[ 4 ].format = CELL_GCM_TEXTURE_LN | CELL_GCM_TEXTURE_NR | CELL_GCM_TEXTURE_A8R8G8B8;
		uDepthPitch		= Asura_PS3_Helper::GetPitchSize( uDepthWidth * 4 ); // 24Bits for Z and 8Bits for Stencil
	}

	bool bInternAllocatedDepthBuffer = false;
	CellGcmEnum eDepthLocation = eLocation;
	if( bHasDepthBuffer )
	{
		if( ( ppxDepthBuffer != NULL ) && ( *ppxDepthBuffer != NULL ) )
		{
			bInternAllocatedDepthBuffer	= false;
			eDepthLocation = Asura_PS3_VideoMemoryManager::IsVramAddress(*ppxDepthBuffer) ? CELL_GCM_LOCATION_LOCAL : CELL_GCM_LOCATION_MAIN;
		}
		else
		{
			bInternAllocatedDepthBuffer	= true;
			eDepthLocation = eLocation;
		}
	}


	for( u_int uCounter = 0; uCounter < 2; ++uCounter )
	{
		switch( eFormat )
		{
			default:
				{
					Asura_OutputToDebugger( ("Unknown format for frame buffer\n") );
				}
			case ASURA_PS3_FRAMEBUFFER_RGBA16F:
				{
					uBytesPerPixel = 8;
					m_xRenderSurface[ uCounter ].colorFormat	= CELL_GCM_SURFACE_F_W16Z16Y16X16;
					break;
				}
			case ASURA_PS3_FRAMEBUFFER_RGBA:
				{
					uBytesPerPixel = 4;
					m_xRenderSurface[ uCounter ].colorFormat	= CELL_GCM_SURFACE_A8R8G8B8;
					break;
				}
			case ASURA_PS3_FRAMEBUFFER_RGBA32F:
				{
					uBytesPerPixel = 16;
					m_xRenderSurface[ uCounter ].colorFormat	= CELL_GCM_SURFACE_F_W32Z32Y32X32;
					break;
				}
			case ASURA_PS3_FRAMEBUFFER_32F:
				{
					uBytesPerPixel = 4;
					m_xRenderSurface[ uCounter ].colorFormat	= CELL_GCM_SURFACE_F_X32;
					break;
				}
		}

		switch( ucNumberOfRenderTargets )
		{
			case 0:
				{
					m_xRenderSurface[ uCounter ].colorTarget	= CELL_GCM_SURFACE_TARGET_NONE;
					break;
				}
			case 1:
				{
					m_xRenderSurface[ uCounter ].colorTarget	= CELL_GCM_SURFACE_TARGET_0;
					break;
				}
			case 2:
				{
					m_xRenderSurface[ uCounter ].colorTarget	= CELL_GCM_SURFACE_TARGET_MRT1;
					break;
				}
			case 3:
				{
					m_xRenderSurface[ uCounter ].colorTarget	= CELL_GCM_SURFACE_TARGET_MRT2;
					break;
				}
			case 4:
				{
					m_xRenderSurface[ uCounter ].colorTarget	= CELL_GCM_SURFACE_TARGET_MRT3;
					break;
				}
		}

		m_xRenderSurface[ uCounter ].colorLocation[ 0 ]	= eLocation;
		m_xRenderSurface[ uCounter ].colorLocation[ 1 ]	= eLocation;
		m_xRenderSurface[ uCounter ].colorLocation[ 2 ]	= eLocation;
		m_xRenderSurface[ uCounter ].colorLocation[ 3 ]	= eLocation;

		m_xRenderSurface[ uCounter ].colorOffset[0]		= 0;
		m_xRenderSurface[ uCounter ].colorOffset[1]		= 0;
		m_xRenderSurface[ uCounter ].colorOffset[2]		= 0;
		m_xRenderSurface[ uCounter ].colorOffset[3]		= 0;

		m_xRenderSurface[ uCounter ].colorPitch[0]		= 64;
		m_xRenderSurface[ uCounter ].colorPitch[1]		= 64;
		m_xRenderSurface[ uCounter ].colorPitch[2]		= 64;
		m_xRenderSurface[ uCounter ].colorPitch[3]		= 64;

		m_xRenderSurface[ uCounter ].depthFormat		= CELL_GCM_SURFACE_Z24S8;
		m_xRenderSurface[ uCounter ].depthLocation		= eDepthLocation;

		m_xRenderSurface[ uCounter ].type				= (bSwizzled) ? CELL_GCM_SURFACE_SWIZZLE : CELL_GCM_SURFACE_PITCH;

		m_xRenderSurface[ uCounter ].antialias			= CELL_GCM_SURFACE_CENTER_1;

		m_xRenderSurface[ uCounter ].width				= uWidth;
		m_xRenderSurface[ uCounter ].height				= uHeight;
		m_xRenderSurface[ uCounter ].x					= 0;
		m_xRenderSurface[ uCounter ].y					= 0;
	}

	const u_int uHeightRound = (eLocation == CELL_GCM_LOCATION_LOCAL) ? 32 : 64;
	const u_int uDepthHeightRound = (eDepthLocation == CELL_GCM_LOCATION_LOCAL) ? 32 : 64;
//BW todo: use smaller rounding for non-tiled buffers

	u_int uColorSize	= roundup( uColorPitch * roundup( uHeight, uHeightRound ), TILED_REGION_SIZE_BOUNDARY );

	u_int uDepthSize;
	if( ucNumberOfRenderTargets > 0 )
	{
		uDepthSize	= roundup( uDepthPitch * roundup( uHeight, uDepthHeightRound ), TILED_REGION_SIZE_BOUNDARY );
	}
	else
	{
		uDepthSize	= uDepthPitch * uHeight;
	}

	u_int		uWastageColourSize	= 0;
	u_int		uWastageDepthSize	= 0;

	// Allocate additional memory, to make total memory usage equivalent to allocating a larger buffer
	if ( uAllocWidth > uWidth || uAllocHeight > uHeight )
	{
		uAllocWidth = Asura_Maths::Max(uAllocWidth,uWidth);
		uAllocHeight = Asura_Maths::Max(uAllocHeight,uHeight);

		u_int	uMaxColorPitch		= CalcColourPitch( uAllocWidth, uBytesPerPixel );

		uWastageColourSize			= roundup( uMaxColorPitch * roundup( uAllocHeight, uHeightRound ), TILED_REGION_SIZE_BOUNDARY ) - uColorSize;
		
		u_int	uMaxDepthWith		= roundup( roundup( uAllocWidth, 256 / 4 ), 0x40 );
		u_int	uMaxDepthPitch		= Asura_PS3_Helper::GetPitchSize( uMaxDepthWith * 4 ); // 24Bits for Z and 8Bits for Stencil

		if( ucNumberOfRenderTargets > 0 )
		{
			uWastageDepthSize		= roundup( uMaxDepthPitch * roundup( uAllocHeight, uHeightRound ), TILED_REGION_SIZE_BOUNDARY ) - uDepthSize;
		}
		else
		{
			uWastageDepthSize	= ( uMaxDepthPitch * uAllocHeight ) - uDepthSize;
		}
	}

	for(int i = 0; i < 5; ++i )
	{
		m_apxBufferAddress[ i ] = NULL;
		m_apxWasteBufferAddress[ i ] = NULL;
	}

	int				iCurrentBuffer = 0;
	for( u_int uCounter = 0; uCounter < ucMaxRenderSurf; ++uCounter )
	{
		for( int i = 0 ; i < ucNumberOfRenderTargets; ++i )
		{
			void* pxColorAddress;
			
			if ( eLocation == CELL_GCM_LOCATION_LOCAL )
			{
				pxColorAddress = ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( uColorSize, TILED_REGION_BASE_BOUNDARY );
				if ( uWastageColourSize > 0 ) m_apxWasteBufferAddress[ iCurrentBuffer ] = ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( uWastageColourSize, TILED_REGION_BASE_BOUNDARY );
			}
			else
			{
				pxColorAddress = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( uColorSize, TILED_REGION_BASE_BOUNDARY );
				if ( uWastageColourSize > 0 ) m_apxWasteBufferAddress[ iCurrentBuffer ] = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( uWastageColourSize, TILED_REGION_BASE_BOUNDARY );
			}
			
			u_int uColorOffset = Asura_PS3_Helper::AddressToOffset( pxColorAddress );

			//Get measure of memory used by buffers & buffer sizes, nums.
			Asura_OutputToDebugger_Retail(( "Frame Buffer %d %d: %4dx%3d %2dbpp, %4d pitch, needs %4dk allocated %4dk, wasted %4dk, total alloc (inc waste) %4dk, @ 0x%08x   - \"%s\"\n",
				uCounter,i,
				uWidth, uHeight, uBytesPerPixel*8, uColorPitch, (uWidth*uHeight*uBytesPerPixel)/1024, uColorSize/1024, uWastageColourSize/1024, (uColorSize + uWastageColourSize) / 1024,  pxColorAddress, szName ));
			g_uTotalMemRequired += ( uWidth * uHeight * uBytesPerPixel );
			g_uTotalMemAllocated += uColorSize + uWastageColourSize;


			m_xRenderSurface[ uCounter ].colorOffset[ i ]	= uColorOffset;
			m_xRenderSurface[ uCounter ].colorPitch[ i ]	= uColorPitch;

			xTextures[ i + uCounter ].offset	= uColorOffset;
			xTextures[ i + uCounter ].pitch		= uColorPitch;
			xTextures[ i + uCounter ].location	= eLocation;
			paxTextureData[ i + uCounter ]		= pxColorAddress;

			Asura_Assert( !bTiled || g_usTile < 15, "Ran out of tile regions, all frame buffers created after this point will have sub-optimal performance" );
			//BW to fix this we could share tiled regions (the limit is on the number of different ones, there's no problem sharing them), particularly for MRT and shadows.
			// n.b. don't share frame buffer with z-buffer, it makes things slower.
			
			if ( bTiled && g_usTile < 15 )
			{
				int iErr = cellGcmSetTileInfo( g_usTile, eLocation, uColorOffset, uColorSize, uColorPitch, CELL_GCM_COMPMODE_DISABLED, 0, (ucBank++)&3 );
				Asura_Assert( iErr == CELL_OK, "cellGcmSetTileInfo: failed - bad parameters?\nThis buffer will not be in tiled memory, the game will be SLOW!" );
				iErr = cellGcmBindTile( g_usTile );
				Asura_Assert( iErr != CELL_GCM_ERROR_INVALID_VALUE, "cellGcmBindTile: Cannot tile this frame buffer, ran out of indices" );
				Asura_Assert( iErr != CELL_GCM_ERROR_ADDRESS_OVERWRAP, "cellGcmBindTile: Cannot tile this frame buffer, tiled regions would overlap!" );
				Asura_Assert( iErr == CELL_OK || iErr == CELL_GCM_ERROR_INVALID_VALUE || iErr == CELL_GCM_ERROR_ADDRESS_OVERWRAP, "cellGcmBindTile: unknown error or stupid parameters" );
				g_usTile++; // each tiled region that we bind must have a unique index
			}

			m_apxBufferAddress[ iCurrentBuffer ] = pxColorAddress;
			++iCurrentBuffer;
		}
		if( m_cDisplayID >= 0 )
		{
			int iRet = cellGcmSetDisplayBuffer( m_cDisplayID + uCounter, m_xRenderSurface[ uCounter ].colorOffset[ 0 ], m_xRenderSurface[ uCounter ].colorPitch[ 0 ], m_xRenderSurface[ uCounter ].width, m_xRenderSurface[ uCounter ].height );
			if( iRet != CELL_OK )
			{
				printf( "Asura_PS3_FrameBuffer::Bind -> cellGcmSetDisplayBuffer failed (game will crash)\n" );
			}
		}
	}

	u_int uDepthOffset;
	void* pxDepthAddress;
	if( bInternAllocatedDepthBuffer )
	{
		if ( eDepthLocation == CELL_GCM_LOCATION_LOCAL )
		{
			pxDepthAddress = ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( uDepthSize, TILED_REGION_BASE_BOUNDARY );
			if ( uWastageDepthSize ) m_apxWasteBufferAddress[ iCurrentBuffer ] = ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( uWastageDepthSize, TILED_REGION_BASE_BOUNDARY );
		}
		else
		{
			pxDepthAddress = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( uDepthSize, TILED_REGION_BASE_BOUNDARY );
			if ( uWastageDepthSize ) m_apxWasteBufferAddress[ iCurrentBuffer ] = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( uWastageDepthSize, TILED_REGION_BASE_BOUNDARY );
		}
		uDepthOffset = Asura_PS3_Helper::AddressToOffset( pxDepthAddress );

		Asura_OutputToDebugger_Retail(( "Depth Buffer: %4dx%3d %2dbpp, needs %4dk allocated %4dk, wasted %4dk, total alloc (inc waste) %4dk, @ 0x%08x   - \"%s\"\n",
			uWidth, uHeight, 4*8, (uWidth*uHeight*4)/1024, uDepthSize/1024, uWastageDepthSize/1024, (uDepthSize + uWastageDepthSize) / 1024,  pxDepthAddress, szName ));
		g_uTotalMemRequired += ( uWidth * uHeight * 4 );
		g_uTotalMemAllocated += uDepthSize + uWastageDepthSize;

		Asura_Assert( !bTiled || g_usTile < 15, "Ran out of tile regions, all frame buffers created after this point will have sub-optimal performance" );
		//BW see comment above

		if ( bTiled && g_usTile < 15 )
		{
			int iErr = cellGcmSetTileInfo( g_usTile, CELL_GCM_LOCATION_LOCAL, uDepthOffset, uDepthSize, uDepthPitch, CELL_GCM_COMPMODE_DISABLED, 0, (ucBank++)&3 );
			Asura_Assert( iErr == CELL_OK, "cellGcmSetTileInfo: failed - bad parameters?\nThis buffer will not be in tiled memory, the game will be SLOW!" );
			iErr = cellGcmBindTile( g_usTile );
			Asura_Assert( iErr != CELL_GCM_ERROR_INVALID_VALUE, "cellGcmBindTile: Cannot tile this frame buffer, ran out of indices" );
			Asura_Assert( iErr != CELL_GCM_ERROR_ADDRESS_OVERWRAP, "cellGcmBindTile: Cannot tile this frame buffer, tiled regions would overlap!" );
			Asura_Assert( iErr == CELL_OK || iErr == CELL_GCM_ERROR_INVALID_VALUE || iErr == CELL_GCM_ERROR_ADDRESS_OVERWRAP, "cellGcmBindTile: unknown error or stupid parameters" );
			g_usTile++; // each tiled region that we bind must have a unique index
		}
		
		m_apxBufferAddress[ iCurrentBuffer ] = pxDepthAddress;
		++iCurrentBuffer;
		if( ppxDepthBuffer )
		{
			*ppxDepthBuffer	= pxDepthAddress;		// return the depth address
		}
		xTextures[ 4 ].offset = uDepthOffset;
		xTextures[ 4 ].pitch = uDepthPitch;
		xTextures[ 4 ].location = eDepthLocation;
		
		// remap components, if I can, so stencil-> alpha & depth -> rgb. this is free & helps simplify a couple of debug things.
		xTextures[ 4 ].remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY,
			CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_FROM_A, CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G,
			CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );
		
		m_pxTextures[ 4 ] = new Asura_PS3_Texture( &xTextures[ 4 ] );
		m_pxTextures[ 4 ]->SetTextureData( pxDepthAddress );
	}
	else
	{
		if( ppxDepthBuffer )
		{
			pxDepthAddress = *ppxDepthBuffer;
			uDepthOffset = Asura_PS3_Helper::AddressToOffset( pxDepthAddress );
		}
		else
		{
			uDepthOffset	= 0;
			uDepthPitch		= 64;
		}
	}

	Asura_OutputToDebugger_Retail(( "Frame Buffer Total: needed %.2fMB allocated %.2fMB\n", (float)g_uTotalMemRequired/(1024.0f*1024.0f), (float)g_uTotalMemAllocated/(1024.0f*1024.0f) ));

	m_xRenderSurface[ 0 ].depthOffset	= uDepthOffset;
	m_xRenderSurface[ 0 ].depthPitch	= uDepthPitch;
	m_xRenderSurface[ 1 ].depthOffset	= uDepthOffset;
	m_xRenderSurface[ 1 ].depthPitch	= uDepthPitch;

	if( !bUseDoubleBuffer )
	{
		m_xRenderSurface[ 1 ].colorOffset[ 0 ]	= m_xRenderSurface[ 0 ].colorOffset[ 0 ];
		m_xRenderSurface[ 1 ].colorPitch[ 0 ]	= m_xRenderSurface[ 0 ].colorPitch[ 0 ];
	}

	Asura_Mem::Copy( &m_axSingleRT[ 0 ], &m_xRenderSurface[ 0 ], sizeof( m_xRenderSurface[ 0 ] ) );
	Asura_Mem::Copy( &m_axSingleRT[ 1 ], &m_xRenderSurface[ 1 ], sizeof( m_xRenderSurface[ 0 ] ) );
	Asura_Mem::Copy( &m_axSingleRT[ 2 ], &m_xRenderSurface[ 0 ], sizeof( m_xRenderSurface[ 0 ] ) );
	for( int i = 0; i < 3; ++i )
	{
		m_axSingleRT[ i ].colorOffset[ 1 ]	= 0;
		m_axSingleRT[ i ].colorOffset[ 2 ]	= 0;
		m_axSingleRT[ i ].colorOffset[ 3 ]	= 0;
		m_axSingleRT[ i ].colorPitch[ 1 ]	= 64;
		m_axSingleRT[ i ].colorPitch[ 2 ]	= 64;
		m_axSingleRT[ i ].colorPitch[ 3 ]	= 64;
		m_axSingleRT[ i ].colorTarget		= CELL_GCM_SURFACE_TARGET_0;
	}
	m_axSingleRT[ 2 ].colorOffset[ 0 ]		= 0;							//Z Pass only
	m_axSingleRT[ 2 ].colorPitch[ 0 ]		= 64;							//Z Pass only
	m_axSingleRT[ 2 ].colorTarget			= CELL_GCM_SURFACE_TARGET_NONE; //Z Pass only

	u_int uMaxTextures = ucMaxRenderSurf;
	if( ucNumberOfRenderTargets > ucMaxRenderSurf )
	{
		uMaxTextures = ucNumberOfRenderTargets;
	}
	for( u_int uCounter = 0; uCounter < uMaxTextures; ++uCounter )
	{
		m_pxTextures[ uCounter ] = new Asura_PS3_Texture( &xTextures[ uCounter ] );
		m_pxTextures[ uCounter ]->SetTextureData( paxTextureData[ uCounter ] );
	}

	m_bIsInSingleRT = false;

	if( bUseZCullMemory && bInternAllocatedDepthBuffer )
	{
		if( m_cDisplayID < 0 )
		{
			printf("You must assign a Display ID to be able to use ZCullMemory\n");
		}
		u_int uZCullStart = 0;
		for( u_int uCounter = 0; uCounter < ucMaxRenderSurf; ++uCounter )
		{
			cellGcmSetZcull( m_cDisplayID + uCounter, m_xRenderSurface[ 0 ].depthOffset, uDepthWidth, roundup( m_xRenderSurface[ 0 ].height, 0x40 ), uZCullStart, CELL_GCM_ZCULL_Z24S8, CELL_GCM_SURFACE_CENTER_1,
					CELL_GCM_ZCULL_LESS, /*CELL_GCM_ZCULL_MSB*/CELL_GCM_ZCULL_LONES,
					//BW ^is that right?
					CELL_GCM_SCULL_SFUNC_ALWAYS, 0x80, 0xff );

			uZCullStart += 2048 * roundup( m_xRenderSurface[ 0 ].height, 0x40 );
		}
	}
	if( bWaitForIdleState )
	{
		Asura_PS3_Graphics::UnlockMutex();
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FrameBuffer::~Asura_PS3_FrameBuffer
// Author       : Catalin Arsenescu (27-March-2007)
////////////////////////////////////////////////////////////////////////////////
Asura_PS3_FrameBuffer::~Asura_PS3_FrameBuffer()
{
	for(int i = 0; i < 5; ++i )
	{
		if( m_apxBufferAddress[ i ] != NULL )
		{
			if ( m_eLocation == CELL_GCM_LOCATION_LOCAL )
			{
				Asura_PS3_VideoMemoryManager::FreeLocal( m_apxBufferAddress[ i ] );
				if ( m_apxWasteBufferAddress[ i ] ) Asura_PS3_VideoMemoryManager::FreeLocal( m_apxWasteBufferAddress[ i ] );
			}
			else
			{
				Asura_PS3_VideoMemoryManager::FreeMain( m_apxBufferAddress[ i ] );
				if ( m_apxWasteBufferAddress[ i ] ) Asura_PS3_VideoMemoryManager::FreeMain( m_apxWasteBufferAddress[ i ] );
			}
			m_apxBufferAddress[ i ] = NULL;
			m_apxWasteBufferAddress[ i ] = NULL;
		}
		delete m_pxTextures[ i ];
		m_pxTextures[ i ] = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FrameBuffer::Bind
// Author       : Catalin Arsenescu (27-March-2007)
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FrameBuffer::Bind( bool bSetViewport )
{
	if( m_ucNumberOfRenderTargets == 1 )
	{
		m_bIsInSingleRT = true;
	}
	else
	{
		Asura_PS3_RenderingStates::RequestAlphaTest(false);
		m_bIsInSingleRT = false;
	}

	Asura_PS3_RenderingStates::SetSurface( &m_xRenderSurface[ m_ucCurrentSurface ] );
	if ( bSetViewport ) Asura_PS3_RenderingStates::SetViewport( m_xRenderSurface[0].width, m_xRenderSurface[0].height );
}

void Asura_PS3_FrameBuffer::BindAsSingleRT( bool bSetViewport )
{
	m_bIsInSingleRT = true;
	Asura_PS3_RenderingStates::SetSurface( &m_axSingleRT[ m_ucCurrentSurface ] );
	if ( bSetViewport ) Asura_PS3_RenderingStates::SetViewport( m_xRenderSurface[0].width, m_xRenderSurface[0].height );
}

void Asura_PS3_FrameBuffer::BindAsZPassOnly( bool bSetViewport )
{
	m_bIsInSingleRT = true;
	Asura_PS3_RenderingStates::SetSurface( &m_axSingleRT[ 2 ] );
	if ( bSetViewport ) Asura_PS3_RenderingStates::SetViewport( m_xRenderSurface[0].width, m_xRenderSurface[0].height );
}

void Asura_PS3_FrameBuffer::RestoreBind( void )
{
	if( m_bIsInSingleRT )
	{
		Bind();
	}
}

Asura_PS3_Texture* Asura_PS3_FrameBuffer::GetRTAsTexture( u_int uRenderTarget )
{
	switch( uRenderTarget )
	{
	case ASURA_RENDER_TARGET_CURRENT_FRONT_BUFFER:
		{
			return m_pxTextures[ m_ucCurrentSurface ];
		}
	default:
		{
			return m_pxTextures[ uRenderTarget ];
		}
	}
}

Asura_PS3_Texture* Asura_PS3_FrameBuffer::GetZBufferAsTexture( void )
{
	return m_pxTextures[ 4 ];
}

void Asura_PS3_FrameBuffer::Flip( void )
{
	if( m_cDisplayID >= 0)
	{
		Asura_PS3_SwapChain::Present( m_cDisplayID + m_ucCurrentSurface );
	}
	else
	{
		Asura_OutputToDebugger( ( "Trying to flip using an invalid display ID\n" ) );
	}
	if( m_bDoubleBuffer )
	{
		m_ucCurrentSurface = ( m_ucCurrentSurface + 1 ) % 2;
	}
}

#ifdef ASURA_ASSERT
void Asura_PS3_FrameBuffer::AssertFlip( void )
{
	if( m_cDisplayID >= 0)
	{
		Asura_PS3_SwapChain::PresentAssert( m_cDisplayID + m_ucCurrentSurface );
	}
	else
	{
		Asura_OutputToDebugger( ( "Trying to flip using an invalid display ID\n" ) );
	}
	if( m_bDoubleBuffer )
	{
		m_ucCurrentSurface = ( m_ucCurrentSurface + 1 ) % 2;
	}
}
#endif

CellGcmSurface* Asura_PS3_FrameBuffer::GetSurface( void )
{
	return &m_xRenderSurface[ m_ucCurrentSurface ];
}

#undef roundup
