///////////////////////////////////////////////////////////////////////////////
//
// File		: Asura_PS3_GCM_IndexBuffer.cpp
// Author	: Andy Tate (22-Jul-2005)
// Modified	: Catalin Arsenescu (27-March-2007)
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Rendering/Asura_PS3_GCM_IndexBuffer.h"

// Engine independent.
#include "Debugging/Asura_Assert.h"

// Engine platform-specific.
#include "Rendering/Asura_PS3_ScratchRenderMemory.h"
#include "Rendering/Asura_PS3_VideoMemoryManager.h"
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_GCM_IndexBuffer::Initialise
// Author		: Andy Tate (22-Jul-2005)
// Modified		: Catalin Arsenescu (27-March-2007)
//
// Description  : Handles the initialisation of the index-buffer.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GCM_IndexBuffer::Initialise( const u_int uNumberOfVertices, u_char ucIndexArrayType, u_char ucLocation, bool bForceLocation )
{
	Asura_Assert	((0 == m_uIndexBufferSize), "Index-buffer has already been created.");
	if( m_uIndexBufferSize != 0 )
	{
		if(m_uIndexBufferSize < uNumberOfVertices * m_ucIndexSize)
		{
			Shutdown();
		}
		else
		{
			return;
		}
	}

	m_ucLocation = ucLocation;
	m_ucIndexArrayType = ucIndexArrayType;
	m_bUsingGlobalAllocator = true;

	if( m_ucIndexArrayType == CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16 )
	{
		m_ucIndexSize = 2;
	}
	else
	{
		m_ucIndexSize = 4;
	}

	m_uIndexBufferSize	= uNumberOfVertices * m_ucIndexSize;
	
	if( !bForceLocation )
	{
		if( Asura_PS3_VideoMemoryManager::OverrideLocation( m_ucLocation ) )
		{
		}
#ifdef ASURA_PS3_USE_BALANCED_MEMORY_USAGE
		else
		{
			if( Asura_PS3_VideoMemoryManager::GetFreeMain() > Asura_PS3_VideoMemoryManager::GetFreeLocal() )
			{
				m_ucLocation	= CELL_GCM_LOCATION_MAIN;
			}
			else
			{
				m_ucLocation	= CELL_GCM_LOCATION_LOCAL;
			}
		}
#endif // ASURA_PS3_USE_BALANCED_MEMORY_USAGE
	}

	// Try to alloc it in the requested memory, but if it's not possible place it in the other.
	if( m_ucLocation == CELL_GCM_LOCATION_LOCAL )
	{
		m_pxIndex	= ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( m_uIndexBufferSize, 128 );
	}
	else
	{
		m_pxIndex	= ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_uIndexBufferSize, 128 );
	}

	if( !m_pxIndex )
	{
		if( m_ucLocation == CELL_GCM_LOCATION_LOCAL )
		{
			m_ucLocation	= CELL_GCM_LOCATION_MAIN;
			m_pxIndex		= ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_uIndexBufferSize, 128 );
		}
		else
		{
			m_ucLocation	= CELL_GCM_LOCATION_LOCAL;
			m_pxIndex		= ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( m_uIndexBufferSize, 128 );
		}
	}

	if(m_pxIndex == NULL)
	{
		Asura_Assertf( false, ("[Asura_PS3_GCM_IndexBuffer::Initialise] - Failed to allocate memory for index buffer (%d vertices)\n", uNumberOfVertices));
	}
	else
	{
		cellGcmAddressToOffset( m_pxIndex, &m_uIndexOffset );
	}
}
/// End of function 'Asura_PS3_GCM_IndexBuffer::Initialise'. ///

void Asura_PS3_GCM_IndexBuffer::InitialiseFromBuffer( const u_int uNumberOfVertices )
{
	m_ucLocation			= CELL_GCM_LOCATION_MAIN;
	m_ucIndexArrayType		= CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16;
	m_bUsingGlobalAllocator	= false;
	m_ucIndexSize			= 2;

	if( uNumberOfVertices > 65530 )
	{
		m_ucIndexArrayType	= CELL_GCM_DRAW_INDEX_ARRAY_TYPE_32;
		m_ucIndexSize		= 4;
	}

	m_uIndexBufferSize		= uNumberOfVertices * m_ucIndexSize;
	m_pxIndex				= Asura_PS3_ScratchRenderMemory::AllocMainVideoMemory( m_uIndexBufferSize );
	cellGcmAddressToOffset( m_pxIndex, &m_uIndexOffset );
}
/// End of function 'Asura_PS3_GCM_IndexBuffer::Initialise'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_GCM_IndexBuffer::Shutdown
// Author       : Andy Tate
// Date         : 22-Jul-2005
//
// Description  : Handles the shutdown of the index-buffer.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GCM_IndexBuffer::Shutdown (void)
{
	// Delete the GL index-buffers...
	m_uIndexBufferSize	= 0;
	if( m_pxIndex && m_bUsingGlobalAllocator )
	{
		if( m_ucLocation == CELL_GCM_LOCATION_LOCAL )
		{
			Asura_PS3_VideoMemoryManager::FreeLocal( m_pxIndex );
		}
		else
		{
			Asura_PS3_VideoMemoryManager::FreeMain( m_pxIndex );
		}
	}
	m_pxIndex = NULL;
}
/// End of function 'Asura_PS3_GCM_IndexBuffer::Shutdown'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_GCM_IndexBuffer::RenderQuadsInline
// Author       : Ben Weston
// Description	: Copies vertex data into the command buffer. Use this if your vertex data is changing.
//				  There is no performance benefit from using an index buffer, since we de-reference the vertices here.
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GCM_IndexBuffer::RenderQuadsInline( Asura_PS3_GCM_VertexBuffer* pxVertexBuffer, const u_int uNumIndices, const u_int uStartIndex ) const
{
	u_int	uVertexSize = pxVertexBuffer->GetSize();
	void*	pxBuffer = NULL;
	char*	pcOut;
	u_int	iSize = uNumIndices * uVertexSize;
	const char*	pcVertexData = reinterpret_cast<const char*>(pxVertexBuffer->GetReadPtr());

	cellGcmSetDrawBeginUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_PRIMITIVE_QUADS );
	cellGcmSetDrawInlineArrayPointerUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), iSize / sizeof( float ), &pxBuffer );
	pcOut = reinterpret_cast<char*>(pxBuffer);

	// copy vertices, dereferencing indices
	for ( u_int i = uStartIndex; i < uStartIndex+uNumIndices; i++ )
	{
		u_int uIdx;
		if ( m_ucIndexArrayType == CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16 )
		{
			uIdx = (reinterpret_cast<uint16*>(m_pxIndex))[i];
		}
		else
		{
			uIdx = (reinterpret_cast<uint32*>(m_pxIndex))[i];
		}

		Asura_Mem::Copy( pcOut, pcVertexData + (uVertexSize * uIdx), uVertexSize );
		pcOut += uVertexSize;
	}

	cellGcmSetDrawEndUnsafeInline(Asura_PS3_Graphics::GetGcmCurrentContext() );

	Asura_Render::AddToTrianglesRendered( uNumIndices / 2 );
}


/// End of file. ///
