//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_RSXProfiling.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_PS3_USE_GPU_PROFILE_BARS

// This file's header.
#include "Rendering/Asura_PS3_RSXProfiling.h"

// Engine independent.
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_CPUProfile_Renderer.h"
#ifdef ASURA_USE_DEBUG_TEXT
#include "Debugging/Asura_DebugText.h"
#endif
#include "Debugging/Asura_Profiling_BarRenderer.h"
#include "Interface/Asura_CommandConsole.h"

// Engine platform-specific.
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_Helper.h"
#include "Rendering/Asura_PS3_VideoMemoryManager.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// Max sections that can be profiled within a single frame. Some will be
// repeated many times - eg. skin rendering - so account for that. Each
// section allowed requires 72 bytes of memory.
#ifndef ASURA_PS3_RSXPROFILING_MAXSECTIONS
#define ASURA_PS3_RSXPROFILING_MAXSECTIONS (2048)
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

// According to the documentation the counters used by cellGcmSetTimeStamp()
// are shared with cellGcmSetReport() and cellGcmSetRenderEnable(). This is
// not currently a problem since we're not doing any conditional rendering.
// However, if we do at any point then we'll have to start worrying about
// setting aside a specific area for this profiling, etc. One thing to
// consider is that the SN Tuner also makes use of these counters, at the
// beginning of the range. To avoid issues with that, start halfway through
// for now. Another option would be check if it's running or not, of course.
static const u_int uCOUNTER_INDEX_OFFSET = 1024;

// This obviously needs to be kept in sync with the Asura_PS3_RSXProfileIndex enumeration.
// This is only used by text dumps
static const char* aszRSX_PROFILE_NAMES[] =
{
	"entire frame",   // ASURA_PS3_RSX_PROFILE__ENTIREFRAME  

	"skybox",         // ASURA_PS3_RSX_PROFILE__SKYBOX       
	"env",            // ASURA_PS3_RSX_PROFILE__ENV          
	"smooth skins",   // ASURA_PS3_RSX_PROFILE__SMOOTHSKIN   
	"obj hiers",      // ASURA_PS3_RSX_PROFILE__OBJHIER      
	"botanicals",     // ASURA_PS3_RSX_PROFILE__BOTANICALS   
	"carpet",         // ASURA_PS3_RSX_PROFILE__CARPET       
	"create shadows", // ASURA_PS3_RSX_PROFILE__CREATESHADOWS
	"apply shadows",  // ASURA_PS3_RSX_PROFILE__APPLYSHADOWS 
	"photons",        // ASURA_PS3_RSX_PROFILE__PHOTONS      
	"pfx",            // ASURA_PS3_RSX_PROFILE__PROJECTORS   
	"fsfx",           // ASURA_PS3_RSX_PROFILE__FOG          
	"pfx",            // ASURA_PS3_RSX_PROFILE__PFX          
	"fsfx",           // ASURA_PS3_RSX_PROFILE__FSFX         
	"2d",             // ASURA_PS3_RSX_PROFILE__2D           
	"bfr resolves",   // ASURA_PS3_RSX_PROFILE__BFRRESOLVES  

    "proj 0",         // ASURA_PS3_RSX_PROFILE__PROJECTSPECIFIC_0
    "proj 1",
    "proj 2",
    "proj 3",
    "proj 4",
    "proj 5",
    "proj 6",
	"proj 7",         // ASURA_PS3_RSX_PROFILE__PROJECTSPECIFIC_7

	"misc"			 // ASURA_PS3_RSX_PROFILE__MISC
};
ASURA_STATIC_ASSERT( sizeof(aszRSX_PROFILE_NAMES) == ASURA_PS3_RSX_PROFILE__NUMINDICES * sizeof(const char*) );

static const u_int uRSX_PROFILE_UNPROFILED_ARGB = 0xFF604020; // brown

static const u_int auRSX_PROFILE_BAR_ARGBS[] =
{

    // The first colour is only used when we wrap and is intentionally
    // different from the real UNPROFILED colour above. This is just so we
    // can pass the timer durations array unmodified - see the rendering code.
    // The choice of colours is intended to be consistent with the 360
    // GPU profiling bars, though possibly nothing else.

    0xFFFFFFFF, // (white)           PI_ENTIREFRAME

	0xFF2222FF, // (dark blue)       ASURA_PS3_RSX_PROFILE__SKYBOX
	0xFF22FF22, // (green)           ASURA_PS3_RSX_PROFILE__ENV
	0xFFFF2222, // (red)             ASURA_PS3_RSX_PROFILE__SMOOTHSKIN
	0xFFFFFF22, // (yellow)          ASURA_PS3_RSX_PROFILE__OBJHIER
	0xFFFF22FF, // (bright pink)     ASURA_PS3_RSX_PROFILE__BOTANICALS
	0xFF222222, // (black)           ASURA_PS3_RSX_PROFILE__CARPET
	0xFFDD5555, // (orangey-red)     ASURA_PS3_RSX_PROFILE__CREATESHADOWS
	0xFF55DDDD, // (dull cyan)       ASURA_PS3_RSX_PROFILE__APPLYSHADOWS
	0xFFDDDD55, // (brighter yellow) ASURA_PS3_RSX_PROFILE__PHOTONS
	0xFFDD55DD, // (duller pink)     ASURA_PS3_RSX_PROFILE__PROJECTORS
	0xFFDDDDDD, // (light grey)      ASURA_PS3_RSX_PROFILE__FOG
	0xFF2222FF, // (dark blue)       ASURA_PS3_RSX_PROFILE__PFX
	0xFF22FFFF, // (cyan)            ASURA_PS3_RSX_PROFILE__FSFX
	0xFFFFFF22, // (yellow)          ASURA_PS3_RSX_PROFILE__2D
	0xFF5555DD, // (violet)          ASURA_PS3_RSX_PROFILE__BFRRESOLVES
    
    0xFFFF0000, // ASURA_PS3_RSX_PROFILE__PROJECTSPECIFIC_0
	0xFFFF8000,
	0xFFFFFF00,
	0xFF00FF00,
	0xFF008080,
	0xFF0000FF,
	0xFF800080,
	0xFFFF00FF, // ASURA_PS3_RSX_PROFILE__PROJECTSPECIFIC_7
	
	0xFF000000, // (black)			ASURA_PS3_RSX_PROFILE__MISC
    

    // Leave it to wrap after that for project bars (so the first one will
    // be the fake PI_ENTIREFRAME colour).
};

// G L O B A L S /////////////////////////////////////////////////////////////

// Best to toggle this by breakpointing BeginFrame() or doing it at the start,
// otherwise things may break horribly. Note this can't be static otherwise
// it gets optimised out. (Another option would be to make it volatile maybe.)
bool g_bEnableRSXProfiling = true;

bool g_bRenderRSXProfilingBars = true;
bool g_bRenderRSXProfilingText = false;
bool g_bOutputRSXProfilingFrameDump = false;

static Asura_PS3_RSXProfileIndex g_eCurrentProfileIndex = ASURA_PS3_RSX_PROFILE__NUMINDICES; // ASURA_PS3_RSX_PROFILE__NUMINDICES => not in a section currently (apart from maybe entire frame).

// First entry in current double buffer corresponds to entire frame. However,
// since the profile index here is known, we use the index value to count 
// the number of sections for this frame. If it's 0 then the buffer is
// invalid and somehow the RSX has not finished the previous frame yet.

// Note that the timer value data is much bigger than it needs to be. At the
// least it could be about half the size because we only need the 64-bit
// timer field within the CellGcmReportData struct; unfortunately I couldn't
// persuade the RSX to transfer this specific address using, eg.,
// cellGcmSetTransferData() since cellGcmAddressToOffset() didn't like the
// pointer returned by cellGcmGetReportDataAddress().

#ifdef ASURA_PS3_DO_NOT_USE_COMMAND_BUFFER
#define ASURA_PS3_RSXPROFILING_BUFFERS	2
#else
#define ASURA_PS3_RSXPROFILING_BUFFERS	3
#endif

static u_int*             g_apuTimerIndexBfrs[ASURA_PS3_RSXPROFILING_BUFFERS]; // Counter indices go here.
static u_int*             g_puTimerIndexDst;      // Current output pointer for indices, pointing into relevant buffer above.
static CellGcmReportData* g_apxTimerValueBfrs[ASURA_PS3_RSXPROFILING_BUFFERS]; // And the timing data goes here, 2 entries per timer index.
static CellGcmReportData* g_pxTimerValueDst;      // Current output pointer for timing data from RSX, pointing into relevant buffer above.
static u_int              g_uCurRSXDoubleBuffer;  // Which halves of the double buffers above the RSX is currently writing to.
static u_int              g_uSectionCount;        // Number of sections written into the buffers so far, including the initial one for the entire frame.

static int	g_aiStack[8];
static int	g_iStackIdx = 0;


// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static u_int GetCounterIdx( u_int uProfileIdx, bool bIsBeginSection )
{
    return( uCOUNTER_INDEX_OFFSET + (static_cast<u_int>(uProfileIdx) << 1) + (bIsBeginSection ? 0 : 1) );
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_PS3_RSXProfiling::Initialise( void )
{
    Asura_Assert( !g_apuTimerIndexBfrs[0], "Already initialised?" );

    for( u_int u = 0; u < ASURA_PS3_RSXPROFILING_BUFFERS; ++u )
    {
        g_apuTimerIndexBfrs[u] = static_cast<u_int*>( ASURA_PS3_VIDEOMEMORY_ALLOCMAIN(ASURA_PS3_RSXPROFILING_MAXSECTIONS * sizeof(u_int), 128) );

        // Invalidate count to start with. This will also be done at the
        // start of each frame for the buffer the RSX is now writing to.
        g_apuTimerIndexBfrs[u][0] = 0;

        g_apxTimerValueBfrs[u] = static_cast<CellGcmReportData*>( ASURA_PS3_VIDEOMEMORY_ALLOCMAIN(ASURA_PS3_RSXPROFILING_MAXSECTIONS * sizeof(CellGcmReportData) * 2, 128) );
    }

    g_uCurRSXDoubleBuffer = 0;
    g_puTimerIndexDst     = NULL;
    g_pxTimerValueDst     = NULL;

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
 	Asura_CommandConsole::AddVar( L"Debug.Profile.GPUBars", &g_bRenderRSXProfilingBars, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
 	Asura_CommandConsole::AddVar( L"Debug.Profile.GPUText", &g_bRenderRSXProfilingText, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
}


void Asura_PS3_RSXProfiling::Shutdown( void )
{
    for( u_int u = 0; u < ASURA_PS3_RSXPROFILING_BUFFERS; ++u )
    {
        Asura_PS3_VideoMemoryManager::FreeMain( g_apuTimerIndexBfrs[u] );
        Asura_PS3_VideoMemoryManager::FreeMain( g_apxTimerValueBfrs[u] );
    }
}


void Asura_PS3_RSXProfiling::BeginFrame( void )
{
    Asura_Assert( g_eCurrentProfileIndex == ASURA_PS3_RSX_PROFILE__NUMINDICES, "Mis-matched BeginSection() / EndSection() calls?" );

    if( !g_bEnableRSXProfiling ) { return; }

    // Point at entry beyond reserved initial one for entire frame.
    g_puTimerIndexDst = g_apuTimerIndexBfrs[g_uCurRSXDoubleBuffer] + 1;
    g_pxTimerValueDst = g_apxTimerValueBfrs[g_uCurRSXDoubleBuffer] + 2;
    g_uSectionCount   = 1;

    // Zero the section count for this frame so we have some way of detecting
    // whether or not the data is valid later.
    g_apuTimerIndexBfrs[g_uCurRSXDoubleBuffer][0] = 0;

    // Begin profiling section for entire frame.
    WriteTimerCmds( Asura_PS3_Graphics::GetGcmCurrentContext(), ASURA_PS3_RSX_PROFILE__ENTIREFRAME, true );
}


void Asura_PS3_RSXProfiling::EndFrame( void )
{
    Asura_Assert( g_eCurrentProfileIndex == ASURA_PS3_RSX_PROFILE__NUMINDICES, "Mis-matched BeginSection() / EndSection() calls?" );
    Asura_Assert( g_iStackIdx == 0, "Mis-matched BeginSection() / EndSection() calls?" );
    
    g_eCurrentProfileIndex = ASURA_PS3_RSX_PROFILE__NUMINDICES;
    g_iStackIdx = 0;

    if( !g_bEnableRSXProfiling ) { return; }

    CellGcmContextData* pxContext = Asura_PS3_Graphics::GetGcmCurrentContext();

    WriteTimerCmds( pxContext, ASURA_PS3_RSX_PROFILE__ENTIREFRAME, false );

    // Additional final commands to get the section count written back at
    // the beginning.
    cellGcmSetTransferLocationInline( pxContext, CELL_GCM_LOCATION_MAIN );
    cellGcmSetInlineTransferInline( pxContext, Asura_PS3_Helper::AddressToOffset(g_apuTimerIndexBfrs[g_uCurRSXDoubleBuffer]), &g_uSectionCount, 1 );

#ifdef ASURA_PS3_DO_NOT_USE_COMMAND_BUFFER
    cellGcmFlush( pxContext );
#endif

	++g_uCurRSXDoubleBuffer;
    g_uCurRSXDoubleBuffer %= ASURA_PS3_RSXPROFILING_BUFFERS;
    g_puTimerIndexDst = NULL;
    g_pxTimerValueDst = NULL;
}

void Asura_PS3_RSXProfiling::BeginSection( Asura_PS3_RSXProfileIndex eProfileIndex )
{
    Asura_Assert( eProfileIndex < ASURA_PS3_RSX_PROFILE__NUMINDICES, "Invalid profile index" );

    if( !g_bEnableRSXProfiling ) { return; }

	if ( g_eCurrentProfileIndex == ASURA_PS3_RSX_PROFILE__NUMINDICES )
	{
		WriteTimerCmds( Asura_PS3_Graphics::GetGcmCurrentContext(), eProfileIndex, true );
		g_eCurrentProfileIndex = eProfileIndex;
	}

	// Keep track of the stack of profile calls but only actually time the top level
	// in future we might want to profile recursively, though GPAD already supports that so there's not much reason.
	Asura_Assert( g_iStackIdx*sizeof(u_int) < sizeof(g_aiStack), "Too many nested RSX profiles." );
	g_aiStack[g_iStackIdx++] = eProfileIndex;
}


void Asura_PS3_RSXProfiling::EndSection( Asura_PS3_RSXProfileIndex eProfileIndex )
{
    if( !g_bEnableRSXProfiling ) { return; }

	g_iStackIdx--;
	Asura_Assert( g_iStackIdx >= 0, "Mis-matched BeginSection() / EndSection() calls?" );
	Asura_Assertf( g_aiStack[g_iStackIdx] == eProfileIndex, ( "RSX Profiling End type:\"%s\" doesn't match corresponding Begin \"%s\"", aszRSX_PROFILE_NAMES[g_aiStack[g_iStackIdx]], aszRSX_PROFILE_NAMES[eProfileIndex] ) );
	
	if ( g_iStackIdx == 0 )
	{
		WriteTimerCmds( Asura_PS3_Graphics::GetGcmCurrentContext(), eProfileIndex, false );
		g_eCurrentProfileIndex = ASURA_PS3_RSX_PROFILE__NUMINDICES;
	}
}


bool Asura_PS3_RSXProfiling::WriteTimerCmds( CellGcmContextData* pxContext, u_int uProfileIdx, bool bIsBeginSection )
{
    Asura_Assert( g_puTimerIndexDst && g_pxTimerValueDst, "Not in frame?" );

    // Check for overflow.
    if( g_uSectionCount >= ASURA_PS3_RSXPROFILING_MAXSECTIONS )
    {
        Asura_Assert( false, "Too many sections being profiled! Increase ASURA_PS3_RSXPROFILING_MAXSECTIONS or render less." );
        return false;
    }

    // Commands to wait until the RSX is fully finished with whatever it is
    // we're profiling. This is presumably desirable for accuracy if our
    // rendering involves, eg., large quads (like PFX) that may tie up the
    // back end for quite some time after that part of the command buffer was
    // processed.
    if( !bIsBeginSection )
    {
        Asura_PS3_Helper::WriteBlockingWaitForBackEndCmds( pxContext );
    }

    // Commands to tell RSX to write out timestamp.
    u_int uCounterIdx = GetCounterIdx( uProfileIdx, bIsBeginSection );
    cellGcmSetTimeStampInline( pxContext, uCounterIdx );

    // Commands to copy profile index to relevant location. The exception is
    // for the entire frame where we'll write a count out when we're
    // completely finished.
    if( bIsBeginSection && uProfileIdx != ASURA_PS3_RSX_PROFILE__ENTIREFRAME )
    {
        cellGcmSetTransferLocationInline( pxContext, CELL_GCM_LOCATION_MAIN );
        cellGcmSetInlineTransferInline( pxContext, Asura_PS3_Helper::AddressToOffset(g_puTimerIndexDst++), &uProfileIdx, 1 );
    }

    // Commands to tell the RSX to copy the timer value back to the PPU.
    {
        // Special case for destination address for entire frame profile, ie.
        // always at the beginning.
        CellGcmReportData* pxDst;
        if( uProfileIdx == ASURA_PS3_RSX_PROFILE__ENTIREFRAME )
        {
            pxDst = g_apxTimerValueBfrs[g_uCurRSXDoubleBuffer] + (bIsBeginSection ? 0 : 1);
        }
        else
        {
            pxDst = g_pxTimerValueDst++;
        }

        u_int uOffset = Asura_PS3_Helper::AddressToOffset( pxDst );
        cellGcmSetTransferReportDataInline( pxContext, uOffset, uCounterIdx, 1 );
    }

    if( !bIsBeginSection )
    {
        if( uProfileIdx != ASURA_PS3_RSX_PROFILE__ENTIREFRAME )
        {
            ++g_uSectionCount;
        }

#ifdef ASURA_ASSERT
        // Bit of additional verification.
        u_int uIndexCount = g_puTimerIndexDst - g_apuTimerIndexBfrs[g_uCurRSXDoubleBuffer];
        u_int uValueCount = g_pxTimerValueDst - g_apxTimerValueBfrs[g_uCurRSXDoubleBuffer];

        Asura_Assert( uIndexCount * 2 == uValueCount, "Mis-matched BeginSection() / EndSection() calls?" );
#endif
    }

    return true;
}

DEBUGVAR u_int dbg_uRSXProfTextX = 60;
#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
DEBUGVAR u_int dbg_uRSXProfTextY = 15;
#else // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
DEBUGVAR u_int dbg_uRSXProfTextY = 2;
#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )

void Asura_PS3_RSXProfiling::RenderBars( void )
{
    if( !g_bEnableRSXProfiling ) { return; }

	const u_int uPreviousBuffer = ( g_uCurRSXDoubleBuffer + 1 )  % ASURA_PS3_RSXPROFILING_BUFFERS;
    const u_int*             const puTimerIndices = g_apuTimerIndexBfrs[ uPreviousBuffer ];
    const CellGcmReportData* const pxTimerData    = g_apxTimerValueBfrs[ uPreviousBuffer ];

    const u_int uSectionCount = puTimerIndices[0];
    if( !uSectionCount )
    {
        // Could render the background divisions and nothing else here maybe
        // but just do nothing for now. This should only happen in the first
        // frame in general.
        return;
    }

    float fTotalFrameDuration;
    float afDurations[ASURA_PS3_RSX_PROFILE__NUMINDICES];
    u_int auHitCounts[ASURA_PS3_RSX_PROFILE__NUMINDICES];
	float fRunningTotalTime = 0.0f;

    // Parse the timing data into totals per profile index. Note that the
    // PI_ENTIREFRAME entry will be 0 and therefore ignored everywhere. It's
    // convenient to special-case it.
    {
        fTotalFrameDuration = static_cast<float>(pxTimerData[1].timer - pxTimerData[0].timer) / 1000000000.0f;

        for( u_int u = 0; u < ASURA_PS3_RSX_PROFILE__NUMINDICES; ++u )
        {
            afDurations[u] = 0.0f;
            auHitCounts[u] = 0;
        }

        for( u_int u = 1; u < uSectionCount; ++u )
        {
            u_int uIdx = puTimerIndices[u];
            Asura_Assert( uIdx != ASURA_PS3_RSX_PROFILE__ENTIREFRAME && uIdx < ASURA_PS3_RSX_PROFILE__NUMINDICES, "Unexpected profile index" );

            afDurations[uIdx] += static_cast<float>(pxTimerData[(u << 1) + 1].timer - pxTimerData[u << 1].timer) / 1000000000.0f;

            ++auHitCounts[uIdx];
        }
    }

    for( u_int u = 0; u < ASURA_PS3_RSX_PROFILE__NUMINDICES; ++u )
    {
        fRunningTotalTime += afDurations[u];
    }

    if( g_bRenderRSXProfilingBars )
    {
        static Asura_Profiling_BarRenderer_Data xData;

        xData.SetUseDefaultBackgroundColours();
        xData.SetBarColours( auRSX_PROFILE_BAR_ARGBS, sizeof(auRSX_PROFILE_BAR_ARGBS) / sizeof(auRSX_PROFILE_BAR_ARGBS[0]), uRSX_PROFILE_UNPROFILED_ARGB );
        xData.SetYOffset( Asura_CPUProfile_Renderer::GetYOffset() + 0.018f );

        // Pass the array unmodified - the first (ENTIREFRAME) entry will be
        // 0 and that bar colour will be unused unless we wrap. See the 
        // declaration of auRSX_PROFILE_BAR_ARGBS.
        xData.SetNewFrameData( fTotalFrameDuration, afDurations, ASURA_PS3_RSX_PROFILE__NUMINDICES );

        Asura_Profiling_BarRenderer::Render( xData );
    }
#ifdef ASURA_USE_DEBUG_TEXT
	if (g_bRenderRSXProfilingText)
	{
		u_int uNumRendered = 0;
		for (int i = 1; i<ASURA_PS3_RSX_PROFILE__NUMINDICES; i++)
		{
			if (afDurations[i] > 0.0f)
			{
				const char* szName = aszRSX_PROFILE_NAMES[i] ? aszRSX_PROFILE_NAMES[i] : "???";
				Asura_DebugText::AddTextBlock( dbg_uRSXProfTextX, dbg_uRSXProfTextY + uNumRendered, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
					"%2d %5.02f%% %s: (%d us)", i, Asura_Maths::FtoD(100.0f * afDurations[i] / fTotalFrameDuration), szName, Asura_Maths::FloatToInt(afDurations[i] * 1000000.0f));

				// render a colour coded square
				float fX = (static_cast<float>(dbg_uRSXProfTextX - 1) * ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
				float fY = 1.0f + (static_cast<float>(dbg_uRSXProfTextY + uNumRendered) * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
				float fW = ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH - 2.0f;
				float fH = ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT - 2.0f;
				Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
				Asura_Render::Overlay.Colour = auRSX_PROFILE_BAR_ARGBS[i];
				Asura_Render::Overlay.Render();

				++uNumRendered;
			}
		}

		{
			// render a colour coded square for the untimed section
			float fX = (static_cast<float>(dbg_uRSXProfTextX - 1) * ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
			float fY = 1.0f + (static_cast<float>(dbg_uRSXProfTextY + uNumRendered) * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
			float fW = ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH - 2.0f;
			float fH = ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT - 2.0f;
			Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
			Asura_Render::Overlay.Colour = uRSX_PROFILE_UNPROFILED_ARGB;
			Asura_Render::Overlay.Render();
			// and the text...
			const float fTime = fTotalFrameDuration - fRunningTotalTime;
			Asura_DebugText::AddTextBlock( dbg_uRSXProfTextX, dbg_uRSXProfTextY + uNumRendered, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
				"   %5.02f%% unprofiled: (%d us)", Asura_Maths::FtoD(100.0f * fTime / fTotalFrameDuration), Asura_Maths::FloatToInt(fTime * 1000000.0f));
		}
	}
#endif // #ifdef ASURA_USE_DEBUG_TEXT

    if( g_bOutputRSXProfilingFrameDump )
    {
        printf( "\n--- RSX profiling data ---\n\nframe number: %u\n", Asura_Timers::GetFrameNumber() );

        float fUnprofiled = fTotalFrameDuration;

        printf( "entire frame: %0.05f (%4.01f%% of 30Hz frame)\n", fTotalFrameDuration, (fTotalFrameDuration / (1.0f / 30.0f)) * 100.0f );

        printf( "\nframe breakdown:\n" );
        for( u_int u = 1; u < ASURA_PS3_RSX_PROFILE__NUMINDICES; ++u )
        {
            float fThisDuration = afDurations[u];
            if( fThisDuration > 0.0f )
            {
                printf( " %2u %14s: %0.05f (%5.02f%%, %3u hits)\n", u, aszRSX_PROFILE_NAMES[u], fThisDuration, (fThisDuration / fTotalFrameDuration) * 100.0f, auHitCounts[u] );
                fUnprofiled -= fThisDuration;
            }
        }

        fUnprofiled = Asura_Maths::Max( fUnprofiled, 0.0f );
        printf( " -      unprofiled:  %0.05f (%5.02f%%)\n\n", fUnprofiled, (fUnprofiled / fTotalFrameDuration) * 100.0f );

        g_bOutputRSXProfilingFrameDump = false;
    }
}

#endif // #ifdef ASURA_PS3_USE_GPU_PROFILE_BARS
