//
// Asura_PS3_TextureManagement.cpp
//
// Created by KJFL - 27 June 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS3_TextureManagement.h"

// Engine independent.
#include "Core/Asura_TextureManagement.h"
#if defined (ASURA_USE_TEXTURE_STREAMING)
#include "Core/Asura_Texture_Streaming.h"
#endif
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_MemAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"

// Engine platform-specific.
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_VideoMemoryManager.h"
#include "cell/gcm/gcm_method_data.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

#if !defined(ASURA_NO_DEBUG_OUTPUT)
//#define ASURA_PS3_LOG_TEXTURE_STREAMING	1
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)

static const u_int uPINK_TEXEL            = 0xFF8080FFu; // To match PS2's pink texture.
static const u_int uWHITE_TEXEL           = 0xFFFFFFFFu;
static const u_int uBLACK_TEXEL           = 0x00000000u;
static const u_int uFLAT_NORMAL_MAP_TEXEL = 0x80808080u; // Flat tangent space normal map texel

// D E B U G V A R S /////////////////////////////////////////////////////////

// G L O B A L S /////////////////////////////////////////////////////////////

static Asura_PS3_Texture	g_axTextureStorage[ASURA_MAX_NO_OF_TEXTURES];
#if defined (ASURA_USE_TEXTURE_STREAMING)
static Asura_PS3_Texture	g_axStreamingTextureStorage[ASURA_MAX_NO_OF_TEXTURES];
#endif
static Asura_PS3_Texture	g_axSpecialTextures[Asura_PS3_TextureManagement::SPECIALTEXTURE_NUM_IDS];
static bool					g_bLoadToLocalMemory	= true;
static bool					g_bLoadFromFile			= false;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static u_int CalculateTexelDataSize( ASURA_PS3_TEXTUREFORMAT eFormat, u_int uWidth, u_int uHeight );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Texture_Management.
//

void Asura_Texture_Management::Platform_LoadTexture( const Asura_String& xName, int iTexID )
{
    if( Asura_Texture_Management::IsTextureLoaded(iTexID) )
    {
        // We expect this to happen whenever LoadTextures() gets called so
        // we can't assert.
        return;
    }

#if _DEBUG
    g_axTextureStorage[iTexID].m_szName = Asura_StringUtil::Duplicate( xName.GetString() );
#endif

    char szPath[256];

    // Construct full file path.
    {
        const char* szInputName = xName.GetString();

        const char* szExt = strrchr( szInputName, '.' );
        if( !szExt )
        {
            // Just append the correct extension to whatever's there.
            szExt = szInputName + xName.GetLength();
        }

        Asura_PrintfToBfr( szPath, "%s%s%.*s.ps3tx", 256, true,
                           "graphics", ('\\' == *szInputName ? "" : "\\"), szExt - szInputName, szInputName );
    }

    Asura_Handle iFileHandle = Asura_FileAccess::Open( szPath );

    if( ASURA_INVALID_HANDLE == iFileHandle )
    {
        g_axTextureStorage[iTexID].SetupDummyTexture( uPINK_TEXEL );
    }
    else
    {
        g_axTextureStorage[iTexID].InitFromFile( iFileHandle );
        Asura_FileAccess::Close( iFileHandle );
    }

	Loaded[iTexID] = s_uCurResetGroup;
}

void Asura_Texture_Management::Platform_LoadTextureFromResource( Asura_Handle iFileHandle, int /*iFileSize*/, int iTexID, u_int /*uChunkFlags*/ )
{
#if _DEBUG
	const Asura_String* pxName = Asura_Texture_Management::GetTextureName( iTexID );
	if( pxName )
	{
		g_axTextureStorage[iTexID].m_szName = Asura_StringUtil::Duplicate( pxName->GetString() );
	}
#endif

	g_axTextureStorage[iTexID].InitFromFile( iFileHandle );
	// (MP - 22.03.2006) Little point in checking return value currently
	// since independent code doesn't cope with failure anyway. Besides,
	// we assume at runtime that assets we're provided with are valid.
}

// (MP - 21.03.2006) This should never even be getting called. Might as well
// omit the definition to get link errors if it ever is.
//void Asura_Texture_Management::ReloadTexture( int /*index*/ )
//{
//}

void Asura_Texture_Management::Platform_DestroyTexture( int iTexID )
{
	g_axTextureStorage[iTexID].Destroy();
}

void Asura_Texture_Management::Platform_Initialise( void )
{
    Asura_PS3_TextureManagement::Initialise();
}

void Asura_Texture_Management::Platform_Reset( void )
{
    Asura_PS3_TextureManagement::Reset();
}

void Asura_Texture_Management::Platform_Shutdown( void )
{
    Asura_PS3_TextureManagement::Shutdown();
}

//
// Asura_PS3_TextureManagement
//
bool Asura_PS3_TextureManagement::s_bTextureBudgetEnabled = false;
int	 Asura_PS3_TextureManagement::s_iTextureBudget = 0;
int	 Asura_PS3_TextureManagement::s_iTextureBudgetExceededBy = 0;

void Asura_PS3_TextureManagement::Initialise( void )
{
	g_axSpecialTextures[SPECIALTEXTURE_ID_PLACEHOLDER   & ~SPECIALTEXTURE_FLAG].SetupDummyTexture( uPINK_TEXEL            );
	g_axSpecialTextures[SPECIALTEXTURE_ID_NOTEXTURE     & ~SPECIALTEXTURE_FLAG].SetupDummyTexture( uWHITE_TEXEL           );
	g_axSpecialTextures[SPECIALTEXTURE_ID_FLATNORMALMAP & ~SPECIALTEXTURE_FLAG].SetupDummyTexture( uFLAT_NORMAL_MAP_TEXEL );
	g_axSpecialTextures[SPECIALTEXTURE_ID_BLACKTEXTURE  & ~SPECIALTEXTURE_FLAG].SetupDummyTexture( uBLACK_TEXEL           );
}

void Asura_PS3_TextureManagement::Reset( void )
{
	g_bLoadToLocalMemory = true;
}

void Asura_PS3_TextureManagement::Shutdown( void )
{
	int i;

	for( i = 0; i < Asura_PS3_TextureManagement::SPECIALTEXTURE_NUM_IDS; ++i )
	{
		g_axSpecialTextures->Destroy();
	}

	for( i = 0; i < ASURA_MAX_NO_OF_TEXTURES; ++i )
	{
		g_axTextureStorage->Destroy();
	}
}

void Asura_PS3_Texture::SetAsTextureIndex( int iIndex, bool bNormalizedTextureCoordinates )
{
#if _DEBUG
	Asura_Assertf( m_xTexture.width != 0, ( "Texture's width is 0. Texture \"%s\" (index %d).\n", m_szName, iIndex ) );
	Asura_Assertf( m_xTexture.height != 0, ( "Texture's height is 0. Texture \"%s\" (index %d).\n", m_szName, iIndex ) );
#else
	Asura_Assert( m_xTexture.width != 0, "Texture's width is 0.\n" );
	Asura_Assert( m_xTexture.height != 0, "Texture's height is 0.\n" );
#endif
	CellGcmTexture* pxTexture = &m_xTexture; 
	if ( m_xTexture.width == 0 || m_xTexture.height == 0 )
	{
		Asura_PS3_Texture* pxPlatformTexture = Asura_PS3_TextureManagement::GetTexture(Asura_PS3_TextureManagement::SPECIALTEXTURE_ID_PLACEHOLDER);	 // avoid a crash
		pxTexture = pxPlatformTexture->GetTexture();
	}
	

	if( bNormalizedTextureCoordinates )
	{
		uint8_t ucFormat = ~CELL_GCM_TEXTURE_UN;
		m_xTexture.format &= ucFormat;
	}
	else
	{
		m_xTexture.format |= CELL_GCM_TEXTURE_UN;
	}
	m_uiTextureFormat = CELL_GCM_METHOD_DATA_TEXTURE_FORMAT( m_xTexture.location, m_xTexture.cubemap, m_xTexture.dimension, m_xTexture.format, m_xTexture.mipmap );
	Asura_PS3_RenderingStates::SetTexture( iIndex, pxTexture, bNormalizedTextureCoordinates );
}

void Asura_PS3_Texture::SetUseNormalizedCoordinates( bool bNormalizedTextureCoordinates )
{
	if( bNormalizedTextureCoordinates )
	{
		uint8_t ucFormat = ~CELL_GCM_TEXTURE_UN;
		m_xTexture.format &= ucFormat;
	}
	else
	{
		m_xTexture.format |= CELL_GCM_TEXTURE_UN;
	}
	m_uiTextureFormat = CELL_GCM_METHOD_DATA_TEXTURE_FORMAT( m_xTexture.location, m_xTexture.cubemap, m_xTexture.dimension, m_xTexture.format, m_xTexture.mipmap );
}

Asura_PS3_Texture* Asura_PS3_TextureManagement::GetTexture( int iTexID )
{
	if( iTexID == ASURA_TEXTURE_NONE )
	{
        // This texture with a single white texel simulates untextured
        // rendering. To render without textures using a fragment shader you
        // have to set the colour accordingly. Just setting a white texture
        // means we don't need to modify all fragment shaders to cope with
        // ASURA_TEXTURE_NONE.
        iTexID = SPECIALTEXTURE_ID_NOTEXTURE;
	}

    // Handle specific requests for any of our special case textures.
    if( iTexID & SPECIALTEXTURE_FLAG )
    {
        iTexID &= ~SPECIALTEXTURE_FLAG;
        Asura_Assert( 0 <= iTexID && iTexID < SPECIALTEXTURE_NUM_IDS, "" );
        return &g_axSpecialTextures[iTexID];
    }

    if( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES )
    {
#ifdef ASURA_USE_TEXTURE_STREAMING
		Asura_PS3_Texture* pxStreamingTexture = &g_axStreamingTextureStorage[iTexID];

		if ( pxStreamingTexture && pxStreamingTexture->GetTextureData() && Asura_Texture_Streaming_System::IsTextureLoaded( iTexID ))
		{
			return pxStreamingTexture;
		}
#endif

        Asura_PS3_Texture* pxRet = &g_axTextureStorage[iTexID];
		Asura_Assert(pxRet, "no texture");
		if (pxRet)
		{
			return pxRet;
		}
    }
    else
    {
        // Invalid texture ID so fall through to use pink texture. However,
        // we should assert in this case since we should never end up
        // requesting an entirely invalid texture ID without some error
        // occurring elsewhere (eg. bad conversion on load,
        // ASURA_MAX_NO_OF_TEXTURES not being high enough, etc.)
        Asura_Assert( false, "Invalid texture ID requested" );
    }

	return &g_axSpecialTextures[SPECIALTEXTURE_ID_PLACEHOLDER & ~SPECIALTEXTURE_FLAG];
}

//
// Asura_PS3_Texture
//

Asura_PS3_Texture::Asura_PS3_Texture( u_char ucLocation )
{
	m_pxTextureData = NULL;
	m_bAllocatedOwnTextureData = false;
	m_xTexture._padding = 0;
	m_xTexture.cubemap = CELL_GCM_FALSE;
	m_xTexture.depth = 1;
	m_xTexture.dimension = CELL_GCM_TEXTURE_DIMENSION_2;
	m_xTexture.format = CELL_GCM_TEXTURE_A8R8G8B8;
	m_xTexture.height = 1;
	m_xTexture.location = ucLocation;
	m_xTexture.mipmap = 1;
	m_xTexture.offset = 0;
	m_xTexture.pitch = 4;
	// ARGB -> ARGB remap
	m_xTexture.remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_A,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );

#if _DEBUG
    m_szName = NULL;
#endif
	InitGCMMacroData();
}

Asura_PS3_Texture::Asura_PS3_Texture( CellGcmTexture* pxTexture )
{
	Asura_Mem::Copy( &m_xTexture, pxTexture, sizeof( CellGcmTexture ) );
	m_pxTextureData = NULL;
	m_bAllocatedOwnTextureData = false;
	InitGCMMacroData();

#if _DEBUG
    m_szName = NULL;
#endif
}

Asura_PS3_Texture::~Asura_PS3_Texture( void )
{
    Destroy();
}

void Asura_PS3_Texture::SetTextureData( void* pxTextureData )
{ 
	m_pxTextureData = pxTextureData; 
}

void Asura_PS3_Texture::CreateTexture( u_short usWidth, u_short usHeight, ASURA_PS3_TEXTUREFORMAT uiFormat, bool bAllocateMemory, u_short usFalgs )
{
	Asura_Assert( !( m_bAllocatedOwnTextureData && m_pxTextureData ), "already allocated texture data" );
	Destroy();

	m_xTexHdr.m_usWidth           = usWidth;
    m_xTexHdr.m_usHeight          = usHeight;
    m_xTexHdr.m_usFlags           = usFalgs;
    m_xTexHdr.m_ucTextureFormatID = uiFormat;
    m_xTexHdr.m_ucNumberOfMipmaps = 0;

	if( usHeight == 1 )
	{
		m_xTexture.dimension = CELL_GCM_TEXTURE_DIMENSION_1;
	}

	if( bAllocateMemory )
	{
		CreateTextureObject();
	}
}

// return true if the full texture is loaded in memory
// return false in case only a part of the texture is loaded in memory
bool Asura_PS3_Texture::InitFromFile( Asura_Handle iFileHandle, Asura_PS3_FileSystem::Handle iCacheHandle, fnAllocateTexture* pfnAllocator )
{
	Asura_Assert( !( m_bAllocatedOwnTextureData && m_pxTextureData ), "already allocated texture data" );
	Destroy();

   // Read header.
	if( iFileHandle != ASURA_INVALID_HANDLE )
	{
		Asura_FileAccess::ReadData( iFileHandle, &m_xTexHdr, sizeof(Asura_PS3_TextureHeader) );
	}
	else
	{
		Asura_PS3_FileSystem::ReadData( iCacheHandle, sizeof( Asura_PS3_TextureHeader ), &m_xTexHdr );
	}

    int iTextureSize;

	g_bLoadFromFile = true;
    iTextureSize = CreateTextureObject(pfnAllocator);
	g_bLoadFromFile = false;
	if( iTextureSize >= 0 )
	{
		if( iFileHandle != ASURA_INVALID_HANDLE )
		{
			Asura_FileAccess::ReadData( iFileHandle, m_pxTextureData, iTextureSize );
			// Write file data to the cache
			if( iCacheHandle != Asura_PS3_FileSystem::INVALID_HANDLE )
			{
				Asura_PS3_FileSystem::WriteData( iCacheHandle, sizeof( Asura_PS3_TextureHeader ), &m_xTexHdr );
				Asura_PS3_FileSystem::WriteData( iCacheHandle, iTextureSize, m_pxTextureData );
			}
		}
		else
		{
			Asura_PS3_FileSystem::ReadData( iCacheHandle, iTextureSize, m_pxTextureData );
		}
	}
	else
	{
		if( iFileHandle != ASURA_INVALID_HANDLE )
		{
			Asura_FileAccess::SkipData( iFileHandle, -iTextureSize );
		}
		return false;
	}

	return true;
}

void Asura_PS3_Texture::SetupDummyTexture( u_int uRGBA, fnAllocateTexture* pfnAllocator )
{
	Asura_Assert( !( m_bAllocatedOwnTextureData && m_pxTextureData ), "already allocated texture data" );
	Destroy();

    // Fake the texture header.
    m_xTexHdr.m_usWidth           = 1;
    m_xTexHdr.m_usHeight          = 1;
    m_xTexHdr.m_usFlags           = 0;
    m_xTexHdr.m_ucTextureFormatID = ASURA_PS3_TEXTUREFORMAT_8888_RGBA;
    m_xTexHdr.m_ucNumberOfMipmaps = 0;

    CreateTextureObject(pfnAllocator);

	((u_int*)m_pxTextureData)[0] = uRGBA;
}

int Asura_PS3_Texture::CalculatePitch( const int iWidth, const int iBlockSize, const int iBytesPerPixel )
{
	int iPitch = ( iWidth + 3 ) / 4;        // calc from PS3 SDK, no need to use floor() because this is an int
	iPitch *= iBlockSize;
	iPitch *= iBytesPerPixel;
	 
	//round up to the next multiple of 64
	return ( iPitch + 63 )&( ~63 );
}

int Asura_PS3_Texture::CreateTextureObject( fnAllocateTexture* pfnAllocator )
{
    const ASURA_PS3_TEXTUREFORMAT eFormat = static_cast<ASURA_PS3_TEXTUREFORMAT>( m_xTexHdr.m_ucTextureFormatID );
	int iPitch, iSize = 0;
	u_int uCurrentHeight, uCurrentWidth;

	m_xTexture.width = uCurrentWidth = m_xTexHdr.m_usWidth;
	m_xTexture.height = uCurrentHeight = m_xTexHdr.m_usHeight;
	m_xTexture.mipmap = m_xTexHdr.m_ucNumberOfMipmaps + 1;

	if( m_xTexture.width == 0 )
	{
		Asura_Assert( false, "Wrong texture exported (0 width)" );
	}

	iPitch = m_xTexture.width;

	switch( eFormat )
	{
	case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_A8R8G8B8;
			iPitch = CalculatePitch( iPitch, 4, 4 );
			m_xTexture.format |= CELL_GCM_TEXTURE_LN;
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_COMPRESSED_DXT45;
			iPitch = 0;
			m_xTexture.format |= CELL_GCM_TEXTURE_SZ;
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_COMPRESSED_DXT1;
			iPitch = 0;
			m_xTexture.format |= CELL_GCM_TEXTURE_SZ;
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_ABGR16F:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_W16_Z16_Y16_X16_FLOAT;
			iPitch = CalculatePitch( iPitch, 4, 8 );
			if( m_xTexHdr.m_usFlags & ASURA_PS3_TEXTURE_SWIZZLE )
			{
				m_xTexture.format |= CELL_GCM_TEXTURE_SZ;
			}
			else
			{
				m_xTexture.format |= CELL_GCM_TEXTURE_LN;
			}
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_YX16F:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_Y16_X16_FLOAT;
			iPitch = CalculatePitch( iPitch, 4, 4 );
			m_xTexture.format |= CELL_GCM_TEXTURE_LN;
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_YX16:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_Y16_X16;
			iPitch = CalculatePitch( iPitch, 4, 4 );
			m_xTexture.format |= CELL_GCM_TEXTURE_LN;
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_X32F:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_X32_FLOAT;
			iPitch = CalculatePitch( iPitch, 4, 4 );
			m_xTexture.format |= CELL_GCM_TEXTURE_LN;
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_DEPTH24_D8:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_DEPTH24_D8;
			iPitch = CalculatePitch( iPitch, 4, 4 );
			m_xTexture.format |= CELL_GCM_TEXTURE_LN;
		}
	case ASURA_PS3_TEXTUREFORMAT_DXT1_A8: //
		{
			m_xTexture.format = CELL_GCM_TEXTURE_COMPRESSED_DXT1;
			iPitch = 0;
			m_xTexture.format |= CELL_GCM_TEXTURE_SZ;
			// ARGB -> ARGB remap
			m_xTexture.remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_G,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );
			break;
		}
	case ASURA_PS3_TEXTUREFORMAT_L8:
		{
			m_xTexture.format = CELL_GCM_TEXTURE_B8;
			iPitch *= 1;
			if( m_xTexHdr.m_usFlags & ASURA_PS3_TEXTURE_SWIZZLE )
			{
				m_xTexture.format |= CELL_GCM_TEXTURE_SZ;
			}
			else
			{
				m_xTexture.format |= CELL_GCM_TEXTURE_LN;
			}
			m_xTexture.remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_G,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );
			break;
		}
	default:
		{
			Asura_Assert( false, "Invalid texture format" );
			return 0;
		}
	}

	m_xTexture.format |= CELL_GCM_TEXTURE_NR;
	m_xTexture.pitch = iPitch;

	for(int i = 0; i < m_xTexture.mipmap; ++i)
	{
		iSize += CalculateTexelDataSize( eFormat, uCurrentWidth, uCurrentHeight );
		uCurrentWidth >>= 1;
		uCurrentHeight >>= 1;
	}

	m_iTextureFaceSize = iSize;
	if( m_xTexHdr.m_usFlags & ASURA_PS3_TEXTURE_CUBEMAP )
	{
		m_xTexture.cubemap = CELL_GCM_TRUE;
		int iPad = 128 - ( iSize % 128 );
		if( iPad != 128 )
		{
			iSize += iPad;
		}
		m_iTextureFaceSize = iSize;
		iSize *= 6;
	}

	if (Asura_PS3_TextureManagement::TextureBudgetEnabled())
	{
		Asura_PS3_TextureManagement::DecrementTextureBudget(iSize);
	}

	int iTempSize = iSize;
	if(Asura_PS3_TextureManagement::TextureBudgetEnabled() && Asura_PS3_TextureManagement::GetTextureBudget() < 0 )
	{	
		iPitch					/= m_xTexture.width;
		m_xTexture.pitch		 = iPitch;
		m_xTexHdr.m_usWidth		 = m_xTexture.width		= uCurrentWidth		= 1;
		m_xTexHdr.m_usHeight	 = m_xTexture.height	= uCurrentHeight	= 1;
		m_xTexture.mipmap		 = 1;
		iSize					 = CalculateTexelDataSize( eFormat, uCurrentWidth, uCurrentHeight );
		if( m_xTexHdr.m_usFlags & ASURA_PS3_TEXTURE_CUBEMAP )
		{
			int iPad = 128 - ( iSize % 128 );
			if( iPad != 128 )
			{
				iSize += iPad;
			}
			iSize *= 6;
		}
	}
	{
		//4 bytes alignment of the chunks
		int iPad = 4 - ( iSize % 4 );
		if( iPad != 4 )
			iSize += iPad; 
	}

	if(pfnAllocator == NULL)
	{
		m_bAllocatedOwnTextureData = true;

#ifdef ASURA_PS3_USE_BALANCED_MEMORY_USAGE
		if( !Asura_PS3_VideoMemoryManager::OverrideLocation( m_xTexture.location ) )
		{
			//BW - rewritten because certain combinations of circumstances could make it try to alloc vram when none was left

			//BW the old logic was as follows:
			// g_bLoadToLocalMemory - flip this with each load to try to evenly divide them between main & local
			//						this is not ideal, but for the time being we can't be sure how best to allocate them.
			// g_bLoadFromFile - only use g_bLoadToLocalMemory if this is true. otherwise just alloc from whichever pool is bigger
			// if any alloc fails, try allocing from the other memory (there was a bug in 1 case where this would not happen, which caused a crash)

			// to simplify this, ignore g_bLoadFromFile, always use g_bLoadToLocalMemory, and check size against free space before alloc.
			// assume main will have more space than local (which is true now that we can access all of main.

			m_pxTextureData = NULL;

			if ( g_bLoadToLocalMemory &&
				Asura_PS3_VideoMemoryManager::GetFreeLocal() > (u_int)iSize+128 )
			{
				m_xTexture.location	= CELL_GCM_LOCATION_LOCAL;
				m_pxTextureData		= ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( iSize, 128 );

				//BW 17/09/08 - bias towards vram a bit more. alternating between main & local was running us out of memory. So only alternate when we've got less vram than main ram.
				if ( Asura_PS3_VideoMemoryManager::GetFreeLocal() < Asura_PS3_VideoMemoryManager::GetFreeMain() )
				{
					g_bLoadToLocalMemory = false; // put the next texture main memory
				}
			}

			if( m_pxTextureData == NULL ) // this will catch the previous alloc failing
			{
				m_xTexture.location	= CELL_GCM_LOCATION_MAIN;
				m_pxTextureData	= ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( iSize, 128 );
				g_bLoadToLocalMemory = true; // try to put the next texture in local memory
			}
		}
#else // ASURA_PS3_USE_BALANCED_MEMORY_USAGE
		Asura_PS3_VideoMemoryManager::OverrideLocation( m_xTexture.location );
		if( m_xTexture.location == CELL_GCM_LOCATION_LOCAL )
		{
			m_pxTextureData = ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( iSize, 128 );
			if( !m_pxTextureData )
			{
				m_xTexture.location = CELL_GCM_LOCATION_MAIN;
				m_pxTextureData = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( iSize, 128);
			}
		}
		else
		{
			m_pxTextureData = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( iSize, 128);
			if( !m_pxTextureData )
			{
				m_xTexture.location = CELL_GCM_LOCATION_LOCAL;
				m_pxTextureData = ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( iSize, 128 );
			}
		}
#endif
	}
	else
	{
		m_pxTextureData	= pfnAllocator(iSize, 128, m_xTexture);
	}

	if ( m_pxTextureData == NULL )
	{
		Asura_Assert( 0, "Failed to allocate texture data\n" );
		return 0;
	}

	cellGcmAddressToOffset( m_pxTextureData, &m_xTexture.offset );

	InitGCMMacroData();

	if(Asura_PS3_TextureManagement::TextureBudgetEnabled() && Asura_PS3_TextureManagement::GetTextureBudget() < 0 )
	{
		char* pxTextureData = static_cast<char*>( m_pxTextureData );
		for( int i = 0; i < iSize; ++i )
		{
			pxTextureData[ i ] = 255;
		}
		return -iTempSize;
	}
	return iSize;
}

void Asura_PS3_Texture::Destroy( fnDeallocateTexture* pfnDeallocator )
{
	if(pfnDeallocator == NULL)
	{
		if( m_pxTextureData && m_bAllocatedOwnTextureData )
		{
			if( m_xTexture.location == CELL_GCM_LOCATION_LOCAL )
			{
				Asura_PS3_VideoMemoryManager::FreeLocal( m_pxTextureData );
			}
			else
			{
				Asura_PS3_VideoMemoryManager::FreeMain( m_pxTextureData );
			}
		}
	}
	else
	{
		pfnDeallocator(m_pxTextureData);
	}
	
	//reset to defaults, in case it gets used again
	m_pxTextureData = NULL;
	m_bAllocatedOwnTextureData = false;
	m_xTexture._padding = 0;
	m_xTexture.cubemap = CELL_GCM_FALSE;
	m_xTexture.depth = 1;
	m_xTexture.dimension = CELL_GCM_TEXTURE_DIMENSION_2;
	m_xTexture.format = CELL_GCM_TEXTURE_A8R8G8B8;
	m_xTexture.height = 1;
	m_xTexture.location = CELL_GCM_LOCATION_LOCAL;
	m_xTexture.mipmap = 1;
	m_xTexture.offset = 0;
	m_xTexture.pitch = 4;
	// ARGB -> ARGB remap
	m_xTexture.remap = CELL_GCM_REMAP_MODE( CELL_GCM_TEXTURE_REMAP_ORDER_XYXY, CELL_GCM_TEXTURE_REMAP_FROM_A,CELL_GCM_TEXTURE_REMAP_FROM_R, CELL_GCM_TEXTURE_REMAP_FROM_G, CELL_GCM_TEXTURE_REMAP_FROM_B, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP, CELL_GCM_TEXTURE_REMAP_REMAP );
#if _DEBUG
	delete[] m_szName;
	m_szName = NULL;
#endif

}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static u_int CalculateTexelDataSize( ASURA_PS3_TEXTUREFORMAT eFormat, u_int uWidth, u_int uHeight )
{
    switch( eFormat )
    {
        case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
		case ASURA_PS3_TEXTUREFORMAT_YX16F:
		case ASURA_PS3_TEXTUREFORMAT_YX16:
		case ASURA_PS3_TEXTUREFORMAT_X32F:
		case ASURA_PS3_TEXTUREFORMAT_DEPTH24_D8:
            return uWidth * uHeight * 4;

        case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
			{
				u_int uW = uWidth, uH = uHeight;

				if( uW < 4 ) uW = 4;
				if( uH < 4 ) uH = 4;
				return uW * uH;
			}

		case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
			{
				u_int uW = uWidth, uH = uHeight;

				if( uW < 4 ) uW = 4;
				if( uH < 4 ) uH = 4;
				return ( ( uW * uH ) >> 1 );
			}

		case ASURA_PS3_TEXTUREFORMAT_DXT1_A8:
			{
				u_int uW = uWidth, uH = uHeight;

				if( uW < 4 ) uW = 4;
				if( uH < 4 ) uH = 4;
				return ( ( uW * uH ) >> 1 );
			}
		case ASURA_PS3_TEXTUREFORMAT_L8:
			return uWidth * uHeight;

		case ASURA_PS3_TEXTUREFORMAT_ABGR16F:
			return uWidth * uHeight * 8;

        default:
            Asura_Assertf( false, ("Unsupported texture format ID: %i", eFormat) );
            return 0;
    }
}

void Asura_PS3_Texture::InitGCMMacroData( void )
{
	m_uiTextureOffset		= m_xTexture.offset;
	m_uiTextureControl1		= m_xTexture.remap;
	m_uiTextureControl3		= CELL_GCM_METHOD_DATA_TEXTURE_CONTROL3( m_xTexture.pitch, m_xTexture.depth );
	m_uiTextureFormat		= CELL_GCM_METHOD_DATA_TEXTURE_FORMAT( m_xTexture.location, m_xTexture.cubemap, m_xTexture.dimension, m_xTexture.format, m_xTexture.mipmap );
	m_uiTextureImageRect	= CELL_GCM_METHOD_DATA_TEXTURE_IMAGE_RECT( m_xTexture.height, m_xTexture.width );
}

bool Asura_PS3_Texture::Is8BitTexture( void )
{
	uint8_t uBaseFormat = m_xTexture.format & ~0x60; //remove SZ/LN and NR/UN flags

	return (uBaseFormat == CELL_GCM_TEXTURE_A8R8G8B8
		|| uBaseFormat == CELL_GCM_TEXTURE_B8
		|| uBaseFormat == CELL_GCM_TEXTURE_COMPRESSED_DXT1
		|| uBaseFormat == CELL_GCM_TEXTURE_COMPRESSED_DXT23
		|| uBaseFormat == CELL_GCM_TEXTURE_COMPRESSED_DXT45
		|| uBaseFormat == CELL_GCM_TEXTURE_G8B8
		|| uBaseFormat == CELL_GCM_TEXTURE_D8R8G8B8
		|| uBaseFormat == (CELL_GCM_TEXTURE_COMPRESSED_B8R8_G8R8 & ~0x60) //apparently these formats always have the LN flag set
		|| uBaseFormat == (CELL_GCM_TEXTURE_COMPRESSED_R8B8_R8G8 & ~0x60)

//less than 8-bit - maybe want to include these?
// 		|| uBaseFormat == CELL_GCM_TEXTURE_A1R5G5B5				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_A4R4G4B4				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_R5G6B5					
// 		|| uBaseFormat == CELL_GCM_TEXTURE_R6G5B5					

//more than 8-bit
// 		|| uBaseFormat == CELL_GCM_TEXTURE_DEPTH24_D8				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_DEPTH24_D8_FLOAT		
// 		|| uBaseFormat == CELL_GCM_TEXTURE_DEPTH16				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_DEPTH16_FLOAT			
// 		|| uBaseFormat == CELL_GCM_TEXTURE_X16					
// 		|| uBaseFormat == CELL_GCM_TEXTURE_Y16_X16				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_R5G5B5A1				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_COMPRESSED_HILO8		
// 		|| uBaseFormat == CELL_GCM_TEXTURE_COMPRESSED_HILO_S8		
//		|| uBaseFormat == CELL_GCM_TEXTURE_W16_Z16_Y16_X16_FLOAT	
// 		|| uBaseFormat == CELL_GCM_TEXTURE_W32_Z32_Y32_X32_FLOAT	
// 		|| uBaseFormat == CELL_GCM_TEXTURE_X32_FLOAT				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_D1R5G5B5				
// 		|| uBaseFormat == CELL_GCM_TEXTURE_Y16_X16_FLOAT			
		);
}

// KJFL - 28 April 2008 - texture streaming platform fns
#if defined (ASURA_USE_TEXTURE_STREAMING)

static Asura_Handle g_iStreamingFileHandle	= ASURA_INVALID_HANDLE;
#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
static Asura_PS3_FileSystem::Handle g_iStreamingCachePS3FileHandle	= Asura_PS3_FileSystem::INVALID_HANDLE;
#endif

// Set up the mem heap for the texture streaming on initialise
void Asura_Texture_Streaming_System::Platform_Initialise()
{
	if (ASURA_TEXTURE_STREAMING_MEMORY_ALLOWED_MB > 0)
	{
		Asura_PS3_Texture_Streaming_MemoryAllocator::Initialise(ASURA_TEXTURE_STREAMING_MEMORY_ALLOWED_MB, s_uMaxNumberOfLoadedTextures);
	}
}

void Asura_Texture_Streaming_System::Platform_Shutdown()
{
	Asura_PS3_Texture_Streaming_MemoryAllocator::Shutdown();
}

// open the streaming file - return true if successful
bool Asura_Texture_Streaming_System::Platform_StartStreaming( void )
{
	Asura_Assert( g_iStreamingFileHandle == ASURA_INVALID_HANDLE, "Code bug: is streaming file still open?" );

	Asura_PS3_Texture_Streaming_MemoryAllocator::Reset();
	g_iStreamingFileHandle = Asura_FileAccess::Open( Asura_Texture_Streaming_System::GetTextureStreamingFilename(), false, true );

#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
	//Clear the cache
	Asura_PS3_FileSystem::ClearCache();
	// Open the cache file for read and write
	Asura_PS3_FileSystem::OpenForReadWrite( "StreamingTextures.ps3", &g_iStreamingCachePS3FileHandle, false, true );
#endif

	// return true if open successful.
	return ( g_iStreamingFileHandle != ASURA_INVALID_HANDLE );
}

// close the streaming file
void Asura_Texture_Streaming_System::Platform_EndStreaming( void )
{
	if ( g_iStreamingFileHandle != ASURA_INVALID_HANDLE )
	{
		Asura_FileAccess::Close( g_iStreamingFileHandle );
	}
#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
	if( g_iStreamingCachePS3FileHandle != Asura_PS3_FileSystem::INVALID_HANDLE )
	{
		Asura_PS3_FileSystem::Close( &g_iStreamingCachePS3FileHandle );
		g_iStreamingCachePS3FileHandle	= Asura_PS3_FileSystem::INVALID_HANDLE;
	}
#endif
	g_iStreamingFileHandle		= ASURA_INVALID_HANDLE;
}

void Asura_Texture_Streaming_System::Platform_StreamTexture( Asura_Streaming_Texture& xTexture )
{
	// Since we have our mem allocator, changing the max number of loaded textures is a bad idea.
	Asura_Assert(s_uMaxNumberOfLoadedTextures == ASURA_TEXTURE_STREAMING_DEFAULT_MAX_LOADED_TEXTURES, "CODE BUG: Changing the maximum number of loaded streaming textures on PS3 is not supported.");

	// seek to the beginning of the texture resource
	bool bSuccess;
#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
	if( xTexture.IsInCache() )
	{
#if defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
		const Asura_TickCount xStart = Asura_Timers::GetCurTicks();
		Asura_OutputToDebugger_Retail(( "Loading texture %d from cache.\n", xTexture.GetTextureIndex() ));
#endif	//defined(ASURA_PS3_LOG_TEXTURE_STREAMING)

		bSuccess = Asura_PS3_FileSystem::SetCurrentPosition( g_iStreamingCachePS3FileHandle, xTexture.GetPositionInFile() );
		if (!bSuccess) return;
		// load the texture from the cache file
		bSuccess = g_axStreamingTextureStorage[ xTexture.GetTextureIndex() ].InitFromFile( ASURA_INVALID_HANDLE, g_iStreamingCachePS3FileHandle );

#if defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
		const Asura_TickCount xTimeInTicks = Asura_Timers::GetElapsedTicks( xStart );
		const float fTimeInSecs = Asura_Timers::TicksToSeconds( xTimeInTicks );
		Asura_OutputToDebugger_Retail(( "That took %.4f seconds.\n", fTimeInSecs ));
#endif	//defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
	}
	else
#endif
	{
#if defined(ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE) && defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
		const Asura_TickCount xStart = Asura_Timers::GetCurTicks();
		Asura_OutputToDebugger_Retail(( "Loading texture %d & putting into cache.\n", xTexture.GetTextureIndex() ));
#endif	//defined(ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE) && defined(ASURA_PS3_LOG_TEXTURE_STREAMING)

		bSuccess = Asura_FileAccess::SeekFile( g_iStreamingFileHandle, xTexture.GetPositionInFile() );
		if (!bSuccess) return;
#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
			if(g_iStreamingCachePS3FileHandle != Asura_PS3_FileSystem::INVALID_HANDLE)
			{
				Asura_PS3_FileSystem::SetCurrentPosition( g_iStreamingCachePS3FileHandle, xTexture.GetPositionInFile() );
			}
			// load the texture BlueRay and save it to cache file
			if(g_xPS3TextureStreamingMemoryAllocator.IsInitialised())
			{
				bSuccess = g_axStreamingTextureStorage[ xTexture.GetTextureIndex() ].InitFromFile( g_iStreamingFileHandle, g_iStreamingCachePS3FileHandle, PS3TextureStreamingMemoryAllocate);
			}
			else
			{
				bSuccess = g_axStreamingTextureStorage[ xTexture.GetTextureIndex() ].InitFromFile( g_iStreamingFileHandle, g_iStreamingCachePS3FileHandle );
			}

	#if defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
			const Asura_TickCount xTimeInTicks = Asura_Timers::GetElapsedTicks( xStart );
			const float fTimeInSecs = Asura_Timers::TicksToSeconds( xTimeInTicks );
			Asura_OutputToDebugger_Retail(( "That took %.4f seconds.\n", fTimeInSecs ));
	#endif	//defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
#else
		if(Asura_PS3_Texture_Streaming_MemoryAllocator::IsInitialised())
		{
			bSuccess = g_axStreamingTextureStorage[ xTexture.GetTextureIndex() ].InitFromFile( g_iStreamingFileHandle, Asura_PS3_FileSystem::INVALID_HANDLE, Asura_PS3_Texture_Streaming_MemoryAllocator::Allocate);
		}
		else
		{
			bSuccess = g_axStreamingTextureStorage[ xTexture.GetTextureIndex() ].InitFromFile( g_iStreamingFileHandle );
		}
#endif
	}

	xTexture.SetStatus( Asura_Streaming_Texture::AST_LOADED );
#ifdef ASURA_PLATFORM_USE_FILE_SYSTEM_CACHE
	// if the full texture is loaded in cache update the flag of the texture
	if( bSuccess )
	{
		xTexture.SetIsInCache( true );
	}
#endif
}

void Asura_Texture_Streaming_System::Platform_UnloadTexture( u_int uTextureID )
{
#if defined(ASURA_PS3_LOG_TEXTURE_STREAMING)
	Asura_OutputToDebugger_Retail(( "Unloading texture %d\n", uTextureID ));
#endif	//defined(ASURA_PS3_LOG_TEXTURE_STREAMING)

	// remove the texture from memory
	if(Asura_PS3_Texture_Streaming_MemoryAllocator::IsInitialised())
	{
		g_axStreamingTextureStorage[uTextureID].Destroy(Asura_PS3_Texture_Streaming_MemoryAllocator::Free);
	}
	else
	{
		g_axStreamingTextureStorage[uTextureID].Destroy();
	}

}

//////////////////////////////////////////////////////////////////////////

bool				Asura_PS3_Texture_Streaming_MemoryAllocator::s_bMemoryIsLocal	= false;
void*				Asura_PS3_Texture_Streaming_MemoryAllocator::s_pMemory			= NULL;
Asura_MemAllocator*	Asura_PS3_Texture_Streaming_MemoryAllocator::s_pxMemAllocator	= NULL;

void Asura_PS3_Texture_Streaming_MemoryAllocator::Initialise(const u_int uSizeInMB, const u_int uMaxTextures)
{
	// Allocate the size of memory needed PLUS 1kb per texture (in case of extra memory used by allocator)...
	const u_int	uSize	= (uSizeInMB + 1) * (1024 * 1024);

	// Attempt to allocate a single piece of memory large enough for our textures...
	s_pMemory	= ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL(uSize, 128);

	if(s_pMemory == NULL)
	{
		s_pMemory = ASURA_PS3_VIDEOMEMORY_ALLOCMAIN(uSize, 128);

		if(s_pMemory != NULL)
		{
			s_bMemoryIsLocal	= false;
		}
	}
	else
	{
		s_bMemoryIsLocal	= true;
	}

	// If we've succeeded, create the memory-allocator...
	if(s_pMemory != NULL)
	{
		s_pxMemAllocator	= new Asura_MemAllocator(s_pMemory, uSize, 128, (uMaxTextures* 2) + 1);
	}
}

void Asura_PS3_Texture_Streaming_MemoryAllocator::Shutdown(void)
{
	delete(s_pxMemAllocator);
	s_pxMemAllocator	= NULL;

	if(s_pMemory != NULL)
	{
		if(s_bMemoryIsLocal)
		{
			Asura_PS3_VideoMemoryManager::FreeLocal(s_pMemory);
		}
		else
		{
			Asura_PS3_VideoMemoryManager::FreeMain(s_pMemory);
		}
		s_pMemory	= NULL;
	}
}

void Asura_PS3_Texture_Streaming_MemoryAllocator::Reset(void)
{
	if (s_pxMemAllocator && s_pxMemAllocator->GetNumUsedBlocks() > 0)
	{
		Asura_Assert(false, "All streaming textures should have been unloaded before this point.");
		s_pxMemAllocator->Wipe();
	}
}

void* Asura_PS3_Texture_Streaming_MemoryAllocator::Allocate(const u_int uSize, const u_int uAlignment, CellGcmTexture& xGCMTexture)
{
	if(s_pxMemAllocator == NULL)
	{
		return(NULL);
	}
	else
	{
		xGCMTexture.location	= (s_bMemoryIsLocal) ? CELL_GCM_LOCATION_LOCAL : CELL_GCM_LOCATION_MAIN;
		return(s_pxMemAllocator->Allocate(uSize, uAlignment));
	}
}

void Asura_PS3_Texture_Streaming_MemoryAllocator::Free(void* pData)
{
	if(s_pxMemAllocator != NULL)
	{
		s_pxMemAllocator->Free(pData);
	}
}

#endif	//defined (ASURA_USE_TEXTURE_STREAMING)


/// End of file. ///
