/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.FXH"
#include "MaterialResponse_Shared.ps3FXH"

const float2	g_xTwoOverScreen;	// dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float2	g_xCameraTanHalfFOV;// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
const float4	g_xNearPlane;		// Fade and fallout controls
const float4	g_xFalloff;
const float		g_fDistanceFade;

const sampler	g_xNormalMaterialBufferSampler		:register( s0 );
const sampler	g_xAlbedoSpecularBufferSampler		:register( s1 );
const sampler	g_xDepthBufferSampler				:register( s2 );

const sampler	g_xProjectedTextureSampler		:register( s3 );
const sampler	g_xDetailTextureSampler			:register( s4 );

const float4x4	g_xProjectionMatrix;
const float4x4	g_xInvViewMatrix;
const float4	g_xDetailTextureScaleAndOffset;
const float4	g_xProjectorPosition;// Position of the projector
const float4	g_xProjectionDirection;
const float4	g_xLightColour;		// Colour/intensity of the light
const float4	g_xNeutralColour = float4( 0.0f, 0.0f, 0.0f, 0.0f );// Neutral colour for the current blend mode (0,0,0 for additive, 1,1,1 for multiply, etc)

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#ifdef SHADER_USE_DETAIL_TEXTURE
static const bool g_bUseDetailTexture = true;
#else
static const bool g_bUseDetailTexture = false;
#endif


/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// The vertex shader recreates the shadow frustum in world space, then projects
// it as an object in the standard camera view

struct VS_Output
{
	float4 m_xPosition				: POSITION;
	float3 m_xCameraDirDistance		: TEXCOORD0;
};

struct VS_Output_PosOnly
{
	float4 m_xPosition				: POSITION;	
};

#ifdef SHADER_DYNAMIC_LIGHT
float4 main( const VS_Output xIn, float2 xScreenPos : WPOS ) : COLOR
{
	const float4 xBufferNormalMaterial		= GetScreenColour( g_xNormalMaterialBufferSampler,		xScreenPos );
	const float4 xBufferAlbedoSpecular		= GetScreenColour( g_xAlbedoSpecularBufferSampler,		xScreenPos );
	const float4 xDepthBufferValue			= GetScreenColour( g_xDepthBufferSampler,				xScreenPos );

	float3	xNormal;
	float	fScreenDepth;
	float3	xAlbedoColour;
	float	hSpecularStrength;
	float	fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormalMaterial, xBufferAlbedoSpecular, xDepthBufferValue, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	xScreenPos -= 0.5f;
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.xy * fScreenDepth * ( xScreenPos * g_xTwoOverScreen.xy - 1.0h ),
													fScreenDepth, 1.0h );

	// Do the projection, and sample the projected texture
	// ------------------------------------------------------------------------------------------------
	float4 xProjectedPosition = mul(  xCameraViewSpacePosition, g_xProjectionMatrix );
	float3 xProjectedPosition2D = xProjectedPosition.xyz / xProjectedPosition.w;

	// DLH - Clip any XYZ value outside of 0-1
	clip( 0.5f - abs( xProjectedPosition2D.x - 0.5f ) );
	clip( 0.5f - abs( xProjectedPosition2D.y - 0.5f ) );
	clip( 0.5f - abs( xProjectedPosition2D.z - 0.5f ) );
	// Handle the falloff
	float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);
	clip ( 0.5f - abs( fDist - 0.5f ) );
	
	float4 f4Tex = h4tex2D( g_xProjectedTextureSampler, xProjectedPosition2D.xy );	
	
	// Handle detail texturing, if we're using it
	// ------------------------------------------------------------------------------------------------							
	if( g_bUseDetailTexture )
	{
		xProjectedPosition.xy = ( xProjectedPosition.xy * g_xDetailTextureScaleAndOffset.z ) + ( g_xDetailTextureScaleAndOffset.xy * xProjectedPosition.w );
		float4 f4DetailTex =  h4tex2Dproj( g_xDetailTextureSampler, xProjectedPosition.xyw );
		f4Tex *= f4DetailTex;
	}

	float fFalloff = min( 1.0f, pow( ( ( 1.0f - fDist ) * g_xFalloff.x ), g_xFalloff.y ) );		
	fFalloff = fFalloff *  g_fDistanceFade ;

	// Deferred dynamic light - material response
	// Diffuse lighting
	float4 xWorldPosition = mul(  xCameraViewSpacePosition, g_xInvViewMatrix );

	float4 xLightDelta = normalize( g_xProjectorPosition - xWorldPosition );

	float3 xLightAttenuatedColour = g_xLightColour.rgb * fFalloff * f4Tex.rgb;

	fFalloff = dot( ( f4Tex.xyz * fFalloff ), float3( 0.333f, 0.333f, 0.333f ) );

	f4Tex.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
		float4( xNormal, 1.0f ), normalize( xIn.m_xCameraDirDistance.xyz ), xLightDelta.xyz,
		xLightAttenuatedColour, xLightAttenuatedColour, fFalloff );

	f4Tex.a = 1.0f;

	return f4Tex;
}
#endif

// -----------------------------------------------------------------------------
// Pixel Shader (FakeLight)
// Projected texture as a fake light (multiply scaled by N.L)
// -----------------------------------------------------------------------------
#ifdef SHADER_FAKE_LIGHT
float4 main( float2 xScreenPos : WPOS) : COLOR
{
	const float4 xBufferNormalMaterial		= GetScreenColour( g_xNormalMaterialBufferSampler,		xScreenPos );
	const float4 xBufferAlbedoSpecular		= GetScreenColour( g_xAlbedoSpecularBufferSampler,		xScreenPos );
	const float4 xDepthBufferValue			= GetScreenColour( g_xDepthBufferSampler,				xScreenPos );

	float3	xNormal;
	float	fScreenDepth;
	float3	xAlbedoColour;
	float	hSpecularStrength;
	float	fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormalMaterial, xBufferAlbedoSpecular, xDepthBufferValue, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	xScreenPos -= 0.5f;
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.xy * fScreenDepth * ( xScreenPos * g_xTwoOverScreen.xy - 1.0h ),
													fScreenDepth, 1.0h );

	// Do the projection, and sample the projected texture
	// ------------------------------------------------------------------------------------------------
	float4 xProjectedPosition = mul(  xCameraViewSpacePosition, g_xProjectionMatrix );
	float3 xProjectedPosition2D = xProjectedPosition.xyz / xProjectedPosition.w;

	// DLH - Clip any XYZ value outside of 0-1
	clip( 0.5f - abs( xProjectedPosition2D.x - 0.5f ) );
	clip( 0.5f - abs( xProjectedPosition2D.y - 0.5f ) );
	clip( 0.5f - abs( xProjectedPosition2D.z - 0.5f ) );
	// Handle the falloff
	float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);
	clip ( 0.5f - abs( fDist - 0.5f ) );

	float4 f4Tex = h4tex2D( g_xProjectedTextureSampler, xProjectedPosition2D.xy );

	// Handle detail texturing, if we're using it
	// ------------------------------------------------------------------------------------------------
	if( g_bUseDetailTexture )
	{
		xProjectedPosition.xy = ( xProjectedPosition.xy * g_xDetailTextureScaleAndOffset.z ) + ( g_xDetailTextureScaleAndOffset.xy * xProjectedPosition.w );
		float4 f4DetailTex =  h4tex2Dproj( g_xDetailTextureSampler, xProjectedPosition.xyw );
		f4Tex = lerp( g_xNeutralColour, f4DetailTex,  f4Tex.r );
	}

	float fFalloff = min( 1.0f, pow( ( ( 1.0f - fDist ) * g_xFalloff.x ), g_xFalloff.y ) );
	fFalloff = fFalloff *  g_fDistanceFade ;

	// 'Fake Light' mode
	half3 xColor = xAlbedoColour;

	float4 xWorldPosition = mul(  xCameraViewSpacePosition, g_xInvViewMatrix );
	float4 xL = normalize( g_xProjectorPosition - xWorldPosition );

	// Scale the falloff by N.L
	float fDot =  dot( xL.xyz, xNormal.xyz );

	f4Tex.rgb = ( 8.0f * xAlbedoColour.xyz * f4Tex.xyz *  fFalloff * fDot );
	f4Tex.a = 1;

	return f4Tex;
}
#endif

// -----------------------------------------------------------------------------
// Pixel Shader (StandardProjector)
// Just project the texture
// -----------------------------------------------------------------------------
#ifdef SHADER_STANDARD_PROJECTOR
float4 main( float2 xScreenPos : WPOS ) : COLOR
{
	const float4 xBufferNormalMaterial		= GetScreenColour( g_xNormalMaterialBufferSampler,		xScreenPos );
	const float4 xBufferAlbedoSpecular		= GetScreenColour( g_xAlbedoSpecularBufferSampler,		xScreenPos );
	const float4 xDepthBufferValue			= GetScreenColour( g_xDepthBufferSampler,				xScreenPos );

	float3	xNormal;
	float	fScreenDepth;
	float3	xAlbedoColour;
	float	hSpecularStrength;
	float	fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormalMaterial, xBufferAlbedoSpecular, xDepthBufferValue, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	xScreenPos -= 0.5f;
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.xy * fScreenDepth * ( xScreenPos * g_xTwoOverScreen.xy - 1.0h ),
													fScreenDepth, 1.0h );
														
	// Do the projection, and sample the projected texture
	// ------------------------------------------------------------------------------------------------
	float4 xProjectedPosition = mul(  xCameraViewSpacePosition, g_xProjectionMatrix );
	float3 xProjectedPosition2D = xProjectedPosition.xyz / xProjectedPosition.w;

	// DLH - Clip any XYZ value outside of 0-1
	clip( 0.5f - abs( xProjectedPosition2D.x - 0.5f ) );
	clip( 0.5f - abs( xProjectedPosition2D.y - 0.5f ) );
	clip( 0.5f - abs( xProjectedPosition2D.z - 0.5f ) );
	// Handle the falloff
	float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);
	clip ( 0.5f - abs( fDist - 0.5f ) );
	
	float4 f4Tex = h4tex2D( g_xProjectedTextureSampler, xProjectedPosition2D.xy );	
		
	// Handle detail texturing, if we're using it
	// ------------------------------------------------------------------------------------------------
	if( g_bUseDetailTexture )
	{
		xProjectedPosition.xy = ( xProjectedPosition.xy * g_xDetailTextureScaleAndOffset.z ) + ( g_xDetailTextureScaleAndOffset.xy * xProjectedPosition.w );
		float4 f4DetailTex =  h4tex2Dproj( g_xDetailTextureSampler, xProjectedPosition.xyw );
		f4Tex = lerp( g_xNeutralColour, f4DetailTex,  f4Tex.r );
	}

	// Handle the falloff
	// ------------------------------------------------------------------------------------------------
	float fFalloff = min( 1.0f, pow( ( ( 1.0f - fDist ) * g_xFalloff.x ), g_xFalloff.y ) );
	fFalloff = fFalloff *  g_fDistanceFade ;
	f4Tex = lerp( g_xNeutralColour, f4Tex,  fFalloff );

	return f4Tex;
}
#endif


// -----------------------------------------------------------------------------
// Pixel Shader (Spotlight)
// Dynamic light, with material response - no projected texture
// -----------------------------------------------------------------------------
#ifdef SHADER_SPOT_LIGHT
float4 main( const VS_Output xIn, float2 xScreenPos : WPOS ) : COLOR
{
	const float4 xBufferNormalMaterial		= GetScreenColour( g_xNormalMaterialBufferSampler,		xScreenPos );
	const float4 xBufferAlbedoSpecular		= GetScreenColour( g_xAlbedoSpecularBufferSampler,		xScreenPos );
	const float4 xDepthBufferValue			= GetScreenColour( g_xDepthBufferSampler,				xScreenPos );

	float3	xNormal;
	float	fScreenDepth;
	float3	xAlbedoColour;
	float	hSpecularStrength;
	float	fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormalMaterial, xBufferAlbedoSpecular, xDepthBufferValue, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	xScreenPos -= 0.5f;
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.xy * fScreenDepth * ( xScreenPos * g_xTwoOverScreen.xy - 1.0h ),
													fScreenDepth, 1.0h );

	// Handle the falloff
	float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);
	clip ( 0.5f - abs( fDist - 0.5f ) );

	float fFalloff = min( 1.0f, pow( ( ( 1.0f - fDist ) * g_xFalloff.x ), g_xFalloff.y ) );
	fFalloff = fFalloff *  g_fDistanceFade ;

	// Calculate the angular falloff
	float4 xWorldPosition = mul(  xCameraViewSpacePosition, g_xInvViewMatrix );

	float fDot = dot( g_xProjectionDirection.xyz, normalize( xWorldPosition.xyz - g_xProjectorPosition.xyz ) );
	fFalloff *= saturate( ( fDot + g_xFalloff.z ) * g_xFalloff.w );

	// Deferred dynamic light - material response		 
	float4 xLightDelta = normalize( g_xProjectorPosition - xWorldPosition );

	half3 xLightAttenuatedColour = g_xLightColour.rgb * fFalloff;

	float4 f4Tex;	
	f4Tex.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
		float4( xNormal, 1.0f ), normalize( xIn.m_xCameraDirDistance.xyz ), xLightDelta.xyz,
		xLightAttenuatedColour, xLightAttenuatedColour, fFalloff );
							
	f4Tex.a = 1;				
		
	return f4Tex;
}
#endif
