//smooth skin shader, include this with various defines to create specific shaders for combinations of these options:
//	DAMAGE_MAPS - extract a per-bone damage value to pass to the pixel shader
//	PRESKINNED_POSITIONS - if the input position data has already been skinned (by SPU skinning)

//This shader uses a lot of constants, so don't allow compiler to use it's own ones, and explicitly define all global constants

//#include "Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const	float4		g_xCameraPosition : C8;
const	float4		g_xUV_Offsets : C9;

const float4x4	g_xWorldViewProjectionMatrix : C0;
const float4x4	g_xWorldMatrix : C4;

//this is the transpose of how we do it in the engine (converted when uploaded), to save constants (only need 3 constants not 4)
const float3x4	g_axTransforms[150] : C10; // * 3

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float3 m_xPosition		: ATTR0;
	float4 m_xUV			: ATTR1;

	int4   m_aiBoneIndices	: ATTR2;
	float4 m_afBoneWeights	: ATTR3;

	float3 m_xNormal		: ATTR4;
	float3 m_xTangent		: ATTR5;
	float3 m_xBinormal		: ATTR6;
};

struct OutputVertex
{
	float4 m_xPosition              : POSITION;

	float4 m_xUVs					: TEXCOORD0;

	float4 m_xCameraDirDistance     : TEXCOORD1; // Direction to camera, and distance from view plane in w
	float4 m_xTangentSpaceCameraDir : TEXCOORD2;

	float3 m_xNormal                : TEXCOORD3;
	float3 m_xTangent               : TEXCOORD4;
	float3 m_xBinormal              : TEXCOORD5;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

//from Shared.fxh, which is not included because it defines too many constants
float4 MaterialResponse_AnimateUVs( const float4 xUVPair )
{
	return xUVPair + g_xUV_Offsets;
}

OutputVertex main( const InputVertex xIn )
{
	// Position and UVs
	float4			xObjectSpacePosition = float4( 0.0f, 0.0f, 0.0f, 1.0f ); 
	float4			afBoneWeights = float4 ( xIn.m_afBoneWeights.xyz, 1-dot(xIn.m_afBoneWeights.xyz,float3(1)) );   //this takes 2 extra instructions, not too bad
	float			fAlbedoBlend = xIn.m_afBoneWeights[3]; //initial blend value

	OutputVertex	xOut;

	float3x4 axBoneMatrix;

	axBoneMatrix  = g_axTransforms[ xIn.m_aiBoneIndices[0] ] * afBoneWeights[0];
	axBoneMatrix += g_axTransforms[ xIn.m_aiBoneIndices[1] ] * afBoneWeights[1];
	axBoneMatrix += g_axTransforms[ xIn.m_aiBoneIndices[2] ] * afBoneWeights[2];
	axBoneMatrix += g_axTransforms[ xIn.m_aiBoneIndices[3] ] * afBoneWeights[3];

#ifdef DAMAGE_MAPS
	// Get the damagemap blend value
	fAlbedoBlend -= axBoneMatrix[2][2];
	
	// now re-compute that value in the matrix
	// would be nice to do these 4 calculations as float4's, but we'd have to build vectors with 1 element
	// from each matrix, which is presumably 1 cycle per element to construct, so instead:
	float2 xZ0 = g_axTransforms[ xIn.m_aiBoneIndices[0] ]._m00_m01 * g_axTransforms[ xIn.m_aiBoneIndices[0] ]._m11_m10;
	float2 xZ1 = g_axTransforms[ xIn.m_aiBoneIndices[1] ]._m00_m01 * g_axTransforms[ xIn.m_aiBoneIndices[1] ]._m11_m10;
	float2 xZ2 = g_axTransforms[ xIn.m_aiBoneIndices[2] ]._m00_m01 * g_axTransforms[ xIn.m_aiBoneIndices[2] ]._m11_m10;
	float2 xZ3 = g_axTransforms[ xIn.m_aiBoneIndices[3] ]._m00_m01 * g_axTransforms[ xIn.m_aiBoneIndices[3] ]._m11_m10;
	float2 xZ = xZ0*afBoneWeights[0]
			   +xZ1*afBoneWeights[1]
			   +xZ2*afBoneWeights[2]
			   +xZ3*afBoneWeights[3];

	axBoneMatrix[2][2] = xZ.x - xZ.y;
#endif

#ifdef PRESKINNED_POSITIONS
	xObjectSpacePosition.xyz	= xIn.m_xPosition.xyz;
#else
	xObjectSpacePosition.xyz	= mul( axBoneMatrix, float4( xIn.m_xPosition.xyz, 1.0f ) );
#endif
	xObjectSpacePosition.w  = 1.0f;

	float3x3		xMatrix;
	xMatrix = (float3x3)axBoneMatrix;

	xOut.m_xNormal	= mul( xMatrix, xIn.m_xNormal   );
	xOut.m_xTangent	= mul( xMatrix, xIn.m_xTangent  );
	xOut.m_xBinormal= mul( xMatrix, xIn.m_xBinormal );

	xOut.m_xPosition = mul( xObjectSpacePosition, g_xWorldViewProjectionMatrix );

	xOut.m_xUVs = MaterialResponse_AnimateUVs( xIn.m_xUV );
	// lerp seems to be backwards to what the damage maps expect
	xOut.m_xTangentSpaceCameraDir.w = saturate(fAlbedoBlend);

	// Direction to camera, and world-space distance to vertex from viewer

	const float4 xWorldSpacePosition = mul( xObjectSpacePosition, g_xWorldMatrix );

	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xWorldSpacePosition.xyz;
	xOut.m_xCameraDirDistance.w = xOut.m_xPosition.w;

	// We're doing lighting in world-space now
	float3x3 xWorldMatrix = (float3x3)g_xWorldMatrix;
	xOut.m_xNormal	 = mul( xOut.m_xNormal,		xWorldMatrix );
	xOut.m_xTangent	 = mul( xOut.m_xTangent,	xWorldMatrix );
	xOut.m_xBinormal = mul( xOut.m_xBinormal,	xWorldMatrix );

	const float3x3 xTangentToWorldMatrix = float3x3( xOut.m_xTangent, xOut.m_xBinormal, xOut.m_xNormal );

	// Note implicit transpose in the mul, by swapping operand order
	xOut.m_xTangentSpaceCameraDir.xyz = mul( xTangentToWorldMatrix, xOut.m_xCameraDirDistance.xyz );

	return xOut;
}
