// Asura_PS3_Multistream_File_Loading.cpp
// Thomas Young - 18.09.08
// Copyright (C) 2008 Rebellion, All Rights Reserved.

#include "Core/Asura.h"

#ifdef ASURA_USE_AUDIO_SYSTEM

// Asura
#include "Core/Asura_Multithread.h"
#include "Core/Threads/Asura_ProducerConsumerQueue.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_Memheap.h"
#include "Core/Asura_Hash.h"
#include "Sound/Asura_Audio_System.h"

// PS3
#include "Sound/Asura_PS3_Multistream_File_Loading.h"

// SDK
#include <cell/mstream.h>

// G L O B A L   V A R I A B L E S ///////////////////////////////////////////

static Asura_MemHeap*				g_pxStreamMemoryHeap	=	0;
static char*						g_pcStreamMemory		=	0;
static Asura_ProducerConsumerQueue	g_xStreamLoadingQueue( 8, 2, ASURA_MULTITHREAD_STACK_SIZE_SMALL );

// S T R U C T U R E S ///////////////////////////////////////////////////////

struct MSFHeader
{
	// copied from Multistream tools docs with correct 10 loop markers instead of 12 in there.

	char	header[3];			//Will be 'MSF'
	char	version;			//Revision of tool (Example: 0x32 = Version 3.2)
	int		compressionType;	//Will be the compression type of the output
	int		channels;			//Number of channels in the output file
	int		sampleSize;			//Number of *bytes* of sample data following the
	int		sampleRate;			//Playback sample rate
	int		miscInfo;			//See "More About the MSF miscInfo Field" section.
	int		LoopMarkers[10];	//10 loop markers. See "MSF Loop Markers" section.
};

struct __attribute__((packed)) WAVEFORMAT
{
	uint16	wFormatTag;			// format type
	uint16	nChannels;			// number of channels (i.e. mono, stereo...)
	uint32	nSamplesPerSec;		// sample rate
	uint32	nAvgBytesPerSec;	// for buffer estimation
	uint16	nBlockAlign;		// block size of data
	uint16	wBitsPerSample;		// Number of bits per sample of mono data 
};

// C O N S T A N T S /////////////////////////////////////////////////////////

static const char	szCHUNK_MSF	[]					= "MSF";
static const int	iMSF_HEADER_SIZE				= 64;
static const int	iSTREAMING_MP3_BUFFER_SIZE		= 24 * 1024;
static const int	iSTREAMING_CUSTOM_BUFFER_SIZE	= 192 * 1024;
static const int	iSTREAMING_BUFFER_HEAP_SIZE		= iSTREAMING_MP3_BUFFER_SIZE * ASURA_PS3_MULTISTREAM_MP3_STREAM_BUFFERS * 8
													+ iSTREAMING_CUSTOM_BUFFER_SIZE * ASURA_PS3_MULTISTREAM_CUSTOM_STREAM_BUFFERS;	//Enough for 8 streaming sounds + 1 custom streaming sound
static const u_int	uMP3_FRAME_HEADER_STATIC_BITS	= 0xFFFFFD0;

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

void Asura_Multistream_Streaming::InitialiseStreaming ()
{
	//This static function sets up the memheap
	g_pcStreamMemory		= static_cast<char*>( Asura_Allocate( iSTREAMING_BUFFER_HEAP_SIZE ) );
	g_pxStreamMemoryHeap	= new Asura_MemHeap( g_pcStreamMemory, iSTREAMING_BUFFER_HEAP_SIZE );

	Asura_Assert( g_pcStreamMemory && g_pxStreamMemoryHeap, "Failed to allocate memory for the PS3 sound streaming" );
}


void Asura_Multistream_Streaming::ShutdownStreaming ()
{
	delete g_pxStreamMemoryHeap;
	Asura_Deallocate( g_pcStreamMemory );
}


static int FindChunkInData ( const char* pcData, int iDataSize, const char *szChunkName )
{
	const char* pcRead = pcData;
	int iChunkNameLen = strlen(szChunkName);

	// Seach all data until chunk name located
	while ( iDataSize > 0 )
	{
		// Found first character of chunk name (check remainder of chunk) 
		if ( *pcRead == *szChunkName)
		{
			// Look for chunk name
			int iCount = 0;
			while ( (pcRead[iCount] == szChunkName[iCount]) && (iCount < iChunkNameLen) ) 
			{ 
				iCount++; 
			}
			if ( iCount == iChunkNameLen ) 
			{ 
				return iCount + int( pcRead ) - int( pcData );	// All counts used therefore chunk matched
			}  
		}						   
		pcRead++;
		iDataSize--;
	}

	return -1;
}


static bool LoadMSFInfo ( void* pBuffer, int iBufferSize, Asura_PS3_Sound_Info& xSoundInfo )
{
	Asura_Assert(iBufferSize >= iMSF_HEADER_SIZE, "buffer too small for MSF");
	const MSFHeader* pxMSFHeader = reinterpret_cast<MSFHeader*>(pBuffer);

	xSoundInfo.SetSampleFormat(pxMSFHeader->compressionType);
	xSoundInfo.SetChannelCount(pxMSFHeader->channels);
	xSoundInfo.SetTotalDataBytes(pxMSFHeader->sampleSize);

	xSoundInfo.SetSampleRate(pxMSFHeader->sampleRate);

	Asura_Assert( xSoundInfo.GetSampleFormat() == CELL_MS_MP3, "Loading a sound which is not an MP3" );
	if ( xSoundInfo.GetSampleFormat() == CELL_MS_MP3 )
	{
		// code based on https://ps3.scedev.net/forums/thread/41486
		CellMSMP3FrameHeader xMP3Hdr;
		int iRet=cellMSMP3GetFrameInfo( &static_cast<char*>(pBuffer)[iMSF_HEADER_SIZE], &xMP3Hdr ); 
		if ( iRet==-1 )
		{
			Asura_Assert(false, "Invalid MP3 header");
			return false;
		}
		const int aiSampsMPG2[3] =	{ 192, 1152, 576  };
		const int aiSampsMPG1[3] =	{ 384, 1152, 1152 };
		
		const int iMPEGver = xMP3Hdr.Sync&( 1 << 19 );
		
		if (iMPEGver == 0) // MPEG Version 2 or 2.5
		{
			xSoundInfo.SetSamplesPerMP3Frame(aiSampsMPG2[xMP3Hdr.Layer-1]); // should be 576
		}
		else // MPEG Version 1
		{
			xSoundInfo.SetSamplesPerMP3Frame(aiSampsMPG1[xMP3Hdr.Layer-1]); // should be 1152
		}
		
		xSoundInfo.SetMP3FrameSize	( xMP3Hdr.PacketSize );
		xSoundInfo.SetBitRate		( xMP3Hdr.BitRate );
	}
	else
	{
		return false;
	}

	// loop offset
	xSoundInfo.SetLoopOffset( 0 );

	// Data start (constant)
	xSoundInfo.SetDataOffset( iMSF_HEADER_SIZE );
	return true;
}


//___________________________________________________________________________________________________


Asura_MultiStream_Buffer::Asura_MultiStream_Buffer ()
{
	m_pcWaveData	= 0;
	m_bLooping		= false;
	m_szFilename	= 0;
	m_xFilenameHash = ASURA_HASH_ID_UNSET;

	Asura_Mem::Set( &m_xSoundInfo, 0, sizeof( Asura_PS3_Sound_Info ) );
}


Asura_MultiStream_Buffer::~Asura_MultiStream_Buffer ()
{
	if( m_pcWaveData )
	{
		Asura_DeallocateAlign( m_pcWaveData );
	}
}


bool Asura_MultiStream_Buffer::Load ( Asura_Chunk_Stream& xStream, const char* const szFilename, const u_int uFilesize )
{
	char acBuffer[ 128 ];	//The buffer must be larger than iMSF_HEADER_SIZE so that we can read an MP3 frame header
	u_int uBytesRead = 0;

	uBytesRead = xStream.Read( 128, acBuffer );

	if( uBytesRead != static_cast<u_int>( 128 ) )
	{
		return false;
	}

	if ( FindChunkInData(acBuffer, 128, szCHUNK_MSF) == 3 )
	{
		if (LoadMSFInfo( acBuffer, 128, m_xSoundInfo ))
		{
			u_int uSampleDataSize = uFilesize - iMSF_HEADER_SIZE;

			m_pcWaveData = static_cast<char*>( Asura_AllocateAlign( uSampleDataSize, ASURA_ALIGN_16_BYTE ) );
			if( m_pcWaveData )
			{
				const u_int uCurrentReadOffset = 128 - iMSF_HEADER_SIZE;
				Asura_Mem::Copy( m_pcWaveData, &acBuffer[iMSF_HEADER_SIZE], uCurrentReadOffset );
				if( xStream.Read( uSampleDataSize - uCurrentReadOffset, &m_pcWaveData[ uCurrentReadOffset ] ) != ( uSampleDataSize - uCurrentReadOffset ) )
				{
					Asura_Assert( false, "Failed to read data from chunk stream" );
					Asura_DeallocateAlign( m_pcWaveData );
					return false;
				}

				m_szFilename = 0;
				#if defined ASURA_AUDIO_STORE_FILENAMES
				m_szFilename = Asura_StringUtil::Duplicate( szFilename );
				#endif
				m_xFilenameHash = Asura_GetHashID( szFilename );
			}

			return true;
		}
	}

	Asura_Assert( false, ("Unable to identify audio data format") );
	return false;
}


//___________________________________________________________________________________________________


Asura_MultiStream_MP3_Stream::Asura_MultiStream_MP3_Stream ( const Asura_Hash_ID xWavHash )
{
	const Asura_Audio_StreamInfo* const pxStreamInfo = Asura_Audio_System::GetStreamInfo( xWavHash );

	if( !pxStreamInfo )
	{
		Asura_Assert( false, "No stream info, cannot play streaming sound" );
		m_eState = STREAM_STATE_ERROR;

		for( int iBuffer = 0; iBuffer < ASURA_PS3_MULTISTREAM_MP3_STREAM_BUFFERS; ++iBuffer )
		{
			m_apStreamBuffer [ iBuffer ] = 0;
		}
	}
	else
	{
		m_iFilehandle					= -1;
		m_xFilenameHash					= xWavHash;
		m_eState						= STREAM_STATE_NEED_PREPARE;

		m_szFilename = 0;
		#if defined ASURA_AUDIO_STORE_FILENAMES
		m_szFilename = Asura_StringUtil::Duplicate( pxStreamInfo->m_szFilename );
		#endif

		for( int iBuffer = 0; iBuffer < ASURA_PS3_MULTISTREAM_MP3_STREAM_BUFFERS; ++iBuffer )
		{
			m_apStreamBuffer			 [ iBuffer ] = static_cast<char*>( g_pxStreamMemoryHeap->Alloc( iSTREAMING_MP3_BUFFER_SIZE, 128 ) );
			m_auStreamBufferSize		 [ iBuffer ] = 0;
			m_bStreamBufferDiscontinuous [ iBuffer ] = false;
			m_afStreamBufferTimes		 [ iBuffer ] = 0.0f;

			Asura_Assert( m_apStreamBuffer[ iBuffer ], "Ran out of memory in the streaming sound heap" );
		}
		m_iActiveBuffer					= 0;
		m_fSeekTime						= 0.0f;
		m_bShouldSeek					= false;
		m_bLooping						= pxStreamInfo ? pxStreamInfo->m_bLooping : false;
		m_bCompleted					= false;
		m_iBufferReadSize				= iSTREAMING_MP3_BUFFER_SIZE;
		m_bShouldCleanRead				= false;
		m_iProcessingBuffer				= 0;
		m_pxStreamInfo					= 0;
		m_uFrameSyncBytes				= 0;
	}
}


Asura_MultiStream_MP3_Stream::~Asura_MultiStream_MP3_Stream ()
{
	while( m_eState == STREAM_STATE_READING )
	{
		Asura_Thread_System::DelayCurrentThread( 0.05f );
	}
	for( int iBuffer = 0; iBuffer < ASURA_PS3_MULTISTREAM_MP3_STREAM_BUFFERS; ++iBuffer )
	{
		if( m_apStreamBuffer[ iBuffer ] )
		{
			g_pxStreamMemoryHeap->Free( m_apStreamBuffer[ iBuffer ] );
		}
	}
	
	if( !m_bCompleted && ( m_eState != STREAM_STATE_ERROR ) )
	{
		Asura_PS3_FileSystem::Close( &m_iFilehandle );
	}
}


bool Asura_MultiStream_MP3_Stream::Prepare ()
{
	if( m_eState != STREAM_STATE_NEED_PREPARE )
	{
		Asura_Assert( false, "Stream does not need preparing" );
		return false;
	}

	m_eState = STREAM_STATE_PREPARING;

	Asura_Assert( m_eState == STREAM_STATE_PREPARING, "Incorrect state in stream" );

	// Get the info for this file stream from the audio system (slightly confusing name sorry)
	m_pxStreamInfo = Asura_Audio_System::GetStreamInfo( m_xFilenameHash );

	m_xSoundInfo.SetDataOffset( m_xSoundInfo.GetDataOffset() );

	if( !m_pxStreamInfo )
	{
		Asura_Assert(false, "Failed to to get stream info" );
		m_eState = STREAM_STATE_ERROR;
		return false;
	}

	// Attempt to pre-open the file (this can take a while for the disk to seek)
	if ( !Asura_PS3_FileSystem::OpenForReading( m_pxStreamInfo->m_szPackage, &m_iFilehandle ) )
	{
		Asura_Assertf( false, ("Unable to open file %s", m_pxStreamInfo->m_szPackage ) );
		m_eState = STREAM_STATE_ERROR;
		return false;
	}

	if( !Asura_PS3_FileSystem::SetCurrentPosition( m_iFilehandle, m_pxStreamInfo->m_uLocation ) )
	{
		Asura_Assert( false, "Failed to seek in file" );
		Asura_PS3_FileSystem::Close( &m_iFilehandle );
		m_eState = STREAM_STATE_ERROR;
		return false;
	}

	// Read the header, we need to read more than the header size to get an MP3 frame header
	char acBuffer[ 128 ];

	if( Asura_PS3_FileSystem::ReadData( m_iFilehandle, 128, &acBuffer[0] ) != static_cast<u_int>( 128 ) )
	{
		Asura_Assertf( false, ("Failed to read data: %s", szASURA_AUDIO_STREAMING_FILENAME ) );
		Asura_PS3_FileSystem::Close( &m_iFilehandle );
		m_eState = STREAM_STATE_ERROR;
		return false;
	}

	if ( FindChunkInData( acBuffer, 128, szCHUNK_MSF ) == 3 )
	{
		if (LoadMSFInfo( acBuffer, 128, m_xSoundInfo ))
		{
			// Save the frame header bytes so we can sync when seeking
			m_uFrameSyncBytes = *( reinterpret_cast< u_int* >( &acBuffer[ iMSF_HEADER_SIZE ] ) );
			m_uFrameSyncBytes &= uMP3_FRAME_HEADER_STATIC_BITS;
		}
		else
		{
			// LoadMSFInfo will provide an assert, no need to duplicate it
			return false;
		}
	}
	else
	{
		Asura_Assertf( false, ("Failed to find MSF info: %s", szASURA_AUDIO_STREAMING_FILENAME ) );
		Asura_PS3_FileSystem::Close( &m_iFilehandle );
		m_eState = STREAM_STATE_ERROR;
		return false;
	}

	Asura_PS3_FileSystem::SetCurrentPosition( m_iFilehandle, m_pxStreamInfo->m_uLocation + m_xSoundInfo.GetDataOffset() );

	m_eState = STREAM_STATE_IDLE;

	return true;
}


void Asura_MultiStream_MP3_Stream::Read ( const bool bThreaded )
{
	if( m_eState != STREAM_STATE_IDLE )
	{
		Asura_Assert( false, "Stream is not idle, cannot read more" );
		return;
	}

	m_eState = STREAM_STATE_READING;

	if( bThreaded )
	{
		#ifdef ASURA_ASSERT
		if( !g_xStreamLoadingQueue.TryToAddToQueue( Asura_MultiStream_MP3_Stream::ThreadFunc_LoadData, this, 0 ) )
		{
			Asura_Assert( false, "Multistream streaming ran out of space on producer consumer queue" );
		}
		#else
		g_xStreamLoadingQueue.TryToAddToQueue( Asura_MultiStream_MP3_Stream::ThreadFunc_LoadData, this, 0 );
		#endif
	}
	else
	{
		// Non-threaded load
		Asura_MultiStream_MP3_Stream::ThreadFunc_LoadData( this, 0 );
	}
}


bool Asura_MultiStream_MP3_Stream::GetBuffer ( char* &pcBuffer, u_int &uBufferSize, bool &bDiscontinuous, float &fBufferStartTime )
{
	if( m_eState != STREAM_STATE_LOADED ) 
	{
		Asura_Assert( false, "Cannot return buffer, wrong stream state" );
		return false;
	}

	pcBuffer			= m_apStreamBuffer				[ m_iActiveBuffer ];
	uBufferSize			= m_auStreamBufferSize			[ m_iActiveBuffer ];
	bDiscontinuous		= m_bStreamBufferDiscontinuous	[ m_iActiveBuffer ];
	fBufferStartTime	= m_afStreamBufferTimes			[ m_iActiveBuffer ];
	
	m_iActiveBuffer = ( m_iActiveBuffer + 1 ) % ASURA_PS3_MULTISTREAM_MP3_STREAM_BUFFERS;

	if( m_bCompleted )
	{
		m_eState = STREAM_STATE_COMPLETED;
	}
	else
	{
		m_eState = STREAM_STATE_IDLE;
	}

	return true;
}


void Asura_MultiStream_MP3_Stream::SeekNextRead ( const float fSeekTime )
{
	if( m_eState == STREAM_STATE_READING )
	{
		Asura_Assert( false, "Cannot seek, wrong stream state" );
		return;
	}

	m_bShouldSeek = true;
	m_fSeekTime = fSeekTime;
}


int Asura_MultiStream_MP3_Stream::GetDefaultBufferSize()
{
	return iSTREAMING_MP3_BUFFER_SIZE;
}


void Asura_MultiStream_MP3_Stream::SetNextReadSize ( const float fDuration )
{
	if( m_eState == STREAM_STATE_READING )
	{
		Asura_Assert( false, "Cannot set next read size, currently reading" );
		return;
	}
	const float fBytesPerSec = static_cast<float>( ( m_xSoundInfo.GetBitRate() * 1000 ) / 8 );
	int iSamples = static_cast<int>( fDuration * fBytesPerSec );
	if( iSamples > iSTREAMING_MP3_BUFFER_SIZE )
	{
		Asura_Assert( false, "Cannot set read size greater than the default" );
		return;
	}

	m_iBufferReadSize = iSamples;
}

void Asura_MultiStream_MP3_Stream::CleanNextRead ()
{
	if( m_eState == STREAM_STATE_READING )
	{
		Asura_Assert( false, "Cannot set 'clean next read' , currently reading" );
		return;
	}
	//Clean next read means the data buffer returned will terminate on a frame boundry
	m_bShouldCleanRead = true;
}


float Asura_MultiStream_MP3_Stream::GetReadPosition ( int iBufferOffset )
{
	if( m_eState < STREAM_STATE_IDLE )
	{
		return 0.0f;
	}

	const float fBytesPerSec = (m_xSoundInfo.GetBitRate() * 1000) / 8;
	float fPlayTime = 0.0f;

	if( iBufferOffset < 0 )
	{
		//We need the previous buffer, and take the buffer size + offset (offset is negative)
		int iBuffer = ( m_iProcessingBuffer - 1 );
		if( iBuffer < 0 )
		{
			iBuffer += ASURA_PS3_MULTISTREAM_MP3_STREAM_BUFFERS;
		}
		fPlayTime = m_afStreamBufferTimes[ iBuffer ] + ( static_cast<float>( m_auStreamBufferSize[ iBuffer ] + iBufferOffset ) / fBytesPerSec );
	}
	else
	{
		//We need the current buffer
		fPlayTime = ( m_afStreamBufferTimes[ m_iProcessingBuffer ] + static_cast<float>( iBufferOffset ) / fBytesPerSec );
	}

	return fPlayTime;
}


void Asura_MultiStream_MP3_Stream::ThreadFunc_LoadData ( void* pxData, u_int uDataSize )
{
	Asura_MultiStream_MP3_Stream* pxThis = reinterpret_cast<Asura_MultiStream_MP3_Stream*>( const_cast<void*>( pxData ) );

	Asura_Assert( pxThis->m_eState == STREAM_STATE_READING, "Incorrect state in stream" );

	const u_int uFileOffset = pxThis->m_pxStreamInfo->m_uLocation;

	//Get the desired read size and reset it to the default
	int iDesiredRead = pxThis->m_iBufferReadSize;
	pxThis->m_iBufferReadSize = iSTREAMING_MP3_BUFFER_SIZE;

	if( pxThis->m_bShouldSeek )
	{
		pxThis->m_bShouldSeek = false;

		//Need to seek, work out what sample matches the time required
		if( pxThis->m_fSeekTime == 0.0f )
		{
			//easy case
			if( !Asura_PS3_FileSystem::SetCurrentPosition( pxThis->m_iFilehandle, pxThis->m_xSoundInfo.GetDataOffset() + uFileOffset ) )
			{
				Asura_PS3_FileSystem::Close( &( pxThis->m_iFilehandle ) );
				pxThis->m_eState = Asura_MultiStream_MP3_Stream::STREAM_STATE_ERROR;
				return;
			}
		}
		else
		{
			if( !pxThis->SeekToTimeInFile( pxThis->m_fSeekTime ) )
			{
				Asura_PS3_FileSystem::Close( &( pxThis->m_iFilehandle ) );
				pxThis->m_eState = Asura_MultiStream_MP3_Stream::STREAM_STATE_ERROR;
				return;
			}
		}

		//We have seeked, so this buffer will not follow previous one
		pxThis->m_bStreamBufferDiscontinuous[pxThis->m_iActiveBuffer] = true;
	}
	else
	{
		pxThis->m_bStreamBufferDiscontinuous[pxThis->m_iActiveBuffer] = false;
	}

	bool bCheckCompletion = false;
	const int iCurrentPosition = static_cast<int>( Asura_PS3_FileSystem::GetCurrentPosition( pxThis->m_iFilehandle ) );
	const int iEndOfData = static_cast<int>( pxThis->GetSampleDataSize() + pxThis->GetSampleDataOffset() + uFileOffset );
	if( ( iCurrentPosition + iDesiredRead ) >= iEndOfData )
	{
		//Dont read past the end of the data
		iDesiredRead = iEndOfData - iCurrentPosition;
		Asura_Assert( iDesiredRead > 2048, "Read a very small buffer, might not have time to process the next one" );
		bCheckCompletion = true;
	}
	else
	{
		// Try to deal with the case where we are left with a very small amount of data to read, resulting in starvation when we cannot process the subsequent buffer quickly enough
		const int iSamplesRemaining = iEndOfData - iCurrentPosition;
		if( iSamplesRemaining < static_cast<int>( static_cast<float>( iSTREAMING_MP3_BUFFER_SIZE ) * 1.5f ) )
		{
			iDesiredRead = Asura_Maths::Min( static_cast<int>( static_cast<float>( iSamplesRemaining ) * 0.6f ), iDesiredRead );
		}
	}

	if( iDesiredRead <= 0 )
	{
		Asura_Assertf( iDesiredRead > 0, ("Error, desired read is negative (or zero)! Current pos: %d, end of data: %d ", iCurrentPosition, iEndOfData ) );
		Asura_PS3_FileSystem::Close( &( pxThis->m_iFilehandle ) );
		pxThis->m_eState = Asura_MultiStream_MP3_Stream::STREAM_STATE_ERROR;
		return;
	}
			
	pxThis->m_afStreamBufferTimes[pxThis->m_iActiveBuffer] = pxThis->GetMP3PlayPosFromFileOffset( iCurrentPosition - uFileOffset );

	pxThis->m_auStreamBufferSize[pxThis->m_iActiveBuffer] = Asura_PS3_FileSystem::ReadData	( pxThis->m_iFilehandle
																							, iDesiredRead
																							, pxThis->m_apStreamBuffer[pxThis->m_iActiveBuffer] );

	if( bCheckCompletion )
	{
		//Not reading a full buffer indicates we have reached the end of the sample data
		if( pxThis->IsLooping() )
		{
			pxThis->m_bShouldSeek	= true;
			pxThis->m_fSeekTime		= 0.0f;
		}
		else
		{
			pxThis->m_bCompleted = true;
			//Close the file
			Asura_PS3_FileSystem::Close( &pxThis->m_iFilehandle );
		}
	}
	else if ( pxThis->m_bShouldCleanRead )
	{
		//Find a frame boundry and modify our returned buffer size with it
		pxThis->m_auStreamBufferSize[pxThis->m_iActiveBuffer] -= 4;	//go four bytes back so we can start checking for sync bytes
		char* pucBuffer = &(pxThis->m_apStreamBuffer[ pxThis->m_iActiveBuffer ])[ pxThis->m_auStreamBufferSize[pxThis->m_iActiveBuffer] ];
		
		//while( ( *reinterpret_cast<unsigned short*>(pucBuffer) & 0xFFF0 ) != 0xFFF0 )
		while( !pxThis->IsMP3FrameStart( *reinterpret_cast<u_int*>(pucBuffer) ) )
		{
			--pxThis->m_auStreamBufferSize[pxThis->m_iActiveBuffer];
			--pucBuffer;
		}
		pxThis->m_bShouldCleanRead = false;
	}

	__eieio();	//io barrier
	pxThis->m_eState = Asura_MultiStream_MP3_Stream::STREAM_STATE_LOADED;
}


float Asura_MultiStream_MP3_Stream::GetMP3PlayPosFromFileOffset ( const int iFileOffset ) const
{
	const float fBytesPerSec = (m_xSoundInfo.GetBitRate() * 1000) / 8;
	return (static_cast<float>(iFileOffset - iMSF_HEADER_SIZE) / fBytesPerSec);
}


inline bool Asura_MultiStream_MP3_Stream::IsMP3FrameStart( const u_int uBytes )
{
	// check if the appropriate bits match our saved MP3 frame header
	// some bits can change between frames (e.g. the padding bit) so we have to mask it

	return bool( ( uBytes & uMP3_FRAME_HEADER_STATIC_BITS ) == ( m_uFrameSyncBytes ) );
}


bool Asura_MultiStream_MP3_Stream::SeekToTimeInFile ( const float fSeekTime )
{
	if( !m_pxStreamInfo ) 
	{
		return false;
	}

	const float fBytesPerSec = static_cast<float>( ( m_xSoundInfo.GetBitRate() * 1000 ) / 8 );	
	const u_int uSeekSample = static_cast<u_int>( fSeekTime * fBytesPerSec ) + m_xSoundInfo.GetDataOffset();

	if( uSeekSample > ( GetSampleDataSize() - 512 ) )
	{
		// This is too late in the data to seek to
		return false;
	}

	//We need to seek to the rough location, then find a frame boundry to read from
	if( !Asura_PS3_FileSystem::SetCurrentPosition( m_iFilehandle, m_pxStreamInfo->m_uLocation + uSeekSample ) )
	{
		return false;
	}

	unsigned char pcTempBuffer[2048];

	u_int uRead = Asura_PS3_FileSystem::ReadData( m_iFilehandle, 2048, pcTempBuffer );

	if( uRead != 2048 )
	{
		return false;
	}

	//Find a frame boundry
	unsigned char* pucBuffer = static_cast<unsigned char*>( &pcTempBuffer[0] );
	u_int uSeek = 0;
	while( !IsMP3FrameStart( *reinterpret_cast<u_int*>(pucBuffer) ) )
	{
		++uSeek;
		++pucBuffer;
	}

	if( !Asura_PS3_FileSystem::SetCurrentPosition( m_iFilehandle, m_pxStreamInfo->m_uLocation + uSeekSample + uSeek ) )
	{
		return false;
	}

	return true;
}


//___________________________________________________________________________________________________


Asura_MultiStream_Custom_Stream::Asura_MultiStream_Custom_Stream ( const Asura_PS3_Audio_MoreDataFunc pfnMoreDataCallback, void* pMoreDataCallbackArg )
{
	m_pfnMoreDataCallback	= pfnMoreDataCallback;
	m_pMoreDataCallbackArg	= pMoreDataCallbackArg;

	m_eState = STREAM_STATE_IDLE;

	for( int iBuffer = 0; iBuffer < ASURA_PS3_MULTISTREAM_CUSTOM_STREAM_BUFFERS; ++iBuffer )
	{
		m_apStreamBuffer			 [ iBuffer ] = static_cast<char*>( g_pxStreamMemoryHeap->Alloc( iSTREAMING_CUSTOM_BUFFER_SIZE, 128 ) );
		m_auStreamBufferSize		 [ iBuffer ] = 0;

		Asura_Assert( m_apStreamBuffer[ iBuffer ], "Ran out of memory in the streaming sound heap" );
	}
	m_iActiveBuffer = 0;	
	m_bCompleted = false;
}


Asura_MultiStream_Custom_Stream::~Asura_MultiStream_Custom_Stream ()
{
	while( m_eState == STREAM_STATE_READING )
	{
		Asura_Thread_System::DelayCurrentThread( 0.05f );
	}
	for( int iBuffer = 0; iBuffer < ASURA_PS3_MULTISTREAM_CUSTOM_STREAM_BUFFERS; ++iBuffer )
	{
		if( m_apStreamBuffer[ iBuffer ] )
		{
			g_pxStreamMemoryHeap->Free( m_apStreamBuffer[ iBuffer ] );
		}
	}
}


bool Asura_MultiStream_Custom_Stream::GetBuffer ( char* &pcBuffer, u_int &uBufferSize )
{
	if( m_eState != STREAM_STATE_LOADED ) 
	{
		Asura_Assert( false, "Cannot return buffer, wrong stream state" );
		return false;
	}

	pcBuffer			= m_apStreamBuffer		[ m_iActiveBuffer ];
	uBufferSize			= m_auStreamBufferSize	[ m_iActiveBuffer ];
	
	m_iActiveBuffer = ( m_iActiveBuffer + 1 ) % ASURA_PS3_MULTISTREAM_CUSTOM_STREAM_BUFFERS;

	if( m_bCompleted )
	{
		m_eState = STREAM_STATE_COMPLETED;
	}
	else
	{
		m_eState = STREAM_STATE_IDLE;
	}

	return true;
}


void Asura_MultiStream_Custom_Stream::Read ( const bool bThreaded )
{
	if( m_eState != STREAM_STATE_IDLE )
	{
		Asura_Assert( false, "Stream is not idle, cannot read more" );
		return;
	}

	if( !m_pfnMoreDataCallback )
	{
		Asura_Assert( false, "No callback to get more data" );
		return;
	}

	if( bThreaded )
	{
		m_eState = STREAM_STATE_READING;

		#ifdef ASURA_ASSERT
		if( !g_xStreamLoadingQueue.TryToAddToQueue( Asura_MultiStream_Custom_Stream::ThreadFunc_GetData, this, 0 ) )
		{
			Asura_Assert( false, "Multistream streaming ran out of space on producer consumer queue" );
		}
		#else
		g_xStreamLoadingQueue.TryToAddToQueue( Asura_MultiStream_Custom_Stream::ThreadFunc_GetData, this, 0 );
		#endif
	}
	else
	{
		m_auStreamBufferSize[ m_iActiveBuffer ] = static_cast<u_int>( GetDefaultBufferSize() );

		if( !m_pfnMoreDataCallback( m_apStreamBuffer[ m_iActiveBuffer ], m_auStreamBufferSize[ m_iActiveBuffer ], m_pMoreDataCallbackArg ) )
		{
			m_bCompleted = true;
		}
		m_eState = STREAM_STATE_LOADED;
	}
}


int Asura_MultiStream_Custom_Stream::GetDefaultBufferSize ()
{
	return iSTREAMING_CUSTOM_BUFFER_SIZE;
}


void Asura_MultiStream_Custom_Stream::ThreadFunc_GetData ( void* pxData, u_int uDataSize )
{
	Asura_MultiStream_Custom_Stream* pxThis = reinterpret_cast<Asura_MultiStream_Custom_Stream*>( const_cast<void*>( pxData ) );

	if( pxThis )
	{
		Asura_Assert( pxThis->m_eState == STREAM_STATE_READING, "Incorrect state in stream" );

		pxThis->m_auStreamBufferSize[ pxThis->m_iActiveBuffer ] = static_cast<u_int>( GetDefaultBufferSize() );

		if( !pxThis->m_pfnMoreDataCallback( pxThis->m_apStreamBuffer[ pxThis->m_iActiveBuffer ], pxThis->m_auStreamBufferSize[ pxThis->m_iActiveBuffer ], pxThis->m_pMoreDataCallbackArg ) )
		{
			pxThis->m_bCompleted = true;
		}

		__eieio();	//io barrier
		pxThis->m_eState = STREAM_STATE_LOADED;
	}
}


#endif //ASURA_USE_AUDIO_SYSTEM

//eof
