///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_Sound_System.cpp
// Author    : Philip Hall
// Created   : 11/09/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifndef ASURA_USE_AUDIO_SYSTEM

// This file's header
#include "Sound/Asura_PS3_Sound_System.h"

// Asura headers
#include "Core/Asura_HeapArray.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_LogFile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Sound/Asura_Sound.h"

// Asura PS3 headers
#include "Sound/Asura_PS3_Sound_Loader.h"
#include "Sound/Asura_PS3_MultiStream_Streaming.h"

#include <sys/ppu_thread.h>

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////

// Loaded sound tracking
Asura_PS3_LoadedSound	Asura_PS3_LoadedSound::s_axLoadedSounds[ASURA_MAX_NUMBER_OF_LOADED_SOUNDS];
int						Asura_PS3_LoadedSound::s_iNumberOfLoadedSounds = 0;
u_int					Asura_PS3_LoadedSound::s_uLoadedSoundsSizeTotal = 0;


#ifdef ASURA_PS3_EXTRA_ASYNC_THREAD
static const int	ASYNC_PPU_THREAD_PRIORITY   = 1000;
static const size_t	ASYNC_THREAD_STACK_SIZE     = 8 * 1024;	// Tiny 8k stack
static const int   	ASYNC_PLAYBACK_QUEUE_LENGTH = 8;
static const int    PACKED_LOOPING_BIT          = 0x10000000;
#endif

// Active sound tracking
Asura_FixedArray<Asura_PS3_ActiveSound, iASURA_PS3_SOUND_MAX_VOICES> Asura_PS3_ActiveSound::s_axActiveSounds;

// Sound system access point
Asura_PS3_Sound_System	Asura_PS3_Sound_System::s_xAPI;

// Repeated error messages
static const char szINVALID_LOAD_INDEX[] = "Invalid load index (out of range)";
static const char szINVALID_SOUND_HANDLE[] = "Invalid sound handle (out of range)";

// This single static doesn't really need a file to itself
#if defined (ASURA_PS3_USE_NULLSOUND) && (ASURA_PS3_USE_NULLSOUND != 0) 
	#include "Sound/Asura_PS3_NullSound.h"
	Asura_PS3_NullSound_Streaming Asura_PS3_NullSound::s_xStreamingAPI; 
#endif	// ASURA_PS3_USE_NULLSOUND

#ifdef ASURA_ASSERT
static bool IsValidPlayingSoundHandle( int iHandle )
{
	return iHandle >= 0 && iHandle < iASURA_PS3_SOUND_MAX_VOICES;
}
#endif


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveSound::Asura_PS3_ActiveSound
// Description : Active Sound constructor - initialises to known defaults
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_ActiveSound::Asura_PS3_ActiveSound()
{
	m_fVolume = 0.0f;			
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_LoadedSound::Asura_PS3_LoadedSound
// Description : Construct a loaded sound - sets initial default values
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_LoadedSound::Asura_PS3_LoadedSound()
{
	m_pcWaveData = NULL;
	m_bStreamed = false;
	m_bNeedsSoundInfo = false;
	m_iSourceFileIndex = -1;
	m_uFileOffset = 0;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_LoadedSound::~Asura_PS3_LoadedSound
// Description : Destroy loaded sound (frees any unfreed memory)
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_LoadedSound::~Asura_PS3_LoadedSound()
{
	Destroy();
}
	
	
////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_LoadedSound::Create
// Description : Create a loaded sound from the given sample information
//
// Parameters  : 
//   const char* pSamples  - Source sample data
//   const Asura_PS3_Sound_Info& xSoundInfo - Sound information
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_LoadedSound::Create(const char* pSamples, const Asura_PS3_Sound_Info& xSoundInfo)
{
	m_xSoundInfo = xSoundInfo;
	m_bStreamed = false;
	m_bNeedsSoundInfo = false;

	const u_int uTotalDataBytes = xSoundInfo.GetTotalDataBytes();
	m_pcWaveData = static_cast<char*>( Asura_AllocateAlign(uTotalDataBytes, ASURA_ALIGN_16_BYTE) );
	if ( !m_pcWaveData )
	{
		Asura_Assert(false, "Unable to allocate space for wave data");
		return;
	}

	// Update stats
	Asura_PS3_LoadedSound::s_uLoadedSoundsSizeTotal += uTotalDataBytes;
	Asura_PS3_LoadedSound::s_iNumberOfLoadedSounds++;
	
	Asura_Mem::Copy(m_pcWaveData, pSamples, uTotalDataBytes);
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_LoadedSound::CreateStreamed
// Description : Create a streamed 'loaded' sound (this is a little hacky)
//
// Parameters  : 
//   int iSourceFileIndex  - Source file to stream from (see sound system)
//   u_int uOffset         - Offset into source file for sound file
//   u_int uSize           - Size of sound file within the source file
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_LoadedSound::CreateStreamed(int iSourceFileIndex, u_int uOffset, u_int uSize)
{
	m_bStreamed = true;
	m_bNeedsSoundInfo = true;

	m_pcWaveData = NULL;
	m_iSourceFileIndex = iSourceFileIndex;
	m_uFileOffset = uOffset;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_LoadedSound::Destroy
// Description : Destroy loaded sound and reset for future usage
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_LoadedSound::Destroy()
{
	if ( m_pcWaveData != NULL )
	{
		// Update stats
		Asura_PS3_LoadedSound::s_uLoadedSoundsSizeTotal -= m_xSoundInfo.GetTotalDataBytes();
		Asura_PS3_LoadedSound::s_iNumberOfLoadedSounds--;

		Asura_DeallocateAlign(m_pcWaveData);
		m_pcWaveData = NULL;
	}
	m_xSoundInfo = Asura_PS3_Sound_Info();
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::Asura_PS3_Sound_System
// Description : PS3 sound system constructor - initialise to known defaults
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_Sound_System::Asura_PS3_Sound_System()
{
	m_iNextSSIndex = 0;
	m_iNextSSIndexAtReset = 0;
	Asura_Mem::Set(m_aszStreamingSourceFiles, 0, sizeof(m_aszStreamingSourceFiles));
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::Initialise
// Description : Initialise the sound system - allocates sound storage
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::Initialise()
{
	#ifndef ASURA_USE_AUDIO_SYSTEM

	if ( !PARENT::Initialise() )
	{
		return false;
	}

#ifdef ASURA_PS3_EXTRA_ASYNC_THREAD
	// Allocate async playback list
	m_xAsyncList.Create(ASYNC_PLAYBACK_QUEUE_LENGTH);

	// Create a thread to perform the async playback (note: it spends most of its time blocked)
	int iResult = sys_ppu_thread_create(&m_xAsyncThread, &Asura_PS3_Sound_System::Async_Playback_Thread, 
	                                     reinterpret_cast<uint64_t>(this), ASYNC_PPU_THREAD_PRIORITY, ASYNC_THREAD_STACK_SIZE, 
	                                     SYS_PPU_THREAD_CREATE_JOINABLE, "Async Audio Thread");
	if ( iResult != 0 )
	{
		Asura_Assertf( false, ("[Asura_PS3_Sound_System::Initialise] - Async playback thread creation failed - streamed sounds will not play!\nError : %d\n", iResult) );
		return false;
	}
#endif

	#endif //ASURA_USE_AUDIO_SYSTEM
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::Shutdown
// Description : Showdown and clean up the sound system
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::Shutdown()
{
	bool bReturn = true;
	#ifndef ASURA_USE_AUDIO_SYSTEM

#ifdef ASURA_PS3_EXTRA_ASYNC_THREAD
	// Wait for async playback thread to complete (blocks)
	m_xAsyncReady.Signal();               // signal to allow exit
	uint64_t uThreadResult = 0;
	sys_ppu_thread_join( m_xAsyncThread, &uThreadResult );
	
	m_xAsyncReady.Destroy();
	m_xAsyncList.Destroy();
#endif

	// Unload all loaded sounds
	for (int iLoop=0; iLoop < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS; iLoop++)
	{
		UnloadSound(iLoop);
	}

	// Reset stats
	Asura_PS3_LoadedSound::s_iNumberOfLoadedSounds = 0;
	Asura_PS3_LoadedSound::s_uLoadedSoundsSizeTotal = 0;
	m_iNextSSIndex = 0;
	m_iNextSSIndexAtReset = 0;

	bReturn = PARENT::Shutdown();
	#endif //ASURA_USE_AUDIO_SYSTEM
	return bReturn;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::Reset
// Description : Unload all the sounds
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Sound_System::Reset()
{
	#ifndef ASURA_USE_AUDIO_SYSTEM
	// Stop sounds playing
	for (int iLoop=0; iLoop < GetMaxConcurrentSounds(); iLoop++ )
	{	
		Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iLoop];
		
		if ( xActiveSound.IsValid() && xActiveSound.IsPlaying() )
		{
			xActiveSound.Stop(0);
		}
	}

	m_iNextSSIndex = m_iNextSSIndexAtReset;

	PARENT::Reset();
	#endif //ASURA_USE_AUDIO_SYSTEM
}

void Asura_PS3_Sound_System::SetResetPoint()
{
	m_iNextSSIndexAtReset = m_iNextSSIndex;
}

void Asura_PS3_Sound_System::Update()
{
	#ifndef ASURA_USE_AUDIO_SYSTEM
	PARENT::Update();
	#endif
}


void Asura_PS3_Sound_System::SetListener( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation )
{
	m_xListenerPosition = xPosition;
	m_xListenerOrientation = xOrientation;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::AddStreamingSoundSourceFile
// Description : Set the current source file for streaming sounds 
//
// Parameters  : 
//   const char* szFilename - Filename to use
//
// Returns     : int - Source file index
///////////////////////////////////////////////////////////////////////////////
int Asura_PS3_Sound_System::AddStreamingSoundSourceFile(const char* szFilename)
{
	if ( m_iNextSSIndex < ASURA_PS3_MAX_STREAMING_FILES )
	{
		strncpy( m_aszStreamingSourceFiles[m_iNextSSIndex], szFilename, ASURA_PS3_MAX_PATH);
		m_iNextSSIndex++;
		return (m_iNextSSIndex-1);
	}
	else
	{
		return -1;
	}
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::AddStreamingSoundData
// Description : Add information about streaming sounds
//
// Parameters  : 
//   int iStreamingFileIndex  - Source file index
//   int iSoundID             - ID of streaming sound 
//   u_int uOffset            - Offset into source file
//   u_int uSize              - Total size of sound (in bytes)
//   Asura_Hash_ID uSoundHash - Sound hash
//
// Returns     : bool - returns true on succes, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::AddStreamingSoundData(int iStreamingFileIndex, int iSoundID, u_int uOffset, u_int uSize, Asura_Hash_ID uSoundHash)
{
	// Flag sound as loaded in Asura_Sound_System and get loaded index.
    int iLoadIndex = Asura_Sound_System::SoundLoadedByPlatformCode(iSoundID, uSoundHash);
    if (iLoadIndex < 0)
    {
		Asura_Assert(false, "[Asura_PS3_Sound_System::AddStreamingSoundData] - Failed to add streamed sound to list of loaded sounds\n");
        return false;
    }

	// Check load index
	Asura_Assert( iLoadIndex >= 0 && iLoadIndex < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, szINVALID_LOAD_INDEX );
    if ( iLoadIndex < 0 || iLoadIndex >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS )
	{
		return false;
	}

	Asura_PS3_LoadedSound& xSound = Asura_PS3_LoadedSound::s_axLoadedSounds[iLoadIndex];
	xSound.CreateStreamed(iStreamingFileIndex, uOffset, uSize);

	return true;
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::LoadSound
// Description : Load sounds from disk - Not currently implemented
//
// Parameters  : 
//   const char* szFilename  - Source filename
//   int iLoadIndex          - Index for loaded sound object
//
// Returns     : bool - Returns true on success 
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::LoadSound(const char* szFilename, int iLoadIndex)
{
	Asura_Assert( iLoadIndex >= 0 && iLoadIndex < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, szINVALID_LOAD_INDEX );
    if ( iLoadIndex < 0 || iLoadIndex >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS )
	{
		return false;
	}
	
	// s_axLoadedSounds[iLoadIndex].Clear();

	Asura_Assert(false, "PS3 does not expect to load sounds direct from disk (yet)");
	return false;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::LoadSoundFromResource
// Description : Load a sound from the given source
//
// Parameters  : 
//   Asura_Handle iFileHandle  - File handle (0=resource stream)
//   int iFileSize             - Size of file within resource
//   int iLoadIndex            - Index for loaded sound object
//
// Returns     : bool - Returns true on success 
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::LoadSoundFromResource( Asura_Handle iFileHandle, int iFileSize, int iLoadIndex )
{
	Asura_Assert( iLoadIndex >= 0 && iLoadIndex < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, szINVALID_LOAD_INDEX );
    if ( iLoadIndex < 0 || iLoadIndex >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS )
	{
		return false;
	}

	Asura_HeapArray<char> xTempBuffer;
	xTempBuffer.SetSize(iFileSize);
	char* pcTempBuffer = &xTempBuffer[0];
    
    if ( !pcTempBuffer )
    {
        Asura_Assert(0, "[Asura_PS3_Sound_System::LoadSoundFromResource] - Failed to create temporary load buffer\n");
        return false;
    }

    if ( !Asura_FileAccess::ReadData(iFileHandle, pcTempBuffer, iFileSize) )
	{
		Asura_Assert(false, "[Asura_PS3_Sound_System::LoadSoundFromResource] - Failed to read sound data\n");
		return false;
	}

	// Temporary information for sample load
	Asura_PS3_Sound_Info xSoundInfo;
	char*   pSamples  = NULL;

	// Load sound from temporary buffer
	if ( !Asura_PS3_Sound_Loader::LoadAuto(&pcTempBuffer[0], iFileSize, pSamples, xSoundInfo) )
	{
		Asura_Assert(false, "[Asura_PS3_Sound_System::LoadSoundFromResource] - Failed to load sound\n");	
		return false;
	}

	// Initialise the sound at the given loadindex
	Asura_PS3_LoadedSound::s_axLoadedSounds[iLoadIndex].Create(pSamples, xSoundInfo);
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::UnloadSound
// Description : Unload a sound (gets the 'loaded' sound to free its memory)
//
// Parameters  : 
//   int iLoadedSoundID - Index for loaded sound
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::UnloadSound(int iLoadedSoundID)
{
	Asura_Assert( iLoadedSoundID >= 0 && iLoadedSoundID < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, szINVALID_LOAD_INDEX );

	if ( Asura_PS3_LoadedSound::s_axLoadedSounds[iLoadedSoundID].IsLoaded() )
	{
		Asura_PS3_LoadedSound::s_axLoadedSounds[iLoadedSoundID].Destroy();
	}
	
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::CreateSound
// Description : Create an 'active' sound suitable for playback 
//
// Parameters  : 
//   int iLoadedSoundID                - Index for source loaded sound
//   Asura_Handle iPlayingSoundHandle  - Playback handle from Asura_Sound_System
//   bool bUseEQ                       - Set true to use the EQ on this sound
//   bool bUseHardware3dPan            - Set true to use Hardware 3D panning
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Sound_System::CreateSound(int iLoadedSoundID, Asura_Handle iPlayingSoundHandle, bool bUseEQ, bool bUseHardware3dPan)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	
	Asura_Assert( iLoadedSoundID >= 0 && iLoadedSoundID < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, szINVALID_LOAD_INDEX );
	Asura_Assert( Asura_PS3_LoadedSound::s_axLoadedSounds[iLoadedSoundID].IsLoaded(), "Sound not loaded" );

	// Create and inititalise an 'active' sound
	return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].Create(iLoadedSoundID, iPlayingSoundHandle, bUseEQ, bUseHardware3dPan);
}


bool Asura_PS3_Sound_System::DestroySound(Asura_Handle iPlayingSoundHandle)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// stop if necessary & cleanup
	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];
	if ( xActiveSound.IsPlaying() || xActiveSound.IsPaused() )
	{
		xActiveSound.Stop(0);
	}

	xActiveSound.Destroy();
	return true;
}


bool Asura_PS3_Sound_System::PlaySound(Asura_Handle iPlayingSoundHandle, bool bLooping)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];

	// Start sound playing (use different method depending on streaming nature)
	if ( xActiveSound.IsStreamed() )
	{
		xActiveSound.SetLooping(bLooping);
		xActiveSound.SetPending(true); // This prevents it being stopped by the sound system before it's started streaming
		xActiveSound.SetSoundCanBeDeleted(false);
		Asura_PS3_MultiStream_Streaming::AddToLoadQueue( PlayStreamedSound, &xActiveSound, sizeof(xActiveSound));
		return true;
	}
	else // Normal immediate (non-streamed) playback
	{
		return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].Play( bLooping );
	}

	return false;
}


bool Asura_PS3_Sound_System::IsSoundPlaying(Asura_Handle iPlayingSoundHandle)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// check if sound is playing (includes paused-but-playing)
	return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].IsPlaying();
}


bool Asura_PS3_Sound_System::IsSoundPreparing(Asura_Handle iPlayingSoundHandle)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);

	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];

	return (xActiveSound.IsStreamed() && xActiveSound.m_bPending);
}


bool Asura_PS3_Sound_System::StopSound(Asura_Handle iPlayingSoundHandle)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// stop sound playing
	return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].Stop(0);
}


bool Asura_PS3_Sound_System::PauseSound(Asura_Handle iPlayingSoundHandle)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// suspend sound
	return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].Pause();
}


bool Asura_PS3_Sound_System::UnpauseSound(Asura_Handle iPlayingSoundHandle)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// resume sound
	return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].Unpause();
}


bool Asura_PS3_Sound_System::SetVolume( Asura_Handle iPlayingSoundHandle, float fVolume )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// set volume
	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetVolume(fVolume);
	return true;
}



bool Asura_PS3_Sound_System::SetWetDryVolume( Asura_Handle iPlayingSoundHandle, float fWetVolume, float fDryVolume)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetWetDryVolume(fWetVolume, fDryVolume);
	return true;
}


bool Asura_PS3_Sound_System::SetPan( Asura_Handle iPlayingSoundHandle, float fPan )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// set panning
	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetPan(fPan);
	return true;
}


bool Asura_PS3_Sound_System::SetPosition(Asura_Handle iPlayingSoundHandle, const Asura_Vector_3& xPosition)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetPosition(xPosition);
	return true;
}


bool Asura_PS3_Sound_System::SetSpatial( Asura_Handle iPlayingSoundHandle, bool bIsSpatial )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// set spatial settings
	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetSpatial(bIsSpatial);
	return true;
}


bool Asura_PS3_Sound_System::SetPitch( Asura_Handle iPlayingSoundHandle, float fPitch )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// set playback pitch
	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetPitch(fPitch);
	return true;
}


bool Asura_PS3_Sound_System::SetUseReverb( Asura_Handle iPlayingSoundHandle, bool bUseReverb )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// set reverb usage
	return Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetUseReverb(bUseReverb);
}


bool Asura_PS3_Sound_System::SetUseDistanceFilter( Asura_Handle iPlayingSoundHandle, bool bUseDistanceFilter )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// set distance filter usage
	Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle].SetUseDistanceFilter(bUseDistanceFilter);
	return true;
}

void Asura_PS3_Sound_System::SetOcclusionLPF(Asura_Handle iPlayingSoundHandle, float fOcclusion )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);

	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];
	xActiveSound.UpdateOcclusionFilter(fOcclusion);
}

bool Asura_PS3_Sound_System::SetVolumePanAndFade(Asura_Handle iPlayingSoundHandle, float fVolume, float fPan, float fFade)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// Set volume, panning and fade
	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];

	xActiveSound.SetVolume(fVolume);
	xActiveSound.SetPan(fPan);
	xActiveSound.SetFade(fFade);
	return true;
}


bool Asura_PS3_Sound_System::SetPanAndFade( Asura_Handle iPlayingSoundHandle, float fPan, float fFade )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// Set panning and fade
	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];

	xActiveSound.SetPan(fPan);
	xActiveSound.SetFade(fFade);
	return true;
}


void Asura_PS3_Sound_System::SetEQType(Asura_Handle iPlayingSoundHandle, Asura_Sound_System::EQType eType, float fLevel)
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// Set EQ for the playing sound
	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];
	xActiveSound.SetEQType(eType);
}


bool Asura_PS3_Sound_System::SetRadii(Asura_Handle iPlayingSoundHandle, float fInner, float fOuter )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// Set Range radii for the playing sound
	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];
	xActiveSound.SetRadii(fInner, fOuter);
	return true;
}

void Asura_PS3_Sound_System::UpdateDistanceFilter(Asura_Handle iPlayingSoundHandle, float fDistance )
{
	Asura_Assert(IsValidPlayingSoundHandle(iPlayingSoundHandle), szINVALID_SOUND_HANDLE);	

	// Set Range radii for the playing sound
	Asura_PS3_ActiveSound& xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iPlayingSoundHandle];
	xActiveSound.UpdateDistanceFilter(fDistance);
}

#ifdef ASURA_PS3_EXTRA_ASYNC_THREAD
////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Sound_System::Async_Playback_Thread
// Description : Asynchronous playback thread
//
// Parameters  : 
//   uint64_t param - Object's 'this' pointer
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Sound_System::Async_Playback_Thread(uint64_t param)
{
	Asura_OutputToDebugger(("[Asura_PS3_Sound_System::Async_Playback_Thread] - Async playback thread started...\n"));

	Asura_PS3_Sound_System* pThis = reinterpret_cast<Asura_PS3_Sound_System*>(param);
	if ( pThis != NULL )
	{
		// Wait infinitely on the semaphore (exits when the semaphore is destroyed in shutdown)
		while ( pThis->m_xAsyncReady.Wait(0) )
		{
			int* piNext = pThis->m_xAsyncList.Peek(0);
			if ( piNext != NULL )
			{
				int iIndex = *piNext;
				pThis->m_xAsyncList.Drop();
				if ( iIndex >= 0 )
				{
					Asura_PS3_ActiveSound::s_axActiveSounds[iIndex & ~PACKED_LOOPING_BIT].Play( (iIndex & PACKED_LOOPING_BIT) != 0 );
					Asura_PS3_ActiveSound::s_axActiveSounds[iIndex & ~PACKED_LOOPING_BIT].SetSoundCanBeDeleted(true);
				}
				else
				{
					Asura_OutputToDebugger(("[Asura_PS3_Sound_System::Async_Playback_Thread] - Requested exit...\n"));
					break;	// Requested exit - add a negative index.
				}
			}
			else
			{
				Asura_OutputToDebugger(("[Asura_PS3_Sound_System::Async_Playback_Thread] - Semaphore signalled but nothing in list, assumed exit request...\n"));
				break;
			}
		}
	}

	Asura_OutputToDebugger(("[Asura_PS3_Sound_System::Async_Playback_Thread] - Async playback thread exited...\n"));
	sys_ppu_thread_exit(0);
}
#endif

void Asura_PS3_Sound_System::PlayStreamedSound( void* pData, u_int uSize )
{
	Asura_Assert(uSize == sizeof(Asura_PS3_ActiveSound), "wrong data size");
	Asura_PS3_ActiveSound* pxActiveSound = reinterpret_cast<Asura_PS3_ActiveSound*>(pData);
	pxActiveSound->Play(pxActiveSound->m_bLooping);
	pxActiveSound->SetSoundCanBeDeleted(true);
}

#endif // #ifndef ASURA_USE_AUDIO_SYSTEM

//eof
