///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_Sound_System.h
// Author    : Philip Hall
// Created   : 11/09/007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////


#ifndef __ASURA_PS3_SOUNDSYSTEM_H__
#define __ASURA_PS3_SOUNDSYSTEM_H__

#ifndef ASURA_USE_AUDIO_SYSTEM

// Asura headers
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_RingBuffer.h"
#include "Sound/Asura_Sound.h"

// Asura PS3 headers
#include "Core/Asura_PS3_Semaphore.h"
#include "Sound/Asura_PS3_Sound_Common.h"
#include "Sound/Asura_PS3_Sound_Loader.h"

// Voices do not necessarily equate to sub-buses
static const int iASURA_PS3_SOUND_MAX_VOICES = 64;
static const int iASURA_PS3_SOUND_MAX_VOICE_CHANNELS = 64 + 24; // allow 24 stereo sounds
static const int ASURA_PS3_SOUND_MAX_SUB_BUSES = 16;

// HACK: Temporary selection of sound system - move these elsewhere!!!
#define ASURA_PS3_USE_MULTISTREAM (1)
//#define ASURA_PS3_USE_NULLSOUND (1)

// Select sound system implementation
#if defined (ASURA_PS3_USE_MULTISTREAM) && (ASURA_PS3_USE_MULTISTREAM != 0) 
// MULTISTREAM
	#include "Sound/Asura_PS3_MultiStream.h"
	typedef class Asura_PS3_MultiStream				Asura_PS3_Sound_API; 
	typedef class Asura_PS3_ActiveMSSound           Asura_PS3_ActiveSound_Base;

#elif defined (ASURA_PS3_USE_NULLSOUND) && (ASURA_PS3_USE_NULLSOUND != 0) 
// NULL SOUND
	#include "Sound/Asura_PS3_NullSound.h"
	typedef class Asura_PS3_NullSound				Asura_PS3_Sound_API; 
	typedef class Asura_PS3_ActiveNullSound			Asura_PS3_ActiveSound_Base;
#else
// Default to NULL SOUND
	#include "Sound/Asura_PS3_NullSound.h"
	typedef class Asura_PS3_NullSound				Asura_PS3_Sound_API; 
	typedef class Asura_PS3_ActiveNullSound			Asura_PS3_ActiveSound_Base;
#endif

static const int ASURA_PS3_MAX_STREAMING_FILES = 4;	 // Maximum number of streaming source files (that's files NOT streams)


////PGH/////////////////////////////////////////
// PS3 loaded sound
////////////////////////////////////////////////
class Asura_PS3_LoadedSound
{
public:
	Asura_PS3_LoadedSound();
	~Asura_PS3_LoadedSound();
	
	void	Create(const char* pSamples, const Asura_PS3_Sound_Info& xSoundInfo);
	void    CreateStreamed(int iSourceFileIndex, u_int uOffset, u_int uSize);
	void	Destroy();

	inline bool  IsLoaded() const			{ return (m_pcWaveData != NULL) || m_bStreamed; }
	inline bool  IsStreamed() const			{ return m_bStreamed; }
	inline bool  NeedsSoundInfo() const     { return m_bNeedsSoundInfo; }
	inline int   GetSampleRate() const		{ return m_xSoundInfo.GetSampleRate(); }
	inline int   GetNumSamples() const		{ return m_xSoundInfo.GetTotalDataBytes() / (m_xSoundInfo.GetSampleBits() >> 3); } 
	inline char* GetSampleData() const		{ return m_pcWaveData; }
	inline u_int GetSampleDataSize() const	{ return m_xSoundInfo.GetTotalDataBytes(); }
	inline int	 GetNumChannels() const		{ return m_xSoundInfo.GetChannelCount(); }
	inline int   GetSourceFileIndex() const { return m_iSourceFileIndex; }
	inline u_int GetFileOffset() const      { return m_uFileOffset; }
	inline const Asura_PS3_Sound_Info& GetSoundInfo() const { return m_xSoundInfo; }
	inline void  SetSoundInfo(const Asura_PS3_Sound_Info& xInfo) { m_xSoundInfo = xInfo; m_bNeedsSoundInfo = false; }

protected:
	char*	             m_pcWaveData;  // Stored waveform samples
	Asura_PS3_Sound_Info m_xSoundInfo;
	bool                 m_bStreamed;
	bool                 m_bNeedsSoundInfo;
	int                  m_iSourceFileIndex;
	u_int                m_uFileOffset;

public:
	// Loaded sound tracking
	static Asura_PS3_LoadedSound	s_axLoadedSounds[ASURA_MAX_NUMBER_OF_LOADED_SOUNDS];
	static int						s_iNumberOfLoadedSounds;
	static u_int					s_uLoadedSoundsSizeTotal;

	ASURA_DECLARE_NO_COPYING(Asura_PS3_LoadedSound);
};



////PGH/////////////////////////////////////////
// Customised PS3 active sound
////////////////////////////////////////////////
class Asura_PS3_ActiveSound : public Asura_PS3_ActiveSound_Base
{
	typedef Asura_PS3_ActiveSound_Base PARENT;
public:
	Asura_PS3_ActiveSound();

public:
	// Active sound tracking (dynamically allocated)
	static Asura_FixedArray<Asura_PS3_ActiveSound, iASURA_PS3_SOUND_MAX_VOICES> s_axActiveSounds;

	ASURA_DECLARE_NO_COPYING(Asura_PS3_ActiveSound);
};



////PGH/////////////////////////////////////////
// PS3 sound system root
////////////////////////////////////////////////
class Asura_PS3_Sound_System : public Asura_PS3_Sound_API
{
	typedef Asura_PS3_Sound_API PARENT;

public:
	Asura_PS3_Sound_System();

	// Initialisation
	bool Initialise();
	bool Shutdown();
	void Reset();
	void SetResetPoint();
	void Update();

	void SetListener( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation );
	int  AddStreamingSoundSourceFile(const char* szFilename);
	bool AddStreamingSoundData(int iStreamingFileIndex, int iSoundID, u_int uOffset, u_int uSize, Asura_Hash_ID uSoundHash);

	// Loading
	bool LoadSound(const char* szFilename, int iLoadIndex);
	bool LoadSoundFromResource(Asura_Handle iFileHandle, int iFileSize, int iLoadIndex);
	bool UnloadSound(int iLoadedIndex);

	bool CreateSound(int iLoadedSoundID, Asura_Handle iPlayingSoundHandle, bool bUseEQ, bool bUseHardware3dPan);
	bool DestroySound(Asura_Handle iPlayingSoundHandle);

	// Playback control
	bool PlaySound(Asura_Handle iPlayingSoundHandle, bool bLooping);
	bool IsSoundPlaying(Asura_Handle iPlayingSoundHandle);
	bool IsSoundPreparing(Asura_Handle iPlayingSoundHandle);
	bool StopSound(Asura_Handle iPlayingSoundHandle);
	bool PauseSound(Asura_Handle iPlayingSoundHandle);
	bool UnpauseSound(Asura_Handle iPlayingSoundHandle);

	// Aural control
	bool SetPosition( Asura_Handle iPlayingSoundHandle, const Asura_Vector_3& xPosition );
	bool SetVolume( Asura_Handle iPlayingSoundHandle, float fVolume );
	bool SetWetDryVolume( Asura_Handle iPlayingSoundHandle, float fWetVolume, float fDryVolume);
	bool SetPan( Asura_Handle iPlayingSoundHandle, float fPan );
	bool SetSpatial( Asura_Handle iPlayingSoundHandle, bool bIsSpatial );
	bool SetPitch( Asura_Handle iPlayingSoundHandle, float fPitch );
	bool SetUseReverb( Asura_Handle iPlayingSoundHandle, bool bUseReverb );
	bool SetUseDistanceFilter( Asura_Handle iPlayingSoundHandle, bool bUseReverb );
	bool SetVolumePanAndFade(Asura_Handle iPlayingSoundHandle, float fVolume, float fPan, float fFade);
	bool SetPanAndFade( Asura_Handle iPlayingSoundHandle, float fPan, float fFade );
	void SetEQType(Asura_Handle iPlayingSoundHandle, Asura_Sound_System::EQType eType, float fLevel);
	bool SetRadii(Asura_Handle iPlayingSoundHandle, float fInner, float fOuter );
	void UpdateDistanceFilter(Asura_Handle iPlayingSoundHandle, float fDistance );
	void SetOcclusionLPF(Asura_Handle iPlayingSoundHandle, float fOcclusion );

	// General Accessors
	const Asura_Vector_3&   GetListener() const    { return m_xListenerPosition; }
	const Asura_Matrix_3x3& GetOrientation() const { return m_xListenerOrientation; }
	const char *GetStreamingFilename(int iIndex) const { return m_aszStreamingSourceFiles[iIndex]; }

	// Get reference to internally selected API
	static Asura_PS3_Sound_System& GetAPI() { return s_xAPI; }

private:
	// Static sound-system access point	(access via GetAPI)
	static Asura_PS3_Sound_System	s_xAPI;

	// async thread function
	static void PlayStreamedSound( void* pData, u_int uSize );

#ifdef ASURA_PS3_EXTRA_ASYNC_THREAD
	// Async playback
	Asura_PS3_Semaphore     m_xAsyncReady;
	Asura_RingBuffer<int>   m_xAsyncList;
	sys_ppu_thread_t        m_xAsyncThread;
	static void Async_Playback_Thread(uint64_t param);
#endif

	// Listener orientation
	Asura_Vector_3 	        m_xListenerPosition;	// Listener position
	Asura_Matrix_3x3        m_xListenerOrientation; // Listener orientation

	// Streaming sound source files
	int   m_iNextSSIndex;
	int   m_iNextSSIndexAtReset;
	char  m_aszStreamingSourceFiles[ASURA_PS3_MAX_STREAMING_FILES][ASURA_PS3_MAX_PATH];

	ASURA_DECLARE_NO_COPYING(Asura_PS3_Sound_System);
};


#endif // #ifndef ASURA_USE_AUDIO_SYSTEM

#endif
