///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_GameData.cpp
// Author    : Philip Hall
// Created   : 14/12/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "Utility/Asura_PS3_GameData.h"

// Asura headers
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Asura PS3 headers
#include "FileAccess/Asura_PS3_FileSystem.h" 

#include <sys/ppu_thread.h>
#include <sys/timer.h>
#include <sys/memory.h>
#include <sysutil/sysutil_sysparam.h>
#include <sysutil/sysutil_common.h>
#include <sysutil/sysutil_gamedata.h>


////PGH/////////////////////////////////////////
// Useful constants
////////////////////////////////////////////////
const int    GAMEDATA_PPU_PRIORITY   = 750;
const size_t GAMEDATA_PPU_STACK_SIZE = (1024 * 16);
const size_t GAMEDATA_MEMORY_CONTAINER_SIZE = (1024 * 1024 * 3);
static Asura_PS3_GameData* g_pThis = NULL;


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_GameData::Asura_PS3_GameData
// Description : Initialise the PS3 game-data wrapper object
//
// Parameters  : 
//   const char* szProductCode   - Product Code
//   bool bAutoDelete            - Set true to delete self when complete
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_GameData::Asura_PS3_GameData(const char* szProductCode, bool bAutoDelete)
{
	m_bRecreateData = false;
	m_bEnoughSpace = false;
	m_iSpaceRequiredKB = 0;
	Asura_PrintfToBfr(m_szProductCode, "%s", CELL_GAMEDATA_DIRNAME_SIZE, true, szProductCode);
	m_szDataPath[0] = 0;
	m_xThreadID = 0;
	m_pfnCheckData    = NULL;
	m_pfnRecreateData = NULL;
	m_bAutoDeleteThis = bAutoDelete;
}


Asura_PS3_GameData::~Asura_PS3_GameData()
{
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_GameData::Begin
// Description : Start the data check
//
// Parameters  : 
//   Asura_PS3_DataCheckCallback pfnCheck   - Data check callback function
//   Asura_PS3_DataCreateCallback pfnCreate - Data creation callback function
//
// Returns     : bool - Returns true if successful
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_GameData::Begin(Asura_PS3_DataCheckCallback pfnCheck, Asura_PS3_DataCreateCallback pfnCreate)
{
	m_pfnCheckData    = pfnCheck;
	m_pfnRecreateData = pfnCreate;
	
	// Launch the thread
	int iResult = sys_ppu_thread_create(&m_xThreadID, &Asura_PS3_GameData::ThreadGameData, 
	                                     reinterpret_cast<uint64_t>(this), GAMEDATA_PPU_PRIORITY, 
	                                     GAMEDATA_PPU_STACK_SIZE, 0, "Storage_Data");

	Asura_Assert(iResult >= 0, "[Asura_PS3_GameData::BeginCheckData] - Unable to create game data storage thread\n");
	return (iResult == CELL_OK);
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_GameData::WaitComplete
// Description : Wait or poll for the game data thread to complete
//
// Parameters  :
//   int iTimeoutMS - Duration to wait in milliseconds (zero or less to poll)
//
// Returns     : bool - Returns true when completed, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_GameData::WaitComplete(int iTimeoutMS)
{
	bool bComplete = ( iTimeoutMS <= 0 ) ? m_xCompleted.TryWait() : m_xCompleted.Wait( iTimeoutMS*1000 );

	if ( bComplete )
	{
		uint64_t uStat;
		sys_ppu_thread_join(m_xThreadID, &uStat);  // join the worker thread back
		m_xThreadID = 0;
	}

	cellSysutilCheckCallback();	 // This is a teeny-tiny-hack
	return bComplete;
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_GameData::ThreadGameData
// Description : Game data handling thread
//
// Parameters  : 
//   uint64_t uArg - Pointer to this object
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GameData::ThreadGameData(uint64_t uArg)
{
	Asura_PS3_GameData* pThis = reinterpret_cast<Asura_PS3_GameData*>(uArg);

	sys_memory_container_t xMemoryContainer;
	if ( sys_memory_container_create( &xMemoryContainer, GAMEDATA_MEMORY_CONTAINER_SIZE) != 0 )
	{
		Asura_Assert( false, "[Asura_PS3_GameData::ThreadGameData] - Unable to allocate system memory container\n" );
	}
	else
	{
		Asura_Assert(g_pThis == NULL, "[Asura_PS3_GameData::ThreadGameData] - Possibly in use multiple times\n");

		g_pThis = pThis;  // Unlike the save utility, there's no user parameter
		int iResult = cellGameDataCheckCreate2( CELL_GAMEDATA_VERSION_CURRENT,
												pThis->m_szProductCode,
												CELL_GAMEDATA_ERRDIALOG_ALWAYS,
												&Asura_PS3_GameData::GameDataStatusCallback,
												xMemoryContainer);
		g_pThis = NULL;

		if (iResult == CELL_GAMEDATA_RET_OK)
		{
			if ((pThis->m_bRecreateData) && (pThis->m_pfnRecreateData != NULL))
			{
				pThis->m_pfnRecreateData(pThis);
			}
		}
		else if ( iResult == static_cast<int>(CELL_GAMEDATA_ERROR_BROKEN) )
		{
			cellGameDataExitBroken();
		}

		iResult = sys_memory_container_destroy( xMemoryContainer );
		if ( iResult < 0 )
		{
			Asura_Assert(false, "[Asura_PS3_SingleStorage::ThreadStorage] - Potential leak in memory container\n");
		}
	}
	pThis->m_xCompleted.Signal();
	if ( pThis->m_bAutoDeleteThis )
	{
		sys_timer_usleep(5000);	     // Short wait to allow semaphore waits to release (we can wait as long as we like really).
		delete pThis;
	}
	sys_ppu_thread_exit(0);
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_GameData::GameDataStatusCallback
// Description : Game status callback for the gamedata utility
//
// Parameters  : 
//   CellGameDataCBResult* pResult  - Result return
//   CellGameDataStatGet* pGet      - Incoming data
//   CellGameDataStatSet* pSet      - Outgoing data
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GameData::GameDataStatusCallback( CellGameDataCBResult* pResult, CellGameDataStatGet* pGet, CellGameDataStatSet* pSet )
{
	Asura_PS3_GameData* pThis = reinterpret_cast<Asura_PS3_GameData*>(g_pThis);

	// Take a copy of the game data path for future file IO operations
	strncpy(pThis->m_szInfoPath, pGet->contentInfoPath, CELL_GAMEDATA_PATH_MAX);
	strncpy(pThis->m_szDataPath, pGet->gameDataPath, CELL_GAMEDATA_PATH_MAX);

	pResult->result = CELL_GAMEDATA_CBRESULT_OK;
	pSet->setParam = &pGet->getParam;
	pSet->reserved = NULL;	

	pResult->reserved = NULL;

	pThis->m_bEnoughSpace = true;

	// Check data
	if ( pGet->isNewData )
	{
		// New data, always needs creating
		pThis->m_bRecreateData = true;
	}
	else
	{
		if ( pThis->m_pfnCheckData != NULL )
		{
			pThis->m_iSpaceRequiredKB = 0;
			pThis->m_bRecreateData = pThis->m_pfnCheckData(pThis, pThis->m_iSpaceRequiredKB);

			if ( pThis->m_iSpaceRequiredKB >  pGet->hddFreeSizeKB )
			{
				pResult->errNeedSizeKB = pThis->m_iSpaceRequiredKB - pGet->hddFreeSizeKB;
				pResult->result = CELL_GAMEDATA_ERROR_NOSPACE;
				pThis->m_bEnoughSpace = false;
			}
			else
			{
				pResult->result = pThis->m_bRecreateData ? CELL_GAMEDATA_CBRESULT_OK : CELL_GAMEDATA_CBRESULT_OK_CANCEL;
			}
		}
		else
		{
			pThis->m_bRecreateData = true;
		}
	}
}
