////////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PS3_SaveLoad.h
// Author : Neil Young
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////


#ifdef ASURA_PS3_SAVEGAMES
#define MAX_SAVE_GAMES 8


////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_PlatformBasics.h"	// for Asura_UnicodeChar

#include <sysutil/sysutil_savedata.h>

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

class Asura_ProducerConsumerQueue;
class Asura_ProtoSaveBlock;

class Asura_PS3_SaveLoad
{
public:
	//these are called by the engine, project code shouldn't need to ever call them
	static void Initialise();
	static void Shutdown();
	//sets the title for save games - should be product title
	static void SetSaveDataTitle( const Asura_UnicodeChar* wszTitle );
	//set the subtitle - should be quick description of save, so profile name, current level or something
	static void SetSaveDataSubTitle( const Asura_UnicodeChar* wszSubTitle );
	//sets the detail text - longer description, should help to "uniquify" the save, so stuff like play time, cash earned, etc
	static void SetSaveDataDetail( const Asura_UnicodeChar* wszDetail );
	//sets the title for the blank new data slot on the list interface
	static void SetNewDataTitle( const Asura_UnicodeChar* wszTitle );
	//sets the project portion of directory (IE, slot) names - this must be product code once you have one
	static void SetProjectDirName( const char* szProjectDirName );
	//sets the function to call to check loaded data (not save blocks) is valid. should return false if profile is corrupt
	static void SetProfileCheckCallback(bool (*pfnCallback)(const void* pData));
	//adds a blank slot to the list load interface - note that the sony UI for this is wrong, they are aware and working on it.
	//not suitable for retail use until that's fixed
	static void SetBlankSlotOnLoad(bool bBlankSlotOnLoad) { s_bBlankSlotOnLoad = bBlankSlotOnLoad; }
	
	//launches list load interface
	static bool LoadProfile( void* pData, const u_int uSizeOfData, Asura_ProtoSaveBlock* pxSaveBlock = 0 );
	//does an autosave if a valid slot is known, otherwise launches list save interface
	static bool SaveProfile( void* pData, const u_int uSizeOfData, Asura_ProtoSaveBlock* pxSaveBlock = 0 );
	//launches list save interface
	static bool ManualSaveProfile( void* pData, const u_int uSizeOfData, Asura_ProtoSaveBlock* pxSaveBlock = 0 );

	//silently counts valid slots
	static bool CountProfiles();

	//launches delete from all interface
	static bool DeleteProfiles();


	enum SAVELOAD_STATUS
	{
		STATUS_IDLE = 0,
		STATUS_SAVING,
		STATUS_SAVED,
		STATUS_LOADING,
		STATUS_LOADED,
		STATUS_DELETING,
		STATUS_DELETED,
		STATUS_COUNTING,
		STATUS_COUNTED
	};

	static SAVELOAD_STATUS GetStatus() { return s_eStatus; }
	//wrapper for the above to check if class is doing *anything*
	static bool IsBusy();

	//do we have a valid autosave slot? (set to true after successful load or save)
	static bool CanAutosave() { return s_bCanAutosave; }
	//use this to forget autosave slot, so calls to SaveProfile will list save until another autosave is chosen
	static void ClearAutosaveFlag() { s_bCanAutosave = false; }
	//set this to cause an existing save block to be left as is for timesaving, if the calling code is confident it hasn't changed
	//flag persists
	static void SetDontOverwriteSaveBlock(bool bDontOverwriteSaveBlock) { s_bDontOverwriteSaveBlock = bDontOverwriteSaveBlock; }
	
	//did we run out of space on last save? if so, we'll likely need to offer delete/retry flow
	static bool FreeSpaceFailOnLastSave() { return s_bFreeSpaceFailOnLastSave; }
	//flag if any error occured on last operation.
	static bool ErrorOnLastOperation() { return s_bErrorOnLastOperation; }
	//was the last operation canceled by the user
	static bool LastOperationCanceled() { return s_bLastOperationCanceled; }
	//how many save slots were counted by last count operation
	static u_int LastProfileCountResult() { return s_uLastProfileCountResult; }


private:

	enum FILE_INDEX
	{
		FILE_INDEX_ICON0 = 0,
		FILE_INDEX_ICON1,
		FILE_INDEX_PIC1,
		FILE_INDEX_SND0,
		FILE_INDEX_LAST_CIF = FILE_INDEX_SND0,

		FILE_INDEX_PROFILE,
		FILE_INDEX_SAVEBLOCKHEADER,
		FILE_INDEX_SAVEBLOCK,
		FILE_INDEX_TOTAL,

	};

	enum SAVE_OPERATION_STEP
	{
		SAVE_OPERATION_STEP_ICON0 = 0,
		SAVE_OPERATION_STEP_ICON1,
		SAVE_OPERATION_STEP_PIC1,
		SAVE_OPERATION_STEP_SND0,
		SAVE_OPERATION_STEP_PROFILE,
		SAVE_OPERATION_STEP_SAVEBLOCKHEADER,
		SAVE_OPERATION_STEP_SAVEBLOCK,
		SAVE_OPERATION_STEP_END,

		NUM_SAVE_OPERATION_STEPS
	};

	enum LOAD_OPERATION_STEP
	{
		LOAD_OPERATION_STEP_PROFILE = 0,
		LOAD_OPERATION_STEP_SAVEBLOCKHEADER,
		LOAD_OPERATION_STEP_SAVEBLOCK,
		LOAD_OPERATION_STEP_END,

		NUM_LOAD_OPERATION_STEPS
	};

	static char* s_aszFileNames[FILE_INDEX_TOTAL];

	// project specific directory name
	static const u_int uPROJECTDIRNAMELENGTH = 10;// "ABCD12345"
	static char s_szProjectDirName[uPROJECTDIRNAMELENGTH];
	// autosave variables
	static char s_szAutosaveFilename[CELL_SAVEDATA_DIRNAME_SIZE];
	static bool s_bCanAutosave;
	static bool s_bIsAutosave;

	static bool (*s_pfnCallback)(const void* pData);

	static const FILE_INDEX s_aeFileIndexForSaveStep[NUM_SAVE_OPERATION_STEPS];
	static const FILE_INDEX s_aeFileIndexForLoadStep[NUM_LOAD_OPERATION_STEPS];
	static const CellSaveDataFileType s_aeFileType[FILE_INDEX_TOTAL];
	static void* s_apFileBuffers[FILE_INDEX_TOTAL];
	static u_int s_auFileBufferSizes[FILE_INDEX_TOTAL];
	static bool s_abExistingFiles[FILE_INDEX_TOTAL];
	static void* s_pNewIconBuffer;
	static u_int s_uNewIconBufferSize;

	static u_int s_uSizeOfSaveBlock;
	static Asura_ProtoSaveBlock* s_pxSaveBlock;

	static bool s_bBlankSlotOnLoad;
	static bool s_bDontOverwriteSaveBlock;

	static bool s_bFreeSpaceFailOnLastSave;
	static bool s_bErrorOnLastOperation;
	static bool s_bLastOperationCanceled;

	static u_int s_uLastProfileCountResult;

	static char s_szNewDataTitle[CELL_SAVEDATA_SYSP_TITLE_SIZE];
	static CellSaveDataListNewData s_xNewData;
	static CellSaveDataNewDataIcon s_xNewDataIcon;
	static char s_szNewDataDirName[CELL_SAVEDATA_DIRNAME_SIZE];

	//what stage we have reached in saving and loading the file
	static u_int s_uSaveStep;
	static u_int s_uLoadStep;

	// status
	static SAVELOAD_STATUS s_eStatus;

	// total number of our files in a save slot
	const static u_int s_uMaxFiles;
	//buffer for sony list data
	static void* s_pListBuffer;
	static u_int s_uListBufferSize;

	// parameters for save file
	static char s_szTitle[CELL_SAVEDATA_SYSP_TITLE_SIZE];
	static char s_szSubTitle[CELL_SAVEDATA_SYSP_SUBTITLE_SIZE];
	static char s_szDetail[CELL_SAVEDATA_SYSP_DETAIL_SIZE];

	// secureFileId used for encrypting and decrypting secure files
	static const char* s_szSecureFileId;

	static u_int s_uNumCallBacksForBar;//for progress bar maths

	static u_int SizeInK(u_int uBytes);
	static u_int ProgressPercent();
	static void SetupBlankSlot(CellSaveDataListGet* pxListGet, CellSaveDataListSet* pxListSet);
	static bool LoadCIFiles();
	static void SetupNewDataIcon();
	static void UnloadCIFiles();
	static void UnloadNewDataIcon();
	static int CalculateSizeNewData(CellSaveDataStatGet* pxStatGet);
	static int CalculateSizeOverwrite(CellSaveDataStatGet* pxStatGet);

	static void SaveDataListCallback( CellSaveDataCBResult* pxResult, CellSaveDataListGet* pxListGet, CellSaveDataListSet* pxListSet );
	static void SaveDataStatusCallback( CellSaveDataCBResult* pxResult, CellSaveDataStatGet* pxStatGet, CellSaveDataStatSet* pxStatSet );
	static void SaveDataFileCallback( CellSaveDataCBResult* pxResult, CellSaveDataFileGet* pxFileGet, CellSaveDataFileSet* pxFileSet );
	static void AutoSaveDataFixedCallback( CellSaveDataCBResult* pxResult, CellSaveDataListGet* pxListGet, CellSaveDataFixedSet* pxFixedSet );
	static void LoadDataListCallback( CellSaveDataCBResult* pxResult, CellSaveDataListGet* pxListGet, CellSaveDataListSet* pxListSet );
	static void LoadDataStatusCallback( CellSaveDataCBResult* pxResult, CellSaveDataStatGet* pxStatGet, CellSaveDataStatSet* pxStatSet );
	static void LoadDataFileCallback( CellSaveDataCBResult* pxResult, CellSaveDataFileGet* pxFileGet, CellSaveDataFileSet* pxFileSet );
	static void CountDataFixedCallback( CellSaveDataCBResult* pxResult, CellSaveDataListGet* pxListGet, CellSaveDataFixedSet* pxFixedSet );

	static Asura_ProducerConsumerQueue* s_pxProducerConsumerQueue;
	static void Thread_Save( void* pData, u_int uSize );
	static void Thread_Load( void* pData, u_int uSize );
	static void Thread_Autosave( void* pData, u_int uSize );

	static void Thread_CountProfiles( void* pData, u_int uSize );

	static void Thread_DeleteProfiles( void* pData, u_int uSize );
};
#endif // #ifdef ASURA_PS3_SAVEGAMES
