
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_Skybox.cpp
// Author : Andy Tate
// Date   : 29-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Asura independent header-files...
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Fog.h"
#include "SpecialFX/Asura_Skybox.h"
#include "Rendering/Asura_Viewport.h"

// Asura PS3 header-files...
#include "ChunkSystem/asura_chunk_resourcefile_ps3.h"
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Rendering/Asura_PS3_DeferredRendering.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_RenderingOptions.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"

//Shaders definition
#include "Rendering/ShadersDefinition/VP/SFX_SkyBox_2D_VP.h"
#include "Rendering/ShadersDefinition/VD/SFX_SkyBox_2D_VD.h"
#include "Rendering/ShadersDefinition/FP/SFX_Skybox_FP.h"
#include "Rendering/ShadersDefinition/FP/ClearAlpha_FP.h"

// Game header-files...

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#define MAX_SKYBOX_VERTICES						8
#define MAX_SKYBOX_FACES						6
#define MAX_SKYBOX_VERTICES_PER_FACE			4

////////////////////////////////////////////////////////////////////////////////
// Structure declarations...
////////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct SkyboxVertex
{
	float x, y; //screen coords
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const u_int uNUM_SKYBOX_VERTICES = 4;

const SkyboxVertex axSKYBOX_VERTEX[ uNUM_SKYBOX_VERTICES ] =
{
	{ -1.0f, -1.0f },
	{  1.0f, -1.0f },
	{  1.0f,  1.0f },
	{ -1.0f,  1.0f }
};

// Vertex-buffer data...

static Asura_PS3_GCM_VertexBuffer		g_xSkyboxVB;

static Asura_PS3_FragmentShaderCopy*			g_pxFragmentShader;
static Asura_PS3_GCM_VertexShader*				g_pxVertexShader;

#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
static bool ReloadSkyboxShaders( bool bHelp )
{
	if( bHelp )
	{
		Asura_CommandConsole::Printf("Reload Skybox Shader");
		return true;
	}

	static const char* aszArgs[] = {
		"-mcgb",
		"-ifcvt none",
		"-texformat d FLOAT_RGBA16",
		"-fuse-nrmh",
		"-nobcolor",
		"-fno-max-psize",
		"-po OutColorPrec=fp16",
		"-fastmath",
		"-fastprecision",
		"-inline all",
		"-unroll all",
		"-nofloatbindings",
		"-DUSE_NEW_SHADER",
		NULL
	};

	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ExportTime/SFX_Skybox_FP.ps3FX",
		ASURA_SHADER_HASH_SKYBOX, ASURA_PS3_SHADER_TYPE_HASH_SKYBOX, aszArgs, NULL, new Asura_PS3_Skybox_Handles() );

	return true;
}
#endif //#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)

void Asura_Skybox::Platform_Initialise (void)
{
#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
	Asura_CommandConsole::AddCmd(L"Render.ReloadShaders.Skybox", ReloadSkyboxShaders, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif
	g_pxVertexShader	= new Asura_PS3_VP_Skybox_2D();

	g_pxVertexShader->InitVertexBufferDeclaration( g_axVertexDeclaration, 1 );

	g_xSkyboxVB.Initialise( sizeof( SkyboxVertex ), uNUM_SKYBOX_VERTICES, CELL_GCM_LOCATION_LOCAL );
	SkyboxVertex* const pxVertex = static_cast< SkyboxVertex* >( g_xSkyboxVB.LockForWriting() );
	Asura_Assert( pxVertex, "Unable to lock skybox vertex buffer." );
	Asura_Mem::Copy( pxVertex, axSKYBOX_VERTEX, sizeof( axSKYBOX_VERTEX ) );
	g_xSkyboxVB.Unlock();
}


//
// Platform_Render
//
void Asura_Skybox::Platform_Render (void)
{
	if( !Asura_PS3_RenderingOptions::GetOption( Asura_PS3_RenderingOptions::OPTION_DRAW_SKYBOX ) )
	{
		return;
	}

	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__SKYBOX, "Asura_Skybox::Platform_Render" );


//	Asura_Assert( GetUseNewSkyboxShader(), "Old-style sky-boxes no longer supported" );


	Asura_PS3_DeferredRendering::BindAsSingleRT();

	// render states
 	Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
	Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_NONE );
	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
	Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0123_ALL );
	Asura_PS3_RenderingStates::SetAlphaBlending(false);
	Asura_PS3_RenderingStates::SetStencilTestEnabled( false );


	// setup FP
	g_pxFragmentShader = Asura_PS3_GCM_ShadersLibrary::GetResettableFP( ASURA_SHADER_HASH_SKYBOX, ASURA_PS3_SHADER_TYPE_HASH_SKYBOX );
	if( g_pxFragmentShader == NULL )
	{
		Asura_Assert( false, "Skybox shader not found" );
	    ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__SKYBOX );
		return;
	}

	float fDistance = Asura_Camera::GetFarClip();
 	g_pxFragmentShader->SetFloats( FP_Distance,				1, &fDistance );

	g_pxFragmentShader->SetFloats( FP_SkyboxColour,			3, s_afSkyboxColour );
	g_pxFragmentShader->SetFloats( FP_SunDir1,				3, s_afSunDir1 );
	g_pxFragmentShader->SetFloats( FP_SunDir2,				3, s_afSunDir2 );
	g_pxFragmentShader->SetFloats( FP_HazeGradientPower1,	1, &s_fHazeStrength1 );
	g_pxFragmentShader->SetFloats( FP_HazeGradientPower2,	1, &s_fHazeStrength2 );
	g_pxFragmentShader->SetFloats( FP_HazeColour1,			3, s_afHazeColour1 );
	g_pxFragmentShader->SetFloats( FP_HazeColour2,			3, s_afHazeColour2 );
	g_pxFragmentShader->SetFloats( FP_SunGradientPower1,	1, &s_fSunStrength1 );
	g_pxFragmentShader->SetFloats( FP_SunGradientPower2,	1, &s_fSunStrength2 );
	g_pxFragmentShader->SetFloats( FP_SunColour1,			3, s_afSunColour1 );
	g_pxFragmentShader->SetFloats( FP_SunColour2,			3, s_afSunColour2 );

	g_pxFragmentShader->Bind();

	// screen-space vertex shader
	
	//Pass in the matrix to transform screen space into a camera->vertex direction in world space
	//We want to do the inverse WorldViewProjection, but the projection matrix can't be inverted (since w & z both come from z)
	//So instead form partial invert of projection matrix, and full invert of view matrix
	const Asura_AMatrix4x4& xProj = Asura_PS3_Graphics::GetProjectionMatrix();
	Asura_AMatrix3x3 xFrustumMatrix(
		1 / xProj.GetElement(0,0), 0, 0,
		0, 1 / xProj.GetElement(1,1), 0,
		0, 0, 1 / xProj.GetElement(3,2) //z scale goes into w, for x/w, y/w (z value is no use to us since z/w = z-buffer value)
	);

	//Effectively we create a transformation to the frustum, so  x=y=z>0 lies along the top-right edge of the frustum, -x=y=z>0 is top left, and so on.
	Asura_AMatrix3x3 xCamToWorld = *Asura_Camera::GetOrientation();
	xFrustumMatrix = xFrustumMatrix * xCamToWorld;
	
	g_pxVertexShader->SetUniformParameter( VP_FrustumMatrix, xFrustumMatrix.GetElements() );

	g_pxVertexShader->Bind(); //I think I don't need to set anything (!)

	g_xSkyboxVB.Bind( g_axVertexDeclaration, 1 );
	g_xSkyboxVB.RenderQuads( 4 );


	// write to the stencil, so any clouds etc can use it as a mask

// huh? shouldn't this go first?
// is this clearing the iojothe rbuffer fdioujhkjhwerkiosdafwhaTY?

	// Switch SCULL modes to enable fast-cull on things using this stencil
	Asura_PS3_Graphics::SwitchStencilCullMode( ASURA_PS3_STENCIL_EQUAL, ASURA_STENCIL_MASK_SKY, ASURA_STENCIL_MASK_SKY );
	// This forces anything that renders to the ASURA_STENCIL_MASK_SKY to shade all pixels, even if they fail the z-test. This is to accommodate z-fail stencil writing (even if it's set to _KEEP).
	// To avoid the overhead of the sky-shader, render the stencil values in a separate pass to the actual sky

	Asura_PS3_RenderingStates::SetStencilTestEnabled( true );
	Asura_PS3_RenderingStates::SetStencilFunction( ASURA_PS3_STENCIL_ALWAYS, ASURA_STENCIL_MASK_SKY, ASURA_STENCIL_MASK_SKY );
	Asura_PS3_RenderingStates::SetStencilMask(ASURA_STENCIL_MASK_SKY);
	Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_REPLACE ); //assume it was cleared to 0 at the start of the frame & nobody overwrote it

	Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_NONE ); //stencil only

	Asura_PS3_FragmentShaderCopy *pxBlankFP = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_GLOBAL, ASURA_GLOBAL_SHADER_BLANK );
	pxBlankFP->Bind();

	g_xSkyboxVB.RenderQuads( 4 );

	Asura_PS3_RenderingStates::SetStencilTestEnabled( false );
	Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP );
	Asura_PS3_DeferredRendering::Bind();
 	Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0123_ALL );
    ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__SKYBOX );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Skybox::Platform_DoRenderSetup
// Author       : Andy Tate
// Date         : 4-Jul-2005
// Modified		: 19-April-2007 (Catalin Arsenescu) -> copied from xBox360
// Description  : Handles the PS3 specific set-up for rendering sky-boxes.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_DoRenderSetup (void)
{
}

/// End of function 'Asura_Skybox::Platform_DoRenderSetup'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Skybox::Platform_Reset
// Author       : Andy Tate
// Date         : 4-Jul-2005
//
// Description  : Handles the PS3 specific skybox resetting.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_Reset (void)
{
}

/// End of function 'Asura_Skybox::Platform_Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Skybox::Platform_Shutdown
// Author       : Andy Tate
// Date         : 4-Jul-2005
//
// Description  : Handles the PS3 specific skybox shutdown.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_Shutdown (void)
{
	// Destroy the vertex-buffers
	g_xSkyboxVB.Destroy();
	delete g_pxVertexShader;

	g_pxVertexShader = NULL;
	g_pxFragmentShader = NULL;
}

/// End of function 'Asura_Skybox::Platform_Shutdown'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Skybox::Platform_ShouldModifyFogForSkyboxObjects
// Author       : Andy Tate
// Date         : 27-Jul-2005
//
// Description  : 
//
// Return value : Always "false".
////////////////////////////////////////////////////////////////////////////////

bool Asura_Skybox::Platform_ShouldModifyFogForSkyboxObjects (void)
{
	return	(false);
}

/// End of function 'Asura_Skybox::Platform_ShouldModifyFogForSkyboxObjects'. ///


/// End of file. ///
