
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_VolatileMemoryManagement.h
// Author : Andy Tate
// Date   : 9-May-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_PSP_VOLATILEMEMORYMANAGEMENT_H__
#define __ASURA_PSP_VOLATILEMEMORYMANAGEMENT_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Handle.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#if !defined(ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS)
#ifdef ASURA_USE_STREAMING
#define ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS	2
#else
#define ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS	1
#endif
#endif	// !defined(ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS)



////////////////////////////////////////////////////////////////////////////////
// Class declarations...
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_PSP_VolatileMemoryBlock
// Author       : Andy Tate
// Date         : 18-May-2006
//
// Description  : Class to handle a single block of allocated volatile memory.
////////////////////////////////////////////////////////////////////////////////

class Asura_PSP_VolatileMemoryBlock
{
public:
	// Public typedefs...
	typedef void AllocationCallbackFunction (u_char* pucDataStart);
	typedef void DeallocationCallbackFunction (u_char* pucDataStart);

	// Public functions...
	Asura_PSP_VolatileMemoryBlock();
	~Asura_PSP_VolatileMemoryBlock();
	void Initialise(const u_int uNumBytes, u_char* pucDataStart, 
					AllocationCallbackFunction* pfnAllocationCallback, 
					DeallocationCallbackFunction* pfnDeallocationCallback);
	void Shutdown();

	// Allocation functions...
	void Deallocate();
	void Reallocate(u_char* pucDataStart);

	// Accessor functions...
	u_int GetSize(void) const { return(m_uSize); }

private:
	// Private variables...
	u_int							m_uSize;					// The number of bytes in this memory-block.
	u_char*							m_pucDataStart;				// A pointer to the start of the data.
	AllocationCallbackFunction*		m_pfnAllocationCallback;	// A pointer to the allocation callback function.
	DeallocationCallbackFunction*	m_pfnDeallocationCallback;	// A pointer to the deallocation callback function.

	// Private functions...
	Asura_PSP_VolatileMemoryBlock  (const Asura_PSP_VolatileMemoryBlock& xOther);
	Asura_PSP_VolatileMemoryBlock&	operator=	(const Asura_PSP_VolatileMemoryBlock& xOther);

};   // End of class 'Asura_PSP_VolatileMemoryBlock'.


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_PSP_VolatileMemoryManagement
// Author       : Andy Tate
// Date         : 9-May-2006
//
// Description  : Class to handle multiple allocated from the volatile memory
//				  on the PSP. This shouldn't really be necessary, but the Sony
//				  libraries you do not let you perform multiple allocations.
////////////////////////////////////////////////////////////////////////////////

class Asura_PSP_VolatileMemoryManagement
{
public:
	// Public typedefs...
	typedef void AllocationCallbackFunction (u_char* pucDataStart);
	typedef void DeallocationCallbackFunction (u_char* pucDataStart);

	// Public functions...
	static	void			Initialise	(void);
	static	void			Shutdown	(void);

	// Allocation/deallocation functions...
	static	Asura_Handle	Allocate	(const u_int uNumBytes, AllocationCallbackFunction* pfnAllocationCallback, DeallocationCallbackFunction* pfnDeallocationCallback);
	static	bool			Deallocate	(Asura_Handle& uHandle);

	// Power-management functions...
	static	void			Suspend		(void);
	static	void			Resume		(void);


protected:
	// Protected variables...


	// Protected functions...


private:
	// Private variables...
	static	bool							s_bInitialised;
	static	bool							s_bSuspended;
	static	bool							s_bPendingInitialise;
	static	u_char*							s_pucDataStart;		// A pointer to the start of the data allocated from volatile memory.
	static	u_int							s_uSizeOfData;		// The size of the data allocated from volatile memory.
	static	u_int							s_uCurrentOffset;	// The current offset within volatile memory.

	static	u_int							s_uNumMemoryBlocksInUse;	// The number of memory-blocks allocated.
	static	Asura_PSP_VolatileMemoryBlock	s_axMemoryBlocks[ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS];	// Array of useable memory-blocks.


	// Private functions...

	Asura_PSP_VolatileMemoryManagement  (void);
	Asura_PSP_VolatileMemoryManagement  (const Asura_PSP_VolatileMemoryManagement& xOther);
	~Asura_PSP_VolatileMemoryManagement ();

	Asura_PSP_VolatileMemoryManagement&	operator=	(const Asura_PSP_VolatileMemoryManagement& xOther);

};   // End of class 'Asura_PSP_VolatileMemoryManagement'.


#endif   // ifndef __ASURA_PSP_VOLATILEMEMORYMANAGEMENT_H__


/// End of file. ///
