//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Maths.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"

#include "Core/Asura_PlatformBasics.h"

// G L O B A L S /////////////////////////////////////////////////////////////

const Asura_Vector_2   Asura_Maths::ZeroVector2( 0.0f, 0.0f );

const Asura_Vector_3   Asura_Maths::ZeroVector3(0.0f, 0.0f, 0.0f);

const Asura_Matrix_3x3 Asura_Maths::IdentityMatrix3x3( 1.0f, 0.0f, 0.0f,
                                                       0.0f, 1.0f, 0.0f,
                                                       0.0f, 0.0f, 1.0f );

const Asura_Quat       Asura_Maths::IdentityQuat( 0.0f, 0.0f, 0.0f, 1.0f );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Bounding_Box.
//

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_Bounding_Box::GetSqdMinDistFromBB().
// PARAMETERS   xOtherBB - The BB to be tested. Both BBs are assumed to be in
//                         the same space.
// RETURNS      Minimum distance between the two BBs.
// DESCRIPTION  See RETURNS.
// AUTHOR       MP - 27.11.2002.
//              Based on GetSqdMinDistFromPoint() - see below.
//
//////////////////////////////////////////////////////////////////////////////
float Asura_Bounding_Box::GetSqdMinDistFromBB( const Asura_Bounding_Box& xOtherBB ) const
{
    float fSqdDistance = 0.0f;
    float fDelta;

    if( xOtherBB.MaxX < MinX )
    {
        fDelta = xOtherBB.MaxX - MinX;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xOtherBB.MinX > MaxX )
    {
        fDelta = xOtherBB.MinX - MaxX;
        fSqdDistance += fDelta * fDelta;
    }

    if( xOtherBB.MaxY < MinY )
    {
        fDelta = xOtherBB.MaxY - MinY;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xOtherBB.MinY > MaxY )
    {
        fDelta = xOtherBB.MinY - MaxY;
        fSqdDistance += fDelta * fDelta;
    }

    if( xOtherBB.MaxZ < MinZ )
    {
        fDelta = xOtherBB.MaxZ - MinZ;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xOtherBB.MinZ > MaxZ )
    {
        fDelta = xOtherBB.MinZ - MaxZ;
        fSqdDistance += fDelta * fDelta;
    }

    return fSqdDistance;
}
// End function Asura_Bounding_Box::GetSqdMinDistFromBB().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_Bounding_Box::GetSqdMinDistFromPoint().
// PARAMETERS   xPoint - Point to be tested. This is assumed to be in the
//                       same space as the BB.
// RETURNS      Minimum distance from the point to the BB.
// DESCRIPTION  See RETURNS.
// AUTHOR       Original code: Magic Software. (Note that we're allowed to
//                             use it under the terms of their license.)
//              Asura port   : MP - 27.11.2002.
//
//////////////////////////////////////////////////////////////////////////////
float Asura_Bounding_Box::GetSqdMinDistFromPoint( const Asura_Vector_3& xPoint ) const
{
    float fSqdDistance = 0.0f;
    float fDelta;

    if( xPoint.x < MinX )
    {
        fDelta = xPoint.x - MinX;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xPoint.x > MaxX )
    {
        fDelta = xPoint.x - MaxX;
        fSqdDistance += fDelta * fDelta;
    }

    if( xPoint.y < MinY )
    {
        fDelta = xPoint.y - MinY;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xPoint.y > MaxY )
    {
        fDelta = xPoint.y - MaxY;
        fSqdDistance += fDelta * fDelta;
    }

    if( xPoint.z < MinZ )
    {
        fDelta = xPoint.z - MinZ;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xPoint.z > MaxZ )
    {
        fDelta = xPoint.z - MaxZ;
        fSqdDistance += fDelta * fDelta;
    }

    return fSqdDistance;
}
// End function Asura_Bounding_Box::GetSqdMinDistFromPoint().

// essentially the same calculation as GetSqdMinDistFromPoint, but this returns the result as an unsquared vector
void Asura_Bounding_Box::CalcMinVectorIntoBox( const Asura_Vector_3& xPoint, Asura_Vector_3& xVector ) const
{
    xVector.x = xVector.y = xVector.z = 0.0f;

    if ( xPoint.x < MinX )
    {
        xVector.x = MinX - xPoint.x;
    }
    else if ( xPoint.x > MaxX )
    {
        xVector.x = MaxX - xPoint.x;
    }
    if ( xPoint.y < MinY )
    {
        xVector.y = MinY - xPoint.y;
    }
    else if ( xPoint.y > MaxY )
    {
        xVector.y = MaxY - xPoint.y;
    }
    if ( xPoint.z < MinZ )
    {
        xVector.z = MinZ - xPoint.z;
    }
    else if ( xPoint.z > MaxZ )
    {
        xVector.z = MaxZ - xPoint.z;
    }
}

void Asura_Bounding_Box::FromOBB( const Asura_Bounding_Box& xOBB, const Asura_Matrix_3x3& xOri, bool bRotateAboutMidpoint /*=false*/ )
{
    const Asura_Vector_3 xCentre  = xOBB.CalculateCentre();
    const Asura_Vector_3 xExtents = xOBB.CalculateExtents();

    float fHalfAABBExtentX = 0.5f * (   Asura_Maths::Abs(xOri.e11 * xExtents.x)
                                      + Asura_Maths::Abs(xOri.e12 * xExtents.y)
                                      + Asura_Maths::Abs(xOri.e13 * xExtents.z) );

    float fHalfAABBExtentY = 0.5f * (   Asura_Maths::Abs(xOri.e21 * xExtents.x)
                                      + Asura_Maths::Abs(xOri.e22 * xExtents.y)
                                      + Asura_Maths::Abs(xOri.e23 * xExtents.z) );

    float fHalfAABBExtentZ = 0.5f * (   Asura_Maths::Abs(xOri.e31 * xExtents.x)
                                      + Asura_Maths::Abs(xOri.e32 * xExtents.y)
                                      + Asura_Maths::Abs(xOri.e33 * xExtents.z) );

    MinX = -fHalfAABBExtentX;
    MaxX =  fHalfAABBExtentX;
    MinY = -fHalfAABBExtentY;
    MaxY =  fHalfAABBExtentY;
    MinZ = -fHalfAABBExtentZ;
    MaxZ =  fHalfAABBExtentZ;

	if(bRotateAboutMidpoint)
	{
		Displace( xCentre );
	}
	else
	{
		Displace( xCentre*xOri );
	}
}

//
// Asura_RandomGenerator.
//

float Asura_RandomGenerator::GetNormalFloat( void )
{
    // We generate two at a time so cache the other one for the next call.
    // The maths below is pretty expensive so this is worth the effort.
    static bool bGotSpare = false;
    static float fSpare;

    if( bGotSpare )
    {
        bGotSpare = false;
        return fSpare;
    }

    // We use a Box-Muller transformation to get two normal distribution
    // random numbers from two uniform distribution ones. Given the latter,
    // u0 and u1, our output values are:
    //
    //     f0 = sqrt(-2 ln u0) * cos(2Pi * u1)
    //     f1 = sqrt(-2 ln u0) * sin(2Pi * u1)
    //
    // However, we can avoid the trigonometry by picking a point inside the
    // 2D unit circle, at the potential expense of having to generate extra
    // input uniform random numbers. See "Numerical Recipes In C" (page 289
    // in 2nd edition) for more details.

    float f0, f1, fMagSqd;
    do
    {
        f0 = 2.0f * GetFloat() - 1.0f;
        f1 = 2.0f * GetFloat() - 1.0f;
        fMagSqd = f0 * f0 + f1 * f1;
    }
    while( !fMagSqd || fMagSqd >= 1.0f );

    float fScale = Asura_Maths::Sqrt( -2.0f * logf(fMagSqd) / fMagSqd );
    
    fSpare = f1 * fScale;
    bGotSpare = true;

    return f0 * fScale;
}

//
// Asura_Vector_2.
//

//Choose a random direction within an angular deviation of this vector.
Asura_Vector_2 Asura_Vector_2::ChooseRandomDirectionInSegment(const float fMaxTheta, const float fMinTheta) const
{
	float fCosTheta, fSinTheta;

	if(x*x + y*y < 0.000001f)
	{
		Asura_Assert(false, "ChooseRandomDirectionInSegment() used with zero starting vector");
		return *this;
	}

	//Work out a random theta in the range
	// 0 <= theta <= fMaxTheta, and take its cosine.
	fCosTheta = Asura_Maths::Cosine((fMaxTheta-fMinTheta)*Asura_Random::GetFloat() + fMinTheta);

	//Randomly decide on which side of the vector the point should lie, and
	//derive the sine from the cosine, using a square root.
	if(Asura_Random::GetIntU()&0x40000000)
	{
		fSinTheta = Asura_Maths::Sqrt(1.f - fCosTheta*fCosTheta);
	}
	else
	{
		fSinTheta = -Asura_Maths::Sqrt(1.f - fCosTheta*fCosTheta);
	}

	//Rotate this through the angle specified
	Asura_Vector_2 xRet;

	xRet.x = fCosTheta * x - fSinTheta * y;
	xRet.y = fSinTheta * x + fCosTheta * y;

	//Choose a random length for the new vector, in the range of 0..1
	xRet.Normalise();

	return xRet;
}

//Choose a vector of random length within a specified angular devation of
//this vector, and with a length in the specified range.
Asura_Vector_2 Asura_Vector_2::ChooseRandomVectorInSegment(const float fMaxTheta, const float fMinTheta, const float fMaxLength, const float fMinLength) const
{
	return ChooseRandomDirectionInSegment(fMaxTheta, fMinTheta) 
			* ((fMaxLength - fMinLength) * Asura_Random::GetFloat() + fMinLength);
}

//Choose a random direction within an angular deviation of this vector.
Asura_Vector_2 Asura_Vector_2::ChooseRandomDirectionInSegmentSeeded(const float fMaxTheta, const float fMinTheta) const
{
	float fCosTheta, fSinTheta;

	if(x*x + y*y < 0.000001f)
	{
		Asura_Assert(false, "ChooseRandomDirectionInSegment() used with zero starting vector");
		return *this;
	}

	//Work out a random theta in the range
	// 0 <= theta <= fMaxTheta, and take its cosine.
	fCosTheta = Asura_Maths::Cosine((fMaxTheta-fMinTheta)*Asura_Random::GetFloat() + fMinTheta);

	//Randomly decide on which side of the vector the point should lie, and
	//derive the sine from the cosine, using a square root.
	if(Asura_Random::GetSeededIntU()&0x40000000)
	{
		fSinTheta = Asura_Maths::Sqrt(1.f - fCosTheta*fCosTheta);
	}
	else
	{
		fSinTheta = -Asura_Maths::Sqrt(1.f - fCosTheta*fCosTheta);
	}

	//Rotate this through the angle specified
	Asura_Vector_2 xRet;

	xRet.x = fCosTheta * x - fSinTheta * y;
	xRet.y = fSinTheta * x + fCosTheta * y;

	//Choose a random length for the new vector, in the range of 0..1
	xRet.Normalise();

	return xRet;
}

//Choose a vector of random length within a specified angular devation of
//this vector, and with a length in the specified range.
Asura_Vector_2 Asura_Vector_2::ChooseRandomVectorInSegmentSeeded(const float fMaxTheta, const float fMinTheta, const float fMaxLength, const float fMinLength) const
{
	return ChooseRandomDirectionInSegmentSeeded(fMaxTheta, fMinTheta) 
		* ((fMaxLength - fMinLength) * Asura_Random::GetFloat() + fMinLength);
}
//
// Asura_Vector_3.
//

//Choose a unit vector in a cone, with an angle subtended to this vector in the specified range.
Asura_Vector_3 Asura_Vector_3::ChooseRandomDirectionInCone(const float fMaxTheta, const float fMinTheta) const
{
	Asura_Vector_3 xDirection = *this;
	xDirection.Normalise();

	if(fMinTheta > fMaxTheta) return xDirection;

	//Now, we construct an arbitrary orthogonal basis about the axis.
	float fAbsX = Asura_Maths::Abs(xDirection.x);
	float fAbsY = Asura_Maths::Abs(xDirection.y);
	float fAbsZ = Asura_Maths::Abs(xDirection.z);

	//Asura_Vector_3 xContructionAxis(1.f, 0.f, 0.f);
	Asura_Vector_3 xNormal;

	//Work out the largest component of xDirection, and construct an appropriate normal,
	//using the unrolled cross product with that axis.
	if(fAbsX > fAbsY)
	{
		if(fAbsY > fAbsZ)
		{	
			//Cross with z-axis.
			xNormal.x = xDirection.y;
			xNormal.y = -xDirection.x;
			xNormal.z = 0.f;
		}
		else
		{
			//Cross with y-axis.
			xNormal.x = -xDirection.z;
			xNormal.y = 0.f;
			xNormal.z = xDirection.x;
		}
	}
	else if(fAbsX > fAbsZ)
	{
		//Cross with z-axis.
		xNormal.x = xDirection.y;
		xNormal.y = -xDirection.x;
		xNormal.z = 0.f;
	}
	else
	{
		//Cross with x-axis.
		xNormal.x = 0.f;
		xNormal.y = xDirection.z;
		xNormal.z = -xDirection.y;
	}

	xNormal.Normalise();

	//Now that we have the first normal, we can construct the second.
	Asura_Vector_3 xNormal2 = xDirection^xNormal;

	//Now, we choose a random 2-vector.
	Asura_Vector_2 xRandom;
	xRandom.RandomDirection();

	//Now we choose a random number between the cosines of the two angles.
	float fV1 = Asura_Maths::Cosine(fMinTheta);
	float fV2 = Asura_Maths::Cosine(fMaxTheta);

	float fRandomFloat = Asura_Random::GetFloat() * (fV1-fV2) + fV2;

	//Just check the normalisation, to be sure.
	if(fRandomFloat > 1.f) fRandomFloat = 1.f;
	else if(fRandomFloat < -1.f) fRandomFloat = -1.f;

	//Scale the primary component to the chosen length.
	xDirection *= fRandomFloat;

	//Work out the correct normalisation factor for the second component...
	//...and add the second component, scaled by that factor.
	xDirection += (xNormal * xRandom.x + xNormal2 * xRandom.y) * Asura_Maths::Sqrt(1.f - fRandomFloat*fRandomFloat);

	return xDirection;
}

//Choose a vector of length within the specified range, and with an angular deviation
//from this vector in the specified range.
Asura_Vector_3 Asura_Vector_3::ChooseRandomVectorInCone(const float fMaxTheta, const float fMinTheta, const float fMaxLength, const float fMinLength) const
{
	float fLength = Asura_Random::GetFloat()*(fMaxLength - fMinLength) + fMinLength;

	return ChooseRandomDirectionInCone(fMaxTheta, fMinTheta)*fLength;
}

//Choose a unit vector in a cone, with an angle subtended to this vector in the specified range.
Asura_Vector_3 Asura_Vector_3::ChooseRandomDirectionInConeSeeded(const float fMaxTheta, const float fMinTheta) const
{
	Asura_Vector_3 xDirection = *this;
	xDirection.Normalise();

	if(fMinTheta > fMaxTheta) return xDirection;

	//Now, we construct an arbitrary orthogonal basis about the axis.
	float fAbsX = Asura_Maths::Abs(xDirection.x);
	float fAbsY = Asura_Maths::Abs(xDirection.y);
	float fAbsZ = Asura_Maths::Abs(xDirection.z);

	//Asura_Vector_3 xContructionAxis(1.f, 0.f, 0.f);
	Asura_Vector_3 xNormal;

	//Work out the largest component of xDirection, and construct an appropriate normal,
	//using the unrolled cross product with that axis.
	if(fAbsX > fAbsY)
	{
		if(fAbsY > fAbsZ)
		{	
			//Cross with z-axis.
			xNormal.x = xDirection.y;
			xNormal.y = -xDirection.x;
			xNormal.z = 0.f;
		}
		else
		{
			//Cross with y-axis.
			xNormal.x = -xDirection.z;
			xNormal.y = 0.f;
			xNormal.z = xDirection.x;
		}
	}
	else if(fAbsX > fAbsZ)
	{
		//Cross with z-axis.
		xNormal.x = xDirection.y;
		xNormal.y = -xDirection.x;
		xNormal.z = 0.f;
	}
	else
	{
		//Cross with x-axis.
		xNormal.x = 0.f;
		xNormal.y = xDirection.z;
		xNormal.z = -xDirection.y;
	}

	xNormal.Normalise();

	//Now that we have the first normal, we can construct the second.
	Asura_Vector_3 xNormal2 = xDirection^xNormal;

	//Now, we choose a random 2-vector.
	Asura_Vector_2 xRandom;
	xRandom.RandomDirectionSeeded();

	//Now we choose a random number between the cosines of the two angles.
	float fV1 = Asura_Maths::Cosine(fMinTheta);
	float fV2 = Asura_Maths::Cosine(fMaxTheta);

	float fRandomFloat = Asura_Random::GetFloat() * (fV1-fV2) + fV2;

	//Just check the normalisation, to be sure.
	if(fRandomFloat > 1.f) fRandomFloat = 1.f;
	else if(fRandomFloat < -1.f) fRandomFloat = -1.f;

	//Scale the primary component to the chosen length.
	xDirection *= fRandomFloat;

	//Work out the correct normalisation factor for the second component...
	//...and add the second component, scaled by that factor.
	xDirection += (xNormal * xRandom.x + xNormal2 * xRandom.y) * Asura_Maths::Sqrt(1.f - fRandomFloat*fRandomFloat);

	return xDirection;
}

//Choose a vector of length within the specified range, and with an angular deviation
//from this vector in the specified range.
Asura_Vector_3 Asura_Vector_3::ChooseRandomVectorInConeSeeded(const float fMaxTheta, const float fMinTheta, const float fMaxLength, const float fMinLength) const
{
	float fLength = Asura_Random::GetFloat()*(fMaxLength - fMinLength) + fMinLength;

	return ChooseRandomDirectionInConeSeeded(fMaxTheta, fMinTheta)*fLength;
}
// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

void Asura_Maths::OrientationFromDirection(Asura_Matrix_3x3 &Orientation, const Asura_Vector_3 &d)
{
    Asura_Vector_3 XVector;
    Asura_Vector_3 YVector;
    Asura_Vector_3 ZVector = d;

    ZVector.Normalise();


    if (ZVector.z==0.0f && ZVector.x==0.0f)
    {
        XVector.x = -ZVector.y;
        XVector.y = ZVector.x;
        XVector.z = 0.0f;

        XVector.Normalise();

        YVector = -XVector^ZVector;
    }
    else
    {
        XVector.x = ZVector.z;
        XVector.y = 0;
        XVector.z = -ZVector.x;

        XVector.Normalise();

        YVector = -XVector^ZVector;
    }

    Orientation.e11 = XVector.x;
    Orientation.e21 = XVector.y;
    Orientation.e31 = XVector.z;

    Orientation.e12 = YVector.x;
    Orientation.e22 = YVector.y;
    Orientation.e32 = YVector.z;

    Orientation.e13 = ZVector.x;
    Orientation.e23 = ZVector.y;
    Orientation.e33 = ZVector.z;
}

#if defined ASURA_PSP_ASM_USE_SN_SYNTAX
// copied from en.prog.vfpu newsgroup posting "Fast Arctangent Functions" by Craig Wright
struct sVector4
{
	float x;
	float y;
	float z;
	float w;

} __attribute__((aligned(16)));

float atan2f_atan66vfpu( float y, float x )
{
	static const char tag[]		= "C. Wright version of atan2f";
	// Constants
	static const sVector4 region_constants =
	{
		1.0f,
		0.26794919243112270647255365849413f,	// tan( pi / 12 )
		-3.7320508075688772935274463415059f,	// -1 / tan( pi / 12 )
		0.57735026918962576450914878050196f		// tan( pi / 6 )
	};

	static const sVector4 equation_constants =
	{
		1.6867629106f,
		0.4378497304f,
		1.6867633134f,
		0.0f
	};

	int y_value __attribute__((forceregister)) = *reinterpret_cast<int *>( &y );
	int x_value __attribute__((forceregister)) = *reinterpret_cast<int *>( &x );
	int negative_zero __attribute__((forceregister)) = 0x80000000;
	int y_sign __attribute__((forceregister));
	int x_sign __attribute__((forceregister));
	int result __attribute__((forceregister));

	__asm
	{
		.set	push
		.set	noreorder

// Load values and test for special cases /////////////////////////////////////

		// Store the sign for y and abs y
		and		y_sign, y_value, negative_zero
		xor		y_value, y_value, y_sign

		// Load our values
		mtv		y_value, s600
		mtv		x_value, s601

		// Test for zeros
		vcmp.p	ez, c600
		bvt		4, Atan2_Zeros
		nop

		// Test for infinities or NaNs
		vcmp.p	es, c600
		bvt		4, Atan2_Infinities
		nop

		// Calculate y/x
		vrcp.s	s700, s601
		vmul.s	s700, s700, s600

		b		Atan2_Start
		nop

// Zeros //////////////////////////////////////////////////////////////////////

	Atan2_Zeros:

		bvf		0, Atan2_NonZeroY
		nop
		and		x_sign, x_value, negative_zero

		Atan2_Case0:	// atan2( +-0, -0 ) = +-pi
		Atan2_Case1:	// atan2( +-0, +0 ) = +-0
		Atan2_Case2:	// atan2( +-0, x < 0 ) = +-pi
		Atan2_Case3:	// atan2( +-0, x > 0 ) = +-0

			beql	x_sign, $zero, Atan2_Finish
			vzero.s	s700
			b		Atan2_Finish
			vcst.s	s700, vfpu_pi

	Atan2_NonZeroY:

		Atan2_Case4:	// atan2( y < 0, +-0 ) = -pi/2
		Atan2_Case5:	// atan2( y > 0, +-0 ) = pi/2

			b		Atan2_Finish
			vcst.s	s700, vfpu_pi_2

// Infinities /////////////////////////////////////////////////////////////////

	Atan2_Infinities:

		bvf		1, Atan2_FiniteX
		nop
		and		x_sign, x_value, negative_zero
		bvf		0, Atan2_FiniteY
		nop

		Atan2_Case9:	// atan2( +-oo, -oo ) = +-3pi/4
		Atan2_Case10:	// atan2( +-oo, +oo ) = +-pi/4

			beq		x_sign, $zero, Atan2_Finish
			vcst.s	s700, vfpu_pi_4
			vpfxt	3,y,z,w
			b		Atan2_Finish
			vmul.s	s700, s700, s000

	Atan2_FiniteY:

		Atan2_Case6:	// atan2( +-y, -oo ) = +-pi
		Atan2_Case7:	// atan2( +-y, +oo ) = +-0

			beql	x_sign, $zero, Atan2_Finish
			vzero.s	s700
			b		Atan2_Finish
			vcst.s	s700, vfpu_pi

	Atan2_FiniteX:

		Atan2_Case8:	// atan2( +-oo, x ) = +-pi/2
			b		Atan2_Finish
			vcst.s	s700, vfpu_pi_2

// Usual cases ////////////////////////////////////////////////////////////////

	Atan2_Start:

		// Load the region constants
		lv.q	c710, &region_constants

		// NOTE: No need to test for infinities here

		// Use the absolute value
		vabs.s	s000, s700

		// Test for complement and region
		vpfxs	x,x,-x,w
		vcmp.t	gt, c000, c710		// c0 :=  x > 1
									// c1 :=  x > tan( pi / 12 )
									// c2 := -x > -1 / tan( pi / 12 )

	Atan2_Complement:

		// Test for complement
		bvf		0, Atan2_NormalRegionTest
		nop

		vrcp.s	s000, s000	// s000 := 1 / x

	Atan2_ComplementedRegionTest:

		// Adjust for region if:
		// => -x > -1 / tan( pi / 12 )
		bvf		2, Atan2_Calculation
		nop
		b		Atan2_RegionAdjustment
		nop

	Atan2_NormalRegionTest:

		// Adjust for region if
		// => x > tan( pi / 12 )
		bvf	1, Atan2_Calculation
		nop

	Atan2_RegionAdjustment:

		// x = ( x - r3 ) / ( 1 + r3.x )

		vmul.q		c010, c000(1,x,0,0), c710(-w,w,0,0)	// s010 := -r3
														// s011 := r3.x
		vadd.p		c010, c000(x,1), c010				// s010 := ( x - r3 )
														// s011 := ( 1 + r3.x )
		vrcp.s		s011, s011							// s011 := 1 / ( 1 + r3.x )
		vmul.s		s000, s010, s011					// s000 := ( x - r3 ) / ( 1 + r3.x )

	Atan2_Calculation:

		// y = ( c0.x + c1.x^3 ) / ( c2 + x^2 );

		// Load the constants
		lv.q	c100, &equation_constants

		vmul.s	s001, s000, s000			// s001 := x^2
		vdot.p	s002, c000(1,y), c100		// s002 := ( c0 + c1.x^2 )
		vadd.s	s003, s102, s001			// s003 := ( c2 + x^2 )
		vrcp.s	s003, s003					// s003 := 1 / ( c2 + x^2 )
		vmul.s	s000, s000, s002			// s000 := ( c0.x + c1.x^3 )
		vmul.s	s700, s000, s003			// s700 := ( c0.x + c1.x^3 ) / ( c2 + x^2 )

	Atan2_CorrectForRegion:

		// We adjusted the region if 
		// c1 :=  x > tan( pi / 12 )
		// c2 := -x > -1 / tan( pi / 12 )
		bvf		1, Atan2_CorrectForComplement
		nop
		bvf		2, Atan2_CorrectForComplement
		//nop

		// y += sixthpi;
		vcst.s	s702, vfpu_pi_6
		vadd.s	s700, s700, s702

	Atan2_CorrectForComplement:

		// We complemented if
		// c0 :=  x > 1
		bvf		0, Atan2_CorrectForQuadrant
		//nop

		// y = halfpi - y;
		vcst.s	s703, vfpu_pi_2
		vsub.s	s700, s703, s700

	Atan2_CorrectForQuadrant:

		// if x < 0, result = pi - result
		bgez	x_value, Atan2_CorrectForSign
		vcst.s	s701, vfpu_pi
		vsub.s	s700, s701, s700

	Atan2_CorrectForSign:
	Atan2_Finish:

		mfv		result, s700
		or		result, result, y_sign

		.set	pop
	}

	return *reinterpret_cast<float*>( &result );
}
#elif 0 && defined ASURA_PSP_USE_VFPU_ASSEMBLER
/*
	Inspired by: http://sources.FreeBSD.org/HEAD/src/lib/msun/src/e_atan2f.c
*/

float atan2f_gccvfpu(register const float y,register const float x)
{
	vfpu_v4 afPI2;
	vfpu_v4 afT;
	vfpu_v4 afYdivX;

	__asm__
	(
		"mtv			%[fY], %y[afXY]\n"
		"mtv			%[fX], %x[afXY]\n"
		"vrcp.s			%[afYdivX], %x[afXY]\n"

		// CC[4] = (%x[afXY] is special) | (%y[afXY] is special)
		// CC[5] = (%x[afXY] is special) & (%y[afXY] is special)
		"vcmp.p			ES, %[afXY]\n"
		"bvt			4, atan2_specials\n"
		" vcst.s		%[afPI2], VFPU_PI_2\n"

		"vmul.s			%[afYdivX], %y[afXY], %x[afYdivX]\n"

		/* If y/x is NaN, NaN is the result. */
		"vcmp.s			EN, %[afYdivX]\n"
		"bvt			0, atan2_done\n"
		" vmov.s		%[afT], %[afYdivX]\n"

		/* if y/x is special */
		"vcmp.s			ES, %[afYdivX]\n"
		"bvtl			0, atan2_specialYdivX\n"
		" vsgn.s		%x[afT], %y[afXY]\n"	// %[afT] = sign(%y[afXY]) = sign(y)

		/* y/x is not special */

		"vmul.s			s011, %x[afYdivX], %x[afYdivX]\n"		// s011 = %[afYdivX] * %[afYdivX]
		"vone.s			s003\n"					// s003 = 1.0f

		"vcmp.s			LT, %x[afT][|x|], s003\n"	// CC[0] = |%x[afT]| < s003
		"vadd.s			s003, s011, s003\n"		// s003 = s011 + s003 = s*s + 1

		"bvfl			0, atan2_specialX\n"	// if (|x| < s003)
		" vrcp.s		s003, s003\n"			// s003 = 1 / s003
		"vrsq.s			s003, s003\n"			// s003 = 1 / sqrt(s003)

		"vmul.s			%x[afT], %x[afT], s003\n" // %x[afT] = %x[afT] * s003 = s / sqrt(s*s + 1)
		"b				2f\n"					// goto 2f
		" vasin.s		%x[afT], %x[afT]\n"		// %x[afT] = asin(%x[afT])

	"atan2_specials:\n"
		/* If both X and Y are special, result is NaN */
		"vfim.s			%x[afT], -1.0\n"		// %x[afT] = -1.0
		"bvt			5, 4f\n"				// if (CC[5]!=0) goto 4f
		" vrsq.s		%x[afT], %x[afT]\n"		// %x[afT] = sqrt(-1.0) = NaN

		/* If either X or X are NaN or Inf */
		"vcmp.p			EN, %[afXY]\n"			// CC[4] = (%y[afXY] is NaN) | (%x[afXY] is NaN)
		"bvt			4, 4f\n"				// if (CC[4]!=0) goto 4f
		" vcmp.p		 EI, c000\n"

		"vsgn.s			%y[afXY], %y[afXY]\n"	// %y[afXY] = sign(%y[afXY])
		"vcst.s			%x[afPI2], VFPU_PI_4\n"	// %x[afXY] = VFPU_PI_4 = PI / 4

		"bvt			0, 3f\n"				// if (CC[0]!=0) goto 3f
		" vone.s			%x[afT]\n"				// %x[afT] = 1.0

		"vsgn.s			s011, %x[afXY]\n"		// s011 = sign(%x[afXY])
		"b				3f\n"					// goto 3f
		" vsub.s		%x[afT], %x[afT], s011\n" // %x[afT] = %x[afT] - s011

	"atan2_specialX:\n"
		"vmul.s			s003, s011, s003\n"		// s003 = s011 * s003 = s*s / sqrt(s*s + 1)
		"vocp.s			s003, s003\n"			// s003 = 1.0 - s003
		"vsqrt.s		s003, s003\n"			// s003 = sqrt(s003)
		"vasin.s		s003, s003\n"			// s003 = asin(s003)
		"vocp.s			%x[afT], s003\n"		// %x[afT] = 1.0 - s003

	"atan2_specialYdivX:\n"
		"vsgn.p			c000[x,0:1], c000[x,-y]\n"	// %y[afXY] = sign(%y[afXY])
												// %x[afXY] = sat0(sign(-%x[afXY]))
		"vcmp.s			EZ, %x[afXY]\n"			// CC[0] = (%x[afXY]==0.0) = (x < 0) ? 1 : 0
		"bvt			0, 3f\n"				// if (x>=0) goto 3f
		" nop\n"

		"vsub.s			%x[afT], %x[afT][2], %x[afT][|x|]\n" // %x[afT] = 2.0f - |%x[afT]|

	"3:\n"
		"vmul.s			%x[afT], %x[afT][|x|], %y[afXY]\n"	// %x[afT] = |%x[afT]| * %y[afXY]
		"vmul.s			%x[afT], %x[afT], %x[afPI2]\n"		// %x[afT] = %x[afT] * %x[afPI2]

	"atan2_done:\n"
		"mfv			%0, %x[afT]\n"				// d    = %x[afT]

	: [fAtan2] "=r" (fAtan2), [afXY] "=j" (afXY), [afT] "=j" (afT),
	  [afPI2] "=j" (afPI2), [afX] "=j" (afX), [afY] "=j" (afY)

	: "r" (x), "r" (y)
	);

	return fAtan2;
}
#elif 1
/*
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 */

/* Somewhat refactored from the original */

#define F_ZERO		(0.0f)
#define F_EPS		(1.0e-30f)

#define F_PI_LO		(8.7422776573e-08f) /* 0xb3bbbd2e */
#define F_PI_4		(7.8539818525e-01f) /* 0x3f490fdb */
#define F_PI_2		(1.5707963705e+00f) /* 0x3fc90fdb */
#define F_PI		(3.1415927410e+00f) /* 0x40490fdb */

float atan2f_sunpro(register const float y,register const float x)
{
	/* When the vector is on the X axis, sign depends on sign of X */
	static const float axisX[] =
		/* (+0,+x), (-0,+x), (+0,-x), (-0,-x) */
		/* (+y,+INF), (-y,+INF), (+,-INF), (-y,-INF) */
		{ F_ZERO, -F_ZERO,
		  F_PI+F_EPS, -(F_PI+F_EPS) };
	/* When the vector is on the Y axis, sign depends on sign of Y */
	static const float axisY[] =
		/* (+,+0), (-,+0) (+,-0), (-,-0) */
		{ F_PI_2+F_EPS, -(F_PI_2+F_EPS),
		  F_PI_2+F_EPS, -(F_PI_2+F_EPS) };
	/* Diagonals */
	static const float axisD[] =
		/* +INF,-INF; -INF,+INF; +INF,-INF; -INF,-INF */
		{ F_PI_4+F_EPS, -(F_PI_4+F_EPS),
		  3.0f*F_PI_4+F_EPS, -(3.0f*F_PI_4+F_EPS) };

	register const int hx = *reinterpret_cast<const int*>(&x);
	register const int hy = *reinterpret_cast<const int*>(&y);
	register const int ix = hx & 0x7fffffff;
	register const int iy = hy & 0x7fffffff;

	/* If x or y are NaN, result is NaN */
	if (ix > 0x7f800000 || iy > 0x7f800000)
		return x+y;

	/* if x is exactly +1.0f, result is 'atanf' */
	if (hx == 0x3f800000)
		return atanf(y);

	/* 2*sign(x)+sign(y) */
	register const int iQuad = ((hy>>31) & 0x01) | ((hx>>30) & 0x02);

    /* when |y| = 0 we are on the X axis */
	if (iy == 0)
	{
		return axisX[iQuad];
	}

    /* when |x| = 0 we are on the Y axis */
	if (ix == 0)
	{
		return axisY[iQuad];
	}

	if (ix == 0x7f800000)
	{
		if (iy == 0x7f800000)
		{
			/* when |y| and |x| are both INF it is like (+-1,+-1) */
			return axisD[iQuad];
		}
		else
		{
			/* when |x| = INF and |y| <> INF we are on the X axis */
			return axisX[iQuad];
	    }
	}

    /* compute exp of y/x */
	register const int k = (iy-ix) >> 23;
	register float z;

	if (k > 60)
	{
		/* |y/x| > 2**60 */
		z = F_PI_2 + 0.5f*F_PI_LO;
	}
	else if (hx < 0 && k < -60)
	{
		/* |y|/x < 2**(-60) */
		z = 0.0f;
	}
	else
	{
		z = atanf(fabsf(y/x));
	}

	/* Now take it to the right quadrant */
	switch (iQuad)
	{
	case 0 /* +,+ */: return  z;
	case 1 /* -,+ */: return -z;
	case 2 /* +,- */: return  F_PI-z+F_PI_LO;
	case 3 /* -,- */: return -(F_PI-z+F_PI_LO);
	}
}
#endif
