//
// Asura_PSP_FileIO.cpp
//
// Created by KOR - 13/04/2005
// Largely copied from the Retry Read sample
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header
#include "Asura_PSP_FileIO.h"

// Asura includes
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// PSP headers
#include "Core/Asura_PSP_Thread_Priorities.h"
#include "Debugging/Asura_PSP_GPIO.h"
#include "FileAccess/tgIoOpen.h"

#include <fatms.h>	// for dir entry stuff on ms0
#include <kernel.h>
#include <mediaman.h>
#include <psperror.h>
#include <stdio.h>
#include <umderror.h>
#if defined(ASURA_PSP_UMD) && defined(ASURA_PSP_KEEP_UMD_SPINNING)
#include <umddevctl.h>
#endif	// defined(ASURA_PSP_UMD) && defined(ASURA_PSP_KEEP_UMD_SPINNING)

#if defined(ASURA_PSP_UMD)
#define UMD_RETRY_DELAY						(1 * 1000 * 1000)

#if defined(ASURA_PSP_KEEP_UMD_SPINNING)
#define KEEP_UMD_SPINNING_THREAD_STACKSIZE	(32 * 1024)
#define KEEP_UMD_SPINNING_DELAY				(2 * 1000 * 1000)

int Asura_PSP_FileIO::s_iKeepUMDSpinningThreadID	= -1;
#endif	// defined(ASURA_PSP_KEEP_UMD_SPINNING)
#else
static bool g_bForceFMVOnMS = true;
#endif	// defined(ASURA_PSP_UMD)

static volatile bool g_bSuspended = false;
static volatile int g_iPendingReadOperations = 0;
static volatile int g_iFileIOSemaLine;	// for semaphore lockup debugging.

// Static members
SceUID			Asura_PSP_FileIO::s_iSemaphoreId		= -1;
Asura_TickCount	Asura_PSP_FileIO::s_xLastOperationTicks	= 0;

static const int ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME = 0;
static const int iASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME_WITH_TIMEOUT = 1 * 1000 * 1000;

#if defined(ASURA_PSP_DEMO) && !defined(ASURA_PSP_DEMO_PATH)
#define ASURA_PSP_DEMO_PATH "SAMPLE/" // trailing slash is required
#endif 

int	Asura_PSP_FileIO::Initialise()
{
	// Init semaphore
	if (s_iSemaphoreId >= 0)
	{
		return SCE_KERNEL_ERROR_OK;
	}

	if ((s_iSemaphoreId = sceKernelCreateSema("ASURA_IO_SEMA", SCE_KERNEL_SA_THFIFO, 1, 1, NULL)) < 0)
	{
		Asura_OutputToDebugger(("*** ERROR: %s(0x%08x)\n", __FUNCTION__, s_iSemaphoreId));
		return s_iSemaphoreId;
	}

#if defined(ASURA_PSP_UMD) && !defined(ASURA_PSP_ARTIST_BUILD)
	sceUmdActivate(SCE_UMD_MODE_POWERON, SCE_UMD_ALIAS_NAME);
	sceUmdWaitDriveStat(SCE_UMD_READABLE);

	tgIoFillCache("DISC0:/", 0, 1);

#if defined(ASURA_PSP_KEEP_UMD_SPINNING)
	s_iKeepUMDSpinningThreadID	= sceKernelCreateThread("Keep UMD spinning thread", KeepUMDSpinningThread, iASURA_THREAD_PRIORITY_KEEP_UMD_SPINNING, KEEP_UMD_SPINNING_THREAD_STACKSIZE, SCE_KERNEL_TH_USE_VFPU, NULL);
	if(s_iKeepUMDSpinningThreadID < 0)
	{
		Asura_Assertf(false, ("sceKernelCreateThread(): 0x%08X\n",s_iKeepUMDSpinningThreadID));
	}
	else
	{
		SceInt32	status	= sceKernelStartThread(s_iKeepUMDSpinningThreadID, 0, NULL);

		if (status == SCE_KERNEL_ERROR_OK)
		{
			sceKernelDeleteThread(s_iKeepUMDSpinningThreadID);
			s_iKeepUMDSpinningThreadID	= -1;
		}
		else
		{
			Asura_Assertf(false, ("sceKernelStartThread(): 0x%08X\n",status));
		}
	}
#endif	// defined(ASURA_PSP_KEEP_UMD_SPINNING)
#endif	// defined(ASURA_PSP_UMD)

	return SCE_KERNEL_ERROR_OK;
}

int	Asura_PSP_FileIO::Shutdown()
{
	// Semaphore shutdown
	SceUID uid;

	if (s_iSemaphoreId >= 0)
	{
		WaitSemaphoreCB();

		uid = s_iSemaphoreId;
		s_iSemaphoreId = -1;

		sceKernelDeleteSema(uid);
	}

#if defined(ASURA_PSP_UMD) && defined(ASURA_PSP_KEEP_UMD_SPINNING)
	// Shutdown this thread...
	sceKernelDeleteThread(s_iKeepUMDSpinningThreadID);
	s_iKeepUMDSpinningThreadID	= -1;
#endif	// defined(ASURA_PSP_UMD) && defined(ASURA_PSP_KEEP_UMD_SPINNING)

	return SCE_KERNEL_ERROR_OK;
}

#if defined ASURA_PSP_UMD && !defined(ASURA_PSP_ARTIST_BUILD)
bool Asura_PSP_FileIO::DiscExists(void)
{
	int	iStatus	= sceUmdGetDriveStat();

	return((iStatus & SCE_UMD_READABLE) != 0);
}

bool Asura_PSP_FileIO::FileExists (const char* szFilePath)
{
	TGIO_CACHE_ENTRY*	pxCacheEntry	= tgIoFindFile(szFilePath);

	return(pxCacheEntry != NULL);
}

int	Asura_PSP_FileIO::Open(const char *szFilepath, int iFlag, SceMode iMode, const bool bAllowRetry)
{
	int iRet;

	/*E Retry Open */
	while (1)
	{
		// (MH - 25.06.2007) combined WaitSemaphore and drive ready check.
		// Some drive ready code below may be unnecessary now but it 
		// shouldn't hurt to leave it in.
		if ((iRet = WaitSemaphoreAndDriveReady(bAllowRetry, __LINE__)) < 0)
		{
			break;
		}

		iRet = tgIoOpen(szFilepath, iFlag, iMode);

		SignalSemaphore();

		if(bAllowRetry)
		{
			if (iRet >= 0)
			{
				break;
			}

			if (iRet == (int)SCE_ERROR_ERRNO_ENOENT ||
				iRet == (int)SCE_ERROR_ERRNO_EUNIT  ||
				iRet == (int)SCE_ERROR_ERRNO_ENODEV ||
				iRet == (int)SCE_ERROR_ERRNO_ENOMEM ||
				iRet == (int)SCE_ERROR_ERRNO_EINVAL ||
				iRet == (int)SCE_ERROR_ERRNO_ENAMETOOLONG)
			{
				Asura_OutputToDebugger(("*** FATAL ERROR: sceIoOpen(0x%08x) (on file %s)\n", iRet, szFilepath));
#ifndef ASURA_PSP_SONY_SUBMISSION
				{
					char string[2048];
					Asura_PrintfToBfr(string, "File not found: %s", 2048, true, szFilepath);
	#ifndef ASURA_ASSERT
					Asura_AssertFired("", 0, "",string, NULL);
	#else //ASURA_ASSERT
					Asura_Assert(false, string);
	#endif	//ASURA_ASSERT
				}
#endif	//ASURA_PSP_SONY_SUBMISSION
			}

			// KOR - 2/3/2006 - flash GPO light 7 to indicate we are waiting on a file.
			static int iLight = 0;
			iLight = !iLight;
			Asura_PSP_GPIO::SetGPO((iLight << 7));
			Asura_OutputToDebugger(("** READ FAILED. RETRY...: %s(0x%08x) - %s\n", __FUNCTION__, iRet, szFilepath));

			sceKernelDelayThread(UMD_RETRY_DELAY);

			if (iRet == (int)SCE_UMD_NO_MEDIUM ||
				iRet == (int)SCE_UMD_NOT_DEVICE_READY ||
				sceUmdGetDriveStat() & SCE_UMD_MEDIA_OUT)
			{
				Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));
				if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_MEDIA_IN, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
				{
					break;
				}
			}
			if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_READABLE, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
			{
				break;
			}
		}
		else
		{
			break;
		}
	}	 /*E Retry Open While */
	return iRet;
}

int	Asura_PSP_FileIO::Getstat(const char* szFilePath, SceIoStat* pxStat)
{
	int iRet;

	/*E Retry GetStat */
	while (1)
	{
		// (MH - 25.06.2007) combined WaitSemaphore and drive ready check.
		// Some drive ready code below may be unnecessary now but it 
		// shouldn't hurt to leave it in.
		bool bAllowRetry = true;
		if ((iRet = WaitSemaphoreAndDriveReady(bAllowRetry, __LINE__)) < 0)
		{
			break;
		}

		iRet					= sceIoGetstat(szFilePath, pxStat);
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();

		SignalSemaphore();

		if (iRet >= 0)
		{
			break;
		}

		if (iRet == (int)SCE_ERROR_ERRNO_EBADF  ||
			iRet == (int)SCE_ERROR_ERRNO_EINVAL ||
			iRet == (int)SCE_ERROR_ERRNO_ENODEV ||
			iRet == (int)SCE_ERROR_ERRNO_EFSIZE ||
			iRet == (int)SCE_ERROR_ERRNO_EFAULT ||
			iRet == (int)SCE_UMD_LBA_OUTOFRANGE ||
			iRet == (int)SCE_ERROR_ERRNO_NOMAIN_ADDR ||
			iRet == (int)SCE_ERROR_ERRNO_ENAMETOOLONG)
		{
			Asura_OutputToDebugger(("*** FATAL ERROR: sceIoRead(0x%08x)\n", iRet));			
		}
		Asura_OutputToDebugger(("** READ FAILED. RETRY...: %s(0x%08x) - %s\n", __FUNCTION__, iRet, szFilePath));		

		sceKernelDelayThread(UMD_RETRY_DELAY);

		if (iRet == (int)SCE_UMD_NO_MEDIUM ||
			iRet == (int)SCE_UMD_NOT_DEVICE_READY ||
			sceUmdGetDriveStat() & SCE_UMD_MEDIA_OUT)
		{
			Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));			
			if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_MEDIA_IN, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
			{
				break;
			}
		}
		if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_READABLE, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
		{
			break;
		}
	}	/*E Retry GetStat While */

	return iRet;
}

int	Asura_PSP_FileIO::GetFileSize(const char* szFilePath)
{
	return tgIoGetFileSize(szFilePath);
}

int	Asura_PSP_FileIO::Read(SceUID iFile, void *pvReadBuffer, SceSize uReadSize, const bool bAllowRetry)
{
	int iRet;
	
	++g_iPendingReadOperations; 

	/*E Retry Read */
	while (1)
	{
		// (MH - 25.06.2007) we were doing the sceUmdWaitDriveStatCB wait for Drive Readable here
		// then locking the semaphore, then doing the read but the problem is that the WaitSemaphore 
		// could stop us until the drive was no longer readable. So I've changed it to lock the semaphore 
		// first and then check the drive is readable so we know that it won't stop being readable 
		// due to power off stuff.
		if (iRet = WaitSemaphoreAndDriveReady(bAllowRetry, __LINE__) < 0)
		{
			break;
		}

		Asura_Assert(!g_bSuspended, "fileIO suspended but we still got a lock");

		// NY - 16 August 2006 - set the first 4 bytes of buffer to a known value
		if (uReadSize >= 4)
		{
			static_cast<u_int*>(pvReadBuffer)[0] = 0xF00D4A11u;
		}
		iRet					= sceIoRead(iFile, pvReadBuffer, uReadSize);
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();

		if(bAllowRetry)
		{
			// NY - 16 August 2006 - kernel claims to have read data in, but we know differently!
			if (uReadSize >= 4 && iRet > 0 && static_cast<u_int*>(pvReadBuffer)[0] == 0xF00D4A11u )
			{									    
				//assume we need to rewind.....
				sceIoLseek(iFile, -iRet, SCE_SEEK_CUR);
			}

			SignalSemaphore();

			// NY - 16 August 2006 - stay in loop if read silently failled.
			if (iRet >= 0 && (static_cast<u_int*>(pvReadBuffer)[0] != 0xF00D4A11u || uReadSize < 4 || iRet == 0))
			{
				break;
			}

			if (iRet == (int)SCE_ERROR_ERRNO_EBADF  ||
				iRet == (int)SCE_ERROR_ERRNO_EINVAL ||
				iRet == (int)SCE_ERROR_ERRNO_ENODEV ||
				iRet == (int)SCE_ERROR_ERRNO_EFSIZE ||
				iRet == (int)SCE_ERROR_ERRNO_EFAULT ||
				iRet == (int)SCE_UMD_LBA_OUTOFRANGE ||
				iRet == (int)SCE_ERROR_ERRNO_NOMAIN_ADDR ||
				iRet == (int)SCE_ERROR_ERRNO_ENAMETOOLONG)
			{
				Asura_OutputToDebugger(("*** FATAL ERROR: sceIoRead(0x%08x)\n", iRet));
			}

			Asura_OutputToDebugger(("** READ FAILED. RETRY...: %s(0x%08x)\n", __FUNCTION__, iRet));

			sceKernelDelayThread(UMD_RETRY_DELAY);

			if (iRet == (int)SCE_UMD_NO_MEDIUM ||
				iRet == (int)SCE_UMD_NOT_DEVICE_READY ||
				sceUmdGetDriveStat() & SCE_UMD_MEDIA_OUT)
			{
				Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));
				if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_MEDIA_IN, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
				{
					break;
				}
			}
			if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_READABLE, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
			{
				break;
			}
		}
		else
		{
			if (uReadSize >= 4 && iRet > 0 && static_cast<u_int*>(pvReadBuffer)[0] == 0xF00D4A11u )
			{									    
				// Read failed for some reason...
				iRet	= SCE_ERROR_ERRNO_EBADF;
			}

			SignalSemaphore();
			break;
		}

	}	/*E Retry Read While */

	if (iRet < 0)
	{
		Asura_Assert(false, "Asura_PSP_FileIO::Read failed");
	}

	if (g_bSuspended)
	{
		Asura_OutputToDebugger(("** Asura_PSP_FileIO::Read now suspended\n"));
        // (MH - 15.08.2006) if fileIO is suspended we had better wait until it resumes.
		// This is so that if the calling function is about to copy from the buffer we just
		// read into to volatile memory, but volatile memory is now unavailable, it waits 
		// until it's available again.
		WaitSemaphore(__LINE__);
		Asura_Assert(!g_bSuspended, "waiting for semaphore didn't wait for resume");
		SignalSemaphore();
		Asura_OutputToDebugger(("** Asura_PSP_FileIO::Read no longer suspended\n"));
	}

	--g_iPendingReadOperations; 

	return iRet;
}

SceOff Asura_PSP_FileIO::Lseek(SceUID iFile, SceOff offset, int iWhence)
{
	int iRet;
	SceOff pointer;

	if ((iRet = WaitSemaphore(__LINE__)) < 0)
	{
		return (SceOff)iRet;
	}

	if ((pointer = sceIoLseek(iFile, offset, iWhence)) < 0)
	{
		SignalSemaphore();
		return pointer;
	}

	if ((iRet = SignalSemaphore()) < 0)
	{
		return (SceOff)iRet;
	}
	
	s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
	return pointer;
}

int	Asura_PSP_FileIO::Close(SceUID iFile)
{
// (MP - 21.07.2006) Unnecessary - we're not doing any asynchronous
// operations. Under 2.71 libraries or above, this causes a warning to be
// output to the debugger.
//	SceIores xDummy;
//	sceIoWaitAsync( iFile , &xDummy );
	
	// JG - 17/08/06: Sometimes we were closing files with ongoing read operations
	// Even worse, the handle was getting messed up and this close operation was affecting to the wrong file
	// This fix works around the fact that we don't now for sure which handle/thread is going to be closed, so to be sure
	// we wait until ALL reading operations have finished
	int iRet;
	if ((iRet = WaitSemaphore(__LINE__)) < 0)
	{
		return (SceOff)iRet;
	}
	
	while (g_iPendingReadOperations)
	{
		// There is an reading operation going, release the semaphore to allow it to complete
		SignalSemaphore(); 

		sceKernelDelayThreadCB(1000); 

		// Grab again the semaphore
		WaitSemaphore(__LINE__); 
	}

	iRet = sceIoClose(iFile);
	Asura_Assertf((iRet >= 0), ("sceIoClose(): Failed %08X", iRet));
	SignalSemaphore();
	return(iRet);
}

SceUID Asura_PSP_FileIO::LoadModule(const char *szFileName, int iFlag, const SceKernelLMOption *pxOpt)
{
	int iRet;

	/*E Retry LoadModule */
	while (1)
	{
		// (MH - 25.06.2007) combined WaitSemaphore and drive ready check.
		// Some drive ready code below may be unnecessary now but it 
		// shouldn't hurt to leave it in.
		bool bAllowRetry = true;
		if ((iRet = WaitSemaphoreAndDriveReady(bAllowRetry, __LINE__)) < 0)
		{
			break;
		}

		iRet					= sceKernelLoadModule(szFileName, iFlag, pxOpt);
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();

		SignalSemaphore();

		if (iRet >= 0)
		{
			break;
		}

		if (iRet == (int)SCE_KERNEL_ERROR_NO_MEMORY ||
			iRet == (int)SCE_KERNEL_ERROR_ILLEGAL_ADDR ||
			iRet == (int)SCE_KERNEL_ERROR_ILLEGAL_OBJECT ||
			iRet == (int)SCE_KERNEL_ERROR_ILLEGAL_CONTEXT ||
			iRet == (int)SCE_KERNEL_ERROR_MODULE_MGR_BUSY ||
			iRet == (int)SCE_ERROR_ERRNO_EBADF  ||
			iRet == (int)SCE_ERROR_ERRNO_EUNIT  ||
			iRet == (int)SCE_ERROR_ERRNO_ENODEV ||
			iRet == (int)SCE_ERROR_ERRNO_ENOMEM ||
			iRet == (int)SCE_ERROR_ERRNO_EINVAL ||
			iRet == (int)SCE_ERROR_ERRNO_EFSIZE ||
			iRet == (int)SCE_ERROR_ERRNO_EFAULT ||
			iRet == (int)SCE_UMD_LBA_OUTOFRANGE ||
			iRet == (int)SCE_ERROR_ERRNO_NOMAIN_ADDR ||
			iRet == (int)SCE_ERROR_ERRNO_ENAMETOOLONG)
		{
			Asura_OutputToDebugger(("*** FATAL ERROR: sceKernelLoadModule(0x%08x)\n", iRet));
		}

		Asura_OutputToDebugger(("** READ FAILED. RETRY...: %s(0x%08x) - %s\n", __FUNCTION__, iRet, szFileName));

		sceKernelDelayThread(UMD_RETRY_DELAY);

		if (iRet == (int)SCE_UMD_NO_MEDIUM ||
			iRet == (int)SCE_UMD_NOT_DEVICE_READY ||
			sceUmdGetDriveStat() & SCE_UMD_MEDIA_OUT) {

			Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));
			if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_MEDIA_IN, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
			{
				break;
			}
		}
		if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_READABLE, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
		{
			break;
		}
	}	/*E Retry LoadModule While */

	return iRet;
}

SceUID Asura_PSP_FileIO::OpenDir(const char* szDirName, const bool bAllowRetry)
{
	int iRet;

	/*E Retry Open */
	while (1)
	{
		// (MH - 25.06.2007) combined WaitSemaphore and drive ready check.
		// Some drive ready code below may be unnecessary now but it 
		// shouldn't hurt to leave it in.
		if ((iRet = WaitSemaphoreAndDriveReady(bAllowRetry, __LINE__)) < 0)
		{
			break;
		}

//		Asura_OutputToDebugger(("* OPENING: %s(%d) FILE: %s\n", __FUNCTION__, __LINE__, szDirName));
		iRet = sceIoDopen(szDirName);

		if(bAllowRetry)
		{
		SignalSemaphore();

		if (iRet >= 0)
		{
			break;
		}

		if (iRet == (int)SCE_ERROR_ERRNO_ENOENT ||
			iRet == (int)SCE_ERROR_ERRNO_EUNIT  ||
			iRet == (int)SCE_ERROR_ERRNO_ENODEV ||
			iRet == (int)SCE_ERROR_ERRNO_ENOMEM ||
			iRet == (int)SCE_ERROR_ERRNO_EINVAL ||
			iRet == (int)SCE_ERROR_ERRNO_ENAMETOOLONG)
		{
				Asura_OutputToDebugger(("*** FATAL ERROR: sceIoDopen(0x%08x) (on file %s)\n", iRet, szDirName));
		}

		// KOR - 2/3/2006 - flash GPO light 7 to indicate we are waiting on a file.
		static int iLight = 0;
		iLight = !iLight;
		Asura_PSP_GPIO::SetGPO((iLight << 7));
			Asura_OutputToDebugger(("** READ FAILED. RETRY...: %s(0x%08x) - %s\n", __FUNCTION__, iRet, szDirName));

		sceKernelDelayThread(UMD_RETRY_DELAY);

		if (iRet == (int)SCE_UMD_NO_MEDIUM ||
			iRet == (int)SCE_UMD_NOT_DEVICE_READY ||
			sceUmdGetDriveStat() & SCE_UMD_MEDIA_OUT)
		{
			Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));
			if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_MEDIA_IN, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
			{
				break;
			}
		}
		if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_READABLE, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
		{
			break;
		}
		}
		else
		{
			if(iRet < 0)
			{
				// Read failed for some reason...
				iRet	= SCE_ERROR_ERRNO_EBADF;
			}

			SignalSemaphore();
			break;
		}
	}	 /*E Retry Open While */

//	Asura_OutputToDebugger(("* OPEN FINISHED - ID: %u\n", iRet));
	return iRet;
}

int Asura_PSP_FileIO::ReadDir(SceUID fd, struct SceIoDirent* pxContents, const bool bAllowRetry)
{
	int iRet;

	++g_iPendingReadOperations; 

#if SCE_DEVKIT_VERSION < 0x03080000
	static SceFatMsDirentPrivate ls_xPrivateData;
#else
	static SceFatMsDirent ls_xPrivateData;
#endif

	memset(pxContents, 0, sizeof(SceIoDirent));
	pxContents->d_private = &ls_xPrivateData;


	/*E Retry Read */
	while (1)
	{
		// (MH - 25.06.2007) combined WaitSemaphore and drive ready check.
		// Some drive ready code below may be unnecessary now but it 
		// shouldn't hurt to leave it in.
		if ((iRet = WaitSemaphoreAndDriveReady(bAllowRetry, __LINE__)) < 0)
		{
			break;
		}

		Asura_Assert(!g_bSuspended, "fileIO suspended but we still got a lock");

		iRet					= sceIoDread(fd, pxContents);
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();

		if(bAllowRetry)
		{
		SignalSemaphore();

		if (iRet >= 0)
		{
			break;
		}

		if (iRet == (int)SCE_ERROR_ERRNO_EBADF  ||
			iRet == (int)SCE_ERROR_ERRNO_EINVAL ||
			iRet == (int)SCE_ERROR_ERRNO_ENODEV ||
			iRet == (int)SCE_ERROR_ERRNO_EFSIZE ||
			iRet == (int)SCE_ERROR_ERRNO_EFAULT ||
			iRet == (int)SCE_UMD_LBA_OUTOFRANGE ||
			iRet == (int)SCE_ERROR_ERRNO_NOMAIN_ADDR ||
			iRet == (int)SCE_ERROR_ERRNO_ENAMETOOLONG)
		{
			Asura_OutputToDebugger(("*** FATAL ERROR: sceIoDread(0x%08x)\n", iRet));
		}

		Asura_OutputToDebugger(("** READ FAILED. RETRY...: %s(0x%08x)\n", __FUNCTION__, iRet));

		sceKernelDelayThread(UMD_RETRY_DELAY);

		if (iRet == (int)SCE_UMD_NO_MEDIUM ||
			iRet == (int)SCE_UMD_NOT_DEVICE_READY ||
			sceUmdGetDriveStat() & SCE_UMD_MEDIA_OUT)
		{
			Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));
			if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_MEDIA_IN, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
			{
				break;
			}
		}
		if ((iRet = sceUmdWaitDriveStatCB(SCE_UMD_READABLE, ASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME)) < 0)
		{
			break;
		}
		}
		else
		{	
			if(iRet < 0)
			{
				// Read failed for some reason...
				iRet	= SCE_ERROR_ERRNO_EBADF;
			}

			SignalSemaphore();
			break;
		}

	}	/*E Retry Read While */

	if (iRet < 0)
	{
		Asura_OutputToDebugger(("Asura_PSP_FileIO::ReadDir failed\n"));
	}

	if (g_bSuspended)
	{
		Asura_OutputToDebugger(("** Asura_PSP_FileIO::ReadDir now suspended\n"));
		// (MH - 15.08.2006) if fileIO is suspended we had better wait until it resumes.
		// This is so that if the calling function is about to copy from the buffer we just
		// read into to volatile memory, but volatile memory is now unavailable, it waits 
		// until it's available again.
		WaitSemaphore(__LINE__);
		Asura_Assert(!g_bSuspended, "waiting for semaphore didn't wait for resume");
		SignalSemaphore();
		Asura_OutputToDebugger(("** Asura_PSP_FileIO::ReadDir no longer suspended\n"));
	}

	--g_iPendingReadOperations; 

	return iRet;
}

int Asura_PSP_FileIO::CloseDir(SceUID fd)
{
	// (MP - 21.07.2006) Unnecessary - we're not doing any asynchronous
	// operations. Under 2.71 libraries or above, this causes a warning to be
	// output to the debugger.
	//	SceIores xDummy;
	//	sceIoWaitAsync( iFile , &xDummy );

//	Asura_OutputToDebugger(("* CLOSING: %s(%d) FILE: %i\n", __FUNCTION__, __LINE__, fd));

	// JG - 17/08/06: Sometimes we were closing files with ongoing read operations
	// Even worse, the handle was getting messed up and this close operation was affecting to the wrong file
	// This fix works around the fact that we don't now for sure which handle/thread is going to be closed, so to be sure
	// we wait until ALL reading operations have finished
	int iRet;
	if ((iRet = WaitSemaphore(__LINE__)) < 0)
	{
		return (SceOff)iRet;
	}

	while (g_iPendingReadOperations)
	{
		// There is an reading operation going, release the semaphore to allow it to complete
		SignalSemaphore(); 

		sceKernelDelayThreadCB(1000); 

		// Grab again the semaphore
		WaitSemaphore(__LINE__); 
	}

	sceIoDclose(fd);

	if ((iRet = SignalSemaphore()) < 0)
	{
		return (SceOff)iRet;
	}

	return SCE_KERNEL_ERROR_OK;
}

#else//defined ASURA_PSP_UMD && !defined(ASURA_PSP_ARTIST_BUILD)

bool Asura_PSP_FileIO::DiscExists(void)
{
	return(true);
}

bool Asura_PSP_FileIO::FileExists (const char* szFilePath)
{
	int	iFileHandle	= Asura_PSP_FileIO::Open(szFilePath, SCE_O_RDONLY, 0777);

	// Did we manage to open the file?
	if(iFileHandle >= 0)
	{
		Asura_PSP_FileIO::Close(iFileHandle);
		return(true);
	}
	else
	{
		return(false);
	}
}

int	Asura_PSP_FileIO::Open(const char *szFilepath, int iFlag, SceMode iMode, const bool bAllowRetry)
{
	int	iResult	= sceIoOpen(szFilepath, iFlag, iMode);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		return(iResult);
	}
}

int	Asura_PSP_FileIO::Getstat(const char* szFilePath, SceIoStat* pxStat)
{
	int iResult	= sceIoGetstat(szFilePath, pxStat);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		return(iResult);
	}
}

int	Asura_PSP_FileIO::GetFileSize(const char* szFilePath)
{
	SceIoStat xFileStat;
	int iError = sceIoGetstat(szFilePath, &xFileStat);
	Asura_Assertf(iError >= 0, ("Error %x returned from sceIoGetstat for file %s", iError, szFilePath));
	return (iError >= 0) ? xFileStat.st_size : 0;
}

int	Asura_PSP_FileIO::Read(SceUID iFile, void *pvReadBuffer, SceSize uReadSize, const bool bAllowRetry)
{
	int	iResult	= sceIoRead(iFile, pvReadBuffer, uReadSize);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		Asura_Assertf(false, ("Asura_PSP_FileIO::Read(): Failed %08X", iResult));
		return(0);
	}
}

SceOff Asura_PSP_FileIO::Lseek(SceUID iFile, SceOff offset, int iWhence)
{
	int	iResult	= sceIoLseek(iFile, offset, iWhence);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		Asura_Assertf(false, ("Asura_PSP_FileIO::Lseek(): Failed %08X", iResult));
		return(0);
	}
}

int	Asura_PSP_FileIO::Close(SceUID iFile)
{
// (MP - 21.07.2006) Unnecessary - we're not doing any asynchronous
// operations. Under 2.71 libraries or above, this causes a warning to be
// output to the debugger.
//	SceIores xDummy;
//	sceIoWaitAsync( iFile , &xDummy );
	int iResult	= sceIoClose(iFile);

	Asura_Assertf((iResult == SCE_KERNEL_ERROR_OK), ("Asura_PSP_FileIO::Close(): Failed %08X", iResult));
	return(iResult);
}

SceUID Asura_PSP_FileIO::LoadModule(const char *szFileName, int iFlag, const SceKernelLMOption *pxOpt)
{
	int	iResult	= sceKernelLoadModule(szFileName, iFlag, pxOpt);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		Asura_Assertf(false, ("Asura_PSP_FileIO::LoadModule(): Failed %08X", iResult));
		return(iResult);
	}
}

SceUID Asura_PSP_FileIO::OpenDir(const char* szDirName, const bool bAllowRetry)
{
	int	iResult	= sceIoDopen(szDirName);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		return(iResult);
	}
}

int Asura_PSP_FileIO::ReadDir(SceUID fd, struct SceIoDirent* pxContents, const bool bAllowRetry)
{
#if SCE_DEVKIT_VERSION < 0x03080000
	static SceFatMsDirentPrivate ls_xPrivateData;
#else
	static SceFatMsDirent ls_xPrivateData;
	ls_xPrivateData.size = sizeof(SceFatMsDirent);
#endif

	memset(pxContents, 0, sizeof(SceIoDirent));
	pxContents->d_private = &ls_xPrivateData;
	int	iResult	= sceIoDread(fd, pxContents);

	if(iResult >= 0)
	{
		s_xLastOperationTicks	= Asura_Timers::GetCurTicks();
		return(iResult);
	}
	else
	{
		Asura_Assertf(false, ("Asura_PSP_FileIO::Lseek(): ReadDir %08X", iResult));
		return(iResult);
	}
}

int Asura_PSP_FileIO::CloseDir(SceUID fd)
{
	int iResult	= sceIoDclose(fd);

	Asura_Assertf((iResult == SCE_KERNEL_ERROR_OK), ("Asura_PSP_FileIO::Close(): Failed %08X", iResult));
	return(iResult);
}

#endif//else defined ASURA_PSP_UMD && !defined(ASURA_PSP_ARTIST_BUILD)

int Asura_PSP_FileIO::LoadStartModule(SceUID *piModuleID, const char *szFileName, const SceKernelLMOption *pxOpt) 
{
	int iRes;
	int iModuleResult;
	char szTotalFilename[256]; 

	// Slightly different from GetFullPath - modules cannot be loaded from memory stick
#if defined (ASURA_PSP_UMD)
	// Prepend disc path to the filename
#ifdef ASURA_PSP_DEMO
	Asura_PrintfToBfr( szTotalFilename, "disc0:/PSP_GAME/USRDIR/%s%s", (sizeof szTotalFilename - 1), true, ASURA_PSP_DEMO_PATH, szFileName );
#else
	Asura_PrintfToBfr( szTotalFilename, "disc0:/PSP_GAME/USRDIR/%s", (sizeof szTotalFilename - 1), true, szFileName );
#endif

	for (int i = 0; szTotalFilename[i]; i++)
	{
		szTotalFilename[i] = (szTotalFilename[i] == '\\') ? '/' : szTotalFilename[i];
	}
#else
	Asura_PrintfToBfr( szTotalFilename,"host0:%s", (sizeof szTotalFilename - 1), true, szFileName );
#endif

	strupr(szTotalFilename);

	if ((*piModuleID = LoadModule(szTotalFilename, 0, pxOpt)) < 0)
	{
		Asura_OutputToDebugger(("*** ERROR: sceKernelLoadModule(0x%08x)\n", *piModuleID));
		Asura_OutputToDebugger(("    Filename: '%s'\n",szTotalFilename));
		return *piModuleID;
	}

	if ((iRes = sceKernelStartModule(*piModuleID, 0, 0, &iModuleResult, NULL)) < 0)
	{
		Asura_OutputToDebugger(("*** ERROR: sceKernelStartModule(0x%08x): 0x%08x\n",
			   *piModuleID,iRes));
		Asura_OutputToDebugger(("    Filename: '%s'\n",szTotalFilename));
		return iRes;
	}

	return iRes;
}

int Asura_PSP_FileIO::UnloadStopModule(SceUID* piModuleID)
{
	int iRes;
	int iModuleResult;

	if ((iRes = sceKernelStopModule(*piModuleID, 0, 0, &iModuleResult, NULL)) < 0)
	{
		Asura_OutputToDebugger(("*** ERROR: sceKernelStopModule(0x%08x): 0x%08x\n", *piModuleID, iRes));
		return iRes;
	}

	if ((iRes = sceKernelUnloadModule(*piModuleID)) < 0)
	{
		Asura_OutputToDebugger(("*** ERROR: sceKernelUnloadModule(0x%08x): 0x%08x\n", *piModuleID, iRes));
		return iRes;
	}

	return iRes;
}

void Asura_PSP_FileIO::Suspend(void)
{
	if (SCE_KERNEL_ERROR_SEMA_ZERO == (unsigned)PollSemaphore())
	{
		Asura_OutputToDebugger( ("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__) );
		WaitSemaphore(__LINE__);
	}
	g_bSuspended = true;
}

void Asura_PSP_FileIO::Resume(void)
{
	g_bSuspended = false;
	SignalSemaphore();
}

bool Asura_PSP_FileIO::IsSuspended(void)
{
	return g_bSuspended;
}

int	Asura_PSP_FileIO::PollSemaphore()
{
	int iRet = sceKernelPollSema(s_iSemaphoreId, 1);
	Asura_OutputToDebuggerAssert(iRet >= 0 || iRet == SCE_KERNEL_ERROR_SEMA_ZERO, ("@@@ERROR@@@ %s: iRet = %d\n", __FUNCTION__, iRet));
	return iRet;
}

#if defined ASURA_PSP_UMD && !defined(ASURA_PSP_ARTIST_BUILD)
// WaitSemaphoreAndDriveReady locks the semaphore and checks the UMD drive is readable.
// The semaphore is unlocked if the wait for the drive fails. Return value is SCE_KERNEL_ERROR_OK
// (0) for success or < 0 on failure.
int Asura_PSP_FileIO::WaitSemaphoreAndDriveReady( bool bAllowRetry, int iLine )
{
	// Lock the semaphore first and then check the drive is readable so we 
	// know that it won't stop being readable due to power off stuff.
	int iRet;

	if ((iRet = WaitSemaphore(iLine)) < 0)
	{
		return iRet;
	}

	if (!(sceUmdGetDriveStat() & SCE_UMD_READABLE))
	{
		if(bAllowRetry)
		{
			// sceUmdWaitDriveStatCB was being called with no timeout which would cause a deadlock if 
			// the poweroff thread was waiting for the semaphore. So now we only wait for a second, and
			// if it's still not readable, release the semaphore, sleep for a second, and then reacquire
			// the semaphore until the drive becomes readable again.
			bool bContinue = true;
			Asura_OutputToDebugger(("** WAITING: %s(%d)\n", __FUNCTION__, __LINE__));
			while (1)
			{
				iRet	= sceUmdWaitDriveStatCB(SCE_UMD_READABLE, iASURA_PSP_UMD_WAIT_DRIVE_STAT_TIME_WITH_TIMEOUT );
				if (iRet == SCE_KERNEL_ERROR_OK)
				{
					Asura_OutputToDebugger(("** sceUmdWaitDriveStatCB OK: %s(%d)\n", __FUNCTION__, __LINE__));
					break;
				}
				else if (iRet == SCE_KERNEL_ERROR_WAIT_TIMEOUT )
				{
					//Asura_OutputToDebugger(("** sceUmdWaitDriveStatCB timed out, will retry: %s(%d)\n", __FUNCTION__, __LINE__));
					SignalSemaphore();
					sceKernelDelayThread(UMD_RETRY_DELAY);
					WaitSemaphore(iLine);
				}
				else
				{
					Asura_OutputToDebugger(("** sceUmdWaitDriveStatCB gave up (%x): %s(%d)\n", iRet, __FUNCTION__, __LINE__));
					SignalSemaphore();
					break;
				}

			}
		}
		else
		{
			Asura_OutputToDebugger(("** NO DISK: %s(%d)\n", __FUNCTION__, __LINE__));
			iRet	= SCE_UMD_NO_MEDIUM;
			SignalSemaphore();
		}
	}

	return iRet;
}
#endif // #if defined ASURA_PSP_UMD && !defined(ASURA_PSP_ARTIST_BUILD)

int	Asura_PSP_FileIO::WaitSemaphore(int iLine)
{
	int iRet = sceKernelWaitSema(s_iSemaphoreId, 1, NULL);
	Asura_OutputToDebuggerAssert(iRet >= 0, ("@@@ERROR@@@ %s: iRet = %d\n", __FUNCTION__, iRet));
	if (iRet >= 0)
	{
		g_iFileIOSemaLine = iLine;
	}
	return iRet;
}

int	Asura_PSP_FileIO::WaitSemaphoreCB()
{
	int iRet = sceKernelWaitSemaCB(s_iSemaphoreId, 1, NULL);
	Asura_OutputToDebuggerAssert(iRet >= 0, ("@@@ERROR@@@ %s: iRet = %d\n", __FUNCTION__, iRet));
	return iRet;
}

int	Asura_PSP_FileIO::SignalSemaphore()
{
	g_iFileIOSemaLine = 0;
	int iRet = sceKernelSignalSema(s_iSemaphoreId, 1);
	Asura_OutputToDebuggerAssert(iRet >= 0, ("@@@ERROR@@@ %s: iRet = %d\n", __FUNCTION__, iRet));
	return iRet;
}

void Asura_PSP_FileIO::GetFullPath (const char* szFilename,
									const u_int uMaxPathLength,
									char* szFullPath, 
									bool bForceMemStickIfUMD /* = false */)
{
	// may need to add "/PSP_GAME/USRDIR/" + filename so allow 17 characters less in UMD case.
#if defined (ASURA_PSP_ARTIST_BUILD)
	Asura_PrintfToBfr(szFullPath, "ms0:%s", uMaxPathLength, true, szFilename);
#elif defined (ASURA_PSP_UMD)
	// Prepend disc (or memory stick) path to the filename

#ifdef ASURA_PSP_DEMO
	const char* szPath = "disc0:/PSP_GAME/USRDIR/" ASURA_PSP_DEMO_PATH;
#else
	const char* szPath = "disc0:/PSP_GAME/USRDIR/";
#endif

	if( bForceMemStickIfUMD )	
	{
		szPath = "ms0:"; 
	}

	Asura_PrintfToBfr(szFullPath, "%s%s", uMaxPathLength, true, szPath, szFilename);
	for (int i = 0; szFullPath[i]; i++)
	{
		szFullPath[i] = (szFullPath[i] == '\\') ? '/' : szFullPath[i];
	}
#else
	Asura_PrintfToBfr(szFullPath, "host0:%s", uMaxPathLength, true, szFilename);
#endif
	strupr(szFullPath);
}

void Asura_PSP_FileIO::GetFullPathForFMV (const char* szFilename,
										  const u_int uMaxPathLength,
										  char* szFullPath)
{
#if !defined(ASURA_PSP_UMD)
	if (g_bForceFMVOnMS)
	{
		Asura_PrintfToBfr(szFullPath, "ms0:%s", uMaxPathLength, true, szFilename);
		strupr(szFullPath);
	}
	else
#endif	// !defined(ASURA_PSP_UMD)
	{
		GetFullPath(szFilename, uMaxPathLength, szFullPath);
	}
}


void Asura_PSP_FileIO::SetReadFMVFromMemStick( bool bTF)
{
#if !defined(ASURA_PSP_UMD)
	g_bForceFMVOnMS = bTF;
#endif
}


#if defined(ASURA_PSP_UMD) && defined(ASURA_PSP_KEEP_UMD_SPINNING)

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_FileIO::KeepUMDSpinning
// Author       : Andy Tate
// Date         : 23-Jun-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_FileIO::KeepUMDSpinning (void)
{
	Asura_OutputToDebugger(("Asura_PSP_FileIO::KeepUMDSpinning()\n"));

#if 1
	u_int	uLBA;

	// Change the disc head position (as little as possible)...
	sceIoDevctl("umd0:", SCE_UMD_GET_LASTLBA, NULL, 0, &uLBA, sizeof(uLBA));
	if(uLBA > 10)
	{
		uLBA	-= 10;
	}
	else
	{
		uLBA	+= 10;
	}
	sceIoDevctl("umd0:", SCE_UMD_SEEK_DISC, &uLBA, sizeof(uLBA), NULL, 0);
#else
	// Activate the disc...
	sceIoDevctl("umd0:", SCE_UMD_ACTIVATE_DRIVE, NULL, 0, NULL, 0);
#endif
}

/// End of function 'Asura_PSP_FileIO::KeepUMDSpinning'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_FileIO::KeepUMDSpinningThread
// Author       : Andy Tate
// Date         : 23-Jun-2006
////////////////////////////////////////////////////////////////////////////////

int Asura_PSP_FileIO::KeepUMDSpinningThread (SceSize args, void *argp)
{
// RDM - 17/08/06 - don't want fp exceptions at all, in any threads
#if !defined ASURA_ENABLE_FLOATING_POINT_EXCEPTIONS \
	|| defined ASURA_PSP_ARTIST_BUILD \
	|| defined ASURA_PSP_SONY_SUBMISSION

	// Disable fp exceptions if not requested
	__asm__ volatile("ctc1 %0, $31" : : "r" (0));
#endif

	for(;;)
	{
#if 1
		const Asura_TickCount	xTicksSinceLastOperation	= Asura_Timers::GetElapsedTicks(s_xLastOperationTicks);
		const float				fSecondsSinceLastOperation	= Asura_Timers::TicksToSeconds(xTicksSinceLastOperation);

		// Do we need to "prod" the disc?
		if(fSecondsSinceLastOperation >= 2.0f)
		{
			KeepUMDSpinning();
		}
#else
		KeepUMDSpinning();
#endif

		// Sleep for a while...
		sceKernelDelayThread(KEEP_UMD_SPINNING_DELAY);
	}

	sceKernelExitThread(SCE_KERNEL_EXIT_SUCCESS);
	return SCE_KERNEL_EXIT_SUCCESS;
}

/// End of function 'Asura_PSP_FileIO::KeepUMDSpinningThread'. ///

#endif	// defined(ASURA_PSP_UMD) && defined(ASURA_PSP_KEEP_UMD_SPINNING)


/// End of file. ///
