//////////////////////////////////////////////////////////////////////////
// tgIoOpen.c
// PSP(TM) Programmer Tool Runtime Library Release 2.6.0
// cached high-speed sector based sceIoOpen() replacement function
//
// Copyright (C) 2004-2005 Sony Computer Entertainment Europe
// All Rights Reserved.
//
// Usage:
//
// 1. if you have your own memory management system, you need to use 
// tgIoSetMemoryManager() to let the cache function know where to get 
// the memory from, otherwise it will just use malloc() and free().
// 
// 2. after you have activated the UMD drive, given it a name and 
// waited for it to be in SCE_UMD_READABLE state, use tgIoFillCache();
// to fill the cache, usually tgIoFillCache("disc0:/");
// 
// 3. now just replace sceIoOpen() calls by tgIoOpen() calls (using the
// same parameters and everything).
// 
// NOTE: you will have to pass tgIoOpen() the FULL pathname to the 
// files from where you have cached the directory.
// i.e. if you have used "disc0:/", you will have to call it like this:
// tgIoOpen("disc0:/PSP_GAME/USR_DIR/MYDATA.DAT", SCE_O_RDONLY, 0);
// that means you can't set a current working dir or use relative paths
// _after_ you have cached a directory.
//
// However, you can do the following:
//
// sceIoChdir("disc0:/PSP_GAME/USRDIR/");
// tgIoFillCache("/");
//
// and then access files like this:
//
// tgIoOpen("/MYDATA.DAT", SCE_O_RDONLY, 0);
//
// Also be aware that the caching is done recursively, so calling it 
// from a thread with a small stack isn't a good idea.
// 
// If the caching does not print any warnings about collisions  in the 
// TTY output, you can remove the TGIO_CACHE_STORE_FILENAME define 
// from the .h file and you will save memory as it's not storing the
// filenames.
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// includes
//////////////////////////////////////////////////////////////////////////
#include <psptypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iofilemgr.h>
#include <iofilemgr_stat.h>
#include <iofilemgr_dirent.h>

#include "tgIoOpen.h"
//#include "tgIoPA.h"

#include "FileAccess/Asura_PSP_FileIO.h"


//////////////////////////////////////////////////////////////////////////
// static variables
//////////////////////////////////////////////////////////////////////////
static TGIO_CACHE *tgIoCurrentCache = NULL;

#ifndef __MWERKS__
static void (*tgio_sort)( TGIO_CACHE *cache ) = tgIoSortCache;
#endif

//////////////////////////////////////////////////////////////////////////
//
// the hashing code below was taken from:
//
// http://burtleburtle.net/bob/hash/doobs.html
//
//////////////////////////////////////////////////////////////////////////
typedef  unsigned int  ub4;   /* unsigned 4-byte quantities */
typedef  unsigned char ub1;   /* unsigned 1-byte quantities */

#define hashsize(n) ((ub4)1<<(n))
#define hashmask(n) (hashsize(n)-1)

/*
--------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bits set, and the deltas of all three
high bits or all three low bits, whether the original value of a,b,c
is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() was built out of 36 single-cycle latency instructions in a 
structure that could supported 2x parallelism, like so:
a -= b; 
a -= c; x = (c>>13);
b -= c; a ^= x;
b -= a; x = (a<<8);
c -= a; b ^= x;
c -= b; x = (b>>13);
...
Unfortunately, superscalar Pentiums and Sparcs can't take advantage 
of that parallelism.  They've also turned some of those single-cycle
latency instructions into multi-cycle latency instructions.  Still,
this is the fastest good hash I could find.  There were about 2^^68
to choose from.  I only looked at a billion or so.
--------------------------------------------------------------------
*/
#define mix(a,b,c) \
{ \
	a -= b; a -= c; a ^= (c>>13); \
	b -= c; b -= a; b ^= (a<<8); \
	c -= a; c -= b; c ^= (b>>13); \
	a -= b; a -= c; a ^= (c>>12);  \
	b -= c; b -= a; b ^= (a<<16); \
	c -= a; c -= b; c ^= (b>>5); \
	a -= b; a -= c; a ^= (c>>3);  \
	b -= c; b -= a; b ^= (a<<10); \
	c -= a; c -= b; c ^= (b>>15); \
}

/*
--------------------------------------------------------------------
hash() -- hash a variable-length key into a 32-bit value
k       : the key (the unaligned variable-length array of bytes)
len     : the length of the key, counting by bytes
initval : can be any 4-byte value
Returns a 32-bit value.  Every bit of the key affects every bit of
the return value.  Every 1-bit and 2-bit delta achieves avalanche.
About 6*len+35 instructions.

The best hash table sizes are powers of 2.  There is no need to do
mod a prime (mod is sooo slow!).  If you need less than 32 bits,
use a bitmask.  For example, if you need only 10 bits, do
h = (h & hashmask(10));
In which case, the hash table should have hashsize(10) elements.

If you are hashing n strings (ub1 **)k, do it like this:
for (i=0, h=0; i<n; ++i) h = hash( k[i], len[i], h);

By Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net.  You may use this
code any way you wish, private, educational, or commercial.  It's free.

See http://burtleburtle.net/bob/hash/evahash.html
Use for hash table lookup, or anything where one collision in 2^^32 is
acceptable.  Do NOT use for cryptographic purposes.
--------------------------------------------------------------------
*/

static ub4 hash( register ub1 *k, register ub4 length, register ub4 initval )
//register ub1 *k;        /* the key */
//register ub4  length;   /* the length of the key */
//register ub4  initval;  /* the previous hash, or an arbitrary value */
{
	register ub4 a,b,c,len;

	/* Set up the internal state */
	len = length;
	a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
	c = initval;         /* the previous hash value */

	/*---------------------------------------- handle most of the key */
	while (len >= 12)
	{
		a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
		b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
		c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
		mix(a,b,c);
		k += 12; len -= 12;
	}

	/*------------------------------------- handle the last 11 bytes */
	c += length;
	switch(len)              /* all the case statements fall through */
	{
		case 11: c+=((ub4)k[10]<<24);
		case 10: c+=((ub4)k[9]<<16);
		case 9 : c+=((ub4)k[8]<<8);
			/* the first byte of c is reserved for the length */
		case 8 : b+=((ub4)k[7]<<24);
		case 7 : b+=((ub4)k[6]<<16);
		case 6 : b+=((ub4)k[5]<<8);
		case 5 : b+=k[4];
		case 4 : a+=((ub4)k[3]<<24);
		case 3 : a+=((ub4)k[2]<<16);
		case 2 : a+=((ub4)k[1]<<8);
		case 1 : a+=k[0];
			/* case 0: nothing left to add */
	}
	mix(a,b,c);
	/*-------------------------------------------- report the result */
	return c;
}

//////////////////////////////////////////////////////////////////////////
// tgIoHash
//
// container function to make the code above useable as callback
//
// in:  char* filename to be hashed
//
// out: unsiged int hash value
//////////////////////////////////////////////////////////////////////////
unsigned int tgIoHash( const char *name )
{
	return( hash((ub1 *)name, strlen(name), 0x04c11db7) );
}

//////////////////////////////////////////////////////////////////////////
// tgIoSetHashFunction
//
// function to set the hashing function callback
//
// in:  pointer to an appropriate hashing function that takes a char* as
//		parameter and returns a 32bit unsigned int value
// out: -
//////////////////////////////////////////////////////////////////////////
static unsigned int (*tgio_hash)( const char *string ) = tgIoHash ;

void tgIoSetHashFunction( unsigned int (*hash_func)( const char *string ) )
{
	if ( hash_func == 0 )
		hash_func = tgIoHash;

	tgio_hash = hash_func;
}

//////////////////////////////////////////////////////////////////////////
// tgIoSetMemoryManager
//
// in:  pointer to a malloc() equivalent function
//		pointer to a free() equivalent function
//
// out: -
//////////////////////////////////////////////////////////////////////////
void *(*tgio_alloc)( size_t size ) = malloc ;
void (*tgio_free)( void *addr ) = free ;

void tgIoSetMemoryManager( void *(*alloc_func)( size_t size ), void (*free_func)( void *addr ) )
{
	if ( alloc_func == 0 )
		alloc_func = malloc;
	if ( free_func == 0 )
		free_func = free;
	tgio_alloc = alloc_func;
	tgio_free = free_func;
}

//////////////////////////////////////////////////////////////////////////
// tgIoCountFiles
//
// in : const char * to the directory we want to count
//
// out: int number of files inside the given directory
//////////////////////////////////////////////////////////////////////////
int tgIoCountFiles( const char *dname , int recursive )
{
	char dname_buffer[256];
	int ret;
	int fileCount = 0;
	SceUID fd;
	SceIoDirent dirent;

	fd = Asura_PSP_FileIO::OpenDir(dname, false);

	if (fd < 0)
		return(-1);

	while(1)
	{ 
		ret = Asura_PSP_FileIO::ReadDir(fd, &dirent, false); 

		if (ret < 0)
		{
			fileCount = -1;
		}

		if (ret <= 0)
		{
			break;
		}

		if (SCE_STM_ISDIR(dirent.d_stat.st_mode))
		{
			#ifdef TGIO_CHECK_FATDIRS
				if(strcmp(".", dirent.d_name)==0)
					continue;
				if(strcmp("..", dirent.d_name)==0)
					continue;
			#endif

			if(recursive)
			{
				sprintf(dname_buffer, "%s%s/", dname, dirent.d_name);
				int iNumFiles = tgIoCountFiles(dname_buffer, recursive);
				if(iNumFiles >= 0)
				{
					fileCount += iNumFiles;
				}
				else
				{
					fileCount = -1;
					break;
				}
			}
		}
		else
		{
			fileCount++;
		}
	}

	Asura_PSP_FileIO::CloseDir(fd);

	return( fileCount );
}


//////////////////////////////////////////////////////////////////////////
// tgIoCacheDir
//
// in : const char* containing the directory name to be cached
//		TGIO_CACHE_ENTRY* to the first entry to start filling
//		int flag for recursive caching (1 = yes, 0 = no)
//
// out: TGIO_CACHE_ENTRY* to the next vailable entry
//////////////////////////////////////////////////////////////////////////
TGIO_CACHE_ENTRY *tgIoCacheDir(const char *dname, TGIO_CACHE_ENTRY *destination, int recursive)
{
	char dname_buffer[256];
	int ret;
	SceUID fd;
	SceIoDirent dirent;

	fd = Asura_PSP_FileIO::OpenDir(dname, false);

	if (fd < 0)
	{
		destination = NULL;
		return(destination);
	}

	while(1)
	{ 
		ret = Asura_PSP_FileIO::ReadDir(fd, &dirent, false); 

		if (ret < 0)
		{
			destination = NULL;
		}

		if (ret <= 0)
		{			
			break;
		}

		if (SCE_STM_ISDIR(dirent.d_stat.st_mode))
		{
			#ifdef TGIO_CHECK_FATDIRS
				if(strcmp(".", dirent.d_name)==0)
					continue;
				if(strcmp("..", dirent.d_name)==0)
					continue;
			#endif

			if(recursive)
			{
				sprintf(dname_buffer, "%s%s/", dname, dirent.d_name);
				destination = tgIoCacheDir(dname_buffer, destination, recursive);
				if(!destination)
				{
					break;
				}
			}
		}
		else
		{
			sprintf(dname_buffer, "%s%s", dname, dirent.d_name);
			unsigned int hash = tgio_hash(dname_buffer);

			#ifdef TGIO_CHECK_COLLISIONS
				TGIO_CACHE_ENTRY *collide = tgIoFindHash(hash);

				#ifdef TGIO_CACHE_STORE_FILENAME
					if(collide)
					{
						if(strcmp(dname_buffer, collide->name)==0)
						{
							//printf("***WARNING: file %s appears twice!\n", test->name);
							continue;
						}
						else
						{
							printf(	"***WARNING: hash collision!\n"
									"***WARNING: 1st file: %s\n"
									"***WARNING: 2nd file: %s\n\n", dname_buffer, collide->name);
						}
					}
				#else
					if(collide)
						continue;
				#endif
			#endif

			destination->hash = hash;
			destination->lba = (int)dirent.d_stat.st_private[0];
			destination->size = (int)dirent.d_stat.st_size;
			#ifdef TGIO_CACHE_STORE_FILENAME
				ret = strlen(dname_buffer)+1;
				destination->name = (char*)tgio_alloc(ret);
				strcpy(destination->name, dname_buffer);
			#endif

			destination++;
		}
	}
	Asura_PSP_FileIO::CloseDir(fd);
	return(destination);
}

//////////////////////////////////////////////////////////////////////////
// tgIoFillCache
//
// function to fill the internal TOC cache for use by tgIoOpen()
//
// in:  char* to the directory to be cached, usualy "disc0:/"
//		TGIO_CACHE* to the cache to be used. set to 0 if you want the
//					function to allocate and init one for you
//		int flag for recursive caching (1 = yes, 0 = no)
//
// out: TGIO_CACHE* to the cached files
//////////////////////////////////////////////////////////////////////////
#ifdef __MWERKS__
	void (*tgio_sort)( TGIO_CACHE *cache ) = tgIoSortCache;
#endif

TGIO_CACHE *tgIoFillCache( const char *dname, TGIO_CACHE *cache, int recursive )
{
	int i;

	if( cache == 0 )
	{
		// allocate cache
		cache = (TGIO_CACHE*)tgio_alloc( sizeof(TGIO_CACHE) );

		// return if not enough memory
		if( cache == 0 )
			return(0);

		cache->numFiles = -1;

		while(cache->numFiles < 0)
		{
			// set number of files in cache
			#ifdef TGIO_CACHE_NUM_FILES
				cache->numFiles = TGIO_CACHE_NUM_FILES ;
			#else
				cache->numFiles = tgIoCountFiles( dname , recursive);
			#endif
		}

		// calc number of bytes to alloc for the array (+1 dummy at the end as marker)
		i = ( cache->numFiles * (sizeof(TGIO_CACHE_ENTRY)) ) + (sizeof(TGIO_CACHE_ENTRY)) ;
		cache->cache = (TGIO_CACHE_ENTRY*)tgio_alloc( i );

		// if no memory, then free cache and return
		if( cache->cache == 0 )
		{
			tgio_free( cache );
			return(0);
		}

		// clear the array
		memset(cache->cache, 0, i );
	}

	tgIoCurrentCache = cache;

	while(tgIoCacheDir(dname, cache->cache, recursive) == NULL)
	{
		memset(cache->cache, 0, i );
	};

	tgio_sort(cache);

//	for( i = 0; i < cache->numFiles; i++ )
//		printf("(%3d) - %p - %s\n", i, (void*)(cache->cache[i].hash), cache->cache[i].name);

	return(cache);
}


//////////////////////////////////////////////////////////////////////////
// tgIoSetCacheSort
//
// function to set the cache sort callback function
//
// in:  function pointer to a sorting function
//
// out: -
//////////////////////////////////////////////////////////////////////////
void tgIoSetCacheSort( void (*sort_func)( TGIO_CACHE *cache ) )
{
	if ( sort_func == 0 )
		sort_func = tgIoSortCache;

	tgio_sort = sort_func;
}

//////////////////////////////////////////////////////////////////////////
// tgIoSortCache
//
// function to sort a cache by the hash values
//
// in:  TGIO_CACHE* to a cache
//
// out: -
//////////////////////////////////////////////////////////////////////////
void tgIoSortCache(TGIO_CACHE *cache)
{
	int i, j;

	// sort the array by Hash
	for( i = 0; i < cache->numFiles-1; i++ )
	{
		for( j = i+1; j < cache->numFiles; j++ )
		{
			if(cache->cache[i].hash > cache->cache[j].hash)
			{
				unsigned int tempLBA = cache->cache[j].lba;
				unsigned int tempHash = cache->cache[j].hash;
				unsigned int tempSize = cache->cache[j].size;

				#ifdef TGIO_CACHE_STORE_FILENAME
					char *tempName = cache->cache[j].name;
					cache->cache[j].name = cache->cache[i].name;
					cache->cache[i].name = tempName;
				#endif

				cache->cache[j].lba = cache->cache[i].lba;
				cache->cache[j].hash = cache->cache[i].hash;
				cache->cache[j].size = cache->cache[i].size;

				cache->cache[i].lba = tempLBA;
				cache->cache[i].hash = tempHash;
				cache->cache[i].size = tempSize;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// tgIoClearCache
//
// function to clear the cache list and free all allocated memory
//
// in:  -
//
// out: -
//////////////////////////////////////////////////////////////////////////
void tgIoClearCache( TGIO_CACHE *cache )
{
	if(cache==0)
	{
		cache = tgIoCurrentCache;
		tgIoCurrentCache = NULL;
	}

	if(cache)
	{
		#ifdef TGIO_CACHE_STORE_FILENAME
			int i;
			for(i=0; i<cache->numFiles; i++)
			{
				if(cache->cache[i].name)
					tgio_free(cache->cache[i].name);
			}
		#endif

		cache->numFiles = -1;
		if(cache->cache)
			tgio_free(cache->cache);
		tgio_free(cache);
	}
}

//////////////////////////////////////////////////////////////////////////
// tgIoGetNumFiles
//
// function to get the number of cached files
//
// in:  -
//
// out: int number of cached files
//////////////////////////////////////////////////////////////////////////
int tgIoGetNumFiles( void )
{
	return( tgIoCurrentCache->numFiles );
}

//////////////////////////////////////////////////////////////////////////
// tgIoGetAllocatedBytes
//
// function to get the number of bytes allocted for the cache
//
// in:  -
//
// out: unsigned int number of total bytes allocated for the cache
//////////////////////////////////////////////////////////////////////////
unsigned int tgIoGetAllocatedBytes( void )
{
	unsigned int result = 0;

	if(tgIoCurrentCache)
	{
		result += sizeof(TGIO_CACHE) + sizeof(TGIO_CACHE_ENTRY) + ( sizeof(TGIO_CACHE_ENTRY) * tgIoCurrentCache->numFiles ) ;
		#ifdef TGIO_CACHE_STORE_FILENAME
			int i;
			for(i=0; i<tgIoCurrentCache->numFiles; i++)
				if( tgIoCurrentCache->cache[i].name )
					result += strlen( tgIoCurrentCache->cache[i].name ) + 1 ;
		#endif
	}

	return(result);
}

//////////////////////////////////////////////////////////////////////////
// tgIoFindHashB
//
// function to find a given hash value in the cache list from a given
// start position
//
// in:  unsigned int hash value to look for
//		TGIO_FLIST* starting position
//
// out: TGIO_FLIST* where a hash match was found
//		or 0 if no match was found
//////////////////////////////////////////////////////////////////////////
TGIO_CACHE_ENTRY *tgIoFindHashB( unsigned int hash, TGIO_CACHE_ENTRY *walkList )
{
	while(walkList->size)
	{
		if(walkList->hash == hash)
			return(walkList);
		walkList++;
	}

	return(0);
}

//////////////////////////////////////////////////////////////////////////
// tgIoFindHash
//
// function to find a given hash value in the cache list
//
// in:  unsigned int hash value to look for
//
// out: TGIO_FLIST* where a hash match was found
//		or 0 if no match was found 
//////////////////////////////////////////////////////////////////////////
TGIO_CACHE_ENTRY *tgIoFindHash( unsigned int hash )
{
	if(tgIoCurrentCache)
	{
		TGIO_CACHE_ENTRY *cache = tgIoCurrentCache->cache;
		if(cache)
		{
			int i;
			for(i=0; i<tgIoCurrentCache->numFiles; i++)
			{
				if(cache->hash == hash)
					return(cache);
				cache++;
			}
		}
	}
	return(0);
}

//////////////////////////////////////////////////////////////////////////
// tgIoFindCacheCollide
//
// function to count the number of collides in the cache
//
// in:  -
//
// out: int number of collides
//////////////////////////////////////////////////////////////////////////
int tgIoFindCacheCollide( void )
{
	int count = 0;

	TGIO_CACHE_ENTRY *root = tgIoCurrentCache->cache;
	TGIO_CACHE_ENTRY *child;

	while(root->size)
	{
		child = root+1;
		if(child->size)
			if( tgIoFindHashB(root->hash, child) )
				count++;
		root = child;
	}

	return(count);
}

//////////////////////////////////////////////////////////////////////////
// tgIoFindFile
//
// function to find a specific file in the cache
//
// in:  char* to the filename
//
// out: TGIO_FLIST* where the file was found
//		or 0 if no match was found 
//////////////////////////////////////////////////////////////////////////
TGIO_CACHE_ENTRY *tgIoFindFile( const char *filename )
{
	if(tgIoCurrentCache)
	{
		TGIO_CACHE_ENTRY *base = tgIoCurrentCache->cache;
		if(base)
		{
			int left=0,
				right=tgIoCurrentCache->numFiles,
				middle=0;
			unsigned int hash = tgio_hash(filename);

			while(left<=right)
			{
				middle = (left+right)>>1 ;

				#ifdef TGIO_CACHE_STORE_FILENAME
					if( ( base[middle].hash == hash ) &&
						( strcmp(filename, base[middle].name ) == 0 ) )
				#else
					if( base[middle].hash == hash )
				#endif
				{
					return( &base[middle] );
				}
				else if( base[middle].hash > hash )
				{
					right = middle-1;
				}
				else
				{
					left = middle+1;
				}
			}
		}
	}
	return(0);
}

//////////////////////////////////////////////////////////////////////////
// tgIoGetCache
//
// in : -
//
// out: TGIO_CACHE* to the currently used cache
//////////////////////////////////////////////////////////////////////////
TGIO_CACHE *tgIoGetCache(void)
{
	return( tgIoCurrentCache );
}

//////////////////////////////////////////////////////////////////////////
// tgIoSetCache
//
// in : TGIO_CACHE* to the new active cache
//
// out: -
//////////////////////////////////////////////////////////////////////////
void tgIoSetCache(TGIO_CACHE *cache)
{
	tgIoCurrentCache = cache;
}

//////////////////////////////////////////////////////////////////////////
// tgIoGetCachedName
//
// in : const char* to the filename we want to have the sector name for
//		char* to a buffer of at least 40 bytes
//
// out: TGIO_CACHE_ENTRY* to the corresponding cache entry
//		or 0 in case of an error
//////////////////////////////////////////////////////////////////////////
TGIO_CACHE_ENTRY *tgIoGetCachedName( const char *filename, char *outname )
{
	TGIO_CACHE_ENTRY *hashedFile = tgIoFindFile(filename);

	if(hashedFile)
		sprintf(outname, "disc0:/sce_lbn0x%x_size0x%x", hashedFile->lba, hashedFile->size);

	return(hashedFile);
}

//////////////////////////////////////////////////////////////////////////
// tgIoGetCachedNameA
//
// in : const char* to the filename we want to have the sector name for
//
// out: const char* to a LBA filename that can be used to open the file
//		or 0 in case of an error
//////////////////////////////////////////////////////////////////////////
const char *tgIoGetCachedNameA( const char *filename )
{
	static char pBuffer[TGIO_STRINGBUFFER_SIZE];
	TGIO_CACHE_ENTRY *hashedFile = tgIoFindFile(filename);

	if(hashedFile)
	{
		sprintf(pBuffer, "disc0:/sce_lbn0x%x_size0x%x", hashedFile->lba, hashedFile->size);
		return(pBuffer);
	}
	else
	{
		return(0);
	}
}

//////////////////////////////////////////////////////////////////////////
// tgIoOpen
//
// function to replace sceIoOpen() by a cached high-speed version using
// sector based file open access
//
// in:  const char* filename of file to be opened
//		int flag (see iomanuser-Reference for list of valid flags)
//		SceMode mode for when flag is set to SCE_O_CREAT
//
// out: SceUID >= 0 filedescriptor
//		or < 0 if error
//////////////////////////////////////////////////////////////////////////
SceUID tgIoOpen( const char *filename, int flag, SceMode mode )
{
	static char buffer[TGIO_STRINGBUFFER_SIZE];

	TGIO_CACHE_ENTRY *hashedFile = tgIoGetCachedName(filename, buffer);

	if(hashedFile)
	{
		#ifdef TGIO_CHECK_FATDIRS
			if(hashedFile->lba)
			{
		#endif
				#ifdef TGIO_ENABLE_PA
					if(tgIoPAIsStarted)
						tgIoPAAdd(filename, hashedFile->lba, hashedFile->size);
				#endif
				return( sceIoOpen(buffer, flag, mode) );
		#ifdef TGIO_CHECK_FATDIRS
			}
		#endif
	}

	#ifdef TGIO_REPORT_ERROR
		printf("***WARNING: tgIoOpen() couldn't find file %s (%p) in cache!\n", filename, (void*)tgio_hash(filename) );
	#endif

	return( sceIoOpen(filename, flag, mode) );
}

//////////////////////////////////////////////////////////////////////////
// tgIoOpenAsync
//
// function to replace sceIoOpen() by a cached high-speed version using
// sector based file open access
//
// in:  const char* filename of file to be opened
//		int flag (see iomanuser-Reference for list of valid flags)
//		SceMode mode for when flag is set to SCE_O_CREAT
//
// out: SceUID >= 0 filedescriptor
//		or < 0 if error
//////////////////////////////////////////////////////////////////////////
SceUID tgIoOpenAsync( const char *filename, int flag, SceMode mode )
{
	static char buffer[TGIO_STRINGBUFFER_SIZE];

	TGIO_CACHE_ENTRY *hashedFile = tgIoGetCachedName(filename, buffer);

	if(hashedFile)
	{
		#ifdef TGIO_CHECK_FATDIRS
			if(hashedFile->lba)
			{
		#endif
				#ifdef TGIO_ENABLE_PA
					if(tgIoPAIsStarted)
						tgIoPAAdd(filename, hashedFile->lba, hashedFile->size);
				#endif
				return( sceIoOpenAsync(buffer, flag, mode) );
		#ifdef TGIO_CHECK_FATDIRS
			}
		#endif
	}

	#ifdef TGIO_REPORT_ERROR
		printf("***WARNING: tgIoOpenAsync() couldn't find file %s (%p) in cache!\n", filename, (void*)tgio_hash(filename) );
	#endif

	return( sceIoOpenAsync(filename, flag, mode) );
}

int tgIoGetFileSize(const char* szFilename)
{
	static char buffer[TGIO_STRINGBUFFER_SIZE];

	TGIO_CACHE_ENTRY *hashedFile = tgIoGetCachedName(szFilename, buffer);

	if(hashedFile)
	{
		return hashedFile->size;
	}

#ifdef TGIO_REPORT_ERROR
	printf("***WARNING: tgIoOpen() couldn't find file %s (%p) in cache!\n", szFilename, (void*)tgio_hash(szFilename) );
#endif
	// Get the file length
	SceIoStat xFileStat;
	int iError = sceIoGetstat( szFilename, &xFileStat ); 

#ifdef TGIO_REPORT_ERROR
	printf("***ERROR: sceIoGetstat() returned error code (%x) for file %s!\n", iError, szFilename );
#endif

	return (iError >= 0) ? xFileStat.st_size : 0;
}