//
// Asura_PSP_MoviePlayback.cpp
//
// Created by KJFL - 13 April 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "MoviePlayback/Asura_PSP_MoviePlayback.h"
#include "MoviePlayback/Asura_PSP_MoviePlayback_Internal.h"

#ifdef ASURA_SUPPORT_PSP_MPEG

#include <libwave.h>
#include <displaysvc.h>
#include <powersvc.h>
#include <libpsmf.h>

#include "Core/Asura_String.h"
#include "Core/Asura_UserInput.h"
#include "Core/Asura_PSP_KernelManagement.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Render.h"

#include "Input/Asura_PlatformInput.h"

#include "Core/Asura_Platform_MemoryManagement.h"
#include "Core/Asura_PSP_Thread_Priorities.h"
#include "Core/Asura_PSP_VolatileMemoryManagement.h"
#include "Debugging/Asura_PSP_ScreenLog.h"
#include "FileAccess/Asura_PSP_FileIO.h"
#include "Rendering/Asura_PSP_MotionBlur.h"
#include "Sound/Asura_PSP_SAS.h"

// D E F I N E S /////////////////////////////////////////////////////////////
#define Asura_PSP_MovieDebugOutput(x)
//#define Asura_PSP_MovieDebugOutput(x) Asura_OutputToDebugger_Retail(x)

// I N T E R N A L   F U N C T I O N S ///////////////////////////////////////
static void Asura_PSP_MovieLockMutex( int iLine );
static void Asura_PSP_MovieUnlockMutex();

// G L O B A L S /////////////////////////////////////////////////////////////
DEBUGVAR bool dbg_bAudioOn = true;
DEBUGVAR int dbg_iStreamNumber = 0;
static bool g_bForceUseOfSoundMem = false;

static Asura_Handle g_uVolatileMemoryHandle	= ASURA_INVALID_HANDLE;
static u_char*		g_pucVolatileMemory		= NULL;
static Asura_Mutex	g_xMovieMutex;
static int			g_iMovieMutexLockLine;
static u_int		g_uNumMovieSuspends;

// with pack size (SCE_MPEG_PACK_SIZE) being 2048, this equates to a 2Mb ring buffer
static const u_int uRINGBUFFER_NO_OF_PACKS = 512;
static const u_int uLCD_FRAME_SIZE  = 512;
static const u_int uNO_OF_DECODED_AUDIO_BUFFERS = 8;


static const u_int AUDIO_OUTPUT_CHANNEL = 2;

// The header for the movie file is an unknown size (apparently). However, Sony indicate
// it should always be contained in the first 2k of the file. I've no idea if this is 
// because the mpeg pack size  is 2048, so I'm keeping it a separate constant.
const u_int uASURA_PSP_MOVIE_HEADER_SIZE = 2048;

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_PSP_Movie* Asura_PSP_MoviePlayback::s_pxCurrentMovie = 0;
Asura_PSP_Movie* Asura_PSP_MoviePlayback::s_pxShuttingDownMovie = 0;
const char* Asura_PSP_MoviePlayback::s_szFilename = 0;
int Asura_PSP_MoviePlayback::s_iAudioStream = 0;
int Asura_PSP_MoviePlayback::s_iVideoStream = 0;
volatile bool Asura_PSP_MoviePlayback::s_bSuspended = false;
volatile bool Asura_PSP_MoviePlayback::s_bResume = false;

bool Asura_PSP_MoviePlayback::s_bDiskOut = false;
bool Asura_PSP_MoviePlayback::s_bDiskReturned = false;
volatile bool Asura_PSP_MoviePlayback::s_bCanDecodeAudio = false;

// C L A S S   F U N C T I O N S /////////////////////////////////////////////
void Asura_PSP_MoviePlayback::PlayFullScreenMovie(const char* const szFilename,
												  int iAudioStream /* = 0 */,
												  int iVideoStream /* = 0 */,
												  void (*pfnRenderSubtitleCB)(const float fTime) /* = NULL */,
												  float fTimeBeforeSkipAllowed /* = 5.00f */,
												  float fAudioVolume /*= 1.0f*/,
												  bool bStreamShouldBePresent /*= true*/)
{
	Asura_Assert(s_pxCurrentMovie == 0, "trying to play more than one movie at once. this call blocks so this should NEVER happen");

	Asura_PSP_Movie xMovie(pfnRenderSubtitleCB, fTimeBeforeSkipAllowed, fAudioVolume);
	bool bBlurEnabled = Asura_PSP_MotionBlur::IsEnabled();
	bool bWasFogged	= Asura_Fog::IsFogEnabled();

	// Fog colour changes the screen clear colour - so disable it to
	// clear the screen to black (seen when using the power switch)...
	Asura_Fog::EnableFog(false);
	Asura_PSP_MotionBlur::Enable(false);

	// Disable any currently playing sounds
	Asura_Sound_System::PauseAllSystems();

	s_szFilename = szFilename;
	s_iAudioStream = iAudioStream;
	s_iVideoStream = iVideoStream;

	do
	{
		// Handle waiting for suspension to be resumed...
		if (s_bSuspended && !s_bResume)
		{
			while (s_bSuspended)
			{
				//we're suspended. loop ad infinitum, as the psp inexplicably waits a few
				//frames after calling the suspend callback before actually suspeneding
				if (s_bResume)
				{
					//resume flag set, get the movie up and running again
					s_bSuspended = false;
					s_bResume = false;
				}
				else
					{
					//just twiddling our thumbs until we resume
					Asura_Render::BeginFrame();
					Asura_Render::EndFrame();
					Asura_UserInput::Update();

					// Ensure we're allowed to do something else...
					Asura_Thread_System::YieldCurrentThread();
				}
			}
		}
		else
		{
			s_bSuspended = false;
			s_bResume = false;
		}
		CheckDiskStatus();
		if (s_bDiskOut && !s_bDiskReturned)
		{
			while (s_bDiskOut)
			{
				if (s_bSuspended)
				{
					s_bDiskOut = false;
					s_bDiskReturned = false;
				}
				CheckDiskStatus();
				//we're suspended. loop ad infinitum, as the psp inexplicably waits a few
				//frames after calling the suspend callback before actually suspeneding
				if (s_bDiskReturned)
				{
					Asura_OutputToDebugger(("FMV handled disk in\n"));
					//resume flag set, get the movie up and running again
					s_bDiskOut = false;
					s_bDiskReturned = false;
				}
				else
				{
					//just twiddling our thumbs until we resume
					Asura_Render::BeginFrame();
					// really need some text here.....
					Asura_Render::EndFrame();
					Asura_UserInput::Update();

					// Ensure we're allowed to do something else...
					Asura_Thread_System::YieldCurrentThread();
				}
			}
		}
		else
		{
			s_bDiskOut = false;
			s_bDiskReturned = false;
		}

		if (!s_bSuspended && !s_bDiskOut)
		{
			// Initialise the movie...
			Asura_PSP_MovieLockMutex(__LINE__);
            // (MH - 13.07.2007) new retry loop to cope with power off during initialise
			bool bInitOk = false;
			bool bRetry = false;
			do
			{
				u_int uNumSuspendsBeforeInitialise = g_uNumMovieSuspends;
				bRetry = false;
				bInitOk = xMovie.Initialise( szFilename, iAudioStream, iVideoStream, bStreamShouldBePresent );
				if (bInitOk)
				{
					s_pxCurrentMovie = &xMovie;
				}
				else
				{
					xMovie.Shutdown();
					// retry if the init failed after a suspension.
					if (uNumSuspendsBeforeInitialise != g_uNumMovieSuspends)
					{
						bRetry = true;
					}
				}

				// reset all the suspension vars before continuing/retrying
				s_bDiskOut = false;
				s_bDiskReturned = false;
				s_bSuspended = false;
				s_bResume = false;
			}
			while (bRetry);

			Asura_PSP_MovieUnlockMutex();

			CheckDiskStatus();
			if (bInitOk && !s_bSuspended && !s_bDiskOut)
			{
				// Play the movie...
				Asura_PSP_MovieLockMutex(__LINE__);
				xMovie.Play();
				Asura_PSP_MovieUnlockMutex();
				s_bCanDecodeAudio = true;

				CheckDiskStatus();
				while(!xMovie.Finished() && !s_bSuspended && !s_bDiskOut)
				{
					// Play the movie...
					Asura_PSP_MovieLockMutex(__LINE__);
					xMovie.Update();
					Asura_PSP_MovieUnlockMutex();
					CheckDiskStatus();
				}
				s_bCanDecodeAudio = false;
			}
		}
	} while(s_bSuspended || s_bDiskOut);

	Asura_PSP_MovieLockMutex(__LINE__);
	s_pxCurrentMovie = NULL;
	s_pxShuttingDownMovie = &xMovie;
	xMovie.Shutdown();
	s_pxShuttingDownMovie = NULL;
	Asura_PSP_MovieUnlockMutex();

	// Restore the fogging...
	Asura_Fog::EnableFog(bWasFogged);
	Asura_PSP_MotionBlur::Enable(bBlurEnabled);

	// Restore sounds
	Asura_Sound_System::UnpauseAllSystems();
}

void Asura_PSP_MoviePlayback::CheckDiskStatus()
{
	if (!s_bSuspended)
	{
		bool bDiskReady = Asura_PSP_KernelManagement::IsUMDReady();
		if (!s_bDiskOut && !bDiskReady)
		{
			Asura_OutputToDebugger(("FMV registered disk out\n"));
			s_bDiskOut = true;
			if (s_pxCurrentMovie)
			{
				if (!s_pxCurrentMovie->Finished())
				{
					s_pxCurrentMovie->Finish();
				}
				s_pxCurrentMovie->Shutdown();
			}
		}
		if (s_bDiskOut && !s_bDiskReturned && bDiskReady)
		{
			Asura_OutputToDebugger(("FMV registered disk in\n"));
			s_bDiskReturned = true;
		}
	}
}

void Asura_PSP_MoviePlayback::Suspend (void)
{
	// NY - 23 June 2006 - skip the rest of the fmv - we need to drop out of the blocking function to suspend / exit
	// if we ever need to resume FMVs (from the start, or, for the adventerous, where we left off)
	// this playback code would need to become non-blocking and polled from the game's update loop
	// NY - 13 July 2006 - still killing the fmv off, but we don't need to drop out of the blocking function, just let it run a mini-main-loop
	//this means it's still there on resume, and can restart the fmv. Hurrah!
	
	Asura_PSP_MovieLockMutex(__LINE__);

	s_bSuspended = true;
	if (s_pxCurrentMovie)
	{
		//under normal cicumstances we shouldn't do this from a side thread, but we're about
		//to suspend the volatile memory manager, and the movie's properties get stomped on when we resume, so this is safe
		if (!s_pxCurrentMovie->Finished())
		{
			s_pxCurrentMovie->Finish();
		}
		s_pxCurrentMovie->Shutdown();
	}
	if (s_pxShuttingDownMovie)
	{
		// We were already trying to shut down. Just make sure the memory gets deallocated.
		s_pxShuttingDownMovie->DeallocateVolatileMemory();
	}

	++g_uNumMovieSuspends;
}

void Asura_PSP_MoviePlayback::Resume (void)
{
	s_bResume = true;
	if (s_pxCurrentMovie)
	{
		Asura_Assert(s_bSuspended, "resuming an fmv we didn't suspend?");
		//just flag to resume, as we're in the resume thread and it's not a good idea to much here!
	}

	Asura_PSP_MovieUnlockMutex();
}

bool Asura_PSP_MoviePlayback::DecodeAudio (short* psCurrentBuffer, float& fVolume)
{
	if(s_pxCurrentMovie == NULL)
	{
		return(false);
	}
	else if(!s_bCanDecodeAudio || s_bSuspended || s_bDiskOut)
	{
		return(false);
	}
	else
	{
		return(s_pxCurrentMovie->DecodeAudio(psCurrentBuffer, fVolume));
	}
}


static void AllocateVolatileMemoryCallback(u_char* pucDataStart)
{
	g_pucVolatileMemory	= pucDataStart;
}

static void DeallocateVolatileMemoryCallback(u_char* pucDataStart)
{
	g_pucVolatileMemory	= NULL;
}


Asura_PSP_Movie::Asura_PSP_Movie(void (*pfnRenderSubtitleCB)(const float fTime),
										float fTimeBeforeSkipAllowed,
										float fAudioVolume)
{
	// set some things to sensible defaults
	m_iFileID = ASURA_INVALID_HANDLE;
	m_uStreamOffset = 0;
	m_uStreamSize = 0;
	m_uBytesRead = 0;
	m_pxVideoStream = NULL;
	m_pxAudioStream = NULL;
	m_pxAvcEsBuf = NULL;
	m_ppucDecodedAudioBuffers = NULL;
	m_iAudioInitFlag = 0;
	m_iNextAudioDecodeBuffer = 0;
	m_iNextAudioPlayBuffer = 0;
	m_iNumberOfFullAudioBuffers = 0;
	m_bKillAudioThread = false;
	m_bUsedSASMemory = false;
	m_bCallsceMpegInitOK = false;
	m_bCallsceMpegCreateOK = false;
	m_pcPSPStream = NULL;
	m_pcWorkBuffer = NULL;
	m_pcAudioBuffer = NULL;
	m_bScreenClearEnabledStore = false;
	m_bFinished = true;
	m_bDoneInitialSeek = false;

	m_pfnRenderSubtitleCB = pfnRenderSubtitleCB;
	m_fTimeBeforeSkipAllowed = fTimeBeforeSkipAllowed;
	m_fAudioVolume = fAudioVolume;
	m_iStartVSyncCount = 0;
	m_iFramesDecoded = 0;

	m_iNumAudioStreams = 0;
	m_iNumVideoStreams = 0;
}

Asura_PSP_Movie::~Asura_PSP_Movie()
{
}

bool Asura_PSP_Movie::Initialise( const char* const szFilename, int iAudioStream, int iVideoStream, bool bStreamShouldBePresent )
{
	m_bFinished = false;
	m_bDoneInitialSeek = false;
	m_bKillAudioThread = false;
	m_bUsedSASMemory = false;

	char acHeaderBuffer[uASURA_PSP_MOVIE_HEADER_SIZE];
	char szTotalFilename[256];

	Asura_PSP_FileIO::GetFullPathForFMV(szFilename, 256, szTotalFilename);
	Asura_PSP_ScreenLog::Log("Playing movie : %s\n",szTotalFilename);

	Asura_PSP_MovieDebugOutput(("Playing movie : %s\n",szTotalFilename));

	// release the mutex while the file access happens.
	Asura_PSP_MovieUnlockMutex();
	m_iFileID = Asura_PSP_FileIO::Open( szTotalFilename , SCE_O_RDONLY, 0777 );
	Asura_PSP_MovieDebugOutput(("Asura_PSP_FileIO::Open : %d\n",m_iFileID));

	if (m_iFileID < 0)
	{
		Asura_PSP_MovieLockMutex(__LINE__);
		return false;
	}

	int iRetValue;

	// Read in the first two kilobytes so we have the header in memory
	int iReadSize = Asura_PSP_FileIO::Read( m_iFileID, acHeaderBuffer, uASURA_PSP_MOVIE_HEADER_SIZE );

	if ( iReadSize < uASURA_PSP_MOVIE_HEADER_SIZE )
	{
		Asura_PSP_MovieLockMutex(__LINE__);
		return false;
	}

	// do the parsing which doesn't require the mpeg object to be initialised.
	bool bOK = ParseHeader1( acHeaderBuffer );
	Asura_PSP_MovieDebugOutput(("ParseHeader1 : %s\n", bOK ? "OK" : "failed"));
	if (!bOK)
	{
		Asura_PSP_MovieLockMutex(__LINE__);
		return false;
	}

	// now the file access is complete we can lock the mutex again
	Asura_PSP_MovieLockMutex(__LINE__);

	iRetValue = sceMpegInit();

	Asura_PSP_MovieDebugOutput(("sceMpegInit : %d\n",iRetValue));

	if ( iRetValue != SCE_MPEG_OK )
	{
		return false;
	}
	m_bCallsceMpegInitOK = true;
	
	SceInt32 iBufferSize = sceMpegRingbufferQueryMemSize( uRINGBUFFER_NO_OF_PACKS );

	Asura_PSP_MovieDebugOutput(("sceMpegRingbufferQueryMemSize : %d\n",iBufferSize));

	if ( iBufferSize < 0 )
	{
		return false;
	}
	
	SceInt32 iWorkSize = sceMpegQueryMemSize (0);

	Asura_PSP_MovieDebugOutput(("sceMpegQueryMemSize : %d\n",iWorkSize));

	if ( iWorkSize < 0 )
	{
		return false;
	}

	// Audio buffer sizes: docs say to allocate 2048+64 bytes for ES size 
	// and 8192 for audio out, with 64-byte alignment.
	int iMaxExpectedAudioEsSize = 2112;
	int iMaxExpectedAudioOutSize = 8192;

	// get a big block of memory for all our requirements.
	u_int uBytesRequired = iBufferSize + iWorkSize + iMaxExpectedAudioEsSize + 63 +
		(iMaxExpectedAudioOutSize * uNO_OF_DECODED_AUDIO_BUFFERS);

	Asura_Assert((g_pucVolatileMemory == NULL), "Volatile memory is already allocated.");
	g_uVolatileMemoryHandle = Asura_PSP_VolatileMemoryManagement::Allocate(uBytesRequired,AllocateVolatileMemoryCallback,DeallocateVolatileMemoryCallback);

	Asura_PSP_MovieDebugOutput(("Asura_PSP_VolatileMemoryManagement::Allocate : %d\n",g_uVolatileMemoryHandle));

	if(g_uVolatileMemoryHandle == ASURA_INVALID_HANDLE)
	{
		return(false);
	}

	char* pcStartOfMemory = reinterpret_cast<char*>(g_pucVolatileMemory);
	char* pcMemoryBlockCur = pcStartOfMemory;
	// we rely later on the fact that m_pcPSPStream is the base ptr so we can free the memory.
	m_pcPSPStream = pcMemoryBlockCur;
	pcMemoryBlockCur += iBufferSize;
	m_pcWorkBuffer = pcMemoryBlockCur;
	pcMemoryBlockCur += iWorkSize;

	iRetValue = sceMpegRingbufferConstruct
				(
					&m_xRingbuffer,
					uRINGBUFFER_NO_OF_PACKS,
					m_pcPSPStream,
					iBufferSize,
					RingbufferCallBack,
					(SceVoid*)this
				);

	Asura_PSP_MovieDebugOutput(("sceMpegRingbufferConstruct : %d\n",iRetValue));

	if ( iRetValue != SCE_MPEG_OK )
	{
		return false;
	}
	
	iRetValue = sceMpegCreate(&m_xMpeg, (SceUChar8*)m_pcWorkBuffer, iWorkSize, &m_xRingbuffer, uLCD_FRAME_SIZE, 0, NULL);

	Asura_PSP_MovieDebugOutput(("sceMpegCreate : %d\n",iRetValue));

	if( iRetValue != SCE_MPEG_OK)
	{
		return false;
	}

	m_bCallsceMpegCreateOK = true;

	bOK = ParseHeader2(acHeaderBuffer);
	Asura_PSP_MovieDebugOutput(("ParseHeader2 : %s\n", bOK ? "OK" : "failed"));
	if (!bOK)
	{
		return false;
	}

	// If we've been passed an invalid video or audio stream, use stream 0.
	if (iAudioStream >= m_iNumAudioStreams)
	{
		if (m_iNumAudioStreams == 0)
		{
			Asura_Assertf(!bStreamShouldBePresent, ("Audio stream %d not present in FMV %s (no audio streams present).", iAudioStream, szFilename));
			iAudioStream = -1;
		}
		else
		{
			Asura_Assertf(!bStreamShouldBePresent, ("Audio stream %d not present in FMV %s (%d audio streams present). Using stream 0.", iAudioStream, szFilename, m_iNumAudioStreams));
			iAudioStream = 0;
		}
	}
	if (iVideoStream >= m_iNumVideoStreams)
	{
		Asura_Assertf(!bStreamShouldBePresent, ("Video stream %d not present in FMV %s (%d video streams present). Using stream 0.", iVideoStream, szFilename, m_iNumVideoStreams));
		iVideoStream = 0;
	}

	// register video stream
	{
		m_pxVideoStream = sceMpegRegistStream( &m_xMpeg, SCE_MPEG_STREAM_M4AVC, iVideoStream );
		
		Asura_PSP_MovieDebugOutput(("sceMpegRegistStream : 0x%x\n", m_pxVideoStream));

		if ( m_pxVideoStream == SCE_NULL )
		{
			Asura_OutputToDebugger(("fatal error : sceMpegRegistStream() AVC\n"));
			return false;
		}


		m_pxAvcEsBuf = sceMpegMallocAvcEsBuf( &m_xMpeg );

		Asura_PSP_MovieDebugOutput(("sceMpegMallocAvcEsBuf : 0x%x\n", m_pxAvcEsBuf));

		if ( m_pxAvcEsBuf == SCE_NULL ) 
		{
			Asura_OutputToDebugger(("fatal error : sceMpegMallocAvcEsBuf()\n"));
			return false;
		}

		iRetValue = sceMpegInitAu( &m_xMpeg, m_pxAvcEsBuf, &m_xAvcAu );

		Asura_PSP_MovieDebugOutput(("sceMpegInitAu : %d\n",iRetValue));

		if ( iRetValue != SCE_MPEG_OK )
		{
			Asura_OutputToDebugger(("fatal error : sceMpegInitAu() AVC %x\n", iRetValue));
			return false;
		}
	}

	// register audio stream
	if ( dbg_bAudioOn && (iAudioStream >= 0) )
	{
		m_pxAudioStream = sceMpegRegistStream( &m_xMpeg, SCE_MPEG_STREAM_ATRAC, iAudioStream );

		if ( m_pxAudioStream == SCE_NULL) 
		{
			Asura_OutputToDebugger( ("fatal error : sceMpegRegistStream() ATRAC\n"));
			return false;
		}

		SceInt32 iAudioEsSize, iAudioOutSize;
	
		/*J ATRAC3plus1AUům */
		iRetValue = sceMpegQueryAtracEsSize( &m_xMpeg, &iAudioEsSize, &iAudioOutSize );
		if (iRetValue != SCE_MPEG_OK)
		{
			Asura_OutputToDebugger( ("fatal error : sceMpegQueryAtracEsSize() %x\n", iRetValue));
			return false;
		}

		Asura_Assert(iAudioEsSize <= iMaxExpectedAudioEsSize, "unexpected Audio ES size");
		Asura_Assert(iAudioOutSize <= iMaxExpectedAudioOutSize, "unexpected Audio out size");

		// align audio buffer to 64 bytes
		m_pcAudioBuffer = reinterpret_cast<SceChar8*>((
			reinterpret_cast<u_int>(pcMemoryBlockCur) + 63) & ~63);
		pcMemoryBlockCur = m_pcAudioBuffer + iMaxExpectedAudioEsSize;

		/*J ATRAC3pluspAU\̂̏ */
		iRetValue = sceMpegInitAu( &m_xMpeg, (SceMpegHandle)m_pcAudioBuffer, &m_xAtracAu );
		if (iRetValue != SCE_MPEG_OK)
		{
			Asura_OutputToDebugger( ("fatal error : sceMpegInitAu() %x\n", iRetValue));
			return false;
		}

		m_ppucDecodedAudioBuffers = new SceUChar8* [uNO_OF_DECODED_AUDIO_BUFFERS];

		for ( u_int u=0; u<uNO_OF_DECODED_AUDIO_BUFFERS; ++u )
		{
			Asura_Assert(!(reinterpret_cast<u_int>(pcMemoryBlockCur) & 63), "memory block no longer aligned");
			m_ppucDecodedAudioBuffers[u] = reinterpret_cast<SceUChar8*>(pcMemoryBlockCur);
			pcMemoryBlockCur += iAudioOutSize;
		}

		// LOCKSEMA
		// This semaphore wasn't even being used, *AND* was leaking!
		// I'll comment it out for now, but if it needs to be around for some less obvious reason,
		// I've added in (further commented-out) code to delete it. Look for the LOCKSEMA comment

//		m_iLockSema = sceKernelCreateSema ("sound_lock_sema", SCE_KERNEL_SA_THFIFO, 1, 1, NULL);
//		Asura_Assert(m_iLockSema >= 0, "sceKernelCreateSema failed");
	}

	Asura_Assert( pcMemoryBlockCur <= &pcStartOfMemory[uBytesRequired], "overran memory buffer");

	//sceKernelChangeThreadPriority(sceKernelGetThreadId(), SCE_KERNEL_USER_LOWEST_PRIORITY );
	Asura_PSP_MovieDebugOutput(("Asura_PSP_Movie::Initialise() done\n"));

	// everything successful
	return true;

}


void Asura_PSP_Movie::Shutdown( void )
{
	#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
		int iRet = 0;
	#endif

	if (m_pxAvcEsBuf)
	{
		#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
			iRet =
		#endif
			sceMpegFreeAvcEsBuf( &m_xMpeg, m_pxAvcEsBuf);

		Asura_OutputToDebuggerAssert( iRet == SCE_OK, ("sceMpegFreeAvcEsBuf failed: %x\n", iRet));
		m_pxAvcEsBuf = NULL;
	}

	if (m_pxVideoStream)
	{
		#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
			iRet =
		#endif
			sceMpegUnRegistStream( &m_xMpeg, m_pxVideoStream );

		Asura_OutputToDebuggerAssert( iRet == SCE_OK, ("sceMpegUnRegistStream failed: %x\n", iRet));
		m_pxVideoStream = NULL;
	}

	if (m_bCallsceMpegCreateOK)
	{
		#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
			iRet =
		#endif
			sceMpegDelete( &m_xMpeg );

		Asura_OutputToDebuggerAssert( iRet == SCE_OK, ("sceMpegDelete failed: %x\n", iRet));

		#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
			iRet =
		#endif
			sceMpegRingbufferDestruct( &m_xRingbuffer );

		Asura_OutputToDebuggerAssert( iRet == SCE_OK, ("sceMpegRingbufferDestruct failed: %x\n", iRet));
		m_bCallsceMpegCreateOK = false;
	}
	
	if (m_pxAudioStream)
	{
		#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
			iRet =
		#endif
			sceMpegUnRegistStream( &m_xMpeg, m_pxAudioStream );

		Asura_OutputToDebuggerAssert( iRet == SCE_OK, ("sceMpegUnRegistStream failed: %x\n", iRet));
		m_pxAudioStream = NULL;
	}
	
//	LOCKSEMA

//	if ( dbg_bAudioOn )
//	{
//		if ( m_iLockSema >= 0 )
//		{
//			sceKernelDeleteSema(m_iLockSema);
//		}
//	}

	if (m_bCallsceMpegInitOK)
	{
		#if !defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG
			iRet =
		#endif
			sceMpegFinish();

		Asura_OutputToDebuggerAssert( iRet == SCE_OK, ("sceMpegFinish failed: %x\n", iRet));
		m_bCallsceMpegInitOK = false;
	}

	if (!m_bUsedSASMemory)
	{
		// m_pcPSPStream is the start of the big block so just deallocate that.
//		Asura_Deallocate( m_pcPSPStream );
	}
	delete [] m_ppucDecodedAudioBuffers;
	m_ppucDecodedAudioBuffers = NULL;

	if (m_iFileID >= 0)
	{
		Asura_PSP_FileIO::Close( m_iFileID );
		m_iFileID = ASURA_INVALID_HANDLE;
	}

	if (m_bUsedSASMemory)
	{
//		Asura_PSP_SAS::RestorePermanentSounds();
	}

	// Wait a few frames
	for (int i=0; i<5; i++)
	{
		sceDisplayWaitVblankStart();
	}

	if (g_pucVolatileMemory)
	{
		DeallocateVolatileMemory();
	}
}

void Asura_PSP_Movie::DeallocateVolatileMemory()
{
	if(g_pucVolatileMemory == NULL)
	{
#if defined(ASURA_PSP_UMD) || defined(ASURA_PSP_ARTIST_BUILD)
		Asura_Assert(false, "Volatile memory is already freed.");
#endif	// defined(ASURA_PSP_UMD) || defined(ASURA_PSP_ARTIST_BUILD)
	}
	else
	{
		Asura_PSP_VolatileMemoryManagement::Deallocate(g_uVolatileMemoryHandle);
	}
}

DEBUGVAR float dbg_fMinTimeBeforeSkipAllowed = 0.25f;
DEBUGVAR bool dbg_bPreReadData = true;

void Asura_PSP_Movie::Play()
{
	m_uBytesRead = 0;
	m_iAudioInitFlag = 1;
	m_iNextAudioDecodeBuffer = 0;
	m_iNextAudioPlayBuffer = 0;
	m_iNumberOfFullAudioBuffers = 0;

	Asura_PSP_Movie* pxMovie = this;

	
#ifdef ASURA_PSP_UMD
	// UMD build. Forbid skipping for a given length of time
	if(m_fTimeBeforeSkipAllowed <= dbg_fMinTimeBeforeSkipAllowed)
	{
		m_fTimeBeforeSkipAllowed = dbg_fMinTimeBeforeSkipAllowed;
	}
#else
	// Non-UMD build. Allow skipping.
	m_fTimeBeforeSkipAllowed = 0.0f;
#endif

	// (AW - 9/1/06) - Disable screen clear to get around bug with GU_Clear. Store the old state so we can restore it.
	m_bScreenClearEnabledStore = Asura_Render::GetEnableScreenClear();
	Asura_Render::SetEnableScreenClear(false);

	// Make sure deferred rendering draws it's list immediately
	Asura_PSP_Graphics::SetEnableDeferredImmediateDraw( true );

	if (dbg_bPreReadData)
	{
		// (AW - 13/1/06) - We had a problem with the video stuttering near the start of playback.
		// This resulted in the sound ending up slightly ahead of the video for the duration of the FMV.
		// Filling the ringbuffer before starting to play the FMV seems to fix this, but I have to 
		// admit that I don't really know why.
		while (sceMpegRingbufferAvailableSize(&m_xRingbuffer) > 0 &&
			   m_uBytesRead < m_uStreamSize
			   && !m_bFinished
			   && Asura_PSP_KernelManagement::IsUMDReady())//just drop out if disk ejected - calling code will handle it
		{
			ReadStream();
		}
	}
	m_iStartVSyncCount = sceDisplayGetVcount();
	m_iFramesDecoded = 0;
}

bool Asura_PSP_Movie::Update()
{
	// if we're not ready to display the current frame, wait a bit
	while ((sceDisplayGetVcount() - m_iStartVSyncCount) < (2 * m_iFramesDecoded))
	{
		// if we could be spending this time streaming stuff in,
		// let's do so.
		if (m_uBytesRead < m_uStreamSize)
		{
			ReadStream();
			DecodeAudio();
		}
		else 
		{
			sceDisplayWaitVblankStart();
		}
	}

	Asura_Render::BeginFrame();

	// if we've read in all the data, and the ringbuffer is empty (i.e. nothing left to decode), then the stream has truly finished.
	if (m_uBytesRead >= m_uStreamSize && sceMpegRingbufferAvailableSize(&m_xRingbuffer) ==  uRINGBUFFER_NO_OF_PACKS)
	{
		DecodeAudioEnd();
		
		DecodeVideoEnd();

		m_bFinished = true;
	}
	else // otherwise we'll stream in any data we can, and decode video & audio from the ringbuffer
	{
		if (m_uBytesRead < m_uStreamSize)
		{
			ReadStream();
		}

		DecodeAudio();
		u_int uSafety = 0;
		while ((2 * m_iFramesDecoded) <= (sceDisplayGetVcount() - m_iStartVSyncCount) && uSafety < 4)
		{
			++uSafety;
			DecodeVideo();
		}
		
	}

	if(m_pfnRenderSubtitleCB)
	{
		m_pfnRenderSubtitleCB(m_iFramesDecoded / 29.97f);			
	}

	Asura_Render::EndFrame();

	// don't let the LCD power off
	sceKernelPowerTick(SCE_KERNEL_POWER_TICK_LCDONLY);


	Asura_UserInput::Update();

	// NY - 27 June 2006 - only check for cross, rather than any key, at request of activision.
	//leaving old code in an ifdef, as I'm not sure all projects will want this...
#ifdef ASURA_PSP_MPEG_ANYKEYSKIP
	if ( (Asura_UserInput::DebouncedHasAnyKeyBeenPressed() && ((m_iFramesDecoded / 29.97f) >= m_fTimeBeforeSkipAllowed) )
		|| m_bFinished)
#else // #ifdef ASURA_PSP_MPEG_ANYKEYSKIP
	if ( (Asura_UserInput::DebouncedKeyPressed(ASURA_PSP_KEYID_PAD1_CROSS) && ((m_iFramesDecoded / 29.97f) >= m_fTimeBeforeSkipAllowed) )
		|| m_bFinished)
#endif // #ifdef ASURA_PSP_MPEG_ANYKEYSKIP
	{
		Finish();
	}
	return m_bFinished;
}

void Asura_PSP_Movie::Finish()
{
	if (!m_bFinished)
	{
		DecodeAudioEnd();
		DecodeVideoEnd();
	}
	m_bFinished = true;
	Asura_Render::SetEnableScreenClear(m_bScreenClearEnabledStore);
	Asura_PSP_Graphics::SetEnableDeferredImmediateDraw( false );

	memset(sceGeEdramGetAddr(), 0, 0x200000);

	m_bKillAudioThread = true;
}

void Asura_PSP_Movie::ReadStream( void )
{
	// find out how many packets we have the capacity to read
	const SceInt32 iCapacitySize = sceMpegRingbufferAvailableSize(&m_xRingbuffer);

	// ask the lib to read that many packets
	if( iCapacitySize > 0 )
	{
		int iRequiredBlocks = 32; // magic number from sample code

		if (iRequiredBlocks > iCapacitySize ) iRequiredBlocks = iCapacitySize;

		const SceInt32 iReadSize = sceMpegRingbufferPut(&m_xRingbuffer, iRequiredBlocks, iCapacitySize);
	
		// increment our counter by how much we've read
		if (iReadSize > 0)
		{
			m_uBytesRead += iReadSize * SCE_MPEG_PACK_SIZE;
		}
	}
}



bool Asura_PSP_Movie::DecodeVideo( void )
{
	SceInt32 iRetValue;
	SceInt32 iRGBNum;
	SceInt32 iAvcAttr;

	iRetValue = sceMpegGetAvcAu( &m_xMpeg, m_pxVideoStream, &m_xAvcAu, &iAvcAttr );
	
	if ( iRetValue != SCE_MPEG_OK )
	{
		return false;
	}
		
	// get a pointer to the render target
	SceUChar8* pucDispBuf = (SceUChar8*)( (unsigned int)sceGeEdramGetAddr() + Asura_PSP_Graphics::GetCurrentRenderingBuffer() );

	// after tracking down some errors, it seems this decode routine wants a pointer to our render target pointer...
	iRetValue = sceMpegAvcDecode( &m_xMpeg, &m_xAvcAu, uLCD_FRAME_SIZE, &pucDispBuf, &iRGBNum );
	
	if ( iRetValue != SCE_MPEG_OK )
	{
		return false;
	}
	++m_iFramesDecoded;
	return true;
}

bool Asura_PSP_Movie::DecodeVideoEnd( void )
{
	SceInt32 iRetValue;
	SceInt32 iRGBNum;

	// get a pointer to the render target
	SceUChar8* pucDispBuf = (SceUChar8*)( (unsigned int)sceGeEdramGetAddr() + Asura_PSP_Graphics::GetCurrentRenderingBuffer() );

	// after tracking down some errors, it seems this decode routine wants a pointer to our render target pointer...
	iRetValue = sceMpegAvcDecodeStop( &m_xMpeg, uLCD_FRAME_SIZE, &pucDispBuf, &iRGBNum );
	
	if ( iRetValue != SCE_MPEG_OK )
	{
		return false;
	}

	return true;
}

bool Asura_PSP_Movie::DecodeAudio( void )
{
	SceInt32 iRetValue;
	SceVoid* pAtracBody;

	if (m_pxAudioStream && IsADecodeAudioBufferFree())
	{
		SceUChar8* pucSoundBuf = GetNextDecodedAudioBuffer();
		if (!pucSoundBuf) return false;

		// see if we have anything to decode
		iRetValue = sceMpegGetAtracAu( &m_xMpeg, m_pxAudioStream, &m_xAtracAu, &pAtracBody );
		if ( iRetValue != SCE_MPEG_OK )
		{
			return false;
		}

		iRetValue = sceMpegAtracDecode( &m_xMpeg, &m_xAtracAu, pucSoundBuf, m_iAudioInitFlag );

		if ( iRetValue != SCE_MPEG_OK )
		{
			return false;
		}

		if (!m_iAudioInitFlag)
		{
			m_xMutex.Lock();
			{
				++m_iNumberOfFullAudioBuffers;

				// increment and wrap around
				if ( ++m_iNextAudioDecodeBuffer >= uNO_OF_DECODED_AUDIO_BUFFERS )
				{
					m_iNextAudioDecodeBuffer = 0;
				}
			}
			m_xMutex.Unlock();
		}


		m_iAudioInitFlag = 0;
	}


	return true;
}

bool Asura_PSP_Movie::DecodeAudioEnd( void )
{
	DecodeAudio();


	return true;
}



bool Asura_PSP_Movie::ParseHeader1( const char* pcHeader )
{
	int iRetValue;

	// Get the number of video and audio streams
	// This needs the psmf header structure
	// Load the psmf prx for this short time we need it.
	SceUID uID;
	SceKernelLMOption uMode;
	uMode.size = sizeof(uMode);
    uMode.mpidtext = SCE_KERNEL_PRIMARY_USER_PARTITION;
    uMode.mpiddata = SCE_KERNEL_PRIMARY_USER_PARTITION;
    uMode.position = SCE_KERNEL_LM_POS_Low;
    uMode.access   = SCE_KERNEL_LMWO_ACCESS_Noseek;

	iRetValue = Asura_PSP_FileIO::LoadStartModule(&uID,"module/psmf.prx",&uMode);

	if (iRetValue < 0)
	{
		// TODO: Do we need to unload-stop here if the load-start failed?
		return false;
	}

	ScePsmf xPsmfData;
	iRetValue = scePsmfSetPsmf(&xPsmfData, pcHeader);

	if ( iRetValue < 0 )
	{
		Asura_PSP_FileIO::UnloadStopModule(&uID);
		return false;
	}

	m_iNumAudioStreams = scePsmfGetNumberOfSpecificStreams(&xPsmfData, SCE_MPEG_STREAM_ATRAC);
	m_iNumVideoStreams = scePsmfGetNumberOfSpecificStreams(&xPsmfData, SCE_MPEG_STREAM_M4AVC);

	Asura_Assert(m_iNumVideoStreams > 0, "No video streams in FMV.");

	// All psmf operations complete, unload the module.
	iRetValue = Asura_PSP_FileIO::UnloadStopModule(&uID);
	if ( iRetValue < 0 )
	{
		return false;
	}

	return true;
}

bool Asura_PSP_Movie::ParseHeader2( const char* pcHeader )
{

	// show the mpeg library the header, and query when the data stream actually starts
	int iRetValue = sceMpegQueryStreamOffset( &m_xMpeg, pcHeader, &m_uStreamOffset );
	
	if ( iRetValue < 0 )
	{
		return false;
	}

	// get the size of the stream too
	iRetValue = sceMpegQueryStreamSize( pcHeader, &m_uStreamSize );

	if ( iRetValue < 0 )
	{
		return false;
	}

	// well, that's okay then
	return true;
}



bool Asura_PSP_Movie::IsADecodeAudioBufferFree( void )
{
    if (m_iNumberOfFullAudioBuffers >= uNO_OF_DECODED_AUDIO_BUFFERS - 2)
	{
		// no free buffers, I'm afraid
		return false;
	}

	if (m_iNextAudioPlayBuffer == m_iNextAudioDecodeBuffer && m_iNumberOfFullAudioBuffers)
	{
		return false;
	}

	return true;
}

SceUChar8* Asura_PSP_Movie::GetNextDecodedAudioBuffer( void )
{
	SceUChar8* pucBuffer = m_ppucDecodedAudioBuffers[ m_iNextAudioDecodeBuffer ];

	return pucBuffer;
}

SceInt32 Asura_PSP_Movie::RingbufferCallBack( SceVoid* pBuf, SceInt32 iNum, SceVoid* pCallbackData )
{
	Asura_PSP_Movie* pxMovie = reinterpret_cast<Asura_PSP_Movie*>(pCallbackData);

	const int iLOADBLOCKSIZE = SCE_MPEG_PACK_SIZE * 8;

	static u_char aucMemBuff[iLOADBLOCKSIZE] __attribute__((aligned(16)));

	int iBytesRead = 0;

	u_char* pucVolBuf = static_cast<u_char*>(pBuf);

	// NY - 26 June 2006 - we can't read straight from UMD to volatile memory, so read to main memory then memcpy over
	// going one pack at a time for now, to minimse main memory usage - this may be way too small...
	// upped to 8 packs for umd
	while (iBytesRead < SCE_MPEG_PACK_SIZE * iNum)
	{
		int iBytesToRead = Asura_Maths::Min(SCE_MPEG_PACK_SIZE * iNum - iBytesRead, iLOADBLOCKSIZE);
		if (pxMovie->m_bFinished)
		{
			break;
		}
		// unlock our mutex for fileaccess
        // (MH - 18.07.2007) moved the initial seek to here from the initialisation phase to reduce mutex unlock/lock occurrences.
		if (!pxMovie->m_bDoneInitialSeek)
		{
			Asura_PSP_MovieUnlockMutex();
			int iRet = Asura_PSP_FileIO::Lseek( pxMovie->m_iFileID, pxMovie->m_uStreamOffset, SCE_SEEK_SET );
			Asura_PSP_MovieLockMutex(__LINE__);
			if (iRet < 0)
			{
				Asura_OutputToDebugger(("Movie initial seek failed\n"));
				return 0;
			}
			pxMovie->m_bDoneInitialSeek = true;
		}
		Asura_PSP_MovieUnlockMutex();
		SceInt32 iReadSize = Asura_PSP_FileIO::Read( pxMovie->m_iFileID, aucMemBuff, iBytesToRead, false );
		Asura_PSP_MovieLockMutex(__LINE__);
		if ( iReadSize < 0 )
		{
			return 0;
		}
		if ( iReadSize == 0)
		{
			break;
		}
		if ( (iReadSize % SCE_MPEG_PACK_SIZE) != 0)
		{
			return -1;
		}

		if (pxMovie->m_bFinished)
		{
			break;
		}
		iBytesRead += iReadSize;
		Asura_Mem::Copy(pucVolBuf, aucMemBuff, iBytesToRead);
		pucVolBuf += iBytesToRead;
	}

	// return number of packs loaded
	return iBytesRead / SCE_MPEG_PACK_SIZE;
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_Movie::DecodeAudio
// Author       : Andy Tate
// Date         : 24-Jul-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_Movie::DecodeAudio (short* psCurrentBuffer,
								   float& fVolume)
{
	if(m_bKillAudioThread)
	{
		return(false);
	}
	else if (m_iNumberOfFullAudioBuffers > 0)
	{
		const SceUChar*	pucDecodedBuffer	= m_ppucDecodedAudioBuffers [m_iNextAudioPlayBuffer];
		const u_int		uSize				= ASURA_PSP_SA_NUM_SAMPLES * ASURA_PSP_SA_NUM_STEREO_CHANNELS * sizeof(short);

		m_xMutex.Lock();
		{
			// Copy the data...
			Asura_Mem::Copy(psCurrentBuffer, pucDecodedBuffer, uSize);

			// Decrease the buffer count...
			m_iNumberOfFullAudioBuffers--;

			// increment and wrap around
			if ( ++(m_iNextAudioPlayBuffer) >= uNO_OF_DECODED_AUDIO_BUFFERS )
			{
				m_iNextAudioPlayBuffer = 0;
			}

			// Store the volume...
			fVolume	= m_fAudioVolume;
		}
		m_xMutex.Unlock();
		return(true);
	}
	else
	{
		return(false);
	}
}

/// End of function 'Asura_PSP_Movie::DecodeAudio'. ///


static void Asura_PSP_MovieLockMutex( int iLine )
{
	g_xMovieMutex.Lock();
	g_iMovieMutexLockLine = iLine;
}

static void Asura_PSP_MovieUnlockMutex()
{
	Asura_Assert(g_iMovieMutexLockLine, "mutex lock line not set when locked?");
	g_iMovieMutexLockLine = 0;
	g_xMovieMutex.Unlock();
}

#endif//def ASURA_SUPPORT_PSP_MPEG
