//
// Asura_PSP_TextureManagement.cpp
//
// Created by KJFL - Tuesday, February 10, 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PSP_TextureManagement.h"

// Engine independent.
#include "Core/Asura_NameTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_ScreenShot_System.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_MovieToTexture_System.h"
#include "Rendering/Asura_PSP_TextureHeader.h"

// PSP headers
#include "Debugging/Asura_PSP_GPIO.h"
#include "Rendering/Asura_PSP_Graphics.h"

Asura_PSP_Texture* g_apxTextureStorage[ASURA_MAX_NO_OF_TEXTURES];

#include <kernel.h>

// RJT: DMA texture upload as opposed to memory copy
#ifdef ASURA_PSP_DEFERREDRENDER
	#define ASURA_PSP_TEXTURE_DMA
#endif

DEBUGVAR bool dbg_bHiResScreenshotForceNoMipmapping = false;




//
// Asura_Texture_Management.
//

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Texture_Management::LoadTexture
// PARAMETERS	Name of texture
//				Index to load to
// RETURNS		None
// DESCRIPTION	Loads a texture to memory and keeps a record of its location
// AUTHOR		KJL
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Texture_Management::Platform_LoadTexture( const Asura_String& name, int index)
{
	// KJFL - 2004/02/10 17:41:56 - intend to only load from a resource file on the PSP
}
// End Of Function 'Asura_Texture_Management::LoadTexture'

void Asura_Texture_Management::Platform_LoadTextureFromResource(Asura_Handle fileHandle, int fileSize, int index, u_int uChunkFlags)
{
	Asura_PSP_Texture*& pxTexture = g_apxTextureStorage[index];

	if (pxTexture)
	{
		Asura_Assert( false, "Texture already loaded to that index" );
		delete pxTexture;
	}

	// KJFL - 2004/02/11 09:31:59 - make a new texture. This will allocate memory for the header,
	// but not yet the memory required for the textures themselves
	// TODO - might it be better to use a mempool for Asura_PSP_Texture? 
	// currently I'm storing the actual texture memory elsewhere because a) it needs to be aligned
	// and b) it makes it more flexible to cope with changes in the PSP hardware
	pxTexture = new Asura_PSP_Texture();

	if ( !pxTexture )
	{
		Asura_Assert( false, "Unable to allocate memory for new texture header" );
		return;
	}

	// read in header
	Asura_FileAccess::ReadData( fileHandle, pxTexture, sizeof(Asura_PSP_TextureHeader) );

	int iDataSize = fileSize - sizeof(Asura_PSP_TextureHeader);

	Asura_Assert( iDataSize>0, "invalid data size");

	int iTextureSize;
	int iClutSize;
	
	switch (pxTexture->m_ucTextureFormatID)
	{
		default:
		{
			Asura_Assertf( false, ("Unknown texture format id <%d>", pxTexture->m_ucTextureFormatID) );
			Asura_FileAccess::SkipData( fileHandle, iDataSize );
			return;
		}
		case ASURA_PSP_TEXTUREFORMAT_8888:
		{
			iTextureSize = pxTexture->m_usHeight * pxTexture->m_usWidth * 4;
			iClutSize = 0;
			break;
		}
		case ASURA_PSP_TEXTUREFORMAT_IDX8:
		{
			iTextureSize = pxTexture->m_usHeight * pxTexture->m_usWidth;
			iClutSize = 256 * sizeof(u_int);
			break;
		}
		case ASURA_PSP_TEXTUREFORMAT_IDX4:
		{
			iTextureSize = (pxTexture->m_usHeight * pxTexture->m_usWidth) >> 1;
			iClutSize = 16 * sizeof(u_int);
			break;
		}
	}

	u_int uMips = pxTexture->m_ucNumberOfMipmaps;
	int iSize = iTextureSize + iClutSize;
	int iShift = 2;
	for (u_int u=1; u<=uMips; u++)
	{
		iSize += iTextureSize >> iShift;
		iShift += 2;
	}

	Asura_Assert( iDataSize == iSize, "texture data mismatch" );
    
	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE);
	{
		// KOR - 24/08/2004 - needs 16 byte alignment (for both texture and CLUT? needs testing...)
		Asura_ResourceAllocateMemoryBlock(pxTexture->m_pTextureBase, iSize, ASURA_ALIGN_16_BYTE);

		int iOffset = 0;
		iSize = iTextureSize;
		for (u_int u=0; u<=uMips; u++)
		{
			Asura_FileAccess::ReadData( fileHandle, (char*)pxTexture->m_pTextureBase + iOffset, iSize );
			iOffset += iSize;
			iSize >>= 2;
		}

		if (iClutSize)
		{
			Asura_ResourceAllocateMemoryBlock(pxTexture->m_pClutBase, iClutSize, ASURA_ALIGN_16_BYTE);
			Asura_FileAccess::ReadData(fileHandle, pxTexture->m_pClutBase, iClutSize);
		}
		sceKernelDcacheWritebackAll();
	}
	Asura_ResourcePopType();
}

void Asura_Texture_Management::ReloadTexture(int /*index*/)
{
	//Not supported on PSP
}

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Texture_Management::Platform_DestroyTexture
// PARAMETERS	Index of texture to destroy
// RETURNS		None
// DESCRIPTION	Deletes the record of the loaded texture
// AUTHOR		KJL
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Texture_Management::Platform_DestroyTexture( int iTexID )
{
	delete g_apxTextureStorage[iTexID];
	g_apxTextureStorage[iTexID] = 0;
}
// End Of Function 'Asura_Texture_Management::Platform_DestroyTexture'

void Asura_Texture_Management::Platform_Initialise( void )
{
	for (u_int u=0; u<ASURA_MAX_NO_OF_TEXTURES; u++)
	{
		g_apxTextureStorage[u] = 0;
	}
}

void Asura_Texture_Management::Platform_Shutdown( void )
{
  
}

void Asura_Texture_Management::Platform_Reset( void )
{
}

//
// Asura_PSP_Texture.
//

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_PSP_Texture::FindPaletteIndex
// PARAMETERS	Colour we want to find in the palette
// RETURNS		Index if is in the palette, -1 if is not
// DESCRIPTION	Finds the index in the palette for a given colour
// AUTHOR		JG
/////////////////////////////////////////////////////////////////////////////////////////
int Asura_PSP_Texture::FindPaletteIndex(const unsigned int uColor)
{
	int iIndex = -1; 
	
	if (m_ucTextureFormatID == ASURA_PSP_TEXTUREFORMAT_IDX8)
	{
		unsigned int* puPalette = static_cast<unsigned int*>(m_pClutBase); 

		for (unsigned int u=0; u < 256; u++)
		{
			if (puPalette[u] == uColor)
			{
				iIndex = u; 
				break; 
			}
		}
	}
	else
	{
		Asura_Assert(false, "Incorrect texture format, colour could not be found"); 
	}

	return iIndex; 
}


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_PSP_Texture::TintPaletteEntriesWithColor
// PARAMETERS	Alpha key to identify the colours to tint, colour used to tint
// RETURNS		NA
// DESCRIPTION	Tint the palette entries flag in the alpha channel with the key using the colour
// AUTHOR		JG
/////////////////////////////////////////////////////////////////////////////////////////
DEBUGVAR unsigned int dbg_uPaletteAlphaDistance = 3;

void Asura_PSP_Texture::TintPaletteEntriesWithColor(const unsigned int uAlphaKey, const unsigned int uColor)
{
	if (m_ucTextureFormatID == ASURA_PSP_TEXTUREFORMAT_IDX8)
	{
		unsigned int* puPalette = static_cast<unsigned int*>(m_pClutBase); 

		for (unsigned int u=0; u < 256; u++)
		{
			unsigned int uAlphaColor = puPalette[u] >> 24;

			if (abs(uAlphaColor - uAlphaKey) < dbg_uPaletteAlphaDistance)
			{
				unsigned int uPaletteColor = puPalette[u];
				
				unsigned int uRed	= (uPaletteColor & 0x000000ff); 
				unsigned int uGreen = (uPaletteColor & 0x0000ff00) >> 8; 
				unsigned int uBlue	= (uPaletteColor & 0x00ff0000) >> 16;  

				unsigned int uBrightest = Asura_Maths::Max(Asura_Maths::Max(uRed, uGreen), uBlue); 
				float fIntensity = uBrightest / 255.0f;

				uRed	= (uColor & 0x000000ff); 
				uGreen	= (uColor & 0x0000ff00) >> 8; 
				uBlue	= (uColor & 0x00ff0000) >> 16;  

				unsigned int uMax = Asura_Maths::Max(Asura_Maths::Max(uRed, uGreen), uBlue); 
				unsigned int uSaturation = 0xff - uMax; 
				uRed += uSaturation; 
				uGreen += uSaturation; 
				uBlue += uSaturation; 

				unsigned int uTintedColor = 0xff000000 | (Asura_Maths::FloatToInt_Near(uBlue * fIntensity) << 16) | (Asura_Maths::FloatToInt_Near(uGreen * fIntensity) << 8) | Asura_Maths::FloatToInt_Near(uRed * fIntensity);
				
				puPalette[u] = (uAlphaColor << 24) | (uTintedColor & 0x00ffffff); 
			}
		}
	}
	else
	{
		Asura_Assert(false, "Incorrect texture format, palette could not be tint"); 
	}
}

//
// Asura_PSP_TextureManagement
//

int		Asura_PSP_TextureManagement::s_iLastTextureIndex = ASURA_TEXTURE_INVALID; 
void*	Asura_PSP_TextureManagement::s_pLastPaletteData = NULL; 
extern Asura_PSP_Texture* g_apxTextureStorage[];

DEBUGVAR bool dbg_bEnableTextures = true;

extern "C"
{
	void* sceGeEdramGetAddr();
};

static int g_iLastNonNoneTexture = ASURA_TEXTURE_NONE;

static int g_iNumMipsMultiplier = 1;
static int g_iNumMipsOffset = 0;

// (SB - 26/02/07) Added this #define in engine config for projects where special effects take up all or most VRAM
#ifdef ASURA_PSP_NO_TEXTURES_IN_VRAM
DEBUGVAR bool dbg_bTexturesInVRAM = false; 
#else
DEBUGVAR bool dbg_bTexturesInVRAM = true; 
#endif
DEBUGVAR bool dbg_bAddNoisePatternToTexturesInVRAM = false; 

enum 
{
	ASURA_PSP_TEXTURE_IGNORE_PALETTE_DATA = 0,
	ASURA_PSP_TEXTURE_USE_PALETTE_DATA
}; 

void Asura_PSP_TextureManagement::SetTextureIndex(const int iTextureIndex)
{
	UploadTexture( iTextureIndex, ASURA_PSP_TEXTURE_USE_PALETTE_DATA ); 
}

void Asura_PSP_TextureManagement::RequestTextureIndex( const int iTextureIndex )
{
	if( ASURA_TEXTURE_INVALID == iTextureIndex )
	{
		s_iLastTextureIndex = ASURA_TEXTURE_INVALID;
	}
	else if( (iTextureIndex != s_iLastTextureIndex) || (ASURA_TEXTURE_INVALID == s_iLastTextureIndex) )
	{
		SetTextureIndex( iTextureIndex );
	}
}

void Asura_PSP_TextureManagement::RequestTextureData( const int iTextureIndex )
{
	if( ASURA_TEXTURE_INVALID == iTextureIndex )
	{
		s_iLastTextureIndex = ASURA_TEXTURE_INVALID;
	}
	else if( (iTextureIndex != s_iLastTextureIndex) || (ASURA_TEXTURE_INVALID == s_iLastTextureIndex) )
	{
		SetTextureData( iTextureIndex );
	}
}

void Asura_PSP_TextureManagement::RequestPaletteDataForTexture( const int iTextureIndex, const void* pPaletteData )
{
	Asura_Assert( iTextureIndex != ASURA_TEXTURE_NONE, "Invalid texture specified when requesting a palette" );
	Asura_PSP_Texture* pxTexture = g_apxTextureStorage[iTextureIndex];
	Asura_Assert( pxTexture, "Invalid texture index!" ); 

	switch (pxTexture->m_ucTextureFormatID)
	{
		default:
		{
			Asura_Assertf( false, ("Invalid texture format id <%d>", pxTexture->m_ucTextureFormatID) );
			break;
		}
		case ASURA_PSP_TEXTUREFORMAT_IDX8:
		{
			RequestPaletteData( 256, pPaletteData ); 
			break;
		}
		case ASURA_PSP_TEXTUREFORMAT_IDX4:
		{
			RequestPaletteData( 16, pPaletteData ); 
			break; 
		}
	}
}

void Asura_PSP_TextureManagement::RequestPaletteData( const unsigned int uNumberOfPaletteEntries, const void* pPaletteData )
{
	if( NULL == pPaletteData )
	{
		s_pLastPaletteData = NULL; 
	}
	else if( (pPaletteData != s_pLastPaletteData) || (NULL == s_pLastPaletteData) )
	{
		SetPaletteData( uNumberOfPaletteEntries, pPaletteData ); 
	}

}

void Asura_PSP_TextureManagement::SetTextureData( const int iTextureIndex )
{
	UploadTexture( iTextureIndex, ASURA_PSP_TEXTURE_IGNORE_PALETTE_DATA ); 
}

void Asura_PSP_TextureManagement::SetPaletteData( const unsigned int uNumberOfPaletteEntries, const void* pPaletteData )
{
	Asura_Assert( uNumberOfPaletteEntries == 16 || uNumberOfPaletteEntries == 256, "Unsupported palette data!" ); 
	Asura_Assert( pPaletteData, "Palette data can not be NULL" ); 

	// Upload the palette data
	GU_ClutMode( SCEGU_PF8888, 0, 0xff, 0 );
	GU_ClutLoad( uNumberOfPaletteEntries / 8, pPaletteData );
}

void Asura_PSP_TextureManagement::UploadTexture( int iTextureIndex, bool bUploadPaletteDataFromTexture /*=true*/ )
{
	Asura_Assert(Asura_Render::IsRendering() || Asura_PSP_Graphics::IsInFrameForAssert(), 
		"calling renderstate function outside of BeginFrame/EndFrame()");

	if( !dbg_bEnableTextures || Asura_PSP_GPIO::GetGPI( ASURA_PSP_GPI_UNTEXTURED ) )
	{
		GU_Disable( SCEGU_TEXTURE );
		return;
	}

	if( iTextureIndex == ASURA_TEXTURE_NONE )
	{
		// send blank texture
		GU_Disable( SCEGU_TEXTURE );
		g_iLastNonNoneTexture = s_iLastTextureIndex;
		s_iLastTextureIndex = iTextureIndex;
	}
	else
	{
		// TODO - is there a nicer way to cache this?
		if( s_iLastTextureIndex == ASURA_TEXTURE_NONE )
		{
			GU_Enable( SCEGU_TEXTURE );
			if( g_iLastNonNoneTexture == iTextureIndex )
			{
				// We're done as it's already loaded
				s_iLastTextureIndex = iTextureIndex;
				return;
			}
		}

		s_iLastTextureIndex = iTextureIndex;

		Asura_PSP_Texture* pxTexture = g_apxTextureStorage[iTextureIndex];
	
		if (!pxTexture)
		{
			// send blank texture
			GU_Disable(SCEGU_TEXTURE);
			return;
		}
		else
		{
			GU_Enable(SCEGU_TEXTURE);
		}

		u_char ucNumMipmaps = pxTexture->m_ucNumberOfMipmaps;
		if (Asura_ScreenShot_System::IsTakingHiResScreenshot() && dbg_bHiResScreenshotForceNoMipmapping)
		{
			ucNumMipmaps = 0;
		}
		GU_TexLevelMode(SCEGU_LOD_SLOPE, -(ucNumMipmaps * g_iNumMipsMultiplier + g_iNumMipsOffset));

		const int iTexbufMode = SCEGU_TEXBUF_FAST;

		u_int uPixelSizeNibbles = 2;
		u_int uWidth = pxTexture->m_usWidth;
		u_int uHeight = pxTexture->m_usHeight;
		u_int uSize = pxTexture->m_usWidth; // Does this need to be different to the width ever? Probably if the mip is ever small enough, but code is needed to handle that if true.
		u_int uOffset = 0;

		switch (pxTexture->m_ucTextureFormatID)
		{
			default:
			{
				Asura_Assertf( false, ("Unknown texture format id <%d>", pxTexture->m_ucTextureFormatID) );
				return;
			}
			case ASURA_PSP_TEXTUREFORMAT_8888:
			{
				GU_TexMode(SCEGU_PF8888, ucNumMipmaps, SCEGU_SINGLE_CLUT, iTexbufMode);
				uPixelSizeNibbles = 8;
				break;
			}
			case ASURA_PSP_TEXTUREFORMAT_IDX8:
			{
				GU_TexMode(SCEGU_PFIDX8, ucNumMipmaps, SCEGU_SINGLE_CLUT, iTexbufMode);
				if( bUploadPaletteDataFromTexture )
				{
					SetPaletteData( 256, pxTexture->m_pClutBase ); 
				}
				uPixelSizeNibbles = 2;
				break;
			}
			case ASURA_PSP_TEXTUREFORMAT_IDX4:
			{
				// KOR - TODO - CLUT for each mip level (only really possible for 4-bit, should improve image quality)
				GU_TexMode(SCEGU_PFIDX4, ucNumMipmaps, SCEGU_SINGLE_CLUT, iTexbufMode);
				if( bUploadPaletteDataFromTexture )
				{
					SetPaletteData( 16, pxTexture->m_pClutBase ); 
				}
				uPixelSizeNibbles = 1;
				break;
			}
		}

		// Are we using copying our textures into VRAM?
		if (dbg_bTexturesInVRAM)
		{
			// Valid handler?
			if (!Asura_PSP_VRAMManagement::IsValidHandler(pxTexture->m_iVRAMHandler))
			{
				// Invalid handler, let's try to allocate space in VRAM for this texture and mipmaps

				// Do we know the texture size (including mipmaps)
				if (!pxTexture->m_uTextureTotalSize)
				{
					// Calculate the total data size 
					pxTexture->CalculateTotalTextureSize(); 
				}

				// Allocate enough space for this texture and mipmaps in VRAM
				pxTexture->m_iVRAMHandler = Asura_PSP_VRAMManagement::AllocateVRAM(pxTexture->m_uTextureTotalSize); 
				
				// Get a pointer to the new allocated space in VRAM 
				char* pVRAM = Asura_PSP_VRAMManagement::GetVRAM(pxTexture->m_iVRAMHandler); 
				
				// Upload all mipmaps to VRAM
				for (u_char u=0; u<=ucNumMipmaps; u++)
				{
					// Size of this level
					unsigned int uTextureSize = (uWidth * uHeight * uPixelSizeNibbles) >> 1; 

					// Succesfull allocation?
					if (pVRAM)
					{
						// Copy the data
#ifdef ASURA_PSP_TEXTURE_DMA
						GU_CopyImage( SCEGU_PF8888, 0, 0, 128, ( uTextureSize >> 9 ) + 1, 128, static_cast< u_char * >( pxTexture->m_pTextureBase ) + uOffset, 0, 0, 128, pVRAM + uOffset );

						// Synchronise the transfer
//						GU_Signal( SCE_GE_SIGNAL_DRAWSYNC, 0 );
#else
						Asura_Mem::Copy(pVRAM + uOffset, static_cast<char*>(pxTexture->m_pTextureBase) + uOffset, uTextureSize); 
#endif
					}

					// Send data to GU (this frame still using the main memory copy, to avoid cache issues)
					GU_TexImage(u, uWidth, uHeight, uSize, static_cast<char*>(pxTexture->m_pTextureBase) + uOffset);

					// Next mipmap level
					uOffset += uTextureSize;
					uWidth >>= 1;
					uHeight >>= 1;
					uSize >>= 1;
				}

				// Adding some noise to the textures in VRAM so they look different to main memory (parallel lines pattern)
				// It helps to see what the manager is doing
				if (dbg_bAddNoisePatternToTexturesInVRAM)
				{
					if (pVRAM)
					{
						// Noise pattern, so we can see which textures are being rendered from VRAM 
						for (unsigned int uNoise = 0; uNoise < uOffset; uNoise += 8)
						{
							pVRAM[uNoise] = 0; 
						}
					}
				}
			}
			else
			{
				// Valid handler, let's upload the texture
				for (u_char u=0; u<=ucNumMipmaps; u++)
				{

					// Get a pointer to texture data in VRAM
					char* pVRAM = Asura_PSP_VRAMManagement::GetVRAM(pxTexture->m_iVRAMHandler); 			

					// Is it a valid pointer?
					if (pVRAM)
					{
						// Yes, upload the texture from VRAM 
						GU_TexImage(u, uWidth, uHeight, uSize, pVRAM + uOffset);
					}
					else
					{
						// No, didn't fit in VRAM when we try to allocate before
						// Upload from main memory
						GU_TexImage(u, uWidth, uHeight, uSize, static_cast<char*>(pxTexture->m_pTextureBase) + uOffset);
					}

					// Next mipmap level
					uOffset += (uWidth * uHeight * uPixelSizeNibbles) >> 1;
					uWidth >>= 1;
					uHeight >>= 1;
					uSize >>= 1;
				}
			}
		}
		else
		{
			// Upload txture data from main memory
			for (u_char u=0; u<=ucNumMipmaps; u++)
			{
				GU_TexImage(u, uWidth, uHeight, uSize, static_cast<char*>(pxTexture->m_pTextureBase) + uOffset);
				uOffset += (uWidth * uHeight * uPixelSizeNibbles) >> 1;
				uWidth >>= 1;
				uHeight >>= 1;
				uSize >>= 1;
			}
		}

		// Texture wrapping/clamping
		GU_TexWrap(
			( Asura_Texture_Management::GetTextureFlags( iTextureIndex ) & AsuraTextureFlag_ClampU ) ? SCEGU_CLAMP : SCEGU_REPEAT,
			( Asura_Texture_Management::GetTextureFlags( iTextureIndex ) & AsuraTextureFlag_ClampV ) ? SCEGU_CLAMP : SCEGU_REPEAT
		);

		// TEMP:
//		GU_TexLevelMode( SCEGU_LOD_AUTO, -0.75f );
//		GU_TexFilter( ( ucNumMipmaps > 1 ) ? SCEGU_LINEAR_MIPMAP_LINEAR : SCEGU_LINEAR, SCEGU_LINEAR );

		// RJT NOTE: Required for SceeGu (as it doesn't flush per-mip like the original libgu!)
		GU_TexFlush();
	}
}

// (SB - 26/02/07) Remove cached textures from VRAM
void Asura_PSP_TextureManagement::InvalidateTexturesInVRAM()
{
	for (unsigned int i = 0; i < ASURA_MAX_NO_OF_TEXTURES; i++)
	{
		if (g_apxTextureStorage[i])
		{
			g_apxTextureStorage[i]->m_iVRAMHandler = Asura_PSP_VRAMManagement::ASURA_PSP_VRAM_INVALID;
		}
	}
}

const Asura_PSP_Texture* Asura_PSP_TextureManagement::GetTexture( const int iTextureIndex )
{
	Asura_PSP_Texture* pxTexture = NULL;

	if ( iTextureIndex >= 0 && iTextureIndex < ASURA_MAX_NO_OF_TEXTURES )
	{
		pxTexture = g_apxTextureStorage[iTextureIndex];
	}
	else
	{
		Asura_Assert( false, "Invalide texture index!" ); 
	}

	return pxTexture; 
}
