//////////////////////////////////////////////////////////////////////////////
//
// Asura_PSP_Sound.h
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     class to manage PSP sound data, for both streamed and non-streamed
//     sounds
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ASURA_PSP_SOUND_H___
#define ASURA_PSP_SOUND_H___ 1

#include "Sound/Asura_Sound.h"

static const u_int uASURA_PSP_MAX_VOICES = 16;
static const u_int uASURA_PSP_STREAMING_MAX_FILENAME_LENGTH = 256;
static const u_int uASURA_PSP_STREAMING_MAX_FILES = 2;

enum ASURA_PSP_LOADED_TYPE
{
	ASURA_PSP_LT_NOT_LOADED,
	ASURA_PSP_LT_RESIDENT,
	ASURA_PSP_LT_STREAMED,
};

struct Asura_PSP_SAS_LoadedSound
{
	// Raw data
	u_char*	m_pucData;
	u_int	m_bAllocatedOwnData:1;	//  set iff we need to call Asura_DeallocateAlign on unload
	u_int	m_uSize:31;
	float	m_fPitchModifier;
};

struct Asura_PSP_Streamed_LoadedSound
{
	const char* m_szFilename;
	u_int		m_uOffset;
	u_int		m_uSize;
};

struct Asura_PSP_LoadedSound
{
	ASURA_PSP_LOADED_TYPE m_eLoadedType;
	union
	{
		Asura_PSP_SAS_LoadedSound		m_xSASLoadedSound;
		Asura_PSP_Streamed_LoadedSound	m_xStreamedLoadedSound;
	};
};

struct Asura_PSP_ActiveSound
{
	int	m_iLoadedSoundIndex;
};


class Asura_PSP_Sound
{
public:

	static void Initialise();
	static void Reset();
	static void Shutdown();
	static void SetResetPoint();

	// Standard playing methods
	static bool CreateSound(int iLoadedSoundID, Asura_Handle iSoundHandle);
	static bool PlaySound(Asura_Handle iSoundHandle, bool bLooping);
	static bool IsSoundPlaying(Asura_Handle iSoundHandle);
	static bool IsSoundPaused(Asura_Handle iSoundHandle);
	static bool StopSound(Asura_Handle iSoundHandle);
	static bool DestroySound(Asura_Handle iSoundHandle);
	static bool PauseSound(Asura_Handle iSoundHandle);
	static bool UnpauseSound(Asura_Handle iSoundHandle);
	// Property setting
	static bool SetSoundVolumePanAndFade(Asura_Handle iSoundHandle, float fVolume, float fPan, float fFade);
	static bool SetSoundPitch(Asura_Handle iSoundHandle, float fPitch);

	static ASURA_PSP_LOADED_TYPE GetLoadedType(int iLoadedIndex);
	static bool IsResident(int iLoadedIndex) { return GetLoadedType(iLoadedIndex) == ASURA_PSP_LT_RESIDENT; }
	static bool IsStreamed(int iLoadedIndex) { return GetLoadedType(iLoadedIndex) == ASURA_PSP_LT_STREAMED; }
	static int GetLoadedIndex(u_int uVoice);

	// GetXXXSound() functions Assert and return NULL if sound is not loaded or is wrong type.
	static const Asura_PSP_SAS_LoadedSound* GetResidentSound(int iLoadedIndex);
	static const Asura_PSP_Streamed_LoadedSound* GetStreamedSound(int iLoadedIndex);

	static void SoundLoaded( int iLoadedIndex, u_char* pucData, bool bAllocatedOwnData, float m_fPitchModifier, u_int m_uSize );
	static void SoundUnloaded( int iLoadedIndex );

	static void InitStreamedSounds( const char* szFilename );
	static void AddStreamingSoundData( int iSoundID, int iOffset, int iSize, Asura_Hash_ID uSoundHash );
	static Asura_Handle GetMusicSoundHandle( void ) { return s_xMusicSoundHandle; }

private:

	static Asura_PSP_ActiveSound s_axActiveSounds[uASURA_PSP_MAX_VOICES];
	static Asura_PSP_LoadedSound s_axLoadedSounds[ASURA_MAX_NUMBER_OF_LOADED_SOUNDS];
	static char s_aszFilenames[uASURA_PSP_STREAMING_MAX_FILES][uASURA_PSP_STREAMING_MAX_FILENAME_LENGTH];
	static bool s_abFileIsPermanent[uASURA_PSP_STREAMING_MAX_FILES];
	static const char* s_szCurFilename;
	static Asura_Handle s_xMusicSoundHandle;
};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ASURA_PSP_SOUND_H___
