//
// Asura_Platform_Beam.cpp
//
// Created by KJFL - Tuesday, March 30, 2004
// First-pass implementation by Andrew Weinkove - 16/04/04. Based on PC implementation.
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "SpecialFX/Asura_Beam_System.h"

#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"

#include "Rendering/Asura_PSP_Graphics.h"
#include "Rendering/Asura_PSP_RenderEnvironment.h"
#include "Rendering/Asura_PSP_TextureManagement.h"
#include <kernel.h>

void Asura_Beam_System::Platform_Initialise()
{
}

void Asura_Beam_System::Platform_Shutdown()
{
}

void Asura_Beam_System::Platform_RenderStart(void)
{
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);

#ifndef ASURA_BEAMS_ALL_MODES
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
#endif
	
	// get a suitable texture
	Asura_PSP_TextureManagement::RequestTextureIndex( s_iTiledTextureTextureID );
}

void Asura_Beam_System::Platform_RenderEnd(void)
{
}

struct Asura_PSP_RenderBeam_Vertex
{
	// fully 32bit - 24 bytes
	float			m_fU;
	float			m_fV;
	unsigned		m_uColour;
	float			m_afPosition[3];
};

const int iASURA_PSP_BEAM_VERTEX_FORMAT = SCEGU_TEXTURE_FLOAT | SCEGU_VERTEX_FLOAT | SCEGU_COLOR_PF8888;

// KOR - 30/09/2004 - Dodgy buffer. Yuck. To be replaced with proper class when I get the time (unify with beam renderer)
#define ASURA_PSP_MAX_BEAMS_PER_FRAME (128)

static Asura_PSP_RenderBeam_Vertex g_axVertices[ASURA_PSP_MAX_BEAMS_PER_FRAME * 8] __attribute__ ((aligned(16)));
static u_int g_uOffset = 0;

void FlushBeams()
{
	g_uOffset = 0;
}

// Clipping precisely to the near plane seems to not work that well, maybe because of rounding errors
// or perhaps the beam caps / width of beam, so we will push the effective near plane forward by a small amount
DEBUGVAR float dbg_fNearPlaneBeamClipAdjust = 0.1f;

// A position on the each plane to clip against
// Near plane, other 4 planes, no far plane clipping required
static Asura_Vector_3 g_axBeamClipPositions[5];
// Corresponding normals to these positions
static Asura_Vector_3 g_axBeamClipNormals[5];

// Ripped from PSP_RenderEnviroment_Zone.
// KOR - 16/03/2006 - actual size is 4096. Was set to 4095, but still had rare drop-outs, probably due to floating point errors
// Hopefully this is conservative enough to catch everything without too much performance hit.
// JG - 28/04/2006: it seems as is not conservative enough. Some further testing in Gun + some reference code from Core suggest that 
// this value should be around 3800 
DEBUGVAR float dbg_fFudgedGuardbandSize = 3800.f;
DEBUGVAR float dbg_fFudgedScreenRatio = 1.05f;

// PrepareForBeamClip, sets up the arrays above so they dont have to be re-calculated in every RenderBeam call
void PrepareForBeamClip()
{
	const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();
	// Near plane normal = camera forward.
    g_axBeamClipNormals[0] = Asura_Vector_3(pxCameraOrientation->e13, pxCameraOrientation->e23, pxCameraOrientation->e33);
	g_axBeamClipNormals[0].Normalise();
	// Near plane position = camera position, but forward enough to match (or just go past) the near plane.
	g_axBeamClipPositions[0] = *Asura_Camera::GetPosition() + g_axBeamClipNormals[0] * (Asura_Camera::GetNearClip() + dbg_fNearPlaneBeamClipAdjust);


	// For the edge planes we will rely on the Frustum setup done in the RenderEnvironmentZone.
	// 'parantly these planes all go thought the camera, which makes finding a position on the plane nice n easy..

	g_axBeamClipPositions[1] = *Asura_Camera::GetPosition();
	g_axBeamClipPositions[2] = *Asura_Camera::GetPosition();
	g_axBeamClipPositions[3] = *Asura_Camera::GetPosition();
	g_axBeamClipPositions[4] = *Asura_Camera::GetPosition();

	
	// And retrieve the guardband normals ... pretty sure this will always be valid.
    Asura_PSP_RenderEnvironment_Zone &xZone = Asura_PSP_RenderEnvironment::GetCurrentZone();

	// Clipping against the frustum seemed ok, but we should really be using the guardband approximation
	// Normals seemed to point upwards, so ill need to negate these.

	//g_axBeamClipNormals[1] = -xZone.m_xFrustum.m_xLeft.m_xNormal;
	//g_axBeamClipNormals[2] = -xZone.m_xFrustum.m_xRight.m_xNormal;
	//g_axBeamClipNormals[3] = -xZone.m_xFrustum.m_xUp.m_xNormal;
	//g_axBeamClipNormals[4] = -xZone.m_xFrustum.m_xDown.m_xNormal;

	g_axBeamClipNormals[1] = -xZone.GetGuardbandLimits().m_xLeft.m_xNormal;
	g_axBeamClipNormals[2] = -xZone.GetGuardbandLimits().m_xRight.m_xNormal;
	g_axBeamClipNormals[3] = -xZone.GetGuardbandLimits().m_xUp.m_xNormal;
	g_axBeamClipNormals[4] = -xZone.GetGuardbandLimits().m_xDown.m_xNormal;
}


bool ClipLineToPlane(Asura_Vector_3& xP1, Asura_Vector_3& xP2, Asura_Vector_3& xPlanePos, Asura_Vector_3& xPlaneNormal)
{
	// First we need to work out which side of the plane each point is on.

	Asura_Vector_3 xP1Side = xP1 - xPlanePos; // Vector from plane to point1
	Asura_Vector_3 xP2Side = xP2 - xPlanePos; // Vector from plane to point2

	float fDotP1 = xP1Side * xPlaneNormal;	  // Dot these with the plane normal
	float fDotP2 = xP2Side * xPlaneNormal;

	bool bP1Ok;
	bool bP2Ok;
	
	if (fDotP1 < 0.f)
	{
		bP1Ok = false;
	}
	else 
	{
		bP1Ok = true;
	}


	if (fDotP2 < 0.f) 
	{
		bP2Ok = false;
	}
	else 
	{
		bP2Ok = true;
	}

	if (!bP1Ok && !bP2Ok) return false; // Both points bad, reject beam totaly.
	if (bP1Ok && bP2Ok) return true; // Both points ok, no clipping needed

	// Need to clip one point so calculate position.

	float fNum = xPlaneNormal * (xPlanePos - xP1);
	float fDem = xPlaneNormal * (xP2-xP1);

	if (fDem == 0.f)
	{
		// Not sure this test should ever hit as the sign checks earlier should prevent it, but just in case reject the beam;
		return false; 
	}

	float fU = fNum/fDem; // Intersection parametric.
	Asura_Vector_3 xP = xP1 + (xP2 - xP1)*fU; // Replace fU in the equation for the line to find the intersection point 

	if (!bP1Ok)
	{
		// Point 1 failed 
		xP1 = xP;
		return true;
	}
	// Point 2 failed
	xP2 = xP;
	return true;
}




void RenderBeam(Asura_Beam* pxBeam)
{
	Asura_Assert(g_uOffset < ASURA_PSP_MAX_BEAMS_PER_FRAME * 8, "Too many beams this frame");
	if(g_uOffset >= ASURA_PSP_MAX_BEAMS_PER_FRAME * 8) return;

#ifdef ASURA_BEAMS_ALL_MODES
	Asura_Render::CurrentStates.RequestTranslucencyMode(pxBeam->GetTranslucencyMode());
#endif

	// calculate the vertex positions for triangle fan 
	Asura_Vector_3 axVertex[8];
	
	// get camera set up
	const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();
	const Asura_Vector_3* pxPos = Asura_Camera::GetPosition();

	// get camera data
	Asura_Vector_3 xCameraSide = Asura_Vector_3(pxCameraOrientation->e11, pxCameraOrientation->e21, pxCameraOrientation->e31);
	Asura_Vector_3 xCameraDown = Asura_Vector_3(pxCameraOrientation->e12, pxCameraOrientation->e22, pxCameraOrientation->e32);
	Asura_Vector_3 xCameraForward = Asura_Vector_3(pxCameraOrientation->e13, pxCameraOrientation->e23, pxCameraOrientation->e33);
	Asura_Vector_3 xCameraPos = Asura_Vector_3(pxPos->x,pxPos->y,pxPos->z);

	// new Beam positions
	Asura_Vector_3 xBeamStart = pxBeam->GetTailPos();//m_xTailPosition;
	Asura_Vector_3 xBeamEnd = pxBeam->CalcCurHeadPos(); 

	// Now we could chose to not clip small beams, its debatable if this would be slower or faster
	// If the beam is within the volume then its going to be slower, but it will be faster if a beam is outside.
	// Esp if the beam is totaly behind the near plane, which I guess on average accounts for half of the beams around!
    //if ((xBeamStart-xBeamEnd).MagnitudeSquared())... { }

	if (!ClipLineToPlane(xBeamStart, xBeamEnd, g_axBeamClipPositions[0], g_axBeamClipNormals[0])) return;
	if (!ClipLineToPlane(xBeamStart, xBeamEnd, g_axBeamClipPositions[1], g_axBeamClipNormals[1])) return;
	if (!ClipLineToPlane(xBeamStart, xBeamEnd, g_axBeamClipPositions[2], g_axBeamClipNormals[2])) return;
	if (!ClipLineToPlane(xBeamStart, xBeamEnd, g_axBeamClipPositions[3], g_axBeamClipNormals[3])) return;
	if (!ClipLineToPlane(xBeamStart, xBeamEnd, g_axBeamClipPositions[4], g_axBeamClipNormals[4])) return;

	// Beam radii
	float fTailRadius, fHeadRadius;
	pxBeam->CalcCurRadii( &fTailRadius, &fHeadRadius );
	
	// calculate new vector directions 
	Asura_Vector_3 xEyeVector = xCameraPos - xBeamStart;
	xEyeVector.Normalise();
	Asura_Vector_3 xPerpVector;

	// Colour is in ABGR format
	Asura_RGBA xTailRGBA, xHeadRGBA;
	pxBeam->CalcCurColours( &xTailRGBA, &xHeadRGBA );
	unsigned int uTailColour = xTailRGBA.ToUIntABGR();
	unsigned int uHeadColour = xHeadRGBA.ToUIntABGR();

	// Get tiled texture info for this beam.
	const Asura_TiledTextureInfo* pxTileInfo = Asura_TiledTextureManagement::GetTileInfoByHash( pxBeam->GetTemplate()->uTileHashID );
	Asura_Assert( pxTileInfo, "Failed to get tile info for beam" ); // Shouldn't fail since we check for it in independent code when trying to create the beam.

	// the effect doesn't work at low angles to sight so we introduce 
	// a tolerance angle in which the head of the beam is brought
	// up to the tail - there is a tolerance to stop things
	// flipping too much with small error
	float fAngleTolerance = 0.0002f;
	float fAngleToLineOfSight= xEyeVector * pxBeam->GetDir();//m_xDirection;
	
	// switch angle sign if neg to reduce later work
	fAngleToLineOfSight = Asura_Maths::Abs(fAngleToLineOfSight);
	
	// work out if in line with camera - if in line draw two 
	// separate triangle fans with the mirrored head and tail
	// else draw a triangle strip with flattend ends facing the camera
	if(fAngleToLineOfSight > 1.0f - fAngleTolerance)
	{		
		if (fAngleToLineOfSight >= 1.0f)
		{
			//alternative to Beam direction needed, use camera side
			xPerpVector = xCameraSide;
		}
		else 
		{
			// cross product ok to use
			xPerpVector = (pxBeam->GetDir()) ^ xEyeVector;
		}

		// draw two triangle fans on each other (tail and head points)

		//cross product ok to use
		xPerpVector = (pxBeam->GetDir()) ^ xEyeVector;
	
		xPerpVector.Normalise();	
		Asura_Vector_3 xUp = xCameraForward ^ xPerpVector;
		xUp.Normalise();
		Asura_Vector_3 xRight = xCameraForward ^ xUp;
		
		// first fan
		axVertex[0] = (xRight	* -fTailRadius) + xBeamStart;
		axVertex[1] = (xUp		*  fTailRadius) + xBeamStart;
		axVertex[2] = (xRight	*  fTailRadius) + xBeamStart;
		axVertex[3] = (xUp		* -fTailRadius) + xBeamStart;

		for (int i=0;i<4;i++)
		{
			g_axVertices[g_uOffset + i].m_afPosition[0] = axVertex[i].x;
			g_axVertices[g_uOffset + i].m_afPosition[1] = axVertex[i].y;
			g_axVertices[g_uOffset + i].m_afPosition[2] = axVertex[i].z;
			
			g_axVertices[g_uOffset + i].m_uColour = uTailColour;
		}

		//skewed texture coords such that you get mirror effect
		g_axVertices[g_uOffset + 0].m_fU = pxTileInfo->m_fu1; g_axVertices[g_uOffset + 0].m_fV = 0.5f * (pxTileInfo->m_fv1 + pxTileInfo->m_fv2);
		g_axVertices[g_uOffset + 1].m_fU = pxTileInfo->m_fu1; g_axVertices[g_uOffset + 1].m_fV = pxTileInfo->m_fv1;
		g_axVertices[g_uOffset + 2].m_fU = 0.5f * (pxTileInfo->m_fu1 + pxTileInfo->m_fu2); g_axVertices[g_uOffset + 2].m_fV = pxTileInfo->m_fv1;
		g_axVertices[g_uOffset + 3].m_fU = pxTileInfo->m_fu1; g_axVertices[g_uOffset + 3].m_fV = pxTileInfo->m_fv1;

		sceKernelDcacheWritebackAll();
		GUM_DrawArray(SCEGU_PRIM_TRIANGLE_FAN, 
						iASURA_PSP_BEAM_VERTEX_FORMAT,
						4, 
						NULL, 
						g_axVertices + g_uOffset);
		g_uOffset += 4;

		//now draw second point in distance
		axVertex[0] = (xRight	* -fHeadRadius) + xBeamEnd;
		axVertex[1] = (xUp		*  fHeadRadius) + xBeamEnd;
		axVertex[2] = (xRight	*  fHeadRadius) + xBeamEnd;
		axVertex[3] = (xUp		* -fHeadRadius) + xBeamEnd;
		
		//copy new data into PSP vertices 
		for (int i=0;i<4;i++)
		{
			g_axVertices[g_uOffset + i].m_afPosition[0] = axVertex[i].x;
			g_axVertices[g_uOffset + i].m_afPosition[1] = axVertex[i].y;
			g_axVertices[g_uOffset + i].m_afPosition[2] = axVertex[i].z;
			
			g_axVertices[g_uOffset + i].m_uColour = uHeadColour;
		}

		//skewed texture coords such that you get mirror effect
		g_axVertices[g_uOffset + 0].m_fU = pxTileInfo->m_fu2; g_axVertices[g_uOffset + 0].m_fV = 0.5f * (pxTileInfo->m_fv1 + pxTileInfo->m_fv2);
		g_axVertices[g_uOffset + 1].m_fU = pxTileInfo->m_fu2; g_axVertices[g_uOffset + 1].m_fV = pxTileInfo->m_fv2;
		g_axVertices[g_uOffset + 2].m_fU = 0.5f * (pxTileInfo->m_fu1 + pxTileInfo->m_fu2); g_axVertices[g_uOffset + 2].m_fV = pxTileInfo->m_fv2;
		g_axVertices[g_uOffset + 3].m_fU = pxTileInfo->m_fu2; g_axVertices[g_uOffset + 3].m_fV = pxTileInfo->m_fv2;

		sceKernelDcacheWritebackAll();
		GUM_DrawArray(SCEGU_PRIM_TRIANGLE_FAN, 
						iASURA_PSP_BEAM_VERTEX_FORMAT,
						4, 
						NULL, 
						g_axVertices + g_uOffset);	
		g_uOffset += 4;

		// keep track of num triangles
	#if defined ASURA_RENDER_TTYPES
		Asura_Render::AddToTrianglesRendered(8,1,
			ASURA_RENDER_TTYPE_OTHER);
	#else
		Asura_Render::AddToTrianglesRendered(8);
	#endif
	}
	else 
	{
		//cross product ok to use
		xPerpVector = (pxBeam->GetDir()) ^ xEyeVector;
	
		xPerpVector.Normalise();	
		Asura_Vector_3 xUp = xCameraForward ^ xPerpVector;
		xUp.Normalise();
		Asura_Vector_3 xRight = xCameraForward ^ xUp;
		
		// vertex 0 at top and nearest
		//		+v0 
		// + v1		+ v2
		// + v3		+ v4
		//		+v5 

		axVertex[0] = (xUp		*  fTailRadius) + xBeamStart;
		axVertex[1] = (xRight	* -fTailRadius) + xBeamStart;
		axVertex[2] = (xRight	*  fTailRadius) + xBeamStart;
		axVertex[3] = (xRight	* -fHeadRadius ) + xBeamEnd;
		axVertex[4] = (xRight	*  fHeadRadius ) + xBeamEnd;
		axVertex[5] = (xUp		* -fHeadRadius ) + xBeamEnd;

		//draw it as a triangle strip
		for (int i=0;i<6;i++)
		{
			g_axVertices[g_uOffset + i].m_afPosition[0] = axVertex[i].x;
			g_axVertices[g_uOffset + i].m_afPosition[1] = axVertex[i].y;
			g_axVertices[g_uOffset + i].m_afPosition[2] = axVertex[i].z;

			g_axVertices[g_uOffset + i].m_uColour = (i < 3) ? uTailColour : uHeadColour;
		}
		g_axVertices[g_uOffset + 0].m_fU = pxTileInfo->m_fu1;
		g_axVertices[g_uOffset + 0].m_fV = pxTileInfo->m_fv1;
		g_axVertices[g_uOffset + 1].m_fU = pxTileInfo->m_fu1;
		g_axVertices[g_uOffset + 1].m_fV = 0.5f * (pxTileInfo->m_fv1 + pxTileInfo->m_fv2);
		g_axVertices[g_uOffset + 2].m_fU = 0.5f * (pxTileInfo->m_fu1 + pxTileInfo->m_fu2);
		g_axVertices[g_uOffset + 2].m_fV = pxTileInfo->m_fv1;
		g_axVertices[g_uOffset + 3].m_fU = 0.5f * (pxTileInfo->m_fu1 + pxTileInfo->m_fu2);
		g_axVertices[g_uOffset + 3].m_fV = pxTileInfo->m_fv2;
		g_axVertices[g_uOffset + 4].m_fU = pxTileInfo->m_fu2;
		g_axVertices[g_uOffset + 4].m_fV = 0.5f * (pxTileInfo->m_fv1 + pxTileInfo->m_fv2);
		g_axVertices[g_uOffset + 5].m_fU = pxTileInfo->m_fu2;
		g_axVertices[g_uOffset + 5].m_fV = pxTileInfo->m_fv2;
	
		sceKernelDcacheWritebackAll();
		GUM_DrawArray(SCEGU_PRIM_TRIANGLE_STRIP, 
						iASURA_PSP_BEAM_VERTEX_FORMAT,
						6, 
						NULL, 
						g_axVertices + g_uOffset);	
		g_uOffset += 6;

		// keep track of num triangles
	#if defined ASURA_RENDER_TTYPES
		Asura_Render::AddToTrianglesRendered(6,1,
			ASURA_RENDER_TTYPE_OTHER);
	#else
		Asura_Render::AddToTrianglesRendered(6);
	#endif
	}
}


void Asura_Beam_System::Platform_Render(void)
{
	Asura_Camera::SetCameraIndex(0);
	Asura_Render::ApplyCameraSettings();
	Asura_PSP_Graphics::SetUseColourDoubling(false);
	
	PrepareForBeamClip();

	for(unsigned j=0; j<s_uNumActiveBeams; j++ )
	{	
		if( !s_apxBeamPtrs[j] ) continue;

#ifdef ASURA_BEAMS_SUPPORT_CAMERA_SELECTION
		if( s_apxBeamPtrs[j]->m_uDontRenderIfCameraIgnoresThisGuid == Asura_Camera::GetEntityToIgnore() ) continue;
		if( (s_apxBeamPtrs[j]->m_uOnlyRenderIfCameraIgnoresThisGuid != ASURA_GUID_UNREGISTERED) && 
			(s_apxBeamPtrs[j]->m_uOnlyRenderIfCameraIgnoresThisGuid != Asura_Camera::GetEntityToIgnore()) ) 
		{
			continue;
		}
#endif //#ifdef ASURA_BEAMS_SUPPORT_CAMERA_SELECTION

		if( s_apxBeamPtrs[j]->IsInvisible() ) continue;
		RenderBeam(s_apxBeamPtrs[j]);
		s_apxBeamPtrs[j]->m_bDontUpdateUntilRenderedOnce = false;
	}
}

