//
// Asura_Wii_MEM2.cpp
//
// Created by KOR - 09 February 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_Wii_MEM2.h"

// Asura headers
#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#endif
#include "Debugging/Asura_Assert.h"


// SDK headers
#include	<revolution/mem.h>

// Static data

bool	Asura_Wii_MEM2::s_bInitialised = false;
u_int	Asura_Wii_MEM2::s_uNumAllocations = 0;
Asura_MEM2_State Asura_Wii_MEM2::m_axPreviousStates[ASURA_MM_STATE_NUM_TYPES];

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
Asura_TemplateHashTree<Asura_MEM2_Node> Asura_Wii_MEM2::s_xAllocationNodeTree;
#endif

static MEMAllocator	s_xMem2Allocator;
static MEMHeapHandle s_xHeapHandle;

const u_int uEXTRA_DEBUG_MEMORY_SIZE = 64 * 1024 * 1024;

// Class implementation
void Asura_Wii_MEM2::Initialise()
{
	Asura_Assert(!s_bInitialised, "Already initialised");
	if(s_bInitialised) { return; }
	void *pvLow = OSGetMEM2ArenaLo();
	void *pvHigh = OSGetMEM2ArenaHi();
	u_int uTotalSize = (u_int)pvHigh - (u_int)pvLow;
	if(uTotalSize > uEXTRA_DEBUG_MEMORY_SIZE)
	{
		// Bring it down to retail size so we don't get any nasty shocks.
		uTotalSize -= uEXTRA_DEBUG_MEMORY_SIZE;
	}
	OSReport("Creating MEM2 heap: %u bytes\n", uTotalSize);
	s_xHeapHandle = MEMCreateExpHeap(pvLow, uTotalSize);
	if ( s_xHeapHandle == MEM_HEAP_INVALID_HANDLE )
	{
		OSHalt("MEM2 heap allocation error.\n");
	}
	else
	{
		OSSetMEM2ArenaLo(pvHigh);
		// Use 32-byte alignment for safety. We should be only dealing with large chunks of data
		// in MEM2 (textures, sounds, etc.), so this overhead should be relatively insignificant.
		MEMInitAllocatorForExpHeap(&s_xMem2Allocator, s_xHeapHandle, 32);
	}

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
	Asura_MemoryManagement_System::SetCheckpointCallback(Checkpoint);
#endif

	s_bInitialised = true;
	s_uNumAllocations = 0;
}

void Asura_Wii_MEM2::Shutdown()
{
	Asura_Assert(s_bInitialised, "Not initialised");
	s_bInitialised = false;
	// TODO - destroy heap here.
	Asura_Assert(!s_uNumAllocations, "Still have allocations on MEM2 shutdown");
}

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
void* Asura_Wii_MEM2::Allocate(size_t uSize, const char* szFile, u_int uLine, bool bOutOfMemoryIsFatal /* = true */)
#else
void* Asura_Wii_MEM2::Allocate(size_t uSize, bool bOutOfMemoryIsFatal /* = true */)
#endif
{
	if(!s_bInitialised)
	{
		Initialise();
	}
	++s_uNumAllocations;
	void* pvData = MEMAllocFromAllocator(&s_xMem2Allocator, uSize);
#ifndef ASURA_SUBMISSION_BUILD
	if(!pvData)
	{
		Asura_Assert(0, "Out of MEM2. You may experience missing sounds, textures, or FMV");
		if(bOutOfMemoryIsFatal)
		{
			const GXColor xFatalBg = { 128, 0, 0, 0 };
			const GXColor xFatalFg = { 255, 255, 255, 0 };
			// Unfortunately we lose our stack when we do this.
			// Might want to investigate a more debuggable version.
			OSFatal( xFatalFg, xFatalBg, "Out of MEM2 memory");
		}
	}
#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
	else
	{
		// Log the data
		Asura_MEM2_Node* pxNode = s_xAllocationNodeTree.AddNode(reinterpret_cast<Asura_Hash_ID>(pvData));
		pxNode->m_szFile = szFile;
		pxNode->m_uLine = uLine;
		pxNode->m_uSize = uSize;
		pxNode->m_uFrame = Asura_Timers::GetFrameNumber();
	}
#endif
#endif
	return pvData;
}

void Asura_Wii_MEM2::Free(void* pvData)
{
	if(!pvData) { return; }
	--s_uNumAllocations;
#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
	if(pvData)
	{
		s_xAllocationNodeTree.DeleteNodeByHash(reinterpret_cast<Asura_Hash_ID>(pvData));
	}
#endif
	MEMFreeToAllocator(&s_xMem2Allocator, pvData);
}

u_int Asura_Wii_MEM2::GetAllocatedBytes()
{
	return MEMGetHeapTotalSize(s_xHeapHandle) - GetFreeBytes();
}

u_int Asura_Wii_MEM2::GetFreeBytes()
{
	return MEMGetTotalFreeSizeForExpHeap(s_xHeapHandle);
}

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED

void Asura_Wii_MEM2::Checkpoint(int iState)
{
	Asura_Assert(0 <= iState && iState < ASURA_MM_STATE_NUM_TYPES, "State out of range");
	if(iState < 0 || iState >= ASURA_MM_STATE_NUM_TYPES) { return; }

	Asura_MEM2_State& xState = m_axPreviousStates[iState];

	const int iNumAllocs = s_uNumAllocations;
	const int iAllocatedBytes = GetAllocatedBytes();

	static int ls_iCounter = 1;

	char szLogfileName[sizeof("MEM2_Checkpoint_LXXX_State_XX-X.txt")];
	if(xState.m_iNumberOfAllocations != iASURA_MEM2_STATE_UNINITIALISED)
	{
		if(xState.m_iNumberOfAllocations != iNumAllocs || xState.m_iTotalMemoryCurrentlyAlloced != iAllocatedBytes)
		{
			// Dump to logfile for diffing.
			Asura_PrintfToBfr(szLogfileName, "MEM2_Checkpoint_Leak_State_%02i-%1i.txt", sizeof(szLogfileName), true, iState, ls_iCounter);
			DumpMallocInfo(szLogfileName);
			Asura_Assertf(0, ("MEM2 environment has changed since the last time in this state. Previous: %i allocs, %i bytes. Current: %i allocs, %i bytes.",
				xState.m_iNumberOfAllocations, xState.m_iTotalMemoryCurrentlyAlloced, iNumAllocs, iAllocatedBytes));
		}
	}

	// Now update the stored state.
	xState.m_iNumberOfAllocations = iNumAllocs;
	xState.m_iTotalMemoryCurrentlyAlloced = iAllocatedBytes;

	// change the counter index
	ls_iCounter ^= 1;
	// Dump to logfile
	Asura_PrintfToBfr(szLogfileName, "MEM2_Checkpoint_Last_State_%02i-%1i.txt", sizeof(szLogfileName), true, iState, ls_iCounter);
	DumpMallocInfo(szLogfileName);
}

static const char* g_szMallocInfoFilename = NULL;
static const char* MALLOC_DUMP_DEFAULT_PATH = "Logfiles\\";
static const char* MALLOC_INFO_DUMP_DEFAULT_FILE = "MEM2MallocInfo.txt";
static const char* MALLOC_SUMMARY_DUMP_DEFAULT_FILE = "MEM2MallocSummary.txt";

void Asura_Wii_MEM2::DumpMallocInfo(const char* szFilenameNoPath /* = NULL */, const char* szPath /* = NULL */)
{
	char szFileName[256];
	Asura_PrintfToBfr(szFileName, "%s%s", 256, true, 
		szPath ? szPath : MALLOC_DUMP_DEFAULT_PATH, 
		szFilenameNoPath ? szFilenameNoPath : MALLOC_INFO_DUMP_DEFAULT_FILE);

	g_szMallocInfoFilename = szFileName;
	Asura_OutputToDebugger(( "Dumping MEM2 malloc info to \"%s\".\n", szFileName ));

	Asura_SimpleLogfileClear( szFileName );

	Asura_SimpleLogfileOutput( (szFileName, "Node address\t    Size\t   Frame\t    Line\tFile\n" ) );

	s_xAllocationNodeTree.Traverse(DumpMallocInfoLine);
}

void Asura_Wii_MEM2::DumpMallocInfoLine(Asura_Hash_ID uHash, const Asura_MEM2_Node* pxNode)
{
	Asura_SimpleLogfileOutput( (g_szMallocInfoFilename, "0x%08x\t%8u\t%8u\t%8u\t%s\n", uHash, pxNode->m_uSize, pxNode->m_uFrame, pxNode->m_uLine, pxNode->m_szFile) );
}

#endif // #ifdef ASURA_MEMORYMANAGEMENT_ENABLED
