//
// Asura_DebugFileStream.cpp
//
// Created by KOR - 09 October 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "FileAccess/Asura_DebugFileStream.h"

// Asura header-files...
#include "Debugging/Asura_Assert.h"

// Wii Header Files
#include "LoadAndSave/Asura_Wii_NAND.h"

#include <stdio.h>

#include <zlib/zlib.h>

// simple helper function for casting void * to file * so we dont need FILE* in global class header
inline FILE* __FP(void *fp) {return (::FILE*)fp;}

// Static Variables
static bool	g_bFileStreamToPC				=	true;

// Debug Variables
DEBUGVAR int	dbg_iMinSpaceNeededToWrite		=	1000000;

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DebugFileStream::Open
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Opens the specified file for either reading or writing only.
//
// Parameters   : The type of open to perform.
//
// Return value : "true" if successful, "false" if failed to open.
////////////////////////////////////////////////////////////////////////////////
bool Asura_DebugFileStream::Open(const OpenType eOpenType)
{
	// Is there already a file open?
	if (m_bIsOpen)
	{
		// Yes - report an error...
		Asura_Assertf	(!m_bIsOpen, ("\"%s\" has already been opened.", m_xFilename.GetString ()));
		return			(false);
	}
	else
	{
		// No - do we want to read the file?
		if (g_bFileStreamToPC)
		{
			switch (m_eOpenType = eOpenType)
			{
				case READ_DATA:
					m_pFile = fopen(m_xFilename.GetString (), "rb");
					break;
				case WRITE_DATA:
					if(m_bCompressed)
					{
						Asura_String xFilename = m_xFilename;
						xFilename += ".gz";
						m_pFile = gzopen(xFilename.GetString (), "wb");
					}
					else
					{
						m_pFile = fopen(m_xFilename.GetString (), "wb");
					}
					break;
				default:
					Asura_Assertf	(false, ("Unexpected \"Asura_FileStream\" open type specified for \"%s\".", m_xFilename.GetString ()));
					break;
			}
		}
		else
		{
			switch (m_eOpenType = eOpenType)
			{
				case READ_DATA:
				{
					// We should never be reading
					Asura_Assert(false, "Error: Should not be reading data");
					return false;
				}
				case WRITE_DATA:
				{
					// We don't actually open the file now, we do it when we read and write.
					// If the file already exists we want to delete it
					if (Asura_Wii_NAND::DoesFileExist(m_xFilename.GetString()))
					{
						Asura_Wii_NAND::DeleteFile(m_xFilename.GetString());
					}
					// We want to check to see if we have a reasonable amount of space left
					if (Asura_Wii_NAND::CheckAvailableSpaceBytes(dbg_iMinSpaceNeededToWrite))
					{
						Asura_Assert(false, "Not enough space to write a new file");
						return false;
					}
					return true;
				}
				default:
				{
					Asura_Assertf	(false, ("Unexpected \"Asura_FileStream\" open type specified for \"%s\".", m_xFilename.GetString ()));
					return false;
				}
			}
		}

		// Set the flags...
		m_bIsOpen	= (NULL != m_pFile);

		// Return whether we were successful...
		return	(m_bIsOpen);
	}
}

/// End of function 'Asura_DebugFileStream::Open'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DebugFileStream::Open
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Opens the specified file for either reading or writing only.
//
// Parameters   : The name of the file to open.
//
// Return value : "true" if successful, "false" if failed to open.
////////////////////////////////////////////////////////////////////////////////

bool Asura_DebugFileStream::Open(const char* szFilename, const OpenType eOpenType, const bool bUseFullPath, const bool bCompressed /* = false */)
{
	// Is there already a file open?
	if (m_bIsOpen)
	{
		// Yes - report an error...
		Asura_Assertf	(!m_bIsOpen, ("Cannot open \"%s\" - \"%s\" is already open.", szFilename, m_xFilename.GetString()));
		return			(false);
	}
	// If we are streaming to NAND then the path name cannot be longer then NAND_MAX_NAME
	if(!g_bFileStreamToPC)
	{
		// We want to remove the directory from the File name (if it exists)
		szFilename = strrchr(szFilename, '\\');
		// We need to move one more along (it currently points to the last '\', we want the string after the '\')
		szFilename++;
		// Now make sure the file name isn't too long
		if(!(strlen(szFilename) <= NAND_MAX_NAME))
		{
			Asura_Assertf	(false,("Filename (%s) is too long", szFilename));
			return			false;
		}
	}
	// No - do we need to use the full path?
	if (bUseFullPath)
	{
		// Store the filename...
		m_xFilename	= szFilename;
	}
	else
	{
		// No - simple copy the filename...
		m_xFilename	= szFilename;
	}
	m_bCompressed = bCompressed;
	// Open this file...
	return	(Open(eOpenType));
}

/// End of function 'Asura_DebugFileStream::Open'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DebugFileStream::ReadData
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Reads the specified amount of data from the stream into the
//				  buffer given.
//
// Parameters   : The number of bytes to be read and a pointer to the address
//				  to read the data into.
//
// Return value : The number of bytes successfully read.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_DebugFileStream::ReadData (const u_int uNumberOfBytes,
									   void* pData)
{
	// Are we able to read?
	if (IsReading())
	{
		// Yes - return the number of bytes read...
		return fread(pData, 1, uNumberOfBytes, __FP(m_pFile));
	}
	else
	{
		// No - cannot read anything...
		return	(0);
	}
}

/// End of function 'Asura_DebugFileStream::ReadData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DebugFileStream::SkipData
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Reads the specified amount of data from the stream into the
//				  buffer given.
//
// Parameters   : A pointer to the address to read the data into and the number
//				  of bytes to be read.
//
// Return value : The number of bytes successfully read.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_DebugFileStream::SkipData (const u_int uNumberOfBytes)
{
	// Are we able to read?
	if (IsReading())
	{
		// Yes - return the number of bytes read...
		return fseek(__FP(m_pFile), uNumberOfBytes, SEEK_CUR);
	}
	else
	{
		// No - cannot read anything...
		return	(0);
	}
}

/// End of function 'Asura_DebugFileStream::SkipData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DebugFileStream::WriteData
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Writes the specified amount of data from the stream into the
//				  buffer given.
//
// Parameters   : The number of bytes to be written and a pointer to the address
//				  to write the data into.
//
// Return value : The number of bytes successfully write.
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR bool dbg_bCheckSpaceBeforeWriting		= true;

u_int Asura_DebugFileStream::WriteData (const u_int uNumberOfBytes,
								   const void* pData)
{
	// Are we able to write?
	if (g_bFileStreamToPC)
	{
		if (IsWriting())
		{
			if(m_bCompressed)
			{
				return gzwrite(m_pFile, pData, uNumberOfBytes);
			}
			else
			{
				// Yes - return the number of bytes read...
				u_int uTotalBytes = 0;
				while (uTotalBytes < uNumberOfBytes)
				{
					u_int uBytesToWrite = uNumberOfBytes;

					if (uBytesToWrite - uTotalBytes > 1024)
						uBytesToWrite = uTotalBytes + 1024;

					u_int uBytes = 0;
					while (uBytes == 0)
					{
						u_int uStartPos = ftell(__FP(m_pFile));
						uBytes = fwrite((const char*)pData + uTotalBytes, 1, uBytesToWrite - uTotalBytes, __FP(m_pFile));
						u_int uEndPos = ftell(__FP(m_pFile));
						if (uEndPos - uStartPos != uBytes)
						{
							fseek(__FP(m_pFile), uStartPos, SEEK_SET);
							uBytes = 0;
						}
						fflush(__FP(m_pFile));
					}
					uTotalBytes += uBytes;

				}
				return uTotalBytes;
			}
		}
		else
		{
			// No - cannot read anything...
			return	(0);
		}
	}
	else
	{
		// First lets check that there is enough room left
		if (dbg_bCheckSpaceBeforeWriting)
		{
			if(Asura_Wii_NAND::CheckAvailableSpaceBytes(uNumberOfBytes))
			{
				Asura_Assert(false, "Not enough space to write");
				return false;
			}
		}
		// As there is lets write the data to the end of the file
		int iSize = 0;
		if(Asura_Wii_NAND::DoesFileExist(m_xFilename.GetString()))
		{
			iSize = Asura_Wii_NAND::GetFileSize(m_xFilename.GetString());
		}
		if (Asura_Wii_NAND::WriteFile(m_xFilename.GetString(),pData, uNumberOfBytes, true, false, iSize))
		{
			return uNumberOfBytes;
		}
	}

	return 0;
}

/// End of function 'Asura_DebugFileStream::WriteData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_DebugFileStream::Close
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Closes the currently open file.
//
// Return value : "true" if closed successfully, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_DebugFileStream::Close (void)
{
	// is the file open?
	if (m_bIsOpen)
	{
		if (g_bFileStreamToPC)
		{
			if(m_bCompressed)
			{
				return gzclose(m_pFile);
			}
			else
			{
				// Yes - return the number of bytes read...
				return fclose(__FP(m_pFile));
			}
		}
		// Else just return true
		return true;
	}
	else
	{
		// No - cannot read anything...
		return	(0);
	}
}

/// End of function 'Asura_DebugFileStream::Close'. ///


/// End of file. ///
