//
// Asura_Wii_Input.cpp
//
// Created by KOR - 09 October 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Input/Asura_Input.h"
#include "Input/Asura_Wii_Input.h"

// Asura headers
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#ifdef ASURA_USE_NEW_INPUT_LEGACY_SUPPORT
#include "Core/Asura_UserInput.h"
#endif
#include "Core/Memory/Asura_Mem.h"
#ifdef ASURA_USE_DEBUG_TEXTURE_VIEWER
#include "Debugging/Asura_Debug_Texture_Viewer.h"
#endif
#include "Debugging/Asura_Logfile.h"
#include "Input/Asura_Controller.h"
#include "Interface/Asura_CommandConsole.h"

#ifdef ASURA_USE_CURSOR
#include "Interface/Asura_Cursor.h"
#endif

#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_GX.h"

#ifdef ASURA_USE_CURSOR
#include "Rendering/Asura_Viewport.h"
#endif

// Platform headers
#include "Core/Asura_Wii_MEM2.h"
#include "Rendering/Asura_Wii_ROMFont.h"

#include "HomeButton/Asura_Wii_HomeButtonMenu.h"

// SDK headers
#include <revolution/kpad.h>
#include <revolution/wpad.h>

// Constants
const u_char uWPAD_TIMEOUT_MINUTES	  = 15;

// Debugvars
DEBUGVAR float dbg_fPos_play_radius  = 0.15f ;	// Set pos sensitivity
DEBUGVAR float dbg_fPos_sensitivity  = 0.75f ;
DEBUGVAR float dbg_fHori_play_radius = 0.00f ;	// Set horizon sensitivity
DEBUGVAR float dbg_fHori_sensitivity = 1.00f ;
DEBUGVAR float dbg_fDist_play_radius = 0.00f ;	// Set dist sensitivity
DEBUGVAR float dbg_fDist_sensitivity = 1.00f ;
DEBUGVAR float dbg_fAcc_play_radius  = 0.10f ;	// Set acc sensitivity
DEBUGVAR float dbg_fAcc_sensitivity  = 1.00f ;

DEBUGVAR float dbg_fRepeat_delay_sec = 0.75f ;	// Set key repeat
DEBUGVAR float dbg_fRepeat_pulse_sec = 0.25f ;

DEBUGVAR float dbg_fAccelerationMultiplier	= 1.f;
DEBUGVAR float dbg_fAccelerationDeadZone	= 0.1f;

// Asura_Input platform implementation
void Asura_Input::Platform_Initialise()
{
	Asura_Wii_Input::Initialise();

	// As good a place as any to set up these mappings
#ifdef ASURA_USE_DEBUG_TEXTURE_VIEWER
	Asura_ControlMap& xControlMap = Asura_Debug_Texture_Viewer::GetControlMap();
	xControlMap.CreateMappings( Asura_Debug_Texture_Viewer::NUM_CONTROL_MAPPINGS );
	xControlMap.AddKeyMapping(Asura_Debug_Texture_Viewer::CONTROL_UP,		ASURA_WII_INPUT_NUNCHUCK_STICK_U, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xControlMap.AddKeyMapping(Asura_Debug_Texture_Viewer::CONTROL_DOWN,		ASURA_WII_INPUT_NUNCHUCK_STICK_D, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xControlMap.AddKeyMapping(Asura_Debug_Texture_Viewer::CONTROL_LEFT,		ASURA_WII_INPUT_NUNCHUCK_STICK_L, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xControlMap.AddKeyMapping(Asura_Debug_Texture_Viewer::CONTROL_RIGHT,	ASURA_WII_INPUT_NUNCHUCK_STICK_R, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xControlMap.AddKeyMapping(Asura_Debug_Texture_Viewer::CONTROL_SELECT,	ASURA_WII_INPUT_A, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xControlMap.AddKeyMapping(Asura_Debug_Texture_Viewer::CONTROL_CHANGETRANSLUCENCY, ASURA_WII_INPUT_B, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
#endif
}

void Asura_Input::Platform_Reset()
{
}

void Asura_Input::Platform_Shutdown()
{
	for(u_int uController = 0; uController < WPAD_MAX_CONTROLLERS; ++uController)
	{
		delete m_apxControllers[uController];
	}

	Asura_Wii_Input::Shutdown();
}

void Asura_Input::Platform_Update()
{
	Asura_Wii_Input::Update();
}

u_int Asura_Input::Platform_GetNumConnectedControllers()
{
	return Asura_Wii_Input::GetNumConnectedControllers();
}

DEBUGVAR float dbg_fCursorClampX = 1.f;
DEBUGVAR float dbg_fCursorClampY = 1.f;

// Function to detect if the controller was connected or disconnected.
// Each time the state of controller will change, this callback will be called.
void Asura_Wii_Controller::ConnectCallback (const s32 iChan, const s32 iReason)
{
	Asura_Wii_Controller* pxWiiController = static_cast<Asura_Wii_Controller*>(Asura_Input::GetController(iChan));

	if( pxWiiController )
	{
		pxWiiController->m_bIsConnected = (iReason != WPAD_ERR_NO_CONTROLLER);
	}
}

// This is potentially wrong - we never check for multiple updates in one frame.  SW
void Asura_Wii_Controller::Platform_Update()
{
	Asura_Mem::Set(m_pfCurrentKeys, 0, sizeof(float) * m_uNumKeys);

	if(!m_bIsConnected) { return; }

#ifdef ASURA_USE_CURSOR
	// RWM - 2007-08-30 - If Cursor disabled, disable wii pointer to save power
	const bool bCursorEnabled = Asura_Cursor::IsCursorMovementEnabled();
	if ( bCursorEnabled != WPADIsDpdEnabled( m_uControllerIndex ) && !Asura_Wii_HomeButtonMenu::IsActive() )
	{
		if ( bCursorEnabled )
		{
			KPADEnableDPD( m_uControllerIndex );
			// KOR - default the cursor position to zero until we have something valid
			m_xDPDPosition.Zero();
		}
		else
		{
			KPADDisableDPD( m_uControllerIndex );
		}
	}
#else
	KPADEnableDPD( m_uControllerIndex );
#endif

	int iKpadReads = KPADRead( m_uControllerIndex, &m_axKpadStatus[0], KPAD_MAX_READ_BUFS ) ;

	if (!Asura_GX::ReadyToDraw()) {	return; }

	if(Asura_Wii_HomeButtonMenu::IsActive()) { return; }

	// Check whether we got anything this frame.
	if(!iKpadReads)
	{
		// If not, assume that the input is the same as the last time.
		Asura_Mem::Copy(m_pfCurrentKeys, m_pfPreviousKeys, sizeof(float) * m_uNumKeys);
		return;
	}

	switch(m_axKpadStatus[0].dev_type)
	{
		case WPAD_DEV_NOT_FOUND:
		case WPAD_DEV_CORE:
		case WPAD_DEV_FREESTYLE:
		case WPAD_DEV_CLASSIC:
		case WPAD_DEV_UNKNOWN:
		{
			m_bUnsupportedExtension = false;
			break;
		}
		default:
		{
			m_bUnsupportedExtension = true;
		}
	}
	
	m_bNunchukAttached = (m_axKpadStatus[0].dev_type == WPAD_DEV_FREESTYLE);

	m_pfCurrentKeys[ASURA_WII_INPUT_A]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_A)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_B]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_B)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
	if(m_bNunchukAttached)
	{
		m_pfCurrentKeys[ASURA_WII_INPUT_C]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_C)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
		m_pfCurrentKeys[ASURA_WII_INPUT_Z]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_Z)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
	}
	m_pfCurrentKeys[ASURA_WII_INPUT_UP]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_UP)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_DOWN]	= (m_axKpadStatus[0].hold & KPAD_BUTTON_DOWN)  ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_LEFT]	= (m_axKpadStatus[0].hold & KPAD_BUTTON_LEFT)  ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_RIGHT]	= (m_axKpadStatus[0].hold & KPAD_BUTTON_RIGHT) ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_PLUS]	= (m_axKpadStatus[0].hold & KPAD_BUTTON_PLUS)  ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_MINUS]	= (m_axKpadStatus[0].hold & KPAD_BUTTON_MINUS) ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_1]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_1)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_2]		= (m_axKpadStatus[0].hold & KPAD_BUTTON_2)	   ? fASURA_CONTROLLER_INPUT_MAX : 0;
	m_pfCurrentKeys[ASURA_WII_INPUT_HOME]	= (m_axKpadStatus[0].hold & KPAD_BUTTON_HOME)  ? fASURA_CONTROLLER_INPUT_MAX : 0;

	#ifdef ASURA_USE_CURSOR
	const bool bOnScreen = (m_axKpadStatus[0].dpd_valid_fg > 0);
	if(bOnScreen)
	{
	const Asura_Vector_2 xHalfViewport(Asura_Viewport::GetWidth() * .5f, Asura_Viewport::GetHeight() * .5f);
		const float fCursorX = Asura_Maths::ClampToRange(m_axKpadStatus[0].pos.x, -dbg_fCursorClampX, dbg_fCursorClampX);
		const float fCursorY = Asura_Maths::ClampToRange(m_axKpadStatus[0].pos.y, -dbg_fCursorClampY, dbg_fCursorClampY);

		m_xDPDPosition.x = ( fCursorX + 1.0f ) * ( xHalfViewport.x );
		m_xDPDPosition.y = ( fCursorY + 1.0f ) * ( xHalfViewport.y );
	}

// 	Asura_Cursor::SetCursorIndex(m_uControllerIndex);
	if(Asura_Cursor::IsCursorMovementEnabled())
	{
		Asura_Cursor::UpdateAbs( m_xDPDPosition.x, m_xDPDPosition.y, m_uControllerIndex );
	}
	else
	{
		Asura_Cursor::UpdateAbs( 0.f, 0.f, m_uControllerIndex );
	}
	Asura_Cursor::SetCursorOffscreen(!bOnScreen, m_uControllerIndex);
	#endif //ASURA_USE_CURSOR

	if (m_axKpadStatus[0].dpd_valid_fg > 0)
	{
	const float fCursorX = Asura_Maths::ClampToRange(m_axKpadStatus[0].pos.x, -1.0f, 1.0f);
	const float fCursorY = Asura_Maths::ClampToRange(m_axKpadStatus[0].pos.y, -1.0f, 1.0f);
		const bool bUpsideDown = (m_axKpadStatus[0].horizon.x < 0.f);
		const float fHorizonY = bUpsideDown ? Asura_Maths::Sign(m_axKpadStatus[0].horizon.y) : m_axKpadStatus[0].horizon.y;
		const float fCursorTilt = Asura_Maths::ArcSine(fHorizonY) * (2.f / Asura_Maths::Pi);
	SetAxisValues(ASURA_WII_INPUT_DPD_R, ASURA_WII_INPUT_DPD_L, fCursorX);
	SetAxisValues(ASURA_WII_INPUT_DPD_U, ASURA_WII_INPUT_DPD_D, fCursorY);
		SetAxisValues(ASURA_WII_INPUT_DPD_TILT_R, ASURA_WII_INPUT_DPD_TILT_L, fCursorTilt);
	}
	else
	{
		SetAxisValues(ASURA_WII_INPUT_DPD_R, ASURA_WII_INPUT_DPD_L, 0.f);
		SetAxisValues(ASURA_WII_INPUT_DPD_U, ASURA_WII_INPUT_DPD_D, 0.f);
		SetAxisValues(ASURA_WII_INPUT_DPD_TILT_R, ASURA_WII_INPUT_DPD_TILT_L, 0.f);
	}

	for(u_int u = 0; u < iKpadReads; ++u)
	{
		u_int uKpadSampleIndex = (iKpadReads-1) - u;

	// Update samples
	if(m_bNunchukAttached)
	{
		SetAxisValues(ASURA_WII_INPUT_NUNCHUCK_STICK_R, ASURA_WII_INPUT_NUNCHUCK_STICK_L, m_axKpadStatus[0].ex_status.fs.stick.x);
		SetAxisValues(ASURA_WII_INPUT_NUNCHUCK_STICK_U, ASURA_WII_INPUT_NUNCHUCK_STICK_D, m_axKpadStatus[0].ex_status.fs.stick.y);

			m_axNunchuckAccSamples[m_uCurrentSample].x = m_axKpadStatus[uKpadSampleIndex].ex_status.fs.acc.x;
			m_axNunchuckAccSamples[m_uCurrentSample].y = m_axKpadStatus[uKpadSampleIndex].ex_status.fs.acc.y;
			m_axNunchuckAccSamples[m_uCurrentSample].z = m_axKpadStatus[uKpadSampleIndex].ex_status.fs.acc.z;
	}

		m_axRemoteAccSamples[m_uCurrentSample].x = m_axKpadStatus[uKpadSampleIndex].acc.x;
		m_axRemoteAccSamples[m_uCurrentSample].y = m_axKpadStatus[uKpadSampleIndex].acc.y;
		m_axRemoteAccSamples[m_uCurrentSample].z = m_axKpadStatus[uKpadSampleIndex].acc.z;
	
		IncrementCurrentSample();
	}

	// Calculate smoothed and deadzoned tilt control.
	Asura_Vector_3 xNunchuckAverageAcc = Asura_Maths::ZeroVector3;
	Asura_Vector_3 xRemoteAverageAcc   = Asura_Maths::ZeroVector3;

	for(u_int u = 0; u < uNUM_ACCELERATION_SAMPLES_TO_AVERAGE_FOR_TILT; ++u)
	{
		xNunchuckAverageAcc += m_axNunchuckAccSamples[u];
		xRemoteAverageAcc   += m_axRemoteAccSamples[u];
	}
	// Remote tilt
	xRemoteAverageAcc *= (1.f / uNUM_ACCELERATION_SAMPLES_TO_AVERAGE_FOR_TILT);

	xRemoteAverageAcc.x = Asura_Maths::Sign(xRemoteAverageAcc.x) * Asura_Maths::Max((Asura_Maths::Abs(xRemoteAverageAcc.x) - dbg_fAccelerationDeadZone), 0.f) / (1.f - dbg_fAccelerationDeadZone);
	xRemoteAverageAcc.y = Asura_Maths::Sign(xRemoteAverageAcc.y) * Asura_Maths::Max((Asura_Maths::Abs(xRemoteAverageAcc.y) - dbg_fAccelerationDeadZone), 0.f) / (1.f - dbg_fAccelerationDeadZone);
	xRemoteAverageAcc.z = Asura_Maths::Sign(xRemoteAverageAcc.z) * Asura_Maths::Max((Asura_Maths::Abs(xRemoteAverageAcc.z) - dbg_fAccelerationDeadZone), 0.f) / (1.f - dbg_fAccelerationDeadZone);

	SetAxisValues(ASURA_WII_INPUT_REMOTE_TILT_L, ASURA_WII_INPUT_REMOTE_TILT_R, Asura_Maths::ClampToRange(-xRemoteAverageAcc.x * dbg_fAccelerationMultiplier, -1.f, 1.f));
	SetAxisValues(ASURA_WII_INPUT_REMOTE_TILT_U, ASURA_WII_INPUT_REMOTE_TILT_D, Asura_Maths::ClampToRange( xRemoteAverageAcc.z * dbg_fAccelerationMultiplier, -1.f, 1.f));

	// Nunchuck tilt
	xNunchuckAverageAcc *= (1.f / uNUM_ACCELERATION_SAMPLES_TO_AVERAGE_FOR_TILT);

	xNunchuckAverageAcc.x = Asura_Maths::Sign(xNunchuckAverageAcc.x) * Asura_Maths::Max((Asura_Maths::Abs(xNunchuckAverageAcc.x) - dbg_fAccelerationDeadZone), 0.f) / (1.f - dbg_fAccelerationDeadZone);
	xNunchuckAverageAcc.y = Asura_Maths::Sign(xNunchuckAverageAcc.y) * Asura_Maths::Max((Asura_Maths::Abs(xNunchuckAverageAcc.y) - dbg_fAccelerationDeadZone), 0.f) / (1.f - dbg_fAccelerationDeadZone);
	xNunchuckAverageAcc.z = Asura_Maths::Sign(xNunchuckAverageAcc.z) * Asura_Maths::Max((Asura_Maths::Abs(xNunchuckAverageAcc.z) - dbg_fAccelerationDeadZone), 0.f) / (1.f - dbg_fAccelerationDeadZone);	

	SetAxisValues(ASURA_WII_INPUT_NUNCHUCK_TILT_L, ASURA_WII_INPUT_NUNCHUCK_TILT_R, Asura_Maths::ClampToRange(-xNunchuckAverageAcc.x * dbg_fAccelerationMultiplier, -1.f, 1.f));
	SetAxisValues(ASURA_WII_INPUT_NUNCHUCK_TILT_U, ASURA_WII_INPUT_NUNCHUCK_TILT_D, Asura_Maths::ClampToRange( xNunchuckAverageAcc.z * dbg_fAccelerationMultiplier, -1.f, 1.f));	
}

void Asura_Wii_Controller::IncrementCurrentSample( void )
		{
	if(++m_uCurrentSample >= uNUM_ACCELERATION_SAMPLES)
		{
		m_uCurrentSample = 0;
		}
		}

bool Asura_Wii_Controller::HasAnyKeyBeenPressed( void ) const
		{
	for(u_int uKey = 0; uKey <= ASURA_WII_INPUT_LAST_BUTTON; ++uKey)
		{
		if(IsKeyPressed(uKey))
			{ 
			return true;
			}
			}
	return false;
		}

bool Asura_Wii_Controller::DebouncedHasAnyKeyBeenPressed( void ) const
		{
	for(u_int uKey = 0; uKey <= ASURA_WII_INPUT_LAST_BUTTON; ++uKey)
			{ 
		if( IsDebouncedKeyPressed(uKey) )
			{
			return true;
			}
		}
	return false;
			}

Asura_Vector_3 Asura_Wii_Controller::GetRemoteAcceleration( const u_int uNumFramesAgo ) const
	{
	return m_axRemoteAccSamples[GetAccelerationArrayIndex(uNumFramesAgo)];
	}

Asura_Vector_3 Asura_Wii_Controller::GetNunchuckAcceleration( const u_int uNumFramesAgo ) const
	{
	return m_axNunchuckAccSamples[GetAccelerationArrayIndex(uNumFramesAgo)];
	}
	
u_int Asura_Wii_Controller::GetAccelerationArrayIndex( const u_int uNumFramesAgo ) const
	{
	u_int uFramesToLookBack;
	// Can't go back further than we have history
	if ( uNumFramesAgo > ( uNUM_ACCELERATION_SAMPLES - 1 ) )
	{
		uFramesToLookBack = ( uNUM_ACCELERATION_SAMPLES - 1 );
		}
		else
		{			
		uFramesToLookBack = uNumFramesAgo;
	}

	int iArrayIndex = m_uCurrentSample - uFramesToLookBack - 1; // Current sample increments in the update, so we need to go back one further

	// Wrap around at zero
	if ( iArrayIndex < 0 )
	{
		iArrayIndex += uNUM_ACCELERATION_SAMPLES;
			}

	Asura_Assert(iArrayIndex < uNUM_ACCELERATION_SAMPLES,	"Wiimote acceleration index out of range");
	Asura_Assert(iArrayIndex >= 0,							"Wiimote acceleration index out of range");

	return static_cast<u_int>(iArrayIndex);
		}

// Asura_Wii_Input implementation

// Static class data
bool Asura_Wii_Input::s_bInitialised = false;
u_int Asura_Wii_Input::s_uMaxAllowedControllers = WPAD_MAX_CONTROLLERS;
DEBUGVAR u_int dbg_uUnusedControllerDisconnectDelayMilliseconds = 5000;

void Asura_Wii_Input::Initialise()
{
	WPADRegisterAllocator(WPAD_Alloc, WPAD_Free);
	WPADSetAutoSleepTime(uWPAD_TIMEOUT_MINUTES);
	KPADInit();
	KPADReset() ;

	s_bInitialised = true;

	// TODO - figure out when exactly to register the controllers
	for(u_int uController = 0; uController < WPAD_MAX_CONTROLLERS; ++uController)
	{
		KPADSetPosParam ( uController, dbg_fPos_play_radius,  dbg_fPos_sensitivity ) ;
		KPADSetHoriParam( uController, dbg_fHori_play_radius, dbg_fHori_sensitivity ) ;
		KPADSetDistParam( uController, dbg_fDist_play_radius, dbg_fDist_sensitivity ) ;
		KPADSetAccParam ( uController, dbg_fAcc_play_radius,  dbg_fAcc_sensitivity ) ;
		KPADSetBtnRepeat( uController, dbg_fRepeat_delay_sec, dbg_fRepeat_pulse_sec ) ;
		KPADEnableAimingMode(uController);

		new Asura_Wii_Controller(uController);
		KPADSetConnectCallback (uController, &Asura_Wii_Controller::ConnectCallback);
	}
}

void Asura_Wii_Input::Shutdown()
{
	s_bInitialised = false;
}

void Asura_Wii_Input::Update()
{
	Asura_Assert(s_bInitialised, "Updating but not initialised");
	if(!Asura_Wii_HomeButtonMenu::IsActive() && Asura_Timers::GetMSecsSinceStart() > dbg_uUnusedControllerDisconnectDelayMilliseconds)
	{
		// Disconnect unused controllers if we're not in the Home Menu.
		for(u_int u = s_uMaxAllowedControllers; u < WPAD_MAX_CONTROLLERS; ++u)
		{
			WPADDisconnect(u);
		}
	}
#ifdef ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR
	while(AreUnsupportedExtensionsAttached())
	{
		ShowUnsupportedExtensionScreen();
		// Update controllers
		for(u_int uController = 0; uController < WPAD_MAX_CONTROLLERS; ++uController)
		{
			Asura_Controller* pxController = Asura_Input::GetController(uController);
			if(pxController)
			{
				pxController->Update();
			}
		}
	}
#endif//def ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR
			}

#ifdef ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR
bool Asura_Wii_Input::AreUnsupportedExtensionsAttached()
{
	for(u_int uController = 0; uController < WPAD_MAX_CONTROLLERS; ++uController)
	{
		const Asura_Wii_Controller* pxController = static_cast<const Asura_Wii_Controller*>(Asura_Input::GetController(uController));
		if(pxController->HasUnsupportedExtension())
		{
			return true;
		}
	}
	return false;
}

DEBUGVAR short dbg_sFontSize = 16;

void Asura_Wii_Input::ShowUnsupportedExtensionScreen()
{
	Asura_Render::BeginFrame();
	Asura_Wii_ROMFont::SetSize(dbg_sFontSize, 0);
	Asura_Wii_ROMFont::RenderString(4 * 8, 4 * dbg_sFontSize, 0, "TODO: LOCALISE: Unsupported extension.");
	Asura_Wii_ROMFont::RenderString(4 * 8, 5 * dbg_sFontSize, 0, "Please detach the extension from the Wii Remote.");
	Asura_Render::EndFrame();
}
#endif//def ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR

void* Asura_Wii_Input::WPAD_Alloc(u32 uSize)
{
	return ASURA_WII_MEM2_ALLOC(uSize);
}

u8 Asura_Wii_Input::WPAD_Free(void *pvAddr)
{
	Asura_Wii_MEM2::Free(pvAddr);
	return 1;
}

u_int Asura_Wii_Input::GetNumConnectedControllers()
{
	for(u_int uController = 0; uController < WPAD_MAX_CONTROLLERS; ++uController)
	{
		const Asura_Wii_Controller* pxController = static_cast<const Asura_Wii_Controller*>(Asura_Input::GetController(uController));
		if(!pxController || !pxController->IsConnected())
		{
			return uController;
		}
	}
	return WPAD_MAX_CONTROLLERS;
}

// Legacy support
#ifdef ASURA_USE_NEW_INPUT_LEGACY_SUPPORT
int Asura_UserInput::Platform_GetNumberOfActiveControllers(bool bConsecutive)
{
	//return Asura_Input::GetNumConnectedControllers();
	u_int uActiveControllers = 0;
	for(u_int u = 0; u < WPAD_MAX_CONTROLLERS; ++u)
	{
		if(Platform_IsControllerActive(u))
		{
			++uActiveControllers;
		}
		else
		{
			// Only count consecutive controllers, otherwise we're going to get confused.
			break;
}
	}
	return uActiveControllers;
}

void Asura_UserInput::Platform_Initialise(void)
{
	Asura_Input::Initialise();
}

void Asura_UserInput::Platform_Shutdown(void)
{
	Asura_Input::Shutdown();
}

void Asura_UserInput::Platform_Reset(void)
{
	Asura_Input::Reset();
}

void Asura_UserInput::Platform_Update(void)
{
	Asura_Input::Update();

	// Fill out CurrentInput
	for(u_int uController = 0; uController < WPAD_MAX_CONTROLLERS; ++uController)
	{
		const u_int uKeysPerController = ASURA_KEYID_PAD2_BUTTON1 - ASURA_KEYID_PAD1_BUTTON1;
		const u_int uStartKey = ASURA_KEYID_PAD1_BUTTON1 + uKeysPerController * uController;
		const u_int uAxesPerController = ASURA_AXIS_PAD2_LSTICK_H - ASURA_AXIS_PAD1_LSTICK_H;
		const u_int uStartAxis = ASURA_AXIS_PAD1_LSTICK_H + uAxesPerController * uController;
		for(u_int uKey = 0; uKey < uASURA_WII_NUM_CONTROLLER_KEYS; ++uKey)
		{
			CurrentInput[uStartKey + uKey] = Asura_Input::GetKeyValue(uController, uKey) ? 255 : 0;
		}

		AxisValue[uStartAxis] = Asura_Input::GetAxisValue(uController, ASURA_WII_INPUT_NUNCHUCK_STICK_R, ASURA_WII_INPUT_NUNCHUCK_STICK_L);
		AxisValue[uStartAxis + 1] = Asura_Input::GetAxisValue(uController, ASURA_WII_INPUT_NUNCHUCK_STICK_U, ASURA_WII_INPUT_NUNCHUCK_STICK_D);

		AxisValue[uStartAxis + 2] = Asura_Input::GetAxisValue(uController, ASURA_WII_INPUT_DPD_R, ASURA_WII_INPUT_DPD_L);
		AxisValue[uStartAxis + 3] = Asura_Input::GetAxisValue(uController, ASURA_WII_INPUT_DPD_U, ASURA_WII_INPUT_DPD_D);
	}
}

bool Asura_UserInput::Platform_IsControllerActive(int iPlayerNum)
{
	const Asura_Wii_Controller* pxWiiController = static_cast<const Asura_Wii_Controller*>(Asura_Input::GetController(iPlayerNum));
	return pxWiiController ? pxWiiController->IsConnected() : false;
}

bool Asura_UserInput::Platform_IsKeyToggled(int iKey)
{
	// KOR - no keyboard input.
	return false;
}


void Asura_UserInput::Platform_UpdateInputBuffer( Asura_UnicodeChar* wszBuffer, u_int uMaxChars, u_int& uInputCursorPos,
												 bool (*pfnHandleInputOverride)(ASURA_KEYID eKey), const Asura_UnicodeChar* wszInputBuffer,
												 u_int uInputBufferSize )
{
	// KOR - no keyboard input.
}
#endif
