/**
*
* \file Asura_Wii_PlatformInput.h
*
* Created by Kim Randell - 30/01/2003 17:39:39 
*
* Copyright (C) 2003 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_WII_PLATFORMINPUT_H__
#define __ASURA_WII_PLATFORMINPUT_H__

//#define ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR (1)

// Asura headers
#include "Core/Asura_Maths.h" // for Asura_Vector_3
#include "Core/Memory/Asura_Mem.h"
#include "Input/Asura_Controller.h"

// SDK headers
#include <revolution/kpad.h>

// Constants
const u_int uNUM_ACCELERATION_SAMPLES = 256;	// Size of the stored acceleration history.
												// 256 is about 1.5 seconds

const u_int uNUM_ACCELERATION_SAMPLES_TO_AVERAGE_FOR_TILT = 12;	// Averaging over a few samples gives a better value for the tilt

enum ASURA_WII_INPUT_ID
{
	// Buttons
	ASURA_WII_INPUT_A = 0,
	ASURA_WII_INPUT_B,
	ASURA_WII_INPUT_C,
	ASURA_WII_INPUT_Z,
	ASURA_WII_INPUT_UP,
	ASURA_WII_INPUT_DOWN,
	ASURA_WII_INPUT_LEFT,
	ASURA_WII_INPUT_RIGHT,
	ASURA_WII_INPUT_PLUS,
	ASURA_WII_INPUT_MINUS,
	ASURA_WII_INPUT_1,
	ASURA_WII_INPUT_2,
	ASURA_WII_INPUT_LAST_BUTTON = ASURA_WII_INPUT_2,
	ASURA_WII_INPUT_HOME,

	// Axes
	ASURA_WII_INPUT_NUNCHUCK_STICK_U,
	ASURA_WII_INPUT_NUNCHUCK_STICK_D,
	ASURA_WII_INPUT_NUNCHUCK_STICK_L,
	ASURA_WII_INPUT_NUNCHUCK_STICK_R,

	ASURA_WII_INPUT_REMOTE_TILT_U,
	ASURA_WII_INPUT_REMOTE_TILT_D,
	ASURA_WII_INPUT_REMOTE_TILT_L,
	ASURA_WII_INPUT_REMOTE_TILT_R,

	ASURA_WII_INPUT_NUNCHUCK_TILT_U,
	ASURA_WII_INPUT_NUNCHUCK_TILT_D,
	ASURA_WII_INPUT_NUNCHUCK_TILT_L,
	ASURA_WII_INPUT_NUNCHUCK_TILT_R,	

	ASURA_WII_INPUT_DPD_U,
	ASURA_WII_INPUT_DPD_D,
	ASURA_WII_INPUT_DPD_L,
	ASURA_WII_INPUT_DPD_R,

	ASURA_WII_INPUT_DPD_TILT_L,
	ASURA_WII_INPUT_DPD_TILT_R,

	ASURA_WII_NUM_INPUTS,
};

// Class definitions
class Asura_Wii_Controller : public Asura_Controller
{
public:
	Asura_Wii_Controller(u_int uControllerIndex)
	: Asura_Controller(uControllerIndex, ASURA_WII_NUM_INPUTS)
	, m_uCurrentSample(0)
	, m_xDPDPosition(Asura_Maths::ZeroVector2)

	, m_uFlags(0)
	{
		Asura_Mem::Set(m_axKpadStatus, 0, sizeof(m_axKpadStatus));
		for(u_int uSample = 0; uSample < uNUM_ACCELERATION_SAMPLES; ++uSample)
		{
			m_axNunchuckAccSamples[uSample].Zero();
			m_axRemoteAccSamples[uSample].Zero();
		}
	}

	virtual void Platform_Update();

	const KPADStatus* GetKpadStatus() const { return m_axKpadStatus; }
	KPADStatus* GetKpadStatus() { return m_axKpadStatus; }

	bool HasUnsupportedExtension() const { return m_bUnsupportedExtension; }
	bool IsConnected() const { return m_bIsConnected; }
	bool IsNunchukAttached() const { return m_bNunchukAttached; }
	const Asura_Vector_2& GetDPDPosition() const { return m_xDPDPosition; }

	virtual bool HasAnyKeyBeenPressed( void ) const;
	virtual bool DebouncedHasAnyKeyBeenPressed( void ) const;

	// Function to detect if the controller was connected or disconnected.
	static void	ConnectCallback (const s32 iChan, const s32 iReason);
	// Expose our records of acceleration over time, to allow proper motion sensing
	Asura_Vector_3 GetRemoteAcceleration(   const u_int uNumFramesAgo = 0 ) const;
	Asura_Vector_3 GetNunchuckAcceleration( const u_int uNumFramesAgo = 0 ) const;

private:

	// Helper functions
	u_int GetAccelerationArrayIndex( const u_int uNumFramesAgo ) const;
	void  IncrementCurrentSample( void );

	Asura_Vector_3	m_axRemoteAccSamples[uNUM_ACCELERATION_SAMPLES];
	Asura_Vector_3	m_axNunchuckAccSamples[uNUM_ACCELERATION_SAMPLES];

	u_int	m_uCurrentSample;

	Asura_Vector_2	m_xDPDPosition;

	union
	{
		u_int m_uFlags;
		struct
		{
			u_int m_bUnsupportedExtension						:1;
			u_int m_bIsConnected								:1;
			u_int m_bNunchukAttached							:1;
		};
	};

	KPADStatus		m_axKpadStatus[ KPAD_MAX_READ_BUFS ];
};

class Asura_Wii_Input
{
public:
	static void Initialise();
	static void Shutdown();
	static void Update();


	static bool IsInitialised() { return s_bInitialised; }

	static u_int GetNumConnectedControllers();

#ifdef ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR
	static bool AreUnsupportedExtensionsAttached();
	static void ShowUnsupportedExtensionScreen();
#endif//def ASURA_WII_INPUT_SHOW_UNSUPPORTED_EXTENSION_ERROR

	static void SetMaxAllowedControllers(u_int uMax) { s_uMaxAllowedControllers = uMax; }
private:
	static void* WPAD_Alloc(u32 uSize);
	static u8	 WPAD_Free(void *pvAddr);

	// Static data
	static bool s_bInitialised;
	static u_int s_uMaxAllowedControllers;
};

// Legacy support
#ifdef ASURA_USE_NEW_INPUT_LEGACY_SUPPORT
#include "Core/Asura_UserInput.h"
enum ASURA_WII_KEYID
{
	ASURA_WII_KEYID_CONTROLLER1_A		= ASURA_KEYID_PAD1_BUTTON1,
	ASURA_WII_KEYID_CONTROLLER1_B		= ASURA_KEYID_PAD1_BUTTON2,
	ASURA_WII_KEYID_CONTROLLER1_C		= ASURA_KEYID_PAD1_BUTTON3,
	ASURA_WII_KEYID_CONTROLLER1_Z		= ASURA_KEYID_PAD1_BUTTON4,
	ASURA_WII_KEYID_CONTROLLER1_UP		= ASURA_KEYID_PAD1_BUTTON5,
	ASURA_WII_KEYID_CONTROLLER1_DOWN	= ASURA_KEYID_PAD1_BUTTON6,
	ASURA_WII_KEYID_CONTROLLER1_LEFT	= ASURA_KEYID_PAD1_BUTTON7,
	ASURA_WII_KEYID_CONTROLLER1_RIGHT	= ASURA_KEYID_PAD1_BUTTON8,
	ASURA_WII_KEYID_CONTROLLER1_PLUS	= ASURA_KEYID_PAD1_BUTTON9,
	ASURA_WII_KEYID_CONTROLLER1_MINUS	= ASURA_KEYID_PAD1_BUTTON10,
	ASURA_WII_KEYID_CONTROLLER1_1		= ASURA_KEYID_PAD1_BUTTON11,
	ASURA_WII_KEYID_CONTROLLER1_2		= ASURA_KEYID_PAD1_BUTTON12,

	ASURA_WII_KEYID_CONTROLLER2_A		= ASURA_KEYID_PAD2_BUTTON1,
	ASURA_WII_KEYID_CONTROLLER2_B		= ASURA_KEYID_PAD2_BUTTON2,
	ASURA_WII_KEYID_CONTROLLER2_C		= ASURA_KEYID_PAD2_BUTTON3,
	ASURA_WII_KEYID_CONTROLLER2_Z		= ASURA_KEYID_PAD2_BUTTON4,
	ASURA_WII_KEYID_CONTROLLER2_UP		= ASURA_KEYID_PAD2_BUTTON5,
	ASURA_WII_KEYID_CONTROLLER2_DOWN	= ASURA_KEYID_PAD2_BUTTON6,
	ASURA_WII_KEYID_CONTROLLER2_LEFT	= ASURA_KEYID_PAD2_BUTTON7,
	ASURA_WII_KEYID_CONTROLLER2_RIGHT	= ASURA_KEYID_PAD2_BUTTON8,
	ASURA_WII_KEYID_CONTROLLER2_PLUS	= ASURA_KEYID_PAD2_BUTTON9,
	ASURA_WII_KEYID_CONTROLLER2_MINUS	= ASURA_KEYID_PAD2_BUTTON10,
	ASURA_WII_KEYID_CONTROLLER2_1		= ASURA_KEYID_PAD2_BUTTON11,
	ASURA_WII_KEYID_CONTROLLER2_2		= ASURA_KEYID_PAD2_BUTTON12,

	ASURA_WII_KEYID_CONTROLLER3_A		= ASURA_KEYID_PAD3_BUTTON1,
	ASURA_WII_KEYID_CONTROLLER3_B		= ASURA_KEYID_PAD3_BUTTON2,
	ASURA_WII_KEYID_CONTROLLER3_C		= ASURA_KEYID_PAD3_BUTTON3,
	ASURA_WII_KEYID_CONTROLLER3_Z		= ASURA_KEYID_PAD3_BUTTON4,
	ASURA_WII_KEYID_CONTROLLER3_UP		= ASURA_KEYID_PAD3_BUTTON5,
	ASURA_WII_KEYID_CONTROLLER3_DOWN	= ASURA_KEYID_PAD3_BUTTON6,
	ASURA_WII_KEYID_CONTROLLER3_LEFT	= ASURA_KEYID_PAD3_BUTTON7,
	ASURA_WII_KEYID_CONTROLLER3_RIGHT	= ASURA_KEYID_PAD3_BUTTON8,
	ASURA_WII_KEYID_CONTROLLER3_PLUS	= ASURA_KEYID_PAD3_BUTTON9,
	ASURA_WII_KEYID_CONTROLLER3_MINUS	= ASURA_KEYID_PAD3_BUTTON10,
	ASURA_WII_KEYID_CONTROLLER3_1		= ASURA_KEYID_PAD3_BUTTON11,
	ASURA_WII_KEYID_CONTROLLER3_2		= ASURA_KEYID_PAD3_BUTTON12,

	ASURA_WII_KEYID_CONTROLLER4_A		= ASURA_KEYID_PAD4_BUTTON1,
	ASURA_WII_KEYID_CONTROLLER4_B		= ASURA_KEYID_PAD4_BUTTON2,
	ASURA_WII_KEYID_CONTROLLER4_C		= ASURA_KEYID_PAD4_BUTTON3,
	ASURA_WII_KEYID_CONTROLLER4_Z		= ASURA_KEYID_PAD4_BUTTON4,
	ASURA_WII_KEYID_CONTROLLER4_UP		= ASURA_KEYID_PAD4_BUTTON5,
	ASURA_WII_KEYID_CONTROLLER4_DOWN	= ASURA_KEYID_PAD4_BUTTON6,
	ASURA_WII_KEYID_CONTROLLER4_LEFT	= ASURA_KEYID_PAD4_BUTTON7,
	ASURA_WII_KEYID_CONTROLLER4_RIGHT	= ASURA_KEYID_PAD4_BUTTON8,
	ASURA_WII_KEYID_CONTROLLER4_PLUS	= ASURA_KEYID_PAD4_BUTTON9,
	ASURA_WII_KEYID_CONTROLLER4_MINUS	= ASURA_KEYID_PAD4_BUTTON10,
	ASURA_WII_KEYID_CONTROLLER4_1		= ASURA_KEYID_PAD4_BUTTON11,
	ASURA_WII_KEYID_CONTROLLER4_2		= ASURA_KEYID_PAD4_BUTTON12,
};

enum ASURA_WII_AXISID
{
	ASURA_WII_AXISID_CONTROLLER1_NUNCHUK_H	= ASURA_AXIS_PAD1_LSTICK_H,
	ASURA_WII_AXISID_CONTROLLER1_NUNCHUK_V	= ASURA_AXIS_PAD1_LSTICK_V,

	ASURA_WII_AXISID_CONTROLLER2_NUNCHUK_H	= ASURA_AXIS_PAD2_LSTICK_H,
	ASURA_WII_AXISID_CONTROLLER2_NUNCHUK_V	= ASURA_AXIS_PAD2_LSTICK_V,

	ASURA_WII_AXISID_CONTROLLER3_NUNCHUK_H	= ASURA_AXIS_PAD3_LSTICK_H,
	ASURA_WII_AXISID_CONTROLLER3_NUNCHUK_V	= ASURA_AXIS_PAD3_LSTICK_V,

	ASURA_WII_AXISID_CONTROLLER4_NUNCHUK_H	= ASURA_AXIS_PAD4_LSTICK_H,
	ASURA_WII_AXISID_CONTROLLER4_NUNCHUK_V	= ASURA_AXIS_PAD4_LSTICK_V,
};

const u_int uASURA_WII_NUM_CONTROLLER_KEYS = 12;
const u_int uASURA_WII_KEYOFFSET_NEXT_PAD = ASURA_WII_KEYID_CONTROLLER2_A - ASURA_WII_KEYID_CONTROLLER1_A;
const u_int uASURA_WII_AXISOFFSET_NEXT_PAD = ASURA_WII_AXISID_CONTROLLER2_NUNCHUK_H - ASURA_WII_AXISID_CONTROLLER1_NUNCHUK_H;

#endif

#endif//__ASURA_WII_PLATFORMINPUT_H__
