//
// Asura_GX.h
//
// Created by KOR - 09 October 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_GX_H__
#define __ASURA_GX_H__

// Asura headers
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_RenderStates.h"
#ifdef ASURA_USE_PROFILING_SYSTEM
#include "Debugging/Asura_Profiling_System.h"
#endif
#include "Rendering/Asura_Colour.h"

// Library headers
#include <revolution/vi/vifuncs.h>
#include <revolution/gx/gxstruct.h>

// Constants
const u_int uASURA_GX_MAX_ENTITY_LIGHTS = 3;
// Note that changing the near clip value will have a large impact on the accuracy of the Z-buffer,
// and consequently any special effects based on it (such as graph fog or line distance fades)
const float fASURA_GX_NEAR_CLIP = .1f; 
const float fASURA_GX_FAR_CLIP = 3000.f;


// Forward declarations
struct Asura_TextureAnimationOutput;

class Asura_GX
{	
public:
	// tokens are used instead of GXDrawDone
	enum GPU_CALLBACK_TOKEN {
		GPU_CALLBACK_STARTFRAME = 1,
		GPU_CALLBACK_ENDFRAME,
		GPU_CALLBACK_SYNC,
		GPU_CALLBACK_SMOOTHSKIN_SYNC,
#ifdef ASURA_USE_PROFILING_SYSTEM
		GPU_CALLBACK_PROFILE_START_FIRST,
		GPU_CALLBACK_PROFILE_START_LAST = GPU_CALLBACK_PROFILE_START_FIRST + ASURA_PROFILING_SYSTEM_MAX_TIMERS,
		GPU_CALLBACK_PROFILE_STOP_FIRST,
		GPU_CALLBACK_PROFILE_STOP_LAST = GPU_CALLBACK_PROFILE_STOP_FIRST + ASURA_PROFILING_SYSTEM_MAX_TIMERS,
#endif
	};
	static void AddDrawSyncCallback(GPU_CALLBACK_TOKEN uToken);


	typedef void (*VertexFormatSetupCB)(void);

	static VertexFormatSetupCB s_pfnVertexFormatSetupCB;

	static void InitVertexFormats();
	static void Initialise(void);
	static void Shutdown(void);
	static void RestoreAll(void);
	static bool HasBeenInitialised(void);
	static void ResetStates(void);

	static void SetViewingArea(void);
	static void SetVideoMode(int width, int height, int depth, bool fsaa = false);
	static void Set3DProjection(bool s_bThreeDee, bool forceupdate = false);

	static void ClearZBuffer(void);
	static void SetZBufferMode(bool Enabled, enum ASURA_ZBUFFERCMP_ID CompareMode, bool WriteEnabled);

	static void BeginFrame(void);
	static void EndFrame(void);
	static void FlushFrame(void);

	static void EndFrameIfNecessary(void);
	static void BeginFrameIfNecessary(void);

	// KOR - 13/05/2003 12:31:46 - Nintendo's MTX library uses 3*4 matrices for orientation + position
	static inline void OriPosToMTX(Asura_Matrix_3x3 xOri, Asura_Vector_3 xPos, Mtx& xMtx)
	{
		MTXRowCol(xMtx, 0, 0) = xOri.e11;
		MTXRowCol(xMtx, 0, 1) = xOri.e12;
		MTXRowCol(xMtx, 0, 2) = xOri.e13;
		MTXRowCol(xMtx, 0, 3) = xPos.x;
		MTXRowCol(xMtx, 1, 0) = xOri.e21;
		MTXRowCol(xMtx, 1, 1) = xOri.e22;
		MTXRowCol(xMtx, 1, 2) = xOri.e23;
		MTXRowCol(xMtx, 1, 3) = xPos.y;
		MTXRowCol(xMtx, 2, 0) = xOri.e31;
		MTXRowCol(xMtx, 2, 1) = xOri.e32;
		MTXRowCol(xMtx, 2, 2) = xOri.e33;
		MTXRowCol(xMtx, 2, 3) = xPos.z;
	}

	static inline u_int ARGBToRGBA(u_int uARGB)
	{
		return (uARGB >> 24) | (uARGB << 8);
	}

	static void SetUpEntityLights(const Asura_Vector_3& xPos, bool bVertexAlpha = false);
	static void SetUpEntitySpecular(const Asura_Vector_3& xPos, int iMaterialID);

	static unsigned int GetBufferWidth() { return s_xRenderModeObj.fbWidth; }
	static unsigned int GetBufferSize() { return VIPadFrameBufferWidth(s_xRenderModeObj.fbWidth) * s_xRenderModeObj.xfbHeight * (unsigned int)VI_DISPLAY_PIX_SZ; }
	static void CopyBuffer();
	static void SwapBuffers();
	static void SetupMaterial(int iMaterialID, u_int uPrevTevStages = 0, u_int uPrevTexGens = 0, const int iTexIDOverride = ASURA_TEXTURE_NONE, float fAlpha = 1.f);
	static void SetupDepthOnly();

	// mutex to control access to rendering. Maybe needed in non-assert builds for other purposes?
	static void LockMutex();
	static void UnlockMutex();

	static void RetraceCallback(u32 retraceCount);
	static void RetraceCallback2(u32 retraceCount);

	static Asura_RGBA GetClearColour();
	static void SetClearColour(Asura_RGBA);

	static void SetResetButtonPressed() { s_bResetHasBeenPressed = true; }
	static void SetPowerButtonPressed() { s_bPowerHasBeenPressed = true; }

	static bool ShouldRenderWireframe() { return s_bWireFrame; }

	static const GXRenderModeObj& GetRenderModeObject() { return s_xRenderModeObj; }
	static const Mtx& GetViewMatrix() { return s_xViewMatrix; }
	static const Mtx44& GetProjMatrix() { return s_xProjMatrix; }

	static void LoadRequiredTextures();
	static int GetWhiteTextureIndex() { return s_iWhiteTextureIndex; }
	static u_int GetScreenHeight();
	static u_int GetScreenWidth();
	static u_int GetScreenStride();
	static void WaitForRenderDone();
	static void *GetFrameBuffer(u_int uSwap = 0);

	static bool ReadyToDraw();

	static void ReturnToMenu();
	static void RestartSystem();
	static void CheckForRestart();

	static void ARGBToGXColor(u_int uARGB, GXColor& xColour);
	static void SetDisableTranslucency(bool bDisable) { s_bDisableTranslucency = bDisable; }
#ifdef ASURA_ASSERT
	// Render states can be invalidated by an assert.
	static void SetRenderStatesValid(bool bValid) { s_bRenderStatesValid = bValid; }
	static bool GetRenderStatesValid() { return s_bRenderStatesValid; }
#endif

#ifdef _DEBUG
	static void ResetGX();
#endif
private:
	static Asura_RGBA s_rgbaClearcol;
	static u_int DiffFifoPtr(const void*, const void*);
	static Mtx s_xViewMatrix;
	static Mtx44 s_xProjMatrix;
	static bool s_bWireFrame;
	static bool s_bResetHasBeenPressed;
	static bool s_bPowerHasBeenPressed;
	static bool s_bFatalError;
	static bool s_bDisableTranslucency;
#ifdef ASURA_ASSERT
	static bool s_bRenderStatesValid;
#endif
	static GXRenderModeObj s_xRenderModeObj;
	static void* s_pvFrameBuffer[2];
	static bool s_bFirstFrame;
	static bool Initialised;
	static void* s_pvDefaultFifo;
	static bool s_bThreeD;
	static float s_fCurrentZOffset;
	static int s_iWhiteTextureIndex;
};


#endif//__ASURA_GX_H__