
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_MemoryManagement.cpp
// Author : Kevin Floyer-Lea
// Date   : 21-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Core/Asura_MemoryManagement.h"

// Asura header-files...
#include "Core/Asura_Multithread.h"
#include "Core/Asura_Timers.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

static Asura_Mutex	g_xMemoryManagementMutex;


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

// (AAM - 09.03.2006) Need Platform_Initialise() and Platform_Shutdown() if ASURA_MEMORY_MANAGEMENT_ENABLED
// is defined, *or* if ASURA_MEMORYMANAGEMENT_PLATFORM_INIT_WHEN_DISABLED is non-zero
#if (defined ASURA_MEMORYMANAGEMENT_ENABLED) || ASURA_MEMORYMANAGEMENT_PLATFORM_INIT_WHEN_DISABLED

void Asura_MemoryManagement_System::Platform_Initialise(void)
{
}

void Asura_MemoryManagement_System::Platform_Shutdown(void)
{
}

#endif	// ASURA_MEMORYMANAGEMENT_PLATFORM_INIT_WHEN_DISABLED

void Asura_MemoryManagement_System::Platform_Lock(void)
{
	g_xMemoryManagementMutex.Lock();
}

void Asura_MemoryManagement_System::Platform_Unlock(void)
{
	g_xMemoryManagementMutex.Unlock();
}

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
u_int Asura_MemoryManagement_System::Platform_GetAddressSpaceStart(void) { return 1070000000; } //base address of our process
u_int Asura_MemoryManagement_System::Platform_GetAddressSpaceEnd(void) { return 1200000000;} //1gb (first half of virtual memory available to our process)
#endif

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED

#ifdef ASURA_MEMORYMANAGEMENT_FILL

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_MemoryManagement_System::Platform_FillMem().
// PARAMETERS   pAddr          - Address, or actual address if aligned.
//              uSize          - Size, or actual size if aligned.
//              eEventID       - Type of memory management event.
//              uAlignment     - Alignment type, or 0 if no alignment.
//              pReportedAddr  - Reported address if memory is aligned.
//              uReportedSize  - Reported size if memory is aligned.
// RETURNS      None.
// DESCRIPTION  Fills specified region of memory. If the caller requested
//              memory aligned on a 64-byte boundary and a multiple of 64
//              bytes in size then we ensure the returned memory is not
//              present in the cache.
// AUTHOR       Morgan Parry - 29.07.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_MemoryManagement_System::Platform_FillMem( void* pAddr,
                                                      u_int uSize,
                                                      ASURA_MM_EVENT_ID eEventID,
                                                      u_int uAlignment,
                                                      void* pReportedAddr /*= NULL*/,
                                                      u_int uReportedSize /*= 0*/ )
{
	if( uAlignment )
    {
        Asura_Assert( pReportedAddr != NULL, "" );
        Asura_Assert( uReportedSize > 0, "" );
    }

    u_char ucFill;
	switch( eEventID )
    {
        case ASURA_NEW:
        case ASURA_NEWALIGN:
        case ASURA_NEWARRAY:
        case ASURA_ALLOC:
            ucFill = ucFillValueAllocate;
            break;

        case ASURA_DELETE:
        case ASURA_DELETEALIGN:
        case ASURA_DELETEARRAY:
        case ASURA_DEALLOC:
            ucFill = ucFillValueFree;
            break;

        default:
            Asura_Assert( false, "Invalid memory manager event ID." );
            return;
    }

	// Fill.
    u_char* pucOut = reinterpret_cast<u_char*>( pAddr );

    {
        while( uSize-- ) { *pucOut++ = ucFill; }
    }
}
// End function Asura_MemoryManagement_System::Platform_FillMem().

#endif // #ifdef ASURA_MEMORYMANAGEMENT_FILL

void Asura_MemoryManagement_System::Platform_PostDeallocate( void )
{
}

#else // #ifdef ASURA_MEMORYMANAGEMENT_ENABLED

void* Asura_PC_AllocateAlign(unsigned int bytes, enum ASURA_ALIGN_ID alignment)
{
    int			iTotalBytes = bytes + alignment + sizeof(void*);    // sizeof(void*) bytes before reported address is ptr to start of block
    char* const	pcMem		= (char*)Asura_Allocate(iTotalBytes);
    
    if (!pcMem) 
    {
        return NULL;
    }

    u_long			ulOffset				= (alignment - ((u_long)(pcMem + sizeof(void*))) & (alignment-1));
	char* const		pcReportedMem			= (char*)((unsigned char*)(pcMem + sizeof(void*)) + ulOffset);
    u_long*	const	pcActualAddressStore	= (u_long*)(pcReportedMem - sizeof(void*));

    *pcActualAddressStore = (u_long)pcMem;
    //printf("Asura_PS2_AllocateAlign: %d bytes, %d align, addr = 0x%8p\n", bytes, alignment, pcReportedMem);
    return pcReportedMem;
}

void Asura_PC_DeallocateAlign(void* address)
{
    if (!address) return;

    u_long*	pulActualAddress = ((u_long*)address) - 1;
    u_long	ulActualAddress	= *pulActualAddress;
    //printf("Asura_PS2_DeallocateAlign: addr = 0x%8p, actual = 0x%8x\n", address, uActualAddress);
    Asura_Deallocate((void*)ulActualAddress);
}

#endif

#if defined ASURA_MEMORYMANAGEMENT_ENABLED

void Asura_MemoryUtility::Platform_Initialise()
{
}

void Asura_MemoryUtility::Platform_Shutdown()
{
}

void* Asura_MemoryUtility::Platform_PreAllocate( size_t size )
{
	return NULL;
}

void* Asura_MemoryUtility::Platform_PostAllocate( void* pAddr )
{
	return pAddr;
}

bool Asura_MemoryUtility::Platform_PreDeallocate( void* pAddr )
{
	return false;
}

void Asura_MemoryUtility::Platform_PostDeallocate( void* pAddr )
{
}

#endif // ASURA_MEMORYMANAGEMENT_ENABLED

#if (defined ASURA_MEMORYMANAGEMENT_ENABLED) || ASURA_MEMORYMANAGEMENT_PLATFORM_INIT_WHEN_DISABLED

#if !ASURA_MEMORYMANAGEMENT_INIT_OBJ_IN_PLATFORM_CODE
	#error This code assumes this define is turned on in this file's header.
#endif

#pragma warning( disable: 4073 )
#pragma init_seg( lib ) // Ensure this is constructed before any of the other globals in our application.
static Asura_MemoryManagement_System g_xMemManagementInitialiserObj;

#endif (defined ASURA_MEMORYMANAGEMENT_ENABLED) || ASURA_MEMORYMANAGEMENT_PLATFORM_INIT_WHEN_DISABLED



#if defined(ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS)

struct XMemAllocation
{
	LPVOID pAddress;
	SIZE_T dwSize;
	u_int uFrameNumber;

	XMemAllocation()
	{
		Reset();
	}

	void Reset()
	{
		pAddress = NULL;
		dwSize = 0;
		uFrameNumber = 0;
	}
};

static u_int			g_uNumXMemAllocations	= 0;
static DWORD			g_dwXMemAllocated		= 0;
static XMemAllocation	g_axXMemAllocations[ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS];
static Asura_Mutex		g_xXMemMutex;

/* ------------------------------------------------------------------------- */

LPVOID XMemAlloc( SIZE_T dwSize, DWORD dwAllocAttributes )
{
	LPVOID		lpAddress		= XMemAllocDefault( dwSize, dwAllocAttributes );
	const DWORD	dwAllocatorID	= (dwAllocAttributes >> 16) & 0xFF;

	if((dwAllocatorID >= eXALLOCAllocatorId_GameMin) && (dwAllocatorID <= eXALLOCAllocatorId_GameMax))
	{
		if(g_uNumXMemAllocations < ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS)
		{
			g_xXMemMutex.Lock();
			{
				for ( int i = 0; i < ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS; ++i )
				{
					if ( !g_axXMemAllocations[i].pAddress )
					{
						++g_uNumXMemAllocations;
						g_dwXMemAllocated	+= dwSize;

						g_axXMemAllocations[i].pAddress = lpAddress;
						g_axXMemAllocations[i].dwSize = dwSize;
						g_axXMemAllocations[i].uFrameNumber = Asura_Timers::GetFrameNumber();
						break;
					}
				}
			}
			g_xXMemMutex.Unlock();
		}
		else
		{
			Asura_Assert( false, "run out of room!" );
		}
	}

	return lpAddress;
}

/* ------------------------------------------------------------------------- */

VOID XMemFree( PVOID pAddress, DWORD dwAllocAttributes )
{
	const DWORD	dwAllocatorID	= (dwAllocAttributes >> 16) & 0xFF;

	if((dwAllocatorID >= eXALLOCAllocatorId_GameMin) && (dwAllocatorID <= eXALLOCAllocatorId_GameMax))
	{
		if(pAddress != NULL)
		{
			g_xXMemMutex.Lock();
			{
				for ( int i = 0; i < ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS; ++i )
				{
					if ( g_axXMemAllocations[i].pAddress == pAddress )
					{
						--g_uNumXMemAllocations;
						g_dwXMemAllocated	-= g_axXMemAllocations[i].dwSize;

						g_axXMemAllocations[i].Reset();
						break;
					}
				}
			}
			g_xXMemMutex.Unlock();
		}
	}

	XMemFreeDefault( pAddress, dwAllocAttributes );
}

/* ------------------------------------------------------------------------- */

SIZE_T XMemSize( PVOID pAddress, DWORD dwAllocAttributes )
{
	DWORD size = XMemSizeDefault( pAddress, dwAllocAttributes );

	const DWORD	dwAllocatorID	= (dwAllocAttributes >> 16) & 0xFF;

	if((dwAllocatorID >= eXALLOCAllocatorId_GameMin) && (dwAllocatorID <= eXALLOCAllocatorId_GameMax))
	{
		g_xXMemMutex.Lock();
		{
			for ( int i = 0; i < ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS; ++i )
			{
				if ( g_axXMemAllocations[i].pAddress == pAddress )
				{
					Asura_Assert( size == g_axXMemAllocations[i].dwSize, "Size wrong!" );
				}
			}
		}
		g_xXMemMutex.Unlock();
	}

	return size;
}

static DWORD CalculatePhysicalAlocationAttributes(const Asura_Xbox360_MemoryManagement::MemoryType eMemoryType, const u_int uXAllocAlignment)
{
	switch(eMemoryType)
	{
		case Asura_Xbox360_MemoryManagement::MT_READ_ONLY:
		{
			return(MAKE_XALLOC_ATTRIBUTES(0,							// Object type
										  FALSE,						// Heap tracks attributes
										  TRUE,							// Must succeed
										  FALSE,						// Fixed size
										  0,							// Allocator ID
										  uXAllocAlignment,				// Alignment
										  XALLOC_MEMPROTECT_READWRITE,	// Memory protection
										  FALSE,						// Zero-initialise
										  XALLOC_MEMTYPE_PHYSICAL));	// Memory type
		}

		case Asura_Xbox360_MemoryManagement::MT_READ_WRITE:
		{
			return(MAKE_XALLOC_ATTRIBUTES(0,							// Object type
										  FALSE,						// Heap tracks attributes
										  TRUE,							// Must succeed
										  FALSE,						// Fixed size
										  0,							// Allocator ID
										  uXAllocAlignment,				// Alignment
										  XALLOC_MEMPROTECT_READWRITE,	// Memory protection
										  FALSE,						// Zero-initialise
										  XALLOC_MEMTYPE_PHYSICAL));	// Memory type
		}

		case Asura_Xbox360_MemoryManagement::MT_READ_WRITE_COMBINE:
		{
			return(MAKE_XALLOC_ATTRIBUTES(0,								// Object type
										  FALSE,							// Heap tracks attributes
										  TRUE,								// Must succeed
										  FALSE,							// Fixed size
										  0,								// Allocator ID
										  uXAllocAlignment,					// Alignment
										  XALLOC_MEMPROTECT_WRITECOMBINE,	// Memory protection
										  FALSE,							// Zero-initialise
										  XALLOC_MEMTYPE_PHYSICAL));		// Memory type
		}

		default:
		{
			Asura_Assert(false, "Unexpected memory-type specified for physical-allocation.");
			return(MAKE_XALLOC_ATTRIBUTES(0,							// Object type
										  FALSE,						// Heap tracks attributes
										  TRUE,							// Must succeed
										  FALSE,						// Fixed size
										  0,							// Allocator ID
										  uXAllocAlignment,				// Alignment
										  XALLOC_MEMPROTECT_READWRITE,	// Memory protection
										  FALSE,						// Zero-initialise
										  XALLOC_MEMTYPE_PHYSICAL));	// Memory type
		}
	}		// End of switch.
}

void* Asura_Xbox360_MemoryManagement::AllocatePhysical(const MemoryType eMemoryType, const u_int uNumBytes, const u_int uXAllocAlignment)
{
	const DWORD	dwAttributes	= CalculatePhysicalAlocationAttributes(eMemoryType, uXAllocAlignment);

	return(XMemAlloc(uNumBytes, dwAttributes));
}

void Asura_Xbox360_MemoryManagement::FreePhysical(const MemoryType eMemoryType, void* pData, const u_int uXAllocAlignment)
{
	if(pData != NULL)
	{
		const DWORD	dwAttributes	= CalculatePhysicalAlocationAttributes(eMemoryType, uXAllocAlignment);

		XMemFree(pData, dwAttributes);
	}
}

#else

static u_int ConvertXAllocAlignmentToNumBytes(const u_int uXAllocAlignment)
{
	switch(uXAllocAlignment)
	{
		case XALLOC_PHYSICAL_ALIGNMENT_DEFAULT:
		{
			return(0);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_4:
		{
			return(4);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_8:
		{
			return(8);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_16:
		{
			return(16);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_32:
		{
			return(32);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_64:
		{
			return(64);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_128:
		{
			return(128);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_256:
		{
			return(256);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_512:
		{
			return(512);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_1K:
		{
			return(1024);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_2K:
		{
			return(2048);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_4K:
		{
			return(4096);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_8K:
		{
			return(8192);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_16K:
		{
			return(16384);
		}

		case XALLOC_PHYSICAL_ALIGNMENT_32K:
		{
			return(32768);
		}

		default:
		{
			Asura_Assert(false, "Unexpected alignment type specified for physical-allocation - assuming default.");
			return(0);
		}
	}	// End of switch.
}

void* Asura_Xbox360_MemoryManagement::AllocatePhysical(const MemoryType eMemoryType, const u_int uNumBytes, const u_int uXAllocAlignment)
{
	const u_int	uAlignmentNumBytes	= ConvertXAllocAlignmentToNumBytes(uXAllocAlignment);

	switch(eMemoryType)
	{
		case MT_READ_ONLY:
		{
			return(XPhysicalAlloc(uNumBytes, MAXULONG_PTR, uAlignmentNumBytes, PAGE_READONLY));
		}

		case MT_READ_WRITE:
		{
			return(XPhysicalAlloc(uNumBytes, MAXULONG_PTR, uAlignmentNumBytes, PAGE_READWRITE));
		}

		case MT_READ_WRITE_COMBINE:
		{
			return(XPhysicalAlloc(uNumBytes, MAXULONG_PTR, uAlignmentNumBytes, PAGE_READWRITE | PAGE_WRITECOMBINE));
		}

		default:
		{
			Asura_Assert(false, "Unexpected memory-type specified for physical-allocation.");
			return(NULL);
		}
	}		// End of switch.
}

void Asura_Xbox360_MemoryManagement::FreePhysical(const MemoryType eMemoryType, void* pData, const u_int uXAllocAlignment)
{
	if(pData != NULL)
	{
		XPhysicalFree(pData);
	}
}

#endif	//defined(ASURA_XBOX360_DEBUG_MAX_XMEM_ALLOCATIONS)


/// End of file. ///
