
//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_VideoToTexture.cpp
// Author    : Jason Turner
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM

//
// Asura independent includes
//
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FMV/Asura_VideoToTexture.h"
#include "Rendering/Asura_Render.h"
#include "Sound/Asura_XAudio2.h"

//
// Asura Xbox360 specific
//
#include "Core/Memory/Asura_Mem.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"

//
// Shader Include Files
//
#include "YUV_to_RGB.h"

//
// 360 includes
//
#include <Xmedia2.h>
#include <xaudio2.h>

//
// Local defines
//

// Enables YUV->RGB colour space conversion in the fragment shader
#define ENABLE_SHADER_CSC
//#define ENABLE_DEBUG_LOG_RENDER_TIME

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData
//
////////////////////////////////////////////////////////////////////////////////

class Asura_VideoToTexture_XboxData
{
public:
	friend class Asura_VideoToTexture;

	Asura_VideoToTexture_XboxData();
	~Asura_VideoToTexture_XboxData();

	char				m_szFilename[MAX_PATH];

private:
	struct QuadVertex
	{
		float m_fX, m_fY;
		float m_fU, m_fV;
	};

	bool InitialiseRenderTarget(const int iTextureRTIndex);
	bool Render();
	void Reset();
	void SwapInPlaceholderTexture();
	void SwapOutPlaceholderTexture();

	static const u_int				m_uNumVerts = 4;
	volatile bool					m_bPlayerInited;
	IXMedia2XmvPlayer*				m_pxXMVPlayer;

	Direct3D_BaseTexture*			m_pPlaceholderTexture;
	bool							m_bRenderingPlaceholderTexture;
	Direct3D_Surface*				m_pEnvRenderTarget;
	Direct3D_Texture*				m_pEnvTexture;
	Direct3D_Texture*				m_pYVideoTexture;
	Direct3D_Texture*				m_pUVideoTexture;
	Direct3D_Texture*				m_pVVideoTexture;

	Asura_Xbox360_Effect*			m_pxEffect;
	Asura_Xbox360_Effect_System::Handle m_xYSamplerHandle;
	Asura_Xbox360_Effect_System::Handle m_xUSamplerHandle;
	Asura_Xbox360_Effect_System::Handle m_xVSamplerHandle;

	Asura_Xbox360_VertexBuffer*		m_pQuadVertexBuffer;
	Direct3D_VertexDeclaration*		m_pxVertexDeclaration;

	DWORD							m_uVideoWidth;
	DWORD							m_uVideoHeight;
	DWORD							m_ulMovieDurationMsec;

	Asura_VideoToTexture*			m_pxParentObject;		// pointer to parent Asura_VideoToTexture object
};

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData::Asura_VideoToTexture_XboxData()
//
////////////////////////////////////////////////////////////////////////////////

Asura_VideoToTexture_XboxData::Asura_VideoToTexture_XboxData()
{
	m_bPlayerInited = false;
	m_pxXMVPlayer = NULL;

	m_pPlaceholderTexture = NULL;
	m_pEnvRenderTarget = NULL;
	m_pEnvTexture = NULL;
	m_pYVideoTexture = NULL;
	m_pUVideoTexture = NULL;
	m_pVVideoTexture = NULL;
	m_pQuadVertexBuffer = NULL;
	m_pxVertexDeclaration = NULL;
	m_pxEffect = NULL;

	m_uVideoWidth = 0;
	m_uVideoHeight = 0;
	m_ulMovieDurationMsec = 0;

	m_pxParentObject = NULL;
	m_bRenderingPlaceholderTexture = true;
};

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData::~Asura_VideoToTexture_XboxData()
//
////////////////////////////////////////////////////////////////////////////////

Asura_VideoToTexture_XboxData::~Asura_VideoToTexture_XboxData()
{
	Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData::~Asura_VideoToTexture_XboxData()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_XboxData::Reset()
{
	if (m_bPlayerInited)
	{
		m_bPlayerInited = false;
	}

	if (!m_bRenderingPlaceholderTexture)
	{
		SwapInPlaceholderTexture();
	}

	if (m_pEnvRenderTarget)
	{
		m_pEnvRenderTarget->Release();
		m_pEnvRenderTarget = NULL;
	}

	if (m_pEnvTexture)
	{
		m_pEnvTexture->Release();
		m_pEnvTexture = NULL;
	}

	if (m_pYVideoTexture)
	{
		m_pYVideoTexture->Release();
		m_pYVideoTexture = NULL;
	}

	if (m_pUVideoTexture)
	{
		m_pUVideoTexture->Release();
		m_pUVideoTexture = NULL;
	}

	if (m_pVVideoTexture)
	{
		m_pVVideoTexture->Release();
		m_pVVideoTexture = NULL;
	}

	bool bNowOwned = Asura_Xbox360_Graphics::AcquireThreadOwnership();
	if (m_pQuadVertexBuffer)
	{
		// unset our vertex buffer in the D3D device
		Asura_Xbox360_Graphics::GetDevice()->SetStreamSource( 0, 0, 0, 0 );
		m_pQuadVertexBuffer->Destroy();
		delete m_pQuadVertexBuffer;
		m_pQuadVertexBuffer = NULL;
	}

	if (m_pxVertexDeclaration)
	{
		m_pxVertexDeclaration->Release();
		m_pxVertexDeclaration = NULL;
	}

	if (m_pxEffect)
	{
		Asura_Xbox360_Effect_System::Destroy(m_pxEffect);
		m_pxEffect = NULL;
	}

	if (m_pxXMVPlayer)
	{
		m_pxXMVPlayer->Stop(XMEDIA_STOP_IMMEDIATE);
		m_pxXMVPlayer->Release();
		m_pxXMVPlayer = NULL;
	}

	if (bNowOwned)
	{
		Asura_Xbox360_Graphics::ReleaseThreadOwnership();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Initialise()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Initialise()
{
	// create and initialise 360 data
	Asura_VideoToTexture_XboxData* pxXboxData = new Asura_VideoToTexture_XboxData;
	Asura_Assert(pxXboxData, "");

	m_pPlatformData = pxXboxData;
	pxXboxData->m_pxParentObject = this;

	pxXboxData->m_pxEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auYUV_to_RGB);
	Asura_Assert(pxXboxData->m_pxEffect, "Unable to create FX shader YUV_to_RGB");
	pxXboxData->m_xYSamplerHandle	= pxXboxData->m_pxEffect->GetParameterHandle("g_xYMap");
	pxXboxData->m_xUSamplerHandle	= pxXboxData->m_pxEffect->GetParameterHandle("g_xUMap");
	pxXboxData->m_xVSamplerHandle	= pxXboxData->m_pxEffect->GetParameterHandle("g_xVMap");

	pxXboxData->m_bPlayerInited = true;
	m_ePlayerState = PLAYER_STATE_INITIALISED;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Shutdown()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture::Platform_Shutdown()
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);
	delete pxXboxData;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Open()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Open(const char* szFilename, const int iTextureRTIndex, u_int uSoundTrackIndex)
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);
	HRESULT iRes;

	Asura_String xFileName(szFilename, MAX_PATH);
	m_xFileName = "FMV\\" + xFileName + ".wmv";

	Asura_PrintfToBfr(pxXboxData->m_szFilename, "game:\\FMV\\%s.wmv", MAX_PATH, 0, szFilename);

	// If we're using the D3D device outside of BeginFrame/EndFrame, we need to acquire the thread first
	bool bNowOwned = Asura_Xbox360_Graphics::AcquireThreadOwnership();

	Direct3D_Device* pxDevice = Asura_Xbox360_Graphics::GetDevice();
	Asura_Assert(pxDevice,"Failed to get Direct3D device for XMV playback");

	XMEDIA_XMV_CREATE_PARAMETERS xParams;
	ZeroMemory(&xParams, sizeof (xParams));

	xParams.dwFlags = 0;
	xParams.createType = XMEDIA_CREATE_FROM_FILE;
	xParams.createFromFile.szFileName = pxXboxData->m_szFilename;

	// disable audio stream if we're not playing sound
	xParams.dwAudioStreamId = (m_eFlags & ASURA_FMV_FLAG_NO_SOUND) ? XMEDIA_STREAM_ID_DONT_USE : uSoundTrackIndex + 1;

	if (m_eFlags & ASURA_FMV_FLAG_LOOP)
	{
		xParams.dwFlags |= XMEDIA_CREATE_FOR_LOOP;
	}

	MEMORYSTATUS xStatus;
	GlobalMemoryStatus( &xStatus );
	const u_int uStartMemFree = xStatus.dwAvailPhys;

	IXMedia2XmvPlayer*	pxXMVPlayer;
	IXAudio2* pxAudioEngine = Asura_XAudio2::GetEngine();
	iRes = XMedia2CreateXmvPlayer(pxDevice, pxAudioEngine, &xParams, &pxXMVPlayer);	
	if(SUCCEEDED(iRes))
	{	
		XMEDIA_VIDEO_DESCRIPTOR xDescriptor;
		iRes = pxXMVPlayer->GetVideoDescriptor(&xDescriptor);
		if(SUCCEEDED(iRes))
		{
			pxXboxData->m_uVideoWidth = xDescriptor.dwWidth;
			pxXboxData->m_uVideoHeight = xDescriptor.dwHeight;
			pxXboxData->m_ulMovieDurationMsec = xDescriptor.dwClipDuration;

			m_xVideoInfo.m_uWidth = xDescriptor.dwWidth;
			m_xVideoInfo.m_uHeight = xDescriptor.dwHeight;
			m_xVideoInfo.m_fDuration = static_cast<float>(xDescriptor.dwClipDuration) / 1000.f;

			pxXboxData->m_pxXMVPlayer = pxXMVPlayer;
			bool bError = pxXboxData->InitialiseRenderTarget(iTextureRTIndex);

			if (!bError)
			{
				return false;
			}

			if (!(m_eFlags & ASURA_FMV_FLAG_NO_SOUND) && (m_xSpeakerPropertyList.Size() > 0))
			{
				IXAudio2SourceVoice *pxVoice;
				iRes = pxXMVPlayer->GetSourceVoice(&pxVoice);
				Asura_Assert(pxVoice, "Can't get XMV source voice");
				m_iSourceVoiceHandle = Asura_XAudio2::CreateWrapperVoice(pxVoice);
				Asura_Assert(m_iSourceVoiceHandle != ASURA_INVALID_HANDLE, "Can't create XMV source voice");
			}

			m_ePlayerState = PLAYER_STATE_FILE_OPENED;
		}
	}
	else
	{
		Asura_Assertf(false, ("Unable to open FMV \"%s\"\n", pxXboxData->m_szFilename));
		m_ePlayerState = PLAYER_STATE_ERROR;
		return false;
	}

	if (bNowOwned)
	{
		Asura_Xbox360_Graphics::ReleaseThreadOwnership();
	}

	//
	// This is not the best way to calculate memory used by the FMV player - we really need a global facility to override XMemAlloc().
	// XMedia2CreateXmvPlayer seems to allocate all the memory up front.
	//
	GlobalMemoryStatus( &xStatus );
	const u_int uEndMemFree = xStatus.dwAvailPhys;
	s_uTotalMemUsed = uStartMemFree - uEndMemFree;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Update()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture::Platform_Update()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Play()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Play()
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);

	if (pxXboxData->m_pxXMVPlayer)
	{
		XMEDIA_PLAYBACK_STATUS xStatus;
		HRESULT uRes = pxXboxData->m_pxXMVPlayer->GetStatus(&xStatus);

		if ((uRes == S_OK) && (xStatus.Status == XMEDIA_PLAYER_PAUSED))
		{
			pxXboxData->m_pxXMVPlayer->Resume();
		}
	}

	m_ePlayerState = PLAYER_STATE_PLAYING;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_Pause()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_Pause()
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);
	bool bPaused = false;

	if (pxXboxData->m_pxXMVPlayer)
	{
		XMEDIA_PLAYBACK_STATUS xStatus;
		HRESULT uRes = pxXboxData->m_pxXMVPlayer->GetStatus(&xStatus);

		if ((uRes == S_OK) && (xStatus.Status == XMEDIA_PLAYER_PLAYING))
		{
			pxXboxData->m_pxXMVPlayer->Pause();
			m_ePlayerState = PLAYER_STATE_PAUSED;
			bPaused = true;
		}
	}

	return bPaused;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_SetPlaySpeed()
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture::Platform_SetPlaySpeed(float fSpeed)
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);

	IXMedia2XmvPlayer* pxXMVPlayer = pxXboxData->m_pxXMVPlayer;
	if (pxXMVPlayer)
	{
		if ((GetPlayerState() == PLAYER_STATE_PLAYING) || (GetPlayerState() == PLAYER_STATE_PAUSED))
		{
			pxXMVPlayer->SetPlaySpeed(fSpeed);
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData::InitialiseRenderTarget
//
// iTextureRTIndex = texture index of render target
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture_XboxData::InitialiseRenderTarget(const int iTextureRTIndex)
{
	HRESULT iResult;

	Asura_Assert (iTextureRTIndex != ASURA_TEXTURE_INVALID, "Render target texture invalid");

	if (iTextureRTIndex == ASURA_TEXTURE_INVALID)
	{
		return false;
	}

	Direct3D_Device* pxDevice = Asura_Xbox360_Graphics::GetDevice();
	Asura_Assert(pxDevice, "Can't get D3D device");

	//
	// Create 3 textures for Y,U,V video frame, output from video player.
	//
	iResult = pxDevice->CreateTexture(
		m_uVideoWidth,
		m_uVideoHeight,
		1,
		0,
		D3DFMT_LIN_L8,
		D3DPOOL_DEFAULT,
		&m_pYVideoTexture,
		NULL );

	iResult = pxDevice->CreateTexture(
		m_uVideoWidth/2,		// U and V planes are half width and height
		m_uVideoHeight/2,
		1,
		0,
		D3DFMT_LIN_L8,
		D3DPOOL_DEFAULT,
		&m_pUVideoTexture,
		NULL );

	iResult = pxDevice->CreateTexture(
		m_uVideoWidth/2,		// U and V planes are half width and height
		m_uVideoHeight/2,
		1,
		0,
		D3DFMT_LIN_L8,
		D3DPOOL_DEFAULT,
		&m_pVVideoTexture,
		NULL );

	if (!m_pYVideoTexture || !m_pUVideoTexture || !m_pVVideoTexture)
	{
		Asura_Assert(false, "Asura_VideoToTexture: Failed to create render targets");
		return false;
	}

	// if we're not doing full screen rendering...
	if (iTextureRTIndex != ASURA_TEXTURE_NONE)
	{
		iResult = pxDevice->CreateRenderTarget(
			m_uVideoWidth,
			m_uVideoHeight,
			D3DFMT_X8R8G8B8,
			D3DMULTISAMPLE_NONE,
			0,
			FALSE,
			&m_pEnvRenderTarget,
			NULL
			);

		//
		// Create temporary texture (the size of the video frame) for Resolving each frame into.
		//
		iResult = pxDevice->CreateTexture(
			m_uVideoWidth,
			m_uVideoHeight,
			1,
			0,
			D3DFMT_X8R8G8B8,
			D3DPOOL_DEFAULT,
			&m_pEnvTexture,
			NULL );

		if (!m_pEnvRenderTarget || !m_pEnvTexture)
		{
			Asura_Assert(false, "Asura_VideoToTexture: Failed to create render targets");
			return false;
		}

		Asura_Xbox360_Texture* pxAsuraTex = Asura_Xbox360_TextureManagement::GetTexture(iTextureRTIndex);

		// save current texture interface
		m_pPlaceholderTexture = pxAsuraTex->GetD3DTexture();
	}

	// Setup vertex buffer
	{
		Direct3D_VertexElement axVertexElements[] =
		{
			{ 0,  0, D3DDECLTYPE_FLOAT2,   D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
			{ 0,  8, D3DDECLTYPE_FLOAT2,   D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
			D3DDECL_END()
		};

		iResult = pxDevice->CreateVertexDeclaration(axVertexElements, &m_pxVertexDeclaration);
		Asura_Assert(m_pxVertexDeclaration, "Failed to create overlay quad vertex declaration.");

		m_pQuadVertexBuffer = new Asura_Xbox360_VertexBuffer();
		m_pQuadVertexBuffer->CreateStatic(sizeof(QuadVertex), m_uNumVerts);
		QuadVertex *pxVerts = static_cast<QuadVertex*>(m_pQuadVertexBuffer->Lock());		

		pxVerts[0].m_fX = -1.0f;
		pxVerts[0].m_fY = 1.0f;
		pxVerts[0].m_fU = 0.0f;
		pxVerts[0].m_fV = 0.0f;

		pxVerts[1].m_fX = 1.0f;
		pxVerts[1].m_fY = 1.0f;
		pxVerts[1].m_fU = 1.0f;
		pxVerts[1].m_fV = 0.0f;

		pxVerts[2].m_fX = 1.0f;
		pxVerts[2].m_fY = -1.0f;
		pxVerts[2].m_fU = 1.0f;
		pxVerts[2].m_fV = 1.0f;

		pxVerts[3].m_fX = -1.0f;
		pxVerts[3].m_fY = -1.0f;
		pxVerts[3].m_fU = 0.0f;
		pxVerts[3].m_fV = 1.0f;

		m_pQuadVertexBuffer->Unlock();
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData::Render
//
// Grabs the current video frame from the player and renders to our render target
// Called between BeginFrame/EndFrame
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture::Platform_Render()
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);

	if ((m_ePlayerState == PLAYER_STATE_PLAYING)
		|| (m_ePlayerState == PLAYER_STATE_PAUSED))
	{
		bool bError = pxXboxData->Render();
		if (!bError)
		{
			m_ePlayerState = PLAYER_STATE_COMPLETE;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture_XboxData::Render
//
// Grabs the current video frame from the player and renders to our render target
// bool = false.... shutdown now
////////////////////////////////////////////////////////////////////////////////

bool Asura_VideoToTexture_XboxData::Render()
{
	HRESULT iResult = S_OK;
	D3DLOCKED_RECT xLockRectY;
	D3DLOCKED_RECT xLockRectU;
	D3DLOCKED_RECT xLockRectV;
	bool bRenderFullScreen = (m_pxParentObject->GetTextureRTIndex() == ASURA_TEXTURE_NONE) ? true : false;

	SCOPEDPIXEVENT((0xffeebbcc, "VideoToTexture Render"));

#ifdef ENABLE_DEBUG_LOG_RENDER_TIME
	u_int uRenderStartTime = Asura_Timers::GetMSecsSinceStart_Continuous();
#endif

	// always clear the viewport, regardless of whether video frames are available
	if (bRenderFullScreen)
	{
		Asura_Xbox360_Graphics::GetDevice()->Clear( 0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0 );
	}

	m_pYVideoTexture->LockRect( 0, &xLockRectY, NULL, 0 );
	m_pUVideoTexture->LockRect( 0, &xLockRectU, NULL, 0 );
	m_pVVideoTexture->LockRect( 0, &xLockRectV, NULL, 0 );

	DWORD uLuminanceSize   = m_uVideoHeight * xLockRectY.Pitch;
	DWORD uChrominanceSize = m_uVideoHeight / 2 * xLockRectU.Pitch;

	XMEDIA_VIDEO_FRAME xVideoFrame;
	ZeroMemory( &xVideoFrame, sizeof( XMEDIA_VIDEO_FRAME ) );
	xVideoFrame.videoFormat = XMEDIA_VIDEO_FORMAT_I420;
	
	xVideoFrame.dwFlags = 0;

	// XMEDIA_PLAY_DISABLE_AV_SYNC stops GetNextFrame() from blocking internally (while it waits for the audio/video sync).
	// We may have to do the A/V sync ourselves if the video is wildly out of sync.
	xVideoFrame.dwFlags |= XMEDIA_PLAY_DISABLE_AV_SYNC;
	xVideoFrame.lTimeToPresent = NULL;
	xVideoFrame.i420.pvYBuffer = xLockRectY.pBits;
	xVideoFrame.i420.pvUBuffer = xLockRectU.pBits;
	xVideoFrame.i420.pvVBuffer = xLockRectV.pBits;
	xVideoFrame.i420.dwYBufferSize = uLuminanceSize;
	xVideoFrame.i420.dwUBufferSize = uChrominanceSize;
	xVideoFrame.i420.dwVBufferSize = uChrominanceSize;
	xVideoFrame.i420.dwYPitch = xLockRectY.Pitch;
	xVideoFrame.i420.dwUPitch = xLockRectU.Pitch;
	xVideoFrame.i420.dwVPitch = xLockRectV.Pitch;

	// Note: if you don't call GetNextFrame():
	//		* it will effectively pause playback on the current frame,
	//		* it will return wrong results from GetStatus() - ie. call it when we're paused
	iResult = m_pxXMVPlayer->GetNextFrame( &xVideoFrame );

	bool bBailOut = false;

	// if we're not looping and we're at the end of the video... then stop
	if (!(m_pxParentObject->GetPlayFlags() & Asura_VideoToTexture::ASURA_FMV_FLAG_LOOP))
	{
		if (iResult == XMEDIA_W_EOF)
		{
			m_pYVideoTexture->UnlockRect( 0 );
			m_pUVideoTexture->UnlockRect( 0 );
			m_pVVideoTexture->UnlockRect( 0 );
			return false;
		}
	}

	// no frame available so don't render anything
	if (iResult == XMEDIA_W_NO_DATA)
	{
		bBailOut = true;
	}
	else if( FAILED( iResult ))
	{
		Asura_OutputToDebuggerAssert(false, ("GetNextFrame() failed: 0x%X\n", iResult));
		bBailOut = true;
	}

	// Unlock *before* potentially bailing out below
	m_pYVideoTexture->UnlockRect( 0 );
	m_pUVideoTexture->UnlockRect( 0 );
	m_pVVideoTexture->UnlockRect( 0 );

	if (bBailOut)
	{
		return true;
	}

	// Make sure we're rendering the texture containing the video frames
	if (m_bRenderingPlaceholderTexture)
	{
		SwapOutPlaceholderTexture();
	}

	Direct3D_Device* pxDevice = Asura_Xbox360_Graphics::GetDevice();
	Asura_Assert(pxDevice, "Unable to get D3D device");

	Direct3D_Surface* pxOldRT = NULL;
	Direct3D_Surface* pxThisRT;

	pxDevice->SetVertexDeclaration(m_pxVertexDeclaration);	

	// Save current render target
	pxDevice->GetRenderTarget( 0, &pxOldRT );

	// render full screen
	if (bRenderFullScreen)
	{
		pxThisRT = Asura_Xbox360_Graphics::GetFinalRT();
	}
	else
	{
		pxThisRT = m_pEnvRenderTarget;
	}

	pxDevice->SetRenderTarget( 0, pxThisRT );

	m_pxEffect->SetTexture(m_xYSamplerHandle, m_pYVideoTexture);
	m_pxEffect->SetTexture(m_xUSamplerHandle, m_pUVideoTexture);
	m_pxEffect->SetTexture(m_xVSamplerHandle, m_pVVideoTexture);

	m_pxEffect->Begin();
	m_pxEffect->Commit();
	
	// Set the stream source
	m_pQuadVertexBuffer->Bind(0);

	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
	Asura_Render::CurrentStates.RequestAlphaTest(false);
	Asura_Render::CurrentStates.RequestTextureOpMode( ASURA_TEXTUREOP_OPAQUE );
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestZBufferEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	pxDevice->DrawPrimitive( D3DPT_QUADLIST, 0, 1 );

	m_pxEffect->End();

	// Render the subtitles over the frame
	if (m_pxParentObject->GetSubtitleCallback())
	{
		m_pxParentObject->GetSubtitleCallback()(m_pxParentObject->GetPlayedTimeInSeconds());
	}

	if (!bRenderFullScreen)
	{
		// Resolve back into placeholder texture
		Asura_Xbox360_Graphics::Resolve(m_pEnvTexture);
	}

	// Put old render target back again
	pxDevice->SetRenderTarget( 0, pxOldRT );
	SAFE_RELEASE(pxOldRT);

	// unset our textures in the D3D device
	pxDevice->SetTexture(0, 0);
	pxDevice->SetTexture(1, 0);
	pxDevice->SetTexture(2, 0);

#ifdef ENABLE_DEBUG_LOG_RENDER_TIME
	u_int uRenderTime = Asura_Timers::GetMSecsSinceStart_Continuous() - uRenderStartTime;
	Asura_OutputToDebugger_Retail(("FMV Render time: %dms\n", uRenderTime));
#endif

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_VideoToTexture::Platform_GetPlayedTimeInSeconds
//
////////////////////////////////////////////////////////////////////////////////

float Asura_VideoToTexture::Platform_GetPlayedTimeInSeconds()
{
	Asura_VideoToTexture_XboxData *pxXboxData = reinterpret_cast<Asura_VideoToTexture_XboxData*>(m_pPlatformData);

	XMEDIA_PLAYBACK_STATUS xStatus;
	if (pxXboxData && pxXboxData->m_pxXMVPlayer && pxXboxData->m_pxXMVPlayer->GetStatus(&xStatus) == S_OK )
	{
		return( static_cast<float> (xStatus.lDuration / 1000));
	} 
	else 	
	{
		return 0.0f;
	}

}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_XboxData::SwapInPlaceholderTexture()
{
	if (m_pxParentObject)
	{
		//
		// Put the old texture data back
		//
		int iTex = m_pxParentObject->GetTextureRTIndex();
		if ((iTex != ASURA_TEXTURE_NONE) && m_pPlaceholderTexture)
		{
			Asura_Xbox360_Texture* pxAsuraTex = Asura_Xbox360_TextureManagement::GetTexture(iTex);
			pxAsuraTex->SetD3DTexture(m_pPlaceholderTexture);
			m_bRenderingPlaceholderTexture = true;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_VideoToTexture_XboxData::SwapOutPlaceholderTexture()
{
	int iTex = m_pxParentObject->GetTextureRTIndex();
	if (iTex != ASURA_TEXTURE_NONE)
	{
		Asura_Xbox360_Texture* pxAsuraTex = Asura_Xbox360_TextureManagement::GetTexture(iTex);

		Asura_Assert(m_pEnvTexture, "Video texture NULL!");
		// replace placeholder texture data
		pxAsuraTex->SetD3DTexture(m_pEnvTexture);

		m_bRenderingPlaceholderTexture = false;
	}
}

#endif // ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM
