
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PlatformInput.cpp
// Author : Kevin Floyer-Lea
// Date   : 23-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Input/Asura_PlatformInput.h"

// Asura platform-independent
#include "Core/Asura_FixedArray.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Misc
#include "Core/Asura_MemoryManagement_DebugOff.h"
#include <xtl.h>
#include "Core/Asura_MemoryManagement_DebugOn.h"

// Deadzone for the gamepad inputs, taken from MS sample code
const short sINPUT_DEADZONE = (short)( 0.24f * float(0x7FFF) );

// Current keyboard state (USB debug keyboard)
Asura_FixedArray<bool, ASURA_KEYID_END_OF_KEYBOARD>		g_abKeyState;
Asura_FixedArray<bool, ASURA_KEYID_END_OF_KEYBOARD>		g_abKeyPressedThisFrame;
Asura_FixedArray<int, 256>								g_aiVKToAsuraKey;
const int			g_iMaxInputBufferSize = 64;
int					g_iInputBufferSize=0;
Asura_UnicodeChar	g_wszInputBuffer[g_iMaxInputBufferSize];

class Asura_Xbox360_Controller
{
	public:

    // Device properties
    XINPUT_CAPABILITIES m_xCaps;

    // Thumb stick values converted to range [-1,+1]
    float		m_fLeftX;
    float		m_fLeftY;

    float		m_fRightX;
    float		m_fRightY;

	// Records which buttons were pressed this frame - only set on
    // the frame that the button is first pressed.
    u_short		m_usPressedButtons;
	
	// The user index associated with this gamepad
    u_int		m_usUserIndex;

	// trigger info, similar to the button state data above
    u_char		m_ucLeftTrigger;
    u_char		m_ucRightTrigger;

	// Flag indicating if pad is currently connected
    bool		m_bConnected;

    // Flags for whether game pad was just inserted or removed
    bool		m_bInserted;
    bool		m_bRemoved;

};

static Asura_Xbox360_Controller g_axControllers[XUSER_MAX_COUNT];

// ancillary converter function taken from MS sample code
static inline float ConvertThumbstickValue( short sThumbstickValue )
{
    if( sThumbstickValue > +sINPUT_DEADZONE )
    {
        return (sThumbstickValue-sINPUT_DEADZONE) / (32767.0f-sINPUT_DEADZONE);
    }
    if( sThumbstickValue < -sINPUT_DEADZONE )
    {
        return (sThumbstickValue+sINPUT_DEADZONE+1.0f) / (32767.0f-sINPUT_DEADZONE);
    }
    return 0.0f;
}



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_GetNumberOfActiveControllers
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Returns the number of active controllers, this may be restricted
//				  to consecutive controllers from port 0.
//
// Parameters   : "true" if we're to check for consecutive controllers.
//
// Return value : The number of active controllers found.
////////////////////////////////////////////////////////////////////////////////

int Asura_UserInput::Platform_GetNumberOfActiveControllers (bool bConsecutive)
{
	int iConnected = 0;
	for(DWORD xUser = 0; xUser < XUSER_MAX_COUNT; ++xUser)
	{
		XINPUT_STATE xState;
		DWORD xResult = XInputGetState(xUser,&xState);
		Asura_Assertf( (xResult == ERROR_SUCCESS) || (xResult == ERROR_DEVICE_NOT_CONNECTED) , ("Failed to get user input state for user %d",xUser) );
		
		if(xResult == ERROR_SUCCESS)
		{
			++iConnected;	
		}
	}

    return iConnected;
}

/// End of function 'Asura_UserInput::Platform_GetNumberOfActiveControllers'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_Initialise
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Xbox360 specific initialisation for user-input.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_UserInput::Platform_Initialise (void)
{
	ZeroMemory( g_axControllers, sizeof( g_axControllers ) );
	ZeroMemory( g_abKeyState, ASURA_KEYID_END_OF_KEYBOARD);
	ZeroMemory( g_abKeyPressedThisFrame, ASURA_KEYID_END_OF_KEYBOARD);
	
	for(int i=0;i<256;i++)
	{
		g_aiVKToAsuraKey[i]=-1;
	}
}

/// End of function 'Asura_UserInput::Platform_Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_Shutdown
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Xbox360 specific shutdown for user-input.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_UserInput::Platform_Shutdown (void)
{
}

/// End of function 'Asura_UserInput::Platform_Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_Reset
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Xbox360 specific user-input reset code.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_UserInput::Platform_Reset (void)
{
	ZeroMemory( g_axControllers, sizeof( g_axControllers ) );
}

/// End of function 'Asura_UserInput::Platform_Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : VKeyToAsuraKey
// Author       : Dave Reed
// Date         : 12-Oct-2006
//
// Description  : Maps from an Xbox360 VKey/Unicode pair to an Asura key ID
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

int VKeyToAsuraKey(int iVKey, int iUnicode)
{				
	switch(iVKey)
	{	
	case VK_TAB:	 	return ASURA_KEYID_TAB;   
	case VK_RETURN:  	return ASURA_KEYID_ENTER; 
	case VK_SHIFT:		return ASURA_KEYID_SHIFT;
	case VK_CONTROL:	return ASURA_KEYID_CTRL;
	case VK_PAUSE:	 	return ASURA_KEYID_PAUSE;
	case VK_ESCAPE:		return ASURA_KEYID_ESCAPE;
	case VK_SPACE:		return ASURA_KEYID_SPACE;
	case VK_END:		return ASURA_KEYID_END;
	case VK_HOME:		return ASURA_KEYID_HOME;
	case VK_LEFT:	 	return ASURA_KEYID_LEFT;  
	case VK_RIGHT:	 	return ASURA_KEYID_RIGHT; 
	case VK_UP:		 	return ASURA_KEYID_UP;	
	case VK_DOWN:	 	return ASURA_KEYID_DOWN;	
	case VK_INSERT:		return ASURA_KEYID_INS;
	case VK_DELETE:	 	return ASURA_KEYID_DEL;   
	case VK_NUMPAD0:	return ASURA_KEYID_NUMPAD0;
	case VK_NUMPAD1:	return ASURA_KEYID_NUMPAD1;
	case VK_NUMPAD2:	return ASURA_KEYID_NUMPAD2;
	case VK_NUMPAD3:	return ASURA_KEYID_NUMPAD3;
	case VK_NUMPAD4:	return ASURA_KEYID_NUMPAD4;
	case VK_NUMPAD5:	return ASURA_KEYID_NUMPAD5;
	case VK_NUMPAD6:	return ASURA_KEYID_NUMPAD6;
	case VK_NUMPAD7:	return ASURA_KEYID_NUMPAD7;
	case VK_NUMPAD8:	return ASURA_KEYID_NUMPAD8;
	case VK_NUMPAD9:	return ASURA_KEYID_NUMPAD9;
	case VK_MULTIPLY:	return ASURA_KEYID_NUMPADMULTIPLY;
	case VK_ADD:		return ASURA_KEYID_NUMPADADD;
	case VK_SUBTRACT:	return ASURA_KEYID_NUMPADSUB;
	case VK_DECIMAL:	return ASURA_KEYID_NUMPADDEL;
	case VK_DIVIDE:		return ASURA_KEYID_NUMPADDIVIDE;
	case VK_F1:			return ASURA_KEYID_F1;
	case VK_F2:			return ASURA_KEYID_F2;
	case VK_F3:			return ASURA_KEYID_F3;
	case VK_F4:			return ASURA_KEYID_F4;
	case VK_F5:			return ASURA_KEYID_F5;
	case VK_F6:			return ASURA_KEYID_F6;
	case VK_F7:			return ASURA_KEYID_F7;
	case VK_F8:			return ASURA_KEYID_F8;
	case VK_F9:			return ASURA_KEYID_F9;
	case VK_F10:		return ASURA_KEYID_F10;
	case VK_F11:		return ASURA_KEYID_F11;
	case VK_F12:		return ASURA_KEYID_F12;
	case VK_F13:		return ASURA_KEYID_F13;
	case VK_F14:		return ASURA_KEYID_F14;
	case VK_F15:		return ASURA_KEYID_F15;
	case VK_F16:		return ASURA_KEYID_F16;
	case VK_NUMLOCK:	return ASURA_KEYID_NUMLOCK;
	case VK_SCROLL:		return ASURA_KEYID_SCROLLOK;

	default:
		if (iUnicode<=0xff)
		{
			// Handle standard ASCII keys
			return iUnicode;
		}
		else
		{
			return 0;
		}
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_Update
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Xbox360 specific user-input update code.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_UserInput::Platform_Update (void)
{
	// Loop through all controllers
	for( u_int u=0; u<XUSER_MAX_COUNT; ++u )
	{
		Asura_Xbox360_Controller& xController = g_axControllers[ u ];

		// Note if the controller was connected last frame, and try to get the input state (if we can't, it's not connected)
		bool bWasConnected = xController.m_bConnected;

		XINPUT_STATE xInputState;
		xController.m_bConnected = ( XInputGetState( u, &xInputState ) == ERROR_SUCCESS );

		// Track insertion and removals
		xController.m_bRemoved  = (  bWasConnected && !xController.m_bConnected );
		xController.m_bInserted = ( !bWasConnected &&  xController.m_bConnected );

		// if the controller isn't connected, we're done
		if( !xController.m_bConnected )	continue;

		// handle a newly inserted controller
		if( xController.m_bInserted )
		{
			// probably can't trust what's in memory so we'll wipe it
			// (previously pressed buttons, user may have swapped controllers etc.)
			ZeroMemory( &xController, sizeof( xController ) );

			xController.m_bConnected = true;
			xController.m_bInserted  = true;

			// fill in the capabilities, currently (2005/10/10) not doing much with this,
			// since we only have one type of controller to play with.
			XInputGetCapabilities( u, XINPUT_FLAG_GAMEPAD, &xController.m_xCaps );
		}

		// Let's look at our input data then. I guess we should be looking at the capabilities
		// here really to see what kind of controller it is?

		// Let's get those thumbsticks into -1 to +1 floats, including the deadzone.
		xController.m_fLeftX = ConvertThumbstickValue( xInputState.Gamepad.sThumbLX );
		xController.m_fLeftY = ConvertThumbstickValue( xInputState.Gamepad.sThumbLY );
		xController.m_fRightX = ConvertThumbstickValue( xInputState.Gamepad.sThumbRX );
		xController.m_fRightY = ConvertThumbstickValue( xInputState.Gamepad.sThumbRY );

		// Which buttons have been pressed?
		xController.m_usPressedButtons = xInputState.Gamepad.wButtons;

		// Triggers are analogue (0..255)...
		xController.m_ucLeftTrigger = xInputState.Gamepad.bLeftTrigger;
		xController.m_ucRightTrigger = xInputState.Gamepad.bRightTrigger;
		
		// Set the user index for this gamepad
		xController.m_usUserIndex = u;
	}

	// tell the independent interface about all of this
	// note we'll have to zero controllers that aren't plugged in

	// Loop through all controllers
	for( u_int u = 0; u < XUSER_MAX_COUNT; ++u )
	{
		Asura_Xbox360_Controller& xController = g_axControllers[u];

		// Update the current axis values held by Asura...
		AxisValue[ u * uASURA_XBOX360_NUM_AXES_IN_PAD + ASURA_AXIS_PAD1_LSTICK_H ] = xController.m_fLeftX;
		AxisValue[ u * uASURA_XBOX360_NUM_AXES_IN_PAD + ASURA_AXIS_PAD1_LSTICK_V ] = xController.m_fLeftY;
		AxisValue[ u * uASURA_XBOX360_NUM_AXES_IN_PAD + ASURA_AXIS_PAD1_RSTICK_H ] = xController.m_fRightX;
		AxisValue[ u * uASURA_XBOX360_NUM_AXES_IN_PAD + ASURA_AXIS_PAD1_RSTICK_V ] = -xController.m_fRightY;

		// Update the current button values held by Asura...
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_A ]			= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_A) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_B ]			= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_B) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_X ]			= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_X) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_Y ]			= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_Y) == 0) ? 0 : 255;

		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_L_BUMPER ]	= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_LEFT_SHOULDER) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_R_BUMPER ]	= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_L_THUMB ]	= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_LEFT_THUMB) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_R_THUMB ]	= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_RIGHT_THUMB) == 0) ? 0 : 255;

		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_START ]		= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_START) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_BACK ]		= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_BACK) == 0) ? 0 : 255;

		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_UP ]			= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_DPAD_UP) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_DOWN ]		= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_DPAD_DOWN) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_LEFT ]		= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_DPAD_LEFT) == 0) ? 0 : 255;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_RIGHT ]		= ((xController.m_usPressedButtons & XINPUT_GAMEPAD_DPAD_RIGHT) == 0) ? 0 : 255;

		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_L_TRIGGER ]	= xController.m_ucLeftTrigger;
		CurrentInput[ u * uASURA_XBOX360_NUM_BUTTONS_IN_PAD + ASURA_XBOX360_KEYID_PAD1_R_TRIGGER ]	= xController.m_ucRightTrigger;
	}



	// ----------------------------------------------
	//   Update the keyboard state (DaveR 12/10/06)
	// ----------------------------------------------
	XINPUT_KEYSTROKE xKeyStroke;
	g_iInputBufferSize = 0;

	while ( XInputGetKeystroke(0,XINPUT_FLAG_KEYBOARD, &xKeyStroke) == ERROR_SUCCESS)
	{		
		if (xKeyStroke.Flags & XINPUT_KEYSTROKE_KEYDOWN)
		{
			int iChar = VKeyToAsuraKey(xKeyStroke.VirtualKey, xKeyStroke.Unicode);

			g_abKeyState[iChar] = true;
			g_abKeyPressedThisFrame[iChar] = true;

			if (xKeyStroke.VirtualKey<256)
			{	
				// Add the keypress to the input buffer (used by the console)
				if (g_iInputBufferSize< g_iMaxInputBufferSize)
				{		
					g_wszInputBuffer[g_iInputBufferSize++] = (Asura_UnicodeChar)iChar;
				}
				g_aiVKToAsuraKey[xKeyStroke.VirtualKey] = (Asura_UnicodeChar)iChar;
			}
		}
		
		if (xKeyStroke.Flags & XINPUT_KEYSTROKE_KEYUP)
		{
			// We don't get the unicode key value on key-up messages, so we use the VK->AsuraKey mapping
			// that we recorded when the key was pressed...
			int iAsuraKey = g_aiVKToAsuraKey[xKeyStroke.VirtualKey];
			if (iAsuraKey>=0 && iAsuraKey<ASURA_KEYID_END_OF_KEYBOARD)
			{				
				g_abKeyState[iAsuraKey] = false;
			}
		}			
	}		
	
	// Update Asura's CurrentInput[] array - flag the keys that were down during this frame
	for(int i=0; i<ASURA_KEYID_END_OF_KEYBOARD; i++)
	{
		if(g_abKeyState[i] || g_abKeyPressedThisFrame[i])
		{
			CurrentInput[i] = 255;
			g_abKeyPressedThisFrame[i] = false;
		}		
	}

	// Update input buffer
	s_wszInputBuffer   = g_wszInputBuffer;
	s_uInputBufferSize = g_iInputBufferSize;

}

/// End of function 'Asura_UserInput::Platform_Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_UpdateInputBuffer
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Updates the specified input buffer using the USB-keyboard.
//
// Parameters   : 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_UserInput::Platform_UpdateInputBuffer (Asura_UnicodeChar* wszBuffer,
												  u_int uMaxChars,
												  u_int& uInputCursorPos,
												  bool (*pfnHandleInputOverride)(ASURA_KEYID eKey), 
												  const Asura_UnicodeChar* wszInputBuffer,
												  u_int uInputBufferSize)
{

}

/// End of function 'Asura_UserInput::Platform_UpdateInputBuffer'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_IsControllerActive
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Checks whether the specified player is active.
//
// Parameters   : The ID of the player to check for.
//
// Return value : "true" if active, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_UserInput::Platform_IsControllerActive (int iPlayerNum)
{
	//we take player to mean the controller port
	bool bConnected = false;
	if( (iPlayerNum >= 0) && (iPlayerNum < XUSER_MAX_COUNT) )
	{
		bConnected = g_axControllers[iPlayerNum].m_bConnected;
	}
	return bConnected;
}

/// End of function 'Asura_UserInput::Platform_IsControllerActive'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_UserInput::Platform_IsKeyToggled
// Author       : Kevin Floyer-Lea
// Date         : 23-Jun-2005
//
// Description  : Checks whether the specified key has been toggled this frame.
//
// Parameters   : The ID of the key to check for.
//
// Return value : "true" if the specified key has been toggled this frame.
////////////////////////////////////////////////////////////////////////////////

bool Asura_UserInput::Platform_IsKeyToggled (int iKey)
{
	if (iKey<ASURA_KEYID_END_OF_KEYBOARD)
	{	
		return g_abKeyState[iKey];
	}

	return false;
}

/// End of function 'Asura_UserInput::Platform_IsKeyToggled'. ///

/// End of file. ///
