//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_LiveSession.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     class to wrap Live! sessions. Based on CSession from Sessions sample
//     code.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Online/Asura_Xbox360_LiveSession.h"

// Asura Xbox360 includes
#include "Core/Asura_Xbox360_OS.h"

// do not include Asura_Lobby.h etc.
#include <Winsockx.h>
#include <xonline.h>

//--------------------------------------------------------------------------------------
// Constants
//--------------------------------------------------------------------------------------
const u_int uASURA_XBOX360_SESSION_MODIFY_FLAGS_ALLOWED = 
	XSESSION_CREATE_USES_ARBITRATION |
	XSESSION_CREATE_INVITES_DISABLED |
	XSESSION_CREATE_JOIN_VIA_PRESENCE_DISABLED |
	XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED;


//--------------------------------------------------------------------------------------
// Name: Session()
// Desc: Session constructor -- initialize values
//--------------------------------------------------------------------------------------
Asura_Xbox360_LiveSession::Asura_Xbox360_LiveSession( u_int uOwnerController, 
													 bool bIsHost, 
													 bool bIsRanked, 
													 bool bAllowJoinInProgress,
													 u_int uPublicSlots, 
													 u_int uPrivateSlots )
{
	m_xSession                  = INVALID_HANDLE_VALUE;
	m_bIsHost                   = bIsHost;
	m_bIsRanked                 = bIsRanked;
	m_bUsingQoS                 = false;
	ZeroMemory( &m_xSessionInfo, sizeof( m_xSessionInfo ) );
	m_ulSessionNonce            = 0;
	m_uSessionFlags             = bIsRanked ? XSESSION_CREATE_LIVE_MULTIPLAYER_RANKED : XSESSION_CREATE_LIVE_MULTIPLAYER_STANDARD;
	if (!bAllowJoinInProgress)
	{
		m_uSessionFlags |= XSESSION_CREATE_JOIN_IN_PROGRESS_DISABLED;
	}
	m_uOwnerController          = uOwnerController;
	ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );

	Asura_Assert(uPrivateSlots + uPublicSlots >= 0, "can't create session with no slots");
	m_uNumPrivateSlotsFilled = 0;
	m_uTotalPrivateSlots = uPrivateSlots;
	m_uNumPublicSlotsFilled = 0;
	m_uTotalPublicSlots = uPublicSlots;
	m_pxPublicXUIDs = new XUID[uPublicSlots];
	m_pxPrivateXUIDs = new XUID[uPrivateSlots];

	m_pxRegistrationResults     = NULL;

	m_eSessionState             = ASURA_XBOX360_LIVESESSION_STATE_NONE;
}


//--------------------------------------------------------------------------------------
// Name: ~Asura_Xbox360_LiveSession()
// Desc: Session destructor -- clean up outstanding sessions
//--------------------------------------------------------------------------------------
Asura_Xbox360_LiveSession::~Asura_Xbox360_LiveSession()
{
	delete[] m_pxRegistrationResults;
	delete[] m_pxPublicXUIDs;
	delete[] m_pxPrivateXUIDs;

	if( m_xSession != INVALID_HANDLE_VALUE )
	{
		StopQoSListener();
#ifdef ASURA_ASSERT
		u_int uRet = 
#endif
		XSessionDelete( m_xSession, NULL );
		Asura_Assertf(uRet == ERROR_SUCCESS, ("XSessionDelete failed with error %u", uRet));
		CloseHandle( m_xSession );
		m_xSession = INVALID_HANDLE_VALUE;
	}
}


//--------------------------------------------------------------------------------------
// Name: CreateSession()
// Desc: Creates a new session 
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::CreateSession()
{
	Asura_Assert( m_eSessionState == ASURA_XBOX360_LIVESESSION_STATE_NONE, "CreateSession called on existing session!" );

	// Configure the slots
	m_uNumPublicSlotsFilled = 0;
	m_uNumPrivateSlotsFilled = 0;

	if( m_bIsHost )
	{
		m_uSessionFlags |= XSESSION_CREATE_HOST;
	}

	// Advertise the session
	ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );

	if( INVALID_HANDLE_VALUE != m_xSession )
	{
		CloseHandle( m_xSession );
		m_xSession = INVALID_HANDLE_VALUE;
	}

#ifdef ASURA_ASSERT
	u_int uRet = 
#endif		
	XSessionCreate(
		m_uSessionFlags,
		m_uOwnerController,
		m_uTotalPublicSlots,
		m_uTotalPrivateSlots,
		&m_ulSessionNonce,
		&m_xSessionInfo,
		&m_xOverlapped,
		&m_xSession );

	Asura_Assertf( uRet == ERROR_IO_PENDING, ("XSessionCreate failed with error %d\n", uRet ));
}


//--------------------------------------------------------------------------------------
// Name: MigrateSession()
// Desc: Called by new host to take ownership of a session, and other members to rejoin it
// Arg: The session to migrate to, or NULL to create a new session (as the owner)
// Ret: The new session information on pending success or NULL on failiure
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::MigrateSession( const XSESSION_INFO* pxNewSession /*= NULL*/ )
{
	Asura_Assert( m_eSessionState == ASURA_XBOX360_LIVESESSION_STATE_IDLE, "MigrateSession called during invalid state" );

	DWORD dwUserIndex = (pxNewSession) ? XUSER_INDEX_NONE : m_uOwnerController;
	if (pxNewSession)
	{
		Asura_Mem::Copy( &m_xMigratingSessionInfo, pxNewSession, sizeof(m_xMigratingSessionInfo));
	}
	else
	{
		Asura_Mem::Set( &m_xMigratingSessionInfo, 0, sizeof (m_xMigratingSessionInfo));
	}
	
	int uRet = XSessionMigrateHost(
			m_xSession,
			dwUserIndex,
			&m_xMigratingSessionInfo,
			&m_xOverlapped
		);

	Asura_Assertf( uRet == ERROR_IO_PENDING, ("XSessionCreate failed with error %d\n", uRet ));
	if (uRet == ERROR_IO_PENDING)
	{
		SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_MIGRATING );
	}
}


//--------------------------------------------------------------------------------------
// Name: ModifySessionFlags()
// Desc: Modify the session, to account for the host's correct flag settings (join)
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::ModifySessionFlags( u_int uFlags )
{
	if( m_eSessionState > ASURA_XBOX360_LIVESESSION_STATE_CREATING && m_eSessionState < ASURA_XBOX360_LIVESESSION_STATE_DELETING )
	{
		if( !m_bIsRanked )
		{

			// turn the allowed modify uFlags off
			m_uSessionFlags &= ~uASURA_XBOX360_SESSION_MODIFY_FLAGS_ALLOWED;
			// and get them from the uFlags input
			m_uSessionFlags |= (uFlags & uASURA_XBOX360_SESSION_MODIFY_FLAGS_ALLOWED);

			Asura_Assert(false, "TODO: make this asynchronous");

			u_int uRet = XSessionModify(
				m_xSession,
				m_uSessionFlags,
				m_uTotalPublicSlots,
				m_uTotalPrivateSlots,
				NULL               // do it synchronously for simplicity; actual titles should
				);                 // be asynchronous

			if( uRet != ERROR_SUCCESS )
			{
#ifdef ASURA_ASSERT
				HRESULT hr = 
#endif
					XGetOverlappedExtendedError( NULL );

				Asura_Assertf( false, ("Failed to modify session uFlags, error 0x%08x\n", hr ));
			}
		}
	}
}


//--------------------------------------------------------------------------------------
// Name: StartQoSListener()
// Desc: Turn on the Quality of Service Listener for the current session
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::StartQoSListener( u_char* pucData, u_int uDataLen, u_int uBitsPerSec )
{
	u_int uFlags = XNET_QOS_LISTEN_SET_DATA;
	if( uBitsPerSec )
	{
		uFlags |= XNET_QOS_LISTEN_SET_BITSPERSEC;
	}
	if( !m_bUsingQoS )
	{
		uFlags |= XNET_QOS_LISTEN_ENABLE;
	}
	u_int uRet = XNetQosListen( &(m_xSessionInfo.sessionID), pucData, uDataLen, uBitsPerSec, uFlags );
	if( ERROR_SUCCESS != uRet )
	{
		Asura_Assertf(false, ( "Failed to start QoS listener, error 0x%08x\n", uRet ));
	}
	m_bUsingQoS = true;
}


//--------------------------------------------------------------------------------------
// Name: StopQoSListener()
// Desc: Turn off the Quality of Service Listener for the current session
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::StopQoSListener()
{
	if( m_bUsingQoS )
	{
		u_int dwRet;
		dwRet = XNetQosListen( &(m_xSessionInfo.sessionID), NULL, 0, 0, XNET_QOS_LISTEN_RELEASE );
		if( ERROR_SUCCESS != dwRet )
		{
			Asura_Assertf(false, ( "Warning: Failed to stop QoS listener, error 0x%08x\n", dwRet ));
		}
		m_bUsingQoS = false;
	}
}


//--------------------------------------------------------------------------------------
// Name: IsSameSession()
// Desc: Returns true if this session info is for the existing session
//--------------------------------------------------------------------------------------
bool Asura_Xbox360_LiveSession::IsSameSession( const XSESSION_INFO& sessionInfo ) const
{
	if( memcmp( &(sessionInfo.sessionID), &(m_xSessionInfo.sessionID),
		sizeof( m_xSessionInfo.sessionID ) ) )
	{
		return false;
	}
	return true;

}


//--------------------------------------------------------------------------------------
// Name: AddLocalPlayers()
// Desc: Add players on the local console to the session
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::AddLocalPlayer( bool bInvited )
{
	const u_int uNumPlayers = 1;
	const DWORD uUserIndex = 0;
	const BOOL invited = bInvited;

	XUID xLocalXuid;
#ifdef ASURA_ASSERT
	DWORD uRet = 
#endif
		XUserGetXUID(Asura_Xbox360_OS::GetActivePlayerIndex(), &xLocalXuid);
	Asura_Assert(uRet == ERROR_SUCCESS, "XUserGetXUID failed");

	// TODO: use an async operation - need to put vars into global storage instead of on stack.
	u_int dwRet = XSessionJoinLocal(
		m_xSession,
		uNumPlayers,
		&uUserIndex,
		&invited,
		NULL );

	if( dwRet != ERROR_SUCCESS )
	{
#ifdef ASURA_ASSERT
		HRESULT hr = 
#endif
			XGetOverlappedExtendedError( NULL );

		Asura_Assertf(false, ( "Failed to add local users to session, error 0x%08x\n", hr ));
	}
	else
	{
		// if invited, first try to grab a private slot, if that fails, use a public slot.
		if( bInvited && (m_uNumPrivateSlotsFilled < m_uTotalPrivateSlots))
		{
			m_pxPrivateXUIDs[m_uNumPrivateSlotsFilled++] = xLocalXuid;
		}
		else
		{
			Asura_Assert( m_uNumPublicSlotsFilled < m_uTotalPublicSlots, "Too many slots filled!\n" );
			if (m_uNumPublicSlotsFilled < m_uTotalPublicSlots)
			{
				m_pxPublicXUIDs[m_uNumPublicSlotsFilled++] = xLocalXuid;
			}
		}
	}
}


//--------------------------------------------------------------------------------------
// Name: AddRemotePlayers()
// Desc: Add players on remote consoles to the session
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::AddRemotePlayers(u_int uNumPlayers, XUID* pxXUIDs, BOOL* pxInvited)
{
	// TODO: use an async operation - need to put vars into global storage instead of on stack.
	u_int dwRet = XSessionJoinRemote(
		m_xSession,
		uNumPlayers,
		pxXUIDs,
		pxInvited,
		NULL );

	if( dwRet != ERROR_SUCCESS )
	{
#ifdef ASURA_ASSERT
		HRESULT hr = 
#endif
			XGetOverlappedExtendedError( NULL );

		Asura_Assertf(false, ( "Failed to add remote users to session, error 0x%08x\n", hr ));
	}
	else
	{
		// if invited, first try to grab a private slot, if that fails, use a public slot.
		for (u_int u = 0; u < uNumPlayers; ++u)
		{
			if( pxInvited[u] && (m_uNumPrivateSlotsFilled < m_uTotalPrivateSlots)  )
			{
				m_pxPrivateXUIDs[m_uNumPrivateSlotsFilled++] = pxXUIDs[u];
			}
			else if (m_uNumPublicSlotsFilled < m_uTotalPublicSlots)
			{
				m_pxPublicXUIDs[m_uNumPublicSlotsFilled++] = pxXUIDs[u];
			}
			else
			{
				Asura_Assert( false, "Too many slots filled!" );
			}
		}
	}
}


//--------------------------------------------------------------------------------------
// Name: RemoveLocalPlayers()
// Desc: Remove players on the local console from the session
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::RemoveLocalPlayer()
{
	// If we're in an arbitrated session and the game is in progress, we have someone
	// dropping in the middle of the session. Force his scores to zero and write stats
	if( ( m_uSessionFlags & XSESSION_CREATE_USES_ARBITRATION ) &&
		( (ASURA_XBOX360_LIVESESSION_STATE_STARTING == m_eSessionState) ||
		(ASURA_XBOX360_LIVESESSION_STATE_IN_GAME == m_eSessionState) ) )
	{
		// TODO: call XSessionWriteStats() 
	}

	u_int uNumPlayers = 1;
	DWORD uUserIndex = 0;

	// TODO: use an async operation - need to put vars into global storage instead of on stack.
	u_int uRet = XSessionLeaveLocal(
		m_xSession,
		uNumPlayers,
		&uUserIndex,
		NULL );

	if( uRet != ERROR_SUCCESS )
	{
#ifdef ASURA_ASSERT
		HRESULT hr = 
#endif
			XGetOverlappedExtendedError( NULL );

		Asura_Assertf( false, ( "Failed to remove local users from session, error 0x%08x\n", hr ));
	}
}


//--------------------------------------------------------------------------------------
// Name: RemoveRemotePlayers()
// Desc: Remove players on a remote console from the session
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::RemoveRemotePlayers(u_int uNumPlayers, XUID* pxXUIDs)
{
	// If we're in an arbitrated session and the game is in progress, we have someone
	// dropping in the middle of the session. Force his scores to zero and write stats
	if( ( m_uSessionFlags & XSESSION_CREATE_USES_ARBITRATION ) &&
		( (ASURA_XBOX360_LIVESESSION_STATE_STARTING == m_eSessionState) ||
		(ASURA_XBOX360_LIVESESSION_STATE_IN_GAME == m_eSessionState) ) )
	{
		// TODO: call XSessionWriteStats() 
	}

	// TODO: use an async operation - need to put vars into global storage instead of on stack.
	u_int uRet = XSessionLeaveRemote(
		m_xSession,
		uNumPlayers,
		pxXUIDs,
		NULL );

	if( uRet != ERROR_SUCCESS )
	{
#ifdef ASURA_ASSERT
		HRESULT hr = 
#endif
			XGetOverlappedExtendedError( NULL );

		Asura_Assertf(false, ( "Failed to remove remote users from session, error 0x%08x\n", hr ));
	}
	else
	{
		// update free slots
		for (u_int uPlayer = 0; uPlayer < uNumPlayers; ++uPlayer)
		{
			bool bInvited = false;
			bool bFound = false;
			for (u_int uSlot = 0; uSlot < m_uNumPrivateSlotsFilled; ++uSlot)
			{
				if (m_pxPrivateXUIDs[uSlot] == pxXUIDs[uPlayer])
				{
					// swap with last entry
					m_pxPrivateXUIDs[uSlot] = m_pxPrivateXUIDs[m_uNumPrivateSlotsFilled - 1];
					bInvited = true;
					bFound = true;
					break;
				}
			}
			if (!bInvited)
			{
				// must be in a public slot
				for (u_int uSlot = 0; uSlot < m_uNumPublicSlotsFilled; ++uSlot)
				{
					if (m_pxPublicXUIDs[uSlot] == pxXUIDs[uPlayer])
					{
						// swap with last entry
						m_pxPublicXUIDs[uSlot] = m_pxPublicXUIDs[m_uNumPublicSlotsFilled - 1];
						bFound = true;
						break;
					}
				}
			}
			if (bFound)
			{
				if( bInvited )
				{
					Asura_Assert(m_uNumPrivateSlotsFilled, "slots error");
					--m_uNumPrivateSlotsFilled;
				}
				else
				{
					Asura_Assert(m_uNumPublicSlotsFilled, "slots error");
					--m_uNumPublicSlotsFilled;
				}
			}
			else
			{
				Asura_Assert(false, "failed to find player to remove");
			}
		}
	}
}


bool Asura_Xbox360_LiveSession::IsPlayerInSession(XUID xXUID) const
{
	// search private slots
	for (u_int uSlot = 0; uSlot < m_uNumPrivateSlotsFilled; ++uSlot)
	{
		if (m_pxPrivateXUIDs[uSlot] == xXUID)
		{
			return true;
		}
	}
	// search public slots
	for (u_int uSlot = 0; uSlot < m_uNumPublicSlotsFilled; ++uSlot)
	{
		if (m_pxPublicXUIDs[uSlot] == xXUID)
		{
			return true;
		}
	}
	return false;
}

XUID Asura_Xbox360_LiveSession::GetXUIDForPrivateSlot(u_int uSlot) const
{
	Asura_Assert( uSlot < m_uNumPrivateSlotsFilled, "invalid slot");
	return m_pxPrivateXUIDs[uSlot];
}

XUID Asura_Xbox360_LiveSession::GetXUIDForPublicSlot(u_int uSlot) const
{
	Asura_Assert( uSlot < m_uNumPublicSlotsFilled, "invalid slot");
	return m_pxPublicXUIDs[uSlot];
}


//--------------------------------------------------------------------------------------
// Name: RegisterForArbitration()
// Desc: Register self for arbitration
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::RegisterForArbitration()
{
	DWORD uRegResultsSize = 0;
	m_pxRegistrationResults = NULL;

	ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );

	// Call once to determine size of results buffer
	u_int ret = XSessionArbitrationRegister(
		m_xSession,
		0,
		m_ulSessionNonce,
		&uRegResultsSize,
		NULL,
		NULL); 

	if( ( ret != ERROR_INSUFFICIENT_BUFFER ) || ( uRegResultsSize == 0 ) )
	{
		Asura_Assertf(false, ( "Failed on first call to XSessionArbitrationRegister, hr=0x%08x\n", ret ));
	}

	m_pxRegistrationResults = (PXSESSION_REGISTRATION_RESULTS)new u_char[ uRegResultsSize ];
	if( m_pxRegistrationResults == NULL )
	{
		Asura_Assert(false, "Failed to allocate buffer.\n" );
	}


	ret = XSessionArbitrationRegister(
		m_xSession,
		0,
		m_ulSessionNonce,
		&uRegResultsSize,
		m_pxRegistrationResults,
		&m_xOverlapped );

	if( ret != ERROR_IO_PENDING )
	{
#ifdef ASURA_ASSERT
		HRESULT hr = 
#endif
			XGetOverlappedExtendedError( NULL );

		Asura_Assertf(false, ( "Failed to register the joined session, hr=0x%08x\n", hr ));
	}

	m_eSessionState = ASURA_XBOX360_LIVESESSION_STATE_REGISTERING;
}


//--------------------------------------------------------------------------------------
// Name: SwitchToState()
// Desc: Changes to a new session state and performs initialization for the new state
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::SwitchToState( ASURA_XBOX360_LIVESESSION_STATE newState )
{
	// Clean up from the previous state
	switch( m_eSessionState )
	{
	case ASURA_XBOX360_LIVESESSION_STATE_IN_GAME:
	case ASURA_XBOX360_LIVESESSION_STATE_FINISHED:
		break;

	case ASURA_XBOX360_LIVESESSION_STATE_REGISTERED:
	{
		if( m_pxRegistrationResults )
		{
			delete[] m_pxRegistrationResults;
			m_pxRegistrationResults = NULL;
		}

		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_NONE:
	case ASURA_XBOX360_LIVESESSION_STATE_CREATING:
	case ASURA_XBOX360_LIVESESSION_STATE_IDLE:
	case ASURA_XBOX360_LIVESESSION_STATE_REGISTERING:
	case ASURA_XBOX360_LIVESESSION_STATE_STARTING:
	case ASURA_XBOX360_LIVESESSION_STATE_ENDING:
	case ASURA_XBOX360_LIVESESSION_STATE_DELETING:
	default:
		break;
	}

	// Initialize the next state
	switch( newState )
	{
	case ASURA_XBOX360_LIVESESSION_STATE_CREATING:
	{
		CreateSession();
		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_REGISTERING:
	{
		RegisterForArbitration();
		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_STARTING:
	{
		ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );
#ifdef ASURA_ASSERT
		u_int uRet = 
#endif
		XSessionStart( m_xSession, 0, &m_xOverlapped );
		Asura_Assertf(uRet == ERROR_IO_PENDING, ("XSessionStart failed with error %u", uRet));
		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_ENDING:
	{
		ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );
#ifdef ASURA_ASSERT
		u_int uRet = 
#endif
		XSessionEnd( m_xSession, &m_xOverlapped );
		Asura_Assertf(uRet == ERROR_IO_PENDING, ("XSessionEnd failed with error %u", uRet));
		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_DELETING:
	{
		ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );
		if( m_xSession != INVALID_HANDLE_VALUE )
		{
			StopQoSListener();
#ifdef ASURA_ASSERT
			u_int uRet = 
#endif
			XSessionDelete( m_xSession, &m_xOverlapped );
			Asura_Assertf(uRet == ERROR_IO_PENDING, ("XSessionDelete failed with error %u", uRet));
		}
		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_ERROR:
	{
		Asura_Assert(false, "Live! Session error");
		break;
	}

	case ASURA_XBOX360_LIVESESSION_STATE_NONE:
	case ASURA_XBOX360_LIVESESSION_STATE_IDLE:
	case ASURA_XBOX360_LIVESESSION_STATE_REGISTERED:
	case ASURA_XBOX360_LIVESESSION_STATE_IN_GAME:
	case ASURA_XBOX360_LIVESESSION_STATE_FINISHED:
	default:
		break;
	}

	m_eSessionState = newState;
}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::Update()
// Desc: Called by sample update, processes overlapped call checks
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::Update()
{
	switch( m_eSessionState )
	{
	case ASURA_XBOX360_LIVESESSION_STATE_IDLE:
	case ASURA_XBOX360_LIVESESSION_STATE_REGISTERED:
	case ASURA_XBOX360_LIVESESSION_STATE_IN_GAME:
		break;
	case ASURA_XBOX360_LIVESESSION_STATE_CREATING:
		UpdateCreating();
		break;
	case ASURA_XBOX360_LIVESESSION_STATE_REGISTERING:
		UpdateRegistering();
		break;
	case ASURA_XBOX360_LIVESESSION_STATE_STARTING:
		UpdateStarting();
		break;
	case ASURA_XBOX360_LIVESESSION_STATE_ENDING:
		UpdateEnding();
		break;
	case ASURA_XBOX360_LIVESESSION_STATE_DELETING:
		UpdateDeleting();
		break;
	case ASURA_XBOX360_LIVESESSION_STATE_MIGRATING:
		UpdateMigrating();
		break;
	default:
		Asura_Assert(false, "unhandled state");
		break;
	}
}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::UpdateCreating()
// Desc: Check for completion of XSessionCreate
//--------------------------------------------------------------------------------------
void Asura_Xbox360_LiveSession::UpdateCreating()
{
	// If the overlapped IO is still going, wait
	if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
	{
		// If there was no error, we're in
		HRESULT hr = XGetOverlappedExtendedError( &m_xOverlapped );

		if( FAILED( hr ) )
		{
			Asura_Assertf( false, ("XSessionCreate overlapped task failed with error 0x%08x", hr));
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_ERROR );
		}
		else
		{
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_IDLE );
		}
	}
}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::UpdateRegistering()
// Desc: Check for completion of XSessionArbitrationRegister
//--------------------------------------------------------------------------------------
VOID Asura_Xbox360_LiveSession::UpdateRegistering()
{
	if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
	{
		if( SUCCEEDED( XGetOverlappedExtendedError( &m_xOverlapped ) ) )
		{
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_REGISTERED );
		}
		else
		{
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_ERROR );
		}
	}
}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::UpdateStarting()
// Desc: Check for completion of XSessionStart
//--------------------------------------------------------------------------------------
VOID Asura_Xbox360_LiveSession::UpdateStarting()
{
	if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
	{
		if( SUCCEEDED( XGetOverlappedExtendedError( &m_xOverlapped ) ) )
		{
			// The session is started; start playing
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_IN_GAME );
		}
		else
		{
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_ERROR );
		}
	}

}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::UpdateEnding()
// Desc: Check for completion of XSessionEnd
//--------------------------------------------------------------------------------------
VOID Asura_Xbox360_LiveSession::UpdateEnding()
{
	if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
	{
		if( SUCCEEDED( XGetOverlappedExtendedError( &m_xOverlapped ) ) )
		{
			// We're done, move to post-game
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_FINISHED );
		}
		else
		{
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_ERROR );
		}
	}
}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::UpdateDeleting()
// Desc: Check for completion of XSessionDelete
//--------------------------------------------------------------------------------------
VOID Asura_Xbox360_LiveSession::UpdateDeleting()
{
	if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
	{
		CloseHandle( m_xSession );
		m_xSession = INVALID_HANDLE_VALUE;
		SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_NONE );
	}
}


//--------------------------------------------------------------------------------------
// Name: Asura_Xbox360_LiveSession::UpdateMigrating()
// Desc: Check for completion of XSessionMigrateHost
//--------------------------------------------------------------------------------------
VOID Asura_Xbox360_LiveSession::UpdateMigrating()
{
	if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
	{
		if( SUCCEEDED( XGetOverlappedExtendedError( &m_xOverlapped ) ) )
		{
			m_xSessionInfo = m_xMigratingSessionInfo;
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_IDLE );
		}
		else
		{
			Asura_Assert(false, "Error changing ownership of live session");
			SwitchToState( ASURA_XBOX360_LIVESESSION_STATE_ERROR );
		}
	}
}

