///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_Xbox360_Storage.cpp
//
// Created by Martin G.Harris (25/2/2009)
// 
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
// Description:
//
//		Classes for reading/writing data to Xbox LIVE storage.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Network/Online/Asura_Xbox360_Storage.h"

#include "Core/Asura_Xbox360_OS.h"

///////////////////////////////////////////////////////////////////////////////
// Asura_Xbox360_StorageTransactionManager
///////////////////////////////////////////////////////////////////////////////
Asura_Xbox360_StorageTransactionManager::Asura_Xbox360_StorageTransactionManager(XSTORAGE_FACILITY eFacility /*= XSTORAGE_FACILITY_PER_USER_TITLE*/)
: m_uUserIndex(0)
, m_eFacility(eFacility)
, m_dwResult(ERROR_SUCCESS)
, m_bError(false)
, m_bTransactionInProgress(false)
{
	Reset();
}

Asura_Xbox360_StorageTransactionManager::~Asura_Xbox360_StorageTransactionManager()
{
	// Cancel any in-progress transaction operation.
	if (m_bTransactionInProgress)
	{	
#if defined(ASURA_ASSERT)
		const BOOL bCompleted =
#endif
			XCancelOverlapped(&m_xOverlapped);
		Asura_Assert(bCompleted, "Cancelled XStorage transaction operation!");
	}
}

///////////
// Reset //
///////////
// Reset the transaction manager ready for the next operation.
void Asura_Xbox360_StorageTransactionManager::Reset()
{	
	// Cancel any in-progress transaction operation.
	if (m_bTransactionInProgress)
	{
#if defined(ASURA_ASSERT)
		const BOOL bCompleted =
#endif
			XCancelOverlapped(&m_xOverlapped);
		Asura_Assert(bCompleted, "Cancelled XStorage transaction operation!");
	}

	m_bTransactionInProgress = false;
	m_bError = false;
	m_dwResult = ERROR_SUCCESS;

	m_uUserIndex = Asura_Xbox360_OS::GetActivePlayerIndex();
	
	ZeroMemory( &m_xOverlapped, sizeof(m_xOverlapped) );
	ZeroMemory( &m_xResults, sizeof(m_xResults) );
}

/////////////////////
// BuildServerPath //
/////////////////////
void Asura_Xbox360_StorageTransactionManager::BuildServerPath( const WCHAR* wszStoredItemName,
															   WCHAR* wszServerPathName,
															   DWORD* pdwServerPathLen )
{
	Asura_Assert(m_eFacility != XSTORAGE_FACILITY_GAME_CLIP, "Game clips not currently supported");

	m_dwResult = XStorageBuildServerPath( m_uUserIndex, m_eFacility,
										  NULL, 0,
										  wszStoredItemName,
										  wszServerPathName, pdwServerPathLen );

	if ( m_dwResult != ERROR_SUCCESS )
	{
		Asura_Assertf(0, ("Error (0x%08x) building XStorage server path", m_dwResult));
		SetError();
	}
}

//////////////////
// GetDataAsync //
//////////////////
// Download the file from the Xbox LIVE server asynchronously.
// If the buffer is not large enough to hold the entire storage item then the download will fail.
// Parameters: Name of file; memory buffer to store the data; and the size of this buffer.
void Asura_Xbox360_StorageTransactionManager::GetDataAsync(const WCHAR* wszStoredItemName, void* pBuffer, u_int uBufferSize)
{
	// Cancel any operation that is in progress.
	Reset();

	if (wszStoredItemName == 0 || *wszStoredItemName == 0)
	{
		Asura_Assert(0, "Null XStorage item name!");
		SetError();
		return;
	}

	if (pBuffer == 0 || uBufferSize == 0)
	{
		Asura_Assert(0, "Null data buffer!");
		SetError();
		return;
	}

	WCHAR wszServerPathName[512];
    DWORD dwServerPathLen = ARRAYSIZE(wszServerPathName);

	BuildServerPath( wszStoredItemName, wszServerPathName, &dwServerPathLen );
	if (HasErrorOccurred()) return;

	// Zero the buffer and structs
	ZeroMemory( pBuffer, uBufferSize );

	// Start reading the file.
    m_dwResult = XStorageDownloadToMemory(m_uUserIndex,
										  wszServerPathName,
										  uBufferSize, static_cast<const BYTE*>(pBuffer),
										  sizeof( XSTORAGE_DOWNLOAD_TO_MEMORY_RESULTS ), &m_xResults,
										  &m_xOverlapped);

	if (m_dwResult == ERROR_IO_PENDING)
	{
		m_bTransactionInProgress = true;
	}
	else
	{
		Asura_Assertf(0, ("Error (0x%08x) downloading XStorage file: %S", m_dwResult, wszStoredItemName));
		SetError();
	}
}

//////////////////
// SetDataAsync //
//////////////////
// Upload the file to the Xbox LIVE server asynchronously.
// The specified buffer must remain valid and unmodified until the upload operation completes.
// Parameters: Name of file; memory containing the data; and the size of this data.
void Asura_Xbox360_StorageTransactionManager::SetDataAsync(const WCHAR* wszStoredItemName, void* pData, u_int uDataSize)
{
	// Cancel any operation that is in progress.
	Reset();

	if (wszStoredItemName == 0 || *wszStoredItemName == 0)
	{
		Asura_Assert(0, "Null XStorage item name!");
		SetError();
		return;
	}

	if (pData == 0 || uDataSize == 0)
	{
		Asura_Assert(0, "Null data buffer!");
		SetError();
		return;
	}

	// Writing files is only allowed for the per-user storage facility,
	// (and for game clips which aren't implemented yet).
	if (m_eFacility != XSTORAGE_FACILITY_PER_USER_TITLE)
	{
		Asura_Assert(0, "Writing only allowed for per-user XStorage");
		SetError();
		return;
	}

	WCHAR wszServerPathName[512];
    DWORD dwServerPathLen = ARRAYSIZE(wszServerPathName);

	BuildServerPath( wszStoredItemName, wszServerPathName, &dwServerPathLen );
	if (HasErrorOccurred()) return;

	// Start writing the file.
    m_dwResult = XStorageUploadFromMemory(m_uUserIndex,
										  wszServerPathName,
										  uDataSize, static_cast<const BYTE*>(pData),
										  &m_xOverlapped);

	if (m_dwResult == ERROR_IO_PENDING)
	{
		m_bTransactionInProgress = true;
	}
	else
	{
		Asura_Assertf(0, ("Error (0x%08x) uploading XStorage file: %S", m_dwResult, wszStoredItemName));
		SetError();
	}
}

/////////////////
// DeleteAsync //
/////////////////
// Delete the file from the Xbox LIVE server asynchronously.
// Parameters: Name of file to delete.
void Asura_Xbox360_StorageTransactionManager::DeleteAsync(const WCHAR* wszStoredItemName)
{
	// Cancel any operation that is in progress.
	Reset();

	if (wszStoredItemName == 0 || *wszStoredItemName == 0)
	{
		Asura_Assert(0, "Null XStorage item name!");
		SetError();
		return;
	}

	WCHAR wszServerPathName[512];
    DWORD dwServerPathLen = ARRAYSIZE(wszServerPathName);

	BuildServerPath( wszStoredItemName, wszServerPathName, &dwServerPathLen );
	if (HasErrorOccurred()) return;

	// Deleting files is only allowed for the per-user storage facility.
	if (m_eFacility != XSTORAGE_FACILITY_PER_USER_TITLE)
	{
		Asura_Assert(0, "Deleting only allowed for per-user XStorage");
		SetError();
		return;
	}

	// Start the process of deleting the file.
	m_dwResult = XStorageDelete( m_uUserIndex, wszServerPathName, &m_xOverlapped );

	if (m_dwResult == ERROR_IO_PENDING)
	{
		m_bTransactionInProgress = true;
	}
	else
	{
		Asura_Assertf(0, ("Error (0x%08x) deleting XStorage file: %S", m_dwResult, wszStoredItemName));
		SetError();
	}
}

////////////
// Update //
////////////
// Poll the IO data to see whether the operation has finished.
// Return true when finished or if an error occurs.
bool Asura_Xbox360_StorageTransactionManager::Update()
{
	// If the overlapped IO is still going then wait until it completes.
	if (m_bTransactionInProgress)
	{
		if ( XHasOverlappedIoCompleted( &m_xOverlapped ) )
		{
			m_bTransactionInProgress = false;
		
			m_dwResult = XGetOverlappedExtendedError( &m_xOverlapped );
			
			if ( m_dwResult != ERROR_SUCCESS )
			{
				// Note: 'file not found' may not be an actual error, since this typically happens
				// the first time the user runs the game, when it tries to read their per user data.
				Asura_Assertf(m_dwResult == XONLINE_E_STORAGE_FILE_NOT_FOUND, ("Error (0x%08x) XStorage transaction failed", m_dwResult));
				SetError();
			}
		}
	}

	return !m_bTransactionInProgress;
}
