//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Voice.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     wrapper for XHV/XHV2 - deals with online voice comms
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_LOBBY

// This file's header.
#include "Network/Online/Asura_Xbox360_Voice.h"

// Asura independent includes
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_NetTransport_System.h"

// Asura 360 includes
#include "Core/Asura_Xbox360_OS.h"
#include "Network/Online/Asura_Xbox360_Friends.h"
#include "Sound/Asura_XAudio2.h"

// C O N S T A N T S / ///////////////////////////////////////////////////////
static const float fASURA_XBOX360_MAX_TIME_BETWEEN_VOICE_SENDS = 0.2f; // 200 ms

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
IXHV2Engine* Asura_Xbox360_Voice::s_pxXHVEngine;
bool Asura_Xbox360_Voice::s_bFriendsOnly;
float Asura_Xbox360_Voice::s_fForceSendTimer;
Asura_Xbox360_Voice_Msg* Asura_Xbox360_Voice::s_pxVoiceDataMsg;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_Xbox360_Voice::Initialise(u_int uMaxRemoteTalkers)
{
	Asura_Assert(uMaxRemoteTalkers <= XHV_MAX_REMOTE_TALKERS, "too many remote talkers");
	uMaxRemoteTalkers = Asura_Maths::Min( uMaxRemoteTalkers, u_int(XHV_MAX_REMOTE_TALKERS) );

	Asura_Assert(!s_pxXHVEngine, "XHV already initialised");
	const u_int uPlayerIndex = Asura_Xbox360_OS::GetActivePlayerIndex();
	if (CheckVoicePrivileges(uPlayerIndex))
	{
		HANDLE  m_xWorkerThread; // This is provided so that the title can use XSetThreadProcessor to schedule the thread on a different processor. 
		XHV_PROCESSING_MODE xMode = XHV_VOICECHAT_MODE;

		// Set up parameters for the voice chat engine
		XHV_INIT_PARAMS xXhvParams = {0};
		xXhvParams.dwMaxRemoteTalkers            = uMaxRemoteTalkers;
		xXhvParams.dwMaxLocalTalkers             = 1;
		xXhvParams.localTalkerEnabledModes       = &xMode;
		xXhvParams.remoteTalkerEnabledModes      = &xMode;
		xXhvParams.dwNumLocalTalkerEnabledModes  = 1;
		xXhvParams.dwNumRemoteTalkerEnabledModes = 1;
		xXhvParams.pXAudio2						 = Asura_XAudio2::GetEngine(); 

		// Create the engine
		HRESULT hr = XHV2CreateEngine( &xXhvParams, &m_xWorkerThread, &s_pxXHVEngine );
		Asura_Assertf(hr == S_OK, ("XHVCreateEngine() failed 0x%x", hr));

		if (hr == S_OK)
		{
			// register local user
			XHV_PROCESSING_MODE xLocalProcMode = XHV_VOICECHAT_MODE;

			hr = s_pxXHVEngine->RegisterLocalTalker( uPlayerIndex );
			Asura_Assertf(hr == S_OK, ("RegisterLocalTalker() failed 0x%x", hr));
			if ( hr == S_OK )
			{
				s_pxXHVEngine->StartLocalProcessingModes( uPlayerIndex, &xLocalProcMode, 1 );
			}
		}

		// allocate voice message
		XUID xLocalXuid;
		hr = XUserGetXUID(uPlayerIndex, &xLocalXuid);
		Asura_Assertf(hr == S_OK, ("XUserGetXUID() failed 0x%x", hr));

		s_pxVoiceDataMsg = new Asura_Xbox360_Voice_Msg(xLocalXuid);
		s_fForceSendTimer = 0.0f;
	}
}

void Asura_Xbox360_Voice::Shutdown()
{
	delete s_pxVoiceDataMsg;
	s_pxVoiceDataMsg = NULL;

	if (!s_pxXHVEngine)
	{
		return;
	}

#ifdef ASURA_ASSERT
	HRESULT hr = 
#endif
		s_pxXHVEngine->UnregisterLocalTalker(Asura_Xbox360_OS::GetActivePlayerIndex());
	Asura_Assertf(hr == S_OK, ("UnregisterLocalTalker() failed 0x%x", hr));

	s_pxXHVEngine->Release();
	s_pxXHVEngine = NULL;
}

bool Asura_Xbox360_Voice::Update()
{
	if (!s_pxXHVEngine)
	{
		return false;
	}

	const u_int uVoiceDataBytes = s_pxVoiceDataMsg->m_usDataSize;
	Asura_Assert(uVoiceDataBytes < uASURA_XBOX360_MAX_VOICE_PAYLOAD, "voice buffer is full");
	BYTE* pucVoiceData = s_pxVoiceDataMsg->m_aucPayload + uVoiceDataBytes;
	DWORD uSize = uASURA_XBOX360_MAX_VOICE_PAYLOAD - uVoiceDataBytes;
	DWORD uNumPackets = 0;
	HRESULT hr = s_pxXHVEngine->GetLocalChatData(Asura_Xbox360_OS::GetActivePlayerIndex(), pucVoiceData, &uSize, &uNumPackets);
	Asura_Assertf(hr == S_OK || hr == E_PENDING, ("GetLocalChatData() failed 0x%x", hr));

	if (hr == S_OK)
	{
		Asura_Assert(uVoiceDataBytes + uSize < USHRT_MAX, "voice data buffer too large");
		s_pxVoiceDataMsg->m_usDataSize = static_cast<u_short>(uVoiceDataBytes + uSize);
		Asura_Assert(s_pxVoiceDataMsg->m_usDataSize <= uASURA_XBOX360_MAX_VOICE_PAYLOAD, "buffer overrun");
	}

	// if we can pending data, increase timer to ensure it gets sent soon
	if (s_pxVoiceDataMsg->m_usDataSize)
	{
		s_fForceSendTimer += Asura_Timers::GetFrameTime();
	}

	// send if the buffer is full or we've not sent voice for a certain time
	if ((s_pxVoiceDataMsg->m_usDataSize == uASURA_XBOX360_MAX_VOICE_PAYLOAD) || 
		(s_fForceSendTimer >= fASURA_XBOX360_MAX_TIME_BETWEEN_VOICE_SENDS) )
	{
		return true;
	}
	return false;
}

bool Asura_Xbox360_Voice::CheckVoicePrivileges(u_int uPlayerIndex)
{
#ifdef ASURA_ASSERT
	DWORD uError;
#endif
	BOOL bResult = FALSE;

#ifdef ASURA_ASSERT
	uError = 
#endif
		XUserCheckPrivilege(uPlayerIndex, XPRIVILEGE_COMMUNICATIONS, &bResult);
	Asura_Assertf(uError == S_OK, ("XUserCheckPrivilege() failed 0x%x", uError));

	if (bResult)
	{
		// Gamer is allowed to talk to all other gamers.
		s_bFriendsOnly = false;
	}
	else
	{
#ifdef ASURA_ASSERT
		uError = 
#endif
			XUserCheckPrivilege(uPlayerIndex, XPRIVILEGE_COMMUNICATIONS_FRIENDS_ONLY, &bResult);
		Asura_Assertf(uError == S_OK, ("XUserCheckPrivilege() failed 0x%x", uError));

		if (bResult)
		{
			// Gamer is allowed to talk only to members of the gamer's friends list.
			s_bFriendsOnly = true;
		}
		else
		{
			// Gamer is not allowed to talk to any other gamer.
			return false;
		}
	}

	return true;
}


void Asura_Xbox360_Voice::RegisterRemotePlayer(const XUID& xXUID)
{
	if (s_pxXHVEngine)
	{
		XHV_PROCESSING_MODE xRemoteProcMode = XHV_VOICECHAT_MODE;

		DWORD uError = s_pxXHVEngine->RegisterRemoteTalker( xXUID, NULL, NULL, NULL );
		Asura_Assertf(uError == S_OK, ("RegisterRemoteTalker() failed 0x%x", uError));
		if ( uError == S_OK )
		{
			s_pxXHVEngine->StartRemoteProcessingModes( xXUID, &xRemoteProcMode, 1 );
		}
	}
}

void Asura_Xbox360_Voice::UnregisterRemotePlayer(const XUID& xXUID)
{
	if (s_pxXHVEngine)
	{
#ifdef ASURA_ASSERT
		DWORD uError = 
#endif
		s_pxXHVEngine->UnregisterRemoteTalker( xXUID );
		Asura_Assertf(uError == S_OK, ("UnregisterRemoteTalker() failed 0x%x", uError));
	}

}

void Asura_Xbox360_Voice::VoiceMsgSent()
{
	Asura_Assert(s_pxVoiceDataMsg, "voice not initialised");
	if (s_pxVoiceDataMsg)
	{
		s_fForceSendTimer = 0.0f;
		s_pxVoiceDataMsg->m_usDataSize = 0;
	}
}

bool Asura_Xbox360_Voice::HandleMessage( const char* pcMessage, int iLength )
{
	const Asura_Xbox360_Voice_Msg* pxMsg = reinterpret_cast<const Asura_Xbox360_Voice_Msg*>(pcMessage);
	Asura_Assert(pxMsg->m_usMsgID == ASURA_NETTRANSPORT_MSG_VOICE_DATA, "unexpected voice message");
	if (pxMsg->m_usMsgID != ASURA_NETTRANSPORT_MSG_VOICE_DATA)
	{
		return false;
	}

	if (s_bFriendsOnly)
	{
		if (!Asura_Xbox360_Friends::IsPlayerAFriend(pxMsg->m_xXuidSrc))
		{
			Asura_OutputToDebuggerOnce(("Asura_Xbox360_Voice::HandleMessage: dropping voice message "
				"due to 'friends only' setting. This message will not appear again.\n"));
			return true;
		}
	}

	BOOL iShouldMute = FALSE;
	const u_int uPlayerIndex = Asura_Xbox360_OS::GetActivePlayerIndex();
	if (XUserMuteListQuery(uPlayerIndex, pxMsg->m_xXuidSrc, &iShouldMute) == ERROR_SUCCESS)
	{
		if (iShouldMute)
		{
			Asura_OutputToDebuggerOnce(("Asura_Xbox360_Voice::HandleMessage: dropping voice message "
				"due to user being muted. This message will not appear again.\n"));
			return true;
		}
	}

	DWORD uSize = pxMsg->m_usDataSize;
#ifdef ASURA_ASSERT
	DWORD uError = 
#endif
	s_pxXHVEngine->SubmitIncomingChatData(pxMsg->m_xXuidSrc, pxMsg->m_aucPayload, &uSize);
	// from the Docs: If pdwSize contains a length on return that is less than the length 
	// that was sent to the function, it is because the incoming packet buffer is full. 
	// In this scenario, simply wait, to allow the buffer to clear out. Then resubmit the remaining packets.
	// (MH - 11.09.2008) Just going to drop them in this case, like the samples do.
	if (uSize < pxMsg->m_usDataSize)
	{
		Asura_OutputToDebugger_Retail(("Asura_Xbox360_Voice::HandleMessage: dropping %u bytes of incoming voice data\n", uSize));
	}
	else
	{
		Asura_Assertf(uError == S_OK, ("SubmitIncomingChatData() failed 0x%x", uError));
	}

	return true;
}

bool Asura_Xbox360_Voice::IsPlayerTalking(const XUID& xXUID)
{
	if (s_pxXHVEngine)
	{
		const u_int uPlayerIndex = Asura_Xbox360_OS::GetActivePlayerIndex();
		XUID xLocalXuid;
#ifdef ASURA_ASSERT
		HRESULT hr = 
#endif
			XUserGetXUID(uPlayerIndex, &xLocalXuid);
		Asura_Assertf(hr == S_OK, ("XUserGetXUID() failed 0x%x", hr));
		if (xXUID == xLocalXuid)
		{
			return s_pxXHVEngine->IsLocalTalking(uPlayerIndex) != 0;
		}
		else
		{
			return s_pxXHVEngine->IsRemoteTalking(xXUID) != 0;
		}
	}
	return false;
}

// Is the headset attached?
bool Asura_Xbox360_Voice::IsHeadsetPresent()
{
	if (s_pxXHVEngine)
	{
		const u_int uPlayerIndex = Asura_Xbox360_OS::GetActivePlayerIndex();
		return s_pxXHVEngine->IsHeadsetPresent(uPlayerIndex) != 0;
	}
	return false;
}

Asura_Xbox360_Voice_MsgHeader::Asura_Xbox360_Voice_MsgHeader(const XUID& xXuid)
{
	m_usMsgID = ASURA_NETTRANSPORT_MSG_VOICE_DATA;
	m_usDataSize = 0;
	m_xXuidSrc = xXuid;
}

#endif // #ifdef ASURA_SUPPORT_LOBBY
