//
// Asura_Xbox360_ObjectRenderer.cpp
//
// Created by KJFL - 09 November 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Rendering/Asura_Xbox360_ObjectRenderer.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_TextureAnimation.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#include "Shadows/Asura_Shadow_System.h"
#include "Shapes/Asura_Shapes.h"

// Asura Xbox 360
#include "Debugging/Asura_Platform_GPUProfile.h"
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_IndexBuffer.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Rendering/Asura_Xbox360_VertexBuffer.h"
#include "Shapes/Asura_Platform_Shape.h"

// Shader Include Files
#include "Object_MaterialResponse.h"
#include "Object_Shadow.h"

// Misc
#include <fxl.h>
#include <xtl.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Remeber to update vertex size in Asura_Chunk_ResourceFile_Xbox360_Object.cpp
// if this changes

const D3DVERTEXELEMENT9 axVERTEX_DECLARATION[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 12, D3DDECLTYPE_FLOAT16_2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT16_2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 1 },
	{ 0, 20, D3DDECLTYPE_FLOAT1,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 2 },
	{ 0, 24, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0 },
	{ 0, 28, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,  0 },
	{ 0, 32, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0 },
	D3DDECL_END()
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

// Standard object effect

static Asura_Xbox360_Effect* g_pxObjectEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_xWorldMatrixHandle = 0;

static Asura_Xbox360_Effect_System::Handle g_xWorldViewProjHandle = 0;

static Asura_Xbox360_MaterialResponse_Handles g_xMaterialResponseHandles;

// Object shadow effect

static Asura_Xbox360_Effect* g_pxObjectShadowEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_xShadowWorldViewProjHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowFrustumNearPlaneHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xOneOverShadowFrustumDepthHandle = 0;

// Vertex declaration

static Direct3D_VertexDeclaration* g_pxObjectVertexDeclaration = NULL;

/******************************************************************************/
/* Asura_Xbox360_ObjectRenderer Static Class Data                             */
/******************************************************************************/

Asura_TemplateHashTree< Asura_Xbox360_ObjectRenderer::Object > Asura_Xbox360_ObjectRenderer::s_xObjectTree;

/******************************************************************************/
/* Asura_Xbox360_ObjectRenderer::Object Class Functions                       */
/******************************************************************************/

Asura_Xbox360_ObjectRenderer::Object::Object()
	: m_uNumberOfStrips(0)
	, m_pxStrips( NULL )
{
}

Asura_Xbox360_ObjectRenderer::Object::~Object()
{
	m_xVertexBuffer.Destroy();
	m_xIndexBuffer.Destroy();

	delete[] m_pxStrips;
}

/******************************************************************************/
/* Asura_Xbox360_ObjectRenderer Static Class Function                         */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::Initialise()
//
// Initialises the Xbox 360 object renderer
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_ObjectRenderer::Initialise()
{
	// Standard object effect

	g_pxObjectEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auObject_MaterialResponse);	

	g_xWorldViewProjHandle = g_pxObjectEffect->GetParameterHandle( "g_xWorldViewProjectionMatrix" );

	g_xMaterialResponseHandles.InitialiseFromEffect( *g_pxObjectEffect );

	g_xWorldMatrixHandle = g_pxObjectEffect->GetParameterHandle( "g_xWorldMatrix" );

	// Object shadow effect

	g_pxObjectShadowEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auObject_Shadow);

	g_xShadowWorldViewProjHandle = g_pxObjectShadowEffect->GetParameterHandle( "g_xWorldViewProj" );
	g_xShadowFrustumNearPlaneHandle = g_pxObjectShadowEffect->GetParameterHandle( "g_fShadowFrustumNearPlane" );
	g_xOneOverShadowFrustumDepthHandle = g_pxObjectShadowEffect->GetParameterHandle( "g_fOneOverShadowFrustumDepth" );

	// Vertex declaration

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxObjectVertexDeclaration );
	Asura_Assert( g_pxObjectVertexDeclaration, "Failed to create object vertex declaration." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::Reset()
//
// Resets the Xbox 360 object renderer. Called between levels
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_ObjectRenderer::Reset()
{
	s_xObjectTree.Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::Shutdown()
//
// Shuts the Xbox 360 object renderer down
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_ObjectRenderer::Shutdown()
{
	Reset();

	Asura_Xbox360_Effect_System::Destroy( g_pxObjectEffect );
	g_pxObjectEffect = NULL;

	Asura_Xbox360_Effect_System::Destroy( g_pxObjectShadowEffect );
	g_pxObjectShadowEffect = NULL;

	g_pxObjectVertexDeclaration->Release();
	g_pxObjectVertexDeclaration = NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::CreateObject()
//
// Creates a new Xbox 360 object mesh
//
// Parameters:		uHashID - Hash ID of the object's name
// Return value:	Pointer to the object
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_ObjectRenderer::Object* Asura_Xbox360_ObjectRenderer::CreateObject( const Asura_Hash_ID uHashID )
{
	#if defined ASURA_ASSERT
		const Object* const pxExistingObject = s_xObjectTree.GetDataByHash( uHashID );
		Asura_Assertf( !pxExistingObject, ("An object with this hash (0x%X) has already been created.", uHashID) );
	#endif

	Object* const pxNewObject = s_xObjectTree.AddNode( uHashID );
	Asura_Assert( pxNewObject, "Unable to create new object." );

	return pxNewObject;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::GetMaterialIndex()
//
// Retrieves the primary material index for the first strip of an object
//
// Parameters:		xShape - Reference to shape
// Return value:	Material ID
//
////////////////////////////////////////////////////////////////////////////////

int Asura_Xbox360_ObjectRenderer::GetMaterialIndex( const Asura_Shape &xShape )
{
	const Object* const pxObject = s_xObjectTree.GetDataByHash( xShape.GetShapeHashID() );
	Asura_Assert( pxObject, "Unable to find object." );

	return pxObject->m_pxStrips[0].m_auMaterialResponseHashIDs[0];
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::RenderObject()
//
// Renders an object mesh on 360
//
// Parameters:		xShape - Reference to the shape to be rendered
//					xShapeInstance - Reference to the specific shape instance
//					uAmbientARGB - Ambient colour to add, and object opacity
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_ObjectRenderer::RenderObject( const Asura_Shape& xShape, const Asura_Shape_Instance& xShapeInstance, const u_int uAmbientARGB )
{
	Object* const pxObject = s_xObjectTree.GetDataByHash( xShape.GetShapeHashID() );
	Asura_Assert( pxObject, "Unable to find object to render." );

	if( pxObject->m_xIndexBuffer.IsEmpty() || pxObject->m_xVertexBuffer.IsEmpty() )
	{
		#if defined ASURA_ASSERT
			#if defined _DEBUG
				const char* const szShapeName = static_cast< const Asura_Platform_Shape_RenderData* >( xShape.GetRenderData() )->m_szName;
			#else
				const char* const szShapeName = "<Need Debug build>";
			#endif
		#endif

		Asura_Assertf( xShape.IsNeverRendered(), ("Unrenderable shape %s hasn't been tagged as 'never render'.", szShapeName) );

		return;
	}

	const bool bTransparentPass = Asura_SortedRender::IsTransparentPass();

	const Asura_Hash_ID uMaterialHash1 = pxObject->m_pxStrips[0].m_auMaterialResponseHashIDs[0];
	const Asura_MaterialResponse* const pxMaterial1
		= Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

	if( pxMaterial1 )
	{
		if( pxMaterial1->GetDontRender() ) return;

		const bool bTransparentStrip = pxMaterial1->IsTransparent();
		if( bTransparentStrip != bTransparentPass ) return;
	}

	SCOPEDPIXEVENT(( D3DCOLOR_XRGB( 0, 0xFF, 0xFF ), "Object 0x%X (Material Response)", xShape.GetShapeHashID() ));
	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_SHAPES);

	// Set material data
	{
		const Asura_Hash_ID uMaterialHash2 = pxObject->m_pxStrips[0].m_auMaterialResponseHashIDs[1];

		const Asura_MaterialResponse* const pxMaterial2 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash2 );

		xShapeInstance.UpdateLighting();

		Asura_Xbox360_MaterialResponse_System::SetEffectParameters( pxMaterial1, pxMaterial2, NULL, g_xMaterialResponseHandles, *g_pxObjectEffect );
		Asura_Xbox360_MaterialResponse_System::SetBlendStates( pxMaterial1 );
		Asura_Xbox360_MaterialResponse_System::SetSHParameters( pxMaterial1, g_xMaterialResponseHandles, *g_pxObjectEffect );
	}

	// Set world and world-view-proj matrices
	{
		D3DXMATRIX xWorldViewProj;

		// Rotation
		Asura_Mem::Set( &xWorldViewProj, 0, sizeof( xWorldViewProj ) );
		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xWorldViewProj.m[u][v] = xShapeInstance.GetOrientation().Element[v][u];
			}
		}

		xWorldViewProj._44 = 1.0f;

		// Translation
		D3DXMATRIX xTemp;
		const Asura_Vector_3& xPosition = xShapeInstance.GetPosition();
		D3DXMatrixTranslation( &xTemp, xPosition.x, xPosition.y, xPosition.z );
		xWorldViewProj *= xTemp;

		g_pxObjectEffect->SetMatrix( g_xWorldMatrixHandle, xWorldViewProj );

		// View-projection
		xWorldViewProj *= Asura_Xbox360_Graphics::GetViewProjectionMatrix();

		g_pxObjectEffect->SetMatrix( g_xWorldViewProjHandle, xWorldViewProj );
	}

	// Set render states
	{
		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestDitheringMode( true );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );

		// Leave wireframe as it was set elsewhere
	}

	// Set up streams and draw

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxObjectVertexDeclaration );
	pxObject->m_xIndexBuffer.Bind();
	pxObject->m_xVertexBuffer.Bind();

	const u_int uTechnique = bTransparentPass ? 1 : 0;

	g_pxObjectEffect->Begin( uTechnique );
	g_pxObjectEffect->Commit();

	// Render each strip

	for( u_int u = 0; u < pxObject->m_uNumberOfStrips; ++u )
	{
		const Strip& xStrip	= pxObject->m_pxStrips[u];

		if( xStrip.m_uNumberOfIndices > 0 )
		{
			SCOPEDPIXEVENT(( D3DCOLOR_XRGB( 0, 0xC0, 0xC0 ), "Strip %u", u ));

			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0, xStrip.m_uStartIndex, xStrip.m_uNumberOfIndices );
			Asura_Render::AddToTrianglesRendered( xStrip.m_uNumberOfTriangles );
		}
	}

	g_pxObjectEffect->End();

	Asura_Xbox360_Graphics::ResetDepthBias();

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );				
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL);				
	Asura_Xbox360_Graphics::SetTranslucencyModeMRT( 2, ASURA_TRANSLUCENCY_OFF );

	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_SHAPES);
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_ObjectRenderer::RenderShadow()
//
// Renders an object shadow on 360
//
// Parameters:		xShape - Reference to the shape to be used
//					xShapeInstance - Reference to the specific shape instance
//					xShadow - Reference to the shadow to be rendered
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_ObjectRenderer::RenderShadow( const Asura_Shape& xShape, const Asura_Shape_Instance& xShapeInstance, const Asura_Shadow& xShadow )
{
	Object* const pxObject = s_xObjectTree.GetDataByHash( xShape.GetShapeHashID() );
	Asura_Assert( pxObject, "Unable to find object to render." );

	if( pxObject->m_xIndexBuffer.IsEmpty() || pxObject->m_xVertexBuffer.IsEmpty() )
	{
		#if defined ASURA_ASSERT
			#if defined _DEBUG
				const char* const szShapeName = static_cast< const Asura_Platform_Shape_RenderData* >( xShape.GetRenderData() )->m_szName;
			#else
				const char* const szShapeName = "<Need Debug build>";
			#endif
		#endif

		Asura_Assertf( xShape.IsNeverRendered(), ("Unrenderable shape %s hasn't been tagged as 'never render'.", szShapeName) );

		return;
	}

	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0, 0xFF, 0xFF ), "Object Shadow 0x%X", xShape.GetShapeHashID() );

	pxObject->m_xIndexBuffer.Bind();
	pxObject->m_xVertexBuffer.Bind();

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxObjectVertexDeclaration );

	// Set world-view-proj matrix

	D3DXMATRIX xWorldViewProj;
	memset( &xWorldViewProj, 0, sizeof( D3DXMATRIX ) );

	for( u_int u = 0; u < 3; ++u )
	{
		for( u_int v = 0; v < 3; ++v )
		{
			xWorldViewProj.m[u][v] = xShapeInstance.GetOrientation().Element[v][u];
		}
	}

	xWorldViewProj._44 = 1.0f;

	const Asura_Vector_3& xPosition = xShapeInstance.GetPosition();
	D3DXMATRIX xTemp;

	D3DXMatrixTranslation( &xTemp, xPosition.x, xPosition.y, xPosition.z );
	xWorldViewProj *= xTemp;

	// Light position

	const Asura_Vector_3& xLightPos = xShadow.GetLightPosition();
	D3DXMatrixTranslation( &xTemp, -xLightPos.x, -xLightPos.y, -xLightPos.z );
	xWorldViewProj *= xTemp;

	// Light orientation

	Asura_Matrix_3x3 xLightOri;
	xShadow.GetLightOrientation(xLightOri);
	memset( &xTemp, 0, sizeof( D3DXMATRIX ) );

	for( u_int u = 0; u < 3; ++u )
	{
		for( u_int v = 0; v < 3; ++v )
		{
			xTemp.m[u][v] = xLightOri.Element[u][v];

			if( v == 1 )
			{
				xTemp.m[u][v] = -xTemp.m[u][v];
			}
		}
	}

	xTemp._44 = 1.0f;

	xWorldViewProj *= xTemp;

	// Projection

	const float fVerticalFOV = Asura_Maths::DegToRad( xShadow.GetFrustum().GetVerticalFOVInDegrees() );
	D3DXMatrixPerspectiveFovLH( &xTemp, fVerticalFOV, 1.0f, xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane() );

	xWorldViewProj *= xTemp;

	g_pxObjectShadowEffect->GetInterface()->SetMatrixF( g_xShadowWorldViewProjHandle, xWorldViewProj );

	// Frustum parameters

	const float fShadowFrustumNearPlane = xShadow.GetFrustum().GetNearPlane();
	g_pxObjectShadowEffect->GetInterface()->SetFloat( g_xShadowFrustumNearPlaneHandle, fShadowFrustumNearPlane );

	const float fOneOverShadowFrustumDepth = 1.0f / (xShadow.GetFrustum().GetFarPlane() - fShadowFrustumNearPlane);
	g_pxObjectShadowEffect->GetInterface()->SetFloat( g_xOneOverShadowFrustumDepthHandle, fOneOverShadowFrustumDepth );

	g_pxObjectShadowEffect->Begin();
	g_pxObjectShadowEffect->Commit();

	// Render each strip

	for( u_int u = 0; u < pxObject->m_uNumberOfStrips; ++u )
	{
		const Strip& xStrip	= pxObject->m_pxStrips[u];

		const Asura_Hash_ID uHash = xStrip.m_auMaterialResponseHashIDs[0];
		const Asura_MaterialResponse* const pxMaterial =
			Asura_MaterialResponse_System::GetMaterialByHashID( uHash );

		if( pxMaterial && (pxMaterial->GetRenderType() == Asura_MaterialResponse::RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW) )
		{
			continue;
		}

		if( xStrip.m_uNumberOfIndices > 0 )
		{
			PIXBeginNamedEvent( D3DCOLOR_XRGB( 0, 0xC0, 0xC0 ), "Strip %u", u );

			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0, xStrip.m_uStartIndex, xStrip.m_uNumberOfIndices );
			Asura_Render::AddToTrianglesRendered( xStrip.m_uNumberOfTriangles );

			PIXEndNamedEvent(); // Strip
		}
	}

	g_pxObjectShadowEffect->End();

	PIXEndNamedEvent(); // Object
}
