//
// Asura_Xbox360_TextureManagement.cpp
//
// Created by KJFL - 27 June 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"
#include "Core/Asura_Multithread.h"

#include "Rendering/Asura_Xbox360_TextureManagement.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureManagement.h"

#if defined (ASURA_USE_TEXTURE_STREAMING)
#include "Core/Asura_Texture_Streaming.h"
#endif

#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_XPR_System.h"

// Misc
#include <xgraphics.h>

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// DDS header structures
//
// On 360, the DirectX headers don't define the DDS header structure at all. See
// DXSDK topic 'DDS File Reference' for more info
//
////////////////////////////////////////////////////////////////////////////////

struct Asura_Xbox360_DDS_PixelFormat
{
	u_int m_uSize;
	u_int m_uFlags;
	u_int m_uFourCC;
	u_int m_uRGBBitCount;
	u_int m_uRBitMask;
	u_int m_uGBitMask;
	u_int m_uBBitMask;
	u_int m_uABitMask;
};

struct Asura_Xbox360_DDS_Caps
{
	u_int m_uCaps1;
	u_int m_uCaps2;
	u_int m_auReserved[2];
};

struct Asura_Xbox360_DDS_Header
{
	u_int m_uMagic;
	u_int m_uSize;
	u_int m_uFlags;
	u_int m_uHeight;
	u_int m_uWidth;
	u_int m_uPitchOrLinearSize;
	u_int m_uDepth;
	u_int m_uMipMapCount;
	u_int m_auReserved[11];
	Asura_Xbox360_DDS_PixelFormat m_xPixelFormat;
	Asura_Xbox360_DDS_Caps m_xCaps;
	u_int m_uReserved2;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Constants for DDS file flags. Just the ones we need
const u_int DDSD_CAPS        = 0x001;
const u_int DDSCAPS_COMPLEX  = 0x008;
const u_int DDSCAPS2_CUBEMAP = 0x200;

const D3DCOLOR uASURA_XBOX360_PINK_TEXEL  = D3DCOLOR_XRGB( 0xFF, 0x80, 0x80 );
const D3DCOLOR uASURA_XBOX360_WHITE_TEXEL = D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF );

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Asura_Xbox360_Texture* g_apxTextureStorage[ ASURA_MAX_NO_OF_TEXTURES ];
#if defined( ASURA_USE_TEXTURE_STREAMING )
static Asura_Xbox360_Texture* g_apxStreamingTextureStorage[ ASURA_MAX_NO_OF_TEXTURES ];
#endif

static Asura_Xbox360_Texture  g_xPlaceholderTexture;
static Asura_Xbox360_Texture  g_xWhiteTexture;

#if defined _DEBUG
uint32 dbg_uMaxTextureSize           = 256;
#else // #if defined _DEBUG
uint32 dbg_uMaxTextureSize           = 2048;
#endif // #if defined _DEBUG

/******************************************************************************/
/* Asura_Texture_Management Static Class Functions                            */
/******************************************************************************/

void Asura_Texture_Management::Platform_DestroyTexture( int iTexID )
{
	if( iTexID < 0 || iTexID >= ASURA_MAX_NO_OF_TEXTURES )
	{
		Asura_Assertf( false, ("Attempting to destroy invalid texture index (%d).", iTexID) );
		return;
	}

	delete g_apxTextureStorage[ iTexID ];
	g_apxTextureStorage[ iTexID ] = NULL;
#if defined( ASURA_USE_TEXTURE_STREAMING )
	delete g_apxStreamingTextureStorage[ iTexID ];
	g_apxStreamingTextureStorage[ iTexID ] = NULL;
#endif
}

void Asura_Texture_Management::Platform_Initialise()
{
	Asura_Xbox360_TextureManagement::Initialise();
}

void Asura_Texture_Management::Platform_LoadTexture( const Asura_String& xName, int iTexID )
{
	Asura_Assertf( IsTextureLoaded( iTexID ) == (g_apxTextureStorage[ iTexID ] != NULL), ("Texture load state mismatch (%d).", iTexID) );
	if( IsTextureLoaded( iTexID ) ) return;

	Asura_String xPath, xFilename, xExtension, xFullPath;
	xName.SplitFilename( &xPath, &xFilename, &xExtension );

	if( xExtension.Compare( "xpr", true ) != 0 )
	{
		// If this isn't an XPR, try loading it as a standard texture file

		xFullPath = "Graphics";
		xFullPath += xName;

		const u_int uFileSize = Asura_FileAccess::GetFileSize( xFullPath.GetString() );

		const Asura_Handle iFileHandle = Asura_FileAccess::Open( xFullPath.GetString() );
		if( iFileHandle != ASURA_INVALID_HANDLE )
		{
			u_char* const pucBuffer = new u_char[ uFileSize ];
			if( pucBuffer )
			{
				if( Asura_FileAccess::ReadData( iFileHandle, pucBuffer, uFileSize ) )
				{
					// Just have to assume that textures loaded from disk aren't sRGB (they're
					// usually used for overlay quads, which we choose not to correct this way)
					Asura_Xbox360_TextureManagement::CreateTextureFromFileInMemory( iTexID, pucBuffer, uFileSize, false );
					Loaded[ iTexID ] = GetTextureResetGroup();
				}
				else
				{
					Asura_Assert( false, "Unable to read from resource file." );
				}
			}
			else
			{
				Asura_Assert( false, "Unable to allocate texture file buffer." );
			}

			Asura_FileAccess::Close( iFileHandle );
			delete[] pucBuffer;

			return;
		}
	}

	// If that failed, try looking for an XPR with the same name

	xFullPath = "Graphics";
	xFullPath += xPath;
	xFullPath += "\\";
	xFullPath += xFilename;
	xFullPath += ".xpr";

	const Asura_Handle iFileHandle = Asura_FileAccess::Open( xFullPath.GetString() );
	if( iFileHandle == ASURA_INVALID_HANDLE )
	{
		Asura_Assertf( false, ("Failed to load texture %s.", xFullPath.GetString()) );
		Loaded[ iTexID ] = ASURA_TEXTURE_ERROR_LOADING;

		return;
	}

	Asura_Chunk_File_Stream xFileStream( iFileHandle );
	Asura_Xbox360_XPR_System::ParseXPR( xFileStream );

	Asura_FileAccess::Close( iFileHandle );
}

void Asura_Texture_Management::Platform_LoadTextureFromResource( Asura_Handle iFileHandle, int iFileSize, int iTexID, u_int uChunkFlags )
{
	if( uChunkFlags & Asura_Chunk_ResourceFile::TF_XBOX360_TEXTURE_XPR )
	{
		// If this is actually an .xpr, parse it appropriately
		Asura_Chunk_File_Stream xStream( iFileHandle );
		Asura_Xbox360_XPR_System::ParseXPR( xStream );
	}
	else
	{
		u_char* const pucBuffer = new u_char[ iFileSize ];
		if( !pucBuffer )
		{
			Asura_Assert( false, "Unable to allocate texture file buffer." );
			Asura_FileAccess::SkipData( iFileHandle, iFileSize );

			return;
		}

		if( !Asura_FileAccess::ReadData( iFileHandle, pucBuffer, iFileSize ) )
		{
			Asura_Assert( false, "Unable to read from resource file." );
			delete[] pucBuffer;

			return;
		}

		const bool bSRGB = ((uChunkFlags & Asura_Chunk_ResourceFile::TF_SRGB) != 0);

		Asura_Xbox360_TextureManagement::CreateTextureFromFileInMemory( iTexID,
			pucBuffer, iFileSize, bSRGB );

		delete[] pucBuffer;
	}
}

void Asura_Texture_Management::Platform_Reset()
{
	Asura_Xbox360_TextureManagement::Reset();
}

void Asura_Texture_Management::Platform_Shutdown()
{
	Asura_Xbox360_TextureManagement::Shutdown();
}

/******************************************************************************/
/* Asura_Xbox360_TextureManagement Static Class Functions                     */
/******************************************************************************/

bool Asura_Xbox360_TextureManagement::s_bTextureBudgetEnabled = false;
int	 Asura_Xbox360_TextureManagement::s_iTextureBudget = 0;
int	 Asura_Xbox360_TextureManagement::s_iTextureBudgetExceededBy = 0;

void Asura_Xbox360_TextureManagement::Initialise()
{
	g_xPlaceholderTexture.SetupPlaceholderTexture( uASURA_XBOX360_PINK_TEXEL );
	g_xPlaceholderTexture.m_szName = Asura_StringUtil::Duplicate( "Placeholder" );

	g_xWhiteTexture.SetupPlaceholderTexture( uASURA_XBOX360_WHITE_TEXEL );
	g_xWhiteTexture.m_szName = Asura_StringUtil::Duplicate( "White" );
}

void Asura_Xbox360_TextureManagement::Reset()
{
}

void Asura_Xbox360_TextureManagement::Shutdown()
{
	Direct3D_Device*	pxDevice	= Asura_Xbox360_Graphics::GetDevice();

	pxDevice->AcquireThreadOwnership();
	pxDevice->UnsetAll();

	g_xPlaceholderTexture.Destroy();
	g_xWhiteTexture.Destroy();
}

void Asura_Xbox360_TextureManagement::CreateTextureFromXPR( const char* const szName, Direct3D_Texture* const pxD3DTexture, void* const pData, Asura_Xbox360_XPR& xXPR, bool bStreamedTexture /*= false*/  )
{
#ifdef ASURA_USE_TEXTURE_STREAMING
	if (bStreamedTexture)
	{
		// Get the independent ID
		const int iTexID = Asura_Texture_Management::AddTexture( szName );

		// Allocate an Xbox 360 texture if required
		if( !g_apxStreamingTextureStorage[ iTexID ] )
		{
			g_apxStreamingTextureStorage[ iTexID ] = new Asura_Xbox360_Texture;
			if( !g_apxStreamingTextureStorage[ iTexID ] )
			{
				Asura_Assert( false, "Unable to allocate memory for texture." );
				return;
			}
		}

		// Initialise from the XPR data
		g_apxStreamingTextureStorage[ iTexID ]->InitFromXPR( szName, pxD3DTexture, pData, xXPR );
	} 
	else
#endif
	{
		// Get the independent ID
		const int iTexID = Asura_Texture_Management::AddTexture( szName );

		// Allocate an Xbox 360 texture if required
		if( !g_apxTextureStorage[ iTexID ] )
		{
			g_apxTextureStorage[ iTexID ] = new Asura_Xbox360_Texture;
			if( !g_apxTextureStorage[ iTexID ] )
			{
				Asura_Assert( false, "Unable to allocate memory for texture." );
				return;
			}
		}

		// Initialise from the XPR data
		g_apxTextureStorage[ iTexID ]->InitFromXPR( szName, pxD3DTexture, pData, xXPR );
		Asura_Texture_Management::SetTextureLoadState( iTexID, Asura_Texture_Management::GetTextureResetGroup() );

	}
}

void Asura_Xbox360_TextureManagement::CreateTextureFromFileInMemory( const int iTexID,
	const void* const pData, const int iDataSize, const bool bSRGB, bool bStreamedTexture /*= false*/ )
{
#ifdef ASURA_USE_TEXTURE_STREAMING
	if (bStreamedTexture)
	{
		// Allocate an Xbox 360 texture if required
		if( !g_apxStreamingTextureStorage[ iTexID ] )
		{
			g_apxStreamingTextureStorage[ iTexID ] = new Asura_Xbox360_Texture;
			if( !g_apxStreamingTextureStorage[ iTexID ] )
			{
				Asura_Assert( false, "Unable to allocate memory for texture." );
				return;
			}
		}

		const Asura_String* pxName = Asura_Texture_Management::GetTextureName( iTexID );
		g_apxStreamingTextureStorage[ iTexID ]->InitFromFileInMemory( pxName ? pxName->GetString() : "Unknown", pData, iDataSize, bSRGB );
	} 
	else
#endif
	{
		// Allocate an Xbox 360 texture if required
		if( !g_apxTextureStorage[ iTexID ] )
		{
			g_apxTextureStorage[ iTexID ] = new Asura_Xbox360_Texture;
			if( !g_apxTextureStorage[ iTexID ] )
			{
				Asura_Assert( false, "Unable to allocate memory for texture." );
				return;
			}
		}

		const Asura_String* const pxName = Asura_Texture_Management::GetTextureName( iTexID );
		const char* const szName = (pxName ? pxName->GetString() : "Unknown");

		g_apxTextureStorage[ iTexID ]->InitFromFileInMemory( szName, pData, iDataSize, bSRGB );

		// The resource chunk loader takes care of the texture load state
	}
}

Asura_Xbox360_Texture* Asura_Xbox360_TextureManagement::GetTexture( const int iTexID )
{
	if (iTexID == ASURA_TEXTURE_NONE)
	{
		return &g_xWhiteTexture;
	}
	if (iTexID < 0 || iTexID > ASURA_MAX_NO_OF_TEXTURES)
	{
		return &g_xPlaceholderTexture;
	}
#ifdef ASURA_USE_TEXTURE_STREAMING
	if (g_apxStreamingTextureStorage[iTexID] && g_apxStreamingTextureStorage[iTexID]->GetD3DTexture() && Asura_Texture_Streaming_System::IsTextureLoaded(iTexID))
	{
		return g_apxStreamingTextureStorage[iTexID];
	}
#endif
	if (g_apxTextureStorage[iTexID])
	{
		return g_apxTextureStorage[iTexID];
	}
	return &g_xPlaceholderTexture;
}

/******************************************************************************/
/* Asura_Xbox360_Texture Class Functions                                      */
/******************************************************************************/

Asura_Xbox360_Texture::Asura_Xbox360_Texture()
	: m_pxD3DTexture( NULL )
	, m_pxXPR( NULL )
	, m_szName( NULL )
{
}

Asura_Xbox360_Texture::~Asura_Xbox360_Texture()
{
	Destroy();
}

void Asura_Xbox360_Texture::Destroy()
{
	SCOPED_DEVICE_OWNERSHIP();

	// MRD 23/04/09: Make sure the texture is unbound from the device before destroying it
	for( u_int u = 0; u < D3DSAMP_MAXSAMPLERS; ++u )
	{
		Direct3D_BaseTexture* pxTexture = NULL;
		if (SUCCEEDED(Asura_Xbox360_Graphics::GetDevice()->GetTexture( u, &pxTexture )))
		{

			if( pxTexture == m_pxD3DTexture )
			{
				Asura_Xbox360_Graphics::GetDevice()->SetTexture( u, NULL );
			}

			SAFE_RELEASE( pxTexture );
		}
	}

	if( m_pxXPR )
	{

		m_pxXPR->Release();
	}
	else
	{
		if( m_pxD3DTexture)
		{
			m_pxD3DTexture->Release();
		}

		delete[] m_szName;
	}

	m_pxD3DTexture = NULL;
	m_pxXPR = NULL;
	m_szName = NULL;
}

void Asura_Xbox360_Texture::InitFromFileInMemory( const char* const szName,
	const void* const pData, const int iDataSize, const bool bSRGB )
{
	Asura_Assertf( !m_pxD3DTexture, ("Trying to create texture from file in memory over existing data (%s).", m_szName ? m_szName : "") );

	const Asura_Xbox360_DDS_Header* const pxDDSHeader = static_cast< const Asura_Xbox360_DDS_Header* >( pData );
	const u_int uHeaderSize = sizeof( Asura_Xbox360_DDS_Header ) - sizeof( u_int ); // Magic value isn't included in the size


	const bool bDDSFile = pxDDSHeader->m_uMagic == 'DDS '; // Big-endian on 360, so this works OK
	const bool bCubemap = iDataSize >= sizeof( Asura_Xbox360_DDS_Header )
						&& bDDSFile 
						&& __loadwordbytereverse( 0, &pxDDSHeader->m_uSize ) ==  uHeaderSize
						&& (__loadwordbytereverse( 0, &pxDDSHeader->m_uFlags ) & DDSD_CAPS)
						&& (__loadwordbytereverse( 0, &pxDDSHeader->m_xCaps.m_uCaps1 ) & DDSCAPS_COMPLEX)
						&& (__loadwordbytereverse( 0, &pxDDSHeader->m_xCaps.m_uCaps2 ) & DDSCAPS2_CUBEMAP);

	u_int uMips = __loadwordbytereverse(0, &pxDDSHeader->m_uMipMapCount);

#if !defined(ASURA_NO_DEBUG_OUTPUT) 
	u_int uWidth = __loadwordbytereverse(0, &pxDDSHeader->m_uWidth);
	u_int uHeight = __loadwordbytereverse(0, &pxDDSHeader->m_uHeight);

	if (bDDSFile)
	{
		u_int uMipCount = 0;
		for (u_int xSmallDim = Asura_Maths::Max(uWidth, uHeight) ;xSmallDim > 0;xSmallDim >>= 1)
		{
			uMipCount += 1;
		}

		if (uMipCount != uMips)
		{
			Asura_OutputToDebugger_Retail(("DDS file %s has Mip Count %d, size is %dx%d, full mip chain would be %d\n", szName, uMips, uWidth, uHeight, uMipCount));
		}
	} else {
		Asura_OutputToDebugger_Retail(("%s is a TGA so not creating mip maps\n", szName));
	}
#endif

	D3DXIMAGE_INFO xImageInfo;
	D3DXIMAGE_INFO* const pxImageInfo = &xImageInfo;

	uMips	= (bDDSFile)
			? Asura_Maths::Max<u_int>(uMips, 1)
			: 1;

	if( bCubemap )
	{
		Direct3D_CubeTexture* pxTexture = NULL;

		if (SUCCEEDED(D3DXCreateCubeTextureFromFileInMemoryEx( Asura_Xbox360_Graphics::GetDevice(), pData, iDataSize, D3DX_DEFAULT_NONPOW2, uMips,
												 0, D3DFMT_UNKNOWN, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, pxImageInfo, NULL, &pxTexture )))
		{
			m_pxD3DTexture = pxTexture;
		}
		else
		{
			m_pxD3DTexture = NULL;
		}

	}
	else
	{
		Direct3D_Texture* pxTexture = NULL;

		if (SUCCEEDED(D3DXCreateTextureFromFileInMemoryEx( Asura_Xbox360_Graphics::GetDevice(), pData, iDataSize, D3DX_DEFAULT_NONPOW2, D3DX_DEFAULT_NONPOW2, uMips,
											 0, D3DFMT_UNKNOWN, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, pxImageInfo, NULL, &pxTexture )))
		{
			m_pxD3DTexture = pxTexture;
		}
		else
		{
			m_pxD3DTexture = NULL;
		}
	}


	bool bResize = false;

#if defined(ASURA_XBOX360_REDUCE_TEXTURES)

	// (JDS - 19/03/2007) - This is a fairly crude way to save memory. Just gets the size data from the texture, releases the existing copy
	// and reparses the in-memory file with clamped max width and height.
	if(xImageInfo.Width > dbg_uMaxTextureSize || xImageInfo.Height > dbg_uMaxTextureSize)
	{
		bResize=true;

		if(xImageInfo.Width > dbg_uMaxTextureSize)
			xImageInfo.Width = dbg_uMaxTextureSize;

		if(xImageInfo.Height > dbg_uMaxTextureSize)
			xImageInfo.Height = dbg_uMaxTextureSize;
	}
#endif

	if(Asura_Xbox360_TextureManagement::TextureBudgetEnabled() && Asura_Xbox360_TextureManagement::GetTextureBudget()<iDataSize)
	{	
		bResize=true;
		xImageInfo.Width  = 4;		
		xImageInfo.Height = 4;
	}

	if (bResize)
	{
		m_pxD3DTexture->Release();

		if( bCubemap )
		{
			Direct3D_CubeTexture* pxTexture = NULL;

			if (SUCCEEDED(D3DXCreateCubeTextureFromFileInMemoryEx( Asura_Xbox360_Graphics::GetDevice(), pData, iDataSize, xImageInfo.Width, uMips,
													 0, D3DFMT_UNKNOWN, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, NULL, NULL, &pxTexture )))
			{
				m_pxD3DTexture = pxTexture;
			}
			else
			{
				m_pxD3DTexture = NULL;
			}

		}
		else
		{
			Direct3D_Texture* pxTexture = NULL;

			if (SUCCEEDED(D3DXCreateTextureFromFileInMemoryEx( Asura_Xbox360_Graphics::GetDevice(), pData, iDataSize, xImageInfo.Width, xImageInfo.Height, uMips,
												 0, D3DFMT_UNKNOWN, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT, 0, NULL, NULL, &pxTexture )))

			{
				m_pxD3DTexture = pxTexture;
			}
			else
			{
				m_pxD3DTexture = NULL;
			}
		}		
	}

	Asura_Assertf(m_pxD3DTexture, ("Failed to create texture %s.", szName));

	if (Asura_Xbox360_TextureManagement::TextureBudgetEnabled())
	{	
		Asura_Xbox360_TextureManagement::DecrementTextureBudget(iDataSize);	
	}

	if( bSRGB && m_pxD3DTexture && m_pxD3DTexture->Format.DataFormat != GPUTEXTUREFORMAT_16 && m_pxD3DTexture->Format.DataFormat != GPUTEXTUREFORMAT_16_16)
	{
		m_pxD3DTexture->Format.SignX = GPUSIGN_GAMMA;
		m_pxD3DTexture->Format.SignY = GPUSIGN_GAMMA;
		m_pxD3DTexture->Format.SignZ = GPUSIGN_GAMMA;

		// Expand to 16-bit when sampling, otherwise we introduce artefacts. See "Picture Perfect"
		// slide 32 (Gamefest 2006). This runs at half speed, but hopefully we won't be bound there

#define EXPAND_FORMAT(x)                                                           \
		case GPUTEXTUREFORMAT_##x:                                                     \
		{                                                                              \
		m_pxD3DTexture->Format.DataFormat = GPUTEXTUREFORMAT_##x##_AS_16_16_16_16; \
		break;                                                                     \
		}

		switch( m_pxD3DTexture->Format.DataFormat )
		{
			EXPAND_FORMAT( 8_8_8_8 );
			EXPAND_FORMAT( DXT1 );
			EXPAND_FORMAT( DXT2_3 );
			EXPAND_FORMAT( DXT4_5 );

		default:
			{
				// Can't expand other formats, but the artists occasionally use things like L8
				// format for PFX (by saving a .dds directly), so don't assert. For now, pump out
				// some debug info

				Asura_OutputToDebugger_Retail(( "Unable to use 16-bit expansion on texture '%s'.\n",
					szName ));

				break;
			}
		}

#undef EXPAND_FORMAT
	}

	m_szName = Asura_StringUtil::Duplicate( szName );
	m_pxXPR = NULL;
}



void Asura_Xbox360_Texture::InitFromXPR( const char* const szName, Direct3D_Texture* const pxD3DTexture, void* const pData, Asura_Xbox360_XPR& xXPR )
{
	Asura_Assertf( !m_pxD3DTexture, ("Trying to create texture from XPR over existing data (%s).", m_szName ? m_szName : "") );

	m_pxD3DTexture = pxD3DTexture;
	m_szName = szName;

	XGOffsetBaseTextureAddress( m_pxD3DTexture, pData, pData );

	m_pxXPR = &xXPR;
	xXPR.AddRef();
}

void Asura_Xbox360_Texture::SetupPlaceholderTexture( const D3DCOLOR uColour )
{
	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	if( !pxDevice )
	{
		Asura_Assert( false, "Trying to create placeholder texture before D3D is initialised." );
		return;
	}

	Asura_Assertf( !m_pxD3DTexture, ("Trying to create placeholder texture over existing data (%s).", m_szName ? m_szName : "") );

	Direct3D_Texture* pxTexture = NULL;
	const HRESULT xResult = pxDevice->CreateTexture( 1, 1, 1, 0, D3DFMT_A8R8G8B8, 0, &pxTexture, NULL );
	m_pxD3DTexture = pxTexture;

	if( FAILED( xResult ) )
	{
		Asura_Assert( false, "Failed to create placeholder texture." );
		return;
	}

	D3DLOCKED_RECT xLockedRect = {0};
	pxTexture->LockRect( 0, &xLockedRect, NULL, 0 );

	if( !xLockedRect.pBits )
	{
		Asura_Assert( false, "Failed to lock placeholder texture." );
		Destroy();

		return;
	}

	*reinterpret_cast< D3DCOLOR* >( xLockedRect.pBits ) = uColour;
	pxTexture->UnlockRect(0);

	m_szName = NULL;
	m_pxXPR = NULL;
}



void Asura_Xbox360_TextureManagement::CreatePlaceholderTexture( const char* const szName)
{
	// Get the independent ID
	const int iTexID = Asura_Texture_Management::AddTexture( szName );

	// Allocate an Xbox 360 texture if required
	if( !g_apxTextureStorage[ iTexID ] )
	{
		g_apxTextureStorage[ iTexID ] = new Asura_Xbox360_Texture;
		if( !g_apxTextureStorage[ iTexID ] )
		{
			Asura_Assert( false, "Unable to allocate memory for texture." );
			return;
		}
	}

	g_apxTextureStorage[iTexID]->SetupPlaceholderTexture( 0 );
}

#if defined (ASURA_USE_TEXTURE_STREAMING)
#include "Core/Asura_Texture_Streaming.h"

static Asura_Handle g_iStreamingFileHandle = ASURA_INVALID_HANDLE;

void Asura_Texture_Streaming_System::Platform_Initialise()
{
}

void Asura_Texture_Streaming_System::Platform_Shutdown()
{
}

bool Asura_Texture_Streaming_System::Platform_StartStreaming( void )
{
	Asura_Assert( g_iStreamingFileHandle == ASURA_INVALID_HANDLE, "Code bug: is streaming file still open?" );

	g_iStreamingFileHandle = Asura_FileAccess::Open( Asura_Texture_Streaming_System::GetTextureStreamingFilename(), false, true );

	// return true if open successful.
	return ( g_iStreamingFileHandle != ASURA_INVALID_HANDLE );
}

void Asura_Texture_Streaming_System::Platform_EndStreaming( void )
{
	if ( g_iStreamingFileHandle != ASURA_INVALID_HANDLE )
	{
		Asura_FileAccess::Close( g_iStreamingFileHandle );
	}

	g_iStreamingFileHandle = ASURA_INVALID_HANDLE;
}

void Asura_Texture_Streaming_System::Platform_StreamTexture( Asura_Streaming_Texture& xTexture )
{
	const int iTextureID = xTexture.GetTextureIndex();

	if( g_iStreamingFileHandle == ASURA_INVALID_HANDLE )
	{
		return;
	}

	int iFileSize = xTexture.GetFileSize();

	if (!Asura_FileAccess::SeekFile( g_iStreamingFileHandle, xTexture.GetPositionInFile() ))
	{
		Asura_Assert( false, "Failed to read texture data from resource." );
		return;
	}

	if (xTexture.GetChunkFlags() & Asura_Chunk_ResourceFile::TF_XBOX360_TEXTURE_XPR)
	{
		Asura_Chunk_File_Stream xStream(g_iStreamingFileHandle);
		Asura_Xbox360_XPR_System::ParseXPR( xStream, true );
	} else {
		// worry here about memory fragmentation? we can easily get the largest file size
		void * const pBuffer = Asura_Allocate( iFileSize );

		if( !pBuffer )
		{
			Asura_Assert( false, "Failed to allocate memory for texture." );
			return;
		}

		if (!Asura_FileAccess::ReadData( g_iStreamingFileHandle, pBuffer, iFileSize))
		{
			Asura_Assert( false, "Failed to read texture data from resource." );
			return;
		}
		Asura_Xbox360_TextureManagement::CreateTextureFromFileInMemory( iTextureID, pBuffer, iFileSize, true, true );
		Asura_Deallocate( pBuffer );
	}
	xTexture.SetStatus( Asura_Streaming_Texture::AST_LOADED );
}

void Asura_Texture_Streaming_System::Platform_UnloadTexture( u_int uTextureID )
{
	if( g_apxStreamingTextureStorage[ uTextureID ] )
	{
		delete g_apxStreamingTextureStorage[ uTextureID ];
		g_apxStreamingTextureStorage[ uTextureID ] = NULL;
	}

	// remove the texture from memory

}

#endif

