//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR_System.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Rendering/Asura_Xbox360_XPR_System.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"

// Misc
#include <xtl.h>

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct Asura_Xbox360_XPR_Header
{
	DWORD m_uMagicValue;
	DWORD m_uHeaderSize;
	DWORD m_uDataSize;
};

struct Asura_Xbox360_XPR_Resource
{
	DWORD m_uType;
	DWORD m_uOffset;
	DWORD m_uSize;
	DWORD m_uNameOffset;
};

enum Asura_Xbox360_XPR_ResourceType
{
	// NB: FourCC codes MS compiler extension

	RESOURCETYPE_USERDATA      = 'USER',
	RESOURCETYPE_TEXTURE       = 'TX2D',
	RESOURCETYPE_CUBEMAP       = 'TXCM',
	RESOURCETYPE_VOLUMETEXTURE = 'TX3D',
	RESOURCETYPE_VERTEXBUFFER  = 'VBUF',
	RESOURCETYPE_INDEXBUFFER   = 'IBUF',
	RESOURCETYPE_VERTEXSHADER  = 'VSHD',
	RESOURCETYPE_PIXELSHADER   = 'PSHD',
	RESOURCETYPE_EOF           = 0xFFFFFFFF,
};

#define IS_TEXTURE(x) (((x) & 0xFFFF0000) == (RESOURCETYPE_TEXTURE & 0xFFFF0000))

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const DWORD uXPR_ALLOC_ATTRIBUTES = MAKE_XALLOC_ATTRIBUTES
(
	0,								// Object type
	FALSE,							// Heap tracks attributes
	TRUE,							// Must succeed
	FALSE,							// Fixed size
	0,								// Allocator ID
	XALLOC_PHYSICAL_ALIGNMENT_4K,	// Alignment
	XALLOC_MEMPROTECT_WRITECOMBINE,	// Memory protection
	FALSE,							// Zero-initialise
	XALLOC_MEMTYPE_PHYSICAL			// Memory type
);

/******************************************************************************/
/* Asura_Xbox360_XPR_System Static Class Variables                            */
/******************************************************************************/

/******************************************************************************/
/* Asura_Xbox360_XPR Class Functions                                          */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR::Asura_Xbox360_XPR()
//
// Constructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_XPR::Asura_Xbox360_XPR()
	: m_pcHeader( NULL )
	, m_pData( NULL )
	, m_uRefCount(0)
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR::Create()
//
// Stores pointers to an XPR bundle
//
// Parameters:		pcHeader - Pointer to XPR header data
//					pData - Pointer to main XPR data
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_XPR::Create( char* const pcHeader, void* const pData )
{
	m_pcHeader = pcHeader;
	m_pData = pData;
	m_uRefCount = 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR::Destroy()
//
// Frees the memory stored in an XPR struct
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_XPR::Destroy()
{
	Asura_Assert( m_uRefCount == 0, "XPR reference count not 0 on destruction." );

	delete[] m_pcHeader;
	XMemFree( m_pData, uXPR_ALLOC_ATTRIBUTES );

	m_pcHeader = NULL;
	m_pData = NULL;
	m_uRefCount = 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR::AddRef()
//
// Increments the reference count on an XPR, returning the new value
//
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Xbox360_XPR::AddRef()
{
	++m_uRefCount;
	return m_uRefCount;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR::Release()
//
// Decrements the reference count on an XPR, returning the new value. If this is
// 0, the XPR is destroyed and its memory is deallocated
//
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Xbox360_XPR::Release()
{
	Asura_Assert( m_uRefCount > 0, "XPR reference count already 0." );
	--m_uRefCount;

	if( m_uRefCount == 0 )
	{
		Destroy();
	}

	return m_uRefCount;
}

/******************************************************************************/
/* Asura_Xbox360_XPR_System Static Class Functions                            */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR_System::Initialise()
//
// Initialises the XPR system
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_XPR_System::Initialise()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR_System::Reset()
//
// Resets the XPR system. Called between level loads
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_XPR_System::Reset()
{
	// We don't actually clean out the XPRs at this point, because textures they contain might be in
	// the 'permanent' reset group. Instead they're reference counted and are automatically deleted
	// when the last resource is unloaded
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR_System::Shutdown()
//
// Shuts the XPR system down
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_XPR_System::Shutdown()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR_System::ParseXPR()
//
// Loads resources from an XPR contained in an Asura chunk stream
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_XPR_System::ParseXPR( Asura_Chunk_Stream& xStream, bool bStreamedTexture /*= false*/ )
{
	// Read the XPR header

	Asura_Xbox360_XPR_Header xXPRHeader;
	xStream.Read( sizeof( Asura_Xbox360_XPR_Header ), &xXPRHeader );

	if( xXPRHeader.m_uMagicValue != 'XPR2' )
	{
		Asura_Assert( false, "XPR header doesn't have correct magic number." );
		return;
	}

	// Read the resource header

	char* const pcHeader = new char[ xXPRHeader.m_uHeaderSize ];
	if( !pcHeader )
	{
		Asura_Assert( false, "Unable to allocate memory for XPR header." );
		return;
	}

	xStream.Read( xXPRHeader.m_uHeaderSize, pcHeader );

	const u_int uNumResources = *reinterpret_cast< const u_int* >( pcHeader );
	if( uNumResources == 0 )
	{
		Asura_Assert( false, "Found an empty XPR chunk." );
		delete[] pcHeader;

		return;
	}

	Asura_Xbox360_XPR_Resource* const pxResources =
		reinterpret_cast< Asura_Xbox360_XPR_Resource* >( pcHeader + 4 );

	// Need to make sure that even we're over the texture budget, we still load shaders. In practice
	// we won't mix them in the same bundle

	bool bUseTextureBudget = Asura_Xbox360_TextureManagement::TextureBudgetEnabled();

	for( u_int u = 0; u < uNumResources; ++u )
	{
		if( !IS_TEXTURE( pxResources[u].m_uType ) )
		{
			bUseTextureBudget = false;
		}
	}

	// Read the resource data
	if( !bUseTextureBudget
	 || (int)xXPRHeader.m_uDataSize < Asura_Xbox360_TextureManagement::GetTextureBudget() )
	{
		void* pData = NULL;

		if( xXPRHeader.m_uDataSize > 0 )
		{
			pData = XMemAlloc( xXPRHeader.m_uDataSize, uXPR_ALLOC_ATTRIBUTES );

			if( !pData )
			{
				Asura_Assert( false, "Unable to allocate memory for XPR data." );
				delete[] pcHeader;
				return;
			}

			xStream.Read( xXPRHeader.m_uDataSize, pData );
		}

		#if !defined ASURA_NO_DEBUG_OUTPUT
			MEMORYSTATUS xStartMemStatus;
			GlobalMemoryStatus( &xStartMemStatus );

			u_int uNumShaders = 0;
		#endif

		// Store the data for later

		Asura_Xbox360_XPR* pxXPR = FindFreeSlot();
		if (!pxXPR)
		{
			Asura_Assert( false, "Unable to allocate memory for XPR data." );
			delete[] pcHeader;
			XMemFree(pData, uXPR_ALLOC_ATTRIBUTES);
			return;
		}
		Asura_Xbox360_XPR& xXPR = *pxXPR;
		xXPR.Create( pcHeader, pData );
		xXPR.AddRef();

		for( u_int u = 0; u < uNumResources; ++u )
		{
			const char* const szName = pcHeader + pxResources[u].m_uNameOffset;
			char* const pcResource = pcHeader + pxResources[u].m_uOffset;

			if( IS_TEXTURE( pxResources[u].m_uType ) )
			{
				Direct3D_Texture* pxTextureHeader =
					reinterpret_cast< Direct3D_Texture* >( pcResource );

				Asura_Xbox360_TextureManagement::CreateTextureFromXPR( szName, pxTextureHeader,
																	   pData, xXPR, bStreamedTexture );
			}
			else if( pxResources[u].m_uType == RESOURCETYPE_PIXELSHADER )
			{
				Asura_Xbox360_MaterialResponse_System::AddPixelShader( szName, pcResource, xXPR );

				#if !defined ASURA_NO_DEBUG_OUTPUT
					++uNumShaders;
				#endif
			}
		}
		xXPR.Release();

		#if !defined ASURA_NO_DEBUG_OUTPUT
			if( uNumShaders > 0 )
			{
				MEMORYSTATUS xEndMemStatus;
				GlobalMemoryStatus( &xEndMemStatus );

				Asura_OutputToDebugger_Retail(( "Loaded %u shaders from an XPR chunk. Memory used = %u bytes (possibly including textures.)\n",
					uNumShaders, xStartMemStatus.dwAvailPhys - xEndMemStatus.dwAvailPhys ));
			}
		#endif
	}
	else
	{
		xStream.Skip( xXPRHeader.m_uDataSize);

		for( u_int u = 0; u < uNumResources; ++u )
		{
			if( IS_TEXTURE( pxResources[u].m_uType ) )
			{
				const char* const szName = pcHeader + pxResources[u].m_uNameOffset;
				Asura_Xbox360_TextureManagement::CreatePlaceholderTexture( szName );
			}
		}

		// MH: fix memory leak with the placeholder textures not freeing the header later.
		delete[] pcHeader;
	}

	if( bUseTextureBudget )
	{	
		Asura_Xbox360_TextureManagement::DecrementTextureBudget( xXPRHeader.m_uDataSize );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_XPR_System::FindFreeSlot()
//
// Finds a free XPR slot
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_XPR* Asura_Xbox360_XPR_System::FindFreeSlot()
{
	return new Asura_Xbox360_XPR;
}
