/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float4x4 g_xProjectionMatrix;
const float4x4 g_xInvViewMatrix;
const float4x4 g_xLightOrientation;

const float4 g_xTwoOverScreen;			// dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float4 g_xCameraTanHalfFOV;		// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.

// Directional light parameters
const float3 g_xLightDirection;

// Omni light parameters
const float4 g_xFalloff;				// x = scale, y = power
const float4 g_xLimitAngles;			// x = -cos(outer-limit), y = 1 / (-cos(outer-limit) + cos(inner-limit)), z = outer-limit
const float4 g_xLightPosScale;			// World-space light pos in xyz, 1/range in w
const float4 g_xLightColour;			// Colour in rgb, intensity in a
const float4 g_xMeshScaleAndPos;		// Position/scale for the light volume

const float  g_fProjCubeLerpDist;

#define DEFINE_SAMPLER( TextureName, Register )											\
	const texture g_x##TextureName;														\
	const sampler2D g_x##TextureName##Sampler : register( Register ) = sampler_state	\
	{																					\
		Texture = <g_x##TextureName>;													\
		AddressU = CLAMP;																\
		AddressV = CLAMP;																\
	};

// On 360, we specify filtering state in the instruction

DEFINE_SAMPLER( NormalBuffer, s5 );
DEFINE_SAMPLER( AlbedoBuffer, s6 );
DEFINE_SAMPLER( DepthMtlBuffer, s7 );

const texture g_xProjCubeMap;
const samplerCUBE g_xProjCubeMapSampler = sampler_state
{
	Texture = <g_xProjCubeMap>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};

const texture g_xProjCubeMapBlurred;
const samplerCUBE g_xProjCubeMapBlurredSampler = sampler_state
{
	Texture = <g_xProjCubeMapBlurred>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};


#undef DEFINE_SAMPLER

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct VS_Output
{
	float4 xPosition  : POSITION;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// -----------------------------------------------------------------------------
// Vertex Shader (Sphere)
// The vertex shader recreates the light frustum in world space, then projects
// it as an object in the standard camera view
// -----------------------------------------------------------------------------

VS_Output VShader_Sphere( const float4 xInPosition : POSITION )
{
	VS_Output xOut;

	float4 xPos = xInPosition;

	xPos.xyz *= g_xMeshScaleAndPos.www;
	xPos.xyz += g_xMeshScaleAndPos.xyz;
	
	xOut.xPosition = mul( xPos, g_xViewProjectionMatrix );	

	return xOut;
}


// -----------------------------------------------------------------------------
// Vertex Shader (indexed Sphere)
// The vertex shader recreates the light frustum in world space, then projects
// it as an object in the standard camera view
// -----------------------------------------------------------------------------

VS_Output VShader_IndexedSphere( const float2 xInData : POSITION )
{
	VS_Output xOut;

	const float	fRadialAngle	= xInData.x;
	const float	fVerticalAngle	= xInData.y;
	const float	fRadialWidth	= sin(fVerticalAngle);
	const float	fRadialHeight	= cos(fVerticalAngle);
	
	float4 xPos;
	
	xPos.x	= fRadialWidth * cos(fRadialAngle);
	xPos.y	= fRadialWidth * sin(fRadialAngle);
	xPos.z	= fRadialHeight;
	xPos.w	= 1.0f;

	xPos.xyz *= g_xMeshScaleAndPos.www;
	xPos.xyz += g_xMeshScaleAndPos.xyz;
	
	xOut.xPosition = mul( xPos, g_xViewProjectionMatrix );	

	return xOut;
}


// -----------------------------------------------------------------------------
// Vertex Shader (indexed Sphere-segment)
// The vertex shader recreates the light frustum in world space, then projects
// it as an object in the standard camera view
// -----------------------------------------------------------------------------

VS_Output VShader_IndexedSphereSegment( const float3 xInData : POSITION )
{
	VS_Output xOut;

	const float	fRadialAngle	= xInData.x;
	const float	fVerticalAngle	= xInData.y * g_xLimitAngles.z;
	const float	fRadialWidth	= sin(fVerticalAngle) * xInData.z;
	const float	fRadialHeight	= cos(fVerticalAngle);
	
	float4 xPos;
	
	xPos.x	= fRadialWidth * cos(fRadialAngle);
	xPos.y	= fRadialWidth * sin(fRadialAngle);
	xPos.z	= min(xInData.z, fRadialHeight);
	xPos.w	= 1.0f;

	xPos.xyz *= g_xMeshScaleAndPos.www;
	xPos.xyz = mul(float4(xPos.xyz, 0), g_xLightOrientation);
	xPos.xyz += g_xMeshScaleAndPos.xyz;
	
	xOut.xPosition = mul( xPos, g_xViewProjectionMatrix );	

	return xOut;
}


// -----------------------------------------------------------------------------
// Pixel Shader (ClearDestAlpha)
// Just output 0
// -----------------------------------------------------------------------------

float4 PShader_ClearDestAlpha( void ) : COLOR
{
	return float4(0,0,0,0);
}


// -----------------------------------------------------------------------------
// Pixel Shader (Photon)
// Material response lighting from a point light - with angular constraints
// and cubemap projection
// -----------------------------------------------------------------------------

float4 PShader_LimitedPhoton( const VS_Output xIn, float2 xScreenPos : VPOS, uniform bool bUseAngularRange, uniform bool bUseProjCube, uniform bool bUseDualProjCube ) : COLOR
{
	float4 xBufferNormal;
	float4 xBufferDepthMtl;
	half4  xBufferAlbedo;

	xScreenPos += g_xVPOSOffsetVector;

	asm
	{
		tfetch2D xBufferNormal,   xScreenPos, g_xNormalBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xBufferDepthMtl, xScreenPos, g_xDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};

	half3 xNormal;
	float fScreenDepth;
	half3 xAlbedoColour;
	half  hSpecularStrength;
	float fMaterial;

	Deferred_GetNormalDepthMaterial( xBufferNormal, float4( 0, 0, 0, 0 ), xBufferDepthMtl, xNormal, fScreenDepth, fMaterial );	

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.x * fScreenDepth *  (xScreenPos.x * g_xTwoOverScreen.x - 1.0),
													g_xCameraTanHalfFOV.y * fScreenDepth * -(xScreenPos.y * g_xTwoOverScreen.y - 1.0),
													fScreenDepth,
													1.0 );

	float4 xOutput;

	float4 xWorldPosition	= mul( xCameraViewSpacePosition, g_xInvViewMatrix );
	float3 xLightDelta		= g_xLightPosScale.xyz - xWorldPosition.xyz;
	half3  xLightStrength	= xLightDelta.xyz * g_xLightPosScale.w;
	half   fLightDistance	= sqrt( dot( xLightStrength.xyz, xLightStrength.xyz ) );
	float3 xCameraDir		= normalize(g_xCameraPosition.xyz - xWorldPosition.xyz);

	// Early-out if not in range of the light
	[branch]
	if( 0.5 - abs( fLightDistance - 0.5 ) >= 0 )
	{
		// Get the albedo+specular
		asm
		{
			tfetch2D xBufferAlbedo, xScreenPos, g_xAlbedoBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		};

		Deferred_GetAlbedoSpecular( xBufferNormal, xBufferAlbedo, xBufferDepthMtl, xAlbedoColour, hSpecularStrength);

		{
			// Full material response (slow)
			half  fLightAttenuation = min( 1, pow( (1.0 - fLightDistance) * g_xFalloff.x, g_xFalloff.y ) );
			
			float3 xOfs = normalize(g_xLightPosScale.xyz-xWorldPosition);
			
			if (bUseAngularRange)
			{
				// Handle the angular falloff		
				float  fDot = dot(xOfs, g_xLightDirection);		
				fLightAttenuation *= saturate((fDot + g_xLimitAngles.x) * g_xLimitAngles.y);
			}
			
			half3 xLightAttenuatedColour = g_xLightColour.rgb * fLightAttenuation;

			if (bUseProjCube)
			{
				float3 xCol;
				
				// Cubemap projector
				float3 xDir = mul(xOfs, g_xLightOrientation);
				xCol = TexFetchCube_LinearLinearPoint( g_xProjCubeMapSampler, xDir);		
				
				if (bUseDualProjCube)
				{			
					float fLerp = min(fLightDistance/g_fProjCubeLerpDist,1.0f);
					xCol = lerp(xCol, TexFetchCube_LinearLinearPoint( g_xProjCubeMapBlurredSampler, xDir), fLerp);
				}
				
				xLightAttenuatedColour *= xCol;
			}			

			xOutput.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
				half4( xNormal, 1 ), xCameraDir, normalize( xLightDelta ), xLightAttenuatedColour,
				xLightAttenuatedColour, dot( xLightAttenuatedColour, float3( 0.333, 0.333, 0.333 ) ) );

			xOutput.a = 1;
		}
	}
	else
	{
		xOutput = float4( 0, 0, 0, 0 );
	}

	return xOutput;
}


// -----------------------------------------------------------------------------
// Pixel Shader (Infinite)
// Material response lighting from a parallel light
// -----------------------------------------------------------------------------

float4 PShader_Infinite( const VS_Output xIn, float2 xScreenPos : VPOS ) : COLOR
{
	float4 xBufferNormal;
	float4 xBufferDepthMtl;
	half4  xBufferAlbedo;

	xScreenPos += g_xVPOSOffsetVector;

	asm
	{
		tfetch2D xBufferNormal,   xScreenPos, g_xNormalBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xBufferDepthMtl, xScreenPos, g_xDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xBufferAlbedo,   xScreenPos, g_xAlbedoBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};

	half3 xNormal;
	float fScreenDepth;
	half3 xAlbedoColour;
	half  hSpecularStrength;
	float fMaterial;

	Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xBufferDepthMtl, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial );

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	const float4 xCameraViewSpacePosition = float4(	g_xCameraTanHalfFOV.x * fScreenDepth * (xScreenPos.x * g_xTwoOverScreen.x  - 1.0),
													g_xCameraTanHalfFOV.y  * fScreenDepth * -(xScreenPos.y * g_xTwoOverScreen.y - 1.0),
													fScreenDepth,
	 	  											1.0);
		   														 
	float4 xOutput;

	float4 xWorldPosition   = mul(  xCameraViewSpacePosition, g_xInvViewMatrix);			
	float3 xLightDelta		= g_xLightPosScale.xyz - xWorldPosition.xyz;
	half3  xLightStrength	= xLightDelta.xyz * (g_xLightPosScale.w);	
	half   fLightDistance	= sqrt( dot( xLightStrength.xyz, xLightStrength.xyz ))	;					   
	float3 xCameraDir		= normalize(g_xCameraPosition.xyz - xWorldPosition.xyz);
	
	clip( 0.5 - abs( fLightDistance - 0.5));
	
	// Handle the falloff
	half  fLightAttenuation = min(1,pow(((1.0f - fLightDistance) * g_xFalloff.x), g_xFalloff.y));			
	
	half3 xLightAttenuatedColour = g_xLightColour * fLightAttenuation;
	xOutput.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
		half4( xNormal, 1), xCameraDir, g_xLightDirection, xLightAttenuatedColour,
		xLightAttenuatedColour, dot(xLightAttenuatedColour, float3(0.333,0.333,0.333) ) );

	xOutput.a = 1;		 	

	return xOutput;
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	// This is now implemented as several 'passes', instead of one ubershader with loads of static
	// branching. We only ever render 1-2 of these passes per photon

	// Pass Type 0 - Clear dest. alpha (clear shadow mask)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_ClearDestAlpha_Sphere
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_ClearDestAlpha();
	}

	// Pass Type 1 - Clear dest. alpha (clear shadow mask)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_ClearDestAlpha_IndexedSphere
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphere();
		PixelShader  = compile ps_3_0 PShader_ClearDestAlpha();
	}

	// Pass Type 2 - Clear dest. alpha (clear shadow mask)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_ClearDestAlpha_IndexedSphereSegment
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphereSegment();
		PixelShader  = compile ps_3_0 PShader_ClearDestAlpha();
	}


	// Pass Type 3 - Render as an infinite light
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_Infinite
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_Infinite();
	}
	
	
	// Pass Type 4 - Render as a sphere photon (point light)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_Photon
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(false, false, false);
	}
	
	// Pass Type 5 - Render as a sphere photon (point light) with cube-map
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_Photon_CubeMap
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(false, true, false);
	}
	
	// Pass Type 6 - Render as an sphere limited-photon (point light) with dual cube-maps
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_Photon_DualCubeMaps
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(false, true, true);
	}
	
	
	// Pass Type 7 - Render as an indexed-sphere photon (point light)
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_IndexedPhoton
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(false, false, false);
	}
	
	// Pass Type 8 - Render as an indexed-sphere photon (point light) with cube-map
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_IndexedPhoton_CubeMap
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(false, true, false);
	}
	
	// Pass Type 9 - Render as a an indexed-sphere photon (point light) with dual cube-maps
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_IndexedPhoton_DualCubeMaps
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(false, true, true);
	}
	
	
	// Pass Type 10 - Render as a limited-photon (point light) with limited angle
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_LimitedPhoton
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(true, false, false);
	}
	
	// Pass Type 11 - Render as a limited-photon (point light) with limited angle and cube-map
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_LimitedPhoton_CubeMap
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(true, true, false);
	}
	
	// Pass Type 12 - Render as a limited-photon (point light) with limited angle and dual cube-maps
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_LimitedPhoton_DualCubeMaps
	{
		VertexShader = compile vs_3_0 VShader_Sphere();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(true, true, true);
	}
	
	
	// Pass Type 13 - Render as a limited-photon (point light) with limited angle
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_OptimisedLimitedPhoton
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphereSegment();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(true, false, false);
	}
	
	// Pass Type 14 - Render as a limited-photon (point light) with limited angle and cube-map
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_OptimisedLimitedPhoton_CubeMap
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphereSegment();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(true, true, false);
	}
	
	// Pass Type 15 - Render as a limited-photon (point light) with limited angle and dual cube-maps
	// ----------------------------------------------------------------------------------------
	pass DeferredPhoton_OptimisedLimitedPhoton_DualCubeMaps
	{
		VertexShader = compile vs_3_0 VShader_IndexedSphereSegment();
		PixelShader  = compile ps_3_0 PShader_LimitedPhoton(true, true, true);
	}
}
