/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

const float4x4 g_xProjectionMatrix;
const float4x4 g_xInvViewMatrix;

const float4   g_xProjectorPosition;						// Position of the projector
const float4   g_xProjectionDirection;
const float4   g_xNearPlane;								// Fade and fallout controls
const float    g_fDistanceFade;
const float4   g_xFalloff;									
const float4   g_xNeutralColour = float4(0,0,0,0);			// Neutral colour for the current blend mode (0,0,0 for additive, 1,1,1 for multiply, etc)

const float4   g_xLightColour;								// Colour/intensity of the light

const bool     g_bUseDetailTexture;							// Detail texturing
const float4   g_xDetailTextureScaleAndOffset;

const float    g_fTwoOverScreenWidth;						// Dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float    g_fTwoOverScreenHeight;

const float    g_fCameraTanHalfHorizFOV;					// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
const float    g_fCameraTanHalfVertFOV;


const texture g_xNormalBuffer;
const sampler g_xNormalBufferSampler : register(s5) = sampler_state
{
	Texture = <g_xNormalBuffer>;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};

const texture g_xAlbedoBuffer;
const sampler g_xAlbedoBufferSampler : register(s6) = sampler_state
{
	Texture = <g_xAlbedoBuffer>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};

const texture g_xDepthMtlBuffer;
const sampler g_xDepthMtlBufferSampler : register(s7) = sampler_state
{
	Texture = <g_xDepthMtlBuffer>;
	AddressU = CLAMP;
	AddressV = CLAMP;
};


const texture g_xProjectedTexture;
// Mipmapping doesn't work with arbitrary dependent texture reads like we do here...
const sampler g_xProjectedTextureSampler : register(s1) = sampler_state
{
	Texture = <g_xProjectedTexture>;
	AddressU =  CLAMP;										// Can't use border mode - arbitrary border colours aren't not well supported
	AddressV =  CLAMP; 
};


const texture g_xDetailTexture;
const sampler g_xDetailTextureSampler : register(s3) = sampler_state
{
	Texture = <g_xDetailTexture>;
	AddressU =  WRAP;
	AddressV =  WRAP; 
};


/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/


/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// The vertex shader recreates the shadow frustum in world space, then projects
// it as an object in the standard camera view
struct VS_Input
{
	float4 m_xPosition				: POSITION0;
};

struct VS_Output
{
	float4 m_xPosition				: POSITION;
	float3 m_xCameraDirDistance		: TEXCOORD0;
};

struct VS_Output_PosOnly
{
	float4 m_xPosition				: POSITION;	
};

VS_Output VShader( const VS_Input xIn)
{
	float4 xPosition = xIn.m_xPosition;
	
	VS_Output xOut;	
	xOut.m_xPosition = mul( xPosition, g_xViewProjectionMatrix );	
	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xPosition;
	
	return( xOut);
}

VS_Output_PosOnly VShader_PosOnly( const VS_Input xIn)
{
	float4 xPosition = xIn.m_xPosition;
	
	VS_Output_PosOnly xOut;	
	xOut.m_xPosition = mul( xPosition, g_xViewProjectionMatrix );		
	
	return( xOut);
}


// -----------------------------------------------------------------------------
// Pixel Shader (Clear Dest Alpha)
// Just output 0 - for clearing the shadow mask
// -----------------------------------------------------------------------------

float4 PShader_ClearDestAlpha(void) : COLOR
{
	return float4(0,0,0,0);
}


// -----------------------------------------------------------------------------
// Pixel Shader (DynamicLight)
// Projected texture as a full dynamic light, with material response
// -----------------------------------------------------------------------------

float4 PShader_DynamicLight( float2 xScreenPos : VPOS,  const float3 xCameraDirDistance : TEXCOORD0) : COLOR
{
		// Fetch the world-space normal and screen depth, then reconstruct the camera view space position
		// ------------------------------------------------------------------------------------------------
		half3 xNormal, xAlbedoColour;
		float fScreenDepth, fMaterial;		
		half  hSpecularStrength;
		
		xScreenPos += g_xVPOSOffsetVector;

		const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * float2(g_fTwoOverScreenWidth, g_fTwoOverScreenHeight) / 2;	
		
		const float4 xBufferNormal     = TexFetch2D_PointPointPoint(g_xNormalBufferSampler, xNormalizedCoords);
		const float4 xBufferDepthMtl   = TexFetch2D_PointPointPoint(g_xDepthMtlBufferSampler, xNormalizedCoords);
		const half4  xBufferAlbedo     = TexFetch2D_PointPointPoint(g_xAlbedoBufferSampler, xNormalizedCoords);	
		
		Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xBufferDepthMtl, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);
		
		const float4 xCameraViewSpacePosition = float4(	g_fCameraTanHalfHorizFOV * fScreenDepth * (xScreenPos.x * g_fTwoOverScreenWidth  - 1.0),
														g_fCameraTanHalfVertFOV  * fScreenDepth * -(xScreenPos.y * g_fTwoOverScreenHeight - 1.0),
														fScreenDepth, 1.0);
							
		// Do the projection, and sample the projected texture
		// ------------------------------------------------------------------------------------------------							
		float4 xProjectedPosition = mul(  xCameraViewSpacePosition, g_xProjectionMatrix);	
		float3 xProjectedPosition2D = xProjectedPosition.xyz / xProjectedPosition.w;
		
		// DLH - Clip any XYZ value outside of 0-1
		clip( 0.5 - abs(xProjectedPosition2D.xyz - 0.5));
		
		float4 f4Tex =   TexFetch2D_LinearLinearNone(g_xProjectedTextureSampler, xProjectedPosition2D.xy);	
		
		// Handle detail texturing, if we're using it
		// ------------------------------------------------------------------------------------------------							
		if (g_bUseDetailTexture)
		{
			xProjectedPosition.xy = (xProjectedPosition.xy * g_xDetailTextureScaleAndOffset.z) + (g_xDetailTextureScaleAndOffset.xy*xProjectedPosition.w);
			float4 f4DetailTex =   TexFetch2DProj_LinearLinearNone(g_xDetailTextureSampler, xProjectedPosition);			
			f4Tex *= f4DetailTex;
		}


		// Handle the falloff
		float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);		
		clip (0.5f-abs(fDist-0.5f));		
		float  fFalloff = min(1,pow(((1.0f - fDist) * g_xFalloff.x), g_xFalloff.y));		
 		fFalloff = fFalloff *  g_fDistanceFade ;
		
		// Deferred dynamic light - material response
		// Diffuse lighting
		float4 xWorldPosition = mul(  xCameraViewSpacePosition, g_xInvViewMatrix);	
 
		float4 xLightDelta = g_xProjectorPosition - xWorldPosition;
					
		half3 xLightAttenuatedColour = g_xLightColour.rgb * fFalloff * f4Tex.rgb;
		
		fFalloff = dot((f4Tex.xyz*fFalloff), float3(0.333f,0.333f,0.333f));		
				
		f4Tex.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
			half4( xNormal, 1), normalize( xCameraDirDistance.xyz ), normalize( xLightDelta ),
			xLightAttenuatedColour, xLightAttenuatedColour, fFalloff );
							
		f4Tex.a = 1;				
		
		return f4Tex;
}


// -----------------------------------------------------------------------------
// Pixel Shader (FakeLight)
// Projected texture as a fake light (multiply scaled by N.L)
// -----------------------------------------------------------------------------

float4 PShader_FakeLight( float2 xScreenPos : VPOS) : COLOR
{

		// Fetch the world-space normal and screen depth, then reconstruct the camera view space position
		// ------------------------------------------------------------------------------------------------
		float4 xBufferNormal;
		float4 xBufferDepthMtl;
		half4  xBufferAlbedo;

		xScreenPos += g_xVPOSOffsetVector;

		asm
		{
			tfetch2D xBufferNormal,   xScreenPos, g_xNormalBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
			tfetch2D xBufferDepthMtl, xScreenPos, g_xDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
			tfetch2D xBufferAlbedo,   xScreenPos, g_xAlbedoBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled				
		};
		
		
		half3 xNormal, xAlbedoColour;
		float fScreenDepth, fMaterial;		
		half  hSpecularStrength;
		
		Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xBufferDepthMtl, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);
			
		const float4 xCameraViewSpacePosition = float4(	g_fCameraTanHalfHorizFOV * fScreenDepth * (xScreenPos.x * g_fTwoOverScreenWidth  - 1.0),
														g_fCameraTanHalfVertFOV  * fScreenDepth * -(xScreenPos.y * g_fTwoOverScreenHeight - 1.0),
														fScreenDepth, 1.0);
							
		// Do the projection, and sample the projected texture
		// ------------------------------------------------------------------------------------------------							
		float4 xProjectedPosition = mul(  xCameraViewSpacePosition, g_xProjectionMatrix);	
		float3 xProjectedPosition2D = xProjectedPosition.xyz / xProjectedPosition.w;
		
		// DLH - Clip any XYZ value outside of 0-1
		clip( 0.5 - abs(xProjectedPosition2D.xyz - 0.5));
		
		float4 f4Tex =   TexFetch2D_LinearLinearNone(g_xProjectedTextureSampler, xProjectedPosition2D.xy);	
		
		// Handle detail texturing, if we're using it
		// ------------------------------------------------------------------------------------------------							
		if (g_bUseDetailTexture)
		{
			xProjectedPosition.xy = (xProjectedPosition.xy * g_xDetailTextureScaleAndOffset.z) + (g_xDetailTextureScaleAndOffset.xy*xProjectedPosition.w);
			float4 f4DetailTex =   TexFetch2DProj_LinearLinearNone(g_xDetailTextureSampler, xProjectedPosition);			
			f4Tex = lerp(g_xNeutralColour, f4DetailTex,  f4Tex.r);
		}

		// Handle the falloff
		// ------------------------------------------------------------------------------------------------							
		float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);		
		clip (0.5f-abs(fDist-0.5f));		
		float  fFalloff = min(1,pow(((1.0f - fDist) * g_xFalloff.x), g_xFalloff.y));		
 		fFalloff = fFalloff *  g_fDistanceFade ;
		
		// 'Fake Light' mode
		half3 xColor = xAlbedoColour;
			
		float4 xWorldPosition = mul(  xCameraViewSpacePosition, g_xInvViewMatrix);	
		float4 xL = normalize(g_xProjectorPosition-xWorldPosition);
			
		// Scale the falloff by N.L
		float fDot =  dot(xL.xyz,xNormal.xyz);
						
		f4Tex.rgb = (8.0f * xAlbedoColour * f4Tex *  fFalloff * fDot);
		f4Tex.a = 1;
	
		return f4Tex;
}


// -----------------------------------------------------------------------------
// Pixel Shader (StandardProjector)
// Just project the texture
// -----------------------------------------------------------------------------

float4 PShader_StandardProjector( float2 xScreenPos : VPOS) : COLOR
{
		// Fetch the world-space normal and screen depth, then reconstruct the camera view space position
		// ------------------------------------------------------------------------------------------------				
		float4 xBufferNormal;
		float4 xBufferDepthMtl;
		half4  xBufferAlbedo;

		xScreenPos += g_xVPOSOffsetVector;


		asm
		{
			tfetch2D xBufferNormal,   xScreenPos, g_xNormalBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
			tfetch2D xBufferDepthMtl, xScreenPos, g_xDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
			tfetch2D xBufferAlbedo,   xScreenPos, g_xAlbedoBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled				
		};		
		
		
		half3 xNormal, xAlbedoColour;
		float fScreenDepth, fMaterial;		
		half  hSpecularStrength;
		Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xBufferDepthMtl, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);
	
		const float4 xCameraViewSpacePosition = float4(	g_fCameraTanHalfHorizFOV * fScreenDepth * (xScreenPos.x * g_fTwoOverScreenWidth  - 1.0),
														g_fCameraTanHalfVertFOV  * fScreenDepth * -(xScreenPos.y * g_fTwoOverScreenHeight - 1.0),
														fScreenDepth, 1.0);
														
		
		// Do the projection, and sample the projected texture
		// ------------------------------------------------------------------------------------------------
		float4 xProjectedPosition = mul(  xCameraViewSpacePosition, g_xProjectionMatrix);	
		float3 xProjectedPosition2D = xProjectedPosition.xyz / xProjectedPosition.w;
		
		// DLH - Clip any XYZ value outside of 0-1
		clip( 0.5 - abs(xProjectedPosition2D.xyz - 0.5));		
		
		float4 f4Tex =   TexFetch2D_LinearLinearNone(g_xProjectedTextureSampler, xProjectedPosition2D.xy);;	
		
		// Handle detail texturing, if we're using it
		// ------------------------------------------------------------------------------------------------
		if (g_bUseDetailTexture)
		{
			xProjectedPosition.xy = (xProjectedPosition.xy * g_xDetailTextureScaleAndOffset.z) + (g_xDetailTextureScaleAndOffset.xy*xProjectedPosition.w);
			float4 f4DetailTex =   TexFetch2DProj_LinearLinearNone(g_xDetailTextureSampler, xProjectedPosition);			
			f4Tex = lerp(g_xNeutralColour, f4DetailTex,  f4Tex.r);
		}


		// Handle the falloff
		// ------------------------------------------------------------------------------------------------
		float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);		
		clip (0.5f-abs(fDist-0.5f));		
		float  fFalloff = min(1,pow(((1.0f - fDist) * g_xFalloff.x), g_xFalloff.y));		
 		fFalloff = fFalloff *  g_fDistanceFade ;
		
		f4Tex = lerp(g_xNeutralColour, f4Tex,  fFalloff);
		
		return f4Tex;
}



// -----------------------------------------------------------------------------
// Pixel Shader (Spotlight)
// Dynamic light, with material response - no projected texture
// -----------------------------------------------------------------------------

float4 PShader_Spotlight( float2 xScreenPos : VPOS,  const float3 xCameraDirDistance : TEXCOORD0) : COLOR
{
		// Fetch the world-space normal and screen depth, then reconstruct the camera view space position
		// ------------------------------------------------------------------------------------------------
		float4 xBufferNormal;
		float4 xBufferDepthMtl;
		half4  xBufferAlbedo;

		xScreenPos += g_xVPOSOffsetVector;


		asm
		{
			tfetch2D xBufferNormal,   xScreenPos, g_xNormalBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
			tfetch2D xBufferDepthMtl, xScreenPos, g_xDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
			tfetch2D xBufferAlbedo,   xScreenPos, g_xAlbedoBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled				
		};
		
		
		half3 xNormal, xAlbedoColour;
		float fScreenDepth, fMaterial;		
		half  hSpecularStrength;			
		Deferred_GetNormalDepthAlbedoSpecularMaterial( xBufferNormal, xBufferAlbedo, xBufferDepthMtl, xNormal, fScreenDepth, xAlbedoColour, hSpecularStrength, fMaterial);
	
		const float4 xCameraViewSpacePosition = float4(	g_fCameraTanHalfHorizFOV * fScreenDepth * (xScreenPos.x * g_fTwoOverScreenWidth  - 1.0),
														g_fCameraTanHalfVertFOV  * fScreenDepth * -(xScreenPos.y * g_fTwoOverScreenHeight - 1.0),
														fScreenDepth, 1.0);
									

		// Handle the distance falloff
		float  fDist  = dot(g_xNearPlane, xCameraViewSpacePosition);		
		clip (0.5f-abs(fDist-0.5f));		
		float  fFalloff = min(1,pow(((1.0f - fDist) * g_xFalloff.x), g_xFalloff.y));		
 		fFalloff = fFalloff *  g_fDistanceFade ;
 		
 		// Calculate the angular falloff 		
 		float4 xWorldPosition = mul(  xCameraViewSpacePosition, g_xInvViewMatrix);	
 		 		
 		float fDot = dot( g_xProjectionDirection.xyz, normalize(xWorldPosition.xyz - g_xProjectorPosition.xyz)); 		 			 	 	 		
 		fFalloff*= saturate((fDot + g_xFalloff.z) * g_xFalloff.w);
 		
		
		// Deferred dynamic light - material response		 
		float4 xLightDelta = g_xProjectorPosition - xWorldPosition;
					
		half3 xLightAttenuatedColour = g_xLightColour.rgb * fFalloff;				
			
		float4 f4Tex;	
		f4Tex.rgb = MaterialResponse_Basic( fMaterial, xAlbedoColour, hSpecularStrength,
			half4( xNormal, 1), normalize( xCameraDirDistance.xyz ), normalize( xLightDelta ),
			xLightAttenuatedColour, xLightAttenuatedColour, fFalloff );
							
		f4Tex.a = 1;				
		
		return f4Tex;
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/


technique Projector
{
	// This is now implemented as several 'passes', instead of one ubershader with loads of static
	// branching. We only ever render 1-2 of these passes per projector.

	// Pass Type 0 - Clear dest. alpha (clear shadow mask)
	// ----------------------------------------------------------------------------------------
	pass Projector_ClearDestAlpha
	{
		VertexShader = compile vs_3_0 VShader_PosOnly();
		PixelShader  = compile ps_3_0 PShader_ClearDestAlpha();
	}
	
	// Pass Type 1 - Render projector as dynamic light
	// ----------------------------------------------------------------------------------------
	pass Projector_DynamicLight
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_DynamicLight();
	}
	
	// Pass Type 2 - Render projector as fake light
	// ----------------------------------------------------------------------------------------
	pass Projector_FakeLight
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_FakeLight();
	}
	
	// Pass Type 3 - Render projector, standard (no lighting calculation - just project it)
	// ----------------------------------------------------------------------------------------
	pass Projector_Standard
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_StandardProjector();
	}
	
	// Pass Type 4 - Render projector, spotlight mode
	// ----------------------------------------------------------------------------------------
	pass Projector_Spotlight
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader  = compile ps_3_0 PShader_Spotlight();
	}	
}
