/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#include "Botanicals_Common.fxh"

const float  g_fNumAngles;		// Num quads per imposter (different orientations)
const float  g_fFirstInstance;

const float3 g_xCameraRightVector;
const float3 g_xCameraUpVector;
const float3 g_xCameraForwardsVector;


/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex_Botanicals_Prelighting VShader( const int iIndex : INDEX )
{
	const float3 axVerts[4]=	{{ -0.5, -1.0  ,1.0 },			// z component is the bone weight
								{   0.5, -1.0  ,1.0 },
								{   0.5,  0.0  ,0.0 },
								{  -0.5,  0.0  ,0.0 }};
										
	const float4 axUSelect[4]=	{{1,0,0,0},{0,0,1,0},{0,0,1,0},{1,0,0,0}};
	const float4 axVSelect[4]=	{{0,1,0,0},{0,1,0,0},{0,0,0,1},{0,0,0,1}};
																													
	int iInstanceIndex   = (iIndex + 0.5) / (4.0 * g_fNumAngles);
	int iAngleIndex      = ((iIndex + 0.5) / 4.0) % g_fNumAngles;
	int iBillboardVertex = (iIndex + 0.5) % 4.0;
	
	iInstanceIndex += g_fFirstInstance;
	
	OutputVertex_Botanicals_Prelighting xOut;

	// Instance properties (from stream 2)
	float3 xInstanceOrientation0;
	float3 xInstanceOrientation1;
	float3 xInstanceOrientation2;
	float4 xInstancePositionAndAlpha;
	float4 xRedSH;
	float4 xGreenSH;
	float4 xBlueSH;
	float  fBoneWeight;
	float4 xBoneOffsetAndScale;
	
	// ImposterAngle properties (from stream 0)
	float4 xSize;
	float4 xUVSet;
	float4 xForwards;
	float4 xAngleFade;
	float4 xLocalPos;
	float4 xTangent;
	float4 xBinormal;
	
	asm
	{
		// Fetch imposterangle properties
		// ------------------------------------------
		vfetch xLocalPos	, iAngleIndex, position0
		vfetch xAngleFade	, iAngleIndex, position1
		vfetch xSize		, iAngleIndex, position2
		vfetch xUVSet		, iAngleIndex, texcoord0
		vfetch xForwards	, iAngleIndex, normal0	
		vfetch xTangent		, iAngleIndex, tangent0	
		vfetch xBinormal	, iAngleIndex, binormal0	
		
		
		// Grab the instance's world matrix
		vfetch xInstanceOrientation0.xyz, iInstanceIndex, texcoord1
		vfetch xInstanceOrientation1.xyz, iInstanceIndex, texcoord2
		vfetch xInstanceOrientation2.xyz, iInstanceIndex, texcoord3
		vfetch xInstancePositionAndAlpha, iInstanceIndex, texcoord4

		// Grab the spherical harmonic
		vfetch xRedSH,   iInstanceIndex, texcoord5
		vfetch xGreenSH, iInstanceIndex, texcoord6
		vfetch xBlueSH,  iInstanceIndex, texcoord7

		// Bone Position Offset
		vfetch xBoneOffsetAndScale, iInstanceIndex, texcoord8
	};
		
	float fInstanceAlpha	= xInstancePositionAndAlpha.w;
	float2 xUV = float2(0.5f,0.5f);	
	

	// Get world pos of imposter origin
	// ----------------------------------------
	const float3x4	xWorldMatrix	= float3x4(float4(xInstanceOrientation0.xyz * xBoneOffsetAndScale.w, xInstancePositionAndAlpha.x),
											   float4(xInstanceOrientation1.xyz * xBoneOffsetAndScale.w, xInstancePositionAndAlpha.y),
											   float4(xInstanceOrientation2.xyz * xBoneOffsetAndScale.w, xInstancePositionAndAlpha.z));
	float4 xWorldPosition;
	xWorldPosition.xyz = xLocalPos;
	xWorldPosition.w = 1;
	xWorldPosition.xyz = mul( xWorldMatrix, xWorldPosition);
	
	// Select billboard corner
	// ----------------------------------------
	float3 xVertexOffset;
	xVertexOffset.xy = (axVerts[iBillboardVertex].xy * xSize.xy * xBoneOffsetAndScale.ww);	
	xVertexOffset.z = 0;
	
	// Facing-the-camera normal
	// ----------------------------------------
	float3 xNormal	= normalize(g_xCameraPosition.xyz - xWorldPosition.xyz); 
	
	
	// Handle angle fade and culling 
	// ----------------------------------------
	float2 x2DNormal = normalize(xNormal.xz);
	float3 xFacing   = mul(xWorldMatrix,xForwards.xyz);
	float2 x2DFacing = normalize(xFacing.xz);
	float fDot= dot(x2DNormal,x2DFacing);			
	float fAng= acos(fDot);
	float fMax= (xAngleFade.x*0.51f);			// Added 0.01f to work around a precision issue that
	float fFadeStart=fMax - xAngleFade.y;		// cause some 360-degree imposters to flicker.

	if (fAng>fMax)
	{		
		// Cull it
		xVertexOffset.xy = 0;				
	}
	else
	{
		if (fAng>fFadeStart)
		{						
			float fImposterFade= (fAng-fFadeStart)/xAngleFade.y;						
			fInstanceAlpha += (0.5f*fImposterFade);
		}
	}
 
	// Orient the billboard
	// ----------------------------------------
	float3 xBillboardUp    = normalize(float3( xInstanceOrientation0.y,xInstanceOrientation1.y,xInstanceOrientation2.y));
	float3 xBillboardFwd   = float3(x2DNormal.x, 0, x2DNormal.y);
	float3 xBillboardRight = cross(xBillboardUp,xBillboardFwd);
	
	// Add the transformed offset
	// ----------------------------------------
	xWorldPosition.xyz += (xBillboardRight * xVertexOffset.x);
	xWorldPosition.xyz += (xBillboardUp    * xVertexOffset.y);
	
	// Movement from the bone
	// ----------------------------------------
	fBoneWeight =  axVerts[iBillboardVertex].z * xBoneOffsetAndScale.w;	
	//xWorldPosition.xyz += mul(xWorldMatrix, xBoneOffset.xyz) * fBoneWeight;	
	xWorldPosition.xyz = Bot_CalcNewBonedPosition( xWorldPosition.xyz, fBoneWeight, mul(xWorldMatrix, xBoneOffsetAndScale.xyz));

	// Project it
	// ----------------------------------------
	xOut.m_xPosition = mul( xWorldPosition, g_xViewProjectionMatrix );	
	
	xOut.m_xUVPair.xz = dot(axUSelect[iBillboardVertex],xUVSet);
	xOut.m_xUVPair.yw = dot(axVSelect[iBillboardVertex],xUVSet);
	
	// Lighting calculations all done in world space now
	// --------------------------------------------------------------------------------------		
	xOut.m_afAlphaCut   = fInstanceAlpha.x;
	
	const float3x3 xMatrix = float3x3( -g_xCameraRightVector,
		 							    g_xCameraUpVector,
									    g_xCameraForwardsVector );
	
	xOut.m_xTangent     = -mul(  xMatrix, -xTangent  );
	xOut.m_xBinormal    =  mul(  xMatrix,  xBinormal );
	xOut.m_xNormal      = -g_xCameraForwardsVector; 
	
		
	// Light it - from front and back, then lerp
	half3 xColour, xBackColour;		
	half4 xNorm = half4(xOut.m_xNormal,1);						
	xOut.m_xColour.rgb = SHLookup( xNorm, xRedSH, xGreenSH, xBlueSH );
		
	xNorm.rgb=-xNorm.rgb;	
	xOut.m_xBackColour.rgb =  SHLookup( xNorm, xRedSH, xGreenSH, xBlueSH );
	
	
	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xWorldPosition.xyz;
	xOut.m_xCameraDirDistance.w   = xOut.m_xPosition.w;
	
	return xOut;
}
 

/*
DeferredOutputPixel PShader( const OutputVertex xIn )
{
	const half4 xAlbedo = float4(1,1,1,1); //tex2D( g_xTexture, xIn.UVA.xy );

	half4 xColour;
	xColour.rgb = xAlbedo * xIn.Colour * 2.0;
	xColour.a = saturate( xAlbedo.a - xIn.UVA.z );

	const float3 xWorldSpaceNormal = normalize( xIn.NormalDistance.xyz );

	DeferredOutputPixel xOut;

	xOut = Deferred_FillOutputPixel( xColour,
									 xWorldSpaceNormal,		// Normal
									 xIn.NormalDistance.w,	// Distance
									 xAlbedo.rgb,			// Albedo
									 0.0,					// Specular mask
									 1.0);					// Material v-coord

	return xOut;
}
*/

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass
	{
		AlphaTestEnable = TRUE;
		AlphaFunc       = GREATER;
		AlphaRef        = 0;
		CullMode        = CW;
		ZEnable         = TRUE;
		ZWriteEnable    = TRUE;
		VertexShader    = compile vs_3_0 VShader();
		PixelShader     = compile ps_3_0 PShader_Botanicals_Prelighting();		
	}
}