/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

// First pass, creating the maps

float4x4 g_xShadowViewProj;

sampler2D g_xTransparencyMask : register(s0) = sampler_state
{
	AddressU = WRAP;
	AddressV = WRAP;
};

float2 g_xTexAnim;

// Second & third passes, Gaussian blur

float g_fBlurSlice;

sampler3D g_xBlurSampler = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
	AddressW = CLAMP;
};

// Fourth pass, applying to scene

float4x4 g_axShadowViewProj[4];
float4x4 g_xInvCamViewProj;

float2 g_xTwoOverScreenDims;
float2 g_xCameraTanHalfFOV;

float4 g_xSplits;

float2 g_xArrayCoeffs;

float g_fPower;
float g_afStrengths[4];

sampler2D g_xDepthSampler = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};

sampler3D g_xShadowMapArray = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
	AddressW = CLAMP;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

float SampleShadow_Depth( const float3 xUVW, const float fOffsetX, const float fOffsetY )
{
	float fResult;

	asm
	{
		tfetch3D fResult.r___, xUVW, g_xBlurSampler, MinFilter=point, MagFilter=point,
			MipFilter=point, VolMinFilter=point, VolMagFilter=point, AnisoFilter=disabled,
			OffsetX=fOffsetX, OffsetY=fOffsetY, UnnormalizedTextureCoords=true
	};

	return fResult;
}

float2 SampleShadow_Moments( const float3 xUVW, const float fOffsetX, const float fOffsetY,
	const bool bUnnormalised )
{
	float2 xResult;

	asm
	{
		tfetch3D xResult.rg__, xUVW, g_xShadowMapArray, MinFilter=linear, MagFilter=linear,
			MipFilter=point, VolMinFilter=point, VolMagFilter=point, AnisoFilter=disabled,
			OffsetX=fOffsetX, OffsetY=fOffsetY, UnnormalizedTextureCoords=bUnnormalised
	};

	return xResult;
}

// Create shadow map, no transparency

float4 VShader_CreateShadowMap( const float4 xPos : POSITION ) : POSITION
{
	return mul( xPos, g_xShadowViewProj );
}

// Create shadow map, 1-bit transparent strips. We need to alpha test against the texture

float4 VShader_CreateShadowMap_Trans( const float4 xPos : POSITION, const float2 xUV : TEXCOORD,
	out float2 xOutUV : TEXCOORD ) : POSITION
{
	xOutUV = xUV + g_xTexAnim;
	return mul( xPos, g_xShadowViewProj );
}

float4 PShader_CreateShadowMap_Trans( const float2 xUV : TEXCOORD ) : COLOR
{
	float4 xResult;
	xResult.rgb = 0.0f;

	asm
	{
		tfetch2D xResult.___a, xUV, g_xTransparencyMask, MinFilter=linear, MagFilter=linear,
			MipFilter=linear, AnisoFilter=disabled
	};

	return xResult;
}

// Blur horizontally then vertically

float4 VShader_Blur( const float4 xPos : POSITION ) : POSITION
{
	return xPos;
}

float4 PShader_BlurHorz( const float2 xViewportPos : VPOS ) : COLOR
{
	const float3 xUVW = { xViewportPos, g_fBlurSlice };

	// Point-sample the five taps, since we can't apply filtering before computing the squares

	float4 xSamples0123;
	float  fSample4;

	// Assume D3DRS_HALFPIXELOFFSET is set
	xSamples0123.x = SampleShadow_Depth( xUVW, -2.0f, 0.0f );
	xSamples0123.y = SampleShadow_Depth( xUVW, -1.0f, 0.0f );
	xSamples0123.z = SampleShadow_Depth( xUVW, +0.0f, 0.0f );
	xSamples0123.w = SampleShadow_Depth( xUVW, +1.0f, 0.0f );
	fSample4       = SampleShadow_Depth( xUVW, +2.0f, 0.0f );

	const float4 xKernel = { 1.f/16.f, 4.f/16.f, 6.f/16.f, 4.f/16.f };

	const float fBlurredDepth = dot( xSamples0123, xKernel ) + fSample4 * xKernel.x;

	// Square the samples
	const float4 xSquaredSamples0123 = xSamples0123 * xSamples0123;
	const float fSquaredSample4 = fSample4 * fSample4;

	// Find the "average" of the squares = E(x^2)
	const float fBlurredSquaredDepth =
		dot( xSquaredSamples0123, xKernel ) + fSquaredSample4 * xKernel.x;

	return float4( fBlurredDepth, fBlurredSquaredDepth, 0.0f, 0.0f );
}

float4 PShader_BlurVert( const float2 xViewportPos : VPOS ) : COLOR
{
	const float3 xUVW = { xViewportPos, g_fBlurSlice };

	// This shader works a little differently to the horizontal pass. Because we've got the squares
	// already, we can use linear filtering to reduce the number of samples needed to four

	float4 xSamples1, xSamples2;

	// Again, assume D3DRS_HALFPIXELOFFSET has been set
	xSamples1.xy = SampleShadow_Moments( xUVW, 0.0f, -1.5f, true );
	xSamples1.zw = SampleShadow_Moments( xUVW, 0.0f, -0.5f, true );
	xSamples2.xy = SampleShadow_Moments( xUVW, 0.0f, +0.5f, true );
	xSamples2.zw = SampleShadow_Moments( xUVW, 0.0f, +1.5f, true );

	// NB kernel slightly different from horz pass, but maths works out the same due to filtering
	const float4 xKernel = { 2.f/16.f, 6.f/16.f, 6.f/16.f, 2.f/16.f };

	const float fBlurredDepth = dot( float4( xSamples1.xz, xSamples2.xz ), xKernel );
	const float fBlurredSquaredDepth = dot( float4( xSamples1.yw, xSamples2.yw ), xKernel );

	return float4( fBlurredDepth, fBlurredSquaredDepth, 0.0f, 0.0f );
}

// Apply shadows to scene

float4 VShader_ApplyShadows( const float4 xPos : POSITION ) : POSITION
{
	return xPos;
}

float Chebychev( const float t, const float m, const float s2 )
{
	const float d = t - m;
	return s2 / (s2 + d*d);
}

float4 PShader_ApplyShadows( float2 xUV : VPOS ) : COLOR
{
	// Correct VPOS for predicated tiling
	xUV += g_xVPOSOffsetVector;

	// Reconstruct camera view space position, assuming D3DRS_HALFPIXELOFFSET is on
	float fViewSpaceDepth;
	asm
	{
		tfetch2D fViewSpaceDepth.r___, xUV, g_xDepthSampler, UnnormalizedTextureCoords=true,
			MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};

	float4 xViewSpacePos =
	{
		g_xCameraTanHalfFOV * fViewSpaceDepth * (xUV * g_xTwoOverScreenDims - 1.0f),
		fViewSpaceDepth,
		1.0f
	};

	// In view space, y goes up (our view matrices correct for Asura's y axis)
	xViewSpacePos.y = -xViewSpacePos.y;

	// Actually this is the inv view matrix only now
	const float4 xWorldSpacePos = mul( xViewSpacePos, g_xInvCamViewProj );

	// Calculate which segment of the cascade to use

	const float4 xDepths = fViewSpaceDepth.xxxx;
	const float4 xResults = (xDepths > g_xSplits);
	const float4 xOne = { 1.0f, 1.0f, 1.0f, 1.0f };
	const float fMapToUse = dot( xResults, xOne );

	// Reproject into shadow map clip space
	float4 xShadowCoords = mul( xWorldSpacePos, g_axShadowViewProj[ fMapToUse ] );

	// Scale from [-1,1] into [0,1] and invert y for texture lookup
	// TODO: incorporate this into the above matrix
	xShadowCoords.xy = xShadowCoords.xy * 0.5f + 0.5f;
	xShadowCoords.y = 1.0f - xShadowCoords.y;

	clip( xShadowCoords.xyz );
	clip( 1.0f - xShadowCoords.xyz );

	float3 xUVW;
	xUVW.xy = xShadowCoords.xy;
	xUVW.z = g_xArrayCoeffs.x + g_xArrayCoeffs.y * fMapToUse;

	// Sample depth from shadow's POV and work out if we're in shadow

	// Again, assume D3DRS_HALFPIXELOFFSET is on
	const float2 xSample = SampleShadow_Moments( xUVW, 0.0f, 0.0f, false );

	const float fMean = xSample.r;
	const float fVariance = saturate( xSample.g - fMean * fMean + 0.0001f );

	const float fChebychev = Chebychev( xShadowCoords.z, fMean, fVariance );

	const float fResult = xShadowCoords.z > fMean
						? 1.0f - pow( fChebychev, g_fPower )
						: 0.0f;

	return float4( 0.0f, 0.0f, 0.0f, saturate( fResult * g_afStrengths[ fMapToUse ] ) );
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass CreateShadowMap
	{
		VertexShader = compile vs_3_0 VShader_CreateShadowMap();
		PixelShader = NULL;
	}

	pass CreateShadowMap_Trans
	{
		VertexShader = compile vs_3_0 VShader_CreateShadowMap_Trans();
		PixelShader = compile ps_3_0 PShader_CreateShadowMap_Trans();
	}

	pass BlurHorz
	{
		VertexShader = compile vs_3_0 VShader_Blur();
		PixelShader = compile ps_3_0 PShader_BlurHorz();
	}

	pass BlurVert
	{
		VertexShader = compile vs_3_0 VShader_Blur();
		PixelShader = compile ps_3_0 PShader_BlurVert();
	}

	pass ApplyShadows
	{
		VertexShader = compile vs_3_0 VShader_ApplyShadows();
		PixelShader = compile ps_3_0 PShader_ApplyShadows();
	}
}