
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// See EdgeAntiAlias_FP.fpx in PS3 for original code!!
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#include "Shared.fxh"



///////////////////////////////////////////////////////////////////////////////
// Vertex shader for screen-space vertices

float4x4			g_xWorldViewProj;
float				g_fDepthOffset;
float4				g_xConstants; // x = screen width, y = screen height, z = 1 / screen width, w = 1 / screen height

int					g_iNumBoneWeightings	: register(i0);

const sampler g_xInputSampler //	: register( s0 );
{
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = POINT;
	MagFilter = POINT;
	MipFilter = POINT;
};

struct EEAAOutVertex
{
	float4 m_xPosition	: POSITION;
	float4 m_xPosition2	: TEXCOORD0;
};

// Vertex shader functions

float4 CalculateEdgeAAVertexPos( float2 xScreenSize, float4 xProjSpacePos )
{
	// Partially move into screen space, but don't convert from homogeneous (don't
	// do /.w), instead do it in pixel shader as this saves interpolation precision.
	// ( Why does that preserve precision? I don't know, ask BW :) )
	float4 xF = float4( xScreenSize.x/2, xScreenSize.y/2, 0, 1 );
	float4 xOutPos = xProjSpacePos*xF.xyzw + xProjSpacePos.w*xF.xyzz;

	// DirectX has inverted y....
	xOutPos.y = xScreenSize.y * xProjSpacePos.w - xOutPos.y;

	return xOutPos;
}

// Pixel shader functions

float4 SampleScreen( float2 xUV, float2 xOneOverScreenSize )
{
	return tex2D( g_xInputSampler, (xUV + float2( 0.5, 0.5 )) * xOneOverScreenSize );
}

// Because of the interpolation imprecision we often can't tell whether a couple of middle pixels
// of the horizontal (vertical) subsegment lay above or below the actual polygon edge, so add
// additional small error area that will blend between fixed xScreenPos-xHVMask and xScreenPos+xHVMask
// pixels, since we don't know which side xScreenPos belongs to (that will also cause artifacts 
// as these pixels are further apart, but it's much less noticeable)
float CalcMidLineErrorK( float2 xDeltaPos )
{
	const float fErrorEps = 0.001;
	return 1 - saturate( (dot( xDeltaPos.xy * xDeltaPos.xy, float2( 1, 1 ) ) - fErrorEps) * 100000.0 );
}

float4 DoEdgeAA( float2 xScreenPos, float4 xVertPos, float2 xOneOverScreenSize )
{
	// Do View->Screen transform here (saves a lot on precision for some reason, copied from PS3 version)
	float2 xPos = xVertPos.xy / xVertPos.w;

	// Determine whether we're a pixel on horizontal or vertical line.	
	// xHVMask will contain (1, 0) if the line is horizontal, and (0, 1) if the line is vertical.
	// I've got no idea whether this will work on PC the same way.
	float2 xDD = float2( ddx( xPos.x ), ddy( xPos.y ) );
	float2 xHVMask = saturate( (xDD.yx - float2( 0.5, 0.5 )) * 3 );
	
	// If we're a pixel on horizontal line, xPos.y will now contain y distance from the actual line.
	// (same for vertical, just in .x)
	xPos = xPos-xScreenPos;

	// See function comments
	float fErrorK = CalcMidLineErrorK( xPos );

	// Colour in red bits that we consider to be in the possible error area.
	//if( fErrorK > 0.5 )		return float4( 10, 0, 0, 1 ) ;

	// For every horizontal (vertical) line subsegment, .y (.x) offset will range from -0.5 to 0.5 if
	// the line goes upward, or 0.5 to -0.5 if it goes downward. We need to convert that into [1, 0] 
	// or [0, 1] blend value used to interpolate between background/on-object samples (we don't know 
	// which one is the background or the object, but it is irrelevant)
	half2 xBlendVal = 0.5 - xPos.xy;

	// Cut off error (although it works without this so it could be removed!)
	xBlendVal	*= xHVMask;
	xPos		*= xHVMask;

	float fBlendVal = dot( xBlendVal, float2( 1, 1 ) );	// it either has x, or y

	float2 xUV1 = floor( xScreenPos + xPos.xy );
	float2 xUV2 = xUV1 + xHVMask;	// make either x or y one

	xUV1 = lerp( xUV1, xScreenPos - xHVMask, fErrorK );
	xUV2 = lerp( xUV2, xScreenPos + xHVMask, fErrorK );

	float4 xColour = lerp( SampleScreen(xUV1, xOneOverScreenSize), SampleScreen(xUV2, xOneOverScreenSize), fBlendVal );
	return xColour;
}

///////////////////////////////////////////////////////////////////////////////

EEAAOutVertex VShader_Environment( const float4 xIn : POSITION )
{
	EEAAOutVertex xOut;

	xOut.m_xPosition = mul( xIn, g_xWorldViewProj );
 	xOut.m_xPosition.z += g_fDepthOffset * 1;

	xOut.m_xPosition2 = CalculateEdgeAAVertexPos( g_xConstants.xy, xOut.m_xPosition );

	return xOut;
}

float3x4 FetchBone( const int fIndex, out float fDamage )
{
	float4 xBoneMat0, xBoneMat1, xBoneMat2;

	// Bones stored in semantics POSITION1-3. Ignore the last row, it's always (0 0 0 1)

	asm
	{
		vfetch xBoneMat0, fIndex, position1, UseTextureCache=true
		vfetch xBoneMat1, fIndex, position2, UseTextureCache=true
		vfetch xBoneMat2, fIndex, position3, UseTextureCache=true
	};

	// Per-bone texturing blend factors ("damage values") are packed into one of the matrix
	// elements. We can reconstruct the row (assuming they're orthogonal) with a cross product
	fDamage = xBoneMat2.z;
	xBoneMat2.xyz = cross( xBoneMat0.xyz, xBoneMat1.xyz );

	return float3x4( xBoneMat0, xBoneMat1, xBoneMat2 );
}


struct SkinInputVertex
{
	float4 m_xPosition        : POSITION;
	float4 m_xBlendIndices    : BLENDINDICES;
	float4 m_xBlendWeights    : BLENDWEIGHT;

	float3 m_xTangent         : TANGENT;
	float3 m_xBinormal        : BINORMAL;
	float3 m_xNormal          : NORMAL;

	float4 m_xUVPair          : TEXCOORD0;
	float  m_fDualTextureLerp : TEXCOORD1;
};

// Disable pointless warning about [optimizeAutoZ] causing /XOautoz- to be ignored
#pragma warning( disable : 3608 )

[optimizeAutoZ( true )]
EEAAOutVertex VShader_Skin( const SkinInputVertex xIn )
{
	EEAAOutVertex xOut;

	// Accumulate the blended skinning matrix and damage value

	float3x4 xBoneMat =
	{
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
		{ 0.0f, 0.0f, 0.0f, 0.0f },
	};

	float fDamageValue = 0.0f;

	for( int i = 0; i < g_iNumBoneWeightings; ++i )
	{
		const float fWeight = xIn.m_xBlendWeights[i];

		[branch]
		if( fWeight > 0.0f )
		{
			float fDamage;
			const float3x4 xMat = FetchBone( xIn.m_xBlendIndices[i], fDamage );

			xBoneMat += xMat * fWeight;
			fDamageValue += fDamage * fWeight;
		}
	}


	// Apply skinning to the input geometry
	const float4 xPos = float4( mul( xBoneMat, xIn.m_xPosition ), 1.0f );

	xOut.m_xPosition = mul( xPos, g_xWorldViewProj );
 	xOut.m_xPosition.z += g_fDepthOffset * 1;

	xOut.m_xPosition2 = CalculateEdgeAAVertexPos( g_xConstants.xy, xOut.m_xPosition );

	return xOut;
}


///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// Pixel shader


float4 PShader_EdgeAA( float2 xScreenPos : VPOS, const EEAAOutVertex xIn ) : COLOR
{
	xScreenPos += g_xVPOSOffsetVector;

	//clip( -1 );
	//return float4( 10, 0, 0, 0 );

	float4 xColour = DoEdgeAA( xScreenPos, xIn.m_xPosition2, g_xConstants.zw );
	xColour.w = 0.5;

/*
SPECKS:

many/most of the specks are caused by near thing's edge's background being another edge:
	- if near edge drawn first, far edge will blend between 2 pixels, at angle where one is actually in foreground it will carry a speck onto the background (outside foreground object)
	- if far drawn first... near will stretch a pixel of far's AA but otherwise should look ok.
possible fix: have a depth cull on read-from, offset on lines should be enough

Black specks in the middle of white:
	- unknown, perhaps maths has a subtle div by zero? negative alpha values? - easily checked, clamp alphas before using ...
*/
	return xColour;
}

float4 PShader_EdgeAADebug( float2 xScreenPos : VPOS, const EEAAOutVertex xIn ) : COLOR
{
	xScreenPos += g_xVPOSOffsetVector;

	float2 xPos = xIn.m_xPosition2.xy / xIn.m_xPosition2.w;
	xPos = xPos-xScreenPos;

	float fErrorK = CalcMidLineErrorK( xPos );
	if( fErrorK > 0.5 )		return float4( 10, 0, 0, 1 ) ;

	return float4( 10, 10, 0, 1 );
}



///////////////////////////////////////////////////////////////////////////////
// Techniques                                                                //
///////////////////////////////////////////////////////////////////////////////

technique EdgeAntiAliasing_Environment
{
	// pass 0
	pass
	{
		VertexShader = compile vs_3_0 VShader_Environment();
		PixelShader = compile ps_3_0 PShader_EdgeAA( );
	}

	// pass 1 - debug
	pass
	{
		VertexShader = compile vs_3_0 VShader_Environment();
		PixelShader = compile ps_3_0 PShader_EdgeAADebug( );
	}
}

technique EdgeAntiAliasing_Skin
{
	// pass 0
	pass
	{
		VertexShader = compile vs_3_0 VShader_Skin();
		PixelShader = compile ps_3_0 PShader_EdgeAADebug( ); 
		//PixelShader = compile ps_3_0 PShader_EdgeAA( );
	}

	// pass 1 - debug
	pass
	{
		VertexShader = compile vs_3_0 VShader_Skin();
		PixelShader = compile ps_3_0 PShader_EdgeAADebug( );
	}
}
