
/******************************************************************************/
/* Xbox360 version of the 'Advanced Bloom/SM3.0 Bloom' shader				  */
/* Dave Reed - 24/01/07														  */
/*																			  */
/* Basically a port of the PC version, but updated to work the 360 way, with  */
/* just one render target which gets resolved to a texture.					  */
/*																			  */
/* I've not attempted to optimize this yet, in case we decide to change it's  */
/* behaviour																  */
/******************************************************************************/

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"

const sampler g_xTexture : register(s0) = sampler_state
{	
	AddressU	= CLAMP;
	AddressV	= CLAMP;
};

const sampler g_xAccumTexture : register(s1) = sampler_state
{	
	AddressU	= CLAMP;
	AddressV	= CLAMP;
};
 
const sampler g_xDepthTexture : register(s2) = sampler_state
{	
	AddressU = CLAMP;
	AddressV = CLAMP;
};


const float2 g_xUVOffset          = float2(0.05f,0.0f);
const bool   g_bBlur 			  = true;
const bool   g_bExtractBrightBits = true;
const float  g_fSaturation	      = 0.0f;
const bool   g_bAccumulate        = false;
const float4 g_xOutputColour	  = float4(1,1,1,1);
const float4 g_xInputThreshold	  = float4(0,0,0,0);
const float  g_fInputScale        = 1.0f;
const float  g_fAccumFade         = 1.0f;
const bool   g_bExtractNearDepthOfField = false;
const float4 g_xFocalDistances;
const float2 g_xUVCorrection;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float2 m_xPosition   : POSITION;
	float2 m_axUVs	     : TEXCOORD0;
};

struct OutputVertex
{
	float4 m_xPosition   : POSITION;
	float4 m_axUVs[3]    : TEXCOORD0;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn)
{
	OutputVertex xOut;

	xOut.m_xPosition = float4(xIn.m_xPosition.x,xIn.m_xPosition.y,0,1);

	// (AAM - 23.02.2009) Used to be 5 float2s, but I'm packing them to save interpolants

	xOut.m_axUVs[0] = float4( xIn.m_axUVs, 0.0f, 0.0f);
	xOut.m_axUVs[1].xy = xIn.m_axUVs + ( -1.5f  * g_xUVOffset);
	xOut.m_axUVs[1].zw = xIn.m_axUVs + ( -0.5f  * g_xUVOffset);
	xOut.m_axUVs[2].xy = xIn.m_axUVs + (  0.5f  * g_xUVOffset);
	xOut.m_axUVs[2].zw = xIn.m_axUVs + (  1.5f  * g_xUVOffset);
	return xOut;
}


float4 PShader( const OutputVertex xIn,  const float2 xScreenPos : VPOS  ) : COLOR
{
	float4 xCol;

	if (g_bBlur)
	{
		xCol = 0.25f * (TexFetch2D_LinearLinearLinear(g_xTexture, xIn.m_axUVs[1].xy)+
				      	TexFetch2D_LinearLinearLinear(g_xTexture, xIn.m_axUVs[1].zw)+
			      		TexFetch2D_LinearLinearLinear(g_xTexture, xIn.m_axUVs[2].xy)+
		  		      	TexFetch2D_LinearLinearLinear(g_xTexture, xIn.m_axUVs[2].zw));
		
		xCol.a *= g_xFocalDistances.z;

	}
	else
	{
		xCol = TexFetch2D_LinearLinearLinear(g_xTexture, xIn.m_axUVs[0].xy);
		xCol.a= 1.0f;
	}		
	
	if (g_bExtractBrightBits)
	{
	     xCol = saturate(xCol - g_xInputThreshold);
	     xCol *= g_fInputScale;
  
	     float fMono = 0.3333f * (xCol.r+xCol.g+xCol.b);		
	     xCol = lerp(float4(fMono,fMono,fMono,1), xCol, g_fSaturation);             	     
	     	     
	     if (g_bAccumulate)
	     {
			// Hacky - this will need improving, especially if the system is updated to handle split-screen...
			// The UVs passed in are correct for sampling from the width x height buffer - this correction is so
			// we sample correctly from the half-dimensions buffer used to accumulate the glow
			//
			// (Values are 0.5/screen width and 0.5/screen height)
			
			xCol += (g_fAccumFade * TexFetch2D_LinearLinearLinear(g_xAccumTexture, xIn.m_axUVs[0].xy + g_xUVCorrection));
	     }	     	
	}
	else
	{
		
	    xCol *= g_xOutputColour;
	}
	
	if (g_bExtractNearDepthOfField)
	{
		float4 xDepth;
		float2 xScaledScreenPos = xScreenPos * 2.0f;
		asm
		{
			tfetch2D xDepth, xScaledScreenPos.xy, g_xDepthTexture, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		};
			
		float fBlurNear = saturate(1.0 - (xDepth.r-g_xFocalDistances.x) / (g_xFocalDistances.y-g_xFocalDistances.x));
		xCol.a = fBlurNear;
	}

	
	return xCol;
}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass
	{
		AlphaTestEnable = FALSE;
		CullMode = NONE;
		ZEnable = FALSE;
		ZWriteEnable = FALSE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader();
	}
}