
// ========================================================================================================================
//
// LightShaft.fx - LightShaft shader - Dave Reed 16/07/07
//
// 2 shaders in one... also contains the shaders for the corona at the origin of the lightshaft
//
// ========================================================================================================================

#include "Shared.fxh"


const float4x4 g_xPosOrient;
const float4   g_xConeGeometry;		// .x/.y = near/far radii.  .z/.w = near/far distances
const float4   g_xNearColour;
const float4   g_xFarColour;
const float4   g_xTextureParams;

const texture g_xTexture;
const sampler g_xTextureSampler : register(s0) = sampler_state
{
	Texture = <g_xTexture>;
	AddressU = WRAP;
	AddressV = CLAMP;
};

const texture g_xTexture2;
const sampler g_xTextureSampler2 : register(s1) = sampler_state
{
	Texture = <g_xTexture2>;
	AddressU = WRAP;
	AddressV = CLAMP;
};

const texture g_xDepthTexture;
const sampler g_xDepthTextureSampler : register(s2) = sampler_state
{
	Texture   = <g_xDepthTexture>;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};



// ------------------------------------------------------------------------------------------------------------------------
// Vertex Formats - Light Shaft
// ------------------------------------------------------------------------------------------------------------------------

struct VS_Input_LightShaft
{
	float4	xPosition					: POSITION0;	
	float4	xUVAB						: TEXCOORD0;	// UV, Alpha, Bottom-of-cone (if 1)
};

struct VS_Output_LightShaft
{
    float4	xPosition					: POSITION0;    
	float3	xDepthDiameterFade			: TEXCOORD0;	
	float3  xNormal						: TEXCOORD1;	
	float3  xConeForwards				: TEXCOORD2;			
	float4  xColour						: TEXCOORD3;
	float4  xUV							: TEXCOORD4;
};


// ========================================================================================================================
// Vertex Shader - Light Shaft
// ========================================================================================================================

VS_Output_LightShaft VS_LightShaft (VS_Input_LightShaft IN)
{
	VS_Output_LightShaft	OUT; 

    // Calculate the vertex position (object space)
    // ------------------------------------------------------------------------			   
    float4 xPos =  ((IN.xPosition * g_xConeGeometry.x + float4(0,0,g_xConeGeometry.z,0)) * (1.0f-IN.xUVAB.w)) +
				   ((IN.xPosition * g_xConeGeometry.y + float4(0,0,g_xConeGeometry.w,0))  * IN.xUVAB.w);
	xPos.w = 1;
								
	// Transform pos and normal to world space
	// ------------------------------------------------------------------------			   
	float3 xNormal = mul(IN.xPosition.xyz, g_xPosOrient);
	OUT.xNormal		= xNormal;
	float4 xWorldPos= mul(xPos,g_xPosOrient);    
    OUT.xPosition	= mul(xWorldPos, g_xViewProjectionMatrix);        

	// Output depth and diameter (for volumetric calculations)
	// ------------------------------------------------------------------------
	float fRadius =  (g_xConeGeometry.x * (1.0f-IN.xUVAB.w)) + (g_xConeGeometry.y * IN.xUVAB.w);
    OUT.xDepthDiameterFade.x = OUT.xPosition.w;	
	OUT.xDepthDiameterFade.y = fRadius * 2;

	// Calculate 'forward' vector perpendicular to the axis of the cone, that
	// we'll dot the normals with to get the alpha/depth for each pixel
	// ------------------------------------------------------------------------
    float4 xAxisPos = (float4(0,0,g_xConeGeometry.z,0) * (1.0f-IN.xUVAB.w)) +
				      (float4(0,0,g_xConeGeometry.w,0)  * IN.xUVAB.w);
	xAxisPos.w=1;
	xAxisPos = mul(xAxisPos, g_xPosOrient);		
	float3 xAxisDir = mul(float3(0,0,1), g_xPosOrient);	
	
	// This is a bit mad... we lerp between 2 different ways of calculating the vector, as
	// if we use the way that works best at a distance, it causes problems when the camera
	// gets close, giving us unwanted sharp edges... this fixes it
	
	float fLerp = min(1.0f,max(0.0f,OUT.xPosition.w / (2.0f*fRadius)));	
	float3 xTargetPos = (xWorldPos.xyz * (1-fLerp)) + (xAxisPos.xyz * fLerp);	
	
	float3 xAxisToCam = normalize(g_xCameraPosition-xTargetPos);
	OUT.xConeForwards = normalize(cross(xAxisDir, normalize(cross(xAxisToCam,xAxisDir))));

	// Calculate fade-out value for when we're looking into the beam
	// ------------------------------------------------------------------------
	float3 xCam = normalize(g_xCameraPosition-xWorldPos);
	float fDot = abs(dot(xAxisDir, xCam));	
	OUT.xDepthDiameterFade.z = max(0,(fDot*fDot*fDot)+0.2f);
	
	// Handle UV scaling and scrolling	
	// ------------------------------------------------------------------------
	OUT.xUV.xy = IN.xUVAB.xy * float2(g_xTextureParams.x,1) + float2(g_xTextureParams.z,0); 		
	OUT.xUV.zw = IN.xUVAB.xy * float2(g_xTextureParams.y,1) + float2(g_xTextureParams.w,0); 		
	
	// Fade out near camera
	// ------------------------------------------------------------------------
	float fDistFade = min(max((OUT.xPosition.w - 0.5) * 0.25f,0.0f), 1.0f);
	
	// Lerp colour
	// ------------------------------------------------------------------------
	OUT.xColour =  fDistFade* 2.0f * ((g_xNearColour * (1.0f-IN.xUVAB.w)) + (g_xFarColour * IN.xUVAB.w));
				
	return OUT;
}

// ========================================================================================================================
// Pixel Shader - Light Shaft
// ========================================================================================================================

float4 PS_LightShaft (const VS_Output_LightShaft IN, float2 xScreenPos : VPOS ) : COLOR
{	
	xScreenPos += g_xVPOSOffsetVector;

	// Calculate approximately how much of the shaft we're looking through
	// ------------------------------------------------------------------------
	float fDot = max(dot(IN.xConeForwards, IN.xNormal),0);
		
	// Do the and do the softness stuff
	// ------------------------------------------------------------------------	
	float4 xDepth; 
	asm
	{
		tfetch2D xDepth, xScreenPos, g_xDepthTextureSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};
	
	float fSoftValue = min(abs(xDepth.r-IN.xDepthDiameterFade.x)/(fDot*IN.xDepthDiameterFade.y),1.0f);					
		
	// Do the texture lookups, modulate by vertex colour
	// ------------------------------------------------------------------------
	float4 xColour = (TexFetch2D_LinearLinearLinear(g_xTextureSampler, IN.xUV.xy) * TexFetch2D_LinearLinearLinear(g_xTextureSampler2, IN.xUV.zw)) * IN.xColour; 					
			
	// Squaring fDot adds some falloff as we get towards the edge of the cone			
	// ------------------------------------------------------------------------
	fDot = max(0, (fDot-0.5f) * 2.0f - IN.xDepthDiameterFade.z);	
	xColour.a *= fSoftValue * (fDot*fDot);
					
	return xColour;
}



// ------------------------------------------------------------------------------------------------------------------------
// Vertex Formats - Corona
// ------------------------------------------------------------------------------------------------------------------------

struct VS_Input_Corona
{
	float4	xPosition					: POSITION0;	
	float4	xDiffuse					: COLOR0;
	float4	xUV							: TEXCOORD0;    
};

struct VS_Output_Corona
{
    float4	xPosition					: POSITION0;    
	float4	xUV							: TEXCOORD0;
	float4 	xDiffuse					: TEXCOORD1;    
};


// ========================================================================================================================
// Vertex Shader - Corona
// ========================================================================================================================

VS_Output_Corona VS_LightShaft_Corona(VS_Input_Corona IN)
{
	VS_Output_Corona	OUT;

    // Calculate the vertex position...
    float4 xPos	    = mul(IN.xPosition,g_xPosOrient);    
    OUT.xPosition	= mul(xPos, g_xViewProjectionMatrix);    
    OUT.xUV         = IN.xUV;
    OUT.xDiffuse    = 2.0f*IN.xDiffuse;

	// Packing our softness values into the UV interpolator
	// xUV.z already contains fSize * fVolumetricDepthScale	
	
	OUT.xUV.w		= OUT.xPosition.w;		// depth

	return OUT;
}


// ========================================================================================================================
// Pixel Shader - Corona
// ========================================================================================================================

float4 PS_LightShaft_Corona(const VS_Output_Corona IN, float2 xScreenPos : VPOS ) : COLOR
{
	// Soft/volumetric corona
		
	xScreenPos += g_xVPOSOffsetVector;
	
	float4 xColour = TexFetch2D_LinearLinearLinear(g_xTextureSampler, IN.xUV.xy); 	
		
	float4 xDepth; 
	asm
	{
		tfetch2D xDepth, xScreenPos, g_xDepthTextureSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};
							
	float fSoftValue = min(abs(xDepth.r-IN.xUV.w)/(IN.xUV.z),1.0f);
	xColour.a *= fSoftValue;
		
	return IN.xDiffuse * xColour;	
}



// ========================================================================================================================
// Techniques
// ========================================================================================================================

technique LightShaft
{
    pass p0
    {		
		// Pass 0 - The lightshaft itself	
		VertexShader	= compile vs_3_0 VS_LightShaft();
		PixelShader		= compile ps_3_0 PS_LightShaft();
    }
    
    pass P1
    {
		// Pass 1 - The corona if you look into the beam
		VertexShader    = compile vs_3_0 VS_LightShaft_Corona();
		PixelShader		= compile ps_3_0 PS_LightShaft_Corona();
    }
}
