/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Fog.fxh"
#include "Shared.fxh"


const sampler g_xTexture = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};

const sampler g_xDepthTexture = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};


const bool   g_bZReadEnable					= false;

const float  g_fTwoOverScreenWidth		    = 0;
const float  g_fTwoOverScreenHeight			= 0;

const float  g_fFirstInstance				= 0;
const float  g_fVolumetricDepthScale		= 0.0f;

const float4 g_xXDelta						= float4(1,0,0,0);
const float4 g_xYDelta						= float4(0,0,1,0);
const float4 g_xLightColour					= float4(1,1,1,1);
const float4 g_xPosOffsetAndSizeScale	    = float4(0,0,0,1);
const float4 g_xNearFarFade;



/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct OutputVertex
{
	float4 Position				 : POSITION;
	float4 Diffuse				 : COLOR;
	float4 UVPair				 : TEXCOORD0;
	float3 xParams				 : TEXCOORD4;		// .x = fDepth, .y = fSize, .z = fCrossFadeBlendWeight
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/




OutputVertex VShader( const int iIndex : INDEX )
{
	const float4 axVerts[4]=	   {	{ -1.0, -1.0  ,0.0,  1.0},
										{  1.0, -1.0  ,0.0,  1.0},
										{  1.0,  1.0  ,0.0,  1.0},
										{ -1.0,  1.0  ,0.0,  1.0}};

	const float4 axUSelector[16] = {    {1,0,0,0}, {0,0,1,0}, {0,0,1,0}, {1,0,0,0},
										{1,0,0,0}, {0,0,1,0}, {0,0,1,0}, {1,0,0,0},
										{0,0,1,0}, {1,0,0,0}, {1,0,0,0}, {0,0,1,0},
										{0,0,1,0}, {1,0,0,0}, {1,0,0,0}, {0,0,1,0}};
										
	const float4 axVSelector[16] = {    {0,1,0,0}, {0,1,0,0}, {0,0,0,1}, {0,0,0,1},
										{0,0,0,1}, {0,0,0,1}, {0,1,0,0}, {0,1,0,0},
										{0,0,0,1}, {0,0,0,1}, {0,1,0,0}, {0,1,0,0},
										{0,1,0,0}, {0,1,0,0}, {0,0,0,1}, {0,0,0,1}};								 
				 
	
	// Integer arithmetic has some rounding errors; see help topic "Vertex Fetching in HLSL"
	// and thread "Hardware instancing oddness" in xds360.graphics
	
	float fInstanceIndex  = ((iIndex + 0.5) / 4.0);				// 4 verts in each instance
	int   iInstanceIndex  = fInstanceIndex;	
	int   iVertexIndex    = iIndex - (iInstanceIndex*4.0);
	iInstanceIndex  = iInstanceIndex + g_fFirstInstance;
	
		
	float4 xInstancePosition;
	float4 xDiffuse;
	float2 xParams;
	float2 xSize;
	float4 xUVSet1;
	float4 xUVSet2;
	
	// Fetch the vertex (from the geometry stream)
	asm
	{
		vfetch xInstancePosition.xyzw, iInstanceIndex, position0	
		vfetch xUVSet1.xyzw,		   iInstanceIndex, texcoord0
		vfetch xUVSet2.xyzw,		   iInstanceIndex, texcoord1
		vfetch xSize.xy,			   iInstanceIndex, texcoord2
		vfetch xParams.xy,			   iInstanceIndex, texcoord3
		vfetch xDiffuse.xyzw,		   iInstanceIndex, color0		
	};
	
	float fRotation = xInstancePosition.w;	
	
	// Handle global size scale
	xSize = xSize * g_xPosOffsetAndSizeScale.w;
	
	OutputVertex xOut;
	
	// Colour and Misc params
	// ---------------------------------------------
	xOut.xParams.z = xParams.y;										// CrossFadeBlendWeight
	xOut.xParams.y = xSize.x*2.0f * g_fVolumetricDepthScale;		// Size
	xOut.Diffuse = xDiffuse;
	xOut.Diffuse.rgb *= g_xLightColour.rgb;


	// Handle the near/far fade (and culling when
	// alpha=0 by setting the size to 0)
	// ---------------------------------------------
		
	float4 xPos;
	xPos.xyz = xInstancePosition.xyz + g_xPosOffsetAndSizeScale.xyz;
	xPos.w   = 1.0f;
				
	float4 x2 = mul( xPos , g_xViewProjectionMatrix);				// <- OPTIMIZE THIS...
	float fDist = x2.w;
		
	float fNearFade = min(1,max(0,(fDist - g_xNearFarFade.y) / (g_xNearFarFade.x - g_xNearFarFade.y)));
	float fFarFade  = min(1,max(0,(fDist - g_xNearFarFade.w) / (g_xNearFarFade.z - g_xNearFarFade.w)));
	xOut.Diffuse.a *= min(fNearFade,fFarFade);
	
	if (xOut.Diffuse.a<=0)
	{
		// Cull it if alpha is 0, by degenerating it to a point. 
		xSize.xy=float2(0,0);
	}
		
	// Get the base vertex			
	// ---------------------------------------------
	float4 xTmp = axVerts[iVertexIndex];
	
	// Scale it	
	// ---------------------------------------------
	xTmp.xy = xTmp.xy * xSize.xy;
	
	// Rotate it
	// ---------------------------------------------
	float2 xRot0,xRot1,xRotated;
	sincos(-fRotation, xRot0.y, xRot0.x);        
	xRot1.x= xRot0.y;							
	xRot1.y=-xRot0.x;	
	xRotated.x = dot(xRot0, xTmp.xy);
	xRotated.y =-dot(xRot1, xTmp.xy);
	
	// Align it
	// ---------------------------------------------
	xTmp.xyz = xRotated.x*g_xXDelta.xyz + xRotated.y*g_xYDelta.xyz;	

	// Position it
	// ---------------------------------------------
	xTmp.xyz = xTmp.xyz + xInstancePosition.xyz + g_xPosOffsetAndSizeScale.xyz;
		
	// Xform to screen
	// ---------------------------------------------
	xOut.Position = mul( xTmp , g_xViewProjectionMatrix );
	
	// Do the UVs
	// ---------------------------------------------
	int  iOrient = xParams.x;	
	int  iTexOrientIndex = iOrient*4 + iVertexIndex;
	xOut.UVPair.x = dot(xUVSet1 , axUSelector[iTexOrientIndex]);
	xOut.UVPair.y = dot(xUVSet1 , axVSelector[iTexOrientIndex]);
	xOut.UVPair.z = dot(xUVSet2 , axUSelector[iTexOrientIndex]);
	xOut.UVPair.w = dot(xUVSet2 , axVSelector[iTexOrientIndex]);
		
	xOut.xParams.x = xOut.Position.w;
		
	return xOut;
}



float4 PShader( const OutputVertex xIn, float2 xScreenPos : VPOS, uniform bool bCrossfadeTextureFrames, uniform bool bVolumetric ) : COLOR
{
	xScreenPos += g_xVPOSOffsetVector;

	float4 xColour	= TexFetch2D_LinearLinearLinear( g_xTexture, xIn.UVPair.xy );    
	
	if( bCrossfadeTextureFrames )
	{
		xColour = lerp( TexFetch2D_LinearLinearLinear( g_xTexture, xIn.UVPair.zw ), xColour, xIn.xParams.z );
	}

	xColour *= xIn.Diffuse;

	if (bVolumetric)
	{
		// Do the soft particle stuff
		float4 xDepth;
		asm
		{
			tfetch2D xDepth, xScreenPos, g_xDepthTexture, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		};
		
	    float fSoftValue = min(abs(xDepth.r-xIn.xParams.x)/xIn.xParams.y,1.0f);
		xColour.a *= fSoftValue;	
	}
		
	if (g_bFogEnabled)
	{
		return ApplyFog(xColour,xIn.xParams.x);
	}
	else
	{	
		return xColour;
	}
}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique Point
{
	pass SingleTexture
	{
		AlphaTestEnable = TRUE;	
		AlphaRef = 0;
		CullMode = NONE;
		ZEnable = <g_bZReadEnable>;
		ZFunc = LESSEQUAL;
		ZWriteEnable = FALSE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader(false, false);
	}

	pass DualTexture
	{
		AlphaTestEnable = TRUE;	
		AlphaRef = 0;
		CullMode = NONE;
		ZEnable = <g_bZReadEnable>;
		ZFunc = LESSEQUAL;
		ZWriteEnable = FALSE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader(true, false);
	}
}

technique Volumetric
{
	pass SingleTexture
	{
		AlphaTestEnable = TRUE;	
		AlphaRef = 0;
		CullMode = NONE;
		ZEnable = <g_bZReadEnable>;
		ZFunc = LESSEQUAL;
		ZWriteEnable = FALSE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader(false, true);
	}

	pass DualTexture
	{
		AlphaTestEnable = TRUE;	
		AlphaRef = 0;
		CullMode = NONE;
		ZEnable = <g_bZReadEnable>;
		ZFunc = LESSEQUAL;
		ZWriteEnable = FALSE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader(true, true);
	}
}
