///////////////////////////////////////////////////////////////////////////////
// This code is ported from Ben's PS3 version and is probably not perfectly
// same as the original, although it should behave almost in the same way.
///////////////////////////////////////////////////////////////////////////////


/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"
#include "Deferred_Shared.fxh"

float4				g_xScreenResolution;		// .xy == screen_resolution, .zw == (2.0, -2.0) / screen_resolution

// x = fn/(f-n); y = f/(f-n); z = n; where n = near clip, f = far clip, corresponding to the depth buffer being read.
float4				g_xDepthProjection;

float4				g_xParams;					// x = sample distance, y = threshold factor, z = 1 if first pass 0 if not

float				g_fDebugSwitch;				// used for debugging only

float2				g_xFocalLen;

const sampler		g_xDfrNormalBufferSampler	// full screen deferred shading normal buffer
{
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};

const sampler		g_xDfrDepthMtlBufferSampler	// full screen deferred shading DepthMtl buffer
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};

const sampler		g_xDepthBufferSampler		// depth buffer used in blur (should be w/2, h/2 of the screen)
{
	AddressU = CLAMP; 
	AddressV = CLAMP;
};

float4 VShader( const float4 xIn : POSITION ) : POSITION
{
	return xIn;
}

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/


void	SmartDepthBlur( float4 xInDepths, float4 xInConfidences, out float fOutDepth, out float fOutConfidence )
{
	// coverage:
	// use whichever depth region occupies the most samples, fade confidence as they approach 50/50

	// pick values that are close enough to each other (within threshold)
	// choose the group of values closest to us unless we have more confidence in some other values
	// i.e. pick the highest confidence, using closest as a tie-breaker.
	// obviously if they're all close enough we use the average, otherwise we use the average of the in-range vals.
	
	// mask values of equal-highest confidence
 	float2 xTemp = max( xInConfidences.xy, xInConfidences.zw );
 	
 	//small threshold value, for precision (possibly not needed) //TODO: check this
	float fMaxConf = max( xTemp.x, xTemp.y ) - 0.001f;
	
	const float fIgnoreValue = -100000;
	// 0 = use this value, fIgnoreValue = ignore this value
	float4 xIgnoreMask = (xInConfidences >= fMaxConf.xxxx) ? float4(0, 0, 0, 0) : float4(fIgnoreValue, fIgnoreValue, fIgnoreValue, fIgnoreValue);
	
	// pick closest of these
	float4 xMaskedVals = xInDepths + xIgnoreMask; // depth vals are always less than fIgnoreValue, so adding fIgnoreValue to them guarantees they won't be closest
	xTemp = max( xMaskedVals.xy, xMaskedVals.zw );
	float fClosest = max( xTemp.x, xTemp.y );
	
	// construct xFlags from all vals within threshold of that
 	float4 d1 = xInDepths;
 	float4 d2 = fClosest.xxxx;
	float4 xDepth = abs( d1-d2 );
//	float4 xFlags = ( xDepth < g_xParams.yyyy*dot(xInConfidences,float4(0.25)) ) ? float4(1) : float4(0);
//now with fading:
	float4 xFlags = saturate( 0.5f + (1.0f - xDepth / g_xParams.yyyy) / 0.4f );
	
	// mul by confidence
	xFlags *= xInConfidences;
	float fFlagMag = dot( xFlags, float4(1, 1, 1, 1) );
	
	// new confidence = 1111.flags / 4;
	fOutConfidence = fFlagMag / 4.0f;
	
	// new val = vals.flags / 1111.flags;
	fOutDepth = dot( xFlags, xInDepths ) / fFlagMag;
}

float4 PShader_DepthDownsampleFromRT( const float2 xScreenVPos : VPOS ) : COLOR
{
	// smart filter:
	//   sample 4 points (which may themselves be sampled from 4 points, etc.)
	//   compute a point that represents the occlusion of this area when viewed from points outside the area
	//   (if it's relevant) we can assume those points are about 1/3rd to 1/2 the width of this area away, approximately horizontally & vertically
	//     - at the time of writing precise positions of those points haven't been decided (it's likely something that will be tweaked for best results).

	float2 xTexCoords = xScreenVPos * 2; //		downsampling!
	
	float2 xSamplingCoords[4];
	xSamplingCoords[0] = xTexCoords + float2(-1,-1) * g_xParams.x;
	xSamplingCoords[1] = xTexCoords + float2( 1,-1) * g_xParams.x;
	xSamplingCoords[2] = xTexCoords + float2(-1, 1) * g_xParams.x;
	xSamplingCoords[3] = xTexCoords + float2( 1, 1) * g_xParams.x;

	float4 xDepthBufferValues[4];
	asm
	{
		tfetch2D xDepthBufferValues[0], xSamplingCoords[0], g_xDfrDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[1], xSamplingCoords[1], g_xDfrDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[2], xSamplingCoords[2], g_xDfrDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[3], xSamplingCoords[3], g_xDfrDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};
	
	float4 xDepthSamples = 
	float4(
			xDepthBufferValues[0].r, 
			xDepthBufferValues[1].r,
			xDepthBufferValues[2].r,
			xDepthBufferValues[3].r
		);
	
	float fConfidence;
	float fDepth;
	
	SmartDepthBlur( xDepthSamples, float4(1, 1, 1, 1), fDepth, fConfidence );

	return float4( fDepth, fConfidence, 0, 1 );
}

float4 PShader_DepthBlur( float2 xScreenVPos : VPOS ) : COLOR
{
	float2 xTexCoords = xScreenVPos;
	
	// explicitly create a 1/2 pixel drift to allow pixels to represent blur over an even-number of pixels.
	//i.e. there's no centre pixel to an 8x8 box, so push things .5 towards the negative (by reading .5 towards the +ve)
	xTexCoords += 0.5; 

	float2 xSamplingCoords[16];
	xSamplingCoords[ 0] = xTexCoords + float2(-1,-1)*g_xParams.x;
	xSamplingCoords[ 1] = xTexCoords + float2(-3,-1)*g_xParams.x;
	xSamplingCoords[ 2] = xTexCoords + float2(-1,-3)*g_xParams.x;
	xSamplingCoords[ 3] = xTexCoords + float2(-3,-3)*g_xParams.x;
	xSamplingCoords[ 4] = xTexCoords + float2( 1,-1)*g_xParams.x;
	xSamplingCoords[ 5] = xTexCoords + float2( 3,-1)*g_xParams.x;
	xSamplingCoords[ 6] = xTexCoords + float2( 1,-3)*g_xParams.x;
	xSamplingCoords[ 7] = xTexCoords + float2( 3,-3)*g_xParams.x;
	xSamplingCoords[ 8] = xTexCoords + float2(-1, 1)*g_xParams.x;
	xSamplingCoords[ 9] = xTexCoords + float2(-3, 1)*g_xParams.x;
	xSamplingCoords[10] = xTexCoords + float2(-1, 3)*g_xParams.x;
	xSamplingCoords[11] = xTexCoords + float2(-3, 3)*g_xParams.x;
	xSamplingCoords[12] = xTexCoords + float2( 1, 1)*g_xParams.x;
	xSamplingCoords[13] = xTexCoords + float2( 3, 1)*g_xParams.x;
	xSamplingCoords[14] = xTexCoords + float2( 1, 3)*g_xParams.x;
	xSamplingCoords[15] = xTexCoords + float2( 3, 3)*g_xParams.x;
	
	float4 xDepthBufferValues[16];
	// [unroll]for( int i =0 ; i < 16; i++ ) <-> doesn't work :( 'error X3500: Microcode Compiler not yet implemented: array lval with non-constant index not supported in inline microcode'

	asm
	{
		tfetch2D xDepthBufferValues[ 0], xSamplingCoords[ 0], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 1], xSamplingCoords[ 1], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 2], xSamplingCoords[ 2], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 3], xSamplingCoords[ 3], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 4], xSamplingCoords[ 4], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 5], xSamplingCoords[ 5], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 6], xSamplingCoords[ 6], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 7], xSamplingCoords[ 7], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 8], xSamplingCoords[ 8], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[ 9], xSamplingCoords[ 9], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[10], xSamplingCoords[10], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[11], xSamplingCoords[11], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[12], xSamplingCoords[12], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[13], xSamplingCoords[13], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[14], xSamplingCoords[14], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthBufferValues[15], xSamplingCoords[15], g_xDepthBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};

	float4 axDepthSamples[4] = {
			float4( xDepthBufferValues[0].x, xDepthBufferValues[4].x, xDepthBufferValues[8].x, xDepthBufferValues[12].x ),
			float4( xDepthBufferValues[1].x, xDepthBufferValues[5].x, xDepthBufferValues[9].x, xDepthBufferValues[13].x ),
			float4( xDepthBufferValues[2].x, xDepthBufferValues[6].x, xDepthBufferValues[10].x, xDepthBufferValues[14].x ),
			float4( xDepthBufferValues[3].x, xDepthBufferValues[7].x, xDepthBufferValues[11].x, xDepthBufferValues[15].x )
		};
	
	float4 axConfidenceSamples[4] = {
			float4( xDepthBufferValues[0].y, xDepthBufferValues[4].y, xDepthBufferValues[8].y, xDepthBufferValues[12].y ),
			float4( xDepthBufferValues[1].y, xDepthBufferValues[5].y, xDepthBufferValues[9].y, xDepthBufferValues[13].y ),
			float4( xDepthBufferValues[2].y, xDepthBufferValues[6].y, xDepthBufferValues[10].y, xDepthBufferValues[14].y ),
			float4( xDepthBufferValues[3].y, xDepthBufferValues[7].y, xDepthBufferValues[11].y, xDepthBufferValues[15].y )
		};

	float4 xConfidencesMiddleStep;
	float4 xDepthsMiddleStep;
	
	SmartDepthBlur( axDepthSamples[0], axConfidenceSamples[0], xDepthsMiddleStep.x, xConfidencesMiddleStep.x );
	SmartDepthBlur( axDepthSamples[1], axConfidenceSamples[1], xDepthsMiddleStep.y, xConfidencesMiddleStep.y );
	SmartDepthBlur( axDepthSamples[2], axConfidenceSamples[2], xDepthsMiddleStep.z, xConfidencesMiddleStep.z );
	SmartDepthBlur( axDepthSamples[3], axConfidenceSamples[3], xDepthsMiddleStep.w, xConfidencesMiddleStep.w );

	float fDepth;
	float fConfidence;
	
	SmartDepthBlur( xDepthsMiddleStep, xConfidencesMiddleStep, fDepth, fConfidence );
	return float4( fDepth, fConfidence, 0, 1 );
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


sampler			g_axDepthDownsampleSampler0
{	
	AddressU = CLAMP; 
	AddressV = CLAMP;
};

sampler			g_axDepthDownsampleSampler1
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};

float4x4		g_xViewMatrix;

float4			g_xTwoOverScreen;		// dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
float4			g_xCameraTanHalfFOV;	// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
float2			g_xDepthAdjustment;

float4			g_xBrightnessContrast;

float4			g_xDistanceFade;

float4			g_axSamplePositions[2]; // = { float4( -1,1, -2,6 ), float4( -6,26, 0,0 ) }; (for testing)

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// used for debugging only
/*
const float4x4 g_xInvViewMatrix;
const float    g_fTwoOverScreenWidth;						// Dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float    g_fTwoOverScreenHeight;
const float    g_fCameraTanHalfHorizFOV;					// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
const float    g_fCameraTanHalfVertFOV;
*/
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Convert z-buffer depth to distance from camera view plane
// ( distance from camera view is (worldPos * viewProj).w, z-buffer depth is ((worldPos * viewProj).z / (worldPos * viewProj).w) )
float DepthToDistanceFromViewPlane( float fDepth )
{
	// invert the projection & perspective. i.e.
	// Z = b-a*z; W=z; depth = Z/W;
	// so z = a/(b-depth)
 	return g_xDepthAdjustment.x / (g_xDepthAdjustment.y - fDepth);
}

// xScreenVPos is in screen space
// returns position in view space
float3 ScreenToViewPosition( float2 xScreenVPos, float fScreenDepth )
{
	float2 xUV = xScreenVPos * g_xScreenResolution.zw;
    xUV = (xUV - float2(1.0, -1.0));
    return float3(xUV / g_xFocalLen * fScreenDepth, fScreenDepth);
}

// This can be useful for testing (to perform calculations in world space instead of view space)
// (in order for this to work you'll have to enable setting of variables in Asura_Direct3D_AmbientOcclusion.cpp, ApplyShadows )
/*
float4 ScreenToWorldPosition( float2 xDepthTexturePixelPos, float fDepth )
{
	const float4 xViewSpacePosition = float4(	g_fCameraTanHalfHorizFOV * fDepth * (xDepthTexturePixelPos.x * g_fTwoOverScreenWidth  - 1.0),
													g_fCameraTanHalfVertFOV  * fDepth * -(xDepthTexturePixelPos.y * g_fTwoOverScreenHeight - 1.0),
													fDepth, 1.0);
	
	float4 xWorldPosition = mul( xViewSpacePosition, g_xInvViewMatrix );
	return xWorldPosition;
}
*/

float CalcOcclusion( float3 xBasePos, float3 xBaseNormal, float3 xSamplePos, float fSampleConfidence )
{
	float3 xDiff = xSamplePos - xBasePos;
	
	float fLength = sqrt( dot( xDiff.xyz,xDiff.xyz ) );

	float fResult = saturate( dot( xBaseNormal, xDiff.xyz ) / fLength ); // we *do* need all this precision & floats.
 		
	float fReduction = 1.0;

//		// reduce more on distant samples to hide the hard-edges resulting from the limited blur range in the data
//		float fPixelRange = 2*sqrt(dot(xSamp,xSamp)); // distance in pixels to fade to
//		float fMaxDist = g_xCameraTanHalfFOV.x * fDepth * fPixelRange * g_xTwoOverScreen.x; // equivalent world-space distance
//		// clamp max dist to give more consistent results at different depths
//		//fMaxDist = min( fMaxDist, 1.0f );
	float fMaxDist = g_xDistanceFade.x * fSampleConfidence; //^ is just not getting nice results, < is surprisingly effective
	float fFadeRange = 0.2f; // as a proportion of 1.0
	fReduction *= saturate( (1.0f - fLength / fMaxDist ) / fFadeRange );

	// reduce the significance of this sample
	// since position is not linearly related to occlusion angle we might want to tweak this curve a bit (so long as it's cheaper than an acos)
	fReduction = pow( fReduction, 1 );

	return fResult * fReduction;
}

float4 SampleOcclusion( float2 xScreenVPos, float fBaseDepth, float3 xBaseViewPos, float3 xBaseNormal, float2 xSamp, sampler xDepthSampler, int iLevel, out float4 xConfidence )
{
	float2 axSamplePos[4] = { xSamp, float2( xSamp.y, -xSamp.x ), -xSamp, float2( -xSamp.y, xSamp.x ) };
	
	float4 xResult;
	
	float fScale = 2;

	float2 xRoundedBasePos = round(xScreenVPos/fScale); // round to closest
	
	[unroll] for ( int i=0; i < 4; i++ )
	{
		float fDepth;

		float2 xActualPos, xTextureCoord;
		
		if ( iLevel == 1 )
		{
			// sample the first downsampled buffer, pixel centres are at the centre of the 4 pixels they represent, /2
			xTextureCoord = xRoundedBasePos + axSamplePos[i]/fScale;
		}
		else
		{
			// sample a buffer blurred from one of the same size, which has a 1/2 pixel offset in the -ve direction.
			xTextureCoord = xRoundedBasePos + axSamplePos[i]/fScale - 0.5;
		}
		xActualPos = xRoundedBasePos*fScale + axSamplePos[i]; // this will be innaccurate with sample pos's that don't deal properly with the downsampled buffer (e.g. non-int ones or odd ones when iLevel>1)
		
		
		// <*> // float4 xTexel = h4tex2D( g_axDepthDownsampleSampler[iLevel-1], xTextureCoord );
		// <*> // fDepth = Deferred_GetDepth( xTexel );
		float4 xTexel;
		asm
		{
			tfetch2D xTexel, xTextureCoord, xDepthSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		};
		
		fDepth = xTexel.x;

		// convert this value to an amount of occlusion, allowing for perspective distortion
 		float3 xAdjPos = ScreenToViewPosition( xActualPos, fDepth );
 		//float3 xAdjPos = ScreenToWorldPosition( xActualPos, fDepth );
 		
		xConfidence[i]	= xTexel.y;
 		xResult[i]		= CalcOcclusion( xBaseViewPos, xBaseNormal, xAdjPos, xTexel.y );
	}
	
	//xResult.yzw = 0.0;
	//xConfidence.yzw = 0.0;
	
	return xResult;
}

float4 PShader_Occlusion( float2 xScreenVPos : VPOS ) : COLOR
{
	xScreenVPos += g_xVPOSOffsetVector;
	
 	// <*> // float4 xNormalDepth = GetNormalDepth(xScreenVPos);
 	// <*> // float fDepth = xNormalDepth.w;
	float3 xNormal;
	float fDepth;
	
	float4 xBufferNormal;
	float4 xBufferDepthMtl;
	asm
	{
		tfetch2D xBufferNormal,   xScreenVPos, g_xDfrNormalBufferSampler,   UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xBufferDepthMtl, xScreenVPos, g_xDfrDepthMtlBufferSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};
	Deferred_GetNormalDepth( xBufferNormal, xBufferDepthMtl, xNormal, fDepth );

 	// transform normal into view space
 	// (to do the occlusion in world space (useful for testing) disable this line, and use ScreenToWorldPosition instead of ScreenToViewPosition)
 	xNormal = mul( xNormal, g_xViewMatrix );

	// get viewspace position of this pixel
 	float3 xPosition = ScreenToViewPosition( xScreenVPos, fDepth );
 	
 	const float fSampDist = 1;//3.0;

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS	(2) //corresponding to declaration in Asura_Direct3D_AmbientOcclusion.cpp

	float4 axOcclusion[NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS] = { float4(0, 0, 0, 0), float4(0, 0, 0, 0) };  //, float4(0, 0, 0, 0) };
	float4 axConfidence[NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS] = { float4(0, 0, 0, 0), float4(0, 0, 0, 0) }; //, float4(0, 0, 0, 0) };
	
	axOcclusion[0] = SampleOcclusion( xScreenVPos, fDepth, xPosition, xNormal, g_axSamplePositions[0].xy, g_axDepthDownsampleSampler0, 1, axConfidence[0] );
#if ( NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS > 1 )
	axOcclusion[1] = SampleOcclusion( xScreenVPos, fDepth, xPosition, xNormal, g_axSamplePositions[0].zw, g_axDepthDownsampleSampler1, 2, axConfidence[1] );
#if ( NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS > 2 )
	axOcclusion[2] = SampleOcclusion( xScreenVPos, fDepth, xPosition, xNormal, g_axSamplePositions[1].xy, g_axDepthDownsampleSampler2, 3, axConfidence[2] );
#endif
#endif
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
 	// reduce occlusion by confidence
 	// lerp towards nearer occlusion, to simulate a small number of peaks poking over the nearer things
	axOcclusion[0] *= axConfidence[0];
	
	float4 xOcclusion = axOcclusion[0];
#if ( NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS > 1 )
	axOcclusion[1] = lerp( axOcclusion[0], axOcclusion[1], axConfidence[1] );
	xOcclusion = max( xOcclusion, axOcclusion[1] );
#if ( NUM_AMBIENT_OCCLUSION_SHADER_SAMPLERS > 2 )
	xOcclusion = max( xOcclusion, axOcclusion[2] );
	axOcclusion[2] = lerp( axOcclusion[1], axOcclusion[2], axConfidence[2] );
#endif
#endif

	// combine occlusion
	xOcclusion = acos( xOcclusion );	//this value is 0 for vertical (fully occluded) and pi/2 for horizontal (not occluded in this direction)
	
	// assume each direction represents approximate occlusion of a 90 degree region, so just average them together
	// might be better to simulate some overlap (to fake the regions that aren't sampled), e.g. for 180 degrees you add the opposing ones xz+=yw then multiply the results vis = x*z
	float fVisibility = dot( xOcclusion, float4(1 / 6.283185, 1 / 6.283185, 1 / 6.283185, 1 / 6.283185) );
	
	// adjust & clamp the effect
	// - to remove artefacts on curved surfaces (which self-shadow since the normals don't match the actual shape)
	// - and to artificially deepen shadows
	fVisibility = saturate(fVisibility*g_xBrightnessContrast.x + g_xBrightnessContrast.y);

//return float4( fVisibility * (axConfidence[2].xyz * axConfidence[2].w+0)*2, 1 ); //neat CMYK debug mode.
//if ( fVisibility > 0.95 ) return float4 ( 1 - frac( fVisibility.xxx * float3(4,16,1)/0.05f ) ,1 ); // debug - emphasize errors in near-unoccluded surfaces (to pick up on flat surface self-occlusion artefacts)
//	return float4( fVisibility.xxx*3, 1 );	 //better brightness (silly HDR doesn't adapt well)
	return float4( fVisibility.xxx, 1 );
}

///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique SSAO
{
	// pass 0
	pass DepthDownsampleFromRT
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_DepthDownsampleFromRT( );
	}
	
	// pass 1
	pass DepthBlur
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_DepthBlur();
	}

	// pass 2
	pass Occlusion
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Occlusion();
		//PixelShader = compile ps_3_0 PShader_OcclusionV2();
	}
}
