//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Sound.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#ifndef ASURA_USE_AUDIO_SYSTEM

#include "Sound/Asura_Sound.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_StringBuffer.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_MemAllocator.h"
#include "Core/Asura_Multithread.h"
#include "Core/Threads/Asura_ProducerConsumerQueue.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif

// Asura Xbox 360
#include "Sound/Asura_Xbox360_Sound.h"
#include "Sound/Asura_Xbox360_Sound_LowPassFilter.h"

// Misc
#include <xaudio.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#ifdef ASURA_USE_OPENAL
	// (AAM - 16.11.2007) OpenAL causes memory leaks on 360, and the reverb never worked anyway
	#error OpenAL is no longer supported on 360. Please undefine ASURA_USE_OPENAL.
#endif

#define XBOX360_SOUND_XAUDIO_DEBUG_HIGH		( 0 && _DEBUG )
#define XBOX360_SOUND_USE_HARDWARE_PAN		( 1 && defined( ASURA_USE_OPENAL ) )

// This is the XMA2FORMAT version number our code is written against.
// If Microsoft bumps the version, we'll need to alter the loader to match
#define ASURA_XBOX360_SOUND_XMA2_VERSION	4

#ifndef ASURA_USE_OPENAL

	// Custom effect identifiers
	enum
	{
		XAUDIOFXID_LOWPASS = XAUDIOFXID_FIRSTCLIENT,
	};

	// Effect table entries
	const XAUDIOFXTABLEENTRY axEFFECT_TABLE_ENTRIES[] =
	{
		// Source effect
		{
			XAUDIOFXID_SOURCE,
			XAudioQuerySourceEffectSize,
			XAudioCreateSourceEffect,
		},

		// Router effect
		{
			XAUDIOFXID_ROUTER,
			XAudioQueryRouterEffectSize,
			XAudioCreateRouterEffect,
		},

		// Render effect
		{
			XAUDIOFXID_RENDER,
			XAudioQueryRenderEffectSize,
			XAudioCreateRenderEffect,
		},

		// Reverb
		{
			XAUDIOFXID_REVERB,
			XAudioQueryReverbEffectSize,
			XAudioCreateReverbEffect,
		},

		// Low pass filter
		{
			XAUDIOFXID_LOWPASS,
			Asura_Xbox360_Sound_LowPassFilter::QuerySize,
			Asura_Xbox360_Sound_LowPassFilter::Create
		},
	};

	// Effect table
	const XAUDIOFXTABLE xEFFECT_TABLE =
	{
		sizeof( axEFFECT_TABLE_ENTRIES ) / sizeof( axEFFECT_TABLE_ENTRIES[0] ),
		axEFFECT_TABLE_ENTRIES
	};

	// Initialisation structs for the reverb effect object
	const XAUDIOFXINIT xEffect = { XAUDIOFXID_REVERB, NULL };
	const XAUDIOFXINIT* pxEffectList = &xEffect;
	const XAUDIOVOICEFXCHAIN xEFFECT_CHAIN = { 1, &pxEffectList };

	// Maps Asura_Sound_System::ReverbType to I3DL2 settings
	static XAUDIOREVERBI3DL2SETTINGS axREVERB_SETTINGS[] =
	{
		// Standard I3DL2 presets
		{ I3DL2_ENVIRONMENT_PRESET_DEFAULT },
		{ I3DL2_ENVIRONMENT_PRESET_GENERIC },
		{ I3DL2_ENVIRONMENT_PRESET_PADDEDCELL },
		{ I3DL2_ENVIRONMENT_PRESET_ROOM },
		{ I3DL2_ENVIRONMENT_PRESET_BATHROOM },
		{ I3DL2_ENVIRONMENT_PRESET_LIVINGROOM },
		{ I3DL2_ENVIRONMENT_PRESET_STONEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_AUDITORIUM },
		{ I3DL2_ENVIRONMENT_PRESET_CONCERTHALL },
		{ I3DL2_ENVIRONMENT_PRESET_CAVE },
		{ I3DL2_ENVIRONMENT_PRESET_ARENA },
		{ I3DL2_ENVIRONMENT_PRESET_HANGAR },
		{ I3DL2_ENVIRONMENT_PRESET_CARPETEDHALLWAY },
		{ I3DL2_ENVIRONMENT_PRESET_HALLWAY },
		{ I3DL2_ENVIRONMENT_PRESET_STONECORRIDOR },
		{ I3DL2_ENVIRONMENT_PRESET_ALLEY },
		{ I3DL2_ENVIRONMENT_PRESET_FOREST },
		{ I3DL2_ENVIRONMENT_PRESET_CITY },
		{ I3DL2_ENVIRONMENT_PRESET_MOUNTAINS },
		{ I3DL2_ENVIRONMENT_PRESET_QUARRY },
		{ I3DL2_ENVIRONMENT_PRESET_PLAIN },
		{ I3DL2_ENVIRONMENT_PRESET_PARKINGLOT },
		{ I3DL2_ENVIRONMENT_PRESET_SEWERPIPE },
		{ I3DL2_ENVIRONMENT_PRESET_UNDERWATER },

		// EAX 2 additions
		{ I3DL2_ENVIRONMENT_PRESET_DRUGGED },
		{ I3DL2_ENVIRONMENT_PRESET_DIZZY },
		{ I3DL2_ENVIRONMENT_PRESET_PSYCHOTIC },

		// Castle presets
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_SMALLROOM },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_SHORTPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_MEDIUMROOM },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_LONGPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_LARGEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_HALL },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_CUPBOARD },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_COURTYARD },
		{ I3DL2_ENVIRONMENT_PRESET_CASTLE_ALCOVE },

		// Factory presets
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_ALCOVE },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_SHORTPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_MEDIUMROOM },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_LONGPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_LARGEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_HALL },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_CUPBOARD },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_COURTYARD },
		{ I3DL2_ENVIRONMENT_PRESET_FACTORY_SMALLROOM },

		// Ice palace presets
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_ALCOVE },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_SHORTPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_MEDIUMROOM },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_LONGPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_LARGEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_HALL },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_CUPBOARD },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_COURTYARD },
		{ I3DL2_ENVIRONMENT_PRESET_ICEPALACE_SMALLROOM },

		// Space station presets
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_ALCOVE },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_MEDIUMROOM },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_SHORTPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_LONGPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_LARGEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_HALL },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_CUPBOARD },
		{ I3DL2_ENVIRONMENT_PRESET_SPACESTATION_SMALLROOM },

		// Wooden galleon presets
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_ALCOVE },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_SHORTPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_MEDIUMROOM },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_LONGPASSAGE },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_LARGEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_HALL },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_CUPBOARD },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_SMALLROOM },
		{ I3DL2_ENVIRONMENT_PRESET_WOODEN_COURTYARD },

		// Sports presets
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_EMPTYSTADIUM },
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_SQUASHCOURT },
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_SMALLSWIMMINGPOOL },
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_LARGESWIMMINGPOOL },
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_GYMNASIUM },
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_FULLSTADIUM },
		{ I3DL2_ENVIRONMENT_PRESET_SPORT_STADIUMTANNOY },

		// Prefab presets
		{ I3DL2_ENVIRONMENT_PRESET_PREFAB_WORKSHOP },
		{ I3DL2_ENVIRONMENT_PRESET_PREFAB_SCHOOLROOM },
		{ I3DL2_ENVIRONMENT_PRESET_PREFAB_PRACTISEROOM },
		{ I3DL2_ENVIRONMENT_PRESET_PREFAB_OUTHOUSE },
		{ I3DL2_ENVIRONMENT_PRESET_PREFAB_CARAVAN },

		// Dome and pipe presets
		{ I3DL2_ENVIRONMENT_PRESET_DOME_TOMB },
		{ I3DL2_ENVIRONMENT_PRESET_PIPE_SMALL },
		{ I3DL2_ENVIRONMENT_PRESET_DOME_SAINTPAULS },
		{ I3DL2_ENVIRONMENT_PRESET_PIPE_LONGTHIN },
		{ I3DL2_ENVIRONMENT_PRESET_PIPE_LARGE },
		{ I3DL2_ENVIRONMENT_PRESET_PIPE_RESONANT },

		// Outdoors presets
		{ I3DL2_ENVIRONMENT_PRESET_OUTDOORS_BACKYARD },
		{ I3DL2_ENVIRONMENT_PRESET_OUTDOORS_ROLLINGPLAINS },
		{ I3DL2_ENVIRONMENT_PRESET_OUTDOORS_DEEPCANYON },
		{ I3DL2_ENVIRONMENT_PRESET_OUTDOORS_CREEK },
		{ I3DL2_ENVIRONMENT_PRESET_OUTDOORS_VALLEY },

		// Mood presets
		{ I3DL2_ENVIRONMENT_PRESET_MOOD_HEAVEN },
		{ I3DL2_ENVIRONMENT_PRESET_MOOD_HELL },
		{ I3DL2_ENVIRONMENT_PRESET_MOOD_MEMORY },

		// Driving simluation presets
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_COMMENTATOR },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_PITGARAGE },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_INCAR_RACER },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_INCAR_SPORTS },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_INCAR_LUXURY },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_FULLGRANDSTAND },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_EMPTYGRANDSTAND },
		{ I3DL2_ENVIRONMENT_PRESET_DRIVING_TUNNEL },

		// City presets
		{ I3DL2_ENVIRONMENT_PRESET_CITY_STREETS },
		{ I3DL2_ENVIRONMENT_PRESET_CITY_SUBWAY },
		{ I3DL2_ENVIRONMENT_PRESET_CITY_MUSEUM },
		{ I3DL2_ENVIRONMENT_PRESET_CITY_LIBRARY },
		{ I3DL2_ENVIRONMENT_PRESET_CITY_UNDERPASS },
		{ I3DL2_ENVIRONMENT_PRESET_CITY_ABANDONED },

		// Misc rooms
		{ I3DL2_ENVIRONMENT_PRESET_DUSTYROOM },
		{ I3DL2_ENVIRONMENT_PRESET_CHAPEL },
		{ I3DL2_ENVIRONMENT_PRESET_SMALLWATERROOM },
	};

	const u_int uNUM_REVERB_SETTINGS = sizeof( axREVERB_SETTINGS ) / sizeof( axREVERB_SETTINGS[0] );

	ASURA_STATIC_ASSERT( uNUM_REVERB_SETTINGS == Asura_Sound_System::ASURA_REVERB_ID_MAX );

#endif

//The minimum memory allocated to an on-demand sound (to avoid fragmentation)
#ifndef XBOX360_SOUND_ON_DEMAND_SOUNDS_MINIMUM_ALLOCATION_SIZE
	#define XBOX360_SOUND_ON_DEMAND_SOUNDS_MINIMUM_ALLOCATION_SIZE (256*1024)
#endif

//The size of the on-demand sound buffer, should be large enough to accomodate Max concurrent sounds * max sound size
#ifndef XBOX360_SOUND_ON_DEMAND_SOUNDS_MAXIMUM_ALLOCATION_SIZE
	#define XBOX360_SOUND_ON_DEMAND_SOUNDS_MAXIMUM_ALLOCATION_SIZE (2*1024*1024)
#endif

#define XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE		16
#define XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_ACTIONS		16
#define XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_PREPARATIONS	8	//for the producer/consumer class
#define XBOX360_SOUND_ON_DEMAND_SOUNDS_PREPARATION_THREADS		1	//for the producer/consumer class

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	static float g_fDistanceFilterScale			= 1.0f;
	static float g_fOcclusionFilterScale		= 1.0f;
#define CONSOLE_VARNAME_AUDIO_DISTANCE_SCALE	L"Audio.Distance.LowPassFilterScale"
#define CONSOLE_VARNAME_AUDIO_OCCLUSION_SCALE	L"Audio.Occlusion.LowPassFilterScale"
#endif

const float fASURA_XBOX360_SOUND_LPF_GAIN	= -30.0f;		//negative gain

/******************************************************************************/
/* Enumerations                                                                */
/******************************************************************************/

enum ASURA_XBOX360_DSP_EFFECTS
{
	ASURA_XBOX360_DSP_EFFECTS_DISTANCE_FILTER	= 0,
	ASURA_XBOX360_DSP_EFFECTS_OCCLUSION_FILTER	= 0,
	ASURA_XBOX360_DSP_EFFECTS_REVERB			= 0
};

enum ASURA_ON_DEMAND_QUEUE_ACTIONS
{
	ASURA_ON_DEMAND_QUEUE_ACTION_PLAY	= 0,
	ASURA_ON_DEMAND_QUEUE_ACTION_PAUSE,
	ASURA_ON_DEMAND_QUEUE_ACTION_RESUME,
	ASURA_ON_DEMAND_QUEUE_ACTION_STOP,
	ASURA_ON_DEMAND_QUEUE_ACTION_PITCH,
	ASURA_ON_DEMAND_QUEUE_ACTION_USES_REVERB,
	ASURA_ON_DEMAND_QUEUE_ACTION_VOLUME_PAN_FADE,
	ASURA_ON_DEMAND_QUEUE_ACTION_DIST_FILTER_GAIN,
	ASURA_ON_DEMAND_QUEUE_ACTION_WET_DRY_VOLUME,
	ASURA_ON_DEMAND_QUEUE_ACTION_OCCLUSION_LPF,

	ASURA_STREAM_QUEUE_ACTION_NUM_ITEMS
};

enum ASURA_XBOX360_ON_DEMAND_SOUND_STATE
{
	ASURA_ON_DEMAND_STATE_VOID					= 0x00000000,	//No state has been set

	ASURA_ON_DEMAND_STATE_PREPARING				= 0x00000001,	//Data is being loaded (threaded out)
	ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE	= 0x00000002,	//Data is loaded, need to wait for Update() to clear queue
	ASURA_ON_DEMAND_STATE_PREPARED				= 0x00000004,	//Finished preparing
	ASURA_ON_DEMAND_STATE_PREPARE_FAILED		= 0x00000008,	//Failed to prepare correctly
	ASURA_ON_DEMAND_STATE_PREPARE_MASK			= 0xFFFFFFF0,

	ASURA_ON_DEMAND_STATE_STOPPED				= 0x00000010,	//These are not so much the playing state
	ASURA_ON_DEMAND_STATE_PLAYING				= 0x00000020,	//as what we have queued as our last action
	ASURA_ON_DEMAND_STATE_PAUSED				= 0x00000040,
	ASURA_ON_DEMAND_STATE_PLAYBACK_MASK			= 0xFFFFFF0F,
	
	ASURA_ON_DEMAND_STATE_NUM_ITEMS
};

/******************************************************************************/
/* Forward declarations                                                       */
/******************************************************************************/

class Asura_Xbox360_OnDemandSound;
class SourceVoice;

static void							ThreadedSoundLoading			(void* pData, u_int uDataSize);
static void							AddToOnDemandSoundUpdates		(Asura_Xbox360_OnDemandSound* pxSound);
static void							RemoveFromOnDemandSoundUpdates	(const Asura_Xbox360_OnDemandSound* pxSound);
static Asura_Xbox360_OnDemandSound* PopNextOnDemandSoundToUpdate	();
static bool							UpdateVoiceWetDryVolumes		(SourceVoice* pxVoice);

/******************************************************************************/
/* Global Variables part 1                                                    */
/******************************************************************************/

static char*					g_pcOnDemandBuffer		= NULL;
static Asura_MemAllocator*		g_pxOnDemandAllocator	= NULL;
static Asura_FixedArray<Asura_Xbox360_OnDemandSound*,XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE>	g_apxOnDemandUpdates;
static Asura_Mutex				g_OnDemandUpdatesMutex;	//protect the list of updatable sounds
static Asura_Xbox360_OnDemandSound* g_apxOnDemandSounds[ASURA_MAX_NUMBER_OF_LOADED_SOUNDS];
static Asura_Xbox360_OnDemandSound*	g_apxOnDemandSoundPool_Global		= NULL;	//On demand sounds which will not be reset
static Asura_Xbox360_OnDemandSound*	g_apxOnDemandSoundPool_Resettable	= NULL;	//On demand sounds which can be deleted on reset

static bool	g_bHitResetPoint = false;

Asura_ProducerConsumerQueue*				g_pxOnDemandSoundQueue = NULL;

/******************************************************************************/
/* Classes                                                                    */
/******************************************************************************/

#ifdef ASURA_USE_OPENAL
	Asura_OpenAL_Sound_API		Asura_Xbox360_Sound::s_xAPI;
	Asura_Xbox360_ActiveSound	Asura_Xbox360_Sound::s_axActiveSounds[ASURA_XBOX360_SOUND_MAX_VOICES];
	RawXMASample				Asura_Xbox360_Sound::s_axXMASamples[ASURA_MAX_NUMBER_OF_LOADED_SOUNDS];
#endif

////////////////////////////////////////////////////////////////////////////////
//
// RIFFChunkHeader, XMA2Format
//
// Helper structures for loading XMA sound data
//
////////////////////////////////////////////////////////////////////////////////

class RIFFChunkHeader
{
public:

	u_int m_uChunkType;
	u_int m_uDataSize;

	bool ReadFromFile( const Asura_Handle iFileHandle, const u_int uExpectedChunkType );
	bool ReadFromFile( const Asura_Handle iFileHandle );
};

#ifndef ASURA_USE_OPENAL
struct XMA2Format : public XMA2WAVEFORMAT
{
	// There's already one XMA2STREAMFORMAT in the parent: add the rest
	XMA2STREAMFORMAT Streams2[ XAUDIOXMASTREAMCOUNT_MAX - 1 ];
};
#endif

////////////////////////////////////////////////////////////////////////////////
//
// SourceBuffer
//
// Represents a loaded sound buffer
//
////////////////////////////////////////////////////////////////////////////////

#ifndef ASURA_USE_OPENAL
struct SourceBuffer
{
	XAUDIOSOURCEBUFFER m_xXAudioSourceBuffer;

	u_int m_uNumStreams;
	u_int m_auStreamChannelCounts[ XAUDIOXMASTREAMCOUNT_MAX ];
	u_int m_auStreamSampleRates[ XAUDIOXMASTREAMCOUNT_MAX ];
};
#endif

////////////////////////////////////////////////////////////////////////////////
//
// SourceVoice
//
// Represents a single playing sound voice
//
////////////////////////////////////////////////////////////////////////////////

#ifndef ASURA_USE_OPENAL
class SourceVoice
{
public:

	SourceVoice();

	IXAudioSourceVoice* m_pxXAudioSourceVoice;
	IXAudioSubmixVoice* m_pxXAudioDrySubmix;
	SourceBuffer* m_pxSourceBuffer;

	// Platform-independent code sets these before calling play, so we have to cache them here
	float m_fVolume;
	float m_fPan;
	float m_fFade;
	float m_fPitch;
	bool m_bReverb;
	float m_fWetVolume;
	float m_fDryVolume;
	float m_fOcclusionFilter;

	// Pause state
	bool m_bPaused;
	bool m_bPausedByPitch;

	//Is this voice associated with an on-demand sound?
	int m_iOnDemandIndex;
};
#endif

class Asura_Xbox360_OnDemandQueuedAction
{
public:

	DWORD m_uIdentifier;
	int m_iSoundID;
	union OnDemandQueuedActionData
	{
		float m_fPitch;
		bool m_bUseReverb;
		float m_fGain;
		struct Asura_OnDemand_VolPanFade
		{
			float m_fVolume;
			float m_fPan;
			float m_fFade;
		} m_xVolPanFade;
		bool m_bLooping;
		struct Asura_OnDemand_WetDry
		{
			float m_fWetVolume;
			float m_fDryVolume;
		} m_xWetDry;
		float m_fOcclusion;
	} m_xData;
};


class Asura_Xbox360_OnDemandSound
{

public:

	Asura_Xbox360_OnDemandSound		();
	~Asura_Xbox360_OnDemandSound	();
	
	bool IsOnDemandSound	() const;
	void SetAsOnDemand		();
	void SetAsNotOnDemand	();
	
	void SetAsResettable	(const bool bResettable);
	bool IsResettable		() const;

	void Prepare	(const int iPlayingSoundID);
	bool Load		();
	void Unload		();
	void Destroy	(const int iSoundHandle);

	Asura_Xbox360_OnDemandQueuedAction*	PopQueuedAction	();

	const u_int		GetState					() const;
	void			SetPreparedState			(const u_int uFlag);
	void			SetPlayingState				(const u_int uFlag);
	
	SourceBuffer*	GetSourceBuffer				()								{ return m_pxSourceBuffer; }
	void			SetSourceBuffer				(SourceBuffer* pxBuffer)		{ m_pxSourceBuffer = pxBuffer; }
	
	const char*		GetFilename () const		{ return m_xFilename.GetString();	} 
	void			SetFilename					(const char* szFilename)	{ m_xFilename = szFilename;	}

	void			SetBaseFileOffset			(const u_int uOffset)			{ m_uBaseFileOffset = uOffset;	}
	u_int			GetBaseFileOffset			() const						{ return m_uBaseFileOffset;		}

	u_int			GetFileSize					() const						{ return m_uFileSize; }
	void			SetFileSize					(u_int val)						{ m_uFileSize = val; }

	//Redirections from Platform functions,
	//if the sound is still preparing they (possibly queue up the request and) return false
	//if the sound is already prepared they return true and do nothing, allowing platform code to execute
	bool QueuePlay					(const int iSoundHandle, const bool bLooping);
	bool QueueStop					(const int iSoundHandle);
	bool QueuePause					(const int iSoundHandle);
	bool QueueResume				(const int iSoundHandle);
	bool QueueSetPitch				(const int iSoundHandle, const float fPitch);
	bool QueueSetUsesReverb			(const int iSoundHandle, const bool bUseReverb);
	bool QueueSetVolumePanAndFade	(const int iSoundHandle, const float fVolume, const float fPan, const float fFade);
	bool QueueSetDistFilterGain		(const int iSoundHandle, const float fGain);
	bool QueueSetWetDryVolume		(const int iSoundHandle, const float fWetVolume, float fDryVolume);
	bool QueueSetOcclusionLPF		(const int iSoundHandle, const float fOcclusion);

	//An unpleasant but necessary function
	void BusyWaitUntilPrepared	() const;

private:
	
	void Clear			();
	Asura_Xbox360_OnDemandQueuedAction* NextFreeActionQueueItem			();
	Asura_Xbox360_OnDemandQueuedAction* GetFirstActionQueueItemByType	(u_int uIdentifier);

	bool			m_bIsOnDemandSound;
	u_int			m_uState;
	
	Asura_String256	m_xFilename;		//File which contains this sound
	u_int			m_uBaseFileOffset;	//Offset of the wavebank data into the file
	u_int			m_uFileSize;		//Size of the file
	SourceBuffer*	m_pxSourceBuffer;	//Our source buffer

	mutable Asura_Mutex	m_xStateMutex;	//Protect calls to state query/setting functions

	Asura_FixedArray<Asura_Xbox360_OnDemandQueuedAction, XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_ACTIONS> m_axActionQueue;
	u_int m_uActionQueueEnd;
	u_int m_uActionQueueStart;

	bool m_bResettable;
};


Asura_Xbox360_OnDemandSound::Asura_Xbox360_OnDemandSound	()
{
	m_bIsOnDemandSound = false;
	m_uState = ASURA_ON_DEMAND_STATE_VOID;
	m_pxSourceBuffer = NULL;
	m_uBaseFileOffset = 0;
	m_uFileSize = 0;
	m_uActionQueueEnd = 0;
	m_uActionQueueStart = 0;
	m_bResettable = false;
}


Asura_Xbox360_OnDemandSound::~Asura_Xbox360_OnDemandSound	()
{
	Clear();
}


const u_int Asura_Xbox360_OnDemandSound::GetState () const
{
	u_int uRet = 0;
	m_xStateMutex.Lock();
	uRet = m_uState;
	m_xStateMutex.Unlock();
	return uRet;
}


void Asura_Xbox360_OnDemandSound::SetPreparedState (const u_int uFlag)
{
	Asura_Assertf(!(uFlag & ASURA_ON_DEMAND_STATE_PREPARE_MASK),("Called SetPreparedState with an invalid flag, flag: %X", uFlag));

	m_xStateMutex.Lock();
	m_uState &= ASURA_ON_DEMAND_STATE_PREPARE_MASK;	//Clear
	m_uState |= uFlag;								//Set
	m_xStateMutex.Unlock();
}


void Asura_Xbox360_OnDemandSound::SetPlayingState (const u_int uFlag)
{
	Asura_Assertf(!(uFlag & ASURA_ON_DEMAND_STATE_PLAYBACK_MASK),("Called SetPreparedState with an invalid flag, flag: &X", uFlag));

	m_xStateMutex.Lock();
	m_uState &= ASURA_ON_DEMAND_STATE_PLAYBACK_MASK;	//Clear
	m_uState |= uFlag;									//Set
	m_xStateMutex.Unlock();
}


bool Asura_Xbox360_OnDemandSound::IsOnDemandSound() const
{
	return m_bIsOnDemandSound;
}


void Asura_Xbox360_OnDemandSound::SetAsOnDemand ()
{
	m_bIsOnDemandSound = true;
}


void Asura_Xbox360_OnDemandSound::SetAsNotOnDemand ()
{
	m_bIsOnDemandSound = false;
}


void Asura_Xbox360_OnDemandSound::SetAsResettable (const bool bResettable)
{
	m_bResettable = bResettable;
}


bool Asura_Xbox360_OnDemandSound::IsResettable () const
{
	return m_bResettable;
}


void Asura_Xbox360_OnDemandSound::Clear()
{
	//Clear state
	SetPlayingState	(0);
	SetPreparedState(0);
	if(m_pxSourceBuffer && m_pxSourceBuffer->m_xXAudioSourceBuffer.pBuffer)
	{
		g_pxOnDemandAllocator->Free(m_pxSourceBuffer->m_xXAudioSourceBuffer.pBuffer);
	}
	m_pxSourceBuffer = NULL;
	m_uActionQueueEnd = 0;
	m_uActionQueueStart = 0;
};


void Asura_Xbox360_OnDemandSound::Prepare (const int iPlayingSoundID)
{
	if((m_uState & ASURA_ON_DEMAND_STATE_PREPARED) || (m_uState & ASURA_ON_DEMAND_STATE_PREPARING))
	{
		//No need to load
		return;
	}

	SetPreparedState(ASURA_ON_DEMAND_STATE_PREPARING);

	//Try to add this to our preparation queue, we have plenty of queue slots so if we cant queue its a pretty safe bet something is wrong
	if(!g_pxOnDemandSoundQueue->TryToAddToQueue(ThreadedSoundLoading,this,sizeof(this)))
	{
		SetPreparedState(ASURA_ON_DEMAND_STATE_PREPARE_FAILED);
		Asura_Assert(false, "Ran out of queue space for on demand sound preparation, on demand sound will not play");
	}
};


bool Asura_Xbox360_OnDemandSound::Load ()
{
	SourceBuffer* pxNewBuffer = GetSourceBuffer();

	HANDLE uFile = CreateFile( GetFilename(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
	if( uFile == INVALID_HANDLE_VALUE )
	{
		Asura_Assertf(false, ("Failed to open file %s",GetFilename()));
		return false;
	}

	const u_int uFileOffset = GetBaseFileOffset();
	const u_int uFileSize	= GetFileSize();

	if(SetFilePointer(uFile,uFileOffset,NULL,FILE_BEGIN) == INVALID_SET_FILE_POINTER)
	{
		Asura_Assertf(false, ("Failed to seek file %s",GetFilename()));
		CloseHandle(uFile);
		return false;
	}

	DWORD uRead;
	RIFFChunkHeader xChunkHeader;

	if(!ReadFile( uFile, &xChunkHeader, sizeof(RIFFChunkHeader), &uRead, NULL))
	{
		Asura_Assertf(false, ("Failed to read file %s",GetFilename()));
		CloseHandle(uFile);
		return false;
	}

	// RIFF chunk contents are the subtype FOURCC then subtype-dependent data
	u_int uSubType;
	if(!ReadFile( uFile, &uSubType, sizeof( u_int ), &uRead, NULL))
	{
		Asura_Assertf(false, ("Failed to read file %s",GetFilename()));
		CloseHandle(uFile);
		return false;
	}
	if( uSubType != 'WAVE' )
	{
		Asura_Assert( false, "Invalid audio chunk." );
		CloseHandle(uFile);
		return false;
	}

	while(SetFilePointer(uFile,0,0,FILE_CURRENT) < uFileOffset + uFileSize )
	{
		if(!ReadFile( uFile, &xChunkHeader, sizeof( RIFFChunkHeader ), &uRead, NULL))
		{
			Asura_Assertf(false, ("Failed to read file %s",GetFilename()));
			CloseHandle(uFile);
			return false;
		}
		xChunkHeader.m_uDataSize = __loadwordbytereverse( 0, &xChunkHeader.m_uDataSize );

		switch( xChunkHeader.m_uChunkType )
		{
			case 'data':
			{
				if (g_pxOnDemandAllocator)
				{
					pxNewBuffer->m_xXAudioSourceBuffer.pBuffer = g_pxOnDemandAllocator->Allocate ( xChunkHeader.m_uDataSize, 4 ); 
				}
				if( !pxNewBuffer->m_xXAudioSourceBuffer.pBuffer )
				{
					Asura_Assert( false, "Failed to allocate memory for sound. Perhaps increase on demand sound buffer size?" );
					CloseHandle(uFile);
					return false;
				}

				// Don't endian-swap this
				if(!ReadFile( uFile, pxNewBuffer->m_xXAudioSourceBuffer.pBuffer, xChunkHeader.m_uDataSize, &uRead, NULL) )
				{
					Asura_Assertf(false, ("Failed to read file %s",GetFilename()));
					CloseHandle(uFile);
					return false;
				}

				pxNewBuffer->m_xXAudioSourceBuffer.BufferSize = xChunkHeader.m_uDataSize;
			}
			break;

			case 'XMA2':
			{
				// Note that XMA2 format chunks are stored big-endian, so no need to endian-swap

				XMA2Format xFormat;
				if( xChunkHeader.m_uDataSize > sizeof( xFormat ) )
				{
					Asura_Assert( false, "XMA2 audio chunk is too big." );
					CloseHandle(uFile);
					return false;
				}

				if(!ReadFile( uFile, &xFormat, xChunkHeader.m_uDataSize, &uRead, NULL) )
				{
					Asura_Assertf(false, ("Failed to read file %s",GetFilename()));
					CloseHandle(uFile);
					return false;
				}

				if ( xFormat.Version != ASURA_XBOX360_SOUND_XMA2_VERSION )
				{
					Asura_Assertf( false, ("XMA2 chunk has invalid version (%u). Should be %u. Level needs re-exporting, or code updating.", xFormat.Version, ASURA_XBOX360_SOUND_XMA2_VERSION) );
					CloseHandle(uFile);
					return false;
				}

				pxNewBuffer->m_uNumStreams = xFormat.NumStreams;
				for( u_int u = 0; u < xFormat.NumStreams; ++u )
				{
					const XMA2STREAMFORMAT& xStreamFormat = xFormat.Streams[u];

					pxNewBuffer->m_auStreamSampleRates[u] = xFormat.SampleRate;
					pxNewBuffer->m_auStreamChannelCounts[u] = xStreamFormat.Channels;

					if( xStreamFormat.Channels == 0 )
					{
						Asura_Assert( false, "" );
					}
				}

				if( xFormat.LoopEnd != 0 )
				{
					// As long as the file has been built with the /l option, these values should be valid
					pxNewBuffer->m_xXAudioSourceBuffer.LoopBegin = xFormat.LoopBegin;
					pxNewBuffer->m_xXAudioSourceBuffer.LoopEnd = xFormat.LoopEnd;
				}
				else
				{
					// This doesn't seamlessly loop the sample. Rebuild the data with /l for correct looping
					pxNewBuffer->m_xXAudioSourceBuffer.LoopBegin = 0;
					pxNewBuffer->m_xXAudioSourceBuffer.LoopEnd = xFormat.SamplesInSource;
				}
			}
			break;

			case 'seek':
			{
				// We don't care about seek tables
				if(SetFilePointer(uFile,xChunkHeader.m_uDataSize,NULL,FILE_CURRENT) == INVALID_SET_FILE_POINTER)
				{
					Asura_Assertf(false, ("Failed to seek file %s",GetFilename()));
					CloseHandle(uFile);
					return false;
				}
			}
			break;

			default:
			{
				Asura_Assert( false, "Unrecognised WAVE subchunk encountered in sound." );
				if(SetFilePointer(uFile,xChunkHeader.m_uDataSize,NULL,FILE_CURRENT) == INVALID_SET_FILE_POINTER)
				{
					Asura_Assertf(false, ("Failed to seek file %s",GetFilename()));
					CloseHandle(uFile);
					return false;
				}
			}
			break;
		}
	}
	CloseHandle(uFile);

	pxNewBuffer->m_xXAudioSourceBuffer.PlayBegin = 0;
	pxNewBuffer->m_xXAudioSourceBuffer.PlayEnd = 0;
	pxNewBuffer->m_xXAudioSourceBuffer.pContext = NULL;

	return true;
}


bool Asura_Xbox360_OnDemandSound::QueuePlay (const int iSoundHandle, const bool bLooping)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");

	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}

	//Mark the sound as 'playing'
	SetPlayingState(ASURA_ON_DEMAND_STATE_PLAYING);

	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}	

	//This sound is not prepared, queue the action
	Asura_Xbox360_OnDemandQueuedAction* pxAction = NextFreeActionQueueItem();
	if(pxAction)
	{
		pxAction->m_uIdentifier			= ASURA_ON_DEMAND_QUEUE_ACTION_PLAY;

		pxAction->m_iSoundID			= iSoundHandle;
		pxAction->m_xData.m_bLooping	= bLooping;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueStop (const int iSoundHandle)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}
	
	//Mark the sound as 'stopped'
	SetPlayingState(ASURA_ON_DEMAND_STATE_STOPPED);

	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	//This sound is not prepared, queue the action
	Asura_Xbox360_OnDemandQueuedAction* pxAction =  NextFreeActionQueueItem();
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_STOP;

		pxAction->m_iSoundID			= iSoundHandle;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueuePause	(const int iSoundHandle)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}

	//Mark the sound as 'paused'
	SetPlayingState(ASURA_ON_DEMAND_STATE_STOPPED);
	
	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}

	//This sound is not prepared, queue the action
	Asura_Xbox360_OnDemandQueuedAction* pxAction =  NextFreeActionQueueItem();
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_PAUSE;

		pxAction->m_iSoundID			= iSoundHandle;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueResume	(const int iSoundHandle)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");

	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}

	//Mark the sound as 'playing'
	SetPlayingState(ASURA_ON_DEMAND_STATE_PLAYING);

	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	
	//This sound is not prepared, queue the action
	Asura_Xbox360_OnDemandQueuedAction* pxAction = NextFreeActionQueueItem();
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_RESUME;

		pxAction->m_iSoundID		= iSoundHandle;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueSetOcclusionLPF	(const int iSoundHandle, const float fOcclusion)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");

	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}

	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}

	//This sound is not prepared, queue the action
	//this is a case where we can replace an existing action of the same type if possible
	Asura_Xbox360_OnDemandQueuedAction* pxAction = GetFirstActionQueueItemByType(ASURA_ON_DEMAND_QUEUE_ACTION_OCCLUSION_LPF);
	if(pxAction)
	{
		pxAction->m_uIdentifier			= ASURA_ON_DEMAND_QUEUE_ACTION_OCCLUSION_LPF;

		pxAction->m_iSoundID			= iSoundHandle;
		pxAction->m_xData.m_fOcclusion	= fOcclusion;
	}
	return false;	//dont continue in the platform function
}


void Asura_Xbox360_OnDemandSound::Destroy	(const int iSoundHandle)
{
	Asura_Assert(GetState() != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	
	//We need to make sure that we are loaded before unloading 
	BusyWaitUntilPrepared();
	RemoveFromOnDemandSoundUpdates(this);
	Clear();
}


bool Asura_Xbox360_OnDemandSound::QueueSetPitch	(const int iSoundHandle, const float fPitch)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}
	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	//This sound is not prepared, queue the action
	//this is a case where we can replace an existing action of the same type if possible
	Asura_Xbox360_OnDemandQueuedAction* pxAction = GetFirstActionQueueItemByType(ASURA_ON_DEMAND_QUEUE_ACTION_PITCH); 
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_PITCH;

		pxAction->m_iSoundID			= iSoundHandle;
		pxAction->m_xData.m_fPitch		= fPitch;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueSetUsesReverb	(const int iSoundHandle, const bool bUseReverb)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}
	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	//This sound is not prepared, queue the action
	//this is a case where we can replace an existing action of the same type if possible
	Asura_Xbox360_OnDemandQueuedAction* pxAction = GetFirstActionQueueItemByType(ASURA_ON_DEMAND_QUEUE_ACTION_USES_REVERB); 
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_USES_REVERB;

		pxAction->m_iSoundID			= iSoundHandle;
		pxAction->m_xData.m_bUseReverb	= bUseReverb;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueSetVolumePanAndFade	(const int iSoundHandle, const float fVolume, const float fPan, const float fFade)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}
	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	//This sound is not prepared, queue the action
	//this is a case where we can replace an existing action of the same type if possible
	Asura_Xbox360_OnDemandQueuedAction* pxAction = GetFirstActionQueueItemByType(ASURA_ON_DEMAND_QUEUE_ACTION_VOLUME_PAN_FADE); 

	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_VOLUME_PAN_FADE;

		pxAction->m_iSoundID						= iSoundHandle;
		pxAction->m_xData.m_xVolPanFade.m_fVolume	= fVolume;
		pxAction->m_xData.m_xVolPanFade.m_fPan		= fPan;
		pxAction->m_xData.m_xVolPanFade.m_fFade		= fFade;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueSetDistFilterGain	(int iSoundHandle, float fGain)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}
	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	//This sound is not prepared, queue the action
	//this is a case where we can replace an existing action of the same type if possible
	Asura_Xbox360_OnDemandQueuedAction* pxAction = GetFirstActionQueueItemByType(ASURA_ON_DEMAND_QUEUE_ACTION_DIST_FILTER_GAIN); 
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_DIST_FILTER_GAIN;

		pxAction->m_iSoundID			= iSoundHandle;
		pxAction->m_xData.m_fGain		= fGain;
	}
	return false;	//dont continue in the platform function
}


bool Asura_Xbox360_OnDemandSound::QueueSetWetDryVolume		(const int iSoundHandle, const float fWetVolume, float fDryVolume)
{
	const u_int uState = GetState();
	Asura_Assert(uState != ASURA_ON_DEMAND_STATE_VOID, "Called a function on an unloaded on-demand sound");
	if(uState & ASURA_ON_DEMAND_STATE_PREPARE_FAILED)
	{
		//This sound was not prepared, return and do not continue in platform function
		return false;
	}
	if(uState & ASURA_ON_DEMAND_STATE_PREPARED)
	{
		//This sound is ready, just return
		return true;
	}
	//This sound is not prepared, queue the action
	//this is a case where we can replace an existing action of the same type if possible
	Asura_Xbox360_OnDemandQueuedAction* pxAction = GetFirstActionQueueItemByType(ASURA_ON_DEMAND_QUEUE_ACTION_WET_DRY_VOLUME); 
	if(pxAction)
	{
		pxAction->m_uIdentifier		= ASURA_ON_DEMAND_QUEUE_ACTION_WET_DRY_VOLUME;

		pxAction->m_iSoundID						= iSoundHandle;
		pxAction->m_xData.m_xWetDry.m_fWetVolume	= fWetVolume;
		pxAction->m_xData.m_xWetDry.m_fDryVolume	= fDryVolume;
	}
	return false;	//dont continue in the platform function
}


void Asura_Xbox360_OnDemandSound::Unload()
{
	//We need to make sure that we are loaded before unloading 
	BusyWaitUntilPrepared();
	
	RemoveFromOnDemandSoundUpdates(this);
	Clear();
	SetAsNotOnDemand();
}


Asura_Xbox360_OnDemandQueuedAction*	Asura_Xbox360_OnDemandSound::PopQueuedAction	()
{
	const u_int uState = GetState();

	if(uState & ASURA_ON_DEMAND_STATE_PREPARING)
	{
		//Busy updating, dont allow access to queued items
		return 0;
	}

	if(m_uActionQueueStart == m_uActionQueueEnd)
	{
		//Queue is empty
		return 0;
	}

	Asura_Xbox360_OnDemandQueuedAction* pxAction = &m_axActionQueue[m_uActionQueueStart];
	m_uActionQueueStart = (m_uActionQueueStart + 1) % XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_ACTIONS;
	return pxAction;
}


Asura_Xbox360_OnDemandQueuedAction* Asura_Xbox360_OnDemandSound::NextFreeActionQueueItem()
{
	//Return the next free space in the queue of actions
	if((m_uActionQueueEnd + 1) % XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_ACTIONS == m_uActionQueueStart)
	{
		Asura_Assert(false,"Out of space in queue for on-demand sound action");
		return 0;
	}

	Asura_Xbox360_OnDemandQueuedAction* pxAction = &m_axActionQueue[m_uActionQueueEnd];
	m_uActionQueueEnd = (m_uActionQueueEnd + 1) % XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_ACTIONS;
	return pxAction;
}


Asura_Xbox360_OnDemandQueuedAction* Asura_Xbox360_OnDemandSound::GetFirstActionQueueItemByType	(u_int uIdentifier)
{
	//Sometimes we can just replace a queued action, when it is something like change volume
	//in which case we call this function to get a queue item we can override (or the next free slot if not)
	for(u_int u = m_uActionQueueStart; u != m_uActionQueueEnd; u = (u + 1) % XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_ACTIONS )
	{
		if(m_axActionQueue[u].m_uIdentifier == uIdentifier)
		{
			return &m_axActionQueue[u];
		}
	}
	//Didnt find a queue item with this id, just return the first new free slot in the queue
	return NextFreeActionQueueItem();
}


void Asura_Xbox360_OnDemandSound::BusyWaitUntilPrepared () const
{
	u_int uMaxIterations = 256;
	const u_int uSleepTime = 25;

	while( (GetState() & ASURA_ON_DEMAND_STATE_PREPARING ) && uMaxIterations )
	{
		Sleep(uSleepTime);
		-- uMaxIterations;
	}
	Asura_Assert(uMaxIterations,"Timed out waiting for on demand sound to load");
}


/******************************************************************************/
/* Global Variables part 2                                                    */
/******************************************************************************/

#ifndef ASURA_USE_OPENAL
	static Asura_FixedArray< SourceBuffer,	ASURA_MAX_NUMBER_OF_LOADED_SOUNDS>	g_axBuffers;
	static Asura_FixedArray< SourceVoice,	ASURA_XBOX360_SOUND_MAX_VOICES>		g_axVoices;

	static IXAudioSubmixVoice* g_pxReverbSubmixVoice = NULL;

	DVSVAR bool cnsl_bDistanceFilter = true;
#endif

/******************************************************************************/
/* Local function declarations                                                */
/******************************************************************************/
	
static void ThreadedSoundLoading (void* pData, u_int uDataSize)
{
	//Ugly const cast (why does the thread function have to pass a const argument?!)
	Asura_Xbox360_OnDemandSound* pxSound = const_cast<Asura_Xbox360_OnDemandSound*>(reinterpret_cast<const Asura_Xbox360_OnDemandSound*>(pData));
	if(pxSound)
	{
		if(pxSound->Load())
		{
			//We are now ready for a call to Update() to clear our queue
			//It is important that this state (ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE) is not the same as 'prepared'
			//we need to ensure that this sound's queue is cleared down before it can be safely treated as 'prepared'
			pxSound->SetPreparedState(ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE);	
			AddToOnDemandSoundUpdates(pxSound);
		}
		else
		{
			Asura_Assert(false,("Failed to prepare on demand sound in %s",pxSound->GetFilename()));
			pxSound->SetPreparedState(ASURA_ON_DEMAND_STATE_PREPARE_FAILED);
		}
	}
}

void Asura_Xbox360_LoadChunkedSoundData (Asura_Chunk_Stream& xStream)
{
	int iSoundID;
	int iOffset;
	int iSize;
	int iNumSounds;
	int iTotalSkip = 0;

	xStream >> iNumSounds;

	//get the filename with some rather unpleasant casting
	Asura_Chunk_File_Stream* pxFileStream = reinterpret_cast<Asura_Chunk_File_Stream*>(&xStream);
	const char* szFilename = Asura_FileAccess::GetFilename(pxFileStream->GetFileHandle());

	Asura_String xFullFile;
	if(strncmp(szFilename,"game:",strlen(szFilename))!=0)
	{
		xFullFile = "game:\\";
		xFullFile += szFilename;
	}
	else
	{
		xFullFile = szFilename;
	}

	//Allocate a pool of on demand sounds
	Asura_Xbox360_OnDemandSound** pxOnDemandSoundPool = &g_apxOnDemandSoundPool_Global;
	if(g_bHitResetPoint)
	{
		//If we are past the reset point start using the resettable pool
		pxOnDemandSoundPool = &g_apxOnDemandSoundPool_Resettable;
	}
	Asura_Assert(!*pxOnDemandSoundPool,"We have already loaded some on demand sounds and not cleared down our associated buffer");
	*pxOnDemandSoundPool = new Asura_Xbox360_OnDemandSound[iNumSounds];
	Asura_Assertf(*pxOnDemandSoundPool,("Failed to allocate %d bytes for on demand sounds, they will not be loaded", (sizeof(Asura_Xbox360_OnDemandSound) * iNumSounds)));
	if(*pxOnDemandSoundPool)
	{
		for(int i=0;i<iNumSounds;++i)
		{
			xStream >> iSoundID;
			xStream >> iOffset;
			xStream >> iSize;

			iTotalSkip += iSize;

			int iLoadedIndex = Asura_Sound_System::SoundLoadedByPlatformCode(ASURA_SID_DYNAMIC, iSoundID);

			if( (iLoadedIndex >= 0) && (iLoadedIndex < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS))
			{
				//Point this on demand loaded sound to the appropriate place in our pool of sounds
				g_apxOnDemandSounds[iLoadedIndex] = &(*pxOnDemandSoundPool)[i];
				Asura_Xbox360_OnDemandSound* pxCurrentSound = g_apxOnDemandSounds[iLoadedIndex];

				pxCurrentSound->SetAsOnDemand();
				pxCurrentSound->SetAsResettable(g_bHitResetPoint);
				pxCurrentSound->SetBaseFileOffset(iOffset);
				pxCurrentSound->SetFilename(xFullFile.GetString());
				pxCurrentSound->SetFileSize(iSize);
			}
			else
			{
				Asura_Assertf(false,("Invalid loaded sound index"));
				continue;
			}
		}
	}

	xStream.Skip(iTotalSkip);
}


static void AddToOnDemandSoundUpdates		(Asura_Xbox360_OnDemandSound* pxSound)
{
	g_OnDemandUpdatesMutex.Lock();
	
	//Add this sound to our list of sounds to update
	u_int uCount = 0;
	while(g_apxOnDemandUpdates[uCount])
	{
		if(g_apxOnDemandUpdates[uCount] == pxSound)
		{
			//This sound is already listed to update
			g_OnDemandUpdatesMutex.Unlock();
			return;
		}
		
		++ uCount;

		if(uCount == XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE)
		{
			Asura_Assert(false,"Too many on-demand sounds to update this frame");
			g_OnDemandUpdatesMutex.Unlock();
			return;
		}
	}
	g_apxOnDemandUpdates[uCount] = pxSound;
	g_OnDemandUpdatesMutex.Unlock();
}

static Asura_Xbox360_OnDemandSound* PopNextOnDemandSoundToUpdate	()
{
	g_OnDemandUpdatesMutex.Lock();
	Asura_Xbox360_OnDemandSound* pRet = g_apxOnDemandUpdates[0];
	if(pRet)
	{
		memmove(&g_apxOnDemandUpdates[0],&g_apxOnDemandUpdates[1],sizeof(Asura_Xbox360_OnDemandSound*) * (XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE - 1));
		g_apxOnDemandUpdates[XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE - 1] = 0;
	}
	g_OnDemandUpdatesMutex.Unlock();
	return pRet;
}

static void RemoveFromOnDemandSoundUpdates	(const Asura_Xbox360_OnDemandSound* pxSound)
{
	g_OnDemandUpdatesMutex.Lock();

	//remove this sound from our list of sounds to update, we assume there is only one instance in the list
	u_int uCount = 0;
	while(g_apxOnDemandUpdates[uCount])
	{
		if(g_apxOnDemandUpdates[uCount] == pxSound)
		{
			//found it
			if(uCount == (XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE - 1))
			{
				g_apxOnDemandUpdates[uCount] = 0;
			}
			else
			{
				memmove(&g_apxOnDemandUpdates[uCount],&g_apxOnDemandUpdates[uCount+1],sizeof(Asura_Xbox360_OnDemandSound*) * ((XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE - uCount) - 1));
			}
			break;
		}

		++ uCount;

		if(uCount == XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE)
		{
			break;
		}
	}
	g_OnDemandUpdatesMutex.Unlock();
}

static bool UpdateVoiceWetDryVolumes (SourceVoice* pxVoice)
{
	// This maths came from the Xbox code...
	// TY 12.12.07 - modify the volumes with the dry/wet volumes
	// TY 04.02.08 - set the volumes differently depending on wether or not we have a dry submix

	// The XAudio voice may not have been created yet; we can only do that at playback time
	if( !pxVoice->m_pxXAudioSourceVoice )
	{
		//No voice to update, this isnt an error, it just means we arent prepared
		return true;
	}

	//-------- set the volume for the dry submix (or master output if not mono) ----------

	if(pxVoice->m_pxXAudioDrySubmix)
	{
		XAUDIOVOICEOUTPUTVOLUME			xOutputVolume;
		XAUDIOVOICEOUTPUTVOLUMEENTRY	axVolumeEntries[1];
		XAUDIOCHANNELVOLUME				xDrySubmixVolume;
		XAUDIOCHANNELVOLUMEENTRY		axDrySubmixSpeakerVolumes[6];

		Asura_Assertf( ( pxVoice->m_fPan >= 0.0f )  && ( pxVoice->m_fPan <= 1.0f ),  ( "Voice Pan value out of range: %.2f",  pxVoice->m_fPan  ) );
		Asura_Assertf( ( pxVoice->m_fFade >= 0.0f ) && ( pxVoice->m_fFade <= 1.0f ), ( "Voice Fade value out of range: %.2f", pxVoice->m_fFade ) );

		const float fMainVolume = pxVoice->m_fDryVolume * pxVoice->m_fVolume;

		const float fLeftVol	= Asura_Maths::Sqrt( 1.0f - pxVoice->m_fPan );
		const float fRightVol	= Asura_Maths::Sqrt( pxVoice->m_fPan );

		const float fFrontVol	= Asura_Maths::Sqrt(pxVoice->m_fFade);
		const float fRearVol	= Asura_Maths::Sqrt(1.0f - pxVoice->m_fFade);

		//TY 19.02.08 - We dont usually output through the Front centre speaker but 360 seems to expect it (i.e. sounds wrong without it)

		axDrySubmixSpeakerVolumes[0].Volume =	fMainVolume * fFrontVol * fLeftVol;		// Front left
		axDrySubmixSpeakerVolumes[1].Volume =	fMainVolume * fFrontVol * fRightVol;	// Front right
		axDrySubmixSpeakerVolumes[2].Volume =	fMainVolume * fFrontVol;				// Front centre
		axDrySubmixSpeakerVolumes[3].Volume =	0.0f;									// LFE
		axDrySubmixSpeakerVolumes[4].Volume =	fMainVolume * fRearVol * fLeftVol;		// Rear left
		axDrySubmixSpeakerVolumes[5].Volume =	fMainVolume * fRearVol * fRightVol;		// Rear left		

		for( u_int u = 0; u < 6; ++u )
		{
			axDrySubmixSpeakerVolumes[u].EntryIndex = static_cast< XAUDIOCHANNEL >(u);
		}

		xDrySubmixVolume.EntryCount = 6;
		xDrySubmixVolume.paEntries = axDrySubmixSpeakerVolumes;

		axVolumeEntries[0].OutputVoiceIndex = 0;
		axVolumeEntries[0].pVolume = &xDrySubmixVolume;

		xOutputVolume.EntryCount = 1;
		xOutputVolume.paEntries = axVolumeEntries;

		//Set volume
		const HRESULT xOutputResult = pxVoice->m_pxXAudioDrySubmix->SetVoiceOutputVolume( &xOutputVolume );
		if(FAILED(xOutputResult))
		{
			return false;
		}
	}
	
	//-------- set the volume for the source voice (outputting to Dry Submix/Master + Reverb Submix) ----------
	
	XAUDIOVOICEOUTPUTVOLUME			xOutputVolume;
	XAUDIOVOICEOUTPUTVOLUMEENTRY	axVolumeEntries[2];
	XAUDIOCHANNELVOLUME				xDryChannelVolume;
	XAUDIOCHANNELVOLUME				xReverbChannelVolume;
	XAUDIOCHANNELVOLUMEENTRY		xDryChannelSpeakerVolume[XAUDIOSPEAKER_COUNT];
	XAUDIOCHANNELVOLUMEENTRY		xReverbChannelVolumeEntry;

	//-------- set the volume for the dry path ----------

	if(pxVoice->m_pxXAudioDrySubmix)
	{
		//If we have a dry submix, we just pass through (the dry submix does all of the necessary volume adjustment)
		xDryChannelSpeakerVolume[0].EntryIndex = 0;
		xDryChannelSpeakerVolume[0].Volume =  1.0f;

		xDryChannelVolume.EntryCount = 1;
		xDryChannelVolume.paEntries = &xDryChannelSpeakerVolume[0];
	}
	else
	{
		//If we are passing straight to the master, set all the channels
		for(u_int u = 0; u < XAUDIOSPEAKER_COUNT; ++u )
		{
			xDryChannelSpeakerVolume[u].EntryIndex = static_cast<XAUDIOCHANNEL>(u);
			xDryChannelSpeakerVolume[u].Volume = pxVoice->m_fVolume * pxVoice->m_fDryVolume; // dry volume scaled by main volume
		}

		xDryChannelVolume.EntryCount = XAUDIOSPEAKER_COUNT;
		xDryChannelVolume.paEntries = &xDryChannelSpeakerVolume[0];
	}

	//set the 'volume entry' for the dry path
	axVolumeEntries[0].OutputVoiceIndex = 0;
	axVolumeEntries[0].pVolume = &xDryChannelVolume;
	
	//-------- set the volume for the reverb submix ----------

	if(pxVoice->m_bReverb)
	{
		xReverbChannelVolumeEntry.EntryIndex = 0;
		xReverbChannelVolumeEntry.Volume = pxVoice->m_fVolume * pxVoice->m_fWetVolume;	// wet volume scaled by main volume

		xReverbChannelVolume.EntryCount = 1;
		xReverbChannelVolume.paEntries = &xReverbChannelVolumeEntry;

		axVolumeEntries[1].OutputVoiceIndex = 1;
		axVolumeEntries[1].pVolume = &xReverbChannelVolume;
	}
	
	//--------- pass the values to xaudio -------------

	if(pxVoice->m_bReverb)
	{
		xOutputVolume.EntryCount = 2;
	}
	else
	{
		xOutputVolume.EntryCount = 1;
	}
	xOutputVolume.paEntries = axVolumeEntries;

	//Set volume
	const HRESULT xOutputResult = pxVoice->m_pxXAudioSourceVoice->SetVoiceOutputVolume( &xOutputVolume );
	if(FAILED(xOutputResult))
	{
		return false;
	}
	return true;
}

/******************************************************************************/
/* RIFFChunkHeader Class Functions                                            */
/******************************************************************************/

bool RIFFChunkHeader::ReadFromFile( const Asura_Handle iFileHandle )
{
	if( Asura_FileAccess::ReadData( iFileHandle, this, sizeof( *this ) ) )
	{
		m_uDataSize = __loadwordbytereverse( 0, &m_uDataSize );
		return true;
	}
	else
	{
		return false;
	}
}

bool RIFFChunkHeader::ReadFromFile( const Asura_Handle iFileHandle, const u_int uExpectedChunkType )
{
	bool success = ReadFromFile( iFileHandle );
	Asura_Assert( m_uChunkType == uExpectedChunkType, "Invalid audio chunk." );
	return success;
}

/******************************************************************************/
/* SourceVoice Class Functions                                                */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// SourceVoice::SourceVoice()
//
// Contructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

#ifndef ASURA_USE_OPENAL
SourceVoice::SourceVoice()
	: m_fVolume( 1.0f )
	, m_fPan( 0.5f )
	, m_fFade( 0.5f )
	, m_fPitch( 1.0f )
	, m_bReverb( false )
	, m_fDryVolume ( 1.0f )
	, m_fWetVolume ( 1.0f )
	, m_fOcclusionFilter ( 0.0f )
	, m_iOnDemandIndex( ASURA_INVALID_HANDLE )
	, m_pxXAudioSourceVoice ( NULL )
	, m_pxXAudioDrySubmix ( NULL )
{
	m_bPaused = m_bPausedByPitch = false;
}
#endif

/******************************************************************************/
/* Asura_Sound_System Static Class Functions                                  */
/******************************************************************************/

/******************************************************************************/
/* System Basics                                                              */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformInitialise()
//
// Initialises the sound system on Xbox 360
//
// Parameters:		None
// Return value:	Maximum number of simultaneous playing voices
//
////////////////////////////////////////////////////////////////////////////////

int Asura_Sound_System::PlatformInitialise()
{
	#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_AUDIO_DISTANCE_SCALE,  &g_fDistanceFilterScale,  0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_AUDIO_OCCLUSION_SCALE, &g_fOcclusionFilterScale, 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif	

	#if XBOX360_SOUND_XAUDIO_DEBUG_HIGH
		XAudioSetDebugPrintLevel( XAUDIODEBUGLEVEL_INFO );
	#endif

	#ifdef ASURA_USE_OPENAL
		
		#if defined ASURA_ASSERT
			bool bResult =
		#endif
		Asura_Xbox360_Sound::GetAPI().Initialise();

		Asura_Assert( bResult, "Failed to initialise OpenAL audio!" );

		#if XBOX360_SOUND_USE_HARDWARE_PAN
			s_bUseHardware3dPan = true;
		#endif

	#else
		// Initialise XAudio
		{
			XAUDIOENGINEINIT xEngineInitParams = {0};

			xEngineInitParams.pEffectTable = &xEFFECT_TABLE;
			xEngineInitParams.SubmixStageCount = 1;
			xEngineInitParams.ThreadUsage = XAUDIOTHREADUSAGE_THREAD4 | XAUDIOTHREADUSAGE_THREAD5;

			#if defined ASURA_ASSERT
				const HRESULT xResult =
			#endif
				XAudioInitialize( &xEngineInitParams );

			Asura_Assert( SUCCEEDED( xResult ), "Failed to initialise XAudio." );
		}

		// Create a submix voice using the reverb effect
		{
			// Route all output channels straight through to the mastering voice

			XAUDIOCHANNELMAPENTRY axSubmixChannelMapEntries[6];
			for( XAUDIOCHANNEL u = 0; u < 6; ++u )
			{
				axSubmixChannelMapEntries[u].InputChannel = u;
				axSubmixChannelMapEntries[u].OutputChannel = u;
				axSubmixChannelMapEntries[u].Volume = 1.0f;
			}

			XAUDIOCHANNELMAP xSubmixChannelMap;
			xSubmixChannelMap.EntryCount = 6;
			xSubmixChannelMap.paEntries = axSubmixChannelMapEntries;

			XAUDIOVOICEOUTPUTENTRY xSubmixOutputEntry;
			xSubmixOutputEntry.pDestVoice = NULL; // Output to mastering voice
			xSubmixOutputEntry.pChannelMap = &xSubmixChannelMap;

			XAUDIOVOICEOUTPUT xSubmixOutput;
			xSubmixOutput.EntryCount = 1;
			xSubmixOutput.paEntries = &xSubmixOutputEntry;

			XAUDIOSUBMIXVOICEINIT xSubmixVoiceInit = {0};

			// Reverb only works with mono or stereo inputs - we'll use mono only
			xSubmixVoiceInit.Format.ChannelCount = 1;
			xSubmixVoiceInit.Format.SampleType = XAUDIOSAMPLETYPE_NATIVE;
			xSubmixVoiceInit.Format.SampleRate = XAUDIOSAMPLERATE_NATIVE;

			xSubmixVoiceInit.MaxOutputVoiceCount = 1;
			xSubmixVoiceInit.MaxChannelMapEntryCount = XAUDIOSPEAKER_COUNT;
			xSubmixVoiceInit.SubmixStage = 0;
			xSubmixVoiceInit.pEffectChain = &xEFFECT_CHAIN;
			xSubmixVoiceInit.pVoiceOutput = &xSubmixOutput;
			xSubmixVoiceInit.pfnProcessCallback = NULL;
			xSubmixVoiceInit.pContext = NULL;

			#if defined ASURA_ASSERT
				const HRESULT xResult =
			#endif
				XAudioCreateSubmixVoice( &xSubmixVoiceInit, &g_pxReverbSubmixVoice );

			Asura_Assert( SUCCEEDED( xResult ) && g_pxReverbSubmixVoice,
				"Failed to create reverb submix voice." );
		}

		#if defined ASURA_DEBUG_VARIABLE_SYSTEM
			Asura_CommandConsole::AddVar( L"Audio.Distance.UseDistanceFilter", &cnsl_bDistanceFilter, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif

		//Initialise our buffer for sounds loaded on demand
		if(XBOX360_SOUND_ON_DEMAND_SOUNDS_MAXIMUM_ALLOCATION_SIZE)
		{
			g_pcOnDemandBuffer = reinterpret_cast<char*>(XPhysicalAlloc( XBOX360_SOUND_ON_DEMAND_SOUNDS_MAXIMUM_ALLOCATION_SIZE, MAXULONG_PTR, 0, PAGE_READWRITE ));
			g_pxOnDemandAllocator = new Asura_MemAllocator(g_pcOnDemandBuffer, XBOX360_SOUND_ON_DEMAND_SOUNDS_MAXIMUM_ALLOCATION_SIZE, 4, XBOX360_SOUND_ON_DEMAND_SOUNDS_MINIMUM_ALLOCATION_SIZE);
		}

		//reset our sounds to update list
		for(u_int u = 0; u < XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_SOUNDS_TO_UPDATE ; ++u)
		{
			g_apxOnDemandUpdates[u] = 0;
		}

		Asura_Mem::Set(g_apxOnDemandSounds, 0, sizeof(Asura_Xbox360_OnDemandSound*) * ASURA_MAX_NUMBER_OF_LOADED_SOUNDS);
		Asura_Assert(!g_apxOnDemandSoundPool_Global, "Global on demand sound pool is not null");
		Asura_Assert(!g_apxOnDemandSoundPool_Resettable, "Resettable on demand sound pool is not null");

		g_pxOnDemandSoundQueue = new Asura_ProducerConsumerQueue(XBOX360_SOUND_ON_DEMAND_SOUNDS_MAX_QUEUED_PREPARATIONS, XBOX360_SOUND_ON_DEMAND_SOUNDS_PREPARATION_THREADS, ASURA_MULTITHREAD_STACK_SIZE_SMALL);

	#endif

	Asura_Xbox360_XACT::Initialise();

	Platform_SetReverb( ASURA_REVERB_ID_NONE, 1.0f );

	return ASURA_XBOX360_SOUND_MAX_VOICES;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformReset()
//
// Resets the sound system on Xbox 360. Called between level loads
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Sound_System::PlatformReset()
{
	for(u_int u = 0; u < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS; ++u)
	{
		if(g_apxOnDemandSounds[u] && g_apxOnDemandSounds[u]->IsResettable())
		{
			g_apxOnDemandSounds[u] = 0;
		}
	}
	delete []g_apxOnDemandSoundPool_Resettable;
	g_apxOnDemandSoundPool_Resettable = 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformShutdown()
//
// Shuts down the sound system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Sound_System::PlatformShutdown()
{
	Asura_Xbox360_XACT::Shutdown();

#ifdef ASURA_USE_OPENAL

	Asura_Xbox360_Sound::GetAPI().Destroy();

#else
	
	g_pxReverbSubmixVoice->Release();
	g_pxReverbSubmixVoice = NULL;

	// Shut down XAudio
	#if defined ASURA_ASSERT
	const ULONG ulRefCount = XAudioShutDown();
	#else
	XAudioShutDown();	
	#endif
	
	Asura_Assert( ulRefCount == 0, "XAudio reference count not zero." );

	delete g_pxOnDemandSoundQueue;
	g_pxOnDemandSoundQueue = 0;

	XPhysicalFree(g_pcOnDemandBuffer);
	delete g_pxOnDemandAllocator;

	Asura_Mem::Set(g_apxOnDemandSounds, 0, sizeof(Asura_Xbox360_OnDemandSound*) * ASURA_MAX_NUMBER_OF_LOADED_SOUNDS);
	
	delete []g_apxOnDemandSoundPool_Global;
	delete []g_apxOnDemandSoundPool_Resettable;
	
	g_apxOnDemandSoundPool_Global		= 0;
	g_apxOnDemandSoundPool_Resettable	= 0;
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformUpdate()
//
// Updates the sound system on Xbox 360. Called once per frame
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Sound_System::PlatformUpdate()
{
#ifdef ASURA_USE_OPENAL

	#ifdef XBOX360_SOUND_USE_HARDWARE_PAN
		Asura_Xbox360_Sound::GetAPI().SetListener( s_axSoundCentre[0], s_axSoundOrientation[0] );
	#endif

	Asura_Xbox360_Sound::GetAPI().Update();

	TraversePlayingSoundHandles( Asura_Xbox360_Sound::UpdateSound );
#endif

	Asura_Xbox360_XACT::Update();

	//Loop over all of our on-demand sounds and execute any queued actions
	Asura_Xbox360_OnDemandSound *pxSound = 0;
	while((pxSound = PopNextOnDemandSoundToUpdate()) != 0)
	{
		Asura_Assert(pxSound->IsOnDemandSound(), "Updating a not on-demand sound as on-demand");
		
		//Loop over and execute any prepared actions
		const u_int uState = pxSound->GetState();

		Asura_Assert((uState & ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE) || (uState & ASURA_ON_DEMAND_STATE_PREPARED),"Updating an on-demand sound in the wrong state");
		
		//If we were waiting for an update, mark this sound as now fully prepared (before we start calling platform functions)
		if(uState & ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE)
		{
			pxSound->SetPreparedState(ASURA_ON_DEMAND_STATE_PREPARED);
		}
		
		Asura_Xbox360_OnDemandQueuedAction* pxAction = 0;
		while((pxAction = pxSound->PopQueuedAction()) != 0)
		{
			#ifdef ASURA_ASSERT
			bool bRet = false;
			#endif
			switch(pxAction->m_uIdentifier)
			{
				case ASURA_ON_DEMAND_QUEUE_ACTION_PLAY:				
				{
					#ifdef ASURA_ASSERT
					bRet = PlatformPlaySound (pxAction->m_iSoundID,pxAction->m_xData.m_bLooping);
					#else
					PlatformPlaySound (pxAction->m_iSoundID,pxAction->m_xData.m_bLooping);
					#endif
				}
				break;
				
				case ASURA_ON_DEMAND_QUEUE_ACTION_PAUSE:			
				{
					#ifdef ASURA_ASSERT
					bRet = PlatformPauseSound (pxAction->m_iSoundID);	
					#else
					PlatformPauseSound (pxAction->m_iSoundID);
					#endif
				}
				break;
				
				case ASURA_ON_DEMAND_QUEUE_ACTION_RESUME:
				{
					#ifdef ASURA_ASSERT
					bRet = PlatformUnpauseSound	(pxAction->m_iSoundID);
					#else
					PlatformUnpauseSound	(pxAction->m_iSoundID);
					#endif
				}
				break;
				
				case ASURA_ON_DEMAND_QUEUE_ACTION_STOP:	
				{
					#ifdef ASURA_ASSERT
					bRet = PlatformStopSound (pxAction->m_iSoundID); 
					#else
					PlatformStopSound (pxAction->m_iSoundID); 
					#endif
				}
				break;
				
				case ASURA_ON_DEMAND_QUEUE_ACTION_PITCH:			
				{
					#ifdef ASURA_ASSERT
					bRet = PlatformSetSoundPitch (pxAction->m_iSoundID, pxAction->m_xData.m_fPitch);
					#else
					PlatformSetSoundPitch (pxAction->m_iSoundID, pxAction->m_xData.m_fPitch);
					#endif
				}
				break;
								
				case ASURA_ON_DEMAND_QUEUE_ACTION_USES_REVERB:		
				{
					#ifdef ASURA_ASSERT
					bRet = Platform_SetSoundUsesReverb (pxAction->m_iSoundID, pxAction->m_xData.m_bUseReverb);
					#else
					Platform_SetSoundUsesReverb (pxAction->m_iSoundID, pxAction->m_xData.m_bUseReverb);
					#endif
				}
				break;

				case ASURA_ON_DEMAND_QUEUE_ACTION_VOLUME_PAN_FADE:
				{
					#ifdef ASURA_ASSERT
					bRet = PlatformSetSoundVolumePanAndFade (	pxAction->m_iSoundID, 
																pxAction->m_xData.m_xVolPanFade.m_fVolume, 
																pxAction->m_xData.m_xVolPanFade.m_fPan,
																pxAction->m_xData.m_xVolPanFade.m_fFade );
					#else
					PlatformSetSoundVolumePanAndFade (	pxAction->m_iSoundID, 
														pxAction->m_xData.m_xVolPanFade.m_fVolume, 
														pxAction->m_xData.m_xVolPanFade.m_fPan,
														pxAction->m_xData.m_xVolPanFade.m_fFade );
					#endif
				}
				break;

				case ASURA_ON_DEMAND_QUEUE_ACTION_DIST_FILTER_GAIN:
				{
					#ifdef ASURA_ASSERT
					bRet = Platform_SetSoundDistanceFilterGain(pxAction->m_iSoundID, pxAction->m_xData.m_fGain);
					#else
					Platform_SetSoundDistanceFilterGain(pxAction->m_iSoundID, pxAction->m_xData.m_fGain);
					#endif
				}
				break;

				case ASURA_ON_DEMAND_QUEUE_ACTION_WET_DRY_VOLUME:
				{
					#ifdef ASURA_ASSERT
					bRet = Platform_SetSoundWetDryVolume(pxAction->m_iSoundID, pxAction->m_xData.m_xWetDry.m_fWetVolume, pxAction->m_xData.m_xWetDry.m_fDryVolume);
					#else
					Platform_SetSoundWetDryVolume(pxAction->m_iSoundID, pxAction->m_xData.m_xWetDry.m_fWetVolume, pxAction->m_xData.m_xWetDry.m_fDryVolume);
					#endif
				}
				break;

				case ASURA_ON_DEMAND_QUEUE_ACTION_OCCLUSION_LPF:
				{
					#ifdef ASURA_ASSERT
					bRet = Platform_SetOcclusionLPF(pxAction->m_iSoundID, pxAction->m_xData.m_fOcclusion);
					#else
					Platform_SetOcclusionLPF(pxAction->m_iSoundID, pxAction->m_xData.m_fOcclusion);
					#endif
				}
				break;

				default:
				{
					Asura_Assertf(false,("Unrecognised on-demand sound queued action ID: %u", pxAction->m_uIdentifier));
				}
			}
			Asura_Assertf(bRet,("Platform Sound function failed (queued action id: %u)",pxAction->m_uIdentifier));
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformSetResetPoint()
//
// Sets a reset point for tracking memory usage. Not implemented on 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Sound_System::PlatformSetResetPoint()
{
	g_bHitResetPoint = true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::Platform_GetMaxSoundsWithEQ()
//
// Queries the maximum number of playing voices that may have EQ effects applied
//
// Parameters:		None
// Return value:	The maximum number of EQ-enabled voices
//
////////////////////////////////////////////////////////////////////////////////

int Asura_Sound_System::Platform_GetMaxSoundsWithEQ()
{
	return ASURA_XBOX360_SOUND_MAX_VOICES;
}

/******************************************************************************/
/* Sample Management                                                          */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformLoadSound()
//
// Loads a sample from an individual sound file. Shouldn't be called on Xbox 360
//
// Parameters:		szFilename - Name of file to load
//					iLoadedSoundIndex - Corresponding engine sound index
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformLoadSound( const char* szFilename, const int iLoadedSoundIndex )
{
	Asura_Assert( false, "Shouldn't be loading individual sound files on Xbox 360." );
	return false;
}

#ifndef ASURA_USE_OPENAL
static bool ParseChunk( const Asura_Handle iFileHandle, SourceBuffer& xNewBuffer )
{
	RIFFChunkHeader xChunkHeader;
	Asura_FileAccess::ReadData( iFileHandle, &xChunkHeader, sizeof( xChunkHeader ) );
	xChunkHeader.m_uDataSize = __loadwordbytereverse( 0, &xChunkHeader.m_uDataSize );

	switch( xChunkHeader.m_uChunkType )
	{
		case 'data':
		{
			xNewBuffer.m_xXAudioSourceBuffer.pBuffer = XPhysicalAlloc( xChunkHeader.m_uDataSize, MAXULONG_PTR, 0, PAGE_READWRITE );
			if( !xNewBuffer.m_xXAudioSourceBuffer.pBuffer )
			{
				Asura_Assert( false, "Failed to allocate memory for sound." );
				return false;
			}

			// Don't endian-swap this
			Asura_FileAccess::ReadData( iFileHandle, xNewBuffer.m_xXAudioSourceBuffer.pBuffer, xChunkHeader.m_uDataSize );

			xNewBuffer.m_xXAudioSourceBuffer.BufferSize = xChunkHeader.m_uDataSize;

			return true;
		}

		case 'XMA2':
		{
			// Note that XMA2 format chunks are stored big-endian, so no need to endian-swap

			XMA2Format xFormat;
			if( xChunkHeader.m_uDataSize > sizeof( xFormat ) )
			{
				Asura_Assert( false, "XMA2 audio chunk is too big." );
				return false;
			}

			Asura_FileAccess::ReadData( iFileHandle, &xFormat, xChunkHeader.m_uDataSize );

			if ( xFormat.Version != ASURA_XBOX360_SOUND_XMA2_VERSION )
			{
				Asura_Assertf( false, ("XMA2 chunk has invalid version (%u). Should be %u. Level needs re-exporting, or code updating.",
					xFormat.Version, ASURA_XBOX360_SOUND_XMA2_VERSION) );

				return false;
			}

			xNewBuffer.m_uNumStreams = xFormat.NumStreams;
			for( u_int u = 0; u < xFormat.NumStreams; ++u )
			{
				const XMA2STREAMFORMAT& xStreamFormat = xFormat.Streams[u];

				xNewBuffer.m_auStreamSampleRates[u] = xFormat.SampleRate;
				xNewBuffer.m_auStreamChannelCounts[u] = xStreamFormat.Channels;

				if( xStreamFormat.Channels == 0 )
				{
					Asura_Assert( false, "" );
				}
			}

			if( xFormat.LoopEnd != 0 )
			{
				// As long as the file has been built with the /l option, these values should be valid
				xNewBuffer.m_xXAudioSourceBuffer.LoopBegin = xFormat.LoopBegin;
				xNewBuffer.m_xXAudioSourceBuffer.LoopEnd = xFormat.LoopEnd;
			}
			else
			{
				// This doesn't seamlessly loop the sample. Rebuild the data with /l for correct looping
				xNewBuffer.m_xXAudioSourceBuffer.LoopBegin = 0;
				xNewBuffer.m_xXAudioSourceBuffer.LoopEnd = xFormat.SamplesInSource;
			}

			return true;
		}

		case 'seek':
		{
			// We don't care about seek tables
			Asura_FileAccess::SkipData( iFileHandle, xChunkHeader.m_uDataSize );

			return true;
		}

		default:
		{
			Asura_Assert( false, "Unrecognised WAVE subchunk encountered in sound." );
			Asura_FileAccess::SkipData( iFileHandle, xChunkHeader.m_uDataSize );

			return true;
		}
	}
}

#endif // !defined ASURA_USE_OPENAL

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformLoadSoundFromResource()
//
// Loads a sound sample from an Asura resource chunk file
//
// Parameters:		iFileHandle - Handle to open resource file
//					iFileSize - Size of this embedded resource
//					iLoadedSoundIndex - Corresponding engine sound index
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformLoadSoundFromResource( Asura_Handle iFileHandle, int iFileSize, int iLoadedSoundIndex )
{
	if ( iLoadedSoundIndex < 0 || iLoadedSoundIndex >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS )
	{
		Asura_Assertf( false, ( "Invalid buffer index (%d) passed to" __FUNCTION__, iLoadedSoundIndex ) );
		return false;
	}

	// NB: RIFFs are always stored little-endian, need to swap chunk contents. FOURCCs
	// should work out ok though because of the way the compiler treats them

#ifdef ASURA_USE_OPENAL

	RawXMASample& xNewSample = Asura_Xbox360_Sound::GetSample( iLoadedSoundIndex );

	const u_int uFileStartPos = Asura_FileAccess::GetFilePosition( iFileHandle );
	RIFFChunkHeader xChunkHeader;

	// Read RIFF header
	if ( !xChunkHeader.ReadFromFile( iFileHandle, 'RIFF' ) )
	{
		Asura_Assert( xChunkHeader.m_uChunkType != 'RIFF', "XMA file not RIFF type." );
		return false;
	}

	u_int uSubType;
	Asura_FileAccess::ReadData( iFileHandle, &uSubType, sizeof( uSubType ) );

	if( uSubType != 'WAVE' )
	{
		Asura_Assert( uSubType == 'WAVE', "Invalid audio chunk." );
		return false;
	}

	while( Asura_FileAccess::GetFilePosition( iFileHandle ) < uFileStartPos + iFileSize )
	{
		xChunkHeader.ReadFromFile( iFileHandle );

		switch ( xChunkHeader.m_uChunkType )
		{
		case 'data':
			{
				Asura_Assert( !xNewSample.m_pucXMAData, "XMA data already allocated." );
				xNewSample.m_uXMADataSize = xChunkHeader.m_uDataSize;
				xNewSample.m_pucXMAData = reinterpret_cast<u_char*>( 
					XPhysicalAlloc( xNewSample.m_uXMADataSize, MAXULONG_PTR, 0, PAGE_READWRITE ) );

				if ( !xNewSample.m_pucXMAData )
				{
					Asura_Assert( xNewSample.m_pucXMAData, "Failed to allocate memory for sound." );
					return false;
				}

				Asura_FileAccess::ReadData( iFileHandle, xNewSample.m_pucXMAData, xNewSample.m_uXMADataSize );
			}
			break;

		case 'XMA2':
			{
				Asura_Assert( !xNewSample.m_pucFormatBlock, "XMA format block already allocated." );

				// TBD I suspect we may need to do this allocation differently in future 
				// to account for the seek tables.
				xNewSample.m_pucFormatBlock = reinterpret_cast<u_char*>( Asura_Allocate( xChunkHeader.m_uDataSize ) );
				
				if ( !xNewSample.m_pucFormatBlock )
				{
					Asura_Assert( xNewSample.m_pucFormatBlock, "Failed to allocate memory for XMA format block." );
					return false;
				}

				Asura_FileAccess::ReadData( iFileHandle, xNewSample.m_pucFormatBlock, xChunkHeader.m_uDataSize );

				#if ASURA_ASSERT && 1	// Bug investigation, GameDS bug ID DS0078302X0
					XMA2WAVEFORMAT* pxWaveFormat = reinterpret_cast<XMA2WAVEFORMAT*>( xNewSample.m_pucFormatBlock );
					Asura_Assertf( pxWaveFormat->LoopEnd >= 128,	// XMADECODER_SUBFRAME_SIZE_IN_SAMPLES
						( "Loop end of %d is less than %d?", pxWaveFormat->LoopEnd, 128 ) );
				#endif
			}
			break;

		case 'seek':
			{
				// TBD the old code ignored the seek table data, and OpenAL doesn't explicitly
				// ask for it. But surely for seeking and looping to work, it does need to be there -
				// presumably immediately following the format block, as it does in the file. Work
				// out what's going on, and if necessary allocate space for the format block and seek
				// tables all in the same place, and copy them in together.
				Asura_FileAccess::SkipData( iFileHandle, xChunkHeader.m_uDataSize );
			}
			break;

		default:
			{
				Asura_Assert( false, "Unrecognised chunk in XMA file." );
				Asura_FileAccess::SkipData( iFileHandle, xChunkHeader.m_uDataSize );
			}
			break;
		}
	}

	Asura_Assert( xNewSample.m_pucXMAData, "XMA data not parsed in!" );
	Asura_Assert( xNewSample.m_uXMADataSize > 0, "Unknown XMA data size!" );
	Asura_Assert( xNewSample.m_pucFormatBlock, "XMA format block not parsed in!" );

#else

	SourceBuffer& xNewBuffer = g_axBuffers[ iLoadedSoundIndex ];

	const u_int uFileStartPos = Asura_FileAccess::GetFilePosition( iFileHandle );

	RIFFChunkHeader xChunkHeader;

	// Read RIFF header
	if( !xChunkHeader.ReadFromFile( iFileHandle, 'RIFF' ) ) return false;
	// RIFF chunk contents are the subtype FOURCC then subtype-dependent data

	u_int uSubType;
	Asura_FileAccess::ReadData( iFileHandle, &uSubType, sizeof( u_int ) );
	if( uSubType != 'WAVE' )
	{
		Asura_Assert( false, "Invalid audio chunk." );
		return false;
	}

	while( Asura_FileAccess::GetFilePosition( iFileHandle ) < uFileStartPos + iFileSize )
	{
		if( !ParseChunk( iFileHandle, xNewBuffer ) )
			return false;
	}

	xNewBuffer.m_xXAudioSourceBuffer.PlayBegin = 0;
	xNewBuffer.m_xXAudioSourceBuffer.PlayEnd = 0;
	xNewBuffer.m_xXAudioSourceBuffer.pContext = NULL;


#endif

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformUnloadSound()
//
// Frees memory associated with a given sound sample
//
// Parameters:		iLoadedSoundIndex - Engine sound index to unload
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformUnloadSound( const int iLoadedSoundIndex )
{
	if( iLoadedSoundIndex < 0 || iLoadedSoundIndex >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS )
	{
		Asura_Assertf( false, ( "Invalid buffer index (%d) passed to" __FUNCTION__, iLoadedSoundIndex ) );
		return false;
	}

#ifdef ASURA_USE_OPENAL

	RawXMASample& xSample = Asura_Xbox360_Sound::GetSample( iLoadedSoundIndex );

	if ( !xSample.m_pucXMAData )
	{
		Asura_Assert( false, "Attempting to destroy invalid audio packet." );
		return false;
	}

	XPhysicalFree( xSample.m_pucXMAData );
	xSample.m_pucXMAData = NULL;

	Asura_Deallocate( xSample.m_pucFormatBlock );
	xSample.m_pucFormatBlock = NULL;

#else

	if(g_apxOnDemandSounds[iLoadedSoundIndex] && g_apxOnDemandSounds[iLoadedSoundIndex]->IsOnDemandSound())
	{
		g_apxOnDemandSounds[iLoadedSoundIndex]->Unload();		//(Buffer is freed in Unload)
	}
	else
	{
		if( !g_axBuffers[ iLoadedSoundIndex ].m_xXAudioSourceBuffer.pBuffer )
		{
			Asura_Assert( false, "Attempting to destroy invalid audio packet." );
			return false;
		}

		XPhysicalFree( g_axBuffers[ iLoadedSoundIndex ].m_xXAudioSourceBuffer.pBuffer );
		g_axBuffers[ iLoadedSoundIndex ].m_xXAudioSourceBuffer.pBuffer = NULL;
	}
#endif

	return true;
}

/******************************************************************************/
/* Playback                                                                   */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformCreateSound()
//
// Allocates a playback voice and associates it with a loaded sound sample
//
// Parameters:		iLoadedSoundIndex - Index of the sound buffer to play
//					iPlayingSoundHandle - Handle for the new playback voice
//					bUseEQ - true iff this voice should use EQ effects
//					bUseHardware3dPan - true iff this is a spatialised voice
//										and the platform handles spatialisation
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformCreateSound( int iLoadedSoundIndex, Asura_Handle iPlayingSoundHandle, bool bUseEQ, bool bUseHardware3dPan )
{	
	if( iLoadedSoundIndex < 0 || iLoadedSoundIndex >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS )
	{
		Asura_Assertf( false, ( "Invalid buffer index (%d) passed to" __FUNCTION__, iLoadedSoundIndex ) );
		return false;
	}

	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ( "Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle ) );
		return false;
	}

#if ASURA_USE_OPENAL

		Asura_Xbox360_ActiveSound& xActiveSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
		RawXMASample& xSample = Asura_Xbox360_Sound::GetSample( iLoadedSoundIndex );
		return xActiveSound.GetAPI().Initialise( bUseHardware3dPan, xSample.m_uXMADataSize, xSample.m_pucFormatBlock, xSample.m_pucXMAData );

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];
	
	if(g_apxOnDemandSounds[iLoadedSoundIndex] && g_apxOnDemandSounds[iLoadedSoundIndex]->IsOnDemandSound())
	{
		xVoice.m_iOnDemandIndex = iLoadedSoundIndex;
		g_apxOnDemandSounds[iLoadedSoundIndex]->SetSourceBuffer(&g_axBuffers[ iLoadedSoundIndex ]);
		g_apxOnDemandSounds[iLoadedSoundIndex]->Prepare(iPlayingSoundHandle);		
	}
	else
	{
		xVoice.m_iOnDemandIndex = ASURA_INVALID_HANDLE;
	}
	xVoice.m_pxSourceBuffer = &g_axBuffers[ iLoadedSoundIndex ];

	return true;

#endif

}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformDestroySound()
//
// Stops and destroys a sound playback voice
//
// Parameters:		iPlayingSoundHandle - Handle of the voice to destroy
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformDestroySound( Asura_Handle iPlayingSoundHandle )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	xSound.GetAPI().Destroy();

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	if(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE)
	{
		bool bCanDestroy = true;
		//We need to check that no other voices are using this on-demand sound before we destroy it
		for(u_int u = 0; u < ASURA_XBOX360_SOUND_MAX_VOICES; ++ u )
		{
			if(u != static_cast<u_int>(iPlayingSoundHandle))
			{
				if( g_axVoices[ u ].m_iOnDemandIndex == xVoice.m_iOnDemandIndex )
				{
					bCanDestroy = false;
					break;
				}
			}
		}
		if(g_apxOnDemandSounds[xVoice.m_iOnDemandIndex] && bCanDestroy)
		{
			//destroy if we are able
			Asura_Assert(g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound(),"Voice and OnDemandSound mismatch");
			g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->Destroy(iPlayingSoundHandle);
		}
		xVoice.m_iOnDemandIndex = ASURA_INVALID_HANDLE; //disassociate this voice with the on-demand sound
	}

	if( !xVoice.m_pxXAudioSourceVoice ) return true;

	const HRESULT xResult = xVoice.m_pxXAudioSourceVoice->Stop(0);
	if( FAILED( xResult ) ) return false;

	xVoice.m_pxXAudioSourceVoice->Release();
	xVoice.m_pxXAudioSourceVoice = NULL;

	if(xVoice.m_pxXAudioDrySubmix)
	{
		xVoice.m_pxXAudioDrySubmix->Release();
		xVoice.m_pxXAudioDrySubmix = NULL;
	}
	xVoice.m_fVolume			= 0.0f;
	xVoice.m_fPan				= 0.0f;
	xVoice.m_fFade				= 0.0f;
	xVoice.m_fPitch				= 0.0f;
	xVoice.m_bReverb			= false;
	xVoice.m_fWetVolume			= 0.0f;
	xVoice.m_fDryVolume			= 0.0f;
	xVoice.m_fOcclusionFilter	= 0.0f;
	xVoice.m_bPaused			= false;
	xVoice.m_bPausedByPitch		= false;
#endif

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformPlaySound()
//
// Begins playback on a source voice
//
// Parameters:		iPlayingSoundHandle - Handle of voice to play
//					bLooping - true iff the sound should loop until stopped
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformPlaySound( Asura_Handle iPlayingSoundHandle, bool bLooping )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	xSound.SetLooping( bLooping );

	bool success = xSound.GetAPI().Play( bLooping );

	if ( success )
	{
		return true;
	}

	return false;

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE)
	{
		Asura_Assert(g_apxOnDemandSounds[xVoice.m_iOnDemandIndex], "Loaded on demand sound is null");
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueuePlay(iPlayingSoundHandle,bLooping))
		{
			return true;
		}
	}

	// For now, create voices on the fly so we can respond to changing sample rates and looping

	// ---- Create the voice ----
	XAUDIOSOURCEVOICEINIT xVoiceParams = {0};

	xVoiceParams.Format.SampleType = XAUDIOSAMPLETYPE_XMA;
	xVoiceParams.Format.NumStreams = static_cast< XAUDIOXMASTREAMCOUNT >( xVoice.m_pxSourceBuffer->m_uNumStreams );

	bool bMono = true;

	for( u_int u = 0; u < xVoice.m_pxSourceBuffer->m_uNumStreams; ++u )
	{
		xVoiceParams.Format.Stream[u].SampleRate = xVoice.m_pxSourceBuffer->m_auStreamSampleRates[u];
		xVoiceParams.Format.Stream[u].ChannelCount = static_cast< XAUDIOCHANNEL >( xVoice.m_pxSourceBuffer->m_auStreamChannelCounts[u] );
		xVoiceParams.Format.Stream[u].DecodeBufferSize = XMASUBFRAME_DECODEROUTPUT_FASTEST;

		if( xVoiceParams.Format.Stream[u].ChannelCount > 1 )
		{
			bMono = false;
		}
	}

	//If this is a mono sound, create a dry submix (where we can do occlusion filtering
	if(bMono)
	{
		// ---- Create the dry submix ----

		//The dry submix has one input and six outputs
		XAUDIOCHANNELMAPENTRY axSubmixChannelMapEntries[XAUDIOSPEAKER_COUNT];
		for( XAUDIOCHANNEL u = 0; u < XAUDIOSPEAKER_COUNT; ++u )
		{
			axSubmixChannelMapEntries[u].InputChannel = 0;	//the dry submix only applies for mono sounds
			axSubmixChannelMapEntries[u].OutputChannel = static_cast< XAUDIOCHANNEL >(u);
			axSubmixChannelMapEntries[u].Volume = 1.0f;
		}

		XAUDIOCHANNELMAP xSubmixChannelMap;
		xSubmixChannelMap.EntryCount = XAUDIOSPEAKER_COUNT;
		xSubmixChannelMap.paEntries = &axSubmixChannelMapEntries[0];

		XAUDIOVOICEOUTPUTENTRY xSubmixOutputEntry;
		xSubmixOutputEntry.pDestVoice = NULL; // Output to mastering voice
		xSubmixOutputEntry.pChannelMap = &xSubmixChannelMap;

		XAUDIOVOICEOUTPUT xSubmixOutput;
		xSubmixOutput.EntryCount = 1;
		xSubmixOutput.paEntries = &xSubmixOutputEntry;

		XAUDIOSUBMIXVOICEINIT xSubmixVoiceInit = {0};

		xSubmixVoiceInit.Format.ChannelCount = 1;
		xSubmixVoiceInit.Format.SampleType = XAUDIOSAMPLETYPE_NATIVE;	//Ignored
		xSubmixVoiceInit.Format.SampleRate = XAUDIOSAMPLERATE_NATIVE;	//Ignored

		xSubmixVoiceInit.MaxOutputVoiceCount = 1;
		xSubmixVoiceInit.MaxChannelMapEntryCount = XAUDIOSPEAKER_COUNT;
		xSubmixVoiceInit.SubmixStage = 0;
		xSubmixVoiceInit.pVoiceOutput = &xSubmixOutput;
		xSubmixVoiceInit.pfnProcessCallback = NULL;
		xSubmixVoiceInit.pContext = NULL;

		// Create a small effect chain to perform occlusion lpf
		XAUDIOVOICEFXCHAIN xMixFXChain;
		XAUDIOFXINIT xEffectsInit;
		LPCXAUDIOFXINIT pxEffectPointer;

		xEffectsInit.EffectId = XAUDIOFXID_LOWPASS;
		xEffectsInit.pContext = NULL;

		pxEffectPointer = &xEffectsInit;

		xMixFXChain.EffectCount = 1; 
		xMixFXChain.papEffects = &pxEffectPointer;

		xSubmixVoiceInit.pEffectChain = &xMixFXChain;

		//...and create
		HRESULT xResult = XAudioCreateSubmixVoice( &xSubmixVoiceInit, &xVoice.m_pxXAudioDrySubmix );
		if(FAILED(xResult))
		{
			return false;
		}
	}
	else
	{
		xVoice.m_pxXAudioDrySubmix = NULL;
	}

	// --- back to creating the voice... ---

	const XAUDIOVOICEINDEX uNumVoiceOutputs = xVoice.m_bReverb ? 2 : 1;

	xVoiceParams.Category = XAUDIOVOICECATEGORY_NONE;
	xVoiceParams.MaxOutputVoiceCount = uNumVoiceOutputs; 
	xVoiceParams.MaxChannelMapEntryCount = XAUDIOSPEAKER_COUNT;
	xVoiceParams.MaxPacketCount = 1;
	xVoiceParams.MaxPitchShift = 1.0f; // Pitch shift in octaves, not absolute pitch in 0-1 scale

	// Sort out speaker mappings
	XAUDIOVOICEOUTPUT xVoiceOutputMap;
	XAUDIOVOICEOUTPUTENTRY axOutputEntries[2];
	XAUDIOCHANNELMAP xDryChannelMap, xWetChannelMap;
	XAUDIOCHANNELMAPENTRY axDryChannelMapEntries[6], xWetChannelMapEntry;

	if(xVoice.m_pxXAudioDrySubmix)
	{
		//With a dry submix we just pass mono through
		axDryChannelMapEntries[0].InputChannel = 0;
		axDryChannelMapEntries[0].OutputChannel = 0;
		axDryChannelMapEntries[0].Volume = 1.0f;

		xDryChannelMap.EntryCount = 1;
		xDryChannelMap.paEntries = axDryChannelMapEntries;

		axOutputEntries[0].pDestVoice  = xVoice.m_pxXAudioDrySubmix;
		axOutputEntries[0].pChannelMap = &xDryChannelMap;
	}
	else
	{
		// If the dry path goes straight to the master, pass all channels through
		for( u_int u = 0; u < XAUDIOSPEAKER_COUNT; ++u )
		{
			axDryChannelMapEntries[u].InputChannel = static_cast< XAUDIOCHANNEL >(u);
			axDryChannelMapEntries[u].OutputChannel = static_cast< XAUDIOCHANNEL >(u);
			axDryChannelMapEntries[u].Volume = 1.0f;
		}

		xDryChannelMap.EntryCount = XAUDIOSPEAKER_COUNT;
		xDryChannelMap.paEntries = axDryChannelMapEntries;

		axOutputEntries[0].pDestVoice  = NULL;
		axOutputEntries[0].pChannelMap = &xDryChannelMap;
	}

	// Wet channel maps (mono)
	xWetChannelMapEntry.InputChannel = 0;
	xWetChannelMapEntry.OutputChannel = 0;
	xWetChannelMapEntry.Volume = 1.0f;

	xWetChannelMap.EntryCount = 1;
	xWetChannelMap.paEntries = &xWetChannelMapEntry;

	axOutputEntries[1].pDestVoice  = g_pxReverbSubmixVoice;
	axOutputEntries[1].pChannelMap = &xWetChannelMap;
	
	xVoiceOutputMap.EntryCount = uNumVoiceOutputs;
	xVoiceOutputMap.paEntries = &axOutputEntries[0];

	xVoiceParams.pVoiceOutput = &xVoiceOutputMap;

	// Create a small effect chain to perform distance filtering
	// we do this on the voice to have it applied to both wet and dry submixes
	XAUDIOVOICEFXCHAIN xVoiceFXChain;
	XAUDIOFXINIT xVoiceEffectsInit;
	LPCXAUDIOFXINIT pxVoiceEffectPointer;

	xVoiceEffectsInit.EffectId = XAUDIOFXID_LOWPASS;
	xVoiceEffectsInit.pContext = NULL;

	pxVoiceEffectPointer = &xVoiceEffectsInit;

	xVoiceFXChain.EffectCount = 1; 
	xVoiceFXChain.papEffects = &pxVoiceEffectPointer;

	xVoiceParams.pEffectChain = &xVoiceFXChain;

	// Create the voice
	XAudioCreateSourceVoice( &xVoiceParams, &xVoice.m_pxXAudioSourceVoice );
	if( !xVoice.m_pxXAudioSourceVoice )
	{
		Asura_Assert( false, "Failed to create audio source voice." );
		return false;
	}
	
	// These are set before play is called
	PlatformSetSoundPitch				( iPlayingSoundHandle, xVoice.m_fPitch );
	PlatformSetSoundVolumePanAndFade	( iPlayingSoundHandle, xVoice.m_fVolume, xVoice.m_fPan, xVoice.m_fFade );
	Platform_SetOcclusionLPF			( iPlayingSoundHandle, xVoice.m_fOcclusionFilter);
	
	// This should be done for us in PlatformSetSoundVolumePanAndFade...?
	//xVoice.m_pxXAudioSourceVoice->SetVoiceOutputVolume( &xVoice.m_xOutputVolume );

	// Fix up the buffer loop count
	xVoice.m_pxSourceBuffer->m_xXAudioSourceBuffer.LoopCount = bLooping ? XAUDIOLOOPCOUNT_INFINITE : 0;

	// Submit the buffer
	// NB: if we change to a pool of persistant voices, call FlushSourceBuffers() to ditch existing packets

	HRESULT xResult = xVoice.m_pxXAudioSourceVoice->SubmitSourceBuffer( &xVoice.m_pxSourceBuffer->m_xXAudioSourceBuffer, XAUDIOSUBMITPACKET_DISCONTINUITY );
	if( FAILED( xResult ) )
	{
		Asura_Assert( false, "Failed to submit source audio buffer." );
		return false;
	}

	// Start the voice playing
	if(!xVoice.m_bPausedByPitch && !xVoice.m_bPaused)
	{
		xResult = xVoice.m_pxXAudioSourceVoice->Start(0);
		return SUCCEEDED( xResult );
	}
	
	return true;
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformStopSound()
//
// Stops a playback voice
//
// Parameters:		iPlayingSoundHandle - Handle of voice to stop
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformStopSound( Asura_Handle iPlayingSoundHandle )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	return xSound.GetAPI().Stop();

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueStop(iPlayingSoundHandle))
		{
			return true;
		}
	}

	if( !g_axVoices[ iPlayingSoundHandle ].m_pxXAudioSourceVoice ) return false;

	const HRESULT xResult = g_axVoices[ iPlayingSoundHandle ].m_pxXAudioSourceVoice->Stop(0);
	return SUCCEEDED( xResult );
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformSoundIsPlaying()
//
// Queries whether a source voice is currently playing
//
// Parameters:		iPlayingSoundHandle - Handle of voice to query
// Return value:	true iff the voice is playing or paused, false if stopped
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformSoundIsPlaying( Asura_Handle iPlayingSoundHandle )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	return xSound.GetAPI().IsPlaying();

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//Handle on-demand sounds
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		//If we are still preparing or waiting for an update, we need to return our playing 'state'
		//if not preparing then the voice will tell us the playing state
		const u_int uState = g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->GetState();
		if(uState & ASURA_ON_DEMAND_STATE_VOID)
		{
			Asura_Assert(false, "Checking the playing state of an uninitialised on-demand sound");
			return false;
		}
		if( (uState & ASURA_ON_DEMAND_STATE_PREPARING) || (uState & ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE))
		{
			if( (uState & ASURA_ON_DEMAND_STATE_PLAYING) || (uState & ASURA_ON_DEMAND_STATE_PAUSED) )
			{
				return true;
			}
			else
			{
				return false;
			}
		}
	}

	if( !xVoice.m_pxXAudioSourceVoice ) return false;

	// Considered as "playing" if we're paused
	if( xVoice.m_bPaused || xVoice.m_bPausedByPitch ) return true;

	XAUDIOSOURCESTATE xVoiceState;
	xVoice.m_pxXAudioSourceVoice->GetVoiceState( &xVoiceState );

	return (xVoiceState & XAUDIOSOURCESTATE_STARTED) && !(xVoiceState & XAUDIOSOURCESTATE_STOPPING);
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformSoundIsPreparing()
//
// Queries whether a sound is preparing to play
//
// Parameters:		iPlayingSoundHandle - Handle of voice to query
// Return value:	true if preparing
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformSoundIsPreparing(Asura_Handle iPlayingSoundHandle)
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	#if ASURA_USE_OPENAL

	return false;

	#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//Check if this is an on demand sound, and if it is preparing
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		const u_int uState = g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->GetState();
		
		if( ( uState & ASURA_ON_DEMAND_STATE_PREPARING ) || ( uState & ASURA_ON_DEMAND_STATE_PREPARE_NEEDUPDATE ) )
		{
			return true;
		}
	}
	return false;

	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformPauseSound()
//
// Pauses a playback voice
//
// Parameters:		iPlayingSoundHandle - Handle of voice to pause
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformPauseSound( Asura_Handle iPlayingSoundHandle )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	return xSound.GetAPI().Pause();

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueuePause(iPlayingSoundHandle))
		{
			return true;
		}
	}

	if( !xVoice.m_pxXAudioSourceVoice )
	{
		Asura_Assert( false, "Trying to pause an invalid voice." );
		return false;
	}

	// Stop the voice

	xVoice.m_bPaused = true;

	const HRESULT xResult = xVoice.m_pxXAudioSourceVoice->Stop(XAUDIOSTOP_IMMEDIATE);
	return SUCCEEDED( xResult );
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformUnpauseSound()
//
// Resumes playback of a voice
//
// Parameters:		iPlayingSoundHandle - Handle of voice to unpause
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformUnpauseSound( Asura_Handle iPlayingSoundHandle )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	return xSound.GetAPI().Unpause( xSound.IsLooping() );

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueResume(iPlayingSoundHandle))
		{
			return true;
		}
	}

	if( !xVoice.m_pxXAudioSourceVoice )
	{
		Asura_Assert( false, "Trying to unpause an invalid voice." );
		return false;
	}

	// Start the voice

	xVoice.m_bPaused = false;

	if( !xVoice.m_bPausedByPitch )
	{
		const HRESULT xResult = xVoice.m_pxXAudioSourceVoice->Start(0);
		return SUCCEEDED( xResult );
	}
	else
	{
		return true;
	}
#endif
}

/******************************************************************************/
/* Global Properties                                                          */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::Platform_SetReverb()
//
// Sets the current reverb type to apply to all spatialised sounds
//
// Parameters:		eType - Reverb type to apply
//					fReverbAmount - Amount of reverb [0,1]
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Sound_System::Platform_SetReverb( ReverbType eType, float fReverbAmount )
{
#if ASURA_USE_OPENAL
	Asura_Xbox360_Sound::GetAPI().SetReverb( eType );
#else

	if( !g_pxReverbSubmixVoice ) return;

	if( eType < 0 || eType >= uNUM_REVERB_SETTINGS )
	{
		Asura_Assertf( false, ("Invalid reverb type (%i) passed to " __FUNCTION__, eType) );
		eType = ASURA_REVERB_ID_NONE;
	}

	XAUDIOFXPARAM xReverbParams;

	// Set reverb params through I3DL2 interface
	{
		XAUDIOREVERBI3DL2SETTINGS xSettings = axREVERB_SETTINGS[ eType ];

		XAUDIOVOLUME fVolume = XAudioMillibelsToVolume( xSettings.lRoom );
		fVolume *= fReverbAmount;

		if( fVolume <= 0.001f )
		{
			// Avoid log(0)
			xSettings.lRoom = -10000;
		}
		else
		{
			const int iMillibels = XAudioVolumeToMillibels( fVolume );
			xSettings.lRoom = Asura_Maths::ClampToRange( iMillibels, -10000, 0 );
		}

		xReverbParams.Data.pBuffer = &xSettings;
		xReverbParams.Data.BufferSize = sizeof( XAUDIOREVERBI3DL2SETTINGS );

		#if defined ASURA_ASSERT
			const HRESULT xResult =
		#endif
			g_pxReverbSubmixVoice->SetEffectParam(	ASURA_XBOX360_DSP_EFFECTS_REVERB, 
													XAUDIOFXPARAMID_REVERB_I3DL2SETTINGS,
													XAUDIOFXPARAMTYPE_DATA, 
													&xReverbParams );

		Asura_Assert( SUCCEEDED( xResult ), "Failed to set reverb." );
	}

#endif
}

/******************************************************************************/
/* Individual Voice Properties                                                */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::Platform_SetEQType()
//
// Sets the send level for a given EQ effect. Applies to all EQ-enabled voices
//
// Parameters:		iHandle - EQ handle to modify
//					eType - Type of EQ effect to apply
//					fLevel - Send level [0, 1]
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Sound_System::Platform_SetEQType( Asura_Handle iHandle, EQType eType, float fLevel )
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformSetSoundPitch()
//
// Sets the pitch of a playback voice
//
// Parameters:		iPlayingSoundHandle - Handle of voice to adjust
//					fPitch - Pitch to set. 0 = paused, 0.5 = -1 octave, 1 = normal pitch
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformSetSoundPitch( Asura_Handle iPlayingSoundHandle, float fPitch )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to" __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	if( fPitch < 0.0f || fPitch > 2.0f )
	{
		//TY 13.02.08 - Dont assert about the pitch being > 2, because of the way it is calculated this is entirely possible
		//Asura_Assertf( false, ("Invalid pitch (%f) passed to Asura_Sound_System::PlatformSetSoundPitch().", fPitch) );
		Asura_Assertf( fPitch >= 0.0f, ("Invalid pitch (%f) passed to Asura_Sound_System::PlatformSetSoundPitch().", fPitch) );
		fPitch = Asura_Maths::ClampToRange( fPitch, 0.0f, 2.0f );
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	if ( fPitch == 0.0f )
	{
		xSound.SetPausedByPitch( true );
		return xSound.GetAPI().Pause();
	}
	else
	{
		if ( xSound.IsPausedByPitch() && !xSound.GetAPI().Unpause( xSound.IsLooping() ) )
		{
			Asura_Assert( false, "Could not unpause sound from pitch!" );
			return false;
		}
		
		return xSound.GetAPI().SetPitch( fPitch );
	}

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueSetPitch(iPlayingSoundHandle,fPitch))
		{
			return true;
		}
	}

	xVoice.m_fPitch = fPitch;

	if( fPitch == 0.0f )
	{
		// Can't set the pitch shift to -infinity. Pause the sound instead

		xVoice.m_bPausedByPitch = true;

		if( xVoice.m_pxXAudioSourceVoice )
		{
			const HRESULT xResult = xVoice.m_pxXAudioSourceVoice->Stop(0);
			return SUCCEEDED( xResult );
		}
		else
		{
			return true;
		}
	}
	else
	{
		// Should be ok setting this pitch

		if( xVoice.m_pxXAudioSourceVoice )
		{
			HRESULT xResult;

			const XAUDIOPITCH fOctaveAdjustment = XAudioSampleRateToPitch( static_cast< XAUDIOSAMPLERATE >( fPitch * 48000.0f ), 48000 );

			xResult = xVoice.m_pxXAudioSourceVoice->SetPitch( fOctaveAdjustment );
			if( FAILED( xResult ) )
			{
				Asura_Assert( false, "Failed to set voice pitch." );
				return false;
			}

			if( xVoice.m_bPausedByPitch && !xVoice.m_bPaused )
			{
				xVoice.m_bPausedByPitch = false;

				xResult = xVoice.m_pxXAudioSourceVoice->Start(0);
				return SUCCEEDED( xResult );
			}
			else
			{
				return true;
			}
		}
		else
		{
			xVoice.m_bPausedByPitch = false;
			return true;
		}
	}
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_System::PlatformSetSoundVolumePanAndFade()
//
// Sets the volume, pan and fade of a playback voice
//
// Parameters:		iPlayingSoundHandle - Handle of source voice
//					fVolume - Volume to set. 0 = silent, 0.5 = -6 dB, 1 = full volume
//					fPan - Pan to set. 0 = left speakers only, 1 = right speakers only
//					fFade - Fade to set. 0 = rear speakers only, 1 = front speakers only
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::PlatformSetSoundVolumePanAndFade( Asura_Handle iPlayingSoundHandle, float fVolume, float fPan, float fFade )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	if( fPan < 0.0f || fPan > 1.0f )
	{
		Asura_Assertf( false, ("Invalid pan (%f) passed to Asura_Sound_System::PlatformSetSoundVolumePanAndFade().", fPan) );
		fPan = Asura_Maths::ClampToRange( fPan, 0.0f, 1.0f );
	}

	if( fFade < 0.0f || fFade > 1.0f )
	{
		Asura_Assertf( false, ("Invalid fade (%f) passed to Asura_Sound_System::PlatformSetSoundVolumePanAndFade().", fFade) );
		fFade = Asura_Maths::ClampToRange( fFade, 0.0f, 1.0f );
	}

	if( fVolume < 0.0f || fVolume > 1.0f )
	{
		Asura_Assertf( false, ("Invalid volume (%f) passed to Asura_Sound_System::PlatformSetSoundVolumePanAndFade().", fVolume) );
		fVolume = Asura_Maths::ClampToRange( fVolume, 0.0f, 1.0f );
	}

#if ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	bool bSetPanFade = xSound.GetAPI().SetForcedPanAndFade( fPan, fFade );
	bool bSetVolume = xSound.GetAPI().SetVolume( fVolume );

	return bSetVolume && bSetPanFade;

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueSetVolumePanAndFade(iPlayingSoundHandle,fVolume,fPan,fFade))
		{
			return true;
		}
	}

	xVoice.m_fVolume	= fVolume;
	xVoice.m_fPan		= fPan;
	xVoice.m_fFade		= fFade;

	if(UpdateVoiceWetDryVolumes(&xVoice))
	{
		return true;
	}
	else
	{
		return false;
	}
#endif
}

bool Asura_Sound_System::Platform_SetSoundDistanceFilterGain (Asura_Handle iPlayingSoundHandle, float fNormalisedGain)
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__,
			iPlayingSoundHandle) );
		return false;
	}

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueSetDistFilterGain(iPlayingSoundHandle,fNormalisedGain))
		{
			return true;
		}
	}

	if( xVoice.m_pxXAudioSourceVoice  )
	{
		//Set our distance value and update our filter

		XAUDIOFXPARAM xGain;
		xGain.Value = fNormalisedGain * fASURA_XBOX360_SOUND_LPF_GAIN;
		#ifdef CONSOLE_VARNAME_AUDIO_DISTANCE_SCALE
		xGain.Value *= g_fDistanceFilterScale;
		#endif
		HRESULT xResult = xVoice.m_pxXAudioSourceVoice->SetEffectParam(	ASURA_XBOX360_DSP_EFFECTS_DISTANCE_FILTER, 
																		Asura_Xbox360_Sound_LowPassFilter::PID_GAIN, 
																		XAUDIOFXPARAMTYPE_NUMERIC, 
																		&xGain );

		if(FAILED(xResult))
		{
			return false;
		}
	}
	return true;
}


bool Asura_Sound_System::Platform_SetSoundWetDryVolume (Asura_Handle iPlayingSoundHandle, float fWetVolume, float fDryVolume)
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueSetWetDryVolume(iPlayingSoundHandle, fWetVolume, fDryVolume))
		{
			return true;
		}
	}

	//Adjust the actual volumes
	xVoice.m_fWetVolume = fWetVolume;
	xVoice.m_fDryVolume = fDryVolume;

	//We may or may not have volume, pan & fade at this stage
	if(UpdateVoiceWetDryVolumes(&xVoice))
	{
		return true;
	}
	else
	{
		return false;
	}
}


bool Asura_Sound_System::Platform_SetOcclusionLPF (const Asura_Handle iPlayingSoundHandle, const float fOcclusion)
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueSetOcclusionLPF(iPlayingSoundHandle, fOcclusion))
		{
			return true;
		}
	}

	//Save the filter value in case we set this before calling play
	xVoice.m_fOcclusionFilter = fOcclusion;

	if( xVoice.m_pxXAudioDrySubmix )
	{
		//Set our distance value and update our filter
		
		XAUDIOFXPARAM xGain;
		#ifdef CONSOLE_VARNAME_AUDIO_OCCLUSION_SCALE
		xGain.Value = fOcclusion * fASURA_XBOX360_SOUND_LPF_GAIN * g_fOcclusionFilterScale;
		#else
		xGain.Value = fOcclusion * fASURA_XBOX360_SOUND_LPF_GAIN;
		#endif
		HRESULT xResult = xVoice.m_pxXAudioDrySubmix->SetEffectParam(	ASURA_XBOX360_DSP_EFFECTS_OCCLUSION_FILTER, 
																		Asura_Xbox360_Sound_LowPassFilter::PID_GAIN, 
																		XAUDIOFXPARAMTYPE_NUMERIC, 
																		&xGain );

		if(FAILED(xResult))
		{
			return false;
		}
	}
	return true;
}

// Hardware panning is currently PC-only (it's the only way under DirectSound to access surround speakers)
// All other platforms allow us to specify speaker volumes directly (or if using Dolby PL2 have some concept of front and back as well as pan).
// TBD on Xbox 360 would it be better to use X3DAudio and/or OpenAL for 3D positioning?
bool Asura_Sound_System::Platform_HardwarePanSetSoundVolume(Asura_Handle iPlayingSoundHandle, float fVolume)
{
#if XBOX360_SOUND_USE_HARDWARE_PAN

	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	return xSound.GetAPI().SetVolume( fVolume );

#else
	Asura_Assert(false, "this platform did not expect to use hardware 3d pan");
	return true;
#endif
}

bool Asura_Sound_System::Platform_HardwarePanSetSoundPosition(Asura_Handle iPlayingSoundHandle, const Asura_Vector_3& xPosition)
{
#if XBOX360_SOUND_USE_HARDWARE_PAN

	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	return xSound.GetAPI().SetPosition( xPosition );

#else
	Asura_Assert(false, "this platform did not expect to use hardware 3d pan");
	return true;
#endif
}

bool Asura_Sound_System::Platform_HardwarePanSetSoundSpatial(Asura_Handle iPlayingSoundHandle, bool bIsSpatial)
{
#if XBOX360_SOUND_USE_HARDWARE_PAN

	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	return xSound.GetAPI().SetSpatial( bIsSpatial );

#else
	Asura_Assert(false, "this platform did not expect to use hardware 3d pan");
	return true;
#endif
}

bool Asura_Sound_System::Platform_HardwarePanSetSoundRadii(Asura_Handle iPlayingSoundHandle, float fInner, float fOuter)
{
#if XBOX360_SOUND_USE_HARDWARE_PAN

	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	return xSound.GetAPI().SetRadii( fInner, fOuter );

#else
	Asura_Assert(false, "this platform did not expect to use hardware 3d pan");
	return true;
#endif
}

bool Asura_Sound_System::Platform_HardwarePanSetSoundForcedPanAndFade(Asura_Handle iPlayingSoundHandle, float fPan, float fFade)
{
#if XBOX360_SOUND_USE_HARDWARE_PAN

	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );

	return xSound.GetAPI().SetForcedPanAndFade( fPan, fFade );

#else
	Asura_Assert(false, "this platform did not expect to use hardware 3d pan");
	return true;
#endif
}

////////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_System::Platform_SetSoundUsesDistanceFilter( Asura_Handle iPlayingSoundHandle, bool bUseFilter )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#ifdef ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	xSound.GetAPI().SetUseDistanceFilter( bUseFilter );

#endif

	return true;
}

bool Asura_Sound_System::Platform_SetSoundUsesReverb( Asura_Handle iPlayingSoundHandle, bool bUseReverb )
{
	if(!IsValidPlayingSoundHandle(iPlayingSoundHandle))
	{
		Asura_Assertf( false, ("Invalid playing sound handle (%d) passed to " __FUNCTION__, iPlayingSoundHandle) );
		return false;
	}

#ifdef ASURA_USE_OPENAL

	Asura_Xbox360_ActiveSound& xSound = Asura_Xbox360_Sound::GetSound( iPlayingSoundHandle );
	xSound.GetAPI().SetUseReverb( bUseReverb );

#else

	SourceVoice& xVoice = g_axVoices[ iPlayingSoundHandle ];

	//handle on-demand sounds, i.e. call a function to queue this action, if queuing is not necessary it will return true and we can execute platform code
	if(		(xVoice.m_iOnDemandIndex != ASURA_INVALID_HANDLE) 
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]
		&&	g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->IsOnDemandSound())
	{
		if(!g_apxOnDemandSounds[xVoice.m_iOnDemandIndex]->QueueSetUsesReverb(iPlayingSoundHandle,bUseReverb))
		{
			return true;
		}
	}

	xVoice.m_bReverb = bUseReverb;

	// This shouldn't fire, because the only place this is called is between PlatformCreateSound()
	// and PlatformPlaySound(). The source voice is only created in the latter
	Asura_Assertf( !xVoice.m_pxXAudioSourceVoice,
		("Trying to set 'uses reverb' on already playing sound %i.", iPlayingSoundHandle) );

#endif

	return true;
} 



////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_OPENAL
Asura_Xbox360_ActiveSound::Asura_Xbox360_ActiveSound() :
	m_bPausedByPitch( false ),
	m_bIsLooping( false )
{

}
#endif

////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_OPENAL
void Asura_Xbox360_Sound::UpdateSound( Asura_Handle iActiveSoundIndex )
{
	Asura_Xbox360_ActiveSound& xActiveSound = s_axActiveSounds[iActiveSoundIndex];
	Asura_Assert( xActiveSound.GetAPI().IsValid(), "invalid sound" );
	xActiveSound.GetAPI().Update();
}
#endif

#endif // #ifndef ASURA_USE_AUDIO_SYSTEM

//eof
