//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Beam_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "SpecialFX/Asura_Beam_System.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"

// Shader Include Files
#include "Beam.h"
// Misc
#include <fxl.h>
#include <xtl.h>

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct BeamVertex
{
	float x, y, z;
	D3DCOLOR m_uColour;
	float u, v;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const D3DVERTEXELEMENT9 axVERTEX_DECLARATION[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3,   D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 12, D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR,    0 },
	{ 0, 16, D3DDECLTYPE_FLOAT2,   D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	D3DDECL_END()
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Asura_Xbox360_Effect* g_pxBeamEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_hEffectBaseTexture = 0;
static Asura_Xbox360_Effect_System::Handle g_hEffectViewProj = 0;

static Direct3D_VertexDeclaration* g_pxBeamVD = NULL;

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// RenderBeam()
//
// Renders a single beam
//
// Parameters:		xBeam - Reference to beam to render
//					fAngleTolerance - Tolerance for the angle effect
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

static void RenderBeam( const Asura_Beam& xBeam, const float fAngleTolerance )
{
	#if defined ASURA_BEAMS_ALL_MODES
		Asura_Render::CurrentStates.RequestTranslucencyMode( xBeam.GetTranslucencyMode() );
	#endif

	const Asura_Matrix_3x3& xCamOri = *Asura_Camera::GetOrientation();
	const Asura_Vector_3& xCamPos = *Asura_Camera::GetPosition();

	const Asura_Vector_3 xCamRight( xCamOri.e11, xCamOri.e21, xCamOri.e31 );
	const Asura_Vector_3 xCamUp( xCamOri.e12, xCamOri.e22, xCamOri.e32 );
	const Asura_Vector_3 xCamForward( xCamOri.e13, xCamOri.e23, xCamOri.e33 );

	const Asura_Vector_3& xBeamStart = xBeam.GetTailPos();
	const Asura_Vector_3& xBeamDir = xBeam.GetDir();
	const Asura_Vector_3 xBeamEnd = xBeam.CalcCurHeadPos();

	float fTailRadius, fHeadRadius;
	xBeam.CalcCurRadii( &fTailRadius, &fHeadRadius );

	Asura_RGBA xTailColour, xHeadColour;
	xBeam.CalcCurColours( &xTailColour, &xHeadColour );

	const Asura_TiledTextureInfo* const pxBeamTile = Asura_TiledTextureManagement::GetTileInfoByHash( xBeam.GetTemplate()->uTileHashID );
	Asura_Assert( pxBeamTile, "Failed to get beam template texture tile." );

	Asura_Vector_3 xEyeVector = xCamPos;
	xEyeVector -= xBeamStart;
	xEyeVector.Normalise();

	// The effect doesn't work at low angles to sight so we introduce 
	// a tolerance angle in which the head of the beam is brought
	// up to the tail - there is a tolerance to stop things
	// flipping too much with small error

	float fAngleToLineOfSight = xEyeVector * xBeamDir;

	// Switch angle sign if -ve to reduce later work
	if( fAngleToLineOfSight < 0.0f )
	{
		fAngleToLineOfSight = -fAngleToLineOfSight;
	}

	Asura_Vector_3 xPerpVector;
	Asura_Vector_3 axVertexPos[6];
	D3DCOLOR uColour;
	HRESULT hResult;
	BeamVertex* pxVertex;

	// Work out if in line with camera - if in line draw two 
	// separate triangle fans with the mirrored head and tail
	// else draw a triangle strip with flattend ends facing the camera

	if( fAngleToLineOfSight > 1.0f - fAngleTolerance )
	{
		if( fAngleToLineOfSight >= 1.0f )
		{
			// Alternative to beam direction needed, use camera side
			xPerpVector = xCamRight;
		}
		else
		{
			// Cross product ok to use
			xPerpVector = xBeamDir ^ xEyeVector;
		}

		// Draw two triangle fans on each other (tail and head points)

		xPerpVector.Normalise();

		Asura_Vector_3 xUp = xCamForward ^ xPerpVector;
		xUp.Normalise();
		const Asura_Vector_3 xRight = xCamForward ^ xUp;

		// First fan

		axVertexPos[0] = xRight;
		axVertexPos[0] *= -fTailRadius;
		axVertexPos[0] += xBeamStart;

		axVertexPos[1] = xUp;
		axVertexPos[1] *= fTailRadius;
		axVertexPos[1] += xBeamStart;

		axVertexPos[2] = xRight;
		axVertexPos[2] *= fTailRadius;
		axVertexPos[2] += xBeamStart;

		axVertexPos[3] = xUp;
		axVertexPos[3] *= -fTailRadius;
		axVertexPos[3] += xBeamStart;

		uColour = xTailColour.ToUIntARGB();

		hResult = Asura_Xbox360_Graphics::GetDevice()->BeginVertices( D3DPT_TRIANGLEFAN, 4, sizeof( BeamVertex ), reinterpret_cast< void** >( &pxVertex ) );
		if( SUCCEEDED( hResult ) )
		{
			pxVertex[0].x = axVertexPos[0].x;
			pxVertex[0].y = axVertexPos[0].y;
			pxVertex[0].z = axVertexPos[0].z;
			pxVertex[0].m_uColour = uColour;
			pxVertex[0].u = pxBeamTile->m_fu1;
			pxVertex[0].v = 0.5f * (pxBeamTile->m_fv1 + pxBeamTile->m_fv2);

			pxVertex[1].x = axVertexPos[1].x;
			pxVertex[1].y = axVertexPos[1].y;
			pxVertex[1].z = axVertexPos[1].z;
			pxVertex[1].m_uColour = uColour;
			pxVertex[1].u = pxBeamTile->m_fu1;
			pxVertex[1].v = pxBeamTile->m_fv1;

			pxVertex[2].x = axVertexPos[2].x;
			pxVertex[2].y = axVertexPos[2].y;
			pxVertex[2].z = axVertexPos[2].z;
			pxVertex[2].m_uColour = uColour;
			pxVertex[2].u = 0.5f * (pxBeamTile->m_fu1 + pxBeamTile->m_fu2);
			pxVertex[2].v = pxBeamTile->m_fv1;

			pxVertex[3].x = axVertexPos[3].x;
			pxVertex[3].y = axVertexPos[3].y;
			pxVertex[3].z = axVertexPos[3].z;
			pxVertex[3].m_uColour = uColour;
			pxVertex[3].u = pxBeamTile->m_fu1;
			pxVertex[3].v = pxBeamTile->m_fv1;

			Asura_Xbox360_Graphics::GetDevice()->EndVertices();
		}

		// Now draw second point in distance

		axVertexPos[0] = xRight;
		axVertexPos[0] *= -fHeadRadius;
		axVertexPos[0] += xBeamEnd;

		axVertexPos[1] = xUp;
		axVertexPos[1] *= fHeadRadius;
		axVertexPos[1] += xBeamEnd;

		axVertexPos[2] = xRight;
		axVertexPos[2] *= fHeadRadius;
		axVertexPos[2] += xBeamEnd;

		axVertexPos[3] = xUp;
		axVertexPos[3] *= -fHeadRadius;
		axVertexPos[3] += xBeamEnd;

		uColour = xHeadColour.ToUIntARGB();

		hResult = Asura_Xbox360_Graphics::GetDevice()->BeginVertices( D3DPT_TRIANGLEFAN, 4, sizeof( BeamVertex ), reinterpret_cast< void** >( &pxVertex ) );
		if( SUCCEEDED( hResult ) )
		{
			pxVertex[0].x = axVertexPos[0].x;
			pxVertex[0].y = axVertexPos[0].y;
			pxVertex[0].z = axVertexPos[0].z;
			pxVertex[0].m_uColour = uColour;
			pxVertex[0].u = pxBeamTile->m_fu2;
			pxVertex[0].v = 0.5f * (pxBeamTile->m_fv1 + pxBeamTile->m_fv2);

			pxVertex[1].x = axVertexPos[1].x;
			pxVertex[1].y = axVertexPos[1].y;
			pxVertex[1].z = axVertexPos[1].z;
			pxVertex[1].m_uColour = uColour;
			pxVertex[1].u = pxBeamTile->m_fu2;
			pxVertex[1].v = pxBeamTile->m_fv2;

			pxVertex[2].x = axVertexPos[2].x;
			pxVertex[2].y = axVertexPos[2].y;
			pxVertex[2].z = axVertexPos[2].z;
			pxVertex[2].m_uColour = uColour;
			pxVertex[2].u = 0.5f * (pxBeamTile->m_fu1 + pxBeamTile->m_fu2);
			pxVertex[2].v = pxBeamTile->m_fv2;

			pxVertex[3].x = axVertexPos[3].x;
			pxVertex[3].y = axVertexPos[3].y;
			pxVertex[3].z = axVertexPos[3].z;
			pxVertex[3].m_uColour = uColour;
			pxVertex[3].u = pxBeamTile->m_fu2;
			pxVertex[3].v = pxBeamTile->m_fv2;

			Asura_Xbox360_Graphics::GetDevice()->EndVertices();
		}
	}
	else
	{
		// Cross product ok to use
		xPerpVector = xBeamDir ^ xEyeVector;
		xPerpVector.Normalise();

		Asura_Vector_3 xUp = xCamForward ^ xPerpVector;
		xUp.Normalise();
		const Asura_Vector_3 xRight = xCamForward ^ xUp;

		// Vertex 0 at top and nearest
		//      + v0 
		// + v1     + v2
		// + v3     + v4
		//      + v5

		axVertexPos[0] = xUp;
		axVertexPos[0] *= fTailRadius;
		axVertexPos[0] += xBeamStart;

		axVertexPos[1] = xRight;
		axVertexPos[1] *= -fTailRadius;
		axVertexPos[1] += xBeamStart;

		axVertexPos[2] = xRight;
		axVertexPos[2] *= fTailRadius;
		axVertexPos[2] += xBeamStart;

		axVertexPos[3] = xRight;
		axVertexPos[3] *= -fHeadRadius;
		axVertexPos[3] += xBeamEnd;

		axVertexPos[4] = xRight;
		axVertexPos[4] *= fHeadRadius;
		axVertexPos[4] += xBeamEnd;

		axVertexPos[5] = xUp;
		axVertexPos[5] *= -fHeadRadius;
		axVertexPos[5] += xBeamEnd;

		hResult = Asura_Xbox360_Graphics::GetDevice()->BeginVertices( D3DPT_TRIANGLESTRIP, 6, sizeof( BeamVertex ), reinterpret_cast< void** >( &pxVertex ) );
		if( SUCCEEDED( hResult ) )
		{
			uColour = xTailColour.ToUIntARGB();

			pxVertex[0].x = axVertexPos[0].x;
			pxVertex[0].y = axVertexPos[0].y;
			pxVertex[0].z = axVertexPos[0].z;
			pxVertex[0].m_uColour = uColour;
			pxVertex[0].u = pxBeamTile->m_fu1;
			pxVertex[0].v = pxBeamTile->m_fv1;

			pxVertex[1].x = axVertexPos[1].x;
			pxVertex[1].y = axVertexPos[1].y;
			pxVertex[1].z = axVertexPos[1].z;
			pxVertex[1].m_uColour = uColour;
			pxVertex[1].u = pxBeamTile->m_fu1;
			pxVertex[1].v = 0.5f * (pxBeamTile->m_fv1 + pxBeamTile->m_fv2);

			pxVertex[2].x = axVertexPos[2].x;
			pxVertex[2].y = axVertexPos[2].y;
			pxVertex[2].z = axVertexPos[2].z;
			pxVertex[2].m_uColour = uColour;
			pxVertex[2].u = 0.5f * (pxBeamTile->m_fu1 + pxBeamTile->m_fu2);
			pxVertex[2].v = pxBeamTile->m_fv1;

			uColour = xHeadColour.ToUIntARGB();

			pxVertex[3].x = axVertexPos[3].x;
			pxVertex[3].y = axVertexPos[3].y;
			pxVertex[3].z = axVertexPos[3].z;
			pxVertex[3].m_uColour = uColour;
			pxVertex[3].u = 0.5f * (pxBeamTile->m_fu1 + pxBeamTile->m_fu2);
			pxVertex[3].v = pxBeamTile->m_fv2;

			pxVertex[4].x = axVertexPos[4].x;
			pxVertex[4].y = axVertexPos[4].y;
			pxVertex[4].z = axVertexPos[4].z;
			pxVertex[4].m_uColour = uColour;
			pxVertex[4].u = pxBeamTile->m_fu2;
			pxVertex[4].v = 0.5f * (pxBeamTile->m_fv1 + pxBeamTile->m_fv2);

			pxVertex[5].x = axVertexPos[5].x;
			pxVertex[5].y = axVertexPos[5].y;
			pxVertex[5].z = axVertexPos[5].z;
			pxVertex[5].m_uColour = uColour;
			pxVertex[5].u = pxBeamTile->m_fu2;
			pxVertex[5].v = pxBeamTile->m_fv2;

			Asura_Xbox360_Graphics::GetDevice()->EndVertices();
		}
	}
}

/******************************************************************************/
/* Asura_Beam_System Static Class Functions                                   */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Beam_System::Platform_Initialise()
//
// Initialises the beam system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Beam_System::Platform_Initialise()
{
	g_pxBeamEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auBeam);

	g_hEffectBaseTexture = g_pxBeamEffect->GetParameterHandle( "BaseTexture" );
	g_hEffectViewProj = g_pxBeamEffect->GetParameterHandle( "ViewProj" );

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxBeamVD );
	Asura_Assert( g_pxBeamVD, "Failed to create beam vertex declaration." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Beam_System::Platform_Shutdown()
//
// Shuts down the beam system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Beam_System::Platform_Shutdown()
{
	g_pxBeamVD->Release();
	g_pxBeamVD = NULL;

	Asura_Xbox360_Effect_System::Destroy( g_pxBeamEffect );
	g_pxBeamEffect = NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Beam_System::Platform_RenderStart()
//
// Begins rendering beams on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Beam_System::Platform_RenderStart()
{
	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF ), "Beams" );

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxBeamVD );

	g_pxBeamEffect->GetInterface()->SetSampler( g_hEffectBaseTexture, Asura_Xbox360_TextureManagement::GetTexture( s_iTiledTextureTextureID )->GetD3DTexture() );
	g_pxBeamEffect->GetInterface()->SetMatrixF( g_hEffectViewProj, Asura_Xbox360_Graphics::GetViewProjectionMatrix() );

	// If the project doesn't allow individual beams to change translucency mode, use glowing for all

	#if !defined ASURA_BEAMS_ALL_MODES
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
	#endif

	g_pxBeamEffect->Begin();
	g_pxBeamEffect->Commit();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Beam_System::Platform_Render()
//
// Renders beams on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Beam_System::Platform_Render()
{
	const u_int uNumCameras = Asura_Camera::GetNumberOfCameras();

	for( u_int u = 0; u < uNumCameras; ++u )
	{
		PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xC0, 0xC0, 0xC0 ), "Camera %u", u );

		Asura_Camera::SetCameraIndex(u);
		Asura_Render::ApplyCameraSettings();

		for( u_int v = 0; v < GetNumActiveBeams(); ++v )
		{
			if( !s_apxBeamPtrs[v] || s_apxBeamPtrs[v]->IsInvisible() ) continue;

			#if defined ASURA_BEAMS_SUPPORT_CAMERA_SELECTION

				if( s_apxBeamPtrs[v]->m_uDontRenderIfCameraIgnoresThisGuid == Asura_Camera::GetEntityToIgnore()
				 || (s_apxBeamPtrs[v]->m_uOnlyRenderIfCameraIgnoresThisGuid != ASURA_GUID_UNREGISTERED && s_apxBeamPtrs[v]->m_uOnlyRenderIfCameraIgnoresThisGuid != Asura_Camera::GetEntityToIgnore()) )
				{
					continue;
				}

			#endif // defined ASURA_BEAMS_SUPPORT_CAMERA_SELECTION

			PIXBeginNamedEvent( D3DCOLOR_XRGB( 0x80, 0x80, 0x80 ), "Beam %u", v );
			RenderBeam( *s_apxBeamPtrs[v], s_fEndEffectAngleTolerance );
			s_apxBeamPtrs[v]->SetDontUpdateUntilRenderedOnce( false );
			PIXEndNamedEvent(); // Beam
		}

		PIXEndNamedEvent(); // Camera
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Beam_System::Platform_RenderEnd()
//
// Finishes rendering beams on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Beam_System::Platform_RenderEnd()
{
	g_pxBeamEffect->End();

	PIXEndNamedEvent(); // Beams
}
