
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_Decal_System.cpp
// Author : Dave Reed
// Date   : 14-Dec-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

// Asura header-files...

#include "Core/Asura.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_HalfFloat.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Light/Asura_GI_Solution.h"
#include "SpecialFX/Asura_Decal_System.h"
#include "SpecialFX/Asura_Decal_Template.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Module.h"
#include "Camera/Asura_Camera.h"

#include "SpecialFX/Asura_Decal_Manager.h"

#include "Debugging/Asura_Logfile.h"

#include "Debugging/Asura_Platform_GPUProfile.h"
#include "SpecialFX/Asura_Platform_Decal_System.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"

// Shader Include Files
#include "Decals.h"
// Misc
#include <xtl.h>

//#define ASURA_XBOX360_DECAL_DEBUGGING

/******************************************************************************/
/* Macros                                                                     */
/******************************************************************************/

#define VECTOR_TO_UINT( xVector )							\
(															\
	  (static_cast< u_int >( xVector.x * 511.0f ) <<  0)	\
	+ (static_cast< u_int >( xVector.y * 511.0f ) << 10)	\
	+ (static_cast< u_int >( xVector.z * 511.0f ) << 20)	\
)


const Direct3D_VertexElement g_axDecalVertexDeclaration[] =
{
	// Using packed formats where possible on 360
	{ 0,  0, D3DDECLTYPE_FLOAT3,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,     0 },
	{ 0, 12, D3DDECLTYPE_FLOAT4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     0 },
	{ 0, 28, D3DDECLTYPE_DEC3N,		D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,		  0 },
	{ 0, 32, D3DDECLTYPE_DEC3N,		D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,	  0 },
	{ 0, 36, D3DDECLTYPE_DEC3N,		D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL,	  0 },	
	{ 0, 40, D3DDECLTYPE_FLOAT16_4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  1 },	 // SH Red
	{ 0, 48, D3DDECLTYPE_FLOAT16_4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  2 },	 // SH Green
	{ 0, 56, D3DDECLTYPE_FLOAT16_4,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  3 },	 // SH Blue


	D3DDECL_END()
};

static Asura_Xbox360_MaterialResponse_Handles g_xMaterialResponseHandles;



// ************************************************************************************************
// Asura_Xbox360_Decal_System                                                 
// ************************************************************************************************

// Asura_Decal_System - Platform functions redirected to Asura_Xbox360_Decal_System
void Asura_Decal_System::Platform_Initialise( void )			{Asura_Xbox360_Decal_System::Initialise();			}
void Asura_Decal_System::Platform_Shutdown( void )				{Asura_Xbox360_Decal_System::Shutdown();			}
void Asura_Decal_System::Platform_Reset( void )					{Asura_Xbox360_Decal_System::Reset();				}

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
void Asura_Decal_System::Platform_RenderStaticDecals( void )    {Asura_Xbox360_Decal_System::RenderStaticDecals();  }
#endif

// And the declaration objects
Direct3D_VertexDeclaration*				Asura_Xbox360_Decal_System::s_pxDecalVertexDecl		  = NULL;

// And the shader
Asura_Xbox360_Effect*					Asura_Xbox360_Decal_System::s_pxDecalEffect = NULL;
Asura_Xbox360_Decal_ShaderParameters	Asura_Xbox360_Decal_System::s_xShaderParams;

// Static decal data
int										Asura_Xbox360_Decal_System::s_iNumStaticDecalBuffers = 0;
Asura_Xbox360_StaticDecalBuffer*		Asura_Xbox360_Decal_System::s_pxStaticDecalBuffers = NULL;


// ------------------------------------------------------------------------------------------------
// Asura_Xbox360_Decal_System::Initialise( )
// ------------------------------------------------------------------------------------------------

void Asura_Xbox360_Decal_System::Initialise( void )
{		
	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration(g_axDecalVertexDeclaration, &s_pxDecalVertexDecl);
	Asura_Assert(s_pxDecalVertexDecl, "Decal renderer couldn't create vertex decl");	

	s_pxDecalEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auDecals);
	Asura_Assert(s_pxDecalEffect, "Decal renderer couldn't create effect");

	if (s_pxDecalEffect)
	{
		s_xShaderParams.Get(s_pxDecalEffect);
	}
}


// ------------------------------------------------------------------------------------------------
// Asura_Xbox360_Decal_System::Reset( )
// ------------------------------------------------------------------------------------------------

void Asura_Xbox360_Decal_System::Reset( void ) 
{
	delete[] s_pxStaticDecalBuffers;		
	s_pxStaticDecalBuffers = NULL;
	s_iNumStaticDecalBuffers = 0;
}


// ------------------------------------------------------------------------------------------------
// Asura_Xbox360_Decal_System::Shutdown( )
// ------------------------------------------------------------------------------------------------

void Asura_Xbox360_Decal_System::Shutdown( void ) 
{
	if (s_pxStaticDecalBuffers)
	{
		delete [] s_pxStaticDecalBuffers;		
		s_pxStaticDecalBuffers = NULL;
		s_iNumStaticDecalBuffers = 0;
	}

	if (s_pxDecalEffect)
	{
		Asura_Xbox360_Effect_System::Destroy(s_pxDecalEffect);
	}

	if (s_pxDecalVertexDecl)
	{
		s_pxDecalVertexDecl->Release();
		s_pxDecalVertexDecl=NULL;
	}
}


// ------------------------------------------------------------------------------------------------
// RenderStaticDecals( )
// ------------------------------------------------------------------------------------------------

void Asura_Xbox360_Decal_System::RenderStaticDecals(void)
{

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_DECALS);
	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xA0, 0xFF, 0x80 ), "Static Decals" );

	PrepareRenderStates();

	const float afFadeValues[4] = {5,1,1,10};

	for(int i=0; i<s_iNumStaticDecalBuffers; i++)
	{
		Asura_Xbox360_StaticDecalBuffer *pxBuffer = &s_pxStaticDecalBuffers[i];

		if (pxBuffer->m_iDecalMaterialIndex<0)
		{			
			pxBuffer->m_iDecalMaterialIndex = Asura_Decal_Manager::GetMaterialIndex(pxBuffer->m_uDecalMaterialHash);						
		}
		
		if (pxBuffer->m_iDecalMaterialIndex>=0)
		{		
			const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(pxBuffer->m_iDecalMaterialIndex+1);
			const Asura_DecalManager_Material *pxMgrMtl = Asura_Decal_Manager::GetDecalMaterialByIndex(pxBuffer->m_iDecalMaterialIndex);

			if (pxBuffer->m_iNumVertices>0 && ((pxBuffer->m_iModuleIndex<0 || Asura_Module_System::IsModuleVisible(pxBuffer->m_iModuleIndex))))
			{
				bool abAABBInFrustum[ASURA_MAX_NO_OF_CAMERAS];
				int  iNumCamerasVisibleFrom = 0;

				for(int j=0; j<Asura_Camera::GetNumberOfCameras(); j++)
				{
					Asura_Camera::SetCameraIndex(j);
					abAABBInFrustum[j] = Asura_Camera::IsAABBInFrustum(pxBuffer->m_xAABB);

					if (abAABBInFrustum[j])
					{					
						iNumCamerasVisibleFrom++;
					}
				}

				if (iNumCamerasVisibleFrom>0)
				{
					if (pxMgrMtl->m_bStaticDecalBackFaceCull)
					{
						Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW); 
					}
					else
					{				
						Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE); 
					}

					int iVertSize;

					Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration(Asura_Xbox360_Decal_System::GetVertexDecl());
					iVertSize = sizeof(Asura_Xbox360_Decal_Vertex);

					for(int iPass=0;iPass<2;iPass++)
					{
						Asura_Xbox360_Effect *pxEffect = Asura_Xbox360_Decal_System::GetEffect();

						pxEffect->Begin(0,iPass);

						// Set the shader parameters
						Asura_Xbox360_Decal_System::SetupMaterial(iPass, pxMtl, (float *)afFadeValues);

						// Render them
						for(int j=0; j<Asura_Camera::GetNumberOfCameras(); j++)
						{
							Asura_Camera::SetCameraIndex(j);

							if (Asura_Camera::IsCameraActive() && abAABBInFrustum[j])
							{
								Asura_Render::ApplyCameraSettings();					
								pxEffect->Commit();
								pxBuffer->m_xVertexBuffer.Bind(0);
								pxBuffer->m_xIndexBuffer.Bind();							
								Asura_Xbox360_Graphics::GetDevice()->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,0,pxBuffer->m_iNumVertices,0,pxBuffer->m_iNumIndices/3);
							}
						}					

						// Flush any effects we may still be using
						pxEffect->End();

					}
				}
			}
		}
	}

	RestoreRenderStates();
	
	PIXEndNamedEvent();
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_DECALS);
#endif
	
}


// ------------------------------------------------------------------------------------------------
// AllocateStaticDecalBuffers() - Called from the resource chunk loader
// ------------------------------------------------------------------------------------------------

void Asura_Xbox360_Decal_System::AllocateStaticDecalBuffers(int iNumBuffers)
{
	if (s_pxStaticDecalBuffers)
	{
		delete [] s_pxStaticDecalBuffers;
		s_pxStaticDecalBuffers = NULL;
		s_iNumStaticDecalBuffers = 0;
	}
	

	s_pxStaticDecalBuffers = new Asura_Xbox360_StaticDecalBuffer[iNumBuffers];
	s_iNumStaticDecalBuffers = iNumBuffers;
}

// ------------------------------------------------------------------------------------------------
// GetStaticDecalBuffer()
// ------------------------------------------------------------------------------------------------

Asura_Xbox360_StaticDecalBuffer* Asura_Xbox360_Decal_System::GetStaticDecalBuffer(int iIndex)
{
	Asura_Assert(iIndex>=0 && iIndex<s_iNumStaticDecalBuffers, "Invalid staticdecal buffer index!");
	return &s_pxStaticDecalBuffers[iIndex];
}


// ------------------------------------------------------------------------------------------------
// PrepareRenderStates()
// ------------------------------------------------------------------------------------------------
void Asura_Xbox360_Decal_System::PrepareRenderStates(void)
{
	// (AAM - 01.04.2008) Force this in case the state cache has been upset
	Asura_Render::CurrentStates.RequestAlphaTest(true);

	// Force this to update, in case a .fx has messed with the states before it gets here
	Asura_Render::CurrentStates.RequestAlphaTest(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);

	// Z-bias them forwards a bit
	Asura_Xbox360_Graphics::SetDecalDepthBias();

	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	pxDevice->SetRenderState_Inline( D3DRS_ALPHABLENDENABLE, TRUE);

	// Don't blend into the depth/material buffer	
	pxDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, 0);		

	// We need this for blending into the 10:10:10:2 normal buffer
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE,  TRUE );
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE1, TRUE );
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE2, TRUE );
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE3, TRUE );
}



// ------------------------------------------------------------------------------------------------
// RestoreRenderStates()
// ------------------------------------------------------------------------------------------------
void Asura_Xbox360_Decal_System::RestoreRenderStates(void)
{
	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE,  FALSE );
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE1, FALSE );
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE2, FALSE );
	pxDevice->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE3, FALSE );

	// Restore states
	Asura_Xbox360_Graphics::ResetDepthBias();
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);	

	D3DBLENDSTATE xNewBlendState;

	pxDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALL);		
	pxDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL);		
	pxDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL);		
	pxDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL);		
	xNewBlendState.SrcBlend = D3DBLEND_ONE;
	xNewBlendState.BlendOp = D3DBLENDOP_ADD;
	xNewBlendState.DestBlend = D3DBLEND_ZERO;	
	xNewBlendState.SrcBlendAlpha = D3DBLEND_ONE;	
	xNewBlendState.BlendOpAlpha = D3DBLENDOP_ADD;
	xNewBlendState.DestBlendAlpha = D3DBLEND_ZERO;
	pxDevice->SetBlendState( 0, xNewBlendState );
	pxDevice->SetBlendState( 1, xNewBlendState );
	pxDevice->SetBlendState( 2, xNewBlendState );
	pxDevice->SetBlendState( 3, xNewBlendState );
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);

}




void  Asura_Xbox360_Decal_System::SetupMaterial(int iPass, const Asura_Decal_Material *pxMtl, float *pfFadeValues)
{

	Asura_Xbox360_Effect *pxEffect = Asura_Xbox360_Decal_System::GetEffect();
	Asura_Xbox360_Decal_ShaderParameters *pxParams = Asura_Xbox360_Decal_System::GetEffectParameters();
	FXLEffect *pxD3DEffect = pxEffect->GetInterface();

	const Asura_MaterialResponse* const pxMaterial 
		= Asura_MaterialResponse_System::GetMaterialByHashID( pxMtl->m_uMaterialResponseHash );

	if (pxMaterial && iPass == 0)
	{
		pxParams->SetMRParameters(pxEffect, pxMaterial);
	}

	const Asura_MaterialResponse::BLEND_MODE eBlendMode = pxMaterial
		? pxMaterial->GetBlendMode()
		: Asura_MaterialResponse::BLEND_MODE_OPAQUE;

	D3DBLENDSTATE xNewBlendState;
	xNewBlendState.BlendOp = D3DBLENDOP_ADD;
	xNewBlendState.BlendOpAlpha = D3DBLENDOP_ADD;
	xNewBlendState.SrcBlend = D3DBLEND_SRCALPHA;
	if (eBlendMode == Asura_MaterialResponse::BLEND_MODE_ADDITIVE)
	{
		xNewBlendState.DestBlend = D3DBLEND_ONE;	
	} 
	else
	{
		xNewBlendState.DestBlend = D3DBLEND_INVSRCALPHA;	
	}
	if (iPass == 0)
	{
		xNewBlendState.SrcBlendAlpha = D3DBLEND_ZERO;
		xNewBlendState.DestBlendAlpha = D3DBLEND_INVSRCALPHA;
	}
	else
	{
		xNewBlendState.SrcBlendAlpha = D3DBLEND_ONE;
		xNewBlendState.DestBlendAlpha = D3DBLEND_ONE;
	}
	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();
	if (iPass == 0)
	{
		pxDevice->SetBlendState( 0, xNewBlendState );
		pxDevice->SetBlendState( 1, xNewBlendState );
		pxDevice->SetBlendState( 2, xNewBlendState );

		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE);
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL);

		if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback())
		{
			Asura_MaterialResponse_System::GetMaterialProjectContextCallback()(
				pxMaterial ? pxMaterial->GetProjectContext() : ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID, 
				ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
		}


		pxD3DEffect->SetVectorF(pxParams->m_xFadeValues, pfFadeValues);

		// UV scrolling
		float afUVScroll[4];

		if (pxMtl->m_fUScrollSpeed>0.0f || pxMtl->m_fVScrollSpeed>0.0f)
		{	
			float fTimer = Asura_Timers::GetLocalScaledTime();		
			afUVScroll[0] = Asura_Maths::Sine(fTimer * pxMtl->m_fUScrollSpeed) * pxMtl->m_fUScrollRange;
			afUVScroll[1] = Asura_Maths::Cosine(fTimer * pxMtl->m_fVScrollSpeed) * pxMtl->m_fVScrollRange;
		}
		else
		{
			afUVScroll[0] = afUVScroll[1] = 0;
		}

		pxD3DEffect->SetVectorF(pxParams->m_xUVScroll, afUVScroll);


	}
	else
	{
		pxDevice->SetBlendState( 2, xNewBlendState );

		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE, 0);
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALPHA);

	}

	if (pxMaterial->GetTexture(Asura_MaterialResponse::NORMAL_MAP) != -1 && iPass == 0)
	{
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE);
	} 
	else 
	{
		pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, 0);
	}

}

// ------------------------------------------------------------------------------------------------
// Platform_AddStrip( )
// ------------------------------------------------------------------------------------------------

void Asura_Decal_Handler::Platform_AddStrip (const Asura_Decal_Vertex* pxStrip, u_int uNumVertices, const Asura_RGBA& xRGBA, const Asura_Decal_Template* pxDecalTemplate) 
{
	Asura_Xbox360_DecalBuffer* pxDecalBuffer = static_cast<Asura_Xbox360_DecalBuffer*>(m_pxFirstDecalBuffer);

	// Do we have a decal buffer for the current decal material?
	while (pxDecalBuffer)
	{
		if (pxDecalBuffer->m_iDecalMaterialIndex == pxDecalTemplate->m_iDecalMaterialIndex)
		{			
			pxDecalBuffer->AddStrip(pxStrip, uNumVertices, pxDecalTemplate, xRGBA);
			return;
		}
		pxDecalBuffer = pxDecalBuffer->m_pxNextDecalBuffer;
	}

	// Need to create a new buffer
	pxDecalBuffer = new Asura_Xbox360_DecalBuffer(pxDecalTemplate);
	pxDecalBuffer->m_pxNextDecalBuffer = static_cast<Asura_Xbox360_DecalBuffer*>(m_pxFirstDecalBuffer);
	m_pxFirstDecalBuffer = pxDecalBuffer;
	pxDecalBuffer->AddStrip(pxStrip,uNumVertices,pxDecalTemplate,xRGBA);
}





// ************************************************************************************************
// Asura_Xbox360_Buffer 
// This stores all decals for one material type (which may be more than one decal template)
// Has a circular buffer of verts, and an index buffer that gets shifted down when full
// ************************************************************************************************


Asura_Xbox360_DecalBuffer::Asura_Xbox360_DecalBuffer(const Asura_Decal_Template* pxDecalTemplate)
{
	const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(pxDecalTemplate->m_iDecalMaterialIndex);

	m_iDecalMaterialIndex = pxDecalTemplate->m_iDecalMaterialIndex;
	m_iMaxDecals		  = pxMtl->m_iMaxDecals;		
	m_apxDecals			  = new Decal[m_iMaxDecals];
	m_iNumVerts			  = m_iMaxDecals*8;		
	m_iNumIndices		  = m_iMaxDecals*18;	
	m_apxIndices		  = new uint16[m_iNumIndices];							 // Worst-case is 18 indices for a decal  (if all 8 verts are used)
	m_iFirstDecal		  = 0;
	m_iNumDecals		  = 0;
	m_iNextVertex		  = 0;
	m_iNumDecalTris		  = 0;									 
	m_pxNextDecalBuffer   = NULL;		
	m_iCurrentIndexBuffer = 0;
	m_bIndexBufferInvalid = false;

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexBuffer(m_iNumVerts*sizeof(Asura_Xbox360_Decal_Vertex),0,0,0,&m_pxVB,0);
	Asura_Assert(m_pxVB, "Decal system failed to create vertex buffer!");

	for(int i=0;i<2;i++)
	{		
		Asura_Xbox360_Graphics::GetDevice()->CreateIndexBuffer(m_iNumIndices,D3DUSAGE_CPU_CACHED_MEMORY,D3DFMT_INDEX16,0,&m_apxIB[i],0);
		Asura_Assert(m_apxIB[i], "Decal system failed to create index buffer!");
	}
}



Asura_Xbox360_DecalBuffer::~Asura_Xbox360_DecalBuffer()
{
	delete m_pxNextDecalBuffer;

	SCOPED_DEVICE_OWNERSHIP();

	Asura_Xbox360_Graphics::GetDevice()->SetIndices(NULL);
	Asura_Xbox360_Graphics::GetDevice()->SetStreamSource(0, NULL, 0, 0);

	if (m_pxVB)
	{
		m_pxVB->Release();
	}

	if (m_apxIB)
	{
		m_apxIB[0]->Release();
		m_apxIB[1]->Release();
	}

	delete [] m_apxDecals;		
	delete [] m_apxIndices;
}



void Asura_Xbox360_DecalBuffer::AddStrip(const Asura_Decal_Vertex* pxStrip, const u_int uNumberOfVertices, const Asura_Decal_Template* pxDecalTemplate, const Asura_RGBA& xRGBA )
{
#ifdef ASURA_XBOX360_DECAL_DEBUGGING
	Asura_OutputToDebugger(("Adding Decal - %d tris\n",uNumberOfVertices-2));
#endif
	if (!m_pxVB)
	{
		// May have no VB at this point if we were out of memory...
		return;
	}

	if (m_iNumDecals>=m_iMaxDecals)
	{
#ifdef ASURA_XBOX360_DECAL_DEBUGGING
		Asura_OutputToDebugger(("Removing %d tris (decal #%d)\n",(m_apxDecals[m_iFirstDecal].iNumIndices/3),m_iFirstDecal));
#endif
		// No room - flush the first one
		m_iNumDecalTris -= (m_apxDecals[m_iFirstDecal].iNumIndices/3);
		m_iFirstDecal++;
		if (m_iFirstDecal>=m_iMaxDecals)
		{
			m_iFirstDecal=0;
		}
		m_iNumDecals--;				
	}		

	// Is there space in the index buffer?
	int iFirstIndex = 0;
	if (m_iNumDecals>0)
	{		
		int iLastDecal = (m_iFirstDecal+m_iNumDecals-1);
		if (iLastDecal>=m_iMaxDecals)
		{
			iLastDecal-=m_iMaxDecals;
		}

		iFirstIndex = m_apxDecals[iLastDecal].iFirstIndex+
			m_apxDecals[iLastDecal].iNumIndices;

		// De we need to shift the index buffer down
		if (iFirstIndex + ((int)((uNumberOfVertices-2)*3)) >= m_iNumIndices)
		{					
			int iShiftDown = m_apxDecals[m_iFirstDecal].iFirstIndex;			
			int iSrc	   = iShiftDown;
			int iDst	   = 0;
			int iDecal	   = m_iFirstDecal;

			for(int i=0;i<m_iNumDecalTris*3;i++)
			{				
				m_apxIndices[iDst++]=m_apxIndices[iSrc++];
			}

			for(int i=0; i<m_iNumDecals; i++)
			{										
				m_apxDecals[iDecal].iFirstIndex-=iShiftDown;										
				iDecal++;
				if (iDecal>=m_iMaxDecals)
				{
					iDecal-=m_iMaxDecals;
				}
			}	

			iFirstIndex = m_apxDecals[iLastDecal].iFirstIndex+
				m_apxDecals[iLastDecal].iNumIndices;
		}
	}

	// Fill in the decal struct
	int iDecal = m_iFirstDecal+m_iNumDecals;			
	if (iDecal>=m_iMaxDecals)
	{
		iDecal-=m_iMaxDecals;
	}
	m_apxDecals[iDecal].iFirstIndex		= iFirstIndex;
	m_apxDecals[iDecal].iNumIndices		= (uNumberOfVertices-2)*3;
	m_apxDecals[iDecal].fCreationTime	= Asura_Timers::GetLocalScaledTime();
	m_iNumDecals++;


	// Fill in the verts+indices		
	if ((m_iNextVertex+(int)uNumberOfVertices)>m_iNumVerts)
	{
		m_iNextVertex = 0;
	}

	// Compute the tangent/binormal
	// We take a single normal for the decal now, to avoid a discontinuity when the
	// decal is 'bent' over a polygon edge.
	Asura_Vector_3 xNormal,xTangent,xBinormal;
	xNormal = Asura_Decal_System::GetCurrentDecalNormal();
	xNormal.Normalise();
	{		
		// There's room for some optimization here...
		Asura_Vector_3 axPositions[3];
		Asura_Vector_2 axUVs[3];
		for(int i=0;i<3;i++)
		{			
			axPositions[i] = pxStrip[i].m_xPosition;
			axUVs[i].x	   = pxStrip[i].m_fU;
			axUVs[i].y     = 1.0f - pxStrip[i].m_fV;
		}
		Asura_Xbox360_Decal_System::CalculateTangentAndBinormal(axPositions, axUVs, xTangent, xBinormal);

		// The calculation above is incomplete, as it doesn't take the normal into account... so finish the job:
		xBinormal.Normalise();
		xTangent  = xBinormal ^  xNormal;
		xTangent.Normalise();
		xBinormal = xNormal ^ xTangent;
		xBinormal.Normalise();
	}


	// Converting from a strip to an indexed list for now - avoids messing with degens, and
	// may actually be faster?
	if (m_iNextVertex+((int)uNumberOfVertices) >= m_iNumVerts)
	{
		m_iNextVertex = 0;
	}

	uint16 *pxIndices = &m_apxIndices[iFirstIndex];
	int     iIndex = 0;

	for(u_int i=2; i<uNumberOfVertices; i++)
	{	
		// Make it back-face-culling-friendly
		if (i&1)
		{
			pxIndices[iIndex++] = (uint16)(m_iNextVertex+i);			
			pxIndices[iIndex++] = (uint16)(m_iNextVertex+i-1);			
			pxIndices[iIndex++] = (uint16)(m_iNextVertex+i-2);			
		}
		else
		{			
			pxIndices[iIndex++] = (uint16)(m_iNextVertex+i-2);			
			pxIndices[iIndex++] = (uint16)(m_iNextVertex+i-1);			
			pxIndices[iIndex++] = (uint16)(m_iNextVertex+i);			
		}
	}


	// Lock the VB with NoOverwrite flag - as the VB is a circular buffer with space for the max number
	// of decals, so we'll never overwrite verts that the GPU hasn't yet rendered.
	Asura_Xbox360_Decal_Vertex *pxVert = NULL;
	m_pxVB->Lock(0, 0, (void**)&pxVert, D3DLOCK_NOOVERWRITE);
	if (pxVert==NULL) return;
	pxVert = &pxVert[m_iNextVertex];

	// Fill in the verts
	for(u_int i=0; i<uNumberOfVertices; i++)
	{						
		pxVert->fX = pxStrip[i].m_xPosition.x;
		pxVert->fY = pxStrip[i].m_xPosition.y;
		pxVert->fZ = pxStrip[i].m_xPosition.z;
		pxVert->fU = pxStrip[i].m_fU;
		pxVert->fV = pxStrip[i].m_fV;

		// (packing some of these values would be a good idea - at least on 360 - as PC support for packed verts seems poor)
		// Tangent frame

		pxVert->uNormal   = VECTOR_TO_UINT(xNormal);
		pxVert->uTangent  = VECTOR_TO_UINT(xTangent);
		pxVert->uBinormal = VECTOR_TO_UINT(xBinormal);

		// Lighting
		pxVert->fAlpha = xRGBA.m_fA;

		pxVert->afSHRed[0] = 0.0f;
		pxVert->afSHRed[1] = 0.0f;
		pxVert->afSHRed[2] = 0.0f;
		pxVert->afSHRed[3] = xRGBA.m_fR;

		pxVert->afSHGreen[0] = 0.0f;
		pxVert->afSHGreen[1] = 0.0f;
		pxVert->afSHGreen[2] = 0.0f;
		pxVert->afSHGreen[3] = xRGBA.m_fG;

		pxVert->afSHBlue[0] = 0.0f;
		pxVert->afSHBlue[1] = 0.0f;
		pxVert->afSHBlue[2] = 0.0f;
		pxVert->afSHBlue[3] = xRGBA.m_fB;
		pxVert->fCreationTime = m_apxDecals[iDecal].fCreationTime;

		pxVert++;			
		m_iNextVertex++;						
	}	

	m_pxVB->Unlock();

	m_iNumDecalTris+= (uNumberOfVertices-2);

	// Index buffer needs repopulating
	m_bIndexBufferInvalid=true;
}




void Asura_Xbox360_DecalBuffer::Render(void)
{	

	if (Asura_Xbox360_Decal_System::GetEffect()==NULL)
	{
		return;
	}

	// Set up the fade values		
	const Asura_Decal_Material *pxMtl = Asura_Decal_System::Project_GetDecalMaterial(m_iDecalMaterialIndex);				
	float fCurrentTime = Asura_Timers::GetLocalScaledTime();
	float afFadeValues[4] = {fCurrentTime, pxMtl->m_fFadeInTime,pxMtl->m_fFadeOutTime, pxMtl->m_fDecalStayTime};
	if (pxMtl->m_fDecalStayTime<0)
	{
		// Not 'forever', but should be long enough?
		// (AAM - 17.01.2007) If this value is too large, PC cards
		// don't render properly. Some register overflow perhaps?
		afFadeValues[3] = 100000.0f;
	}

	// Remove faded-out decals here - so it doesn't keep rendering them invisibly.
	while (m_iNumDecals>0 && (fCurrentTime > m_apxDecals[m_iFirstDecal].fCreationTime + (afFadeValues[3]+pxMtl->m_fFadeOutTime)))
	{
		// Decal is faded out, remove it
		m_iNumDecalTris -= (m_apxDecals[m_iFirstDecal].iNumIndices/3);
		m_iFirstDecal++;
		if (m_iFirstDecal>=m_iMaxDecals)
		{
			m_iFirstDecal=0;
		}
		m_iNumDecals--;			
		m_bIndexBufferInvalid = true;
	}			



	// Repopulate the index buffer. We do it here, not in the add decal function, as that can often occur more than
	// once in a frame, and doing this more than once would be bad (cause lock stalls).
	//
	// It's a bit annoying having to do it like this, but the 360 doesn't support range-locking of vert/index buffers		
	if (m_bIndexBufferInvalid)
	{		
		Asura_Xbox360_Decal_Vertex *pxIndices = NULL;
		m_iCurrentIndexBuffer=1-m_iCurrentIndexBuffer;
		m_apxIB[m_iCurrentIndexBuffer]->Lock(0,0, (void **)&pxIndices,0 );		
		if (pxIndices==NULL) return;
		int iFirstIndex = m_apxDecals[m_iFirstDecal].iFirstIndex;
		memcpy(&pxIndices[0], &m_apxIndices[iFirstIndex], m_iNumDecalTris*3*2);
		m_apxIB[m_iCurrentIndexBuffer]->Unlock();
		m_bIndexBufferInvalid = false;

#ifdef ASURA_XBOX360_DECAL_DEBUGGING
		Asura_OutputToDebugger(("Index Buffer: "));
		for(int i=0;i<(m_iNumDecalTris*3);i++)
		{			
			Asura_OutputToDebugger(("%d,",  m_apxIndices[i+iFirstIndex]));
		}
		Asura_OutputToDebugger(("\n"));
#endif
	}


	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration(Asura_Xbox360_Decal_System::GetVertexDecl());


	Asura_Xbox360_Effect *pxEffect = Asura_Xbox360_Decal_System::GetEffect();
	

	if (m_iNumDecals>0)
	{
		// Render them
		for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
		{
			Asura_Camera::SetCameraIndex(i);

			if (Asura_Camera::IsCameraActive())
			{
				Asura_Render::ApplyCameraSettings();					
				
				for (int iPass = 0; iPass < 2; ++iPass)
				{
					// Set the shader parameters
					Asura_Xbox360_Decal_System::SetupMaterial(iPass, pxMtl, afFadeValues);

					pxEffect->Begin(0,i);
					pxEffect->Commit();
					Asura_Xbox360_Graphics::GetDevice()->SetIndices(m_apxIB[m_iCurrentIndexBuffer]);
					Asura_Xbox360_Graphics::GetDevice()->SetStreamSource(0, m_pxVB, 0, sizeof(Asura_Xbox360_Decal_Vertex));
					Asura_Xbox360_Graphics::GetDevice()->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,0,m_iNumVerts,0,m_iNumDecalTris);
					pxEffect->End();
				}
			}
		}
	}


	if (m_pxNextDecalBuffer)
	{
		m_pxNextDecalBuffer->Render();
	}
	
}





// ------------------------------------------------------------------------------------------------
// Platform_Render( )
// ------------------------------------------------------------------------------------------------

void Asura_Decal_Handler::Platform_Render (void)
{
	if (m_pxFirstDecalBuffer == 0) return;

	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_DECALS);
	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0x80, 0xA0, 0x60 ), "Decals" );

	Asura_Xbox360_Decal_System::PrepareRenderStates();

	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE); 

	static_cast<Asura_Xbox360_DecalBuffer*>(m_pxFirstDecalBuffer)->Render();

	Asura_Xbox360_Decal_System::RestoreRenderStates();	

	PIXEndNamedEvent();
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_DECALS);
}





// Copied straight from Tools_Tangentspace.
static void Swap( float& xValue1, float& xValue2 ) { float xTemp = xValue1; xValue1 = xValue2; xValue2 = xTemp; }
static void Swap( Asura_Vector_3& xValue1, Asura_Vector_3& xValue2 ) { Asura_Vector_3 xTemp = xValue1; xValue1 = xValue2; xValue2 = xTemp; }

void Asura_Xbox360_Decal_System::CalculateTangentAndBinormal( const Asura_Vector_3* pxPositions,
															  const Asura_Vector_2* pxTextureCoords,
																	Asura_Vector_3& rxTangent,
																	Asura_Vector_3& rxBinormal )
{
	float			afU [3];
	float			afV [3];
	Asura_Vector_3	axPositions [3];

	// Initialise the local data...
	for (int i = 0; i < 3; i++)
	{
		axPositions [i]	= pxPositions [i];
		afU [i]			= pxTextureCoords [i].x;
		afV [i]			= pxTextureCoords [i].y;
	}

	// *** calc the tangent vector ***
	{
		// first sort vertices based on v coord
		if (afV [0] < afV [1])
		{
			Swap	(afU [0],			afU [1]);
			Swap	(afV [0],			afV [1]);
			Swap	(axPositions [0],	axPositions [1]);
		}
		if (afV [0] < afV [2])
		{
			Swap	(afU [0],			afU [2]);
			Swap	(afV [0],			afV [2]);
			Swap	(axPositions [0],	axPositions [2]);
		}
		if (afV [1] < afV [2])
		{
			Swap	(afU [1],			afU [2]);
			Swap	(afV [1],			afV [2]);
			Swap	(axPositions [1],	axPositions [2]);
		}

		// find offset along edge2 to the 'middle' v coord
		float fLambda;
		float fDivisor = (afV [2] - afV [0]);

		if (Asura_Maths::Abs(fDivisor) < 0.0001f)
		{
			fLambda = 1.0f;
		}
		else
		{
			fLambda = (afV [1] - afV [0]) / fDivisor;
		}

		// find the vertex position associated with this
		Asura_Vector_3 xPos = (axPositions [2] - axPositions [0]) * fLambda + axPositions [0];

		// tangent is the ray from the middle vertex to the lerped vertex
		rxTangent = xPos - axPositions [1];

		// make sure tangent points in the right direction
		float fInterpolatedU = (afU [2] - afU [0]) * fLambda + afU [0];

		if (fInterpolatedU < afU [1])
		{
			rxTangent = -rxTangent;
		}
	}

	// *** calc the binormal vector ***
	{
		// first sort vertices based on u coord
		if (afU [0] < afU [1])
		{
			Swap	(afU [0],			afU [1]);
			Swap	(afV [0],			afV [1]);
			Swap	(axPositions [0],	axPositions [1]);
		}
		if (afU [0] < afU [2])
		{
			Swap	(afU [0],			afU [2]);
			Swap	(afV [0],			afV [2]);
			Swap	(axPositions [0],	axPositions [2]);
		}
		if (afU [1] < afU [2])
		{
			Swap	(afU [1],			afU [2]);
			Swap	(afV [1],			afV [2]);
			Swap	(axPositions [1],	axPositions [2]);
		}

		// find offset along edge2 to the 'middle' v coord
		float fLambda;
		float fDivisor = (afU [2] - afU [0]);

		if (Asura_Maths::Abs(fDivisor) < 0.0001f)
		{
			fLambda = 1.0f;
		}
		else
		{
			fLambda = (afU [1] - afU [0]) / fDivisor;
		}

		// find the vertex position associated with this
		Asura_Vector_3 xPos = (axPositions [2] - axPositions [0]) * fLambda + axPositions [0];

		// binormal is the ray form the middle vertex to the lerped vertex
		rxBinormal = xPos - axPositions [1];

		// make sure binormal points in the right direction
		float fInterpolatedV = (afV [2] - afV [0]) * fLambda + afV [0];

		if (fInterpolatedV > afV [1])
		{
			rxBinormal = -rxBinormal;
		}
	}
}

