//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Skybox.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "SpecialFX/Asura_Skybox.h"

// Asura platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Render.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_IndexBuffer.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Rendering/Asura_Xbox360_VertexBuffer.h"

#include "Debugging/Asura_Platform_GPUProfile.h"

// Shader Include Files
#include "SkyBox.h"


// Misc
#include <fxl.h>
#include <xtl.h>

//#define ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER	1

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct SkyboxVertex
{
	float x, y, z;
	float u, v;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

enum SkyboxFaceID
{
	SKYBOX_FACE_DOWN,
	SKYBOX_FACE_FRONT,
	SKYBOX_FACE_BACK,
	SKYBOX_FACE_LEFT,
	SKYBOX_FACE_RIGHT,
	SKYBOX_FACE_UP,

	NUM_SKYBOX_FACES
};

static const char* const g_aszFaceNames[NUM_SKYBOX_FACES] =
{
	"Down",
	"Front",
	"Back",
	"Left",
	"Right",
	"Up"
};

static const u_int	g_uNumQuadsPerAxis		= 1;
static const u_int	g_uNumQuadsPerFace		= g_uNumQuadsPerAxis * g_uNumQuadsPerAxis;
#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
static const u_int	g_uNumVerticesPerFace	= (g_uNumQuadsPerAxis + 1) * (g_uNumQuadsPerAxis + 1);
static const u_int	g_uNumIndicesPerFace	= g_uNumQuadsPerFace * 4;
#else
static const u_int	g_uNumVerticesPerFace	= g_uNumQuadsPerFace * 4;
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
static const u_int	g_uTotalNumVertices		= g_uNumVerticesPerFace * NUM_SKYBOX_FACES;

// Remember that in Asura, up is negative. Also, the skybox interface is just
// broken. Front and back are switched, as are left and right
static const Asura_Vector_3	g_xSkyboxFrontFaceStartPosition(-1.0f, -1.0f,  1.0f);
static const Asura_Vector_3	g_xSkyboxFrontFaceDirection1(0.0f,  2.0f,  0.0f);
static const Asura_Vector_3	g_xSkyboxFrontFaceDirection2(2.0f,  0.0f,  0.0f);

const D3DVERTEXELEMENT9 axVERTEX_DECLARATION[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 12, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	D3DDECL_END()
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static int g_aiSkyboxTexID[ Asura_Skybox::ASURA_SKYBOX_NUM_TEXTURES ] = {0};

#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
static Asura_Xbox360_IndexBuffer g_xSkyboxIB;
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
static Asura_Xbox360_VertexBuffer g_xSkyboxVB;
static Asura_Xbox360_Effect* g_pxSkyboxEffect;

static Asura_Xbox360_Effect_System::Handle g_xSkyboxFactorParameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xFogParameter					= 0;
static Asura_Xbox360_Effect_System::Handle g_xDistanceParameter				= 0;
static Asura_Xbox360_Effect_System::Handle g_xTextureHandle					= 0;
static Asura_Xbox360_Effect_System::Handle g_xWorldViewProjHandle			= 0;
static Asura_Xbox360_Effect_System::Handle g_xFrustumToCamDir				= 0;
static Asura_Xbox360_Effect_System::Handle g_xSkyboxColourParameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xSunDir1Parameter				= 0;
static Asura_Xbox360_Effect_System::Handle g_xSunDir2Parameter				= 0;
static Asura_Xbox360_Effect_System::Handle g_xHazeGradientPower1Parameter	= 0;
static Asura_Xbox360_Effect_System::Handle g_xHazeGradientPower2Parameter	= 0;
static Asura_Xbox360_Effect_System::Handle g_xHazeColour1Parameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xHazeColour2Parameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xSunGradientPower1Parameter	= 0;
static Asura_Xbox360_Effect_System::Handle g_xSunGradientPower2Parameter	= 0;
static Asura_Xbox360_Effect_System::Handle g_xSunColour1Parameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xSunColour2Parameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xDisableHazeParameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xDisable2HazeParameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xDisableSunParameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xDisable2SunParameter			= 0;
static Asura_Xbox360_Effect_System::Handle g_xMirrorBelowHorizonParameter	= 0;
static Asura_Xbox360_Effect_System::Handle g_x2MirrorBelowHorizonParameter	= 0;

static Direct3D_VertexDeclaration* g_pxSkyboxVD = NULL;

/******************************************************************************/
/* Asura_Skybox Static Class Functions                                        */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
// Function     : SetXYZForFace
// Author       : Andy Tate
// Date         : 23-Jul-2007
////////////////////////////////////////////////////////////////////////////////

static SkyboxVertex* SetXYZForFace(const Asura_Vector_3& xStart,
								   const Asura_Vector_3& xIncrement1,
								   const Asura_Vector_3& xIncrement2,
								   SkyboxVertex* pxOutputVertex)
{
#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
	for(u_int x = 0; x < (g_uNumQuadsPerAxis + 1); x++)
	{
		Asura_Vector_3	xXYZ(xStart + (xIncrement1 * static_cast<float>(x)));

		for(u_int y = 0; y < (g_uNumQuadsPerAxis + 1); y++, pxOutputVertex++)
		{
			// Generate this quad...
			pxOutputVertex->x	= xXYZ.GetX();
			pxOutputVertex->y	= xXYZ.GetY();
			pxOutputVertex->z	= xXYZ.GetZ();

			// Increment the values...
			xXYZ	+= xIncrement2;
		}
	}
#else
	for(u_int x = 0; x < g_uNumQuadsPerAxis; x++)
	{
		Asura_Vector_3	xXYZ0(xStart + (xIncrement1 * static_cast<float>(x)));
		Asura_Vector_3	xXYZ1(xXYZ0 + xIncrement2);
		Asura_Vector_3	xXYZ2(xXYZ0 + xIncrement1 + xIncrement2);
		Asura_Vector_3	xXYZ3(xXYZ0 + xIncrement1);

		for(u_int y = 0; y < g_uNumQuadsPerAxis; y++, pxOutputVertex += 4)
		{
			// Generate this quad...
			pxOutputVertex[0].x	= xXYZ0.GetX();
			pxOutputVertex[0].y	= xXYZ0.GetY();
			pxOutputVertex[0].z	= xXYZ0.GetZ();

			pxOutputVertex[1].x	= xXYZ1.GetX();
			pxOutputVertex[1].y	= xXYZ1.GetY();
			pxOutputVertex[1].z	= xXYZ1.GetZ();

			pxOutputVertex[2].x	= xXYZ2.GetX();
			pxOutputVertex[2].y	= xXYZ2.GetY();
			pxOutputVertex[2].z	= xXYZ2.GetZ();

			pxOutputVertex[3].x	= xXYZ3.GetX();
			pxOutputVertex[3].y	= xXYZ3.GetY();
			pxOutputVertex[3].z	= xXYZ3.GetZ();

			// Increment the values...
			xXYZ0	+= xIncrement2;
			xXYZ1	+= xIncrement2;
			xXYZ2	+= xIncrement2;
			xXYZ3	+= xIncrement2;
		}
	}
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)

	return(pxOutputVertex);
}

/// End of function 'SetXYZForFace'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : SetUVForFace
// Author       : Andy Tate
// Date         : 23-Jul-2007
////////////////////////////////////////////////////////////////////////////////

static SkyboxVertex* SetUVForFace(const Asura_Vector_2& xStart,
								  const Asura_Vector_2& xIncrement1,
								  const Asura_Vector_2& xIncrement2,
								  SkyboxVertex* pxOutputVertex)
{
#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
	for(u_int x = 0; x < (g_uNumQuadsPerAxis + 1); x++)
	{
		Asura_Vector_2	xUV(xStart + (xIncrement1 * static_cast<float>(x)));

		for(u_int y = 0; y < (g_uNumQuadsPerAxis + 1); y++, pxOutputVertex++)
		{
			// Generate this quad...
			pxOutputVertex->u	= xUV.x;
			pxOutputVertex->v	= xUV.y;

			// Increment the values...
			xUV	+= xIncrement2;
		}
	}
#else
	for(u_int x = 0; x < g_uNumQuadsPerAxis; x++)
	{
		Asura_Vector_2	xUV0(xStart + (xIncrement1 * static_cast<float>(x)));
		Asura_Vector_2	xUV1(xUV0 + xIncrement2);
		Asura_Vector_2	xUV2(xUV0 + xIncrement1 + xIncrement2);
		Asura_Vector_2	xUV3(xUV0 + xIncrement1);

		for(u_int y = 0; y < g_uNumQuadsPerAxis; y++, pxOutputVertex += 4)
		{
			// Generate this quad...
			pxOutputVertex[0].u	= xUV0.x;
			pxOutputVertex[0].v	= xUV0.y;

			pxOutputVertex[1].u	= xUV1.x;
			pxOutputVertex[1].v	= xUV1.y;

			pxOutputVertex[2].u	= xUV2.x;
			pxOutputVertex[2].v	= xUV2.y;

			pxOutputVertex[3].u	= xUV3.x;
			pxOutputVertex[3].v	= xUV3.y;

			// Increment the values...
			xUV0	+= xIncrement2;
			xUV1	+= xIncrement2;
			xUV2	+= xIncrement2;
			xUV3	+= xIncrement2;
		}
	}
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)

	return(pxOutputVertex);
}

/// End of function 'SetUVForFace'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : CreateVerticesForFace
// Author       : Andy Tate
// Date         : 20-Jul-2007
////////////////////////////////////////////////////////////////////////////////

static SkyboxVertex* CreateVerticesForFace(const Asura_Euler_3& xRotationEuler,
										   SkyboxVertex* pxOutputVertex)
{
	const Asura_Quat		xRotationQuat(xRotationEuler);
	const float				fIncrement	= 1.0f / g_uNumQuadsPerAxis;

	const Asura_Vector_2	xStartUV(0.0f, 0.0f);
	const Asura_Vector_2	xIncUV1(0.0f, fIncrement);
	const Asura_Vector_2	xIncUV2(fIncrement, 0.0f);

	Asura_Vector_3 xStartXYZ = xRotationQuat.RotateVector(g_xSkyboxFrontFaceStartPosition);
	Asura_Vector_3 xIncXYZ1 = xRotationQuat.RotateVector(Asura_Vector_3(g_xSkyboxFrontFaceDirection1 * fIncrement));
	Asura_Vector_3 xIncXYZ2 = xRotationQuat.RotateVector(Asura_Vector_3(g_xSkyboxFrontFaceDirection2 * fIncrement));

	SetXYZForFace(xStartXYZ, xIncXYZ1, xIncXYZ2, pxOutputVertex);
	return(SetUVForFace(xStartUV, xIncUV1, xIncUV2, pxOutputVertex));
}

/// End of function 'CreateVerticesForFace'. ///



#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)

////////////////////////////////////////////////////////////////////////////////
// Function     : CreateIndices
// Author       : Andy Tate
// Date         : 23-Jul-2007
////////////////////////////////////////////////////////////////////////////////

static u_short* CreateIndices(u_short* pusOutputIndex)
{
	u_short	usStartIndex	= 0;

	for(u_short x = 0; x < g_uNumQuadsPerAxis; x++)
	{
		const u_short	usNextStartIndex	= usStartIndex + (g_uNumQuadsPerAxis + 1);

		for(u_short y = 0; y < g_uNumQuadsPerAxis; y++)
		{
			*(pusOutputIndex++)	= usStartIndex + y;
			*(pusOutputIndex++)	= usStartIndex + y + 1;
			*(pusOutputIndex++)	= usNextStartIndex + y + 1;
			*(pusOutputIndex++)	= usNextStartIndex + y;
		}

		usStartIndex	= usNextStartIndex;
	}

	return(pusOutputIndex);
}

/// End of function 'CreateIndices'. ///

#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)



////////////////////////////////////////////////////////////////////////////////
//
// Asura_Skybox::Platform_DoRenderSetup()
//
// Called after each level is loaded to initialise skybox data
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_DoRenderSetup()
{
	const float	fIncrement	= 1.0f / g_uNumQuadsPerAxis;

	// Assume the textures have already been loaded in the level chunks

	for( u_int u = 0; u < ASURA_SKYBOX_NUM_TEXTURES; ++u )
	{
		g_aiSkyboxTexID[u] = Asura_Texture_Management::GetTextureIndex( GetTexture(u) );
	}

	// Need to check if some textures are repeats upside-down and adjust tex co-ords as necessary

	SkyboxVertex* const pxVertex = static_cast< SkyboxVertex* >( g_xSkyboxVB.Lock() );
	Asura_Assert( pxVertex, "Failed to lock skybox vertex buffer." );

	if( GetBackTextureIsFrontUpsideDown() )
	{
		const Asura_Vector_2	xStartUV(0.0f, 1.0f);
		const Asura_Vector_2	xIncUV1(0.0f, -fIncrement);
		const Asura_Vector_2	xIncUV2(fIncrement, 0.0f);

		SetUVForFace(xStartUV, xIncUV1, xIncUV2, &pxVertex[ASURA_SKYBOX_BK * g_uNumVerticesPerFace]);
	}
	else
	{
		const Asura_Vector_2	xStartUV(0.0f, 0.0f);
		const Asura_Vector_2	xIncUV1(0.0f, fIncrement);
		const Asura_Vector_2	xIncUV2(fIncrement, 0.0f);

		SetUVForFace(xStartUV, xIncUV1, xIncUV2, &pxVertex[ASURA_SKYBOX_BK * g_uNumVerticesPerFace]);
	}

	if( GetRightTextureIsLeftUpsideDown() )
	{
		const Asura_Vector_2	xStartUV(0.0f, 1.0f);
		const Asura_Vector_2	xIncUV1(0.0f, -fIncrement);
		const Asura_Vector_2	xIncUV2(fIncrement, 0.0f);

		SetUVForFace(xStartUV, xIncUV1, xIncUV2, &pxVertex[ASURA_SKYBOX_RT * g_uNumVerticesPerFace]);
	}
	else
	{
		const Asura_Vector_2	xStartUV(0.0f, 0.0f);
		const Asura_Vector_2	xIncUV1(0.0f, fIncrement);
		const Asura_Vector_2	xIncUV2(fIncrement, 0.0f);

		SetUVForFace(xStartUV, xIncUV1, xIncUV2, &pxVertex[ASURA_SKYBOX_RT * g_uNumVerticesPerFace]);
	}

	g_xSkyboxVB.Unlock();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Skybox::Platform_Initialise()
//
// Initialises 360 skybox structures
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_Initialise()
{
	// Create the vertex buffer...
	g_xSkyboxVB.CreateStatic( sizeof( SkyboxVertex ), g_uTotalNumVertices );

	SkyboxVertex*	pxVertex = static_cast< SkyboxVertex* >( g_xSkyboxVB.Lock() );
	Asura_Assert( pxVertex, "Unable to lock skybox vertex buffer." );
	{
		const Asura_Euler_3	xFrontRotation(0.0f, 0.0f, 0.0f);
		const Asura_Euler_3	xLeftRotation(0.0f, Asura_Maths::Deg270, 0.0f);
		const Asura_Euler_3	xBackRotation(0.0f, Asura_Maths::Deg180, 0.0f);
		const Asura_Euler_3	xRightRotation(0.0f, Asura_Maths::Deg90, 0.0f);
		const Asura_Euler_3	xUpRotation(Asura_Maths::Deg90, 0.0f, 0.0f);
		const Asura_Euler_3	xDownRotation(Asura_Maths::Deg270, 0.0f, 0.0f);

		pxVertex	= CreateVerticesForFace(xDownRotation, pxVertex);
		pxVertex	= CreateVerticesForFace(xFrontRotation, pxVertex);
		pxVertex	= CreateVerticesForFace(xBackRotation, pxVertex);
		pxVertex	= CreateVerticesForFace(xLeftRotation, pxVertex);
		pxVertex	= CreateVerticesForFace(xRightRotation, pxVertex);
		pxVertex	= CreateVerticesForFace(xUpRotation, pxVertex);
	}
	g_xSkyboxVB.Unlock();

#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
	// Create the index-buffer...
	g_xSkyboxIB.CreateStatic(sizeof(u_short) * g_uNumIndicesPerFace);

	u_short*	pusIndices	= static_cast<u_short*>(g_xSkyboxIB.Lock());
	Asura_Assert( pxVertex, "Unable to lock skybox index-buffer." );
	{
		CreateIndices(pusIndices);
	}
	g_xSkyboxIB.Unlock();
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)

	// Create the effect...
	g_pxSkyboxEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auSkybox);

	g_xSkyboxFactorParameter		= g_pxSkyboxEffect->GetParameterHandle( "SkyboxFactor" );
	g_xFogParameter					= g_pxSkyboxEffect->GetParameterHandle( "Fog" );
	g_xDistanceParameter			= g_pxSkyboxEffect->GetParameterHandle( "g_xDistance" );
	g_xTextureHandle				= g_pxSkyboxEffect->GetParameterHandle( "BaseTexture" );
	g_xWorldViewProjHandle			= g_pxSkyboxEffect->GetParameterHandle( "g_xWorldViewProjMatrix" );
	g_xFrustumToCamDir				= g_pxSkyboxEffect->GetParameterHandle( "g_xFrustumToCamDir" );
	g_xSkyboxColourParameter		= g_pxSkyboxEffect->GetParameterHandle( "g_xSkyboxColour" );
	g_xSunDir1Parameter				= g_pxSkyboxEffect->GetParameterHandle( "g_xSunDir1" );
	g_xSunDir2Parameter				= g_pxSkyboxEffect->GetParameterHandle( "g_xSunDir2" );
	g_xHazeGradientPower1Parameter	= g_pxSkyboxEffect->GetParameterHandle( "g_xHazeGradientPower1" );
	g_xHazeGradientPower2Parameter	= g_pxSkyboxEffect->GetParameterHandle( "g_xHazeGradientPower2" );
	g_xHazeColour1Parameter			= g_pxSkyboxEffect->GetParameterHandle( "g_xHazeColour1" );
	g_xHazeColour2Parameter			= g_pxSkyboxEffect->GetParameterHandle( "g_xHazeColour2" );
	g_xSunGradientPower1Parameter	= g_pxSkyboxEffect->GetParameterHandle( "g_xSunGradientPower1" );
	g_xSunGradientPower2Parameter	= g_pxSkyboxEffect->GetParameterHandle( "g_xSunGradientPower2" );
	g_xSunColour1Parameter			= g_pxSkyboxEffect->GetParameterHandle( "g_xSunColour1" );
	g_xSunColour2Parameter			= g_pxSkyboxEffect->GetParameterHandle( "g_xSunColour2" );
	g_xDisableHazeParameter			= g_pxSkyboxEffect->GetParameterHandle( "g_bDisableHaze" );
	g_xDisable2HazeParameter		= g_pxSkyboxEffect->GetParameterHandle( "g_bDisable2Haze" );
	g_xDisableSunParameter			= g_pxSkyboxEffect->GetParameterHandle( "g_bDisableSun" );
	g_xDisable2SunParameter			= g_pxSkyboxEffect->GetParameterHandle( "g_bDisable2Sun" );
	g_xMirrorBelowHorizonParameter	= g_pxSkyboxEffect->GetParameterHandle( "g_bMirrorBelowHorizon" );
	g_x2MirrorBelowHorizonParameter	= g_pxSkyboxEffect->GetParameterHandle( "g_b2MirrorBelowHorizon" );

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxSkyboxVD );
	Asura_Assert( g_pxSkyboxVD, "Failed to create skybox vertex declaration." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Skybox::Platform_Render()
//
// Renders the current skybox
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_Render()
{
	Direct3D_Device*	pxD3DDevice	= Asura_Xbox360_Graphics::GetDevice();

	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0, 0xFF, 0 ), "Skybox" );
	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_SKYBOX);
	Asura_Xbox360_Graphics::SetSkyboxDepthBias();

	Asura_Render::ApplyCameraSettings();

	const float fFogFactor = Asura_Fog::IsFogEnabled() ? Asura_Fog::GetSkyboxFogBlendFactor() : 0.0f;

	// Oddly, skybox colours are floats yet range from 0 to 255
	const float fSkyboxFactor = 1.0f - fFogFactor;
	const FXLVECTOR xSkyboxColour = { fSkyboxFactor * GetR() / 255.0f, fSkyboxFactor * GetG() / 255.0f, fSkyboxFactor * GetB() / 255.0f, 1.0f };
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xSkyboxFactorParameter, xSkyboxColour );

	const Asura_RGBA& xAsuraFogColour = Asura_Fog::GetFogColour();
	const FXLVECTOR xFogColour = { xAsuraFogColour.m_fR * fFogFactor, xAsuraFogColour.m_fG * fFogFactor, xAsuraFogColour.m_fB * fFogFactor, 1.0f };
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xFogParameter, xFogColour );

	// Set the skybox to output a depth beyond the far clip plane, so it doesn't get fogged
	g_pxSkyboxEffect->GetInterface()->SetFloat( g_xDistanceParameter, Asura_Camera::GetFarClip() + 500.0f );

	D3DXMATRIX xViewProj = Asura_Xbox360_Graphics::GetViewProjectionMatrix();
	/*
	const Asura_Vector_3* pxCameraPos = Asura_Camera::GetPosition();

	Asura_Angle fAngle;
	Asura_Skybox::GetOrientationAroundYAxis( fAngle );
	D3DXMATRIX xTransform;
	D3DXMatrixRotationY( &xTransform, fAngle );

	D3DXMATRIX xTranslation;
	D3DXMatrixTranslation( &xTranslation, pxCameraPos->x, pxCameraPos->y, pxCameraPos->z );

	xTransform *= xTranslation;
	xViewProj = xTransform * xViewProj;
	*/
	// Screen-space vertex shader (should make it same as PS3, but there's a glitch in my math somewhere so I'll use
	// the simple approach for now)
	D3DXMATRIX xView = Asura_Xbox360_Graphics::GetViewMatrix();
	const D3DXMATRIX & xProj = Asura_Xbox360_Graphics::GetProjectionMatrix(); 

	// Remove camera translation so that when we unproject we simply get camera->vertex direction 
	xView._41 = 0.0f;
	xView._42 = 0.0f;
	xView._43 = 0.0f;

	D3DXMATRIX xTempMat = xView * xProj;

	float fDet;
	D3DXMATRIX xFrustumToCamDir;
	D3DXMatrixIdentity( &xFrustumToCamDir );
	D3DXMatrixInverse( &xFrustumToCamDir, &fDet, &xTempMat );


	FXLVECTOR	xSkyboxColour2;
	FXLVECTOR	xSunDir1;
	FXLVECTOR	xSunDir2;
	FXLVECTOR	xHazeColour1;
	FXLVECTOR	xHazeColour2;
	FXLVECTOR	xSunColour1;
	FXLVECTOR	xSunColour2;
	float		fHazeStrength1;
	float		fHazeStrength2;
	float		fSunStrength1;
	float		fSunStrength2;
	bool		bTemp;

	Asura_Skybox::GetSkyboxColour( (float*)&xSkyboxColour2 );
	Asura_Skybox::GetHaze( (float*)&xHazeColour1, (float*)&xHazeColour2, bTemp, bTemp, bTemp, bTemp );
	Asura_Skybox::GetSun( (float*)&xSunColour1, (float*)&xSunColour2, bTemp, bTemp );
	Asura_Skybox::GetHazeStrength( &fHazeStrength1, &fHazeStrength2 );
	Asura_Skybox::GetSunSize( &fSunStrength1, &fSunStrength2 );
	Asura_Skybox::GetSunDir( (float*)&xSunDir1, (float*)&xSunDir2 );

	g_pxSkyboxEffect->GetInterface()->SetMatrixF( g_xWorldViewProjHandle, xViewProj );
	g_pxSkyboxEffect->GetInterface()->SetMatrixF( g_xFrustumToCamDir, xFrustumToCamDir );

	g_pxSkyboxEffect->GetInterface()->SetFloat( g_xHazeGradientPower1Parameter, fHazeStrength1 );
	g_pxSkyboxEffect->GetInterface()->SetFloat( g_xHazeGradientPower2Parameter, fHazeStrength2 );
	g_pxSkyboxEffect->GetInterface()->SetFloat( g_xSunGradientPower1Parameter, fSunStrength1 );
	g_pxSkyboxEffect->GetInterface()->SetFloat( g_xSunGradientPower2Parameter, fSunStrength2 );

	g_pxSkyboxEffect->GetInterface()->SetVector( g_xSkyboxColourParameter, xSkyboxColour2 );
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xSunDir1Parameter, xSunDir1 );
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xSunDir2Parameter, xSunDir2 );
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xHazeColour1Parameter, xHazeColour1 );
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xHazeColour2Parameter, xHazeColour2 );
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xSunColour1Parameter, xSunColour1 );
	g_pxSkyboxEffect->GetInterface()->SetVector( g_xSunColour2Parameter, xSunColour2 );

	g_pxSkyboxEffect->GetInterface()->SetBool( g_xDisableHazeParameter, s_bDisableHaze );
	g_pxSkyboxEffect->GetInterface()->SetBool( g_xDisable2HazeParameter, s_bDisable2Haze );
	g_pxSkyboxEffect->GetInterface()->SetBool( g_xDisableSunParameter, s_bDisableSun );
	g_pxSkyboxEffect->GetInterface()->SetBool( g_xDisable2SunParameter, s_bDisable2Sun );
	g_pxSkyboxEffect->GetInterface()->SetBool( g_xMirrorBelowHorizonParameter, s_b1MirrorBelowHorizon );
	g_pxSkyboxEffect->GetInterface()->SetBool( g_x2MirrorBelowHorizonParameter, s_b2MirrorBelowHorizon );

	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );

	// We'll write to all RTs. Output distance as a special 'far away' value, so fog doesn't affect the sky
	pxD3DDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE,  D3DCOLORWRITEENABLE_ALL );
	pxD3DDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );
	pxD3DDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL );
	pxD3DDevice->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL );

	g_xSkyboxVB.Bind();
	pxD3DDevice->SetVertexDeclaration( g_pxSkyboxVD );
#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
	g_xSkyboxIB.Bind();
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)

	if( GetUseNewSkyboxShader() )
	{
		g_pxSkyboxEffect->Begin( 1 );

		Asura_Xbox360_Graphics::GetFullScreenVB().Bind();
		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( Asura_Xbox360_Graphics::GetFullScreenVD() );

		g_pxSkyboxEffect->Commit();
		Asura_Xbox360_Graphics::GetDevice()->DrawVertices( D3DPT_RECTLIST, 0, 3 );
	}
	else
	{
		g_pxSkyboxEffect->Begin();

		// Doesn't match independent interface because that's broken
		for( u_int u = 0; u < NUM_SKYBOX_FACES; ++u )
		{
			PIXBeginNamedEvent( D3DCOLOR_XRGB( 0, 0xC0, 0 ), g_aszFaceNames[u] );

			g_pxSkyboxEffect->SetTexture( g_xTextureHandle, g_aiSkyboxTexID[u] );
			g_pxSkyboxEffect->Commit();

#if defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
			pxD3DDevice->DrawIndexedVertices(D3DPT_QUADLIST, u * g_uNumVerticesPerFace, 0, g_uNumIndicesPerFace);
#else
			pxD3DDevice->DrawVertices( D3DPT_QUADLIST, u * g_uNumVerticesPerFace, g_uNumVerticesPerFace );
#endif	// defined(ASURA_USE_XBOX360_SKYBOX_INDEX_BUFFER)
			Asura_Render::AddToTrianglesRendered(g_uNumQuadsPerFace * 2);

			PIXEndNamedEvent(); // Face
		}
	}



	g_pxSkyboxEffect->End();

	Asura_Xbox360_Graphics::ResetDepthBias();
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_SKYBOX);
	PIXEndNamedEvent(); // Skybox
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Skybox::Platform_Reset()
//
// Called when level is unloaded etc. Resets 360 skybox structures
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_Reset()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Skybox::Platform_ShouldModifyFogForSkyboxObjects()
//
// Queries whether fog should be adjusted for objects in the sky
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Skybox::Platform_ShouldModifyFogForSkyboxObjects()
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Skybox::Platform_Shutdown()
//
// Shuts down 360 skybox structures
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Skybox::Platform_Shutdown()
{
	Asura_Xbox360_Effect_System::Destroy( g_pxSkyboxEffect );
	g_xSkyboxVB.Destroy();

	g_pxSkyboxVD->Release();
	g_pxSkyboxVD = NULL;
}
