// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 PS2 UNIX PSP-INFRA PS3
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdCore/bdCrypto/bdDHKey.h>

#include <bdCore/bdUtilities/bdTrulyRandom.h>

#define BD_LOG_LEVEL "dhkey"

bdDHKey::bdDHKey()
: m_madeKey(false)
{
	ltc_mp = ltm_desc;
}


#if defined BD_DH_USE_ECC

bdDHKey::~bdDHKey()
{
	if (m_madeKey)
	{
		ecc_free(&m_key);
		m_madeKey = false;
	}
}

bdBool bdDHKey::create()
{
	bdInt error;
	prng_state prng;

	if (register_prng(&yarrow_desc) == -1)
	{
		return false;
	}

	const bdUInt size = 128;
	bdUByte8 random[size];
	bdTrulyRandom::getInstance()->getRandomUByte8(random, size);

	if ((error = yarrow_start(&prng)) != CRYPT_OK)
	{
		BD_ERR(BD_LOG_LEVEL, "Start error %s.", error_to_string(error));
		return false;
	}

	if ((error = yarrow_add_entropy(random, size, &prng)) != CRYPT_OK)
	{
		BD_ERR(BD_LOG_LEVEL, "Add_entropy error %s.", error_to_string(error));
		return false;
	}

	if ((error = yarrow_ready(&prng)) != CRYPT_OK)
	{
		BD_ERR(BD_LOG_LEVEL, "Ready error %s.", error_to_string(error));
		return false;
	}

	if ((error = ecc_make_key(&prng, find_prng("yarrow"), BD_DH_KEY_SIZE, &m_key)) != CRYPT_OK)
	{
		BD_ASSERT(g_assertFalse, "bdDHKey::create, unable to create private key.");
		return false;
	}

	m_madeKey = true;

	return true;
}

bdBool bdDHKey::importKey(const bdUByte8 *const pubKey, bdUInt keylen)
{
	const bdInt error = ecc_import(pubKey, keylen, &m_key);

	if (error != CRYPT_OK)
	{
		BD_WARN(BD_LOG_LEVEL, "Unable to import public key.");
		return false;
	}

	m_madeKey = true;

	return true;
}

bdBool bdDHKey::generateSecret(bdDHKey &other, bdUByte8 secret[BD_DH_SHARED_SECRET_SIZE])
{
	unsigned long tmpSize = BD_DH_SHARED_SECRET_SIZE;

	const bdInt error = ecc_shared_secret(&m_key, &(other.m_key), secret, &tmpSize);
	if (error != CRYPT_OK)
	{
		BD_WARN(BD_LOG_LEVEL, "Unable to generate secret key. Error: %s", error_to_string(error));
		return false;
	}

	// if the shared secret is less than the buffer size just fill the rest with
	// zeros.
	if(tmpSize < BD_DH_SHARED_SECRET_SIZE)
	{
		bdMemset(secret + tmpSize, 0, BD_DH_SHARED_SECRET_SIZE - tmpSize);
	}

	BD_ASSERT(tmpSize <= BD_DH_SHARED_SECRET_SIZE, 
		"bdDHKey::generateSecret, shared secret wrong size.");

	return true;
}

bdBool bdDHKey::exportKey(bdUByte8 *pubKey, bdUInt &keylen)
{
	unsigned long tmpKeylen = keylen;
	bdInt error = ecc_export(pubKey, &tmpKeylen, PK_PUBLIC, &m_key);
	if (error != CRYPT_OK)
	{
		BD_ASSERT(g_assertFalse, "bdDHKey::export, unable to export public key.");
		return false;
	}
	keylen = tmpKeylen;

	if (keylen < BD_DH_EXPORTED_KEY_SIZE)
	{
		bdMemset(pubKey + keylen, 0, BD_DH_EXPORTED_KEY_SIZE - keylen);
	}

	BD_ASSERT(keylen <= BD_DH_EXPORTED_KEY_SIZE, "bdDHKey::exportKey, wrong length!");

	keylen = BD_DH_EXPORTED_KEY_SIZE;

	return true;

}

#else

bdDHKey::~bdDHKey()
{
	if(m_madeKey)
	{
		dh_free(&m_key);
		m_madeKey = false;
	}
}

bdBool bdDHKey::create()
{
	bdInt error;
	prng_state prng;
	
	if (register_prng(&yarrow_desc) == -1) 
	{
		return false;
	}
	
	const bdUInt size = 128;
	bdUByte8 random[size];
	bdTrulyRandom::getInstance()->getRandomUByte8(random, size);
	
	if ((error = yarrow_start(&prng)) != CRYPT_OK) 
	{
		BD_ERR(BD_LOG_LEVEL, "Start error %s." , error_to_string(error));
		return false;
	}
	
	if ((error = yarrow_add_entropy(random, size, &prng)) != CRYPT_OK) 
	{
		BD_ERR(BD_LOG_LEVEL, "Add_entropy error %s." , error_to_string(error));
		return false;
	}
	
	if ((error = yarrow_ready(&prng)) != CRYPT_OK) 
	{
		BD_ERR(BD_LOG_LEVEL, "Ready error %s." , error_to_string(error));
		return false;
	}
	if ((error = dh_make_key(&prng, find_prng("yarrow"), 128, &m_key)) != CRYPT_OK)
	{
		bdBool fail = false;
		BD_ASSERT(fail, "bdDHKey::create, unable to create private key.");
		return false;
	}


	m_madeKey = true;
	return true;
}

bdBool bdDHKey::importKey(const bdUByte8 *const pubKey, bdUInt keylen)
{
	const bdInt error = dh_import(pubKey, keylen, &m_key);

	if (error != CRYPT_OK) 
	{
		BD_WARN(BD_LOG_LEVEL, "Unable to import public key.");
		return false;
	}

	m_madeKey = true;

	return true;
}

bdBool bdDHKey::generateSecret(bdDHKey &other, bdUByte8 secret[BD_DH_SHARED_SECRET_SIZE])
{
	unsigned long tmpSize = BD_DH_SHARED_SECRET_SIZE;

	const bdInt error = dh_shared_secret(&m_key, &(other.m_key), secret, &tmpSize);
	if (error != CRYPT_OK) 
	{
		BD_WARN(BD_LOG_LEVEL, "Unable to generate secret key. Error: %s", error_to_string(error));
		return false;
	}

	// if the shared secret is less than the buffer size just fill the rest with
	// zeros.
	if(tmpSize < BD_DH_SHARED_SECRET_SIZE)
	{
		bdMemset(secret + tmpSize, 0, BD_DH_SHARED_SECRET_SIZE - tmpSize);
	}

	BD_ASSERT(tmpSize <= BD_DH_SHARED_SECRET_SIZE, 
				"bdDHKey::generateSecret, shared secret wrong size.");

	return true;
}

bdBool bdDHKey::exportKey(bdUByte8 *pubKey, bdUInt &keylen)
{
	unsigned long tmpKeylen = keylen;
	bdInt error = dh_export(pubKey, &tmpKeylen, PK_PUBLIC, &m_key);
	if(error != CRYPT_OK)
	{
		bdBool fail = false;
		BD_ASSERT(fail, "bdDHKey::export, unable to export public key.");
		return false;
	}
	keylen = tmpKeylen;

	// if the exported key size is less than the buffer size just fill the rest with
	// zeros.
	if(keylen < BD_DH_EXPORTED_KEY_SIZE)
	{
		bdMemset(pubKey + keylen, 0, BD_DH_EXPORTED_KEY_SIZE - keylen);
	}

	BD_ASSERT(keylen <= BD_DH_EXPORTED_KEY_SIZE, "bdDHKey::exportKey, wrong length!");

	keylen = BD_DH_EXPORTED_KEY_SIZE;

	return true;
}
#endif // BD_DH_USE_ECC
