// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 PS2 UNIX OSX 
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: A wrapper around a socket to make it threaded.

#ifndef BD_THREADED_SOCKET_H
#define BD_THREADED_SOCKET_H

#include <bdCore/bdMemory/bdMemory.h>
#include <bdCore/bdContainers/bdByteBuffer.h>
#include <bdCore/bdContainers/bdQueue.h>
#include <bdCore/bdSocket/bdSocket.h>
#include <bdCore/bdThread/bdRunnable.h>
#include <bdCore/bdThread/bdMutex.h>
#include <bdCore/bdThread/bdSignal.h>

BD_REFERENCE(bdByteBuffer);

class bdThread;
class bdSignal;
class bdSendThread;
class bdReceiveThread;

class bdThreadedSocket : public bdSocket
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		bdThreadedSocket();

		virtual ~bdThreadedSocket();

		void setIncomingBufferSize(const bdUInt size);

		//
		virtual bdBool create(const bdBool broadcast = true);

		virtual bdSocketStatusCode bind(const bdPort port);

		virtual bdSocketStatusCode bind(const bdAddr &addr);
		//
		virtual bdInt sendTo(const bdAddr &addr,
							 const void *data,
							 const bdUInt length);

		//
		virtual bdInt receiveFrom(bdAddr &addr,
								  void *data,
								  const bdUInt size);

		// See base class comment.
		virtual bdBool close();

	protected:

		/// Separate receiving thread
		bdThread *m_receiveThread;
		bdReceiveThread *m_receiver;

		/// Separate sending thread
		bdThread *m_sendThread;
		bdSendThread *m_sender;

		bdUInt m_incomingBufferSize;


};

class bdSendThread : public bdRunnable
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		bdSendThread(const bdInt handle);

		virtual ~bdSendThread();

		/// Version of sendTo() that makes use of a separate thread
		/// \param addr[in] The destination address
		/// \param data[in] A pointer to the data to be sent
		/// \param length[in] Number of bytes to be sent
		virtual bdInt sendTo(const bdAddr &addr,
							 const void *data,
							 const bdUInt length);

	protected:

		/// Inherited from bdRunnable
		virtual bdUInt run(void *args);

		/// Called to indicate the thread to stop
		virtual void stop();

		void waitForBind();

	protected:

		class bdSendData
		{
			public:

				BD_DECLARE_NEW_AND_DELETE_OPERATORS

				inline bdSendData(const bdAddr& addr,
								  bdByteBufferRef data)
				: m_addr(addr),
				  m_data(data)
				{
				}

				bdAddr m_addr;
				bdByteBufferRef m_data;
		};

	protected:

		/// A signal for synchronizing access to the buffer.
		bdSignal m_signal;

		bdMutex m_lock;

		bdQueue<bdSendData> m_toSend;

		bdInt m_handle;

		// Flag to indicate whether the socket has been bound to a port
		// Needed to safely start the receive thread
		friend class bdThreadedSocket;
		bdBool m_isBound;
		bdSignal m_bindSignal;

};


class bdReceiveThread : public bdRunnable
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		bdReceiveThread(const bdInt handle,
						const bdUInt maxQueueSize);

		virtual ~bdReceiveThread();

		/// Non-blocking version of recvfrom() that makes use of a separate thread
		/// \param buf[in] A pointer to a buffer in which the incoming data is stored
		/// \param len[in] The maximum length of the receiving buffer
		/// \param flags[in]
		/// \param from[in] Pointer to an address structure to store the source address
		/// \param fromlen[in] Pointer to a variable that holds the size fo the source address
		/// \param addr[in]
		/// \param data[in]
		/// \param size[in]
		virtual bdInt receiveFrom(bdAddr &addr,
								  void *data,
								  const bdUInt size);

	protected:

		/// Inherited from bdRunnable
		virtual bdUInt run(void *args);

	protected:

		class bdReceiveData
		{
			public:

				BD_DECLARE_NEW_AND_DELETE_OPERATORS

				inline bdReceiveData(const bdInt status,
									 const bdAddr& addr,
									 bdByteBufferRef data)
				: m_status(status),
				  m_addr(addr),
				  m_data(data)
				{
				}

				bdInt m_status;
				bdAddr m_addr;
				bdByteBufferRef m_data;
		};

	protected:

		bdMutex m_lock;

		bdQueue<bdReceiveData> m_received;

		bdUInt m_recvQueueSize;

		bdUInt m_maxRecvQueueSize;

		// A handle to the socket
		bdInt m_handle;
};

#endif // BD_THREADED_SOCKET_H
