// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: Provides functions for writing to a bdStreamBuf.

#ifndef BD_O_STREAM_H
#define BD_O_STREAM_H

#include <bdCore/bdMemory/bdMemory.h>
#include <bdCore/bdStreams/bdStreambuf.h>
#include <bdCore/bdString/bdString.h>

class bdOStream;

/// This namespace contains the output stream manipulator functions.
/// These functions are envoked on a bdOStream using
/// bdOStream::operator<< (bdOStreamManipulator::function f) or
/// simply using the defined macros. Eg os << bdEndl or os << bdFlush.
namespace bdOStreamManipulator
{
	/// Defines the type of a bdOStream manipulator function.
	typedef void (*function)(bdOStream&);

	/// Inserts a new line character into the stream and flushes it.
	/// \param os[in] The stream to add a new line character to.
	void endl(bdOStream& os);

	/// Flushes the stream.
	/// \param os[in] The stream to be flushed.
	void flush(bdOStream& os);
}

/// Provides member functions to assist in writing data to a bdStreambuf.
/// The operator<< functions provide mechanisms to write data formatted in
/// a human readable way to the bdStreambuf.  While most of the other
/// functions are used to perform raw unformatted output.
class bdOStream
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		/// Constructor. Initializes an input stream with a given
		/// bdStreambuf for writing.
		/// \param streambuf[in] The stream buffer to use when
		///  constructing this object
		bdOStream(bdStreambuf* streambuf);

		/// Check the status of this output stream and its bdStreambuf.
		/// \return
		///  - True if no errors have occured.
		///  - False otherwise.
		bdBool isOk() const;

		/// Writes a signed character to the bdStreambuf.
		/// \param c[in] The character to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdNChar8 &c);

		///// Writes an unsigned character to the bdStreambuf.
		///// \param c[in] The character to write to the stream.
		///// \return A reference to \a this stream.
		//bdOStream& operator<< (const bdUChar8 &c);

		/// Writes a human readable 16 bit signed integer to the bdStreambuf.
		/// \param i[in] The integer to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdInt16 &i);

		/// Writes a human readable 16 bit unsigned integer to the bdStreambuf.
		/// \param u[in] The unsigned integer to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdUInt16 &u);

		/// Writes a human readable 32 bit signed integer to the bdStreambuf.
		/// \param i[in] The integer to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdInt &i);

		/// Writes a human readable 32 bit unsigned integer to the bdStreambuf.
		/// \param u[in] The unsigned integer to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdUInt &u);

		/// Writes a human readable 64 bit signed integer to the bdStreambuf.
		/// \param i[in] The integer to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdInt64 &i);

		/// Writes a human readable 64 bit unsigned integer to the bdStreambuf.
		/// \param u[in] The unsigned integer to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdUInt64 &u);

		/// Writes a human readable 32 bit floating point value to the
		/// bdStreambuf. The string written to the bdStreambuf will have
		/// the default precision and width as a similar call to
		/// bdSnprintf().
		/// \param f[in] The float to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdFloat32 &f);

		/// Writes a human readable 64 bit floating point value to the
		/// bdStreambuf. The string written to the bdStreambuf will have
		/// the default precision and width as a similar call to
		/// bdSnprintf.
		/// \param f[in] The float to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdFloat64 &f);

		/// Writes the string pointed to by \a s directly into the bdStreamBuf.
		/// The string pointed to by \a s must be null terminated. The null
		/// terminator is not written to the bdStreamBuf.
		/// \param s[in] A pointer to the string to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdNChar8 *const s);

		///// Identical to operator<< (const bdChar8 *const s).
		///// \param s[in] A pointer to the string to write to the stream.
		///// \return A reference to \a this stream.
		//bdOStream& operator<< (const bdUChar8 *const s);

		/// Writes a string to the bdStreamBuf.
		/// \param s[in] A reference to the string to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdString &s);

		/// Writes a human readable boolean to the bdStreambuf.
		/// If b is true '1' is written otherwise '0' is written.
		/// \param b[in] The boolean to write to the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (const bdBool &b);

		/// This function is intended for use with stream manipulators.
		/// E.g. bdEndl (bdOStreamManipulator::endl) and bdFlush
		/// (bdOStreamManipulator::flush).
		/// \param f[in]	A bdOStreamManipulator::function to call on
		///					the stream.
		/// \return A reference to \a this stream.
		bdOStream& operator<< (bdOStreamManipulator::function f);

		/// Flushes the bdStreambuf, writing out any buffered writes.
		/// This operation may not have the desired effect if a particular
		/// bdStreambuf implementation is buffering writes and does not
		/// implement bdStreambuf::bufFlush().
		void flush();

		/// Writes nbytes from the array pointed to by buf into the
		/// bdStreambuf. No attempt is made to convert the data into
		/// a human readable form.
		/// \param buf[in] A pointer to the buffer to read the data to be written.
		/// \param nbytes[in] The number of bytes to write.
		/// \return The number of bytes actually written.
		/// \note If this differs from \a nbytes an error has occured.
		bdUInt write(const bdNChar8 *const buf,
					   const bdUInt nbytes);

		/// Sets the next location to be written to in the bdStreambuf.
		/// \param offset[in] The number of bytes to move starting at \a from.
		/// \param from[in]		The initial position from which to move the write
		///						position by \a offset bytes.
		/// \return
		///  - True if the operation was successful.
		///  - False otherwise.
		/// \note This will fail if the underlying bsStreambuf does not
		///  implement bdStreambuf::bufSeek().
		bdBool seek(const bdUInt offset,
					const bdStreambuf::SeekFrom from);

		/// Gets the position of the next write location relative to the
		/// start of the bdStreambuf.
		/// \param offset[out] The offset position of the next write.
		/// \return
		///  - True to indicate success.
		///  - False otherwise.
		/// \note This function may fail if the underlying bdStreambuf does
		///  not implement bdStreambuf::bufTell(). In this case the function
		///  will return false and offset will remain unaltered.
		bdBool tell(bdUInt &offset) const;

		/// An accessor method for the underlying buffer.
		/// \return a pointer to the underlying bdStreambuf.
		bdStreambuf* getStreambuf();

	protected:

		/// The underlying bdStreambuf.
		bdStreambuf *m_streambuf;
};

/// Convenience macro for invoking the bdOStreamManipulator::endl
/// manipulator on a bdOStream.
#define bdEndl bdOStreamManipulator::endl

/// Convenience macro for invoking the bdOStreamManipulator::flush
/// manipulator on a bdOStream.
#define bdFlush bdOStreamManipulator::flush

#endif // BD_O_STREAM_H
