// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: A platform independent timing class interface.

#ifndef BD_STOPWATCH_H
#define BD_STOPWATCH_H

#include <bdCore/bdMemory/bdMemory.h>

/// A platform independent timing class interface.
/// A class that times the duration of operations and events.
/// The main purpose is to provide an operating system independent interface
/// to the programmer.
class bdStopwatch
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		/// Default constructor
		bdStopwatch();

		/// Copy constructor
		bdStopwatch(const bdStopwatch &other);

		/// Start the stopwatch running.
		/// The current timings are not cleared before starting the timer.
		/// To clear the current timings call reset().
		void start();

		/// Stop the stopwatch.
		/// This records a timestamp. The elapsed time can be calculated using
		/// getTimeInSeconds(). stop() can be called multiple times for one
		/// call to start().
		void stop();

		/// Clear the current timings.
		/// The time values stored in the stopwatch are reset to zero.
		void reset();

		/// Return the elapsed time between the call to start() and the last
		/// call to stop().
		/// \return		A float in the form of 'a.b' where 'a' is the second
		///				count and 'b' is the fraction of the second.
		bdFloat32 getTimeInSeconds() const;

		/// Get the elapsed time between the call to start() and now.
		/// This is equivalent to calling stop() followed by
		/// getTimeInSeconds(), but is a better alternative since stop()
		/// is non-const.
		/// \return		A float in the form of 'a.b' where 'a' is the second
		///				count and 'b' is the fraction of the second.
		bdFloat32 getElapsedTimeInSeconds() const;

	protected:

		/// The last start time.
		bdUInt64 m_start;

		/// The last stop time.
		bdUInt64 m_stop;
};

#endif // BD_STOPWATCH_H
