// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include "Simulation.h"

#define TITLE_ID (0x12F7)
#define USERNAME "ChatExample"
#define PASSWORD "MyPassword"

Simulation::Simulation()
: m_taskStatus(READY),
  m_task(START_NETWORK),
  m_changedScreen(true),
  m_authGetHostByName(BD_NULL),
  m_authService(BD_NULL),
  m_lsgGetHostByName(BD_NULL),
  m_remoteTask(BD_NULL),
  m_subscribingChannelID(0),
  m_pumpLobbyConnection(false),
  m_lobbyErrorCode(BD_NO_ERROR)
{
}

Simulation::~Simulation()
{
}

bdBool Simulation::init(bdInt32 ,
						bdNChar8* [])
{
	bdBool ok = true;

	// start logging
	// we want to hear about all warnings and errors
	bdLogImpl *log = bdLog::getInstance();
	log->subscribe("warn", &m_logSubscriber);
	log->subscribe("err", &m_logSubscriber);
	log->subscribe("info", &m_logSubscriber);

	return ok;
}

bdBool Simulation::step(const bdFloat32)
{
	const bdBool quitKey = getScreen().wasKeyPressed(BD_QUIT);

	TaskStatus oldTaskStatus = m_taskStatus;

	if(m_taskStatus == READY)
	{
		startTask();
	}

	if(m_taskStatus == PENDING || m_taskStatus == DONE || m_taskStatus == FAILED)
	{
		// pump even if the status is DONE or FAILED to read out any results
		// and close any remote task handles
		pumpTask();
	}

	if(m_taskStatus == DONE || m_taskStatus == FAILED || m_taskStatus == SKIPPED)
	{
		// incase the task fails or completes in a single loop
		if(oldTaskStatus == READY)
		{
			oldTaskStatus = m_taskStatus;
		}

		displayTaskResult();
	}

	m_changedScreen = (oldTaskStatus != m_taskStatus);

	if(m_pumpLobbyConnection)
	{
		bdBool lostConnection = true;
		bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
		if(lobbyService)
		{
			const bdLobbyConnection::Status status = lobbyService->getStatus();
			switch(status)
			{
				case bdConnection::BD_NOT_CONNECTED:
				case bdConnection::BD_CONNECTING:
				case bdConnection::BD_DISCONNECTING:
				case bdConnection::BD_DISCONNECTED:
				{
					lostConnection = true;
				}
				break;
				case bdConnection::BD_CONNECTED:
				{
					lostConnection = false;
				}
				break;
			}
		}

		if(lostConnection)
		{
			// set up for a reconnect
			m_pumpLobbyConnection = false;
			m_task = STOP_NETWORK;
			m_taskStatus = READY;
			m_changedScreen = true;
			CloseRemoteTask(&m_remoteTask);
			m_lastTaskDescription = "Lost connection to lobby.";
		}
	}

	checkMessageStore();

	m_changedScreen = (oldTaskStatus != m_taskStatus);
	getScreen().printToScreen(m_screenText + m_messageText);

	if(m_changedScreen)
	{
		m_screenText = "";
	}

	return !quitKey;
}

void Simulation::quit()
{
	bdLogImpl *log = bdLog::getInstance();
	log->unsubscribe("warn", &m_logSubscriber);
	log->unsubscribe("err", &m_logSubscriber);
	log->unsubscribe("info", &m_logSubscriber);
	
	stopNetwork();

	delete m_authGetHostByName;
	delete m_authService;
	delete m_lsgGetHostByName;
	
	if(m_remoteTask)
	{
		m_remoteTask->close();
	}
}

void Simulation::startTask()
{	
	const bdBool skipKey = getScreen().wasKeyPressed(SKIP_KEY);

	if (skipKey)
	{
		m_taskStatus = SKIPPED;
	}
	else
	{
		const bdBool startKey = getScreen().wasKeyPressed(START_KEY);
		bdString taskDescription("Unknown task");

		switch(m_task)
		{
			case START_NETWORK:
			{
				startNetworkDescription(taskDescription);
				if(startKey)
				{
					startNetwork();
				}
			}
			break;
			case CHECK_DNS:
			{
				checkDNSDescription(taskDescription);
				if(startKey)
				{
					startCheckDNS();
				}
			}
			break;
			case CREATE_ACCOUNT:
			{
				createAccountDescription(taskDescription);
				if(startKey)
				{
					startCreateAccount();
				}
			}
			break;
			case AUTHENTICATE_ACCOUNT:
			{
				authenticateAccountDescription(taskDescription);
				if(startKey)
				{
					startAuthenticateAccount();
				}
			}
			break;
			case LOBBY_CONNECT:
			{
				lobbyServiceConnectDescription(taskDescription);
				if(startKey)
				{
					startLobbyServiceConnect();
				}
			}
			break;
			case GET_CHANNEL_LIST:
			{
				getChannelListDescription(taskDescription);
				if(startKey)
				{
					startGetChannelList();
				}
			}
			break;
			case SUBSCRIBE_TO_CHANNEL:
			{
				subscribeToChannelDescription(taskDescription);
				if(startKey)
				{
					startSubscribeToChannel();
				}
			}
			break;
			case SEND_TO_CHANNEL:
			{
				sendToChannelDescription(taskDescription);
				if(startKey)
				{
					startSendToChannel();
				}
			}
			break;
			case GET_CHANNEL_MEMBERS:
			{
				getChannelMembersDescription(taskDescription);
				if(startKey)
				{
					startGetChannelMembers();
				}
			}
			break;
			case WHISPER_TO_MEMBER:
			{
				whisperToMemberDescription(taskDescription);
				if(startKey)
				{
					startWhisperToMember();
				}
			}
			break;
			case UNSUBSCRIBE_FROM_CHANNEL:
			{
				unsubscribeFromChannelDescription(taskDescription);
				if(startKey)
				{
					startUnsubscribeFromChannel();
				}
			}
			break;
			case LOBBY_DISCONNECT:
			{
				lobbyServiceDisconnectDescription(taskDescription);
				if(startKey)
				{
					startLobbyServiceDisconnect();
				}
			}
			break;
			case STOP_NETWORK:
			{
				stopNetworkDescription(taskDescription);
				if(startKey)
				{
					stopNetwork();
				}
			}
			break;
		}
		
		if(m_changedScreen)
		{
			m_screenText += taskDescription;
			m_screenText += "\n" START_TEXT;
			m_screenText += "\n" SKIP_TEXT;

			if(m_lastTaskDescription.getLength())
			{
				m_screenText += "\n\n";
				m_screenText += m_lastTaskDescription;				
			}
		}
	}
}

void Simulation::pumpTask()
{
	bdString taskDescription("Unknown task");

	switch(m_task)
	{
		case START_NETWORK:
		{
			startNetworkDescription(taskDescription);
			pumpStartNetwork();
		}
		break;
		case CHECK_DNS:
		{
			checkDNSDescription(taskDescription);
			pumpCheckDNS();
		}
		break;
		case CREATE_ACCOUNT:
		{
			createAccountDescription(taskDescription);
			pumpCreateAccount();
		}
		break;
		break;
		case AUTHENTICATE_ACCOUNT:
		{
			authenticateAccountDescription(taskDescription);
			pumpAuthenticateAccount();
		}
		break;
		case LOBBY_CONNECT:
		{
			lobbyServiceConnectDescription(taskDescription);
			pumpLobbyServiceConnect();
		}
		break;
		case GET_CHANNEL_LIST:
		{
			getChannelListDescription(taskDescription);
			pumpGetChannelList();
		}
		break;
		case SUBSCRIBE_TO_CHANNEL:
		{
			subscribeToChannelDescription(taskDescription);
			pumpSubscribeToChannel();
		}
		break;
		case SEND_TO_CHANNEL:
		{
			sendToChannelDescription(taskDescription);
			pumpSendToChannel();
		}
		break;
		case GET_CHANNEL_MEMBERS:
		{
			getChannelMembersDescription(taskDescription);
			pumpGetChannelMembers();
		}
		break;
		case WHISPER_TO_MEMBER:
		{
			whisperToMemberDescription(taskDescription);
			pumpWhisperToMember();
		}
		break;
		case UNSUBSCRIBE_FROM_CHANNEL:
		{
			unsubscribeFromChannelDescription(taskDescription);
			pumpUnsubscribeFromChannel();
		}
		break;
		case LOBBY_DISCONNECT:
		{
			lobbyServiceDisconnectDescription(taskDescription);
			pumpLobbyServiceDisconnect();
		}
		break;
		case STOP_NETWORK:
		{
			// nothing to to here
		}
		break;
	}

	if(m_changedScreen)
	{
		m_screenText += taskDescription;
	}
}

void Simulation::displayTaskResult()
{
	bdString taskDescription("Unknown task result");

	Task nextTaskIfDone = m_task;
	Task nextTaskIfFailed = m_task;
	Task nextTaskIfSkipped = m_task;

	bdBool displayLobbyError = true;

	switch(m_task)
	{
		case START_NETWORK:
		{
			startNetworkDescription(m_lastTaskDescription);
			nextTaskIfDone = CHECK_DNS;
			nextTaskIfFailed = START_NETWORK;
			nextTaskIfSkipped = START_NETWORK;
			displayLobbyError = false;
		}
		break;
		case CHECK_DNS:
		{
			checkDNSDescription(m_lastTaskDescription);
			nextTaskIfDone = CREATE_ACCOUNT;
			nextTaskIfFailed = CHECK_DNS;
			nextTaskIfSkipped = CREATE_ACCOUNT;
			displayLobbyError = false;
		}
		break;
		case CREATE_ACCOUNT:
		{
			createAccountDescription(m_lastTaskDescription);
			checkAuthServiceError();
			nextTaskIfDone = AUTHENTICATE_ACCOUNT;
			nextTaskIfFailed = CREATE_ACCOUNT;
			nextTaskIfSkipped = AUTHENTICATE_ACCOUNT;			
		}
		break;
		case AUTHENTICATE_ACCOUNT:
		{
			authenticateAccountDescription(m_lastTaskDescription);
			checkAuthServiceError();
			nextTaskIfDone = LOBBY_CONNECT;
			nextTaskIfFailed = AUTHENTICATE_ACCOUNT;
			nextTaskIfSkipped = LOBBY_CONNECT;			
		}
		break;
		case LOBBY_CONNECT:
		{
			lobbyServiceConnectDescription(m_lastTaskDescription);
			nextTaskIfDone = GET_CHANNEL_LIST;
			nextTaskIfFailed = LOBBY_CONNECT;
			nextTaskIfSkipped = GET_CHANNEL_LIST;
		}
		break;
		case GET_CHANNEL_LIST:
		{
			getChannelListDescription(m_lastTaskDescription);
			checkChannelChatError();
			nextTaskIfDone = SUBSCRIBE_TO_CHANNEL;
			nextTaskIfFailed = GET_CHANNEL_LIST;
			nextTaskIfSkipped = SUBSCRIBE_TO_CHANNEL;
		}
		break;
		case SUBSCRIBE_TO_CHANNEL:
		{
			subscribeToChannelDescription(m_lastTaskDescription);
			checkChannelChatError();
			nextTaskIfDone = SEND_TO_CHANNEL;
			nextTaskIfFailed = SUBSCRIBE_TO_CHANNEL;
			nextTaskIfSkipped = SEND_TO_CHANNEL;
		}
		break;
		case SEND_TO_CHANNEL:
		{
			sendToChannelDescription(m_lastTaskDescription);
			checkChannelChatError();
			nextTaskIfDone = GET_CHANNEL_MEMBERS;
			nextTaskIfFailed = SEND_TO_CHANNEL;
			nextTaskIfSkipped = GET_CHANNEL_MEMBERS;
		}
		break;
		case GET_CHANNEL_MEMBERS:
		{
			getChannelMembersDescription(m_lastTaskDescription);
			checkChannelChatError();
			nextTaskIfDone = WHISPER_TO_MEMBER;
			nextTaskIfFailed = GET_CHANNEL_MEMBERS;
			nextTaskIfSkipped = WHISPER_TO_MEMBER;
		}
		break;
		case WHISPER_TO_MEMBER:
		{
			whisperToMemberDescription(m_lastTaskDescription);
			checkChannelChatError();
			nextTaskIfDone = UNSUBSCRIBE_FROM_CHANNEL;
			nextTaskIfFailed = WHISPER_TO_MEMBER;
			nextTaskIfSkipped = UNSUBSCRIBE_FROM_CHANNEL;
		}
		break;
		case UNSUBSCRIBE_FROM_CHANNEL:
		{
			unsubscribeFromChannelDescription(m_lastTaskDescription);
			checkChannelChatError();
			nextTaskIfDone = LOBBY_DISCONNECT;
			nextTaskIfFailed = UNSUBSCRIBE_FROM_CHANNEL;
			nextTaskIfSkipped = LOBBY_DISCONNECT;
		}
		break;
		case LOBBY_DISCONNECT:
		{
			lobbyServiceDisconnectDescription(m_lastTaskDescription);			
			nextTaskIfDone = STOP_NETWORK;
			nextTaskIfFailed = LOBBY_DISCONNECT;
			nextTaskIfSkipped = STOP_NETWORK;
		}
		break;
		case STOP_NETWORK:
		{
			stopNetworkDescription(m_lastTaskDescription);
			nextTaskIfDone = START_NETWORK;
			nextTaskIfFailed = STOP_NETWORK;
			nextTaskIfSkipped = CREATE_ACCOUNT;
		}
		break;
	}

	switch(m_taskStatus)
	{
		case READY:
		case PENDING:
		{
			BD_WARN(BD_LOG_LEVEL, "Unexpected task status.");
		}
		break;
		case DONE:
		{
			m_task = nextTaskIfDone;
		}
		break;
		case FAILED:
		{
			m_task = nextTaskIfFailed;
		}
		break;
		case SKIPPED:
		{
			m_task = nextTaskIfSkipped;
		}
		break;
	}

	// if task uses sets a lobby error code and has failed
	// display the error description
	if(displayLobbyError && (m_taskStatus == FAILED))
	{
		bdNChar8 buf[32];
		LobbyErrorCodeToString(m_lobbyErrorCode, buf, sizeof(buf));
		m_lastTaskDescription += "\nError: ";
		m_lastTaskDescription += buf;
		m_lobbyErrorCode = BD_NO_ERROR;
	}

	m_taskStatus = READY;
}

void Simulation::startNetworkDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Start Network",				// READY
											"Starting Network",				// PENDING
											"Started Network",				// DONE
											"Starting Network Failed",		// FAILED
											"Starting Network Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startNetwork()
{
	// fail safe
	m_taskStatus = FAILED;

	bdNetImpl *const net = bdNet::getInstance();
	if(net)
	{
		bdNetStartParams params;

		// platform specific startup parameters
		bdPlatformParams platformParams;
		//TODO: check the return value here
		bdExamplePlatformInit::init(params, &platformParams);

		if(net->start(params))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpStartNetwork()
{
	// fail safe
	m_taskStatus = FAILED;

	bdNetImpl *const net = bdNet::getInstance();
	if(net)
	{
		net->pump();

		const bdNetImpl::bdNetStatus networkStatus = net->getStatus();
		switch(networkStatus)
		{
			case bdNetImpl::BD_NET_BIND_FAILED:
			case bdNetImpl::BD_NET_ONLINE_FAILED:
			case bdNetImpl::BD_NET_INIT_FAILED:
			case bdNetImpl::BD_NET_STOPPED:
			{
				// error conditions
				m_taskStatus = FAILED;
			}
			break;
			case bdNetImpl::BD_NET_PENDING:
			{
				// wait
				m_taskStatus = PENDING;
			}
			break;
			case bdNetImpl::BD_NET_DONE:
			{
				// complete
				m_taskStatus = DONE;
			}
		}
	}
}

void Simulation::checkDNSDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Start DNS Lookup",			// READY
											"Performing DNS Lookup",	// PENDING
											"DNS Lookup Succeeded",		// DONE
											"DNS Lookup Failed",		// FAILED
											"DNS Lookup Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == DONE)
	{
		bdNChar8 addrString[BD_INET_ADDR_STRING_SIZE];
		m_authAddr.toString(addrString, sizeof(addrString));
		description +=  "\n " AUTH_ADDRESS ":";
		description += addrString;
		description +=  "\n";

		m_lsgAddr.toString(addrString, sizeof(addrString));
		description +=  " " LSG_ADDRESS ":";
		description += addrString;
	}

}		

void Simulation::startCheckDNS()
{
	const TaskStatus authDNSStatus = StartGetHostByName(AUTH_ADDRESS, &m_authGetHostByName, m_authAddr);
	const TaskStatus lsgDNSStatus = StartGetHostByName(LSG_ADDRESS, &m_lsgGetHostByName, m_lsgAddr);
	if((authDNSStatus == FAILED) || (lsgDNSStatus == FAILED))
	{
		m_taskStatus = FAILED;
	}

	if((authDNSStatus == PENDING) || (lsgDNSStatus == PENDING))
	{
		m_taskStatus = PENDING;
	}

	if((authDNSStatus == DONE) && (lsgDNSStatus == DONE))
	{
		m_taskStatus = DONE;
	}
}

void Simulation::pumpCheckDNS()
{
	TaskStatus authDNSStatus = DONE;
	TaskStatus lsgDNSStatus = DONE;

	if(m_authGetHostByName)
	{
		authDNSStatus = PumpGetHostByName(&m_authGetHostByName, m_authAddr);
	}

	if(m_lsgGetHostByName)
	{
		lsgDNSStatus = PumpGetHostByName(&m_lsgGetHostByName, m_lsgAddr);
	}

	if((authDNSStatus == FAILED) || (lsgDNSStatus == FAILED))
	{
		m_taskStatus = FAILED;
	}

	if((authDNSStatus == PENDING) || (lsgDNSStatus == PENDING))
	{
		m_taskStatus = PENDING;
	}

	if((authDNSStatus == DONE) && (lsgDNSStatus == DONE))
	{
		m_taskStatus = DONE;
	}
}

bdAuthService* Simulation::getAuthService()
{
	if(m_authService == BD_NULL)
	{
		const bdAddr authAddr(m_authAddr, AUTH_PORT);
		m_authService = new bdAuthService(TITLE_ID, authAddr);
	}

	return m_authService;
}

const bdAuthInfo *Simulation::getAuthInfo()
{
	if(m_authService != BD_NULL)
	{
		return m_authService->getAuthInfo();
	}

	return BD_NULL;
}

void Simulation::checkAuthServiceError()
{
	bdAuthService *const authService = getAuthService();
	if(authService)
	{
		m_lobbyErrorCode = authService->getErrorCode();
	}
}

void Simulation::pumpAuthServiceTask()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService)
	{
		m_taskStatus = PENDING;
		const bdAuthService::bdStatus status = authService->getStatus();
		if(status == bdAuthService::BD_READY)
		{			
			const bdLobbyErrorCode lobbyErrorCode = authService->getErrorCode();
			if(lobbyErrorCode == BD_AUTH_NO_ERROR)
			{
				m_taskStatus = DONE;
			}
			else
			{
				m_taskStatus = FAILED;
			}
		}
	}
}

void Simulation::createAccountDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Create Account (" USERNAME "/" PASSWORD ")",	// READY
											"Creating Account",								// PENDING
											"Created Account",								// DONE
											"Create Account Failed",						// FAILED
											"Create Account Skipped"};						// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startCreateAccount()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService)
	{
		if(authService->createAccount(USERNAME, PASSWORD))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpCreateAccount()
{
	pumpAuthServiceTask();
}

void Simulation::authenticateAccountDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Authenticate Account (" USERNAME "/" PASSWORD ")",	// READY
											"Authenticating Account",									// PENDING
											"Authenticated Account",									// DONE
											"Authenticate Account Failed",								// FAILED
											"Authenticate Account Skipped"};							// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startAuthenticateAccount()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService )
	{
		if(authService->authorizeAccount(USERNAME, PASSWORD))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpAuthenticateAccount()
{
	pumpAuthServiceTask();
}

void Simulation::lobbyServiceConnectDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Connect to Lobby Service",				// READY
											"Connecting to Lobby Service",			// PENDING
											"Connected to Lobby Service",			// DONE
											"Connect to Lobby Service Failed",		// FAILED
											"Connect to Lobby Service Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startLobbyServiceConnect()
{
	// fail safe
	m_taskStatus = FAILED;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		const bdAddr lsgAddr(m_lsgAddr, LSG_PORT);
		const bdBool connecting = lobbyService->connect(lsgAddr, getAuthInfo());
		if(connecting)
		{
			m_taskStatus = PENDING;
		}
		// no need to keep the bdAuthService instance after we are connected to the lobby
		delete m_authService;
		m_authService = BD_NULL;
	}
}

void Simulation::pumpLobbyServiceConnect()
{
	// fail safe
	m_taskStatus = FAILED;

	m_pumpLobbyConnection = false;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		const bdLobbyConnection::Status connectionStatus = lobbyService->getStatus();
		switch(connectionStatus)
		{			
			case bdConnection::BD_NOT_CONNECTED:
			case bdConnection::BD_DISCONNECTING:
			case bdConnection::BD_DISCONNECTED:
			{
				m_taskStatus = FAILED;
			}
			break;
			case bdConnection::BD_CONNECTING:
			{
				m_taskStatus = PENDING;
			}
			break;
			case bdConnection::BD_CONNECTED:
			{
				m_taskStatus = DONE;
				m_pumpLobbyConnection = true;
			}
			break;			
		}
	}
}

void Simulation::checkChannelChatError()
{
	const bdChannelChat *const channelchat = GetChannelChat();	
	if(channelchat)
	{
		m_lobbyErrorCode = channelchat->getErrorCode();
	}
}

void Simulation::getChannelListDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Get Channel List",				// READY
											"Getting Channel List",			// PENDING
											"Got Channel List",				// DONE
											"Get Channel List Failed",		// FAILED
											"Get Channel List Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == DONE)
	{
		description += "\nChannel list:";
		m_channelList.print(description);
	}
}

void Simulation::startGetChannelList()
{
	// fail safe
	m_taskStatus = FAILED;

	bdChannelChat *const channelChat = GetChannelChat();
	if(channelChat)
	{
		m_remoteTask = channelChat->getChannelList(0);
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	}
}

void Simulation::pumpGetChannelList()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);

	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdChannelChat *const channelChat = GetChannelChat();
		if(channelChat)
		{
			bdGetChannelsResult channelsResult;
			const bdBool taskHandled = channelChat->handleTask(m_remoteTask, &channelsResult);
			if(taskHandled)
			{
				m_taskStatus = DONE;
				m_channelList.clear();

				bdBool ok = true;
				const bdUInt numResults = channelsResult.getNumResults();			
				for(bdUInt i = 0; ok && (i < numResults); i++)
				{
					ok  = m_channelList.add(channelsResult.m_channelList[i]);
				}
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::subscribeToChannelDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Subscribe to Channel",				// READY
											"Subscribing to Channel",			// PENDING
											"Subscribed to Channel",			// DONE
											"Subscribe to Channel Failed",		// FAILED
											"Subscribe to Channel Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == FAILED && m_subscribedChannels.getSize() == 0)
	{
		description += "\nNo channels to subscribe to.";
	}
}

void Simulation::startSubscribeToChannel()
{
	// fail safe
	m_taskStatus = FAILED;

	bdChannelChat *const channelChat = GetChannelChat();
	if(channelChat && (m_channelList.getSize() != 0))
	{
		m_subscribingChannelID = m_channelList[0].m_channelID;
		m_remoteTask = channelChat->subscribeToChannel(0, m_subscribingChannelID);
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	}
}

void Simulation::pumpSubscribeToChannel()
{
	pumpGenericChannelChatTask();

	if(m_taskStatus == DONE)
	{
		m_subscribedChannels.pushBack(m_subscribingChannelID);		
	}

	if(m_taskStatus == DONE || m_taskStatus == FAILED)
	{
		m_subscribingChannelID = 0;
	}
}

void Simulation::sendToChannelDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Send to Channel",				// READY
											"Sending to Channel",			// PENDING
											"Sent to Channel",				// DONE
											"Send to Channel Failed",		// FAILED
											"Send to Channel Skipped"};		// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == FAILED && m_subscribedChannels.getSize() == 0)
	{
		description += "\nNot subscribed to any channels.";
	}
}

void Simulation::startSendToChannel()
{
	// fail safe
	m_taskStatus = FAILED;

	bdChannelChat *const channelChat = GetChannelChat();
	if(channelChat && (m_subscribedChannels.getSize() > 0))
	{		
		bdNChar8 str[128];
		bdStrlcpy(str, "Example message", sizeof(str));		
		const bdUInt strLength = bdStrnlen(str, sizeof(str));

		bdByteBuffer payload(128);
		const bdBool ok = payload.write(str, strLength+1);
		if(ok)
		{
			m_remoteTask = channelChat->sendToChannel(0, m_subscribedChannels[0], payload);
			m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
		}
	}
}

void Simulation::pumpSendToChannel()
{
	pumpGenericChannelChatTask();
}

void Simulation::getChannelMembersDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Get Channel Members",				// READY
											"Getting Channel Members",			// PENDING
											"Got Channel Members",				// DONE
											"Get Channel Members Failed",		// FAILED
											"Get Channel Members Skipped"};		// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == DONE)
	{
		description += "\nChannel Members:\n";
		m_memberList.print(description);
	}

	if((m_taskStatus == FAILED) && (m_channelList.getSize() == 0))
	{
		description += "\nNo Channels to get members of.";
	}
}

void Simulation::startGetChannelMembers()
{
	// fail safe
	m_taskStatus = FAILED;

	bdChannelChat *const channelChat = GetChannelChat();
	if(channelChat && (m_channelList.getSize() != 0))
	{
		m_remoteTask = channelChat->getChannelMembers(0, m_channelList[0].m_channelID);
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	}
}

void Simulation::pumpGetChannelMembers()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);

	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdChannelChat *const channelChat = GetChannelChat();
		if(channelChat)
		{
			bdGetChannelMembersResult membersResult;
			const bdBool taskHandled = channelChat->handleTask(m_remoteTask, &membersResult);
			if(taskHandled)
			{
				m_taskStatus = DONE;
				m_memberList.clear();

				bdBool ok = true;
				const bdUInt numResults = membersResult.getNumResults();			
				for(bdUInt i = 0; ok && (i < numResults); i++)
				{
					ok  = m_memberList.add(membersResult.m_memberList[i]);
				}
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::whisperToMemberDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Whisper to Member",			// READY
											"Whispering to Member",			// PENDING
											"Whispered to Member",			// DONE
											"Whisper to Member Failed",		// FAILED
											"Whisper to Member Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == FAILED)
	{
		if(m_memberList.getSize() == 0)
		{
			description += "\nNo members to whisper to.";
		}

		if(m_channelList.getSize() == 0)
		{
			description += "\nNot subscribed to any channels.";
		}
	}
}

void Simulation::startWhisperToMember()
{
	// fail safe
	m_taskStatus = FAILED;

	bdChannelChat *const channelChat = GetChannelChat();
	const bdUInt numMembers = m_memberList.getSize();
	const bdUInt numChannels = m_channelList.getSize();
	if(channelChat && (numMembers != 0) && (numChannels != 0))
	{		
		// Choose a userID that is not the one this instance is connected with
		const bdOnlineUserID localUserID = bdAuthUtility::getUserID(USERNAME);	

		bdBool found = false;
		for(bdUInt i=0; !found && (i < numMembers); i++)
		{
			bdOnlineUserID userID = m_memberList[i].m_userID;
			found = (userID != localUserID);
			if(found)
			{			
				bdNChar8 str[128];
				bdStrlcpy(str, "Whisper message", sizeof(str));		
				const bdUInt strLength = bdStrnlen(str, sizeof(str));

				bdByteBuffer payload(128);
				const bdBool ok = payload.write(str, strLength+1);

				if(ok)
				{
					// assumes m_channelList[0] corresponds to m_memberList
					m_remoteTask = channelChat->whisperToChannelMember(0,
																	   m_channelList[0].m_channelID,
																	   userID,
																	   payload);
					m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
				}
			}
		}		
	}
}

void Simulation::pumpWhisperToMember()
{
	pumpGenericChannelChatTask();
}

void Simulation::unsubscribeFromChannelDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Unsubscribe from Channel",				// READY
											"Unsubscribing from Channel",			// PENDING
											"Unsubscribed from Channel",			// DONE
											"Unsubscribe from Channel Failed",		// FAILED
											"Unsubscribe from Channel Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if((m_taskStatus == FAILED) && m_subscribedChannels.isEmpty())
	{
		description += "\nNot subscribed to any channels.";
	}
}

void Simulation::startUnsubscribeFromChannel()
{
	// fail safe
	m_taskStatus = FAILED;

	bdChannelChat *const channelChat = GetChannelChat();
	if(channelChat && !m_subscribedChannels.isEmpty())
	{
		m_remoteTask = channelChat->unsubscribeToChannel(0, m_subscribedChannels[0]);
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
		m_subscribedChannels.removeAt(0);
	}
}

void Simulation::pumpUnsubscribeFromChannel()
{
	pumpGenericChannelChatTask();
}

void Simulation::lobbyServiceDisconnectDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Disconnect from Lobby Service",			// READY
											"Disconnecting from Lobby Service",			// PENDING
											"Disconnected from Lobby Service",			// DONE
											"Disconnect from Lobby Service Failed",		// FAILED
											"Disconnect from Lobby Service Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startLobbyServiceDisconnect()
{
	// fail safe
	m_taskStatus = FAILED;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		lobbyService->disconnect();
		m_taskStatus = PENDING;
	}
}

void Simulation::pumpLobbyServiceDisconnect()
{
	// fail safe
	m_taskStatus = FAILED;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		const bdLobbyConnection::Status status = lobbyService->getStatus();
		switch(status)
		{
			case bdConnection::BD_DISCONNECTED:
			case bdConnection::BD_NOT_CONNECTED:
			{
				m_taskStatus = DONE;
			}
			break;
			case bdConnection::BD_CONNECTING:
			case bdConnection::BD_CONNECTED:
			{
				m_taskStatus = FAILED;
			}
			break;
			case bdConnection::BD_DISCONNECTING:
			{
				m_taskStatus = PENDING;
			}
			break;
		}
	}
}

void Simulation::stopNetworkDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Stop Network",					// READY
											"Stopping Network",				// PENDING
											"Stopped Network",				// DONE
											"Stopping Network Failed",		// FAILED
											"Stopping Network Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::stopNetwork()
{
	// fail safe
	m_taskStatus = FAILED;

	bdNetImpl *const net = bdNet::getInstance();
	if(net)
	{
		net->stop();
		if(net->getStatus() == bdNetImpl::BD_NET_STOPPED)
		{
			bdExamplePlatformInit::quit();
			m_taskStatus = DONE;
		}		
	}
}

void Simulation::pumpGenericChannelChatTask()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);

	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdChannelChat *const channelChat = GetChannelChat();
		if(channelChat)
		{
			const bdBool taskHandled = channelChat->handleTask(m_remoteTask, BD_NULL);
			if(taskHandled)
			{
				m_taskStatus = DONE;
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::checkMessageStore()
{
	if(m_messageTimer.getElapsedTimeInSeconds() > 10)
	{
		m_messageTimer.reset();
		m_messageText = "";
	}

	bdMessageStore *const messageStore = GetMessageStore();
	if(messageStore)
	{
		bdUInt numMessages = messageStore->getNumMessages();
		if(numMessages > 0)
		{
			m_messageText = "";
			m_messageTimer.start();
		}
		
		for(bdUInt i=0; i<numMessages; i++)
		{
			bdLobbyMessageRef msg = messageStore->getAt(0);
			switch(msg->m_msgType) 
			{
				case BD_CHANNEL_CHAT_BROADCAST_MSG:
				{
					bdChannelChatBroadcastMessageRef channelMsg = reference_cast<bdChannelChatBroadcastMessage>(msg);

					m_messageText += "\nReceived channel message: \"";
					
					bdNChar8 chatText[128];
					bdMemset(chatText, '\0', sizeof(chatText));
					channelMsg->m_data->read(chatText, sizeof(chatText));
					m_messageText += chatText;
					m_messageText += "\"\n";
				}
				break;
				case BD_CHANNEL_CHAT_WHISPER_MSG:
				{
					bdChannelChatWhisperMessageRef channelMsg = reference_cast<bdChannelChatWhisperMessage>(msg);

					m_messageText += "\nReceived channel whisper: \"";
					
					bdNChar8 chatText[128];
					bdMemset(chatText, '\0', sizeof(chatText));
					channelMsg->m_data->read(chatText, sizeof(chatText));
					m_messageText += chatText;
					m_messageText += "\"\n";
				}
				break;
				case BD_CHANNEL_USER_SUBSCRIBED_MSG:
				{
					bdChannelChatSubscribeMessageRef channelMsg = reference_cast<bdChannelChatSubscribeMessage>(msg);
					m_messageText += "\nReceived channel notification. User \"";
					m_messageText += channelMsg->m_originatorName;
					m_messageText += "\" subscribed\n";
				}
				break;
				case BD_CHANNEL_USER_UNSUBSCRIBED_MSG:
				{
					bdChannelChatUnsubscribeMessageRef channelMsg = reference_cast<bdChannelChatUnsubscribeMessage>(msg);
					m_messageText += "\nReceived channel notification. User \"";
					m_messageText += channelMsg->m_originatorName;
					m_messageText += "\" unsubscribed\n";
				}
				break;
				default:
				{
					// don't care...
				}
				break;
			}
			messageStore->removeAt(0);
		}
	}
}
