// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// This example illustrates how secure dedicated servers are supported in DemonWare.
// This is only relevant if your design accommodates trusted dedicated servers
// that report statistics on behalf of the players that connect to them. Such
// servers are often called "ranked servers".
// Such a design implied that credentials are required to run the server. When
// using Matchmaking+, a special license code is required so that the dedicated
// server can authenticate. Furthermore, the server's license ID (a public value 
// derived from the secret license code) can be used by a client to obtain an 
// authentication ticket so it can prove its identity to the dedicated server.
// This provides the basis for a highly secure dedicated server infrastructure.
// If your design specifies the use of game servers (either dedicated- or "listen"-
// servers), but you do not intend to use them to report ranked stats, then this
// example is not relevant to you.
// Please note that it is up to the studio/publisher to manage the distribution
// of license codes to dedicated server administrators.

#include "Simulation.h"
#include "MatchMakingQueries.h"
#include "LeaderboardEntry.h"

#include <bdLobby/bdLobbyCommon/bdAuthCommon.h>
#include <bdConnection/bdMessage/bdReceivedMessage.h>
#include <bdConnection/bdConnectionStore.h>

#define TITLE_ID (0x1267)

#define USERNAME "HostExample"
#define PASSWORD "MyPassword"

/// The key the dedicated server uses when authorizing with Demonware's Authorization server.
#define HOST_LICENSE_CODE "FQ3F-TL2B-09T4-SZ8Y-34R9"

/// The key that came on the clients cd, so that the client can create an account
#define CLIENT_LICENSE_CODE "ZC2B-EM8H-71N2-KU5A-39O4"

Simulation::Simulation()
: m_taskStatus(READY),
  m_task(START_NETWORK),
  m_changedScreen(true),
  m_authGetHostByName(BD_NULL),
  m_authService(BD_NULL),
  m_lsgGetHostByName(BD_NULL),
  m_remoteTask(BD_NULL),
  m_runningAsHost(false),
  m_pumpLobbyConnection(false),
  m_lobbyErrorCode(BD_NO_ERROR)
{
}

Simulation::~Simulation()
{
}

bdBool Simulation::init(bdInt32 ,
						bdNChar8* [])
{
	bdBool ok = true;

	// start logging
	// we want to hear about all warnings and errors
	bdLogImpl *log = bdLog::getInstance();
	log->subscribe("warn", &m_logSubscriber);
	log->subscribe("err", &m_logSubscriber);
	log->subscribe("info", &m_logSubscriber);

	// Register this instance as a dispatch interceptor so it can pick up
	// messages from clients if it is running as a dedicated host.
	bdNet::getInstance()->registerDispatchInterceptor(this);

	return ok;
}

bdBool Simulation::step(const bdFloat32)
{
	const bdBool quitKey = getScreen().wasKeyPressed(BD_QUIT);

	TaskStatus oldTaskStatus = m_taskStatus;

	if(m_taskStatus == READY)
	{
		startTask();
	}

	if(m_taskStatus == PENDING || m_taskStatus == DONE || m_taskStatus == FAILED)
	{
		// pump even if the status is DONE or FAILED to read out any results
		// and close any remote task handles
		pumpTask();
	}

	if(m_taskStatus == DONE || m_taskStatus == FAILED || m_taskStatus == SKIPPED)
	{
		// incase the task fails or completes in a single loop
		if(oldTaskStatus == READY)
		{
			oldTaskStatus = m_taskStatus;
		}

		displayTaskResult();
	}

	m_changedScreen = (oldTaskStatus != m_taskStatus);

	if(m_pumpLobbyConnection)
	{
		bdBool lostConnection = true;
		bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
		if(lobbyService)
		{
			const bdLobbyConnection::Status status = lobbyService->getStatus();
			switch(status)
			{
				case bdConnection::BD_NOT_CONNECTED:
				case bdConnection::BD_CONNECTING:
				case bdConnection::BD_DISCONNECTING:
				case bdConnection::BD_DISCONNECTED:
				{
					lostConnection = true;
				}
				break;
				case bdConnection::BD_CONNECTED:
				{
					lostConnection = false;
				}
				break;
			}
		}

		if(lostConnection)
		{
			// set up for a reconnect
			m_pumpLobbyConnection = false;
			m_task = STOP_NETWORK;
			m_taskStatus = READY;
			m_changedScreen = true;
			CloseRemoteTask(&m_remoteTask);
			m_lastTaskDescription = "Lost connection to lobby.";
		}
	}

	getScreen().printToScreen(m_screenText);

	if(m_changedScreen)
	{
		m_screenText = "";
	}

	// Pump the connection layer.
	if(bdNet::getInstance()->getStatus() == bdNetImpl::BD_NET_DONE)
	{
		bdNet::getInstance()->sendAll();
		bdNet::getInstance()->receiveAndDispatchAll();
	}
	
	return !quitKey;
}

void Simulation::quit()
{
	bdLogImpl *log = bdLog::getInstance();
	log->unsubscribe("warn", &m_logSubscriber);
	log->unsubscribe("err", &m_logSubscriber);
	log->unsubscribe("info", &m_logSubscriber);
	
	stopNetwork();

	delete m_authGetHostByName;
	delete m_authService;
	delete m_lsgGetHostByName;
	
	if(m_remoteTask)
	{
		m_remoteTask->close();
	}
}

void Simulation::startTask()
{	
	const bdBool skipKey = getScreen().wasKeyPressed(SKIP_KEY);

	if (skipKey)
	{
		m_taskStatus = SKIPPED;
	}
	else
	{
		const bdBool startKey = getScreen().wasKeyPressed(START_KEY);
		bdString taskDescription("Unknown task");

		switch(m_task)
		{
			case START_NETWORK:
			{
				startNetworkDescription(taskDescription);
				if(startKey)
				{
					startNetwork();
				}
			}
			break;
			case CHECK_DNS:
			{
				checkDNSDescription(taskDescription);
				if(startKey)
				{
					startCheckDNS();
				}
			}
			break;
			case START_HOST:
			{
				startHostDescription(taskDescription);
				if(startKey)
				{
					// We are not performing an online operation here, so
					// we just set the flag and set the task status to DONE
					// so that we immediately jump to the next state.
					m_runningAsHost = true;
					m_taskStatus = DONE;
				}
			}
			break;
			case CREATE_ACCOUNT:
			{
				createAccountDescription(taskDescription);
				if(startKey)
				{
					startCreateAccount();
				}
			}
			break;
			case AUTHENTICATE_CLIENT_ACCOUNT:
			{
				authenticateClientAccountDescription(taskDescription);
				if(startKey)
				{
					startAuthenticateClientAccount();
				}
			}
			break;
			case AUTHENTICATE_HOST_LICENSE:
				{
					authenticateHostLicenseDescription(taskDescription);
					if(startKey)
					{
						startAuthenticateHostLicense();
					}
				}
			break;
			case LOBBY_CONNECT:
			{
				lobbyServiceConnectDescription(taskDescription);
				if(startKey)
				{
					startLobbyServiceConnect();
				}
			}
			break;
			case ADVERTIZE_SESSION:
			{
				advertizeSessionDescription(taskDescription);
				if(startKey)
				{
					startAdvertizeSession();
				}
			}
			break;
			case WAIT_FOR_TICKETS:
			{
				waitForTicketsDescription(taskDescription);
				if(startKey)
				{
					startWaitForTickets();
				}
			}
			break;
			case WRITE_DEDICATED_STATS:
			{
				writeDedicatedStatsDescription(taskDescription);
				if(startKey)
				{
					startWriteDedicatedStats();
				}
			}
			break;
			case FIND_SESSION:
			{
				findSessionDescription(taskDescription);
				if(startKey)
				{
					startFindSession();
				}
			}
			break;
			case CLIENT_GETTING_HOST_TICKET:
			{
				authenticateClientForHostDescription(taskDescription);
				if(startKey)
				{
					startAuthenticateClientForHost();
				}
			}
			break;
			case CLIENT_CONNECTING_TO_HOST:
			{
				connectToHostDescription(taskDescription);
				if(startKey)
				{
					startConnectToHost();
				}
			}
			break;
			case DELETE_SESSION:
			{
				deleteSessionDescription(taskDescription);
				if(startKey)
				{
					startDeleteSession();
				}
			}
			break;
			case LOBBY_DISCONNECT:
			{
				lobbyServiceDisconnectDescription(taskDescription);
				if(startKey)
				{
					startLobbyServiceDisconnect();
				}
			}
			break;
			case STOP_NETWORK:
			{
				stopNetworkDescription(taskDescription);
				if(startKey)
				{
					stopNetwork();
				}
			}
			break;
		}
		
		if(m_changedScreen)
		{
			m_screenText += taskDescription;
			m_screenText += "\n" START_TEXT;
			m_screenText += "\n" SKIP_TEXT;

			if(m_lastTaskDescription.getLength())
			{
				m_screenText += "\n\n";
				m_screenText += m_lastTaskDescription;				
			}
		}
	}
}

void Simulation::pumpTask()
{
	bdString taskDescription("Unknown task");

	switch(m_task)
	{
		case START_NETWORK:
		{
			startNetworkDescription(taskDescription);
			pumpStartNetwork();
		}
		break;
		case CHECK_DNS:
		{
			checkDNSDescription(taskDescription);
			pumpCheckDNS();
		}
		break;
		case CREATE_ACCOUNT:
		{
			createAccountDescription(taskDescription);
			pumpCreateAccount();
		}
		break;
		break;
		case AUTHENTICATE_CLIENT_ACCOUNT:
		{
			authenticateClientAccountDescription(taskDescription);
			pumpAuthenticateClientAccount();
		}
		break;
		case AUTHENTICATE_HOST_LICENSE:
			{
				authenticateHostLicenseDescription(taskDescription);
				pumpAuthenticateHostLicense();
			}
			break;
		case LOBBY_CONNECT:
		{
			lobbyServiceConnectDescription(taskDescription);
			pumpLobbyServiceConnect();
		}
		break;
		case ADVERTIZE_SESSION:
		{
			advertizeSessionDescription(taskDescription);
			pumpAdvertizeSession();
		}
		break;
		case WAIT_FOR_TICKETS:
		{
			waitForTicketsDescription(taskDescription);
			pumpWaitForTickets();
		}
		break;
		case WRITE_DEDICATED_STATS:
		{
			writeDedicatedStatsDescription(taskDescription);
			pumpWriteDedicatedStats();
		}
		break;
		case FIND_SESSION:
		{
			findSessionDescription(taskDescription);
			pumpFindSession();
		}
		break;
		case CLIENT_GETTING_HOST_TICKET:
		{
			authenticateClientForHostDescription(taskDescription);
			pumpAuthenticateClientForHost();
		}
		break;
		case CLIENT_CONNECTING_TO_HOST:
		{
			connectToHostDescription(taskDescription);
			pumpConnectToHost();
		}
		break;
		case DELETE_SESSION:
		{
			deleteSessionDescription(taskDescription);
			pumpDeleteSession();
		}
		break;
		case LOBBY_DISCONNECT:
		{
			lobbyServiceDisconnectDescription(taskDescription);
			pumpLobbyServiceDisconnect();
		}
		break;
		case STOP_NETWORK:
		{
			// nothing to to here
		}
		break;
		default:
		{
			// don't care..
		}
		break;
	}

	if(m_changedScreen)
	{
		m_screenText += taskDescription;
	}
}

void Simulation::displayTaskResult()
{
	bdString taskDescription("Unknown task result");

	Task nextTaskIfDone = m_task;
	Task nextTaskIfFailed = m_task;
	Task nextTaskIfSkipped = m_task;

	bdBool displayLobbyError = true;

	switch(m_task)
	{
		case START_NETWORK:
		{
			startNetworkDescription(m_lastTaskDescription);
			nextTaskIfDone = CHECK_DNS;
			nextTaskIfFailed = START_NETWORK;
			nextTaskIfSkipped = START_NETWORK;
			displayLobbyError = false;
		}
		break;
		case CHECK_DNS:
		{
			checkDNSDescription(m_lastTaskDescription);
			nextTaskIfDone = START_HOST;
			nextTaskIfFailed = CHECK_DNS;
			nextTaskIfSkipped = START_HOST;
			displayLobbyError = false;
		}
		break;
		case START_HOST:
		{
			nextTaskIfDone = AUTHENTICATE_HOST_LICENSE;
			nextTaskIfFailed = CHECK_DNS;
			nextTaskIfSkipped = CREATE_ACCOUNT;
			displayLobbyError = false;
		}
		break;
		case CREATE_ACCOUNT:
		{
			createAccountDescription(m_lastTaskDescription);
			checkAuthServiceError();
			nextTaskIfDone = AUTHENTICATE_CLIENT_ACCOUNT;
			nextTaskIfFailed = CREATE_ACCOUNT;
			nextTaskIfSkipped = AUTHENTICATE_CLIENT_ACCOUNT;			
		}
		break;
		case AUTHENTICATE_CLIENT_ACCOUNT:
		{
			authenticateClientAccountDescription(m_lastTaskDescription);
			checkAuthServiceError();
			nextTaskIfDone = LOBBY_CONNECT;
			nextTaskIfFailed = AUTHENTICATE_CLIENT_ACCOUNT;
			nextTaskIfSkipped = LOBBY_CONNECT;		
		}
		break;
		case AUTHENTICATE_HOST_LICENSE:
		{
			authenticateHostLicenseDescription(taskDescription);
			checkAuthServiceError();
			nextTaskIfDone = LOBBY_CONNECT;
			nextTaskIfFailed = AUTHENTICATE_HOST_LICENSE;
			nextTaskIfSkipped = LOBBY_CONNECT;
		}
		break;
		case LOBBY_CONNECT:
		{
			lobbyServiceConnectDescription(m_lastTaskDescription);
			if(m_runningAsHost)
			{
				nextTaskIfDone = ADVERTIZE_SESSION;
				nextTaskIfFailed = LOBBY_CONNECT;
				nextTaskIfSkipped = ADVERTIZE_SESSION;
			}
			else
			{
				nextTaskIfDone = FIND_SESSION;
				nextTaskIfFailed = LOBBY_CONNECT;
				nextTaskIfSkipped = FIND_SESSION;
			}
		}
		break;
		case ADVERTIZE_SESSION:
		{
			advertizeSessionDescription(m_lastTaskDescription);
			checkMatchMakingError();
			nextTaskIfDone = WAIT_FOR_TICKETS;
			nextTaskIfFailed = ADVERTIZE_SESSION;
			nextTaskIfSkipped = WAIT_FOR_TICKETS;
		}
		break;
		case WAIT_FOR_TICKETS:
		{
			waitForTicketsDescription(m_lastTaskDescription);
			nextTaskIfDone = WAIT_FOR_TICKETS;
			nextTaskIfFailed = WAIT_FOR_TICKETS;
			nextTaskIfSkipped = WRITE_DEDICATED_STATS;
		}
		break;
		case WRITE_DEDICATED_STATS:
		{
			writeDedicatedStatsDescription(m_lastTaskDescription);
			checkStatsError();
			nextTaskIfDone = DELETE_SESSION;
			nextTaskIfFailed = WRITE_DEDICATED_STATS;
			nextTaskIfSkipped = DELETE_SESSION;
		}
		break;
		case FIND_SESSION:
		{
			findSessionDescription(m_lastTaskDescription);
			checkMatchMakingError();
			nextTaskIfDone = CLIENT_GETTING_HOST_TICKET;
			nextTaskIfFailed = FIND_SESSION;
			nextTaskIfSkipped = CLIENT_GETTING_HOST_TICKET;
		}
		break;
		case CLIENT_GETTING_HOST_TICKET:
		{
			authenticateClientForHostDescription(taskDescription);
			checkAuthServiceError();
			nextTaskIfDone = CLIENT_CONNECTING_TO_HOST;
			nextTaskIfFailed = FIND_SESSION;
			nextTaskIfSkipped = CLIENT_CONNECTING_TO_HOST;
		}
		break;
		case CLIENT_CONNECTING_TO_HOST:
		{
			connectToHostDescription(taskDescription);
			nextTaskIfDone = LOBBY_DISCONNECT;
			nextTaskIfFailed = FIND_SESSION;
			nextTaskIfSkipped = LOBBY_DISCONNECT;
		}
		break;
		case DELETE_SESSION:
		{
			deleteSessionDescription(m_lastTaskDescription);
			checkMatchMakingError();
			nextTaskIfDone = LOBBY_DISCONNECT;
			nextTaskIfFailed = DELETE_SESSION;
			nextTaskIfSkipped = LOBBY_DISCONNECT;
		}
		break;
		case LOBBY_DISCONNECT:
		{
			lobbyServiceDisconnectDescription(m_lastTaskDescription);			
			nextTaskIfDone = STOP_NETWORK;
			nextTaskIfFailed = LOBBY_DISCONNECT;
			nextTaskIfSkipped = STOP_NETWORK;
		}
		break;
		case STOP_NETWORK:
		{
			stopNetworkDescription(m_lastTaskDescription);
			nextTaskIfDone = START_NETWORK;
			nextTaskIfFailed = STOP_NETWORK;
			nextTaskIfSkipped = START_HOST;
		}
		break;
	}

	switch(m_taskStatus)
	{
		case READY:
		case PENDING:
		{
			BD_WARN(BD_LOG_LEVEL, "Unexpected task status.");
		}
		break;
		case DONE:
		{
			m_task = nextTaskIfDone;
		}
		break;
		case FAILED:
		{
			m_task = nextTaskIfFailed;
		}
		break;
		case SKIPPED:
		{
			m_task = nextTaskIfSkipped;
		}
		break;
	}

	// if task uses sets a lobby error code and has failed
	// display the error description
	if(displayLobbyError && (m_taskStatus == FAILED))
	{
		bdNChar8 buf[32];
		LobbyErrorCodeToString(m_lobbyErrorCode, buf, sizeof(buf));
		m_lastTaskDescription += "\nError: ";
		m_lastTaskDescription += buf;
		m_lobbyErrorCode = BD_NO_ERROR;
	}

	m_taskStatus = READY;
}

void Simulation::startNetworkDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Start Network",				// READY
											"Starting Network",				// PENDING
											"Started Network",				// DONE
											"Starting Network Failed",		// FAILED
											"Starting Network Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startNetwork()
{
	bdNetImpl *const net = bdNet::getInstance();
	bdNetStartParams params;
	params.m_onlineGame = true;

	// platform specific startup parameters
	bdPlatformParams platformParams;
	//TODO: check the return value here
	bdExamplePlatformInit::init(params, &platformParams);

	if(net->start(params))
	{
		m_taskStatus = PENDING;
	}
	else
	{
		m_taskStatus = FAILED;
	}
}

void Simulation::pumpStartNetwork()
{
	bdNetImpl *const net = bdNet::getInstance();
	net->pump();

	const bdNetImpl::bdNetStatus networkStatus = net->getStatus();
	switch(networkStatus)
	{
		case bdNetImpl::BD_NET_BIND_FAILED:
		case bdNetImpl::BD_NET_ONLINE_FAILED:
		case bdNetImpl::BD_NET_INIT_FAILED:
		case bdNetImpl::BD_NET_STOPPED:
		{
			// error conditions
			m_taskStatus = FAILED;
		}
		break;
		case bdNetImpl::BD_NET_PENDING:
		{
			// wait
			m_taskStatus = PENDING;
		}
		break;
		case bdNetImpl::BD_NET_DONE:
		{
			// complete
			m_taskStatus = DONE;
		}
	}
}

void Simulation::checkDNSDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Start DNS Lookup",			// READY
											"Performing DNS Lookup",	// PENDING
											"DNS Lookup Succeeded",		// DONE
											"DNS Lookup Failed",		// FAILED
											"DNS Lookup Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == DONE)
	{
		bdNChar8 addrString[BD_INET_ADDR_STRING_SIZE];
		m_authAddr.toString(addrString, sizeof(addrString));
		description +=  "\n " AUTH_ADDRESS ":";
		description += addrString;
		description +=  "\n";

		m_lsgAddr.toString(addrString, sizeof(addrString));
		description +=  " " LSG_ADDRESS ":";
		description += addrString;
	}
}		

void Simulation::startCheckDNS()
{
	const TaskStatus authDNSStatus = StartGetHostByName(AUTH_ADDRESS, &m_authGetHostByName, m_authAddr);
	const TaskStatus lsgDNSStatus = StartGetHostByName(LSG_ADDRESS, &m_lsgGetHostByName, m_lsgAddr);
	if((authDNSStatus == FAILED) || (lsgDNSStatus == FAILED))
	{
		m_taskStatus = FAILED;
	}

	if((authDNSStatus == PENDING) || (lsgDNSStatus == PENDING))
	{
		m_taskStatus = PENDING;
	}

	if((authDNSStatus == DONE) && (lsgDNSStatus == DONE))
	{
		m_taskStatus = DONE;
	}
}

void Simulation::pumpCheckDNS()
{
	TaskStatus authDNSStatus = DONE;
	TaskStatus lsgDNSStatus = DONE;

	if(m_authGetHostByName)
	{
		authDNSStatus = PumpGetHostByName(&m_authGetHostByName, m_authAddr);
	}

	if(m_lsgGetHostByName)
	{
		lsgDNSStatus = PumpGetHostByName(&m_lsgGetHostByName, m_lsgAddr);
	}

	if((authDNSStatus == FAILED) || (lsgDNSStatus == FAILED))
	{
		m_taskStatus = FAILED;
	}

	if((authDNSStatus == PENDING) || (lsgDNSStatus == PENDING))
	{
		m_taskStatus = PENDING;
	}

	if((authDNSStatus == DONE) && (lsgDNSStatus == DONE))
	{
		m_taskStatus = DONE;
	}
}

void Simulation::startHostDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Start Host",			// READY
											"Starting Host",		// PENDING
											"Start Host Succeeded",		// DONE
											"Start Host Failed",		// FAILED
											"Start Host Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

bdAuthService* Simulation::getAuthService()
{
	if(m_authService == BD_NULL)
	{
		const bdAddr authAddr(m_authAddr, AUTH_PORT);
		m_authService = new bdAuthService(TITLE_ID, authAddr);
	}

	return m_authService;
}

void Simulation::checkAuthServiceError()
{
	bdAuthService *const authService = getAuthService();
	if(authService)
	{
		m_lobbyErrorCode = authService->getErrorCode();
	}
}

void Simulation::pumpAuthServiceTask()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService)
	{
		m_taskStatus = PENDING;
		const bdAuthService::bdStatus status = authService->getStatus();
		if(status == bdAuthService::BD_READY)
		{			
			const bdLobbyErrorCode lobbyErrorCode = authService->getErrorCode();
			if(lobbyErrorCode == BD_AUTH_NO_ERROR)
			{
				m_taskStatus = DONE;
			}
			else
			{
				m_taskStatus = FAILED;
			}
		}
	}
}

void Simulation::createAccountDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Create Account (" USERNAME "/" PASSWORD ")",	// READY
											"Creating Account",								// PENDING
											"Created Account",								// DONE
											"Create Account Failed",						// FAILED
											"Create Account Skipped"};						// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startCreateAccount()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService)
	{
		if(authService->createAccount(USERNAME, PASSWORD, CLIENT_LICENSE_CODE))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpCreateAccount()
{
	pumpAuthServiceTask();
}

void Simulation::authenticateClientAccountDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Authenticate Account (" USERNAME "/" PASSWORD ")",	// READY
											"Authenticating Account",									// PENDING
											"Authenticated Account",									// DONE
											"Authenticate Account Failed",								// FAILED
											"Authenticate Account Skipped"};						// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startAuthenticateClientAccount()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService )
	{
		if(authService->authorizeAccount(USERNAME, PASSWORD))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpAuthenticateClientAccount()
{
	pumpAuthServiceTask();
}

void Simulation::authenticateHostLicenseDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Authenticate Host License (" HOST_LICENSE_CODE ")",	// READY
		"Authenticating License",									// PENDING
		"Authenticated License",									// DONE
		"Authenticate License Failed",								// FAILED
		"Authenticate License Skipped"};							// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startAuthenticateHostLicense()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService )
	{
		if(authService->authorizeDedicatedHost(HOST_LICENSE_CODE))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpAuthenticateHostLicense()
{
	pumpAuthServiceTask();
}

void Simulation::lobbyServiceConnectDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Connect to Lobby Service",				// READY
											"Connecting to Lobby Service",			// PENDING
											"Connected to Lobby Service",			// DONE
											"Connect to Lobby Service Failed",		// FAILED
											"Connect to Lobby Service Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startLobbyServiceConnect()
{
	// fail safe
	m_taskStatus = FAILED;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		const bdAddr lsgAddr(m_lsgAddr, LSG_PORT);
		const bdBool connecting = lobbyService->connect(lsgAddr, getAuthService()->getAuthInfo());
		if(connecting)
		{
			m_taskStatus = PENDING;
		}
		// no need to keep the bdAuthService instance after we are connected to the lobby
		delete m_authService;
		m_authService = BD_NULL;
	}
}

void Simulation::pumpLobbyServiceConnect()
{
	// fail safe
	m_taskStatus = FAILED;

	m_pumpLobbyConnection = false;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		const bdLobbyConnection::Status connectionStatus = lobbyService->getStatus();
		switch(connectionStatus)
		{			
		case bdConnection::BD_NOT_CONNECTED:
		case bdConnection::BD_DISCONNECTING:
		case bdConnection::BD_DISCONNECTED:
			{
				m_taskStatus = FAILED;
			}
			break;
		case bdConnection::BD_CONNECTING:
			{
				m_taskStatus = PENDING;
			}
			break;
		case bdConnection::BD_CONNECTED:
			{
				m_taskStatus = DONE;
				m_pumpLobbyConnection = true;
			}
			break;			
		}
	}
}

void Simulation::checkMatchMakingError()
{
	const bdMatchMaking *const matchmaking = GetMatchMaking();	
	if(matchmaking)
	{
		m_lobbyErrorCode = matchmaking->getErrorCode();
	}
}

void Simulation::checkStatsError()
{
	const bdStats *const stats = GetStats();	
	if(stats)
	{
		m_lobbyErrorCode = stats->getErrorCode();
	}
}

void Simulation::printSessionList(bdString& string) const
{
	const bdUInt numSessions = m_foundSessions.getSize();
	bdNChar8 buf[32];
	bdSnprintf(buf, sizeof(buf), "%u session(s) found:\n", numSessions);
	string += buf;
	m_foundSessions.print(string);
}

void Simulation::advertizeSessionDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Advertize Session",			// READY
											"Advertizing Session",			// PENDING
											"Advertized Session",			// DONE
											"Advertize Session Failed",		// FAILED
											"Advertize Session Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startAdvertizeSession()
{
	// fail safe
	m_taskStatus = FAILED;

	bdNetImpl *const net = bdNet::getInstance();
	bdMatchMaking *const matchmaking = GetMatchMaking();

	if(matchmaking && net)
	{
		// the session details
		const bdCommonAddrRef hostAddress = net->getLocalCommonAddr();
		const bdByte8 numPublicOpen    = 16;
		const bdByte8 numPrivateOpen   = 0;
		const bdByte8 numPublicFilled  = 0;
		const bdByte8 numPrivateFilled = 1;
		bdUInt64 licenseID = bdAuthUtility::getLicenseID(HOST_LICENSE_CODE);

		m_session.setHostAddr(hostAddress);
		m_session.setPublicOpen(numPublicOpen);
		m_session.setPublicFilled(numPublicFilled);
		m_session.setPrivateOpen(numPrivateOpen);
		m_session.setPrivateFilled(numPrivateFilled);
		m_session.setLicenseID(reinterpret_cast<bdUByte8*>(&licenseID), sizeof(licenseID));

		// send to the lobby service
		m_remoteTask = matchmaking->createSession(0, &m_session);
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	}
}

void Simulation::pumpAdvertizeSession()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	
	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdMatchMaking *const matchmaking = GetMatchMaking();
		if(matchmaking)
		{
			bdSessionCreateResult advertizeResult;		
			const bdBool taskHandled = matchmaking->handleTask(m_remoteTask, &advertizeResult);
			if(taskHandled)
			{
				m_taskStatus = DONE;
				m_session.setSecurityID(advertizeResult.m_secID);
				m_session.setSecurityKey(advertizeResult.m_secKey);

				// store the security  id and key returned by the server
				bdSecurityKeyMap *keyMap = bdNet::getInstance()->getKeyMap();
				keyMap->registerKey(m_session.getSecurityID(), m_session.getSecurityKey());
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::waitForTicketsDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Wait for Client Tickets",				// READY
											"Waiting for Client Tickets",			// PENDING
											"Received Client Ticket",				// DONE
											"Decryption of Client Ticket Failed",	// FAILED
											"Waiting for Client Tickets Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);

	if(m_taskStatus == DONE && !m_connectedUsers.isEmpty())
	{
		bdAuthTicket &ticket = m_connectedUsers[m_connectedUsers.getSize() - 1];
		description += "\nUsername: ";
		description += ticket.m_username;
		description += "\n";
	}
}

void Simulation::startWaitForTickets()
{
	m_taskStatus = PENDING;
}

void Simulation::pumpWaitForTickets()
{
}

void Simulation::writeDedicatedStatsDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Write Stats",						// READY
											"Writing Stats",					// PENDING
											"Write Stats Complete",				// DONE
											"Write Stats Failed",				// FAILED
											"Write Stats Skipped"};				// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startWriteDedicatedStats()
{
	// fail safe
	m_taskStatus = FAILED;

	bdStats *const stats = GetStats();
	if(stats && !m_connectedUsers.isEmpty())
	{
		// We use a bdArray for convenience here. Most games will have a max 
		// number of players, and could therefore use a non-dynamic array for this.
		//bdArray<SampleLeaderboard1> statsData(32);
		bdArray<LeaderboardEntry> statsData;

		// fill in stats data (using fake data here) for all connected users
		bdUInt i;
		for (i = 0; i < m_connectedUsers.getSize(); i++)
		{
			LeaderboardEntry s;

			// the stats values
			const bdInt64 rating = 10;
			const bdInt64 kills = 5;
			const bdInt32 deaths = 1;
			const bdInt32 assists = 3;
			const bdInt32 started = 20;
			const bdInt32 completed = 15;
			const bdFloat32 killRate = 0.26f;

			s.setRating(rating);
			s.setKills(kills);
			s.setDeaths(deaths);
			s.setAssists(assists);
			s.setStarted(started);
			s.setCompleted(completed);
			s.setKillRate(killRate);
			s.setEntityName(m_connectedUsers[i].m_username);
			s.setEntityID(m_connectedUsers[i].m_userID);

			statsData.pushBack(s);
		}

		// send to the lobby service
		bdStatsInfo *statsPtr = statsData.begin();
		m_remoteTask = stats->writeMultipleStats(	0,
													LEADERBOARD_ID,
													bdStats::STAT_WRITE_REPLACE,
													&statsPtr,
													statsData.getSize());
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);		
	}
}

void Simulation::pumpWriteDedicatedStats()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);

	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdStats *const stats = GetStats();
		if(stats)
		{
			const bdBool taskHandled = stats->handleTask(m_remoteTask, BD_NULL);
			if(taskHandled)
			{
				m_taskStatus = DONE;
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::findSessionDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Find Session",							// READY
											"Finding Session",						// PENDING
											"Find Session Complete",				// DONE
											"Find Session Failed",					// FAILED
											"Find Session Skipped"};				// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
	
	if(m_taskStatus == DONE)
	{
		description += "\n";
		printSessionList(description);
	}
}

void Simulation::startFindSession()
{
	// fail safe
	m_taskStatus = FAILED;

	bdMatchMaking *const matchmaking = GetMatchMaking();
	if(matchmaking)
	{
		bdUInt numResults = 5;
		FindSessions queryParams(numResults);
		m_remoteTask = matchmaking->findSessions(&queryParams);

		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	}
}

void Simulation::pumpFindSession()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);

	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdMatchMaking *const matchmaking = GetMatchMaking();
		if(matchmaking)
		{
			FindSessionsResult sessionInfo;	
			const bdBool taskHandled = matchmaking->handleTask(m_remoteTask, &sessionInfo);
			if(taskHandled)
			{
				m_taskStatus = DONE;
				m_foundSessions.clear();

				if(sessionInfo.getNumResults() > 0)
				{
					m_session = sessionInfo.m_customMMInfo[0];
					m_foundSessions.add(m_session);

					// register the session key for later use.
					bdSecurityKeyMap *keyMap = bdNet::getInstance()->getKeyMap();
					keyMap->registerKey(m_session.getSecurityID(), m_session.getSecurityKey());
				}
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::authenticateClientForHostDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Authenticate Account (" USERNAME "/" PASSWORD ") for Host",	// READY
											"Authenticating Account for Host",								// PENDING
											"Authenticated Account for Host",								// DONE
											"Authenticate Account for Host Failed",							// FAILED
											"Authenticate Account for Host Skipped"};						// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startAuthenticateClientForHost()
{
	// fail safe
	m_taskStatus = FAILED;

	bdAuthService *const authService = getAuthService();
	if(authService && m_foundSessions.getSize()>0)
	{
		bdUInt64 foundHostLicenseID;
		m_session.getLicenseID(reinterpret_cast<bdUByte8*>(&foundHostLicenseID), sizeof(foundHostLicenseID));
		if(authService->authorizeAccount(USERNAME, foundHostLicenseID, PASSWORD))
		{
			m_taskStatus = PENDING;
		}
	}
}

void Simulation::pumpAuthenticateClientForHost()
{
	pumpAuthServiceTask();
}

void Simulation::connectToHostDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Connect to Host",							// READY
											"Connecting to Host",						// PENDING
											"Connected to Host",						// DONE
											"Connect to Host Failed",					// FAILED
											"Connect to Host Skipped"};					// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startConnectToHost()
{
	// fail safe
	m_taskStatus = FAILED;

	//  connect to the server address
	bdConnectionStore *store = bdNet::getInstance()->getConnectionStore();
	m_connection = store->create(m_session.getHostAddr(), m_session.getSecurityID());
	if(m_connection)
	{
		if(m_connection->connect())
		{
			m_taskStatus = PENDING;
			
			// Register this as a listener so we know when the connection has
			// succeeded.
			m_connection->registerListener(this);
		}
	}
}

void Simulation::pumpConnectToHost()
{
}

void Simulation::sendCredentialsToDedicatedServer()
{
	bdMessageRef msg = new bdMessage(DEDICATED_SERVER_CREDENTIALS_MSG);
	bdBitBufferRef buffer = msg->getPayload();
	buffer->writeUInt32(getAuthService()->getAuthInfo()->m_IVSeed);
	buffer->writeBlob(getAuthService()->getAuthInfo()->m_data, BD_AUTH_TICKET_SIZE);
	m_connection->send(msg, true);
}

void Simulation::deleteSessionDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Delete Session",			// READY
											"Deleting Session",			// PENDING
											"Deleted Session",			// DONE
											"Delete Session Failed",		// FAILED
											"Delete Session Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startDeleteSession()
{
	// fail safe
	m_taskStatus = FAILED;
	
	bdMatchMaking *const matchmaking = GetMatchMaking();
	if(matchmaking)
	{
		// send to the lobby service
		m_remoteTask = matchmaking->deleteSession(0, m_session.getSecurityID());
		m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);
	}
}

void Simulation::pumpDeleteSession()
{
	m_taskStatus = CheckRemoteTaskStatus(m_remoteTask);

	if(m_taskStatus == DONE)
	{
		// fail safe
		m_taskStatus = FAILED;

		bdMatchMaking *const matchmaking = GetMatchMaking();
		if(matchmaking)
		{
			const bdBool taskHandled = matchmaking->handleTask(m_remoteTask, BD_NULL);
			if(taskHandled)
			{
				m_taskStatus = DONE;
			}
		}
	}

	if((m_taskStatus == DONE) || (m_taskStatus == FAILED))
	{
		CloseRemoteTask(&m_remoteTask);
	}
}

void Simulation::lobbyServiceDisconnectDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Disconnect from Lobby Service",			// READY
											"Disconnecting from Lobby Service",			// PENDING
											"Disconnected from Lobby Service",			// DONE
											"Disconnect from Lobby Service Failed",		// FAILED
											"Disconnect from Lobby Service Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::startLobbyServiceDisconnect()
{
	// fail safe
	m_taskStatus = FAILED;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		lobbyService->disconnect();
		m_taskStatus = PENDING;
	}
}

void Simulation::pumpLobbyServiceDisconnect()
{
	// fail safe
	m_taskStatus = FAILED;

	bdLobbyServiceImpl *const lobbyService = bdLobbyService::getInstance();
	if(lobbyService)
	{
		const bdLobbyConnection::Status status = lobbyService->getStatus();
		switch(status)
		{
		case bdConnection::BD_DISCONNECTED:
		case bdConnection::BD_NOT_CONNECTED:
			{
				m_taskStatus = DONE;
			}
			break;
		case bdConnection::BD_CONNECTING:
		case bdConnection::BD_CONNECTED:
			{
				m_taskStatus = FAILED;
			}
			break;
		case bdConnection::BD_DISCONNECTING:
			{
				m_taskStatus = PENDING;
			}
			break;
		}
	}
}

void Simulation::stopNetworkDescription(bdString& description)
{
	const bdNChar8 *const descriptions[] = {"Stop Network",					// READY
											"Stopping Network",				// PENDING
											"Stopped Network",				// DONE
											"Stopping Network Failed",		// FAILED
											"Stopping Network Skipped"};	// SKIPPED

	const bdUInt numDescriptions = sizeof(descriptions) / sizeof(bdNChar8*);
	IndexStrings(descriptions, numDescriptions, m_taskStatus, description);
}

void Simulation::stopNetwork()
{
	bdNetImpl *const net = bdNet::getInstance();
	net->stop();
	if(net->getStatus() == bdNetImpl::BD_NET_STOPPED)
	{
		bdExamplePlatformInit::quit();
		m_taskStatus = DONE;
	}
	else
	{
		m_taskStatus = FAILED;
	}
}

bdBool Simulation::accept(const bdReceivedMessage &receivedMsg)
{
	bdBool handled = false;
	const bdMessageRef msg = receivedMsg.getMessage();
	const bdBitBufferRef buffer = msg->getPayload();
	if(msg.notNull() && buffer.notNull() && (msg->getType()  == DEDICATED_SERVER_CREDENTIALS_MSG))	
	{
		bdUInt32 IVSeed;
		bdBool ok = buffer->readUInt32(IVSeed);

		bdUByte8 ticket[BD_AUTH_TICKET_SIZE];
		bdUInt32 ticketSize = BD_AUTH_TICKET_SIZE;
		ok = ok && buffer->readBlob(ticket, ticketSize);

		if(ok && (ticketSize == BD_AUTH_TICKET_SIZE))
		{
			bdUByte8 initialVector[BD_TIGER_HASH_SIZE];
			bdCryptoUtils::calculateInitialVector(IVSeed, initialVector);

			bdByte8 licenseKey[BD_ENCRYPTION_KEY_SIZE];
			bdAuthUtility::getLicenseKey(HOST_LICENSE_CODE, licenseKey);

			bdCryptoUtils::decrypt(licenseKey, initialVector, ticket, ticket, BD_AUTH_TICKET_SIZE);

			bdAuthTicket authTicket;
			authTicket.deserialize(ticket);
			if (authTicket.m_magicNumber != BD_MAGIC_NUMBER)
			{
				BD_ERR("err", "Decryption of auth ticket failed.");
				m_taskStatus = FAILED;
			}
			else if (static_cast<bdUInt32>(time(NULL)) > authTicket.m_timeExpires)
			{
				BD_ERR("err", "Timestamp expired.");
				m_taskStatus = FAILED;
			}
			else
			{
				m_connectedUsers.pushBack(authTicket);

				m_taskStatus = DONE;
			}
		}
		else
		{
			BD_ERR("err", "Error deserializing message from client.");
			m_taskStatus = FAILED;
		}
		handled = true;
	}
	return handled;
}

void Simulation::onConnect(bdConnectionRef connection)
{
	sendCredentialsToDedicatedServer();
	m_taskStatus = DONE;
}

void Simulation::onConnectFailed(bdConnectionRef connection)
{
	m_taskStatus = FAILED;
}
