// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#ifndef BD_CHANNEL_CHAT_H
#define BD_CHANNEL_CHAT_H

#include <bdCore/bdCore.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdLobbyErrorCode.h>
#include <bdLobby/bdLobbyCommon/bdQueryResult.h>

class bdRemoteTask;
class bdRemoteTaskManager;

BD_REFERENCE(bdConnection);

// Channel Chat operation types.
#define BD_CHANNEL_CHAT_SUBSCRIBE				(1u)
#define BD_CHANNEL_CHAT_UNSUBSCRIBE				(2u)
#define BD_CHANNEL_CHAT_DISCOVER				(3u)
#define BD_CHANNEL_CHAT_SEND					(4u)
#define BD_CHANNEL_CHAT_WHISPER					(5u)
#define BD_CHANNEL_CHAT_GET_MEMBERS				(6u)

/// Client side interface to the DemonWare channel based chat service.
/// Use the functions contained in this class to perform operations on the server.
/// Each of these function returns a task handle, which should be pumped to 
/// completion by calling its getStatus() function. When the status of a task is 
/// bdRemoteTask::BD_DONE, you can pass it into the bdChannelChat::handleTask()
/// function. Once this is done, the error code for the task can be accessed via
/// bdChannelChat::getErrorCode(). If you expect a result set to be returned from
/// the server as a result of the task you started, you must pass an instance of 
/// a bdQueryResult into bdChannelChat::handleTask. The implementation must be
/// specialized to match the results you are expecting. Standard implementations
/// are provided.
class bdChannelChat
{
	public:
		BD_DECLARE_NEW_AND_DELETE_OPERATORS		
		
		enum bdChannelType
		{
			BD_BUILT_IN_CHANNEL,
			BD_PUBLIC_USER_CHANNEL,
			BD_PRIVATE_USER_CHANNEL
		};

		enum bdChannelMemberType
		{
			BD_CHANNEL_MEMBER,
			BD_CHANNEL_OWNER
		};

		~bdChannelChat();

		/// Adds a user to the specified channel. 
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param channelID[in]		The ID of the chat channel.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note		A BD_CHANNEL_USER_SUBSCRIBED_MSG will be delivered to all
		///				channel members.
		bdRemoteTask *subscribeToChannel(bdUByte8 controllerIndex, 
										 bdUInt64 channelID);

		/// Removes a user from the specified channel.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param channelID[in]		The ID of the chat channel.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note		A BD_CHANNEL_USER_UNSUBSCRIBED_MSG will be delivered to all
		///				channel members.
		bdRemoteTask* unsubscribeToChannel(bdUByte8 controllerIndex, 
										   bdUInt64 channelID);

		/// Discover the available channels.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdChannelChat::handleTask with an instance of 
		///			bdGetChannelsResult.
		bdRemoteTask *getChannelList(bdUByte8 controllerIndex);

		/// Sends a message to the specified channel.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param channelID[in]		The ID of the chat channel.
		/// \param payload[in]			The payload of the message to send.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note		A BD_CHANNEL_CHAT_BROADCAST_MSG will be delivered to all 
		///				members of the channel.
		bdRemoteTask* sendToChannel(bdUByte8 controllerIndex, 
									bdUInt64 channelID,
									bdByteBuffer& payload);

		/// Sends a message to the specified channel member.  The user must be a 
		/// member of the same subchannel as the sender.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param channelID[in]		The ID of the chat channel.
		/// \param userID[in]			The ID of the recipient.
		/// \param payload[in]			The payload of the message to send.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note		A BD_CHANNEL_CHAT_WHISPER_MSG will be delivered to the
		///				user specified by userID.
		bdRemoteTask* whisperToChannelMember(bdUByte8 controllerIndex, 
											 bdUInt64 channelID,
											 bdOnlineUserID userID,
											 bdByteBuffer& payload);

		/// Retrieves the userIDs and usernames of the members of a specified 
		/// channel, which are also members of the same sub-channel as the requestor.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param channelID[in]		The ID of the chat channel.
		/// \return A pointer to the task that handles the request. BD_NULL is returned
		///			if an error occurred. 
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdChannelChat::handleTask with an instance of 
		///			bdGetChannelMembersResult.
		bdRemoteTask *getChannelMembers(bdUByte8 controllerIndex, 
										bdUInt64 channelID);


		/// Handles a completed task. 
		/// This retrieves the error code returned from the server, and fills out
		/// a result set of data, if appropriate. The specified task must have 
		/// been created by this class. 
		/// If the task is expected to return results, the second parameter should
		/// be specified. bdQueryReply is the base-class for classes that handle
		/// result sets returned from the server. A specialized class for each kind
		/// of result set is located in the same directory as this class.
		/// \param remoteTask[in]	A pointer to the remote task to handle.
		/// \param queryResult[in]	An optional pointer to an object that handles 
		///							the expected type of result data.
		/// \return - True if the task was successfully handled.
		///			- False otherwise. Call bdChannelChat::getErrorCode() to get the 
		///			  error condition.
		bdBool handleTask(const bdRemoteTask *remoteTask, 
						  bdQueryResult *queryResult = BD_NULL);

		/// Return the error code of the last operation that failed.
		/// \return The last error code
		bdLobbyErrorCode getErrorCode() const;

	protected:

		friend class bdLobbyServiceImpl;

		/// Constructs a channel based chat manager with the given task handler.
		/// Note: never call this constructor directly, use bdLobbyService::getChannelChatt() instead.
		/// \param remoteTaskManager[in] A pointer to the remote task manager that handles
		/// requests and replies
		bdChannelChat(bdRemoteTaskManager *const remoteTaskManager);

		bdLobbyErrorCode m_errorCode;

		bdRemoteTaskManager *m_remoteTaskManager;

};

#endif // BD_CHANNEL_CHAT_H
