// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE:	The client-side API for writing, deleting and reading bdKeyPair's

#ifndef BD_KEY_ARCHIVE_H
#define BD_KEY_ARCHIVE_H

#include <bdCore/bdCore.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>

#include <bdLobby/bdKeyArchive/bdKeyPairDict.h>

class bdRemoteTask;
class bdRemoteTaskManager;
class bdQueryResult;

// bdKeyArchive operations
#define BD_KEYARCHIVE_WRITE_PAIRS			(1u)
#define BD_KEYARCHIVE_DELETE_PAIRS			(2u)
#define BD_KEYARCHIVE_READ_PAIRS			(3u)
#define BD_KEYARCHIVE_READ_ALL_PAIRS		(4u)
#define BD_KEYARCHIVE_DELETE_USERS_PAIRS	(5u)

class bdKeyArchive
{
	public:
		
		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		~bdKeyArchive();

		/// Writes a bdkeyPairDicts contents to the server.
		/// Usage scenarios:
		/// Standard Client(users): can only write his own keyPairs, or a teams keyPairs if the client is a 
		///  member of the team.
		/// Dedicated Server: can write any users keyPairs, and any teams keyPairs
		/// \param entries[in]	A bdKeyPairDict which contains bdKeyPairs 
		///						(combination of bdArchiveKey & bdArchiveValue) to be written to the server
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		bdRemoteTask* write(const bdKeyPairDict &entries);

		/// Reads a bdkeyPairs from the server, into a bdKeyPairDict
		/// Both clients and dedicated servers can read all users and all teams bdKeyPairs.
		/// \param entries[in]	A bdKeyPairDict which contains sets of bdArchiveKeys per bdEntityID to be read.
		///						Any values contained in the bdKeyPairDict are ignored.
		/// \param readDedicated[in]	Specifies wheather to read keypairs which were written by users, or 
		///								to read keypairs which were written by dedicated servers.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		bdRemoteTask* read(const bdKeyPairDict &entries,
						   const bdBool readDedicated);

		/// Reads all keyPairs belonging to a specific entityID
		/// Both clients and dedicated servers can read all users and all teams bdKeyPairs.
		/// \param entityID[in]	The entityID whose keyPairs will be returned.
		/// \param isTeam[in]	Specifies if the requested entityID is a team or user ID.
		/// \param readDedicated[in]	Specifies wheather to read keypairs which were written by users, or 
		///								to read keypairs which were written by dedicated servers.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		bdRemoteTask* readAll(const bdEntityID entityID,
							  const bdBool isTeam,
							  const bdBool readDedicated);

		/// Removes a bdkeyPairDicts contents from the server.
		/// Usage scenarios:
		/// Standard Client(users): can only remove his own keyPairs, or a teams keyPairs if the client 
		///  is a member of the team.
		/// Dedicated Server: can remove any users keyPairs, and any teams keyPairs.
		/// \param entries[in]	A bdKeyPairDict which contains bdKeyPairs 
		///						(combination of bdArchiveKey & bdArchiveValue) to be removed from the server.
		///						Any values contained in the bdKeyPairDict are ignored.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		bdRemoteTask* remove(const bdKeyPairDict &entries);

		/// Removes all bdkeyPairs belonging to a user on the server
		/// Only the entityID who owns the keypairs can delete them.
		/// \param entityID[in]	The entityID whose keyPairs will be deleted.
		/// \param isTeam[in]	Specifies if the requested entityID is a team or user ID.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		bdRemoteTask* removeAll(const bdEntityID entityID,
								const bdBool isTeam);

		/// Handle a completed task. 
		/// This retrieves the error code returned from the server, and fills out
		/// a result set of data, if appropriate. The specified task must have 
		/// been created by this class. 
		/// If the task is expected to return results, the second parameter should
		/// be specified. 
		/// \param remoteTask[in]	A pointer to the remote task to handle.
		/// \param queryResult[in]	An optional pointer to an object that handles 
		///							the expected type of result data.
		/// \return - True if the task was successfully handled.
		///			- False otherwise. Call bdKeyArchive::getErrorCode() to get the 
		///			  error condition.
		bdBool handleTask(const bdRemoteTask *remoteTask, 
						  bdQueryResult *queryResult = BD_NULL);

		/// Return the error code of the last operation that failed.
		/// \return The last error code
		bdLobbyErrorCode getErrorCode() const;

	protected:

		friend class bdLobbyServiceImpl;

		bdKeyArchive(bdRemoteTaskManager *const remoteTaskManager);

		/// remote task manager
		bdRemoteTaskManager *m_remoteTaskManager;

		/// error code store for look up
		bdLobbyErrorCode m_errorCode;
};

#endif //BD_KEY_ARCHIVE_H
