// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: Constitutes the main entry point for all of the DemonWare Lobby Services.
// This class provides access to the individual services and maintains the connection
// to the gateway.

#ifndef BD_LOBBY_SERVICES_H
#define BD_LOBBY_SERVICES_H

#include <bdCore/bdCore.h>

#include <bdLobby/bdLobbyConnection.h>
#include <bdLobby/bdLobbyConnectionListener.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdMessageStore/bdMessageStore.h>

#define BD_PROTOCOL_VERSION (180)

BD_REFERENCE(bdAuthInfo);
BD_REFERENCE(bdLobbyConnection);

class bdMessageStore;
class bdMatchMaking;
class bdStats;
class bdProfiles;
class bdTeams;
class bdFriends;
class bdStorage;
class bdChannelChat;
class bdTitleUtilities;
class bdKeyArchive;
class bdRemoteTaskManager;


#undef connect

/// Constitutes the main entry point for all of the DemonWare Lobby Services.
/// This class provides access to the individual services and maintains the connection
/// to the gateway.
class bdLobbyServiceImpl :public bdLobbyConnectionListener
{
	public:
		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		/// Default constructor. 
		bdLobbyServiceImpl();

		/// Destructor.
		~bdLobbyServiceImpl();

		enum bdMessagingMode
		{
			BD_RETAIN_PUSH_MESSAGES,
			BD_DISCARD_PUSH_MESSAGES
		};

#if (!defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)

		/// Connects to the Lobby Service Gateway.
		/// \param lobbyServiceAddr[in]		The address of the LSG
		/// \param authInfo[in]				Authentication ticket
		bdBool connect(	bdAddr lobbyServiceAddr, 
						const bdAuthInfo *authInfo,
						const bdBool useEncryption = true);

		/// Connects to the Lobby Service Gateway.
		/// \param lobbyServiceAddr[in]	The address of the LSG
		/// \param authInfo[in]			Authentication ticket, which should be 
		///								obtained from one of the authorize* 
		///								functions in bdAuthService.
		/// \param sendBuffer[in]		A pointer to a chunk of memory to be 
		///								used as send buffer.
		/// \param sendBufSize[in]		size of the send buffer
		/// \param recvBuffer[in]		A pointer to a chunk of memory to be 
		///								used as receive buffer.
		/// \param recvBufSize[in]		size of the receive buffer
		/// \param useEncryption[in]	Flag to indicate whether the connection is to be encrypted.
		bdBool connect(bdAddr lobbyServiceAddr, 
					   const bdAuthInfo *authInfo,
                       void *sendBuffer,
					   bdUWord sendBufSize,
					   void *recvBuffer, 
					   bdUWord recvBufSize,
					   const bdBool useEncryption = true);

#endif
#if (defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
		/// Xbox-specific version of the connect call, for use when accessing
		/// the Lobby Service Gateway via Microsoft's Live Server Platform (LSP).
		/// \note XNetRegisterKey must be called before connect
		/// is called.
		/// \param	lobbyServiceAddr[in]	The address of the LSG.
		/// \param	id[in]					The security ID of the key previously
		///									registered for accessing LSP.
		/// \param	authInfo[in]			A ticket constructed locally, instead
		///									of via the bdAuthService.
		bdBool connect(	bdCommonAddrRef lobbyServiceAddr,
						bdSecurityID &id,
						const bdAuthInfo *authInfo);

#elif defined BD_PLATFORM_XENON
		/// Xenon-specific version of the connect call, for use when accessing
		/// the Lobby Service Gateway via Microsoft's Live Server Platform (LSP).
		/// \note XNetRegisterKey must be called before connect
		/// is called.
		/// \param	lobbyServiceAddr[in]	The address of the LSG.
		/// \param	id[in]					The security ID of the key previously
		///									registered for accessing LSP.
		/// \param	authInfo[in]			A ticket constructed locally, instead
		///									of via the bdAuthService.
		bdBool connect(	bdAddr lobbyServiceAddr,
						bdSecurityID &id,
						const bdAuthInfo *authInfo);
#endif 
		/// Disconnects from the Lobby Services Gateway.
		void disconnect();

		/// Gets the status of the underlying connection to the gateway.
		bdLobbyConnection::Status getStatus();

		/// Gets an instance of the message store.
		bdMessageStore* getMessageStore();

		/// Gets an instance of the matchmaking lobby service.
		bdMatchMaking* getMatchMaking();

		/// Gets an instance of the profiles lobby service.
		bdProfiles* getProfilesManager();

		/// Gets an instance of the stats lobby service.
		bdStats* getStats();

		/// Gets an instance of the teams lobby service.
		bdTeams* getTeams();

		/// Gets an instance of the friends lobby service.
		bdFriends* getFriends();

		/// Gets an instance of the storage lobby service.
		bdStorage* getStorage();

		// Gets an instance of the channel based chat lobby service.
		bdChannelChat* getChannelChat();

		// Gets an instance of the title utilities lobby service.
		bdTitleUtilities* getTitleUtilities();

		// Gets an instance of the key archive lobby service.
		bdKeyArchive* getKeyArchive();

		/// Sets the messaging mode.
		/// \param messagingMode[in]	If BD_RETAIN_PUSH_MESSAGES is specified,
		///								messages that are pushed from the LSG 
		///								are stored in the internal	message store. 
		///								BD_DISCARD_PUSH_MESSAGES discards all push
		///								messages.
		void setMessagingMode(bdLobbyServiceImpl::bdMessagingMode messagingMode);

		/// Gets the lobby service task manager. Useful for custom services.
		bdRemoteTaskManager* getTaskManager();

	protected:

		friend class bdLobbyConnection;

		/// Event handler for connection related events
		void onConnect(bdLobbyConnectionRef connection);

		/// Event handler for connection related events
		void onDisconnect(bdLobbyConnectionRef connection);

		bdRemoteTaskManager *m_taskManager;

		/// An instance of the user profiles lobby service
		bdProfiles *m_profiles;

		/// An instance of the message store
		/// The message store is not created dynamically so that
		/// messages can still be retrieved after disconnection from
		/// the server
		bdMessageStore m_messageStore;

		bdMessagingMode m_messagingMode;

		/// An instance of the matchmaking lobby service
		bdMatchMaking *m_matchMaking;

		/// An instance of the stats lobby service
		bdStats *m_statsManager;

		/// An instance of the friends lobby service
		bdFriends *m_friendsManager;

		/// An instance of the teams lobby service
		bdTeams* m_teamsManager;

		/// An instance of the storage lobby service
		bdStorage *m_storageManager;

		/// An instance of the channel chat lobby service
		bdChannelChat *m_channelChatManager;

		/// An instance of the title utilities lobby service
		bdTitleUtilities *m_titleUtilitiesManager;

		/// An instance of the key archive lobby service
		bdKeyArchive *m_keyArchive;

		/// Title ID
		bdTitleID m_titleID;

		/// The connection to the Lobby Services Gateway
		bdLobbyConnectionRef m_lobbyConnection;

		bdAuthInfo m_authInfo;

		bdBool m_encryptedConnection;

		bdUInt	m_errorCode;
		
};

/// This class provides a point of access to the singleton instance.
/// Implementation details can be found in bdLobbyServiceImpl.
class bdLobbyService : public bdSingleton<bdLobbyServiceImpl>
{
	BD_DECLARE_NEW_AND_DELETE_OPERATORS

	protected:
		/// Default constructor.
		bdLobbyService();
};

#endif //BD_LOBBY_SERVICES_H
