// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE:	The client-side API for writing, deleting and reading user or team
//			statistics.

#ifndef BD_STATS_H
#define BD_STATS_H

#include <bdCore/bdCore.h>
#include <bdLobby/bdLobbyCommon/bdQueryParams.h>
#include <bdLobby/bdLobbyErrorCode.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>

class bdRemoteTask;
class bdRemoteTaskManager;
class bdStatsInfo;
class bdStatsQueryResult;
class bdQueryResult;
class bdQueryParams;
class bdStatsSubmission;

// Stats operations
#define BD_STAT_WRITE				(1u)
#define BD_STAT_DELETE				(2u)
#define BD_STAT_READ_BY_ENTITY_ID	(3u)
#define BD_STAT_READ_MULTIPLE_IDS	(4u)
#define BD_STAT_READ_LEADERBOARD	(5u)
#define BD_STAT_CUSTOM_READ			(6u)
#define BD_STAT_WRITE_MULTI_STATS	(7u)
#define BD_STAT_START_ARBITRATED_SESSION	(8u)
#define BD_STAT_WRITE_ARBITRATED_STATS		(9u)
#define BD_STAT_WRITE_MULTI_STATS_180	(10u)

/// The client-side API for writing, deleting and reading user or team statistics.
/// The statistics service manages an arbitrary number of leaderboards for a
/// title ID. A user or team can have data in multiple leaderboards. A user or
/// a team is generically referred to as an \a entity or the purposes of this class.
/// Each entry in a leaderboard is a row of information about one entity's performance.
/// Each entry has a compulsory \a rating attribute, on which the leaderboard
/// is ranked.
/// Each leaderboard can also define an arbitrary number of custom attributes for
/// each entry, which can be used to store user or team information relevant to the
/// leaderboard, and which are returned in the result data for queries of that
/// leaderboard. Custom attributes must be configured for a title ID in consultation
/// with the DemonWare support team.
/// Leaderboards are typically queried directly by entity ID (i.e. the user ID or
/// the team ID), or by ranked position.
/// Use the functions contained in this class to perform operations on the server.
/// Each of these function returns a task handle, which should be pumped to
/// completion by calling its getStatus() function. When the status of a task is
/// bdRemoteTask::BD_DONE, you can pass it into the bdStats::handleTask()
/// function. Once this is done, the error code for the task can be accessed via
/// bdStats::getErrorCode(). If you expect a result set to be returned from
/// the server as a result of the task you started, you must pass an instance of
/// a bdQueryResult into bdStats::handleTask. The implementation must
/// match the results you are expecting. If the result set being returned contains
/// customized attributes, you will need a specialized implementation of bdQueryResult.
class bdStats
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		enum bdWriteType
		{
			STAT_WRITE_REPLACE = 0,
			STAT_WRITE_ADD,
			STAT_WRITE_MAX,
			STAT_WRITE_MIN
		};

		~bdStats();

		/// Write an entire entry to a leaderboard.
		/// If no entry exists on the leaderboard, this function
		/// will create one.
		/// \param controllerIndex[in]	The controller index of the requestor, whose
		///								stats entry is to be written.
		/// \param leaderBoardID[in]	The ID of the leaderboard to write the stats to.
		/// \param writeType[in]		The type of stats update. Options are:
		///								- STAT_WRITE_REPLACE: replaces existing stats
		///								- STAT_WRITE_ADD: add new stats to
		///								  existing stat values.
		///								- STAT_WRITE_MAX: uses max value compared between
		///								  existing stats and new stats passed in query.
		///								- STAT_WRITE_MIN: uses min value compared between
		///								  existing stats and new stats passed in query.
		/// \param stats[in]			The bdStatsInfo object describing the stats
		///								to write.
		///								This will normally be an instance of a derived
		///								class, which matches the custom attributes of
		///								entries on the specified leaderboard.
		/// \param teamID[in]			Optional team ID to write the stats for.
		///								If specified (defaults to 0), stats for the
		///								local user specified by controllerIndex will
		///								NOT be written, but instead stats for the
		///								specified teamID will be written instead.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_INVALID_LEADERBOARD_ID - if leaderBoardID is not a valid leaderboard
		///			or a teamID has been specified and the leaderboard is not a team leaderboard.
		///			BD_ACCESS_DENIED - if this user is attempting is not a member of the team
		///			specified by teamID or the teams service is not enabled.
		///			BD_PARAM_MISMATCHED_TYPE - if the format of the stats parameter does not
		///			match that of the leaderboard specified by leaderBoardID.
		bdRemoteTask* writeStats(const bdUByte8 controllerIndex,
								 const bdLeaderBoardID leaderBoardID,
								 const bdWriteType writeType,
								 bdStatsInfo *stats,
								 const bdTeamID teamID = 0);


		/// Write statistical data for multiple users to a leaderboard.
		/// This may only be used by dedicated servers, ie those that have been
		/// authenticated by bdAuthService::authorizeDedicatedHost.
		/// If no entry exists on the leaderboard, this function
		/// will create one.
		/// \param controllerIndex[in]	The controller index of the requestor.
		/// \param leaderBoardID[in]	The ID of the leaderboard to write the stats to.
		/// \param writeType[in]		The type of stats update. Options are:
		///								- STAT_WRITE_REPLACE: replaces existing stats
		///								- STAT_WRITE_ADD: add new stats to
		///								  existing stat values.
		///								- STAT_WRITE_MAX: uses max value compared between
		///								  existing stats and new stats passed in query.
		///								- STAT_WRITE_MIN: uses min value compared between
		///								  existing stats and new stats passed in query.
		/// \param stats[in]			Pointer to an array of bdStatsInfo*.
		///								Each bdStatsInfo* is an object describing
		///								the stats to write.
		///								This will normally be an instance of a derived
		///								class, which matches the custom attributes of
		///								entries on the specified leaderboard.
		///								This writes to a single leaderboard.  Leaderboards are either
		///								user leaderboards or team leaderboards.
		///								As such, this fn will only write user stats to a user leaderboards,
		///								and team stats to a team leaderboard.  Sending mixed stats sets will
		///								only result in the incorrect sets not being written.
		/// \param numStats[in]			The number of entries that the dedicated server
		///								is reporting stats for.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_INVALID_LEADERBOARD_ID - if leaderBoardID is not a valid leaderboard.
		///			BD_ACCESS_DENIED - if this user is attempting is not a host.
		///			BD_PARAM_MISMATCHED_TYPE - if the format of the stats parameter does not
		///			match that of the leaderboard specified by leaderBoardID.
		///			BD_INVALID_STATS_SET - if no stats were passed up in the stats parameter.
		bdRemoteTask* writeMultipleStats(const bdUByte8 controllerIndex,
										 const bdLeaderBoardID leaderBoardID,
										 const bdWriteType writeType,
										 bdStatsInfo **stats,
										 const bdUInt numStats);

		/// Write statistical data for multiple users or teams to multiple leaderboards.
		/// This function behaves similarly to calling \a writeStats multiple times,
		/// but results in a single MM+ call and is much faster.
		///
		/// Dedicated servers may use this function to write statistics for multiple 
		/// users and/or teams to multiple leaderboards. If some of the userIDs / teamIDs
		/// provided are invalid (don't exist), entries for them will be ignored, but all
		/// other entries will be written as usual.
		///
		/// Ordinary users may use this function to write their own statistics, or statistics
		/// for teams they are members of. Attempts to provide userIDs of other users,
		/// wrong teamIDs, etc. will result in the entire write being denied.
		///
		/// This function produces a bdWriteMultiStatsResult. It contains a list of bdOnlineUserIDs
		/// whose stats were written, and a list of team IDs (bdEntityIDs) whose stats were written.
		/// This is provided for the benefit of dedicated servers, to identify any user IDs / team IDs
		/// in the submitted data that were reject by MM+. When used by an ordinary user this result
		/// should be ignored, as either all of data is accepted and written, or an error is returned
		/// and no statistics entries are changed.
		///
		/// \param stats[in]			Pointer to a vector of statistics submission objects.
		/// \param numStats[in]			Number of statistic submission objects in the vector
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success, accompanied by a bdWriteMultiStatsResult.
		///			BD_INVALID_STATS_SET - when some of the stats submissions were incorrectly
		///			initialized; when bdStatsInfo objects provided don't match the MM+ database
		///			configuration.
		///			BD_ACCESS_DENIED - when an ordinary user attempts to write for other users, or
		///			for teams that the user is not part of; when attempting to write to arbitrated
		///			leaderboards; when an ordinary user attempts to write to a dedicated leaderboard.
		/// \note	If \a stats array contains multiple entries for the same entityID on the same
		///			leaderboard, this function will run with an unpredictable result.
		bdRemoteTask* writeMultipleStats(bdStatsSubmission *stats,
										 const bdUInt numStats);

		/// Delete an entry from a leaderboard.
		/// This can be used to reset an entry, since the stats will be recreated
		/// the next time an entry is written to this leaderboard.
		/// \param controllerIndex[in]	The controller index of the requestor, whose
		///								stats will be written.
		/// \param leaderBoardID[in]	The ID of the leaderboard to delete stats from.
		/// \param teamID[in]			Optional team ID to delete the stats for.
		///								If specified (defaults to 0), stats for the
		///								local user specified by controllerIndex will
		///								NOT be deleted, but instead stats for the
		///								specified teamID will be deleted instead.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_INVALID_LEADERBOARD_ID - if leaderBoardID is not a valid leaderboard or not
		///			a valid teams leaderboard.
		///			BD_ACCESS_DENIED - if this user is not a member of a team specified by
		///			teamID or if teams is not enabled.
		bdRemoteTask* deleteStats(const bdUByte8 controllerIndex,
								  const bdLeaderBoardID leaderBoardID,
								  const bdTeamID teamID = 0);

		/// Read an entry from a specified leaderboard, based on user ID or team
		/// ID.
		/// \param leaderBoardID[in]	The ID of the leaderboard to read from.
		/// \param entityID[in]			The team/user ID to read stats for.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by
		///			invoking bdStats::handleTask(). An instance of a class
		///			derived from bdStatsQueryResult, which matches the results
		///			expected from the server, should be passed into it to receive
		///			the results.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_INVALID_LEADERBOARD_ID - if leaderBoardID is not a valid leaderboard.
		/// \note	This function call is deprecated. It is included for backwards 
		///			compatibility with version 1.76. It will be removed in the next release after
		/// 1.80
		/// \deprecated
		bdRemoteTask* readStats(const bdLeaderBoardID leaderBoardID,
								const bdEntityID entityID);

		/// Read entries from a specified leaderboard, based on an array of user
		/// IDs or teamIDs.
		/// \param leaderBoardID[in]	The ID of the leaderboard to read from.
		/// \param *entityID[in]		An array of team IDs or user IDs, whose
		///								stats will be read.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by
		///			invoking bdStats::handleTask(). An instance of a class
		///			derived from bdStatsQueryResult, which matches the results
		///			expected from the server, should be passed into it to receive
		///			the results.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_INVALID_LEADERBOARD_ID - if leaderBoardID is not a valid leaderboard.
		///			BD_ACCESS_DENIED - if this user is attempting is not a host.
		///			BD_PARAM_MISMATCHED_TYPE - if the the number of entityIDs serialized
		///			is not the same as numEntityIDs.
		bdRemoteTask* readStats(const bdLeaderBoardID leaderBoardID,
								const bdEntityID *entityID,
								const bdUInt numEntityIDs);

		/// Read entries from a specified leaderboard, based on ranked position.
		/// This function can be used in one of two ways:
		/// -	Read a specified number of entries, starting at a specified rank.
		/// -	Read a specified number of entries, using a specified user ID or
		///		team ID as a pivot. The operation will attempt to return entries
		///		ranked above and below the specified ID, with the specified ID
		///		at the center rank.
		/// \param leaderBoardID[in]	The ID of the leaderboard to read entries
		///								from.
		/// \param entityID [in]		EntityID of the user or team to use as a
		///								center-point for a range of stats to be returned.
		///								When this parameter is specified, the
		///								firstRank parameter is ignored.
		///								This parameter is optional and may be BD_NULL.
		/// \param firstRank [in]		The highest rank for which stats entries
		///								should be returned.
		///								A value of one specifies the top of the
		///								leaderboard.
		///								If the entityID parameter is specified,
		///								this parameter is ignored.
		/// \param maxResults [in]		The maximum number of stats entries to return.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by
		///			invoking bdStats::handleTask(). An instance of a class
		///			derived from bdStatsQueryResult, which matches the results
		///			expected from the server, should be passed into it to receive
		///			the results.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_INVALID_LEADERBOARD_ID - if leaderBoardID is not a valid leaderboard.
		bdRemoteTask* readLeaderboardStats(const bdLeaderBoardID leaderBoardID,
										   const bdEntityID entityID,
										   const bdUInt64 firstRank,
										   const bdUInt32 maxResults);

		/// Starts an arbitrated session on the server, one is allowed per bdSecurityID.
		/// This returns a nonce value which is valid for this arbitrated session.
		/// Can be re-called to re-retrieve arbitratedSessionNonce.
		/// \param matchMakingSecID[in]	The bdSecurityID obtained from MatchMaking service,
		///								upon session creation.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		/// \todo	Review arbitration commentary, ensure it is sync with the 
		///			documentation & server side code.
		bdRemoteTask* startArbitratedSession (const bdSecurityID matchMakingSecID);

		/// Writes a set of statistics for all users in a single arbitrated stats session.
		/// \param leaderBoardID[in]	The ID of the leaderboard to read entries from.
		/// \param secID[in]			The MatchMaking sessions bdSecurity,
		///								which was used to start the Arbitrated Stats session.
		/// \param nonce[in]			The nonce returned by startArbitratedSession() for the
		///								corresponding bdSecurity.
		/// \param writeType[in]		The type of stats update.
		///								Only the STAT_WRITE_ADD write type
		///								may be used with writeArbitratedStats.
		///								- STAT_WRITE_ADD: add new stats to
		///								  existing stat values.
		/// \param stats[in]			Array of all players stats data.
		///								Pointer to an array of bdStatsInfo*.
		///								Each bdStatsInfo* is an object describing
		///								the stats to write.
		///								This will normally be an instance of a derived
		///								class, which matches the custom attributes of
		///								entries on the specified leaderboard.
		///								This writes to a single leaderboard.  Leaderboards are either
		///								user leaderboards or team leaderboards.
		///								As such, this fn will only write user stats to a user leaderboards,
		///								and team stats to a team leaderboard.  Sending mixed stats sets will
		///								only result in the incorrect sets not being written.
		/// \param numStats[in]			The number of sets of stats data to be written.
		///								Effectively the maximum number of players in the Arbitrated
		///								Stats session.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_PARAM_PARSE_ERROR - if the writeType specified is not allowed.
		///			BD_INVALID_LEADERBOARD_ID - if the leaderboard specified does not exist.
		///			BD_INVALID_STATS_SET - if the parameters in stats do not match the
		///			schema of the specified leaderboard.
		bdRemoteTask* writeArbitratedStats (const bdUInt32 leaderBoardID,
											const bdSecurityID secID,
											const bdUInt64 nonce,
											const bdWriteType writeType,
											bdStatsInfo **stats,
											const bdUInt32 numStats);

		/// Handle a completed task.
		/// This retrieves the error code returned from the server, and fills out
		/// a result set of data, if appropriate. The specified task must have
		/// been created by this class.
		/// If the task is expected to return results, the second parameter should
		/// be specified. bdStatsQueryResult is the base-class for classes that handle
		/// statistics result sets returned from the server. Special implementations
		/// of bdStatsQueryResult must be used to handle results that contain
		/// customized leaderboard attributes.
		/// \param remoteTask[in]	A pointer to the remote task to handle.
		/// \param queryResult[in]	An optional pointer to an object that handles
		///							the expected type of result data.
		/// \return - True if the task was successfully handled.
		///			- False otherwise. Call bdStats::getErrorCode() to get the
		///			  error condition.
		bdBool handleTask(const bdRemoteTask *remoteTask,
						  bdQueryResult *queryResult = BD_NULL);

		/// Get the error code of a failed task.
		/// This returns an error code that can be matched against the bdLobbyErrorCode
		/// enum to determine the problem.
		/// \return The error code.
		bdLobbyErrorCode getErrorCode() const;

	protected:

		friend class bdLobbyServiceImpl;

		bdStats(bdRemoteTaskManager *const remoteTaskManager);

		/// remote task manager
		bdRemoteTaskManager *m_remoteTaskManager;

		/// error code store for look up
		bdLobbyErrorCode m_errorCode;
};

#endif // BD_STATS_H
