// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: Client-side interface to the storage service. 

#ifndef BD_STORAGE_MANAGEMENT_H
#define BD_STORAGE_MANAGEMENT_H

#include <bdCore/bdCore.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdLobbyErrorCode.h>
#include <bdLobby/bdLobbyCommon/bdQueryParams.h>
#include <bdLobby/bdLobbyCommon/bdQueryResult.h>

#include <bdLobby/bdStorage/bdLobbyFile.h>

class bdRemoteTask;
class bdRemoteTaskManager;

BD_REFERENCE(bdConnection);

// Storage Operations
#define BD_STORAGE_UPLOAD			(1u)
#define BD_STORAGE_UPDATE			(2u)
#define BD_STORAGE_UPDATE_PROPS		(3u)
#define BD_STORAGE_REMOVE			(4u)
#define BD_STORAGE_GETFILE			(5u)
#define BD_STORAGE_LISTBYOWNER		(7u)
#define BD_STORAGE_LISTALLPUBFILES	(8u)


/// Client-side interface to the storage service. Manages uploads and downloads
/// of small files.
/// The storage service enables persistent storage of small binary files. Storage
/// files can be either user-owned, in which case they are uploaded by clients 
/// with this interface, or they are publisher-owned, in which case they are 
/// uploaded in consultation with DemonWare support. A typical usage scenario 
/// is the implementation of roaming player profiles, by storing preferences in 
/// user-owned files. Another scenario is the implementation of "Message Of The
/// Day", by upload published-owned files.
/// Use the functions contained in this class to perform operations on the server.
/// Each of these function returns a task handle, which should be pumped to 
/// completion by calling its getStatus() function. When the status of a task is 
/// bdRemoteTask::BD_DONE, you can pass it into the bdStorage::handleTask()
/// function. Once this is done, the error code for the task can be accessed via
/// bdStorage::getErrorCode(). If you expect a result to be returned from
/// the server as a result of the task you started, you must pass a specialized 
/// instance of bdQueryResult into bdStorage::handleTask. The implementation must 
/// match the results you are expecting. Standard implementations are provided
/// for standard types of result sets. 
class bdStorage
{
	public:
		BD_DECLARE_NEW_AND_DELETE_OPERATORS		

		~bdStorage();

		/// Create a new file.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param fileName[in]			The name of the file to be stored.
		/// \param fileData[in]			A pointer to a bdLobbyFile object, which
		///								contains the actual file data.
		/// \param fileVisibility[in]	The visibility of the file. If set to
		///								bdLobbyFileHeader::BD_VISIBLE_PRIVATE, only
		///								the file owner will be able to access the file.
		/// \param ownerType[in]		The owner type of this file. This should 
		///								normally be set to bdLobbyFileHeader::BD_OWNER_USER.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdStorage::handleTask() and passing in an instance
		///			of bdUploadFileResult, which will then contain the file ID
		///			that was assigned by the server.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask *uploadFile(bdUByte8 controllerIndex, 
								 const bdNChar8 *fileName, 
								 const bdLobbyFile* fileData,
								 bdLobbyFileHeader::bdVisibility fileVisibility, 
								 bdLobbyFileHeader::bdOwnerType ownerType);

		/// Update an existing file's data.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param fileID [in]			The unique file ID of the file to be updated.
		/// \param fileData[in]			A pointer to a file object containing the
		///								new file data.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_NO_FILE - if the fileID specified does not represent a file.
		///			BD_PERMISSION_DENIED - if this user does not own the file specified.
		bdRemoteTask* updateFile(bdUByte8 controllerIndex, 
								 bdUInt64 fileID, 
								 bdLobbyFile* fileData);

		/// Update an existing file's properties.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param fileID[in]			The unique file ID of the file to be updated.
		/// \param fileName[in]			The name of the file to be stored.
		/// \param fileVisibility[in]	The visibility of the file. If set to
		///								bdLobbyFileHeader::BD_VISIBLE_PRIVATE, only
		///								the file owner will be able to access the file.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_NO_FILE - if the fileID specified does not represent a file.
		///			BD_PERMISSION_DENIED - if this user does not own the file specified.
		bdRemoteTask* updateFileProperties(bdUByte8 controllerIndex,
										   bdUInt64 fileID,
                                           const bdNChar8 *fileName,
                                           bdLobbyFileHeader::bdVisibility fileVisibility);

		/// Remove a file from storage.
		/// \param controllerIndex[in]	The controller index of the requestor,
		///								who must be the owner of the file.
		/// \param fileID[in]			The ID of the file to be removed.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_NO_FILE - if the fileID specified does not represent a file.
		///			BD_PERMISSION_DENIED - if this user does not own the file specified.
		bdRemoteTask *removeFile(bdUByte8 controllerIndex,
								 bdUInt64 fileID);

		/// Get a file from storage.
		/// \param controllerIndex[in]	The controller index of the requestor.
		/// \param fileID[in]			The ID of the file to be retrieved.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdStorage::handleTask() and passing in an instance
		///			of bdGetFileResult, which will then contain a bdLobbyFile
		///			object for the retrieved file.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_NO_FILE - if the fileID specified does not represent a file.
		///			BD_PERMISSION_DENIED - if this user does not have permission to 
		///			read the specified file.
		bdRemoteTask *getFile(bdUByte8 controllerIndex,
							  bdUInt64 fileID);	
		

		/// Get a list of files owned by a specified user.
		/// \param controllerIndex[in]	The controller index of the requestor.
		/// \param ownerID[in]			The ID of the user whose files should be
		///								returned.
		/// \param fileName[in]			An optional string with the name of the desired
		///								file. Set to \a BD_NULL for any name. Note that
		///								a user can have multiple files named with the same
		///								name and all of those will be returned. Match is
		///								exact (case-sensitive).
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdStorage::handleTask() and passing in an instance
		///			of bdListFileResult, which will then contain a bdLobbyFile
		///			object for the retrieved file.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask *listFilesByOwner(bdUByte8 controllerIndex, 
									   bdOnlineUserID ownerID,
									   bdUInt32 startDate,
									   bdUInt16 maxNumResults,
									   const bdNChar8 *fileName = BD_NULL);
		
		/// Get a list of all files owned by the publisher (i.e. non-user files).
		/// \param controllerIndex[in]	The controller index of the requestor.
		/// \param startDate[in]		A timestamp specifying the oldest file
		///								that should be returned. This is in the
		///								form of seconds since midnight Jan 1 1970, UTC.
		/// \param fileName[in]			An optional string with the name of the desired
		///								file. Set to \a BD_NULL for any name. Note that
		///								there can be multiple files named with the same
		///								name and all of those will be returned. Match is
		///								exact (case-sensitive).
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdStorage::handleTask() and passing in an instance
		///			of bdListFileResult, which will then contain a bdLobbyFile
		///			object for the retrieved file.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask *listAllPublisherFiles(bdUByte8 controllerIndex,
											bdUInt32 startDate,
											bdUInt16 maxNumResults,
											const bdNChar8 *fileName = BD_NULL);

		/// Handle a completed task. 
		/// This retrieves the error code returned from the server, and fills out
		/// a result set of data, if appropriate. The specified task must have 
		/// been created by this class. 
		/// If the task is expected to return results, the second parameter should
		/// be specified. bdQueryResult is the base-class for classes that handle
		/// result sets returned from the server. Specialized implementations
		/// of bdQueryResult exist in the same directory as this class for each
		/// kind of storage task that can be started.
		/// \param remoteTask[in]	A pointer to the remote task to handle.
		/// \param queryResult[in]	An optional pointer to an object that handles 
		///							the expected type of result data.
		/// \return - True if the task was successfully handled.
		///			- False otherwise. Call bdStorage::getErrorCode() to get the 
		///			  error condition.
		bdBool handleTask(const bdRemoteTask *remoteTask, 
						  bdQueryResult *queryResult = BD_NULL);

		/// Return the error code of the last operation that failed.
		/// \return The last error code
		bdLobbyErrorCode getErrorCode() const;

	protected:

		friend class bdLobbyServiceImpl;

		/// Constructs a storage manager with the given task handler.
		/// Note: never call this constructor directly, use bdLobbyService::getStorage() instead.
		/// \param remoteTaskManager[in] A pointer to the remote task manager that handles
		/// requests and replies
		bdStorage(bdRemoteTaskManager *const remoteTaskManager);

		bdLobbyErrorCode m_errorCode;

		bdRemoteTaskManager *m_remoteTaskManager;

};

#endif // BD_STORAGE_MANAGEMENT_H
