// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: The gateway to our network engine.

#ifndef BD_NET_H
#define BD_NET_H

#include <bdCore/bdCore.h>
#include <bdCore/bdCrypto/bdCryptoConfig.h>
#include <bdCore/bdSocket/bdInetAddr.h>
#include <bdCore/bdSocket/bdSecurityKeyMap.h>


#include <bdNet/bdDispatch/bdDispatcher.h>

#if (!defined(BD_PLATFORM_XBOX)) && (!defined(BD_PLATFORM_XENON)) && (!defined(BD_PLATFORM_PSP_ADHOC))
#include <bdNet/bdUPnP.h>
#endif

class bdConnectionListener;
class bdIPDiscoveryClient;
class bdNATTypeDiscoveryClient;
class bdGetHostByName;
class bdConnectionStore;
class bdSocketRouter;

#if !defined BD_PLATFORM_XBOX
# define BD_DEFAULT_PORT (3074)
#else
# define BD_DEFAULT_PORT (1000)
#endif

class bdNetStartParams
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		inline bdNetStartParams()
		: m_onlineGame(false),
		  m_gamePort(BD_DEFAULT_PORT),
		  m_socket(BD_NULL),
		  m_upnpTimeout(3)
		{
			m_natTravHosts.pushBack("stun.eu.demonware.net");
			m_natTravHosts.pushBack("stun.us.demonware.net");
			m_natTravHosts.pushBack("stun.jp.demonware.net");
			m_natTravHosts.pushBack("stun.au.demonware.net");
			m_natTravPort = 3074;
			m_timeout = 10;
		}

		/// Whether or not start an online game. An online game needs
		/// to contact the NAT server to find its public IP address
		/// and NAT Type.
		bdBool m_onlineGame;

		/// The port the engine listens on.
		bdPort m_gamePort;

		/// The socket the engine uses to send and receive data.
		/// \note: This socket will be deleted when bdNet::stop is called.
		bdDefaultSocket *m_socket;

		/// The address of the NAT traversal server.
		bdArray<bdString> m_natTravHosts;

		/// The port for the NAT traversal server.
		bdPort m_natTravPort;

		/// An array of IP addresses that this machine listens on.
		/// If more than one IP address is given the m_socket will be bound
		/// to address "0.0.0.0" but if there is only 1 IP address the socket
		/// will be bound to that address. This facilitates running several
		/// different instances on the same port but different addresses which
		/// is often needed for dedicated servers.
		bdArray<bdInetAddr> m_localAddresses;

		/// The amount of time in seconds to wait for a DNS lookup (gethostbyname)
		bdFloat32 m_timeout;

		/// Timeout for discovery of UPnP NATs. Set this to 0 for no UPnP traversal
		bdUInt m_upnpTimeout;
};

/// The gateway to our network engine.
/// This is the central access point for our network engine.
/// It must be started before any network messages can be sent/received.
class bdNetImpl
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		enum bdNetStatus
		{
			BD_NET_BIND_FAILED = -3,
			BD_NET_ONLINE_FAILED = -2,
			BD_NET_INIT_FAILED = -1,
			BD_NET_STOPPED,
			BD_NET_PENDING,
			BD_NET_DONE
		};

		/// Start the net engine. This will perform NAT traversal and UPnP discovery
		/// if the m_online flag in bdNetStartParams has been set to true.
		/// pump() needs to be called until completion.
		/// \param[in] params : A structure containing various startup parameters
		/// \note If the m_socket member of this class is set you need to ensure
		///       that create() has been called on it before calling this function.
		///       If using the default socket create() will be called internally in
		///       this function.
		bdBool start(const bdNetStartParams& params = bdNetStartParams());

		/// Pumps the startup process of the network subsystem.
		void pump();

		/// Stop the net engine.
		/// Closes all open connections.
		/// \note This call will take a relatively long time (~1000ms) to
		/// complete as it does a controlled shutdown of the connections.
		void stop();

		/// Gets the current state of the net engine
		/// \return The current state
		bdNetStatus getStatus() const;

		/// Send as many of the queued packets as possible.
		/// This function will loop through all the live connections
		/// and try to send any queued packets.
		void sendAll();

		/// Receive as many packets as possible.
		/// This will drive bdConnectionManager which tries to
		/// receive data from the listening socket and send it to the
		/// relevant connection.
		/// Then takes all messages from the connections and tries to dispatch them
		/// to their relevant callback objects or registered interceptors
		void receiveAndDispatchAll();

		/// Registers an interceptor with the dispatcher.
		/// \param interceptor[in]	The interceptor to be registered with this
		///							dispatcher.
		void registerDispatchInterceptor(bdDispatchInterceptor *const interceptor);

		/// Removes an existing interceptor from the dispatcher.
		/// \param interceptor[in]	The interceptor to be removed from this
		///							dispatcher.
		void unregisterDispatchInterceptor(bdDispatchInterceptor *const interceptor);

		/// Gets a pointer to the internal connection store.
		/// Applications can register connection listeners with the connection store
		/// in order to be notified about connection related events.
		/// \return A pointer to the connection store.
		bdConnectionStore *getConnectionStore();

		/// Gets a pointer to the internal security key map. 
		/// \return A pointer to the internal security key map. 
		bdSecurityKeyMap *getKeyMap();

		/// Retrieves the local common address used for this machine.
		/// A common address is used to uniquely identify a machine on the network.
		/// This function has to be called after bdNet::start();
		/// \return A reference to the local common address.
		bdCommonAddrRef getLocalCommonAddr() const;

		/// Gets a reference to the startup parameters that where provided.
		/// \return A reference to the startup parameters
		const bdNetStartParams& getParams();

		/// Gets a pointer to the internal socket router.
		/// \return A pointer to the internal socket router.
		bdSocketRouter* getSocketRouter();


	protected:

		/// Grant friendship to bdSingleton<bdNetImpl> so that it can
		/// construct the singleton instance of this class.
		friend class bdSingleton<bdNetImpl>;

		/// Default constructor.
		/// This is protected so that this class cannot be instantiated, except
		/// as a singleton.
		bdNetImpl();

		/// Virtual destructor.
		virtual ~bdNetImpl();

		bdBool findFreePort(bdAddr& addr);

		bdAddr getBindAddr();

	protected:

		/// Stores the startup parameters that were used
		bdNetStartParams m_params;

		bdCommonAddrRef m_localCommonAddr;

		bdECCKey *m_ECCKey;

		/// A store of security key/id pairs
		bdSecurityKeyMap m_keyStore;

		/// A class to handle input and output of data to and from connections.
		bdConnectionStore *m_connectionStore;

		bdSocketRouter *m_socketRouter;

		/// A class to dispatch packets to relevant dispatch interceptors.
		bdDispatcher m_dispatcher;

		/// Internal status
		bdNetStatus m_status;

		bdIPDiscoveryClient *m_ipDiscClient;

		bdNATTypeDiscoveryClient *m_natTypeDiscClient;

		bdGetHostByName *m_getHostByName;

		bdArray<bdAddr> m_natTravAddrs;

		bdUInt m_currentNatTravAddrIndex;

		bdUInt m_currentNatTravHostIndex;

#if (!defined(BD_PLATFORM_XBOX)) && (!defined(BD_PLATFORM_XENON)) && (!defined(BD_PLATFORM_PSP_ADHOC))
		bdUPnP* m_UPnP;
		bdBool m_UPnPFinished;
		bdUPnP::bdUPnPState m_UPnPState;
#endif

#if (!defined(BD_PLATFORM_XBOX)) && (!defined(BD_PLATFORM_XENON)) && (!defined(BD_PLATFORM_PSP_ADHOC))
		bdDefaultSocket m_tmpSocket;
#endif
};

/// This class provides a point of access to the singleton instance.
/// Implementation details can be found in bdNetImpl.
class bdNet : public bdSingleton<bdNetImpl>
{
	BD_DECLARE_NEW_AND_DELETE_OPERATORS

	protected:
		/// Default constructor.
		bdNet();
};

#endif // BD_NET_H

