// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: An abstract class that represents an outgoing network stream.


#ifndef BD_OUT_STREAM_H
#define BD_OUT_STREAM_H

#include <bdCore/bdCore.h>

#include <bdConnection/bdConnection.h>

#include <bdNet/bdStream/bdStreamTypes.h>

BD_REFERENCE(bdOutStream);

/// Abstract class that represents an outgoing network stream.
class bdOutStream : public bdReferencable
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		/// Constructor that specifies the type, destination, reliability setting and whether
		/// or not the data should be encrypted.
		/// \param connection[in] The destination of the stream.
		/// \param type[in] The stream type.
		/// \param reliable[in] Whether or not the stream should be reliable.
		/// \param priority[in] The priority relative to other streams.
		/// The higher the priority the more important.
		bdOutStream(const bdConnectionRef connection,
					const bdStreamType type,
					const bdBool reliable = false,
					const bdUByte8 priority = 0);

		/// Virtual destructor.
		virtual ~bdOutStream();

		/// Pure virtual function that indicates if the stream has data available.
		/// \return true if data available.
		virtual bdBool hasData() = 0;

		/// Pure virtual function that will write at most \c maxWrite bytes to the \c buffer and \c unencBuffer.
		/// This function is called by the stream store to send data when hasData() has returned true.
		/// \param buffer[out] The buffer to write the data that should be encrypted to.
		/// \param unencBuffer[out] The buffer to write the data that should be \b not encrypted to.
		/// \param maxNumBytes[in] The maximum number of bytes to write to the buffers.
		/// \return
		///		- True, if data is written to the buffer.
		/// 	- False, otherwise.
		virtual bdBool getData(bdBitBufferRef buffer,
							   bdBitBufferRef unencBuffer,
							   const bdUInt maxNumBytes) = 0;

		/// Returns the stream's connection.
		/// \return The connection.
		bdConnectionRef getConnection() const;

		/// Returns the stream's type.
		/// \return The type of the stream.
		bdStreamType getType() const;

		/// Returns whether or not the stream should be reliable.
		bdBool getReliable() const;

		/// Every stream has a locally unique ID. The ID is unique to the host
		/// the stream was originally created upon. Combined with the common addr
		/// of the host it represents a globally unique stream identifier.
		bdStreamNum getNum() const;

		/// Returns the priority of the stream. The priority is used to calculate which
		/// streams get preference when bandwidth is low. Streams with higher priority
		/// will get to use available bandwidth first.
		/// \return The stream priority.
		bdUByte8 getPriority() const;

		/// Closes the stream. Closed streams are considered to have been stopped before sending
		/// all their data.
		void close();

		/// Sets the stream as finished. This indicates all data has been sent and the stream
		/// can be shutdown.
		void finished();

		/// Returns true if the stream is finished.
		bdBool isFinished() const;

		/// Returns true if the stream is closed (but not finished).
		bdBool isClosed() const;

	protected:

		bdConnectionRef m_connection;

		bdStreamType m_type;

		bdBool m_reliable;

		bdStreamNum m_num;

		bdUByte8 m_priority;

		bdBool m_finished;

		bdBool m_closed;

		static bdStreamNum m_staticNum;
};

#endif // BD_OUT_STREAM_H



