// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#ifndef BD_STREAM_STORE_H
#define BD_STREAM_STORE_H

#include <bdCore/bdCore.h>

#include <bdConnection/bdConnectionListener.h>

#include <bdNet/bdBandwidthMonitor/bdBandwidthListener.h>
#include <bdNet/bdDispatch/bdDispatchInterceptor.h>
//#include <bdNet/bdStream/bdStream.h>
//#include <bdNet/bdStream/bdOutStream.h>
//#include <bdNet/bdStream/bdInStream.h>
#include <bdNet/bdStream/bdStreamId.h>

class bdStreamListener;
class bdDispatcher;

BD_REFERENCE(bdOutStream);
BD_REFERENCE(bdInStream);


/// This class manages all aspects of network streams. It creates remote
/// streams for locally added ones, sends data available on local streams
/// to their remote counterparts, receives data for incoming streams and
/// informs listeners of changes of incoming stream states.
/// This class is also bandwidth aware, meaning that it will suspend all sending
/// if the bandwidth monitor indicates we have reached the high watermark and the
/// rest of the time it will only send as much data as each connection has available
/// to it.
class bdStreamStore : public bdDispatchInterceptor,
					  public bdBandwidthListener,
					  public bdConnectionListener
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		/// Constructor
		bdStreamStore(bdDispatcher *dispatcher);

		/// Virtual destructor.
		virtual ~bdStreamStore();

		/// Add a locally created stream.
		/// \param stream[in] The stream to add.
		bdBool registerOutStream(const bdOutStreamRef stream);

		bdBool closeInStream(const bdInStreamRef stream);

		/// Add an event listener. See bdStreamListener for more information.
		/// \param listener[in] The listener to add.
		void registerListener(bdStreamListener *const listener);

		/// Remove an event listener.
		/// \param listener[in] The listener to remove.
		void unregisterListener(bdStreamListener *const listener);

		/// Needs to be called for stream store to do any work. It should be called
		/// at a high frequency (once per frame) and at least as high a frequency as
		/// any stream added through addStream() needs.
		void doWork();

	protected:

		virtual bdBool accept(const bdReceivedMessage& receivedMsg);

		void handleStreamCreate(const bdReceivedMessage& receivedMsg);

		void handleOutStreamClose(const bdReceivedMessage& receivedMsg);

		void handleInStreamFinished(const bdReceivedMessage& receivedMsg);

		void handleInStreamClose(const bdReceivedMessage& receivedMsg);

		void handleStreamData(const bdReceivedMessage& receivedMsg);

		bdBool sendCreateInStream(const bdOutStreamRef outStream);

		bdBool sendFinishedInStream(const bdOutStreamRef outStream);

		bdBool sendCloseInStream(const bdOutStreamRef outStream);

		bdBool sendCloseOutStream(const bdStreamType streamNum,
								  const bdConnectionRef connection);

		void maintainListener(const bdConnectionRef connection,
							  const bdBool registerListener);

		void sortOutStreams();

		virtual void onDisconnect(bdConnectionRef connection);

		virtual void onSendHighWaterMark(bdUInt32 bytesSent);

		virtual void onSendLowWaterMark(bdUInt32 bytesSent);

		class bdOutStreamRefLessThan
		{
			public:

				bdBool operator() (const bdOutStreamRef& x,
								   const bdOutStreamRef& y) const;

		};

	protected:

		bdFastArray<bdStreamListener*> m_listeners;

		bdArray<bdOutStreamRef> m_outStreams;

		bdHashMap<bdStreamId, bdInStreamRef, bdStreamIdHash> m_inStreams;

		bdBool m_okToSend;

		bdDispatcher *m_dispatcher;
};

#endif // BD_STREAM_STORE_H

