// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 PS2 UNIX OSX PSP-INFRA PS3
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#ifndef BD_NAT_TRAV_CLIENT_H
#define BD_NAT_TRAV_CLIENT_H

#include <bdCore/bdCore.h>

#include <bdSocket/bdPacketInterceptor.h>
#include <bdSocket/bdNAT/bdNAT.h>

class bdNATTravListener;
class bdNATTraversalPacket;
class bdServiceBandwidthArbitrator;

// Nat traversal works according to the following table.
// F - full cone
// M - Moderate
// S - Strict
// X - Can Connect
// 0 - Can't connect
//
//     F    M     S
//    ----------------
// F | X |  X  |  X  |
//   |---------------|
// M | X |  X  |  0  |
//   |---------------|
// S | X |  0  |  0  |
//   |---------------|

/// NAT Traversal client code.
///
/// This class attempts to establish a communication channel with a remote 
/// host, possibly traversing multiple NAT devices. It is aware of possibility 
/// that the remote host is on the same LAN as the local host, behind one NAT.
///
/// This is meant to run on clients, driven by \a bdSocketRouter.
class bdNATTravClient : public bdPacketInterceptor
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS;

		/// Defines the possible states of a bdNATTravClient.
		enum bdNATTravClientStatus
		{
			BD_NAT_TRAV_UNINITIALIZED,
			BD_NAT_TRAV_INITIALIZED,
			BD_NAT_TRAV_RUNNING
		};

		/// Basic constructor.
		/// When instantiated using this constructor, init() must be called 
		/// before this class can be used.
		bdNATTravClient();

		/// Constructor.
		/// \param introducers[in] An array of introducers to send keep alives 
		///						   to. The first element in the array points to
		///						   the introducer we will use for introduction
		///						   requests. i.e. The first element must point 
		///						   to a working introducer.
		/// \param socket[in] The socket used for sending packets.
		/// \param bandArb[in] A pointer to an instance of bandwidth 
		///					   arbitrator, shared between connection services 
		///					   (such as QoS). Pointer must be valid for the
		///					   entire lifetime of this class instance.
		bdNATTravClient(const bdArray<bdAddr> &introducers,
						bdDefaultSocket *socket,
						bdServiceBandwidthArbitrator *bandArb);

		/// Default destructor.
		virtual ~bdNATTravClient();

		/// Initialize this class.
		/// This function must be called if the basic constructor was used to 
		/// create the class.
		/// \param introducers[in] An array of introducers to send keep alives 
		///						   to. The first element in the array points to
		///						   the introducer we will use for introduction
		///						   requests. i.e. The first element must point 
		///						   to a working introducer.
		/// \param socket[in] The socket used for sending packets.
		/// \param bandArb[in] A pointer to an instance of bandwidth 
		///					   arbitrator, shared between connection services 
		///					   (such as QoS). Pointer must be valid for the
		///					   entire lifetime of this class instance.
		/// \return
		///   - True if the class was initialized correctly.
		///   - False otherwise.
		bdBool init(const bdArray<bdAddr> &introducers,
					bdDefaultSocket *socket,
					bdServiceBandwidthArbitrator *bandArb);

		/// Perform traversal to a remote host.
		/// \param remote[in] Remote host traversal should try to connect.
		/// \param listener[in] Traversal result listener.
		/// \param throttle[in] Set to true to use bandwidth limiting on this 
		///						connection (through \a bdServiceBandwidthArbitrator).
		/// \return
		///   - True if the connect attempt was successfully started.
		///   - False otherwise.
		bdBool connect(bdCommonAddrRef remote, 
					   bdNATTravListener *listener,
					   bdBool throttle = false);

		/// Performs regular task. Should be called every frame.
		void pump();

		/// Stops all activity in this class.
		/// This will reset the class to the uninitialized state.
		/// After this function has been called, it is safe to call init again.
		/// \return 
		///   - True if the quit was error free.
		///   - False otherwise.
		bdBool quit();

		/// Derived from \a bdPacketInterceptor.
		virtual bdBool acceptPacket(bdDefaultSocket *socket, bdAddr addr, void *data, const bdUInt size, const bdUByte8 type);

		/// Abandons traversal to a particular remote host.
		/// \param remote[in] Address of the remote host we no longer want to connect to.
		void cancelConnect(bdCommonAddrRef remote);

		/// Get the status of this class.
		/// \return The status of the class.
		bdNATTravClientStatus getStatus();

		/// Set the introducers that this class should use.
		/// Invoking this while the NAT Trav client is running may 
		/// result in unusual behaviour. 
		/// It is intended for use when switching from LAN
		/// to online mode.
		/// \param introducers[in] The new array of introducers.
		/// \return 
		///  - True if the introducers were successfully set.
		///  - False otherwise.
		bdBool setupIntroducers(const bdArray<bdAddr> &introducers);

	protected:

		/// Handle incoming packet.
		/// \param addr[in] Address the packet came from.
		/// \param packet[in] Parsed packet.
		void receiveFrom(const bdAddr &addr, bdNATTraversalPacket &packet);

		/// Data structure to hold information about pending traversal request.
		class bdNATTravClientData
		{
			public:
				/// Traversal goes through several stages.
				enum bdNATTravClientDataState
				{
					BD_NTCDS_INITIAL,
					BD_NTCDS_STAGE_1,
					BD_NTCDS_STAGE_2,
					BD_NTCDS_STAGE_3
				};

				/// Default constructor.
				bdNATTravClientData();

				/// Constructor.
				/// \param local[in] Common address of the host.
				/// \param remote[in] Common address of the host we're trying to connect to.
				/// \param listener[in] Listener to notify.
				bdNATTravClientData(bdCommonAddrRef local,
									bdCommonAddrRef remote, 
									bdNATTravListener* listener);

				/// Notify listeners about successful discovery.
				/// \param remote[in] Remote address of the discovery.
				/// \param realAddr[in] Discovered address to be used for sending data
				///  to the remote peer.
				void callOnNATAddrDiscovery(bdCommonAddrRef remote, const bdAddr &realAddr);

				/// Notify listeners about failed discovery.
				/// \param remote[in] Common address of the host traversal was attempted to
				void callOnNATAddrDiscoveryFailed(bdCommonAddrRef remote);

				/// Current state.
				bdNATTravClientDataState m_state;
				/// Our local CA.
				bdCommonAddrRef m_local;
				/// CA of the remote host we're connecting to.
				bdCommonAddrRef m_remote;

				/// Primary listener for NAT discovery event.
				bdNATTravListener *m_listener;
				/// Second listener for NAT discovery event, if any.
				bdNATTravListener *m_secondaryListener;
				/// Number of times we attempted to send requests so far.
				bdUInt m_tries;
				/// Time since we sent a request.
				bdShortTimer m_lastSent;
				/// True, if this traversal request can be throttled by bandwidth arbitrator.
				bdBool m_throttled;
				/// Total age of the request. Used to purge starved requests.
				bdStopwatch m_age;

		};

		/// Send keep-alives to all introducers.
		/// \return True, if successful.
		bdBool sendKeepAlive();

		/// Send stage 1 traversal packet.
		/// \param data[in] Request we're sending it for.
		/// \return True, if successful.
		bdBool sendStage1(bdNATTravClientData &data);

		/// Send stage 2 traversal packet.
		/// \param data[in] Request we're sending it for.
		/// \return True, if successful.
		bdBool sendStage2(bdNATTravClientData &data);

		/// Send stage 3 traversal packet.
		/// \param data[in] Request we're sending it for.
		/// \return True, if successful.
		bdBool sendStage3(bdNATTravClientData &data);

		/// Compute an HMAC signature based on an identifier and a pair of addresses.
		/// \param identifier[in] The identifier.
		/// \param src[in] First address.
		/// \param dest[in] Second address.
		/// \param data[in] Buffer to receive an HMAC signature based on the identifier
		///  and a pair of addresses.
		/// \return true on success, false otherwise
		bdBool doHMac(const bdUInt32 identifier, 
					  const bdAddr &src, 
					  const bdAddr &dest,
					  bdUByte8 data[BD_NAT_TRAV_HMAC_SIZE]) const;

	protected:

		/// Our bandwidth arbitrator.
		bdServiceBandwidthArbitrator *m_bandArb;

		/// All known introducers.
		bdArray<bdAddr> m_introducers;

		/// Socket for outgoing data.
		bdDefaultSocket *m_socket;

		// Map of the address to where we are sending the packets to
		// data required to callback.
		bdHashMap<bdUInt32, bdNATTravClientData> m_callbacks;

		/// Timer for sending keep alives to introducers.
		bdStopwatch m_keepAliveTimer;

		/// Our random secret key for signing requests.
		bdUByte8 m_secretKey[BD_ECC_SHARED_SECRET_SIZE];

		/// The status of this Nat traversal client.
		bdNATTravClientStatus m_status;
};

#endif // BD_NAT_TRAV_CLIENT_H
