// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 PS2 UNIX OSX PSP-INFRA PS3
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdSocket/bdSocketRouter.h>

#include <bdCore/bdSocket/bdCommonAddrInfo.h>

#include <bdSocket/bdAddressMap.h>
#include <bdSocket/bdDTLS/bdDTLSAssociation.h>

#define BD_LOG_LEVEL "bdSocket/socket"

bdSocketRouter::bdSocketRouter()
: m_socket(BD_NULL),
  m_keyMap(BD_NULL),
  m_ECCKey(BD_NULL),
  m_status(BD_SOCKET_ROUTER_UNINITIALIZED)
{
}

bdSocketRouter::bdSocketRouter(bdDefaultSocket *socket, 
							   bdSecurityKeyMap *keyMap, 
							   bdECCKey *key, 
							   const bdArray<bdAddr> &introducers)
: m_socket(socket),
  m_natTrav(introducers, socket, &m_qosBandwidth),
  m_keyMap(keyMap),
  m_ECCKey(key),
  m_qosProber(socket, &m_natTrav, &m_qosBandwidth),
  m_status(BD_SOCKET_ROUTER_INITIALIZED)
{
	registerInterceptor(&m_natTrav);
	registerInterceptor(&m_qosProber);
	m_keyMap->registerListener(this);
}

bdSocketRouter::~bdSocketRouter()
{
	if(m_keyMap)
	{
		m_keyMap->unregisterListener();
	}

	bdHashMap<bdAddr, bdDTLSAssociation *, bdAddrHash>::Iterator it = m_dtls.getIterator();
	while(it)
	{
		delete m_dtls.getValue(it);
		m_dtls.next(it);
	}
	
	m_socket = BD_NULL;
	m_keyMap = BD_NULL;
	m_ECCKey = BD_NULL;
}

bdBool bdSocketRouter::init(	bdDefaultSocket *socket, 
								bdSecurityKeyMap *keyMap, 
								bdECCKey *key, 
								const bdArray<bdAddr> &introducers)

{
	bdBool result = true;

	if(BD_SOCKET_ROUTER_UNINITIALIZED != m_status)
	{
		result = false;
		BD_WARN(BD_LOG_LEVEL, "init() called multiple times.");
	}
	else if(BD_NULL == socket)
	{
		result = false;
		BD_ERR(BD_LOG_LEVEL, "A valid socket pointer is required by this class");
	}
	else if (BD_NULL == keyMap)
	{
		result = false;
		BD_ERR(BD_LOG_LEVEL, "A valid key map pointer is required by this class");
	}
	else if (BD_NULL == key)
	{
		result = false;
		BD_ERR(BD_LOG_LEVEL, "A valid key pointer is required by this class");
	}
	else
	{
		m_socket = socket;
		m_keyMap = keyMap;
		m_ECCKey = key;

		result = m_natTrav.init(introducers, socket, &m_qosBandwidth);
		result = result && m_qosProber.init(socket, &m_natTrav, &m_qosBandwidth);

		m_status = BD_SOCKET_ROUTER_INITIALIZED;

		registerInterceptor(&m_natTrav);
		registerInterceptor(&m_qosProber);

		m_keyMap->registerListener(this);
	}

	return result;
}

bdBool bdSocketRouter::quit()
{
	bdBool result = false;

	if(BD_SOCKET_ROUTER_UNINITIALIZED == m_status)
	{
		result = false;
		BD_WARN(BD_LOG_LEVEL, "quit() called multiple times.");
	}
	else
	{
		// stop listening for key map events
		if(m_keyMap)
		{
			m_keyMap->unregisterListener();
		}

		// Delete all active DTLS associations
		bdHashMap<bdAddr, bdDTLSAssociation *, bdAddrHash>::Iterator it = m_dtls.getIterator();
		while(it)
		{
			delete m_dtls.getValue(it);
			m_dtls.next(it);
		}
		
		// empty the dtls list
		m_dtls.clear();

		// clear pointers
		m_socket = BD_NULL;
		m_keyMap = BD_NULL;
		m_ECCKey = BD_NULL;

		// call quit on helper classes
		result = m_natTrav.quit();
		result = m_qosProber.quit() && result;

		// unregister interceptors
		unregisterInterceptor(&m_natTrav);
		unregisterInterceptor(&m_qosProber);

		// clear any remaining interceptors
		m_interceptors.clear();

		// clear the common addr -> addr handle lookup table
		m_caToAddrMap.clear();

		m_status = BD_SOCKET_ROUTER_UNINITIALIZED;
	}

	return result;
}

bdBool bdSocketRouter::connect(bdAddrHandleRef &addrHandle)
{
	bdBool success = false;
	if(BD_SOCKET_ROUTER_INITIALIZED != m_status)
	{
		success = false;
		BD_WARN(BD_LOG_LEVEL, "Cannot connect before class has been initialized");
	}
	else if(addrHandle.isNull())
	{
		success = false;
		BD_WARN(BD_LOG_LEVEL, "Cannot connect using null addr handle ref");
	}
	else
	{
		bdCommonAddrRef remote = addrHandle->m_commonAddr;
		BD_INFO(BD_LOG_LEVEL, "	Connecting to address.");
		bdNChar8 commonAddrInfo[BD_COMMON_ADDR_INFO_SIZE];
		bdCommonAddrInfo::getInfo(remote, commonAddrInfo, BD_COMMON_ADDR_INFO_SIZE);
		BD_INFO(BD_LOG_LEVEL, commonAddrInfo);

		const bdAddrHandle::bdAddrHandleStatus status = addrHandle->getStatus();
		switch(status)
		{
			case bdAddrHandle::BD_ADDR_NOT_CONSTRUCTED:
			case bdAddrHandle::BD_ADDR_NOT_RESOLVED:
			case bdAddrHandle::BD_ADDR_ERROR:
			case bdAddrHandle::BD_ADDR_UNRESOLVED:
			{
				// different debug output, based on addr handle status
				if(bdAddrHandle::BD_ADDR_ERROR == status)
				{
					BD_INFO(BD_LOG_LEVEL, "Address resolution failed. Trying again.");
				}
				else if (bdAddrHandle::BD_ADDR_UNRESOLVED  == status)
				{
					BD_INFO(BD_LOG_LEVEL, "Address has been unregistered. Starting again.");
				}
				else
				{
					BD_INFO(BD_LOG_LEVEL, "Address not resolved. Sending to NAT Trav.");
				}

				const bdBool added = m_caToAddrMap.put(remote, addrHandle);
				if(false == added)
				{
					BD_WARN(BD_LOG_LEVEL, "Remote addr already exists in map");
					BD_INFO(BD_LOG_LEVEL, "Sending to NAT Trav anyway....");
					//////////////////////////////////////////////////////////////////////////
					// Fixme - Review me - Todo
					// This code is intended not to alter the behaviour of this code.
					// todo: review whether we need this. it would seem not to make sense...
					//       as the addr handle that is registered differs from the one passed 
					//       to this function.
					//
					// Perhaps we need to remove the addr handle first or somesuch.
					//////////////////////////////////////////////////////////////////////////
				}
				// consider only connecting when addr is added successfully.
				// cf comment above.
				success = m_natTrav.connect(remote, this);
			}
			break;
			case bdAddrHandle::BD_ADDR_RESOLVED:
			{
				BD_INFO(BD_LOG_LEVEL, "Address already resolved. Nothing to do.");
			}
			break;
		}
	}

	return success;
}


bdSocketStatus bdSocketRouter::getStatus(const bdAddrHandleRef addrHandle) const
{
	bdSocketStatus status = BD_SOCKET_IDLE;

	if(addrHandle.isNull())
	{
		status = BD_SOCKET_LOST;
		BD_WARN(BD_LOG_LEVEL, "Cannot get status of null addr handle ref");
	}
	else
	{
		bdNChar8 addrStr[BD_ADDR_STRING_SIZE];
		bdAddressMap::getInstance()->addrToString(addrHandle, addrStr, sizeof(addrStr));

		switch(addrHandle->getStatus())
		{
			case bdAddrHandle::BD_ADDR_NOT_CONSTRUCTED:
			{
				status = BD_SOCKET_IDLE;
			}
			break;
			case bdAddrHandle::BD_ADDR_NOT_RESOLVED:
			{	
				status = BD_SOCKET_PENDING;
			}
			break;
			case bdAddrHandle::BD_ADDR_ERROR:
			{
				BD_INFO(BD_LOG_LEVEL, "Address handle error. Socket lost. %s", addrStr);
				status = BD_SOCKET_LOST;
			}
			break;
			case bdAddrHandle::BD_ADDR_RESOLVED:
			{
				const bdAddr &realAddr = addrHandle->getRealAddr();

				bdDTLSAssociation *dtlsSocket = BD_NULL;
				const bdBool found = m_dtls.get(realAddr, dtlsSocket);

				if(found && dtlsSocket)
				{
					status = dtlsSocket->getStatus();
				}
				else if(false == found)
				{
					BD_ERR(BD_LOG_LEVEL, "DTLS not found for %s", addrStr);
					status = BD_SOCKET_LOST;
				}
				else
				{
					BD_ERR(BD_LOG_LEVEL, "Invalid state! %s", addrStr);
					status = BD_SOCKET_LOST;
				}
			}
			break;
			case bdAddrHandle::BD_ADDR_UNRESOLVED:
			{	
				BD_INFO(BD_LOG_LEVEL, "Address handle unresolved. Socket lost. %s", addrStr);
				status = BD_SOCKET_LOST;
			}
			break;
		}
	}

	return status;
}


bdInt bdSocketRouter::sendTo(const bdAddrHandleRef addrHandle, 
							 const void *data, 
							 const bdUInt length)
{
	bdInt val = BD_NET_ERROR;
	if(BD_SOCKET_ROUTER_INITIALIZED != m_status)
	{
		BD_WARN(BD_LOG_LEVEL, "Cannot send data before class has been initialized");
	}
	else if(addrHandle.isNull())
	{
		BD_WARN(BD_LOG_LEVEL, "Cannot send data to null addr handle ref");
	}
	else
	{
		const bdAddr &realAddr = addrHandle->getRealAddr();

		bdDTLSAssociation *dtlsSocket = BD_NULL;
		m_dtls.get(realAddr, dtlsSocket);

		if(dtlsSocket)
		{
			val = dtlsSocket->sendTo(realAddr, data, length);
		}
		else
		{
			BD_ERR(BD_LOG_LEVEL, "Not established! Call connect() first.");
			bdNChar8 commonAddrInfo[BD_COMMON_ADDR_INFO_SIZE];
			bdCommonAddrInfo::getBriefInfo(addrHandle->m_commonAddr, commonAddrInfo, BD_COMMON_ADDR_INFO_SIZE);
			BD_INFO(BD_LOG_LEVEL, commonAddrInfo);
		}
	}

	return val;
}

bdInt bdSocketRouter::receiveFrom(bdAddrHandleRef &addrHandle, 
								  void *data,
								  const bdUInt size)
{
	bdInt val = BD_NET_WOULD_BLOCK;

	if(BD_SOCKET_ROUTER_INITIALIZED != m_status)
	{
		BD_WARN(BD_LOG_LEVEL, "Cannot receive data before class has been initialized");
	}
	else
	{
		bdBool receiving = true;
		bdAddr realAddr;

		while(receiving)
		{
			val = m_socket->receiveFrom(realAddr, data, size);

			if(val > 0)
			{
				const bdUByte8 type = *(reinterpret_cast<const bdUByte8 *>(data));

				// special case for DTLS data
				if(type < BD_DTLS_TYPE_MAX)
				{
					bdDTLSAssociation *dtlsSocket = BD_NULL;
					const bdBool dtlsFound = m_dtls.get(realAddr, dtlsSocket);

					if(dtlsFound && dtlsSocket)
					{
						val = dtlsSocket->receiveFrom(realAddr, data, static_cast<bdUInt>(val), addrHandle, data, size);

						if(val >= 0)
						{
							// done receiving. we have something to pass it to the next level.
							receiving = false; 
						}
					}
					else
					{
						dtlsSocket = new bdDTLSAssociation(m_socket,
															m_keyMap, 
															m_ECCKey, 
															realAddr, 
															addrHandle);

						val = dtlsSocket->receiveFrom(realAddr, data, static_cast<bdUInt>(val), addrHandle, data, size);

						if(dtlsSocket->getStatus() == BD_SOCKET_CONNECTED)
						{
							const bdUWord addrStringSize = 100;
							bdNChar8 addrString[addrStringSize];
							realAddr.toString(addrString, addrStringSize);

							BD_INFO(BD_LOG_LEVEL, "New incoming DTLS connection from %s.", addrString);

							const bdBool dtlsAdded = m_dtls.put(realAddr, dtlsSocket);
							if(false == dtlsAdded)
							{
								BD_ERR(BD_LOG_LEVEL, "New DTLS not added to list.");
							}

							if(val >= 0)
							{
								// done receiving. we have something to pass it to the next level.
								receiving = false; 
							}
						}
						else
						{
							delete dtlsSocket;
						}
					}
				}
				else
				{
					// pass packet on to interceptors until it is accepted.
					bdBool accepted = false;
					for(bdUInt i = 0; !accepted && (i < m_interceptors.getSize()); i++)
					{
						accepted = m_interceptors[i]->acceptPacket(m_socket, realAddr, data, static_cast<bdUInt>(val) , type);
					}
				}
			}
			else if (val < 0)
			{
				switch(val)
				{
					case BD_NET_CONNECTION_RESET:
						// Connection reset is potentially harmful as it can be spoofed
						// and used to reset connections.
						// If a genuine connection reset occurs, DTLS layer will correctly time out.
						// In debug mode we pay attention to it because time outs are very high.
#if defined BD_DEBUG
						if(bdAddressMap::getInstance()->getAddressMapping(realAddr, addrHandle))
						{
							if(addrHandle->getStatus() != bdAddrHandle::BD_ADDR_RESOLVED)
							{
								addrHandle->setRealAddr(realAddr);
								// done receiving. we have something to pass it to the next level.
								receiving = false; 
							}
						}
						break;
#else
						BD_INFO(BD_LOG_LEVEL, "Connection reset packet received. Ignoring.");
						break;
#endif
					case BD_NET_WOULD_BLOCK:
						// nothing left at this level.
						receiving = false;
						break;
					case BD_NET_MSG_SIZE:
						receiving = false;
						break;
					default:
						BD_INFO(BD_LOG_LEVEL, "Unknown error code: %d.", val);
						receiving = false;
						break;
				}
			}
			else
			{
				// This is not something that is expected to occur.
				BD_ERR(BD_LOG_LEVEL, "socket receiveFrom returned 0 ");
				receiving = false;
			}
		}
	}
	return val;
}

void bdSocketRouter::pump()
{
	m_natTrav.pump();
	m_qosProber.pump();

	bdQueue<bdAddr> toRemove;
	
	bdHashMap<bdAddr, bdDTLSAssociation *, bdAddrHash>::Iterator it = m_dtls.getIterator();
	while(it)
	{
		const bdAddr &addr = m_dtls.getKey(it);
		bdDTLSAssociation *conn = m_dtls.getValue(it);
		conn->pump();
		
		switch(conn->getStatus())
		{
			case BD_SOCKET_IDLE:
				BD_WARN(BD_LOG_LEVEL, "Invalid state!");
				break;
			case BD_SOCKET_PENDING:
			case BD_SOCKET_CONNECTED:
				break;
			case BD_SOCKET_LOST:
				{
					bdNChar8 addrString[BD_ADDR_STRING_SIZE];
					conn->getAddrHandle()->m_realAddr.toString(addrString, BD_ADDR_STRING_SIZE);
					
					BD_INFO(BD_LOG_LEVEL, "DTLS closed. Removing to %s.", addrString);
					
					conn->getAddrHandle()->setStatus(bdAddrHandle::BD_ADDR_UNRESOLVED);
					toRemove.enqueue(addr);
				}
				break;

		}

		m_dtls.next(it);
	}
	m_dtls.releaseIterator(it);

	while(!toRemove.isEmpty())
	{
		const bdAddr &addr = toRemove.peek();
		it = m_dtls.getIterator(addr);

		BD_ASSERT(it != BD_NULL, "Couldn't find dlts connection in map.");

		bdDTLSAssociation *dtls = m_dtls.getValue(it);

		delete dtls;
		m_dtls.remove(it);

		bdNChar8 addrString[BD_ADDR_STRING_SIZE];
		addr.toString(addrString, BD_ADDR_STRING_SIZE);
		BD_INFO(BD_LOG_LEVEL, "removed dtls connection to %s", addrString);

		toRemove.dequeue();
		m_dtls.releaseIterator(it);
	}
}

void bdSocketRouter::onNATAddrDiscovery(bdCommonAddrRef remote, const bdAddr &realAddr)
{
	bdAddrHandleRef addrHandle;

	if(BD_SOCKET_ROUTER_INITIALIZED != m_status)
	{
		BD_WARN(BD_LOG_LEVEL, "Cannot react to NAT addr discovery event before class has been initialized");
	}
	else if(m_caToAddrMap.get(remote, addrHandle))
	{
		m_caToAddrMap.remove(remote);

		addrHandle->setRealAddr(realAddr);

		bdDTLSAssociation *conn;
		if(!m_dtls.get(realAddr, conn))
		{
			bdNChar8 addrString[BD_ADDR_STRING_SIZE];
			realAddr.toString(addrString, BD_ADDR_STRING_SIZE);

			BD_INFO(BD_LOG_LEVEL, "Address lookup succeeded. Creating DTLS conn to %s.", addrString);

			conn = new bdDTLSAssociation(m_socket, m_keyMap, m_ECCKey, realAddr, addrHandle);
			conn->connect();

			m_dtls.put(realAddr, conn);
		}
		else
		{
			BD_WARN(BD_LOG_LEVEL, "already connected.");
		}
	}
	else
	{
		BD_ERR(BD_LOG_LEVEL, "address lookup failed.");
		bdNChar8 commonAddrInfo[BD_COMMON_ADDR_INFO_SIZE];
		bdCommonAddrInfo::getBriefInfo(remote, commonAddrInfo, BD_COMMON_ADDR_INFO_SIZE);
		BD_INFO(BD_LOG_LEVEL, commonAddrInfo);

	}
}

void bdSocketRouter::onNATAddrDiscoveryFailed(bdCommonAddrRef remote)
{
	BD_INFO(BD_LOG_LEVEL, "Address lookup failed");

	bdNChar8 commonAddrInfo[BD_COMMON_ADDR_INFO_SIZE];
	bdCommonAddrInfo::getInfo(remote, commonAddrInfo, BD_COMMON_ADDR_INFO_SIZE);
	BD_INFO(BD_LOG_LEVEL, commonAddrInfo);

	bdAddrHandleRef addrHandle;

	if (m_caToAddrMap.get(remote, addrHandle))
	{
		if (addrHandle->getStatus() == bdAddrHandle::BD_ADDR_NOT_RESOLVED)
		{
			addrHandle->setStatus(bdAddrHandle::BD_ADDR_ERROR);
		}
	}

	if(!m_caToAddrMap.remove(remote))
	{
		BD_ERR(BD_LOG_LEVEL, "address lookup failed.");
	}
}

void bdSocketRouter::registerInterceptor(bdPacketInterceptor *const interceptor)
{
	if(BD_SOCKET_ROUTER_INITIALIZED == m_status)
	{
#ifdef BD_DEBUG
		bdUInt32 index;
		if(m_interceptors.findFirst(interceptor, index))
		{
			BD_WARN(BD_LOG_LEVEL, "same listener registered multiple times.");
		}
	#endif // BD_DEBUG

		m_interceptors.pushBack(interceptor);
	}
	else
	{
		BD_ERR(BD_LOG_LEVEL, "Cannot register interceptors until this class is initialized");
	}
}

void bdSocketRouter::unregisterInterceptor(bdPacketInterceptor *const interceptor)
{
	m_interceptors.removeAllKeepOrder(interceptor);
}

bdQoSProbe* bdSocketRouter::getQoSProber()
{
	return &m_qosProber;
}

void bdSocketRouter::onSecurityKeyRemove(const bdSecurityID &id)
{
	bdHashMap<bdAddr, bdDTLSAssociation *, bdAddrHash>::Iterator it = m_dtls.getIterator();
	while (it)
	{
		bdDTLSAssociation *conn = m_dtls.getValue(it);
		if (*(conn->getLocalSecurityId()) == id)
		{
			conn->getAddrHandle()->setStatus(bdAddrHandle::BD_ADDR_UNRESOLVED);
			delete conn;
			m_dtls.remove(it);
			
		}
		else
		{
			m_dtls.next(it);	
		}
	}
	m_dtls.releaseIterator(it);
}

bdBool bdSocketRouter::setupIntroducers(const bdArray<bdAddr> &introducers)
{
	bdBool result = false;
	if(BD_SOCKET_ROUTER_INITIALIZED == m_status)
	{
		result = m_natTrav.setupIntroducers(introducers);
	}
	return result;
}

