//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_AlienQueen_GS2Move.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/AlienQueen/AVP_Behaviour_AlienQueen_GS2Move.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#endif

// axon
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Behaviour/BehaviourTypes/AlienQueen/AVP_Behaviour_AlienQueen_Movement.h"
#include "Core/BrainTypes/AVP_Brain_Actor_AlienQueen.h"

// avp
#include "Entity/PhysicalObject/Actor/Vertebrates/AlienQueen/AVP_ServerActor_AlienQueen.h"

//////////////////////////////////////////////////////////////////////////////
// C L A S S  F U N C T I O N S
// AVP_Behaviour_AlienQueen_GS2Move 

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move - Constructor
// Author       : Callum Hope CAH
// Date         : 13/03/09
// Scope		: Public
////////////////////////////////////////////////////////////////////////////////

AVP_Behaviour_AlienQueen_GS2Move::AVP_Behaviour_AlienQueen_GS2Move( AVP_Brain_Actor_AlienQueen* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_GS2MOVE )
, m_eCurrentState( UNSET )
, m_fTimeInState( 0.0f )
, m_fArrivalRange( 3.0f )
, m_xDestination( Asura_Maths::ZeroVector3 )
, m_bDestinationSet( false )
, m_uGS2NodeGuid( ASURA_GUID_UNREGISTERED )
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::Activate
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Virtual Protected
////////////////////////////////////////////////////////////////////////////////

bool AVP_Behaviour_AlienQueen_GS2Move::Activate()
{
#ifndef ASURA_NO_DEBUG_OUTPUT
	GetAlienQueenBrain()->SetRenderRoute( true );
#endif

	return PARENT::Activate();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::Deactivate
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Virtual Protected
////////////////////////////////////////////////////////////////////////////////

void AVP_Behaviour_AlienQueen_GS2Move::Deactivate()
{
	Axon_Behaviour* pxChild = GetChild();
	while( pxChild )
	{
		UnlinkChild( pxChild );
		GetProjectBrain()->DestroyBehaviour( pxChild );
		pxChild = GetChild();
	}

	PARENT::Deactivate();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::GetToPosition
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Public
//
// Desc			: Takes a destination and sets in motion the procedure for moving to a position. 
////////////////////////////////////////////////////////////////////////////////

bool AVP_Behaviour_AlienQueen_GS2Move::GetToPosition( const Asura_Vector_3& xDestination, Asura_Guid uGS2NodeGuid /*= ASURA_GUID_UNREGISTERED*/  )
{
	SetInError( false );
	SetFinished( false );

	// switching to state UNSET forces a call to CalculateMove in the update function
	SwitchToState( UNSET );

	m_xDestination = xDestination;
	m_bDestinationSet = true;

	//set the destination node guid if one was passed (for use but the child move behaviour to be created.)
	m_uGS2NodeGuid = uGS2NodeGuid;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::Update
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Virtual Protected. 
//
// Desc			: Update routiens depending on the current state. 
////////////////////////////////////////////////////////////////////////////////

void AVP_Behaviour_AlienQueen_GS2Move::Update( const float fTimeslice )
{
	PARENT::Update( fTimeslice );

	if( IsFinished() || IsInError() )
	{
		return;
	}

	m_fTimeInState += fTimeslice;
	switch( m_eCurrentState )
	{
		////////////////////
		case UNSET:
		////////////////////
		{
			Asura_Assert( m_bDestinationSet, "GetToPosition not called" );
			if( GetAlienQueenBrain()->CanPlanMove() &&  m_bDestinationSet )
			{
				CalculateMove();
			}
			break;
		}
		////////////////////
		case MOVE_TO_POSITION:
		////////////////////
		{
			AVP_Navigation::NavData& xNavData = GetAlienQueenBrain()->GetNavData();
			Axon_Behaviour* pxMovementChild = GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT );
			if( !pxMovementChild || pxMovementChild->IsInError() )
			{
				if(!GetToPosition(xNavData.m_xDestinationPosition))
				{
					SetInError(true);
					SwitchToState(FINISHED);
				}
			}
			else if(pxMovementChild->IsFinished())
			{
				SwitchToState(FINISHED);
			}
			break;
		}
		////////////////////
		case FINISHED:
		////////////////////
		{
			break;
		}
		////////////////////
		default:
		////////////////////
		{
			Asura_Assert(false, "unexpected state");
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::RenderDebugInfo (+Info)
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Virtual Protected	
//
// Desc			: Standard debug drawing functino for an entity. 
////////////////////////////////////////////////////////////////////////////////

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
	"UNSET",
	"GET TO POSITION",
	"FINISHED"
};

void AVP_Behaviour_AlienQueen_GS2Move::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	// PRINT OUT THE CURRENT STATE
	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "%s", s_aszStateStrings[ m_eCurrentState ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::GetAlienQueen Brain
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Protected		
//
// Desc			: Accessor to the alien brain controlling this behaviour. 
////////////////////////////////////////////////////////////////////////////////

AVP_Brain_Actor_AlienQueen* AVP_Behaviour_AlienQueen_GS2Move::GetAlienQueenBrain() const
{
	return static_cast<AVP_Brain_Actor_AlienQueen*>( GetProjectBrain() );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::SwitchToState
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Private
//
// Desc			: Deinitialisatoins and initialisations for each state. 
////////////////////////////////////////////////////////////////////////////////

void AVP_Behaviour_AlienQueen_GS2Move::SwitchToState( const State eNextState )
{
	switch( m_eCurrentState )
	{
		////////////////////
		case UNSET:
		////////////////////
		{
			break;
		}
		////////////////////
		case MOVE_TO_POSITION:
		////////////////////
		{
			if ( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT ) )
			{
				DestroyChildWithType( AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT );
			}
			break;
		}
		////////////////////
		case FINISHED:
		default:
		////////////////////
		{
			Asura_Assert(false, "unexpected state");
			break;
		}
	}

	m_eCurrentState = eNextState;
	m_fTimeInState  = 0.0f;
	switch( m_eCurrentState )
	{
		////////////////////
		case MOVE_TO_POSITION:
		////////////////////
		{
			AVP_Behaviour_AlienQueen_Movement* pxMovementBehaviour =
				static_cast<AVP_Behaviour_AlienQueen_Movement*>( CreateChildWithType( AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT ) );
			if( ! pxMovementBehaviour )
			{
				SetInError(true);
				SwitchToState( FINISHED );
			}
			else
			{
				//set the destination GS2 node id if we have one. 
				if( m_uGS2NodeGuid != ASURA_GUID_UNREGISTERED )
				{
					pxMovementBehaviour->SetDestinationGS2NodeGuid( m_uGS2NodeGuid );
				}

				// set speed to medium as default on starting a move.
				pxMovementBehaviour->GetToPosition( GetAlienQueenBrain()->GetNavData().m_xDestinationPosition, AVP_Navigation::MM_MEDIUM, true );
			}
			break;
		}
		////////////////////
		case FINISHED:
		////////////////////
		{
			m_bDestinationSet = false;
			SetFinished( true );
			GetAlienQueenBrain()->SetInGameScene( false );
			break;
		}
		////////////////////
		case UNSET:
		////////////////////
		{
			break;
		}
		////////////////////
		default:
		////////////////////
		{
			Asura_Assert(false, "unexpected state");
			break;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Behaviour_AlienQueen_GS2Move::CalculateMove
// Author       : Callum Hope CAH
// Date         : 16/03/09
// Scope		: Private
//
// Desc			: Takes the Navigation data, if possible to get to the destinations or we are not already there 
//				  it finds a path and switches to move state. 
////////////////////////////////////////////////////////////////////////////////

void AVP_Behaviour_AlienQueen_GS2Move::CalculateMove()
{
	AVP_Navigation::NavData& xNavData = GetAlienQueenBrain()->GetNavData();

	xNavData.m_xStartPosition = GetAlienQueenBrain()->GetActor()->GetPosition();
	xNavData.m_xDestinationPosition = m_xDestination;
	
	// Our current waypoint ID should already be up to date, as we keep it up to date when moving.
	// Update our destination waypoint ID.
	const AVP_Navigation::TestParams xTestParams( xNavData.m_uPathfinderFlags, ASURA_NAVVERTEX_MAX_ZONE_ID );
	if( ! AVP_Navigation::GetNearestWaypointID( m_xDestination, xNavData.m_usDestinationWPID, xTestParams ) )
	{
		SetInError();
		SwitchToState( FINISHED );
		return;
	}

	if( ( xNavData.m_xStartPosition - xNavData.m_xDestinationPosition ).MagnitudeSquared() < m_fArrivalRange * m_fArrivalRange )
	{
		// If we're already near enough the destination then there's nothing to do.
		SwitchToState( FINISHED );
	}
	else
	{
		// Clear out the old route and find the new one.
		xNavData.ResetRoute();

		AVP_Navigation::AlienQueenAStarFunctor xMoveFunctor;
		if( ! AVP_Navigation::FindRoute( xNavData, &xMoveFunctor ) )
		{
			SetInError();
			SwitchToState( FINISHED );
		}

		SwitchToState( MOVE_TO_POSITION );
		GetAlienQueenBrain()->SetInGameScene( true );
	}

	return;
}

