//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Dropship.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/Dropship/AVP_Behaviour_Dropship.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#include "Debugging/Asura_DebugLineRender.h"
#endif

// axon
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Behaviour/BehaviourTypes/Dropship/AVP_Behaviour_Dropship_FireMissiles.h"
#include "Behaviour/BehaviourTypes/Dropship/AVP_Behaviour_Dropship_Movement.h"
#include "Core/BrainTypes/AVP_Brain_Actor_Dropship.h"
#include "Core/AVP_Navigation_Types.h"

// avp
#include "Entity/PhysicalObject/Actor/Vehicles/Dropship/AVP_ServerVehicle_Dropship.h"

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Dropship
//
//////////////////////////////////////////////////////////////////////////////

AVP_Behaviour_Dropship::AVP_Behaviour_Dropship( AVP_Brain_Actor_Dropship* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_DROPSHIP )
, m_eCurrentState( UNSET )
, m_fTimeInState( 0.0f )
, m_bWasFired(false)
, m_bWantToFire(false)
, m_eMissileFireState(MFS_DEACTIVATED)
{
}

void AVP_Behaviour_Dropship::Deactivate()
{
	Axon_Behaviour* pxChild = GetChild();
	while ( pxChild )
	{
		UnlinkChild( pxChild );
		GetProjectBrain()->DestroyBehaviour( pxChild );
		pxChild = GetChild();
	}

    PARENT::Deactivate();
}

DEBUGVAR float dbg_fTimeBetweenFire = 10.0f;
DEBUGVAR float dbg_fHealthToMoveToBackPodsPercent = 50.0f;

void AVP_Behaviour_Dropship::PerFrameUpdate( const float fTimeslice )
{
    PARENT::PerFrameUpdate( fTimeslice );

	
    if ( IsFinished() || IsInError() )
    {
        return;
    }

    m_fTimeInState += fTimeslice;

	switch ( m_eCurrentState )
	{
	case UNSET:
		SwitchToState(FLY_AROUND);
		break;
	case FLY_AROUND:
		{
			Asura_Vector_3 xTarget;
			if (GetDropshipBrain()->GetTarget(xTarget))
			{
				SwitchToState(FLY_AROUND_AIM_AT_TARGET);
			}
		}
		break;
	case FLY_AROUND_AIM_AT_TARGET:
		{
			Asura_Vector_3 xTarget;
			if (!GetDropshipBrain()->GetTarget(xTarget))
			{
				SwitchToState(FLY_AROUND);
			}
			else
			{
				if (GetDropshipBrain()->GetUseMissiles())
				{
					switch (m_eMissileFireState)
					{
					case MFS_DEACTIVATED:
						SwitchToState(OPEN_FRONT_PODS);
						return;
					case MFS_USE_FRONT_MISSILES:
						{
							float fMaxHealth = GetDropshipBrain()->GetDropshipActor()->GetMaxHealth();
							float fCurrentHealth = GetDropshipBrain()->GetDropshipActor()->GetHealth();
							if (fCurrentHealth/fMaxHealth * 100.0f < dbg_fHealthToMoveToBackPodsPercent)
							{
								SwitchToState(OPEN_BACK_PODS);
								return;
							}
						}
						break;
					case MFS_USE_BACK_MISSILES:
						break;
					default:
						Asura_Assert(false, "Unexpected State");
					}

					if (m_fTimeInState > dbg_fTimeBetweenFire && m_eMissileFireState != MFS_DEACTIVATED)
					{
						SwitchToState(STOP_TO_FIRE);
					}

				}
			}
		}
		break;
	case OPEN_FRONT_PODS:
		if (GetDropshipBrain()->AreFrontPodsOpened())
		{
			m_eMissileFireState = MFS_USE_FRONT_MISSILES;
			SwitchToState(FLY_AROUND_AIM_AT_TARGET);
		}
		break;
	case OPEN_BACK_PODS:
		if (GetDropshipBrain()->AreBackPodsOpened())
		{
			m_eMissileFireState = MFS_USE_BACK_MISSILES;
			SwitchToState(FLY_AROUND_AIM_AT_TARGET);
		}
		break;
	case STOP_TO_FIRE:
		{
			AVP_Behaviour_Dropship_Movement * pxBehaviour = static_cast< AVP_Behaviour_Dropship_Movement* >( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_DROPSHIP_MOVEMENT ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				Asura_Assert(false, "Dropship movement behaviour failed");
			}
			else
			{
				if (pxBehaviour->IsHovering())
				{
					SwitchToState(FIRE);
				}
			}
		}
		break;
	case FIRE:
		{
			AVP_Behaviour_Dropship_FireMissiles * pxBehaviour = static_cast< AVP_Behaviour_Dropship_FireMissiles* >( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_DROPSHIP_FIRE_MISSILES ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				SwitchToState(FLY_AROUND_AIM_AT_TARGET);
			}
		}
		break;
	default:
		Asura_Assert(false, "Unexpected state");
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
    "UNSET",
	"FLY_AROUND",
	"FLY_AROUND_AIM_AT_TARGET",
	"STOP_TO_FIRE",
	"FIRE",
	"OPEN_FRONT_PODS",
	"OPEN_BACK_PODS",
	"NUM_STATES"

};

void AVP_Behaviour_Dropship::RenderDebugInfo()
{
    PARENT::RenderDebugInfo();

    Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
        ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "%s", s_aszStateStrings[ m_eCurrentState ] );
    GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

AVP_Brain_Actor_Dropship* AVP_Behaviour_Dropship::GetDropshipBrain() const
{
    return static_cast< AVP_Brain_Actor_Dropship* >( GetProjectBrain() );
}


void AVP_Behaviour_Dropship::SwitchToState( const State eNextState )
{
	
	switch ( m_eCurrentState )
	{
	case UNSET: // create the movement behaviour that will stay alive for the duration of the dropship
		{
			AVP_Behaviour_Dropship_Movement * pxBehaviour = static_cast< AVP_Behaviour_Dropship_Movement* >( CreateChildWithType( AVP_BEHAVIOUR_TYPE_DROPSHIP_MOVEMENT ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				Asura_Assert(false, "Dropship movement behaviour failed");
			}
			else
			{
				pxBehaviour->FollowPath();
			}
		}
		break;
	case FLY_AROUND:
	case FLY_AROUND_AIM_AT_TARGET:
	case STOP_TO_FIRE:
	case OPEN_FRONT_PODS:
	case OPEN_BACK_PODS:
		break;
	case FIRE:
		{
			DestroyChildWithType( AVP_BEHAVIOUR_TYPE_DROPSHIP_FIRE_MISSILES );
			AVP_Behaviour_Dropship_Movement * pxMoveBehaviour = static_cast< AVP_Behaviour_Dropship_Movement* >( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_DROPSHIP_MOVEMENT ) );
			if ( !pxMoveBehaviour || pxMoveBehaviour->IsInError() || pxMoveBehaviour->IsFinished())
			{
				Asura_Assert(false, "Dropship movement behaviour failed");
			}
			else
			{
				pxMoveBehaviour->FollowPath();
			}
		}
		break;
	default:
		Asura_Assert(false, "Unexpected state");
	}

	m_eCurrentState = eNextState;
	m_fTimeInState  = 0.0f;
	switch ( m_eCurrentState )
	{
	case FLY_AROUND:
		{
			AVP_Behaviour_Dropship_Movement * pxBehaviour = static_cast< AVP_Behaviour_Dropship_Movement* >( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_DROPSHIP_MOVEMENT ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				Asura_Assert(false, "Dropship movement behaviour failed");
			}
			else
			{
				pxBehaviour->AimAtTargetMode(false);
			}
		}

		break;
	case FLY_AROUND_AIM_AT_TARGET:
		{
			AVP_Behaviour_Dropship_Movement * pxBehaviour = static_cast< AVP_Behaviour_Dropship_Movement* >( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_DROPSHIP_MOVEMENT ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				Asura_Assert(false, "Dropship movement behaviour failed");
			}
			else
			{
				pxBehaviour->AimAtTargetMode(true);
			}

		}
		break;
	case STOP_TO_FIRE:
		{
			AVP_Behaviour_Dropship_Movement * pxBehaviour = static_cast< AVP_Behaviour_Dropship_Movement* >( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_DROPSHIP_MOVEMENT ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				Asura_Assert(false, "Dropship movement behaviour failed");
			}
			else
			{
				pxBehaviour->Hover();
			}
		}
		break;
	case FIRE:
		{
			AVP_Behaviour_Dropship_FireMissiles * pxBehaviour = static_cast< AVP_Behaviour_Dropship_FireMissiles* >( CreateChildWithType( AVP_BEHAVIOUR_TYPE_DROPSHIP_FIRE_MISSILES ) );
			if ( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished())
			{
				SwitchToState(FLY_AROUND_AIM_AT_TARGET);
			}
			else
			{
				GetDropshipBrain()->RefreshMissValue();
				pxBehaviour->SetNumberOfMissiles(4);
				if (m_eMissileFireState == MFS_USE_FRONT_MISSILES)
				{
					pxBehaviour->UseFrontPods();
				}
				else
				{
					pxBehaviour->UseBackPods();
				}
			}
		}
		break;
	case OPEN_FRONT_PODS:
		GetDropshipBrain()->OpenFrontPods();
		break;
	case OPEN_BACK_PODS:
		GetDropshipBrain()->OpenBackPods();
		break;
	case UNSET:
	default:
		Asura_Assert(false, "Unexpected state");
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
