///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Knowledge_System_AlienQueen.h
// Author : David Wooldridge
// Date   : 19/11/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////


// core
#include "Core/Asura.h"

// header
#include "Knowledge/AVP_Knowledge_System_AlienQueen.h"

// asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#include "Debugging/Asura_DebugLineRender.h"
#endif


// avp
#include "Core/BrainTypes/AVP_Brain_Actor.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/Controller/AVP_ServerEntity_Squad_Marine.h"
#include "Entity/Locations/ShootAt/AVP_ServerEntity_ShootAt.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AlienQueen/AVP_ServerActor_AlienQueen.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ServerActor_Vertebrate.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ServerActor_Human.h"
#include "Entity/PhysicalObject/AVP_ServerEntity_PhysicalObject.h"
#include "Entity/PlayerInfo/AVP_ServerEntity_PlayerInfo.h"

// axon
#include "Knowledge/Axon_Knowledge.h"
#include "Knowledge/SenseTypes/AVP_Sense_Hearing_Actor.h"
#include "Knowledge/SenseTypes/AVP_Sense_Vision_Actor.h"
#include "Tactical/AVP_Tactical_System.h"
#include "Core/AVP_Waypoint_Manager.h"



AVP_Knowledge_System_AlienQueen::AVP_Knowledge_System_AlienQueen( AVP_Brain_Actor* pxBrain )
: PARENT( pxBrain )
, m_xPriorityTargetKnowledge()
, m_xTargetList()
{
	m_xPriorityTargetKnowledge.Reset();
}


AVP_Knowledge_System_AlienQueen::~AVP_Knowledge_System_AlienQueen()
{
	m_xTargetList.Clear();
}

#ifndef ASURA_NO_DEBUG_OUTPUT
bool AVP_Knowledge_System_AlienQueen::Verify()
{
	return ( PARENT::Verify() && m_pxVision && m_pxHearing );
}
#endif


void AVP_Knowledge_System_AlienQueen::Initialise()
{
	PARENT::Initialise();

	m_xPriorityTargetKnowledge.Reset();
}


void AVP_Knowledge_System_AlienQueen::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	int iVersion;
	xStream >> iVersion;

	PARENT::ReadFromChunkStream( xStream );

	if ( iVersion == CURRENT_CHUNK_VERSION )
	{
	}
	else
	{
		Asura_Assertf( iVersion == CURRENT_CHUNK_VERSION, ( "AVP_Knowledge_System_Actor::ReadFromChunkStream - unhandled chunk version (%d)", iVersion ) );
	}
}


void AVP_Knowledge_System_AlienQueen::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << CURRENT_CHUNK_VERSION;

	PARENT::WriteToChunkStream( xStream );
}

void AVP_Knowledge_System_AlienQueen::PerFrameUpdate(const float fTimeslice)
{
	UpdateTargetKnowledge(fTimeslice);
}

void AVP_Knowledge_System_AlienQueen::Update(const float fTimeslice)
{
	PARENT::Update(fTimeslice);
}

///////////////////////////////////////////////////////////////////////////////
// Function		: AVP_Knowledge_System_AlienQueen::UpdateTargetKnowledge
// Author		: Callum Hope
// Date			: 9/4/2009 (re-written)
// Scope		: Public
//
// Desc			: At the moment the Priority target is the Player if they are on the target list or the closest potential target. 
///////////////////////////////////////////////////////////////////////////////

void AVP_Knowledge_System_AlienQueen::UpdateTargetKnowledge(const float fTimeslice)
{
	float fBestDistance = -1.0f;

	// each time we update the knowledge reset the current best target incase the last one has been killed and is no longer on the list
	m_xPriorityTargetKnowledge.m_uSourceGuid = ASURA_GUID_UNREGISTERED;
	m_xPriorityTargetKnowledge.m_uSourceClassification = AsuraEntityClass_NotSet;

	RemoveDeadTargetsFromList();

	if( m_xTargetList.GetNumEntries() == 0 )
	{
		// if we have an empty list set the Priority target guid to unregistered and return
		return;
	}

	Asura_Hashed_List_It< Axon_Knowledge > xTargetListIter( &m_xTargetList );
	while ( !xTargetListIter.Done() )
	{
		Axon_Knowledge* pxCurrentKnowledge = xTargetListIter.GetCurrent();

		if ( pxCurrentKnowledge )
		{
			AVP_ServerEntity_Actor* pxEntity = AVP_ServerEntity_Actor::FindEntity( pxCurrentKnowledge->m_uSourceGuid );
			
			if( pxEntity )
			{
				// update the position and direction of the entity in the knowledge.
				pxCurrentKnowledge->m_xPosition = pxEntity->GetPosition();

				// work out the distance of this target from the Queen and direction to it to update knowledge info. 
				Asura_Vector_3 xDirectionToTarget = pxEntity->GetPosition() - GetActorBrain()->GetActor()->GetPosition();
				pxCurrentKnowledge->m_fDistance = xDirectionToTarget.Magnitude();
				pxCurrentKnowledge->m_xDirection = xDirectionToTarget;

				if( pxEntity->IsPlayer() && ( pxEntity->GetSpeciesByClassification() != AVP_Settings::AVP_SPECIES_ALIEN ) )
				{
					// if the player is in the list and is not an alien then mark them as the priority and return. 
					m_xPriorityTargetKnowledge = *pxCurrentKnowledge;
					return;
				}

				if( ( pxCurrentKnowledge->m_fDistance < fBestDistance ) || ( fBestDistance == -1.0f ) )
				{
					// if this entity is closer to the queen than the previous best target OR this is the first target looked at set is as the current best so far. 
					m_xPriorityTargetKnowledge = *pxCurrentKnowledge;
					fBestDistance = pxCurrentKnowledge->m_fDistance;
				}
			}
		}

		xTargetListIter.Next();
	}

	//// This Needs changing in future as its single player only
	//AVP_ServerEntity_PlayerInfo* pxPlayerInfo = AVP_ServerEntity_PlayerInfo::FindEntity( 1000 ); 

	//if( pxPlayerInfo )
	//{
	//	// get the actual player actor from the puppet guid 
	//	AVP_ServerActor_Vertebrate* pxPlayer = AVP_ServerActor_Vertebrate::FindEntity( pxPlayerInfo->GetPuppetGuid() );

	//	if( pxPlayer )
	//	{
	//		if( pxPlayerInfo->GetPuppetSpecies() != AVP_Settings::AVP_SPECIES_ALIEN )
	//		{
	//			m_xPriorityTargetKnowledge.m_uSourceGuid	          = pxPlayer->GetGuid();
	//			m_xPriorityTargetKnowledge.m_uSourceClassification  = pxPlayer->GetClassification();
	//			m_xPriorityTargetKnowledge.m_xPosition		      = pxPlayer->GetPosition();
	//			
	//			if( GetBrain() )
	//			{
	//				AVP_ServerEntity_Actor* pxOwner = AVP_ServerEntity_Actor::FindEntity( GetBrain()->GetOwnerGuid() );
	//				
	//				if( pxOwner )
	//				{
	//					const Asura_Vector_3 xToPlayer = pxPlayer->GetPosition() - pxOwner->GetPosition();
	//					m_xPriorityTargetKnowledge.m_fDistance = xToPlayer.Magnitude();
	//				}
	//			}
	//		}
	//		else
	//		{
	//			m_xPriorityTargetKnowledge.m_uSourceGuid = ASURA_GUID_UNREGISTERED;
	//		}
	//	}
	//}

}

///////////////////////////////////////////////////////////////////////////////
// Function		: AVP_Knowledge_System_AlienQueen::AddEntityToTargetList
// Author		: Callum Hope
// Date			: 8/4/2009
// Scope		: Public
//
// Desc			: Takes a GUID of an entity, finds it and adds it to the Target List, if its allowed. 
///////////////////////////////////////////////////////////////////////////////

void AVP_Knowledge_System_AlienQueen::AddEntityToTargetList( Asura_Guid uGuid, int iClassification )
{
	// CAH we need to go right down to the asura server level here as its the first common ground between player info and ai entities.
	Asura_Entity_Server* pxEntity = Asura_Entity_Server::FindEntity( uGuid );
	//AVP_ServerEntity_Actor* pxEntity = AVP_ServerEntity_Actor::FindEntity( uGuid );

	if( pxEntity )
	{	
		// supposed to use the classification from the message, but this has been found to be in-accurate or not set most of the time. 
		switch( pxEntity->GetClassification() )
		{
			////////////////////
			case AVP_ENTITYCLASS_HUMAN:
			case AVP_ENTITYCLASS_ANDROID:
			case AVP_ENTITYCLASS_CIVILIAN:
			case AVP_ENTITYCLASS_MARINE:
			////////////////////
			{
				// all these are fine and just single entities, simply add them. 
				AddTarget( uGuid, iClassification );
				break;
			}
			////////////////////
			case AVP_ENTITYCLASS_PLAYER:
			case AVP_ENTITYCLASS_PLAYERINFO:
			////////////////////
			{
				AVP_ServerEntity_PlayerInfo* pxPlayerInfo = AVP_ServerEntity_PlayerInfo::FindEntity( uGuid ); 

				if( pxPlayerInfo )
				{
					// get the actual player actor from the puppet guid 
					AVP_ServerActor_Vertebrate* pxPlayer = AVP_ServerActor_Vertebrate::FindEntity( pxPlayerInfo->GetPuppetGuid() );

					if( pxPlayer )
					{
						if( ( pxPlayerInfo->GetPuppetSpecies() != AVP_Settings::AVP_SPECIES_ALIEN )
							&& ( pxPlayer->GetClassification() != AVP_ENTITYCLASS_SENTRYGUN ) )
						{
							// if we have a player and they are not an alien of any type or a sentry gun (sentry gun check because they have species type Human so smart gun doesnt target them.)
							AddTarget( pxPlayer->GetGuid(), pxPlayer->GetClassification() );
						}
					}
				}
				break;
			}
			////////////////////
			case AVP_ENTITYCLASS_MARINESQUAD:
			////////////////////
			{
				AVP_ServerEntity_Squad_Marine *pxSquad = AVP_ServerEntity_Squad_Marine::FindEntity( uGuid );

				if( pxSquad )
				{
					Asura_Guid uMemberGuid;

					for (u_int uCount = 0; uCount < pxSquad->GetNumSquadMembers(); uCount++ )
					{
						uMemberGuid = pxSquad->GetMemberGuid( uCount );

						if( uMemberGuid != ASURA_GUID_UNREGISTERED)
						{
							AVP_ServerActor_Human* pxSquadMember = AVP_ServerActor_Human::FindEntity( uMemberGuid );

							if( pxSquadMember )
							{
								if( ( pxSquadMember->GetClassification() == AVP_ENTITYCLASS_MARINE )
									|| ( pxSquadMember->GetClassification() == AVP_ENTITYCLASS_ANDROID ) )
								{
									AddTarget( uMemberGuid, pxSquadMember->GetClassification() );
								}
								else
								{
									Asura_Assert( false, "AVP_Knowledge_System_AlienQueen::AddEntityToTargetList\n A message is trying to add a squad that has members that are not androids or marines." );
								}
							}
						}
					}
				}
				break;
			}
			////////////////////
			default:
			////////////////////
			{
				// if this is not an entity that the queen is concerned with then just return, assert for degbug purposes. 
				Asura_Assertf( false, ( "AVP_Knowledge_System_AlienQueen::AddEntityToTargetList\n An entity is trying to be added to the queen's target list that she does not attack\n Entity %d with Classification %d not added", uGuid, pxEntity->GetClassification() ) );
				return;
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Function		: AVP_Knowledge_System_AlienQueen::AddTarget
// Author		: Callum Hope
// Date			: 8/4/2009
// Scope		: Private
//
// Desc			: Small Sub-routein of repetative task, just here to save space elseware in file. 
///////////////////////////////////////////////////////////////////////////////

bool AVP_Knowledge_System_AlienQueen::AddTarget( Asura_Guid uGuid, int iClassification )
{
	AVP_ServerEntity_PhysicalObject* pxEntity = AVP_ServerEntity_PhysicalObject::FindEntity( uGuid );

	if( pxEntity && ! pxEntity->IsDead() )
	{
		// if the entity exists and is not dead check to see if it is already in the target list, if so just return;
		if( m_xTargetList.FindObject( uGuid ) )
		{
			return false;
		}

		// so we are set to add the new entity as a target, create the new memory block to hold the data
		Axon_Knowledge *pxKnowledgeToAdd = CreateKnowledge();
		Asura_Assert( pxKnowledgeToAdd, "AVP_Knowledge_System_AlienQueen::AddTarget\n Unable to allocate knowledge from knowledge pool" );

		if( pxKnowledgeToAdd )
		{
			// now find the distance away from the queen by getting the magnitude of the difference vector from the queen to the entity. 
			Asura_Vector_3 xDirectionToTarget = pxEntity->GetPosition() - GetActorBrain()->GetActor()->GetPosition();
			float fDistance = xDirectionToTarget.Magnitude();
			
			// fill in all the data to the knowledge, not all will be used but do it anyway incase, so its there when it is used. 
			pxKnowledgeToAdd->m_uKnowledgeID          = GetNewKnowledgeID();
			pxKnowledgeToAdd->m_uSourceGuid           = uGuid;
			pxKnowledgeToAdd->m_uSourceClassification = iClassification;
			pxKnowledgeToAdd->m_iSenseID              = AXON_SENSE_TYPE_INVALID;
			pxKnowledgeToAdd->m_uTypeInformation      = 0;
			pxKnowledgeToAdd->m_fAge                  = 0.0f;
			pxKnowledgeToAdd->m_fThreatValue          = 0.0f;
			pxKnowledgeToAdd->m_xPosition             = pxEntity->GetPosition();
			pxKnowledgeToAdd->m_xDirection            = xDirectionToTarget;
			pxKnowledgeToAdd->m_fDistance             = fDistance;

			// Add it to the knowledge list.
			m_xTargetList.AddToList( pxKnowledgeToAdd );

			return true;
		}
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Function		: AVP_Knowledge_System_AlienQueen::RemoveEntityFromTargetList
// Author		: Callum Hope
// Date			: 8/4/2009
// Scope		: Public
//
// Desc			: Takes a GUID of an entity, and if there is an entity in the target list with the ID removes it. 
///////////////////////////////////////////////////////////////////////////////

void AVP_Knowledge_System_AlienQueen::RemoveEntityFromTargetList( Asura_Guid uGuid )
{
	Asura_Entity_Server* pxEntity = Asura_Entity_Server::FindEntity( uGuid );

	if( pxEntity )
	{
		switch( pxEntity->GetClassification() )
		{
			////////////////////
			case AVP_ENTITYCLASS_PLAYER:
			////////////////////
			{
				AVP_ServerEntity_PlayerInfo* pxPlayerInfo = AVP_ServerEntity_PlayerInfo::FindEntity( uGuid ); 

				if( pxPlayerInfo )
				{
					// get the actual player actor from the puppet guid 
					AVP_ServerActor_Vertebrate* pxPlayer = AVP_ServerActor_Vertebrate::FindEntity( pxPlayerInfo->GetPuppetGuid() );

					if( pxPlayer )
					{
						RemoveTarget( pxPlayer->GetGuid() );
					}
				}
				break;
			}
			////////////////////
			case AVP_ENTITYCLASS_MARINESQUAD:
			////////////////////
			{
				AVP_ServerEntity_Squad_Marine *pxSquad = AVP_ServerEntity_Squad_Marine::FindEntity( uGuid );

				if( pxSquad )
				{
					Asura_Guid uMemberGuid;

					for (u_int uCount = 0; uCount < pxSquad->GetNumSquadMembers(); uCount++ )
					{
						uMemberGuid = pxSquad->GetMemberGuid( uCount );

						if( uMemberGuid != ASURA_GUID_UNREGISTERED)
						{
							AVP_ServerActor_Human* pxSquadMember = AVP_ServerActor_Human::FindEntity( uMemberGuid );

							if( pxSquadMember )
							{
								RemoveTarget( pxSquadMember->GetGuid() );
							}
						}
					}
				}
				break;
			}
			////////////////////
			default:
			////////////////////
			{
				RemoveTarget( uGuid );
				break;
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Function		: AVP_Knowledge_System_AlienQueen::RemoveTarget
// Author		: Callum Hope
// Date			: 9/4/2009
// Scope		: Private
//
// Desc			: Small Sub-routein of repetative task, just here to save space elseware in file. 
///////////////////////////////////////////////////////////////////////////////

bool AVP_Knowledge_System_AlienQueen::RemoveTarget( Asura_Guid uGuid )
{
	Axon_Knowledge* pxToDelete = m_xTargetList.FindObject( uGuid ) ;

	if( pxToDelete )
	{
		m_xTargetList.RemoveFromList( pxToDelete );
		DestroyKnowledge( pxToDelete );
		return true;
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Function		: AVP_Knowledge_System_AlienQueen::RemoveDeadTargetsFromList
// Author		: Callum Hope
// Date			: 9/4/2009
// Scope		: Private
//
// Desc			: Small Sub-routein made for update target knowledge procedure, keeps the update function tidy. 
///////////////////////////////////////////////////////////////////////////////

void AVP_Knowledge_System_AlienQueen::RemoveDeadTargetsFromList()
{
	if( m_xTargetList.GetNumEntries() == 0 )
	{
		return;
	}

	Asura_Hashed_List_Deleter_It< Axon_Knowledge > xTargetListIter( &m_xTargetList );
	while ( ! xTargetListIter.Done() )
	{
		Axon_Knowledge* pxCurrentKnowledge = xTargetListIter.GetCurrent();

		if ( pxCurrentKnowledge )
		{
			AVP_ServerEntity_Actor* pxEntity = AVP_ServerEntity_Actor::FindEntity( pxCurrentKnowledge->m_uSourceGuid );

			if( pxEntity )
			{
				if( pxEntity->IsDead() )
				{
					// if the target is dead, remove it from the list and carry on. 
					xTargetListIter.DeleteCurrent();
					DestroyKnowledge( pxCurrentKnowledge );
					continue;
				}
			}
		}
		xTargetListIter.Next();
	}
}
