
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Actor_AttachmentHandler.h
// Author : Steven Lovesey
// Date   : 6-Apr-2009
//
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_ACTOR_ATTACHMENTHANDLER_H__
#define __AVP_ACTOR_ATTACHMENTHANDLER_H__

///////////////////
// Include files //
///////////////////

#include "Core/Asura.h"
#include "Core/Asura_Maths.h"

////////////////////////////////
// Forward class declarations //
////////////////////////////////

class Asura_Blueprint;
class AVP_ClientEntity_Actor;
class AVP_EntityDamage_ApplyToEntity;
class Asura_Animation;
class Asura_Message;

///////////////////////
// Class declaration //
///////////////////////

class AVP_Actor_Attachment
{

public:
	// Construction / Destruction.

	AVP_Actor_Attachment();
	virtual ~AVP_Actor_Attachment();

	Asura_Hash_ID m_uSkinID;
	Asura_Hash_ID m_uAnimID;
	Asura_Hash_ID m_uMarkerID;
	Asura_Hash_ID m_uPhysicsBlueprint;

	float m_fChanceOfDropOnDeath;
	float m_fChanceOfDropOnHeadbit;
	bool  m_bDropOnDecapitation;

	bool 	m_bDeleteOnDeath;
	bool	m_bDeleteOnHeadbit;
	bool	m_bDeleteOnDecapitation;

	bool	m_bUpdateFirstFrame;

	float	m_fAnimationStartTimeParametric; 

	Asura_Animation* m_pxAnimation; // Dont need to fill this out when creating attachments.

};

class AVP_Actor_AttachmentHandler
{

public:

	// Construction / Destruction.
	// Construction is done at actor creation.
	AVP_Actor_AttachmentHandler(AVP_ClientEntity_Actor* pxOwner);
	// Destruction is done at the shutdown of the animation handler after which we will have trouble finding our skin,
	// So we cant leave it till the destructor.
	virtual ~AVP_Actor_AttachmentHandler();

	// Create a new attachment or move an existing one to a new place.
	// Pass this a filled out AVP_Actor_Attachment structure describing its setup.
	bool AddAttachment(const AVP_Actor_Attachment& xAttachmentData);

	// Create a new attachment or move an existing one to a new place from blueprint specification
	bool AddAttachment(const Asura_Blueprint* pxBlueprint);

	// Detach the attachment, creating a new physics object in the world.
	// Note: Creation of physics object triggered on server on non players only.
	// This is due to possible issues with creating objects near the camera that would look ugly (face mask pops into view as you die..)
	// Also with players they may be in first or third person mode, meaning the attachments might not even exist on the server
	bool DetachAttachment(Asura_Hash_ID uSkin);

	// Destroy the attachment entirely.
	bool DestroyAttachment(Asura_Hash_ID uSkin);

	// Call on the death of the actor to trigger dropping/deleting of attachments.
	void OnDeath();

	// Call on the decapitation (head removal) of the actor to trigger dropping/deleting of attachments.
	void OnDecapitation();

	// Call when head is bitten or otherwise badly damaged to trigger dropping/deleting of attachments.
	void OnHeadBit();

	// Message handling.
	bool HandleMessage(const Asura_Message* pxMessage);

	// Damage handling, call in response to owner damage.
	void OnDamage(const AVP_EntityDamage_ApplyToEntity& xDamage); 

	// Destruction handling, call when owner is removed from the world.
	void OnDestruction();

	// Return the last damage impulse felt by this attachment handlers actor.
	const Asura_Vector_3& GetLastDamageImpulse() const { return m_xLastDamageImpulse; }

protected:

	// Destroy all attachments entirely.
	bool DestroyAllAttachments();

	// Remove attachment with this index;
	bool DestroyAttachmentByIndex(u_int uIndex);

	// Detach the attachment with this index, creating a new physics object in the world.
	// Note: Creation of physics object triggered on server on non players only.
	// This is due to possible issues with creating objects near the camera that would look ugly (face mask pops into view as you die..)
	// Also with players they may be in first or third person mode, meaning the attachments might not even exist on the server
	bool DetachAttachmentByIndex(u_int uIndex);

	// This function performs the creation of the skin/animation for the attachment 
	// and links it to the owner as specified in the attachment with this index.
	bool OnAttach(u_int uIndex);

	// This function performs the creation of the skin/animation for the attachment
	// and links it to the owner as specified in the attachment with this index.
	bool OnDetach(u_int uIndex);

	// Creates a physics object to represent the attachment with this index.
	bool CreateDroppedAttachment(u_int uIndex);

private:

	u_int					m_uNumAttachments;		// How many attachments are stored here.
	u_int					m_uAttachmentStorage;	// Space for this many attachments.
	AVP_Actor_Attachment*	m_pxAttachments;		// Pointer to attachment storage.
	AVP_ClientEntity_Actor* m_pxOwner;				// Pointer to owner.
	Asura_Vector_3			m_xLastDamageImpulse;   // Last impulse felt by actor from damage.

};

#endif   // ifndef __AVP_ACTOR_ATTACHMENTHANDLER_H__

/// End of file. ///
