
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Actor_DynamicStat.h
// Author : Andy Tate
// Date   : 6-Oct-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_ACTOR_DYNAMICSTAT_H__
#define __AVP_ACTOR_DYNAMICSTAT_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_Actor_DynamicStat
// Author       : Andy Tate
// Date         : 6-Oct-2006
//
// Description  : Abstract class to handle the updating of actor stats, the stat
//				  can be drained and refilled over a period of time as required.
////////////////////////////////////////////////////////////////////////////////

class AVP_Actor_DynamicStat
{
public:
	// Public typedefs...
	typedef float UpdateFunction (const float fTimeStep, const float fCurrent, const float fDifference, const float fChargeParameter);

	// Public functions...
	AVP_Actor_DynamicStat  (const float fInitial, const float fMin = 0.0f, const float fMax = Asura_Maths::LargeFloat);
//	~AVP_Actor_DynamicStat ();

	// Manipulation functions...
	void	SetToMaximum	(const bool bImmediate)	{ UpdateValues(m_fMaxValue, bImmediate); }
	void	SetToMinimum	(const bool bImmediate)	{ UpdateValues(m_fMinValue, bImmediate); }

	void	Set				(const float fValue, const bool bImmediate)	{ UpdateValues(fValue, bImmediate); }
	void	Add				(const float fValue, const bool bImmediate)	{ UpdateValues(m_fCurrentValue + fValue, bImmediate); }
	void	Remove			(const float fValue, const bool bImmediate)	{ UpdateValues(m_fCurrentValue - fValue, bImmediate); }

	float	GetMax			() const { return m_fMaxValue; }
	void	SetMax			(const float fValue);

	void	Update			(const float fTimeStep, const float fChargeParameter, UpdateFunction* pfnDrain, UpdateFunction* pfnRefill);
	void	Reset			(void);

	// Accessor functions...
	float	GetCurrentValue			(void) const				{ return(m_fCurrentValue); }

	float	GetTimeAtDesiredValue	(void) const				{ return(m_fTimeAtDesiredValue); }

	bool	CanAdd					(const float fValue) const	{ return((m_fCurrentValue + fValue) <= m_fMaxValue); }
	bool	CanRemove				(const float fValue) const	{ return((m_fCurrentValue - fValue) >= m_fMinValue); }
	bool	IsAtMinimum				(void) const				{ return(m_fCurrentValue <= m_fMinValue); }
	bool	IsAtMaximum				(void) const				{ return(m_fCurrentValue >= m_fMaxValue); }


protected:
	// Protected variables...
	float	m_fDesiredValue;		// The value we're to aim for.
	float	m_fCurrentValue;		// The current stat value.

	float	m_fTimeAtDesiredValue;	// The number of seconds we've been at the desired value.

	float	m_fInitialValue;		// The initial value for this stat.
	float	m_fMinValue;			// The minimum value for this stat.
	float	m_fMaxValue;			// The maximum value for this stat.

	// Protected functions...
	void	UpdateValues	(const float fValue, const bool bImmediate);


private:
	// Private variables...

	// Private functions...
	AVP_Actor_DynamicStat  (void);
	AVP_Actor_DynamicStat  (const AVP_Actor_DynamicStat& xOther);

	AVP_Actor_DynamicStat&	operator=	(const AVP_Actor_DynamicStat& xOther);

};   // End of class 'AVP_Actor_DynamicStat'.


#endif   // ifndef __AVP_ACTOR_DYNAMICSTAT_H__


/// End of file. ///
