////////////////////////////////////////////////////////////////
//
// File:		AVP_Actor_Selector.cpp
// Author:		Chris Wallace
// Date:		16-March-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/AVP_Actor_Selector.h"


#include "GUIMenu/Asura_GUIMenu_Element_Image.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Base.h"
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"
#include "Core/Asura_TiledTextureManagement.h"

#include "Actor/AVP_Actor_InputHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_HUDHandler.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/AVP_Inventory_HashIDs.h"

#include "Inventory/Items/Weapons/AVP_InventoryWeapon_SmartGun.h"

#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Sound/Asura_Sound_Event_System.h"
#include "Sound/Asura_Sound_Event_Storage.h"

DEBUGVAR float dbg_fKeepOpenAfterButtonPressTime	= 0.1f;
DEBUGVAR float dbg_fKeepOpenTime					= 2.0f;
DEBUGVAR float dbg_fFadeUpTime						= 0.2f;
DEBUGVAR float dbg_fFadeDownTime					= 0.2f;
DEBUGVAR float dbg_fPressTimeDelay					= 0.1f;

AVP_Actor_Selector::AVP_Actor_Selector( AVP_ClientEntity_Actor& xClient, const u_int uNumSlots, const int* piSlotKey, const Selector_Directions* peSlotDir, const Asura_Hash_ID uInventoryGroupHash, bool* pbShowConsoleVar )
: m_pxSlot( NULL )
, m_piSlotKey( piSlotKey )
, m_peSlotGuiElement( NULL )
, m_peSlotDirections( peSlotDir )
, m_pxControlMap( NULL )
, m_pbShowConsoleVar( pbShowConsoleVar )
, m_uOwnerGuid( xClient.GetGuid() )
, m_uInvGroupHash( uInventoryGroupHash )
, m_uEmptySlotImageHash( ASURA_HASH_ID_UNSET )
, m_uFlags( 0 )
, m_fKeepOpenTimer( dbg_fKeepOpenTime )
, m_fFadeAmount( 0.0f )
, m_fKeyPressedTime( 0.0f )
{
    m_uNumSlots = uNumSlots;

	m_pxSlot = new AVP_Actor_SelectorSlot[ uNumSlots ];
	Initialise( xClient, m_uInvGroupHash );
}

AVP_Actor_Selector::~AVP_Actor_Selector()
{
	delete[] m_pxSlot;
}

void AVP_Actor_Selector::Update( const AVP_ClientEntity_Actor& xClient )
{
	if ( IsActive() )
	{
		const AVP_Actor_InputHandler* pxInputHandler = xClient.GetInputHandler();
		m_fFadeAmount += Asura_Timers::GetGameTimeSlice()/dbg_fFadeUpTime;

		if(pxInputHandler)
		{
			u_int uNumKeysPressed(0);
			int uSelected(-1);
			for ( u_int uItr = 0; uItr < m_uNumSlots; ++uItr )
			{
				if ( pxInputHandler->GetControlMapValue( m_piSlotKey[ uItr ] ) > 0.0f )
				{
					uSelected = uItr;
					++uNumKeysPressed;
				}
			}
	
			if ( uNumKeysPressed == 0 && pxInputHandler->GetControlMapValue( AVP_Actor_InputHandler::WEAPON_SELECTOR_HOLD ) > 0.0f )
			{
				uSelected = m_uCurrentItem;
				++uNumKeysPressed;
			}

			const float fTimeSlice = Asura_Timers::GetGameTimeSlice();

			if( uNumKeysPressed == 0 )
			{
				m_fKeepOpenTimer -= fTimeSlice;
				m_fKeyPressedTime = 0.0f;
				if ( m_fKeepOpenTimer <= 0.0f )
				{
					m_bReadyToClose = true;
				}
			}
			else
			{
				m_bReadyToClose = false;
				m_fKeepOpenTimer = dbg_fKeepOpenAfterButtonPressTime;
				
				if ( uNumKeysPressed == 1 )
				{
					m_fKeyPressedTime += fTimeSlice;
				}

				if ( ((m_fKeyPressedTime > dbg_fPressTimeDelay) && (static_cast< int >( m_uCurrentItem ) != uSelected)) || !m_bSlotSelected )
				{
					SelectItemSlot( xClient, uSelected );
				}
			}
		}
	}
	else
	{
		if(m_fFadeAmount > 0.0f)
		{
			m_fFadeAmount -= Asura_Timers::GetGameTimeSlice() / dbg_fFadeDownTime;
		}
		else
		{
			m_bReadyToClose = false;
		}
	}

	*m_pbShowConsoleVar = m_fFadeAmount > 0.0f;

	// Set the properties of the GuiMenu elements
	m_fFadeAmount = Asura_Maths::ClampToRange( m_fFadeAmount, 0.0f, 1.0f );
}

void AVP_Actor_Selector::Activate( AVP_ClientEntity_Actor& xClient )
{
	AVP_Actor_InputHandler* pxInputHandler = xClient.GetInputHandler();

	if(!m_bActive)
	{
		m_bActive = true;

		AVP_GUIMenu_System::Activate(AVP_GUIMenu_System::GUIMENU_ID_WEAPONSELECTOR);
 		SelectItemSlot(xClient, m_uCurrentItem); // re-select previous
// 		Asura_Hash_ID uWeaponSelectPage = AVP_GUIMenu_System::GetMenuHash( AVP_GUIMenu_System::GUIMENU_ID_WEAPONSELECTOR );
// 		Asura_GUIMenu_Menu_Page* pxPage = Asura_GUIMenu_System::GetMenuPageByID(uWeaponSelectPage);
// 		if(pxPage)
// 		{
// 			pxPage->DeselectActiveWidget();
// 		}	

		Initialise( xClient, m_uInvGroupHash );
		m_bReadyToClose = false;
		m_fKeepOpenTimer = dbg_fKeepOpenTime;
		m_fKeyPressedTime = dbg_fPressTimeDelay;

		if ( m_pxControlMap )
		{
			pxInputHandler->PushControls( m_pxControlMap );
		}

		PlaySound(uAVP_BLUEPRINT_PARAM_SELECTOR_SOUND_OPEN, xClient);
	}
}

void AVP_Actor_Selector::Deactivate( AVP_ClientEntity_Actor& xClient )
{
	if(m_bActive)
	{
		AVP_Inventory_System*	pxInventorySystem	= xClient.GetInventorySystem();
		AVP_Inventory_Group*	pxGroup				= (pxInventorySystem == NULL) ? NULL : pxInventorySystem->GetGroup( m_uInvGroupHash );

		if( pxGroup )
		{
			// Make sure it is actually valid to change weapons.
			if(m_pxSlot[m_uCurrentItem].m_uInvItemHash != pxGroup->GetCurrentItem()) // Did we chose a new weapon?
			{
				if(CanUseSelector(xClient))
				{
					pxGroup->SetCurrentItem( m_pxSlot[m_uCurrentItem].m_uInvItemHash );

					PlaySound(uAVP_BLUEPRINT_PARAM_SELECTOR_SOUND_ACCEPT, xClient);
				}
				else
				{
					PlaySound(uAVP_BLUEPRINT_PARAM_SELECTOR_SOUND_CANCEL, xClient);
				}
			}
			else
			{
				PlaySound(uAVP_BLUEPRINT_PARAM_SELECTOR_SOUND_CANCEL, xClient);
			}
		}

		m_bActive = false;
		AVP_GUIMenu_System::Deactivate(AVP_GUIMenu_System::GUIMENU_ID_WEAPONSELECTOR);


		AVP_Actor_InputHandler* pxInputHandler = xClient.GetInputHandler();
		if (m_pxControlMap && pxInputHandler)
		{
			pxInputHandler->PopControls();
		}
	}
}

bool AVP_Actor_Selector::CanUseSelector( AVP_ClientEntity_Actor& xClient )
{
	AVP_Inventory_System*	pxInventorySystem	= xClient.GetInventorySystem();
	AVP_Inventory_Group*	pxGroup				= (pxInventorySystem == NULL) ? NULL : pxInventorySystem->GetGroup( m_uInvGroupHash );

	if( pxGroup )
	{
		const AVP_Inventory_Item* pxCurrentItem = pxGroup->GetCurrentInventoryItem();
		if ( !pxCurrentItem || pxCurrentItem->IsIdling(true) || pxCurrentItem->IsUsing(true) )
		{
			return true;
		}
	}
	return false;
}

void AVP_Actor_Selector::Initialise( const AVP_ClientEntity_Actor& xClient, const Asura_Hash_ID uGroupHash )
{
	m_uInvGroupHash = uGroupHash;
	m_bSlotSelected = false;

	const AVP_Inventory_System*	pxInventorySystem	= xClient.GetInventorySystem();
	const AVP_Inventory_Group*	pxGroup				= (pxInventorySystem == NULL) ? NULL : pxInventorySystem->GetGroup(uGroupHash);

	for( u_int uItr = 0; uItr < m_uNumSlots; ++uItr )
	{
		m_pxSlot[ uItr ].m_bInUse = false;
	}

	if ( pxGroup != NULL )
	{
		//Get the item at the top of the list
		if(pxGroup->GetNumItems() <= m_uNumSlots)
		{
			if(pxGroup->GetHashInParentSystem() == uAVP_INVENTORY_GROUPHASH_WEAPON)
			{
				Asura_Hash_ID uStartingID = uAVP_INVENTORY_WEAPONHASH_WEAPON01;
				
				for(u_int u=0;u<m_uNumSlots;u++) // reset to 0
				{
					AVP_Vertebrate_HUDHandler::s_auWeaponSelectorDispCondIndex[m_peSlotDirections[u]] = 0;
				}

				for(u_int u=0;u<m_uNumSlots;u++)
				{
					Asura_Hash_ID uWeaponHash = uStartingID+u;
					const AVP_Inventory_Item* pxItem = pxGroup->GetItem(uWeaponHash);
					if(pxItem)
					{
						u_int uSlot = GetItemSlotForInvItem( pxGroup, uWeaponHash );

						if(m_peSlotDirections)
						{
							AVP_Vertebrate_HUDHandler::s_auWeaponSelectorDispCondIndex[m_peSlotDirections[uSlot]] = pxItem->GetSelectorDisplayConditionIndex();
						}
						
						m_pxSlot[ uSlot ].Initialise( pxGroup, uWeaponHash );
						if(pxItem->GetItemType() == AVP_INVENTORYITEMTYPE_SMARTGUN && (uSlot+1 < m_uNumSlots))
						{
							if(m_peSlotDirections)
							{
								AVP_Vertebrate_HUDHandler::s_auWeaponSelectorDispCondIndex[m_peSlotDirections[uSlot+1]] = pxItem->GetSelectorDisplayConditionIndex();
							}
							m_pxSlot[ uSlot+1 ].Initialise( pxGroup, uWeaponHash );
							break;
						}
					}
				}
			}
			else
			{
				//Get the item at the top of the list
				Asura_Hash_ID uItem = pxGroup->FindLowestPriorityItem();
				while ( uItem != ASURA_HASH_ID_UNSET )
				{
					u_int uSlot = GetItemSlotForInvItem( pxGroup, uItem );
					if ( uSlot < m_uNumSlots )
					{
						m_pxSlot[ uSlot ].Initialise( pxGroup, uItem );
// 						if ( pxGroup->GetCurrentItem() == uItem )
// 						{
// 							m_uCurrentItem = uSlot;
// 						}
						uItem = pxGroup->GetNextItem( uItem, false, true );
					}
				}
			}
		}
	}
}

void AVP_Actor_Selector::SelectItemSlot( const AVP_ClientEntity_Actor& xClient, u_int uItemSlot )
{
	Asura_Assert( uItemSlot < m_uNumSlots, "Bad item index" );
	if ( uItemSlot < m_uNumSlots && m_pxSlot[ uItemSlot ].m_bInUse && IsItemSelectable( xClient, uItemSlot ) )
	{
		m_uCurrentItem = uItemSlot;

		AVP_Vertebrate_HUDHandler::s_uCurrentWeaponDispCondSlot = m_uCurrentItem;

		Asura_GUIMenu_Menu_Page* pxPage = Asura_GUIMenu_System::GetMenuPageByID(AVP_GUIMenu_System::GetMenuHash(AVP_GUIMenu_System::GUIMENU_ID_WEAPONSELECTOR));
		if(pxPage)
		{
			Asura_GUIMenu_Root* pxRoot = AVP_GUIMenu_System::GetHUDElement( m_peSlotGuiElement[ uItemSlot ], pxPage->GetID(), false);
			if(pxRoot && pxRoot->IsWidget())
			{
				Asura_GUIMenu_Widget_Base* pxWidget = static_cast<Asura_GUIMenu_Widget_Base*>(pxRoot);
				pxWidget->SetAsFocusedObject(true);
				pxPage->SelectWidget(pxWidget->GetID());
				pxPage->SetStartupOverrideWidgetID(pxWidget->GetID());
			}
		}
		
		PlaySound(uAVP_BLUEPRINT_PARAM_SELECTOR_SOUND_CHANGE, xClient);

		m_fKeyPressedTime = 0.0f;
		m_uPrevSelected = uItemSlot;
		m_bSlotSelected = true;
	}
}

void AVP_Actor_Selector::SelectNextItem(const AVP_ClientEntity_Actor& xClient, bool bWrap)
{
	u_int uTempItem = m_uCurrentItem;

	do
	{	
		if ( bWrap || uTempItem < ( m_uNumSlots - 1 ) )
		{
			if ( uTempItem == ( m_uNumSlots - 1 ) )
			{
				uTempItem = 0;
			}
			else
			{
				++uTempItem;
			}
		}
	}
	while ( !m_pxSlot[ uTempItem ].m_bInUse || !IsItemSelectable( xClient, uTempItem ) );

	m_uCurrentItem = uTempItem;
}

void AVP_Actor_Selector::SelectPreviousItem(const AVP_ClientEntity_Actor& xClient, bool bWrap)
{
	u_int uTempItem = m_uCurrentItem;

	do
	{	
		if ( bWrap || uTempItem > 0 )
		{
			if ( uTempItem == 0 )
			{
				uTempItem = m_uNumSlots - 1;
			}
			else
			{
				uTempItem--;
			}
		}
	}
	while ( uTempItem >= m_uNumSlots || !m_pxSlot[ uTempItem ].m_bInUse || !IsItemSelectable( xClient, uTempItem ) );

	m_uCurrentItem = uTempItem;
}

bool AVP_Actor_Selector::IsItemSelectable( const AVP_ClientEntity_Actor& xClient, int iItemIndex ) const
{
	const AVP_Inventory_System*	pxInventorySystem	= xClient.GetInventorySystem();
	const AVP_Inventory_Group*	pxGroup				= ( pxInventorySystem == NULL ) ? NULL : pxInventorySystem->GetGroup( m_uInvGroupHash );

	if ( pxGroup != NULL )
	{
		return pxGroup->IsItemSelectable( m_pxSlot[ iItemIndex ].m_uInvItemHash );
	}
	else
	{
		return false;
	}
}

bool AVP_Actor_Selector::IsItemEmpty( const AVP_ClientEntity_Actor& xClient, int iItemIndex ) const
{
	const AVP_Inventory_System*	pxInventorySystem	= xClient.GetInventorySystem();
	const AVP_Inventory_Group*	pxGroup				= ( pxInventorySystem == NULL ) ? NULL : pxInventorySystem->GetGroup( m_uInvGroupHash );

	if ( pxGroup != NULL )
	{
		const AVP_Inventory_Item* pxItem = pxGroup->GetItem( m_pxSlot[ iItemIndex ].m_uInvItemHash );
		if(pxItem)
		{
			return pxItem->AreAllSubTypesEmpty();
		}
	}

	return true;
}

u_int AVP_Actor_Selector::GetItemSlotForInvItem( const AVP_Inventory_Group* pxGroup, const Asura_Hash_ID uItemHash )
{
	const AVP_Inventory_Item*	pxItem	= pxGroup->GetItem(uItemHash);
	int iSlot = pxItem != NULL ? pxItem->GetFixedSlot() : -1;
#ifdef ASURA_ASSERT
	if ( iSlot > 0 && iSlot < static_cast< int >( m_uNumSlots ) )
	{
		Asura_Assert( !m_pxSlot[ iSlot ].m_bInUse, "Slot already in use" );
	}
#endif
	for ( u_int uItr = 0; iSlot < 0 && uItr < m_uNumSlots; ++uItr )
	{
		if ( !m_pxSlot[ uItr ].m_bInUse )
		{
			iSlot = uItr;
		}
	}
	Asura_Assertf( iSlot >= 0 && iSlot < static_cast< int >( m_uNumSlots ), ( "No slot for item 0x%08X", uItemHash ) );
	return Asura_Maths::ClampToRange< int >( iSlot, 0, m_uNumSlots );
}

u_int AVP_Actor_Selector::GetItemHashFromSlot( u_int uIndexFromBluePrints )
{
	u_int uIndex = uIndexFromBluePrints - 1; // Blueprints use values starting from 1.

	if(uIndex < m_uNumSlots)
	{
		return m_pxSlot[uIndex].m_uInvItemHash;
	}
	return ASURA_HASH_ID_UNSET;
}

AVP_Actor_Selector::AVP_Actor_SelectorSlot::AVP_Actor_SelectorSlot()
: m_uInvItemHash ( ASURA_HASH_ID_UNSET )
, m_uSelectedImageHash( ASURA_HASH_ID_UNSET )
, m_uNotSelectedImageHash( ASURA_HASH_ID_UNSET )
, m_uSelectedEmptyImageHash( ASURA_HASH_ID_UNSET )
, m_uNotSelectedEmptyImageHash( ASURA_HASH_ID_UNSET )
, m_uNotSelectableImageHash( ASURA_HASH_ID_UNSET )
, m_uFlags( 0 )
{
	
}

void AVP_Actor_Selector::AVP_Actor_SelectorSlot::Initialise( const AVP_Inventory_Group* pxGroup, const Asura_Hash_ID uItemHash )
{
	const AVP_Inventory_Item*	pxItem	= (pxGroup)?pxGroup->GetItem(uItemHash):NULL;

	if(pxItem)
	{
		m_uInvItemHash = uItemHash;
		m_uSelectedImageHash = pxItem->GetMenuImageHash(true);
		m_uNotSelectedImageHash = pxItem->GetMenuImageHash(false);
		m_uSelectedEmptyImageHash = pxItem->GetMenuEmptyImageHash(true);
		m_uNotSelectedEmptyImageHash = pxItem->GetMenuEmptyImageHash(false);
		m_uNotSelectableImageHash = pxItem->GetNotSelectableMenuImageHash();
		m_bInUse = true;
	}
}

void AVP_Actor_Selector::PlaySound(Asura_Hash_ID uBlueprintParamHash,  const AVP_ClientEntity_Actor& xClient)
{
	Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_PLAYERS, xClient.GetBlueprintHashID());
	if(pxBlueprint)
	{
		Asura_Hash_ID uSoundEventHash = pxBlueprint->GetParameterAsHash(uBlueprintParamHash ,ASURA_HASH_ID_UNSET);

		const Asura_Sound_Event* pxSoundEvent = Asura_Sound_Event_Storage::FindEvent(uSoundEventHash);
		Asura_Sound_Event_System::PlayEvent(pxSoundEvent, NULL);
	}
}