////////////////////////////////////////////////////////////////
//
// File:		AVP_Vertebrate_MeleeCounter_System.cpp
// Author:		Chris Wallace
// Date:		22-Jan-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////

///////////////////////////////////////////////
// #includes
///////////////////////////////////////////////
#include "Core/Asura.h"

// This files header
#include "Actor/Vertebrates/AVP_Vertebrate_MeleeCounter_System.h"

#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_Broadphase.h"
#include "Physics/Asura_Physics_MovementQuery.h"
#include "Physics/Asura_Physics_RigidBody.h"

#include "Actor/Vertebrates/AVP_Vertebrate_Description.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Focus.h"
#include "Actor/AVP_Actor_PhysicsObjectHandler.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_LOS.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/Container/AVP_ServerEntity_Container.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_Vertebrate_Messages.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ServerActor_Vertebrate.h"
#include "Entity/PhysicalObject/PhysicsObject/AVP_ServerEntity_PhysicsObject.h"
#include "Entity/AVP_Messages.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/AVP_InventoryItem_Weapon.h"
#include "Moves/Vertebrates/AVP_MovesMelee.h"
#include "Moves/Vertebrates/Predator/AVP_Predator_MovesMelee.h"

#include "Core/BrainTypes/AVP_Brain_Actor.h"
#include "Tactical/AVP_Tactical_System.h"


///////////////////////////////////////////////
// Constants
///////////////////////////////////////////////

DEBUGVAR float dbg_fDefaultDamageRadius = 2.0f;
DEBUGVAR float dbg_fDefaultDamageAIBackAttackFactor = 0.5f;
DEBUGVAR float dbg_fDefaultAttackDetectionRadius = 4.0f;
DEBUGVAR float dbg_fDefaultTargetDetectionRadius = 12.0f;
DEBUGVAR float dbg_fDefaultTargetDetectionCosAngle = 0.866f;
DEBUGVAR float dbg_fDefaultTargetDetectionOriginBackwardsOffset = 0.5f;
DEBUGVAR float dbg_fDefaultBlockDetectionCosAngle = 0.707f;
DEBUGVAR float dbg_fDefaultBlockDamageProportion = 0.5f;

DEBUGVAR float dbg_fDefaultMinorMeleeDamageAmount		= 30.0f;
DEBUGVAR float dbg_fDefaultModerateMeleeDamageAmount	= 60.0f;
DEBUGVAR float dbg_fDefaultMajorMeleeDamageAmount		= 120.0f;

DEBUGVAR float dbg_fDefaultMaxSuckDistance = 4.0f;

DEBUGVAR float dbg_fMeleeDamageImpulseStrength		= 10.0f;
DEBUGVAR float dbg_fFatalDamageAmount		= 1000000.0f;

static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONSTART		= 0xD8F7957D; // Asura_GetHashID("FirstPersonStart");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONHIT		= 0x0615EF6E; // Asura_GetHashID("FirstPersonHit");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONCOUNTERED	= 0x8FB29F36; // Asura_GetHashID("FirstPersonCountered");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONMISS		= 0xBCAA4481; // Asura_GetHashID("FirstPersonMiss");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONSTART		= 0x7AACDA66; // Asura_GetHashID("ThirdPersonStart");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONHIT		= 0x52F95A97; // Asura_GetHashID("ThirdPersonHit");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONCOUNTERED	= 0x19F9479F; // Asura_GetHashID("ThirdPersonCountered");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONMISS		= 0x0C343E78; // Asura_GetHashID("ThirdPersonMiss");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_MAXDISTANCE			= 0x86798D59; // Asura_GetHashID("MaxDistance");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_DAMAGE				= 0xB02E8B6F; // Asura_GetHashID("Damage");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_ATTACKIMPULSE		= 0x17FD95AD; // Asura_GetHashID("AttackImpulse");
static const Asura_Hash_ID uAVP_POUNCE_BLUEPRINT_PARAM_MINIMUMDROP			= 0x3C2007FD; // Asura_GetHashID("MinimumDrop");

DEBUGVAR float dbg_fPounceDefaultMinDrop = -5.0f;

/////////////////////////////////////////////////////////////////////
//
// Class :	AVP_Vertebrate_MeleeCounter_System
//
//////////////////////////////////////////////////////////////////////
const char* AVP_Vertebrate_MeleeCounter_System::s_szPounceAttackBaseParameterName = "PounceAttack";

const u_short AVP_Vertebrate_MeleeCounter_System::s_ausLosClassificationExceptions[NUM_LOS_CLASSIFICATION_EXCEPTIONS] =
	{
		AVP_ENTITYCLASS_HUMANFLARE,
		AVP_ENTITYCLASS_PROJECTILE_FLAMETHROWER,
		AVP_ENTITYCLASS_VENT,
	};

AVP_Vertebrate_MeleeCounter_System::AVP_Vertebrate_MeleeCounter_System(AVP_ServerActor_Vertebrate& xOwner)
: m_xOwner(xOwner)
, m_uMeleeQueueSize(0)
, m_uQueuedCounter(AVP_MELEEATTACK_TYPE_NONE)
, m_fDamageRange(dbg_fDefaultDamageRadius)
, m_fDamageAIBackAttackFactor(dbg_fDefaultDamageAIBackAttackFactor)
, m_fAttackDetectionRange(dbg_fDefaultAttackDetectionRadius)
, m_fTargetDetectionRange(dbg_fDefaultTargetDetectionRadius)
, m_fTargetDetectionCosAngle(dbg_fDefaultTargetDetectionCosAngle)
, m_fTargetDetectionOriginBackwardsOffset(dbg_fDefaultTargetDetectionOriginBackwardsOffset)
, m_fBlockDetectionCosAngle(dbg_fDefaultBlockDetectionCosAngle)
, m_fBlockDamageProportion(dbg_fDefaultBlockDamageProportion)
, m_fMinorDamageAmount(dbg_fDefaultMinorMeleeDamageAmount)
, m_fModerateDamageAmount(dbg_fDefaultModerateMeleeDamageAmount)
, m_fMajorDamageAmount(dbg_fDefaultMajorMeleeDamageAmount)
, m_fMaxSuckDistance(dbg_fDefaultMaxSuckDistance)
, m_uCurrentTarget(ASURA_GUID_UNREGISTERED)
, m_uRequestedTarget(ASURA_GUID_UNREGISTERED)
, m_xStoredTargetPosition(0.0f, 0.0f, 0.0f)
, m_uCurrentAttackType(AVP_MELEEATTACK_TYPE_NONE)
, m_uCurrentVariation(0)
, m_uCurrentGrappleBlueprint(ASURA_HASH_ID_UNSET)
, m_bCurrentMoveForceMiss(false)
, m_fCurrentDmgMultiplier(1.0f)
, m_uNumHitEntities(0)
, m_uFlags(0)
, m_fActiveGrappleAttackerValue(0.0f)
, m_fActiveGrappleVictimValue(0.0f)
, m_fMinPounceAttackDistance(0.0f)
, m_uNumPounceAttackVariations(0)
, m_pxPounceAttacks(NULL)
, m_uNumLeftQuickAttackVariations(1)
, m_uNumRightQuickAttackVariations(1)
, m_eNextQuickAttackSide(QUICK_ATTACK_RIGHT)
{ 
}

AVP_Vertebrate_MeleeCounter_System::~AVP_Vertebrate_MeleeCounter_System()
{
	delete[] m_pxPounceAttacks;
}

void AVP_Vertebrate_MeleeCounter_System::InitialiseFromBlueprint(const Asura_Blueprint& xBlueprint)
{
	Asura_Value	xValue;

	if (xBlueprint.GetParameterValueByHashID(0x1D13BD3A, xValue)) //Asura_GetHashID("MeleeDamageRange")
	{
		m_fDamageRange = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0xd79c4f49, xValue)) //Asura_GetHashID("MeleeDamageAIBackAttackFactor")
	{
		m_fDamageAIBackAttackFactor = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x605AA724, xValue)) //Asura_GetHashID("MeleeStandardAttackRange")
	{
		m_fAttackDetectionRange = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x2731651D, xValue)) //Asura_GetHashID("MeleeTargetDetectionRange")
	{
		m_fTargetDetectionRange = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x2647A4B3, xValue)) //Asura_GetHashID("MeleeTargetDetectionAngle")
	{
		m_fTargetDetectionCosAngle = Asura_Maths::Cosine(Asura_Maths::DegToRad(xValue.GetFloatValue()));
	}
	if (xBlueprint.GetParameterValueByHashID(0x8A246D7D, xValue)) //Asura_GetHashID("MeleeTargetDetectionOriginBackwardsOffset")
	{
		m_fTargetDetectionOriginBackwardsOffset = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x9D483947, xValue)) //Asura_GetHashID("MeleeBlockDetectionAngle")
	{
		m_fBlockDetectionCosAngle = Asura_Maths::Cosine(Asura_Maths::DegToRad(xValue.GetFloatValue()));
	}
	if (xBlueprint.GetParameterValueByHashID(0xDB22F559, xValue)) //Asura_GetHashID("MeleeBlockDamageTakenPercentage")
	{
		m_fBlockDamageProportion = xValue.GetFloatValue() * 0.01f;
	}
	if (xBlueprint.GetParameterValueByHashID(0x33CD609A, xValue)) //Asura_GetHashID("MeleeDamageAmountMinor")
	{
		m_fMinorDamageAmount = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x74C61DFE, xValue)) //Asura_GetHashID("MeleeDamageAmountModerate")
	{
		m_fModerateDamageAmount = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x33C9AE9E, xValue)) //Asura_GetHashID("MeleeDamageAmountMajor")
	{
		m_fMajorDamageAmount = xValue.GetFloatValue();
	}
	if (xBlueprint.GetParameterValueByHashID(0x2BA3B7AF, xValue)) //Asura_GetHashID("MeleeMaxSuckDistance")
	{
		m_fMaxSuckDistance = xValue.GetFloatValue();
	}
}

void AVP_Vertebrate_MeleeCounter_System::PostLoadInitialise()
{
	const Asura_Blueprint* pxBlueprint = GetExtraMeleeDataBlueprint();
	if (pxBlueprint)
	{
		Asura_Value xValue;

		const char* szLeftQuickAttackBaseParameterName;
		const char* szRightQuickAttackBaseParameterName;
		if (GetQuickAttackBaseParameterNames(szLeftQuickAttackBaseParameterName, szRightQuickAttackBaseParameterName))
		{
			Asura_Hash_ID uBlueprintID = ASURA_HASH_ID_UNSET;
			if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_MELEE_FIRSTPERSON, xValue))
			{
				uBlueprintID = xValue.GetHashIDValue();
			}

			if ((uBlueprintID == ASURA_HASH_ID_UNSET) && pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_MELEE_THIRDPERSON, xValue))
			{
				uBlueprintID = xValue.GetHashIDValue();
			}

			if (uBlueprintID != ASURA_HASH_ID_UNSET)
			{
				const Asura_Blueprint* pxMeleeBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uBlueprintID);
				if (pxMeleeBlueprint)
				{
					m_uNumLeftQuickAttackVariations = pxMeleeBlueprint->CountEnumeratedParameters(szLeftQuickAttackBaseParameterName, 2);
					m_uNumRightQuickAttackVariations = pxMeleeBlueprint->CountEnumeratedParameters(szRightQuickAttackBaseParameterName, 2);
				}
			}
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_MINPOUNCEDISTANCE, xValue))
		{
			m_fMinPounceAttackDistance = xValue.GetFloatValue();
		}

		AVP_Blueprint_ParameterEnumerator_PounceVariations_MeleeCounterData xEnumerator(GetPounceBlueprintType(), *this);
		pxBlueprint->EnumerateParameters(GetPounceAttackBaseParameterName(), 2, xEnumerator);
	}
}

bool AVP_Vertebrate_MeleeCounter_System::HandleMessage( const Asura_Message* pxMessage )
{
	const u_short usMsgID = pxMessage->GetMsgID();

	if (usMsgID == AVP_MSGID_REQUEST_MELEE || usMsgID == AVP_MSGID_REQUEST_MELEE_EX)
	{
		const AVP_Message_Vertebrate_RequestMelee* pxData = static_cast<const AVP_Message_Vertebrate_RequestMelee*>(pxMessage->GetData());

		const u_int uType = pxData->GetType();
		const Asura_Guid uTargetGuid = pxData->GetTargetGuid();

		bool bForceMiss = false;
		float fDmgMultiplier = 1.0f;

		if (usMsgID == AVP_MSGID_REQUEST_MELEE_EX )
		{
			const AVP_Message_Vertebrate_RequestMelee_Ex* pxDataEx = static_cast<const AVP_Message_Vertebrate_RequestMelee_Ex*>(pxData);
			bForceMiss =  pxDataEx->GetForceMiss();
			fDmgMultiplier = pxDataEx->GetDamageMultiplier();
		}

		if (GetAttackTypeCanCounter(uType))
		{
			// If this can counter, empty the queue and flag for a queued counter.
			m_uQueuedCounter = uType;

			// Check the special melee types.
			if (uType == AVP_MELEEATTACK_TYPE_BLOCK)
			{
				m_bBlockQueued = true;
			}
			else if (uType == AVP_MELEEATTACK_TYPE_UNBLOCK)
			{
				m_bUnblockQueued = true;
			}
			// Otherwise, add this to the queue.
			else
			{
				ClearAttackRequests();
				m_axQueuedMeleeAttacks[0].m_uType = uType;
				m_axQueuedMeleeAttacks[0].m_uTarget = uTargetGuid;
				m_axQueuedMeleeAttacks[0].m_bForceMiss = bForceMiss;
				m_axQueuedMeleeAttacks[0].m_fDmgMultiplier = fDmgMultiplier;

				m_uMeleeQueueSize = 1;
				RequestTarget(uTargetGuid);
				// Make sure this gets set.
				if (uTargetGuid != ASURA_GUID_UNREGISTERED)
				{
					UpdateCurrentTarget();
				}
			}
		}
		else
		{
			// Check the special melee types.
			if (uType == AVP_MELEEATTACK_TYPE_BLOCK)
			{
				m_bBlockQueued = true;
			}
			else if (uType == AVP_MELEEATTACK_TYPE_UNBLOCK)
			{
				m_bUnblockQueued = true;
			}
			// Otherwise, add this to the queue, if we can.
			else if (m_uMeleeQueueSize < MAX_QUEUED_MELEE_ATTACKS)
			{
				m_axQueuedMeleeAttacks[m_uMeleeQueueSize].m_uType = uType;
				m_axQueuedMeleeAttacks[m_uMeleeQueueSize].m_uTarget = uTargetGuid;
				m_axQueuedMeleeAttacks[m_uMeleeQueueSize].m_bForceMiss = bForceMiss;
				m_axQueuedMeleeAttacks[m_uMeleeQueueSize].m_fDmgMultiplier = fDmgMultiplier;
				++m_uMeleeQueueSize;
			}
		}
			
		return true;
	}


	if (usMsgID == AVP_MSGID_ACTIVEGRAPPLE_VALUE)
	{
		Asura_Assert(pxMessage->From == m_xOwner.GetGuid() || pxMessage->From == m_uCurrentTarget, "Active grapple result from a wrong entity.");

		// We need to find out if we're the attacker or the victim.
		const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
		if (pxClient)
		{
			const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
			if (pxMovementSystem)
			{
				const Asura_Move* pxCurrentMove = pxMovementSystem->GetCurrentMove();
				if (pxCurrentMove)
				{
					bool bValueUsed = false;

					// The puppet is always the victim.
					const bool bVictim = pxCurrentMove->GetHash() == AVP_Move_Puppet::GetStaticHash();

					if (bVictim)
					{
						// If we're the victim, we need to pass this result on to our attacker.
						const AVP_Move_Puppet* pxPuppetMove = static_cast<const AVP_Move_Puppet*>(pxCurrentMove);
						const Asura_Guid uAttackerGuid = pxPuppetMove->GetCurrentPuppeteerGuid();
						AVP_ServerActor_Vertebrate* pxAttacker = AVP_ServerActor_Vertebrate::FindEntity(uAttackerGuid);
						if (pxAttacker)
						{
							pxAttacker->HandleMessage(pxMessage);
						}
					}
					else
					{
						// As the the attacker, handle the value.
						const AVP_Message_Vertebrate_ActiveGrappleValue* pxData = static_cast<const AVP_Message_Vertebrate_ActiveGrappleValue*>(pxMessage->GetData());
						const float fValue = pxData ? pxData->GetValue() : 0.0f;
						const bool bFromVictim = pxMessage->From != m_xOwner.GetGuid();

						if (bFromVictim)
						{
							if (!m_bActiveGrappleReceivedVictimValue)
							{
								m_fActiveGrappleVictimValue = fValue;
								m_bActiveGrappleReceivedVictimValue = true;
								bValueUsed = true;
							}
						}
						else
						{
							if (!m_bActiveGrappleReceivedAttackerValue)
							{
								m_fActiveGrappleAttackerValue = fValue;
								m_bActiveGrappleReceivedAttackerValue = true;
								bValueUsed = true;
							}
						}
					}

					if (bValueUsed)
					{
						// Send the result immediately, even if we haven't yet received the second result. This allows us to have a prediction of what the result will be.
						ApplyActiveGrappleResult();
					}
				}
			}
		}

		return true;
	}

	return false;
}

void AVP_Vertebrate_MeleeCounter_System::Update( void )
{
	// This initial version uses the clients on the server to work out whether attacks and counters are possible.
	// If this turns out to be too unresponsive in multiplayer, we have to consider moving a lot of the melee functionality to the client (at least for players), leaving the server as basically an arbitrator only.
	const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
	if (!pxClient) { return; }

	const AVP_Movement_System* pxMovementSystem = static_cast<const AVP_Movement_System*>(pxClient->GetMovementSystem());
	if (!pxMovementSystem) { return; }

	u_int uPreviousAttackType = m_uCurrentAttackType;
	Asura_Guid uPreviousTarget = m_uCurrentTarget;
	bool bBlockPreviouslyActive = IsBlockActive();

	bool bAttackPermitted = true;
	bool bBlockPermitted = true;
	bool bCounterPermitted = true;

	u_int uMoveAnimFlags = 0;
	const AVP_Move* pxCurrentMove = static_cast<const AVP_Move*>(pxMovementSystem->GetCurrentMove());
	if (pxCurrentMove)
	{
		if (pxCurrentMove->GetShouldClearMeleeRequestQueue())
		{
			ClearAttackRequests();
		}
		if (pxCurrentMove->GetShouldClearCounterRequest())
		{
			m_uQueuedCounter = AVP_MELEEATTACK_TYPE_NONE;
		}
		if (m_bAlwaysUpdateTarget && pxCurrentMove->GetAllowMeleeTargetUpdate())
		{
			UpdateCurrentTarget();
		}

		uMoveAnimFlags = pxCurrentMove->GetAnimFlagsForLastFrame();

		bAttackPermitted = ((uMoveAnimFlags & AVP_ANIMFLAG_ALLOW_NEW_MELEE) != 0);
		bBlockPermitted = ((uMoveAnimFlags & AVP_ANIMFLAG_ALLOW_NEW_BLOCK) != 0) || pxCurrentMove->GetAllowsBlock();
		bCounterPermitted = ((uMoveAnimFlags & AVP_ANIMFLAG_ALLOW_NEW_COUNTER) != 0);

		// The block allowed anim flag can override the move's allow block flag.
		if (!bBlockPermitted)
		{
			BreakBlock();
		}

		// Handling for randomising the result of difficulty-free grapples
		if (((uMoveAnimFlags & AVP_ANIMFLAG_SHOW_MELEE_COUNTERBAR) != 0) && m_bActiveGrappleReceivedAttackerValue && m_bActiveGrappleReceivedVictimValue)
		{
			ApplyActiveGrappleResult();
		}

		if(!m_bCurrentMoveForceMiss && !m_bCurrentAttackCountered && ((uMoveAnimFlags & AVP_ANIMFLAG_MELEE_DAMAGE) != 0))
		{
			u_int uPreviousHitEntities = m_uNumHitEntities;
			CheckForAndDealMeleeDamage();
			if ((uPreviousHitEntities == 0) && (m_uNumHitEntities > 0))
			{
				// First hit, let our entity know so that the moves play the correct anim.
				SendMeleeAttackHitMessage();
			}
		}
		else if (!m_bCurrentAttackCountered && ((uMoveAnimFlags & AVP_ANIMFLAG_GRAPPLE_APPLY_DAMAGE) != 0))
		{
			ApplyGrappleDamage(m_uCurrentGrappleBlueprint);
		}
		else
		{
			// Clear melee hits, as the current attack is complete. Done here in case one move can hit twice.
			m_uNumHitEntities = 0;
			m_bCurrentAttackHasHitEnv = false;
		}
	}
	else
	{
		// Keep updating the target if we're not in a move.
		if (m_bAlwaysUpdateTarget)
		{
			UpdateCurrentTarget();
		}

		// Clear melee hits, as the current attack is complete.
		m_uNumHitEntities = 0;
		m_bCurrentAttackHasHitEnv = false;
		m_uCurrentAttackType = AVP_MELEEATTACK_TYPE_NONE;
		m_uCurrentGrappleBlueprint = ASURA_HASH_ID_UNSET;
	}

	const bool bCanAttackOrCounter = bAttackPermitted || bCounterPermitted;
	const bool bHasAttackOrCounter = (m_uQueuedCounter != AVP_MELEEATTACK_TYPE_NONE) || (m_uMeleeQueueSize > 0);

	if (bCanAttackOrCounter && bHasAttackOrCounter)
	{
		const AVP_ClientEntity_Actor* pxTarget = AVP_ClientEntity_Actor::FindEntity(m_uCurrentTarget);

		AVP_Inventory_System* pxMasterInventory = m_xOwner.GetMasterInventory();
		AVP_Inventory_Item* pxWeapon = pxMasterInventory ? pxMasterInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL;
		const AVP_Inventory_Item* pxStim = pxMasterInventory ? pxMasterInventory->GetItem(uAVP_BLUEPRINT_PARAM_STIM) : NULL;
		const bool bStimIdling = pxStim ? pxStim->IsIdling(true) : true;
		const bool bHasWeapon = pxWeapon != NULL;
		const bool bIsWeaponIdling = bHasWeapon && pxWeapon->IsIdling(true);
		const bool bIsUsingWeapon = bHasWeapon && pxWeapon->IsUsing(false);
		const bool bWeaponInMeleeUsage = bIsUsingWeapon && (pxWeapon->GetCurrentUsageType() == AVP_InventoryItem_Weapon::UT_MELEE);
		const bool bCanChangeUsage = pxWeapon ? pxWeapon->GetCanUseUsageType(AVP_InventoryItem_Weapon::UT_MELEE, AVP_Inventory_Item::TT_NONE) : true;

		if (bCounterPermitted && (m_uQueuedCounter != AVP_MELEEATTACK_TYPE_NONE))
		{
			if (bStimIdling && bCanChangeUsage && (!bHasWeapon || bIsWeaponIdling || bWeaponInMeleeUsage || !pxClient->IsPlayer()))
			{
				// See if we can actually counter anything.
				// In the first version, we can only counter our current target, and even then only if they're attacking us.
				if (pxTarget)
				{
					u_int uAnimFlags = pxTarget->GetFlagsForAllAnimsLastFrame();

					if ((uAnimFlags & AVP_ANIMFLAG_MELEE_COUNTER) != 0)
					{
						const AVP_ServerActor_Vertebrate* pxServerTarget = AVP_ServerActor_Vertebrate::FindEntity(m_uCurrentTarget);
						if (pxServerTarget && (pxServerTarget->GetCurrentMeleeTarget() == m_xOwner.GetGuid()))
						{
							// OK, so the anim is at the correct point, and they are attacking us. Let's get countering.
							if (PerformCounter())
							{
								// By entering a counter, we can no longer trigger an attack this frame.
								bAttackPermitted = false;
								bBlockPermitted = false;

								// Tell the weapon to start being used.
								if (bHasWeapon && !bWeaponInMeleeUsage)
								{
									Asura_Assert(!bIsUsingWeapon, "Should not be using a weapon already.");
									pxWeapon->RequestBeginUsage(AVP_InventoryItem_Weapon::UT_MELEE);
								}
							}
						}
					}
				}
			}

			if (pxClient->IsPlayer() || bIsWeaponIdling)
			{
				// Counter requests are always used up as soon as possible.
				m_uQueuedCounter = AVP_MELEEATTACK_TYPE_NONE;
			}
		}

		if (bAttackPermitted)
		{
			while (m_uMeleeQueueSize > 0)
			{
				// Set up the requested target for this attack.
				RequestTarget(m_axQueuedMeleeAttacks[0].m_uTarget);
				UpdateCurrentTarget();
				RequestForceMissAndDmgMultiplier(m_axQueuedMeleeAttacks[0].m_bForceMiss, m_axQueuedMeleeAttacks[0].m_fDmgMultiplier);

				u_int uType = m_axQueuedMeleeAttacks[0].m_uType;

				bool bAttackStarted = false;
				if (bStimIdling && bCanChangeUsage && (!bHasWeapon || bIsWeaponIdling || bWeaponInMeleeUsage))
				{
					bAttackStarted = StartMeleeMove(uType);
				}

				--m_uMeleeQueueSize;

				for (u_int u=0; u<m_uMeleeQueueSize; ++u)
				{
					m_axQueuedMeleeAttacks[u] = m_axQueuedMeleeAttacks[u+1];
				}


				if (bAttackStarted)
				{
					bBlockPermitted = false;

					if (uPreviousAttackType != AVP_MELEEATTACK_TYPE_NONE)
					{
						SendIncomingMeleeFinishedMessage(uPreviousTarget);
					}
					// Send a message to our current target telling them we're trying to hit them. Allows AIs to respond in some way.
					if (GetSendIncomingMeleeMessage(m_uCurrentAttackType))
					{
						SendIncomingMeleeStartedMessage(m_uCurrentTarget);
					}
					// Tell the weapon to start being used.
					if (bHasWeapon && !bWeaponInMeleeUsage)
					{
						Asura_Assert(!bIsUsingWeapon, "Should not be using a weapon already.");
						pxWeapon->RequestBeginUsage(AVP_InventoryItem_Weapon::UT_MELEE);
					}

					InitialiseActiveGrapple();

					break;
				}
				else
				{
					SendMeleeFailedToStartMessage(uType);
				}

			}
		}
	}

	// Finally, update blocking states.
	if (bBlockPermitted && m_bBlockQueued && GetCanBlock())
	{
		SetBlockActive(true);
	}
	else if (m_bUnblockQueued && GetCanUnblock())
	{
		SetBlockActive(false);
	}

	// Always use up block requests.
	m_bBlockQueued = false;
	m_bUnblockQueued = false;

	// Check for melee ending
	const bool bMeleeBecomeInactive = (uPreviousAttackType != AVP_MELEEATTACK_TYPE_NONE) && (m_uCurrentAttackType == AVP_MELEEATTACK_TYPE_NONE);
	const bool bBlockBecomeInactive = bBlockPreviouslyActive && !IsBlockActive();

	if (bMeleeBecomeInactive || bBlockBecomeInactive)
	{
		if (bMeleeBecomeInactive)
		{
			// Just completed an attack, tell our target that's happened.
			if (m_uCurrentTarget != ASURA_GUID_UNREGISTERED)
			{
				SendIncomingMeleeFinishedMessage(m_uCurrentTarget);
			}
		}

		// And update our current weapon to let it know we're no longer using it.
		AVP_Inventory_System* pxMasterInventory = m_xOwner.GetMasterInventory();
		AVP_Inventory_Item* pxWeapon = pxMasterInventory ? pxMasterInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL;

		if (pxWeapon)
		{
			Asura_Assert(pxWeapon->GetCurrentUsageType() == AVP_InventoryItem_Weapon::UT_MELEE, "Weapon usage type should be melee.");
			pxWeapon->RequestEndUsage();
		}
	}
}


bool AVP_Vertebrate_MeleeCounter_System::IsTargetInDamageRange() const
{
	const AVP_ServerEntity_PhysicalObject* pxTarget = AVP_ServerEntity_PhysicalObject::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return false; }

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;

	return (xOffset.MagnitudeSquared() <= (m_fDamageRange * m_fDamageRange));
}

bool AVP_Vertebrate_MeleeCounter_System::GetHasAttackTarget() const
{
	// Only actors can be attack targets.
	const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return false; }

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;

	return (xOffset.MagnitudeSquared() <= (m_fAttackDetectionRange * m_fAttackDetectionRange));
}


void AVP_Vertebrate_MeleeCounter_System::RequestTarget(Asura_Guid uRequestedTarget)
{
	// For now, we'll just store the request.
	m_uRequestedTarget = uRequestedTarget;
}

void AVP_Vertebrate_MeleeCounter_System::RequestForceMissAndDmgMultiplier(bool bForceMiss, float fDmgMultiplier)
{
	// For now, we'll just store the request.
	m_bCurrentMoveForceMiss = bForceMiss;
	m_fCurrentDmgMultiplier = fDmgMultiplier;
}

void AVP_Vertebrate_MeleeCounter_System::ClearAttackRequests()
{
	for (u_int u = 0; u < m_uMeleeQueueSize; ++u)
	{
		SendMeleeFailedToStartMessage(m_axQueuedMeleeAttacks[u].m_uType);
	}

	m_uMeleeQueueSize = 0;
}

bool AVP_Vertebrate_MeleeCounter_System::TestBlock(const AVP_ServerEntity_Actor& xAttacker) const
{
	if (!IsBlockActive()) { return false; }

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xToAttacker;
	xAttacker.GetMeleeOrigin(xToAttacker);
	xToAttacker -= xOwnerOrigin;

	const Asura_Vector_3& xAimDirection = m_xOwner.GetAimDirection();

	// Make sure they are in front of us before the next stage. Can get away without normalising anything at this point as we're just checking sign. Need to use our character's down direction though, not world down.
	const Asura_Matrix_3x3& xOri = m_xOwner.GetOrientation();
	const Asura_Vector_3 xDown(xOri.e12, xOri.e22, xOri.e32);
	Asura_Vector_3 xAimWithoutVertical = xAimDirection - (xDown * (xAimDirection * xDown));
	if ((xToAttacker * xAimWithoutVertical) < 0.0f) { return false; }

	// Offset the to-attacker vector by the backwards offset, as with our own attacks.
	xToAttacker += xAimDirection * m_fTargetDetectionOriginBackwardsOffset;
	xToAttacker.Normalise();

	return (xToAttacker * xAimDirection) >= m_fBlockDetectionCosAngle;
}


bool AVP_Vertebrate_MeleeCounter_System::GetCurrentAttackCanBeCountered() const
{
	// Currently, this is based on having a current attack, and our client entity having the can-be-countered anim flag.
	if (m_uCurrentAttackType == AVP_MELEEATTACK_TYPE_NONE) { return false; }

	// TODO - Is this sensible? Fixes some issues, but maybe some sweeping attacks should be able to strike several enemies and then be countered by another.
	if (m_uNumHitEntities > 0) { return false; }

	const AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity(m_xOwner.GetGuid());
	if (!pxClient) { return false; }

	const u_int uAnimFlags = pxClient->GetFlagsForAllAnimsLastFrame();
	return (uAnimFlags & AVP_ANIMFLAG_MELEE_COUNTER) != 0;
}


bool AVP_Vertebrate_MeleeCounter_System::GetCanStartNewAttack() const
{
	if (m_uCurrentAttackType == AVP_MELEEATTACK_TYPE_NONE) { return true; }

	// Check the anim flag too
	const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
	if (!pxClient) { return true; }

	const AVP_Movement_System* pxMovementSystem = static_cast<const AVP_Movement_System*>(pxClient->GetMovementSystem());
	if (!pxMovementSystem) { return true; }

	const Asura_Move* pxCurrentMove = pxMovementSystem->GetCurrentMove();
	if (!pxCurrentMove) { return true; }

	const u_int uAnimFlags = pxCurrentMove->GetAnimFlagsForLastFrame();
	return (uAnimFlags & AVP_ANIMFLAG_ALLOW_NEW_MELEE) != 0;
}



bool AVP_Vertebrate_MeleeCounter_System::StartMeleeMove(u_int uType)
{
	Asura_Hash_ID uMoveHash = GetMeleeMoveHash(uType);
	if (uMoveHash == ASURA_HASH_ID_UNSET) { return false; }

	m_uCurrentAttackType = uType;
	m_uNumHitEntities = 0;
	m_bCurrentAttackCountered = false;
	m_bCurrentAttackHasHitEnv = false;
	m_bCurrentAttackShouldHit = IsCurrentTargetInSuckRange();

	Asura_Hash_ID uGrappleSetBlueprintParam = GetMeleeMoveGrappleSetBlueprintParam(uType);

	if (uGrappleSetBlueprintParam == ASURA_HASH_ID_UNSET)
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Msg_Melee_Move));

		if (pxMessage)
		{
			AVP_Msg_Melee_Move* pxData = static_cast<AVP_Msg_Melee_Move*>(pxMessage->GetData());

			m_uCurrentVariation = PickMeleeMoveVariation(uMoveHash);

			pxData->ClearFlags();
			pxData->SetVariation(m_uCurrentVariation);
			pxData->SetPendingAttackType(uType);
			pxData->SetTargetGuid(m_uCurrentTarget); // Not all moves will need to know this, but a fair number will.
			pxData->SetTargetEntityIsPuppet(GetMeleeMoveStartsPuppet(uMoveHash));
			pxData->SetMoveID(uMoveHash);
			pxData->SetIgnoredByOwner(false);
			pxData->SetSuckPermitted( m_xOwner.IsPlayer() ? IsCurrentTargetInSuckRange() : false); // AI never want to use suck mechanism

			pxMessage->SetToGuid(m_xOwner.GetGuid());
			pxMessage->SetFromGuid(m_xOwner.GetGuid());

			Asura_Network::Send(pxMessage);
		}
	}
	else
	{
		const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
		m_uCurrentGrappleBlueprint = GetGrappleBlueprintFromGrappleSetParameter(pxTarget, uGrappleSetBlueprintParam);

		if (m_uCurrentGrappleBlueprint != ASURA_HASH_ID_UNSET)
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Msg_Grapple_Move));

			if (pxMessage)
			{
				AVP_Msg_Grapple_Move* pxData = static_cast<AVP_Msg_Grapple_Move*>(pxMessage->GetData());
				if (pxData)
				{
					pxData->ClearFlags();
					pxData->SetMoveID(uMoveHash);
					pxData->SetIgnoredByOwner(false);
					pxData->SetTargetGuid(m_uCurrentTarget);
					pxData->SetTargetEntityIsPuppet(true); // We must be setting up a puppet if using a grapple.
					pxData->SetGrappleDescriptionBlueprint(m_uCurrentGrappleBlueprint);
				}

				pxMessage->SetToGuid(m_xOwner.GetGuid());
				pxMessage->SetFromGuid(m_xOwner.GetGuid());

				Asura_Network::Send(pxMessage);
			}
		}
	}

	m_bBlockActive = false;

	return true;
}


bool AVP_Vertebrate_MeleeCounter_System::PerformCounter()
{
	const AVP_ServerActor_Vertebrate* pxTarget = AVP_ServerActor_Vertebrate::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return false; }

	const AVP_Vertebrate_MeleeCounter_System* pxTargetMelee = pxTarget->GetMeleeSystem();
	if (!pxTargetMelee) { return false; }

	Asura_Hash_ID uCounterMoveHash;
	u_int uNewAttackType;
	bool bMakeEnemyIntoPuppet;
	Asura_Hash_ID uGrappleSetBlueprintParam;
	if (!pxTargetMelee->GetCanCounter(m_xOwner, m_uQueuedCounter, uCounterMoveHash, uNewAttackType, bMakeEnemyIntoPuppet, uGrappleSetBlueprintParam)) { return false; }

	// We're starting a counter, which may put us into a particular attack. Set that up.
	m_uCurrentAttackType = uNewAttackType;
	m_uCurrentGrappleBlueprint = ASURA_HASH_ID_UNSET;
	m_uNumHitEntities = 0;
	m_bCurrentAttackCountered = false;
	m_bCurrentAttackHasHitEnv = false;
	m_bBlockActive = false;

	// Start a counter move, if there is one
	if (uCounterMoveHash != ASURA_HASH_ID_UNSET)
	{
		if (uGrappleSetBlueprintParam == ASURA_HASH_ID_UNSET)
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Msg_Melee_Move));

			if (pxMessage)
			{
				AVP_Msg_Melee_Move* pxData = static_cast<AVP_Msg_Melee_Move*>(pxMessage->GetData());

				pxData->ClearFlags();
				pxData->SetVariation(0); // TODO - Do we need variations here?
				pxData->SetPendingAttackType(uNewAttackType);
				pxData->SetTargetGuid(m_uCurrentTarget); // Not all counters will need to know this, but some will.
				pxData->SetMoveID(uCounterMoveHash);
				pxData->SetTargetEntityIsPuppet(bMakeEnemyIntoPuppet);
				pxData->SetIgnoredByOwner(false);

				pxMessage->SetToGuid(m_xOwner.GetGuid());
				pxMessage->SetFromGuid(m_xOwner.GetGuid());

				Asura_Network::Send(pxMessage);
			}
		}
		else
		{
			m_uCurrentGrappleBlueprint = GetGrappleBlueprintFromGrappleSetParameter(pxTarget, uGrappleSetBlueprintParam);

			if (m_uCurrentGrappleBlueprint == ASURA_HASH_ID_UNSET) { return false; }

			Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Msg_Grapple_Move));

			if (pxMessage)
			{
				AVP_Msg_Grapple_Move* pxData = static_cast<AVP_Msg_Grapple_Move*>(pxMessage->GetData());
				if (pxData)
				{
					pxData->ClearFlags();
					pxData->SetMoveID(uCounterMoveHash);
					pxData->SetIgnoredByOwner(false);
					pxData->SetTargetGuid(m_uCurrentTarget);
					pxData->SetTargetEntityIsPuppet(true); // We must be setting up a puppet if using a grapple.
					pxData->SetGrappleDescriptionBlueprint(m_uCurrentGrappleBlueprint);
				}

				pxMessage->SetToGuid(m_xOwner.GetGuid());
				pxMessage->SetFromGuid(m_xOwner.GetGuid());

				Asura_Network::Send(pxMessage);
			}

			InitialiseActiveGrapple();
		}
	}

	// Also let the enemy know of the counter.
	Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_COUNTER_MELEE);

	if (pxMessage)
	{
		pxMessage->SetToGuid(m_uCurrentTarget);
		pxMessage->SetFromGuid(m_xOwner.GetGuid());

		Asura_Network::Send(pxMessage);
	}

	return true;
}

DEBUGVAR float dbg_fMeleeLOSConeAngle = Asura_Maths::Pi * 0.05f;
DEBUGVAR float dbg_fMeleeLOSRangeModifier = 0.75f; // We usually allow attacks to hit further than they should so they connect more reliably. We don't want to do that without lock-on, as it just looks silly when you hit walls from several metres away.
DEBUGVAR float dbg_fDamageMovementQueryRangeModifier = 0.25f;

void AVP_Vertebrate_MeleeCounter_System::CheckForAndDealMeleeDamage()
{
	if (m_uNumHitEntities >= MAX_HIT_MELEE_ENTITIES) { return; }

	const bool bNoTarget = (m_uCurrentTarget == ASURA_GUID_UNREGISTERED);
	const AVP_ServerEntity_Actor* pxActorTarget = bNoTarget ? NULL : AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
	const bool bTargetInDamageRange = bNoTarget ? false : IsTargetInDamageRange();
	const bool bNonActorTarget = bNoTarget ? false : (pxActorTarget == NULL);
	const bool bDoLOSTest = (bNoTarget || !bTargetInDamageRange || bNonActorTarget) && m_xOwner.IsPlayer() && (!m_bCurrentAttackShouldHit || !IsCurrentTargetInSuckRange());
	const bool bLOSTestCanHitEntities = bNoTarget || bNonActorTarget || (!bTargetInDamageRange && !GetAttackOnlyHitsCurrentTarget());

	if (bDoLOSTest)
	{
		// Special case - if we have no target, fire a LOS forwards and see if we can hit any ragdolls.
		Asura_Vector_3 xHeadPos;
		Asura_Matrix_3x3 xHeadOri;
		if (!m_xOwner.GetWorldSpaceMarkerPoint(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_EYES), &xHeadPos, &xHeadOri)) { return; }

		Asura_Vector_3 xDir(xHeadOri.e13, xHeadOri.e23, xHeadOri.e33);
		// Add a touch of jitter, so that multi-frame attacks have more chance of missing thin blockages
		xDir = xDir.ChooseRandomDirectionInCone(dbg_fMeleeLOSConeAngle);
		Asura_LOS_InputData xInput(xHeadPos, xDir, m_xOwner.GetGuid(), m_fDamageRange * dbg_fMeleeLOSRangeModifier, false);
		const Asura_LOS_OutputData* pxOutput = DoMeleeLOSTest(&xInput);

		if (pxOutput->m_bHitEnvironment)
		{
			if (m_bCurrentAttackHasHitEnv) { return; }

			m_bCurrentAttackHasHitEnv = true;

			// Send an environmental damage message.
			Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_APPLY_TO_ENVIRONMENT, sizeof(AVP_Message_EntityDamage_ApplyToEnvironment));
			if (pxMessage)
			{
				const Asura_LOS_OutputData& xLOSOutput = *pxOutput;

				const Asura_Vector_3 xPointOfImpact(xHeadPos + (xDir * xLOSOutput.m_fDistanceToIntersection));

				const AVP_EntityDamage_ApplyToEnvironment xDamage(GetAttackDamageType(), m_xOwner.GetGuid(), xPointOfImpact, xDir, xLOSOutput.m_xNormal, xLOSOutput.m_uMaterialResponseHashID, xLOSOutput.m_usPolyFlags, xLOSOutput.m_uEnvironmentFaceIndex);

				AVP_Message_EntityDamage_ApplyToEnvironment* pxData = static_cast<AVP_Message_EntityDamage_ApplyToEnvironment*>(pxMessage->GetData());

				pxData->SetDamage(xDamage);

				pxMessage->SetFromGuid(m_xOwner.GetGuid());
				pxMessage->SetToGuid(ASURA_GUID_ALL_CLIENTMACHINES);
				Asura_Network::Send(pxMessage);
			}

			return;
		}

		// If we can't hit entities, we're done.
		if (!bLOSTestCanHitEntities) { return; }

		const Asura_Guid uHitEntity = pxOutput->m_uHitEntity;
		if (!uHitEntity || (uHitEntity == ASURA_GUID_UNREGISTERED)) { return; }

		for (u_int u=0; u<m_uNumHitEntities; ++u)
		{
			if (m_auHitEntities[u] == uHitEntity) { return; }
		}

		// OK, that should be enough to allow us to knock the ragdoll around. Also, damage allies in the way.
		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
		if(pxMessage)
		{
			AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
			const AVP_EntityDamage_RequestVictim	xDamage(GetAttackDamageType(),
															m_xOwner.GetGuid(),
															uMELEE_MARKER_POINT_HASH,
															uHitEntity,
															GetAttackDamage(),
															xHeadPos + (xDir * pxOutput->m_fDistanceToIntersection),
															xDir,
															pxOutput->m_xNormal,
															GetAttackImpulseStrength(),
															pxOutput->m_iHitBoneID,
															false,
															m_xOwner.GetGuid(),
															pxOutput->m_iHitAttachmentID);
			pxData->SetDamage(xDamage);

			pxMessage->SetFromGuid(m_xOwner.GetGuid());
			pxMessage->SetToGuid(uHitEntity);
			Asura_Network::Send(pxMessage);
		}

		m_auHitEntities[m_uNumHitEntities++] = uHitEntity;

		return;
	}

	// We know we have a valid actor target by this point.
	if (m_bCurrentAttackShouldHit)
	{
		// A reasonable check for initial distance is whether we're still within suck range.
		if (!IsCurrentTargetInSuckRange())
		{
			return;
		}

		// Do a movement query towards the target and see if we can get close enough.
		const AVP_Actor_PhysicsObjectHandler* pxOwnerPhysHandler = m_xOwner.GetPhysicsObjectHandler();
		if (!pxOwnerPhysHandler)
		{
			// See if our client has it.
			const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
			if (pxClient)
			{
				pxOwnerPhysHandler = pxClient->GetPhysicsObjectHandler();
			}
		}
		const AVP_Actor_PhysicsObjectHandler* pxTargetPhysHandler = pxActorTarget->GetPhysicsObjectHandler();
		if (!pxTargetPhysHandler)
		{
			// See if its client has it.
			const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_uCurrentTarget);
			if (pxClient)
			{
				pxTargetPhysHandler = pxClient->GetPhysicsObjectHandler();
			}
		}

		if (!pxOwnerPhysHandler || !pxTargetPhysHandler)
		{
			return;
		}

		const Asura_Physics_Object* pxOwnerPhysObj = pxOwnerPhysHandler->GetPhysicsObject();
		const Asura_Physics_Object* pxTargetPhysObj = pxTargetPhysHandler->GetPhysicsObject();
		if (!pxOwnerPhysObj || !pxTargetPhysObj)
		{
			return;
		}
		const Asura_AMatrix3x3 xOri(m_xOwner.GetOrientation());

		// Aim just in front of the target
		Asura_AVector3 xStart(m_xOwner.GetPosition());
		Asura_AVector3 xTarget(pxActorTarget->GetPosition()); // Not the final target, but is the first part of calculating it.
		Asura_AVector3 xOffset = xTarget - xStart;
		Asura_AVector3 xFlatOffset(xOffset.GetX(), 0.0f, xOffset.GetZ());
		if (xFlatOffset.ApproximatelyEquals(Asura_Maths::ZeroAVector3))
		{
			xFlatOffset.SetX(xOri.GetElement(0, 2));
			xFlatOffset.SetZ(xOri.GetElement(2, 2));
			if (xFlatOffset.ApproximatelyEquals(Asura_Maths::ZeroAVector3))
			{
				xFlatOffset.SetZ(1.0f);
			}
		}
		xFlatOffset.Normalise();

		xTarget -= xFlatOffset * m_fDamageRange * dbg_fDamageMovementQueryRangeModifier;

		Asura_OBB xOBB(pxOwnerPhysObj->GetCollisionObject()->GetBoundingBox(), xOri);
		Asura_Physics_MovementQuery::ASURA_PMQ_COLLISION_PROXY eProxy = (pxOwnerPhysObj->GetCollisionProxy() == Asura_Physics_Object::COLLISION_PROXY_SPHERESTACK) ? Asura_Physics_MovementQuery::ASURA_PMQCP_SPHERESTACK : Asura_Physics_MovementQuery::ASURA_PMQCP_OBB;

		Asura_Physics_MovementQuery xQuery(xStart, xTarget - xStart, xOBB, eProxy, pxOwnerPhysObj->GetFlagsOfPolysToIgnore(), false, true, true, false, true, false, pxOwnerPhysObj->GetCollisionGroups(), pxOwnerPhysObj->GetCollisionGroupsToCollideWith());
		Asura_Physics_MovementQuery_Results xResults(xQuery);

		// See if the query took us within damage range.
		xOffset = xResults.GetEndPosition() - xTarget;
		if (xOffset.MagnitudeSquared() > (m_fDamageRange * m_fDamageRange))
		{
			return;
		}
	}
	else if (!bTargetInDamageRange)
	{
		return;
	}

	for (u_int u=0; u<m_uNumHitEntities; ++u)
	{
		if (m_auHitEntities[u] == m_uCurrentTarget) { return; }
	}

	ApplyMeleeDamage(m_uCurrentTarget);
}

void AVP_Vertebrate_MeleeCounter_System::ApplyMeleeDamage(Asura_Guid uDamagedGuid)
{
	const AVP_ServerEntity_PhysicalObject* pxTarget = AVP_ServerEntity_PhysicalObject::FindEntity(uDamagedGuid);

	// Calculate position and direction of damage data.
	if (pxTarget)
	{
		float fDamage = m_fCurrentDmgMultiplier * GetAttackDamage();
		if (fDamage > 0.0f)
		{
			// Check if the attack is blocked, and reduce damage accordingly.
			bool bBlocked = false;
			if (!IsAttackUnblockable())
			{
				// We need to get the entity as a vertebrate for this to work.
				const AVP_ServerActor_Vertebrate* pxTargetVertebrate = AVP_ServerActor_Vertebrate::FindEntity(uDamagedGuid);

				const AVP_Vertebrate_MeleeCounter_System* pxMeleeSystem = pxTargetVertebrate ? pxTargetVertebrate->GetMeleeSystem() : NULL;
				if (pxMeleeSystem && pxMeleeSystem->TestBlock(m_xOwner))
				{
					bBlocked = true;
					fDamage *= pxMeleeSystem->GetBlockDamageProportion();
				}
			}
	
			// Stealth attack detection
			bool bStealthDamage = false; 

			const Asura_Matrix_3x3& xTargetOrientM = pxTarget->GetOrientation();
			const Asura_Vector_3 xTargetForward( xTargetOrientM.e13, xTargetOrientM.e23, xTargetOrientM.e33 );
			const Asura_Matrix_3x3& xOurOrientM = m_xOwner.GetOrientation();
			const Asura_Vector_3 xOurForward( xOurOrientM.e13, xOurOrientM.e23, xOurOrientM.e33 );
			if ( xOurForward * xTargetForward > 0.0f )
			{
				if ( !m_xOwner.IsPlayer() )
				{
					fDamage *= GetAIBackAttackDamageFactor();
				}
				bStealthDamage = true; 
			}

			Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
			if(pxMessage)
			{
				AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());

				const float fImpulseStrength = GetAttackImpulseStrength();

				Asura_Vector_3 xPosition;
				const AVP_ClientEntity_PhysicalObject* pxClient = AVP_ClientEntity_PhysicalObject::FindEntity(uDamagedGuid);
				if (!pxClient || !pxClient->GetMarkerLocationFromHash(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_TARGET_2), &xPosition, NULL))
				{
					xPosition = pxTarget->GetPosition() + Asura_Vector_3(0.0f, -1.2f, 0.0f);
				}

				const int iBoneHit = GetAttackBoneHit(uDamagedGuid);

				Asura_Vector_3 xDirection;
				GetAttackDirection(pxTarget, xDirection);
				const Asura_Vector_3 xNormal(-xDirection);

				const AVP_EntityDamage_RequestVictim	xDamage(GetAttackDamageType(),
																m_xOwner.GetGuid(),
																uMELEE_MARKER_POINT_HASH,
																uDamagedGuid,
																fDamage,
																xPosition,
																xDirection,
																xNormal,
																fImpulseStrength,
																iBoneHit,
																bBlocked,
																m_xOwner.GetGuid(),
																-1,
																ASURA_HASH_ID_UNSET,
																bStealthDamage);

				// Inform the client regardless of whether we can be damaged...
				pxData->SetDamage(xDamage);

				pxMessage->SetFromGuid(m_xOwner.GetGuid());
				pxMessage->SetToGuid(uDamagedGuid);
				Asura_Network::Send(pxMessage);
			}
		}

		m_auHitEntities[m_uNumHitEntities++] = m_uCurrentTarget;
	}
}

void AVP_Vertebrate_MeleeCounter_System::ApplyGrappleDamage() const
{
	// First version just applies damage with a zero impulse.
	const AVP_ServerEntity_PhysicalObject* pxTarget = AVP_ServerEntity_PhysicalObject::FindEntity(m_uCurrentTarget);

	if (pxTarget)
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
		if(pxMessage)
		{
			AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());

			// TODO - May well need calculation of grapple impulse strength and direction, like with the normal attacks.
			const float				fDamage				= GetGrappleDamage();
			const float				fImpulseStrength	= 0.0f;
			const Asura_Vector_3&	xPosition			= pxTarget->GetPosition();

			const Asura_Vector_3	xNormal(0.0f, -1.0f, 0.0f);
			const Asura_Vector_3	xDirection(0.0f, 1.0f, 0.0f);

			{
				const AVP_EntityDamage_RequestVictim	xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_GRAPPLE,
																m_xOwner.GetGuid(),
																uMELEE_MARKER_POINT_HASH,
																m_uCurrentTarget,
																fDamage,
																xPosition,
																xDirection,
																xNormal,
																fImpulseStrength,
																-1,
																false,
																m_xOwner.GetGuid());

				// Inform the client regardless of whether we can be damaged...
				pxData->SetDamage(xDamage);

				pxMessage->SetFromGuid(m_xOwner.GetGuid());
				pxMessage->SetToGuid(m_uCurrentTarget);
				Asura_Network::Send(pxMessage);
			}
		}
	}
}

void AVP_Vertebrate_MeleeCounter_System::ApplyGrappleDamage(const Asura_Hash_ID uGrappleDescriptionID) const
{
	if (uGrappleDescriptionID == ASURA_HASH_ID_UNSET)
	{
		// Old system, go back to the old function.
		ApplyGrappleDamage();
	}
	else
	{
		// Get the correct damage value from the description.
		const Asura_Blueprint* pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, uGrappleDescriptionID);
		if (!pxGrappleDescription)
		{
			ApplyGrappleDamage();
		}
		else
		{
			Asura_Value xValue;
			if (!pxGrappleDescription->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DAMAGE, xValue))
			{
				ApplyGrappleDamage();
			}
			else
			{
				Asura_Hash_ID uDamageBlueprintID = xValue.GetHashIDValue();
				if (uDamageBlueprintID == ASURA_HASH_ID_UNSET)
				{
					ApplyGrappleDamage();
				}
				else
				{
					const Asura_Blueprint* pxDamageBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEDESIGN, uDamageBlueprintID);
					if (!pxDamageBlueprint)
					{
						ApplyGrappleDamage();
					}
					else
					{
						// OK, we managed to get a grapple damage blueprint. Get the correct damage value based on the result of the active grapple.
						Asura_Hash_ID uAmountParameter;
						Asura_Hash_ID uTypeParameter;
						Asura_Guid uSource;
						Asura_Guid uTarget;
						if (m_bActiveGrappleAttackerVictorious)
						{
							uSource = m_xOwner.GetGuid();
							uTarget = m_uCurrentTarget;
							if (m_bActiveGrappleCriticalVictory)
							{
								uAmountParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERCRITICALWIN_AMOUNT;
								uTypeParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERCRITICALWIN_TYPE;
							}
							else
							{
								uAmountParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERWIN_AMOUNT;
								uTypeParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERWIN_TYPE;
							}
						}
						else
						{
							uSource = m_uCurrentTarget;
							uTarget = m_xOwner.GetGuid();
							if (m_bActiveGrappleCriticalVictory)
							{
								uAmountParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMCRITICALWIN_AMOUNT;
								uTypeParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMCRITICALWIN_TYPE;
							}
							else
							{
								uAmountParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMWIN_AMOUNT;
								uTypeParameter = uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMWIN_TYPE;
							}
						}

						const float fDamage = pxDamageBlueprint->GetParameterAsFloat(uAmountParameter, 0.0f);

						const AVP_ServerEntity_PhysicalObject* pxTarget = AVP_ServerEntity_PhysicalObject::FindEntity(m_uCurrentTarget);
						if (pxTarget)
						{
							Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
							if(pxMessage)
							{
								AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());

								// TODO - May well need calculation of grapple impulse strength and direction, like with the normal attacks.
								const float				fImpulseStrength	= 0.0f;

								Asura_Hash_ID uDamageType = uAVP_BLUEPRINT_NAME_DAMAGE_GRAPPLE;
								if (pxDamageBlueprint->GetParameterValueByHashID(uTypeParameter, xValue))
								{
									uDamageType = xValue.GetHashIDValue();
								}

								Asura_Vector_3 xPosition;
								const AVP_ClientEntity_PhysicalObject* pxClient = AVP_ClientEntity_PhysicalObject::FindEntity(uTarget);
								if (!pxClient || !pxClient->GetMarkerLocationFromHash(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_TARGET_2), &xPosition, NULL))
								{
									xPosition = pxTarget->GetPosition() + Asura_Vector_3(0.0f, -1.2f, 0.0f);
								}
								Asura_Vector_3 xDirection;
								GetAttackDirection(pxTarget, xDirection);
								const Asura_Vector_3 xNormal(-xDirection);

								{
									const AVP_EntityDamage_RequestVictim	xDamage(uDamageType,
																					uSource,
																					uMELEE_MARKER_POINT_HASH,
																					uTarget,
																					fDamage,
																					xPosition,
																					xDirection,
																					xNormal,
																					fImpulseStrength,
																					-1,
																					false,
																					m_xOwner.GetGuid());

									// Inform the client regardless of whether we can be damaged...
									pxData->SetDamage(xDamage);

									pxMessage->SetFromGuid(m_xOwner.GetGuid());
									pxMessage->SetToGuid(m_uCurrentTarget);
									Asura_Network::Send(pxMessage);
								}
							}
						}
					}
				}
			}
		}
	}
}

bool AVP_Vertebrate_MeleeCounter_System::GetAttackTypeCanCounter(u_int uType) const
{
	return uType == AVP_MELEEATTACK_TYPE_COUNTER;
}

Asura_Hash_ID AVP_Vertebrate_MeleeCounter_System::GetMeleeMoveHash(u_int uType) const
{
	if (uType == AVP_MELEEATTACK_TYPE_NORMAL)
	{
		// Make sure our current weapon allows it.
		const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
		if (pxClient)
		{
			Asura_Hash_ID uFirstPersonBlueprint, uThirdPersonBlueprint;
			pxClient->GetCurrentMeleeBlueprintIDs(uFirstPersonBlueprint, uThirdPersonBlueprint);

			if ((uFirstPersonBlueprint != ASURA_HASH_ID_UNSET) || (uThirdPersonBlueprint != ASURA_HASH_ID_UNSET))
			{
				const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, (uFirstPersonBlueprint != ASURA_HASH_ID_UNSET) ? uFirstPersonBlueprint : uThirdPersonBlueprint);
				if (pxBlueprint)
				{
					Asura_Value xValue;
					if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_HASDEFAULTATTACK, xValue))
					{
						if (!xValue.GetBoolValue())
						{
							return ASURA_HASH_ID_UNSET;
						}
					}
				}
			}
			else
			{
				// Also fail in this case, as without any melee blueprint we can't get any animations.
				return ASURA_HASH_ID_UNSET;
			}
		}
		return AVP_Move_Melee_DefaultAttack_Start::GetStaticHash();
	}

	return ASURA_HASH_ID_UNSET;
}

u_int AVP_Vertebrate_MeleeCounter_System::PickMeleeMoveVariation(Asura_Hash_ID uMoveHash) const
{
	// By default, look at the move we've been provided with and see how many variations it has.
	const AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity(m_xOwner.GetGuid());
	if (!pxClient) { return 0; }

	const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
	if (!pxMovementSystem) { return 0; }

	const AVP_Move* pxMove = static_cast<const AVP_Move*>(pxMovementSystem->GetMove(uMoveHash));
	if (!pxMove) { return 0; }

	const u_int uNumVariations = pxMove->GetNumVariations();
	if (uNumVariations == 0) { return 0; }

	return Asura_Random::GetIntU(0, uNumVariations - 1);
}

Asura_Hash_ID AVP_Vertebrate_MeleeCounter_System::GetMeleeMoveGrappleSetBlueprintParam(u_int uType) const
{
	return ASURA_HASH_ID_UNSET;
}


bool AVP_Vertebrate_MeleeCounter_System::GetMeleeMoveStartsPuppet(Asura_Hash_ID uMoveHash) const
{
	// By default, look at the move we've been provided with and see how many variations it has.
	const AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity(m_xOwner.GetGuid());
	if (!pxClient) { return false; }

	const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	const AVP_Move* pxMove = static_cast<const AVP_Move*>(pxMovementSystem->GetMove(uMoveHash));
	if (!pxMove) { return false; }

	return pxMove->HasPuppet();
}

bool AVP_Vertebrate_MeleeCounter_System::GetCanCounter(const AVP_ServerEntity_Actor& xEnemy, const u_int uQueuedCounterType, Asura_Hash_ID& uMoveHash, u_int& uNewAttackType, bool& bMakeEnemyIntoPuppet, Asura_Hash_ID& uGrappleSetBlueprintParam) const
{
	bMakeEnemyIntoPuppet = false;
	uNewAttackType = AVP_MELEEATTACK_TYPE_COUNTER;
	uGrappleSetBlueprintParam = ASURA_HASH_ID_UNSET;

	const float fMaxDistance = GetCurrentMaxCounterDistance();

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	xEnemy.GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;

	if (xOffset.MagnitudeSquared() > (fMaxDistance * fMaxDistance)) { return false; }

	switch (xEnemy.GetClassification())
	{
	case AVP_ENTITYCLASS_ALIEN:
	case AVP_ENTITYCLASS_PREDALIEN:
		return GetCanCounter_AlienCounter(xEnemy, uQueuedCounterType, uMoveHash, uNewAttackType, bMakeEnemyIntoPuppet, uGrappleSetBlueprintParam);

	case AVP_ENTITYCLASS_ANDROID:
	case AVP_ENTITYCLASS_CIVILIAN:
	case AVP_ENTITYCLASS_HUMAN:
	case AVP_ENTITYCLASS_MARINE:
		return GetCanCounter_HumanCounter(xEnemy, uQueuedCounterType, uMoveHash, uNewAttackType, bMakeEnemyIntoPuppet, uGrappleSetBlueprintParam);

	case AVP_ENTITYCLASS_PREDATOR:
		return GetCanCounter_PredatorCounter(xEnemy, uQueuedCounterType, uMoveHash, uNewAttackType, bMakeEnemyIntoPuppet, uGrappleSetBlueprintParam);


	default:
		uMoveHash = ASURA_HASH_ID_UNSET;
		return false;
	}
}

void AVP_Vertebrate_MeleeCounter_System::SetPounceAttackVariationData(u_int uNumVariations, float* pfMaxDistances, float* pfMinDrops, float* pfDamageValues, float* pfAttackImpulses)
{
	// TODO - If we're going to end up calling this twice all the time, do we need to do something better where we don't keep reallocating things?
	if (m_uNumPounceAttackVariations > 0)
	{
		delete[] m_pxPounceAttacks;
	}

	m_uNumPounceAttackVariations = uNumVariations;

	m_pxPounceAttacks = new PounceAttackData[uNumVariations];

	for (u_int u=0; u<uNumVariations; ++u)
	{
		m_pxPounceAttacks[u].m_fMaxDistance = pfMaxDistances[u];
		m_pxPounceAttacks[u].m_fMinDrop = pfMinDrops[u];
		m_pxPounceAttacks[u].m_fDamage = pfDamageValues[u];
		m_pxPounceAttacks[u].m_fAttackImpulse = pfAttackImpulses[u];
	}
}


void AVP_Vertebrate_MeleeCounter_System::SearchForNewTarget()
{
	RequestTarget(ASURA_GUID_UNREGISTERED);
	UpdateCurrentTarget();
}


void AVP_Vertebrate_MeleeCounter_System::MakePlayer()
{
	m_bAlwaysUpdateTarget = true;
}

void AVP_Vertebrate_MeleeCounter_System::MakeAI()
{
	m_bAlwaysUpdateTarget = false;
}


bool AVP_Vertebrate_MeleeCounter_System::GetCanBlock() const
{
	if (IsBlockActive()) { return false; }

	// Check if we're playing with our weapons.  TODO - needs improvements both here and in the inventory; this is currently very nasty, and very restricting, but doesn't look as bad as not checking what the weapons are doing.
	const AVP_Inventory_System* pxInventory = m_xOwner.GetMasterInventory();
	if (!pxInventory) { return true; }

	const AVP_Inventory_Item* pxStim = pxInventory->GetItem(uAVP_BLUEPRINT_PARAM_STIM);
	const bool bStimIdling = pxStim ? pxStim->IsIdling(true) : true;

	const AVP_Inventory_Item* pxCurrentWeapon = pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON);
	if (!pxCurrentWeapon) { return bStimIdling; }

	const bool bIsWeaponIdling = pxCurrentWeapon->IsIdling(true);
	const bool bIsUsingWeapon = pxCurrentWeapon->IsUsing(false);
	const bool bWeaponInMeleeUsage = bIsUsingWeapon && (pxCurrentWeapon->GetCurrentUsageType() == AVP_InventoryItem_Weapon::UT_MELEE);

	return (bIsWeaponIdling || bWeaponInMeleeUsage) && bStimIdling;
}

void AVP_Vertebrate_MeleeCounter_System::SetBlockActive(bool bActive)
{
	if (bActive == IsBlockActive()) { return; }

	m_bBlockActive = bActive;

	if (bActive && (m_uCurrentAttackType != AVP_MELEEATTACK_TYPE_NONE))
	{
		// Just completed an attack, tell our target that's happened.
		if (m_uCurrentTarget != ASURA_GUID_UNREGISTERED)
		{
			SendIncomingMeleeFinishedMessage(m_uCurrentTarget);
		}
		m_uNumHitEntities = 0;
		m_bCurrentAttackCountered = false;
		m_bCurrentAttackHasHitEnv = false;
		m_uCurrentAttackType = AVP_MELEEATTACK_TYPE_NONE;
		m_uCurrentGrappleBlueprint = ASURA_HASH_ID_UNSET;
	}

	AVP_Inventory_System* pxInventory = m_xOwner.GetMasterInventory();
	if (pxInventory)
	{
		AVP_Inventory_Item* pxCurrentWeapon = pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON);
		if (pxCurrentWeapon)
		{
			const bool bIsUsingWeapon = pxCurrentWeapon->IsUsing(false);
			const bool bWeaponInMeleeUsage = bIsUsingWeapon && (pxCurrentWeapon->GetCurrentUsageType() == AVP_InventoryItem_Weapon::UT_MELEE);

			if (bActive)
			{
				if (!bWeaponInMeleeUsage)
				{
					Asura_Assert(!bIsUsingWeapon, "Should not be using a weapon already.");
					pxCurrentWeapon->RequestBeginUsage(AVP_InventoryItem_Weapon::UT_MELEE);
				}
			}
			else
			{
				Asura_Assert(pxCurrentWeapon->GetCurrentUsageType() == AVP_InventoryItem_Weapon::UT_MELEE, "Weapon usage type should be melee.");
				pxCurrentWeapon->RequestEndUsage();
			}
		}
	}
}

// Default, generic counter moves for all species. Subclasses can set up special attack-specific counters.
bool AVP_Vertebrate_MeleeCounter_System::GetCanCounter_AlienCounter(const AVP_ServerEntity_Actor& xEnemy, const u_int uQueuedCounterType, Asura_Hash_ID& uMoveHash, u_int& uNewAttackType, bool& bMakeEnemyIntoPuppet, Asura_Hash_ID& uGrappleSetBlueprintParam) const
{
	uMoveHash = ASURA_HASH_ID_UNSET;
	return false;
}

bool AVP_Vertebrate_MeleeCounter_System::GetCanCounter_HumanCounter(const AVP_ServerEntity_Actor& xEnemy, const u_int uQueuedCounterType, Asura_Hash_ID& uMoveHash, u_int& uNewAttackType, bool& bMakeEnemyIntoPuppet, Asura_Hash_ID& uGrappleSetBlueprintParam) const
{
	uMoveHash = AVP_Move_Melee_DefaultCounter::GetStaticHash();

	return (!xEnemy.IsPlayer());
}

bool AVP_Vertebrate_MeleeCounter_System::GetCanCounter_PredatorCounter(const AVP_ServerEntity_Actor& xEnemy, const u_int uQueuedCounterType, Asura_Hash_ID& uMoveHash, u_int& uNewAttackType, bool& bMakeEnemyIntoPuppet, Asura_Hash_ID& uGrappleSetBlueprintParam) const
{
	// At this level, only the standard counter can be used.
	if (uQueuedCounterType != AVP_MELEEATTACK_TYPE_COUNTER)
	{
		uMoveHash = ASURA_HASH_ID_UNSET;
		return false;
	}

	const AVP_Inventory_System* pxInventory = xEnemy.GetInventorySystem();
	if (!pxInventory)
	{
		uMoveHash = ASURA_HASH_ID_UNSET;
		return false;
	}

	const AVP_InventoryItem_Weapon* pxWeapon = static_cast<const AVP_InventoryItem_Weapon*>(pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON));
	if (!pxWeapon)
	{
		uMoveHash = ASURA_HASH_ID_UNSET;
		return false;
	}

	if (pxWeapon->GetItemType() == AVP_INVENTORYITEMTYPE_WEAPON_MELEE)
	{
		uMoveHash = AVP_Predator_Move_Melee_FarCounter::GetStaticHash();
		return true;
	}

	// Every other weapon uses the default counter.
	uMoveHash = AVP_Move_Melee_DefaultCounter::GetStaticHash();
	return true;
}

float AVP_Vertebrate_MeleeCounter_System::GetCurrentMaxCounterDistance() const
{
	return m_fAttackDetectionRange;
}

float AVP_Vertebrate_MeleeCounter_System::GetAttackDamage() const
{
	return m_fMinorDamageAmount;
}

Asura_Hash_ID AVP_Vertebrate_MeleeCounter_System::GetAttackDamageType() const
{
	return uAVP_BLUEPRINT_NAME_DAMAGE_MELEE;
}

float AVP_Vertebrate_MeleeCounter_System::GetAttackImpulseStrength() const
{
	return dbg_fMeleeDamageImpulseStrength;
}

void AVP_Vertebrate_MeleeCounter_System::GetAttackDirection(const AVP_ServerEntity_PhysicalObject* pxTarget, Asura_Vector_3& xDirection) const
{
	if (!pxTarget)
	{
		xDirection = m_xOwner.GetAimDirection();
		return;
	}

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	const AVP_ServerActor_Vertebrate* pxVertTarget = AVP_ServerActor_Vertebrate::FindEntity(pxTarget->GetGuid());
	if (pxVertTarget)
	{
		pxVertTarget->GetChestPosition(xDirection);
	}
	else
	{
		pxTarget->GetMeleeOrigin(xDirection);
	}
	xDirection -= xOwnerOrigin;

	xDirection.Normalise();
}

float AVP_Vertebrate_MeleeCounter_System::GetGrappleDamage() const
{
	return dbg_fFatalDamageAmount;
}

Asura_Hash_ID AVP_Vertebrate_MeleeCounter_System::GetAttackBlueprintIDForGrapple(Asura_Guid uHitEntity) const
{
	const Asura_Hash_ID uParamID = GetAttackGrappleSetBlueprintParamID();
	if (uParamID == ASURA_HASH_ID_UNSET) { return ASURA_HASH_ID_UNSET; }

	const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(uHitEntity);

	return GetGrappleBlueprintFromGrappleSetParameter(pxTarget, uParamID);
}

bool AVP_Vertebrate_MeleeCounter_System::GetSendIncomingMeleeMessage(u_int uAttackType) const
{
	if (uAttackType == AVP_MELEEATTACK_TYPE_NONE) { return false; }

	if (uAttackType == AVP_MELEEATTACK_TYPE_NORMAL) { return IsCurrentTargetInSuckRange(); }

	return true;
}

bool AVP_Vertebrate_MeleeCounter_System::IsCurrentTargetInSuckRange() const
{
	const AVP_ServerActor_Vertebrate* pxTarget = AVP_ServerActor_Vertebrate::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return false; }

	const float fMinSuckDistance = AVP_Move_Melee_GenericAttack_Start::GetMinSuckDistance();

	const Asura_Vector_3& xTargetPos = pxTarget->GetPosition();
	const Asura_Vector_3& xOwnerPos = m_xOwner.GetPosition();
	Asura_Vector_3 xFromTarget = xOwnerPos - xTargetPos;
	const float fMagSquared = xFromTarget.MagnitudeSquared();

	return (fMagSquared <= (m_fMaxSuckDistance * m_fMaxSuckDistance)) && (fMagSquared > (fMinSuckDistance * fMinSuckDistance));
}

bool AVP_Vertebrate_MeleeCounter_System::IsMeleeDamageType(const Asura_Hash_ID uTypeHash)
{
	switch (uTypeHash)
	{
	default:
		return false;

	case uAVP_BLUEPRINT_NAME_DAMAGE_BLADE:
	case uAVP_BLUEPRINT_NAME_DAMAGE_CLAW:
	case uAVP_BLUEPRINT_NAME_DAMAGE_MELEE:
	case uAVP_BLUEPRINT_NAME_DAMAGE_TAIL:
		return true;
	}
}

u_int AVP_Vertebrate_MeleeCounter_System::GetPounceVariation() const
{
	const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return 0; }

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;
	const float fDistSquared = xOffset.MagnitudeSquared();
	const float fDrop = xOffset.y; // Intentionally allowing negative as most pounces have a minimum that lets you jump up.

	float fLeastDistance = Asura_Maths::LargeFloat;
	float fGreatestDrop = -Asura_Maths::LargeFloat;
	u_int uBestIndex = m_uNumPounceAttackVariations;

	for (u_int u=0; u<m_uNumPounceAttackVariations; ++u)
	{
		const float fCurrentDistance = m_pxPounceAttacks[u].m_fMaxDistance;
		const float fCurrentDrop = m_pxPounceAttacks[u].m_fMinDrop;

		// Only use the drop as the deciding factor between pounces if the value is greater than 0 (i.e. an actual drop).
		const bool bCurrentIsDrop = fCurrentDrop > 0.0f;
		const bool bNoPreviousDrop = (fDrop <= 0.0f) || (fGreatestDrop <= 0.0f);
		const bool bPassPreviousDrop = bNoPreviousDrop || (fCurrentDrop > fGreatestDrop);
		if (bPassPreviousDrop && (fDrop >= fCurrentDrop) && (!bNoPreviousDrop || bCurrentIsDrop || (fCurrentDistance < fLeastDistance)) && (fDistSquared <= (fCurrentDistance * fCurrentDistance)))
		{
			uBestIndex = u;
			fLeastDistance = fCurrentDistance;
			fGreatestDrop = fCurrentDrop;
		}
	}

	Asura_Assert(uBestIndex != m_uNumPounceAttackVariations, "Shouldn't be picking a variation if we can't pounce.");
	return uBestIndex;
}

bool AVP_Vertebrate_MeleeCounter_System::GetCanUseShortestPounceVariation(Asura_Guid uTarget) const
{
	if (m_uNumPounceAttackVariations == 0) { return false; }

	u_int uShortestVariation = m_uNumPounceAttackVariations;
	float fShortestDistance = Asura_Maths::LargeFloat;
	float fChosenDrop = -Asura_Maths::LargeFloat;

	for (u_int u=0; u<m_uNumPounceAttackVariations; ++u)
	{
		// Ignore drop attacks
		if (m_pxPounceAttacks[u].m_fMinDrop > 0.0f) { continue; }
		if (m_pxPounceAttacks[u].m_fMaxDistance >= fShortestDistance) { continue; }

		uShortestVariation = u;
		fShortestDistance = m_pxPounceAttacks[u].m_fMaxDistance;
		fChosenDrop = m_pxPounceAttacks[u].m_fMinDrop;
	}

	if (uShortestVariation == m_uNumPounceAttackVariations) { return false; }

	const AVP_ServerActor_Vertebrate* pxTarget = AVP_ServerActor_Vertebrate::FindEntity(uTarget);
	if (!pxTarget) { return false; }

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;
	const float fDistSquared = xOffset.MagnitudeSquared();

	if ((xOffset.y >= fChosenDrop) && (fDistSquared <= (fShortestDistance * fShortestDistance)))
	{
		// Finally, do a LOS test to see if there's anything between us
		Asura_Vector_3 xDirection = xOffset;
		const float fDistance = xDirection.Normalise();
		u_short uFlagsToIgnore = AsuraPolygonFlag_NoCollisions;
		Asura_LOS_InputData xInput(xOwnerOrigin, xDirection, m_xOwner.GetGuid(), fDistance, false, uFlagsToIgnore);
		const Asura_LOS_OutputData* pxOutput = DoMeleeLOSTest(&xInput);

		return (!pxOutput->m_bHitEnvironment && (pxOutput->m_uHitEntity == 0) || (pxOutput->m_uHitEntity == uTarget));
	}
	else
	{
		return false;
	}
}


void AVP_Vertebrate_MeleeCounter_System::InitialiseActiveGrapple()
{
	m_fActiveGrappleAttackerValue = 0.0f;
	m_fActiveGrappleVictimValue = 0.0f;
	m_bActiveGrappleCriticalVictory = false;
	m_bActiveGrappleVictimAdvantage = false;

	// See what each participant is.
	const bool bOwnerIsPlayer = m_xOwner.IsPlayer();
	const AVP_ServerActor_Vertebrate* pxVictim = AVP_ServerActor_Vertebrate::FindEntity(m_uCurrentTarget);
	const bool bVictimIsPlayer = pxVictim && pxVictim->IsPlayer();

	// See if we have a grapple description blueprint
	Asura_Blueprint* pxGrappleDescription = NULL;
	if (m_uCurrentGrappleBlueprint != ASURA_HASH_ID_UNSET)
	{
		pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, m_uCurrentGrappleBlueprint);
	}

	bool bOwnerHasActiveGrapple = true;
	bool bVictimHasActiveGrapple = true;
	if (pxGrappleDescription)
	{
		Asura_Value xValue;
		if (pxGrappleDescription->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DIFFICULTY_ATTACKER, xValue))
		{
			bOwnerHasActiveGrapple = xValue.GetIntValue() > 0;
		}
		else
		{
			bOwnerHasActiveGrapple = false;
		}

		if (pxGrappleDescription->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DIFFICULTY_VICTIM, xValue))
		{
			bVictimHasActiveGrapple = xValue.GetIntValue() > 0;
		}
		else
		{
			bVictimHasActiveGrapple = false;
		}
	}

	// We only wait for further input if a character is a player with an active grapple or both are AI
	if (bOwnerIsPlayer || bVictimIsPlayer)
	{
		m_bActiveGrappleReceivedAttackerValue = !bOwnerIsPlayer || !bOwnerHasActiveGrapple;
		m_bActiveGrappleReceivedVictimValue = !bVictimIsPlayer || !bVictimHasActiveGrapple;
	}
	else
	{
		m_bActiveGrappleReceivedAttackerValue = false;
		m_bActiveGrappleReceivedVictimValue = false;
	}

	// If we have no active grapple for anyone, but at least one player, randomise the critical/non-critical result for the appropriate victor (the player if we have one, otherwise the attacker).
	if (!bOwnerHasActiveGrapple && !bVictimHasActiveGrapple)
	{
		if (bOwnerIsPlayer)
		{
			if (Asura_Random::GetInt(0, 1) == 1)
			{
				m_fActiveGrappleAttackerValue = 1.0f;
				m_bActiveGrappleCriticalVictory = true;
			}
		}
		else if (bVictimIsPlayer)
		{
			m_bActiveGrappleVictimAdvantage = true;
			if (Asura_Random::GetInt(0, 1) == 1)
			{
				m_fActiveGrappleVictimValue = 1.0f;
				m_bActiveGrappleCriticalVictory = true;
			}
		}
	}
	else
	{
		// The victim only gets the advantage if the attacker is using an active grapple and the victim is not.
		m_bActiveGrappleVictimAdvantage = (bOwnerIsPlayer && bOwnerHasActiveGrapple) && (!bVictimIsPlayer || !bVictimHasActiveGrapple);
	}

	// Initial result values.
	m_bActiveGrappleAttackerVictorious = !m_bActiveGrappleVictimAdvantage;
}

void AVP_Vertebrate_MeleeCounter_System::ApplyActiveGrappleResult()
{
	// Update our own storage of the details.
	if (m_fActiveGrappleAttackerValue == m_fActiveGrappleVictimValue)
	{
		// In equality, the advantaged character wins.
		m_bActiveGrappleAttackerVictorious = !m_bActiveGrappleVictimAdvantage;
	}
	else
	{
		// Otherwise, the highest score wins.
		m_bActiveGrappleAttackerVictorious = m_fActiveGrappleAttackerValue > m_fActiveGrappleVictimValue;
	}
	// If the victim is invincible, they will always win. It's always the attacker that applies the results, so the current target is the victim.
	const AVP_ClientEntity_Actor* pxVictim = AVP_ClientEntity_Actor::FindEntity(m_uCurrentTarget);
	if (pxVictim && pxVictim->IsInvulnerable())
	{
		m_bActiveGrappleAttackerVictorious = false;
	}

	m_bActiveGrappleCriticalVictory = Asura_Maths::Max(m_fActiveGrappleAttackerValue, m_fActiveGrappleVictimValue) >= 1.0f;

	// Send these details to our entity for passing to clients.
	Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ACTIVEGRAPPLE_RESULT, sizeof(AVP_Message_Vertebrate_ActiveGrappleResult));
	if (pxMessage)
	{
		AVP_Message_Vertebrate_ActiveGrappleResult* pxData = static_cast<AVP_Message_Vertebrate_ActiveGrappleResult*>(pxMessage->GetData());

		pxData->SetAttackerVictorious(m_bActiveGrappleAttackerVictorious);
		pxData->SetCriticalVictory(m_bActiveGrappleCriticalVictory);
		pxData->SetFinalResult(m_bActiveGrappleReceivedVictimValue && m_bActiveGrappleReceivedAttackerValue);

		pxMessage->To = m_xOwner.GetGuid(); // Send this via the server so AIs can make use of it.
		pxMessage->From = m_xOwner.GetGuid();

		Asura_Network::Send(pxMessage);
	}

	if (!m_bActiveGrappleAttackerVictorious)
	{
		// Also inform the AI if they got countered. No need to do this via message sending.
		AVP_Brain_Actor* pxBrain = m_xOwner.GetActorBrain();
		if (pxBrain)
		{
			// TODO - check if it's a problem leaving most data uninitialised.
			Asura_Message xMessage;
			xMessage.SetMsgID(AVP_MSGID_COUNTER_MELEE);
			pxBrain->DoHandleMessage(&xMessage);
		}
	}
}

const Asura_LOS_OutputData* AVP_Vertebrate_MeleeCounter_System::DoMeleeLOSTest(const Asura_LOS_InputData* pxInput)
{
	return AVP_LOS::TestAgainstAlmostEverything(pxInput, NUM_LOS_CLASSIFICATION_EXCEPTIONS, s_ausLosClassificationExceptions);
}

Asura_Hash_ID AVP_Vertebrate_MeleeCounter_System::GetGrappleBlueprintFromGrappleSetParameter(const AVP_ServerEntity_Actor* pxVictim, const Asura_Hash_ID uGrappleSetBlueprintParam)
{
	if (!pxVictim) { return ASURA_HASH_ID_UNSET; }

	const Asura_Hash_ID uGrappleSetBlueprintID = pxVictim->GetCurrentGrappleBlueprintID();
	if (uGrappleSetBlueprintID == ASURA_HASH_ID_UNSET) { return ASURA_HASH_ID_UNSET; }

	const Asura_Blueprint* pxGrappleSet = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, uGrappleSetBlueprintID);
	if (!pxGrappleSet) { return ASURA_HASH_ID_UNSET; }

	Asura_Value xValue;
	if (pxGrappleSet->GetParameterValueByHashID(uGrappleSetBlueprintParam, xValue))
	{
		// Just using the name provided.
		return xValue.GetHashIDValue();
	}
	else
	{
		// We might have set up variations, so try the enumeration technique instead.
		AVP_Blueprint_ParameterEnumerator_GrappleSelector xEnumerator;
		pxGrappleSet->EnumerateParameters(uGrappleSetBlueprintParam, 2, xEnumerator);

		return xEnumerator.GetSelectedGrappleBlueprint();
	}
}


void AVP_Vertebrate_MeleeCounter_System::SendMeleeAttackHitMessage()
{
	// Some moves will need to know which enemy was hit in the end, so that data is sent along. This assumes it'll only ever care about the first enemy hit (probably fine).

	// See if this hit should trigger a grapple.
	const Asura_Guid uHitEntity = m_auHitEntities[0];
	m_uCurrentGrappleBlueprint = GetAttackBlueprintIDForGrapple(uHitEntity);

	if (m_uCurrentGrappleBlueprint == ASURA_HASH_ID_UNSET)
	{
		// Nope, normal hit message
		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_MELEE_ATTACK_HIT, sizeof(AVP_Msg_Melee_Attack_Hit));

		if (pxMessage)
		{
			AVP_Msg_Melee_Attack_Hit* pxData = static_cast<AVP_Msg_Melee_Attack_Hit*>(pxMessage->GetData());
			if (pxData)
			{
				pxData->SetHitEntity(m_auHitEntities[0]);
				pxData->SetPendingVariation(m_uCurrentVariation);
				pxData->SetPendingAttackType(m_uCurrentAttackType);
				pxData->SetDamageType(GetAttackDamageType());
				pxData->SetIsRightHanded(GetAttackIsRightHanded());
			}

			pxMessage->SetToGuid(m_xOwner.GetGuid());
			pxMessage->SetFromGuid(m_xOwner.GetGuid());

			Asura_Network::Send(pxMessage);
		}
	}
	else
	{
		// Yep, special grapple-triggering message. Hopefully the move will be set up to deal with this.
		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_MELEE_ATTACK_PREPAREGRAPPLE, sizeof(AVP_Msg_Melee_Attack_PrepareGrapple));

		if (pxMessage)
		{
			AVP_Msg_Melee_Attack_PrepareGrapple* pxData = static_cast<AVP_Msg_Melee_Attack_PrepareGrapple*>(pxMessage->GetData());
			if (pxData)
			{
				pxData->SetHitEntity(uHitEntity);
				pxData->SetGrappleDescriptionBlueprint(m_uCurrentGrappleBlueprint);
			}

			pxMessage->SetToGuid(m_xOwner.GetGuid());
			pxMessage->SetFromGuid(m_xOwner.GetGuid());

			Asura_Network::Send(pxMessage);
		}

		InitialiseActiveGrapple();
	}
}


void AVP_Vertebrate_MeleeCounter_System::SendMeleeFailedToStartMessage(u_int uAttackType)
{
	// Some moves will need to know which enemy was hit in the end, so that data is sent along. This assumes it'll only ever care about the first enemy hit (probably fine).
	Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_MELEE_ATTACK_FAILED_TO_START, sizeof(AVP_Msg_Melee_Attack_FailedToStart));

	if (pxMessage)
	{
		AVP_Msg_Melee_Attack_FailedToStart* pxData = static_cast<AVP_Msg_Melee_Attack_FailedToStart*>(pxMessage->GetData());
		if (pxData)
		{
			pxData->SetType(uAttackType);
		}

		pxMessage->SetToGuid(m_xOwner.GetGuid());
		pxMessage->SetFromGuid(m_xOwner.GetGuid());

		Asura_Network::Send(pxMessage);
	}
}

void AVP_Vertebrate_MeleeCounter_System::UpdateCurrentTarget()
{
	Asura_Guid uInitialTarget = m_uCurrentTarget;

	if (m_uRequestedTarget != ASURA_GUID_UNREGISTERED)
	{
		// TODO - check if this target is actually reasonable.
		m_uCurrentTarget = m_uRequestedTarget;
	}
	else
	{
		m_uCurrentTarget = ASURA_GUID_UNREGISTERED;

		const Asura_Matrix_3x3& xOri = m_xOwner.GetOrientation();
		const Asura_Vector_3 xDown(xOri.e12, xOri.e22, xOri.e32);

		Asura_Vector_3 xPos;
		m_xOwner.GetMeleeOrigin(xPos);
		Asura_Vector_3 xDir = m_xOwner.GetAimDirection();
		xDir -= xDown * (xDir * xDown);
		xDir.Normalise();

		// Do a broadphase query to significantly reduce the number of entities we check here.
		Asura_Physics_Broadphase_Query xQuery;
		Asura_Bounding_Box xBox;
		xBox.SetToPoint(xPos);
		xBox.Grow(m_fTargetDetectionRange);
		xQuery.MakeQuery(xBox, true, true, true, true);

		float fBestCosAngle = m_fTargetDetectionCosAngle;
		float fBestDistSquared = 10000.0f;
		float fMaxDist = m_fTargetDetectionRange;
		bool bBestInAttackRange = false;
		bool bHasActorTarget = false;
		bool bHasHighThreatActorTarget = false;

		const Asura_Guid uOwnerGuid = m_xOwner.GetGuid();

		// Go through all the entries in turn (there shouldn't be that many).
		const Asura_Physics_Broadphase_Entry* pxEntry = xQuery.GetFirstEntry();

		while (pxEntry)
		{
			const Asura_Physics_RigidBodyInstance* pxInstance = pxEntry->GetInstance();
			const Asura_Guid uEntityGuid = pxInstance->GetAssociatedEntityGuid();
			pxEntry = pxEntry->GetNextEntry();

			if (uEntityGuid == uOwnerGuid) { continue; }

			// See if this is an actor or another kind of object.
			const AVP_ServerEntity_Actor* pxActor = AVP_ServerEntity_Actor::FindEntity(uEntityGuid);

			if (pxActor)
			{
				if (!pxActor->IsDead() && pxActor->IsEntityActive())
				{
					if (AVP_Tactical_System::GetAllegiance(m_xOwner, *pxActor) != AVP_ALLEGIANCE_FRIEND)
					{
						// Get its location and see if it's in range
						Asura_Vector_3 xEnemyPosition;
						pxActor->GetMeleeOrigin(xEnemyPosition);
						Asura_Vector_3 xToEnemyVector = xEnemyPosition - xPos;
						float fDistSquared = xToEnemyVector.MagnitudeSquared();

						const AVP_Inventory_System* pxInventory = pxActor->GetInventorySystem();
						const AVP_Inventory_Item* pxWeapon = pxInventory ? pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL;
						bool bHighThreatTarget = pxActor->IsDestructible() && (pxWeapon ? (pxWeapon->GetItemType() != AVP_INVENTORYITEMTYPE_UNARMED) : true);

						// Only vertebrates should be high-threat targets.
						bHighThreatTarget &= (AVP_ServerActor_Vertebrate::FindEntity(uEntityGuid) != NULL);

						const bool bNewHighThreatTarget = bHighThreatTarget && !bHasHighThreatActorTarget;
						const bool bSameThreatLevel = (bHighThreatTarget == bHasHighThreatActorTarget);

						if (fDistSquared < (fMaxDist*fMaxDist))
						{
							bool bInAttackRange = (fDistSquared < (m_fAttackDetectionRange * m_fAttackDetectionRange)) || GetCanUseShortestPounceVariation(uEntityGuid);

							// Shove the origin back a bit for the angle test, so very close enemies don't get missed.
							Asura_Vector_3 xAdjustedToEnemyVector = xToEnemyVector + (xDir * m_fTargetDetectionOriginBackwardsOffset);

							// Flatten the test
							xAdjustedToEnemyVector -= xDown * (xAdjustedToEnemyVector * xDown);

							xAdjustedToEnemyVector.Normalise();
							float fCosAngle = xAdjustedToEnemyVector * xDir;
							if (fCosAngle > m_fTargetDetectionCosAngle)
							{
								// Don't go from high threat to low threat, unless the old target was out of attack range and the new target is in range.
								if (!bSameThreatLevel && !bNewHighThreatTarget)
								{
									if (bBestInAttackRange || !bInAttackRange)
									{
										continue;
									}
								}

								// Conversely, don't go to a high threat target over a low-threat target if you can't actually attack the high-threat target.
								if (!bInAttackRange && bBestInAttackRange)
								{
									continue;
								}

								if (bNewHighThreatTarget || (fCosAngle > fBestCosAngle) || ((fCosAngle == fBestCosAngle) && (fDistSquared < fBestDistSquared)))
								{
									// Finally, make sure there isn't something in the way.
									Asura_Vector_3 xEnemyDir = xToEnemyVector;
									float fDist = xEnemyDir.Normalise();
									Asura_LOS_InputData xInput(xPos, xEnemyDir, uOwnerGuid, fDist, false, m_xOwner.GetFlagsToIgnore() | AsuraPolygonFlag_NoCollisions);
									const Asura_LOS_OutputData* pxOutput = DoMeleeLOSTest(&xInput);
									if (!pxOutput->m_bHitEnvironment && (!pxOutput->m_uHitEntity || (pxOutput->m_uHitEntity == uEntityGuid)))
									{
										m_uCurrentTarget = uEntityGuid;
										fBestCosAngle = fCosAngle;
										fBestDistSquared = fDistSquared;
										bBestInAttackRange = bInAttackRange;
										bHasActorTarget = true;
										bHasHighThreatActorTarget = bHighThreatTarget;
									}
								}
							}
						}
					}
				}
			}
			else
			{
				// Only bother looking for something else if we don't have an actor target.
				if (bHasActorTarget) { continue; }

				// Don't target non-actors if they've already been hit this attack.
				bool bAlreadyHit = false;
				for (u_int u=0; u<m_uNumHitEntities; ++u)
				{
					if (uEntityGuid == m_auHitEntities[u])
					{
						bAlreadyHit = true;
						break;
					}
				}
				if (bAlreadyHit) { continue; }

				float fDistSquared = Asura_Maths::LargeFloat;
				Asura_Vector_3 xToTargetVector(0.0f, 0.0f, 0.0f);
				Asura_Vector_3 xTargetPosition(0.0f, 0.0f, 0.0f);

				const AVP_ServerEntity_PhysicsObject* pxPhysObj = AVP_ServerEntity_PhysicsObject::FindEntity(uEntityGuid);

				if (pxPhysObj && !pxPhysObj->IsDead() && pxPhysObj->IsEntityActive() && pxPhysObj->IsDestructible())
				{
					// Get its location and see if it's in range
					pxPhysObj->GetMeleeOrigin(xTargetPosition);
					xToTargetVector = xTargetPosition - xPos;
					fDistSquared = xToTargetVector.MagnitudeSquared();
				}
				else
				{
					// Last thing we're interested in that it could be: a container instance.
					const AVP_ServerEntity_Container* pxContainer = AVP_ServerEntity_Container::FindEntity(uEntityGuid);
					if (pxContainer && pxContainer->IsEntityActive() && pxContainer->IsDamageable())
					{
						// Use the centre of the physics object for this.
						xTargetPosition = pxInstance->GetCentreOfMass();
						xToTargetVector = xTargetPosition - xPos;
						fDistSquared = xToTargetVector.MagnitudeSquared();
					}
				}

				if (fDistSquared < (fMaxDist*fMaxDist))
				{
					// Shove the origin back a bit for the angle test, so very close enemies don't get missed.
					xToTargetVector += xDir * m_fTargetDetectionOriginBackwardsOffset;

					xToTargetVector.y = 0.0f;
					xToTargetVector.Normalise();
					float fCosAngle = xToTargetVector * xDir;
					if (fCosAngle > m_fTargetDetectionCosAngle)
					{
						if ((fCosAngle > fBestCosAngle) || ((fCosAngle == fBestCosAngle) && (fDistSquared < fBestDistSquared)))
						{
							m_uCurrentTarget = uEntityGuid;
							fBestCosAngle = fCosAngle;
							fBestDistSquared = fDistSquared;
							m_xStoredTargetPosition = xTargetPosition;
						}
					}
				}
			}
		}
	}

	// If our target has changed and we're currently in an attack, let those involved know.
	if ((uInitialTarget != m_uCurrentTarget) && (m_uCurrentAttackType != AVP_MELEEATTACK_TYPE_NONE))
	{
		SendIncomingMeleeFinishedMessage(uInitialTarget);
		SendIncomingMeleeStartedMessage(m_uCurrentTarget);
	}
}

void AVP_Vertebrate_MeleeCounter_System::SendIncomingMeleeStartedMessage(Asura_Guid uTarget)
{
	if ((uTarget != ASURA_GUID_UNREGISTERED) && !m_bCurrentAttackSentIncomingMeleeMessage)
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_INCOMING_MELEE_STARTED);
		if (pxMessage)
		{
			pxMessage->To = uTarget;
			pxMessage->From = m_xOwner.GetGuid();

			Asura_Network::Send(pxMessage);
		}

		m_bCurrentAttackSentIncomingMeleeMessage = true;
	}
}

void AVP_Vertebrate_MeleeCounter_System::SendIncomingMeleeFinishedMessage(Asura_Guid uTarget)
{
	if ((uTarget != ASURA_GUID_UNREGISTERED) && m_bCurrentAttackSentIncomingMeleeMessage)
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_INCOMING_MELEE_FINISHED);
		if (pxMessage)
		{
			pxMessage->To = uTarget;
			pxMessage->From = m_xOwner.GetGuid();

			Asura_Network::Send(pxMessage);
		}

		m_bCurrentAttackSentIncomingMeleeMessage = false;
	}
}

////////////////////////////////////////////////////////////////
// Parameter enumerators for the additional pounce melee data //
////////////////////////////////////////////////////////////////

AVP_Blueprint_ParameterEnumerator_PounceVariations::AVP_Blueprint_ParameterEnumerator_PounceVariations(Asura_Hash_ID uPounceBlueprintType)
: m_uPounceBlueprintType(uPounceBlueprintType)
, m_uNumVariations(0)
{
}

bool AVP_Blueprint_ParameterEnumerator_PounceVariations::Enumerate(const u_int uIndex, const char* const szName, const Asura_Value& xValue)
{
	Asura_Assert(m_uNumVariations < MAX_POUNCE_VARIATIONS, "Too many pounce variations, ignoring the rest.");
	if (m_uNumVariations >= MAX_POUNCE_VARIATIONS) { return false; }

	Asura_Hash_ID uBlueprintID = xValue.GetHashIDValue();
	if (uBlueprintID == ASURA_HASH_ID_UNSET) { return true; } // It's probably valid to not have a blueprint, if a little odd to have selected a predator blueprint without them.

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(m_uPounceBlueprintType, uBlueprintID);
	if (!pxBlueprint) { return false; }

	Asura_Value xPounceValue;

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONSTART, xPounceValue))
	{
		m_auFirstPersonStartAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auFirstPersonStartAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONHIT, xPounceValue))
	{
		m_auFirstPersonHitAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auFirstPersonHitAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONCOUNTERED, xPounceValue))
	{
		m_auFirstPersonCounteredAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auFirstPersonCounteredAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_FIRSTPERSONMISS, xPounceValue))
	{
		m_auFirstPersonMissAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auFirstPersonMissAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONSTART, xPounceValue))
	{
		m_auThirdPersonStartAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auThirdPersonStartAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONHIT, xPounceValue))
	{
		m_auThirdPersonHitAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auThirdPersonHitAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONCOUNTERED, xPounceValue))
	{
		m_auThirdPersonCounteredAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auThirdPersonCounteredAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_THIRDPERSONMISS, xPounceValue))
	{
		m_auThirdPersonMissAnims[m_uNumVariations] = xPounceValue.GetHashIDValue();
	}
	else
	{
		m_auThirdPersonMissAnims[m_uNumVariations] = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_MAXDISTANCE, xPounceValue))
	{
		m_afMaxDistances[m_uNumVariations] = xPounceValue.GetFloatValue();
	}
	else
	{
		m_afMaxDistances[m_uNumVariations] = 0.0f;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_DAMAGE, xPounceValue))
	{
		m_afDamageValues[m_uNumVariations] = xPounceValue.GetFloatValue();
	}
	else
	{
		m_afDamageValues[m_uNumVariations] = 0.0f;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_ATTACKIMPULSE, xPounceValue))
	{
		m_afAttackImpulses[m_uNumVariations] = xPounceValue.GetFloatValue();
	}
	else
	{
		m_afAttackImpulses[m_uNumVariations] = dbg_fMeleeDamageImpulseStrength;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_POUNCE_BLUEPRINT_PARAM_MINIMUMDROP, xPounceValue))
	{
		m_afMinDrops[m_uNumVariations] = xPounceValue.GetFloatValue();
	}
	else
	{
		m_afMinDrops[m_uNumVariations] = dbg_fPounceDefaultMinDrop;
	}

	++m_uNumVariations;

	return true;
}

// Melee/counter system subclass.
AVP_Blueprint_ParameterEnumerator_PounceVariations_MeleeCounterData::AVP_Blueprint_ParameterEnumerator_PounceVariations_MeleeCounterData(Asura_Hash_ID uPounceBlueprintType, AVP_Vertebrate_MeleeCounter_System& xMeleeCounterSystem)
: PARENT(uPounceBlueprintType)
, m_xMeleeCounterSystem(xMeleeCounterSystem)
{
}

void AVP_Blueprint_ParameterEnumerator_PounceVariations_MeleeCounterData::EndEnumeration(const bool bSuccess)
{
	// Success shouldn't matter as that basically means we tried to set more variations than we're allowed.
	if (m_uNumVariations > 0)
	{
		m_xMeleeCounterSystem.SetPounceAttackVariationData(m_uNumVariations, m_afMaxDistances, m_afMinDrops, m_afDamageValues, m_afAttackImpulses);
	}
}

// Moves subclass.
AVP_Blueprint_ParameterEnumerator_PounceVariations_MoveData::AVP_Blueprint_ParameterEnumerator_PounceVariations_MoveData(	Asura_Hash_ID uPounceBlueprintType,
																															AVP_Move_Melee_Pounce_Attack_Start& xStartMove,
																															AVP_Move_Melee_Pounce_Attack_Return_Hit& xHitMove,
																															AVP_Move_Melee_Pounce_Attack_Return_Countered& xCounteredMove,
																															AVP_Move_Melee_Pounce_Attack_Return_Miss& xMissMove)
: PARENT(uPounceBlueprintType)
, m_xStartMove(xStartMove)
, m_xHitMove(xHitMove)
, m_xCounteredMove(xCounteredMove)
, m_xMissMove(xMissMove)
{
}

void AVP_Blueprint_ParameterEnumerator_PounceVariations_MoveData::EndEnumeration(const bool bSuccess)
{
	// Success shouldn't matter as that basically means we tried to set more variations than we're allowed.
	if (m_uNumVariations > 0)
	{
		m_xStartMove.SetFirstPersonAnimHashesForVariations(m_auFirstPersonStartAnims, m_uNumVariations);
		m_xStartMove.SetThirdPersonAnimHashesForVariations(m_auThirdPersonStartAnims, m_uNumVariations);

		m_xHitMove.SetFirstPersonAnimHashesForVariations(m_auFirstPersonHitAnims, m_uNumVariations);
		m_xHitMove.SetThirdPersonAnimHashesForVariations(m_auThirdPersonHitAnims, m_uNumVariations);

		m_xCounteredMove.SetFirstPersonAnimHashesForVariations(m_auFirstPersonCounteredAnims, m_uNumVariations);
		m_xCounteredMove.SetThirdPersonAnimHashesForVariations(m_auThirdPersonCounteredAnims, m_uNumVariations);

		m_xMissMove.SetFirstPersonAnimHashesForVariations(m_auFirstPersonMissAnims, m_uNumVariations);
		m_xMissMove.SetThirdPersonAnimHashesForVariations(m_auThirdPersonMissAnims, m_uNumVariations);
	}
}

// Melee/counter system subclass.
AVP_Blueprint_ParameterEnumerator_PounceVariations_FocusData::AVP_Blueprint_ParameterEnumerator_PounceVariations_FocusData(Asura_Hash_ID uPounceBlueprintType, AVP_Vertebrate_Focus& xFocus)
: PARENT(uPounceBlueprintType)
, m_xFocus(xFocus)
{
}

void AVP_Blueprint_ParameterEnumerator_PounceVariations_FocusData::EndEnumeration(const bool bSuccess)
{
	// Success shouldn't matter as that basically means we tried to set more variations than we're allowed.
	if (m_uNumVariations > 0)
	{
		m_xFocus.SetPounceAttackVariationData(m_uNumVariations, m_afMaxDistances, m_afMinDrops);
	}
}

// Grapple variation selector.
AVP_Blueprint_ParameterEnumerator_GrappleSelector::AVP_Blueprint_ParameterEnumerator_GrappleSelector()
: m_uNumVariations(0)
, m_uSelectedVariation(0)
{
}

bool AVP_Blueprint_ParameterEnumerator_GrappleSelector::Enumerate(const u_int uIndex, const Asura_Value& xValue)
{
	Asura_Assert(m_uNumVariations < MAX_GRAPPLE_VARIATIONS, "Too many grapple variations, ignoring the rest.");
	if (m_uNumVariations >= MAX_GRAPPLE_VARIATIONS) { return false; }

	Asura_Hash_ID uBlueprintID = xValue.GetHashIDValue();
	if (uBlueprintID == ASURA_HASH_ID_UNSET) { return true; } // It's probably valid to not have a blueprint, the inheritance might be easier to set up that way.

	// Store off this ID.
	m_auGrappleBlueprints[m_uNumVariations++] = uBlueprintID;

	return true;
}

void AVP_Blueprint_ParameterEnumerator_GrappleSelector::EndEnumeration(const bool bSuccess)
{
	// Hopefully it shouldn't matter whether the enumeration was successful, we just need to pick a variation.
	if (m_uNumVariations > 0)
	{
		m_uSelectedVariation = Asura_Random::GetIntU(0, m_uNumVariations-1);
	}
}

Asura_Hash_ID AVP_Blueprint_ParameterEnumerator_GrappleSelector::GetSelectedGrappleBlueprint() const
{
	if (m_uSelectedVariation >= m_uNumVariations) // Should just mean there were no variations, hence the assert.
	{
		Asura_Assert(m_uNumVariations == 0, "Failed to select a grapple variation, but some did exist?");
		return ASURA_HASH_ID_UNSET;
	}
	else
	{
		return m_auGrappleBlueprints[m_uSelectedVariation];
	}
}

/// End of file. ///
