
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Alien_JumpHandler.cpp
// Author : Andy Tate
// Date   : 4-Sep-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/Vertebrates/Alien/AVP_Alien_JumpHandler.h"

// Asura independent header-files...
#include "Blueprint/Asura_Blueprint_System.h"
#if !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Debugging/Asura_DebugLineRender.h"
#include "Physics/Asura_Physics_Environment.h"
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Physics/Asura_Physics.h"
#if defined(ASURA_USE_DEBUG_TEXT)
#include "Debugging/Asura_DebugText.h"
#endif	// defined(ASURA_USE_DEBUG_TEXT)
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/Vertebrates/Alien/AVP_Alien_Focus.h"
#include "Actor/Vertebrates/Alien/AVP_Alien_PlayerAnimStateData.h"
#include "Actor/Vertebrates/Alien/AVP_Alien_HUDHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Description.h"
#include "Attractors/AVP_Attractors_System.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Alien/AVP_ClientActor_Alien.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ClientActor_Human.h"
#include "Entity/Vent/AVP_ClientEntity_Vent.h"
#include "Mesh/AVP_PolygonFlags.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesJump.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesVent.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR bool	dbg_bAlienJumpHandlerRenderDebugInfo	= false;	//true;	//

DVSVAR float dvs_fMinJumpAlongSurfaceDistance = 5.0f;

static const Asura_AVector3	g_xUp(0.0f, -1.0f, 0.0f);


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#if !defined(ASURA_NO_DEBUG_OUTPUT)
const u_int AVP_Alien_JumpHandler::s_auJumpTypeDebugColours[NUM_ALIEN_JT] =
{
	0x00000000,		// JT_NONE
	0x80FF0000,		// JT_UNDEFINED

	0x80FFFF00,		// JT_JUMP_CONCAVE_90_FLOOR_TO_WALL
	0x80FFFF00,		// JT_JUMP_CONCAVE_90_WALL_TO_FLOOR
	0x80FFFF00,		// JT_JUMP_CONCAVE_90_CEILING_TO_WALL
	0x80FFFF00,		// JT_JUMP_CONCAVE_90_WALL_TO_CEILING
	0x80FFFF00,		// JT_JUMP_CONCAVE_90_WALL_TO_WALL

	0x80FF00FF,		// JT_JUMP_180_FLOOR_TO_CEILING,
	0x80FF00FF,		// JT_JUMP_180_CEILING_TO_FLOOR,
	0x80FF00FF,		// JT_JUMP_180_WALL_TO_WALL,

	0x800000FF,		// JT_DROP

	0x80FFFFFF,		// JT_ENTER_VENT
};
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

void AVP_Alien_JumpHandler::CreateConsoleVars()
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.MinLeapAlongSurfaceDistance", &dvs_fMinJumpAlongSurfaceDistance, 0.0f, 100.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_JumpHandler::CalculateJumpType
// Author       : Andy Tate
// Date         : 21-Sep-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Alien_JumpHandler::AlienJumpType AVP_Alien_JumpHandler::CalculateJumpType(const Asura_Vector_3& xStartNormal,
																			  const Asura_Vector_3& xEndNormal) const
{
	const AVP_Actor_PhysicsObjectHandler::GroundType	eStartGroundType	= AVP_Actor_PhysicsObjectHandler::CalculateGroundType(xStartNormal);
	const AVP_Actor_PhysicsObjectHandler::GroundType	eEndGroundType		= AVP_Actor_PhysicsObjectHandler::CalculateGroundType(xEndNormal);

	// Where are we jumping from?
	switch(eStartGroundType)
	{
		case AVP_Actor_PhysicsObjectHandler::ON_FLOOR:
		{
			switch(eEndGroundType)
			{
				case AVP_Actor_PhysicsObjectHandler::ON_FLOOR:
				{
					return(JT_JUMP_ALONG_SURFACE);
				}

				case AVP_Actor_PhysicsObjectHandler::ON_CEILING:
				{
					return(JT_JUMP_180_FLOOR_TO_CEILING);
				}

				case AVP_Actor_PhysicsObjectHandler::ON_WALL:
				{
					return(JT_JUMP_CONCAVE_90_FLOOR_TO_WALL);
				}

				default:
				{
					Asura_Assert(false, "Unexpected ground-type for end-normal.");
					return(JT_NORMAL);
				}
			}		// End of switch.
		}

		case AVP_Actor_PhysicsObjectHandler::ON_CEILING:
		{
			switch(eEndGroundType)
			{
				case AVP_Actor_PhysicsObjectHandler::ON_FLOOR:
				{
					return(JT_JUMP_180_CEILING_TO_FLOOR);
				}

				case AVP_Actor_PhysicsObjectHandler::ON_CEILING:
				{
					return(JT_JUMP_ALONG_SURFACE);
				}

				case AVP_Actor_PhysicsObjectHandler::ON_WALL:
				{
					return(JT_JUMP_CONCAVE_90_CEILING_TO_WALL);
				}

				default:
				{
					Asura_Assert(false, "Unexpected ground-type for end-normal.");
					return(JT_NORMAL);
				}
			}		// End of switch.
		}

		case AVP_Actor_PhysicsObjectHandler::ON_WALL:
		{
			switch(eEndGroundType)
			{
				case AVP_Actor_PhysicsObjectHandler::ON_FLOOR:
				{
					return(JT_JUMP_CONCAVE_90_WALL_TO_FLOOR);
				}

				case AVP_Actor_PhysicsObjectHandler::ON_CEILING:
				{
					return(JT_JUMP_CONCAVE_90_WALL_TO_CEILING);
				}

				case AVP_Actor_PhysicsObjectHandler::ON_WALL:
				{
					const float	fDotProduct	= xStartNormal * xEndNormal;

					// Wall - but what type of wall is this?
					if(fDotProduct > 0.707f)
					{
						// Roughly the same orientation...
						return(JT_JUMP_ALONG_SURFACE);
					}
					else if (fDotProduct < -0.707f)
					{
						// Roughly 180 degrees...
						return(JT_JUMP_180_WALL_TO_WALL);
					}
					else
					{
						// Between 45 and 135 degrees - assume 90 degrees...
						return(JT_JUMP_CONCAVE_90_WALL_TO_WALL);
					}
				}

				default:
				{
					Asura_Assert(false, "Unexpected ground-type for end-normal.");
					return(JT_NORMAL);
				}
			}		// End of switch.
		}

		default:
		{
			Asura_Assert(false, "Unexpected ground-type for start-normal.");
			return(JT_NORMAL);
		}
	}		/// End of switch.
}

/// End of function 'AVP_Alien_JumpHandler::CalculateJumpType'. ///





////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_JumpHandler::CheckForJumpType
// Author       : Andy Tate
// Date         : 4-Sep-2007
////////////////////////////////////////////////////////////////////////////////
DEBUGVAR float dbg_fCanJumpFloorOffset = -0.1f;
DEBUGVAR float dbg_fAlienDuctSwitchLOSDistance = 1.0f;

bool AVP_Alien_JumpHandler::CheckForJumpType(void)
{
	Asura_Assert(m_xOwner.IsPlayer(), "Only players should be checking for jumps in this function.");
	
	AVP_Alien_HUDHandler::SetSurfaceInRange( false );
	
	const Asura_Movement_System*		pxMovementSystem	= m_xOwner.GetMovementSystem();

	Asura_Vector_3						xCameraPosition;
	Asura_Matrix_3x3					xCameraOrientation;

	// check for vent usage
	const AVP_Alien_Focus* pxFocus = static_cast< const AVP_Alien_Focus* >( m_xOwner.GetFocus() );
	if ( pxFocus != NULL && pxFocus->GetVentTarget() != ASURA_GUID_UNREGISTERED )
	{
		m_uJumpTargetGuid = pxFocus->GetVentTarget();
		m_uJumpType	= JT_ENTER_VENT;
		return true;
	}

	const u_int uAttractorID = m_xOwner.GetConstrainedMovementAttractorID();
	if (uAttractorID != uASURA_ATTRACTOR_ID_NONE)
	{
		const Asura_Attractor_Compiled* pxAttractor = Asura_Attractor_System::GetInstance()->GetAttractorByID(uAttractorID);
		if (pxAttractor && pxAttractor->GetType() == uAVP_ATTRACTOR_TYPE_DUCT)
		{
			const Asura_Vector_3& xPosition = m_xOwner.GetCurrentPosition();
			Asura_Vector_3 xNormal;
			pxAttractor->GetNormal(xNormal);
			DuctAttractorCallbackData xData(xNormal, xPosition);

			Asura_Attractor_System::GetInstance()->CollectAttractors(xPosition, xNormal, 0.0f, dbg_fAlienDuctSwitchLOSDistance, TestDuctAttractors_Callback, &xData);

			if (xData.m_uNewAttractor != uASURA_ATTRACTOR_ID_NONE)
			{
				m_uJumpType	= JT_DUCT_SURFACE_SWITCH;
				m_xJumpDestinationNormal = xData.m_xNewAttractorNormal;
				m_xJumpDestinationPosition = xData.m_xPosOnNewAttractor;
				m_fJumpDestinationDistance = (m_xJumpDestinationPosition - xPosition).MagnitudeSquared();
				return true;
			}

			return false;
		}
	}

	// Always perform the LOS-test from the camera...
	Asura_Hash_ID uCameraHashID = AVP_Vertebrate_Description::GetMarkerPointHashID( AVP_Vertebrate_Description::MPID_EYES ); 
	
	if( (pxMovementSystem != NULL) && !pxMovementSystem->IsMoveInProgress() &&
		m_xOwner.GetWorldSpaceMarkerPoint( uCameraHashID, &xCameraPosition, &xCameraOrientation) ) // Camera
	{
		const Asura_Matrix_3x3&		xStartOrientation	= m_xOwner.GetCurrentOrientation();

		const Asura_Vector_3		xStartNormal(-xStartOrientation.e12, -xStartOrientation.e22, -xStartOrientation.e32);
		const Asura_Vector_3		xCameraDirection(xCameraOrientation.e13, xCameraOrientation.e23, xCameraOrientation.e33);

		// Initialise the variables...
		m_fJumpDestinationDistance	= m_fMaxLOSRange;
		m_uJumpTargetGuid			= ASURA_GUID_UNREGISTERED;

		const Asura_Guid			uOwnerGuid	= m_xOwner.GetGuid();
		Asura_LOS_InputData			xLOSInput(xCameraPosition, xCameraDirection, uOwnerGuid, m_fMaxLOSRange, true, AsuraPolygonFlag_NoCollisions | AVP_PolygonFlag_IgnoreForAlienCollision | AVP_PolygonFlag_IgnoreForPlayerCollision );

		const Asura_LOS_OutputData*	pxLOSOutput			= Asura_LOS::TestAgainstEverything(&xLOSInput);;
		// Make sure we aren't going to jump to an "ignore for reorientation" poly.
		if ((pxLOSOutput->m_usPolyFlags & AVP_PolygonFlag_IgnoreForReorientation) != 0)
		{
			return false;
		}

		const float					fDistance			= pxLOSOutput->m_fDistanceToIntersection;
		const Asura_Guid			uTargetGuid			= pxLOSOutput->m_uHitEntity;
		const bool					bHasTargetEntity	= ((uTargetGuid != 0) && (uTargetGuid != ASURA_GUID_UNREGISTERED));

		// Did we have an intersection?
		if (fDistance < m_fMaxLOSRange) 
		{
			bool bSuccess;

			if(bHasTargetEntity)
			{
				// Check physical objects and containers.
				bSuccess = AVP_ClientEntity_PhysicalObject::CanUseForSurfaceReorientation(uTargetGuid);
			}
			else
			{
				bSuccess = pxLOSOutput->m_bHitEnvironment;
			}

			const Asura_Vector_3& xNormal = pxLOSOutput->m_xNormal;
			const Asura_Vector_3 xDestination = xLOSInput.m_xOrigin + (xLOSInput.m_xDirection * fDistance);
			const AlienJumpType eJumpType = CalculateJumpType(xStartNormal, xNormal);

			if (bSuccess && (eJumpType != JT_UNDEFINED) && (eJumpType != JT_NORMAL))
			{
				// Check if we can actually place the box where we want to.
				Asura_AMatrix3x3 xTargetOri;
				CalculateTargetOrientation(eJumpType, m_xOwner, xNormal, xTargetOri);

				const bool bOnFloor = AVP_Actor_PhysicsObjectHandler::CalculateGroundType(xNormal) == AVP_Actor_PhysicsObjectHandler::ON_FLOOR;
				const AVP_Actor_AnimStateData* pxStateData = AVP_Alien_PlayerAnimStateData::GetSingleton().GetAnimStateData( m_xOwner.GetAnimStateBlueprintType(), true, bOnFloor, false, false, false, AVP_Vertebrate_PlayerAnimStateHandler::MOVESPEED_IDLE);
				const Asura_Physics_Object* pxPhysObj = m_xOwner.GetPhysicsObject();
				if (pxStateData && pxPhysObj)
				{
					const AVP_Collision_Descriptor* pxCollDesc = AVP_Collision_Descriptor::GetDataByHash(pxStateData->GetCollisionDescriptorHash());
					const Asura_Collision_Object* pxCollObj = pxCollDesc ? pxCollDesc->GetCollisionBox() : NULL;

					if (pxCollObj)
					{
						Asura_Matrix_3x3 xUnalignedOri;
						xTargetOri.ToMat3(xUnalignedOri);

						const Asura_Vector_3 xTestPos = xDestination + Asura_Vector_3(Asura_AVector3(0.0f, dbg_fCanJumpFloorOffset, 0.0f) * xTargetOri);
						const Asura_Handle iHandleToIgnore = pxPhysObj->GetHandle();
						bSuccess = Asura_Physics_System::CanPlaceObjectAt(xTestPos, xUnalignedOri, *pxCollObj, pxPhysObj->GetFlagsOfPolysToIgnore(), iHandleToIgnore, 0xFFFF, true, true);
					}
				}
			}

			if (bSuccess)
			{
#if !defined(ASURA_NO_DEBUG_OUTPUT)
				m_uJumpDestinationFaceIndex	= bHasTargetEntity ? 0 : pxLOSOutput->m_uEnvironmentFaceIndex;
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)
				// Update the destination info...
				m_xJumpDestinationNormal	= xNormal;
				m_xJumpDestinationPosition	= xDestination;
				m_fJumpDestinationDistance	= fDistance;

				m_uJumpType = eJumpType;
				AVP_Alien_HUDHandler::SetSurfaceInRange( true );

				return true;
			}
			else
			{
				m_uJumpType	= JT_NORMAL;
				return(false);
			}
		}
		else
		{
			m_uJumpType	= JT_NORMAL;
			return(false);
		}
	}
	else
	{
		return(false);
	}
}

/// End of function 'AVP_Alien_JumpHandler::CheckForJumpType'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_JumpHandler::AVP_Alien_JumpHandler
// Author       : Andy Tate
// Date         : 4-Sep-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Alien_JumpHandler::AVP_Alien_JumpHandler(AVP_ClientActor_Vertebrate& m_xOwner)
: PARENT(m_xOwner)
, m_xJumpDestinationNormal(Asura_Maths::ZeroVector3)
, m_xJumpDestinationPosition(Asura_Maths::ZeroVector3)
, m_uJumpTargetGuid(ASURA_GUID_UNREGISTERED)
, m_fJumpDestinationDistance(0.0f)
, m_fMaxLOSRange(10.0f)
#if !defined(ASURA_NO_DEBUG_OUTPUT)
, m_uJumpDestinationFaceIndex(0)
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)
{
	const Asura_Hash_ID		uBlueprintHash	= m_xOwner.GetBlueprintHashID();
	const Asura_Hash_ID		uTypeHash		= uAVP_BLUEPRINT_TYPE_PLAYERS;

	const Asura_Blueprint*	pxBlueprint		= Asura_Blueprint_System::GetBlueprintByHashID(uTypeHash, uBlueprintHash);
	if(pxBlueprint)
	{
		Asura_Value xValue;

		if (pxBlueprint->GetParameterValueByHashID(0xF0978876, xValue))	// Asura_GetHashID("Jump.LOSLength", xValue))
		{
			m_fMaxLOSRange = xValue.GetFloatValue();
		}
	}
}

/// End of function 'AVP_Alien_JumpHandler::AVP_Alien_JumpHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_JumpHandler::Update
// Author       : Andy Tate
// Date         : 4-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Alien_JumpHandler::Update(void)
{
	// Call the parent function...
	PARENT::Update();

	// Can we jump?
	if(CheckForJumpType())
	{
		// We've found a type of jump - nothing more to do...
	}
	else
	{
		m_uJumpType	= JT_NORMAL;
	}
}

/// End of function 'AVP_Alien_JumpHandler::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_JumpHandler::Render
// Author       : Andy Tate
// Date         : 4-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Alien_JumpHandler::Render(void) const
{
	// Call the base-function...
	PARENT::Render();

	// Do we need to do any debug rendering?
	if(dbg_bAlienJumpHandlerRenderDebugInfo)
	{
#if !defined(ASURA_NO_DEBUG_OUTPUT)
		switch(m_uJumpType)
		{
			case JT_NORMAL:
			{
				break;
			}

			case JT_UNDEFINED:

			case JT_JUMP_CONCAVE_90_FLOOR_TO_WALL:
			case JT_JUMP_CONCAVE_90_WALL_TO_FLOOR:
			case JT_JUMP_CONCAVE_90_CEILING_TO_WALL:
			case JT_JUMP_CONCAVE_90_WALL_TO_CEILING:
			case JT_JUMP_CONCAVE_90_WALL_TO_WALL:

			case JT_JUMP_180_FLOOR_TO_CEILING:
			case JT_JUMP_180_CEILING_TO_FLOOR:
			case JT_JUMP_180_WALL_TO_WALL:
			{
				const Asura_AVector3&	xVertex0	= Asura_Physics_Environment::GetFaceVertex(m_uJumpDestinationFaceIndex, 0);
				const Asura_AVector3&	xVertex1	= Asura_Physics_Environment::GetFaceVertex(m_uJumpDestinationFaceIndex, 1);
				const Asura_AVector3&	xVertex2	= Asura_Physics_Environment::GetFaceVertex(m_uJumpDestinationFaceIndex, 2);

				const u_int				uColour		= s_auJumpTypeDebugColours[m_uJumpType];

				Asura_DebugLineRender::AddArrow(m_xJumpDestinationPosition, Asura_Vector_3(m_xJumpDestinationPosition + m_xJumpDestinationNormal), uColour, 0.125f, 0.100f);

				Asura_DebugLineRender::AddLine(xVertex0, xVertex1, uColour);
				Asura_DebugLineRender::AddLine(xVertex1, xVertex2, uColour);
				Asura_DebugLineRender::AddLine(xVertex2, xVertex0, uColour);
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected jump-type when debug-rendering.");
				break;
			}
		}		// End of switch.
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)

#if defined(ASURA_USE_DEBUG_TEXT)
		// Output the GUID...
		Asura_DebugText::AddTextBlock(2, 2, 0xFFFFFFFF, 0, "m_fJumpDestinationDistance = %.3f", m_fJumpDestinationDistance);
		Asura_DebugText::AddTextBlock(2, 3, 0xFFFFFFFF, 0, "m_uJumpTargetGuid = %d", m_uJumpTargetGuid);
#endif	// defined(ASURA_USE_DEBUG_TEXT)
	}
}

/// End of function 'AVP_Alien_JumpHandler::Render'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_JumpHandler::HandleJump
// Author       : Andy Tate
// Date         : 4-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Alien_JumpHandler::HandleJump(const float fJumpValue)
{
	const AVP_Vertebrate_Focus* pxFocus = m_xOwner.GetFocus();
	if (pxFocus && pxFocus->GetActive() && (pxFocus->GetMeleeTarget() != ASURA_GUID_UNREGISTERED))
	{
		return PARENT::HandleJump( fJumpValue );
	}

	if( (fJumpValue > 0.0f) && ( (m_xOwner.GetConstrainedMovementAttractorID() != uASURA_ATTRACTOR_ID_NONE) || (m_xOwner.GetGroundType() != AVP_Actor_PhysicsObjectHandler::OFF_GROUND)) )
	{
		// See if we already have a movement-controlling move in progress, and disallow jumps if we do.
		const Asura_Movement_System* pxMovementSystem = m_xOwner.GetMovementSystem();
		const AVP_Move* pxCurrentMove = pxMovementSystem ? static_cast<const AVP_Move*>(pxMovementSystem->GetCurrentMove()) : NULL;
		if (pxCurrentMove && (pxCurrentMove->IsDisableMovement() || (pxCurrentMove->GetAVPAnimationData() && !pxCurrentMove->GetAVPAnimationData()->GetDontCalculateVelocity())))
		{
			return false;
		}

// 		AVP_Alien_StateHandler&	xStateHandler	= static_cast<AVP_ClientActor_Alien*>(&m_xOwner)->GetAlienStateHandler();

		switch(m_uJumpType)
		{
			case JT_NORMAL:
			case JT_UNDEFINED:
			{
				if ( m_xOwner.GetGroundType() == AVP_Actor_PhysicsObjectHandler::ON_FLOOR )
				{
					return PARENT::HandleJump( fJumpValue );
				}
				else
				{
					return false;
// 					m_uJumpType = JT_DROP;
// 					return SendJumpMoveStartMessage( m_uOwnerGuid, m_uJumpType, m_xJumpDestinationPosition, m_xJumpDestinationNormal ) != ASURA_HASH_ID_UNSET;
				}
			}

			case JT_JUMP_ALONG_SURFACE:
			{
				if ( !m_xOwner.IsShortRouteAvailable( m_xJumpDestinationPosition, dvs_fMinJumpAlongSurfaceDistance, m_fMaxLOSRange ) )
				{
					return SendJumpMoveStartMessage( m_xOwner.GetGuid(), m_uJumpType, m_xJumpDestinationPosition, m_xJumpDestinationNormal ) != ASURA_HASH_ID_UNSET;
				}
				else
				{
					if ( m_xOwner.GetGroundType() == AVP_Actor_PhysicsObjectHandler::ON_FLOOR )
					{
						return PARENT::HandleJump( fJumpValue );
					}
					else
					{
						return false;
						// 					m_uJumpType = JT_DROP;
						// 					return SendJumpMoveStartMessage( m_uOwnerGuid, m_uJumpType, m_xJumpDestinationPosition, m_xJumpDestinationNormal ) != ASURA_HASH_ID_UNSET;
					}
				}
			}

			case JT_JUMP_CONCAVE_90_FLOOR_TO_WALL:
			case JT_JUMP_CONCAVE_90_WALL_TO_FLOOR:
			case JT_JUMP_CONCAVE_90_CEILING_TO_WALL:
			case JT_JUMP_CONCAVE_90_WALL_TO_CEILING:
			case JT_JUMP_CONCAVE_90_WALL_TO_WALL:

			case JT_JUMP_180_FLOOR_TO_CEILING:
			case JT_JUMP_180_CEILING_TO_FLOOR:
			case JT_JUMP_180_WALL_TO_WALL:

			case JT_DUCT_SURFACE_SWITCH:
			{
				const Asura_Guid	uOwnerGuid	= m_xOwner.GetGuid();
				
				return SendJumpMoveStartMessage( uOwnerGuid, m_uJumpType, m_xJumpDestinationPosition, m_xJumpDestinationNormal ) != ASURA_HASH_ID_UNSET;
			}

			case JT_ENTER_VENT:
			{
				const AVP_ClientEntity_Vent* pxVent	= AVP_ClientEntity_Vent::FindEntity( m_uJumpTargetGuid );
				Asura_AVector3 xABeginPos;
				Asura_AVector3 xAFinalPos;
				Asura_AMatrix3x3 xABeginOri;
				Asura_AMatrix3x3 xAFinalOri;
				if ( pxVent != NULL && pxVent->IsEntityActive() )
				{
					Asura_Vector_3 xOffset = m_xOwner.GetCurrentPosition() - pxVent->GetCurrentPosition();
					const Asura_Matrix_3x3 xVentOri = pxVent->GetCurrentOrientation();
					const bool bInFront = (xOffset * Asura_Vector_3(xVentOri.e13, xVentOri.e23, xVentOri.e33)) >= 0.0f;

					const bool bSuccess = bInFront ? pxVent->GetBestEntryPoint( m_xOwner, xABeginPos, xABeginOri, xAFinalPos, xAFinalOri ) : pxVent->GetBestExitPoint( m_xOwner, xABeginPos, xABeginOri, xAFinalPos, xAFinalOri );

					if ( bSuccess )
					{
						const Asura_Guid uOwnerGuid	= m_xOwner.GetGuid();

						Asura_Vector_3 xBeginPos( xABeginPos );
						Asura_Vector_3 xFinalPos( xAFinalPos );
						Asura_Matrix_3x3 xBeginOri;
						xABeginOri.ToMat3( xBeginOri );
						Asura_Matrix_3x3 xFinalOri;
						xAFinalOri.ToMat3( xFinalOri );
						return ASURA_HASH_ID_UNSET != SendVentMoveStartMessage( uOwnerGuid, xBeginPos, xBeginOri, xFinalPos, xFinalOri );
					}
				}
				Asura_Assert( false, "Unable to calculate vent mode data" );
				return false;
			}
			
			default:
			{
				Asura_Assert( false, "Unexpected type of jump-type." );
				return false;
			}
		}		// End of switch.
	}
	else
	{
		return false;
	}
}

Asura_Hash_ID AVP_Alien_JumpHandler::SendJumpMoveStartMessage( const Asura_Guid uOwnerGuid, const u_int uJumpType, const Asura_Vector_3& xTargetPosition, const Asura_Vector_3& xTargetNormal )
{
	Asura_Hash_ID uMoveHash( ASURA_HASH_ID_UNSET );
	switch ( uJumpType )
	{
		case AVP_Alien_JumpHandler::JT_DROP:
		{
			uMoveHash = AVP_Move_Alien_JumpDropStart::GetStaticHash();
		}
		break;

		//case AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_FLOOR_TO_WALL:
		//{
		//	uMoveHash = AVP_Move_Alien_JumpSurfaceConcave90Begin::GetStaticHash();
		//	break;
		//}

		case JT_JUMP_ALONG_SURFACE:
		{
			// for now piggyback on the concave 90 move maybe we'll do a new move later, maybe
		}

		case AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_FLOOR_TO_WALL:
		case AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_WALL_TO_FLOOR:
		case AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_CEILING_TO_WALL:
		case AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_WALL_TO_CEILING:
		case AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_WALL_TO_WALL:
		{
			uMoveHash = AVP_Move_Alien_JumpSurfaceConcave90::GetStaticHash();
			break;
		}

		case AVP_Alien_JumpHandler::JT_JUMP_180_FLOOR_TO_CEILING:
		case AVP_Alien_JumpHandler::JT_JUMP_180_CEILING_TO_FLOOR:
		case AVP_Alien_JumpHandler::JT_JUMP_180_WALL_TO_WALL:
		{
			uMoveHash = AVP_Move_Alien_JumpSurface180::GetStaticHash();
			break;
		}

		case AVP_Alien_JumpHandler::JT_DUCT_SURFACE_SWITCH:
		{
			uMoveHash = AVP_Move_Alien_DuctSurfaceSwitch::GetStaticHash();
			break;
		}

		default:
		{
			Asura_Assert( false, "Unhandled type of jump for move message creation" );
			return NULL;
		}
	}

	if ( ASURA_HASH_ID_UNSET != uMoveHash )
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_MOVE_START, sizeof( AVP_Message_Alien_Move_JumpSurfaceStart ) );

		if ( pxMessage != NULL )
		{
			AVP_Message_Alien_Move_JumpSurfaceStart* pxData = static_cast< AVP_Message_Alien_Move_JumpSurfaceStart* >( pxMessage->GetData() );

			pxMessage->SetFromGuid( uOwnerGuid );
			pxMessage->SetToGuid( uOwnerGuid );

			pxData->ClearFlags();
			pxData->SetMoveID( uMoveHash );
			pxData->SetForceAbort( true );
			pxData->SetIgnoredByOwner( false );

			pxData->SetDestinationNormal( xTargetNormal );
			pxData->SetDestinationPosition( xTargetPosition );
			// Quick hack - forward the message directly to the client...
			AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity( uOwnerGuid );
			if ( pxClient->HandleMessage( pxMessage ) )
			{
				// ...before informing all other client's as well...
				pxData->SetIgnoredByOwner( true );
				Asura_Network::Send( pxMessage );
			}
			else
			{
				Asura_Assert( false, "Client failed to handle message." );
				Asura_Network::DestroyMessage(pxMessage);
				uMoveHash = ASURA_HASH_ID_UNSET;
			}

			pxClient->SetConstrainedMovementAttractorID(uASURA_ATTRACTOR_ID_NONE);
		}
		else
		{
			Asura_Assert( false, "Unable to create message" );
			uMoveHash = ASURA_HASH_ID_UNSET;
		}
	}
	else
	{
		Asura_Assert( false, "Unable to get move hash" );
		uMoveHash = ASURA_HASH_ID_UNSET;
	}
	return uMoveHash;
}

// AI version of the function, dictating animations from the gamescene data it has.
Asura_Hash_ID AVP_Alien_JumpHandler::SendVentMoveStartMessage( const Asura_Guid uOwnerGuid,
															   const Asura_Vector_3& xBeginPosition, const Asura_Matrix_3x3& xBeginOri,
															   const Asura_Vector_3& xFinalPosition, const Asura_Matrix_3x3& xFinalOri,
															   Asura_Hash_ID uBeginAnimHash, Asura_Hash_ID uFinalAnimHash )
{
	Asura_Hash_ID uMoveHash = AVP_Move_Alien_AI_TraverseVent_Begin::GetStaticHash();
	Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_MOVE_START, sizeof( AVP_Message_Alien_Move_VentStart ) );
	if ( pxMessage != NULL )
	{
		AVP_Message_Alien_Move_VentStart* pxData = static_cast< AVP_Message_Alien_Move_VentStart* >( pxMessage->GetData() );

		pxMessage->SetFromGuid( uOwnerGuid );
		pxMessage->SetToGuid( uOwnerGuid );

		pxData->ClearFlags();
		pxData->SetMoveID( uMoveHash );
		pxData->SetForceAbort( true );
		pxData->SetIgnoredByOwner( false );

		pxData->SetBeginPos( xBeginPosition );
		pxData->SetBeginOri( xBeginOri );
		pxData->SetFinalPos( xFinalPosition );
		pxData->SetFinalOri( xFinalOri );
		pxData->SetBeginAnimHash( uBeginAnimHash );
		pxData->SetFinalAnimHash( uFinalAnimHash );

		// Quick hack - forward the message directly to the client...
		AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity( uOwnerGuid );
		if ( pxClient->HandleMessage( pxMessage ) )
		{
			// ...before informing all other client's as well...
			pxData->SetIgnoredByOwner( true );
			Asura_Network::Send( pxMessage );

			return uMoveHash;
		}
		else
		{
			Asura_Assert( false, "Client failed to handle message." );
			Asura_Network::DestroyMessage(pxMessage);
			return ASURA_HASH_ID_UNSET;
		}
	}
	else
	{
		Asura_Assert( false, "Unable to create message" );
		return ASURA_HASH_ID_UNSET;
	}
}

// Player version of the function, not bothering with any anim ids.
Asura_Hash_ID AVP_Alien_JumpHandler::SendVentMoveStartMessage( const Asura_Guid uOwnerGuid,
															   const Asura_Vector_3& xBeginPosition, const Asura_Matrix_3x3& xBeginOri,
															   const Asura_Vector_3& xFinalPosition, const Asura_Matrix_3x3& xFinalOri )
{
	Asura_Hash_ID uMoveHash = AVP_Move_Alien_Player_TraverseVent_Begin::GetStaticHash();
	Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_MOVE_START, sizeof( AVP_Message_Alien_Move_VentStart ) );
	if ( pxMessage != NULL )
	{
		AVP_Message_Alien_Move_VentStart* pxData = static_cast< AVP_Message_Alien_Move_VentStart* >( pxMessage->GetData() );

		pxMessage->SetFromGuid( uOwnerGuid );
		pxMessage->SetToGuid( uOwnerGuid );

		pxData->ClearFlags();
		pxData->SetMoveID( uMoveHash );
		pxData->SetForceAbort( true );
		pxData->SetIgnoredByOwner( false );

		pxData->SetBeginPos( xBeginPosition );
		pxData->SetBeginOri( xBeginOri );
		pxData->SetFinalPos( xFinalPosition );
		pxData->SetFinalOri( xFinalOri );
		pxData->SetBeginAnimHash( ASURA_HASH_ID_UNSET );
		pxData->SetFinalAnimHash( ASURA_HASH_ID_UNSET );

		// Quick hack - forward the message directly to the client...
		AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity( uOwnerGuid );
		if ( pxClient->HandleMessage( pxMessage ) )
		{
			// ...before informing all other client's as well...
			pxData->SetIgnoredByOwner( true );
			Asura_Network::Send( pxMessage );

			// Also make sure we leave anything we're already on.
			pxClient->SetConstrainedMovementAttractorID(uASURA_ATTRACTOR_ID_NONE);
			return uMoveHash;
		}
		else
		{
			Asura_Assert( false, "Client failed to handle message." );
			Asura_Network::DestroyMessage(pxMessage);
			return ASURA_HASH_ID_UNSET;
		}
	}
	else
	{
		Asura_Assert( false, "Unable to create message" );
		return ASURA_HASH_ID_UNSET;
	}
}

void AVP_Alien_JumpHandler::CalculateTargetOrientation(const AVP_Alien_JumpHandler::AlienJumpType eType, const AVP_ClientActor_Vertebrate& xVert, const Asura_Vector_3& xTargetNormal, Asura_AMatrix3x3& xTargetOrientation)
{
	switch (eType)
	{
		default:
		{
			Asura_Assert(false, "Cannot calculate a target orientation for this jump type");
			xTargetOrientation.Identity();
		}
		break;

		case JT_JUMP_ALONG_SURFACE:
		case JT_JUMP_180_FLOOR_TO_CEILING:
		case JT_JUMP_180_CEILING_TO_FLOOR:
		case JT_JUMP_180_WALL_TO_WALL:
		{
			Asura_AVector3 xTargetForwards;
			xVert.GetAimDirection(xTargetForwards);
			Asura_AVector3 xTargetDown(-xTargetNormal);
			Asura_AVector3 xTargetSide;

			// Re-create our orthogonal vectors...
			xTargetSide		= xTargetDown ^ xTargetForwards;
			xTargetForwards	= xTargetSide ^ xTargetDown;

			xTargetSide.Normalise();
			xTargetForwards.Normalise();

			// ...and use these to re-create the target orientation...
			xTargetOrientation.SetColumn0(xTargetSide);
			xTargetOrientation.SetColumn1(xTargetDown);
			xTargetOrientation.SetColumn2(xTargetForwards);
		}
		break;

		case JT_JUMP_CONCAVE_90_FLOOR_TO_WALL:
		case JT_JUMP_CONCAVE_90_WALL_TO_FLOOR:
		case JT_JUMP_CONCAVE_90_CEILING_TO_WALL:
		case JT_JUMP_CONCAVE_90_WALL_TO_CEILING:
		case JT_JUMP_CONCAVE_90_WALL_TO_WALL:
		{
			const Asura_Matrix_3x3&	xCurrentOrientation = xVert.GetCurrentOrientation();
			Asura_AVector3 xForwards;
			xVert.GetAimDirection(xForwards);

			const Asura_AVector3 xCurrentDown(xCurrentOrientation.e12, xCurrentOrientation.e22, xCurrentOrientation.e32);
			xForwards -= xCurrentDown * (xCurrentDown * xForwards);
			if (xForwards.ApproximatelyEquals(Asura_Maths::ZeroAVector3))
			{
				xForwards = Asura_AVector3(xCurrentOrientation.e13, xCurrentOrientation.e23, xCurrentOrientation.e33);
			}
			else
			{
				xForwards.Normalise();
			}
			// The above should already have made down and forwards perpendicular.
			Asura_AVector3 xRight = xCurrentDown ^ xForwards;
			Asura_AMatrix3x3 xAimOri;
			xAimOri.SetColumn0(xRight);
			xAimOri.SetColumn1(xCurrentDown);
			xAimOri.SetColumn2(xForwards);

			Asura_Quat xAimQuat;
			xAimQuat.FromMatrix(xAimOri);

			Asura_Quat xTargetQuat;
			Asura_Quat xRotation;

			// Find out how much we need to rotate and perform that rotation...
			xRotation.RotationFromAToB(Asura_Vector_3(-xCurrentDown), xTargetNormal);
			xTargetQuat	= xRotation * xAimQuat;

			xTargetQuat.ToMatrix(xTargetOrientation);
		}
		break;

		case JT_DROP:
		{
			Asura_AVector3 xRight;
			const Asura_AVector3 xDown(0.0f, 1.0f, 0.0f);
			Asura_AVector3 xForward;

			xVert.GetAimDirection(xForward);
			xRight = xDown ^ xForward;
			if (xRight.NormaliseReturningMagnitude() == 0.0f) 
			{
				xRight = Asura_AVector3(1.0f, 0.0f, 0.0f);
				xForward = Asura_AVector3(0.0f, 0.0f, 1.0f);
			}
			else
			{
				xForward = xRight ^ xDown;
				xForward.Normalise();
			}

			xTargetOrientation.SetColumn0(xRight);
			xTargetOrientation.SetColumn1(xDown);
			xTargetOrientation.SetColumn2(xForward);
		}
		break;
	}
}


DEBUGVAR float dbg_fDuctSurfaceSwitchMaxNormalDP = -0.5f;

void AVP_Alien_JumpHandler::TestDuctAttractors_Callback(void* pData, const Asura_Attractor_Compiled* const pxAttractor)
{
	if (pData && pxAttractor)
	{
		DuctAttractorCallbackData* pxData = static_cast<DuctAttractorCallbackData*>(pData);

		// Make sure it is an opposite surface.
		Asura_Vector_3 xNormal;
		pxAttractor->GetNormal(xNormal);
		if ((xNormal * pxData->m_xCurrentAttractorNormal) > dbg_fDuctSurfaceSwitchMaxNormalDP) { return; }

		// Find the point to jump to.
		Asura_Vector_3 xTargetPos;
		Asura_Vector_3 xOffset;
		const Asura_Vector_3& xNode0Pos = pxAttractor->GetNode(0)->GetPosition();
		const Asura_Vector_3 xAttractorVector = pxAttractor->GetNode(1)->GetPosition() - xNode0Pos;
		Asura_Maths::GetLeastOffsetBetweenLineSegmentAndRay(pxData->m_xCurrentPosition, pxData->m_xCurrentAttractorNormal, xNode0Pos, xAttractorVector, xOffset, &xTargetPos);

		// If we already have a target, see if this is closer.
		if (pxData->m_uNewAttractor != uASURA_ATTRACTOR_ID_NONE)
		{
			if ((pxData->m_xPosOnNewAttractor - pxData->m_xCurrentPosition).MagnitudeSquared() <= xOffset.MagnitudeSquared())
			{
				return;
			}
		}

		// New best.
		pxData->m_uNewAttractor = pxAttractor->GetID();
		pxData->m_xNewAttractorNormal = xNormal;
		pxData->m_xPosOnNewAttractor = xTargetPos;
	}
}

/// End of file. ///
