////////////////////////////////////////////////////////////////
//
// File:		AVP_Predator_MeleeCounter_System.cpp
// Author:		Chris Wallace
// Date:		25-Jan-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////

///////////////////////////////////////////////
// #includes
///////////////////////////////////////////////

#include "Core/Asura.h"

// This files header
#include "Actor/Vertebrates/Predator/AVP_Predator_MeleeCounter_System.h"

#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Group.h"
#include "GUIMenu/Asura_GUIMenu_Element_Image.h"
#include "Interface/Asura_CommandConsole.h"

#include "Actor/Vertebrates/AVP_Vertebrate_Description.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Alien/AVP_ServerActor_Alien.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ServerActor_Human.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ClientActor_Predator.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ServerActor_Predator.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predalien/AVP_ServerActor_Predalien.h"
#include "GUIMenu/AVP_GUIMenu_System.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/AVP_InventoryItem_PredatorCloak.h"
#include "Inventory/Items/AVP_InventoryItem_Weapon.h"
#include "Moves/Vertebrates/Predator/AVP_Predator_MovesMelee.h"

///////////////////////////////////////////////
// #defines
///////////////////////////////////////////////

DEBUGVAR float dbg_fPredatorFinisherImpulseStrength = 20.0f;

const char* AVP_Predator_MeleeCounter_System::s_szLeftQuickAttackBaseParameterName = "QuickAttackLeftStart";
const char* AVP_Predator_MeleeCounter_System::s_szRightQuickAttackBaseParameterName = "QuickAttackRightStart";

static const Asura_Vector_3 xDOWN(0.0f, 1.0f, 0.0f);

DVSVAR bool dvs_bUseHackedTestCounterIndicator = false;

bool AVP_Predator_MeleeCounter_System::s_bHackedTestCounterIndicatorOn = false;

/////////////////////////////////////////////////////////////////////
//
// Class :	AVP_Predator_MeleeCounter_System
//
//////////////////////////////////////////////////////////////////////

AVP_Predator_MeleeCounter_System::AVP_Predator_MeleeCounter_System(AVP_ServerActor_Predator& xOwner)
: AVP_Vertebrate_MeleeCounter_System(xOwner)
{
}

AVP_Predator_MeleeCounter_System::~AVP_Predator_MeleeCounter_System()
{
}


#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
void AVP_Predator_MeleeCounter_System::CreateConsoleVars()
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar(L"AVP.Predator.UseTestCounterIndicator", &dvs_bUseHackedTestCounterIndicator, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif //ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar(L"AVP.Predator.TestCounterIndicatorOn", &s_bHackedTestCounterIndicatorOn, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
}
#endif


void AVP_Predator_MeleeCounter_System::Update()
{
	PARENT::Update();

	// HACK - If this is the player, and their target can currently be countered, update the consolevar to display this.
	// Since this is just a test, the hack is fine. Any real implementation (if it's decided it's needed) will have to be completely different.
	if (m_xOwner.IsPlayer() && AVP_Settings::IsSinglePlayer())
	{
		s_bHackedTestCounterIndicatorOn = false;
		if (dvs_bUseHackedTestCounterIndicator)
		{
			const AVP_ServerActor_Vertebrate* pxTarget = AVP_ServerActor_Vertebrate::FindEntity(m_uCurrentTarget);
			if (pxTarget)
			{
				const AVP_Vertebrate_MeleeCounter_System* pxTargetMelee = pxTarget->GetMeleeSystem();
				if (pxTargetMelee)
				{
					Asura_Hash_ID uJunk1;
					u_int uJunk2;
					bool bJunk3;
					Asura_Hash_ID uJunk4;
					if (pxTargetMelee->GetCanCounter(m_xOwner, AVP_MELEEATTACK_TYPE_COUNTER, uJunk1, uJunk2, bJunk3, uJunk4))
					{
						// Check both our and their anim flags.
						const AVP_ClientActor_Vertebrate* pxTargetClient = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentTarget);
						if (pxTargetClient && !pxTargetClient->IsDead() && ((pxTargetClient->GetFlagsForAllAnimsLastFrame() & AVP_ANIMFLAG_MELEE_COUNTER) != 0))
						{
							const AVP_ClientActor_Vertebrate* pxOwnerClient = AVP_ClientActor_Vertebrate::FindEntity(m_xOwner.GetGuid());
							if (pxOwnerClient)
							{
								const Asura_Move* pxCurrentMove = pxOwnerClient->GetMovementSystem() ? pxOwnerClient->GetMovementSystem()->GetCurrentMove() : NULL;
								s_bHackedTestCounterIndicatorOn = pxCurrentMove ? ((pxCurrentMove->GetAnimFlagsForLastFrame() & AVP_ANIMFLAG_ALLOW_NEW_COUNTER) != 0) : true;

								Asura_GUIMenu_Widget_Group* pxItem = static_cast< Asura_GUIMenu_Widget_Group* >( AVP_GUIMenu_System::GetHUDElement( AVP_GUIMenu_System::GROUP_COUNTER ) );
								if ( pxItem )
								{
									Asura_Vector_3 xChestPos;
									pxTargetClient->GetChestPosition( xChestPos );
									Asura_Camera::ProjectWorldCoordsToScreenSpace( xChestPos, true );
									Asura_Vector_2 xScreenPos(xChestPos.x, xChestPos.y);
									for ( u_int uItr = 0; uItr < pxItem->GetContainerCount(); ++uItr )
									{
										Asura_GUIMenu_Element_Image* pxChild = static_cast< Asura_GUIMenu_Element_Image* >( pxItem->GetContainerItem( uItr ) );
										Asura_Bounding_Rectangle xBounds(pxChild->GetScaledWorldBounds());
										xBounds.Displace(xScreenPos - xBounds.GetDimension() * 0.5f - xBounds.GetTopLeft());
										pxChild->SetScaledWorldBounds( xBounds );
									}
								}
							}
						}
					}
				}
			}
		}
	}
}


bool AVP_Predator_MeleeCounter_System::StartMeleeMove(u_int uType)
{
	// Only do adjustments to attack type if we're using the wristblades.
	const AVP_Inventory_System* pxInventory = m_xOwner.GetInventorySystem();
	if (pxInventory)
	{
		const AVP_InventoryItem_Weapon* pxWeapon = static_cast<const AVP_InventoryItem_Weapon*>(pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON));
		if (pxWeapon && (pxWeapon->GetItemType() == AVP_INVENTORYITEMTYPE_WEAPON_MELEE))
		{

			if (((uType == AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK) || (uType == AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE)) && ((m_uCurrentTarget == ASURA_GUID_UNREGISTERED) || GetTargetTooCloseToPounce()))
			{
				// If primary attack has been pressed and the enemy is close (or we have no target), we just want to attack them normally instead.
				if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK)
				{
					uType = AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE;
				}
				else
				{
					uType = AVP_MELEEATTACK_TYPE_NORMAL;
				}
			}
			else if (((uType == AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE) || (uType == AVP_MELEEATTACK_TYPE_NORMAL)) && (m_uCurrentTarget != ASURA_GUID_UNREGISTERED) && !GetTargetTooCloseToPounce() && !IsCurrentTargetInSuckRange() && GetCanUseShortestPounceVariation(m_uCurrentTarget))
			{
				if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE)
				{
					uType = AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK;
				}
				else
				{
					uType = AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE;
				}
			}

			// Awkwardly, if we want to be sure the character is off the ground we have to see what each of the client and the server think, as only one actually bothers to update it from anything other than ON_FLOOR.
			const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
			bool bOffGround = m_xOwner.GetGroundType() == AVP_Actor_PhysicsObjectHandler::OFF_GROUND;
			bOffGround |= pxClient && (pxClient->GetGroundType() == AVP_Actor_PhysicsObjectHandler::OFF_GROUND);
			if (bOffGround)
			{
				// If off the ground, all jumping attacks are disabled.
				if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK)
				{
					uType = AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE;
				}
				else if (uType == AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE)
				{
					uType = AVP_MELEEATTACK_TYPE_NORMAL;
				}
				else if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE)
				{
					return false;
				}
			}

			// First grab - only available when up close to a human.
			if ((uType == AVP_MELEEATTACK_TYPE_NORMAL) && GetHasAttackTarget())
			{
				const AVP_ServerActor_Human* pxHumanTarget = AVP_ServerActor_Human::FindEntity(m_uCurrentTarget);
				const AVP_ServerActor_Alien* pxAlienTarget = AVP_ServerActor_Alien::FindEntity(m_uCurrentTarget);
				if (pxHumanTarget || pxAlienTarget)
				{
					// We need to know which side we are to decide on whether to do a stealth kill.
					const Asura_Matrix_3x3& xVictimOri = pxHumanTarget ? pxHumanTarget->GetOrientation() : pxAlienTarget->GetOrientation();
					const Asura_Matrix_3x3& xOwnerOri = m_xOwner.GetOrientation();
					const Asura_Vector_3 xVictimForward(xVictimOri.e13, xVictimOri.e23, xVictimOri.e33);
					const Asura_Vector_3 xOwnerForward(xOwnerOri.e13, xOwnerOri.e23, xOwnerOri.e33);
					if ((xVictimForward * xOwnerForward) > 0.0f)
					{
						// Behind the enemy.
						uType = AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY;
					}
					else
					{
						// Facing the enemy.
						uType = AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY;
					}
				}
			}

			// for predalien 
			if ( ((uType == AVP_MELEEATTACK_TYPE_NORMAL) ||(uType == AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE)) && GetHasAttackTarget() )
			{
				const AVP_ServerActor_Predalien* pxPredalienTarget = AVP_ServerActor_Predalien::FindEntity(m_uCurrentTarget);
				if (pxPredalienTarget)
				{
					if (pxPredalienTarget->IsKnockedDown())
					{
						if (pxPredalienTarget->GetGrappleCount() == 0)
						{
							uType = AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIEN1;
						}
						else if (pxPredalienTarget->GetGrappleCount() == 1)
						{
							uType = AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIEN2;
						}
						else 
						{
							uType = AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIENKILL;
						}
					}
					else
					{
						uType = AVP_MELEEATTACK_TYPE_NORMAL;
					}
				}

			}
		}
	}

	if(PARENT::StartMeleeMove(uType))
	{
		if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE)
		{
			// Switch the side.
			SwitchQuickAttackSide();
		}

		AVP_InventoryItem_Cloak* pxCloak = static_cast< AVP_InventoryItem_Cloak* >( m_xOwner.GetInventorySystem()->GetItem( uAVP_INVENTORY_ITEMHASH_CLOAK ) );
		if ( pxCloak != NULL && pxCloak->IsUsing( false ) )
		{
			const float fCloakedEnergyCost = GetCloakedEnergyDrain( uType );
			if ( fCloakedEnergyCost < 0.0f )
			{
				pxCloak->RequestEndUsage( AVP_InventoryItem_Cloak::TT_AUTOMATIC );
			}
			else if ( fCloakedEnergyCost > 0.0f )
			{
				AVP_ClientActor_Predator* pxPred = AVP_ClientActor_Predator::FindEntity( m_xOwner.GetGuid() );
				AVP_Actor_DynamicStat* pxDynamicAmmo = pxPred != NULL ? pxPred->GetDynamicEnergy() : NULL ;
				if ( pxDynamicAmmo != NULL )
				{
					pxDynamicAmmo->Remove( fCloakedEnergyCost, true );
				}
			}
		}
		return(true);
	}
	else
	{
		return(false);
	}
}

bool AVP_Predator_MeleeCounter_System::PerformCounter()
{
	if (!PARENT::PerformCounter())
	{
		return false;
	}

	AVP_ServerActor_Vertebrate* pxTarget = AVP_ServerActor_Vertebrate::FindEntity(m_uCurrentTarget);
	if (pxTarget && pxTarget->GetClassification() == AVP_ENTITYCLASS_PREDALIEN && m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_PREDALIEN_KNOCKDOWN) 
	{
		static_cast<AVP_ServerActor_Predalien*>(pxTarget)->KnockDown();
	}
	return true;

}


bool AVP_Predator_MeleeCounter_System::GetCanPounce(const AVP_ServerEntity_Actor* pxTarget) const
{
	switch (pxTarget->GetClassification())
	{
	default:
		break;
	case AVP_ENTITYCLASS_ALIENQUEEN: // these types can not be pounce.
	case AVP_ENTITYCLASS_PREDALIEN:
	case AVP_ENTITYCLASS_ALIENFACEHUGGER:		
		return false;
	}


	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;

	if (xOffset.MagnitudeSquared() <= (m_fMinPounceAttackDistance * m_fMinPounceAttackDistance))
	{
		return false;
	}
	
	const float fDistSquared = xOffset.MagnitudeSquared();
	const float fDrop = xOffset.y; // Intentionally allowing negative as most pounces have a minimum that lets you jump up.

	for (u_int u=0; u<m_uNumPounceAttackVariations; ++u)
	{
		const float fCurrentDistance = m_pxPounceAttacks[u].m_fMaxDistance;

		if ((fDrop >= m_pxPounceAttacks[u].m_fMinDrop) && (fDistSquared <= (fCurrentDistance * fCurrentDistance)))
		{
			// We just need to return that at least one pounce is valid for this to succeed.
			return true;
		}
	}

	return false;

}


void AVP_Predator_MeleeCounter_System::CheckForAndDealMeleeDamage()
{
	if (m_uNumHitEntities >= MAX_HIT_MELEE_ENTITIES) { return; }

	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERHIT_FATAL:
	case AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERHIT_REMOVELIMB:
		// These are guaranteed hits.
		ApplyMeleeDamage(m_uCurrentTarget);
		break;

	default:
		PARENT::CheckForAndDealMeleeDamage();
		break;
	}
}

DEBUGVAR float dbg_fCounterDismemberRagdollImpulse = 10.0f;
DEBUGVAR float dbg_fCounterDismemberArmImpulse = 0.05f;
DEBUGVAR float dbg_fCounterDismemberFaceImpulse = 0.3f;
DEBUGVAR float dbg_fCounterDismemberDirectionRandomValue = 0.5f;

void AVP_Predator_MeleeCounter_System::ApplyMeleeDamage(Asura_Guid uDamagedGuid)
{
	// If we're applying damage, we've hit the enemy. Always decloak in that situation.
	static_cast<AVP_ServerActor_Predator&>(m_xOwner).Decloak(false);

	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERHIT_FATAL:
	case AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERHIT_REMOVELIMB:
		{
			// Remove either the appropriate arm, or the face.
			const AVP_ServerEntity_PhysicalObject* pxTarget = AVP_ServerEntity_PhysicalObject::FindEntity(uDamagedGuid);

			if (pxTarget)
			{
				Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_DISMEMBER, sizeof(AVP_Msg_Dismember));
				if(pxMessage)
				{
					// AJG - 25/06/08 - For SPS, we'll be using the throw counter with quick attacks, and randomising the limb removal for the power attack. Leaving the old code here for now though, as I suspect we'll return to it after SPS.
					//const bool bLimbRemove = m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERHIT_REMOVELIMB;
					const bool bFatal = m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERHIT_FATAL;
					const bool bLimbRemove = Asura_Random::GetIntU(0, 1) == 0;

					AVP_Msg_Dismember* pxData = static_cast<AVP_Msg_Dismember*>(pxMessage->GetData());

					Asura_Vector_3 xPosition;
					if (!pxTarget->GetWorldSpaceMarkerPoint(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_TARGET_1), &xPosition, NULL))
					{
						xPosition = pxTarget->GetPosition();
					}

					Asura_Vector_3 xAimDirection = m_xOwner.GetAimDirection();
					xAimDirection.y = 0.0f;
					// Add a little randomness to the direction so it doesn't look exactly the same every time.
					xAimDirection.x += Asura_Random::GetFloat(-dbg_fCounterDismemberDirectionRandomValue, dbg_fCounterDismemberDirectionRandomValue);
					xAimDirection.y += Asura_Random::GetFloat(-dbg_fCounterDismemberDirectionRandomValue, 0.0f);
					xAimDirection.z += Asura_Random::GetFloat(-dbg_fCounterDismemberDirectionRandomValue, dbg_fCounterDismemberDirectionRandomValue);

					if (xAimDirection.ApproximatelyEquals(Asura_Maths::ZeroVector3))
					{
						const Asura_Matrix_3x3& xOri = m_xOwner.GetOrientation();
						xAimDirection.x = xOri.e13;
						xAimDirection.y = xOri.e23;
						xAimDirection.z = xOri.e33;
					}
					else
					{
						xAimDirection.Normalise();
					}

					int iElement = AVP_SET_ALIEN_FACE;
					if (bLimbRemove)
					{
						// In this case, we need to know if we're knocking off the left arm or the right arm.
						const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_xOwner.GetGuid());
						if (pxClient)
						{
							const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
							if (pxMovementSystem)
							{
								const Asura_Move* pxMove = pxMovementSystem->GetCurrentMove();
								if (pxMove && (pxMove->GetHash() == AVP_Predator_Move_Melee_CounterAlienJumpAttack_Hit_Right::GetStaticHash()))
								{
									iElement = AVP_SET_RIGHT_ARM;
								}
								else
								{
									iElement = AVP_SET_LEFT_ARM;
								}
							}
						}
					}
					pxData->SetElement(iElement);
					pxData->SetDamagePosition(xPosition);
					pxData->SetRagdollDirection(xAimDirection);
					pxData->SetRagdollImpulse(dbg_fCounterDismemberRagdollImpulse);
					pxData->SetGibDirection(xAimDirection);
					pxData->SetGibImpulse(bLimbRemove ? dbg_fCounterDismemberArmImpulse : dbg_fCounterDismemberFaceImpulse);
					pxData->SetDamage(bFatal ? -1.0f : m_fModerateDamageAmount);
					pxData->SetFromGuid(m_xOwner.GetGuid()); 

					pxMessage->SetFromGuid(m_xOwner.GetGuid());
					pxMessage->SetToGuid(m_uCurrentTarget);

					Asura_Network::Send(pxMessage);
				}
			}
		}
		break;

	default:
		PARENT::ApplyMeleeDamage(uDamagedGuid);
		break;
	}
}

bool AVP_Predator_MeleeCounter_System::GetAttackTypeCanCounter(u_int uType) const
{
	switch (uType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYSLASH:
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		return true;

	default:
		return PARENT::GetAttackTypeCanCounter(uType);
	}
}

void AVP_Predator_MeleeCounter_System::ApplyGrappleDamage() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIEN1:
	case AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIEN2:
		break;
	case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
	case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
	case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
		{
			// From the pickup, the trophy kill occurs.
			const AVP_ServerEntity_PhysicalObject* pxTarget = AVP_ServerEntity_PhysicalObject::FindEntity(m_uCurrentTarget);

			if (pxTarget)
			{
				// The alien has this from the stealth pickup if downed, whereas the human only has this from the front.
				const bool bAlien = pxTarget->GetClassification() == AVP_ENTITYCLASS_ALIEN;
				bool bDownedAlien = bAlien;
				bool bStealthKill = (m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY);
				if (bAlien)
				{
					// We need to check if this is a downed alien. Unfortunately, the safest approach here is to check what move we're using.
					const AVP_ClientActor_Predator* pxClient = AVP_ClientActor_Predator::FindEntity(m_xOwner.GetGuid());
					if (pxClient)
					{
						const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
						if (pxMovementSystem)
						{
							const Asura_Move* pxMove = pxMovementSystem->GetCurrentMove();
							bDownedAlien = pxMove && (pxMove->GetHash() == AVP_Predator_Move_Melee_PickupDownedAlien_Grapple_Success::GetStaticHash());
						}
					}
				}
				if (!bDownedAlien && bStealthKill)
				{
					PARENT::ApplyGrappleDamage();
				}
				else
				{
					Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_DISMEMBER, sizeof(AVP_Msg_Dismember));
					if(pxMessage)
					{
						AVP_Msg_Dismember* pxData = static_cast<AVP_Msg_Dismember*>(pxMessage->GetData());

						Asura_Vector_3 xPosition;
						if (!pxTarget->GetWorldSpaceMarkerPoint(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_TARGET_2), &xPosition, NULL))
						{
							xPosition = pxTarget->GetPosition();
						}

						pxData->SetElement(bAlien ? AVP_SET_ALIEN_FACE : AVP_SET_HEAD);
						pxData->SetDamagePosition(xPosition);
						pxData->SetRagdollDirection(Asura_Vector_3(0.0f, 1.0f, 0.0f));
						pxData->SetRagdollImpulse(0.0f);
						pxData->SetGibDirection(bAlien ? Asura_Vector_3(0.0f, 1.0f, 0.0f) : Asura_Maths::ZeroVector3); // The zero vector tells it not to create the head gib (the attachment is done in a different way).
						pxData->SetGibImpulse(0.0f);
						pxData->SetDamage(-1.0f);
						pxData->SetFromGuid(m_xOwner.GetGuid()); 

						pxMessage->SetFromGuid(m_xOwner.GetGuid());
						pxMessage->SetToGuid(m_uCurrentTarget);

						Asura_Network::Send(pxMessage);
					}
				}
			}
		}
		break;
	default:
		PARENT::ApplyGrappleDamage();
		break;
	}
}

Asura_Hash_ID AVP_Predator_MeleeCounter_System::GetMeleeMoveHash(u_int uType) const
{
	// The primary cancel can always trigger, as the primary slash is only available in wristblades.
	if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYCANCEL)
	{
		return (m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYSLASH) ? AVP_Predator_Move_MeleeCharge_End::GetStaticHash() : ASURA_HASH_ID_UNSET;
	}

	const AVP_Inventory_System* pxInventory = m_xOwner.GetInventorySystem();
	if (!pxInventory) { return ASURA_HASH_ID_UNSET; }

	const AVP_InventoryItem_Weapon* pxWeapon = static_cast<const AVP_InventoryItem_Weapon*>(pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON));
	if (!pxWeapon) { return ASURA_HASH_ID_UNSET; }

	if (pxWeapon->GetItemType() == AVP_INVENTORYITEMTYPE_WEAPON_MELEE)
	{
		// If this is a melee weapon type, it must be the wristblades. That gives us loads of extra options.
		switch (uType)
		{
		case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYSLASH:
			return AVP_Predator_Move_MeleeCharge_Start::GetStaticHash();

		case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
			// This has a special power-up attack if we're in the middle of the PRIMARYSLASH state.
			if (m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYSLASH)
			{
				// We need to find out if we've finished charging up. That's defined by the current move being the charge loop.
				const AVP_ClientActor_Predator* pxClient = AVP_ClientActor_Predator::FindEntity(m_xOwner.GetGuid());
				if (!pxClient) { return ASURA_HASH_ID_UNSET; }
				if (!pxClient->GetMovementSystem()) { return ASURA_HASH_ID_UNSET; }
				if (!pxClient->GetMovementSystem()->GetCurrentMove()) { return ASURA_HASH_ID_UNSET; }

				const Asura_Hash_ID uMoveHash = pxClient->GetMovementSystem()->GetCurrentMove()->GetHash();
				if (uMoveHash == AVP_Predator_Move_MeleeCharge_Loop::GetStaticHash())
				{
					return AVP_Predator_Move_Melee_Wristblade_PowerAttack_Start::GetStaticHash();
				}
				else
				{
					return AVP_Predator_Move_Melee_Wristblade_QuickAttack_Start::GetStaticHash();
				}
			}
			// Otherwise, it'll always be the quick attack.
			else
			{
				return AVP_Predator_Move_Melee_Wristblade_QuickAttack_Start::GetStaticHash();
			}

		case AVP_MELEEATTACK_TYPE_NORMAL:
			return AVP_Predator_Move_Melee_Wristblade_SlowAttack::GetStaticHash();

		case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK:
		case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE:
		case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
			return GetCanPounce() ? AVP_Move_Melee_Pounce_Attack_Start::GetStaticHash() : ASURA_HASH_ID_UNSET;

		case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYFINISHER:
			return AVP_Predator_Move_Melee_QuickFinisher_Start::GetStaticHash();

		case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
		case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
			{
				const AVP_ServerActor_Human* pxHumanTarget = AVP_ServerActor_Human::FindEntity(m_uCurrentTarget);
				if (pxHumanTarget)
				{
					if (uType == AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY)
					{
						return AVP_Predator_Move_Melee_StealthPickupHuman_Start::GetStaticHash();
					}
					else
					{
						return AVP_Predator_Move_Melee_PickupHuman_Start::GetStaticHash();
					}
				}
				else
				{
					// The alien is more complicated - we can only trigger that from the specific counter (or the following animflag-marked animation) at the moment
					const AVP_ServerActor_Alien* pxAlienTarget = AVP_ServerActor_Alien::FindEntity(m_uCurrentTarget);
					if (pxAlienTarget)
					{
						if (pxAlienTarget->IsDowned() || (m_uCurrentAttackType == AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERTHROW))
						{
							return AVP_Predator_Move_Melee_PickupDownedAlien_Start::GetStaticHash();
						}
						else
						{
							const bool bStealth = uType == AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY;
							if (pxAlienTarget->IsStanding())
							{
								return bStealth ? AVP_Predator_Move_Melee_StealthPickupAlienStanding_Start::GetStaticHash() : AVP_Predator_Move_Melee_PickupAlienStanding_Start::GetStaticHash();
							}
							else
							{
								return bStealth ? AVP_Predator_Move_Melee_StealthPickupAlienCrouching_Start::GetStaticHash() : AVP_Predator_Move_Melee_PickupAlienCrouching_Start::GetStaticHash();
							}
						}
					}

					return AVP_Predator_Move_Melee_Wristblade_SlowAttack::GetStaticHash();
				}
			}

		case AVP_PREDATOR_MELEEATTACK_TYPE_POWERATTACK:
			return AVP_Predator_Move_Melee_Wristblade_PowerAttack_Start::GetStaticHash();

		case AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIEN1:
			return AVP_Predator_Move_Melee_PredalienPounceAttack_Start1::GetStaticHash();

		case AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIEN2:
			return AVP_Predator_Move_Melee_PredalienPounceAttack_Start2::GetStaticHash();

		case AVP_PREDATOR_MELEEATTACK_TYPE_GRAPPLE_PREDALIENKILL:
			return AVP_Predator_Move_Melee_PredalienPounceAttack_StartKill::GetStaticHash();

		default:
			return ASURA_HASH_ID_UNSET;
		}
	}

	// All other weapons only have standard melee.
	return PARENT::GetMeleeMoveHash(uType);
}


u_int AVP_Predator_MeleeCounter_System::PickMeleeMoveVariation(Asura_Hash_ID uMoveHash) const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		// If we're using the quick attack, we need to pick between the left and right variations.
		if (uMoveHash == AVP_Predator_Move_Melee_Wristblade_QuickAttack_Start::GetStaticHash())
		{
			if (m_eNextQuickAttackSide == QUICK_ATTACK_LEFT)
			{
				return Asura_Random::GetIntU(0, m_uNumLeftQuickAttackVariations - 1);
			}
			else
			{
				Asura_Assert(m_eNextQuickAttackSide == QUICK_ATTACK_RIGHT, "Invalid quick attack side.");
				return Asura_Random::GetIntU(m_uNumLeftQuickAttackVariations, m_uNumLeftQuickAttackVariations + m_uNumRightQuickAttackVariations - 1);
			}
		}
		// Otherwise, just use the default.
		return PARENT::PickMeleeMoveVariation(uMoveHash);

	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK:
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE:
	case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
		{
			return GetPounceVariation();
		}

	default:
		return PARENT::PickMeleeMoveVariation(uMoveHash);
	}
}

DEBUGVAR float dbg_fPounceMaxCounterDistance = 10.0f;

float AVP_Predator_MeleeCounter_System::GetCurrentMaxCounterDistance() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK:
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE:
		return dbg_fPounceMaxCounterDistance;

	default:
		return PARENT::GetCurrentMaxCounterDistance();
	}
}

float AVP_Predator_MeleeCounter_System::GetCloakedEnergyDrain( u_int uType ) const
{
	return 0.0f;
}

float AVP_Predator_MeleeCounter_System::GetAttackDamage() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK:
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE:
		return m_pxPounceAttacks[m_uCurrentVariation].m_fDamage;

	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYFINISHER:
		return m_fMajorDamageAmount;

	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		// This depends on what move we've ended up with.
		{
			const AVP_ClientActor_Predator* pxClient = AVP_ClientActor_Predator::FindEntity(m_xOwner.GetGuid());
			if (!pxClient) { return m_fModerateDamageAmount; }
			if (!pxClient->GetMovementSystem()) { return m_fModerateDamageAmount; }
			if (!pxClient->GetMovementSystem()->GetCurrentMove()) { return m_fModerateDamageAmount; }

			const Asura_Hash_ID uMoveHash = pxClient->GetMovementSystem()->GetCurrentMove()->GetHash();
			if (uMoveHash == AVP_Predator_Move_Melee_Wristblade_PowerAttack_Start::GetStaticHash())
			{
				return m_fMajorDamageAmount;
			}
			else
			{
				return m_fModerateDamageAmount;
			}
		}

	case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
	case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
	case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
		return 0.0f;

	case AVP_PREDATOR_MELEEATTACK_TYPE_POWERATTACK:
		return m_fMajorDamageAmount;

	default:
		return PARENT::GetAttackDamage();
	}
}

Asura_Hash_ID AVP_Predator_MeleeCounter_System::GetAttackDamageType() const
{
	const AVP_Inventory_System* pxInventory = m_xOwner.GetInventorySystem();
	if (!pxInventory) { return ASURA_HASH_ID_UNSET; }

	const AVP_InventoryItem_Weapon* pxWeapon = static_cast<const AVP_InventoryItem_Weapon*>(pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON));
	if (!pxWeapon) { return ASURA_HASH_ID_UNSET; }

	if (pxWeapon->GetItemType() == AVP_INVENTORYITEMTYPE_WEAPON_MELEE)
	{
		// The melee weapon is the wristblades.
		return uAVP_BLUEPRINT_NAME_DAMAGE_BLADE;
	}
	else
	{
		return uAVP_BLUEPRINT_NAME_DAMAGE_MELEE;
	}
}

float AVP_Predator_MeleeCounter_System::GetAttackImpulseStrength() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYFINISHER:
		return dbg_fPredatorFinisherImpulseStrength;

	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK:
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE:
		return m_pxPounceAttacks[m_uCurrentVariation].m_fAttackImpulse;

	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		// This depends on what move we've ended up with.
		{
			const AVP_ClientActor_Predator* pxClient = AVP_ClientActor_Predator::FindEntity(m_xOwner.GetGuid());
			if (!pxClient) { return PARENT::GetAttackImpulseStrength(); }
			if (!pxClient->GetMovementSystem()) { return PARENT::GetAttackImpulseStrength(); }
			if (!pxClient->GetMovementSystem()->GetCurrentMove()) { return PARENT::GetAttackImpulseStrength(); }

			const Asura_Hash_ID uMoveHash = pxClient->GetMovementSystem()->GetCurrentMove()->GetHash();
			if (uMoveHash == AVP_Predator_Move_Melee_Wristblade_PowerAttack_Start::GetStaticHash())
			{
				return dbg_fPredatorFinisherImpulseStrength;
			}
			else
			{
				return PARENT::GetAttackImpulseStrength();
			}
		}

	case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
	case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
	case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
		return 0.0f;

	case AVP_PREDATOR_MELEEATTACK_TYPE_POWERATTACK:
		return dbg_fPredatorFinisherImpulseStrength;

	default:
		return PARENT::GetAttackImpulseStrength();
	}
}

static const Asura_Matrix_3x3 xSMALL_ROTATION_LEFT(0.866f, 0.0f, 0.500f, 0.0f, 1.0f, 0.0f, -0.500f, 0.0f, 0.866f);
static const Asura_Matrix_3x3 xSMALL_ROTATION_RIGHT(0.866f, 0.0f, -0.500f, 0.0f, 1.0f, 0.0f, 0.500f, 0.0f, 0.866f);

void AVP_Predator_MeleeCounter_System::GetAttackDirection(const AVP_ServerEntity_PhysicalObject* pxTarget, Asura_Vector_3& xDirection) const
{
	PARENT::GetAttackDirection(pxTarget, xDirection);

	// With left and right attacks, fudge the direction a bit.
	switch (m_uCurrentAttackType)
	{
	default:
		break;

	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		// This depends on what move we've ended up with. Only the quick attack applies a direction modifier.
		{
			const AVP_ClientActor_Predator* pxClient = AVP_ClientActor_Predator::FindEntity(m_xOwner.GetGuid());
			if (!pxClient) { break; }
			if (!pxClient->GetMovementSystem()) { break; }
			if (!pxClient->GetMovementSystem()->GetCurrentMove()) { break; }

			const Asura_Hash_ID uMoveHash = pxClient->GetMovementSystem()->GetCurrentMove()->GetHash();
			if (uMoveHash == AVP_Predator_Move_Melee_Wristblade_QuickAttack_Start::GetStaticHash())
			{
				if (m_uCurrentVariation < m_uNumLeftQuickAttackVariations)
				{
					xDirection = xDirection * xSMALL_ROTATION_LEFT;
				}
				else
				{
					xDirection = xDirection * xSMALL_ROTATION_RIGHT;
				}
			}
		}
		break;
	}
}

bool AVP_Predator_MeleeCounter_System::GetAttackIsRightHanded() const
{
	switch (m_uCurrentAttackType)
	{
		default: { break; }
		
		// This depends on what move we've ended up with. Only the quick attack applies a direction modifier.
		case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		{
			const AVP_ClientActor_Predator* pxClient = AVP_ClientActor_Predator::FindEntity(m_xOwner.GetGuid());
			if (!pxClient) { break; }
			if (!pxClient->GetMovementSystem()) { break; }
			if (!pxClient->GetMovementSystem()->GetCurrentMove()) { break; }

			const Asura_Hash_ID uMoveHash = pxClient->GetMovementSystem()->GetCurrentMove()->GetHash();
			if (uMoveHash == AVP_Predator_Move_Melee_Wristblade_QuickAttack_Start::GetStaticHash())
			{
				if (m_uCurrentVariation < m_uNumLeftQuickAttackVariations)
				{
					return false;
				}
			}
		}
		break;
	}

	return true;
}

int AVP_Predator_MeleeCounter_System::GetAttackBoneHit(Asura_Guid uDamagedGuid) const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		{
			// Pick a random skin element to hit, choosing a bone from that.
			// Some elements are alien-specific non-fatal-to-lose ones. If they're picked for non-aliens we'll get a bone ID of -1, which should be fine.
			const AVP_ServerActor_Vertebrate* pxTarget = AVP_ServerActor_Vertebrate::FindEntity(uDamagedGuid);
			if (!pxTarget) { return -1; }

			// Because of the amount of damage caused, this simple approach that always hits a detachable limb will still only occasionally cause limb removal.
			const int iRandom = Asura_Random::GetInt(0, 1);
			const u_int uSkinElement = (iRandom == 0) ? AVP_SET_LEFT_ARM : AVP_SET_RIGHT_ARM;

			return pxTarget->GetAnyBoneIDFromElement(uSkinElement);
		}

	default:
		return PARENT::GetAttackBoneHit(uDamagedGuid);
	}
}

float AVP_Predator_MeleeCounter_System::GetGrappleDamage() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_COUNTERGRAPPLE_MINOR:
		return m_fMinorDamageAmount;

	default:
		return PARENT::GetGrappleDamage();
	}
}

bool AVP_Predator_MeleeCounter_System::IsAttackUnblockable() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSATTACK:
	case AVP_PREDATOR_MELEEATTACK_TYPE_FOCUSPOUNCE:
	case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYFINISHER:
	case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
	case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
		return true;

	default:
		return PARENT::IsAttackUnblockable();
	}
}

bool AVP_Predator_MeleeCounter_System::GetAttackOnlyHitsCurrentTarget() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYSLASH:
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
		return false;

	default:
		return PARENT::IsAttackUnblockable();
	}
}

Asura_Hash_ID AVP_Predator_MeleeCounter_System::GetAttackGrappleSetBlueprintParamID() const
{
	switch (m_uCurrentAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
		return uAVP_BLUEPRINT_PARAM_GRAPPLESET_PREDATOR_FRONT;

	case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
		return uAVP_BLUEPRINT_PARAM_GRAPPLESET_PREDATOR_BACK;

	case AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE:
		// See if we're hitting the front or the back.
		{
			const AVP_ServerActor_Human* pxHumanTarget = AVP_ServerActor_Human::FindEntity(m_uCurrentTarget);
			const AVP_ServerActor_Alien* pxAlienTarget = AVP_ServerActor_Alien::FindEntity(m_uCurrentTarget);
			if (pxHumanTarget || pxAlienTarget)
			{
				// We need to know which side we are to decide on whether to do a stealth kill.
				const Asura_Matrix_3x3& xVictimOri = pxHumanTarget ? pxHumanTarget->GetOrientation() : pxAlienTarget->GetOrientation();
				const Asura_Matrix_3x3& xOwnerOri = m_xOwner.GetOrientation();
				const Asura_Vector_3 xVictimForward(xVictimOri.e13, xVictimOri.e23, xVictimOri.e33);
				const Asura_Vector_3 xOwnerForward(xOwnerOri.e13, xOwnerOri.e23, xOwnerOri.e33);
				if ((xVictimForward * xOwnerForward) > 0.0f)
				{
					// Behind the enemy.
					return uAVP_BLUEPRINT_PARAM_GRAPPLESET_PREDATOR_BACK;
				}
				else
				{
					// Facing the enemy.
					return uAVP_BLUEPRINT_PARAM_GRAPPLESET_PREDATOR_FRONT;
				}
			}
		}

	default:
		return PARENT::GetAttackGrappleSetBlueprintParamID();
	}
}

bool AVP_Predator_MeleeCounter_System::GetSendIncomingMeleeMessage(u_int uAttackType) const
{
	switch (uAttackType)
	{
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYSLASH:
	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYCANCEL:
		return false;

	case AVP_PREDATOR_MELEEATTACK_TYPE_PRIMARYRELEASE:
	case AVP_PREDATOR_MELEEATTACK_TYPE_PICKUPENEMY:
	case AVP_PREDATOR_MELEEATTACK_TYPE_STEALTHPICKUPENEMY:
		return IsCurrentTargetInSuckRange();

	default:
		return PARENT::GetSendIncomingMeleeMessage(uAttackType);
	}
}

bool AVP_Predator_MeleeCounter_System::IsCurrentTargetInSuckRange() const
{
	if (!m_xOwner.IsPlayer()) // prevent suck for AI
	{
		const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
		if (!pxTarget) { return false; }

		const Asura_Vector_3& xTargetPos = pxTarget->GetPosition();
		const Asura_Vector_3& xOwnerPos = m_xOwner.GetPosition();
		Asura_Vector_3 xFromTarget = xOwnerPos - xTargetPos;
		const float fMagSquared = xFromTarget.MagnitudeSquared();

		// since AI doesnt use suck anyhow and it just used to flag when a pounce should be used 
		// ignore minimal distance

		return (fMagSquared <= (m_fDamageRange * m_fDamageRange) );
	}
	else
	{
		return PARENT::IsCurrentTargetInSuckRange();
	}

}


bool AVP_Predator_MeleeCounter_System::GetCanPounce() const
{
	if (GetTargetTooCloseToPounce()) { return false; }

	const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return false; }

	switch (pxTarget->GetClassification())
	{
	default:
		break;
	case AVP_ENTITYCLASS_ALIENQUEEN: // these types can not be pounce.
	case AVP_ENTITYCLASS_PREDALIEN:
	case AVP_ENTITYCLASS_ALIENFACEHUGGER:		
		return false;
	}

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;
	const float fDistSquared = xOffset.MagnitudeSquared();
	const float fDrop = xOffset.y; // Intentionally allowing negative as most pounces have a minimum that lets you jump up.

	for (u_int u=0; u<m_uNumPounceAttackVariations; ++u)
	{
		const float fCurrentDistance = m_pxPounceAttacks[u].m_fMaxDistance;

		if ((fDrop >= m_pxPounceAttacks[u].m_fMinDrop) && (fDistSquared <= (fCurrentDistance * fCurrentDistance)))
		{
			// We just need to return that at least one pounce is valid for this to succeed.
			return true;
		}
	}

	return false;
}

bool AVP_Predator_MeleeCounter_System::GetTargetTooCloseToPounce() const
{
	const AVP_ServerEntity_Actor* pxTarget = AVP_ServerEntity_Actor::FindEntity(m_uCurrentTarget);
	if (!pxTarget) { return false; }

	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	pxTarget->GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;

	return (xOffset.MagnitudeSquared() <= (m_fMinPounceAttackDistance * m_fMinPounceAttackDistance));
}

const Asura_Blueprint* AVP_Predator_MeleeCounter_System::GetExtraMeleeDataBlueprint() const
{
	const AVP_Inventory_System* pxInventory = m_xOwner.GetInventorySystem();
	if (pxInventory)
	{
		const AVP_InventoryGroup_Weapons* pxWeaponsGroup = static_cast<const AVP_InventoryGroup_Weapons*>(pxInventory->GetGroup(uAVP_INVENTORY_GROUPHASH_WEAPON));
		if (pxWeaponsGroup)
		{
			const AVP_Inventory_Item* pxMeleeItem = pxWeaponsGroup->GetItemOfType(AVP_INVENTORYITEMTYPE_WEAPON_MELEE);
			if (pxMeleeItem)
			{
				return Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, pxMeleeItem->GetBlueprintHash());
			}
		}
	}

	return NULL;
}

bool AVP_Predator_MeleeCounter_System::GetQuickAttackBaseParameterNames(const char*& szLeftAttackName, const char*& szRightAttackName) const
{
	szLeftAttackName = s_szLeftQuickAttackBaseParameterName;
	szRightAttackName = s_szRightQuickAttackBaseParameterName;
	return true;
}

bool AVP_Predator_MeleeCounter_System::InsidePounceRange() const
{
	if( m_uCurrentTarget == ASURA_GUID_UNREGISTERED ){ return false; }

	if( GetCanUseShortestPounceVariation( m_uCurrentTarget ) )
	{
		return true;
	}
	return false;
}

Asura_Hash_ID AVP_Predator_MeleeCounter_System::GetPounceBlueprintType() const
{
	return uAVP_BLUEPRINT_TYPE_MOVESPREDATOR;
}

/// End of file. ///
