//
// AVP_CameraShake_System.h
//
// Created by RDM - 02 November 2006 - Ported by CJW - 02/09/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __AVP_CAMERASHAKE_SYSTEM_H__
#define __AVP_CAMERASHAKE_SYSTEM_H__

/************/
/* Includes */
/************/

// asura
#include "Core/Asura_Maths.h"
#include "Network/Asura_Guid.h"

// AVP

/*************/
/* Defines   */
/*************/

#ifndef ASURA_NO_DEBUG_OUTPUT
#define AVP_SHAKE_DEBUG
#endif

/*************/
/* Constants */
/*************/

//System constants
static const u_int uAVP_MAX_CAMERA_SHAKES = 8;

/***********/
/* Enums   */
/***********/

enum AVP_CAMERASHAKE_CURVE
{
	AVP_CAMERASHAKE_CURVE_NONE = 0,
	AVP_CAMERASHAKE_CURVE_CONST_INC,
	AVP_CAMERASHAKE_CURVE_CONST_DEC,
	AVP_CAMERASHAKE_CURVE_INC_INC,
	AVP_CAMERASHAKE_CURVE_INC_DEC,
	AVP_CAMERASHAKE_CURVE_DEC_INC,
	AVP_CAMERASHAKE_CURVE_DEC_DEC,
	AVP_CAMERASHAKE_CURVE_PULSE,
	/* New curves go here*/
	AVP_CAMERASHAKE_CURVE_MAX,
};

enum AVP_CAMERASHAKE_FLAG
{
	AVP_CAMERASHAKE_FLAG_SOURCE_ONLY			= 0x1,
	AVP_CAMERASHAKE_FLAG_SOURCE_IGNORE_DISTANCE	= 0x2,
};


/***********/
/* Classes */
/***********/

class AVP_CameraShake_Parameters
{
public:
	AVP_CameraShake_Parameters():m_uCurve(0),m_fDuration(0.0f),m_fMaxDist(1.0f),m_uFlags(0) {}
	//Shake
	Asura_Vector_3 m_xAmplitude;
	Asura_Vector_3 m_xFrequency;
	u_int m_uCurve;

	//Common
	float m_fDuration;
	float m_fMaxDist;
	u_int m_uFlags;
};

/*************************************************************
 *
 * Class Name    : AVP_CameraShake
 *
 * Author        : Andrew McDonald
 *
 * Description   : Stores data for a camera shake
 *
 *************************************************************/
class AVP_CameraShake
{
public:

	AVP_CameraShake() : m_bActive(false) {}

	void Start( AVP_CameraShake_Parameters& xParameters, Asura_Guid uSourceGuid );
	void Start( AVP_CameraShake_Parameters& xParameters, const Asura_Vector_3& xSourcePosition );
	void Start( AVP_CameraShake_Parameters& xParameters );
	void Deactivate( void )						{ m_bActive = false; }

	void Update( bool bIgnoreTime = false );
	void Apply( Asura_Matrix_3x3& xOrientation, float fDistCurve );
	bool CalculateDistanceCurve( Asura_Vector_3& xCamPos, float& fDistanceCurve );

	void SetScale( float fScale )				{ m_fScale = fScale; }

	const Asura_Vector_3& GetPosition(void)		{ return m_xPosition; }
	Asura_Guid GetSourceGuid(void)				{ return m_uSourceGuid; }

	bool IsActive() const						{ return m_bActive; }
	bool IsActiveForCamera( int iCamera ) const;
	bool IsIgnoringDistanceForCamera( int iCamera) const;

protected:
	void SetCommonParameters( AVP_CameraShake_Parameters& xParameters );

#ifdef AVP_SHAKE_DEBUG
	void DebugOutputShake( float fAmplitude );
#endif

	float GetCurve( u_int uCurve );

	bool m_bActive;
	bool m_bPositionless;

	u_int m_uFlags;

	Asura_Vector_3 m_xAmplitude;
	Asura_Vector_3 m_xInitialAmplitude;
	Asura_Vector_3 m_xFrequency;
	float m_fTimer;
	float m_fDuration;
	float m_fMaxDistSqrd;
	float m_fScale;
	
	u_int m_uCurve;

	Asura_Vector_3 m_xRandom;

	Asura_Guid m_uSourceGuid;
	int m_iSourceCamera;
	Asura_Vector_3 m_xPosition;

	static const float s_fNLog;
	static const float s_fInvNLog;

	// Constant Gradient
	static inline float Curve_Const_Inc( float fValue ) { return fValue; }
	static inline float Curve_Const_Dec( float fValue ) { return 1.0f - fValue; }
	// Increasing Gradient
	static inline float Curve_Inc_Inc( float fValue ) { return Asura_Maths::Power(fValue, s_fNLog); }
	static inline float Curve_Inc_Dec( float fValue ) { return 1.0f - Asura_Maths::Power(fValue, s_fNLog); }
	// Decreasing Gradient
	static inline float Curve_Dec_Inc( float fValue ) { return Asura_Maths::Power(fValue, s_fInvNLog); }
	static inline float Curve_Dec_Dec( float fValue ) { return 1.0f - Asura_Maths::Power(fValue, s_fInvNLog); }
	// Other
	static inline float Curve_Pulse (float fValue) { return Asura_Maths::Sine(fValue*Asura_Maths::Pi); }

	//An array of pointers to curve functions (in case you were wondering)
	static float (*s_apfnCurveFns[AVP_CAMERASHAKE_CURVE_MAX])(float);
};
/* end of class AVP_CameraShake. */


/*************************************************************
 *
 * Class Name    : AVP_CameraShake_System
 *
 * Author        : Richard May
 *
 * Description   : Stores, updates and controls the camera shakes
 *
 *************************************************************/
class AVP_CameraShake_System
{
public:
	/* Public functions. */
	static void Initialise( void );
	static void Shutdown( void );
	static void Reset( void );

	static void Update( void );
	static void ApplyShakes( void );
	
	static void StartShake( Asura_Hash_ID uBlueprint, Asura_Guid uSourceGuid = ASURA_GUID_UNREGISTERED );
	static void StartShake( AVP_CameraShake_Parameters& xParameters, Asura_Guid uSourceGuid );
	static void StartShake( AVP_CameraShake_Parameters& xParameters, const Asura_Vector_3& xSourcePosition );
	static void StartShake( AVP_CameraShake_Parameters& xParameters );	
	static void ScaleShake( Asura_Guid uSourceGuid, float fScale );
	static void StopShake( Asura_Guid uSourceGuid );
	static void StopAllShakes( void );

	static bool IsSourceLocalPlayer( Asura_Guid uSourceGuid );

protected:
	/* Helper functions. */
#ifdef AVP_SHAKE_DEBUG
	static void InitialiseDebug();
#endif
	static bool ShouldCreateShake( float fDuration, Asura_Guid uSourceGuid, const Asura_Vector_3* const pxPosition, float fMaxDist, u_int uFlags, u_int* puSlot=NULL );
	
	/* Data */
	static AVP_CameraShake s_axCameraShakes[uAVP_MAX_CAMERA_SHAKES];
};
/* end of class AVP_CameraShake_System. */



/*************************************************************
 *
 * Class Name    : AVP_Msg_StartCameraShake
 *
 * Author        : Paul Cardy
 *
 * Description   : Message sent to a client to start a blueprinted
 *				   camera shake
 *
 *************************************************************/

struct AVP_Msg_CameraShake
{
	enum { CURRENT_VERSION = 0 };
	u_int m_uVersion;
	Asura_Hash_ID m_uShake;
	Asura_Guid m_uSourceGuid;
};

/***************/
/* End of File */
/***************/

#endif//__AVP_CAMERASHAKE_SYSTEM_H__
