
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Chunk_SurfaceOverlay.cpp
// Author : Chris Wallace
// Date   : 2-Aug-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "ChunkSystem/AVP_Chunk_SurfaceOverlay.h"

// Asura independent header-files...
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_String.h"

#include "Core/AVP_SurfaceEffects.h"


#if defined(ASURA_TOOLS)
// Tools header-files...
#include "Material/ToolsAVP_SurfaceOverlay.h"
#endif	// defined(ASURA_TOOLS)


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Version	Author	Comment
// ~~~~~~~	~~~~~~	~~~~~~~
// 4		EP		Added EffectTypes
////////////////////////////////////////////////////////////////////////////////
const int	AVP_Chunk_SurfaceOverlay::s_iChunkVersion	= 4;



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Chunk_SurfaceOverlay::Process
// Author       : Chris Wallace
// Date         : 2-Aug-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Chunk_SurfaceOverlay::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	if(Version < 3)
	{
		Asura_Assert(false, "AVP_Chunk_SurfaceOverlay: chunk version < 3 not supported - aborting chunk-read.");
		return;
	}

	if(Version > s_iChunkVersion)
	{
		Asura_Assert(false, "AVP_Chunk_SurfaceOverlay: Unexpected chunk version found - aborting chunk-read.");
		return;
	}
	

	// Current stuff
	int	iNumEffectMaps;

	xStream >> iNumEffectMaps;
	
	for(int i = 0; i < iNumEffectMaps; i++)
	{
		Asura_Hash_ID uSurfaceHash;
		Asura_Hash_ID uDamageHash;

		xStream >> uSurfaceHash;
		xStream >> uDamageHash;

		AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetDamageMap(uSurfaceHash, uDamageHash, true);
		if(pxEffectMap)
		{
			Asura_Hash_ID uSelectionHash;
			xStream >> uSelectionHash;
			pxEffectMap->SetHash(uSelectionHash);
		}
	}	

	int	iNumCollisionMaps;

	xStream >> iNumCollisionMaps;

	for(int i = 0; i < iNumCollisionMaps; i++)
	{
		Asura_Hash_ID uCollider;
		Asura_Hash_ID uCollidee;

		xStream >> uCollider;
		xStream >> uCollidee;

		AVP_SurfaceEffect_Map* pxCollisionMap = AVP_SurfaceEffect_System::GetCollisionMap(uCollider, uCollidee, true);
		if(pxCollisionMap)
		{
			Asura_Hash_ID uSelectionHash;
			xStream >> uSelectionHash;
			pxCollisionMap->SetHash(uSelectionHash);
		}
	}

	if( Version>=4 )
	{
		// Effect Types
		int	iNumEffectTypeMaps;

		xStream >> iNumEffectTypeMaps;

		for(int i = 0; i < iNumEffectTypeMaps; i++)
		{
			Asura_Hash_ID uSurfaceHash;
			Asura_Hash_ID uEffectTypeHash;

			xStream >> uSurfaceHash;
			xStream >> uEffectTypeHash;

			AVP_SurfaceEffect_Map* pxEffectTypeMap = AVP_SurfaceEffect_System::GetEffectTypeMap(uSurfaceHash, uEffectTypeHash, true);
			if(pxEffectTypeMap)
			{
				Asura_Hash_ID uSelectionHash;
				xStream >> uSelectionHash;
				pxEffectTypeMap->SetHash(uSelectionHash);
			}
		}
	}
}

/// End of function 'AVP_Chunk_SurfaceOverlay::Process'. ///

#if defined (ASURA_TOOLS)

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Chunk_SurfaceOverlay::Create
// Author       : Chris Wallace
// Date         : 2-Aug-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Chunk_SurfaceOverlay::Create(Asura_Chunk_Stream& xStream)
{
	// Begin writing...
	xStream.WriteChunk_Start(AVP_CHUNKID_SURFACEEFFECTS, s_iChunkVersion, 0);
	{
		////////////////////////////
		// Damage Maps
		////////////////////////////

		int iNumEffectMaps = 0;

		u_int uNumSurfaceTypes = ToolsAVP_SurfaceOverlay::GetNumSurfaceTypes();
		u_int uNumDamageTypes = ToolsAVP_SurfaceOverlay::GetNumDamageTypes();

		for(u_int u = 0; u < uNumSurfaceTypes; u++)
		{
			Asura_Hash_ID uSurfaceHash	= ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( u );

			for(u_int v = 0; v < uNumDamageTypes; v++)
			{
				Asura_Hash_ID uDamageHash	= ToolsAVP_SurfaceOverlay::GetDamageTypeHashFromIndex( v );

				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetDamageMap( uSurfaceHash, uDamageHash, false );
				if ( pxEffectMap && pxEffectMap->GetHash() != ASURA_HASH_ID_UNSET )
				{
					iNumEffectMaps++;
				}
			}	
		}	

		xStream << iNumEffectMaps;

		for(u_int u = 0; u < uNumSurfaceTypes; u++)
		{
			Asura_Hash_ID uSurfaceHash	= ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( u );

			for(u_int v = 0; v < uNumDamageTypes; v++)
			{
				Asura_Hash_ID uDamageHash	= ToolsAVP_SurfaceOverlay::GetDamageTypeHashFromIndex( v );

				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetDamageMap(uSurfaceHash, uDamageHash, false);
				if(pxEffectMap && pxEffectMap->GetHash() != ASURA_HASH_ID_UNSET)
				{
					xStream << uSurfaceHash;
					xStream << uDamageHash;
					xStream << pxEffectMap->GetHash();
				}
			}	
		}
	}

		////////////////////////////
		// Collision Maps
		////////////////////////////
	{
		u_int uNumSurfaceTypes = ToolsAVP_SurfaceOverlay::GetNumSurfaceTypes();

		int iNumEffectMaps = 0;

		for(u_int u = 0; u < uNumSurfaceTypes; u++)
		{
			Asura_Hash_ID uCollider	= ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( u );

			for(u_int v = 0; v < uNumSurfaceTypes; v++)
			{
				Asura_Hash_ID uCollidee	= ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( v );

				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetCollisionMap(uCollider, uCollidee, false);
				if(pxEffectMap && pxEffectMap->GetHash() != ASURA_HASH_ID_UNSET)
				{
					iNumEffectMaps++;
				}
			}	
		}	

		xStream << iNumEffectMaps;

		for(u_int u = 0; u < uNumSurfaceTypes; u++)
		{
			Asura_Hash_ID uCollider	= ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( u );

			for(u_int v = 0; v < uNumSurfaceTypes; v++)
			{
				Asura_Hash_ID uCollidee	= ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( v );

				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetCollisionMap(uCollider, uCollidee, false);
				if(pxEffectMap && pxEffectMap->GetHash() != ASURA_HASH_ID_UNSET)
				{
					xStream << uCollider;
					xStream << uCollidee;
					xStream << pxEffectMap->GetHash();
				}
			}	
		}	
	}

		////////////////////////////
		// EffectType Maps
		////////////////////////////
	{
		u_int uNumSurfaceTypes = ToolsAVP_SurfaceOverlay::GetNumSurfaceTypes();
		u_int uNumEffectTypes = ToolsAVP_SurfaceOverlay::GetNumEffectTypes();
		int iNumEffectMaps = 0;

		for(u_int u = 0; u < uNumSurfaceTypes; u++)
		{
			Asura_Hash_ID uSurfaceHash = ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( u );

			for(u_int v = 0; v < uNumEffectTypes; v++)
			{
				Asura_Hash_ID uEffectTypeHash = ToolsAVP_SurfaceOverlay::GetEffectTypeHashFromIndex( v );

				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetEffectTypeMap(uSurfaceHash, uEffectTypeHash, false);
//				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetCollisionMap(uEffectTypeHash, uSurfaceHash, false);
				if(pxEffectMap && pxEffectMap->GetHash() != ASURA_HASH_ID_UNSET)
				{
					iNumEffectMaps++;
				}
			}	
		}	

		xStream << iNumEffectMaps;

		for(u_int u = 0; u < uNumSurfaceTypes; u++)
		{
			Asura_Hash_ID uSurfaceHash = ToolsAVP_SurfaceOverlay::GetSurfaceTypeHashFromIndex( u );

			for(u_int v = 0; v < uNumEffectTypes; v++)
			{
				Asura_Hash_ID uEffectTypeHash = ToolsAVP_SurfaceOverlay::GetEffectTypeHashFromIndex( v );

				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetEffectTypeMap(uSurfaceHash, uEffectTypeHash, false);
//				AVP_SurfaceEffect_Map* pxEffectMap = AVP_SurfaceEffect_System::GetCollisionMap(uEffectTypeHash, uSurfaceHash, false);
				if(pxEffectMap && pxEffectMap->GetHash() != ASURA_HASH_ID_UNSET)
				{
					xStream << uSurfaceHash;
					xStream << uEffectTypeHash;
					xStream << pxEffectMap->GetHash();
				}
			}	
		}	


	}
	xStream.WriteChunk_End();
}

/// End of function 'AVP_Chunk_SurfaceOverlay::Create'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Chunk_SurfaceOverlay::Save
// Author       : Chris Wallace
// Date         : 2-Aug-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Chunk_SurfaceOverlay::Save(const char* szFilename)
{
	Asura_SaveBlock					xSaveBlock;
	Asura_Chunk_SaveBlock_Stream	xStream(&xSaveBlock);

	// Fill-in the save-block...
	Create(xStream);

	// Terminate the Asura-file...
	xStream.WriteAsuraTerminator();

	// ...before saving the file to disc...
	return(xSaveBlock.SaveFile(szFilename, false));
}

/// End of function 'AVP_Chunk_SurfaceOverlay::Save'. ///

#endif

/// End of file. ///
