
//////////////////////////////////////////////////////////////////////////////
//
// AVP_Interpolator.h
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef AVP_INTERPOLATOR_H___
#define AVP_INTERPOLATOR_H___

#define AVP_INERPOLATOR_DEBUG

#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"

#define AVP_INTERPOLATOR_NUM_POINTS 2
#define AVP_INTERPOLATOR_TIMEOUT	5.0f

// A class for handling the spherical linear interpolation of direction vectors
// intended for smoothing out look/aim directions received in network update messages
template < typename T >
class AVP_Interpolator
{
public:
	AVP_Interpolator( const T& fInitialValue, float fInitialTime = -1.0f );
	AVP_Interpolator();
	virtual ~AVP_Interpolator() {}

	void Update( const T& xNewValue, float fTime = -1.0f );
	void GetCurrent( T& xOutValue ) const;
	void Reset( const T& xNewValue, float fTime = -1.0f );

	virtual void Interpolate( const T& xFromValue, const T& xToValue, const float fFromTime, const float fToTime, const float fCurrentTime, T& xOutValue ) const;

private:
	T				m_axValues[ AVP_INTERPOLATOR_NUM_POINTS ];
	float			m_afTimes[ AVP_INTERPOLATOR_NUM_POINTS ];
	mutable T		m_xCurrentValue; // These are mutable as they are modified in the getter.
	mutable u_int	m_uFrameStamp;
	u_int			m_uNextSlot;
};

class AVP_Angle_Interpolator : public AVP_Interpolator< Asura_Angle >
{
public:
	AVP_Angle_Interpolator( const Asura_Angle& xInitialDirection, float fInitialTime = -1.0f ) : PARENT( xInitialDirection, fInitialTime ) {}
	AVP_Angle_Interpolator() : PARENT() {}
	virtual ~AVP_Angle_Interpolator() {}

	virtual void Interpolate( const Asura_Angle& xFromVal, const Asura_Angle& xToVal, const float fFromTime, const float fToTime, const float fCurrentTime, Asura_Angle& xOutVal ) const;
private:
	typedef AVP_Interpolator< Asura_Angle > PARENT;
};

class AVP_Direction_Interpolator : public AVP_Interpolator< Asura_AVector3 >
{
public:
	AVP_Direction_Interpolator( const Asura_AVector3& xInitialDirection, float fInitialTime = -1.0f ) : PARENT( xInitialDirection, fInitialTime ) {}
	AVP_Direction_Interpolator() : PARENT() {}
	virtual ~AVP_Direction_Interpolator() {}

	virtual void Interpolate( const Asura_AVector3& xFromVal, const Asura_AVector3& xToVal, const float fFromTime, const float fToTime, const float fCurrentTime, Asura_AVector3& xOutVal ) const;
private:
	typedef AVP_Interpolator< Asura_AVector3 > PARENT;
};

template < typename T >
AVP_Interpolator< T >::AVP_Interpolator( const T& xInitialValue, float fInitialTime /* = -1.0f */ )
: m_xCurrentValue( xInitialValue )
, m_uFrameStamp( 0 )
, m_uNextSlot( 1 )
{
	m_afTimes[ 0 ] = fInitialTime < 0.0f ? Asura_Timers::GetServerAbsoluteTime() : fInitialTime;
	m_axValues[ 0 ] = xInitialValue;
	for ( u_int uItr = 1; uItr < AVP_INTERPOLATOR_NUM_POINTS; ++uItr )
	{
		m_afTimes[ uItr ] = -1.0f;
	}
}

template < typename T >
AVP_Interpolator< T >::AVP_Interpolator()
: m_uFrameStamp(0)
, m_uNextSlot(0)
{
	for ( u_int uItr = 0; uItr < AVP_INTERPOLATOR_NUM_POINTS; ++uItr )
	{
		m_afTimes[ uItr ] = -1.0f;
	}
}

template < typename T >
void AVP_Interpolator< T >::Update( const T& xNewValue, float fTime /* = -1.0f */ )
{
	if ( fTime < 0.0f )
	{
		fTime = Asura_Timers::GetServerAbsoluteTime();
	}
	u_int uPrevSlot = m_uNextSlot - 1;
	if ( uPrevSlot >= AVP_INTERPOLATOR_NUM_POINTS )
	{
		uPrevSlot = AVP_INTERPOLATOR_NUM_POINTS - 1;
	}
	// ignore out of order packets or duplicate updates
	if ( fTime > m_afTimes[ uPrevSlot ] )
	{
		m_afTimes[ m_uNextSlot ] = fTime;
		m_axValues[ m_uNextSlot ] = xNewValue;

		if ( ++m_uNextSlot == AVP_INTERPOLATOR_NUM_POINTS )
		{
			m_uNextSlot = 0;
		}
	}
}

template < typename T >
void AVP_Interpolator< T >::GetCurrent( T& xOutValue ) const
{
	if ( m_uFrameStamp != Asura_Timers::GetFrameNumber() )
	{
		m_uFrameStamp = Asura_Timers::GetFrameNumber();
		const float fTime = Asura_Timers::GetServerAbsoluteTime();
		u_int uPrevSlot = m_uNextSlot - 1;
		if ( uPrevSlot >= AVP_INTERPOLATOR_NUM_POINTS )
		{
			uPrevSlot = AVP_INTERPOLATOR_NUM_POINTS - 1;
		}

		Asura_Assert( m_afTimes[ uPrevSlot ] > 0.0f, "No data in interpolator" );
		if ( m_afTimes[ uPrevSlot ] > 0.0f )
		{
			bool bDone = false;
			for ( u_int uItr = m_uNextSlot; uItr != uPrevSlot && !bDone; )
			{
				if ( m_afTimes[ uItr ] > 0.0f && m_afTimes[ uItr ] > ( fTime - AVP_INTERPOLATOR_TIMEOUT ) )
				{
					Interpolate( m_axValues[ uItr ], m_axValues[ uPrevSlot ], m_afTimes[ uItr ], m_afTimes[ uPrevSlot ], fTime, m_xCurrentValue );
					bDone = true;
				}

				if ( ++uItr == AVP_INTERPOLATOR_NUM_POINTS )
				{
					uItr = 0;
				}
			}

			if ( !bDone )
			{
				// then there is only one entry in the interpolator, better use that
				m_xCurrentValue = m_axValues[ uPrevSlot ];
			}
		}
	}
	xOutValue = m_xCurrentValue;
}

template < typename T >
void AVP_Interpolator< T >::Reset( const T& xNewValue, float fTime /* = -1.0f */ )
{
	m_xCurrentValue = xNewValue;
	m_axValues[ 0 ] = xNewValue;
	m_afTimes[ 0 ] = fTime < 0.0f ? Asura_Timers::GetServerAbsoluteTime() : fTime;
	for ( u_int uItr = 1; uItr < AVP_INTERPOLATOR_NUM_POINTS; ++uItr )
	{
		m_afTimes[ uItr ] = -1.0f;
	}
	m_uNextSlot = 1;
}

template < typename T >
void AVP_Interpolator< T >::Interpolate( const T& xFromVal, const T& xToVal, const float fFromTime, const float fToTime, const float fCurrentTime, T& xOutVal ) const
{
	float fParametric = Asura_Maths::ClampToRange( ( fCurrentTime - fToTime ) / ( fToTime - fFromTime ), 0.0f , 1.0f );
	xOutVal = xFromVal * ( 1.0f - fParametric );
	xOutVal = xOutVal + ( xToVal * fParametric );
}

#endif // AVP_INTERPOLATORS_H___
