
///////////////////////////////////////////////////////////////////////////////
//
// File		: AVP_Maths.cpp
// Author	: Paul Humphries
// Date		: 2-Mar-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

// Core Asura
#include "Core/Asura.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Timers.h"

// This file's header
#include "AVP_Maths.h"

// Asura debugging
#include "Debugging/Asura_DebugLineRender.h"


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Maths::ForceMajorAxisPositive
// Author       : Andy Tate
// Date         : 11-Mar-2008
////////////////////////////////////////////////////////////////////////////////

bool AVP_Maths::ForceMajorAxisPositive(Asura_AVector3_InParam xInput,
									   Asura_AVector3_OutParam xOutput)
{
	float	fSign;

	// Find the sign of the major-axis...
	if(Asura_Maths::Abs(xInput.GetX()) > Asura_Maths::Abs(xInput.GetY()))
	{
		fSign	= (Asura_Maths::Abs(xInput.GetX()) > Asura_Maths::Abs(xInput.GetZ()))
				? Asura_Maths::Sign(xInput.GetX())
				: Asura_Maths::Sign(xInput.GetZ());
	}
	else
	{
		fSign	= (Asura_Maths::Abs(xInput.GetY()) > Asura_Maths::Abs(xInput.GetZ()))
				? Asura_Maths::Sign(xInput.GetY())
				: Asura_Maths::Sign(xInput.GetZ());
	}

	// Update the output value and return whether we flipped the axis...
	xOutput	= xInput * fSign;
	return(fSign < 0.0f);
}

/// End of function 'AVP_Maths::ForceMajorAxisPositive'. ///


// Given an orientation matrix, extract the turn component only into a second orientation matrix

void AVP_Maths::ExtractTurnFrom(const Asura_Matrix_3x3& xOrientation, Asura_Matrix_3x3& xResult)
{
	Asura_Euler_3 xEuler;
	xEuler.FromMatrix(xOrientation);
	xEuler.Element[0] = 0.0f;
	xEuler.ToMatrix(xResult);
}

// Given an orientation matrix, extract the tilt component only into a second orientation matrix

void AVP_Maths::ExtractTiltFrom(const Asura_Matrix_3x3& xOrientation, Asura_Matrix_3x3& xResult)
{
	Asura_Euler_3 xEuler;
	xEuler.FromMatrix(xOrientation);
	xEuler.Element[1] = 0.0f;
	xEuler.ToMatrix(xResult);
}

// Perform a Line Of Sight test against everything, ignoring the given Guid

float AVP_Maths::LOSEverything(const Asura_Vector_3& xStart, const Asura_Vector_3& xNonNormalisedDirection, float fMaximumDistance, Asura_Guid uGuidToIgnore)
{
	Asura_Vector_3 xDirection = xNonNormalisedDirection;
	xDirection.Normalise();
	const Asura_LOS_InputData xInputData (xStart, xDirection, uGuidToIgnore, fMaximumDistance);
	const Asura_LOS_OutputData* pxOutputData = Asura_LOS::TestAgainstEverything(&xInputData);
	return pxOutputData->m_fDistanceToIntersection;
}

// Perform a Line Of Sight test against the environment

float AVP_Maths::LOSEnvironment(const Asura_Vector_3& xStart, const Asura_Vector_3& xNonNormalisedDirection, float fMaximumDistance)
{
	Asura_Vector_3 xDirection = xNonNormalisedDirection;
	xDirection.Normalise();
	const Asura_LOS_InputData xInputData (xStart, xDirection, ASURA_GUID_UNREGISTERED, fMaximumDistance);
	const Asura_LOS_OutputData* pxOutputData = Asura_LOS::TestAgainstEnvironment(&xInputData);
	return pxOutputData->m_fDistanceToIntersection;
}

// Write the mirror image (in the YZ plane) of the input vector to the output vector

void AVP_Maths::MirrorYZ(const Asura_Vector_3& xInput, Asura_Vector_3& xOutput)
{
	xOutput.x = -xInput.x;
	xOutput.y = xInput.y;
	xOutput.z = xInput.z;
}

// Write the mirror image (in the YZ plane) of the given euler (i.e. a rotation in the opposite direction around the y axis)

void AVP_Maths::MirrorYZ(const Asura_Euler_3& xInput, Asura_Euler_3& xOutput)
{
	xOutput.x = xInput.x;
	xOutput.y = -xInput.y;
	xOutput.z = xInput.z;
}

// Turn the given euler towards the target euler linearly

void AVP_Maths::TurnLinearlyTowards(Asura_Euler_3& xCurrent, const Asura_Euler_3& xTarget, Asura_Angle fAngularVelocity)
{
	Asura_Angle fTargetTurn = xTarget[1];
	Asura_Angle fCurrentTurn = xCurrent[1];

	if (fTargetTurn - fCurrentTurn > Asura_Maths::Deg180) fCurrentTurn += Asura_Maths::Deg360;
	if (fTargetTurn - fCurrentTurn < -Asura_Maths::Deg180) fCurrentTurn -= Asura_Maths::Deg360;
	xCurrent[1] = fCurrentTurn;

	Asura_Angle fStep = fAngularVelocity * Asura_Timers::GetGameTimeSlice();
	if (xCurrent[1] + fStep < xTarget[1]) xCurrent[1] += fStep;
	else if (xCurrent[1] - fStep > xTarget[1]) xCurrent[1] -= fStep;
	else xCurrent[1] = xTarget[1];

	xCurrent[1]	= Asura_Maths::RadiansToZeroCentredRange(xCurrent[1]);
}

// Turn the given euler towards the target euler exponentially i.e. move a fraction towards the target
// Note that although the tilt(x) and roll(z) components are also moved exponentially towards their target values, the cyclic
// aspect is not dealt with for these components i.e. this method is not intended for tilt or roll differences > 180 degrees

void AVP_Maths::TurnExponentiallyTowards(Asura_Euler_3& xCurrent, const Asura_Euler_3& xTarget, Asura_Angle fInitialAngularVelocity)
{
	Asura_Angle fTargetTurn = xTarget[1];
	Asura_Angle fCurrentTurn = xCurrent[1];

	if (fTargetTurn - fCurrentTurn > Asura_Maths::Deg180) fCurrentTurn += Asura_Maths::Deg360;
	if (fTargetTurn - fCurrentTurn < -Asura_Maths::Deg180) fCurrentTurn -= Asura_Maths::Deg360;
	xCurrent[1] = fCurrentTurn;

	xCurrent += (xTarget - xCurrent) * Asura_Maths::Min(fInitialAngularVelocity * Asura_Timers::GetGameTimeSlice(), 1.0f);

	xCurrent[1]	= Asura_Maths::RadiansToZeroCentredRange(xCurrent[1]);
}

// Move the given float towards the target float at the speed given

void AVP_Maths::LinearlyTowards(float& fCurrent, const float fTarget, const float fSpeed)
{
	float fStep = fSpeed * Asura_Timers::GetGameTimeSlice();
	if (fCurrent + fStep < fTarget)
	{
		fCurrent += fStep;
	}
	else if (fCurrent - fStep > fTarget)
	{
		fCurrent -= fStep;
	}
	else
	{
		fCurrent = fTarget;
	}
}

// Move the given value exponentially towards the target value i.e. move a fraction towards the target value

void AVP_Maths::ExponentiallyTowards(float& fCurrent, const float fTarget, const float fInitialSpeed)
{
	fCurrent += (fTarget - fCurrent) * Asura_Maths::Min(fInitialSpeed * Asura_Timers::GetGameTimeSlice(), 1.0f);
}

void AVP_Maths::RayOffsetFromPoint( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, const Asura_Vector_3& xPoint, Asura_Vector_3& xOffset )
{
	Asura_Assert( Asura_Maths::Abs( xRayDirection.MagnitudeSquared() - 1.0f ) < 1.0e-4f, "Direction should be normalised" );
	Asura_Vector_3 xDisplacement( xPoint - xRayOrigin );
	xOffset = xRayDirection * ( xDisplacement * xRayDirection ) - xDisplacement;
}

#ifndef ASURA_NO_DEBUG_OUTPUT

// Draw a one shot line given the start, non-normalised direction, length and (optional) colour

void AVP_Maths::DrawLine(const Asura_Vector_3& xStart, const Asura_Vector_3& xNonNormalisedDirection, float fLength, u_int uColour /*= 0xffffffff*/)
{
	Asura_Vector_3 xDirection = xNonNormalisedDirection;
	xDirection.Normalise();
	Asura_DebugLineRender::AddLine(xStart, xStart + xDirection * fLength, uColour);
}

// Create a timed line set consisting of one line with the start, non-normalised direction, length and (optional) colour

void AVP_Maths::CreateLine(const Asura_Vector_3& xStart, const Asura_Vector_3& xNonNormalisedDirection, float fLength, u_int uColour /*= 0xffffffff*/, float fDuration /*= 0.0f*/, float fFadeDuration /*= 60.0f*/)
{
	Asura_Vector_3 xDirection = xNonNormalisedDirection;
	xDirection.Normalise();
	Asura_DebugLineRender_LineSet* pxLineSet = Asura_DebugLineRender::CreateTimedLineSet(1, fDuration, fFadeDuration);
	pxLineSet->AddLine(xStart, xStart + xDirection * fLength, uColour);
}

// Create a timed line set consisting of two lines converging at the end point

void AVP_Maths::CreateArrow(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, u_int uColour /*= 0xffffffff*/, float fDuration /*= 0.0f*/, float fFadeDuration /*= 60.0f*/)
{
	Asura_DebugLineRender_LineSet* pxLineSet = Asura_DebugLineRender::CreateTimedLineSet(1, fDuration, fFadeDuration);
	float fDistance = (xStart - xEnd).Magnitude();
	pxLineSet->AddLine(xStart, xEnd, uColour);
	pxLineSet->AddLine(xStart + Asura_Vector_3(fDistance*0.1f, -fDistance*0.1f, 0.0f), xEnd, uColour);
}

#endif // ASURA_NO_DEBUG_OUTPUT

bool AVP_Maths::RayIntersectsOBB( const Asura_AVector3& xRayOrigin, const Asura_AVector3& xRayDirection, const Asura_OBB& xOBB, float* pfDistance /*= NULL*/, Asura_AVector3* pxHitPosition /*= NULL*/, bool* pbInside /*= NULL*/, Asura_AVector3* pxNormal /*= NULL*/ )
{
	// Probably much quicker to do a SAT test as I only want a boolean hit/no-hit but I can't be bothered right now

	// First, test against the bounding sphere.
	float fRadius = Asura_Maths::Sqrt( xOBB.GetHalfExtents() * xOBB.GetHalfExtents() );
	if ( Asura_Maths::RayIntersectsSphere( Asura_Vector_3( xRayOrigin ), Asura_Vector_3( xRayDirection ), Asura_Vector_3( xOBB.GetCentre() ), fRadius ) )
	{
		// Convert LOS test from world space to object space
		Asura_AVector3 xOSOrigin	= xRayOrigin;
		Asura_AVector3 xOSDirection	= xRayDirection;

		Asura_AMatrix3x3 xMat = xOBB.GetOrientation();
		xMat.Transpose();

		xOSOrigin -= xOBB.GetCentre();

		xOSOrigin = xOSOrigin * xMat;
		xOSDirection = xOSDirection * xMat;

		Asura_Bounding_Box xBB;
		xBB.MinX = -xOBB.GetHalfExtentX();
		xBB.MaxX = xOBB.GetHalfExtentX();
		xBB.MinY = -xOBB.GetHalfExtentY();
		xBB.MaxY = xOBB.GetHalfExtentY();
		xBB.MinZ = -xOBB.GetHalfExtentZ();
		xBB.MaxZ = xOBB.GetHalfExtentZ();
		
		Asura_Vector_3 xHitPosition3;
		Asura_Vector_3 xHitNormal3;
		// Now, do a LOS test against the rotated BB.
		if ( Asura_Maths::RayIntersectsBoundingBox( Asura_Vector_3( xOSOrigin ), Asura_Vector_3( xOSDirection ), xBB, pfDistance, pxHitPosition ? &xHitPosition3 : NULL, pbInside, pxNormal ? &xHitNormal3 : NULL ) )
		{
			// put output back in worldspace
			if ( pxHitPosition )
			{
				*pxHitPosition = Asura_AVector3( xHitPosition3 );
				*pxHitPosition = *pxHitPosition * xOBB.GetOrientation();
				*pxHitPosition += xOBB.GetCentre();
			}
			if ( pxNormal )
			{
				*pxNormal = Asura_AVector3( xHitNormal3 );
				*pxNormal = *pxNormal * xOBB.GetOrientation();
			}
			return true;
		}
	}
	return false;
}

bool AVP_Maths::IsAngleInRange( float fAngle, float fACWBound, float fCWBound, float* pfLerpParametric /*= NULL*/ )
{
	// force our angles into the correct ranges
	fAngle = Asura_Maths::RadiansToZeroMinimumRange( fAngle );
	fACWBound = Asura_Maths::RadiansToZeroMinimumRange( fACWBound );
	fCWBound = Asura_Maths::RadiansToZeroMinimumRange( fCWBound );

	if ( fACWBound >= fCWBound )
	{
		if ( fAngle >= fACWBound )
		{
			fCWBound += 2.0f * Asura_Maths::Pi;
		}
		else
		{
			fAngle += 2.0f * Asura_Maths::Pi;
			fCWBound += 2.0f * Asura_Maths::Pi;
		}
	}
	if ( pfLerpParametric )
	{
		*pfLerpParametric = ( fAngle - fACWBound ) / ( fCWBound - fACWBound );
	}
	return fACWBound <= fAngle && fAngle <= fCWBound;
}

void AVP_Maths::FindClosestPointOnEdge( Asura_AVector3_InParam xPoint, Asura_AVector3_InParam xV0, Asura_AVector3_InParam xV1, Asura_AVector3_OutParam xClosestPointOnEdge )
{
	Asura_AVector3 xEdgeDirection = xV1 - xV0;
	float fEdgeLength = xEdgeDirection.NormaliseReturningMagnitude();
	float fDistanceAlongEgde = Asura_Maths::ClampToRange( ( xPoint - xV0 ) * xEdgeDirection, 0.0f, fEdgeLength );
	xClosestPointOnEdge = xV0 + xEdgeDirection * fDistanceAlongEgde;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Maths::OffsetDirectionInCone
// Author       : Chris Wallace
// Date         : 03-Feb-2009
// 
// Description	: Based on Asura_Vector_3::RandomDirectionInConeSeeded, this 
//				  version takes a random vector (-1.0f to 1.0) for the amount
//				  to move away from the original direction, this is then scaled 
//				  based on MaxTheta but is otherwise untouched.
//
////////////////////////////////////////////////////////////////////////////////

Asura_Vector_3 AVP_Maths::OffsetDirectionInCone(Asura_Vector_3& xOriginalDirection, Asura_Vector_2 xNewDirection, const float fMaxTheta)
{
	Asura_Vector_3 xDirection = xOriginalDirection;
	xDirection.Normalise();

	if(fMaxTheta < 0.0f) return xDirection;

	//Now, we construct an arbitrary orthogonal basis about the axis.
	float fAbsX = Asura_Maths::Abs(xDirection.x);
	float fAbsY = Asura_Maths::Abs(xDirection.y);
	float fAbsZ = Asura_Maths::Abs(xDirection.z);

	//Asura_Vector_3 xContructionAxis(1.f, 0.f, 0.f);
	Asura_Vector_3 xNormal;

	//Work out the largest component of xDirection, and construct an appropriate normal,
	//using the unrolled cross product with that axis.
	if(fAbsX > fAbsY)
	{
		if(fAbsY > fAbsZ)
		{	
			//Cross with z-axis.
			xNormal.x = xDirection.y;
			xNormal.y = -xDirection.x;
			xNormal.z = 0.f;
		}
		else
		{
			//Cross with y-axis.
			xNormal.x = -xDirection.z;
			xNormal.y = 0.f;
			xNormal.z = xDirection.x;
		}
	}
	else if(fAbsX > fAbsZ)
	{
		//Cross with z-axis.
		xNormal.x = xDirection.y;
		xNormal.y = -xDirection.x;
		xNormal.z = 0.f;
	}
	else
	{
		//Cross with x-axis.
		xNormal.x = 0.f;
		xNormal.y = xDirection.z;
		xNormal.z = -xDirection.y;
	}

	xNormal.Normalise();

	//Now that we have the first normal, we can construct the second.
	Asura_Vector_3 xNormal2 = xDirection^xNormal;

	//Now we choose a random number between the cosines of the two angles.
	float fScale = Asura_Maths::Cosine(fMaxTheta);

	//Just check the normalisation, to be sure.
	if(fScale > 1.f) fScale = 1.f;
	else if(fScale < -1.f) fScale = -1.f;

	//Scale the primary component to the chosen length.
	xNewDirection *= fScale;

	//Work out the correct normalisation factor for the second component...
	//...and add the second component, scaled by that factor.
	xDirection += (xNormal * xNewDirection.x + xNormal2 * xNewDirection.y) * Asura_Maths::Sqrt(1.f - fScale*fScale);

	return xDirection;
}

/// End of file. ///

