//////////////////////////////////////////////////////////////////////////////
//
// AVP_Profile.h
//
// (AJK - 04.08.2008) Created
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __AVP_PROFILE__
#define __AVP_PROFILE__

// Asura
#include "Core/Asura_StringBuffer.h"
#include "ChunkSystem/Asura_SaveBlock.h"

// AVP
#include "Core/Profile/AVP_Profile_Controls.h"
#include "Core/Profile/AVP_Profile_Graphics.h"
#include "Core/Profile/AVP_Profile_Sounds.h"

class AVP_Profile_LoadAndSave;

class AVP_Profile
{
	// friend class here so the AVP_Profile_LoadAndSave system can have access to the data without the use of a large numbers of accessors. 
	// This class has been split in half to seperate the load/save functionality
	// This is the best way to allow the Load/Save procedure access to the data (that goes both ways) whilest on the whole preserving encapsulation.
	// The data needed from this class by AVP_Profile_LoadAndSave has no context outside of these 2 classes and with this set up it is all still private to others. 
	// All Functions in AVP_Profile_LoadAndSave that talk to AVP_Profile are private in AVP_Profile_LoadAndSave
	friend class AVP_Profile_LoadAndSave;

public:
	enum E_PROFILE_FILE_STATES
	{
		PROFILE_FILE_OK,
		PROFILE_FILE_NOT_FOUND,
		PROFILE_FILE_NOT_FOUND_NOUSER,
		PROFILE_FILE_CORRUPT,
		PROFILE_DEVICE_ERROR,
	};

	enum OPTION_SCREEN_GUIDS
	{
		FRONTEND_CONTROL,
		FRONTEND_GRAPHICS,
		FRONTEND_SOUND,

		MARINE_CONTROL,
		MARINE_GRAPHICS,
		MARINE_SOUND,

		ALIEN_CONTROL,
		ALIEN_GRAPHICS,
		ALIEN_SOUND,

		PREDATOR_CONTROL,
		PREDATOR_GRAPHICS,
		PREDATOR_SOUND,

		NUM_OPTION_SCREEN_GUIDS
	};

	static bool ConsoleCmd_Save         ( bool bHelp);
	static bool ConsoleCmd_Restore      ( bool bHelp);
	static bool ConsoleCmd_OpenOptionsMenu( bool bHelp);
	static bool ConsoleCmd_OpenConfirmCancelMenu( bool bHelp);
	static bool ConsoleCmd_ResetDefaults( u_int uElementIndex, bool bInterrogate, bool bHelp );

	void UpdateChange();
	bool IsChanged() const { return m_bChanged; }

	AVP_Profile();
	void Shutdown();

    void SetToDefaults();

	void ApplySettings();

	E_PROFILE_FILE_STATES Load();
	void Save();
	void Restore();

	AVP_Profile_Controls&	GetControls() { return m_xControls; }
	AVP_Profile_Graphics&	GetGraphics() { return m_xGraphics; }
	AVP_Profile_Sounds&		GetSounds()   { return m_xSounds;   }

	void  SetNumber(const u_int uNumber) { m_uNumber = uNumber; }
	u_int GetNumber() const { return m_uNumber; }
	const Asura_StringBuffer<128>& GetName() const { return m_xName; }
	void  SetName(Asura_StringBuffer<128>& xName)   { m_xName = xName;}

	void Update();

private:

	void SetupConsoleCommands();
	void SetupConsoleVariables();
	void RemoveConsoleVariables();

	static u_int GetCurSpeciesOptionsMenuHash();
	static u_int GetCurSpeciesConfirmCancelMenuHash();
	static u_int GetCurSpeciesPauseMenuHash();

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream);

	void Platform_SetToDefaults	();
	bool Platform_FillFilename	(Asura_StringBuffer<128>& xFilename);
	bool Platform_SaveFile		(Asura_SaveBlock& xSaveBlock);
	bool Platform_DeleteFile	();
	bool Platform_Initialise	();

	static bool	s_bConsoleCommandsValid;
	static bool	s_bConsoleVariablesValid;

    bool                    m_bChanged;
	u_int                   m_uNumber;
	Asura_StringBuffer<128> m_xName;

	AVP_Profile_Controls    m_xControls;
	AVP_Profile_Graphics    m_xGraphics;
	AVP_Profile_Sounds      m_xSounds;
    
	//Backup profile options
	AVP_Profile_Controls    m_xBackupControls;
	AVP_Profile_Graphics    m_xBackupGraphics;
	AVP_Profile_Sounds      m_xBackupSounds;

	enum 
	{
		CURRENT_CHUNK_VERSION = 1,
	};

	static Asura_Hash_ID s_uaOptionScreenGuids[NUM_OPTION_SCREEN_GUIDS];

	static const u_int s_uAVP_PROFILE_ALIEN_OPTIONS_MENU		   = 0x08DC0D26; //Asura_GetHashID( "Alien_Options"); 
	static const u_int s_uAVP_PROFILE_MARINE_OPTIONS_MENU		   = 0xC2A1D921; //Asura_GetHashID( "Marine_Options"); 
	static const u_int s_uAVP_PROFILE_PREDATOR_OPTIONS_MENU		   = 0x918DC956; //Asura_GetHashID( "Predator_Options"); 

	static const u_int s_uAVP_PROFILE_ALIEN_CONFIRM_CANCEL_MENU    = 0x6893BD51; //Asura_GetHashID( "Alien_Confirm_Cancel"); 
	static const u_int s_uAVP_PROFILE_MARINE_CONFIRM_CANCEL_MENU   = 0x7F8A90F6; //Asura_GetHashID( "Marine_Confirm_Cancel"); 
	static const u_int s_uAVP_PROFILE_PREDATOR_CONFIRM_CANCEL_MENU = 0xBDEEEB21; //Asura_GetHashID( "Predator_Confirm_Cancel"); 

	static const u_int s_uAVP_PROFILE_ALIEN_PAUSE_MENU             = 0xA8EDAE7E; //Asura_GetHashID( "Alien_Pause"); 
	static const u_int s_uAVP_PROFILE_MARINE_PAUSE_MENU			   = 0x741C3D39; //Asura_GetHashID( "Marine_Pause"); 
	static const u_int s_uAVP_PROFILE_PREDATOR_PAUSE_MENU		   = 0x3A40B6AE; //Asura_GetHashID( "Predator_Pause"); 
};

#endif // #ifndef __AVP_PROFILE__