
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_ConnectionState_EndGameInitialisation.cpp
// Author : Andy Tate
// Date   : 25-Feb-2009
//
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Core/States/Connection/AVP_ConnectionState_EndGameInitialisation.h"

// Asura independent header-files...
#include "Core/Asura_Multithread.h"
#include "Debugging/Asura_DebugText.h"
#include "Entity/Asura_Entity_System.h"
#include "Sound/Asura_Audio_Systems_Manager.h"

// Game independent header-files...
#include "Core/AVP_Connection.h"
#include "Core/AVP_Core.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_GameStatus.h"
#include "Entity/AVP_Messages.h"
#include "Entity/AVP_Server_Messages.h"
#include "GUIMenu/AVP_GUIMenu_System.h"
#include "Network/AVP_Player.h"
#include "Network/AVP_Server.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

volatile bool	AVP_ConnectionState_EndGameInitialisation::s_bInitialisationComplete	= false;


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ConnectionState_EndGameInitialisation::PerformInitialiseCompletion
// Author       : Andy Tate
// Date         : 25-Feb-2009
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_EndGameInitialisation::PerformInitialiseCompletion(void)
{
	// This is here to initialise the occlusion (which is not done during the loading/initialising
	// phase to prevent simultaneous blueprint lookups (which aren't thread safe as they use hash trees))
	Asura_Entity_System::PostCreateInitialise();

	// We're all done...
	AVP_GameStatus::ResetStatusChanges();
	s_bInitialisationComplete	= true;
}

/// End of function 'AVP_ConnectionState_EndGameInitialisation::PerformInitialiseCompletion'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ConnectionState_EndGameInitialisation::Enter
// Author       : Andy Tate
// Date         : 25-Feb-2009
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_EndGameInitialisation::Enter(const u_int uPreviousState)
{
	// Actually perform/begin the initialisation...
	s_bInitialisationComplete	= false;

	AVP_Core::SetShouldUpdateSoundSystem(false);

#if defined(AVP_ASYNCHRONOUS_LOADING)
	Asura_Thread_System::CreateThread("PerformGameInitialisation", PerformInitialiseCompletion, this);
#else
	PerformInitialiseCompletion();
#endif	//defined(AVP_ASYNCHRONOUS_LOADING)
}

/// End of function 'AVP_ConnectionState_EndGameInitialisation::Enter'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ConnectionState_EndGameInitialisation::Update
// Author       : Andy Tate
// Date         : 25-Feb-2009
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_EndGameInitialisation::Update(void)
{
	if(Asura_Network::IsServer())
	{
		AVP_Server::UpdateClientInitialisation();
	}

	Asura_Network::Update();

	// Is it time to start the game?
	if(s_bInitialisationComplete)
	{

		if(Asura_Network::IsServer())
		{
			Asura_Network::EndOfFrameUpdate();
		}
		else
		{
			Asura_Network::FlushBufferedMessages();
		}

		GetParentMachine().ChangeState(AVP_Connection::STATE_PLAYING);
	}
}

/// End of function 'AVP_ConnectionState_EndGameInitialisation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ConnectionState_EndGameInitialisation::Render
// Author       : Andy Tate
// Date         : 25-Feb-2009
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_EndGameInitialisation::Render(void) const
{
#if !defined(ASURA_TOOLS)
	AVP_Core::BeginRender(false);
	{
		AVP_GUIMenu_System::Render(true);
#if defined(ASURA_USE_DEBUG_TEXT)
		Asura_DebugText::AddTextBlock(50, 20, 0xFFFFFFFF, 0, "EndGameInitialisation: %.02f secs", GetParentMachine().GetTimeInCurrentState());
#endif	// defined(ASURA_USE_DEBUG_TEXT)
	}
	AVP_Core::EndRender();
#endif	// defined(ASURA_TOOLS)
}

/// End of function 'AVP_ConnectionState_EndGameInitialisation::Render'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ConnectionState_EndGameInitialisation::Exit
// Author       : Andy Tate
// Date         : 25-Feb-2009
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_EndGameInitialisation::Exit(const u_int uNextState)
{
	AVP_Core::SetShouldUpdateSoundSystem(true);

#ifdef ASURA_USE_GUI_MENU_SYSTEM
	AVP_GUIMenu_System::Deactivate();

	if (AVP_Settings::IsMultiPlayer())
	{
		// Any of these loading screens could be open by now, depending on the player species selection
		AVP_GUIMenu_System::Deactivate( AVP_GUIMenu_System::GUIMENU_ID_LOADING, AVP_Settings::AVP_SPECIES_ALIEN ); 
		AVP_GUIMenu_System::Deactivate( AVP_GUIMenu_System::GUIMENU_ID_LOADING, AVP_Settings::AVP_SPECIES_PREDATOR ); 		
		AVP_GUIMenu_System::Deactivate( AVP_GUIMenu_System::GUIMENU_ID_LOADING, AVP_Settings::AVP_SPECIES_HUMAN ); 	
	}
#endif //#ifdef ASURA_USE_GUI_MENU_SYSTEM

	AVP_GameStatus::ResetStatusChanges();

	// Send a message telling this remote player is ready to play, and requesting activation of this actor
	AVP_Player* pxPlayer = AVP_Game_Core::GetPlayer(); 

	if (pxPlayer)
	{
		AVP_Message_JoinReady xData; 
		xData.SetPlayerGuid( pxPlayer->GetGuid() ); 

		Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_JOINGAME_READY, sizeof(xData), pxPlayer->GetClassification(), &xData ); 
		if (pxMessage)
		{
			pxMessage->SetToGuid(ASURA_GUID_SERVER);
			pxMessage->SetFromGuid(pxPlayer->GetGuid());
			pxMessage->UseGuaranteedDelivery(); 

			Asura_Network::Send(pxMessage);
		}
		else
		{
			Asura_Assert( false, "Failed to create join game ready message, the server won't know that you are ready to play" ); 
		}
	}
	else
	{
		Asura_Assert( AVP_Settings::GetNumberOfLocalPlayers() == 0, "Couldn't find a player to activate" ); 
	}

#if defined(ASURA_MEMORYMANAGEMENT_ENABLED)
	//Asura_MemoryManagement_System::DumpMallocSummary();
#endif	//defined(ASURA_MEMORYMANAGEMENT_ENABLED)
}

/// End of function 'AVP_ConnectionState_EndGameInitialisation::Exit'. ///


/// End of file. ///
