
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_ConnectionState_Loading.cpp
// Author : Anthony Gregson
// Date   : 22/11/07
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED
#include "Core/Asura_MemoryManagement.h"
#endif
// This file's header...
#include "Core/States/Connection/AVP_ConnectionState_Loading.h"

// Asura independent header-files...
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_MemoryManagement_Enums.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Texture_Streaming.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "Core/Resources/Asura_Resource_System.h"
#include "Debugging/Asura_DebugText.h"
#include "Entity/Asura_Entity_System.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Network/Asura_Network.h"
#include "Rendering/Asura_Module.h"
#include "Teleport/Asura_Teleport_System.h"

// Axon independent header-files...
#include "Core/Axon_System.h"

// Game independent header-files...
#include "Actor/AVP_Actor_HUDHandler.h"
#include "Core/AVP_Connection.h"
#include "Core/AVP_Core.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_Settings.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "LoadAndSave/AVP_WeaponData_System.h"
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
#include "GUIMenu/AVP_GUIMenu_System.h"
#endif
#include "Network/AVP_Player.h"

// Tools header-files...


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

volatile bool AVP_ConnectionState_Loading::s_bLoadCompleted	= false;

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function	: AVP_ConnectionState_Loading::PerformLoad
// Author	: Anthony Gregson
// Date		: 22/11/07
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_Loading::PerformLoad(const void* pUserData)
{
	const LoadingData* pxLoadingData = reinterpret_cast<const LoadingData*>(pUserData);
	bool bLoadPlayerData = true; 
	if (pxLoadingData)
	{
		bLoadPlayerData = !pxLoadingData->m_bOnlyLoadLevelFile;
	}


	// Load the level file...
	if(pxLoadingData == NULL || pxLoadingData->m_szFilename == NULL)
	{
		Asura_Assert(false, "No level file to load.");
	}
	else
	{
#ifdef ASURA_USE_STREAMING
		Asura_Streaming_Zone* pxZone = Asura_Streaming_System::CreateZone(Asura_Streaming_System::MISSION_DATA, "Empty mission");
		if (pxZone)
		{
			Asura_Streaming_System::CreateEmptyZone(Asura_Streaming_System::MISSION_DATA, pxZone->GetHashID());
		}

		pxZone = Asura_Streaming_System::CreateZone(Asura_Streaming_System::ZONE_DATA, szLevelFile);
		if (pxZone)
		{
			Asura_Streaming_System::LoadZone(Asura_Streaming_System::ZONE_DATA, pxZone->GetHashID(), true);
		}
#else //#ifdef ASURA_USE_STREAMING
		LoadLevel(pxLoadingData->m_szFilename);
#endif //#else //#ifdef ASURA_USE_STREAMING
	}

	if (bLoadPlayerData)
	{
		// Load in required character files
		LoadCharacterFiles();
	
		// Load in the required weapon files
		LoadWeaponFiles();
	}

	// Finished loading...
	s_bLoadCompleted	= true;
}

/// End of function 'AVP_ConnectionState_Loading::PerformLoad'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function	: AVP_ConnectionState_Loading::Enter
// Author	: Anthony Gregson
// Date		: 22/11/07
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_Loading::Enter(const u_int uPreviousState)
{
	AVP_Core::SetShouldUpdateSoundSystem(false);
	m_xLoadingData.m_bOnlyLoadLevelFile = false; 

	switch (uPreviousState)
	{
		case AVP_Connection::STATE_INVALID:
		{
			AVP_Core::InitialiseNetwork(AVP_Settings::GetExternalNetworkActive());
			//Fall thru
		}

		case AVP_Connection::STATE_WAITFORPLAYERS:
		{
			AVP_Core::InitialiseEffects();
			break;
		}
		case AVP_Connection::STATE_MAPCHANGE:
		{
			// JG - 05 May 2009 - Doesn't seem like we will be able to use this flag yet
			//m_xLoadingData.m_bOnlyLoadLevelFile = true; 
			break;
		}

		default: break;
	}

	// Tools tends to allocate stuff all over the place, but we tend not to worry about it either since it's often just because a change has caused more files to be loaded.
#if defined(ASURA_MEMORYMANAGEMENT_ENABLED) && !defined(ASURA_TOOLS)
	Asura_MemoryManagement_System::Checkpoint(ASURA_MM_STATE_PRE_LEVEL_LOAD);
#endif
	// Load the blueprint data...
	m_xLoadingData.m_szFilename		= AVP_Settings::GetLevelName();
	m_xStartTick					= Asura_Timers::GetCurTicks();
	s_bLoadCompleted				= false;

	// Ensure we've got the correct default data...
	AVP_Settings::RefreshBlueprintData(m_xLoadingData.m_szFilename);

#if !defined(ASURA_TOOLS)
	//GJDH - Activate loading menu here
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
	AVP_GUIMenu_System::Activate( AVP_GUIMenu_System::GUIMENU_ID_LOADING );
#endif
#else
	// Do we need to load now?
	if(!AVP_Settings::HasSetLevelName())
	{
		// must have quit to game from tools, so don't reload env.
		PerformLoad(NULL);
	}
	else
#endif	// !defined(ASURA_TOOLS)
	{
#if defined(AVP_ASYNCHRONOUS_LOADING)
		Asura_Thread_System::CreateThread("AVP_ConnectionState_Loading::PerformLoad", PerformLoad, &m_xLoadingData );
#endif	// defined(AVP_ASYNCHRONOUS_LOADING)	
	}
}

/// End of function 'AVP_ConnectionState_Loading::Enter'. ///



////////////////////////////////////////////////////////////////////////////////
// Function	: AVP_ConnectionState_Loading::Update
// Author	: Anthony Gregson
// Date		: 22/11/07
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_Loading::Update()
{
#if defined(ASURA_TOOLS)
	// We don't need to load any more...
	s_bLoadCompleted	= true;
#else
#if !defined(AVP_ASYNCHRONOUS_LOADING)
	const float	fTimeInState	= GetParentMachine().GetTimeInCurrentState();

	if(fTimeInState > 1.0f)
	{
#if defined(AVP_USE_DEVELOPMENT_FRONT_END) || defined(AVP_USE_GAME_FRONTEND)
		PerformLoad(m_szFilename);
#endif	// defined(AVP_USE_DEVELOPMENT_FRONT_END) | defined(AVP_USE_GAME_FRONTEND)
	}
#endif	// !defined(AVP_ASYNCHRONOUS_LOADING)
#endif	// defined(ASURA_TOOLS)

	if(s_bLoadCompleted)
	{
		GetParentMachine().ChangeState(AVP_Connection::STATE_WAITFORGUID);
	}
	else
	{
		// While loading, keep the network updated, so we can get notice of client guids and the like while this is going on.
		Asura_Network::Update();
#ifdef AVP_ASYNCHRONOUS_LOADING // Probably irrelevant, since we almost certainly have to be loading asynchronously to hit this code.
		Asura_Thread_System::YieldCurrentThread();
#endif //#ifdef AVP_ASYNCHRONOUS_LOADING
	}
}

/// End of function 'AVP_ConnectionState_Loading::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function	: AVP_ConnectionState_Loading::Render
// Author	: Anthony Gregson
// Date		: 22/11/07
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_Loading::Render() const
{
#if !defined(ASURA_TOOLS)
	AVP_Core::BeginRender(false);
	{
		AVP_GUIMenu_System::Render(true);

#if defined(ASURA_USE_DEBUG_TEXT)
		{
			const Asura_TickCount	xElapsedTicks	= Asura_Timers::GetElapsedTicks(m_xStartTick);
			const float				fElapsedSecs	= Asura_Timers::TicksToSeconds(xElapsedTicks);

			Asura_DebugText::AddTextBlock(50, 20, 0xFFFFFFFF, 0, "Loading: %.02f secs", fElapsedSecs);
		}
#endif	// defined(ASURA_USE_DEBUG_TEXT)
	}
	AVP_Core::EndRender();
#endif	// defined(ASURA_TOOLS)
}

/// End of function 'AVP_ConnectionState_Loading::Render'. ///



////////////////////////////////////////////////////////////////////////////////
// Function	: AVP_ConnectionState_Loading::Exit
// Author	: Anthony Gregson
// Date		: 22/11/07
////////////////////////////////////////////////////////////////////////////////

void AVP_ConnectionState_Loading::Exit(const u_int uNextState)
{
	AVP_Core::SetShouldUpdateSoundSystem(true);
}

/// End of function 'AVP_ConnectionState_Loading::Exit'. ///

void AVP_ConnectionState_Loading::LoadCharacterFiles()
{
	bool bCharLoadOK = true;

	if(AVP_Settings::IsSinglePlayer())
	{
		const AVP_Settings::Species	eSpecies	= AVP_Settings::GetPlayerSpecies();
		switch(eSpecies)
		{
			case AVP_Settings::AVP_SPECIES_PREDATOR:
			{
				Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_PLAYER_PREDATOR, szAVP_RESOURCE_BANK_NAME_PLAYER_PREDATOR, uAVP_RESOURCE_BANK_TRG_PLAYER_PREDATOR);
				bCharLoadOK	&= AVP_Core::LoadFile("Characters\\SP_Predator." ASURA_STANDARD_FILE_EXTENSION);
				bCharLoadOK &= AVP_Core::LoadFile("Graphics\\Predator_HUD_Elements."ASURA_STANDARD_FILE_EXTENSION);
				bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Predator_InGame." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);
				break;
			}
			case AVP_Settings::AVP_SPECIES_ALIEN:
			{
				Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_PLAYER_ALIEN, szAVP_RESOURCE_BANK_NAME_PLAYER_ALIEN, uAVP_RESOURCE_BANK_TRG_PLAYER_ALIEN);
				bCharLoadOK &= AVP_Core::LoadFile("Characters\\SP_Alien." ASURA_STANDARD_FILE_EXTENSION);
				bCharLoadOK &= AVP_Core::LoadFile("Graphics\\Alien_HUD_Elements."ASURA_STANDARD_FILE_EXTENSION);
				bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Alien_InGame." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);
				break;
			}

			default:
			{
				// We have got this far and no character type has been either selected or found
				// Load in a marine for now and set the type.
				Asura_Assert(false, "Unknown character type.");
				AVP_Settings::SetPlayerSpecies(AVP_Settings::AVP_SPECIES_HUMAN);
			}
			// fall through to human
	
			case AVP_Settings::AVP_SPECIES_HUMAN:
			{
				Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_PLAYER_MARINE, szAVP_RESOURCE_BANK_NAME_PLAYER_MARINE, uAVP_RESOURCE_BANK_TRG_PLAYER_MARINE);
				bCharLoadOK &= AVP_Core::LoadFile("Characters\\SP_Marine." ASURA_STANDARD_FILE_EXTENSION);
				bCharLoadOK &= AVP_Core::LoadFile("Graphics\\Marine_HUD_Elements."ASURA_STANDARD_FILE_EXTENSION);
				bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Marine_InGame." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);
				break;
			}
		}
	}
	else
	{
		// Load all skins and HUDs
		Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_PLAYER_PREDATOR, szAVP_RESOURCE_BANK_NAME_PLAYER_PREDATOR, uAVP_RESOURCE_BANK_TRG_PLAYER_PREDATOR);
		bCharLoadOK	&= AVP_Core::LoadFile("Characters\\MP_Predator." ASURA_STANDARD_FILE_EXTENSION);
		bCharLoadOK &= AVP_Core::LoadFile("Graphics\\Predator_HUD_Elements."ASURA_STANDARD_FILE_EXTENSION);
		bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Predator_InGame." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);

		Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_PLAYER_ALIEN, szAVP_RESOURCE_BANK_NAME_PLAYER_ALIEN, uAVP_RESOURCE_BANK_TRG_PLAYER_ALIEN);
		bCharLoadOK &= AVP_Core::LoadFile("Characters\\MP_Alien." ASURA_STANDARD_FILE_EXTENSION);
		bCharLoadOK &= AVP_Core::LoadFile("Graphics\\Alien_HUD_Elements."ASURA_STANDARD_FILE_EXTENSION);
		bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Alien_InGame." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);

		Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_PLAYER_MARINE, szAVP_RESOURCE_BANK_NAME_PLAYER_MARINE, uAVP_RESOURCE_BANK_TRG_PLAYER_MARINE);
		bCharLoadOK &= AVP_Core::LoadFile("Characters\\MP_Marine." ASURA_STANDARD_FILE_EXTENSION);
		bCharLoadOK &= AVP_Core::LoadFile("Graphics\\Marine_HUD_Elements."ASURA_STANDARD_FILE_EXTENSION);
		bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Marine_InGame." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);

		Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_MULTIPLAYER, szAVP_RESOURCE_BANK_NAME_MULTIPLAYER, uAVP_RESOURCE_BANK_TRG_MULTIPLAYER);
		bCharLoadOK &= AVP_Core::LoadFile("GUIMenu\\Multiplayer." ASURA_GUIMENU_STANDARD_FILE_EXTENSION);
	}

	if(!bCharLoadOK)
	{
#if defined(ASURA_ASSERT)
		Asura_Assert(false, "Failed to load in Character file(s).");
#else
		Asura_OutputToDebugger_Retail(("*** Failed to load in Character file(s). ***\n"));
#endif	//defined(ASURA_ASSERT)
	}

	Asura_TiledTextureManagement::SetTextureHandlesFromNames();
}

bool AVP_ConnectionState_Loading::LoadWeaponFiles()
{
	bool bLoadOK = true;
	for ( u_int uCharItr = 0; uCharItr < AVP_Game_Core::NUM_CHARACTERTYPES; ++uCharItr )
	{
		AVP_Game_Core::CharacterType eCharType = static_cast< AVP_Game_Core::CharacterType >( uCharItr );
		const u_int uNumWeapons = AVP_WeaponData_System::GetNumWeaponsForCharacterType( eCharType );
		for ( u_int uWeaponItr = 0; uWeaponItr < uNumWeapons; ++uWeaponItr )
		{
			// always load the tequila sp weapon for now
			if ( AVP_WeaponData_System::GetShouldLoadWeapon( uWeaponItr, eCharType ) || Asura_GetHashID( AVP_WeaponData_System::GetWeaponNameFromIndex( uWeaponItr, eCharType ) ) == uAVP_INVENTORY_WEAPONHASH_TEQUILA_SP )
			{
				const char* szWeaponName = AVP_WeaponData_System::GetWeaponNameFromIndex( uWeaponItr, eCharType );
				const Asura_Hash_ID uWeaponHash = Asura_GetHashID( szWeaponName );
				Asura_String xWeaponFilename("Characters\\");
				xWeaponFilename += szWeaponName;
				xWeaponFilename += "." ASURA_STANDARD_FILE_EXTENSION;
				Asura_Resource_System::CreateResourceBank( uWeaponHash, szWeaponName, uAVP_RESOURCE_BANK_TRG_PLAYER_WEAPONS );
				bool bWeaponLoadOK = AVP_Core::LoadFile( xWeaponFilename.GetString() );
				bLoadOK &= bWeaponLoadOK;
				if ( !bWeaponLoadOK )
				{
#if defined(ASURA_ASSERT)
					Asura_Assertf(false, ("Failed to load in Weapon file. %s", xWeaponFilename.GetString()));
#else
					Asura_OutputToDebugger_Retail(("*** Failed to load in Weapon file. ***\n", xWeaponFilename.GetString()));
#endif	//defined(ASURA_ASSERT)
				}
			}
		}
	}
	return bLoadOK;
}

void AVP_ConnectionState_Loading::LoadLevel(const char* szLevelFile)
{
	Asura_Resource_System::CreateResourceBank(uAVP_RESOURCE_BANK_HASH_LEVEL, szAVP_RESOURCE_BANK_NAME_LEVEL, uAVP_RESOURCE_BANK_TRG_LEVEL);
	if(AVP_Core::LoadFile(szLevelFile))
	{
		char sTempBuffer[256];

		// create an Asura string to allow easy manipulation
		Asura_String xLevelNameWithPathAndExtension( szLevelFile );
		Asura_String xLevelNameStripped;

		// strip the path and extension of the level file then use the level name to find and set the correct environment
		xLevelNameWithPathAndExtension.SplitFilename( NULL, &xLevelNameStripped, NULL );
		Asura_PrintfToBfr( sTempBuffer, "GUIMenu\\%s." ASURA_GUIMENU_STANDARD_FILE_EXTENSION, 256, true, xLevelNameStripped.GetString() );

		// check and double check, if the file exists try to load it. 
		if( Asura_FileAccess::FileExists( sTempBuffer ) )
		{
			if( AVP_Core::LoadFile( sTempBuffer ) )
			{	
				AVP_Actor_HUDHandler::SetHUDSubLevelLoaded( true );
				AVP_Actor_HUDHandler::SetHUDSubLevelHashID( Asura_GetHashID( xLevelNameStripped.GetString() ) );
			}
			else
			{
				AVP_Actor_HUDHandler::SetHUDSubLevelLoaded( false );
				AVP_Actor_HUDHandler::SetHUDSubLevelHashID( 0 );
			}
		}
		else
		{
			AVP_Actor_HUDHandler::SetHUDSubLevelLoaded( false );
			AVP_Actor_HUDHandler::SetHUDSubLevelHashID( 0 );
		}

		AddLevelBlueprintWeaponsToLoadList( szLevelFile );
	}
	else
	{
#if defined(ASURA_ASSERT)
		Asura_Assertf(false, ("Failed to load level \"%s\".", szLevelFile));
#else
		Asura_OutputToDebugger_Retail(("*** Failed to load level \"%s\". ***\n", szLevelFile));
#endif
	}
}

void AVP_ConnectionState_Loading::AddLevelBlueprintWeaponsToLoadList( const char* szLevelFile )
{
	const bool bIsSinglePlayer = AVP_Settings::IsSinglePlayer();
	if (bIsSinglePlayer)
	{
		AVP_Game_Core::CharacterType eCharType = AVP_Game_Core::CHARACTERTYPE_UNKNOWN;
		const AVP_Settings::Species	eSpecies	= AVP_Settings::GetPlayerSpecies();
		switch(eSpecies)
		{
			case AVP_Settings::AVP_SPECIES_ALIEN: eCharType = AVP_Game_Core::SINGLE_PLAYER_ALIEN; break;
			case AVP_Settings::AVP_SPECIES_HUMAN: eCharType = AVP_Game_Core::SINGLE_PLAYER_MARINE; break;
			case AVP_Settings::AVP_SPECIES_PREDATOR: eCharType = AVP_Game_Core::SINGLE_PLAYER_PREDATOR; break;
			default: Asura_Assert( false, "Invalid character type" ); return;
		}
		
		const Asura_Blueprint* pxLevelBP = AVP_Game_Core::GetLevelBlueprint( szLevelFile, eCharType );
		if ( pxLevelBP )
		{
			Asura_Hash_ID auWeaponHashesInLevel[ uAVP_MAX_WEAPONS_IN_LEVEL ];
			AVP_Blueprint_ParameterEnumerator_LevelWeapons xWeaponEnumerator( auWeaponHashesInLevel );
			u_int uNumWeapons = pxLevelBP->EnumerateParameters( "Weapon", 2, xWeaponEnumerator );

			if ( uNumWeapons < uAVP_MAX_WEAPONS_IN_LEVEL )
			{
				Asura_Value xVal;
				if ( pxLevelBP->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_HACKINGTOOL, xVal ) && xVal.GetHashIDValue() != ASURA_HASH_ID_UNSET )
				{
					auWeaponHashesInLevel[ uNumWeapons ] = xVal.GetHashIDValue();
					++uNumWeapons;
				}
			}

			for ( u_int uWeaponItr = 0; uWeaponItr < uNumWeapons; ++uWeaponItr )
			{
				u_int uWeaponIndex = AVP_WeaponData_System::GetWeaponIndexFromHash( auWeaponHashesInLevel[ uWeaponItr ], eCharType );
				AVP_WeaponData_System::SetShouldLoadWeapon( uWeaponIndex, eCharType );
			}
		}
	}
	else
	{
		for (int iSpeciesIndex = 0; iSpeciesIndex < AVP_Settings::AVP_NUM_SPECIES; iSpeciesIndex++)
		{
			AVP_Game_Core::CharacterType eCharType = AVP_Game_Core::CHARACTERTYPE_UNKNOWN;
			const AVP_Settings::Species	eSpecies	= static_cast<AVP_Settings::Species>(iSpeciesIndex);
			switch(eSpecies)
			{
				case AVP_Settings::AVP_SPECIES_ALIEN: eCharType = AVP_Game_Core::MULTIPLAYER_ALIEN; break;
				case AVP_Settings::AVP_SPECIES_HUMAN: eCharType = AVP_Game_Core::MULTIPLAYER_MARINE; break;
				case AVP_Settings::AVP_SPECIES_PREDATOR: eCharType = AVP_Game_Core::MULTIPLAYER_PREDATOR; break;
				default: Asura_Assert( false, "Invalid character type" ); return;
			}
			
			const Asura_Blueprint* pxLevelBP = AVP_Game_Core::GetLevelBlueprint( szLevelFile, eCharType );
			if ( pxLevelBP )
			{
				Asura_Hash_ID auWeaponHashesInLevel[ uAVP_MAX_WEAPONS_IN_LEVEL ];
				AVP_Blueprint_ParameterEnumerator_LevelWeapons xWeaponEnumerator( auWeaponHashesInLevel );
				u_int uNumWeapons = pxLevelBP->EnumerateParameters( "Weapon", 2, xWeaponEnumerator );

				if ( uNumWeapons < uAVP_MAX_WEAPONS_IN_LEVEL )
				{
					Asura_Value xVal;
					if ( pxLevelBP->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_HACKINGTOOL, xVal ) && xVal.GetHashIDValue() != ASURA_HASH_ID_UNSET )
					{
						auWeaponHashesInLevel[ uNumWeapons ] = xVal.GetHashIDValue();
						++uNumWeapons;
					}
				}

				for ( u_int uWeaponItr = 0; uWeaponItr < uNumWeapons; ++uWeaponItr )
				{
					u_int uWeaponIndex = AVP_WeaponData_System::GetWeaponIndexFromHash( auWeaponHashesInLevel[ uWeaponItr ], eCharType );
					AVP_WeaponData_System::SetShouldLoadWeapon( uWeaponIndex, eCharType );
				}
			}
		}

		// Load all pickup weapons
		Asura_Blueprint_ConstIterator xPickupBPItr( uAVP_BLUEPRINT_TYPE_PICKUPS );
		while ( !xPickupBPItr.Done() && xPickupBPItr.GetCurrent() )
		{
			AVP_Settings::Species eSpecies = static_cast < AVP_Settings::Species >( xPickupBPItr.GetCurrent()->GetParameterAsInt( 0x843f8d15, AVP_Settings::AVP_SPECIES_UNKNOWN ) );

			AVP_Game_Core::CharacterType eCharType = AVP_Game_Core::CHARACTERTYPE_UNKNOWN;
			switch ( eSpecies )
			{
				case AVP_Settings::AVP_SPECIES_ALIEN: { eCharType = AVP_Game_Core::MULTIPLAYER_ALIEN; } break;
				case AVP_Settings::AVP_SPECIES_HUMAN: { eCharType = AVP_Game_Core::MULTIPLAYER_MARINE; } break;
				case AVP_Settings::AVP_SPECIES_PREDATOR: { eCharType = AVP_Game_Core::MULTIPLAYER_PREDATOR; } break;
#ifdef ASURA_TOOLS
				default: { Asura_Assertf( false, ( "Invalid charater type in pickup blueprint %s", xPickupBPItr.GetCurrent()->GetName() ) ); } break;
#else
				default: { Asura_Assertf( false, ( "Invalid charater type in pickup blueprint %d", xPickupBPItr.GetCurrent()->GetHashID() ) ); } break;
#endif
			}

			Asura_Hash_ID uWeaponItemHash = xPickupBPItr.GetCurrent()->GetParameterAsHash( uAVP_BLUEPRINT_PARAM_PICKUP_MULTIPLAYERINVENTORYITEMBLUEPRINT, ASURA_HASH_ID_UNSET );
			if ( uWeaponItemHash != ASURA_HASH_ID_UNSET )
			{
				u_int uWeaponIndex = AVP_WeaponData_System::GetWeaponIndexFromHash( uWeaponItemHash, eCharType );
				AVP_WeaponData_System::SetShouldLoadWeapon( uWeaponIndex, eCharType );
			}
			xPickupBPItr.Next();
		} // pickup BP loop
	}
}

/// End of file. ///
