///////////////////////////////////////////////////////////////////////////////
//
// File: AVP_ServerEntity_Container.cpp
//
// Created by Martin G.Harris (20/8/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// Project-specific server entity that manages a container instance.
//
///////////////////////////////////////////////////////////////////////////////


#include "Core/Asura.h"
#include "Entity/Container/AVP_ServerEntity_Container.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes

// Asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Container/Asura_ContainerRegistry.h"
#include "Network/Asura_Message.h"

// AvP
#include "Entity/AVP_Messages.h"
#include "Entity/AVP_server_Messages.h"
#include "Entity/Container/AVP_ClientEntity_Container.h"
#include "Entity/Damage/AVP_EntityDamage_Apply.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/Damage/AVP_EntityDamage_Request.h"

// Tools
#ifdef ASURA_TOOLS
#include "Entity/Container/ToolsAVP_Entity_Container.h"
#endif //#ifdef ASURA_TOOLS


// Register the server entity.
Asura_ContainerRegisterServerEntity<AVP_ServerEntity_Container> xRegisterAVPServerEntity;


// Constants
const int iAVP_CONTAINER_CHUNK_VERSION = 1001;

///////////////////////////////////////////////////////////////////////////////
// AVP_ServerEntity_Container
///////////////////////////////////////////////////////////////////////////////
AVP_ServerEntity_Container::AVP_ServerEntity_Container(Asura_Guid uGuid)
: Asura_ServerEntity_Container(uGuid)
, m_uProjectFlags(0)
, m_fLightingTimer(0.0f)
, m_xOldOrientation(Asura_Maths::IdentityQuat)
, m_xOldPosition(Asura_Maths::ZeroVector3)
{
	m_xPrevLightingBB.Reset();
	InvalidateLightingBB();
	m_bIgnoreForSurfaceReorientation = true;
}

AVP_ServerEntity_Container::~AVP_ServerEntity_Container()
{
}


/////////////////////////
// ReadFromChunkStream //
/////////////////////////
// Reads this entity from the specified chunk-stream.
void AVP_ServerEntity_Container::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	// Peek version
	if ( xStream.PeekInt() < 1000 )
	{
		// Old version without a project-specific entity.
		PARENT::ReadFromChunkStream( xStream );
		return;
	}
	
	// New project-specfic entity.
	int iVersion;
	xStream >> iVersion;
	if (iVersion < 1000 || iVersion > iAVP_CONTAINER_CHUNK_VERSION)
	{
		Asura_Assertf(0, ("Unknown chunk version %d for container entity", iVersion));
		return;
	}

	PARENT::ReadFromChunkStream(xStream);

	// Put any project-specific variables here...
	if (iVersion >= 1001)
	{
		// Read in the project flags, but preserve some of them.
		const bool bInitialisedLightingBB = m_bInitialisedLightingBB;
		const bool bHasLights			  = m_bHasLights;
		const bool bPreviousHasLights	  = m_bPreviousHasLights;
		
		Asura_ReadFromStream_Bitfield_Wrapper<u_int> xFlagsBitfield(m_uProjectFlags);
		xStream >> xFlagsBitfield;
		
		m_bInitialisedLightingBB = bInitialisedLightingBB;
		m_bHasLights			 = bHasLights;
		m_bPreviousHasLights	 = bPreviousHasLights;
	}
}

////////////////////////
// WriteToChunkStream //
////////////////////////
// Writes this entity to the specified chunk-stream.
void AVP_ServerEntity_Container::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << iAVP_CONTAINER_CHUNK_VERSION;
	
	PARENT::WriteToChunkStream(xStream);

	// Put any project-specific variables here...
	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xFlagsBitfield(m_uProjectFlags);
	xFlagsBitfield.Add(m_bIgnoreForSurfaceReorientation, 1);
	xFlagsBitfield.Add(m_bInitialisedLightingBB, 1);
	xFlagsBitfield.Add(m_bHasLights, 1);
	xFlagsBitfield.Add(m_bPreviousHasLights, 1);

	xStream << xFlagsBitfield;
}

///////////////////
// HandleMessage //
///////////////////
bool AVP_ServerEntity_Container::HandleMessage(const Asura_Message* pxMessage)
{
	switch (pxMessage->GetMsgID())
	{
		case AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM:
		{
			const AVP_Message_EntityDamage_RequestVictim* pxData = static_cast<const AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
			AVP_EntityDamage_RequestVictim xDamage;
			
			pxData->GetDamage(xDamage);
			xDamage.Apply();
			return true;
		}

		// We want default handling of these messages here, as only actors will do anything with these but more than that can be a target.
		case AVP_MSGID_INCOMING_MELEE_STARTED:
		case AVP_MSGID_INCOMING_MELEE_FINISHED:
			return true;

		default:
			break;
	}

	return PARENT::HandleMessage(pxMessage);
}

//////////////////////////////////
// FillProjectCreateMessageData //
//////////////////////////////////
// Fill in the specified number of bytes of project-specific data.
void AVP_ServerEntity_Container::FillProjectCreateMessageData(void* pData) const
{
	char* pcData = reinterpret_cast<char*>(pData);
	*pcData = static_cast<char>((m_bIgnoreForSurfaceReorientation ? 1 : 0));
}

////////////
// Update //
////////////

DEBUGVAR float dbg_fLightingUpdateTimer = 3.0f;

void AVP_ServerEntity_Container::Update()
{
	PARENT::Update();

	if ( m_bHasLights )
	{
		m_fLightingTimer -= Asura_Timers::GetGameTimeSlice();
		if ( m_fLightingTimer < 0.0f )
		{
			m_fLightingTimer = Asura_Random::GetFloat( 0.5f * dbg_fLightingUpdateTimer, dbg_fLightingUpdateTimer );
			if ( !GetPosition().ApproximatelyEquals( m_xOldPosition ) || !GetOrientationQuat().ApproximatelyEquals( m_xOldOrientation ) )
			{
				m_xOldPosition = GetPosition();
				m_xOldOrientation = GetOrientationQuat();
				InvalidateLightingBB();
				SendLightChangeMessage();
			}
		}
	}
}

////////////////////////
// TakeDamageExtended //
////////////////////////
// Apply damage.
void AVP_ServerEntity_Container::TakeDamageExtended(const AVP_EntityDamage_RequestVictim& xDamage, const bool bForceFatal, const bool bSendMessageToClient )
{
	if (!IsEntityActive()) return;

	// Ideally we should use the actual health value for the fatal case.
	const float fDamageToApply = (bForceFatal ? 100000.0f : xDamage.GetAmount());
	
	// Apply damage to the server instance
	Asura_Container_ApplyDamageOutput xOutput;
	ApplyDamage(xOutput, xDamage.GetAttachmentID(), fDamageToApply);


	if( bSendMessageToClient )
	{
		// Inform client for damage effects.
		const float	fImpulseStrength = xDamage.GetImpulseStrength();

		Asura_Vector_3	xDirection;
		Asura_Vector_3	xNormal;
		Asura_Vector_3	xPointOfImpact;

		xDamage.GetDirection(xDirection);
		xDamage.GetNormal(xNormal);
		xDamage.GetPointOfImpact(xPointOfImpact);
		xDirection.Normalise();

		const Asura_Hash_ID uTypeHashType	= xDamage.GetTypeHash();
		const Asura_Guid uVictimGuid		= GetGuid();

		const u_int uDestroyedSkinElements = 0;
		const u_int uGibbedSkinElements = 0;
		
		// Send the damage message and use this to handle any events...
		const int iBoneID = xDamage.GetBoneID();
		const Asura_Guid uAttackerGuid = xDamage.GetFromGuid();
		const Asura_Guid uAttackerParentGuid = xDamage.GetParentGuid();
		const float fDamageAmount = (xOutput.m_fOldHealth - xOutput.m_fNewHealth);
		const int iAttachmentID = xDamage.GetAttachmentID();
		const Asura_Hash_ID uMatResponseHash = xDamage.GetMaterialResponseHashID();

		// Note: currently m_fAmount is set to the actual reduction in the health.
		const AVP_EntityDamage_ApplyToEntity xNewDamage(uTypeHashType,
														uAttackerGuid,
														xPointOfImpact,
														xDirection,
														xNormal,
														uVictimGuid,
														xOutput.m_fNewHealth,
														fDamageAmount,
														fImpulseStrength,
														iBoneID,
														uDestroyedSkinElements,
														uGibbedSkinElements,
														uAttackerParentGuid,
														iAttachmentID,
														uMatResponseHash);

		Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_APPLY_TO_ENTITY, sizeof(AVP_Message_EntityDamage_ApplyToEntity));

		if (pxMessage)
		{
			AVP_Message_EntityDamage_ApplyToEntity*	pxData = static_cast<AVP_Message_EntityDamage_ApplyToEntity*>(pxMessage->GetData());

			// Inform the client regardless of whether we can be damaged...
			pxData->SetDamage(xNewDamage);
			Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
		}
	}
}

///////////////////////////
// GetContainerLuminance //
///////////////////////////
// Estimate the luminance at the specified waypoint position.
float AVP_ServerEntity_Container::GetContainerLuminance(const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal) const
{
	if (IsEntityActive() && !m_xInputState.IsOff())
	{
		// Make sure that the lighting BB is initialised.
		InitLightingBB();
		if (!m_bHasLights)
		{
			// Container has no lights currently active.
			return 0.0f;
		}

		if (!m_xLightingBB.Contains(xPosition))
		{
			// Waypoint is outside the range of the lights.
			return 0.0f;
		}

		const Asura_ClientContainerInstance* const pxClientInstance = GetClientInstance();
		Asura_Assertf(pxClientInstance, ("Failed to find container client instance for entity %u", GetGuid()));
		if (pxClientInstance)
		{
			AVP_Container_Functor_GetLuminance xFunctor(xPosition, xNormal);
			pxClientInstance->Traverse(xFunctor);
			return xFunctor.GetLuminance();
		}
	}

	return 0.0f;
}

//////////////////////////
// InvalidateLightingBB //
//////////////////////////
void AVP_ServerEntity_Container::InvalidateLightingBB()
{
	m_xLightingBB.Reset();
	m_bInitialisedLightingBB = false;
	m_bHasLights = false;
	m_bPreviousHasLights = false;
}

/////////////////////
// UpdateLightingBB //
/////////////////////
void AVP_ServerEntity_Container::UpdateLightingBB() const
{
	if (m_bInitialisedLightingBB && !m_bHasLights)
	{
		// Previous state didn't have lights, so we only need the BB for the new state;
		// there is no need to do a union of new and old states.
		m_xLightingBB.Reset();
	}

	m_bInitialisedLightingBB = true;
	m_bPreviousHasLights = m_bHasLights;
	m_bHasLights = false;

	const Asura_ClientContainerInstance* const pxClientInstance = GetClientInstance();
	if (pxClientInstance)
	{
		AVP_Container_Functor_SetLightingBB xFunctor(m_xLightingBB);
		pxClientInstance->Traverse(xFunctor);
		m_bHasLights = xFunctor.HasLights();
	}
}

////////////////////
// InitLightingBB //
////////////////////
void AVP_ServerEntity_Container::InitLightingBB() const
{
	// Initialise the lighting BB for the first time.
	if (!m_bInitialisedLightingBB)
	{
		UpdateLightingBB();
	}
}

/////////////////////////////
// OnEntityActivateMessage //
/////////////////////////////
void AVP_ServerEntity_Container::OnEntityActivateMessage()
{
	PARENT::OnEntityActivateMessage();

	if (!m_bInitialisedLightingBB)
	{
		// Not initialised yet, so initialise it.
		UpdateLightingBB();

		if (m_bHasLights)
		{
			// Activated a container with lights, need to recalculate the waypoint luminance.
			SendLightChangeMessage();
		}
	}
	else if (m_bPreviousHasLights)
	{
		// Previously had lights when deactivated. Reuse the BB which was calculated when it was active.
		m_bHasLights = true;
		m_bPreviousHasLights = false;
		
		SendLightChangeMessage();
	}
	else
	{
		// Since it had no lights when it was deactivated, it shouldn't have any when activated.
		m_bHasLights = false;
	}
}

///////////////////////////////
// OnEntityDeactivateMessage //
///////////////////////////////
void AVP_ServerEntity_Container::OnEntityDeactivateMessage()
{
	// Make sure that the lighting BB has already been initialised.
	InitLightingBB();

	PARENT::OnEntityDeactivateMessage();

	if (m_bHasLights)
	{
		// Light is switching off, send light change message.
		// Preserve the BB, so that we can use it upon reactivating.
		SendLightChangeMessage();
		m_bHasLights = false;
		m_bPreviousHasLights = true;
	}
	else
	{
		m_bPreviousHasLights = false;
	}
}


////////////////////
// PreStateChange //
////////////////////
// Called on state change caused by receiving a message, or due to damage.
void AVP_ServerEntity_Container::PreStateChange()
{
	// First time that state changes, make sure that we have initialised the
	// lighting BB, before we do the change.
	InitLightingBB();
}

/////////////////////
// PostStateChange //
/////////////////////
void AVP_ServerEntity_Container::PostStateChange()
{
	Asura_Assert(m_bInitialisedLightingBB, "LightingBB wasn't initialised before state change!");
	UpdateLightingBB();

	// If either the previous state or the new state has lights then we need to recalculate the
	// waypoint lighting.
	if (m_bPreviousHasLights || m_bHasLights)
	{
		SendLightChangeMessage();
	}
}

///////////////////
// SetLightingBB //
///////////////////
// Store a BB for the volume affected by the change in lighting (photons and projectors).
// This is roughly a union of the BB for the current lighting and that for the previous state's lighting.
void AVP_ServerEntity_Container::SetLightingBB( Asura_Bounding_Box& xBB ) const
{
	// Make sure that the lighting BB has been initialised.
	InitLightingBB();

	// Return the cached value.
	xBB = m_xLightingBB;
}

////////////////////////////
// SendLightChangeMessage //
////////////////////////////
void AVP_ServerEntity_Container::SendLightChangeMessage() const
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_LIGHT_CHANGE_EVENT_CODE, sizeof( AVP_Message_LightChange ) );
	if ( pxMessage )
	{
		AVP_Message_LightChange* pxData = static_cast< AVP_Message_LightChange* >( pxMessage->GetData() );
		if ( pxData )
		{
			pxData->SetOldBB( m_xPrevLightingBB );
			SetLightingBB( m_xPrevLightingBB );
			pxData->SetNewBB( m_xPrevLightingBB );

			Send( pxMessage, ASURA_GUID_SERVER );
		}
	}
}

///////////////////////////
// FindBlueprintFromGuid //
///////////////////////////
// Find the blueprint in the root collection of the entity with the specified Guid.
// If uBlueprintTypeHash != ASURA_HASH_ID_UNSET, then only return blueprints of the specified type.
// Returns NULL if no relevant blueprint found.
const Asura_Container_Blueprint* AVP_ServerEntity_Container::FindBlueprintFromGuid(Asura_Guid uGuid, Asura_Hash_ID uBlueprintTypeHash)
{
	// Note: Currently we don't care whether the blueprints are enabled or not (provided that there is only a single one).
	const bool bIncludeDisabled = true;

	// Check for server entities.
	const AVP_ServerEntity_Container* pxServer = AVP_ServerEntity_Container::FindEntity(uGuid);
	if (pxServer)
	{
		return pxServer->FindBlueprint(uBlueprintTypeHash, bIncludeDisabled);
	}

	// Check for client entities.
	const AVP_ClientEntity_Container* pxClient = AVP_ClientEntity_Container::FindEntity(uGuid);
	if (pxClient)
	{
		return pxClient->FindBlueprint(uBlueprintTypeHash, bIncludeDisabled);
	}

	return 0;
}

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM
