//--------------------------------------------------------------------------------------
//	File	:	AVP_ServerNode_Objective_Marine.cpp
//	Author	:	Gareth Thomas
//	Created	:	23 / 01 / 2008
//	Copyright (C) 2006 Rebellion, All Rights Reserved.
//--------------------------------------------------------------------------------------

// Core
#include "Core/Asura.h"

// Header
#include "Entity/Locations/Objective/Marine/AVP_ServerEntity_Objective_Marine.h"

// Asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif
#include "Network/Asura_Message.h"

// AVP
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Entity_EventIDs.h"
#include "Entity/AVP_Messages.h"

//------------------------------------------------------------------------------
// Constants, Defines, Statics and Globals
//------------------------------------------------------------------------------
Asura_Hashed_List<AVP_ServerNode_Objective_Marine> AVP_ServerNode_Objective_Marine::s_xList;

// const char* AVP_ServerNode_Objective_Marine::s_aszObjectiveTypeNames[MAX_OBJECTIVE_TYPES] =
// {
// 	"Move to Goal at slow speed, stop to engage threats.",
// 	"Move to Goal at medium speed, do not stop to engage threats.",
// 	"Move to Goal at medium speed, stop to engage threats.",
// 	"Move to Goal at slow speed, stop to engage threats, scan area en route.",
// };

const char* AVP_ServerNode_Objective_Marine::s_aszObjectiveMoveModeNames[ MAX_OBJECTIVE_MOVEMENT_MODES_TYPES ] = 
{
	"MOVE SLOW - STOP TO ENGAGE",
	"MOVE SLOW - DONT STOP TO ENGAGE",
	"MOVE MEDIUM - STOP TO ENGAGE",
	"MOVE MEDIUM  - DONT STOP TO ENGAGE",
	"ADVANCE SLOW - ( Not In Use Yet - Do Not Use )",
	"ADVANCE MEDIUM - ( Not In Use Yet - Do Not Use )",
	"RETREAT SLOW - ( Not In Use Yet - Do Not Use )",
	"RETREAT MEDIUM - ( Not In Use Yet - Do Not Use )",
};

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
AVP_ServerNode_Objective_Marine::AVP_ServerNode_Objective_Marine(const Asura_Guid uGuid)
: PARENT(uGuid)
, m_xPosition(Asura_Maths::ZeroVector3)
, m_bUseSpecificGameSceneNodes(0)
, m_bObjectiveCompleteOnMessage(0)
, m_bObjectiveHasReceivedCompleteMessage(false)
, m_bRunToAction(false)
, m_pxRequiredGameScenes(NULL)
, m_pxOptionalGameScenes(NULL)
, m_uNumRequiredGameScenes(0)
, m_uNumOptionalGameScenes(0)
{
	Classification = AVP_ENTITYCLASS_AI_MARINE_OBJECTIVE;

	SetNeverUpdateClient();
	SetNeverUpdates(true);

	s_xList.AddToList(this);
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
AVP_ServerNode_Objective_Marine::~AVP_ServerNode_Objective_Marine(void)
{
	if ( m_pxRequiredGameScenes )
	{
		delete [] m_pxRequiredGameScenes;
	}

	if ( m_pxOptionalGameScenes )
	{
		delete [] m_pxOptionalGameScenes;
	}

	s_xList.RemoveFromList(this);
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::PostLoadInitialise(void)
{
	PARENT::PostLoadInitialise();

	if ( GetNumberOfVolumes() == 0 )
	{
		Asura_Assertf(false, ( "Objective / AI Goal Node with Guid %u has NO VOLUME!", GetGuid() ) );
	}
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
bool AVP_ServerNode_Objective_Marine::HandleMessage(const Asura_Message* pxMessage)
{
	switch (pxMessage->GetMsgID())
	{
		case AVP_MSGID_AI_OBJECTIVE_COMPLETE:
		{
			m_bObjectiveHasReceivedCompleteMessage = true;
			return true;
		}

		default:
		{
			return PARENT::HandleMessage(pxMessage);
		}
	}

}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::Reset(void)
{
	m_bObjectiveHasReceivedCompleteMessage = false;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	//Version
	xStream << CURRENT_CHUNK_VERSION;

	//Parent
	PARENT::WriteToChunkStream(xStream);

	xStream << m_xPosition;

	xStream << m_bUseSpecificGameSceneNodes;
	xStream << m_uNumRequiredGameScenes;

	for (u_int u = 0; u < m_uNumRequiredGameScenes; ++u)
	{
		xStream << m_pxRequiredGameScenes[u];
	}

	xStream << m_uNumOptionalGameScenes;

	for (u_int u = 0; u < m_uNumOptionalGameScenes; ++u)
	{
		xStream << m_pxOptionalGameScenes[u];
	}

	xStream << m_bObjectiveCompleteOnMessage;

	xStream << m_bRunToAction;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	//Version
	u_int uVersion;
	xStream	>> uVersion;

	//Parent
	PARENT::ReadFromChunkStream(xStream);

	if (uVersion == 0)
	{
		m_iNumVolumes = 0;
		m_uGamePlayVolumeGuid = ASURA_GUID_UNREGISTERED;
	}
	
	if (uVersion == 1)
	{
		xStream >> m_xPosition;
		xStream >> m_uGamePlayVolumeGuid;
		xStream >> m_bUseSpecificGamePlayVolume;
		xStream >> m_iNumVolumes;
		m_pxVolumeGroup->ReadFromChunkStream(xStream);
	}

	if (uVersion == 2)
	{
		xStream >> m_xPosition;
		xStream >> m_uGamePlayVolumeGuid;
		xStream >> m_bUseSpecificGamePlayVolume;
		xStream >> m_iNumVolumes;
		m_pxVolumeGroup->ReadFromChunkStream(xStream);

		xStream >> m_bUseSpecificGameSceneNodes;

		xStream >> m_uNumRequiredGameScenes;
		m_pxRequiredGameScenes = new Asura_Guid[m_uNumRequiredGameScenes];

		for (u_int u = 0; u < m_uNumRequiredGameScenes; ++u)
		{
			xStream >> m_pxRequiredGameScenes[u];
		}

		xStream >> m_uNumOptionalGameScenes;
		m_pxOptionalGameScenes = new Asura_Guid[m_uNumOptionalGameScenes];

		for (u_int u = 0; u < m_uNumOptionalGameScenes; ++u)
		{
			xStream >> m_pxOptionalGameScenes[u];
		}
	}

	if (uVersion == 3)
	{
		xStream >> m_xPosition;
		xStream >> m_uGamePlayVolumeGuid;
		xStream >> m_bUseSpecificGamePlayVolume;
		xStream >> m_iNumVolumes;
		m_pxVolumeGroup->ReadFromChunkStream(xStream);

		xStream >> m_bUseSpecificGameSceneNodes;

		xStream >> m_uNumRequiredGameScenes;
		m_pxRequiredGameScenes = new Asura_Guid[m_uNumRequiredGameScenes];

		for (u_int u = 0; u < m_uNumRequiredGameScenes; ++u)
		{
			xStream >> m_pxRequiredGameScenes[u];
		}

		xStream >> m_uNumOptionalGameScenes;
		m_pxOptionalGameScenes = new Asura_Guid[m_uNumOptionalGameScenes];

		for (u_int u = 0; u < m_uNumOptionalGameScenes; ++u)
		{
			xStream >> m_pxOptionalGameScenes[u];
		}

		xStream >> m_bObjectiveCompleteOnMessage;

		xStream >> m_bRunToAction;
	}

	if (uVersion == CURRENT_CHUNK_VERSION)
	{
		xStream >> m_xPosition;
		xStream >> m_bUseSpecificGameSceneNodes;

		xStream >> m_uNumRequiredGameScenes;
		m_pxRequiredGameScenes = new Asura_Guid[m_uNumRequiredGameScenes];

		for (u_int u = 0; u < m_uNumRequiredGameScenes; ++u)
		{
			xStream >> m_pxRequiredGameScenes[u];
		}

		xStream >> m_uNumOptionalGameScenes;
		m_pxOptionalGameScenes = new Asura_Guid[m_uNumOptionalGameScenes];

		for (u_int u = 0; u < m_uNumOptionalGameScenes; ++u)
		{
			xStream >> m_pxOptionalGameScenes[u];
		}

		xStream >> m_bObjectiveCompleteOnMessage;

		xStream >> m_bRunToAction;
	}
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::SetRequiredGuids(Asura_Guid* pxGuids, u_int uNumGuids)
{
	if ( m_pxRequiredGameScenes )
	{
		delete [] m_pxRequiredGameScenes;
	}

	m_pxRequiredGameScenes = new Asura_Guid[ uNumGuids ];

	for ( u_int u = 0; u < uNumGuids; ++u )
	{
		m_pxRequiredGameScenes[u] = pxGuids[u];
	}

	m_uNumRequiredGameScenes = uNumGuids;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::SetOptionalGuids(Asura_Guid* pxGuids, u_int uNumGuids)
{
	if ( m_pxOptionalGameScenes )
	{
		delete [] m_pxOptionalGameScenes;
	}

	m_pxOptionalGameScenes = new Asura_Guid[ uNumGuids ];

	for ( u_int u = 0; u < uNumGuids; ++u )
	{
		m_pxOptionalGameScenes[u] = pxGuids[u];
	}

	m_uNumOptionalGameScenes = uNumGuids;
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::SendObjectiveReachedMessage(void)
{
	SendMessagesForEvent(AVP_ENTITY_EVENT_MARINE_AI_OBJECTIVE_COMPLETE);
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::SendCompleteEvent(void)
{
	SendMessagesForEvent(AVP_ENTITY_EVENT_MARINE_AI_OBJECTIVE_ACTIONS_COMPLETE);
}

//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
AVP_ServerNode_Objective_Marine* AVP_ServerNode_Objective_Marine::Create(const Asura_Guid uGuid)
{
	return(new AVP_ServerNode_Objective_Marine(uGuid));
}

#ifdef ASURA_TOOLS
//--------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------
void AVP_ServerNode_Objective_Marine::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	PARENT::ApplyChangesForTemplateInstance(pxInstance);

	m_xPosition = pxInstance->TransformPosition(m_xPosition);
}
#endif