//////////////////////////////////////////////////////////////////////////////
//
// AVP_ServerEntity_DamageableObject.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     Physical object which can be affected by explosions (initially static)
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/PhysicalObject/PhysicsObject/DamageableObject/AVP_ServerEntity_DamageableObject.h"

// Engine independent.
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/AVP_Maths.h"

// Project
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/PhysicalObject/AVP_ServerEntity_PhysicalObject.h"
#include "Entity/PhysicalObject/PhysicsObject/DamageableObject/AVP_DamageableObject_Messages.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static const u_int	g_uCurrentChunkVersion	 = 3;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////
//
// AVP_ServerEntity_DamageableObject
//

Asura_Hashed_List<AVP_ServerEntity_DamageableObject> AVP_ServerEntity_DamageableObject::s_xList;


AVP_ServerEntity_DamageableObject::AVP_ServerEntity_DamageableObject(const Asura_Guid uGuid)
: Asura_Entity_Server(uGuid)
, m_bUsingDamagedSkin(false)
, m_fDamageThresholdSkin(5.0f)
, m_fDamageThresholdDynamic(20.0f)
, m_bBeginsStatic(false)
{
	Classification = AVP_ENTITYCLASS_DAMAGEABLEOBJECT;
	s_xList.AddToList(this);

	for (u_int i = 0; i < 6; i++)
	{
		m_auDamagedSkinFileHashID[i] = ASURA_HASH_ID_UNSET;
		m_auDamagedSkinHashID[i] = ASURA_HASH_ID_UNSET;
		m_auDamagedAnimFileHashID[i] = ASURA_HASH_ID_UNSET;
		m_auDamagedAnimHashID[i] = ASURA_HASH_ID_UNSET;
	}

	SetEntityActiveFromLoad( true );
}

AVP_ServerEntity_DamageableObject::~AVP_ServerEntity_DamageableObject(void)
{
	s_xList.RemoveFromList(this);
}

void AVP_ServerEntity_DamageableObject::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// Write the version..
	xStream << g_uCurrentChunkVersion;

	// Write the parent data
	PARENT::WriteToChunkStream( xStream );

	// Write the begins static flag
	xStream << m_bBeginsStatic;

	// Write the damage threshold for the skin to change
	xStream << m_fDamageThresholdSkin;

	// Write the damage threshold for changing to a dynamic object
	xStream << m_fDamageThresholdDynamic;

	// Write the skins and anims
	for (int i = 0; i < 6; i++)
	{
		xStream << m_auDamagedSkinFileHashID[i];
		xStream << m_auDamagedSkinHashID[i];
		xStream << m_auDamagedAnimFileHashID[i];
		xStream << m_auDamagedAnimHashID[i];
	}
}

void AVP_ServerEntity_DamageableObject::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	// Get the version..
	xStream >> uVersion;

	// Check the version
	if (uVersion > g_uCurrentChunkVersion)
	{
		Asura_Assertf(false, ("Unexpected chunk-version for destructible - aborting read."));
		return;
	}

	// Read the parent data
	PARENT::ReadFromChunkStream(xStream);

	// Read the 'begins static' flag
	if (uVersion > 1)
	{
		xStream >> m_bBeginsStatic;
	}

	// Read the damage threshold for changing to a damaged skin
	if (uVersion > 0)
	{
		xStream >> m_fDamageThresholdSkin;
	}

	// Read the damage threshold for changing to a dynamic object
	if (uVersion > 2)
	{
		xStream >> m_fDamageThresholdDynamic;
	}
	else
	{
		m_fDamageThresholdDynamic = m_fDamageThresholdSkin;
	}

	// Read the return to static flag (no longer used)
	if (uVersion < 2)
	{
		bool bReturnToStatic;
		xStream >> bReturnToStatic;
	}

	// Read the skins and anims
	if (uVersion > 0)
	{
		for (int i = 0; i < 6; i++)
		{
			xStream >> m_auDamagedSkinFileHashID[i];
			xStream >> m_auDamagedSkinHashID[i];
			xStream >> m_auDamagedAnimFileHashID[i];
			xStream >> m_auDamagedAnimHashID[i];
		}
	}

	// Change type to static or dynamic
	if (m_bBeginsStatic)
	{
		ChangeType(ASURA_PHYSICS_OBJECT_STATIC);
	}
	else
	{
		ChangeType(ASURA_PHYSICS_OBJECT_DYNAMIC);
	}
}

void AVP_ServerEntity_DamageableObject::TakeDamageExtended(const AVP_EntityDamage_RequestVictim& xDamage,
												   const bool bForceFatal, const bool bSendMessageToClient)
{
	PARENT::TakeDamageExtended(xDamage, bForceFatal, bSendMessageToClient);

	bool bSendUpdate = false;

	// if not using damaged skin and damage above damage skin threshold, change to the appropriate damaged skin
	if (!m_bUsingDamagedSkin && m_fDamageThresholdSkin >= 0.0f && xDamage.GetAmount() >= m_fDamageThresholdSkin)
	{
		int iMaxIndex = 0;
		Asura_Vector_3	xPOI;
		Asura_Vector_3	xNormal;

		xDamage.GetPointOfImpact(xPOI);
		xDamage.GetNormal(xNormal);

		// Method using damage position rather than normals - ifdef'd out as it may be required at a later stage..
#if 0
		Asura_Vector_3 xExtents = m_pxDynamicObject->GetBoundingBox().CalculateExtents();
		Asura_Vector_3 xCentre = m_pxDynamicObject->GetBoundingBox().CalculateCentre();
		Asura_Matrix_3x3 xOrientation = m_pxDynamicObject->GetOrientation();

		Asura_Vector_3 xFromCentre(xPoint - xCentre);

		float fMaxResult = 0.0f;
		float fResult;
		Asura_Vector_3 xComponentOrientated;

		xOrientation.Transpose();

		for (int i = 0; i < 6; i++)
		{
			OrientComponent(xCentre, xExtents, xOrientation, i, xComponentOrientated);
			float fMagnitude = xComponentOrientated.Normalise();
			xComponentOrientated *= 1.0f/fMagnitude;
			//xComponentOrientated *= 0.5f;
			fResult = xFromCentre * xComponentOrientated;
			if (fResult > fMaxResult)
			{
				fMaxResult = fResult;
				iMaxIndex = i;
			}
		}
#else
		// Transpose the normal from world space back into object space
		// and look for closet +- axis
		const Asura_Matrix_3x3&	xOrientation(m_pxDynamicObject->GetOrientation());
		const Asura_Vector_3	xNormalInObjectSpace(xOrientation.ApplyTranspose(xNormal));
		const float				fNormalDotXAxis	= xNormalInObjectSpace[0];
		const float				fThreshold	= 0.707f;	// cos(45)

		if(fNormalDotXAxis < -fThreshold)
		{
			iMaxIndex	= FACE_LEFT;
		}
		else if(fNormalDotXAxis > fThreshold)
		{
			iMaxIndex	= FACE_RIGHT;
		}
		else
		{
			const float	fNormalDotYAxis	= xNormalInObjectSpace[1];

			if(fNormalDotYAxis < -fThreshold)
			{
				iMaxIndex	= FACE_TOP;
			}
			else if(fNormalDotYAxis > fThreshold)
			{
				iMaxIndex	= FACE_BOTTOM;
			}
			else
			{
				const float	fNormalDotZAxis	= xNormalInObjectSpace[2];

				if(fNormalDotZAxis < -fThreshold)
				{
					iMaxIndex	= FACE_FRONT;
				}
				else if(fNormalDotZAxis > fThreshold)
				{
					iMaxIndex	= FACE_BACK;
				}
			}
		}
#endif

		SetSkinAndAnimFileID(m_auDamagedSkinFileHashID[iMaxIndex], m_auDamagedAnimFileHashID[iMaxIndex]);
		SetSkinAndAnimID(m_auDamagedSkinHashID[iMaxIndex], m_auDamagedAnimHashID[iMaxIndex]);

		m_bUsingDamagedSkin = true;								// now damaged (don't damage again)

		bSendUpdate = true;
	}
	if (m_pxDynamicObject->GetType() == ASURA_PHYSICS_OBJECT_STATIC && m_fDamageThresholdDynamic >= 0.0f && xDamage.GetAmount() >= m_fDamageThresholdDynamic)
	{
		m_pxDynamicObject->ChangeType(ASURA_PHYSICS_OBJECT_DYNAMIC);
		SetAlwaysUpdateClient(true);

		bSendUpdate = true;
	}

	if( bSendUpdate )
	{
		MakeClientUpdateMessage();
	}
}

void AVP_ServerEntity_DamageableObject::OrientComponent(const Asura_Vector_3& xCentre, const Asura_Vector_3& xExtents, const Asura_Matrix_3x3& xOrientation, const int iIndex, Asura_Vector_3& xResult)
{
	// iIndex: 0  1  2  3  4  5  --->  +X  +Y  +Z  -X  -Y  -Z   i.e. right top back left bottom front
	Asura_Vector_3 xExtentsCopy = xExtents;

	int iAxis = iIndex % 3;
	if (iAxis != 0) xExtentsCopy[0] = 0.0f;
	if (iAxis != 1) xExtentsCopy[1] = 0.0f;
	if (iAxis != 2) xExtentsCopy[2] = 0.0f;

	if (iIndex > 2)
	{
		xExtentsCopy[iAxis] = -xExtentsCopy[iAxis];
	}
	xResult = xExtentsCopy * xOrientation;
}

void AVP_ServerEntity_DamageableObject::MakeClientUpdateMessage( void )
{
	Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_CLIENTUPDATE, sizeof(AVP_Message_DamageableObject_ClientUpdate), GetClassification() );
	if( !pxMsg ) { return; }

	AVP_Message_DamageableObject_ClientUpdate* pxMsgData = static_cast<AVP_Message_DamageableObject_ClientUpdate*>( pxMsg->GetData() );
	pxMsgData->FillDamageable( this );

	Send( pxMsg, ASURA_GUID_MY_CLIENT_ENTITIES );
}

void AVP_ServerEntity_DamageableObject::MakeCreateMessage( Asura_Guid uRequestingClient, bool bSendGuaranteed /*= false*/ )
{
	Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_CLIENTCREATE, sizeof(AVP_Message_DamageableObject_ClientCreate), GetClassification() );
	if( !pxMsg ) { return; }

	AVP_Message_DamageableObject_ClientCreate* pxMsgData = static_cast<AVP_Message_DamageableObject_ClientCreate*>( pxMsg->GetData() );
	pxMsgData->FillDamageable( this );

	pxMsg->SetToGuid( ASURA_GUID_MY_CLIENT_ENTITIES );
	pxMsg->SetFromGuid( GetGuid() );
	if (bSendGuaranteed)
	{
		pxMsg->UseGuaranteedDelivery();
	}
	Asura_Network::SendToSpecificMachine( pxMsg, uRequestingClient );
}

#ifdef ASURA_TOOLS

void AVP_ServerEntity_DamageableObject::CreateFromTools(Asura_Guid uGuid)
{ 
	AVP_ServerEntity_DamageableObject *pxDamageableObject = new AVP_ServerEntity_DamageableObject(uGuid);

	pxDamageableObject->SetupPhysicsObject(ASURA_PHYSICS_OBJECT_DYNAMIC, false);
}

bool AVP_ServerEntity_DamageableObject::CanBeIncludedInTemplate(void) const
{
	return GetClassification() == AVP_ENTITYCLASS_DAMAGEABLEOBJECT;
}

void AVP_ServerEntity_DamageableObject::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	PARENT::ApplyChangesForTemplateInstance( pxInstance );
}

#endif // #ifdef ASURA_TOOLS

void AVP_ServerEntity_DamageableObject::InitialiseFromBlueprint( const Asura_Blueprint& xBlueprint )
{
    PARENT::InitialiseFromBlueprint( xBlueprint );

    Asura_Value xValue;
    if (xBlueprint.GetParameterValueByHashID(Asura_GetHashID("DamageThresholdSkin"), xValue))
    {
        m_fDamageThresholdSkin = xValue.GetFloatValue();
    }
    if (xBlueprint.GetParameterValueByHashID(Asura_GetHashID("DamageThresholdDynamic"), xValue))
    {
        m_fDamageThresholdDynamic = xValue.GetFloatValue();
    }
    if (xBlueprint.GetParameterValueByHashID(Asura_GetHashID("BeginsStatic"), xValue))
    {
        m_bBeginsStatic = xValue.GetBoolValue();
    }

	if (m_bBeginsStatic)
	{
		ChangeType(ASURA_PHYSICS_OBJECT_STATIC);
	}
	else
	{
		ChangeType(ASURA_PHYSICS_OBJECT_DYNAMIC);
	}

    char *s[] = {"+X", "+Y", "+Z", "-X", "-Y", "-Z"};

    for (int i = 0; i < 6; i++)
    {
        Asura_String xSkin = "DamagedSkin";
        xSkin += s[i];
        if (xBlueprint.GetParameterValueByHashID(Asura_GetHashID(xSkin.GetString()), xValue))
        {
            m_auDamagedSkinHashID[i] = static_cast<u_int>(xValue.GetIntValue());
        }
        Asura_String xAnim = "DamagedAnim";
        xAnim += s[i];
        if (xBlueprint.GetParameterValueByHashID(Asura_GetHashID(xAnim.GetString()), xValue))
        {
            m_auDamagedSkinHashID[i] = static_cast<u_int>(xValue.GetIntValue());
        }
    }
}
