//////////////////////////////////////////////////////////////////////////////
//
// AVP_ClientEntity_Vent.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     Project override of asura lift class
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Vent/AVP_ClientEntity_Vent.h"

#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Shapes/Asura_Shapes.h"
#include "Entity/PhysicalObject/Asura_ClientEntity_PhysicalObject.h"

// Project
#include "Attractors/AVP_Attractors_System.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Network/Asura_Message.h"

#include "Entity/Vent/AVP_Vent_Messages.h"

Asura_Hashed_List<AVP_ClientEntity_Vent> AVP_ClientEntity_Vent::s_xList;

AVP_ClientEntity_Vent::AVP_ClientEntity_Vent()
: m_pxExitPoints(NULL)
, m_pxShapeInstance( NULL )
, m_bIsDamaged(false)
, m_bUseExitPointsOnly(true)
{
	Classification = AVP_ENTITYCLASS_VENT;
	AddToLineOfSightList();
	s_xList.AddToList(this);
}

void AVP_ClientEntity_Vent::ProcessClientCreateMessage(const AVP_Message_Vent_ClientCreate* pxData, float fServerTime)
{
	TeleportTo( pxData->GetPosition(), pxData->GetOrientation() );

	m_bIsDamaged = pxData->GetStartDamaged();
	SetEntityActiveFromLoad( pxData->GetStartActive() );
	SetEntityActive( pxData->GetStartActive() );

	m_uUndamagedSkinHash = pxData->GetUndamagedSkinHash();
	m_uDamagedSkinHash = pxData->GetDamagedSkinHash();

	if ( ( !m_bIsDamaged && m_uUndamagedSkinHash != ASURA_HASH_ID_UNSET ) || ( m_bIsDamaged && m_uDamagedSkinHash != ASURA_HASH_ID_UNSET ) )
	{
		m_pxShapeInstance = new Asura_Shape_Instance();
		m_pxShapeInstance->SetNewPosAndOri( GetCurrentPosition(), GetCurrentOrientation() );
		m_pxShapeInstance->SetSkinAndAnimID( m_bIsDamaged ? m_uDamagedSkinHash : m_uUndamagedSkinHash, ASURA_HASH_ID_UNSET, 0, 0.0f );
	}

	m_uNumExitPoints = pxData->GetNumExitPoints();
	if (m_uNumExitPoints > 0)
	{
		m_pxExitPoints = new ExitPoint[m_uNumExitPoints];
		if (m_pxExitPoints)
		{
			for (u_int u=0; u<m_uNumExitPoints; ++u)
			{
				m_pxExitPoints[u].m_xPosition = Asura_AVector3(pxData->GetExitPointPosition(u));
				m_pxExitPoints[u].m_xNormal = Asura_AVector3(pxData->GetExitPointNormal(u));
			}
		}
	}
}

AVP_ClientEntity_Vent::~AVP_ClientEntity_Vent() 
{
	delete m_pxShapeInstance;

	delete[] m_pxExitPoints;

	s_xList.RemoveFromList(this);
}

void AVP_ClientEntity_Vent::Create(const Asura_Message* pxMessage)
{
	Asura_Assert( pxMessage->GetMsgID() == ASURA_MSG_CLIENTCREATE, "" );
	Asura_Assert( pxMessage->GetEntityClassification() == AVP_ENTITYCLASS_VENT, "" );

	if ( pxMessage->GetMsgID() == ASURA_MSG_CLIENTCREATE )
	{
		const AVP_Message_Vent_ClientCreate* pxData = static_cast< const AVP_Message_Vent_ClientCreate* >( pxMessage->GetData() );

		AVP_ClientEntity_Vent* pxNewVent = new AVP_ClientEntity_Vent();
		if (pxNewVent)
		{
			pxNewVent->ProcessClientCreateMessage(pxData, pxMessage->GetSentTime());
		}
	}
}


static const float fVENT_EXIT_NORMAL_SIMILARITY_VALUE = 0.001f;
static const float fCOS45 = 0.707f;

DEBUGVAR Asura_AVector3 dbg_xExitBeginOffset(0.0f, 0.75f, -1.0f);

bool AVP_ClientEntity_Vent::GetBestExitPoint(const Asura_ClientEntity_PhysicalObject& xClient, Asura_AVector3& xExitBeginPos, Asura_AMatrix3x3& xExitBeginOri, Asura_AVector3& xExitTargetPos, Asura_AMatrix3x3& xExitTargetOri) const
{
	const Asura_AVector3 xClientPos(xClient.GetCurrentPosition());

	const Asura_Matrix_3x3& xClientOri = xClient.GetCurrentOrientation();
	const Asura_AVector3 xClientUp(-xClientOri.e12, -xClientOri.e22, -xClientOri.e32);
	const Asura_AVector3 xClientForward(xClientOri.e13, xClientOri.e23, xClientOri.e33);

	float fBestDP = -2.0f; // Make sure it's an impossible value - -1 is valid since we are allowed to flip 180 degrees
	float fBestDrop = -Asura_Maths::LargeFloat; // Deciding factor if multiple exits have the same normal and the normal is roughly perpendicular to the current up vector.
	float fBestForward = -Asura_Maths::LargeFloat; // Deciding factor if multiple exits have the same normal and the normal is roughly the same as or opposite to the current up vector.
	u_int uBestPoint = m_uNumExitPoints;

	for (u_int u=0; u<m_uNumExitPoints; ++u)
	{
		const float fDP = m_pxExitPoints[u].m_xNormal * xClientUp;
		const Asura_AVector3 xOffset = m_pxExitPoints[u].m_xPosition - xClientPos;
		const float fDrop = xOffset * -xClientUp;
		const float fForward = xOffset * xClientForward;

		bool bNewBest = false;
		if (fDP > (fBestDP + fVENT_EXIT_NORMAL_SIMILARITY_VALUE))
		{
			bNewBest = true;
		}
		else if (fDP >= (fBestDP - fVENT_EXIT_NORMAL_SIMILARITY_VALUE))
		{
			// Depending on whether the normal is roughly perpendicular or not we check different values.
			// For consistency, use the current best's value.
			if (Asura_Maths::Abs(fBestDP) < fCOS45)
			{
				// Approaching perpendicular, use the furthest down.
				bNewBest = fDrop > fBestDrop;
			}
			else
			{
				// Otherwise, use the furthest forwards.
				bNewBest = fForward > fBestForward;
			}
		}

		if (bNewBest)
		{
			fBestDP = fDP;
			fBestDrop = fDrop;
			fBestForward = fForward;
			uBestPoint = u;
		}
	}

	if (uBestPoint != m_uNumExitPoints)
	{
		// We need to get a position to jump to alongside this vent, and then the final target using the point we've just found.
		// The position and orientation next to the vent must be based on the orientation we're going to end up at to be as non-jarring as possible.
		xExitTargetPos = m_pxExitPoints[uBestPoint].m_xPosition;
		const Asura_AVector3& xTargetNormal = m_pxExitPoints[uBestPoint].m_xNormal;

		const Asura_AVector3 xVentPos(GetCurrentPosition());
		const Asura_Matrix_3x3& xVentOri = GetCurrentOrientation();
		const Asura_AVector3 xVentForward(xVentOri.e13, xVentOri.e23, xVentOri.e33);

		Asura_AVector3 xTargetForward;
		Asura_AVector3 xBeginDown;
		const float fForwardNormalDP = xVentForward * xTargetNormal;
		if (Asura_Maths::Abs(fForwardNormalDP) < fCOS45)
		{
			// Roughly perpendicular to the vent, so we just want to point outwards
			xTargetForward = xVentForward;
			xBeginDown = -xTargetNormal;
		}
		else
		{
			// Roughly towards or away from the the vent, work out which side of the vent we're going to end up on.
			const Asura_AVector3 xVentDown(xVentOri.e12, xVentOri.e22, xVentOri.e32);
			Asura_AVector3 xOffset = xExitTargetPos - xVentPos;
			Asura_AVector3 xOffsetOnVentPlane = xOffset - (xVentForward * (xVentForward * xOffset));
			xOffsetOnVentPlane.Normalise();
			const float fDownOffsetDP = xOffsetOnVentPlane * xVentDown;
			if (Asura_Maths::Abs(fDownOffsetDP) >= fCOS45)
			{
				// Below or above the vent, use the down vector.
				xTargetForward = (fDownOffsetDP > 0.0f) ? xVentDown : -xVentDown;
				xBeginDown = (fDownOffsetDP > 0.0f) ? xVentDown : -xVentDown;
			}
			else
			{
				// Alongside the vent. Work out which side.
				const Asura_AVector3 xVentRight(xVentOri.e11, xVentOri.e21, xVentOri.e31);
				const float fRightOffsetDP = xOffset * xVentRight;
				xTargetForward = (fRightOffsetDP > 0.0f) ? xVentRight : -xVentRight;
				xBeginDown = (fRightOffsetDP > 0.0f) ? xVentRight : -xVentRight;
			}
			// If we're jumping to the far plane, reverse our begin down vector.
			if (fForwardNormalDP < 0.0f)
			{
				xBeginDown = -xBeginDown;
			}
		}
		// Get the right vector of our target orientation.
		Asura_AVector3 xTargetRight = xTargetForward ^ xTargetNormal;
		xTargetRight.Normalise();
		// The forward may not actually be perpendicular to the normal, but the right now is. Recalculate the forward.
		xTargetForward = xTargetNormal ^ xTargetRight;
		xExitTargetOri.SetColumn0(xTargetRight);
		xExitTargetOri.SetColumn1(-xTargetNormal);
		xExitTargetOri.SetColumn2(xTargetForward);

		// Similar for the begin orientation. Note we always use the vent forward as our begin forward.
		Asura_AVector3 xBeginRight = xBeginDown ^ xVentForward;
		xBeginRight.Normalise();
		xBeginDown = xVentForward ^ xBeginRight;
		xExitBeginOri.SetColumn0(xBeginRight);
		xExitBeginOri.SetColumn1(xBeginDown);
		xExitBeginOri.SetColumn2(xVentForward);

		// Finally, now we have our being orientation, we can work out our begin position.
		xExitBeginPos = xVentPos + (dbg_xExitBeginOffset * xExitBeginOri);

		return true;
	}

	return false;
}

DEBUGVAR float dbg_fEntryPointAttractorBBRange = 1.0f;
DEBUGVAR float dbg_fEntryPointLOSRange = 2.0f;

DEBUGVAR Asura_AVector3 dbg_xEntryBeginOffset(0.0f, 0.75f, -1.25f);
DEBUGVAR float dbg_fEntryTargetOffsetDistance = 1.0f;

bool AVP_ClientEntity_Vent::GetBestEntryPoint(const Asura_ClientEntity_PhysicalObject& xClient, Asura_AVector3& xEntryBeginPos, Asura_AMatrix3x3& xEntryBeginOri, Asura_AVector3& xEntryTargetPos, Asura_AMatrix3x3& xEntryTargetOri) const
{
	// Taking the simple approach that makes a lot of assumptions first.
	// First, work out what the most appropriate start would be.
	const Asura_Vector_3& xClientPos = xClient.GetCurrentPosition();
	const Asura_Matrix_3x3& xClientOri = xClient.GetCurrentOrientation();

	const Asura_Vector_3& xVentPos = GetCurrentPosition();
	const Asura_Matrix_3x3& xVentOri = GetCurrentOrientation();

	const Asura_Vector_3 xClientRight(xClientOri.e11, xClientOri.e21, xClientOri.e31);
	const Asura_Vector_3 xClientDown(xClientOri.e12, xClientOri.e22, xClientOri.e32);
	const Asura_Vector_3 xVentRight(xVentOri.e11, xVentOri.e21, xVentOri.e31);
	const Asura_Vector_3 xVentDown(xVentOri.e12, xVentOri.e22, xVentOri.e32);
	const Asura_Vector_3 xVentForward(xVentOri.e13, xVentOri.e23, xVentOri.e33);

	Asura_Vector_3 xBestAttractorEntryDown;

	const float fForwardDownDP = xVentForward * xClientDown;

	if (Asura_Maths::Abs(fForwardDownDP) < fCOS45)
	{
		// Looking head-on at the vent, try to maintain client down.
		const float fDownDownDP = xVentDown * xClientDown;
		if (fDownDownDP > 0.0f)
		{
			xBestAttractorEntryDown = xVentDown;
		}
		else
		{
			xBestAttractorEntryDown = -xVentDown;
		}
	}
	else
	{
		// Looking across the vent. Try to maintain client right.
		const float fRightRightDP = xVentRight * xClientRight;

		// The vent faces out of the duct, so we want to match against the opposite to what you might expect.
		if (fRightRightDP >= 0.0f)
		{
			xBestAttractorEntryDown = -xVentDown;
		}
		else
		{
			xBestAttractorEntryDown = xVentDown;
		}
	}

	// Entity begin forward is always vent backwards, which by definition is perpendicular to the down vectors above so we can get a right vector from this too.
	const Asura_Vector_3 xEntryForwards = -xVentForward;

	const Asura_Vector_3 xAttractorEntryRight = xBestAttractorEntryDown ^ xEntryForwards;

	// Next, we need to know if we're jumping to an attractor or just to the other side.
	Asura_Matrix_3x3 xAttractorEntryOri(xAttractorEntryRight.x, xBestAttractorEntryDown.x, xEntryForwards.x, xAttractorEntryRight.y, xBestAttractorEntryDown.y, xEntryForwards.y, xAttractorEntryRight.z, xBestAttractorEntryDown.z, xEntryForwards.z);
	EntryPointCallbackData xAttractorData(*this, xClientPos, xClientOri, xAttractorEntryOri);
	Asura_Bounding_Box xBB;
	xBB.SetToPoint(xVentPos);
	xBB.Grow(dbg_fEntryPointAttractorBBRange);
	Asura_Attractor_System::GetInstance()->CollectAttractors(xBB, TestDuctAttractors_Callback, &xAttractorData);

	if (xAttractorData.m_uBestAttractorID != uASURA_ATTRACTOR_ID_NONE)
	{
		// Found a valid attractor. This is everything we need, so set up our values to pass back.
		xEntryTargetOri = Asura_AMatrix3x3(xAttractorData.m_xBestOrientationOnAttractor);
		xEntryTargetPos = Asura_AVector3(xAttractorData.m_xBestPosOnAttractor) + (Asura_AVector3(0.0f, 0.0f, dbg_fEntryTargetOffsetDistance) * xEntryTargetOri);
		xEntryBeginOri = Asura_AMatrix3x3(xAttractorEntryOri);
		xEntryBeginPos = Asura_AVector3(xVentPos) + (dbg_xEntryBeginOffset * xEntryBeginOri);

		return true;
	}

	return false;
}

void AVP_ClientEntity_Vent::Render()
{
	if ( m_pxShapeInstance != NULL )
	{
		m_pxShapeInstance->Render();
	}
}

void AVP_ClientEntity_Vent::Damage()
{
	m_bIsDamaged = true;
	if ( m_pxShapeInstance != NULL )
	{
		if ( m_uDamagedSkinHash != ASURA_HASH_ID_UNSET )
		{
			m_pxShapeInstance->SetSkinAndAnimID( m_uDamagedSkinHash, ASURA_HASH_ID_UNSET, 0, 0.0f );
		}
		else
		{
			delete m_pxShapeInstance;
		}
	}
	else if ( m_uDamagedSkinHash != ASURA_HASH_ID_UNSET )
	{
		m_pxShapeInstance = new Asura_Shape_Instance();
		m_pxShapeInstance->SetNewPosAndOri( GetCurrentPosition(), GetCurrentOrientation() );
		m_pxShapeInstance->SetSkinAndAnimID( m_uDamagedSkinHash, ASURA_HASH_ID_UNSET, 0, 0.0f );
	}
}

DEBUGVAR float dbg_fVentLOSSphereRadius = 0.5f;

bool AVP_ClientEntity_Vent::LOSTest( const class Asura_LOS_InputData* pxInput, class Asura_LOS_OutputData* pxOutput )
{
	if ( m_pxShapeInstance != NULL )
	{
		return m_pxShapeInstance->LOSTest( GetCurrentPosition(), GetCurrentOrientation(), pxInput, pxOutput );
	}
	else
	{
		// Do a sphere test so we can still hit them even without a shape.
		float fDistanceToIntersection;
		if (Asura_Maths::RayIntersectsSphere(pxInput->m_xOrigin, pxInput->m_xDirection, GetCurrentPosition(), dbg_fVentLOSSphereRadius, &fDistanceToIntersection))
		{
			// Make sure this is a new best.
			if (fDistanceToIntersection < pxOutput->m_fDistanceToIntersection)
			{
				pxOutput->m_bHitEnvironment = false;
				pxOutput->m_bStartedInside = fDistanceToIntersection <= 0.0f;
				pxOutput->m_fDistanceToIntersection = fDistanceToIntersection;
				pxOutput->m_iHitAttachmentID = Asura_AnimationAttachment::INVALID_ID;
				pxOutput->m_iHitBoneID = -1;
				pxOutput->m_sMaterialID = ASURA_MATERIAL_NONE;
				pxOutput->m_uEnvironmentFaceIndex = 0;
				pxOutput->m_uHitEntity = Guid;
				pxOutput->m_uMaterialResponseHashID = ASURA_HASH_ID_UNSET;
				pxOutput->m_usPolyFlags = 0;
				pxOutput->m_xNormal = -pxInput->m_xDirection;

				return true;
			}
		}

		return false;
	}
}

void AVP_ClientEntity_Vent::GetLineOfSightAABB(Asura_Bounding_Box& xAABB)
{
	if (m_pxShapeInstance)
	{
		Asura_Animation* pxAnim = m_pxShapeInstance->GetAnimation();
		if (pxAnim)
		{
			pxAnim->GetCoarseBoundingBox( xAABB, GetCurrentPosition(), GetCurrentOrientation() );
		}
		else
		{
			xAABB = m_pxShapeInstance->GetBB();
		}
	}
	else
	{
		// Need an AABB around the sphere.
		xAABB.SetToPoint(GetCurrentPosition());
		xAABB.Grow(dbg_fVentLOSSphereRadius);
	}
}

void AVP_ClientEntity_Vent::PostCreateInitialise()
{
	// Check for nearby duct env hooks.
	Asura_Bounding_Box xBB;
	xBB.SetToPoint(GetCurrentPosition());
	xBB.Grow(dbg_fEntryPointAttractorBBRange);

	Asura_Attractor_System::GetInstance()->CollectAttractors(xBB, TestForNearbyEnvHooks_Callback, this);
}


static const float fDUCT_ENTRY_NORMAL_SIMILARITY_VALUE = 0.001f;
DEBUGVAR float dbg_fMaxDistanceBetweenDuctAndVentForEntry = 3.0f;

void AVP_ClientEntity_Vent::TestDuctAttractor(AVP_ClientEntity_Vent::EntryPointCallbackData& xData, const Asura_Attractor_Compiled& xAttractor) const
{
	if (xAttractor.GetType() != uAVP_ATTRACTOR_TYPE_DUCT) { return; }

	const bool bHasPossibleAttractor = xData.m_uBestAttractorID != uASURA_ATTRACTOR_ID_NONE;

	const Asura_Vector_3 xEntityRight(xData.m_xEntityOrientation.e11, xData.m_xEntityOrientation.e21, xData.m_xEntityOrientation.e31);
	const Asura_Vector_3 xEntityDown(xData.m_xEntityOrientation.e12, xData.m_xEntityOrientation.e22, xData.m_xEntityOrientation.e32);
	const Asura_Vector_3 xEntityForward(xData.m_xEntityOrientation.e13, xData.m_xEntityOrientation.e23, xData.m_xEntityOrientation.e33);

	Asura_Vector_3 xNormal;
	xAttractor.GetNormal(xNormal);

	const float fNormalDP = xNormal * xEntityDown;

	if (bHasPossibleAttractor)
	{
		// See if our normal is better. Want to maintain the character's down vector as priority, so need a normal as opposed as possible.
		const Asura_Vector_3 xPreviousNormal(-xData.m_xBestOrientationOnAttractor.e12, -xData.m_xBestOrientationOnAttractor.e22, -xData.m_xBestOrientationOnAttractor.e32);
		const float fPreviousNormalDP = xPreviousNormal * xEntityDown;

		if (fNormalDP > (fPreviousNormalDP + fDUCT_ENTRY_NORMAL_SIMILARITY_VALUE)) { return; }
	}

	const Asura_Vector_3& xVentPos = GetCurrentPosition();
	const Asura_Matrix_3x3& xVentOri = GetCurrentOrientation();
	const Asura_Vector_3 xVentBackwards(-xVentOri.e13, -xVentOri.e23, -xVentOri.e33);

	const Asura_Vector_3& xNode0Pos = xAttractor.GetNode(0)->GetPosition();
	const Asura_Vector_3& xNode1Pos = xAttractor.GetNode(1)->GetPosition();
	const Asura_Vector_3 xAttractorVector = xNode1Pos - xNode0Pos;

	float fIntersectionDist;
	if (!Asura_Maths::RayIntersectsCylinder(xVentPos, xVentBackwards, xNode0Pos, xAttractorVector, dbg_fEntryPointLOSRange, &fIntersectionDist)) { return; }

	// We need to get the position on the attractor.
	Asura_Vector_3 xTargetPos = xVentPos + (xVentBackwards * fIntersectionDist);
	Asura_Vector_3 xAttractorDir = xAttractorVector;
	xAttractorDir.Normalise();
	xTargetPos -= xNode0Pos;
	xTargetPos = xAttractorDir * (xTargetPos * xAttractorDir);
	xTargetPos += xNode0Pos;

	// We don't actually care about the size of the least offset, we care about the offset between the found location and the vent position.
	Asura_Vector_3 xTargetOffset = xTargetPos - xVentPos;

	if (xTargetOffset.MagnitudeSquared() > (dbg_fMaxDistanceBetweenDuctAndVentForEntry * dbg_fMaxDistanceBetweenDuctAndVentForEntry)) { return; }

	// Work out what the orientation on this attractor would be.
	Asura_Vector_3 xTargetRight;
	const Asura_Vector_3 xEntryBeginRight(xData.m_xEntryBeginOrientation.e11, xData.m_xEntryBeginOrientation.e21, xData.m_xEntryBeginOrientation.e31);
	if (Asura_Maths::Abs(xEntryBeginRight * xNormal) > 0.999f)
	{
		// Normal roughly matches our right direction, shouldn't happen but for safety we'll deal with it by using down instead.
		xTargetRight = Asura_Vector_3(xData.m_xEntryBeginOrientation.e12, xData.m_xEntryBeginOrientation.e22, xData.m_xEntryBeginOrientation.e32);
	}
	else
	{
		xTargetRight = xEntryBeginRight;
	}

	Asura_Vector_3 xTargetForward = xNormal ^ xTargetRight;
	xTargetForward.Normalise();
	// Recalculate right to guarantee perpendicular.
	xTargetRight = xTargetForward ^ xNormal;

	// Make sure we're ending up on the correct side of the vent.
	if ((xTargetOffset * xTargetForward) < -0.001f) { return; }

	xData.m_uBestAttractorID = xAttractor.GetID();
	xData.m_xBestPosOnAttractor = xTargetPos;
	xData.m_xBestOrientationOnAttractor = Asura_Matrix_3x3(xTargetRight.x, -xNormal.x, xTargetForward.x, xTargetRight.y, -xNormal.y, xTargetForward.y, xTargetRight.z, -xNormal.z, xTargetForward.z);
}

void AVP_ClientEntity_Vent::TestDuctAttractors_Callback(void* pData, const Asura_Attractor_Compiled* const pxAttractor)
{
	if (pData && pxAttractor)
	{
		EntryPointCallbackData* pxData = static_cast<EntryPointCallbackData*>(pData);
		pxData->m_xVent.TestDuctAttractor(*pxData, *pxAttractor);
	}
}

DEBUGVAR float dbg_fVentMaxNearbyHookDistance = 5.0f;

void AVP_ClientEntity_Vent::TestForNearbyEnvHooks_Callback(void* pData, const Asura_Attractor_Compiled* const pxAttractor)
{
	if (pData && pxAttractor && (pxAttractor->GetType() == uAVP_ATTRACTOR_TYPE_DUCT))
	{
		// Found a nearby duct, non-exit-point use is now allowed as long as it's within a sensible range.
		AVP_ClientEntity_Vent* pxVent = static_cast<AVP_ClientEntity_Vent*>(pData);
		float fDistanceFromHookAxis;
		pxAttractor->GetRelativePosition(pxVent->GetCurrentPosition(), 0.0f, NULL, NULL, &fDistanceFromHookAxis);
		if (fDistanceFromHookAxis <= dbg_fVentMaxNearbyHookDistance)
		{
			pxVent->SetUseExitPointsOnly(false);
		}
	}
}
