//////////////////////////////////////////////////////////////////////////////
//
// AVP_ServerEntity_Vent.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     Physical object which can be affected by explosions (initially static)
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Vent/AVP_ServerEntity_Vent.h"

// Engine independent.
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"

// Project
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Locations/OrientedNode/GamesceneNode/Asura_ServerNode_GS2Node.h"
#include "Entity/Locations/Spline/GamesceneSpline/Asura_ServerNode_GS2Spline.h"
#include "Entity/Vent/AVP_Vent_Messages.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static const u_int	g_uCurrentChunkVersion	= 2;

Asura_Hashed_List<AVP_ServerEntity_Vent> AVP_ServerEntity_Vent::s_xList;


AVP_ServerEntity_Vent::AVP_ServerEntity_Vent(const Asura_Guid uGuid)
: Asura_Entity_Server(uGuid)
, m_bIsDamaged( false )
, m_uUndamagedSkinHash( ASURA_HASH_ID_UNSET )
, m_uDamagedSkinHash( ASURA_HASH_ID_UNSET )
, m_uNumVentNodes( 0 )
, m_uNumExitPoints(0)
, m_pxExitPoints(NULL)
{
	Classification = AVP_ENTITYCLASS_VENT;
	s_xList.AddToList(this);
	SetNeverUpdateClient();
	SetEntityActiveFromLoad( true );
}

AVP_ServerEntity_Vent::~AVP_ServerEntity_Vent(void)
{
	delete[] m_pxExitPoints;

	s_xList.RemoveFromList(this);
}

void AVP_ServerEntity_Vent::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// Write the version..
	xStream << g_uCurrentChunkVersion;

	// Write the parent data
	PARENT::WriteToChunkStream( xStream );

	xStream << m_xPosition;
	xStream << m_xOrientation;

	xStream << m_bIsDamaged;

	xStream << m_uUndamagedSkinHash;
	xStream << m_uDamagedSkinHash;

	Asura_Assert( m_uNumVentNodes <= AVP_MAX_VENT_NODES, "Chunk Error: There are too many nodes on this vent" );
	xStream << Asura_Maths::Min( m_uNumVentNodes, static_cast< u_int >( AVP_MAX_VENT_NODES ) );
	for ( u_int uItr = 0; uItr < m_uNumVentNodes && uItr < AVP_MAX_VENT_NODES; ++uItr )
	{
		xStream << m_auVentNode[ uItr ];
	}

	xStream << m_uNumExitPoints;
	for (u_int u=0; u<m_uNumExitPoints; ++u)
	{
		xStream << m_pxExitPoints[u].m_xPosition;
		xStream << m_pxExitPoints[u].m_xOrientation;
	}
}

void AVP_ServerEntity_Vent::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	// Get the version..
	xStream >> uVersion;

	// Check the version
	if (uVersion > g_uCurrentChunkVersion)
	{
		Asura_Assertf(false, ("Unexpected chunk-version for vent - aborting read."));
		return;
	}

	Asura_Assert( uVersion >= 1, "Vent version 0 no longer supported, if this is a problem speak to Will" );
	if (uVersion == 2)
	{
		// Read the parent data
		PARENT::ReadFromChunkStream(xStream);

		xStream >> m_xPosition;
		xStream >> m_xOrientation;

		xStream >> m_bIsDamaged;

		xStream >> m_uUndamagedSkinHash;
		xStream >> m_uDamagedSkinHash;

		xStream >> m_uNumVentNodes;
		Asura_Assert( m_uNumVentNodes <= AVP_MAX_VENT_NODES, "Chunk Error: There are too many nodes on this vent" );
		for ( u_int uItr = 0; uItr < m_uNumVentNodes && uItr < AVP_MAX_VENT_NODES; ++uItr )
		{
			xStream >> m_auVentNode[ uItr ];
		}

		xStream >> m_uNumExitPoints;
		if (m_uNumExitPoints > 0)
		{
			Asura_Assert(!m_pxExitPoints, "Shouldn't have allocated exit point data yet. Memory leak!");
			m_pxExitPoints = new ExitPoint[m_uNumExitPoints];
			Asura_Assert(m_pxExitPoints, "Failed to allocate exit points! If you're in tools, saving the level may delete any previously set up.");
			if (m_pxExitPoints)
			{
				for (u_int u=0; u<m_uNumExitPoints; ++u)
				{
					xStream >> m_pxExitPoints[u].m_xPosition;
					xStream >> m_pxExitPoints[u].m_xOrientation;
				}
			}
			else
			{
				m_uNumExitPoints = 0;
			}
		}
	}
	else if ( uVersion == 1 )
	{
		// Read the parent data
		PARENT::ReadFromChunkStream(xStream);

		xStream >> m_xPosition;
		xStream >> m_xOrientation;

		xStream >> m_bIsDamaged;

		xStream >> m_uUndamagedSkinHash;
		xStream >> m_uDamagedSkinHash;

		xStream >> m_uNumVentNodes;
		Asura_Assert( m_uNumVentNodes <= AVP_MAX_VENT_NODES, "Chunk Error: There are too many nodes on this vent" );
		for ( u_int uItr = 0; uItr < m_uNumVentNodes && uItr < AVP_MAX_VENT_NODES; ++uItr )
		{
			xStream >> m_auVentNode[ uItr ];
		}

#ifdef ASURA_TOOLS
		// If we're in tools, create a couple of default points ready for placement.
		CreateDefaultExitPoints();
#endif //#ifdef ASURA_TOOLS
	}
}

void AVP_ServerEntity_Vent::MakeCreateMessage( Asura_Guid uRequestingClient, bool bSendGuaranteed /*= false*/ )
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_CLIENTCREATE, sizeof( AVP_Message_Vent_ClientCreate ), GetClassification());
	if(pxMessage)
	{
		AVP_Message_Vent_ClientCreate* pxData = static_cast<AVP_Message_Vent_ClientCreate*>(pxMessage->GetData());
		pxData->FillVent(this);

		pxMessage->SetToGuid( ASURA_GUID_MY_CLIENT_ENTITIES );
		pxMessage->SetFromGuid( GetGuid() );
		if (bSendGuaranteed)
		{
			pxMessage->UseGuaranteedDelivery();
		}
		Asura_Network::SendToSpecificMachine( pxMessage, uRequestingClient );
	}
}

#ifdef ASURA_TOOLS

static const Asura_Vector_3 xDEFAULT_EXIT_1_OFFSET(0.0f, -0.5f, 1.0f);
static const Asura_Vector_3 xDEFAULT_EXIT_2_OFFSET(0.0f, 0.5f, 1.0f);

void AVP_ServerEntity_Vent::CreateDefaultExitPoints()
{
	const Asura_Quat xOriQuat(m_xOrientation);

	Asura_Vector_3 xExitPos = m_xPosition + (xDEFAULT_EXIT_1_OFFSET * m_xOrientation);
	CreateExitPoint(xExitPos, xOriQuat);

	xExitPos = m_xPosition + (xDEFAULT_EXIT_2_OFFSET * m_xOrientation);
	CreateExitPoint(xExitPos, xOriQuat);
}

void AVP_ServerEntity_Vent::CreateExitPoint(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation)
{
	Asura_Assert(m_uNumExitPoints < AVP_Message_Vent_ClientCreate::MAX_EXIT_POINTS, "Too many player exit points, ask a programmer to increase the limit.");
	if (m_uNumExitPoints >= AVP_Message_Vent_ClientCreate::MAX_EXIT_POINTS) { return; }

	// We're in tools, so we don't really care about newing and deleting pointers as needed.
	ExitPoint* pxNewExits = new ExitPoint[m_uNumExitPoints+1];
	Asura_Assert(pxNewExits, "Failed to allocate new player exit points, point not created.");
	if (!pxNewExits) { return; }

	// Copy the old data.
	if (m_uNumExitPoints > 0)
	{
		Asura_Mem::Copy(pxNewExits, m_pxExitPoints, sizeof(ExitPoint) * m_uNumExitPoints);
	}
	// Add the new point.
	pxNewExits[m_uNumExitPoints].m_xPosition = xPosition;
	pxNewExits[m_uNumExitPoints].m_xOrientation = xOrientation;
	++m_uNumExitPoints;

	// Clean up the old data and connect up our new list.
	delete[] m_pxExitPoints;
	m_pxExitPoints = pxNewExits;
}

void AVP_ServerEntity_Vent::DeleteExitPoint(u_int uIndex)
{
	Asura_Assert(uIndex < m_uNumExitPoints, "Invalid player exit to delete.");
	if (uIndex >= m_uNumExitPoints) { return; }
	Asura_Assert(m_uNumExitPoints > 1, "Cannot delete the final player exit point.");
	if (m_uNumExitPoints <= 1) { return; }

	// We're in tools, so we don't really care about newing and deleting pointers as needed.
	ExitPoint* pxNewExits;
	pxNewExits = new ExitPoint[m_uNumExitPoints-1]; // We know m_uNumExitPoints is greater than 1.
	Asura_Assert(pxNewExits, "Failed to allocate new player exit points, point not deleted.");
	if (!pxNewExits) { return; }

	// Copy the old data.
	u_int uNewIndex = 0;
	u_int uOldIndex = 0;
	for (; uOldIndex<m_uNumExitPoints; ++uOldIndex)
	{
		if (uOldIndex != uIndex)
		{
			pxNewExits[uNewIndex] = m_pxExitPoints[uOldIndex];
			++uNewIndex;
		}
	}
	--m_uNumExitPoints;

	// Clean up the old data and connect up our new list.
	delete[] m_pxExitPoints;
	m_pxExitPoints = pxNewExits;
}

void AVP_ServerEntity_Vent::CopyExitPoints(const u_int uNumPoints, const ExitPoint* pxExitPoints)
{
	ExitPoint* pxNewExits = new ExitPoint[uNumPoints];
	Asura_Assert(pxNewExits, "Failed to allocate exit points array.");
	if (!pxNewExits) { return; }

	delete[] m_pxExitPoints;
	m_pxExitPoints = pxNewExits;

	for (u_int u=0; u<uNumPoints; ++u)
	{
		m_pxExitPoints[u] = pxExitPoints[u];
	}

	m_uNumExitPoints = uNumPoints;
}

void AVP_ServerEntity_Vent::ApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance )
{
	PARENT::ApplyChangesForTemplateInstance(pxInstance);

	m_xPosition = pxInstance->TransformPosition( m_xPosition );
	m_xOrientation = pxInstance->TransformMatrix( m_xOrientation );

	for ( u_int uNodeItr = 0; uNodeItr < m_uNumVentNodes; ++uNodeItr )
	{
		m_auVentNode[ uNodeItr ] = pxInstance->ConvertGuid( m_auVentNode[ uNodeItr ] );
	}

	for (u_int u=0; u<m_uNumExitPoints; ++u)
	{
		m_pxExitPoints[u].m_xPosition = pxInstance->TransformPosition(m_pxExitPoints[u].m_xPosition);
		m_pxExitPoints[u].m_xOrientation = pxInstance->TransformQuat(m_pxExitPoints[u].m_xOrientation);
	}
}

bool AVP_ServerEntity_Vent::CanBeIncludedInTemplate(void) const
{
	return GetClassification() == AVP_ENTITYCLASS_VENT;
}

#endif // #ifdef ASURA_TOOLS

bool AVP_ServerEntity_Vent::AreNodesConnected( const u_int uStartNode, const u_int uEndNode )
{
	const Asura_ServerNode_GS2Node* pxStartNode = Asura_ServerNode_GS2Node::FindEntity( uStartNode );
	const Asura_ServerNode_GS2Node* pxEndNode = Asura_ServerNode_GS2Node::FindEntity( uEndNode );
	if ( pxStartNode == NULL || pxEndNode == NULL )
	{
		return false;
	}

	bool bConnected = false;

	for ( u_int uItr = 0; uItr < pxStartNode->GetNumberLinks(); ++uItr )
	{
		Asura_ServerNode_GS2Spline* pxSpline = Asura_ServerNode_GS2Spline::FindEntity( pxStartNode->GetLink( uItr ) );
		bConnected |= ( pxSpline != NULL && pxSpline->GetTargetGuid() == uEndNode );
	}

	return bConnected;
}
