//////////////////////////////////////////////////////////////////////////////
//
// AVP_FSFX.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined(ASURA_USE_FSFX)

#include "FSFX/AVP_FSFX.h"

#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_TextureManagement.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"

/******************************************************************************/
/* Macros                                                                     */
/******************************************************************************/

#if defined ASURA_TOOLS
	// All supported platforms enabled by default
	#define SET_PLATFORM_ACTIVE_FLAGS() m_iPlatformsActive = GetPlatformsSupportedID(); 
#else
	// All platforms...
	#define SET_PLATFORM_ACTIVE_FLAGS() m_iPlatformsActive = (1 << NUM_PLATFORMS) - 1; 
#endif // defined ASURA_TOOLS

/******************************************************************************/
/* Asura_FSFX_Component Static Class Functions                                */
/******************************************************************************/

Asura_FSFX_Component* Asura_FSFX_Component::Project_CreateOfType( const Type eType )
{
	switch( (int)eType )
	{
		case AVP_FSFX_USE_STENCIL:
		{
			return new AVP_FSFX_Component_UseStencil;
		}
		case AVP_FSFX_DISPLACEMENT_MAP:
		{
			return new AVP_FSFX_Component_DisplacementMap;
		}
		case AVP_FSFX_EDGES:
		{
			return new AVP_FSFX_Component_Edges;
		}

		default:
		{
			return NULL;
		}
	}
}

/******************************************************************************/
/* AVP_FSFX_Component_UseStencil Class Functions                              */
/******************************************************************************/

AVP_FSFX_Component_UseStencil::AVP_FSFX_Component_UseStencil()
	: m_uBitsToCheck( 0 ), m_uBitValues( 0 )
{
	SET_PLATFORM_ACTIVE_FLAGS();
}

Asura_FSFX_Component* AVP_FSFX_Component_UseStencil::CreateCopy()
{
	AVP_FSFX_Component_UseStencil* pxComp = new AVP_FSFX_Component_UseStencil;
	*pxComp = *this;
	return pxComp;
}

void AVP_FSFX_Component_UseStencil::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

	#if defined ASURA_ASSERT
		u_int uPos = xStream.GetPosition();
	#endif

	switch( iVersion )
	{
		case 1:
		{
			bool bEnableStencil;
			xStream >> bEnableStencil;

			if (bEnableStencil)
			{
				m_uBitsToCheck = AVP_MATERIAL_RESPONSE_PROJECT_FLAG_ALIEN;
			} else {
				m_uBitsToCheck = 0;
			}
			m_uBitValues = 0;
			break;
		}	
		case 2:
			{
				xStream >> m_uBitsToCheck;
				xStream >> m_uBitValues;
				m_uBitsToCheck &= AVP_MATERIAL_RESPONSE_PROJECT_FLAG_ALLUSED;
				m_uBitValues &= AVP_MATERIAL_RESPONSE_PROJECT_FLAG_ALLUSED;

				break;
			}

		default:
		{
			Asura_Assert( false, "AVP_FSFX_Component_UseStencil version not supported. Skipping." );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in AVP_FSFX_Component_UseStencil::Deserialize." );
}

#if defined ASURA_TOOLS

void AVP_FSFX_Component_UseStencil::Serialize( Asura_Chunk_Stream& xStream ) const
{
	// bools are chunked as chars to avoid platform size differences
	int iSize = 2 * sizeof( u_int );
	WriteStandardHeader( xStream, iSize );

	u_int uPos = xStream.GetPosition();


	xStream << m_uBitsToCheck;
	xStream << m_uBitValues;

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in AVP_FSFX_Component_UseStencil::Serialize." );
}

#endif // defined ASURA_TOOLS


/******************************************************************************/
/* AVP_FSFX_Component_DisplacementMap Class Functions                              */
/******************************************************************************/

int AVP_FSFX_Component_DisplacementMap::s_iReferences = 0;

AVP_FSFX_Component_DisplacementMap::AVP_FSFX_Component_DisplacementMap()
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_fAmount = 0.0f;
	m_xTextureFilename = "";
	m_uTextureHash = ASURA_HASH_ID_UNSET;
	m_iTextureID = ASURA_TEXTURE_INVALID;

	if (s_iReferences++ == 0)
	{
		Platform_Initialise();
	}
}

AVP_FSFX_Component_DisplacementMap::~AVP_FSFX_Component_DisplacementMap()
{
	if (--s_iReferences == 0)
	{
		Platform_Shutdown();
	}
}


Asura_FSFX_Component* AVP_FSFX_Component_DisplacementMap::CreateCopy()
{
	AVP_FSFX_Component_DisplacementMap* pxComp = new AVP_FSFX_Component_DisplacementMap;
	*pxComp = *this;
	return pxComp;
}

void AVP_FSFX_Component_DisplacementMap::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	u_int uPos = xStream.GetPosition();
#endif

	m_iTextureID = ASURA_TEXTURE_INVALID;
	m_xTextureFilename = "";
	m_fAmount = 0.0f;

	switch( iVersion )
	{
	case 1:
		{
			xStream >> m_uTextureHash;
			xStream >> m_fAmount;
#if defined( ASURA_TOOLS )
			xStream >> m_xTextureFilename;
#else // #if defined( ASURA_TOOLS )
			xStream.SkipString();
#endif
			UpdateTexture();

			break;
		}	
	default:
		{
			Asura_Assert( false, "AVP_FSFX_Component_DisplacementMap version not supported. Skipping." );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in AVP_FSFX_Component_DisplacementMap::Deserialize." );
}

#if defined ASURA_TOOLS

void AVP_FSFX_Component_DisplacementMap::Serialize( Asura_Chunk_Stream& xStream ) const
{
	// bools are chunked as chars to avoid platform size differences
	int iSize = (2 + m_xTextureFilename.GetSizeInInts()) * sizeof( u_int );
	WriteStandardHeader( xStream, iSize );
	u_int uPos = xStream.GetPosition();
	
	xStream << m_uTextureHash;
	xStream << m_fAmount;
	xStream << m_xTextureFilename;


	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in AVP_FSFX_Component_DisplacementMap::Serialize." );
}

#endif // defined ASURA_TOOLS

void AVP_FSFX_Component_DisplacementMap::SetTexture(const Asura_String &xTextureName, int iNewTexture) 
{
	m_iTextureID = iNewTexture;
	m_xTextureFilename = xTextureName;
	m_uTextureHash = m_iTextureID != ASURA_TEXTURE_NONE && m_iTextureID != ASURA_TEXTURE_INVALID ? Asura_GetHashID_IgnoreExtension( m_xTextureFilename.GetString() ) : ASURA_HASH_ID_UNSET;
}


void AVP_FSFX_Component_DisplacementMap::UpdateTexture(  ) 
{
#if defined( ASURA_TOOLS )
	bool bWasNew = false;
	Asura_Assert( m_xTextureFilename.GetLength() > 0, "Invalid texture name on FSFX noise component" );
	if ( ASURA_TEXTURE_INVALID == m_iTextureID )
	{
		const int iPrevTexResetGrp = Asura_Texture_Management::GetTextureResetGroup();
		Asura_Texture_Management::SetTextureResetGroup( ASURA_TEXTURE_LOADED_PERMANENT );

		m_iTextureID = Asura_Texture_Management::AddTexture( m_xTextureFilename, &bWasNew );
		// Make the hash value 
		m_uTextureHash = Asura_GetHashID_IgnoreExtension( m_xTextureFilename.GetString() );

		Asura_Texture_Management::SetTextureResetGroup( iPrevTexResetGrp );
	}

	if ( ASURA_TEXTURE_NONE == m_iTextureID )
	{
		m_uTextureHash = ASURA_HASH_ID_UNSET;
		return;
	}

	if ( ASURA_TEXTURE_INVALID == m_iTextureID )
	{
		m_uTextureHash = ASURA_HASH_ID_UNSET;
		return;
	}

	if( true == bWasNew )
	{
		Asura_Texture_Management::LoadTextures();
	}
#else
	if ( m_uTextureHash == ASURA_HASH_ID_UNSET )
	{
		m_iTextureID = ASURA_TEXTURE_NONE;
		return;
	}
	m_iTextureID = Asura_Texture_Management::GetIndexFromHashID( m_uTextureHash );
	Asura_Assert( this->m_iTextureID != ASURA_TEXTURE_INVALID, "" );
#endif // #if defined( ASURA_TOOLS )

}




/******************************************************************************/
/* AVP_FSFX_Component_Edges Class Functions                                   */
/******************************************************************************/

int AVP_FSFX_Component_Edges::s_iReferences = 0;

AVP_FSFX_Component_Edges::AVP_FSFX_Component_Edges()
{
	SET_PLATFORM_ACTIVE_FLAGS();

	m_xTintColour = Asura_RGBA(135.0f/255.0f, 111.0f/255.0f, 0.0, 0.0f);
	m_fBlurWidth = 75.0f;
	m_fEDWidth = 10.0f;
	m_fAmount = 10.0f;
	m_fLowThreshold = 0.0f;
	m_fHighThreshold = 40.0f;
	m_uMethod = 1;


	if (s_iReferences++ == 0)
	{
		Platform_Initialise();
	}
}

AVP_FSFX_Component_Edges::~AVP_FSFX_Component_Edges()
{
	if (--s_iReferences == 0)
	{
		Platform_Shutdown();
	}
}


Asura_FSFX_Component* AVP_FSFX_Component_Edges::CreateCopy()
{
	AVP_FSFX_Component_Edges* pxComp = new AVP_FSFX_Component_Edges;
	*pxComp = *this;
	return pxComp;
}

void AVP_FSFX_Component_Edges::Deserialize( Asura_Chunk_Stream& xStream )
{
	int iVersion, iSize;
	ReadStandardHeader( xStream, iVersion, iSize );

#if defined ASURA_ASSERT
	u_int uPos = xStream.GetPosition();
#endif


	switch( iVersion )
	{
	case 1:
		{
			xStream >> m_xTintColour.m_fR;
			xStream >> m_xTintColour.m_fG;
			xStream >> m_xTintColour.m_fB;
			m_xTintColour.m_fA = 0.0f;

			xStream >> m_fBlurWidth;
			xStream >> m_fEDWidth;
			xStream >> m_fAmount;
			xStream >> m_fLowThreshold;
			xStream >> m_fHighThreshold;
			xStream >> m_uMethod;

			break;
		}
	default:
		{
			Asura_Assert( false, "AVP_FSFX_Component_Edges version not supported. Skipping." );
			xStream.Skip( iSize );
			break;
		}
	}

	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in AVP_FSFX_Component_Edges::Deserialize." );
}

#if defined ASURA_TOOLS

void AVP_FSFX_Component_Edges::Serialize( Asura_Chunk_Stream& xStream ) const
{
	// bools are chunked as chars to avoid platform size differences
	int iSize = 9 * sizeof( u_int );
	WriteStandardHeader( xStream, iSize );
	u_int uPos = xStream.GetPosition();

	xStream << m_xTintColour.m_fR;
	xStream << m_xTintColour.m_fG;
	xStream << m_xTintColour.m_fB;

	xStream << m_fBlurWidth;
	xStream << m_fEDWidth;
	xStream << m_fAmount;
	xStream << m_fLowThreshold;
	xStream << m_fHighThreshold;
	xStream << m_uMethod;



	Asura_Assert( xStream.GetPosition() == uPos + iSize, "Size mismatch in AVP_FSFX_Component_Edges::Serialize." );
}

#endif // defined ASURA_TOOLS



#endif	// defined(ASURA_USE_FSFX)


/// End of file. ///
