//////////////////////////////////////////////////////////////////////////////
//
// AVP_DevLobby.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(AVP_USE_DEV_LOBBY)

// This file's header.
#include "FrontEnd/AVP_DevLobby.h"

// Asura includes
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_HashedLocalisedText.h"
#include "Fileaccess/Asura_FileAccess.h"
#include "GUIMenu/Asura_GUIMenu_System.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_NetlinkManager.h"
#include "Network/Online/Asura_Lobby.h"
#include "Network/Online/Asura_Matchmaking_System.h"
#include "Rendering/Asura_Viewport.h"

// AVP includes
#include "Core/AVP_Core.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_GameStatus.h"
#include "Core/AVP_Settings.h"
#include "Core/AVP_StateMachine_Core.h"
#include "GUIMenu/AVP_GUIMenu_System.h"
#include "Multiplayer/AVP_Multiplayer_GameRules.h"
#include "Network/AVP_Client.h"
#include "Network/AVP_Server.h"
#include "Network/AVP_VoiceChannels.h"


// I N T E R N A L   C L A S S E S ///////////////////////////////////////////

// we're encoding game version from Asura_NetTransport_System and the ranked 
// match status in the AVP_LOBBY_ATTRIBUTE_GAME_VERSION attribute. Ranked 
// match status is only filtered automatically on 360 so this is needed to 
// prevent PS3 and PC from returning the wrong type of match. Ideally this 
// would be done as standard by the lobby system but search criteria are all 
// done in project code now.
struct AVP_DevLobby_GameVer 
{
	u_short m_uFlags;
	u_short m_uVersion;
};

union AVP_DevLobby_GameVer_Union
{
	int m_iGameVer;
	AVP_DevLobby_GameVer m_xGameVer;
};

struct AVP_DevLobby_AdvancedSettingsA
{
	u_int	m_bAllowMixedTeams		: 1,
			m_uSpeciesRatioPredator	: 3,
			m_uSpeciesRatioHuman	: 3,
			m_uSpeciesRatioAlien	: 3,
			m_uNumberOfTeams		: 2, 
			m_uTeamOneSpecies		: 2, 
			m_uTeamTwoSpecies		: 2, 
			m_uTeamThreeSpecies		: 2,
			m_bRespawnUponDeath		: 1,
			m_uRespawnType			: 2,
			m_uLives				: 6,
			m_bFriendlyFire			: 1;
			// 4 spare bits
};

union AVP_DevLobby_AdvancedSettingsA_Union
{
	u_int	m_uAdvancedSettings; 
	AVP_DevLobby_AdvancedSettingsA m_xAdvancedSettings; 
};

struct AVP_DevLobby_AdvancedSettingsB
{
	u_int	m_uScoreLimit			: 16,	// Plenty
			m_uTimeLimit			: 6;	// In minutes, up to an hour
			// 10 spare bits
};

union AVP_DevLobby_AdvancedSettingsB_Union
{
	u_int	m_uAdvancedSettings; 
	AVP_DevLobby_AdvancedSettingsB m_xAdvancedSettings; 
};

// C O N S T A N T S /////////////////////////////////////////////////////////
static const int iAVP_LOBBY_INVALID_SEARCH_FILTER = -1;
static const u_int uAVP_DEVLOBBY_MAX_SKILL_THRESHOLD = 19;

// check num attributes is correct
ASURA_STATIC_ASSERT(AVP_NUM_LOBBY_ATTRIBUTES == ASURA_LOBBY_NUM_INT_ATTRIBUTES);

// G L O B A L   D A T A /////////////////////////////////////////////////////
static const char* szMULTIPLAYER_MAP_PATH = "Envs\\MP\\";


// L O C A L   F U N C T I O N S /////////////////////////////////////////////
static Asura_Hash_ID GetSpeciesNameTextHashID(AVP_Settings::Species eSpecies)
{
	switch(eSpecies)
{
	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return 0x593E169A;	// Asura_GetHashID("Species_Predator");
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return 0xF0305E04;	// Asura_GetHashID("Species_Alien");
	case AVP_Settings::AVP_SPECIES_HUMAN:
		return 0x29BEAE25;	// Asura_GetHashID("Species_Marine");
	default:
		return ASURA_HASH_ID_UNSET;
	}
}

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool AVP_DevLobby::s_bIsInitialised;
bool AVP_DevLobby::s_bSetupComplete;
bool AVP_DevLobby::s_bQuitRequested;
bool AVP_DevLobby::s_bAllowRankedMatch = true;
AVP_DevLobby_GameSettings AVP_DevLobby::s_xLobbySettings;
int AVP_DevLobby::s_aiLobbySearchFilters[AVP_NUM_LOBBY_ATTRIBUTES];
bool AVP_DevLobby::s_bGametypeChanged;
int AVP_DevLobby::s_iPrevGametype;
bool AVP_DevLobby::s_bAutoJoinOnSearchComplete;
Asura_LinkList<AVP_DevLobby_MapData*> AVP_DevLobby::s_xMapList;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void AVP_DevLobby::Initialise()
{
	// set online game to true. On PC this may be set to false by 
	// Platform_Initialise() if using LAN.
	AVP_Settings::SetOnlineGame(true);
	Platform_Initialise();

	s_bSetupComplete = false;
	s_bQuitRequested = false;
	
	const bool bLobbySystemWasInitialised = Asura_Lobby_System::IsInitialised();
	if (!bLobbySystemWasInitialised)
	{
		Asura_Lobby_System::Initialise();
		// Allow joining games in progress
		Asura_Lobby_System::SetAllowJoinInProgress( true ); 

		// No need to repeatedly create the map list, just do it the first time we go into multiplayer.
		CreateMapList();
	}

	for (u_int u = 0; u < AVP_NUM_LOBBY_ATTRIBUTES; ++u)
	{
		s_aiLobbySearchFilters[u] = iAVP_LOBBY_INVALID_SEARCH_FILTER;
	}

	s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP] = s_xMapList.Size(); // Any map is signified by the limit value instead, since the menu elements don't support a -1 value.

	Asura_Matchmaking_System::Initialise();

	// open first page: either post game page or top lobby menu
	if (bLobbySystemWasInitialised)
	{
		Asura_Assert((Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_ENDING_GAME) || (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_TERMINATING) || (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID), "unexpected lobby state");
		Asura_GUIMenu_System::OpenMenuPage( 0x3B3C1630 /* Asura_GetHashID("DevLobby_PostGame") */ );
	}
	else
	{
		Asura_GUIMenu_System::OpenMenuPage( 0x06950657 /* Asura_GetHashID("DevLobby_Main") */ );
	}

	s_bIsInitialised = true;
}

void AVP_DevLobby::AddConsoleCmdsAndVars()
{
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.AllowRankedMatch",			&s_bAllowRankedMatch,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.AllowMixedTeams",			&s_xLobbySettings.m_bAllowMixedTeams,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.RespawnUponDeath",			&s_xLobbySettings.m_bRespawnUponDeath,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );	
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.FriendlyFire",				&s_xLobbySettings.m_bFriendlyFire,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.Quit",						RequestQuit,							Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.CreateGame",				CreateGame,								Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.SearchForGame",			SearchForGame,							Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.BrowseForGame",			BrowseForGame,							Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.JoinGame",					JoinGame,								Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.ListGame",					ListGame,								Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.StartGame",				StartGame,								Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.ChangeSpecies",			ChangeSpecies,							Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.ChangeTeam",				ChangeTeam,								Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.AlterGameType",			AlterGameType,							Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.ApplyAdvancedSettings",	ApplyAdvancedSettings,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.AllowChangingTeam",		ConsoleCmd_AllowChangingTeam,			Asura_CommandConsole::REMOVE_AT_SHUTDOWN ); 
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.AllowChangingSpecies", 	ConsoleCmd_AllowChangingSpecies,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.IsRoleBasedGame",			ConsoleCmd_IsRoleBasedGame,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.PopulateMapSelectorNames", PopulateMapSelectorName,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.PopulateMapSearchNames",	PopulateMapSearchName,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.DevLobby.PopulateMapDisplayNames",	PopulateMapDisplayName,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	// hack - allow hosting tiny games to help testing.
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.MaxPlayers",			&s_xLobbySettings.m_uMaxPlayers,						2 /* uAVP_MAX_NET_PLAYERS_MIN */,	uAVP_MAX_NET_PLAYERS_MAX,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.MapIndex",				&s_xLobbySettings.m_iMapIndex,							0,									Asura_Maths::Max(static_cast<int>(s_xMapList.Size()) - 1, 0),	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.Gametype",				&s_xLobbySettings.m_iGametypeIndex,						0,									AVP_Settings::AVP_NUM_GAMETYPES - 1,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.Species",				&s_xLobbySettings.m_iLocalSpecies,						0, 									AVP_Settings::AVP_NUM_SPECIES - 1,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.SpeciesRatioPredator", &s_xLobbySettings.m_iSpeciesRatioPredator,				0, 									7,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.SpeciesRatioHuman",	&s_xLobbySettings.m_iSpeciesRatioHuman, 				0, 									7,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.SpeciesRatioAlien",	&s_xLobbySettings.m_iSpeciesRatioAlien, 				0, 									7,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.NumberOfTeams",		&s_xLobbySettings.m_iNumberOfTeams,						2, 									3,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.TeamOneSpecies",		&s_xLobbySettings.m_iTeamOneSpecies,					0, 									2,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.TeamTwoSpecies",		&s_xLobbySettings.m_iTeamTwoSpecies,					0, 									2,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.TeamThreeSpecies",		&s_xLobbySettings.m_iTeamThreeSpecies,					0, 									2,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.RespawnType",			&s_xLobbySettings.m_iRespawnType,						0, 									3,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.Lives",				&s_xLobbySettings.m_iLives,								0, 									1023,									Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.ScoreLimit",			&s_xLobbySettings.m_iScoreLimit,						0, 									65535,									Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.TimeLimit",			&s_xLobbySettings.m_iTimeLimit,							0, 									63,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	ASURA_STATIC_ASSERT( iAVP_LOBBY_INVALID_SEARCH_FILTER == -1 );
	Asura_CommandConsole::AddVar( L"SeaS",		&s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP],		iAVP_LOBBY_INVALID_SEARCH_FILTER,	
		static_cast<int>(s_xMapList.Size()), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.SearchFilter.Gametype",&s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAMETYPE],	iAVP_LOBBY_INVALID_SEARCH_FILTER,	AVP_Settings::AVP_NUM_GAMETYPES - 1,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	Asura_CommandConsole::SetPointerToModifiedFlag(L"AVP.DevLobby.Gametype", &s_bGametypeChanged);
}

void AVP_DevLobby::Shutdown()
{
	// shutdown the matchmaking system, telling it whether we should leave any 
	// lobbies we're currently in.
	const bool bLeaveLobby = s_bQuitRequested;
	Asura_Matchmaking_System::Shutdown(bLeaveLobby);

	// only shutdown the lobby if we quit from it, otherwise it remains active for the game duration.
	if (IsQuitRequested() || AVP_GameStatus::GetQuitApplication())
	{
		Asura_Lobby_System::Shutdown();

		// Also only clear the map list on quit as it only gets created when the lobby system initialises.
		ClearMapList();
	}

	Platform_Shutdown();

	s_bIsInitialised = false;
}

bool AVP_DevLobby::Update()
{
	if (!Asura_NetLinkManager::IsConnected())
	{
		// connection dropped, any further usage of the network functions could crash.
		// The game should display a "connection dropped" dialog. It could do this by 
		// trapping the netlink state ASURA_NETLINK_CONNECTION_DROPPED in the main menu
		// and switching to the appropriate error menu, then calling Asura_NetLinkManager::ResetDroppedStatus()
		// either directly or using the appropriate console command.
		s_bQuitRequested = true;
		return true;
	}

	Asura_CommandConsole::Update();

	if (!Asura_Matchmaking_System::IsIdle())
	{
		Asura_Matchmaking_System::Update();
		if (Asura_Matchmaking_System::HasErrorOccurred())
		{
			s_bQuitRequested = true;
			return true;
		}
		if (!Asura_Matchmaking_System::IsFinished())
		{
			return false;
		}
	}
	else
	{
		// pick a server if we've done a search
		if (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID && 
			Asura_Lobby_System::GetSearchState() == ASURA_LOBBY_SEARCHSTATE_SEARCH_COMPLETE &&
			s_bAutoJoinOnSearchComplete )
		{
			const u_int uNumSearchResults = Asura_Lobby_System::GetNumSearchResults();
			u_int uIndexToJoin = 0;
			u_int uMostPlayers = 0;
			for (u_int u = 0; u < uNumSearchResults; ++u)
			{
				// pick the one with the most players, if it has some room in it
				const u_int uNumPlayers = Asura_Lobby_System::GetSearchResultByIndex(u).m_uNumPlayers;
				const u_int uMaxPlayers = Asura_Lobby_System::GetSearchResultByIndex(u).m_uMaxPlayers;
				if (uNumPlayers < uMaxPlayers && uNumPlayers > uMostPlayers)
				{
					uIndexToJoin = u;
					uMostPlayers = uNumPlayers;
				}
			}
			if (uIndexToJoin < uNumSearchResults)
			{
				Asura_Lobby_System::JoinLobby(uIndexToJoin);
				s_xLobbySettings.m_uMaxPlayers = Asura_Lobby_System::GetSearchResultByIndex(uIndexToJoin).m_uMaxPlayers;
			}
		}
	}

	// keep lobby data up to date on clients, force gametype to be valid
	UpdateLobbySettings();

	// keep team names up-to-date
	UpdateTeamNames();

	// voice channels settings for the Lobby
	SetLobbyVoiceChannelsSettings(); 

	// check the lobby gamestate to see if we should load up the game
	if (Asura_Lobby_System::IsReadyToJoinGame())
	{
		s_bSetupComplete = true;
		InitialiseGameSettings();
	}

	return s_bSetupComplete || s_bQuitRequested;
}

void AVP_DevLobby::UpdateLobbySettings()
{
	// force gametype to be allowed value
	if (s_bGametypeChanged)
	{
		CheckGametypeIsAllowed();
	}
	s_iPrevGametype = s_xLobbySettings.m_iGametypeIndex;

	// keep lobby data up to date on clients
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	if (pxLobby)
	{
		if (!Asura_Lobby_System::IsHost())
		{
			s_xLobbySettings.m_iGametypeIndex = pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAMETYPE);
			s_xLobbySettings.m_iMapIndex = GetMapIndexByHash(pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_MAP));
			AVP_Settings::SetGameType( static_cast<AVP_Settings::GameType>(s_xLobbySettings.m_iGametypeIndex) );
			AVP_Multiplayer_GameRules::Refresh(); 
			
			// Update the advanced settings after refreshing the game rules, to make sure all default values are overridden
			AVP_DevLobby_AdvancedSettingsA_Union xAdvSettingsA; 
			xAdvSettingsA.m_uAdvancedSettings = pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_ADVANCED_SETTINGS_A); 
			s_xLobbySettings.m_bAllowMixedTeams = (xAdvSettingsA.m_xAdvancedSettings.m_bAllowMixedTeams == 1); 
			s_xLobbySettings.m_iSpeciesRatioPredator = xAdvSettingsA.m_xAdvancedSettings.m_uSpeciesRatioPredator; 
			s_xLobbySettings.m_iSpeciesRatioHuman = xAdvSettingsA.m_xAdvancedSettings.m_uSpeciesRatioHuman; 
			s_xLobbySettings.m_iSpeciesRatioAlien = xAdvSettingsA.m_xAdvancedSettings.m_uSpeciesRatioAlien; 
			s_xLobbySettings.m_iNumberOfTeams = xAdvSettingsA.m_xAdvancedSettings.m_uNumberOfTeams; 
			s_xLobbySettings.m_iTeamOneSpecies = xAdvSettingsA.m_xAdvancedSettings.m_uTeamOneSpecies; 
			s_xLobbySettings.m_iTeamTwoSpecies = xAdvSettingsA.m_xAdvancedSettings.m_uTeamTwoSpecies; 
			s_xLobbySettings.m_iTeamThreeSpecies = xAdvSettingsA.m_xAdvancedSettings.m_uTeamThreeSpecies; 
			s_xLobbySettings.m_bRespawnUponDeath = xAdvSettingsA.m_xAdvancedSettings.m_bRespawnUponDeath;
			s_xLobbySettings.m_iRespawnType = xAdvSettingsA.m_xAdvancedSettings.m_uRespawnType; 
			s_xLobbySettings.m_iLives = xAdvSettingsA.m_xAdvancedSettings.m_uLives;
			s_xLobbySettings.m_bFriendlyFire = xAdvSettingsA.m_xAdvancedSettings.m_bFriendlyFire; 
			
			AVP_DevLobby_AdvancedSettingsB_Union xAdvSettingsB; 
			xAdvSettingsB.m_uAdvancedSettings = pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_ADVANCED_SETTINGS_B); 
			s_xLobbySettings.m_iScoreLimit = xAdvSettingsB.m_xAdvancedSettings.m_uScoreLimit; 
			s_xLobbySettings.m_iTimeLimit = xAdvSettingsB.m_xAdvancedSettings.m_uTimeLimit; 

			// Update game rules with host settings
			ApplyAdvancedSettings(false); 
		}

		// set lobby requested team to match preferred species. This will need to be more complex at some point.
		if (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_PRE_GAME)
		{
			if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
			{
				if (AVP_Multiplayer_GameRules::GetAllowMixedTeams() == false) // Species based team deathmatch
				{
					if (AVP_Multiplayer_GameRules::GetTeamSpeciesAreMutuallyExclusive())	// No two teams with the same species
					{
						u_char ucPreferredTeam = ASURA_LOBBY_INVALID_TEAM;
						AVP_Settings::Species ePreferredSpecies = static_cast<AVP_Settings::Species>(s_xLobbySettings.m_iLocalSpecies);
						if (ePreferredSpecies == AVP_Multiplayer_GameRules::GetTeamOneSpecies())
						{
							ucPreferredTeam = 0;
						}
						else if (ePreferredSpecies == AVP_Multiplayer_GameRules::GetTeamTwoSpecies())
						{
							ucPreferredTeam = 1;
						}
						else if (ePreferredSpecies == AVP_Multiplayer_GameRules::GetTeamThreeSpecies() && AVP_Multiplayer_GameRules::GetNumberOfTeams() > 2)
						{
							ucPreferredTeam = 2;
						}
						else
						{
							// Our selected species isn't allowed. Let's rotate teams
							ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam();
							ucPreferredTeam = (ucPreferredTeam + 1) % static_cast<u_char>(AVP_Multiplayer_GameRules::GetNumberOfTeams());
							AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
							AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
							// Player species is determined by the team in this case
							AVP_Settings::Species ePlayerSpecies = AVP_Multiplayer_GameRules::GetValidPlayerSpecies(ePreferredSpecies, ePlayerTeam, AVP_Settings::AVP_ROLE_NORMAL, false); 
							s_xLobbySettings.m_iLocalSpecies = static_cast<int>(ePlayerSpecies); 
						}
						Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam);
					}
					else
					{
						// Species settings are ambiguous in this case, player should be assigned to a team
						u_char ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam();
						// Check if the team settings are correct
						if (ucPreferredTeam + 1 > AVP_Multiplayer_GameRules::GetNumberOfTeams()) 
						{
							// Assign a random team
							ucPreferredTeam = static_cast<u_char>(Asura_Random::GetIntU( 0, AVP_Multiplayer_GameRules::GetNumberOfTeams() - 1 ));  
							AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
							AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
							// Player species is determined by the team in this case
							AVP_Settings::Species ePlayerSpecies = AVP_Multiplayer_GameRules::GetValidPlayerSpecies(AVP_Settings::AVP_SPECIES_UNKNOWN, ePlayerTeam, AVP_Settings::AVP_ROLE_NORMAL, false); 
							s_xLobbySettings.m_iLocalSpecies = static_cast<int>(ePlayerSpecies); 
							Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam);
						}
						else
						{
							AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
							// Player species is determined by the team in this case
							AVP_Settings::Species ePlayerSpecies = AVP_Multiplayer_GameRules::GetValidPlayerSpecies(AVP_Settings::AVP_SPECIES_UNKNOWN, ePlayerTeam, AVP_Settings::AVP_ROLE_NORMAL, false); 
							s_xLobbySettings.m_iLocalSpecies = static_cast<int>(ePlayerSpecies); 
						}
					}
				}
				else	// Mixed teams
				{
					// Make sure we pick up a team to start with
					u_char ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam();
					// Check if the team settings are correct
					if (ucPreferredTeam + 1 > AVP_Multiplayer_GameRules::GetNumberOfTeams()) 
					{
						// Assign a random team
						ucPreferredTeam = static_cast<u_char>(Asura_Random::GetIntU( 0, AVP_Multiplayer_GameRules::GetNumberOfTeams() - 1 ));  
						AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
						AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
						Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam);
					}		
				}
			}

			// set lobby requested species
			Asura_Assert(s_xLobbySettings.m_iLocalSpecies >= 0 && s_xLobbySettings.m_iLocalSpecies < AVP_Settings::AVP_NUM_SPECIES, "Invalid species selection"); 
			u_char ucSpecies = static_cast<u_char>(s_xLobbySettings.m_iLocalSpecies);
			Asura_Lobby_System::SetLocalPlayersProjectData(&ucSpecies); 
		}
	}
}

void AVP_DevLobby::InitialiseGameSettings()
{
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	Asura_Assert(pxLobby, "need a valid lobby pointer!");
	if (pxLobby)
	{
		// get lobby attributes
		int iGametype = pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAMETYPE);
		int iLevel = GetMapIndexByHash(pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_MAP));
		Asura_Assert(iLevel < static_cast<int>(s_xMapList.Size()), "invalid level index from lobby data");
		iLevel = Asura_Maths::Min(iLevel, static_cast<int>(s_xMapList.Size() - 1));
		const char* szLevelName = GetMapFilename(iLevel);

		AVP_Settings::SetMaxNetPlayers(pxLobby->GetMaxPlayers());
		AVP_Settings::SetGameType( static_cast<AVP_Settings::GameType>(iGametype) );
		AVP_Multiplayer_GameRules::Refresh();
		ApplyAdvancedSettings(false); 
		AVP_Settings::SetLevelName(szLevelName);

		const Asura_Lobby_Player* pxLocalPlayer = pxLobby->FindLocalPlayer();
		Asura_Assert(pxLocalPlayer, "couldn't get local player from lobby system");
		if (pxLocalPlayer)
		{
			if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
			{
				u_char ucTeam = pxLocalPlayer->GetTeamIndex();
				Asura_Assert(ucTeam < AVP_Multiplayer_GameRules::GetNumberOfTeams(), "invalid team");
				AVP_Settings::SetRequestedMultiplayerTeam(static_cast<AVP_Settings::Team>(ucTeam + 1)); // 0 is TEAM_NONE.
				// Voice chat is directed to our team mates only
				Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY ); 
				Asura_Lobby_System::ConnectLocalPlayerToVoiceChannel(ucTeam + 1); 
				Asura_Lobby_System::SetLocalPlayersActiveVoiceChannel(ucTeam + 1); 
			}
			else
			{
				AVP_Settings::SetRequestedMultiplayerTeam( AVP_Settings::AVP_TEAM_NONE ); 
				// Talk to everyone using the in-game global voice channel
				Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY ); 
				Asura_Lobby_System::ConnectLocalPlayerToVoiceChannel( uAVP_VOICE_CHANNEL_GLOBAL ); 
				Asura_Lobby_System::SetLocalPlayersActiveVoiceChannel( uAVP_VOICE_CHANNEL_GLOBAL ); 
			}
		}
	}

	bool bServer = Asura_Lobby_System::IsHost();
	AVP_Settings::SetIsServer( bServer );
	AVP_Settings::SetExternalNetworkActive(true);

	if (!bServer)
	{
		if (Asura_Lobby_System::GetGameAddress())
		{
			AVP_Settings::SetAddressToConnectTo( Asura_Lobby_System::GetGameAddress() );
		}
	}

	// TODO: the way these are set will change
	AVP_Settings::SetPlayerSpecies(static_cast<AVP_Settings::Species>(s_xLobbySettings.m_iLocalSpecies));
	AVP_Game_Core::SetRespawnSpecies(static_cast<AVP_Settings::Species>(s_xLobbySettings.m_iLocalSpecies));
}

void AVP_DevLobby::UpdateTeamNames()
{
	if (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_PRE_GAME)
	{
		if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
		{
			const Asura_Hash_ID uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
#if 0
			if (AVP_Multiplayer_GameRules::GetAllowMixedTeams())
			{
				// teams are team 1,2,3
				Asura_Lobby_System::SetTeamName(0, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920974) );// "team1"
				Asura_Lobby_System::SetTeamName(1, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920975) );// "team2"
				Asura_Lobby_System::SetTeamName(2, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920976) );// "team3"
			}
			else
			{
				// teams are per species
				Asura_Lobby_System::SetTeamName(0, Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamOneSpecies())));
				Asura_Lobby_System::SetTeamName(1, Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamTwoSpecies())));
				Asura_Lobby_System::SetTeamName(2, Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamThreeSpecies())));
			}
#else
			// HACK: to see slots available per team
			static Asura_UnicodeChar wszTeam1NameAndSlots[128]; 
			static Asura_UnicodeChar wszTeam2NameAndSlots[128]; 
			static Asura_UnicodeChar wszTeam3NameAndSlots[128]; 
		
			if (AVP_Multiplayer_GameRules::GetAllowMixedTeams())
			{
				// teams are team 1,2,3
				Asura_PrintfToBfr( wszTeam1NameAndSlots, L"%s [%d:%d:%d]", 128, true, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920974), 
					CalculateAvailableSpeciesSlotsInTeam( 0, 1 ),
					CalculateAvailableSpeciesSlotsInTeam( 2, 1 ),
					CalculateAvailableSpeciesSlotsInTeam( 1, 1 ));
				Asura_Lobby_System::SetTeamName(0, wszTeam1NameAndSlots );// "team1"
				Asura_PrintfToBfr( wszTeam2NameAndSlots, L"%s [%d:%d:%d]", 128, true, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920975), 
					CalculateAvailableSpeciesSlotsInTeam( 0, 2 ),
					CalculateAvailableSpeciesSlotsInTeam( 2, 2 ),
					CalculateAvailableSpeciesSlotsInTeam( 1, 2 ));
				Asura_Lobby_System::SetTeamName(1, wszTeam2NameAndSlots );// "team2"
				Asura_PrintfToBfr( wszTeam3NameAndSlots, L"%s [%d:%d:%d]", 128, true, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920976), 
					CalculateAvailableSpeciesSlotsInTeam( 0, 3 ),
					CalculateAvailableSpeciesSlotsInTeam( 2, 3 ),
					CalculateAvailableSpeciesSlotsInTeam( 1, 3 ));
				Asura_Lobby_System::SetTeamName(2, wszTeam3NameAndSlots );// "team3"
			}
			else
			{
				// teams are per species
				Asura_PrintfToBfr( wszTeam1NameAndSlots, L"%s [%d]", 128, true,  Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamOneSpecies())), 
					CalculateAvailableSpeciesSlotsInTeam( static_cast<int>(AVP_Multiplayer_GameRules::GetTeamOneSpecies()), 1 ));
				Asura_Lobby_System::SetTeamName(0, wszTeam1NameAndSlots );
				Asura_PrintfToBfr( wszTeam2NameAndSlots, L"%s [%d]", 128, true,  Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamTwoSpecies())), 
					CalculateAvailableSpeciesSlotsInTeam( static_cast<int>(AVP_Multiplayer_GameRules::GetTeamTwoSpecies()), 2 ));
				Asura_Lobby_System::SetTeamName(1, wszTeam2NameAndSlots );
				Asura_PrintfToBfr( wszTeam3NameAndSlots, L"%s [%d]", 128, true,  Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamThreeSpecies())), 
					CalculateAvailableSpeciesSlotsInTeam( static_cast<int>(AVP_Multiplayer_GameRules::GetTeamThreeSpecies()), 3 ));
				Asura_Lobby_System::SetTeamName(2, wszTeam3NameAndSlots);
			}
#endif
		}
	}
}

void AVP_DevLobby::SetLobbyVoiceChannelsSettings()
{
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_GLOBAL ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_TEAM1 ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_TEAM2 ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_TEAM3 ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_DEATH ); 
	Asura_Lobby_System::ConnectLocalPlayerToVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY );
	Asura_Lobby_System::SetLocalPlayersActiveVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY ); 
}

void AVP_DevLobby::CreateMapList()
{
	Asura_Assert(s_xMapList.Size() == 0, "Map list already created.");
	Asura_FileAccess::FindFiles(szMULTIPLAYER_MAP_PATH, AVP_EXPORTED_LEVEL_FILE_EXTENSION, CreateMapList_Callback);

	// Remove the console-variables related to the number of maps, and re-add them to make sure that the maximum is correct...
	if (Asura_CommandConsole::VarExists(L"AVP.DevLobby.MapIndex"))			Asura_CommandConsole::RemoveVar(L"AVP.DevLobby.MapIndex");
	if (Asura_CommandConsole::VarExists(L"AVP.DevLobby.SearchFilter.Map"))	Asura_CommandConsole::RemoveVar(L"AVP.DevLobby.SearchFilter.Map");

	Asura_CommandConsole::AddVar( L"AVP.DevLobby.MapIndex",					&s_xLobbySettings.m_iMapIndex,						0,									Asura_Maths::Max(static_cast<int>(s_xMapList.Size()) - 1, 0),	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.DevLobby.SearchFilter.Map",			&s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP],	iAVP_LOBBY_INVALID_SEARCH_FILTER,	s_xMapList.Size(),												Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}

void AVP_DevLobby::CreateMapList_Callback(const char* szFilename)
{
	AVP_DevLobby_MapData* pxData = new AVP_DevLobby_MapData(szFilename);
	if (pxData)
	{
		s_xMapList.PushBack(pxData);
	}
}

void AVP_DevLobby::ClearMapList()
{
	for (Asura_LinkList_It<AVP_DevLobby_MapData*> xIt(s_xMapList); !xIt.Done();)
	{
		// Delete the data and remove the entry from the list.
		AVP_DevLobby_MapData** ppxData = xIt.GetCurrent();
		AVP_DevLobby_MapData* pxData = ppxData ? *ppxData : NULL;

		xIt.DeleteCurrent();
		delete pxData;
	}
}

int AVP_DevLobby::GetMapIndexByHash(Asura_Hash_ID uHash)
{
	int iIndex = 0;
	for (Asura_LinkList_It<AVP_DevLobby_MapData*> xIt(s_xMapList); !xIt.Done(); xIt.Next())
	{
		AVP_DevLobby_MapData** ppxData = xIt.GetCurrent();
		if (ppxData && *ppxData && (*ppxData)->GetHash() == uHash)
		{
			// Found it, index is already correct.
			break;
		}
		else
		{
			++iIndex;
		}
	}
	return iIndex; // Intentionally returning one greater than the maximum valid index if we don't have the map, as that allows us to display unknown.
}

Asura_Hash_ID AVP_DevLobby::GetMapHashByIndex(int iIndex)
{
	const AVP_DevLobby_MapData* pxData = NULL;
	if (iIndex < static_cast<int>(s_xMapList.Size()))
	{
		const Asura_ListElt<AVP_DevLobby_MapData*>* pxElement = s_xMapList[iIndex];
		if (pxElement)
		{
			pxData = pxElement->Data();
		}
	}

	if (pxData)
	{
		return pxData->GetHash();
	}
	else
	{
		return ASURA_HASH_ID_UNSET;
	}
}

const char* AVP_DevLobby::GetMapFilename(int iIndex)
{
	const AVP_DevLobby_MapData* pxData = NULL;
	if (iIndex < static_cast<int>(s_xMapList.Size()))
	{
		const Asura_ListElt<AVP_DevLobby_MapData*>* pxElement = s_xMapList[iIndex];
		if (pxElement)
		{
			pxData = pxElement->Data();
		}
	}

	if (pxData)
	{
		return pxData->GetFilename();
	}
	else
	{
		Asura_Assert(false, "Failed to get map filename.");
		return "";
	}
}

bool AVP_DevLobby::CreateGame( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Lobby_Attributes xAttributes;
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_MAP, GetMapHashByIndex(s_xLobbySettings.m_iMapIndex));
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAMETYPE, s_xLobbySettings.m_iGametypeIndex);
		// now fill out the game version. We've encoded game version and ranked status in here.
		AVP_DevLobby_GameVer_Union xGameVer;
		xGameVer.m_xGameVer.m_uVersion = Asura_NetTransport_System::Project_GetVersionID();
		xGameVer.m_xGameVer.m_uFlags = Asura_Lobby_System::IsUsingRankedLobbies() ? 1 : 0;
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAME_VERSION, xGameVer.m_iGameVer);

		// set the "advanced settings". These are for information rather than for searching.
		AVP_DevLobby_AdvancedSettingsA_Union xAdvSettingsA; 
		xAdvSettingsA.m_xAdvancedSettings.m_bAllowMixedTeams = s_xLobbySettings.m_bAllowMixedTeams; 
		xAdvSettingsA.m_xAdvancedSettings.m_uNumberOfTeams = s_xLobbySettings.m_iNumberOfTeams;  
		xAdvSettingsA.m_xAdvancedSettings.m_uSpeciesRatioPredator = s_xLobbySettings.m_iSpeciesRatioPredator; 
		xAdvSettingsA.m_xAdvancedSettings.m_uSpeciesRatioHuman = s_xLobbySettings.m_iSpeciesRatioHuman; 
		xAdvSettingsA.m_xAdvancedSettings.m_uSpeciesRatioAlien = s_xLobbySettings.m_iSpeciesRatioAlien; 
		xAdvSettingsA.m_xAdvancedSettings.m_uTeamOneSpecies = s_xLobbySettings.m_iTeamOneSpecies;
		xAdvSettingsA.m_xAdvancedSettings.m_uTeamTwoSpecies = s_xLobbySettings.m_iTeamTwoSpecies;
		xAdvSettingsA.m_xAdvancedSettings.m_uTeamThreeSpecies = s_xLobbySettings.m_iTeamThreeSpecies;
		xAdvSettingsA.m_xAdvancedSettings.m_bRespawnUponDeath = s_xLobbySettings.m_bRespawnUponDeath;
		xAdvSettingsA.m_xAdvancedSettings.m_uRespawnType = s_xLobbySettings.m_iRespawnType; 
		xAdvSettingsA.m_xAdvancedSettings.m_uLives = s_xLobbySettings.m_iLives;
		xAdvSettingsA.m_xAdvancedSettings.m_bFriendlyFire = s_xLobbySettings.m_bFriendlyFire;

		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_ADVANCED_SETTINGS_A, xAdvSettingsA.m_uAdvancedSettings); 

		AVP_DevLobby_AdvancedSettingsB_Union xAdvSettingsB;
		xAdvSettingsB.m_xAdvancedSettings.m_uScoreLimit = s_xLobbySettings.m_iScoreLimit;
		xAdvSettingsB.m_xAdvancedSettings.m_uTimeLimit = s_xLobbySettings.m_iTimeLimit; 
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_ADVANCED_SETTINGS_B, xAdvSettingsB.m_uAdvancedSettings);

		// set the lobby skill to be the host's skill
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_SKILL, Asura_Lobby_System::GetLocalPlayerSkill());

		// initialise the local game settings
		AlterGameType(false); 
		ApplyAdvancedSettings(false);

		int iNumTeams = AVP_Multiplayer_GameRules::GetNumberOfTeams();
		Asura_Assert(iNumTeams > 0 && iNumTeams <= UCHAR_MAX, "unexpected num teams");

		Asura_Lobby_System::CreateLobby( s_xLobbySettings.m_uMaxPlayers, static_cast<u_char>(iNumTeams), xAttributes );
		
	}
	return true;
}

void AVP_DevLobby::CreateGameForAutomatch()
{
	// called from automatic matchmaking.
	// set options first - use the search filter for gametype.
	// If it was not set pick one at random
	int iGametype = s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAMETYPE];
	if (iGametype == iAVP_LOBBY_INVALID_SEARCH_FILTER)
	{
		iGametype = Asura_Random::GetInt(0, AVP_Settings::AVP_NUM_GAMETYPES - 1);
	}
	s_xLobbySettings.m_iGametypeIndex = iGametype;
	// ensure the gametype is allowed
	CheckGametypeIsAllowed();
	s_iPrevGametype = s_xLobbySettings.m_iGametypeIndex;

	// create the game using the console command
	CreateGame(false);
}

void AVP_DevLobby::SetSearchFilters( u_int uSkillThreshold, Asura_Lobby_SearchCriterion* axFilters, u_int& uNumFilters )
{
	u_int uNumSearchCriteria = 0;

	// first fill out the game version. We've encoded game version and ranked status in here.
	AVP_DevLobby_GameVer_Union xGameVer;
	xGameVer.m_xGameVer.m_uVersion = Asura_NetTransport_System::Project_GetVersionID();
	xGameVer.m_xGameVer.m_uFlags = Asura_Lobby_System::IsUsingRankedLobbies() ? 1 : 0;
	s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAME_VERSION] = xGameVer.m_iGameVer;

	for (u_int u = 0; u < AVP_NUM_LOBBY_ATTRIBUTES; ++u)
	{
		if (s_aiLobbySearchFilters[u] != iAVP_LOBBY_INVALID_SEARCH_FILTER)
		{
			if (u == AVP_LOBBY_ATTRIBUTE_MAP)
			{
				// Special case for the map - we need to send the hash id instead of the index, and ignore the value if it's at the limit.
				if (s_aiLobbySearchFilters[u] != static_cast<int>(s_xMapList.Size()))
				{
					u_int uFilter = uNumSearchCriteria++;
					axFilters[uFilter].Set(u, ASURA_LOBBY_SEARCH_OPERATOR_EQ, GetMapHashByIndex(s_aiLobbySearchFilters[u]));
				}
			}
			else
			{
				u_int uFilter = uNumSearchCriteria++;
				axFilters[uFilter].Set(u, ASURA_LOBBY_SEARCH_OPERATOR_EQ, s_aiLobbySearchFilters[u]);
			}
		}
	}

#ifdef ASURA_USE_MERIT_SYSTEM
	// now set the skill filters: (skill > myskill - skillthreshold) && (skill < myskill + skillthreshold)
	const int iMySkill = Asura_Lobby_System::GetLocalPlayerSkill();
	const int iMinSkill = iMySkill - static_cast<int>(uSkillThreshold);
	const int iMaxSkill = iMySkill + static_cast<int>(uSkillThreshold);
	axFilters[uNumSearchCriteria++].Set(AVP_LOBBY_ATTRIBUTE_SKILL, ASURA_LOBBY_SEARCH_OPERATOR_GE, iMinSkill);
	axFilters[uNumSearchCriteria++].Set(AVP_LOBBY_ATTRIBUTE_SKILL, ASURA_LOBBY_SEARCH_OPERATOR_LE, iMaxSkill);
#endif

	uNumFilters = uNumSearchCriteria;
}

bool AVP_DevLobby::SearchForGame( bool bHelp )
{
	if (!bHelp)
	{
		// create the search filters

		// first fill out the game version. We've encoded game version and ranked status in here.
		AVP_DevLobby_GameVer_Union xGameVer;
		xGameVer.m_xGameVer.m_uVersion = Asura_NetTransport_System::Project_GetVersionID();
		xGameVer.m_xGameVer.m_uFlags = Asura_Lobby_System::IsUsingRankedLobbies() ? 1 : 0;
		s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAME_VERSION] = xGameVer.m_iGameVer;

		Asura_Lobby_SearchCriterion axSearchCriteria[AVP_NUM_LOBBY_ATTRIBUTES];
		u_int uNumSearchCriteria = 0;
		SetSearchFilters(uAVP_DEVLOBBY_MAX_SKILL_THRESHOLD, axSearchCriteria, uNumSearchCriteria);
		Asura_Assert(uNumSearchCriteria <= AVP_NUM_LOBBY_ATTRIBUTES, "array overwrite");

		// start the search
		s_bAutoJoinOnSearchComplete = true;
		Asura_Lobby_System::SearchForLobby(&axSearchCriteria[0], uNumSearchCriteria);
	}
	return true;
}

bool AVP_DevLobby::BrowseForGame( bool bHelp )
{
	if (!bHelp)
	{
		// create the search filters
		Asura_Assert(!Asura_Lobby_System::IsUsingRankedLobbies(),"Can't browse ranked games");

		if (!Asura_Lobby_System::IsUsingRankedLobbies())
		{
			// first fill out the game version. We've encoded game version and ranked status in here.
			AVP_DevLobby_GameVer_Union xGameVer;
			xGameVer.m_xGameVer.m_uVersion = Asura_NetTransport_System::Project_GetVersionID();
			xGameVer.m_xGameVer.m_uFlags = 0;

			s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAME_VERSION]	= xGameVer.m_iGameVer;
			s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP]				= iAVP_LOBBY_INVALID_SEARCH_FILTER;
			s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAMETYPE]		= iAVP_LOBBY_INVALID_SEARCH_FILTER;

			Asura_Lobby_SearchCriterion axSearchCriteria[AVP_NUM_LOBBY_ATTRIBUTES];
			u_int uNumSearchCriteria = 0;
			SetSearchFilters(uAVP_DEVLOBBY_MAX_SKILL_THRESHOLD, axSearchCriteria, uNumSearchCriteria);
			Asura_Assert(uNumSearchCriteria <= AVP_NUM_LOBBY_ATTRIBUTES, "array overwrite");

			// start the search
			s_bAutoJoinOnSearchComplete = false;
			Asura_Lobby_System::SearchForLobby(&axSearchCriteria[0], uNumSearchCriteria);
		}
	}
	return true;
}

bool AVP_DevLobby::JoinGame( u_int uGameIndex, bool bInterrogate, bool bHelp )
{
	if (bHelp || bInterrogate)
	{

	}
	else
	{
		Asura_Assert( uGameIndex < Asura_Lobby_System::GetNumSearchResults(), "Invalid game index");
		if ((Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID)
			&& (Asura_Lobby_System::GetSearchState() == ASURA_LOBBY_SEARCHSTATE_SEARCH_COMPLETE)
			&& (uGameIndex < Asura_Lobby_System::GetNumSearchResults()))
		{
			Asura_Lobby_System::JoinLobby( uGameIndex );
		}
	}

	return true;
}

bool AVP_DevLobby::ListGame( u_int uGameIndex, bool bInterrogate, bool bHelp )
{
	if (bHelp || bInterrogate)
	{

	}
	else
	{
		if ((Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID)
			&& (Asura_Lobby_System::GetSearchState() == ASURA_LOBBY_SEARCHSTATE_SEARCH_COMPLETE))
		{
			if (uGameIndex < Asura_Lobby_System::GetNumSearchResults())
			{
				const Asura_Lobby_ListEntry& xGame = Asura_Lobby_System::GetSearchResultByIndex( uGameIndex );
				Asura_CommandConsole::Printf( "%S (%d/%d)", xGame.m_awcDescription, xGame.m_uNumPlayers, xGame.m_uMaxPlayers );

				return true;
			}
		}
	}

	return false;
}

bool AVP_DevLobby::StartGame( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Lobby_System::StartGame();
	}
	return true;
}

bool AVP_DevLobby::ChangeSpecies( bool bHelp )
{
	if (!bHelp)
	{
		s_xLobbySettings.m_iLocalSpecies = (s_xLobbySettings.m_iLocalSpecies + 1) % AVP_Settings::AVP_NUM_SPECIES;
		Asura_Assert(s_xLobbySettings.m_iLocalSpecies >= 0 && s_xLobbySettings.m_iLocalSpecies < AVP_Settings::AVP_NUM_SPECIES, "Invalid species selection"); 
		u_char ucSpecies = static_cast<u_char>(s_xLobbySettings.m_iLocalSpecies);
		Asura_Lobby_System::SetLocalPlayersProjectData(&ucSpecies); 
	}
	return true;
}

bool AVP_DevLobby::ChangeTeam( bool bHelp )
{
	if (!bHelp)
	{
		u_char ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam(); 
		ucPreferredTeam = (ucPreferredTeam + 1) % static_cast<u_char>(AVP_Multiplayer_GameRules::GetNumberOfTeams());
		Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam); 
		AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
		AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
	} 
	return true; 
} 

bool AVP_DevLobby::AlterGameType( bool bHelp )
{
	if (!bHelp)
	{
		AVP_Settings::GameType eNewGameType = static_cast<AVP_Settings::GameType>(s_xLobbySettings.m_iGametypeIndex); 
		if (eNewGameType != AVP_Settings::GetGameType())
		{
			AVP_Settings::SetGameType( eNewGameType );
			AVP_Multiplayer_GameRules::Refresh(); 
		}
	}
	return true;
}

bool AVP_DevLobby::ApplyAdvancedSettings( bool bHelp )
{
	// JG - 05 March 2009 - Commented this lot out until we decide when / if this settings are exposed to the host
	//if (!bHelp)
	//{
	//	AVP_Multiplayer_GameRules::SetAllowMixedTeams( s_xLobbySettings.m_bAllowMixedTeams ); 
	//	AVP_Multiplayer_GameRules::SetPlayerRatioPredator( s_xLobbySettings.m_iSpeciesRatioPredator ); 
	//	AVP_Multiplayer_GameRules::SetPlayerRatioHuman( s_xLobbySettings.m_iSpeciesRatioHuman ); 
	//	AVP_Multiplayer_GameRules::SetPlayerRatioAlien( s_xLobbySettings.m_iSpeciesRatioAlien ); 
	//	AVP_Multiplayer_GameRules::SetNumberOfTeams( s_xLobbySettings.m_iNumberOfTeams ); 
	//	AVP_Multiplayer_GameRules::SetTeamOneSpecies( static_cast<AVP_Settings::Species>( s_xLobbySettings.m_iTeamOneSpecies )); 
	//	AVP_Multiplayer_GameRules::SetTeamTwoSpecies( static_cast<AVP_Settings::Species>( s_xLobbySettings.m_iTeamTwoSpecies )); 
	//	AVP_Multiplayer_GameRules::SetTeamThreeSpecies( static_cast<AVP_Settings::Species>( s_xLobbySettings.m_iTeamThreeSpecies )); 
	//	AVP_Multiplayer_GameRules::SetRespawnUponDeath( s_xLobbySettings.m_bRespawnUponDeath ); 
	//	//AVP_Multiplayer_GameRules::SetRespawnType( static_cast<AVP_Settings::RespawnType>( s_xLobbySettings.m_iRespawnType )); 
	//	AVP_Multiplayer_GameRules::SetLivesAvailablePerTeam( s_xLobbySettings.m_iLives ); 
	//	AVP_Multiplayer_GameRules::SetFriendlyFireOn( s_xLobbySettings.m_bFriendlyFire ); 
	//	AVP_Multiplayer_GameRules::SetPlayerScoreLimit( s_xLobbySettings.m_iScoreLimit ); 
	//	AVP_Multiplayer_GameRules::SetTimeLimit( s_xLobbySettings.m_iTimeLimit );
	//}
	return true; 
}

bool AVP_DevLobby::PopulateMapSelectorName(u_int uElementIndex, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		return true;
	}
	else
	{
		return PopulateMapName(uElementIndex, false, false);
	}
}

bool AVP_DevLobby::PopulateMapSearchName(u_int uElementIndex, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		return true;
	}
	else
	{
		return PopulateMapName(uElementIndex, true, false);
	}
}

bool AVP_DevLobby::PopulateMapDisplayName(u_int uElementIndex, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		return true;
	}
	else
	{
		return PopulateMapName(uElementIndex, false, true);
	}
}

bool AVP_DevLobby::PopulateMapName(u_int uElementIndex, bool bIncludeAny, bool bIncludeUnknown)
{
	const AVP_DevLobby_MapData* pxData = NULL;
	if (uElementIndex < s_xMapList.Size())
	{
		const Asura_ListElt<AVP_DevLobby_MapData*>* pxElement = s_xMapList[uElementIndex];
		if (pxElement)
		{
			pxData = pxElement->Data();
		}
	}

	if (pxData)
	{
		Asura_CommandConsole::Printf(L"%S", pxData->GetDisplayName());
		return true;
	}
	else if (uElementIndex == s_xMapList.Size())
	{
		if (bIncludeAny)
		{
			// TODO - Localise this.
			Asura_CommandConsole::Printf(L"Any");
			return true;
		}
		else if (bIncludeUnknown)
		{
			// TODO - Localise this.
			Asura_CommandConsole::Printf(L"<UNKNOWN>");
			return true;
		}
	}
	return false;
}

bool AVP_DevLobby::ConsoleCmd_AllowChangingTeam( bool bHelp )
{
	if (!bHelp)
	{
		Asura_CommandConsole::Printf( AllowChangingTeam() ? "1" : "0" );
	}
	return true;
}

bool AVP_DevLobby::ConsoleCmd_AllowChangingSpecies( bool bHelp )
{
	if (!bHelp)
	{
		Asura_CommandConsole::Printf( AllowChangingSpecies() ? "1" : "0" );
	}
	return true;
}

bool AVP_DevLobby::ConsoleCmd_IsRoleBasedGame( bool bHelp )
{
	if (!bHelp)
	{
		Asura_CommandConsole::Printf( AVP_Multiplayer_GameRules::IsRoleBasedGame() ? "1" : "0" );
	}
	return true;
}

bool AVP_DevLobby::AllowChangingTeam()
{
	if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
	{
		if (AVP_Multiplayer_GameRules::GetAllowMixedTeams())
		{
			return true; 
		}
		else
		{
			if (!AVP_Multiplayer_GameRules::GetTeamSpeciesAreMutuallyExclusive())
			{
				return true;
			}
		}
	}
	return false; 
}

bool AVP_DevLobby::AllowChangingSpecies()
{
	if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
	{
		if (!AVP_Multiplayer_GameRules::GetAllowMixedTeams())
		{
			if (!AVP_Multiplayer_GameRules::GetTeamSpeciesAreMutuallyExclusive())
			{
				return false;
			}
		}
	}

	// In role based games players don't select their species
	if (AVP_Multiplayer_GameRules::IsRoleBasedGame())
	{
		return false;
	}

	return true; 
}

int AVP_DevLobby::CalculateAvailableSpeciesSlotsInTeam( int iSpecies, int iTeam )
{
	// Sanity checks
	Asura_Assert( iSpecies >= 0 && iSpecies <= 2, "Invalid Species"); 
	Asura_Assert( iTeam >= 1 && iTeam <= 3, "Invalid Team"); 

	AVP_Settings::Species eSpecies = static_cast<AVP_Settings::Species>(iSpecies); 
	AVP_Settings::Team eTeam = static_cast<AVP_Settings::Team>(iTeam); 
	
	int iMaxNumberOfPlayers = s_xLobbySettings.m_uMaxPlayers; 
	int iNumberOfTeams = s_xLobbySettings.m_iNumberOfTeams; 
	int iMaxNumberOfPlayersInTeam = AVP_Multiplayer_GameRules::CalculateMaxNumberOfPlayersInTeam( iMaxNumberOfPlayers, iNumberOfTeams, eTeam );

	int iTotalNumberOfSpeciesSlots = iMaxNumberOfPlayersInTeam; 
	if (AVP_Multiplayer_GameRules::GetAllowMixedTeams())
	{
		iTotalNumberOfSpeciesSlots = AVP_Multiplayer_GameRules::CalculateMaxNumberOfPlayerSpeciesInTeam( iMaxNumberOfPlayersInTeam, eSpecies, eTeam ); 
	}
	else
	{
		iTotalNumberOfSpeciesSlots = AVP_Multiplayer_GameRules::CalculateMaxNumberOfPlayerSpeciesInTeam( iMaxNumberOfPlayers, eSpecies, eTeam ); 
	}

	// Now let's see how many of these have been used
	if (iTotalNumberOfSpeciesSlots > 0)
	{
		const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
		if (pxLobby)
		{
			u_char ucTeam = static_cast<u_char>(iTeam) - 1; 
			u_int uActualNumberOfPlayersInTeam = pxLobby->GetNumPlayersForTeam( ucTeam ); 
			for (u_int uIndex = 0; uIndex < uActualNumberOfPlayersInTeam; ++uIndex)
			{
				const Asura_Lobby_Player& xPlayer = pxLobby->GetTeamPlayerByIndex(ucTeam, uIndex); 
				u_char ucSpecies = *(xPlayer.GetProjectData()); 
				if (iSpecies == ucSpecies)
				{
					// Slot taken
					iTotalNumberOfSpeciesSlots =  iTotalNumberOfSpeciesSlots - 1; 
				}
			}
		}
	}

	Asura_Assert( iTotalNumberOfSpeciesSlots >= 0, "Invalid number of slots" ); 
	return iTotalNumberOfSpeciesSlots; 
}


void AVP_DevLobby::Render()
{
	// Render black quad over entire screen.
	Asura_Render::Overlay.Colour = 0xff000000;
	Asura_Render::Overlay.TextureIndex = -1;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
	Asura_Render::Overlay.SetRectXY(0.0f, 0.0f, static_cast<float>(Asura_Viewport::GetWidth()), static_cast<float>(Asura_Viewport::GetHeight()));
	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();

	Asura_GUIMenu_System::Render();
}


bool AVP_DevLobby::IsGametypeAllowed(int iGametype)
{
	// this is basically a hack to allow us to restrict gamemodes for milestones.
	switch (iGametype)
	{
		case AVP_Settings::AVP_GAMETYPE_DEATH_MATCH:
		case AVP_Settings::AVP_GAMETYPE_SPECIES_TEAM_DEATH_MATCH:
		//case AVP_Settings::AVP_GAMETYPE_DISRUPTION_DOMINATION:
		case AVP_Settings::AVP_GAMETYPE_HUNT:
		case AVP_Settings::AVP_GAMETYPE_SURVIVOR:
		case AVP_Settings::AVP_GAMETYPE_MIXED_SPECIES_TEAM_DEATH_MATCH:
		case AVP_Settings::AVP_GAMETYPE_DOMINATION:
			// allowed
			return true;
		default:
			// not allowed
			return false;
	}

}
void AVP_DevLobby::CheckGametypeIsAllowed()
{
	// ensure that if gametype was modified by console that it is a valid value
	// this is basically a hack to allow us to restrict gamemodes for milestones.
	int iIncrement = s_xLobbySettings.m_iGametypeIndex > s_iPrevGametype ? 1 : -1;
	while (!IsGametypeAllowed(s_xLobbySettings.m_iGametypeIndex))
	{
		if (s_xLobbySettings.m_iGametypeIndex >= AVP_Settings::AVP_NUM_GAMETYPES - 1)
		{
			iIncrement = -1;
		}
		if (s_xLobbySettings.m_iGametypeIndex <= 0)
		{
			iIncrement = 1;
		}
		s_xLobbySettings.m_iGametypeIndex += iIncrement;
	}
}

// AVP_DevLobby_CreateGameSettings
AVP_DevLobby_GameSettings::AVP_DevLobby_GameSettings()
{
	// set up default values
	m_uMaxPlayers = 18;
	m_iMapIndex =  0;
	m_iGametypeIndex = AVP_Settings::AVP_GAMETYPE_DEATH_MATCH;
	m_iLocalSpecies = static_cast<int>(Asura_Random::GetSeededIntU() % AVP_Settings::AVP_NUM_SPECIES);
	// TODO: this advanced settings should be a copy of the default settings in the multiplayer rules. 
	// I'm setting them manually here to keep the same behaviour that we had in multiplayer showcase build
	// (3 teams, species based, in this particular team order), but in the future will copy blueprint data here
	m_bAllowMixedTeams = false; 
	m_iSpeciesRatioPredator = 1; 
	m_iSpeciesRatioHuman = 2;
	m_iSpeciesRatioAlien = 4;
	m_iNumberOfTeams = 3;
	m_iTeamOneSpecies = AVP_Settings::AVP_SPECIES_ALIEN; 
	m_iTeamTwoSpecies = AVP_Settings::AVP_SPECIES_HUMAN; 
	m_iTeamThreeSpecies = AVP_Settings::AVP_SPECIES_PREDATOR;
	m_bRespawnUponDeath = true; 
	m_iRespawnType = 0; // This is highly game type dependent. 
	m_iLives = 0;	// Infinite lives
	m_bFriendlyFire = false; 
	m_iScoreLimit = 20;
	m_iTimeLimit = 10;
}

AVP_DevLobby_MapData::AVP_DevLobby_MapData(const char* szFilename)
{
	Asura_PrintfToBfr(m_acFilename, szFilename, 256, true);
	m_uHash = Asura_GetHashID(szFilename);
	// This is messy, but since we'll have to replace this with a localised string at some point anyway it doesn't seem a major issue right now.
	Asura_String xFilename(szFilename);
	Asura_String xName;
	xFilename.SplitFilename(NULL, &xName, NULL);
	if (xName.StartsWith("MP_"))
	{
		xName = xName.SubString(3, xName.GetLength() - 3);
	}
	Asura_PrintfToBfr(m_acDisplayName, xName.GetString(), 32, true);
}

// Asura platform implementations, if this is not already defined in the game lobby define it here. 
#if !( defined( AVP_USE_GAME_LOBBY ) )

// ASURA CODE SHARED BY GAME AND DEV LOBBIES /////////////////////////////////
void Asura_Lobby_System::Project_PrintErrorCodeToConsole()
{
	const Asura_Hash_ID uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
	u_int uStringHash = ASURA_HASH_ID_UNSET;
	switch (s_iNetworkDisconnectReason)
	{
		case ASURA_NDR_TIMEOUT:
		{
			uStringHash = 0x05353E2A;//Asura_GetHashID("Error_Timeout")
			break;
		}
		case ASURA_NDR_SERVERQUIT:
		{
			uStringHash = 0x0B5B2C4E;//Asura_GetHashID("Error_HostQuit")
			break;
		}
		// PS3 error codes copied from SDK documentation.
		case 0x80022b13:
		{
			uStringHash = 0x7988D3B1; // Asura_GetHashID("Error_PS3RoomDoesNotExist")
			break;
		}
		case 0x80022b08:
		{
			uStringHash = 0x4E828A48; // Asura_GetHashID("Error_PS3UserBlocked")
			break;
		}
		case 0x80022b19:
		{
			uStringHash = 0xAA8044C3; // Asura_GetHashID("Error_PS3RoomFull")
			break;
		}
		default:
		{
			uStringHash = 0x8CE918B9;// Asura_GetHashID("Error_Code_Label")
			const Asura_UnicodeChar* wszString = Asura_HashedLocalisedText::GetString(uPageHash, uStringHash);
			Asura_CommandConsole::Printf( L"%s 0x%x", wszString, s_iNetworkDisconnectReason );
			return;
		}
	}

	const Asura_UnicodeChar* wszString = Asura_HashedLocalisedText::GetString(uPageHash, uStringHash);
	Asura_CommandConsole::Printf( L"%s", wszString );
}

void Asura_Lobby_System::Project_GetInviteText( const Asura_UnicodeChar*& wszSubject, const Asura_UnicodeChar*& wszBody )
{
	const Asura_Hash_ID uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
	const Asura_Hash_ID uStringHashSubject = 0x0B7511F6; // Asura_GetHashID("Invite_Subject")
	const Asura_Hash_ID uStringHashBody = 0xA657C118; // Asura_GetHashID("Invite_Body")
	wszSubject = Asura_HashedLocalisedText::GetString(uPageHash, uStringHashSubject);
	wszBody = Asura_HashedLocalisedText::GetString(uPageHash, uStringHashBody);
}
#endif //#if !( defined( AVP_USE_GAME_LOBBY ) )

#endif	//defined(AVP_USE_DEV_LOBBY)

// E N D   O F   F I L E /////////////////////////////////////////////////////
