
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Inventory_Item.cpp
// Author : Andy Tate
// Date   : 25-Oct-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Inventory/AVP_Inventory_Item.h"

// Asura independent header-files...
#include "Animation/Asura_AnimationAttachment.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"

// Game platform header-files...
#include "Actor/AVP_Actor_AttachmentHandler.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Moves/AVP_Movement_System.h"
#include "Inventory/AVP_AmmoStore.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_InventoryItem_Messages.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/AVP_InventoryItem_DynamicSubType.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestStateChange
// Author       : Andy Tate
// Date         : 25-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestStateChange(const State eNewState)
{
	if(IsStateValid(eNewState))
	{
		m_eNewState		= eNewState;
		m_bHasNewState	= true;
		return(true);
	}
	else
	{
		Asura_Assert(false, "Cannot change state - the specified state is not valid for this inventory-item.");
		return(false);
	}
}

/// End of function 'AVP_Inventory_Item::RequestStateChange'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::OnStateChange
// Author       : Andy Tate
// Date         : 26-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::OnStateChange(const u_int uOldState,
									   const u_int uNewState)
{
	switch(uOldState)
	{
		case INACTIVE:
		{
			SetEntityActiveFlag(true);
			break;
		}
	}		// End of switch.

	switch(uNewState)
	{
		case INACTIVE:
		{
			SetEntityActiveFlag(false);
			break;
		}

		case IDLING:
		{
			m_uCurrentUsageType	= UT_NONE;

			if(m_uPendingUsageType != UT_NONE)
			{
				RequestUsing(IDLING, m_uPendingUsageType, m_uPendingTransitionType);
			}
			break;
		}

		default:
		{
			break;
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::OnStateChange'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::GetSubTypeIndexFromUsageType
// Author       : Andy Tate
// Date         : 12-Dec-2007
////////////////////////////////////////////////////////////////////////////////

int AVP_Inventory_Item::GetSubTypeIndexFromUsageType(const u_int uUsageType) const
{
	Asura_Assert((m_ppxSubTypes == NULL), "This inventory-item has sub-types but no mapping to usage-types.");
	return(-1);
}

/// End of function 'AVP_Inventory_Item::GetSubTypeIndexFromUsageType'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::CreateAbortMoveMessage
// Author       : Andy Tate
// Date         : 5-Dec-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::CreateAbortMoveMessage(void)
{
	AVP_Inventory_System&	xParentSystem	= GetParentSystem();
	AVP_ClientEntity_Actor*	pxClient		= xParentSystem.GetParentClient();

	// Only client-side items should be sending start-move messages, the
	// moves themselves suppress further messages since the inventory-
	// system itself sends state-change messages...
	if(pxClient == NULL)
	{
		return(false);
	}
	else
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(ASURA_MSG_MOVE_ABORT, sizeof(Asura_Msg_Move));

		if(pxMessage == NULL)
		{
			return(false);
		}
		else
		{
			const AVP_Inventory_System&	xParentSystem	= GetParentSystem();
			const Asura_Guid			uOwnerGuid		= xParentSystem.GetOwnerGuid();
			Asura_Msg_Move*		pxData			= static_cast<Asura_Msg_Move*>(pxMessage->GetData());

			// Create the packet...
			pxData->ClearFlags();
			pxData->SetIgnoredByOwner(false);
			pxData->SetForceAbort(true);

			// Send the packet...
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->SetToGuid(uOwnerGuid);

			pxClient->HandleMessage(pxMessage);
			Asura_Network::DestroyMessage(pxMessage);
			return(true);
		}
	}
}

/// End of function 'AVP_Inventory_Item::CreateAbortMoveMessage'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::CreateStartMoveMessage
// Author       : Andy Tate
// Date         : 16-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::CreateStartMoveMessage(const Asura_Hash_ID uMoveHash)
{
	// Begin the appropriate move...
	if(uMoveHash == ASURA_HASH_ID_UNSET)
	{
		return(false);
	}
	else
	{
		AVP_Inventory_System&	xParentSystem	= GetParentSystem();
		AVP_ClientEntity_Actor*	pxClient		= xParentSystem.GetParentClient();

		// Only client-side items should be sending start-move messages, the
		// moves themselves suppress further messages since the inventory-
		// system itself sends state-change messages...
		if(pxClient == NULL)
		{
			return(false);
		}
		else
		{
			Asura_Message*	pxMessage	= Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(Asura_Msg_Move));

			if(pxMessage == NULL)
			{
				return(false);
			}
			else
			{
				const AVP_Inventory_System&	xParentSystem	= GetParentSystem();
				const Asura_Guid			uOwnerGuid		= xParentSystem.GetOwnerGuid();
				Asura_Msg_Move*		pxData			= static_cast<Asura_Msg_Move*>(pxMessage->GetData());

				// Create the packet...
				pxData->ClearFlags();
				pxData->SetMoveID(uMoveHash);
				pxData->SetIgnoredByOwner(false);
				pxData->SetForceAbort(true);

				// Send the packet...
				pxMessage->SetFromGuid(uOwnerGuid);
				pxMessage->SetToGuid(uOwnerGuid);

				pxClient->HandleMessage(pxMessage);
				Asura_Network::DestroyMessage(pxMessage);
				return(true);
			}
		}
	}
}

/// End of function 'AVP_Inventory_Item::CreateStartMoveMessage'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::DoesClientStateMatchServerState
// Author       : Andy Tate
// Date         : 6-Dec-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::DoesClientStateMatchServerState(const AVP_ClientEntity_Actor& xClient) const
{
	const AVP_Inventory_System&	xParentSystem	= GetParentSystem();

	if(xParentSystem.IsServerSide())
	{
		const Asura_Hash_ID			uItemHash		= GetHashInParentGroup();
		const Asura_Hash_ID			uGroupHash		= m_xParentGroup.GetHashInParentSystem();

		const AVP_Inventory_System*	pxClientSystem	= xClient.GetInventorySystem();
		const AVP_Inventory_Group*	pxClientGroup	= (pxClientSystem == NULL) ? NULL : pxClientSystem->GetGroup(uGroupHash);
		const AVP_Inventory_Item*	pxClientItem	= (pxClientGroup == NULL) ? NULL : pxClientGroup->GetItem(uItemHash);

		if(pxClientItem == NULL)
		{
			Asura_Assert(false, "Cannot find corresponding client-side inventory-item.");
			return(false);
		}
		else
		{
			const State	eServerState	= GetCurrentState();
			const State	eClientState	= pxClientItem->GetCurrentState();

			return(eServerState == eClientState);
		}
	}
	else
	{
		Asura_Assert(false, "This function cannot be called from a client-side inventory-item.");
		return(false);
	}
}

/// End of function 'AVP_Inventory_Item::DoesClientStateMatchServerState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestIdling
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestIdling(const State eSourceState,
									   const bool bInstant,
									   const u_int uTransitionType)
{
	if (!GetCanUseIdling(bInstant, uTransitionType))  { return false; }

	bool	bSuccess	= false;

	// Decide which state to change to next...
	switch(eSourceState)
	{
		case INACTIVE:
		{
			if(bInstant)
			{
				bSuccess	= RequestStateChange(IDLING);
			}
			else
			{
				if(IsStateValid(ACTIVATING))
				{
					bSuccess	= RequestStateChange(ACTIVATING);
				}
				else if(IsStateValid(BEGIN_IDLING))
				{
					bSuccess	= RequestStateChange(BEGIN_IDLING);
				}
				else if(IsStateValid(IDLING))
				{
					bSuccess	= RequestStateChange(IDLING);
				}
				else
				{
					Asura_Assert(false, "State \"IDLING\" is not valid - you cannot begin idling.");
				}
			}
			break;
		}

		case BEGIN_IDLING:
		{
			break;
		}

		case END_IDLING: // Added for the cloak. It it possible to have the code request am automatic de-cloak while in this state when it is important to have it succeed.
		{
			if(bInstant)
			{
				bSuccess	= RequestStateChange(IDLING);
			}
			else
			{
				if(IsStateValid(BEGIN_IDLING))
				{
					bSuccess	= RequestStateChange(BEGIN_IDLING);
				}
				else if(IsStateValid(IDLING))
				{
					bSuccess	= RequestStateChange(IDLING);
				}
				else
				{
					Asura_Assert(false, "State \"IDLING\" is not valid - you cannot begin idling.");
				}
			}
			break;
		}

		case IDLING:
		{
			//Asura_Assert( false, "Changing state to \"IDLING\" when we already were in this state. Requesting state change again" ); 
			if(bInstant)
			{
				bSuccess	= RequestStateChange(IDLING);
			}
			else
			{
				if(IsStateValid(BEGIN_IDLING))
				{
					bSuccess	= RequestStateChange(BEGIN_IDLING);
				}
				else if(IsStateValid(IDLING))
				{
					bSuccess	= RequestStateChange(IDLING);
				}
				else
				{
					Asura_Assert(false, "State \"IDLING\" is not valid - you cannot begin idling.");
				}
			}
			break;
		}

		case USING:
		{
			Asura_Assert(!bInstant, "Cannot transition from \"USING\" state to \"IDLING\" instantly.");

			if(IsStateValid(END_USING))
			{
				bSuccess	= RequestStateChange(END_USING);
			}
			else if(IsStateValid(BEGIN_IDLING))
			{
				bSuccess	= RequestStateChange(BEGIN_IDLING);
			}
			else if(IsStateValid(IDLING))
			{
				bSuccess	= RequestStateChange(IDLING);
			}
			else
			{
				Asura_Assert(false, "State \"IDLING\" is not valid - you cannot begin idling.");
			}
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected state to commence idle from.");
			break;
		}
	}		// End of switch.

	// Was our state-change successful?
	if(bSuccess)
	{
		const AVP_Inventory_System&	xParentSystem	= GetParentSystem();

		// Do we need to send a message?
		if(xParentSystem.IsMaster() && (eSourceState != INACTIVE))
		{
			Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_ITEM_BEGIN_IDLING, sizeof(AVP_Message_InventoryItem_BeginIdling));

			if(pxMessage != NULL)
			{
				const Asura_Guid						uOwnerGuid	= xParentSystem.GetOwnerGuid();
				AVP_Message_InventoryItem_BeginIdling*	pxData		= static_cast<AVP_Message_InventoryItem_BeginIdling*>(pxMessage->GetData());

				if(pxData != NULL)
				{
					const u_int							uUsageType	= GetCurrentUsageType();
					const AVP_InventoryItem_SubType*	pxSubType	= GetSubTypeFromUsageType(uUsageType);
					const u_int							uQuantity	= (pxSubType == NULL) ? 0 : Asura_Maths::FloatToInt(pxSubType->GetCurrentQuantity());
					const Asura_Hash_ID					uGroupHash	= m_xParentGroup.GetHashInParentSystem();

					pxData->SetGroupHash(uGroupHash);
					pxData->SetItemHash(m_uHashInParentGroup);
					pxData->SetTransitionType(uTransitionType);
					pxData->SetUsageType(uUsageType);
					pxData->SetQuantity(uQuantity);
				}

				pxMessage->SetFromGuid(uOwnerGuid);
				pxMessage->SetToGuid(uOwnerGuid);
				pxMessage->UseGuaranteedDelivery();
				Asura_Network::Send(pxMessage);
			}
		}

		// Update the source and destination states...
		m_eSourceState				= eSourceState;
		m_eDestinationState			= IDLING;

		m_uCurrentTransitionType	= uTransitionType;
		return(true);
	}
	else
	{
		return(false);
	}
}

/// End of function 'AVP_Inventory_Item::RequestIdling'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestNextIdlingState
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestNextIdlingState(void)
{
	const State	eCurrentState	= GetCurrentState();

	// Decide which state to change to next...
	switch(eCurrentState)
	{
		case ACTIVATING:
		case END_USING:
		{
			if(IsStateValid(BEGIN_IDLING))
			{
				return(RequestStateChange(BEGIN_IDLING));
			}
			else if(IsStateValid(IDLING))
			{
				return(RequestStateChange(IDLING));
			}
			else
			{
				Asura_Assert(false, "State \"IDLING\" is not valid - you cannot begin idling.");
				return(false);
			}
		}

		case BEGIN_IDLING:
		{
			if(IsStateValid(IDLING))
			{
				return(RequestStateChange(IDLING));
			}
			else
			{
				Asura_Assert(false, "State \"IDLING\" is not valid - you cannot begin idling.");
				return(false);
			}
		}

		default:
		{
			Asura_Assert(false, "Unexpected state to commence idling from.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::RequestNextIdlingState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestUsing
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestUsing(const State eSourceState,
									  const u_int uUsageType,
									  const u_int uTransitionType)
{
	const AVP_Inventory_System&		xParentSystem		= GetParentSystem();

	if( !GetCanUseUsageType(uUsageType, uTransitionType) && xParentSystem.IsMaster() )
	{
		return(false);
	}
	else
	{
		bool	bSuccess	= false;

		// Decide which state to change to next...
		switch(eSourceState)
		{
			case INACTIVE:
			{
				if(m_bHasNewState && (m_eNewState == IDLING))
				{
					m_uPendingUsageType			= uUsageType;
					m_uPendingTransitionType	= uTransitionType;
					return(true);
				}
				else
				{
					Asura_Assert(false, "Item is inactive when usage is requested - must be activated first.");
					return(false);
				}
			}

			case END_USING:
			case BEGIN_IDLING:
			{
				m_uPendingUsageType	= uUsageType;
				return(true);
			}

			case IDLING:
			{
				// Validation...
				Asura_Assert((GetCurrentUsageType() == UT_NONE), "Usage-type has not been reset.");

				// Decide which state to change to next...
				if(IsStateValid(END_IDLING))
				{
					bSuccess	= RequestStateChange(END_IDLING);
				}
				else if(IsStateValid(BEGIN_USING))
				{
					bSuccess	= RequestStateChange(BEGIN_USING);
				}
				else if(IsStateValid(USING))
				{
					bSuccess	= RequestStateChange(USING);
				}
				else
				{
					Asura_Assert(false, "State \"USING\" is not valid - you cannot begin usage.");
				}
				break;
			}

			case USING:
			{
				// Very much a special case, requesting using from the using state is usually blocked before this in GetCanUseUsageType.
				// If it is explicitly allowed through (like with the minigun) then just switch to the new usage type.
				m_uCurrentUsageType = uUsageType;
				bSuccess = true;
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected state to commence usage from.");
				return(false);
			}
		}		// End of switch.

		// Was our state-change successful?
		if(bSuccess)
		{
			// Do we need to send a message?
			if(xParentSystem.IsMaster())
			{
				Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_ITEM_BEGIN_USING, sizeof(AVP_Message_InventoryItem_BeginUsing));

				if(pxMessage != NULL)
				{
					const Asura_Guid						uOwnerGuid	= xParentSystem.GetOwnerGuid();
					AVP_Message_InventoryItem_BeginUsing*	pxData		= static_cast<AVP_Message_InventoryItem_BeginUsing*>(pxMessage->GetData());

					if(pxData != NULL)
					{
						const AVP_InventoryItem_SubType*	pxSubType	= GetSubTypeFromUsageType(uUsageType);
						const u_int							uQuantity	= (pxSubType == NULL) ? 0 : Asura_Maths::FloatToInt(pxSubType->GetCurrentQuantity());
						const Asura_Hash_ID					uGroupHash	= m_xParentGroup.GetHashInParentSystem();

						pxData->SetGroupHash(uGroupHash);
						pxData->SetItemHash(m_uHashInParentGroup);
						pxData->SetTransitionType(uTransitionType);
						pxData->SetUsageType(uUsageType);
						pxData->SetQuantity(uQuantity);
					}

					pxMessage->SetFromGuid(uOwnerGuid);
					pxMessage->SetToGuid(uOwnerGuid);
					pxMessage->UseGuaranteedDelivery();
					Asura_Network::Send(pxMessage);
				}
			}

			// Update the source and destination states...
			m_eSourceState				= eSourceState;
			m_eDestinationState			= USING;

			m_uCurrentUsageType			= uUsageType;
			m_uPendingUsageType			= UT_NONE;

			m_uCurrentTransitionType	= uTransitionType;
			return(true);
		}
		else
		{
			return(false);
		}
	}
}

/// End of function 'AVP_Inventory_Item::RequestUsing'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestNextUsingState
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestNextUsingState(void)
{
	const State	eCurrentState	= GetCurrentState();

	// Validation...
	Asura_Assert((GetCurrentUsageType() != UT_NONE), "Usage-type has not been set.");
	Asura_Assert((GetDestinationState() == USING), "Destination-state has not been set.");

	// Decide which state to change to next...
	switch(eCurrentState)
	{
		case END_IDLING:
		{
			if(IsStateValid(BEGIN_USING))
			{
				return(RequestStateChange(BEGIN_USING));
			}
			else if(IsStateValid(USING))
			{
				return(RequestStateChange(USING));
			}
			else
			{
				Asura_Assert(false, "State \"USING\" is not valid - you cannot begin usage.");
				return(false);
			}
		}

		case BEGIN_USING:
		{
			if(IsStateValid(USING))
			{
				return(RequestStateChange(USING));
			}
			else
			{
				Asura_Assert(false, "State \"USING\" is not valid - you cannot begin usage.");
				return(false);
			}
		}

		default:
		{
			Asura_Assert(false, "Unexpected state to commence usage from.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::RequestNextUsingState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestInactive
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestInactive(const State eSourceState,
										 const bool bInstant,
										 const u_int uTransitionType)
{
	bool	bSuccess	= false;
	
	if(bInstant)
	{
		bSuccess	= RequestStateChange(INACTIVE);
	}
	else
	{
		// Decide which state to change to next...
		switch(eSourceState)
		{
			case USING:
			{
				if(IsStateValid(END_USING))
				{
					bSuccess	= RequestStateChange(END_USING);
				}
				else if(IsStateValid(BEGIN_IDLING))
				{
					bSuccess	= RequestStateChange(BEGIN_IDLING);
				}
				else if(IsStateValid(IDLING))
				{
					bSuccess	= RequestStateChange(IDLING);
				}
				else if(IsStateValid(END_IDLING))
				{
					bSuccess	= RequestStateChange(END_IDLING);
				}
				else if(IsStateValid(DEACTIVATING))
				{
					bSuccess	= RequestStateChange(DEACTIVATING);
				}
				else if(IsStateValid(INACTIVE))
				{
					bSuccess	= RequestStateChange(INACTIVE);
				}
				else
				{
					Asura_Assert(false, "State \"INACTIVE\" is not valid - you cannot begin deactivating.");
				}
				break;
			}

			case IDLING:
			{
				if(IsStateValid(END_IDLING))
				{
					bSuccess	= RequestStateChange(END_IDLING);
				}
				else if(IsStateValid(DEACTIVATING))
				{
					bSuccess	= RequestStateChange(DEACTIVATING);
				}
				else if(IsStateValid(INACTIVE))
				{
					bSuccess	= RequestStateChange(INACTIVE);
				}
				else
				{
					Asura_Assert(false, "State \"INACTIVE\" is not valid - you cannot begin deactivating.");
				}
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected state to commence deactivation from.");
				break;
			}
		}		// End of switch.
	}

	// Was our state-change successful?
	if(bSuccess)
	{
		m_eSourceState				= eSourceState;
		m_eDestinationState			= INACTIVE;

		m_uCurrentTransitionType	= uTransitionType;
		return(true);
	}
	else
	{
		return(false);
	}
}

/// End of function 'AVP_Inventory_Item::RequestInactive'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestNextInactiveState
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestNextInactiveState(void)
{
	const State	eCurrentState	= GetCurrentState();

	// Decide which state to change to next...
	switch(eCurrentState)
	{
		case END_IDLING:
		{
			if(IsStateValid(DEACTIVATING))
			{
				return(RequestStateChange(DEACTIVATING));
			}
			else if(IsStateValid(INACTIVE))
			{
				return(RequestStateChange(INACTIVE));
			}
			else
			{
				Asura_Assert(false, "State \"INACTIVE\" is not valid - you cannot begin deactivating.");
				return(false);
			}
		}

		case DEACTIVATING:
		{
			if(IsStateValid(INACTIVE))
			{
				return(RequestStateChange(INACTIVE));
			}
			else
			{
				Asura_Assert(false, "State \"INACTIVE\" is not valid - you cannot begin deactivating.");
				return(false);
			}
		}

		default:
		{
			Asura_Assert(false, "Unexpected state to commence deactivation from.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::RequestNextInactiveState'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::AVP_Inventory_Item
// Author       : Andy Tate
// Date         : 25-Oct-2006
////////////////////////////////////////////////////////////////////////////////

AVP_Inventory_Item::AVP_Inventory_Item(AVPInventoryItemType eType, AVP_Inventory_Group& xParentGroup,
									   const Asura_Hash_ID uItemHash,
									   const Asura_Hash_ID uBlueprintHash,
									   Asura_Char const* const szName)
: m_xParentGroup(xParentGroup)
, m_szName(szName)
, m_uBlueprintHash(uBlueprintHash)
, m_uEntityGuid(ASURA_GUID_UNREGISTERED)
, m_uPriority(0)
, m_ppxSubTypes(NULL)
, m_uMenuImageHash(ASURA_HASH_ID_UNSET)
, m_uMenuSelectedImageHash(ASURA_HASH_ID_UNSET)
, m_uMenuEmptyImageHash(ASURA_HASH_ID_UNSET)
, m_uMenuSelectedEmptyImageHash(ASURA_HASH_ID_UNSET)
, m_uNotSelectableImageHash(ASURA_HASH_ID_UNSET)
, m_uLinkMPHash( ASURA_HASH_ID_UNSET )
, m_uCommonItemHash( ASURA_HASH_ID_UNSET )
, m_eItemType( eType )
, m_fLastMessageTimestamp( 0.0f )
, m_pxContainerAttachment(NULL)
, m_uValidStateBits(0)
, m_fTimeInState(0.0f)
, m_uSelectorDisplayConditionIndex(0)
, m_uContainerAttachment_MarkerPoint(ASURA_HASH_ID_UNSET)
, m_uGeneralFlags(0)
, m_uStateFlags(0)
, m_uUsageFlags(0)
{
	// The only valid states by default are inactive and idling...
	SetStateAsValid(INACTIVE);
	SetStateAsValid(IDLING);

	// Register with the parent-system...
	m_uHashInParentGroup	= (xParentGroup.AddItem(uItemHash, this)) ? uItemHash : ASURA_HASH_ID_UNSET;

	GetStaticDataFromBlueprint();

	CreateContainerAttachment();
}

/// End of function 'AVP_Inventory_Item::AVP_Inventory_Item'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::~AVP_Inventory_Item
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Inventory_Item::~AVP_Inventory_Item()
{
	DestroyContainerAttachment();

	for ( u_int u = 0; u < m_uNumSubTypes; ++u )
	{
		delete m_ppxSubTypes[ u ];
		m_ppxSubTypes[ u ] = NULL;
	}
	delete [] m_ppxSubTypes;
	m_ppxSubTypes = NULL;
}

/// End of function 'AVP_Inventory_Item::~AVP_Inventory_Item'. ///

void AVP_Inventory_Item::InitialiseFromBlueprint()
{
	for ( u_int u = 0; u < m_uNumSubTypes; ++u )
	{
		m_ppxSubTypes[ u ]->InitialiseFromBlueprint();
	}
}

void AVP_Inventory_Item::FillCommonItemHashFromPickups(Asura_Hash_ID uItemHash)
{
	const Asura_Blueprint* pxItemBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, uItemHash );
	if ( pxItemBlueprint )
	{
		Asura_Value xValue;
		if ( pxItemBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_PICKUP, xValue ) )
		{
			Asura_Hash_ID uPickupHash = xValue.GetHashIDValue();

			// Get common item hash from pickup
			if ( uPickupHash != ASURA_HASH_ID_UNSET )
			{
				const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_PICKUPS, uPickupHash );
				Asura_Hash_ID uInvItemParamHash = AVP_Settings::IsMultiPlayer() ? uAVP_BLUEPRINT_PARAM_PICKUP_MULTIPLAYERINVENTORYITEMBLUEPRINT : uAVP_BLUEPRINT_PARAM_PICKUP_SINGLEPLAYERINVENTORYITEMBLUEPRINT;
				if ( pxBlueprint->GetParameterValueByHashID( uInvItemParamHash, xValue ) )
				{
					m_uCommonItemHash = xValue.GetHashIDValue();
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::ReadFromChunkStream
// Author       : Andy Tate
// Date         : 4-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;

	if ( uVersion == uCURRENT_CHUNK_VERSION )
	{
		// Some sub-types are automatically added during the inventory initialisation.
		// If we don't clear them here we'll get a memory leak.
		InitialiseSubTypes( 0 );

		u_int uNumSubTypes;
		xStream >> uNumSubTypes;
		m_uNumSubTypes = uNumSubTypes;
		if ( m_uNumSubTypes )
		{
			m_ppxSubTypes = new AVP_InventoryItem_SubType*[ m_uNumSubTypes ];
			for ( u_int u = 0; u < m_uNumSubTypes; ++u )
			{
				int iSubType;
				xStream >> iSubType;

				if(iSubType!=AVP_INVENTORYITEMSUBTYPE_NONE)
				{
					Asura_Hash_ID uBlueprintHashID;
					xStream >> uBlueprintHashID;

					m_ppxSubTypes[ u ] = CreateSubType( iSubType, uBlueprintHashID );
					if ( m_ppxSubTypes[ u ] )
					{
						m_ppxSubTypes[ u ]->ReadFromChunkStream( xStream );
					}
					else
					{
						// We have to assume no-one's going to be removing types and classes,
						// at least without checking to see if any items of that type are saved out.
						// If this is going to happen we could save the size of the item and skip the data here.
						Asura_Assert( false, "AVP_Inventory_Item::ReadFromChunkStream - unable to create sub type" );
					}
				}
				else
				{
					Asura_Assert( false, "AVP_Inventory_Item::ReadFromChunkStream - invalid subtype in chunk stream" );
				}
			}
		}
	}
	else if ( uVersion == 0 )
	{
	}
	else
	{
		Asura_Assertf( false, ( "AVP_Inventory_Item::ReadFromChunkStream - unhandled chunk version (%u)", uVersion ) );
	}
}

/// End of function 'AVP_Inventory_Item::ReadFromChunkStream'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::WriteToChunkStream
// Author       : Andy Tate
// Date         : 4-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << uCURRENT_CHUNK_VERSION;

	xStream << m_uNumSubTypes;
	for ( u_int u = 0; u < m_uNumSubTypes; ++u )
	{
		if(m_ppxSubTypes[ u ])
		{
			xStream << m_ppxSubTypes[ u ]->GetType();
			xStream << m_ppxSubTypes[ u ]->GetBlueprintHashID();
			m_ppxSubTypes[ u ]->WriteToChunkStream( xStream );
		}
		else
		{
			xStream << (int)AVP_INVENTORYITEMSUBTYPE_NONE;
			Asura_Assert(false, "AVP_Inventory_Item::WriteToChunkStream : NULL subtype found in inventory item, skipping data, this is not good!");
		}
	}
}

/// End of function 'AVP_Inventory_Item::WriteToChunkStream'. ///


AVP_InventoryItem_SubType* AVP_Inventory_Item::CreateSubType( const int iSubType, const Asura_Hash_ID uBlueprintHashID )
{
	if ( iSubType == AVP_INVENTORYITEMSUBTYPE_BASE )
	{
		return new AVP_InventoryItem_SubType( iSubType, uBlueprintHashID );
	}
	else if ( iSubType == AVP_INVENTORYITEMSUBTYPE_DYNAMIC )
	{
		return new AVP_InventoryItem_DynamicSubType( iSubType, uBlueprintHashID );
	}
	else
	{
		Asura_Assertf( false, ( "AVP_Inventory_Item::CreateSubType - unable to create sub type %d", iSubType ) );
		return NULL;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestBeginUsage
// Author       : Andy Tate
// Date         : 27-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestBeginUsage(const u_int uUsageType,
										   const u_int uTransitionType)
{
	const AVP_Inventory_System&	xParentSystem	= m_xParentGroup.GetParentSystem();

	if(xParentSystem.IsMaster())
	{
		return(RequestUsing(uUsageType, uTransitionType));
	}
	else
	{
		Asura_Assert(false, "Cannot request begin-usage of an inventory-item from the slave.");
		return(false);
	}
}

/// End of function 'AVP_Inventory_Item::RequestBeginUsage'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestEndUsage
// Author       : Andy Tate
// Date         : 27-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestEndUsage(const u_int uTransitionType)
{
	const AVP_Inventory_System&	xParentSystem	= m_xParentGroup.GetParentSystem();

	if(xParentSystem.IsMaster())
	{
		return(RequestIdling(false, uTransitionType));
	}
	else if(xParentSystem.IsServerSide())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_ITEM_REQUEST_END_USAGE, sizeof(AVP_Message_InventoryItem_RequestEndUsage));

		// We're the slave - we need to ask the master whether we can do this...
		if(pxMessage != NULL)
		{
			const Asura_Guid							uOwnerGuid	= xParentSystem.GetOwnerGuid();
			AVP_Message_InventoryItem_RequestEndUsage*	pxData		= static_cast<AVP_Message_InventoryItem_RequestEndUsage*>(pxMessage->GetData());

			if(pxData != NULL)
			{
				const Asura_Hash_ID	uGroupHash	= m_xParentGroup.GetHashInParentSystem();

				pxData->SetGroupHash(uGroupHash);
				pxData->SetItemHash(m_uHashInParentGroup);
				pxData->SetTransitionType(uTransitionType);
			}

			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->SetToGuid(ASURA_GUID_MY_CLIENT_ENTITIES);
			pxMessage->UseGuaranteedDelivery();
			Asura_Network::Send(pxMessage);
		}
		return(true);
	}
	else
	{
		Asura_Assert(false, "Cannot request end-usage of an inventory-item from a client-side slave.");
		return(false);
	}
}

/// End of function 'AVP_Inventory_Item::RequestEndUsage'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::InitialiseSubTypes
// Author       : Andy Tate
// Date         : 30-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::InitialiseSubTypes(const u_int uNumSubTypes)
{
	// Remove the existing sub-types...
	for ( u_int u = 0; u < m_uNumSubTypes; ++u )
	{
		delete m_ppxSubTypes[ u ];
		m_ppxSubTypes[ u ] = NULL;
	}
	delete [] m_ppxSubTypes;
	m_ppxSubTypes = NULL;

	// Ensure that we're inactive to start with...
	m_uNumSubTypes = uNumSubTypes;

	// Initialise the sub-types...
	if ( uNumSubTypes > 0 )
	{
		m_ppxSubTypes = new AVP_InventoryItem_SubType* [uNumSubTypes];
		for(u_int u = 0; u < uNumSubTypes; u++)
		{
			m_ppxSubTypes[u] = NULL;
		}
	}
}

/// End of function 'AVP_Inventory_Item::InitialiseSubTypes'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::AreAllSubTypesEmpty
// Author       : Andy Tate
// Date         : 15-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::AreAllSubTypesEmpty(void) const
{
	for(u_int u = 0; u < m_uNumSubTypes; u++)
	{
		if(!m_ppxSubTypes[u]->IsEmpty() || GetParentSystem().GetAmmoStore()->GetAmmoQuantity(m_ppxSubTypes[u]->GetAmmoType()) > 0)
		{
			return(false);
		}
	}

	return(true);
}

/// End of function 'AVP_Inventory_Item::AreAllSubTypesEmpty'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::IsAnySubTypeEmpty
// Author       : Andy Tate
// Date         : 15-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::IsAnySubTypeEmpty(void) const
{
	for(u_int u = 0; u < m_uNumSubTypes; u++)
	{
		if(m_ppxSubTypes[u]->IsEmpty())
		{
			return(true);
		}
	}

	return(false);
}

/// End of function 'AVP_Inventory_Item::IsAnySubTypeEmpty'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestActivation
// Author       : Andy Tate
// Date         : 25-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestActivation(const bool bInstant,
										   const u_int uTransitionType)
{
	switch(m_eCurrentState)
	{
		case INACTIVE:
		case DEACTIVATING:
		{
			return(RequestIdling(bInstant, uTransitionType));
		}

		case ACTIVATING:
		case BEGIN_IDLING:
		case IDLING:
		case END_IDLING:
		case BEGIN_USING:
		case USING:
		case END_USING:
		{
			return(true);
		}

		default:
		{
			Asura_Assert(false, "Unexpected state found during activation.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::RequestActivation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::RequestDeactivation
// Author       : Andy Tate
// Date         : 25-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::RequestDeactivation(const bool bInstant,
											 const u_int uTransitionType)
{
	switch(m_eCurrentState)
	{
		case INACTIVE:
		{
			// Already inactive, ignore
			return (true); 
		}
		case DEACTIVATING:
		{
			if(m_bHasNewState)
			{
				Asura_Assert(false, "Inventory-item deactivation is requested on the same frame as activation");
				return(false);
			}
			else
			{
				return(true);
			}
		}

		case ACTIVATING:
		case BEGIN_IDLING:
		case IDLING:
		case END_IDLING:
		case BEGIN_USING:
		case USING:
		case END_USING:
		{
			return(RequestInactive(bInstant, uTransitionType));
		}

		default:
		{
			Asura_Assert(false, "Unexpected state found during deactivation.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::RequestDeactivation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::HandleMessage
// Author       : Andy Tate
// Date         : 15-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::HandleMessage(const Asura_Message* pxMessage)
{
	const AVP_Inventory_System&	xParentSystem	= GetParentSystem();
	const u_int					uMessageID		= pxMessage->GetMsgID();

	if (pxMessage->GetSentTime() < m_fLastMessageTimestamp)
	{
		// Old message, ignore!
		Asura_Assertf( false, ("Ignoring old message %s", Asura_Network::GetMsgIDDesc( pxMessage->GetMsgID() ))); 
		return true; 
	}

	switch(uMessageID)
	{
		case AVP_MSGID_ITEM_BEGIN_IDLING:
		{
			if(xParentSystem.IsSlave())
			{
				const AVP_Message_InventoryItem_BeginIdling*	pxData	= static_cast<const AVP_Message_InventoryItem_BeginIdling*>(pxMessage->GetData());

				if(pxData != NULL)
				{
					const u_int					uTransitionType	= pxData->GetTransitionType();
					const u_int					uUsageType		= pxData->GetUsageType();
					AVP_InventoryItem_SubType*	pxSubType		= GetSubTypeFromUsageType(uUsageType);

					RequestIdling(false, uTransitionType);

					if(pxSubType != NULL)
					{
						const float	fQuantity	= static_cast<float>(pxData->GetQuantity());

						pxSubType->SetCurrentQuantity(fQuantity);
					}
				}
			}

			if(xParentSystem.IsServerSide())
			{
				const Asura_Guid	uOwnerGuid	= xParentSystem.GetOwnerGuid();

				// We're a master AND a server - ensure that we forward this to the client...
				Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, uOwnerGuid);
			}
			
			m_fLastMessageTimestamp = pxMessage->GetSentTime(); 
			
			return(true);
		}

		case AVP_MSGID_ITEM_BEGIN_USING:
		{
			if(xParentSystem.IsSlave())
			{
				const AVP_Message_InventoryItem_BeginUsing*	pxData	= static_cast<const AVP_Message_InventoryItem_BeginUsing*>(pxMessage->GetData());

				if(pxData != NULL)
				{
					const u_int					uTransitionType	= pxData->GetTransitionType();
					const u_int					uUsageType		= pxData->GetUsageType();
					AVP_InventoryItem_SubType*	pxSubType		= GetSubTypeFromUsageType(uUsageType);

					RequestUsing(uUsageType, uTransitionType);

					if(pxSubType != NULL)
					{
						const float	fQuantity	= static_cast<float>(pxData->GetQuantity());

						pxSubType->SetCurrentQuantity(fQuantity);
					}
				}
			}

			if(xParentSystem.IsServerSide())
			{
				const Asura_Guid	uOwnerGuid	= xParentSystem.GetOwnerGuid();

				// We're a master AND a server - ensure that we forward this to the client...
				Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, uOwnerGuid);
			}

			m_fLastMessageTimestamp = pxMessage->GetSentTime(); 

			return(true);
		}

		case AVP_MSGID_ITEM_REQUEST_END_USAGE:
		{
			if(xParentSystem.IsMaster())
			{
				const AVP_Message_InventoryItem_BeginIdling*	pxData	= static_cast<const AVP_Message_InventoryItem_BeginIdling*>(pxMessage->GetData());

				if(pxData != NULL)
				{
					const u_int	uTransitionType	= pxData->GetTransitionType();

					RequestEndUsage(uTransitionType);
				}
			}

			m_fLastMessageTimestamp = pxMessage->GetSentTime(); 

			return(true);
		}

		case AVP_MSGID_ATTACH_WEAPON:
		{
			AVP_Inventory_System&				xParentSystem		= GetParentSystem();
			AVP_ClientEntity_Actor*				pxClient			= xParentSystem.GetParentClient();
			AVP_Actor_AnimationHandler*			pxAnimationHandler	= pxClient == NULL ? NULL : pxClient->GetAnimationHandler();
			AVP_Actor_AttachmentHandler*		pxAttachmentHandler	= pxClient == NULL ? NULL : pxClient->GetAttachmentHandler();

			if(pxAnimationHandler)
			{
				pxAnimationHandler->AttachItem( this );
				
				if(pxAttachmentHandler)
				{
					Asura_Hash_ID uSkinHash = (GetAnimation())?GetAnimation()->GetSkinHashID():ASURA_HASH_ID_UNSET;
					if(uSkinHash != ASURA_HASH_ID_UNSET)
					{
						pxAttachmentHandler->DestroyAttachment(uSkinHash);
					}
				}
			}
			
			SetHidden(false);
			
			return true;
		}

		case AVP_MSGID_DETACH_WEAPON:
		{
			AVP_Inventory_System&				xParentSystem		= GetParentSystem();
			AVP_ClientEntity_Actor*				pxClient			= xParentSystem.GetParentClient();
			AVP_Actor_AnimationHandler*			pxAnimationHandler	= pxClient == NULL ? NULL : pxClient->GetAnimationHandler();
			AVP_Actor_AttachmentHandler*		pxAttachmentHandler	= pxClient == NULL ? NULL : pxClient->GetAttachmentHandler();

			if(pxAnimationHandler)
			{
				pxAnimationHandler->DetachItem( this );

				if(pxAttachmentHandler)
				{
					AVP_Actor_Attachment xAttachment;

					Asura_Hash_ID uBlueprintHash = static_cast<const AVP_Message_Actor_Attachment*>(pxMessage->GetData())->m_uBlueprintHash;

					if(uBlueprintHash != ASURA_HASH_ID_UNSET)
					{
						Asura_Value xTempValue;
						Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_ATTACHMENTS, uBlueprintHash);
						if(pxBlueprint)
						{
							if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_MARKERPOINT, xTempValue))
							{
								xAttachment.m_uMarkerID = xTempValue.GetHashIDValue();
							} 
							else
							{
								Asura_Assertf(false, ("BLUEPRINT ERROR: Failed to find HashID 'MarkerPoint' in specified attachment-blueprint. %0X", pxBlueprint->GetHashID()));
								return(false);
							}


							if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_DEFAULTANIM, xTempValue))
							{
								xAttachment.m_uAnimID = xTempValue.GetHashIDValue();
							}
							else
							{
								Asura_Assertf(false, ("BLUEPRINT ERROR: Failed to find HashID 'DefaultAnim' in specified attachment-blueprint. %0X", pxBlueprint->GetHashID()));
								return(false);
							}

							if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACHMENT_DROPONDECAPITATION, xTempValue))
							{
								if (xTempValue.GetBoolValue()) xAttachment.m_bDropOnDecapitation = true;
							}
							else
							{
								Asura_Assertf(false, ("BLUEPRINT WARNING: Failed to find bool 'DropOnDecapitation' in specified attachment-blueprint. %0X", pxBlueprint->GetHashID()));
							}

							if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACHMENT_CHANCEOFDROPDEATH, xTempValue))
							{
								xAttachment.m_fChanceOfDropOnDeath = xTempValue.GetFloatValue();
							}
							else
							{
								Asura_Assertf(false, ("BLUEPRINT WARNING: Failed to find float 'ChanceOfDropDeath' in specified attachment-blueprint. %0X", pxBlueprint->GetHashID()));
							}

							if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACHMENT_CHANCEOFDROPHEADBIT, xTempValue))
							{
								xAttachment.m_fChanceOfDropOnHeadbit = xTempValue.GetFloatValue();
							}
							else
							{
								Asura_Assertf(false, ("BLUEPRINT WARNING: Failed to find float 'ChanceOfDropHeadbit' in specified attachment-blueprint. %0X", pxBlueprint->GetHashID()));
							}
						}
						
						xAttachment.m_uSkinID = (GetAnimation())?GetAnimation()->GetSkinHashID():ASURA_HASH_ID_UNSET; // Skin comes from the weapon, not the blueprint
						
						if(xAttachment.m_uSkinID != ASURA_HASH_ID_UNSET && xAttachment.m_uMarkerID != ASURA_HASH_ID_UNSET)
						{
							pxAttachmentHandler->AddAttachment(xAttachment);
						}
					}		
				}
			}
			SetHidden(true);

			return true;
		}

		default:
		{
			return(false);
		}
	}		// End of switch.
}

/// End of function 'AVP_Inventory_Item::HandleMessage'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::Update
// Author       : Andy Tate
// Date         : 25-Oct-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::Update(void)
{
	AVP_Inventory_System&	xParentSystem	= GetParentSystem();
	AVP_ClientEntity_Actor*	pxClient		= xParentSystem.GetParentClient();

	// Keep updating until we've finished all instantaneous transitions...
	do
	{
		// Do we need to change state?
		if(m_bHasNewState)
		{
			do
			{
				m_bHasNewState	= false;
				m_fTimeInState	= 0.0f;

				OnStateChange(m_eCurrentState, m_eNewState);

				if(!m_bHasNewState)
				{
					m_ePreviousState	= m_eCurrentState;
					m_eCurrentState		= m_eNewState;
				}

			} while(m_bHasNewState);
		}
		else
		{
			m_fTimeInState	+= Asura_Timers::GetGameTimeSlice();
		}

		// What are we supposed to be doing?
		if(pxClient == NULL)
		{
			AVP_ServerEntity_Actor*	pxServer	= xParentSystem.GetParentServer();

			if(pxServer == NULL)
			{
				Asura_Assert(false, "Cannot find either client or server for this inventory-item.");
			}
			else
			{
				UpdateServerState(*pxServer);
			}
		}
		else
		{
			UpdateClientState(*pxClient);
		}

	} while(m_bHasNewState);
}

/// End of function 'AVP_Inventory_Item::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::Render
// Author       : Andy Tate
// Date         : 25-Oct-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::Render(void) const
{
	// Nothing to do here - override if necessary...
}

/// End of function 'AVP_Inventory_Item::Render'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::CreateDebugString
// Author       : Andy Tate
// Date         : 6-Dec-2007
////////////////////////////////////////////////////////////////////////////////

u_int AVP_Inventory_Item::CreateDebugString(const u_int uMaxChars,
											char* szText) const
{
	u_int	uOffset	= Asura_PrintfToBfr(szText, "\"%s\" - ST %d UT %d", uMaxChars, true, m_szName, m_eCurrentState, m_uCurrentUsageType);

	return(uOffset);
}

/// End of function 'AVP_Inventory_Item::CreateDebugString'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::GetParentSystem
// Author       : Andy Tate
// Date         : 22-Nov-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Inventory_System& AVP_Inventory_Item::GetParentSystem(void)
{
	return(m_xParentGroup.GetParentSystem());
}

/// End of function 'AVP_Inventory_Item::GetParentSystem'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::GetParentSystem
// Author       : Andy Tate
// Date         : 22-Nov-2007
////////////////////////////////////////////////////////////////////////////////

const AVP_Inventory_System& AVP_Inventory_Item::GetParentSystem(void) const
{
	return(m_xParentGroup.GetParentSystem());
}

/// End of function 'AVP_Inventory_Item::GetParentSystem'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::IsSelectable
// Author       : Andy Tate
// Date         : 15-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Inventory_Item::IsSelectable(void) const
{
	return((m_ppxSubTypes == NULL) || !AreAllSubTypesEmpty());
}

/// End of function 'AVP_Inventory_Item::IsSelectable'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::SetEntityActiveFlag
// Author       : Chris Wallace
// Date         : 15-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::SetEntityActiveFlag(bool bActive)
{
	Asura_Guid uGuid = GetEntityGuid();

	if(uGuid != ASURA_GUID_UNREGISTERED)
	{
		if(bActive)
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_ACTIVATE);
			if (pxMessage)
			{
				pxMessage->SetFromGuid(GetParentSystem().GetOwnerGuid());
				pxMessage->SetToGuid(uGuid);
				pxMessage->SentTime = Asura_Timers::GetServerAbsoluteTime();
				Asura_Network::Send(pxMessage);
			}
		}
		else
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
			if (pxMessage)
			{
				pxMessage->SetFromGuid(GetParentSystem().GetOwnerGuid());
				pxMessage->SetToGuid(uGuid);
				pxMessage->SentTime = Asura_Timers::GetServerAbsoluteTime();
				Asura_Network::Send(pxMessage);
			}
		}
	}
}

/// End of function 'AVP_Inventory_Item::SetEntityActiveFlag'. ///


bool AVP_Inventory_Item::GetCanUseUsageType(const u_int uUsageType, const u_int uTransitionType) const
{
	// TODO - this is almost certainly too simplistic, but should work until we have inventory items that don't use moves to do their actions.
	const AVP_Inventory_System&		xParentSystem		= GetParentSystem();
	const AVP_ClientEntity_Actor*	pxClient			= xParentSystem.GetParentClient();
	const Asura_Movement_System*	pxMovementSystem	= (pxClient == NULL) ? NULL : pxClient->GetMovementSystem();

	if(!IsIdling(false))
	{
		Asura_Assert(false, "Unexpected state to commence usage from.");
		return false;
	}

	return (pxMovementSystem == NULL) || !pxMovementSystem->IsMoveInProgress();
}


const Asura_Blueprint* AVP_Inventory_Item::GetBlueprint() const
{
	return Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, GetBlueprintHash());
}


Asura_Hash_ID AVP_Inventory_Item::GetPickupBlueprintHash() const
{
	Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, m_uBlueprintHash);

	if(pxBlueprint)
	{
		Asura_Value xValue;
		if(pxBlueprint->GetParameterValueByHashID(0xC5150A7C, xValue)) // "Pickup"
		{
		   return xValue.GetHashIDValue();
		}
		else
		{
			return ASURA_HASH_ID_UNSET;
		}
	}
	else
	{
		Asura_Assert(false, "Inventory item has no blueprint hash!");
		return ASURA_HASH_ID_UNSET;
	}
}

void AVP_Inventory_Item::GetStaticDataFromBlueprint()
{
	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, GetBlueprintHash() );
	if ( pxBlueprint )
	{
		Asura_Value xValue;
		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_MARKERPOINT_LINK, xValue ) )
		{
			m_uLinkMPHash = xValue.GetHashIDValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_IMAGE_NOTSELECTED, xValue ) )
		{
			m_uMenuImageHash = xValue.GetHashIDValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_IMAGE_SELECTED, xValue ) )
		{
			m_uMenuSelectedImageHash = xValue.GetHashIDValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_IMAGE_NOTSELECTED_EMPTY, xValue ) )
		{
			m_uMenuEmptyImageHash = xValue.GetHashIDValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_IMAGE_SELECTED_EMPTY, xValue ) )
		{
			m_uMenuSelectedEmptyImageHash = xValue.GetHashIDValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_IMAGE_UNSELECTABLE, xValue ) )
		{
			m_uNotSelectableImageHash = xValue.GetHashIDValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_WEAPONSELECTOR_INDEX, xValue ) )
		{
			m_uSelectorDisplayConditionIndex = xValue.GetIntValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CONTAINERATTACHMENT_MARKERPOINT, xValue ) )
		{
			m_uContainerAttachment_MarkerPoint = xValue.GetIntValue();
		}

		FillCommonItemHashFromPickups( GetBlueprintHash() );
	}
	else
	{
		Asura_Assertf( false, ( "AVP_Inventory_Item::GetStaticDataFromBlueprint - could not find blueprint %u to initialise from.", GetBlueprintHash() ) );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::CreateContainerAttachment
// Author       : Emil Popovic
// Date         : 28-Apr-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::CreateContainerAttachment()
{
	// We need to create the effect only on client side.
	if( GetParentSystem().IsClientSide() )
	{	
		if(m_uContainerAttachment_MarkerPoint != ASURA_HASH_ID_UNSET) // CW - Dont create attachment if markerpoint is not set
		{
			// First read, if we have defined container.
			const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, GetBlueprintHash() );
			if ( pxBlueprint )
			{
				Asura_Value xValue;
				if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CONTAINERATTACHMENT, xValue ) )
				{
					Asura_Hash_ID uContainerHash = xValue.GetHashIDValue();

					if( uContainerHash!=ASURA_HASH_ID_UNSET) 
					{
						m_pxContainerAttachment = new Asura_AnimationAttachment_ClientContainerInstance( uContainerHash, Asura_Container_InputState::On() );
					}
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::DestroyContainerAttachment
// Author       : Emil Popovic
// Date         : 28-Apr-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::DestroyContainerAttachment()
{
	if (m_pxContainerAttachment != NULL)
	{
		if (m_pxContainerAttachment->IsLinked())
		{
			m_pxContainerAttachment->UnlinkAttachment(true);
		}
		else
		{
			delete m_pxContainerAttachment;
		}
		m_pxContainerAttachment = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::OnAttachItem
// Author       : Emil Popovic
// Date         : 28-Apr-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::OnAttachItem( Asura_Animation* pxWeaponAnimation )
{
	// We can link container only if it exist
	if( m_pxContainerAttachment && pxWeaponAnimation && m_uContainerAttachment_MarkerPoint!=ASURA_HASH_ID_UNSET )
	{
		if(!m_pxContainerAttachment->LinkAttachment( pxWeaponAnimation, m_uContainerAttachment_MarkerPoint, false ))
		{
			delete m_pxContainerAttachment; // If we fail to link the attachment, destroy it.
			m_pxContainerAttachment = NULL;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Inventory_Item::OnDetachItem
// Author       : Emil Popovic
// Date         : 28-Apr-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Inventory_Item::OnDetachItem()
{
	// We can unlink container only if it exist
	if( m_pxContainerAttachment )
	{
		if (m_pxContainerAttachment->IsLinked())
		{
			// Also make sure, we do not destroy the container.
			m_pxContainerAttachment->UnlinkAttachment(false);
		}
	}
}


/// End of file. ///
