
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Inventory_Item.h
// Author : Andy Tate
// Date   : 25-Oct-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_INVENTORY_ITEM_H__
#define __AVP_INVENTORY_ITEM_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Inventory/AVP_Inventory_ItemTypes.h"
#include "Debugging/Asura_Assert.h"

////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation;
class Asura_Blueprint;
class Asura_Chunk_Stream;
class Asura_Message;
class AVP_Inventory_Group;
class AVP_InventoryItem_SubType;
class AVP_Inventory_System;
class AVP_ClientEntity_Actor;
class AVP_ServerEntity_Actor;
class Asura_AnimationAttachment_ClientContainerInstance;

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_Inventory_Item
// Author       : Andy Tate
// Date         : 25-Oct-2006
//
// Description  : Abstract class to handle the operation a particular item in
//				  the actor's inventory.
////////////////////////////////////////////////////////////////////////////////

class AVP_Inventory_Item
{
public:
	// Public enumerated types...
	enum ItemUsageType
	{
		UT_NONE,

		NUM_ITEM_USAGE_TYPES
	};

	enum ItemTransitionType
	{
		TT_NONE,

		NUM_ITEM_TRANSITION_TYPES
	};

	// Public functions...
	AVP_Inventory_Item(AVPInventoryItemType eType, AVP_Inventory_Group& xParentGroup, const Asura_Hash_ID uItemHash, const Asura_Hash_ID uBlueprintHash, Asura_Char const* const szName );
	virtual ~AVP_Inventory_Item();

	AVPInventoryItemType GetItemType() const { return m_eItemType; }

	void					SetBlueprintHash(const Asura_Hash_ID uHash)	{ m_uBlueprintHash = uHash; }
	Asura_Hash_ID			GetBlueprintHash(void) const				{ return(m_uBlueprintHash); }
	const Asura_Blueprint*	GetBlueprint() const;

	u_int GetSelectorDisplayConditionIndex(void) const { return m_uSelectorDisplayConditionIndex; }

	virtual void InitialiseFromBlueprint();

	virtual void PostLoadInitialise() {}

	// Chunk processing functions...
	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

    // Request functions...
	bool	RequestActivation(const bool bInstant = false, const u_int uTransitionType = 0);
    bool	RequestBeginUsage(const u_int uUsageType, const u_int uTransitionType = 0);
    bool	RequestEndUsage(const u_int uTransitionType = 0);
	bool	RequestDeactivation(const bool bInstant = false, const u_int uTransitionType = 0);

	// Client side animation functions...
	virtual Asura_Animation*		CreateAnimation()								{ return NULL; }
	virtual void					DestroyAnimation()								{}
	virtual Asura_Animation*		GetAnimation( bool bCreateIfNecessary = false )	{ return NULL; }
	virtual const Asura_Animation*	GetAnimation() const							{ return NULL; }
	virtual void					InitialiseAnimation()							{}
	Asura_Hash_ID					GetLinkMarkerPoint() const						{ return m_uLinkMPHash; }

	// Sub-type functions...
	void	InitialiseSubTypes(const u_int uNumSubTypes);

	bool	IsAnySubTypeEmpty(void) const;
	bool	AreAllSubTypesEmpty(void) const;

	// Virtual manipulation functions...
	virtual bool	HandleMessage(const Asura_Message* pxMessage);
	virtual	void	Update(void);
	virtual	void	Render(void) const;

	// Virtual query functions...
	virtual	bool	IsSelectable(void) const;

	virtual bool	IsEncumberment(void) const { return false; }

	// State query functions...
	bool	IsActive(void) const					{ return(IsActive(GetCurrentState())); }
	bool	IsInactive(void) const					{ return(IsInactive(GetCurrentState())); }
	bool	IsActivating(void) const				{ return(IsActivating(GetCurrentState())); }
	bool	IsIdling(const bool bFullyIdling) const	{ return(IsIdling(GetCurrentState(), bFullyIdling)); }
	bool	IsUsing(const bool bFullyUsing) const	{ return(IsUsing(GetCurrentState(), bFullyUsing)); }
	bool	IsDeactivating(void) const				{ return(IsDeactivating(GetCurrentState())); }

	virtual int		GetFixedSlot() const { return -1; }

	// Debugging functions...
	virtual u_int	CreateDebugString(const u_int uMaxChars, char* szText) const;

    // Pickup Hash ID... Used to generate a pickup from an inventory item
	Asura_Hash_ID GetPickupBlueprintHash() const;	

	// Accessor functions...
	AVP_Inventory_System&		GetParentSystem(void);
	const AVP_Inventory_System&	GetParentSystem(void) const;

	Asura_Char const*	GetName(void) const	{ return(m_szName); }

	Asura_Hash_ID	GetHashInParentGroup(void) const	{ return(m_uHashInParentGroup); }

	u_int	GetNumSubTypes(void) const				{ return(m_uNumSubTypes); }
	u_int	GetCurrentUsageType(void) const			{ return(m_uCurrentUsageType); }
	u_int	GetCurrentTransitionType(void) const	{ return(m_uCurrentTransitionType); }

	void		SetEntityGuid(const Asura_Guid uGuid)	{ m_uEntityGuid = uGuid; }
	Asura_Guid	GetEntityGuid(void) const				{ return(m_uEntityGuid); }

	void	SetPriority(const u_int uPriority)	{ m_uPriority = uPriority; }
	u_int	GetPriority(void) const				{ return(m_uPriority); }

	AVP_InventoryItem_SubType*			GetSubType(const u_int uSubType)		{ Asura_Assert( uSubType < m_uNumSubTypes, "Not a valid sub type" ); return( ( m_ppxSubTypes == NULL || uSubType >= m_uNumSubTypes ) ? NULL : m_ppxSubTypes[uSubType] ); }
	const AVP_InventoryItem_SubType*	GetSubType(const u_int uSubType) const	{ Asura_Assert( uSubType < m_uNumSubTypes, "Not a valid sub type" ); return( ( m_ppxSubTypes == NULL || uSubType >= m_uNumSubTypes ) ? NULL : m_ppxSubTypes[uSubType] ); }

	Asura_Hash_ID	GetMenuImageHash(const bool bSelected) const	{ return((bSelected)?m_uMenuSelectedImageHash:m_uMenuImageHash); }
	Asura_Hash_ID	GetMenuEmptyImageHash(const bool bSelected) const	{ return((bSelected)?m_uMenuSelectedEmptyImageHash:m_uMenuEmptyImageHash); }
	Asura_Hash_ID	GetNotSelectableMenuImageHash() const	{ return(m_uNotSelectableImageHash); }
	void			SetMenuImageHashes(Asura_Hash_ID uHash, Asura_Hash_ID uSelectedHash, Asura_Hash_ID uNotSelectableImageHash) { m_uMenuImageHash = uHash; m_uMenuSelectedImageHash = uSelectedHash; m_uNotSelectableImageHash = uNotSelectableImageHash; }
	Asura_Hash_ID   GetCommonItemHash() const { return m_uCommonItemHash; }

	void	SetChangeOnZeroQuantity(const bool bSet)	{ m_bChangeOnZeroQuantity = bSet; }
	bool	GetChangeOnZeroQuantity(void) const			{ return(m_bChangeOnZeroQuantity); }

	virtual int	GetSubTypeIndexFromUsageType(const u_int uUsageType) const;

	bool HasPendingStateChange() const { return m_eDestinationState != m_eCurrentState; }

	bool IsHidden(void) const { return m_bHidden; } 
	void SetHidden(bool bHidden) { m_bHidden = bHidden; }

	virtual bool GetCanUseIdling(const bool bInstant, const u_int uTransitionType) const { return true; }
	virtual bool GetCanUseUsageType(const u_int uUsageType, const u_int uTransitionType) const;

protected:
	// Protected enumerated types...
	enum State
	{
		INACTIVE,
		ACTIVATING,

		BEGIN_IDLING,
		IDLING,
		END_IDLING,

		BEGIN_USING,
		USING,
		END_USING,

		DEACTIVATING,

		MAX_STATES
	};

	// Protected variables...
	AVP_Inventory_Group&		m_xParentGroup;				// A reference to the group which contains this item.
	Asura_Char const*			m_szName;					// The name of this inventory-item.
	Asura_Hash_ID				m_uHashInParentGroup;		// The hash-ID that can be used to access this group from the parent-group;

	Asura_Hash_ID				m_uBlueprintHash;			// Hash-ID of the blueprint used for this item.
	Asura_Guid					m_uEntityGuid;				// The GUID of the entity used for this item.
	u_int						m_uPriority;				// The priority of this item.

	AVP_InventoryItem_SubType**	m_ppxSubTypes;				// Array of pointers to the individual quantity data for each sub-type.

	Asura_Hash_ID				m_uMenuImageHash;			// Hash-ID of the image to be used on menus for this item
	Asura_Hash_ID				m_uMenuSelectedImageHash;	// Hash-ID of the image to be used on menus for this item when selected
	Asura_Hash_ID				m_uMenuEmptyImageHash;			// Hash-ID of the image to be used on menus for this item
	Asura_Hash_ID				m_uMenuSelectedEmptyImageHash;	// Hash-ID of the image to be used on menus for this item when selected
	Asura_Hash_ID				m_uNotSelectableImageHash;	// Hash-ID of the image to be used on menus for this item when un-selectable

	Asura_Hash_ID				m_uLinkMPHash;				// Hash-ID of the marker point to attach animation to on parent/owner
	Asura_Hash_ID               m_uCommonItemHash;       // Hash-ID of the pickup

	AVPInventoryItemType		m_eItemType;

	float						m_fLastMessageTimestamp;	// Server timestamp for last handled message, to avoid reordering issues

	Asura_AnimationAttachment_ClientContainerInstance* m_pxContainerAttachment;

	// Protected functions...
	bool	DoesClientStateMatchServerState(const AVP_ClientEntity_Actor& xClient) const;

	// Request functions...
	bool	RequestIdling(const State eSourceState, const bool bInstant, const u_int uTransitionType = TT_NONE);
	bool	RequestUsing(const State eSourceState, const u_int uUsageType, const u_int uTransitionType = TT_NONE);
	bool	RequestInactive(const State eSourceState, const bool bInstant, const u_int uTransitionType = TT_NONE);

	bool	RequestIdling(const bool bInstant, const u_int uTransitionType = TT_NONE)	{ return(RequestIdling(GetCurrentState(), bInstant, uTransitionType)); }
	bool	RequestUsing(const u_int uUsageType, const u_int uTransitionType = TT_NONE)	{ return(RequestUsing(GetCurrentState(), uUsageType, uTransitionType)); }
	bool	RequestInactive(const bool bInstant, const u_int uTransitionType = TT_NONE)	{ return(RequestInactive(GetCurrentState(), bInstant, uTransitionType)); }

	bool	RequestNextIdlingState(void);
	bool	RequestNextUsingState(void);
	bool	RequestNextInactiveState(void);

	// Function to create start-move messages used by items...
	bool	CreateAbortMoveMessage(void);
	bool	CreateStartMoveMessage(const Asura_Hash_ID uMoveHash);

	// State update functions...
	virtual void	OnStateChange(const u_int uOldState, const u_int uNewState);

	// Abstract state functions...
	virtual void	UpdateClientState(AVP_ClientEntity_Actor& xClient)	= 0;
	virtual void	UpdateServerState(AVP_ServerEntity_Actor& xServer)	= 0;

	virtual AVP_InventoryItem_SubType* CreateSubType( const int iSubType, const Asura_Hash_ID uBlueprintHashID );

	friend class AVP_Actor_AnimationHandler;
	virtual void OnAttachItem( Asura_Animation* pxWeaponAnimation );	// Called when this item is attached to owner. (so we might create container and attach it to this object)
	virtual void OnDetachItem();										// Called just before item is detached from owner. (so we might clean up the detaching of container, etc...)

	// Virtual accessor functions...

	// Accessor functions...
	float	GetTimeInState(void) const	{ return(m_fTimeInState); }

	State	GetCurrentState(void) const		{ return(static_cast<State>(m_eCurrentState)); }
	State	GetPreviousState(void) const	{ return(static_cast<State>(m_ePreviousState)); }
	State	GetSourceState(void) const		{ return(static_cast<State>(m_eSourceState)); }
	State	GetDestinationState(void) const	{ return(static_cast<State>(m_eDestinationState)); }

	void	SetStateAsValid(const State eState)		{ m_uValidStateBits |= (1 << eState); }
	bool	IsStateValid(const State eState) const	{ return((m_uValidStateBits & (1 << eState)) != 0); }

	bool	IsActive(const State eState) const							{ return(eState != INACTIVE); }
	bool	IsInactive(const State eState) const						{ return(eState == INACTIVE); }
	bool	IsActivating(const State eState) const						{ return(eState == ACTIVATING); }
	bool	IsIdling(const State eState, const bool bFullyIdling) const	{ return((bFullyIdling) ? (eState == IDLING) : ((eState == BEGIN_IDLING) || (eState == IDLING) || (eState == END_IDLING) || (m_eDestinationState == IDLING) || (m_eSourceState == IDLING && ( m_eCurrentState != USING && m_eCurrentState != INACTIVE )))); }
	bool	IsUsing(const State eState, const bool bFullyUsing) const	{ return((bFullyUsing) ? (eState == USING) : ((eState == BEGIN_USING) || (eState == USING) || (eState == END_USING) || (m_eDestinationState == USING) || (m_eSourceState == USING && ( m_eCurrentState != IDLING && m_eCurrentState != INACTIVE )))); }
	bool	IsDeactivating(const State eState) const					{ return(eState == DEACTIVATING); }

	AVP_InventoryItem_SubType*			GetSubTypeFromUsageType(const u_int uUsageType)			{ const int iIndex = GetSubTypeIndexFromUsageType(uUsageType); return((iIndex >= 0) ? GetSubType(iIndex) : NULL); }
	const AVP_InventoryItem_SubType*	GetSubTypeFromUsageType(const u_int uUsageType) const	{ const int iIndex = GetSubTypeIndexFromUsageType(uUsageType); return((iIndex >= 0) ? GetSubType(iIndex) : NULL); }

	virtual void SetEntityActiveFlag(bool bActive);

	// Should not be used unser normal conditions. Required to force a change of usage type in the shotgun
	// from double to single shot if only one shell left.
	void SetCurrentUsageType(u_int uUsageType) { m_uCurrentUsageType = uUsageType; }

private:
	void GetStaticDataFromBlueprint();

	u_int	m_uValidStateBits;	// A bitfield used to specific which states are valid for this item.
	float	m_fTimeInState;		// The number of seconds we've been in this state.
	u_int	m_uSelectorDisplayConditionIndex; // For the guimenu to hide or show items in the selector menu page
	Asura_Hash_ID m_uContainerAttachment_MarkerPoint;	// Marker point on weapon to specify, where we should attach the container to it.

	union
	{
		u_int		m_uGeneralFlags;				// Single access point for the general-flags.

		struct
		{
			u_int	m_uNumSubTypes					 : 4,	// The number of sub-types within this item.
					m_bChangeOnZeroQuantity			 : 1,	// Flag to say whether we change item on zero quantity.
					m_bHidden						 : 1,	// forces removal of attachment (so AI can use keyboards etc)

													 : 0;	// 26 spare-bits
		};
	};

	union
	{
		u_int		m_uStateFlags;				// Single access point for the state-flags.

		struct
		{
			u_int	m_ePreviousState	: 4,	// The previous state of this item.
					m_eCurrentState		: 4,	// The current state of this item.
					m_eNewState			: 4,	// The state to change to.
					m_bHasNewState		: 1,	// Flag to say that we need to change state.

					m_eSourceState		: 4,	// The source state of this item (no necessarily the new one).
					m_eDestinationState	: 4,	// The destination state of this item (no necessarily the new one).

										: 0;	// 7 spare-bits
		};
	};

	union
	{
		u_int		m_uUsageFlags;						// Single access point for the usage-flags.

		struct  
		{
			u_int	m_uCurrentUsageType			: 4,	// The type of usage for this item.
					m_uCurrentTransitionType	: 4,	// The type of transition currently being performed.

					m_uPendingUsageType			: 4,	// The type of usage for this item.
					m_uPendingTransitionType	: 4,	// The type of transition currently being performed.

					m_uPreviousUsageType		: 4,	// The type of the previous usage

												: 0;	// 16 spare-bits
		};
	};

	// Private functions...
	AVP_Inventory_Item(void);
	AVP_Inventory_Item(const AVP_Inventory_Item& xOther);

	AVP_Inventory_Item&	operator=(const AVP_Inventory_Item& xOther);

	// NOTE: This function is private - all actual state-change requests are made via specific
	// functions which check the validity of the subsequent states avoiding invalid ones...
	bool	RequestStateChange(const State eNewState);
	void    FillCommonItemHashFromPickups(Asura_Hash_ID uItemHash);

	void CreateContainerAttachment();
	void DestroyContainerAttachment();

	static const u_int uCURRENT_CHUNK_VERSION = 1;
};   // End of class 'AVP_Inventory_Item'.


#endif   // ifndef __AVP_INVENTORY_ITEM_H__


/// End of file. ///
