
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_InventoryWeapon_Disc.cpp
// Author : Chris Wallace
// Date   : 04-Dec-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Inventory/Items/Weapons/AVP_InventoryWeapon_Disc.h"

// Asura independent header-files...
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"

// Game independent header-files...
#include "Actor/Vertebrates/AVP_Vertebrate_Focus.h"
#include "Actor/Vertebrates/Predator/AVP_Predator_AnimationHandler.h"
#include "Actor/Vertebrates/Predator/AVP_Predator_InputHandler.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Game_Core.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ClientActor_Vertebrate.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ServerActor_Vertebrate.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ClientActor_Predator.h"
#include "Entity/PhysicalObject/Projectile/AVP_ClientProjectile_Disc.h"
#include "Entity/PhysicalObject/Projectile/AVP_ServerProjectile_Disc.h"
#include "Entity/PhysicalObject/Projectile/AVP_Projectile_Messages.h"
#include "Entity/PhysicalObject/Projectile/AVP_GenericProjectileData.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/Weapons/AVP_RangedWeapon_ClientEffectsHandler.h"
#include "Moves/AVP_Movement_System.h"

DEBUGVAR float dbg_fMaxTimeUsing = 20.0f;

AVP_InventoryWeapon_Disc::AVP_InventoryWeapon_Disc( AVP_Inventory_Group& xParentGroup, const Asura_Hash_ID uHash, Asura_Hash_ID uBlueprintHash )
: PARENT( xParentGroup, uHash, uBlueprintHash )
, m_bIsInLaserTargetMode( false )
{
	m_eItemType = AVP_INVENTORYITEMTYPE_PREDATORDISC;
}

int AVP_InventoryWeapon_Disc::GetSubTypeIndexFromUsageType(const u_int uUsageType) const
{
	switch(uUsageType)
	{
		case UT_THROW:
		case UT_BOUNCE_THROW:
		{
			return( ST_PRIMARY );
		}

		case UT_NONE:
		case UT_RELOAD_ALL:
		case UT_MELEE:
		{
			return(-1);
		}

		default:
		{
			Asura_Assert(false, "Unexpected usage-type for this inventory-item.");
			return(-1);
		}
	}
}

void AVP_InventoryWeapon_Disc::StartUsing( const u_int uUsageType )
{
	PARENT::StartUsing( uUsageType );

	switch( uUsageType )
	{
		case UT_THROW:
		case UT_BOUNCE_THROW:
		{
			HandleStartFiringEffects( uUsageType );
		}
		break;

		default:
		{
			// do nothing
		}
		break;
	}
}

void AVP_InventoryWeapon_Disc::StopUsing( const u_int uUsageType )
{
	switch ( uUsageType )
	{
		case UT_THROW:
		case UT_BOUNCE_THROW:
		{
			HandleStopFiringEffects( uUsageType, true );

			AVP_ClientEntity_Actor* pxActor = GetParentSystem().GetParentClient();
			AVP_Actor_AnimationHandler* pxAnimationHandler = pxActor ? pxActor->GetAnimationHandler() : NULL;
			if ( pxAnimationHandler != NULL )
			{
				pxAnimationHandler->AttachItem( this );
			}

			AVP_InventoryItem_SubType* pxSubType = GetSubTypeFromUsageType( GetCurrentUsageType() );
			pxSubType->SetCurrentQuantity( 1.0f ); // if we increase, we could end up with  more than one...
		}
		break;

		default:
		{
			// do nothing
		}
		break;
	}
}

void AVP_InventoryWeapon_Disc::SendCreateProjectileMessage( const AVP_GenericProjectileData* pxProjectileData, const AVP_RangedWeapon_SubType* pxSubType, u_int uUsageType, Asura_Hash_ID uMuzMP, float fRange, float fMaxRange, const Asura_Vector_3& xActorVel, const Asura_Vector_3& xFirePosition, const Asura_Vector_3& xShotDirection, const Asura_Matrix_3x3& xShotOrientation, Asura_Guid* puTargetGuid /*=NULL*/ )
{
	Asura_Guid uOwnerGuid = GetParentSystem().GetOwnerGuid();
	if ( GetParentSystem().IsMaster() )
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_CREATE_PROJECTILE_DISC, sizeof( AVP_Message_CreateProjectile_Disc ) );
		if ( pxMessage != NULL )
		{
			AVP_Message_CreateProjectile_Disc* pxData = static_cast< AVP_Message_CreateProjectile_Disc* >( pxMessage->GetData() );
			
			// Fill-in the message...
			pxData->SetProjectileBlueprintHash( pxProjectileData->GetProjectileHash() );
			pxData->SetPosition( xFirePosition );
			pxData->SetOrientation( Asura_Quat(xShotOrientation) );
			pxData->SetActorVelocity( xActorVel );

			float fDamageScale = 1.0f;
			if ( uUsageType == UT_CHARGED_SHOT_PRIMARY || uUsageType == UT_CHARGED_SHOT_SECONDARY )
			{
				fDamageScale = pxSubType->GetDamageMultiplierPerCharge() * m_fChargedPower;
			}
			pxData->SetDamageScale( fDamageScale );
			pxData->SetLaunch( true );
			// TODO we might need to do the local handling hack to make this more responsive on remote clients
			pxData->SetBounceMode( uUsageType == UT_BOUNCE_THROW );
			pxData->SetLaserTargetMode( IsInLaserTargetMode() );

			if(puTargetGuid)
			{
				pxData->SetTargetGuid(*puTargetGuid);
			}
			else
			{
				pxData->SetTargetGuid(ASURA_GUID_UNREGISTERED);
			}
						
			pxMessage->SetToGuid( uOwnerGuid );
			pxMessage->SetFromGuid( uOwnerGuid );

/*			// EP - 27 February 2009 - This code is working but it is blocked. Because disc is not following specific targets any more. Instead it is following predator's laser. But I am keeping this code for awhile, because I am not sure if designers didn't change their mind again. :-)
			AVP_ClientActor_Predator* pxPred = AVP_ClientActor_Predator::FindEntity( GetParentSystem().GetOwnerGuid() );
			AVP_Vertebrate_Focus* pxFocus = pxPred != NULL ? pxPred->GetFocus() : NULL;
			if ( pxFocus != NULL && uUsageType != UT_BOUNCE_THROW )
			{
				pxData->SetTargetData( pxFocus->GetLockOnTarget(), 0xe5a85d03 ); // target_1
			}
			else
			{
				pxData->SetTargetData( ASURA_GUID_UNREGISTERED, ASURA_HASH_ID_UNSET );
			}
*/

			Asura_Network::Send( pxMessage );
		}
	}
}


void AVP_InventoryWeapon_Disc::SendTurnLaserTrackingMessageToPredator( bool bTurnLaserTrackingOn )
{
	if ( GetParentSystem().IsMaster() )
	{
		Asura_Message* pxMessage = NULL;
		
		if( bTurnLaserTrackingOn )
		{
			pxMessage = Asura_Network::CreateMessage( AVP_MSGID_DISC_LASER_TRACKING_ON );
		}
		else
		{
			pxMessage = Asura_Network::CreateMessage( AVP_MSGID_DISC_LASER_TRACKING_OFF );
		}

		if ( pxMessage )
		{
			pxMessage->SetFromGuid( GetParentSystem().GetOwnerGuid() );
			pxMessage->SetToGuid( GetParentSystem().GetOwnerGuid() );

			Asura_Network::Send( pxMessage );
		}
	}
}


void AVP_InventoryWeapon_Disc::SendRetrieveDiscMessage(void)
{
	if ( GetParentSystem().IsMaster() )
	{
		Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_RETURN_DISC );
		if ( pxMessage )
		{
			pxMessage->SetFromGuid( GetParentSystem().GetOwnerGuid() );
			pxMessage->SetToGuid( GetEntityGuid() );

			Asura_Network::Send( pxMessage );
		}
	}
}

void AVP_InventoryWeapon_Disc::ProcessPlayerInput( const AVP_ClientEntity_Actor& xActor )
{
	const AVP_Actor_InputHandler* pxInputHandler = xActor.GetInputHandler();
	const State eCurrentState = GetCurrentState();
	if ( pxInputHandler != NULL && ( GetTimeInState() > 0.0f ) )
	{
		if ( pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::PRIMARY_ATTACK_TAP ) > 0.0f )
		{
			if ( eCurrentState == USING && m_ppxSubTypes[ ST_PRIMARY ] && m_ppxSubTypes[ ST_PRIMARY ]->IsEmpty() )
			{
				SendRetrieveDiscMessage();
			}
			else if ( eCurrentState == IDLING )
			{
				RequestUsing( UT_BOUNCE_THROW );	// UT_THROW	// All throws should be bouncing.
			}
		}

		if ( pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::SECONDARY_ATTACK_TAP ) > 0.0f ) // start laser aiming mode on tap
		{
			if ( IsInLaserTargetMode()==false )
			{
				SetLaserTargetMode( true );

				SendTurnLaserTrackingMessageToPredator( true );	// Send information to predator that disk needs to switch its behavior.
				return;
			}
		}
		else if ( pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::SECONDARY_ATTACK_HELD) == 0.0f ) // stop laser aiming on release
		{
			if ( IsInLaserTargetMode() )
			{
				SetLaserTargetMode( false );

				SendTurnLaserTrackingMessageToPredator( false );
				return;
			}
		}
	}
}

void AVP_InventoryWeapon_Disc::CalculateFiringDataForSubType( const AVP_ClientEntity_Actor& xActor, Asura_Hash_ID uMuzMP, float fMuzVel, float fMaxRange, Asura_Vector_3& xFirePosition, Asura_Vector_3& xFireDirection, Asura_Matrix_3x3& xMuzzleOri, float& fRange, Asura_Guid* puTargetGuid /*= NULL*/ )
{
	if(!xActor.GetMarkerLocationFromHash( uMuzMP, &xFirePosition, &xMuzzleOri ))
	{
		xFirePosition	= xActor.GetCurrentPosition();
		xMuzzleOri		= xActor.GetCurrentOrientation();
	}
	fRange = fMaxRange;
	Asura_AVector3 xAimDir;
	xActor.GetAimDirection( xAimDir );
	xFireDirection = xAimDir;
}

void AVP_InventoryWeapon_Disc::UpdateClientState( AVP_ClientEntity_Actor& xClient )
{
	PARENT::UpdateClientState( xClient );

	AVP_InventoryItem_SubType* pxSubType = GetSubTypeFromUsageType( GetCurrentUsageType() );
	AVP_Actor_AnimationHandler* pxAnimationHandler = xClient.GetAnimationHandler();
	bool bAction = ( m_pxClientEffectsHandler != NULL && m_pxClientEffectsHandler->GetActorAnimFlagsForLastFrame( xClient ) & AVP_ANIMFLAG_ACTION_1 ) != 0;

	if ( bAction )
	{
		if ( GetCurrentState() == USING && ( GetCurrentUsageType() == UT_BOUNCE_THROW || GetCurrentUsageType() == UT_THROW ) && pxSubType->GetCurrentQuantity() > 0.0f )
		{
			if ( pxAnimationHandler != NULL )
			{
				pxAnimationHandler->DetachItem( this );
				DestroyAnimation();
			}
			
			if ( GetParentSystem().IsMaster() )
			{
				Use();
			}
			pxSubType->SetCurrentQuantity( 0.0f );
		}
	}

	if ( GetCurrentState() == USING && GetTimeInState() > dbg_fMaxTimeUsing )
	{
		RequestIdling( false );
	}
}

void AVP_InventoryWeapon_Disc::UpdateServerState( AVP_ServerEntity_Actor& xServer )
{
	PARENT::UpdateServerState( xServer );

	AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity( GetParentSystem().GetOwnerGuid() );
	if (!pxClient) { return; }

	bool bAction = false;
	const AVP_Inventory_System* pxInventory = pxClient->GetInventorySystem();
	if (pxInventory)
	{
		const AVP_InventoryWeapon_Disc* pxClientDiscItem = static_cast< const AVP_InventoryWeapon_Disc* >( pxInventory->GetItem( m_uBlueprintHash ) );
		if ( pxClientDiscItem != NULL )
		{
			bAction = ( pxClientDiscItem->m_pxClientEffectsHandler != NULL && pxClientDiscItem->m_pxClientEffectsHandler->GetActorAnimFlagsForLastFrame(  *pxClient ) & AVP_ANIMFLAG_ACTION_1 ) != 0;
		}	
	}
	
	if ( bAction )
	{
		AVP_InventoryItem_SubType* pxSubType = GetSubTypeFromUsageType( GetCurrentUsageType() );

		if ( GetCurrentState() == USING && ( GetCurrentUsageType() == UT_BOUNCE_THROW || GetCurrentUsageType() == UT_THROW ) && pxSubType && (pxSubType->GetCurrentQuantity() > 0.0f) )
		{
			if ( GetParentSystem().IsMaster() )
			{
				Use();
			}
			pxSubType->SetCurrentQuantity( 0.0f );
		}
	}

	if ( GetCurrentState() == USING && GetTimeInState() > dbg_fMaxTimeUsing )
	{
		RequestIdling( false );
	}
}


void AVP_InventoryWeapon_Disc::SetLaserTargetMode( bool bLaserTargetMode )	
{

	m_bIsInLaserTargetMode = bLaserTargetMode;


	// Inform disc about current change targetting.
	AVP_ServerProjectile_Disc* pxDiscProjectile = AVP_ServerProjectile_Disc::FindEntity( GetEntityGuid() );
	if( pxDiscProjectile )
	{
		pxDiscProjectile->SetLaserTargetMode( bLaserTargetMode );
	}

	// Also hide or show laser
	if( GetParentSystem().IsClientSide() && m_pxClientEffectsHandler )
	{

		if( m_bIsInLaserTargetMode )
		{
			const AVP_Inventory_System&	xParentSystem	= GetParentSystem();
			AVP_ClientEntity_Actor* pxClient( AVP_ClientEntity_Actor::FindEntity( xParentSystem.GetOwnerGuid() ) );
			AVP_Actor_AnimationHandler* pxAnimHandler	= pxClient != NULL ? pxClient->GetAnimationHandler() : NULL;
			Asura_Animation*			pxAnimation		= pxAnimHandler != NULL ? pxAnimHandler->GetAnimation() : NULL;
		
			if( pxAnimation )
			{
				m_pxClientEffectsHandler->ActivateLaser( *pxAnimation );
			}
		}
		else
		{
			m_pxClientEffectsHandler->DeactivateLaser();
		}
	}
}



/// End of file. ///
