
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_InventoryWeapon_SmarGun.cpp
// Author : Chris Wallace
// Date   : 22-Sept-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Inventory/Items/Weapons/AVP_InventoryWeapon_SmartGun.h"

// Asura independent header-files...
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "Sound/Asura_Sound_Event_System.h"

// Game independent header-files...
#include "Actor/Vertebrates/AVP_Vertebrate_InputHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Focus.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Settings.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Inventory/AVP_AmmoStore.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/Weapons/AVP_RangedWeapon_ClientEffectsHandler.h"
#include "GUIMenu/AVP_GUIMenu_System.h"
#include "GUIMenu/Asura_GUIMenu_Root.h"
#include "GUIMenu/Asura_GUIMenu_Element_Image.h"

DEBUGVAR float dbg_fSpreadLerpRate = 0.2f;

static const Asura_Hash_ID	s_uBellyBoxMarkerHash = 0x9906a463;	// smartgun_box
static const int			s_aiArmIKBone[4] = { 2, 3, 4, -1 }; // Arm_2, Arm_1, BellyBox

AVP_InventoryWeapon_SmartGun::AVP_InventoryWeapon_SmartGun( AVP_Inventory_Group& xParentGroup, const Asura_Hash_ID uHash, const Asura_Hash_ID uBlueprintHash )
: AVP_InventoryWeapon_RangedWeapon(xParentGroup, uHash, uBlueprintHash)
, m_bTrackingMode(false)
, m_fRange(1000.f)
, m_fSpinAcceleration(15.0f/6.0f)
, m_fSpinDeceleration(3.0f/6.0f)
, m_fSpinSpeed(0.0f)
, m_fSpinMaxSpeed(10.0f/6.0f)
, m_fSpinAngle(0.0f)
, m_fSpinInputScale(1.0f)
, m_uSpinLoopSoundHash(ASURA_HASH_ID_UNSET)
, m_iSpinLoopSoundHandle(ASURA_INVALID_HANDLE)
{
	m_eItemType = AVP_INVENTORYITEMTYPE_SMARTGUN;
	m_xMinigunDirLastFrame.Zero();
	m_xFirePosition.Zero();
	m_xMuzzleOri.Identity();
}

void AVP_InventoryWeapon_SmartGun::StartUsing( const u_int uUsageType )
{
	PARENT::StartUsing( uUsageType );

	switch( uUsageType )
	{
		case UT_N_BULLET_BURST_PRIMARY:
		case UT_FULL_AUTO_PRIMARY:
		{
			m_xMinigunDirLastFrame.Zero(); // reset on start firing
		}
		break;

		default:
		{
			// do nothing
		}
		break;
	}
}

void AVP_InventoryWeapon_SmartGun::InitialiseFromBlueprint()
{
	PARENT::InitialiseFromBlueprint();

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, GetBlueprintHash() );

	if ( pxBlueprint )
	{
		Asura_Value xValue;

		if ( pxBlueprint->GetParameterValueByHashID( 0xa00782ac /*"Spin.Acceleration"*/, xValue ) )
		{
			m_fSpinAcceleration = xValue.GetFloatValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( 0xe0e5d84b /*"Spin.Deceleration"*/, xValue ) )
		{
			m_fSpinDeceleration = xValue.GetFloatValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( 0x49a73d8f /*"Spin.MaxSpeed"*/, xValue ) )
		{
			m_fSpinMaxSpeed = xValue.GetFloatValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( 0xF20117F5 /*"SoundSpinLoop"*/, xValue ) )
		{
			m_uSpinLoopSoundHash = xValue.GetHashIDValue();
		}
	}
}

void AVP_InventoryWeapon_SmartGun::HandleStartFiringEffects( const u_int uUsageType )
{
	// Overridden so don't play weapon anim
	if ( m_pxClientEffectsHandler != NULL )
	{
		m_pxClientEffectsHandler->SetTracerCount( 0 );

		int iSubType = GetSubTypeIndexFromUsageType( uUsageType );
		if ( iSubType >= 0 )
		{
			const SubType eSubType = static_cast< SubType >( iSubType );
			AVP_ClientEntity_Actor* pxOwner = GetParentSystem().GetParentClient();
			Asura_Assert( pxOwner != NULL, "No client owner" );
			if ( pxOwner != NULL )
			{
				AVP_RangedWeapon_SubType* pxSubType = static_cast<AVP_RangedWeapon_SubType*>(GetSubType(iSubType));
				if((GetCurrentUsageType() != UT_FULL_AUTO_PRIMARY && GetCurrentUsageType() != UT_N_BULLET_BURST_PRIMARY) || (pxSubType && (!pxSubType->IsEmpty())))
				{
					m_pxClientEffectsHandler->PlaySound( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_BEGIN, eSubType );
					m_pxClientEffectsHandler->PlayActorAnim( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_BEGIN, eSubType );
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryWeapon_SmartGun::HandleFireEffects
// Author       : Andy Tate
// Date         : 20-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryWeapon_SmartGun::HandleFireEffects( const u_int uUsageType, const Asura_Vector_3& xFirePosition, const Asura_Matrix_3x3& xFireOrientation, const float fRange, const bool bFirstShot )
{
	if ( m_pxClientEffectsHandler != NULL )
	{
		const SubType eSubType = static_cast< SubType >( GetSubTypeIndexFromUsageType( uUsageType ) );

		if ( bFirstShot )
		{
			AVP_ClientEntity_Actor*		pxOwner			= GetParentSystem().GetParentClient();
			AVP_Actor_InputHandler*		pxInputHandler	= pxOwner ? pxOwner->GetInputHandler() : NULL;
			AVP_Actor_AnimationHandler* pxAnimHandler	= pxOwner != NULL ? pxOwner->GetAnimationHandler() : NULL;
			Asura_Animation*			pxAnimation		= pxAnimHandler != NULL ? pxAnimHandler->GetAnimation() : NULL;
			Asura_Assert ( pxAnimation != NULL, "No owner client animation found" );
			if ( pxAnimation != NULL )
			{
				// Muzzle flash PFX
				m_pxClientEffectsHandler->DoMuzzleFlash( AVP_RangedWeapon_ClientEffectsHandler::FLASH_FIRE, *pxOwner, *pxAnimation, xFirePosition, xFireOrientation, eSubType );

				// Also play muzzle flash "Hot" when certain number of bullets left the barrel already.
				if( m_bCanPlayHotMuzzleFlash )
				{
					float fMuzzleFlash_Hot_NumBulletsToBegin = m_pxClientEffectsHandler->GetWeaponHotMuzzleFlash_NumBulletsToBegin(eSubType);
					if( GetAmmoFiredSoFar()>=fMuzzleFlash_Hot_NumBulletsToBegin && fMuzzleFlash_Hot_NumBulletsToBegin!=0.0f )
					{
						m_bCanPlayHotMuzzleFlash = false;
						m_pxClientEffectsHandler->DoMuzzleFlash( AVP_RangedWeapon_ClientEffectsHandler::FLASH_HOT, *pxOwner, *pxAnimation, xFirePosition, xFireOrientation, eSubType );
					}
				}
			}

			AVP_RangedWeapon_SubType* pxSubType = static_cast<AVP_RangedWeapon_SubType*>(GetSubType(eSubType));
			if(pxSubType)
			{
				if (pxSubType->GetCurrentQuantity() > 1.0f)
				{
					// Sound
					if ( !m_pxClientEffectsHandler->IsSoundPlaying( AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_BEGIN, eSubType ) )
					{
						if ( m_pxClientEffectsHandler->IsSoundPlaying( AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType ) )
						{
							m_pxClientEffectsHandler->UpdateSoundPosition( xFirePosition, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType );
						}
						else
						{
							m_pxClientEffectsHandler->PlaySound( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType );
						}
					}
				}

				if ( pxInputHandler )
				{
					pxInputHandler->AddRecoilVelocity( pxSubType->GetWeaponRecoil() );
				}
			}

			// Actor Anim
			if(!m_pxClientEffectsHandler->IsActorAnimPlaying(*pxOwner,AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_BEGIN, eSubType ))
			{
				m_pxClientEffectsHandler->PlayActorAnim( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType );
			}

			m_pxClientEffectsHandler->DoRecoilCameraShake(GetParentSystem().GetOwnerGuid(), eSubType);
		}

		if ( ((AVP_RangedWeapon_SubType*)GetSubType( eSubType ))->GetRateOfFire() >= 1.0f )
		{
			// Beam effect
			m_pxClientEffectsHandler->DoTracerBeam( xFirePosition, xFireOrientation, fRange, eSubType, 1 );
		}
		else
		{
			// Beam effect
			m_pxClientEffectsHandler->DoTracerBeam( xFirePosition, xFireOrientation, fRange, eSubType, 2 );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryWeapon_SmartGun::HandleStopFiringEffects
// Author       : Andy Tate
// Date         : 27-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryWeapon_SmartGun::HandleStopFiringEffects( const u_int uUsageType, const bool bStopWepaonAnims )
{
	if ( m_pxClientEffectsHandler != NULL )
	{
		AVP_ClientEntity_Actor*		pxOwner			= GetParentSystem().GetParentClient();
		Asura_Assert( pxOwner != NULL, "No client owner" );
		if ( pxOwner != NULL )
		{
			const SubType eSubType = static_cast< SubType >( GetSubTypeIndexFromUsageType( uUsageType ) );

			m_pxClientEffectsHandler->StopChargePFX( eSubType );
			if ( m_pxClientEffectsHandler->IsActorAnimPlaying( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType ) ||
				m_pxClientEffectsHandler->IsActorAnimPlaying( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_BEGIN, eSubType ))
			{
				m_pxClientEffectsHandler->StopActorAnim( *pxOwner, true );
			}
	
			if(!m_bFirstShot) // Don't play these if we never got off the first frame.
			{
				if(m_pxClientEffectsHandler->IsSoundPlaying(AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType))
				{
					m_pxClientEffectsHandler->PlaySound( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_END, eSubType );
				}

				if ( !m_pxClientEffectsHandler->IsActorAnimPlaying( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_END, eSubType ) )
				{
					m_pxClientEffectsHandler->PlayActorAnim( *pxOwner, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_END, eSubType );
				}
			}

			/// Stop loop effects
			m_pxClientEffectsHandler->StopSound( AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, eSubType );
			m_pxClientEffectsHandler->StopSound( AVP_RangedWeapon_ClientEffectsHandler::EFFECT_CHARGING, eSubType );
		}
	}
}

void AVP_InventoryWeapon_SmartGun::ProcessPlayerInput( const AVP_ClientEntity_Actor& xActor )
{
	const bool bNoAmmo = m_ppxSubTypes[ ST_PRIMARY ]->IsEmpty();
	const AVP_Actor_InputHandler* pxInputHandler = xActor.GetInputHandler();
	const State eCurrentState = GetCurrentState();

	bool bAllowReload = ShouldAllowReload();

	if ( pxInputHandler != NULL && ( GetTimeInState() > 0.0f ) && !xActor.IsMovingFast() )
	{
		if ( ( pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::RELOAD ) > 0.0f ) && bAllowReload && CanReload() )
		{
			RequestUsing( UT_RELOAD_ALL );
		}
		else
		{
			// Store this so we can do half spin on consoles
			m_fSpinInputScale = pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::PRIMARY_ATTACK_HELD );

			if ( pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::PRIMARY_ATTACK_HELD ) > 0.0f && ( !bNoAmmo ) )
			{
				if ( eCurrentState == IDLING )
				{
					RequestUsing( UT_FULL_AUTO_PRIMARY );
				}
			}
			
			if ( pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::SECONDARY_ATTACK_TAP ) > 0.0f)
			{
				SetTrackingMode(!GetTrackingMode());
			}
		}
	}
}

void AVP_InventoryWeapon_SmartGun::OnStateChange(const u_int uOldState,	const u_int uNewState)
{
	if(uNewState == INACTIVE)
	{
		m_fSpinSpeed = 0.0f; // stop the spin
	}
	PARENT::OnStateChange(uOldState, uNewState);
}

void AVP_InventoryWeapon_SmartGun::UpdateClientState( AVP_ClientEntity_Actor& xClient )
{
	const State	eCurrentState	= GetCurrentState();
	const u_int	uUsageType		= GetCurrentUsageType();
	int			iSubType		= ST_PRIMARY;//GetSubTypeIndexFromUsageType( uUsageType );

	// sanity check
	if(iSubType > static_cast<int>(m_uNumSubTypes) ) 
	{
		Asura_Assert(false, "Updating a Minigun that only has one subtype, this should never happen. Are the Blueprints correct?");
		return;
	}

	AVP_RangedWeapon_SubType*	pxSubType	= iSubType < 0 ? NULL : static_cast< AVP_RangedWeapon_SubType* >( m_ppxSubTypes[ iSubType ] );

	AVP_InventoryItem_Weapon::UpdateClientState( xClient ); // skip ranged wepaon, we cover it all here

	// decrement the player shot timers
	if ( m_fPrimaryShotTimer > 0.0f )
	{
		m_fPrimaryShotTimer -= Asura_Timers::GetGameTimeSlice();
	}

	if ( m_fSecondaryShotTimer > 0.0f )
	{
		m_fSecondaryShotTimer -= Asura_Timers::GetGameTimeSlice();
	}

	// handle weapon cool down
	if ( m_fSpreadCoolDownTimer > 0.0f )
	{
		m_fSpreadCoolDownTimer -= Asura_Timers::GetGameTimeSlice();
		m_fCurrentSpreadParametric = Asura_Maths::Lerp( m_fCurrentSpreadParametric, m_fTargetSpreadParametric, dbg_fSpreadLerpRate );
	}
	else
	{
		m_fTargetSpreadParametric = m_fCurrentSpreadParametric = Asura_Maths::ClampToRange( m_fCurrentSpreadParametric - m_fSpreadCoolDown * Asura_Timers::GetGameTimeSlice(), 0.0f, 1.0f );
	}

	// Update lock on laser target effect
	UpdateLaser( xClient );

	// if we're moving fast then stop the actor anim
	if ( xClient.IsMovingFast() && m_pxClientEffectsHandler != NULL && uUsageType != UT_RELOAD_ALL )
	{
		m_pxClientEffectsHandler->DestroyActorAnim( xClient );
	}

	if ( eCurrentState != INACTIVE )
	{
		UpdateTargetLock( xClient );
		UpdateSpin(eCurrentState, uUsageType);
	}

	switch ( eCurrentState )
	{
		case IDLING:
		{
			if ( ( GetTimeInState() > 0.0f ) && !xClient.IsMovingFast() )
			{
				CheckForAutomaticReload();
			}
		}
		break;

		case USING:
		{
			m_bFirstShot = false; // As this weapon spins up, we don't want a first shot as such, it causes issues.

			switch(uUsageType)
			{
				case UT_N_BULLET_BURST_PRIMARY:
				{
					if ( ( xClient.IsMovingFast() || (!IsPlayerIsStillFiring( xClient, iSubType ) ) || (!pxSubType || pxSubType->IsEmpty()) ) && GetTimeInState() > 0.0f )
					{
						m_fBurstBulletsToFire = 0.0f;

						if ( GetParentSystem().IsMaster() )
						{
							RequestIdling(false);
						}
					}
					else
					{
						const AVP_Actor_InputHandler* pxInputHandler = xClient.GetInputHandler();

						if(GetParentSystem().IsMaster())
						{
							if(pxInputHandler)
							{
								if((pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::PRIMARY_ATTACK_HELD )) > 0.0f)
								{
									if(m_fSpinSpeed >= m_fSpinMaxSpeed)
									{
										m_fSpinSpeed = m_fSpinMaxSpeed;

										if(m_fBurstBulletsToFire > 0.0f)
										{
											Use();
										}
										else
										{
											if ( !m_pxClientEffectsHandler->IsActorAnimPlaying( xClient ) && GetTimeInState() > 0.0f )
											{
												m_fBurstBulletsToFire = 0.0f;
												RequestIdling(false);
											}
										}
									}
								}
							}
						}
						else
						{
							if (m_fSpinSpeed >= m_fSpinMaxSpeed)
							{
								m_fSpinSpeed = m_fSpinMaxSpeed;
								
								if (m_fBurstBulletsToFire > 0.0f)
								{
									Use();
								}
								else
								{
									if ( !m_pxClientEffectsHandler->IsActorAnimPlaying( xClient ) && GetTimeInState() > 0.0f )
									{
										m_fBurstBulletsToFire = 0.0f;
										//RequestIdling(false);
									}
								}
							}
						}
					}
				}
				break;

				case UT_FULL_AUTO_PRIMARY:
				{
					if ( ( xClient.IsMovingFast() || (!IsPlayerIsStillFiring( xClient, iSubType ) ) || (!pxSubType || pxSubType->IsEmpty()) ) && GetTimeInState() > 0.0f )
					{
						if ( GetParentSystem().IsMaster() )
						{
							RequestIdling(false);
						}
					}
					else
					{
						const AVP_Actor_InputHandler* pxInputHandler = xClient.GetInputHandler();

						if(GetParentSystem().IsMaster())
						{
							if(pxInputHandler)
							{
								if((pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::PRIMARY_ATTACK_HELD )) > 0.0f)
								{
									if(m_fSpinSpeed >= m_fSpinMaxSpeed)
									{
										m_fSpinSpeed = m_fSpinMaxSpeed;
										Use();
									}
									else
									{
										if ( m_pxClientEffectsHandler->IsActorAnimPlaying( xClient, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, static_cast<SubType>(iSubType) ) ||
											m_pxClientEffectsHandler->IsActorAnimPlaying( xClient, AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_BEGIN, static_cast<SubType>(iSubType) ))
										{
											m_pxClientEffectsHandler->StopActorAnim( xClient, true );
										}

										m_pxClientEffectsHandler->StopSound( AVP_RangedWeapon_ClientEffectsHandler::EFFECT_FIRE_LOOP, static_cast<SubType>(iSubType) );
										m_pxClientEffectsHandler->StopSound( AVP_RangedWeapon_ClientEffectsHandler::EFFECT_CHARGING, static_cast<SubType>(iSubType) );
									}
								}
							}
						}
						else
						{
							if(m_fSpinSpeed >= m_fSpinMaxSpeed)
							{
								m_fSpinSpeed = m_fSpinMaxSpeed;
								Use();
							}
						}
					}
				}
				break;

				case UT_RELOAD_ALL:
				{
					// If the reload anim is finished...
					if( ( !m_pxClientEffectsHandler->IsActorAnimPlaying( xClient ) || m_pxClientEffectsHandler->GetActorAnimFlagsForLastFrame( *GetParentSystem().GetParentClient() ) & AVP_ANIMFLAG_ACTION_1 ) != 0 ) // Ugh!
					{
						if(!m_bReloadHandled)
						{
							Use();
							m_bReloadHandled = true;
						}
					}
					if( !m_pxClientEffectsHandler->IsActorAnimPlaying( xClient ) )
					{
						if(GetParentSystem().IsMaster())
						{
							RequestIdling(false);
						}
						m_bReloadHandled = false;
					}
				}
				break;

				default:
				{
					break; // don't do anything, the derived class will probably handle it
				}
			}		
		}

		default:
			break;
	}
}

void AVP_InventoryWeapon_SmartGun::UpdateServerState( AVP_ServerEntity_Actor& xServer )
{
	AVP_InventoryItem_Weapon::UpdateServerState( xServer ); // Skip ranged weapon, all handled here

	const State	eCurrentState					= GetCurrentState();
	const AVP_Inventory_System&	xParentSystem	= GetParentSystem();
	const u_int	uUsageType						= GetCurrentUsageType();
	const int	iSubType						= ST_PRIMARY;//GetSubTypeIndexFromUsageType( uUsageType );

	if( iSubType > static_cast<int>(m_uNumSubTypes) ) 
	{
		Asura_Assert(false, "Updating a Minigun that only has one subtype, this should never happen. Are the Blueprints correct?");
		return;
	}

	AVP_RangedWeapon_SubType*	pxSubType		= iSubType < 0 ? NULL : static_cast< AVP_RangedWeapon_SubType* >( m_ppxSubTypes[ iSubType ] );

	// cooldown
	if ( m_fSpreadCoolDownTimer > 0.0f )
	{
		m_fSpreadCoolDownTimer -= Asura_Timers::GetGameTimeSlice();
	}
	else
	{
		m_fCurrentSpreadParametric = Asura_Maths::ClampToRange( m_fCurrentSpreadParametric - m_fSpreadCoolDown * Asura_Timers::GetGameTimeSlice(), 0.0f, 1.0f );
	}

	if ( xParentSystem.IsMaster() ) // if we're the master system ( this will be the case for the AI )
	{
		// Decide what to do based on our current state...
		switch ( eCurrentState )
		{
			case USING:
			{
				switch(uUsageType)
				{
					case UT_N_BULLET_BURST_PRIMARY:
					{
						AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(GetParentSystem().GetOwnerGuid());
						AVP_Inventory_System* pxClientSystem = (pxClient)?pxClient->GetInventorySystem():NULL;
						AVP_InventoryWeapon_SmartGun* pxWeapon = (pxClientSystem)?static_cast<AVP_InventoryWeapon_SmartGun*>(pxClientSystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON)):NULL;

						if ( pxWeapon && pxWeapon->m_fSpinSpeed >= pxWeapon->m_fSpinMaxSpeed )
						{
							if (m_fBurstBulletsToFire > 0.0f)
							{
								Use();
							}
							else
							{
								if ( !HasPendingStateChange() && GetTimeInState() > 0.0f )
								{
									m_fBurstBulletsToFire = 0.0f;
									RequestIdling( false );
								}
							}
						}

						if( ( pxSubType != NULL ? pxSubType->IsEmpty() : true ) && GetTimeInState() > 0.0f )
						{
							m_fBurstBulletsToFire = 0.0f;
							RequestIdling(false);
						}
						break;
					}

					case UT_FULL_AUTO_PRIMARY:
					{
						AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(GetParentSystem().GetOwnerGuid());
						AVP_Inventory_System* pxClientSystem = (pxClient)?pxClient->GetInventorySystem():NULL;
						AVP_InventoryWeapon_SmartGun* pxWeapon = (pxClientSystem)?static_cast<AVP_InventoryWeapon_SmartGun*>(pxClientSystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON)):NULL;

						if(pxWeapon && pxWeapon->m_fSpinSpeed >= pxWeapon->m_fSpinMaxSpeed)
						{
							Use();
						}

						if( ( pxSubType != NULL ? pxSubType->IsEmpty() : true ) && GetTimeInState() > 0.0f )
						{
							RequestIdling(false);
						}
						break;
					}

					case UT_RELOAD_ALL:
					{
						if ( !HasPendingStateChange() && GetTimeInState() > 0.0f )
						{
							const AVP_ClientEntity_Actor* pxClient( AVP_ClientEntity_Actor::FindEntity( xParentSystem.GetOwnerGuid() ) );
							const AVP_Inventory_System* pxClientSystem( pxClient != NULL ? pxClient->GetInventorySystem() : NULL );
							const AVP_InventoryWeapon_RangedWeapon* pxClientWeapon( static_cast< const AVP_InventoryWeapon_RangedWeapon* >( pxClientSystem != NULL ? pxClientSystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL ) );
							const AVP_RangedWeapon_ClientEffectsHandler* pxClientEffectsHandler( pxClientWeapon != NULL ? pxClientWeapon->GetClientEffectsHandler() : NULL );
							if ( ( pxClientEffectsHandler == NULL || !pxClientEffectsHandler->IsActorAnimPlaying( *pxClient ) || ( ( pxClientEffectsHandler->GetActorAnimFlagsForLastFrame( *pxClient ) & AVP_ANIMFLAG_ACTION_1 ) != 0 ) ) )// Ugh!
							{
								Use();
								RequestIdling( false );
							}
						}
					}
					break;
			
					default:
					{
						Asura_Assertf( false, ("Unhandled case, usage type: %d", uUsageType));
						break;
					}
					
				}	
			}
			break;

			case END_USING:
			{
				const AVP_ClientEntity_Actor* pxClient( AVP_ClientEntity_Actor::FindEntity( xParentSystem.GetOwnerGuid() ) );
				const AVP_Inventory_System* pxClientSystem( pxClient != NULL ? pxClient->GetInventorySystem() : NULL );
				const AVP_InventoryWeapon_SmartGun* pxClientWeapon( static_cast< const AVP_InventoryWeapon_SmartGun* >( pxClientSystem != NULL ? pxClientSystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL ) );
				const AVP_RangedWeapon_ClientEffectsHandler* pxClientEffectsHandler( pxClientWeapon != NULL ? pxClientWeapon->m_pxClientEffectsHandler : NULL );
				if ( ( pxClientEffectsHandler == NULL || !pxClientEffectsHandler->IsActorAnimPlaying( *pxClient ) ) && GetTimeInState() > 0.0f )// Ugh!
				{
					if ( GetDestinationState() == INACTIVE )
					{
						RequestNextInactiveState();
					}
					else
					{
						RequestNextIdlingState();
					}
				}			
			}
			break;

			case IDLING:
			{
				if ( ( GetTimeInState() > 0.0f ) && !xServer.IsMovingFast() )
				{
					CheckForAutomaticReload();
				}
			}
			break;

		default:
			{
				// Nothing to do...
			}
			break;
		}
	}
}

bool AVP_InventoryWeapon_SmartGun::IsPlayerIsStillFiring( const AVP_ClientEntity_Actor& xActor, int iSubType ) const
{
	if(xActor.GetMovementSystem() && xActor.GetMovementSystem()->IsMoveInProgress())
	{
		return false;
	}

	const AVP_Actor_InputHandler* pxInputHandler = xActor.GetInputHandler();
	if ( pxInputHandler != NULL && ( GetTimeInState() > 0.0f ) )
	{
		switch ( iSubType )
		{
			case ST_PRIMARY:
			{
				return (pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::PRIMARY_ATTACK_HELD ) > 0.0f && pxInputHandler->GetControlMapValue( AVP_Vertebrate_InputHandler::FAST ) == 0.0f);
			}

			default:
			{

			}
		}
	}
	//else we're either not a player or we've not been in the state for long enough.
	return true;
}

void AVP_InventoryWeapon_SmartGun::InitialiseAnimation()
{
	Asura_Animation* pxAnimation = GetAnimation( false );
	if ( pxAnimation && pxAnimation->GetParent() )
	{
		Asura_Assert( m_xIKSolver.GetFirstChain() == NULL, "IK Solver already initialised, clearing it..." );
		if ( m_xIKSolver.GetFirstChain() )
		{
			m_xIKSolver.Clear();
		}

		m_xArmIKChain.SetBonesArrays( s_aiArmIKBone, NULL );
		m_xArmIKChain.SetMarkerPointHash( s_uBellyBoxMarkerHash );
		m_xArmIKChain.SetAnim( pxAnimation->GetParent() );
		m_xArmIKChain.SetUseOri( true );
		m_xArmIKChain.SetPosSmoothing( false );
		m_xArmIKChain.SetOriSmoothing( false );
		m_xArmIKChain.SetExtendLimit( 0.0f );
		m_xArmIKChain.SetAxesFromRestPose( pxAnimation->GetSkin()->BoneRestPose );
		m_xIKSolver.AddChain( m_xArmIKChain );

		m_xIKSolver.Enable();
		pxAnimation->SetIKSolver( &m_xIKSolver );
	}
}

void AVP_InventoryWeapon_SmartGun::DestroyAnimation()
{
	Asura_Animation* pxAnimation = GetAnimation( false );
	if ( pxAnimation )
	{
		pxAnimation->SetIKSolver( NULL );
		m_xIKSolver.Disable();
		m_xIKSolver.Shutdown();
	}

	if(m_pxClientEffectsHandler)
	{
		m_pxClientEffectsHandler->StopPermenantWeaponAnim();
	}

	PARENT::DestroyAnimation();
}

void AVP_InventoryWeapon_SmartGun::CalculateFiringDataForSubType( const AVP_ClientEntity_Actor& xActor, Asura_Hash_ID uMuzMP, float fMuzVel, float fMaxRange, Asura_Vector_3& xFirePosition, Asura_Vector_3& xFireDirection, Asura_Matrix_3x3& xMuzzleOri, float& fRange, Asura_Guid* puTargetGuid /*= NULL*/ )
{
	if(!xActor.GetMarkerLocationFromHash( uMuzMP, &xFirePosition, &xMuzzleOri ))
	{
		xFirePosition	= xActor.GetCurrentPosition();
		xMuzzleOri		= xActor.GetCurrentOrientation();
	}

	if ( xActor.IsPlayer() )
	{
		xFireDirection = m_xMinigunDirLastFrame; 
		fRange = m_fRange;
		xFirePosition = m_xFirePosition;
		xMuzzleOri = m_xMuzzleOri;
	}
	else // if AI
	{
		Asura_AVector3 xTemp;
		xActor.GetAimDirection( xTemp );
		xFireDirection = Asura_Vector_3( xTemp );
		fRange = 1000.0f; //xFireDirection.Normalise();
	}
}

void AVP_InventoryWeapon_SmartGun::UpdateSpin(State eState, u_int uUsageType)
{
	bool bSpinUp = false;

	if(eState == USING && ( uUsageType == UT_FULL_AUTO_PRIMARY || uUsageType == UT_N_BULLET_BURST_PRIMARY ) )
	{
		if(m_fSpinSpeed <= (m_fSpinMaxSpeed*m_fSpinInputScale))
		{
			bSpinUp = true;
		}
	}

	if(bSpinUp)
	{
		m_fSpinSpeed += m_fSpinAcceleration * Asura_Timers::GetGameTimeSlice();
	}
	else
	{
		m_fSpinSpeed -= m_fSpinDeceleration * Asura_Timers::GetGameTimeSlice();
	}

	m_fSpinSpeed = Asura_Maths::ClampToRange(m_fSpinSpeed, 0.0f, m_fSpinMaxSpeed);

	m_fSpinAngle += m_fSpinSpeed * Asura_Timers::GetGameTimeSlice();

	while(m_fSpinAngle > 1.0f)
	{
		m_fSpinAngle -= 1.0f; // Loop from 0.0f to 1.0f
	}

	if(m_pxClientEffectsHandler)
	{
		if(!m_pxClientEffectsHandler->IsPermenantWeaponAnimPlaying())
		{
			m_pxClientEffectsHandler->PlayPermenantWeaponAnim();
		}

		m_pxClientEffectsHandler->SetWeaponPermenantAnimParametricTimer(m_fSpinAngle);
	}

	const Asura_Sound_Event* pxSpinLoop = Asura_Sound_Event_Storage::FindEvent(m_uSpinLoopSoundHash);
	if(pxSpinLoop)
	{
		if(m_fSpinSpeed > 0.0f)
		{
			if (m_iSpinLoopSoundHandle != ASURA_INVALID_HANDLE && !Asura_Sound_Event_System::IsPlaying(m_iSpinLoopSoundHandle))
			{
				Asura_Sound_Event_System::Stop( m_iSpinLoopSoundHandle );
			}
			if(m_iSpinLoopSoundHandle == ASURA_INVALID_HANDLE)
			{
				Asura_Sound_Event_System::PlayEvent(pxSpinLoop, &m_iSpinLoopSoundHandle, NULL, 0.0f); // start silent
			}
			Asura_Sound_Event_System::SetPitchScale(m_iSpinLoopSoundHandle, m_fSpinSpeed);
			Asura_Sound_Event_System::SetVolumeScale(m_iSpinLoopSoundHandle, m_fSpinSpeed);
		}
		else
		{
			if(m_iSpinLoopSoundHandle != ASURA_INVALID_HANDLE)
			{
				Asura_Sound_Event_System::Stop(m_iSpinLoopSoundHandle);
			}
		}
	}
}

void AVP_InventoryWeapon_SmartGun::UpdateTargetLock( const AVP_ClientEntity_Actor& xActor )
{
	const AVP_RangedWeapon_SubType*	pxSubType	= static_cast< AVP_RangedWeapon_SubType* >( m_ppxSubTypes[ 0 ] );

	if(!pxSubType) return;

	const AVP_Actor_AnimationHandler* pxAnimHandler = xActor.GetAnimationHandler();
	const Asura_Hash_ID	uMuzMP		= (pxAnimHandler && pxAnimHandler->IsFirstPerson()) ? pxSubType->GetMuzzleMarkerPointFirstPerson() : pxSubType->GetMuzzleMarkerPointThirdPerson();
	const float			fMaxRange	= pxSubType->GetMaxRange();

	Asura_Vector_3 xFireDirection(0.f,0.f,0.f);

	const AVP_Vertebrate_Focus* pxFocus = xActor.GetFocus();
	if ( pxFocus )
	{
		if(pxFocus->HasValidLockOnTarget())
		{
			Asura_Vector_3			xTargetVelocity;
			Asura_Vector_3			xTargetPosition( Asura_Maths::ZeroVector3 );
			AVP_ClientEntity_Actor*	pxTarget	= AVP_ClientEntity_Actor::FindEntity( pxFocus->GetLockOnTarget() );
			Asura_Assert( pxTarget != NULL, "Unable to get tagged actor" );
			if ( pxTarget )
			{
				pxTarget->GetCurrentVelocity(xTargetVelocity);
				if ( !pxTarget->GetMarkerLocationFromHash( 0xe5a85d03, &xTargetPosition, NULL ) ) // target_1
				{
					xTargetPosition = pxTarget->GetCurrentPosition();
					Asura_Assert(false, "Failed to find target markerpoint");
				}
			}

			CalculateFiringDataForMovingTarget( xActor, uMuzMP, 1.0f, xTargetPosition, Asura_Maths::ZeroVector3, m_xFirePosition, m_xMuzzleOri, xFireDirection);
			xFireDirection.Normalise();
			m_fRange = ( xTargetPosition - m_xFirePosition ).Magnitude();

			m_xMinigunDirLastFrame.Lerp(xFireDirection, 6.0f*Asura_Timers::GetGameTimeSlice()); // Lerp old dir towards desired dir
			// TODO - Set players aim direction here ( shouldn't affect look direction )
		}
		else
		{
			Asura_Matrix_3x3	xFireOrientation;
			m_fRange = CalculateFiringDataForCamera( xActor, uMuzMP, fMaxRange, m_xFirePosition, xFireOrientation );
			m_xMinigunDirLastFrame = Asura_Vector_3( xFireOrientation.e13, xFireOrientation.e23, xFireOrientation.e33 );
		}
	}
	else
	{
		Asura_Matrix_3x3	xFireOrientation;
		m_fRange = CalculateFiringDataForCamera( xActor, uMuzMP, fMaxRange, m_xFirePosition, xFireOrientation );
		m_xMinigunDirLastFrame = Asura_Vector_3( xFireOrientation.e13, xFireOrientation.e23, xFireOrientation.e33 );
	}
}

// void AVP_InventoryWeapon_SmartGun::RenderTargetLock( void )
// {
// 	Asura_GUIMenu_Root* pxItem = AVP_GUIMenu_System::GetHUDElement( AVP_GUIMenu_System::SMARTGUN_RETICLE, AVP_GUIMenu_System::GetMenuHash( AVP_GUIMenu_System::GUIMENU_ID_HUD_ELEMENTS ), false );
// 	if( pxItem )
// 	{
// 		Asura_Vector_3 xHitLocation = m_xFirePosition + (m_xMinigunDirLastFrame*m_fRange);
// 		Asura_Camera::ProjectWorldCoordsToScreenSpace( xHitLocation, true ); 
// 
// 		Asura_Bounding_Rectangle xNewBoundingBox( pxItem->GetScaledWorldBounds() ); 
// 		Asura_Vector_2 xInstancePosition( xHitLocation.GetX(), xHitLocation.GetY() ); 
// 		Asura_Vector_2 xIconOrigin( pxItem->GetPosOrigin() );
// 		Asura_Vector_2 xIconSize( xNewBoundingBox.GetDimension() ); 
// 		xIconOrigin.x *= xIconSize.x;
// 		xIconOrigin.y *= xIconSize.y;
// 		Asura_Vector_2 xInstanceDisplacement = xInstancePosition - ( pxItem->GetScaledWorldPos() + xIconOrigin );
// 		xNewBoundingBox.Displace( xInstanceDisplacement );
// 		pxItem->SetScaledWorldBounds( xNewBoundingBox );
// 		pxItem->Render( ASURA_GUIMENU_ACTION_FLAG_ALL );
// 	}
// }

void AVP_InventoryWeapon_SmartGun::GetScreenCoordsForReticle(float& fX, float& fY) const
{
	Asura_Vector_3 xHitLocation = m_xFirePosition + (m_xMinigunDirLastFrame*m_fRange);
	Asura_Camera::ProjectWorldCoordsToScreenSpace( xHitLocation, true ); 

	fX = xHitLocation.GetX();
	fY = xHitLocation.GetY(); 
}

/// End of file. ///
