//////////////////////////////////////////////////////////////////////////////
//
// AVP_Leaderboards.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     platform-independent leaderboards code
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_LEADERBOARDS

// This file's header.
#include "Network/AVP_Leaderboards.h"

// Asura includes
#include "Debugging/Asura_Assert.h"
#include "Network/Online/Asura_Leaderboards.h"

// AVP includes
#include "Entity/PlayerInfo/AVP_ClientEntity_PlayerInfo.h"


// G L O B A L   D A T A /////////////////////////////////////////////////////
static const u_int g_auLeaderboardNumStats[AVP_NUM_LEADERBOARDS] = 
{
	AVP_NUM_KILLS_STATS,
	AVP_NUM_SCORE_STATS,
	AVP_NUM_MATCHES_STATS,
	AVP_NUM_MYSTATS_STATS,
	AVP_NUM_INTERNAL_STATS,
};

static const u_int g_axKillsLBStatIDs[AVP_NUM_KILLS_STATS] = 
{
	AVP_STATID_TOTAL_KILLS,			// AVP_TOTAL_KILLS,
	AVP_STATID_KILLS_AS_MARINE,		// AVP_KILLS_AS_MARINE,
	AVP_STATID_KILLS_AS_ALIEN,		// AVP_KILLS_AS_ALIEN,
	AVP_STATID_KILLS_AS_PREDATOR,	// AVP_KILLS_AS_PREDATOR,
	AVP_STATID_KILLS_TO_DEATH_RATIO,// AVP_KILLS_TO_DEATH_RATIO,
	AVP_STATID_KILLS_KILL_STREAK,	// AVP_KILLS_KILL_STREAK,
	AVP_STATID_KILLS_TOTAL_DEATHS,	// AVP_KILLS_TOTAL_DEATHS,
};

static const u_int g_axScoreLBStatIDs[AVP_NUM_SCORE_STATS] = 
{
	AVP_STATID_SCORE_TOTAL,			//AVP_SCORE_TOTAL,
	AVP_STATID_SCORE_MARINE,		//AVP_SCORE_MARINE,
	AVP_STATID_SCORE_ALIEN,			//AVP_SCORE_ALIEN,
	AVP_STATID_SCORE_PREDATOR,		//AVP_SCORE_PREDATOR,
	AVP_STATID_SCORE_TIME_PLAYED,	//AVP_SCORE_TIME_PLAYED,
};

static const u_int g_axMatchesLBStatIDs[AVP_NUM_MATCHES_STATS] = 
{
	AVP_STATID_MATCHES_TOTAL,			//AVP_MATCHES_TOTAL,
	AVP_STATID_MATCHES_AS_MARINE,		//AVP_MATCHES_AS_MARINE,
	AVP_STATID_MATCHES_AS_ALIEN,		//AVP_MATCHES_AS_ALIEN,
	AVP_STATID_MATCHES_AS_PREDATOR,		//AVP_MATCHES_AS_PREDATOR,
	AVP_STATID_MATCHES_WON,				//AVP_MATCHES_WON,
	AVP_STATID_MATCHES_LOST,			//AVP_MATCHES_LOST,
	AVP_STATID_WINNING_STREAK,			//AVP_MATCHES_WINNING_STREAK,
};

static const u_int g_axMyStatsLBStatIDs[AVP_NUM_MYSTATS_STATS] = 
{
	AVP_STATID_MYSTATS_TOTAL_RANK,		//AVP_MYSTATS_TOTAL_RANK,
	AVP_STATID_MYSTATS_RANK_MARINE,		//AVP_MYSTATS_RANK_MARINE,
	AVP_STATID_MYSTATS_RANK_ALIEN,		//AVP_MYSTATS_RANK_ALIEN,
	AVP_STATID_MYSTATS_RANK_PREDATOR,	//AVP_MYSTATS_RANK_PREDATOR,
	AVP_STATID_MYSTATS_HEADSHOTS,		//AVP_MYSTATS_HEADSHOTS,
	AVP_STATID_MYSTATS_HEADBITES,		//AVP_MYSTATS_HEADBITES,
	AVP_STATID_MYSTATS_TROPHIES,		//AVP_MYSTATS_TROPHIES,
};

static const u_int g_axInternalLBStatIDs[AVP_NUM_INTERNAL_STATS] = 
{
	AVP_STATID_QUIT_EARLY_COUNT,		//AVP_INTERNAL_QUIT_EARLY_COUNT,
};

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void AVP_Leaderboards::AddConsoleCmdsAndVars()
{
	// add meaningful names here?
}


// GetAssociatedStatID() returns the AVP_STAT_IDENTIFIER for a given leaderboard
// and column (where 0 is the ranked stat column). 
u_int AVP_Leaderboards::GetAssociatedStatID( u_int uLeaderboard, u_int uColumn )
{
	if (uLeaderboard >= AVP_NUM_LEADERBOARDS || 
		uColumn >= g_auLeaderboardNumStats[uLeaderboard])
	{
		Asura_Assert(false, "invalid leaderboard or column");
		return AVP_STATID_INVALID;
	}

	switch (uLeaderboard)
	{
		case AVP_LEADERBOARD_KILLS:		return g_axKillsLBStatIDs[uColumn];
		case AVP_LEADERBOARD_SCORE:		return g_axScoreLBStatIDs[uColumn];
		case AVP_LEADERBOARD_MATCHES:	return g_axMatchesLBStatIDs[uColumn];
		case AVP_LEADERBOARD_MYSTATS:	return g_axMyStatsLBStatIDs[uColumn];
		case AVP_LEADERBOARD_INTERNAL:	return g_axInternalLBStatIDs[uColumn];
	}

	Asura_Assert(false, "unexpected leaderboard/column");
	return AVP_STATID_INVALID;
}

// GetStatResultById() returns results for the match which has just ended. It's up to other systems
// to keep track of stats which span multiple matches and to add up the stats for this match with 
// totals from all previous matches.
u_int AVP_Leaderboards::GetStatResultById(u_int uStatID, const AVP_Multiplayer_PlayerScore& xScore)
{
	switch (uStatID)
	{
	case AVP_STATID_TOTAL_KILLS:				return xScore.GetNumberOfKills();
	case AVP_STATID_KILLS_AS_MARINE:			return xScore.GetNumberOfKillsAsMarine();
	case AVP_STATID_KILLS_AS_ALIEN:				return xScore.GetNumberOfKillsAsAlien();
	case AVP_STATID_KILLS_AS_PREDATOR:			return xScore.GetNumberOfKillsAsPredator();
	case AVP_STATID_KILLS_KILL_STREAK:			return xScore.GetKillStreak();
	case AVP_STATID_KILLS_TOTAL_DEATHS:			return xScore.GetNumberOfTimesKilled();
	case AVP_STATID_SCORE_TOTAL:				return xScore.GetIndividualScore();
	case AVP_STATID_SCORE_MARINE:				return 0; // TODO
	case AVP_STATID_SCORE_ALIEN:				return 0; // TODO
	case AVP_STATID_SCORE_PREDATOR:				return 0; // TODO
	case AVP_STATID_SCORE_TIME_PLAYED:			return 0; // TODO
	case AVP_STATID_MATCHES_TOTAL:				return 1;
	case AVP_STATID_MATCHES_AS_MARINE:			return 0; // TODO
	case AVP_STATID_MATCHES_AS_ALIEN:			return 0; // TODO
	case AVP_STATID_MATCHES_AS_PREDATOR:		return 0; // TODO
	case AVP_STATID_MATCHES_WON:				return 0; // TODO
	case AVP_STATID_MATCHES_LOST:				return 0; // TODO
	case AVP_STATID_MYSTATS_TOTAL_RANK:			return 0; // TODO
	case AVP_STATID_MYSTATS_RANK_MARINE:		return 0; // TODO
	case AVP_STATID_MYSTATS_RANK_ALIEN:			return 0; // TODO
	case AVP_STATID_MYSTATS_RANK_PREDATOR:		return 0; // TODO
	case AVP_STATID_MYSTATS_HEADSHOTS:			return xScore.GetNumberOfHeadshots();
	case AVP_STATID_MYSTATS_HEADBITES:			return xScore.GetNumberOfHeadbites();
	case AVP_STATID_MYSTATS_TROPHIES:			return xScore.GetNumberOfTrophies();
	default:
		break;
	}

	Asura_Assert(false, "stat ID not found");
	return 0;
}

bool AVP_Leaderboards::IsAdditiveStat( u_int uStatID )
{
	return uStatID > AVP_STATID_INVALID && uStatID < AVP_NUM_ADDITIVE_STATS;
}


// Asura_Leaderboard_System project functions
u_int Asura_Leaderboard_System::Project_GetNumLeaderboards()
{
	return AVP_NUM_LEADERBOARDS;
}

u_int Asura_Leaderboard_System::Project_GetNumLeaderboardColumns( u_int uLeaderboard )
{
	Asura_Assert(uLeaderboard < AVP_NUM_LEADERBOARDS, "invalid leaderboard");
	return g_auLeaderboardNumStats[uLeaderboard];
}

u_int Asura_Leaderboard_System::Project_GetRowsPerPage()
{
	// TODO: this needs to be bigger e.g. 25 but keep it low to test page up/down code.
	return 3;
}

ASURA_LB_DATAFORMAT Asura_Leaderboard_System::Project_GetDataFormat(u_int uLeaderboard, u_int uColumn)
{
	// only the kills leaderboard has a float
	if (uLeaderboard == AVP_LEADERBOARD_KILLS && uColumn == AVP_KILLS_TO_DEATH_RATIO)
	{
		return ASURA_LB_DATAFORMAT_FLOAT;
	}
	return ASURA_LB_DATAFORMAT_UINT;
}

// Project_GetDataFormatString should return a printf-style specifier for column e.g. "%.02f", "%.f"
const char* Asura_Leaderboard_System::Project_GetDataFormatString(u_int uLeaderboard, u_int uColumn)
{
	if (Project_GetDataFormat(uLeaderboard, uColumn) == ASURA_LB_DATAFORMAT_FLOAT)
	{
		return "%.02f";
	}
	return "%.f";
}

#endif
