
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_CalculateSurfaceReorientationNormal.cpp
// Author : Andy Tate
// Date   : 5-Feb-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Physics/AVP_CalculateSurfaceReorientationNormal.h"

// Asura independent header-files...
#if !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Debugging/Asura_DebugLineRender.h"
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Physics/Asura_Physics_CCD.h"
#include "Physics/Asura_Physics_Object.h"

// Game independent header-files...
#include "Core/AVP_Plane.h"
#include "Physics/AVP_Physics_OBBTests.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

// Debug variables...
DEBUGVAR float	dbg_fFloorCosLimit		= 0.707f;	// cosine(45 degrees)

#if !defined(ASURA_NO_DEBUG_OUTPUT)
static bool	dbg_bDebugRenderTriangles		= false;	//true;	//
static bool	dbg_bDebugRenderSurfaceNormals	= false;	//true;	//
static bool	dbg_bDebugRenderContactPoints	= false;	//true;	//
static bool	dbg_bDebugRenderVelocities		= false;	//true;	//

// Debug constants...
static const u_int	g_uMaxDebugRenderColours					= 8;
static const u_int	g_uMaxDebugRenderContactPointsPerTriangle	= 12;

static const u_int g_auDebugRenderColours[g_uMaxDebugRenderColours] =
{
	0x80FFFFFF,
	0x80FFFF00,
	0x80FF00FF,
	0x8000FFFF,
	0x80FF0000,
	0x8000FF00,
	0x800000FF,
	0x80000000,
};

#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#if !defined(ASURA_NO_DEBUG_OUTPUT)
static u_int	g_uCurrentDebugColour	= 0;
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::CalculateTriangleConvexValue
// Author       : Andy Tate
// Date         : 13-Feb-2008
////////////////////////////////////////////////////////////////////////////////

float AVP_CalculateSurfaceReorientationNormal::CalculateTriangleConvexValue(const Asura_Physics_Tri& xTriangle,
																			const Asura_OBB& xPhysicsOBB,
																			Asura_AVector3_InParam xTriangleNormal) const
{
	const AVP_Plane			xPlane(xTriangle.GetCentre(), xTriangleNormal);

	const Asura_AVector3&	xCentre				= xPhysicsOBB.GetCentre();
	const Asura_AMatrix3x3&	xOrientation		= xPhysicsOBB.GetOrientation();
	const float				fClosestDistance	= xPlane.CalculateClosestDistance(xCentre);

	Asura_AVector3			xHalfExtents(xPhysicsOBB.GetHalfExtents());
	Asura_AVector3			xFaceNormalInOBBSpace(xOrientation.ApplyTranspose(xTriangleNormal));

	float					fMaxRotatedDistance;
	float					fConvexValue;

	// Calculate the maximum length of the BB in the direction
	//  of the normal (ignoring the upwards component)...
	xFaceNormalInOBBSpace.Max(xFaceNormalInOBBSpace, -xFaceNormalInOBBSpace);
	xFaceNormalInOBBSpace.SetY(0.0f);
	if ( xFaceNormalInOBBSpace.ApproximatelyEquals( Asura_Maths::ZeroAVector3 ) )
	{
		fConvexValue = Asura_Maths::LargeFloat;
	}
	else
	{
		xFaceNormalInOBBSpace.Normalise();

		xHalfExtents.SetY(0.0f);

		// NOTE: only allow rotation if we are more than one-tenth of the way inside the plane,
		// this stops the BB from tipping into the surface and losing contacts on the second face.
		fMaxRotatedDistance	= xHalfExtents * xFaceNormalInOBBSpace;
		fConvexValue		= fClosestDistance / fMaxRotatedDistance;
	}

	return(fConvexValue);
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal::CalculateTriangleConvexValue'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::IsWithinCosLimits
// Author       : Andy Tate
// Date         : 12-Feb-2008
////////////////////////////////////////////////////////////////////////////////

bool AVP_CalculateSurfaceReorientationNormal::IsWithinCosLimits(Asura_AVector3_InParam xNormal, Asura_AVector3_InParam xInputVelocity) const
{
	Asura_AVector3	xUpVector;
	float			fUpDotProduct;

	// Compare our up-vector with the normal...
	GetUpVector(xUpVector);

	fUpDotProduct	= xUpVector * xNormal;

	if(fUpDotProduct < m_fCosLimit)
	{
		// This normal is too far away to be considered...
		return(false);
	}
	else
	{
		// Do we need to check our directional limits too?
		if(m_bUseDirectionCosLimits)
		{
			// We can only do this if there is some difference between our normal and the up-vector...
			float fVelMag = xInputVelocity.Magnitude();
			if(fUpDotProduct >= dbg_fFloorCosLimit || fVelMag < 0.01f )
			{
				return(true);
			}
			else
			{
				Asura_AVector3	xFwdVector;
				Asura_AVector3	xNormalAlongUpVector;
				Asura_AVector3	xNormalPerpendicularToUpVector;

				// Calculate the normal perpendicular to our up-vector...
				xNormalAlongUpVector			= xUpVector * fUpDotProduct;
				xNormalPerpendicularToUpVector	= xNormal - xNormalAlongUpVector;

				if ( !xNormalPerpendicularToUpVector.ApproximatelyEquals( Asura_Maths::ZeroAVector3 ) )
				{
					xNormalPerpendicularToUpVector.Normalise();
				}

				GetForwardsVector(xFwdVector);
				
				float fNormVelDotProduct = ( xNormalPerpendicularToUpVector * xInputVelocity ) / fVelMag;
				if ( fNormVelDotProduct >= 0.0f )
				{
					// if we're moving away from the normal then it's OK to accept it, we only want to reject ones we're moving towards at glancing angles
					return true;
				}
				else
				{
					float fFwdVelDotProcuct = Asura_Maths::ClampToRange( ( xFwdVector * xInputVelocity ) / fVelMag, -1.0f, 1.0f );
					float fCosLimit = 0.0f;
					if ( fFwdVelDotProcuct >= 0.0f )
					{
						fCosLimit = Asura_Maths::Lerp( m_fSidewaysCosLimit, m_fForwardsCosLimit, fFwdVelDotProcuct );
					}
					else
					{
						fCosLimit = Asura_Maths::Lerp( m_fSidewaysCosLimit, m_fBackwardsCosLimit, -fFwdVelDotProcuct );
					}

					return ( fNormVelDotProduct <= fCosLimit );
				}
			}
		}
		else
		{
			// We're not checking any other limits - therefore we are successful...
			return(true);
		}
	}
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal::IsWithinCosLimits'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::CalculateTriangleNormal
// Author       : Andy Tate
// Date         : 9-Aug-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_CalculateSurfaceReorientationNormal::CalculateTriangleNormal(const Asura_Physics_Tri& xTriangle,
																	  const Asura_OBB& xPhysicsOBB,
																	  const Asura_OBB& xQueryOBB,
																	  Asura_AVector3_InParam xInputVelocity,
																	  Asura_AVector3_OutParam xOutputNormal) const
{
	// Get the normal from this triangle...
	xTriangle.GetNormal(xOutputNormal);

	// Do we need to consider this triangle?
	if(Asura_Physics_CCD::DoesOBBIntersectTri(xQueryOBB, xTriangle))
	{
		bool	bValidTriangle	= IsWithinCosLimits(xOutputNormal, xInputVelocity);

		if(m_bUseDirectionCosLimits && !bValidTriangle)
		{
			const float	fConvexValue	= CalculateTriangleConvexValue(xTriangle, xPhysicsOBB, xOutputNormal);

			bValidTriangle	= (fConvexValue < 0.999f);
		}

		if(bValidTriangle)
		{
			const AVP_Plane			xPlane(xTriangle.GetCentre(), xOutputNormal);

			const Asura_AVector3&	xHalfExtents		= xQueryOBB.GetHalfExtents();
			const Asura_AVector3&	xCentre				= xQueryOBB.GetCentre();
			const Asura_AMatrix3x3&	xOrientation		= xQueryOBB.GetOrientation();
			const float				fClosestDistance	= xPlane.CalculateClosestDistance(xCentre);

			Asura_AVector3			xFaceNormalInOBBSpace(xOrientation.ApplyTranspose(xOutputNormal));

			float					fMaxRotatedDistance;
			float					fDistanceOverPlane;

			// Calculate the maximum length of the BB in the direction of the normal...
			xFaceNormalInOBBSpace.Max(xFaceNormalInOBBSpace, -xFaceNormalInOBBSpace);

			// NOTE: only allow rotation if we are more than one-tenth of the way inside the plane,
			// this stops the BB from tipping into the surface and losing contacts on the second face.
			fMaxRotatedDistance	= xHalfExtents * xFaceNormalInOBBSpace;
			fDistanceOverPlane	= fClosestDistance / fMaxRotatedDistance;

			if(fDistanceOverPlane < -0.500f)
			{
				return(false);
			}
			else
			{
#if !defined(ASURA_NO_DEBUG_OUTPUT)
				// Increase the colour currently being used...
				g_uCurrentDebugColour	= (g_uCurrentDebugColour + 1) % g_uMaxDebugRenderColours;

				// Decide whether to render any debug data...
				if(dbg_bDebugRenderTriangles)
				{
					const u_int				uColour		= g_auDebugRenderColours[g_uCurrentDebugColour];
					const Asura_AVector3*	pxVertices	= xTriangle.GetVertexArray();

					Asura_DebugLineRender::AddLine(pxVertices[0], pxVertices[1], uColour);
					Asura_DebugLineRender::AddLine(pxVertices[1], pxVertices[2], uColour);
					Asura_DebugLineRender::AddLine(pxVertices[2], pxVertices[0], uColour);
				}

				if(dbg_bDebugRenderSurfaceNormals)
				{
					const u_int		uColour		= g_auDebugRenderColours[g_uCurrentDebugColour];
					Asura_AVector3	xCentre(xTriangle.GetCentre());
					Asura_AVector3	xNormal;

					xTriangle.GetNormal(xNormal);

					Asura_DebugLineRender::AddArrow(Asura_Vector_3(xCentre), Asura_Vector_3(xCentre + xNormal), uColour);
				}

				if(dbg_bDebugRenderContactPoints)
				{
					Asura_AVector3	xPenetrationAxis;
					float			fPenetration;

					if(Asura_Physics_CCD::DoesOBBIntersectTri(xQueryOBB, xTriangle, fPenetration, xPenetrationAxis))
					{
						Asura_AVector3							axContactPoints[g_uMaxDebugRenderContactPointsPerTriangle];

						const u_int								uNumContactPoints	= AVP_Physics_OBBTests::FindContactPoints_OBBTri(xQueryOBB, xTriangle, xPenetrationAxis, axContactPoints);
						const u_int								uColour				= g_auDebugRenderColours[g_uCurrentDebugColour];

						for(u_int u = 0; u < uNumContactPoints; u++)
						{
							Asura_DebugLineRender::AddArrow(Asura_Vector_3(axContactPoints[u]), Asura_Vector_3(axContactPoints[u] + (xOutputNormal * 0.200f)), uColour, 0.250f, 0.050f);
						}
					}
				}
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)

				return(true);
			}
		}
		else
		{
			return(false);
		}
	}
	else
	{
		return(false);
	}
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal::CalculateTriangleNormal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::CalculateVelocityAlongTriangle
// Author       : Andy Tate
// Date         : 8-Oct-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_CalculateSurfaceReorientationNormal::CalculateVelocityAlongTriangle(const Asura_Physics_Tri& xTriangle,
																			 Asura_AVector3_InParam xInputVelocity,
																			 Asura_AVector3_OutParam xOutputVelocity) const
{
	if(xInputVelocity.ApproximatelyEquals(Asura_Maths::ZeroAVector3))
	{
		return(false);
	}
	else
	{
		Asura_AVector3	xUpVector;
		Asura_AVector3	xOutputNormal;
		Asura_Quat		xRotation;

		// Get the normal from this triangle...
		xTriangle.GetNormal(xOutputNormal);

		// Calculate the difference between the 2 normals and update the velocity here...
		GetUpVector(xUpVector);
		xRotation.RotationFromAToB(xUpVector, xOutputNormal);
		xOutputVelocity = xRotation.RotateVector(xInputVelocity);

#if !defined(ASURA_NO_DEBUG_OUTPUT)
		// Decide whether to render any debug data...
		if(dbg_bDebugRenderVelocities)
		{
			const u_int		uColour		= g_auDebugRenderColours[g_uCurrentDebugColour];
			Asura_AVector3	xCentre(xTriangle.GetCentre());

			Asura_DebugLineRender::AddArrow(Asura_Vector_3(xCentre), Asura_Vector_3(xCentre + (xOutputVelocity * 0.200f)), uColour);
		}
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)

		return(true);
	}
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal::CalculateVelocityAlongTriangle'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::AVP_CalculateSurfaceReorientationNormal
// Author       : Andy Tate
// Date         : 6-Feb-2008
////////////////////////////////////////////////////////////////////////////////

AVP_CalculateSurfaceReorientationNormal::AVP_CalculateSurfaceReorientationNormal(const float fCosLimit,
																				 const u_int uMaxNormals)
: m_fCosLimit(fCosLimit)
, m_xCurrentOrientation(Asura_Maths::IdentityAMatrix3x3)
, m_xCurrentPosition(Asura_Maths::ZeroAVector3)
, m_xCurrentNormal(0.0f, -1.0f, 0.0f)
, m_xCurrentVelocityAlongNormal(Asura_Maths::ZeroAVector3)
, m_xCurrentVelocityPerpendicularToNormal(Asura_Maths::ZeroAVector3)
, m_xDesiredNormal(0.0f, -1.0f, 0.0f)
, m_xDesiredVelocity(Asura_Maths::ZeroAVector3)
, m_uFlags(0)
{
	m_axNormals		= new Asura_AVector3 [uMaxNormals];
	m_uMaxNormals	= uMaxNormals;
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal::AVP_CalculateSurfaceReorientationNormal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::~AVP_CalculateSurfaceReorientationNormal
// Author       : Andy Tate
// Date         : 6-Feb-2008
////////////////////////////////////////////////////////////////////////////////

AVP_CalculateSurfaceReorientationNormal::~AVP_CalculateSurfaceReorientationNormal()
{
	delete[](m_axNormals);
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal::~AVP_CalculateSurfaceReorientationNormal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_CalculateSurfaceReorientationNormal::Update
// Author       : Andy Tate
// Date         : 5-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_CalculateSurfaceReorientationNormal::Update(Asura_Physics_Object& xPhysicsObject,
													 Asura_AVector3_InParam xInputVelocity,
													 Asura_AVector3_InParam xInputNormal)
{
	float	fSpeedAlongNormal;

	// Update the position/orientation data...
	m_xCurrentOrientation.FromMat3(xPhysicsObject.GetOrientation());
	m_xCurrentPosition	= xPhysicsObject.GetPosition();

	// Calculate the normal and the velocity data relative to it...
	GetUpVector(m_xCurrentNormal);

	fSpeedAlongNormal	= m_xCurrentNormal * xInputVelocity;

	if(fSpeedAlongNormal == 0.0f)
	{
		m_xCurrentVelocityAlongNormal.Zero();
		m_xCurrentVelocityPerpendicularToNormal	= xInputVelocity;
	}
	else
	{
		m_xCurrentVelocityAlongNormal			= m_xCurrentNormal * fSpeedAlongNormal;
		m_xCurrentVelocityPerpendicularToNormal	= xInputVelocity - m_xCurrentVelocityAlongNormal;
	}

	// Reset the desired data...
	m_xDesiredNormal.Zero();
	m_xDesiredVelocity.Zero();

	m_uNumNormalsUsed		= 0;
#if !defined(ASURA_NO_DEBUG_OUTPUT)
	g_uCurrentDebugColour	= 0;
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)
}

/// End of function 'AVP_CalculateSurfaceReorientationNormal_EnvCache::Update'. ///


/// End of file. ///
