
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_CalculateSurfaceReorientationNormal.h
// Author : Andy Tate
// Date   : 5-Feb-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_CALCULATESURFACEREORIENTATIONNORMAL_H__
#define __AVP_CALCULATESURFACEREORIENTATIONNORMAL_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Physics_Object;
class Asura_Physics_Tri;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_CalculateSurfaceReorientationNormal
// Author       : Andy Tate
// Date         : 5-Feb-2008
//
// Description  : Class used to calculate the normal to be use when applying
//				  surface-reorientation to a physics-object.
////////////////////////////////////////////////////////////////////////////////

class AVP_CalculateSurfaceReorientationNormal
{
public:
	// Public functions...
	virtual ~AVP_CalculateSurfaceReorientationNormal();

	// Virtual functions...
	virtual void	Update(Asura_Physics_Object& xPhysicsObject, Asura_AVector3_InParam xInputVelocity, Asura_AVector3_InParam xInputNormal);

	// Accessor functions...
	bool GetNormal(const u_int uIndex, Asura_AVector3_OutParam xNormal) const
	{
		if(uIndex < m_uNumNormalsUsed)
		{
			xNormal	= m_axNormals[uIndex];
			return(true);
		}
		else
		{
			return(false);
		}
	}

	u_int	GetMaxNormals(void) const									{ return(m_uMaxNormals); }
	u_int	GetNumNormalsUsed(void) const								{ return(m_uNumNormalsUsed); }

	void	GetDesiredNormal(Asura_AVector3_OutParam xNormal) const		{ xNormal = m_xDesiredNormal; }
	void	GetDesiredVelocity(Asura_AVector3_OutParam xVelocity) const	{ xVelocity = m_xDesiredVelocity; }

	void ResetDirectionCosLimits(void)
	{
		m_bUseDirectionCosLimits	= false;
	}

	void SetDirectionCosLimits(const float fForwards, const float fSideways, const float fBackwards)
	{
		m_fForwardsCosLimit			= fForwards;
		m_fSidewaysCosLimit			= fSideways;
		m_fBackwardsCosLimit		= fBackwards;
		m_bUseDirectionCosLimits	= true;
	}

	void SetWalkWalkingDisabled( bool bDisable ) { m_bWallWalkingDisabled = bDisable; }

	void SetCosLimit( const float fCos ) { m_fCosLimit = fCos; }

protected:
	// Protected variables...
	float				m_fCosLimit;								// Limit beyond which we reject polygons for surface re-orientation.
	float				m_fForwardsCosLimit;						// Limit beyond which we reject polygons for surface re-orientation.
	float				m_fSidewaysCosLimit;						// Limit beyond which we reject polygons for surface re-orientation.
	float				m_fBackwardsCosLimit;						// Limit beyond which we reject polygons for surface re-orientation.

	Asura_AMatrix3x3	m_xCurrentOrientation;						// The orientation we're currently using.
	Asura_AVector3		m_xCurrentPosition;							// The position we're currently using.
	Asura_AVector3		m_xCurrentNormal;							// The normal we're currently using.
	Asura_AVector3		m_xCurrentVelocityAlongNormal;				// The component of the velocity along the current normal.
	Asura_AVector3		m_xCurrentVelocityPerpendicularToNormal;	// The component of the velocity perpendicular to the current normal.

	Asura_AVector3		m_xDesiredNormal;							// The current surface-normal.
	Asura_AVector3		m_xDesiredVelocity;							// The velocity used when calculating the surface-normal.

	// Protected functions...
	AVP_CalculateSurfaceReorientationNormal(const float fCosLimit, const u_int uMaxNormals);

	// Useful internal calculation functions...
	float	CalculateTriangleConvexValue(const Asura_Physics_Tri& xTriangle, const Asura_OBB& xPhysicsOBB, Asura_AVector3_InParam xTriangleNormal) const;

	bool	IsWithinCosLimits(Asura_AVector3_InParam xNormal, Asura_AVector3_InParam xInputVelocity) const;
	bool	CalculateTriangleNormal(const Asura_Physics_Tri& xTriangle, const Asura_OBB& xPhysicsOBB, const Asura_OBB& xQueryOBB, Asura_AVector3_InParam xInputVelocity, Asura_AVector3_OutParam xOutputNormal) const;
	bool	CalculateVelocityAlongTriangle(const Asura_Physics_Tri& xTriangle, Asura_AVector3_InParam xInputVelocity, Asura_AVector3_OutParam xOutputVelocity) const;

	// Accessor functions...
	void	GetSideVector(Asura_AVector3_OutParam xSide) const			{ xSide = m_xCurrentOrientation.GetColumn0(); }
	void	GetUpVector(Asura_AVector3_OutParam xUp) const				{ xUp = -m_xCurrentOrientation.GetColumn1(); }
	void	GetForwardsVector(Asura_AVector3_OutParam xForwards) const	{ xForwards = m_xCurrentOrientation.GetColumn2(); }

	bool AddNormal(Asura_AVector3_InParam xNormal)
	{
		if(m_uNumNormalsUsed < m_uMaxNormals)
		{
			m_axNormals[m_uNumNormalsUsed++]	 = xNormal;
			m_xDesiredNormal					+= xNormal;
			return(true);
		}
		else
		{
			return(false);
		}
	}

private:
	// Private variables...
	Asura_AVector3*	m_axNormals;						// Array of normals already used.

	union
	{
		u_int		m_uFlags;							// Single access-point for the flags.

		struct
		{
			u_int	m_uMaxNormals				: 8,	// The maximum number of normals allowed.
					m_uNumNormalsUsed			: 8,	// The number of normals used.
					m_bUseDirectionCosLimits	: 1,	// Flag to say whether to use the direction limits.
					m_bWallWalkingDisabled      : 1,

												: 0;	
		};
	};

	// Private functions...
	AVP_CalculateSurfaceReorientationNormal(void);
	AVP_CalculateSurfaceReorientationNormal(const AVP_CalculateSurfaceReorientationNormal& xOther);

	AVP_CalculateSurfaceReorientationNormal&	operator=(const AVP_CalculateSurfaceReorientationNormal& xOther);

};   // End of class 'AVP_CalculateSurfaceReorientationNormal'.


#endif   // ifndef __AVP_CALCULATESURFACEREORIENTATIONNORMAL_H__


/// End of file. ///
