
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor.cpp
// Author : Andy Tate
// Date   : 7-Feb-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Physics/AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor.h"

// Asura independent header-files...
#include "Debugging/Asura_Logfile.h"

// Game independent header-files...
#include "Physics/AVP_CalculateSurfaceReorientationNormal.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float			dbg_fSmoothingDurationForMinAngle	= 0.200f;
DEBUGVAR float			dbg_fSmoothingDurationForMaxAngle	= 0.100f;
DEBUGVAR Asura_Angle	dbg_fSmoothingMinAngle				= Asura_Maths::Pi / 12;	// 15 degrees
DEBUGVAR Asura_Angle	dbg_fSmoothingMaxAngle				= Asura_Maths::Pi / 6;	// 30 degrees


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor(void)
: m_xVelocityInterceptor(Asura_Maths::ZeroAVector3)
{
	// Nothing more to initialise...
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Reset
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Reset(Asura_AVector3_InParam xNormal)
{
	// Reset the velocity interceptor...
	m_xVelocityInterceptor.Reset(xNormal);

	// Call the parent function...
	PARENT::Reset(xNormal);
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update
// Author       : Andy Tate
// Date         : 6-Mar-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update(const float fTimeStep,
																			Asura_AVector3_InParam xDesiredNormal,
																			Asura_AVector3_InOutParam xCurrentNormal)
{
	if( (dbg_fSmoothingDurationForMinAngle > 0.0f) &&
		(dbg_fSmoothingDurationForMaxAngle > 0.0f) )
	{
		// Are we effectively at our destination?
		if(xCurrentNormal.ApproximatelyEquals(xDesiredNormal))
		{
			xCurrentNormal	= xDesiredNormal;
		}
		else
		{
			const float			fDotProduct			= xDesiredNormal * xCurrentNormal;
			const Asura_Angle	fAngle				= Asura_Maths::ArcCosine(fDotProduct);
			const float			fScale				= (Asura_Maths::Abs(fAngle) - dbg_fSmoothingMinAngle) / (dbg_fSmoothingMaxAngle - dbg_fSmoothingMinAngle);
			const float			fRotationParametric	= Asura_Maths::ClampToRange(fScale, 0.0f, 1.0f);
			const float			fDuration			= Asura_Maths::Lerp(dbg_fSmoothingDurationForMinAngle, dbg_fSmoothingDurationForMaxAngle, fRotationParametric);

			Asura_AVector3		xTargetNormal;

			// Update the current state of the velocity-interceptor...
			m_xVelocityInterceptor.GetTarget(xTargetNormal);

			if(xTargetNormal == xDesiredNormal)
			{
				m_xVelocityInterceptor.AddTime(fTimeStep);
			}
			else
			{
				m_xVelocityInterceptor.SetTarget(xDesiredNormal, fDuration);
			}

			// Calculate the latest normal using the velocity-interceptor...
			m_xVelocityInterceptor.CalculatePosition(xCurrentNormal);
			xCurrentNormal.Normalise();
		}
	}
	else
	{
		xCurrentNormal	= xDesiredNormal;
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update(const float fTimeStep,
																			const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																			Asura_AVector3_InOutParam xCurrentNormal)
{
	Asura_AVector3	xDesiredNormal;

	// Fetch the desired normal to use...
	xNormalCalculation.GetDesiredNormal(xDesiredNormal);

	// Call the other version of this function...
	AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update(fTimeStep, xDesiredNormal, xCurrentNormal);
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_VelocityInterceptor::Update'. ///


/// End of file. ///
