//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// AVP_Platform_Profile_LoadAndSave.cpp
// Thomas Young / Callum Hope - 8,1,09
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// File functionality to deal with saving data to file on the 360 platform
// No actual profile functionality, just profile reading saving, loading, and writing of data used in profiles. 
// *** This file is the Main thread interface to the thread running the Asura_Xbox360_Profile_System ***
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(AVP_SUPPORT_PROFILE_LOADANDSAVE)

#include "Core/Asura_HashedLocalisedText.h"
#include "Core/Asura_Xbox360_Profile_System.h"
#include "GUIMenu/Asura_Xbox360_Profile_GUIMenu.h"
#include "Interface/Asura_CommandConsole.h"

#include "LoadAndSave/AVP_Profile_LoadAndSave.h"

// G L O B A L   V A R I A B L E S ///////////////////////////////////////////

// CAH: Yes global variables are bad, 
// but these are only conserned with the 360 side of load/saving, meaning that they are saving memory for other platform builds being here like this. 

static bool		g_bSaveRequestedProfile		= false;
static bool		g_bLoadRequestedProfile		= false;
static bool		g_bFlowRequestedProfile		= false;

// CAH TEMP here until we decide what the names will be and how we want to store multi-languages of them
static char*				g_szBaseFile_Name360		= "AliensVsPredator";
static Asura_UnicodeChar*	g_szBaseFile_DisplayName360 = L"Aliens Vs Predator";

// CAH TEMP until checkpoint data and save/load system is sorted out. 
static char*	g_szTEMP_CheckpointData	= 0;

// TY TODO: Replace these with prehashed values when they are finalised
static const Asura_Hash_ID	uTEXT_PAGE							= Asura_GetHashID( "Menu" );
static const Asura_Hash_ID	u360_CHECKPOINT_FILE_DISPLAY_TEXT	= Asura_GetHashID( "360_CHECKPOINT_FILE_DISPLAY_TEXT" );

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Xbox360_Profile_Pause
// Scope		: Local Public Helper
// Author       : Callum Hope
// Date         : 14/1/09
//
// Desc			: CALLBACK FUNCTION used exclusively by the 360 for event on an ingame pause request. 
////////////////////////////////////////////////////////////////////////////////

void Xbox360_Profile_Pause ()
{
	// TODO: if we are in the game, pause
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Xbox360_Profile_Unpause
// Scope		: Local Public Helper
// Author       : Callum Hope
// Date         : 14/1/09
//
// Desc			: CALLBACK FUNCTION used exclusively by the 360 for event on an ingame UNpause request. 
////////////////////////////////////////////////////////////////////////////////

void Xbox360_Profile_Unpause ()
{
	// TODO: if we are in the game, unpause
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Xbox360_Profile_ReturnToStart
// Scope		: Local Public Helper
// Author       : Callum Hope
// Date         : 14/1/09
//
// Desc			: CALLBACK FUNCTION used exclusively by the 360 for event on an ingame restart mission request. 
////////////////////////////////////////////////////////////////////////////////

void Xbox360_Profile_ReturnToStart ()
{
	// TODO:	if we are loading, unwind and return to start screen
	//			if we are in game, quit and return to start screen
	//			if we are in the menus, close and open the start screen
	// in SS2 this was flagged in the core along with the state machine, may need to
	// do something similar here, although the state machine is a lot more complicated
}

////////////////////////////////////////////////////////////////////////////////
// Function     : ConsoleCmd_SelectDevice
// Scope		: Local Public Helper
// Author       : Callum Hope
// Date         : 14/1/09
//
// Desc			: Function to set the Storage device on the 360 to be saved to. 
//				  Called from the menu system should the player not be logged into a profile on game start. 
////////////////////////////////////////////////////////////////////////////////

bool ConsoleCmd_SelectDevice( bool bHelp )
{
	if( bHelp )
	{
		return true;
	}

	if( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		Asura_Xbox360_Profile_System::SelectStorageDevice ( AVP_Profile_LoadAndSave::GetProfileSaveData(), AVP_Profile_LoadAndSave::GetMaxSaveSize() );
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_Initialise
// Scope		: static private
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Sets up variables to be used globally and initialises anything to do with Asura_Xbox360_Profile_System
//				  Also begins the Asura_Xbox360_Profile_System profile flow, which is not initialisaiotion but need to be done before the menus load. 
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Platform_Initialise()
{
	// find the size of the profile data and set the global holder for it to be used whenever a save or load occurs. 
	s_iSaveFile_MaxSize = sizeof( s_xProfileSaveData );

	// set the first save slot for the profile saves. 
	Platform_SetLoadSaveSlot( 0 );

	// Initialise the Asura save system for the 360
	Asura_Xbox360_Profile_System::Initialise();

	// Give the save system the name of the file to look for on disk and the text displayed to the user in menus for it.  
	Asura_Xbox360_Profile_System::SetBaseFilename( g_szBaseFile_Name360, g_szBaseFile_DisplayName360 );

	// Now begin the Profile flow, sets the updates going for the save system. 
	Platform_BeginProfileFlow();

	// Initialise the GUI subsystem concerned with the Load and save process. 
	Asura_Xbox360_Profile_GUIMenu::Initialise();

	// Register our callbacks with the profile guimenu system
	Asura_Xbox360_Profile_GUIMenu::RegisterCallback( ASURA_XBOX360_PROFILE_GUIMENU_CALLBACK_PAUSE			, Xbox360_Profile_Pause			);
	Asura_Xbox360_Profile_GUIMenu::RegisterCallback( ASURA_XBOX360_PROFILE_GUIMENU_CALLBACK_UNPAUSE			, Xbox360_Profile_Unpause		);
	Asura_Xbox360_Profile_GUIMenu::RegisterCallback( ASURA_XBOX360_PROFILE_GUIMENU_CALLBACK_RETURN_TO_START	, Xbox360_Profile_ReturnToStart	);

	// Add our console commands
	Asura_CommandConsole::AddCmd( L"Xbox360Profile.SelectDevice", ConsoleCmd_SelectDevice, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_Update
// Scope		: Static Private. 
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Per-frame update, because we are interacting with a asyncronous thread essentially this is ues to handle all of the requests
//				  For actions that could not be done until the other thread is idle again. 
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Platform_Update()
{
	Asura_Xbox360_Profile_System::Update();
	Asura_Xbox360_Profile_GUIMenu::Update();

	// If we have Just carried out a load and the Loadsystem is Idle (meaning the load is done) copy the data over to the Profile system to be used from now on. 
	if( ( s_eLoadingState == AVP_PROFILELOAD_STATE_JUSTLOADED ) && ( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE ) )
	{
		Platform_CopyDataToProfileFromLoadAndSave();
		//MKC - Applying the settigns we have just loaded so everyting is set to the right values
		s_pxProfile->ApplySettings();
		s_eLoadingState = AVP_PROFILELOAD_STATE_IDLE;
	}

	// this is essentially a job queue, but has been done with an 'if - else' clause to prioritise any actions by what they are, not what order they were requested. 
	if( g_bFlowRequestedProfile && ( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE ) )
	{
		// if a flow start was requested and the system is now free do it now. 
		Platform_BeginProfileFlow();
		g_bFlowRequestedProfile = false;
	}
	else if( g_bSaveRequestedProfile && ( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE ) )
	{
		// if we have requested a save but havent been able to do so, do it now if the system is no longer busy. 
		Platform_SaveProfileData();
		g_bSaveRequestedProfile = false;
	}
	else if( ( s_eLoadingState == AVP_PROFILELOAD_STATE_LOADREQUESTED ) && ( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE ) )
	{
		// if we have requested a load but could do it do it now if the system is no longer busy. 
		Platform_LoadProfileData();
		g_bLoadRequestedProfile = false;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_Shutdown
// Scope		: static private
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Just a standard shut down for the variout elements of Asura_Xbox360_Profile system. 
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Platform_Shutdown()
{
	Asura_Xbox360_Profile_System::Shutdown ();
	Asura_Xbox360_Profile_GUIMenu::Shutdown();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_BeginProfileFlow
// Scope		: Static Private 
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Function to do the initial load from file and start the Asura_Xbox360_Profile_System updates
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Platform_BeginProfileFlow()
{
	// get the current data in AVP_Profile and copy to the save data structure. at this stage it will be default data. 
	Platform_CopyDataToLoadAndSaveFromProfile();

	if( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		// if the load/save system if idle then begin the flow process (set it up and start update calls)
		Asura_Xbox360_Profile_System::BeginProfileFlow( &s_xProfileSaveData, s_iSaveFile_MaxSize );

		// set the just loaded tab so that when the Begin process is complete the update function will copy the data over to AVP_Profile
		s_eLoadingState = AVP_PROFILELOAD_STATE_JUSTLOADED;
		g_bFlowRequestedProfile = false;
	}
	else
	{
		// if the system is busy then raise the flag to get this done in the update when we can.
		g_bFlowRequestedProfile = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_SaveProfileData
// Scope		: Static Private
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Routein to actually (and finally) tell the Asura_Xbox360_Profile_System to save the profile data to disk. 
// Return		: bool, on the 360 this is always true, as its used by other platforms. (PC)
////////////////////////////////////////////////////////////////////////////////

bool AVP_Profile_LoadAndSave::Platform_SaveProfileData()
{ 
	// get the current data in AVP_Profile and copy to the save data structure. 
	Platform_CopyDataToLoadAndSaveFromProfile();
	
	if( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		// if the 360 save system is idle then initiate a save. 
		Asura_Xbox360_Profile_System::SaveProfileData ( &s_xProfileSaveData, s_iSaveFile_MaxSize );
		g_bSaveRequestedProfile = false;
	}
	else
	{
		// otherwise raise a flag to say we want to save and save it when possible in the update funciton.
		g_bSaveRequestedProfile = true;
	}

	// the return here does not mean much for the 360 so just return positive to avoid asserts. 
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_LoadProfileData
// Scope		: Static Private
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Routein to actually (and finally) tell the Asura_Xbox360_Profile_System to load the profile data on disk. 
//				  We dont have to worry about there not being a file on disk because on the call to Asura_Xbox360_Profile_System::BeginProfileFlow one was made if there wasnt one. 
// Return		: Return an error feedback state for the AVP_Prifile system to use. 
////////////////////////////////////////////////////////////////////////////////

AVP_Profile::E_PROFILE_FILE_STATES AVP_Profile_LoadAndSave::Platform_LoadProfileData( )
{
	if( Asura_Xbox360_Profile_System::GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		// if the 360 save system is idle then initiate a load. 
		Asura_Xbox360_Profile_System::LoadProfileData( &s_xProfileSaveData, s_iSaveFile_MaxSize );
		
		// if the save/load system is not in a state of error then set the just loaded flag and return positive feedback
		if( Asura_Xbox360_Profile_System::GetState() != ASURA_XBOX360_PROFILE_STATE_ERROR )
		{
			s_eLoadingState = AVP_PROFILELOAD_STATE_JUSTLOADED;
			return AVP_Profile::PROFILE_FILE_OK;
		}
		else
		{
			// set the loading state back to idle as we did not load and are not loading. 
			s_eLoadingState = AVP_PROFILELOAD_STATE_IDLE;
			
			// return the correct code depending on what went wrong with the load. 
			switch( Asura_Xbox360_Profile_System::GetErrorState() )
			{
				case ASURA_XBOX360_PROFILE_ERROR_LOGOUT:
				case ASURA_XBOX360_PROFILE_ERROR_WRONG_USER:
				case ASURA_XBOX360_PROFILE_ERROR_NO_DEVICE_SELECTED:
				case ASURA_XBOX360_PROFILE_ERROR_NO_PROFILE_SELECTED:
				{
					return AVP_Profile::PROFILE_FILE_NOT_FOUND_NOUSER;
				}
				case ASURA_XBOX360_PROFILE_ERROR_NONE:
				{
					return AVP_Profile::PROFILE_FILE_OK;
				}
				case ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE:
				case ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED:
				{
					return AVP_Profile::PROFILE_DEVICE_ERROR;
				}
				case ASURA_XBOX360_PROFILE_ERROR_CORRUPT:
				default:
				{
					return AVP_Profile::PROFILE_FILE_CORRUPT;
				}
			}
		}
	}
	else
	{
		// if the load/save system is busy then raise a flag to say we want to save and save it when possible in the update funciton.
		s_eLoadingState = AVP_PROFILELOAD_STATE_LOADREQUESTED;
		return AVP_Profile::PROFILE_FILE_OK;
	}
}

// comment to do on implementation
void AVP_Profile_LoadAndSave::Platform_SetLoadSaveSlot( int iSlot )
{
	s_iSlot = iSlot;
}

// comment to do on implementation
void AVP_Profile_LoadAndSave::Platform_SaveCheckpoint( )
{
	Asura_UnicodeChar szDisplayName[128];

	// Build the display filename, I have made something up here, but obviously you can make it whatever you want
	Asura_PrintfToBfr( szDisplayName, Asura_HashedLocalisedText::GetString( uTEXT_PAGE, u360_CHECKPOINT_FILE_DISPLAY_TEXT ), 128, true, Asura_Random::GetInt( 1, 10 ) );

	Asura_Xbox360_Profile_System::SaveCheckpoint( s_iSlot, g_szTEMP_CheckpointData, 1024, szDisplayName );
}

// comment to do on implementation
void AVP_Profile_LoadAndSave::Platform_LoadCheckpoint( )
{
	Asura_Mem::Set( g_szTEMP_CheckpointData, 0, 1024 );
	Asura_Xbox360_Profile_System::LoadCheckpoint( s_iSlot, g_szTEMP_CheckpointData, 1024 );
}

// comment to do on implementation
AVP_Profile_State AVP_Profile_LoadAndSave::Platform_GetState( )
{
	// Convert the 360 profile state to something the RW profile system can use
	Asura_Xbox360_Profile_State eState;

	eState = Asura_Xbox360_Profile_System::GetState();

	switch ( eState )
	{
		case ASURA_XBOX360_PROFILE_STATE_IDLE:				return AVP_PROFILE_STATE_IDLE;
		case ASURA_XBOX360_PROFILE_STATE_PROCESSING:		return AVP_PROFILE_STATE_BUSY;
		case ASURA_XBOX360_PROFILE_STATE_LOAD_PROFILE:		return AVP_PROFILE_STATE_LOADING_PROFILE;
		case ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE:		return AVP_PROFILE_STATE_LOADING_PROFILE;
		case ASURA_XBOX360_PROFILE_STATE_LOAD_CHECKPOINT:	return AVP_PROFILE_STATE_LOADING_CHECKPOINT;
		case ASURA_XBOX360_PROFILE_STATE_SAVE_CHECKPOINT:	return AVP_PROFILE_STATE_SAVING_CHECKPOINT;
		case ASURA_XBOX360_PROFILE_STATE_ERROR:				return AVP_PROFILE_STATE_BUSY;				// Profile GUI system will handle errors
		case ASURA_XBOX360_PROFILE_STATE_QUERY:				return AVP_PROFILE_STATE_BUSY;				// ... and queries
	}

	Asura_Assert( false, "Unknown profile state returned" );

	return AVP_PROFILE_STATE_IDLE;
}

// comment to do on implementation
AVP_Profile_Error AVP_Profile_LoadAndSave::Platform_GetLastError( )
{
	// The 360 profile GUI system will handle errors for us
	return AVP_PROFILE_ERROR_NONE;
}

// comment to do on implementation
const Asura_UnicodeChar* AVP_Profile_LoadAndSave::Platform_GetSlotDisplayText ( const int iSlot )
{
	return Asura_Xbox360_Profile_System::GetSlotDisplayText( iSlot );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_CopyDataToLoadAndSaveFromProfile
// Scope		: Static Private
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Function to stop repetition in other code. 
//				  Copies all of the data in AVP_Profile that needs to be saved into the data structure that is saved. 
//				  This can be done like this because of the Friend relationship AVP_Prifile has with AVP_Profile_LoadAndSave
//				  all functions that talk to friend class AVP_Profile are private in AVP_Prifile_SaveAndLoad
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Platform_CopyDataToLoadAndSaveFromProfile()
{
	// this copying if fine due to AVP_Profile_LoadAndSave's friend relationship with AVP_Profile. 
	s_xProfileSaveData.m_xControls = s_pxProfile->m_xControls;
	s_xProfileSaveData.m_xGraphics = s_pxProfile->m_xGraphics;
	s_xProfileSaveData.m_xSounds = s_pxProfile->m_xSounds;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Platform_CopyDataToProfileFromLoadAndSave
// Scope		: Static Private
// Author       : Callum Hope
// Date         : 16/1/09
//
// Desc			: Function to stop repetition in other code. 
//				  Copies all of the data that has been loaded from file back to the AVP_Profile class where it can be used by the rest of the code. 
//				  This can be done like this because of the Friend relationship AVP_Prifile has with AVP_Profile_LoadAndSave
//				  all functions that talk to friend class AVP_Profile are private in AVP_Prifile_SaveAndLoad
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Platform_CopyDataToProfileFromLoadAndSave()
{
	// this copying if fine due to AVP_Profile_LoadAndSave's friend relationship with AVP_Profile. 
	s_pxProfile->m_xControls = s_xProfileSaveData.m_xControls;
	s_pxProfile->m_xGraphics = s_xProfileSaveData.m_xGraphics;
	s_pxProfile->m_xSounds = s_xProfileSaveData.m_xSounds;
}

#endif	//defined(AVP_SUPPORT_PROFILE_LOADANDSAVE)


/// End of file. ///
