/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "..\..\..\..\..\Asura\Xbox360\Shaders\Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float4x4 g_xWorldViewProjectionMatrix;
const float4x3 g_axBoneMatrices[80];

const float3 g_xBlurColour;
const float2 g_xBlurParams = {2.0f / 512.0f, 25.0f}; // x = blur kernel size, y = max distance


const texture g_xDepthTexture;
const texture g_xSceneTexture;

const sampler g_xSceneTextureSampler  = sampler_state
{
	Texture   = <g_xSceneTexture>;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};

const sampler g_xDepthTextureSampler  = sampler_state
{
	Texture   = <g_xDepthTexture>;
	AddressU  = CLAMP;
	AddressV  = CLAMP;
};



/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 m_xPosition     : POSITION;

	int4   m_aiBoneIndices : BLENDINDICES;
	float4 m_afBoneWeights : BLENDWEIGHT;

	float3 m_xNormal       : NORMAL;

};

struct OutputVertex_Deferred
{
	float4 m_xPosition      : POSITION;
	float4  m_fColour       : TEXCOORD0;
	float4  m_fColour2      : TEXCOORD1;
};


struct VertexFS
{
	float4 m_xPosition		: POSITION;
	float2 m_xUV			: TEXCOORD;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

float4x3 GetBoneMatrix_Constants( const float fIndex )
{
	return g_axBoneMatrices[ fIndex ];
}

float4x3 GetBoneMatrix_VertexStream( const float fIndex )
{
	float4 xBoneMat0, xBoneMat1, xBoneMat2;

	// Bones stored in semantics POSITION1-3. Ignore the last row, it's always (0 0 0 1)

	asm
	{
		vfetch xBoneMat0, fIndex, position1, UseTextureCache=true
		vfetch xBoneMat1, fIndex, position2, UseTextureCache=true
		vfetch xBoneMat2, fIndex, position3, UseTextureCache=true
	};

	// Stored as 3x4 in the vertex stream
	return transpose( float3x4( xBoneMat0, xBoneMat1, xBoneMat2 ) );
}

float EncodeDepth(float f)
{
	return 1.0f - saturate(f / g_xBlurParams.y);
}
float DecodeDepth(float f)
{
	return (1.0f - f) * g_xBlurParams.y;
}

OutputVertex_Deferred VShader( const InputVertex xIn, uniform bool bNewRenderer )
{
	// Position and UVs
	float4   xObjectSpacePosition = float4(0,0,0,1); 
	float4x3 axBoneMatrix;
	
	OutputVertex_Deferred xOut;
				
	// Decrementing loop counter doesn't seem to work, and
	// so when we finish we want the most heavily weighted bone - BoneWeights[0] - in 
	// axBoneMatrix, hence the '3-i' indexing
	for(int i=0; i<4; i++)
	{
		if( bNewRenderer )
		{
			axBoneMatrix = GetBoneMatrix_VertexStream( xIn.m_aiBoneIndices[3-i] );
		}
		else
		{
			axBoneMatrix = GetBoneMatrix_Constants( xIn.m_aiBoneIndices[3-i] );
		}

		// Recompute the matrix elements that we've stored per-bone-effects values in
		axBoneMatrix[2] = cross (axBoneMatrix[0], axBoneMatrix[1]);
		xObjectSpacePosition.xyz += mul( xIn.m_xPosition, axBoneMatrix) * xIn.m_afBoneWeights[3-i];
	}


	xOut.m_xPosition = mul( xObjectSpacePosition, g_xWorldViewProjectionMatrix );;
	float fDepth = EncodeDepth(xOut.m_xPosition.w);
//	float fDepth = EncodeDepth(max(mul( float4(0.0f, 0.0f, 0.0f, 1.0f ), g_xWorldViewProjectionMatrix ).w, 0.0f));
	xOut.m_fColour = float4(g_xBlurColour.rgb * pow(smoothstep(0.05f, 0.3f, fDepth ), 2.0f), fDepth);
	xOut.m_fColour2 = float4(0.0f, 0.0f, 0.0f, smoothstep(0.0f, 0.05f, fDepth));
	return xOut;
}
struct OutputPixel
{
	half4 m_xBufferColour    : COLOR0;
	half4 m_xBufferAlpha	: COLOR1;
};

OutputPixel PShader( const OutputVertex_Deferred xIn)
{
	OutputPixel xOut;
	xOut.m_xBufferColour = xIn.m_fColour;
	xOut.m_xBufferAlpha = xIn.m_fColour2;
	return xOut;
}

float4 VShader_FS( const VertexFS xIn ) : POSITION
{
	return xIn.m_xPosition;
}

float4 Avg4(sampler xSampler, const float2 xUV)
{
	float4 xSample[4];

	asm { 
		tfetch2D xSample[0], xUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.0, OffsetY = 0.0 
		tfetch2D xSample[1], xUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 1.0, OffsetY = 0.0
		tfetch2D xSample[2], xUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.0, OffsetY = 1.0
		tfetch2D xSample[3], xUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 1.0, OffsetY = 1.0
	};

	return (xSample[0] + xSample[1] + xSample[2] + xSample[3]) / 4;
}


float4 BlurLine(sampler xSampler, const float2 xUV, const float2 xOffset)
{
	float4 xInitialColour;
	float4 xSample;

	float2 xNewUV = xUV - xOffset * 5.0f;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	xInitialColour = xSample;

	xNewUV = xUV - xOffset * 3.0f;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	xInitialColour += xSample;

	xNewUV = xUV - xOffset;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	xInitialColour += xSample;

	xNewUV = xUV + xOffset;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	xInitialColour += xSample;

	xNewUV = xUV + xOffset * 3.0f;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	xInitialColour += xSample;

	xNewUV = xUV + xOffset * 5.0f;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	xInitialColour += xSample;

	return xInitialColour * 1.0f / 6.0f;
}

float BlurDepth(sampler xSampler, const float2 xUV, const float2 xOffset)
{
	float fDepth;

	float4 xInitialColour;
	float4 xSample;

	float2 xNewUV = xUV - xOffset * 3.0f;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	fDepth = xSample.a;

	xNewUV = xUV - xOffset;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	fDepth += xSample.a;

	xNewUV = xUV + xOffset;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	fDepth += xSample.a;

	xNewUV = xUV + xOffset * 3.0f;
	asm {
		tfetch2D xSample, xNewUV, xSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=linear, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5 
	};
	fDepth += xSample.a;

	return fDepth / 4.0f;
}




float4 PShader_BlurHoriz( const float2 xScreenPos : VPOS ) : COLOR
{
 	float fDepth = BlurDepth(g_xSceneTextureSampler, xScreenPos, float2(g_xBlurParams.x, 0.0f));
	float2 f2BlurSize = float2(g_xBlurParams.x, 0.0f);
 	f2BlurSize /= (0.25f + DecodeDepth(fDepth));
 	return float4(BlurLine(g_xSceneTextureSampler, xScreenPos,  f2BlurSize).rgb, fDepth);
}

float4 PShader_BlurVert( const float2 xScreenPos : VPOS ) : COLOR
{
	float fDepth = BlurDepth(g_xSceneTextureSampler, xScreenPos, float2(0.0f, g_xBlurParams.x));
	float2 f2BlurSize = float2(0.0f, g_xBlurParams.x);
 	f2BlurSize /= (0.25f + DecodeDepth(fDepth));
	float4 xInitialColour = BlurLine(g_xSceneTextureSampler, xScreenPos, f2BlurSize);


 	return float4(xInitialColour.rgb * 2.0f, 1.0f);
}

float4 PShader_ApplyToScreen( const float2 xScreenPos : VPOS ) : COLOR
{
 	float4 xRet = Avg4(g_xSceneTextureSampler, xScreenPos / 2);

	return float4(pow(smoothstep(0.01f, 1.0f, xRet), 1.0f).xyz * (1.0f - xRet.a), 1.0f);
}




/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

// Perform material response using the SH
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader( false );
		PixelShader = compile ps_3_0 PShader( );
	}
}

technique FS
{
	pass
	{
		VertexShader = compile vs_3_0 VShader_FS();
		PixelShader = compile ps_3_0 PShader_BlurHoriz();
	}
	pass
	{
		VertexShader = compile vs_3_0 VShader_FS();
		PixelShader = compile ps_3_0 PShader_BlurVert();
	}

	pass
	{
		ALPHABLENDENABLE = TRUE;
		SRCBLEND = ONE;
		DESTBLEND = ONE;
		SRCBLENDALPHA = ZERO;
		DESTBLENDALPHA = ZERO;
		BLENDOP = ADD;

		VertexShader = compile vs_3_0 VShader_FS();
		PixelShader = compile ps_3_0 PShader_ApplyToScreen();
	}

}

// This will go when the new skin renderer is fully switched on, we'll always pull the bones from
// the secondary vertex stream
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader( true );
		PixelShader = compile ps_3_0 PShader( );
	}
}