///////////////////////////////////////////////////////////////////////////////
//
// File: AVP_Platform_CloakEffect.cpp
//
// Created by Andrew Southgate 03/12/2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Handles rendering of the predators cloak
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Core/Asura_TextureManagement.h"
#include "SpecialFX/AVP_CloakEffect.h"
#include "SpecialFX/AVP_Platform_CloakEffect.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#include "Rendering/Asura_Viewport.h"
#include "Interface/Asura_CommandConsole.h"

// Asura 360
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_CharacterRenderer.h"
#include "Rendering/Asura_Xbox360_SkinRenderer.h"
#include "Debugging/Asura_Platform_GPUProfile.h"

// Shader Include Files
#include "CloakEffect.h"


namespace
{
	Asura_Xbox360_Effect*		g_pxCloakEffect				= 0;

	Asura_Xbox360_MaterialResponse_Handles				g_xMaterialsResponseHandles;

	Asura_Xbox360_Effect_System::Handle					g_xWorldMatrixHandle;
	Asura_Xbox360_Effect_System::Handle					g_xWorldViewMatrixHandle;
	Asura_Xbox360_Effect_System::Handle					g_xWorldViewProjectionMatrixHandle;
	Asura_Xbox360_Effect_System::Handle					g_xObjectSpaceCameraPositionHandle;
	Asura_Xbox360_Effect_System::Handle					g_xSceneTextureHandle;
	Asura_Xbox360_Effect_System::Handle					g_xNoiseTextureHandle;
	Asura_Xbox360_Effect_System::Handle					g_xOffsetHandle;
	Asura_Xbox360_Effect_System::Handle					g_xParamsHandle;
	Asura_Xbox360_Effect_System::Handle					g_xLightningColourHandle;
	Asura_Xbox360_Effect_System::Handle					g_xTransitionValueHandle;
	Asura_Xbox360_Effect_System::Handle					g_xBoneMatricesHandle;
	Asura_Xbox360_Effect_System::Handle					g_xUseDamageMapsHandle;
	Asura_Xbox360_Effect_System::Handle					g_xCloakFadeOutHandle;
	int													g_iNoiseTexture;
}

void AVP_Platform_CloakEffect::Initialise()
{
	CaptureEffect();
	g_iNoiseTexture = Asura_Texture_Management::AddTexture("\\SpecialFX\\AVP_PredCloak.dds");

}

void AVP_Platform_CloakEffect::CaptureEffect(void)
{
#define LOADPARAMETER(codeName, shaderName) { codeName = g_pxCloakEffect->GetParameterHandle(shaderName); Asura_Assert(codeName, "Could not get shader parameter '" shaderName "'."); }

	g_pxCloakEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auCloakEffect);

	Asura_Assert( g_pxCloakEffect, "Failed to load predator cloak effect." );

	if (g_pxCloakEffect)
	{	

		g_xMaterialsResponseHandles.InitialiseFromEffect(*g_pxCloakEffect);

		LOADPARAMETER(g_xWorldMatrixHandle, "g_xWorldMatrix");
		LOADPARAMETER(g_xWorldViewMatrixHandle, "g_xWorldViewMatrix");
		LOADPARAMETER(g_xWorldViewProjectionMatrixHandle, "g_xWorldViewProjectionMatrix");
		LOADPARAMETER(g_xObjectSpaceCameraPositionHandle, "g_xObjectSpaceCameraPosition");
		LOADPARAMETER(g_xNoiseTextureHandle, "g_xNoiseTextureSampler");
		LOADPARAMETER(g_xSceneTextureHandle, "g_xSceneSampler");
		
		LOADPARAMETER(g_xOffsetHandle, "g_fOffset");
		LOADPARAMETER(g_xParamsHandle, "g_fParams");
		LOADPARAMETER(g_xCloakFadeOutHandle, "g_xCloakFadeOut");
		
		LOADPARAMETER(g_xLightningColourHandle, "g_fLightningColour");
		LOADPARAMETER(g_xTransitionValueHandle, "g_fTransitionValue");
		LOADPARAMETER(g_xBoneMatricesHandle, "g_axBoneMatrices");
		LOADPARAMETER(g_xUseDamageMapsHandle, "g_bUseDamageMaps");
		
	}
}

void AVP_Platform_CloakEffect::Shutdown()
{
	Asura_Xbox360_Effect_System::Destroy( g_pxCloakEffect );
	g_pxCloakEffect = NULL;

}

void AVP_Platform_CloakEffect::RenderSkin( const AVP_CloakEffect *pxCloakEffect,  const Asura_Animation& xAnimation,
										const Asura_Hierarchy_Skin& xSkin,
										Asura_AMatrix4x4_InParam xTransform,
										Asura_AMatrix4x4_InParam xInverseTransform,
										const Asura_AMatrix4x4* const pxBoneTransforms )
{
	//	const float fAlpha = xAnimation.GetAlpha();
	//	const int iTexIDOverride = xAnimation.GetAnimTexID();

	const Asura_Xbox360_Skin* pxNewSkin = NULL;
	Asura_Xbox360_CharacterRenderer_Skin* pxOldSkin = NULL;
	
	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		pxNewSkin = Asura_Xbox360_SkinRenderer::GetSkin( xSkin.m_uNameHashID );
		if( !pxNewSkin ) return;
	}
	else
	{
		pxOldSkin = Asura_Xbox360_CharacterRenderer::GetCharacter( xSkin.m_uNameHashID );
		if( !pxOldSkin ) return;
	}

	if( !xSkin.NumberOfBones ) return;

	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xFF, 0, 0xFF ), "Character %s", xSkin.Name.GetString() );
	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_SKINS);

#ifdef ASURA_USE_DAMAGE_MAPS	
	const bool bUseDamageMaps = xAnimation.IsUsingDamageMap();
#else
	const bool bUseDamageMaps = false;
#endif

	Asura_Xbox360_Effect *pxEffect = g_pxCloakEffect;

	const bool bTransparentPass = Asura_SortedRender::IsTransparentPass();
	const bool bFullyCloaked = pxCloakEffect->GetMaxBoneValue() - 1.0f >= (pxCloakEffect->GetLightningScale() - 0.00001f);


	// Set world and world-view-proj matrices
	{
		D3DXMATRIX xWorldViewProj;

		xTransform.StoreColsUnaligned(&xWorldViewProj._11);

		// world
		pxEffect->SetMatrix( g_xWorldMatrixHandle, xWorldViewProj );
		// view
		xWorldViewProj *= Asura_Xbox360_Graphics::GetViewMatrix();
		pxEffect->SetMatrix( g_xWorldViewMatrixHandle, xWorldViewProj );

		// projection
		xWorldViewProj *= Asura_Xbox360_Graphics::GetProjectionMatrix();
		pxEffect->SetMatrix( g_xWorldViewProjectionMatrixHandle, xWorldViewProj );
	}

	const u_int uNumBoneMaps = (pxOldSkin ? pxOldSkin->GetNumBoneMaps() : 0);

	// Set bone matrices
	if (uNumBoneMaps == 0 && !Asura_Xbox360_SkinRenderer::IsEnabled())
	{
		D3DXMATRIX axBoneMatrix[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
		memset( axBoneMatrix, 0, sizeof( axBoneMatrix ) );

		for( int i = 0; i < xSkin.NumberOfBones; ++i )
		{
			pxBoneTransforms[i].StoreColsUnaligned(&axBoneMatrix[i]._11);

#ifdef ASURA_USE_DAMAGE_MAPS
			if (bUseDamageMaps)
			{
				axBoneMatrix[i]._33 = xAnimation.GetBoneDamageMapValue(i);
			}
#endif
			axBoneMatrix[i]._32 = pxCloakEffect->GetBoneValue(i);
		}

		pxEffect->GetInterface()->SetMatrixArrayF4x4( g_xBoneMatricesHandle, axBoneMatrix[0], xSkin.NumberOfBones );

		pxEffect->SetBool( g_xUseDamageMapsHandle, bUseDamageMaps );
	}

	// Set render states
	{

		Asura_Render::CurrentStates.RequestAlphaTest( true );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestDitheringMode( true );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferEnabled( false );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

		// Leave wireframe as it was set elsewhere
	}

	// Skin shader needs object-space camera position
	{
		Asura_AVector4	xObjectSpaceCamPos(*Asura_Camera::GetPosition(), 1.0f);

		float			afParam[4];

		xObjectSpaceCamPos	= xObjectSpaceCamPos * xInverseTransform;
		xObjectSpaceCamPos.StoreUnaligned(afParam);

		pxEffect->SetVector( g_xObjectSpaceCameraPositionHandle, afParam );
	}

	// Need to upload this separately for new renderer, since we can't easily pack it into the
	// bone matrices
	pxEffect->SetFloat( g_xTransitionValueHandle, pxCloakEffect->GetTransitionValue() );

	if( pxOldSkin )
	{
		pxOldSkin->GetIB().Bind();
		pxOldSkin->GetVB().Bind();
		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( Asura_Xbox360_CharacterRenderer::GetVertexDeclaration() );
	}
	else
	{
		pxNewSkin->GetIB().Bind();
		pxNewSkin->GetVB().Bind();
		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( Asura_Xbox360_SkinRenderer::GetVertexDeclaration() );

		Asura_Xbox360_SkinRenderer::CopyBones( pxBoneTransforms, xSkin.NumberOfBones, xAnimation );
	}

	// Render one strip for each material

	u_int uStartIndex = 0;

	u_int uTechnique =  bTransparentPass ? bFullyCloaked + 1 : 0;
	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		uTechnique += 3;
	}

	pxEffect->Begin( uTechnique );

	g_pxCloakEffect->SetTexture(g_xNoiseTextureHandle, g_iNoiseTexture);

	if (bTransparentPass==false)
	{	
		// No fogging if we're doing the opaque pass
		Asura_Xbox360_MaterialResponse_System::SetFogParameters( false, NULL, g_xMaterialsResponseHandles, *g_pxCloakEffect);
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF);
	} else {
		g_pxCloakEffect->SetTexture(g_xSceneTextureHandle, Asura_Xbox360_Graphics::GetSceneResolveTexture());
		Asura_Xbox360_Graphics::ResetDepthBias();
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL);
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

	}

	if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback() && bTransparentPass)
	{
		Asura_MaterialResponse_System::GetMaterialProjectContextCallback() (ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID, ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
	}

	const u_int uNumStrips =
		(pxOldSkin ? pxOldSkin->GetNumMaterials() : pxNewSkin->GetNumStrips());

	for( u_int u = 0; u < uNumStrips; ++u )
	{
		const Asura_Xbox360_CharacterRenderer_Strip* pxOldStrip =
			(pxOldSkin ? &pxOldSkin->GetStrip(u) : NULL);

		const Asura_Xbox360_SkinStrip* pxNewStrip =
			(pxNewSkin ? &pxNewSkin->GetStrip(u) : NULL);

#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		const int iSkinElement = (pxOldStrip ? pxOldStrip->m_iElementID : pxNewStrip->m_iElementID);
		if (xAnimation.GetSkinElementVisibility(iSkinElement))
#endif
		{			

			// Set bone matrices
			if (uNumBoneMaps > 0 && !Asura_Xbox360_SkinRenderer::IsEnabled())
			{
				const Asura_Xbox360_CharacterRenderer_Strip& xStrip = pxOldSkin->GetStrip(u);

				D3DXMATRIX axBoneMatrix[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
				memset( axBoneMatrix, 0, sizeof( axBoneMatrix ) );
				const Asura_Xbox360_CharacterRenderer_Skin::BoneMap &xBoneMap = pxOldSkin->GetBoneMap(xStrip.m_uBoneMap);

				for( u_int i = 0; i < xBoneMap.m_uSize; ++i )
				{
					u_int uBone = xBoneMap.m_puEntries[i];
					pxBoneTransforms[uBone].StoreColsUnaligned(&axBoneMatrix[i]._11);

#ifdef ASURA_USE_DAMAGE_MAPS
					if (bUseDamageMaps)
					{
						axBoneMatrix[i]._33 = xAnimation.GetBoneDamageMapValue(uBone);
					}
#endif
					axBoneMatrix[i]._32 = pxCloakEffect->GetBoneValue(uBone);
				}

				pxEffect->GetInterface()->SetMatrixArrayF4x4( g_xBoneMatricesHandle, axBoneMatrix[0], xBoneMap.m_uSize );

				pxEffect->SetBool( g_xUseDamageMapsHandle, bUseDamageMaps );
			}

			const Asura_Hash_ID uMaterialHash1 = pxOldStrip
				? pxOldStrip->m_auMaterialResponseHashIDs[0]
				: pxNewStrip->m_auMaterials[0];

			const Asura_Hash_ID uMaterialHash2 = pxOldStrip
				? pxOldStrip->m_auMaterialResponseHashIDs[1]
				: pxNewStrip->m_auMaterials[1];

			const Asura_MaterialResponse* const pxMaterial1 
				= Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

			bool bVisible = true;

			if( pxMaterial1 )
			{
				if( pxMaterial1->GetDontRender() )
				{
					bVisible = false;
				}
			}
			if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback() && !bTransparentPass)
			{
				Asura_MaterialResponse_System::GetMaterialProjectContextCallback() (pxMaterial1->GetProjectContext(), ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
			}

			if( bVisible )
			{
				PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xC0, 0, 0xC0 ), "Strip %u (primary material 0x%X)", u , uMaterialHash1 );

				// Set material data
				{
					const Asura_MaterialResponse* const pxMaterial2 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash2 );

					Asura_Xbox360_MaterialResponse_System::SetEffectParameters( pxMaterial1, pxMaterial2, &xAnimation, g_xMaterialsResponseHandles, *g_pxCloakEffect );
//					Asura_Xbox360_MaterialResponse_System::SetBlendStates( pxMaterial1 );
					Asura_Xbox360_MaterialResponse_System::SetSHParameters( pxMaterial1, g_xMaterialsResponseHandles, *g_pxCloakEffect);

					if (bTransparentPass)
					{
						Asura_Xbox360_MaterialResponse_System::SetFogParameters( true, pxMaterial1, g_xMaterialsResponseHandles, *g_pxCloakEffect );
					}
				}

				for (int iTiling = 0; iTiling < (bTransparentPass ? Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses() : 1); ++iTiling)
				{
					if (Asura_Xbox360_Graphics::IsTiling2())
					{
						Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
					}
					Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);

					pxEffect->Commit();

					const u_int uNumIndices =
						(pxOldStrip ? pxOldStrip->m_uNumIndices : pxNewStrip->m_uNumIndices);

					Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0, uStartIndex, uNumIndices );
				}

				PIXEndNamedEvent(); // Material



			}
		}

		uStartIndex += (pxOldStrip ? pxOldStrip->m_uNumIndices : pxNewStrip->m_uNumIndices);
	}
	if (Asura_Xbox360_Graphics::IsTiling2())
	{
		Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
	}

	pxEffect->End();

	Asura_Render::AddToTrianglesRendered( pxOldSkin ? pxOldSkin->GetNumTriangles() : pxNewSkin->GetNumTriangles() );

	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

	Asura_Xbox360_Graphics::ResetDepthBias();
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );

	PIXEndNamedEvent(); // Character
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_SKINS);
}

void AVP_CloakEffect::Platform_Initialise(void)
{
	AVP_Platform_CloakEffect::Initialise();
}

void AVP_CloakEffect::Platform_Shutdown(void)
{
	AVP_Platform_CloakEffect::Shutdown();
}


void AVP_CloakEffect::Platform_BeginRender(void) const
{
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );				
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL );				
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL );				
}

static float s_afVariations[AVP_CloakEffect::NUM_CLOAK_EFFECT_PERMUTATIONS][4] =
{
	{ 0.0f, 0.0f, 0.0f, 0.0f},
	{ 0.5f, 0.0f, 0.0f, 0.0f},
	{ 0.0f, 0.5f, 0.0f, 0.0f},
	{ 0.5f, 0.5f, 0.0f, 0.0f},
	{ 0.0f, 0.0f, 0.5f, 0.0f},
	{ 0.5f, 0.0f, 0.5f, 0.0f},
	{ 0.0f, 0.5f, 0.5f, 0.0f},
	{ 0.5f, 0.5f, 0.5f, 0.0f},
};

void AVP_CloakEffect::Platform_RenderSkin(const Asura_Animation& xAnimation,
										  const Asura_Hierarchy_Skin& xSkin) const
{
	if (Asura_SortedRender::IsTransparentPass())
	{
		Asura_Xbox360_Graphics::Resolve( Asura_Xbox360_Graphics::GetSceneResolveTexture() );
	}

	float fTime = Asura_Timers::GetLocalScaledTime() * m_fVariationSpeed;
	float afOffset[4] = {
		Asura_Maths::Sine(fTime) * m_fVariationScale + s_afVariations[m_uSeed][0], 
		Asura_Maths::Cosine(fTime * 1.53f) * m_fVariationScale + s_afVariations[m_uSeed][1], 
		Asura_Maths::Cosine(fTime * 1.17f) * m_fVariationScale + s_afVariations[m_uSeed][2], 
		1.0f
	};
	g_pxCloakEffect->SetVector(g_xOffsetHandle, afOffset);

	float xParams[4] = {m_fEffectScale, 1.0f / m_fLightningScale, 1.0f, 1.0f};
	g_pxCloakEffect->SetVector(g_xParamsHandle, xParams);

	g_pxCloakEffect->SetVector(g_xLightningColourHandle, &m_xLightningColour.m_fR);

	float fCosAng = Asura_Maths::Cosine(m_fMaxVisibilityAngle);
	float afFadeOut[4] = {1.0f / m_fMaxVisibilityDistance, 
		fCosAng, fCosAng < 1.0f ? 1.0f / (1.0f - fCosAng) : 0.0f, 0.0f};

	g_pxCloakEffect->SetVector(g_xCloakFadeOutHandle, afFadeOut);


	AVP_Platform_CloakEffect::RenderSkin(this, xAnimation, xSkin, Asura_Hierarchy_System::s_xBaseTransform, Asura_Hierarchy_System::s_xInverseBaseTransform, Asura_Hierarchy_System::s_axBoneTransforms);
}

void AVP_CloakEffect::Platform_EndRender(void) const
{

	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );				
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL );				
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL );				

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );

	Asura_Xbox360_Graphics::ResetDepthBias();

	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
}


/// End of file. ///
