#ifdef ASURA_USE_AXON
//////////////////////////////////////////////////////////////////////////////
//
// Axon_BehaviourParams.h
//
// Copyright (C) 2008 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __AXON_BEHAVIOURPARAMS_H__
#define __AXON_BEHAVIOURPARAMS_H__ 1

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

class Asura_Chunk_Stream;
class Asura_Message;
class Asura_ServerEntity_Node;
class Asura_Tabs_Dialog;
class Asura_Vector_2;
class Tools_Entity_Node;
class Tools_Parameter;
class Tools_EntityTemplateInstance;
template< typename T > class List;

//////////////////////////////////////////////////////////////////////////////
//
// Axon_BehaviourSet_IDs
// These are the IDs of the different behaviour sets.
// A behaviour set comprises one or more behaviour params of whatever type in a tree structure.
// These params are used to create the tree of behaviours in-game when starting on a GS2 object.
//
// Note that because this enum was introduced late to the code-base the values are a bit odd.
//
//////////////////////////////////////////////////////////////////////////////

enum Axon_BehaviourSet_IDs
{
	AXON_BS_ID_NONE = 0,

	AXON_BS_ID_CORE_NODE = 7,
	AXON_BS_ID_CORE_SPLINE = 8,
	AXON_BS_ID_CORE_ATTRACTOR_EN_ROUTE = 9,

	AXON_BS_ID_FIRST_PROJECT_SPECIFIC = 1000
};

//////////////////////////////////////////////////////////////////////////////
//
// Axon_BehaviourParams_Types
// These are the types of the individual behaviour params classes.
//
//////////////////////////////////////////////////////////////////////////////

// NOTE:
// These values are saved out to chunk buffers, so do not reorder or change them.
enum Axon_BehaviourParams_Types
{
	AXON_BP_TYPE_NONE					= 0,

	// The blank behaviour parameters type just contains a behaviour ID and nothing else.
	AXON_BP_TYPE_BLANK					= 1,

    AXON_BP_TYPE_GS2_NODE				= 2,
    AXON_BP_TYPE_GS2_SPLINE				= 3,
	AXON_BP_TYPE_GS2_ATTRACTOR_ENROUTE	= 4,

	AXON_BP_TYPE_FIRST_PROJECT_SPECIFIC	= 1000
};

//////////////////////////////////////////////////////////////////////////////
//
// Axon_BehaviourParams
//
// 99% of this code is to support the idea of these parameters existing in trees.
// This isn't going to be implemented until phase 2.  :(
//
//////////////////////////////////////////////////////////////////////////////

class Axon_BehaviourParams
{
public:
	Axon_BehaviourParams( const Axon_BehaviourParams& xParams );
	Axon_BehaviourParams( const u_int uBehaviourType );
	Axon_BehaviourParams( const u_int uParamsType, const u_int uBehaviourType );
	virtual ~Axon_BehaviourParams();

	virtual bool IsBaseOrDerivativeOf( const u_int uParamsType ) const { return ( m_uParamsType == AXON_BP_TYPE_NONE ); }
	u_int GetParamsType() const { return m_uParamsType; }
	u_int GetBehaviourType() const { return m_uBehaviourType; }

	void LinkChild( Axon_BehaviourParams* pxBehaviourParams );
	bool UnlinkChild( Axon_BehaviourParams* pxBehaviourParams );

	Axon_BehaviourParams* GetChild()   const { return m_pxChild; }
	Axon_BehaviourParams* GetSibling() const { return m_pxSibling; }

	bool DoHandleMessage( const Asura_Message* pxMessage );

	// Creates the behaviour params tree that represents the chosen behaviour set.
	static void CreateTree( const int iBehaviourSet, Axon_BehaviourParams*& pxBehaviourParams );
	// Converts between the old behaviour params tree that represented the chosen behaviour set and a new tree that now represents it.
	// Returns whether it did any work.
	static bool ConvertTree( const int iBehaviourSet, const Axon_BehaviourParams* pxOldBehaviourParams, Axon_BehaviourParams*& pxNewBehaviourParams );
	static void CreateTreeFromChunkStream( Asura_Chunk_Stream& xStream, Axon_BehaviourParams*& pxBehaviourParams );
	static void WriteTreeToChunkStream( const Axon_BehaviourParams* pxBehaviourParams, Asura_Chunk_Stream& xStream );

protected:
	void LinkSibling( Axon_BehaviourParams* pxBehaviourParams );
	bool UnlinkSibling( Axon_BehaviourParams* pxBehaviourParams );

	virtual void ReadFromChunkStream( Asura_Chunk_Stream& xStream );
	virtual void WriteToChunkStream( Asura_Chunk_Stream& xStream ) const;
	virtual bool HandleMessage( const Asura_Message* pxMessage ) { return false; }

private:
	int GetNumChildren() const;

	// Called during reading in from a chunk to re-create the behaviour parameter tree.
	static Axon_BehaviourParams* Create( const u_int uParamType, const u_int uBehaviourType );
	// Project_Create is called from Create to allow the project to create project-specific behaviour params.
	static Axon_BehaviourParams* Project_Create( const u_int uParamType, const u_int uBehaviourType );
	// Project_CreateTree is called from a GS2 object when it has a new behaviour set on it.
	// This should create the tree of all behaviour parameters associated with that behaviour set.
	static bool					 Project_CreateTree( const int iBehaviourSet, Axon_BehaviourParams*& pxBehaviourParams );
	// Project_ConvertTree exists to allow old trees to be updated if the behaviour parameters associated with a behaviour set have changed.
	static bool					 Project_ConvertTree( const int iBehaviourSet, const Axon_BehaviourParams* pxOldBehaviourParams, Axon_BehaviourParams*& pxNewBehaviourParams );

private:
	u_int				  m_uParamsType;
	u_int				  m_uBehaviourType;
	Axon_BehaviourParams* m_pxChild;
	Axon_BehaviourParams* m_pxSibling;

	enum { CURRENT_MASTER_CHUNK_VERSION = 0 };
	enum { CURRENT_CHUNK_VERSION = 0 };

#ifdef ASURA_TOOLS
	// Tools only interface and parameters.
public:
	void DoSetToolsNode( Tools_Entity_Node* pxToolsNode );
	void DoAddOverlayControls( Asura_Tabs_Dialog* pxTabsDialog, Asura_Vector_2& xOffset );
	void DoRender();
	void DoUpdate();
	void DoUpdateWhileSelected();
	void DoStartPreview();
	// Returns whether any of the parameter tree is still previewing.
	bool DoUpdatePreview();
	void DoStopPreview();
	void DoAddFilesToExportList( List< Asura_String >& xNameList ) const;
	void DoAddFilesToExportListFromMessage( u_short usMessageID, void* pData, List< Asura_String >& xNameList ) const;
	bool DoHandlesThisMessage( const int iMessageID );
	bool DoOwnsThisMessageBlock( const u_short usMessageBlockID );
	bool DoGetParameter( const Asura_Hash_ID uParamID, Tools_Parameter* pxParam ) const;
	bool DoSetParameter( const Asura_Hash_ID uParamID, const Tools_Parameter* pxParam );
	void DoApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance );
	bool DoVerify( const bool bAutomatic );
	Axon_BehaviourParams* DoClone() const;

protected:
	virtual void SetToolsNode( Tools_Entity_Node* pxToolsNode ) { m_pxToolsNode = pxToolsNode; }
	Tools_Entity_Node* GetToolsNode() const { return m_pxToolsNode; }

	virtual void AddOverlayControls( Asura_Tabs_Dialog* pxTabsDialog, Asura_Vector_2& xOffset ) {}
	virtual void Render() {}
	virtual void Update() {}
	virtual void UpdateWhileSelected() {}
	virtual void StartPreview() {}
	virtual bool UpdatePreview() { return true; }
	virtual void StopPreview() {}
	virtual void AddFilesToExportList( List< Asura_String >& xNameList ) const {}
	virtual void AddFilesToExportListFromMessage( u_short usMessageID, void* pData, List< Asura_String >& xNameList ) const {}
	virtual bool HandlesThisMessage( const int iMessageID ) { return false; }
	virtual bool OwnsThisMessageBlock( const u_short usMessageBlockID ) { return false; }
	virtual bool GetParameter( const Asura_Hash_ID uParamID, Tools_Parameter* pxParam ) const { return false; }
	virtual bool SetParameter( const Asura_Hash_ID uParamID, const Tools_Parameter* pxParam ) { return false; }
	virtual void ApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance ) {}
	virtual bool Verify( const bool bAutomatic ) { return true; }
	virtual Axon_BehaviourParams* Clone() const;

private:
	Tools_Entity_Node* m_pxToolsNode;
#endif
};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif //__AXON_BEHAVIOURPARAMS_H__
#endif // ASURA_USE_AXON
