//////////////////////////////////////////////////////////////////////////////
//
// Axon_Brain.cpp
//
// Created by NJD - 17 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_AXON

// header
#include "Core/Axon_Brain.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Camera/Asura_Camera.h"
#endif
#include "Core/Asura_Callback.h"
#include "Core/Asura_Scheduler.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Network/Asura_Message.h"

// axon
#include "Behaviour/Axon_Behaviour.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_GS2Actor.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_GS2Attractor.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_GS2Node.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_GS2Spline.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Types.h"
#include "Core/Axon_Messages.h"
#include "Core/Axon_System.h"

#ifdef ASURA_USE_NEW_NAV_STUFF
#include "Behaviour/BehaviourTypes/Axon_Behaviour_FollowActor.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Idle.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_LookAt.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Movement.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_UseGuardZone.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Wander.h"
#endif // #ifdef ASURA_USE_NEW_NAV_STUFF

// tools
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Axon_Brain
//
//////////////////////////////////////////////////////////////////////////////

u_int Axon_Brain::s_uCurrentBehaviourID = AXON_INVALID_BEHAVIOUR_ID;

Axon_Brain::Axon_Brain( const Asura_Guid uOwnerGuid )
: m_uOwnerGuid( uOwnerGuid )
, m_iBehaviourUpdateHandle( ASURA_INVALID_HANDLE )
, m_pxFirstBehaviour( NULL )
, m_uBrainFlags( 0 )
, m_xMessageQueue()
#ifndef ASURA_NO_DEBUG_OUTPUT
, m_xDebugTextPos( Asura_Maths::ZeroVector3 )
, m_bDebugTextPosOnScreen( false )
#endif
{
    Axon_System::RegisterBrain( this );
}

Axon_Brain::~Axon_Brain()
{
    Asura_Assert( !m_bCallbacksRegistered, "Axon_Brain::~Axon_Brain - callbacks are still registered - this is highly likely to cause a crash" );
    Axon_System::UnregisterBrain( this );
}

void Axon_Brain::Initialise()
{
    OnActivate();
    m_bActive = true;
}

void Axon_Brain::PostLoadInitialise()
{
	for ( Axon_Behaviour* pxTopBehaviour = m_pxFirstBehaviour; pxTopBehaviour; pxTopBehaviour = pxTopBehaviour->GetStackChild() )
	{
		pxTopBehaviour->DoPostLoadInitialise();
	}
}

void Axon_Brain::Shutdown()
{
    Asura_Assert( !m_bUpdatingBehaviours,           "Axon_Brain::Shutdown - please do not shutdown whilst behaviours are mid-update" );
    Asura_Assert( !m_xMessageQueue.GetNumEntries(), "Axon_Brain::Shutdown - we should not have messages still stored at this point" );

	if ( IsActive() )
	{
		m_bActive = false;
		OnDeactivate();
	}
	// Force these to be called in case we've created behaviours or registered callbacks outside the control of activation/deactivation code.
	UnregisterCallbacks();
	DestroyAllBehaviours();
}

void Axon_Brain::DoHandleMessage( const Asura_Message* pxMessage )
{
    if ( m_bUpdatingBehaviours )
    {
        // Queue this message and handle it after we've finished updating everything.
        Asura_Message* pxNewMessage = Asura_Network::DuplicateMessage( pxMessage, false );
        m_xMessageQueue.AddToList( pxNewMessage );
        return;
    }

	if ( !HandleMessage( pxMessage ) )
	{
		Asura_Assertf( false, ("Axon_Brain::DoHandleMessage - message %s not handled", Asura_Network::GetMsgIDDesc(pxMessage->GetMsgID()) ));
	}
}

void Axon_Brain::DoPerFrameUpdate( const float fTimeslice )
{
    if ( m_bPaused || !m_bActive ) { return; }

    if ( m_pxFirstBehaviour )
    {
        m_bUpdatingBehaviours = true;
        m_pxFirstBehaviour->DoPerFrameUpdate( fTimeslice );
        m_bUpdatingBehaviours = false;
    }

    // If we've queued any messages whilst the behaviours were updating then handle them now.
    if ( m_xMessageQueue.GetNumEntries() )
    {
        Asura_Message* pxMessage = NULL;
        while ( ( pxMessage = m_xMessageQueue.PopTailObject() ) != NULL )
        {
            if ( !HandleMessage( pxMessage ) )
            {
                Asura_Assert( false, "Axon_Brain::DoPerFrameUpdate - queued message not handled" );
            }
            Asura_Network::DestroyMessage( pxMessage );
        }
    }
}

#ifndef ASURA_NO_DEBUG_OUTPUT
bool Axon_Brain::SetOwnerDebugTextStartPos()
{
	m_bDebugTextPosOnScreen = false;
	Asura_ServerEntity_PhysicalObject *pxOwnerPhyObj = Asura_ServerEntity_PhysicalObject::FindEntity( GetOwnerGuid() );
	if( pxOwnerPhyObj )
	{
		m_xDebugTextPos = pxOwnerPhyObj->GetPosition();
		m_xDebugTextPos.y -= 1.5f; // Move start position off the ground a bit up to the body.
		m_bDebugTextPosOnScreen = true;
	}
	return m_bDebugTextPosOnScreen;
}

void Axon_Brain::DoRenderDebugInfo()
{
	// Set up the on screen text position to use.
	if ( SetOwnerDebugTextStartPos() )
	{
		m_bDebugTextPosOnScreen = Asura_Camera::ProjectWorldCoordsToScreenSpace( m_xDebugTextPos, true );
	}

	// Render behaviour debug info
	if ( DebugTextPosOnScreen() )
	{
		if ( m_pxFirstBehaviour )
		{
			m_pxFirstBehaviour->DoRenderDebugInfo();
		}
	}
}
#endif

void Axon_Brain::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	DestroyAllBehaviours();

	int iVersion;
    xStream >> iVersion;

    if ( iVersion == CURRENT_CHUNK_VERSION )
    {
        Asura_ReadFromStream_Bitfield_Wrapper< u_int > xBitfieldWrapper( m_uBrainFlags );
        xStream >> xBitfieldWrapper;

        int iNumTopBehaviours;
        xStream >> iNumTopBehaviours;

        for ( int iStackIndex = 0; iStackIndex < iNumTopBehaviours; ++iStackIndex )
        {
            u_int uBehaviourType;
            xStream >> uBehaviourType;
            Asura_Assert( uBehaviourType != AXON_BEHAVIOUR_TYPE_UNSET, "Axon_Brain::ReadFromChunkStream - bad behaviour type" );

            Axon_Behaviour* pxTopBehaviour = CreateBehaviour( uBehaviourType );
            Asura_Assert( pxTopBehaviour, "Axon_Brain::ReadFromChunkStream - could not create behaviour" );
            if ( pxTopBehaviour )
            {
                AddToBottomOfStack( pxTopBehaviour );
                pxTopBehaviour->ReadFromChunkStream( xStream );
            }
        }
    }
    else if ( iVersion == 3 )
    {
        u_int uPaused;
        xStream >> uPaused;
        m_bPaused = ( uPaused == 0 ) ? false : true;

        int iNumTopBehaviours;
        xStream >> iNumTopBehaviours;

        for ( int iStackIndex = 0; iStackIndex < iNumTopBehaviours; ++iStackIndex )
        {
            u_int uBehaviourType;
            xStream >> uBehaviourType;
            Asura_Assert( uBehaviourType != AXON_BEHAVIOUR_TYPE_UNSET, "Axon_Brain::ReadFromChunkStream - bad behaviour type" );

            Axon_Behaviour* pxTopBehaviour = CreateBehaviour( uBehaviourType );
            Asura_Assert( pxTopBehaviour, "Axon_Brain::ReadFromChunkStream - could not create behaviour" );
            if ( pxTopBehaviour )
            {
                AddToBottomOfStack( pxTopBehaviour );
                pxTopBehaviour->ReadFromChunkStream( xStream );
            }
        }
    }
    else if ( iVersion == 2 )
    {
        int iNumTopBehaviours;
        xStream >> iNumTopBehaviours;

        for ( int iStackIndex = 0; iStackIndex < iNumTopBehaviours; ++iStackIndex )
        {
            int iBehaviourType;
            xStream >> iBehaviourType;
            Asura_Assert( iBehaviourType != AXON_BEHAVIOUR_TYPE_UNSET, "Axon_Brain::ReadFromChunkStream - bad behaviour type" );

            Axon_Behaviour* pxTopBehaviour = CreateBehaviour( iBehaviourType );
            Asura_Assert( pxTopBehaviour, "Axon_Brain::ReadFromChunkStream - could not create behaviour" );
            if ( pxTopBehaviour )
            {
                AddToBottomOfStack( pxTopBehaviour );
                pxTopBehaviour->ReadFromChunkStream( xStream );
            }
        }
    }
    else if ( iVersion == 1 )
    {
        int iNumTopBehaviours;
        xStream >> iNumTopBehaviours;

        for ( int iStackIndex = 0; iStackIndex < iNumTopBehaviours; ++iStackIndex )
        {
            int iBehaviourType;
            xStream >> iBehaviourType;
            Asura_Assert( iBehaviourType != AXON_BEHAVIOUR_TYPE_UNSET, "Axon_Brain::ReadFromChunkStream - bad behaviour type" );

            Axon_Behaviour* pxTopBehaviour = CreateBehaviour( iBehaviourType );
            Asura_Assert( pxTopBehaviour, "Axon_Brain::ReadFromChunkStream - could not create behaviour" );
            if ( pxTopBehaviour )
            {
                AddToBottomOfStack( pxTopBehaviour );
                pxTopBehaviour->ReadFromChunkStream( xStream );;

                int iNumSiblingBehaviours;
                xStream >> iNumSiblingBehaviours;
                for ( int iSiblingIndex = 0; iSiblingIndex < iNumSiblingBehaviours; ++iSiblingIndex )
                {
                    xStream >> iBehaviourType;
                    Asura_Assert( iBehaviourType != AXON_BEHAVIOUR_TYPE_UNSET, "Axon_Brain::ReadFromChunkStream - bad behaviour type" );

                    Axon_Behaviour* pxSiblingBehaviour = CreateBehaviour( iBehaviourType );
                    Asura_Assert( pxSiblingBehaviour, "Axon_Brain::ReadFromChunkStream - could not create behaviour" );
                    if ( pxSiblingBehaviour )
                    {
                        pxTopBehaviour->LinkSibling( pxSiblingBehaviour );
                        pxSiblingBehaviour->ReadFromChunkStream( xStream );
                    }
                }
            }
        }
    }
    else if ( iVersion == 0 )
    {
    }
    else
    {
        Asura_Assertf( false, ( "Axon_Brain::ReadFromChunkStream - unhandled chunk version (%d)", iVersion ) );
    }

    if ( m_bCallbacksRegistered )
    {
        RegisterCallbacks();
    }
}

void Axon_Brain::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
    xStream << static_cast< int >( CURRENT_CHUNK_VERSION );

    Asura_WriteToStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uBrainFlags );
    xFlagsBitfield.Add( 0,                      1 ); // not writing the m_bUpdatingBehaviours field
    xFlagsBitfield.Add( m_bPaused,              1 );
    xFlagsBitfield.Add( m_bActive,              1 );
    xFlagsBitfield.Add( m_bCallbacksRegistered, 1 );
    xStream << xFlagsBitfield;

    // First count the number of top behaviours down the stack and write this number out.
    int iNumTopBehaviours = 0;
    for ( Axon_Behaviour* pxTopBehaviour = m_pxFirstBehaviour; pxTopBehaviour; pxTopBehaviour = pxTopBehaviour->GetStackChild() )
    {
        ++iNumTopBehaviours;
    }
    xStream << iNumTopBehaviours;

    // Iterate down the top behaviours and write their data out. They can decide themselves how much of their children's data they want to save.
    for ( Axon_Behaviour* pxTopBehaviour = m_pxFirstBehaviour; pxTopBehaviour; pxTopBehaviour = pxTopBehaviour->GetStackChild() )
    {
        xStream << pxTopBehaviour->GetType();
        pxTopBehaviour->WriteToChunkStream( xStream );
    }
}

bool Axon_Brain::IsOnGamescene() const
{
	return ( m_pxFirstBehaviour &&
		( m_pxFirstBehaviour->GetType() == AXON_BEHAVIOUR_TYPE_GS2_ACTOR ||
		  m_pxFirstBehaviour->GetChildWithType( m_pxFirstBehaviour, AXON_BEHAVIOUR_TYPE_GS2_ACTOR, true ) ) );
}

#ifdef ASURA_TOOLS
void Axon_Brain::ApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance )
{
	// We do not need to convert this guid at all! it has already been done. the current m_uOwnerGuid is correct.
	// m_uOwnerGuid = pxInstance->ConvertGuid( m_uOwnerGuid );

    for ( Axon_Behaviour* pxTopBehaviour = m_pxFirstBehaviour; pxTopBehaviour; pxTopBehaviour = pxTopBehaviour->GetStackChild() )
    {
        pxTopBehaviour->DoApplyChangesForTemplateInstance( pxInstance );
    }
}

void Axon_Brain::Copy( const Axon_Brain& xSourceBrain )
{
	Asura_Chunk_Memory_Stream xStream;
	xStream.SetSwapEndian( false );
	xSourceBrain.WriteToChunkStream( xStream );
	xStream.SetPosition( 0 );
	ReadFromChunkStream( xStream );
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Axon_Brain::OnActivate()
{
	RegisterCallbacks();
}

void Axon_Brain::OnDeactivate()
{
	UnregisterCallbacks();
	DestroyAllBehaviours();
}

void Axon_Brain::OnPause()
{
	if ( m_pxFirstBehaviour )
	{
		m_pxFirstBehaviour->DoPause();
	}
	UnregisterCallbacks();
}

void Axon_Brain::OnUnpause()
{
	RegisterCallbacks();
	if ( m_pxFirstBehaviour )
	{
		m_pxFirstBehaviour->DoUnpause();
	}
}

void Axon_Brain::RegisterCallbacks()
{
    if ( m_iBehaviourUpdateHandle == ASURA_INVALID_HANDLE )
	{
	    Asura_Scheduler& xBehaviourUpdateScheduler = Axon_System::GetBehaviourUpdateScheduler();
	    Asura_Callback_MemberFunction_1Arg<void, float,  Axon_Brain > xBehaviourUpdateCallback( this, &Axon_Brain::DoBehaviourUpdate );
		m_iBehaviourUpdateHandle = xBehaviourUpdateScheduler.RegisterCallback( 0.1f, &xBehaviourUpdateCallback );
	}
    m_bCallbacksRegistered = true;
}

void Axon_Brain::UnregisterCallbacks()
{
    m_bCallbacksRegistered = false;
    if ( m_iBehaviourUpdateHandle != ASURA_INVALID_HANDLE )
    {
        Axon_System::GetBehaviourUpdateScheduler().UnregisterCallback( m_iBehaviourUpdateHandle );
		m_iBehaviourUpdateHandle = ASURA_INVALID_HANDLE;
    }
}

bool Axon_Brain::HandleMessage( const Asura_Message* pxMessage )
{
    bool bHandledMessage   = false;
    bool bPassToBehaviours = false;

	switch ( pxMessage->GetMsgID() )
	{
        case ASURA_MSG_ACTIVATE:
        {
            if ( !m_bActive )
            {
				OnActivate();
                m_bActive = true;
            }
           
            bHandledMessage = true;
			break;
        }

        case ASURA_MSG_DEACTIVATE:
        {
            if ( m_bActive )
            {
                m_bActive = false;
				OnDeactivate();
            }

            bHandledMessage = true;
			break;
        }

        case ASURA_MSG_AXON_SET_PAUSED:
        {
            if ( m_bActive )
            {
                const Axon_Msg_Set_Paused* pxSetPausedData = static_cast< const Axon_Msg_Set_Paused* >( pxMessage->GetData() );
                if ( pxSetPausedData )
                {
                    if ( pxSetPausedData->m_bPause != static_cast< bool >( m_bPaused ) )
                    {
                        if ( pxSetPausedData->m_bPause )
                        {
							OnPause();
                        }
                        else
                        {
							OnUnpause();
                        }

                        m_bPaused = pxSetPausedData->m_bPause;
                    }
                }
            }

            // Say we handled this regardless of whether anything changed.
            bHandledMessage = true;
			break;
        }

		case ASURA_MSG_AXON_GS2_ATTRACTOR_PULSE:
		case ASURA_MSG_AXON_GS2_TELEPORT_TO_NODE:
		{
			// These are messages that should trigger the gamescene behaviour.
			// Check to see if the behaviour exists. If it doesn't, create it and push it onto the stack.
			if ( !IsOnGamescene() )
			{
				Axon_Behaviour* pxBehaviour = CreateActivateAndPushBehaviour( AXON_BEHAVIOUR_TYPE_GS2_ACTOR );
				if ( pxBehaviour )
				{
					pxBehaviour->DoHandleMessage( pxMessage );
				}
				bHandledMessage = true;
			}
			else
			{
				bPassToBehaviours = true;
			}

			break;
		}

		case ASURA_MSG_AXON_GS2_ANIM_CHANGE:
		case ASURA_MSG_AXON_GS2_ATTRACTOR_CANCEL:
		case ASURA_MSG_AXON_GS2_DONT_STOP_AT_NODE:
		case ASURA_MSG_AXON_GS2_EXIT_GAMESCENE:
		case ASURA_MSG_AXON_GS2_LEAVE_NODE:
		case ASURA_MSG_AXON_GS2_PATHFIND_TO_NODE:
		case ASURA_MSG_AXON_GS2_TAKE_LINK:
		{
			if ( IsOnGamescene() )
			{
				bPassToBehaviours = true;
			}
			else
			{
				bHandledMessage = true;
			}

			break;
		}

		default:
		{
            // Maybe the behaviours will know something about this message.
            bPassToBehaviours = true;
			break;
		}
	}

    if ( bPassToBehaviours && m_pxFirstBehaviour )
    {
        bHandledMessage |= m_pxFirstBehaviour->DoHandleMessage( pxMessage );
    }

	return bHandledMessage;
}

void Axon_Brain::UpdateBehaviours( const float fTimeslice )
{
	if ( m_bPaused || !m_bActive ) { return; }

    if ( m_pxFirstBehaviour )
    {
        m_bUpdatingBehaviours = true;

        Asura_Assert( ( m_pxFirstBehaviour->GetSibling() == NULL ), "Axon_Brain::UpdateBehaviours - the top behaviour should not have siblings" );
		m_pxFirstBehaviour->DoUpdate( fTimeslice );
        if ( m_pxFirstBehaviour->IsFinished() || m_pxFirstBehaviour->IsInError() )
        {
            // The top behaviour has finished or is in error and must be removed.
            Axon_Behaviour* pxOldBehaviour = PopAndDeactivateBehaviour();
            DestroyBehaviour( pxOldBehaviour );
        }

        m_bUpdatingBehaviours = false;
    }

    // If we've queued any messages whilst the behaviours were updating then handle them now.
    if ( m_xMessageQueue.GetNumEntries() )
    {
        Asura_Message* pxMessage = NULL;
        while ( ( pxMessage = m_xMessageQueue.PopTailObject() ) != NULL )
        {
            if ( !HandleMessage( pxMessage ) )
            {
                Asura_Assert( false, "Axon_Brain::UpdateBehaviours - queued message not handled" );
            }
            Asura_Network::DestroyMessage( pxMessage );
        }
    }
}

void Axon_Brain::DestroyAllBehaviours()
{
    Axon_Behaviour* pxBehaviour = PopAndDeactivateBehaviour();
    while ( pxBehaviour )
    {
        DestroyBehaviour( pxBehaviour );
        pxBehaviour = PopAndDeactivateBehaviour();
    }
}

Axon_Behaviour* Axon_Brain::CreateActivateAndPushBehaviour( const u_int uBehaviourType, const Asura_Message* pxMessage /*= NULL*/ )
{
	Axon_Behaviour* pxBehaviour = CreateBehaviour( uBehaviourType, pxMessage );
	if ( !pxBehaviour )
	{
		return NULL;
	}

	if ( m_pxFirstBehaviour )
	{
		m_pxFirstBehaviour->DoPause();
	}

	if ( pxBehaviour->DoActivate() )
	{
		pxBehaviour->SetStackChild( m_pxFirstBehaviour );
		m_pxFirstBehaviour = pxBehaviour;
		return pxBehaviour;
	}
	else
	{
		pxBehaviour->DoDeactivate();
		DestroyBehaviour( pxBehaviour );
		if ( m_pxFirstBehaviour )
		{
			m_pxFirstBehaviour->DoUnpause();
		}
		return NULL;
	}
}

Axon_Behaviour* Axon_Brain::PopAndDeactivateBehaviour()
{
    if ( m_pxFirstBehaviour )
    {
        Axon_Behaviour* pxOldFirstBehaviour = m_pxFirstBehaviour;
		m_pxFirstBehaviour = pxOldFirstBehaviour->GetStackChild();
		pxOldFirstBehaviour->DoDeactivate();
        if ( m_pxFirstBehaviour )
        {
            m_pxFirstBehaviour->DoUnpause();
        }

        return pxOldFirstBehaviour;
    }
    else
    {
        return NULL;
    }
}

void Axon_Brain::AddToBottomOfStack( Axon_Behaviour* pxBehaviour )
{
	if ( m_pxFirstBehaviour == NULL )
	{
		m_pxFirstBehaviour = pxBehaviour;
	}
	else
	{
		Axon_Behaviour* pxStackedBehaviour = m_pxFirstBehaviour;
		while ( pxStackedBehaviour->GetStackChild() != NULL )
		{
			pxStackedBehaviour = pxStackedBehaviour->GetStackChild();
		}
		pxStackedBehaviour->SetStackChild( pxBehaviour );
	}
}

size_t Axon_Brain::GetLargestCoreBehaviourSize()
{
    size_t uLargestBehaviourSize = 0;
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_GS2Actor );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_GS2AttractorEnRoute );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_GS2Node );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_GS2Spline );

	#ifdef ASURA_USE_NEW_NAV_STUFF
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_FollowActor );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_Idle );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_LookAt );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_Movement );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_UseGuardZone );
	AXON_CALC_MAX_BEHAVIOUR_SIZE( uLargestBehaviourSize, Axon_Behaviour_Wander );
	#endif // #ifdef ASURA_USE_NEW_NAV_STUFF

	return uLargestBehaviourSize;
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif
