//////////////////////////////////////////////////////////////////////////////
//
// main.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <cell/dma.h>
#include <cell/sync.h>
#include <edge/geom/edgegeom.h>
#include <sys/spu_thread.h>
#include <spu_intrinsics.h>
#include <spu_printf.h>
#include <string.h>

// Engine independent.
#include "Core/Memory/Asura_Mem_ResetOnlyAllocator.h"
#include "Core/Asura_Maths.h"

// Engine platform-specific.
#include "Debugging/Asura_Assert.h"
//#include "Rendering/ShadersDefinition/VD/Env_VD.h"
#include "Rendering/Asura_PS3_EnvironmentRenderer_SPUJob.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// Problem with (adjusted) indices >= 4096? 4096 x 16 (ie. vtx size) = 65536...
// This should be a multiple of 8. Note that this alone doesn't always help
// because some strips or even just tris still reference huge ranges of
// vertices, eg. (24, 4850, 4850), which rathers scuppers this whole thing.
#define FIXED_MAX_BATCH_SIZE (2048)

#define DEBUG_OUTPUT_LEVEL (0)

#if DEBUG_OUTPUT_LEVEL == 3
    #define DEBUG_TRACE_INSANE( args ) \
        { spu_printf args ; }
    #define DEBUG_TRACE_DETAILED( args ) \
        { spu_printf args ; }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { spu_printf args ; }
#elif DEBUG_OUTPUT_LEVEL == 2
    #define DEBUG_TRACE_INSANE( args ) \
        { }
    #define DEBUG_TRACE_DETAILED( args ) \
        { spu_printf args ; }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { spu_printf args ; }
#elif DEBUG_OUTPUT_LEVEL == 1
    #define DEBUG_TRACE_INSANE( args ) \
        { }
    #define DEBUG_TRACE_DETAILED( args ) \
        { }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { spu_printf args ; }
#else
    #define DEBUG_TRACE_INSANE( args ) \
        { }
    #define DEBUG_TRACE_DETAILED( args ) \
        { }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { }
#endif

#define STACK_CHECK_SIZE (0)

#if STACK_CHECK_SIZE
    #define PREPARE_STACK_CHECK() \
    { \
        u_char* pucStart = reinterpret_cast<u_char*>( 256 * 1024 - uSTACK_SIZE_TO_ALLOW - STACK_CHECK_SIZE ); \
        memset( pucStart, 0xAB, STACK_CHECK_SIZE ); \
    }

    #define DO_STACK_CHECK( _code ) \
    { \
        const u_char* pucStart = reinterpret_cast<u_char*>( 256 * 1024 - uSTACK_SIZE_TO_ALLOW - STACK_CHECK_SIZE ); \
        for( u_int u = 0; u < STACK_CHECK_SIZE; ++u ) \
        { \
            if( *pucStart != 0xAB ) \
            { \
                /* If only the end of this area is modified then the stack */ \
                /* is probably overgrown. If it's just the beginning then  */ \
                /* probably something's written too high.                  */ \
                spu_printf( "SPU ERROR: stack splatted or overgrown! (code: %i)", _code ); \
                \
                /* Halt to try and make it easier to find the offending.   */ \
                /* location. Also pass the specified code, so that this    */ \
                /* can also be used to signal where this check was being   */ \
                /* done.                                                   */ \
                si_stopd( si_from_int(_code), si_from_int(_code), si_from_int(_code) ); \
            } \
        } \
    }
#else
    #define PREPARE_STACK_CHECK() { }
    #define DO_STACK_CHECK( _code ) { }
#endif

// G L O B A L S /////////////////////////////////////////////////////////////

static const u_int uSTACK_SIZE_TO_ALLOW = 4 * 1024; // Needs to be a multiple of 16.

extern u_int _end; // End of executable (plus static data, etc.)

static union
{
    Asura_PS3_EnvironmentRenderer_SPUJob g_xNextJob;
    char g_acNextJobBuffer[(sizeof(Asura_PS3_EnvironmentRenderer_SPUJob) + 15) & ~15];
} __attribute__(( aligned(16) ));

static union
{
    EdgeGeomViewportInfo g_xViewportInfo;
    char g_acViewportInfoBuffer[(sizeof(EdgeGeomViewportInfo) + 15) & ~15];
} __attribute__(( aligned(16) ));

// We'll propogate the new index count to all components of this vector and
// then write back the appropriately aligned component to main RAM (because
// the low 4 bits in the ls and ea args to cellDmaSmallPut() must match).
static vec_uint4 g_xOutputIndexCount;

static Asura_Mem_ResetOnlyAllocator g_xAllocator;

// These have to be 0 under the current SDK.
static const u_int uDMA_TID = 0;
static const u_int uDMA_RID = 0;

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static inline u_int ConvertToTriList( u_short* pusIn, u_int uNumInputIndices,
                                      u_short* pusOut,
                                      u_short* pusMinIdx, u_short* pusMaxIdx,
                                      u_short* pusEndIndices, u_int* puNumEndIndices, bool* pbForceFlipOnFirstTri )
{
    u_short* const pusOutStart = pusOut;

    DEBUG_TRACE_INSANE(( "\nSTRIP INPUT:\n" ));
    for( u_int u = 0; u < (uNumInputIndices >> 2); ++u )
    {
        DEBUG_TRACE_INSANE(( "%6u %6u %6u %6u\n", pusIn[u << 2], pusIn[(u << 2) + 1], pusIn[(u << 2) + 2], pusIn[(u << 2) + 3] ));
    }
    for( u_int u = uNumInputIndices & ~3; u < uNumInputIndices; ++u )
    {
        DEBUG_TRACE_INSANE(( "%6u ", pusIn[u] ));
    }
    DEBUG_TRACE_INSANE(( "\n" ));

    // Input strip is CCW,CW,CCW,CW,... and output tri list needs
    // to be CCW,CCW,CCW,CCW,...
    bool bFlipWinding = false;
    bool bNeedToForceFlipNextRestart = false;

    u_int uNumIndicesLeft;

    // Patch in anything from the end of the previous batch.
    u_int uNumEndIndices = *puNumEndIndices;
    if( uNumEndIndices )
    {
        // It's possible the last index of the first triangle will be a
        // restart index - we didn't get to consider it last time around.
        // (Note that because of the way we handle restarts in the main
        // conversion loop below, the other 2 can't be restarts unless the
        // strip pointlessly contains multiple restarts in a row. We're not
        // bothering to cope with that.)
        if( uNumEndIndices == 2 && *pusIn == 0xFFFF )
        {
            // Swallow this first restart index.
            ++pusIn;
            uNumIndicesLeft = uNumInputIndices - 1;
        }
        else
        {
            // Just prepend the extra indices. We allowed for such things
            // when calculating buffer sizes.
            for( u_int u = uNumEndIndices; u--; )
            {
                *--pusIn = pusEndIndices[u];
            }
            uNumIndicesLeft = uNumInputIndices + uNumEndIndices;

            // May need to force a flip for the first tri in this case.
            if( *pbForceFlipOnFirstTri )
            {
                bNeedToForceFlipNextRestart = true;
            }
        }
    }
    else
    {
        uNumIndicesLeft = uNumInputIndices;
    }

    // In any case, pretend we've got a restart index pending to kick-start
    // the loop.
    u_short usIdx0 = 0xFFFF; // Don't need to initialise 0 or 1 - just to
    u_short usIdx1 = 0xFFFF; // keep compiler quiet and for logging purposes.
    u_short usIdx2 = 0xFFFF;

    u_short usMinIdx = 0xFFFF, usMaxIdx = 0;
    bool bNeedToFlipFirstTriNextBatch = false;

    DEBUG_TRACE_INSANE(( "\nCONVERSION LOG:\n" ));

    while( true )
    {
        // Restart index?
        if( usIdx2 == 0xFFFF )
        {
            DEBUG_TRACE_INSANE(( "got restart\n" ));

            if( uNumIndicesLeft < 3 )
            {
                // Retain any remaining input for next time.
                uNumEndIndices = uNumIndicesLeft;
                if( uNumIndicesLeft >= 1 ) { pusEndIndices[0] = *pusIn++; }
                if( uNumIndicesLeft == 2 ) { pusEndIndices[1] = *pusIn++; }

                break;
            }

            usIdx0 = *pusIn++;
            usIdx1 = *pusIn++;
            usIdx2 = *pusIn++;

            uNumIndicesLeft -= 3;

            DEBUG_TRACE_INSANE(( "loaded fresh tri: %6u %6u %6u\n", static_cast<u_int>(usIdx0), static_cast<u_int>(usIdx1), static_cast<u_int>(usIdx2) ));

            usMinIdx = Asura_Maths::Min( usMinIdx, usIdx0 );
            usMaxIdx = Asura_Maths::Max( usMaxIdx, usIdx0 );
            usMinIdx = Asura_Maths::Min( usMinIdx, usIdx1 );
            usMaxIdx = Asura_Maths::Max( usMaxIdx, usIdx1 );

            // Reset winding as necessary.
            bFlipWinding = bNeedToForceFlipNextRestart;
            bNeedToForceFlipNextRestart = false;
        }

        usMinIdx = Asura_Maths::Min( usMinIdx, usIdx2 );
        usMaxIdx = Asura_Maths::Max( usMaxIdx, usIdx2 );

        DEBUG_TRACE_INSANE(( "cur tri: %6u %6u %6u", static_cast<u_int>(usIdx0), static_cast<u_int>(usIdx1), static_cast<u_int>(usIdx2) ));

        // Don't write out degenerates. They will be present to fix up the
        // winding after joining (real) strips.
        if( usIdx0 != usIdx1 && usIdx0 != usIdx2 && usIdx1 != usIdx2 )
        {
            *pusOut++ = usIdx0;

            DEBUG_TRACE_INSANE(( " (not degenerate; %sflipping)", (bFlipWinding ? "" : "not ") ));

            if( bFlipWinding )
            {
                *pusOut++ = usIdx2;
                *pusOut++ = usIdx1;
            }
            else
            {
                *pusOut++ = usIdx1;
                *pusOut++ = usIdx2;
            }
        }
        else
        {
            DEBUG_TRACE_INSANE(( " (degenerate)" ));
        }

        DEBUG_TRACE_INSANE(( "\n" ));

        bFlipWinding = !bFlipWinding;

        // End of strip?
        if( !uNumIndicesLeft )
        {
            // Keep some strip context for next time.
            pusEndIndices[0] = usIdx1;
            pusEndIndices[1] = usIdx2;
            uNumEndIndices = 2;
            bNeedToFlipFirstTriNextBatch = bFlipWinding;

            break;
        }

        usIdx0 = usIdx1;
        usIdx1 = usIdx2;
        usIdx2 = *pusIn++;

        --uNumIndicesLeft;

        DEBUG_TRACE_INSANE(( "new tri: %6u %6u %6u (%u indices remaining, pusIn=0x%X, pusOut=0x%X)\n",
                             static_cast<u_int>(usIdx0), static_cast<u_int>(usIdx1), static_cast<u_int>(usIdx2), uNumIndicesLeft, reinterpret_cast<u_int>(pusIn), reinterpret_cast<u_int>(pusOut) ));
    }

    *pusMinIdx             = usMinIdx;
    *pusMaxIdx             = usMaxIdx;
    *puNumEndIndices       = uNumEndIndices;
    *pbForceFlipOnFirstTri = bNeedToFlipFirstTriNextBatch;

    u_int uNumOutputIndices = pusOut - pusOutStart;

    DEBUG_TRACE_INSANE(( "\nTRI LIST VERSION:\n" ));
    for( u_int u = 0; u < uNumOutputIndices / 3; ++u )
    {
        DEBUG_TRACE_INSANE(( "%6u %6u %6u\n", pusOutStart[u * 3], pusOutStart[u * 3 + 1], pusOutStart[u * 3 + 2] ));
    }

    return uNumOutputIndices;
}

int main( uint32_t uSPUNum, uint64_t uJobQueueAddr, uint64_t uViewportInfoAddr )
{
    DEBUG_TRACE_SUMMARY(( "[%u] GeomCull_Env_SPU started\n", uSPUNum ));

    PREPARE_STACK_CHECK();

    // Calculate available memory.
    u_int uEndOfExecutable = (reinterpret_cast<u_int>(&_end) + 15) & ~15;
    u_int uTotalMemoryAvailable = 256 * 1024 - (uEndOfExecutable + uSTACK_SIZE_TO_ALLOW + STACK_CHECK_SIZE);
    void* pvStartOfAvailableMemory = reinterpret_cast<void*>( uEndOfExecutable );

    DEBUG_TRACE_SUMMARY(( "[%u] memory available: %.1fKB\n", uSPUNum, static_cast<float>(uTotalMemoryAvailable) / 1024.0f ));

    enum
    {
        DMATAG_GET_VIEWPORT_INFO,
        DMATAG_GET_JOB,
        DMATAG_GET_VTX_BFR,
        DMATAG_GET_IDX_BFR,

        DMATAG_PUT_IDX_BFR,
        DMATAG_PUT_IDX_COUNT,

        DMATAG_GET_CAPE,
        DMATAG_WEAR_CAPE,
        DMATAG_FLY,
    };

    cellDmaGet( g_acViewportInfoBuffer, uViewportInfoAddr, sizeof(g_acViewportInfoBuffer), DMATAG_GET_VIEWPORT_INFO, uDMA_TID, uDMA_RID );

    CellSyncLFQueuePopContainer xQueuePopContainer;
    cellSyncLFQueuePopContainerInitialize( &xQueuePopContainer, g_acNextJobBuffer, DMATAG_GET_JOB );

    DO_STACK_CHECK( 1 );

    while( true )
    {
        g_xAllocator.Initialise( pvStartOfAvailableMemory, uTotalMemoryAvailable, false );

        cellSyncLFQueuePopBegin( uJobQueueAddr, &xQueuePopContainer );
        cellSyncLFQueuePopEnd( uJobQueueAddr, &xQueuePopContainer );
		cellDmaWaitTagStatusAll( (1 << DMATAG_GET_VIEWPORT_INFO) | (1 << DMATAG_GET_JOB) );

        if( g_xNextJob.uBFCPolicyOrTerminator == Asura_PS3_EnvironmentRenderer_SPUJob::TERMINATOR )
        {
            DEBUG_TRACE_SUMMARY(( "[%u] got terminator (debug: 0x%X)\n", uSPUNum, g_xNextJob.uDebug ));
            break;
        }

        u_int uTotalInputIdxCount = g_xNextJob.uInputIdxCount;
        u_int uTotalInputVtxCount = g_xNextJob.uInputVtxCount;

        DEBUG_TRACE_SUMMARY(( "[%u] got job (%u i, %u v; out count addr: 0x%08X; bfcpol=%u; debug=0x%X)\n",
                              uSPUNum, uTotalInputIdxCount, uTotalInputVtxCount, g_xNextJob.puOutputIdxCount, g_xNextJob.uBFCPolicyOrTerminator, g_xNextJob.uDebug ));

        u_int uCullingFlavour;
        switch( g_xNextJob.uBFCPolicyOrTerminator )
        {
            default:
            case Asura_PS3_EnvironmentRenderer_SPUJob::BFC_NONE:
                uCullingFlavour = EDGE_GEOM_CULL_FRUSTUM;
                break;

            case Asura_PS3_EnvironmentRenderer_SPUJob::BFC_BACK:
                uCullingFlavour = EDGE_GEOM_CULL_BACKFACES_AND_FRUSTUM;
                break;

            case Asura_PS3_EnvironmentRenderer_SPUJob::BFC_FRONT:
                uCullingFlavour = EDGE_GEOM_CULL_FRONTFACES_AND_FRUSTUM;
                break;
        }

//TEMP
//TEMP
if( g_xNextJob.uDebug & 0x10000000 )
{
    uCullingFlavour = EDGE_GEOM_CULL_NONE;
}
//TEMP
//TEMP

        // Calculate the 'scratch buffer' size for the whole job. This is
        // actually something tied to the SPURS memory layouts and
        // specifically here just needs to be large enough for Edge to store
        // its uniform tables in there. This is 1 QW per vertex per table but
        // we only have 1 table. To avoid having to allocate a buffer to
        // transfer in the 3-vector input vertices and then separately have
        // this other buffer that just gets the 4-vector versions we'll cheat
        // and transfer the former in at the end of the scratch buffer and
        // convert almost in place. This means we only need 16 bytes per
        // input vertex instead of 12+16.
        u_int uScratchBfrSize = ((uTotalInputVtxCount + 3) & ~3) << 4; // Edge decompresses 4 verts at a time so padding is required.
        uScratchBfrSize += 16; // Extra padding to ensure we can align input transfer without risk of overwrite.

        // Similarly, calculate the index buffer size. We have to assume
        // the worst case in terms of the input strip - ie. the maximum
        // number of tris per input index, ie. a perfect strip - to avoid
        // overlap in the worst cases.
        u_int uInputIndexBfrSize = ((uTotalInputIdxCount - 2) * 3 * 2 + 15) & ~15;
        // Also allow some extra slop for internal transfer alignments (when
        // we put the input strip at the end, etc.)
        uInputIndexBfrSize += 32;

        u_int uBytesAvailableForBfrs = g_xAllocator.GetNumFreeBytes() & ~15;
        uBytesAvailableForBfrs -= 64; // Account for alignment between buffers.

        DO_STACK_CHECK( 2 );

        // Figure out batch size.
        u_int uMaxInputIndicesPerBatch;
#if !FIXED_MAX_BATCH_SIZE
        if( uScratchBfrSize + uInputIndexBfrSize <= uBytesAvailableForBfrs )
        {
            // Whole job will fit. The buffers can stay those sizes.
            uMaxInputIndicesPerBatch = uTotalInputIdxCount;
        }
        else
#endif
        {
            // Need to split into batches. As with the full strip case, we
            // need to allow for perfect strips when avoiding overlap. Since
            // no index will reference more than one vertex we can assume
            // for I input strip indices the memory requirement will be
            //
            //     (I - 2) * 3 * 2   +    I * 16
            //          [indices]       [vertices]
            //
            // ie. <= I * 22 bytes
            //
            // Though generally the index range should be quite controlled -
            // ie. 1000 indices will have a range of values of around 1000 -
            // we also need to allow some slop in the vertex buffer size for
            // when this isn't quite the case.
            //
            // Finally we'll need to allow for alignment within the index
            // buffer (for putting the input strip data at the end) and to
            // allow for output padding within batches (for alignment
            // purposes).

            const u_int uBytesPerInputIndex = 22; // See above.

            const u_int uIndexRangeSlop = 128; // +/- 64
            u_int uBytesForIndexRangeSlop = uIndexRangeSlop << 4;

            u_int uIndexBfrPadBytes = 64; // Up to 32 bytes for alignment stuff, up to 64 bytes for output padding.

#if FIXED_MAX_BATCH_SIZE
            uMaxInputIndicesPerBatch = FIXED_MAX_BATCH_SIZE;
#else
            uMaxInputIndicesPerBatch = (uBytesAvailableForBfrs - (uBytesForIndexRangeSlop + uIndexBfrPadBytes)) / uBytesPerInputIndex;
#endif

            // Align that to a multiple of 8 so input transfer size is 16-
            // byte aligned for both indices (8 * 2) and vertices (8 * 12).
            uMaxInputIndicesPerBatch &= ~7;

            uInputIndexBfrSize = ((uMaxInputIndicesPerBatch - 2) * 3 * 2 + 15) & ~15;
            uInputIndexBfrSize += uIndexBfrPadBytes;

            // Recalculate scratch buffer size now - it's just the rest.
            uScratchBfrSize = uBytesAvailableForBfrs - uInputIndexBfrSize;
        }

        u_short* const pusIdxBfr = static_cast<u_short*>( g_xAllocator.Allocate(uInputIndexBfrSize, 16) );
        void* const pvScratchBfr = g_xAllocator.Allocate( uScratchBfrSize, 16 );

        DEBUG_TRACE_DETAILED(( "[%u] idx bfr: 0x%X (%i bytes)\n[%u] scratch bfr: 0x%X (%u bytes)\n", uSPUNum, pusIdxBfr, uInputIndexBfrSize, uSPUNum, pvScratchBfr, uScratchBfrSize ));

        u_int uRemainingInputIndices = uTotalInputIdxCount;
        uint64_t uInputIdxSrcAddr = reinterpret_cast<uint64_t>(g_xNextJob.pvInputIdxBfr);
        uint64_t uOutputIdxDstAddr = reinterpret_cast<uint64_t>(g_xNextJob.pvOutputIdxBfr);
        u_int uTotalOutputIndices = 0;

        DO_STACK_CHECK( 3 );

        DEBUG_TRACE_DETAILED(( "[%u] max input indices per batch: %u%s\n", uSPUNum, uMaxInputIndicesPerBatch, (uMaxInputIndicesPerBatch == uTotalInputIdxCount ? " (whole job)" : "") ));

        // Some state for resuming strip conversion in batches after the
        // first.
        u_short ausPrevBatchEndIndices[2] = { 0, 0 }; // Note that these will not be offset.
        u_int uNumEndIndicesFromPrevBatch = 0;
        bool bForceFlipOnFirstTri = false;

        // Process indices in batches.
        while( true )
        {
            u_int uNumInputIndicesThisBatch = Asura_Maths::Min( uRemainingInputIndices, uMaxInputIndicesPerBatch );
            u_int uInputIdxBfrTransferSize = (uNumInputIndicesThisBatch * sizeof(u_short) + 15) & ~15;
            DEBUG_TRACE_DETAILED(( "[%u] batch of %u indices (%u remaining)\n", uSPUNum, uNumInputIndicesThisBatch, uRemainingInputIndices - uNumInputIndicesThisBatch ));

            // Put input indices at the end so we can convert in-place.
            u_short* pusInputIdxDst = reinterpret_cast<u_short*>( ((reinterpret_cast<u_int>(pusIdxBfr) + uInputIndexBfrSize) - uInputIdxBfrTransferSize) & ~15 );

            // Fetch indices. Wait for previous index output to complete -
            // we're probably about to overwrite at least some of it.
            cellDmaWaitTagStatusAll( 1 << DMATAG_PUT_IDX_BFR );
            DEBUG_TRACE_DETAILED(( "[%u] fetching %u bytes of indices (0x%X -> 0x%X)\n", uSPUNum, uInputIdxBfrTransferSize, static_cast<u_int>(uInputIdxSrcAddr), pusInputIdxDst ));
            cellDmaLargeGet( pusInputIdxDst, uInputIdxSrcAddr, uInputIdxBfrTransferSize, DMATAG_GET_IDX_BFR, uDMA_TID, uDMA_RID );

            DEBUG_TRACE_DETAILED(( "[%u] converting %u strip indices (inc %u left over)\n", uSPUNum, uNumInputIndicesThisBatch, uNumEndIndicesFromPrevBatch ));

            // Convert to tri list.
            cellDmaWaitTagStatusAll( 1 << DMATAG_GET_IDX_BFR );
            u_short usMinIdx, usMaxIdx;
            u_int uNumConvertedIndices = ConvertToTriList( pusInputIdxDst, uNumInputIndicesThisBatch,
                                                           static_cast<u_short*>(pusIdxBfr),
                                                           &usMinIdx, &usMaxIdx,
                                                           ausPrevBatchEndIndices, &uNumEndIndicesFromPrevBatch, &bForceFlipOnFirstTri );
            DEBUG_TRACE_DETAILED(( "[%u] got %u indices (%u tris)\n", uSPUNum, uNumConvertedIndices, uNumConvertedIndices / 3 ));

            // Need to align vertex transfer to a multiple of 48 bytes (so 16
            // byte aligned for DMA and a multiple of the vertex size), ie. a
            // multiple of 4 vertices.
            u_int uIndexOffset = usMinIdx & ~3;
            u_int uNumVerticesToTransfer = (((usMaxIdx + 1) - uIndexOffset) + 3) & ~3;
            DEBUG_TRACE_DETAILED(( "[%u] aligned index range: [%u, %u)\n",
                                   uSPUNum, uIndexOffset, uIndexOffset + uNumVerticesToTransfer ));

#if FIXED_MAX_BATCH_SIZE
            Asura_Assert( uNumVerticesToTransfer <= 4096, "LIMIT" );
#endif

            // Double-check the indices weren't outside the range we planned
            // for and so that the padded 4-vector versions will fit. Abort
            // if not because things are going to go horribly wrong anyway in
            // that case.
            if( uNumVerticesToTransfer * 4 * sizeof(float) > uScratchBfrSize )
            {
                // Need to increase the index range slop in this case,
                // most likely.
                spu_printf( "[%u] SPU ERROR: Vertex range too large - skipping! (range is %u after alignment)\n", uSPUNum, uNumVerticesToTransfer );

                // Tell the PPU to skip this strip.
                g_xOutputIndexCount = spu_splats( 0u );
                void* pvSrcAddr = reinterpret_cast<void*>( reinterpret_cast<u_int>(&g_xOutputIndexCount) + (reinterpret_cast<u_int>(g_xNextJob.puOutputIdxCount) & 0xF) );
                cellDmaSmallPut( pvSrcAddr, reinterpret_cast<uint64_t>(g_xNextJob.puOutputIdxCount), sizeof(u_int), DMATAG_PUT_IDX_COUNT, uDMA_TID, uDMA_RID );

                break;
            }

            // Calculate offset input address and transfer size based on the
            // index range we've determined.
            uint64_t uInputVtxSrcAddr = reinterpret_cast<uint64_t>( g_xNextJob.pvInputVtxBfr );
            uInputVtxSrcAddr += uIndexOffset * 12;
            u_int uInputVtxBfrTransferSize = (uNumVerticesToTransfer * 3 * sizeof(float) + 15) & ~15;

            // Determine where the vertices are going - they need to be put
            // at the end of the available buffer so that we can convert in
            // place.
            u_int uInputVtxDstAddr = reinterpret_cast<u_int>(pvScratchBfr) + uScratchBfrSize;
            uInputVtxDstAddr -= uInputVtxBfrTransferSize;
            uInputVtxDstAddr &= ~15; // Round down - we padded above to make sure this is safe.
            void* const pvInputVtxDst = reinterpret_cast<void*>(uInputVtxDstAddr);

            DO_STACK_CHECK( 4 );

            // Fetch vertices.
            DEBUG_TRACE_DETAILED(( "[%u] fetching %u bytes of vertices (0x%X -> 0x%X)\n",
                                   uSPUNum, uInputVtxBfrTransferSize, static_cast<u_int>(uInputVtxSrcAddr), pvInputVtxDst ));
            cellDmaLargeGet( pvInputVtxDst, uInputVtxSrcAddr, uInputVtxBfrTransferSize, DMATAG_GET_VTX_BFR, uDMA_TID, uDMA_RID );

            DO_STACK_CHECK( 5 );

            // Offset indices for this vertex buffer subset.
            if( uIndexOffset )
            {
                for( u_int u = 0; u < uNumConvertedIndices; ++u )
                {
                    pusIdxBfr[u] -= uIndexOffset;
                }
            }

            // Setup basic config data for Edge.
            // NB: The uniform table count is 0 here because it's calculated as
            // (num attributes - 1). You're supposed to add 1 if you're culling
            // but not if you specify the 'fast path' flag since that then
            // transforms and culls in-place.
            EdgeGeomSpuConfigInfo xEdgeCfg;
            xEdgeCfg.flagsAndUniformTableCount = 0 | EDGE_GEOM_FLAG_STATIC_GEOMETRY_FAST_PATH;
            xEdgeCfg.commandBufferHoleSize = 0; // Irrelevant.
            xEdgeCfg.inputFlavor = EDGE_GEOM_INPUT_FLAVOR_F32c3;
            xEdgeCfg.outputFlavor = EDGE_GEOM_OUTPUT_FLAVOR_F32c3;
            xEdgeCfg.shapeFlavor = 0; // Irrelevant.
            xEdgeCfg.indexesFlavorAndSkinningFlavor = EDGE_GEOM_INDEXES_U16_TRIANGLE_LIST_CCW << 4;
            xEdgeCfg.numVertexes = uNumVerticesToTransfer;
            xEdgeCfg.numIndexes = uNumConvertedIndices;
            xEdgeCfg.indexesOffset = 0; // Irrelevant.

            DO_STACK_CHECK( 6 );

            EdgeGeomLocalToWorldMatrix xLocalToWorldMat =
            {
                { 1.0f, 0.0f, 0.0f, 0.0f,
                  0.0f, 1.0f, 0.0f, 0.0f,
                  0.0f, 0.0f, 1.0f, 0.0f, }
            };

            DO_STACK_CHECK( 7 );

            // Initialise the Edge library.
            EdgeGeomSpuContext xEdgeContext;
            edgeGeomInitialize( &xEdgeContext,
                                &xEdgeCfg,
                                pvScratchBfr, uScratchBfrSize,
                                pvStartOfAvailableMemory, uTotalMemoryAvailable,
                                0, // Irrelevant - we'll be dealing with all transfers.
                                &g_xViewportInfo,
                                &xLocalToWorldMat,
                                NULL );

            DO_STACK_CHECK( 8 );

            // Decompress vertices, ie. pad 3-vectors to 4-vectors, in-place,
            // or fake it if this not the first batch.
            cellDmaWaitTagStatusAll( 1 << DMATAG_GET_VTX_BFR );
            DEBUG_TRACE_DETAILED(( "[%u] decompressing vertices (flavour %u)\n", uSPUNum, xEdgeCfg.inputFlavor ));
            edgeGeomDecompressVertexes( &xEdgeContext, pvInputVtxDst, NULL );

            DO_STACK_CHECK( 9 );

            // Generate modified index buffer in-place once they've finished
            // transferring in.
            edgeGeomDecompressIndexes( &xEdgeContext, pusIdxBfr ); // This just sets up an internal ptr - no decompression is required.
            DEBUG_TRACE_DETAILED(( "[%u] culling (flavour %u)\n", uSPUNum, uCullingFlavour ));
            u_int uOutputIndexCount = edgeGeomCullTriangles( &xEdgeContext, uCullingFlavour );

            // Fix up indices again.
            if( uIndexOffset )
            {
                for( u_int u = 0; u < uOutputIndexCount; ++u )
                {
                    pusIdxBfr[u] += uIndexOffset;
                }
            }

            uRemainingInputIndices -= uNumInputIndicesThisBatch;

            // If there are more indices to process then we need to make sure
            // that the output is qw-aligned, since otherwise any DMA transfer
            // out for the next batch will then be problematic.
            u_int uPaddedOutputIndexCount = uOutputIndexCount;
            if( uRemainingInputIndices && (uOutputIndexCount & 7) )
            {
                u_short* pusOut = pusIdxBfr + uOutputIndexCount;

                DEBUG_TRACE_DETAILED(( "[%u] %u output indices (%u tris) - padding required\n", uSPUNum, uOutputIndexCount, uOutputIndexCount / 3 ));

                // Pad with degenerate triangles until we hit alignment.
                // Note that this should output at most 21 indices and we
                // padded for this in buffer size calculations above.
                u_short usPadIdx = *(pusOut - 1); // TODO: Can we use restart index here instead?
                while( uPaddedOutputIndexCount & 7 )
                {
                    *pusOut++ = usPadIdx;
                    *pusOut++ = usPadIdx;
                    *pusOut++ = usPadIdx;

                    uPaddedOutputIndexCount += 3;
                }
            }

            uTotalOutputIndices += uPaddedOutputIndexCount;

            // Copy modified indices back to main RAM.
            u_int uOutputIdxBfrTransferSize = (uPaddedOutputIndexCount * sizeof(u_short) + 15) & ~15;
            DEBUG_TRACE_DETAILED(( "[%u] outputting %u indices (%u bytes) to 0x%X\n", uSPUNum, uPaddedOutputIndexCount, uOutputIdxBfrTransferSize, static_cast<u_int>(uOutputIdxDstAddr) ));
            cellDmaLargePut( pusIdxBfr, uOutputIdxDstAddr, uOutputIdxBfrTransferSize, DMATAG_PUT_IDX_BFR, uDMA_TID, uDMA_RID );

            if( uRemainingInputIndices )
            {
                uInputIdxSrcAddr += uInputIdxBfrTransferSize;
                uOutputIdxDstAddr += uOutputIdxBfrTransferSize;
            }
            else
            {
                break;
            }
        }

        // Output the final index count.
        cellDmaWaitTagStatusAll( 1 << DMATAG_PUT_IDX_COUNT );
        g_xOutputIndexCount = spu_splats( uTotalOutputIndices );
        void* pvSrcAddr = reinterpret_cast<void*>( reinterpret_cast<u_int>(&g_xOutputIndexCount) + (reinterpret_cast<u_int>(g_xNextJob.puOutputIdxCount) & 0xF) );
        cellDmaSmallPut( pvSrcAddr, reinterpret_cast<uint64_t>(g_xNextJob.puOutputIdxCount), sizeof(u_int), DMATAG_PUT_IDX_COUNT, uDMA_TID, uDMA_RID );
    }

    cellDmaWaitTagStatusAll( 0xFFFFFFFF );
    sys_spu_thread_exit( EXIT_SUCCESS );
}
