//
// Asura_Physics_Solver_SPU_Main.cpp
//
// Created by KJFL - 9 July 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//


// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <cell/dma.h>
#include <cell/sync.h>
#include <sys/spu_thread.h>
#include <spu_intrinsics.h>
#include <spu_printf.h>
#include <string.h>

// Engine independent.
#include "Core/Memory/Asura_Mem_ResetOnlyAllocator.h"
#include "Core/Asura_Maths.h"
#include "Physics/Asura_Physics_BodyState.h"
#include "Physics/Asura_Physics_Solver.h"

// Engine platform-specific.
//#include "Debugging/Asura_Assert.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// G L O B A L S /////////////////////////////////////////////////////////////

static const u_int uSTACK_SIZE_TO_ALLOW = 4 * 1024; // Needs to be a multiple of 16.

extern u_int _end; // End of executable (plus static data, etc.)



#define DEBUG_OUTPUT_LEVEL (0)

#if DEBUG_OUTPUT_LEVEL == 3
    #define DEBUG_TRACE_INSANE( args ) \
        { spu_printf args ; }
    #define DEBUG_TRACE_DETAILED( args ) \
        { spu_printf args ; }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { spu_printf args ; }
#elif DEBUG_OUTPUT_LEVEL == 2
    #define DEBUG_TRACE_INSANE( args ) \
        { }
    #define DEBUG_TRACE_DETAILED( args ) \
        { spu_printf args ; }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { spu_printf args ; }
#elif DEBUG_OUTPUT_LEVEL == 1
    #define DEBUG_TRACE_INSANE( args ) \
        { }
    #define DEBUG_TRACE_DETAILED( args ) \
        { }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { spu_printf args ; }
#else
    #define DEBUG_TRACE_INSANE( args ) \
        { }
    #define DEBUG_TRACE_DETAILED( args ) \
        { }
    #define DEBUG_TRACE_SUMMARY( args ) \
        { }
#endif

static union
{
    Asura_Physics_Solver_Data g_xNextJob;
    char g_acNextJobBuffer[(sizeof(Asura_Physics_Solver_Data) + 15) & ~15];
} __attribute__(( aligned(16) ));

static Asura_Mem_ResetOnlyAllocator g_xAllocator;

// These have to be 0 under the current SDK.
static const u_int uDMA_TID = 0;
static const u_int uDMA_RID = 0;

static char acDataBuffer[192*1024] __attribute__((aligned(128)));

static Asura_AMatrix3x3 axOrientation[256];
static Asura_AVector3	axPosition[256];

#define STACK_CHECK_SIZE (0)

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

int main( uint32_t uSPUNum, uint64_t uJobQueueAddr, uint64_t uSPUBarrierAddr, uint64_t uPPUBarrierAddr )
{
    DEBUG_TRACE_SUMMARY(( "[%u] Physics_Solver_SPU\n", uSPUNum ));
    DEBUG_TRACE_DETAILED(( "[%u] job q @ 0x%X, spu barrier @ 0x%X, ppu barrier @ 0x%X\n", uSPUNum, static_cast<uint32_t>(uJobQueueAddr), static_cast<uint32_t>(uSPUBarrierAddr), static_cast<uint32_t>(uPPUBarrierAddr) ));

    // Calculate available memory.
    u_int uEndOfExecutable = (reinterpret_cast<u_int>(&_end) + 15) & ~15;
    u_int uTotalMemoryAvailable = 256 * 1024 - (uEndOfExecutable + uSTACK_SIZE_TO_ALLOW + STACK_CHECK_SIZE);
    void* pvStartOfAvailableMemory = reinterpret_cast<void*>( uEndOfExecutable );

    DEBUG_TRACE_SUMMARY(( "[%u] mem available: %.1fKB\n", uSPUNum, static_cast<float>(uTotalMemoryAvailable) / 1024.0f ));

    enum
    {
        DMATAG_GET_JOB,
		DMA_TAG_DATA_IN,
		DMA_TAG_DATA_IN2,
		DMA_TAG_DATA_OUT,
    };

    CellSyncLFQueuePopContainer xQueuePopContainer;
    cellSyncLFQueuePopContainerInitialize( &xQueuePopContainer, g_acNextJobBuffer, DMATAG_GET_JOB );

    while( true )
    {
        while( true )
        {
            cellSyncLFQueuePopBegin( uJobQueueAddr, &xQueuePopContainer );
            cellSyncLFQueuePopEnd( uJobQueueAddr, &xQueuePopContainer );
		    cellDmaWaitTagStatusAll( (1 << DMATAG_GET_JOB) );

            if( g_xNextJob.m_uNumberOfSolverBodies == 0 )
            {
                break;
            }
		    else
		    {
                DEBUG_TRACE_SUMMARY(( "[%u] %ub task: %u bodies, %u contacts, %u pairing contacts\n", uSPUNum, g_xNextJob.m_uDataSizeInBytes, g_xNextJob.m_uNumberOfSolverBodies, g_xNextJob.m_uNumberOfSingleBodyContacts,
				    g_xNextJob.m_uNumberOfPairingContacts));
		    }

		    cellDmaWaitTagStatusAll( (1<<DMA_TAG_DATA_OUT) ); 
		    cellDmaLargeGet( acDataBuffer, g_xNextJob.m_uDataAddress, g_xNextJob.m_uDataSizeInBytes, DMA_TAG_DATA_IN, uDMA_TID, uDMA_RID );
    	
		    u_int uNumberOfSolverBodies = g_xNextJob.m_uNumberOfSolverBodies;
		    u_int uNumberOfPairingContacts = g_xNextJob.m_uNumberOfPairingContacts;
		    u_int uNumberOfSingleBodyContacts = g_xNextJob.m_uNumberOfSingleBodyContacts;
   			u_int uNumberOfConeTwistConstraints = g_xNextJob.m_uNumberOfConeTwistConstraints;
	
		    u_int uBodyDataSize = uNumberOfSolverBodies * sizeof( Asura_Physics_Solver_Body);
		    u_int uPairingContactsDataSize = uNumberOfPairingContacts * sizeof( Asura_Physics_Solver_PairingContact);
    		u_int uSingleContactDataSize = uNumberOfSingleBodyContacts * sizeof( Asura_Physics_Solver_SingleBodyContact);
	
		    char* pcBufferStart = acDataBuffer;
		    Asura_Physics_Solver_Body* pxBodies = reinterpret_cast<Asura_Physics_Solver_Body*>( pcBufferStart );
		    Asura_Physics_Solver_PairingContact* pxPairingContacts = reinterpret_cast<Asura_Physics_Solver_PairingContact*>( pcBufferStart + uBodyDataSize );
		    Asura_Physics_Solver_SingleBodyContact* pxSingleBodyContacts = reinterpret_cast<Asura_Physics_Solver_SingleBodyContact*>( pcBufferStart + uPairingContactsDataSize + uBodyDataSize );
			Asura_Physics_Solver_ConeTwistConstraint* pxConeTwistConstraints = reinterpret_cast<Asura_Physics_Solver_ConeTwistConstraint*>(pcBufferStart + uPairingContactsDataSize + uBodyDataSize + uSingleContactDataSize );

		    DEBUG_TRACE_INSANE(( "Asura_Physics_Solver_Body = %d bytes\n", sizeof(Asura_Physics_Solver_Body) ));
		    DEBUG_TRACE_INSANE(( "Asura_Physics_Solver_PairingContact = %u bytes\n", sizeof(Asura_Physics_Solver_PairingContact) ));
		    DEBUG_TRACE_INSANE(( "Asura_Physics_Solver_SingleBodyContact = %u bytes\n", sizeof(Asura_Physics_Solver_SingleBodyContact) ));
    		
		    DEBUG_TRACE_INSANE(( "Asura_AVector3 = %u bytes\n", sizeof(Asura_AVector3) ));
		    DEBUG_TRACE_INSANE(( "Asura_AVector3 = %u bytes\n", sizeof(Asura_AMatrix3x3) ));

		    cellDmaWaitTagStatusAll( (1<<DMA_TAG_DATA_IN) ); 

		    // DMA in the orientation and centre of mass data now - we won't need it till later so
		    // we don't have to block on it
		    for (u_int u=0; u<uNumberOfSolverBodies; ++u)
		    {
			    Asura_Physics_BodyState* pxBody = pxBodies[u].m_pxParentBody;

			    u_int uOrientation = (u_int)(&pxBody->m_xOrientation);
			    u_int uCentreOfMass = (u_int)(&pxBody->m_xCentreOfMass);
    	
			    cellDmaGet( &(axOrientation[u]), uOrientation, sizeof(Asura_AMatrix3x3), DMA_TAG_DATA_IN2, uDMA_TID, uDMA_RID );
			    cellDmaGet( &(axPosition[u]), uCentreOfMass, sizeof(Asura_AVector3), DMA_TAG_DATA_IN2, uDMA_TID, uDMA_RID );
		    }

		    u_int uNumberOfIterations = 20;
    		
			// Fix up the body pointers to be aimed at SPU memory rather than PPU memory.
			// Also precalculate anything we can prior to the iterations taking place.
		    for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
		    {
			    Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];

			    xContact.m_pxBody = pxBodies + (((u_int)xContact.m_pxBody - g_xNextJob.m_uDataAddress)/sizeof(Asura_Physics_Solver_Body));

		    }
		    for (u_int u=0; u<uNumberOfPairingContacts; ++u)
		    {
			    Asura_Physics_Solver_PairingContact& xContact = pxPairingContacts[u];

			    xContact.m_pxBodyA = pxBodies + (((u_int)xContact.m_pxBodyA - g_xNextJob.m_uDataAddress)/sizeof(Asura_Physics_Solver_Body));
			    xContact.m_pxBodyB = pxBodies + (((u_int)xContact.m_pxBodyB - g_xNextJob.m_uDataAddress)/sizeof(Asura_Physics_Solver_Body));

		    }
			for ( u_int u=0; u<uNumberOfConeTwistConstraints; ++u )
			{
				Asura_Physics_Solver_ConeTwistConstraint& xConstraint = pxConeTwistConstraints[u];

				xConstraint.m_pxBodyA = pxBodies + (((u_int)xConstraint.m_pxBodyA - g_xNextJob.m_uDataAddress)/sizeof(Asura_Physics_Solver_Body));
				xConstraint.m_pxBodyB = pxBodies + (((u_int)xConstraint.m_pxBodyB - g_xNextJob.m_uDataAddress)/sizeof(Asura_Physics_Solver_Body));

				xConstraint.CalculateJacobians();
			}

			// Use the clamped contacts when the cone-twists are present (which currently is when it's a ragdoll)
			// KJFL - TODO - this is a trifle inelegant.
			bool bUseClampedContacts = (uNumberOfConeTwistConstraints > 0);

		    for (u_int uIterations = 0; uIterations<uNumberOfIterations; ++uIterations)
		    {
				// do all cone twist joints
				for ( u_int u=0; u<uNumberOfConeTwistConstraints; ++u )
				{
					Asura_Physics_Solver_ConeTwistConstraint::Solve( pxConeTwistConstraints[u] );
				}
	
			    // do all pairing contacts
			    for (u_int u=0; u<uNumberOfPairingContacts; ++u)
			    {
				    Asura_Physics_Solver_PairingContact& xContact = pxPairingContacts[u];

				    Asura_Physics_Solver_PairingContact::Solve_LowPenetration( xContact );
			    }
    			
				if ( bUseClampedContacts )
				{
					// and then all vs environment contacts
					for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
					{
						Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];

						Asura_Physics_Solver_SingleBodyContact::Solve_Clamped( xContact );
					} 
				}
				else
				{
					// and then all vs environment contacts
					for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
					{
						Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];

						Asura_Physics_Solver_SingleBodyContact::Solve_LowPenetration( xContact );
					} 
				}

			    // do all pairing contacts
			    for (u_int u=0; u<uNumberOfPairingContacts; ++u)
			    {
				    Asura_Physics_Solver_PairingContact& xContact = pxPairingContacts[u];
				    if (xContact.m_fPenetration >= fHIGH_PENETRATION_THRESHOLD )
				    {
					    Asura_Physics_Solver_PairingContact::Solve_HighPenetration( xContact );
				    }
				}

				if ( !bUseClampedContacts )
				{
					// and then all vs environment contacts
					for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
					{
						Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];
						if (xContact.m_fPenetration >= fHIGH_PENETRATION_THRESHOLD )
						{
							Asura_Physics_Solver_SingleBodyContact::Solve_HighPenetration( xContact );
						}

					} 
				}
    			
		    }

		    cellDmaWaitTagStatusAll( (1<<DMA_TAG_DATA_IN2) ); 

		    for (u_int u=0; u<uNumberOfSolverBodies; ++u)
		    {
				// Skip any infinite mass objects
				if ( pxBodies[u].m_fInvMass == 0.0f ) continue;

			    Asura_Physics_BodyState* pxBody = pxBodies[u].m_pxParentBody;

				u_int uLinearVelocity = (u_int)(&pxBody->m_xLinearVelocity);
			    u_int uAngularVelocity = (u_int)(&pxBody->m_xAngularVelocity);
			    u_int uOrientation = (u_int)(&pxBody->m_xOrientation);
			    u_int uCentreOfMass = (u_int)(&pxBody->m_xCentreOfMass);
    	
			    cellDmaPut( &(pxBodies[u].m_xLinearVelocity), uLinearVelocity, sizeof(Asura_AVector3), DMA_TAG_DATA_OUT, uDMA_TID, uDMA_RID );
			    cellDmaPut( &(pxBodies[u].m_xAngularVelocity), uAngularVelocity, sizeof(Asura_AVector3), DMA_TAG_DATA_OUT, uDMA_TID, uDMA_RID );

			    Asura_AVector3& xCentreOfMass = axPosition[u];
			    Asura_AMatrix3x3& xOrientation = axOrientation[u];
    			
			    Asura_AMatrix3x3 xDelta;
			    xDelta.Tilde( pxBodies[u].m_xSeparationAngularVelocity );		
			    xDelta = xOrientation*xDelta;
    	
			    xOrientation.Add( xDelta );
			    xOrientation.Orthogonalise();
			    xCentreOfMass += pxBodies[u].m_xSeparationLinearVelocity;

			    cellDmaPut( &xOrientation, uOrientation, sizeof(Asura_AMatrix3x3), DMA_TAG_DATA_OUT, uDMA_TID, uDMA_RID );
			    cellDmaPut( &xCentreOfMass, uCentreOfMass, sizeof(Asura_AVector3), DMA_TAG_DATA_OUT, uDMA_TID, uDMA_RID );
		    }
        }

        DEBUG_TRACE_SUMMARY(( "[%u] terminator\n", uSPUNum ));

        // Signal that we're done.
        cellSyncBarrierNotify( uSPUBarrierAddr );

        DEBUG_TRACE_SUMMARY(( "[%u] waiting for all threads to complete\n", uSPUNum ));

        // Wait until all threads have signalled as well, so that we don't
        // go consuming their terminators.
        cellSyncBarrierWait( uSPUBarrierAddr );

        // Ensure all outgoing DMAs are complete before we notify the PPU
        // (since that will result in our thread group being suspended pretty
        // much immediately).
        cellDmaWaitTagStatusAll( 0xFFFFFFFF );

        DEBUG_TRACE_SUMMARY(( "[%u] continuing\n", uSPUNum ));

        // Signal the PPU that we're resuming our queue polling so that it
        // can go ahead and suspend the thread group.
        cellSyncBarrierNotify( uPPUBarrierAddr );
    }
}
