//////////////////////////////////////////////////////////////////////////////
//
// main.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <cell/atomic.h>
#include <cell/dma.h>
#include <cell/sync/barrier.h>
#include <cell/sync/lfqueue.h>
#include <edge/geom/edgegeom.h>
#include <sdk_version.h>
#include <sys/spu_thread.h>
#include <spu_intrinsics.h>
#include <spu_printf.h>

// Asura
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS3.h"	// for shared definition of ASURA_PS3_SPUSKINNING_BUFFER_SIZE, since character export must chop them up to fit
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"

#include "Rendering/Asura_PS3_SPU_SkinningTypes.h"

//
// Debug printfs
//
// these are very slow and affect the timing, so any bugs related to synchronisation will be affected (usually "fixed") by switching on debug output

#if (0)	// just mark the start & end
	#define PING(a) { if (a) { spu_printf("SPU %d: start\n", g_iSPUNum); } else { spu_printf("SPU %d: end\n", g_iSPUNum); } }
#else
	#define PING(a)
#endif

#if (0)	// general debug info
	#define DEBUG_PRINTF(args) spu_printf args
#else
	#define DEBUG_PRINTF(args)
#endif

#if (0) // debug task info
	#define DEBUG_DUMP_TASK()																\
	{																						\
		spu_printf("\nSPU %d: Task info:-\n", g_iSPUNum);									\
		spu_printf("  m_uInputAddr %016llx\n", g_xJobInfo.m_uInputAddr);					\
		spu_printf("  m_uVertPosOutAddr %016llx\n", g_xJobInfo.m_uVertPosOutAddr);			\
		spu_printf("  m_uNumVerts %08x\n", g_xJobInfo.m_uNumVerts);							\
		spu_printf("  m_eMode %s\n", g_xJobInfo.m_eMode ? "shadow" : "full" );				\
		spu_printf("\n");																	\
	}
#else
	#define DEBUG_DUMP_TASK()
#endif

#if (0) // DMA debug
#define DEBUG_ATOMIC( pLocal, pMain, bWrite )											\
		{																								\
		if ( bWrite )																			\
			{																						\
			spu_printf("SPU %d: Atomic via %08p to %016llx\n", g_iSPUNum, pLocal, pMain);		\
			}																						\
			else																					\
			{																						\
			spu_printf("SPU %d: Atomic via %08p from %016llx\n", g_iSPUNum, pLocal, pMain);		\
			}																						\
		}

#define DEBUG_GET( pLocal, pMain, uSize ) { spu_printf("SPU %d: DMA GET %d bytes to %08p from %016llx\n", g_iSPUNum, uSize, pLocal, pMain); Asura_Assert((((u_int)(pLocal))&0xf) == (((u_int)(pMain))&0xf), "ALIGNMENT ERROR!\n"); }
#define DEBUG_PUT( pLocal, pMain, uSize ) { spu_printf("SPU %d: DMA PUT %d bytes from %08p to %016llx\n", g_iSPUNum, uSize, pLocal, pMain); Asura_Assert((((u_int)(pLocal))&0xf) == (((u_int)(pMain))&0xf), "ALIGNMENT ERROR!\n"); }

#else
#define DEBUG_ATOMIC( pLocal, pMain, bWrite )
//still assert about alignment, SPU crashes so it's damn useful to know if this happened
#define DEBUG_GET( pLocal, pMain, uSize ) Asura_Assertf((((u_int)(pLocal))&0xf) == (((u_int)(pMain))&0xf), ("ALIGNMENT ERROR! SPU %d: DMA GET %d bytes to %08p from %016llx\n", g_iSPUNum, uSize, pLocal, pMain));
#define DEBUG_PUT( pLocal, pMain, uSize ) Asura_Assertf((((u_int)(pLocal))&0xf) == (((u_int)(pMain))&0xf), ("ALIGNMENT ERROR! SPU %d: DMA PUT %d bytes from %08p to %016llx\n", g_iSPUNum, uSize, pLocal, pMain));
#endif

#if (0)	// vertex shading debug
	static int g_iCurrentVertex;
	#define DEBUG_VERTEX_ID(var) g_iCurrentVertex = var;
	#if (1)
		//debug a specific vertex
		#define DEBUG_VERTEX 200
		#define DEBUG_VERTEX_TEST if ( g_iCurrentVertex == DEBUG_VERTEX )
	#else
		//debug all vertices
		#define DEBUG_VERTEX_TEST
	#endif

	#define DEBUG_VERTEX_SCALAR(var) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: "#var" %.2f\n", g_iSPUNum, g_iCurrentVertex, var ); }
	#define DEBUG_VERTEX_VECTOR(var) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: "#var" (%.2f, %.2f, %.2f, %.2f)\n", g_iSPUNum, g_iCurrentVertex, var.GetX(), var.GetY(), var.GetZ(), var.GetW() ); }

 	#define DEBUG_VERTEX_TYPE(format,var) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: "#var" "#format"\n", g_iSPUNum, g_iCurrentVertex, var ); }
 	#define DEBUG_VERTEX_TYPE4(format,var0,var1,var2,var3)																			\
 		DEBUG_VERTEX_TEST																											\
 		{																															\
 			spu_printf( "SPU %d: vertex %d: "#var0" "#format", "#format", "#format", "#format"\n",									\
 											g_iSPUNum, g_iCurrentVertex, var0, var1, var2, var3 );									\
 		}
 		
 	#define DEBUG_VERTEX_PRINTF(args) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: ", g_iSPUNum, g_iCurrentVertex ); spu_printf(args); }
#else
	#define DEBUG_VERTEX_ID(var)
	#define DEBUG_VERTEX_SCALAR(var)
	#define DEBUG_VERTEX_VECTOR(var)
	#define DEBUG_VERTEX_TYPE(szFormat,var)
	#define DEBUG_VERTEX_TYPE4(szFormat,var0,var1,var2,var3)
	#define DEBUG_VERTEX_PRINTF(args)
#endif

#ifdef NDEBUG
	//not in release builds where this will affect performance
	#define CHECK_BUFFER(start,size,dataptr,datasize)												\
	{																								\
		Asura_Assertf( (char*)dataptr >= (char*)start, ( "Buffer underrun! %08x+%x not in %08x+%x\n", dataptr, datasize, start, size ) );						\
		Asura_Assertf( size - ((char*)dataptr - (char*)start) >= datasize, ( "Buffer overrun! %08x+%x not in %08x+%x\n", dataptr, datasize, start, size ) );	\
	}

	//verbose version (always prints something)
	#define CHECK_BUFFERV(start,size,dataptr,datasize)												\
	{																								\
		Asura_Assertf( (char*)dataptr >= (char*)start, ( "Buffer underrun! %08x+%x not in %08x+%x\n", dataptr, datasize, start, size ) );						\
		Asura_Assertf( size - ((char*)dataptr - (char*)start) >= datasize, ( "Buffer overrun! %08x+%x not in %08x+%x\n", dataptr, datasize, start, size ) );	\
		spu_printf( "buffer from 0x%08x to 0x%08x ("#start", "#size")\ninserting @ 0x%08x to 0x%08x ("#dataptr", "#datasize")\n", (char*)start, (char*)start+size, (char*)dataptr, (char*)dataptr+datasize );		\
	}
#else
	#define CHECK_BUFFER
	#define CHECK_BUFFERV
#endif


//
// Defines
//


//tags, 0->31
enum {
	DMA_TAG_VERTS	= 0, //buffer 0 = +0, buffer 1 = +1
	DMA_TAG_INDICES	= 2, //
	DMA_TAG_GET_JOB = 4,
	DMA_TAG_FLAG_COMPLETE,
	DMA_TAG_BONES,
	DMA_TAG_GET_BATCH,
	DMA_TAG_AA_EDGES
};

#define SIZEOF_INDEX 2	// unsigned shorts, no guarantee this will work if you change it.

//
// Variables
//

extern u_int _end; // End of executable (plus static data, etc.)

// static const u_int uSTACK_SIZE_TO_ALLOW = 4 * 1024; // Needs to be a multiple of 16.
// 

// These are not used in this SDK, but they recommend 0 for future compatibility (see Reference_System-Programming_Note section 4)
static const u_int uDMA_TID = 0;
static const u_int uDMA_RID = 0;

static int g_iSPUNum = -1;

//BW dammit what header are these in?
#define max(A,B) (((A)>(B))?(A):(B))
#define min(A,B) (((A)<(B))?(A):(B))


// Bone & vertex buffers
 
//input verts share a buffer with the largest type of output verts (_NoPos) (inputs alloced from end, out's alloced from start)
// to avoid overlap we must add on the size of 1 output vertex (so last/first doesn't overlap)
// also all buffers need an extra 16 bytes to allow for alignment (must use same main memory and local memory alignment)
//BW TODO overlap pos and in buffers (since we can do that now there's no no nopos no buffer).
const u_int		g_uInOutBuffSize = (ASURA_PS3_SPUSKINNING_NUM_VERTICES*sizeof(Asura_PS3_SPU_Skinning_InputVertex)+16 + 127)&~127; //round to 128 bytes
const u_int		g_uOutPosBuffSize = (ASURA_PS3_SPUSKINNING_NUM_VERTICES*sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly)+16 + 15)&~15; //round to 16 bytes
const u_int		g_uIdxBuffSize = (ASURA_PS3_SPUSKINNING_NUM_INDICES*SIZEOF_INDEX+16 + 15)&~15; //round to 16 bytes

static char		g_acInOutBuff[2][g_uInOutBuffSize] __attribute__((aligned(128)));
static char		g_acOutPosBuff[2][g_uOutPosBuffSize] __attribute__((aligned(16)));
static u_short	g_auIdxBuff[2][g_uIdxBuffSize/SIZEOF_INDEX] __attribute__((aligned(16)));


//edge data = 6 edges per vertex, each entry stores a short giving index of the other vertex on this edge, the sign of that value indicates how many polys this edge appears in after culling (if only 1 it's in the silhouette)
//index of the array is the lower valued of the 2 vertices on the edge (e.g. an edge from vert 4 to vert 19 that's on 1 poly would be at index 6*4 and store a value of -19).
#define MAX_EDGES_ON_VERT (6)
const u_int		g_uEdgeOnVertSize = (ASURA_PS3_SPUSKINNING_NUM_VERTICES-1)*MAX_EDGES_ON_VERT*SIZEOF_INDEX;
static short g_aEdgeOnVert[g_uEdgeOnVertSize/SIZEOF_INDEX];


Asura_AMatrix4x4 g_axBoneMatrices[256]; // this is probably a high enough limit for anything we want to do.

static u_int g_uStartVertex;


static Asura_PS3_SPU_SkinningJob g_xJobInfo __attribute__((aligned(16)));

static u_int g_uAtomicBuf[32] __attribute__((aligned(128))); //buffer used by atomic operations
static u_int g_auAligningBuffer[5] __attribute__((aligned(16))); //for small dma's, so all possible alignments of a 4-byte word can be handled

// static u_int g_uInputVertexSize;
// static u_int g_uOutputVertexSize;

static u_int g_uTransferBuff;
static u_int g_uProcessBuff;

static uint64_t g_pAAIndicesOut;

// some debug stats
static u_int g_uTotalVerts = 0;
static u_int g_uTotalBatches = 0;

//
// Functions
//

//
// HalfToFloatVec
//
// Convert 4 halfs in LSBs of input int register into floats
//
inline vec_float4 HalfToFloatVec( vec_uint4 xHalfs )
{
	// 	half (1|5|10) -> float (1|8|23)
	// 	seeeeeffffffffff
	//
	// 	s000eeeeeffffffffff0000000000000
	// 	+01110000	//offset so that 10000 -> 10000000 (exponent of 0, represented as halfway through the uint5 or uint8 range)

	const vec_uint4 x8000 = {0x8000,0x8000,0x8000,0x8000};
	const vec_uint4 x7fff = {0x7fff,0x7fff,0x7fff,0x7fff};
	const vec_uint4 xHalfExp = {0x7c00,0x7c00,0x7c00,0x7c00};
	const vec_uint4 x38000000 = {0x38000000,0x38000000,0x38000000,0x38000000};
	const vec_uint4 x0 = {0,0,0,0};
	const vec_uint4 xff800000 = {0xff800000,0xff800000,0xff800000,0xff800000};

	vec_uint4 xTemp = spu_sl(xHalfs & x8000, 16) + spu_sl(xHalfs & x7fff,13);
	xTemp += x38000000;
	
	vec_float4 xResult = reinterpret_cast<vec_float4>(xTemp);
	
	// do we need to fix any denormalised values?
	vec_uint4 xExps = xHalfs & xHalfExp;
	if ( vec_any_eq(xExps,x0) )
	{
		//fix denormalized values & zero, by subtracting the same float value without a mantissa (which is + or - 0x38000000)
		//e.g. we've changed 0x0000 into 0x38000000, we then subtract 0x38000000 to get 0.0
		//e.g. 0x0010 becomes 0x38020000, the bit that's set is at the correct point, but floats have an implicit high bit set, so subtract that.
		xTemp &= xff800000;
		vec_float4 xF = reinterpret_cast<vec_float4>(xTemp);
	
		//BW this could easily be optimised to use vectors
		for ( int i=0; i<4; i++ )
		{
			if ( xExps[i] == 0 )
			{
				xResult[i] = xResult[i] - xF[i];
			}
		}
	}
	
	return xResult;
}

//BW todo: see if the compiler gets what I'm trying to do here or if it's dumb, if so help it out with more intrinsics and static consts etc.
inline void Unpack11_11_10( Asura_AVector4& xOut, int iPacked )
{
// 	DEBUG_VERTEX_PRINTF(("Unpack11_11_10\n"));
// 	DEBUG_VERTEX_TYPE("%08x",iPacked);

	const vec_int4 xShift = { 1<<21, 1<<10, 1, 0 }; //shift left first to cause sign extend
	const vec_int4 xMask = { 0xffe00000, 0xffe00000, 0xffc00000, 0 }; //1 less bit of precision in z
	const vec_float4 xScale = { 1024.0f/1023.0f, 1024.0f/1023.0f, 512.0f/511.0f, 0.0f };

	vec_int4 xInts = {iPacked,iPacked,iPacked,iPacked};
	xInts *= xShift;
	xInts &= xMask;

	vec_float4 xVec = spu_convtf(xInts,31); //convert to floats, divide by 1<<30, neato
	xVec *= xScale;

// 	DEBUG_VERTEX_TYPE4("%08x",xInts[0],xInts[1],xInts[2],xInts[3]);
// 	DEBUG_VERTEX_TYPE4("%.2f",xVec[0],xVec[1],xVec[2],xVec[3]);
	
	xOut.SetVector( xVec );
}

//BW lowest bits sometimes change => there's a small precision error, but otherwise pack === unpack
inline u_int Pack11_11_10( Asura_AVector4& xVec )
{
// 	DEBUG_VERTEX_PRINTF(("Pack11_11_10\n"));
// 	DEBUG_VERTEX_VECTOR(xVec);

	const vec_float4 xScale = { 1023.0f, 1023.0f*(float)(1<<11), 511.0f*(float)(1<<22) }; //stick the shift in here, saves some work
	const vec_int4 xMask = { 0x7ff, 0x7ff<<11, 0x3ff<<22, 0 };
	vec_float4 xVS = xVec.GetVector();

	xVS *= xScale;
	vec_int4 xInts = spu_convts(xVS,0);
	xInts &= xMask;

// 	DEBUG_VERTEX_TYPE4("%08x",xInts[0],xInts[1],xInts[2],xInts[3]);
// 	DEBUG_VERTEX_TYPE("%08x",(xInts[0] | xInts[1] | xInts[2]));

	return xInts[0] | xInts[1] | xInts[2];
}


void ShadeVertexPosOnly( Asura_PS3_SPU_Skinning_OutputVertex_PosOnly *pxOut, Asura_PS3_SPU_Skinning_InputVertex *pxIn )
{
	Asura_AVector4	xInPos( pxIn->m_xPosition, 1.0f ); 
	Asura_AVector4	axBonePos[4]; 
	float			fAlbedoBlend = 0.0f;
	xInPos.SetW(1.0f);

//	vec_float4 xWeights = HalfToFloatVec( (vec_uint4){pxIn->m_afBoneWeights[0], pxIn->m_afBoneWeights[1], pxIn->m_afBoneWeights[2], pxIn->m_afBoneWeights[3]} );
	float afWeights[4] = { pxIn->m_aucBoneWeights[0], pxIn->m_aucBoneWeights[1], pxIn->m_aucBoneWeights[2], 0 }; 
	afWeights[0] /= 255.0f;
	afWeights[1] /= 255.0f;
	afWeights[2] /= 255.0f;
	afWeights[3] = 1.0f - afWeights[0] - afWeights[1] - afWeights[2];

	axBonePos[0] = g_axBoneMatrices[ pxIn->m_aucBoneIndices[0] ].Apply( xInPos );
	axBonePos[1] = g_axBoneMatrices[ pxIn->m_aucBoneIndices[1] ].Apply( xInPos );
	axBonePos[2] = g_axBoneMatrices[ pxIn->m_aucBoneIndices[2] ].Apply( xInPos );
	axBonePos[3] = g_axBoneMatrices[ pxIn->m_aucBoneIndices[3] ].Apply( xInPos );

	Asura_AVector4 xOut	= (axBonePos[0] * afWeights[0])
						+ (axBonePos[1] * afWeights[1])
						+ (axBonePos[2] * afWeights[2])
						+ (axBonePos[3] * afWeights[3]);
	pxOut->m_afPosition[0] = xOut.GetX();
	pxOut->m_afPosition[1] = xOut.GetY();
	pxOut->m_afPosition[2] = xOut.GetZ();
}

// read what the next batch of data is to process, and remove it from the queue so another SPU doesn't duplicate this work
// return false if no more data
bool GetNextBatch( uint64_t& pVertsIn, uint64_t& pVertPosOut, u_int& uNumVerts, u_int uMaxVerts, uint64_t& pIdxIn, uint64_t& pIdxOut, u_int& uNumIdx )
{
	if ( g_xJobInfo.m_eMode == SPU_SKINNING_POS_AND_CULL )
	{
		// input data is a batch list, load the prescribed batch
		Asura_PS3_SPU_Skinning_Batch xBatch;
		
		DEBUG_GET( &xBatch, g_xJobInfo.m_uInputAddr, sizeof(Asura_PS3_SPU_Skinning_Batch) );
		cellDmaGet( &xBatch, g_xJobInfo.m_uInputAddr, sizeof(Asura_PS3_SPU_Skinning_Batch), DMA_TAG_GET_BATCH, uDMA_TID, uDMA_RID );
		cellDmaWaitTagStatusAll( 1 << DMA_TAG_GET_BATCH ); //BW really we shouldn't have to wait here...

		g_xJobInfo.m_uInputAddr += sizeof(Asura_PS3_SPU_Skinning_Batch);
		
		// detect the terminator
		if ( xBatch.m_uNumVerts == 0 )
		{
			DEBUG_PRINTF(("SPU %d: Character finished\n", g_iSPUNum));
			return false;
		}
		
		g_uTotalVerts += xBatch.m_uNumVerts;
		g_uTotalBatches++;

		pVertsIn = xBatch.m_uVertAddr;
		uNumVerts = xBatch.m_uNumVerts;
		
		Asura_Assert( uNumVerts <= uMaxVerts, "SPU_Skinning GetNextBatch - BATCH IS BIGGER THAN uMaxVerts!!! PANIC!!!\n" );
		//BW hopefully this will only fire if characters haven't been re-exported
		
		pIdxIn = xBatch.m_uStartIndexAddr;
		uNumIdx = xBatch.m_uNumIndices;

		DEBUG_PRINTF(("SPU %d: Got *Actual* Batch: read from %016llx num verts %u\n", g_iSPUNum, pVertsIn, uNumVerts));
	}
	else
	{
		// input data is a raw vertex data, load as much as we can.

		if ( g_uStartVertex >= g_xJobInfo.m_uNumVerts )
		{
			//move on to the next character
			//BW todo start loading the next character's verts! (like so: if g_xNextTask.uNumVerts > 0 etc.)
			DEBUG_PRINTF(("SPU %d: Character finished\n", g_iSPUNum));

			return false;
		}

		uNumVerts = min( g_xJobInfo.m_uNumVerts - g_uStartVertex, uMaxVerts );

		pVertsIn = g_xJobInfo.m_uInputAddr + g_uStartVertex*sizeof(Asura_PS3_SPU_Skinning_InputVertex);

		pIdxIn = 0;
		uNumIdx = 0;

		DEBUG_PRINTF(("SPU %d: Got Batch: start Vertex %u num verts %u\n", g_iSPUNum, g_uStartVertex, uNumVerts));
	}

	// get the write pointers
	pVertPosOut = g_xJobInfo.m_uVertPosOutAddr;
	pIdxOut = g_xJobInfo.m_uIndicesOutAddr;

	// move the write pointers forwards
	g_xJobInfo.m_uVertPosOutAddr += sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly)*uNumVerts;
	g_xJobInfo.m_uIndicesOutAddr += SIZEOF_INDEX*uNumIdx;

	g_uStartVertex += uNumVerts;
	
	return true;
}

int main( uint32_t uSPUNum, uint64_t uJobQueueAddr, uint64_t uGcmLabelAddr, uint64_t uStartupInfo )
{
	g_uTransferBuff = 0;
	bool bDidAnything = false;
	
	g_iSPUNum = static_cast<int>(uSPUNum);

	PING(true);
	DEBUG_PRINTF(("SPU %d: Running\n", g_iSPUNum));	
	
	DEBUG_PRINTF(("SPU %d: stack size %d bytes\n", g_iSPUNum, 256*1024 - reinterpret_cast<u_int>(&_end))); //all our alloced mem is static, so everything that's left is the stack
	
	
	// first load in constants that won't change with each job
	Asura_PS3_SPU_Skinning_StartupInfo xStartupInfo __attribute__((aligned(128)));
	DEBUG_GET( &xStartupInfo, uStartupInfo, sizeof(xStartupInfo) );
	cellDmaGet( &xStartupInfo, uStartupInfo, sizeof(xStartupInfo), DMA_TAG_GET_BATCH, 0, 0 );
	cellDmaWaitTagStatusAll( 0xffffffff );
	uint64_t uBarrierAddr = xStartupInfo.m_uBarrierAddr;

	// setup viewport info so culling can discard triangles that don't intersect any pixel centres (i.e. really small ones and ones beyond edges of screen)
	int x = xStartupInfo.m_auScissorArea[0];
	int y = xStartupInfo.m_auScissorArea[1];
	int w = xStartupInfo.m_auScissorArea[2];
	int h = xStartupInfo.m_auScissorArea[3];
	// values relating to depth are hard coded for a 0 to 1 range, which is hard coded throughout the Asura engine. If it ever becomes variable this will need updating.
	EdgeGeomViewportInfo xViewportInfo = {
		{x,y,w,h}, // uint16_t scissorArea[4];
		{0.0f,1.0f}, // float depthRange[2];
		{},	// float viewProjectionMatrix[16]; - filled in later
		{w/2, h/2, 0.5f, 0.0f},	// float viewportScales[4];
		{w/2 + x, h/2 + y, 0.5f, 0.0f},	// float viewportOffsets[4];
		CELL_GCM_SURFACE_CENTER_1,	//uint8_t sampleFlavor;
		{} //pad
	};


	CellSyncLFQueuePopContainer xQueuePopContainer;
	cellSyncLFQueuePopContainerInitialize( &xQueuePopContainer, &g_xJobInfo, DMA_TAG_GET_JOB );

	// task loop (1 task = 1 character's  skin)
	while (true)
	{
		int iRet = cellSyncLFQueuePopBegin( uJobQueueAddr, &xQueuePopContainer ); // busy-wait for data if queue's empty
		iRet |= cellSyncLFQueuePopEnd( uJobQueueAddr, &xQueuePopContainer );
		if ( iRet != CELL_OK )
		{
			spu_printf("Spu_Skinning: cellSyncLFQueuePop failed. return code %08x\n", cellSyncLFQueuePopBegin);
			break;
		}
		cellDmaWaitTagStatusAll( 1 << DMA_TAG_GET_JOB );


		DEBUG_PRINTF(("SPU %d: Got task\n", g_iSPUNum));
		DEBUG_DUMP_TASK();


		if ( g_xJobInfo.m_uNumVerts == 0 ) // detect terminator
		{
			break;
		}
		
		bDidAnything = true;
		

		//arrange in/out positions to prevent output verts overwriting input verts until those inputs have been read (i.e. a vert can overwrite itself but nothing later in the list)
 		//u_int uMaxVerts = uUseableBufferSize / max(sizeof(Asura_PS3_SPU_Skinning_InputVertex),sizeof(Asura_PS3_SPU_Skinning_OutputVertex_NoPos)+sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly));
		//u_int uInStart = ASURA_PS3_SPUSKINNING_BUFFER_SIZE-uMaxVerts*sizeof(Asura_PS3_SPU_Skinning_InputVertex);
		u_int uOutStart = 0;
		
// 			DEBUG_PRINTF(("g_uInputVertexSize %d\n", g_uInputVertexSize));
// 			DEBUG_PRINTF(("g_uOutputVertexSize %d\n", g_uOutputVertexSize));
// 			DEBUG_PRINTF(("Batch size %d verts\n\n", uMaxVerts));

		uint64_t apVertPosOut[2];
		uint64_t apVertNoPosOut[2];
		u_int uNumVerts[2];
		uint64_t apIndicesOut[2];
		u_int auIndicesStart[2];
		u_int uNumIndices[2];
		u_int auInStart[2];
		bool bDataLastTime = false;
		
		g_pAAIndicesOut = g_xJobInfo.m_uAAIndicesOutAddr;
		int iAAEdgesAvailable = SPU_SKINNING_MAX_AA_EDGES;


		//load bones
		cellDmaLargeGetf( g_axBoneMatrices, g_xJobInfo.m_xShaderConstants.m_uBoneArrayAddr, g_xJobInfo.m_xShaderConstants.m_uNumBones*sizeof(Asura_AMatrix4x4), DMA_TAG_BONES, uDMA_TID, uDMA_RID );

		g_uStartVertex = 0;
		u_int uIndexOffset = 0;

		// vertex batch loop
		do
		{
			bool bDataThisTime = false;

			uint64_t pVertsIn;
			uint64_t pIndicesIn;

			if ( GetNextBatch( pVertsIn, apVertPosOut[g_uTransferBuff], uNumVerts[g_uTransferBuff], ASURA_PS3_SPUSKINNING_NUM_VERTICES,
								pIndicesIn, apIndicesOut[g_uTransferBuff], uNumIndices[g_uTransferBuff] ) )
			{
				auInStart[g_uTransferBuff] = (g_uInOutBuffSize-uNumVerts[g_uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex))&~0xf;
				
				// move downwards to achieve alignment
				auInStart[g_uTransferBuff] += (pVertsIn&0xf) - 0x10;

				CHECK_BUFFER( g_acInOutBuff[g_uTransferBuff], g_uInOutBuffSize, &g_acInOutBuff[g_uTransferBuff][auInStart[g_uTransferBuff]], uNumVerts[g_uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex) );
				
				// queue a get after the preceding put from the same buffer
				// Get with a fence, to make sure the preceding put with the same tag gets finished first.

 				DEBUG_GET( &g_acInOutBuff[g_uTransferBuff][auInStart[g_uTransferBuff]], pVertsIn, uNumVerts[g_uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex) );
				cellDmaUnalignedGetf( &g_acInOutBuff[g_uTransferBuff][auInStart[g_uTransferBuff]], pVertsIn, uNumVerts[g_uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex), DMA_TAG_VERTS+g_uTransferBuff, uDMA_TID, uDMA_RID );
				// this assumes input data is 16-byte aligned, output data doesn't assume this at present.
			
				if ( pIndicesIn )
				{
					//compute a start position from alignment
					auIndicesStart[g_uTransferBuff] = (pIndicesIn&0xf)/2;

					CHECK_BUFFER( g_auIdxBuff[g_uTransferBuff], g_uIdxBuffSize, &g_auIdxBuff[g_uTransferBuff][auIndicesStart[g_uTransferBuff]], uNumIndices[g_uTransferBuff]*SIZEOF_INDEX );

					DEBUG_GET( &g_auIdxBuff[g_uTransferBuff][auIndicesStart[g_uTransferBuff]], pIndicesIn, uNumIndices[g_uTransferBuff]*SIZEOF_INDEX );
					cellDmaUnalignedGetf( &g_auIdxBuff[g_uTransferBuff][auIndicesStart[g_uTransferBuff]], pIndicesIn, uNumIndices[g_uTransferBuff]*SIZEOF_INDEX, DMA_TAG_INDICES+g_uTransferBuff, uDMA_TID, uDMA_RID );
				}
				
				bDataThisTime = true;
			}

			g_uProcessBuff = 1 - g_uTransferBuff;
			

			if ( bDataLastTime )
			{
				cellDmaWaitTagStatusAll( (1<<DMA_TAG_BONES) | (1<<(DMA_TAG_VERTS+g_uProcessBuff)) | (1<<(DMA_TAG_INDICES+g_uProcessBuff)) ); //wait for loading to complete (DMA_TAG_VERTEX_DATA_BUF0 or 1)


				u_int uNoPosAlignment = 0xf & apVertNoPosOut[g_uProcessBuff]; //output data might not be aligned (because output vertex size is 68 bytes)
				u_int uPosAlignment = 0xf & apVertPosOut[g_uProcessBuff]; //output data might not be aligned (because output vertex size is 68 bytes)
				char* pInData = &g_acInOutBuff[g_uProcessBuff][auInStart[g_uProcessBuff]];
				char* pInDataEnd = pInData + uNumVerts[g_uProcessBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex);

				char* pOutPosStart = g_acOutPosBuff[g_uProcessBuff] + uOutStart + uPosAlignment;

				//place other vertex data table after positions (n.b. we might not be using it)
				char* pOutNoPosStart = g_acInOutBuff[g_uProcessBuff] + uOutStart + uNoPosAlignment;

				char* pOutPos = pOutPosStart;
				char* pOutNoPos = pOutNoPosStart;


				CHECK_BUFFER( g_acOutPosBuff[g_uProcessBuff], g_uOutPosBuffSize, pOutPosStart, uNumVerts[g_uProcessBuff]*sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly) );
				
				//
				// SHADE ALL VERTICES
				//
				switch ( g_xJobInfo.m_eMode )
				{
					case SPU_SKINNING_POS_AND_CULL:
					{
						while ( pInData < pInDataEnd )
						{
							DEBUG_VERTEX_ID(((char*)pOutPos - (char*)pOutPosStart)/sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly));
							
							CHECK_BUFFER(g_acOutPosBuff[g_uProcessBuff],g_uOutPosBuffSize,pOutPos, sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly));
							CHECK_BUFFER(g_acInOutBuff[g_uProcessBuff],g_uInOutBuffSize,pInData, sizeof(Asura_PS3_SPU_Skinning_InputVertex));


							//we use the same buffer for input & output, arranged so that in the worst case each output vertex will overwrite an input vertex
							//BW to test: if this causes problems could extend the buffers to allow a 1 vertex space
							ShadeVertexPosOnly(reinterpret_cast<Asura_PS3_SPU_Skinning_OutputVertex_PosOnly*>(pOutPos), reinterpret_cast<Asura_PS3_SPU_Skinning_InputVertex*>(pInData));

							pOutPos += sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly);
							pInData += sizeof(Asura_PS3_SPU_Skinning_InputVertex);
						}
						break;
					}

					case SPU_SKINNING_POS_ONLY_NO_CULL:
					{
						int iVertCount=0;
						while ( pInData < pInDataEnd )
						{
							DEBUG_VERTEX_ID(iVertCount++);

							//we use the same buffer for input & output, arranged so that in the worst case each output vertex will overwrite an input vertex
							//BW to test: if this causes problems could extend the buffers to allow a 1 vertex space
							ShadeVertexPosOnly(reinterpret_cast<Asura_PS3_SPU_Skinning_OutputVertex_PosOnly*>(pOutPos), reinterpret_cast<Asura_PS3_SPU_Skinning_InputVertex*>(pInData));
							
							pOutPos += sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly);
							pInData += sizeof(Asura_PS3_SPU_Skinning_InputVertex);
						}
						break;
					}
				}

				
				if ( g_xJobInfo.m_eMode != SPU_SKINNING_POS_ONLY_NO_CULL )
				{
					//
					// CULL TRIANGLES
					//

					//BW ultimately all this could be filled in by the tools, for now do it manually
					u_short *puIndices = &g_auIdxBuff[g_uProcessBuff][auIndicesStart[g_uProcessBuff]];

					//for some bizarre reason edge needs the vertex buffer aligned to 16-bytes (although each position is 12-bytes!)
					u_int uAlignFudge = 4 - ((u_int)pOutPosStart&0xf) / 4;
					//so to give it an aligned buffer, offset all verts by 4 - the number of words out of alignment that it is.
					//i.e. if it's 4-bytes out, that's 1-word, so pretend there's (4-1) vertices before the start of the buffer
					//(total size 36 bytes, giving alignment of "vertex3" the first actual vertex, as 4-bytes)


					EdgeGeomSpuConfigInfo xEdgeCfg;
					xEdgeCfg.flagsAndUniformTableCount = 0 | EDGE_GEOM_FLAG_STATIC_GEOMETRY_FAST_PATH;
					xEdgeCfg.commandBufferHoleSize = 0; // Irrelevant.
#if CELL_SDK_VERSION < 0x240000
					xEdgeCfg.inputFlavor = EDGE_GEOM_INPUT_FLAVOR_F32c3;
					xEdgeCfg.outputFlavor = EDGE_GEOM_OUTPUT_FLAVOR_F32c3; // Irrelevant - I hope
					xEdgeCfg.shapeFlavor = 0; // Irrelevant.
#else
					xEdgeCfg.inputVertexFormatId = EDGE_GEOM_SPU_VERTEX_FORMAT_F32c3;
					xEdgeCfg.outputVertexFormatId = EDGE_GEOM_RSX_VERTEX_FORMAT_F32c3; // Irrelevant - I hope
					xEdgeCfg.vertexDeltaFormatId = 0; // Irrelevant.
#endif
					xEdgeCfg.indexesFlavorAndSkinningFlavor = EDGE_GEOM_INDEXES_U16_TRIANGLE_LIST_CCW << 4;
					xEdgeCfg.numVertexes = uNumVerts[g_uProcessBuff]+uAlignFudge;
					xEdgeCfg.numIndexes = uNumIndices[g_uProcessBuff];
					xEdgeCfg.indexesOffset = 0xffffffff;

					EdgeGeomLocalToWorldMatrix xLocalToWorldMat =
					{
						{ 1.0f, 0.0f, 0.0f, 0.0f,
						0.0f, 1.0f, 0.0f, 0.0f,
						0.0f, 0.0f, 1.0f, 0.0f, }
					};


					g_xJobInfo.m_xShaderConstants.m_xWorldViewProjectionMatrix.StoreRowsUnaligned( xViewportInfo.viewProjectionMatrix );
					

					// Initialise the Edge library.
					EdgeGeomSpuContext xEdgeContext;
					edgeGeomInitialize( &xEdgeContext,
						&xEdgeCfg,
						NULL, 0, //pvScratchBfr, uScratchBfrSize,
						NULL, 0, //pvStartOfAvailableMemory, uTotalMemoryAvailable,
						0, // Irrelevant - we'll be dealing with all transfers.
						&xViewportInfo,
						&xLocalToWorldMat,
						NULL );


					//index buffer needs to be aligned and indices need to be wrt our local vertex table, not the global one
					u_short* pAlignedIndices = reinterpret_cast<u_short*>(reinterpret_cast<u_int>(puIndices)&~0xf);

					CHECK_BUFFER( g_auIdxBuff[g_uProcessBuff], g_uIdxBuffSize, pAlignedIndices, uNumIndices[g_uProcessBuff]*SIZEOF_INDEX );

					for ( int i=0; i < (int)uNumIndices[g_uProcessBuff]; i++ )
					{
						pAlignedIndices[i] = puIndices[i] - uIndexOffset + uAlignFudge;
					}

					// EDGE_GEOM_FLAG_STATIC_GEOMETRY_FAST_PATH should, hopefully, stop these from doing anything 
					edgeGeomDecompressVertexes( &xEdgeContext, pOutPosStart - uAlignFudge*12, NULL );
					edgeGeomDecompressIndexes( &xEdgeContext, pAlignedIndices );

					u_int uOutputIndexCount = edgeGeomCullTriangles( &xEdgeContext, EDGE_GEOM_CULL_BACKFACES_AND_FRUSTUM );

					

//spu_printf("uOutputIndexCount %d / %d\n", uOutputIndexCount, uNumIndices[g_uProcessBuff]);

					//adjust the alignment to match main memory output buffer & correct the index for our big vertex table
					for ( int i=(int)uOutputIndexCount-1; i >= 0; i-- )
					{
						puIndices[i] = pAlignedIndices[i] + uIndexOffset - uAlignFudge;
					}

					// pad it with 0xffff - this is not as efficient as actually truncating the list (and changing the draw call)
					// but the overhead for the RSX to parse this is so small it's not worth the effort of fixing.
					// N.B. even if we reduce the size of the output index list, there's no need to make the length variable (since the allocated space will be constant)
					for ( int i=(int)uOutputIndexCount; i < (int)uNumIndices[g_uProcessBuff]; i++ )
					{
						puIndices[i] = 0xffff;
					}
					

					if ( iAAEdgesAvailable > 0 && g_pAAIndicesOut != 0 )
					{					
						//Find Silhouette edges for edge anti-aliasing
						// the method employed here is to mark edges as they're found on the culled triangles, then output those that were found only once.

						//wait for completion of previous aa download
						cellDmaWaitTagStatusAll( 1<<DMA_TAG_AA_EDGES );

						//init to 0, this is fine because the other vert must be > 0 (since any edge containing 0 will use that as the index to this table), and it clears the counters
						Asura_Mem::Set(g_aEdgeOnVert, 0, (uNumVerts[g_uProcessBuff]-1)*MAX_EDGES_ON_VERT*sizeof(short));
						
						for ( u_short* puCulledTri = puIndices; puCulledTri < &puIndices[uOutputIndexCount]; puCulledTri += 3 )
						{
							// mark the edge
							for ( int j=0; j<3; j++ )
							{
								int k = j+1; if ( k == 3 ) k = 0;
								short p0 = min(puCulledTri[j],puCulledTri[k]) - uIndexOffset;
								short p1 = max(puCulledTri[j],puCulledTri[k]) - uIndexOffset;

								short *pEdge = &g_aEdgeOnVert[p0*MAX_EDGES_ON_VERT]; //edge list for lower vertex

								//check to see if p1 is already in the list
								for ( int l=0; l < MAX_EDGES_ON_VERT-1; l++ )
								{
									//I've written this to be easy to roll out & easily implemented with conditional instructions, though I dunno if the compiler's smart enough to realise.
									if ( *pEdge != 0 && abs(*pEdge) != p1 ) pEdge++;
								}

								//if the list fills up this will replace the last edge in the list, but won't overrun. This means some edges won't get AA'd but it's no big deal.
								*pEdge = (*pEdge==p1)?-p1:p1; //flag if we've seen it before (-ve)
								//*pEdge = p1 - (*pEdge)*2; //equivalent without conditional
							}
						}

						//output the flagged edges
						short* pWhat = g_aEdgeOnVert;
						u_short* puAAEdges = reinterpret_cast<u_short*>(pWhat); // reuse the edge list, be careful we don't overtake.
						puAAEdges += (g_pAAIndicesOut&0xf)/2;
						
						u_short* puOut = puAAEdges;
						for ( int iEdge = 0; iEdge < (int)(uNumVerts[g_uProcessBuff]-1)*MAX_EDGES_ON_VERT; iEdge++ )
						{
							if ( g_aEdgeOnVert[iEdge] > 0 && g_aEdgeOnVert[iEdge] != 0 // if this edge is flagged (-ve) it has been seen by more than 1 poly, so ignore. we want just the ones on 1 poly.
								 && (char*)puOut < (char*)&g_aEdgeOnVert[iEdge] ) //this will very rarely fail, and in that rare case I'd rather have no AA on an edge than corrupt nonsense.
							{
								//output the edge as a silhouette edge for anti-aliasing
								u_short p0 = iEdge/MAX_EDGES_ON_VERT;
								u_short p1 = abs(g_aEdgeOnVert[iEdge]);
								
								Asura_Assert( p0 < uNumVerts[g_uProcessBuff], "Sanity check failed Loop index calc wierdness." );
								Asura_Assert( p0 < p1, "Sanity check failed degenerate edge impossibility paradox conundrum." );
								Asura_Assertf( p1 < uNumVerts[g_uProcessBuff], ("Other point on edge is garbage, should have been cleared before use. 0x%04x / 0x%04x @ 0x%08x", p1, uNumVerts[g_uProcessBuff], &g_aEdgeOnVert[iEdge] ) );

								*(puOut++) = p0+uIndexOffset;
								*(puOut++) = p1+uIndexOffset;

							}
						}

						u_int uNumAAEdges = (puOut - puAAEdges)/2;


						//DMA out the data

						// output anti-alias edges
						u_int uSize = min(iAAEdgesAvailable,(int)uNumAAEdges)*2*SIZEOF_INDEX;
						DEBUG_PUT( puAAEdges, g_pAAIndicesOut, uSize );
						cellDmaUnalignedPut( puAAEdges, g_pAAIndicesOut, uSize, DMA_TAG_AA_EDGES, uDMA_TID, uDMA_RID );

						g_pAAIndicesOut += uSize;
						iAAEdgesAvailable -= uSize/(2*SIZEOF_INDEX);
					
					} // if ( iAAEdgesAvailable > 0 )


					// force it to wait for previous PUT (fences & barriers don't seem to be enough to guarantee these are written in strict order)
					// DAMMIT WHY CANT I FENCE/BARRIER ON A DIFFERENT TAG???!?!?!?!?!?!?!
					// I want to wait on just the put, but this will wait on the get too. dammit.
					// well we were gonna wait on it in a moment anyway.
					cellDmaWaitTagStatusAll( 1<<(DMA_TAG_INDICES+g_uTransferBuff) );
					// To be more precise: put must wait for previous put, regardless of double-buffering, because they both write to the same buffer and unaligned puts have glitches if you don't wait (only seen it in shorts, not words, so likely only affects index buffer)
					// However get only needs to wait for the put in the same double-buffer (so it can overwrite the indices with the next strip)
					// And of course all this "waiting" could be done with fences so SPU doesn't actually wait.
					
					// output indices
					DEBUG_PUT( &g_auIdxBuff[g_uProcessBuff][auIndicesStart[g_uProcessBuff]], apIndicesOut[g_uProcessBuff], uNumIndices[g_uProcessBuff]*SIZEOF_INDEX );
					cellDmaUnalignedPut( &g_auIdxBuff[g_uProcessBuff][auIndicesStart[g_uProcessBuff]], apIndicesOut[g_uProcessBuff], uNumIndices[g_uProcessBuff]*SIZEOF_INDEX, DMA_TAG_INDICES+g_uProcessBuff, uDMA_TID, uDMA_RID );
				}  // if ( g_xJobInfo.m_eMode != SPU_SKINNING_POS_ONLY_NO_CULL )

				// output transformed positions
				DEBUG_PUT( pOutPosStart, apVertPosOut[g_uProcessBuff], sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly) * uNumVerts[g_uProcessBuff] );
				cellDmaUnalignedPut( pOutPosStart, apVertPosOut[g_uProcessBuff], sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly) * uNumVerts[g_uProcessBuff], DMA_TAG_VERTS+g_uProcessBuff, uDMA_TID, uDMA_RID );
				
				
				uIndexOffset += uNumVerts[g_uProcessBuff];
			} // if ( bDataLastTime )

			bDataLastTime = bDataThisTime;
			g_uTransferBuff = g_uProcessBuff;

		} while ( bDataLastTime ); // vertex batch loop

//spu_printf( "Think there's %d edges\n", SPU_SKINNING_MAX_AA_EDGES - iAAEdgesAvailable );


		//pad AA edge table with ffff's
		if ( iAAEdgesAvailable > 0 && g_pAAIndicesOut != 0 )
		{
			cellDmaWaitTagStatusAll( 1<<DMA_TAG_AA_EDGES );
			Asura_Mem::Set( g_aEdgeOnVert, 0xff, g_uEdgeOnVertSize );

			while ( iAAEdgesAvailable > 0 )
			{
				cellDmaWaitTagStatusAll( 1<<DMA_TAG_AA_EDGES );

				short* pData = g_aEdgeOnVert;
				pData += (g_pAAIndicesOut&0xf)/2;
				u_int uSize = min( (u_int)iAAEdgesAvailable*2*SIZEOF_INDEX, g_uEdgeOnVertSize-0x10 );
//spu_printf( "padding %d\n", uSize );

				DEBUG_PUT( pData, g_pAAIndicesOut, uSize );
				cellDmaUnalignedPut( pData, g_pAAIndicesOut, uSize, DMA_TAG_AA_EDGES, uDMA_TID, uDMA_RID );
				
				iAAEdgesAvailable -= uSize/(2*SIZEOF_INDEX);
				g_pAAIndicesOut += uSize;
			}
		}


		if ( g_xJobInfo.m_eMode == SPU_SKINNING_POS_AND_CULL )
		{
			//BW todo write new num indices to the command buffer
			//BW at the moment I'm just padding with 0xffff's, and RSX has got a lot quicker, not sure it's worth doing this.
		}
	
	} // task loop


	// wait for it all to finish
	cellDmaWaitTagStatusAll( 0xFFFFFFFF );


	//
	// terminate
	//

	// wait for all the spus before we tell the RSX it's data is ready
	cellSyncBarrierNotify(uBarrierAddr);
	cellSyncBarrierWait(uBarrierAddr);


	// This was the last character, tell the RSX the character data is ready
	// strictly speaking we want only the last SPU to finish to do this - this could cause a problem if RSX catches up to SPUs and SPUs have differing task sizes.
	if ( uGcmLabelAddr != 0 )
	{
		//BW apparently atomic ops don't work on an RSX label, use DMA instead
		// 			cellAtomicStore32( g_uAtomicBuf, g_xJobInfo.uGcmLabelAddr, g_xJobInfo.uGcmLabelValue );
		//low 4 bits must be the same
		u_int* pAligned = reinterpret_cast<u_int*>(&g_auAligningBuffer[(uGcmLabelAddr&0xf)>>2]);
		*pAligned = uSPU_SKINNING_RSX_FLAG;
		DEBUG_PUT( pAligned, uGcmLabelAddr, sizeof(u_int) );
		cellDmaSmallPut( pAligned, uGcmLabelAddr, sizeof(u_int), DMA_TAG_FLAG_COMPLETE, 0, 0 );
	}
	else
	{
		if ( bDidAnything ) spu_printf("SPU %d: error uGcmLabelAddr = NULL, SPU code was probably invoked without any data\n", g_iSPUNum);
	}

	DEBUG_PRINTF(("SPU %d: Terminating\n", g_iSPUNum));
	
	PING(false);

	cellDmaWaitTagStatusAll( 0xFFFFFFFF );
	
#if (0)
	// The SPU has an annoying habit of working fine if I do printfs during processing
	// (usually because it affects synchronisation, but on one occasion I forgot to return a value from a bool function, and enabling DEBUG_PRINTF made the compiler implicitly return true rather than false from that function)
	// So sometimes I'll want to just dump info here.
	
	spu_printf("SPU %d: Processing complete\n", g_iSPUNum);
	spu_printf("SPU %d: Total Batches %d\n", g_iSPUNum, g_uTotalBatches);
	spu_printf("SPU %d: Total Verts %d\n", g_iSPUNum, g_uTotalVerts);
#endif	
		
	sys_spu_thread_exit( EXIT_SUCCESS );
}




