//////////////////////////////////////////////////////////////////////////////
//
// Tools_Blueprint_Export.cpp
//
// Created by HE - 24 April 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Blueprint/Tools_Blueprint_Export.h"

// asura
#include "Blueprint/Asura_Blueprint_System.h"
#include "Blueprint/Asura_Blueprint.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

//tools
#include "Blueprint/Tools_Blueprint_Edit_Object.h"
#include "Misc/Tools_ProgressIndicator.h"
#include "PC/FileSystem/Tools_FileSystem.h"

//libs
#include <stdarg.h>
#include <errno.h>

//////////////////////////////////////////////////////////////////////////////
//
// Tools_Blueprint_XMLExporter
//
//////////////////////////////////////////////////////////////////////////////

Tools_Blueprint_XMLExporter::Tools_Blueprint_XMLExporter( bool bRaw, const char* szPath, const char* szStyleSheet ) : PARENT(szPath, szStyleSheet)
{
	m_pxBlueprintEO = new Tools_Blueprint_Edit_Object();
	m_iNumBlueprintsDone = 0;
	m_iNumHashTypes = Asura_Blueprint_System::GetNumTypes();
	m_pxProgressBar1 = new Tools_ProgressIndicator("(Very approximate) total");
	m_pxProgressBar2 = new Tools_ProgressIndicator("Blueprints exported");
	m_bExportRaw = bRaw;
}

Tools_Blueprint_XMLExporter::~Tools_Blueprint_XMLExporter()
{
	delete m_pxBlueprintEO;
	delete m_pxProgressBar1;
	delete m_pxProgressBar2;
}

void Tools_Blueprint_XMLExporter::StartXMLExport()
{
	m_pxProgressBar1->Update(0, m_iNumHashTypes, true );
	m_pxProgressBar2->Update(0, 0, true );

	//start the export - one XML file for every blueprint type
	const int iNumTypes = Asura_Blueprint_System::GetNumTypes();
	int		  iType		= 0;

	for(Asura_Blueprint_Type_ConstIterator xIterator; !xIterator.Done(); xIterator.Next(), iType++)
	{
		m_pxProgressBar1->Update( iType, iNumTypes, true );

		Asura_Hash_ID uTypeHashID = xIterator.GetCurrentHash();
		int iNumRootBlueprints = Asura_Blueprint_System::GetNumRootBlueprintsInType(uTypeHashID);

		//recursively write the blueprints to the buffer, starting from the root blueprints
		for ( int j = 0; j < iNumRootBlueprints; ++j)
		{
			Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetRootBlueprint(uTypeHashID, j);

			WriteBlueprintToXML(pxBlueprint, uTypeHashID);

		}

		WriteXMLToFile( Asura_Blueprint_System::GetTypeName(uTypeHashID)->GetString() );

	}

	m_pxProgressBar1->Finish();
	m_pxProgressBar2->Finish();
}

void Tools_Blueprint_XMLExporter::WriteBlueprintToXML( Asura_Blueprint* pxBlueprint, Asura_Hash_ID uTypeHashID ){

	Asura_Hash_ID uThisBptHashID = pxBlueprint->GetHashID();
	int iNumChildren = Asura_Blueprint_System::GetNumChildBlueprints( uTypeHashID, uThisBptHashID, true );

	// <blueprint name="BlueprintName" numChildren="42">
	AddToBuffer( GetTabs() );
	AddToBuffer( "<blueprint name=\"" );
	AddToBufferEscaped( pxBlueprint->GetName().GetString() );
	AddToBuffer( "\" numChildren=\"" );
	AddToBuffer( iNumChildren );
	AddToBuffer("\" >\n" );

	IncTabs();

	// Starting with this blueprint, look at the parameters and write them out if they haven't been overridden.
	// Repeat for each parent going up the tree to get all inherited parameters for this blueprint.

    Asura_Blueprint* pxLookingAt = pxBlueprint;
    while ( pxLookingAt )
    {
        for ( int i = 0; i < pxLookingAt->GetNumParameters(); ++i )
        {
            Asura_Parameter* pxParameter = pxLookingAt->GetParameterByIndex( i );
            Asura_Value xFinalValue;
            if ( pxParameter && pxParameter->GetOverrideType() == ABOT_NONE && pxBlueprint->GetParameterValueByHashID( pxParameter->GetHashID(), xFinalValue ) )
            {
                
				AddToBuffer( GetTabs() );
				AddToBuffer( "<parameter name=\"" );
				AddToBufferEscaped( pxParameter->GetName().GetString() );
				AddToBuffer( "\" " );

				if( m_bExportRaw )
				{
					switch( xFinalValue.GetValueType() )
					{
						case Asura_Value::VT_INT:
						{
							AddToBuffer( "final_value=\"" );
							AddToBuffer( xFinalValue.GetIntValue() );
							break;
						}
						case Asura_Value::VT_FLOAT:
						{
							AddToBuffer( "final_value=\"" );
							AddToBuffer( xFinalValue.GetFloatValue() );
							break;
						}
						case Asura_Value::VT_BOOL:
						{
							AddToBuffer( "final_value=\"" ); 
							AddToBuffer( xFinalValue.GetBoolValue() ? "true" : "false" );
							break;
						}
						case Asura_Value::VT_HASH:
						{
							AddToBuffer( "final_value=\"" );
							AddToBuffer( xFinalValue.GetHashIDValue() );
                            AddToBuffer( " " );
                            AddToBufferEscaped( xFinalValue.GetHashStringValue() );
							break;
						}
						default:
						{
							Asura_Assertf(false, ("Unknown parameter type: %d", xFinalValue.GetValueType() ));
							break;
						}
					}
				}
				else
				{
					// We want the ParameterValue, whether it be skin/texture/colour/whatever.
                    const int iTemplatedType = m_pxBlueprintEO->GetTemplatedTypeFromTemplateID( pxParameter->GetTemplate() );
                    Asura_String xValueString;
                    m_pxBlueprintEO->GetValueString( xFinalValue, iTemplatedType, xValueString );

					// If the value is unset we need to change the output slightly as angle brackets will confuse the XML.
					AddToBuffer( "final_value=\"" ); AddToBufferEscaped( xValueString.GetString() );
				}
				
				// <parameter name="name" final_value="arthur" />
				AddToBuffer( "\" />\n" );				

            }
        }

        pxLookingAt = pxLookingAt->GetParent();
    }

	++m_iNumBlueprintsDone;
	m_pxProgressBar2->Update(m_iNumBlueprintsDone, 0, true );

	//handle children recursively
	for (int i = 0; i < iNumChildren; ++i )
	{
		Asura_Blueprint* xChildBpt = Asura_Blueprint_System::GetChildBlueprint( uTypeHashID, uThisBptHashID, true, i );
		WriteBlueprintToXML( xChildBpt, uTypeHashID );
	}

	DecTabs();
	AddToBuffer( GetTabs() );
	AddToBuffer( "</blueprint>\n" );

}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
