
//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE        :  Tools_BotanicalCreation_Rendering.cpp
//
//  Created by Dave Reed - 13/01/06
//
//  Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//
//  Botanical rendering in the tools
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_BOTANICALS

#include "BotanicalCreation/Tools_Mode_BotanicalCreation.h"
#include "BotanicalCreation/Tools_BotanicalCreation_BotanicalShape.h"
#include "BotanicalCreation/Tools_BotanicalCreation_Rendering.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Misc/Tools_Render.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Direct3d.h"
#include "Rendering/Asura_Direct3d_Shaders.h"
#include "Rendering/Asura_Fog.h"
#include "Camera/Asura_Camera.h"
#include "Rendering/Tools_RenderModules.h"
#include "Rendering/Asura_Direct3D_Shaders.h"

//#define ASURA_BOTANICALS_FLAT_LIGHTING


#define TOOLS_BOTANICALRENDERABLE_SHADERCONST_FOR_WORLDMTX 30

// ------------------------------------------------------------------------------------------------
// Functions to set the states for fog. 
//
// Currently called one per batch - which is bad. (If we continue to use this module
// to render botanicals in placement mode, then it needs optimizing - as it's all very slow, as it
// was intended to only be used to render a single instance in BotanicalCreation mode)
// ------------------------------------------------------------------------------------------------

static void Tools_BotanicalRenderable_StartFoggedRendering()
{	
	// turn on fog if we have to...
	if (Asura_Fog::IsFogRenderable())
	{
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);

		if (Asura_Fog::IsGraphFogEnabled())
		{
			Asura_Direct3D_Shaders::RequestVertexShader(ASURA_D3D_VS_TOOLS_BOTANICALS_GRAPHFOG);		
		}
		else
		{		
			Asura_Direct3D_Shaders::RequestVertexShader(ASURA_D3D_VS_TOOLS_BOTANICALS_LINEARFOG);		
		}

		// This sets the fog constants (and also updates the viewproj matrix)
		Tools_RenderModule_System::SetupShaderConstantsForFog();
	}

	// After using .fx effects, these can be screwed up
	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);	
}


static void Tools_BotanicalRenderable_EndFoggedRendering()
{
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, FALSE);
}


static void Tools_BotanicalRenderable_SetWorldMatrix(D3DXMATRIX &xWorldMat)
{
	if (Asura_Fog::IsFogRenderable())
	{
		// Using the shader
		D3DXMATRIX xTempMat;		
		D3DXMatrixTranspose( &xTempMat, &xWorldMat);
		Asura_Direct3D::D3DDevice->SetVertexShaderConstantF(TOOLS_BOTANICALRENDERABLE_SHADERCONST_FOR_WORLDMTX, (float *)&xTempMat, 4);
	}
	else
	{
		// Using fixed-function
		Asura_Direct3D::D3DDevice->SetTransform( D3DTS_WORLD, &xWorldMat );
	}
}


// ------------------------------------------------------------------------------------------------
// Mesh Renderable
// ------------------------------------------------------------------------------------------------
Tools_BotanicalRenderable_Mesh::Tools_BotanicalRenderable_Mesh(Tools_BotanicalElementLOD* pxLOD)
							   :m_xLightingSolution(true,false)
{
	// We now take a copy of the mesh, so we can render it ourselves (culling on/off,
	// lighting, alpha test, etc). This is slow, as it doesn't use vertex buffers or shaders,
	// and the list isn't stripped/indexed, so there's lots of duplicated verts

	m_pxVertices=0;
	m_iNumTriangles=0;
	m_pxLOD=pxLOD;

	if (pxLOD->m_pxMesh)
	{
		Asura_Mesh *pxMesh=pxLOD->m_pxMesh;

		// Get the texture from the first poly
		m_iTextureID=Asura_Material_System::GetTextureID(m_pxLOD->m_pxMesh->TextureDataPtr[0].m_iMaterialID);

		if (!pxMesh->AreNormalsValid())
		{
			pxMesh->CreateNormals();
		}

		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			if (pxMesh->PolygonDataPtr[i].IsQuad())
				m_iNumTriangles+=2;
			else
				m_iNumTriangles+=1;
		}		

		m_pxVertices = new D3DGenericVertex[m_iNumTriangles*3];
		D3DGenericVertex *pxVtx=m_pxVertices;
		
		for(int i=0;i<pxMesh->NumberOfPolygons;i++)
		{
			Asura_Mesh_Polygon *pxPoly=&pxMesh->PolygonDataPtr[i];

			int iTrisInPoly=pxPoly->NumberOfVertices()-2;

			for(int j=0;j<iTrisInPoly;j++)
			{			
				for(int k=0;k<3;k++)
				{
					const int iaVertTable[2][3]={{0,1,2},{0,2,3}};
					int   iVertInPoly=iaVertTable[j][k];					
					int   iVertexID=pxPoly->VertexID[iVertInPoly];
					
					Asura_Mesh_Vertex *pxSrcVtx = &pxMesh->VertexDataPtr[iVertexID];
					
					pxVtx->X = pxSrcVtx->Position.x;
					pxVtx->Y = pxSrcVtx->Position.y;
					pxVtx->Z = pxSrcVtx->Position.z;										
					pxVtx->U = pxMesh->TextureDataPtr[i].U[0][iVertInPoly];
					pxVtx->V = pxMesh->TextureDataPtr[i].V[0][iVertInPoly];		

#ifdef ASURA_BOTANICALS_FLAT_LIGHTING
					pxVtx->Normal = Asura_Vector_3(0,-1,0);
#else
					pxVtx->Normal = pxMesh->NormalsDataPtr[i].VertexNormal[iVertInPoly];				
#endif

					pxVtx++;
				}
			}
		}
	}
}



Tools_BotanicalRenderable_Mesh::~Tools_BotanicalRenderable_Mesh()
{
	if (m_pxVertices!=NULL)
	{
		delete [] m_pxVertices;
	}
}


void Tools_BotanicalRenderable_Mesh::Render(float fFade, Asura_Vector_3 *pxPosition, Asura_Matrix_3x3 *pxOrientation, float fScale,uint32 uFlags)
{
	SCOPEDPIXEVENT((0xff00ff00, L"(Tools) Render Botanical Mesh"));

	// Update the lighting
	m_xLightingSolution.SetPosition((*pxPosition));
	m_xLightingSolution.RecalculateLighting(true, false);
	
	// Get the SH representation of the scene lights
	const Asura_SphericalHarmonics_Order2_RGB& xSH = Asura_GI_Solution::s_xSHForCurrentRender;	

	// Update the vertex RGBs
	D3DGenericVertex *pxVtx=m_pxVertices;

	if (uFlags&&TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTRED)
	{		
		for(int i=0;i<m_iNumTriangles*3;i++)
		{
			pxVtx->Diffuse=0xffff0000;
			pxVtx++;
		}
	}
	else if (uFlags&&TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTBLUE)
	{		
		for(int i=0;i<m_iNumTriangles*3;i++)
		{
			pxVtx->Diffuse=0xff0000ff;
			pxVtx++;
		}
	}
	else
	{	
		if (Tools_Mode_BotanicalCreation::IsLightingEnabled())
		{	
			for(int i=0;i<m_iNumTriangles*3;i++)
			{
				// Is this correct? It seems to work pretty well... clamping to 0 gave much nicer 
				// results
				#define LIGHTING_CLAMP(x) x //((x>0)?x:0)

				Asura_Vector_3 xNormal = pxVtx->Normal*(*pxOrientation);

				Asura_RGBA xLighting;
				xLighting.m_fR = LIGHTING_CLAMP(xSH.m_afRedCoefficients[0])   
					+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[1] * xNormal.x) 
					+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[2] * xNormal.y) 
					+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[3] * xNormal.z);

				xLighting.m_fG = LIGHTING_CLAMP(xSH.m_afGreenCoefficients[0])
					+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[1] * xNormal.x)
					+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[2] * xNormal.y)
					+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[3] * xNormal.z);

				xLighting.m_fB = LIGHTING_CLAMP(xSH.m_afBlueCoefficients[0])  
					+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[1] * xNormal.x)
					+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[2] * xNormal.y)
					+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[3] * xNormal.z);
				
				pxVtx->Diffuse=xLighting.ToUIntARGB(true) |0xff000000;	
				pxVtx++;
			}
		}
		else
		{
			for(int i=0;i<m_iNumTriangles*3;i++)
			{
				pxVtx->Diffuse=0xff808080;	
				pxVtx++;
			}
		}
	}


	// Select the texture
	Asura_Render::CurrentStates.RequestTextureIndex(0, m_iTextureID);								

	// Render them
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS_2X);
	Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);

	int nFade=(int)((1.0f-fFade)*127.0f);
	Asura_Render::CurrentStates.RequestAlphaTest( true , 0x80+nFade);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	
	// Not seen an Asura Matrix -> D3D Matrix function anywhere yet
	// Also putting the scale in here
	D3DXMATRIX matWorld;	
	matWorld._11 = pxOrientation->e11*fScale;	matWorld._12 = pxOrientation->e21*fScale;	matWorld._13 = pxOrientation->e31*fScale;
	matWorld._21 = pxOrientation->e12*fScale;	matWorld._22 = pxOrientation->e22*fScale;	matWorld._23 = pxOrientation->e32*fScale;
	matWorld._31 = pxOrientation->e13*fScale;	matWorld._32 = pxOrientation->e23*fScale;	matWorld._33 = pxOrientation->e33*fScale;	
	matWorld._41 = pxPosition->x;
	matWorld._42 = pxPosition->y;
	matWorld._43 = pxPosition->z;
	matWorld._44 = 1.0f;
	matWorld._14 = matWorld._24 = matWorld._34 = 0.0f;
	Tools_BotanicalRenderable_SetWorldMatrix(matWorld);

	// (AAM - 21.08.2006) New renderers won't leave the view-proj matrix here
	#if defined ASURA_TARGET_SM3

		const D3DXMATRIX& xViewProj = Asura_Direct3D::GetViewProjectionMatrix();

		D3DXMATRIX xViewProjTranspose;
		D3DXMatrixTranspose( &xViewProjTranspose, &xViewProj );

		Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( 0, xViewProjTranspose, 4 );

	#endif

	// Draw it
	Tools_BotanicalRenderable_StartFoggedRendering();
	Asura_Direct3D::D3DDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, m_iNumTriangles, (void*) m_pxVertices, sizeof( m_pxVertices[0]));
	Tools_BotanicalRenderable_EndFoggedRendering();

	Asura_Render::CurrentStates.RequestAlphaTest(false);

	// Restore the world matrix
	D3DXMatrixTranslation( &matWorld, 0,0,0 );
	Tools_BotanicalRenderable_SetWorldMatrix(matWorld);	
	
	
}




// ------------------------------------------------------------------------------------------------
// Billboards Renderable
// ------------------------------------------------------------------------------------------------


Tools_BotanicalRenderable_Billboards::Tools_BotanicalRenderable_Billboards(const Tools_BotanicalElementLOD *pxLOD)
									 :m_xLightingSolution(true,false)
{
	m_pxLOD=pxLOD;
	m_pxBillboards=NULL;
	m_iNumBillboards=0;

	if (pxLOD->m_pxMesh)
	{
		// Get the texture from the first poly
		m_iTextureID=Asura_Material_System::GetTextureID(m_pxLOD->m_pxMesh->TextureDataPtr[0].m_iMaterialID);

		// Count the number of quads in the mesh
		Asura_Mesh_Polygon *pxPolys=pxLOD->m_pxMesh->PolygonDataPtr;

		for(int i=0;i<pxLOD->m_pxMesh->NumberOfPolygons;i++)
		{
			if (pxPolys[i].IsQuad())
			{
				m_iNumBillboards++;
			}
		}

		if (m_iNumBillboards>0)
		{
			m_pxBillboards = new Tools_Botanicals_Billboard[m_iNumBillboards];
			int iIdx=0;

			float fMinX=100000, fMaxX=-100000;
			float fMinY=100000, fMaxY=-100000;
			float fMinZ=100000, fMaxZ=-100000;

			for(int i=0;i<pxLOD->m_pxMesh->NumberOfPolygons;i++)
			{
				if (pxPolys[i].IsQuad())
				{					
				
					Asura_Vector_3 xCentre=Asura_Vector_3(0,0,0);

					Tools_Botanicals_Billboard *pxBB=&m_pxBillboards[iIdx++];

					Tools_Mode_BotanicalCreation::CreateBillboardFromQuad(pxBB,pxLOD->m_pxMesh,i);
															
					// Track the extents of the billboards
					if (pxBB->xPos.x<fMinX) fMinX=pxBB->xPos.x; else if (pxBB->xPos.x>fMaxX) fMaxX=pxBB->xPos.x;
					if (pxBB->xPos.y<fMinY) fMinY=pxBB->xPos.y; else if (pxBB->xPos.y>fMaxY) fMaxY=pxBB->xPos.y;
					if (pxBB->xPos.z<fMinZ) fMinZ=pxBB->xPos.z; else if (pxBB->xPos.z>fMaxZ) fMaxZ=pxBB->xPos.z;

				}
			}

			m_xCentre.x=0.5f*(fMinX+fMaxX);
			m_xCentre.y=0.5f*(fMinY+fMaxY);
			m_xCentre.z=0.5f*(fMinZ+fMaxZ);


			// Now create the D3D verts. We'll update the positions on the CPU each frame,
			// but don't need to touch the UVs or index list
			m_pxVertices=new D3DGenericVertex[m_iNumBillboards*4];
			m_pIndices=new uint16[m_iNumBillboards*6];

			int iIndexIndex=0;
			int iVertexIndex=0;

			for(int i=0;i<m_iNumBillboards;i++)
			{
				Tools_Botanicals_Billboard *pxBB=&m_pxBillboards[i];

				m_pIndices[iIndexIndex++] = static_cast<uint16>( iVertexIndex + 0 );
				m_pIndices[iIndexIndex++] = static_cast<uint16>( iVertexIndex + 0 );
				m_pIndices[iIndexIndex++] = static_cast<uint16>( iVertexIndex + 1 );
				m_pIndices[iIndexIndex++] = static_cast<uint16>( iVertexIndex + 3 );
				m_pIndices[iIndexIndex++] = static_cast<uint16>( iVertexIndex + 2 );
				m_pIndices[iIndexIndex++] = static_cast<uint16>( iVertexIndex + 2 );

				for(int j=0;j<4;j++)
				{				
					uint32 uWhichU[4]={0,1,1,0};
					uint32 uWhichV[4]={0,0,1,1};
					m_pxVertices[iVertexIndex].Diffuse = 0xff808080;
					m_pxVertices[iVertexIndex].U = pxBB->fU[uWhichU[j]];
					m_pxVertices[iVertexIndex].V = pxBB->fV[uWhichV[j]];
					iVertexIndex++;
				}	

				// Also create a normal, for lighting. Assumes the cluster of
				// billboards are a spherical-ish shape. We need something better
				// than this
#ifdef ASURA_BOTANICALS_FLAT_LIGHTING
				pxBB->xNormal=Asura_Vector_3(0,-1,0);
#else
				pxBB->xNormal=pxBB->xPos-m_xCentre;
//				pxBB->xNormal.y=0;
 				pxBB->xNormal.Normalise();
#endif
			}			
		}
	}
}


Tools_BotanicalRenderable_Billboards::~Tools_BotanicalRenderable_Billboards()
{
	if (m_pxBillboards)
	{
		delete [] m_pxBillboards;
	}
	
	if (m_pxVertices)
	{	
		delete [] m_pxVertices;
	}

	if (m_pIndices)
	{
		delete [] m_pIndices;
	}
}





void Tools_BotanicalRenderable_Billboards::Render(float fFade, Asura_Vector_3 *pxPosition, Asura_Matrix_3x3 *pxOrientation, float fScale,uint32 uFlags)
{
	SCOPEDPIXEVENT((0xff00ff00, L"(Tools) Render Botanical Billboards"));

	// Create the fwd and up vectors for the billboards	
	Asura_Matrix_3x3 xMat = *Asura_Camera::GetOrientation();
	Asura_Vector_3 xBillboardUp=Asura_Vector_3(0.0f,-1.0f,0.0f)*xMat;
	Asura_Vector_3 xBillboardRight=Asura_Vector_3(1.0f,0,0.0f)*xMat;
	Asura_Vector_3 xLightingNormal=Asura_Vector_3(0.0f,0,-1.0f)*xMat;
	uint32 uBillboardCol;

	// Recalc the lighting. This should only need doing if the position
	// or lights have changed - but as we're in the editor, we do it 
	// every time	
	//
	// This is rather hacky - we should really have one LightingSolution object
	// per instance - but I've left it here so it's easier to change if we
	// do something completely different....
	if (Tools_Mode_BotanicalCreation::IsLightingEnabled())
	{
	
		m_xLightingSolution.SetPosition((*pxPosition)+m_xCentre);
		m_xLightingSolution.RecalculateLighting(true, false);

		// Get the SH representation of the scene lights
		const Asura_SphericalHarmonics_Order2_RGB& xSH = Asura_GI_Solution::s_xSHForCurrentRender;	

		Asura_RGBA xLighting;
		xLighting.m_fR	= LIGHTING_CLAMP(xSH.m_afRedCoefficients[0])   
			+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[1] * xLightingNormal.x) 
			+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[2] * xLightingNormal.y) 
			+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[3] * xLightingNormal.z);

		xLighting.m_fG	= LIGHTING_CLAMP(xSH.m_afGreenCoefficients[0])
			+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[1] * xLightingNormal.x)
			+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[2] * xLightingNormal.y)
			+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[3] * xLightingNormal.z);

		xLighting.m_fB=   LIGHTING_CLAMP(xSH.m_afBlueCoefficients[0])  
			+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[1] * xLightingNormal.x)
			+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[2] * xLightingNormal.y)
			+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[3] * xLightingNormal.z);

			uBillboardCol=xLighting.ToUIntARGB(true) |0xff000000;
	}


	if (m_pxBillboards && m_pxVertices && m_pIndices)
	{
		int iVertexIndex=0;
		
		// Set up each billboard. This would all be much faster done in a shader.		
		for (int i=0;i<m_iNumBillboards;i++)
		{
			uint32 uCol;
			Tools_Botanicals_Billboard* pxBB=&m_pxBillboards[i];

			// Transform centre position to world space (inc. rotation/scaling)
			Asura_Vector_3 xWorldPos=(*pxPosition)+((pxBB->xPos*fScale)*(*pxOrientation));			

			if (uFlags&&TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTRED)
			{
				uCol=0xffff0000;
			}
			else if (uFlags&&TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTBLUE)
			{
				uCol=0xff0000ff;
			}
			else
			{
				if (Tools_Mode_BotanicalCreation::IsLightingEnabled())
				{			
					uCol=uBillboardCol;
				}
				else
				{
					uCol=0xff808080;
				}
			}

			// Set the colour and (world space) position for each vertex
			for(int j=0;j<4;j++)
			{
				const float afAmountOfRight[4] = {-0.5f,0.5f,0.5f,-0.5f};
				const float afAmountOfUp[4]    = {0.5f,0.5f,-0.5f,-0.5f};

				D3DGenericVertex *pxVertex=&m_pxVertices[iVertexIndex];

				Asura_Vector_3 xPos = xWorldPos+ 
									(xBillboardUp*afAmountOfUp[j]*pxBB->fHeight*fScale)+
									(xBillboardRight*afAmountOfRight[j]*pxBB->fWidth*fScale);

				pxVertex->X=xPos.x;
				pxVertex->Y=xPos.y;
				pxVertex->Z=xPos.z;		
				pxVertex->Diffuse=uCol;
				iVertexIndex++;
			}
		}


		// Select the texture		
		Asura_Render::CurrentStates.RequestTextureIndex(0,m_iTextureID);								

		// Render them
		Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS_2X);
		Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
		Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);

		int nFade=(int)((1.0f-fFade)*127.0f);
		Asura_Render::CurrentStates.RequestAlphaTest( true , 0x80+nFade);
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);	
				
		Tools_BotanicalRenderable_StartFoggedRendering();
		Asura_Direct3D::D3DDevice->DrawIndexedPrimitiveUP( D3DPT_TRIANGLESTRIP, 0, m_iNumBillboards*4, (m_iNumBillboards*6)-2, m_pIndices, D3DFMT_INDEX16, (void*) m_pxVertices, sizeof( m_pxVertices[0]));
		Tools_BotanicalRenderable_EndFoggedRendering();

		Asura_Render::CurrentStates.RequestAlphaTest(false);		
	}
}


// ------------------------------------------------------------------------------------------------
// Imposters Renderable
// ------------------------------------------------------------------------------------------------

Tools_BotanicalRenderable_Imposters::Tools_BotanicalRenderable_Imposters(Tools_BotanicalElementLOD *pxLOD)
								    :m_xLightingSolution(true,false)
{
	m_pxLOD=pxLOD;
}


Tools_BotanicalRenderable_Imposters::~Tools_BotanicalRenderable_Imposters()
{
}


void Tools_BotanicalRenderable_Imposters::Render(float fFade, Asura_Vector_3 *pxPosition, Asura_Matrix_3x3 *pxOrientation, float fScale,uint32 uFlags)
{
	SCOPEDPIXEVENT((0xff00ff00, L"(Tools) Render Botanical Imposters"));

	uint32 uBillboardCol;

	// Fix for disappearing imposters in BotanicalCreation mode when fog is enabled
	if (Asura_Fog::IsFogRenderable())
	{
		D3DXMATRIX matWorld;
		D3DXMatrixTranslation( &matWorld, 0,0,0 );
		Tools_BotanicalRenderable_SetWorldMatrix(matWorld);	
	}

	// Set renderstates
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS_2X);
	Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);
	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);	

	Asura_Vector_3 xBillboardUp=Asura_Vector_3(0,-1,0);
	Asura_Vector_3 xFwd;
	Asura_Camera::GetForwardVector(xFwd);

	Asura_Vector_3 xLightingNormal=-xFwd;		

	xBillboardUp=xBillboardUp*(*pxOrientation);	
	Asura_Vector_3 xBillboardRight=xBillboardUp^xFwd;
	xBillboardRight.Normalise();

	if (Tools_Mode_BotanicalCreation::IsLightingEnabled())
	{

		// Update the lighting
		m_xLightingSolution.SetPosition((*pxPosition));
		m_xLightingSolution.RecalculateLighting(true, false);

		// Get the SH representation of the scene lights
		const Asura_SphericalHarmonics_Order2_RGB& xSH = Asura_GI_Solution::s_xSHForCurrentRender;	

		Asura_RGBA xLighting;
		xLighting.m_fR	= LIGHTING_CLAMP(xSH.m_afRedCoefficients[0])   
			+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[1] * xLightingNormal.x) 
			+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[2] * xLightingNormal.y) 
			+ LIGHTING_CLAMP(xSH.m_afRedCoefficients[3] * xLightingNormal.z);

		xLighting.m_fG	= LIGHTING_CLAMP(xSH.m_afGreenCoefficients[0])
			+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[1] * xLightingNormal.x)
			+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[2] * xLightingNormal.y)
			+ LIGHTING_CLAMP(xSH.m_afGreenCoefficients[3] * xLightingNormal.z);

		xLighting.m_fB=   LIGHTING_CLAMP(xSH.m_afBlueCoefficients[0])  
			+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[1] * xLightingNormal.x)
			+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[2] * xLightingNormal.y)
			+ LIGHTING_CLAMP(xSH.m_afBlueCoefficients[3] * xLightingNormal.z);

		uBillboardCol = xLighting.ToUIntARGB(true);
	}

	// For each imposter poly
	for(Asura_ListElt<Tools_Botanicals_Imposter*>*pxElt=m_pxLOD->m_xImposterList.Front();pxElt;pxElt=pxElt->Next())
	{
		D3DGenericVertex xVerts[4];

		Tools_Botanicals_Imposter *pxImposter = pxElt->Data();

		if (pxImposter->m_bValid)
		{	
			uint32 uCol;
			
#ifdef ASURA_BOTANICALS_FLAT_LIGHTING
			Asura_Vector_3 xNormal=Asura_Vector_3(0,-1,0);
#else
			Asura_Vector_3 xNormal= *Asura_Camera::GetPosition();
			xNormal=xNormal-*pxPosition;
			xNormal.Normalise();				
#endif

			// Light it
			if (uFlags&&TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTRED)
			{
				uCol=0xffff0000;
			}
			else if (uFlags&&TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTBLUE)
			{
				uCol=0xff0000ff;
			}
			else
			{			
				if (Tools_Mode_BotanicalCreation::IsLightingEnabled())
				{			
					uCol=uBillboardCol |0xff000000;
				}
				else
				{
					uCol=0xff808080;
				}
			}

			Tools_Botanicals_Billboard *pxBB=&pxImposter->m_xBillboard;
			
			// Create the billboard verts
			for(int i=0;i<4;i++)
			{
				const float afAmountOfUp[4]={1,1,0,0};
				const float afAmountOfRight[4]={-0.5f,0.5f,0.5f,-0.5f};
				const int   aiUSelect[4]={0,1,1,0};
				const int   aiVSelect[4]={0,0,1,1};

				Asura_Vector_3 xPos=(*pxPosition)+pxBB->xPos
									+(xBillboardUp*pxBB->fHeight*fScale*afAmountOfUp[i])
									+(xBillboardRight*pxBB->fWidth*fScale*afAmountOfRight[i]);

				xVerts[i].X=xPos.x;
				xVerts[i].Y=xPos.y;
				xVerts[i].Z=xPos.z;
				xVerts[i].Diffuse=uCol;
				xVerts[i].U = pxBB->fU[aiUSelect[i]];
				xVerts[i].V = pxBB->fV[aiVSelect[i]];
			}


			// Calculate the visibility/fade for this imposter 			
			xNormal=pxImposter->m_xBillboard.xNormal* (*pxOrientation);
			xNormal.y=0;
			xNormal.Normalise();

			Asura_Vector_3 xToCamera= *Asura_Camera::GetPosition()- *pxPosition;
			xToCamera.y=0;
			xToCamera.Normalise();
			
			float fDot= xNormal*xToCamera;			
			float fAng= acos(fDot)/(Asura_Maths::Pi/180.0f);
			float fMax= pxImposter->m_fAngleRange*0.5f;
			float fFadeStart=fMax-pxImposter->m_fFadeRange;

			if (fAng<=fMax)
			{		
				float fImposterFade=1.0f;
				if (fAng>fFadeStart)
				{
					fImposterFade=fAng-fFadeStart;
					fImposterFade=1.0f-(fImposterFade/pxImposter->m_fFadeRange);
				}

				// Set the fade value
				int nFade=(int)((1.0f-(fImposterFade*fFade))*127.0f);
				Asura_Render::CurrentStates.RequestAlphaTest( true , 0x80+nFade);

				// Select the texture		
				Asura_Render::CurrentStates.RequestTextureIndex(0,pxImposter->m_iTextureIndexID);								

				Tools_BotanicalRenderable_StartFoggedRendering();

				// Draw it
				Asura_Direct3D::D3DDevice->DrawPrimitiveUP(D3DPT_TRIANGLEFAN, 2, (void*)xVerts, sizeof(xVerts[0]));			

				Tools_BotanicalRenderable_EndFoggedRendering();
			}
		}

	}

	// Restore states
	Asura_Render::CurrentStates.RequestAlphaTest(false);	
	
}


#endif //ASURA_USE_BOTANICALS