//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE        :  Tools_Mode_BotanicalCreation.cpp
//
//  Created by Dave Reed - 04/01/06
//
//  Copyright (C) 2005 Rebellion (Derby), All Rights Reserved.
//
//  Mode for creating 'BotanicalShapes' - setting up LODs, etc
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined( ASURA_USE_BOTANICALS) && defined( ASURA_TOOLS)

#include "Core/Asura_UserInput.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"
#include "FileAccess/Asura_FileAccess.h"
#include "ChunkSystem/Asura_Chunk_Botanicals_Definition.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Render.h"
#include "Misc/Tools_MessageLog.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Render_Options.h"
#include "Tools_Overlays.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Movement/Tools_Movement.h"
#include "Camera/Asura_Camera.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_CommandConsole.h"
#include "Interface/Asura_Menubar_System.h"
#include "Interface/Asura_Menubar.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Dialog.h"
#include "Load&Save/Tools_SaveAsura.h"

#include "Shapes/Asura_Shape_System.h"
#include "Shapes/Asura_Shapes.h"

#include "BotanicalCreation/Tools_BotanicalCreation_BotanicalShape.h"
#include "BotanicalCreation/Tools_Mode_BotanicalCreation.h"
#include "BotanicalCreation/Tools_BotanicalCreation_EditLODs.h"
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
#include "BotanicalCreation/Tools_BotanicalCreation_BasicInteractive.h"
#endif
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
#include "BotanicalCreation/Tools_BotanicalCreation_StateTables.h"
#endif

#include "Botanical/Asura_Direct3D_Botanicals_Game_Render.h"
#include "Botanicals/Asura_Botanicals_InstanceManager.h"

#include "Tools_System.h"

static Asura_Vector_3   g_xBotanicalCreationModeOrigin(0,0,0);
static Asura_Vector_3   g_xBotanicalCreationModePosition(0,-1,-2);
static Asura_Matrix_3x3 g_xBotanicalCreationModeOrientation(1,0,0, 0,1,0, 0,0,1);

float			Tools_Mode_BotanicalCreation::s_fLODSliderValue=1.0f;
float			Tools_Mode_BotanicalCreation::s_fLODSliderMin=0.0f;
float			Tools_Mode_BotanicalCreation::s_fLODSliderMax=1.0f;
Asura_Vector_2	Tools_Mode_BotanicalCreation::s_xPreviewOverlayPosition;
Asura_Vector_2	Tools_Mode_BotanicalCreation::s_xMainOverlayPosition=Asura_Vector_2(4,4);
bool			Tools_Mode_BotanicalCreation::s_bShowChecks[3]={true,false,false};
bool			Tools_Mode_BotanicalCreation::s_bShowChecksChanged[3]={false,false,false};
bool            Tools_Mode_BotanicalCreation::s_bManuallySetLOD=false;
bool			Tools_Mode_BotanicalCreation::s_bMainOverlayOpen=false;
Asura_BotanicalShape* Tools_Mode_BotanicalCreation::s_pxCurrentBotanicalShape=NULL;

bool			Tools_Mode_BotanicalCreation::s_bUseLocalCamera=true;
bool			Tools_Mode_BotanicalCreation::s_bCameraModeChanged=false;
bool			Tools_Mode_BotanicalCreation::s_bLightingEnabled=true;

void Tools_Mode_BotanicalCreation::Initialise(void)
{
	Asura_Direct3D_Botanicals_Game_Render::ClearDataSet();

	Tools_Render_Options::HideEnvironment = true;	

	// We can't (without changing engine code) turn off rendering of the environment.
	// So we copy a 'design pattern' used in Hierarchy mode.... and move the camera to
	// an empty space a long way from the game world!...	
	m_xPreModeCameraPosition    = Tools_Movement::GetCameraObject()->Position;
	m_xPreModeCameraOrientation = Tools_Movement::GetCameraObject()->Orientation;

	s_bUseLocalCamera=true;
	s_bCameraModeChanged=true;

	// Create the overlays
	s_bMainOverlayOpen=false;	
	
	CreatePreviewControlsOverlay();
	CreateMainOverlay();

	
}


Asura_Vector_3 Tools_Mode_BotanicalCreation::GetOrigin() 
{
	return g_xBotanicalCreationModeOrigin;
}


void Tools_Mode_BotanicalCreation::CreateMainOverlay()
{
	if (s_bMainOverlayOpen)
	{
		CloseMainOverlay();
	}

	Asura_Overlay* pxNewOverlay;
	Asura_Dialog*  pxNewDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
	
	pxNewOverlay = new Asura_Overlay("Botanical Creation",false,&s_xMainOverlayPosition);
	
	// Add main menu options
	pxNewDialog = new Asura_Function_Dialog("Create from Shape file (.asr)",xOffset,CreateFromShapeFile);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=1.5f*Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_Function_Dialog("Load BotanicalShape (.btd)",xOffset,LoadBotanicalShape);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=1.5f*Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_Function_Dialog("Edit LODs", xOffset, CreateLODsOverlay);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=1.5f*Asura_Dialog::DefaultSpacing();

#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	pxNewDialog = new Asura_Function_Dialog("Edit Basic Physics", xOffset, CreateBasicInteractiveOverlay);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=1.5f*Asura_Dialog::DefaultSpacing();
#endif

#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	pxNewDialog = new Asura_Function_Dialog("Edit State Tables", xOffset, CreateStateTableOverlay);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=1.5f*Asura_Dialog::DefaultSpacing();
#endif

	pxNewDialog = new Asura_Tickbox_Dialog("Single Shape View", xOffset, &s_bUseLocalCamera, &s_bCameraModeChanged);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=1.5f*Asura_Dialog::DefaultSpacing();

	//submit the overlay
	pxNewOverlay->m_uID = TOOLS_OVERLAY_BOTANICALCREATION_MAIN;
	Asura_Overlay_System::SubmitNewOverlay(pxNewOverlay,false, true, false);

	s_bMainOverlayOpen=true;	
	
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	Tools_BotanicalCreation_BasicInteractive::SetMode( Tools_BotanicalCreation_BasicInteractive::MODE_NORMAL);
#endif
}


void Tools_Mode_BotanicalCreation::CloseMainOverlay()
{
	if (s_bMainOverlayOpen)
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_BOTANICALCREATION_MAIN);
		s_bMainOverlayOpen=false;			
	}
}

void Tools_Mode_BotanicalCreation::CreateLODsOverlay()
{
	if (GetCurrentBotanicalShape()!=NULL)
	{	
		if (GetCurrentBotanicalShape()->GetToolsShape()!=NULL)
		{		
			Tools_SaveAsura::SetBotanicalCreationModified(true);
			Tools_BotanicalCreation_EditLODs::CreateOverlay();	
			return;
		}
	}

	// Nothing to edit
	Message_Overlay xDlg("Nothing to edit! (Create/Load a BotanicalShape first...)");
	xDlg.DoModal();
}

void Tools_Mode_BotanicalCreation::CloseLODsOverlay()
{	
	Tools_BotanicalCreation_EditLODs::CloseOverlay();		
}

#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
void Tools_Mode_BotanicalCreation::CreateBasicInteractiveOverlay( void)
{
	Asura_BotanicalShape* pxCurrentBotanicalShape = GetCurrentBotanicalShape();
	if( NULL != pxCurrentBotanicalShape)
	{	
		if( pxCurrentBotanicalShape->GetToolsShape()!=NULL)
		{				
			Tools_BotanicalCreation_BasicInteractive::CreateOverlay();	
			return;
		}
	}

	// Nothing to edit
	Message_Overlay xDlg("Nothing to edit! (Create/Load a BotanicalShape first...)");
	xDlg.DoModal();
}

void Tools_Mode_BotanicalCreation::CloseBasicInteractiveOverlay( void)
{
//	Tools_BotanicalCreation_BasicInteractive::CloseOverlay();		
	Tools_BotanicalCreation_BasicInteractive::SetMode( Tools_BotanicalCreation_BasicInteractive::MODE_LITE);			
}
#endif

#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
void Tools_Mode_BotanicalCreation::CreateStateTableOverlay( void)
{
	if (GetCurrentBotanicalShape()!=NULL)
	{	
		if (GetCurrentBotanicalShape()->GetToolsShape()!=NULL)
		{	
			Tools_SaveAsura::SetBotanicalCreationModified(true);
			Tools_BotanicalCreation_StateTables::CreateOverlay();	
			return;
		}
	}

	// Nothing to edit
	Message_Overlay xDlg("Nothing to edit! (Create/Load a BotanicalShape first...)");
	xDlg.DoModal();
}

void Tools_Mode_BotanicalCreation::CloseStateTableOverlay( void)
{
	Tools_BotanicalCreation_StateTables::CloseOverlay();		
}
#endif


void Tools_Mode_BotanicalCreation::CreatePreviewControlsOverlay(void)
{
	// Create the numeric LOD slider 
	//
	// Create the buttons 
	// - All / Element / Lod	
	Asura_Overlay* pxNewOverlay;
	Asura_Dialog*  pxNewDialog;
	Asura_Vector_2 xOffset;

	//create an overlay at the bottom of the screen for controlling LOD preview
	s_xPreviewOverlayPosition.x = 0;
	s_xPreviewOverlayPosition.y = (Asura_Viewport::GetHeight()/14.0f)-4.0f;
	pxNewOverlay = new Asura_Overlay("LOD Control",false,&s_xPreviewOverlayPosition);
	pxNewOverlay->m_bOnPanel = true;


	//create the LOD slider control
	xOffset.x = 0;
	xOffset.y = 0;
	float slideWidth = (Asura_Viewport::GetWidth()/10.0f);
	pxNewDialog = new Asura_NumberSlider_Dialog(xOffset,&s_fLODSliderValue,&s_fLODSliderMin,&s_fLODSliderMax,slideWidth,"%.03f");
	pxNewDialog->m_iControlID = DialogID_LODSlider;
	pxNewOverlay->AddDialog(pxNewDialog);

	xOffset.x = 3.0f;
	xOffset.y = 1.7f;

	pxNewDialog = new Asura_Dialog("LOD Preview:",xOffset);
	xOffset.x+=11;
	pxNewOverlay->AddDialog(pxNewDialog);

	// Create the checkboxes showing which parts to display
	for(int n=0;n<3;n++)
	{	
		char *szNames[]={"Show All Elements","Show Current Element","Show Current LOD"};
		pxNewDialog = new Asura_Tickbox_Dialog(szNames[n],xOffset,&s_bShowChecks[n],&s_bShowChecksChanged[n]);
		pxNewOverlay->AddDialog(pxNewDialog);
		xOffset.x+=20;
	}

	pxNewDialog = new Asura_Tickbox_Dialog("Manual LOD Slider",xOffset,&s_bManuallySetLOD);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.x+=20;

	//submit the overlay
	pxNewOverlay->m_uID = TOOLS_OVERLAY_BOTANICALCREATION_PREVIEW;
	Asura_Overlay_System::SubmitNewOverlay(pxNewOverlay,false, true, false);
}


// Temporary calculation for Duncan to replace... Min LOD at 100 metres, linear scale
float Tools_Mode_BotanicalCreation::CalculateLOD(Asura_Vector_3 *pxPos, float fInvMaxDrawDistance)
{		
	float fLod=1.0f;
	Asura_Vector_3 *pxCameraPos=(Asura_Vector_3*)Asura_Camera::GetPosition();
	Asura_Vector_3 xDist= *pxCameraPos-*pxPos;

	// Avoid divide by zeroes
	if (xDist.x==0.0f && xDist.y==0.0f && xDist.z==0.0f)
	{
		return 1.0f;
	}

	float fDist=xDist.Magnitude();

	fLod = 1.0f - (fInvMaxDrawDistance * fDist);

	// DaveR - 21/06/06 - Clamp to zero
	if (fLod<0.0f) 
	{
		fLod=0.0f;
	}

	return fLod;
}


		
void Tools_Mode_BotanicalCreation::SetRenderStateToShowAllElements()
{
	s_bShowChecks[0]=true;
	for(int n=1;n<3;n++)
	{
		s_bShowChecks[n]=false;
	}
}

void Tools_Mode_BotanicalCreation::Update(void)
{

	if (s_bCameraModeChanged)
	{	
		if (s_bUseLocalCamera)
		{		
			Tools_Movement::TeleportCameraToPosition(&g_xBotanicalCreationModePosition,&g_xBotanicalCreationModeOrientation);
		}
		else
		{
			Tools_Movement::TeleportCameraToPosition(&m_xPreModeCameraPosition,&m_xPreModeCameraOrientation);			
		}
		s_bCameraModeChanged=false;
	}

	if (s_bUseLocalCamera)
	{
		g_xBotanicalCreationModePosition    = Tools_Movement::GetCameraObject()->Position;
		g_xBotanicalCreationModeOrientation = Tools_Movement::GetCameraObject()->Orientation;
	}
	else
	{
		m_xPreModeCameraPosition    = Tools_Movement::GetCameraObject()->Position;
		m_xPreModeCameraOrientation = Tools_Movement::GetCameraObject()->Orientation;
	}



	Tools_BotanicalCreation_EditLODs::Update();
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	Tools_BotanicalCreation_BasicInteractive::Update();
#endif
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	Tools_BotanicalCreation_StateTables::Update();
#endif

	// Calculate the LOD (0-1 value). Temporary calculation for Duncan to replace...
	if((!s_bManuallySetLOD) && (s_pxCurrentBotanicalShape))
	{
		float fInvDist = s_pxCurrentBotanicalShape->GetToolsShape()->m_fInvMaxDrawDistance;
		s_fLODSliderValue=CalculateLOD(&g_xBotanicalCreationModeOrigin, fInvDist);

	}


	

#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	// Simulate radio buttons with checkboxes
	for(int n=0;n<3;n++)
	{
		if (s_bShowChecksChanged[n])
		{
			s_bShowChecksChanged[n]=false;

			if (s_bShowChecks[n]==false)
			{
				s_bShowChecks[n]=true;
			}
			else
			{
				for(int i=0;i<3;i++)
				{
					if (i!=n)
					{				 
						s_bShowChecks[i]=false;
					}
				}
			}
			
			Tools_BotanicalCreation_StateTables::ClearSectionVersionVisibleFilters();
		}
	}
#endif	//defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
}



void Tools_Mode_BotanicalCreation::RenderUI(void)
{
	// Render a floor grid (1m squares)
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	for(int n=-50;n<=50;n+=1)
	{
		float f=(float)n;
		f=0.2f*f;
		
		uint32  uCol=0x10ffffff;
		
		if ((n%5)==0)
		{
			uCol=0x40ffffff;
		}

		if (n==0)
		{
			uCol=0x8000ff00;
		}		

		Asura_Vector_3 x1,x2;  // <- the sort of case where v would be much better than arbitrary x as a prefix for vectors...

		x1=g_xBotanicalCreationModeOrigin;
		x2=g_xBotanicalCreationModeOrigin;
		x1.x-=10.0f;
		x2.x+=10.0f;
		x1.z+=f;
		x2.z+=f;
		Asura_Render::RenderLine(x1,x2,uCol);

		x1=g_xBotanicalCreationModeOrigin;
		x2=g_xBotanicalCreationModeOrigin;
		x1.z-=10.0f;
		x2.z+=10.0f;
		x1.x+=f;
		x2.x+=f;
		Asura_Render::RenderLine(x1,x2,uCol);
	}
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	
	// Lines are now batched - so flush them
	Asura_Render::FlushDisplay();
	
	// Render the object (or selected elements)

	if (s_pxCurrentBotanicalShape)
	{
		Tools_BotanicalShape *pxShape=s_pxCurrentBotanicalShape->GetToolsShape();

		if (pxShape)
		{
			int iElement=0;
			int iLOD=0;
			
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
//			uint32 uSectionVersionIndex = 0;
#endif

			const u_int uSectionCount = pxShape->m_xHashTree_xSections.GetNumberOfNodes();
			for( u_int uSectionIndex = 0; uSectionIndex < uSectionCount; uSectionIndex += 1)
			{
				Tools_BotanicalSection* pxToolSection = pxShape->m_xHashTree_xSections.GetDataByIndex( uSectionIndex);
				
				const u_int uVersionCount = pxToolSection->m_xHashTree_xVersions.GetNumberOfNodes();
				for( u_int uVersionIndex = 0; uVersionIndex < uVersionCount; uVersionIndex += 1)
				{
					{											
						Tools_BotanicalVersion* pxToolVersion = pxToolSection->m_xHashTree_xVersions.GetDataByIndex( uVersionIndex);
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
						if( false == Tools_BotanicalCreation_StateTables::IsSectionVersionVisible( pxToolSection, pxToolVersion))
						{
							continue;
						}

						
						const Asura_CollisionMesh *pxCollMesh = Tools_BotanicalCreation_StateTables::GetVisibleSectionVersionCollisionMesh( pxToolSection, pxToolVersion);

						if (pxCollMesh)
						{
							// Render it
							uint32 uPolys = pxCollMesh->GetNumberOfPolys();

							for(uint32 uPoly=0; uPoly<uPolys; uPoly++)
							{
								Asura_Vector_3 axVerts[4];
								uint32 uVertCount = pxCollMesh->GetNumVerticesInPoly(uPoly);
								pxCollMesh->GetVertices(axVerts,uPoly);

								for(uint32 uVert=0; uVert<uVertCount; uVert++)
								{
									Asura_Render::RenderLine(axVerts[uVert], axVerts[(uVert+1)%uVertCount], 0xffffff00);
								}
							}
						}						
						
#endif
						for(Asura_ListElt<Tools_BotanicalElement*>* pxEltElt = pxToolVersion->m_xList_pxElements.Front(); pxEltElt;pxEltElt=pxEltElt->Next())
						{
							Tools_BotanicalElement *pxElement=pxEltElt->Data();
							iLOD=0;

							for(Asura_ListElt<Tools_BotanicalElementLOD*>* pxLODElt=pxElement->m_xLODList.Front(); pxLODElt;pxLODElt=pxLODElt->Next())
							{
								bool bDontRender=false;
								Tools_BotanicalElementLOD *pxLOD=pxLODElt->Data();
								
						
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
								if (Tools_BotanicalCreation_EditLODs::IsActive() && false == Tools_BotanicalCreation_StateTables::IsSectionVersionVisibleFiltered())
								{
									if (!s_bShowChecks[0] && Tools_BotanicalCreation_EditLODs::GetSelectedElement()!=iElement)
									{
										bDontRender=true;
									}

									if (s_bShowChecks[2] && Tools_BotanicalCreation_EditLODs::GetSelectedLOD()!=iLOD)
									{
										bDontRender=true;
									}
								}
								//static bool GetSelectedElement() {return s_iSelectedElement;};
								//static bool GetSelectedLOD()     {return s_iSelectedLOD;};
#endif	// defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES

								if (!bDontRender)
								{
									float fLOD=GetCurrentLODValue();

									pxLOD->Render(pxElement->m_eType,fLOD,&g_xBotanicalCreationModeOrigin,(Asura_Matrix_3x3 *)&Asura_Maths::IdentityMatrix3x3,1.0f, 0);
								}
													
								iLOD++;
							}

							iElement++;
						}
					}
				}
			}
		}
	}

	// This renders the imposter angle ranges
	Tools_BotanicalCreation_EditLODs::Render();
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	Tools_BotanicalCreation_BasicInteractive::Render();
#endif
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	Tools_BotanicalCreation_StateTables::Update();
#endif
}


void Tools_Mode_BotanicalCreation::Shutdown(void)
{
	Tools_Render_Options::HideEnvironment = false;

	// Force a reload - in case we've said no to 'save changes'
	ReloadShape();

	// Deselect the current shape
	// (Removed, as we don't need to any more - we can keep it selected
	//  whilst using other modes(
//	s_pxCurrentBotanicalShape=NULL;
	
	// Put the camera back where it was before entering this mode
	Tools_Movement::TeleportCameraToPosition(&m_xPreModeCameraPosition,&m_xPreModeCameraOrientation);

	// Remove the preview controls bar
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_BOTANICALCREATION_PREVIEW);

	// Close the regular overlays
	CloseMainOverlay();
	CloseLODsOverlay();
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	CloseBasicInteractiveOverlay();
#endif
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	CloseStateTableOverlay();
#endif
	
	Asura_Direct3D_Botanicals_Game_Render::MakeDataSet( Asura_Botanicals_InstanceManager::GetInstanceList());

}

void Tools_Mode_BotanicalCreation::PreShutdown()
{
/*
	//Prompt for saving before leaving the mode , since this won't be picked up by Tools_SaveAsura::IsModified() at the moment
	//RAR 12/07/2007
	if(s_pxCurrentBotanicalShape)
	{
		if(ShowYesNoOverlay("Save changes to current BotanicalShape?"))
		{
			Save();			
		}
	}
*/
}



void Tools_Mode_BotanicalCreation::CreateFromShapeFile(void)
{
	// Prompt to save changes
	if (!SaveChanges()) return;

	// File open dialog
	Tools_FileSystem_DialogInfo xInfo;

	xInfo.DefaultExtension = "";
	xInfo.DefaultFilename  = ""; 
	xInfo.DefaultPath = "Objects\\";
	xInfo.FileFilter = "Shape Files\0*.asr\0";
	xInfo.WindowTitle = "Load Shape...";
	xInfo.AllowMultipleSelect = false;
	if (!xInfo.DefaultFilename) xInfo.DefaultFilename = "";

	// Get the name, and strip the extension
	Asura_String name = Tools_FileSystem::GetLoadFilename(xInfo);

	if(name.IsNull())
	{
		// Cancelled
		return;
	}

	
	// Call the main Create function
	Asura_BotanicalShape *pxShape=CreateBotanicalShape(name.GetString(),true, true);

	// Clean this up - make the current shape an Asura_BotanicalShape
	s_pxCurrentBotanicalShape=pxShape;

	// Need to do this whenever we reload the shape
	Tools_BotanicalCreation_EditLODs::Reset();
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	Tools_BotanicalCreation_BasicInteractive::Reset();
#endif
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	Tools_BotanicalCreation_StateTables::Reset();
#endif
}



void Tools_Mode_BotanicalCreation::LoadBotanicalShape(void)
{
	// Prompt to save changes
	if (!SaveChanges()) return;

	// File open dialog
	Tools_FileSystem_DialogInfo xInfo;

	xInfo.DefaultExtension = "";
	xInfo.DefaultFilename  = ""; 
	xInfo.DefaultPath = "Objects\\";
	xInfo.FileFilter = "BotanicalShape Definitions\0*.btd\0";
	xInfo.WindowTitle = "Load BotanicalShape...";
	xInfo.AllowMultipleSelect = false;
	if (!xInfo.DefaultFilename) xInfo.DefaultFilename = "";

	// Get the name, and strip the extension
	Asura_String name = Tools_FileSystem::GetLoadFilename(xInfo);

	if(name.IsNull())
	{
		// Cancelled
		return;
	}

	// Call the main Create function
	Asura_BotanicalShape *pxShape=CreateBotanicalShape(name.GetString(),true, false);

	// Clean this up - make the current shape an Asura_BotanicalShape
	s_pxCurrentBotanicalShape=pxShape;

	// Need to do this whenever we reload the shape
	Tools_BotanicalCreation_EditLODs::Reset();
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	Tools_BotanicalCreation_BasicInteractive::Reset();
#endif
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	Tools_BotanicalCreation_StateTables::Reset();
#endif
}




bool Tools_Mode_BotanicalCreation::SaveChanges(void)
{
	if (s_pxCurrentBotanicalShape)
	{	
		// Nothing to edit
		YesNoCancel_Overlay xDlg("Save changes to current BotanicalShape? (You will lose any changes if you click 'no')");
		if (!xDlg.DoModal())
		{
			return false;
		}
		else
		{
			if(xDlg.GetResult())
			{
				if( false == Save())
				{					
					Message_Overlay xDlg("Failed to Save the current BotanicalShape");
					
					xDlg.DoModal();
					
					return( false);
				}
			}
			else
			{
				ReloadShape();
			}
		}
	}
	return true;
}


void Tools_Mode_BotanicalCreation::ReloadShape(void)
{
	// This is needed to undo any changes to BotanicalShapes that we're using in placement
	// mode, if we've changed them but not saved
	if (s_pxCurrentBotanicalShape)
	{		
		Tools_BotanicalShape* pxToolsShape=s_pxCurrentBotanicalShape->GetToolsShape();
		CreateBotanicalShape( pxToolsShape->GetName(),true,false);		
	}
}


const char *Tools_Mode_BotanicalCreation::GetCurrentDefinitionFilename(void)
{
	Tools_BotanicalShape* pxToolsShape=s_pxCurrentBotanicalShape->GetToolsShape();
	return pxToolsShape->GetName();
}


bool Tools_Mode_BotanicalCreation::Save(void)
{

	if (s_pxCurrentBotanicalShape)
	{
		Tools_BotanicalShape* pxToolsShape=s_pxCurrentBotanicalShape->GetToolsShape();
		
		if (pxToolsShape)
		{
			bool bResult=Asura_Chunk_Botanicals_Definition::Create( GetCurrentDefinitionFilename() );

			return bResult;
		}				
	}

	return true;
}


void Tools_Mode_BotanicalCreation::QuickSave(void)
{	
	Tools_SaveAsura::SaveBotanicalCreation();
}



void Tools_Mode_BotanicalCreation::PreSave()
{		
}


void Tools_Mode_BotanicalCreation::PostSave()
{		
}

static bool Tools_Mode_BotanicalCreation_CreateBotanicalShape_IsStringValid( const char* pszBuffer, const char* pszAllowed)
{
	const u_int uStringSize = u_int( strlen( pszBuffer));
	for( u_int uBufferIndex = 0; uBufferIndex < uStringSize; uBufferIndex += 1)
	{
		const char cCharacter = pszBuffer[ uBufferIndex];
		const char szTest[ 2] = { cCharacter, 0};
		if( false == Asura_StringUtil::Contains( pszAllowed, szTest))
		{
			return( false);
		}
	}
	return( true);
}

static const char* Tools_Mode_BotanicalCreation_CreateBotanicalShape_FindNameSection( const char* pszBuffer, const char* pszFindHeader, const char* pszFindFooter, u_int& uLength)
{
	const char* pszFound = NULL;
	
	{			
		int iIndexString = ~0;
		bool bContains = Asura_StringUtil::Contains( pszBuffer, pszFindHeader, &iIndexString);
		if( false == bContains)
		{
			return( NULL);
		}
		
		pszFound = pszBuffer + iIndexString + strlen( pszFindHeader);
	}

	{			
		int iIndexString = ~0;
		bool bContains = Asura_StringUtil::Contains( pszFound, pszFindFooter, &iIndexString);
		if( true == bContains)
		{
			uLength = iIndexString;
		}
		else
		{
			uLength = u_int( strlen( pszFound));
		}
	}

	return( pszFound);
}

Tools_BotanicalElement* Tools_BotanicalVersion::FindElement( const char* pszElementName)
{
	for( Asura_ListElt< Tools_BotanicalElement*> *pxItElement = this->m_xList_pxElements.Front(); NULL != pxItElement; pxItElement = pxItElement->Next())
	{
		Tools_BotanicalElement* pxElement = pxItElement->Data();
		if( NULL == pxElement)
		{
			continue;
		}
		
		if( 0 != pxElement->m_xName.Compare( pszElementName))
		{
			continue;
		}
		
		return( pxElement);
	}
	
	return( NULL);
}

Asura_BotanicalShape* Tools_Mode_BotanicalCreation::CreateBotanicalShape( const char *szFileNameIn, bool bForceReload,bool bFirstTime)
{	
	Asura_BotanicalShape* pxShape=NULL;
	char  szFileNameMinusExtension[256];
	char  szFileToLoad[256];

	Asura_String xStrippedFileName;
	xStrippedFileName = Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir(szFileNameIn);
	char *szFileName = xStrippedFileName.GetString();


	// Strip the extension (could be .asr or .btd at this point)
	int i= int( strlen(szFileName));
	if (i>3 && szFileName[i-4]=='.' && i<256)
	{
		strcpy(szFileNameMinusExtension,szFileName);
		szFileNameMinusExtension[i-4]=0;
	}
	else
	{
		// Keep the current shape open
		return s_pxCurrentBotanicalShape;
	}

	// Does this shape already exist?  If so, we need to clear it's tools representation now
	pxShape=Asura_Botanicals_ShapeManager::GetShapeByHash(Asura_GetHashID(szFileNameMinusExtension));
	
	if (pxShape)
	{
		// Are we reloading it?
		if (!bForceReload)
		{
			return pxShape;
		}

		// Meshes aren't usually discarded on deletion of a BotanicalShape - it's normally
		// done when the mesh system is reset. But to reload meshes, we have to remove the
		// old ones first
		Tools_BotanicalShape* pxOldShape=pxShape->GetToolsShape();

		// If there's already tools data there, delete it, as we're recreating it now
		if (pxOldShape!=NULL)
		{
			pxOldShape->DiscardMeshes();
			delete pxOldShape;
			pxShape->SetToolsShape(NULL);
		}
	}

	// Get the base name (minus extension and path)
	const char* szBaseName=Tools_FileSystem::StripPath(szFileNameMinusExtension);
	
	u_int uMeshesBeforeLoad = Asura_Mesh_System::GetNumberOfMeshes();

#if 0
	Asura_LinkList<Asura_String> xListMeshNamesOriginal;
	for( u_int uMeshIndex = 0; uMeshIndex < uMeshesBeforeLoad; uMeshIndex++)
	{
		// Create list of meshes to discard	
		Asura_Mesh*		pxMesh = Asura_Mesh_System::GetMesh( uMeshIndex );

		xListMeshNamesOriginal.Push_Back( pxMesh->Name);
	}
#endif
	
	// Load the shape
	strcpy(szFileToLoad,szFileNameMinusExtension);
	strcat(szFileToLoad,".asr");
	LoadAsuraFile(szFileToLoad);

	Tools_SaveAsura::SetBotanicalCreationModified(false);

	// I think we need this, to get any new textures to actually load
	Asura_Texture_Management::LoadTextures();

//	Asura_LinkList<Asura_String> xListMeshNamesFiltered;
	
	// Search for the components we expect - name_x_y - and build up the elements. Set default LODs	
	Tools_BotanicalShape* pxBotanicalShape=new Tools_BotanicalShape(szFileNameMinusExtension);

	// Search for the components we expect - name_x_y - and build up the elements. Set default LODs	
	int iElement=0;
	int iTotalLODCount=0;
	
	// DLH - 23 November 2006 12:32:10 - Find ALL meshes that start with the base name and save them
	{	
		Asura_String xMeshNameStart( szBaseName);
		
		xMeshNameStart.MakeLowercase();
		const char* pszMeshNameStart = xMeshNameStart.GetString();
		const u_int uIndexEndOfStart = xMeshNameStart.GetLength();

		u_int uMeshes = Asura_Mesh_System::GetNumberOfMeshes();
		for( u_int uMeshIndex = 0; uMeshIndex < uMeshes; uMeshIndex++)
		{
			Asura_Mesh* pxMesh = Asura_Mesh_System::GetMesh( uMeshIndex );
			
			if( pxMesh->m_iSourceFileType != AsuraFile_Shapes)
			{
				continue;
			}
			
			Asura_String xDecodeNameString = pxMesh->Name;
			
			xDecodeNameString.MakeLowercase();

			Asura_String xNameTestString = xDecodeNameString;

			// Strip LOD suffixes.
			for( u_int u = 0; u < 2; ++u )
			{
				int iLen = xNameTestString.GetLength();
				if( (iLen > 2) && (xNameTestString[iLen - 2] == '_') )
				{
					if(isdigit(xNameTestString[iLen - 1]) || xNameTestString[iLen - 1] == 'i')
					{
						xNameTestString.Truncate( iLen - 2 );
					}
				}
			}

			if( xNameTestString.Compare(pszMeshNameStart) )
			{
				continue;
			}

			/*
			int iIndexStringHeader = ~0;
			//bool bContainsHeader = xDecodeNameString.Contains( pszMeshNameStart, &iIndexStringHeader);
			
			if( false == bContainsHeader || 0 != iIndexStringHeader)
			{
				continue;
			}*/
			
			Asura_String xSectionName( "");
			Asura_String xVersionName( "");
			Asura_String xElementName( "");
			Asura_String xLODName( "");
			
			const char* pszDecodeString = &xDecodeNameString[ uIndexEndOfStart];

			// DLH - 23 November 2006 16:29:04 - Decode the Element_LOD_Section_Version
			{			
				u_int uFoundLength = 0;
				const char* pszFound = Tools_Mode_BotanicalCreation_CreateBotanicalShape_FindNameSection( pszDecodeString, "_", "_", uFoundLength);
				if( NULL == pszFound)
				{
					Asura_Assertf( false, ("Malformed name for Botanical '%s'\n Mesh '%s'\nCannot resolve the Element number.", szBaseName, pxMesh->Name.GetString()));
					continue;
				}
				
				xElementName = pszFound;
				xElementName.Truncate( uFoundLength);
				pszDecodeString = pszFound + uFoundLength;
			}
			
			{			
				u_int uFoundLength = 0;
				const char* pszFound = Tools_Mode_BotanicalCreation_CreateBotanicalShape_FindNameSection( pszDecodeString, "_", "_", uFoundLength);
				if( NULL == pszFound)
				{
					Asura_Assertf( false, ("Malformed name for Botanical '%s'\n Mesh '%s'\nCannot resolve the LOD number.", szBaseName, pxMesh->Name.GetString()));
					continue;
				}
				
				xLODName = pszFound;
				xLODName.Truncate( uFoundLength);
				pszDecodeString = pszFound + uFoundLength;
			}
			
			{			
				u_int uFoundLength = 0;
				const char* pszFound = Tools_Mode_BotanicalCreation_CreateBotanicalShape_FindNameSection( pszDecodeString, "_s", "_", uFoundLength);
				if( NULL != pszFound)
				{
					xSectionName = pszFound;
					xSectionName.Truncate( uFoundLength);
				}
			}
			
			{			
				u_int uFoundLength = 0;
				const char* pszFound = Tools_Mode_BotanicalCreation_CreateBotanicalShape_FindNameSection( pszDecodeString, "_v", "_", uFoundLength);
				if( NULL != pszFound)
				{
					xVersionName = pszFound;
					xVersionName.Truncate( uFoundLength);
				}
			}
			
			{
				const char* pszValidElement = "0123456789i";
				const char* pszValidLOD = "0123456789";
				const char* pszValidCharacters = "0123456789qwertyuiopasdfghjklzxcvbnm";
				bool bIsElementNameValid = Tools_Mode_BotanicalCreation_CreateBotanicalShape_IsStringValid( xElementName.GetString(), pszValidElement);
				bool bIsLODNameValid = Tools_Mode_BotanicalCreation_CreateBotanicalShape_IsStringValid( xLODName.GetString(), pszValidLOD);
				bool bIsSectionNameValid = Tools_Mode_BotanicalCreation_CreateBotanicalShape_IsStringValid( xSectionName.GetString(), pszValidCharacters);
				bool bIsVersionNameValid = Tools_Mode_BotanicalCreation_CreateBotanicalShape_IsStringValid( xVersionName.GetString(), pszValidCharacters);
				if( false == bIsElementNameValid || false == bIsLODNameValid || false == bIsSectionNameValid || false == bIsVersionNameValid)
				{
					Tools_MessageLog::AddMessage( "Botanicals",
												  "Invalid Mesh Name",
												  "Botanical shape '%s': is trying to use mesh '%s' which has an invalid mesh name.",
												  szBaseName,
												  pxMesh->Name
												  );

					Tools_MessageLog::ShowOverlay();
					continue;
				}
			}
			
			// DLH - 23 November 2006 16:28:29 - Get the Section Pointer
			const Asura_Hash_ID xSectionNameHash = xSectionName.GetHash();
			Tools_BotanicalSection* pxSection = pxBotanicalShape->m_xHashTree_xSections.GetDataByHash( xSectionNameHash);
			if( NULL == pxSection)
			{
				pxSection = pxBotanicalShape->m_xHashTree_xSections.AddNode( xSectionNameHash);
				pxSection->m_xName = xSectionName;
			}
			
			// DLH - 23 November 2006 16:28:17 - Get the Version Pointer
			const Asura_Hash_ID xVersionNameHash = xVersionName.GetHash();
			Tools_BotanicalVersion* pxVersion = pxSection->m_xHashTree_xVersions.GetDataByHash( xVersionNameHash);
			if( NULL == pxVersion)
			{
				pxVersion = pxSection->m_xHashTree_xVersions.AddNode( xVersionNameHash);
				pxVersion->m_xName = xVersionName;
			}
			
			// DLH - 23 November 2006 16:27:01 - Get an Element Pointer
			Tools_BotanicalElement* pxElement = pxVersion->FindElement( xElementName.GetString());
			if( NULL == pxElement)
			{
				pxElement = new Tools_BotanicalElement();
				pxElement->m_xName = xElementName;
				iElement += 1;
				
				if( 0 == xElementName.Compare( "i") )
				{
					pxElement->m_eType = TOOLS_BOTANICALS_ELEMENTTYPE_IMPOSTER;
				}
				
				{
					const int iElementValueImposter = 0x7FFFFFF;
					int iElementValue = iElementValueImposter;
					bool bConvertedValueStatus = Asura_StringUtil::GetIntValue( xElementName.GetString(), &iElementValue);
					
					Asura_Assertf( true == bConvertedValueStatus || iElementValueImposter == iElementValue, (" '%s' did not convert to a number but the value has been set to %d. It should be %d", xElementName.GetString(), iElementValue, iElementValueImposter));

#if defined( ASURA_ASSERT)
					int iAssertCheckOrderLastValue = 0;
#endif
					bool bAddedToList = false;
					for(Asura_ListElt<Tools_BotanicalElement*>* pxElt = pxVersion->m_xList_pxElements.Front(); pxElt;pxElt=pxElt->Next())
					{
						Tools_BotanicalElement* pxElementInList = pxElt->Data();
						
						const char* pszElementInListName = pxElementInList->m_xName.GetString();
						int iElementInListValue = iElementValueImposter;
						Asura_StringUtil::GetIntValue( pszElementInListName, &iElementInListValue);
						
#if defined( ASURA_ASSERT)
						Asura_Assert( iAssertCheckOrderLastValue <= iElementInListValue, "Values seem to be out of order.");
						iAssertCheckOrderLastValue = iElementInListValue;
#endif
						
						if( iElementInListValue < iElementValue)
						{
							continue;
						}
						
						pxVersion->m_xList_pxElements.Insert_Before( pxElt, pxElement);
						bAddedToList = true;
						break;
					}
					
					if( false == bAddedToList)
					{
						pxVersion->m_xList_pxElements.Push_Back( pxElement);
					}
					
#if defined( _DEBUG) && defined( ASURA_ASSERT)
					{
						int iAssertCheckOrderLastValue = 0;
						for(Asura_ListElt<Tools_BotanicalElement*>* pxElt = pxVersion->m_xList_pxElements.Front(); pxElt;pxElt=pxElt->Next())
						{
							Tools_BotanicalElement* pxElementInList = pxElt->Data();
							
							const char* pszElementInListName = pxElementInList->m_xName.GetString();
							int iElementInListValue = iElementValueImposter;
							Asura_StringUtil::GetIntValue( pszElementInListName, &iElementInListValue);
							
							Asura_Assert( iAssertCheckOrderLastValue <= iElementInListValue, "Values seem to be out of order.");;
							iAssertCheckOrderLastValue = iElementInListValue;
						}						
					}
#endif
				}
				
			}
			
			// DLH - 23 November 2006 16:27:18 - Get a LOD Pointer
			if( TOOLS_BOTANICALS_ELEMENTTYPE_IMPOSTER == pxElement->m_eType)
			{
				Tools_BotanicalElementLOD* pxLOD = NULL;
				
				{
					for( Asura_ListElt<Tools_BotanicalElementLOD*>* pxElt = pxElement->m_xLODList.Front(); pxElt;pxElt=pxElt->Next())
					{
						Tools_BotanicalElementLOD* pxLODInList = pxElt->Data();
						
						if( NULL == pxLODInList)
						{
							continue;
						}
						pxLOD = pxLODInList;
						break;
					}
						
					if( NULL == pxLOD)
					{
						pxLOD = new Tools_BotanicalElementLOD;
						pxLOD->m_xName = "i";
						iTotalLODCount += 1;
						pxElement->m_xLODList.Push_Back( pxLOD);
					}
				}
				
				pxLOD->AddImposter( pxMesh);
			}
			else
			{
				Tools_BotanicalElementLOD* pxLOD = pxElement->GetLODByName( xLODName.GetString());
				if( NULL == pxLOD)
				{
					pxLOD = new Tools_BotanicalElementLOD;
					pxLOD->m_xName = xLODName;
					iTotalLODCount += 1;
					
					int iLODValue = 0;
					Asura_StringUtil::GetIntValue( xLODName.GetString(), &iLODValue);
					
					bool bAddedToList = false;
					for(Asura_ListElt<Tools_BotanicalElementLOD*>* pxElt = pxElement->m_xLODList.Front(); pxElt;pxElt=pxElt->Next())
					{
						Tools_BotanicalElementLOD* pxLODInList = pxElt->Data();
						
						const char* pszLODInListName = pxLODInList->m_xName.GetString();
						int iLODInListValue = 0;
						Asura_StringUtil::GetIntValue( pszLODInListName, &iLODInListValue);
						
						if( iLODInListValue < iLODValue)
						{
							continue;
						}
						
						pxElement->m_xLODList.Insert_Before( pxElt, pxLOD);
						bAddedToList = true;
						break;
					}
					
					if( false == bAddedToList)
					{
						pxElement->m_xLODList.Push_Back( pxLOD);
					}
					
	#if defined( _DEBUG) && defined( ASURA_ASSERT)
					{
						int iAssertCheckOrderLastValue = 0;
						for(Asura_ListElt<Tools_BotanicalElementLOD*>* pxElt = pxElement->m_xLODList.Front(); pxElt;pxElt=pxElt->Next())
						{
							Tools_BotanicalElementLOD* pxLODInList = pxElt->Data();
							
							const char* pszLODInListName = pxLODInList->m_xName.GetString();
							int iLODInListValue = 0;
							Asura_StringUtil::GetIntValue( pszLODInListName, &iLODInListValue);
							
							Asura_Assert( iAssertCheckOrderLastValue <= iLODInListValue, "Values seem to be out of order.");
							iAssertCheckOrderLastValue = iLODInListValue;
						}						
					}
	#endif
				}
				
				// DLH - 23 November 2006 16:27:56 - Setup the Mesh Data
				Asura_Assertf( NULL == pxLOD->m_pxMesh, ("Botanical '%s' :\n'%s' '%s' '%s' '%s' already seems to have a mesh! '%s'", szBaseName, xSectionName.GetString(), xVersionName.GetString(), xElementName.GetString(), xLODName.GetString(), pxLOD->m_pxMesh->Name.GetString()));
				pxLOD->m_pxMesh = pxMesh;
			}
		}
	}
	

	// Discard any unused meshes
	Asura_LinkList<Asura_String> xDiscardList;
	for( int i = uMeshesBeforeLoad;i<Asura_Mesh_System::GetNumberOfMeshes();i++)
	{
		// Create list of meshes to discard	
		Asura_Mesh*		pxMesh = NULL;
		pxMesh = Asura_Mesh_System::GetMesh( i );

		if (!pxBotanicalShape->ContainsMesh(pxMesh))
		{
			xDiscardList.Push_Back(pxMesh->Name);
		}		
	}

	// Then delete them
	for (Asura_ListElt<Asura_String> *pxElt=xDiscardList.Front(); pxElt; pxElt=pxElt->Next())
	{
		Asura_Mesh*	 pxMesh = Asura_Mesh_System::GetMesh(pxElt->Data());
		if (pxMesh)
		{
			Asura_Mesh_System::DeleteMesh(pxMesh);
		}
	}
	
	// If we've got any valid components, add it to the shape manager, else print warnings
	// and bail out

	if (pxBotanicalShape->GetNumElements()>0)
	{
				
		if (!pxShape)
		{			
			pxShape=Asura_Botanicals_ShapeManager::AddShape(Asura_GetHashID(szFileNameMinusExtension));
		}
		
		pxShape->SetToolsShape(pxBotanicalShape);		
		pxBotanicalShape->SetDefaultLODs();

		if( Tools_System::GetCurrentMode() == TOOLS_MODE_BOTANICALCREATION )
		{
			char szTmp[256];
			sprintf(szTmp,"Botanical shape %s: found %d elements, and a total of %d LOD meshes",szBaseName,iElement,iTotalLODCount);
			Tools_Timed_Message_System::AddMessage(0x00ff00, szTmp);
		}
	}
	else
	{
		Tools_MessageLog::AddMessage( "Botanicals",
									  "Missing Mesh",
									  "Error: botanical shape %s contains no LOD meshes. Mesh names should be ShapeName_X_Y where ShapeName matches the EXPORT filename and X=Element, Y=LOD (both in the range 0-9)",
									  szBaseName );

		Tools_MessageLog::ShowOverlay( "Botanicals", "Missing Mesh" );

		delete pxBotanicalShape;

		// No LODs found
		//Message_Overlay xDlg("Failed - No LOD meshes found\n\nMesh names should be ShapeName_X_Y\nwhere ShapeName matches the EXPORT filename\nand X=Element, Y=LOD (both in the range 0-9)");
		//xDlg.DoModal();

		return NULL;
	}
	

	// If we're reloading one, load the definition file now, to set up the LODs/imposters/etc
	if (!bFirstTime)
	{
		sprintf(szFileToLoad,"%s.btd",szFileNameMinusExtension);
		LoadAsuraFile(szFileToLoad);
	}	

	// Create renderables for each LOD
	pxBotanicalShape->UpdateRenderables();

#if (defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES)
	pxBotanicalShape->ValidateStateTable();
#endif

	return pxShape;
}


bool Tools_Mode_BotanicalCreation::IsLightingEnabled(void)
{
	// Can't use lighting when in single-object mode, as there's
	// no valid GI data
	if (s_bUseLocalCamera && s_bMainOverlayOpen)
	{
		return false;
	}
	else
	{
		return s_bLightingEnabled; 
	}
}




void Tools_Mode_BotanicalCreation::CreateBillboardFromQuad(Tools_Botanicals_Billboard *pxBB,Asura_Mesh *pxMesh,int iPoly)
{
	Asura_Mesh_Polygon *pxPolys=pxMesh->PolygonDataPtr;

	Asura_Vector_3 xCentre=Asura_Vector_3(0,0,0);

	Asura_Vector_3 axVerts[4];

	for(int j=0;j<4;j++)
	{
		Asura_Mesh_Vertex *pxVert=&pxMesh->VertexDataPtr[pxPolys[iPoly].VertexID[j]];												
		xCentre=xCentre+pxVert->Position;												

		axVerts[j] = pxVert->Position;
	}

	pxBB->xPos=xCentre*0.25f;		

	Asura_Vector_3 x1 = axVerts[0]-axVerts[1];
	Asura_Vector_3 x2 = axVerts[2]-axVerts[1];
	x1.Normalise();
	x2.Normalise();
	Asura_Vector_3 xNormal = x1^x2;
	xNormal.Normalise();

	Asura_Vector_3 xRight = xNormal ^ Asura_Vector_3(0,1,0);
	xRight.Normalise();

	float fD = - ((xRight.x * xCentre.x)+(xRight.y * xRight.y)+(xRight.z * xCentre.z));

	float fBestTopLeftScore = 1000000;
	int   iTopLeftIndex = 0;
	
	for(int i=0; i<4; i++)
	{
		float fRightness = (axVerts[i].x * xRight.x) +
						   (axVerts[i].y * xRight.y) +			
						   (axVerts[i].z * xRight.z) + fD;			

		// It's on the left side of the splitting plane
		if (fRightness<0)
		{
			if (axVerts[i].y<fBestTopLeftScore)
			{
				fBestTopLeftScore = axVerts[i].y;
				iTopLeftIndex = i;
			}
		}
	}

	int iTopRightIndex = (iTopLeftIndex+1)&3;
	int iBotRightIndex = (iTopLeftIndex+2)&3;
	int iBotLeftIndex  = (iTopLeftIndex+3)&3;

	Asura_Vector_3 xDist = axVerts[iTopRightIndex] - axVerts[iTopLeftIndex];
	pxBB->fWidth = xDist.Magnitude();

	xDist = axVerts[iBotLeftIndex] - axVerts[iTopLeftIndex];
	pxBB->fHeight = xDist.Magnitude();

	pxBB->fU[0]=pxMesh->TextureDataPtr[iPoly].U[0][iTopLeftIndex];
	pxBB->fV[0]=pxMesh->TextureDataPtr[iPoly].V[0][iTopLeftIndex];
	pxBB->fU[1]=pxMesh->TextureDataPtr[iPoly].U[0][iBotRightIndex];
	pxBB->fV[1]=pxMesh->TextureDataPtr[iPoly].V[0][iBotRightIndex];

}


#endif //ASURA_USE_BOTANICALS
