//////////////////////////////////////////////////////////////////////////////
//
// Tools_Chunk_MaterialResponse.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#include "ChunkSystem/Tools_Chunk_MaterialResponse.h"

// Asura Platform-independent
//#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_List.h"

// Tools
#include "Materials/Tools_MaterialResponse_System.h"
#include "Texturing/Tools_Texturing_MaterialResponse.h"

/******************************************************************************/
/* Tools_Chunk_MaterialResponse Static Class Data                             */
/******************************************************************************/

const u_int Tools_Chunk_MaterialResponse::uCURRENT_VERSION = 26;

Asura_String Tools_Chunk_MaterialResponse::s_xLoadingFilename;

/******************************************************************************/
/* Tools_Chunk_MaterialResponse Class Functions                               */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Tools_Chunk_MaterialResponse::SetLoadingFilename()
//
// Stores the name of the file being currently processed. The chunk loader will
// use this to set the internal name of the loaded materials
//
// Parameters:		szFilename - Filename to use
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Chunk_MaterialResponse::SetLoadingFilename( const char* const szFilename )
{
	s_xLoadingFilename = szFilename;
}

////////////////////////////////////////////////////////////////////////////////
//
// Tools_Chunk_MaterialResponse::Process()
//
// Processes a chunk containing material responses and adds them to the system.
// Use SetLoadingFilename() before this to set the name for the loaded materials
//
// Parameters:		iFileHandle - Handle to file from which to read the chunk
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Tools_Chunk_MaterialResponse::Process( const Asura_Handle iFileHandle )
{
	if( Version < 0 || Version > uCURRENT_VERSION )
	{
		Asura_Assertf( false, ("Material response chunk has invalid version number (%d). You may need new code.", Version) );
		return false;
	}

	if(Version <= 18)
	{
		Asura_Assertf( false, ("This chunk (%d) is too old to work with loading materials on demand.", Version) );
		return false;
	}

	Asura_Chunk_File_Stream xStream( iFileHandle );

	int iNumMaterials;
	xStream >> iNumMaterials;

	//Version 18 and below had two textures per slot ,and so need to go through the fun conversion process.
	//RAR 13/06/2007
	if(Version <= 18)
	{
		for( int i = 0; i < iNumMaterials; ++i )
		{
			// This does assume there's only one material per file
			Asura_String xMaterialName;
			Tools_MaterialResponse_System::MakeMaterialNameFromPath( s_xLoadingFilename, xMaterialName );

			Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddMaterial( xMaterialName.GetHash() );

			if( !pxNewMaterial )
			{
				Asura_Assert( false, "Unable to allocate memory for material." );
				return false;
			}
			//allocate and initialise storage for old texture slots
			pxNewMaterial->m_puOldTextureHashIDs = new Asura_Hash_ID[Tools_MaterialResponse::OLD_NUMBER_OF_TEXTURE_SLOTS];
			pxNewMaterial->m_pxOldTextureFilenames = new Asura_String[Tools_MaterialResponse::OLD_NUMBER_OF_TEXTURE_SLOTS];
			pxNewMaterial->m_puOldTextureAnimHashIDs = new Asura_Hash_ID[2];

			Asura_Mem::Set(pxNewMaterial->m_puOldTextureHashIDs,0,sizeof(Asura_Hash_ID)*Tools_MaterialResponse::OLD_NUMBER_OF_TEXTURE_SLOTS);
			Asura_Mem::Set(pxNewMaterial->m_puOldTextureAnimHashIDs,0,sizeof(Asura_Hash_ID)*2);


			pxNewMaterial->m_xName = xMaterialName;

			if( Version < 11 )
			{
				// (AAM - 09.02.2007) Used to read material name here, but it's redundant since we can
				// use the filename instead. Also, we're using the full path as identifier now
				xStream.SkipString();
			}

			if( Version >= 12 )
			{
				xStream >> pxNewMaterial->m_uParentHash;
			}

			xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_1 ];

			xStream >> pxNewMaterial->m_fAmbientContribution;
			xStream >> pxNewMaterial->m_fColouriseAmbient;
			xStream >> pxNewMaterial->m_xAmbientColour;

			float fDiffuseContribution = 0.0f;
			if( Version < 14 )
			{
				xStream >> fDiffuseContribution;
			}

			pxNewMaterial->m_xDiffuseGraph.ReadFromChunkStream( xStream );
			xStream >> pxNewMaterial->m_fColouriseDiffuse;
			xStream >> pxNewMaterial->m_xDiffuseColour;

			xStream >> pxNewMaterial->m_fMetalicity;

			float fSpecularContribution = 0.0f;
			if( Version < 14 )
			{
				xStream >> fSpecularContribution;
			}

			pxNewMaterial->m_xSpecularGraph.ReadFromChunkStream( xStream );
			xStream >> pxNewMaterial->m_fColouriseSpecular;
			xStream >> pxNewMaterial->m_xSpecularColour;

			float fTransparency = 0.0f;
			if( Version >= 1 )
			{
				if( Version < 14 )
				{
					xStream >> fTransparency;
				}

				pxNewMaterial->m_xTransparencyGraph.ReadFromChunkStream( xStream );
			}

			if( Version >= 2 )
			{
				xStream >> pxNewMaterial->m_fSelfIlluminationContribution;
				xStream >> pxNewMaterial->m_fColouriseSelfIllumination;
				xStream >> pxNewMaterial->m_xSelfIlluminationColour;
			}

			if( Version >= 3 )
			{
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_NORMAL_MAP_1 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_SPECULAR_MAP_1 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_TRANSPARENCY_MAP_1 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_1 ];
			}

			if( Version >= 4 )
			{
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_2 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_NORMAL_MAP_2 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_SPECULAR_MAP_2 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_TRANSPARENCY_MAP_2 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_2 ];
			}

			if ( Version >= 5 )
			{
				// rather do a loop through texture slots here but previous versions of data structure have it in this order
				for( u_int u = 0; u<2; ++u )
				{
					xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_1 + u ];
					xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_NORMAL_MAP_1 + u ];
					xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_SPECULAR_MAP_1 + u ];
					xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_TRANSPARENCY_MAP_1 + u ];
					xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_1 + u ];
				}
			}

			if ( Version >= 6 )
			{
				for( u_int u = 0; u<2; ++u )
				{
					xStream >> pxNewMaterial->m_puOldTextureAnimHashIDs[u];
				}
			}

			if ( Version >= 7 )
			{
				// Can't stream in an enum value, so we're using a temp u_int
				u_int uTemp;
				xStream >> uTemp;
					
				pxNewMaterial->m_eBlendMode = (enum Asura_MaterialResponse::BLEND_MODE) uTemp;
			}

			if ( Version >= 8 )
			{
				xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_1 ];
				xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2 ];
			}

			if ( Version >= 9 )
			{
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_1 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2 ];

				if( Version < 17 )
				{
					// Use the old 'diffuse reflectivity' value for new reflectivity
					xStream >> pxNewMaterial->m_fReflectivity;
				}
			}

			if ( Version >= 10 && Version < 17 )
			{
				// Ditch the old 'specular reflectivity' value
				float fSpecularReflectivity;
				xStream >> fSpecularReflectivity;
			}

			if( Version < 13 )
			{
				// Graphs changed to run [-1,1] instead of [0,1], and discrete samples removed

				pxNewMaterial->m_xDiffuseGraph.SetNumDiscreteSamples(0);
				Asura_Graph_Point< float >* pxControlPoint = pxNewMaterial->m_xDiffuseGraph.GetControlPoints();
				while( pxControlPoint )
				{
					pxControlPoint->m_fTime = pxControlPoint->m_fTime * 2.0f - 1.0f;
					pxControlPoint = pxControlPoint->m_pxNext;
				}

				pxNewMaterial->m_xSpecularGraph.SetNumDiscreteSamples(0);
				pxControlPoint = pxNewMaterial->m_xSpecularGraph.GetControlPoints();
				while( pxControlPoint )
				{
					pxControlPoint->m_fTime = pxControlPoint->m_fTime * 2.0f - 1.0f;
					pxControlPoint = pxControlPoint->m_pxNext;
				}

				pxNewMaterial->m_xTransparencyGraph.SetNumDiscreteSamples(0);
				pxControlPoint = pxNewMaterial->m_xTransparencyGraph.GetControlPoints();
				while( pxControlPoint )
				{
					pxControlPoint->m_fTime = pxControlPoint->m_fTime * 2.0f - 1.0f;
					pxControlPoint = pxControlPoint->m_pxNext;
				}
			}

			if( Version < 14 )
			{
				// Contribution scalars rolled into graph values

				Asura_Graph_Point< float >* pxControlPoint = pxNewMaterial->m_xDiffuseGraph.GetControlPoints();
				while( pxControlPoint )
				{
					pxControlPoint->m_xValue *= fDiffuseContribution;
					pxControlPoint = pxControlPoint->m_pxNext;
				}
				pxNewMaterial->m_xDiffuseGraph.SetClampingValues( true, true, 0.0f, fTOOLS_MATERIAL_RESPONSE_GRAPH_MAX );

				pxControlPoint = pxNewMaterial->m_xSpecularGraph.GetControlPoints();
				while( pxControlPoint )
				{
					pxControlPoint->m_xValue *= fSpecularContribution;
					pxControlPoint = pxControlPoint->m_pxNext;
				}
				pxNewMaterial->m_xSpecularGraph.SetClampingValues( true, true, 0.0f, fTOOLS_MATERIAL_RESPONSE_GRAPH_MAX );

				// Also flipped transparency graph to make it more intuitive

				pxControlPoint = pxNewMaterial->m_xTransparencyGraph.GetControlPoints();
				while( pxControlPoint )
				{
					pxControlPoint->m_xValue = 100.0f - pxControlPoint->m_xValue * fTransparency;
					pxControlPoint = pxControlPoint->m_pxNext;
				}
				// Transparency can only go 0-100%
				pxNewMaterial->m_xTransparencyGraph.SetClampingValues( true, true, 0.0f, 100.0f );
			}

			if ( Version >= 15 )
			{	
				// Special FX properties
				u_int uTemp;
				xStream >> uTemp;
				pxNewMaterial->m_eSpecialFXMode = (enum Asura_MaterialResponse::SPECIALFX_MODE) uTemp;

				xStream >> pxNewMaterial->m_xWaterReflectionColour;
				xStream >> pxNewMaterial->m_xWaterRefractionColour;
				xStream >> pxNewMaterial->m_xWaterFogColour;
				xStream >> pxNewMaterial->m_fWaterFogStrength;
				
				for(int i=0;i<6;i++)
				{
					xStream >> pxNewMaterial->m_afWaterUVControl[i];		
				}

				xStream >> pxNewMaterial->m_fWaterRippleScale;
				xStream >> pxNewMaterial->m_fWaterEdgeSoftness;

				for(int i=0;i<6;i++)
				{
					xStream >> pxNewMaterial->m_afWaterWaveControl[i];		
				}
			}

			if( Version >= 16 )
			{
#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
				xStream >> pxNewMaterial->m_uSurfaceTypeHash;
#else
				xStream >> pxNewMaterial->m_uSurfaceType;
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
			}

			if( Version >= 17 )
			{
				xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_1 ];
				xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_2 ];

				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_1 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_2 ];

				xStream >> pxNewMaterial->m_fReflectivity;
			}
			else if( pxNewMaterial->GetSpecialFXMode() != Asura_MaterialResponse::SPECIALFX_MODE_WATER )
			{
				// Changed to cube maps for non-water reflections. Better ditch the old sphere maps,
				// they just won't work
				pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_1 ] = "";
				pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2 ] = "";

				pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_1 ] = ASURA_HASH_ID_UNSET;
				pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2 ] = ASURA_HASH_ID_UNSET;
			}

			if( Version >= 18 )
			{
				xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_PARALLAX_MAP_1 ];
				xStream >> pxNewMaterial->m_pxOldTextureFilenames[ Tools_MaterialResponse::OLD_PARALLAX_MAP_2 ];

				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_PARALLAX_MAP_1 ];
				xStream >> pxNewMaterial->m_puOldTextureHashIDs[ Tools_MaterialResponse::OLD_PARALLAX_MAP_2 ];

				xStream >> pxNewMaterial->m_fParallaxScale;
			}

			//copy first texture of each type from old values to current values
			for(int iSlot=0;iSlot<Tools_MaterialResponse::OLD_NUMBER_OF_TEXTURE_SLOTS;iSlot+=2)
			{
				int iDestSlot = iSlot/2;

				pxNewMaterial->m_auTextureHashIDs[iDestSlot] = pxNewMaterial->m_puOldTextureHashIDs[iSlot];
				pxNewMaterial->m_axTextureFilenames[iDestSlot] = pxNewMaterial->m_pxOldTextureFilenames[iSlot];
			}
			pxNewMaterial->m_uTextureAnimHashID = pxNewMaterial->m_puOldTextureAnimHashIDs[0];

			// Before the decal flag was added, transparent blend modes just meant decal-type transparency anyway
			if(pxNewMaterial->GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_OPAQUE)
			{
				pxNewMaterial->m_uChunkedFlags	&= ~Tools_MaterialResponse::BITMASK_DECAL;
			}
			else
			{
				pxNewMaterial->m_uChunkedFlags	|= Tools_MaterialResponse::BITMASK_DECAL;
			}
		}
	}
	else
	{
		//Version 19 and onwards , these are after the big material overhaul.
		for( int i = 0; i < iNumMaterials; ++i )
		{
			// This does assume there's only one material per file
			Asura_String xMaterialName;
			Tools_MaterialResponse_System::MakeMaterialNameFromPath( s_xLoadingFilename, xMaterialName );

			Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddMaterial( xMaterialName.GetHash() );

			if( !pxNewMaterial )
			{
				Asura_Assert( false, "Unable to allocate memory for material." );
				return false;
			}

			// KJFL - 26 October 2007 - 
			// If the material has previously been marked as unloaded, the name will already be correct
			// and we shouldn't over-write it. If we're loading on demand midway through something, 
			// the Tools will probably already have pointers to the existing string.
			if ( !pxNewMaterial->m_bIsLoaded )
			{
				pxNewMaterial->m_bIsLoaded = true;
			}
			else
			{
				pxNewMaterial->m_xName = xMaterialName;
			}

			xStream >> pxNewMaterial->m_uParentHash;

			int iSavedTextureSlots;
			xStream >> iSavedTextureSlots;
			Asura_Assert(iSavedTextureSlots<=Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS,"Too many texture slots in saved material");

			for(int iSlot = 0; iSlot < iSavedTextureSlots; iSlot++)
			{
				xStream >> pxNewMaterial->m_auTextureHashIDs[iSlot];
				xStream >> pxNewMaterial->m_axTextureFilenames[iSlot];
			}

			xStream >> pxNewMaterial->m_fAmbientContribution;
			xStream >> pxNewMaterial->m_fColouriseAmbient;
			xStream >> pxNewMaterial->m_xAmbientColour;

			pxNewMaterial->m_xDiffuseGraph.ReadFromChunkStream( xStream );
			xStream >> pxNewMaterial->m_fColouriseDiffuse;
			xStream >> pxNewMaterial->m_xDiffuseColour;

			xStream >> pxNewMaterial->m_fMetalicity;

			pxNewMaterial->m_xSpecularGraph.ReadFromChunkStream( xStream );
			xStream >> pxNewMaterial->m_fColouriseSpecular;
			xStream >> pxNewMaterial->m_xSpecularColour;

			pxNewMaterial->m_xTransparencyGraph.ReadFromChunkStream( xStream );

			xStream >> pxNewMaterial->m_fSelfIlluminationContribution;
			xStream >> pxNewMaterial->m_fColouriseSelfIllumination;
			xStream >> pxNewMaterial->m_xSelfIlluminationColour;

			xStream >> pxNewMaterial->m_uTextureAnimHashID;

			// Can't stream in an enum value, so we're using a temp u_int

			u_int uTemp;
			xStream >> uTemp;
			pxNewMaterial->m_eBlendMode = (enum Asura_MaterialResponse::BLEND_MODE) uTemp;

			// Special FX properties
			xStream >> uTemp;
			pxNewMaterial->m_eSpecialFXMode = (enum Asura_MaterialResponse::SPECIALFX_MODE) uTemp;

			xStream >> pxNewMaterial->m_xWaterReflectionColour;
			xStream >> pxNewMaterial->m_xWaterRefractionColour;
			xStream >> pxNewMaterial->m_xWaterFogColour;
			xStream >> pxNewMaterial->m_fWaterFogStrength;
			
			for(int i=0;i<6;i++)
			{
				xStream >> pxNewMaterial->m_afWaterUVControl[i];
			}

			xStream >> pxNewMaterial->m_fWaterRippleScale;
			
			if( Version >= 22 )
			{
				xStream >> pxNewMaterial->m_fWaterDistortionScale;
			}

			xStream >> pxNewMaterial->m_fWaterEdgeSoftness;

			for(int i=0;i<6;i++)
			{
				xStream >> pxNewMaterial->m_afWaterWaveControl[i];
			}

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
			xStream >> pxNewMaterial->m_uSurfaceTypeHash;
#else
			xStream >> pxNewMaterial->m_uSurfaceType;
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
			xStream >> pxNewMaterial->m_fReflectivity;
			xStream >> pxNewMaterial->m_fParallaxScale;

			xStream >> pxNewMaterial->m_uSecondMaterialID;

			if( Version >= 26 )
			{
				xStream >> pxNewMaterial->m_uProjectContext;
			} else {
				pxNewMaterial->m_uProjectContext = 0;
			}


			if( Version >= 24 )
			{
				xStream >> pxNewMaterial->m_fAlbedoDetailMapUVScale;
				xStream >> pxNewMaterial->m_fAlbedoDetailMapValueScale;
				xStream >> pxNewMaterial->m_fNormalDetailMapUVScale;
				xStream >> pxNewMaterial->m_fNormalDetailMapValueScale;
			}

			if( Version >= 25 )
			{
				u_int	uRenderType;

				xStream >> pxNewMaterial->m_uChunkedFlags;
				xStream >> uRenderType;

				pxNewMaterial->m_eRenderType	= static_cast<Asura_MaterialResponse::RENDER_TYPE>(uRenderType);
			}
			else if( Version >= 23 )
			{
				union
				{
					u_int		m_uChunkedFlags;

					struct
					{
						u_int	m_bDontRender			: 1;
						u_int	m_bAlwaysCastShadows	: 1;
						u_int	m_bDecal				: 1;
					};
				} xOldChunkedFlags;

				// Switched to using a bitfield for the flags
				Asura_ReadFromStream_Bitfield_Wrapper< u_int >
					xBitfield( xOldChunkedFlags.m_uChunkedFlags );

				xStream >> xBitfield;

				if(xOldChunkedFlags.m_bDecal)
				{
					pxNewMaterial->m_uChunkedFlags	|= Tools_MaterialResponse::BITMASK_DECAL;
				}
				else
				{
					pxNewMaterial->m_uChunkedFlags	&= ~Tools_MaterialResponse::BITMASK_DECAL;
				}

				if(xOldChunkedFlags.m_bDontRender)
				{
					if(xOldChunkedFlags.m_bAlwaysCastShadows)
					{
						pxNewMaterial->m_eRenderType	= Asura_MaterialResponse::RENDER_TYPE_DONT_RENDER_BUT_CAST_SHADOW;
					}
					else
					{
						pxNewMaterial->m_eRenderType	= Asura_MaterialResponse::RENDER_TYPE_DEFAULT;
					}
				}
				else
				{
					pxNewMaterial->m_eRenderType	= Asura_MaterialResponse::RENDER_TYPE_DEFAULT;
				}
			}
			else
			{
				// bools are chunked as u_chars
				u_char ucDecal;

				if( Version >= 20 )
				{
					xStream >> ucDecal;
				}
				else
				{
					// Before the decal flag was added, transparent blend modes just meant decal-type transparency anyway
					ucDecal = (pxNewMaterial->GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_OPAQUE);
				}

				if(ucDecal)
				{
					pxNewMaterial->m_uChunkedFlags	|= Tools_MaterialResponse::BITMASK_DECAL;
				}
				else
				{
					pxNewMaterial->m_uChunkedFlags	&= ~Tools_MaterialResponse::BITMASK_DECAL;
				}

				if( Version >= 21 )
				{
					u_char	ucDontRender;
					u_char	ucAlwaysCastShadows;

					xStream >> ucDontRender;
					xStream >> ucAlwaysCastShadows;

					if(ucDontRender)
					{
						if(ucAlwaysCastShadows)
						{
							pxNewMaterial->m_eRenderType	= Asura_MaterialResponse::RENDER_TYPE_DONT_RENDER_BUT_CAST_SHADOW;
						}
						else
						{
							pxNewMaterial->m_eRenderType	= Asura_MaterialResponse::RENDER_TYPE_DEFAULT;
						}
					}
					else
					{
						pxNewMaterial->m_eRenderType	= Asura_MaterialResponse::RENDER_TYPE_DEFAULT;
					}
				}
			}

			// (AAM - 04.01.2008) Now I do want to use the discrete samples functionality. Reset
			// after each load in case the graph width has changed since the chunk was saved
			pxNewMaterial->m_xDiffuseGraph.SetNumDiscreteSamples( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH );
			pxNewMaterial->m_xSpecularGraph.SetNumDiscreteSamples( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH );
			pxNewMaterial->m_xTransparencyGraph.SetNumDiscreteSamples( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH );
		}
	}

	//Asura_PC_MaterialResponse_UpdateTexture();
	#ifdef ASURA_MAIN
	Asura_PC_MaterialResponse_InvalidTexture();
	#endif

	return true;
}

/******************************************************************************/
/* Tools_Chunk_MaterialResponse Static Class Functions                        */
/******************************************************************************/

void Tools_Chunk_MaterialResponse::Create( Asura_Chunk_Stream& xStream, Asura_List< Tools_MaterialResponse >& xMaterialList )
{
	xStream.WriteChunk_Start( TOOLS_CHUNK_MATERIAL_RESPONSE, uCURRENT_VERSION, 0 );

	xStream << xMaterialList.GetNumEntries();

	for( Asura_List_ConstIt< Tools_MaterialResponse > xIter( &xMaterialList ); !xIter.Done(); xIter.Next() )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = xIter.GetCurrent();

		// (AAM - 09.02.2007) Used to write out name here, but it's not needed

		xStream << pxCurrMaterial->m_uParentHash;

		//Save all the texture hashes and filenames
		xStream << (int) (Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS);
		for(int iSlot = 0; iSlot < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; iSlot++)
		{
			xStream << pxCurrMaterial->m_auTextureHashIDs[iSlot];
			xStream << pxCurrMaterial->m_axTextureFilenames[iSlot];
		}

		xStream << pxCurrMaterial->m_fAmbientContribution;
		xStream << pxCurrMaterial->m_fColouriseAmbient;
		xStream << pxCurrMaterial->m_xAmbientColour;

		pxCurrMaterial->m_xDiffuseGraph.WriteToChunkStream( xStream, false );
		xStream << pxCurrMaterial->m_fColouriseDiffuse;
		xStream << pxCurrMaterial->m_xDiffuseColour;

		xStream << pxCurrMaterial->m_fMetalicity;
		pxCurrMaterial->m_xSpecularGraph.WriteToChunkStream( xStream, false );
		xStream << pxCurrMaterial->m_fColouriseSpecular;
		xStream << pxCurrMaterial->m_xSpecularColour;

		pxCurrMaterial->m_xTransparencyGraph.WriteToChunkStream( xStream, false );

		xStream << pxCurrMaterial->m_fSelfIlluminationContribution;
		xStream << pxCurrMaterial->m_fColouriseSelfIllumination;
		xStream << pxCurrMaterial->m_xSelfIlluminationColour;

		xStream << pxCurrMaterial->m_uTextureAnimHashID;

		// stream out enum value as a u_int
		xStream << (u_int)pxCurrMaterial->m_eBlendMode;

		// Special FX properties
		xStream << (u_int)pxCurrMaterial->m_eSpecialFXMode;

		xStream << pxCurrMaterial->m_xWaterReflectionColour;
		xStream << pxCurrMaterial->m_xWaterRefractionColour;
		xStream << pxCurrMaterial->m_xWaterFogColour;
		xStream << pxCurrMaterial->m_fWaterFogStrength;

		for(int i=0;i<6;i++)
		{
			xStream << pxCurrMaterial->m_afWaterUVControl[i];
		}

		xStream << pxCurrMaterial->m_fWaterRippleScale;
		xStream << pxCurrMaterial->m_fWaterDistortionScale;
		xStream << pxCurrMaterial->m_fWaterEdgeSoftness;

		for(int i=0;i<6;i++)
		{
			xStream << pxCurrMaterial->m_afWaterWaveControl[i];
		}

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
		xStream << pxCurrMaterial->m_uSurfaceTypeHash;
#else
		xStream << pxCurrMaterial->m_uSurfaceType;
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
		xStream << pxCurrMaterial->m_fReflectivity;
		xStream << pxCurrMaterial->m_fParallaxScale;
		xStream << pxCurrMaterial->m_uSecondMaterialID;

		xStream << pxCurrMaterial->m_uProjectContext;

		xStream << pxCurrMaterial->m_fAlbedoDetailMapUVScale;
		xStream << pxCurrMaterial->m_fAlbedoDetailMapValueScale;
		xStream << pxCurrMaterial->m_fNormalDetailMapUVScale;
		xStream << pxCurrMaterial->m_fNormalDetailMapValueScale;

		xStream << pxCurrMaterial->m_uChunkedFlags;

		xStream << static_cast<u_int>(pxCurrMaterial->m_eRenderType);
	}

	xStream.WriteChunk_End();
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Tools_Chunk_MaterialResponseNames **
**-----------------------------------*/

#define TOOLS_CHUNK_MATERIAL_RESPONSE_NAMES_VERSION 0

bool Tools_Chunk_MaterialResponseNames::Process(Asura_Chunk_Stream& xStream) const
{
	// Is this the correct version?

	if (Version > TOOLS_CHUNK_MATERIAL_RESPONSE_NAMES_VERSION)
	{
		Asura_Assertf(false, ("Unknown chunk version %d for material response names",Version));
		return false;
	}

	int iNumNames;
	xStream >> iNumNames;

	char acBuffer[256];
	for(int i=0;i<iNumNames;i++)
	{
		xStream.ReadString(acBuffer, sizeof(acBuffer));

		Tools_MaterialResponse_System::AddToMaterialNameTree(acBuffer);
	}

	return true;
}

bool Tools_Chunk_MaterialResponseNames::Process(Asura_Handle iFileHandle) const
{
	Asura_Chunk_File_Stream	xStream(iFileHandle);
	return Process(xStream);
}
void Tools_Chunk_MaterialResponseNames::Create(const List<const char*>& xMaterialList,class Asura_Chunk_Stream& xStream)
{
	if(!xMaterialList.size()) return;

	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(TOOLS_CHUNK_MATERIAL_RESPONSE_NAMES, TOOLS_CHUNK_MATERIAL_RESPONSE_NAMES_VERSION, uChunkFlags);

	xStream << xMaterialList.size();

	for(CLIF<const char*> xMaterialLIF(&xMaterialList); !xMaterialLIF.done(); xMaterialLIF.next())
	{
		xStream << (Asura_String)xMaterialLIF();
	}
	
	xStream.WriteChunk_End();
}

void Tools_Chunk_MaterialResponseNames::Create(const Asura_HashTree& xMaterialTree, Asura_Chunk_Stream& xStream)
{
	//get the hashes of the materials
	int iNumMaterials = xMaterialTree.GetNumberOfNodes();
	if(!iNumMaterials) return;

	Asura_Hash_ID* puMaterialIDs = new Asura_Hash_ID[iNumMaterials];
	xMaterialTree.MakeHashList(puMaterialIDs);

	//and from that , get a list of material names
	List<const char*> xMaterialList;
	for(int i=0;i<iNumMaterials;i++)
	{
		const char* szName = Tools_MaterialResponse_System::GetMaterialNameByHashID(puMaterialIDs[i]);
		if(szName) xMaterialList.add_entry(szName);
	}
	delete [] puMaterialIDs;

	//create the chunk
	Create(xMaterialList, xStream);
}

void Tools_Chunk_MaterialResponseNames::Create(const List<const char*>& xMaterialList,class Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create(xMaterialList, xStream);
}

void Tools_Chunk_MaterialResponseNames::Create(const Asura_HashTree& xMaterialTree,class Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create(xMaterialTree, xStream);
}




#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
