
///////////////////////////////////////////////////////////////////////////////
//
// File   : Tools_Chunk_MaterialResponseOverrides.cpp
// Author : Andy Tate
// Date   : 4-May-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)

// This file's header...
#include "ChunkSystem/Tools_Chunk_MaterialResponseOverrides.h"

// Asura independent header-files...
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_MaterialResponse_System.h"

#if defined(ASURA_TOOLS)
// Tools header-files...
#include "Materials/Tools_MaterialResponse_System.h"
#endif	// defined(ASURA_TOOLS)


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const u_int	g_uChunkVersion	= 1;


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_Chunk_MaterialResponseOverrides::Process
// Author       : Andy Tate
// Date         : 4-May-2007
////////////////////////////////////////////////////////////////////////////////

bool Tools_Chunk_MaterialResponseOverrides::Process(const Asura_Handle iFileHandle) const
{
	if(Version > g_uChunkVersion)
	{
		Asura_Assert(false, "Unexpected chunk version found - aborting chunk-read.");
		return(false);
	}
	else
	{
		Asura_Chunk_File_Stream	xStream(iFileHandle);

		u_int	uNumOverrideMaps;

		// Read each override map in-turn...
		xStream >> uNumOverrideMaps;

		for(u_int u = 0; u < uNumOverrideMaps; u++)
		{
			u_int								uNumOverrides;
			Asura_String						xOverrideMapName;
			Asura_Hash_ID						uOverrideMapHash;
			Asura_MaterialResponse_OverrideMap*	pxOverrideMap;

			// Read the name of this override-map...
			xStream >> xOverrideMapName;
			uOverrideMapHash	= xOverrideMapName.GetHash();

			// Read each override in-turn...
			xStream >> uNumOverrides;

			// Ensure we have such an override-map...
			pxOverrideMap = Asura_MaterialResponse_System::GetOverrideMap(uOverrideMapHash);
			if(pxOverrideMap == NULL)
			{
				pxOverrideMap = Asura_MaterialResponse_System::CreateOverrideMap( uOverrideMapHash, uNumOverrides );
				pxOverrideMap->SetName(xOverrideMapName.GetString());
			}

			for(u_int v = 0; v < uNumOverrides; v++)
			{
				Asura_Hash_ID						uMaterialHash;
				Asura_Hash_ID						uOverrideHash;
				Asura_MaterialResponse_Override*	pxOverride;

				Asura_String	xMaterialName;
				Asura_String	xOverrideName;

				// Read the name of this override...
				xStream >> xMaterialName;
				xStream >> xOverrideName;

				uMaterialHash	= xMaterialName.GetHash();
				uOverrideHash	= xOverrideName.GetHash();

				//Skip any blank entries that were saved as a result of missing materials.RAR 21/11/2007
				if(!uMaterialHash)	continue;
				
				// Ensure we have such an override...
				pxOverride	= pxOverrideMap->GetOverride(uMaterialHash);

				if(pxOverride == NULL)
				{
					pxOverride	= pxOverrideMap->CreateOverride(uMaterialHash);
					pxOverride->SetOverrideName( xOverrideName );
					pxOverride->SetOverrideHashID(uOverrideHash);
				}
			}
		}

		return(true);
	}
}

/// End of function 'Tools_Chunk_MaterialResponseOverrides::Process'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_Chunk_MaterialResponseOverrides::Create
// Author       : Andy Tate
// Date         : 4-May-2007
////////////////////////////////////////////////////////////////////////////////

void Tools_Chunk_MaterialResponseOverrides::Create(Asura_Chunk_Stream& xStream)
{
	// Begin writing...
	xStream.WriteChunk_Start(TOOLS_CHUNK_MATERIAL_RESPONSE_OVERRIDES, g_uChunkVersion, 0);
	{
		const char**	pszOverrideMapNames	= Asura_MaterialResponse_System::CreateOverrideMapNameArray();
		const u_int		uNumOverrideMaps	= Asura_MaterialResponse_System::GetNumOverrideMaps();

		// Write each override map in-turn...
		xStream << uNumOverrideMaps;

		for(u_int u = 0; u < uNumOverrideMaps; u++)
		{
			const Asura_String							xOverrideMapName(pszOverrideMapNames[u]);
			const Asura_Hash_ID							uOverrideMapHash	= xOverrideMapName.GetHash();
			const Asura_MaterialResponse_OverrideMap*	pxOverrideMap		= Asura_MaterialResponse_System::GetOverrideMap(uOverrideMapHash);

			if(pxOverrideMap == NULL)
			{
				Asura_Assert(false, "Cannot find override-map to write.");
			}
			else
			{
				Asura_Hash_ID* puMaterialHashIDsToOverride = pxOverrideMap->CreateMaterialHashIDsToOverrideArray();
				const u_int	   uNumOverrides		       = pxOverrideMap->GetNumOverrides();

				// Write the name of this override-map...
				xStream << xOverrideMapName;

				// Write each override in-turn...
				xStream << uNumOverrides;

				for(u_int v = 0; v < uNumOverrides; v++)
				{
					const Asura_Hash_ID						uMaterialHash	= puMaterialHashIDsToOverride[v];
					const Asura_MaterialResponse_Override*	pxOverride		= pxOverrideMap->GetOverride(uMaterialHash);

					if(pxOverride == NULL)
					{
						Asura_Assert(false, "Failed to find override to write-out.");
					}
					else
					{
						const Tools_MaterialResponse*	pxMaterialResponse	= Tools_MaterialResponse_System::GetMaterialByHashID(uMaterialHash);
						

						if(pxMaterialResponse == NULL)
						{
							const Asura_String	xEmptyString;

							Asura_Assertf(false, ("Failed to find material-response for hash-ID %08X", uMaterialHash));
							xStream << xEmptyString;//blank material name
							xStream << xEmptyString;//blank override name
						}
						else
						{	
							xStream << pxMaterialResponse->GetName();
							xStream << pxOverride->GetOverrideName();
						}
					}
				}

				// Tidy-up our previously created data...
				pxOverrideMap->DestroyMaterialHashIDsToOverrideArray( puMaterialHashIDsToOverride );
			}
		}

		// Tidy-up our previously created data...
		Asura_MaterialResponse_System::DestroyOverrideMapNameArray(pszOverrideMapNames);
	}
	xStream.WriteChunk_End();
}

/// End of function 'Tools_Chunk_MaterialResponseOverrides::Create'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_Chunk_MaterialResponseOverrides::Save
// Author       : Andy Tate
// Date         : 4-May-2007
////////////////////////////////////////////////////////////////////////////////

bool Tools_Chunk_MaterialResponseOverrides::Save(const char* szFilename)
{
	Asura_SaveBlock					xSaveBlock;
	Asura_Chunk_SaveBlock_Stream	xStream(&xSaveBlock);

	// Fill-in the save-block...
	Create(xStream);

	// Terminate the Asura-file...
	xStream.WriteAsuraTerminator();

	// ...before saving the file to disc...
	return(xSaveBlock.SaveFile(szFilename, false));
}

/// End of function 'Tools_Chunk_MaterialResponseOverrides::Save'. ///

#endif	// defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)


/// End of file. ///
