//////////////////////////////////////////////////////////////////////////////
//
// Tools_Chunk_StaticDecals.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM

#include "ChunkSystem/Tools_Chunk_StaticDecals.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_ModuleLinking.h"//for module index conversion
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_Handle.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Module.h"

// Tools
#include "Decals/Tools_StaticDecal_System.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Objects/Tools_Object.h"
#include "Objects/Tools_Object_System.h"

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const int iCURRENT_VERSION = 3;


/******************************************************************************/
/* Tools_Chunk_StaticDecals Class Functions                                   */
/******************************************************************************/

bool Tools_Chunk_StaticDecals::Process( Asura_Chunk_Stream& xStream ) const
{
	if( Version > iCURRENT_VERSION )
	{
		Asura_Assertf( false, ("Unknown static decals chunk version (%d). You may need a new Tools build.", Version) );
		return false;
	}

	bool bUseModuleIndexConversion = false;
	if(Version >= 3)
	{
		u_int uNumberOfModules;
		xStream >> uNumberOfModules;

		//check number of modules with the module index conversion thing
		Asura_ModuleIndex_Conversion::CheckNumberOfModules(uNumberOfModules,Asura_Module_System::GetNumberOfModules());
		bUseModuleIndexConversion = true;
	}
	else
	{
		Asura_Assertf(false, ("Old static decal chunk version detected (loaded version: %d, current version: %d). This chunk doesn't support changes to module data and may not render correctly if the number of modules changes. Please go into Decals mode, recalculate all, and save the file.", Version, iCURRENT_VERSION));
		Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Static decal file out of date. Please recalculate all and save the file.");
	}


	// Remove any existing decal buffer (should never happen, but just in case)
	Tools_StaticDecal_System::Reset();

	
	int iNumBuffers;	
	xStream >> iNumBuffers;

	Tools_StaticDecal_System::AllocateDecalBuffers(iNumBuffers);
	
	for(int i=0; i<iNumBuffers; i++)
	{
		Tools_StaticDecal_Buffer *pxBuffer = Tools_StaticDecal_System::GetDecalBufferByIndex(i);

		// Read the buffer params
		xStream >> pxBuffer->m_iModuleIndex;
		//Deal with changes to module indices
		if(bUseModuleIndexConversion && pxBuffer->m_iModuleIndex>=0)
		{
			pxBuffer->m_iModuleIndex = Asura_ModuleIndex_Conversion::ConvertIndex(pxBuffer->m_iModuleIndex);				
			if(pxBuffer->m_iModuleIndex < 0)
			{
				//Old module for this decal buffer no longer exists, so reset to uninitialised
				pxBuffer->m_iModuleIndex = -2;
			}
		}

		xStream >> pxBuffer->m_uDecalMaterial;

		// Read the decals
		xStream >> pxBuffer->m_iNumDecals;

		if (Version>0)
		{
			xStream >> pxBuffer->m_xAABB;
		}

		pxBuffer->m_ppxDecals = new Tools_StaticDecal*[pxBuffer->m_iNumDecals];

		for (int j=0; j < pxBuffer->m_iNumDecals; j++)
		{
			Tools_StaticDecal *pxDecal = new Tools_StaticDecal();
			pxBuffer->m_ppxDecals[j] = pxDecal;

			xStream >> pxDecal->m_xPosition;
			xStream >> pxDecal->m_xOrient;
			xStream >> pxDecal->m_iModuleIndex;			
			xStream >> pxDecal->m_uDecalSetHash;
			xStream >> pxDecal->m_iDecalIndex;
			xStream >> pxDecal->m_fScale;
			xStream >> pxDecal->m_fAlpha;
			xStream >> pxDecal->m_uUsage;

			//Deal with changes to module indeces
			if(bUseModuleIndexConversion && pxDecal->m_iModuleIndex>=0)
			{
				pxDecal->m_iModuleIndex = Asura_ModuleIndex_Conversion::ConvertIndex(pxDecal->m_iModuleIndex);				
				if(pxDecal->m_iModuleIndex < 0)
				{
					//Old module for this decal no longer exists ,so reset to uninitialised
					pxDecal->m_iModuleIndex = -2;
				}
			}
		}

		// Read the indices
		xStream >> pxBuffer->m_iNumIndices;
		
		pxBuffer->m_iNumPolys = pxBuffer->m_iNumIndices/3;

		if (pxBuffer->m_iNumIndices>0)
		{		
			pxBuffer->m_piIndices = new int[pxBuffer->m_iNumIndices];

			for (int j=0; j < pxBuffer->m_iNumIndices; j++)
			{
				xStream >> pxBuffer->m_piIndices[j];
			}
		}
		else
		{
			pxBuffer->m_piIndices = NULL;
		}

		// Read the vertices
		xStream >> pxBuffer->m_iNumVerts;

		if (pxBuffer->m_iNumVerts>0)
		{
			pxBuffer->m_pxVerts = new Tools_StaticDecal_Vertex[pxBuffer->m_iNumVerts];
	
			for (int j=0; j < pxBuffer->m_iNumVerts; j++)
			{
				if( Version < 2 )
				{
					xStream >> pxBuffer->m_pxVerts[j].xPos;
					xStream >> pxBuffer->m_pxVerts[j].xNormal;
					xStream >> pxBuffer->m_pxVerts[j].xBinormal;
					xStream >> pxBuffer->m_pxVerts[j].xTangent;
					u_int uColour;
					xStream >> uColour;
					Asura_RGBA xCol;
					xCol.FromUIntARGB(uColour);
					pxBuffer->m_pxVerts[j].fAlpha = xCol.m_fA;
					pxBuffer->m_pxVerts[j].xSHRed = Asura_AVector4(0.0f, 0.0f, 0.0f, xCol.m_fR);
					pxBuffer->m_pxVerts[j].xSHGreen = Asura_AVector4(0.0f, 0.0f, 0.0f, xCol.m_fG);
					pxBuffer->m_pxVerts[j].xSHBlue = Asura_AVector4(0.0f, 0.0f, 0.0f, xCol.m_fB);

					xStream >> pxBuffer->m_pxVerts[j].fU;
					xStream >> pxBuffer->m_pxVerts[j].fV;
				
					float fUnused;
					xStream >> fUnused;		// Always 0 for now (it's the 'creation time' for non-static decals)
				} else{
					xStream >> pxBuffer->m_pxVerts[j].xPos;
					xStream >> pxBuffer->m_pxVerts[j].fU;
					xStream >> pxBuffer->m_pxVerts[j].fV;
					xStream >> pxBuffer->m_pxVerts[j].fCreationTime;
					xStream >> pxBuffer->m_pxVerts[j].fAlpha;

					xStream >> pxBuffer->m_pxVerts[j].xNormal;
					xStream >> pxBuffer->m_pxVerts[j].xTangent;
					xStream >> pxBuffer->m_pxVerts[j].xBinormal;

					xStream >> *(Asura_RGBA*)&pxBuffer->m_pxVerts[j].xSHRed;
					xStream >> *(Asura_RGBA*)&pxBuffer->m_pxVerts[j].xSHGreen;
					xStream >> *(Asura_RGBA*)&pxBuffer->m_pxVerts[j].xSHBlue;
				}

			}
		}
		else
		{
			pxBuffer->m_pxVerts = NULL;
		}
	}


	return true;
}

bool Tools_Chunk_StaticDecals::Process( const int iFileHandle ) const
{
	Asura_Chunk_File_Stream xStream( iFileHandle );
	return Process( xStream );
}

/******************************************************************************/
/* Tools_Chunk_ToolsLights Static Class Functions                             */
/******************************************************************************/

void Tools_Chunk_StaticDecals::Create( Asura_Chunk_Stream& xStream )
{
	xStream.WriteChunk_Start( TOOLS_CHUNK_STATICDECALS, iCURRENT_VERSION, 0 );


	xStream << (u_int)Asura_Module_System::GetNumberOfModules();

	int iNumBuffers = Tools_StaticDecal_System::GetNumDecalBuffers();
	xStream << iNumBuffers;

	for(int i=0; i<iNumBuffers; i++)
	{
		Tools_StaticDecal_Buffer *pxBuffer = Tools_StaticDecal_System::GetDecalBufferByIndex(i);

		// Output the params
		xStream << pxBuffer->m_iModuleIndex;
		xStream << pxBuffer->m_uDecalMaterial;

		// Output the decals
		xStream << pxBuffer->m_iNumDecals;

		// Output the bounding box
		xStream << pxBuffer->m_xAABB;

		for (int j=0; j < pxBuffer->m_iNumDecals; j++)
		{
			Tools_StaticDecal *pxDecal = pxBuffer->m_ppxDecals[j];

			xStream << pxDecal->m_xPosition;
			xStream << pxDecal->m_xOrient;
			xStream << pxDecal->m_iModuleIndex;			
			xStream << pxDecal->m_uDecalSetHash;
			xStream << pxDecal->m_iDecalIndex;
			xStream << pxDecal->m_fScale;
			xStream << pxDecal->m_fAlpha;
			xStream << pxDecal->m_uUsage;			
		}

		// Output the indices
		xStream << pxBuffer->m_iNumIndices;

		for (int j=0; j < pxBuffer->m_iNumIndices; j++)
		{
			xStream << pxBuffer->m_piIndices[j];
		}

		// Output the vertices
		xStream << pxBuffer->m_iNumVerts;

		for (int j=0; j < pxBuffer->m_iNumVerts; j++)
		{
			xStream << pxBuffer->m_pxVerts[j].xPos;
			xStream << pxBuffer->m_pxVerts[j].fU;
			xStream << pxBuffer->m_pxVerts[j].fV;
			xStream << pxBuffer->m_pxVerts[j].fCreationTime;
			xStream << pxBuffer->m_pxVerts[j].fAlpha;

			xStream << pxBuffer->m_pxVerts[j].xNormal;
			xStream << pxBuffer->m_pxVerts[j].xBinormal;
			xStream << pxBuffer->m_pxVerts[j].xTangent;

			xStream << *(Asura_RGBA*)&pxBuffer->m_pxVerts[j].xSHRed;
			xStream << *(Asura_RGBA*)&pxBuffer->m_pxVerts[j].xSHGreen;
			xStream << *(Asura_RGBA*)&pxBuffer->m_pxVerts[j].xSHBlue;

		}
	}

	xStream.WriteChunk_End();
}


bool Tools_Chunk_StaticDecals::Create( char *szFileName )
{

	Asura_SaveBlock* pxSaveBlock = new Asura_SaveBlock();

	Asura_Assert(pxSaveBlock, "Unable to create saveblock");

	if(pxSaveBlock)
	{
		Asura_Chunk_SaveBlock_Stream	xStream (pxSaveBlock);
		
	    if(!Tools_Level_Export::IsExport())
		{
    		Asura_Chunk_ModuleNames::Create(xStream);
		}
				
		Create( xStream );

		xStream.WriteAsuraTerminator();

		bool bSuccess=pxSaveBlock->SaveFile(szFileName, false );

		delete pxSaveBlock;

		if (bSuccess)
		{
			return true;
		}
	}

	return false;
}


#endif // ASURA_USE_STATIC_DECAL_SYSTEM