

//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE : Tools_StaticDecal_System.cpp
//
//  Created by Dave Reed - 07/08/07
//
//  Copyright (C) 2007 Rebellion (Derby), All Rights Reserved.
//
//  Static Decal management/creation for tools
//
//////////////////////////////////////////////////////////////////////////////////////////


#include "Core/Asura.h"

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM

#include "Core/Asura_Maths.h"
#include "Camera/Asura_Camera.h"
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"
#include "Tools_StaticDecal_System.h"
#include "Objects/Tools_Object_System.h"
#include "Botanicals/Asura_Botanicals_InstanceManager.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Debugging/Asura_Logfile.h"
#include "Light/Asura_GI_Solution.h"

int						    Tools_StaticDecal_System::s_iNumDecalBuffers = 0;
Tools_StaticDecal_Buffer**  Tools_StaticDecal_System::s_ppxDecalBuffers = 0;

// ----------------------------------------------------------------------------------------
// The 'workspace' is a set of buffers used when updating the geometry for a decal buffer, 
// at which point we don't know how many verts we're going to need. 
// ----------------------------------------------------------------------------------------
int							Tools_StaticDecal_System::s_iWorkspaceNumVerts   = 0;
int							Tools_StaticDecal_System::s_iWorkspaceNumIndices = 0;
int							Tools_StaticDecal_System::s_iWorkspaceMaxVerts	 = 0;
int							Tools_StaticDecal_System::s_iWorkspaceMaxIndices = 0;
int*						Tools_StaticDecal_System::s_piWorkspaceIndices   = NULL;
Tools_StaticDecal_Vertex*	Tools_StaticDecal_System::s_pxWorkspaceVerts     = NULL;
uint32					    Tools_StaticDecal_System::s_uCurrentAlpha		 = 0;

Tools_StaticDecal*			Tools_StaticDecal_System::s_pxSelectedDecal			 = NULL;
Tools_StaticDecal_Buffer*   Tools_StaticDecal_System::s_pxSelectedDecalBuffer    = NULL;
int							Tools_StaticDecal_System::s_iSelectedDecalFirstIndex = 0;
int							Tools_StaticDecal_System::s_iSelectedDecalNumPolys   = 0;


// ----------------------------------------------------------------------------------------
// Initialise()
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::Initialise(void)
{
	s_ppxDecalBuffers = NULL;
	s_iNumDecalBuffers = 0;

	s_piWorkspaceIndices = NULL;
	s_iWorkspaceMaxVerts = 0;

	s_piWorkspaceIndices = NULL;
	s_iWorkspaceMaxIndices = 0;
}



// ----------------------------------------------------------------------------------------
// Shutdown()
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::Shutdown(void)
{
	if (s_ppxDecalBuffers)
	{
		for(int i=0;i<s_iNumDecalBuffers;i++)
		{
			delete s_ppxDecalBuffers[i];
		}

		delete [] s_ppxDecalBuffers;
		s_ppxDecalBuffers = NULL;
	}
	
	s_iNumDecalBuffers = 0;

	if (s_piWorkspaceIndices)
	{
		delete [] s_piWorkspaceIndices;
		s_piWorkspaceIndices = NULL;
		s_iWorkspaceMaxIndices = 0;
	}

	if (s_pxWorkspaceVerts)
	{
		delete [] s_pxWorkspaceVerts;
		s_piWorkspaceIndices = NULL;
		s_iWorkspaceMaxVerts = 0;
	}
}



// ----------------------------------------------------------------------------------------
// Reset()
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::Reset(void)
{
	Shutdown();
	Initialise();
}



// ----------------------------------------------------------------------------------------
// AllocateDecalBuffers()
// Called from the tools chunk loader
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::AllocateDecalBuffers(int iNumDecalBuffers)
{
	s_ppxDecalBuffers = new Tools_StaticDecal_Buffer*[iNumDecalBuffers];
	s_iNumDecalBuffers = iNumDecalBuffers;

	for (int i=0; i<iNumDecalBuffers; i++)
	{
		s_ppxDecalBuffers[i] = new Tools_StaticDecal_Buffer();
	}
}


// ----------------------------------------------------------------------------------------
// GetNumDecalBuffers()
// ----------------------------------------------------------------------------------------
int Tools_StaticDecal_System::GetNumDecalBuffers(void)
{
	return s_iNumDecalBuffers;
}



// ----------------------------------------------------------------------------------------
// GetNumDecals()
// ----------------------------------------------------------------------------------------
int Tools_StaticDecal_System::GetNumDecals(void)
{
	int iCount = 0;

	for(int i=0;i<s_iNumDecalBuffers;i++)
	{
		iCount+=s_ppxDecalBuffers[i]->m_iNumDecals;
	}

	return iCount;
}


// ----------------------------------------------------------------------------------------
// GetNumDecalPolys()
// ----------------------------------------------------------------------------------------
int Tools_StaticDecal_System::GetNumDecalPolys(void)
{
	int iCount = 0;

	for(int i=0;i<s_iNumDecalBuffers;i++)
	{
		iCount+=s_ppxDecalBuffers[i]->m_iNumPolys;
	}

	return iCount;
}


// ----------------------------------------------------------------------------------------
// UpdateBuffers() - Create vertex/index buffers for any decalbuffers flagged as dirty
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::UpdateBuffers(bool bForceAll,  uint32 uForceSpecificMaterial, bool bOnExport)
{
	if (bForceAll)
	{
		if (!bOnExport)
		{		
			Tools_SaveAsura::SetStaticDecalsModified(true);
		}

		// For a proper recalc, we first need to ensure the decal module indices are valid
		// Iterate through the tools objects here, updating them
		for(LIF<Tools_Object*> olif(&Tools_Object_System::ObjectList);!olif.done();olif.next())
		{
			if (olif()->Type == TOT_StaticDecal)
			{					
				Tools_StaticDecal_Object *pxToolsObject = (Tools_StaticDecal_Object *)olif();
				UpdateDecal(pxToolsObject);							
			}
		}
	}


	for(int i=0;i<s_iNumDecalBuffers;i++)
	{
		if (s_ppxDecalBuffers[i]->m_bDirty || bForceAll)
		{
			UpdateVertices(s_ppxDecalBuffers[i]);
			s_ppxDecalBuffers[i]->m_bDirty = false;
		}
	}
}


// ----------------------------------------------------------------------------------------
// OptimizeBuffer() - Remove any duplicated verts. This is done before export only
// ----------------------------------------------------------------------------------------

static bool VertsEqual(Tools_StaticDecal_Vertex &xA, Tools_StaticDecal_Vertex &xB)
{
	if (!(xA.xSHRed == xB.xSHRed)) return false;
	if (!(xA.xSHGreen == xB.xSHGreen)) return false;
	if (!(xA.xSHBlue == xB.xSHBlue)) return false;

	if (xA.fAlpha!=xB.fAlpha) return false;

	if (abs(xA.fU - xB.fU)>0.001f) return false;
	if (abs(xA.fV - xB.fV)>0.001f) return false;	
	
	if (!xA.xPos.ApproximatelyEquals(xB.xPos)) return false;
	if (!xA.xNormal.ApproximatelyEquals(xB.xNormal)) return false;
	if (!xA.xTangent.ApproximatelyEquals(xB.xTangent)) return false;
	if (!xA.xBinormal.ApproximatelyEquals(xB.xBinormal)) return false;

	return true;
}


void Tools_StaticDecal_System::OptimizeBuffer(int iBuffer)
{
	Tools_StaticDecal_Buffer *pxBuffer = s_ppxDecalBuffers[iBuffer];

	if (pxBuffer->m_iNumVerts>0)
	{
		Tools_StaticDecal_Vertex *pxNewVerts = new Tools_StaticDecal_Vertex[pxBuffer->m_iNumVerts];
		int iNumNewVerts = 0 ;
		int *piRemapTable = new int[pxBuffer->m_iNumVerts];

		for(int i=0; i<pxBuffer->m_iNumVerts; i++)
		{
			Tools_StaticDecal_Vertex xVert = pxBuffer->m_pxVerts[i];
			bool bFound=false;

			for(int j=0; (j<iNumNewVerts && !bFound); j++)
			{
				if (VertsEqual(pxNewVerts[j], xVert))
				{
					bFound=true;
					piRemapTable[i] = j;
				}
			}

			if (!bFound)
			{
				pxNewVerts[iNumNewVerts] = xVert;
				piRemapTable[i] = iNumNewVerts;
				iNumNewVerts++;
			}						
		}

		for(int i=0; i<pxBuffer->m_iNumIndices; i++)
		{
			pxBuffer->m_piIndices[i] = piRemapTable[pxBuffer->m_piIndices[i]];
		}

		delete [] pxBuffer->m_pxVerts;
		pxBuffer->m_pxVerts = pxNewVerts;
		
		delete [] piRemapTable;

		int iSaved = pxBuffer->m_iNumVerts-iNumNewVerts;		
		float fPercent = 100.0f*(((float)(iSaved)) / ((float)(pxBuffer->m_iNumVerts)));

		pxBuffer->m_iNumVerts = iNumNewVerts;		
		
		Asura_OutputToDebugger_Retail(( "INFO: Optimizing decal buffer - removed %d duplicated verts (%0.02f%%)\n", iSaved, fPercent ));
	}
}




// ----------------------------------------------------------------------------------------
// SetSelectedDecal() - Sets the selected decal (will get it's wireframe drawn)
// ----------------------------------------------------------------------------------------

void Tools_StaticDecal_System::SetSelectedDecal(Tools_StaticDecal_Object *pxToolsDecal)
{
	if (pxToolsDecal==NULL)
	{
		s_pxSelectedDecal = NULL;
		s_pxSelectedDecalBuffer = NULL;
	}
	else
	{
		s_pxSelectedDecal = pxToolsDecal->m_pxDecal;
		s_pxSelectedDecalBuffer = GetDecalBuffer(pxToolsDecal->m_pxDecal);
	}
}


// ----------------------------------------------------------------------------------------
// RenderWireframe() - Renders the wireframe of the selected decal 
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::RenderWireframe(void)
{
	Asura_Vector_3 xOfs;
	Asura_Camera::GetForwardVector(xOfs);
	xOfs.Normalise();
	xOfs = xOfs * -0.05f;

	if (s_pxSelectedDecal && s_pxSelectedDecalBuffer)
	{	
		for(int i=0;i<s_iNumDecalBuffers;i++)
		{
			Tools_StaticDecal_Buffer *pxBuffer = s_ppxDecalBuffers[i];

			if (pxBuffer == s_pxSelectedDecalBuffer )
			{						

				if (pxBuffer->m_iNumIndices>=s_iSelectedDecalFirstIndex+(s_iSelectedDecalNumPolys*3))
				{				
					for(int j=0; j<s_iSelectedDecalNumPolys; j++)
					{
						Asura_Vector_3 xA = xOfs+pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+0]].xPos;
						Asura_Vector_3 xB = xOfs+pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+1]].xPos;
						Asura_Vector_3 xC = xOfs+pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+2]].xPos;
						Asura_Render::RenderLine(xA, xB, 0xFFFFFFFF);
						Asura_Render::RenderLine(xB, xC, 0xFFFFFFFF);
						Asura_Render::RenderLine(xC, xA, 0xFFFFFFFF);

						/*
						// For debugging
						for (int k=0;k<3;k++)
						{
							Asura_Vector_3 xA = (xOfs*1.1f)+pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+k]].xPos;
							Asura_Vector_3 xN = xA + pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+k]].xNormal * 0.1f;
							Asura_Vector_3 xT = xA + pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+k]].xTangent * 0.1f;
							Asura_Vector_3 xB = xA + pxBuffer->m_pxVerts[pxBuffer->m_piIndices[s_iSelectedDecalFirstIndex+j*3+k]].xBinormal * 0.1f;

							Asura_Render::RenderLine(xA, xN, 0xFFFF0000);
							Asura_Render::RenderLine(xA, xT, 0xFF00FF00);
							Asura_Render::RenderLine(xA, xB, 0xFF0000FF);
						}
						*/
					}		
				}
			}
		}
	}	
}


// ----------------------------------------------------------------------------------------
// UpdateDecal() - If a decal has changed (or just been added), make sure it's in the
// right buffer, update it's params, and mark any affects buffers as dirty
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::UpdateDecal(Tools_StaticDecal_Object *pxToolsDecal)
{
	
	// Find module that it belongs in
	int iModule = Asura_Botanicals_InstanceManager::GetModuleIndexFromPosition(pxToolsDecal->Position, pxToolsDecal->Orientation, true);	
	bool bListChanged = false;

	if(pxToolsDecal->m_pxDecal)
	{
		Tools_StaticDecal *pxDecal = pxToolsDecal->m_pxDecal;

		if (iModule != pxDecal->m_iModuleIndex || pxToolsDecal->m_uDecalSetHash != pxDecal->m_uDecalSetHash)		
		{
			bListChanged = true;
		}
	}
	else
	{
		bListChanged = true;
	}

	if (bListChanged)
	{
		if(pxToolsDecal->m_pxDecal)
		{
			// Remove decal from the old list
			RemoveDecal(pxToolsDecal);
		}

		// Add decal to the new list			
		pxToolsDecal->m_pxDecal = AddDecal(iModule, pxToolsDecal->m_uDecalSetHash);
	}


	// Update the decal, and mark it's buffer as dirty
	Tools_StaticDecal *pxDecal = pxToolsDecal->m_pxDecal;
	Asura_Assert(pxDecal, "Fatal - failed to create decal");
	
	pxDecal->m_iModuleIndex  = iModule;
	pxDecal->m_xPosition	 = pxToolsDecal->Position;
	pxDecal->m_xOrient		 = pxToolsDecal->Orientation;
	pxDecal->m_uDecalSetHash = pxToolsDecal->m_uDecalSetHash;	
	pxDecal->m_iDecalIndex   = pxToolsDecal->m_iDecalIndex;
	pxDecal->m_fAlpha		 = pxToolsDecal->m_fAlpha;
	pxDecal->m_fScale		 = pxToolsDecal->m_fScale;

	Tools_StaticDecal_Buffer *pxBuffer = GetDecalBuffer(pxDecal->m_iModuleIndex, pxDecal->m_uDecalSetHash);
	if (pxBuffer)
	{
		pxBuffer->m_bDirty = true;
	}	
}



// ----------------------------------------------------------------------------------------
// RemoveDecal() - Remove a decal from it's DecalBuffer, destroying the buffer if empty
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::RemoveDecal(Tools_StaticDecal_Object *pxToolsDecal)
{
	Tools_StaticDecal_Buffer *pxBuffer = GetDecalBuffer(pxToolsDecal->m_pxDecal);
	Asura_Assert(pxBuffer, "Fatal - Decal buffer has gone missing...");

	// Locate the decal, remove it
	int iIndex = -1;	

	for (int i=0; i< pxBuffer->m_iNumDecals; i++)
	{
		if (pxBuffer->m_ppxDecals[i] == pxToolsDecal->m_pxDecal)
		{
			iIndex = i;
		}
	}

	if (iIndex>=0)
	{
		pxBuffer->m_iNumDecals--;

		if (pxBuffer->m_iNumDecals>0)
		{	
			Tools_StaticDecal **ppxDecals = pxBuffer->m_ppxDecals;
			pxBuffer->m_ppxDecals = new Tools_StaticDecal*[pxBuffer->m_iNumDecals];	

			for (int i=0; i< pxBuffer->m_iNumDecals; i++)
			{
				if (i<iIndex)
					pxBuffer->m_ppxDecals[i] = ppxDecals[i];
				else
					pxBuffer->m_ppxDecals[i] = ppxDecals[i+1];				
			}

			delete ppxDecals[iIndex];
			delete [] ppxDecals;

			pxBuffer->m_bDirty = true;
		}
		else
		{			
			delete pxBuffer->m_ppxDecals[0];
			delete [] pxBuffer->m_ppxDecals;
			pxBuffer->m_ppxDecals = NULL;
			RemoveDecalBuffer(pxBuffer);
		}	
	}
}


// ----------------------------------------------------------------------------------------
// AddDecal() - Internal function to add a decal, creating the DecalBuffer if needed
// ----------------------------------------------------------------------------------------
Tools_StaticDecal *Tools_StaticDecal_System::AddDecal(int iModule, uint32 uDecalSetHash)
{
	// This will create the buffer if needed:
	Tools_StaticDecal_Buffer *pxBuffer = GetDecalBuffer(iModule, uDecalSetHash);
	Asura_Assert(pxBuffer, "Fatal - Decal buffer has gone missing...");

	// Now add the decal
	Tools_StaticDecal **ppxDecals = pxBuffer->m_ppxDecals;

	pxBuffer->m_iNumDecals++;
	pxBuffer->m_ppxDecals = new Tools_StaticDecal*[pxBuffer->m_iNumDecals];

	if (ppxDecals)
	{
		for (int i=0; i< pxBuffer->m_iNumDecals-1; i++)
		{
			 pxBuffer->m_ppxDecals[i] = ppxDecals[i];
		}

		delete [] ppxDecals;
	}
		
	pxBuffer->m_ppxDecals[pxBuffer->m_iNumDecals-1] = new Tools_StaticDecal();
	return pxBuffer->m_ppxDecals[pxBuffer->m_iNumDecals-1];
}



// ----------------------------------------------------------------------------------------
// GetDecalBuffer() - Find the decal buffer that a Tools_StaticDecal belongs to
// ----------------------------------------------------------------------------------------
Tools_StaticDecal_Buffer *Tools_StaticDecal_System::GetDecalBuffer(Tools_StaticDecal *pxDecal)
{
	return GetDecalBuffer(pxDecal->m_iModuleIndex, pxDecal->m_uDecalSetHash);
}

Tools_StaticDecal_Buffer *Tools_StaticDecal_System::GetDecalBuffer(int iModule, uint32 uDecalSetHash)
{		
	uint32  uMaterial = 0;

	Asura_DecalManager_DecalSet *pxSet = Asura_Decal_Manager::GetDecalSet(uDecalSetHash);

	if (pxSet)
	{
		uMaterial = pxSet->m_uMtlNameHash;
	}

	for(int i=0; i<s_iNumDecalBuffers; i++)
	{
		if (s_ppxDecalBuffers[i]->m_iModuleIndex   == iModule &&
			s_ppxDecalBuffers[i]->m_uDecalMaterial == uMaterial)
		{
			return s_ppxDecalBuffers[i];
		}
	}

	// Not found - need to add a new decal buffer
	Tools_StaticDecal_Buffer **ppxOldBuffers = s_ppxDecalBuffers;
	s_iNumDecalBuffers++;
	s_ppxDecalBuffers = new Tools_StaticDecal_Buffer*[s_iNumDecalBuffers];	

	if (ppxOldBuffers)
	{
		for(int i=0; i<s_iNumDecalBuffers-1; i++)	
		{
			s_ppxDecalBuffers[i] = ppxOldBuffers[i];
		}

		delete [] ppxOldBuffers;
	}

	s_ppxDecalBuffers[s_iNumDecalBuffers-1] = new Tools_StaticDecal_Buffer;
	s_ppxDecalBuffers[s_iNumDecalBuffers-1]->m_iModuleIndex   = iModule;
	s_ppxDecalBuffers[s_iNumDecalBuffers-1]->m_uDecalMaterial = uMaterial;
	return s_ppxDecalBuffers[s_iNumDecalBuffers-1];
}



// ----------------------------------------------------------------------------------------
// RemoveDecalBuffer() - Removes a DecalBuffer
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::RemoveDecalBuffer(Tools_StaticDecal_Buffer *pxBuffer)
{

	int iIndex =-1;

	for(int i=0; i<s_iNumDecalBuffers; i++)
	{
		if (s_ppxDecalBuffers[i] == pxBuffer)
		{
			iIndex = i;
		}
	}

	if (iIndex>=0)
	{
		Tools_StaticDecal_Buffer **ppxOldBuffers = s_ppxDecalBuffers;
		s_iNumDecalBuffers--;

		if (s_iNumDecalBuffers>0)
		{
			s_ppxDecalBuffers = new Tools_StaticDecal_Buffer*[s_iNumDecalBuffers];	

			for (int i=0; i< s_iNumDecalBuffers; i++)
			{
				if (i<iIndex)
					s_ppxDecalBuffers[i] = ppxOldBuffers[i];
				else
					s_ppxDecalBuffers[i] = ppxOldBuffers[i+1];				
			}
		}
		else
		{
			s_ppxDecalBuffers = NULL;
		}

		delete pxBuffer;
		delete [] ppxOldBuffers;
		
	}

}


// ----------------------------------------------------------------------------------------
// CreateToolsObjects() - Create the tools objects for all static decals in the system
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::CreateToolsObjects(void)
{
	// Create all the tool objects (when entering Decals mode)

	for(int i=0;i<s_iNumDecalBuffers;i++)
	{
		Tools_StaticDecal_Buffer *pxBuffer = s_ppxDecalBuffers[i];

		for(int j=0; j<pxBuffer->m_iNumDecals; j++)
		{
			Tools_StaticDecal *pxDecal = pxBuffer->m_ppxDecals[j];

			Tools_StaticDecal_Object *pxToolsDecal = new Tools_StaticDecal_Object(pxDecal->m_xPosition, pxDecal->m_xOrient, false);
			
			pxToolsDecal->m_pxDecal		  = pxDecal;
			pxToolsDecal->m_uDecalSetHash = pxDecal->m_uDecalSetHash;	
			pxToolsDecal->m_iDecalIndex	  = pxDecal->m_iDecalIndex;
			pxToolsDecal->m_fAlpha		  = pxDecal->m_fAlpha;
			pxToolsDecal->m_fScale		  = pxDecal->m_fScale;
		}
	}
}




// ----------------------------------------------------------------------------------------
// ClearWorkspaceVerts() - Clear out temporary buffer used to build vert/idx buffers
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::ClearWorkspaceVerts()
{	
	s_iWorkspaceNumVerts = 0;
	s_iWorkspaceNumIndices = 0;
}



// ----------------------------------------------------------------------------------------
// AddStrip() - Called back from Asura_Decal_System::AddDecal, builds up vert/idx
// data for a DecalBuffer in the workspace buffers
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::AddStrip(const Asura_Decal_Vertex* pxStrip,
										u_int uNumVertices,
										const Asura_RGBA& xDiffuseRGBA,
										const Asura_Decal_Template* pxDecalTemplate,
										bool bCalculateStatiSH)
{

	Asura_Assert(uNumVertices<=9, "Unexpectedly large number of verts in decal strip");

	// Do we need to grow the workspace vertex buffer?
	if (s_iWorkspaceNumVerts+static_cast<int>(uNumVertices) >= s_iWorkspaceMaxVerts)
	{
		Tools_StaticDecal_Vertex* pxVerts = new Tools_StaticDecal_Vertex[s_iWorkspaceMaxVerts+128];
		if (s_pxWorkspaceVerts)
		{
			for(int i=0; i<s_iWorkspaceMaxVerts; i++)
			{
				pxVerts[i] = s_pxWorkspaceVerts[i];
			}		

			delete[] s_pxWorkspaceVerts;
		}
		s_pxWorkspaceVerts = pxVerts;
		s_iWorkspaceMaxVerts+=128;
	}

	// Do we need to grow the workspace index buffer?
	if (s_iWorkspaceNumIndices+static_cast<int>(uNumVertices*3) >= s_iWorkspaceMaxIndices)
	{
		int *pxIndices = new int[s_iWorkspaceMaxIndices+512];
		if (s_piWorkspaceIndices)
		{
			for(int i=0; i<s_iWorkspaceMaxIndices; i++)
			{
				pxIndices[i] = s_piWorkspaceIndices[i];
			}		

			delete[] s_piWorkspaceIndices;
		}
		s_piWorkspaceIndices = pxIndices;
		s_iWorkspaceMaxIndices+=512;
	}


	// Compute the tangent/binormal
	// We take a single normal for the decal now, to avoid a discontinuity when the
	// decal is 'bent' over a polygon edge.
	Asura_Vector_3 xNormal,xTangent,xBinormal;
	xNormal = Asura_Decal_System::GetCurrentDecalNormal();
	xNormal.Normalise();
	{		
		// There's room for some optimization here...
		Asura_Vector_3 axPositions[3];
		Asura_Vector_2 axUVs[3];
		for(int i=0;i<3;i++)
		{			
			axPositions[i] = pxStrip[i].m_xPosition;
			axUVs[i].x	   = pxStrip[i].m_fU;
			axUVs[i].y     = 1.0f - pxStrip[i].m_fV;
		}
		Asura_Decal_System::CalculateTangentAndBinormal(axPositions, axUVs, xTangent, xBinormal);

		// The calculation above is incomplete, as it doesn't take the normal into account... so finish the job:
		xBinormal.Normalise();
		xTangent  = xBinormal ^ xNormal;
		xTangent.Normalise();
		xBinormal = xNormal ^ xTangent;
		xBinormal.Normalise();
	}


	// Now convert the strip to an indexed list, in our workspace
	for(uint32 i=0; i<uNumVertices; i++)
	{
		Tools_StaticDecal_Vertex *pxVert = &s_pxWorkspaceVerts[s_iWorkspaceNumVerts+i];

		pxVert->xPos	  = pxStrip[i].m_xPosition;
		pxVert->fU		  = pxStrip[i].m_fU;
		pxVert->fV		  = pxStrip[i].m_fV;			
		pxVert->fCreationTime = 0.0f;
		pxVert->xNormal   = xNormal;
		pxVert->xTangent  = xTangent;
		pxVert->xBinormal = xBinormal;
		if (bCalculateStatiSH)
		{
			Asura_SphericalHarmonics_Order2_RGB xSH;
			Asura_GI_Solution::GetStaticIrradiance(xSH, pxVert->xPos);
			pxVert->xSHRed = Asura_AVector4(xSH.m_afRedCoefficients[1], xSH.m_afRedCoefficients[2], xSH.m_afRedCoefficients[3], xSH.m_afRedCoefficients[0]);
			pxVert->xSHGreen = Asura_AVector4(xSH.m_afGreenCoefficients[1], xSH.m_afGreenCoefficients[2], xSH.m_afGreenCoefficients[3], xSH.m_afGreenCoefficients[0]);
			pxVert->xSHBlue = Asura_AVector4(xSH.m_afBlueCoefficients[1], xSH.m_afBlueCoefficients[2], xSH.m_afBlueCoefficients[3], xSH.m_afBlueCoefficients[0]);
		} else {
			pxVert->xSHRed = Asura_AVector4(0.0f, 0.0f, 0.0f, xDiffuseRGBA.m_fR);
			pxVert->xSHGreen = Asura_AVector4(0.0f, 0.0f, 0.0f, xDiffuseRGBA.m_fG);
			pxVert->xSHBlue = Asura_AVector4(0.0f, 0.0f, 0.0f, xDiffuseRGBA.m_fB);
		}
		pxVert->fAlpha = static_cast<float>(s_uCurrentAlpha) * 1.0f / 255.0f * xDiffuseRGBA.m_fA;
	}

	// loop over polys
	for(uint32 i=0; i<(uNumVertices-2); i++)
	{
		if (i&1)
		{					
			s_piWorkspaceIndices[s_iWorkspaceNumIndices+(i*3)+0] = s_iWorkspaceNumVerts+i;
			s_piWorkspaceIndices[s_iWorkspaceNumIndices+(i*3)+1] = s_iWorkspaceNumVerts+i+1;
			s_piWorkspaceIndices[s_iWorkspaceNumIndices+(i*3)+2] = s_iWorkspaceNumVerts+i+2;		
		}
		else
		{		
			s_piWorkspaceIndices[s_iWorkspaceNumIndices+(i*3)+0] = s_iWorkspaceNumVerts+i+2;
			s_piWorkspaceIndices[s_iWorkspaceNumIndices+(i*3)+1] = s_iWorkspaceNumVerts+i+1;
			s_piWorkspaceIndices[s_iWorkspaceNumIndices+(i*3)+2] = s_iWorkspaceNumVerts+i;					
		}
	}

	// Update the counts
	s_iWorkspaceNumVerts+=uNumVertices;
	s_iWorkspaceNumIndices+=((uNumVertices-2)*3);	
}



// ----------------------------------------------------------------------------------------
// UpdateVertices() - Rebuild vertex/index buffers for a DecalBuffer
// ----------------------------------------------------------------------------------------
void Tools_StaticDecal_System::UpdateVertices(Tools_StaticDecal_Buffer *pxBuffer)
{
	// First we create the decals in a temporary buffer. This buffer will grow as needed
	// (but not be resized down as decals are removed)

	bool bAllowBackFaceProjection = true;

	ClearWorkspaceVerts();

	Asura_DecalManager_Material *pxMaterial = Asura_Decal_Manager::GetMaterial(pxBuffer->m_uDecalMaterial);

	if (pxMaterial)
	{
		bAllowBackFaceProjection = pxMaterial->m_bStaticDecalBackFaceProject;
	}


	Asura_Decal_System::BeginStaticDecalCreation(bAllowBackFaceProjection);	
	
	for(int i=0; i<pxBuffer->m_iNumDecals; i++)
	{		
		Tools_StaticDecal *pxDecal = pxBuffer->m_ppxDecals[i];		

		if (pxDecal == s_pxSelectedDecal)
		{
			s_pxSelectedDecalBuffer    = pxBuffer;
			s_iSelectedDecalFirstIndex = s_iWorkspaceNumIndices;
		}
		
		Asura_DecalManager_DecalSet *pxSet = Asura_Decal_Manager::GetDecalSet(pxDecal->m_uDecalSetHash);
		
		if (pxSet)
		{		
			s_uCurrentAlpha = (uint32)(pxDecal->m_fAlpha * 255.0f);

			uint32 uDecalTemplate = pxSet->m_iFirstAsuraDecalTemplate + pxDecal->m_iDecalIndex;			
			Asura_Decal_System::AddDecal(pxDecal->m_xPosition, pxDecal->m_xOrient, pxDecal->m_fScale, uDecalTemplate);
		}

		if (pxDecal == s_pxSelectedDecal)
		{			
			s_iSelectedDecalNumPolys = (s_iWorkspaceNumIndices-s_iSelectedDecalFirstIndex)/3;
		}
	}

	pxBuffer->m_iNumPolys = s_iWorkspaceNumIndices/3;

	Asura_Decal_System::EndStaticDecalCreation();

	// Now copy the decals into our VB/IB for this DecalBuffer. 
	if (pxBuffer->m_pxVerts)
	{
		delete [] pxBuffer->m_pxVerts;
		pxBuffer->m_pxVerts = NULL;
		pxBuffer->m_iNumVerts = 0;
	}

	if (s_iWorkspaceNumVerts>0)
	{
		pxBuffer->m_pxVerts = new Tools_StaticDecal_Vertex[s_iWorkspaceNumVerts];
		memcpy(pxBuffer->m_pxVerts, s_pxWorkspaceVerts, s_iWorkspaceNumVerts * sizeof(Tools_StaticDecal_Vertex));
		pxBuffer->m_iNumVerts = s_iWorkspaceNumVerts;

		// Update the bounding box
		for(int i=0; i<s_iWorkspaceNumVerts; i++)
		{		
			pxBuffer->m_xAABB.Set(s_pxWorkspaceVerts[i].xPos);
		}
	} 

	if (pxBuffer->m_piIndices)
	{
		delete [] pxBuffer->m_piIndices;
		pxBuffer->m_piIndices = NULL;
		pxBuffer->m_iNumIndices = 0;
	}

	if (s_iWorkspaceNumIndices>0)
	{
		pxBuffer->m_piIndices = new int[s_iWorkspaceNumIndices];
		memcpy(pxBuffer->m_piIndices, s_piWorkspaceIndices, s_iWorkspaceNumIndices * sizeof(int));
		pxBuffer->m_iNumIndices = s_iWorkspaceNumIndices;
	} 

}

#endif // ASURA_USE_STATIC_DECAL_SYSTEM