///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_ContainerState.h
//
// Created by Martin G.Harris (26/6/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_CONTAINER_STATE_H__
#define __TOOLS_CONTAINER_STATE_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Entity/Container/Asura_ContainerState.h"

// Forward declarations
class Asura_Chunk_Stream;
class Tools_Container_ChildIterator;
class Tools_Container;
class Asura_Container_Collection;

// Renaming a state has the habit of breaking container entities and static messages.
// Store the last few hashes, so that we have some chance of fixing stuff up automatically.
const u_int uNUM_OLD_STATE_HASHES_TO_STORE = 8;

//////////////////////////
// Tools_ContainerState //
//////////////////////////
// Object representing a container state definition belonging to a specific collection.
// Each state stores information about the states of any child containers.
class Tools_ContainerState
{
public:
	Tools_ContainerState();
	~Tools_ContainerState();

	Asura_Hash_ID GetStateHash() const { return m_uStateHash; }
	const Asura_String& GetStateName() const { return m_xStateName; }

	u_int GetNumChildStates() const { return m_uNumChildStates; }
 
	void SetInitialStateName(const Asura_String& xName)
	{
		m_xStateName = xName;
		m_uStateHash = m_xStateName.GetHash();
	}

	// Remove the specified input for the child.
	void RemoveChild(Asura_Hash_ID uChildContainerHash, u_int uInputIndex);
	
	// Remove all the inputs for the child.
	void RemoveChildAllInputs(Asura_Hash_ID uChildContainerHash);

	// Add, remove or change as necessary.
	void ChangeChild(Asura_Hash_ID uChildContainerHash, Asura_Hash_ID uChildStateHash, u_int uInputIndex);

	// Replace the child if it exists.
	void ReplaceChild(const Asura_Hash_ID uOldChildHash, const Asura_Hash_ID uNewChildHash);

	// Get child state, returns "off" state hash if child is disabled or unknown.
	Asura_Hash_ID GetChildState(u_int uInputIndex, Asura_Hash_ID uChildContainerHash) const;

	// Access by index.
	const Asura_ContainerStateEntry* GetStateEntry(u_int uIndex) const;

	// Chunkstream
	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	// Note: these must match the corresponding definition in Asura_ContainerState.
	// Hash and string representing the "on" state.
//	static Asura_Hash_ID GetOnStateHash() { return 0x00000ddf; } // Asura_GetHashID("on");
	static const char* GetOnStateName() { return "on"; }

	// Hash and string representing the "off" state.
//	static Asura_Hash_ID GetOffStateHash() { return 0x0001ad6f; } // Asura_GetHashID("off");
	static const char* GetOffStateName() { return "off"; }

	// On/off states can't be modified.
	bool IsInvariant() const;

	// Check for errors. Return number of errors found.
	void Verify(const Tools_Container* pxParent, const Tools_Container_ChildIterator& xIt) const;
	
	// Silently fix up certain errors that the user can't fix.
	void FixUpErrors(const Tools_Container* pxParent);

	// Replaced a state hash associated with a specified container collection. Fix up state.
	// Return true if modified.
	bool ReplaceState(const Asura_Hash_ID uContainerHash, const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash);

	//
	// Damage overrides
	//
	u_int GetNumDamageOverrides() const { return m_uNumDamageOverrides; }

	// Add a damage override.
	void AddDamageOverride(float fMaxHealthPercent, Asura_Hash_ID uStateHash);
	
	// Remove a damage override.
	void RemoveDamageOverride(u_int uIndex);

	// Get damage override by index.
	const Asura_ContainerState_DamageOverride* GetDamageOverride(u_int uIndex) const;

	// Replaced a state hash associated with this container collection. Fix up damage overrides.
	// Return true if modified.
	bool ReplaceDamageOverrideState(const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash);

	// We've renamed this state, store the old hash.
	void SaveOldStateHash(const Asura_Hash_ID uOldStateHash);
	bool HasOldStateHash(const Asura_Hash_ID uOldStateHash) const;

	// Input mask (bit 0 corresponds to input slot 0 etc, set to 1 for enabled).
	u_int* GetInputMask() { return &m_uInputSlotMask; }

	// Is this state enabled for the specified input slot.
	bool IsEnabledForInputSlot(u_int uInputIndex) const;

protected:
	void FreeAll();

	// Add a new child state.
	void AddChild(Asura_Hash_ID uChildContainerHash, Asura_Hash_ID uChildStateHash, u_int uInputIndex);

protected:
	Asura_String m_xStateName;		// Name of the state.
	Asura_Hash_ID m_uStateHash;		// Hash of the state name.

	// Mask indicating for which input slots this state is valid
	// (allows the designer to say that some states only appear in certain slots).
	// This is a Tools-only thing and is only weakly enforced by the UI.
	u_int m_uInputSlotMask;

	// States of child containers.
	u_int m_uNumChildStates;
	Asura_ContainerStateEntry* m_pxChildState;

	// Damage overrides
	u_int m_uNumDamageOverrides;
	Asura_ContainerState_DamageOverride* m_pxDamageOverride;

	// Store the last few state name hashes. Oldest is at the end of the array, newest at the start.
	Asura_Hash_ID m_auOldStateHashes[uNUM_OLD_STATE_HASHES_TO_STORE];
};


/////////////////////////////////
// Tools_ContainerStateMachine //
/////////////////////////////////
// Set of state definitions comprising a state machine.
// Each container collection has its own state machine.
class Tools_ContainerStateMachine
{
public:
	Tools_ContainerStateMachine();
	~Tools_ContainerStateMachine();

	u_int GetNumInputs() const { return m_uNumInputs; } // Number of inputs (usually 1).
	
	// Return true if the number of inputs was changed.
	bool AddInput();
	bool RemoveInput();

	bool AddDefaultStates();

	u_int GetNumStates() const { return m_uNumStates; } // Total number of states.
	u_int GetNumVariantStates() const; // Number of states excluding invariant ones (i.e. on/off).

	const Asura_String& GetStateName(u_int uIndex) const;
	const Asura_Hash_ID GetStateHash(u_int uIndex) const;

	void AddState(const Asura_String& xStateName);
	void RemoveState(Asura_Hash_ID uStateHash);

	// Replace the child for all states.
	void ReplaceChild(const Asura_Hash_ID uOldChildHash, const Asura_Hash_ID uNewChildHash);

	// Remove the child for all states and all inputs.
	void RemoveChildAllInputs(const Asura_Hash_ID uChildContainerHash);
	
	// Get child state, returns "off" state hash if child is disabled or unknown.
	Asura_Hash_ID GetChildState(Asura_Hash_ID uStateHash, Asura_Hash_ID uChildContainerHash, u_int uInputIndex) const;

	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	const Tools_ContainerState* FindState(Asura_Hash_ID uStateHash) const;
	Tools_ContainerState* FindState(Asura_Hash_ID uStateHash);

	const Tools_ContainerState* GetState(u_int uIndex) const;
	Tools_ContainerState* GetState(u_int uIndex);

	// Check for errors. Return number of errors found.
	void Verify(const Tools_Container* pxParent, const Tools_Container_ChildIterator& xIt) const;
	// Silently fix up certain errors that the user can't fix.
	void FixUpErrors(const Tools_Container* pxParent);

	// Does the hash belong to state that has been renamed? If so then return the new state hash.
	Asura_Hash_ID GetHashForRenamedState(const Asura_Hash_ID uOldStateHash) const;

	// Copy the data to the Asura container collection.
	void CopyDataToAsura(Asura_Container_Collection* pxAsuraCollection) const;

	// Replaced a state hash associated with a specified container collection. Fix up state machine.
	// Return true if modified.
	bool ReplaceState(const Asura_Hash_ID uContainerHash, const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash);

	// Replace the state hash in the damage overrides.
	// Return true if modified.
	bool ReplaceDamageOverrideState(const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash);

protected:
	void FreeAll();

	// Find index of state or return -1 if not found.
	int FindStateIndex(Asura_Hash_ID uStateHash) const;
	bool IsValid(int iIndex) const;

protected:
	// States
	u_int m_uNumStates;
	Tools_ContainerState** m_ppxState;

	// Inputs
	u_int m_uNumInputs;
};


////////////////////////////////////
// Tools_ContainerSortedStateList //
////////////////////////////////////
// Generate an alphabetically sorted state list for the given state machine.
class Tools_ContainerSortedStateList
{
public:
	Tools_ContainerSortedStateList(Tools_ContainerStateMachine& xStateMachine);
	~Tools_ContainerSortedStateList();

	u_int GetNumStates() const { return m_xStateMachine.GetNumStates(); }

	// Return the state for the given index.
	Tools_ContainerState* GetState(u_int uIndex) const;

private:
	static int Compare_Callback(const void* pFirst, const void* pSecond);
	int Compare(u_int uFirst, u_int uSecond) const;

private:
	Tools_ContainerStateMachine& m_xStateMachine;
	u_int* m_puIndices;

	static Tools_ContainerSortedStateList* s_pxThis;
};

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__TOOLS_CONTAINER_STATE_H__

