///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Container_System.h
//
// Created by Martin G.Harris (6/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_CONTAINER_SYSTEM_H__
#define __TOOLS_CONTAINER_SYSTEM_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_Maths.h"
#include "Core/Asura_HashTree.h"
#include "Interface/Asura_Tree_Dialog.h"

#include "Entity/Container/Tools_Container.h"

// Forward declarations
class Asura_Overlay;
class Tools_ModeDataFile;

/////////////////////////////
// Tools_Container_Options //
/////////////////////////////
// Some user definable options that are saved out.
class Tools_Container_Options
{
public:
	Tools_Container_Options();

	bool ShouldDisplaySoundEvents()	  const { return m_bDisplaySoundEvents; }

	float GetEntityOriginMeshSize()	  const { return m_fEntityOriginMeshSize; }
	float GetModeOriginMeshSize()	  const { return m_fModeOriginMeshSize; }
	float GetModeAttachmentMeshSize() const { return m_fModeAttachmentMeshSize; }

protected:
	bool m_bDisplaySoundEvents;

	float m_fEntityOriginMeshSize;
	float m_fModeOriginMeshSize;
	float m_fModeAttachmentMeshSize;

	// Friends
	friend class Tools_Container_Overlay_Manager;
};

//////////////////////////////////
// Tools_ContainerItemReference //
//////////////////////////////////
// Helper class for referring to a selected item.
class Tools_ContainerItemReference
{
public:
	Tools_ContainerItemReference(const Asura_TreeDialog_Item* pxItem);

	Asura_Hash_ID m_uFileHash;
	Asura_Hash_ID m_uGroupHash;
	Asura_Hash_ID m_uContainerHash;

	Tools_ContainerFile*  m_pxFile;
	Tools_ContainerGroup* m_pxGroup;
	Tools_Container*	  m_pxContainer;
};

/////////////////////////////////////
// Tools_Container_Overlay_Manager //
/////////////////////////////////////
// Handles all the overlay related code for the main tree of groups and containers.
class Tools_Container_Overlay_Manager
{
public:
	Tools_Container_Overlay_Manager();

	// Main tree overlay.
	bool IsOverlayOpen() const { return m_bOverlayIsOpen; }
	void ToggleOverlay();

	void CreateOverlay();
    void DestroyOverlay();
	void UpdateOverlay();

	void AddOverlayItems(Asura_Overlay& xOverlay, Asura_Vector_2& xOffset);

	Asura_Vector_2* GetPosition() { return &m_xOverlayPosition; }

	// Options
	const Tools_Container_Options& GetOptions() const { return m_xOptions; }

	// Refresh things that have been renamed.
	void RefreshContainer(const Tools_Container* pxContainer);
	void RefreshGroup(const Tools_ContainerGroup* pxGroup);

	// Close gadget on the overlay has been clicked.
	void ClosingByCloseGadget();

	// Load/save data concerning the overlay, such as, position.
	void TransferModeData(const int iMode);

	// Selection.
	static bool IsFileSelectionValid()		{ return ASURA_HASH_ID_UNSET != s_uSelectedFileHash; }
	static bool IsGroupSelectionValid()		{ return ASURA_HASH_ID_UNSET != s_uSelectedGroupHash; }
	static bool IsContainerSelectionValid()	{ return ASURA_HASH_ID_UNSET != s_uSelectedContainerHash; }
	static bool IsGroupOnlySelectionValid() { return IsGroupSelectionValid() && !IsContainerSelectionValid(); }

	static bool IsFileSelectionValidAndWriteable()		{ return IsFileSelectionValid() && IsFileWriteable(); }
	static bool IsGroupSelectionValidAndWriteable()		{ return IsGroupSelectionValid() && IsFileWriteable(); }
	static bool IsContainerSelectionValidAndWriteable()	{ return IsContainerSelectionValid() && IsFileWriteable(); }
	static bool IsGroupOnlySelectionValidAndWriteable() { return IsGroupOnlySelectionValid() && IsFileWriteable(); }

	static bool IsContainerValidForMoving();
	static bool IsContainerValidForOpenInMode();
	static bool IsGroupValidForOpenInMode();

	static bool IsFileWriteable();

	static Asura_Hash_ID GetSelectedFileHash()		{ return s_uSelectedFileHash; }
	static Asura_Hash_ID GetSelectedGroupHash()		{ return s_uSelectedGroupHash; }
	static Asura_Hash_ID GetSelectedContainerHash() { return s_uSelectedContainerHash; }

	// Callbacks
	static void TreeDoubleClicked();
	static void DragAndDropTreeItem(Asura_TreeDialog_Item* pxDropTarget);
	static int GetTreeItemIconIndex(Asura_TreeDialog_Item* pxItem);
	static void RenameTreeItem(Asura_TreeDialog_Item* pxItem, const char* szName);

	// Tree.
	Asura_TreeDialog_Tree& GetTree() { return m_xTree; }
		
	Asura_TreeDialog_Item* GetFileItem(Asura_Hash_ID uFileHash);
	Asura_TreeDialog_Item* GetGroupItem(Asura_Hash_ID uGroupHash);
	Asura_TreeDialog_Item* GetContainerItem(Asura_Hash_ID uContainerHash);


	// Editing overlay.
	void OpenEditOverlay(Asura_Hash_ID uEditContainerHash, bool bSortToFront = true);
	void CloseEditOverlay();
	bool IsEditOverlayOpen() { return m_bEditOverlayIsOpen; }
	void ToggleEditOverlay();

	void EditSelectedContainer() { OpenEditOverlay(s_uSelectedContainerHash); }
	void ClosingEditByCloseGadget();

	Asura_Vector_2& GetEditPosition() { return m_xEditOverlayPosition; }
	Asura_Hash_ID GetEditingContainerHash()	{ return m_uEditingContainerHash; }

	void BringToFrontNextUpdate() { m_bEditOverlayBringToFront = true; }
	void RebuildEditOverlay() { m_bEditOverlayNeedToRebuild = true; }

protected:
	// Non-recursive search for item.
	Asura_TreeDialog_Item* FindItemByData(Asura_TreeDialog_Item* pxStartItem, Asura_Hash_ID uSearchHash);

	void TransferToolsModeData( Tools_ModeDataFile& xFile );

	// Move group to another file.
	static void MoveGroup(Asura_Hash_ID uSelectedGroupHash, Asura_Hash_ID uTargetFileHash);

	// Move container to another group.
	static void MoveContainer(Asura_Hash_ID uSelectedContainerHash, Asura_Hash_ID uTargetGroupHash);

	static Asura_Hash_ID GetSingleGlobalCollectionFromGroup();

	static void AddContextMenu(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext);

	static void ContextMenu_Edit();
	static void ContextMenu_Rename();
	static void ContextMenu_Delete();
	static void ContextMenu_ToggleScope();
	static void ContextMenu_CreateFile();
	static void ContextMenu_CreateGroup();
	static void ContextMenu_CreateContainer();
	static void ContextMenu_OpenInMode();
	static void ContextMenu_OpenGroupInMode();

	enum { CURRENT_SAVE_VERSION = 3 };

protected:
	//  Main tree overlay.
	Asura_Vector_2 m_xOverlayPosition;
	bool m_bOverlayIsOpen;
	bool m_bModeDataLoaded;
	Asura_Vector_2 m_xTreeDialogueDimensions;

	// Options
	Tools_Container_Options m_xOptions;

	// Current selection.
	Asura_Dialog* m_pxSelectFileDialog;
	Asura_Dialog* m_pxSelectGroupDialog;
	Asura_Dialog* m_pxSelectContainerDialog;
	Asura_Dialog* m_pxSelectTypeDialog;

	// Tree of groups and containers.
	Asura_TreeDialog_Tree m_xTree;

	// Editing overlay.
	bool m_bEditOverlayIsOpen;
	bool m_bEditOverlayBringToFront;
	bool m_bEditOverlayNeedToRebuild;
	Asura_Vector_2 m_xEditOverlayPosition;
	Asura_Hash_ID m_uEditingContainerHash;

	// Static data
	static Asura_Hash_ID s_uSelectedFileHash;
	static Asura_Hash_ID s_uSelectedGroupHash;
	static Asura_Hash_ID s_uSelectedContainerHash;


	enum TAB_NAMES
    {
        TAB_EDITING,
		TAB_DISPLAY,
        TAB_TOOLS,

        NUM_TAB_NAMES
    };
    static const char* const s_aszTabNames[ NUM_TAB_NAMES ];
};

////////////////////////////
// Tools_Container_System //
////////////////////////////
class Tools_Container_System
{
public:
	static void Initialise();
	static void Shutdown();
	static void Update();

	static bool IsInitialised() { return s_bInitialised; }

	// Overlay
	// Main overlay
	static void ToggleOverlay() { s_xOverlayManager.ToggleOverlay(); }
	static bool	IsOverlayOpen() { return s_xOverlayManager.IsOverlayOpen(); }
	static Tools_Container_Overlay_Manager& GetOverlayManager() { return s_xOverlayManager; }

	// Editing overlay
	static void ToggleEditOverlay() { s_xOverlayManager.ToggleEditOverlay(); }
	static bool	IsEditOverlayOpen() { return s_xOverlayManager.IsEditOverlayOpen(); }
	static Asura_Vector_2& GetEditPosition() { return s_xOverlayManager.GetEditPosition(); }
	static void ClosingEditByCloseGadget() { s_xOverlayManager.ClosingEditByCloseGadget(); }

	// Options
	static const Tools_Container_Options& GetOptions() { return s_xOverlayManager.GetOptions(); }
	static bool ShouldDisplaySoundEvents() { return GetOptions().ShouldDisplaySoundEvents(); }
	static bool ShouldRenderPFX();

	// Modification/save
	static bool IsModified();
	static void SaveModified();
	static void SavePrompt(bool bLastChanceToSave);
	static void SetFileModified(Asura_Hash_ID uFileHash);

	static void WarnIfContainerIsReadOnly(Asura_Hash_ID uContainerHash);

	// Callbacks
	static void CreateNewFile();
	static void CreateNewGroup();
	static void CreateNewContainer();
	static void CreateNewContainerFromMode(const u_int uInsertClass);
	static Tools_Container* CreateNewContainerAsCopyFromMode(Asura_Hash_ID uCopyContainerHash, Tools_Container* pxDestinationContainer);

	static void MoveSelectedGroup();
	static void RenameSelectedGroup();
	static void DeleteSelectedGroup();

	static void EditSelectedContainer();
	static void MoveSelectedContainer();
	static void RenameSelectedContainer();
	static void DeleteSelectedContainer();

	static void SaveAndCloseOverlay();
	static void Reload();

	static void VerifyContainerInformation();
	static void IncrementErrorCount();

	static void ToggleScope();
	static bool IsValidForToggleScope();

	// Selection
	static void SelectContainer(Asura_Hash_ID uContainerHash, bool bOpenEditOverlay);
	static void BringEditOverlayToFrontNextUpdate() { s_xOverlayManager.BringToFrontNextUpdate(); }

	// Choose container from list.
	// Input:
	// Set uContainerHash   to the initial selection (which will be highlighted in the list).
	// Set uRemoveContained to exclude all containers that contain the specified container.
	//	                    Used to prevent circular references from being set up.  
	// Set bIncludeNone     to include a "*None*" option (if selected returns ASURA_HASH_ID_UNSET).
	// Set bCollectionOnly  to include only collection containers.
	// Set uGroupScopeHash  to include containers in the specified group-scope,
	//					    otherwise only global-scope containers will be included.
	// Output:
	// Returns false if cancelled, true if selection made.
	// uContainerHash contains selected value.
	static bool SelectContainerModalOverlay(Asura_Hash_ID& uContainerHash, const Asura_Hash_ID uRemoveContained,
											bool bIncludeNone, bool bCollectionOnly, Asura_Hash_ID uGroupScopeHash);

	// Choose container collection or create option. Return false if cancelled.
	static bool CreateOrSelectModalOverlay(Asura_Hash_ID& uContainerHash);

	// Create a collection. Return false if cancelled.
	static bool CreateCollectionModalOverlay(Asura_Hash_ID& uContainerHash);

	// Select a file. Option to only list writeable files. Return false if cancelled.
	static bool SelectFileModalOverlay(Asura_Hash_ID& uFileHash, bool bOnlyListWriteable);

	// Select a group from the specified file. Optional create group option. Return false if cancelled.
	static bool SelectGroupModalOverlay(const Asura_Hash_ID uFileHash, Asura_Hash_ID& uGroupHash, bool bCreateGroupOption);

	// Create new group in specified file.  Return false if cancelled. Returns hash of created group name.
	static bool CreateNewGroup(const Asura_Hash_ID uFileHash, Asura_Hash_ID& uCreatedGroupHash);

	// Create new collection container in specified file and group.  Return false if cancelled. Returns hash of created container name.
	static bool CreateNewCollection(const Asura_Hash_ID uFileHash, const Asura_Hash_ID uGroupHash,
									Asura_Hash_ID& uContainerHash, const char* szDefaultName = 0);

	// Select a container classification. Return true if Ok, false if cancelled.
	static bool SelectContainerClassModalOverlay(u_int& uClass, const u_int uExcludeClass = AsuraContainerClass_NotSet);

	// Select a marker point. Return true if Ok, false if cancelled.
	static bool SelectMarkerPointModalOverlay(Asura_String& xMarkerName, Asura_Hash_ID uGroupHash);

	// Search
	static Tools_ContainerFile*  FindFileByHash(Asura_Hash_ID uFilenameHash);
	static Tools_ContainerGroup* FindGroupByHash(Asura_Hash_ID uGroupHash);
	static Tools_Container* FindContainerByHash(Asura_Hash_ID uContainerHash);

	static Tools_ContainerFile* FindFileByFilename(const Asura_String& xFilename);
	static Tools_ContainerFile* FindFileByShortNameHash(Asura_Hash_ID uShortNameHash);

	static Tools_Container* FindEditingContainer()
	{
		return FindContainerByHash(s_xOverlayManager.GetEditingContainerHash());
	}

	// Get file tree
	static const Asura_TemplateHashTree<Tools_ContainerFile>& GetFileTree() { return s_xFileTree; }

	// Add new file or load an old one.
	static void RegisterFile(Tools_ContainerFile& xFile, bool bNewFile = false);
	// Add new group or load an old one.
	static void RegisterGroup(Tools_ContainerGroup& xGroup, bool bNewGroup = false);
	// Add new container or load an old one. Return true if successful.
	static bool RegisterContainer(Tools_Container* pxContainer, bool bNewContainer = false, bool bOpenEditOverlay = false);

	// Verification
	static const char* GetCategory();
	static const char* GetSubCategory();

	// Helpers
	
	// Check whether the specified container (uContainerHash) equals or contains the searched for container (uSearchHash).
	// Return true if it does.
	static bool ContainerSearch(const Asura_Hash_ID uContainerHash, const Asura_Hash_ID uSearchHash);
	
	// Container has been modified, refresh the entities.
	static void SetContainerModified(const Asura_Hash_ID uContainerChanged, bool bMinorChange);

	// The specified attachment has become a marker attachment. Copy child hash data for any collections that contain it.
	static void CopyDataForMarkerAttachment(const Asura_Hash_ID uContainerHash);

	// Helper for container creation.
	static Tools_Container* CreateNewContainerFromMode_Helper(u_int uClassification,
															  u_int uAutoGenNumber,
															  const char* const szParentName,
															  const Tools_ContainerGroup* const pxGroup);
	
	// Delete a container if it is orphaned. Only delete those with group-scope.
	static void DeleteIfOrphaned(const Asura_Hash_ID uContainerHash, bool bPromptUser = false);
	// Return true if the container is known to be orphaned.
	static bool IsOrphaned(const Asura_Hash_ID uContainerHash);
	static bool IsOrphaned(const Tools_Container* pxContainer);

	// Render any tools specific stuff.
	static void ToolsRender(const Asura_Hash_ID uContainerHash, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation);

	// Replaced a state hash associated with a specified container collection. Fix up other state machines.
	static void ReplaceState(const Asura_Hash_ID uContainerHash, const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash);

	// Change rendering of PFX in Tools.
	static void SetShowPFX(bool bShowPFX);

	// Helpers for renaming containers and groups.
	static bool IsContainerNameValid(Asura_String& xUserString, const Tools_ContainerGroup* pxGroup,
									 Asura_Hash_ID uOldContainerHash);
	static bool IsGroupNameValid(Asura_String& xUserString, const Tools_ContainerGroup& xGroup, bool bNewGroup);

	// Pop up a dialogue to rename the specified container.
	static void RenameContainerUsingDialogue(Tools_Container* pxContainer);

	// Set container/group name. Need to check that the new name is valid before calling.
	static void RenameContainer(const Asura_String& xNewName, Tools_Container* pxContainer);
	static void RenameGroup(const Asura_String& xNewName, Tools_ContainerGroup* pxGroup);

	static void MoveSelectedGroup_Helper(Tools_ContainerGroup* pxGroup, Tools_ContainerFile* pxOldFile, Tools_ContainerFile* pxNewFile);
	static void MoveSelectedContainer_Helper(Asura_Hash_ID uSelectedContainerHash, Asura_Hash_ID uTargetGroupHash);

	// Can this container be moved to a new group.
	static bool IsContainerValidForMoving(Asura_Hash_ID uContainerHash);

	// Return true if Ok, i.e. no errors.
	static bool VerifyEntity(const class Asura_ServerEntity_Container* pxServer);
	static bool VerifyContainer(const Tools_Container* pxContainer);
	static bool VerifyContainerMessages();

	// Helper for "open group in container mode".
	static Asura_Hash_ID GetSingleGlobalCollectionFromGroup(Asura_Hash_ID uGroupHash);

	// Is the specified container state valid? uContainerHash should refer to a collection.
	static bool IsContainerStateValid(Asura_Hash_ID uContainerHash, Asura_Hash_ID uStateHash, u_int uInputIndex);
	static bool IsContainerStateValid(Asura_Hash_ID uContainerHash, const Asura_Container_InputState& xInputState);

private:
	// Load/delete
	static void LoadAllFiles();
	static void PostLoadInitialise();
	static void DeleteAllContainers();

	// Set bPromptUser to prompt user.
	// Set bRecursive to recursively delete orphaned containers with group-scope.
	static bool DeleteContainer(const Asura_Hash_ID uContainerHash, bool bPromptUser, bool bRecursive);

	// Save group of containers to a file.
	static void SaveFile(Tools_ContainerFile* pxFile);

	// Dialogue tree functions.
	static void AddFileToDialogueTree(const Tools_ContainerFile* pxFile, bool bSelect = false);
	static void AddGroupToDialogueTree(const Tools_ContainerGroup* pxGroup, bool bSelect = false);
	static void AddContainerToDialogueTree(const Tools_Container* pxContainer, Asura_Hash_ID uGroupHash, bool bSelect = false);
	
	static void RemoveGroupFromDialogueTree(const Tools_ContainerGroup* pxGroup);
	static void RemoveContainerFromDialogueTree(const Tools_Container* pxContainer);

	// Set name functions.
	static bool SetFilenameDialogue(Tools_ContainerFile& xFile);
	static bool SetGroupNameDialogue(Tools_ContainerGroup& xGroup, bool bNewGroup);

	static bool SetContainerNameDialogue(Asura_String& xString, const Tools_ContainerGroup* pxGroup,
										 Asura_Hash_ID uOldContainerHash, bool bAllowAutoAccept = false);

	static void VerifyEntities();
	static void VerifyStaticMessages(bool bLogFixes = true);

	static bool IsValidForScopeChangeToGroup(const Tools_Container* pxContainerToChange);

private:
	static Tools_Container_Overlay_Manager s_xOverlayManager;
	static bool s_bInitialised;

	static Asura_TemplateHashTree<Tools_ContainerFile>	s_xFileTree;
	static Asura_TemplateHashTree<Tools_ContainerGroup> s_xGroupTree;
	static Asura_TemplateHashTree<Tools_Container*>		s_xContainerTree;

	// Filename of the file currently being processed.
	static const char* s_pxCurrentFilename;

	static int s_iNumErrors;
};



#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__TOOLS_CONTAINER_SYSTEM_H__

