///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Entity_Container.cpp
//
// Created by Martin G.Harris (6/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Tools_Entity_Container.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "Core/Asura_LOS.h"
#include "Dynamics/Asura_Collision_Object.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_Entity_Event.h"
#include "Entity/Container/Asura_ClientContainerInstance.h"
#include "Entity/Container/Asura_Container_System.h"
#include "Entity/Container/Asura_ContainerRegistry.h"
#include "Entity/Container/Asura_ServerEntity_Container.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Overlay_System.h"
#include "Mesh/Asura_Mesh.h"
#include "Network/Asura_Message.h"
#include "Physics/Asura_Physics_Object.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Render.h"

// Tools
#include "Tools_System.h"
#include "Entity/Tools_Entity_Data.h"
#include "Entity/Container/Tools_Container.h"
#include "Entity/Container/Tools_ContainerRegistry.h"
#include "Entity/Container/Tools_ContainerState.h"
#include "Entity/Container/Tools_Mode_Container.h"
#include "Entity/Container/Tools_Container_System.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Waypointing/Tools_Mode_Waypointing.h"

// Register the Tools entity.
Tools_ContainerRegisterToolsEntity<Tools_Entity_Container> xRegisterEntity(false);

///////////////////////////////////////////////////////////////////////////////
// Tools_ContainerState_DropDownList_Dialog
///////////////////////////////////////////////////////////////////////////////

// List of drop down list dialogues.
Asura_List<Tools_ContainerState_DropDownList_Dialog> Tools_ContainerState_DropDownList_Dialog::s_xList;

Tools_ContainerState_DropDownList_Dialog::Tools_ContainerState_DropDownList_Dialog( const char* szName,
																				    const Asura_Vector_2& xOffset,
																				    const Asura_Hash_ID* puContainerHash,
																				    Asura_Hash_ID* puStateHash,
																				    bool* pbChanged,
																					bool bAlwaysAddOn,
																					bool bAlwaysAddOff,
																					bool bAlwaysAllowNone,
																					u_int uInputIndex,
																					const u_int* puInputIndex)
: Asura_Hash_DropDownList_Dialog(xOffset)
, m_puContainerHash(puContainerHash)
, m_uNumStates(0)
, m_pxStateMachine(0)
, m_iValidCount(0)
, m_bAlwaysAddOn(bAlwaysAddOn)
, m_bAlwaysAddOff(bAlwaysAddOff)
, m_bAlwaysAllowNone(bAlwaysAllowNone)
, m_iOnIndex(-1)
, m_iOffIndex(-1)
, m_uInputIndex(uInputIndex)
, m_puInputIndex(puInputIndex)
{
	InitialiseFromContainer();
	Initialise(xOffset, 10, puStateHash, pbChanged, szName);

	SetOnlyUpdateSelectionIndexOnClick(true);

	s_xList.AddToList(this);
}

Tools_ContainerState_DropDownList_Dialog::~Tools_ContainerState_DropDownList_Dialog()
{
	s_xList.RemoveFromList(this);
}

void Tools_ContainerState_DropDownList_Dialog::PerFrameUpdate()
{
	Asura_Hash_DropDownList_Dialog::PerFrameUpdate();

	if (m_puInputIndex)
	{
		if (*m_puInputIndex != m_uInputIndex)
		{
			// Input slot has been changed. Refresh.
			m_uInputIndex = *m_puInputIndex;
			Refresh();
		}
	}
}

// The states may have changed refresh all of the dialogues.
void Tools_ContainerState_DropDownList_Dialog::RefreshAllDialogues()
{
	for (Asura_List_It<Tools_ContainerState_DropDownList_Dialog> xIt(s_xList); 
		 !xIt.Done();
		 xIt.Next())
	{
		Tools_ContainerState_DropDownList_Dialog* pxDialogue = xIt.GetCurrent();
		if (pxDialogue)
		{
			pxDialogue->Refresh();
		}
	}
}

const char* Tools_ContainerState_DropDownList_Dialog::GetNameByIndex(int iIndex) const
{
	if (iIndex == m_iOnIndex)
	{
		return Tools_ContainerState::GetOnStateName();
	}
	else if (iIndex == m_iOffIndex)
	{
		return Tools_ContainerState::GetOffStateName();
	}

	if (m_pxStateMachine)
	{
		const u_int uNumInputs = m_pxStateMachine->GetNumInputs();
		const Tools_ContainerState* const pxState = m_pxStateMachine->GetState(iIndex);
		if (pxState)
		{
			if (m_uInputIndex >= uNumInputs || !pxState->IsEnabledForInputSlot(m_uInputIndex))
			{
				const Asura_Hash_ID uStateHash = pxState->GetStateHash();
				const bool bAllowOn  = (m_bAlwaysAddOn  && uStateHash == Asura_ContainerState::GetOnStateHash());
				const bool bAllowOff = (m_bAlwaysAddOff && uStateHash == Asura_ContainerState::GetOffStateHash());

				// If m_bAlwaysAddOn/Off is set then don't skip the state if it is on/off.
				if ( !bAllowOn && !bAllowOff )
				{
					// State is disallowed for this input slot. Skip it.
					return 0;
				}
			}
		}

		return m_pxStateMachine->GetStateName(iIndex).GetString();
	}

	return GetUnknownHashText();
}

void Tools_ContainerState_DropDownList_Dialog::Refresh()
{
	++m_iValidCount;
	InitialiseFromContainer();
}

void Tools_ContainerState_DropDownList_Dialog::InitialiseFromContainer()
{
	m_uNumStates = 0;
	m_pxStateMachine = 0;

	if (!m_puContainerHash) return;

	Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(*m_puContainerHash);
	if (!pxContainer) return;
	
	m_pxStateMachine = pxContainer->GetStateMachine();
	if (!m_pxStateMachine) return;
	
	m_uNumStates = m_pxStateMachine->GetNumStates();

	// If requested then add "on" and "off" if they are missing.
	if (m_bAlwaysAddOn)
	{
		if (m_pxStateMachine->FindState(Asura_ContainerState::GetOnStateHash()) == 0)
		{
			m_iOnIndex = static_cast<int>(m_uNumStates);
			++m_uNumStates;
		}
	}

	if (m_bAlwaysAddOff)
	{
		if (m_pxStateMachine->FindState(Asura_ContainerState::GetOffStateHash()) == 0)
		{
			m_iOffIndex = static_cast<int>(m_uNumStates);
			++m_uNumStates;
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
// Tools_ContainerState_MessageEditing_Dialog
///////////////////////////////////////////////////////////////////////////////
// Drop down dialogue for selecting container state in Tools message editing.

// Tools_ContainerState_MessageEditing_Dialog
Tools_ContainerState_MessageEditing_Dialog::Tools_ContainerState_MessageEditing_Dialog(const char* szName,
																					   Asura_Vector_2& xOffset,
																					   Asura_Guid uEntity,
																					   Asura_Hash_ID* puStateHash)
: Tools_ContainerState_DropDownList_Dialog(szName, xOffset, GetContainerHashPointer(uEntity), puStateHash, 0, false, false, true, 0, 0)
{
}


Tools_ContainerState_MessageEditing_Dialog::Tools_ContainerState_MessageEditing_Dialog(const char* szName,
																					   Asura_Vector_2& xOffset,
																					   Asura_Guid uEntity,
																					   Asura_Hash_ID* puStateHash,
																					   const u_int* puInputIndex)
: Tools_ContainerState_DropDownList_Dialog(szName, xOffset, GetContainerHashPointer(uEntity), puStateHash, 0, false, false, true,
										   (puInputIndex ? *puInputIndex : 0), puInputIndex)
{
}

// Function to force the container hash to be initialised before the base constructor is called.
Asura_Hash_ID* Tools_ContainerState_MessageEditing_Dialog::GetContainerHashPointer(Asura_Guid uEntity)
{
	const Asura_ServerEntity_Container* pxServerEntity = Asura_ServerEntity_Container::FindEntity(uEntity);
	Asura_Assert(pxServerEntity, "Server entity not found.");
	if (pxServerEntity)
	{
		m_uContainerHash = pxServerEntity->GetContainerHash();
	}
	else
	{
		m_uContainerHash = ASURA_HASH_ID_UNSET;
	}
	return &m_uContainerHash;
}


//////////////////////////////////////////
// Tools_Container_MultipleState_Dialog //
//////////////////////////////////////////

const u_int uMAX_NUM_INPUT_NUMBER_STRINGS = 10;
static const char* s_aszInputNumberStrings[uMAX_NUM_INPUT_NUMBER_STRINGS] =
{
	"(0) ", "(1) ", "(2) ", "(3) ", "(4) ", "(5) ", "(6) ", "(7) ", "(8) ", "(9) "
};


// Add dialogues for multiple container input states to the overlay.
Tools_Container_MultipleState_Dialog::Tools_Container_MultipleState_Dialog(	const char* szBaseName,
																			Asura_Overlay& xOverlay,
																			Asura_Vector_2& xOffset,
																			const Asura_Hash_ID* puContainerHash,
																			const u_int uNumInputStates,
																			Asura_Hash_ID* puStateHash,
																			bool* pbChanged,
																			bool bInstanceOverlay)
	: Asura_Dialog(szBaseName, xOffset)
	, m_fMaxCoordX(xOffset.x)
	, m_bInstanceOverlay(bInstanceOverlay)
	, m_bChanged(false)
	, m_pbChanged(pbChanged ? pbChanged : &m_bChanged)
{
	// Hide the parent dialogue which is unused.
	SetHidden(true);

	const bool bAlwaysAddOn		= false;
	const bool bAlwaysAddOff	= m_bInstanceOverlay;
	const bool bAlwaysAllowNone = !m_bInstanceOverlay;

	for (u_int u=0; u<uNumInputStates; ++u)
	{
		// Use the base name for the first state and number the rest.
		const char* const szTitle = ((u>0 && u<uMAX_NUM_INPUT_NUMBER_STRINGS) ? s_aszInputNumberStrings[u] : szBaseName);
		Asura_Dialog* pxDialog = new Tools_ContainerState_DropDownList_Dialog(szTitle, xOffset, puContainerHash, &puStateHash[u],
																			  m_pbChanged, bAlwaysAddOn, bAlwaysAddOff, bAlwaysAllowNone, u, 0);
		xOverlay.AddDialog(pxDialog);
		m_xChildDialogueList.AddToList(pxDialog);

		m_fMaxCoordX = Asura_Maths::Max(m_fMaxCoordX, pxDialog->GetMaxCoords().x);
	
		if (u == uNumInputStates-1)
		{
			xOffset.y += Asura_Dialog::LargeSpacing();
		}
		else
		{
			xOffset.y += Asura_Dialog::DefaultSpacing();
		}
	}
}

// Add dialogues for multiple container input states to the overlay.
Tools_Container_MultipleState_Dialog::Tools_Container_MultipleState_Dialog(Asura_Subwindow_Dialog* pxSubwindowDialog,
																		   bool bEnabled,
																		   Asura_Vector_2& xOffset,
																		   const Asura_Hash_ID* puContainerHash,
																		   const u_int uNumInputStates,
																		   Asura_Hash_ID* puStateHash)
	: Asura_Dialog("", xOffset)
	, m_fMaxCoordX(xOffset.x)
	, m_bInstanceOverlay(false)
	, m_bChanged(false)
	, m_pbChanged(&m_bChanged)
{
	// Hide the parent dialogue which is unused.
	SetHidden(true);

	const bool bAlwaysAddOn	    = false;
	const bool bAlwaysAddOff	= true;
	const bool bAlwaysAllowNone = false;

	for (u_int u=0; u<uNumInputStates; ++u)
	{
		// Label the states if there is more than one.
		const char* const szTitle = (uNumInputStates>1 && u<uMAX_NUM_INPUT_NUMBER_STRINGS) ? s_aszInputNumberStrings[u] : "";
		Asura_Dialog* pxDialog = new Tools_ContainerState_DropDownList_Dialog(szTitle, xOffset, puContainerHash, &puStateHash[u],
																			  m_pbChanged, bAlwaysAddOn, bAlwaysAddOff, bAlwaysAllowNone, u, 0);

		pxDialog->SetEnabled(bEnabled);
		pxSubwindowDialog->AddDialog(pxDialog);
		m_xChildDialogueList.AddToList(pxDialog);

		m_fMaxCoordX = Asura_Maths::Max(m_fMaxCoordX, pxDialog->GetMaxCoords().x);
	
		if (u != uNumInputStates-1)
		{
			xOffset.y += Asura_Dialog::DefaultSpacing();
		}
	}
}

////////////////////
// PerFrameUpdate //
////////////////////
void Tools_Container_MultipleState_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	if (m_bInstanceOverlay && HasChanged())
	{
		// When any sub-dialogue changes we need to refresh the instance overlay.
		*m_pbChanged = false;
		Tools_Mode_Container::RequestPreviewStateFromDialogue();
	}
}

////////////////////////////
// SetHelpTextForChildren //
////////////////////////////
void Tools_Container_MultipleState_Dialog::SetHelpTextForChildren(const char* szHelpText)
{
	for (Asura_List_It<Asura_Dialog> xIt(m_xChildDialogueList); !xIt.Done(); xIt.Next())
	{
		Asura_Dialog* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->SetHelpText(szHelpText);
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
// Tools_Entity_Container
///////////////////////////////////////////////////////////////////////////////

// Static data
Asura_Hashed_List<Tools_Entity_Container> Tools_Entity_Container::s_xList(6);

u_int Tools_Entity_Container::s_uOutlineColour = 0;

// Tools_Entity_Container
Tools_Entity_Container::Tools_Entity_Container(Asura_Entity* pxEntity)
: Tools_Entity_Object(pxEntity)
, m_pxServerEntity(0)
, m_pxInstance(0)
, m_bBlocksWaypointGeneration(false)
, m_bChangeInputState(false)
{
	Type = TOT_Entity_Container;
	SetMesh("ToolsShape_Diamond");
	SetScale(Tools_Container_System::GetOptions().GetEntityOriginMeshSize());

	if (pxEntity && pxEntity->GetClassification() == AsuraEntityClass_Container)
	{
		m_pxServerEntity = static_cast<Asura_ServerEntity_Container*>(pxEntity);
	}

	Asura_Assert(m_pxServerEntity, "Server entity null.");

	// Load any skins, anims etc.
	LoadRequiredFiles();

	s_xList.AddToList(this);

	// Try to fix up errors caused by renamed states, before creating the instance.
	FixUpErrors();

	CreateInstance();
}

Tools_Entity_Container::~Tools_Entity_Container()
{
	s_xList.RemoveFromList(this);
	DestroyInstance();
}

//////////////////////
// GetContainerHash //
//////////////////////
Asura_Hash_ID Tools_Entity_Container::GetContainerHash() const
{
	if (m_pxServerEntity)
	{
		return m_pxServerEntity->GetContainerHash();
	}

	return ASURA_HASH_ID_UNSET;
}

///////////////////
// GetInputState //
///////////////////
const Asura_Container_InputState& Tools_Entity_Container::GetInputState() const
{
	if (m_pxServerEntity)
	{
		return m_pxServerEntity->GetInputState();
	}

	return Asura_Container_InputState::Off();
}

/////////////////////
// DestroyInstance //
/////////////////////
// Destroy the instance.
void Tools_Entity_Container::DestroyInstance()
{
	Asura_Container_System::DestroyInstance(m_pxInstance);
	m_pxInstance = 0;
}

////////////////////
// CreateInstance //
////////////////////
// Create an instance.
void Tools_Entity_Container::CreateInstance()
{
	// Destroy any existing instance first.
	DestroyInstance();

	Asura_Assert(m_pxServerEntity, "Null server entity pointer");
	if (!m_pxServerEntity) return;

	// Check that the container exists.
	const Tools_Container* pxToolsContainer = GetToolsContainer();
	if (!pxToolsContainer)
	{
		const Asura_Hash_ID uContainerHash = GetContainerHash();
		if (uContainerHash != ASURA_HASH_ID_UNSET && Tools_Container_System::IsInitialised())
		{
			Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Container (0x%08x) not found.", uContainerHash);
		}
		return;
	}
	
	// Check for circular references before making the instance, otherwise it would crash.
	if (pxToolsContainer->CheckIsTooDeep())
	{
		return;
	}

	// Create a new instance.
	const Asura_Hash_ID uContainerHash = GetContainerHash();
	Asura_Container_InputState& xInputState = m_pxServerEntity->m_xInputState;
	
	const Tools_ContainerStateMachine* const pxMachine = pxToolsContainer->GetStateMachine();
	const u_int uNumInputs = (pxMachine ? pxMachine->GetNumInputs() : 1);
	if (xInputState.GetNumInputs() != uNumInputs)
	{
		// Change the number of inputs in the server entity.
		xInputState.SetNumInputs(uNumInputs);
	}

	if (uContainerHash != ASURA_HASH_ID_UNSET && !xInputState.IsOff())
	{
		if (Asura_Container_System::GetTree().GetNumberOfNodes() > 0)
		{
			m_pxInstance = Asura_Container_System::CreateClientInstance(uContainerHash);
			Asura_Assertf(m_pxInstance, ("Failed to create container instance (0x%08x)", uContainerHash));

			if (m_pxInstance)
			{
				// We are creating the whole thing from scratch, don't play transition-only effects.
				Asura_Container_System::SetDoingCreate(true);
				m_pxInstance->ChangeState(xInputState);
				Asura_Container_System::SetDoingCreate(false);

				Asura_ContainerInstance_Properties xData(0, GetGuid());
				m_pxInstance->Initialise(0, xData);
				m_pxInstance->TeleportTo(Position, Orientation);

				if (m_pxServerEntity)
				{
					// Currently the server entity doesn't create the server instance if tools is active;
					// so just set the flag to be true.
					// TODO if we change the code to create a server instance then this line can be removed.
					m_pxServerEntity->m_bShouldBlockWaypointLinks = true;
				}
			}
		}
	}
}

////////////////////////
// RefreshMinorChange //
////////////////////////
void Tools_Entity_Container::RefreshMinorChange()
{
	if (m_pxInstance)
	{
		m_pxInstance->TeleportTo(Position, Orientation);
	}
}

/////////////////////////////////////
// RefreshMinorChangeAllInstances  //
/////////////////////////////////////
void Tools_Entity_Container::RefreshMinorChangeAllInstances()
{
	Asura_Hashed_List_It<Tools_Entity_Container> xIt(GetList(), false);
	for ( ; !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxEntity = xIt.GetCurrent();
		if (pxEntity)
		{
			pxEntity->RefreshMinorChange();
		}
	}
}

/////////////////////////
// DestroyAllInstances //
/////////////////////////
// Delete the instances in all the tools entities.
void Tools_Entity_Container::DestroyAllInstances()
{
	Asura_Hashed_List_It<Tools_Entity_Container> xIt(GetList(), false);
	for ( ; !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxEntity = xIt.GetCurrent();
		if (pxEntity)
		{
			pxEntity->DestroyInstance();
		}
	}
}

////////////////////////
// CreateAllInstances //
////////////////////////
// Create the instances in all the tools entities.
void Tools_Entity_Container::CreateAllInstances()
{
	Asura_Hashed_List_It<Tools_Entity_Container> xIt(GetList(), false);
	for ( ; !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxEntity = xIt.GetCurrent();
		if (pxEntity)
		{
			pxEntity->CreateInstance();
		}
	}
}


////////////
// Create //
////////////
void Tools_Entity_Container::Create()
{
	// Get a new GUID
	Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();
	
	// Set up tools data and auto-create a unique name (prompt for new name if there is a clash).
	// Note: this name is now usually replaced by the call to SetNameFromContainer().
	Tools_Entity_Data* pxToolsData = CreateToolsDataWithDefaultName("Container", "Container", uGuid, GetNamePrefix());
	if (!pxToolsData) return;

	// Create the server object and set its data.
	Asura_ServerEntity_Container* pxServerObject = Asura_ContainerRegistry::CreateServerEntity(uGuid);
	Asura_Assert(pxServerObject, "Null server object");
	if (pxServerObject)
	{
		pxServerObject->ToolsData = pxToolsData;
	}

	// Create the tools object and select it.
	Tools_Entity_Container* pxToolsObject = Tools_ContainerRegistry::CreateToolsEntity(pxServerObject);
	Tools_Object_System::SelectObject(pxToolsObject);
	
	// By default we will automatically name containers.
	if (pxToolsObject)
	{
		pxToolsObject->SetNameFromContainer();
	}
}

void Tools_Entity_Container::Create(Asura_Entity* pxEntity)
{
	Tools_ContainerRegistry::CreateToolsEntity(pxEntity);
}

/////////////
// Destroy //
/////////////
void Tools_Entity_Container::Destroy()
{
	PARENT::Destroy();
}

////////////
// Render //
////////////
void Tools_Entity_Container::Render()
{
	// Only render the origin marker in entity mode.
	if (Tools_System::IsInObjectMode())
	{
		PARENT::Render();
	
		// Render any blocked waypoint links.
		if ( IsSelected() && m_pxServerEntity)
		{
			for ( unsigned char uc = 0; uc < m_pxServerEntity->m_ucNumLinksToBlock; ++uc )
			{
				const Asura_AI_NavConnection_Index& xNavConnection = m_pxServerEntity->m_pxLinksToBlock[ uc ];

				Asura_AI_NavVertex* pxWaypoint1 = Asura_AI_Navigation_System::GetWaypoint( xNavConnection.m_usWaypointIndex );
				Asura_Assert( pxWaypoint1, "Failed to find waypoint" );
				if ( !pxWaypoint1 ) continue;

				if ( xNavConnection.m_usConnectionIndex >= pxWaypoint1->m_ucNumConnections )
				{
					continue;
				}

				Asura_AI_NavVertex* pxWaypoint2 = Asura_AI_Navigation_System::GetWaypoint( pxWaypoint1->m_pxConnections[ xNavConnection.m_usConnectionIndex ].m_usIndex );
				Asura_Assert( pxWaypoint2, "Failed to find waypoint" );
				if ( !pxWaypoint2 ) continue;

				Asura_Vector_3 xPos1 = pxWaypoint1->m_xPosition;
				xPos1.y -= 0.5f;
				Asura_Vector_3 xPos2 = pxWaypoint2->m_xPosition;
				xPos2.y -= 0.5f;
				Asura_Render::RenderLine( xPos1, xPos2, 0xffffffff );
			}
		}	
	}
}

///////////////////
// RenderPreFSFX //
///////////////////
void Tools_Entity_Container::RenderPreFSFX()
{
	if (!m_pxInstance) return;

	if (IsBeingRenderedAsIcon())
	{
		return;
	}

	// Set outline colour.
	s_uOutlineColour = 0;

	Tools_Object* pxRootObject = GetObjectToSelect();
	if (pxRootObject && pxRootObject->ShouldRenderSelectionOutline())
	{
		if (pxRootObject->RenderAsSelected())
		{
			s_uOutlineColour = 0xffff0000;
		}
		else if (pxRootObject->IsGroupSelected())
		{
			s_uOutlineColour = 0xff0000ff;
		}
	}

	// Render the in-game meshes, shadows, effects etc.
	m_pxInstance->Render(Position, Orientation);
	m_pxInstance->RenderShadow(0);

	// Clear outline colour
	s_uOutlineColour = 0;
}

////////////
// Update //
////////////
void Tools_Entity_Container::Update()
{
	PARENT::Update();

	if (m_pxServerEntity && m_bChangeInputState)
	{
		m_bChangeInputState = false;
		Changed = true;

		// Selection of state hash changed. Recreate the instance.
		CreateInstance();

		// If this container is the primary selection then copy the state change to any others
		// in the selected group.
		if (Tools_Object_System::SelectedObject == this)
		{
			CopyStateChangeToOthersSelected();
		}
	}

	if (m_pxInstance)
	{
		m_pxInstance->Update();
	}
}

////////////////////////
// AddOverlayControls //
////////////////////////
void Tools_Entity_Container::AddOverlayControls(Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset)
{
	PARENT::AddOverlayControls(pxOverlay, xOffset);

	// Refresh the dynamic string labels.
	RefreshDialogue();

	// Add any project overlay controls here.
	AddProjectOverlayControls(pxOverlay, xOffset);

	Asura_ServerEntity_Container* const pxServer = GetServerEntity();

	Asura_Dialog* pxDialog = 0;

	xOffset.y += Asura_Dialog::SmallSpacing();
	pxDialog = new Asura_Function_Dialog("Rename using Container Name", xOffset, SetNameFromContainer_Callback);
	pxDialog->SetHelpText("Automatically generate the name from the container name. Multiple objects can be selected to rename several entities at once.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Edit in Container Mode", xOffset, EditContainer_Callback, 0, IsContainerValid);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Dynamic_Dialog(&m_xContainerName, xOffset);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Select Container", xOffset, SelectContainer_Callback);
	pxDialog->SetHelpText("Select the container from a list. Multiple objects can be selected to set several at once.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	if (pxServer)
	{
		Asura_Container_InputState& xInputState = pxServer->m_xInputState;
		const char* const szTitle = (xInputState.GetNumInputs() > 1) ? "(0) Initial State" : "Initial State";
		Tools_Container_MultipleState_Dialog* pxMultiDialog = new Tools_Container_MultipleState_Dialog( szTitle,
																										*pxOverlay,
																										xOffset,
																										&pxServer->m_uContainerHash,
																										xInputState.GetNumInputs(),
																										xInputState.GetStateHashArray(),
																										&m_bChangeInputState,
																										false );

		pxMultiDialog->SetHelpTextForChildren("Select the initial state. Multiple objects can be selected to set several at once.");
		pxOverlay->AddDialog(pxMultiDialog);
		xOffset.y += Asura_Dialog::DefaultSpacing();
	}
}

///////////////////
// ResizeOverlay //
///////////////////
void Tools_Entity_Container::ResizeOverlay()
{
	Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay( GetOverlayID() );
	if (pxOverlay)
	{
		pxOverlay->ResizeOverlay();
	}
}

/////////////////////////////////////
// CopyStateChangeToOthersSelected //
/////////////////////////////////////
// The user has changed the intial state setting for an entity, check whether due to multiple selection
// there are any others that need changing.
void Tools_Entity_Container::CopyStateChangeToOthersSelected()
{
	// Multiple selection allowed. Change state for any that are of the correct type.
	if (Tools_Object_System::SelectedObjectList.size() <= 0) return;
	
	const Asura_Container_InputState& xInputState = GetInputState();

	for (LIF<Tools_Object*> xObjectLIF(&Tools_Object_System::SelectedObjectList); !xObjectLIF.done(); xObjectLIF.next())
	{
		Tools_Object* pxObject = xObjectLIF();
		if (!pxObject) continue;
		if (pxObject->Type != TOT_Entity_Container) continue;
		
		Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);
		if (pxToolsContainer->m_pxServerEntity == 0)
		{
			continue;
		}

		if (pxToolsContainer->GetInputState() == xInputState)
		{
			// Already have the same state.
			continue;
		}

		// Check that the container entity actually has the desired state.
		if (pxToolsContainer->GetContainerHash() != GetContainerHash())
		{
			// Different type of container, but may have the same state.
			const bool bValidState = Tools_Container_System::IsContainerStateValid(pxToolsContainer->GetContainerHash(), xInputState);
			if (!bValidState)
			{
				// Not a valid state. So skip this entity.
				continue;
			}
		}

		// Copy directly to their server entity.
		pxToolsContainer->m_pxServerEntity->m_xInputState = xInputState;
		pxToolsContainer->m_bChangeInputState = true;		
	}
}

//////////////////////////////
// SelectContainer_Callback //
//////////////////////////////
void Tools_Entity_Container::SelectContainer_Callback()
{
	Tools_Object* pxObject = Tools_Object_System::SelectedObject;
	if (!pxObject) return;
	if (pxObject->Type != TOT_Entity_Container) return;
	
	Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);
	Asura_ServerEntity_Container* pxServer = pxToolsContainer->GetServerEntity();
	if (!pxServer) return;

	Asura_Hash_ID uNewContainerHash = pxServer->GetContainerHash();

	// Include only global-scope collection containers.
	// We could allow others (i.e. non-collections), but they aren't all that useful.
	const bool bOk = Tools_Container_System::SelectContainerModalOverlay(uNewContainerHash, ASURA_HASH_ID_UNSET, true, true, ASURA_HASH_ID_UNSET);
	if (!bOk) return; // Cancelled

	// If the old state is still valid then preserve it. Otherwise clear the state.
	const Asura_Container_InputState& xOldInputState = pxServer->GetInputState();
	const bool bValidOldState = Tools_Container_System::IsContainerStateValid(uNewContainerHash, xOldInputState);

	Asura_Container_InputState xNewInputState;
	if (bValidOldState)
	{
		xNewInputState = xOldInputState;
	}
	else
	{
		xNewInputState.Reset();
		const Tools_Container* pxToolsContainer = Tools_Container_System::FindContainerByHash(uNewContainerHash);
		if (pxToolsContainer)
		{
			xNewInputState.SetNumInputs(pxToolsContainer->GetNumInputs());
		}
	}

	// Multiple selection allowed. Select for any that are of the correct type.
	if (Tools_Object_System::SelectedObjectList.size() > 0)
	{
		for(LIF<Tools_Object*> xObjectLIF(&Tools_Object_System::SelectedObjectList);!xObjectLIF.done();xObjectLIF.next())
		{
			Tools_Object* pxObject = xObjectLIF();
			if (pxObject && pxObject->Type == TOT_Entity_Container)
			{
				Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);
				pxToolsContainer->SelectContainer(uNewContainerHash, xNewInputState);
			}
		}
	}
}

/////////////////////
// SelectContainer //
/////////////////////
void Tools_Entity_Container::SelectContainer(const Asura_Hash_ID uNewContainerHash, const Asura_Container_InputState& xNewInputState)
{	
	Asura_ServerEntity_Container* pxServer = GetServerEntity();
	if (!pxServer) return;

	if (pxServer->GetContainerHash() == uNewContainerHash &&
		pxServer->GetInputState()    == xNewInputState)
	{
		return; // No change
	}

	const bool bChangedNumInputs = (pxServer->m_xInputState.GetNumInputs() != xNewInputState.GetNumInputs());

	pxServer->SetContainerHash(uNewContainerHash);
	pxServer->m_xInputState = xNewInputState;
	
	RefreshDialogue();
	RefreshStateDialogue();
	ResizeOverlay();
	LoadRequiredFiles();
	CreateInstance();

	Changed = true;

	// Auto rename, if necessary, so that the Tools name matches the container name.
	if (ShouldAutoRename())
	{
		SetNameFromContainer();
	}

	// Rebuild the overlay if the number of inputs has changed.
	if (bChangedNumInputs)
	{
		const Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayID());
		if (pxOverlay)
		{
			CreateObjectOverlay();
		}
	}
}

////////////////////////////
// EditContainer_Callback //
////////////////////////////
void Tools_Entity_Container::EditContainer_Callback()
{
	Tools_Object* pxObject = Tools_Object_System::SelectedObject;
	if (!pxObject) return;
	if (pxObject->Type != TOT_Entity_Container) return;
	
	Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);
	const Asura_Hash_ID uContainerHash = pxToolsContainer->GetContainerHash();
	const Asura_Guid uGuid = pxToolsContainer->GetGuid();

	// New behaviour is to switch to container preview mode, with this container selected.
	Tools_Mode_Container::RequestModeForSelectedContainerEntity(uContainerHash, uGuid);
}


//////////////////////
// IsContainerValid //
//////////////////////
bool Tools_Entity_Container::IsContainerValid()
{
	Tools_Object* pxObject = Tools_Object_System::SelectedObject;
	if (!pxObject) return false;
	if (pxObject->Type != TOT_Entity_Container) return false;
	
	Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);
	const Asura_Hash_ID uContainerHash = pxToolsContainer->GetContainerHash();
	
	return (Tools_Container_System::FindContainerByHash(uContainerHash) != 0);
}

/////////////////
// OnSelection //
/////////////////
void Tools_Entity_Container::OnSelection()
{
	PARENT::OnSelection();

	// If only the edit overlay is open then we need to close and reopen it ourselves.
	const bool bOpenEdit = ( Tools_Container_System::IsEditOverlayOpen() &&
							!Tools_Container_System::IsOverlayOpen() );
	Tools_Container_System::SelectContainer( GetContainerHash(), bOpenEdit );
}

///////////////////////
// GetToolsContainer //
///////////////////////
Tools_Container* Tools_Entity_Container::GetToolsContainer() const
{
	if (!m_pxServerEntity) return 0;
	
	const Asura_Hash_ID uContainerHash = m_pxServerEntity->GetContainerHash();
	return Tools_Container_System::FindContainerByHash(uContainerHash);
}

/////////////////////
// RefreshDialogue //
/////////////////////
void Tools_Entity_Container::RefreshDialogue()
{
	m_xContainerName = "Container : ";

	if (m_pxServerEntity)
	{
		const Asura_Hash_ID uContainerHash = m_pxServerEntity->GetContainerHash();
		const Tools_Container* const pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
		if (pxContainer)
		{
			Asura_String xFullName;
			pxContainer->MakeFullName(xFullName);

			m_xContainerName += xFullName;
		}
		else if (uContainerHash == ASURA_HASH_ID_UNSET)
		{
			m_xContainerName += "*None*";
		}
		else
		{
			m_xContainerName += "*Not Found*";
		}
	}
}

//////////////////////////
// RefreshStateDialogue //
//////////////////////////
// List of container states may have changed, refresh the relevant dialogues.
void Tools_Entity_Container::RefreshStateDialogue()
{
	Tools_ContainerState_DropDownList_Dialog::RefreshAllDialogues();
}

/////////////
// Refresh //
/////////////
// Main container overlay has changed we need to refresh the dialogue if an object of this type is selected.
void Tools_Entity_Container::Refresh()
{
	Tools_Object* pxObject = Tools_Object_System::SelectedObject;
	if (!pxObject) return;
	if (pxObject->Type != TOT_Entity_Container) return;
	
	Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);

	pxToolsContainer->RefreshDialogue();
	pxToolsContainer->RefreshStateDialogue();
	pxToolsContainer->ResizeOverlay();
}

//////////////////////////////
// RefreshContainerModified //
//////////////////////////////
// The specified container has changed or been deleted, recreate any affected instance trees.
void Tools_Entity_Container::RefreshContainerModified(const Asura_Hash_ID uContainerChanged, bool bMinorChange)
{
	const bool bDeleting = (Tools_Mode_Container::GetDeletionCounter() > 0);

	Asura_Hashed_List_It<Tools_Entity_Container> xIt(GetList(), false);
	for ( ; !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxEntity = xIt.GetCurrent();
		if (pxEntity)
		{
			if (Tools_Container_System::ContainerSearch(pxEntity->GetContainerHash(), uContainerChanged))
			{
				if (bDeleting)
				{
					pxEntity->DestroyInstance();
				}
				else if (bMinorChange)
				{
					pxEntity->RefreshMinorChange();
				}
				else
				{
					pxEntity->CreateInstance();
				}
			}
			else if (!bDeleting && !pxEntity->m_pxInstance)
			{
				pxEntity->CreateInstance();
			}
		}
	}
}

///////////////////////////
// ChangedNumberOfInputs //
///////////////////////////
// The number of inputs has changed for a collection we may need to update the object overlay.
void Tools_Entity_Container::ChangedNumberOfInputs(const Asura_Hash_ID uContainerChanged, u_int uNumInputs)
{
	Asura_Hashed_List_It<Tools_Entity_Container> xIt(GetList(), false);
	for ( ; !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxEntity = xIt.GetCurrent();
		if (pxEntity && pxEntity->GetContainerHash() == uContainerChanged)
		{
			// Rebuild the overlay if it is open.
			Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(pxEntity->GetOverlayID());
			if (pxOverlay)
			{
				// Recreate the overlay.
				if (pxEntity->m_pxServerEntity)
				{
					pxEntity->m_pxServerEntity->m_xInputState.SetNumInputs(uNumInputs);
				}
				pxEntity->CreateObjectOverlay();
			}
		}
	}
}

/////////////////////////
// UpdateWhileSelected //
/////////////////////////
void Tools_Entity_Container::UpdateWhileSelected()
{
	PARENT::UpdateWhileSelected();
}

///////////////////////////////
// UpdatePositionInAsuraData //
///////////////////////////////
void Tools_Entity_Container::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{
	if (m_pxServerEntity)
	{
		//	Update the position if it has changed
		if	(m_pxServerEntity->GetPosition() != Position)
		{
			m_pxServerEntity->SetPosition(Position);
			Changed	= true;
		}

		//	Update the orientation if it has changed
		const Asura_Quat xToolsOriQuat(Orientation);

		if (m_pxServerEntity->GetOrientationQuat() != xToolsOriQuat)
		{
			m_pxServerEntity->SetOrientationQuat(xToolsOriQuat);
			Changed	= true;
		}
	}

	if (m_pxInstance)
	{
		m_pxInstance->TeleportTo(Position, Orientation);
	}

	PARENT::UpdatePositionInAsuraData(bUpdateLinkedObjects);
}

///////////////
// PrintInfo //
///////////////
void Tools_Entity_Container::PrintInfo(float& fX,float& fY)
{
	PARENT::PrintInfo(fX, fY);

	Asura_Text_Output::Print(fX, fY++, 0xffffffff, 0, "%s", m_xContainerName.GetString());
}

///////////////////
// ObjectIsInLOS //
///////////////////
bool Tools_Entity_Container::ObjectIsInLOS(const Asura_Vector_3& xAlpha, const Asura_Vector_3& xBeta, float& fLowest)
{
	if (m_pxInstance)
	{
		Asura_LOS_InputData xInput(xAlpha, xBeta, ASURA_GUID_UNREGISTERED);
		Asura_LOS_OutputData xOutput;
		xOutput.Reset(fLowest);
		if (m_pxInstance->LOSTest(&xInput, &xOutput))
		{
			fLowest = xOutput.m_fDistanceToIntersection;
			return true;
		}
	}

	return PARENT::ObjectIsInLOS(xAlpha, xBeta, fLowest);
}

//////////////////////////////////////
// GetPolygonAndPositionUnderCursor //
//////////////////////////////////////
// This function is currently only used by the snap to vertex code.
bool Tools_Entity_Container::GetPolygonAndPositionUnderCursor(int& iPolygonIndex, Asura_Vector_3& xPosition)
{
	if (m_pxInstance)
	{
		const Asura_Vector_3& xCameraPosition = *Asura_Camera::GetPosition();
		const Asura_Matrix_3x3& xCameraOrientation = *Asura_Camera::GetOrientation();
		
		// Get the z direction for the mouse
		Asura_Vector_3 xBeta;
		Asura_Cursor::GetLOSDirection(xBeta);

		// Transform this direction into worldspace
		xBeta = xBeta * xCameraOrientation;

		float fDistance = ASURA_LOS_MAX_DISTANCE;

		// Do a LOS test to find out what is under the cursor.
		if ( ObjectIsInLOS(xCameraPosition, xBeta, fDistance) )
		{
			// Don't bother with getting a polygon, just get a position.
			iPolygonIndex = -1;
			xPosition = xCameraPosition + (xBeta * fDistance);
			return true;
		}
	}

	// Give up and just us the Tools mesh instead.
	return PARENT::GetPolygonAndPositionUnderCursor(iPolygonIndex, xPosition);
}

////////////////////////////////////////////
// GetPolygonNormalAndPositionUnderCursor //
////////////////////////////////////////////
bool Tools_Entity_Container::GetPolygonNormalAndPositionUnderCursor(Asura_Vector_3& xNormal, Asura_Vector_3& xPosition)
{
	if (m_pxInstance)
	{
		const Asura_Vector_3& xCameraPosition = *Asura_Camera::GetPosition();
		const Asura_Matrix_3x3& xCameraOrientation = *Asura_Camera::GetOrientation();
		
		// Get the z direction for the mouse
		Asura_Vector_3 xBeta;
		Asura_Cursor::GetLOSDirection(xBeta);

		// Transform this direction into worldspace
		xBeta = xBeta * xCameraOrientation;

		// Do a LOS test to find out what is under the cursor.
		Asura_LOS_InputData xInput(xCameraPosition, xBeta, ASURA_GUID_UNREGISTERED);
		Asura_LOS_OutputData xOutput;
		xOutput.Reset(ASURA_LOS_MAX_DISTANCE);

		if (m_pxInstance->LOSTest(&xInput, &xOutput))
		{
			xPosition = xCameraPosition + (xBeta * xOutput.m_fDistanceToIntersection);
			xNormal = xOutput.m_xNormal;
			return true;
		}
	}

	// Give up and just us the Tools mesh instead.
	return PARENT::GetPolygonNormalAndPositionUnderCursor(xNormal, xPosition);
}

//////////////////////////
// AddFilesToExportList //
//////////////////////////
void Tools_Entity_Container::AddFilesToExportList(List<Asura_String>& xNameList) const
{
	PARENT::AddFilesToExportList(xNameList);

	const Tools_Container* const pxContainer = GetToolsContainer();
	if (pxContainer && !pxContainer->CheckIsTooDeep())
	{
		pxContainer->AddFilesToExportList(xNameList);
	}
}

///////////////////////
// LoadRequiredFiles //
///////////////////////
// Load any files needed for tools e.g. skins, anims etc.
void Tools_Entity_Container::LoadRequiredFiles()
{
	const Tools_Container* const pxContainer = GetToolsContainer();
	if (pxContainer && !pxContainer->CheckIsTooDeep())
	{
		pxContainer->LoadRequiredFiles();
	}
}

////////////////////////
// HandlesThisMessage //
////////////////////////
bool Tools_Entity_Container::HandlesThisMessage(int iMessageID)
{
	if (iMessageID == ASURA_MSG_CONTAINER_CHANGE_STATE || iMessageID == ASURA_MSG_CONTAINER_CHANGE_STATE_EXTENDED)
	{
		return true;
	}

	return PARENT::HandlesThisMessage(iMessageID);
}

//////////////////////
// HandlesThisEvent //
//////////////////////
bool Tools_Entity_Container::HandlesThisEvent(int iEventID) const
{
	if ( iEventID == ASURA_ENTITY_EVENT_ONDESTRUCTION ) 
	{
		return true;
	}
	else if( (ASURA_ENTITY_EVENT_ONHEALTHPERCENT_0 <= iEventID) && (iEventID <= ASURA_ENTITY_EVENT_ONHEALTHPERCENT_99) ) 
	{
		// Note: this event only works if the entity has an enabled health container.
		// If it has more than one health container then they will all generate events.
		return true;
	}

	return PARENT::HandlesThisEvent( iEventID );
}

//////////////////
// GetStateName //
//////////////////
// Get the name of the container state for the given entity.
const char* Tools_Entity_Container::GetStateName(Asura_Guid uGuid, Asura_Hash_ID uStateHash)
{
	if (uStateHash == Asura_ContainerState::GetOnStateHash())
	{
		return Tools_ContainerState::GetOnStateName();
	}
	else if (uStateHash == Asura_ContainerState::GetOffStateHash())
	{
		return Tools_ContainerState::GetOffStateName();
	}
	else if (uStateHash == ASURA_HASH_ID_UNSET)
	{
		return "*None*";
	}

	const Tools_Entity_Container* const pxEntity = FindEntity(uGuid);
	if (pxEntity)
	{
		const Asura_Hash_ID uContainerHash = pxEntity->GetContainerHash();
		const Tools_Container* const pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
		if (pxContainer)
		{
			const Tools_ContainerStateMachine* const pxMachine = pxContainer->GetStateMachine();
			if (pxMachine)
			{
				const Tools_ContainerState* const pxState = pxMachine->FindState(uStateHash);
				if (pxState)
				{
					const Asura_String& xName = pxState->GetStateName();
					return xName.GetString();
				}
			}
		}
	}
	else if (!Tools_System::IsActive())
	{
		// In-game there are no tools objects. State probably exists, but we don't know its name.
		static char szBuffer[16];
		Asura_PrintfToBfr(szBuffer, "0x%08x", 16, true, uStateHash);
		return szBuffer;
	}

	return "*Unknown*";
}

///////////////////////////////////
// SetNameFromContainer_Callback //
///////////////////////////////////
void Tools_Entity_Container::SetNameFromContainer_Callback()
{
	// Multiple selection allowed. Rename any that are of the correct type.
	if (Tools_Object_System::SelectedObjectList.size() > 0)
	{
		for(LIF<Tools_Object*> xObjectLIF(&Tools_Object_System::SelectedObjectList);!xObjectLIF.done();xObjectLIF.next())
		{
			Tools_Object* pxObject = xObjectLIF();
			if (pxObject && pxObject->Type == TOT_Entity_Container)
			{
				Tools_Entity_Container* pxToolsContainer = static_cast<Tools_Entity_Container*>(pxObject);
				pxToolsContainer->SetNameFromContainer();
			}
		}
	}
}

//////////////////////////
// SetNameFromContainer //
//////////////////////////
// Automatically set the Tools name using the container name.
void Tools_Entity_Container::SetNameFromContainer()
{
	Asura_ServerEntity_Container* pxServerEntity = GetServerEntity();
	if (!pxServerEntity) return;

	const Asura_Hash_ID uContainerHash = pxServerEntity->GetContainerHash();
	const char* const szToolsName = GetToolsName();

	const Tools_Container* const pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
	if (!pxContainer)
	{
		// Not found or not set. Just use the guid.
		// See note before ShouldAutoRename(); this name fits the required pattern.
		Asura_String256 xTestName("Container %u", GetGuid());

		if (szToolsName && xTestName.Compare(szToolsName) == 0)
		{
			// Nothing to do. We can reuse the same name.
			return;
		}

		SetToolsName(xTestName.GetString());
		return;
	}

	const Asura_String& xBaseName = pxContainer->GetContainerName();

	for (u_int uNumber = 1; uNumber < 1000; ++uNumber)
	{
		// See note before ShouldAutoRename(); this name fits the required pattern.
		const Asura_String256 xTestName("Container: %s %03u", xBaseName.GetString(), uNumber);
		
		if ( szToolsName && xTestName.Compare(szToolsName) == 0 )
		{
			// Nothing to do. We can reuse the same name.
			return;
		}

		bool bNameInUse = false;

		for (LIF<Tools_Object*> olif(&Tools_Object_System::ObjectList); !olif.done(); olif.next())
		{
			if (xTestName.Compare(olif()->GetName()) == 0)
			{
				// Name already in use
				bNameInUse = true;
				break;
			}
		}

		if (!bNameInUse)
		{
			// Found an unused name.
			SetToolsName(xTestName.GetString());
			return;
		}
	}
}

//////////////////
// SetToolsName //
//////////////////
void Tools_Entity_Container::SetToolsName(const char* szName)
{
	if (Entity && Entity->ToolsData)
	{
		Entity->ToolsData->Name = szName;
		SetEntityModified();
	}
}

//////////////////
// GetToolsName //
//////////////////
const char* Tools_Entity_Container::GetToolsName() const
{
	if (Entity && Entity->ToolsData)
	{
		return Entity->ToolsData->Name.GetString();
	}
	return 0;
}

//////////////////////
// ShouldAutoRename //
//////////////////////
// Assume that if the name starts with "Container" and ends with a number, then
// it was automatically named and the name should be updated if the container is changed.
// Not a perfect test, but better than nothing. This allows containers to be manually renamed.
// Both: "Container 100123" and "Container: SomeContainerName 001" fit the required pattern.
bool Tools_Entity_Container::ShouldAutoRename() const
{
	if (Entity && Entity->ToolsData)
	{
		const Asura_String& xToolsName = Entity->ToolsData->Name;

		if (xToolsName.StartsWith("Container"))
		{
			// Reverse the string to bring the number at the end to the start;
			// it's reversed, but we only care whether the number is there, not its value.
			Asura_String xCopy = xToolsName;
			xCopy.Reverse();

			int iDummy = 0;
			const bool bIsNumber = Asura_StringUtil::GetIntValue( xCopy.GetString(), &iDummy, false );
			return bIsNumber;
		}
	}

	return false;
}

////////////
// Verify //
////////////
// Returns whether the object's associated game data is valid.
// bAutomatic signifies whether the verification has been automatically or manually requested.
bool Tools_Entity_Container::Verify( const bool bAutomatic )
{
	return Tools_Container_System::VerifyEntity(m_pxServerEntity);
}

/////////////////
// FixUpErrors //
/////////////////
// Fix up errors caused by renaming states automatically.
void Tools_Entity_Container::FixUpErrors(bool bLogFixes /*= true*/)
{
	if (!m_pxServerEntity) return;

	const Tools_Container* const pxContainer = GetToolsContainer();
	if (!pxContainer) return;

	const Tools_ContainerStateMachine* const pxStateMachine = pxContainer->GetStateMachine();
	if (!pxStateMachine) return;

	Asura_Container_InputState& xInputState = m_pxServerEntity->m_xInputState;

	for (u_int u=0; u<xInputState.GetNumInputs(); ++u)
	{
		const Asura_Hash_ID uStateHash = xInputState.GetStateHash(u);
		if (pxStateMachine->FindState(uStateHash))
		{
			// State is valid.
			continue;
		}

		// State not found. Check whether it corresponds to a renamed state.
		const Asura_Hash_ID uNewStateHash = pxStateMachine->GetHashForRenamedState(uStateHash);
		if (uNewStateHash != uStateHash && uNewStateHash != ASURA_HASH_ID_UNSET)
		{
			// Need to fix up the entity.
			xInputState.SetStateHash(u, uNewStateHash);

			// Mark as modified.
			SetEntityModified();

			if (bLogFixes)
			{
				// Log the change.
				const Tools_ContainerState* pxState = pxStateMachine->FindState(uNewStateHash);
				const char* szStateName = pxState ? pxState->GetStateName().GetString() : "*Unknown*";

				Tools_MessageLog::AddMessage( Tools_Container_System::GetCategory(), Tools_Container_System::GetSubCategory(), new Tools_MessageLogAction_GotoEntity(GetGuid()), "Entity %u : container state renamed (0x%08x) -> '%s'. Fixed.", GetGuid(), uStateHash, szStateName );
			}
		}
	}
}



///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_BlockNavData
///////////////////////////////////////////////////////////////////////////////
// Parent classes for functors to remove/reduce waypoints that are blocked by container mesh instances.
class Asura_Container_Functor_BlockNavData: public Asura_Container_Functor
{
public:
	Asura_Container_Functor_BlockNavData() : m_bBlocksWaypoints(false) {}

	bool GetBlocksWaypoints() const {return m_bBlocksWaypoints; }

	virtual void Process(const Asura_ClientContainerInstance* pxEntry)
	{
		if (pxEntry->GetClassification() == AsuraContainerClass_Mesh)
		{
			ProcessMesh(static_cast<const Asura_ClientContainerInstance_Mesh*>(pxEntry));
		}
	}

private:
	void ProcessMesh(const Asura_ClientContainerInstance_Mesh* const pxMesh)
	{
		const Asura_Container_Mesh* const pxContainerMesh = pxMesh->GetContainer();
		if (!pxContainerMesh) return;

		if (!pxContainerMesh->ShouldBlockNavDataGeneration())
		{
			return; // Not flagged to block waypoint generation.
		}
		if (pxContainerMesh->HasNoCollisions())
		{
			return; // No collisions.
		}

		const Asura_Animation* const pxAnimation = pxMesh->GetAnimation();
		if (!pxAnimation) return;

		const Asura_Physics_Object* const pxPhysicsObject = pxMesh->GetPhysicsObject();
		if (!pxPhysicsObject) return;

		const Asura_Collision_Object* const pxCollObj = pxPhysicsObject->GetCollisionObject();
		if (pxCollObj)
		{
			m_bBlocksWaypoints = true;
			Callback(pxCollObj, pxPhysicsObject->GetPosition(), pxPhysicsObject->GetOrientation());
		}
	}

protected:
	virtual void Callback(const Asura_Collision_Object* pxCollObject, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation) = 0;

private:
	bool m_bBlocksWaypoints;
};

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_RemoveBlockedWaypoints
///////////////////////////////////////////////////////////////////////////////
// Functor to remove waypoints that are blocked by container mesh instances.
class Asura_Container_Functor_RemoveBlockedWaypoints : public Asura_Container_Functor_BlockNavData
{
protected:
	virtual void Callback(const Asura_Collision_Object* pxCollObject, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
	{
		Tools_Mode_Waypointing::RemoveWaypointsBlockedByCollisionObject(pxCollObject, xPosition, xOrientation);
	}
};

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_ReduceWaypointRadius
///////////////////////////////////////////////////////////////////////////////
// Functor to reduce waypoint's radius due to blocking by container mesh instances.
class Asura_Container_Functor_ReduceWaypointRadius : public Asura_Container_Functor_BlockNavData
{
public:
	Asura_Container_Functor_ReduceWaypointRadius(Tools_Nav_Vertex* pxNavVertex)
		: m_pxNavVertex(pxNavVertex)
	{}

protected:
	virtual void Callback(const Asura_Collision_Object* pxCollObject, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
	{
		// Transform waypoint position to object space.
		Asura_Vector_3 xAlpha = m_pxNavVertex->m_xPosition - xPosition;
		xAlpha = xOrientation.ApplyTranspose( xAlpha );

		const float fDistanceSq = pxCollObject->GetBoundingBox().GetSqdMinDistFromPoint( xAlpha );
		if ( fDistanceSq < m_pxNavVertex->m_fRadius * m_pxNavVertex->m_fRadius )
		{
			m_pxNavVertex->m_fRadius = Asura_Maths::Sqrt( fDistanceSq );
		}
	}

private:
	Tools_Nav_Vertex* m_pxNavVertex;
};

////////////////////////////
// RemoveBlockedWaypoints //
////////////////////////////
// Remove any waypoints blocked by the container entity.
void Tools_Entity_Container::RemoveBlockedWaypoints() const
{
	if (m_pxInstance)
	{
		Asura_Container_Functor_RemoveBlockedWaypoints xFunctor;
		m_pxInstance->Traverse(xFunctor);

		// Cache information about whether this entity blocks waypoint generation.
		m_bBlocksWaypointGeneration = xFunctor.GetBlocksWaypoints();
	}
}

//////////////////////////
// ReduceWaypointRadius //
//////////////////////////
// Reduce the radius of any nearby waypoints as necessary.
void Tools_Entity_Container::ReduceWaypointRadius(Tools_Nav_Vertex* pxNavVertex) const
{
	if (m_pxInstance)
	{
		Asura_Container_Functor_ReduceWaypointRadius xFunctor(pxNavVertex);
		m_pxInstance->Traverse(xFunctor);
	}
}

///////////////////////////////////
// RecalculateWaypointLinkBlocks //
///////////////////////////////////
void Tools_Entity_Container::RecalculateWaypointLinkBlocks()
{
    if ( m_pxServerEntity )
    {
        // Remove the old link blocks.
        m_pxServerEntity->UpdateWaypointLinkBlocks( true );

        // Calculate the new link blocks.
        Asura_ServerEntity_Container::InitialiseWaypointLinkBlockCalculations();
        m_pxServerEntity->CalculateWaypointLinkBlocks();
        Asura_ServerEntity_Container::ShutdownWaypointLinkBlockCalculations();

        // Apply the new link blocks.
        m_pxServerEntity->UpdateWaypointLinkBlocks();
    }
}

////////////////////////
// OnGroupDeselection //
////////////////////////
void Tools_Entity_Container::OnGroupDeselection()
{
	PARENT::OnGroupDeselection();

	// No point in calculating link blocks on an object being destroyed.
	if ( !IsBeingDestroyed() )
	{
		// Recalculate blocked links.
		RecalculateWaypointLinkBlocks();
	}
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_CullWaypointLinks
///////////////////////////////////////////////////////////////////////////////
// Functor to set which container mesh instances are taken in to account when culling waypoint links.
class Asura_Container_Functor_CullWaypointLinks : public Asura_Container_Functor
{
public:
	Asura_Container_Functor_CullWaypointLinks(bool bBegin) : m_bBegin(bBegin) {}

	virtual void Process(const Asura_ClientContainerInstance* pxEntry)
	{
		if (pxEntry->GetClassification() == AsuraContainerClass_Mesh)
		{
			ProcessMesh(static_cast<const Asura_ClientContainerInstance_Mesh*>(pxEntry));
		}
	}

protected:
	void ProcessMesh(const Asura_ClientContainerInstance_Mesh* const pxMesh)
	{
		const Asura_Container_Mesh* const pxContainerMesh = pxMesh->GetContainer();
		if (!pxContainerMesh) return;
	
		if (pxContainerMesh->HasNoCollisions())
		{
			return; // No collisions.
		}

		const Asura_Animation* const pxAnimation = pxMesh->GetAnimation();
		if (!pxAnimation) return;

		const Asura_Physics_Object* const pxPhysicsObject = pxMesh->GetPhysicsObject();
		if (!pxPhysicsObject) return;

		const Asura_Collision_Object* const pxCollObj = pxPhysicsObject->GetCollisionObject();
		if (!pxCollObj) return;

		// Only static objects that are flagged as blocking nav data generation will actually cull links.
		// If we are beginning the culling then set everything else to be ignored. After culling restore the flag.
		const bool bCullLinks = (pxPhysicsObject->GetType() == ASURA_PHYSICS_OBJECT_STATIC) &&
								 pxContainerMesh->ShouldBlockNavDataGeneration();
		
		const bool bIgnore = (m_bBegin && !bCullLinks);

		if (bIgnore != pxPhysicsObject->IsIgnore())
		{
			const_cast<Asura_Physics_Object*>(pxPhysicsObject)->SetIgnore(bIgnore);
		}	
	}

	bool m_bBegin;
};

////////////////////////////
// BeginCullWaypointLinks //
////////////////////////////
void Tools_Entity_Container::BeginCullWaypointLinks()
{
	for (Asura_Hashed_List_It<Tools_Entity_Container> xIt(Tools_Entity_Container::GetList()); !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxToolsEntity = xIt.GetCurrent();
		if (pxToolsEntity && pxToolsEntity->m_pxInstance)
		{
			Asura_Container_Functor_CullWaypointLinks xFunctor(true);
			pxToolsEntity->m_pxInstance->Traverse(xFunctor);
		}
	}
}

//////////////////////////
// EndCullWaypointLinks //
//////////////////////////
void Tools_Entity_Container::EndCullWaypointLinks()
{
	for (Asura_Hashed_List_It<Tools_Entity_Container> xIt(Tools_Entity_Container::GetList()); !xIt.Done(); xIt.Next())
	{
		Tools_Entity_Container* pxToolsEntity = xIt.GetCurrent();
		if (pxToolsEntity && pxToolsEntity->m_pxInstance)
		{
			Asura_Container_Functor_CullWaypointLinks xFunctor(false);
			pxToolsEntity->m_pxInstance->Traverse(xFunctor);
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_ProcessMeshVertices
///////////////////////////////////////////////////////////////////////////////
// Traverse the instance looking for meshes and call a callback function for each vertex.
class Asura_Container_Functor_ProcessMeshVertices : public Asura_Container_Functor
{
public:
	Asura_Container_Functor_ProcessMeshVertices() {}

	virtual void Process(const Asura_ClientContainerInstance* pxEntry)
	{
		if (pxEntry->GetClassification() == AsuraContainerClass_Mesh)
		{
			ProcessMesh(static_cast<const Asura_ClientContainerInstance_Mesh*>(pxEntry));
		}
	}

	// Called for each vertex on each mesh. Position is in world-space.
	virtual void ProcessVertex(const Asura_Vector_3& xPointPosition) = 0;

protected:
	void ProcessMesh(const Asura_ClientContainerInstance_Mesh* const pxMesh)
	{
		const Asura_Animation* pxAnimation = pxMesh->GetAnimation();
		if (!pxAnimation) return;

		const Asura_Hierarchy_Skin* pxSkin = pxAnimation->GetSkin();
		if (!pxSkin) return;

		const Asura_Hierarchy_Tween* pxTweenStorage = pxAnimation->GetTweenStore();
		if (!pxTweenStorage->IsPositionDataValid()) return;

		const Asura_Quat& xOriQuat = pxAnimation->GetOrientQuat();
		Asura_Matrix_3x3 xOrientation;
		xOriQuat.ToMatrix(xOrientation);

		const Asura_Vector_3& xPosition = pxAnimation->GetPosition();

		// Calculate the world positions of all the bones
    	for (int i = 0; i < pxSkin->NumberOfBones; i++)
		{
			Asura_Matrix_3x3 xBoneOri;
			Asura_Vector_3 xBonePos;

			pxTweenStorage->GetBoneData(i).GetOrientation(xBoneOri);
			pxTweenStorage->GetBoneData(i).GetPosition(xBonePos);

			if (i == 0)
			{
				// Root
				axBonePosition[0] = xBonePos * xOrientation + xPosition;
				axBoneOrientation[0] = xOrientation * xBoneOri;
			}
			else
			{
				// Get parent
				const int iParentIndex = pxSkin->ParentBone[i];
				Asura_Assert(iParentIndex >= 0 && iParentIndex < i, "Invalid bone index");

				axBonePosition[i] = xBonePos * axBoneOrientation[iParentIndex] + axBonePosition[iParentIndex];
				axBoneOrientation[i] = axBoneOrientation[iParentIndex] * xBoneOri;
			}
		}

		if (pxSkin->IsSkinnedModel())
		{
			// Calculate position of each vertex in skin
			for (int i=0; i<pxSkin->NumberOfVertices; i++)
			{
				Asura_Vector_3 xVertPos(0.0f, 0.0f, 0.0f);

				for (int b=0; b<4; b++)
				{
					const float fWeight = pxSkin->VertexData[i].Weight[b];
					if (fWeight == -1.0f) break;
					
					const int iBone = pxSkin->VertexData[i].BoneID[b];

					const Asura_Vector_3 xBoneVertex = (pxSkin->VertexData[i].Offset[b] * axBoneOrientation[iBone]) + axBonePosition[iBone];
		
					xVertPos += xBoneVertex*fWeight;
				}

				ProcessVertex(xVertPos);
			}
		}
		else
		{
			Asura_Assert(pxSkin->SectionData, "Null section data");
			if (!pxSkin->SectionData) return;

			for (int iBone=0; iBone<pxSkin->NumberOfBones; iBone++)
			{
				const Asura_Mesh* pxSectionMesh = pxSkin->SectionData[iBone].Mesh;
				if (!pxSectionMesh) continue;

				// Search the section mesh for the nearest point
				for (int i=0; i<pxSectionMesh->NumberOfVertices; i++)
				{
					Asura_Vector_3 xVertPos = pxSectionMesh->VertexDataPtr[i].Position;

					// Transform the point into world space
					xVertPos = xVertPos * axBoneOrientation[iBone];
					xVertPos += axBonePosition[iBone];

					ProcessVertex(xVertPos);
				}
			}		
		}
	}

	// Storage for the bone data.
	Asura_Vector_3 axBonePosition[ASURA_MAX_NO_OF_BONES_PER_MODEL];
	Asura_Matrix_3x3 axBoneOrientation[ASURA_MAX_NO_OF_BONES_PER_MODEL];
};

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_GetSnapPoint
///////////////////////////////////////////////////////////////////////////////
// Functor to find the closest snap point.
class Asura_Container_Functor_GetSnapPoint : public Asura_Container_Functor_ProcessMeshVertices
{
public:
	Asura_Container_Functor_GetSnapPoint(const Asura_Vector_3& xSearchPos, const Asura_Vector_3& xInitialBestSnapPos)
		: m_xSearchPosition(xSearchPos)
		, m_xBestSnapPoint(0.0f, 0.0f, 0.0f)
		, m_fMinDist(ASURA_LOS_MAX_DISTANCE)
	{
		// We've been given an initial best snap position. Process it.
		ProcessVertex(xInitialBestSnapPos);
	}

	const Asura_Vector_3& GetSnapPoint() const { return m_xBestSnapPoint; }

	// Found a point; see if it is closer to the search position than the previous points.
	virtual void ProcessVertex(const Asura_Vector_3& xPointPosition)
	{
		const float fDist = (xPointPosition - m_xSearchPosition).Magnitude();
		if (fDist < m_fMinDist)
		{
			m_fMinDist = fDist;
			m_xBestSnapPoint = xPointPosition;
		}
	}

protected:
	// Input
	Asura_Vector_3 m_xSearchPosition; // Start position for search.
	
	// Output
	Asura_Vector_3 m_xBestSnapPoint;
	float m_fMinDist;
};


///////////////////////////////////////////////////////////////////////////////
// Asura_Container_Functor_GetSnapBB
///////////////////////////////////////////////////////////////////////////////
// Functor to determine the BB in local object-space (or optionally in world-space).
class Asura_Container_Functor_GetSnapBB : public Asura_Container_Functor_ProcessMeshVertices
{
public:
	Asura_Container_Functor_GetSnapBB(const Asura_Vector_3& xEntityPos, const Asura_Matrix_3x3& xEntityOri, bool bWorldSpace = false)
		: m_xEntityPos(xEntityPos)
		, m_xEntityOri(xEntityOri)
		, m_bFoundPoint(false)
		, m_bWorldSpace(bWorldSpace)
	{
		m_xOutputBB.Reset();
	}

	bool HasFoundPoint() const { return m_bFoundPoint; }
	const Asura_Bounding_Box& GetBB() const { return m_xOutputBB; }


	// Found a point in world-space. Add it to the bounding box.
	virtual void ProcessVertex(const Asura_Vector_3& xPointPosition)
	{
		m_bFoundPoint = true;

		if (m_bWorldSpace)
		{
			m_xOutputBB.Set(xPointPosition);
		}
		else
		{
			// Convert the point from world-space to object-space.
			const Asura_Vector_3 xObjectPos = m_xEntityOri.ApplyTranspose( xPointPosition - m_xEntityPos );
			m_xOutputBB.Set(xObjectPos);
		}
	}

protected:
	// Input
	const Asura_Vector_3& m_xEntityPos;
	const Asura_Matrix_3x3& m_xEntityOri;
	
	// Output
	Asura_Bounding_Box m_xOutputBB;
	bool m_bFoundPoint;
	bool m_bWorldSpace;
};


//////////////////
// GetSnapPoint //
//////////////////
// Search the children for the closest point that we can use as a snap point.
Asura_Vector_3 Tools_Entity_Container::GetSnapPoint(const Asura_Vector_3& xPos) const
{
	// Start by checking the diamond-shaped vertices in the Tools mesh.
	const Asura_Vector_3 xSnapForToolsMesh = PARENT::GetSnapPoint(xPos);

	if (m_pxInstance)
	{
		Asura_Container_Functor_GetSnapPoint xFunctor(xPos, xSnapForToolsMesh);
		m_pxInstance->Traverse(xFunctor);
		return xFunctor.GetSnapPoint();
	}

	// Give up and just use the tools mesh.
	return xSnapForToolsMesh;
}

///////////////////////////////////
// GetBoundingBoxForSnapToVertex //
///////////////////////////////////
// Get bounding box in local object space for use by snap to vertex code.
void Tools_Entity_Container::GetBoundingBoxForSnapToVertex(Asura_Bounding_Box& xOutputBB) const
{
	if (m_pxInstance)
	{
		Asura_Container_Functor_GetSnapBB xFunctor(Position, Orientation, false);
		m_pxInstance->Traverse(xFunctor);
		if (xFunctor.HasFoundPoint())
		{
			xOutputBB = xFunctor.GetBB();
			return;
		}
	}

	// Give up and use the diamond shaped tools mesh instead.
	PARENT::GetBoundingBoxForSnapToVertex(xOutputBB);
}

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM
