///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Mode_Container.cpp
//
// Created by Martin G.Harris (5/6/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Tools_Mode_Container.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "Core/Asura_LOS.h"
#include "Core/Asura_StringBuffer.h"
#include "Entity/Container/Asura_ClientContainerInstance.h"
#include "Entity/Container/Asura_Container_System.h"
#include "Entity/Container/Asura_ServerEntity_Container.h"
#include "Rendering/Asura_Text_Output.h"

// Tools
#include "Entity/Container/Tools_Container_Attachment.h"
#include "Entity/Container/Tools_Container_Collection.h"
#include "Entity/Container/Tools_Container_System.h"
#include "Entity/Container/Tools_ContainerFile.h"
#include "Entity/Container/Tools_ContainerRegistry.h"
#include "Entity/Container/Tools_ContainerState.h"
#include "Entity/Container/Tools_Entity_Container.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Menubar.h"
#include "Interface/Asura_Menubar_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Tools_ContextMenu_System.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Render_Options.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Objects/Tools_Object_System.h"
#include "Tools_System.h"


// Helper
extern int SortTreeAlphabetically( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo );

// Static data
Tools_Mode_Container* Tools_Mode_Container::s_pxThis = 0;
Asura_Hash_ID Tools_Mode_Container::s_uContainerToEdit = ASURA_HASH_ID_UNSET;

Asura_Container_InputState Tools_Mode_Container::s_xStateToPreview;
Asura_Container_InputState Tools_Mode_Container::s_xStateToPreviewOld;

Asura_Guid Tools_Mode_Container::s_uEntityToEdit = ASURA_GUID_UNREGISTERED;
bool Tools_Mode_Container::s_bOpenFromFileOverlay = false;

Asura_Vector_3 Tools_Mode_Container::s_xPosition(0.0f, 0.0f, 0.0f);

// Note: Asura_Maths::IdentityMatrix3x3 may not have been set up yet!
Asura_Matrix_3x3 Tools_Mode_Container::s_xOrientation(1.0f, 0.0f, 0.0f,
													  0.0f, 1.0f, 0.0f,
													  0.0f, 0.0f, 1.0f);

Asura_Vector_2 Tools_Mode_Container::s_xInstanceOverlayPosition(30.0f, 5.0f);
Asura_Vector_2 Tools_Mode_Container::s_xInstanceTreeDimensions(30.0f, 45.0f);
bool Tools_Mode_Container::s_bInstanceOverlayOpen = true;
Tools_Container_InstanceTree Tools_Mode_Container::s_xInstanceTree;
int Tools_Mode_Container::s_iDeletionCounter = 0;
Tools_Container_Clipboard Tools_Mode_Container::s_xClipboard;
Tools_Container_UndoStorage Tools_Mode_Container::s_xUndoStorage;

bool s_bIconForAttachments = true;

///////////////////////////////////////////////////////////////////////////////
// Tools_Overlay_ContainerInstance
///////////////////////////////////////////////////////////////////////////////

class Tools_Overlay_ContainerInstance : public Asura_Overlay
{
public :
	Tools_Overlay_ContainerInstance(const char* szName, bool bAlwaysOnTop, Asura_Vector_2* pxPosition)
	: Asura_Overlay(szName, bAlwaysOnTop, pxPosition)
	{}

	~Tools_Overlay_ContainerInstance()
	{
		if (IsClosingByCloseGadget())
		{
			// Overlay was closed using the close gadget.
			Tools_Mode_Container::s_bInstanceOverlayOpen = false;
		}
	}
};

///////////////////////////////////////////////////////////////////////////////
// Tools_Container_InstanceTree
///////////////////////////////////////////////////////////////////////////////
Tools_Container_InstanceTree::Tools_Container_InstanceTree()
: m_pxHighlightedAttachment(0)
, m_uSavedSelectedContainerHash(ASURA_HASH_ID_UNSET)
, m_uNumCollapsedStored(0)
{
	ResetCollapsed();
}

void Tools_Container_InstanceTree::SetSelectedAttachment(Tools_Container_Object* pxHighlightedAttachment)
{
	m_xTree.SetSelectedItem(0);
	m_pxHighlightedAttachment = pxHighlightedAttachment;
}

void Tools_Container_InstanceTree::SetSelectedItem(Asura_TreeDialog_Item* pxItem)
{
	m_xTree.SetSelectedItem(pxItem);
	m_pxHighlightedAttachment = 0;
}

void Tools_Container_InstanceTree::SetSelectedContainer(Asura_Hash_ID uContainerHash)
{
	// Clear existing selection.
	SetSelectedItem(0);

	// Select specified container if it exists in the tree.
	Asura_TreeDialog_Item* pxItem = m_xTree.FindByData(static_cast<int>(uContainerHash));
	if (pxItem)
	{
		SetSelectedItem(pxItem);
	}
	else if (s_bIconForAttachments)
	{
		// Check for inlined attachments.
		Tools_Container_Object* pxToolsObject = Tools_Mode_Container::FindToolsObjectByContainerHash(uContainerHash);
		if (pxToolsObject && pxToolsObject->GetClassification() == AsuraContainerClass_Attachment)
		{
			SetSelectedAttachment(pxToolsObject);
		}
	}
}

void Tools_Container_InstanceTree::InitialiseSelection()
{
	if (m_uSavedSelectedContainerHash != ASURA_HASH_ID_UNSET)
	{
		SetSelectedContainer(m_uSavedSelectedContainerHash);
		m_uSavedSelectedContainerHash = ASURA_HASH_ID_UNSET;
		RestoreCollapsed();
	}
}

void Tools_Container_InstanceTree::ClearSelection()
{
	m_uSavedSelectedContainerHash = ASURA_HASH_ID_UNSET;
}

void Tools_Container_InstanceTree::DeleteAll()
{
	// Save the selected container, so that we can reselect it later.
	if (m_xTree.GetNumVisibleItems() > 0)
	{
		m_uSavedSelectedContainerHash = GetSelectedContainerHash();
		StoreCollapsed();
	}

	m_xTree.DeleteAll();
	m_pxHighlightedAttachment = 0;
}

void Tools_Container_InstanceTree::AddItem(const char* szLabel, Asura_Hash_ID uContainerHash)
{
	m_xTree.AddItem(szLabel, uContainerHash);
}

Asura_TreeDialog_Item* Tools_Container_InstanceTree::InsertItem(int ( *pfnCompFunc )( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo ),
									Asura_TreeDialog_Item* pxParent,
									const char* szLabel,
									Asura_Hash_ID uContainerHash)
{
	return m_xTree.InsertItem(pfnCompFunc, pxParent, szLabel, uContainerHash);
}

Tools_Container_Object* Tools_Container_InstanceTree::GetParentOfSelectedToolsObject()
{
	Tools_Container_Object* const pxToolsObject = GetSelectedToolsObject();
	Tools_Container_Object* const pxToolsParent = (pxToolsObject ? pxToolsObject->GetParent() : 0);
	return pxToolsParent;
}

Asura_Hash_ID Tools_Container_InstanceTree::GetSelectedContainerHash()
{
	Tools_Container_Object* pxToolsObject = GetSelectedToolsObject();
	return (pxToolsObject ? pxToolsObject->GetContainerHash() : ASURA_HASH_ID_UNSET);
}

Tools_Container_Object* Tools_Container_InstanceTree::GetSelectedToolsObject()
{
	const Asura_TreeDialog_Item* const pxItem = m_xTree.GetSelectedItem();
	if (pxItem)
	{
		m_pxHighlightedAttachment = 0;
		return static_cast<Tools_Container_Object*>(pxItem->GetUserPointer());
	}
	return m_pxHighlightedAttachment;
}

void Tools_Container_InstanceTree::TraverseUpdateStateInfo(Asura_TreeDialog_Item* pxItem)
{
	// Loop over the siblings
	for(; pxItem; pxItem = pxItem->GetSibling())
	{
		u_int uColour = 0;
		
		void* pUserPointer = pxItem->GetUserPointer();
		if (pUserPointer)
		{
			Tools_Container_Object* pxToolsObject = static_cast<Tools_Container_Object*>(pUserPointer);
			
			if (pxToolsObject->IsOff())
			{
				uColour = ASURA_GADGET_DISABLED_COLOUR;
			}
		}

		pxItem->SetColourOverride(uColour);

		// Recurse over children
		if (pxItem->GetChild())
		{
			TraverseUpdateStateInfo(pxItem->GetChild());
		}
	}
}

void Tools_Container_InstanceTree::ResetCollapsed()
{
	m_uNumCollapsedStored = 0;
	for (u_int u=0; u<MAX_NUM_COLLAPSED; ++u)
	{
		m_auCollapsed[u] = ASURA_HASH_ID_UNSET;
	}
}

void Tools_Container_InstanceTree::StoreCollapsed()
{
	ResetCollapsed();
	TraverseStoreCollapsedInfo(m_xTree.GetRoot());
}

void Tools_Container_InstanceTree::TraverseStoreCollapsedInfo(Asura_TreeDialog_Item* pxItem)
{
	// Loop over the siblings
	for(; pxItem; pxItem = pxItem->GetSibling())
	{
		if (!pxItem->IsExpanded())
		{
			if (m_uNumCollapsedStored<MAX_NUM_COLLAPSED)
			{
				m_auCollapsed[m_uNumCollapsedStored++] = static_cast<u_int>(pxItem->GetUserData());
			}
		}
		// Recurse over children
		if (pxItem->GetChild())
		{
			TraverseStoreCollapsedInfo(pxItem->GetChild());
		}
	}
}

void Tools_Container_InstanceTree::RestoreCollapsed()
{
	for (u_int u=0; u<m_uNumCollapsedStored; ++u)
	{
		const Asura_Hash_ID uContainerHash = m_auCollapsed[u];
		Asura_TreeDialog_Item* pxItem = m_xTree.FindByData(uContainerHash);
		if (pxItem)
		{
			pxItem->Expand(false);
		}
	}

	ResetCollapsed();
}


///////////////////////////////////////////////////////////////////////////////
// Tools_Container_Clipboard
///////////////////////////////////////////////////////////////////////////////
// Class to help cut/paste/copy stuff.
Tools_Container_Clipboard::Tools_Container_Clipboard()
{
	Reset();
}

Tools_Container_Clipboard::~Tools_Container_Clipboard()
{
}

void Tools_Container_Clipboard::Reset()
{
	m_uCopyContainerHash = ASURA_HASH_ID_UNSET;
	m_uParentContainerHash = ASURA_HASH_ID_UNSET;
	m_uGrandparentContainerHash = ASURA_HASH_ID_UNSET;

	m_bShouldCut = false;
}

bool Tools_Container_Clipboard::IsEmpty() const
{
	return m_uCopyContainerHash == ASURA_HASH_ID_UNSET ||
		(m_uParentContainerHash == ASURA_HASH_ID_UNSET && m_bShouldCut);
}

void Tools_Container_Clipboard::SetCut()
{
	Reset();
	Tools_Mode_Container::GetSelected(m_uCopyContainerHash, m_uParentContainerHash, m_uGrandparentContainerHash);
	m_bShouldCut = true;
}

void Tools_Container_Clipboard::SetCopy()
{
	Reset();
	Tools_Mode_Container::GetSelected(m_uCopyContainerHash, m_uParentContainerHash, m_uGrandparentContainerHash);
	m_bShouldCut = false;
}

// Paste to the destination container.
void Tools_Container_Clipboard::DoPaste(Tools_Container* pxPasteToContainer)
{
	if (pxPasteToContainer &&
		pxPasteToContainer->GetClassification() == AsuraContainerClass_Collection &&
		pxPasteToContainer->CanAddChild())
	{
		if (IsCopy())
		{
			DoCopy(pxPasteToContainer);
		}
		else if (IsCut())
		{
			DoCut(pxPasteToContainer);
		}
	}

	Reset();
}

// Move container from the old location to the new location.
void Tools_Container_Clipboard::DoCut(Tools_Container* pxPasteToContainer)
{
	Tools_Container* const pxContainer = Tools_Container_System::FindContainerByHash(m_uCopyContainerHash);
	if (!pxContainer) return; // Copy container no longer exists.

	Tools_Container* const pxParent = Tools_Container_System::FindContainerByHash(m_uParentContainerHash);
	if (!pxParent) return;	// Parent container no longer exists.

	Tools_Container_ChildIterator xIt(pxParent);
	if (!xIt.IsMember(m_uCopyContainerHash))
	{
		// Copy container is no longer a child of this parent.
		return;
	}

	const bool bParentIsAttachment = (pxParent->GetClassification() == AsuraContainerClass_Attachment);
	Tools_Container* const pxGrandparentContainer = Tools_Container_System::FindContainerByHash(m_uGrandparentContainerHash);
	if (bParentIsAttachment && !pxGrandparentContainer)
	{
		// Grandparent no longer exists.
		return;
	}

	// Exclude case where we are pasting back in to the same collection.
	const Asura_Hash_ID uPasteToHash = pxPasteToContainer->GetContainerHash();
	if (uPasteToHash == m_uParentContainerHash)
	{
		// Pasting back to the parent. Do nothing.
		return;
	}
	else if (bParentIsAttachment && uPasteToHash == m_uGrandparentContainerHash)
	{
		// Pasting back to the grandparent (the parent is just an attachment).
		// Do nothing.
		return;
	}

	// Make sure instances are deleted.
	Tools_Mode_Container::MarkDeletionStart();

	// Copies any attachment information, but it may not work properly in the new collection.
	// Copy the selected container, or if it has an attachment copy that instead.
	const Asura_Hash_ID uHashToCopy = (bParentIsAttachment ? m_uParentContainerHash : m_uCopyContainerHash);
	Tools_Container* const pxPasteFromCollection = (bParentIsAttachment ? pxGrandparentContainer : pxParent);

	// Try to preserve some state data and copy it to the new parent collection.
	// It'll only copy states which exist in both the old and the new parent collections.
	Tools_Mode_Container::s_xUndoStorage.StoreStateData(pxPasteFromCollection, m_uCopyContainerHash, uHashToCopy);

	// Add the container to the new location.
	pxPasteToContainer->AddChild(uHashToCopy);

	// Remove from the old location.
	if (pxPasteFromCollection)
	{
		pxPasteFromCollection->RemoveChild(uHashToCopy);
	}

	// Try to restore some state data.
	Tools_Mode_Container::s_xUndoStorage.RestoreStateData(pxPasteToContainer, m_uCopyContainerHash, uHashToCopy);

	// Make sure instances are recreated.
	Tools_Mode_Container::MarkDeletionStop();

	// Reselect the pasted container.
	Tools_Mode_Container::SelectInstanceContainer(m_uCopyContainerHash);
}

// Make a copy of the container in the new location.
void Tools_Container_Clipboard::DoCopy(Tools_Container* pxPasteToContainer)
{
	Tools_Container* const pxContainer = Tools_Container_System::FindContainerByHash(m_uCopyContainerHash);
	if (!pxContainer) return; // Copy container no longer exists.

	// Check whether there is an attachment that needs copying too.
	const Tools_Container_Attachment* pxOldAttachment = 0;

	const Tools_Container* const pxParent = Tools_Container_System::FindContainerByHash(m_uParentContainerHash);
	if (pxParent && pxParent->GetClassification() == AsuraContainerClass_Attachment)
	{
		pxOldAttachment = static_cast<const Tools_Container_Attachment*>(pxParent);
		if (pxOldAttachment->GetChildContainerHash() != m_uCopyContainerHash)
		{
			// Copy container is no longer a child of this parent.
			pxOldAttachment = 0;
		}
	}
	const bool bParentIsAttachment = (pxOldAttachment != 0);

	// Check for a grandparent.
	Tools_Container* const pxGrandparentContainer = Tools_Container_System::FindContainerByHash(m_uGrandparentContainerHash);
	if (bParentIsAttachment && !pxGrandparentContainer)
	{
		// Grandparent no longer exists.
		return;
	}

	Tools_Container* const pxNewContainer = Tools_Container_System::CreateNewContainerAsCopyFromMode(m_uCopyContainerHash, pxPasteToContainer);
	
	if (!pxNewContainer || pxNewContainer != Tools_Mode_Container::GetContainerSelected())
	{
		 // Failed to create or failed to select new container.
		return;
	}

	const Asura_Hash_ID uNewContainerHash = pxNewContainer->GetContainerHash();
	Asura_Hash_ID uNewChildHash = uNewContainerHash;

	// Insert an attachment if needed.
	if (pxOldAttachment && Tools_Mode_Container::IsValidForInsertAttachment())
	{
		Tools_Mode_Container::s_pxThis->InsertAttachment();
		Tools_Container* const pxNewAttach = Tools_Mode_Container::GetContainerSelected();

		if (pxNewAttach &&
			pxNewAttach->GetClassification() == AsuraContainerClass_Attachment &&
			pxNewAttach->GetContainerHash() != pxOldAttachment->GetContainerHash())
		{
			Tools_Container_Attachment* const pxNewAttachment = static_cast<Tools_Container_Attachment*>(pxNewAttach);

			// Copy the attachment data over. This overwrites the child data so we need to restore that.
			pxNewAttachment->CopyDataFromContainer(pxOldAttachment);
			pxNewAttachment->SetChildContainer(uNewContainerHash);

			uNewChildHash = pxNewAttachment->GetContainerHash();
		}
	}

	// Try to copy the state data to the new container.
	const Asura_Hash_ID uHashToCopy = (bParentIsAttachment ? m_uParentContainerHash : m_uCopyContainerHash);
	const Tools_Container* const pxPasteFromCollection = (bParentIsAttachment ? pxGrandparentContainer : pxParent);

	Tools_Mode_Container::s_xUndoStorage.StoreStateData(pxPasteFromCollection, uNewContainerHash, uHashToCopy);	
	Tools_Mode_Container::s_xUndoStorage.RestoreStateData(pxPasteToContainer, uNewContainerHash, uNewChildHash);
}


///////////////////////////////////////////////////////////////////////////////
// Tools_Container_UndoStorage
///////////////////////////////////////////////////////////////////////////////
Tools_Container_UndoStorage::Tools_Container_UndoStorage()
{
}

Tools_Container_UndoStorage::~Tools_Container_UndoStorage()
{
	while (!m_xAttachmentList.IsEmpty())
	{
		PopTailEntry_Attachment();
	}

	while (!m_xStateList.IsEmpty())
	{
		PopTailEntry_State();
	}
}

void Tools_Container_UndoStorage::PopTailEntry_Attachment()
{
	UndoData_Attachment* const pxData = m_xAttachmentList.PopTailObject();
	if (pxData)
	{
		delete pxData;
	}
}

void Tools_Container_UndoStorage::PopTailEntry_State()
{
	UndoData_State* const pxData = m_xStateList.PopTailObject();
	if (pxData)
	{
		delete pxData;
	}
}

void Tools_Container_UndoStorage::StoreData(const Tools_Container_Attachment* pxContainerToStore)
{
	if (!pxContainerToStore) return;

	const Asura_Hash_ID uChildHash = pxContainerToStore->GetChildContainerHash();
	if (uChildHash == ASURA_HASH_ID_UNSET) return;

	// Add a new entry to the head.
	UndoData_Attachment* const pxData = new UndoData_Attachment;
	pxData->m_uKeyContainerHash = uChildHash;

	// Determine size of chunk.
	Asura_Chunk_Dummy_Stream xDummy;
	pxContainerToStore->WriteToChunkStream(xDummy);
	
	const u_int uNumBytes = xDummy.GetPosition();
	pxData->m_pucStreamData = new u_char[uNumBytes];
	pxData->m_uNumBytes = uNumBytes;

	// Store chunk.
	Asura_Chunk_Memory_Stream xStream(uNumBytes, pxData->m_pucStreamData);
	xStream.SetSwapEndian(false);
	pxContainerToStore->WriteToChunkStream(xStream);

	// Add to head of list.
	m_xAttachmentList.AddToList(pxData);

	// Limit the size of the undo list, by dropping tail (i.e. oldest) entries if necessary.
	const int iMaxSizeOfList = 16;
	if (m_xAttachmentList.GetNumEntries() > iMaxSizeOfList)
	{
		PopTailEntry_Attachment();
	}
}

void Tools_Container_UndoStorage::RestoreData(Tools_Container* pxContainerToRestore, const Asura_Hash_ID uKeyContainerHash)
{
	if (!pxContainerToRestore) return;
	if (pxContainerToRestore->GetClassification() != AsuraContainerClass_Attachment) return;

	for (Asura_List_It<UndoData_Attachment> xIt(m_xAttachmentList); !xIt.Done(); xIt.Next())
	{
		UndoData_Attachment* const pxData = xIt.GetCurrent();
		if (pxData && pxData->m_uKeyContainerHash == uKeyContainerHash)
		{
			// Found it.
			if (pxData->m_pucStreamData)
			{
				const UndoData_Attachment* const pxConstData = pxData;
				Asura_Chunk_Memory_Stream xStream(pxConstData->m_uNumBytes, pxConstData->m_pucStreamData);
				Tools_Container_Attachment xTemporary;
				xTemporary.m_uClassification = AsuraContainerClass_Attachment;
				xTemporary.ReadFromChunkStream(xStream);

				// Restore the data.
				pxContainerToRestore->CopyDataFromContainer(&xTemporary);
			}

			// Delete the node.
			xIt.DeleteCurrent();
			delete pxData;

			return;
		}
	}
}

// Search for state undo information.
Tools_Container_UndoStorage::UndoData_State* Tools_Container_UndoStorage::FindState(Asura_Hash_ID uKeyContainerHash)
{
	for (Asura_List_It<UndoData_State> xIt(m_xStateList); !xIt.Done(); xIt.Next())
	{
		UndoData_State* const pxData = xIt.GetCurrent();
		if (pxData && pxData->m_uKeyContainerHash == uKeyContainerHash)
		{
			return pxData;
		}
	}
	return 0;
}

// Store/restore state information.
void Tools_Container_UndoStorage::StoreStateData(const Tools_Container* pxParentContainer,
												 const Asura_Hash_ID uKeyContainerHash,
												 const Asura_Hash_ID uChildContainerHash)
{
	if (!pxParentContainer) return;
	if (pxParentContainer->GetClassification() != AsuraContainerClass_Collection) return;

	if (uKeyContainerHash == ASURA_HASH_ID_UNSET) return;
	if (uChildContainerHash == ASURA_HASH_ID_UNSET) return;

	const Tools_ContainerStateMachine* const pxMachine = pxParentContainer->GetStateMachine();
	if (!pxMachine) return;

	const u_int uNumStates = pxMachine->GetNumStates();
	if (uNumStates == 0) return;

	// Search for an existing entry, so that we can append data; or create a new one.
	UndoData_State* pxUndoState = Tools_Container_UndoStorage::FindState(uKeyContainerHash);
	if (!pxUndoState)
	{
		pxUndoState = new UndoData_State;
		pxUndoState->m_uKeyContainerHash = uKeyContainerHash;

		// Add to head of list.
		m_xStateList.AddToList(pxUndoState);

		// Limit the size of the undo list, by dropping tail (i.e. oldest) entries if necessary.
		const int iMaxSizeOfList = 16;
		if (m_xStateList.GetNumEntries() > iMaxSizeOfList)
		{
			PopTailEntry_State();
		}
	}

	const Tools_Container* const pxChildContainer = Tools_Container_System::FindContainerByHash(uChildContainerHash);
	const u_int uNumChildInputs = (pxChildContainer ? pxChildContainer->GetNumInputs() : 1);

	// Store the existing states.
	for (u_int u=0; u<uNumStates; ++u)
	{
		const Tools_ContainerState* const pxState = pxMachine->GetState(u);
		if (!pxState || pxState->IsInvariant()) continue;

		// Loop over the child inputs if there is more than one.
		for (u_int v=0; v<uNumChildInputs; ++v)
		{
			const Asura_Hash_ID uChildStateHash = pxState->GetChildState(v, uChildContainerHash);

			if (!Asura_ContainerState::IsOff(uChildStateHash))
			{
				// Save the child state.
				pxUndoState->SetState(pxState->GetStateHash(), uChildStateHash, v);
			}
			else
			{
				// Remove the child state if it exists.
				pxUndoState->RemoveState(pxState->GetStateHash(), v);
			}
		}
	}
}

void Tools_Container_UndoStorage::RestoreStateData(Tools_Container* pxNewParentContainer,
												   const Asura_Hash_ID uKeyContainerHash,
												   const Asura_Hash_ID uChildContainerHash)
{
	if (!pxNewParentContainer) return;
	if (pxNewParentContainer->GetClassification() != AsuraContainerClass_Collection) return;

	if (uKeyContainerHash == ASURA_HASH_ID_UNSET) return;
	if (uChildContainerHash == ASURA_HASH_ID_UNSET) return;

	Tools_ContainerStateMachine* const pxMachine = pxNewParentContainer->GetStateMachine();
	if (!pxMachine) return;

	const u_int uNumStates = pxMachine->GetNumStates();
	if (uNumStates == 0) return;

	const UndoData_State* const pxUndoState = Tools_Container_UndoStorage::FindState(uKeyContainerHash);
	if (!pxUndoState) return;

	const Tools_Container* const pxChildContainer = Tools_Container_System::FindContainerByHash(uChildContainerHash);
	const u_int uNumChildInputs = (pxChildContainer ? pxChildContainer->GetNumInputs() : 1);

	// Restore any states that exist.
	for (u_int u=0; u<uNumStates; ++u)
	{
		Tools_ContainerState* const pxState = pxMachine->GetState(u);
		if (!pxState || pxState->IsInvariant()) continue;

		// Loop over the child inputs if there is more than one.
		for (u_int v=0; v<uNumChildInputs; ++v)
		{
			const Asura_Hash_ID uChildStateHash = pxUndoState->GetState(pxState->GetStateHash(), v);
			if (!Asura_ContainerState::IsOff(uChildStateHash))
			{
				pxState->ChangeChild(uChildContainerHash, uChildStateHash, v);
			}
		}
	}
}

////////////////////
// UndoData_State //
////////////////////
Tools_Container_UndoStorage::UndoData_State::UndoData_State()
: m_uKeyContainerHash(ASURA_HASH_ID_UNSET)
{
	for (u_int u=0; u<Tools_Container_UndoStorage::MAX_NUM_STATES; ++u)
	{
		m_auCollectionStateHash[u] = ASURA_HASH_ID_UNSET;
		m_auChildStateHash[u] = ASURA_HASH_ID_UNSET;
		m_auInputIndex[u] = 0;
	}
}

bool Tools_Container_UndoStorage::UndoData_State::IsMatch(u_int uIndex, Asura_Hash_ID uStateHash, u_int uInputIndex) const
{
	return m_auCollectionStateHash[uIndex] == uStateHash && m_auInputIndex[uIndex] == uInputIndex;
}

// Set the child state for the given collection state.
void Tools_Container_UndoStorage::UndoData_State::SetState(Asura_Hash_ID uStateHash, Asura_Hash_ID uChildStateHash, u_int uInputIndex)
{
	for (u_int u=0; u<Tools_Container_UndoStorage::MAX_NUM_STATES; ++u)
	{
		if (IsMatch(u, uStateHash, uInputIndex) || m_auCollectionStateHash[u] == ASURA_HASH_ID_UNSET)
		{
			m_auCollectionStateHash[u] = uStateHash;
			m_auChildStateHash[u] = uChildStateHash;
			m_auInputIndex[u] = uInputIndex;
			return;
		}
	}
}

// Unset the child state, effectively removing it.
void Tools_Container_UndoStorage::UndoData_State::RemoveState(Asura_Hash_ID uStateHash, u_int uInputIndex)
{
	for (u_int u=0; u<Tools_Container_UndoStorage::MAX_NUM_STATES; ++u)
	{
		if (m_auCollectionStateHash[u] == ASURA_HASH_ID_UNSET) return;
		
		if (IsMatch(u, uStateHash, uInputIndex))
		{
			m_auChildStateHash[u] = ASURA_HASH_ID_UNSET;
			m_auInputIndex[u] = 0;
			return;
		}
	}
}

// Get child state for given collection state (and given input index).
Asura_Hash_ID Tools_Container_UndoStorage::UndoData_State::GetState(Asura_Hash_ID uStateHash, u_int uInputIndex) const
{
	for (u_int u=0; u<Tools_Container_UndoStorage::MAX_NUM_STATES; ++u)
	{
		if (m_auCollectionStateHash[u] == ASURA_HASH_ID_UNSET)
		{
			return ASURA_HASH_ID_UNSET;
		}
		if ( IsMatch(u, uStateHash, uInputIndex) )
		{
			return m_auChildStateHash[u];
		}
	}

	return ASURA_HASH_ID_UNSET;
}

///////////////////////////////////////////////////////////////////////////////
// Tools_Mode_Container
///////////////////////////////////////////////////////////////////////////////
Tools_Mode_Container::Tools_Mode_Container()
: m_bOverlayWasOpen(false)
, m_bEditOverlayWasOpen(false)
, m_pxInstance(0)
, m_pxMainToolsObject(0)
, m_iMenuHandle(ASURA_INVALID_HANDLE)
, m_uPreviousInstanceContainerHash(ASURA_HASH_ID_UNSET)
{
	s_pxThis = this;
}

Tools_Mode_Container::~Tools_Mode_Container()
{
	s_pxThis = NULL;
}

//////////////////////
// TransferModeData //
//////////////////////
void Tools_Mode_Container::TransferModeData(int iMode)
{
	Tools_ModeDataFile xFile( "ContainerMode.dat", iMode );
 
	int iVersion = CURRENT_MODE_SAVE_VERSION;
    xFile.Transfer( iVersion );

    if ( iVersion >= 0 )
    {
        xFile.Transfer( s_xInstanceOverlayPosition );
		xFile.Transfer( s_xInstanceTreeDimensions );
		xFile.Transfer( s_bInstanceOverlayOpen );
    }

	if ( iVersion >= 1 )
	{
		xFile.Transfer( s_uContainerToEdit );

		// Only save/restore the first state input. TODO We could save/load them all.
		u_int& uStateHashToPreview = *s_xStateToPreview.GetStateHashArray();
		xFile.Transfer( uStateHashToPreview );
	}
}

///////////////////////////
// IsInstanceOverlayOpen //
///////////////////////////
bool Tools_Mode_Container::IsInstanceOverlayOpen()
{
	return s_bInstanceOverlayOpen;
}

///////////////////////////
// ToggleInstanceOverlay //
///////////////////////////
void Tools_Mode_Container::ToggleInstanceOverlay()
{
	if (IsInstanceOverlayOpen())
	{
		DestroyInstanceOverlay();
	}
	else
	{
		CreateInstanceOverlay();
	}
}

///////////////////////////
// CreateInstanceOverlay //
///////////////////////////
void Tools_Mode_Container::CreateInstanceOverlay()
{
	if (s_bInstanceOverlayOpen)
	{
		DestroyInstanceOverlay();
	}

	Asura_String256 xTitle("Container Instance");
	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(s_uContainerToEdit);
	if (pxContainer)
	{
		xTitle += ": ";
		xTitle += pxContainer->GetContainerName().GetString();
	}

	Asura_Overlay* pxNewOverlay = new Tools_Overlay_ContainerInstance( xTitle.GetString(), false, &s_xInstanceOverlayPosition);
    if ( !pxNewOverlay ) return;
    
	pxNewOverlay->m_uID = reinterpret_cast< u_int >( s_pxThis );

	if (s_pxThis)
	{
		Asura_Vector_2 xOffset = Asura_Vector_2( 1.0f, 3.0f );
		s_pxThis->AddInstanceOverlayItems(*pxNewOverlay, xOffset);
    }
	
	Asura_Overlay_System::SubmitNewOverlay( pxNewOverlay );

	s_bInstanceOverlayOpen = true;
}

////////////////////////////
// DestroyInstanceOverlay //
////////////////////////////
void Tools_Mode_Container::DestroyInstanceOverlay()
{
	s_bInstanceOverlayOpen = false;
	Asura_Overlay_System::CloseOverlays( reinterpret_cast< u_int >( s_pxThis ) );
}

/////////////////////////////
// AddInstanceOverlayItems //
/////////////////////////////
void Tools_Mode_Container::AddInstanceOverlayItems(Asura_Overlay& xOverlay, Asura_Vector_2& xOffset)
{
	RefreshInstanceOverlay(false);

	const int iThis = reinterpret_cast<int>(this);
	Asura_Dialog* pxDialog = 0;

	Asura_Vector_2 xSavedOffset = xOffset;
	
	// State to preview
	const char* const szTitle = (s_xStateToPreview.GetNumInputs() > 1) ? "(0) Preview State" : "Preview State";
	Tools_Container_MultipleState_Dialog* pxMultiDialog = new Tools_Container_MultipleState_Dialog( szTitle,
																									xOverlay,
																									xOffset,
																									&s_uContainerToEdit,
																									s_xStateToPreview.GetNumInputs(),
																									s_xStateToPreview.GetStateHashArray(),
																									0,
																									true);
	xOverlay.AddDialog(pxMultiDialog);

	xSavedOffset.x = 1.0f + Asura_Maths::Max(xOffset.x + 21.0f, pxMultiDialog->GetMaxCoordX());


	// Retrigger PFX, sound events etc.
	pxDialog = new Asura_Function_Dialog("Retrigger", xSavedOffset, Retrigger, iThis, 0);
	pxDialog->SetHelpText( "Retrigger any non-looping and transitional effects, such as PFX, sound events etc." );
	xOverlay.AddDialog( pxDialog );


	// Instance tree
	Asura_Tree_Dialog* pxNewTreeDialogue = new Asura_Tree_Dialog( xOffset, &s_xInstanceTree.GetTree(), s_xInstanceTreeDimensions.x, s_xInstanceTreeDimensions.y );
    pxNewTreeDialogue->SetDoubleClickedFunction( EditChild, iThis );
	pxNewTreeDialogue->SetDragDropCallback( Asura_Callback_Void_1Arg<void,Asura_TreeDialog_Item*>(DragAndDropTreeItem));
	pxNewTreeDialogue->SetIconCallback( Asura_Callback_Void_1Arg<int, Asura_TreeDialog_Item*>(GetTreeItemIconIndex) );
	pxNewTreeDialogue->SetIconClickCallback( Asura_Callback_Void_2Arg<void, Asura_TreeDialog_Item*, bool>(TreeIconClicked));
	pxNewTreeDialogue->SetUsePerItemHelp(true);
    xOverlay.AddDialog( pxNewTreeDialogue );
    xOffset.y += s_xInstanceTreeDimensions.y + 1.0f;

	//add control for resizing the tree
	pxDialog = new Asura_Resize_Dialog(pxNewTreeDialogue, &s_xInstanceTreeDimensions);
	pxDialog->SetRelativePosition(pxNewTreeDialogue, Asura_Vector_2(-pxDialog->GetWidth(), -pxDialog->GetHeight()), true, true);
    xOverlay.AddDialog( pxDialog );

	const float fColumnIdentX = 15.0f;

	// Options
	pxDialog = new Asura_Function_Dialog("Add", xOffset, AddChild, iThis, 0, IsValidForAdd);
	pxDialog->SetHelpText( "Add an already existing container to the selected collection." );
	xOverlay.AddDialog(pxDialog);
	xOffset.x += 9.0f;

	pxDialog = new Asura_Function_Dialog("Remove/Delete", xOffset, RemoveChild, iThis, 0, IsInstanceContainerSelectionValidForRemoval);
	pxDialog->SetHelpText( "Remove selected container. Delete the container and any children if they are no longer being used and have group-scope." );
	xOverlay.AddDialog(pxDialog);
	xOffset.x += 15.0f;
//	xOffset.x += fColumnIdentX;

	pxDialog = new Asura_Function_Dialog( "Edit", xOffset, EditChild, iThis, 0, IsInstanceContainerSelectionValid );
    pxDialog->SetHelpText( "Edit the currently selected container." );
    xOverlay.AddDialog( pxDialog );
	xOffset.x -= 24.0f;
//	xOffset.x -= fColumnIdentX;
    xOffset.y += Asura_Dialog::LargeSpacing();

	// Add creation functions.
	pxDialog = new Asura_Function_Dialog("Create", xOffset, CreateContainer, static_cast<int>(AsuraContainerClass_NotSet), 0, IsValidForCreate);
	pxDialog->SetHelpText( "Create container and add to the selected parent collection." );
	xOverlay.AddDialog( pxDialog );
	xOffset.x += fColumnIdentX;

	pxDialog = new Asura_Function_Dialog("Create with Attachment", xOffset, CreateContainer, static_cast<int>(AsuraContainerClass_Attachment), 0, IsValidForCreateWithAttachment);
	pxDialog->SetHelpText( "Create container and add to the selected parent collection. Insert a new attachment between them." );
	xOverlay.AddDialog( pxDialog );
	xOffset.x -= fColumnIdentX;
	xOffset.y += Asura_Dialog::LargeSpacing();

	// Add insertion/removal functions.
	pxDialog = new Asura_Function_Dialog("Insert Attachment", xOffset, InsertAttachment, iThis, 0, IsValidForInsertAttachment);
	pxDialog->SetHelpText( "Insert an attachment between selected container and its parent." );
	xOverlay.AddDialog( pxDialog );
	xOffset.x += fColumnIdentX;

	pxDialog = new Asura_Function_Dialog("Remove Attachment", xOffset, RemoveAttachment, iThis, 0, IsValidForRemoveAttachment);
	pxDialog->SetHelpText( "Remove attachment from the selected container." );
	xOverlay.AddDialog( pxDialog );
	xOffset.x -= fColumnIdentX;
	xOffset.y += Asura_Dialog::LargeSpacing();

	xOverlay.MakeDialogsRelativeToResizableDialog();

	xOverlay.SetContextMenuCallback(AddContextMenu, 0);
}

//////////////////////////
// GetTreeItemIconIndex //
//////////////////////////
int Tools_Mode_Container::GetTreeItemIconIndex(Asura_TreeDialog_Item* pxItem)
{
	if (pxItem)
	{
		void* const pUserPointer = pxItem->GetUserPointer();
		if (pUserPointer)
		{
			Tools_Container_Object* const pxToolsObject = static_cast<Tools_Container_Object*>(pUserPointer);
			if (s_bIconForAttachments)
			{
				if (pxToolsObject->IsParentAnAttachment())
				{
					const Tools_Container_Object* const pxParent = pxToolsObject->GetParent();
					if (pxParent && pxParent->GetContainerHash() == s_xInstanceTree.GetSelectedContainerHash())
					{
						return ASURA_TREEDIALOG_ICON_ATTACHMENT_HIGHLIGHT;
					}
					
					return ASURA_TREEDIALOG_ICON_ATTACHMENT;
				}
			}
			else if (pxItem->GetNumChildren())
			{
				if (pxItem->IsExpanded())
				{
					return ASURA_TREEDIALOG_ICON_OPENEDFOLDER;
				}
				else	
				{
					return ASURA_TREEDIALOG_ICON_CLOSEDFOLDER;
				}
			}
		}
	}

	return -1;
}

/////////////////////
// TreeIconClicked //
/////////////////////
void Tools_Mode_Container::TreeIconClicked(Asura_TreeDialog_Item* pxItem, bool bDrag)
{
	if (!pxItem) return;
	if (bDrag) return;

	if (s_bIconForAttachments)
	{
		void* const pUserPointer = pxItem->GetUserPointer();
		if (pUserPointer)
		{
			Tools_Container_Object* const pxToolsObject = static_cast<Tools_Container_Object*>(pUserPointer);
			if (pxToolsObject->IsParentAnAttachment())
			{
				Tools_Container_Object* const pxParent = pxToolsObject->GetParent();
				if (pxParent)
				{
					s_xInstanceTree.SetSelectedAttachment(pxParent);
					return;
				}
			}
		}
	}

	if (pxItem->GetNumChildren())
	{
		// Clicked on something else, so just expand/unexpand it.
		pxItem->Expand(!pxItem->IsExpanded());
	}
}

/////////////////////////
// DragAndDropTreeItem //
/////////////////////////
// Dragged a container in the tree.
void Tools_Mode_Container::DragAndDropTreeItem(Asura_TreeDialog_Item* pxDropTarget)
{
	if (!pxDropTarget) return;
	if (!IsValidForCutOrCopy()) return;

	Tools_Container_Object* pxDropTargetToolsObject = static_cast<Tools_Container_Object*>(pxDropTarget->GetUserPointer());
	if (!pxDropTargetToolsObject) return;
	
	const u_int uClassification = pxDropTargetToolsObject->GetClassification();
	if (uClassification != AsuraContainerClass_Collection)
	{
		return;
	}

	const Asura_Hash_ID uTargetContainerHash = pxDropTargetToolsObject->GetContainerHash();
	Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uTargetContainerHash);
	if (!pxContainer) return;

	s_xClipboard.SetCut();
	s_xClipboard.DoPaste(pxContainer);
}


////////////////////////////
// RefreshInstanceOverlay //
////////////////////////////
void Tools_Mode_Container::RefreshInstanceOverlay(bool bDeleting)
{
	s_xInstanceTree.DeleteAll();
	if (bDeleting)
	{
		UpdateInstanceOverlayStateInfo();
		return;
	}
	
	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(s_uContainerToEdit);
	if (pxContainer && pxContainer->CheckIsTooDeep())
	{
		s_xInstanceTree.AddItem("Circular Reference", s_uContainerToEdit);
		return;
	}

	// Populate the instance overlay tree.
	if (m_pxMainToolsObject)
	{
		m_pxMainToolsObject->AddToInstanceOverlayTree(s_xInstanceTree.GetRoot());
		UpdateInstanceOverlayStateInfo();

		// Try to reselect the container that was previously selected.
		s_xInstanceTree.InitialiseSelection();
	}
}

//////////////////////
// InsertItemInTree //
//////////////////////
// Insert item in instance overlay tree. Return new item.
Asura_TreeDialog_Item* Tools_Mode_Container::InsertItemInTree(Asura_TreeDialog_Item* pxParent,
															  const char* szLabel,
															  Asura_Hash_ID uContainerHash,
															  Tools_Container_Object* pxToolsObject)
{
	Asura_TreeDialog_Item* pxItem = s_xInstanceTree.InsertItem(SortTreeAlphabetically, pxParent, szLabel, uContainerHash);
	if (pxItem)
	{
		pxItem->SetUserPointer(pxToolsObject);
	}
	return pxItem;
}

////////////////////////////////////
// FindToolsObjectByContainerHash //
////////////////////////////////////
Tools_Container_Object* Tools_Mode_Container::FindToolsObjectByContainerHash(const Asura_Hash_ID uContainerHash)
{
	if (s_pxThis && s_pxThis->m_pxMainToolsObject)
	{
		return s_pxThis->m_pxMainToolsObject->FindContainer(uContainerHash);
	}
	return 0;
}

///////////////////////////
// RefreshStateToPreview //
///////////////////////////
void Tools_Mode_Container::RefreshStateToPreview()
{
	// Check that all of the inputs are valid. If not then replace them with default values.
	bool bOk = true;
	for (u_int u=0; u<s_xStateToPreview.GetNumInputs(); ++u)
	{
		const Asura_Hash_ID uStateHash = s_xStateToPreview.GetStateHash(u);
		if (uStateHash == Asura_ContainerState::GetOffStateHash())
		{
			// "Off" is always valid in tools container mode.
			continue;
		}

		// Check that the state is valid for the given slot.
		const bool bValidState = Tools_Container_System::IsContainerStateValid(s_uContainerToEdit, uStateHash, u);
		if (!bValidState)
		{
			if (u==0 && Tools_Container_System::IsContainerStateValid(s_uContainerToEdit, Asura_ContainerState::GetOnStateHash(), 0))
			{
				// Turn on.
				bOk = false;
				s_xStateToPreview.SetStateHash(u, Asura_ContainerState::GetOnStateHash());
			}
			else
			{
				// Turn off.
				bOk = false;
				s_xStateToPreview.SetStateHash(u, Asura_ContainerState::GetOffStateHash());
			}
		}
	}

	if (!bOk)
	{
		s_xStateToPreviewOld = s_xStateToPreview;
		UpdateInstanceOverlayStateInfo();
	}
}

//////////////////////////////////
// GetInstanceContainerSelected //
//////////////////////////////////
Asura_Hash_ID Tools_Mode_Container::GetInstanceContainerSelected()
{
	return s_xInstanceTree.GetSelectedContainerHash();
}

//////////////////////////
// GetContainerSelected //
//////////////////////////
// Get a container for the selected item.
Tools_Container* Tools_Mode_Container::GetContainerSelected()
{
	return Tools_Container_System::FindContainerByHash(GetInstanceContainerSelected());
}

////////////////////////////////
// GetContainerSelectedParent //
////////////////////////////////
// Get a container for the parent of the selected item.
Tools_Container* Tools_Mode_Container::GetContainerSelectedParent()
{
	Tools_Container_Object* const pxToolsParent = s_xInstanceTree.GetParentOfSelectedToolsObject();
	if (pxToolsParent)
	{
		return Tools_Container_System::FindContainerByHash(pxToolsParent->GetContainerHash());
	}
	return 0;
}

//////////////////////////////////////////////
// GetContainerSelectedParentForCanAddChild //
//////////////////////////////////////////////
// Search for the first parent of the selected container for which a child can be added.
// Usually this is the container itself or its parent collection.
Tools_Container* Tools_Mode_Container::GetContainerSelectedParentForCanAddChild()
{
	Tools_Container* pxContainer = 0;

	for (Tools_Container_Object* pxToolsObject = s_xInstanceTree.GetSelectedToolsObject() ;
		 pxToolsObject ;
		 pxToolsObject = pxToolsObject->GetParent() )
	{
		pxContainer = Tools_Container_System::FindContainerByHash(pxToolsObject->GetContainerHash());

		if (!pxContainer || pxContainer->CanAddChild())
		{
			break;
		}
	}

	return pxContainer;
}

/////////////////
// GetSelected //
/////////////////
// Get hashes of the selected container, its parent, and its grandparent.
void Tools_Mode_Container::GetSelected(Asura_Hash_ID& uSelectedHash, Asura_Hash_ID& uParentHash, Asura_Hash_ID& uGrandparentHash)
{
	uSelectedHash	 = ASURA_HASH_ID_UNSET;
	uParentHash		 = ASURA_HASH_ID_UNSET;
	uGrandparentHash = ASURA_HASH_ID_UNSET;

	const Tools_Container_Object* pxToolsSelected = s_xInstanceTree.GetSelectedToolsObject();
	if (!pxToolsSelected) return;
	uSelectedHash = pxToolsSelected->GetContainerHash();

	const Tools_Container_Object* pxToolsParent = pxToolsSelected->GetParent();
	if (!pxToolsParent) return;
	uParentHash = pxToolsParent->GetContainerHash();

	const Tools_Container_Object* pxToolsGrandparent = pxToolsParent->GetParent();
	if (!pxToolsGrandparent) return;
	uGrandparentHash = pxToolsGrandparent->GetContainerHash();
}


///////////////
// EditChild //
///////////////
void Tools_Mode_Container::EditChild()
{
	const Asura_Hash_ID uContainerHash = GetInstanceContainerSelected();
	Tools_Container_System::SelectContainer(uContainerHash, true);
}

//////////////
// AddChild //
//////////////
void Tools_Mode_Container::AddChild()
{
	Tools_Container* const pxContainer = GetContainerSelected();
	if (pxContainer && pxContainer->CanAddChild())
	{
		pxContainer->AddChildFromList();
	}
}

/////////////////
// RemoveChild //
/////////////////
void Tools_Mode_Container::RemoveChild()
{
	const Asura_Hash_ID uSelectedHash = s_xInstanceTree.GetSelectedContainerHash();
	if (uSelectedHash == s_uContainerToEdit) return;

	Tools_Container_Object* const pxToolsParent = s_xInstanceTree.GetParentOfSelectedToolsObject();
	if (!pxToolsParent) return;

	const bool bParentIsAttachment = (pxToolsParent->GetClassification() == AsuraContainerClass_Attachment);
	
	const Asura_Hash_ID uParentHash = pxToolsParent->GetContainerHash();

	Tools_Container* pxParentContainer = Tools_Container_System::FindContainerByHash(uParentHash);
	if (pxParentContainer)
	{	
		// Ask the user.
		const Tools_Container* pxChildContainer = Tools_Container_System::FindContainerByHash(uSelectedHash);
		if (pxChildContainer &&
			!pxChildContainer->PromptForDeletion(true))
		{
			return; // No
		}

		const bool bWasEditOpen = Tools_Container_System::IsEditOverlayOpen();
		pxParentContainer->RemoveChild(uSelectedHash);

		if (uParentHash != ASURA_HASH_ID_UNSET)
		{
			// Select the parent.
			Tools_Container_System::SelectContainer(uParentHash, bWasEditOpen);
			Tools_Mode_Container::SelectInstanceContainer(uParentHash);

			if (bParentIsAttachment)
			{
				// Remove the parent attachment too.
				RemoveAttachment();
			}
		}
	}
}

/////////////////////////////////////////////////
// IsInstanceContainerSelectionValidForRemoval //
/////////////////////////////////////////////////
bool Tools_Mode_Container::IsInstanceContainerSelectionValidForRemoval()
{
	const Asura_Hash_ID uSelectedHash = s_xInstanceTree.GetSelectedContainerHash();
	if (uSelectedHash == s_uContainerToEdit)
	{
		return false;
	}

	if (uSelectedHash == ASURA_HASH_ID_UNSET)
	{
		return false;
	}

	Tools_Container_Object* const pxToolsParent = s_xInstanceTree.GetParentOfSelectedToolsObject();
	if (!pxToolsParent)
	{
		return false;
	}

	return true;
}

//////////////////////
// IsValidForCreate //
//////////////////////
// Can create and add existing container?
bool Tools_Mode_Container::IsValidForCreate()
{
	const Tools_Container* const pxContainer = GetContainerSelectedParentForCanAddChild();
	if (pxContainer && pxContainer->CanAddChild())
	{
		return true;
	}

	return false;
}

////////////////////////////////////
// IsValidForCreateWithAttachment //
////////////////////////////////////
// Can create and add existing container with attachment?
bool Tools_Mode_Container::IsValidForCreateWithAttachment()
{
	const Tools_Container* const pxContainer = GetContainerSelectedParentForCanAddChild();
	if (pxContainer && pxContainer->CanAddChild() && pxContainer->GetClassification() != AsuraContainerClass_Attachment)
	{
		return true;
	}

	return false;
}

///////////////////
// IsValidForAdd //
///////////////////
// Can add existing container?
bool Tools_Mode_Container::IsValidForAdd()
{
	const Tools_Container* const pxContainer = GetContainerSelected();
	if (pxContainer && pxContainer->CanAddChild())
	{
		return true;
	}

	return false;
}

////////////////////////////////
// IsValidForInsertAttachment //
////////////////////////////////
// Can insert if neither selected nor parent of selected is an attachment.
bool Tools_Mode_Container::IsValidForInsertAttachment()
{
	const Tools_Container* const pxContainer = GetContainerSelected();
	if (pxContainer && !Tools_ContainerRegistry::CanUseAttachment(pxContainer->GetClassification()))
	{
		return false;
	}

	const Tools_Container* const pxParentContainer = GetContainerSelectedParent();
	if (!pxParentContainer || pxParentContainer->GetClassification() == AsuraContainerClass_Attachment)
	{
		return false;
	}
	
	return true;
}

////////////////////////////////
// IsValidForRemoveAttachment //
////////////////////////////////
// Can remove if the selected item is an attachment.
bool Tools_Mode_Container::IsValidForRemoveAttachment()
{
	const Tools_Container* const pxContainer = GetContainerSelected();
	if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Attachment)
	{
		return true;
	}

	Tools_Container* pxParent = Tools_Mode_Container::GetContainerSelectedParent();
	if (pxParent && pxParent->GetClassification() == AsuraContainerClass_Attachment)
	{
		return true;
	}

	return false;
}

/////////////////////////
// IsValidForCutOrCopy //
/////////////////////////
// Can cut/copy the selected item?
// For now, don't allow attachments or collections to be cut/copied.
bool Tools_Mode_Container::IsValidForCutOrCopy()
{
	const Tools_Container* const pxContainer = GetContainerSelected();
	if (!pxContainer) return false;
	
	const u_int uClass = pxContainer->GetClassification();
	switch (uClass)
	{
	case AsuraContainerClass_NotSet:
	case AsuraContainerClass_Attachment:
	case AsuraContainerClass_Collection:
		return false;
	
	default:
		break;
	}
	
	const Tools_Container* const pxParent = GetContainerSelectedParent();
	if (!pxParent) return false;

	return true;
}

/////////////////////
// IsValidForPaste //
/////////////////////
// Can paste the selected item?
// Only allow pasting in to collections.
bool Tools_Mode_Container::IsValidForPaste()
{
	if (s_xClipboard.IsEmpty()) return false;

	const Tools_Container* const pxContainer = GetContainerSelected();
	if (!pxContainer) return false;

	const u_int uClass = pxContainer->GetClassification();
	return uClass == AsuraContainerClass_Collection;
}

//////////////////////
// InsertAttachment //
//////////////////////
void Tools_Mode_Container::InsertAttachment()
{
	const Tools_Container* const pxContainer = GetContainerSelected();
	if (!pxContainer) return;
	const Asura_Hash_ID uContainerHash = pxContainer->GetContainerHash();

	// Find parent container and group.
	Tools_Container* const pxParent = GetContainerSelectedParent();
	if (!pxParent) return;

	const Asura_Hash_ID uGroupHash = pxParent->GetGroupHash();
	Tools_ContainerGroup* const pxGroup = Tools_Container_System::FindGroupByHash(uGroupHash);
	if (!pxGroup)
	{
		Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Container group (0x%08x) not found.", uGroupHash);
		return;
	}

	// Auto-generate a suitable name and create class.
	const u_int uAutoGenNumber = pxGroup->GetAutoGenNumber();

	// Use root name by default, rather than parent name, as it should be shorter.
	const Asura_Hash_ID uRootHash = GetContainerToEditHash();
	const Tools_Container* pxRoot = Tools_Container_System::FindContainerByHash(uRootHash);
	if (!pxRoot) return;
	
	const char* const szParentName = pxRoot->GetContainerName().GetString();
	Asura_Assert(szParentName, "Collection container name is null");

	// Create new attachment.
	const u_int uClassification = AsuraContainerClass_Attachment;
	Tools_Container* pxNewContainer = Tools_Container_System::CreateNewContainerFromMode_Helper(uClassification, uAutoGenNumber, szParentName, pxGroup);
	if (!pxNewContainer)
	{
		return;
	}

	// Link attachment to selected container.
	pxNewContainer->AddChild(uContainerHash);

	const Asura_Hash_ID uNewContainerHash = pxNewContainer->GetContainerHash();

	// Restore any data to the attachment, if the attachment had previously been removed.
	s_xUndoStorage.RestoreData(pxNewContainer, uContainerHash);

	// Register the new container.
	Tools_Container_System::RegisterContainer(pxNewContainer, true, true);
	pxNewContainer = 0;

	// Add attachment to parent, remove old link if necesary.
	pxParent->ReplaceChild(uContainerHash, uNewContainerHash);
	pxParent->SetModified();

	SelectInstanceContainer(uNewContainerHash);
}

//////////////////////
// RemoveAttachment //
//////////////////////
void Tools_Mode_Container::RemoveAttachment()
{
	// Check whether the selected container is an attachment.
	Tools_Container_Object* pxToolsObject = s_xInstanceTree.GetSelectedToolsObject();
	if (!pxToolsObject) return;

	if (pxToolsObject->GetClassification() != AsuraContainerClass_Attachment)
	{
		// Check whether the parent is an attachment.
		pxToolsObject = pxToolsObject->GetParent();
		if (!pxToolsObject || pxToolsObject->GetClassification() != AsuraContainerClass_Attachment)
		{
			// Give up.
			return;
		}
	}

	Asura_Hash_ID uContainerHash = pxToolsObject->GetContainerHash();
	Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
	if (!pxContainer || pxContainer->GetClassification() != AsuraContainerClass_Attachment)
	{
		return;
	}

	Tools_Container_Attachment* const pxAttachment = static_cast<Tools_Container_Attachment*>(pxContainer);
	const Asura_Hash_ID uAttachmentHash = pxAttachment->GetContainerHash();
	const Asura_Hash_ID uChildHash = pxAttachment->GetChildContainerHash();

	Tools_Container_Object* pxToolsParent = pxToolsObject->GetParent();
	if (!pxToolsParent) return;

	Tools_Container* const pxParentContainer = Tools_Container_System::FindContainerByHash(pxToolsParent->GetContainerHash());
	if (!pxParentContainer) return;

	// Store a copy of the attachment, in case user changes his mind and wants to add it back.
	s_xUndoStorage.StoreData(pxAttachment);

	// Make sure instances are deleted.
	MarkDeletionStart();
	Tools_Container_System::SetContainerModified(uChildHash, false);

	pxParentContainer->ReplaceChild(uAttachmentHash, uChildHash);

	Tools_Container_System::DeleteIfOrphaned(uAttachmentHash);

	// Make sure instances are recreated.
	MarkDeletionStop();

	// Reselect the child.
	Tools_Mode_Container::SelectInstanceContainer(uChildHash);
}

///////////////
// Retrigger //
///////////////
void Tools_Mode_Container::Retrigger()
{
	if (m_pxInstance && !s_xStateToPreview.IsOff())
	{
		// Destroy all of the container instances. Set all the inputs to be "off".
		Asura_Container_InputState xOffState = s_xStateToPreview;
		xOffState.SetOff();
		m_pxInstance->ChangeState(xOffState);

		// Recreate them, but pretend that it is a state change, rather than a creation.
		Asura_Container_System::SetDoingStateChange(true);
		m_pxInstance->ChangeState(s_xStateToPreview);
		Asura_Container_System::SetDoingStateChange(false);

		Asura_ContainerInstance_Properties xData;
		m_pxInstance->Initialise(0, xData);
		m_pxInstance->TeleportTo(s_xPosition, s_xOrientation);
	}
}

/////////////////////
// CreateContainer //
/////////////////////
void Tools_Mode_Container::CreateContainer(int iInsertClass)
{
	const u_int uInsertClass = static_cast<u_int>(iInsertClass);
	Tools_Container_System::CreateNewContainerFromMode(uInsertClass);
}

///////////////////////////
// UpdateInstanceOverlay //
///////////////////////////
void Tools_Mode_Container::UpdateInstanceOverlay()
{
	if (!s_bInstanceOverlayOpen) return;

	const Asura_Hash_ID uContainerHash = GetInstanceContainerSelected();
	const bool bChanged = (uContainerHash != m_uPreviousInstanceContainerHash);
	m_uPreviousInstanceContainerHash = uContainerHash;

	if (uContainerHash != ASURA_HASH_ID_UNSET && bChanged)
	{
		// Don't change selection during volume editing.
		const Tools_Object* pxSelected = Tools_Object_System::SelectedObject;
		if (!pxSelected || pxSelected->Type != TOT_VolumeEdit)
		{
			// Selection has changed.
			// Find the relevant tools object and select it.
			for(LIF<Tools_Object*> xIt(&Tools_Object_System::ObjectList); !xIt.done(); xIt.next())
			{
				const int iType = xIt()->Type;
				if (iType == TOT_Container_Object_Attachment ||
					iType == TOT_Container_Object_Collection ||
					iType == TOT_Container_Object_Misc)
				{
					Tools_Container_Object* pxObject = static_cast<Tools_Container_Object*>(xIt());
					if (pxObject->GetContainerHash() == uContainerHash)
					{
						Tools_Object_System::SelectObject(pxObject);
						break;
					}
				}
			}
		}
	}
}

/////////////////////////////
// SelectInstanceContainer //
/////////////////////////////
// TODO if we have multiple instances of the same container then we need to select the correct one.
void Tools_Mode_Container::SelectInstanceContainer(const Asura_Hash_ID uContainerHash)
{
	if (!s_pxThis) return;

	if (s_pxThis->m_uPreviousInstanceContainerHash == uContainerHash)
	{
		return;
	}

	s_xInstanceTree.SetSelectedContainer(uContainerHash);
}

///////////////
// PrintInfo //
///////////////
void Tools_Mode_Container::PrintInfo(float& fX, float& fY)
{
	Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(s_uContainerToEdit);
	if (pxContainer)
	{
		Asura_String xName;
		pxContainer->MakeFullName(xName);
		//const Asura_String& xName = pxContainer->GetContainerName();
		Asura_Text_Output::Print(fX, fY++, 0xffffffff, 0, "Container: %s", xName.GetString());
		Asura_Text_Output::Print(fX, fY++, 0xffffffff, 0, "Container hash: 0x%08x", pxContainer->GetContainerHash());
		fY++;
	}
}

////////////////////
// InitialiseMode //
////////////////////
void Tools_Mode_Container::InitialiseMode()
{
	PARENT::InitialiseMode();

	Tools_Render_Options::SetPFXControlledByMode(true);

	s_xClipboard.Reset();

	// Create the menubar.
	m_iMenuHandle = Asura_Menubar_System::CreateMenuSection();
	Asura_Menubar_Title* pxTitle = Asura_Menubar_System::AddTitle(m_iMenuHandle, "Container Mode");

	// Duplicate of entry on main "Utility" menu. Main file overlay.
	pxTitle->AddItem("File Overlay", Tools_Container_System::ToggleOverlay, Tools_Container_System::IsOverlayOpen );
	// Editing overlay.
	pxTitle->AddItem("Edit Overlay", Tools_Container_System::ToggleEditOverlay, Tools_Container_System::IsEditOverlayOpen);
	// Instance overlay.
	pxTitle->AddItem("Instance Overlay", ToggleInstanceOverlay, IsInstanceOverlayOpen);

	const Asura_Hash_ID uTempContainerToEdit    = s_uContainerToEdit;
	const Asura_Container_InputState xTempStateToPreview = s_xStateToPreview;

	TransferModeData(TOOLSMODEDATAFILE_LOAD);

	// Record which overlays are open.
	m_bOverlayWasOpen = Tools_Container_System::IsOverlayOpen();
	m_bEditOverlayWasOpen = Tools_Container_System::IsEditOverlayOpen();

	const bool bRestore = (s_uEntityToEdit != ASURA_GUID_UNREGISTERED || s_bOpenFromFileOverlay);
	if (bRestore)
	{
		// Restore the requested values, which were overwritten by the call to TransferModeData.
		s_uContainerToEdit = uTempContainerToEdit;
		s_xStateToPreview  = xTempStateToPreview;
	}

	// Check whether user has entered directly from entity mode, by requesting to edit a specific entity.
	if (s_uEntityToEdit != ASURA_GUID_UNREGISTERED)
	{
		if (Tools_Container_System::FindContainerByHash(s_uContainerToEdit) == 0)
		{
			// Couldn't find the container, stop trying to edit the entity.
			s_uEntityToEdit = ASURA_GUID_UNREGISTERED;
			Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Container (0x%08x) not found.", s_uContainerToEdit);
		}
	}

	bool bOk = true;

	if (s_uEntityToEdit == ASURA_GUID_UNREGISTERED && !s_bOpenFromFileOverlay)
	{
		// Prompt user to select or create container collection.
		bOk = Tools_Container_System::CreateOrSelectModalOverlay(s_uContainerToEdit);
	
		if (bOk && s_uContainerToEdit == ASURA_HASH_ID_UNSET)
		{
			// Prompt for file, existing or new group, new container name, and create collection.
			bOk = Tools_Container_System::CreateCollectionModalOverlay(s_uContainerToEdit);
			if (bOk)
			{
				// Default to "on" state for new collections.
				s_xStateToPreview.SetStateHash(0, Asura_ContainerState::GetOnStateHash());
			}
		}
	}

	// Clear this flag now.
	s_bOpenFromFileOverlay = false;

	// Check that the container exists.
	Tools_Container* pxContainer = 0;
	if (bOk)
	{
		pxContainer = Tools_Container_System::FindContainerByHash(s_uContainerToEdit);
		Asura_Assert(pxContainer, "Container pointer null");
		bOk = (pxContainer != 0);
	}

	// Check for circular references.
	if (pxContainer && pxContainer->CheckIsTooDeep(true))
	{
		bOk = false;
	}

	if (!bOk)
	{
		// Cancelled, switch to standby. Don't bother prompting for save.
		Tools_System::RequestStandbyMode(true);
		// Clear parameters.
		s_uEntityToEdit = ASURA_GUID_UNREGISTERED;
		s_uContainerToEdit = ASURA_HASH_ID_UNSET;
		Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_INFO, "Exited container mode, changing to standby.");
		return;
	}

	// Set the number of state inputs from the collection's state machine.
	const Tools_ContainerStateMachine* const pxMachine = pxContainer->GetStateMachine();
	s_xStateToPreview.SetNumInputs(pxMachine ? pxMachine->GetNumInputs() : 1);

	pxContainer->LoadRequiredFiles();
	
	// Recursively verify this container.
	Tools_Container_System::VerifyContainer(pxContainer);

	// If the file is writeable then test for, and if needed automatically fix up, certain types of error.
	const Tools_ContainerFile* pxFile = Tools_Container_System::FindFileByHash(pxContainer->GetFileHash());
	if (pxFile && pxFile->IsWriteable())
	{
		pxContainer->FixUpErrors();
	}

	// Warn if container file is read only.
	Tools_Container_System::WarnIfContainerIsReadOnly(s_uContainerToEdit);

	Tools_Object_System::EnableAllControls(true);
	
	// Create main tools object and select it.
	CreateToolsInstance();
	CreateInstance();

	if (m_pxMainToolsObject)
	{
		if (s_uEntityToEdit != ASURA_GUID_UNREGISTERED)
		{
			// Came from entity mode, camera is probably pointing at the entity anyway.
			Tools_Object_System::SelectObject(m_pxMainToolsObject);
		}
		else
		{
			Tools_Object_System::GotoAndSelect(m_pxMainToolsObject);
		}
	}

	if (Tools_Container_System::IsOverlayOpen())
	{
		// Select specified collection in tree view, if overlay is open.
		Tools_Container_System::SelectContainer(s_uContainerToEdit, m_bEditOverlayWasOpen);
	}

	// These objects just get in the way at the moment. Hide them.
	Tools_Object_System::SetTypeHidden(TOT_Container_Object_Collection, ObjectHideFlag_Mode, true);
	Tools_Object_System::SetTypeHidden(TOT_Container_Object_Misc, ObjectHideFlag_Mode, true);

	if (s_bInstanceOverlayOpen)
	{
		CreateInstanceOverlay();
	}

	s_xStateToPreviewOld = s_xStateToPreview;
}

//////////////////
// ShutdownMode //
//////////////////
void Tools_Mode_Container::ShutdownMode()
{
	// If we are reselecting container mode then don't close the file overlay as it is annoying.
	if (Tools_System::NextMode != TOOLS_MODE_CONTAINER)
	{
		// Close the overlay if it wasn't open when we entered this mode.
		if (!m_bOverlayWasOpen)
		{
			if (Tools_Container_System::IsOverlayOpen())
			{
				Tools_Container_System::ToggleOverlay();
			}
		}
	}

	// Close the edit overlay if it wasn't open when we entered this mode.
	if (!m_bEditOverlayWasOpen)
	{
		if (Tools_Container_System::IsEditOverlayOpen())
		{
			Tools_Container_System::GetOverlayManager().CloseEditOverlay();
		}
	}
	
	// Destroy main object and client instance.
	DestroyToolsInstance();
	DestroyInstance();

	Tools_Render_Options::SetPFXControlledByMode(false);

	PARENT::ShutdownMode();

	TransferModeData(TOOLSMODEDATAFILE_SAVE);

	if (s_bInstanceOverlayOpen)
	{
		DestroyInstanceOverlay();
	}

	Tools_Object_System::EnableAllControls(false);

	s_uEntityToEdit = ASURA_GUID_UNREGISTERED;

	// Shutdown the menu
	if(m_iMenuHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Menubar_System::DestroyMenuSection(m_iMenuHandle);
		m_iMenuHandle = ASURA_INVALID_HANDLE;
	}

	RefreshInstanceOverlay(true);

	s_xClipboard.Reset();

	Tools_Container_Object::ResetStoredHiddenFlag();
}

///////////////////////////////////////////
// RequestModeForSelectedContainerEntity //
///////////////////////////////////////////
// Request this mode for a specific container entity.
void Tools_Mode_Container::RequestModeForSelectedContainerEntity(Asura_Hash_ID uContainerHash, Asura_Guid uEntityGuid)
{
	if (uContainerHash == ASURA_HASH_ID_UNSET) return;
	if (uEntityGuid == ASURA_GUID_UNREGISTERED) return;

	Asura_ServerEntity_Container* pxServer = Asura_ServerEntity_Container::FindEntity(uEntityGuid);
	if (!pxServer) return;

	// Set up initial parameters for this mode.
	s_uContainerToEdit = uContainerHash;
	s_uEntityToEdit = uEntityGuid;
	
	s_xStateToPreview = pxServer->GetInputState();
	s_xStateToPreviewOld = s_xStateToPreview;

	s_xPosition = pxServer->GetPosition();
	pxServer->GetOrientationQuat().ToMatrix(s_xOrientation);
	s_bOpenFromFileOverlay = false;
	
	// Request a switch to the mode.
	Tools_System::RequestContainerMode();
}

/////////////////////////////////////
// RequestModeForSelectedContainer //
/////////////////////////////////////
void Tools_Mode_Container::RequestModeForSelectedContainer(Asura_Hash_ID uContainerHash)
{
	if (uContainerHash == ASURA_HASH_ID_UNSET) return;

	// Set up initial parameters for this mode.
	s_uContainerToEdit = uContainerHash;
	s_uEntityToEdit = ASURA_GUID_UNREGISTERED;
	s_bOpenFromFileOverlay = true;
	
	// Request a switch to the mode.
	Tools_System::RequestContainerMode();
}

///////////////
// QuickSave //
///////////////
void Tools_Mode_Container::QuickSave()
{
	Tools_Container_System::SaveModified();
}

////////////////
// IsModified //
////////////////
bool Tools_Mode_Container::IsModified()
{
	return Tools_Container_System::IsModified();
}

/////////////////
// PreShutdown //
/////////////////
void Tools_Mode_Container::PreShutdown()
{
	Tools_Container_System::SavePrompt(false);
}

/////////////////////
// DestroyInstance //
/////////////////////
// Destroy the instance.
void Tools_Mode_Container::DestroyInstance()
{
	Asura_Container_System::DestroyInstance(m_pxInstance);
	m_pxInstance = 0;
	m_uPreviousInstanceContainerHash = ASURA_HASH_ID_UNSET;
}

////////////////////
// CreateInstance //
////////////////////
// Create an instance.
void Tools_Mode_Container::CreateInstance()
{
	// Destroy any existing instance first.
	DestroyInstance();

	if (s_uContainerToEdit == ASURA_HASH_ID_UNSET)
	{
		return;
	}

	// Check that the container exists.
	const Tools_Container* pxToolsContainer = Tools_Container_System::FindContainerByHash(s_uContainerToEdit);
	if (!pxToolsContainer)
	{
		if (Tools_Container_System::IsInitialised())
		{
			Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Container (0x%08x) not found.", s_uContainerToEdit);
		}
		return;
	}

	// Check for circular references before making the instance, otherwise it would crash.
	if (pxToolsContainer->CheckIsTooDeep())
	{
		return;
	}

	RefreshStateToPreview();

	// Create a new instance.
	if (Asura_Container_System::GetTree().GetNumberOfNodes() > 0)
	{
		if (!s_xStateToPreview.IsOff())
		{
			m_pxInstance = Asura_Container_System::CreateClientInstance(s_uContainerToEdit);
			Asura_Assertf(m_pxInstance, ("Failed to create container instance (0x%08x)", s_uContainerToEdit));
		}

		if (m_pxInstance)
		{
			// We are creating the whole thing from scratch, don't play transition-only effects.
			Asura_Container_System::SetDoingCreate(true);
			m_pxInstance->ChangeState(s_xStateToPreview);
			Asura_Container_System::SetDoingCreate(false);

			Asura_ContainerInstance_Properties xData;
			m_pxInstance->Initialise(0, xData);
			m_pxInstance->TeleportTo(s_xPosition, s_xOrientation);
		}
	}
}

//////////////////////////
// DestroyToolsInstance //
//////////////////////////
void Tools_Mode_Container::DestroyToolsInstance()
{
	// The instance tree contains pointers to the Tools objects, so we need to clean those up too.
	// Just delete the whole tree.
	s_xInstanceTree.DeleteAll();

	if (m_pxMainToolsObject)
	{
		m_pxMainToolsObject->Destroy();
	}
	m_pxMainToolsObject = 0;
}

/////////////////////////
// CreateToolsInstance //
/////////////////////////
void Tools_Mode_Container::CreateToolsInstance()
{
	DestroyToolsInstance();

	m_pxMainToolsObject = new Tools_Container_Object_Origin(s_uContainerToEdit);
	if (m_pxMainToolsObject)
	{
		m_pxMainToolsObject->TeleportTo(s_xPosition, s_xOrientation);
	}
}

/////////////////////////
// DestroyAllInstances //
/////////////////////////
// Destroy both tools and client instances.
void Tools_Mode_Container::DestroyAllInstances()
{
	if (s_pxThis)
	{
		s_pxThis->DestroyToolsInstance();
		s_pxThis->DestroyInstance();
	}
}

////////////////////////
// CreateAllInstances //
////////////////////////
// Create both tools and client instances.
void Tools_Mode_Container::CreateAllInstances()
{
	if (s_pxThis)
	{
		s_pxThis->CreateToolsInstance();
		s_pxThis->CreateInstance();
	}
}

/////////////////
// EarlyRender //
/////////////////
void Tools_Mode_Container::EarlyRender()
{
	PARENT::EarlyRender();

	if (!m_pxInstance) return;

	// Set outline colour.
	Tools_Entity_Container::SetOutlineColour(0);

	/*
	Tools_Object* pxRootObject = GetObjectToSelect();
	if (pxRootObject && pxRootObject->ShouldRenderSelectionOutline())
	{
		if (pxRootObject->RenderAsSelected())
		{
			s_uOutlineColour = 0xffff0000;
		}
		else if (pxRootObject->IsGroupSelected())
		{
			s_uOutlineColour = 0xff0000ff;
		}
	}
	*/

	// Render the in-game meshes, shadows, effects etc.
	m_pxInstance->Render(s_xPosition, s_xOrientation);
	m_pxInstance->RenderShadow(0);

	// Clear outline colour
	//Tools_Entity_Container::SetOutlineColour(0);
}

////////////
// Update //
////////////
void Tools_Mode_Container::Update()
{
	PARENT::Update();

	if (m_pxMainToolsObject && m_pxMainToolsObject->HasBeenMoved())
	{
		// Origin has been moved by user, make sure client instance is up to date.
		if (m_pxInstance)
		{
			m_pxInstance->TeleportTo(s_xPosition, s_xOrientation);	
		}
		m_pxMainToolsObject->ClearMovedFlag();
	}

	if (m_pxInstance)
	{
		m_pxInstance->Update();
	}

	HandleKeyboard();
	UpdateInstanceOverlay();
}

////////////////////
// HandleKeyboard //
////////////////////
void Tools_Mode_Container::HandleKeyboard()
{
	
}

/////////////////
// CreateChild //
/////////////////
// Create tools objects.
Tools_Container_Object* Tools_Mode_Container::CreateChild(Asura_Hash_ID uContainerHash)
{
	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
	if (pxContainer)
	{
		switch (pxContainer->GetClassification())
		{
		case AsuraContainerClass_Collection:
			return new Tools_Container_Object_Collection(uContainerHash);

		case AsuraContainerClass_Attachment:
			return new Tools_Container_Object_Attachment(uContainerHash);

		default:
			// Other types.
			return new Tools_Container_Object_Misc(uContainerHash);
		}
	}
	else if (uContainerHash != ASURA_HASH_ID_UNSET)
	{
		// Container not found. Generate a default placeholder object.
		return new Tools_Container_Object_Misc(uContainerHash);
	}

	return 0;
}

/////////////
// Refresh //
/////////////
// Refresh the instance when a container has been modified.
void Tools_Mode_Container::Refresh(const Asura_Hash_ID uContainerChanged, bool bMinorChange)
{
	const bool bDeleting = (Tools_Mode_Container::GetDeletionCounter() > 0);

	// Rebuild the client and tools instances if necessary.	
	if (Tools_Container_System::ContainerSearch(s_uContainerToEdit, uContainerChanged))
	{
		if (bDeleting)
		{
			DestroyInstance();
			DestroyToolsInstance();
			RefreshInstanceOverlay(true);
		}
		else if (bMinorChange)
		{
			// Minor change, just make sure positions are up to date.
			if (m_pxMainToolsObject)
			{
				m_pxMainToolsObject->TeleportTo(s_xPosition, s_xOrientation);
			}
			if (m_pxInstance)
			{
				m_pxInstance->TeleportTo(s_xPosition, s_xOrientation);
			}
		}
		else
		{
			CreateInstance();
			CreateToolsInstance();
			RefreshInstanceOverlay(false);
		}
	}
}

//////////////////////////////
// RefreshContainerModified //
//////////////////////////////
// Refresh the instance when a container has been modified.
void Tools_Mode_Container::RefreshContainerModified(const Asura_Hash_ID uContainerChanged, bool bMinorChange)
{
	if (s_pxThis)
	{
		s_pxThis->Refresh(uContainerChanged, bMinorChange);
	}
}

///////////////////////////
// ChangedNumberOfInputs //
///////////////////////////
// Refresh the instance overlay when the number of collection inputs has been changed.
void Tools_Mode_Container::ChangedNumberOfInputs(const Asura_Hash_ID uContainerChanged, u_int uNumInputs)
{
	if (uContainerChanged == s_uContainerToEdit)
	{
		const bool bOpen = IsInstanceOverlayOpen();
		if (bOpen)
		{	
			DestroyInstanceOverlay();
		}

		DestroyAllInstances();
		
		s_xStateToPreview.SetNumInputs(uNumInputs);
		s_xStateToPreviewOld = s_xStateToPreview;
		if (s_pxThis)
		{
			s_pxThis->RefreshInstanceOverlay(true);
			s_pxThis->RefreshStateToPreview();
		}

		CreateAllInstances();

		if (bOpen)
		{
			// Recreate the instance overlay, to make sure that the preview state dialogue is updated.
			CreateInstanceOverlay();
		}
	}
}

///////////////////
// RefreshRename //
///////////////////
// Refresh the instance overlay when a container has been renamed.
void Tools_Mode_Container::RefreshRename(const Asura_Hash_ID uContainerRenamed)
{
	if (uContainerRenamed == s_uContainerToEdit)
	{
		if (IsInstanceOverlayOpen())
		{
			// Recreate the instance overlay, to make sure that the title is updated.
			CreateInstanceOverlay();
		}
	}
}

/////////////////////////
// RequestPreviewState //
/////////////////////////
// Change the preview state
void Tools_Mode_Container::RequestPreviewState(const Asura_Container_InputState& xInputState, Asura_Hash_ID uCollectionHash)
{
	if (s_xStateToPreview == xInputState) return;	// No change
	if (s_uContainerToEdit != uCollectionHash) return;	// Only root collection can be previewed.

	s_xStateToPreview = xInputState;
	s_xStateToPreviewOld = s_xStateToPreview;

	if (s_pxThis)
	{
		s_pxThis->Refresh(uCollectionHash, false);
	}
}

/////////////////////////////////////
// RequestPreviewStateFromDialogue //
/////////////////////////////////////
// Dialogue for "preview state" on instance overlay has been changed.
void Tools_Mode_Container::RequestPreviewStateFromDialogue()
{
	if (s_pxThis)
	{
		s_pxThis->Refresh(s_uContainerToEdit, false);

		// Refresh the collection edit overlay if needed.
		Tools_Container_Collection::ChangedPreviewStateFromDialogue(s_uContainerToEdit, s_xStateToPreview, s_xStateToPreviewOld);
		s_xStateToPreviewOld = s_xStateToPreview;
	}
}

///////////////////////
// MarkDeletionStart //
///////////////////////
// Mark beginning of deletion process.
void Tools_Mode_Container::MarkDeletionStart()
{
	// Deletion process started.
	++s_iDeletionCounter;
}

//////////////////////
// MarkDeletionStop //
//////////////////////
// Mark end of deletion process.
void Tools_Mode_Container::MarkDeletionStop()
{
	--s_iDeletionCounter;
	Asura_Assert(s_iDeletionCounter>=0, "Unbalanced deletion counter");

	if (s_iDeletionCounter == 0 && s_pxThis)
	{
		// Deletion finished. Recreate instances.
		if (!s_pxThis->m_pxInstance)
		{
			s_pxThis->CreateInstance();
		}

		if (!s_pxThis->m_pxMainToolsObject)
		{
			s_pxThis->CreateToolsInstance();
			s_pxThis->RefreshInstanceOverlay(false);
		}
	}

	if (s_iDeletionCounter == 0)
	{
		// Make sure entity is recreated.
		Tools_Entity_Container::RefreshContainerModified(ASURA_HASH_ID_UNSET, false);
	}
}

////////////////////////////////////
// UpdateInstanceOverlayStateInfo //
////////////////////////////////////
// Need to update the state info in the main instance overlay.
void Tools_Mode_Container::UpdateInstanceOverlayStateInfo()
{
	if (s_pxThis && s_pxThis->m_pxMainToolsObject)
	{
		s_pxThis->m_pxMainToolsObject->ChangeState(s_xStateToPreview);
		s_pxThis->s_xInstanceTree.TraverseUpdateStateInfo(s_xInstanceTree.GetRoot());
	}
}

/////////////////////////
// AddContextMenuItems //
/////////////////////////
void Tools_Mode_Container::AddContextMenu(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext)
{
	Tools_ContextMenu_Item* pxItem = 0;
	Tools_ContextMenu_Section* pxSection = 0;

	// Use the section menu for a tenuously named 'edit' for now
	pxSection = xMenu.GetSection(TOOLS_CONTEXTMENU_SECTION_SELECTION);
	pxSection->SetHeading("Edit menu");

	// Edit
	pxItem = new Tools_ContextMenu_Item_Function("Edit", ContextMenu_Edit);
	pxItem->SetEnabled(IsInstanceContainerSelectionValid());
	pxItem->SetHelpText("Edit this container.");
	pxSection->AddItem(pxItem);

	pxSection->AddSeparator();

	// Insert/remove attachment.
	pxItem = new Tools_ContextMenu_Item_Function("Insert attachment", ContextMenu_InsertAttachment);
	pxItem->SetEnabled(IsValidForInsertAttachment());
	pxSection->AddItem(pxItem);

	pxItem = new Tools_ContextMenu_Item_Function("Remove attachment", ContextMenu_RemoveAttachment);
	pxItem->SetEnabled(IsValidForRemoveAttachment());
	pxSection->AddItem(pxItem);

	pxSection->AddSeparator();

	// Rename
	pxItem = new Tools_ContextMenu_Item_Function("Rename", ContextMenu_Rename);
	pxSection->AddItem(pxItem);

	pxItem = new Tools_ContextMenu_Item_Function("Toggle show name", ContextMenu_ToggleShowName);
	pxItem->SetHelpText("Toggle whether it shows or hides the container name in the tree.");
	pxSection->AddItem(pxItem);

	pxSection->AddSeparator();

	// Remove/delete
	pxItem = new Tools_ContextMenu_Item_Function("Remove/Delete", ContextMenu_Delete);
	pxItem->SetEnabled(IsInstanceContainerSelectionValidForRemoval());
	pxItem->SetHelpText("Remove this container. Delete it, and any children, if they are no longer being used and have group-scope.");
	pxSection->AddItem(pxItem);

	pxSection->AddSeparator();

	// Cut/copy/paste
	pxItem = new Tools_ContextMenu_Item_Function("Cut", ContextMenu_Cut);
	pxItem->SetEnabled(IsValidForCutOrCopy());
	pxItem->SetHelpText("Cut this container to the clipboard. Remove it from this collection upon pasting.");
	pxSection->AddItem(pxItem);

	pxItem = new Tools_ContextMenu_Item_Function("Copy", ContextMenu_Copy);
	pxItem->SetEnabled(IsValidForCutOrCopy());
	pxItem->SetHelpText("Copy this container to the clipboard.");
	pxSection->AddItem(pxItem);


	pxItem = new Tools_ContextMenu_Item_Function("Paste", ContextMenu_Paste);
	pxItem->SetEnabled(IsValidForPaste());
	pxItem->SetHelpText("Paste the container from the clipboard to the specified collection.");
	pxSection->AddItem(pxItem);

}

void Tools_Mode_Container::ContextMenu_InsertAttachment()
{
	if (s_pxThis)
	{
		s_pxThis->InsertAttachment();
	}
}

void Tools_Mode_Container::ContextMenu_RemoveAttachment()
{
	if (s_pxThis)
	{
		s_pxThis->RemoveAttachment();
	}
}

void Tools_Mode_Container::ContextMenu_Edit()
{
	if (s_pxThis)
	{
		s_pxThis->EditChild();
	}
}

void Tools_Mode_Container::ContextMenu_Rename()
{
	Tools_Container* pxContainer = GetContainerSelected();
	Tools_Container_System::RenameContainerUsingDialogue(pxContainer);
}

void Tools_Mode_Container::ContextMenu_ToggleShowName()
{
	Tools_Container* pxContainer = GetContainerSelected();
	if (pxContainer)
	{
		pxContainer->ToggleShowName();

		// Mark file as being modified, so that the change will be saved.
		Tools_Container_System::SetFileModified(pxContainer->GetFileHash());

		if (s_pxThis)
		{
			s_pxThis->RefreshInstanceOverlay(false);
		}
	}
	
}

void Tools_Mode_Container::ContextMenu_Delete()
{
	if (s_pxThis)
	{
		s_pxThis->RemoveChild();
	}
}

void Tools_Mode_Container::ContextMenu_Cut()
{
	s_xClipboard.SetCut();
}

void Tools_Mode_Container::ContextMenu_Copy()
{
	s_xClipboard.SetCopy();
}

void Tools_Mode_Container::ContextMenu_Paste()
{
	s_xClipboard.DoPaste(GetContainerSelected());
}

/////////////////////////////////////////////////////////////////////
// Tools_Container_Object_Origin
/////////////////////////////////////////////////////////////////////
bool Tools_Container_Object_Origin::s_bWasHidden = false;

Tools_Container_Object_Origin::Tools_Container_Object_Origin(Asura_Hash_ID uContainerHash)
: Tools_PositionEdit_Object(&Tools_Mode_Container::s_xPosition, &Tools_Mode_Container::s_xOrientation, "Origin", 0 ,&m_bMoved)
, m_pxChild(0)
, m_uContainerHash(uContainerHash)
, m_bMoved(false)
{
	Type = TOT_Container_Object_Origin;
	SetMesh("ToolsShape_RedDiamond");
	SetScale(Tools_Container_System::GetOptions().GetModeOriginMeshSize());
	
	m_xInputState.Reset();
	m_xInputState = Asura_Container_InputState::On();

	// Check for circular references before creating the child.
	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
	if (pxContainer && !pxContainer->CheckIsTooDeep())
	{
		m_pxChild = Tools_Mode_Container::CreateChild(uContainerHash);
		if (m_pxChild)
		{
			m_pxChild->SetUniqueID(0);
			m_pxChild->PostCreateInitialise();
		}
	}

	// Restore hidden state.
	if (s_bWasHidden)
	{
		SetHiddenFlag(ObjectHideFlag_User, true);
	}
}

Tools_Container_Object_Origin::~Tools_Container_Object_Origin()
{
	Tools_Container_Object::ResetStoredHiddenFlag();

	if (m_pxChild)
	{
		m_pxChild->Destroy();
	}

	// Save hidden state for origin.
	s_bWasHidden = IsHiddenFlagSet(ObjectHideFlag_User);
}

void Tools_Container_Object_Origin::PrintInfo(float& fX, float& fY)
{
	Tools_Mode_Container::PrintInfo(fX, fY);
	Tools_Object::PrintInfo(fX, fY);
}

void Tools_Container_Object_Origin::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{
	PARENT::UpdatePositionInAsuraData(bUpdateLinkedObjects);
	TeleportTo(Position, Orientation);
}

void Tools_Container_Object_Origin::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri)
{
	// Make sure all the other tools objects are moved too.
	if (m_pxChild)
	{
		m_pxChild->TeleportTo(Position, Orientation);
	}
}

void Tools_Container_Object_Origin::OnSelection()
{
	PARENT::OnSelection();

	if (m_uContainerHash != ASURA_HASH_ID_UNSET)
	{
		Tools_Container_System::SelectContainer(m_uContainerHash, false);
		Tools_Mode_Container::SelectInstanceContainer(m_uContainerHash);
	}
}

void Tools_Container_Object_Origin::AddToInstanceOverlayTree(Asura_TreeDialog_Item* pxParent)
{
	if (!pxParent) return;

	if (m_pxChild)
	{
		m_pxChild->AddToInstanceOverlayTree(pxParent);
	}
	else
	{
		Asura_TreeDialog_Item* const pxItem = Tools_Mode_Container::InsertItemInTree(pxParent, "*Not Found*", m_uContainerHash, 0);
		if (pxItem)
		{
			Asura_String256 xName("0x%08x", m_uContainerHash);
			pxItem->SetHelpText(xName.GetString());
			pxItem->Expand(true);
		}
	}
}

Tools_Container_Object* Tools_Container_Object_Origin::FindContainer(Asura_Hash_ID uContainerHash)
{
	return m_pxChild ? m_pxChild->FindContainer(uContainerHash) : 0;
}

void Tools_Container_Object_Origin::ChangeState(const Asura_Container_InputState& xInputState)
{
	m_xInputState = xInputState;
	if (m_pxChild)
	{
		m_pxChild->ChangeState(m_xInputState);
	}
}


/////////////////////////////////////////////////////////////////////
// Tools_Container_Object
/////////////////////////////////////////////////////////////////////

// Static data
u_int Tools_Container_Object::s_auHiddenObjectID[MAX_NUM_HIDDEN_STORED];
Asura_Hash_ID Tools_Container_Object::s_auHiddenObjectHash[MAX_NUM_HIDDEN_STORED];

Tools_Container_Object::Tools_Container_Object(Asura_Hash_ID uContainerHash)
: Tools_Object(TOT_Container_Object_Misc)
, m_szDescription(0)
, m_uContainerHash(uContainerHash)
, m_uClassification(AsuraContainerClass_NotSet)
, m_uUniqueID(0xffffffff)
, m_uMovedByUserFrameNumber(0)
, m_pxParent(0)
{
	m_xInputState.Reset();
	m_xInputState = Asura_Container_InputState::On();

	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
	if (pxContainer)
	{
		m_szDescription = pxContainer->GetDescription();
		m_xName = pxContainer->GetContainerName();
		m_uClassification = pxContainer->GetClassification();
	}
	else if (uContainerHash == ASURA_HASH_ID_UNSET)
	{
		m_szDescription = "*Not set*";
		m_xName = "*None*";
	}
	else
	{
		m_szDescription = "*Unknown Type*";
		m_xName = "*Not Found*";
	}

	// Create the children.
	bool bAddedChild = false;
	for ( Tools_Container_ChildIterator xIt(pxContainer); !xIt.Done(); xIt.Next() )
	{
		const Asura_Hash_ID uHash = xIt.GetCurrentHash();
		Tools_Container_Object* pxChild = Tools_Mode_Container::CreateChild(uHash);
		if (pxChild)
		{
			pxChild->SetParent(this);
			m_xChildren.AddToList(pxChild);
			bAddedChild = true;
		}
	}

	// If this is an attachment with no children then add a placeholder "none" object as a child.
	if (!bAddedChild && m_uClassification == AsuraContainerClass_Attachment)
	{
		Tools_Container_Object* pxChild = new Tools_Container_Object_Misc(ASURA_HASH_ID_UNSET);
		if (pxChild)
		{
			pxChild->SetParent(this);
			m_xChildren.AddToList(pxChild);
		}
	}
}

Tools_Container_Object::~Tools_Container_Object()
{
	m_pxParent = 0;
	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren) ; !xIt.Done(); xIt.Next() )
	{
		Tools_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->Destroy();
		}
	}

	// Save the user-hidden flag.
	const bool bHiddenByObject = IsHiddenFlagSet(ObjectHideFlag_User);
	if (bHiddenByObject)
	{
		SetWasHidden(GetUniqueID(), GetContainerHash());
	}
}

void Tools_Container_Object::PostCreateInitialise()
{
	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren) ; !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->PostCreateInitialise();
		}
	}

	// Restore the user-hidden flag.
	const bool bWasHidden = WasHidden(GetUniqueID(), GetContainerHash());
	if (bWasHidden)
	{
		SetHiddenFlag(ObjectHideFlag_User, true);
	}
}

void Tools_Container_Object::ResetStoredHiddenFlag()
{
	for (int i=0; i<MAX_NUM_HIDDEN_STORED; ++i)
	{
		s_auHiddenObjectID[i] = 0;
		s_auHiddenObjectHash[i] = ASURA_HASH_ID_UNSET;
	}
}

// Store data about user-hidden objects, so that we can restore the flags upon rebuilding the objects.
void Tools_Container_Object::SetWasHidden(u_int uUniqueID, Asura_Hash_ID uContainerHash)
{
	for (int i=0; i<MAX_NUM_HIDDEN_STORED; ++i)
	{
		if (s_auHiddenObjectHash[i] == ASURA_HASH_ID_UNSET)
		{
			s_auHiddenObjectID[i]   = uUniqueID;
			s_auHiddenObjectHash[i] = uContainerHash;
			break;
		}
	}
}

bool Tools_Container_Object::WasHidden(u_int uUniqueID, Asura_Hash_ID uContainerHash)
{
	for (int i=0; i<MAX_NUM_HIDDEN_STORED; ++i)
	{
		if (s_auHiddenObjectHash[i] == ASURA_HASH_ID_UNSET)
		{
			break; // Not found.
		}

		if (s_auHiddenObjectHash[i] == uContainerHash &&
			s_auHiddenObjectID[i] == uUniqueID)
		{
			return true;
		}
	}
	return false;
}


void Tools_Container_Object::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri)
{
	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren) ; !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->TeleportTo(xPosition, xOri);
		}
	}

	if (m_uMovedByUserFrameNumber != Asura_Timers::GetFrameNumber())
	{
		Position = xPosition;
		Orientation = xOri;
	}
}

void Tools_Container_Object::PrintInfo(float& fX, float& fY)
{
	Tools_Mode_Container::PrintInfo(fX, fY);
	PARENT::PrintInfo(fX, fY);

	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_uContainerHash);
	if (pxContainer)
	{
		const bool bGroupScope = pxContainer->IsGroupScope();
		Asura_Text_Output::Print(fX, fY++, 0xffffffff, 0, "Scope: %s", (bGroupScope ? "Group" : "Global"));
	}
}

void Tools_Container_Object::OnSelection()
{
	PARENT::OnSelection();

	if (m_uContainerHash != ASURA_HASH_ID_UNSET)
	{
		Tools_Container_System::SelectContainer(m_uContainerHash, false);
		Tools_Mode_Container::SelectInstanceContainer(m_uContainerHash);
	}
}

bool Tools_Container_Object::IsParentAnAttachment() const
{
	return m_pxParent && m_pxParent->GetClassification() == AsuraContainerClass_Attachment;
}

Tools_Container_Object* Tools_Container_Object::FindContainer(Asura_Hash_ID uContainerHash)
{
	if (m_uContainerHash == uContainerHash)
	{
		return this;
	}

	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren) ; !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			Tools_Container_Object* pxResult = pxCurrent->FindContainer(uContainerHash);
			if (pxResult)
			{
				return pxResult;
			}
		}
	}

	return 0;
}

const class Asura_ClientContainerInstance* Tools_Container_Object::GetClientInstance() const
{
	if (m_uClassification == AsuraContainerClass_Attachment) return 0;

	Asura_ClientContainerInstance* pxClientInstanceRoot = Tools_Mode_Container::GetClientInstance();
	if (pxClientInstanceRoot)
	{
		Asura_ClientContainerInstance* pxFound = pxClientInstanceRoot->FindFromUniqueID(GetUniqueID());
		if (pxFound)
		{
			if (pxFound->GetContainerHash() == m_uContainerHash)
			{
				return pxFound;
			}
		}
	}

	return 0;
}

//////////////////////////////
// AddToInstanceOverlayTree //
//////////////////////////////
// Add container to the instance tree below the specified parent.
void Tools_Container_Object::AddToInstanceOverlayTree(Asura_TreeDialog_Item* pxParent)
{
	if (m_uContainerHash == ASURA_HASH_ID_UNSET)
	{
		Asura_TreeDialog_Item* const pxChildItem = Tools_Mode_Container::InsertItemInTree(pxParent, "*None*", m_uContainerHash, this);
		if (pxChildItem)
		{
			Asura_String256 xName("Not set");
			pxChildItem->SetHelpText(xName.GetString());
		}
		return;
	}

	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_uContainerHash);
	if (!pxContainer)
	{
		Asura_TreeDialog_Item* const pxItem = Tools_Mode_Container::InsertItemInTree(pxParent, "*Not Found*", m_uContainerHash, this);
		if (pxItem)
		{
			Asura_String256 xName("0x%08x", m_uContainerHash);
			pxItem->SetHelpText(xName.GetString());
		}
		return;
	}

	Asura_TreeDialog_Item* pxItem = 0;
	
	if (s_bIconForAttachments && m_uClassification == AsuraContainerClass_Attachment)
	{
		// Skip the attachment, which will be an icon.
		pxItem = pxParent;
	}
	else
	{
		// Build up a label.
		Asura_String256 xLabel;
		pxContainer->AddInstanceName(xLabel, false, true);

		pxItem = Tools_Mode_Container::InsertItemInTree(pxParent, xLabel.GetString(), m_uContainerHash, this);
		if (!pxItem) return;

		Asura_String xName;
		pxContainer->MakeFullName(xName);
		pxItem->SetHelpText(xName.GetString());
	}


	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren); !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->AddToInstanceOverlayTree(pxItem);
		}
	}

	// Expand everything initially. Later, after creating the tree,
	// we'll collapse stuff using the stored "collapse data".
	if (pxItem)
	{
		pxItem->Expand(true);
	}
}

void Tools_Container_Object::ChangeState(const Asura_Container_InputState& xInputState)
{
	m_xInputState = xInputState;
}


/////////////////////////////////////////////////////////////////////
// Tools_Container_Object_Misc
/////////////////////////////////////////////////////////////////////
Tools_Container_Object_Misc::Tools_Container_Object_Misc(Asura_Hash_ID uContainerHash)
: Tools_Container_Object(uContainerHash)
{
	Type = TOT_Container_Object_Misc;
	SetMesh("ToolsShape_CutsceneTrackPoint");
}

///////////////////////////////
// UpdatePositionInAsuraData //
///////////////////////////////
void Tools_Container_Object_Misc::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{
	PARENT::UpdatePositionInAsuraData(bUpdateLinkedObjects);

	if (m_pxParent && m_pxParent->Type == TOT_Container_Object_Attachment)
	{
		m_pxParent->Position = Position;
		m_pxParent->Orientation = Orientation;
		m_pxParent->UpdatePositionInAsuraData(bUpdateLinkedObjects);
	}
}

bool Tools_Container_Object_Misc::ShouldTestForLOS()
{
	// Ignore objects that aren't being rendered, except for meshes which are rendered elsewhere.
	if (m_uClassification != AsuraContainerClass_Mesh)
	{
		if (IsHidden()) return false;
	}

	const Asura_ClientContainerInstance* pxFound = GetClientInstance();
	if (pxFound && pxFound->GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED))
	{
		// Mesh was hidden by the user.
		return false;
	}

	if (!Tools_Render_Options::IsPointWithinViewDistance(Position)) return false;

	return true;
}


bool Tools_Container_Object_Misc::ObjectIsInLOS(const Asura_Vector_3& xOrigin, const Asura_Vector_3& xDir, float& fLowest)
{
	Asura_ClientContainerInstance* pxInstance = Tools_Mode_Container::GetClientInstance();
	if (pxInstance)
	{
		Asura_ClientContainerInstance* pxFound = pxInstance->FindFromUniqueID(m_uUniqueID);
		if (pxFound)
		{
			if (pxFound->GetContainerHash() != m_uContainerHash)
			{
				Asura_Assert(0, "Found wrong container instance");
				return false;
			}

			if (pxFound->GetClassification() != AsuraContainerClass_Mesh)
			{
				return false;
			}

			Asura_ClientContainerInstance_Mesh* pxMesh = static_cast<Asura_ClientContainerInstance_Mesh*>(pxFound);
			Asura_Animation* pxAnimation = pxMesh->GetAnimation();
			
			if (pxAnimation)
			{
				const float fRadius = pxAnimation->GetCoarseBoundingSphereRadius();
				const Asura_Vector_3& xCentre = pxAnimation->GetBoundingSphereCentre();
				
				if (Asura_Maths::RayIntersectsSphere(xOrigin, xDir, xCentre, fRadius))
				{
					Asura_LOS_InputData xLOSInput(xOrigin, xDir, ASURA_GUID_UNREGISTERED, fLowest);
					xLOSInput.m_uPolyFlagsToIgnore = 0; //(default would ignore polygons flagged ignore-for-los... which would be bad)
						
					Asura_LOS_OutputData xLOSOutput;
					xLOSOutput.m_fDistanceToIntersection = fLowest;

					Asura_Matrix_3x3 xOriMat;
					const Asura_Quat& xQuat = pxAnimation->GetOrientQuat();
					xQuat.ToMatrix(xOriMat);

					if (pxAnimation->LOSTest(pxAnimation->GetPosition(), xOriMat, &xLOSInput, &xLOSOutput))
					{
						// A hit, write back the distance.
						fLowest = xLOSOutput.m_fDistanceToIntersection;
						return true;
					}
				}
			}
		}
	}

	return false;
}

////////////
// Update //
////////////
void Tools_Container_Object_Misc::Update()
{
	PARENT::Update();
	UpdateMeshHidden();
}

///////////////////
// SetMeshHidden //
///////////////////
// Allow the user to hide some of the mesh container instances.
void Tools_Container_Object_Misc::UpdateMeshHidden()
{
	if (m_uClassification != AsuraContainerClass_Mesh)
	{
		return;
	}

	Asura_ClientContainerInstance* pxClientInstanceRoot = Tools_Mode_Container::GetClientInstance();
	if (!pxClientInstanceRoot) return;
	
	Asura_ClientContainerInstance* pxFound = pxClientInstanceRoot->FindFromUniqueID(GetUniqueID());
	if (!pxFound) return;
	if (pxFound->GetContainerHash() != m_uContainerHash) return;
	if (pxFound->GetClassification() != AsuraContainerClass_Mesh ) return;

	Asura_ClientContainerInstance_Mesh* pxMeshInstance = static_cast<Asura_ClientContainerInstance_Mesh*>(pxFound);

	const bool bHiddenByObject = IsHiddenFlagSet(ObjectHideFlag_User);
	const bool bHiddenByType = Tools_Object_System::GetTypeHiddenFlag(Type, ObjectHideFlag_HideByTypeOverlay);

	if (bHiddenByObject || bHiddenByType)
	{
		pxMeshInstance->SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED);
	}
	else
	{
		pxMeshInstance->ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED);
	}
}

/////////////////////////////////////////////////////////////////////
// Tools_Container_Object_Attachment
/////////////////////////////////////////////////////////////////////
Tools_Container_Object_Attachment::Tools_Container_Object_Attachment(Asura_Hash_ID uContainerHash)
: Tools_Container_Object(uContainerHash)
, m_uChildHash(ASURA_HASH_ID_UNSET)
, m_xParentPos(Asura_Maths::ZeroVector3)
, m_xParentOri(Asura_Maths::IdentityMatrix3x3)
{
	Type = TOT_Container_Object_Attachment;
	SetMesh("ToolsShape_CutsceneTrackPoint");
	SetScale(Tools_Container_System::GetOptions().GetModeAttachmentMeshSize());

	const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
	if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Attachment)
	{
		const Tools_Container_Attachment* pxAttachment = static_cast<const Tools_Container_Attachment*>(pxContainer);
		m_uChildHash = pxAttachment->GetChildContainerHash();
	}
}

Tools_Container_Object_Attachment::~Tools_Container_Object_Attachment()
{
}

void Tools_Container_Object_Attachment::PrintInfo(float& fX, float& fY)
{
	PARENT::PrintInfo(fX, fY);

	const Tools_Container* pxChildContainer = Tools_Container_System::FindContainerByHash(m_uChildHash);
	if (pxChildContainer)
	{
		Asura_String xName;
		pxChildContainer->MakeFullName(xName);
		//const Asura_String& xName = pxChildContainer->GetContainerName();
		fY++;
		Asura_Text_Output::Print(fX, fY++, 0xffffffff, 0, "Attached: %s  (%s)", xName.GetString(), pxChildContainer->GetDescription());
	}
}

///////////////////////////////
// UpdatePositionInAsuraData //
///////////////////////////////
void Tools_Container_Object_Attachment::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{
	PARENT::UpdatePositionInAsuraData(bUpdateLinkedObjects);

	// The tools object has been moved, update the attachment offset.
	Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_uContainerHash);
	if (!pxContainer || pxContainer->GetClassification() != AsuraContainerClass_Attachment)
	{
		Asura_Assert(0, "Failed to find attachment container");
		return;
	}

	Tools_Container_Attachment* pxAttachment = static_cast<Tools_Container_Attachment*>(pxContainer);
	
	// Find the attachment position and quaternion, which will transform from
	// m_xParentPos, m_xParentOri to Position, Orientation.
	// TODO this doesn't take account of any marker points.

	Asura_Matrix_3x3 xOriTemp = m_xParentOri;
	xOriTemp.Transpose();
	xOriTemp = xOriTemp * Orientation;
	const Asura_Quat xAttachQuat(xOriTemp);

	const Asura_Vector_3 xPosTemp = Position - m_xParentPos;
	const Asura_Vector_3 xAttachPosition = m_xParentOri.ApplyTranspose(xPosTemp);
	
	m_uMovedByUserFrameNumber = Asura_Timers::GetFrameNumber();

	pxAttachment->SetPosOriFromMode(xAttachPosition, xAttachQuat);

	// Make sure that the child is moved too.
	TeleportTo(m_xParentPos, m_xParentOri);
}

////////////////
// TeleportTo //
////////////////
void Tools_Container_Object_Attachment::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri)
{
	m_xParentPos = xPosition;
	m_xParentOri = xOri;

	// Apply offset.
	const Asura_Container* pxContainer = Asura_Container_System::FindContainer(m_uContainerHash);
	if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Attachment)
	{
		// TODO this doesn't take account of any marker points.
		const Asura_Container_Attachment* pxAttachment = static_cast<const Asura_Container_Attachment*>(pxContainer);
		Asura_Vector_3 xNewPos;
		Asura_Matrix_3x3 xNewOri;
		pxAttachment->ApplyOffset(xPosition, xOri, xNewPos, xNewOri, 0);

		// Move to attachment position.
		PARENT::TeleportTo(xNewPos, xNewOri);
	}
	else
	{
		Asura_Assert(0, "Failed to find attachment container");
	}	
}

/////////////////
// SetUniqueID //
/////////////////
void Tools_Container_Object_Attachment::SetUniqueID(u_int uUniqueID)
{
	m_uUniqueID = uUniqueID;

	// There should be at most a single child and it has the same ID as the attachment.
	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren); !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->SetUniqueID(uUniqueID);
		}
	}
}

void Tools_Container_Object_Attachment::ChangeState(const Asura_Container_InputState& xInputState)
{
	m_xInputState = xInputState;
	
	// Just pass the state on to any child.
	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren) ; !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			pxCurrent->ChangeState(xInputState);
		}
	}
}


/////////////////////////////////////////////////////////////////////
// Tools_Container_Object_Collection
/////////////////////////////////////////////////////////////////////
Tools_Container_Object_Collection::Tools_Container_Object_Collection(Asura_Hash_ID uContainerHash)
: Tools_Container_Object(uContainerHash)
{
	Type = TOT_Container_Object_Collection;
	SetMesh("ToolsShape_Diamond");
}

/////////////////
// SetUniqueID //
/////////////////
void Tools_Container_Object_Collection::SetUniqueID(u_int uUniqueID)
{
	m_uUniqueID = uUniqueID;

	const Asura_Container* pxAsuraContainer = Asura_Container_System::FindContainer(m_uContainerHash);
	const Asura_Container_Collection* pxAsuraCollection = 0;

	if (pxAsuraContainer && pxAsuraContainer->GetClassification() == AsuraContainerClass_Collection)
	{
		pxAsuraCollection = static_cast<const Asura_Container_Collection*>(pxAsuraContainer);
	}

	Asura_Assert(pxAsuraCollection, "Couldn't find the asura collection");
	if (!pxAsuraCollection) return;


	int iMultiplicity = 0;
	Asura_Hash_ID uPreviousHash = ASURA_HASH_ID_UNSET;

	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren); !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxCurrent = xIt.GetCurrent();
		if (pxCurrent)
		{
			const Asura_Hash_ID uChildHash = pxCurrent->GetContainerHash();
			if (uPreviousHash == uChildHash)
			{
				++iMultiplicity;
			}
			else
			{
				iMultiplicity = 0;
			}
			uPreviousHash = uChildHash;

			const int iIndex = pxAsuraCollection->FindChildIndex(uChildHash, iMultiplicity);
			if (iIndex<0)
			{
				Asura_Assert(0, "Failed to set unique ID on child");
				continue;
			}

			const u_int uChildID = (uUniqueID<<8) + iIndex + 1;
			pxCurrent->SetUniqueID(uChildID);
		}
	}
}

// Work out the state of all the children.
void Tools_Container_Object_Collection::ChangeState(const Asura_Container_InputState& xInputState)
{
	m_xInputState = xInputState;

	Asura_Container_OutputState xOutputState;

	const Asura_Container* const pxContainer = Asura_Container_System::FindContainer(m_uContainerHash);
	Asura_Assertf(pxContainer, ("Container (0x%08x) not found", m_uContainerHash));
	if (pxContainer)
	{
		if (pxContainer->GetClassification() == AsuraContainerClass_Collection)
		{
			const Asura_Container_Collection* const pxCollection = static_cast<const Asura_Container_Collection*>(pxContainer);
			pxCollection->ConvertState(xInputState, 100.0f, xOutputState);
		}
	}

	for (Asura_List_It<Tools_Container_Object> xIt(m_xChildren) ; !xIt.Done(); xIt.Next() )
	{
		Tools_Container_Object* pxChild = xIt.GetCurrent();
		if (!pxChild) continue;

		const Asura_Hash_ID uChildContainerHash = pxChild->GetContainerHash();
		
		Asura_Container_InputState xChildInputState;
		xOutputState.GetChildInputState(uChildContainerHash, xChildInputState);

		pxChild->ChangeState(xChildInputState);
	}
}

///////////////////////////////
// UpdatePositionInAsuraData //
///////////////////////////////
void Tools_Container_Object_Collection::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{
	PARENT::UpdatePositionInAsuraData(bUpdateLinkedObjects);

	if (m_pxParent && m_pxParent->Type == TOT_Container_Object_Attachment)
	{
		m_pxParent->Position = Position;
		m_pxParent->Orientation = Orientation;
		m_pxParent->UpdatePositionInAsuraData(bUpdateLinkedObjects);
	}
}


#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

