///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Entity_ConsoleVar.cpp
//
// Created by Martin G.Harris (20/6/2007)
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

// Need command console for the console variables.
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#include "Core/Asura.h"
#include "Entity/Intangible/Tools_Entity_ConsoleVar.h"

// Asura
#include "Camera/Asura_Camera.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Entity/Intangible/Asura_ClientEntity_ConsoleVar.h"
#include "Entity/Intangible/Asura_ServerEntity_ConsoleVar.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Text_Output.h"

// Tools
#include "Interface/Asura_CommandConsole.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Objects/Tools_Object_System.h"

const u_int uBUFFER_LENGTH = 256;


const int iCONSOLE_VAR_NUM_TYPES = 4; // Including invalid.

// List of types.
const char* g_apszTypeList[] =
{
	"*Invalid*",	// TYPE_INVALID = -1
	"Boolean",		// TYPE_BOOL  = 0
	"Integer",		// TYPE_INT	  = 1
	"Float",		// TYPE_FLOAT = 2
};


/////////////////////////////
// Tools_Entity_ConsoleVar //
/////////////////////////////
Tools_Entity_ConsoleVar::Tools_Entity_ConsoleVar(Asura_Entity* pxEntity)
: Tools_Entity_Object(pxEntity)
, m_iTypeFromList(0)
, m_bTypeChangedFlag(false)
{
	Type = TOT_Entity_ConsoleVar;
	
	Mesh = Asura_Mesh_System::GetMesh("ToolsShape_ConsoleVariable");

	if (pxEntity)
	{
		CopyVarNameFromClient();
		CopyTypeFromClient();
	}
}

Tools_Entity_ConsoleVar::~Tools_Entity_ConsoleVar()
{
}

///////////////////////////
// CopyVarNameFromClient //
///////////////////////////
void Tools_Entity_ConsoleVar::CopyVarNameFromClient()
{
	const Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());
	Asura_Assert(pxClient, "Null client entity");
	if (pxClient)
	{
		// Copy string from client.
		const Asura_Char* szVarName = pxClient->GetVarName();
		if (szVarName)
		{
			m_xVarName = szVarName;
		}
	}
}

////////////////////////
// CopyTypeFromClient //
////////////////////////
void Tools_Entity_ConsoleVar::CopyTypeFromClient()
{
	const Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());
	Asura_Assert(pxClient, "Null client entity");
	if (pxClient && pxClient->m_uVarNameHash != ASURA_HASH_ID_UNSET)
	{
		m_iTypeFromList = (static_cast<int>(pxClient->GetType()) + 1);
		if (m_iTypeFromList<0 || m_iTypeFromList>= iCONSOLE_VAR_NUM_TYPES)
		{
			m_iTypeFromList = 0; // Corresponds to invalid.
		}
	}
}

////////////
// Create //
////////////
void Tools_Entity_ConsoleVar::Create()
{
	// Get a new GUID
	const Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();

	// Set up tools data and auto-create a unique name (prompt for new name if there is a clash).
	Tools_Entity_Data* pxToolsData = CreateToolsDataWithDefaultName("Console Variable", "ConsoleVar", uGuid, GetNamePrefix());

	if (pxToolsData == NULL)
	{
		Asura_Assert(0, "Tools_Entity_ConsoleVar::Create() - no tools data");
		return;
	}

	Asura_ServerEntity_ConsoleVar* pxServerObject = new Asura_ServerEntity_ConsoleVar(uGuid);
	if (pxServerObject == NULL)
	{
		Asura_Assert(0, "Tools_Entity_ConsoleVar::Create() - no server object" );
		return;
	}

	pxServerObject->ToolsData = pxToolsData;

	Asura_ClientEntity_ConsoleVar* pxClientObject = new Asura_ClientEntity_ConsoleVar(uGuid);
	Asura_Assert(pxClientObject, "Tools_Entity_ConsoleVar::Create() - no client object");

	// Create the game specific tools object and select it.
	Tools_Entity_ConsoleVar* pxToolsObject = new Tools_Entity_ConsoleVar(pxServerObject);

	if (pxToolsObject == NULL)
	{
		Asura_Assert(0, "Tools_Entity_ConsoleVar::Create() - no tools object");
		return;
	}

	Tools_Object_System::SelectObject(pxToolsObject);

	pxToolsObject->Position		= *Asura_Camera::GetPosition();

//	Asura_Euler_3 xEuler(Asura_Maths::Deg90, 0.0f, 0.0f);
//	xEuler.ToMatrix(pxToolsObject->Orientation);
	pxToolsObject->Orientation	= Asura_Maths::IdentityMatrix3x3;

	Tools_Timed_Message_System::AddMessage(0xffffff, "Created new Console Variable Entity");
	Tools_SaveAsura::SetEntityModified();
}

////////////////////////
// AddOverlayControls //
////////////////////////
void Tools_Entity_ConsoleVar::AddOverlayControls(Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset)
{
	PARENT::AddOverlayControls(pxNewOverlay, xOffset);

	xOffset.y+=Asura_Dialog::DefaultSpacing();

	Asura_Dialog* pxDialog = new Asura_Function_Dialog("Change Variable name", xOffset, ChangeVarName);
	pxNewOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Select entity to track", xOffset, SelectEntityToTrack);
	pxDialog->SetHelpText("Select an entity, whose internal data will be copied to the console variable, whenever it changes.");
	pxNewOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	Asura_ServerEntity_ConsoleVar* pxServer = Asura_ServerEntity_ConsoleVar::FindEntity(GetGuid());
	if (pxServer)
	{
		pxDialog = new Asura_FlagTickbox_Dialog("Count time upwards", xOffset, &pxServer->m_uFlags, Asura_ServerEntity_ConsoleVar::FLAG_TIME_COUNT_UP, &Changed);
		pxDialog->SetHelpText("When tracking time triggers, use time elapsed (rather than time remaining).");
		pxNewOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();
	}

	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_DropDown_List_Dialog(xOffset, g_apszTypeList, &m_iTypeFromList, iCONSOLE_VAR_NUM_TYPES, iCONSOLE_VAR_NUM_TYPES, 0, &m_bTypeChangedFlag, "Type" );
	pxNewOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	xOffset.y+=Asura_Dialog::DefaultSpacing();
}

////////////
// Update //
////////////
void Tools_Entity_ConsoleVar::Update()
{
	PARENT::Update();

	if (m_bTypeChangedFlag)
	{
		// We have changed the type in the dialog.
		m_bTypeChangedFlag = false;

		// If they selected invalid then cancel the change.
		if (m_iTypeFromList == 0)
		{
			// Cancel the change
			CopyTypeFromClient();
			return;
		}

		bool bAllowed = true;

		const Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());
		Asura_Assert(pxClient, "Null client entity");
		if (pxClient && pxClient->m_uVarNameHash != ASURA_HASH_ID_UNSET)
		{
			const Asura_Hash_ID uVarNameHash = pxClient->m_uVarNameHash;

			int iNumber = 0;
			const bool bOwned = Asura_ClientEntity_ConsoleVar::GetNumberOfReferences(uVarNameHash, iNumber);

			if (!bOwned)
			{
				Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Console variable is defined in code, can't change the type!");
				
				// Cancel change
				bAllowed = false;
			}
			else if (iNumber > 1)
			{
				// Change will affect multiple entities.
				Asura_String xTitle( "This change will affect more than one ConsoleVar entity. Make the change?" );

				YesNo_Overlay xYesNoOverlay( xTitle );
				if (!xYesNoOverlay.DoModal())
				{
					// Cancel change
					bAllowed = false;
				}
			}
			
			if (bAllowed)
			{
				// Modify the entities and console variable.
				const Asura_ClientEntity_ConsoleVar::TYPE eNewType = static_cast<Asura_ClientEntity_ConsoleVar::TYPE>(m_iTypeFromList-1);
				Asura_ClientEntity_ConsoleVar::ChangeType(uVarNameHash, eNewType);

				SetEntityModified();
			}
			else
			{
				// Cancel the change
				CopyTypeFromClient();
			}
		}
	}
	else
	{
		// Copy back from client, is case the type has been changed.
		CopyTypeFromClient();
	}
}

///////////////////
// ChangeVarName //
///////////////////
void Tools_Entity_ConsoleVar::ChangeVarName()
{
	if (!ThisEntity) return;

	Tools_Entity_ConsoleVar* pxThis = static_cast<Tools_Entity_ConsoleVar*>(ThisEntity);
	const Asura_Guid uGuid = pxThis->GetGuid();

	// Make sure that this is in valid range.
	if (pxThis->m_iTypeFromList<0 || pxThis->m_iTypeFromList>= iCONSOLE_VAR_NUM_TYPES)
	{
		pxThis->m_iTypeFromList = 0; // Corresponds to invalid.
	}

	Asura_ClientEntity_ConsoleVar::TYPE eType = static_cast<Asura_ClientEntity_ConsoleVar::TYPE>(pxThis->m_iTypeFromList-1);
	if (eType == Asura_ClientEntity_ConsoleVar::TYPE_INVALID)
	{
		eType = Asura_ClientEntity_ConsoleVar::TYPE_INT;
	}

	GetString_Overlay xOverlay("Console Variable name:", pxThis->m_xVarName.GetString(), uBUFFER_LENGTH);
	if (xOverlay.DoModal())
	{
		Asura_String xNewName = xOverlay.GetString();
		if (!xNewName.Contains("."))
		{
			xNewName = "Unclassified.";

			// The DVS system renames variables which don't contain a '.', so we better do the same.
			if (xOverlay.GetString() && *xOverlay.GetString() != 0)
			{
				xNewName += xOverlay.GetString();
			}
			else
			{
				xNewName += "Null";
			}
		}

		pxThis->m_xVarName = xNewName;

		ThisEntity->SetEntityFileModified();

		// Change name on client.
		Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(uGuid);
		Asura_Assert(pxClient, "Couldn't find client");
		if (pxClient)
		{
			Asura_CommandConsole_VarRepository::Var* pxVar = Asura_CommandConsole_VarRepository::Find(xNewName.GetString());
			if (pxVar)
			{
				// If console variable already exists (eg was added in code) then try to copy the type.
				eType = Asura_ClientEntity_ConsoleVar::ConvertType(pxVar->m_eType);
			}

			pxClient->ChangeVarName(xNewName.GetString(), eType);
		}

		// Fix up server type too.
		Asura_ServerEntity_ConsoleVar* pxServer = Asura_ServerEntity_ConsoleVar::FindEntity(uGuid);
		Asura_Assert(pxServer, "Couldn't find server");
		if (pxServer)
		{
			pxServer->m_eType = eType;
		}

		// Copy type back to list dialog.
		if (pxThis)
		{
			pxThis->CopyTypeFromClient();
		}

		// Make sure that we save.
		ThisEntity->SetEntityModified();
	}
}

/////////////////////////
// SelectEntityToTrack //
/////////////////////////
void Tools_Entity_ConsoleVar::SelectEntityToTrack()
{
	if (!ThisEntity) return;
	
	Asura_ServerEntity_ConsoleVar* pxServer = Asura_ServerEntity_ConsoleVar::FindEntity(ThisEntity->GetGuid());

	Tools_Entity_ConsoleVar_ChooseEntityToTrack* pxChooseObject =
		new Tools_Entity_ConsoleVar_ChooseEntityToTrack( &pxServer->m_uEntityToTrackGuid);
	
	if ( pxChooseObject )
	{
		Tools_Object_System::ChooseObject( "Choose an entity to track", pxChooseObject );

		// Make sure that we save.
		ThisEntity->SetEntityModified();
	}
}

////////////////////////
// HandlesThisMessage //
////////////////////////
bool Tools_Entity_ConsoleVar::HandlesThisMessage(int iMessageID)
{
	if (iMessageID == ASURA_MSG_CHANGE_CONSOLEVAR)
	{
		return true;
	}

	return PARENT::HandlesThisMessage(iMessageID);
}

///////////////
// PrintInfo //
///////////////
void Tools_Entity_ConsoleVar::PrintInfo(float& fX, float& fY)
{
	const u_int uCOLOUR_WHITE = 0xffffffff;

	PARENT::PrintInfo(fX, fY);

	fY++;
	Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Name: %s", m_xVarName.GetString());
	
	Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());
	if (pxClient)
	{
		const Asura_CommandConsole_VarRepository::Var* pxVar = pxClient->GetVar();
		if (pxVar)
		{
			// Copy data to this entity.
			const Asura_CommandConsole_VarRepository::Var::Data& xData = pxVar->m_xData;
			switch (pxClient->m_eType)
			{
			case Asura_ClientEntity_ConsoleVar::TYPE_BOOL:
			{
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Type: Boolean");				
				const bool bData = (xData.pbData ? *xData.pbData : false);
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Value: %s", bData ? "true" : "false");
				break;
			}

			case Asura_ClientEntity_ConsoleVar::TYPE_INT:
			{
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Type: Integer");
				const int iData = (xData.piData ? *xData.piData : 0);
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Value: %d", iData);
				break;
			}

			case Asura_ClientEntity_ConsoleVar::TYPE_FLOAT:
			{
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Type: Float");
				const float fData = (xData.pfData ? *xData.pfData : 0.0f);
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Value: %f", fData);
				break;
			}

			case Asura_ClientEntity_ConsoleVar::TYPE_INVALID:
			{
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable Type: *INVALID*");
				break;
			}

			default:
				Asura_Assert(0, "Unknown Console Variable type!");
				break;
			}
		}

		const Asura_Hash_ID uVarNameHash = pxClient->m_uVarNameHash;
		if (uVarNameHash != ASURA_HASH_ID_UNSET)
		{
			int iNumber = 0;
			const bool bOwned = Asura_ClientEntity_ConsoleVar::GetNumberOfReferences(uVarNameHash, iNumber);

			if (iNumber>0 || !bOwned)
			{
				fY++;
			}
			if (iNumber>0)
			{
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "References to this variable: %d", iNumber);
			}
			if (!bOwned)
			{
				Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Variable is defined in code.");
			}
		}
	}

	// Display some info about the entity we are tracking: its guid, classification and what data we are copying.
	Asura_ServerEntity_ConsoleVar* pxServer = Asura_ServerEntity_ConsoleVar::FindEntity(GetGuid());
	if (pxServer && pxServer->m_uEntityToTrackGuid != ASURA_GUID_UNREGISTERED)
	{
		const char * pszString = "(entity not found)";

		const Asura_Entity_Server* pxTracked = Asura_Entity_Server::FindEntity(pxServer->m_uEntityToTrackGuid);		
		if (pxTracked)
		{
			const u_int uClass = pxTracked->GetClassification();

			switch (uClass)
			{
			default:
				pszString = "(unexpected classification)";
				break;

			case AsuraEntityClass_CountedTrigger:
				pszString = "(counted trigger - tracks counter)";
				break;

			case AsuraEntityClass_TimeTrigger:
				if (pxServer->IsTimeCountUp())
				{
					pszString = "(time trigger - tracks time elapsed)";
				}
				else
				{
					pszString = "(time trigger - tracks time remaining)";
				}
				break;
			}
		}

		fY++;
		Asura_Text_Output::Print(fX, fY++, uCOLOUR_WHITE, 0, "Entity to track: %d %s", pxServer->m_uEntityToTrackGuid, pszString);
	}
}

//////////
// Copy //
//////////
void Tools_Entity_ConsoleVar::Copy(const Tools_Object* pxObject)
{
	PARENT::Copy(pxObject);
	
	
	Asura_ServerEntity_ConsoleVar* pxFromEntity = Asura_ServerEntity_ConsoleVar::FindEntity(pxObject->GetGuid());
	Asura_ServerEntity_ConsoleVar* pxToEntity   = Asura_ServerEntity_ConsoleVar::FindEntity(GetGuid());

	if (!pxFromEntity || !pxToEntity)
	{
		Asura_Assert(0,"Couldn't find server entity to do copy");
		return;
	}

	// Copy any server data.
	pxToEntity->m_eType = pxFromEntity->m_eType;
	pxToEntity->m_uFlags = pxFromEntity->m_uFlags;
	pxToEntity->m_uEntityToTrackGuid = pxFromEntity->m_uEntityToTrackGuid;

	
	Asura_ClientEntity_ConsoleVar* pxClientFromEntity = Asura_ClientEntity_ConsoleVar::FindEntity(pxObject->GetGuid());
	Asura_ClientEntity_ConsoleVar* pxClientToEntity   = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());

	if (!pxClientFromEntity || !pxClientToEntity)
	{
		Asura_Assert(0,"Couldn't find client entity to do copy");
		return;
	}

	// Copy any client data.
	// Now we will have two entities pointing at the same variable, one of which will own it.
	pxClientToEntity->ChangeVarName(pxClientFromEntity->GetVarName(), pxClientFromEntity->GetType());
	pxClientToEntity->m_uFlags = pxClientFromEntity->m_uFlags;

	// Copy data to tools entity.
	Tools_Entity_Object* pxToolsToEntity = Tools_Entity_Object::FindEntity(GetGuid());
	if (pxToolsToEntity)
	{
		if (pxToolsToEntity->Type == TOT_Entity_ConsoleVar)
		{
			Tools_Entity_ConsoleVar* pxTools = static_cast<Tools_Entity_ConsoleVar*>(pxToolsToEntity);
			pxTools->CopyVarNameFromClient();
		}
	}
}

////////////////////////////
// CreateObjectOfThisType //
////////////////////////////
Tools_Object* Tools_Entity_ConsoleVar::CreateObjectOfThisType() const
{
	// NOTE: Create() doesn't return the created object, but does select it.
	Tools_Object* pxOldSelection = Tools_Object_System::SelectedObject;
	Create();
	Tools_Object* pxNewSelection = Tools_Object_System::SelectedObject;

	return (pxOldSelection!=pxNewSelection) ? pxNewSelection : 0;
}

//////////////////////////////////////
// AddMessageOverlayControls_Change //
//////////////////////////////////////
const int iNUM_OPERATIONS = 4;

const char* g_apszOperationList[iNUM_OPERATIONS] =
{
	"Set",				//FLAG_OPERATION_SET
	"Add",				//FLAG_OPERATION_ADD
	"Multiply",			//FLAG_OPERATION_MULTIPLY
	"Toggle Boolean",	//FLAG_OPERATION_TOGGLE_BOOLEAN
};

const char* g_apszOperationHelp[iNUM_OPERATIONS] =
{
	"Set console variable to specified value.",
	"Add data to console variable (not relevant for Booleans).",
	"Multiply console variable by the float data value (not relevant for Booleans).",
	"Toogle boolean value (only relevant for Booleans, data is ignored)."
};

// Convert float to an integer representing the boolean.
class Tools_Dialog_BooleanStoredAsFloat : public Asura_Dialog_SharedValue<int>
{
public:
	Tools_Dialog_BooleanStoredAsFloat(float* pfVariable)
		: m_pfVariable(pfVariable)
	{}

	virtual void GetValue(int* pxValue, bool* pbMixed) const
	{
		if (pbMixed)
		{
			*pbMixed = false;
		}
		if (pxValue && m_pfVariable)
		{
			*pxValue = (*m_pfVariable != 0.0f) ? 1 : 0;
		}
	}

	virtual void SetValue(const int& xValue)
	{
		if (m_pfVariable)
		{
			*m_pfVariable = (xValue ? 1.0f : 0.0f);
		}
	}

public:
	float* m_pfVariable;
};

// Convert float to an integer using a well defined conversion.
class Tools_Dialog_IntegerStoredAsFloat : public Asura_Dialog_SharedValue<int>
{
public:
	Tools_Dialog_IntegerStoredAsFloat(float* pfVariable)
		: m_pfVariable(pfVariable)
	{}

	virtual void GetValue(int* pxValue, bool* pbMixed) const
	{
		if (pbMixed)
		{
			*pbMixed = false;
		}
		if (pxValue && m_pfVariable)
		{
			*pxValue = Asura_Maths::FloatToInt_Near(*m_pfVariable);
		}
	}

	virtual void SetValue(const int& xValue)
	{
		if (m_pfVariable)
		{
			*m_pfVariable = static_cast<float>(xValue);
		}
	}

public:
	float* m_pfVariable;
};


void Tools_Entity_ConsoleVar::AddMessageOverlayControls_Change(Asura_StaticMessage* pxMessage,
															   Asura_Overlay* pxOverlay,
															   Asura_Vector_2& xOffset)
{
	// MGH 25/6/2007: This will work so long as we never have two message editing dialogs open at once.
	// The data is stored as a float whatever the type actually is.
	static Tools_Dialog_BooleanStoredAsFloat xBoolean(0);
	static Tools_Dialog_IntegerStoredAsFloat xInteger(0);

	Asura_Assert(pxMessage && pxMessage->m_pData, "Null pointer");
	Asura_Message_Change_ConsoleVar* pxData = static_cast<Asura_Message_Change_ConsoleVar*>(pxMessage->m_pData);

	// List of operations
	Asura_DropDown_List_Dialog* pxListDialog = new Asura_DropDown_List_Dialog(xOffset, g_apszOperationList, &pxData->m_iFlags, iNUM_OPERATIONS, iNUM_OPERATIONS, 0, 0, "Operation" );
	pxListDialog->SetListHelpStrings(g_apszOperationHelp);
	pxOverlay->AddDialog(pxListDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	xOffset.y += Asura_Dialog::DefaultSpacing();

	// Float
	Asura_Dialog* pxDialog = 0;
	pxDialog = new Asura_Numerical_Dialog("Float data", xOffset, &pxData->m_fData, -99999.0f, 99999.0f, 1.0f, "%.3f");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	// Integer
	xInteger.m_pfVariable = &pxData->m_fData;

	pxDialog = new Asura_Numerical_Dialog("Integer data", xOffset, &xInteger, -99999, +99999, 1, "%d");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	// Boolean
	xBoolean.m_pfVariable = &pxData->m_fData;

	pxDialog = new Asura_Numerical_Dialog("Boolean data (1 is true)", xOffset, &xBoolean, 0, 1, 1, "%d");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	xOffset.y += Asura_Dialog::DefaultSpacing();
}

/////////////////////////////////////////////////
// Tools_Entity_ConsoleVar_ChooseEntityToTrack //
/////////////////////////////////////////////////
// Only certain types of server entities can be tracked by the console variable.
bool Tools_Entity_ConsoleVar_ChooseEntityToTrack::CanBeSelected( Tools_Object* pxObject )
{
	if (pxObject)
	{
		const int iType = pxObject->Type;

		if (iType == TOT_Entity_TimeTrigger ||
			iType == TOT_Entity_CountedTrigger)
		{
			return true;
		}
	}
	
	return false;
}

#endif //ASURA_SUPPORT_COMMAND_CONSOLE
