// Tools_Entity_DialogueController.cpp
// Thomas Young - 17.11.08
// Copyright (C) 2008 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Sound/Tools_Entity_DialogueController.h"

#ifdef ASURA_USE_DIALOGUE_SYSTEM

#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "Entity/Sound/Asura_ServerEntity_DialogueController.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Entity/Asura_Entity_Event.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_Text_Output.h"
#include "Sound/Asura_Dialogue.h"
#include "Sound/Asura_Dialogue_System.h"

#include "Entity/Tools_Entity_Data.h"
#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
#include "Entity/Tools_MessageEditor.h"
#endif
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Dialog.h"
#include "Objects/Tools_Object_System.h"
#include "Sound/Tools_PickDialogueEvent_Overlay.h"
#include "Sound/Tools_PickDialogueLine_Overlay.h"
#include "Sound/Tools_Dialogue_System.h"
#include "Sound/Tools_DialogueEvent_Overlay.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

static const char* const g_aszTabs[] = 
{ 
	"Lines", 
	"Character Map" 
};

static char* g_aszPlaybackTypes[] =
{
	"All Sequential",
	"One By One",
	"Random",
	"CD Player Random",
};

static const float fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT = Asura_Dialog::DefaultSpacing() + 0.35f;

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

List<Asura_Dialogue_CharacterMapping>	Tools_Entity_DialogueController::s_xRecentCharacterMappings;
Tools_Entity_DialogueController*		Tools_Entity_DialogueController::s_pxThis						= 0;

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

// ___________________________ Tools_ChooseObject_CharacterMap_Controller _______________________________________


class Tools_ChooseObject_CharacterMap_Controller : public Tools_ChooseObject
{
public:

	Tools_ChooseObject_CharacterMap_Controller( Tools_Entity_DialogueController* pxDialogueController )
	{
		m_pxDialogueController = pxDialogueController;
	}

	virtual bool CanBeSelected(Tools_Object* pxObject)
	{
		return ( pxObject->GetGuid() != ASURA_GUID_UNREGISTERED );
	}

	virtual void ObjectSelected(Tools_Object* pxObject)
	{
		Asura_Guid uGuid = ASURA_GUID_UNREGISTERED;
		if( pxObject )
		{
			uGuid = pxObject->GetGuid();
		}
		m_pxDialogueController->EditCharacterMap_GuidSelected( uGuid );
	}

	virtual void OnFinish()
	{
		delete this;
	}

	Tools_Entity_DialogueController* m_pxDialogueController;
};

// _______________________________ Tools_Entity_DialogueController _______________________________________________

Tools_Entity_DialogueController::Tools_Entity_DialogueController( Asura_Entity* pxEntity ) : Tools_Entity_Activatable_Object( pxEntity )
{
	Type = TOT_Entity_DialogueController;
	SetMesh( "ToolsShape_DialogueCon" );

	m_iSelectedCharacter		= 0;
	m_iNumCharacterMapStrings	= 0;
	m_pszCharacterMapStrings	= 0;
	m_uCharacterMappingValidity = static_cast<u_int>( -1 );

	m_eAnimation				= ANIMATION_TYPE_NONE;
	m_iAnimationLine			= -1;
	m_fAnimationTimer			= 0.0f;
	m_pxOverlay					= 0;
	m_pxRandDelayDialogs[0]		= 0;
	m_pxRandDelayDialogs[1]		= 0;

	UpdateCharacterMappings ();

	// Print some messages if there are problems with the server entity, it should have validated its lines on load
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	if( pxDialogueEntity )
	{
		if( !pxDialogueEntity->IsEveryLinePresent() && !pxDialogueEntity->IsEveryLineValid() )
		{
			// If missing lines and some lines are invalid give a single general message rather than 2 specific ones
			Tools_MessageLog::AddMessage( Tools_MessageLog::DIALOGUE, "Verification", NULL, "Dialogue controller '%s' (GUID %u) has extra lines and missing lines", GetName(), GetGuid() );
		}
		else
		{
			if( !pxDialogueEntity->IsEveryLinePresent() )
			{
				Tools_MessageLog::AddMessage( Tools_MessageLog::DIALOGUE, "Verification", NULL, "Dialogue controller '%s' (GUID %u) has missing lines", GetName(), GetGuid() );
			}
			if( !pxDialogueEntity->IsEveryLineValid() )
			{
				Tools_MessageLog::AddMessage( Tools_MessageLog::DIALOGUE, "Verification", NULL, "Dialogue controller '%s' (GUID %u) has extraneous lines", GetName(), GetGuid() );
			}
		}
		
	}
}

Tools_Entity_DialogueController::~Tools_Entity_DialogueController()
{
	ClearAnimationData ();
	DeleteCharacterMapStrings ();
}


void Tools_Entity_DialogueController::Create()
{
	Tools_Object* pxObject = Create( false );
	Tools_Object_System::SelectObject( pxObject );
}

Tools_Entity_DialogueController* Tools_Entity_DialogueController::Create( bool bGenerateName )
{
	//get a name , and the position data
	Asura_Guid uGuid = Asura_Entity_System::GetNextToolsCreatedGuid();
	Tools_Entity_Data* pxToolsData = 0;
	
	if(bGenerateName)
	{
		pxToolsData = Tools_Entity_Object::CreateToolsDataWithDefaultName("Dialogue Controller", "Dialogue Controller",uGuid,GetNamePrefix());
	}
	else
	{
		pxToolsData = Tools_Entity_Object::CreateToolsData("Dialogue Controller",0,GetNamePrefix());
	}

	if(!pxToolsData) 
	{
		return 0;
	}

	//create the asura data
	Asura_ServerEntity_DialogueController* pxAsuraEntity = new Asura_ServerEntity_DialogueController( );
	pxAsuraEntity->ToolsData = pxToolsData;

	//create the tools object
	Tools_Entity_DialogueController* pxObject = new Tools_Entity_DialogueController( pxAsuraEntity );

	Tools_SaveAsura::SetEntityModified();

	return pxObject;
}


Tools_Object* Tools_Entity_DialogueController::CreateObjectOfThisType() const
{
	return Create( true );	
}

void Tools_Entity_DialogueController::Destroy()
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	for( u_int uLine = 0; uLine < pxDialogueEntity->m_xLines.Size(); ++uLine )
	{
		Asura_StaticMessage_System::DeleteMessageBlock( pxDialogueEntity->m_xLines[uLine].m_usStartMsgBlock );
		pxDialogueEntity->m_xLines[uLine].m_usStartMsgBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;
		
		Asura_StaticMessage_System::DeleteMessageBlock( pxDialogueEntity->m_xLines[uLine].m_usEndMsgBlock );
		pxDialogueEntity->m_xLines[uLine].m_usEndMsgBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;
	}
		
	Tools_Entity_Activatable_Object::Destroy();
}

void Tools_Entity_DialogueController::Copy(const Tools_Object* pxObject)
{
	Tools_Entity_Activatable_Object::Copy(pxObject);

	const Tools_Entity_DialogueController* pxFromObject = (const Tools_Entity_DialogueController*) pxObject;

	Asura_ServerEntity_DialogueController* pxFromEntity = (Asura_ServerEntity_DialogueController*) pxFromObject->Entity;
	Asura_ServerEntity_DialogueController* pxToEntity = (Asura_ServerEntity_DialogueController*) Entity;

	pxToEntity->m_uEventHash = pxFromEntity->m_uEventHash;
	pxToEntity->m_pxEvent = pxFromEntity->m_pxEvent;

	pxToEntity->m_bDeactivateLinesOnUse = pxFromEntity->m_bDeactivateLinesOnUse;
	pxToEntity->m_bUseRandomDelay = pxFromEntity->m_bUseRandomDelay;
	pxToEntity->m_fMinRandomDelay = pxFromEntity->m_fMinRandomDelay;
	pxToEntity->m_fMaxRandomDelay =pxFromEntity-> m_fMaxRandomDelay;

	pxToEntity->m_xMappings.Clear();
	pxToEntity->m_xMappings.SetSize( pxFromEntity->m_xMappings.Size() );
	for( u_int uMapping = 0; uMapping < pxFromEntity->m_xMappings.Size(); ++uMapping )
	{
		pxToEntity->m_xMappings[uMapping] = pxFromEntity->m_xMappings[uMapping];
	}

	//Delete any old message blocks attached to the lines
	for( u_int uLine = 0; uLine < pxToEntity->m_xLines.Size(); ++uLine )
	{
		Asura_StaticMessage_System::DeleteMessageBlock( pxToEntity->m_xLines[uLine].m_usStartMsgBlock );
		Asura_StaticMessage_System::DeleteMessageBlock( pxToEntity->m_xLines[uLine].m_usEndMsgBlock );
	}
	
	pxToEntity->m_xLines.Clear();
	pxToEntity->m_xLines.SetSize( pxFromEntity->m_xLines.Size() );
	for( u_int uLine = 0; uLine < pxFromEntity->m_xLines.Size(); ++uLine )
	{
		//copy the message blocks
		pxToEntity->m_xLines[uLine] = pxFromEntity->m_xLines[uLine];

		pxToEntity->m_xLines[uLine].m_usStartMsgBlock	= Asura_StaticMessage_System::CopyMessageBlock( pxFromEntity->m_xLines[uLine].m_usStartMsgBlock	, pxFromObject->GetGuid(), GetGuid());
		pxToEntity->m_xLines[uLine].m_usEndMsgBlock		= Asura_StaticMessage_System::CopyMessageBlock( pxFromEntity->m_xLines[uLine].m_usEndMsgBlock	, pxFromObject->GetGuid(), GetGuid());
	}

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::UpdateMessageEditorTable();
	#endif

}

void Tools_Entity_DialogueController::AddOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset)
{
	Tools_Entity_Activatable_Object::AddOverlayControls(pxOverlay,xOffset);

	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	// Validate the lines in the controller in case anything has changed in tools
	pxDialogueEntity->ValidateLines();

	m_pxOverlay = pxOverlay;
	
	const float fOriginX = xOffset.x;
	
	Asura_Dialog* pxDialog;
	char szText[200];

	xOffset.x += 30.0f;
	xOffset.y -= 5.6f;

	if( pxDialogueEntity->m_pxEvent )
	{
		Asura_PrintfToBfr( szText, "Dialogue Event: %s", 200, true, pxDialogueEntity->m_pxEvent->GetName() );
	}
	else
	{
		Asura_PrintfToBfr( szText, "Dialogue Event: *Not Set*", 200, true );
	}
	
	pxDialog = new Asura_Function_Dialog( szText, xOffset, _SelectEvent );
	pxDialog->SetHelpText("Select the dialogue event to associate with this controller");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Edit event properties", xOffset, _ViewEvent );
	pxDialog->SetHelpText("Show the properties of the dialogue event associated with this controller");
	pxDialog->SetEnabled( pxDialogueEntity->m_pxEvent != 0 );
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	
	// Show some details about this event
	if( pxDialogueEntity->m_pxEvent )
	{
		Asura_PrintfToBfr( szText, "File: %s", 200, true, pxDialogueEntity->m_pxEvent->GetFilename() );
	}
	else
	{
		Asura_PrintfToBfr( szText, "File: N/A", 200, true );
	}
	pxDialog = new Asura_Dialog( szText, xOffset );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	if( pxDialogueEntity->m_pxEvent )
	{
		Asura_PrintfToBfr( szText, "Playback Type: %s", 200, true, g_aszPlaybackTypes[ pxDialogueEntity->m_pxEvent->GetPlaybackType() ] );
	}
	else
	{
		Asura_PrintfToBfr( szText, "Playback Type: N/A", 200, true );
	}
	pxDialog = new Asura_Dialog( szText, xOffset );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	if( pxDialogueEntity->m_pxEvent )
	{
		Asura_PrintfToBfr( szText, "Play Probability: %d", 200, true, pxDialogueEntity->m_pxEvent->GetPlayProbability() );
	}
	else
	{
		Asura_PrintfToBfr( szText, "Play Probability: N/A", 200, true );
	}
	pxDialog = new Asura_Dialog( szText, xOffset );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	if( pxDialogueEntity->m_pxEvent )
	{
		Asura_PrintfToBfr( szText, "Replay Delay: %.2f", 200, true, pxDialogueEntity->m_pxEvent->GetReplayDelayTime() );
	}
	else
	{
		Asura_PrintfToBfr( szText, "Replay Delay: N/A", 200, true );
	}
	pxDialog = new Asura_Dialog( szText, xOffset );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	xOffset.x = fOriginX;

	if( pxDialogueEntity->m_uEventHash != ASURA_HASH_ID_UNSET )
	{
		if( !pxDialogueEntity->m_pxEvent )
		{
			pxDialog = new Asura_Dialog( "The event associated with this controller was not found", xOffset );
			xOffset.y += Asura_Dialog::DefaultSpacing();
		}
		else
		{
			Asura_Tabs_Dialog* pxTabs = new Asura_Tabs_Dialog( xOffset, g_aszTabs, 2, 200 );
			pxOverlay->AddDialog( pxTabs );
			xOffset.y = pxTabs->m_xMaxCoord.y + 1.0f;

			//redirect all subsequent dialogs to the tab dialog
			pxOverlay->SetCurrentParentDialog( pxTabs );	

			// Add controls to the 'lines' tab
			pxTabs->SelectTab( 0 );

			pxDialog = new Asura_Tickbox_Dialog( "Deactivate lines once played", xOffset, &pxDialogueEntity->m_bDeactivateLinesOnUse );
			pxDialog->SetHelpText("Tick to have dialogue lines deactivated once they have been played");
			pxOverlay->AddDialog(pxDialog);
			xOffset.y += Asura_Dialog::DefaultSpacing();

			Asura_Tickbox_Dialog* pxTickDialog = new Asura_Tickbox_Dialog( "Use Random Delay", xOffset, &pxDialogueEntity->m_bUseRandomDelay );
			pxTickDialog->SetClickedFunction( _ClickedRandDelay );
			pxTickDialog->SetHelpText("Tick to have each line delay further playback for a random amount of time");
			pxOverlay->AddDialog(pxTickDialog);
			xOffset.y += Asura_Dialog::DefaultSpacing() + 0.5f;

			m_pxRandDelayDialogs[0] = new Asura_Numerical_Dialog("Min", xOffset, &pxDialogueEntity->m_fMinRandomDelay, 0.0f, 120.0f, 0.1f, "%.1f");
			m_pxRandDelayDialogs[0]->SetHelpText("Set the minimum delay time for use with random delays");
			m_pxRandDelayDialogs[0]->SetEnabled( pxDialogueEntity->m_bUseRandomDelay );
			pxOverlay->AddDialog(m_pxRandDelayDialogs[0]);
			xOffset.x += 12.0f;

			m_pxRandDelayDialogs[1] = new Asura_Numerical_Dialog("Max", xOffset, &pxDialogueEntity->m_fMaxRandomDelay, 0.0f, 120.0f, 0.1f, "%.1f");
			m_pxRandDelayDialogs[1]->SetHelpText("Set the maximum delay time for use with random delays");
			m_pxRandDelayDialogs[1]->SetEnabled( pxDialogueEntity->m_bUseRandomDelay );
			pxOverlay->AddDialog(m_pxRandDelayDialogs[1]);
			xOffset.x -= 12.0f;
			xOffset.y += Asura_Dialog::LargeSpacing();

			if( !pxDialogueEntity->IsEveryLinePresent() )
			{
				xOffset.x += 35.0f;
				xOffset.y -= 5.0f;

				pxDialog = new Asura_Panel_Dialog( xOffset, Asura_Vector_2( 20.5f, 1.0f ), 0x7AFFD800 );
				pxOverlay->AddDialog(pxDialog);

				pxDialog = new Asura_Dialog( "Not all lines from the event are added", xOffset );
				pxOverlay->AddDialog(pxDialog);

				xOffset.y += 5.0f;
				xOffset.x -= 35.0f;
			}

			// Add all the lines in the server entity

			xOffset.x += 1.5f;

			pxDialog = new Asura_Dialog( "Active", xOffset );
			pxOverlay->AddDialog( pxDialog );
			xOffset.x += 5.0f;

			pxDialog = new Asura_Dialog( "Weight", xOffset );
			pxOverlay->AddDialog( pxDialog );
			xOffset.x += 7.0f;

			pxDialog = new Asura_Dialog( "Line ID Number", xOffset );
			pxOverlay->AddDialog( pxDialog );
			xOffset.x += 22.0f;

			#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
			xOffset.x += 0.5f;
			#else
			pxDialog = new Asura_Dialog( "On Start", xOffset );
			pxOverlay->AddDialog( pxDialog );
			xOffset.x += 5.5f;

			pxDialog = new Asura_Dialog( "On End", xOffset );
			pxOverlay->AddDialog( pxDialog );
			xOffset.x += 5.0f;
			#endif

			pxDialog = new Asura_Dialog( "Play Next", xOffset );
			pxOverlay->AddDialog( pxDialog );

			xOffset.x = fOriginX;
			xOffset.y += Asura_Dialog::DefaultSpacing() + 0.4f;

			const int iIconTexture = Tools_Dialog::GetIconTexture();

			for( int iLine = 0; iLine < static_cast<int>( pxDialogueEntity->m_xLines.Size() ); ++iLine )
			{
				// Move some of the control y positions based on the animation status
				if( ( m_eAnimation == ANIMATION_TYPE_REMOVE ) && ( m_iAnimationLine == iLine ) )
				{
					// We are doing the animation for this line being removed
					xOffset.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
				}

				if( ( m_eAnimation == ANIMATION_TYPE_MOVE_DOWN ) )
				{
					if ( iLine == m_iAnimationLine )
					{
						// This line is moving up
						xOffset.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
					}
					else if( iLine == ( m_iAnimationLine + 1 ) )
					{
						// This line is moving down
						xOffset.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
						xOffset.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
					}
				}

				if( ( m_eAnimation == ANIMATION_TYPE_MOVE_UP ) )
				{
					if ( iLine == m_iAnimationLine )
					{
						// This line is moving down
						xOffset.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
						xOffset.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
					}
					else if( iLine == ( m_iAnimationLine - 1 ) )
					{
						// This line is moving up
						xOffset.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
					}
				}

				Asura_DialogueController_Line* pxLine = &( pxDialogueEntity->m_xLines[ iLine ] );

				Asura_PrintfToBfr( szText, "%d:", 200, true, ( iLine + 1 ) );
				pxDialog = new Asura_StringDisplay_Dialog( szText, xOffset );
				pxOverlay->AddDialog( pxDialog );
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 1.5f;

				pxDialog = new Asura_Tickbox_Dialog( "", xOffset, &pxLine->m_bActive );
				pxDialog->SetHelpText("Tick to make this line playable");
				pxOverlay->AddDialog(pxDialog);
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 3.0f;

				bool bRandomPlayback = false;
				if( pxDialogueEntity->m_pxEvent && ( pxDialogueEntity->m_pxEvent->GetPlaybackType() == ASURA_DPT_RANDOM ) )
				{
					bRandomPlayback = true;
				}

				pxDialog = new Asura_Numerical_Dialog("", xOffset, &pxLine->m_ucWeight, 0, 100, 5, "%d");
				pxDialog->SetHelpText("Set the chance for this line to play if using random playback");
				pxDialog->SetEnabled( bRandomPlayback );
				pxOverlay->AddDialog(pxDialog);
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 9.0f;

				const Asura_Dialogue_Line* pxDialogueLine = Asura_Dialogue_Storage::FindLine( pxLine->m_uDialogueLine );
				
				if( !pxLine->m_bIsValid )
				{
					pxDialog = new Asura_Panel_Dialog( xOffset, Asura_Vector_2( 22.0f, fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT * 0.6f ), 0x7AC20008 );
					pxOverlay->AddDialog(pxDialog);
					StoreAnimationData( iLine, pxDialog );
				}

				// Build the preview text for this line
				const Asura_UnicodeChar* wszText = pxDialogueLine ? pxDialogueLine->GetText() : 0;
				Asura_PrintfToBfr(szText,"%S",sizeof(szText)-3, true, wszText ? wszText : L"*None*");
				//add ellipsis if too long
				if( strlen(szText) == sizeof(szText)-4 )
				{
					strcat(szText,"...");
				}
				//replace newline characters by spaces
				for( char* pcNewLine = strchr(szText,'\n');pcNewLine;pcNewLine = strchr(pcNewLine,'\n') )
				{
					*pcNewLine = ' ';
				}

				pxDialog = new Asura_Function_Dialog( pxDialogueLine ? pxDialogueLine->GetName() : "*None*", xOffset, _SelectLine, iLine );
				pxDialog->SetHelpText( szText );
				pxOverlay->AddDialog(pxDialog);
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 23.0f;

				Asura_PrintfToBfr( szText, "%d", 200, true, Asura_StaticMessage_System::GetNumberOfMessagesInBlock( pxLine->m_usStartMsgBlock ) );
				pxDialog = new Asura_Function_Dialog( szText, xOffset, _EditStartMsgs, iLine );
				pxDialog->SetHelpText("Edit the start messages sent for this line entry");
				pxOverlay->AddDialog(pxDialog);
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 5.0f;
				
				Asura_PrintfToBfr( szText, "%d", 200, true, Asura_StaticMessage_System::GetNumberOfMessagesInBlock( pxLine->m_usEndMsgBlock ) );
				pxDialog = new Asura_Function_Dialog( szText, xOffset, _EditEndMsgs, iLine );
				pxDialog->SetHelpText("Edit the end messages sent for this line entry");
				pxOverlay->AddDialog(pxDialog);
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 5.5f;

				pxDialog = new Asura_Tickbox_Dialog( "", xOffset, &pxLine->m_bPlayNext );
				pxDialog->SetHelpText("Tick this to have the next line of dialogue automatically played on completion of this one");
				pxOverlay->AddDialog(pxDialog);
				StoreAnimationData( iLine, pxDialog );
				xOffset.x += 4.5f;

				// The icon buttons are a bit big so offset y to centralise them
				xOffset.y -= 0.2f;

				if( iLine > 0 )
				{
					Asura_Icon_Dialog* pxIconDlg = new Asura_Icon_Dialog( xOffset, iIconTexture, TOOLS_ICON_ARROW_UP );
					pxIconDlg->SetClickedFunction( _MoveLineUp, iLine );
					pxIconDlg->SetHelpText( "Move this line up in the ordering" );
					pxOverlay->AddDialog( pxIconDlg );
					StoreAnimationData( iLine, pxIconDlg );
				}
				xOffset.x += 2.0f;

				if( iLine != static_cast<int>( pxDialogueEntity->m_xLines.Size() - 1 ) )
				{
					Asura_Icon_Dialog* pxIconDlg = new Asura_Icon_Dialog( xOffset, iIconTexture, TOOLS_ICON_ARROW_DOWN );
					pxIconDlg->SetClickedFunction( _MoveLineDown, iLine );
					pxIconDlg->SetHelpText( "Move this line down in the ordering" );
					pxOverlay->AddDialog( pxIconDlg );
					StoreAnimationData( iLine, pxIconDlg );
				}
				xOffset.x += 3.0f;

				Asura_Icon_Dialog* pxIconDlg = new Asura_Icon_Dialog( xOffset, iIconTexture, TOOLS_ICON_DELETE );
				pxIconDlg->SetClickedFunction( _DeleteLine, iLine );
				pxIconDlg->SetHelpText( "Remove this line from the controller line list" );
				pxOverlay->AddDialog( pxIconDlg );
				StoreAnimationData( iLine, pxIconDlg );

				xOffset.y += 0.2f;

				xOffset.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;

				if(		( ( m_eAnimation == ANIMATION_TYPE_MOVE_DOWN ) && ( iLine == ( m_iAnimationLine + 1 ) ) )
					||	( ( m_eAnimation == ANIMATION_TYPE_MOVE_UP	 ) && ( iLine == m_iAnimationLine ) ) )
				{
					xOffset.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
				}

				xOffset.x = fOriginX;
			}
			
			if( ( m_eAnimation == ANIMATION_TYPE_REMOVE ) && ( static_cast<int>( pxDialogueEntity->m_xLines.Size() ) == m_iAnimationLine ) )
			{
				xOffset.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
			}
			
			xOffset.x += 13.5f;
			
			pxDialog = new Asura_Function_Dialog( "Add a line...", xOffset, _AddLine );
			pxDialog->SetHelpText("Select a dialogue line from the associated event to add to this controller");
			pxOverlay->AddDialog(pxDialog);
			StoreAnimationData( 1000, pxDialog );

			xOffset.x = fOriginX;
			
			// Add controls to the 'Character map' tab
			pxTabs->SelectTab( 1 );

			xOffset.y = pxTabs->m_xMaxCoord.y + 1.0f;

			//character mapping list
			if( !m_pszCharacterMapStrings ) 
			{
				UpdateCharacterMapStrings();
			}

			Asura_List_Dialog* pxListDialog;
			pxListDialog = new Asura_List_Dialog( xOffset, m_pszCharacterMapStrings, &m_iSelectedCharacter, 6, m_iNumCharacterMapStrings, NULL );
			pxListDialog->SetDoubleClickFunction( _EditCharMap, 0 );
			pxListDialog->SetHelpText("Map of characters in dialogue to guids/entities in level");
			pxListDialog->m_iControlID = Asura_GetHashID("Character Map");
			pxOverlay->AddDialog(pxListDialog);
			xOffset.y = pxListDialog->m_xMaxCoord.y + 1.0f;

			pxDialog = new Asura_Function_Dialog("Edit Character Mapping", xOffset, _EditCharMap, 0 );
			pxDialog->SetHelpText("Edit mapping for selected entry in list");
			pxOverlay->AddDialog(pxDialog);
			xOffset.y += Asura_Dialog::DefaultSpacing();

			pxDialog = new Asura_Function_Dialog("Clear Character Mapping", xOffset, _ClearCharMap );
			pxDialog->SetHelpText("Clear mapping for selected entry in list");
			pxOverlay->AddDialog(pxDialog);
			xOffset.y += Asura_Dialog::DefaultSpacing();

			pxTabs->SelectTab( 0 );
		}
	}
}


void Tools_Entity_DialogueController::UpdateOverlayControls(Asura_Overlay* pxOverlay)
{
	Tools_Entity_Activatable_Object::UpdateOverlayControls(pxOverlay);


	//update the text on the choose dialogue control
	Asura_ServerEntity_DialogueController* pxDialogueEntity = (Asura_ServerEntity_DialogueController*) Entity;
	char szText[100];
	if(pxDialogueEntity->m_pxEvent)
	{
		Asura_PrintfToBfr(szText,"Dialogue : %s",sizeof(szText),true,pxDialogueEntity->m_pxEvent->GetName());
	}
	else
	{
		Asura_PrintfToBfr(szText,"Dialogue : 0x%x",sizeof(szText),true,pxDialogueEntity->m_uEventHash);
	}

	pxOverlay->UpdateDialogLabel(Asura_GetHashID("Dialogue"),szText);

}

void Tools_Entity_DialogueController::UpdateCharacterMappings()
{
	m_uCharacterMappingValidity = Tools_Dialogue_System::GetCharacterMappingValidity();
	
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	if( !pxDialogueEntity->m_pxEvent ) 
	{
		return;
	}

	//get a list of all characters used by the event
	List<Asura_Hash_ID> xCharacterList;

	List<const Asura_Dialogue_Event*> xEventQueue;
	xEventQueue.add_entry( pxDialogueEntity->m_pxEvent );

	for(LIF<const Asura_Dialogue_Event*> xEventLIF(&xEventQueue);!xEventLIF.done();xEventLIF.next())
	{
		const Asura_Dialogue_Event* pxEvent = xEventLIF();
		for(u_int i=0;i<pxEvent->GetNumPlayables();i++)
		{
			const Asura_Dialogue_Playable* pxPlayable = pxEvent->GetPlayable(i);
			if(pxPlayable->IsLine())
			{
				const Asura_Dialogue_Line* pxChildLine = (const Asura_Dialogue_Line*) pxPlayable;
				if( !xCharacterList.contains( pxChildLine->GetCharacter() ) ) 
				{
					xCharacterList.add_entry( pxChildLine->GetCharacter() );
				}

			}
			else
			{
				const Asura_Dialogue_Event* pxChildEvent = (const Asura_Dialogue_Event*) pxPlayable;
				if( !xEventQueue.contains( pxChildEvent ) ) 
				{
					xEventQueue.add_entry( pxChildEvent );
				}
			}
		}
	}

	Asura_HeapArray<Asura_Dialogue_CharacterMapping> xNewList;
	xNewList.SetSize( xCharacterList.size() );

	for(u_int iNewMap = 0; iNewMap < xNewList.Size(); ++iNewMap )
	{
		Asura_Hash_ID uCharacter = xCharacterList[iNewMap];
		
		//see if there was a mapping for this character in the old list
		Asura_Guid uGuid = ASURA_GUID_UNREGISTERED;
		bool bNewEntry = true;
		for(u_int iOldMap = 0; iOldMap < pxDialogueEntity->m_xMappings.Size(); iOldMap++ )
		{
			if( pxDialogueEntity->m_xMappings[iOldMap].m_uCharacterHashID == uCharacter )
			{
				uGuid = pxDialogueEntity->m_xMappings[iOldMap].m_uCharacterGuid;
				bNewEntry = false;
				break;
			}
		}

		if(bNewEntry)
		{
			uGuid = GetDefaultCharacterMapping(uCharacter);
		}

		xNewList[ iNewMap ].m_uCharacterHashID = uCharacter;
		xNewList[ iNewMap ].m_uCharacterGuid = uGuid;
	}

	//replace the old list ( i.e. do a rubbish manual copy)
	pxDialogueEntity->m_xMappings.Clear();
	pxDialogueEntity->m_xMappings.SetSize( xNewList.Size() );
	for(u_int iNewMap = 0; iNewMap < xNewList.Size(); ++iNewMap )
	{
		pxDialogueEntity->m_xMappings[ iNewMap ] = xNewList[ iNewMap ];
	}

	// Clean up
	if(IsSelected())
	{
		//Update the list on the overlay
		UpdateCharacterMapStrings();
	}
	else
	{
		//Old character map strings (if any) will be invalid , but don't need them yet
		DeleteCharacterMapStrings();
	}
}


void Tools_Entity_DialogueController::AddToRecentCharacterMappings(const Asura_Dialogue_CharacterMapping& xMapping)
{
	//remove existing entry for this hash (if any)
	for(LIF<Asura_Dialogue_CharacterMapping> xMappingLIF(&s_xRecentCharacterMappings); !xMappingLIF.done(); xMappingLIF.next())
	{
		if(xMappingLIF().m_uCharacterHashID == xMapping.m_uCharacterHashID)
		{
			xMappingLIF.delete_current();
			break;
			
		}
	}
	//and add a new entry
	s_xRecentCharacterMappings.add_entry_start(xMapping);
}


Asura_Guid Tools_Entity_DialogueController::GetDefaultCharacterMapping(Asura_Hash_ID uCharacterHash)
{
	//check for recently done mappings first
	for( LIF<Asura_Dialogue_CharacterMapping> xMappingLIF(&s_xRecentCharacterMappings); !xMappingLIF.done(); xMappingLIF.next() )
	{
		if(xMappingLIF().m_uCharacterHashID == uCharacterHash)
		{
			return xMappingLIF().m_uCharacterGuid;
		}
	}

	//failing that , check the mappings of existing triggers in the level
	for( Asura_Hashed_List_It<Asura_ServerEntity_DialogueController> xIt( Asura_ServerEntity_DialogueController::GetList() ); !xIt.Done(); xIt.Next() )
	{
		Asura_ServerEntity_DialogueController* pxEntity = xIt.GetCurrent();
		if( pxEntity->IsDestroyed() )
		{
			continue;
		}

		const Asura_HeapArray<Asura_Dialogue_CharacterMapping>* pxList = &pxEntity->m_xMappings;

		if( pxList )
		{
			for( u_int i = 0; i < pxList->Size(); i++ )
			{
				const Asura_Dialogue_CharacterMapping xMapping = (*pxList)[i];
				if(		( xMapping.m_uCharacterHashID	== uCharacterHash ) 
					&&	( xMapping.m_uCharacterGuid		!= ASURA_GUID_UNREGISTERED) )
				{
					return xMapping.m_uCharacterGuid;
				}
			}
		}
	}

	return ASURA_GUID_UNREGISTERED;
}


void Tools_Entity_DialogueController::UpdateCharacterMapStrings()
{
	DeleteCharacterMapStrings();

	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );
	const Asura_HeapArray<Asura_Dialogue_CharacterMapping>* pxList = &pxDialogueEntity->m_xMappings;

	if( pxList->Size() )
	{
		m_iNumCharacterMapStrings = pxList->Size();
		m_pszCharacterMapStrings = new char*[ pxList->Size() ];

		for(u_int i=0;i<pxList->Size();i++)
		{
			const Asura_Dialogue_CharacterMapping& xMapping = (*pxList)[i];
		
			//get the character name
			const char* szCharacter = Tools_Dialogue_System::GetCharacterNameByHash(xMapping.m_uCharacterHashID);

			char szCharacterBuffer[100];
			if(!szCharacter)
			{
				Asura_PrintfToBfr(szCharacterBuffer,"Character 0x%x",sizeof(szCharacterBuffer),true,xMapping.m_uCharacterHashID);
				szCharacter = szCharacterBuffer;
			}

			//get the name of the target object
			char szEntity[100];
			Tools_Object* pxObject = Tools_Object_System::GetObjectFromGuid(xMapping.m_uCharacterGuid);
			if(pxObject)
			{
				Asura_PrintfToBfr(szEntity,"%s",sizeof(szEntity),true,pxObject->GetName());
			}
			else if(xMapping.m_uCharacterGuid == ASURA_GUID_UNREGISTERED)
			{
				strcpy(szEntity,"*Not Set*");
			}
			else
			{
				Asura_PrintfToBfr(szEntity,"Guid %d",sizeof(szEntity),true,xMapping.m_uCharacterGuid);
			}

			//build the full string for the list
			char szString[100];
			Asura_PrintfToBfr(szString,"%s - %s",sizeof(szString),true,szCharacter,szEntity);

			m_pszCharacterMapStrings[i] = new char[strlen(szString)+1];
			strcpy(m_pszCharacterMapStrings[i],szString);
		}
	}
	else
	{
		m_iNumCharacterMapStrings = 1;
		m_pszCharacterMapStrings = new char*[1];
		m_pszCharacterMapStrings[0] = new char[10];
		strcpy(m_pszCharacterMapStrings[0],"*None*");
	}
	
	if(m_iSelectedCharacter >= m_iNumCharacterMapStrings) 
	{
		m_iSelectedCharacter = m_iNumCharacterMapStrings - 1;
	}

	//update the overlay if needed
	if( IsSelected() )
	{
		Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayID());
		if(pxOverlay)
		{
			Asura_List_Dialog* pxDialog = (Asura_List_Dialog*)pxOverlay->GetDialog( Asura_GetHashID("Character Map") );
			if(pxDialog)
			{
				pxDialog->ReplaceStrings( m_iNumCharacterMapStrings, m_pszCharacterMapStrings );
			}
		}
	}
}

void Tools_Entity_DialogueController::DeleteCharacterMapStrings()
{
	for(int i=0;i<m_iNumCharacterMapStrings;i++)
	{
		delete [] m_pszCharacterMapStrings[i];
	}
	delete [] m_pszCharacterMapStrings;
	m_pszCharacterMapStrings = 0;

	m_iNumCharacterMapStrings = 0;

}

void Tools_Entity_DialogueController::EditCharMap()
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );
	
	//get the selected mapping
	if( m_iSelectedCharacter >= static_cast<int>( pxDialogueEntity->m_xMappings.Size() ) )
	{
		return;
	}
	const Asura_Dialogue_CharacterMapping& xMapping = pxDialogueEntity->m_xMappings[ m_iSelectedCharacter ];

	//sort out the prompt
	char szPrompt[200];
	const char* szCharacter = Tools_Dialogue_System::GetCharacterNameByHash( xMapping.m_uCharacterHashID );
	Asura_PrintfToBfr(szPrompt,"Select the entity to use for %s",sizeof(szPrompt),true,szCharacter ? szCharacter : "this character");

	//get the initial selection
	Tools_Object* pxObject = Tools_Object_System::GetObjectFromGuid( xMapping.m_uCharacterGuid );
	
	//initialise and start the object choosing process
	Tools_ChooseObject_CharacterMap_Controller* pxChooseObject = new Tools_ChooseObject_CharacterMap_Controller( this );
	Tools_Object_System::ChooseObject( szPrompt, pxChooseObject, pxObject);
}

void Tools_Entity_DialogueController::EditCharacterMap_GuidSelected(Asura_Guid uGuid)
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );
	
	//update the character mapping in the entity
	if( m_iSelectedCharacter >= static_cast<int>( pxDialogueEntity->m_xMappings.Size() ) )
	{
		return;
	}

	if( pxDialogueEntity->m_xMappings[ m_iSelectedCharacter ].m_uCharacterGuid != uGuid)
	{
		pxDialogueEntity->m_xMappings[ m_iSelectedCharacter ].m_uCharacterGuid = uGuid;
		
		SetEntityModified();
		UpdateCharacterMapStrings();

		AddToRecentCharacterMappings( pxDialogueEntity->m_xMappings[ m_iSelectedCharacter ] );
	}
}

void Tools_Entity_DialogueController::ClearCharMap()
{	
	EditCharacterMap_GuidSelected( ASURA_GUID_UNREGISTERED );
}

bool Tools_Entity_DialogueController::HandlesThisMessage ( int iMessageID )
{
	if(		( iMessageID == ASURA_MSG_DIALOGUE_CONTROLLER_TOGGLE_LINE	) 
		||	( iMessageID == ASURA_MSG_DIALOGUE_CONTROLLER_TRIGGER_LINE		) 
		||	( iMessageID == ASURA_MSG_DIALOGUE_CONTROLLER_INTERRUPT		) )
	{
		return true;
	}
	return PARENT::HandlesThisMessage( iMessageID );
}

bool Tools_Entity_DialogueController::HandlesThisEvent(int iEventID) const
{
	if( iEventID == ASURA_ENTITY_EVENT_ONFINISHED )
	{
		return true;
	}

	return Tools_Entity_Activatable_Object::HandlesThisEvent(iEventID);
}

void Tools_Entity_DialogueController::AddFilesToExportList(List<Asura_String>& nameList) const
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	Tools_Level_Export::AddDialogueEventToExportList( pxDialogueEntity->m_uEventHash );
}


void Tools_Entity_DialogueController::UpdateWhileSelected()
{
	Tools_Entity_Activatable_Object::UpdateWhileSelected();

	if(m_uCharacterMappingValidity != Tools_Dialogue_System::GetCharacterMappingValidity())
	{
		UpdateCharacterMappings();
	}

	// Do our animations
	if( m_eAnimation != ANIMATION_TYPE_NONE )
	{
		m_fAnimationTimer += Asura_Timers::GetActualFrameTime() * 2.0f;

		if( m_fAnimationTimer >= 1.0f )
		{
			// Finished animating
			m_fAnimationTimer = 1.0f;
			m_eAnimation = ANIMATION_TYPE_NONE;
		}

		// Update our dialog positions
		Asura_List_It<Animated_Line_Entry> xAnimListIter( m_xAnimationList );

		for( ; !xAnimListIter.Done(); xAnimListIter.Next() )
		{
			Animated_Line_Entry* pxEntry = xAnimListIter.GetCurrent();
			if( pxEntry )
			{
				const float fLerpFactor = Asura_Maths::Power( m_fAnimationTimer, 0.3f );
				// Lerp to the new position
				Asura_Vector_2 xPosition( Asura_Maths::Lerp( pxEntry->m_xOrigin.x, pxEntry->m_xDestination.x, fLerpFactor )
										, Asura_Maths::Lerp( pxEntry->m_xOrigin.y, pxEntry->m_xDestination.y, fLerpFactor ) );

				pxEntry->m_pxDialog->MoveDialog( xPosition );
			}
		}

		m_pxOverlay->ResizeOverlay();

		// did we finish? if so clear our data
		if( m_eAnimation == ANIMATION_TYPE_NONE )
		{
			ClearAnimationData();
		}
	}
}

void Tools_Entity_DialogueController::PrintInfo(float& fX,float& fY)
{
	Tools_Entity_Activatable_Object::PrintInfo(fX,fY);

	//Display some information about the targeted event
	Asura_ServerEntity_DialogueController* pxDialogueEntity = (Asura_ServerEntity_DialogueController*) Entity;
	const Asura_Dialogue_Event* pxEvent = pxDialogueEntity->m_pxEvent;
	if(!pxEvent) return;

	Asura_Text_Output::Print(fX,fY++,0xffffffff,0,"Play Type : %s",g_aszPlaybackTypes[ pxEvent->GetPlaybackType() ] );
	Asura_Text_Output::Print(fX,fY++,0xffffffff,0,"Num Lines : %u",pxDialogueEntity->m_xLines.Size());
}

bool Tools_Entity_DialogueController::OwnsThisMessageBlock(unsigned short usMessageBlockID)
{
	// Returns true if the message block belongs to this entity

	// Loop over all of our lines
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>(Entity);
	
	if ( pxDialogueEntity )
	{
		for( u_int uLine = 0; uLine< pxDialogueEntity->m_xLines.Size(); ++uLine )
		{
			if(		pxDialogueEntity->m_xLines[ uLine ].m_usStartMsgBlock == usMessageBlockID
				||	pxDialogueEntity->m_xLines[ uLine ].m_usEndMsgBlock	== usMessageBlockID )
			{
				return true;
			}
		}
	}
	return Tools_Entity_Activatable_Object::OwnsThisMessageBlock(usMessageBlockID);
}


bool Tools_Entity_DialogueController::GetEventDescForMessageBlock(unsigned short usMessageBlockID,Asura_String& xOutputDesc)
{
	// Gets description of what causes message block to be sent, returns false if not implemented, or message isn't found.

	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>(Entity);

	if ( pxDialogueEntity )
	{
		for( u_int uLine = 0; uLine< pxDialogueEntity->m_xLines.Size(); ++uLine )
		{
			if(	pxDialogueEntity->m_xLines[ uLine ].m_usStartMsgBlock == usMessageBlockID )
			{
				xOutputDesc = "On Starting Dialogue Line";
				return true;
			}
			else if( pxDialogueEntity->m_xLines[ uLine ].m_usStartMsgBlock == usMessageBlockID )
			{
				xOutputDesc = "On Completing Dialogue Line";
				return true;
			}
		}
	}

	return Tools_Entity_Activatable_Object::GetEventDescForMessageBlock(usMessageBlockID, xOutputDesc);
}



void Tools_Entity_DialogueController::OnSelection	()
{
	//Remember which instance is selected so that button presses have an instance to use
	s_pxThis = this;
	PARENT::OnSelection();
}


void Tools_Entity_DialogueController::OnDeselection	()
{
	s_pxThis = 0;
	ClearAnimationData();
	PARENT::OnDeselection();
}


void Tools_Entity_DialogueController::SelectEvent ()
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	Asura_Hash_ID uEventHash = Tools_PickDialogueEvent_Overlay::PickEvent("Select a dialogue event to associate", pxDialogueEntity->m_uEventHash);
	if( uEventHash && uEventHash != pxDialogueEntity->m_uEventHash )
	{
		//update the entity with the new event
		pxDialogueEntity->SetEventHashID( uEventHash );
		
		UpdateCharacterMappings();

		#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
		Tools_MessageEditor::UpdateMessageEditorTable();
		#endif

		DeleteOverlay();
		CreateOverlay();

		SetEntityModified();
	}
}


void Tools_Entity_DialogueController::ViewEvent ()
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	Tools_DialogueEvent_Overlay::EditEvent( pxDialogueEntity->m_uEventHash );
}


void Tools_Entity_DialogueController::SelectLine ( int iLine )
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	const Asura_Hash_ID uPicked = Tools_PickDialogueLine_Overlay::PickLine( "Select a line to use", 0, false, pxDialogueEntity->m_uEventHash );

	if( !Tools_PickDialogueLine_Overlay::WasCancelled() )
	{
		pxDialogueEntity->m_xLines[ iLine ].m_uDialogueLine = uPicked;
	}

	DeleteOverlay();
	CreateOverlay();

	SetEntityModified();
}


void Tools_Entity_DialogueController::EditStartMsgs ( int iLine )
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = (Asura_ServerEntity_DialogueController*) Entity;

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::StartMessageEditingMode(&pxDialogueEntity->m_xLines[iLine].m_usStartMsgBlock);
	#else
	Tools_Object_System::EditMessageBlock( "Start Messages", &pxDialogueEntity->m_xLines[ iLine ].m_usStartMsgBlock );
	#endif
}


void Tools_Entity_DialogueController::EditEndMsgs ( int iLine )
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = (Asura_ServerEntity_DialogueController*) Entity;

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::StartMessageEditingMode(&pxDialogueEntity->m_xLines[iLine].m_usEndMsgBlock);
	#else
	Tools_Object_System::EditMessageBlock( "End Messages", &pxDialogueEntity->m_xLines[ iLine ].m_usEndMsgBlock );
	#endif
}


void Tools_Entity_DialogueController::DeleteLine ( int iLine )
{
	if( m_eAnimation != ANIMATION_TYPE_NONE )
	{
		return;
	}

	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	if( !pxDialogueEntity )
	{
		return;
	}

	// Remove the line entry from the serverentity
	Asura_StaticMessage_System::DeleteMessageBlock( pxDialogueEntity->m_xLines[ iLine ].m_usStartMsgBlock );
	Asura_StaticMessage_System::DeleteMessageBlock( pxDialogueEntity->m_xLines[ iLine ].m_usEndMsgBlock );
	pxDialogueEntity->m_xLines.RemoveElement( static_cast<u_int>( iLine ) );

	// Animate this line being remove
	m_eAnimation		= ANIMATION_TYPE_REMOVE;
	m_iAnimationLine	= iLine;
	m_fAnimationTimer	= 0.0f;

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::UpdateMessageEditorTable();
	#endif

	DeleteOverlay();
	CreateOverlay();

	SetEntityModified();
}


void Tools_Entity_DialogueController::MoveLineUp ( int iLine )
{
	if( m_eAnimation != ANIMATION_TYPE_NONE )
	{
		return;
	}

	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	if( !pxDialogueEntity )
	{
		return;
	}

	if( iLine == 0 )
	{
		return;
	}
	// Swap the lines in the serverentity
	Asura_DialogueController_Line xTempLine = pxDialogueEntity->m_xLines[ iLine ];

	pxDialogueEntity->m_xLines[ iLine ] = pxDialogueEntity->m_xLines[ iLine - 1 ];
	pxDialogueEntity->m_xLines[ iLine - 1 ] = xTempLine;

	// Animate this line being remove
	m_eAnimation		= ANIMATION_TYPE_MOVE_UP;
	m_iAnimationLine	= iLine;
	m_fAnimationTimer	= 0.0f;

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::UpdateMessageEditorTable();
	#endif

	DeleteOverlay();
	CreateOverlay();

	SetEntityModified();
}

void Tools_Entity_DialogueController::MoveLineDown ( int iLine )
{
	if( m_eAnimation != ANIMATION_TYPE_NONE )
	{
		return;
	}

	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	if( !pxDialogueEntity )
	{
		return;
	}

	if( iLine == static_cast<int>( pxDialogueEntity->m_xLines.Size() ) - 1 )
	{
		return;
	}

	// Swap the lines in the serverentity
	Asura_DialogueController_Line xTempLine = pxDialogueEntity->m_xLines[ iLine ];

	pxDialogueEntity->m_xLines[ iLine ] = pxDialogueEntity->m_xLines[ iLine + 1 ];
	pxDialogueEntity->m_xLines[ iLine + 1 ] = xTempLine;

	// Animate this line being remove
	m_eAnimation		= ANIMATION_TYPE_MOVE_DOWN;
	m_iAnimationLine	= iLine;
	m_fAnimationTimer	= 0.0f;

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::UpdateMessageEditorTable();
	#endif

	DeleteOverlay();
	CreateOverlay();

	SetEntityModified();
}


void Tools_Entity_DialogueController::ClickedRandDelay ()
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	if( m_pxRandDelayDialogs[0] )
	{
		m_pxRandDelayDialogs[0]->SetEnabled( pxDialogueEntity->m_bUseRandomDelay );
	}
	if( m_pxRandDelayDialogs[1] )
	{
		m_pxRandDelayDialogs[1]->SetEnabled( pxDialogueEntity->m_bUseRandomDelay );
	}
}


void Tools_Entity_DialogueController::AddLine ( )
{
	Asura_ServerEntity_DialogueController* pxDialogueEntity = static_cast<Asura_ServerEntity_DialogueController*>( Entity );

	Tools_PickDialogueLine_Overlay::PickLine( "Select dialogue lines to add", ASURA_HASH_ID_UNSET, true, pxDialogueEntity->m_uEventHash );
	
	if( !Tools_PickDialogueLine_Overlay::WasCancelled() && ( Tools_PickDialogueLine_Overlay::GetNumSelected() > 0 ) )
	{
		int iNextArrayEntry = pxDialogueEntity->m_xLines.Size();

		pxDialogueEntity->m_xLines.Resize( iNextArrayEntry + Tools_PickDialogueLine_Overlay::GetNumSelected() );

		for( int iLine = 0; iLine < Tools_PickDialogueLine_Overlay::GetNumSelected(); ++iLine )
		{
			Asura_Hash_ID uSelection = Tools_PickDialogueLine_Overlay::GetSelection( iLine );

			pxDialogueEntity->m_xLines[ iNextArrayEntry ].SetDefaults();
			pxDialogueEntity->m_xLines[ iNextArrayEntry ].m_uDialogueLine = uSelection;
			
			++iNextArrayEntry;
		}

		#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
		Tools_MessageEditor::UpdateMessageEditorTable();
		#endif
		
		DeleteOverlay();
		CreateOverlay();

		SetEntityModified();
	}
}


void Tools_Entity_DialogueController::StoreAnimationData ( const int iLine, Asura_Dialog* pxDialog )
{
	if( m_eAnimation == ANIMATION_TYPE_NONE )
	{
		// Not animating, done
		return;
	}

	Asura_Vector_2 xDestination = pxDialog->m_xOffset;

	switch( m_eAnimation )
	{
		case ANIMATION_TYPE_REMOVE:
		{
			if( iLine < m_iAnimationLine )
			{
				return;
			}
			xDestination.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
		}
		break;

		case ANIMATION_TYPE_MOVE_UP:
		{
			if( iLine == ( m_iAnimationLine - 1 ) )
			{
				// This line is the one which was moved up
				xDestination.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
			}
			else if( iLine == m_iAnimationLine  )
			{
				// This line is the one which was moved down
				xDestination.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
			}
			else
			{
				// No change
				return;
			}
		}
		break;

		case ANIMATION_TYPE_MOVE_DOWN:
		{
			if( iLine == ( m_iAnimationLine + 1 ) )
			{
				// This line is the one which was moved down
				xDestination.y += fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
			}
			else if( iLine == m_iAnimationLine )
			{
				// This line is the one which was moved up
				xDestination.y -= fTOOLS_DIALOGUE_LINE_ENTRY_HEIGHT;
			}
			else
			{
				// No change
				return;
			}
		}
		break;
	}

	Animated_Line_Entry* pxEntry = new Animated_Line_Entry;

	pxEntry->m_pxDialog		= pxDialog;
	pxEntry->m_xOrigin		= pxDialog->m_xOffset;
	pxEntry->m_xDestination = xDestination;
	
	m_xAnimationList.AddToList( pxEntry );
}


void Tools_Entity_DialogueController::ClearAnimationData ( )
{
	while( !m_xAnimationList.IsEmpty() )
	{
		Animated_Line_Entry* pxEntry = m_xAnimationList.PopHeadObject();
		delete pxEntry;
	}
	m_eAnimation				= ANIMATION_TYPE_NONE;
	m_iAnimationLine			= -1;
	m_fAnimationTimer			= 0.0f;
	m_pxOverlay					= 0;
}


void Tools_Entity_DialogueController::AddOverlayControls_ToggleLine	(Asura_StaticMessage* pxMessage, Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset)
{
	Asura_Msg_Dialogue_ToggleLine* pxData = static_cast<Asura_Msg_Dialogue_ToggleLine*>( pxMessage->m_pData );

	Asura_Dialog* pxDialog = 0;

	Asura_ServerEntity_DialogueController* pxEntity = Asura_ServerEntity_DialogueController::FindEntity( pxMessage->m_uSendTo );

	if( !pxEntity )
	{
		Asura_Assert( false, "Failed to find entity" );
		return;
	}

	pxDialog = new Tools_DialogueLineHash_DropDownList_Dialog( "Dialogue line", xOffset, &pxData->m_uLineID, pxEntity );
	pxDialog->SetHelpText("Select a dialogue line to affect.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog( "Active", xOffset, &pxData->m_bEnable );
	pxDialog->SetHelpText( "Should this line be active" );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();
}


void Tools_Entity_DialogueController::AddOverlayControls_Interrupt	(Asura_StaticMessage* pxMessage, Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset)
{
	Asura_Msg_Dialogue_Interrupt* pxData = static_cast<Asura_Msg_Dialogue_Interrupt*>( pxMessage->m_pData );

	Asura_Dialog* pxDialog = 0;

	pxDialog = new Asura_Tickbox_Dialog( "Immediate", xOffset, &pxData->m_bImmediate );
	pxDialog->SetHelpText( "Stop the current line from playing in addition to any further playback" );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();
}

void Tools_Entity_DialogueController::AddOverlayControls_TriggerLine	(Asura_StaticMessage* pxMessage, Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset)
{
	Asura_Msg_Dialogue_TriggerLine* pxData = static_cast<Asura_Msg_Dialogue_TriggerLine*>( pxMessage->m_pData );

	Asura_Dialog* pxDialog = 0;

	Asura_ServerEntity_DialogueController* pxEntity = Asura_ServerEntity_DialogueController::FindEntity( pxMessage->m_uSendTo );

	if( !pxEntity )
	{
		Asura_Assert( false, "Failed to find entity" );
		return;
	}

	pxDialog = new Tools_DialogueLineHash_DropDownList_Dialog( "Dialogue line", xOffset, &pxData->m_uLineID, pxEntity, "*Not Set*" );
	pxDialog->SetHelpText("Select a dialogue line to trigger.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();
}



// ______________________________ Tools_DialogueLineHash_DropDownList_Dialog ___________________________________


Tools_DialogueLineHash_DropDownList_Dialog::Tools_DialogueLineHash_DropDownList_Dialog	( const char* szName
																						, Asura_Vector_2& xOffset
																						, Asura_Hash_ID* puHash
																						, const Asura_ServerEntity_DialogueController* pxEntity
																						, const char* szZeroHashText )
																						: Asura_Hash_DropDownList_Dialog( xOffset )
{
	m_pxEntity = pxEntity;
	m_szZeroHashText = szZeroHashText;
	
	Initialise( xOffset, 10, puHash, &m_bChanged, szName );
}

u_int Tools_DialogueLineHash_DropDownList_Dialog::GetNumberOfHashValues() const 
{ 
	// Find this dialogue event and return the number of lines
	Asura_Assert( m_pxEntity, "Failed to find dialogue controller" );
	if( m_pxEntity )
	{
		return m_pxEntity->m_xLines.Size();
	}
	return 0;
}

const char* Tools_DialogueLineHash_DropDownList_Dialog::GetNameByIndex(int iIndex) const
{
	// Find this dialogue event and return the number of lines
	Asura_Assert( m_pxEntity, "Failed to find event associated with dialogue controller" );
	if( m_pxEntity )
	{
		const Asura_Dialogue_Line* pxDialogueLine = Asura_Dialogue_Storage::FindLine( m_pxEntity->m_xLines[iIndex].m_uDialogueLine );
		
		if( pxDialogueLine )
		{
			return pxDialogueLine->GetName();
		}
		
	}
	return "Unknown";
}



#endif //ASURA_USE_DIALOGUE_SYSTEM