
////////////////////////////////////////////////////////////////////////////////
//
// Tools_Entity_FMVController.cpp
// Author    : Jason Turner
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM

#include "Entity/Asura_Entity_Event.h"
#include "Entity/Tools_Entity_FMVController.h"
#include "Entity/FMVController/Asura_ServerEntity_FMVController.h"
#include "Entity/FMVController/Asura_ServerEntity_FMVSpeaker.h"
#include "Entity/FMVController/Asura_ClientEntity_FMVController.h"
#include "FMV/Asura_VideoToTexture.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Materials/Tools_Overlay_PickMaterialResponse.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Misc/Tools_Render.h"
#include "Network/Asura_Message.h"

#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Objects/Tools_Object_System.h"

#include "PC/FileSystem/Tools_FileSystem.h"


////////////////////////////////////////////////////////////////////////////////
// Tools_ChooseObject_FMVSpeaker
////////////////////////////////////////////////////////////////////////////////

Tools_ChooseObject_FMVSpeaker::Tools_ChooseObject_FMVSpeaker( Tools_Entity_FMVController* pxSourceObj, bool bAdd) :
		m_pxSourceObj( pxSourceObj ),
		m_bAddRemoveFlag (bAdd)
{
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

bool Tools_ChooseObject_FMVSpeaker::CanBeSelected( Tools_Object* pxObject )
{
	if (pxObject)
	{
		// only allow speaker entities
		if (pxObject->Type == TOT_Entity_FMVSpeaker)
		{
			//
			// If we're Adding - Only allow selection of unlinked speakers
			// If we're Removing - Only allow selection of linked speakers
			//
			Asura_ServerEntity_FMVController* pxControllerEntity = static_cast<Asura_ServerEntity_FMVController*>(m_pxSourceObj->m_pxServerEntity);
			Asura_TemplateTree_ConstIterator<Asura_Guid, Asura_Guid> xIterator(pxControllerEntity->GetSpeakerList());

			while(!xIterator.Done())
			{
				const Asura_Guid* puSpeakerGuid = xIterator.GetCurrent();

				if (puSpeakerGuid)
				{
					if (pxObject->GetGuid() == *puSpeakerGuid)
					{
						return (!m_bAddRemoveFlag);
					}
				}

				xIterator.Next();
			}
			return m_bAddRemoveFlag;
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_ChooseObject_FMVSpeaker::ObjectSelected( Tools_Object* pxObject )
{
	if (pxObject)
	{
		for ( CLIF<Tools_Object*> xIter( GetSelectionList() ); !xIter.done(); xIter.next() )
		{
			if (m_bAddRemoveFlag)
			{
				m_pxSourceObj->AddLinkedSpeaker(xIter()->GetGuid());
			}
			else
			{
				m_pxSourceObj->RemoveLinkedSpeaker(xIter()->GetGuid());
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

// class implementation
Tools_Entity_FMVController::Tools_Entity_FMVController(Asura_Entity* pxEntity)
:Tools_Entity_Activatable_Object(pxEntity)
,m_bNeedsUpdate (false)
{
	Type = TOT_Entity_FMVController;
	Mesh = Asura_Mesh_System::GetMesh("ToolsShape_Cutscene");
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::AddOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset)
{
	PARENT::AddOverlayControls(pxOverlay,xOffset);
	Asura_Dialog* pxDialog;
	char szText[256];
	Asura_Vector_2 xStartOffset = xOffset;

	Asura_ServerEntity_FMVController* pxAsuraEntity = (Asura_ServerEntity_FMVController*) Entity;

	//
	// Dividing line
	//
	pxDialog = new Asura_Divider_Dialog( xOffset );
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	//
	// FMV select button
	//
	{
		//get text for fmv selection control
		if(pxAsuraEntity->m_xFMVName.IsBlank())
		{
			strcpy(szText,"FMV : <Not Set>");
		}
		else
		{
			Asura_PrintfToBfr(szText,"FMV : %s", 100, true, pxAsuraEntity->m_xFMVName.GetString());
		}

		pxDialog = new Asura_Function_Dialog(szText, xOffset, SelectFMV, (int)this);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y += Asura_Dialog::LargeSpacing();
	}

	//
	// Loop tickbox
	//
	{
		Asura_FunctionTickbox_Dialog* pxTickBox = new Asura_FunctionTickbox_Dialog( "Loop", xOffset, this, GetLoopChecked, SetLoopChecked );
		pxTickBox->SetHelpText("At the end, the FMV will loop to the start (if checked)");
		pxOverlay->AddDialog(pxTickBox);
		xOffset.y += Asura_Dialog::LargeSpacing();
	}

	xOffset.x = xStartOffset.x;

	bool bMuted = (pxAsuraEntity->m_uFlags & Asura_VideoToTexture::ASURA_FMV_FLAG_NO_SOUND) ? true : false;

	//
	// Mute Audio tickbox
	//
	{
		Asura_FlagTickbox_Dialog* pxTickBox = new Asura_FlagTickbox_Dialog( "Mute Audio", xOffset, &pxAsuraEntity->m_uFlags, Asura_VideoToTexture::ASURA_FMV_FLAG_NO_SOUND, &m_bNeedsUpdate);
		pxTickBox->SetHelpText("Mutes the embedded FMV audio stream");
		pxOverlay->AddDialog(pxTickBox);
		xOffset.y += Asura_Tickbox_Dialog::DefaultSpacing();
	}

	//
	// Speaker links
	//
	unsigned int uSpeakerTotal = pxAsuraEntity->GetSpeakerList().GetNumberOfNodes();
	
	Asura_PrintfToBfr(szText, "Linked Speakers ( %u / %u )%s:", 256, true, uSpeakerTotal, Asura_ServerEntity_FMVController::MAX_LINKED_SPEAKERS, uSpeakerTotal ? "" : " - Full Volume");
	pxDialog = new Asura_StringDisplay_Dialog(szText, xOffset);
	if (bMuted)
	{
		pxDialog->SetEnabled(false);
	}
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();
	xOffset.x += Asura_Dialog::DefaultSpacing() * 2;

	if (uSpeakerTotal <= Asura_ServerEntity_FMVController::MAX_LINKED_SPEAKERS)
	{
		pxDialog = new Asura_Function_Dialog("Add Linked Speaker", xOffset, OnClickAddSpeakerLink, (int)this);
		pxDialog->SetHelpText("Add A Speaker Link.");
		if (bMuted)
		{
			pxDialog->SetEnabled(false);
		}

		pxOverlay->AddDialog(pxDialog);
		xOffset.y += Asura_Dialog::DefaultSpacing();
	}

	if (uSpeakerTotal > 0)
	{
		pxDialog = new Asura_Function_Dialog("Remove Linked Speaker", xOffset, OnClickRemoveSpeakerLink, (int)this);
		pxDialog->SetHelpText("Remove A Speaker Link.");
		if (bMuted)
		{
			pxDialog->SetEnabled(false);
		}
		pxOverlay->AddDialog(pxDialog);
		xOffset.y += Asura_Dialog::DefaultSpacing();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::Create()
{
	Tools_Object* pxObject = Create(false);
	Tools_Object_System::SelectObject(pxObject);
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

Tools_Entity_FMVController* Tools_Entity_FMVController::Create(bool bGenerateName)
{
	//get a name and the position data
	Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();
	Tools_Entity_Data* pxToolsData = 0;
	if(bGenerateName)
	{
		pxToolsData = Tools_Entity_Object::CreateToolsDataWithDefaultName("FMV Controller","FMV Controller",uGuid,GetNamePrefix());
	}
	else
	{
		pxToolsData = Tools_Entity_Object::CreateToolsData("FMV Controller",0,GetNamePrefix());
	}

	if(!pxToolsData)
	{
		Asura_Assert(0, "Tools_Entity_FMVController::Create() - no tools data");
		return NULL;
	}

	//create the asura data
	Asura_ServerEntity_FMVController* pxServerObject = new Asura_ServerEntity_FMVController(uGuid);
	if(!pxServerObject)
	{
		Asura_Assert(0, "Tools_Entity_FMVController::Create() - no server object" );
		return NULL;
	}

	pxServerObject->ToolsData = pxToolsData;

	Asura_ClientEntity_FMVController* pxClientObject = new Asura_ClientEntity_FMVController(uGuid);
	Asura_Assert(pxClientObject, "Tools_Entity_FMVController::Create() - no client object");

	// Create the tools object
	Tools_Entity_FMVController* pxToolsObject = new Tools_Entity_FMVController(pxServerObject);
	if (!pxToolsObject)
	{
		Asura_Assert(0, "Tools_Entity_FMVController::Create() - no tools object");
		return NULL;
	}

	Tools_Object_System::SelectObject(pxToolsObject);
	Tools_Timed_Message_System::AddMessage(0xffffff, "Created new FMV Controller Entity");
	Tools_SaveAsura::SetEntityModified();

	return pxToolsObject;
}

////////////////////////////////////////////////////////////////////////////////
//
// return true if this entity needs this callback type
////////////////////////////////////////////////////////////////////////////////

bool Tools_Entity_FMVController::HandlesThisEvent(int iEventID) const
{
	if(iEventID == ASURA_ENTITY_EVENT_ONDESTRUCTION ||
		iEventID == ASURA_ENTITY_EVENT_ONACTIVATION ||
		iEventID == ASURA_ENTITY_EVENT_ONDEACTIVATION ||
		iEventID == ASURA_ENTITY_EVENT_ONDEATH
		)
	{
		return true;
	}

	return PARENT::HandlesThisEvent(iEventID);
}

////////////////////////////////////////////////////////////////////////////////
//
// return true if this entity accepts this message
////////////////////////////////////////////////////////////////////////////////

bool Tools_Entity_FMVController::HandlesThisMessage(int iMessageID)
{
	if(	iMessageID == ASURA_MSG_ACTIVATE ||
		iMessageID == ASURA_MSG_DEACTIVATE ||
		iMessageID == ASURA_MSG_FMV_CONTROLLER_PLAY
		)
	{
		return true;
	}

	return PARENT::HandlesThisMessage(iMessageID);
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

bool Tools_Entity_FMVController::GetLoopChecked(void *pData)
{
	Tools_Entity_FMVController* pxToolsEntity = static_cast<Tools_Entity_FMVController*>(pData);
	Asura_ServerEntity_FMVController* pxServerEntity = static_cast<Asura_ServerEntity_FMVController*>(pxToolsEntity->m_pxServerEntity);

	return (pxServerEntity->m_uFlags & Asura_VideoToTexture::ASURA_FMV_FLAG_LOOP);
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::SetLoopChecked(bool bValue, void *pData)
{
	Tools_Entity_FMVController* pxToolsEntity = static_cast<Tools_Entity_FMVController*>(pData);
	Asura_ServerEntity_FMVController* pxServerEntity = static_cast<Asura_ServerEntity_FMVController*>(pxToolsEntity->m_pxServerEntity);
	Asura_ClientEntity_FMVController* pxClientEntity = static_cast<Asura_ClientEntity_FMVController*>(pxToolsEntity->m_pxClientEntity);

	if (bValue)
	{
		pxServerEntity->m_uFlags |= Asura_VideoToTexture::ASURA_FMV_FLAG_LOOP;
		pxClientEntity->m_uFlags |= Asura_VideoToTexture::ASURA_FMV_FLAG_LOOP;
	}
	else 
	{
		pxServerEntity->m_uFlags &= ~Asura_VideoToTexture::ASURA_FMV_FLAG_LOOP;
		pxClientEntity->m_uFlags &= ~Asura_VideoToTexture::ASURA_FMV_FLAG_LOOP;
	}

	pxToolsEntity->SetEntityModified();	
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::SelectFMV()
{
	Tools_FileSystem_DialogInfo	xFileSelector;

	xFileSelector.DefaultExtension		= "*.bik";
	xFileSelector.DefaultFilename		= "";
	xFileSelector.DefaultPath			= NULL;
	xFileSelector.FileFilter			= "Movie files (*.bik; *.wmv; *.avi)\0*.bik; *.wmv; *.avi;\0";
	xFileSelector.WindowTitle			= "Choose source FMV...";

	xFileSelector.AllowMultipleSelect	= false;

	const Asura_String	xSelectedFilename(Tools_FileSystem::GetLoadFilename(xFileSelector));

	if(!xSelectedFilename.IsBlank())
	{
		Asura_String	xWorkingDirectory;
		
		xWorkingDirectory	 = Asura_FileAccess::PlatformGetAsuraFilesDir();
		xWorkingDirectory	+= "\\FMV\\";

		if(xSelectedFilename.StartsWith(xWorkingDirectory.GetString()))
		{
			const u_int		uWorkingDirectoryLength	= xWorkingDirectory.GetLength();
			const u_int		uSelectedFilenameLength	= xSelectedFilename.GetLength();
			const u_int		uRelativeFilenameLength	= uSelectedFilenameLength - uWorkingDirectoryLength;

			Asura_String	xFilename(xSelectedFilename.SubString(uWorkingDirectoryLength, uRelativeFilenameLength));

			xFilename.StripExtension();

			Asura_ServerEntity_FMVController* pxServerEntity = static_cast<Asura_ServerEntity_FMVController*>(m_pxServerEntity);
			Asura_ClientEntity_FMVController* pxClientEntity = static_cast<Asura_ClientEntity_FMVController*>(m_pxClientEntity);

			pxServerEntity->m_xFMVName = xFilename;
			pxClientEntity->m_xFMVName = xFilename;

			m_bNeedsUpdate = true;
		}
		else
		{
			Asura_String	xMessage;

			xMessage	 = "Chosen file is not under \"";
			xMessage	+= xWorkingDirectory;
			xMessage	+= "\").";

			ShowMessageOverlay(xMessage.GetString());
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

Tools_Object* Tools_Entity_FMVController::CreateObjectOfThisType() const
{
	return Create(true);	
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::Copy(const Tools_Object* pxObject)
{
	PARENT::Copy(pxObject);

	const Tools_Entity_FMVController* pxFromObject = (const Tools_Entity_FMVController*) pxObject;

	Asura_ServerEntity_FMVController* pxServerFromEntity = (Asura_ServerEntity_FMVController*) pxFromObject->Entity;
	Asura_ServerEntity_FMVController* pxServerToEntity = (Asura_ServerEntity_FMVController*) Entity;

	pxServerToEntity->m_xFMVName = pxServerFromEntity->m_xFMVName;
	pxServerToEntity->m_uFlags = pxServerFromEntity->m_uFlags;

	Asura_ClientEntity_FMVController* pxClientToEntity = (Asura_ClientEntity_FMVController *) m_pxClientEntity;
	Asura_Assert(pxClientToEntity, "Couldn't find client entity to do copy");

	pxClientToEntity->m_xFMVName = pxServerFromEntity->m_xFMVName;
	pxClientToEntity->m_uFlags = pxServerFromEntity->m_uFlags;
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::Update(void)
{
	if (m_bNeedsUpdate)
	{
		SetEntityModified();	

		//update the FMV controller's overlay
		DeleteOverlay();
		CreateOverlay();

		m_bNeedsUpdate = false;
	}
}
////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::Render(void)
{
	PARENT::Render();

	Asura_ServerEntity_FMVController* pxServerEntity = static_cast<Asura_ServerEntity_FMVController*>(m_pxServerEntity);
	Asura_TemplateTree_Iterator<Asura_Guid, Asura_Guid> xIterator(pxServerEntity->m_xLinkedSpeakerGuids);

	// Render FMV Speaker links
	while(!xIterator.Done())
	{
		Asura_Guid*	puSpeakerGuid = xIterator.GetCurrent();

		if (puSpeakerGuid)
		{
			Asura_ServerEntity_FMVSpeaker* pxSpeaker = Asura_ServerEntity_FMVSpeaker::FindEntity(*puSpeakerGuid);
			if(pxSpeaker)
			{
				Tools_Render::RenderCylinder(Position, pxSpeaker->GetPosition(), 0.02f, 0xaf7fff00, false);
			}
		}

		xIterator.Next();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::OnClickAddSpeakerLink(int iThis)
{
	Tools_Entity_FMVController* pxSourceObj = (Tools_Entity_FMVController*) iThis;
	Asura_Assert (pxSourceObj, "");
	Tools_ChooseObject_FMVSpeaker* pxChooseObject = new Tools_ChooseObject_FMVSpeaker(pxSourceObj, true);
	Tools_Object_System::ChooseObject("Select The Speaker To Add Link", pxChooseObject);
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::OnClickRemoveSpeakerLink(int iThis)
{
	Tools_Entity_FMVController* pxSourceObj = (Tools_Entity_FMVController*) iThis;
	Asura_Assert (pxSourceObj, "");

	Tools_ChooseObject_FMVSpeaker* pxChooseObject = new Tools_ChooseObject_FMVSpeaker(pxSourceObj, false);
	Tools_Object_System::ChooseObject("Select The Speaker To Remove Link", pxChooseObject);
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::AddLinkedSpeaker(Asura_Guid uSpeakerGuid)
{
	Asura_ServerEntity_FMVController* pxControllerEntity = static_cast<Asura_ServerEntity_FMVController*>(m_pxServerEntity);

	if (pxControllerEntity->GetSpeakerList().GetNumberOfNodes() < Asura_ServerEntity_FMVController::MAX_LINKED_SPEAKERS)
	{
		Asura_ServerEntity_FMVSpeaker* pxSpeakerEntity = Asura_ServerEntity_FMVSpeaker::FindEntity(uSpeakerGuid);
		Asura_Assert(pxSpeakerEntity, "Can't find speaker entity!");

		// store speaker Guid in this Controller
		pxControllerEntity->AddLinkedSpeaker(uSpeakerGuid);

		// store this Controller Guid in Speaker object
		pxSpeakerEntity->AddLinkedController(GetGuid());

		m_bNeedsUpdate = true;
	}
}
////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_FMVController::RemoveLinkedSpeaker(Asura_Guid uSpeakerGuid)
{
	Asura_ServerEntity_FMVController* pxControllerEntity = static_cast<Asura_ServerEntity_FMVController*>(m_pxServerEntity);

	if (pxControllerEntity->GetSpeakerList().GetNumberOfNodes() > 0)
	{
		Asura_ServerEntity_FMVSpeaker* pxSpeakerEntity = Asura_ServerEntity_FMVSpeaker::FindEntity(uSpeakerGuid);
		Asura_Assert(pxSpeakerEntity, "Can't find speaker entity!");

		// remove speaker Guid from this Controller
		pxControllerEntity->RemoveLinkedSpeaker(uSpeakerGuid);

		// remove this Controller Guid from Speaker object
		pxSpeakerEntity->RemoveLinkedController(GetGuid());

		m_bNeedsUpdate = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

#endif // ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM
