//
// Tools_EntityTemplate_System.h
//
// Created by Richard Rice - 07/04/2004
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//


#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _TOOLS_ENTITYTEMPLATE_SYSTEM_H_
#define _TOOLS_ENTITYTEMPLATE_SYSTEM_H_


#include "Core/Asura_Maths.h"
#include "Entity/Triggers/Asura_ServerEntity_AdvancedVolumeTrigger.h"
#include "Network/Asura_Guid.h"

#include "misc/list_tem.hpp"


/*
About ApplyChangesForTemplateInstance()...

When an instance of an entity is created from a template file, it is done by first creating the entity
using the normal chunk loader , and then ApplyChangesForTemplateInstance() is used to make any changes
needed to get it in the correct position , and sort out any references the entity has to other objects etc.

IN PARTICULAR , FAILURE TO UPDATE MESSAGE BLOCKS WILL SCREW UP LEVELS BIG TIME.

These things will need updating :

1) Message blocks. Updated using Tools_EntityTemplateInstance::ConvertMessageBlock(). Failure to do this
will result in the entity having a reference to a message block that it doesn't own , which will cause
the stolen message block being deleted with this entity.

2) Sound IDs. Updated using Tools_EntityTemplateInstance::ConvertSoundID(). This will convert entity and
environment sound IDs stored in the template file to values that are valid in the level file

3) Entity Guids. Updated using Tools_EntityTemplateInstance::ConvertGuid().

4) Sound Guids. Updated using Tools_EntityTemplateInstance::ConvertSoundGuid(). Only really of concern
to the sound controller entity.

5) Positions ,orientations and bounding boxes (Assuming they are relative to the world). Updated using TransformPosition(),TransformMatrix(),
TransformQuat() and TransformBoundingBox(). Alternatively GetRelativePosition() and GetRelativeMatrix()
can be used instead to determine how to asjust any positional data.


RAR 29/04/2004
*/

/////////////////////////////////////////////////////////////////
//forward declarations

class Tools_EntityTemplate;
class Tools_EntityTemplateInstance;

class Tools_AsuraFileChunkLister;
class Asura_Chunk_Entity;
class Asura_Chunk_Entity_ToolsData;
class Asura_Chunk_StaticMessages;
class Asura_Chunk_SoundNames;
class Asura_Chunk_Phonons;
class Asura_Chunk_SoundEvent_Phonon;
class Asura_Chunk_GlobalIllumination;
class Asura_Chunk_Corona;
class Asura_ServerEntity_Template;
class Asura_Chunk_TextureNames;
class Asura_Chunk_StreamingBackgroundSound;
struct Asura_StaticMessage;

/////////////////////////////////////////////////////////////////

//Class for storing Template information for all templates used in a level

class Tools_EntityTemplate_System
{
public :
	Tools_EntityTemplate_System();
	virtual ~Tools_EntityTemplate_System();

	//get pointer to single instance of this class
	static Tools_EntityTemplate_System* GetThis() {return s_pxThis;}

	static void Initialise();
	static void Shutdown();

	static Tools_EntityTemplate* GetTemplate(const char* pszName);
	//Forces template to be reloaded the next time it is needed.
	static void InvalidateTemplate(const char* pszName);
	//Forces all templates to be reloaded
	static void InvalidateTemplates();

	//Adds guids of destroyed members of template instances to a list for later recycling.
	//These should be safe to reuse since there shouldn't be any external objects referencing them , 
	//and it should prevent the guids being churned through too quickly whenever templates are updated
	static void RecycleEntityGuid(Asura_Guid uGuid);
	static void RecycleSoundGuid(Asura_Guid uGuid);
	static void RecycleLightID(u_int uID);
	static void RecycleCoronaID(u_int uID);
	static void RecycleStreamingSoundID(u_int uID);

	static Asura_Guid GetNextEntityGuid();
	static Asura_Guid GetNextSoundGuid(bool bMissionFile);
	static u_int GetNextLightID();
	static u_int GetNextCoronaID();
	static u_int GetNextStreamingSoundID();
	
	//virtual functions to be overidden by projects

	//For updating any guids in the extra message data for a static message
	//Derived classes should call the parent version as well , in case it is required for future
	//asura messages.
	virtual void ConvertStaticMessage(Asura_StaticMessage* pxMessage, const Tools_EntityTemplateInstance* pxInstance) const;


private :
	static List<Tools_EntityTemplate*> s_pxTemplateList;
	static List<Asura_Guid> s_xSpareEntityGuids;
	static List<Asura_Guid> s_xSpareMissionSoundGuids;
	static List<Asura_Guid> s_xSpareCommonSoundGuids;
	static List<u_int> s_xSpareLightIDs;
	static List<u_int> s_xSpareCoronaIDs;
	static List<u_int> s_xSpareStreamingSoundIDs;

	static Tools_EntityTemplate_System* s_pxThis;	


};

/////////////////////////////////////////////////////////////////

//Class for storing information about one template used by a level

class Tools_EntityTemplate
{
public :
	Tools_EntityTemplate(const char* pszName);
	~Tools_EntityTemplate();

	const char* GetName() const {return m_pszName;}
	bool IsValid() const {return m_bValid;}

	void CreateTemplateMembers(class Tools_EntityTemplateInstanceMain* pxInstance);


	const Asura_Vector_3& GetPosition() const;
	const Asura_Matrix_3x3& GetOrient() const;
	int GetVersion() const;

	//converts message block from value stored in template to valid value for level
	u_short ConvertMessageBlock(u_short usBlockID) const;

	//gets the index in the guid array for a guid , returning -1 if not found
	int GetIndexFromGuid(Asura_Guid uGuid);
	//
	int GetIndexFromID(enum Tools_EntityTemplate_ChildType eType,u_int uID);

	//convert sound IDs from entity sound IDs in template file to entity sound IDs in the level
	int ConvertSoundID(int iSoundID);

	//gets a list of templates that this template depends on
	bool GetDependencies(List<const char*>& xNameList);

	static bool IsUpdatingFromTemplate(){return s_bUpdatingFromTemplate;}

	static void SetUpdatingFromTemplate(bool bSet){ s_bUpdatingFromTemplate = bSet;}

private :
	void CreateEntities(class Tools_EntityTemplateInstanceMain* pxInstance);
	void CreatePhonons(class Tools_EntityTemplateInstanceMain* pxInstance);
	void CreateLights(class Tools_EntityTemplateInstanceMain* pxInstance);
	void CreateCoronas(class Tools_EntityTemplateInstanceMain* pxInstance);
	#ifdef ASURA_USE_ATTRACTORS
	void CreateAttractors(class Tools_EntityTemplateInstanceMain* pxInstance);
	void UpdateAttractors(class Tools_EntityTemplateInstanceMain* pxInstance);
	#endif
	void CreateStreamingSounds(class Tools_EntityTemplateInstanceMain* pxInstance);
	void UpdateStreamingSounds(class Tools_EntityTemplateInstanceMain* pxInstance);


	void CleanUp();

	char* m_pszName;
	bool m_bValid;

	Tools_AsuraFileChunkLister* m_pxChunkLister;

	const Asura_Chunk_Entity* m_pxTemplateEntity;
	const Asura_Chunk_Entity_ToolsData* m_pxTemplateToolsData;

	List<struct Tools_EntityChunkPair*> m_xEntityList;

//	List<const Asura_Chunk_Entity*> m_xEntityList;
//	List<const Asura_Chunk_Entity_ToolsData*>	m_xToolsDataList;

	const Asura_Chunk_StaticMessages* m_pxMessageChunk;

	List<const Asura_Chunk_SoundNames*> m_xSoundChunkList;
	const Asura_Chunk_SoundEvent_Phonon* m_pxPhononChunk;
	
	#ifdef ASURA_USE_ATTRACTORS
	const class Asura_Chunk_Tools_Attractors* m_pxAttractorChunk;
	#endif


	const Asura_Chunk_GlobalIllumination* m_pxLightChunk;

	const Asura_Chunk_Corona* m_pxCoronaChunk;
	const Asura_Chunk_TextureNames* m_pxCoronaTextureChunk;
	const Asura_Chunk_TextureNames* m_pxEntityTextureChunk;

	const Asura_Chunk_StreamingBackgroundSound* m_pxStreamingSoundChunk;

	struct Tools_EntityTemplateInfo* m_pxTemplateInfo;


	//array of guids of entities in the template file
	int m_iNumGuids;
	Asura_Guid* m_puGuids;

	static bool s_bUpdatingFromTemplate;
	
};

struct Tools_EntityChunkPair
{
	const Asura_Chunk_Entity* m_pxEntityChunk;
	const Asura_Chunk_Entity_ToolsData* m_pxToolsChunk;
};


//Base class for an instance of updating from a template or entity file.
//(Keeping the old name to avoid change everything).
class Tools_EntityTemplateInstance
{
public :
	virtual ~Tools_EntityTemplateInstance(){}

	//Get position and orientation of this instance relative to the original version in the template file
	virtual const Asura_Vector_3& GetRelativePosition() const = 0;
	virtual const Asura_Matrix_3x3& GetRelativeMatrix() const = 0;
	virtual bool ShouldUseRelativeMatrix() const {return true;}

	virtual bool OnlyUpdatesPosition() const {return false;}

	//Returns true if the move is only temporary for recentring on level exports.
	//(Added as an attempt to fix objects becoming unmerged from the env).
	virtual bool IsTemporaryMove() const {return false;}

	//Transform a position in template file to be correct for this instance
	Asura_Vector_3 TransformPosition(const Asura_Vector_3& xPosition) const;
	//Transform a matrix in template file to be correct for this instance
	Asura_Matrix_3x3 TransformMatrix(const Asura_Matrix_3x3& xMat) const;
	//Transform a quat in template file to be correct for this instance
	Asura_Quat TransformQuat(const Asura_Quat& xQuat) const;
	//Transform a euler in template file to be correct for this instance
	Asura_Euler_3 TransformEuler(const Asura_Euler_3 xEuler) const;
	//Transform a bounding box in template file to be correct for this instance
	//Note that since the bounding box is axially aligned , this will only be 
	//approximately correct if the rotation isn't suitable
	Asura_Bounding_Box TransformBoundingBox(const Asura_Bounding_Box& xBox) const;
    // Transform an oriented volume.
    Asura_OrientedVolume TransformOrientedVolume( const Asura_OrientedVolume& xVolume ) const;

	//converts message block from value stored in template to valid value for level
	virtual u_short ConvertMessageBlock(u_short usBlockID) const {return usBlockID;}

	//convert guid in template file to guid of equivalent entity in this instance
	virtual Asura_Guid ConvertGuid(Asura_Guid uGuid) const {return uGuid;}
	//and the same for sound guids
	virtual Asura_Guid ConvertSoundGuid(Asura_Guid uGuid) const {return uGuid;}

	//convert sound IDs from entity sound IDs in template file to entity sound IDs in the level
	virtual int ConvertSoundID(int iSoundID) const {return iSoundID;}

	//convert attractor IDs from IDs in template file to IDs in the level
	//(This won't work on the attractor node IDs yet , but nothing refers to them in the files yet)
	virtual u_int ConvertAttractorID(u_int uAttractorID) const {return uAttractorID;}

	//convert streaming sound ID's from IDs in the template file to IDs in the level
	virtual u_int ConvertStreamingSoundID(u_int uSoundID) const {return uSoundID;}

};


//Stores working data for updating one instance of a template
class Tools_EntityTemplateInstanceMain : public Tools_EntityTemplateInstance
{
	friend class Tools_EntityTemplate;

public :
	Tools_EntityTemplateInstanceMain(Asura_ServerEntity_Template* pxEntity,Tools_EntityTemplate* pxTemplate);


	//Get position and orientation of this instance relative to the original version in the template file
	const Asura_Vector_3& GetRelativePosition() const {return m_xRelPos;}
	const Asura_Matrix_3x3& GetRelativeMatrix()const {return m_xRelOrient;}

	//converts message block from value stored in template to valid value for level
	u_short ConvertMessageBlock(u_short usBlockID) const;

	//convert guid in template file to guid of equivalent entity in this instance
	Asura_Guid ConvertGuid(Asura_Guid uGuid) const;
	//and the same for sound guids
	Asura_Guid ConvertSoundGuid(Asura_Guid uGuid) const;

	//convert sound IDs from entity sound IDs in template file to entity sound IDs in the level
	int ConvertSoundID(int iSoundID) const;

	//convert attractor IDs from IDs in template file to IDs in the level
	//(This won't work on the attractor node IDs yet , but nothing refers to them in the files yet)
	u_int ConvertAttractorID(u_int uAttractorID) const;

	//convert streaming sound ID's from IDs in the template file to IDs in the level
	virtual u_int ConvertStreamingSoundID(u_int uSoundID) const;

	u_int GetTemplatePlatforms() const;
	u_int GetTemplateLightSets() const;
	Asura_Guid GetTemplateGuid() const;
	Asura_Hash_ID GetReplacementSkinID() const;
	bool ShouldMergeWithEnvironment() const;

private :

	//update the array in the template entity
	void SetIDList(enum Tools_EntityTemplate_ChildType eType,const List<u_int>& xIDList);
	class Tools_EntityTemplate_ChildIDArray* GetChildIDArray(enum Tools_EntityTemplate_ChildType eType);

	void UpdateVersion(int iVersion);
	void UpdateInputsAndOutputs(const struct Tools_EntityTemplateInfo* pxTemplateInfo);

	Asura_ServerEntity_Template* m_pxEntity;
	Tools_EntityTemplate* m_pxTemplate;

	Asura_Vector_3 m_xRelPos;
	Asura_Matrix_3x3 m_xRelOrient;

	#ifdef ASURA_USE_ATTRACTORS
	List<class Tools_Attractor*> m_xAttractorList;
	#endif
};

//Variation of template instance that just applies an offset.
//Used by export code to recentre environment
class Tools_EntityTemplateInstance_ApplyOffset : public Tools_EntityTemplateInstance
{
public :
	//Offset is value to be added to entity positions
	Tools_EntityTemplateInstance_ApplyOffset(const Asura_Vector_3& xOffset) {m_xOffset = xOffset;}

	virtual const Asura_Vector_3& GetRelativePosition() const {return m_xOffset;}
	virtual const Asura_Matrix_3x3& GetRelativeMatrix() const {return Asura_Maths::IdentityMatrix3x3;}
	virtual bool ShouldUseRelativeMatrix() const {return false;}

	virtual bool OnlyUpdatesPosition() const {return true;}
	virtual bool IsTemporaryMove() const {return true;}


private :
		Asura_Vector_3 m_xOffset;
};


//Variation of template instance that just converts message blocks.
//Used by the duplication code.
class Tools_EntityTemplateInstance_MessageBlocks : public Tools_EntityTemplateInstance
{
	friend class Tools_EntityTemplate;

public :

	Tools_EntityTemplateInstance_MessageBlocks(Asura_Guid uFromGuid, Asura_Guid uToGuid) { m_uFromGuid = uFromGuid; m_uToGuid = uToGuid; }

	const Asura_Vector_3& GetRelativePosition() const { return Asura_Maths::ZeroVector3; }
	const Asura_Matrix_3x3& GetRelativeMatrix() const { return Asura_Maths::IdentityMatrix3x3; }
	bool ShouldUseRelativeMatrix() const {return false;}

	//converts message block from value stored in template to valid value for level
	u_short ConvertMessageBlock(u_short usBlockID) const;

private:
	Asura_Guid m_uFromGuid;
	Asura_Guid m_uToGuid;
};


#endif //_TOOLS_ENTITYTEMPLATE_SYSTEM_H_
