//
// Tools_GUIMenu_ImportText.h
//
// Created by Tom Perry - 16/09/2008
//
// These classes contain the functionality to import a group of text via an overlay.
// There is an abstract class for these overlays that can be built on.
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_GUIMENU_IMPORTTEXT_
#define __TOOLS_GUIMENU_IMPORTTEXT_

#ifdef ASURA_USE_GUI_MENU_SYSTEM

// Independent
#include "Core/Asura_Hash.h"
#include "Core/Asura_Handle.h"
#include "Interface/Asura_Overlay.h"

// Tools
#include "GUIMenu/Tools_GUIMenu_Root.h"
#include "Misc/Tools_Modal_Overlay.h"

// This is an abstract class for building import overlays onto
class Tools_GUIMenu_ImportText : public Tools_Modal_Overlay
{
public:
	Tools_GUIMenu_ImportText(const Asura_String& xTitle);
	virtual ~Tools_GUIMenu_ImportText();

	// This static function brings up the browser so the user can search for the file he wishes to import the text from
	static void SelectFileToImportText(Tools_GUIMenu_Root* pxToolsParentItem,bool bImportAll, Asura_Vector_2 xPos);

protected:
	// This insert text:
	// xPos = The position the text will be inserted to. This will be increased to be at the bottom left
	//		  of the text, plus any offsets.
	// iCurLevel = This is the current level of the text, and will determine which style is used. The styles are kept
	//			   in an array, and this is the index of that array.
	Tools_GUIMenu_Root* InsertText(Asura_Vector_2 &xPos, int iCurLevel);

	// This inserts a group.
	// xPos = The position the group will be inserted to. This will be increased to be at the bottom left
	//		  of the group, plus any offsets.
	// iCurLevel = This is the current level of the group. Elements from the text list will be added to this group
	//			   until some text with a higher level is found. Then this group will close.
	// bCreateGroup = The group will only be created if this is true. Otherwise the text is added to the menu without
	//				  the enclosing group
	Tools_GUIMenu_Root* InsertGroup(Asura_Vector_2 &xPos, int iCurLevel, bool bCreateGroup = true);

	// This gets the level of the text. The level is determined by a prefix of the ID:
	// (no prefix)  = Default level (0)
	// +			= Level 1
	// ++			= Level 2
	// +++			= Level 3 
	// etc. The highest level is determined by "m_iLevelCap". If a line of text has a higher level then that it is lower
	// to meet the cap.
	int GetTextLevel(Asura_Hash_ID uHashID);

	// This function links all the groups together so make editing the sequence of names easier later
	// pxGroup = A pointer to the group that contains the groups to be linked together
	void LinkGroups(Tools_GUIMenu_Root* pxGroup);

	// These function control the style list which is used on the drop down dialogs in the overlays
	void UpdateStyleList();
	void DeleteStyleList();
	static int SortStyleCallback(const void* pItem1,const void*pItem2);

	// The amount of "levels" of text the overlya caters for
	int m_iLevelCap;

	//for the list of styles
	int m_iNumStyles;
	Tools_GUIMenu_Root** m_ppxStyleItems;
	char** m_pszStyleNames;

	// This contains style choice for each level (as an int, relating to the index of m_ppxStyleItems)
	int* m_aiStyleChoice;

	// Offsets used when importing the text (the displacement between the bottom left and top left of
	// following lines of text)
	float m_fXOffset;
	float m_fYOffset;

	// These are used to see how many text elements we have left to insert
	u_int m_uAmountOfTextToInsert;
	u_int m_uNumberOfTextsInserted;

	// This is the Hash ID of the text we are currently up to when importing
	Asura_Hash_ID m_uHashID;

	// The list of text IDs to inset
	List<Asura_Hash_ID> m_xSelectionList;

	// These variables are static as they are used by the static function of this class:
	// The current position choosen to import the text
	static Asura_Vector_2 s_xPosChoosen;
	// The hash of the current text file being used
	static Asura_Hash_ID s_uTextPage;
	// The parent object
	static Tools_GUIMenu_Root* s_pxToolsParentItem;

private:
	// An abstract function all dervived classes must fill out
	virtual void InitialiseOverlay(class Asura_Overlay* Overlay) =0;
};

class Tools_GUIMenu_ImportText_Overlay : public Tools_GUIMenu_ImportText
{
public:
	Tools_GUIMenu_ImportText_Overlay( const Asura_String& xTitle);
	virtual ~Tools_GUIMenu_ImportText_Overlay();

	// This shows a tool tip for the line we are hovering over. The tool tip shows the text the ID represents.
	static void ShowTooltipForLine(int iIndex, const char* szLineName);

protected:
	virtual void Update();

private:
	virtual void InitialiseOverlay(class Asura_Overlay* Overlay);

	// These relate to the list of text IDs, which can be filtered via a search
	void BuildLineList();
	void DeleteLineList();

	bool DoesLineSatisfySearchFilter(int iIndex, const char* szWildcardSearchString);
	
	Asura_Hash_ID GetLineHash(int iIndex) const;

    static void CancelSelected( void ) { }

	#define OVERLAY_BOOL_CALLBACK(__FnName)\
	static bool __FnName(int iThis){return ((Tools_GUIMenu_ImportText_Overlay*)iThis)->__FnName();}\
	virtual bool __FnName();
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis){((Tools_GUIMenu_ImportText_Overlay*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_BOOL_CALLBACK(WantsGroup);
	OVERLAY_VOID_CALLBACK(OkaySelected)

	#undef OVERLAY_BOOL_CALLBACK
	#undef OVERLAY_VOID_CALLBACK

	// Do we want all the text to be grouped together?
	bool m_bGroupTextTogether;

	// These variabels all relate to the list of IDs:
	// The search text and option for filtering
	Asura_String m_xSearchText;
	int m_iSearchOption;
	bool m_bFilterChanged;
	bool m_bLineChanged;
	// Do we want the list sorted or not?
	bool m_bSortList;
	bool m_bPrevSortList;
	int m_iLineIndex;
	Asura_Hash_ID m_uMostRecentSelectedLine;//hash of most recent line selected by user
	Asura_Hash_ID m_uSelectedLine;//hash of currently selected line
	//list of line names
	u_int m_uLineListSize;
	const char** m_apszLineList;
	class Asura_StringTree* m_pxLineStringTree;

	class Asura_List_Dialog* m_pxLineListDialog;
};

class Tools_GUIMenu_ImportAllText_Overlay : public Tools_GUIMenu_ImportText
{
public:
	Tools_GUIMenu_ImportAllText_Overlay( const Asura_String& xTitle);
	virtual ~Tools_GUIMenu_ImportAllText_Overlay();

private:
	virtual void InitialiseOverlay(class Asura_Overlay* Overlay);

	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis){((Tools_GUIMenu_ImportAllText_Overlay*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(OkaySelected)

	#undef OVERLAY_VOID_CALLBACK

    static void CancelSelected( void ) {  }
};

#endif // ASURA_USE_GUI_MENU_SYSTEM
#endif //__TOOLS_GUIMENU_IMPORTTEXT_OVERLAY__
