//
// Tools_Hierarchy_BodyPart_System.cpp
//
// Created by Richard Rice - 9/10/02
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Hierarchy/Tools_Hierarchy_BodyPart_System.h"

#include "Animation/Asura_Hierarchy_BodyPart_System.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

#include "Misc/Tools_IniFile.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_EnumerationReader.h"

#include "Load&Save/LevelExport/Tools_Level_Export.h"


//empty macro for displaying errors... because I haven't decided where to put them yet
#define INIFILE_ERROR(x) ShowIniFileError x

void ShowIniFileError(const char* pszFormat,...)
{
	char szBuffer[200];

	va_list ap;
	va_start(ap, pszFormat);
	Asura_PrintfToBfr(szBuffer,pszFormat,ap,200);
	va_end(ap);

	//chuck the errors through the level export system at the moment
	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_ERROR,"%s",szBuffer);

}


////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Tools_Hierarchy_BodyPart_System **
**---------------------------------*/

Tools_Hierarchy_BodyPart_System* Tools_Hierarchy_BodyPart_System::s_pxThis;

Tools_Hierarchy_BodyPart_System::Tools_Hierarchy_BodyPart_System()
{
	s_pxThis = this;
	m_pxBodyPartEnum = 0;
	m_pxDeathTypeEnum = 0;
}

Tools_Hierarchy_BodyPart_System::~Tools_Hierarchy_BodyPart_System()
{
	s_pxThis = 0;
}


void Tools_Hierarchy_BodyPart_System::_ReadIniFile()
{
		
	//load the ini file
	const char* pszFilename = GetIniFileName();
	Asura_Assert(pszFilename,"");

	Tools_IniFile xIniFile(pszFilename);
	if(!xIniFile.LoadIniFile())
	{
		INIFILE_ERROR(("Failed to load %s",pszFilename));
		return;
	}

	//read in the enumeration strings
	pszFilename = GetEnumFileName();
	Asura_Assert(pszFilename,"");
	m_pxBodyPartEnum = new Tools_EnumerationReader;
	m_pxDeathTypeEnum = new Tools_EnumerationReader;

	if(!m_pxBodyPartEnum->ReadEnumeration(pszFilename,"enum AsuraBodyPartType"))
	{
		INIFILE_ERROR(("%s",m_pxBodyPartEnum->GetError()));
	}
	
	if(!m_pxDeathTypeEnum->ReadEnumeration(pszFilename,"enum AsuraDeathType"))
	{
		INIFILE_ERROR(("%s",m_pxDeathTypeEnum->GetError()));
	}

	//process the inifile
	Asura_Hierarchy_BodyPart_System::LoadFromIniFile(&xIniFile);


	delete m_pxBodyPartEnum;
	m_pxBodyPartEnum = 0;
	delete m_pxDeathTypeEnum;
	m_pxDeathTypeEnum = 0;
}


void Tools_Hierarchy_BodyPart_System::AddAnimation(const char* pszName)
{
	Asura_Assert(s_pxThis,"");
	Asura_Assert(pszName,"");

	//get a filename from the animation name
	Asura_String xFilename;
	if(Tools_SaveAsura::FindAnimationRecursive("Characters",pszName,xFilename))
	{
		int iJunkIndex;
		Asura_Hierarchy_BodyPart_System::s_pxAnimationTree->FindNodeIndexByName(xFilename,true,iJunkIndex);
	}
	else
	{
		INIFILE_ERROR(("Failed to locate animation file for %s",pszName));
	}

}

int Tools_Hierarchy_BodyPart_System::GetBodyPartEnum(const char* pszName)
{
	Asura_Assert(s_pxThis,"");
	Asura_Assert(s_pxThis->m_pxBodyPartEnum,"");
	Asura_Assert(pszName,"");

	int iIndex = ASURA_BODYPART_MAX;
	if(!s_pxThis->m_pxBodyPartEnum->GetEnumerationValue(pszName,iIndex))
	{
		INIFILE_ERROR(("Failed to find enumeration value for body part %s",pszName));
	}

	return iIndex;
}

int Tools_Hierarchy_BodyPart_System::GetDeathTypeEnum(const char* pszName)
{
	Asura_Assert(s_pxThis,"");
	Asura_Assert(s_pxThis->m_pxDeathTypeEnum,"");
	Asura_Assert(pszName,"");

	int iIndex = ASURA_BODYPART_MAX;
	if(!s_pxThis->m_pxDeathTypeEnum->GetEnumerationValue(pszName,iIndex))
	{
		INIFILE_ERROR(("Failed to find enumeration value for death type %s",pszName));
	}

	return iIndex;
}


////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Asura_Hierarchy_BodyPart_System **
**---------------------------------*/

void Asura_Hierarchy_BodyPart_System::LoadFromIniFile(Tools_IniFile* pxIniFile)
{
	Asura_Assert(pxIniFile,"");
	Reset();

	//get a list of root sections , each section should correspond to a different body type
	List<char*> xSectionList = pxIniFile->GetSectionNameList();
	if(!xSectionList.size())
	{
		INIFILE_ERROR(("No body types found"));
		return;
	}

	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		s_iNumberOfBodyTypes = xSectionList.size();
		Asura_ResourceAllocateArray(s_pxBodyTypes, s_iNumberOfBodyTypes);

		LIF<char*> xSectionLIF(&xSectionList);
		for(int i=0;!xSectionLIF.done();xSectionLIF.next(),i++)
		{
			//step into the next section and read in the data for the body type
			pxIniFile->GetIniTextSection(xSectionLIF(),true);
			s_pxBodyTypes[i].LoadFromIniFile(pxIniFile);
			pxIniFile->GetParentSection();
		}
	}
	Asura_ResourcePopType();
}

////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Asura_Hierarchy_BodyType **
**--------------------------*/

void Asura_Hierarchy_BodyType::LoadFromIniFile(Tools_IniFile* pxIniFile)
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		//read in the list of skins
		List<Asura_Hash_ID> xSkinList;
		while(pxIniFile->GetNextIniTextLine())
		{
			Asura_String xTemp;
			if(pxIniFile->GetString(xTemp))
			{
				Asura_Hash_ID uSkinHash = Asura_GetHashID(xTemp.GetString());
				if(!xSkinList.contains(uSkinHash))
				{
					xSkinList.add_entry(uSkinHash);
				}
				else
				{
					INIFILE_ERROR(("Skin %s appears more than once for %s",xTemp.GetString(),pxIniFile->GetCurrentSectionName()));
				}
			}
		}
		//copy the list into an array
		m_iNumberOfSkins = xSkinList.size();
		Asura_ResourceAllocateArray(m_puSkins, m_iNumberOfSkins);

		LIF<Asura_Hash_ID> xSkinLIF(&xSkinList);
		for(int i=0;i<m_iNumberOfSkins;i++,xSkinLIF.next())
		{
			m_puSkins[i] = xSkinLIF();
		}

		//The subsections should correspond the different body parts
		List<char*> xSectionList = pxIniFile->GetSectionNameList();

		if(xSectionList.size()<1)
		{
			INIFILE_ERROR(("No body part entries found for %s",pxIniFile->GetCurrentSectionName()));
		}
		else
		{
			m_iNumberOfBodyParts = xSectionList.size();
			Asura_ResourceAllocateArray(m_pxBodyParts, m_iNumberOfBodyParts);

			LIF<char*> xSectionLIF(&xSectionList);
			for(int i=0;i<m_iNumberOfBodyParts && !xSectionLIF.done();i++,xSectionLIF.next())
			{

				pxIniFile->GetIniTextSection(xSectionLIF(),true);
				m_pxBodyParts[i].LoadFromIniFile(pxIniFile);
				pxIniFile->GetParentSection();
			}
		}
	}
	Asura_ResourcePopType();
}

////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Asura_Hierarchy_BodyPart **
**--------------------------*/

void Asura_Hierarchy_BodyPart::LoadFromIniFile(Tools_IniFile* pxIniFile)
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		m_iBodyPartType = Tools_Hierarchy_BodyPart_System::GetBodyPartEnum(pxIniFile->GetCurrentSectionName());

		//read in the list of hierarchy sections
		List<Asura_Hash_ID> xHashList;
		while(pxIniFile->GetNextIniTextLine())
		{
			Asura_String xTemp;
			if(pxIniFile->GetString(xTemp))
			{
				Asura_Hash_ID uSectionHash = Asura_GetHashID(xTemp.GetString());
				if(!xHashList.contains(uSectionHash))
				{
					xHashList.add_entry(uSectionHash);
				}
				else
				{
					INIFILE_ERROR(("Section %s appears more than once for %s",xTemp.GetString(),pxIniFile->GetCurrentSectionName()));
				}
			}
		}
		//copy the list into an array
		m_iNumberOfSections = xHashList.size();
		Asura_ResourceAllocateArray(m_puSections, m_iNumberOfSections);

		LIF<Asura_Hash_ID> xHashLIF(&xHashList);
		for(int i=0;i<m_iNumberOfSections;i++,xHashLIF.next())
		{
			m_puSections[i] = xHashLIF();
		}

		//The subsections should correspon to different death animations
		List<char*> xSectionList = pxIniFile->GetSectionNameList();
		if(xSectionList.size())
		{
			m_iNumberOfDeathTypes = xSectionList.size();
			Asura_ResourceAllocateArray(m_pxDeathTypes, m_iNumberOfDeathTypes);

			LIF<char*> xSectionLIF(&xSectionList);
			for(int i=0;i<m_iNumberOfDeathTypes && !xSectionLIF.done();i++,xSectionLIF.next())
			{

				pxIniFile->GetIniTextSection(xSectionLIF(),true);
				m_pxDeathTypes[i].LoadFromIniFile(pxIniFile);
				pxIniFile->GetParentSection();
			}
		}
	}
	Asura_ResourcePopType();
}

////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Asura_Hierarchy_DeathType **
**---------------------------*/

void Asura_Hierarchy_DeathType::LoadFromIniFile(Tools_IniFile* pxIniFile)
{
	Asura_ResourcePushType(ASURA_RESOURCE_ANIMATION);
	{
		m_iDeathType = Tools_Hierarchy_BodyPart_System::GetDeathTypeEnum(pxIniFile->GetCurrentSectionName());

		//read in the list of animations
		List<Asura_Hash_ID> xHashList;
		while(pxIniFile->GetNextIniTextLine())
		{
			Asura_String xTemp;
			if(pxIniFile->GetString(xTemp))
			{
				Asura_Hash_ID uAnimHash = Asura_GetHashID(xTemp.GetString());
				if(!xHashList.contains(uAnimHash))
				{
					xHashList.add_entry(uAnimHash);

					//add to the global list of all animations used as well
					Tools_Hierarchy_BodyPart_System::AddAnimation(xTemp.GetString());
				}
				else
				{
					INIFILE_ERROR(("Animation %s appears more than once for %s",xTemp.GetString(),pxIniFile->GetCurrentSectionName()));
				}
			}
		}
		//copy the list into an array
		m_iNumberOfAnimations = xHashList.size();
		Asura_ResourceAllocateArray(m_puAnimations, m_iNumberOfAnimations);

		LIF<Asura_Hash_ID> xHashLIF(&xHashList);
		for(int i=0;i<m_iNumberOfAnimations;i++,xHashLIF.next())
		{
			m_puAnimations[i] = xHashLIF();
		}
	}
	Asura_ResourcePopType();
}
