//
// Tools_IPImageEx.cpp
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "ImageProcessing/Tools_IPImageEx.h"

#include <string.h>
//#include <stdio.h>

#include "ImageProcessing/Tools_Histogram.h"
#include "ImageProcessing/Tools_Paletize.h"
#include "ImageProcessing/Tools_ColourMap.h"
#ifndef M_PI
	#define M_PI 3.1415926535897932384626433832795
#endif

// Standard filters

static double IPDefaultFilterFunc(double t)
{
	/* f(t) = 2|t|^3 - 3|t|^2 + 1, -1 <= t <= 1 */
	if(t < 0.0) t = -t;
	if(t < 1.0) return((2.0 * t - 3.0) * t * t + 1.0);
	return(0.0);
}
IPFilter const ip_filtDefault(1.0, IPDefaultFilterFunc);

static double IPBoxFilterFunc(double t)
{
    if((t > -0.5) && (t <= 0.5)) return(1.0);
    return(0.0);
}
IPFilter const ip_filtBox(0.51, IPBoxFilterFunc);

static double IPTriangleFilterFunc(double t)
{
    if(t < 0.0) t = -t;
    if(t < 1.0) return(1.0 - t);
    return(0.0);
}
IPFilter const ip_filtTriangle(1.0, IPTriangleFilterFunc);

static double IPBellFilterFunc(double t)
{
	/* box (*) box (*) box */
    if(t < 0.) t = -t;
    if(t < .5) return(.75 - (t * t));
    if(t < 1.5) {
        t = (t - 1.5);
        return(.5 * (t * t));
    }
    return(0.0);
}
IPFilter const ip_filtBell(1.5, IPBellFilterFunc);

static double IPBSplineFilterFunc(double t)
{
	/* box (*) box (*) box (*) box */
    double tt;

    if(t < 0.) t = -t;
    if(t < 1.) {
        tt = t * t;
        return((.5 * tt * t) - tt + (2.0 / 3.0));
    } else if(t < 2.) {
        t = 2. - t;
        return((1.0 / 6.0) * (t * t * t));
    }
    return(0.0);
}
IPFilter const ip_filtBSpline(2.0, IPBSplineFilterFunc);

static double IPL3SinC(double x)
{
    x *= M_PI;
    if(x != 0.) return(sin(x) / x);
    return(1.0);
}
static double IPLanczos3FilterFunc(double t)
{
    if(t < 0.) t = -t;
    if(t < 3.0) return(IPL3SinC(t) * IPL3SinC(t/3.0));
    return(0.0);
}
IPFilter const ip_filtLanczos3(3.0, IPLanczos3FilterFunc);

static double IPMitchellFilterFunc(double t)
{
	double const B = 1.0/3.0;
	double const C = 1.0/3.0;

    double tt;

    tt = t * t;
    if(t < 0.) t = -t;
    if(t < 1.0) {
        t = (((12.0 - 9.0 * B - 6.0 * C) * (t * tt))
             + ((-18.0 + 12.0 * B + 6.0 * C) * tt)
             + (6.0 - 2. * B));
        return(t / 6.0);
    } else if(t < 2.0) {
        t = (((-1.0 * B - 6.0 * C) * (t * tt))
             + ((6.0 * B + 30.0 * C) * tt)
             + ((-12.0 * B - 48.0 * C) * t)
             + (8.0 * B + 24. * C));
        return(t / 6.0);
    }
    return(0.0);
}
IPFilter const ip_filtMitchell(2.0, IPMitchellFilterFunc);

void Tools_IPImageEx::ClearHist() const
{
	delete pHist;
	pHist = NULL;
}

void Tools_IPImageEx::MakeHist() const
{
	pHist = new HHHashHistogram;
	
	if (NOT_PALETTIZED == GetPaletteSize())
	{
		if (HasMask())
		{
			for (unsigned y=0; y<GetHeight(); ++y)
				for (unsigned x=0; x<GetWidth(); ++x)
				{
					if (GetColourRow(y)[x] != GetMaskColour())
						*pHist += HHPixel(GetColourRow(y)[x],1);
				}
		}
		else
		{
			for (unsigned y=0; y<GetHeight(); ++y)
				for (unsigned x=0; x<GetWidth(); ++x)
					*pHist += HHPixel(GetColourRow(y)[x],1);
		}
	}
	else
	{
		if (HasMask())
		{
			for (unsigned y=0; y<GetHeight(); ++y)
				for (unsigned x=0; x<GetWidth(); ++x)
				{
					if (GetIndexRow(y)[x] != GetMaskIndex())
						*pHist += HHPixel(GetPalette()[GetIndexRow(y)[x]],1);
				}
		}
		else
		{
			for (unsigned y=0; y<GetHeight(); ++y)
				for (unsigned x=0; x<GetWidth(); ++x)
					*pHist += HHPixel(GetPalette()[GetIndexRow(y)[x]],1);
		}
	}
	
	bModified = false;
}

unsigned Tools_IPImageEx::CountColoursUsed() const
{
	if (bModified && pHist) ClearHist();
		
	if (!pHist) MakeHist();
	
	return static_cast<unsigned>(pHist->NumEntries() + (HasMask() ? 1 : 0));
}

bool Tools_IPImageEx::CreateByChangingColourDepth(Tools_IPImageEx const & src, unsigned psize)
{
	if (!src.IsValid() || &src == this) return false;
	
	CreateNew(src.GetWidth(),src.GetHeight(),psize);
	
	if (NOT_PALETTIZED == psize)
	{
		if (NOT_PALETTIZED == src.GetPaletteSize())
		{
			for (unsigned y=0; y<GetHeight(); ++y)
				memcpy(GetColourRow(y),src.GetColourRow(y),GetWidth()*sizeof(IPColour));
			
			if (src.HasMask())
			{
				SetMaskColour(src.GetMaskColour());
			}
		}
		else
		{
			for (unsigned y=0; y<GetHeight(); ++y)
				for (unsigned x=0; x<GetWidth(); ++x)
					GetColourRow(y)[x] = src.GetPalette()[src.GetIndexRow(y)[x]];
					
			if (src.HasMask())
			{
				SetMaskColour(src.GetPalette()[src.GetMaskIndex()]);
			}
		}
	}
	else if (psize >= src.GetPaletteSize() && src.GetPaletteSize() != NOT_PALETTIZED)
	{
		memcpy(GetPalette(),src.GetPalette(),src.GetPaletteSize()*sizeof(IPColour));
		
		for (unsigned y=0; y<GetHeight(); ++y)
			memcpy(GetIndexRow(y),src.GetIndexRow(y),GetWidth()*sizeof(IPIndex));
			
		if (src.HasMask())
		{
			SetMaskIndex(src.GetMaskIndex());
		}
	}
	else
	{
		if (src.bModified && src.pHist) src.ClearHist();
		
		if (!src.pHist) src.MakeHist();
		
		PG_ColourOcTree paltree;
		
		for (unsigned long nc = src.pHist->BeginRead(); nc; --nc)
		{
			HHPixel const & rPix = src.pHist->ReadPixelAndAdvance();
			paltree.Store(rPix,rPix.n);
		}
		src.pHist->EndRead();
		paltree.Correct_Statistics();
		
		PG_Pixel * aPalette = new PG_Pixel [psize];
		unsigned nPalOffset = src.HasMask() ? 1 : 0;
		
		psize = paltree.Compute_Palette(aPalette + nPalOffset, psize - nPalOffset);
		
		for (unsigned i=nPalOffset; i<psize+nPalOffset; ++i)
			GetPalette()[i] = aPalette[i];
			
		delete[] aPalette;
		
		Tools_ColourMap colmap(GetPalette() + nPalOffset, psize);
		
		if (NOT_PALETTIZED == src.GetPaletteSize())
		{
			if (src.HasMask())
			{
				for (unsigned y=0; y<GetHeight(); ++y)
					for (unsigned x=0; x<GetWidth(); ++x)
					{
						if (src.GetColourRow(y)[x] == src.GetMaskColour())
							GetIndexRow(y)[x] = 0;
						else
							GetIndexRow(y)[x] = static_cast<IPIndex>(colmap.NearestColour(src.GetColourRow(y)[x])) + 1;
					}
					
				GetPalette()[0] = src.GetMaskColour();
			}
			else
			{
				for (unsigned y=0; y<GetHeight(); ++y)
					for (unsigned x=0; x<GetWidth(); ++x)
						GetIndexRow(y)[x] = static_cast<IPIndex>(colmap.NearestColour(src.GetColourRow(y)[x]));
			}
		}
		else
		{
			if (src.HasMask())
			{
				for (unsigned y=0; y<GetHeight(); ++y)
					for (unsigned x=0; x<GetWidth(); ++x)
					{
						if (src.GetIndexRow(y)[x] == src.GetMaskIndex())
							GetIndexRow(y)[x] = 0;
						else
							GetIndexRow(y)[x] = static_cast<IPIndex>(colmap.NearestColour(src.GetPalette()[src.GetIndexRow(y)[x]])) + 1;
					}
						
				GetPalette()[0] = src.GetPalette()[src.GetMaskIndex()];
			}
			else
			{
				for (unsigned y=0; y<GetHeight(); ++y)
					for (unsigned x=0; x<GetWidth(); ++x)
						GetIndexRow(y)[x] = static_cast<IPIndex>(colmap.NearestColour(src.GetPalette()[src.GetIndexRow(y)[x]]));
			}
		}
			
		if (src.HasMask())
		{
			SetMaskIndex(0);
		}
	}
	return true;
}

bool Tools_IPImageEx::EnsurePalettized()
{
	if (!IsValid()) return false;
	
	if (NOT_PALETTIZED != GetPaletteSize()) return true;
	
	if (bModified && pHist) ClearHist();
	
	if (!pHist) MakeHist();

	Tools_IPImage imgTmp;

	MoveTo(&imgTmp);
	
	unsigned nPalStart = imgTmp.HasMask() ? 1U : 0U;
	
	CreateNew(imgTmp.GetWidth(),imgTmp.GetHeight(),pHist->NumEntries()+nPalStart);
	
	for (unsigned i=0; i<pHist->NumEntries(); ++i)
	{
		GetPalette()[i+nPalStart] = pHist->GetNthEntry(i);
	}
	
	for (unsigned y=0; y<GetHeight(); ++y)
		for (unsigned x=0; x<GetWidth(); ++x)
			GetIndexRow(y)[x] = pHist->GetEntryIndex(imgTmp.GetColourRow(y)[x])+nPalStart;

	if (imgTmp.HasMask())
	{
		GetPalette()[0] = imgTmp.GetMaskColour();
		SetMaskIndex(0);
	}

	imgTmp.MoveAlphaTo(this);

	bModified = false;
	
	return true;
}

bool Tools_IPImageEx::RemoveUnusedColours()
{
	if (!IsValid()) return false;
	
	// cannot remove colours if not palettized
	if (NOT_PALETTIZED == GetPaletteSize()) return true;
	
	// if the histogram is valid and has the same number of entries as the palette size
	// then there are no unused colours
	if (!bModified && pHist && pHist->NumEntries()==GetPaletteSize()) return true;
	
	bool * abClrIndecesUsed = new bool [GetPaletteSize()];
	
	{for (unsigned i=0; i<GetPaletteSize(); ++i)
	{
		abClrIndecesUsed[i] = false;
	}}
	
	if (HasMask())
		abClrIndecesUsed[GetMaskIndex()]=true;
		
	{for (unsigned y=0; y<GetHeight(); ++y)
		for (unsigned x=0; x<GetWidth(); ++x)
			abClrIndecesUsed[GetIndexRow(y)[x]] = true;
	}
			
	IPIndex * aiClrIndexMap = new IPIndex [GetPaletteSize()];
	
	unsigned nNewPaletteSize = 0;
	
	for (unsigned i=0; i<GetPaletteSize(); ++i)
	{
		aiClrIndexMap[i] = nNewPaletteSize;
		
		if (abClrIndecesUsed[i])
		{
			GetPalette()[nNewPaletteSize++] = GetPalette()[i];
		}
	}
	
	delete [] abClrIndecesUsed;
	
	if (nNewPaletteSize == GetPaletteSize())
	{
		// no changes required if the palette size is not changing
		delete [] aiClrIndexMap;
		return true;
	}
	
	if (HasMask())
		SetMaskIndex(aiClrIndexMap[GetMaskIndex()]);
		
	for (unsigned y=0; y<GetHeight(); ++y)
		for (unsigned x=0; x<GetWidth(); ++x)
			GetIndexRow(y)[x] = aiClrIndexMap[GetIndexRow(y)[x]];
			
	delete [] aiClrIndexMap;
	
	return SetPaletteSize(nNewPaletteSize);
}

class IPFilterSourcingTable
{
	public:
		class DstPixel
		{
			public:
				class SrcPixel
				{
					public:
						SrcPixel() : m_pdblWeight(new double) {}
						SrcPixel(SrcPixel const & r) : m_pdblWeight(new double(*r.m_pdblWeight)), m_uIndex(r.m_uIndex) {}
						SrcPixel & operator = (SrcPixel const & r) { m_uIndex = r.m_uIndex; *m_pdblWeight = *r.m_pdblWeight; return *this; }
						~SrcPixel() { delete m_pdblWeight; }
					
						unsigned GetIndex() const { return m_uIndex; }
						void SetIndex(unsigned i) { m_uIndex = i; }
						double const & GetWeight() const { return *m_pdblWeight; }
						void SetWeight(double dbl) const { *m_pdblWeight = dbl; }
						
					private:
						unsigned m_uIndex;
						double * m_pdblWeight; // so struct size is 8
				};
			
				DstPixel() : m_pSources(NULL), m_uNumSources(0) {}
				~DstPixel() { if (m_pSources) delete[] m_pSources; }
			
				unsigned GetNumContributors() const { return m_uNumSources; }
				SrcPixel const & GetContributor(unsigned i) const { return m_pSources[i]; }
				
				void Create(unsigned uSrcLeft, unsigned uSrcRight, double dblSrcCentMinusHalf, double dblScale, IPFilter::Func pfnFilter)
				{
					if (m_pSources) delete[] m_pSources;

					if (uSrcRight < uSrcLeft)
					{
						m_pSources = NULL;
						m_uNumSources = 0;
						return;
					}
					
					m_uNumSources = uSrcRight+1 - uSrcLeft;
					m_pSources = new SrcPixel[m_uNumSources];
					
					double dblTotalWeight = 0.0;
					
					for (unsigned j=uSrcLeft; j<=uSrcRight; ++j)
					{
						double dblWeight = pfnFilter((dblSrcCentMinusHalf-double(j))*dblScale);
						dblTotalWeight += dblWeight;
						m_pSources[j-uSrcLeft].SetIndex(j);
						m_pSources[j-uSrcLeft].SetWeight(dblWeight);
					}
					
					double dblInvWeight = 1.0 / dblTotalWeight;
					
					for (unsigned i=0; i<m_uNumSources; ++i)
					{
						m_pSources[i].SetWeight(m_pSources[i].GetWeight()*dblInvWeight);
					}
				}
				
			private:
				DstPixel(DstPixel const &);
				DstPixel & operator = (DstPixel const &);
			
				SrcPixel * m_pSources;
				unsigned m_uNumSources;
		};
		
		IPFilterSourcingTable() : m_pContributions(NULL) {}
		~IPFilterSourcingTable() { if (m_pContributions) delete[] m_pContributions; }
		
		DstPixel const & GetSources(unsigned uPixel) const { return m_pContributions[uPixel]; }

		void Create(unsigned uDestWidth, unsigned uSrcWidth, double dblScale, IPFilter const & filter)
		{
			if (m_pContributions) delete[] m_pContributions;
			
			m_pContributions = new DstPixel[uDestWidth];
			
			double dblInvScale = 1.0 / dblScale;
			double dblClampedScale = dblScale > 1.0 ? 1.0 : dblScale;
			
			double dblSrcFilterWidthMinusOne = filter.dblWidth / dblClampedScale;
			
			for (unsigned i=0; i<uDestWidth; ++i)
			{
				double dblSrcCentMinusHalf = (double(i) + 0.5) * dblInvScale - 0.5;
				
				int iLeft = int(floor(dblSrcCentMinusHalf - dblSrcFilterWidthMinusOne));
				int iRight = int(ceil(dblSrcCentMinusHalf + dblSrcFilterWidthMinusOne));
				
				m_pContributions[i].Create(
					iLeft < 0 ? 0U : unsigned(iLeft) >= uSrcWidth ? uSrcWidth-1U : unsigned(iLeft),
					iRight < 0 ? 0U : unsigned(iRight) >= uSrcWidth ? uSrcWidth-1U : unsigned(iRight),
					dblSrcCentMinusHalf,
					dblClampedScale,
					filter.pfnFunc
				);
			}
		}
		
	private:
		IPFilterSourcingTable(IPFilterSourcingTable const &);
		IPFilterSourcingTable & operator = (IPFilterSourcingTable const &);
		
		DstPixel * m_pContributions;
};

bool Tools_IPImageEx::CreateByScaling(Tools_IPImageEx const & src, unsigned width, unsigned height, float xScale, float yScale, IPFilter const & filter)
{
	if (!src.IsValid() || &src == this) return false;
	
	IPFilterSourcingTable fst;
	
	// Create temporary to hold transpose of width-scaled image
	
	Tools_IPImage imgTmp;
	
	imgTmp.CreateNew(src.GetHeight(), width, NOT_PALETTIZED);
	
	// Create sourcing table for scaling width
	
	fst.Create(width, src.GetWidth(), xScale, filter);
	
	// Possible mask for the temporary

	bool * * aabTmpMask;
	
	if (src.HasMask())
	{
		// -- SORRY, there's some cutting and pasting gone on here :-(

		aabTmpMask = new bool * [imgTmp.GetHeight()];
		if (!aabTmpMask) return false;
		*aabTmpMask = new bool [imgTmp.GetWidth() * imgTmp.GetHeight()];
		if (!*aabTmpMask)
		{
			delete[] aabTmpMask;
			return false;
		}
		for (unsigned i=1; i<imgTmp.GetHeight(); ++i)
		{
			aabTmpMask[i] = &aabTmpMask[i-1][imgTmp.GetWidth()];
		}
		
		if (src.IsPalettized())
		{
			for (unsigned ySrc=0; ySrc<src.GetHeight(); ++ySrc)
			{
				IPIndex const * pSrcRow = src.GetIndexRow(ySrc);
				
				for (unsigned yDst=0; yDst<imgTmp.GetHeight(); ++yDst)
				{
					IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
					
					double dblRSum = 0.0;
					double dblGSum = 0.0;
					double dblBSum = 0.0;
					double dblNonMaskWeight = 1.0;
					bool bAnyMasks = false;
					
					for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
					{
						IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
						
						IPIndex ixPalette = pSrcRow[spContributor.GetIndex()];
						
						if (src.GetMaskIndex() == ixPalette)
						{
							bAnyMasks = true;
							dblNonMaskWeight -= spContributor.GetWeight();
						}
						else
						{
							IPColour const & colour = src.GetPalette()[ixPalette];
							
							dblRSum += double(colour.r) * spContributor.GetWeight();
							dblGSum += double(colour.g) * spContributor.GetWeight();
							dblBSum += double(colour.b) * spContributor.GetWeight();
						}
					}
					
					if (bAnyMasks)
					{
						if (dblNonMaskWeight <= 0.5)
						{
							aabTmpMask[yDst][ySrc] = true;
							continue;
						}
						
						double dblInvNonMaskWeight = 1.0 / dblNonMaskWeight;
						dblRSum *= dblInvNonMaskWeight;
						dblGSum *= dblInvNonMaskWeight;
						dblBSum *= dblInvNonMaskWeight;
					}
					
					aabTmpMask[yDst][ySrc] = false;
					
					IPColour & colPixel = imgTmp.GetColourRow(yDst)[ySrc];
					
					colPixel.r = (unsigned char)(dblRSum>0.0 ? dblRSum<255.0 ? unsigned char(dblRSum + 0.5) : 255 : 0);
					colPixel.g = (unsigned char)(dblGSum>0.0 ? dblGSum<255.0 ? unsigned char(dblGSum + 0.5) : 255 : 0);
					colPixel.b = (unsigned char)(dblBSum>0.0 ? dblBSum<255.0 ? unsigned char(dblBSum + 0.5) : 255 : 0);
				}
			}
		}
		else
		{
			for (unsigned ySrc=0; ySrc<src.GetHeight(); ++ySrc)
			{
				IPColour const * pSrcRow = src.GetColourRow(ySrc);
				
				for (unsigned yDst=0; yDst<imgTmp.GetHeight(); ++yDst)
				{
					IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
					
					double dblRSum = 0.0;
					double dblGSum = 0.0;
					double dblBSum = 0.0;
					double dblNonMaskWeight = 1.0;
					bool bAnyMasks = false;
					
					for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
					{
						IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
						
						IPColour colour = pSrcRow[spContributor.GetIndex()];
						
						if (src.GetMaskColour() == colour)
						{
							bAnyMasks = true;
							dblNonMaskWeight -= spContributor.GetWeight();
						}
						else
						{
							dblRSum += double(colour.r) * spContributor.GetWeight();
							dblGSum += double(colour.g) * spContributor.GetWeight();
							dblBSum += double(colour.b) * spContributor.GetWeight();
						}
					}
					
					if (bAnyMasks)
					{
						if (dblNonMaskWeight <= 0.5)
						{
							aabTmpMask[yDst][ySrc] = true;
							continue;
						}
						
						double dblInvNonMaskWeight = 1.0 / dblNonMaskWeight;
						dblRSum *= dblInvNonMaskWeight;
						dblGSum *= dblInvNonMaskWeight;
						dblBSum *= dblInvNonMaskWeight;
					}
					
					aabTmpMask[yDst][ySrc] = false;
					
					IPColour & colPixel = imgTmp.GetColourRow(yDst)[ySrc];
					
					colPixel.r = (unsigned char)(dblRSum>0.0 ? dblRSum<255.0 ? unsigned char(dblRSum + 0.5) : 255 : 0);
					colPixel.g = (unsigned char)(dblGSum>0.0 ? dblGSum<255.0 ? unsigned char(dblGSum + 0.5) : 255 : 0);
					colPixel.b = (unsigned char)(dblBSum>0.0 ? dblBSum<255.0 ? unsigned char(dblBSum + 0.5) : 255 : 0);
				}
			}
		}
	}
	else
	{
		aabTmpMask = NULL;
		
		if (src.IsPalettized())
		{
			for (unsigned ySrc=0; ySrc<src.GetHeight(); ++ySrc)
			{
				IPIndex const * pSrcRow = src.GetIndexRow(ySrc);
				
				for (unsigned yDst=0; yDst<imgTmp.GetHeight(); ++yDst)
				{
					IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
					
					double dblRSum = 0.0;
					double dblGSum = 0.0;
					double dblBSum = 0.0;
					
					for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
					{
						IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
						
						//IPIndex ixPalette = pSrcRow[spContributor.GetIndex()];
						
						IPColour const & colour = src.GetPalette()[pSrcRow[spContributor.GetIndex()]];
						
						dblRSum += double(colour.r) * spContributor.GetWeight();
						dblGSum += double(colour.g) * spContributor.GetWeight();
						dblBSum += double(colour.b) * spContributor.GetWeight();
					}
					
					IPColour & colPixel = imgTmp.GetColourRow(yDst)[ySrc];
					
					colPixel.r = (unsigned char)(dblRSum>0.0 ? dblRSum<255.0 ? unsigned char(dblRSum + 0.5) : 255 : 0);
					colPixel.g = (unsigned char)(dblGSum>0.0 ? dblGSum<255.0 ? unsigned char(dblGSum + 0.5) : 255 : 0);
					colPixel.b = (unsigned char)(dblBSum>0.0 ? dblBSum<255.0 ? unsigned char(dblBSum + 0.5) : 255 : 0);
				}
			}
		}
		else
		{
			for (unsigned ySrc=0; ySrc<src.GetHeight(); ++ySrc)
			{
				IPColour const * pSrcRow = src.GetColourRow(ySrc);
				
				for (unsigned yDst=0; yDst<imgTmp.GetHeight(); ++yDst)
				{
					IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
					
					double dblRSum = 0.0;
					double dblGSum = 0.0;
					double dblBSum = 0.0;
					
					for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
					{
						IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
						
						IPColour colour = pSrcRow[spContributor.GetIndex()];
						
						dblRSum += double(colour.r) * spContributor.GetWeight();
						dblGSum += double(colour.g) * spContributor.GetWeight();
						dblBSum += double(colour.b) * spContributor.GetWeight();
					}
					
					IPColour & colPixel = imgTmp.GetColourRow(yDst)[ySrc];
					
					colPixel.r = (unsigned char)(dblRSum>0.0 ? dblRSum<255.0 ? unsigned char(dblRSum + 0.5) : 255 : 0);
					colPixel.g = (unsigned char)(dblGSum>0.0 ? dblGSum<255.0 ? unsigned char(dblGSum + 0.5) : 255 : 0);
					colPixel.b = (unsigned char)(dblBSum>0.0 ? dblBSum<255.0 ? unsigned char(dblBSum + 0.5) : 255 : 0);
				}
			}
		}
	}
	
	// Do the alpha channel if there is one
	
	if (src.HasAlphaChannel())
	{
		imgTmp.CreateAlphaChannel();
		
		for (unsigned ySrc=0; ySrc<src.GetHeight(); ++ySrc)
		{
			IPAlphaValue const * pSrcRow = src.GetAlphaRow(ySrc);
			
			for (unsigned yDst=0; yDst<imgTmp.GetHeight(); ++yDst)
			{
				IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
				
				double dblSum = 0.0;
				
				for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
				{
					IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
					
					dblSum += double(pSrcRow[spContributor.GetIndex()]) * spContributor.GetWeight();
				}
				
				imgTmp.GetAlphaRow(yDst)[ySrc] = dblSum>0.0 ? dblSum<255.0 ? IPAlphaValue(dblSum + 0.5) : 255 : 0;
			}
		}
	}
	
	// Create this image
	
	CreateNew(width, height, NOT_PALETTIZED);
	
	// Create sourcing table for scaling height
	
	fst.Create(height, src.GetHeight(), yScale, filter);
	
	if (aabTmpMask)
	{
		if (src.IsPalettized())
		{
			SetMaskColour(src.GetPalette()[src.GetMaskIndex()]);
		}
		else
		{
			SetMaskColour(src.GetMaskColour());
		}
	
		for (unsigned ySrc=0; ySrc<imgTmp.GetHeight(); ++ySrc)
		{
			IPColour const * pSrcRow = imgTmp.GetColourRow(ySrc);
			bool const * pMaskRow = aabTmpMask[ySrc];
			
			for (unsigned yDst=0; yDst<GetHeight(); ++yDst)
			{
				IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
				
				double dblRSum = 0.0;
				double dblGSum = 0.0;
				double dblBSum = 0.0;
				double dblNonMaskWeight = 1.0;
				bool bAnyMasks = false;
				
				for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
				{
					IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
					
					if (pMaskRow[spContributor.GetIndex()])
					{
						bAnyMasks = true;
						dblNonMaskWeight -= spContributor.GetWeight();
					}
					else
					{
						IPColour colour = pSrcRow[spContributor.GetIndex()];
						
						dblRSum += double(colour.r) * spContributor.GetWeight();
						dblGSum += double(colour.g) * spContributor.GetWeight();
						dblBSum += double(colour.b) * spContributor.GetWeight();
					}
				}
				
				if (bAnyMasks)
				{
					if (dblNonMaskWeight <= 0.5)
					{
						GetColourRow(yDst)[ySrc] = GetMaskColour();
						continue;
					}
					
					double dblInvNonMaskWeight = 1.0 / dblNonMaskWeight;
					dblRSum *= dblInvNonMaskWeight;
					dblGSum *= dblInvNonMaskWeight;
					dblBSum *= dblInvNonMaskWeight;
				}
				
				IPColour & colPixel = GetColourRow(yDst)[ySrc];
				
				colPixel.r = (unsigned char)(dblRSum>0.0 ? dblRSum<255.0 ? unsigned char(dblRSum + 0.5) : 255 : 0);
				colPixel.g = (unsigned char)(dblGSum>0.0 ? dblGSum<255.0 ? unsigned char(dblGSum + 0.5) : 255 : 0);
				colPixel.b = (unsigned char)(dblBSum>0.0 ? dblBSum<255.0 ? unsigned char(dblBSum + 0.5) : 255 : 0);
				
				if (colPixel == GetMaskColour())
				{
					if (colPixel.b) --colPixel.b;
					else ++colPixel.b;
				}
			}
		}
		delete[] *aabTmpMask;
		delete[] aabTmpMask;
	}
	else
	{
		for (unsigned ySrc=0; ySrc<imgTmp.GetHeight(); ++ySrc)
		{
			IPColour const * pSrcRow = imgTmp.GetColourRow(ySrc);
			
			for (unsigned yDst=0; yDst<GetHeight(); ++yDst)
			{
				IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
				
				double dblRSum = 0.0;
				double dblGSum = 0.0;
				double dblBSum = 0.0;
				
				for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
				{
					IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
					
					IPColour colour = pSrcRow[spContributor.GetIndex()];
					
					dblRSum += double(colour.r) * spContributor.GetWeight();
					dblGSum += double(colour.g) * spContributor.GetWeight();
					dblBSum += double(colour.b) * spContributor.GetWeight();
				}
				
				IPColour & colPixel = GetColourRow(yDst)[ySrc];
				
				colPixel.r = (unsigned char)(dblRSum>0.0 ? dblRSum<255.0 ? unsigned char(dblRSum + 0.5) : 255 : 0);
				colPixel.g = (unsigned char)(dblGSum>0.0 ? dblGSum<255.0 ? unsigned char(dblGSum + 0.5) : 255 : 0);
				colPixel.b = (unsigned char)(dblBSum>0.0 ? dblBSum<255.0 ? unsigned char(dblBSum + 0.5) : 255 : 0);
			}
		}
	}
	
	// Do the alpha channel if there is one
	
	if (imgTmp.HasAlphaChannel())
	{
		CreateAlphaChannel();
		
		for (unsigned ySrc=0; ySrc<imgTmp.GetHeight(); ++ySrc)
		{
			IPAlphaValue const * pSrcRow = imgTmp.GetAlphaRow(ySrc);
			
			for (unsigned yDst=0; yDst<GetHeight(); ++yDst)
			{
				IPFilterSourcingTable::DstPixel const & dpSources = fst.GetSources(yDst);
				
				double dblSum = 0.0;
				
				for (unsigned i=0; i<dpSources.GetNumContributors(); ++i)
				{
					IPFilterSourcingTable::DstPixel::SrcPixel const & spContributor = dpSources.GetContributor(i);
					
					dblSum += double(pSrcRow[spContributor.GetIndex()]) * spContributor.GetWeight();
				}
				
				GetAlphaRow(yDst)[ySrc] = dblSum>0.0 ? dblSum<255.0 ? IPAlphaValue(dblSum + 0.5) : 255 : 0;
			}
		}
	}
	
	return true;
}

