//
// Tools_PS3_Texture_Writer.cpp
//
// Created by KJFL - 28 June 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "ImageProcessing/Tools_PS3_Texture_Writer.h"

#include "Core/Asura_TextureFlags.h"
#include "Debugging/Asura_Assert.h"

#include "Load&Save/Tools_TextureConversionData.h"
#include "ImageProcessing/Tools_NvTextureTools.h"

#ifdef ASURA_MAIN
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#endif
#include <float.h>

static inline u_int Asura_PS3_CountLeadingZero( u_int x )
{
	u_int y;
	u_int n = 32;

	y = x >> 16; if (y != 0) { n = n - 16; x = y; }
	y = x >>  8; if (y != 0) { n = n -  8; x = y; }
	y = x >>  4; if (y != 0) { n = n -  4; x = y; }
	y = x >>  2; if (y != 0) { n = n -  2; x = y; }
	y = x >>  1; if (y != 0) { return n - 2; }
	return n - x;
}

static inline u_int Asura_PS3_UtilLog2( u_int x )
{
	return 31 - Asura_PS3_CountLeadingZero( x );
}

static void Asura_PS3_ConvertSwizzle( u_char*& dst, u_char*& src, const u_int width, const u_int depth, const u_int xpos, const u_int ypos, const u_int level )
{
	if( level == 1 )
	{
		switch( depth )
		{
		case 16: //FP32
			{
				*( ( u_int*& ) dst )++ = *( ( u_int* ) src + ( ypos * width + xpos ) );
				break;
			}
		case 8: //FP16
			{
				*( ( u_int*& ) dst )++ = *( ( u_int* ) src + ( ypos * width + xpos ) );
				break;
			}
		case 4: //RGBA or ARGB
			{
				*( ( u_int*& ) dst )++ = *( ( u_int* ) src + ( ypos * width + xpos ) );
				break;
			}
		case 3: //RGB
			{
				*dst++ = src[ ( ypos * width + xpos ) * depth ];
				*dst++ = src[ ( ypos * width + xpos ) * depth + 1 ];
				*dst++ = src[ ( ypos * width + xpos ) * depth + 2 ];
				break;
			}
		case 2:
			{
				*( ( u_short*& ) dst )++ = *( ( u_short* ) src + ( ypos * width + xpos ) );
				break;
			}
		case 1: //B8
			{
				*( ( u_char*& ) dst )++ = *( ( u_char* ) src + ( ypos * width + xpos ) );
				break;
			}
		default:
			Asura_Assert( false, "Invalid depth size ( supported 16, 8, 4, 3, 2, 1 )" );
		}
		return;
	}
	else
	{
		Asura_PS3_ConvertSwizzle( dst, src, width, depth, xpos, ypos, level - 1 );
		Asura_PS3_ConvertSwizzle( dst, src, width, depth,	xpos + ( 1U << ( level - 2 ) ), ypos, level - 1 );
		Asura_PS3_ConvertSwizzle( dst, src, width, depth, xpos, ypos + ( 1U << ( level - 2 ) ), level - 1 );
		Asura_PS3_ConvertSwizzle( dst, src, width, depth, xpos + ( 1U << ( level - 2 ) ), ypos + ( 1U << ( level - 2 ) ), level - 1 );
	}
}

void Asura_PS3_ConvertLinearToSwizzle( u_char *dst, u_char *src, u_int width, const u_int height, u_int depth )
{
	switch( depth )
	{
	case 8:
		{
			width *= 2;
			depth = 4;
			break;
		}
	case 16:
		{
			width *= 4;
			depth = 4;
			break;
		}
	default:
		{
			break;
		}
	}
	if( width == height )
	{
		Asura_PS3_ConvertSwizzle( dst, src, width, depth, 0, 0, Asura_PS3_UtilLog2( width ) + 1 );
	}
	else if( width > height )
	{
		u_int baseLevel = Asura_PS3_UtilLog2( width ) - ( Asura_PS3_UtilLog2( width ) - Asura_PS3_UtilLog2( height ) );
		for (u_int i = 0; i < ( 1UL << ( Asura_PS3_UtilLog2( width ) - Asura_PS3_UtilLog2( height ) ) ); ++i )
		{
				Asura_PS3_ConvertSwizzle( dst, src, width, depth, ( 1U << baseLevel ) * i, 0, baseLevel + 1 );
		}
	}
	else if( width < height )
	{
		u_int baseLevel = Asura_PS3_UtilLog2( height ) - ( Asura_PS3_UtilLog2( height ) - Asura_PS3_UtilLog2( width ) );
		for ( u_int i = 0; i < ( 1UL << ( Asura_PS3_UtilLog2( height ) - Asura_PS3_UtilLog2( width ) ) ); ++i )
		{
				Asura_PS3_ConvertSwizzle( dst, src, width, depth, 0, ( 1U << baseLevel ) * i, baseLevel + 1 );
		}
	}
}

Tools_PS3_Texture_Writer* Tools_PS3_Texture_Writer::s_pxCurTextureWriter;

volatile bool			g_bCubemap		= false;
volatile int			g_iSurface		= 0;

NV_ERROR_CODE Tools_PS3_Texture_Writer::MipMapCallback(const void* pPixelData, size_t uCount, const MIPMapData* pxMipMapData, void* pUserData)
{
	if (!pxMipMapData)
	{
		// must be some header data
		return NV_OK;
	}

	// do we have space?
	if (MAXIMUM_NUMBER_OF_IMAGES == s_pxCurTextureWriter->m_uNumberOfImages)
	{
		return NV_OK;
	}

	u_int uMipLevel = pxMipMapData->mipLevel;

	// Is this an expected mipmap level? (zero corresponds to the 'original' image)
	if ( uMipLevel < 0 || uMipLevel > MAXIMUM_NUMBER_OF_IMAGES )
	{
		Asura_Assertf( false, ("Invalid mipmap level: %d", uMipLevel ));
		return NV_FAIL;
	}
	
	Tools_PS3_ImageData& xImage = g_bCubemap ? s_pxCurTextureWriter->m_axImages[g_iSurface][s_pxCurTextureWriter->m_uNumberOfImages] : s_pxCurTextureWriter->m_axImage[s_pxCurTextureWriter->m_uNumberOfImages];

	if( xImage.m_uHeight < 1 || xImage.m_uWidth < 1 )
	{
		return NV_OK;
	}

	switch( s_pxCurTextureWriter->m_xHeader.m_ucTextureFormatID )
	{
		default:
	
		case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
		{
			// Is the size of this image what we were expecting?
			if ( uCount != xImage.m_uImageSizeInBytes)
			{
				Asura_Assertf( false, ("Mipmap size == %d, but ImageSize == %d", uCount, xImage.m_uImageSizeInBytes) );
				return NV_FAIL;
			}

			xImage.m_puPixelData = new u_int[uCount];

			if(xImage.m_puPixelData == NULL)
			{
				return(NV_FAIL);
			}
			else
			{
				//memcpy( xImage.m_puPixelData, pPixelData, uCount );
				const u_char* pucSource = reinterpret_cast<const u_char*>(pPixelData);
				u_char* pucDest = reinterpret_cast<u_char*>(xImage.m_puPixelData);
				for (u_int u=0; u<uCount/4; u++)
				{
					// KJFL - 14 July 2005 - uncompressed PS3 textures go RGBA (A in lowest byte)
					// BUT it is also endianess swapped. 
					pucDest[0] = pucSource[2];
					pucDest[1] = pucSource[1];
					pucDest[2] = pucSource[0];
					pucDest[3] = pucSource[3];

					pucDest+=4;
					pucSource+=4;
				}
			}

			break;
		}

		case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
		case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
		case ASURA_PS3_TEXTUREFORMAT_DXT1_A8:
		{
			// KJFL - 15 July 2005 - 
			// Touch up image size to that reported by mipmap generator.
			// Compressed textures may have different sizes to those you would expect,
			// e.g. DXT5 1x1, 2x2 & 4x4 all take 16 bytes due to compression scheme.
			xImage.m_puPixelData = new u_int[uCount];

			if(xImage.m_puPixelData == NULL)
			{
				return(NV_FAIL);
			}
			else
			{
				// KJFL - 14 July 2005 - it would appear DXTs are the correct endianess for the GPU
				// or at least, at the moment they are. ;)
				memcpy( xImage.m_puPixelData, pPixelData, uCount );
			}

			break;
		}

		case ASURA_PS3_TEXTUREFORMAT_L8:
		{
			if ( uCount != xImage.m_uImageSizeInBytes)
			{
				Asura_Assertf( false, ("Mipmap size == %d, but ImageSize == %d", uCount, xImage.m_uImageSizeInBytes) );
				return NV_FAIL;
			}
			xImage.m_puPixelData = new u_int[uCount];
			if(xImage.m_puPixelData == NULL)
			{
				return(NV_FAIL);
			}
			else
			{
				Asura_PS3_ConvertLinearToSwizzle( ( u_char* )xImage.m_puPixelData, ( u_char* ) pPixelData, xImage.m_uWidth, xImage.m_uHeight, 1 );
			}
			break;
		}
		case ASURA_PS3_TEXTUREFORMAT_ABGR16F:
		{
			if ( uCount != xImage.m_uImageSizeInBytes)
			{
				Asura_Assertf( false, ("Mipmap size == %d, but ImageSize == %d", uCount, xImage.m_uImageSizeInBytes) );
				return NV_FAIL;
			}
			xImage.m_puPixelData = new u_int[uCount];
			if(xImage.m_puPixelData == NULL)
			{
				return(NV_FAIL);
			}
			else
			{
				u_int* puPixelData = new u_int[uCount];

				if(puPixelData == NULL)
				{
					return(NV_FAIL);
				}
				else
				{
					Asura_PS3_ConvertLinearToSwizzle( ( u_char* )puPixelData, ( u_char* ) pPixelData, xImage.m_uWidth, xImage.m_uHeight, 8 );
					const u_char* pucSource = reinterpret_cast<const u_char*>(puPixelData);
					u_char* pucDest = reinterpret_cast<u_char*>(xImage.m_puPixelData);
					for( u_int i = 0; i < uCount - 1; i += 2 )
					{
						pucDest[ i     ] = pucSource[ i + 1 ];
						pucDest[ i + 1 ] = pucSource[ i     ];
					}
					delete puPixelData;
				}
			}
			break;
		}


	}

	s_pxCurTextureWriter->m_uNumberOfImages++;

	return NV_OK;
}


Tools_PS3_Texture_Writer::Tools_PS3_Texture_Writer(const u_int uWidth, const u_int uHeight, const u_int uPitch, 
												   const void* pPixels0, const void* pPixels1, const void* pPixels2,
												   const void* pPixels3, const void* pPixels4, const void* pPixels5,
												   const TextureConversionData* pxConvertData, 
												   nvTextureFormats eNvTextureFormat,  bool& bTextureWriterSucess,
												   u_int uMaxMipMapDrops, bool bAlphaOnly )
{
	//CUBE MAPS

	bTextureWriterSucess = false;

	m_bDropExtraMipLevel = false;

	enum ASURA_PS3_TEXTUREFORMAT eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
	g_bCubemap = false;

	switch( eNvTextureFormat )
	{
	case kDXT1:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT1_RGB;
			if( bAlphaOnly )
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT1_A8;
			}
			break;
		}
	case kDXT5:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
			break;
		}
	case kDXT5_NM:
		{
			//We could use COMPRESSED_HILO_S8 - it's 16-bits per pixel but perfect for normal maps, it saves us a bit of maths.
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;

#if 0
			//	We want to remove the "hidden feature" lowering normal map texture sizes
			if ( (pxConvertData->m_iTextureFlags & AsuraTextureFlag_HighQualityCompression) == 0 )
			{
				//we don't have an equivalent to xbox's low-quality normal compression, so instead bump the normal map down by 1 mip level.
				m_bDropExtraMipLevel = true;
			}
#endif

			break;
		}
	case kA8:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT1_A8;
			break;
		}
	case kL8:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_L8;
			break;
		}
	}

	m_uMaxResultingMipmaps = uMaxMipMapDrops;
	m_bDropExtraMipLevel = m_bDropExtraMipLevel || ((pxConvertData->m_iTextureFlags & AsuraTextureFlag_ReduceExtraForConsoles) != 0);
	m_uMaximumTextureSize = pxConvertData ? pxConvertData->m_uMaximumTextureSize : 0;
	
	if ( (pxConvertData->m_iTextureFlags & AsuraTextureFlag_NoMipMaps) != 0 ) m_uMaxResultingMipmaps = 1;

	m_uNumberOfImages = 0;

	for( int i = 0; i < 6; ++i )
	{
		m_axImages[i][0].m_uHeight = uHeight;
		m_axImages[i][0].m_uWidth = uWidth;

		// JAT 19/07/07 - need to set the image size correctly here to handle textures which are <4 in width or height before mip-mapping
		int iW = m_axImages[i][ 0 ].m_uWidth;
		int iH = m_axImages[i][ 0 ].m_uHeight;
		if( iW < 4 ) iW = 4;
		if( iH < 4 ) iH = 4;

		switch (eTextureFormat)
		{
		case ASURA_PS3_TEXTUREFORMAT_L8:
			{
				m_axImages[i][0].m_uImageSizeInBytes = iW * iH;
				break;
			}
		case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
		case ASURA_PS3_TEXTUREFORMAT_DXT1_A8:
			{
				m_axImages[i][0].m_uImageSizeInBytes = ( ( iW * iH ) >> 1 ) ;
				break;
			}
		default:
		case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
		case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
				m_axImages[i][0].m_uImageSizeInBytes = iW * iH;
				break;
			}
		}

		for (u_int u = 1; u < MAXIMUM_NUMBER_OF_IMAGES; ++u)
		{
			m_axImages[i][ u ].m_uHeight = m_axImages[i][ u - 1 ].m_uHeight >> 1;
			m_axImages[i][ u ].m_uWidth	 = m_axImages[i][ u - 1 ].m_uWidth >> 1;
			iW = m_axImages[i][ u ].m_uWidth;
			iH = m_axImages[i][ u ].m_uHeight;

			if( iW < 4 ) iW = 4;
			if( iH < 4 ) iH = 4;

			switch (eTextureFormat)
			{
			case ASURA_PS3_TEXTUREFORMAT_L8:
				{
					m_axImages[i][u].m_uImageSizeInBytes = iW * iH;
					break;
				}
			case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
			case ASURA_PS3_TEXTUREFORMAT_DXT1_A8:
				{
					m_axImages[i][u].m_uImageSizeInBytes = ( ( iW * iH ) >> 1 ) ;
					break;
				}
			default:
			case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
			case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
				{
					eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
					m_axImages[i][u].m_uImageSizeInBytes = iW * iH;
					break;
				}
			}
		}
	}

	// fill in texture header info
	m_xHeader.m_ucTextureFormatID	= static_cast<u_char>(eTextureFormat);
	m_xHeader.m_usWidth				= static_cast<u_short>(uWidth);
	m_xHeader.m_usHeight			= static_cast<u_short>(uHeight);
	m_xHeader.m_ucNumberOfMipmaps	= static_cast<u_char>(MAXIMUM_NUMBER_OF_IMAGES); // Will be reduced at write-out time if more than possible
	m_xHeader.m_usFlags				= ASURA_PS3_TEXTURE_CUBEMAP;

	// get the images please
	g_bCubemap = true;
	g_iSurface = 0;
	m_uNumberOfImages = 0;
	const bool bSuccess0 = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels0, pxConvertData );
	g_iSurface = 1;
	m_uNumberOfImages = 0;
	const bool bSuccess1 = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels1, pxConvertData );
	g_iSurface = 2;
	m_uNumberOfImages = 0;
	const bool bSuccess2 = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels2, pxConvertData );
	g_iSurface = 3;
	m_uNumberOfImages = 0;
	const bool bSuccess3 = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels3, pxConvertData );
	g_iSurface = 4;
	m_uNumberOfImages = 0;
	const bool bSuccess4 = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels4, pxConvertData );
	g_iSurface = 5;
	m_uNumberOfImages = 0;
	const bool bSuccess5 = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels5, pxConvertData );

	bTextureWriterSucess = bSuccess0 & bSuccess1 & bSuccess2 & bSuccess3 & bSuccess4 & bSuccess5;
}

Tools_PS3_Texture_Writer::Tools_PS3_Texture_Writer(const u_int uWidth, const u_int uHeight, const u_int uPitch, 
												   const void* pPixels, const TextureConversionData* pxConvertData, 
												   nvTextureFormats eNvTextureFormat,  bool& bTextureWriterSucess,
												   u_int uMaxMipMapDrops, bool bAlphaOnly )
{
	//NOT CUBE MAPS
	
	bTextureWriterSucess = false;
	
	enum ASURA_PS3_TEXTUREFORMAT eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
	g_bCubemap = false;

	m_bDropExtraMipLevel = false;

	switch( eNvTextureFormat )
	{
	case kDXT1:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT1_RGB;
			if( bAlphaOnly )
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT1_A8;
			}
			else if (pxConvertData->m_iTextureFlags & AsuraTextureFlag_HighQualityCompression)
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_8888_RGBA;
				eNvTextureFormat = k8888;
			}
			break;
		}
	case kDXT5:
		{
			if (pxConvertData->m_iTextureFlags & AsuraTextureFlag_HighQualityCompression)
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_8888_RGBA;
				eNvTextureFormat = k8888;
			}
			else
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
			}
			break;
		}
	case kDXT5_NM:
		{
			//We could use COMPRESSED_HILO_S8 - it's 16-bits per pixel but perfect for normal maps, it saves us a bit of maths.
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;

#if 0
			//	We want to remove the "hidden feature" lowering normal map texture sizes
			if ( (pxConvertData->m_iTextureFlags & AsuraTextureFlag_HighQualityCompression) == 0 )
			{
				//we don't have an equivalent to xbox's low-quality normal compression, so instead bump the normal map down by 1 mip level.
				m_bDropExtraMipLevel = true;
			}
#endif

			break;
		}
	case kA8:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT1_A8;   //BW L8 doesn't work for alpha only, Asura_PC_CompressDXT() reads the RGB channel, not the alpha, only works on DXT1_A8. - the run-time doesn't have a problem and can use L8 as A8 (on PS3 they're the same)
			break;
		}
	case kL8:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_L8;
			break;
		}
	case kA16B16G16R16F:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_ABGR16F;
			break;
		}
	}

	// KJFL - 15 July 2005 - will at some point want to check alpha values:
	//
	// Export DXT1 - no alpha
	// Export DXT5 - with alpha
	//
	// Currently just putting out DXT5

	// scan through texture to see if any mid-range alpha values are used
/*	bool b8BitAlpha = false;
	for( u_int uH = 0; uH < uHeight; ++uH )
	{
		int addr = uH * uPitch + 3;
		for( u_int uW = 0; uW < uWidth; ++uW )
		{
			u_char ucAlpha = *(((u_char*)pPixels) + addr );
			addr += 4;

			if( ucAlpha != 0 && ucAlpha != 255 )
			{
				b8BitAlpha = true;
				break;
			}
		}
		if( b8BitAlpha ) break;
	}
	*/
    m_uMaxResultingMipmaps = uMaxMipMapDrops;
	m_bDropExtraMipLevel = m_bDropExtraMipLevel || ((pxConvertData->m_iTextureFlags & AsuraTextureFlag_ReduceExtraForConsoles) != 0);
	m_uMaximumTextureSize = pxConvertData ? pxConvertData->m_uMaximumTextureSize : 0;

	if ( (pxConvertData->m_iTextureFlags & AsuraTextureFlag_NoMipMaps) != 0 ) m_uMaxResultingMipmaps = 1;

	m_uNumberOfImages = 0;

	m_axImage[0].m_uHeight = uHeight;
	m_axImage[0].m_uWidth = uWidth;
 
	// JAT 19/07/07 - need to set the image size correctly here to handle textures which are <4 in width or height before mip-mapping
	int iW = m_axImage[ 0 ].m_uWidth;
	int iH = m_axImage[ 0 ].m_uHeight;
	if( iW < 4 ) iW = 4;
	if( iH < 4 ) iH = 4;

	switch (eTextureFormat)
	{
		case ASURA_PS3_TEXTUREFORMAT_ABGR16F:
		{
			m_axImage[0].m_uImageSizeInBytes = ( m_axImage[ 0 ].m_uWidth * m_axImage[ 0 ].m_uHeight ) * 8;
			break;
		}
		case ASURA_PS3_TEXTUREFORMAT_L8:
		{
			m_axImage[0].m_uImageSizeInBytes = m_axImage[ 0 ].m_uWidth * m_axImage[ 0 ].m_uHeight;
			break;
		}
		case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
 		case ASURA_PS3_TEXTUREFORMAT_DXT1_A8:
		{
			m_axImage[0].m_uImageSizeInBytes = ( ( iW * iH ) >> 1 ) ;
			break;
		}
		case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
		{
			m_axImage[0].m_uImageSizeInBytes = iW * iH * 4;
			break;
		}

		default:
		case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
		{
			eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
			m_axImage[0].m_uImageSizeInBytes = iW * iH;
			break;
		}
	}

	for (u_int u = 1; u < MAXIMUM_NUMBER_OF_IMAGES; ++u)
	{
		m_axImage[ u ].m_uHeight = m_axImage[ u - 1 ].m_uHeight >> 1;
		m_axImage[ u ].m_uWidth	 = m_axImage[ u - 1 ].m_uWidth >> 1;
		iW = m_axImage[ u ].m_uWidth;
		iH = m_axImage[ u ].m_uHeight;

		if(eTextureFormat == ASURA_PS3_TEXTUREFORMAT_8888_RGBA)
		{
			if( iW < 1 ) iW = 1;
			if( iH < 1 ) iH = 1;
		}
		else
		{
			if( iW < 4 ) iW = 4;
			if( iH < 4 ) iH = 4;
		}

		switch (eTextureFormat)
		{
			case ASURA_PS3_TEXTUREFORMAT_ABGR16F:
				{
					m_axImage[u].m_uImageSizeInBytes = m_axImage[ u ].m_uWidth * m_axImage[ u ].m_uHeight * 8;
					break;
				}
			case ASURA_PS3_TEXTUREFORMAT_L8:
				{
					m_axImage[u].m_uImageSizeInBytes = m_axImage[ u ].m_uWidth * m_axImage[ u ].m_uHeight;
					break;
				}
			case ASURA_PS3_TEXTUREFORMAT_DXT1_RGB:
 			case ASURA_PS3_TEXTUREFORMAT_DXT1_A8:
			{
				m_axImage[u].m_uImageSizeInBytes = ( ( iW * iH ) >> 1 ) ;
				break;
			}
			case ASURA_PS3_TEXTUREFORMAT_8888_RGBA:
			{
				m_axImage[u].m_uImageSizeInBytes = iW * iH * 4;
				break;
			}
			default:
			case ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA:
			{
				eTextureFormat = ASURA_PS3_TEXTUREFORMAT_DXT5_RGBA;
				m_axImage[u].m_uImageSizeInBytes = iW * iH;
				break;
			}
		}
	}

	// fill in texture header info
	m_xHeader.m_ucTextureFormatID	= static_cast<u_char>(eTextureFormat);
	m_xHeader.m_usWidth				= static_cast<u_short>(uWidth);
	m_xHeader.m_usHeight			= static_cast<u_short>(uHeight);
	m_xHeader.m_ucNumberOfMipmaps	= static_cast<u_char>(MAXIMUM_NUMBER_OF_IMAGES); // Will be reduced at write-out time if more than possible
	switch( eTextureFormat )
	{
	case ASURA_PS3_TEXTUREFORMAT_L8:
	case ASURA_PS3_TEXTUREFORMAT_ABGR16F:
		{
			m_xHeader.m_usFlags				= ASURA_PS3_TEXTURE_SWIZZLE;
			break;
		}
	default:
		{
			m_xHeader.m_usFlags				= ASURA_PS3_TEXTURE_NORMAL;
			break;
		}
	}

	// get the images please
	const bool bSuccess = GetImages( eNvTextureFormat, uWidth, uHeight, uPitch, pPixels, pxConvertData );

	// swizzling/endianess change required?
	//ConvertPixelData( eTextureFormat );
	bTextureWriterSucess = bSuccess;
}

Tools_PS3_Texture_Writer::~Tools_PS3_Texture_Writer()
{
	// delete the images' pixel data etc

	for (u_int u = 0; u < MAXIMUM_NUMBER_OF_IMAGES; ++u )
	{
		m_axImage[ u ].Reset();
	}

	for(int i = 0; i < 6; ++i )
	{
		for (u_int u = 0; u < MAXIMUM_NUMBER_OF_IMAGES; ++u )
		{
			m_axImages[ i ][ u ].Reset();
		}
	}
}

#define WRITE_DATA(_data,_datasize) \
{\
	WriteFile(xFile,_data,_datasize,&uBytesWritten,0);\
	uTotalWritten+=uBytesWritten;\
	if((unsigned)_datasize!=uBytesWritten)\
	{\
		CloseHandle(xFile);\
		return false;\
	}\
}

// check endianess
#define SWAP_SHORT_ENDIANESS(__a)								\
{																\
	union														\
	{															\
		short twobytes;											\
		char bytes[2];											\
	};															\
	twobytes = *(short*)(&(__a));								\
	bytes[0] = (char)(bytes[0] ^ bytes[1]);						\
	bytes[1] = (char)(bytes[1] ^ bytes[0]);						\
	bytes[0] = (char)(bytes[0] ^ bytes[1]);						\
	*(short*)(&(__a)) = twobytes;								\
}										


bool Tools_PS3_Texture_Writer::WriteOutFile(const char* szOutputFilename)
{
	unsigned long uBytesWritten = 0;
	unsigned long uTotalWritten = 0;

	u_int uImageToUse = 0;

	if ( m_bDropExtraMipLevel && ( (uImageToUse + 1) < m_uNumberOfImages ) )
	{
		++uImageToUse;
	}

	//Drop mip maps to get beneath the texture size specified in the conversion data, if needed
	if(m_uMaximumTextureSize)
	{
		while ( ( m_axImage[uImageToUse].m_uWidth > m_uMaximumTextureSize || m_axImage[uImageToUse].m_uHeight > m_uMaximumTextureSize)
				&& uImageToUse < m_uNumberOfImages )
		{
			uImageToUse++;
		}
	}


	u_char ucMaxMipMaps = static_cast<u_char>(m_uMaxResultingMipmaps);

	// PS3 needs textures down to 1x1 level
	HANDLE xFile = CreateFile(szOutputFilename,GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);

	if(xFile == INVALID_HANDLE_VALUE) return false;

	if( g_bCubemap )
	{
		for	(	 m_xHeader.m_ucNumberOfMipmaps=0;
				(m_xHeader.m_ucNumberOfMipmaps < ucMaxMipMaps)
			&& ((m_xHeader.m_ucNumberOfMipmaps + uImageToUse + 1) < m_uNumberOfImages)
			&& ((m_axImages[0][m_xHeader.m_ucNumberOfMipmaps + uImageToUse + 1].m_uHeight) >= 1)
			&& ((m_axImages[0][m_xHeader.m_ucNumberOfMipmaps + uImageToUse + 1].m_uWidth) >= 1);
			++m_xHeader.m_ucNumberOfMipmaps
			)
		{
		}
			m_xHeader.m_usWidth = static_cast<u_short>(m_axImages[0][uImageToUse].m_uWidth);
			m_xHeader.m_usHeight = static_cast<u_short>(m_axImages[0][uImageToUse].m_uHeight);
	}
	else
	{
		for	(	 m_xHeader.m_ucNumberOfMipmaps=0;
				(m_xHeader.m_ucNumberOfMipmaps < ucMaxMipMaps)
			&& ((m_xHeader.m_ucNumberOfMipmaps + uImageToUse + 1) < m_uNumberOfImages)
			&& ((m_axImage[m_xHeader.m_ucNumberOfMipmaps + uImageToUse + 1].m_uHeight) >= 1)
			&& ((m_axImage[m_xHeader.m_ucNumberOfMipmaps + uImageToUse + 1].m_uWidth) >= 1);
			++m_xHeader.m_ucNumberOfMipmaps
			)
		{
		}
			m_xHeader.m_usWidth = static_cast<u_short>(m_axImage[uImageToUse].m_uWidth);
			m_xHeader.m_usHeight = static_cast<u_short>(m_axImage[uImageToUse].m_uHeight);
	}


	SWAP_SHORT_ENDIANESS( m_xHeader.m_usFlags );
	SWAP_SHORT_ENDIANESS( m_xHeader.m_usHeight );
	SWAP_SHORT_ENDIANESS( m_xHeader.m_usWidth );

	// write out the header
	WRITE_DATA( (&m_xHeader), sizeof(m_xHeader) );

	if( g_bCubemap )
	{
		u_int addr = 0;
		char acPad[ 128 ];

		for( int i = 0; i < 6; ++i )
		{
			for (u_int u = 0; u <= m_xHeader.m_ucNumberOfMipmaps; ++u )
			{
				// write out image data
				WRITE_DATA( ( m_axImages[i][ uImageToUse + u ].m_puPixelData ), ( m_axImages[ i ] [ uImageToUse + u ].m_uImageSizeInBytes ) );
				addr += m_axImages[ i ] [ uImageToUse + u ].m_uImageSizeInBytes;
			}
			int iPad = 128 - ( addr % 128 );
			if( iPad != 128 )
			{
				WRITE_DATA( ( acPad ), iPad );
				addr += iPad;
			}
		}
	}
	else
	{
		for (u_int u=0; u<=m_xHeader.m_ucNumberOfMipmaps; u++)
		{
			// write out image data
			WRITE_DATA( (m_axImage[uImageToUse+u].m_puPixelData), (m_axImage[uImageToUse+u].m_uImageSizeInBytes) );
		}
	}

	CloseHandle(xFile);

	return true;
	
}


bool Tools_PS3_Texture_Writer::GetImages( const nvTextureFormats eTextureFormat, const u_int uWidth, const u_int uHeight, 
										  const u_int uPitch, const void* pPixels, const struct TextureConversionData* pxConvertData )
{
	bool bSuccess = true;

	Asura_Assert(pxConvertData, "must pass in conversion data");
	if(!pxConvertData)
	{
		return false;
	}

	// setup options for NVidia texture conversion tools.
    s_pxCurTextureWriter = this;
	{
		const int iTextureFlags = pxConvertData->m_iTextureFlags;

		nvCompressionOptions xOptions;
		xOptions.SetDefaultOptions();
		xOptions.bDitherColor = true;
		xOptions.textureFormat = eTextureFormat;
		xOptions.mipFilterType = kMipFilterKaiser;

		// (MH - 20.12.2007) added sharpening of first few mip levels (down to 64x64)
		if (iTextureFlags & AsuraTextureFlag_SharpenMipMaps)
		{
			Tools_NvTextureTools::SetOptionsForSharpenMipMaps(xOptions, uWidth, uHeight);
		}

		//The memory management system incorrectly picks up the type of new being used by nvidia library
		//when compiled into ResourcePacker. So disabling the checking here. RAR 12/05/2006
		Asura_MemoryManagement_System::SetIgnoreNewAndDeleteTypeMismatch(true);

		const u_char* pucPixelData = reinterpret_cast<const u_char*>(pPixels);

		NV_ERROR_CODE eError = Asura_PC_CompressDXT(
			pucPixelData,//raw_data
			uWidth,
			uHeight,
			uPitch,//pitch
// BW want to compress L8s and verify TGA8->A8s
// can't I do it through this? (i.e. 1 format, for "single channel runtime compressed", plus logic here to remap channels)
// => create a test case, L8 from RGBA - compressed  & A8 from TGA8 (andy reckons => RGB => A=1 =>bug!)
			nvBGRA,	// pixel order
			&xOptions,//compression options
			MipMapCallback,	// mip callback
			NULL);	// subrect (use defaults).

		Asura_MemoryManagement_System::SetIgnoreNewAndDeleteTypeMismatch(false);

		if( eError != NV_OK) 
		{
			bSuccess = false;
// DLH - 03 May 2007 11:45:19 - Cannot obtain the error string from CompressDXT as the static memory that it lives in is deallocated when compression DLL is Freed 
#if 0 && defined( ASURA_ASSERT)
			const char* pszCompressionError = Asura_PC_CompressDXT_GetErrorString(eError);
			Asura_Assertf(eError == NV_OK, ("Error in Asura_PC_CompressDXT: %s", pszCompressionError));
#endif
#if 1 && defined( ASURA_ASSERT)
			Asura_Assertf(eError == NV_OK, ("Error in Asura_PC_CompressDXT"));
#endif
		}
	}
    s_pxCurTextureWriter = NULL;

	return( bSuccess);
}
