//
// Tools_ColourPicker.h
//
// Created by Richard Rice - 28/07/2024
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_COLOURPICKER_H__
#define __TOOLS_COLOURPICKER_H__

#include "Core/Asura_String.h"
#include "Interface/Asura_Dialog.h"
#include "Misc/Tools_Colour.h"




//Class for opening a colour picking overlay.
class Tools_ColourPicker
{
public :
	//Creates a colour picker.
	//pszTitle - Title for overlay when created
	//pfRed - Red component for editing
	//pfGreen - Green component for editing
	//pfBlue - Blue component for editing
	//fMax - Maximum value for RGB (typically 1 or 255)
	//pbChangedFlag - optional flag to be set when colour is changed
	Tools_ColourPicker(const char* pszTitle,float* pfRed,float* pfGreen,float* pfBlue, float fMax,bool* pbChangedFlag);
	Tools_ColourPicker(const char* pszTitle,Asura_Dialog_SharedValue<float>* pxSharedRed,Asura_Dialog_SharedValue<float>* pxSharedGreen,Asura_Dialog_SharedValue<float>* pxSharedBlue, float fMax,bool* pbChangedFlag);
	~Tools_ColourPicker();

	//Enables alpha to be edited at same time.. needs to occur before overlay is opened
	void EnableAlphaEditing(float* pfAlpha);
	bool IsAlphaEditingEnabled() const {return m_pfAlpha!=0;}

	//Enables HDR colour values to be selected
	void EnableHDR( const bool bHDR ) { m_bHDR = bHDR; }
	bool ISHDREnabled() const { return m_bHDR; }

	//Opens overlay if not already open
	void OpenOverlay();
	//Closes overlay if currently open
	void CloseOverlay();
	//Returns true if the overlay associated with this colour picker is open	
	bool IsOverlayOpen() const;

	//Modal overlay version
	bool DoModal();

	//Adds an extra dialog control to the overlay if it is open.
	//The control is assumed to have been created with an offset relative to the bottom left of the preset colour controls
	void AddDialog(class Asura_Dialog* pxDialog);

	// KOR - for externally changing the colour. Used when switching nodes of a colour graph.
	void SetColour(float fRed, float fGreen, float fBlue);
	void SetColour(u_int uARGB,bool bFlagAsChanged);

	//get current colour values in range 0 to 255
	float GetRed() const {return m_fRed;}
	float GetGreen() const {return m_fGreen;}
	float GetBlue() const {return m_fBlue;}
	float GetAlpha() const {return IsAlphaEditingEnabled() ? m_fAlpha : 255.0f;}

	static void LoadCustomColours();
	static void SaveCustomColours();
	
private :
	static void _Apply(int iThis){((Tools_ColourPicker*)iThis)->Apply();}
	void Apply();

	static void _Okay(int iThis){((Tools_ColourPicker*)iThis)->Okay();}
	void Okay();

	static void _Cancel(int iThis){((Tools_ColourPicker*)iThis)->Cancel();}
	void Cancel();

	void AddOverlayControls( Asura_Overlay* pxOverlay, Asura_Vector_2 &xOffset );
	void InitColourValues();

	void Update();

	static void TransferModeData(int iMode);

	//values passed in from constructor
	float* m_pfRed;
	float* m_pfGreen;
	float* m_pfBlue;
	Asura_Dialog_SharedValue<float>* m_pxSharedRed;
	Asura_Dialog_SharedValue<float>* m_pxSharedGreen;
	Asura_Dialog_SharedValue<float>* m_pxSharedBlue;
	float* m_pfAlpha;
	float m_fMax;
	bool* m_pbChangedFlag;
	Asura_String m_xTitle;

	//internal stuff
	float m_fRed;
	float m_fGreen;
	float m_fBlue;
	float m_fAlpha;
	Tools_Colour_Pair m_xColourPair;

	float m_fInitialRed;
	float m_fInitialGreen;
	float m_fInitialBlue;
	float m_fInitialAlpha;

	float m_fHDRScale;
	float m_fInitialHDRScale;

	bool m_bChanged;
	bool m_bModal;
	bool m_bHDR;

	Asura_Vector_2 m_xOffsetForExtraDialogs;


	int m_iOverlayID;

	static bool s_bCustomColoursLoaded;


	friend class Tools_ColourPicker_Overlay;
};



//Creates an coloured box that will open a colour picker when clicked.
class Tools_ColourPicker_Dialog : public Asura_Dialog
{
public :
	//pszTitle - Title for overlay when created
	//xOffset - The position of this dialog.
	//xSize - The size of the box in overlay/text units
	//pfRed - Red component for editing
	//pfGreen - Green component for editing
	//pfBlue - Blue component for editing
	//fMax - Maximum value for RGB (typically 1 or 255)
	//pbChangedFlag - optional flag to be set when colour is changed
	Tools_ColourPicker_Dialog(const char* pszTitle,Asura_Vector_2 xOffset,Asura_Vector_2 xSize,float* pfRed,float* pfGreen,float* pfBlue, float fMax,bool* pbChangedFlag);
	Tools_ColourPicker_Dialog(const char* pszTitle,Asura_Vector_2 xOffset,Asura_Vector_2 xSize,Asura_Dialog_SharedValue<float>* pxSharedRed,Asura_Dialog_SharedValue<float>* pxSharedGreen,Asura_Dialog_SharedValue<float>* pxSharedBlue, float fMax,bool* pbChangedFlag);

	//Sets an alpha value to be edited as well. Its assumed to have the same range as the RGB values.
	//(Can't be bothered with a shared value version yet , since I don't need it)
	void EnableAlphaEditing(float* pfAlpha);

	void EnableHDR(const bool bHDR) { if(m_pxPicker) m_pxPicker->EnableHDR(bHDR); }

	~Tools_ColourPicker_Dialog();

private :
	static void OnClicked(int iThis);

	Tools_ColourPicker* m_pxPicker;

	
};

//For picking a colour using an ARGB u_int (alpha won't be changed , unless EnableAlphaEditing is used)
class Tools_ARGBColourPicker_Dialog : public Asura_Dialog
{
public :
	//If bAddLabel then pszTitle will be shown to the right of the colour picker square
	Tools_ARGBColourPicker_Dialog(const char* pszTitle,Asura_Vector_2 xOffset,Asura_Vector_2 xSize,u_int* puARGB,bool* pbChangedFlag,bool bAddLabel = false);
	void EnableAlphaEditing(bool bEnable);
	void SetModal(bool bModal) { m_bModal = bModal; }

	~Tools_ARGBColourPicker_Dialog();

	virtual void PerFrameUpdate();
	virtual Asura_Vector_2 GetTextOffset() const {return m_xTextOffset;}

private :
	static void OnClicked(int iThis);

	Tools_ColourPicker* m_pxPicker;
	u_int* m_puARGB;
	float m_fRed;
	float m_fGreen;
	float m_fBlue;
	float m_fAlpha;
	bool m_bChanged;
	bool *m_pbChanged;	
	bool m_bEditAlpha;
	bool m_bModal;
	Asura_Vector_2 m_xTextOffset;
};



#endif //__TOOLS_COLOURPICKER_H__

