//
// Tools_ContextMenu_System.cpp
//
// Created by Richard Rice - 30/10/2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Interface/Tools_ContextMenu_System.h"

#include "Core/Asura_UserInput.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_Windows.h"
#include "Debugging/Asura_CPUProfile_Renderer.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Viewport.h"
#include "Sound/Asura_Audio_Systems_Manager.h"
#include "Sound/Asura_Sound.h"
#include "SpecialFX/Asura_Skybox.h"

#include "Tools_System.h"
#include "Tools_Mode.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Objects/Tools_Object_System.h"

//default menu background colour
static const u_int g_uBackgroundColour = 0xff0000ff;
static const u_int g_uSectionHeadingBackgroundColour = 0xff5f5fcf;
static const u_int g_uHighlightBackgroundColour = 0xff7f7fff;
static const u_int g_uHeadingsBackgroundColour = 0xff000000;//0xff7f7fff;

//colour for lines on menu
static const u_int g_uOutlineColour = 0xff000000;

//padding to the left/right of the menu item text
static const u_int g_uMenuItemHorizontalBorder = 12;

//Height in pixels for most menu items
static const u_int g_uMenuItemDefaultHeight = 14;
//Height for the separator
static const u_int g_uSeparatorHeight = 4;

//text colours
static const u_int g_uStandardTextColour = 0xffffffff;
static const u_int g_uHeadingsTextColour = 0xffffffff;
static const u_int g_uDisabledTextColour = 0xff7f7f7f;
static const u_int g_uDefaultActionTextColour = 0xffff3f3f;

static const u_int g_uTranslucencyColour = 0x80000000;



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Tools_ContextMenu_System **
**--------------------------*/

List<Tools_ContextMenu_DefaultAction> Tools_ContextMenu_System::s_xDefaultActionList;
bool Tools_ContextMenu_System::s_bRightClickUsed = false;


void Tools_ContextMenu_System::Update()
{
	if (!s_bRightClickUsed && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
	{
		Tools_ContextMenu xMenu;

		Tools_ContextMenu_Context xContext;
		xContext.m_bShiftHeld = Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT);
		xContext.m_bControlHeld = Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL);
		xContext.m_bAltHeld = Asura_UserInput::KeyPressed(ASURA_KEYID_ALT);

		if(Asura_Overlay_System::CursorIsOverAnOverlay())
		{
			xContext.m_pxOverlay = Asura_Overlay_System::GetCurrentOverlay();
		}


		//allow the overlay to add stuff to the menus
		if(xContext.m_pxOverlay)
		{
			xContext.m_pxOverlay->AddContextMenuItems(xMenu, xContext);
		}
		else if(Tools_Object_System::AreControlsEnabled())
		{
			//object system enabled , so allow it to add stuff
			Tools_Object_System::AddContextMenuItems(xMenu,xContext);
		}

		//allow the current mode to add stuff to the menu
		Tools_Mode* pxMode = Tools_System::GetCurrentModeClass();
		if(pxMode) pxMode->AddContextMenuItems(xMenu, xContext);


		//If anything has been added to the menus ,then show them
		if(xMenu.HasAnyEntries())
		{
			xMenu.ShowMenu();			
		}
	}
	s_bRightClickUsed = false;
}



void Tools_ContextMenu_System::AddToDefaultActionList(Asura_Hash_ID uHeadingID, Asura_Hash_ID uTargetID)
{
	if(!uHeadingID || !uTargetID) return;
	if(uHeadingID == uTargetID) return;

	//remove any existing entry for this heading
	for(LIF<Tools_ContextMenu_DefaultAction> xActionLIF(&s_xDefaultActionList); !xActionLIF.done(); xActionLIF.next())
	{
		if(xActionLIF().m_uHeadingID == uHeadingID)
		{
			xActionLIF.delete_current();
			break;
		}
	}

	//Arbitrarily delete the last entry if the list is getting too large ,in case heading names
	//are being generate in such a way as to be flooding the list
	if(s_xDefaultActionList.size()>=64) s_xDefaultActionList.delete_last_entry();

	Tools_ContextMenu_DefaultAction xAction;
	xAction.m_uHeadingID = uHeadingID;
	xAction.m_uTargetID = uTargetID;
	s_xDefaultActionList.add_entry_start(xAction);
}

Asura_Hash_ID Tools_ContextMenu_System::GetDefaultAction(Asura_Hash_ID uHeadingID)
{
	for(LIF<Tools_ContextMenu_DefaultAction> xActionLIF(&s_xDefaultActionList); !xActionLIF.done(); xActionLIF.next())
	{
		if(xActionLIF().m_uHeadingID == uHeadingID)
		{
			return xActionLIF().m_uTargetID;
			break;
		}
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------**
** Tools_ContextMenu **
**-------------------*/

bool Tools_ContextMenu::s_bMenuActive = false;

Tools_ContextMenu::Tools_ContextMenu()
{
	m_iPosX = 0;
	m_iPosY = 0;
	m_iCursorOffsetX = 0;
	m_iCursorOffsetY = 0;
	m_pxCurrentSection = 0;
	m_bReadyToClose = false;
	m_pxClickedItem = 0;
	m_szHelpText = 0;
}

Tools_ContextMenu::~Tools_ContextMenu()
{
	while(m_xSectionList.size())
	{
		delete m_xSectionList.first_entry();
		m_xSectionList.delete_first_entry();
	}
}


void Tools_ContextMenu::ShowMenu()
{
	// pause profiling to avoid assert
	ASURA_CPUPROFILE_SCOPED_PAUSE

	InitialiseBeforeDisplay();

	s_bMenuActive = true;
	while(!m_bReadyToClose)
	{
		Asura_Windows::CheckForWindowsMessages();

		
		Asura_Cursor::EnableCursorMovement(true);
		Asura_UserInput::Update();
		Asura_Cursor::OncePerFrameUpdate();

		//just updating the overlay system.. don't want the top menus to be used while the overlay is up
		//also lockout overlays , since the context menus will be using any input
		Asura_Overlay_System::Update(true);

		//do this menu's update
		Update();

		if(Asura_Windows::IsApplicationActive())
		{
			Tools_Render::BeginRenderWorld();
			Asura_Render::CurrentStates.RequestWireFrameMode(Asura_Render::GetWireframe());
			Asura_Skybox::Render();
			Tools_Render::RenderEnvironment();
			Tools_Object_System::Render();

			#if defined ASURA_TARGET_SM3
				if( Tools_Render::LightingIsOn() )
				{
					Asura_DeferredShading::Render();
				}
				else
				{
					// JAT 18/10/07 - clear out any shadows that might have been added in to the list but not rendered
					Asura_Shadow_System::FinishedRenderingShadows();
				}
			#endif

			Tools_Mode* const pxCurrMode = Tools_System::GetCurrentModeClass();
			const bool bRenderMode = pxCurrMode && pxCurrMode->ShouldRenderDuringModalOverlay();

			if(bRenderMode)
			{
				pxCurrMode->RenderWorld_PreToneMap();
			}

			#if defined ASURA_USE_HDR
				// No call to special fx system in this loop, so do this here
				Asura_Render::ApplyToneMapping();
			#endif

			if(bRenderMode)
			{
				pxCurrMode->RenderWorld_PostToneMap();
			}


			Tools_Render::EndRenderWorld();
			Asura_Render::CurrentStates.RequestWireFrameMode(false);
			if(bRenderMode)
			{
				pxCurrMode->RenderUI();
			}
			Asura_Overlay_System::Render();
			Tools_Timed_Message_System::Render();
			//render this menu
			Render(*this);

			Asura_Cursor::Render();
			Asura_Render::EndFrame();
		}

		#if defined ASURA_USE_AUDIO_SYSTEM
		Asura_Audio_Systems_Manager::Update();
		#else
		Asura_Sound_System::Update();
		#endif
	
		Asura_Timers::HandlePrimaryTimer();

	}

	RestoreCursorPosition();
	s_bMenuActive = false;

	//Update input one final time in an attempt to debounce any key press that
	//may have been used to close the overlay.
	Asura_Windows::CheckForWindowsMessages();
	Asura_UserInput::Update();

	if(m_pxClickedItem) m_pxClickedItem->HandleClick();
}

void Tools_ContextMenu::InitialiseBeforeDisplay()
{
	//reset a bunch of working data
	m_bReadyToClose = false;
	m_pxCurrentSection =0;
	m_pxClickedItem = 0;
	m_szHelpText = 0;

	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		xSectionLIF()->InitialiseBeforeDisplay();
	}


	//determine the sizes of the various menu pieces , and place at a suitable point on screen
	int iCursorX, iCursorY;
	Asura_Cursor::GetScreenSpaceCoords(iCursorX, iCursorY);

 	CalculateSize();
 	SetPosition(iCursorX, iCursorY);

	//Set the default action to use when clicking on each menu heading
	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		xSectionLIF()->SetupDefaultActionForHeading();
	}
}

void Tools_ContextMenu::Update()
{
	int iCursorX,iCursorY;
	Asura_Cursor::GetScreenSpaceCoords(iCursorX, iCursorY);

	Tools_ContextMenu_Item* pxItemUnderCursor = 0;
	
	//Check the current branch of the menu to see if the mouse is still over it
	if(m_pxCurrentSection)
	{
		pxItemUnderCursor = m_pxCurrentSection->CheckCursorPosition(iCursorX, iCursorY);
	}

	if(!pxItemUnderCursor)
	{
		//failing that , check all of the menus
		for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
		{
			pxItemUnderCursor = xSectionLIF()->CheckCursorPosition(iCursorX, iCursorY);
			if(pxItemUnderCursor)
			{
				//get the old active branch to close any submenus
				if(m_pxCurrentSection) m_pxCurrentSection->SetCurrentSection(false);

				//and tell the new one that it has become current
				m_pxCurrentSection = xSectionLIF();
				m_pxCurrentSection->SetCurrentSection(false);

				break;
			}
		}
	}
	

	if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
	{
		if(pxItemUnderCursor)
		{
			//clicked on the menus , so close if the item clicked on desires it
			if(pxItemUnderCursor->CanBeClicked())
			{
				m_pxClickedItem = pxItemUnderCursor;
				m_bReadyToClose = true;

				//remember this button as the default to use when clicking on the menu headings
				Asura_Hash_ID uHeadingID = m_pxCurrentSection ? m_pxCurrentSection->GetHeadingID() : 0;
				Asura_Hash_ID uTargetID = pxItemUnderCursor->GetID();

				Tools_ContextMenu_System::AddToDefaultActionList(uHeadingID, uTargetID);
			}
		}
		else
		{
			//clicked off the menus , so just close
			m_bReadyToClose = true;
		}
	}
	else if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
	{
		m_bReadyToClose = true;
	}

	//update tool tips
	if(Asura_Cursor::IsStatic() && pxItemUnderCursor)
	{
		m_szHelpText = pxItemUnderCursor->GetHelpText();
	}
	else
	{
		m_szHelpText = 0;
	}
}

void Tools_ContextMenu::Render(const Tools_ContextMenu& xContextMenu)
{
	//render each menu section , leaving the current one till last , so it gets drawn on top
	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		if(m_pxCurrentSection != xSectionLIF())
		{
			xSectionLIF()->Render(*this);
		}
	}
	if(m_pxCurrentSection) m_pxCurrentSection->Render(*this);

	Asura_Render::FlushDisplay();

	if(m_szHelpText && *m_szHelpText)
	{
		Asura_Overlay_System::RenderHelpText(m_szHelpText);
	}

}

Tools_ContextMenu_Section* Tools_ContextMenu::GetSection(int iSection)
{

	static const char* s_aszSectionNames[4] =
	{
		"Selection",
		"Display",
		"Object",
		"Mode",
	};
	
	if(iSection<0 || iSection >=4)
	{
		Asura_Assert(0,"Invalid section index");
		return 0;
	}	

	const char* szName = s_aszSectionNames[iSection];

	//search for an existing menu section with this name
	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		if(!stricmp(szName, xSectionLIF()->GetName()))
		{
			return xSectionLIF();
		}
	}

	//doesn't exist , so create it

	Tools_ContextMenu_Section* pxSection = new Tools_ContextMenu_Section(szName,true);
	m_xSectionList.add_entry(pxSection);

	pxSection->SetExpandsRight(((iSection>>1) & 1) ? false : true);
	pxSection->SetExpandsDown((iSection & 1) ? false : true);

	return pxSection;
}

void Tools_ContextMenu::CalculateSize()
{
	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		xSectionLIF()->CalculateSize();
	}
}

void Tools_ContextMenu::SetPosition(int iCursorX, int iCursorY)
{
	int iViewWidth = Asura_Viewport::GetWidth();
	int iViewHeight = Asura_Viewport::GetHeight();


	//determine the extents required by the menu , to see if they will fit on screen at this position
	int iMinX = 0;
	int iMinY = 0;
	int iMaxX = 0;
	int iMaxY = 0;

	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		const Tools_ContextMenu_Section* pxSection = xSectionLIF();

		if(pxSection->GetExpandsRight())
		{
			iMaxX = Asura_Maths::Max(iMaxX, pxSection->GetWidth());
		}
		else
		{
			iMinX = Asura_Maths::Min(iMinX, -pxSection->GetWidth());
		}

		if(pxSection->GetExpandsDown())
		{
			iMaxY = Asura_Maths::Max(iMaxY, pxSection->GetHeight());
		}
		else
		{
			iMinY = Asura_Maths::Min(iMinY, -pxSection->GetHeight());
		}

	}

	m_iPosX = iCursorX;
	m_iPosY = iCursorY;

	//centre the extents on the desired position , and shift things around if needed to get them to fit
	iMinX += m_iPosX;
	iMaxX += m_iPosX;
	iMinY += m_iPosY;
	iMaxY += m_iPosY;

	if(iMaxX > iViewWidth)
	{
		int iOffset = iViewWidth - iMaxX;
		m_iPosX += iOffset;
		iMinX += iOffset;
		iMaxX += iOffset;
	}
	if(iMinX < 0)
	{
		int iOffset = -iMinX;
		m_iPosX += iOffset;
		iMinX += iOffset;
		iMaxX += iOffset;
	}

	if(iMaxY > iViewHeight)
	{
		int iOffset = iViewHeight - iMaxY;
		m_iPosY += iOffset;
		iMinY += iOffset;
		iMaxY += iOffset;
	}
	if(iMinY < 0)
	{
		int iOffset = -iMinY;
		m_iPosY += iOffset;
		iMinY += iOffset;
		iMaxY += iOffset;
	}



	for(LIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		xSectionLIF()->SetPosition(m_iPosX, m_iPosY);
	}

	//If the menu was moved to fit on screen , then move the cursor to centre on it
	m_iCursorOffsetX = iCursorX - m_iPosX;
	m_iCursorOffsetY = iCursorY - m_iPosY;

	if(m_iCursorOffsetX || m_iCursorOffsetY)
	{
		Asura_Cursor::SetScreenSpaceCoords(m_iPosX,m_iPosY);
	}
}

void Tools_ContextMenu::RestoreCursorPosition()
{
	//If the cursor was moved when creating the menu , then move it back
	if(m_iCursorOffsetX || m_iCursorOffsetY)
	{
		int iCursorX, iCursorY;
		Asura_Cursor::GetScreenSpaceCoords(iCursorX, iCursorY);

		iCursorX += m_iCursorOffsetX;
		iCursorY += m_iCursorOffsetY;

		Asura_Cursor::SetScreenSpaceCoords(iCursorX, iCursorY);		
	}
}

bool Tools_ContextMenu::HasAnyEntries() const
{
	//The first item in each section will just be the section heading , so search for one with something other
	//that that
	for(CLIF<Tools_ContextMenu_Section*> xSectionLIF(&m_xSectionList); !xSectionLIF.done(); xSectionLIF.next())
	{
		if(xSectionLIF()->GetNumItems() > 1) return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Tools_ContextMenu_Section **
**---------------------------*/

Tools_ContextMenu_Section::Tools_ContextMenu_Section(const char* szName, bool bCreateHeadingItem)
{
	m_xName = szName;
	m_bNeedToAddSeparator = false;
	m_iWidth = 0;
	m_iHeight = 0;
	m_iPosX = 0;
	m_iPosY = 0;
	m_bRight = true;
	m_bDown = true;
	m_pxCurrentItem = 0;
	m_pxHeading = 0;

	if(bCreateHeadingItem)
	{
		//create a section heading item with the same name
		m_pxHeading = new Tools_ContextMenu_Item_SectionHeading( szName);
		AddItem(m_pxHeading);
	}
}

Tools_ContextMenu_Section::~Tools_ContextMenu_Section()
{
	while(m_xItemList.size())
	{
		delete m_xItemList.first_entry();
		m_xItemList.delete_first_entry();
	}
}


void Tools_ContextMenu_Section::Render(const Tools_ContextMenu& xContextMenu)
{
	//render backdrop for menu section
	Asura_Render::Overlay.SetRectXY( (float)m_iPosX, (float)m_iPosY, (float)m_iWidth, (float)m_iHeight);
	Asura_Render::Overlay.Colour = g_uBackgroundColour;		
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;

	#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
	if (xContextMenu.GetCurrentSection () != this)
	{
		Asura_Render::Overlay.Colour &= 0x00ffffff;
		Asura_Render::Overlay.Colour |= g_uTranslucencyColour;
		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
	}
	#endif

	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;

	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();

	//render each menu item
	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		xItemLIF()->Render(xContextMenu);
	}

	unsigned int uOutlineColour = g_uOutlineColour;

	#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
	if (xContextMenu.GetCurrentSection () != this)
	{
		uOutlineColour = (g_uOutlineColour & 0x00ffffff) | g_uTranslucencyColour;
	}
	#endif

	//render menu outline.
	float fWidth = (float) m_iWidth;
	float fHeight = (float) m_iHeight;
	float fX = (float) m_iPosX;
	float fY = (float) m_iPosY;
	Asura_Render::Render2DLine(fX,fY,fX+fWidth,fY,uOutlineColour);
	Asura_Render::Render2DLine(fX+fWidth,fY,fX+fWidth,fY+fHeight,uOutlineColour);
	Asura_Render::Render2DLine(fX+fWidth,fY+fHeight,fX,fY+fHeight,uOutlineColour);
	Asura_Render::Render2DLine(fX,fY+fHeight,fX,fY,uOutlineColour);

	//If there is a current menu item , then let it draw any child menus
	if(m_pxCurrentItem) m_pxCurrentItem->RenderSubmenus();
}

void Tools_ContextMenu_Section::CalculateSize()
{
	m_iWidth = 0;
	m_iHeight = 0;

	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		xItemLIF()->CalculateSize();
		m_iWidth = Asura_Maths::Max(m_iWidth, xItemLIF()->GetMinWidth());
		m_iHeight += xItemLIF()->GetHeight();
	}

	//Update the menu items so they all use the maximum width found
	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		xItemLIF()->SetWidth(m_iWidth);
	}
}

void Tools_ContextMenu_Section::SetPosition(int iX, int iY)
{
	//set position in this , converting it to the top-left position
	m_iPosX = iX;
	m_iPosY = iY;

	if(!m_bRight) m_iPosX -= m_iWidth;
	if(!m_bDown) m_iPosY -= m_iHeight;

	//set the position of each child item
	int iItemY = iY;
	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		if(!m_bDown) iItemY -= xItemLIF()->GetHeight();
		xItemLIF()->SetPosition(m_iPosX, iItemY);
		if(m_bDown) iItemY += xItemLIF()->GetHeight();
	}
}

void Tools_ContextMenu_Section::SetCurrentSection(bool bSet)
{
	if(m_pxCurrentItem && !bSet)
	{
		//this will close any submenus that were open
		m_pxCurrentItem->SetCurrentItem(false);
		m_pxCurrentItem = 0;
	}
}

bool Tools_ContextMenu_Section::Contains(int iX, int iY) const
{
	if(iX >= m_iPosX && iX < (m_iPosX+m_iWidth) &&
	   iY >= m_iPosY && iY < (m_iPosY+m_iHeight))
	{
		return true;		
	}
	return false;
}

Tools_ContextMenu_Item* Tools_ContextMenu_Section::CheckCursorPosition(int iX,int iY)
{
	//Check the current item (and any submenus) to see if the mouse is still over it
	if(m_pxCurrentItem)
	{
		Tools_ContextMenu_Item* pxItem = m_pxCurrentItem->CheckCursorPosition(iX, iY);
		if(pxItem)
		{
			return pxItem;
		}
	}

	
	//failing that , check all of the items
	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		Tools_ContextMenu_Item* pxItem = xItemLIF()->CheckCursorPosition(iX, iY);
		if(pxItem)
		{
			//get the old active item to close any submenus
			if(m_pxCurrentItem) m_pxCurrentItem->SetCurrentItem(false);

			//and tell the new one that it has become current
			m_pxCurrentItem = xItemLIF();
			m_pxCurrentItem->SetCurrentItem(true);

			return pxItem;
		}
	}
	return 0;
}

void Tools_ContextMenu_Section::AddItem(Tools_ContextMenu_Item* pxItem)
{
	if(m_bNeedToAddSeparator)
	{
		m_bNeedToAddSeparator = false;
		//probably no point in having a separator as the first entry
		int iNonHeadingItems = m_xItemList.size();
		if(m_pxHeading) iNonHeadingItems--;
		if(iNonHeadingItems)
		{
			AddItem(new Tools_ContextMenu_Item_Separator);
		}
	}

	m_xItemList.add_entry(pxItem);
	pxItem->SetParentSection(this);
}

Asura_Hash_ID Tools_ContextMenu_Section::GetHeadingID() const
{
	return m_pxHeading ? m_pxHeading->GetID() : 0;
}

void Tools_ContextMenu_Section::SetupDefaultActionForHeading()
{
	Asura_Hash_ID uTargetID = m_pxHeading ? Tools_ContextMenu_System::GetDefaultAction(m_pxHeading->GetID()) : 0;

	Tools_ContextMenu_Item* pxItem = FindDefaultAction(uTargetID);
	if(m_pxHeading)
	{
		m_pxHeading->SetDefaultAction(pxItem);
	}
}

Tools_ContextMenu_Item* Tools_ContextMenu_Section::FindDefaultAction(Asura_Hash_ID uTargetID)
{
	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		Tools_ContextMenu_Item* pxItem = xItemLIF()->FindDefaultAction(uTargetID);
		if(pxItem) return pxItem;
	}

	return 0;
}

void Tools_ContextMenu_Section::InitialiseBeforeDisplay()
{
	m_pxCurrentItem = 0;
	for(LIF<Tools_ContextMenu_Item*> xItemLIF(&m_xItemList); !xItemLIF.done(); xItemLIF.next())
	{
		xItemLIF()->InitialiseBeforeDisplay();
	}
}

void Tools_ContextMenu_Section::SetHeading(const char* szHeading)
{
	if(!m_pxHeading) return;

	m_pxHeading->SetName(szHeading);
	m_pxHeading->SetID(szHeading);
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Tools_ContextMenu_Item **
**------------------------*/

Tools_ContextMenu_Item::Tools_ContextMenu_Item( const char* szName)
{
	m_xName = szName;
	m_uID = szName ? Asura_GetHashID(szName) : 0;
	m_pxParent = 0;
	m_bHighlight = false;
	m_iHeight = g_uMenuItemDefaultHeight;
	m_iPosX = 0;
	m_iPosY = 0;
	m_bEnabled = true;
	m_bDefaultAction = false;
	
	CalculateWidth();
 }

void Tools_ContextMenu_Item::Render(const Tools_ContextMenu& xContextMenu)
{
	//render different coloured background if needed
	u_int uBackgroundColour = GetBackgroundColour();

	if (IsHeading())
	{
		uBackgroundColour = g_uHeadingsBackgroundColour;
	}

	if(uBackgroundColour != g_uBackgroundColour)
	{
		Asura_Render::Overlay.SetRectXY( (float)m_iPosX, (float)m_iPosY, (float)m_iWidth, (float)m_iHeight);
		Asura_Render::Overlay.Colour = uBackgroundColour;		
		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		if (xContextMenu.GetCurrentSection () != this->GetParentSection ())
		{
			Asura_Render::Overlay.Colour &= 0x00ffffff;
			Asura_Render::Overlay.Colour |= 0x32000000;
			Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
		}
		else
		#else
		{
			Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
		}
		#endif
		Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;

		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();
	}

	//render the text
	if(!m_xName.IsBlank())
	{
		Asura_Handle iFontHandle = Asura_Text_Output::GetSmallFontHandle();	
		Asura_Font_System::SetFont(iFontHandle);

		Asura_Bounding_Rectangle xRect;
		xRect.MinX = (float) (m_iPosX + g_uMenuItemHorizontalBorder);
		xRect.MaxX = (float) (m_iPosX + m_iWidth - g_uMenuItemHorizontalBorder);
		xRect.MinY = (float) (m_iPosY);
		xRect.MaxY = (float) (m_iPosY + m_iHeight);

		u_int uFormat = 0;
		if(GetParentSection()->GetExpandsRight())
		{
			uFormat |= ASURA_TEXTFORMAT_LEFTJUSTIFIED;
		}
		else
		{
			uFormat |= ASURA_TEXTFORMAT_RIGHTJUSTIFIED;
		}

		u_int uTextColour = g_uStandardTextColour;
		if(!m_bEnabled)
		{
			uTextColour = g_uDisabledTextColour;
		}
		else if(m_bDefaultAction)
		{
			uTextColour = g_uDefaultActionTextColour;
		}

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		if (xContextMenu.GetCurrentSection () != this->GetParentSection ())
		{
			uTextColour = g_uTranslucencyColour | 0x00ffffff;
		}
		#endif

		if (IsHeading())
		{
			//uFormat |= ASURA_TEXTFORMAT_DROPSHADOWS;
			uTextColour = g_uHeadingsTextColour;
		}

		Asura_Font_System::Print(xRect, 0.0f, 0.0f, 0.0f, 1.0f, uTextColour, uFormat, m_xName.GetString());
	}
}


void Tools_ContextMenu_Item::SetPosition(int iX, int iY)
{
	m_iPosX = iX;
	m_iPosY = iY;
}

void Tools_ContextMenu_Item::SetWidth( int iWidth)
{
	Asura_Assert(iWidth >= m_iMinWidth, "");
	m_iWidth = iWidth;
}

bool Tools_ContextMenu_Item::Contains(int iX, int iY) const
{
	if(iX >= m_iPosX && iX < (m_iPosX+m_iWidth) &&
	   iY >= m_iPosY && iY < (m_iPosY+m_iHeight))
	{
		return true;		
	}
	return false;
}

Tools_ContextMenu_Item* Tools_ContextMenu_Item::CheckCursorPosition(int iX, int iY)
{
	return Contains(iX, iY) ? this : 0;
}

u_int Tools_ContextMenu_Item::GetBackgroundColour() const
{
	return m_bHighlight ? g_uHighlightBackgroundColour : g_uBackgroundColour;
}

void Tools_ContextMenu_Item::SetParentSection(Tools_ContextMenu_Section* pxParent)
{
	Asura_Assert(!m_pxParent, "Parent already set");
	m_pxParent = pxParent;
}

Tools_ContextMenu_Item* Tools_ContextMenu_Item::FindDefaultAction(Asura_Hash_ID uTargetID)
{
	if(uTargetID == m_uID && IsEnabled())
	{
		m_bDefaultAction = true;
		return this;
	}
	return 0;
}

void Tools_ContextMenu_Item::InitialiseBeforeDisplay()
{
	m_bHighlight = false;
	m_bDefaultAction = false;
}

void Tools_ContextMenu_Item::SetID(const char* szIDName)
{
	m_uID = szIDName ? Asura_GetHashID(szIDName) : 0 ;
}

void Tools_ContextMenu_Item::SetName(const char* szName)
{
	m_xName = szName;
	CalculateWidth();
}

void Tools_ContextMenu_Item::CalculateWidth()
{
	if(!m_xName.IsBlank())
	{
		float fWidth = Asura_Text_Output::WidthOfString(m_xName.GetString(), ASURA_TEXT_FORMAT_SMALLFONT);
		m_iMinWidth = Asura_Maths::FloatToInt(fWidth * ASURA_TEXT_OUTPUT_FONT_WIDTH);
	}
	else
	{
		m_iMinWidth = 0;
	}
	m_iMinWidth += g_uMenuItemHorizontalBorder*2;
	m_iWidth = m_iMinWidth;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Tools_ContextMenu_Item_Submenu **
**--------------------------------*/

Tools_ContextMenu_Item_Submenu::Tools_ContextMenu_Item_Submenu(const char* szName)
:Tools_ContextMenu_Item(szName)
{
	m_bSubmenuOpen = false;
	m_pxMenuSection = new Tools_ContextMenu_Section(szName, false);
}

Tools_ContextMenu_Item_Submenu::~Tools_ContextMenu_Item_Submenu()
{
	delete m_pxMenuSection;
}


void Tools_ContextMenu_Item_Submenu::Render(const Tools_ContextMenu& xContextMenu)
{
	PARENT::Render(xContextMenu);

	u_int uTextColour = IsEnabled() ? g_uStandardTextColour : g_uDisabledTextColour;
	
	//draw the arrow symbol on the left or right
	if(GetParentSection()->GetExpandsRight())
	{
		float fX = (float) (GetPosX()+GetWidth()+1-g_uMenuItemHorizontalBorder);
		float fY = (float) GetPosY();

		Asura_Text_Output::ConvertPixelsToPrintCoords(fX, fY);

		char cSymbol = 5;
		Asura_Text_Output::Print(fX,fY,uTextColour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",cSymbol);
	}
	else
	{
		float fX = (float) (GetPosX()+1);
		float fY = (float) GetPosY();

		Asura_Text_Output::ConvertPixelsToPrintCoords(fX, fY);

		char cSymbol = 1;
		Asura_Text_Output::Print(fX,fY,uTextColour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",cSymbol);
	}

}

void Tools_ContextMenu_Item_Submenu::RenderSubmenus(const Tools_ContextMenu& xContextMenu)
{
	if(m_bSubmenuOpen)
	{
		m_pxMenuSection->Render(xContextMenu);
	}
}

void Tools_ContextMenu_Item_Submenu::SetCurrentItem(bool bSet)
{
	if(bSet && IsEnabled())
	{
		//determine whether the menu should be opened to the left or the right
		int iViewWidth = Asura_Viewport::GetWidth();

		int iWidthRequired = GetParentSection()->GetWidth();

		int iAvailableRight = iViewWidth - (GetPosX()+GetWidth());
		int iAvailableLeft = GetPosX();

		bool bParentRight = GetParentSection()->GetExpandsRight();
		bool bRight;

		if(iAvailableRight >= iWidthRequired && iAvailableLeft >= iWidthRequired)
		{
			//can fit either way , so match with parent
			bRight = bParentRight;
		}
		else if(iAvailableRight >= iWidthRequired)
		{
			bRight = true;
		}
		else if(iAvailableLeft >= iWidthRequired)
		{
			bRight = false;
		}
		else
		{
			bRight = (iAvailableRight > iAvailableLeft) ? true : false;
		}

		int iSubMenuX = bRight ? (GetPosX() + GetWidth()) : GetPosX();

		//check for up or down
		int iViewHeight = Asura_Viewport::GetHeight();

		int iHeightRequired = GetParentSection()->GetHeight();

		int iAvailableUp = GetPosY()+GetHeight();
		int iAvailableDown = iViewHeight - GetPosY(); 

		bool bParentDown = GetParentSection()->GetExpandsDown();
		bool bDown;

		if(iAvailableDown >= iHeightRequired && iAvailableUp >= iHeightRequired)
		{
			//can fit either way , so match with parent
			bDown = bParentDown;
		}
		else if(iAvailableDown >= iHeightRequired)
		{
			bDown = true;
		}
		else if(iAvailableUp >= iHeightRequired)
		{
			bDown = false;
		}
		else
		{
			bDown = (iAvailableDown > iAvailableUp) ? true : false;
		}

		int iSubMenuY = bDown ? (GetPosY()) : (GetPosY()+GetHeight());

			
		//set the direction and position for the submenu
		m_pxMenuSection->SetExpandsRight(bRight);
		m_pxMenuSection->SetExpandsDown(bDown);
		m_pxMenuSection->SetPosition(iSubMenuX, iSubMenuY);

		
		
		m_pxMenuSection->SetCurrentSection(true);
		m_bSubmenuOpen = true;
	}
	else
	{
		m_pxMenuSection->SetCurrentSection(false);
		m_bSubmenuOpen = false;
	}

	Tools_ContextMenu_Item::SetCurrentItem(bSet);
}


void Tools_ContextMenu_Item_Submenu::CalculateSize()
{
	//determine the dimensions of the child menu
	m_pxMenuSection->CalculateSize();
}

Tools_ContextMenu_Item* Tools_ContextMenu_Item_Submenu::CheckCursorPosition(int iX, int iY)
{
	//check the submenu if open
	if(m_bSubmenuOpen)
	{
		Tools_ContextMenu_Item* pxItem = m_pxMenuSection->CheckCursorPosition(iX, iY);
		if(pxItem)
		{
			return pxItem;
		}		
	}
	//otherwise check the parent class to see if the cursor is over this
	return PARENT::CheckCursorPosition(iX, iY);
}

Tools_ContextMenu_Item* Tools_ContextMenu_Item_Submenu::FindDefaultAction(Asura_Hash_ID uTargetID)
{
	Tools_ContextMenu_Item* pxItem = m_pxMenuSection->FindDefaultAction(uTargetID);
	if(pxItem) SetDefaultAction(true);

	return pxItem;
}

void Tools_ContextMenu_Item_Submenu::InitialiseBeforeDisplay()
{
	PARENT::InitialiseBeforeDisplay();

	m_pxMenuSection->InitialiseBeforeDisplay();
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Tools_ContextMenu_Item_Function **
**---------------------------------*/

Tools_ContextMenu_Item_Function::Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(void))
:Tools_ContextMenu_Item(szName)
{
	InitialiseFunctionPointers();

	Asura_Assert(pfnFunction, "");	
	m_pfnFunction = pfnFunction;
}

Tools_ContextMenu_Item_Function::Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(void*), void* pUserData)
:Tools_ContextMenu_Item(szName)
{
	InitialiseFunctionPointers();

	Asura_Assert(pfnFunction, "");	
	m_pfnFunction_Ptr = pfnFunction;
	m_pUserData = pUserData;
}

Tools_ContextMenu_Item_Function::Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(void*,u_int), void* pUserData, u_int uUserData)
:Tools_ContextMenu_Item(szName)
{
	InitialiseFunctionPointers();

	Asura_Assert(pfnFunction, "");	
	m_pfnFunction_PtrUint = pfnFunction;
	m_pUserData = pUserData;
	m_uUserData = uUserData;
}

Tools_ContextMenu_Item_Function::Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(int), int iUserData)
:Tools_ContextMenu_Item(szName)
{
	InitialiseFunctionPointers();

	Asura_Assert(pfnFunction, "");	
	m_pfnFunction_Int = pfnFunction;
	m_uUserData = static_cast< u_int >( iUserData );
}



void Tools_ContextMenu_Item_Function::HandleClick()
{
	if(IsEnabled())
	{
		if(m_pfnFunction)
		{
			m_pfnFunction();
		}
		else if(m_pfnFunction_Ptr)
		{
			m_pfnFunction_Ptr(m_pUserData);
		}
		else if(m_pfnFunction_Int)
		{
			m_pfnFunction_Int(static_cast< int >( m_uUserData ));
		}
		else if(m_pfnFunction_PtrUint)
		{
			m_pfnFunction_PtrUint(m_pUserData,m_uUserData);
		}
	}
}

void Tools_ContextMenu_Item_Function::InitialiseFunctionPointers()
{
	m_pfnFunction = 0;
	m_pfnFunction_Ptr = 0;
	m_pfnFunction_Int = 0;
	m_pfnFunction_PtrUint = 0;

	m_pUserData = 0;
	m_uUserData = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------**
** Tools_ContextMenu_Item_Tick **
**-----------------------------*/

Tools_ContextMenu_Item_Tick::Tools_ContextMenu_Item_Tick(const char* szName, bool* pbTicked, bool* pbChanged /*= 0*/)
:Tools_ContextMenu_Item(szName)
{
	Asura_Assert(pbTicked, "");

	m_pbTicked = pbTicked;
	m_pbChanged = pbChanged;

	m_bTicked = pbTicked ? *pbTicked : false;
}

void Tools_ContextMenu_Item_Tick::Render(const Tools_ContextMenu& xContextMenu)
{
	PARENT::Render(xContextMenu);

	u_int uTextColour = IsEnabled() ? g_uStandardTextColour : g_uDisabledTextColour;
	
	if(m_bTicked)
	{
		char cSymbol = 6;

		//draw the tick symbol on the left or right
		if(GetParentSection()->GetExpandsRight())
		{
			float fX = (float) (GetPosX()+GetWidth()+1-g_uMenuItemHorizontalBorder);
			float fY = (float) GetPosY();

			Asura_Text_Output::ConvertPixelsToPrintCoords(fX, fY);

			Asura_Text_Output::Print(fX,fY,uTextColour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",cSymbol);
		}
		else
		{
			float fX = (float) (GetPosX()+1);
			float fY = (float) GetPosY();

			Asura_Text_Output::ConvertPixelsToPrintCoords(fX, fY);

			Asura_Text_Output::Print(fX,fY,uTextColour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",cSymbol);
		}
	}
}

void Tools_ContextMenu_Item_Tick::HandleClick()
{
	if(IsEnabled())
	{
		if(m_pbTicked) *m_pbTicked = !*m_pbTicked;
		if(m_pbChanged) *m_pbChanged = true;
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Tools_ContextMenu_Item_Separator **
**----------------------------------*/

Tools_ContextMenu_Item_Separator::Tools_ContextMenu_Item_Separator()
:Tools_ContextMenu_Item(0)
{
	SetHeight(g_uSeparatorHeight);
}

void Tools_ContextMenu_Item_Separator::Render(const Tools_ContextMenu& xContextMenu)
{
	//draw a horizontal line
	float fLeft = (float) GetPosX();
	float fRight = fLeft + (float) GetWidth();
	float fY = (float) GetPosY() + GetHeight()/2.0f;

	unsigned int uOutlineColour = g_uOutlineColour;

	#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
	if (xContextMenu.GetCurrentSection () != GetParentSection ())
	{
		uOutlineColour = (g_uOutlineColour & 0x00ffffff) | g_uTranslucencyColour;
	}
	#endif

	Asura_Render::Render2DLine(fLeft,fY,fRight,fY,uOutlineColour);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------------**
** Tools_ContextMenu_Item_SectionHeading **
**---------------------------------------*/

Tools_ContextMenu_Item_SectionHeading::Tools_ContextMenu_Item_SectionHeading(const char* szName)
:Tools_ContextMenu_Item(szName)
{
	m_pxDefaultActionItem = 0;
}

u_int Tools_ContextMenu_Item_SectionHeading::GetBackgroundColour() const
{
	return IsHighlight() ? g_uHighlightBackgroundColour : g_uSectionHeadingBackgroundColour;
}
