//
// Tools_AsrFileManager.cpp
//
// Created by Richard Rice - 18/10/2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Load&Save/Tools_AsrFileManager.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Memory/Asura_Mem.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Tree_Dialog.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Env_System.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Module.h"
#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING)


#include "Tools_Overlays.h"
#include "Load&Save/Tools_LoadedFileList.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_IniFile.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Module_Selection.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Render_Options.h"
#include "PC/FileSystem/Tools_FileSystem.h"



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------**
** Tools_Overlay_AsrFileManager **
**-----------------------------*/

#define TOOLS_ASRFILE_DISABLED_COLOUR (0xaaaaaa)


class Tools_Overlay_AsrFileManager : public Asura_Overlay
{
public :
	Tools_Overlay_AsrFileManager();

	void Initialise();
	void PerFrameUpdate();

private :
	void CreateTree();

	static int TreeItemCompare(const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo);

	//gets the details for for the current file based on the selection from the tree
	Tools_AsrFileState* GetCurrentFile();

	void AddMeshToTree(const Asura_Mesh* pxMesh);
	

	//overlay callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis){((Tools_Overlay_AsrFileManager*)iThis)->__FnName();}\
	void __FnName();
	#define OVERLAY_BOOL_CALLBACK(__FnName)\
	static bool __FnName(int iThis){return ((Tools_Overlay_AsrFileManager*)iThis)->__FnName();}\
	bool __FnName();

	OVERLAY_VOID_CALLBACK(TransferModules);
	OVERLAY_BOOL_CALLBACK(CanTransferModules);

	OVERLAY_VOID_CALLBACK(CreateFile);
	OVERLAY_VOID_CALLBACK(MergeFile);

	OVERLAY_VOID_CALLBACK(AddToSelection);
	OVERLAY_BOOL_CALLBACK(CanAddToSelection);


	#undef OVERLAY_VOID_CALLBACK
	#undef OVERLAY_BOOL_CALLBACK
		
	Asura_TreeDialog_Tree m_xTree;
	bool m_bVisibleFlagsChanged;

	static Asura_String s_xInitialSelection;//name of the initial selection to use in the tree
};

Asura_String Tools_Overlay_AsrFileManager::s_xInitialSelection;


Tools_Overlay_AsrFileManager::Tools_Overlay_AsrFileManager()
:Asura_Overlay("Env .ASR Files", false, &Tools_AsrFileManager::s_xOverlayPosition)
,m_bVisibleFlagsChanged(false)
{
	m_uID = TOOLS_OVERLAY_ASRFILEMANAGER;
}

void Tools_Overlay_AsrFileManager::Initialise()
{
	Asura_Dialog* pxDialog = 0;
	Asura_Vector_2 xOffset(1.0f,3.0f);
	
	//start of tabs
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset, 0, 0);
	AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	SetCurrentParentDialog(pxTabsDialog);

	float fTabsStartY = xOffset.y;
	float fTabsEndY = xOffset.y;

	//start of main tab
	pxTabsDialog->AddTab("Main");
	xOffset.y = fTabsStartY;



	CreateTree();
	pxDialog = new Asura_Tree_Dialog( xOffset, &m_xTree, 20.0f,20.0f);
	AddDialog(pxDialog);
	xOffset.y= pxDialog->m_xMaxCoord.y + 1.0f;

	//add control for resizing the tree
	SetCurrentParentDialog(0);
	pxDialog = new Asura_Resize_Dialog(pxDialog, &Tools_AsrFileManager::s_xUserSetTreeSize);
	AddDialog(pxDialog);
	SetCurrentParentDialog(pxTabsDialog);


	pxDialog = new Asura_Function_Dialog("Select Modules", xOffset, AddToSelection, (int)this);
	pxDialog->SetIsEnabledFunction(CanAddToSelection, (int)this);
	pxDialog->SetHelpText("Adds the module or contents of the file that has been selected in the tree view to the selection used by the module selection system.\nOnly available in modes where module selection is enabled.");
	AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Function_Dialog("Transfer Modules", xOffset, TransferModules, (int)this);
	pxDialog->SetIsEnabledFunction(CanTransferModules, (int)this);
	pxDialog->SetHelpText("Transfers the modules selected within the module selection system to the selected file.\nThis only works when module selection is enabled , such as in module mode.");
	AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Function_Dialog("Create File", xOffset, CreateFile, (int)this);
	pxDialog->SetHelpText("Create a new .ASR file for storing the environment modules.");
	AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Merge File", xOffset, MergeFile, (int)this);
	pxDialog->SetHelpText("Merges the contents of the selected .ASR file into another file , and deletes the old file..");
	AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	//end of main tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//start of visibility tab
	pxTabsDialog->AddTab("Visible");
	xOffset.y = fTabsStartY;

	//add a tickbox for each file
	for(LIF<Tools_AsrFileState*> xFileLIF(&Tools_AsrFileManager::s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		pxDialog = new Asura_Tickbox_Dialog( xFileLIF()->m_xStrippedName.GetString(), xOffset, &xFileLIF()->m_bVisible, &m_bVisibleFlagsChanged);
		pxDialog->SetHelpText("If unticked then the modules belonging to this file will be hidden");
		AddDialog(pxDialog);
		xOffset.y += Asura_Dialog::DefaultSpacing();
	}


	//end of visibility tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//end of tabs
	xOffset.y = fTabsEndY + 1.0f;
	SetCurrentParentDialog(0);

	static u_int s_uSelectedTab = 0;
	pxTabsDialog->SetSelectedTabPointer(&s_uSelectedTab);


	MakeDialogsRelativeToResizableDialog();
	Asura_Overlay_System::SubmitNewOverlay(this);
}

void Tools_Overlay_AsrFileManager::CreateTree()
{
	//add entries for each file to the top layer of the tree.
	for(CLIF<Tools_AsrFileState*> xFileLIF(&Tools_AsrFileManager::s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		if(xFileLIF()->m_bToBeDeleted) continue;

		u_int uColour = 0;
		if(!Tools_SaveAsura::CanSaveFile(xFileLIF()->m_xRelativeName.GetString(),false))
		{
			uColour = TOOLS_ASRFILE_DISABLED_COLOUR;
		}

		Asura_TreeDialog_Item* pxItem = m_xTree.InsertItem(TreeItemCompare,xFileLIF()->m_xStrippedName.GetString());
		pxItem->SetColourOverride(uColour);
	}

	//add each module mesh as a child of the appropriate file
	for(int i=0; i<Asura_Module_System::GetNumberOfModules(); i++)
	{
		const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		AddMeshToTree(pxMesh);
	}

	if(!s_xInitialSelection.IsBlank())
	{
		Asura_TreeDialog_Item* pxSelectedItem = m_xTree.FindByName(s_xInitialSelection.GetString());
		if(pxSelectedItem)
		{
			m_xTree.SetSelectedItem(pxSelectedItem);
		}
		s_xInitialSelection.MakeNull();
	}
}

int Tools_Overlay_AsrFileManager::TreeItemCompare(const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo)
{
	return stricmp(xOne.GetName(),xTwo.GetName());
}

void Tools_Overlay_AsrFileManager::AddMeshToTree(const Asura_Mesh* pxMesh)
{
	pxMesh = pxMesh->GetDefaultPlatformMesh();
		
	if(pxMesh->m_xStrippedFilename.IsBlank())
	{
		Asura_Assert(0,"Module mesh is missing the name of its file");
		return;
	}

	//get the parent for this mesh
	Asura_TreeDialog_Item* pxParent = m_xTree.GetRoot()->FindChildByName(pxMesh->m_xStrippedFilename.GetString(), true);
	if(!pxParent)
	{
		Asura_Assertf(0,("Couldn't find an entry for %s in the env .asr tree",pxMesh->m_xStrippedFilename.GetString()));
	}

	//and add as a child
	Asura_TreeDialog_Item* pxItem = m_xTree.InsertItem(TreeItemCompare, pxParent, pxMesh->Name.GetString());
	pxItem->SetColourOverride( pxParent->GetColourOverride());
}

bool Tools_Overlay_AsrFileManager::CanTransferModules()
{
	return Tools_Module_Selection::IsActive();
}

void Tools_Overlay_AsrFileManager::TransferModules()
{
	//make sure some modules have been seleted
	if(!Tools_Module_Selection::GetNumSelected())
	{
		ShowMessageOverlay("No modules have been selected in the module selection system");
		return;
	}

	//determine the destination file.
	Tools_AsrFileState* pxDestFile = GetCurrentFile();
	if(!pxDestFile)
	{
		ShowMessageOverlay("A destination file needs to be selected first by selecting an entry from the tree view");
		return;
	}

	//the destination needs to be writeable
	if(!Tools_SaveAsura::CanSaveFile(pxDestFile->m_xRelativeName.GetString(),false))
	{
		ShowMessageOverlay("Can't transfer modules since the destination isn't writeable");
		return;
	}

	if(!ShowYesNoOverlay("Transfer the %d selected module(s) to %s?",Tools_Module_Selection::GetNumSelected(),pxDestFile->m_xStrippedName.GetString()))
	{
		return;
	}

	//attempt to move all the selected modules
	int iNumMoved = 0;
	int iNumFailedMoves = 0;

	for(CLIF<int> xSelLIF(Tools_Module_Selection::GetSelectionList()); !xSelLIF.done(); xSelLIF.next())
	{
		int iModule = xSelLIF();
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModule);
		pxMesh = pxMesh->GetDefaultPlatformMesh();

		//make sure the source file is writeable
		Tools_AsrFileState* pxSourceFile = Tools_AsrFileManager::GetFile(pxMesh->m_xStrippedFilename.GetString());
		if(!pxSourceFile)
		{
			Asura_Assert(0,"Failed to find file that mesh belongs to");
			continue;
		}

		if(pxSourceFile == pxDestFile) continue;

		if(!Tools_SaveAsura::CanSaveFile(pxSourceFile->m_xRelativeName.GetString(),false))
		{
			iNumFailedMoves++;
			continue;
		}

		//update all the platform meshes for this module with the new filename
		for(Asura_Mesh* pxPlatMesh = pxMesh; pxPlatMesh; pxPlatMesh = pxPlatMesh->m_pxNextPlatformMesh)
		{
			pxPlatMesh->m_xStrippedFilename = pxDestFile->m_xStrippedName;
		}

		//update the tree
		Asura_TreeDialog_Item* pxItem = m_xTree.FindByName(pxMesh->Name.GetString());
		m_xTree.DeleteItem(pxItem);
		AddMeshToTree(pxMesh);

		//mark the source and destination as modified
		pxSourceFile->m_bModified = true;
		pxDestFile->m_bModified = true;

		iNumMoved++;		
	}

	Tools_AsrFileManager::UpdateModuleVisibility();

	char acMessagePart1[200];
	Asura_PrintfToBfr(acMessagePart1,"Transferred %d module(s) to %s\n",sizeof(acMessagePart1),true,iNumMoved, pxDestFile->m_xStrippedName.GetString());

	char acMessagePart2[200] = "";
	if(iNumFailedMoves)
	{
		Asura_PrintfToBfr(acMessagePart2,"Failed to transfer %d module(s) because the source file(s) weren't writeable",sizeof(acMessagePart2),true,iNumFailedMoves, pxDestFile->m_xStrippedName.GetString());
	}

	ShowMessageOverlay("%s%s",acMessagePart1,acMessagePart2);
}

Tools_AsrFileState* Tools_Overlay_AsrFileManager::GetCurrentFile()
{
	//determine the current file by looking at the selected item in the tree view
	const Asura_TreeDialog_Item* pxSelection = m_xTree.GetSelectedItem();
	if(!pxSelection) return 0;

	//get the parent of the selection if needed
	while(pxSelection->GetDepth() > 1)
	{
		pxSelection = pxSelection->GetParent();
	}

	return Tools_AsrFileManager::GetFile(pxSelection->GetName());
}

void Tools_Overlay_AsrFileManager::CreateFile()
{
	//make sure the file list is writeable 
	if(!Tools_AsrFileManager::CanSaveFileListFile(false))
	{
		ShowMessageOverlay("Can't create a new file since the file containing the list of .asr files for the level isn't writeable");
		return;
	}

	//ask for the name for the new file
	GetString_Overlay xOverlay("Enter name for the new module ASR file","");
	if(!xOverlay.DoModal()) return;

	const char* szString = xOverlay.GetString();
	if(!szString || !*szString) return;

	if(strchr(szString,'\\') || strchr(szString,'/'))
	{
		ShowMessageOverlay("Filename shouldn't contain a path");
		return;
	}

	//build the stripped filename
	char szStrippedName[MAX_PATH];
	strcpy(szStrippedName,szString);
	char* pcDot = strchr(szStrippedName, '.');
	if(pcDot) pcDot = 0;
	if(!*szStrippedName) return;
	strcat(szStrippedName,".asr");
	
	//make sure we don't already have an entry for this file
	//(if there is an existing entry that is about to be deleted it can be ressurected though)
	Tools_AsrFileState* pxExistingFile = Tools_AsrFileManager::GetFile(szStrippedName);
	if(pxExistingFile && !pxExistingFile->m_bToBeDeleted)
	{
		ShowMessageOverlay("A file with that name already exists");
		return;
	}

	//build the relative filename
	const Asura_String& xMainRelativeName = Tools_AsrFileManager::s_pxMainFile->m_xRelativeName;
	
	Asura_String xPath;
	xMainRelativeName.SplitFilename(&xPath, 0, 0);


	char szRelativeName[MAX_PATH];
	if(xPath.IsBlank())
	{
		strcpy(szRelativeName,szStrippedName);
	}
	else
	{
		Asura_PrintfToBfr(szRelativeName,"%s\\%s",sizeof(szRelativeName),true,xPath.GetString(),szStrippedName);
	}

	//make sure the file is okay for saving
	if(!Tools_SaveAsura::CanSaveFile(szRelativeName,false))
	{
		ShowMessageOverlay("Can't create file because %s isn't writeable",szRelativeName);
		return;

	}

	//add an entry to the file list
	Tools_AsrFileState* pxFile;
	if(pxExistingFile)
	{
		pxExistingFile->m_xRelativeName = szRelativeName;
		pxExistingFile->m_xStrippedName = szStrippedName;
		pxExistingFile->m_bToBeDeleted = false;
		pxFile = pxExistingFile;
	}
	else
	{
		pxFile = Tools_AsrFileManager::AddToFileList(szRelativeName,false);
	}
	pxFile->m_bModified = true;
	Tools_AsrFileManager::s_bUsingMultipleAsrFiles = true;

	//Update the filelist text file
	Tools_AsrFileManager::SetFileListFileModified();

	Tools_AsrFileManager::UpdateModuleVisibility();

	//store the name of the new file to use as the new selection when the overlay is rebuilt
	s_xInitialSelection = pxFile->m_xStrippedName.GetString();

	//rebuild the overlay to update the tickboxes
	//'this' will be deleted beyond this point , so this needs to be the last thing that happens
	Tools_AsrFileManager::CloseOverlay();
	Tools_AsrFileManager::CreateOverlay();
}

void Tools_Overlay_AsrFileManager::MergeFile()
{
	//make sure a file is selected
	Tools_AsrFileState* pxSourceFile = GetCurrentFile();
	if(!pxSourceFile)
	{
		ShowMessageOverlay("An .ASR file needs to be selected in the tree view first");
		return;
	}
	//ignore the main file
	if(pxSourceFile->m_bMainFile)
	{
		ShowMessageOverlay("Can't use merge file on the main asura file");
		return;
	}

	//the text file needs to be writeable
	if(!Tools_AsrFileManager::CanSaveFileListFile(false))
	{
		ShowMessageOverlay("Can't merge a file into another since the file containing the list of .asr files for the level isn't writeable");
		return;
	}


	//the source file needs to be writeable
	if(!Tools_SaveAsura::CanSaveFile(pxSourceFile->m_xRelativeName.GetString(),false))
	{
		ShowMessageOverlay("Can't merge the contents of %s into another file because it isn't writeable",pxSourceFile->m_xStrippedName.GetString());
		return;
	}

	//ask which file to merge the file into
	char acPrompt[200];
	Asura_PrintfToBfr(acPrompt,"Merge %s into which file?",sizeof(acPrompt),true,pxSourceFile->m_xStrippedName.GetString());
	ListSelection_Overlay xOverlay(acPrompt);

	for(LIF<Tools_AsrFileState*> xFileLIF(&Tools_AsrFileManager::s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		if(xFileLIF() == pxSourceFile) continue;
		if(xFileLIF()->m_bToBeDeleted) continue;

		xOverlay.AddItem(xFileLIF()->m_xStrippedName.GetString(), (int) xFileLIF());
	}
	if(!xOverlay.DoModal()) return;

	Tools_AsrFileState* pxDestFile = (Tools_AsrFileState*) xOverlay.GetSelectionData();
	
	
	
	
	//See if there are any modules in the source file
	bool bAnyModulesToMove = false;

	for(int i=0; i<Asura_Module_System::GetNumberOfModules(); i++)
	{
		const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		pxMesh = pxMesh->GetDefaultPlatformMesh();

		Asura_Assert(!pxMesh->m_xStrippedFilename.IsBlank(), "");

		if(!pxMesh->m_xStrippedFilename.Compare(pxSourceFile->m_xStrippedName))
		{
			bAnyModulesToMove = true;
			break;
		}
	}

	Tools_AsrFileState* pxMainFile = Tools_AsrFileManager::s_pxMainFile;
	Asura_Assert(pxMainFile->m_bMainFile, "");

	//If there are any modules to move then the main file needs to be writeable
	if(!Tools_SaveAsura::CanSaveFile(pxDestFile->m_xRelativeName.GetString(),false))
	{
		ShowMessageOverlay("Can't merge the contents of %s into %s because the destination file isn't writeable",pxSourceFile->m_xStrippedName.GetString(),pxDestFile->m_xStrippedName.GetString());
		return;
	}

	//remove the old file from the tree
	Asura_TreeDialog_Item* pxFileItem = m_xTree.GetRoot()->FindChildByName(pxSourceFile->m_xStrippedName.GetString(), true);
	m_xTree.DeleteItem(pxFileItem);

	
	//flag each mesh in the old file to use the new file
	if(bAnyModulesToMove)
	{
		for(int i=0; i<Asura_Module_System::GetNumberOfModules(); i++)
		{
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
			pxMesh = pxMesh->GetDefaultPlatformMesh();

			Asura_Assert(!pxMesh->m_xStrippedFilename.IsBlank(), "");

			if(!pxMesh->m_xStrippedFilename.Compare(pxSourceFile->m_xStrippedName))
			{
				for(Asura_Mesh* pxPlatMesh = pxMesh; pxPlatMesh; pxPlatMesh = pxPlatMesh->m_pxNextPlatformMesh)
				{
					pxPlatMesh->m_xStrippedFilename = pxDestFile->m_xStrippedName;
				}

				//update the tree
				AddMeshToTree(pxMesh);

			}
		}
		pxDestFile->m_bModified = true;
	}

	pxSourceFile->m_bModified = true;
	pxSourceFile->m_bToBeDeleted = true;

	//flag the file list as modified
	Tools_AsrFileManager::SetFileListFileModified();

	//see if there is more than one asr file left in use
	int iNumAsrFiles = 0;
	for(LIF<Tools_AsrFileState*> xFileLIF(&Tools_AsrFileManager::s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		if(xFileLIF()->m_bToBeDeleted) continue;

		iNumAsrFiles++;
	}
		
	Tools_AsrFileManager::s_bUsingMultipleAsrFiles = (iNumAsrFiles > 1);

	Tools_AsrFileManager::UpdateModuleVisibility();

	//store the current selection for re-selecting it when the overlay is rebuilt
	Asura_TreeDialog_Item* pxSelection = m_xTree.GetSelectedItem();
	if(pxSelection)
	{
		s_xInitialSelection = pxSelection->GetName();
	}

	//rebuild the overlay to update the tickboxes
	//'this' will be deleted beyond this point , so this needs to be the last thing that happens
	Tools_AsrFileManager::CloseOverlay();
	Tools_AsrFileManager::CreateOverlay();

}

void Tools_Overlay_AsrFileManager::PerFrameUpdate()
{
	Asura_Overlay::PerFrameUpdate();

	if(m_bVisibleFlagsChanged)
	{
		m_bVisibleFlagsChanged = false;
		Tools_AsrFileManager::UpdateModuleVisibility();
	}
}

bool Tools_Overlay_AsrFileManager::CanAddToSelection()
{
	return Tools_Module_Selection::IsActive();
}

void Tools_Overlay_AsrFileManager::AddToSelection()
{
	const Asura_TreeDialog_Item* pxSelection = m_xTree.GetSelectedItem();
	if(!pxSelection) return;

	const char* szName = pxSelection->GetName();
	bool bFileSelected = (pxSelection->GetDepth() == 1);

	//Go through the modules and either select the module with the matching name ,or select
	//all modules in the matching file

	for(int i=0; i<Asura_Module_System::GetNumberOfModules(); i++)
	{
		const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		pxMesh = pxMesh->GetDefaultPlatformMesh();

		if(bFileSelected)
		{
			if(!pxMesh->m_xStrippedFilename.Compare(szName))
			{
				Tools_Module_Selection::AddToSelection(i);
			}
		}
		else
		{
			if(!pxMesh->Name.Compare(szName))
			{
				Tools_Module_Selection::AddToSelection(i);
				break;
			}
		}
	}
	
	
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Tools_AsrFileManager **
**----------------------*/

bool Tools_AsrFileManager::s_bUsingMultipleAsrFiles = false;

int Tools_AsrFileManager::s_iNumModulesAllocated = 0;
int Tools_AsrFileManager::s_iNumModulesLoaded = 0;
Tools_TempLoadModule** Tools_AsrFileManager::s_ppxModuleStorage;

List<Tools_AsrFileState*> Tools_AsrFileManager::s_xFileList;
Tools_AsrFileState* Tools_AsrFileManager::s_pxMainFile = 0;
bool Tools_AsrFileManager::s_bFileListFileModified = false;

Asura_Vector_2 Tools_AsrFileManager::s_xOverlayPosition(30.0f,10.0f);
Asura_Vector_2 Tools_AsrFileManager::s_xUserSetTreeSize(0.0f,0.0f);



#if !defined(ASURA_USE_STREAMING)
static int SortTempLoadModuleCallback(const void* p1, const void* p2)
{
	const Tools_TempLoadModule** ppxModule1 = (const Tools_TempLoadModule**) p1;
	const Tools_TempLoadModule** ppxModule2 = (const Tools_TempLoadModule**) p2;

	const Asura_Mesh* pxMesh1 = (*ppxModule1)->m_pxModule->Mesh;
	const Asura_Mesh* pxMesh2 = (*ppxModule2)->m_pxModule->Mesh;

	return stricmp(pxMesh1->Name.GetString(), pxMesh2->Name.GetString());
}
#endif

void Tools_AsrFileManager::Shutdown()
{
	Reset();
}

void Tools_AsrFileManager::Reset()
{
	while(s_xFileList.size())
	{
		delete s_xFileList.first_entry();
		s_xFileList.delete_first_entry();
	}
	s_pxMainFile = 0;
}



bool Tools_AsrFileManager::LoadMainAsuraFile(const char* szFilename)
{
	#if defined(ASURA_USE_STREAMING)
	//Ignoring support for multiple asr files for streaming for now.
	//May be a trivial change to get it working , but it will require further investigation
	Asura_Streaming_Zone*	pxZone	= Asura_Streaming_System::CreateZone(Asura_Streaming_System::ZONE_DATA, szFilename);
	const Asura_Hash_ID		uHash	= pxZone->GetHashID();

	bool bLoaded = Asura_Streaming_System::LoadZone(Asura_Streaming_System::ZONE_DATA, uHash, false);
	
	AddToFileList(szFilename,true);
	#else
	//check for a file containing a list of .asr files to load
	Asura_String xPath,xName;
	Asura_String(szFilename).SplitFilename(&xPath,&xName,0);

	if(xPath.IsNull())
	{
		xPath = "";
	}
	else if(!xPath.IsBlank())
	{
		xPath += "\\";
	}

	char szFileListName[MAX_PATH];
	Asura_PrintfToBfr(szFileListName,"%s%s_AsrList.txt",sizeof(szFileListName),true,xPath.GetString(), xName.GetString());

	Tools_IniFile xIni(szFileListName);
	xIni.LoadIniFile();

	bool bLoaded = true;
	
	//load each file listed
	while(xIni.GetNextIniTextLine())
	{
		Asura_String xAsrFilename;
		if(!xIni.GetString(xAsrFilename)) continue;

		Asura_String xAsrName;
		xAsrFilename.SplitFilename(0,&xAsrName,0);

		//leave the main asr file till last
		if(!xName.Compare(xAsrName)) continue;

//		char szFullFilename[MAX_PATH];
//		Asura_PrintfToBfr(szFullFilename,"%s%s.asr",sizeof(szFullFilename),true,xPath.GetString(), xAsrName.GetString());

		s_bUsingMultipleAsrFiles = true;

		bLoaded &= LoadAsuraFile(xAsrFilename.GetString());

		AddToFileList(xAsrFilename.GetString(),false);
	}
	//load the main file
	bLoaded &= LoadAsuraFile(szFilename);
	
	AddToFileList(szFilename,true);		
    	
	if(s_iNumModulesLoaded)
	{
		//Finished loading all the asr files , so now need to construct the environment from all the module
		//data that has been stored up
		
		//Sort all the loaded modules alphabetically
		qsort(s_ppxModuleStorage, s_iNumModulesLoaded, sizeof(Tools_TempLoadModule*), SortTempLoadModuleCallback);

		//Create the env and allocate the module storage
		Asura_Environment* pxEnvironment = Asura_Environment_System::CreateEnvironment(xName.GetString());

		Asura_Environment_System::SetCurrentEnvironment(pxEnvironment);

		pxEnvironment->CreateModules(s_iNumModulesLoaded);

		Asura_Module_Handler& xModuleHandler = Asura_Module_System::GetCurrentHandler();
        xModuleHandler.AllocateModuleStorage( *pxEnvironment );

		//Memcpy the data into the module and environment system
		//zero our copy of the memory afterwards to prevent it from being deleted
		Asura_Environment_Module* pxDestEnvModule = pxEnvironment->Modules;

		for(int i=0;i<s_iNumModulesLoaded;i++)
		{
			Asura_Module* pxDestModule = xModuleHandler.GetModuleData(i);

			Asura_Environment_Module* pxSourceEnvModule = s_ppxModuleStorage[i]->m_pxEnvModule;
			Asura_Module* pxSourceModule = s_ppxModuleStorage[i]->m_pxModule;

			Asura_Mem::Copy(pxDestEnvModule,pxSourceEnvModule,sizeof(Asura_Environment_Module));
			Asura_Mem::Set(pxSourceEnvModule, 0, sizeof(Asura_Environment_Module));

			Asura_Mem::Copy(pxDestModule, pxSourceModule,sizeof(Asura_Module));
			Asura_Mem::Set(pxSourceModule, 0, sizeof(Asura_Module));

			delete s_ppxModuleStorage[i];
			s_ppxModuleStorage[i] = 0;

	        pxDestEnvModule++;

			//warn about duplicated module names
			if(i)
			{
				//The modules are sorted by name , so only need to check adjacent pairs for matches.
				Asura_Mesh* pxMesh1 = Asura_Module_System::GetMeshForModule(i-1);
				Asura_Mesh* pxMesh2= Asura_Module_System::GetMeshForModule(i);

				//In practice comparing the mesh pointers would be sufficient , since it is only possible to have one
				//mesh of a given name
				if(!stricmp(pxMesh1->Name.GetString(), pxMesh2->Name.GetString()))
				{
					Asura_Assertf(0,("There are two modules sharing the same name (%s).\nThis will probably happen as a result of using multiple asr files for the environment.\nSaving changes to the .asr files will likely result in loss of data , since only one of the meshes will have been loaded succesfully.",pxMesh1->Name.GetString()));
				}
			}

		}
	}

	//delete any temporary module storage
	delete [] s_ppxModuleStorage;
	s_ppxModuleStorage = 0;
	s_iNumModulesAllocated = 0;
	s_iNumModulesLoaded = 0;

	#endif

	return bLoaded;
}

void Tools_AsrFileManager::AllocateModuleStorage(int iNumModules)
{
	Asura_Assert(s_iNumModulesAllocated == s_iNumModulesLoaded,"Expecting all previously allocated modules to have been used by now");

	//allocate an enlarged array , and copy over the old values
	int iNumModulesAllocated = iNumModules + s_iNumModulesAllocated;
	Tools_TempLoadModule** ppxModuleStorage = new Tools_TempLoadModule*[iNumModulesAllocated];

	Asura_Mem::Copy(ppxModuleStorage,s_ppxModuleStorage,sizeof(Tools_TempLoadModule*)*s_iNumModulesAllocated);
	//zero the new modules
	Asura_Mem::Set(ppxModuleStorage+s_iNumModulesAllocated, 0, sizeof(Tools_TempLoadModule*)*iNumModules);

	delete [] s_ppxModuleStorage;
	s_ppxModuleStorage = ppxModuleStorage;
	s_iNumModulesAllocated = iNumModulesAllocated;
}

Tools_TempLoadModule* Tools_AsrFileManager::GetNextModule()
{
	if(s_iNumModulesLoaded < s_iNumModulesAllocated)
	{
		//return the next slot in the allocated module store
		Tools_TempLoadModule** ppxModule = &s_ppxModuleStorage[s_iNumModulesLoaded];
		s_iNumModulesLoaded++;

		*ppxModule = new Tools_TempLoadModule;
		return *ppxModule;
	}
	else
	{
		Asura_Assert(s_iNumModulesLoaded < s_iNumModulesAllocated,"Used up all the allocated module storage already");
		return 0;
	}
}

Tools_AsrFileState* Tools_AsrFileManager::AddToFileList(const char* szFilename, bool bMainFile)
{
	Tools_AsrFileState* pxFile = new Tools_AsrFileState;

	pxFile->m_xRelativeName = Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir(szFilename);
	pxFile->m_xStrippedName = Tools_FileSystem::StripPath(szFilename);
	pxFile->m_bModified = false;
	pxFile->m_bMainFile = bMainFile;
	pxFile->m_bToBeDeleted = false;
	pxFile->m_bVisible = true;

	if(bMainFile)
	{
		Asura_Assert(!s_pxMainFile,"Should only be one main asr file");
		s_pxMainFile = pxFile;
	}

	//add to a sorted list
	LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList);

	for(; !xFileLIF.done(); xFileLIF.next())
	{
		if(pxFile->m_xStrippedName.Compare(xFileLIF()->m_xStrippedName) < 0)
		{
			s_xFileList.add_entry_before(pxFile,xFileLIF());
			break;
		}
	}
	if(xFileLIF.done())
	{
		s_xFileList.add_entry(pxFile);
	}

	return pxFile;
}

void Tools_AsrFileManager::SetAsrModified(const Asura_Mesh* pxMesh)
{
	Asura_Assert(pxMesh,"");
	if(pxMesh->m_xStrippedFilename.IsBlank())
	{
		Asura_Assert(0,"No filename available for mesh");
		return;
	}
	else
	{
		SetAsrModified(pxMesh->m_xStrippedFilename.GetString());
	}
}

void Tools_AsrFileManager::SetAsrModified(const char* szStrippedName)
{
	Asura_Assert(szStrippedName && *szStrippedName, "");

	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();
		if(!stricmp(szStrippedName, pxFile->m_xStrippedName.GetString()))
		{
			if(!pxFile->m_bModified)
			{
				Tools_SaveAsura::WarnIfReadOnly(pxFile->m_xRelativeName.GetString());
				pxFile->m_bModified = true;
			}
			return;
		}
	}

	Asura_Assert(0,"Failed to find the file being modified");
}

void Tools_AsrFileManager::SetAllAsrModified()
{
	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();

		if(!pxFile->m_bModified)
		{
			Tools_SaveAsura::WarnIfReadOnly(pxFile->m_xRelativeName.GetString());
			pxFile->m_bModified = true;
		}
	}
}

bool Tools_AsrFileManager::IsModified()
{
	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();

		if(pxFile->m_bModified)
		{
			return true;
		}
	}
	if(s_bFileListFileModified) return true;

	return false;
}

void Tools_AsrFileManager::SetFilesToCheckForExternalModification()
{
	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();

		Tools_LoadedFileList::TrackExternalModification(pxFile->m_xRelativeName.GetString());
	}
}

void Tools_AsrFileManager::FlagEnvironmentFilesAsLoaded()
{
	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();

		Tools_LoadedFileList::AddFile(pxFile->m_xRelativeName.GetString());
	}
}


void Tools_AsrFileManager::SaveAll()
{
	if(Tools_Level_Export::IsExport())
	{
		//during exports we just save one file that contains all the meshes
		//get the filename from Tools_SaveAsura since the environment filename
		//will have been modified for the export
		Tools_SaveAsura::SaveMainAsrFile(Tools_SaveAsura::GetMainFilename());
		return;
	}

	//save all our files
	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done();)
	{
		Tools_AsrFileState* pxFile = xFileLIF();
		{
			if(pxFile->m_bToBeDeleted)
			{
				DeleteFile(pxFile->m_xRelativeName.GetString());
				delete pxFile;
				xFileLIF.delete_current();
			}
			else 
			{
				if(Tools_SaveAsura::SaveMainAsrFile(pxFile->m_xRelativeName.GetString()))
				{
					pxFile->m_bModified = false;
				}
				xFileLIF.next();
			}
		}
	}

	if(s_bFileListFileModified)
	{
		SaveFileListFile();
	}
	
}

void Tools_AsrFileManager::SaveModified()
{
	if(Tools_Level_Export::IsExport())
	{
		Asura_Assert(0,"Tools_AsrFileManager::SaveModified() shouldn't be being called during exports");
		return;
	}

	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done();)
	{
		Tools_AsrFileState* pxFile = xFileLIF();
		if(pxFile->m_bToBeDeleted)
		{
			DeleteFile(pxFile->m_xRelativeName.GetString());
			delete pxFile;
			xFileLIF.delete_current();
		}
		else
		{
			if(pxFile->m_bModified)
			{
				if(Tools_SaveAsura::SaveMainAsrFile(pxFile->m_xRelativeName.GetString()))
				{
					pxFile->m_bModified = false;
				}
			}
			xFileLIF.next();
		}
	}
	if(s_bFileListFileModified)
	{
		SaveFileListFile();
	}
}

void Tools_AsrFileManager::SetFileListFileModified()
{
	if(s_bFileListFileModified) return;
	s_bFileListFileModified = true;

	Asura_String xFilename;
	GetFileListFilename(xFilename);

	Tools_SaveAsura::WarnIfReadOnly(xFilename.GetString());

}

void Tools_AsrFileManager::GetFileListFilename(Asura_String& xFilename)
{
	Asura_Assert(s_xFileList.size(),"");

	//build the filename based on the name of the main asr file
	const Asura_String& xAsrFilename = s_pxMainFile->m_xRelativeName;

	Asura_String xPath,xName;
	xAsrFilename.SplitFilename(&xPath, &xName, 0);

	if(xPath.IsNull())
	{
		xPath = "";
	}
	else if(!xPath.IsBlank())
	{
		xPath += "\\";
	}

	char szFileListName[MAX_PATH];
	Asura_PrintfToBfr(szFileListName,"%s%s_AsrList.txt",sizeof(szFileListName),true,xPath.GetString(), xName.GetString());

	xFilename = szFileListName;
}

bool Tools_AsrFileManager::CanSaveFileListFile(bool bShowMessage)
{
	Asura_String xFilename;
	GetFileListFilename(xFilename);

	return Tools_SaveAsura::CanSaveFile(xFilename.GetString(), bShowMessage);
}


void Tools_AsrFileManager::MakeBackups()
{
	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();

		Tools_SaveAsura::BackupFile(pxFile->m_xRelativeName.GetString());
	}
}

void Tools_AsrFileManager::PreSaveAllAs(const char* szNewFilename)
{
	//Update all meshes that were associated with any of the asr files so that they
	//are associated with the new filename
	const char* szNewStrippedFilename = Tools_FileSystem::StripPath(szNewFilename);

	for(int i=0; i<Asura_Mesh_System::GetNumberOfMeshes(); i++)
	{
		Asura_Mesh* pxMesh = Asura_Mesh_System::GetMesh(i);
		if(!pxMesh) continue;
		if(pxMesh->m_xStrippedFilename.IsBlank()) continue;

		for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
		{
			Tools_AsrFileState* pxFile = xFileLIF();
			if(!pxFile->m_xStrippedName.Compare(pxMesh->m_xStrippedFilename))
			{
				//update this mesh to use the new file
				pxMesh->m_xStrippedFilename = szNewStrippedFilename;
				break;
			}
		}
		
	}
	
	//delete our current list of files and add a single entry for the new filename
	while(s_xFileList.size())
	{
		delete s_xFileList.first_entry();
		s_xFileList.delete_first_entry();
	}
	s_pxMainFile = 0;
	AddToFileList(szNewFilename,true);

	//close the overlay with the tree view , since the old contents will be wrong
	CloseOverlay();

	s_bUsingMultipleAsrFiles = false;
}

void Tools_AsrFileManager::CreateOverlay()
{
	#ifdef ASURA_USE_STREAMING
	//Some changes would probably need to be made to the loading process to get
	//this working with streaming builds.
	ShowMessageOverlay("Not currently implemented for streaming builds");
	#else
	if(IsOverlayOpen())
	{
		//already exists ,so just bring the overlay to the front
		Asura_Overlay_System::BringOverlayToTop(TOOLS_OVERLAY_ASRFILEMANAGER);
		return;
	}

	Tools_Overlay_AsrFileManager* pxOverlay = new Tools_Overlay_AsrFileManager();
	pxOverlay->Initialise();
	#endif
}

void Tools_AsrFileManager::CloseOverlay()
{
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_ASRFILEMANAGER);
}

void Tools_AsrFileManager::ToggleOverlay()
{
	if(IsOverlayOpen())
	{
		CloseOverlay();
	}
	else
	{
		CreateOverlay();
	}
}

bool Tools_AsrFileManager::IsOverlayOpen()
{
	return Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_ASRFILEMANAGER);
}

Tools_AsrFileState* Tools_AsrFileManager::GetFile(const char* szStrippedName)
{
	Asura_Assert(szStrippedName && *szStrippedName,"");
	if(!szStrippedName || !*szStrippedName)
	{
		return 0;
	}

	for(LIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
	{
		Tools_AsrFileState* pxFile = xFileLIF();

		if(!pxFile->m_xStrippedName.Compare(szStrippedName))
		{
			return pxFile;
		}
	}

	return 0;
}

void Tools_AsrFileManager::SaveFileListFile()
{
	Asura_String xFilename;
	GetFileListFilename(xFilename);

	if(!Tools_SaveAsura::CanSaveFile(xFilename.GetString(),true))
	{
		return;
	}

	FILE* pFile = fopen(xFilename.GetString(),"wt");
	bool bOkay = true;

	if(pFile)
	{
		//write out a simple text file containing the names of all the files other than the main file
		fprintf(pFile, "#List of additional environment asr files that contain modules/meshes for the level :\n\n");

	
		for(CLIF<Tools_AsrFileState*> xFileLIF(&s_xFileList); !xFileLIF.done(); xFileLIF.next())
		{
			const Tools_AsrFileState* pxFile = xFileLIF();
			if(pxFile->m_bMainFile) continue;
			if(pxFile->m_bToBeDeleted) continue;

			fprintf(pFile, "%s\n", pxFile->m_xRelativeName.GetString());
		}

		fclose(pFile);
	}
	else
	{
		bOkay = false;
	}

	if(bOkay)
	{
		Tools_Timed_Message_System::AddMessage(0x00ff00,"Saved %s successfully.",Tools_FileSystem::StripPath(xFilename.GetString()));
		s_bFileListFileModified = false;
	}
	else
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Unable to save %s! (Write-protected file?)",xFilename.GetString());
	}
}


void Tools_AsrFileManager::UpdateModuleVisibility()
{
	//check each module and update the hidden flag based on the visible flag for its file
	for(int i=0; i<Asura_Module_System::GetNumberOfModules(); i++)
	{
		const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		pxMesh = pxMesh->GetDefaultPlatformMesh();
		
		Asura_Assert(!pxMesh->m_xStrippedFilename.IsBlank(), "");

		const Tools_AsrFileState* pxFile = GetFile(pxMesh->m_xStrippedFilename.GetString());

		if(pxFile->m_bVisible)
		{
			Tools_Render_Options::RemoveRenderFlagModule(i, ToolsRenderFlag_AsrFileManagerHide);
		}
		else
		{
			Tools_Render_Options::AddRenderFlagModule(i, ToolsRenderFlag_AsrFileManagerHide);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Tools_TempLoadModule::Tools_TempLoadModule()
{
	m_pxModule = new Asura_Module;
	m_pxEnvModule = new Asura_Environment_Module;
}

Tools_TempLoadModule::~Tools_TempLoadModule()
{
	delete m_pxModule;
	delete m_pxEnvModule;
}
