//
// Tools_AsuraFileContents.cpp
//
// Created by Richard Rice - 31/01/03
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Load&Save/Tools_AsuraFileContents.h"

#include "ChunkSystem/Asura_Chunk_FileInfo.h"
#include "ChunkSystem/Asura_Chunk_Fragment.h"
#include "ChunkSystem/Asura_Chunk_Hierarchy.h"
//#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Mesh.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Endian.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"

#include "PC/FileSystem/Tools_FileSystem.h"

Tools_AsuraFileContents::Tools_AsuraFileContents()
{
	m_bEnvironment = false;
	m_iFileType = AsuraFile_Error;
	m_bDetermineFileTypeOnly = false;
	m_iNumAnimTables = 0;
	m_iNumTiledTexture = 0;
}

Tools_AsuraFileContents::~Tools_AsuraFileContents()
{
	DeleteAll();
}

void Tools_AsuraFileContents::DeleteAll()
{
	m_bEnvironment = false;
	m_iFileType = AsuraFile_Error;
	m_iNumAnimTables = 0;

	while(m_xMeshList.size()) m_xMeshList.delete_first_entry();
	while(m_xSkinList.size()) m_xSkinList.delete_first_entry();
	while(m_xAnimList.size()) m_xAnimList.delete_first_entry();
	while(m_xFragmentList.size()) m_xFragmentList.delete_first_entry();
}

bool Tools_AsuraFileContents::LoadAsuraFile(const char* pszFilename,bool bCheckNetwork /*=true*/)
{
	char acChunkBuffer[512];

	Asura_Assert(pszFilename,"");

	//ditch any old file information
	DeleteAll();

	m_xFilename = pszFilename;

	//get network filename if needed
	char szFilename[MAX_PATH];
	if(bCheckNetwork)
	{
		Asura_FileAccess::PlatformGetNetworkFilename(pszFilename,szFilename,MAX_PATH);
		pszFilename = szFilename;
	}
	
	//open the file
	Asura_Handle iHandle = Asura_FileAccess::Open(pszFilename);
	if(iHandle == ASURA_INVALID_HANDLE)
	{
		//May be better to let the caller deal with a false return value , rather than asserting here... RAR 07/12/2005
		//Asura_Assertf(0,("Failed to load %s",pszFilename));
		return false;
	}
	
	//Check the file header to confirm that this an asura file , and to determine the compression
	if(!Asura_FileAccess::ProcessAsuraFileHeader(iHandle))
	{
		Asura_FileAccess::Close(iHandle);
		return false;
	}
   

	// KOR - 11/09/2003 16:57:01 - Check for big-endian asura file extension (.asrgc)
	// (this happens for Gamecube character dialogue files)
	// KJFL - 27 June 2005 - extended to generic 'BigEndian' format
	{
        if( Asura_StringUtil::EndsWith(pszFilename, ASURA_STANDARD_FILE_EXTENSION_BE_SUFFIX, false) )
		{
			Asura_FileAccess::Close( iHandle );
			return true;
		}
	}
	//Start processing the chunks
	
	unsigned uFilePosition = Asura_FileAccess::GetFilePosition(iHandle);
	
	Asura_Chunk_Header* chunk = reinterpret_cast<Asura_Chunk_Header*>( acChunkBuffer );

	Asura_FileAccess::ReadData(iHandle,acChunkBuffer,4);

	m_iFileType = AsuraFile_Unknown;

	bool bOkay = true;
	while(chunk->Valid())
	{
		switch(chunk->ID)
		{ 
			case ASURA_CHUNK_ENVIRONMENT_MODULELIST :
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Header) - 4 );
				m_bEnvironment = true;
				break;

			case ASURA_CHUNK_HIERARCHY_ANIM :
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Hierarchy_Anim) - 4 );
				ProcessChunk((Asura_Chunk_Hierarchy_Anim*) acChunkBuffer, iHandle);
				break;

			case ASURA_CHUNK_HIERARCHY_SKIN :
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Hierarchy_Skin) - 4 );
				ProcessChunk((Asura_Chunk_Hierarchy_Skin*) acChunkBuffer, iHandle);
				break;

			case ASURA_CHUNK_MESH_VERTICES :
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Mesh_Vertices) - 4 );
				ProcessChunk((Asura_Chunk_Mesh_Vertices*) acChunkBuffer, iHandle);
				break;

			case ASURA_CHUNK_FRAGMENT :
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Fragment) - 4 );
				ProcessChunk((Asura_Chunk_Fragment*) acChunkBuffer, iHandle);
				break;
				
			case ASURA_CHUNK_FILEINFO :
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_FileInfo) - 4 );
				ProcessChunk((Asura_Chunk_FileInfo*) acChunkBuffer, iHandle);
				break; 

			case ASURA_CHUNK_ANIMATIONTABLE :
				m_iNumAnimTables++;
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Header) - 4 );
				break;

			case ASURA_CHUNK_TILEDTEXTURE :
				++m_iNumTiledTexture;
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Header) - 4 );
				break;


			default:
			{
				//Read rest of header in.
				Asura_FileAccess::ReadData( iHandle, &acChunkBuffer[4], sizeof(Asura_Chunk_Header) - 4 );

				break;
			}
		
		}
		if(m_bDetermineFileTypeOnly && m_iFileType != AsuraFile_Unknown)
		{
			//no need to check further
			break;
		}


		//Advance to the end of the chunk
		unsigned uNewFilePosition = Asura_FileAccess::GetFilePosition(iHandle);
		unsigned uExpectedPosition = uFilePosition+chunk->Size;
		if(uNewFilePosition != uExpectedPosition)
		{
			if(uExpectedPosition<uNewFilePosition)
			{
				//Have read in too much. Chunk reading process will be completely screwed
				Asura_Assertf(0,("Read too much while processing chunk '%c%c%c%c' . Read %u bytes , was expecting %u bytes",chunk->TextID[0],chunk->TextID[1],chunk->TextID[2],chunk->TextID[3],(unsigned)(uNewFilePosition-uFilePosition),(unsigned)chunk->Size));
				bOkay = false;
				break;
			}
			else
			{
				//Have read in too little. Skip to end of chunk
				Asura_FileAccess::SkipData(iHandle,uExpectedPosition-uNewFilePosition);

				uNewFilePosition = Asura_FileAccess::GetFilePosition(iHandle);
				if(uNewFilePosition != uExpectedPosition)
				{
					Asura_Assertf(0,("Error skipping to end of chunk '%c%c%c%c'. Read %u bytes , needed to read %u bytes.",chunk->TextID[0],chunk->TextID[1],chunk->TextID[2],chunk->TextID[3],(unsigned)(uNewFilePosition-uFilePosition),(unsigned)chunk->Size));
					bOkay = false;
					break;
				}
			}	
		}

		uFilePosition = uNewFilePosition;
		if(!Asura_FileAccess::ReadData(iHandle,acChunkBuffer,4))
		{
			Asura_Assert(0,"Error reading file");
			bOkay = false;
			break;
		}
	}
	Asura_FileAccess::Close( iHandle );

	if(!bOkay)
	{
		m_iFileType = AsuraFile_Error;
	}
	return bOkay;
}


void Tools_AsuraFileContents::ProcessChunk(const Asura_Chunk_Hierarchy_Anim* pxChunk, Asura_Handle iHandle)
{
	if(pxChunk->Version>=6)
	{
		//skip to the name
		Asura_FileAccess::SkipData(iHandle,8);
	}
	const Asura_String xName = Asura_FileAccess::ReadString(iHandle);
	m_xAnimList.add_entry(xName);

}

void Tools_AsuraFileContents::ProcessChunk(const Asura_Chunk_Hierarchy_Skin* pxChunk, Asura_Handle iHandle)
{
	
	const Asura_String xName = Asura_FileAccess::ReadString(iHandle);
	m_xSkinList.add_entry(xName);
}

void Tools_AsuraFileContents::ProcessChunk(const Asura_Chunk_Mesh_Vertices* pxChunk, Asura_Handle iHandle)
{
	
	const Asura_String xName = Asura_FileAccess::ReadString(iHandle);
	m_xMeshList.add_entry(xName);
}

void Tools_AsuraFileContents::ProcessChunk(const Asura_Chunk_Fragment* pxChunk, Asura_Handle iHandle)
{
	//read in the fragment hash
	Asura_Hash_ID uHash;
	Asura_FileAccess::ReadData( iHandle, &uHash, sizeof(Asura_Hash_ID));
	m_xFragmentList.add_entry(uHash);

}

void Tools_AsuraFileContents::ProcessChunk(const class Asura_Chunk_FileInfo* pxChunk, Asura_Handle iHandle)
{
	if(pxChunk->Version>=1)
	{
		//skip file size
		Asura_FileAccess::SkipData(iHandle,sizeof(u_int));
		//and get the file type
		Asura_FileAccess::ReadData( iHandle, &m_iFileType, sizeof(int));
	}
}

bool Tools_AsuraFileContents::IsEnvironmentFile(const char* pszFilename,bool bCheckNetwork /*= true*/)
{
	return GetAsuraFileType(pszFilename,bCheckNetwork)==AsuraFile_Environment;
}

AsuraFileType Tools_AsuraFileContents::GetFileType() const
{
	if(m_iFileType == AsuraFile_Unknown)
	{
		return GuessFileType();
	}
	return (AsuraFileType) m_iFileType;
}

AsuraFileType Tools_AsuraFileContents::GetAsuraFileType(const char* pszFilename,bool bCheckNetwork)
{
	Tools_AsuraFileContents xFileContents;
	xFileContents.m_bDetermineFileTypeOnly = true;
	xFileContents.LoadAsuraFile(pszFilename);

	return xFileContents.GetFileType();
}

AsuraFileType Tools_AsuraFileContents::GuessFileType() const
{
	if(m_iFileType != AsuraFile_Unknown) return (AsuraFileType)m_iFileType;

	if(m_bEnvironment)
	{
		return AsuraFile_Environment;
	}
	
	//check for template files via the extension
	Asura_String xExtension;
	m_xFilename.SplitFilename(0,0,&xExtension);
	if(xExtension.IsBlank()) return AsuraFile_Unknown;

	if(!xExtension.Compare("AsuraTemplate"))
	{
		return AsuraFile_Template;
	}
	if(!xExtension.Compare("FaceAnim"))
	{
		return AsuraFile_FaceAnim;
	}

	//Only attempt to guess for files with asr extension
	if(!xExtension.Compare("asr"))
	{
		if(m_xFragmentList.size())
		{
			//some sort of fragment file
			if(m_xAnimList.size())
			{
				return AsuraFile_AnimFragment;
			}
			else
			{
				return AsuraFile_Fragment;
			}
		}
	
		if(m_xSkinList.size())
		{
			return AsuraFile_Skin;	
		}	
	
		if(m_xAnimList.size())
		{
			return AsuraFile_Anim;
		}
	
		if(m_xMeshList.size())
		{
			return AsuraFile_Shapes;
		}
		if(m_iNumAnimTables)
		{
			return AsuraFile_AnimTable;
		}
		if(m_iNumTiledTexture)
		{
			return AsuraFile_TextureTiler;
		}
	}
	

	return AsuraFile_Unknown;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Tools_AsuraFileChunkLister **
**----------------------------*/

Tools_AsuraFileChunkLister::Tools_AsuraFileChunkLister()
{
	m_pcFileBuffer = 0;
	m_bBigEndian = false;
	m_bIndividualChunkAllocations = false;
}

Tools_AsuraFileChunkLister::~Tools_AsuraFileChunkLister()
{
	DeleteAll();
}

void Tools_AsuraFileChunkLister::DeleteAll()
{
	while(m_xChunkList.size())
	{
		if(m_bIndividualChunkAllocations)
		{
			delete [] (char*)(m_xChunkList.first_entry());
		}

		m_xChunkList.delete_first_entry();
	}
	delete [] m_pcFileBuffer;
	m_pcFileBuffer = 0;
	m_bBigEndian = false;
	m_bIndividualChunkAllocations = false;

}

bool Tools_AsuraFileChunkLister::LoadAsuraFile(const char* pszFilename,bool bCheckNetwork /* = true */,bool bFailureHandledExternally /* = false */)
{
	Asura_Assert(pszFilename,"");

	//ditch any old file information
	DeleteAll();

	//get network filename if needed
	char szFilename[MAX_PATH];
	if(bCheckNetwork)
	{
		Asura_FileAccess::PlatformGetNetworkFilename(pszFilename,szFilename,MAX_PATH);
		pszFilename = szFilename;
	}
	
	//read and decompress the file
	
	Tools_ReadAsuraFileOutput xReadFileOutput;
	m_pcFileBuffer = Tools_FileSystem::ReadAsuraFile(pszFilename,&xReadFileOutput,false,bFailureHandledExternally);

	if(!m_pcFileBuffer)
	{
		DeleteAll();
		return false;
	}

	u_int uFileSize = xReadFileOutput.m_uBufferSize;
	
	//Extract the chunks , trying both endianesses
	//Some files (I can't remember which) have the contents as being big endian ,even if the compression isn't
	for(int iBigEndian = m_bBigEndian ? 1 : 0; iBigEndian < 2; iBigEndian++)
	{
		
		//skip past identifier to get to the first chunk
		const char* pcChunkBuffer = m_pcFileBuffer;
		DWORD uFilePosition = 0;
	

		const Asura_Chunk_Header* pxChunk = (Asura_Chunk_Header*) pcChunkBuffer;
	
		bool bOkay = true;

		while(pxChunk->Valid())
		{
			m_xChunkList.add_entry(pxChunk);

			u_int uChunkSize = pxChunk->Size;
			if(iBigEndian) uChunkSize = Asura_Endian::SwapBytes(uChunkSize);

			if(uChunkSize < sizeof(Asura_Chunk_Header))
			{
				bOkay = false;
				break;
			}
            
			//advance to the next chunk
			pcChunkBuffer += uChunkSize;
			uFilePosition += uChunkSize;
			pxChunk = (Asura_Chunk_Header*) pcChunkBuffer;

			if(uFilePosition>=uFileSize)
			{
				//something wrong with the file , we've gone too far before finding a terminator
				bOkay = false;
				break;
			}
		}

		if(bOkay)
		{
			//make sure the rest of the file is zeroes
			for(u_int uPos=uFilePosition; uPos<uFileSize; uPos++)
			{
				if(*pcChunkBuffer++ != 0)
				{
					bOkay = false;
					break;
				}
			}
		}

		if(bOkay)
		{
			m_bBigEndian = iBigEndian ? true : false;
			return true;
		}

		//delete all the chunks ,and start again
		while(m_xChunkList.size()) m_xChunkList.delete_first_entry();

	}

	return false;
}


bool Tools_AsuraFileChunkLister::LoadAsuraFile(const char* pszFilename, const int* piChunkIDArray, int iChunkIDArraySize,bool bCheckNetwork /* = true */,bool bFailureHandledExternally /* = false */)
{
	Asura_Assert(pszFilename,"");

	//ditch any old file information
	DeleteAll();


	//get network filename if needed
	char szFilename[MAX_PATH];
	if(bCheckNetwork)
	{
		Asura_FileAccess::PlatformGetNetworkFilename(pszFilename,szFilename,MAX_PATH);
		pszFilename = szFilename;
	}

	if( Asura_StringUtil::EndsWith(pszFilename, ASURA_STANDARD_FILE_EXTENSION_BE_SUFFIX, false) )
	{
		//disallow files with bigendian extensions when using this version for now
		Asura_Assert(0,"Big endian files not handled");
		return false;
	}

	
	//open the file
	Asura_Handle iHandle = Asura_FileAccess::Open(pszFilename);
	if(iHandle == ASURA_INVALID_HANDLE)
	{
		return false;
	}

	//Check the file header to confirm that this an asura file , and to determine the compression
	if(!Asura_FileAccess::ProcessAsuraFileHeader(iHandle))
	{
		Asura_FileAccess::Close(iHandle);
		return false;
	}

	
	
	//Start processing the chunks
	
	unsigned uFilePosition = Asura_FileAccess::GetFilePosition(iHandle);
	
	Asura_Chunk_Header xChunkHeader;;
	char* pcChunkHeader = (char*)&xChunkHeader;

	Asura_FileAccess::ReadData(iHandle,&xChunkHeader,4);

	bool bOkay = true;
	while(xChunkHeader.Valid())
	{
		//read in the rest of the chunk header
		Asura_FileAccess::ReadData(iHandle,pcChunkHeader+4,sizeof(Asura_Chunk_Header)-4);

		bool bReadChunk = false;
		for(int i=0; i<iChunkIDArraySize; i++)
		{
			if(xChunkHeader.ID == piChunkIDArray[i])
			{
				bReadChunk = true;
				break;
			}
		}
		
		//if this is one of the listed chunk IDs , then allocate a buffer for it and read it in
		if(bReadChunk)
		{
			char* pcChunk = new char[xChunkHeader.Size];
			Asura_Mem::Copy(pcChunk,pcChunkHeader,sizeof(Asura_Chunk_Header));

			u_int uRemainder = xChunkHeader.Size - sizeof(Asura_Chunk_Header);
			if(uRemainder)
			{
				Asura_FileAccess::ReadData(iHandle,pcChunk+sizeof(Asura_Chunk_Header),uRemainder);
			}
			m_xChunkList.add_entry((const Asura_Chunk_Header*)(pcChunk));

			m_bIndividualChunkAllocations = true;
		}

		//Advance to the end of the chunk
		unsigned uNewFilePosition = Asura_FileAccess::GetFilePosition(iHandle);
		unsigned uExpectedPosition = uFilePosition+xChunkHeader.Size;
		if(uNewFilePosition != uExpectedPosition)
		{
			if(uExpectedPosition<uNewFilePosition)
			{
				//Have read in too much. Chunk reading process will be completely screwed
				Asura_Assertf(0,("Read too much while processing chunk '%c%c%c%c' . Read %u bytes , was expecting %u bytes",xChunkHeader.TextID[0],xChunkHeader.TextID[1],xChunkHeader.TextID[2],xChunkHeader.TextID[3],(unsigned)(uNewFilePosition-uFilePosition),(unsigned)xChunkHeader.Size));
				bOkay = false;
				break;
			}
			else
			{
				//Have read in too little. Skip to end of chunk
				Asura_FileAccess::SkipData(iHandle,uExpectedPosition-uNewFilePosition);

				uNewFilePosition = Asura_FileAccess::GetFilePosition(iHandle);
				if(uNewFilePosition != uExpectedPosition)
				{
					Asura_Assertf(0,("Error skipping to end of chunk '%c%c%c%c'. Read %u bytes , needed to read %u bytes.",xChunkHeader.TextID[0],xChunkHeader.TextID[1],xChunkHeader.TextID[2],xChunkHeader.TextID[3],(unsigned)(uNewFilePosition-uFilePosition),(unsigned)xChunkHeader.Size));
					bOkay = false;
					break;
				}
			}	
		}

		uFilePosition = uNewFilePosition;
		if(!Asura_FileAccess::ReadData(iHandle,&xChunkHeader,4))
		{
			Asura_Assert(0,"Error reading file");
			bOkay = false;
			break;
		}
	}
	Asura_FileAccess::Close( iHandle );

	if(!bOkay)
	{
		DeleteAll();
	}

	return bOkay;
}


const Asura_Chunk_Header* Tools_AsuraFileChunkLister::GetChunkByID(int iChunkID)
{
	for(CLIF<const Asura_Chunk_Header*> xChunkLIF(&m_xChunkList);!xChunkLIF.done();xChunkLIF.next())
	{
		if(xChunkLIF()->ID == iChunkID)
		{
			return xChunkLIF();
		}
	}
	return 0;
}

void  Tools_AsuraFileChunkLister::ListChunksByID(int iChunkID,List<const Asura_Chunk_Header*>& xOutputList)
{
	for(CLIF<const Asura_Chunk_Header*> xChunkLIF(&m_xChunkList);!xChunkLIF.done();xChunkLIF.next())
	{
		if(xChunkLIF()->ID == iChunkID)
		{
			xOutputList.add_entry(xChunkLIF());
		}
	}
}
