//
// Tools_Sound_Export.cpp
//
// Created by Richard Rice - 23/05/03
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Load&Save/Tools_Sound_Export.h"

#include "Debugging/Asura_Assert.h"

#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Sound/Conversion/Tools_AT3_Conversion.h"
#include "Sound/Conversion/Tools_Bink_Conversion.h"
#include "Sound/Conversion/Tools_DSPADPCM_Conversion.h"
#include "Sound/Conversion/Tools_MSF_Conversion.h"
#include "Sound/Conversion/Tools_Vag_Conversion.h"
#include "Sound/Conversion/Tools_XMA_Conversion.h"
#include "Sound/Conversion/Tools_xWMA_Conversion.h"

// RAR 08/07/03 - Bumped to version 2 to force regeneration of all files, to deal with possible corruption of some old files.
// MH  18/07/03 - up to v3 to force min length on streaming vags
// SK  12/10/06 - up to v4 to force vags to resample using the r8brain dll
// RAR 16/11/06 - bumped to v5 to fix sounds that were having their sample rates increased when not required
// AAM 01/06/07 - bumped to v6 to force 360 sounds to reconvert for new XMA2 format
// RAR 11/07/06 - V7 - Added bitrate for PSP AT3 conversion
// MH  15/01/08 - V8 - bumped to force PS3 to reconvert to MP3
#define TOOLS_SOUND_DAT_FILE_VERSION 8

struct ToolsSoundExportFile
{
	FILETIME m_xSourceModifiedTime;
	FILETIME m_xDestModifiedTime;
	bool m_bFirstSampleTooBig;
};

#ifndef DISABLE_R8BRAIN_DLL

HINSTANCE		Tools_Sound_Export::m_hDllInstance_R8B = NULL;
r8b_executeMem	Tools_Sound_Export::m_fnR8B_ExecuteMem = NULL;
r8b_execute		Tools_Sound_Export::m_fnR8B_Execute = NULL;
r8b_getWAVInfo	Tools_Sound_Export::m_fnR8B_GetWavInfo = NULL;

#endif // DISABLE_R8BRAIN_DLL

unsigned short Tools_Sound_Export::s_usProjectVersion = 0;

bool Tools_Sound_Export::ExportVagFile(const char* pszInputFilename,const char* pszOutputFilename,const struct ToolsSoundConversionData* pxConvertData,u_int* puFailureCode /*= 0*/,float fMinLength /*= 0.0f*/)
{
	Asura_Assert(pszInputFilename,"");
	Asura_Assert(pszOutputFilename,"");
	Asura_Assert(pxConvertData,"");

	if(puFailureCode) *puFailureCode = Tools_ConvertVagFile_Success;

	//see if the file will need creating
	ToolsSoundExportFile xExportData;
	bool bNeedToGenerate = false;

	if(DoesFileNeedExporting(pszInputFilename,pszOutputFilename,pxConvertData,&xExportData,SOUND_CONVERSION_DEFAULT))
	{
		//check network version
		Asura_String xShadowFilename;
		Tools_FileSystem::GetShadowConversionFilename(pszOutputFilename,&xShadowFilename);
		//check network version
		if(xShadowFilename.IsBlank() || DoesFileNeedExporting(pszInputFilename,xShadowFilename.GetString(),pxConvertData,&xExportData,SOUND_CONVERSION_DEFAULT))
		{
			//need to regenerate file then
			bNeedToGenerate = true;
		}
		else
		{
			bNeedToGenerate = !Tools_FileSystem::CopyFromConvertedFileShadow(pszOutputFilename);
		}
	}


	if(!bNeedToGenerate)
	{
		//file is up to date , or wasn't needed
		if(xExportData.m_bFirstSampleTooBig)
		{
			if(puFailureCode) *puFailureCode = Tools_ConvertVagFile_FirstSampleTooBig;
			return false;
		}
		else
		{
			return true;
		}
	}
	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",pszInputFilename);

	//write the vag file
	u_int uFailureCode = Tools_ConvertVagFile_Success;

	Tools_FileSystem::MakeWriteable(pszOutputFilename);

	bool bVagCreated = ConvertVagFile(pszInputFilename,pszOutputFilename,pxConvertData->m_iMaxFrequency,
				   		pxConvertData->m_bLoop,pxConvertData->m_uFirstSample,pxConvertData->m_uMaxSamples,
				   		&uFailureCode, fMinLength);

	if(puFailureCode) *puFailureCode = uFailureCode;

//	if(bVagCreated)
//	{
//		Tools_FileSystem::CopyFileTimes(pszInputFilename,pszOutputFilename);
//	}


	if(!bVagCreated && uFailureCode !=Tools_ConvertVagFile_FirstSampleTooBig)
	{
		return false;
	}

	if(uFailureCode ==Tools_ConvertVagFile_FirstSampleTooBig)
	{
		//if the file wasn't created because the first sample was too large , make sure and old
		//vag file isn't lying around
		DeleteFile(pszOutputFilename);
	}

	//write the conversion data
	WriteConversionData(pszInputFilename,pszOutputFilename,pxConvertData,(uFailureCode == Tools_ConvertVagFile_FirstSampleTooBig));

	Tools_FileSystem::CopyToConvertedFileShadow(pszInputFilename, pszOutputFilename);

	return bVagCreated;
}

bool Tools_Sound_Export::ExportDSPFile(const char* szInputFilename,const char* szOutputFilename,const struct ToolsSoundConversionData* pxConvertData,u_int* puFailureCode /*= 0*/)
{
	Asura_Assert(szInputFilename,"No input filename given");
	Asura_Assert(szOutputFilename,"No output filename given");
	Asura_Assert(pxConvertData,"No conversion data given");

	if(puFailureCode)
	{
		*puFailureCode = Tools_ConvertDSPFile_Success;
	}

	// See if the file will need creating
	if(UpdateFromShadow(szInputFilename, szOutputFilename, pxConvertData, SOUND_CONVERSION_DEFAULT))
	{
		// Up to date.
		return true;
	}

	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",szInputFilename);

	// Write the DSP file
	Tools_FileSystem::MakeWriteable(szOutputFilename);
	if(!ConvertDSPFile(szInputFilename,szOutputFilename,pxConvertData, puFailureCode))
	{
		return false;
	}

	OnExportFinish(szInputFilename, szOutputFilename, pxConvertData);

	return true;
}

bool Tools_Sound_Export::ExportXMAFile( const char* szInputFilename, const char* szOutputFilename, const ToolsSoundConversionData* pxConvertData, u_int* puFailureCode /*= 0*/ )
{
	if( !szInputFilename || !szOutputFilename /*|| !pxConvertData*/ )
	{
		Asura_Assert( false, "NULL parameter passed to Tools_Sound_Export::ExportXMAFile()." );

		if( puFailureCode )
		{
			*puFailureCode = Tools_ConvertXMAFile_InvalidParameter;
		}

		return false;
	}

	if( puFailureCode )
	{
		*puFailureCode = Tools_ConvertXMAFile_Success;
	}

	// See if the file will need creating
	if(UpdateFromShadow(szInputFilename, szOutputFilename, pxConvertData, SOUND_CONVERSION_DEFAULT))
	{
		// Up to date.
		return true;
	}

	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",szInputFilename);

	// Write the XMA file

	u_int uFailureCode = 0;

	Tools_FileSystem::MakeWriteable( szOutputFilename );
	const bool bXMACreated = Tools_ConvertXMAFile( szInputFilename, szOutputFilename, pxConvertData, &uFailureCode );

	if( puFailureCode )
	{
		*puFailureCode = uFailureCode;
	}

	if( !bXMACreated ) return false;

	OnExportFinish(szInputFilename, szOutputFilename, pxConvertData);

	return true;
}

bool Tools_Sound_Export::ExportxWMAFile( const char* szInputFilename, const char* szOutputFilename,
	const ToolsSoundConversionData* pxConvertData, u_int* puFailureCode /*= 0*/ )
{
	if( !szInputFilename || !szOutputFilename /*|| !pxConvertData*/ )
	{
		Asura_Assert( false, "NULL parameter passed to Tools_Sound_Export::ExportxWMAFile()." );

		if( puFailureCode )
		{
			*puFailureCode = Tools_ConvertxWMAFile_InvalidParameter;
		}

		return false;
	}

	if( puFailureCode )
	{
		*puFailureCode = Tools_ConvertxWMAFile_Success;
	}

	// See if the file will need creating
	if(UpdateFromShadow(szInputFilename, szOutputFilename, pxConvertData, SOUND_CONVERSION_DEFAULT))
	{
		// Up to date.
		return true;
	}

	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",szInputFilename);

	// Write the xWMA file

	u_int uFailureCode = 0;

	Tools_FileSystem::MakeWriteable( szOutputFilename );
	const bool bxWMACreated =
		Tools_ConvertxWMAFile( szInputFilename, szOutputFilename, pxConvertData, &uFailureCode );

	if( puFailureCode )
	{
		*puFailureCode = uFailureCode;
	}

	if( !bxWMACreated ) return false;

	OnExportFinish(szInputFilename, szOutputFilename, pxConvertData);

	return true;
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Tools_Sound_Export::ExportMSFFile
// Description : Expore a wave file to a PS3 MSF file (via MSenc)
//				 (copied from XMA export sequence)
//
// Parameters  :
//   const char* szInputFilename                    - Input file name
//   const char* szOutputFilename                   - Output file name
//   const ToolsSoundConversionData* pxConvertData  - Conversion information
//
// Returns     : int - Returns zero on success, nonzero error code on failure
///////////////////////////////////////////////////////////////////////////////
int Tools_Sound_Export::ExportMSFFile( const char* szInputFilename, const char* szOutputFilename, const ToolsSoundConversionData* pxConvertData )
{
	if( !szInputFilename || !szOutputFilename  )
	{
		Asura_Assert( false, "[Tools_Sound_Export::ExportMSFFile] - NULL filename parameter(s)\n" );
		return Tools_ConvertMSFFile_Failed;
	}

	// See if the file will need creating
	if(UpdateFromShadow(szInputFilename, szOutputFilename, pxConvertData, SOUND_CONVERSION_DEFAULT))
	{
		// File is up-to-date.
		return Tools_ConvertMSFFile_Success;
	}

	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",szInputFilename);

	Tools_FileSystem::MakeWriteable( szOutputFilename );

	// Write the MSF file
	int iResult = Tools_ConvertMSFFile( szInputFilename, szOutputFilename, pxConvertData );

	if ( iResult == Tools_ConvertMSFFile_Success )
	{
		OnExportFinish(szInputFilename, szOutputFilename, pxConvertData);
	}

	return iResult;
}


bool Tools_Sound_Export::ExportAT3File( const char* szInputFilename, const char* szOutputFilename, const ToolsSoundConversionData* pxConvertData, u_int* puFailureCode /*= 0*/ )
{
	if( !szInputFilename || !szOutputFilename /*|| !pxConvertData*/ )
	{
		Asura_Assert( false, "NULL parameter passed to Tools_Sound_Export::ExportAT3File()." );

		if( puFailureCode )
		{
			*puFailureCode = Tools_ConvertAT3File_FailedToReadWav;
		}

		return false;
	}

	if( puFailureCode )
	{
		*puFailureCode = Tools_ConvertAT3File_Success;
	}

	// See if the file will need creating
	ToolsSoundExportFile xExportData;
	bool bNeedToGenerate = false;

	if( DoesFileNeedExporting( szInputFilename, szOutputFilename, pxConvertData, &xExportData,SOUND_CONVERSION_AT3 ) )
	{
		// Check network version
		Asura_String xShadowFilename;
		Tools_FileSystem::GetShadowConversionFilename( szOutputFilename, &xShadowFilename );

		if( xShadowFilename.IsBlank() || DoesFileNeedExporting( szInputFilename, xShadowFilename.GetString(), pxConvertData, &xExportData, SOUND_CONVERSION_AT3 ) )
		{
			// Need to regenerate file then
			bNeedToGenerate = true;
		}
		else
		{
			bNeedToGenerate = !Tools_FileSystem::CopyFromConvertedFileShadow( szOutputFilename );
		}
	}

	if(!bNeedToGenerate)
	{
		//Force file to be regenerated if zero-sized.
		//This will deal with some old files that weren't being converted correctly
		//RAR 02/07/2007
		u_int uFileSize = 0;
		Tools_FileSystem::GetSizeOfFile(szOutputFilename,uFileSize);
		if(!uFileSize) bNeedToGenerate = true;

	}

	if( !bNeedToGenerate )
	{
		// File is up to date, or wasn't needed
		return true;
	}

	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",szInputFilename);


	// Write the AT3 file

	u_int uFailureCode = 0;

	Tools_FileSystem::MakeWriteable( szOutputFilename );
	const bool bAT3Created = Tools_ConvertAT3File( szInputFilename, szOutputFilename, pxConvertData, &uFailureCode );

	if( puFailureCode )
	{
		*puFailureCode = uFailureCode;
	}

	if( !bAT3Created ) return false;

	OnExportFinish(szInputFilename, szOutputFilename, pxConvertData);

	return true;
}

bool Tools_Sound_Export::ExportBinkFile(const char* szInputFilename, const char* szOutputFilename,const struct ToolsSoundConversionData* pxConvertData,u_int* puFailureCode /*= 0*/)
{
	Asura_Assert(szInputFilename,"No input filename given");
	Asura_Assert(szOutputFilename,"No output filename given");
	Asura_Assert(pxConvertData,"No conversion data given");

	if(puFailureCode)
	{
		*puFailureCode = Tools_ConvertBinkFile_Success;
	}

	// See if the file will need creating
	if(UpdateFromShadow(szInputFilename, szOutputFilename, pxConvertData, SOUND_CONVERSION_DEFAULT))
	{
		// Up to date.
		return true;
	}
	
	Tools_Level_Export::AddMessage(Tools_Level_Export::MT_INFORMATION,"Converting %s",szInputFilename);
	
	// Write the Bink file
	Tools_FileSystem::MakeWriteable(szOutputFilename);
	if(!Tools_ConvertBinkFile(szInputFilename,szOutputFilename,pxConvertData, puFailureCode))
	{
		// Failed.
		return false;
	}

	OnExportFinish(szInputFilename, szOutputFilename, pxConvertData);

	return true;
}


bool Tools_Sound_Export::DoesFileNeedExporting(const char* pszInputFilename,const char* pszOutputFilename,const struct ToolsSoundConversionData* pxConvertData,ToolsSoundExportFile* pxSavedExportData, ConversionType eType)
{
	char szFilename[MAX_PATH];
	unsigned long uBytesTransferred;
	bool bOkay = true;

	strcpy(szFilename,pszOutputFilename);

	//replace the fileextension
	char* dotPos = strrchr(szFilename,'.');
	if(dotPos)*dotPos = 0;
	strcat(szFilename,".dat");

	//open the file
	HANDLE handle = CreateFile(szFilename,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
	if(handle==INVALID_HANDLE_VALUE) return true;

	ToolsSoundConversionData xSavedConvertData;
	ToolsSoundExportFile xSavedExportData;

	//read in the version number
	int iVersion;
	ReadFile(handle,&iVersion,sizeof(int),&uBytesTransferred,0);

	const unsigned short* const pusVersion = reinterpret_cast<unsigned short*>( &iVersion );

	const int iEngine	= 0;
	const int iProject	= 1;

	// Check the engine version

	//old versions just aren't supported at all
	if(pusVersion[iEngine] < 5) bOkay = false;
	//and versions that are too new can't be handled
	if(pusVersion[iEngine] > TOOLS_SOUND_DAT_FILE_VERSION) bOkay = false;

	//Version 5->6 was just for forcing the xbox360 sounds to be reconverted
	if(pusVersion[iEngine] == 5 && Tools_Level_Export::s_eExportType == Tools_Level_Export::Export_Xbox360) bOkay = false;

	// version 8 is to force PS3 to be reconverted to MP3
	if (pusVersion[iEngine] < 8 && Tools_Level_Export::s_eExportType == Tools_Level_Export::Export_PS3) bOkay = false;

	//determine the size of the conversion data based on the version
	u_int uConversionDataSize = sizeof(ToolsSoundConversionData);
	if(pusVersion[iEngine] <= 6) uConversionDataSize = sizeof(ToolsSoundConversionData_V6);


	// Check the project version
	if( pusVersion[iProject] < s_usProjectVersion )
	{
		bOkay = false;
	}


	//read in the rest
	if(bOkay)
	{
		ReadFile(handle,&xSavedExportData,sizeof(ToolsSoundExportFile),&uBytesTransferred,0);
		ReadFile(handle,&xSavedConvertData,uConversionDataSize,&uBytesTransferred,0);

		if(xSavedConvertData.m_iMaxFrequency != pxConvertData->m_iMaxFrequency ||
		xSavedConvertData.m_bLoop != pxConvertData->m_bLoop ||
		xSavedConvertData.m_uFirstSample != pxConvertData->m_uFirstSample ||
		xSavedConvertData.m_uMaxSamples != pxConvertData->m_uMaxSamples)
		{
			bOkay = false;
		}

		if(eType == SOUND_CONVERSION_AT3)
		{
			//AT3 conversion needs the bitrate to match
			if(xSavedConvertData.m_uAT3BitRate != pxConvertData->m_uAT3BitRate) bOkay =  false;
		}

		if(pxSavedExportData) *pxSavedExportData = xSavedExportData;
	}

	CloseHandle(handle);

	if(!bOkay) return true;

	//check modified times
	FILETIME xModifiedTime;
	if(!Tools_FileSystem::GetModifiedTime(pszInputFilename,&xModifiedTime)) return true;
	if(Tools_FileSystem::ApproxCompareFileTime(&xModifiedTime,&xSavedExportData.m_xSourceModifiedTime)!=0) return true;

	if(!xSavedExportData.m_bFirstSampleTooBig)
	{
		//The converted file is supposed to exist , make sure it does , and has the right filetime
		if(!Tools_FileSystem::GetModifiedTime(pszOutputFilename,&xModifiedTime)) return true;
		if(Tools_FileSystem::ApproxCompareFileTime(&xModifiedTime,&xSavedExportData.m_xDestModifiedTime)!=0) return true;
	}

	//file shouldn't need re-exporting in that case
	return false;
}

bool Tools_Sound_Export::UpdateFromShadow(const char* pszInputFilename, const char* pszOutputFilename, const ToolsSoundConversionData* pxConvertData, ConversionType eType)
{
	ToolsSoundExportFile xExportData;
	bool bNeedToGenerate = false;

	if(DoesFileNeedExporting(pszInputFilename,pszOutputFilename,pxConvertData,&xExportData, eType))
	{
		//check network version
		Asura_String xShadowFilename;
		Tools_FileSystem::GetShadowConversionFilename(pszOutputFilename,&xShadowFilename);
		//check network version
		if(xShadowFilename.IsBlank() || DoesFileNeedExporting(pszInputFilename,xShadowFilename.GetString(),pxConvertData,&xExportData,eType))
		{
			//need to regenerate file then
			bNeedToGenerate = true;
		}
		else
		{
			bNeedToGenerate = !Tools_FileSystem::CopyFromConvertedFileShadow(pszOutputFilename);
		}
	}

	return !bNeedToGenerate;
}

void Tools_Sound_Export::OnExportFinish(const char* szInputFilename, const char* szOutputFilename, const ToolsSoundConversionData* pxConvertData)
{
	// Write the conversion data
	WriteConversionData( szInputFilename, szOutputFilename, pxConvertData, false );

	Tools_FileSystem::CopyToConvertedFileShadow( szInputFilename, szOutputFilename );
}


void Tools_Sound_Export::WriteConversionData(const char* pszInputFilename,const char* pszOutputFilename,const ToolsSoundConversionData* pxConvertData,bool bFirstSampleTooBig)
{
	//fill in the export data
	ToolsSoundExportFile xExportData;
	memset(&xExportData,0,sizeof(xExportData));
	xExportData.m_bFirstSampleTooBig = bFirstSampleTooBig;
	if(!Tools_FileSystem::GetModifiedTime(pszInputFilename,&xExportData.m_xSourceModifiedTime))
	{
		return;
	}
	Tools_FileSystem::GetModifiedTime(pszOutputFilename,&xExportData.m_xDestModifiedTime);




	char szFilename[MAX_PATH];
	unsigned long uBytesTransferred;

	strcpy(szFilename,pszOutputFilename);

	//replace the fileextension
	char* dotPos = strrchr(szFilename,'.');
	if(dotPos)*dotPos = 0;
	strcat(szFilename,".dat");

	Asura_Assertf(!Tools_FileSystem::IsReadOnly(szFilename),("%s is readonly. It will be made writeable",szFilename));
	Tools_FileSystem::MakeWriteable(szFilename);


	//open the file
	HANDLE handle = CreateFile(szFilename,GENERIC_WRITE,0,0,CREATE_ALWAYS,0,0);
	if(handle==INVALID_HANDLE_VALUE) return;

	//and save
	int iVersion = 0;
	unsigned short* const pusVersion = reinterpret_cast< unsigned short* >( &iVersion );

	// save the project and engine version numbers
	pusVersion[0] = TOOLS_SOUND_DAT_FILE_VERSION;
	pusVersion[1] = s_usProjectVersion;

	WriteFile(handle,&iVersion,sizeof(iVersion),&uBytesTransferred,0);
	WriteFile(handle,&xExportData,sizeof(xExportData),&uBytesTransferred,0);
	WriteFile(handle,pxConvertData,sizeof(*pxConvertData),&uBytesTransferred,0);

	CloseHandle(handle);
}


ToolsSoundConversionData::ToolsSoundConversionData()
{
	m_iMaxFrequency = 22050;
	m_bLoop = false;
	m_uFirstSample = 0;
	m_uMaxSamples = 0;
	m_bLoadImmediate = true;
	m_fMinLength = 0.0f;
	m_uAT3BitRate = 64;
}

bool Tools_Sound_Export::Resample_Sound(const Tools_SoundResampleData& xInputProperties,const Tools_SoundResampleData& xOutputProperties,const void* pInputBuffer,void* pOutputBuffer)
{
	//make sure the paramters are ones that we deal with
	if(xInputProperties.m_uNumberOfChannels != 1 || xOutputProperties.m_uNumberOfChannels != 1)
	{
		Asura_Assert(xInputProperties.m_uNumberOfChannels == 1 || xOutputProperties.m_uNumberOfChannels == 1,"Only supports mono sounds at the moment");
		return false;
	}

	if(!(xInputProperties.m_uBitsPerSample == 8 || xInputProperties.m_uBitsPerSample == 16))
	{
		Asura_Assert(xInputProperties.m_uBitsPerSample == 8 || xInputProperties.m_uBitsPerSample == 16,"Sound must be 8 or 16 bit");
		return false;
	}

	if(!(xOutputProperties.m_uBitsPerSample == 8 || xOutputProperties.m_uBitsPerSample == 16))
	{
		Asura_Assert(xOutputProperties.m_uBitsPerSample == 8 || xOutputProperties.m_uBitsPerSample == 16,"Sound must be 8 or 16 bit");
		return false;
	}

	if(!(pInputBuffer && pOutputBuffer))
	{
		Asura_Assert(pInputBuffer && pOutputBuffer,"");
		return false;
	}

	if(!xInputProperties.m_uNumberOfSamples)
	{
		Asura_Assert(xInputProperties.m_uNumberOfSamples,"");
		return false;
	}


	float fFrequencyScale = xOutputProperties.m_uNumberOfSamples/(float)xInputProperties.m_uNumberOfSamples;

	const unsigned char* pucInput8Bit = (unsigned char*) pInputBuffer;
	const short* psInput16Bit = (short*) pInputBuffer;

	unsigned char* pucOutput8Bit = (unsigned char*) pOutputBuffer;
	short* psOutput16Bit = (short*) pOutputBuffer;


	for(u_int uOutputIndex = 0;uOutputIndex < xOutputProperties.m_uNumberOfSamples;uOutputIndex++)
	{
		int iSample = 0;
		u_int uInputIndex = (u_int)(uOutputIndex/fFrequencyScale);
		if(uInputIndex<xInputProperties.m_uNumberOfSamples)
		{
			if(xInputProperties.m_uBitsPerSample == 8)
			{
				//convert 8 bit sample to 16 bit
				iSample = ((int)pucInput8Bit[uInputIndex])*256;
				iSample -= 32768;
			}
			else
			{
				iSample = psInput16Bit[uInputIndex];
			}
		}
		//make sure the value is in range
		if(iSample>0x7fff) iSample = 0x7fff;
		if(iSample<-0x7fff) iSample = -0x7fff;

		if(xOutputProperties.m_uBitsPerSample == 8)
		{
			//convert sample to 8 bit
			iSample += 32768;
			iSample /= 256;
			pucOutput8Bit[uOutputIndex] = (unsigned char) iSample;
		}
		else
		{
			psOutput16Bit[uOutputIndex] = (short) iSample;
		}

	}

	return true;
}

// -----------------------------------------------------------------------------------------

#ifndef DISABLE_R8BRAIN_DLL

//Unloads R8 brain dll if it gets loaded by anything
class Tools_R8Brain_Cleanup
{
public :
	Tools_R8Brain_Cleanup(){}
	~Tools_R8Brain_Cleanup()
	{
		if(Tools_Sound_Export::m_hDllInstance_R8B)
		{
			Tools_Sound_Export::Shutdown_R8Brain_Dll();
		}
	}
};
static Tools_R8Brain_Cleanup s_xR8Brain_Cleanup;


bool Tools_Sound_Export::Init_R8Brain_Dll( void)
{
	bool bRet = false;

	m_fnR8B_ExecuteMem	= NULL;
	m_fnR8B_Execute		= NULL;
	m_fnR8B_GetWavInfo	= NULL;

	m_hDllInstance_R8B = LoadLibrary("r8b.dll");
	Asura_Assert( m_hDllInstance_R8B != NULL, "Could not find the r8b.dll");

	if( m_hDllInstance_R8B != NULL)
	{
		m_fnR8B_Execute = (r8b_execute) GetProcAddress( m_hDllInstance_R8B, "r8b_execute");
		m_fnR8B_ExecuteMem = (r8b_executeMem) GetProcAddress( m_hDllInstance_R8B, "r8b_executeMem");
		m_fnR8B_GetWavInfo = (r8b_getWAVInfo) GetProcAddress( m_hDllInstance_R8B, "r8b_getWavInfo");

		if( (m_fnR8B_ExecuteMem != NULL) || (m_fnR8B_Execute != NULL) || (m_fnR8B_GetWavInfo != NULL))
		{
			bRet = true;
		}
		else
		{
			Asura_Assert( m_fnR8B_Execute != NULL, "Unable to find r8b_execute function in the r8b.dll");
			Asura_Assert( m_fnR8B_ExecuteMem != NULL, "Unable to find r8b_executeMem function in the r8b.dll");
			Asura_Assert( m_fnR8B_GetWavInfo != NULL, "Unable to find r8b_getWavInfo function in the r8b.dll");

			Shutdown_R8Brain_Dll();
		}
	}

	return bRet;
}

// -----------------------------------------------------------------------------------------

bool Tools_Sound_Export::Shutdown_R8Brain_Dll( void)
{
	bool bRet = false;

	Asura_Assert( m_hDllInstance_R8B == NULL, "Attempting to free the r8b.dll without it being loaded.");

	if( m_hDllInstance_R8B != NULL)
	{
		FreeLibrary( m_hDllInstance_R8B);

		m_hDllInstance_R8B	= NULL;
		m_fnR8B_Execute		= NULL;
		m_fnR8B_ExecuteMem	= NULL;
		m_fnR8B_GetWavInfo	= NULL;

		bRet = true;
	}

	return bRet;
}

// -----------------------------------------------------------------------------------------

void* Tools_Sound_Export::ResampleSound_R8Brain( R8BInputFormat* pxInputFormat, int iNewSampleRate, int iNewBitsDepth, int* piOutSize)
{
	unsigned char* pOutBuffer = NULL;
	int iAllocSize = 0;
	int iOutputSize = 0;
	int iCancel = 0;

	R8BError error;

	if(!m_hDllInstance_R8B) Init_R8Brain_Dll();

	Asura_Assert( m_hDllInstance_R8B != NULL, "Calling ResampleSound_R8Brain without the dll being loaded.");
	Asura_Assert( m_fnR8B_ExecuteMem != NULL, "Calling ResampleSound_R8Brain without the function pointer being set.");

	if( (m_hDllInstance_R8B != NULL) && (m_fnR8B_ExecuteMem != NULL))
	{
		//if( pxInputFormat->iChannels != 1)// || xOutputProperties.m_uNumberOfChannels != 1)
		//{
		//	Asura_Assert( pxInputFormat->iChannels == 1 ,"Only supports mono sounds at the moment");// || xOutputProperties.m_uNumberOfChannels == 1,"Only supports mono sounds at the moment");
		//	return false;
		//}

		//if(!(pxInputFormat->iBitDepth == 8 || pxInputFormat->iBitDepth == 16))
		//{
		//	Asura_Assert(pxInputFormat->iBitDepth == 8 || pxInputFormat->iBitDepth == 16,"Input Sound must be 8 or 16 bit");
		//	return false;
		//}

		if(!(iNewBitsDepth == 8 || iNewBitsDepth == 16))
		{
			Asura_Assert(iNewBitsDepth == 8 || iNewBitsDepth == 16,"Output Sound must be 8 or 16 bit");
			return false;
		}

		if(!pxInputFormat->iLength)
		{
			Asura_Assert(pxInputFormat->iLength,"Number of samples in wav is zero.");
			return false;
		}

		float fOutSize = ceilf((((float)pxInputFormat->iLength * (float)iNewSampleRate) / (float)pxInputFormat->iRate) + 0.5f);
		iAllocSize = (int)fOutSize;

//		unsigned char* pOutBuffer = new unsigned char[(int)iAllocSize * pxInputFormat->iBitDepth];
		pOutBuffer = new unsigned char[(int)iAllocSize * iNewBitsDepth * pxInputFormat->iChannels];
		if( pOutBuffer != NULL)
		{
			(m_fnR8B_ExecuteMem)( (R8BInputFormat*)pxInputFormat,
				(void*)pOutBuffer,
				iAllocSize,
				iNewSampleRate,
				iNewBitsDepth,
				4,	// always do best possible quality
				&iCancel,
				0,
				0,
				&error,
				&iOutputSize);
		}
	}

	if( piOutSize != NULL)
	{
		*piOutSize = iOutputSize;
	}

	return (void*)pOutBuffer;
}

#endif // DISABLE_R8BRAIN_DLL

// -----------------------------------------------------------------------------------------

