//
// Tools_CollisionMesh.cpp
//
// Created by Richard Rice - 18/03/2004
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Mesh/Asura_CollisionMesh.h"

#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_CollMeshGenData.h"

#include "Misc/list_tem.hpp"
#include "Misc/Tools_Adjacency.h"


//////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------**
** ToolsMergedPolygon **
**--------------------*/

class ToolsMergedPolygon
{
public :
	ToolsMergedPolygon(const Asura_Mesh* pxMesh,int iPolyIndex,bool bAddVerts);

	bool AddPolygon(const Asura_Mesh_Polygon* pxPoly);

	static void SplitIntoConvexPolygons(List<ToolsMergedPolygon*>& xPolygonList);
	static void SplitIntoQuads(List<ToolsMergedPolygon*>& xPolygonList);

	void FillInPolygon(Asura_CollisionMesh_Poly* pxPoly) const;
	int GetPolyIndex(){return m_iPolyIndex;}

private :

	void SplitIntoConvexPolygonsMain(List<ToolsMergedPolygon*>& xOutputList);
	void RemoveCollinearVertices();
	void SplitIntoQuadsMain(List<ToolsMergedPolygon*>& xOutputList);

	//checks to make sure 3 vertices form a convex corner
	bool IsConvex(int v1,int v2,int v3);
	//checks to make sure edge doesn't intersect any edges in the polygon
	bool IsEdgeOkay(int v1,int v2);
	bool EdgesIntersect(int v1,int v2,int v3,int v4) const;
	void RemoveDodgyPoints();

	ToolsMergedPolygon* ToolsMergedPolygon::SplitPolygon(int v1,int v2);


	const Asura_Mesh* m_pxMesh;
	int m_iPolyIndex;
	Asura_Vector_3 m_xNormal;

	List<int> m_xVerts;
};


const static float COLLISION_MESH_TOLERANCE = 0.001f;

ToolsMergedPolygon::ToolsMergedPolygon(const Asura_Mesh* pxMesh,int iPolyIndex,bool bAddVerts)
{
	Asura_Assert(pxMesh,"");
	Asura_Assert(iPolyIndex>=0 && iPolyIndex<pxMesh->NumberOfPolygons,"");

	m_pxMesh = pxMesh;
	m_iPolyIndex = iPolyIndex;
	m_xNormal = pxMesh->NormalsDataPtr[m_iPolyIndex].FaceNormal;

	if(bAddVerts)
	{
		const Asura_Mesh_Polygon* pxPoly = &m_pxMesh->PolygonDataPtr[m_iPolyIndex];

		int iNumVerts = pxPoly->NumberOfVertices();
		for(int i=0;i<iNumVerts;i++)
		{
			m_xVerts.add_entry(pxPoly->VertexID[i]);
		}
	}
}

bool ToolsMergedPolygon::AddPolygon(const Asura_Mesh_Polygon* pxPoly)
{
	Asura_Assert(pxPoly,"");
	int iNumVerts = pxPoly->NumberOfVertices();
	
	//If this polygon can be merged in then there should be 2 or more consecutive vertices
	//in the merged polygon that match vertices in the passed polygon.

	//Search for the first vertex that matches , first finding a vertex that doesn't match , in case
	//the series of matched vertices straddles the start/end of list
	

	LIL<int> xVertLIL(&m_xVerts);
	int iCount = 0;
	bool bFoundUnmatched = false;
	int iMatchIndex;
	for(;;xVertLIL.next(),iCount++)
	{
		if(iCount>=m_xVerts.size())
		{
			return false;//failed to find an appropriate starting vertex
		}
		
		bool bSharedVert = false;
		for(int i=0;i<iNumVerts;i++)
		{
			if(pxPoly->VertexID[i] == xVertLIL())
			{
				bSharedVert = true;
				iMatchIndex = i;
				break;
			}
		}
		if(bFoundUnmatched)
		{
			if(bSharedVert)
			{
				break;//this is the start of the matched vertices
			}
		}
		else
		{
			if(!bSharedVert)
			{
				bFoundUnmatched = true;//found our unmatched vert , can now search for first matched
			}
		}
	}

	LIL<int> xMatchLIL(xVertLIL);

	//Now continue and find how many vertices in a row match.
	//We need to head forwards in the merged polygon , and backwards through the passed polygon
	int iMatchCount = 0;
	for(int i=0;i<iNumVerts;i++,xVertLIL.next())
	{
		int iIndex = (iNumVerts+iMatchIndex-i)%iNumVerts;

		if(xVertLIL() == pxPoly->VertexID[iIndex])
		{
			iMatchCount++;
		}
		else
		{
			break;
		}
	}

	if(iMatchCount<2) return false; //need to have at least 2 matching verts

	//Now need to check to make sure any of the remaining verts don't match in another
	//run elswhere , to avoid getting 'holes' in the merged polygon
	for(int i=iMatchCount;i<iNumVerts;i++)
	{
		int iIndex = (iNumVerts+iMatchIndex-i)%iNumVerts;
		if(m_xVerts.contains(pxPoly->VertexID[iIndex]))
		{
			return false;
		}
	}

	//Merge appears to be okay , so time to add in the vertices.
	//If there were more than 2 matched verts , then we need to remove
	//any between the first and last ones

	for(int i=1;i<iMatchCount-1;i++)
	{
		int iIndex = (iNumVerts+iMatchIndex-i)%iNumVerts;
		m_xVerts.delete_entry(pxPoly->VertexID[iIndex]);
	}

	//now need to add in any of the unmatched vertices in order
	for(int i=0;i<iNumVerts-iMatchCount;i++)
	{
		int iPrevIndex = (iMatchIndex+i)%iNumVerts;
		int iIndex = (iPrevIndex+1)%iNumVerts;
		m_xVerts.add_entry_after(pxPoly->VertexID[iIndex],pxPoly->VertexID[iPrevIndex]);
	}

	return true;
}

void ToolsMergedPolygon::RemoveDodgyPoints()
{
	//check each set of 3 vertices in turn to see if the middle one can be removed
	LIL<int> xVertLIL(&m_xVerts);

	int iFirst = xVertLIL();
	xVertLIL.next();

	int iMiddle = xVertLIL();
	xVertLIL.next();

	//remove points that cause the boundary of the polygon to double back on itself...
	//Find sets of collinear poits where the middle one isn't in between the first and last.
	for(int iLoop = m_xVerts.size();iLoop && m_xVerts.size()>3;iLoop--)
	{
		bool bRemove = false;
		int iLast = xVertLIL();

		//get the 2 sides
		Asura_Vector_3 xSide1 = m_pxMesh->VertexDataPtr[iMiddle].Position-m_pxMesh->VertexDataPtr[iFirst].Position;
		Asura_Vector_3 xSide2 = m_pxMesh->VertexDataPtr[iLast].Position-m_pxMesh->VertexDataPtr[iMiddle].Position;


		//get a normal for the first side lying on the polygon
		Asura_Vector_3 xSideNormal = xSide1^m_xNormal;
		xSideNormal.Normalise();

		//get the distance that side2 moves in the direction of that normal
		float fDist = xSide2*xSideNormal;

		//points are collinear if the distance is small enough
		if(Asura_Maths::Abs(fDist)<COLLISION_MESH_TOLERANCE && xSide1*xSide2<0)
		{
			bRemove = true;
		}
		else if(xSide1.Magnitude()<COLLISION_MESH_TOLERANCE)//also remove point if its too close to the previous one
		{
			bRemove = true;
		}

		if(bRemove)
		{
			//can delete the middle point
			m_xVerts.delete_entry(iMiddle);
			iMiddle = xVertLIL();
			xVertLIL.next();
		}
		else
		{
			//advance to the next then
			iFirst = iMiddle;
			iMiddle = xVertLIL();
			xVertLIL.next();
		}
	}
}

void ToolsMergedPolygon::SplitIntoConvexPolygons(List<ToolsMergedPolygon*>& xPolygonList)
{
	for(LIF<ToolsMergedPolygon*> xPolyLIF(&xPolygonList);!xPolyLIF.done();xPolyLIF.next())
	{
		xPolyLIF()->SplitIntoConvexPolygonsMain(xPolygonList);
	}
}

void ToolsMergedPolygon::SplitIntoConvexPolygonsMain(List<ToolsMergedPolygon*>& xPolygonList)
{
	Asura_Assert(m_xVerts.size()>2,"");

	RemoveDodgyPoints();
	RemoveCollinearVertices();

	if(m_xVerts.size()==3)
	{
		//can't split up a triangle anymore
		return;
	}

	while(true)
	{
		//search for a concave corner
		LIL<int> xVertLIL(&m_xVerts);
		bool bConvex = true;
		for(int i=0;i<m_xVerts.size();i++,xVertLIL.next())
		{
			if(!IsConvex(xVertLIL.get_previous(),xVertLIL(),xVertLIL.get_next()))
			{
				bConvex = false;
				break;
			}
		}

		if(bConvex)
		{
			//already convex , no more needs doing
			return;
		}

		int iBefore = xVertLIL.get_previous();
		int iVert = xVertLIL();
		int iAfter = xVertLIL.get_next();

		//search for a place that this can be split that will make this corner convex , and
		//give the shortest edge

		int iBestVert = -1;
		float fBestDist = Asura_Maths::LargeFloat;
		int iBestConvexScore = 0;


		LIL<int> xVertLIL2(&m_xVerts);		
		for(int i=0;i<m_xVerts.size();xVertLIL2.next(),i++)
		{
			if(xVertLIL2() == iBefore || xVertLIL2() == iVert || xVertLIL2() == iAfter) continue;

			//make sure the new edge won't cross any existing edges
			if(!IsEdgeOkay(iVert,xVertLIL2())) continue;

			//See which corners at this vertex will be made convex
			int iConvexScore = 0;
			if(IsConvex(iBefore,iVert,xVertLIL2())) iConvexScore += 4;
			if(IsConvex(xVertLIL2(),iVert,iAfter)) iConvexScore += 4;

			//and check at the other end of the edde (lower priority though)
			if(IsConvex(iVert,xVertLIL2(),xVertLIL2.get_next())) iConvexScore++;
			if(IsConvex(xVertLIL2.get_previous(),xVertLIL2(),iVert)) iConvexScore++;
			
			if(iConvexScore < iBestConvexScore) continue;
			

			//get the distance
			Asura_Vector_3 xOffset = m_pxMesh->VertexDataPtr[iVert].Position - m_pxMesh->VertexDataPtr[xVertLIL2()].Position;
			float fDist = xOffset.Magnitude();

			if(fDist<fBestDist)
			{
				iBestVert = xVertLIL2();
				fBestDist = fDist;
				iBestConvexScore = iConvexScore;
			}
		}
		if(iBestConvexScore == 0)
		{
			//no improvement at all?
			return;
		}
	
		//Need to split off the vertices between the 2 chosen verts into a new polygon
		ToolsMergedPolygon* pxNewPoly = SplitPolygon(xVertLIL(),iBestVert);

		//add to the polygon list after this one
		xPolygonList.add_entry_after(pxNewPoly,this);

	}
}

bool ToolsMergedPolygon::IsConvex(int v1,int v2,int v3)
{
	//get the 2 sides
	Asura_Vector_3 xSide1 = m_pxMesh->VertexDataPtr[v2].Position-m_pxMesh->VertexDataPtr[v1].Position;
	Asura_Vector_3 xSide2 = m_pxMesh->VertexDataPtr[v3].Position-m_pxMesh->VertexDataPtr[v2].Position;

	//get a normal for the first side lying on the polygon , pointing into the polygon
	Asura_Vector_3 xSideNormal = m_xNormal^xSide1;
	xSideNormal.Normalise();

	//get the distance that side2 moves in the direction of that normal
	float fDist = xSide2*xSideNormal;

	//this should be at least 0 for polygon to be convex , however allow very slightly negative
	//to have some tolerance for edges that aren't quite straight
	return fDist >= -COLLISION_MESH_TOLERANCE;
	
}

bool ToolsMergedPolygon::IsEdgeOkay(int v1,int v2)
{
	//need to check this edge against each other edge to see if it intersects
	CLIF<int> xVertLIF(&m_xVerts);

	int iPrev = xVertLIF();
	xVertLIF.next();

	for(;!xVertLIF.done();xVertLIF.next())
	{
		if(EdgesIntersect(v1,v2,iPrev,xVertLIF()))
		{
			return false;
		}
		iPrev = xVertLIF();
	}

	if(EdgesIntersect(v1,v2,iPrev,m_xVerts.first_entry()))
	{
		return false;
	}

	//Okay , this edge doesn't intersect any in the original polygon.
	//Next we need to make sure it doesn't head out of the polygon immediately.

	//Find the verts before and after v1
	int iBefore = -1;
	int iAfter = -1;
	LIL<int> xVertLIL(&m_xVerts);

	for(int i=0;i<m_xVerts.size();i++,xVertLIL.next())
	{
		if(xVertLIL() == v1)
		{
			iBefore = xVertLIL.get_previous();
			iAfter = xVertLIL.get_next();
		}
	}
	Asura_Assert(iBefore != -1,"");

	//If v2 is the vert immediately before or after v1 , then its bound to be okay
	if(v2==iBefore || v1==iAfter) return true;

	//If the corner iBefore-v1-iAfter is convex then the edge should be inside
	//if both iBefore-v1-v2 and v1-iAfter-v2 are convex
	//If that corner is concave , then we just need one of the other corners to be convex.

	bool bConvex1 = IsConvex(iBefore,v1,iAfter);
	bool bConvex2 = IsConvex(iBefore,v1,v2);
	bool bConvex3 = IsConvex(v1,iAfter,v2);

	if(bConvex1)
	{
		return bConvex2 && bConvex3;
	}

	return bConvex2 || bConvex3;
}

bool ToolsMergedPolygon::EdgesIntersect(int v1,int v2,int v3,int v4) const
{
	if(v1==v3 || v1==v4) return false;
	if(v2==v3 || v2==v4) return false;
	
	Asura_Vector_3 xSide1 = m_pxMesh->VertexDataPtr[v2].Position-m_pxMesh->VertexDataPtr[v1].Position;

	//get 2 normals on polygon
	Asura_Vector_3 xNormal1 = xSide1;
	xNormal1.Normalise();
	Asura_Vector_3 xNormal2 = m_xNormal ^ xNormal1;
	xNormal2.Normalise();

	//get coordinates of points using these basis vectors
	Asura_Vector_2 p1,p2,p3,p4;
	
	p1.x = m_pxMesh->VertexDataPtr[v1].Position * xNormal1;
	p1.y = m_pxMesh->VertexDataPtr[v1].Position * xNormal2;
	p2.x = m_pxMesh->VertexDataPtr[v2].Position * xNormal1;
	p2.y = m_pxMesh->VertexDataPtr[v2].Position * xNormal2;
	p3.x = m_pxMesh->VertexDataPtr[v3].Position * xNormal1;
	p3.y = m_pxMesh->VertexDataPtr[v3].Position * xNormal2;
	p4.x = m_pxMesh->VertexDataPtr[v4].Position * xNormal1;
	p4.y = m_pxMesh->VertexDataPtr[v4].Position * xNormal2;

	//the line p1 to p2 should be parallel to the x axis , so find out where
	//p3 to p4 crosses the y value

	float fTolerance = COLLISION_MESH_TOLERANCE;

	if(Asura_Maths::Abs(p3.y-p4.y)<=fTolerance) return false;

	if(Asura_Maths::Abs(p3.y-p1.y)<=fTolerance)
	{
		if(p1.x <= p3.x+fTolerance && p3.x <= p2.x+fTolerance) return true;
		if(p2.x <= p3.x+fTolerance && p3.x <= p1.x+fTolerance) return true;
	}

	if(Asura_Maths::Abs(p4.y-p1.y)<=fTolerance)
	{
		if(p1.x <= p4.x+fTolerance && p4.x <= p2.x+fTolerance) return true;
		if(p2.x <= p4.x+fTolerance && p4.x <= p1.x+fTolerance) return true;
	}



	if(p3.y<p1.y-fTolerance && p4.y<p1.y-fTolerance) return false;
	if(p3.y>p1.y+fTolerance && p4.y>p1.y+fTolerance) return false;

	float fLerp = (p1.y-p3.y)/(p4.y-p3.y);

	float fXIntersect = p3.x + fLerp*(p4.x-p3.x);

	if(p1.x<fXIntersect-fTolerance && p2.x<fXIntersect-fTolerance) return false;
	if(p1.x>fXIntersect+fTolerance && p2.x>fXIntersect+fTolerance) return false;

	return true;

}

void ToolsMergedPolygon::RemoveCollinearVertices()
{
	//check each set of 3 vertices in turn to see if the middle one can be removed
	LIL<int> xVertLIL(&m_xVerts);

	int iFirst = xVertLIL();
	xVertLIL.next();

	int iMiddle = xVertLIL();
	xVertLIL.next();

	for(int iLoop = m_xVerts.size();iLoop && m_xVerts.size()>3;iLoop--)
	{
		int iLast = xVertLIL();

		//get the 2 sides
		Asura_Vector_3 xSide1 = m_pxMesh->VertexDataPtr[iLast].Position-m_pxMesh->VertexDataPtr[iFirst].Position;
		Asura_Vector_3 xSide2 = m_pxMesh->VertexDataPtr[iMiddle].Position-m_pxMesh->VertexDataPtr[iFirst].Position;

		//get a normal for the first side lying on the polygon
		Asura_Vector_3 xSideNormal = xSide1^m_xNormal;
		xSideNormal.Normalise();

		//get the distance that side2 moves in the direction of that normal
		float fDist = xSide2*xSideNormal;

		//points are collinear if the distance is small enough
		if(Asura_Maths::Abs(fDist)<COLLISION_MESH_TOLERANCE)
		{
			//can delete the middle point
			m_xVerts.delete_entry(iMiddle);
			iMiddle = xVertLIL();
			xVertLIL.next();
		}
		else
		{
			//advance to the next then
			iFirst = iMiddle;
			iMiddle = xVertLIL();
			xVertLIL.next();
		}
	}
}


void ToolsMergedPolygon::SplitIntoQuads(List<ToolsMergedPolygon*>& xPolygonList)
{
	for(LIF<ToolsMergedPolygon*> xPolyLIF(&xPolygonList);!xPolyLIF.done();xPolyLIF.next())
	{
		xPolyLIF()->SplitIntoQuadsMain(xPolygonList);
	}
}

void ToolsMergedPolygon::SplitIntoQuadsMain(List<ToolsMergedPolygon*>& xPolygonList)
{
	
 	RemoveCollinearVertices();

	#if 1
	LIL<int> xVertLIL(&m_xVerts);

	while(m_xVerts.size()>4)
	{
		//move the iterator back twice , so we lose the first and last vertex from this polygon
		xVertLIL.previous();
		xVertLIL.previous();
		
		
		ToolsMergedPolygon* pxQuad = new ToolsMergedPolygon(m_pxMesh,m_iPolyIndex,false);
		
		//transfer 4 vertices to a new quad
		pxQuad->m_xVerts.add_entry(xVertLIL());
		xVertLIL.next();
		//second and third vertex needs to be removed from this polygon
		pxQuad->m_xVerts.add_entry(xVertLIL());
		xVertLIL.delete_current();
		pxQuad->m_xVerts.add_entry(xVertLIL());
		xVertLIL.delete_current();
		pxQuad->m_xVerts.add_entry(xVertLIL());

		//add to the polygon list after this one
		xPolygonList.add_entry_after(pxQuad,this);

	}
	#endif

}

ToolsMergedPolygon* ToolsMergedPolygon::SplitPolygon(int v1,int v2)
{
	Asura_Assert(m_xVerts.contains(v1),"");
	Asura_Assert(m_xVerts.contains(v2),"");

	ToolsMergedPolygon* pxNewPoly = new ToolsMergedPolygon(m_pxMesh,m_iPolyIndex,false);

	//search for the first shared vertex
	LIL<int> xVertLIL(&m_xVerts);

	for(;;xVertLIL.next())
	{
		if(xVertLIL() == v1) break;
	}

	//add the first shared point
	pxNewPoly->m_xVerts.add_entry(xVertLIL());
	xVertLIL.next();

	//transfer the intermediate points
	for(;xVertLIL()!=v2;)
	{
		pxNewPoly->m_xVerts.add_entry(xVertLIL());
		xVertLIL.delete_current();
	}

	//add the end shared point
	pxNewPoly->m_xVerts.add_entry(xVertLIL());

	Asura_Assert(pxNewPoly->m_xVerts.size() > 2,"");

	return pxNewPoly;
	
	
}

void ToolsMergedPolygon::FillInPolygon(Asura_CollisionMesh_Poly* pxPoly) const
{
	Asura_Assert(m_xVerts.size() == 3 || m_xVerts.size() == 4,"");

	CLIF<int> xVertLIF(&m_xVerts);

	for(int i=0;!xVertLIF.done();xVertLIF.next(),i++)
	{
		pxPoly->m_usVertexID[i] = (u_short) xVertLIF();
	}
	if(m_xVerts.size()==3)
	{
		pxPoly->m_usVertexID[3] = 0xffff;
	}
	
}

//////////////////////////////////////////////////////////////////////////////////////////////////
#if 0
void Asura_CollisionMesh::UpdateFromMesh(Asura_Mesh* pxAsuraMesh)
{
	Asura_ResourcePushType(ASURA_RESOURCE_COLLISION_MESH);
	{
		//Delete any old collision mesh data. RAR 28/03/03
		Asura_ResourceDeallocateArray(m_pxVertices);
		Asura_ResourceDeallocateArray(m_pxPolys);
		Asura_ResourceDeallocateArray(m_pusPolygonFlags);
		Asura_ResourceDeallocateArray(m_psMaterialIDs);

		m_uNumberOfVertices = 0;
		m_pxVertices = 0;
		m_uNumberOfPolys = 0;
		m_pxPolys = 0;
		m_pusPolygonFlags = 0;
		m_psMaterialIDs = 0;

		const Asura_Mesh_CollMeshGenData* pxCollGenData = pxAsuraMesh->m_pxCollMeshGenData;

		// Set up normals if needed
		if (!pxAsuraMesh->AreNormalsValid())
		{
			pxAsuraMesh->CreateNormals();
		}

		// Set up BB if needed.
		if (!pxAsuraMesh->BoundingBoxDataValid)
		{
			pxAsuraMesh->CalculateBoundingBox();
		}

		Asura_Assert(pxAsuraMesh->VertexDataValid, "Asura mesh not set up yet");
		Asura_Assert(pxAsuraMesh->AreNormalsValid(), "Asura mesh not set up yet");


		// Bounding volumes
		m_xBoundingBox = pxAsuraMesh->BoundingBox;
		m_fBoundingRadius = pxAsuraMesh->BoundingRadius;
		ResetBB();

		//initialise array of polygons that have been dealt with
		bool* pbPolygonDone = new bool[pxAsuraMesh->NumberOfPolygons];
		for(int i=0;i<pxAsuraMesh->NumberOfPolygons;i++)
		{
			pbPolygonDone[i] = false;
		}


		//generate polygon adjacencies
		Tools_MeshAdjacency xMeshAdj;
		xMeshAdj.Initialise(pxAsuraMesh);

		List<ToolsMergedPolygon*> xPolygonList;

		//Check each polygon , and mark any that won't be needed as dealt with already
		if(pxAsuraMesh->PolygonFlagDataPtr)
		{
			for(int i=0;i<pxAsuraMesh->NumberOfPolygons;i++)
			{
				if((pxAsuraMesh->PolygonFlagDataPtr[i].Flags & AsuraPolygonFlag_IgnoreLOS) && 
					(pxAsuraMesh->PolygonFlagDataPtr[i].Flags & AsuraPolygonFlag_NoCollisions))
				{
					pbPolygonDone[i] = true;
				}
				if(!s_bIncludeDontExportPolys && (pxAsuraMesh->PolygonFlagDataPtr[i].Flags & AsuraPolygonFlag_DontExport))
				{
					pbPolygonDone[i] = true;
				}
			}
		}

		if(pxCollGenData && pxCollGenData->m_pxPolygons)
		{
			for(int i=0;i<pxAsuraMesh->NumberOfPolygons;i++)
			{
				if(pxCollGenData->m_pxPolygons[i].m_bReplaceByBox)
				{
					pbPolygonDone[i] = true;
				}
			}
		}



		for(int i=0;i<pxAsuraMesh->NumberOfPolygons;i++)
		{
			if(pbPolygonDone[i]) continue;

			List<int> xPolyQueue;


			ToolsMergedPolygon* pxMerge = new ToolsMergedPolygon(pxAsuraMesh,i,true);
			xPolygonList.add_entry(pxMerge);

			//and add the polygon to the queue
			xPolyQueue.add_entry(i);
			pbPolygonDone[i] = true;

			Asura_Vector_3 xNormal = pxAsuraMesh->NormalsDataPtr[i].FaceNormal;
			int iFlags = 0;
			if(pxAsuraMesh->PolygonFlagDataPtr) iFlags = pxAsuraMesh->PolygonFlagDataPtr[i].Flags;

			//move out from this polygon and determine all connected coplanar polygons.
			for(LIF<int> xPolyLIF(&xPolyQueue);!xPolyLIF.done();xPolyLIF.next())
			{
				int iPolyIndex = xPolyLIF();
				//check each adjaccent polygon to see if it can be added
				const Tools_Polygon_Adjacency* pxPolyAdj = xMeshAdj.GetPolygonAdjacencies(iPolyIndex);

				for(const Tools_Adjacent_Polygon* pxAdj = pxPolyAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
				{
					if(pxAdj->GetNumSharedVertices() != 2) continue;//need a shared edge

					int iPolyIndex2 = pxAdj->GetPolygonIndex();

					if(pbPolygonDone[iPolyIndex2]) continue;//skip any polygons that have already been dealt with

					if(pxAsuraMesh->PolygonFlagDataPtr)
					{
						//make sure flags are the same
						if(iFlags != pxAsuraMesh->PolygonFlagDataPtr[iPolyIndex2].Flags)
						{
							continue;
						}
					}

					//is this polygon coplanar?
					float fDot = pxAsuraMesh->NormalsDataPtr[iPolyIndex2].FaceNormal * xNormal;
					if(fDot < 0.9999f) continue;

					//add to the merged polygon if possible
					if(pxMerge->AddPolygon(&pxAsuraMesh->PolygonDataPtr[iPolyIndex2]))
					{
						xPolyQueue.add_entry(iPolyIndex2);
						pbPolygonDone[iPolyIndex2] = true;
					}
				}
			}
		}

		//Make all the polygons convex
		ToolsMergedPolygon::SplitIntoConvexPolygons(xPolygonList);
		//and split up into tris/quads
		ToolsMergedPolygon::SplitIntoQuads(xPolygonList);

		delete [] pbPolygonDone;

		//Allocate space for polygons
		u_int uNumUsedPolys = xPolygonList.size();
		m_uNumberOfPolys = xPolygonList.size();
		if(pxCollGenData)
		{
			m_uNumberOfPolys += pxCollGenData->m_uNumberOfBoxes*6;//add space for boxes
		}
		Asura_ResourceAllocateArray(m_pxPolys, m_uNumberOfPolys);

		//Initialise vertex conversion array
		m_uNumberOfVertices = 0;
		int* piVertConv = new int[pxAsuraMesh->NumberOfVertices];
		for(int i = 0;i<pxAsuraMesh->NumberOfVertices;i++)
		{
			piVertConv[i] = -1;
		}

		//and reverse polygon conversion
		int* piAntiPolyConv = new int[m_uNumberOfPolys];

		//Fill in polygons , and determine which vertices are used
		LIF<ToolsMergedPolygon*> xPolygonLIF(&xPolygonList);
		Asura_CollisionMesh_Poly* pxCollPoly = m_pxPolys;

		for(int i=0;!xPolygonLIF.done();xPolygonLIF.next(),pxCollPoly++,i++)
		{
			xPolygonLIF()->FillInPolygon(pxCollPoly);
			piAntiPolyConv[i] = xPolygonLIF()->GetPolyIndex();

			//convert the vertex indeces
			for(int j=0;j<4;j++)
			{
				if(pxCollPoly->m_usVertexID[j] == 0xffff) continue;

				int iVert = pxCollPoly->m_usVertexID[j];
				if(piVertConv[iVert] == -1)
				{
					piVertConv[iVert] = m_uNumberOfVertices++;
				}

				pxCollPoly->m_usVertexID[j] = (u_short)piVertConv[iVert];
			}
		}

		//leave vertex space for bounding boxes
		u_int uNumUsedVertices = m_uNumberOfVertices;
		if(pxCollGenData)
		{
			m_uNumberOfVertices += pxCollGenData->m_uNumberOfBoxes*8;
		}

		//Fill in vertex positions
		Asura_ResourceAllocateArray(m_pxVertices, m_uNumberOfVertices);
		for(int i=0;i<pxAsuraMesh->NumberOfVertices;i++)
		{
			if(piVertConv[i] != -1)
			{
				m_pxVertices[piVertConv[i]] = pxAsuraMesh->VertexDataPtr[i].Position;
			}
		}
		delete [] piVertConv;

		if(pxCollGenData && pxCollGenData->m_pxBoxes && pxCollGenData->m_pxPolygons)
		{

			//generate bounding box polygons
			for(u_int i=0;i<pxCollGenData->m_uNumberOfBoxes;i++)
			{
				const Asura_Mesh_CollBoxInfo* pxBox = &pxCollGenData->m_pxBoxes[i];
				Asura_Bounding_Box xBox;
				xBox.Reset();
				int iPolyIndex = -1;//index of one polygon that is part of box , for taking flags etc

				//search for all polygons that are part of the box , and extend the bounding box to include them
				for(int j=0;j<pxAsuraMesh->NumberOfPolygons;j++)
				{
					if(pxCollGenData->m_pxPolygons[j].m_usGroupID != (u_short) pxBox->m_uGroupID)
					{
						continue;
					}

					int iPolyFlags = pxAsuraMesh->PolygonFlagDataPtr ? pxAsuraMesh->PolygonFlagDataPtr[j].Flags : 0;

					if((iPolyFlags & AsuraPolygonFlag_IgnoreLOS) && 
						(iPolyFlags & AsuraPolygonFlag_NoCollisions))
					{
						continue;
					}
					if(!s_bIncludeDontExportPolys && (iPolyFlags & AsuraPolygonFlag_DontExport))
					{
						continue;
					}


					//first polygon found?
					if(iPolyIndex == -1) iPolyIndex = j;

					//grab the verts , and add them to the box
					const Asura_Mesh_Polygon* pxPoly = &pxAsuraMesh->PolygonDataPtr[j];

					for(int k=0;k<pxPoly->NumberOfVertices();k++)
					{
						Asura_Vector_3 xVec = pxAsuraMesh->VertexDataPtr[pxPoly->VertexID[k]].Position;
						xVec = pxBox->m_xOrient.ApplyTranspose(xVec);//vector needs to be rotated so as to be using the axes of the box

						xBox.Set(xVec);
					}
				}

				if(iPolyIndex == -1)
				{
					//this was a dud , so skip it
					continue;
				}

				//generate the vertices
				Asura_Vector_3* pxVerts = m_pxVertices + uNumUsedVertices;
				for(int j=0;j<8;j++)
				{
					pxVerts[j].x = (j&1) ? xBox.MaxX : xBox.MinX;
					pxVerts[j].y = (j&2) ? xBox.MaxY : xBox.MinY;
					pxVerts[j].z = (j&4) ? xBox.MaxZ : xBox.MinZ;

					//rotate back
					pxVerts[j] = pxVerts[j] * pxBox->m_xOrient;
				}

				//fill in the polygon vertex indeces
				Asura_CollisionMesh_Poly* pxPolys = m_pxPolys + uNumUsedPolys;

				const u_short ausVertInds[6][4] =
				{
					{0,2,3,1},
					{4,5,7,6},
					{1,3,7,5},
					{0,4,6,2},
					{0,1,5,4},
					{2,6,7,3},
				};

				for(int j=0;j<6;j++)
				{
					piAntiPolyConv[j+uNumUsedPolys] = iPolyIndex;
					for(int k=0;k<4;k++)
					{
						pxPolys[j].m_usVertexID[k] = (u_short)(ausVertInds[j][k] + uNumUsedVertices);
					}
				}


				//update the used polygon and vertex counts
				uNumUsedPolys += 6;
				uNumUsedVertices += 8;

			}

		}

		//correct number of polygons and vertices
		m_uNumberOfPolys = uNumUsedPolys;
		m_uNumberOfVertices = uNumUsedVertices;




		const bool bHasPolygonData = pxAsuraMesh->PolygonFlagDataValid;
		const bool bHasTextureData = pxAsuraMesh->TextureDataValid;

		// TODO set up m_iOverallFlags from somewhere!
		// Needs to be setable in tools, so that overall flags can be set.

		if(m_uNumberOfPolys != 0)
		{
			// MC - 03/12/03: About time something was implemented for this!
			// For now, predict the overall flags based on the first poly, and check 
			// all subsequent polys agree with it in the loop, later.

			if (bHasPolygonData)
			{
				m_usOverallFlags = (unsigned short)pxAsuraMesh->PolygonFlagDataPtr[0].Flags;
			}
			else
			{
				// (AW - 11.02.2004) - If there is no polygon flag data whatsoever then it the
				// overall flags need to be reset accordingly.
				m_usOverallFlags = 0;
			}

			// AW - 24/03/03 - Same situation for material IDs.
			// MC - These seem to be handled by the call to SetAllMaterialIDs in LoadShape?

		}



		if (bHasPolygonData)
		{
			Asura_ResourceAllocateArray(m_pusPolygonFlags, m_uNumberOfPolys);
		}

		if (bHasTextureData)
		{
			Asura_ResourceAllocateArray(m_psMaterialIDs, m_uNumberOfPolys);
		}



		//go through polygons again , and check polygon flags
		bool bCanDiscardFlags = true;
		bool bCanDiscardMaterialIDs = true;
		xPolygonLIF.restart();

		for(u_int i=0;i<m_uNumberOfPolys;i++)
		{
			u_int uMeshPoly = (u_int) piAntiPolyConv[i];

			if (m_pusPolygonFlags)
			{
				m_pusPolygonFlags[i] = (unsigned short)pxAsuraMesh->PolygonFlagDataPtr[uMeshPoly].Flags;

				if (m_pusPolygonFlags[i] != m_usOverallFlags)
				{
					bCanDiscardFlags = false;
				}
				m_usOverallFlags &= m_pusPolygonFlags[i];
			}
			if (m_psMaterialIDs)
			{
				m_psMaterialIDs[i] = (short)pxAsuraMesh->TextureDataPtr[uMeshPoly].m_iMaterialID;
				if (m_psMaterialIDs[i] != m_sOverallMaterial)
				{
					bCanDiscardMaterialIDs = false;
				}
			}
		}
		delete [] piAntiPolyConv;

		//clear up temporary polygon data
		while(xPolygonList.size())
		{
			delete xPolygonList.first_entry();
			xPolygonList.delete_first_entry();
		}

		// Discard the poly flags if they aren't needed.
		if (m_pusPolygonFlags && 
			bCanDiscardFlags)
		{
			Asura_ResourceDeallocateArray(m_pusPolygonFlags);
			m_pusPolygonFlags = NULL;
		}
		// Discard the material IDs if they aren't needed.
		if (m_psMaterialIDs && 
			bCanDiscardMaterialIDs)
		{
			Asura_ResourceDeallocateArray(m_psMaterialIDs);
			m_psMaterialIDs = NULL;
		}
	}
	Asura_ResourcePopType();
}
#endif


