
///////////////////////////////////////////////////////////////////////////////
//
// File   : Tools_MeshRestartIndex.cpp
// Author : Andy Tate
// Date   : 20-Sep-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "MeshProcessing/Tools_MeshRestartIndex.h"

// Asura independent header-files...
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Asura platform header-files...

// Tools header-files...
#include "Load&Save/LevelExport/Tools_Level_Export.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#if 0
#define DEBUG_PRINTF(s)	Asura_OutputToDebugger(s)
#else
#define DEBUG_PRINTF(s)
#endif


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_MeshRestartIndex::IsValidForCurrentPlatform
// Author       : Andy Tate
// Date         : 20-Sep-2005
//
// Description  : Checks whether the current export platform is able to handle
//				  restart indices in it's index-buffers.
//
// Parameters   : A reference to the restart value to be used.
//
// Return value : "true" if valid, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Tools_MeshRestartIndex::IsValidForCurrentPlatform (u_short& usRestartValue)
{
	// Which platform are we exporting for?
	switch (Tools_Level_Export::s_eExportType)
	{
		case Tools_Level_Export::Export_PC:
		case Tools_Level_Export::Export_PS2:
		case Tools_Level_Export::Export_PSP:
		case Tools_Level_Export::Export_Wii:
		{
			// Not yet supported...
			return	(false);
		}

		case Tools_Level_Export::Export_PS3:
		case Tools_Level_Export::Export_Xbox360:
		{
			// Here we go...
			usRestartValue	= 0xFFFF;

			return	(true);
		}

		default:
		{
			// Shouldn't get here...
			Asura_Assert	(false, "Tools_MeshRestartIndex::IsValidForCurrentPlatform(): Unexpected platform type.");
			return			(false);
		}
	}		// End of switch.
}

/// End of function 'Tools_MeshRestartIndex::IsValidForCurrentPlatform'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_MeshRestartIndex::ReplaceDegeneratesInTriStrip
// Author       : Andy Tate
// Date         : 20-Sep-2005
//
// Description  : 
//
// Parameters   : 
//
// Return value : The number of indices following the replacement of degenerates.
////////////////////////////////////////////////////////////////////////////////

u_int Tools_MeshRestartIndex::ReplaceDegeneratesInTriStrip (const u_int uNumIndices,
															const u_short* pusInputIndices,
															u_short* pusOutputIndices)
{
	u_short	usRestartIndex;

	// Do we need to bother replacing?
	if (IsValidForCurrentPlatform(usRestartIndex))
	{
		u_int	uCurrentInputIndex	= 0;
		u_int	uCurrentOutputIndex	= 0;

		// Yes - do we have enough indices?
		DEBUG_PRINTF	(("Tools_MeshRestartIndex::ReplaceDegeneratesInTriStrip(): uNumIndices = %d...\n", uNumIndices));

		if (6 < uNumIndices)
		{
			u_short*	pusIndices			= new u_short [uNumIndices];
			u_int*		puIndexCount		= new u_int [uNumIndices];

			u_int		uOriginalInputIndex	= 0;
			int			iNumDistinctIndices	= 0;

			// Reduce to individual indices with their counts...
			while (uNumIndices > uCurrentInputIndex)
			{
				const u_short	usIndex	= pusInputIndices [uCurrentInputIndex++];

				// Is this index the same as the last one?
				DEBUG_PRINTF	(("%04X\n", usIndex));

				if ( (0 < iNumDistinctIndices) && (usIndex == pusIndices [iNumDistinctIndices - 1]) )
				{
					// Yes - simply increase the count...
					puIndexCount [iNumDistinctIndices - 1]++;
				}

				else
				{
					// No - store this index and initialise the count to 1...
					puIndexCount [iNumDistinctIndices]	= 1;
					pusIndices [iNumDistinctIndices++]	= usIndex;
				}
			}

			// Now create the output data...
			for (int i = 0; i < iNumDistinctIndices; i++)
			{
				// Are we talking about the last couple of indices?
				if (i < (iNumDistinctIndices - 2))
				{
					// Output the current index...
					pusOutputIndices [uCurrentOutputIndex++]	= pusIndices [i];
					DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], pusOutputIndices [uCurrentOutputIndex - 1]));

					// Do we need a restart index?
					if ( (1 < puIndexCount [i]) && (1 < puIndexCount [i + 1]) )
					{
						bool	bDegenerateToStartStrip;

						// Yes - do so...
						pusOutputIndices [uCurrentOutputIndex++]	= usRestartIndex;

						for (u_int c = 1; c < puIndexCount [i]; c++)
						{
							DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], usRestartIndex));
						}

						// Acculmulate the original index...
						uOriginalInputIndex		+= puIndexCount [i++];
						bDegenerateToStartStrip	 = (0 == ((uOriginalInputIndex + puIndexCount [i]) & 1));

						// Also write the next index...
						pusOutputIndices [uCurrentOutputIndex++]	= pusIndices [i];

						// Should our next index be rendered counter-clockwise?
						if (bDegenerateToStartStrip)
						{
							// Yes - better add a degenerate to preserve the winding...
							pusOutputIndices [uCurrentOutputIndex++]	= pusIndices [i];

							for (u_int c = 2; c < puIndexCount [i]; c++)
							{
								DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], usRestartIndex));
							}

							for (u_int c = 0; c < 2; c++)
							{
								DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], pusOutputIndices [uCurrentOutputIndex - 1]));
							}
						}

						else
						{
							for (u_int c = 1; c < puIndexCount [i]; c++)
							{
								DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], usRestartIndex));
							}

							DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], pusOutputIndices [uCurrentOutputIndex - 1]));
						}
					}

					else
					{
						// No - write out any duplicate indices...
						for (u_int c = 1; c < puIndexCount [i]; c++)
						{
							pusOutputIndices [uCurrentOutputIndex++]	= pusIndices [i];
							DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], pusOutputIndices [uCurrentOutputIndex - 1]));
						}
					}
				}

				else
				{
					// Yes - write this index...
					pusOutputIndices [uCurrentOutputIndex++]	= pusIndices [i];
					DEBUG_PRINTF	(("%04X => %04X\n", pusIndices [i], pusOutputIndices [uCurrentOutputIndex - 1]));
				}

				// Acculmulate the original index...
				uOriginalInputIndex	+= puIndexCount [i];
			}

			// Validation...
			Asura_Assert	((uNumIndices == uOriginalInputIndex), "Mismatch in count between input indices and those used for tri-strip generation.");

			// Finally, remove the arrays...
			delete []	(pusIndices);
			delete []	(puIndexCount);
		}

		else
		{
			// Ensure we copy all of the indices...
			while (uNumIndices > uCurrentInputIndex)
			{
				pusOutputIndices [uCurrentOutputIndex++]	= pusInputIndices [uCurrentInputIndex++];
				DEBUG_PRINTF	(("%04X => %04X\n", pusInputIndices [uCurrentInputIndex - 1], pusOutputIndices [uCurrentOutputIndex - 1]));
			}
		}

		for (u_int i = 0; i < uCurrentOutputIndex; i++)
		{
			DEBUG_PRINTF	(("%04X\n", pusOutputIndices [i]));
		}

		DEBUG_PRINTF	(("...uCurrentInputIndex = %d, uCurrentOutputIndex = %d\n", uCurrentInputIndex, uCurrentOutputIndex));


		// Return the number of output indices...
		return	(uCurrentOutputIndex);
	}

	else
	{
		// No - are our input and output buffers the same?
		if (pusOutputIndices != pusInputIndices)
		{
			// No - so we'd better copy the data...
			for (u_int i = 0; i < uNumIndices; i++)
			{
				pusOutputIndices [i]	= pusInputIndices [i];
			}
		}
		
		// Return the same number of indices...
		return	(uNumIndices);
	}
}

/// End of function 'Tools_MeshRestartIndex::ReplaceDegeneratesInTriStrip'. ///


/// End of file. ///
