//////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_Processor.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "MeshProcessing/Tools_SmoothSkin_Processor.h"

// Engine independent.
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Core/Asura_Material_System.h"
#include "Materials/Tools_MaterialResponse_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Mesh/Asura_Mesh.h"

#include "Debugging/Asura_Logfile.h"

// Tools.
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "MeshProcessing/Tools_TangentSpace.h"
#include "Misc/Tools_HashTable.h"

#if defined(ASURA_SUPPORT_SKIN_ELEMENTS) || defined(ASURA_SUPPORT_STRIP_BY_ID)
# include "Rendering/Asura_Project_Skin_Elements.h"
#endif

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Tools_SmoothSkin_Processor::s_bAllowMultipleTextures = false;

Tools_SmoothSkin_Processor::ColourGenerationHook* Tools_SmoothSkin_Processor::s_pfnColourGenerationFunction = NULL;
Tools_SmoothSkin_Processor::VertexModificationHook* Tools_SmoothSkin_Processor::s_pfnVertexModificationFunction = NULL;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static inline bool IsTriDegenerate( const u_short ausVertexIndices[3] )
{
    return( ausVertexIndices[0] == ausVertexIndices[1]
            ||
            ausVertexIndices[0] == ausVertexIndices[2]
            ||
            ausVertexIndices[1] == ausVertexIndices[2] );
}


static inline bool ShouldExportPoly( const Asura_Mesh* pxMesh, int iPolyIdx )
{
    if( pxMesh->PolygonFlagDataValid )
    {
        if( pxMesh->PolygonFlagDataPtr[iPolyIdx].Flags & (AsuraPolygonFlag_DontExport | AsuraPolygonFlag_DontRender) )
        {
            return false;
        }
    }

    return true;
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Tools_SmoothSkin_SubElement.
//

Tools_SmoothSkin_SubElement::Tools_SmoothSkin_SubElement( void )
    : m_uNumTris( 0 )
    , m_pusIndices( NULL )
{
}


Tools_SmoothSkin_SubElement::~Tools_SmoothSkin_SubElement( void )
{
    delete[] m_pusIndices;
}

//
// Tools_SmoothSkin_Material.
//

Tools_SmoothSkin_Material::Tools_SmoothSkin_Material( void )
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	: m_pxSubElements( NULL )
#else
    : m_iMaterialID( ASURA_MATERIAL_NONE )
	, m_pxSubElements( NULL )
#endif
    
{
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	m_auMaterialResponseHashID[0] = ASURA_HASH_ID_UNSET;
	m_auMaterialResponseHashID[1] = ASURA_HASH_ID_UNSET;
	#endif
}


Tools_SmoothSkin_Material::~Tools_SmoothSkin_Material( void )
{
    delete[] m_pxSubElements;
}

//
// Tools_SmoothSkin_Vertex.
//

void Tools_SmoothSkin_Vertex::Set( u_int                 uMaterialIdx,
                                   const Asura_Vector_3& xPos,
                                   const Asura_Vector_3& xNormal,
                                   float                 fU,
                                   float                 fV,
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
								   u_int				 uFlags,
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
								   u_int				 uElement,
#endif
                                   const short           asBoneIndices[4],
								   const float           afBoneWeights[4],
								   const u_int			 uRGBAColour /* = 0xffffffff */ )
{
    m_xNormal = xNormal;
    m_xUV.x   = fU;
    m_xUV.y   = fV;

    Asura_Assert( uMaterialIdx < 65536, "" );
    m_xHashData.m_usMaterialIdx = static_cast<u_short>( uMaterialIdx );

    m_xHashData.m_xPos = xPos;

    // Copy bone data.
    u_int uBone = 0;
    for( ; uBone < 4 && afBoneWeights[uBone] > 0.0f; ++uBone )
    {
        m_xHashData.m_xBoneIndices.m_aucIndices[uBone] = static_cast<u_char>( asBoneIndices[uBone] );
        m_xHashData.m_afBoneWeights[uBone]           = afBoneWeights[uBone];
    }

    // Record weighting count.
    m_xHashData.m_usNumWeightings = static_cast<u_short>( uBone );

    // Fill unused entries.
    for( ; uBone < 4; ++uBone )
    {
        m_xHashData.m_xBoneIndices.m_aucIndices[uBone] = 0;
        m_xHashData.m_afBoneWeights[uBone]           = 0.0f;
    }

#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
	m_xHashData.m_uFlags = uFlags;
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
	m_xHashData.m_uElement = uElement;
#endif

	m_xHashData.m_uRGBAColour = uRGBAColour;

	m_uHash = Asura_GetHashID_Binary( &m_xHashData, sizeof(HashData) );
}


void Tools_SmoothSkin_Vertex::SetTangentAndBinormal( const Asura_Vector_3& xTangent, const Asura_Vector_3& xBinormal )
{
    m_xTangent  = xTangent;
    m_xBinormal = xBinormal;
}

void Tools_SmoothSkin_Vertex::SetDualPass( const Asura_Vector_2& xUV2, float fAlpha )
{
	m_xUV2.x   = xUV2.x;
	m_xUV2.y   = xUV2.y;
	m_fAlpha = fAlpha;
}


//
// Tools_SmoothSkin_Processor.
//

Tools_SmoothSkin_Processor::Tools_SmoothSkin_Processor( void ) : 
	m_bIgnoreDoubleSidedFlag( false ),
    m_bIncludeNormalMapData( false ),
	m_bIgnoreIDFromPolyFlags(false),
    m_pxInputSkin( NULL ),
    m_pxInputMesh( NULL ),
    m_uNumTris( 0 ),
    m_uNumMaterials( 0 ),
    m_pxMaterials( NULL ),
	#if !defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
    m_ppxMaterialLookupTable( NULL ),
	#endif
    m_uNumElements( 0 ),
    m_pxElements( NULL ),
    m_uNumVertices( 0 ),
    m_pxVertices( NULL ),
    m_pxVertexHashTable( NULL ),
	m_uNumBoneMapTables(0),
	m_ppxBoneMapTables(0),
	m_puBoneMapIndexes(0),
	m_puVertexBoneMapIndexes(0),
	m_piElementMap( NULL )
{
}


Tools_SmoothSkin_Processor::~Tools_SmoothSkin_Processor( void )
{
    delete[] m_pxElements;
    delete   m_pxVertexHashTable;
    delete[] m_pxVertices;
    delete[] m_pxMaterials;
	
	// Maps elements in this skin to element IDs
	if (m_piElementMap)
	{	
		delete[] m_piElementMap;
	}

	#if !defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
    // Special case for lookup table, since we adjust the pointer so that we
    // can dereference it using ASURA_MATERIAL_NONE as an index. See
    // CreateMaterialList().
    if( m_ppxMaterialLookupTable )
    {
        delete[] --m_ppxMaterialLookupTable;
    }
	#endif

	for (u_int i = 0; i < m_uNumBoneMapTables; ++i)
	{
		delete m_ppxBoneMapTables[i];
	}

	delete [] m_ppxBoneMapTables;
	delete [] m_puBoneMapIndexes;
	delete [] m_puVertexBoneMapIndexes;
}


bool Tools_SmoothSkin_Processor::Initialise( Asura_Hierarchy_Skin* pxInputSkin )
{
#define CHECK( e ) { if( !(e) ) { return false; } }

    Asura_Assert( pxInputSkin && pxInputSkin->GetMesh(), "" );
	if( !(pxInputSkin && pxInputSkin->GetMesh()) )		return false;

    // Ensure normals are up to date.
    pxInputSkin->CalculateSkinNormalData();

    // Can use const pointers internally now we've processed the normals.
    m_pxInputSkin = pxInputSkin;
    m_pxInputMesh = m_pxInputSkin->GetMesh();

    CHECK( m_pxInputMesh->PolygonDataValid && m_pxInputMesh->TextureDataValid && m_pxInputMesh->VertexDataValid );

    Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "Processing skin '%s'...", m_pxInputSkin->Name.GetString() );
    
    CHECK( CreateMaterialList()   );
    CHECK( ProcessIntoElements()  );
    CHECK( ProcessIntoMaterials() );
    
    GatherStats();

    return true;

#undef CHECK
}


u_int Tools_SmoothSkin_Processor::CalculateMaterialIdx( const Tools_SmoothSkin_Material* pxMaterial ) const
{
	return( pxMaterial - m_pxMaterials );
}


u_int Tools_SmoothSkin_Processor::CalculateVertexIdx( const Tools_SmoothSkin_Vertex* pxVertex ) const
{
    return( pxVertex - m_pxVertices );
}


bool Tools_SmoothSkin_Processor::CreateMaterialList( void )
{
    Asura_Assert( !m_pxMaterials, "" );

    m_uNumTris      = 0;
    m_uNumMaterials = 0;

	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		u_int* puTrisPerTexture = new u_int[ 1 + ASURA_MAX_NO_OF_TEXTURES ];
		if( !puTrisPerTexture ) { return false; }

		// Fiddle with the pointers. We'll use ASURA_MATERIAL_NONE and ASURA_TEXTURE_NONE too
		++puTrisPerTexture;

		//temporary storage for all the material pairs in the skin
		List<Tools_PolygonMaterialResponseIDs> xMaterialPairList;


		puTrisPerTexture[ ASURA_TEXTURE_NONE ] = 0;
		Asura_Mem::Set( puTrisPerTexture, 0, sizeof( u_int ) * ASURA_MAX_NO_OF_TEXTURES );

	#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		m_ppxMaterialLookupTable = new Tools_SmoothSkin_Material*[1 + ASURA_MAX_NO_OF_MATERIALS];
		if( !m_ppxMaterialLookupTable ) { return false; }

		u_int* puTrisPerTexture = new u_int[1 + ASURA_MAX_NO_OF_TEXTURES];
		if( !puTrisPerTexture ) { return false; }

		// Fiddle with array ptrs so that ASURA_MATERIAL_NONE and
		// ASURA_TEXTURE_NONE are valid indices. (Of course, we're assuming that
		// they're both defined to be -1...)
		++m_ppxMaterialLookupTable;
		++puTrisPerTexture;

		for( int i = ASURA_MATERIAL_NONE; i < ASURA_MAX_NO_OF_MATERIALS; ++i ) { m_ppxMaterialLookupTable[i] = NULL; }
		for( int i = ASURA_TEXTURE_NONE; i < ASURA_MAX_NO_OF_TEXTURES; ++i ) { puTrisPerTexture[i] = 0; }

	#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

    // Iterate through the polys, counting tris and finding all distinct
    // materials.
    for( int iPoly = 0; iPoly < m_pxInputMesh->NumberOfPolygons; ++iPoly )
    {
        if( !ShouldExportPoly(m_pxInputMesh, iPoly) ) { continue; }

		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			Tools_PolygonMaterialResponseIDs xMaterialPair(m_pxInputMesh->TextureDataPtr[ iPoly ].m_auMaterialResponseHashID[0],m_pxInputMesh->TextureDataPtr[ iPoly ].m_auMaterialResponseHashID[1]);
		#else
			int iMaterialID = m_pxInputMesh->TextureDataPtr[iPoly].m_iMaterialID;
		#endif

        const u_int uNumTrisInPoly = ( m_pxInputMesh->PolygonDataPtr[iPoly].IsQuad() ? 2 : 1 );
        m_uNumTris += uNumTrisInPoly;
        if( ShouldPolyBeDoubleSided(iPoly) )
        {
            //add extra tris for double sided polygons
            m_uNumTris += uNumTrisInPoly;
        }

		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			
			
			const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( xMaterialPair.GetMaterial(0) );
			if( !pxMaterial )
			{
				puTrisPerTexture[ ASURA_TEXTURE_NONE ] += uNumTrisInPoly;
			}
			else
			{
				puTrisPerTexture[ pxMaterial->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) ] += uNumTrisInPoly;
			}

			if( !xMaterialPairList.contains(xMaterialPair) )
			{
				xMaterialPairList.add_entry(xMaterialPair);
				++m_uNumMaterials;
			}

		#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			puTrisPerTexture[Asura_Material_System::GetTextureID(iMaterialID)] += uNumTrisInPoly;        

			// Adjust material ID to array index.
			if( ASURA_MATERIAL_NONE != iMaterialID ) { iMaterialID -= ASURA_MATERIAL_FIRSTID; }

			Asura_Assert( ASURA_MATERIAL_NONE <= iMaterialID && iMaterialID < ASURA_MAX_NO_OF_MATERIALS, "" );

			// Already encountered a poly using this material?
			if( !m_ppxMaterialLookupTable[iMaterialID] )
			{
				// At this point, we just want a non-NULL pointer. We'll set the
				// correct value in the loop below.
				++m_ppxMaterialLookupTable[iMaterialID];

				++m_uNumMaterials;
			}

		#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
    }

    // Cope with perverse meshes with no polys to export.
    if( !m_uNumTris ) { return true; }

    // Deal with meshes with multiple textures. Firstly, we need to output
    // an error message. Beyond that, we should try to cope. I think the most
    // obvious way is to remove tris using the offending texture ID(s),
    // where the "correct" one is regarded as the one used by most tris.
    if (!s_bAllowMultipleTextures)
	{
        bool  bMultiple     = false;
        u_int uHighestCount = 0;
        int   iIDOfHighest  = INT_MIN;

        for( int i = ASURA_TEXTURE_NONE; i < ASURA_MAX_NO_OF_TEXTURES; ++i )
        {
            if( !puTrisPerTexture[i] ) { continue; }

            bool bHighest;
            if( INT_MIN == iIDOfHighest )
            {
                bHighest = true;
            }
            else 
            {
                bMultiple = true;
                bHighest = puTrisPerTexture[i] > uHighestCount;
            }

            if( bHighest )
            {
                uHighestCount = puTrisPerTexture[i];
                iIDOfHighest  = i;
            }
        }

        if( bMultiple )
        {
            // Report error.
            Tools_Level_Export::AddMessage( Tools_Level_Export::MT_ERROR,
                                            "Skin '%s' is using multiple textures. All but one will be removed.",
                                            m_pxInputSkin->Name.GetString() );

            // Report (just in the logfile) which texture has been chosen.
            {
                const char* szTexName = ( ASURA_TEXTURE_NONE == iIDOfHighest ? "-none-" : Asura_Texture_Management::GetTextureName(iIDOfHighest)->GetString() );
                Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                                "    Keeping all materials using texture '%s'.",
                                                szTexName );
            }

            // Remove all invalid ones.
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			for(LIF<Tools_PolygonMaterialResponseIDs> xMaterialLIF(&xMaterialPairList); !xMaterialLIF.done();)
			{
				const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID(xMaterialLIF().GetMaterial(0));
				const int iTexID = pxMaterial ? pxMaterial->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) : ASURA_TEXTURE_NONE;
				if( iTexID != iIDOfHighest )
				{
					const char* const szMaterialName = (pxMaterial ? pxMaterial->GetName().GetString() : "-none-");
					const char* const szTexName = (iTexID == ASURA_TEXTURE_NONE ? "-none-" : Asura_Texture_Management::GetTextureName( iTexID )->GetString());
					Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
													"	Removed material '%s' which used texture '%s'.)",
													szMaterialName,
													szTexName );

					xMaterialLIF.delete_current();
					--m_uNumMaterials;
				}
				else
				{
					xMaterialLIF.next();
				}
			}
			

#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

            for( int iMaterialID = ASURA_MATERIAL_NONE; iMaterialID < ASURA_MAX_NO_OF_MATERIALS; ++iMaterialID )
            {
                if( m_ppxMaterialLookupTable[iMaterialID] )
                {
                    int iRealMaterialID = ( ASURA_MATERIAL_NONE == iMaterialID ? ASURA_MATERIAL_NONE : ASURA_MATERIAL_FIRSTID + iMaterialID );
                    int iTexID = Asura_Material_System::GetTextureID( iRealMaterialID );
                    if( iTexID != iIDOfHighest )
                    {
                        // Report (just in the logfile) the one(s) we're
                        // removing.
                        const char* szMaterialName = ( ASURA_MATERIAL_NONE == iRealMaterialID ? "-none-" : Asura_Material_System::GetMaterial(iRealMaterialID)->m_xMaterialName.GetString() );
                        if( !szMaterialName ) { szMaterialName = "-unnamed-"; }
                        const char* szTexName = ( ASURA_TEXTURE_NONE == iTexID ? "-none-" : Asura_Texture_Management::GetTextureName(iTexID)->GetString() );
                        Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                                        "    Removed material %i ('%s') which used texture '%s'.)",
                                                        iRealMaterialID,
                                                        szMaterialName,
                                                        szTexName );

                        // And remove from lookup table.
                        m_ppxMaterialLookupTable[iMaterialID] = NULL;
                        --m_uNumMaterials;
                    }
                }
            }

#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

            // Fix up tri count.
            u_int uTrisRemoved = 0;
            for( int iTexID = ASURA_TEXTURE_NONE; iTexID < ASURA_MAX_NO_OF_TEXTURES; ++iTexID )
            {
                if( iTexID != iIDOfHighest ) { uTrisRemoved += puTrisPerTexture[iTexID]; }
            }
            Asura_Assert( uTrisRemoved && uTrisRemoved < m_uNumTris, "" );
            Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                            "    Total of %u tris removed (%u -> %u).",
                                            uTrisRemoved,
                                            m_uNumTris,
                                            m_uNumTris - uTrisRemoved );
            m_uNumTris -= uTrisRemoved;
        }
    }
    // End code to deal with multiple textures.

    delete[] --puTrisPerTexture;
    puTrisPerTexture = NULL;

    m_pxMaterials = new Tools_SmoothSkin_Material[m_uNumMaterials];
    if( !m_pxMaterials ) { return false; }

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

    // Do two passes through the materials. First, add opaque materials.
    // Second pass, add non-opaque ones.
	Tools_SmoothSkin_Material* pxOut = m_pxMaterials;
	for( u_int uPass = 0; uPass < 2; ++uPass )
	{
		for(LIF<Tools_PolygonMaterialResponseIDs> xMaterialLIF(&xMaterialPairList); !xMaterialLIF.done(); xMaterialLIF.next())
		{

			const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( xMaterialLIF().GetMaterial(0) );

			// Check if we're dealing with this material in this pass.
			const bool bOpaque = (!pxMaterial || pxMaterial->GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_OPAQUE);
			if( bOpaque != (0 == uPass) ) continue;

			// Store material details.
			pxOut->m_auMaterialResponseHashID[0] = xMaterialLIF().GetMaterial(0);
			pxOut->m_auMaterialResponseHashID[1] = xMaterialLIF().GetMaterial(1);

			++pxOut;
		}
	}

#else //!defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

    int aiToSkinConversionList[ASURA_MAX_NO_OF_MATERIALS];

    // Prepare conversion list.
    {
        //Need to consider the meshes of all the LOD versions of the skin to get the correct
        //conversion list.
        //This is a tad bit dubious for files that have only ever been exported from Max, since
        //there isn't any guarantee that the ordering will be the same. But since the extra materials
        //are generated in tools , any multiple material skins should hopefully have been saved
        //in tools at least once.
        //RAR 06/10/2005
        List<Asura_Mesh*> xMeshList;
        Tools_SaveAsura::ListMeshesForSkin(m_pxInputSkin->GetRootSkin(),xMeshList);

        Tools_SaveAsura::PrepareMeshMaterialConversionLists( xMeshList, aiToSkinConversionList, NULL );
    }

    // Do two passes through the materials. First, add opaque materials.
    // Second pass, add non-opaque ones.
    Tools_SmoothSkin_Material* pxOut = m_pxMaterials;
    for( u_int uPass = 0; uPass < 2; ++uPass )
    {
        for( int iMaterialID = ASURA_MATERIAL_NONE; iMaterialID < ASURA_MAX_NO_OF_MATERIALS; ++iMaterialID )
        {
            if( !m_ppxMaterialLookupTable[iMaterialID] ) { continue; }

            const int iRealMaterialID = ( ASURA_MATERIAL_NONE == iMaterialID ? ASURA_MATERIAL_NONE : ASURA_MATERIAL_FIRSTID + iMaterialID );

            // Check if we're dealing with this material in this pass.
            const u_int uMaterialFlags = Asura_Material_System::GetMaterialFlags( iRealMaterialID );
            const bool bOpaque = ( 0 == (uMaterialFlags & (AsuraMaterialFlag_Transparent | AsuraMaterialFlag_PS2AdditiveAlpha)) );
            if( bOpaque != (0 == uPass) ) { continue; }

            // Fix up lookup table ptr.
            m_ppxMaterialLookupTable[iMaterialID] = pxOut;
        
            // Store material details.
            pxOut->m_iMaterialID           = iRealMaterialID;
            pxOut->m_iMaterialIDWithinSkin = ( ASURA_MATERIAL_NONE == iMaterialID ? ASURA_MATERIAL_NONE : aiToSkinConversionList[iMaterialID] );

            ++pxOut;
        }
    }
#endif

    Asura_Assert( static_cast<u_int>(pxOut - m_pxMaterials) == m_uNumMaterials, "" );

    return true;
}

namespace {

	class BoneTable
	{
	public:
		class Strip;

		class BoneUsageID
		{
			static u_int m_uMaxNumBones;
			u_int*      m_puBonesUsed;

		public:
			static void SetMaxNumBones(u_int uMaxNumBones);

			BoneUsageID();
			~BoneUsageID();
			BoneUsageID(const BoneUsageID&);
			void operator=(const BoneUsageID&);

			BoneUsageID& operator +=(const BoneUsageID&);
			BoneUsageID& operator +=(u_int uBoneIndex);
			bool isBoneUsed(u_int uBoneIndex) const;
			bool isBoneUsed(const BoneUsageID&) const;
			u_int GetNumNotUsed(const BoneUsageID&) const;
			u_int GetNumUsed() const;
		};

		class Bone
		{
			friend class Strip;
			u_int				m_uBoneIndex;
			const Strip**		m_pxStrips;
			u_int				m_uNumStrips;

			void			AddStrip(const Strip* pxStrip);

		public:
			Bone() : m_uBoneIndex(0), m_pxStrips(NULL), m_uNumStrips(0) {};
			~Bone() {delete[] m_pxStrips;}

			void			SetBoneIndex(u_int uBoneIndex) {m_uBoneIndex = uBoneIndex;}
			u_int			GetBoneIndex() const {return m_uBoneIndex;}


			u_int			GetNumStrips() const {return m_uNumStrips;}
			const Strip*	GetStrip(u_int uStrip) const {return m_pxStrips[uStrip];}
		};
		class Strip
		{
		public:
			u_int				m_uMaterial;
			u_int				m_uElement;
			Strip*				m_pxNext;
			Strip*				m_pxPrev;
			const Bone**		m_pxBones;
			u_int				m_uNumBones;
			u_int				m_uMergeIndex;
			BoneUsageID			m_xUsage;

		public:
			Strip() : m_pxNext(NULL), m_pxPrev(NULL), m_uMaterial(0), m_uElement(0), m_pxBones(NULL), m_uNumBones(0) {};
			~Strip() {delete[] m_pxBones;}

			void		SetMatElem(u_int uMaterial, u_int uElem) { m_uMaterial = uMaterial; m_uElement = uElem;}
			void		SetMergeIndex(u_int uMergeIndex) { m_uMergeIndex = uMergeIndex;}
			u_int		GetMaterial() const { return m_uMaterial;}
			u_int		GetElement() const { return m_uElement;}
			u_int		GetMergeIndex() const { return m_uMergeIndex;}

			void		AddBone(Bone* pxBone);

			const BoneUsageID &GetUsage() const {return m_xUsage;}

			u_int		GetNumBones() const {return m_uNumBones;}
			const Bone*	GetBone(u_int uBone) const {return m_pxBones[uBone];}

			void Unlink(Strip *&pxHead)
			{
				if (m_pxNext)
				{
					m_pxNext->m_pxPrev = m_pxPrev;
				} else if (m_pxPrev) {
					m_pxPrev->m_pxNext = NULL;
				}

				if (m_pxPrev)
				{
					m_pxPrev->m_pxNext = m_pxNext;
				} else if ((pxHead = m_pxNext) != NULL)
				{
					m_pxNext->m_pxPrev = NULL;
				}

				m_pxNext = NULL;
				m_pxPrev = NULL;
			}

			void LinkAfter(Strip *pxItem)
			{
				if (pxItem)
				{
					m_pxNext = pxItem->m_pxNext;
					m_pxPrev = pxItem;
					if (pxItem->m_pxNext)
					{
						pxItem->m_pxNext->m_pxPrev = this;
					}
					pxItem->m_pxNext = this;
				}
			}
			void LinkBefore(Strip *pxItem)
			{
				if (pxItem)
				{
					m_pxNext = pxItem;
					m_pxPrev = pxItem->m_pxPrev;
					if (pxItem->m_pxPrev)
					{
						pxItem->m_pxPrev->m_pxNext = this;
					}
					pxItem->m_pxPrev = this;
				}
			}
			Strip *GetNext() 
			{
				return m_pxNext;
			}
			Strip *GetPrev() 
			{
				return m_pxPrev;
			}
		};
		struct VertexElement {
			u_int uMaterial, uElement;
		};
		struct BoneTableEntry {
			u_int uSize;
			u_int *pEntries;
			u_int *pInverse;

			BoneTableEntry(u_int uNumEntries, u_int uNumBones) : uSize(uNumEntries), pEntries(new u_int[uNumEntries]), pInverse(new u_int[uNumBones]) {}
			~BoneTableEntry() {delete [] pEntries;delete [] pInverse;}
		};
	private:
		Bone *m_pxBones;
		Strip *m_pxStrips;
		Strip **m_ppxMerged;
		// for each merged strip there is an int array

		BoneTableEntry **m_ppxBoneTable;
		VertexElement *m_pxVertexTable;

		const u_int m_uNumBones;
		const u_int m_uNumMaterials;
		const u_int m_uNumElements;
		u_int m_uNumMerged;

		void Merge(u_int uMaxNumBones);
		void RemapVertex	(const Tools_SmoothSkin_Processor &xProcessor);

		const Bone*	GetBone(u_int uBone) const
		{
			Asura_Assert(uBone < m_uNumBones, "");
			return &m_pxBones[uBone];
		}
		const Strip*	GetStrip(u_int uStrip) const
		{
			Asura_Assert(uStrip < m_uNumMaterials * m_uNumElements, "");
			return &m_pxStrips[uStrip];
		}
		const Strip*	GetStrip(u_int uMaterial, u_int uElement) const 
		{
			Asura_Assert(uMaterial < m_uNumMaterials && uElement < m_uNumElements, "");
			return GetStrip(uMaterial * m_uNumElements + uElement);
		}


		Bone*	GetBone(u_int uBone) 
		{
			Asura_Assert(uBone < m_uNumBones, "");
			return &m_pxBones[uBone];
		}
		Strip*	GetStrip(u_int uStrip) 
		{
			Asura_Assert(uStrip < m_uNumMaterials * m_uNumElements, "");
			return &m_pxStrips[uStrip];
		}
		Strip*	GetStrip(u_int uMaterial, u_int uElement) 
		{
			Asura_Assert(uMaterial < m_uNumMaterials && uElement < m_uNumElements, "");
			return GetStrip(uMaterial * m_uNumElements + uElement);
		}
	public:
		BoneTable(u_int uNumBones, u_int uNumMaterials, u_int uNumElements);
		~BoneTable();

		void Process(const Tools_SmoothSkin_Processor &xProcessor, u_int uMaxNumBones);
		
		u_int GetBoneMapIndex(u_int uMaterial, u_int uElement) const;
		u_int GetNumBoneMaps() const;
		VertexElement &GetVertex(u_int uVertexIndex) const
		{
			return m_pxVertexTable[uVertexIndex];
		}
		const BoneTableEntry *GetBoneMap(u_int uBoneMapIndex) const;


	};

	u_int BoneTable::BoneUsageID::m_uMaxNumBones = 0;

	void BoneTable::BoneUsageID::SetMaxNumBones(u_int uMaxNumBones)
	{

		m_uMaxNumBones = uMaxNumBones;
	}

	BoneTable::BoneUsageID::BoneUsageID() : m_puBonesUsed(new u_int[(m_uMaxNumBones + 31)/32])
	{
		memset(m_puBonesUsed, 0, ((m_uMaxNumBones + 31)/32) * 4);
	}
	BoneTable::BoneUsageID::BoneUsageID(const BoneUsageID& xCopyFrom) : m_puBonesUsed(new u_int[(m_uMaxNumBones + 31)/32])
	{
		memcpy(m_puBonesUsed, xCopyFrom.m_puBonesUsed, ((m_uMaxNumBones + 31)/32) * 4);
	}

	BoneTable::BoneUsageID::~BoneUsageID()
	{
		delete[]  m_puBonesUsed;
	}

	void BoneTable::BoneUsageID::operator=(const BoneUsageID& xCopyFrom)
	{
		memcpy(m_puBonesUsed, xCopyFrom.m_puBonesUsed, ((m_uMaxNumBones + 31)/32) * 4);
	}

	BoneTable::BoneUsageID& BoneTable::BoneUsageID::operator +=(const BoneUsageID&xBoneUsage)
	{
		for (u_int i = 0; i < (m_uMaxNumBones + 31)/32; ++i)
		{
			m_puBonesUsed[i] |= xBoneUsage.m_puBonesUsed[i];
		}
		return *this;
	}
	BoneTable::BoneUsageID& BoneTable::BoneUsageID::operator +=(u_int uBoneIndex)
	{
		m_puBonesUsed[uBoneIndex / 32] |= (1u<<(uBoneIndex & 31));
		return *this;
	}
	bool BoneTable::BoneUsageID::isBoneUsed(u_int uBoneIndex)  const
	{
		return (m_puBonesUsed[uBoneIndex / 32] & (1u<<(uBoneIndex & 31))) != 0;
	}
	bool BoneTable::BoneUsageID::isBoneUsed(const BoneUsageID& xBoneUsage)  const
	{
		for (u_int i = 0; i < (m_uMaxNumBones + 31)/32; ++i)
		{
			if ((m_puBonesUsed[i] | xBoneUsage.m_puBonesUsed[i]) != m_puBonesUsed[i])
			{
				return false;
			}
		}
		return true;
	}

	u_int BoneTable::BoneUsageID::GetNumNotUsed(const BoneUsageID& xBoneUsage) const
	{
		u_int uNotUsed = 0;
		for (u_int i = 0; i < (m_uMaxNumBones + 31)/32; ++i)
		{
			if ((m_puBonesUsed[i] | xBoneUsage.m_puBonesUsed[i]) != m_puBonesUsed[i])
			{
				for (u_int j = 0; j < 32; ++j)
				{
					if (xBoneUsage.m_puBonesUsed[i] & (1<<j))
					{
						uNotUsed += !(m_puBonesUsed[i] & (1<<j));
					}
				}
			}
		}
		return uNotUsed;
	}

	u_int BoneTable::BoneUsageID::GetNumUsed() const
	{
		u_int uUsed = 0;
		for (u_int i = 0; i < (m_uMaxNumBones + 31)/32; ++i)
		{
			if (m_puBonesUsed[i] != 0)
			{
				for (int j = 0; j < 32; ++j)
				{
					uUsed += (m_puBonesUsed[i] & (1<<j)) != 0;
				}
			}
		}
		return uUsed;
	}



	void BoneTable::Bone::AddStrip(const Strip* pxStrip)
	{
		const Strip** pxStrips = new const Strip*[m_uNumStrips + 1];
		Asura_Mem::Copy(pxStrips, m_pxStrips, sizeof(Strip*) * m_uNumStrips);
		pxStrips[m_uNumStrips++] = pxStrip;
		delete[] m_pxStrips;
		m_pxStrips = pxStrips;
	}
	void BoneTable::Strip::AddBone(Bone* pxBone)
	{
		if (m_xUsage.isBoneUsed(pxBone->GetBoneIndex()))
		{
			return;
		}

		const Bone** pxBones = new const Bone*[m_uNumBones + 1];
		Asura_Mem::Copy(pxBones, m_pxBones, sizeof(Bone*) * m_uNumBones);
		pxBones[m_uNumBones++] = pxBone;
		delete[] m_pxBones;
		m_pxBones = pxBones;

		m_xUsage += pxBone->GetBoneIndex();

		pxBone->AddStrip(this);
	}

	BoneTable::BoneTable(u_int uNumBones, u_int uNumMaterials, u_int uNumElements) 
	: m_pxBones(new Bone[uNumBones]), m_pxStrips(new Strip[uNumMaterials * uNumElements]),
		m_ppxMerged(new Strip*[uNumMaterials * uNumElements]), m_ppxBoneTable(new BoneTableEntry*[uNumMaterials * uNumElements]), 
		m_pxVertexTable(NULL),
		m_uNumBones(uNumBones), m_uNumMaterials(uNumMaterials), m_uNumElements(uNumElements), m_uNumMerged(0)
	{
		for (u_int uBoneIdx = 0; uBoneIdx < uNumBones; ++uBoneIdx)
		{
			GetBone(uBoneIdx)->SetBoneIndex(uBoneIdx);
		}
		for (u_int uMatIdx = 0; uMatIdx < uNumMaterials; ++uMatIdx)
		{
			for (u_int uElemIdx = 0; uElemIdx < uNumElements; ++uElemIdx)
			{
				GetStrip(uMatIdx, uElemIdx)->SetMatElem(uMatIdx, uElemIdx);
				m_ppxMerged[uMatIdx * uNumElements + uElemIdx] = NULL;
				m_ppxBoneTable [uMatIdx * uNumElements + uElemIdx] = NULL;
			}
		}
	}

	BoneTable::~BoneTable() 
	{
		delete[] m_pxBones;
		delete[] m_pxStrips;
		delete[] m_ppxMerged;
		for (u_int i = 0; i < m_uNumMerged; ++i)
		{
			delete  m_ppxBoneTable[i];
		}
		delete [] m_ppxBoneTable;
		delete [] m_pxVertexTable;
	}


	void BoneTable::Process(const Tools_SmoothSkin_Processor &xProcessor, u_int uMaxNumBones)
	{
		const u_int uNumMaterials = xProcessor.GetNumMaterials();
		const u_int uNumElements = xProcessor.GetNumElements();


		for (u_int uMatIdx = 0; uMatIdx < uNumMaterials; ++uMatIdx)
		{
			const Tools_SmoothSkin_Material& xMaterial(xProcessor.GetMaterial(uMatIdx));

			// Loop through each sub-element
			for (u_int uElemIdx = 0; uElemIdx < uNumElements; ++uElemIdx)
			{
				Strip *pxStrip = GetStrip(uMatIdx, uElemIdx);
				const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement(uElemIdx);
				const u_short *pusIndicies = xSubElement.GetIndices();

				for (u_int uTri = 0; uTri < xSubElement.GetNumTris() * 3; ++uTri)
				{
					const Tools_SmoothSkin_Vertex *pxVertex = &xProcessor.GetVertex(pusIndicies[uTri]);

					const float *pfBoneWeights = pxVertex->GetBoneWeights();
					const u_char *pusBoneIndices = pxVertex->GetBoneIndices();
					for (u_int uBone = 0; uBone < pxVertex->GetNumWeightings(); ++uBone)
					{
						if (pfBoneWeights[uBone] > 0.0f)
						{
							Bone *pxBone = GetBone(pusBoneIndices[uBone]);

							pxStrip->AddBone(pxBone);
							Asura_Assert(pxStrip->GetNumBones() == pxStrip->GetUsage().GetNumUsed(), "");
						}
					}
				}
			}
		}
		Merge(uMaxNumBones);
		RemapVertex(xProcessor);
	}



	void BoneTable::Merge(u_int uMaxNumBones)
	{
		if (m_uNumElements * m_uNumMaterials > 0)
		{
			Strip *pxHead = NULL;

			for (u_int i = 0; i < m_uNumElements * m_uNumMaterials; ++i)
			{
				Strip *pxStrip = GetStrip(i);

				Asura_Assertf(pxStrip->GetNumBones() <= uMaxNumBones, ("Strip from skin uses more than %d Bones in a single material/element pair", uMaxNumBones));

				Strip *pxStripPrev = NULL;
				for (Strip *pxStripIter = pxHead; pxStripIter; (pxStripIter = pxStripIter->m_pxNext), (pxStripPrev = pxStripIter))
				{
					if (pxStripIter->GetNumBones() < pxStrip->GetNumBones())
					{
						pxStripPrev = pxStripIter->GetPrev();
						break;
					} 
				}
				if (pxStripPrev)
				{
					pxStrip->LinkAfter(pxStripPrev);
				} else{
					pxStrip->LinkBefore(pxHead);
					pxHead = pxStrip;
				}
			}

			while (pxHead)
			{
				BoneUsageID uUsage = pxHead->GetUsage();
				u_int uBestDiff;
				Strip *pxBest;
				Strip *pxMerged = pxHead; 
				u_int uNumMerged = 1;
				pxHead = pxHead->GetNext();
				pxMerged->Unlink(pxHead);
				pxMerged->SetMergeIndex(m_uNumMerged);

				while ((uBestDiff = (uMaxNumBones - uUsage.GetNumUsed()) + 1) > 1)
				{
					pxBest = NULL;

					for (Strip *pxStrip = pxHead, *pxNext; pxStrip; pxStrip = pxNext)
					{
						pxNext = pxStrip->GetNext();

						u_int uBones = uUsage.GetNumNotUsed(pxStrip->GetUsage());

						if (uBones == 0)
						{        
							pxStrip->Unlink(pxHead);
							pxStrip->LinkAfter(pxMerged);
							pxStrip->SetMergeIndex(m_uNumMerged);
							++uNumMerged;
						} else if (uBones < uBestDiff) {
							uBestDiff = uBones;
							pxBest = pxStrip;
						}
					}

					if (pxBest)
					{
						uUsage += pxBest->GetUsage();
						pxBest->Unlink(pxHead);
						pxBest->LinkAfter(pxMerged);
						pxBest->SetMergeIndex(m_uNumMerged);
						++uNumMerged;
					} else {
						// end the while loop
						break;
					}
				} 
				BoneTableEntry *pxBoneTableEntry = new BoneTableEntry(uUsage.GetNumUsed(), m_uNumBones);
				for (u_int u = 0, uBoneIndex = 0; u < m_uNumBones; ++u)
				{
					if (uUsage.isBoneUsed(u))
					{
						Asura_Assert(uBoneIndex < pxBoneTableEntry->uSize, "");
						pxBoneTableEntry->pEntries[uBoneIndex] = u;
						pxBoneTableEntry->pInverse[u] = uBoneIndex;
						++uBoneIndex;
					} else {
						pxBoneTableEntry->pInverse[u] = 0xFFFFFFFF;
					}
				}

				m_ppxBoneTable[m_uNumMerged] = pxBoneTableEntry;
				m_ppxMerged[m_uNumMerged++] = pxMerged;

//				Asura_OutputToDebugger(("Created %d Merged strips using %d bones\n", uNumMerged, pxBoneTableEntry->uSize));
			}
		}
	}

	void BoneTable::RemapVertex	(const Tools_SmoothSkin_Processor &xProcessor)
	{
		u_int uNumVerts = xProcessor.GetNumVertices();

		m_pxVertexTable = new VertexElement[uNumVerts];

		u_int *puVertexUsageCounts = new u_int[uNumVerts];
		u_int *puVertexUsageIndexes = new u_int[uNumVerts];

		memset (puVertexUsageCounts, 0, uNumVerts * sizeof(u_int));
		memset (puVertexUsageIndexes, 0, uNumVerts * sizeof(u_int));


		for (u_int uMergedIndex = 0; uMergedIndex < m_uNumMerged; ++uMergedIndex)
		{
			for (Strip *pxStrip = m_ppxMerged[uMergedIndex]; pxStrip; pxStrip = pxStrip->GetNext())
			{
				const Tools_SmoothSkin_Material& xMaterial(xProcessor.GetMaterial(pxStrip->GetMaterial()));
				const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement(pxStrip->GetElement());
				const u_short *pusIndicies = xSubElement.GetIndices();

				for (u_int uTri = 0; uTri < xSubElement.GetNumTris() * 3; ++uTri)
				{
					u_int uIndex = pusIndicies[uTri];
					if (puVertexUsageIndexes[uIndex] != uMergedIndex+1)
					{
						puVertexUsageIndexes[uIndex] = uMergedIndex+1;
						++puVertexUsageCounts[uIndex];

						m_pxVertexTable[uIndex].uElement = pxStrip->GetElement();
						m_pxVertexTable[uIndex].uMaterial= pxStrip->GetMaterial();
					}
				}
			}
		}

		delete [] puVertexUsageCounts;
		delete [] puVertexUsageIndexes;
	}

	u_int BoneTable::GetBoneMapIndex(u_int uMaterial, u_int uElement) const
	{
		return GetStrip(uMaterial, uElement)->GetMergeIndex();
	}
	u_int BoneTable::GetNumBoneMaps() const
	{
		return m_uNumMerged;
	}
	const BoneTable::BoneTableEntry *BoneTable::GetBoneMap(u_int uBoneMapIndex) const
	{
		return m_ppxBoneTable[uBoneMapIndex];
	}
}

void Tools_SmoothSkin_Processor::CreateBoneRemapTables(u_int uMaxNumBones)
{
	if ((u_int)m_pxInputSkin->NumberOfBones < uMaxNumBones)
	{
		return;
	}


	BoneTable::BoneUsageID::SetMaxNumBones(m_pxInputSkin->NumberOfBones);
	
	const u_int uNumBones = m_pxInputSkin->NumberOfBones;

	BoneTable xBoneTable(uNumBones,GetNumMaterials(),GetNumElements());

	xBoneTable.Process(*this, uMaxNumBones);

	m_ppxBoneMapTables = new Tools_SmoothSkin_BoneMap*[m_uNumBoneMapTables = xBoneTable.GetNumBoneMaps()];

	for (u_int i = 0; i < m_uNumBoneMapTables; ++i)
	{
		const BoneTable::BoneTableEntry *pxBoneMap = xBoneTable.GetBoneMap(i);
		m_ppxBoneMapTables[i] = new Tools_SmoothSkin_BoneMap(pxBoneMap->uSize, uNumBones);
		memcpy(m_ppxBoneMapTables[i]->m_puBoneMap, pxBoneMap->pEntries, sizeof(u_int) * pxBoneMap->uSize);
		memcpy(m_ppxBoneMapTables[i]->m_puInvBoneMap, pxBoneMap->pInverse, sizeof(u_int) * uNumBones);
	}
	m_puBoneMapIndexes = new u_int[GetNumMaterials() * GetNumElements()];
	for (u_int m = 0; m < GetNumMaterials(); ++m)
	{
		for (u_int e = 0; e < GetNumElements(); ++e)
		{
			m_puBoneMapIndexes[m * GetNumElements() + e] = xBoneTable.GetBoneMapIndex(m,e);
		}
	}
	m_puVertexBoneMapIndexes = new u_int[GetNumVertices()];
	for (u_int v = 0; v < GetNumVertices(); ++v)
	{
		m_puVertexBoneMapIndexes[v] = xBoneTable.GetBoneMapIndex(xBoneTable.GetVertex(v).uMaterial, xBoneTable.GetVertex(v).uElement);
	}

}




void Tools_SmoothSkin_Processor::GatherStats( void ) const
{
    u_int auVertsPerWeightingCount[4] = { 0, 0, 0, 0 };

    u_int uMaxWeightings   = 0;
    u_int uTotalWeightings = 0;

    for( u_int uVtx = 0; uVtx < m_uNumVertices; ++uVtx )
    {
        u_int uNumWeightings = GetVertex(uVtx).GetNumWeightings();
        // JAT 11/04/06 added extra text thanks to MH
        Asura_Assertf( 0 < uNumWeightings && uNumWeightings <= 4,  ("unexpected number of weightings in skins' %s', should be 1 to 4, value is %u",  m_pxInputSkin->Name.GetString(), uNumWeightings) );

        // JAT 11/04/06 try to cope gracefully with the problem
        uNumWeightings = Asura_Maths::Min( uNumWeightings, 4u );
        uNumWeightings = Asura_Maths::Max( uNumWeightings, 1u );

        uMaxWeightings = Asura_Maths::Max( uMaxWeightings, uNumWeightings );

        uTotalWeightings += uNumWeightings;

        ++auVertsPerWeightingCount[uNumWeightings - 1];
    }

    float fAvgWeightings = static_cast<float>(uTotalWeightings) / static_cast<float>(m_uNumVertices);

    // Log.
    {
        Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                        "    Weighting stats:\n        Max per vertex: %u\n        Avg per vertex: %.1f",
                                        uMaxWeightings,
                                        fAvgWeightings );

        Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "        Num verts with each weighting count:" );
        for( u_int u = 0; u < 4; ++u )
        {
            Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "            %u: %u", u + 1, auVertsPerWeightingCount[u] );
        }
    }
}





// Elements are now used for detachable sections of skins - for wounds and
// detachable skins. Element IDs are defined in Asura_Project_Skin_Elements.cpp/h,
// and are the 'body part IDs'. Beware - These are *not* the same as element 
// indices within the skin processor - m_piElementMap maps element indices to
// the 'body part IDs'. These IDs are now stored in the strip structure

int  Tools_SmoothSkin_Processor::GetElementID(u_int uElement) const
{
#if defined(ASURA_SUPPORT_SKIN_ELEMENTS) || defined(ASURA_SUPPORT_STRIP_BY_ID)
	return m_piElementMap[uElement];
#else
	return 0;
#endif
}


#if defined(ASURA_SUPPORT_SKIN_ELEMENTS) || defined(ASURA_SUPPORT_STRIP_BY_ID)

bool Tools_SmoothSkin_Processor::ProcessIntoElements( void )
{
	// Allocate space for the maximum number of vertices we can create. We
	// should never actually reach this but it's quicker to allocate a block
	// up front.
	u_int uOriginalNumVertices = 3 * m_uNumTris;
	m_pxVertices = new Tools_SmoothSkin_Vertex[uOriginalNumVertices];
	m_uNumVertices = 0;

	// Allocate hash table.
	m_pxVertexHashTable = new VertexHashTable( 1 << Asura_Maths::GetFloorLog2(uOriginalNumVertices >> 1) );
	if( !m_pxVertexHashTable ) { return false; }


	// First, we need to count the number of elements used in the skin, and build a remapping table	
	m_uNumElements = 0;

	for( int iPoly = 0; iPoly < m_pxInputMesh->NumberOfPolygons; ++iPoly )
	{
		int iElement = 0;
#if defined(ASURA_SUPPORT_SKIN_ELEMENTS)
		if (m_pxInputMesh->PolygonFlagDataValid)
		{
			iElement = Asura_Project_Skin_Elements::UnpackElementIDFromPolyFlags(m_pxInputMesh->PolygonFlagDataPtr[iPoly].Flags);
		}
#else if defined(ASURA_SUPPORT_STRIP_BY_ID)
		if (!m_bIgnoreIDFromPolyFlags)
		{
			iElement = Asura_Project_Skin_Elements::UnpackElementIDFromPolyFlags(m_pxInputMesh->GetProjectPolygonData(iPoly));
		}
#endif

		// Is this element included?
		bool bFound = false;

		for(u_int u=0;u<m_uNumElements;u++)
		{
			if (m_piElementMap[u] == iElement)
			{
				bFound=true;
			}
		}

		if (!bFound)
		{
			// Resize the element map			
			int *piOldMap = m_piElementMap;
			m_piElementMap = new int[m_uNumElements+1];

			if (piOldMap)
			{
				memcpy(m_piElementMap, piOldMap, sizeof(int)*m_uNumElements);				
				delete [] piOldMap;
			}

			m_piElementMap[m_uNumElements]=iElement;
			m_uNumElements++;
		}
	}
	

	m_pxElements = new Tools_SmoothSkin_Element[m_uNumElements];
	if( !m_pxElements ) { return false; }

	for (uint32 uElement = 0; uElement<m_uNumElements; uElement++)
	{
		// Worst-case tri count, as we don't know the actual count here
		m_pxElements[uElement].m_uNumTris = 0; 		

		m_pxElements[uElement].m_pusIndices = new u_short[3 * m_uNumTris];
		if( !m_pxElements[uElement].m_pusIndices ) { return false; }

		m_pxElements[uElement].m_puNumTrisPerMaterial = new u_int[m_uNumMaterials];
		if( !m_pxElements[uElement].m_puNumTrisPerMaterial ) { return false; }
		memset( m_pxElements[uElement].m_puNumTrisPerMaterial, 0, m_uNumMaterials * sizeof(u_int) );
	}

	//Initialise tangent and binormal calculation
	//RAR 09/07/2007
	Tools_Mesh_TangentSpace xMeshTangentSpace;
	if( m_bIncludeNormalMapData )
	{
		xMeshTangentSpace.Initialise(m_pxInputMesh);
	}


	for(uint32 uElement = 0; uElement<m_uNumElements; uElement++)
	{	
		u_int uUsedIndices = 0;

		for( int iPoly = 0; iPoly < m_pxInputMesh->NumberOfPolygons; ++iPoly )
		{
			if( !ShouldExportPoly(m_pxInputMesh, iPoly))
			{
				continue; 
			}

#if defined(ASURA_SUPPORT_SKIN_ELEMENTS)
			// Determine which element this poly belongs in			
			if (m_pxInputMesh->PolygonFlagDataValid)
			{
				int iElementID = Asura_Project_Skin_Elements::UnpackElementIDFromPolyFlags(m_pxInputMesh->PolygonFlagDataPtr[iPoly].Flags);
#else if defined(ASURA_SUPPORT_STRIP_BY_ID)
			{
				int iElementID = 0;
				if (!m_bIgnoreIDFromPolyFlags)
				{
					iElementID = Asura_Project_Skin_Elements::UnpackElementIDFromPolyFlags(m_pxInputMesh->GetProjectPolygonData(iPoly));
				}
#endif

				// If it's not in the current element, remove it
				if (iElementID != m_piElementMap[uElement])
				{
					continue;
				}
			}


#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			Tools_SmoothSkin_Material* pxMaterial = 0;
			for(u_int iMat=0; iMat<m_uNumMaterials; iMat++)
			{
				if(m_pxMaterials[iMat].m_auMaterialResponseHashID[0] == m_pxInputMesh->TextureDataPtr[ iPoly ].m_auMaterialResponseHashID[0] &&
					m_pxMaterials[iMat].m_auMaterialResponseHashID[1] == m_pxInputMesh->TextureDataPtr[ iPoly ].m_auMaterialResponseHashID[1])
				{
					pxMaterial = &m_pxMaterials[iMat];
					break;
				}
			}


#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			int iMaterialID = m_pxInputMesh->TextureDataPtr[iPoly].m_iMaterialID;
			if( ASURA_MATERIAL_NONE != iMaterialID ) { iMaterialID -= ASURA_MATERIAL_FIRSTID; }

			Tools_SmoothSkin_Material* pxMaterial = m_ppxMaterialLookupTable[iMaterialID];

#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			// Ignore polys that we need to remove due to the skin using
			// multiple textures.
			if( !pxMaterial ) { continue; }
	
			const u_int uMaterialIdx = CalculateMaterialIdx( pxMaterial );
			const bool  bIsQuad      = m_pxInputMesh->PolygonDataPtr[iPoly].IsQuad();
			const u_int uTrisInPoly  = ( bIsQuad ? 2 : 1 );

			bool bDoubleSided = ShouldPolyBeDoubleSided( iPoly );


			//If double sided we will need to do 2 passes of adding the vertices , with
			//normals and vertex ordering being reversed for the second side of the polygons
			//RAR 30/01/2004
			for( int iMirror=0;iMirror < (bDoubleSided ? 2 : 1);iMirror++)
			{
				m_pxElements[uElement].m_puNumTrisPerMaterial[uMaterialIdx] += uTrisInPoly;

				for( int iVtx = 0; iVtx < m_pxInputMesh->PolygonDataPtr[iPoly].NumberOfVertices(); ++iVtx )
				{
					Tools_SmoothSkin_Vertex& xVertex = m_pxVertices[m_uNumVertices];

					int iVtxToRead = iVtx;
					if(iMirror)
					{
						//Swap verts 0 and 2 when exporting mirrored polgons.
						//This should cause triangles to be exported as 2,1,0 and quads as 2,1,0 and 0,3,2 , resulting
						//in both being the reversed versions of the non-mirrored polygons.
						//RAR 06/03/2007
						if(iVtxToRead == 0) iVtxToRead = 2;
						else if(iVtxToRead == 2) iVtxToRead = 0;
					}

					// Prepare vertex data.
					{
						const int iMeshVertexID = m_pxInputMesh->PolygonDataPtr[iPoly].VertexID[iVtxToRead];

						Asura_Vector_3 xNormal = m_pxInputMesh->NormalsDataPtr[iPoly].VertexNormal[iVtxToRead];
						if( iMirror )
						{
							xNormal *= -1.0f;
						}

						Asura_Vector_3 xTangent, xBinormal;
						
						if( m_bIncludeNormalMapData )
						{
							xMeshTangentSpace.GetTangentAndBinormal(iPoly, iMeshVertexID, 0, xNormal, xTangent, xBinormal);							
						}


						// (SB - 14/03/07) Adding support for vertex colour via colour generation hook
						u_int uVertexColour = 0xffffffff;
						if (s_pfnColourGenerationFunction != NULL)
						{
							uVertexColour = s_pfnColourGenerationFunction(*m_pxInputSkin, *m_pxInputMesh, iPoly, iMeshVertexID);
						}
						

						const Asura_Mesh_TextureInfo* pxTexInfo = m_pxInputMesh->GetTextureInfoForExport(iPoly);
						
						xVertex.Set( uMaterialIdx,
							m_pxInputMesh->VertexDataPtr[iMeshVertexID].Position,
							xNormal,
							pxTexInfo->U[0][iVtxToRead],
							pxTexInfo->V[0][iVtxToRead],
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
							m_pxInputMesh->GetVertexFlags(iMeshVertexID),
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
							uElement,
#endif
							m_pxInputSkin->VertexData[iMeshVertexID].BoneID,
							m_pxInputSkin->VertexData[iMeshVertexID].Weight,
							uVertexColour );

						// (SB - 25/05/07) Allow project to modify the vertices if it needs to
						if (s_pfnVertexModificationFunction != NULL)
						{
							s_pfnVertexModificationFunction(xVertex);
						}


						xVertex.SetDualPass(Asura_Vector_2(pxTexInfo->U[1][iVtxToRead], pxTexInfo->V[1][iVtxToRead]), pxTexInfo->m_afAlpha[iVtxToRead]);


						if( m_bIncludeNormalMapData )
						{
							xVertex.SetTangentAndBinormal(xTangent, xBinormal);
						}
					}

					// Attempt to insert into the hash table. If this is a duplicate,
					// we'll get back a pointer to the matching vertex and we use
					// that instead.
					const Tools_SmoothSkin_Vertex* pxVtxToUse = m_pxVertexHashTable->Insert( &xVertex );
					if( !pxVtxToUse ) { return false; }

					// If it wasn't a duplicate, update count.
					if( &xVertex == pxVtxToUse ) { ++m_uNumVertices; }

					// Record index.
					{
						u_int uIdx = CalculateVertexIdx( pxVtxToUse );
						Asura_Assert( uIdx < 65536, "Too many distinct vertices in this skin." );
						m_pxElements[uElement].m_pusIndices[uUsedIndices++] = static_cast<u_short>( uIdx );
					}
				}


				// Inc the triangle count
				m_pxElements[uElement].m_uNumTris ++;

				// If this is a quad, we need to fix up the output indices. Currently
				// the end of the array, representing the current poly, is {0,1,2,3}.
				// We need {0,1,2,2,3,0} instead.
				if( bIsQuad )
				{
					m_pxElements[uElement].m_uNumTris ++;
					m_pxElements[uElement].m_pusIndices[uUsedIndices    ] = m_pxElements[uElement].m_pusIndices[uUsedIndices - 1];
					m_pxElements[uElement].m_pusIndices[uUsedIndices - 1] = m_pxElements[uElement].m_pusIndices[uUsedIndices - 2];
					m_pxElements[uElement].m_pusIndices[uUsedIndices + 1] = m_pxElements[uElement].m_pusIndices[uUsedIndices - 4];
					uUsedIndices += 2;					
				}

				// Analyse the tri(s) we've just written and remove them again if
				// they're degenerate.
				u_int uCheck = uUsedIndices - uTrisInPoly * 3;
				while( uCheck < uUsedIndices )
				{
					if( IsTriDegenerate(&m_pxElements[uElement].m_pusIndices[uCheck]) )
					{
						// Move second tri of quad down if necessary.
						if( uCheck + 3 < uUsedIndices )
						{
							Asura_Assert( bIsQuad, "" );
							memcpy( &m_pxElements[uElement].m_pusIndices[uCheck + 0], &m_pxElements[uElement].m_pusIndices[uCheck + 3], 3 * sizeof(u_int) );
						}

						uUsedIndices -= 3;

						// (MH - 26.07.2004) correct tri counts for removed degenerates
						m_pxElements[uElement].m_uNumTris -= 1;
						m_pxElements[uElement].m_puNumTrisPerMaterial[uMaterialIdx] -= 1;
					}
					else
					{
						uCheck += 3;
					}
				}
			}
		}

#if _DEBUG
		Asura_Assert( uUsedIndices == 3 * m_pxElements[uElement].m_uNumTris, "" );
#endif
	}

#if _DEBUG
	{
		u_int uTotal = 0;
		u_int uOrigTotal = 0;
		for( u_int uElement = 0; uElement < m_uNumElements; ++uElement )
		{	
			for( u_int u = 0; u < m_uNumMaterials; ++u )
			{
				uTotal += m_pxElements[uElement].m_puNumTrisPerMaterial[u]; 
			}

			uOrigTotal+=m_pxElements[uElement].m_uNumTris;
		}		
		
		Asura_Assert( uTotal == uOrigTotal, "" );		
	}
#endif	



	Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "    Found %u distinct vertices.", m_uNumVertices );
	Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "    %u tris.", m_uNumTris );

	return true;
}

#else

bool Tools_SmoothSkin_Processor::ProcessIntoElements( void )
{
    // Allocate space for the maximum number of vertices we can create. We
    // should never actually reach this but it's quicker to allocate a block
    // up front.
    u_int uOriginalNumVertices = 3 * m_uNumTris;
    m_pxVertices = new Tools_SmoothSkin_Vertex[uOriginalNumVertices];
    m_uNumVertices = 0;

    // Allocate hash table.
    m_pxVertexHashTable = new VertexHashTable( 1 << Asura_Maths::GetFloorLog2(uOriginalNumVertices >> 1) );
    if( !m_pxVertexHashTable ) { return false; }

    // (MP - 16.04.2004) TODO: Proper element splitting code, if/when it's
    // necessary.
    // ... For now, just create a single element containing everything. Final
    // splitting code would be pretty similar, except obviously it would
    // distribute the polygons to the relevant elements.

    m_uNumElements = 1;
    m_pxElements = new Tools_SmoothSkin_Element[m_uNumElements];
    if( !m_pxElements ) { return false; }

    m_pxElements[0].m_uNumTris = m_uNumTris;

    m_pxElements[0].m_pusIndices = new u_short[3 * m_pxElements[0].m_uNumTris];
    if( !m_pxElements[0].m_pusIndices ) { return false; }

    m_pxElements[0].m_puNumTrisPerMaterial = new u_int[m_uNumMaterials];
    if( !m_pxElements[0].m_puNumTrisPerMaterial ) { return false; }
    memset( m_pxElements[0].m_puNumTrisPerMaterial, 0, m_uNumMaterials * sizeof(u_int) );


    u_int uUsedIndices = 0;
    for( int iPoly = 0; iPoly < m_pxInputMesh->NumberOfPolygons; ++iPoly )
    {
        if( !ShouldExportPoly(m_pxInputMesh, iPoly) ) { continue; }

		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			Tools_SmoothSkin_Material* pxMaterial = 0;
			for(u_int iMat=0; iMat<m_uNumMaterials; iMat++)
			{
				if(m_pxMaterials[iMat].m_auMaterialResponseHashID[0] == m_pxInputMesh->TextureDataPtr[ iPoly ].m_auMaterialResponseHashID[0] &&
					m_pxMaterials[iMat].m_auMaterialResponseHashID[1] == m_pxInputMesh->TextureDataPtr[ iPoly ].m_auMaterialResponseHashID[1])
				{
					pxMaterial = &m_pxMaterials[iMat];
					break;
				}
			}


		#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			int iMaterialID = m_pxInputMesh->TextureDataPtr[iPoly].m_iMaterialID;
			if( ASURA_MATERIAL_NONE != iMaterialID ) { iMaterialID -= ASURA_MATERIAL_FIRSTID; }

			Tools_SmoothSkin_Material* pxMaterial = m_ppxMaterialLookupTable[iMaterialID];

		#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

        // Ignore polys that we need to remove due to the skin using
        // multiple textures.
        if( !pxMaterial ) { continue; }

        const u_int uMaterialIdx = CalculateMaterialIdx( pxMaterial );
        const bool  bIsQuad      = m_pxInputMesh->PolygonDataPtr[iPoly].IsQuad();
        const u_int uTrisInPoly  = ( bIsQuad ? 2 : 1 );

        bool bDoubleSided = ShouldPolyBeDoubleSided( iPoly );

        // Calculate tangent and binormal for the poly.
        Asura_Vector_3 xTangent, xBinormal;
        if( m_bIncludeNormalMapData )
        {
            Tools_TangentSpace::CalculateTangentAndBinormalForPoly( m_pxInputMesh, iPoly, 0,
                                                                    xTangent, xBinormal );
        }

        //If double sided we will need to do 2 passes of adding the vertices , with
        //normals and vertex ordering being reversed for the second side of the polygons
        //RAR 30/01/2004
        for( int iMirror=0;iMirror < (bDoubleSided ? 2 : 1);iMirror++)
        {
            m_pxElements[0].m_puNumTrisPerMaterial[uMaterialIdx] += uTrisInPoly;

            for( int iVtx = 0; iVtx < m_pxInputMesh->PolygonDataPtr[iPoly].NumberOfVertices(); ++iVtx )
            {
                Tools_SmoothSkin_Vertex& xVertex = m_pxVertices[m_uNumVertices];

                int iVtxToRead = iVtx;
				if(iMirror)
				{
					//Swap verts 0 and 2 when exporting mirrored polgons.
					//This should cause triangles to be exported as 2,1,0 and quads as 2,1,0 and 0,3,2 , resulting
					//in both being the reversed versions of the non-mirrored polygons.
					//RAR 06/03/2007
					if(iVtxToRead == 0) iVtxToRead = 2;
					else if(iVtxToRead == 2) iVtxToRead = 0;
				}

                // Prepare vertex data.
                {
                    const int iMeshVertexID = m_pxInputMesh->PolygonDataPtr[iPoly].VertexID[iVtxToRead];

                    Asura_Vector_3 xNormal = m_pxInputMesh->NormalsDataPtr[iPoly].VertexNormal[iVtxToRead];
                    if( iMirror )
                    {
                        xNormal *= -1.0f;
                    }

					const Asura_Mesh_TextureInfo* pxTexInfo = m_pxInputMesh->GetTextureInfoForExport(iPoly);
                    
                    xVertex.Set( uMaterialIdx,
                                 m_pxInputMesh->VertexDataPtr[iMeshVertexID].Position,
                                 xNormal,
                                 pxTexInfo->U[0][iVtxToRead],
                                 pxTexInfo->V[0][iVtxToRead],
                                 m_pxInputSkin->VertexData[iMeshVertexID].BoneID,
                                 m_pxInputSkin->VertexData[iMeshVertexID].Weight );
					xVertex.SetDualPass(Asura_Vector_2(pxTexInfo->U[1][iVtxToRead], pxTexInfo->V[1][iVtxToRead]), pxTexInfo->m_afAlpha[iVtxToRead]);

                    if( m_bIncludeNormalMapData )
                    {
                        xVertex.SetTangentAndBinormal( Tools_TangentSpace::ClampPerpendicularToNormal(xNormal, xTangent),
                                                       Tools_TangentSpace::ClampPerpendicularToNormal(xNormal, xBinormal) );
                    }
                }

                // Attempt to insert into the hash table. If this is a duplicate,
                // we'll get back a pointer to the matching vertex and we use
                // that instead.
                const Tools_SmoothSkin_Vertex* pxVtxToUse = m_pxVertexHashTable->Insert( &xVertex );
                if( !pxVtxToUse ) { return false; }

                // If it wasn't a duplicate, update count.
                if( &xVertex == pxVtxToUse ) { ++m_uNumVertices; }

                // Record index.
                {
                    u_int uIdx = CalculateVertexIdx( pxVtxToUse );
                    Asura_Assert( uIdx < 65536, "Too many distinct vertices in this skin." );
                    m_pxElements[0].m_pusIndices[uUsedIndices++] = static_cast<u_short>( uIdx );
                }
            }


            // If this is a quad, we need to fix up the output indices. Currently
            // the end of the array, representing the current poly, is {0,1,2,3}.
            // We need {0,1,2,2,3,0} instead.
            if( bIsQuad )
            {
                m_pxElements[0].m_pusIndices[uUsedIndices    ] = m_pxElements[0].m_pusIndices[uUsedIndices - 1];
                m_pxElements[0].m_pusIndices[uUsedIndices - 1] = m_pxElements[0].m_pusIndices[uUsedIndices - 2];
                m_pxElements[0].m_pusIndices[uUsedIndices + 1] = m_pxElements[0].m_pusIndices[uUsedIndices - 4];
                uUsedIndices += 2;
            }

            // Analyse the tri(s) we've just written and remove them again if
            // they're degenerate.
            u_int uCheck = uUsedIndices - uTrisInPoly * 3;
            while( uCheck < uUsedIndices )
            {
                if( IsTriDegenerate(&m_pxElements[0].m_pusIndices[uCheck]) )
                {
                    // Move second tri of quad down if necessary.
                    if( uCheck + 3 < uUsedIndices )
                    {
                        Asura_Assert( bIsQuad, "" );
                        memcpy( &m_pxElements[0].m_pusIndices[uCheck + 0], &m_pxElements[0].m_pusIndices[uCheck + 3], 3 * sizeof(u_int) );
                    }

                    uUsedIndices -= 3;

                    // (MH - 26.07.2004) correct tri counts for removed degenerates
                    m_pxElements[0].m_uNumTris -= 1;
                    m_pxElements[0].m_puNumTrisPerMaterial[uMaterialIdx] -= 1;
                }
                else
                {
                    uCheck += 3;
                }
            }
        }
    }

#if _DEBUG
    {
        u_int uTotal = 0;
        for( u_int u = 0; u < m_uNumMaterials; ++u ) {  uTotal += m_pxElements[0].m_puNumTrisPerMaterial[u]; }
        Asura_Assert( uTotal == m_pxElements[0].m_uNumTris, "" );

        Asura_Assert( uUsedIndices == 3 * m_pxElements[0].m_uNumTris, "" );
    }
#endif

    Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "    Found %u distinct vertices.", m_uNumVertices );
    Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "    %u tris.", m_uNumTris );

    return true;
}

#endif

bool Tools_SmoothSkin_Processor::ProcessIntoMaterials( void )
{
    // Allocate sub-elements and their index arrays.
    for( u_int uMaterial = 0; uMaterial < m_uNumMaterials; ++uMaterial )
    {
        Tools_SmoothSkin_Material& xMaterial = m_pxMaterials[uMaterial];

        xMaterial.m_pxSubElements = new Tools_SmoothSkin_SubElement[m_uNumElements];
        for( u_int uElement = 0; uElement < m_uNumElements; ++uElement )
        {
            const Tools_SmoothSkin_Element& xElement = m_pxElements[uElement];
            Tools_SmoothSkin_SubElement& xSubElement = xMaterial.m_pxSubElements[uElement];
            
            if( xElement.m_puNumTrisPerMaterial[uMaterial] )
            {
                xSubElement.m_pusIndices = new u_short[3 * xElement.m_puNumTrisPerMaterial[uMaterial]];
                if( !xSubElement.m_pusIndices ) { return false; }
            }
        }
    }

    // Copy indices.
    for( u_int uElement = 0; uElement < m_uNumElements; ++uElement )
    {
        const Tools_SmoothSkin_Element& xElement = m_pxElements[uElement];

        for( u_int uTri = 0; uTri < xElement.m_uNumTris; ++uTri )
        {
            // Obviously a tri is assumed to have the same materials on all
            // three vertices.
            const u_int uMaterialIdx = m_pxVertices[xElement.m_pusIndices[uTri * 3]].GetMaterialIdx();

            Tools_SmoothSkin_Material&   xMaterial   = m_pxMaterials[uMaterialIdx];
            Tools_SmoothSkin_SubElement& xSubElement = xMaterial.m_pxSubElements[uElement];
            
            u_int uVertOut = 3 * xSubElement.m_uNumTris++;

            Asura_Assert( xSubElement.m_pusIndices, "" );
            xSubElement.m_pusIndices[uVertOut + 0] = xElement.m_pusIndices[uTri * 3 + 0];
            xSubElement.m_pusIndices[uVertOut + 1] = xElement.m_pusIndices[uTri * 3 + 1];
            xSubElement.m_pusIndices[uVertOut + 2] = xElement.m_pusIndices[uTri * 3 + 2];
        }

#if _DEBUG
        for( u_int uMaterial = 0; uMaterial < m_uNumMaterials; ++uMaterial )
        {
            Asura_Assert( m_pxMaterials[uMaterial].m_pxSubElements[uElement].m_uNumTris == xElement.m_puNumTrisPerMaterial[uMaterial], "" );
        }
#endif
    }

    return true;
}


bool Tools_SmoothSkin_Processor::ShouldPolyBeDoubleSided( int iPolyIdx ) const
{
    //check for double sided flag being disabled
    if( m_bIgnoreDoubleSidedFlag ) { return false; }

    //check the mesh in that case to see if the polygon is flagged as such
    if( m_pxInputMesh->PolygonFlagDataValid )
    {
        if( m_pxInputMesh->PolygonFlagDataPtr[iPolyIdx].Flags & (AsuraPolygonFlag_DoubleSided) )
        {
            return true;
        }
    }

    return false;
}

//
// Tools_SmoothSkin_Element.
//

Tools_SmoothSkin_Element::Tools_SmoothSkin_Element( void )
    : m_uNumTris( 0 )
    , m_pusIndices( NULL )
    , m_puNumTrisPerMaterial( NULL )
{
}


Tools_SmoothSkin_Element::~Tools_SmoothSkin_Element( void )
{
    delete[] m_puNumTrisPerMaterial;
    delete[] m_pusIndices;
}

