//////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_Processor.h
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___TOOLS_SMOOTHSKIN_PROCESSOR_H___
#define ___TOOLS_SMOOTHSKIN_PROCESSOR_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Engine independent.
#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	#include "Core/Asura_MaterialFlags.h"
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

const float TOOLS_SMOOTHSKIN_PROCESSOR_UV_EPSILON_SQD = 0.005f * 0.005f;

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

template<typename T> class Tools_HashTable;
class Asura_Hierarchy_Skin;
class Asura_Mesh;

class Tools_SmoothSkin_Element;
class Tools_SmoothSkin_Material;
class Tools_SmoothSkin_Processor;
class Tools_SmoothSkin_SubElement;
class Tools_SmoothSkin_Vertex;
class Tools_SmoothSkin_BoneMap;

// C L A S S E S /////////////////////////////////////////////////////////////

// This is the main class here, providing functionality for processing
// Asura_Hierarchy_Skin instances into simple lists, where the skin mesh is
// split by material and then element.
class Tools_SmoothSkin_Processor
{

    public:

		// T Y P E D E F S ///////////////////////////////////////////////////

		typedef u_int (ColourGenerationHook)(const Asura_Hierarchy_Skin& xSkin, const Asura_Mesh& xMesh, const int iPolyID, const int iVertexID);
		typedef void (VertexModificationHook)(Tools_SmoothSkin_Vertex& xVertex);

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        Tools_SmoothSkin_Processor( void );
        ~Tools_SmoothSkin_Processor( void );

        // M E T H O D S /////////////////////////////////////////////////////

        // Set various options. These should be done before calling
        // Initialise(), if required.
        void SetIgnoreDoubleSidedFlag( bool bIgnore ) { m_bIgnoreDoubleSidedFlag = bIgnore; }
        void SetIncludeNormalMapData( bool bInclude ) { m_bIncludeNormalMapData = bInclude; }
        void SetIgnoreIDFromPolyFlags( bool bIgnore ) { m_bIgnoreIDFromPolyFlags = bIgnore; }

        // Initialisation.
        bool Initialise( Asura_Hierarchy_Skin* pxSkin );

        // Query methods.
        inline const Tools_SmoothSkin_Element& GetElement( u_int uIdx ) const;
        inline const Tools_SmoothSkin_Material& GetMaterial( u_int uIdx ) const;
        inline const Tools_SmoothSkin_Vertex& GetVertex( u_int uIdx ) const;
        const Tools_SmoothSkin_Vertex* GetVertices( void ) const { return m_pxVertices; }
        const Asura_Hierarchy_Skin* GetInputSkin( void ) const { return m_pxInputSkin; }
        u_int GetNumTris( void ) const { return m_uNumTris; }
        u_int GetNumElements( void ) const { return m_uNumElements; }
        u_int GetNumMaterials( void ) const { return m_uNumMaterials; }
        u_int GetNumVertices( void ) const { return m_uNumVertices; }
		int   GetElementID( u_int uElement ) const;
		bool ShouldIncludeNormalMapData( void ) const { return m_bIncludeNormalMapData; }

		// Static methods (called by project tools startup).
		static void SetAllowMultipleTextures( bool bAllow ) { s_bAllowMultipleTextures = bAllow; }

		static void SetColourGenerationCallback(ColourGenerationHook* const pfnColourGenerationFunction = NULL) { s_pfnColourGenerationFunction = pfnColourGenerationFunction; }
		static void SetVertexModificationCallback(VertexModificationHook* const pfnVertexModificationFunction = NULL) { s_pfnVertexModificationFunction = pfnVertexModificationFunction; }

		void CreateBoneRemapTables(u_int uMaxNumBones);

		u_int GetNumBoneMaps() const {return m_uNumBoneMapTables;}
		u_int GetBoneMapIndex(u_int uMaterial, u_int uElement) const {return m_puBoneMapIndexes ? m_puBoneMapIndexes[uMaterial * m_uNumElements + uElement] : 0xFFFFFFFF;}
		const Tools_SmoothSkin_BoneMap *GetBoneMap(u_int uBoneMapIndex) const {return m_ppxBoneMapTables && uBoneMapIndex < m_uNumBoneMapTables ? m_ppxBoneMapTables[uBoneMapIndex] : NULL;}
		u_int GetVertexBoneMapIndex(u_int uVertex) const {return m_puVertexBoneMapIndexes ? m_puVertexBoneMapIndexes[uVertex] : 0xFFFFFFFF;}


    private:

        // T Y P E S /////////////////////////////////////////////////////////

        typedef Tools_HashTable<Tools_SmoothSkin_Vertex> VertexHashTable;

        // M E T H O D S /////////////////////////////////////////////////////

        u_int CalculateMaterialIdx( const Tools_SmoothSkin_Material* pxMaterial ) const;
        u_int CalculateVertexIdx( const Tools_SmoothSkin_Vertex* pxVertex ) const;

        bool CreateMaterialList( void );
        bool ProcessIntoElements( void );
        bool ProcessIntoMaterials( void );

        void GatherStats( void ) const;

        bool ShouldPolyBeDoubleSided( int iPolyIdx ) const;

        void CalculateTangentAndBinormalForPoly( int iPolyIdx, Asura_Vector_3* pxTangent, Asura_Vector_3* pxBinormal ) const;

        // S T A T E /////////////////////////////////////////////////////////

        // Options.
        bool m_bIgnoreDoubleSidedFlag;
        bool m_bIncludeNormalMapData;
        bool m_bIgnoreIDFromPolyFlags;

        // Input data.
        const Asura_Hierarchy_Skin* m_pxInputSkin;
        const Asura_Mesh*           m_pxInputMesh;

        // Number of triangles in skin. This will not necessarily match up
        // with the count in the input data since we filter out ones we're
        // not interested in (don't render, don't export, ...)
        u_int                       m_uNumTris;

        // Details of all materials used by this skin. The material array
        // is sorted by material ID in ascending order.
        u_int                       m_uNumMaterials;
        Tools_SmoothSkin_Material*  m_pxMaterials;
		#if !defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)
        Tools_SmoothSkin_Material** m_ppxMaterialLookupTable; // Material ID -> internal material ptr.
		#endif

        u_int                       m_uNumElements;
        Tools_SmoothSkin_Element*   m_pxElements;
		int						   *m_piElementMap;

        // This is the master vertex list, containing only distinct vertices.
        // All the other classes index into this list.
        u_int                       m_uNumVertices;
        Tools_SmoothSkin_Vertex*    m_pxVertices;

        // To accelerate queries and the elimination of duplicates, we
        // maintain a hash table as well.
        VertexHashTable*            m_pxVertexHashTable;

		// To allow us to have skins larger than the maximum number of
		// bones we can render at once we have bone map tables;
		// How many bone map tables do we have
		u_int						m_uNumBoneMapTables;
		// an array of pointers to each table
		Tools_SmoothSkin_BoneMap**	m_ppxBoneMapTables;
		// an array specifiying which bone map each strip uses
		u_int*						m_puBoneMapIndexes;
		// an array specifying which bone map each vertex uses
		u_int*						m_puVertexBoneMapIndexes;


		// static class data
		static bool s_bAllowMultipleTextures;

		static ColourGenerationHook* s_pfnColourGenerationFunction;
		static VertexModificationHook* s_pfnVertexModificationFunction;

};
// End class Tools_SmoothSkin_Processor.

// This class manages a list of all the sub-elements for a given material.
// All materials have one sub-element for each element in the parent skin,
// though some will be empty if no polygons in that element use the
// corresponding material.
class Tools_SmoothSkin_Material
{
    
    public:
        
        // M E T H O D S /////////////////////////////////////////////////////

        // Query methods.
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		Asura_Hash_ID GetMaterialResponseHashID(int iIndex) const { return m_auMaterialResponseHashID[iIndex]; }

		// Dummies for old material ID. All the other platform exporters won't compile without
		int GetMaterialID() const { return ASURA_MATERIAL_NONE; }
		int GetMaterialIDWithinSkin() const { return 0; }
#else
        int GetMaterialID( void ) const { return m_iMaterialID; }
        int GetMaterialIDWithinSkin( void ) const { return m_iMaterialIDWithinSkin; }
#endif

        inline const Tools_SmoothSkin_SubElement& GetSubElement( u_int uIdx ) const;

    private:
        
        // F R I E N D S /////////////////////////////////////////////////////

        friend Tools_SmoothSkin_Processor;

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        Tools_SmoothSkin_Material( void );
        ~Tools_SmoothSkin_Material( void );
        
        // S T A T E /////////////////////////////////////////////////////////
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		Asura_Hash_ID                 m_auMaterialResponseHashID[2];
#else
        int                           m_iMaterialID;
        int                           m_iMaterialIDWithinSkin;
#endif

        Tools_SmoothSkin_SubElement*  m_pxSubElements;
        
};
// End class Tools_SmoothSkin_Material.

class Tools_SmoothSkin_BoneMap
{
public:
	u_int GetNumEntries() const {return m_uNumEntries;}
	u_int GetBone(u_int uIndex) const {return m_puBoneMap[uIndex];}
	u_int GetEntryForBone(u_int uIndex) const {return m_puInvBoneMap[uIndex];}

	// just needed for serializing
	const u_int *GetMap() const {return m_puBoneMap;}
private:
	// F R I E N D S /////////////////////////////////////////////////////

	friend Tools_SmoothSkin_Processor;
	Tools_SmoothSkin_BoneMap(u_int uNumEntries, u_int uNumBones) : m_uNumEntries(uNumEntries), m_puBoneMap(new u_int[uNumEntries]), m_puInvBoneMap(new u_int [uNumBones]) {} 
	~Tools_SmoothSkin_BoneMap(){delete [] m_puBoneMap; delete [] m_puInvBoneMap; }

	u_int    m_uNumEntries;
	u_int*   m_puBoneMap;
	u_int*   m_puInvBoneMap;
};
// End class Tools_SmoothSkin_BoneMap.

// This is just temporary until element-splitting code is properly integrated.
class Tools_SmoothSkin_Element
{

    public:

        // M E T H O D S /////////////////////////////////////////////////////

        u_int GetNumTris( void ) const { return m_uNumTris; }

    private:
        
        // F R I E N D S /////////////////////////////////////////////////////

        friend Tools_SmoothSkin_Processor;

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        Tools_SmoothSkin_Element( void );
        ~Tools_SmoothSkin_Element( void );

        // S T A T E /////////////////////////////////////////////////////////

        u_int    m_uNumTris;
        u_short* m_pusIndices;           // 3 per tri, indexing into processor's master vertex list.
        u_int*   m_puNumTrisPerMaterial;

};
// End class Tools_SmoothSkin_Element.

// A sub-element contains all the polygons from the element using a
// particular material. For each element, there will be exactly one sub-
// element per material. The mesh data is stored in plain tri-lists, without
// any platform-specific modifications.
class Tools_SmoothSkin_SubElement
{
									
    public:

        // M E T H O D S /////////////////////////////////////////////////////

        u_int GetNumTris( void ) const { return m_uNumTris; }
        u_short GetIndex( u_int uTri, u_int uVert ) const { return m_pusIndices[uTri * 3 + uVert]; }
        const u_short* GetIndices( void ) const { return m_pusIndices; }

    private:
        
        // F R I E N D S /////////////////////////////////////////////////////

        friend Tools_SmoothSkin_Material;
        friend Tools_SmoothSkin_Processor;

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        Tools_SmoothSkin_SubElement( void );
        ~Tools_SmoothSkin_SubElement( void );

        // S T A T E /////////////////////////////////////////////////////////

        u_int    m_uNumTris;
        u_short* m_pusIndices; // 3 per tri, indexing into processor's master vertex list. NULL iff m_uNumTris is 0.
        
};
// End class Tools_SmoothSkin_SubElement.

// Simple platform-independent smooth skin vertex.
class Tools_SmoothSkin_Vertex
{

    public:

        // C O N S T R U C T O R S   /   D E S T R U C T O R S # /////////////

        Tools_SmoothSkin_Vertex( void ) 
        : m_uHash( ASURA_HASH_ID_UNSET ) 
        , m_xTangent(0.0f,0.0f,0.0f)
        , m_xBinormal(0.0f,0.0f,0.0f)
        { }

        // O P E R A T O R S /////////////////////////////////////////////////
    
        bool operator==( const Tools_SmoothSkin_Vertex& xOther ) const
        {
            // Quick rejection for data that just requires exact comparisons.
            if( m_uHash != xOther.m_uHash ) { return false; }

            // Special cases.
            {
                if(!NormalApproximatelyEquals(m_xNormal,xOther.m_xNormal)) return false;
				if(!NormalApproximatelyEquals(m_xTangent,xOther.m_xTangent)) return false;
				if(!NormalApproximatelyEquals(m_xBinormal,xOther.m_xBinormal)) return false;
								

                const Asura_Vector_2 xUVDelta( xOther.m_xUV - m_xUV );
                if( xUVDelta.MagnitudeSquared() >= TOOLS_SMOOTHSKIN_PROCESSOR_UV_EPSILON_SQD ) { return false; }
            }

            // Hash will probably always be sufficient but we'd better check
            // everything else just in case.
            return(    m_xHashData.m_usMaterialIdx		== xOther.m_xHashData.m_usMaterialIdx
                    && m_xHashData.m_xPos				== xOther.m_xHashData.m_xPos
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
					&& m_xHashData.m_uFlags				== xOther.m_xHashData.m_uFlags
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
					&& m_xHashData.m_uElement			== xOther.m_xHashData.m_uElement
#endif
                    && m_xHashData.m_xBoneIndices.m_uAllIndices == xOther.m_xHashData.m_xBoneIndices.m_uAllIndices
                    && m_xHashData.m_afBoneWeights[0]	== xOther.m_xHashData.m_afBoneWeights[0]
                    && m_xHashData.m_afBoneWeights[1]	== xOther.m_xHashData.m_afBoneWeights[1]
                    && m_xHashData.m_afBoneWeights[2]	== xOther.m_xHashData.m_afBoneWeights[2]
                    && m_xHashData.m_afBoneWeights[3]	== xOther.m_xHashData.m_afBoneWeights[3]         ); 
        }

        // M E T H O D S /////////////////////////////////////////////////////

        // Set methods.
        void Set( u_int                 uMaterialIdx,
                  const Asura_Vector_3& xPos,
                  const Asura_Vector_3& xNormal,
                  float                 fU,
                  float                 fV,
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
				  u_int					uVertexFlags,
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
				  u_int					uElement,
#endif
				  const short           asBoneIndices[4],
				  const float           afBoneWeights[4],
				  const u_int			uRGBAColour = 0xffffffff );

        // Additional optional set methods, necessary depending on what
        // options are set in the parent Tools_SmoothSkin_Processor.
        void SetTangentAndBinormal( const Asura_Vector_3& xTangent, const Asura_Vector_3& xBinormal );
		void SetDualPass( const Asura_Vector_2& xUV2, float fAlpha );

        // Query methods. Some of this data will only be valid if the
        // appropriate option is turned on in the parent
        // Tools_SmoothSkin_Processor.
        Asura_Hash_ID         GetHashID( void )        const { return m_uHash; }
        u_int                 GetMaterialIdx( void )   const { return m_xHashData.m_usMaterialIdx; }
        const Asura_Vector_3& GetPos( void )           const { return m_xHashData.m_xPos; }
        const Asura_Vector_3& GetNormal( void )        const { return m_xNormal; }
        const Asura_Vector_2& GetTexCoords( void )     const { return m_xUV; }
		const Asura_Vector_2& GetTexCoords2( void )    const { return m_xUV; }
        float                 GetU( void )             const { return m_xUV.x; }
        float                 GetV( void )             const { return m_xUV.y; }
		float                 GetU2( void )            const { return m_xUV2.x; }
		float                 GetV2( void )            const { return m_xUV2.y; }
		float                 GetAlpha( void )         const { return m_fAlpha; }

		u_int                 GetNumWeightings( void ) const { return m_xHashData.m_usNumWeightings; }
        const u_char*         GetBoneIndices( void )   const { return m_xHashData.m_xBoneIndices.m_aucIndices; }
        const float*          GetBoneWeights( void )   const { return m_xHashData.m_afBoneWeights; }
        const Asura_Vector_3& GetTangent( void )       const { return m_xTangent; }
        const Asura_Vector_3& GetBinormal( void )      const { return m_xBinormal; }
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
		const u_int			  GetFlags( void )		   const { return m_xHashData.m_uFlags; }
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
		const u_int			  GetElement( void )		   const { return m_xHashData.m_uElement; }
#endif
		const u_int			  GetColour( void )		   const { return m_xHashData.m_uRGBAColour; }

		static bool NormalApproximatelyEquals(const Asura_Vector_3& xNorm1, const Asura_Vector_3& xNorm2)
		{
			const float fNrmTol = 0.996194698092f;   // +/- 5 (where this value is cos 5)

			if(xNorm1 * xNorm2 < fNrmTol)
			{
				//Failed first test (which assumes that they are both unit length vectors)
				//Allow equality however if the normals are both zero
				if(!(xNorm1.ApproximatelyEquals(Asura_Maths::ZeroVector3) && xNorm2.ApproximatelyEquals(Asura_Maths::ZeroVector3)))
				{
					return false;
				}
			}
			return true;
		}

    private:

        // T Y P E S /////////////////////////////////////////////////////////

        union BoneIndices
        {
            u_char m_aucIndices[4];
            u_int  m_uAllIndices;
        };

        // F R I E N D S /////////////////////////////////////////////////////

        friend Tools_SmoothSkin_Processor;

        // S T A T E /////////////////////////////////////////////////////////

        // Meta data.
        Asura_Hash_ID m_uHash;

        // Data that is not simply checked for exact equality when comparing
        // vertices.
        Asura_Vector_3 m_xNormal;
        Asura_Vector_2 m_xUV;
		Asura_Vector_2 m_xUV2;
		float m_fAlpha;

        // Data where we do just look for exact equality.
        struct HashData
        {
            Asura_Vector_3	m_xPos;
            u_short			m_usMaterialIdx;    // NB: This is an index into the internal list of materials, not the global material ID.
            u_short			m_usNumWeightings;
            BoneIndices		m_xBoneIndices;
            float			m_afBoneWeights[4];
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
			u_int			m_uFlags;
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
			u_int			m_uElement;
#endif
			u_int			m_uRGBAColour;
        }
        m_xHashData;

        // Data that is not checked for equality at all.
        // (MP - 09.05.2006) This behaviour follows what we were doing in
        // the Tools_CharacterProcessor_NormalMapped code. I'm not sure it's
        // the right thing to do though.
        Asura_Vector_3 m_xTangent;
        Asura_Vector_3 m_xBinormal;

};
// End class Tools_SmoothSkin_Vertex.

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Tools_SmoothSkin_Processor.
//

inline const Tools_SmoothSkin_Element& Tools_SmoothSkin_Processor::GetElement( u_int uIdx ) const { return m_pxElements[uIdx]; }
inline const Tools_SmoothSkin_Material& Tools_SmoothSkin_Processor::GetMaterial( u_int uIdx ) const { return m_pxMaterials[uIdx]; }
inline const Tools_SmoothSkin_Vertex& Tools_SmoothSkin_Processor::GetVertex( u_int uIdx ) const { return m_pxVertices[uIdx]; }

//
// Tools_SmoothSkin_Material.
//

inline const Tools_SmoothSkin_SubElement& Tools_SmoothSkin_Material::GetSubElement( u_int uIdx ) const { return m_pxSubElements[uIdx]; }

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___TOOLS_SMOOTHSKIN_PROCESSOR_H___
