//////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_Processor_PSP.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "MeshProcessing/Tools_SmoothSkin_Processor_PSP.h"

// Stdlib independent.
#include <memory.h>

// Engine independent.
#include "Animation/Asura_Hierarchy.h" // to get skin name
#include "Core/Asura_Mempool.h"


// Tools.
#include "MeshProcessing/Asura_TriStripper.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Tools_SmoothSkin_PSP_Processor.
//

Tools_SmoothSkin_PSP_Processor::Tools_SmoothSkin_PSP_Processor( void )
    : m_pxSrcProcessor( NULL )
    , m_uNumberOfGroups( 0 )
    , m_pxGroups( NULL )
{
}

bool Tools_SmoothSkin_PSP_Processor::Initialise( const Tools_SmoothSkin_Processor* pxSrcProcessor )
{
    Asura_Assert( pxSrcProcessor, "" );
    m_pxSrcProcessor = pxSrcProcessor;

    // At the moment this code assumes a single element in the source skin.
    Asura_Assert( 1 == m_pxSrcProcessor->GetNumElements(),  "" );

    bool bOK = false;

    const u_int uNumMaterials = m_pxSrcProcessor->GetNumMaterials();
    
    u_int uPolySubTotal = 0;

    Asura_TrackingHandlePool<Tools_SmoothSkin_PSP_BoneSet> xBoneSetPool(8);

    int** ppiPolygonGroups = new int*[uNumMaterials];
    if( !ppiPolygonGroups ) { goto CleanUp; }

    for( u_int u = 0; u < uNumMaterials; ++u ) { ppiPolygonGroups[u] = NULL; }

    m_uNumberOfGroups = 0;
    for( u_int uMaterial = 0; uMaterial < uNumMaterials; ++uMaterial )
    {
        const Tools_SmoothSkin_Material&   xMaterial   = m_pxSrcProcessor->GetMaterial( uMaterial );
        const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement( 0 );

        u_int uNumberOfTris = xSubElement.GetNumTris();

        int* piPolygonGroup = new int[uNumberOfTris];
        if( !piPolygonGroup ) { goto CleanUp; }

        ppiPolygonGroups[uMaterial] = piPolygonGroup;

        const int iInvalidGroup = -1;

        // flag all the polygons in the current element as 'not in a group'
        for (u_int u=0; u<uNumberOfTris; u++)
        {
            piPolygonGroup[u] = iInvalidGroup;
        }

        for (u_int uTri=0; uTri<uNumberOfTris; uTri++)
        {
            // find a polygon not in a group
            if( iInvalidGroup != piPolygonGroup[uTri] ) { continue; }

            // lets make a new group and declare this polygon part of it
            Asura_Handle iHandle = xBoneSetPool.Allocate();
            piPolygonGroup[uTri] = iHandle;

            Asura_Assert( iHandle - m_uNumberOfGroups == 0, "these should be in sync" );
            
            Tools_SmoothSkin_PSP_BoneSet& xBoneSet = *(xBoneSetPool.GetPointer(iHandle));

            // so, what's the bone set used by this polygon?
            bool bSuccess = xBoneSet.AddBonesOfTri( uTri, m_pxSrcProcessor, xSubElement );
        
            // if not successful we're in trouble - a polygon needs more than four different bones
			Asura_Assertf( bSuccess, ( "Polygon using more than 4 weights!\nSkin name:%s", pxSrcProcessor->GetInputSkin()->Name.GetString() ) );
        
			if (!bSuccess)
            {
				// This function is only really here to deal with 'broken' meshes.
				// N.B. Weightings for this bone will probably be wrong!
                xBoneSet.AddBonesOfTriUsingMostImportantWeightingsOnly( uTri, m_pxSrcProcessor, xSubElement );
			}

            xBoneSet.SetMaterialIndex( uMaterial );

            // scan through the remaining tris and add them to the group if we can
            for (u_int u = uTri+1; u < uNumberOfTris; u++)
            {
                if (iInvalidGroup != piPolygonGroup[u]) continue;

                if ( xBoneSet.AddBonesOfTri(u, m_pxSrcProcessor, xSubElement) )
                {
                    piPolygonGroup[u] = iHandle;
                }
            }

            uPolySubTotal += xBoneSet.GetNumberOfTris();
            m_uNumberOfGroups++;
        }
    }

    m_pxGroups = new Tools_SmoothSkin_PSP_Group[m_uNumberOfGroups];
    if( !m_pxGroups ) { goto CleanUp; }

    for( u_int uGroup = 0; uGroup < m_uNumberOfGroups; ++uGroup )
    {
        // initialise a new group
        Tools_SmoothSkin_PSP_Group& xGroup = m_pxGroups[uGroup];

        Tools_SmoothSkin_PSP_BoneSet* pxBoneSet = xBoneSetPool.GetPointer(uGroup);

        u_int uMaterialIdx = pxBoneSet->GetMaterialIndex();

        // allocate index space
        xGroup.m_uNumberOfIndices = pxBoneSet->GetNumberOfTris()*3;
        xGroup.m_puIndices = new u_int[xGroup.m_uNumberOfIndices];
        
        // allocate enough vertices for worst case scenario
        xGroup.m_pxVertices = new Tools_SmoothSkin_PSP_Vertex[xGroup.m_uNumberOfIndices];

        if( !xGroup.m_puIndices || !xGroup.m_pxVertices ) { goto CleanUp; }

        xGroup.m_iMaterial = uMaterialIdx;

        // get the four bones for this group.
        for (u_int u=0; u<4; u++)
        {
            xGroup.m_aucBoneIDs[u] = static_cast<u_char>(pxBoneSet->GetBone(u));
        }

		if (uGroup!=0) ReorderBonesToMinimiseRenderingStateChanges( xGroup, m_pxGroups[ uGroup - 1 ] );

        // we now want to extract polygons used by this group. Since this will very likely be a subset
        // of the original sub element, we'll need to construct a new vertex list and renumber
        // them as we go.
        u_int uNumberOfVertices = 0;
        u_int uNumberOfIndices = 0;

        const Tools_SmoothSkin_SubElement& xSubElement = m_pxSrcProcessor->GetMaterial(uMaterialIdx).GetSubElement(0);

        const u_int uNumberOfTris = xSubElement.GetNumTris();
        
//      const u_int uNumberOfVerticesInElement = m_pxTriList->GetElement(uTriListElement)->m_uNumberOfVertices;

        // initialise the lookup table we'll be using to keep track of used vertices
        int* piIndexLookup = new int[m_pxSrcProcessor->GetNumVertices()];
        if( !piIndexLookup ) { goto CleanUp; }
        memset( piIndexLookup, -1, m_pxSrcProcessor->GetNumVertices() * sizeof(int) );
        
        for (u_int uTri=0; uTri<uNumberOfTris; uTri++)
        {
            // search through for triangles in this element which are part of the group
            if (static_cast<u_int>(ppiPolygonGroups[uMaterialIdx][uTri]) != uGroup) continue;

            for (u_int u=0; u<3; u++)
            {
                const u_int uIndex = xSubElement.GetIndex( uTri, u );

                // if we haven't used this index buffer, add it to the lookup table
                // and copy across the vertex
                if( -1 == piIndexLookup[uIndex] )
                {
                    piIndexLookup[uIndex] = uNumberOfVertices;
                    xGroup.m_pxVertices[uNumberOfVertices].Set( m_pxSrcProcessor->GetVertex(uIndex), xGroup.m_aucBoneIDs );
                    uNumberOfVertices++;
                }
            
                xGroup.m_puIndices[uNumberOfIndices++] = piIndexLookup[uIndex];
            }
        }
        
        delete[] piIndexLookup;

        xGroup.m_uNumberOfVertices = uNumberOfVertices;

        Asura_Assert( uNumberOfIndices == xGroup.m_uNumberOfIndices, "number of indices does not match up" );

    }


	// stripify the groups
	for( u_int uGroup = 0; uGroup < m_uNumberOfGroups; ++uGroup )
    {
        Tools_SmoothSkin_PSP_Group& xGroup = m_pxGroups[uGroup];

		// damn, data is in u_ints, but stripifier takes u_shorts
		u_short* pusListIndices = new u_short[ xGroup.m_uNumberOfIndices ];
		for (u_int u=0; u<xGroup.m_uNumberOfIndices; ++u)
		{
			pusListIndices[u] = (u_short)(xGroup.m_puIndices[u]);
		}

        Tools_TriStripper::Initialise();

        u_short* pusStripIndices;
        u_int uNumStripIndices = Tools_TriStripper::CreateStrips( xGroup.m_uNumberOfIndices/3, pusListIndices, &pusStripIndices );

        Tools_TriStripper::Shutdown();

		delete [] pusListIndices;

		delete [] xGroup.m_puIndices;

		xGroup.m_puIndices = new u_int[ uNumStripIndices ];
		xGroup.m_uNumberOfIndices = uNumStripIndices;
	
		for (u_int u=0; u<xGroup.m_uNumberOfIndices; ++u)
		{
			xGroup.m_puIndices[u] = pusStripIndices[u];
		}

		delete [] pusStripIndices;

	}

    bOK = true;

CleanUp:
    
    if( ppiPolygonGroups )
    {
        for( u_int u = uNumMaterials; u--; )
        {
            delete[] ppiPolygonGroups[u];
        }
        delete[] ppiPolygonGroups;
    }

    return bOK;
}


			
		

void Tools_SmoothSkin_PSP_Processor::ReorderBonesToMinimiseRenderingStateChanges( Tools_SmoothSkin_PSP_Group& xCurrentGroup, const Tools_SmoothSkin_PSP_Group& xComparisonGroup )
{
	// if any bones in xCurrentGroup match xComparisionGroup, we'd like to keep them in the same bone slot (0-3)
	// so that state changes are minimised on the PSP
	u_char aucReorderedBones[4] = { 0xff, 0xff, 0xff, 0xff };
	bool abProcessedBone[4];

	for ( u_int uBone=0; uBone<4; ++uBone )
	{
		abProcessedBone[ uBone ] = false;

		for ( u_int uBone2=0; uBone2<4; ++uBone2 )
		{
			if ( xCurrentGroup.m_aucBoneIDs[uBone] == xComparisonGroup.m_aucBoneIDs[uBone2] )
			{ 
				aucReorderedBones[uBone2] = xCurrentGroup.m_aucBoneIDs[uBone];
				abProcessedBone[ uBone ] = true;
				break;
			}
		}
	}

	for ( u_int uBone=0; uBone<4; ++uBone )
	{
		if ( abProcessedBone[uBone]) continue;

		// find a blank space for the bone
		for ( u_int uBone2=0; uBone2<4; ++uBone2 )
		{
			if ( aucReorderedBones[ uBone2 ] == 0xff )
			{
				aucReorderedBones[ uBone2 ] = xCurrentGroup.m_aucBoneIDs[uBone];
				break;
			}
		}
	}
	
	for ( u_int uBone=0; uBone<4; ++uBone )
	{
		xCurrentGroup.m_aucBoneIDs[uBone] = aucReorderedBones[uBone];
	}

}

Tools_SmoothSkin_PSP_Processor::~Tools_SmoothSkin_PSP_Processor( void )
{
    delete[] m_pxGroups;
}

u_int Tools_SmoothSkin_PSP_Processor::GetTotalNumberOfVertices( void ) const
{
    u_int uTotalNumberOfVertices = 0;
    for (u_int u=0; u<m_uNumberOfGroups; u++)
    {
        uTotalNumberOfVertices += m_pxGroups[u].m_uNumberOfVertices;
    }
    return uTotalNumberOfVertices;
}

u_int Tools_SmoothSkin_PSP_Processor::GetTotalNumberOfIndices( void ) const
{
    u_int uTotalNumberOfIndices= 0;
    for (u_int u=0; u<m_uNumberOfGroups; u++)
    {
        uTotalNumberOfIndices += m_pxGroups[u].m_uNumberOfIndices;
    }
    return uTotalNumberOfIndices;
}

const Tools_SmoothSkin_PSP_Group* Tools_SmoothSkin_PSP_Processor::GetGroup( const u_int uGroup ) const
{
    if (uGroup >= m_uNumberOfGroups)
    {
        Asura_Assert( uGroup < m_uNumberOfGroups, "Group requested does not exist" );
        return 0;
    }

    return &m_pxGroups[uGroup];
}

//
// Tools_SmoothSkin_PSP_Group.
//

Tools_SmoothSkin_PSP_Group::Tools_SmoothSkin_PSP_Group()
{
    m_puIndices = 0;
    m_pxVertices = 0;
}

Tools_SmoothSkin_PSP_Group::~Tools_SmoothSkin_PSP_Group()
{
    delete[] m_puIndices;
    delete[] m_pxVertices;
}

//
// Tools_SmoothSkin_PSP_Group.
//

void Tools_SmoothSkin_PSP_Vertex::Set( const Tools_SmoothSkin_Vertex& xVertex, const u_char* pucBoneIDs )
{
	// Full range of short gets translated to (-1.f, 1.f). Most of our smooth skins are about 2m tall, so go with 4m range to be safe.
	// 32768 / 4 = 8192
    // (NY - 15.02.2007) changed to 8m to match ps2 (which also packs into 16bits, to accuracy is presumably ok) since simpsons has skins over 4m tall!
	// 32768 / 8 = 4096
	// version bumped so code can handle old data still
	m_asPosition[0] = static_cast<short>(xVertex.GetPos().x * 4096.f);
	m_asPosition[1] = static_cast<short>(xVertex.GetPos().y * 4096.f);
	m_asPosition[2] = static_cast<short>(xVertex.GetPos().z * 4096.f);
	// Normal gets mapped from (-1, 1) to (-16384, 16384)
	m_asNormal[0] = static_cast<short>(xVertex.GetNormal().x * 16384.f);
	m_asNormal[1] = static_cast<short>(xVertex.GetNormal().y * 16384.f);
	m_asNormal[2] = static_cast<short>(xVertex.GetNormal().z * 16384.f);
	m_usU = static_cast<u_short>(xVertex.GetU() * 32768.f);
	m_usV = static_cast<u_short>(xVertex.GetV() * 32768.f);
	m_uRGBAColour = xVertex.GetColour();

    // the 'assuming only using most important bone version'
    bool bBoneIndicesMatch = false;

    const u_char* pucBoneIndices = xVertex.GetBoneIndices();
    for( u_int u = 0; u < 4; ++u )
    {
		m_aucWeight[u] = 0;
	
		for ( u_int v = 0; v < 4; ++v )
		{
			if( pucBoneIndices[v] == pucBoneIDs[u] )
			{
	            bBoneIndicesMatch = true;
				int iWeight = Asura_Maths::FloatToInt_Near( xVertex.GetBoneWeights()[v] * 128.0f ) ;

				if (iWeight < 0) iWeight = 0;
				else if (iWeight > 128) iWeight = 128;

				m_aucWeight[u] = static_cast<u_char>(iWeight);
				break;
			}
		}
    }

	// sometimes weights don't add up to "one" due to rounding errors

	// find the error
	int iWeightError = -128;
	u_int uMostImportantBone = 0;
	for( u_int u = 0; u < 4; ++u )
    {
		iWeightError += m_aucWeight[u];

		if (m_aucWeight[ uMostImportantBone ] < m_aucWeight[u])
		{
			uMostImportantBone = u;
		}
	}

	// if the error is non-zero find the most important bone and tweak its weight to eliminate the error
	if ( iWeightError != 0 )
	{
		m_aucWeight[uMostImportantBone] = static_cast<u_char>(m_aucWeight[uMostImportantBone] - iWeightError);
	}

	

    Asura_Assert( bBoneIndicesMatch, "no bone indices matched" );
}

//
// Tools_SmoothSkin_PSP_BoneSet.
//

bool Tools_SmoothSkin_PSP_BoneSet::AddBonesOfTri( u_int uTri, const Tools_SmoothSkin_Processor* pxSrcProcessor, const Tools_SmoothSkin_SubElement& xSubElement )
{
    u_int uOriginalNumberOfBonesUsed = m_uNumberOfUsedBones;

    for (u_int v=0; v<3; v++)
    {
        u_int uVertexIndex = xSubElement.GetIndex( uTri, v );
        Asura_Assert( uVertexIndex < pxSrcProcessor->GetNumVertices(), "invalid index" );
        
        const Tools_SmoothSkin_Vertex& xVertex = pxSrcProcessor->GetVertex( uVertexIndex );
        const u_char* pucBoneIndices = xVertex.GetBoneIndices();

        // could possibly add all vertices - for now adding most important
        // need to cull off the 'noise' vertices with weights >5% or something
        for (u_int u=0; u<4; u++)
        //u_int u=0;
        {
            bool bSuccess = AddBone( pucBoneIndices[u] );

            // if not successful revert to the bone set we had at the beginning         
            if (!bSuccess)
            {
                m_uNumberOfUsedBones = uOriginalNumberOfBonesUsed;
                return false;
            }
        }
    }

    m_uNumberOfTrianglesAdded++;
    return true;
}

bool Tools_SmoothSkin_PSP_BoneSet::AddBonesOfTriUsingMostImportantWeightingsOnly( u_int uTri, const Tools_SmoothSkin_Processor* pxSrcProcessor, const Tools_SmoothSkin_SubElement& xSubElement )
{
	// This function is only really here to deal with 'broken' meshes.
    u_int uOriginalNumberOfBonesUsed = m_uNumberOfUsedBones;

    for (u_int v=0; v<3; v++)
    {
        u_int uVertexIndex = xSubElement.GetIndex( uTri, v );
        Asura_Assert( uVertexIndex < pxSrcProcessor->GetNumVertices(), "invalid index" );
        
        const Tools_SmoothSkin_Vertex& xVertex = pxSrcProcessor->GetVertex( uVertexIndex );
        const u_char* pucBoneIndices = xVertex.GetBoneIndices();

        // could possibly add all vertices - for now adding most important
        // need to cull off the 'noise' vertices with weights >5% or something
        // for (u_int u=0; u<4; u++)
        u_int u=0;
        {
            bool bSuccess = AddBone( pucBoneIndices[u] );

            // if not successful revert to the bone set we had at the beginning         
            if (!bSuccess)
            {
                m_uNumberOfUsedBones = uOriginalNumberOfBonesUsed;
                return false;
            }
        }
    }

    m_uNumberOfTrianglesAdded++;
    return true;
}
