/*
  Tools_Adjacency.h

  Created by Richard Rice - 24/06/2002

  Copyright 2002 Rebellion, All Rights Reserved.
*/

/*23456789.123456789.123456789.123456789.123456789.123456789.123456789.123456*/
#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __TOOLS_ADJACENCY_H__
#define __TOOLS_ADJACENCY_H__

#include "Core/Asura_Maths.h"
#include "MeshProcessing/Tools_MeshAdjacency.h"
#include "Core/Asura_HashTree.h"


class Tools_Adjacency_Handler
{
public:
	Tools_Adjacency_Handler();
	~Tools_Adjacency_Handler();


	void Initialise();
	void Shutdown();

	bool IsInitialised() { return m_bInitialised; }

	void Invalidate();
	void InvalidateModule(int iModuleId);

	void InvalidateCollision();
	void InvalidateCollisionModule(int iModuleId);

	//Adjacencies in Asura_Mesh
	const Tools_Polygon_Adjacency*
	  GetPolygonAdjacencies(int iModuleId,int iPolyId);

	const Tools_Polygon_Adjacency*
	  GetVertexAdjacencies(int iModuleId,int iVertexId);

	//Get vertices in mesh that aren't used by any polygons
	int GetNumLooseVertices(int iModuleId);
	int GetLooseVertex(int iModuleId,int iIndex);


	//Adjacencies in Asura_CollisionMesh
	const Tools_Polygon_Adjacency*
	  GetCollPolygonAdjacencies(int iModuleId,int iPolyId);

	const Tools_Polygon_Adjacency*
	  GetCollVertexAdjacencies(int iModuleId,int iVertexId);


private:
	void InitialiseModuleArray();
	void AddLink(int iModuleIndex,Asura_Hash_ID uLinkZoneHash,int iLinkModuleIndex,bool bCollision);
	void DeleteLink(int iModuleIndex,Asura_Hash_ID uLinkZoneHash,int iLinkModuleIndex,bool bCollision);

	Asura_Hash_ID m_uZoneHash;

	bool m_bInitialised;
	
	class Tools_Module_Adjacency* m_paxModules;
	class Tools_Module_Adjacency* m_paxCollModules;
	int m_iModulesCount;

	friend class Tools_Module_Adjacency;
	friend class Tools_Adjacency;

	#if defined ASURA_PC_SSE
		static const __m128 s_afToleranceP;
		static const __m128 s_afToleranceN;
	#endif
};

class Tools_Adjacency
{
public:
	static void Initialise();
	static void Shutdown();

	static bool IsInitialised() { return s_bInitialised; }

	//invalidates both render and collision adjacencies
	static void Invalidate();
	static void InvalidateModule(int iModuleId) {GetCurrentHandler().InvalidateModule(iModuleId);}

	//invalidates collision adjacencies
	static void InvalidateAllCollisionModules();
	static void InvalidateCollisionModule(int iModuleId) {GetCurrentHandler().InvalidateCollisionModule(iModuleId);}


	//Adjacencies in Asura_Mesh
	static const Tools_Polygon_Adjacency* GetPolygonAdjacencies(int iModuleId,int iPolyId)
	{ return GetCurrentHandler().GetPolygonAdjacencies(iModuleId,iPolyId); }

	static const Tools_Polygon_Adjacency* GetVertexAdjacencies(int iModuleId,int iVertexId)
	{ return GetCurrentHandler().GetVertexAdjacencies(iModuleId,iVertexId);}

	//Get vertices in mesh that aren't used by any polygons
	static int GetNumLooseVertices(int iModuleId)
	{ return GetCurrentHandler().GetNumLooseVertices(iModuleId); }

	static int GetLooseVertex(int iModuleId,int iIndex)
	{ return GetCurrentHandler().GetLooseVertex(iModuleId,iIndex); }


	//Adjacencies in Asura_CollisionMesh
	static const Tools_Polygon_Adjacency* GetCollPolygonAdjacencies(int iModuleId,int iPolyId)
	{ return GetCurrentHandler().GetCollPolygonAdjacencies(iModuleId,iPolyId); }

	static const Tools_Polygon_Adjacency* GetCollVertexAdjacencies(int iModuleId,int iVertexId)
	{ return GetCurrentHandler().GetCollVertexAdjacencies(iModuleId,iVertexId);}



	static bool IsDistanceWithinTolerance(const Asura_Vector_3& xDistance)
	{
		static const float fTolerance = 0.001f;
		
		return(xDistance.x>=-fTolerance && xDistance.x<=fTolerance
			&& xDistance.y>=-fTolerance && xDistance.y<=fTolerance
			&& xDistance.z>=-fTolerance && xDistance.z<=fTolerance);
	}

	#if defined(ASURA_USE_STREAMING)
	// Handler functions...
	static Tools_Adjacency_Handler* CreateHandler (const Asura_Hash_ID uHash);
	static bool DestroyHandler(const Asura_Hash_ID uHash);

	static void CreateAndSetHandler (const Asura_Hash_ID uHash) { s_pxCurrentHandler = CreateHandler(uHash); }
	static void SetCurrentHandler (Tools_Adjacency_Handler* pxHandler);
	static void SetCurrentHandler (const Asura_Hash_ID uHash) { s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static Tools_Adjacency_Handler&	GetCurrentHandler(void) { return(*s_pxCurrentHandler); }
	static Tools_Adjacency_Handler*	GetHandler(const Asura_Hash_ID uHash){ return(s_xHandlerTree.GetDataByHash(uHash)); }
	#else
	// Private functions...
	static Tools_Adjacency_Handler& GetCurrentHandler (void){ return(s_xHandler); }
	#endif	// defined(ASURA_USE_STREAMING)


private:
	static void InvalidateCallback(Asura_Hash_ID uHash,Tools_Adjacency_Handler* pxHandler){pxHandler->Invalidate();}
	static void InvalidateCollisionCallback(Asura_Hash_ID uHash,Tools_Adjacency_Handler* pxHandler){pxHandler->InvalidateCollision();}

	static bool s_bInitialised;

	#if defined(ASURA_USE_STREAMING)
	// Private variables...
	static	Tools_Adjacency_Handler* s_pxCurrentHandler;		// Currently active handler.
	static	Asura_TemplateHashTree<Tools_Adjacency_Handler> s_xHandlerTree;			// Tree of all of the handlers available.
	#else
	// Private variables...
	static	Tools_Adjacency_Handler s_xHandler;				// Single static-instance of a zone.
	#endif

};

class Tools_Module_Adjacency : public Tools_MeshAdjacency
{
public:

	Tools_Module_Adjacency();
	~Tools_Module_Adjacency();

	void Initialise(Asura_Hash_ID m_uZoneHash,int iModuleId, bool bCollisionMesh);
	
	void AddLink(const Asura_Hash_ID uZoneHash,const int iOtherId);
	void DeleteLink(const Asura_Hash_ID uZoneHash,const int iOtherId);
	void InvalidateLinkedModules();

protected:

	virtual int GetModuleIndex() const { return m_iModuleId; }
	virtual Asura_Hash_ID GetZoneHash() const {return m_uZoneHash;}

	virtual void CalculateVertexAdjacencies();
	virtual void CalculatePolygonAdjacencies();

	void CalculateAllAdjacencies();
	void CalculateAdjacenciesWithModule(class Asura_Module_Handler* pxOtherModuleHandler,const Asura_Hash_ID uOtherZoneHash,const int iOtherId);


	int m_iModuleId;
	Asura_Hash_ID m_uZoneHash;

	/*
		Keeps track of which modules may have references to each other,
		so we know to invalidate them when the number of polygons or
		vertices change.
	*/

	struct Tools_Adjacency_LinkedModule* m_pxLinkedModules;
	int m_iLinkedModulesCapacity;
	int m_iLinkedModulesCount;
	bool m_bCollisionMesh;
};

struct Tools_Adjacency_LinkedModule
{
	Asura_Hash_ID m_uZoneHash;
	int m_iModuleIndex;
};


//working space for matching vertices in adjacent modules
class Tools_Adjacency_VertexMatch
{
public :
	static void Initialise();
	static void Shutdown();

	static void StartVertMatch(int iNumVerts);
	static void EndVertMatch();

	static void AddEntry(int iThisVert,int iThatVert);

	static int GetNumMatchedVerts(int iVert){return s_pusNumMatches[iVert];}
	static int GetMatchedVert(int iVert,int iIndex) {return s_ppusMatchedVerts[iVert][iIndex];}

	static bool IsMatched(int iThisVert,int iThatVert);

private :
	static int s_iNumVerts;

	static u_short* s_pusNumMatches;
	static u_short** s_ppusMatchedVerts;

	static u_short* s_pusBuffer;
	static int s_iBufferSize;
	static int s_iSpaceUsed;

	static int s_iMostRecentVert;
};


#endif
