//
// Tools_Help.cpp
//
// Created by Kevin Lea - 05/10/2001 16:52:35
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Misc/Tools_Help.h"

// Stdlib independent.
#include <string.h>

// Engine independent.
#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Viewport.h"


////////////////////////////////////////////////////////////////////////////////////////////
/*-----------**
** Help Text **
**-----------*/
static const char* HelpHelpStrings[] =
{
	"Left :        Previous Topic",
	"Right :       Next Topic",
	"Up :          Scroll up",
	"Down :        Scroll down",
	"Escape :      Exit help",
	NULL,
};

////////////////////////////////////////////////////////////////////////////////////////////
/*------------**
** Tools_Help **
**------------*/

List<ToolsHelpTopic*>		Tools_Help::HelpTopicList;
LIL<ToolsHelpTopic*>		Tools_Help::HelpTopicLIL(&Tools_Help::HelpTopicList);
bool						Tools_Help::DisplayingHelp;

void Tools_Help::Initialise()
{
	DisplayingHelp = false;
	AddHelpTopic("Help",HelpHelpStrings);
}

void Tools_Help::Shutdown()
{
	RemoveHelpTopic("Help");
	while(HelpTopicList.size())
	{
		delete HelpTopicList.first_entry();
		HelpTopicList.delete_first_entry();
	}
}

void Tools_Help::Update()
{
	if(!HelpTopicList.size()) return;

   	if(!Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
	{
		if(DisplayingHelp)
		{
			
			//check for keys
			if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_LEFT))
			{
				HelpTopicLIL.previous();
			}
			if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_RIGHT))
			{
				HelpTopicLIL.next();
			}
			if(Asura_UserInput::KeyPressed(ASURA_KEYID_DOWN))
			{
				//scroll down
				HelpTopicLIL()->Offset+=Asura_Timers::GetFrameTime()*5.0f;
			}
			if(Asura_UserInput::KeyPressed(ASURA_KEYID_UP))
			{
				//scroll up
				HelpTopicLIL()->Offset-=Asura_Timers::GetFrameTime()*5.0f;
			}
		}
		else
		{
			if(Asura_UserInput::DebouncedKeyPressed('h') && Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
			{
				DisplayingHelp = true;
			}
		}
		if(DisplayingHelp && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ESCAPE))
		{
			DisplayingHelp = false;
			//Update input again to prevent escape from being picked up for other stuff.
			Asura_UserInput::Update();

		}
	}

}

void Tools_Help::Render()
{
	if(!HelpTopicList.size()) return;
	if(!DisplayingHelp) return;

	float width = (float)Asura_Viewport::GetWidth();
	float height = (float)Asura_Viewport::GetHeight();
	
	//darken everythin currently on screem
	Asura_Render::Overlay.SetRectXY(0,0,width,height);
	Asura_Render::Overlay.SetRectUV(0,0,1,1);
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.Colour=0x9f000000;
	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();


	ToolsHelpTopic* topic = HelpTopicLIL();
	
	//render the title
	float x = width/2;
	float y = 1.0f;
	Asura_Text_Output::ConvertPixelsToPrintCoords(x,y);
	Asura_Text_Output::Print(x,y,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,topic->Title);

	//convert screen width/height into print coords
	Asura_Text_Output::ConvertPixelsToPrintCoords(width,height);

	//draw the lines of text
	x = 1.0f;
	y = 4.0f;


	if((int)topic->Offset>=topic->NumStrings-1)
	{
		topic->Offset = (float)(topic->NumStrings-1);
	}
	if(topic->Offset<0.0f) topic->Offset = 0.0f;

	for(int i=(int)topic->Offset;i<topic->NumStrings;i++)
	{
		y = 4+i - topic->Offset;
		Asura_Text_Output::Print(x,y,0xffffffff,0,topic->HelpStrings[i]);
		if(y>height) break;
	}

}

void Tools_Help::AddHelpTopic( const char* Title, const char** Strings)
{
	Asura_Assert(Title,"Expecting a title");
	Asura_Assert(Strings,"Expecting some help strings");
	Asura_Assert(Strings[0],"Expecting at least one help string");

	
	//make sure we haven't already got this topic
	for(LIF<ToolsHelpTopic*> help_lif(&HelpTopicList);!help_lif.done();help_lif.next())
	{
		if(!_stricmp(help_lif()->Title,Title))
		{
			Asura_Assert(0,"Help topic has already been added");
			return;
		}
	}

	//allocate an new help topic
	ToolsHelpTopic* topic = new ToolsHelpTopic(Title,Strings);

	//add to the list
	HelpTopicList.add_entry(topic);
	//reset the iterator if this is the first entry
	if(HelpTopicList.size()==1)
	{
		HelpTopicLIL.restart();
	}
}

void Tools_Help::RemoveHelpTopic( const char* Title)
{
	Asura_Assert(Title,"Expecting a title");
	//search for the topic
	for(LIF<ToolsHelpTopic*> help_lif(&HelpTopicList);!help_lif.done();help_lif.next())
	{
		if(!_stricmp(help_lif()->Title,Title))
		{
			//adjust the iterator
			if(HelpTopicLIL() == help_lif())
			{
				HelpTopicLIL.previous();
			}
			//remove it
			delete help_lif();
			help_lif.delete_current();
			return;
		}
	}
	Asura_Assert(0,"Failed to find the help topic being removed");

}

void Tools_Help::ShowHelpTopic(const char* szTitle)
{
	Asura_Assert(szTitle,"Expecting a title");
	if(!HelpTopicList.size()) return;

	//change to the requested page
	for(int i=0;i<HelpTopicList.size();i++)
	{
		if(!_stricmp(HelpTopicLIL()->Title,szTitle))
		{
			DisplayingHelp = true;
			break;
		}
		HelpTopicLIL.next();

	}

	
}

///////////////////////////////////////////////////////////////////////////////////////////
/*----------------**
** ToolsHelpTopic **
**----------------*/

ToolsHelpTopic::ToolsHelpTopic( const char* _Title, const char** _HelpStrings)
{
	Title = _Title;
	HelpStrings = _HelpStrings;
	NumStrings = 0;
	Offset = 0.0f;

	//count the strings
	while(HelpStrings[NumStrings])
	{
		NumStrings++;
	}
}
