#include "Core/Asura.h"

// This file's header.
#include "Misc/Tools_IniFile.h"

// Engine independent.
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
#include "FileAccess/Asura_FileAccess.h"
#endif






Tools_IniFile::Tools_IniFile(const char* filename,bool bCheckNetwork /*= true*/)
{
	Initialise();

	if(filename)
	{
		#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
		char szFilename[MAX_PATH];
		if(bCheckNetwork)
		{
			Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy(filename,szFilename,MAX_PATH);
			filename = szFilename;
		}
		#endif

		Filename = new char[strlen(filename)+1];
		strcpy(Filename,filename);
	}

}

Tools_IniFile::~Tools_IniFile()
{
	Shutdown();
}

void Tools_IniFile::Initialise()
{
	Filename = 0;

	IniTextStart = 0;
	IniTextPos = 0;
	LinePos = 0;
	RootSection = 0;
	CurrentSection = 0;
	CurrentRootSection = 0;
	Loaded = false;
	Error = false;
	LineIndex = 0;

	LineBufferSize = 64;
	LineBuffer = new char[LineBufferSize];
}

void Tools_IniFile::Shutdown()
{
	delete [] Filename;
	delete [] IniTextStart;

	while(EnumerationList.size())
	{
		delete [] EnumerationList.first_entry().EnumString;
		EnumerationList.delete_first_entry();
	}

	delete RootSection;
	delete [] LineBuffer;
}

void Tools_IniFile::Reset()
{
	Shutdown();
	Initialise();
}

bool Tools_IniFile::SetFilename(const char* filename,bool bCheckNetwork /*=true*/)
{
	//if the file has already been loaded , it is to late to change the filename
	if(Loaded) return false;

	//replace the filename
	delete [] Filename;
	Filename = NULL;
	if(filename)
	{
		#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
		char szFilename[MAX_PATH];
		if(bCheckNetwork)
		{
			Asura_FileAccess::PlatformGetNetworkFilename(filename,szFilename,MAX_PATH);
			filename = szFilename;
		}
		#endif
		
		Filename = new char[strlen(filename)+1];
		strcpy(Filename,filename);
	}

	//reset error flag
	Error = false;

	return true;
}


bool Tools_IniFile::LoadIniFile()
{
	if(Error) return false;
	if(Loaded) return true;
	if(!Filename) return false;
	
	if(!IniTextStart)
	{
		HANDLE file=CreateFile(Filename,GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS, 0);
		if(file==INVALID_HANDLE_VALUE)
		{
			Error = true;
			return false;
		}

		int fileSize;
		unsigned long bytesRead;

		//copy the file contents into a buffer
		fileSize= GetFileSize(file,0);
		IniTextStart=new char[fileSize+1];
		IniTextStart[fileSize] = 0;
		ReadFile(file,IniTextStart,fileSize,&bytesRead,0);
		CloseHandle(file);

		BuildSectionList();
		LoadIniFileEnumerations();

		//go to start of ini file
		GetRootSection();
	}
	Loaded = true;
	return true;

}

const char* Tools_IniFile::GetIniTextSection(const char* sectionName,bool stepIn /*= false*/)
{
	if(!LoadIniFile()) return 0;

	CurrentSection = CurrentRootSection->GetSubSection(sectionName);
	
	if(CurrentSection)
	{
		IniTextPos = CurrentSection->SectionStart;
		if(stepIn)
		{
			CurrentRootSection = CurrentSection;
		}
		LineIndex = CurrentSection->SectionStartLine;

	}
	else
	{
		IniTextPos = 0;
		LineIndex = 0;
	}



	return IniTextPos;
}

const char* Tools_IniFile::GetIniTextSectionByIndex(int iSectionIndex,bool stepIn /*= false*/)
{
	if(!LoadIniFile()) return 0;

	CurrentSection = CurrentRootSection->GetSubSectionByIndex(iSectionIndex);
	
	if(CurrentSection)
	{
		IniTextPos = CurrentSection->SectionStart;
		if(stepIn)
		{
			CurrentRootSection = CurrentSection;
		}
		LineIndex = CurrentSection->SectionStartLine;

	}
	else
	{
		IniTextPos = 0;
		LineIndex = 0;
	}


	return IniTextPos;
}


const char* Tools_IniFile::GetParentSection()
{
	if(!LoadIniFile()) return 0;

	if(CurrentRootSection->Parent)
	{
		CurrentRootSection = CurrentRootSection->Parent;
	}
	CurrentSection = CurrentRootSection;
	IniTextPos = CurrentSection->SectionStart;
	LineIndex = CurrentSection->SectionStartLine;

	return IniTextPos;
}

const char* Tools_IniFile::GetRootSection()
{
	if(!LoadIniFile()) return 0;

	CurrentRootSection = RootSection;
	CurrentSection = CurrentRootSection;
	IniTextPos = CurrentSection->SectionStart;
   	LineIndex = CurrentSection->SectionStartLine;

	return IniTextPos;
}

const char* Tools_IniFile::GetCurrentSectionName()
{
	if(CurrentSection)
	{
		return CurrentSection->SectionName;
	}
	return NULL;
}

#if 1
void Tools_IniFile::BuildSectionList()
{
	CurrentRootSection = CurrentSection = RootSection = new IniTextSection(0);
	RootSection->SectionStart = IniTextStart;
	RootSection->SectionStartLine = 1;
	IniTextSection* ParentSection = RootSection;
	
	IniTextPos = IniTextStart;
	LineIndex=1;

	char* pos = IniTextStart;

	while(*pos)
	{
		//skip white space
		while(*pos)
		{
			if(*pos==' ' || *pos=='\t') pos++;
			else break;
		}
		
		if(*pos == '[')
		{
			*pos++ = 0;
			char* sectionName = pos;


			while(*pos && *pos!='\n' && *pos!='\r' && *pos!=']') pos++;
			if(*pos == ']')
			{
				*pos++ = 0;
				//find start of next line after section heading
				while(*pos && *pos!='\n' && *pos!='\r') pos++;
				if(*pos == '\r') pos++;
				if(*pos == '\n') pos++;
				LineIndex++;
					
				char* sectionStart = pos;

				IniTextSection* section = new IniTextSection(ParentSection);
				section->SectionName = sectionName;
				section->SectionStart = sectionStart;
				section->SectionStartLine = LineIndex-1;
				continue;
			}
		}
		else if(*pos == '{')
		{
			*pos++ = ' ';
			if(ParentSection->SubSections.size())
			{
				ParentSection = ParentSection->SubSections.last_entry();
			}
			else
			{
				Asura_Assert(0,"Unexpected '{' in inifile");
			}
		}
		else if(*pos == '}')
		{
			*pos++ = ' ';
			if(ParentSection->Parent)
			{
				ParentSection = ParentSection->Parent;
			}
			else
			{
				Asura_Assert(0,"Unexpected '}' in inifile");
			}
		}
		//go to the start of the next line
		while(*pos && *pos!='\n' && *pos!='\r') pos++;
		if(*pos == '\r') pos++;
		if(*pos == '\n') pos++;
		LineIndex++;
	}

	
}
#else

void Tools_IniFile::BuildSectionList()
{
	CurrentRootSection = CurrentSection = RootSection = new IniTextSection(0);
	RootSection->SectionStart = IniTextStart;
	RootSection->SectionStartLine = 1;
	
	IniTextPos = IniTextStart;
	LineIndex=1;

	char* pos = IniTextStart;

	while(*pos)
	{
		//skip white space
		while(*pos)
		{
			if(*pos==' ' || *pos=='\t') pos++;
			else break;
		}
		
		if(*pos == '[')
		{
			*pos++ = 0;
			char* sectionName = pos;


			while(*pos && *pos!='\n' && *pos!='\r' && *pos!=']') pos++;
			if(*pos == ']')
			{
				*pos++ = 0;
				//find start of next line after section heading
				while(*pos && *pos!='\n' && *pos!='\r') pos++;
				if(*pos == '\r') pos++;
				if(*pos == '\n') pos++;
				LineIndex++;
					
				char* sectionStart = pos;

				IniTextSection* section = new IniTextSection(RootSection);
				section->SectionName = sectionName;
				section->SectionStart = sectionStart;
				section->SectionStartLine = LineIndex-1;
				continue;
			}
		}
		else if(*pos == '{')
		{
			*pos++ = 0;
			char* sectionName = pos;


			while(*pos && *pos!='\n' && *pos!='\r' && *pos!='}') pos++;
			if(*pos == '}')
			{
				*pos++ = 0;
				//find start of next line after section heading
				while(*pos && *pos!='\n' && *pos!='\r') pos++;
				if(*pos == '\r') pos++;
				if(*pos == '\n') pos++;
				LineIndex++;
					
				char* sectionStart = pos;

				if(RootSection->SubSections.size())
				{
					IniTextSection* section = new IniTextSection(RootSection->SubSections.last_entry());
					section->SectionName = sectionName;
					section->SectionStart = sectionStart;
					section->SectionStartLine = LineIndex-1;
				}
				continue;
			}
		}
		//go to the start of the next line
		while(*pos && *pos!='\n' && *pos!='\r') pos++;
		if(*pos == '\r') pos++;
		if(*pos == '\n') pos++;
		LineIndex++;
	}

	
}
#endif


bool Tools_IniFile::GetNextLine()
{
	LineBuffer[0] = 0;
	LinePos = LineBuffer;
	
	if(!IniTextPos || *IniTextPos == 0)
	{
		return false;
	}
	
	//get the length of the next line
	int length = 0;
	char* lineStart = IniTextPos;
	while(*IniTextPos && *IniTextPos!='\n' && *IniTextPos!='\r')
	{
		length++;
		IniTextPos++;
	}
	if(length>=LineBufferSize)
	{
		//need to enlarge the linebuffer
		LineBufferSize = (length+16)&~15;
		delete [] LineBuffer;
		LineBuffer = new char[LineBufferSize];
		LinePos = LineBuffer;
	}
	strncpy(LineBuffer,lineStart,length);
	LineBuffer[length] = 0;

	//skip past the endofline characters
	if(*IniTextPos == '\r')
	{
		IniTextPos++;
	}
	if(*IniTextPos == '\n')
	{
		IniTextPos++;
	}
	LineIndex++;

	bool inString = false;
	while(*LinePos)
	{
		if(*LinePos == '\"')
		{
			inString = !inString;
		}
		if(*LinePos == '#' && !inString)
		{
			*LinePos = 0;
			break;
		}
		LinePos++;
	}

	LinePos = LineBuffer;

	return true;
}

List<char*> Tools_IniFile::GetSectionNameList()
{
	List<char*> sect_list;
	if(LoadIniFile())
	{
		CurrentRootSection->GetSectionNameList(sect_list);
	}
	return sect_list;
}

int Tools_IniFile::GetNumberOfSections()
{
	if(LoadIniFile())
	{
		return CurrentRootSection->GetNumberOfSections();
	}
	return 0;
}

bool Tools_IniFile::GetNextIniTextLine()
{
	if(!LoadIniFile()) return false;

	LineBuffer[0] = 0;
	LinePos = 0;

	if(!IniTextPos) return false;
	
	while(GetNextLine())
	{
		LinePos = LineBuffer;

		while(*LinePos)
		{
			if(*LinePos == ' ' || *LinePos == '\t' || *LinePos == '\n' || *LinePos == '\r' )
			{
				LinePos++;
			}
			else
			{
				LinePos = LineBuffer;
				return true;
			}
		}
	}

	IniTextPos = 0;
	return false;
}


bool Tools_IniFile::GetNextOnLine(int& output)
{
	if(!LoadIniFile()) return false;

	if(!LinePos) return false;

	while(*LinePos)
	{
		if(*LinePos == ' ' || *LinePos == '\t' || *LinePos == '\n' || *LinePos == '\r' )
		{
			LinePos++;
		}
		else
		{
			break;
		}
	}

	if((*LinePos>='a' && *LinePos<='z') || (*LinePos>='A' && *LinePos<='Z'))
	{
		if(!EnumValFromString(LinePos,output)) return false;
		
	}
	else if(LinePos[0] == '0' && LinePos[1] == 'x')
	{
		if(!sscanf(LinePos,"%x",&output)) return false;
	}
	else
	{
		if(!sscanf(LinePos,"%d",&output)) return false;
	}

	//skip past next comma
	LinePos = strchr(LinePos,',');
	if(LinePos) LinePos++;
	return true;
}

bool Tools_IniFile::GetNextOnLine(float& output)
{
	if(!LinePos) return false;

	while(*LinePos)
	{
		if(*LinePos == ' ' || *LinePos == '\t' || *LinePos == '\n' || *LinePos == '\r' )
		{
			LinePos++;
		}
		else
		{
			break;
		}
	}

	if((*LinePos>='a' && *LinePos<='z') || (*LinePos>='A' && *LinePos<='Z'))
	{
		int temp;
		if(!EnumValFromString(LinePos,temp)) return false;
		output = (float) temp;
		
	}
	else if(LinePos[0] == '0' && LinePos[1] == 'x')
	{
		int temp;
		if(!sscanf(LinePos,"%x",&temp)) return false;
		output = (float) temp;
	}
	else
	{
		if(!sscanf(LinePos,"%f",&output)) return false;
	}

	//skip past next comma
	LinePos = strchr(LinePos,',');
	if(LinePos) LinePos++;
	return true;
}

bool Tools_IniFile::GetNextOnLine(Asura_String& output)
{
	if(!LinePos) return false;
	output.MakeNull();

	while(*LinePos)
	{
		if(*LinePos == '\"')
		{
			LinePos++;	
			break;
		}

		if(*LinePos == ',') return false;
		LinePos++;
	}

	char* secondQuote = strchr(LinePos,'\"');
	if(!secondQuote) return false;
	
	*secondQuote = 0;
	if(strlen(LinePos))
	{
		output = LinePos;
	}
	LinePos = secondQuote+1;
	
	//skip past next comma
	LinePos = strchr(LinePos,',');
	if(LinePos) LinePos++;
	return true;
}

bool Tools_IniFile::GetString(Asura_String& output)
{
	if(!LineBuffer) return false;
	output.MakeNull();

	char* pcPos = LineBuffer;

	//skip initial spaces
	while(*pcPos == ' ' || *pcPos == '\t') pcPos++;
	//copy into a temp buffer
	char szBuffer[256];
	strncpy(szBuffer,pcPos,256);
	szBuffer[255] = 0;

	//find the last non-space
	char* pcTerminatorPos = szBuffer;
	pcPos = szBuffer;
	while(*pcPos)
	{
		if(*pcPos != ' ' && *pcPos != '\t')
		{
			pcTerminatorPos = pcPos+1;
		}
		pcPos++;
	}
	*pcTerminatorPos = 0;

	//copy this string into the output
	output = szBuffer;

	return !output.IsBlank();
}



void Tools_IniFile::LoadIniFileEnumerations()
{
	if(EnumerationList.size()) return;

	GetIniTextSection("Enumerations");
	while(GetNextIniTextLine())
	{
		int index;
		char buffer[256];
		char* equals = strchr(LineBuffer,'=');
		if(equals)
		{
			*equals = 0;
			equals++;
			if(sscanf(LineBuffer,"%s",buffer) && sscanf(equals,"%d",&index))
			{
				IniFileEnum ini;
				ini.EnumString = new char[strlen(buffer)+1];
				strcpy(ini.EnumString,buffer);
				ini.EnumValue = index;
				
				EnumerationList.add_entry(ini);
			}
		}
	}
}

bool Tools_IniFile::EnumValFromString(const char* buffer,int& output)
{
	char enumString[256] = "";
	int length = 0;

	while(*buffer && *buffer!=' ' && *buffer!='\t' && *buffer!='\n' && *buffer !='\r' && *buffer!=',')
	{
		enumString[length++] = *buffer++;
	}
	
	for(LIF<IniFileEnum> ilif(&EnumerationList);!ilif.done();ilif.next())
	{
		if(!_stricmp(ilif().EnumString,enumString))
		{
			output = ilif().EnumValue;
			return true;
		}
	}
	return false;
}



bool Tools_IniFile::GetStringInCurrentSection(const char* identifier,Asura_String& output, bool bStripQuotes /*=false*/)
{
	if(!LoadIniFile()) return false;
	if(!CurrentSection) return false;
	return GetStringFromBuffer(identifier,CurrentSection->SectionStart,output,bStripQuotes)!=false;
}

bool Tools_IniFile::GetIntInCurrentSection(const char* identifier,int& output)
{
	if(!LoadIniFile()) return false;
	if(!CurrentSection) return false;

	Asura_String entry;
	if(GetStringFromBuffer(identifier,CurrentSection->SectionStart,entry))
	{
		const char* text = entry.GetString();
		
		if((*text>='a' && *text<='z') || (*text>='A' && *text<='Z'))
		{
			if(!EnumValFromString(text,output)) return false;
			
		}
		else if(text[0] == '0' && text[1] == 'x')
		{
			if(!sscanf(text,"%x",&output)) return false;
		}
		else
		{
			if(!sscanf(text,"%d",&output)) return false;
		}
		return true;

	}
	return false;

}


bool Tools_IniFile::GetUIntInCurrentSection(const char* identifier,u_int& output)
{
	if(!LoadIniFile()) return false;
	if(!CurrentSection) return false;

	Asura_String entry;
	if(GetStringFromBuffer(identifier,CurrentSection->SectionStart,entry))
	{
		const char* text = entry.GetString();
		int iTemp;

		if((*text>='a' && *text<='z') || (*text>='A' && *text<='Z'))
		{
			if(!EnumValFromString(text,iTemp)) return false;
			
		}
		else if(text[0] == '0' && text[1] == 'x')
		{
			if(!sscanf(text,"%x",&iTemp)) return false;
		}
		else
		{
			if(!sscanf(text,"%d",&iTemp)) return false;
		}

		output = static_cast<u_int> (iTemp);
		return true;

	}
	return false;

}

bool Tools_IniFile::GetFloatInCurrentSection(const char* identifier,float& output)
{
	if(!LoadIniFile()) return false;
	if(!CurrentSection) return false;

	Asura_String entry;
	if(GetStringFromBuffer(identifier,CurrentSection->SectionStart,entry))
	{
		const char* text = entry.GetString();
		
		if(!sscanf(text,"%f",&output)) return false;
		
		return true;

	}
	return false;
}

bool Tools_IniFile::GetBoolInCurrentSection(const char* identifier,bool& output)
{
	if(!LoadIniFile()) return false;
	if(!CurrentSection) return false;

	Asura_String entry;
	if(GetStringFromBuffer(identifier,CurrentSection->SectionStart,entry))
	{
		const char* text = entry.GetString();
		
		if(!_stricmp(text,"Yes") || !_stricmp(text,"True"))
		{
			output = true;
			return true;
		}
		if(!_stricmp(text,"No") || !_stricmp(text,"False"))
		{
			output = false;
			return true;
		}
		int temp;
		if(sscanf(text,"%d",&temp) == 1)
		{
			output = temp >0;
			return true;
		}
		Asura_Assertf( false, ("Tools Ini File '%s' has unknown bool value in identifier '%s'", this->Filename, identifier));
	}
	return false;

}

void Tools_IniFile::ListIdentifiersInCurrentSection(List<Asura_String>& xIdentifierList)
{
	if(!LoadIniFile()) return;
	if(!CurrentSection) return;

	ListIdentifiersInBuffer(CurrentSection->SectionStart,xIdentifierList);
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------**
** IniTextSection **
**----------------*/
IniTextSection::IniTextSection(IniTextSection* parent)
{
	SectionName = 0;
	SectionStart = 0;
	Parent = parent;
	if(Parent)
	{
		Parent->SubSections.add_entry(this);
	}
}

IniTextSection::~IniTextSection()
{
	while(SubSections.size())
	{
		delete SubSections.first_entry();
		SubSections.delete_first_entry();
	}
}

IniTextSection* IniTextSection::GetSubSection(const char* name)
{
	if(!name)
	{
		return this;
	}

	for(LIF<IniTextSection*> sublif(&SubSections);!sublif.done();sublif.next())
	{
		if(!_stricmp(name,sublif()->SectionName))
		{
			return sublif();
		}
	}
	return 0;
}

IniTextSection* IniTextSection::GetSubSectionByIndex(int iIndex)
{

	int iCount = 0;
	for(LIF<IniTextSection*> sublif(&SubSections);!sublif.done();sublif.next())
	{
		if(!_stricmp("Enumerations",sublif()->SectionName)) continue;

		if(iCount == iIndex)
		{
			return sublif();
		}
		iCount++;
	}
	return 0;
}


void  IniTextSection::GetSectionNameList(List<char*>& sect_list)
{
	while(sect_list.size()) sect_list.delete_first_entry();
	
	for(LIF<IniTextSection*> sublif(&SubSections);!sublif.done();sublif.next())
	{
		if(_stricmp("Enumerations",sublif()->SectionName))
		{
			sect_list.add_entry(sublif()->SectionName);
		}
	}
}

int IniTextSection::GetNumberOfSections()
{
	int iCount = 0;
	for(LIF<IniTextSection*> sublif(&SubSections);!sublif.done();sublif.next())
	{
		if(_stricmp("Enumerations",sublif()->SectionName))
		{
			iCount++;
		}
	}
	return iCount;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------**
** Tools_TabbedStringTable **
**-------------------------*/
Tools_TabbedStringTable::Tools_TabbedStringTable(const char* pszFilename,bool bCheckNetwork /*= true*/)
{
	m_iWidth = 0;
	m_iHeight = 0;
	m_pppcTable = 0;
	m_pcBuffer = 0;
	m_bError = false;

	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char szFilename[MAX_PATH];
	if(bCheckNetwork)
	{
		Asura_FileAccess::PlatformGetNetworkFilename(pszFilename,szFilename,MAX_PATH);
		pszFilename = szFilename;
	}
	#endif


	//first need to load the file , and read it into memory
	HANDLE hFile=CreateFile(pszFilename,GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS, 0);
	if(hFile==INVALID_HANDLE_VALUE)
	{
		m_bError = true;
		return;
	}

	int iFileSize;
	unsigned long uBytesRead;

	//copy the file contents into a buffer
	iFileSize= GetFileSize(hFile,0);
	m_pcBuffer=new char[iFileSize+1];
	m_pcBuffer[iFileSize] = 0;
	ReadFile(hFile,m_pcBuffer,iFileSize,&uBytesRead,0);
	CloseHandle(hFile);

	//now scan through the file working out the width and height
	char* pcPos = m_pcBuffer;
	if(!*pcPos)
	{
		m_bError = true;
		return;
	}

	while(*pcPos)
	{
		m_iHeight++;

		int iTabs = 0;
		//count the tabs on this line
		while(*pcPos && *pcPos!='\r' && *pcPos!='\n')
		{
			if(*pcPos == '\t') iTabs++;
			pcPos++;
		}	
		m_iWidth = max(iTabs+1,m_iWidth);

		//skip to the next line
		if(*pcPos == '\r') pcPos++;
		if(*pcPos == '\n') pcPos++;

	}
	int i,j;

	m_pppcTable = new char**[m_iHeight];
	for(i=0;i<m_iHeight;i++)
	{
		m_pppcTable[i] = new char*[m_iWidth];
		for(j=0;j<m_iWidth;j++)
		{
			m_pppcTable[i][j] = 0;
		}
	}

	//now scan though the buffer a second time , filling in the table,
	//and inserting terminators
	pcPos = m_pcBuffer;
	i=0;
	j=0;

	while(*pcPos)
	{
		m_pppcTable[i][j] = pcPos;

		//skip to next tab or newline
		while(*pcPos && *pcPos!='\t' && *pcPos!='\n' && *pcPos!='\r') pcPos++;

		if(*pcPos == '\t')
		{
			*pcPos++ = 0;
			j++;
		}
		else if(*pcPos == '\r')
		{
			*pcPos++ = 0;
			pcPos++;
			i++;
			j=0;
		}
		else if(*pcPos == '\n')
		{
			*pcPos++ = 0;
			i++;
			j=0;
		}
	}
}

Tools_TabbedStringTable::~Tools_TabbedStringTable()
{
	delete [] m_pcBuffer;
	for(int i=0;i<m_iHeight;i++)
	{
		delete [] m_pppcTable[i];
	}
	delete [] m_pppcTable;
}

const char* Tools_TabbedStringTable::GetEntry(int iLine,int iCol) const
{
	if(iLine<m_iHeight && iCol<m_iWidth)
	{
		return m_pppcTable[iLine][iCol];
	}
	return 0;	
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Tools_UnicodeTabbedStringTable **
**--------------------------------*/

//Perforce messes up unicode text files by replacing newlines with this character
static const Asura_UnicodeChar g_wcMangledUnicodeNewline = ('\n' << 8) | '\r';

Tools_UnicodeTabbedStringTable::Tools_UnicodeTabbedStringTable(const char* pszFilename,bool bCheckNetwork /*= true*/)
{
	m_iWidth = 0;
	m_iHeight = 0;
	m_pppwcTable = 0;
	m_pcBuffer = 0;
	m_bError = false;

	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char szFilename[MAX_PATH];
	if(bCheckNetwork)
	{
		Asura_FileAccess::PlatformGetNetworkFilename(pszFilename,szFilename,MAX_PATH);
		pszFilename = szFilename;
	}
	#endif


	//first need to load the file , and read it into memory
	HANDLE hFile=CreateFile(pszFilename,GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS, 0);
	if(hFile==INVALID_HANDLE_VALUE)
	{
		m_bError = true;
		return;
	}

	int iFileSize;
	unsigned long uBytesRead;

	//copy the file contents into a buffer
	iFileSize= GetFileSize(hFile,0);
	m_pcBuffer=new char[iFileSize+2];
	*(Asura_UnicodeChar*)&m_pcBuffer[iFileSize] = L'\0';
	ReadFile(hFile,m_pcBuffer,iFileSize,&uBytesRead,0);
	CloseHandle(hFile);

	Asura_UnicodeChar* pwcUnicodeText = (Asura_UnicodeChar*) m_pcBuffer;

	//make sure this is unicode
	if(*pwcUnicodeText != 0xFEFF)
	{
		m_bError = true;
		return;
	}
	pwcUnicodeText++;

	
	//now scan through the file working out the width and height
	Asura_UnicodeChar* pwcPos = pwcUnicodeText;
	if(!*pwcPos)
	{
		m_bError = true;
		return;
	}

	int iLine = 0;
	while(*pwcPos)
	{
		iLine++;

		int iTabs = 0;
		//count the tabs on this line
		bool bCheckForEnteringQuotes = true;
		bool bInQuotes = false;

		while(*pwcPos)
		{
			Asura_Assertf(*pwcPos != g_wcMangledUnicodeNewline,("Unicode text file '%s' appears to have been mangled by perforce.\nPlease make sure that it is flagged as binary in perforce.\nYou will also need to re-export the file from its excel spreadsheet , and possible force at least one change to make sure the file gets updated.",pszFilename));
	
			
			if(bInQuotes)
			{
				if(*pwcPos == L'\"')
				{
					if(pwcPos[1] != L'\"')
					{
						//single quote ,so reached end of quotes
						bInQuotes = false;
						pwcPos++;
					}
					else
					{
						pwcPos+=2;
					}

				}
				else
				{
					pwcPos++;
				}
			}
			else
			{
				if(*pwcPos == L'\r' || *pwcPos == L'\n')
				{
					//end of line
					break;
				}
				else if(*pwcPos == L'\t')
				{
					iTabs++;
					pwcPos++;
					bCheckForEnteringQuotes = true;

				}
				else if(bCheckForEnteringQuotes && *pwcPos == L'\"')
				{
					//quoted entry
					bCheckForEnteringQuotes = false;
					bInQuotes = true;	
					pwcPos++;
					
					//Something in this cell , so update count of max used width/height
					if(iTabs+1>m_iWidth) m_iWidth = iTabs+1;
					m_iHeight = iLine;
				}
				else
				{
					bCheckForEnteringQuotes = false;
					pwcPos++;

					//Something in this cell , so update count of max used width/height
					if(iTabs+1>m_iWidth) m_iWidth = iTabs+1;
					m_iHeight = iLine;
				}
			}
		}

		//skip to the next line
		if(*pwcPos == L'\r') pwcPos++;
		if(*pwcPos == L'\n') pwcPos++;

	}
	if(m_iWidth < 1) m_iWidth = 1;
	if(m_iHeight < 1) m_iHeight = 1;

	int i,j;

	m_pppwcTable = new Asura_UnicodeChar**[m_iHeight];
	m_pppwcTable[0] = new Asura_UnicodeChar*[m_iWidth*m_iHeight];
	Asura_Mem::Set(m_pppwcTable[0], 0, m_iWidth*m_iHeight*sizeof(Asura_UnicodeChar*));
	for(int i=1; i<m_iHeight; i++)
	{
		m_pppwcTable[i] = m_pppwcTable[0]+(i*m_iWidth);
	}

	//now scan though the buffer a second time , filling in the table,
	//and inserting terminators
	pwcPos = pwcUnicodeText;
	i=0;
	j=0;

	while(*pwcPos)
	{
		Asura_UnicodeChar* pwcCellStart = pwcPos;
		

		bool bCheckForEnteringQuotes = true;
		bool bInQuotes = false;

		Asura_UnicodeChar* pwcDest;

		//search for next line or tab
		while(*pwcPos)
		{
			if(bInQuotes)
			{
				if(*pwcPos == L'\"')
				{
					if(pwcPos[1] != L'\"')
					{
						//single quote ,so reached end of quotes
						*pwcDest++ = 0;
						bInQuotes = false;
						pwcPos++;
					}
					else
					{
						//two quotes. copy one and skip the other.
						*pwcDest++ = L'\"';
						pwcPos+=2;
					}

				}
				else
				{
					*pwcDest++ = *pwcPos++;
				}
			}
			else
			{
				if(*pwcPos == L'\r' || *pwcPos == L'\n' || *pwcPos == L'\t')
				{
					break;
				}
				else if(bCheckForEnteringQuotes && *pwcPos == L'\"')
				{
					//First character in entry was a quote , so move the start position after it
					//Also will need to copy quoted strings as we go , since some doubled up quotes may need removing
					bCheckForEnteringQuotes = false;
					bInQuotes = true;	
					pwcPos++;
					pwcCellStart = pwcPos;
					
					pwcDest = pwcPos;

				}
				else
				{
					bCheckForEnteringQuotes = false;
					pwcPos++;
				}
			}
		}

		if(*pwcPos == L'\t' || *pwcPos == L'\r' || *pwcPos == L'\n')
		{
			//end of text for the current cell
			Asura_UnicodeChar wCellTerminator = *pwcPos;
			*pwcPos++ = 0;
			if(j<m_iWidth && i<m_iHeight)
			{
				m_pppwcTable[i][j] = pwcCellStart;
			}
			else
			{
				//expectng any strings outside of the used width/height to be blank
				Asura_Assert(*pwcCellStart == 0,"Found non-blank string in area of table previously determined to be empty");
			}
			
			if(wCellTerminator == L'\t')
			{
				j++;
			}
			else if(wCellTerminator == L'\r')
			{
				pwcPos++;
				i++;
				j=0;
			}
			else if(wCellTerminator == L'\n')
			{
				i++;
				j=0;
			}
		}
	}
}

Tools_UnicodeTabbedStringTable::~Tools_UnicodeTabbedStringTable()
{
	delete [] m_pcBuffer;
	if(m_pppwcTable)
	{
		delete [] m_pppwcTable[0];
	}
	delete [] m_pppwcTable;
}

const Asura_UnicodeChar* Tools_UnicodeTabbedStringTable::GetEntry(int iLine,int iCol) const
{
	if(iLine<m_iHeight && iCol<m_iWidth)
	{
		return m_pppwcTable[iLine][iCol];
	}
	return 0;	
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////

void ListIdentifiersInBuffer(const char* pszBuffer,List<Asura_String>& xIdentifierList)
{
	char szTempString[200];
	
	while(*pszBuffer)
	{
		//skip past white space at the beginning of the line
		while(*pszBuffer && *pszBuffer!='#')
		{
			if(*pszBuffer==' ' || *pszBuffer=='\t') pszBuffer++;
			else break;
		}

		if(*pszBuffer!='#')
		{
			const char* pcIdentifierStart = pszBuffer;
			const char* pcLastNonSpace = 0;

			//search for the equals sign after the identifier
			while(*pszBuffer)
			{
				//stop when we reach the end of the line  or an equals
				if(*pszBuffer=='=' || *pszBuffer=='\r' || *pszBuffer=='\n') break;
				
				if(*pszBuffer !=' ' && *pszBuffer!='\t') pcLastNonSpace = pszBuffer;
				pszBuffer++;

			}

			//if we've found an equals sign , and some nonspace character , add to the identifier list
			if(*pszBuffer == '=' && pcLastNonSpace)
			{
				int iLength = ((int)pcLastNonSpace-(int)pcIdentifierStart)+1;
				if(iLength>199) iLength = 199;

				strncpy(szTempString,pcIdentifierStart,iLength);
				szTempString[iLength] = 0;

				xIdentifierList.add_entry(szTempString);
			}
		}
		//skip to next line
		while(*pszBuffer && *pszBuffer!='\r' && *pszBuffer!='\n') pszBuffer++;
		
		if(*pszBuffer=='\r')pszBuffer++;
		if(*pszBuffer=='\n')pszBuffer++;
		
	}

}


bool GetStringFromBuffer(const char* szIdentifier,const char* szBuffer,Asura_String& xOutput, bool bStripQuotes)
{
	char acTempString[200];
	char* pcString=&acTempString[0];
	int iIDLength=strlen(szIdentifier);
	
	while(*szBuffer)
	{
		//skip past white space at the beginning of the line
		while(*szBuffer && *szBuffer!='#')
		{
			if(*szBuffer==' ' || *szBuffer=='\t') szBuffer++;
			else break;
		}
		
		if(*szBuffer!='#')
		{
			//see if the string matches the identifier
			if(!_strnicmp(szBuffer,szIdentifier,iIDLength))
			{
				//skip past the identifier
				szBuffer+=iIDLength;

				
				//skip past white space
				while(*szBuffer)
				{
					if(*szBuffer==' ' || *szBuffer=='\t') szBuffer++;
					else break;
				}

				//have we found the equals sign?
				if(*szBuffer=='=')
				{
					//skip past the sign
					szBuffer++;
					//skip past white space
					while(*szBuffer)
					{
						if(*szBuffer==' ' || *szBuffer=='\t') szBuffer++;
						else if (bStripQuotes && *szBuffer=='\"') szBuffer++;
						else break;
					}
				
					//now copy the text up until the end of the line
					char* pcLastNonSpace=0;
					while(*szBuffer)
					{
						if(*szBuffer=='\n' || *szBuffer=='\r' || *szBuffer=='#')
						{
							*pcString=0;
							break;
						}
						else if(*szBuffer==' ' || *szBuffer=='\t')
						{
							*pcString++=*szBuffer++;
						}
						else if (bStripQuotes && *szBuffer=='\"')
						{
							*pcString++=*szBuffer++;
						}
						else
						{
							pcLastNonSpace=pcString;
							*pcString++=*szBuffer++;
						}

					}
					//finally get rid of trailing spaces on the string
					if(pcLastNonSpace)
					{
						pcLastNonSpace++;
						*pcLastNonSpace=0;
					}

					
					//copy the string to the destination
					xOutput = acTempString;

					return true;

				}
				
				
			}
		}
		//skip to next line
		while(*szBuffer && *szBuffer!='\r' && *szBuffer!='\n') szBuffer++;
		
		if(*szBuffer=='\r')szBuffer++;
		if(*szBuffer=='\n')szBuffer++;
	}

	return false;
}

bool GetIntFromBuffer(const char* identifier,const char* buffer,int& output)
{
	Asura_String tempString;
	if(GetStringFromBuffer(identifier,buffer,tempString))
	{
		return (sscanf(tempString.GetString(),"%d",&output) == 1);
	}
	return false;
}

bool GetBoolFromBuffer(const char* identifier,const char* buffer,bool& output)
{
	Asura_String tempString;
	if(GetStringFromBuffer(identifier,buffer,tempString))
	{
		if(!tempString.Compare("Yes") || !tempString.Compare("True"))
		{
			output = true;
			return true;
		}
		if(!tempString.Compare("No") || !tempString.Compare("False"))
		{
			output = false;
			return true;
		}
		int temp;
		if(sscanf(tempString.GetString(),"%d",&temp) == 1)
		{
			output = temp >0;
			return true;
		}
	}
	return false;
}

bool GetFloatFromBuffer(const char* identifier,const char* buffer,float& output)
{
	Asura_String entry;
	if(GetStringFromBuffer(identifier,buffer,entry))
	{
		const char* text = entry.GetString();
		
		if(!sscanf(text,"%f",&output)) return false;
		
		return true;
	}
	return false;
}

