//
// Tools_MessageLog.cpp
//
// Created by Richard Rice - 27/07/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Misc/Tools_MessageLog.h"

#include "Core/Asura_NameTree.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Text_Output.h"

#include "Tools_Overlays.h"
#include "Tools_System.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/list_tem.hpp"
#include "Misc/Tools_Timed_Messages.h"


////////////////////////////////////////////////////////////////////////////////////////////////
class Tools_Overlay_MessageLog : public Asura_Overlay
{
public :
	Tools_Overlay_MessageLog( const char *szName, bool bAlwaysOnTop, Asura_Vector_2* pxPosition = 0, u_int* puMinimised = 0 )
	:Asura_Overlay(szName, bAlwaysOnTop, pxPosition, puMinimised){}
	
protected :
	//make sure this overlay is only ever visible while tools is active
	virtual bool IsHidden() const
	{
		if(!Tools_System::IsActive()) return true;
		return Asura_Overlay::IsHidden();
	}
	
};


////////////////////////////////////////////////////////////////////////////////////////////////



static const int g_iMaxShortTextLength = 120;
static const u_int g_uMaxMessagesToSave = 10000;


bool Tools_MessageLog::s_bInitialised = false;
u_int Tools_MessageLog::s_uNextID = 1;
List<Tools_MessageLogEntry*> Tools_MessageLog::s_xMessageList;

Asura_Vector_2 Tools_MessageLog::s_xPosition(10.0f,2.0f);

Asura_String Tools_MessageLog::s_xSearchString;
bool Tools_MessageLog::s_bSearchStringChanged = false;

Asura_NameTree Tools_MessageLog::s_xCatNameTree;
const char** Tools_MessageLog::s_pszCatList;
u_int Tools_MessageLog::s_uCatListSize;
int Tools_MessageLog::s_iCatIndex;
bool Tools_MessageLog::s_bCatIndexChanged = false;
Asura_String Tools_MessageLog::s_xCatFilter;
Asura_DropDown_List_Dialog* Tools_MessageLog::s_pxCatListDialog;

Asura_NameTree Tools_MessageLog::s_xSubCatNameTree;
const char** Tools_MessageLog::s_pszSubCatList;
u_int Tools_MessageLog::s_uSubCatListSize;
int Tools_MessageLog::s_iSubCatIndex;
bool Tools_MessageLog::s_bSubCatIndexChanged = false;
Asura_String Tools_MessageLog::s_xSubCatFilter;
Asura_DropDown_List_Dialog* Tools_MessageLog::s_pxSubCatListDialog;

const char** Tools_MessageLog::s_pszMessageStrings;
const char** Tools_MessageLog::s_pszMessageHelpStrings;
Tools_MessageLogEntry** Tools_MessageLog::s_ppxMessages;
u_int Tools_MessageLog::s_uMessageStringsSize;
int Tools_MessageLog::s_iMessageIndex;
bool Tools_MessageLog::s_bMessageIndexChanged;
Asura_List_Dialog* Tools_MessageLog::s_pxMessageListDialog;
u_int Tools_MessageLog::s_uMostRecentSelectedID = 0;

u_int Tools_MessageLog::s_uNextIDWhenStringsWereLastUpdated = 0;

u_int Tools_MessageLog::s_uShowOverlayOptions = 0xffffffff;

const char* Tools_MessageLog::s_aszCatStrings[ NUM_CATEGORIES ] =
{
	"",
	"None",

	// Mode-type categories.
	"Blueprint",
	"Botanicals",
	"Cutscene",
	"Dialogue",
	"Entities",
	"Gui Menu",
	"IK",
	"Material Response",
	"Mesh Edit",
	"PFX",
	"Sound Event",
	"Waypointing",

	// Event-type categories.
	"Export",
	"Verification"
};

static const char* g_szFilterNone = "*All*";

static Asura_Vector_2 g_xUserSetListSize(0.0f,0.0f);

void Tools_MessageLog::Initialise()
{
	Asura_Assert(!s_bInitialised,"Already initialised");
	s_bInitialised = true;

    TransferModeData( TOOLSMODEDATAFILE_LOAD );
}


void Tools_MessageLog::Shutdown()
{
    TransferModeData( TOOLSMODEDATAFILE_SAVE );

	Asura_Assert(s_bInitialised,"Not initialised");
	CloseOverlay();

	DeleteCatList();
	DeleteSubCatList();
	DeleteMessageStrings();

	while(s_xMessageList.size())
	{
		delete s_xMessageList.first_entry();
		s_xMessageList.delete_first_entry();
	}

	s_bInitialised = false;
}

void Tools_MessageLog::Update()
{
	

	if(!OverlayExists()) return;
	

	bool bNeedToUpdateMainList = false;

	//deal with changes to the filters
	if(s_bCatIndexChanged)
	{
		if(s_iCatIndex)
		{
			s_xCatFilter = s_pszCatList[s_iCatIndex];
		}
		else
		{
			s_xCatFilter.MakeNull();
		}
		UpdateSubCatList();
		bNeedToUpdateMainList = true;
	}

	if(s_bSubCatIndexChanged)
	{
		if(s_iSubCatIndex)
		{
			s_xSubCatFilter = s_pszSubCatList[s_iSubCatIndex];
		}
		else
		{
			s_xSubCatFilter.MakeNull();
		}

		bNeedToUpdateMainList = true;
	}

	if(s_bSearchStringChanged)
	{
		s_bSearchStringChanged = false;
		bNeedToUpdateMainList = true;
	}

	if(bNeedToUpdateMainList || s_uNextIDWhenStringsWereLastUpdated != s_uNextID)
	{
		UpdateMessageStrings();
	}
	else if(s_bMessageIndexChanged)
	{
		if(s_ppxMessages)
		{
			//keep track of the selection most recently actively picked by the user
			s_uMostRecentSelectedID = s_ppxMessages[s_iMessageIndex]->m_uID;
		}
	}

	s_bCatIndexChanged = false;
 	s_bSubCatIndexChanged = false;
	s_bMessageIndexChanged = false;

}

void Tools_MessageLog::AddMessage(const char* szCat,const char* szSubCat,const char* szFormat,...)
{
	if(!s_bInitialised) return;

	va_list xArgs;
	va_start(xArgs, szFormat);
    AddMessageInternal( szCat,szSubCat,NULL,szFormat,xArgs);
	va_end(xArgs);

}

void Tools_MessageLog::AddMessage(Category eCat,const char* szSubCat,const char* szFormat,...)
{
	if(!s_bInitialised) return;

	if ( eCat < NO_CATEGORY || NUM_CATEGORIES <= eCat )
	{
		eCat = NO_CATEGORY;
	}

	va_list xArgs;
	va_start(xArgs, szFormat);
	AddMessageInternal( s_aszCatStrings[ eCat ],szSubCat,NULL,szFormat,xArgs);
	va_end(xArgs);

}

void Tools_MessageLog::AddMessage(const char* szCat,const char* szSubCat,Tools_MessageLogAction* pxAction,const char* szFormat,...)
{
	if(!s_bInitialised) return;

	va_list xArgs;
	va_start(xArgs, szFormat);
    AddMessageInternal( szCat,szSubCat,pxAction,szFormat,xArgs);
	va_end(xArgs);

}

void Tools_MessageLog::AddMessage(Category eCat,const char* szSubCat,Tools_MessageLogAction* pxAction,const char* szFormat,...)
{
	if(!s_bInitialised) return;

	if ( eCat < NO_CATEGORY || NUM_CATEGORIES <= eCat )
	{
		eCat = NO_CATEGORY;
	}

	va_list xArgs;
	va_start(xArgs, szFormat);
	AddMessageInternal( s_aszCatStrings[ eCat ],szSubCat,pxAction,szFormat,xArgs);
	va_end(xArgs);

}


void Tools_MessageLog::AddMessageInternal(const char* szCat,const char* szSubCat,Tools_MessageLogAction* pxAction,const char* szFormat,va_list xArgs)
{
	if(!s_bInitialised) return;

	char szBuffer[512];
	Asura_PrintfToBfr(szBuffer,szFormat,xArgs,sizeof(szBuffer),true);


	//fill in message log entry
	Tools_MessageLogEntry* pxEntry = new Tools_MessageLogEntry;
	pxEntry->m_xText = szBuffer;
	pxEntry->m_xCategory = (szCat && *szCat) ? szCat : 0;
	pxEntry->m_xSubCategory = (szSubCat && *szSubCat) ? szSubCat : 0;
	pxEntry->m_uID = s_uNextID++;
	pxEntry->m_pxAction = pxAction;
	pxEntry->m_bShown = false;

	//get shortened version of text for list overlay , in case it is too long
	if(strlen(szBuffer) > g_iMaxShortTextLength)
	{
		strcpy(&szBuffer[g_iMaxShortTextLength-3],"...");
		pxEntry->m_xShortText = szBuffer;
	}
	s_xMessageList.add_entry(pxEntry);

	//make sure there aren't too many messages
	if(s_xMessageList.size() > g_uMaxMessagesToSave)
	{
		while(s_xMessageList.size() > g_uMaxMessagesToSave/2)
		{
			delete s_xMessageList.first_entry();
			s_xMessageList.delete_first_entry();
		}

		if(OverlayExists())
		{
			UpdateMessageStrings();
		}
	}
}

void Tools_MessageLog::DeleteMessages(const char* szCat,const char* szSubCat)
{
	u_int uDeleted = 0;
	for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();)
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIF();
		if(szCat && *szCat && !pxMessage->m_xCategory.IsBlank())
		{
			if(pxMessage->m_xCategory.Compare(szCat))
			{
				//category doesn't match
				xMessageLIF.next();
				continue;
			}
		}

		if(szSubCat && *szSubCat && !pxMessage->m_xSubCategory.IsBlank())
		{
			if(pxMessage->m_xSubCategory.Compare(szSubCat))
			{
				//subcategory doesn't match
				xMessageLIF.next();
				continue;
			}
		}

		//message should be deleted
		delete pxMessage;
		xMessageLIF.delete_current();
		uDeleted++;
	}

	if(uDeleted && OverlayExists())
	{
		UpdateMessageStrings();
	}
}

void Tools_MessageLog::DeleteMessages(Category eCat,const char* szSubCat)
{
	if ( eCat < NO_CATEGORY || NUM_CATEGORIES <= eCat )
	{
		eCat = NO_CATEGORY;
	}
	DeleteMessages( s_aszCatStrings[ eCat ], szSubCat );
}

void Tools_MessageLog::DeleteAllMessages()
{
	//clear out the message list
	u_int uDeleted = 0;
	while(s_xMessageList.size())
	{
		delete s_xMessageList.first_entry();
		s_xMessageList.delete_first_entry();

		uDeleted++;
	}

	if(uDeleted && OverlayExists())
	{
		UpdateMessageStrings();
	}
}

void Tools_MessageLog::DeleteVisibleMessages()
{
	//clear out all messages currently being shown
	u_int uDeleted = 0;
	for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();)
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIF();
		if(pxMessage->m_bShown)
		{
			//message should be deleted
			delete pxMessage;
			xMessageLIF.delete_current();
			uDeleted++;
		}
		else
		{
			xMessageLIF.next();
		}
	}

	if(uDeleted && OverlayExists())
	{
		UpdateMessageStrings();
	}
}

void Tools_MessageLog::TransferModeData( const int iMode )
{
    Tools_ModeDataFile xFile( "MessageLog.dat", iMode );

	if( false == xFile.GetError())
	{
		xFile.Transfer( s_xPosition );
		xFile.Transfer( s_uShowOverlayOptions );
		xFile.Transfer( g_xUserSetListSize );
	}
}

void Tools_MessageLog::ShowOverlay()
{
	if(!Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_MESSAGELOG))
	{
		//Default to showing the end of the list when opening the overlay		
		s_uMostRecentSelectedID = 0xffffffff;
		
		CreateOverlay();
	}
	Asura_Overlay_System::BringOverlayToTop(TOOLS_OVERLAY_MESSAGELOG);	
}

void Tools_MessageLog::ShowOverlay(const char* szCat,const char* szSubCat,u_int uInitialSelectedID /*= 0*/)
{
	//initialise filters
	s_xCatFilter = szCat;
	s_xSubCatFilter = szSubCat;
	s_xSearchString = "";
	//Set initial selection
	if(uInitialSelectedID)
	{
		s_uMostRecentSelectedID = uInitialSelectedID;
	}
	else if(!OverlayExists())
	{
		s_uMostRecentSelectedID = 0xffffffff;
	}
	

	if(OverlayExists())
	{
		//if overlay already existed , then we need to force an update of the various lists
		UpdateCatList();
		UpdateSubCatList();
		UpdateMessageStrings();

		Asura_Overlay_System::BringOverlayToTop(TOOLS_OVERLAY_MESSAGELOG);	
	}
	else
	{
		CreateOverlay();
	}
}

void Tools_MessageLog::ShowOverlay(Category eCat,const char* szSubCat,u_int uInitialSelectedID /*= 0*/)
{
	if ( eCat < NO_CATEGORY || NUM_CATEGORIES <= eCat )
	{
		eCat = NO_CATEGORY;
	}

	ShowOverlay( s_aszCatStrings[ eCat ], szSubCat, uInitialSelectedID );
}

void Tools_MessageLog::CloseOverlay()
{
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MESSAGELOG);
}

bool Tools_MessageLog::OverlayExists()
{
	return Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_MESSAGELOG);
}

void Tools_MessageLog::CreateOverlay()
{
	Asura_Overlay* pxOverlay = new Tools_Overlay_MessageLog("Message Log",false,&s_xPosition);
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset(1.0f,3.0f);

	//start of tabs
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset,0,0);
	pxOverlay->AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	pxOverlay->SetCurrentParentDialog(pxTabsDialog);

	float fTabsStartY = xOffset.y;
	float fTabsEndY = xOffset.y;

	//Main tab
	pxTabsDialog->AddTab("Main");
	xOffset.y = fTabsStartY;


	s_pxCatListDialog = 0;
	float fMaxPrintY = Asura_Viewport::GetHeight() / ASURA_TEXT_OUTPUT_FONT_HEIGHT;
	const float fSomeMagicNumberConstant = 24.0f;
	const int _iMaxLines = Asura_Maths::FloatToInt_Chop((fMaxPrintY - fSomeMagicNumberConstant)/1.2f);
	const int iMaxLines = Asura_Maths::Max( 1, _iMaxLines);

	//add drop down lists for filters
	UpdateCatList();
	s_bCatIndexChanged = false;
	pxDialog = s_pxCatListDialog = new Asura_DropDown_List_Dialog(xOffset,s_pszCatList,&s_iCatIndex,iMaxLines,s_uCatListSize,NULL,&s_bCatIndexChanged,"Category");
	pxDialog->SetHelpText("Limits the list to showing entries using this category");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	UpdateSubCatList();
	s_bSubCatIndexChanged = false;
	pxDialog  = s_pxSubCatListDialog = new Asura_DropDown_List_Dialog(xOffset,s_pszSubCatList,&s_iSubCatIndex,iMaxLines,s_uSubCatListSize,NULL,&s_bSubCatIndexChanged,"Subcategory");
	pxDialog->SetHelpText("Limits the list to showing entries using this subcategory");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	//Add search option
	s_bSearchStringChanged = false;
	if(s_xSearchString.IsNull()) s_xSearchString = "";
	pxDialog = new Asura_String_Dialog("Find",xOffset,&s_xSearchString,30,&s_bSearchStringChanged);
	pxDialog->SetHelpText("Limits the list to showing those entries that contain the specified string");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();


	UpdateMessageStrings();
	s_bMessageIndexChanged = false;
	pxDialog = s_pxMessageListDialog = new Asura_List_Dialog(xOffset,s_pszMessageStrings,&s_iMessageIndex,iMaxLines,s_uMessageStringsSize,GotoMessage,&s_bMessageIndexChanged);
	s_pxMessageListDialog->SetDoubleClickFunction(GotoMessage,0);
	s_pxMessageListDialog->SetListHelpStrings(s_pszMessageHelpStrings);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += iMaxLines*1.2f + 1.0f;


	pxDialog = new Asura_Function_Dialog("Delete All",xOffset,DeleteAllMessages);
	pxDialog->SetHelpText("Deletes all messages from the log.");
	pxDialog->SetRelativePosition(s_pxMessageListDialog, true, false);
	pxOverlay->AddDialog(pxDialog);

	xOffset.x = pxDialog->m_xMaxCoord.x + 1.0f;
	pxDialog = new Asura_Function_Dialog("Delete Visible",xOffset,DeleteVisibleMessages);
	pxDialog->SetHelpText("Deletes all messages that are visible using the current filters.");
	pxDialog->SetRelativePosition(s_pxMessageListDialog, true, false);
	pxOverlay->AddDialog(pxDialog);

	xOffset.x = pxDialog->m_xMaxCoord.x + 1.0f;
	pxDialog = new Asura_Function_Dialog("Copy To Clipboard",xOffset,CopyToClipboard);
	pxDialog->SetHelpText("Copies the visible contents of the message log to the windows clipboard.");
	pxDialog->SetRelativePosition(s_pxMessageListDialog, true, false);
	pxOverlay->AddDialog(pxDialog);
	xOffset.x = 1.0f;
	xOffset.y += Asura_Dialog::LargeSpacing();
	
	
	pxDialog = new Asura_Function_Dialog("Go To",xOffset,GotoMessage);
	pxDialog->SetIsEnabledFunction(IsEnabled_GotoMessage);
	pxDialog->SetHelpText("Go to whatever the selected message refers to ,if implemented for that message.\nRight clicking on the list has the same effect.");
	pxDialog->SetRelativePosition(s_pxMessageListDialog, true, false);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	//End of main tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//Options tab
	pxTabsDialog->AddTab("Options");
	xOffset.y = fTabsStartY;

	pxDialog = new Asura_Dialog("Show overlay for :",xOffset);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Missing textures",xOffset,&s_uShowOverlayOptions,TOOLS_MESSAGELOG_TEXTURES);
	pxDialog->SetHelpText("Automatically show this overlay for missing textures.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Sound",xOffset,&s_uShowOverlayOptions,TOOLS_MESSAGELOG_SOUND);
	pxDialog->SetHelpText("Automatically show this overlay for sound event errors.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Dialogue",xOffset,&s_uShowOverlayOptions,TOOLS_MESSAGELOG_DIALOGUE);
	pxDialog->SetHelpText("Automatically show this overlay for dialogue errors.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Axon",xOffset,&s_uShowOverlayOptions,TOOLS_MESSAGELOG_AXON);
	pxDialog->SetHelpText("Automatically show this overlay for axon errors.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Cutscene",xOffset,&s_uShowOverlayOptions,TOOLS_MESSAGELOG_CUTSCENE);
	pxDialog->SetHelpText("Automatically show this overlay for cutscene errors.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Entities",xOffset,&s_uShowOverlayOptions,TOOLS_MESSAGELOG_ENTITIES);
	pxDialog->SetHelpText("Automatically show this overlay for entity errors.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();



	//End of options tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//End of tabs
	xOffset.y = fTabsEndY + 1.0f;
	pxOverlay->SetCurrentParentDialog(0);
	pxTabsDialog->SelectTab(0);

	//add control for resizing the main message log list dialog
	pxDialog = new Asura_Resize_Dialog(s_pxMessageListDialog, &g_xUserSetListSize);
	pxOverlay->AddDialog(pxDialog);
	

	pxOverlay->m_uID = TOOLS_OVERLAY_MESSAGELOG;
	Asura_Overlay_System::SubmitNewOverlay( pxOverlay);
}

void Tools_MessageLog::UpdateCatList()
{
	DeleteCatList();

	//build list of categories used by all messages
	for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();xMessageLIF.next())
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIF();
		if(!pxMessage->m_xCategory.IsBlank())
		{
			int iJunkIndex;
			s_xCatNameTree.FindNodeIndexByName(pxMessage->m_xCategory.GetString(),true,iJunkIndex);
		}
	}

	//Include the currently selected category as well
	if(!s_xCatFilter.IsBlank())
	{
		int iJunkIndex;
		s_xCatNameTree.FindNodeIndexByName(s_xCatFilter.GetString(),true,iJunkIndex);
	}

	//Get a sorted list ,adding an entry for all cats
	s_uCatListSize = s_xCatNameTree.GetNumberOfNodes()+1;
	s_pszCatList = new const char*[s_uCatListSize];
	s_pszCatList[0] = g_szFilterNone;
	s_xCatNameTree.MakeSortedNameList(s_pszCatList+1);

	//find the current selection in the list
	s_iCatIndex = 0;
	if(!s_xCatFilter.IsBlank())
	{
		for(u_int i=1;i<s_uCatListSize;i++)
		{
			if(!s_xCatFilter.Compare(s_pszCatList[i]))
			{
				s_iCatIndex = i;
				break;
			}
		}
		if(!s_iCatIndex) s_xCatFilter.MakeNull();
	}

	//update list if it exists
	if(s_pxCatListDialog && OverlayExists())
	{
		s_pxCatListDialog->ReplaceStrings(s_uCatListSize,s_pszCatList);
        s_pxCatListDialog->SelectEntry( s_iCatIndex );
	}
}

void Tools_MessageLog::DeleteCatList()
{
	s_xCatNameTree.Reset();
	delete [] s_pszCatList;
	s_pszCatList = 0;
	s_uCatListSize = 0;
}

void Tools_MessageLog::UpdateSubCatList()
{
	DeleteSubCatList();

	//build list of sub-categories in the current category
	for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();xMessageLIF.next())
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIF();
		if(!pxMessage->m_xSubCategory.IsBlank())
		{
			if(s_xCatFilter.IsBlank() || !s_xCatFilter.Compare(pxMessage->m_xCategory))
			{
				int iJunkIndex;
				s_xSubCatNameTree.FindNodeIndexByName(pxMessage->m_xSubCategory.GetString(),true,iJunkIndex);
			}
		}
	}

	//Get a sorted list ,adding an entry for all cats
	s_uSubCatListSize = s_xSubCatNameTree.GetNumberOfNodes()+1;
	s_pszSubCatList = new const char*[s_uSubCatListSize];
	s_pszSubCatList[0] = g_szFilterNone;
	s_xSubCatNameTree.MakeSortedNameList(s_pszSubCatList+1);

	//find the current selection in the list
	s_iSubCatIndex = 0;
	if(!s_xSubCatFilter.IsBlank())
	{
		for(u_int i=1;i<s_uSubCatListSize;i++)
		{
			if(!s_xSubCatFilter.Compare(s_pszSubCatList[i]))
			{
				s_iSubCatIndex = i;
				break;
			}
		}
		if(!s_iSubCatIndex) s_xSubCatFilter.MakeNull();
	}

	//update list if it exists
	if(s_pxSubCatListDialog && OverlayExists())
	{
		s_pxSubCatListDialog->ReplaceStrings(s_uSubCatListSize,s_pszSubCatList);
        s_pxSubCatListDialog->SelectEntry( s_iSubCatIndex );
	}
}

void Tools_MessageLog::DeleteSubCatList()
{
	s_xSubCatNameTree.Reset();
	delete [] s_pszSubCatList;
	s_pszSubCatList = 0;
	s_uSubCatListSize = 0;
}


void Tools_MessageLog::UpdateMessageStrings()
{
	DeleteMessageStrings();

	char szSearchString[50] = "";
	if(!s_xSearchString.IsBlank())
	{
		Asura_PrintfToBfr(szSearchString,"*%s*",sizeof(szSearchString),true,s_xSearchString.GetString());
	}

	//determine which messages to show
	s_uMessageStringsSize = 0;
	for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();xMessageLIF.next())
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIF();
		pxMessage->m_bShown = false;

		if(!s_xCatFilter.IsBlank())
		{
			if(pxMessage->m_xCategory.IsBlank()) continue;
			if(s_xCatFilter.Compare(pxMessage->m_xCategory)) continue;
		}

		if(!s_xSubCatFilter.IsBlank())
		{
			if(pxMessage->m_xSubCategory.IsBlank()) continue;
			if(s_xSubCatFilter.Compare(pxMessage->m_xSubCategory)) continue;
		}

		//check search string
		if(!s_xSearchString.IsBlank())
		{
			if(!Asura_StringUtil::CompareAgainstWildcardString(pxMessage->m_xText.GetString(),szSearchString,true)) continue;
		}

		pxMessage->m_bShown = true;
		s_uMessageStringsSize++;
	}

 	s_iMessageIndex = 0;
	if(s_uMessageStringsSize)
	{
		//allocate array for the strings and IDs
		s_pszMessageStrings = new const char*[s_uMessageStringsSize];
		s_pszMessageHelpStrings = new const char*[s_uMessageStringsSize];
		s_ppxMessages = new Tools_MessageLogEntry*[s_uMessageStringsSize];

		//go over messages again , copying into the arrays.
		//also determine the current selection
		u_int uIndex = 0;
		for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();xMessageLIF.next())
		{
			Tools_MessageLogEntry* pxMessage = xMessageLIF();
			if(!pxMessage->m_bShown) continue;

			s_pszMessageStrings[uIndex] = pxMessage->m_xShortText.IsBlank() ? pxMessage->m_xText.GetString() : pxMessage->m_xShortText.GetString();
			s_pszMessageHelpStrings[uIndex] = pxMessage->m_xText.GetString();
			s_ppxMessages[uIndex] = pxMessage;

			if(pxMessage->m_uID <= s_uMostRecentSelectedID)
			{
				s_iMessageIndex = (int)uIndex;
			}

			uIndex++;
		}
	}
	else
	{
		//empty list
		s_uMessageStringsSize = 1;
		
		s_pszMessageStrings = new const char*[1];
		s_pszMessageStrings[0] = "*No Messages*";
		s_pszMessageHelpStrings = new const char*[1];
		s_pszMessageHelpStrings[0] = "";
		s_ppxMessages = 0;		

	}

	//update list if it exists
	if(s_pxMessageListDialog && OverlayExists())
	{
		s_pxMessageListDialog->ReplaceStrings(s_uMessageStringsSize,s_pszMessageStrings);
		s_pxMessageListDialog->SetListHelpStrings(s_pszMessageHelpStrings);
	}

	s_uNextIDWhenStringsWereLastUpdated = s_uNextID;
}

void Tools_MessageLog::DeleteMessageStrings()
{
	delete [] s_pszMessageStrings;
	s_pszMessageStrings = 0;

	delete [] s_pszMessageHelpStrings;
	s_pszMessageHelpStrings = 0;

	delete [] s_ppxMessages;
	s_ppxMessages = 0;

	s_uMessageStringsSize = 0;
}



void Tools_MessageLog::GotoMessage()
{
	if(!s_ppxMessages) return;
	if(s_ppxMessages[s_iMessageIndex]->m_pxAction)
	{
		s_ppxMessages[s_iMessageIndex]->m_pxAction->Execute();
	}
}

bool Tools_MessageLog::IsEnabled_GotoMessage()
{
	if(!s_ppxMessages) return false;
	return s_ppxMessages[s_iMessageIndex]->m_pxAction != 0;
}

void Tools_MessageLog::CopyToClipboard()
{
	//Get total length of text to copy
	u_int uLength = 0;
	for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();xMessageLIF.next())
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIF();
		if(pxMessage->m_bShown)
		{
			//+2 to add \n\r to the end of each entry
			uLength += pxMessage->m_xText.GetLength() + 2;
		}
	}

	if( OpenClipboard(NULL) )
	{
		HGLOBAL hMem = GlobalAlloc( GMEM_MOVEABLE, uLength+1 );
		if( hMem )
		{
			char* pcDest = reinterpret_cast<char*>( GlobalLock(hMem) );
			*pcDest = 0;

			//copy all the messages into the clipboard memory
			for(LIF<Tools_MessageLogEntry*> xMessageLIF(&s_xMessageList);!xMessageLIF.done();xMessageLIF.next())
			{
				Tools_MessageLogEntry* pxMessage = xMessageLIF();
				if(pxMessage->m_bShown)
				{
					strcpy(pcDest,pxMessage->m_xText.GetString());
					pcDest += pxMessage->m_xText.GetLength();
					//add newline
					strcpy(pcDest,"\r\n");
					pcDest +=2;
				}
			}

			EmptyClipboard();
			SetClipboardData( CF_TEXT, hMem );

			Tools_Timed_Message_System::AddUnloggedMessage(0xffffff,"Message log copied to clipboard.");
		}

		CloseClipboard();
	}
}

bool Tools_MessageLog::AreThereAnyMessages(const char* szCat, const char* szSubCat, u_int uMinID /*= 0*/)
{
	for(LIB<Tools_MessageLogEntry*> xMessageLIB(&s_xMessageList);!xMessageLIB.done();xMessageLIB.next())
	{
		Tools_MessageLogEntry* pxMessage = xMessageLIB();
		pxMessage->m_bShown = false;

		//make sure the message ID is high enough
		if(pxMessage->m_uID < uMinID)
		{
			//messages are created and added in ascending ID order , so no point in searching further
			return false;
		}

		//If a category has been specified  , make sure the message is in that category
		if(szCat && *szCat)
		{
			if(pxMessage->m_xCategory.IsBlank()) continue;
			if(pxMessage->m_xCategory.Compare(szCat)) continue;
		}

		//and for subcategory
		if(szSubCat && *szSubCat)
		{
			if(pxMessage->m_xSubCategory.IsBlank()) continue;
			if(pxMessage->m_xSubCategory.Compare(szSubCat)) continue;
		}

		//found a message that qualifies
		return true;
	}
	return false;
}

bool Tools_MessageLog::AreThereAnyMessages(Category eCat, const char* szSubCat, u_int uMinID /*= 0*/)
{
	if ( eCat < NO_CATEGORY || NUM_CATEGORIES <= eCat )
	{
		eCat = NO_CATEGORY;
	}
	return AreThereAnyMessages(s_aszCatStrings[ eCat ], szSubCat, uMinID);
}
