//
// Tools_Module_Selection.cpp
//
// Created by Richard Rice - 15/10/2002
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Debugging/Asura_Assert.h"

#include "Misc/Tools_Module_Selection.h"

#include "Rendering/Asura_Module.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_Cursor.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Mesh/Asura_Env_System.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Mesh.h"
#include "Interface/Asura_Dialog.h"


#include "Misc/Tools_Prompted_Selection.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_LOS.h"

#include "Tools_Overlays.h"
#include "Misc/Tools_Help.h"

#include "Misc/Tools_Render.h"
#include "Misc/Tools_Render_Options.h"
#include "Misc/Tools_StringUtil.h"
#include "Misc/Tools_Modal_Overlay.h"

///////////////////////////////////////////////////////////////////////////////////////
/*--------------**
** Help Strings **
**--------------*/

static const char* g_apszModuleSelectionHelpStrings [] =
{
	"LMB :         Select module",
	"Shift+LMB :   Deselect module",
	"",
	"L :           Clear selection",
	"",
	"Shift+Z :     Undo selection",
	"Shift+Y :     Redo selection",
	"Shift+< :     Undo selection",
	"Shift+> :     Redo selection",
	NULL,
};


////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Tools_Module_Selection **
**------------------------*/

Tools_Module_Selection* Tools_Module_Selection::s_pxThis = 0;
									 
Tools_Module_Selection::Tools_Module_Selection(bool bAllowMultipleSelect)
{
	s_pxThis = this;

	m_bAllowMultipleSelect = bAllowMultipleSelect;
	m_bSelectionOverlayOpen = true;
	m_bButtonDownForSelection = false;
	m_bShowSelection = true;
	m_iHighlitModule = -1;
	m_iSelectionMode = 0;
	m_iSelectionModeListOffset = 0;
	m_bIsActive = false;
	m_uSelectionOverlayMinimised = 0;

	m_xSelectionOverlayPosition = Asura_Vector_2(3.0f,6.0f);

	//load values from .dat file
	TransferModeData(TOOLSMODEDATAFILE_LOAD);

	m_xSelectionUndo.Initialise();


}

Tools_Module_Selection::~Tools_Module_Selection()
{
	//save values to .dat file
	TransferModeData(TOOLSMODEDATAFILE_SAVE);

	//unhide any modules that were hidden
	Unhide();

	s_pxThis = 0;

}

void Tools_Module_Selection::Activate()
{
	Asura_Assert(s_pxThis,"");

	Tools_Render_Options::ActivateHideFlag(ToolsRenderFlag_ModuleSelectionHide);

	s_pxThis->CreateSelectionOverlay();

	//add help
	Tools_Help::AddHelpTopic("Module Selection",g_apszModuleSelectionHelpStrings);

	s_pxThis->m_bIsActive = true;


}

void Tools_Module_Selection::Deactivate()
{
	//remove help
	Tools_Help::RemoveHelpTopic("Module Selection");
	//unhide any hidden modules
	Tools_Render_Options::DeactivateHideFlag(ToolsRenderFlag_ModuleSelectionHide);
	//close the overlay
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MODULE_SELECTION);

	s_pxThis->m_bIsActive = false;

}

int Tools_Module_Selection::GetSelected()
{
	Asura_Assert(s_pxThis,"");
	if(s_pxThis->m_xSelectedModules.size())
	{
		return s_pxThis->m_xSelectedModules.first_entry();
	}
	return -1;
}

int Tools_Module_Selection::GetHighlitModule()
{
	Asura_Assert(s_pxThis,"");
	return s_pxThis->m_iHighlitModule;
}

int Tools_Module_Selection::GetNumSelected()
{
	Asura_Assert(s_pxThis,"");
	return s_pxThis->m_xSelectedModules.size();
}


bool Tools_Module_Selection::IsSelected(int iModuleIndex)
{
	Asura_Assert(s_pxThis,"");
	return s_pxThis->m_xSelectedModules.contains(iModuleIndex);
}

const List<int>* Tools_Module_Selection::GetSelectionList()
{
	Asura_Assert(s_pxThis,"");
	return &s_pxThis->m_xSelectedModules;

}

Asura_Vector_3 Tools_Module_Selection::GetCentreOfSelection()
{
	if(!GetNumSelected()) return Asura_Maths::ZeroVector3;

	Asura_Bounding_Box xBB;
	xBB.Reset();

	for(CLIF<int> xSelectedLIF(GetSelectionList());!xSelectedLIF.done();xSelectedLIF.next())
	{
		//get the BB for each selected module
		Asura_Vector_3 xPosition = Asura_Module_System::GetPositionOfModule(xSelectedLIF());
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xSelectedLIF());

		Asura_Bounding_Box xModBB = pxMesh->GetBoundingBox();
		xModBB.Displace(xPosition);

		//expand the overall BB
		xBB.Set(xModBB);
	}

	return xBB.CalculateCentre();
}

void Tools_Module_Selection::TransferModeData(int iMode)
{
	//open the file
	Tools_ModeDataFile xFile("ModuleSelection.dat",iMode);

	//tranfer the data
	//(none of these should be removed , and new data should go on the end, otherwise
	//the files will get screwed up)
	bool bJunk = false;
	if( false == xFile.GetError())
	{
		xFile.Transfer(m_xSelectionOverlayPosition);
		xFile.Transfer(bJunk);
		xFile.Transfer(m_uSelectionOverlayMinimised);
	}
}


void Tools_Module_Selection::Initialise(bool bAllowMultipleSelect /*= true*/)
{
	Asura_Assert(!s_pxThis,"Module selection has already been initialised");
	new Tools_Module_Selection(bAllowMultipleSelect);
}

void Tools_Module_Selection::Shutdown()
{
	Asura_Assert(s_pxThis,"");
	delete s_pxThis;

}

void Tools_Module_Selection::UpdateMain()
{
	//Flag the selection undo to start a new undo batch
	m_xSelectionUndo.MarkUndoStart();


	//if the mouse button has been release , reset the ButtonHeldForSelection flag
	if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
	{
		m_bButtonDownForSelection = false;
	}

	m_iHighlitModule = -1;

	if(!Tools_Prompted_Selection::IsSelecting())
	{
		//act upon keyboard input
		HandleKeyboard();

		//and handle selection using the mouse
		HandleMouseSelection();
	}
}

void Tools_Module_Selection::RenderMain( void ) const
{
    bool bCursorModuleIsSelected = false;
	if( m_bShowSelection )
    {
	    for(CLIF<int> xSelectedLIF(&m_xSelectedModules);!xSelectedLIF.done();xSelectedLIF.next())
	    {
            const int iNextModule = xSelectedLIF();

            if( iNextModule == m_iHighlitModule )
            {
                bCursorModuleIsSelected = true;

                if( Tools_Render::HighlightPolyOrObjectUnderCursor )
                {
                    continue;
                }
            }

		    Tools_Render::RenderMeshOutline( iNextModule, 0xffff0000 );
	    }
    }
	
	if( Tools_Render::HighlightPolyOrObjectUnderCursor && m_iHighlitModule >= 0 )
    {
	    float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
	    fAlpha = 255.0f * fAlpha * fAlpha;

        // If module is both selected and has the cursor over it, render in
        // light blue (ie. mixture of blue for selected and white for the
        // cursor being over it.)
        u_int uColour = (bCursorModuleIsSelected ? 0xbbbbff : 0xffffff) | (Asura_Maths::FloatToInt( fAlpha ) << 24);

	    Tools_Render::RenderMeshOutline( m_iHighlitModule, uColour );
    }
}

void Tools_Module_Selection::ToggleSelectionOverlay(void)
{
	Asura_Assert(s_pxThis,"");
	s_pxThis->m_bSelectionOverlayOpen = !s_pxThis->m_bSelectionOverlayOpen;

	if(s_pxThis->m_bSelectionOverlayOpen)
	{
		s_pxThis->m_uSelectionOverlayMinimised = 0;
		s_pxThis->CreateSelectionOverlay();
	}
	else
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MODULE_SELECTION);
	}
}

bool Tools_Module_Selection::SelectionOverlayIsOn(void)
{
	Asura_Assert(s_pxThis,"");
	return s_pxThis->m_bSelectionOverlayOpen;
}

void Tools_Module_Selection::CreateSelectionOverlay()
{
	if(!m_bSelectionOverlayOpen) return;

	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Selection",false,&m_xSelectionOverlayPosition,&m_uSelectionOverlayMinimised);
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);

	if(m_bAllowMultipleSelect)
	{
		//strings for selection mode options
		static char* s_apszSelectionModeStrings[] =
		{
			"Object",
			"Region",
		};

	
		//heading for selection mode list box
		pxDialog = new Asura_Dialog("Selection Mode :",xOffset);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=1.0f;

		//selection mode list box
		pxDialog = new Asura_List_Dialog(xOffset,s_apszSelectionModeStrings,&m_iSelectionModeListOffset,&m_iSelectionMode,2,2,NULL);
		pxDialog->SetHelpText("Object: Select one module at a time.\nRegion : Select all module in the same region.");
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=3.4f;
	}
	
	pxDialog = new Asura_Tickbox_Dialog("Show Selection (Alt+F2)",xOffset,&m_bShowSelection);
	pxDialog->SetHelpText("Highlight selected modules with a blue outline");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	if(m_bAllowMultipleSelect)
	{
		pxDialog = new Asura_Function_Dialog("Select All",xOffset,_SelectAll);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();

		pxDialog = new Asura_Function_Dialog("Clear Selection (L)",xOffset,_DeselectAll);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();

		pxDialog = new Asura_Function_Dialog("Select By Name",xOffset,_SelectByName);
		pxDialog->SetHelpText("Selects all modules with names matching a given search string.\n* can be used as a wildcard in the search string.");
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();

		pxDialog = new Asura_Function_Dialog("Invert Selection",xOffset,_InvertSelection);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();
	}

	pxDialog = new Asura_Function_Dialog("Hide Selected",xOffset,_HideSelection);
	pxDialog->SetHelpText("Hides all selected modules , and deselects them.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Hide Unselected",xOffset,_HideUnselected);
	pxDialog->SetHelpText("Hides any modules that aren't selected.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("Unhide all",xOffset,_Unhide);
 	pxDialog->SetHelpText("Unhides all previously hidden modules.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	if(m_bAllowMultipleSelect)
	{
		pxDialog = new Asura_Function_Dialog("Unhide and select",xOffset,_UnhideAndSelect);
		pxDialog->SetHelpText("Unhides and selects all previously hidden modules.");
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();
	}
	
	pxDialog = new Asura_Function_Dialog("Undo (Shift+Z)",xOffset,_Undo);
	pxDialog->SetHelpText("Module selection undo.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Redo (Shift+Y)",xOffset,_Redo);
	pxDialog->SetHelpText("Module selection redo.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	
	pxOverlay->m_uID = TOOLS_OVERLAY_MODULE_SELECTION;

	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}


void Tools_Module_Selection::HandleKeyboard()
{
	if(Asura_Overlay_System::OverlaysHaveKeyboardFocus()) return;
	
	if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
	{
		if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_F2))
		{
			m_bShowSelection = !m_bShowSelection;
		}
		
	}
	else if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
	{
	}
	else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
	{
		if( Asura_UserInput::DebouncedKeyPressed( 'z' ) || Asura_UserInput::DebouncedKeyPressed( ',' )) // 'shift+z' or 'shift+<' for selection undo
		{
			Undo();
		}
		if ( Asura_UserInput::DebouncedKeyPressed( 'y' ) || Asura_UserInput::DebouncedKeyPressed( '.' )) // 'shift+y' or 'shift+>' for selection redo
		{
			Redo();
		}
	}
	else
	{
		if(Asura_UserInput::DebouncedKeyPressed('l'))// 'L' for clear selection
		{
			DeselectAll();
		}
	}
}

void Tools_Module_Selection::HandleMouseSelection()
{
	// If the user is doing anything with overlays or menus,
	// ignore any mousebutton/key presses etc. 
	if (Asura_Interface::UsingMenusOrOverlays()) return;
	
	
	
	//see if the mouse is over a polygon
	int iModuleIndex;
  	bool bOverModule = Tools_LOS::FindModuleUnderCursor(iModuleIndex);
	
	if (bOverModule) 		
	{
		//remember this module, so we can highlight it during Render()			
		m_iHighlitModule = iModuleIndex;
	}
	
	//if over a module , check for selection/deselection
	if (Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
	{
		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) || m_bButtonDownForSelection)
		{
			m_bButtonDownForSelection = true;
			if (bOverModule)
			{
				
				if (Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
				{
					switch(m_iSelectionMode+m_iSelectionModeListOffset)
					{
						case ModuleSelectionMode_Object :
							Deselect(iModuleIndex);
							break;
						case ModuleSelectionMode_Region :
							DeselectRegion(iModuleIndex);
							break;
					}
				}
				else
				{
					switch(m_iSelectionMode+m_iSelectionModeListOffset)
					{
						case ModuleSelectionMode_Object :
							Select(iModuleIndex,m_bAllowMultipleSelect);
							break;

						case ModuleSelectionMode_Region :
							SelectRegion(iModuleIndex);
							break;
					}
				}
			}

		}
	}
}


void Tools_Module_Selection::SelectAll()
{
	Asura_Assert(m_bAllowMultipleSelect,"");
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();

	for(int i=0;i<iNumberOfModules;i++)
	{
		Select(i,true);
	}

}

void Tools_Module_Selection::DeselectAll()
{
	while(m_xSelectedModules.size())
	{
		Deselect(m_xSelectedModules.first_entry());
	}
}

void Tools_Module_Selection::InvertSelection()
{
	Asura_Assert(m_bAllowMultipleSelect,"");

	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		if(IsSelected(i))
		{
			Deselect(i);
		}
		else
		{
			Select(i,true);
		}
	}
}

void Tools_Module_Selection::SelectRegion(int iModuleIndex)
{
	Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
	unsigned uRegion = pxEnv->GetModuleRegion(iModuleIndex);

	//select all modules that share this region
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		if(pxEnv->GetModuleRegion(i)==uRegion)
		{
			Select(i,true);
		}
	}
}

void Tools_Module_Selection::DeselectRegion(int iModuleIndex)
{
	Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
	unsigned uRegion = pxEnv->GetModuleRegion(iModuleIndex);

	//deselect all modules that share this region
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		if(pxEnv->GetModuleRegion(i)==uRegion)
		{
			Deselect(i);
		}
	}
}




void Tools_Module_Selection::Select(int iModuleIndex,bool bMultiSelect)
{
	
	//don't allow selection of hidden modules
	if(Tools_Render_Options::IsHidden(iModuleIndex) || Tools_Render_Options::AreAllPolygonsInModuleHidden(iModuleIndex))
	{
		return;
	}
	
	if(!m_xSelectedModules.contains(iModuleIndex))
	{
		m_xSelectionUndo.LogModuleSelected(iModuleIndex);

		if(!bMultiSelect)
		{
			DeselectAll();
		}

		m_xSelectedModules.add_entry(iModuleIndex);
	}
}

void Tools_Module_Selection::Deselect(int iModuleIndex)
{
	if(m_xSelectedModules.contains(iModuleIndex))
	{
		m_xSelectionUndo.LogModuleDeselected(iModuleIndex);

		m_xSelectedModules.delete_entry(iModuleIndex);
	}
}


void Tools_Module_Selection::HideSelection()
{
	//go through the selection hiding everything
	for(LIF<int> xSelectionLIF(&m_xSelectedModules);!xSelectionLIF.done();xSelectionLIF.next())
	{
		Tools_Render_Options::AddRenderFlagModule(xSelectionLIF(),ToolsRenderFlag_ModuleSelectionHide);
	}

	//clear the selection
	DeselectAll();
}

void Tools_Module_Selection::HideUnselected()
{
	//Hide anything not in the selection list
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();

	for(int i=0;i<iNumberOfModules;i++)
	{
		if(!IsSelected(i))
		{
			Tools_Render_Options::AddRenderFlagModule(i,ToolsRenderFlag_ModuleSelectionHide);
		}
	}

	//clear the selection
	DeselectAll();
}


void Tools_Module_Selection::Unhide() const
{
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		Tools_Render_Options::RemoveRenderFlagModule(i,ToolsRenderFlag_ModuleSelectionHide);
	}
}


void Tools_Module_Selection::UnhideAndSelect()
{
	Asura_Assert(m_bAllowMultipleSelect,"");

	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		if(Tools_Render_Options::GetRenderFlags(i) & ToolsRenderFlag_ModuleSelectionHide)
		{
			//unhide this module
			Tools_Render_Options::RemoveRenderFlagModule(i,ToolsRenderFlag_ModuleSelectionHide);
			//and select it
			Select(i,true);

		}
	}
}

void Tools_Module_Selection::Undo()
{

	if(m_xSelectionUndo.Undo())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Selection Undo");
	}
}
void Tools_Module_Selection::Redo()
{
	if(m_xSelectionUndo.Redo())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Selection Redo");
	}
}

void Tools_Module_Selection::DeleteModuleMain(int iModuleIndex)
{
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();

	//deselect the module being deleted
	Deselect(iModuleIndex);

	if(IsSelected(iNumberOfModules-1))
	{
		//this module gets renumbered... so deselect and select again with its new index
		Deselect(iNumberOfModules-1);
		Select(iModuleIndex,true);
	}

	//undo will be invalid
	m_xSelectionUndo.FlushUndoBuffer();
}

void Tools_Module_Selection::SelectByName()
{
	Asura_Assert(m_bAllowMultipleSelect,"");

	//ask for name
	static char s_szName[100] = "";

	GetString_Overlay xOverlay("Enter search string for module(s) to select.",s_szName);
	if(!xOverlay.DoModal()) return;

	strcpy(s_szName,xOverlay.GetString());

	int iFound = 0;
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		//have a match?
		if(Tools_StringUtil::WildcardMatch(pxMesh->Name.GetString(),s_szName))
		{
			Select(i,true);
			iFound++;
		}
	}

	Tools_Timed_Message_System::AddMessage(0xffffff,"Found %d module(s)",iFound);

}



////////////////////////////////////////////////////////////////////////////////////////
/*---------------------**
** ModuleSelectionUndo **
**---------------------*/

void ModuleSelectionUndo::ProcessUndoData(void* pData)
{
	Asura_Assert(Tools_Module_Selection::s_pxThis,"");

	ModuleSelectionUndoData* pxUndoData = (ModuleSelectionUndoData*) pData;
	
	switch(pxUndoData->m_eUndoType)
	{
		case ModuleSelectionUndoData::UndoType_Added :
			//deselect the module
			Tools_Module_Selection::s_pxThis->Deselect(pxUndoData->m_iModuleIndex);
			//convert the undo data
			pxUndoData->m_eUndoType = ModuleSelectionUndoData::UndoType_Removed;
			break;


		case ModuleSelectionUndoData::UndoType_Removed :
			//select the module
			Tools_Module_Selection::s_pxThis->Select(pxUndoData->m_iModuleIndex,true);
			//convert the undo data
			pxUndoData->m_eUndoType = ModuleSelectionUndoData::UndoType_Added;
			break;

		default :
			Asura_Assert(0,"Unexpected undo type for module selection");
			break;
	}
}

void ModuleSelectionUndo::LogModuleSelected(int iModule)
{
	ModuleSelectionUndoData* pxUndo = (ModuleSelectionUndoData*) AllocateNextUndoData();
	if(!pxUndo) return;

	pxUndo->m_eUndoType = ModuleSelectionUndoData::UndoType_Added;
	pxUndo->m_iModuleIndex = iModule;
}

void ModuleSelectionUndo::LogModuleDeselected(int iModule)
{
	ModuleSelectionUndoData* pxUndo = (ModuleSelectionUndoData*) AllocateNextUndoData();
	if(!pxUndo) return;

	pxUndo->m_eUndoType = ModuleSelectionUndoData::UndoType_Removed;
	pxUndo->m_iModuleIndex = iModule;
}




