//
// Tools_ProgressIndicator.cpp
//
// Created by Richard Rice - 13/02/2004
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Misc/Tools_ProgressIndicator.h"


#include "Core/Asura_Timers.h"
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_String.h"
#include "Core/Asura_UserInput.h"
#include "Core/Asura_Windows.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Direct3D_BackBufferCapture.h"



#include "misc/list_tem.hpp"

static List<Tools_ProgressIndicator*> g_xIndicatorList;
static bool g_bInitialised = false;
static u_int g_uFrameStamp = 0xffffffff;
static LARGE_INTEGER g_xTimer;
static u_int g_uMinTicksForFirstFrame = 0;
static u_int g_uMinTicksPerFrame = 0;
static u_int g_uTicksPerSecond = 0;
static bool g_bFirstFrameShown = false;

static const int MAX_PROGRESS_MESSAGE = 15;
static int g_iNextProgressMessage = 0;
static int g_iNumProgressMessage = 0;
static char g_aszProgressMessage[MAX_PROGRESS_MESSAGE][256];

Tools_ProgressIndicator::Tools_ProgressIndicator(const char* pszText /*= 0*/, bool bPreserveScreen /*= true*/)
{
	g_xIndicatorList.add_entry(this);
	m_szText[0] = 0;
	m_uProgress = 0;
	m_uMax = 0;
    m_bShowAsPercentage = true;
	m_bActive = false;
	m_bPreserveScreen = bPreserveScreen;
	m_bHasCapturedScreen = false;
	m_bShowTimeRemaining = false;

	// Reset history
	Asura_Mem::Set( m_axProgressSpeedHistory, 0, sizeof( m_axProgressSpeedHistory ) );

	m_iLastProgressTimeStamp = 0;
	m_iTimeElapsed = 0;

	if(pszText)
	{
		SetText("%s",pszText);
	}
}

Tools_ProgressIndicator::~Tools_ProgressIndicator()
{
	if(m_bActive) Finish();
	g_xIndicatorList.delete_entry(this);

    // If this is the last stacked indicator then clear out any messages
    // that have been added.
    if( !g_xIndicatorList.size() )
    {
        g_iNextProgressMessage = 0;
        g_iNumProgressMessage = 0;
    }
}

void Tools_ProgressIndicator::SetText(const char* pszText,...)
{
	va_list ap;
	va_start(ap,pszText);
    Asura_PrintfToBfr( m_szText, pszText, ap, 256, true );
	va_end(ap);

	m_bActive = true;
}

bool Tools_ProgressIndicator::Update(u_int uProgress /*= 0*/,u_int uMax /*= 0*/,bool bForceRender /*= false*/)
{
	m_uProgress = uProgress;
	m_uMax = uMax;

	UpdateTimers();

	if( m_bShowTimeRemaining )
	{
		CalculateTimeRemaining();
	}

	// if we're going to update the screen, check for windows messages and user input
	// (so people can query user input etc)
	Asura_Windows::CheckForWindowsMessages();
    Asura_UserInput::Update();

	//show on screen if enough time has passed
	Render(bForceRender);

	// Return false if ESC is pressed so caller can quit
	if( Asura_UserInput::HasAnyKeyBeenPressed() )
	{
		if( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) )
		{
			return false;
		}
	}

	return true;
}

void Tools_ProgressIndicator::Finish()
{
	m_bActive = false;
}

void Tools_ProgressIndicator::CalculateTimeRemaining()
{
	LARGE_INTEGER xTimer;
	QueryPerformanceCounter( &xTimer );

	if( m_iLastProgressTimeStamp == 0 )
	{
		// No calculation in first call
		Asura_Assert( m_iTimeElapsed == 0, "Elapsed time not zero on first frame" );
		m_iLastProgressTimeStamp = xTimer.QuadPart;
		m_uLastProgress = m_uProgress;
		return;
	}

	const LONGLONG iTimeSinceLastUpdate = xTimer.QuadPart - m_iLastProgressTimeStamp;
	m_iTimeElapsed += iTimeSinceLastUpdate;

	// Update progress speed, if there was progress
	if( m_uLastProgress < m_uProgress )
	{
		// Update history
		for( int iIdx = NUM_STORED_HISTORY - 1; iIdx > 0; iIdx-- )
		{
			if( m_axProgressSpeedHistory[iIdx - 1].iLength == 0 ) continue;

			// Shift values
			m_axProgressSpeedHistory[iIdx] = m_axProgressSpeedHistory[iIdx - 1];
		}

		m_axProgressSpeedHistory[0].uProgress = m_uProgress - m_uLastProgress;
		m_axProgressSpeedHistory[0].iLength = iTimeSinceLastUpdate;
	}

	// Remaining time calculation
	m_iTimeRemaining = ProcessSpeedHistory( m_uMax - m_uProgress );

	// Update timer & progress vars
	m_iLastProgressTimeStamp = xTimer.QuadPart;
	m_uLastProgress = m_uProgress;
}

LONGLONG Tools_ProgressIndicator::ProcessSpeedHistory( u_int uProgressLeft )
{
	// The algorythm: To predict the time needed for the completion of uProgressLeft work,
	// we'll look back into the past and see how much time it took to complete the same amount
	// of work (multiplied by our heuristic, k). If our history doesn't hold information on
	// that much progress yet, use the total average. This way if progress speed stabilises 
	// after a certain time interval, our prediction will also be correct and stable after some
	// time. (AN 12/05/09)
	//
	// Credit goes to a certain Gareth McCaughan.
	// See http://www.mccaughan.org.uk/g/remarks/time-left.html for more info.

	if( !m_uProgress )
	{
		Asura_Assert( 0, "Can't process speed history with 0 progress" );
		return 0;
	}

	const u_int uHEURISTIC_K_INVERTED = 3; // Apparently, k should be [0.25..1], so 1/k should be [4..1]
	uProgressLeft /= uHEURISTIC_K_INVERTED;

	LONGLONG iTimeRemaining = 0;

	for( int iIdx = 0; iIdx < NUM_STORED_HISTORY; iIdx++ )
	{
		if( m_axProgressSpeedHistory[iIdx].uProgress == 0 ) break; // Zero progress is never stored, this must mean that the array is empty from this point onwards

		// If the amount of progress left lies between interval boundaries...
		if( uProgressLeft < m_axProgressSpeedHistory[iIdx].uProgress )
		{
			// ...only add the amount that remains
			// Using long double for added precisity, this won't add much extra overhead since it'll only happen once for every call of this function. (AN 12/05/09)
			iTimeRemaining += static_cast<LONGLONG>( m_axProgressSpeedHistory[iIdx].iLength * ( static_cast<long double>( uProgressLeft ) / m_axProgressSpeedHistory[iIdx].uProgress ) );
			uProgressLeft = 0;
		}
		else
		{
			// ...else add the whole amount
			iTimeRemaining += m_axProgressSpeedHistory[iIdx].iLength;
			uProgressLeft -= m_axProgressSpeedHistory[iIdx].uProgress;
		}

		if( uProgressLeft == 0 ) break;
	}

	// History wasn't long enough, use total average for remaining progress
	if( uProgressLeft > 0 )
	{
		// Using long double here as well, for the same reason. TODO: avoid code duplication somehow..? (AN 12/05/09)
		iTimeRemaining += static_cast<LONGLONG>( m_iTimeElapsed * ( static_cast<long double>( uProgressLeft ) / m_uProgress ) );
	}

	// Don't forget to multiply our end result with the inverse of k, the heuristic!
	return iTimeRemaining * uHEURISTIC_K_INVERTED;
}

void Tools_ProgressIndicator::UpdateTimers()
{
	LARGE_INTEGER xTimer;
	QueryPerformanceCounter(&xTimer);

	if( !g_bInitialised )
	{
		//first time... get timer information
		LARGE_INTEGER xFrequency;
		QueryPerformanceFrequency(&xFrequency);

		g_uMinTicksForFirstFrame = xFrequency.LowPart * 2; //2 seconds
		g_uMinTicksPerFrame = xFrequency.LowPart / 4; // quarter second
		g_uTicksPerSecond = xFrequency.LowPart;

		g_bInitialised = true;
	}

	u_int uFrame = Asura_Timers::GetFrameNumber();
	if( g_uFrameStamp != uFrame )
	{
		//we've moved onto another frame , so reset the time till we draw anything
		g_uFrameStamp = uFrame;
		g_xTimer = xTimer;
		g_bFirstFrameShown = false;
	}
}

void Tools_ProgressIndicator::Render( bool bForceRender )
{
	if( !Asura_Text_Output::IsInitialised() )
	{
		//text rendering not available yet , so forget about showing anything
		return;
	}

	LARGE_INTEGER xTimer;
	QueryPerformanceCounter( &xTimer );

	if( !bForceRender )
	{
		//has enough time passed for a frame to be shown?
		u_int uTimeSince = static_cast<u_int>( xTimer.QuadPart-g_xTimer.QuadPart );
		if(uTimeSince < (g_bFirstFrameShown ? g_uMinTicksPerFrame : g_uMinTicksForFirstFrame))
		{
			return;
		}
	}

	g_bFirstFrameShown = true;
	g_xTimer = xTimer;

	Tools_ProgressIndicator* pxFirstProgress = g_xIndicatorList.size() ? g_xIndicatorList.first_entry() : 0;
	if(pxFirstProgress && pxFirstProgress->m_bPreserveScreen && !pxFirstProgress->m_bHasCapturedScreen)
	{
		//capture the current screen , so text can be rendered over it
		Asura_Direct3D_BackBufferCapture::CaptureCurrentBackBuffer();
		pxFirstProgress->m_bHasCapturedScreen = true;
	}

	bool bFrameBegun = Asura_Direct3D::HasFrameBegun();
	if(!bFrameBegun)
	{
		Asura_Render::BeginFrame();

#ifdef ASURA_TARGET_SM3
		// End HDR rendering prematurely
		Asura_Render::ApplyToneMapping();
#endif
	}

	//make sure that we aren't in wireframe mode
	Asura_Render::CurrentStates.RequestWireFrameMode(false);


	float fWidth = (float)Asura_Viewport::GetWidth();
	float fHeight = (float)Asura_Viewport::GetHeight();
	
	if(pxFirstProgress && pxFirstProgress->m_bPreserveScreen && !Asura_Direct3D_BackBufferCapture::IsCaptureDirty())
	{

		//render old back buffer as backdrop for the text
		D3DQuadVertex axVerts[4];
		short ausIndeces[6] = {0,1,2,2,3,0};
		for(int i=0;i<4;i++)
		{
			axVerts[i].X = (i==1 || i==2) ? fWidth-0.5f : -0.5f;
			axVerts[i].Y = (i==2 || i==3) ? fHeight-0.5f : -0.5f;
			axVerts[i].Z = 0;
			axVerts[i].RHW = 1000.0f;
			axVerts[i].Diffuse = 0xffffffff;
			axVerts[i].U = (i==1 || i==2) ? 1.0f : 0.0f;
			axVerts[i].V = (i==2 || i==3) ? 1.0f : 0.0f;
		}

	 	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
	 	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_CUSTOM);

		Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_QUADVERTEX);
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
		Asura_Direct3D::D3DDevice->SetTexture(0,Asura_Direct3D_BackBufferCapture::GetCaptureTexture());

		Asura_Direct3D::D3DDevice->DrawIndexedPrimitiveUP
			(
			D3DPT_TRIANGLELIST,
			0,
			4,
			2,
			(void*)ausIndeces,
			D3DFMT_INDEX16,
			(void*)axVerts,
			sizeof(D3DQuadVertex)
			);

		Asura_Render::FlushDisplay();
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
		Asura_Render::CurrentStates.RequestTextureIndex(0, ASURA_TEXTURE_INVALID);
	}
	else
	{
		//black out the screen
		Asura_Render::Overlay.SetRectXY(0,0,fWidth,fHeight);
		Asura_Render::Overlay.SetRectUV(0,0,1,1);
		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
		Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.Colour=0xff000000;
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();
	}

	//show some text
	float fPrintX = 0;
	float fPrintY = fHeight/2;
	Asura_Text_Output::ConvertPixelsToPrintCoords(fPrintX,fPrintY);
	
	for(LIF<Tools_ProgressIndicator*> xIndicatorLIF(&g_xIndicatorList);!xIndicatorLIF.done();xIndicatorLIF.next())
	{
		Tools_ProgressIndicator* pxInd = xIndicatorLIF();

		if(pxInd->m_bActive)
		{
			if(pxInd->m_uMax)
			{
                if( pxInd->m_bShowAsPercentage )
                {
				    //we have a 'max' progress , so show a percent
				    Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%s : %.1f%%",pxInd->m_szText,pxInd->m_uProgress*100.0f/pxInd->m_uMax);
                }
                else
                {
				    Asura_Text_Output::Print( fPrintX, fPrintY++, 0xffffffff, ASURA_TEXT_FORMAT_LEFTJUSTIFIED, "%s : %u / %u", pxInd->m_szText, pxInd->m_uProgress, pxInd->m_uMax );
                }

				if( pxInd->m_bShowTimeRemaining )
				{
					const int iElapsedInSec = static_cast<int>( pxInd->m_iTimeElapsed / g_uTicksPerSecond );
					const int iRemainingInSec = static_cast<int>( pxInd->m_iTimeRemaining / g_uTicksPerSecond );

					Asura_Text_Output::Print( fPrintX, fPrintY++, 0xffffffff, ASURA_TEXT_FORMAT_LEFTJUSTIFIED, " Elapsed: %d:%.2d  Remaining: %d:%.2d",
						iElapsedInSec / 60, iElapsedInSec % 60, iRemainingInSec / 60, iRemainingInSec % 60 );
				}
			}
			else if(pxInd->m_uProgress)
			{
				//only have a progress number , so show it as is
				Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%s : %d",pxInd->m_szText,pxInd->m_uProgress);
			}
			else
			{
				//no numbers at all , so just show the text
				Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%s",pxInd->m_szText);
			}
		}
	}

	if( g_iNumProgressMessage )
	{
		//display other progress messages
		fPrintX = 0;
		fPrintY = fHeight/2;
		Asura_Text_Output::ConvertPixelsToPrintCoords(fPrintX,fPrintY);

		int iIndex = g_iNextProgressMessage - 1;
		for(int i=0;i<g_iNumProgressMessage;i++,iIndex--)
		{
			if(iIndex==-1) iIndex = MAX_PROGRESS_MESSAGE-1;
			Asura_Text_Output::Print(fPrintX,--fPrintY,0xffffffff,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%s",g_aszProgressMessage[iIndex]);
		}
	}

	Asura_Render::EndFrame();

	if(bFrameBegun)
	{
		Asura_Render::BeginFrame();
	}

}

void Tools_ProgressIndicator::AddMessage(const char* pszText,...)
{
	va_list ap;
	va_start(ap,pszText);
    Asura_PrintfToBfr( g_aszProgressMessage[g_iNextProgressMessage], pszText, ap, 256, true );
	va_end(ap);

	g_iNextProgressMessage = (g_iNextProgressMessage+1)%MAX_PROGRESS_MESSAGE;

	if(g_iNumProgressMessage<MAX_PROGRESS_MESSAGE)
	{
		g_iNumProgressMessage++;
	}

	Render(false);
	
}

