//
// Tools_XMLExporter.cpp
//
// Created by HE - 10 May 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Misc/Tools_XMLExporter.h"

// asura
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"

//tools
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"

//libs
#include <errno.h>

const int iTABARRAYSIZE = 100;
const u_int uDEFAULTMAXBUFFERSIZE = 65536;

Tools_XMLExporter::Tools_XMLExporter( const char* szPath, const char* szStyleSheet )
{
	m_iNumTabs = 0;
	m_szTabs = new char[iTABARRAYSIZE];
	//fill the tabs buffer with 0s so IncTabs works correctly
	Asura_Mem::Set(m_szTabs, 0, iTABARRAYSIZE);
	m_uBufferStringLength = 0;

	//buffer
	m_uMaxBufferSize = uDEFAULTMAXBUFFERSIZE;
	m_szBuffer = new char[m_uMaxBufferSize];
	m_pcBufferPos = m_szBuffer;
	Asura_Mem::Set(m_szBuffer, 0, m_uMaxBufferSize);

	m_szPath = szPath ? szPath : "logfiles";

	m_szStyleSheet = szStyleSheet ? szStyleSheet : "";
	Asura_StringUtil::ReplaceAllChars( m_szStyleSheet, '\\', '/' ); 

}

Tools_XMLExporter::~Tools_XMLExporter()
{
	delete[] m_szTabs;
	delete[] m_szBuffer;
}



bool Tools_XMLExporter::WriteXMLToFile( const char* szFileName )
{
	char szFileNameAndPath[128];		//should leave us just over 100 chars for the filename
	Asura_PrintfToBfr(szFileNameAndPath, "%s/%s.xml", 128, true, m_szPath, szFileName );
	
	if(!WriteToFile(szFileNameAndPath))
	{
		//if the write fails, we still need to clear the buffer for the next time round
		Clear();
		return false;
	}

	Clear();
	Tools_Timed_Message_System::AddMessage(0xffffff,"Exported ../%s successfully.", szFileNameAndPath);
	return true;
}


bool Tools_XMLExporter::WriteToFile( char* szFileName )
{
	//do some checking in advance to test for e.g. read-only files
	if ( !Tools_SaveAsura::CanSaveFile(szFileName, true) ) return false;

	FILE* fileHandle = fopen( szFileName, "w+" );
	
	if( !fileHandle )
	{
		if( errno == ENOENT )
		{
			Tools_FileSystem::EnsureDirectoryExists( szFileName );
		}

		fileHandle = fopen( szFileName, "w+" );
	}
	
	if( fileHandle )
	{
		fputs( "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n", fileHandle);
		fputs( m_szStyleSheet, fileHandle);
		fputs( "\n<asura>\n", fileHandle);
		fwrite ( m_szBuffer, sizeof(char), m_pcBufferPos - m_szBuffer, fileHandle );
		fputs( "</asura>", fileHandle);
		fclose( fileHandle );
		return true;
	}
	else
	{
		Asura_Assert(false, "File error");
		return false;
	}
}

void Tools_XMLExporter::IncTabs()
//once we hit the maximum number of tabs to store, we'll stop indenting
{
	++m_iNumTabs;
	if( m_iNumTabs < iTABARRAYSIZE ) m_szTabs[m_iNumTabs-1] = '\t';
}

void Tools_XMLExporter::DecTabs()
{
	//clamp the number of tabs to 0 if things go wrong (which they really shouldn't)
	if( m_iNumTabs <= 0 )
	{
		Asura_Assert(false, "Tab mismatch");
	}
	else
	{
		--m_iNumTabs;
		if( m_iNumTabs < iTABARRAYSIZE ) m_szTabs[m_iNumTabs] = '\0';
	}
}

void Tools_XMLExporter::AddToBuffer( const Asura_Char* szInString )
{
	if(!szInString || strcmp( szInString , "" ) == 0) return;

	size_t uInStringLength = strlen( szInString );

	//if necessary, reallocate the buffer to double the size
	bool bResized = false;
	while( uInStringLength + m_uBufferStringLength >= m_uMaxBufferSize )
	{
		m_uMaxBufferSize *= 2;
		bResized = true;
	}

	if( bResized )
	{
		char* szNewBuffer = new char[m_uMaxBufferSize];
		Asura_Mem::Copy(szNewBuffer, m_szBuffer, m_uBufferStringLength);
		delete[] m_szBuffer;
		m_szBuffer = szNewBuffer;
		m_pcBufferPos = m_szBuffer + m_uBufferStringLength;
	}

	Asura_Mem::Copy(m_pcBufferPos, szInString, uInStringLength);
	m_uBufferStringLength += uInStringLength;
	m_pcBufferPos += uInStringLength;

	long long int iBufferStringLength = m_uBufferStringLength;
	Asura_Assert( ( m_pcBufferPos - m_szBuffer == iBufferStringLength ), "Pointer problems :(" );

}

void Tools_XMLExporter::AddToBuffer(int iData)
{
	char acBuffer[16];
	Asura_PrintfToBfr( acBuffer, "%i", 16, true, iData );
	AddToBuffer( acBuffer );
}

void Tools_XMLExporter::AddToBuffer( const float fData )
{
	char acBuffer[16];
	Asura_PrintfToBfr( acBuffer, "%f", 16, true, fData );
	AddToBuffer( acBuffer );
}

void Tools_XMLExporter::AddToBuffer( const u_int uData )
{
	char acBuffer[16];
	Asura_PrintfToBfr( acBuffer, "%u", 16, true, uData );
	AddToBuffer( acBuffer );
}

void Tools_XMLExporter::AddToBuffer( const bool bData )
{	
	AddToBuffer( bData ? "true" : "false" );
}

void Tools_XMLExporter::AddToBufferHex( const Asura_Hash_ID uData )
{
	char acBuffer[16];
	Asura_PrintfToBfr( acBuffer, "%X", 16, true, uData );
	AddToBuffer( acBuffer );
}

void Tools_XMLExporter::AddToBufferEscaped( const Asura_Char* szInString )
{
	if(!szInString) return;

	const char* szIllegalChars = "<>&\""; //characters we need to escape

	char acTemp[1024] = "";
	size_t uIllegalCharPoint = strcspn ( szInString, szIllegalChars );

	if( uIllegalCharPoint != strlen(szInString) )
	//there's a reserved character that needs escaping
	{
		//pointer to how much we've escaped so far
		const char* pcReadInputTo = szInString;
		u_int uBufferRemaining = sizeof(acTemp) - strlen(acTemp);
		do
		{
			//put the chars that don't need escaping into acTemp
			strncat( acTemp, pcReadInputTo, min(uIllegalCharPoint, uBufferRemaining) );

			//add the escaped char
			switch(pcReadInputTo[uIllegalCharPoint])
			{
			case '<':
				strcat ( acTemp, "&lt;" );
				break;

			case '>':
				strcat ( acTemp, "&gt;" );
				break;
			case '&':
				strcat ( acTemp, "&amp;" );
				break;
			case '"':
				strcat ( acTemp, "&quot;" );
				break;
			case '\'':
				strcat ( acTemp, "&apos;" );
				break;
			}
			
			//we've now written the char that needed escaping, so advance the pointer
			pcReadInputTo += uIllegalCharPoint + 1;
			uBufferRemaining = sizeof(acTemp) - strlen(acTemp);
			
			//check if we need to escape any more
			uIllegalCharPoint = strcspn ( pcReadInputTo, szIllegalChars ) ;
		} while ( uIllegalCharPoint != strlen(pcReadInputTo));

		//put the remainder of the string into acTemp
		strncat( acTemp, pcReadInputTo, min(uIllegalCharPoint, uBufferRemaining) );
		AddToBuffer(acTemp);
	}
	else { AddToBuffer( szInString ); }

}

void Tools_XMLExporter::Clear()
{
	m_szBuffer[0] = '\0';
	m_uBufferStringLength = 0;
	m_pcBufferPos = m_szBuffer;
}