#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_MOVEMENT_H_
#define __TOOLS_MOVEMENT_H_

#include "Core/Asura_Maths.h"
#include "Objects/Tools_Object.h"

	

#define ToolsTeleportFlag_Absolute						0x00000001
#define ToolsTeleportFlag_RelativeLocal					0x00000002
#define ToolsTeleportFlag_RelativeWorld					0x00000004
#define ToolsTeleportFlag_MoveCamera					0x00000008
#define ToolsTeleportFlag_MoveSelected					0x00000010
#define ToolsTeleportFlag_MoveCameraAndObjectIfAttached	0x00000020
#define ToolsTeleportFlag_UsePosition			 		0x00000040
#define ToolsTeleportFlag_UseOrient				 		0x00000080
#define ToolsTeleportFlag_FromOverlay					0x00000100 //Teleport was caused by teleport overlay. (Added to help with project specific functions)



//keeps track of the current flying object , and updates it each frame
//(Many of the functions that refer to the camera will most likely affect the 'player' 
//instead in 3rd person mode)
class Tools_Movement
{
	public :
		Tools_Movement();
		~Tools_Movement();
		
		static Tools_Movement* GetThis() {return s_pxThis;}

		static void Initialise(void);
		static void Update(void);
		static void Render();
		static void Shutdown(void);

		//Enables/Disables movement controls
		static void Enable(bool bEnable) {s_bEnabled = bEnable;}


		//called once per frame to update whether mouse rotation is enabled
		static void UpdateMouseRotationState();
		static bool IsUsingMouseRotation(){return s_uMouseRotationState!=0;}
	

		//set the current flying object...
		//pass 0 if it should revert back to moving the camera
		static void SetFlyingObject(Tools_Object*);

		static Tools_Object* GetFlyingObject(){return CurrentObject;}
		static Tools_Object* GetCameraObject(){return CameraObject;}

		static void SaveCameraPosition(int index);
		static void RestoreCameraPosition(int index);

		static void DoTeleport();
		static void TeleportCameraToPosition( const Asura_Vector_3* pxPos, const Asura_Matrix_3x3* pxOri);
		
		static void CreateTeleportOverlay();


		static void SaveCameraPositionFile();
		static void LoadCameraPositionFile();

//		static void ToggleThirdPersonCamera();
//		static bool ThirdPersonCameraIsOn();

		static void FindModule();


		//movement scale when moving the camera/player
		static float CameraMovementScale;
		static float CameraRotationScale;
		//movement scale for other objects
		static float ObjectMovementScale;
		static float ObjectRotationScale;

		//if set to true causes the camera to stay with the currently flown object
		static bool AttachCamera;

		//if true then rotation is relative to the current orientation , instead of world axes
		static bool UseRelativeRotation;

		static bool InvertYAxis;

		//stops movement using the numberpad , should those keys be needed for something else
		static bool DisableNumberPadControls;

//		static bool ThirdPersonCamera;

		static float MouseRotationMultiplier;

		static bool IsUsingDynamics();
		static void ToggleUseDynamics();

		//If true then the directional keys will move relative the world axes instead of the object axes
		static bool IsUsingWorldRelativeMovement();
		static void ToggleUseWorldRelativeMovement();


		static void LoadSettingsFromIniFile();
		static void SaveSettingsToIniFile();

		//option for holding down Alt to activate mouse rotation
		static bool IsUsingAltForMouseRotate(){return s_bUseAltForMouseRotate;}
		static void SetUseAltForMouseRotate(bool bSet) {s_bUseAltForMouseRotate = bSet;}

		//Determines whether the use of MMB for rotation is a toggle
		static bool IsMMBToggled(){return s_bToggleMiddleMouseButton;}
		static void SetMMBToggled(bool bSet){s_bToggleMiddleMouseButton = bSet;}
		static void CancelToggledMMB();
		static void CancelToggledAlt();

		//Determines whether the mouse wheel get used for setting speed
		static bool IsUsingMouseWheelForSpeed(){return s_bUseMouseWheelForSpeed;}
		static void SetUseMouseWheelForSpeed(bool bSet) {s_bUseMouseWheelForSpeed = bSet;}

		
		//Get the centre point for orbit movement
		static const Asura_Vector_3& GetOrbitCentre();

		static void AddOverlayControls(class Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset);


	protected :
		virtual void AddTeleportOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset);
		virtual void TeleportCameraAndOrObject(const Asura_Vector_3* pxPos, const Asura_Matrix_3x3* pxOri,u_int uFlags);
 
	
	private :
		static void TransferModeData(int iLoadOrSave);

		static void UpdateTeleportOverlaySettings();

	protected :
		static u_int s_uTeleportFlags;
		static u_int s_uTeleportFlagsBackup;

	private :
		static bool s_bEnabled;

		//current object to fly. if NULL then...
		static Tools_Object* CurrentObject;
		//fly the camera (player/user) instead
		static Tools_Object* CameraObject;

		static Asura_Vector_3* SavedPosition;
		static Asura_Matrix_3x3* SavedOrientation;

		static Asura_Vector_3 s_xTeleportPosition;
		static Asura_Euler_3 s_xTeleportOrient;

		
		static float s_fSpeedDisplayTimer;

		static bool s_bUseCameraDynamics;//use dynamics when moving the camera
		static bool s_bUseObjectDynamics;//use dynamics when flying an object

		static bool s_bObjectUseWorldRelativeMovement;
		static bool s_bCameraUseWorldRelativeMovement;

		static bool s_bUseAltForMouseRotate;
		static bool s_bToggleMiddleMouseButton;
		static bool s_bUseMouseWheelForSpeed;

		static u_int s_uMouseRotationState;

		static Asura_Vector_2 s_xOptionsOverlayPosition;

		static Tools_Movement* s_pxThis;
		
};




//for handling movement of the player/camera
class Tools_Player_Object : public Tools_Object
{
	public :
		Tools_Player_Object();

		virtual void UpdateMovement(float* pMovementScale,float* pRotationScale);

		virtual void UpdatePositionFromAsuraData();
		virtual void UpdatePositionInAsuraData(bool /*updateLinkedObjects*/);
		virtual bool IsSelectable(){return false;}
		virtual bool AllowMovementUndo(){return false;}
		virtual bool ShouldPersistBetweenModes() const {return true;}
		virtual void Render();


	protected :
		float m_fHeight;//last used player height

		
};



#endif //__TOOLS_MOVEMENT_H_
