//
// Tools_Object_Gizmo.h
//
// Created by Richard Rice - 06/09/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//


#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_OBJECT_GIZMO_H__
#define __TOOLS_OBJECT_GIZMO_H__

#include "Objects/Tools_Object.h"

class Tools_Object_System;

//Deals with creating the gizmos for the selected object.
class Tools_Gizmo_Handler
{
public :
	static void ToggleGizmos(Tools_Object* pxObject);
	static void DeleteGizmos();

	static void ObjectDeleted(Tools_Object* pxObject);
	static void SelectionChanged(Tools_Object* pxNewSelection); 
	

	static void ToggleWorldAxis() {s_bWorldAxis = !s_bWorldAxis;}
	static bool IsWorldAxis() {return s_bWorldAxis;}

	static float GetSnapMoveGranularity(){return s_fSnapMoveGranularity;}
	static float GetSnapRotateGranularity(){return s_fSnapRotateGranularity;}
	static bool IsUniformScale(){return s_bUniformScale;}

	static float GetGizmoScale(){return s_fGizmoScale;}

	static void ShowOverlay();
	static void CloseOverlay();

private :
	static List<class Tools_Object_Gizmo*> s_xGizmoList;
	static Tools_Object* s_pxCurrentObject;
	static bool s_bWorldAxis;

	static float s_fSnapMoveGranularity;
	static Asura_Angle s_fSnapRotateGranularity;
	static bool s_bUniformScale;
	static Asura_Vector_2 s_xOverlayPosition;

	static float s_fGizmoScale;

	friend class Tools_Object_System;//for saving/loading settings
	friend class Tools_Object_Gizmo;
};

//Base class for the gizmos
class Tools_Object_Gizmo : public Tools_Object
{
public :
	Tools_Object_Gizmo(Tools_Object* pxParent,int iAxis,float fSpokePos);
	~Tools_Object_Gizmo();

	void SetControlledObject(Tools_Object* pxObject) {m_pxControlledObject = pxObject;}


protected :
	virtual void Render();
	virtual void Update();
	virtual bool CanBeDestroyed(){return false;}
	virtual bool IsSelectable(){return false;}
	virtual bool IsTemporaryEditObject() const {return true;}

	virtual bool ShouldRenderAsUI() const {return true;}


	virtual void AddContextMenuItems(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext);
	static void ContextMenu_ShowOverlay();


	//updates position/orient/scale based on the position and size of the parent
	void UpdatePosition();
	virtual void UpdateOrientation();

	//Gets the axis for this gizmo
	int GetAxis() const {return m_iAxis;}

	//Gets the normal vector in the direction of the specified axis
	Asura_Vector_3 GetAxisDir(int iAxis) const;
	Asura_Vector_3 GetAxisDir() const {return GetAxisDir(m_iAxis);}


	Tools_Object* m_pxControlledObject;

	int m_iAxis;
	float m_fSpokePosition;//position along gizmo spoke
};

//Deals with moving the object along an axis
class Tools_Object_MoveGizmo : public Tools_Object_Gizmo
{
public :
	Tools_Object_MoveGizmo(Tools_Object* pxParent,int iAxis);

protected :
	virtual bool IsDraggable() const {return true;}
	virtual void StartDrag(const Asura_Vector_2& xCursorPos);
	virtual void UpdateDrag(const Asura_Vector_2& xCursorPos);

	virtual const char* GetName(){return "Move Gizmo";}
	virtual const char* GetTypeDescription(){return "Move Gizmo";}

	bool GetPointNearestCursor(const Asura_Vector_2& xCursorPos,Asura_Vector_3& xOutput);

private :
	Asura_Vector_3 m_xInitialObjectPos;
	Asura_Vector_3 m_xInitialGizmoPos;
	Asura_Vector_3 m_xInitialPoint;
	bool m_bDragOkay;

};

//Deals with moving the object within a plane
class Tools_Object_2AxisMoveGizmo : public Tools_Object_Gizmo
{
public :
	Tools_Object_2AxisMoveGizmo(Tools_Object* pxParent,int iAxis1,int iAxis2);

protected :
	virtual bool IsDraggable() const {return true;}
	virtual void StartDrag(const Asura_Vector_2& xCursorPos);
	virtual void UpdateDrag(const Asura_Vector_2& xCursorPos);

	virtual const char* GetName(){return m_xName.GetString();}
	virtual const char* GetTypeDescription(){return "Move Gizmo";}

	bool GetPointNearestCursor(const Asura_Vector_2& xCursorPos,Asura_Vector_3& xOutput);

	virtual void UpdateOrientation();

private :
	Asura_String m_xName;

	Asura_Vector_3 m_xInitialObjectPos;
	Asura_Vector_3 m_xInitialGizmoPos;
	Asura_Vector_3 m_xInitialPoint;
	bool m_bDragOkay;

	int m_iSecondaryAxis;

};


//Deals with rotating about an axis
class Tools_Object_RotateGizmo : public Tools_Object_Gizmo
{
public :
	Tools_Object_RotateGizmo(Tools_Object* pxParent,int iAxis);

protected :
	virtual bool IsDraggable() const {return true;}
	virtual void StartDrag(const Asura_Vector_2& xCursorPos);
	virtual void UpdateDrag(const Asura_Vector_2& xCursorPos);

	virtual void UpdateOrientation();

	virtual const char* GetName(){return "Rotate Gizmo";}
	virtual const char* GetTypeDescription(){return "Rotate Gizmo";}


	bool GetPointOnPlane(const Asura_Vector_2& xCursorPos,Asura_Vector_3& xOutput);



private :
	Asura_Matrix_3x3 m_xInitialObjectOrient;
	Asura_Vector_3 m_xRotateAxisDir;
	Asura_Vector_3 m_xInitialPoint;
	bool m_bDragOkay;



};

//Deals with moving the object along an axis
class Tools_Object_ScaleGizmo : public Tools_Object_Gizmo
{
public :
	Tools_Object_ScaleGizmo(Tools_Object* pxParent,int iAxis);

protected :
	virtual bool IsDraggable() const {return true;}
	virtual void StartDrag(const Asura_Vector_2& xCursorPos);
	virtual void UpdateDrag(const Asura_Vector_2& xCursorPos);

	virtual const char* GetName(){return "Scale Gizmo";}
	virtual const char* GetTypeDescription(){return "Scale Gizmo";}
	virtual bool IsHidden();

	bool GetCursorDistanceFromObject( const Asura_Vector_2& xCursorPos,float& fOutput);


private :
	Asura_Vector_3 m_xInitialObjectPos;
	Asura_Vector_3 m_xInitialScale;
	float m_fInitialDistance;
	bool m_bDragOkay;

};



#endif //__TOOLS_OBJECT_GIZMO