//////////////////////////////////////////////////////////////////////
// Tools_VolumeCuboidEdit_Object.cpp							//
// SJL 14/11/07														//
// Copyright (C) 2007 Rebellion, All Rights Reserved.			    //
//////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "Objects/Tools_VolumeCuboidEdit_Object.h"

#include "Camera/Asura_Camera.h"
#include "Core/Asura_Los.h"
#include "Core/Asura_UserInput.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Cursor.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"

#include "Tools_Overlays.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_LOS.h"
#include "Misc/Tools_Help.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Objects/Tools_Object_System.h"
#include "Objects/Tools_VolumeGroupEdit_Controller.h"

static const float s_fMaxExtents = 30000.f; // Fairly arbitary.

static const char* s_apszVolumeEditHelpStrings [] =
{
	"Esc :           End volume editing",
	"",
	"Z :             Set bounding box to match module under cursor",
	"X :             Expand bounding box to include module",
	"9 :             Move selected face in",
	"Shift+ 9 :      Move selected face in (faster)",
	"0 :             Move selected face out",
	"Shift + 0 :     Move selected face out (faster)",
	"B :             Snap selected face to point under cursor",
	"G :             As above ,but also moves the opposite face in the opposite direction",
	"",
	"LMB :           Select nearest face under cursor",
	"Shift+LMB :     Select furthest face under cursor",
	"1 :             Select Min X face",
	"2 :             Select Max X face",
	"3 :             Select Min Y face",
	"4 :             Select Max Y face",
	"5 :             Select Min Z face",
	"6 :             Select Max Z face",
	0
};

Tools_VolumeCuboidEdit_Object::Tools_VolumeCuboidEdit_Object(const char* name, Asura_Volume_Cuboid* pxVolume,  bool* pbModifiedFlag /*=0*/)
:Tools_VolumeBaseEdit_Object(name, pxVolume, pbModifiedFlag)
{

	m_pxVolume = pxVolume->GetVolumeBox();
	m_xWorldVolume = *m_pxVolume;

	m_eSelectedFace = FACE_NONE;

	SetMaxExtents (Asura_Vector_3( s_fMaxExtents, s_fMaxExtents, s_fMaxExtents ));
	SetMinExtents (Asura_Vector_3( -s_fMaxExtents, -s_fMaxExtents, -s_fMaxExtents ));

	m_bOverlayVolumeChanged = false;
	m_bMinOverlayVolumeChanged = false;
	m_bMaxOverlayVolumeChanged = false;
	m_bEditingRelativeVolume = false;

	SetDefaultPositionFromVolume();

	m_pszHelpTitle = "Cuboid Volume";
	m_apszHelpTexts = s_apszVolumeEditHelpStrings;

}


Tools_VolumeCuboidEdit_Object::~Tools_VolumeCuboidEdit_Object()
{

}

void Tools_VolumeCuboidEdit_Object::UpdatePositionInAsuraData(bool /*updateLinkedObjects*/)
{
	UpdateFromRelativeVolume();
}

void Tools_VolumeCuboidEdit_Object::UpdateFromRelativeVolume()
{
	UpdateWorldVolumeFromRelative();

	if(m_bEditingRelativeVolume)
	{
		*m_pxVolume = m_xRelativeVolume;
	}
	else
	{
		*m_pxVolume = m_xRelativeVolume;
		m_pxVolume->Displace(Position);

		if(m_bUseOrientation)
		{
			if(m_pxOrientQuat)
			{
				m_pxOrientQuat->FromMatrix(Orientation);
			}
			else if(m_pxOrientMat)
			{
				*m_pxOrientMat = Orientation;
			}
			//If editing the orientation , then the bounding box will need to be offset to take into account
			//that it will be using a different pivot point(the centre of the volume) than the volume object
			Asura_Vector_3 xOffset = GetRelativePivot();
			m_pxVolume->Displace(-xOffset);
			xOffset = xOffset*Orientation;
			m_pxVolume->Displace(xOffset);



		}
		else
		{
			if(m_pxOrientQuat)
			{
				*m_pxOrientQuat = Asura_Maths::IdentityQuat;
			}
			else if(m_pxOrientMat)
			{
				*m_pxOrientMat = Asura_Maths::IdentityMatrix3x3;
			}
		}
	}

	if(m_pbModifiedFlag) *m_pbModifiedFlag = true;
}

void Tools_VolumeCuboidEdit_Object::SetDefaultPositionFromVolume()
{
	if(m_pxOrientMat)
	{
		Orientation = *m_pxOrientMat;
	}
	else if(m_pxOrientQuat)
	{
		m_pxOrientQuat->ToMatrix(Orientation);	
	}
	else
	{
		Orientation.Identity();
	}



	Position = m_xWorldVolume.CalculateCentre();
	if(PreferPositionAtBaseOfVolume())
	{
		Position.y = m_xWorldVolume.MaxY;
		if(Mesh && Mesh->BoundingBoxDataValid)
		{
			//offset position by MaxY of the mesh , so the bottom face lies on the
			//bottom face of the bounding box
			float fOffset = Mesh->BoundingBox.MaxY;
			Position.y -= fOffset;
			if(Position.y<m_xWorldVolume.MinY) Position.y = m_xWorldVolume.MinY;

		}
	}
	m_xRelativeVolume = m_xWorldVolume;
	m_xRelativeVolume.Displace(-Position);
	m_xOverlayVolume = m_xRelativeVolume;

	//negate the minimum valuse , to make them more convenient for the overlay controls
	m_xOverlayVolume.MinX = -m_xOverlayVolume.MinX;
	m_xOverlayVolume.MinY = -m_xOverlayVolume.MinY;
	m_xOverlayVolume.MinZ = -m_xOverlayVolume.MinZ;
	m_bOverlayVolumeChanged = false;
	m_bMinOverlayVolumeChanged = false;
	m_bMaxOverlayVolumeChanged = false;

	if(m_bUseOrientation)
	{
		Asura_Vector_3 xCentre = GetWorldPivot();
		Asura_Vector_3 xOffset = xCentre-Position;
		xOffset = xOffset*Orientation;

		Position = xCentre-xOffset;
	}

	
}

void Tools_VolumeCuboidEdit_Object::UpdateWorldVolumeFromRelative()
{
	m_xWorldVolume = m_xRelativeVolume;
	m_xWorldVolume.Displace(Position);

	if(m_bUseOrientation)
	{
		//If editing the orientation , then the bounding box will need to be offset to take into account
		//that it will be using a different pivot point(the centre of the volume) than the volume object
		Asura_Vector_3 xOffset = GetRelativePivot();
		m_xWorldVolume.Displace(-xOffset);
		xOffset = xOffset*Orientation;
		m_xWorldVolume.Displace(xOffset);
	}
}

void Tools_VolumeCuboidEdit_Object::AddOverlayControls(class Asura_Overlay* pxOverlay,Asura_Vector_2& xOffset)
{
	PARENT::AddOverlayControls(pxOverlay, xOffset);

	Asura_Dialog* pxDialog;

	const char* pszExtentText = "Volume extents (metres) relative to the volume marker object.";

	pxDialog = new Asura_Numerical_Dialog("1: Min X",xOffset,&m_xOverlayVolume.MinX,m_xMinExtents.x,m_xMaxExtents.x,0.1f,"%.3f",&m_bMinOverlayVolumeChanged);
	pxDialog->SetHelpText(pszExtentText);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("2: Max X",xOffset,&m_xOverlayVolume.MaxX,m_xMinExtents.x,m_xMaxExtents.x,0.1f,"%.3f",&m_bMaxOverlayVolumeChanged);
	pxDialog->SetHelpText(pszExtentText);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("3: Min Y",xOffset,&m_xOverlayVolume.MinY,m_xMinExtents.y,m_xMaxExtents.y,0.1f,"%.3f",&m_bMinOverlayVolumeChanged);
	pxDialog->SetHelpText(pszExtentText);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("4: Max Y",xOffset,&m_xOverlayVolume.MaxY,m_xMinExtents.y,m_xMaxExtents.y,0.1f,"%.3f",&m_bMaxOverlayVolumeChanged);
	pxDialog->SetHelpText(pszExtentText);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("5: Min Z",xOffset,&m_xOverlayVolume.MinZ,m_xMinExtents.z,m_xMaxExtents.z,0.1f,"%.3f",&m_bMinOverlayVolumeChanged);
	pxDialog->SetHelpText(pszExtentText);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("6: Max Z",xOffset,&m_xOverlayVolume.MaxZ,m_xMinExtents.z,m_xMaxExtents.z,0.1f,"%.3f",&m_bMaxOverlayVolumeChanged);
	pxDialog->SetHelpText(pszExtentText);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
}


void Tools_VolumeCuboidEdit_Object::Render()
{
	//render the flyable bit
	Tools_Object::Render();

	Asura_Render::CurrentStates.RequestWireFrameMode(false);

	u_int uColour;
	u_int uSelectedFaceColour;
	u_int uWireframeColour;	//colour to use as additional highlight when doing solid volumes
	if(IsSelected())
	{
		uColour = 0xafafff;
		uSelectedFaceColour = 0xff0000;
		uWireframeColour = 0x0000ff;
	}
	else if(GetController() && (GetController()->IsSelected() || GetController()->IsHighlighted()))
	{
		uColour = 0xafffaf;
		uSelectedFaceColour = uColour;
		uWireframeColour = 0x00ff00;
	}
	else
	{
		uColour = 0xafafaf;
		uSelectedFaceColour = uColour;
		uWireframeColour = 0xffffff;
	}
	if(s_bWireframe)
	{
		uColour |= 0xff000000;
		uSelectedFaceColour |= 0xff000000;
	}
	else
	{
		uColour |= 0x70000000;
		uSelectedFaceColour |= 0x70000000;
	}

	Asura_Matrix_3x3 xOrient = m_bUseOrientation ? Orientation : Asura_Maths::IdentityMatrix3x3;
	Tools_Render::RenderCube(m_xRelativeVolume,Position,xOrient,uColour,s_bWireframe,m_eSelectedFace,uSelectedFaceColour);

	if(!s_bWireframe)
	{
		//do a wireframe outline as well
		Tools_Render::RenderCube(m_xRelativeVolume,Position,xOrient,uWireframeColour|0xff000000,true,m_eSelectedFace,uSelectedFaceColour|0xff000000);
	}


	Asura_Render::CurrentStates.RequestWireFrameMode(Asura_Render::GetWireframe());
}




void Tools_VolumeCuboidEdit_Object::UpdateWhileSelected()
{
	if(!Asura_Overlay_System::OverlaysHaveKeyboardFocus())
	{
		if(!Asura_UserInput::KeyPressed(ASURA_KEYID_ALT) && !Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
		{
			if ( !Asura_UserInput::KeyPressed( ASURA_KEYID_SHIFT ) )
			{
				if(Asura_UserInput::DebouncedKeyPressed('x'))
				{
					ExpandToIncludeModule();
				}
				if(Asura_UserInput::DebouncedKeyPressed('z'))
				{
					SetToModule();
				}
				if(Asura_UserInput::DebouncedKeyPressed('1'))
				{
					SetSelectedFace(FACE_MIN_X);
				}
				if(Asura_UserInput::DebouncedKeyPressed('2'))
				{
					SetSelectedFace(FACE_MAX_X);
				}
				if(Asura_UserInput::DebouncedKeyPressed('3'))
				{
					SetSelectedFace(FACE_MIN_Y);
				}
				if(Asura_UserInput::DebouncedKeyPressed('4'))
				{
					SetSelectedFace(FACE_MAX_Y);
				}
				if(Asura_UserInput::DebouncedKeyPressed('5'))
				{
					SetSelectedFace(FACE_MIN_Z);
				}
				if(Asura_UserInput::DebouncedKeyPressed('6'))
				{
					SetSelectedFace(FACE_MAX_Z);
				}
				if( Asura_UserInput::DebouncedKeyPressed('7') ||
					Asura_UserInput::DebouncedKeyPressed('8') )
				{
					SetSelectedFace(FACE_NONE);
				}
			}
			//increase/decrease side.
			if(m_eSelectedFace != FACE_NONE)
			{
				float* pfSelectedFaceValue = 0;
				float* pfOppositeFaceValue = 0;

				switch(m_eSelectedFace)
				{
				case FACE_MIN_X :
					pfSelectedFaceValue = &m_xOverlayVolume.MinX;
					pfOppositeFaceValue = &m_xOverlayVolume.MaxX;
					break;
				case FACE_MIN_Y :
					pfSelectedFaceValue = &m_xOverlayVolume.MinY;
					pfOppositeFaceValue = &m_xOverlayVolume.MaxY;
					break;
				case FACE_MIN_Z :
					pfSelectedFaceValue = &m_xOverlayVolume.MinZ;
					pfOppositeFaceValue = &m_xOverlayVolume.MaxZ;
					break;
				case FACE_MAX_X :
					pfSelectedFaceValue = &m_xOverlayVolume.MaxX;
					pfOppositeFaceValue = &m_xOverlayVolume.MinX;
					break;
				case FACE_MAX_Y :
					pfSelectedFaceValue = &m_xOverlayVolume.MaxY;
					pfOppositeFaceValue = &m_xOverlayVolume.MinY;
					break;
				case FACE_MAX_Z :
					pfSelectedFaceValue = &m_xOverlayVolume.MaxZ;
					pfOppositeFaceValue = &m_xOverlayVolume.MinZ;
					break;
				}	

				float fAdjustment = 0;
				if(Asura_UserInput::KeyPressed('0'))
				{
					fAdjustment = Asura_Timers::GetFrameTime();
				}
				if(Asura_UserInput::KeyPressed('9'))
				{
					fAdjustment = -Asura_Timers::GetFrameTime();
				}
				if(fAdjustment)
				{
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
					{
						fAdjustment*=10.0f;
					}
					*pfSelectedFaceValue += fAdjustment;
					if(*pfSelectedFaceValue<-*pfOppositeFaceValue) *pfSelectedFaceValue = -*pfOppositeFaceValue;
					m_bOverlayVolumeChanged = true;
					m_bChanged = true;
				}
			}
		}

	}

	if(m_bMinOverlayVolumeChanged)
	{
		if(m_xOverlayVolume.MinX<-m_xOverlayVolume.MaxX) m_xOverlayVolume.MinX=-m_xOverlayVolume.MaxX;
		if(m_xOverlayVolume.MinY<-m_xOverlayVolume.MaxY) m_xOverlayVolume.MinY=-m_xOverlayVolume.MaxY;
		if(m_xOverlayVolume.MinZ<-m_xOverlayVolume.MaxZ) m_xOverlayVolume.MinZ=-m_xOverlayVolume.MaxZ;
		m_bMinOverlayVolumeChanged = false;
		m_bOverlayVolumeChanged = true;
	}
	else if(m_bMaxOverlayVolumeChanged)
	{
		if(m_xOverlayVolume.MaxX<-m_xOverlayVolume.MinX) m_xOverlayVolume.MaxX=-m_xOverlayVolume.MinX;
		if(m_xOverlayVolume.MaxY<-m_xOverlayVolume.MinY) m_xOverlayVolume.MaxY=-m_xOverlayVolume.MinY;
		if(m_xOverlayVolume.MaxZ<-m_xOverlayVolume.MinZ) m_xOverlayVolume.MaxZ=-m_xOverlayVolume.MinZ;
		m_bMaxOverlayVolumeChanged = false;
		m_bOverlayVolumeChanged = true;
	}

	if(m_bOverlayVolumeChanged)
	{
		//update relative volume from overlay volume
		m_xRelativeVolume = m_xOverlayVolume;
		m_xRelativeVolume.MinX = - m_xRelativeVolume.MinX;
		m_xRelativeVolume.MinY = - m_xRelativeVolume.MinY;
		m_xRelativeVolume.MinZ = - m_xRelativeVolume.MinZ;
		m_bChanged = true;
		m_bOverlayVolumeChanged = false;
	}
	else
	{
		//update overlay volume from relative volume
		m_xOverlayVolume = m_xRelativeVolume;
		m_xOverlayVolume.MinX = - m_xOverlayVolume.MinX;
		m_xOverlayVolume.MinY = - m_xOverlayVolume.MinY;
		m_xOverlayVolume.MinZ = - m_xOverlayVolume.MinZ;
	}
	PARENT::UpdateWhileSelected();
}


void Tools_VolumeCuboidEdit_Object::ExpandToIncludeModule()
{
	int moduleIndex;
	if(!Tools_LOS::FindModuleUnderCursor(moduleIndex)) return;

	//work out the world bounding box for this module
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(moduleIndex);

	Asura_Vector_3 offset;		
	Asura_Module_System::GetPositionOfModule(moduleIndex,offset);

	//get the bounding box
	if(!mesh->BoundingBoxDataValid) mesh->CalculateBoundingBox();
	Asura_Bounding_Box box = mesh->BoundingBox;

	//add on the modules location
	box.Displace(offset);


	for(int i=0;i<8;i++)
	{
		//get corner of mesh bounding box
		Asura_Vector_3 xCorner;
		xCorner.x = (i&1) ? box.MinX : box.MaxX;
		xCorner.y = (i&2) ? box.MinY : box.MaxY;
		xCorner.z = (i&4) ? box.MinZ : box.MaxZ;

		xCorner = GetRelativePosition(xCorner);


		//expand box to include this point
		m_xRelativeVolume.Set(xCorner);

	}

	UpdateFromRelativeVolume();

	if(m_pbModifiedFlag) *m_pbModifiedFlag = true;
}


void Tools_VolumeCuboidEdit_Object::SetToModule()
{
	int moduleIndex;
	if(!Tools_LOS::FindModuleUnderCursor(moduleIndex)) return;

	//work out the world bounding box for this module
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(moduleIndex);

	Asura_Vector_3 offset;		
	Asura_Module_System::GetPositionOfModule(moduleIndex,offset);

	//get the bounding box
	if(!mesh->BoundingBoxDataValid) mesh->CalculateBoundingBox();
	Asura_Bounding_Box box = mesh->BoundingBox;

	//add on the modules location
	box.Displace(offset);

	//set the volume equal to this box
	m_xWorldVolume = box;
	if(m_pxOrientQuat)
	{
		*m_pxOrientQuat = Asura_Maths::IdentityQuat;
	}
	else if(m_pxOrientMat)
	{
		*m_pxOrientMat = Asura_Maths::IdentityMatrix3x3;
	}

	SetDefaultPositionFromVolume();

	if(m_bEditingRelativeVolume)
	{
		*m_pxVolume = m_xRelativeVolume;
	}
	else
	{
		*m_pxVolume = m_xWorldVolume;
	}

	if(m_pbModifiedFlag) *m_pbModifiedFlag = true;

}

void Tools_VolumeCuboidEdit_Object::SnapFaceToCursor(const bool bAlsoExpandOppositeFace)
{
	if(m_eSelectedFace == FACE_NONE) return;

	int iPolyIndex;
	int iModuleIndex;
	Asura_Vector_3 xPosition;

	//find the location of the cursor
	if(!Tools_LOS::FindPolyAndPositionUnderCursor(iPolyIndex,iModuleIndex,xPosition)) return;

	xPosition = GetRelativePosition(xPosition);

	Asura_Bounding_Box xInitialRelativeVolume = m_xRelativeVolume;

	switch	(m_eSelectedFace)
	{
	case FACE_MIN_X :
		m_xRelativeVolume.MinX = xPosition.x;
		if(m_xRelativeVolume.MinX>m_xRelativeVolume.MaxX) m_xRelativeVolume.MinX=m_xRelativeVolume.MaxX;
		break;
	case FACE_MAX_X :
		m_xRelativeVolume.MaxX = xPosition.x;
		if(m_xRelativeVolume.MinX>m_xRelativeVolume.MaxX) m_xRelativeVolume.MaxX=m_xRelativeVolume.MinX;
		break;
	case FACE_MIN_Y :
		m_xRelativeVolume.MinY = xPosition.y;
		if(m_xRelativeVolume.MinY>m_xRelativeVolume.MaxY) m_xRelativeVolume.MinY=m_xRelativeVolume.MaxY;
		break;
	case FACE_MAX_Y :
		m_xRelativeVolume.MaxY = xPosition.y;
		if(m_xRelativeVolume.MinY>m_xRelativeVolume.MaxY) m_xRelativeVolume.MaxY=m_xRelativeVolume.MinY;
		break;
	case FACE_MIN_Z :
		m_xRelativeVolume.MinZ = xPosition.z;
		if(m_xRelativeVolume.MinZ>m_xRelativeVolume.MaxZ) m_xRelativeVolume.MinZ=m_xRelativeVolume.MaxZ;
		break;
	case FACE_MAX_Z :
		m_xRelativeVolume.MaxZ = xPosition.z;
		if(m_xRelativeVolume.MinZ>m_xRelativeVolume.MaxZ) m_xRelativeVolume.MaxZ=m_xRelativeVolume.MinZ;
		break;
	default :
		Asura_Assert(0,"Unexpected case");
		break;

	}

	if(bAlsoExpandOppositeFace)
	{
		switch	(m_eSelectedFace)
		{
			case FACE_MIN_X :
				m_xRelativeVolume.MaxX -= m_xRelativeVolume.MinX - xInitialRelativeVolume.MinX;
				if(m_xRelativeVolume.MinX>m_xRelativeVolume.MaxX) m_xRelativeVolume.MaxX=m_xRelativeVolume.MinX;
				break;
			case FACE_MAX_X :
				m_xRelativeVolume.MinX -= m_xRelativeVolume.MaxX - xInitialRelativeVolume.MaxX;
				if(m_xRelativeVolume.MinX>m_xRelativeVolume.MaxX) m_xRelativeVolume.MinX=m_xRelativeVolume.MaxX;
				break;
			case FACE_MIN_Y :
				m_xRelativeVolume.MaxY -= m_xRelativeVolume.MinY - xInitialRelativeVolume.MinY;
				if(m_xRelativeVolume.MinY>m_xRelativeVolume.MaxY) m_xRelativeVolume.MaxY=m_xRelativeVolume.MinY;
				break;
			case FACE_MAX_Y :
				m_xRelativeVolume.MinY -= m_xRelativeVolume.MaxY - xInitialRelativeVolume.MaxY;
				if(m_xRelativeVolume.MinY>m_xRelativeVolume.MaxY) m_xRelativeVolume.MinY=m_xRelativeVolume.MaxY;
				break;
			case FACE_MIN_Z :
				m_xRelativeVolume.MaxZ -= m_xRelativeVolume.MinZ - xInitialRelativeVolume.MinZ;
				if(m_xRelativeVolume.MinZ>m_xRelativeVolume.MaxZ) m_xRelativeVolume.MaxZ=m_xRelativeVolume.MinZ;
				break;
			case FACE_MAX_Z :
				m_xRelativeVolume.MinZ -= m_xRelativeVolume.MaxZ - xInitialRelativeVolume.MaxZ;
				if(m_xRelativeVolume.MinZ>m_xRelativeVolume.MaxZ) m_xRelativeVolume.MinZ=m_xRelativeVolume.MaxZ;
				break;
			default :
				Asura_Assert(0,"Unexpected case");
				break;

		}
	}


	UpdateFromRelativeVolume();

	if( m_xRelativeVolume.MinX>0 || m_xRelativeVolume.MinY>0 || m_xRelativeVolume.MinZ>0 ||
		m_xRelativeVolume.MaxX<0 || m_xRelativeVolume.MaxY<0 || m_xRelativeVolume.MaxZ<0)
	{
		SetDefaultPositionFromVolume();
	}

	if(m_pbModifiedFlag) *m_pbModifiedFlag = true;
}

void Tools_VolumeCuboidEdit_Object::SelectFaceUnderCursor( const bool bGetFurthest )
{
	//get the z direction for the mouse
	Asura_Vector_3 xZvec;
	Asura_Cursor::GetLOSDirection( xZvec );

	// Kevin Lea - 22/06/2001 12:49:20 - and transform this direction into worldspace
	const Asura_Vector_3*   pxCameraPosition    = Asura_Camera::GetPosition();
	const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();

	xZvec = xZvec * (*pxCameraOrientation);

	VolumeFace eNearestFace  = FACE_NONE;
	VolumeFace eFurthestFace = FACE_NONE;
	float      fNearest      = ASURA_LOS_MAX_DISTANCE;
	float      fFurthest     = 0.0f;

	Asura_Vector_3 axVertices[4];
	float          fDistance;
	bool           fFacingToward;
	for ( int i=(int)FACE_MIN_X; i<(int)FACE_END_ENUM; ++i )
	{
		GetFaceVertices( (VolumeFace)i, axVertices );

		fDistance = Asura_Maths::RayIntersectsTri( *pxCameraPosition,
			xZvec, axVertices[0], axVertices[1], axVertices[2], fFacingToward );
		if ( 0.0f <= fDistance && fDistance < fNearest )
		{
			fNearest     = fDistance;
			eNearestFace = (VolumeFace)i;
		}
		if ( fFurthest < fDistance && fDistance < ASURA_LOS_MAX_DISTANCE )
		{
			fFurthest     = fDistance;
			eFurthestFace = (VolumeFace)i;
		}

		fDistance = Asura_Maths::RayIntersectsTri( *pxCameraPosition,
			xZvec, axVertices[2], axVertices[3], axVertices[0], fFacingToward );
		if ( 0.0f <= fDistance && fDistance < fNearest )
		{
			fNearest     = fDistance;
			eNearestFace = (VolumeFace)i;
		}
		if ( fFurthest < fDistance && fDistance < ASURA_LOS_MAX_DISTANCE )
		{
			fFurthest     = fDistance;
			eFurthestFace = (VolumeFace)i;
		}
	}

	SetSelectedFace( bGetFurthest ? eFurthestFace : eNearestFace );
}

void Tools_VolumeCuboidEdit_Object::SetSelectedFace(enum VolumeFace eFace)
{
	m_eSelectedFace = eFace;

}

void Tools_VolumeCuboidEdit_Object::GetFaceVertices( enum VolumeFace eFace, Asura_Vector_3 axVertices[] )
{
	switch ( eFace )
	{
	case FACE_MIN_X :
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x = m_xWorldVolume.MinX;
			axVertices[i].y = (i&2) ? m_xWorldVolume.MaxY : m_xWorldVolume.MinY;
			axVertices[i].z = ((i&1)^((i&2)>>1)) ? m_xWorldVolume.MaxZ : m_xWorldVolume.MinZ;
		}
		break;

	case FACE_MIN_Y :
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x = (i&2) ? m_xWorldVolume.MaxX : m_xWorldVolume.MinX;
			axVertices[i].y = m_xWorldVolume.MinY;
			axVertices[i].z = ((i&1)^((i&2)>>1)) ? m_xWorldVolume.MaxZ : m_xWorldVolume.MinZ;
		}
		break;

	case FACE_MIN_Z :
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x = (i&2) ? m_xWorldVolume.MaxX : m_xWorldVolume.MinX;
			axVertices[i].y = ((i&1)^((i&2)>>1)) ? m_xWorldVolume.MaxY : m_xWorldVolume.MinY;
			axVertices[i].z = m_xWorldVolume.MinZ;
		}
		break;

	case FACE_MAX_X :
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x = m_xWorldVolume.MaxX;
			axVertices[i].y = (i&2) ? m_xWorldVolume.MaxY : m_xWorldVolume.MinY;
			axVertices[i].z = ((i&1)^((i&2)>>1)) ? m_xWorldVolume.MaxZ : m_xWorldVolume.MinZ;
		}
		break;

	case FACE_MAX_Y :
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x = (i&2) ? m_xWorldVolume.MaxX : m_xWorldVolume.MinX;
			axVertices[i].y = m_xWorldVolume.MaxY;
			axVertices[i].z = ((i&1)^((i&2)>>1)) ? m_xWorldVolume.MaxZ : m_xWorldVolume.MinZ;
		}
		break;

	case FACE_MAX_Z :
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x = (i&2) ? m_xWorldVolume.MaxX : m_xWorldVolume.MinX;
			axVertices[i].y = ((i&1)^((i&2)>>1)) ? m_xWorldVolume.MaxY : m_xWorldVolume.MinY;
			axVertices[i].z = m_xWorldVolume.MaxZ;
		}
		break;

	default:
		for ( int i=0; i<4; ++i )
		{
			axVertices[i].x =
				axVertices[i].y =
				axVertices[i].z = 0.0f;
		}
		break;
	}
	Asura_Vector_3 xCentre = GetWorldPivot();

	for ( int i=0; i<4; ++i )
	{
		if(m_bUseOrientation)
		{
			Asura_Vector_3 xOffset = xCentre-axVertices[i];
			xOffset = xOffset*Orientation;
			axVertices[i] = xCentre - xOffset;
		}

	}
}


bool Tools_VolumeCuboidEdit_Object::ObjectIsInLOS(const Asura_Vector_3& xStart, const Asura_Vector_3& xDir, float& fLowest)
{
	bool bInLOS = Tools_Object::ObjectIsInLOS(xStart,xDir,fLowest);

	//If selected , test against the volume as well.
	if(IsSelected())
	{
		//Do test against each face.
		//Not using RayIntersectsBoundingBox since I'm not sure I want to use its behaviour for tests starting from within.
		bool fFacingToward;
		float fDistance;
		Asura_Vector_3 axVertices[4];

		for ( int i=(int)FACE_MIN_X; i<(int)FACE_END_ENUM; ++i )
		{
			GetFaceVertices( (VolumeFace)i, axVertices );

			fDistance = Asura_Maths::RayIntersectsTri( xStart,
				xDir, axVertices[0], axVertices[1], axVertices[2], fFacingToward );
			if ( 0.0f <= fDistance && fDistance < fLowest )
			{
				fLowest = fDistance;
				bInLOS = true;
			}

			fDistance = Asura_Maths::RayIntersectsTri( xStart,
				xDir, axVertices[2], axVertices[3], axVertices[0], fFacingToward );
			if ( 0.0f <= fDistance && fDistance < fLowest )
			{
				fLowest = fDistance;
				bInLOS = true;
			}
		}
	}

	return bInLOS;
}

void Tools_VolumeCuboidEdit_Object::Displace(const Asura_Vector_3& xOffset)
{
	Position += xOffset;
	m_xWorldVolume.Displace(xOffset);
}

void Tools_VolumeCuboidEdit_Object::RotateAbout(const Asura_Vector_3& xCentre, const Asura_Matrix_3x3& xOri)
{
	m_bUseOrientation = true;
	Orientation = xOri * Orientation;
	Position -= xCentre;
	Position = Position * xOri;
	Position += xCentre;
}

Asura_Vector_3 Tools_VolumeCuboidEdit_Object::GetRelativePosition(const Asura_Vector_3& xPosition) const
{
	if(m_bUseOrientation)
	{
		//determine the centre of the volume in world space
		Asura_Vector_3 xCentreFromObject = GetRelativePivot();

		Asura_Vector_3 xCentre = Position + (xCentreFromObject*Orientation);

		//get position relative to the centre of the volume
		Asura_Vector_3 xRel = xPosition - xCentre;

		//transform it into the space of the unrotated bounding box
		Asura_Matrix_3x3 xOrientInverse = Orientation;
		xOrientInverse.Transpose();
		xRel = xRel * xOrientInverse;

		//and make relative to the position of th object in the unrotated volume
		return xRel + xCentreFromObject;

	}
	else
	{
		return (xPosition - Position);
	}
}