//
// Tools_FileSystem.h
//
// Created by Kevin Lea - 02/07/2001 12:36:27
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_FILESYSTEM_H__
#define __TOOLS_FILESYSTEM_H__

#include "Core/Asura_String.h"
#include "Misc/list_tem.hpp"

class Tools_FileSystem_DialogInfo
{
	public:
		Tools_FileSystem_DialogInfo():DefaultFilename(0),DefaultPath(0),FileFilter(0),DefaultExtension(0),
			WindowTitle(0),AllowMultipleSelect(false){}

		Tools_FileSystem_DialogInfo(const Asura_Char* szWindowTitle,
									const Asura_Char* szDefaultPath, const Asura_Char* szDefaultFilename, const Asura_Char* szDefaultExtension,
									const Asura_Char* szFileFilter, const bool bAllowMulitpleSelect = false)
		: DefaultFilename(szDefaultFilename)
		, DefaultPath(szDefaultPath)
		, FileFilter(szFileFilter)
		, DefaultExtension(szDefaultExtension)
		, WindowTitle(szWindowTitle)
		, AllowMultipleSelect(bAllowMulitpleSelect)
		{
		}

		const Asura_Char* DefaultFilename;
		const Asura_Char* DefaultPath;
		const Asura_Char* FileFilter;
		const Asura_Char* DefaultExtension;

		const Asura_Char* WindowTitle;

		bool AllowMultipleSelect;
};

//for folder browsing function
struct Tools_FileSystem_SelectFolderInfo
{
	Tools_FileSystem_SelectFolderInfo()
	{
		StartDirectory = "";
		DisplayText = "";
		InitialSelection = "";
	}

	Tools_FileSystem_SelectFolderInfo(const Asura_Char* szStartDirectory, const Asura_Char* szDisplayText, const Asura_Char* szInitialSelection)
	: StartDirectory(szStartDirectory)
	, DisplayText(szDisplayText)
	, InitialSelection(szInitialSelection)
	{
	}

	//input
	const Asura_Char* StartDirectory;//root directory to browse from
	const Asura_Char* DisplayText;
	const Asura_Char* InitialSelection;//initially selected directory (NULL if not needed)
	//output
	Asura_Char SelectedFolder[MAX_PATH];

	enum
	{
	   Error_None = 0,
	   Error_DirectoryDoesntExist	
	}Error;
};

//defines for FindFiles and FindDirectories
#define TOOLS_FILEFIND_RECURSIVE	   0x00000001 //Recurse search through subdirectories
#define TOOLS_FILEFIND_IGNOREBACKUPDIR 0x00000002 //Skip any directory called 'backup'

//error codes for ReadAsuraFile()
enum
{
	TOOLS_READASURAFILE_OK,
	TOOLS_READASURAFILE_FAILED_TO_OPEN_FILE,
	TOOLS_READASURAFILE_NOT_ASURA_FILE,
	TOOLS_READASURAFILE_ERROR_READING_FILE,
};

//optional output from ReadAsuraFile
struct Tools_ReadAsuraFileOutput
{
	u_int m_uBufferSize;//size of the buffer that was allocated
	u_int m_uErrorCode;//error code from above enumeration
	u_int m_uCompression;//compression type of original file from enumeration in Asura_FileAccess
};

class Tools_FileSystem
{
	public:
		static Asura_Char* Tools_FileSystem::GetSaveFilename(Tools_FileSystem_DialogInfo& info);
		static Asura_String Tools_FileSystem::GetLoadFilename(Tools_FileSystem_DialogInfo& info);

		//gets the next selected filename (needs a call to GetLoadFilename first to get the first filename)
		static Asura_String Tools_FileSystem::GetNextLoadFilename();

		//browses for a subdirectory of a given directory
		static bool Tools_FileSystem::SelectFolder(Tools_FileSystem_SelectFolderInfo& info);

		static bool DeleteFileByName( const char* szFilename );
		// The folder must be empty.
		static bool DeleteFolderByName( const char* szFilename );

		static void FindFiles( const char* szSearchString, List<Asura_String>& xOutputList, unsigned uFlags = 0 );
		static void FindFilesLocalAndNetwork( const char* szSearchString, List<Asura_String>& xOutputList, unsigned uFlags = 0 );
		static void FindDirectories( const char* szSearchString, List<Asura_String>& xOutputList, unsigned uFlags = 0 );
		static void FindDirectoriesLocalAndNetwork( const char* szSearchString, List<Asura_String>& xOutputList, unsigned uFlags = 0 );

		//Searches locally and on the network for asura files matching the search string , and then loads them.
		static void Tools_FileSystem::LoadAsuraFiles(const char* szSearchString, unsigned uFlags = 0);

		//Loads and decompresses an asura file into a buffer allocated by the function , without actually parsing it
		//Returns 0 if it fails for some reason
		//pxOuput is an optional pointer for getting additional output
		//If bIncludeFileHeader is true then the buffer will include the 8 byte file header for an uncompressed file
		//at the start. Otherwise the buffer will start at the first chunk
		static char* ReadAsuraFile(const char* szFilename, Tools_ReadAsuraFileOutput* pxOutput = 0,bool bIncludeFileHeader = false,bool bFailureHandledExternally = false);

		//Creates the directory specified if it doesn't already exist (or directory that a file is contained in
		//if a filename is specified)
		static void EnsureDirectoryExists(const char * n, bool bAllowDotInDirectoryName = false);

		static const char* StripPath(const char* pszPath);
		static char* StripPath(char* pszPath);

        static bool FileExists( const char* szFilename );
		static bool FileExists_CheckNetwork( const char* szFilename );
        static bool FileExists_CheckNetworkOnly( const char* szFilename );
		static bool IsReadOnly( const char* szFilename );
		static bool IsDirectory( const char* szFilename );

		static bool MakeWriteable(const char* szFilename);

		// Wrapper for "Asura_FileAccess::Open()" which opens the specified file from
		// the network or local machine depending on which is appropriate.
		static Asura_Handle Open_CheckNetwork(const char* szFilename);

		//returns true if both files exist , and have the same modified time
		//tolerance is tolerance level in seconds for doing the comparison
		static bool DoModifiedTimesMatch(const char* pszSourceFilename,const char* pszDestFilename,float fTolerance = 5.0f);

		//Compares two filetimes with a tolerance level.
		//returns 0 if approximately equal
		//returns -1 if xTime1 is less than xTime2 by more than the tolerance level
		//returns +1 if xTime1 is greater than xTime2 by more than the tolerance level
		static int ApproxCompareFileTime(const FILETIME* pxTime1,const FILETIME* pxTime2,float fTolerance = 5.0f);

		//copies the modified(and accessed/created times I think) from first file to second
		static void CopyFileTimes(const char* pszSourceFilename,const char* pszDestFilename);

		static bool GetModifiedTime(const char* pszFilename,FILETIME* pxModifiedTime);

		//Generates a filename to use for the converted versions of graphics etc during level exports.
		static void GenerateFilenameForConvertedFile(const char* pszOriginalFilename,const char* pszSubdirectoryName,const char* pszExtension,Asura_String& xOutput);

		//Gets the network shadow filename for a converted DDS file (or similar) , from the local
		//filename for it
		//If uDataHash is non-zero then it gets tacked onto the filename for the shadow file
		static bool GetShadowConversionFilename(const char* pszLocalFilename,Asura_String* pxOutputFilename, Asura_Hash_ID uDataHash = 0);
		
		//Copies the local converted file (+ .dat file) to the network shadows
		//szOriginalFile is the source file that was converted
		//pszLocalFilename is the output from the conversion
		//If uDataHash is non-zero then it gets tacked onto the filename for the shadow file
		static bool CopyToConvertedFileShadow(const char* szOriginalFile, const char* pszLocalFilename, Asura_Hash_ID uDataHash = 0);
		
		//Copies the coverted file (+ .dat file) from the network shadows to the local paths
		static bool CopyFromConvertedFileShadow(const char* pszLocalFilename, Asura_Hash_ID uDataHash = 0);
		
		static bool CopyBetweenConvertedFileShadow(const char* pszSourceFilename,const char* pszDestFilename);

		//returns true if file exists , is writeable , and isn't the network copy
		//(ie ,indicates that its one that is currently checked out/being hacked by the user)
		static bool IsWriteableLocalFile(const char* pszFilename);

		//Gets the size of the specified file.
		//If this fails it will return false and set the output size to 0
		static bool GetSizeOfFile(const char* szFilename, u_int& uOutput);

private:
		// Internal helper functions
		static void FindFilesInternal( const char* szSearchString, List<Asura_String>& xOutputList, unsigned uFlags );
		static void FindDirectoriesInternal( const char* szSearchString, List<Asura_String>& xOutputList, unsigned uFlags );
};

#endif//__TOOLS_FILESYSTEM_H__
