/*

DMA subsystem emulation
=======================

This source file handles the emulation of the DMA subsystem (not the VIF or the GIF).

It is initialised with DMA_startDmaTransfer() (when the transfer would normally be
started by writing directly to Dn.CHCR), DMA_sceDmaSend() or DMA_sceDmaSendN().

Once initialised, use the function DMA_emulateCycle() to obtain a new quadword. The
emulator will automatically parse DMA tags and check addresses.

For each call to DMA_emulateCycle, check the error status to see if anything went wrong.
If nothing went wrong, then gDmaTransferQWord will point to the next DMA quadword to
be transferred (if TTE=1, then this qword will point to the tag). The boolean variable
gDmaTransferQWordIsTag indicates whether the qword pointed to is a tag or not.

When gDmaTransferQWord is NULL, the transfer is over.


CHANGES
-------
03/10/2002 - made a bit more codewarrior friendly
25/04/02 - Added warning to DMA check, to warn if using REFE+QWC=0+ADDR=0 (i.e. null dma tag) to end transfer
13/06/02 - Fixed bug where end of memory address was being validated at incorrect address (was using -16 instead of -4)

*/

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "PS2/PS2Strip/DMADebug/DMAEmu.h"

// Stdlib independent.
#include <stdio.h>
#include <string.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"

// PS2Strip.
#include "PS2/PS2Strip/PS2Strip_PS2Stuff.h"

// DMADebug.
#include "PS2/PS2Strip/DMADebug/Debug.h"
#include "PS2/PS2Strip/DMADebug/RTF.h"

// B E G I N   N A M E S P A C E /////////////////////////////////////////////

namespace DMADebug
{

// C O N S T A N T S /////////////////////////////////////////////////////////

static const char *gDmaIdNames[] =
{
	"REFE", "CNT", "NEXT", "REF", "REFS", "CALL", "RET", "END"
};

static const char *gDmaChanNames[] =
{
	"VIF0", "VIF1", "GIF", "fromIPU", "toIPU",
	"SIF0", "SIF1", "SIF2", "fromSPR", "toSPR"
};

static const char *gPceDescription[] = { "No priority", "*RESERVED*", "Disabled", "Enabled" };

// T Y P E S /////////////////////////////////////////////////////////////////

struct DMA_REGISTER 
{
    PS2Strip_DMA_Reg_CHCR chcr; // channel control
    void*                 madr; // memory address
    u_int                 qwc;  // transfer count
    PS2Strip_DMA_Tag*     tadr; // tag address
    void*                 as0;  // address stack
    void*                 as1;  // address stack
    void*                 sadr; // spr address
};

// G L O B A L S /////////////////////////////////////////////////////////////

int gChan;		// The number of the channel we are examining
static DMA_REGISTER *d;
static void *gNextTAddr;		// pointer to where the next tag address will be.

static u_int gTagCount[8];		// Holds how many tags of each type we have encountered.
static u_int gTagTransLengthTotal[8];	// And the total length of data transferred by each type.

// Globally accessible QWord which is what the DMA emulator puts onto the 'bus'.
u_int *gDmaTransferQWord;
bool gDmaTransferQWordIsTag;		// Is this a DMA tag?

// Upper limit on memory size, used for bounds checking
u_int gUpperMainMemLimit = MEM_MAX_128;
static bool gEnableMirrors = false;		// do we consider mirrored memory spaces to be valid?
						// This is used because some parts of VU memory are mirrored in more than
						// one place in memory. By default, this is disabled.

static bool gDumpDmaTags = true;
static bool gDumpDmaStats = true;

//  -------------
//	DMA REGISTERS
//	-------------

// --------- Registers common to all DMA channels
static u_int	d_ctrl;
static u_int	d_stat;
static u_int	d_pcr;
static u_int	d_sqwc;
static u_int	d_rbor;
static u_int	d_rbsr;
static u_int	d_stadr;
static u_int	d_enabler;

static DMA_REGISTER dmaRegs[10];

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static inline void setADR( void* pvDest, void* pvSrc )
{
    if( pvDest != reinterpret_cast<void*>(-1) )
    {
        pvDest = pvSrc;
    }
}

static inline void setQWC( u_int& uDest, u_int uSrc )
{
    uDest = uSrc;
}

static inline void setCHCR( PS2Strip_DMA_Reg_CHCR& xCHCR, u_int uMOD, u_int uDIR )
{
    xCHCR.MOD = uMOD;
    xCHCR.DIR = uDIR;
    xCHCR.STR = 1;
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

// (MP - 16.11.2002) In the original lib, this is supposed to capture the
// current DMA state so it can accurately simulate the DMA transfer. We'll
// just use initial state. This data has come from a mixture of checking
// register state on the PS2 and consulting the (v7) hardware manuals.
static void saveDmaRegisters( void )
{
	d_ctrl    = 0x00000001; //(u_int)DGET_D_CTRL();
	d_stat    = 0x00800000; //(u_int)DGET_D_STAT();
	d_pcr     = 0x00800000; //(u_int)DGET_D_PCR();
	d_sqwc    = 0x00000000; //(u_int)DGET_D_SQWC();
	d_rbor    = 0x00000000; //(u_int)DGET_D_RBOR();
	d_rbsr    = 0x00000000; //(u_int)DGET_D_RBSR();
	d_stadr   = 0x00000000; //(u_int)DGET_D_STADR();
	d_enabler = 0x00000000; //(u_int)DGET_D_ENABLER();

	for( u_int u = 0; u < 10; ++u )
    {
//		sceDmaChan *dc = sceDmaGetChan(i);
		dmaRegs[u].chcr.u32 = 0; //dc->chcr;
		dmaRegs[u].madr     = 0; //(void *)dc->madr;
		dmaRegs[u].qwc      = 0; //(u_int)dc->qwc;
		dmaRegs[u].tadr     = 0; //(sceDmaTag *)dc->tadr;
		dmaRegs[u].as0      = 0; //(void *)dc->as0;
		dmaRegs[u].as1      = 0; //(void *)dc->as1;
		dmaRegs[u].sadr     = 0; //(void *)dc->sadr;
	}
}

// ==========================================================
// CRC functions, for checking to see if data is getting corrupted
// ==========================================================

static int gTableInitialised = 0;
static u_int gCrc;
static u_int crc32_table[256];

#define CRC32_POLY (0x04c11db7)

static void initCrc32(void)
{
	int i, j;
	u_long c;
	for(i = 0; i < 256; i++) {
		for(c = i << 24, j = 8; j > 0; --j) {
			c = c & 0x80000000 ? (c << 1) ^ CRC32_POLY : (c << 1);
		}
		crc32_table[i] = c;
	}
}

static u_int addCRC(const u_char *buf, int len) {
	for(; len > 0; ++buf, --len) {
		gCrc = (gCrc << 8) ^ crc32_table[(gCrc >> 24) ^ *buf];
	}
	return ~gCrc;
}

int DMA_getCRC(u_int *result) {
	int err = DMA_ERR_NO_ERROR;

	if(gTableInitialised == 0) {
		initCrc32();
		gTableInitialised = 1;
	}

	gCrc = 0xffffffff;
	do {
		err |= DMA_emulateDMACycle();
		if(gDmaTransferQWord != NULL) {
			*result = addCRC( reinterpret_cast<const u_char*>(gDmaTransferQWord), 16 );
		}
	} while(!IsFatal(err) && gDmaTransferQWord != NULL);
	return err;
}




// =================================================================
// MINOR SUPPORT FUNCTIONS
// =================================================================


// Output the statistics that have been accumulated so far
// 

static int outputDmaStatistics( void )
{
	int err = DMA_ERR_NO_ERROR;
	int i;
	u_int totalTags = 0, totalSize = 0;
	char *p;

	err |= DEBUG_addToOutputBuffer("==== DMA STATISTICS ====");
	err |= DEBUG_addToOutputBuffer("  Tags      # Tags    QWords Txfr'd");
	err |= DEBUG_addToOutputBuffer("  ----      ------    -------------");

	for(i = 0; i < 8; i++) {
		if(gTagCount[i] != 0) {
			p = gTempString;
			p += sprintf(p, "  %-12s%4d%6s%6d", gDmaIdNames[i],
					gTagCount[i], "", gTagTransLengthTotal[i]);

			totalTags += gTagCount[i];
			totalSize += gTagTransLengthTotal[i];
			err |= DEBUG_addToOutputBuffer(gTempString);
		}
	}
	err |= DEBUG_addToOutputBuffer(   "            ======    =============");
	sprintf(gTempString, "  %12s%4d%6s%6d", "", totalTags, "", totalSize);
	err |= DEBUG_addToOutputBuffer(gTempString);

	return err;
}


// -------------------------------
// ERROR HANDLING
// -------------------------------


// This function acts as a wrapper to DEBUG_AddError. If an address is supplied,
// it will add the address, and address contents to the error string.
// Depending on the severity, it will call DEBUG_addError or DEBUG_addWarning.
// Parameters:
//		errnum - the error number. See Debug.h for those.
//		severity - severity of the error, ie a warning, fatal error or non-fatal error.
//		addr - address where the error was found. Can be NULL.
//		error - string describing the error that was found.

static int AddSysError( int iErrNum, int iSeverity, void* pvAddr, const char* szError )
{
	static char err[1024];

	DMADebug_Puts( "-->Entering AddSysError" );
	sprintf(err, "%s : ", gDmaChanNames[gChan]);
	if( pvAddr )
    {
		pvAddr = DMA_convertToMainMemAddress( pvAddr );
		sprintf( err + strlen(err),
                 DMADEBUG_FMT_ADDRESS_HEX ":[" DMADEBUG_FMT_MEMCONTENTS_64 "] ",
                 reinterpret_cast<u_int>(pvAddr),
                 *(reinterpret_cast<u_int*>(pvAddr) + 1),
                 *(reinterpret_cast<u_int*>(pvAddr)) );
	}
	switch(iSeverity)
    {
		case DMA_ERR_WARNING:
			strcat(err, DMADEBUG_RTF_COLOUR_WARNING);
			break;
		case DMA_ERR_FATAL_CONT:
		case DMA_ERR_FATAL_HALT:
			strcat(err, DMADEBUG_RTF_COLOUR_ERROR);
			break;
	}
	strcat(err, szError);
	strcat(err, DMADEBUG_RTF_COLOUR_NORMAL);
	DMADebug_Puts( "<--Leaving AddSysError" );
	return DEBUG_addError(iErrNum, iSeverity, err);
}

// ----------------------------------------
// MEMORY addres validation and conversion
// ----------------------------------------
	// returns TRUE if the u_int falls in main memory or scratchpad ram.
bool DMA_isValidAddress( void* pvAddr )
{
	u_int a = reinterpret_cast<u_int>( pvAddr );
	u_int b = (a & 0xf0000000) >> 28;
	u_int upperVu0Mem;

	if(b == 2 || b == 3) {
		a = a & 0x0fffffff;
	}
	upperVu0Mem = (gEnableMirrors)? 0x11007fff : 0x11004fff;
	return
			(a >= 0x00100000 && a <= gUpperMainMemLimit) ||		// Main memory (not including kernal area)
			(a >= 0x70000000 && a <= 0x70003fff) ||		// Scratch pad
			(a >= 0x11000000 && a <= 0x11000fff) ||		// VU 0 MicroMem
			(a >= 0x11004000 && a <= upperVu0Mem) ||	// VU 0 Memory
			(a >= 0x1100c000 && a <= 0x1100ffff) ||		// VU 1 Memory
			(a >= 0x11008000 && a <= 0x1100bfff) 		// VU 1 MicroMem
			;
}

// Converts from DMA ADDR field format to a normal EE main memory address.
void* DMA_convertToMainMemAddress( void* pvAddr )
{
	u_int a = reinterpret_cast<u_int>( pvAddr );

// (MP - 18.11.2002) Can't really do this offline.
//	if( a & 0x80000000 )
//    {
//		a = ( (a & 0x0ffffff0) | 0x70000000 );
//	}
//
//    // If Cached or Uncached Accelerated, change to Uncached.
//	if((a>>28) == 0 || (a>>28) == 3)
//    {
//		a = ( (a & 0x0fffffff) | 0x20000000 );
//	}
//
//	DMADebug_Printf(("Converted %p to main memory address: %p.\n", addr, (u_int *)a));

	return( reinterpret_cast<void*>(a) );
}

static void* convertToDmaAddress( void* pvAddr )
{
	u_int a = reinterpret_cast<u_int>( pvAddr );
	if( (a & 0xf0000000) == 0x70000000)
    {
		a &= 0x0fffffff;
		a |= 0x80000000;
	}
	return( reinterpret_cast<void*>(a) );
}

// Validates an address in DMA format, ie top bit is SPR.
static int validateDmaAddress(void *address, const char *description)
{
	int err = DMA_ERR_NO_ERROR;

// (MP - 18.11.2002) Can't really do this offline...
//	DMADebug_Puts("--> Entering validateDmaAddress");
//
//		// Extra information - if the address is 0x7***, then they've just given
//		// the scratch pad address without setting the top bit.
//	if(((u_int)address & 0xf0000000) == 0x70000000) {
//		sprintf(gTempString, "%s at %p looks like a scratchpad address, but bit 31 is not set.", description, address);
//		err |= AddSysError(WARN_DMA_LOOKS_SPR, DMA_ERR_WARNING, NULL, gTempString);
//	}
//
//	if(((u_int)address & 0xf) != 0) {
//		sprintf(gTempString, "%s at %p is not quadword aligned.", description, address);
//		err |= AddSysError(ERR_DMA_ALIGNMENT, DMA_ERR_FATAL_HALT, NULL, gTempString);
//	}
//
//	if(!DMA_isValidAddress(DMA_convertToMainMemAddress(address))) {
//		sprintf(gTempString, "%s (0x%08x) is not in a valid address range", description, (u_int)address);
//		err |= AddSysError(ERR_DMA_INVALID_ADDRESS, DMA_ERR_FATAL_HALT, NULL, gTempString);
//	}
//	DMADebug_Puts("<-- Leaving validateDmaAddress");

	return err;
}




//==============
static void putTransferSourceAddress(void *addr) {
	if(gChan == PS2STRIP_DMA_CHAN_ID_FROMSPR || gChan == PS2STRIP_DMA_CHAN_ID_TOSPR) {
		d->sadr = (void *)((u_int)addr & 0x3ff0);
	} else {
		d->madr = addr;
	}
}
//==============
static void *getTransferSourceAddress(void) {
	void *addr;
	if(gChan == PS2STRIP_DMA_CHAN_ID_FROMSPR || gChan == PS2STRIP_DMA_CHAN_ID_TOSPR) {
		addr = d->sadr;
	} else {
		addr = d->madr;
	}
	return addr;
}
//==============
static void incrementTransferAddress(void) {
	char *saddr;
	saddr = (char *)getTransferSourceAddress();
	saddr += 16;
	putTransferSourceAddress((void *)saddr);
}

//==============
// Sets the global QWord which is transferred by the DMA
// This emulates the qword being transferred.
static void setDmaTransferQWord(void *address, bool isTag)
{
	gDmaTransferQWord = reinterpret_cast<u_int*>( address );
	gDmaTransferQWordIsTag = isTag;
}

//==============
		// Turns off the DMA
static void endDmaTransfer(void) {
	d->chcr.STR = 0;
    gDmaTransferQWord = NULL;
    gDmaTransferQWordIsTag = false;

	if(gDumpDmaStats && d->chcr.MOD == 1) {		// Only worth outputting stats if we're in Source Chain mode
		outputDmaStatistics();
	}
}

//==============
// Read from MADR, pop it in the global register, increment MADR and decrease QWC.
static int transferQWord(void)
{
	int err = DMA_ERR_NO_ERROR;
	void *saddr;
	saddr = getTransferSourceAddress();
	err |= validateDmaAddress(saddr, "DMA data");
	saddr = DMA_convertToMainMemAddress(saddr);
	setDmaTransferQWord( saddr, false );
	
	incrementTransferAddress(); // Increment MADR/SADR
	d->qwc--;
	return err;
}

//==============

static int interpretSourceChainTag(void) {
	int err = DMA_ERR_NO_ERROR;
	u_int *tagAddress, *addr, lo, hi;
	u_int tte, id, qwc, irq, pce, callStackDepth;		// holds DMA channel info
	static char s[1024];
	
	int i, numBlocks;		// number of different areas in memory that this DMA tag will transfer from.
						// This has a min of 0 and a max of 2.
	u_int *startTransAddr[2], *endTransAddr[2];
	
	DMADebug_Puts("--> Entering interpretSourceChainTag");

		// --------- Get the tag (tagAddress in d->tadr is assumed to be correct)
		// --------- and suck all the data out of it.
	tagAddress = reinterpret_cast<u_int*>( DMA_convertToMainMemAddress(d->tadr) );
	lo = tagAddress[0];
	hi = tagAddress[1];
	id = (lo >> 28) & 0x7;
	qwc = lo & 0xffff;
	tte = d->chcr.TTE;
	irq = lo >> 31;
	pce = (lo >> 26) & 0x3;
	addr = (u_int *)tagAddress[1];
	callStackDepth = d->chcr.ASP;

		// ------- Dump what we have so far

	s[0] = 0;

		// Add the name of the channel, the TTE state, the tagAddress of the tag, and the
		// contents of the tag, followed by the tag ID (the type of tag).
	DEBUG_addField(DBGMSG_ENTRY_TITLE, s, DMADEBUG_RTF_COLOUR_DMATITLE "DMA %s%s %-4s" DMADEBUG_RTF_COLOUR_NORMAL " " DMADEBUG_FMT_DATASRCADDR,
				// Generate output prefix (indentation for 'CALL/RET')
			(callStackDepth >= 1) ? "->" : "  ",
			(callStackDepth == 2) ? "->" : "  ",
			gDmaIdNames[id],
			(u_int)tagAddress);

	DEBUG_addField(DBGMSG_ENTRY_FIELD, s, "["DMADEBUG_FMT_MEMCONTENTS_64 "]", hi, lo);			// Tag contents (Hi/Lo)
	DEBUG_addField(DBGMSG_ENTRY_FIELD, s, "CHAN="DMADEBUG_FMT_DATA_STR,  gDmaChanNames[gChan]);		// Channel name
	DEBUG_addField(DBGMSG_ENTRY_FIELD, s, "TTE="DMADEBUG_FMT_DATA_DEC, tte);									// TTE status
	DEBUG_addField(DBGMSG_ENTRY_FIELD, s, "IRQ="DMADEBUG_FMT_DATA_DEC, irq);									// 'interrupt'
	DEBUG_addField(DBGMSG_ENTRY_FIELD, s, "PCE="DMADEBUG_RTF_COLOUR_DATA "%11s" DMADEBUG_RTF_COLOUR_NORMAL, gPceDescription[pce]);		// PCE status
	DEBUG_addField(DBGMSG_ENTRY_FIELD, s, "QWC:" DMADEBUG_FMT_DATA_4HEX DMADEBUG_RTF_COLOUR_NORMAL, qwc);						// QuadWordCount

		// ---------------- Tag field boundary checks

	if(pce == 1) {
		err |= AddSysError(ERR_RESERVED_PCE, DMA_ERR_FATAL_CONT, tagAddress, "PCE field is 1, which is reserved.");
	}

		// Warning check on ignored bits above the QuadWordCount
	if(((lo >> 16) & 0x3ff) != 0) {
		err |= AddSysError(WARN_UNUSED_TAG_BITS_NON_ZERO, DMA_ERR_WARNING, tagAddress,
				"Bits 16-25 in this DMA tag are not 0. You may be trying to transfer >= 1Mb?");
	}

		// Sanity check to see that certain tags are not used on certain DMA channels.
	if(id == DMA_REFS && !(gChan == PS2STRIP_DMA_CHAN_ID_VIF1 || gChan == PS2STRIP_DMA_CHAN_ID_SIF1 || gChan == PS2STRIP_DMA_CHAN_ID_GIF)) {
		err |= AddSysError(ERR_INVALID_TAG_FOR_CHANNEL, DMA_ERR_FATAL_CONT, tagAddress,
			"DMA tag of type REFS encountered, but this is only permitted on VIF1, GIF & SIF1 channels!");
	}
	if(	(id == DMA_CALL || id == DMA_RET) && !(gChan == PS2STRIP_DMA_CHAN_ID_VIF0 || gChan == PS2STRIP_DMA_CHAN_ID_VIF1 || gChan == PS2STRIP_DMA_CHAN_ID_GIF)) {
		err |= AddSysError(ERR_INVALID_TAG_FOR_CHANNEL, DMA_ERR_FATAL_CONT, tagAddress,
			 "DMA tag of type CALL/RET encountered, but this is only permitted on VIF0, VIF1 & GIF channels!");
	}

			// If any of these tags aren't being used, we must check ADDR field for sanity
	if(!(id == DMA_CNT || id == DMA_RET || id == DMA_END)) {
		if(qwc == 0 && id == DMA_REFE && addr == 0) {		// Warn REFE with QWC=0 and ADDR=0, (i.e. whole DMA tag is 0)
			err |= AddSysError(WARN_DMA_NULL_REFE, DMA_ERR_WARNING, tagAddress,
				"REFE tag's QWC=0 and trying to transfer from ADDR=0 - accidentally nulled DMA tag?");
		} else {
			err |= validateDmaAddress(addr, "ADDR field");
			addr = reinterpret_cast<u_int*>( DMA_convertToMainMemAddress(addr) );
			if(IsFatal(err)) {
				err |= AddSysError(ERR_DMA_INVALID_ADDRESS, DMA_ERR_FATAL_CONT, tagAddress, "DMA tag ADDR field is invalid");
			}
		}
	}

		// --------------- Display tagAddress ranges transferred

		// Fill in the arrays holding details of the tagAddress ranges to transfer
		// (There can be 0, 1 or 2 tagAddress ranges per DMA tag depending on the tag ID, TTE and QWC)
	numBlocks = 0;
	if(id == DMA_REF || id == DMA_REFE || id == DMA_REFS) {
		if(tte) {
			startTransAddr[numBlocks] = tagAddress;
			endTransAddr[numBlocks] = tagAddress + 4;
			numBlocks++;
		}
		if(qwc > 0) {
			startTransAddr[numBlocks] = reinterpret_cast<u_int*>( DMA_convertToMainMemAddress(addr) );
			endTransAddr[numBlocks] = startTransAddr[numBlocks] + qwc * 4;
			numBlocks++;
		}
	} else {
		if(tte) {
			startTransAddr[numBlocks] = tagAddress;
		} else {
			startTransAddr[numBlocks] = tagAddress + 4;
		}
		if(qwc > 0 || tte) {
			endTransAddr[numBlocks] = tagAddress + 4 + qwc * 4;
			numBlocks++;
		}
	}

		// Memory bounds checking on the data ranges being transferred
	for(i = 0; i < numBlocks; i++) {
		int error;

		error = DMA_ERR_NO_ERROR;
		error |= validateDmaAddress(convertToDmaAddress(startTransAddr[i]), "Transfer start address");
		error |= validateDmaAddress(convertToDmaAddress(endTransAddr[i]-4), "Transfer end address");
		if(IsFatal(error)) {
			sprintf(gTempString, "Data to be transferred falls outside main mem/scratch pad RAM. "
					"[%08x -> %08x]", (u_int)startTransAddr[i], (u_int)endTransAddr[i]);
			err |= AddSysError(ERR_DMA_INVALID_ADDRESS, DMA_ERR_FATAL_CONT, NULL, gTempString);
			break;
		}
	}

	DMADebug_Puts("About to display blocks");

		// Display the areas of memory that are being transferred. (0, 1 or 2 areas)
	if(numBlocks == 0) {
		DEBUG_addField(DBGMSG_ENTRY_FIELD_NL, s, "Transfers from: ["DMADEBUG_RTF_COLOUR_ADDRESS"No data is transferred"DMADEBUG_RTF_COLOUR_NORMAL"]");
	} else {
		DMADebug_Puts("++++blocks");
		for(i = 0; i < numBlocks; i++) {
			DEBUG_addField(DBGMSG_ENTRY_FIELD_NL, s, "Transfers: ["DMADEBUG_FMT_ADDRESS_HEX "->" DMADEBUG_FMT_ADDRESS_HEX ") [len(qwords):"DMADEBUG_FMT_DATA_HEX "]%s",
				(u_int)startTransAddr[i], (u_int)endTransAddr[i], ((u_int)endTransAddr[i] - (u_int)startTransAddr[i]) >> 4,
				(i == 0 && tte)?" (Includes DMA tag)":"");
		}
	}

			// ------------ Update statistics

	gTagCount[id]++;
	for(i = 0; i < numBlocks; i++) {
		gTagTransLengthTotal[id] += ((u_int)endTransAddr[i] - (u_int)startTransAddr[i]) >> 4;
	}

	DMADebug_Puts("About to calculate next tag address");

			// ------------ Calculate next Tag address, and set MADDR and QWC.

	d->qwc = qwc;

	switch(id) {
		case DMA_CNT:
			gNextTAddr = tagAddress + 4 + qwc * 4;
			d->madr = convertToDmaAddress(tagAddress + 4);
			break;

		case DMA_REF:
		case DMA_REFS:
		case DMA_REFE:
			if(id == DMA_REFE) {
				gNextTAddr = NULL;
			} else {
				gNextTAddr = convertToDmaAddress(tagAddress + 4);
			}
			d->madr = convertToDmaAddress(addr);
			break;

		case DMA_END:
			gNextTAddr = NULL;
			d->madr = convertToDmaAddress(tagAddress + 4);
			break;

		case DMA_RET:
			DMADebug_Printf(("RET encountered - ASP = %d\n", d->chcr.ASP));
			if(d->chcr.ASP == 0)
            {
				gNextTAddr = NULL;
				err |= AddSysError(WARN_PREMATURE_END, DMA_ERR_WARNING, tagAddress,
							 "RET tag encountered, but the call stack is not empty!");
			}
            else if(d->chcr.ASP == 1)
            {
				gNextTAddr = d->as0;
			}
            else if(d->chcr.ASP == 2)
            {
				gNextTAddr = d->as1;
			}
            else
            {
				Asura_Assert( false, "Logic error - ASP of CHCR is > 2" );
			}
			d->madr = convertToDmaAddress(tagAddress + 4);
			break;

		case DMA_NEXT:
			gNextTAddr = convertToDmaAddress(addr);
			d->madr = convertToDmaAddress(tagAddress + 4);
			break;

		case DMA_CALL:
			DMADebug_Printf(("CALL encountered - ASP = %d\n", d->chcr.ASP));
			gNextTAddr = convertToDmaAddress(addr);
			if(d->chcr.ASP == 0) {
				d->as0 = tagAddress + 4 + qwc * 4;
				d->chcr.ASP++;
			} else if(d->chcr.ASP == 1) {
				d->as1 = tagAddress + 4 + qwc * 4;
				d->chcr.ASP++;
			} else if(d->chcr.ASP == 2) {
				gNextTAddr = NULL;
				err |= AddSysError(ERR_CALLSTACK_OVERFLOW, DMA_ERR_FATAL_HALT, tagAddress,
					"DMA tag of type CALL encountered, but call stack is full!");
			}
			d->madr = convertToDmaAddress(tagAddress + 4);
			break;
	}

	if(gDumpDmaTags) {
		err |= DEBUG_addToOutputBuffer(s);
	}

	DMADebug_Puts("<-- Leaving interpretSourceChainTag");

	return err;
}

// =================================================================
//                   MAJOR SUPPORT FUNCTIONS
// =================================================================
	// The DMA Cycle in source Chain mode can be viewed like this:
	// 1) if QWC > 0 then transfer from MADR until QWC = 0;
	// 2) examine tag in TADR, which sets ASP, MADR, QWC and gNextTaddr.
	// 3) if QWC > 0 then transfer from MADR until QWC = 0;
	// 4) if tag == RET && ASP == 0, || tag = END || tag = REFE, then end.
	// 5) TADR = gNextTaddr
	// 6) GOTO 2
// To help duplicate this loop, we use 'gNextSCProcessingStage' to tell us what the next stage should be.
// This is necessary, because when QWC reaches 0, we have to know whether to move onto stage 2 or stage 4.
// 'nextSCProcessingStage' should be set to INTERPRET_TAG if QWC > 0 when the Source Chain transfer is started.
enum PROCESSING_STAGE { INTERPRET_TAG, CHECK_DMA_END_TRANSFER };
static PROCESSING_STAGE gNextSCProcessingStage;

static int emulateChainDmaCycle(void)
{
	int err = DMA_ERR_NO_ERROR;
	void *tadr;

	bool finished = false;
	DMADebug_Puts("--> Entering emulateChainDmaCycle");
	do {
		if(d->qwc != 0) {
			DMADebug_Puts("}}}}}}} Transferred Qword");
			err |= transferQWord();
			finished = true;
		} else {
			if(gNextSCProcessingStage == INTERPRET_TAG) {
				DMADebug_Puts(" ---- Interpret tag");
				d->tadr = reinterpret_cast<PS2Strip_DMA_Tag*>( gNextTAddr ); // set the next tag address
				tadr = d->tadr;
				err |= validateDmaAddress(tadr, "DMA Tag");
				if(IsFatal(err)) {
					break;
				}
				tadr = DMA_convertToMainMemAddress(tadr);
				DMADebug_Puts(" About to interpret source tag");

					// Tag address is OK, read it, interpret it and set gNextTAddr, MADR, QWC
				err |= interpretSourceChainTag();
				DMADebug_Puts(" Finished interpreting source tag");
				if(d->chcr.TTE == 1) {
					setDmaTransferQWord( tadr, true );
					finished = true;
				}
				gNextSCProcessingStage = CHECK_DMA_END_TRANSFER;
				DMADebug_Puts(" ==== Interpret tag");

			} else {
				u_int id;
				DMADebug_Puts(" ---- Check end");
				tadr = DMA_convertToMainMemAddress(d->tadr);
				id = (*(u_int*)tadr >> 28) & 0x7;
				if(id == DMA_REFE || (id == DMA_RET && d->chcr.ASP == 0) || id == DMA_END) {
					if((id == DMA_REFE || id == DMA_END) && d->chcr.ASP != 0) {
						err |= AddSysError(WARN_PREMATURE_END, DMA_ERR_WARNING, (u_int*)d->tadr,
								 "END/REFE tag encountered, but the call stack is not empty!");
					}
					endDmaTransfer();
					finished = true;
				}
                else if(id == DMA_RET && d->chcr.ASP > 0)
                {
					d->chcr.ASP--;
				}
			
				gNextSCProcessingStage = INTERPRET_TAG;
				DMADebug_Puts(" ==== Check end");
			}
		}
	} while(!IsFatal(err) && !finished);
	DMADebug_Puts("<-- Leaving emulateChainDmaCycle");
	return err;
}


static int emulateNormalDmaCycle(void) {
	int err = DMA_ERR_NO_ERROR;

	DMADebug_Puts("--> Entering emulateNormalDmaCycle");
	if(d->qwc == 0) {
		endDmaTransfer();
	} else {
		err |= transferQWord();
	}
	DMADebug_Puts("<-- Leaving emulateNormalDmaCycle");
	return err;
}



		// This function assumes that the DMA transfer has been initialised with DMA_startTransfer.
		// It will emulate a DMA cycle, placing a pointer to the word being transferred into
		// gDmaTransferQWord (it's NULL if there's nothing to transfer). At the end of the function,
		// gDmaTransferQWord will point to the word to be transferred, or NULL if we're at the end.

		// gDmaTeansferQWordIsTag will be TRUE if the word pointed to is a tag.

int DMA_emulateDMACycle(void) {
	int err = DMA_ERR_NO_ERROR;

	DMADebug_Puts("--> Entering DMA_emulateDMACycle");

	gDmaTransferQWord = NULL;
	gDmaTransferQWordIsTag = false;

	if(d->chcr.MOD == 0) {
		err |= emulateNormalDmaCycle();
	} else {
		err |= emulateChainDmaCycle();
	}

	DMADebug_Puts("<-- Leaving DMA_emulateDMACycle");
	return err;
}


static int startup(int channelNum) {
	int err = DMA_ERR_NO_ERROR;
	int i;

	saveDmaRegisters();
	VIF_saveVifRegisters();
	GIF_saveGifRegisters();

	GIF_reset();			// added V1.0, even though it's possible to transfer data to the GIF when it isn't idle.
							// however, the likelihood of wanting to do this is pretty low.

	DEBUG_reset();			// Clear output buffers

	gChan = channelNum;
	d = &dmaRegs[gChan];
	gNextTAddr = d->tadr;

	gNextSCProcessingStage = INTERPRET_TAG;

	for(i = 0; i < 8; i++) {
		gTagCount[i] = 0;
		gTagTransLengthTotal[i] = 0;
	}

	gDmaTransferQWord = NULL;
	gDmaTransferQWordIsTag = false;

	return err;
}

static int startupSanityChecks(void)
{
	int err = DMA_ERR_NO_ERROR;

		// Sanity checks on register settings
	if((d->qwc != 0) && d->chcr.MOD == 1) {		// if mode = chain && qwc != 0)
		err |= AddSysError(WARN_DMA_QWC_NOT_ZERO, DMA_ERR_WARNING, gNextTAddr,
				"The DMA transfer was started in Chain Mode, but QWC != 0");
	}
	if(d->qwc == 0 && d->chcr.MOD == 0) {
		err |= AddSysError(WARN_DMA_NORMAL_QWC_ZERO, DMA_ERR_WARNING, NULL,	
				"'Normal' Mode DMA transfer started, but QWC == 0!");
	}
	if((gChan == PS2STRIP_DMA_CHAN_ID_VIF0 || gChan == PS2STRIP_DMA_CHAN_ID_VIF1 || gChan == PS2STRIP_DMA_CHAN_ID_GIF) && d->chcr.ASP != 0) {
		err |= AddSysError(WARN_DMA_ASP_NOT_ZERO, DMA_ERR_WARNING, NULL,
				"ASP (DMA call stack depth) is not zero at beginning of transfer.");
	}
	if((gChan == PS2STRIP_DMA_CHAN_ID_VIF1 || gChan == PS2STRIP_DMA_CHAN_ID_SIF2) && d->chcr.DIR == 0) {
		err |= AddSysError(ERR_DMA_INVALID_DIR, DMA_ERR_FATAL_HALT, NULL,
			"Sorry, these modules are not designed to handle DMA DIR = 0 (to memory)");
	}

	if(d->chcr.MOD != 0 && d->chcr.MOD != 1) {
		err |= AddSysError(ERR_DMA_INVALID_MOD, DMA_ERR_FATAL_HALT, NULL,
			"Sorry, these modules are not designed to handle DMA MOD == 2 or 3");
	}

    // TTE=1 and Mode=Normal is suspicious
	if(d->chcr.TTE == 1 && d->chcr.MOD == 0) {
		err |= AddSysError(WARN_DMA_TTE_IN_NORM_MODE, DMA_ERR_WARNING, NULL,
			"TTE is set to 1 during a 'Normal' mode transfer");
	}

    // TTE=1 on non VIF channel
	if(d->chcr.TTE == 1 && d->chcr.MOD == 1 && gChan != PS2STRIP_DMA_CHAN_ID_VIF0 && gChan != PS2STRIP_DMA_CHAN_ID_VIF1) {
		err |= AddSysError(WARN_DMA_TTE_IN_NONVIF_CHANNEL, DMA_ERR_WARNING, NULL,
			"TTE is set to 1 for a transfer to the GIF channel");
	}

	return err;
}


// ===================================================

void DMA_parseConfig( const char *cs )
{
			// Parse config string
	gDumpDmaTags = true;
	gDumpDmaStats = true;
	gUpperMainMemLimit = MEM_MAX_128;
	gEnableMirrors = false;
	if(cs != NULL) {
		if(strstr(cs, "-nodmatags") != NULL) {
			gDumpDmaTags = false;
		}
		if(strstr(cs, "-nodmastats") != NULL) {
			gDumpDmaStats = false;
		}
		if(strstr(cs, "-mem32") != NULL) {
			gUpperMainMemLimit = MEM_MAX_32;
		}
		if(strstr(cs, "-usemirrors") != NULL) {
			gEnableMirrors = true;
		}
	}
}


		// This sets up the emulator for a DMA transfer. After this call, you can call DMA_emulateCycle to
		// get the next qword to be transferred.
		// Use this if you set CHCR.STR manually.

int DMA_startDmaTransfer( int channelNum, u_int chcr )
{
	int err = DMA_ERR_NO_ERROR;

	err |= startup(channelNum);
	d->chcr = *(reinterpret_cast<PS2Strip_DMA_Reg_CHCR*>(&chcr));

	err |= startupSanityChecks();
	return err;
}



		// This is similar to the DMA_startDmaTransfer function above.
		// Use this if you start the transfer by calling sceDmaSend()

		// Send by Source Chain
int DMA_sceDmaSend( PS2STRIP_DMA_CHAN_ID eChanID, void* pvTag )
{
	int err = DMA_ERR_NO_ERROR;

	err |= startup( eChanID );

	gNextTAddr = pvTag;
	setADR(d->tadr, pvTag);
	setQWC(d->qwc,  0);
	setCHCR(d->chcr, 1, 1);

	err |= startupSanityChecks();
	return err;
}


		// This is similar to the DMA_startDmaTransfer function above.
		// Use this if you start the transfer by calling sceDmaSendN()

		// Send by Normal
int DMA_sceDmaSendN( PS2STRIP_DMA_CHAN_ID eChanID, void* pvAddr, int iNumQWs )
{
	int err = DMA_ERR_NO_ERROR;

	err |= startup( eChanID );

	setADR(d->madr, pvAddr);
	setQWC(d->qwc,  iNumQWs);
	setCHCR(d->chcr, 0, 1);

	err |= startupSanityChecks();
	return err;
}

// E N D   N A M E S P A C E /////////////////////////////////////////////////

}
