//
// Tools_Mode_ShapeData.cpp
//
// Created by Andrew Weinkove - 07/04/2003
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Shapes/Tools_Mode_ShapeData.h"

#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_DebugText.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Dialog.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Module.h"
#include "Shapes/Asura_Shapes.h"

#include "Tools_Overlays.h"
#include "Load&Save/Tools_AsuraFileContents.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Objects/Tools_Object_System.h"
#include "Objects/Tools_VolumeEdit_Object.h"
#include "PC/FileSystem/Tools_FileSystem.h"


Asura_Vector_2 Tools_Mode_ShapeData::s_xShapeDataOverlayPosition(2, 2);
bool Tools_Mode_ShapeData::s_bChangedCustomBB = false;
bool Tools_Mode_ShapeData::s_bChanged = false;
bool Tools_Mode_ShapeData::s_bChangedBB = false;
bool Tools_Mode_ShapeData::s_bUseCustomBB = false;
bool Tools_Mode_ShapeData::s_bHasNoCollisions = false;
bool Tools_Mode_ShapeData::s_bIsNeverRendered = false;
unsigned Tools_Mode_ShapeData::s_uFlags = 0;
unsigned Tools_Mode_ShapeData::s_uVolumeEditID = 0;
Asura_Bounding_Box Tools_Mode_ShapeData::s_xBB;
Asura_Vector_3 Tools_Mode_ShapeData::s_xOffset(0.0f, 0.0f, 0.0f);
Asura_Mesh* Tools_Mode_ShapeData::s_pxMesh = 0;
Asura_String Tools_Mode_ShapeData::s_xFragmentName;

void Tools_Mode_ShapeData::InitialiseMode()
{
	Tools_Object_System::EnableAllControls(true);
	s_uVolumeEditID = 0;
	s_pxMesh = 0;
	RecalculateOffset();
	UpdateFragmentName();

	CreateShapeDataOverlay();
}

void Tools_Mode_ShapeData::ShutdownMode()
{
	if(s_uVolumeEditID)
	{
		Tools_VolumeEdit_Object::Delete(s_uVolumeEditID);
	}
	Tools_Object_System::EnableAllControls(false);

	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SHAPEDATA);
}


void Tools_Mode_ShapeData::Update()
{
	RecalculateOffset();
	if (s_bChangedCustomBB)
	{
		if (s_bUseCustomBB)
		{
			s_uFlags |= ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB;
			SetBoxToDefaults();
		}
		else
		{
			s_uFlags &= ~ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB;
			if(s_uVolumeEditID)
			{
				Tools_VolumeEdit_Object::Delete(s_uVolumeEditID);
			}
		}
		Tools_SaveAsura::SetMainFileModified(true);
		s_bChangedCustomBB = false;
	}
	if (s_bChanged)
	{
		if (s_bHasNoCollisions)
		{
			s_uFlags |= ASURA_SHAPE_DATA_FLAGS_HAS_NO_COLLISIONS;
		}
		else
		{
			s_uFlags &= ~ASURA_SHAPE_DATA_FLAGS_HAS_NO_COLLISIONS;
		}
		if(s_bIsNeverRendered)
		{
			s_uFlags |= ASURA_SHAPE_DATA_FLAGS_NEVER_RENDERED;
		}
		else
		{
			s_uFlags &=~ ASURA_SHAPE_DATA_FLAGS_NEVER_RENDERED;
		}
		Tools_SaveAsura::SetMainFileModified(true);
		s_bChanged = false;
	}
}

void Tools_Mode_ShapeData::RenderWorld_PostToneMap()
{
	if (s_uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB)
	{
		s_xBB.Displace(s_xOffset);
		Tools_Render::RenderCube(s_xBB, 0xffafafff, true);
		s_xBB.Displace(-s_xOffset);
	}
}

bool Tools_Mode_ShapeData::IsModified()
{
	return Tools_SaveAsura::IsMainFileModified();
}

void Tools_Mode_ShapeData::QuickSave()
{
	Tools_SaveAsura::QuickSaveFile();
}

unsigned Tools_Mode_ShapeData::GetFlags()
{
	return s_uFlags;
}

Asura_Bounding_Box *Tools_Mode_ShapeData::GetBB()
{
	if (s_uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB)
	{
		return &s_xBB;
	}
	return NULL;
}

Asura_Hash_ID Tools_Mode_ShapeData::GetHashID()
{
	return GetHashID(Tools_SaveAsura::GetMainFilename(),true);
}

Asura_Hash_ID Tools_Mode_ShapeData::GetHashID(const char* pszFilename,bool bAssertIfNotObjectDir)
{
	// AW - 29/04/03 - An explanation of the following:
	// Martin(MGH) had experienced a problem when setting up custom BB for files
	// whose paths or filenames exceeded 8 characters. On his machine, the file and 
	// path names were truncated to the form xxxxxx~#, this then messed up the hash
	// calculation with the result that the system did not work.
	// I am unable to reproduce this bug on my machine, and it only occured on his
	// when double-clicking on a file name (using command line or right-clicking and
	// choosing 'open with' did not cause the problem.) I am not sure how many people
	// will be affected by this, so I will assume that it is widespread.
	// The following 2 lines get the longer version of the filename and path.
	char szLongFilenameBuffer[255];
	GetLongPathName(pszFilename, szLongFilenameBuffer, 255);

	char* szRelativePath = strstr( szLongFilenameBuffer, "objects" );
	if (!szRelativePath)
	{
		szRelativePath = strstr( szLongFilenameBuffer, "Objects" );
	}
	if (!szRelativePath)
	{
		szRelativePath = strstr( szLongFilenameBuffer, "OBJECTS" );
	}
	char* szObjectsRelativePath = szRelativePath;
	if(!szObjectsRelativePath)
	{
		if(bAssertIfNotObjectDir)
		{
			Asura_Assert( NULL != szObjectsRelativePath, "Serious error - Object not in 'objects' directory (or perhaps it's because the test is case sensitive)" );
		}
		return ASURA_HASH_ID_UNSET;
	}

	Asura_String xHashingName;
	if( NULL != szObjectsRelativePath )
	{
		xHashingName = szObjectsRelativePath;
	}
	else
	{
		xHashingName = szLongFilenameBuffer;
	}

	// Convert the filename to a HASH value
	Asura_Hash_ID uHashID = Asura_GetHashID( xHashingName.GetString() );
	Asura_Assert( ASURA_HASH_ID_UNSET != uHashID, "Error - Hash value equals the 'UNSET' value" );
	return uHashID;
};

void Tools_Mode_ShapeData::SetData(Asura_Hash_ID uHashID,unsigned uFlags, Asura_Bounding_Box *pxBB)
{
	static bool s_bFileHashInitialised = false;
	static Asura_Hash_ID s_uFileHash;
	if(!s_bFileHashInitialised)
	{
		//get the hash ID of the main file being loaded
		extern char* DefaultFileToLoad;
		s_uFileHash = GetHashID(DefaultFileToLoad,false);
		s_bFileHashInitialised;
	}

	//ignore the values passed , unless they are for the main asura file
	if(uHashID != s_uFileHash) return;


	s_uFlags = uFlags;
	if (uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB)
	{
		s_bUseCustomBB = true;
		Asura_Assert(pxBB, "Expect valid BB");
		s_xBB = *pxBB;
	}
	if (uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_NO_COLLISIONS)
	{
		s_bHasNoCollisions = true;
	}
	if (uFlags & ASURA_SHAPE_DATA_FLAGS_NEVER_RENDERED)
	{
		s_bIsNeverRendered = true;
	}

}

void Tools_Mode_ShapeData::CreateShapeDataOverlay()
{
	Asura_Overlay* pxNewOverlay;
	pxNewOverlay = new Asura_Overlay("Shape Properties",false,&s_xShapeDataOverlayPosition);
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
	
	AddOverlayControls(pxNewOverlay,xOffset);

	pxNewOverlay->m_uID = TOOLS_OVERLAY_SHAPEDATA;
	Asura_Overlay_System::SubmitNewOverlay(pxNewOverlay,false);

	UpdateOverlay();
}

void Tools_Mode_ShapeData::AddOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset)
{
	
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Tickbox_Dialog("Uses Custom Dynamics BB",xOffset,&s_bUseCustomBB, &s_bChangedCustomBB);
	pxDialog->SetHelpText("Shape uses a dynamics bounding box that differs from the meshes bounding box.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Has no dynamics collisions",xOffset,&s_bHasNoCollisions, &s_bChanged);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Never Rendered",xOffset,&s_bIsNeverRendered, &s_bChanged);
	pxDialog->SetHelpText("If this flag is set then no render mesh will be exported during level exports.\nThis is for meshes that are only going to be used for collisions.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("Edit Dynamics BB",xOffset,EditBoundingBox, NULL, HasBoundingBox);
	pxDialog->SetHelpText("Edit the custom dynamics bounding box.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Fragment: ",xOffset,SelectFragmentType);
	pxDialog->SetHelpText("Select the file containing the fragments for this shape.");
	pxDialog->m_iControlID = DialogID_Fragment;
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
}

void Tools_Mode_ShapeData::UpdateOverlay()
{
	Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(TOOLS_OVERLAY_SHAPEDATA);
	if(pxOverlay)
	{
		//update the fragment type text
		char szText[200];
		Asura_PrintfToBfr(szText,"Fragment: %s",200,true,s_xFragmentName.GetString());

		pxOverlay->UpdateDialogLabel(DialogID_Fragment,szText);
	}
}


void Tools_Mode_ShapeData::EditBoundingBox()
{
	//see if the volume edit object already exists
	Tools_SaveAsura::SetMainFileModified(true);
	Tools_VolumeEdit_Object* pxVolumeObject = Tools_VolumeEdit_Object::Find(s_uVolumeEditID);

	if(!pxVolumeObject)
	{
		pxVolumeObject = new Tools_VolumeEdit_Object("Dynamics Bounding Box",&s_xBB,0,&s_bChangedBB);
		s_uVolumeEditID = pxVolumeObject->GetID();
	}
	pxVolumeObject->HiddenFlags |= ObjectHideFlag_ObjectSystem;

	Tools_Object_System::SelectObject(pxVolumeObject);
}

bool Tools_Mode_ShapeData::HasBoundingBox()
{
	return s_bUseCustomBB;
}

void Tools_Mode_ShapeData::SetBoxToDefaults()
{
	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		if(pxMesh && !pxMesh->IsDetailMesh())
		{
			s_xBB = pxMesh->BoundingBox;
			s_xOffset = Asura_Module_System::GetPositionOfModule(i);
		}
	}
}


void Tools_Mode_ShapeData::RecalculateOffset()
{
	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		if(pxMesh && !pxMesh->IsDetailMesh())
		{
			s_pxMesh = pxMesh;
			s_xOffset = Asura_Module_System::GetPositionOfModule(i);
		}
	}
	Asura_Assert(s_pxMesh,"Failed to find an appropriate mesh");
}

void Tools_Mode_ShapeData::SelectFragmentType()
{
	//ask for which fragment file to use
	Tools_FileSystem_DialogInfo	xFileSelector;

	xFileSelector.DefaultExtension		= ".ASR";
	xFileSelector.DefaultFilename		= "";
	xFileSelector.DefaultPath			= "Objects\\";
	xFileSelector.FileFilter			= "Fragment Files\0*.ASR;\0";
	xFileSelector.WindowTitle			= "Select Fragment File";
	xFileSelector.AllowMultipleSelect	= false;

	Asura_String xFullName = Tools_FileSystem::GetLoadFilename( xFileSelector );

	if(xFullName.IsNull())
	{
		//if the shape currently has a fragement type , ask if it should be cleared
		if(!s_pxMesh->m_uFragmentHash) return;

		YesNo_Overlay xOverlay("Remove current fragment type?");
		if(!xOverlay.DoModal()) return;

		s_pxMesh->m_uFragmentHash = ASURA_HASH_ID_UNSET;
	}
	else
	{
		//make sure the file is a fragment file...
		Tools_AsuraFileContents xContents;
		if(!xContents.LoadAsuraFile(xFullName.GetString()))
		{
			Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to load %s",Tools_FileSystem::StripPath(xFullName.GetString()));
			return;
		}

		if(!xContents.IsFragmentFile())
		{
			Tools_Timed_Message_System::AddMessage(0xff0000,"%s isn't a fragment file",Tools_FileSystem::StripPath(xFullName.GetString()));
			return;
		}

		//generate the hash ID for the fragment type , by getting filenames with path and extension
		Asura_String xName;
		xFullName.SplitFilename(0,&xName,0);

		s_pxMesh->m_uFragmentHash = Asura_GetHashID(xName.GetString());
	}
 	
 	Tools_SaveAsura::SetMainFileModified(true);

	UpdateFragmentName();
	UpdateOverlay();
}

void Tools_Mode_ShapeData::UpdateFragmentName()
{
	s_xFragmentName = "*None*";
	if(!s_pxMesh || !s_pxMesh->m_uFragmentHash)
	{
		return;
	}

	Asura_String xFilename = Asura_HashedFilename_System::GetFilename(s_pxMesh->m_uFragmentHash,false);	
	if(xFilename.IsNull())
	{
		s_xFragmentName = "*Unknown*";
		return;
	}

	xFilename.SplitFilename(0,&s_xFragmentName,0);
}

