//
// Tools_Mode_Smoothing.cpp
//
// Created by Kevin Lea - 16/07/2001 16:22:53
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Smoothing/Tools_Mode_Smoothing.h"

#include "Core/Asura_UserInput.h"
#include "Rendering/Asura_Module.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Surface.h"
#include "Rendering/Asura_Render.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Menubar_System.h"
#include "Interface/Asura_Menubar.h"
#include "Interface/Asura_Overlay.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Text_Output.h"
#include "Interface/Asura_Cursor.h"
#include "Core/Asura_Maths.h"
#include "Interface/Asura_Dialog.h"

#include "Tools_Overlays.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_Polygon_Selection.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Help.h"
#include "Misc/Tools_Render_Options.h"
#include "Misc/Tools_Adjacency.h"
#include "Rendering/Tools_BatchRenderPolygonOutline.h"
#include "Rendering/Tools_RenderModules.h"
#include "Texturing/Tools_Texturing_Find.h"

#include "Interface/Asura_Interface.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"

#include "Misc/Tools_Render.h"


///////////////////////////////////////////////////////////////////////////////////////
/*--------------**
** Help Strings **
**--------------*/
static const char* SmoothingModeHelpStrings [] =
{
	"Click on a number from 1 to 31 on the right hand bar to select the current smoothing group.",
	"The letter 'C' on that bar is used to represent the curved polygon flag.",
	"The background colour for the numbers will change from blue to red if the mouse is",
	"over a polygon that has that smoothing group set.",
	"",
	"F :           Apply the current smoothing group",
	"Shift+F :     Remove the current smoothing group",
	"R :           Remove all smoothing groups (but not the curved flag)",
	"",
	"C :           Make curved",
	"Shift+C :     Make not curved",
	"",
	"N :           Previous smoothing group",
	"M :           Next smoothing group",
	"",
	"Ctrl+Z :      Undo",
	"Ctrl+Y :      Redo",
	"< :           Undo",
	"> :           Redo",
	NULL
};

///////////////////////////////////////////////////////////////////////////////////////

//static member variables
Asura_Vector_2					Tools_Mode_Smoothing::SmoothingOverlayPosition = Asura_Vector_2(3.0f,28.0f);
bool							Tools_Mode_Smoothing::SmoothingOverlayOpen = true;
bool							Tools_Mode_Smoothing::HighlightAll = false;
bool							Tools_Mode_Smoothing::HighlightCurrentGroup = true;
bool							Tools_Mode_Smoothing::s_bDiscoLights = false;

int 							Tools_Mode_Smoothing::CurrentSmoothingGroup;
Asura_Handle					Tools_Mode_Smoothing::MenuHandle;
SmoothingModeUndo			 	Tools_Mode_Smoothing::SmoothingUndo;
Asura_Vector_2 					Tools_Mode_Smoothing::SmoothGroupOverlayMin;
Asura_Vector_2 					Tools_Mode_Smoothing::SmoothGroupOverlayMax;

ModulePolygonAdjacencies** 		Tools_Mode_Smoothing::Adjacencies;		
float 							Tools_Mode_Smoothing::AutosmoothAngle = 20.0f;//in degrees


void Tools_Mode_Smoothing::Initialise(void)
{
	//Initalise polygon selection
	Tools_Polygon_Selection::Activate();
	SmoothingUndo.Initialise();

	//reinitiaise stuff
	CurrentSmoothingGroup = 1;
	Adjacencies = NULL;

	//create the menubar
	MenuHandle = Asura_Menubar_System::CreateMenuSection();
	Asura_Menubar_Title* title;

	//load values from .dat file
	TransferModeData(TOOLSMODEDATAFILE_LOAD);


	title = Asura_Menubar_System::AddTitle(MenuHandle,"Overlays");
	title->AddItem("Smoothing",Tools_Mode_Smoothing::ToggleSmoothingOverlay,Tools_Mode_Smoothing::SmoothingOverlayIsOn);
	title->AddItem("Selection",Tools_Polygon_Selection::ToggleSelectionOverlay,Tools_Polygon_Selection::SelectionOverlayIsOn);
	title->AddItem("Find",Tools_Texturing_Find::ToggleFindOverlay,Tools_Texturing_Find::FindOverlayIsOn);

	//create overlays
	CreateSmoothingOverlay();

	Adjacencies = new ModulePolygonAdjacencies*[Asura_Module_System::GetNumberOfModules()];
	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Adjacencies[i] = NULL;
	}
	//add help
	Tools_Help::AddHelpTopic("Smoothing Mode",SmoothingModeHelpStrings);
}


void Tools_Mode_Smoothing::Shutdown(void)
{
	Tools_Help::RemoveHelpTopic("Smoothing Mode");
	SmoothingUndo.Shutdown();
	Tools_Polygon_Selection::Deactivate();

	//save values to .dat file
	TransferModeData(TOOLSMODEDATAFILE_SAVE);

	//shutdown the menu
	if(MenuHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Menubar_System::DestroyMenuSection(MenuHandle);
		MenuHandle = ASURA_INVALID_HANDLE;
	}

	//close overlays
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SMOOTHING);
	
	if(Adjacencies)
	{
		for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
		{
			delete Adjacencies[i]; 
		}
		delete [] Adjacencies;
	}

	Tools_Render_Options::SetUseDiscoLights(false);
}

void Tools_Mode_Smoothing::TransferModeData(int mode)
{
	//open the file
	Tools_ModeDataFile file("SmoothingMode.dat",mode);

	//transfer the data
	//(none of these should be removed , and new data should go on the end, otherwise
	//the files will get screwed up)
	file.Transfer(SmoothingOverlayPosition);
	file.Transfer(SmoothingOverlayOpen);
	file.Transfer(HighlightAll);
	file.Transfer(HighlightCurrentGroup);
}


void Tools_Mode_Smoothing::Update(void)
{
	//Flag the undo to start a new undo batch
	SmoothingUndo.MarkUndoStart();

	//smoothing group overlay
	HandleSmoothGroupOverlay();


	if(MouseIsOverSmoothGroupOverlay())
	{
		Tools_Polygon_Selection::Update(PolygonSelectionUpdate_NoMouseSelection);
	}
	else
	{
		Tools_Polygon_Selection::Update();
	}

	//act upon keyboard input
	Tools_Mode_Smoothing::HandleKeyboard();

	Tools_Render_Options::SetUseDiscoLights(s_bDiscoLights);
}

void Tools_Mode_Smoothing::RenderWorld_PostToneMap(void)
{
	//show outlines indicating smoothing groups
	HighlightSmoothing();

	Tools_Polygon_Selection::Render();
}

void Tools_Mode_Smoothing::RenderUI(void)
{
	Tools_Polygon_Selection::RenderUI();

	RenderSmoothGroupOverlay();
}

bool Tools_Mode_Smoothing::IsModified()
{
	return Tools_SaveAsura::IsMainFileModified();
}


void Tools_Mode_Smoothing::HandleKeyboard()
{
	if(Asura_Overlay_System::OverlaysHaveKeyboardFocus()) return;
 	if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT)) return;
 	if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
 	{
		if(Asura_UserInput::DebouncedKeyPressed('z'))// 'ctrl+z' for  undo
		{
			Undo();
		}
		if(Asura_UserInput::DebouncedKeyPressed('y'))// 'ctrl+y' for  redo
		{
			Redo();
		}
	}
	else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
	{
		if(Asura_UserInput::DebouncedKeyPressed('f'))// 'Shift + F' to remove smoothing group
		{
			RemoveSmoothingGroup();
		}
		if(Asura_UserInput::DebouncedKeyPressed('c'))// 'Shift+C' to remove curviness
		{
			RemoveCurveFlag();
		}

	}
	else
	{
		if(Asura_UserInput::DebouncedKeyPressed('f'))// 'F' to apply smoothing group 
		{
			ApplySmoothingGroup();
		}
		if(Asura_UserInput::DebouncedKeyPressed('c'))// 'C' to make curved
		{
			ApplyCurveFlag();
		}
		if (Asura_UserInput::DebouncedKeyPressed('r'))//'R' to remove smoothing groups
		{
			ClearAllSmoothingGroups();
		}

		if(Asura_UserInput::DebouncedKeyPressed('n'))
		{
			CurrentSmoothingGroup--;
			if(CurrentSmoothingGroup<1)
			{
				CurrentSmoothingGroup = 31;
			}
		}
		if(Asura_UserInput::DebouncedKeyPressed('m'))
		{
			CurrentSmoothingGroup++;
			if(CurrentSmoothingGroup>31)
			{
				CurrentSmoothingGroup = 1;
			}
		}
		if(Asura_UserInput::DebouncedKeyPressed(','))// '<' for  undo
		{
			Undo();
		}
		if(Asura_UserInput::DebouncedKeyPressed('.'))// '>' for  redo
		{
			Redo();
		}

	}
}


void Tools_Mode_Smoothing::ToggleSmoothingOverlay(void)
{
	SmoothingOverlayOpen = !SmoothingOverlayOpen;

	if (SmoothingOverlayOpen)
	{
		CreateSmoothingOverlay();
	}
	else
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SMOOTHING);
	}
}
bool Tools_Mode_Smoothing::SmoothingOverlayIsOn(void)
{
	return SmoothingOverlayOpen;
}

void Tools_Mode_Smoothing::CreateSmoothingOverlay(void)
{
	if(!SmoothingOverlayOpen) return;

	Asura_Overlay* newOverlay;
	newOverlay = new Asura_Overlay("Smoothing Mode",false,&SmoothingOverlayPosition);
	
	Asura_Dialog* newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(1.0f,3.0f);
			
	//use a number slider thingy until I can think of a better way of choosing the smoothing group
	//newDialog  = new Asura_Numerical_Dialog("Smoothing Group",offset,&(CurrentSmoothingGroup),1.0f,32.0f,1.0f,"%.0f");
	//newOverlay->AddDialog(newDialog);

	//offset.y+=2.0f;
	newDialog = new Asura_Tickbox_Dialog("Highlight All",offset,&(Tools_Mode_Smoothing::HighlightAll));
	newDialog->SetHelpText("Colour code all polygons according to their smoothing group.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Highlight Current",offset,&(Tools_Mode_Smoothing::HighlightCurrentGroup));
	newOverlay->AddDialog(newDialog);
	newDialog->SetHelpText("Highlights all polygons using the current smooting group with a green outline.");
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Disco Lights",offset,&(Tools_Mode_Smoothing::s_bDiscoLights));
	newOverlay->AddDialog(newDialog);
	newDialog->SetHelpText("Draws everything with a set of coloured lights for showing up vertex normal dicontinuities.");
	offset.y+=Asura_Dialog::LargeSpacing();

	
	newDialog = new Asura_Function_Dialog("Select Group",offset,Tools_Mode_Smoothing::SelectGroup);
	newDialog->SetHelpText("Select all polygons using the current smoothing group.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Deselect Group",offset,Tools_Mode_Smoothing::DeselectGroup);
	newDialog->SetHelpText("Deselect all polygons using the current smoothing group.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	
	newDialog = new Asura_Function_Dialog("Apply Smoothing Group (F)",offset,Tools_Mode_Smoothing::ApplySmoothingGroup);
	newDialog->SetHelpText("Adds the selected smoothing group to all selected polygons.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Remove Smoothing Group (Shift+F)",offset,Tools_Mode_Smoothing::RemoveSmoothingGroup);
	newDialog->SetHelpText("Removes the selected smoothing group from all selected polygons.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Add Curve Flag (C)",offset,Tools_Mode_Smoothing::ApplyCurveFlag);
	newDialog->SetHelpText("Adds the curved flag to all selected polygons (allowing them to be subdivided as curved if needed).");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Remove Curve Flag (Shift+C)",offset,Tools_Mode_Smoothing::RemoveCurveFlag);
	newDialog->SetHelpText("Removes the curved flag from all selected polygons.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Remove All Smoothing Groups (R)",offset,Tools_Mode_Smoothing::ClearAllSmoothingGroups);
	newDialog->SetHelpText("Removes all smoothing groups from all selected polygons.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog  = new Asura_Numerical_Dialog("Autosmooth Angle",offset,&(AutosmoothAngle),0.0f,180.0f,1.0f,"%.0f");
	newDialog->SetHelpText("Angle in degrees for autosmooth. Any adjacent polygons within this angle of eachother will be given a shared smoothing group.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Autosmooth Selection",offset,Tools_Mode_Smoothing::Autosmooth);
	newDialog->SetHelpText("Automatically assign smoothing groups to all selected polygons based on the angle between them.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Function_Dialog("Undo (Ctrl+Z)",offset,Tools_Mode_Smoothing::Undo);
	newDialog->SetHelpText("Undo changes to smoothing/curve flag.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();
	
	newDialog = new Asura_Function_Dialog("Redo (Ctrl+Y)",offset,Tools_Mode_Smoothing::Redo);
	newDialog->SetHelpText("Redo changes to smoothing/curve flag.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();
	
	
	newOverlay->m_uID = TOOLS_OVERLAY_SMOOTHING;

	Asura_Overlay_System::SubmitNewOverlay(newOverlay,false);
}

void Tools_Mode_Smoothing::ApplySmoothingGroup()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	if(CurrentSmoothingGroup == 32)
	{
		ApplyCurveFlag();
		return;
	}
	
	//convert the smoothing group into its bitfield value
	Asura_Assert(CurrentSmoothingGroup>=1 && CurrentSmoothingGroup<=32,"Smoothing group out of range");
	int SmoothBit = 1<<((int)CurrentSmoothingGroup-1);

	SetSmoothingGroup(SmoothBit,true);

	Tools_Timed_Message_System::AddMessage(0xffffff,"Added smoothing group %d",(int)CurrentSmoothingGroup);
}

void Tools_Mode_Smoothing::RemoveSmoothingGroup()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	if(CurrentSmoothingGroup == 32)
	{
		RemoveCurveFlag();
		return;
	}
	
	//convert the smoothing group into its bitfield value
	Asura_Assert(CurrentSmoothingGroup>=1 && CurrentSmoothingGroup<=32,"Smoothing group out of range");
	int SmoothBit = 1<<((int)CurrentSmoothingGroup-1);

//	bool add = (SmoothBit == NotCurvedSmoothingFlag) ? true : false;
	SetSmoothingGroup(SmoothBit,false);

	Tools_Timed_Message_System::AddMessage(0xffffff,"Removed smoothing group %d",(int)CurrentSmoothingGroup);
}

void Tools_Mode_Smoothing::ApplyCurveFlag()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}

	SetSmoothingGroup(NotCurvedSmoothingFlag,false);

	Tools_Timed_Message_System::AddMessage(0xffffff,"Added curved flag",(int)CurrentSmoothingGroup);
}

void Tools_Mode_Smoothing::RemoveCurveFlag()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}

	SetSmoothingGroup(NotCurvedSmoothingFlag,true);

	Tools_Timed_Message_System::AddMessage(0xffffff,"Removed curved flag",(int)CurrentSmoothingGroup);
}

void Tools_Mode_Smoothing::SetSmoothingGroup(int smooth_bit,bool add)
{
	//log for undo
	PrepareUndo();

	//remove from each polygon in the selection
	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			
		if(add)
			mesh->SmoothingGroupDataPtr[polygon.PolyIndex] |= smooth_bit;
		else
			mesh->SmoothingGroupDataPtr[polygon.PolyIndex] &=~ smooth_bit;

		//update the mesh in the hardware
		ResetMeshesAroundPolygon(polygon);
	}
}


void Tools_Mode_Smoothing::ClearAllSmoothingGroups()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	
	//log for undo
	PrepareUndo();

	//remove from each polygon in the selection
	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			
		//clear all but the curvy bit
		mesh->SmoothingGroupDataPtr[polygon.PolyIndex] &= 0x80000000;

		//update the mesh in the hardware
		ResetMeshesAroundPolygon(polygon);
	}

	Tools_Timed_Message_System::AddMessage(0xffffff,"Smoothing groups removed from selection");
	
}

int SmoothingGroupColours[31] =
{
	0x0000ff,0xffff00,0xff00ff,0x00ffff,
	0xff7f00,0xff007f,0x7fff00,0x00ff7f,
	0x7f00ff,0x007fff,0xff7f7f,0x7fff7f,
	0x7f7fff,0xff9f5f,0xff5f9f,0x9fff5f,
	0x5fff9f,0x9f5fff,0x5f9fff,0x7f0000,
	0x007f00,0x00007f,0x7f7f00,0x7f007f,
	0x007f7f,0x3f0000,0x003f00,0x00003f,
	0x7f7f7f
};

void Tools_Mode_Smoothing::HighlightSmoothing()
{
	if(!HighlightAll && !HighlightCurrentGroup) return;

	int SmoothBit = 1<<(CurrentSmoothingGroup-1);

	
	//Check the smoothing group of each polygon of each mesh
	//and choose an outline for it as needed

	Tools_BatchRenderPolygonOutline::BeginBatch();
	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		// are there any other reasons to reject the module as a whole?

		// module hidden
		if (!Asura_Module_System::IsModuleVisible( i )) continue;

		// module too far from camera
		if(!Tools_Render_Options::IsModuleWithinViewDistance( i )) continue;

		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(i);
		Asura_Assert(mesh,"Failed to get the mesh for the module");

		for(int j=0;j<mesh->NumberOfPolygons;j++)
		{
			int colour = 0;
			//get the smoothing groups for this polygon
			int smooth = mesh->SmoothingGroupDataPtr[j];
			smooth ^=NotCurvedSmoothingFlag;//so that the bit is set if it is curved

			if(HighlightCurrentGroup)
			{
				//does the polygon have this group?
				if(smooth & SmoothBit)
				{
 					colour = 0xff00ff00;
				}
			}
			if(!colour)
			{
				if(HighlightAll && (smooth &~NotCurvedSmoothingFlag))
				{
					//showing all smoothing groups.. find the colour for this group
					int bit = 1;
					for(int k=0;k<31;k++,bit<<=1)
					{
						if(smooth & bit)
						{
							colour = SmoothingGroupColours[k]|0xff000000;
							break;
						}
					}
					
				}
			}
			if(colour)
			{
				Tools_BatchRenderPolygonOutline::AddOutlineToBatch( j, i, colour );
			}
		}
	}
	Tools_BatchRenderPolygonOutline::EndBatch();
}


void Tools_Mode_Smoothing::Undo()
{
	if(SmoothingUndo.Undo())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Undo");
	}
}
void Tools_Mode_Smoothing::Redo()
{
	if(SmoothingUndo.Redo())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Redo");
	}
}


void Tools_Mode_Smoothing::PrepareUndo()
{
	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		SmoothingUndo.LogSmoothingChanged(polygon);
		Tools_SaveAsura::SetModuleModified(polygon.ModuleIndex);
	}
}

void Tools_Mode_Smoothing::DetermineSmoothGroupOverlayPosition()
{
	//position the overlay on the right hand side , middle of the screen
	float width = 10.0f*3.0f;
	float height = 14.0f*33.0f;

	SmoothGroupOverlayMax.x = (float )Asura_Viewport::GetWidth();
	SmoothGroupOverlayMin.x = SmoothGroupOverlayMax.x - width;
	
	SmoothGroupOverlayMin.y = (Asura_Viewport::GetHeight() - height)/2.0f;
	SmoothGroupOverlayMax.y = (Asura_Viewport::GetHeight() + height)/2.0f;

}

bool Tools_Mode_Smoothing::MouseIsOverSmoothGroupOverlay()
{
	//get the current mouse position
	int mx,my;
	Asura_Cursor::GetScreenSpaceCoords(mx,my);

	if(mx<(int)SmoothGroupOverlayMin.x) return false;
	if(mx>(int)SmoothGroupOverlayMax.x) return false;
	if(my<(int)SmoothGroupOverlayMin.y) return false;
	if(my>(int)SmoothGroupOverlayMax.y) return false;

	return true;
}

void Tools_Mode_Smoothing::HandleSmoothGroupOverlay()
{
	DetermineSmoothGroupOverlayPosition();
	if (Asura_Interface::UsingMenusOrOverlays()) return;

	//if the mouse has been clicked on the 'overlay' change the current smooth group
 	if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
	{
		if(!MouseIsOverSmoothGroupOverlay()) return;

		//get the current mouse position
		int mx,my;
		Asura_Cursor::GetScreenSpaceCoords(mx,my);

		//determine the group clicked on from the position
		int group = 1+ (int)((my - SmoothGroupOverlayMin.y)*33.0f/(SmoothGroupOverlayMax.y-SmoothGroupOverlayMin.y));
		if(group !=32)//(there is a gap between smoothing group 31 and the curve flag)
		{
			CurrentSmoothingGroup = group;
			if(CurrentSmoothingGroup<1) CurrentSmoothingGroup = 1;
			if(CurrentSmoothingGroup>32) CurrentSmoothingGroup = 32;
		}
	}
}

void Tools_Mode_Smoothing::RenderSmoothGroupOverlay()
{
	DetermineSmoothGroupOverlayPosition();

	
	//get the polygon currently under the cursor , so we can check its smoothing groups
	int poly_smooth_groups = 0;
	PolygonID polygon;
	if(Tools_Polygon_Selection::GetHighlitPolygon(polygon))
	{
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		poly_smooth_groups = mesh->SmoothingGroupDataPtr[polygon.PolyIndex];
		//reverse the curvy bit , so that it is considered to be set if the polygon is curved
		poly_smooth_groups ^= NotCurvedSmoothingFlag;
	}
	
	//draw the background
	Asura_Render::Overlay.TextureIndex = -1;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

	float x = SmoothGroupOverlayMin.x;
	float y = SmoothGroupOverlayMin.y;

	int smooth_bit = 1;
	for(int i=1;i<=32;i++)
	{
		if(i==32)
		{
			Asura_Render::Overlay.SetRectXY(x,y,30.0f,14.0f);
			Asura_Render::Overlay.Colour = 0x7f0000ff;
			Asura_Render::Overlay.Render();
			y+=14.0f;
		}

		Asura_Render::Overlay.SetRectXY(x,y,30.0f,14.0f);
		Asura_Render::Overlay.Colour = (smooth_bit & poly_smooth_groups) ? 0x7fff0000: 0x7f0000ff;
		Asura_Render::Overlay.Render();
		y+=14.0f;
		smooth_bit<<=1;
	}
	Asura_Render::FlushDisplay();

	//now draw the numbers
	y = SmoothGroupOverlayMin.y;
	x = (SmoothGroupOverlayMin.x+SmoothGroupOverlayMax.x)/2.0f;
	Asura_Text_Output::ConvertPixelsToPrintCoords(x,y);

	for(int i=1;i<=32;i++)
	{
		int colour = (i==CurrentSmoothingGroup) ? 0xff00ff00 : 0xffffffff;
	
		if(i==32)
		{
			y+=1.0f;//gap for the curve flag
			Asura_Text_Output::Print(x,y,colour,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"C",i);
		}
		else
		{
			Asura_Text_Output::Print(x,y,colour,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"%d",i);
		}

		y+=1.0f;
	}

	#if 0
	//////test... show adjacencies
	if(polygon.ModuleIndex>=0)
	{
		for(LIF<PolygonAndAngle> adlif(&Adjacencies[polygon.ModuleIndex]->PolygonArray[polygon.PolyIndex].AdjacencyList);!adlif.done();adlif.next())
		{
			Tools_Render::RenderPolygonOutline(adlif().Polygon.PolyIndex,adlif().Polygon.ModuleIndex,0xff0000ff);
		}
	}
	#endif


}


void Tools_Mode_Smoothing::Autosmooth()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}

	//get the angle in radians
	Asura_Angle Angle = (AutosmoothAngle/180.0f)*Asura_Maths::Pi;

	//first need to initialise some working info for the selected polygons
	PolyAutosmoothInfo** AutosmoothInfo = new PolyAutosmoothInfo*[Asura_Module_System::GetNumberOfModules()];
	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		AutosmoothInfo[i] = NULL;	
	}

	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();
	int last_module_index = -1;
	//find out which modules we have selected polygons for
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		Asura_Assert(polygon.ModuleIndex>=last_module_index,"Expecting the selection iteration to go through modules in ascending order");

		if(last_module_index!=polygon.ModuleIndex)
		{
			Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			//allocate space for the polygons in this module
			AutosmoothInfo[polygon.ModuleIndex] = new PolyAutosmoothInfo[mesh->NumberOfPolygons];

			//make sure we know the adjacencies for this module
			if(!Adjacencies[polygon.ModuleIndex])
			{
				Adjacencies[polygon.ModuleIndex] = new ModulePolygonAdjacencies(polygon.ModuleIndex);
			}
	
			last_module_index=polygon.ModuleIndex;

		}
	}

	#define GetAutosmoothInfo(poly) (&AutosmoothInfo[poly.ModuleIndex][poly.PolyIndex])
	
	//iterate again working out smoothing groups
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//done this polygon yet?
		if(AutosmoothInfo[polygon.ModuleIndex][polygon.PolyIndex].done) continue;

		GetAutosmoothInfo(polygon)->done = true;
		//get a list of all polygons that are connected by a chain of smooth polygons to this
		//also search for a smoothing group allowed by all of them
		List<PolygonID> queue;
		queue.add_entry(polygon);
		int allowed_groups = 0x7fffffff;

		LIF<PolygonID> qlif(&queue);
		for(;!qlif.done();qlif.next())
		{
			//check all the adjacent polygons
			for(LIF<PolygonAndAngle> adlif(&Adjacencies[qlif().ModuleIndex]->AdjacencyLists[qlif().PolyIndex]);!adlif.done();adlif.next())
			{
				//only interested in adjacent polygons that have been selected
				if(!Tools_Polygon_Selection::IsSelected(adlif().Polygon)) continue;

				if(adlif().Angle<=Angle)
				{
					//smooth
					//add to the queue , if it hasn't already been done
					PolyAutosmoothInfo* inf = GetAutosmoothInfo(adlif().Polygon);
					if(!inf->done)
					{
						inf->done = true;
						queue.add_entry(adlif().Polygon);

						//update the allowed group
						allowed_groups &= inf->allowed_groups;
					}
				}

			}
		}

		//see if there is any smoothing group that is currently allowed by all the polygons
		//in the queue. If there is , it becomes our preferred smoothing group
		int preferred_group = 0;
		if(allowed_groups)
		{
			for(preferred_group = 1;!(allowed_groups & preferred_group);preferred_group<<=1){}
		}

		//go through the polygons in the queue working out smoothing group flags to set
		for(qlif.restart();!qlif.done();qlif.next())
		{
   			PolyAutosmoothInfo* inf1 = GetAutosmoothInfo(qlif());
			//check each polygon adjacent to this one
			for(LIF<PolygonAndAngle> adlif(&Adjacencies[qlif().ModuleIndex]->AdjacencyLists[qlif().PolyIndex]);!adlif.done();adlif.next())
			{
				if(adlif().Angle>Angle) continue;
				if(!Tools_Polygon_Selection::IsSelected(adlif().Polygon)) continue;

 				PolyAutosmoothInfo* inf2 = GetAutosmoothInfo(adlif().Polygon);

				//if the polygons already share a smoothing group , then we are done
				if(inf1->smooth_groups & inf2->smooth_groups) continue;

				int group = 0;//group to use
				int allowed = inf1->allowed_groups & inf2->allowed_groups;

				if(!allowed)
				{
					Asura_Assert(allowed,"Ran out of smoothing groups while autosmoothing");
					continue;
				}

				//use the preferred group if allowed
				if(allowed & preferred_group)
				{
					group = preferred_group;
				}
				else
				{
					//Can't use the preferred group

					//Ideally we want to use a smoothing group bit that is already set in one of the polygons
					if(inf1->smooth_groups & allowed)
					{
						allowed &=inf1->smooth_groups;
					}
					else if (inf2->smooth_groups & allowed)
					{
						allowed &=inf2->smooth_groups;
					}

					//find the first bit set in allowed
					for(group = 1;!(group & allowed);group<<=1){};
				}

				//add the chosen smoothing group to the polygons
				inf1->smooth_groups |= group;
				inf2->smooth_groups |= group;

				//disallow this smoothing_group from all adjacent non-smooth polygons
				for(LIF<PolygonAndAngle> adlif1(&Adjacencies[qlif().ModuleIndex]->AdjacencyLists[qlif().PolyIndex]);!adlif1.done();adlif1.next())
				{
					if(!Tools_Polygon_Selection::IsSelected(adlif1().Polygon)) continue;
					if(adlif1().Angle<=Angle) continue;
					GetAutosmoothInfo(adlif1().Polygon)->allowed_groups &=~ group;

				}
								
				for(LIF<PolygonAndAngle> adlif2(&Adjacencies[adlif().Polygon.ModuleIndex]->AdjacencyLists[adlif().Polygon.PolyIndex]);!adlif2.done();adlif2.next())
				{
					if(!Tools_Polygon_Selection::IsSelected(adlif2().Polygon)) continue;
					if(adlif2().Angle<=Angle) continue;
					GetAutosmoothInfo(adlif2().Polygon)->allowed_groups &=~ group;

				}
			}

		}

	}

	//now have all the smoothing groups calculated , prepare undo and copy the results over
	PrepareUndo();
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			
		//clear all but the curvy bit
		mesh->SmoothingGroupDataPtr[polygon.PolyIndex] &= 0x80000000;
		//set the smoothing groups
		mesh->SmoothingGroupDataPtr[polygon.PolyIndex] |= AutosmoothInfo[polygon.ModuleIndex][polygon.PolyIndex].smooth_groups;
	
		//update the mesh in the hardware
		ResetMeshesAroundPolygon(polygon);
	}

	//clearup the working space
	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		delete [] AutosmoothInfo[i];	
	}
	delete [] AutosmoothInfo;

	Tools_Timed_Message_System::AddMessage(0xffffff,"Finished autosmoothing selection");

}

void Tools_Mode_Smoothing::ResetMeshesAroundPolygon(const PolygonID& xPolygon)
{
	//reset the mesh for this polygon
	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);
	Asura_Assert(pxMesh,"");
	Asura_Mesh_System::PlatformResetRenderMesh(pxMesh);
	pxMesh->m_bSmoothedAcrossModules = false;

	Tools_RenderModule_System::UpdateModuleMesh(pxMesh->Index);

	//get the list of adjacent polyongs
	const Tools_Polygon_Adjacency* pxPolyAdj = Tools_Adjacency::GetPolygonAdjacencies(xPolygon.ModuleIndex,xPolygon.PolyIndex);

	for(const Tools_Adjacent_Polygon* pxAdj = pxPolyAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
	{
		//ignore adjacencies between zones. RAR 02/05/2006		
		if(pxAdj->IsBetweenZones()) continue;
		
		if(pxAdj->GetModuleIndex() == xPolygon.ModuleIndex) continue;

		pxMesh = Asura_Module_System::GetMeshForModule(pxAdj->GetModuleIndex());
		Asura_Assert(pxMesh,"");

		Asura_Mesh_System::PlatformResetRenderMesh(pxMesh);
		pxMesh->m_bSmoothedAcrossModules = false;

		Tools_RenderModule_System::UpdateModuleMesh(pxMesh->Index);
	}
}

void Tools_Mode_Smoothing::SelectGroup()
{
	SelectGroupMain(true);
}

void Tools_Mode_Smoothing::DeselectGroup()
{
	SelectGroupMain(false);
}

void Tools_Mode_Smoothing::SelectGroupMain(bool bSelect)
{
	int iSmoothBit = 1<<(CurrentSmoothingGroup-1);

	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		Asura_Assert(pxMesh,"Failed to get the mesh for the module");

		for(int j=0;j<pxMesh->NumberOfPolygons;j++)
		{
			int iSmooth = pxMesh->SmoothingGroupDataPtr[j];
			iSmooth ^=NotCurvedSmoothingFlag;//so that the bit is set if it is curved

			if(iSmooth & iSmoothBit)
			{
				if(bSelect)
					Tools_Polygon_Selection::SelectPolygon(i,j);
				else
					Tools_Polygon_Selection::DeselectPolygon(i,j);
			}
		}
	}
}

void Tools_Mode_Smoothing::HandleMeshChange(int iModuleIndex)
{
	//Invalidate any undos
	SmoothingUndo.FlushUndoBuffer();

	//Invalidate all adjacenies , since adjacent modules will be invalid as well , and we don't
	//have a convenient store of that information.
	if(Adjacencies)
	{
		for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
		{
			delete Adjacencies[i]; 
			Adjacencies[i] = 0;
		}
	}
}



/////////////////////////////////////////////////////////////////////////////////////////
/*-------------------**
** SmoothingModeUndo **
**-------------------*/


void SmoothingModeUndo::ProcessUndoData(void* data)
{
	SmoothingUndoData* undo_data = (SmoothingUndoData*) data;
	
	switch(undo_data->undo_type)
	{
		case SmoothingUndoData::UndoType_SmoothingChanged :
		{
			//get the mesh
			Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(undo_data->polygon.ModuleIndex);

			//store current smoothing group , and update from undo data
			int group = mesh->SmoothingGroupDataPtr[undo_data->polygon.PolyIndex];
			mesh->SmoothingGroupDataPtr[undo_data->polygon.PolyIndex] = undo_data->smoothing_groups;

			//convert the undo data
			undo_data->smoothing_groups = group;

			//make sure the mesh gets updated in the hardware
			Tools_Mode_Smoothing::ResetMeshesAroundPolygon(undo_data->polygon);

			Tools_SaveAsura::SetMeshModified(mesh);
			break;
		}

		default :
			Asura_Assert(0,"Unexpected undo type for smoothing");
			break;
	}
}

void SmoothingModeUndo::LogSmoothingChanged(const PolygonID& polygon)
{
	//allocate undo data
	SmoothingUndoData* undo = (SmoothingUndoData*) AllocateNextUndoData();
	if(!undo) return;

	//get the mesh data
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
	Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");

	
	//fill in the undo data	
	undo->undo_type = SmoothingUndoData::UndoType_SmoothingChanged;
	undo->polygon = polygon;
	undo->smoothing_groups  = mesh->SmoothingGroupDataPtr[polygon.PolyIndex];
}








/////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** ModulePolygonAdjacencies **
**--------------------------*/



ModulePolygonAdjacencies::ModulePolygonAdjacencies(int module)
{
	ModuleIndex = module;
	NumPolygons = 0;
	AdjacencyLists = NULL;

	//get the mesh for this module
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(ModuleIndex);
	if(!mesh->AreNormalsValid())mesh->CreateNormals();

	//allocate the array of polygon adjacencies
	NumPolygons = mesh->NumberOfPolygons;
	AdjacencyLists = new List<PolygonAndAngle>[NumPolygons];

	//Start by searching for all adjacencies that are within the module

	for(int p1=0;p1<NumPolygons;p1++)
	{
		//get the polygon data
		Asura_Mesh_Polygon* poly1 = &mesh->PolygonDataPtr[p1];
		int num_verts1 = poly1->NumberOfVertices();


		for(int p2=p1+1;p2<NumPolygons;p2++)
		{
			Asura_Mesh_Polygon* poly2 = &mesh->PolygonDataPtr[p2];
			int num_verts2 = poly2->NumberOfVertices();

			//count the shared vertices
			int shared = 0;
			for(int v1=0;v1<num_verts1;v1++)
			{
				for(int v2=0;v2<num_verts2;v2++)
				{
					if(poly1->VertexID[v1] == poly2->VertexID[v2])
					{
						shared++;
						break;
					}
				}
			}

			if(shared>1)
			{
				//these 2 polygons are adjacent
				//determine the angle between them
				float dot = mesh->NormalsDataPtr[p1].FaceNormal*mesh->NormalsDataPtr[p2].FaceNormal;
				if(dot>1.0f) dot = 1.0f;
				else if(dot<-1.0f) dot = -1.0f;
				Asura_Angle angle = Asura_Maths::ArcCosine(dot);

				AdjacencyLists[p1].add_entry(PolygonAndAngle(ModuleIndex,p2,angle));
				AdjacencyLists[p2].add_entry(PolygonAndAngle(ModuleIndex,p1,angle));
			}
		}
	}

	float tolerance = 0.004f;
	
	//Now need to check for adjacencies between modules

	//Get bounding box for this module , and add a little tolerance
	Asura_Bounding_Box box1 = mesh->BoundingBox;
	box1.MinX-=tolerance;
	box1.MinY-=tolerance;
	box1.MinZ-=tolerance;
	box1.MaxX+=tolerance;
	box1.MaxY+=tolerance;
	box1.MaxZ+=tolerance;

	Asura_Vector_3 loc1;
	Asura_Module_System::GetPositionOfModule(ModuleIndex,loc1);

	for(int m2=0;m2<Asura_Module_System::GetNumberOfModules();m2++)
	{
	
		
		if(m2 == ModuleIndex) continue;
		Asura_Mesh* mesh2 = Asura_Module_System::GetMeshForModule(m2);
		if(!mesh2->AreNormalsValid())mesh2->CreateNormals();

		//determine the offset to add to convert points so that they are local module 1
		Asura_Vector_3 loc2;
		Asura_Module_System::GetPositionOfModule(m2,loc2);
		Asura_Vector_3 offset = loc2-loc1;

		//get the bounding box relative to the first module (plus a little tolerance)
		Asura_Bounding_Box box2 = mesh2->BoundingBox;
		box2.MinX+=offset.x-tolerance;
		box2.MinY+=offset.y-tolerance;
		box2.MinZ+=offset.z-tolerance;
		box2.MaxX+=offset.x+tolerance;
		box2.MaxY+=offset.y+tolerance;
		box2.MaxZ+=offset.z+tolerance;

		//do the 2 modules touch?
		if(!box1.Intersects(box2)) continue;

		//get the intersection of the boxes
		Asura_Bounding_Box intersect;
		intersect.MinX = max(box1.MinX,box2.MinX);
		intersect.MinY = max(box1.MinY,box2.MinY);
		intersect.MinZ = max(box1.MinZ,box2.MinZ);
		intersect.MaxX = min(box1.MaxX,box2.MaxX);
		intersect.MaxY = min(box1.MaxY,box2.MaxY);
		intersect.MaxZ = min(box1.MaxZ,box2.MaxZ);

		//get a list of polygons in the second mesh that have at least 2 verts
		//in the intersection
		List<int> p2_list;
		for(int p2=0;p2<mesh2->NumberOfPolygons;p2++)
		{
			Asura_Mesh_Polygon* poly2 = &mesh2->PolygonDataPtr[p2];
			int num_verts2 = poly2->NumberOfVertices();
			int verts_in_intersect = 0;
			for(int v2=0;v2<num_verts2;v2++)
			{
				Asura_Vector_3 temp = mesh2->VertexDataPtr[poly2->VertexID[v2]].Position+offset;
				if(intersect.Contains(temp))
				{
					verts_in_intersect++;
				}
				
			}
			if(verts_in_intersect>1)
			{
				p2_list.add_entry(p2);
			}
			
		}

		if(!p2_list.size()) continue;

		//check each polygon in the first mesh against this list
		for(int p1=0;p1<mesh->NumberOfPolygons;p1++)
		{
			Asura_Mesh_Polygon* poly1 = &mesh->PolygonDataPtr[p1];
			int num_verts1 = poly1->NumberOfVertices();

			//check that this polygon has 2 verts in the intersection
			int verts_in_intersect = 0;
			for(int v1=0;v1<num_verts1;v1++)
			{
				if(intersect.Contains(mesh->VertexDataPtr[poly1->VertexID[v1]].Position))
				{
					verts_in_intersect++;
				}
			}
			if(verts_in_intersect<2) continue;

			//check against the list of polygons in the second mesh
			for(LIF<int> p2lif(&p2_list);!p2lif.done();p2lif.next())
			{
				Asura_Mesh_Polygon* poly2 = &mesh2->PolygonDataPtr[p2lif()];
				int num_verts2 = poly2->NumberOfVertices();

				//look for 2 shared vertices
				//count the shared vertices
				int shared = 0;
				for(int v1=0;v1<num_verts1;v1++)
				{
					for(int v2=0;v2<num_verts2;v2++)
					{
						Asura_Vector_3 diff = mesh->VertexDataPtr[poly1->VertexID[v1]].Position;
						diff -= (mesh2->VertexDataPtr[poly2->VertexID[v2]].Position+offset);

						if(diff.x>=-tolerance && diff.x<=tolerance && 
						   diff.y>=-tolerance && diff.y<=tolerance && 
						   diff.z>=-tolerance && diff.z<=tolerance)
						{
							shared++;
							break;
						}
					}
				}

				if(shared>1)
				{
					//these 2 polygons are adjacent
					//determine the angle between them
					float dot = mesh->NormalsDataPtr[p1].FaceNormal*mesh2->NormalsDataPtr[p2lif()].FaceNormal;
					if(dot>1.0f) dot = 1.0f;
					else if(dot<-1.0f) dot = -1.0f;
					Asura_Angle angle = Asura_Maths::ArcCosine(dot);

					AdjacencyLists[p1].add_entry(PolygonAndAngle(m2,p2lif(),angle));
				}

			}
		}
	}

}


ModulePolygonAdjacencies::~ModulePolygonAdjacencies()
{
	delete [] AdjacencyLists;
}









