//
// Tools_ModuleSmoothing.cpp
//
// Created by Richard Rice - 05/11/2002
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Debugging/Asura_Assert.h"

#include "Smoothing/Tools_ModuleSmoothing.h"

#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Surface.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Bicubic_Patch.h"

#ifdef ASURA_TOOLS
#include "Load&Save/Tools_SaveAsura.h"
#endif
#include "MeshProcessing/Tools_TangentSpace.h"
#include "Misc/Tools_Adjacency.h"

#if defined(ASURA_TARGET_SM3)
#include "Lighting/Tools_EnvironmentalLightTransport.h"
#endif



void Tools_ModuleSmoothing::CheckModuleNormals()
{
	
	#ifdef ASURA_TOOLS
	//only want to do smoothing across modules if an environment file is being edited
	if(Tools_SaveAsura::GetFileType()!=AsuraFile_Environment) return;
	#endif

	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);

		if(!pxMesh->AreNormalsValid() || !pxMesh->m_bSmoothedAcrossModules)
		{
		 	CalculateNormals(i);
		}
	}
}

void Tools_ModuleSmoothing::CalculateNormals(int iModuleIndex)
{
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<=Asura_Module_System::GetNumberOfModules(),"");
	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModuleIndex);

	//make sure face normals are up to date
	if(!pxMesh->m_bFaceNormalsValid)
	{
		pxMesh->CreateFaceNormals();
	}

	Asura_Vector_3 xModulePos = Asura_Module_System::GetPositionOfModule(iModuleIndex);


	for(int i=0;i<pxMesh->NumberOfPolygons;i++)
	{
		Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[i];
		int iNumVerts = pxPoly->NumberOfVertices();
		const Asura_Vector_3 xPolyNormal = pxMesh->NormalsDataPtr[i].FaceNormal;

		int iSmoothGroup = pxMesh->GetSmoothingGroup(i); 
		for(int j=0;j<iNumVerts;j++)
		{
			Asura_Vector_3 xNormal;
			xNormal.Zero();
						
			if(iSmoothGroup)
			{
				Asura_Vector_3 xNormalsFound[16];
				int iNumNormalsFound = 0;

				//get a list of polygons using this vertex
				int iVertexID = pxPoly->VertexID[j];
				const Tools_Polygon_Adjacency* pxVertAdj = Tools_Adjacency::GetVertexAdjacencies(iModuleIndex,iVertexID);

				Asura_Vector_3 xVertPos = pxMesh->VertexDataPtr[iVertexID].Position + xModulePos;

				for(const Tools_Adjacent_Polygon* pxAdj = pxVertAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
				{
					int iModuleIndex2 = pxAdj->GetModuleIndex();
					int iPolygonIndex2 = pxAdj->GetPolygonIndex();

					#ifdef ASURA_USE_STREAMING
					Asura_Module_Handler* pxModHandler2 = Asura_Module_System::GetHandler(pxAdj->GetZoneHash());
					#else
					Asura_Module_Handler* pxModHandler2 = &Asura_Module_System::GetCurrentHandler();
					#endif

					Asura_Mesh* pxMesh2 = pxModHandler2->GetMeshForModule(iModuleIndex2);
					Asura_Assert(pxMesh2,"");
					if(!pxMesh2->m_bFaceNormalsValid) pxMesh2->CreateFaceNormals();

					int iSmoothGroup2 = pxMesh2->GetSmoothingGroup(iPolygonIndex2);

					//add on the face normal for this polygon , if any smoothing groups are shared
					if(iSmoothGroup & iSmoothGroup2)
					{
						Asura_Vector_3 xRelativeVertPos = xVertPos - pxModHandler2->GetPositionOfModule(iModuleIndex2);

						Asura_Vector_3 xNormalToAdd = pxMesh2->GetVertexNormalContribution(iPolygonIndex2,xRelativeVertPos);

						//Don't attempt to smooth with polygons that are facing too much away from each other ,
						//regardless of what the smoothing groups say.
						//RAR 24/10/2008
						if(xNormalToAdd * xPolyNormal < -0.5f)
						{
							continue;	
						}
						
						//make sure this normal hasn't already been added
						bool bAlreadyFound = false;
						for(int l=0;l<iNumNormalsFound;l++)
						{
							if(xNormalsFound[l] == xNormalToAdd)
							{
								bAlreadyFound = true;
								break;
							}

						}
						if(!bAlreadyFound)
						{
							xNormalsFound[iNumNormalsFound++] = xNormalToAdd;
							xNormal += xNormalToAdd;

							if(iNumNormalsFound>=16) iNumNormalsFound = 15;
						}
					}
			
				}
			}
			//normalise the vertex normal
			float fLength = xNormal.Magnitude();
			if(fLength)
			{
				xNormal /= fLength;
			}
			else
			{
				xNormal = pxMesh->GetVertexNormalContribution(i,j);
			}

			pxMesh->NormalsDataPtr[i].VertexNormal[j] = xNormal;
		}
	}
	pxMesh->m_bVertexNormalsValid = true;
	pxMesh->m_bSmoothedAcrossModules = true;
}





void Asura_Surface_Mesh_System::ProcessModuleMeshIntoSurfaces(int iModuleIndex)
{
	Asura_Bicubic_Patch xPatch;
	Asura_Surface xTempSurface;


	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModuleIndex);
	Asura_Assert(pxMesh,"");

	Asura_Vector_3 xPosition;
	Asura_Module_System::GetPositionOfModule(iModuleIndex,xPosition);
	
	if(!Tools_Adjacency::IsInitialised())
	{
		//Haven't got the tools adjacency information , just process into surfaces ignoring other modules
		ProcessMeshIntoSurfaces(*pxMesh,xPosition);
		return;
	}
	
	Asura_Surface_Mesh* pxSurfaceMesh = &s_axSurfaceMeshStorage[pxMesh->Index];
	
	pxSurfaceMesh->NumberOfSurfaces = pxMesh->NumberOfPolygons;
	pxSurfaceMesh->Surfaces = new Asura_Surface[pxMesh->NumberOfPolygons];
	pxSurfaceMesh->SurfaceDataValid = true;


	#if defined(ASURA_TARGET_SM3)
	Tools_Mesh_TangentSpace xMeshTangentSpace;
	xMeshTangentSpace.Initialise(iModuleIndex);
	#endif
	

	int i;

	// KJL - 21-May-01 15:24:08 - make surfaces
	for (i=0; i<pxMesh->NumberOfPolygons; i++)
	{
		//Initialise each surface from its polygon data - RAR
		pxSurfaceMesh->Surfaces[i].Initialise(pxMesh,i);

		#if defined(ASURA_TARGET_SM3)

		Asura_Surface& xSurface = pxSurfaceMesh->Surfaces[i];

		for (int u=0; u<xSurface.NumberOfVertices; ++u)
		{
			Tools_EnvironmentalLightTransport::GetTransportHarmonic
			(
				xSurface.m_axLightTransportHarmonic[u],
				iModuleIndex,
				xSurface.VertexID[u]
			);

			Tools_EnvironmentalLightTransport::GetDiffuseForVertex
			(
				xSurface.m_axDiffuseSH[u],
				xSurface.VertexID[u],
				iModuleIndex
			);

			// JAT 20/07/07 include the vertex painting colour 
			Tools_EnvironmentalLightTransport::UpdateLightingFromVertexPainting( xSurface.m_axDiffuseSH[u], pxMesh, i, u );

			// VGP 13Feb2006 Calculate brightest light direction & intensity for specular
			xSurface.m_axDiffuseSH[u].GetBrightestDirection(xSurface.m_axBrightestLightVector[u]);

			xMeshTangentSpace.GetTangentAndBinormal(i, xSurface.VertexID[u], 0, xSurface.LNormalID[u], xSurface.m_axTangent[u], xSurface.m_axBinormal[u]);
			xMeshTangentSpace.GetTangentAndBinormal(i, xSurface.VertexID[u], 1, xSurface.LNormalID[u], xSurface.m_axTangent2[u], xSurface.m_axBinormal2[u]);
		}

		if (xSurface.NumberOfVertices == 3)
		{
			xSurface.m_axLightTransportHarmonic[3] = xSurface.m_axLightTransportHarmonic[2];
			xSurface.m_axDiffuseSH[3] = xSurface.m_axDiffuseSH[2];

			// VGP 13Feb2006
			xSurface.m_axBrightestLightVector[3] = xSurface.m_axBrightestLightVector[2];

			xSurface.m_axTangent[3] = xSurface.m_axTangent[2];
			xSurface.m_axBinormal[3] = xSurface.m_axBinormal[2];
			xSurface.m_axTangent2[3] = xSurface.m_axTangent2[2];
			xSurface.m_axBinormal2[3] = xSurface.m_axBinormal2[2];
		}
		
		#endif	

	}


	//Sort out end caps , and blend curves between surfaces

	for (i=0; i<pxMesh->NumberOfPolygons; i++)
	{
	  	if (pxSurfaceMesh->Surfaces[i].NumberOfVertices==0) continue;

		// Kevin Lea - 22/08/2001 11:08:51 - if flagged as 'not curved' skip surface
		//if ((pxMesh->SmoothingGroupDataPtr[i]&(1<<31))) continue;

		Asura_Surface* pxSurface = &pxSurfaceMesh->Surfaces[i];

		//check each adjacent polygon
		const Tools_Polygon_Adjacency* pxPolyAdj = Tools_Adjacency::GetPolygonAdjacencies(iModuleIndex,i);

		for(const Tools_Adjacent_Polygon* pxAdj = pxPolyAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
		{
			//ignore adjacencies between zones. RAR 02/05/2006
			if(pxAdj->IsBetweenZones()) continue;
			
			if(pxAdj->GetNumSharedVertices()<2) continue;//need a shared edge

			int iModuleIndex2 = pxAdj->GetModuleIndex();
			int iPolygonIndex2 = pxAdj->GetPolygonIndex();

			Asura_Surface* pxSurface2;
			Asura_Mesh* pxMesh2;
			Asura_Vector_3 xOffset;

			//dealing with polygon in same module?
			if(iModuleIndex==iModuleIndex2)
			{
				//Compare the polygon indeces , to avoid considering each pair more than once
				if(iPolygonIndex2<i) continue;
			 	// If flagged as 'not curved' skip surface
				//if ((pxMesh->SmoothingGroupDataPtr[iPolygonIndex2]&(1<<31))) continue;

				pxSurface2 = &pxSurfaceMesh->Surfaces[iPolygonIndex2];
				pxMesh2 = pxMesh;
				xOffset.Zero();

			}
			else
			{
				pxMesh2 = Asura_Module_System::GetMeshForModule(iModuleIndex2);
				Asura_Assert(pxMesh2,"");

				if(!pxMesh2->AreNormalsValid()) pxMesh2->CreateNormals();

			 	// If flagged as 'not curved' skip surface
				//if ((pxMesh2->SmoothingGroupDataPtr[iPolygonIndex2]&(1<<31))) continue;


				//create a temporary surface for the polygon from the other mesh
				xTempSurface.Initialise(pxMesh2,iPolygonIndex2);
				pxSurface2 = &xTempSurface;

				//get the offset between the modules
				Asura_Module_System::GetPositionOfModule(iModuleIndex2,xOffset);
				xOffset -= xPosition;
			}

 			

			//find the shared vertices

			int iNumShared = 0;
			int aiIndex1[2];
			int aiIndex2[2];
			
			for (int a=0; a<pxSurface->NumberOfVertices; a++)
			{
				for (int b=0; b<pxSurface2->NumberOfVertices; b++)
				{
					int iVert1 = pxSurface->VertexID[a];
					int iVert2 = pxSurface2->VertexID[b];
					
					bool bMatch = false;
					if(pxMesh == pxMesh2)
					{
						//same mesh , can just match based on ID
						bMatch = iVert1==iVert2;
					}
					else
					{
						//different mesh , match based on distance
						Asura_Vector_3 xDiff = xOffset + pxMesh2->VertexDataPtr[iVert2].Position - pxMesh->VertexDataPtr[iVert1].Position;

						if(Tools_Adjacency::IsDistanceWithinTolerance(xDiff))	
						{
							bMatch = true;
						}
					}

					if(bMatch)
					{
						aiIndex1[iNumShared] = a;
						aiIndex2[iNumShared] = b;
						iNumShared++;
						break;
					}
					
				}
				if(iNumShared == 2) break;
			}
			if(iNumShared != 2) continue;
			
			
			
			// KJL 21/05/00 15:05:02 - found shared edge.

			//work out which sides these vertices belong to
			int is,io,io2,js,jo,jo2;

			if(aiIndex1[1] == (aiIndex1[0]+1)%pxSurface->NumberOfVertices)
			{
				io = 0;
				io2 = 1;
			}
			else if(aiIndex1[0] == (aiIndex1[1]+1)%pxSurface->NumberOfVertices)
			{
				io = 1;
				io2 = 0;
			}
			else
			{
				//something odd going on
				continue;
			}
			is = aiIndex1[io];
			
			if(aiIndex2[1] == (aiIndex2[0]+1)%pxSurface2->NumberOfVertices)
			{
				jo = 0;
				jo2 = 1;
			}
			else if(aiIndex2[0] == (aiIndex2[1]+1)%pxSurface2->NumberOfVertices)
			{
				jo = 1;
				jo2 = 0;
			}
			else
			{
				continue;
			}
			js = aiIndex2[jo];

			if(io == jo)
			{
				//directions of the sides on both polygons is the same , this means that the polygons
				//will be facing differnt directions , so probably shouldn't count as connected
				continue;
			}

			
			if(pxSurface->IsCurved && pxSurface2->IsCurved)
			{
				pxSurface->CNormalID[is][io] += pxSurface2->CNormalID[js][jo];
				pxSurface->CNormalID[is][io2] += pxSurface2->CNormalID[js][jo2];
				pxSurface->CNormalID[is][io].Normalise();
				pxSurface->CNormalID[is][io2].Normalise();

				pxSurface2->CNormalID[js][jo] = pxSurface->CNormalID[is][io];
				pxSurface2->CNormalID[js][jo2] = pxSurface->CNormalID[is][io2];

			}
			else if(pxSurface2->IsCurved)
			{
				//first surface is an endcap, take normals from second
				pxSurface->IsEndCap = true;
				pxSurface->CNormalID[is][io] = pxSurface2->CNormalID[js][jo];
				pxSurface->CNormalID[is][io2] = pxSurface2->CNormalID[js][jo2];
				
			}
			else if(pxSurface->IsCurved)
			{
				//second surface is an endcap, take normals from first
				pxSurface2->IsEndCap = true;
				pxSurface2->CNormalID[js][jo] = pxSurface->CNormalID[is][io];
				pxSurface2->CNormalID[js][jo2] = pxSurface->CNormalID[is][io2];
						
			}
			if(pxSurface->m_bHasNoCurveFlag || pxSurface2->m_bHasNoCurveFlag)
			{
				pxSurface->m_abStraightEdge[is] = true;
				pxSurface2->m_abStraightEdge[js] = true;
			}
			
			
		}
	}


	{
		for (int pID=0; pID<pxSurfaceMesh->NumberOfSurfaces; pID++)
		{
			Asura_Surface* surfaceDataPtr = &pxSurfaceMesh->Surfaces[pID];
			if (surfaceDataPtr->NumberOfVertices==3)
			{
				Asura_Vector_3& v0 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[0]].Position;
				Asura_Vector_3& v1 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[1]].Position;
				Asura_Vector_3& v2 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[2]].Position;
 
				surfaceDataPtr->ControlPoint[0][0] = v0;
				surfaceDataPtr->ControlPoint[3][0] = v1;
				surfaceDataPtr->ControlPoint[3][3] = v1;
				surfaceDataPtr->ControlPoint[0][3] = v2;
		
				// KJL 17/05/00 15:10:29 - tri edge 1
				{

					if(surfaceDataPtr->m_abStraightEdge[0])
					{
						surfaceDataPtr->ControlPoint[1][0] = (v0*2.0f+v1)/3.0f;
						surfaceDataPtr->ControlPoint[2][0] = (v0+v1*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v1-v0;
						float edgelength = edgedir.Magnitude()/3.0f;
						
						Asura_Vector_3& n0 = surfaceDataPtr->CNormalID[0][0];
						Asura_Vector_3& n1 = surfaceDataPtr->CNormalID[0][1];
						Asura_Vector_3 normal0 = (edgedir)^(n0);
						Asura_Vector_3 normal1 = (edgedir)^(n1);
						
						normal0.Normalise();
						normal1.Normalise();

						surfaceDataPtr->ControlPoint[1][0] = surfaceDataPtr->ControlPoint[0][0]+(n0^normal0)*edgelength;
						surfaceDataPtr->ControlPoint[2][0] = surfaceDataPtr->ControlPoint[3][0]-(n1^normal1)*edgelength;
					}
				}

				// KJL 17/05/00 15:10:29 - tri edge 2
				{

					if(surfaceDataPtr->m_abStraightEdge[1])
					{
						surfaceDataPtr->ControlPoint[1][3] = (v1*2.0f+v2)/3.0f;
						surfaceDataPtr->ControlPoint[2][3] = (v1+v2*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v2-v1;
						float edgelength = edgedir.Magnitude()/3.0f;

						Asura_Vector_3& n1 = surfaceDataPtr->CNormalID[1][0];
						Asura_Vector_3& n2 = surfaceDataPtr->CNormalID[1][1];
						Asura_Vector_3 normal2 = (edgedir)^(n2);
						Asura_Vector_3 normal1 = (edgedir)^(n1);
						
						normal2.Normalise();
						normal1.Normalise();

						surfaceDataPtr->ControlPoint[1][3] = surfaceDataPtr->ControlPoint[0][3]-(n2^normal2)*edgelength;
						surfaceDataPtr->ControlPoint[2][3] = surfaceDataPtr->ControlPoint[3][0]+(n1^normal1)*edgelength;
					}
				}

				// KJL 17/05/00 15:10:29 - tri edge 3
				{

					if(surfaceDataPtr->m_abStraightEdge[2])
					{
						surfaceDataPtr->ControlPoint[0][2] = (v2*2.0f+v0)/3.0f;
						surfaceDataPtr->ControlPoint[0][1] = (v2+v0*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v0-v2;
						float edgelength = edgedir.Magnitude()/3.0f;

						Asura_Vector_3& n2 = surfaceDataPtr->CNormalID[2][0];
						Asura_Vector_3& n0 = surfaceDataPtr->CNormalID[2][1];
						Asura_Vector_3 normal2 = (edgedir)^(n2);
						Asura_Vector_3 normal0 = (edgedir)^(n0);
					
						normal2.Normalise();
						normal0.Normalise();
					
						surfaceDataPtr->ControlPoint[0][2] = surfaceDataPtr->ControlPoint[0][3]+(n2^normal2)*edgelength;
						surfaceDataPtr->ControlPoint[0][1] = surfaceDataPtr->ControlPoint[0][0]-(n0^normal0)*edgelength;
					}
				}

				// KJL 17/05/00 15:20:31 - duplicate points
				surfaceDataPtr->ControlPoint[2][1] = surfaceDataPtr->ControlPoint[2][0];
				surfaceDataPtr->ControlPoint[2][2] = surfaceDataPtr->ControlPoint[2][3];
				surfaceDataPtr->ControlPoint[3][1] = surfaceDataPtr->ControlPoint[3][0];
				surfaceDataPtr->ControlPoint[3][2] = surfaceDataPtr->ControlPoint[3][0];	

				// KJL 17/05/00 15:22:28 - interpolated
				surfaceDataPtr->ControlPoint[1][1] = surfaceDataPtr->ControlPoint[0][1]+surfaceDataPtr->ControlPoint[1][0]-surfaceDataPtr->ControlPoint[0][0];
				surfaceDataPtr->ControlPoint[1][2] = surfaceDataPtr->ControlPoint[0][2]+surfaceDataPtr->ControlPoint[1][3]-surfaceDataPtr->ControlPoint[0][3];
			}
			else
			{
				Asura_Vector_3& v0 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[0]].Position;
				Asura_Vector_3& v1 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[1]].Position;
				Asura_Vector_3& v2 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[2]].Position;
				Asura_Vector_3& v3 = pxMesh->VertexDataPtr[surfaceDataPtr->VertexID[3]].Position;
 
				surfaceDataPtr->ControlPoint[0][0] = v0;
				surfaceDataPtr->ControlPoint[3][0] = v1;
				surfaceDataPtr->ControlPoint[3][3] = v2;
				surfaceDataPtr->ControlPoint[0][3] = v3;
				
				// KJL 17/05/00 15:10:29 - quad edge 1
				{

					if(surfaceDataPtr->m_abStraightEdge[0])
					{
						surfaceDataPtr->ControlPoint[1][0] = (v0*2.0f+v1)/3.0f;
						surfaceDataPtr->ControlPoint[2][0] = (v0+v1*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v1-v0;
						float edgelength = edgedir.Magnitude()/3.0f;

						Asura_Vector_3& n0 = surfaceDataPtr->CNormalID[0][0];
						Asura_Vector_3& n1 = surfaceDataPtr->CNormalID[0][1];
						Asura_Vector_3 normal0 = (edgedir)^(n0);
						Asura_Vector_3 normal1 = (edgedir)^(n1);
						
						normal0.Normalise();
						normal1.Normalise();

						surfaceDataPtr->ControlPoint[1][0] = surfaceDataPtr->ControlPoint[0][0]+(n0^normal0)*edgelength;
						surfaceDataPtr->ControlPoint[2][0] = surfaceDataPtr->ControlPoint[3][0]-(n1^normal1)*edgelength;
					}
				}

				// KJL 17/05/00 15:10:29 - quad edge 2
				{

					if(surfaceDataPtr->m_abStraightEdge[1])
					{
						surfaceDataPtr->ControlPoint[3][2] = (v2*2.0f+v1)/3.0f;
						surfaceDataPtr->ControlPoint[3][1] = (v2+v1*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v2-v1;
						float edgelength = edgedir.Magnitude()/3.0f;

						Asura_Vector_3& n1 = surfaceDataPtr->CNormalID[1][0];
						Asura_Vector_3& n2 = surfaceDataPtr->CNormalID[1][1];
						Asura_Vector_3 normal2 = (edgedir)^(n2);
						Asura_Vector_3 normal1 = (edgedir)^(n1);
						
						normal2.Normalise();
						normal1.Normalise();

						surfaceDataPtr->ControlPoint[3][2] = surfaceDataPtr->ControlPoint[3][3]-(n2^normal2)*edgelength;
						surfaceDataPtr->ControlPoint[3][1] = surfaceDataPtr->ControlPoint[3][0]+(n1^normal1)*edgelength;
					}
				}

				// KJL 17/05/00 15:10:29 - quad edge 3
				{

					if(surfaceDataPtr->m_abStraightEdge[2])
					{
						surfaceDataPtr->ControlPoint[2][3] = (v2*2.0f+v3)/3.0f;
						surfaceDataPtr->ControlPoint[1][3] = (v2+v3*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v3-v2;
						float edgelength = edgedir.Magnitude()/3.0f;

						Asura_Vector_3& n2 = surfaceDataPtr->CNormalID[2][0];
						Asura_Vector_3& n3 = surfaceDataPtr->CNormalID[2][1];
						Asura_Vector_3 normal2 = (edgedir)^(n2);
						Asura_Vector_3 normal3 = (edgedir)^(n3);
						
						normal2.Normalise();
						normal3.Normalise();

						surfaceDataPtr->ControlPoint[2][3] = surfaceDataPtr->ControlPoint[3][3]+(n2^normal2)*edgelength;
						surfaceDataPtr->ControlPoint[1][3] = surfaceDataPtr->ControlPoint[0][3]-(n3^normal3)*edgelength;
					}
				}


				// KJL 17/05/00 15:10:29 - quad edge 4
				{

					if(surfaceDataPtr->m_abStraightEdge[3])
					{
						surfaceDataPtr->ControlPoint[0][2] = (v3*2.0f+v0)/3.0f;
						surfaceDataPtr->ControlPoint[0][1] = (v3+v0*2.0f)/3.0f;
					}
					else
					{
						Asura_Vector_3 edgedir = v0-v3;
						float edgelength = edgedir.Magnitude()/3.0f;

						Asura_Vector_3& n3 = surfaceDataPtr->CNormalID[3][0];
						Asura_Vector_3& n0 = surfaceDataPtr->CNormalID[3][1];
						Asura_Vector_3 normal3 = (edgedir)^(n3);
						Asura_Vector_3 normal0 = (edgedir)^(n0);
						
						normal3.Normalise();
						normal0.Normalise();

						surfaceDataPtr->ControlPoint[0][2] = surfaceDataPtr->ControlPoint[0][3]+(n3^normal3)*edgelength;
						surfaceDataPtr->ControlPoint[0][1] = surfaceDataPtr->ControlPoint[0][0]-(n0^normal0)*edgelength;
					}
				}

			
				// KJL 17/05/00 15:22:28 - interpolated
				surfaceDataPtr->ControlPoint[1][1] = surfaceDataPtr->ControlPoint[0][1]+surfaceDataPtr->ControlPoint[1][0]-surfaceDataPtr->ControlPoint[0][0];
				surfaceDataPtr->ControlPoint[1][2] = surfaceDataPtr->ControlPoint[0][2]+surfaceDataPtr->ControlPoint[1][3]-surfaceDataPtr->ControlPoint[0][3];
				surfaceDataPtr->ControlPoint[2][1] = surfaceDataPtr->ControlPoint[2][0]+surfaceDataPtr->ControlPoint[3][1]-surfaceDataPtr->ControlPoint[3][0];
				surfaceDataPtr->ControlPoint[2][2] = surfaceDataPtr->ControlPoint[3][2]+surfaceDataPtr->ControlPoint[2][3]-surfaceDataPtr->ControlPoint[3][3];
			}

			{
				for(int x=0; x<4; x++)
				for(int y=0; y<4; y++)
				{
					xPatch.ControlPoint[x][y]=xPosition+surfaceDataPtr->ControlPoint[x][y];
					surfaceDataPtr->ControlPoint[x][y] = xPatch.ControlPoint[x][y];
					//TranslatePointIntoViewspace(&xPatch.ControlPoint[x][y]);
				}
			}

			surfaceDataPtr->ControlPointsValid=true;
				
			for(int n=0; n<4; n++)
			{
				surfaceDataPtr->Normal[n] = surfaceDataPtr->LNormalID[n];
			}
			if (surfaceDataPtr->NumberOfVertices==3)
			{
				surfaceDataPtr->Normal[2] = surfaceDataPtr->LNormalID[1];
				surfaceDataPtr->Normal[3] = surfaceDataPtr->LNormalID[2];
				surfaceDataPtr->U[3] = surfaceDataPtr->U[2];
				surfaceDataPtr->U[2] = surfaceDataPtr->U[1];
				surfaceDataPtr->V[3] = surfaceDataPtr->V[2];
				surfaceDataPtr->V[2] = surfaceDataPtr->V[1];
				surfaceDataPtr->U2[3] = surfaceDataPtr->U2[2];
				surfaceDataPtr->U2[2] = surfaceDataPtr->U2[1];
				surfaceDataPtr->V2[3] = surfaceDataPtr->V2[2];
				surfaceDataPtr->V2[2] = surfaceDataPtr->V2[1];
				surfaceDataPtr->Colour[3] = surfaceDataPtr->Colour[2];
				surfaceDataPtr->Colour[2] = surfaceDataPtr->Colour[1];

#if defined(ASURA_TARGET_SM3)

				surfaceDataPtr->m_axLightTransportHarmonic[3] = surfaceDataPtr->m_axLightTransportHarmonic[2];
				surfaceDataPtr->m_axLightTransportHarmonic[2] = surfaceDataPtr->m_axLightTransportHarmonic[1];

				surfaceDataPtr->m_axDiffuseSH[3] = surfaceDataPtr->m_axDiffuseSH[2];
				surfaceDataPtr->m_axDiffuseSH[2] = surfaceDataPtr->m_axDiffuseSH[1];

				// VGP 13Feb2006
				surfaceDataPtr->m_axBrightestLightVector[3] = surfaceDataPtr->m_axBrightestLightVector[2];
				surfaceDataPtr->m_axBrightestLightVector[2] = surfaceDataPtr->m_axBrightestLightVector[1];

				surfaceDataPtr->m_axTangent[3] = surfaceDataPtr->m_axTangent[2];
				surfaceDataPtr->m_axTangent[2] = surfaceDataPtr->m_axTangent[1];
				surfaceDataPtr->m_axTangent2[3] = surfaceDataPtr->m_axTangent2[2];
				surfaceDataPtr->m_axTangent2[2] = surfaceDataPtr->m_axTangent2[1];

				surfaceDataPtr->m_axBinormal[3] = surfaceDataPtr->m_axBinormal[2];
				surfaceDataPtr->m_axBinormal[2] = surfaceDataPtr->m_axBinormal[1];
				surfaceDataPtr->m_axBinormal2[3] = surfaceDataPtr->m_axBinormal2[2];
				surfaceDataPtr->m_axBinormal2[2] = surfaceDataPtr->m_axBinormal2[1];

#endif
			}

		

			for (int i=0; i<4; i++)
			{
				float u,v;
				switch(i)
				{
					default:
					case 0:
						u=v=0.0f;
						break;
					case 1:
						u=1.0f;
						v=0.0f;
						break;
					case 2:
						u=v=1.0f;
						break;
					case 3:
						u=0.0f;
						v=1.0f;
						break;
				}
				surfaceDataPtr->du0dv0[i] = xPatch.EvaluateCurve(u,v);
				surfaceDataPtr->du2dv0[i] = xPatch.EvaluateCurve_d2u(u,v);
				surfaceDataPtr->du0dv2[i] = xPatch.EvaluateCurve_d2v(u,v);
				//surfaceDataPtr->Normal[i] = xPatch.EvaluateCurve_Normal(u,v);
			
			}

		
			surfaceDataPtr->SubdivideDataValid=true;

		}
	}
}

