// Tools_Audio_DSP.cpp
// Thomas Young / Andrew McDonald - 17.03.09
// Copyright (C) 2009 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_AUDIO_SYSTEM

#include "Sound/Tools_Audio_DSP.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_DSP.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_List.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_String.h"
#include "Core/Asura_StringBuffer.h"
#include "Core/Asura_UserInput.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_XAudio2.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"

// Tools
#include "Tools_Overlays.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_Dialog.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Timed_Messages.h"

// G L O B A L   V A R I A B L E S ///////////////////////////////////////////

extern unsigned int Asura_OverlayBrightness;

// C O N S T A N T S /////////////////////////////////////////////////////////

const char* const aszTOOLS_DSP_EFFECT_NAMES[] =	// Should match Asura_Audio_DSPType enum
{
	"Simple Filter",
	"Add Noise",
	"Flanger",
	"Chorus"
};

const char* const aszTOOLS_DSP_FILTER_TYPES[] =	// Should match Asura_DSP_Params_SimpleFilter::FILTER_TYPE enum
{
	"Lowpass",
	"Highpass",
	"Bandpass",
	"Bandstop"
};

const u_int uTOOLS_DSP_NUM_EFFECT_NAMES = sizeof( aszTOOLS_DSP_EFFECT_NAMES ) / sizeof( aszTOOLS_DSP_EFFECT_NAMES[0] );

ASURA_STATIC_ASSERT( uTOOLS_DSP_NUM_EFFECT_NAMES == ASURA_DSP_NUM_TYPES );

static const float fDSP_CHAIN_SUBWINDOW_NODE_WIDTH				= 120.0f;
static const float fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT				= 75.0f;
static const float fDSP_CHAIN_SUBWINDOW_NODE_STARTX				= 20.0f;
static const float fDSP_CHAIN_SUBWINDOW_NODE_STARTY				= 20.0f;
static const float fDSP_CHAIN_SUBWINDOW_NODE_SPACING			= 30.0f;
static const float fDSP_EFFECT_ADD_SYMBOL_WIDTH					= 7.0f;
static const float fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER	= 8.0f;
static const u_int uDSP_CHAIN_SUBWINDOW_BACKGROUND_COLOUR		= 0x00505070;
static const u_int uDSP_EFFECT_SELECTION_COLOUR					= 0x60999999;
static const u_int uDSP_EFFECT_SHADOW_COLOUR					= 0x40000000;
static const u_int uDSP_EFFECT_BACK_COLOUR						= 0x00B1D7FC;
static const u_int uDSP_EFFECT_BACK_HIGHLIGHTED_COLOUR			= 0x00C5E9FF;
static const u_int uDSP_EFFECT_BORDER_COLOUR					= 0x00000000;
static const u_int uDSP_EFFECT_ADD_BACK_COLOUR					= 0x80B1D7FC;
static const u_int uDSP_EFFECT_ADD_SYMBOL_COLOUR				= 0x00AEF46F;
static const u_int uDSP_EFFECT_ADD_BORDER_COLOUR				= 0xA0000000;
static const u_int uDSP_EFFECT_ADD_TEXT_COLOUR					= 0x00000000;
static const u_int uDSP_EFFECT_DSP_TITLE_COLOUR					= 0x00000000;
static const u_int uDSP_EFFECT_ICON_BACK_COLOUR					= 0x00EEEEEE;
static const u_int uDSP_EFFECT_ICON_FRONT_COLOUR				= 0x00000000;
static const u_int uDSP_EFFECT_ICON_FRONT_COLOUR_FADE			= 0x00888888;
static const u_int uDSP_EFFECT_ICON_BORDER_COLOUR				= 0x00333333;
static const u_int uDSP_ARROW_COLOUR							= 0x00EEEEEE;

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////

//___________________ Tools_Audio_DSPChainListOverlay ________________________________

class Tools_Audio_DSPChainListOverlay : public Asura_Overlay
{
	public:

		enum DialogID
		{
			DialogID_AddEffect,
			DialogID_EffectProperty,
		};

		explicit		Tools_Audio_DSPChainListOverlay		( Asura_Vector_2* const pxPos );
		virtual			~Tools_Audio_DSPChainListOverlay	( );

		void			UpdateChainListStrings				( );

		static bool 	IsOpen								( )		{ return s_pxThis != NULL; }

		static Tools_Audio_DSPChainListOverlay* Instance	( )		{ return s_pxThis; }

	protected:

		void GetChainNames( const char** const pszNames );
		Asura_Audio_DSPChain* GetSelectedChain ();

		static void OnSelectionChange	( );
		static void OnCreateChain		( );
		static void OnEditChain			( );
		static void OnDeleteChain		( );

		static Tools_Audio_DSPChainListOverlay* s_pxThis;

		Asura_List_Dialog* m_pxChainListDlg;

		const char** m_pszChainNames;

		int m_iNumChainNames;
		int m_iSelectedChain;
};

//___________________ Tools_Audio_DSPChainOverlay ________________________________

class Tools_Audio_DSPChainOverlay : public Asura_Overlay
{
	friend class Tools_Audio_DSPChainSubwindow;

	public:

		explicit Tools_Audio_DSPChainOverlay			( Asura_Vector_2* const pxPos, Asura_Audio_DSPChain* pxChain, const int iSelectedEffect = -1 );
		virtual ~Tools_Audio_DSPChainOverlay			( );

		static bool IsOpen								( )									{ return s_pxThis != NULL; }
		static bool IsEditingThisChain					( Asura_Audio_DSPChain* pxChain)	{ return s_pxThis ? ( s_pxThis->m_pxActiveChain == pxChain ) : false; }

		static Tools_Audio_DSPChainOverlay* Instance	( )									{ return s_pxThis; }

	protected:

		// _____________ Functions _______________

		void			AddOverlayControls		( );
		
		static void 	SelectEffect			( const int iEffect );
		
		static void 	AddEffect				( );
		
		static void 	OnMoveUpChain			( );
		static void 	OnMoveDownChain			( );
		static void 	OnDeleteEffect			( );

		static void 	MoveUpChain				( const int iEffectIndex );

		static void 	RecreateDialog			( );

		virtual void	PerFrameUpdate			( );
		
		// Effect specific functions

		static void		UpdateFilterParams		( );
		static bool		IsFilterValid			( int iIgnore );

		// _____________ Variables _______________
		
		static Tools_Audio_DSPChainOverlay* s_pxThis;	// Should only be one of these around

		Asura_Audio_DSPChain* m_pxActiveChain;

		Tools_Audio_DSPChainSubwindow* m_pxEffectSubwindow;

		const char** m_pszEffectNames;

		int m_iNumEffectNames;
		int m_iSelectedEffectInChain;

		// Effect specific variables

		bool m_bFilterParamsChanged;
		int m_iSavedFilterType;
};

//___________________ Tools_Audio_DSPChainSubwindow ________________________________

class Tools_Audio_DSPChainSubwindow : public Asura_Subwindow_Dialog
{
	public:

		Tools_Audio_DSPChainSubwindow  ( const Asura_Vector_2& xOffset, const float fWidth, const float fHeight );
		virtual ~Tools_Audio_DSPChainSubwindow ();

		void	Update( const Asura_Vector_2& xOrigin );
		void	Render( const Asura_Vector_2& xOrigin );

		void	RenderDSPIcon( const Asura_Vector_2& xOrigin, const Asura_Audio_DSPType eDSPType );
		
	protected:

		virtual bool IsPointInDialog (const Asura_Vector_2& xOrigin,const Asura_Vector_2& xPos) const;

	private:

		int	m_iHoveringOver;
};

//___________________ Tools_Audio_DSPEffectPicker ________________________________

class Tools_Audio_DSPEffectPicker : public Tools_Modal_Overlay
{
	public:

		Tools_Audio_DSPEffectPicker		( );
		~Tools_Audio_DSPEffectPicker	( );

		int GetSelectedEffectType		( )	{ return m_iReturnedSelection; }

	protected:
	
		virtual void InitialiseOverlay(class Asura_Overlay* Overlay);

		static void OnAccept	( );
		static void OnDecline	( );

		void Accept				( );
		void Decline			( );
		
		static Tools_Audio_DSPEffectPicker* s_pxThis;

		int m_iListSelection;
		int m_iReturnedSelection;

		Asura_List_Dialog* m_pxEffectList;
};

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

// Tools_Audio_DSP

Asura_Vector_2	Tools_Audio_DSP::s_xChainListOverlayPos		( 0.0f, 0.0f );
Asura_Vector_2	Tools_Audio_DSP::s_xChainOverlayPos			( 0.0f, 0.0f );
bool			Tools_Audio_DSP::s_bModified				= false;

// Tools_Audio_DSPChainListOverlay

Tools_Audio_DSPChainListOverlay* Tools_Audio_DSPChainListOverlay::s_pxThis = NULL;

// Tools_Audio_DSPChainOverlay

Tools_Audio_DSPChainOverlay* Tools_Audio_DSPChainOverlay::s_pxThis = NULL;

// Tools_DSPChain_DropDownList_Dialog

int Tools_DSPChain_DropDownList_Dialog::s_iChainListValidID = 0;

// Tools_Audio_DSPEffectPicker

Tools_Audio_DSPEffectPicker* Tools_Audio_DSPEffectPicker::s_pxThis = NULL;

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

static void RenderDashedLine( const float fX1, const float fY1, const float fX2, const float fY2, const u_int uCol )
{
	const u_int uSegments = 13; // use an odd number
	for( u_int uSeg = 0; uSeg < uSegments; uSeg += 2 )
	{
		const float fPosition1 = 1.0f / static_cast<float>( uSegments ) * static_cast<float>( uSeg );
		const float fPosition2 = 1.0f / static_cast<float>( uSegments ) * static_cast<float>( uSeg + 1 );

		const float fNewX1 = Asura_Maths::Lerp( fX1, fX2, fPosition1 );
		const float fNewX2 = Asura_Maths::Lerp( fX1, fX2, fPosition2 );
		const float fNewY1 = Asura_Maths::Lerp( fY1, fY2, fPosition1 );
		const float fNewY2 = Asura_Maths::Lerp( fY1, fY2, fPosition2 );

		Asura_Render::Render2DLine( fNewX1, fNewY1, fNewX2, fNewY2, uCol );
	}
}


// _______________________ Tools_Audio_DSP __________________________


void Tools_Audio_DSP::CreateOverlay()
{
	Tools_Audio_DSP::TransferModeData( TOOLSMODEDATAFILE_LOAD );

	Tools_Audio_DSPChainListOverlay* const pxNewOverlay = new Tools_Audio_DSPChainListOverlay( &s_xChainListOverlayPos );

	Asura_Assert( pxNewOverlay, "Failed to allocate DSP overlay." );
	if( !pxNewOverlay ) return;

	Asura_Overlay_System::SubmitNewOverlay( pxNewOverlay );
}


void Tools_Audio_DSP::DestroyOverlay()
{
	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_DSP_CHAINLIST );
	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_DSP_CHAIN );
}


void Tools_Audio_DSP::ToggleOverlay()
{
	if( Tools_Audio_DSPChainListOverlay::IsOpen() )
	{
		DestroyOverlay();
	}
	else
	{
		CreateOverlay();
	}
}

bool Tools_Audio_DSP::IsOverlayOpen ( )
{
	return ( Tools_Audio_DSPChainListOverlay::IsOpen() || Tools_Audio_DSPChainOverlay::IsOpen() );
}


void Tools_Audio_DSP::TransferModeData( const int iMode )
{
	//open the file
	Tools_ModeDataFile file( "DSPChains.dat", iMode );

	if( false == file.GetError() )
	{
		file.Transfer( s_xChainListOverlayPos );
		file.Transfer( s_xChainOverlayPos );
	}
}

//_____________________ Tools_Audio_DSPChainListOverlay __________________________________


Tools_Audio_DSPChainListOverlay::Tools_Audio_DSPChainListOverlay( Asura_Vector_2* const pxPos )
	: Asura_Overlay( "DSP Chain list", false, pxPos )
	, m_pxChainListDlg( NULL )
	, m_pszChainNames( NULL )
	, m_iNumChainNames(0)
	, m_iSelectedChain(0)
{
	m_uID = TOOLS_OVERLAY_DSP_CHAINLIST;

	// Add dialogs
	Asura_Vector_2 xOffset( 1.0f, 2.0f );
	Asura_Dialog* pxDialog;
	
	//static Asura_Vector_2 xChainListDlgSize( 20.0f, 20.0f );

	// Chain list
	m_pxChainListDlg = new Asura_List_Dialog( xOffset, m_pszChainNames, &m_iSelectedChain, 10, 0, OnSelectionChange, NULL, true );
	m_pxChainListDlg->SetExpandWidthToFillOverlay( true );
	//m_pxChainListDlg->SetAllowResize( false, &xChainListDlgSize );
	m_pxChainListDlg->SetSingleClickFunction( Asura_Callback_Void< void >( OnSelectionChange ) );
	AddDialog( m_pxChainListDlg );
	xOffset.y += m_pxChainListDlg->m_xMaxCoord.y;

	// Create button
	pxDialog = new Asura_Function_Dialog( "Create DSP Chain", xOffset, OnCreateChain );
	pxDialog->SetHelpText( "Creates a new empty DSP chain." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Create button
	pxDialog = new Asura_Function_Dialog( "Edit Selected DSP Chain", xOffset, OnEditChain );
	pxDialog->SetHelpText( "Creates a new empty DSP chain." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Delete button
	pxDialog = new Asura_Function_Dialog( "Delete Selected DSP Chain", xOffset, OnDeleteChain );
	pxDialog->SetHelpText( "Deletes the selected DSP chain." );
	AddDialog( pxDialog );

	xOffset.x = m_pxChainListDlg->m_xMaxCoord.x + 2.0f;
	xOffset.y = 2.0f;

	UpdateChainListStrings();

	Asura_Assert( !s_pxThis, "Created more than one DSP overlay." );
	s_pxThis = this;
}


Tools_Audio_DSPChainListOverlay::~Tools_Audio_DSPChainListOverlay()
{
	delete[] m_pszChainNames;
	s_pxThis = NULL;

	Tools_Audio_DSP::TransferModeData( TOOLSMODEDATAFILE_SAVE );
}

void Tools_Audio_DSPChainListOverlay::GetChainNames( const char** const pszNames )
{
	const int iNumChains = Asura_Audio_System::GetNumDSPChains();

	for( int i = 0; i < iNumChains; ++i )
	{
		pszNames[i] = 0;

		Asura_Audio_DSPChain* pxChain = Asura_Audio_System::GetDSPChainByIndex( i );
		if( pxChain )
		{
			pszNames[i] = pxChain->m_xName.GetString();
		}
	}

	// TY 28.04.09 - Dont sort them, we need them to be in the order they are stored in the audio system
	// Asura_SortAndSearch::Quicksort( pszNames, iNumChains, Asura_SortAndSearch::SortStringsCaseInsensitive );
}


void Tools_Audio_DSPChainListOverlay::UpdateChainListStrings()
{
	// We're about to delete the name list, so store a pointer to the name in the
	// object itself instead
	const Asura_Audio_DSPChain* const pxChain = GetSelectedChain();
	const char* const szSelectedName = (pxChain ? pxChain->m_xName.GetString() : NULL);

	// Rebuild the name list

	const int iNumChains = Asura_Audio_System::GetNumDSPChains();
	const char** pszChainNames = new const char*[ iNumChains ];

	Asura_Assert( pszChainNames, "Failed to allocate DSP chain name array." );
	if( !pszChainNames ) return;

	GetChainNames( pszChainNames );

	delete[] m_pszChainNames;
	m_pszChainNames = pszChainNames;
	m_iNumChainNames = iNumChains;

	if( m_pxChainListDlg )
	{
		m_pxChainListDlg->ReplaceStrings( iNumChains, pszChainNames );
	}
	
	// Find the new index of our selection

	for( int i = 0; i < iNumChains; ++i )
	{
		const char* const szChainName = pszChainNames[i];
		if( Asura_StringUtil::CompareAgainstWildcardString( szChainName, szSelectedName, false ) )
		{
			// The list dialog will automatically update following this variable
			m_iSelectedChain = i;
			break;
		}
	}

	Tools_DSPChain_DropDownList_Dialog::RefreshChains();
}



Asura_Audio_DSPChain* Tools_Audio_DSPChainListOverlay::GetSelectedChain()
{
	if( m_iSelectedChain < 0 || m_iSelectedChain >= m_iNumChainNames ) return NULL;

	return Asura_Audio_System::GetDSPChainByIndex( m_iSelectedChain );
}


void Tools_Audio_DSPChainListOverlay::OnSelectionChange()
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) 
	{
		return;
	}

	Asura_Audio_DSPChain* pxChain = s_pxThis->GetSelectedChain();

	if( !pxChain )
	{
		return;
	}

	if( Tools_Audio_DSPChainOverlay::IsOpen() )
	{
		Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_DSP_CHAIN );

		Tools_Audio_DSPChainOverlay* const pxNewOverlay = new Tools_Audio_DSPChainOverlay( &Tools_Audio_DSP::s_xChainOverlayPos, pxChain );

		Asura_Assert( pxNewOverlay, "Failed to allocate DSP overlay." );
		if( !pxNewOverlay ) 
		{
			return;
		}

		Asura_Overlay_System::SubmitNewOverlay( pxNewOverlay );
	}
}


void Tools_Audio_DSPChainListOverlay::OnCreateChain()
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	GetString_Overlay xNamePicker( "New DSP Chain", "" );
	if( !xNamePicker.DoModal() ) return;

	const char* const szName = xNamePicker.GetString();
	const Asura_Hash_ID uNameHash = Asura_GetHashID( szName );

	if( uNameHash == ASURA_HASH_ID_UNSET ) return;

	if( Asura_Audio_System::GetDSPChainByHash( uNameHash ) )
	{
		ShowMessageOverlay( "DSP chain '%s' already exists.", szName );
		return;
	}

	Asura_Audio_DSPChain xChain;
	xChain.m_xName = szName;

	Asura_Audio_System::AddDSPChain( xChain.m_xName.GetHash(), &xChain );
	Tools_Audio_DSP::Modified() = true;

	s_pxThis->UpdateChainListStrings();
}


void Tools_Audio_DSPChainListOverlay::OnEditChain()
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	if( Tools_Audio_DSPChainOverlay::IsEditingThisChain( s_pxThis->GetSelectedChain() ) )
	{
		// No need to recreate
		return;
	}

	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_DSP_CHAIN );

	Tools_Audio_DSPChainOverlay* const pxNewOverlay = new Tools_Audio_DSPChainOverlay( &Tools_Audio_DSP::s_xChainOverlayPos, s_pxThis->GetSelectedChain() );

	Asura_Assert( pxNewOverlay, "Failed to allocate DSP overlay." );
	if( !pxNewOverlay ) return;

	Asura_Overlay_System::SubmitNewOverlay( pxNewOverlay );
}


void Tools_Audio_DSPChainListOverlay::OnDeleteChain()
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	if( s_pxThis->m_iSelectedChain >= s_pxThis->m_iNumChainNames ) return;

	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_DSP_CHAIN );

	const char* const szSelectedChain = s_pxThis->m_pszChainNames[ s_pxThis->m_iSelectedChain ];
	const Asura_Hash_ID uSelectedHash = Asura_GetHashID( szSelectedChain );

	Asura_Audio_System::RemoveDSPChain( uSelectedHash );
	
	s_pxThis->m_iSelectedChain = -1;

	Tools_Audio_DSP::Modified() = true;

	s_pxThis->UpdateChainListStrings();
}


//_____________________ Tools_Audio_DSPChainOverlay __________________________________

Tools_Audio_DSPChainOverlay::Tools_Audio_DSPChainOverlay( Asura_Vector_2* const pxPos, Asura_Audio_DSPChain* pxChain, const int iSelectedEffect )
: Asura_Overlay( "DSP Chain", false, pxPos )
, m_pxEffectSubwindow( NULL )
, m_pszEffectNames( NULL )
, m_iNumEffectNames(0)
, m_iSelectedEffectInChain( iSelectedEffect )
, m_pxActiveChain( pxChain )
{
	m_uID = TOOLS_OVERLAY_DSP_CHAIN;

	Asura_Assert( m_pxActiveChain, "Expected DSP chain here");
	Asura_Assert( !s_pxThis, "Created more than one DSP overlay" );
	s_pxThis = this;
	
	SetTitle( pxChain->m_xName.GetString() );

	m_bFilterParamsChanged = false;
	m_iSavedFilterType = -1;

	AddOverlayControls();
}


Tools_Audio_DSPChainOverlay::~Tools_Audio_DSPChainOverlay()
{
	s_pxThis = 0;
}

void Tools_Audio_DSPChainOverlay::AddOverlayControls ()
{
	// Add dialogs
	Asura_Vector_2 xOffset( 1.0f, 2.0f );
	
	const float fSubwindowWidth		= 75.0f;
	const float fSubwindowHeight	= ( fDSP_CHAIN_SUBWINDOW_NODE_STARTY * 2.0f + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT ) / ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	// Chain list
	m_pxEffectSubwindow = new Tools_Audio_DSPChainSubwindow( xOffset, fSubwindowWidth, fSubwindowHeight );
	AddDialog( m_pxEffectSubwindow );

	if( m_iSelectedEffectInChain != -1 )
	{
		Asura_Dialog* pxDialog;

		// If we have a selected effect, add the basic controls
		xOffset.y += m_pxEffectSubwindow->m_xMaxCoord.y;

		Asura_Vector_2 xSavedPos = xOffset;

		pxDialog = new Asura_Function_Dialog( "Move effect earlier", xOffset, OnMoveUpChain );
		pxDialog->SetHelpText( "Moves the effect earlier in the chain" );
		AddDialog( pxDialog );

		xOffset.y += Asura_Dialog::DefaultSpacing();

		pxDialog = new Asura_Function_Dialog( "Move effect later", xOffset, OnMoveDownChain );
		pxDialog->SetHelpText( "Moves the effect down (later) in the chain" );
		AddDialog( pxDialog );

		xOffset.y += Asura_Dialog::LargeSpacing();

		pxDialog = new Asura_Function_Dialog( "Remove Effect", xOffset, OnDeleteEffect );
		pxDialog->SetHelpText( "Remove this effect from the chain" );
		AddDialog( pxDialog );

		xOffset = xSavedPos;
		xOffset.x += 15.0f;

		pxDialog = new Asura_Divider_Dialog( xOffset, 0xFFFFFF, Asura_Divider_Dialog::VERTICAL | Asura_Divider_Dialog::EXTEND_POSITIVE );
		AddDialog( pxDialog );

		xOffset.x += 2.0f;

		if( !m_pxActiveChain )
		{
			return;
		}

		switch( m_pxActiveChain->m_ppxEffects[ m_iSelectedEffectInChain ]->GetType() )
		{
			case ASURA_DSP_SIMPLE_FILTER:
			{
				Asura_DSP_Params_SimpleFilter* pxEffect = reinterpret_cast< Asura_DSP_Params_SimpleFilter* >( m_pxActiveChain->m_ppxEffects[ m_iSelectedEffectInChain ] );

				pxDialog = new Asura_Dialog( "Simple Filter", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Dialog( "Filters the signal, attenuating certain frequencies", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				m_iSavedFilterType = static_cast<int>( pxEffect->m_xParams.m_uType );

				pxDialog = new Asura_DropDown_List_Dialog( xOffset, aszTOOLS_DSP_FILTER_TYPES, &pxEffect->m_xParams.m_uType, 4, 4, 0, &s_pxThis->m_bFilterParamsChanged,"Filter type", true );
				pxDialog->SetHelpText( "Filter type, this will affect the meaning of the other controls shown here" );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				switch( pxEffect->m_xParams.m_uType )
				{
					case Asura_DSP_Params_SimpleFilter::FILTER_TYPE_LOWPASS:
					case Asura_DSP_Params_SimpleFilter::FILTER_TYPE_HIGHPASS:
					{
						pxDialog = new Asura_Numerical_Dialog( "Frequency", xOffset, &pxEffect->m_xParams.m_fFrequency, 100.0f, 22000.0f, 50.0f, "%.1f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "Cut off frequency" );
						AddDialog( pxDialog );
						xOffset.x += 25.0f;

						pxDialog = new Asura_Numerical_Dialog( "Attenuation", xOffset, &pxEffect->m_xParams.m_fAttenuation, 0.0f, 0.9f, 0.1f, "%.2f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "Gain of the ATTENUATED frequencies\n( 0.0 = -30db, 1.0 = Original amplitude )" );
						AddDialog( pxDialog );
						xOffset.y += Asura_Function_Dialog::DefaultSpacing();

						pxDialog = new Asura_Numerical_Dialog( "Gain adjustment", xOffset, &pxEffect->m_xParams.m_fGain, 0.5f, 3.0f, 0.1f, "%.2f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "The filtered signal will be adjusted by this much\nuse this to make small amplitude adjustments" );
						AddDialog( pxDialog );
						xOffset.x -= 25.0f;
					}
					break;

					case Asura_DSP_Params_SimpleFilter::FILTER_TYPE_BANDPASS:
					case Asura_DSP_Params_SimpleFilter::FILTER_TYPE_BANDSTOP:
					{
						pxDialog = new Asura_Numerical_Dialog( "Frequency 1", xOffset, &pxEffect->m_xParams.m_fFrequency, 100.0f, 22000.0f, 50.0f, "%.1f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "Pass/Stop band frequency 1 (left edge)" );
						AddDialog( pxDialog );
						xOffset.x += 25.0f;

						pxDialog = new Asura_Numerical_Dialog( "Attenuation", xOffset, &pxEffect->m_xParams.m_fAttenuation, 0.0f, 0.9f, 0.1f, "%.2f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "Gain of the ATTENUATED frequencies\n( 0.0 = -30db, 1.0 = Original amplitude )" );
						AddDialog( pxDialog );
						xOffset.y += Asura_Function_Dialog::DefaultSpacing();

						pxDialog = new Asura_Numerical_Dialog( "Gain adjustment", xOffset, &pxEffect->m_xParams.m_fGain, 0.5f, 3.0f, 0.1f, "%.2f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "The filtered signal will be adjusted by this much\nuse this to make small amplitude adjustments" );
						AddDialog( pxDialog );
						xOffset.x -= 25.0f;

						pxDialog = new Asura_Numerical_Dialog( "Frequency 2", xOffset, &pxEffect->m_xParams.m_fFrequency2, 100.0f, 22000.0f, 50.0f, "%.1f", &s_pxThis->m_bFilterParamsChanged );
						pxDialog->SetHelpText( "Pass/Stop band frequency 2 (right edge)" );
						AddDialog( pxDialog );
					}
					break;
				}
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				pxDialog = new Asura_Dialog( "Filter generation failed, please try different parameters.", xOffset );
				pxDialog->SetHidden( true );
				AddDialog( pxDialog );
				
			}
			break;

			case ASURA_DSP_ADD_NOISE:
			{
				Asura_DSP_Params_AddNoise* pxEffect = reinterpret_cast< Asura_DSP_Params_AddNoise* >( m_pxActiveChain->m_ppxEffects[ m_iSelectedEffectInChain ] );

				pxDialog = new Asura_Dialog( "Add Noise", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Dialog( "Distorts the playing audio by adding static to the signal", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Effect strength", xOffset, &pxEffect->m_xParams.m_fNoiseAmount, 0.0f, 1.0f,0.1f, "%.2f", &Tools_Audio_DSP::s_bModified );
				AddDialog( pxDialog );
			}
			break;

			case ASURA_DSP_FLANGER:
			{
				Asura_DSP_Params_Flanger* pxEffect = reinterpret_cast< Asura_DSP_Params_Flanger* >( m_pxActiveChain->m_ppxEffects[ m_iSelectedEffectInChain ] );

				pxDialog = new Asura_Dialog( "Flanger", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Dialog( "A swept comb filter (a.k.a. flanger) effect", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Cycle range", xOffset, &pxEffect->m_xParams.m_fRange, 0.1f, 4.0f, 0.1f, "%.2f", &Tools_Audio_DSP::s_bModified );
				AddDialog( pxDialog );
				pxDialog->SetHelpText( "Number of seconds to do one sweep (1/2 the period)" );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Wet/Dry Mix", xOffset, &pxEffect->m_xParams.m_fWetDry, 0.0f, 1.0f,0.1f, "%.2f", &Tools_Audio_DSP::s_bModified );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Gain adjustment", xOffset, &pxEffect->m_xParams.m_fGain, 0.5f, 3.0f, 0.1f, "%.2f", &s_pxThis->m_bFilterParamsChanged );
				AddDialog( pxDialog );
			}
			break;

			case ASURA_DSP_CHORUS:
			{
				Asura_DSP_Params_Chorus* pxEffect = reinterpret_cast< Asura_DSP_Params_Chorus* >( m_pxActiveChain->m_ppxEffects[ m_iSelectedEffectInChain ] );

				pxDialog = new Asura_Dialog( "Chorus", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Dialog( "Emulates the signal being emitted from multiple sources", xOffset );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Delay Time (ms)", xOffset, &pxEffect->m_xParams.m_fDelay, 15.0f, 25.0f, 0.5f, "%.1f", &Tools_Audio_DSP::s_bModified );
				AddDialog( pxDialog );
				pxDialog->SetHelpText( "Offset of the old signal" );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Sweep Depth (ms)", xOffset, &pxEffect->m_xParams.m_fSweepDepth, 0.0f, 8.0f, 0.5f, "%.1f", &Tools_Audio_DSP::s_bModified );
				AddDialog( pxDialog );
				pxDialog->SetHelpText( "The amount to vary the delay" );
				xOffset.y += Asura_Function_Dialog::LargeSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Wet/Dry Mix", xOffset, &pxEffect->m_xParams.m_fWetDry, 0.0f, 1.0f,0.1f, "%.2f", &Tools_Audio_DSP::s_bModified );
				AddDialog( pxDialog );
				xOffset.y += Asura_Function_Dialog::DefaultSpacing();

				pxDialog = new Asura_Numerical_Dialog( "Gain adjustment", xOffset, &pxEffect->m_xParams.m_fGain, 0.5f, 3.0f, 0.1f, "%.2f", &s_pxThis->m_bFilterParamsChanged );
				AddDialog( pxDialog );
			}
			break;
		}
	}

	ResizeOverlay();
}

void Tools_Audio_DSPChainOverlay::RecreateDialog( )
{
	if( !s_pxThis )
	{
		return;
	}

	const int iSelectedEffect = s_pxThis->m_iSelectedEffectInChain;
	Asura_Audio_DSPChain* pxChain = s_pxThis->m_pxActiveChain;

	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_DSP_CHAIN );

	s_pxThis = new Tools_Audio_DSPChainOverlay( &Tools_Audio_DSP::s_xChainOverlayPos, pxChain, iSelectedEffect );

	Asura_Assert( s_pxThis, "Failed to allocate DSP overlay." );
	if( !s_pxThis ) 
	{
		return;
	}

	Asura_Overlay_System::SubmitNewOverlay( s_pxThis );
}

void Tools_Audio_DSPChainOverlay::SelectEffect( const int iEffect )
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	s_pxThis->m_iSelectedEffectInChain = iEffect;

	Asura_Overlay_System::SetCallbackFunctionToCall( RecreateDialog );
}


void Tools_Audio_DSPChainOverlay::AddEffect()
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	if( !s_pxThis->m_pxActiveChain ) return;

	Tools_Audio_DSPEffectPicker* pxPicker = new Tools_Audio_DSPEffectPicker();

	if( !pxPicker )
	{
		return;
	}

	pxPicker->ShowOverlay();

	if( pxPicker->GetSelectedEffectType() == -1 )
	{
		return;
	}

	s_pxThis->m_pxActiveChain->AddEffect( static_cast< Asura_Audio_DSPType >( pxPicker->GetSelectedEffectType() ) );

	// Now any special processing to do when adding effects...

	if( pxPicker->GetSelectedEffectType() == ASURA_DSP_SIMPLE_FILTER )
	{
		// For filters we should calc the coefficients immediately, we do it here because
		// it only needs to be done on creation or changing of the parameters
		Asura_DSP_Params_SimpleFilter* pxEffect = reinterpret_cast< Asura_DSP_Params_SimpleFilter* >( s_pxThis->m_pxActiveChain->m_ppxEffects[ s_pxThis->m_pxActiveChain->m_uNumEffects - 1 ] );
		pxEffect->ReCalcCoefficients();
	}

	delete pxPicker;

	Tools_Audio_DSP::Modified() = true;

	RecreateDialog();
}


void Tools_Audio_DSPChainOverlay::OnDeleteEffect( )
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	if( !s_pxThis->m_pxActiveChain ) return;

	s_pxThis->m_pxActiveChain->DeleteEffect( s_pxThis->m_iSelectedEffectInChain );
	s_pxThis->m_iSelectedEffectInChain = -1;

	Tools_Audio_DSP::Modified() = true;

	RecreateDialog();
}


void Tools_Audio_DSPChainOverlay::MoveUpChain( const int iEffectIndex )
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	if( !s_pxThis->m_pxActiveChain ) return;

	s_pxThis->m_pxActiveChain->MoveEffectUp( s_pxThis->m_iSelectedEffectInChain );

	Tools_Audio_DSP::Modified() = true;

	RecreateDialog();
}


void Tools_Audio_DSPChainOverlay::OnMoveUpChain( )
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	MoveUpChain( s_pxThis->m_iSelectedEffectInChain );
}


void Tools_Audio_DSPChainOverlay::OnMoveDownChain( )
{
	Asura_Assert( s_pxThis, "No DSP overlay exists." );
	if( !s_pxThis ) return;

	// Instead of moving this one, actually move the next one up-chain, to reuse code
	MoveUpChain( s_pxThis->m_iSelectedEffectInChain + 1 );
}


void Tools_Audio_DSPChainOverlay::PerFrameUpdate ()
{
	//Asura_Overlay::Update();

	if( m_bFilterParamsChanged )
	{
		m_bFilterParamsChanged = false;

		UpdateFilterParams();
	}
}


void Tools_Audio_DSPChainOverlay::UpdateFilterParams ( )
{
	// When one of the filter params is changed we need to recalc the coefficients since this is what we use when processing
	if( s_pxThis && s_pxThis->m_pxActiveChain )
	{
		if( s_pxThis->m_pxActiveChain->m_ppxEffects[ s_pxThis->m_iSelectedEffectInChain ]->GetType() == ASURA_DSP_SIMPLE_FILTER )
		{
			Asura_DSP_Params_SimpleFilter* pxEffect = reinterpret_cast< Asura_DSP_Params_SimpleFilter* >( s_pxThis->m_pxActiveChain->m_ppxEffects[ s_pxThis->m_iSelectedEffectInChain ] );

			pxEffect->ReCalcCoefficients();

			if( s_pxThis->m_iSavedFilterType != static_cast<int>( pxEffect->m_xParams.m_uType ) )
			{
				Asura_Overlay_System::SetCallbackFunctionToCall( RecreateDialog );
			}

			Tools_Audio_DSP::Modified() = true;
		}
	}
}

bool Tools_Audio_DSPChainOverlay::IsFilterValid ( int iIgnore )
{
	if( s_pxThis && s_pxThis->m_pxActiveChain )
	{
		if( s_pxThis->m_pxActiveChain->m_ppxEffects[ s_pxThis->m_iSelectedEffectInChain ]->GetType() == ASURA_DSP_SIMPLE_FILTER )
		{
			Asura_DSP_Params_SimpleFilter* pxEffect = reinterpret_cast< Asura_DSP_Params_SimpleFilter* >( s_pxThis->m_pxActiveChain->m_ppxEffects[ s_pxThis->m_iSelectedEffectInChain ] );

			return pxEffect->m_xParams.m_bValid;
		}
	}
	return false;
}

//___________________ Tools_Audio_DSPChainSubwindow ________________________________

Tools_Audio_DSPChainSubwindow::Tools_Audio_DSPChainSubwindow	( const Asura_Vector_2& xOffset, const float fWidth, const float fHeight ) 
																: Asura_Subwindow_Dialog ( xOffset, fWidth, fHeight, false )
{
	m_iHoveringOver = -1;
}


Tools_Audio_DSPChainSubwindow::~Tools_Audio_DSPChainSubwindow ()
{

}


void Tools_Audio_DSPChainSubwindow::Update( const Asura_Vector_2& xOrigin )
{
	Asura_Subwindow_Dialog::Update( xOrigin );

	if( !Tools_Audio_DSPChainOverlay::Instance() )
	{
		return;
	}
	if( !Tools_Audio_DSPChainOverlay::Instance()->m_pxActiveChain )
	{
		return;
	}

	float fCursorX, fCursorY;
	Asura_Cursor::GetScreenCoords(fCursorX, fCursorY);

	if( !IsPointInDialog( xOrigin, Asura_Vector_2( fCursorX, fCursorY ) ) )
	{
		return;
	}
	
	m_iHoveringOver = -1;

	fCursorX *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
	fCursorY *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	Asura_Vector_2 xActualOrigin = xOrigin + m_xOffset; 

	xActualOrigin.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xActualOrigin.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	Asura_Audio_DSPChain* pxActiveChain = Tools_Audio_DSPChainOverlay::Instance()->m_pxActiveChain;

	for( int iNode = 0; iNode < static_cast<int>( pxActiveChain->m_uNumEffects + 1 ); ++iNode )
	{
		const float fStartX = xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_STARTX + ( fDSP_CHAIN_SUBWINDOW_NODE_SPACING + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH ) * static_cast<float>( iNode );
		const float fEndX = fStartX + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH;
		const float fStartY = xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_STARTY;
		const float fEndY = fStartY + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT;


		if	(	( fCursorX > fStartX	)
			&&	( fCursorX < fEndX		)
			&&	( fCursorY > fStartY	)
			&&	( fCursorY < fEndY		) )
		{
			m_iHoveringOver = iNode;
			break;
		}
	}
	
	if( m_iHoveringOver == -1 )
	{
		return;
	}

	if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_MOUSEBUTTON1 ) )
	{
		if( m_iHoveringOver == static_cast<int>( pxActiveChain->m_uNumEffects ) )
		{
			Asura_Overlay_System::SetCallbackFunctionToCall( Tools_Audio_DSPChainOverlay::AddEffect );
			return;
		}
		else
		{
			Tools_Audio_DSPChainOverlay::SelectEffect( m_iHoveringOver );
			return;
		}
	}
}


void Tools_Audio_DSPChainSubwindow::Render( const Asura_Vector_2& xOrigin )
{
	if( !Tools_Audio_DSPChainOverlay::Instance() )
	{
		return;
	}
	if( !Tools_Audio_DSPChainOverlay::Instance()->m_pxActiveChain )
	{
		return;
	}

	Asura_Audio_DSPChain* pxActiveChain = Tools_Audio_DSPChainOverlay::Instance()->m_pxActiveChain;

	Asura_Vector_2 xActualOrigin = xOrigin + m_xOffset;

	xActualOrigin.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xActualOrigin.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	// Background
	Asura_Render::Overlay.Colour			= uDSP_CHAIN_SUBWINDOW_BACKGROUND_COLOUR | Asura_OverlayBrightness;
	Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
	Asura_Render::Overlay.SetRectXY ( xActualOrigin.x, xActualOrigin.y, GetWidth() * ASURA_TEXT_OUTPUT_FONT_WIDTH, GetHeight() * ASURA_TEXT_OUTPUT_FONT_HEIGHT );
	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();

	xActualOrigin.x += fDSP_CHAIN_SUBWINDOW_NODE_STARTX;
	xActualOrigin.y += fDSP_CHAIN_SUBWINDOW_NODE_STARTY;

	u_int uCol;

	// Render the effect blocks
	for( u_int uEffect = 0; uEffect < pxActiveChain->m_uNumEffects; ++uEffect )
	{
		// Is this selected?
		if( Tools_Audio_DSPChainOverlay::Instance()->m_iSelectedEffectInChain == static_cast<int>( uEffect ) )
		{
			// Draw a selection block
			Asura_Render::Overlay.Colour			= uDSP_EFFECT_SELECTION_COLOUR;
			Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
			Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
			Asura_Render::Overlay.SetRectXY ( xActualOrigin.x - fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER
											, xActualOrigin.y - fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER
											, fDSP_CHAIN_SUBWINDOW_NODE_WIDTH + fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER * 2.0f
											, fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT + fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER * 2.0f );
			Asura_Render::Overlay.Render();
			Asura_Render::FlushDisplay();

			// Border
			uCol = uDSP_EFFECT_BORDER_COLOUR | Asura_OverlayBrightness;

			const float fLeft = xActualOrigin.x - fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER; 
			const float fRight = xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH;
			const float fTop = xActualOrigin.y - fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER;
			const float fBottom = xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_SELECTION_BORDER + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT;

			Asura_Render::Render2DLine( fLeft, fTop, fLeft, fBottom, uCol );
			Asura_Render::Render2DLine( fRight, fTop, fRight, fBottom, uCol );
			Asura_Render::Render2DLine( fLeft, fTop, fRight, fTop, uCol );
			Asura_Render::Render2DLine( fLeft, fBottom, fRight, fBottom, uCol );
			
		}

		// Shadow
		Asura_Render::Overlay.Colour			= uDSP_EFFECT_SHADOW_COLOUR;
		Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY ( xActualOrigin.x + 3.0f, xActualOrigin.y + 3.0f, fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT );
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		// Block
		if( m_iHoveringOver == static_cast<int>( uEffect ) )
		{
			uCol = uDSP_EFFECT_BACK_HIGHLIGHTED_COLOUR | Asura_OverlayBrightness;
		}
		else
		{
			uCol = uDSP_EFFECT_BACK_COLOUR | Asura_OverlayBrightness;
		}

		Asura_Render::Overlay.Colour			= uCol;
		Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY ( xActualOrigin.x, xActualOrigin.y, fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT );
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		// Border
		uCol = uDSP_EFFECT_BORDER_COLOUR | Asura_OverlayBrightness;

		Asura_Render::Render2DLine( xActualOrigin.x, xActualOrigin.y, xActualOrigin.x, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, uCol );																			// Left
		Asura_Render::Render2DLine( xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y, xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, uCol );		// Right
		Asura_Render::Render2DLine( xActualOrigin.x, xActualOrigin.y, xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y, uCol );																			// Top
		Asura_Render::Render2DLine( xActualOrigin.x, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, uCol );		// Bottom
		
		const Asura_Audio_DSPType eDSPType = pxActiveChain->m_ppxEffects[ uEffect ]->GetType();
		// Text
		float fTextWidth = Asura_Text_Output::WidthOfString( aszTOOLS_DSP_EFFECT_NAMES[ static_cast<u_int>( eDSPType ) ], ASURA_TEXT_FORMAT_SMALLFONT );
		Asura_Text_Output::Print ( ( xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH * 0.5f ) / ASURA_TEXT_OUTPUT_FONT_WIDTH - fTextWidth * 0.5f, ( xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT * 0.15f ) / ASURA_TEXT_OUTPUT_FONT_HEIGHT, uDSP_EFFECT_DSP_TITLE_COLOUR | Asura_OverlayBrightness, ASURA_TEXT_FORMAT_SMALLFONT, aszTOOLS_DSP_EFFECT_NAMES[ static_cast<u_int>( eDSPType ) ] );

		RenderDSPIcon( xActualOrigin, eDSPType );

		if( uEffect != 0 )
		{
			// If this is not the first block, add an arrow from the previous block
			const float fArrowStartX = ( xActualOrigin.x - fDSP_CHAIN_SUBWINDOW_NODE_SPACING ) + 10.0f;
			const float fArrowEndX = xActualOrigin.x - 10.0f;
			const float fArrowY = xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT * 0.5f;

			uCol = uDSP_ARROW_COLOUR | Asura_OverlayBrightness;

			Asura_Render::Render2DLine( fArrowStartX, fArrowY , fArrowEndX, fArrowY, uCol );
			Asura_Render::Render2DLine( fArrowEndX, fArrowY , fArrowEndX - 5.0f, fArrowY + 5.0f, uCol );
			Asura_Render::Render2DLine( fArrowEndX, fArrowY , fArrowEndX - 5.0f, fArrowY - 5.0f, uCol );
		}

		xActualOrigin.x += ( fDSP_CHAIN_SUBWINDOW_NODE_SPACING + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH );
	}

	if( pxActiveChain->m_uNumEffects != ASURA_DSP_MAX_EFFECTS_PER_CHAIN )
	{
		// Render an extra block the user can hit to add an effect
		// Block
		if( m_iHoveringOver == static_cast<int>( pxActiveChain->m_uNumEffects ) )
		{
			// highlighting the add block
			uCol = uDSP_EFFECT_BACK_COLOUR | Asura_OverlayBrightness;
		}
		else
		{
			uCol = uDSP_EFFECT_ADD_BACK_COLOUR;
		}

		Asura_Render::Overlay.Colour			= uCol;
		Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY ( xActualOrigin.x, xActualOrigin.y, fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT );
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		// Plus symbol...
		const float fSymbolCentreX = xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH * 0.5f;
		const float fSymbolCentreY = xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT * 0.65f;
		const float fSymbolTotalSize = fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT * 0.5f;
		const float fHalfWidth = fDSP_EFFECT_ADD_SYMBOL_WIDTH * 0.5f;

		uCol = uDSP_EFFECT_ADD_SYMBOL_COLOUR | Asura_OverlayBrightness;

		// ...Bar down...
		Asura_Render::Overlay.Colour			= uCol;
		Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY ( fSymbolCentreX - fHalfWidth, fSymbolCentreY - fSymbolTotalSize * 0.5f, fDSP_EFFECT_ADD_SYMBOL_WIDTH, fSymbolTotalSize );
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		// ...Bar accross (1)...
		Asura_Render::Overlay.Colour			= uCol;
		Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY ( fSymbolCentreX - fSymbolTotalSize * 0.5f, fSymbolCentreY - fHalfWidth, ( fSymbolTotalSize * 0.5f ) - fHalfWidth , fDSP_EFFECT_ADD_SYMBOL_WIDTH );
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		// ...Bar accross (2)...
		Asura_Render::Overlay.Colour			= uCol;
		Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY ( fSymbolCentreX + fHalfWidth, fSymbolCentreY - fHalfWidth, ( fSymbolTotalSize * 0.5f ) - fHalfWidth , fDSP_EFFECT_ADD_SYMBOL_WIDTH );
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		// Border
		uCol = uDSP_EFFECT_ADD_BORDER_COLOUR;

		RenderDashedLine( xActualOrigin.x, xActualOrigin.y, xActualOrigin.x, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, uCol );																			// Left
		RenderDashedLine( xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y, xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, uCol );		// Right
		RenderDashedLine( xActualOrigin.x, xActualOrigin.y, xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y, uCol );																				// Top
		RenderDashedLine( xActualOrigin.x, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, xActualOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH, xActualOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT, uCol );		// Bottom

		// Text
		float fTextWidth = Asura_Text_Output::WidthOfString( "Add effect", ASURA_TEXT_FORMAT_SMALLFONT );
		Asura_Text_Output::Print ( fSymbolCentreX / ASURA_TEXT_OUTPUT_FONT_WIDTH - fTextWidth * 0.5f, ( fSymbolCentreY - ( fSymbolTotalSize * 0.5f + 25.0f ) ) / ASURA_TEXT_OUTPUT_FONT_HEIGHT, uDSP_EFFECT_ADD_TEXT_COLOUR | Asura_OverlayBrightness, ASURA_TEXT_FORMAT_SMALLFONT, "Add Effect" );
	}

	Asura_Subwindow_Dialog::Render( xOrigin );
}


void Tools_Audio_DSPChainSubwindow::RenderDSPIcon( const Asura_Vector_2& xOrigin, const Asura_Audio_DSPType eDSPType )
{
	float fLeft = xOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH * 0.25f;
	float fTop = xOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT * 0.5f;
	float fRight = xOrigin.x + fDSP_CHAIN_SUBWINDOW_NODE_WIDTH * 0.75f;
	float fBottom = xOrigin.y + fDSP_CHAIN_SUBWINDOW_NODE_HEIGHT * 0.85f;
	float fWidth = fRight - fLeft;
	float fHeight = fBottom - fTop;
	
	// Back
	Asura_Render::Overlay.Colour			= uDSP_EFFECT_ICON_BACK_COLOUR | Asura_OverlayBrightness;
	Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
	Asura_Render::Overlay.SetRectXY ( fLeft, fTop, fWidth, fHeight );
	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();

	// Border
	u_int uCol = uDSP_EFFECT_ICON_BORDER_COLOUR | Asura_OverlayBrightness;

	Asura_Render::Render2DLine( fLeft, fTop, fLeft, fBottom, uCol );
	Asura_Render::Render2DLine( fRight, fTop, fRight, fBottom, uCol );
	Asura_Render::Render2DLine( fLeft, fTop, fRight, fTop, uCol );
	Asura_Render::Render2DLine( fLeft, fBottom, fRight, fBottom, uCol );

	fLeft += fWidth * 0.1f;
	fRight -= fWidth * 0.1f;
	fTop += fHeight * 0.1f;
	fBottom -= fHeight * 0.1f;
	fWidth = fRight - fLeft;
	fHeight = fBottom - fTop;

	uCol = uDSP_EFFECT_ICON_FRONT_COLOUR | Asura_OverlayBrightness;
	u_int uColFade = uDSP_EFFECT_ICON_FRONT_COLOUR_FADE | Asura_OverlayBrightness;

	switch( eDSPType )
	{
		case ASURA_DSP_SIMPLE_FILTER:
		{
			Asura_Render::Render2DLine( fLeft, fTop, fLeft + fWidth * 0.4f, fTop, uCol );
			Asura_Render::Render2DLine( fLeft + fWidth * 0.4f, fTop, fLeft + fWidth * 0.6f, fBottom, uCol );
			Asura_Render::Render2DLine( fLeft + fWidth * 0.6f, fBottom, fRight, fBottom, uCol );
		}
		break;
		
		case ASURA_DSP_ADD_NOISE:
		{
			Asura_Render::Render2DLine( fLeft, fTop + fHeight * 0.5f, fLeft + fWidth * 0.3f, fTop + fHeight * 0.5f, uCol );

			Asura_Render::Render2DLine( fLeft + fWidth * 0.3f, fTop + fHeight * 0.5f, fLeft + fWidth * 0.4f, fTop + fHeight * 0.1f, uCol );
			Asura_Render::Render2DLine( fLeft + fWidth * 0.4f, fTop + fHeight * 0.1f, fLeft + fWidth * 0.5f, fTop + fHeight * 0.7f, uCol );
			Asura_Render::Render2DLine( fLeft + fWidth * 0.5f, fTop + fHeight * 0.7f, fLeft + fWidth * 0.6f, fTop + fHeight * 0.2f, uCol );
			Asura_Render::Render2DLine( fLeft + fWidth * 0.6f, fTop + fHeight * 0.2f, fLeft + fWidth * 0.7f, fTop + fHeight * 0.5f, uCol );
			
			Asura_Render::Render2DLine( fLeft + fWidth * 0.7f, fTop + fHeight * 0.5f, fLeft + fWidth, fTop + fHeight * 0.5f, uCol );
		}
		break;

		case ASURA_DSP_FLANGER:
		{
			// Right about here I start to realise its pretty hard to come up with a unique icon for every DSP effect...

			float fLastPos = 0.0f;
			const float fInc = 0.1f;
			for( float fPos = fInc; fPos <= 1.0f; fPos += fInc )
			{
				Asura_Render::Render2DLine	( fLeft + ( fLastPos * fWidth )
											, fTop + fHeight * 0.5f + Asura_Maths::Sine( fLastPos * 2.0f * Asura_Maths::Pi ) * fHeight * 0.5f
											, fLeft + ( fPos * fWidth )
											, fTop + fHeight * 0.5f + Asura_Maths::Sine( fPos * 2.0f * Asura_Maths::Pi ) * fHeight * 0.5f
											, uCol );

				if( fPos < 0.5f )
				{
					Asura_Render::Render2DLine	( fLeft + ( fLastPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fLastPos * 2.0f * Asura_Maths::Pi ) * fHeight * 0.5f - fHeight * 0.25f
												, fLeft + ( fPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fPos * 2.0f * Asura_Maths::Pi ) * fHeight * 0.5f - fHeight * 0.25f
												, uColFade );
				}
				else if( fPos > 0.5f )
				{
					Asura_Render::Render2DLine	( fLeft + ( fLastPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fLastPos * 2.0f * Asura_Maths::Pi ) * fHeight * 0.5f + fHeight * 0.25f
												, fLeft + ( fPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fPos * 2.0f * Asura_Maths::Pi ) * fHeight * 0.5f + fHeight * 0.25f
												, uColFade );
				}

				fLastPos = fPos;
			}
		}
		break;

		case ASURA_DSP_CHORUS:
		{
			float fLastPos = 0.0f;
			const float fInc = 0.1f;
			for( float fPos = fInc; fPos <= 1.0f; fPos += fInc )
			{
				if( fPos > fInc )
				{
					Asura_Render::Render2DLine	( fLeft + ( fLastPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fLastPos * 2.0f * Asura_Maths::Pi * 1.2f ) * fHeight * 0.5f
												, fLeft + ( fPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fPos * 2.0f * Asura_Maths::Pi * 1.2f ) * fHeight * 0.5f
												, uColFade );
				}

				if( fPos < ( 1.0f - fInc ) )
				{
					Asura_Render::Render2DLine	( fLeft + ( fLastPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fLastPos * 2.0f * Asura_Maths::Pi * 1.2f + Asura_Maths::Pi * 0.2f ) * fHeight * 0.5f
												, fLeft + ( fPos * fWidth )
												, fTop + fHeight * 0.5f + Asura_Maths::Sine( fPos * 2.0f * Asura_Maths::Pi * 1.2f + Asura_Maths::Pi * 0.2f ) * fHeight * 0.5f
												, uCol );
				}

				fLastPos = fPos;
			}
		}
		break;
	}
}


bool Tools_Audio_DSPChainSubwindow::IsPointInDialog (const Asura_Vector_2& xOrigin,const Asura_Vector_2& xPos) const
{
	if(		( xPos.x > m_xOffset.x + xOrigin.x )
		&&	( xPos.x < m_xOffset.x + xOrigin.x + GetWidth() )
		&&	( xPos.y > m_xOffset.y + xOrigin.y )
		&&	( xPos.y < m_xOffset.y + xOrigin.y + GetHeight() ) )
	{
		return true;
	}

	return false;
}

//__________________ Tools_Audio_DSPEffectPicker _______________________

Tools_Audio_DSPEffectPicker::Tools_Audio_DSPEffectPicker ( ) : Tools_Modal_Overlay( Asura_String( "Select an effect" ) )
{
	Asura_Assert( !s_pxThis, "Do not create multiple instances of Tools_Audio_DSPEffectPicker" );

	s_pxThis = this;

	m_iReturnedSelection = -1;
	m_iListSelection = -1;
}


Tools_Audio_DSPEffectPicker::~Tools_Audio_DSPEffectPicker ( )
{
	s_pxThis = 0;
}


void Tools_Audio_DSPEffectPicker::InitialiseOverlay(class Asura_Overlay* Overlay)
{
	// Add dialogs
	Asura_Vector_2 xOffset( 1.0f, 2.0f );
	Asura_Dialog* pxDialog;

	// Chain list
	m_pxEffectList = new Asura_List_Dialog( xOffset, aszTOOLS_DSP_EFFECT_NAMES, &m_iListSelection, 10, uTOOLS_DSP_NUM_EFFECT_NAMES, 0, NULL, true );
	m_pxEffectList->SetExpandWidthToFillOverlay( true );
	Overlay->AddDialog( m_pxEffectList );
	xOffset.y += m_pxEffectList->m_xMaxCoord.y;

	pxDialog = new Asura_Function_Dialog( "Select", xOffset, OnAccept );
	pxDialog->SetHelpText( "Select this effect for the chain" );
	Overlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();
	
	pxDialog = new Asura_Function_Dialog( "Cancel", xOffset, OnDecline );
	Overlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();
}


void Tools_Audio_DSPEffectPicker::Accept ( )
{
	m_iReturnedSelection = m_iListSelection;
	CloseOverlay();
}


void Tools_Audio_DSPEffectPicker::Decline( )
{
	m_iReturnedSelection = -1;
	CloseOverlay();
}

void Tools_Audio_DSPEffectPicker::OnAccept ( )	
{ 
	s_pxThis->Accept(); 
}

void Tools_Audio_DSPEffectPicker::OnDecline ( )
{ 
	s_pxThis->Decline(); 
}

//__________________ Tools_DSPChain_DropDownList_Dialog _______________________


Tools_DSPChain_DropDownList_Dialog::Tools_DSPChain_DropDownList_Dialog(		const Asura_Vector_2& xOffset
																		,	const int iNumLines
																		,	Asura_Hash_ID* const puHashID
																		,	bool* const pbChanged
																		,	const char* const szName /*= NULL*/ )
																		: Asura_Hash_DropDownList_Dialog( xOffset )
{
	SetOnlyUpdateSelectionIndexOnClick( true );
	Initialise( xOffset, iNumLines, puHashID, pbChanged, szName );
}


const char* Tools_DSPChain_DropDownList_Dialog::GetNameByIndex( int iIndex ) const
{
	const Asura_Audio_DSPChain* const pxChain = Asura_Audio_System::GetDSPChainByIndex( iIndex );
	return pxChain ? pxChain->m_xName.GetString() : NULL;
}


u_int Tools_DSPChain_DropDownList_Dialog::GetNumberOfHashValues	( )	const
{
	return Asura_Audio_System::GetNumDSPChains(); 
}


void Tools_DSPChain_DropDownList_Dialog::RefreshChains()
{
	// Changing this value causes all dialogs of this class to automatically regenerate their lists
	++s_iChainListValidID;
}




#endif // defined ASURA_USE_AUDIO_SYSTEM
