//
// Tools_Dialogue_System.h
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_DIALOGUE_SYSTEM__
#define __TOOLS_DIALOGUE_SYSTEM__

#ifdef ASURA_USE_DIALOGUE_SYSTEM

#include "Core/Asura_NameTree.h"

// Forward declarations
template <class T> class List;
template <class T> class Asura_TemplateHashTree;

class Asura_Dialogue_Line;
class Tools_Dialogue_HashToPath;
class Tools_UnicodeTabbedStringTable;

class Tools_Dialogue_System
{
	friend class Asura_Chunk_Dialogue_ToolsGlobalSettings;
	friend class Asura_Chunk_Dialogue_ToolsHashToPath;
	friend class Tools_Overlay_Dialogue;

public :
	Tools_Dialogue_System();
	~Tools_Dialogue_System();

	static void Initialise();
	static void Shutdown();

	//Update needs to be called each frame if the dialogue system overlay controls have been added to a mode's overlay
	static void Update();

	//returns true if any dialogue files are modified
	static bool IsModified();	
	//Save all dialogue files
	static void SaveAll();
	//save all files marked as modified
	static void SaveModified();

	//sets the file containing the event as modified
	static void SetModified(class Asura_Dialogue_Event* pxEvent);

	//szFilename is filename without path/extension
	static void SetModified(const char* szFilename);

	//Checks to see if the file for the dialogue event can be saved
	static bool CanSaveFile(class Asura_Dialogue_Event* pxEvent);
	static bool CanSaveFile(const char* szFilename);

	static void SetAudioTemplatesModified();

	//Used by Tools_SaveAsura to mark the sound event files as ones to be tracked
	static void SetFilesToCheckForExternalModification();

	//Used by Tools_SaveAsura to load the .dlg specific to a given level.
	static void LoadLevelSpecificFile();


	//Gets default file for new events
	static const char* GetDefaultFile() {return "Default";}

	//Gets the dummy filename used to represent the level specific dialogues
	//(The file itself will share the same name as the level)
	static const char* GetLevelSpecificFile() {return "Level";}

	static bool IsLevelSpecific(const class Asura_Dialogue_Event* pxEvent);

	//Registers a new event that has been created in this trip into tools
	//This is used to track which ones can be safely transferred between files.
	static void RegisterNewEvent(const class Asura_Dialogue_Event* pxEvent);

	static bool CanTransferEvent(const class Asura_Dialogue_Event* pxEvent,const char* szDestFile);

	//main dialogue overlay
	static void ToggleOverlay();
	static bool OverlayIsOn();
	static bool OverlayShouldBeOpen() {return s_bOverlayShouldBeOpen;}
	
	//Adds some sound event editing controls to another overlay
	static void AddOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset);
	static void CloseOverlays();
	static void CreateOverlay();


	//Create a new event with name chosen by user
	static Asura_Hash_ID CreateEvent(const char* szEventName,const List<Asura_Hash_ID>* pxSampleList = 0);
	static Asura_Hash_ID CreateLineTemplate(const char* szName = 0);

	//Deletes the specified event
	static bool DeleteEvent(Asura_Hash_ID uHash,bool bConfirm);
	static bool DeleteLineTemplate(Asura_Hash_ID uHash,bool bConfirm);

	//get the character names
	static int GetNumberOfCharacters(){return s_xCharacterNameTree.GetNumberOfNodes();}
	static const char* GetCharacterNameByIndex(int iIndex){return s_pszCharacterNames[iIndex];}
	static const char* GetCharacterNameByHash(Asura_Hash_ID uHash);

	//Returns true if the event contains any lines spoken by the character (recursing through sub-events if needed)
	static bool DoesEventContainCharacter(const class Asura_Dialogue_Event* pxEvent,Asura_Hash_ID uCharacterID);	

	static void UpdateReferencesToTemplates();

	static void SetEverHadAudioTemplates(){s_bEverHadAudioTemplates = true;}

	static u_int GetDefaultSampleRatePS2() {return s_uDefaultSampleRatePS2;}
	static u_int GetDefaultSampleRatePSP() {return s_uDefaultSampleRatePSP;}

	//Checks dialogue events for errors ,and shows a log of them if any are found
	static void Verify(bool bAutomatic);

	//minimises/un-minimises overlays when doing message editing
	static void PreMessageEditing();
	static void PostMessageEditing();

	//Used by the asura system to tell tools the current subtitle to render
	static void SetSubtitle(const Asura_UnicodeChar*);

	//Shows the current subtitle from the dialogue preview
	static void RenderSubtitle(float& fX,float& fY);

	//controls for editing the play line static message
	static void AddMessageOverlayControls_PlayLine(struct Asura_StaticMessage* pxMessage,class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset);

	//Adds the dialogue line used by the message to the export list
	static void AddFilesToExportListFromMessage(const struct Asura_StaticMessage* pxMessage);

	//Gets the sample rate to use for a dialogue line
	static int GetSampleRate(int iPlatform, const class Asura_Dialogue_Line* pxLine);

	//Get/set the most recent file selected by the user. (Used as the default for new events)
	static const char* GetMostRecentFile() {return s_xMostRecentFile.GetString();}
	static void SetMostRecentFile(const char* szName) {s_xMostRecentFile = szName;}

	//Used when events have changed to indicate that the character mappings may be out of date
	//Allows dialogue controllers/triggers to know when they need to recalculate what needs
	//to be shown in their overlays.
	static u_int GetCharacterMappingValidity(){return s_uCharacterMappingValidity;}
	static void InvalidateCharacterMappings(){s_uCharacterMappingValidity++;}


protected :
	//Gets the animation ID from the string in the dialogue line spreadsheet.
	//A value of zero should be used if there is no animation.
	//Default behaviour is just to get a hash value.
	virtual u_int ParseAnimationString(const char* szString) {return Asura_GetHashID(szString);}

	//Allows the project to modify the dialogue line name that has been read in from the table.
	//It should fill in the output name and output wav name and return true , if the initial name isn't suitable to use as is.
	virtual bool Project_GetLineAndWavName(const char* szInitialName,Asura_String& xOutputLineName,Asura_String& xOutputWavName) {return false;}

	//Used when writing out the text file in ListDialogueLinesInExports() to modify the name of the dialogue line that gets written.
	//Required by Simpsons to stick the x_'s and d_'s back onto the names
	virtual void Project_ModifyLineNameForLogFile(Asura_String& xLineName){};

	static void (*s_pfnPostLoadLevelSpecificFileCallback)(void);

	typedef void(*FindExtraDialogueHeadingsCallback)(const Tools_UnicodeTabbedStringTable& xTable, const char* szFilename);
	static FindExtraDialogueHeadingsCallback s_pfnFindExtraDialogueHeadingsCallback;

	typedef void (*PostProcessDialogueLineCallback)(const Tools_UnicodeTabbedStringTable& xTable, int iTableRow, const Asura_Dialogue_Line* pxLine);
	static PostProcessDialogueLineCallback s_pfnPostProcessDialogueLineCallback;

	//gets full filename from HashToPath tree
	static const char* GetFullFilenameFromHash(Asura_Hash_ID uHash);


private :

	static void TransferModeData(int iMode);

	static bool SaveFile(const char* szFilename);
	static bool SaveAudioTemplates();


	static void EditEvent();
	static void EditLine();
	static void EditAudioTemplate();
	static void Verify() {Verify(false);}

	//adds path and extension to get full filename
	static void GetFullFilename(const char* szFilename,Asura_String& xFullFilename);

	static void LoadAudioTemplates();

	static void LoadDialogueLines();
	static void LoadDialogueLines(const char* szFilename);
	static void LoadDialogueFiles();

	static void ReloadDialogueLines();

	static void EditMessage_SelectLine(int iData);

	static void UpdateHashToPathTree();

	static void ListDialogueLinesInExports();
	static void OutputDialogueExportDetails( Asura_StringTree& xLines, const char* const szExportFileName, FILE*& pFile );

		

	static List<Asura_String> s_xModifiedFiles;
	static bool s_bAudioTemplatesModified;

	static class Asura_HashTree s_xNewEventTree;

	static class Asura_NameTree s_xCharacterNameTree;
	static const char** s_pszCharacterNames;
	static Asura_Hash_ID* s_puCharacterHashes;

	static bool s_bEverHadAudioTemplates;

	static bool s_bGlobalSettingsChanged;
	static u_int s_uDefaultSampleRatePS2;
	static u_int s_uDefaultSampleRatePSP;

	static Asura_Vector_2 s_xOverlayPosition;
	static bool s_bOverlayShouldBeOpen;

	static List<int> s_xMinimisedOverlays;

	static Asura_String s_xMostRecentFile;

	static Asura_TemplateHashTree<Tools_Dialogue_HashToPath> s_xHashToPathTree;

	static Tools_Dialogue_System* s_pxThis;

	static u_int s_uCharacterMappingValidity;
};


//For hash tree containing mapping from line/wav hashes to filenames
class Tools_Dialogue_HashToPath
{
public :
	Asura_Hash_ID m_uHash;
	Asura_String m_xFilename;	
};



#endif //ASURA_USE_DIALOGUE_SYSTEM
#endif //__TOOLS_DIALOGUE_SYSTEM__
