//
// Tools_Mode_MetaMusic.cpp
//
// Created by Richard Rice - 20/01/04
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifndef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM

#include "Sound/Tools_Mode_MetaMusic.h"

#include "Core/Asura_UserInput.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_Menubar_System.h"
#include "Interface/Asura_Menubar.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Rendering/Asura_Text_Output.h"
#include "Sound/Asura_MetaMusic_System.h"
#include "Sound/Asura_Music.h"
#ifdef ASURA_USE_STREAMING
#include "Streaming/Asura_Streaming_System.h"
#endif
#include "Zone/Asura_Zone.h"
#include "Zone/Asura_Zone_Controller.h"

#include "Tools_Overlays.h"
#include "Entity/Tools_Entity_Object.h"
#include "Entity/Tools_Entity_MusicTrigger_Object.h"
#include "Entity/Tools_Entity_MetaMusicTrigger_Object.h"
#include "Entity/Tools_SecondaryEntityFile_System.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_IniFile.h"
#include "Objects/Tools_Object_System.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Sound/Tools_Sound_System.h"

#ifdef ASURA_USE_METAMUSICZONES
#include "Sound/Tools_MetaMusicZone_Editor.h"
#include "Zone/Tools_Zone_Editor.h"
#endif



static const char* g_apszStateStrings[] =
{
	"Intro",
	"Ambient Outro",
	"Tension Outro",
	"Action Outro",
	"Ambient",
	"Ambient to Tension",
	"Ambient to Action",
	"Tension",
	"Tension to Ambient",
	"Tension to Action",
	"Action",
	"Action to Ambient",
	"Action to Tension",
};

static const char* g_apszAbreviatedStateStrings[] =
{
	"Intro",
	"Amb Out",
	"Ten Out",
	"Act Out",
	"Ambient",
	"Amb->Ten",
	"Amb->Act",
	"Tension",
	"Ten->Amb",
	"Ten->Act",
	"Action",
	"Act->Amb",
	"Act->Ten",
};


Tools_Mode_MetaMusic* Tools_Mode_MetaMusic::s_pxThis = 0;

void Tools_Mode_MetaMusic::InitialiseMode()
{
	m_xModeOverlayPosition = Asura_Vector_2(3.0f,10.0f);
	m_xMusicPropOverlayPosition = Asura_Vector_2(26.0f,10.0f);
	m_iMetaMusicNameArraySize = 0;
	m_ppszMetaMusicNames = 0;
	m_pxMetaMusicNameTree = 0;
	m_iMusicIndex = -1;
	m_iMusicListIndex = 0;
	m_bMusicListIndexChanged = true;
	m_pxCurrentTrack = 0;
	m_bModified = false;

	m_iFilenameArraySize = 0;
	m_ppszFilenames = 0;
	m_iFilenameListSel = 0;
	m_iFilenameListOffset = 0;

	m_uSelectedTab = 0;

	m_fFileVolume = 1.0f;
	m_bFileVolumeModified = false;
	#ifdef ASURA_USE_STREAMING
	m_bEditGlobalTracks = false;
	m_bEditGlobalTracksChanged = false;
	#endif

	//create the menubar
	m_iMenuHandle = Asura_Menubar_System::CreateMenuSection();
	Asura_Menubar_Title* pxTitle;

	pxTitle = Asura_Menubar_System::AddTitle(m_iMenuHandle,"Meta Music Mode");
	pxTitle->AddItem("Entity Files",Tools_SecondaryEntityFile_System::ToggleOverlay,Tools_SecondaryEntityFile_System::OverlayIsOn);


	m_pxHandler = &Asura_MetaMusic_System::GetCurrentHandler();

	//load values from .dat file
	TransferModeData(TOOLSMODEDATAFILE_LOAD);

	CreateModeOverlay();
	Tools_SecondaryEntityFile_System::CreateOverlay();


	//initialise object system
	Tools_Object_System::EnableAllControls(true);
	//create all the static objects
	Tools_Entity_Object::CreateAllObjects();
	//only allow selection of music triggers and similar
	Tools_Object_System::SetSelectionFilter(TOT_Entity_MusicTrigger);
	Tools_Object_System::SetSelectionFilter(TOT_Entity_MetaMusicTrigger);
	Tools_Object_System::SetSelectionFilter(TOT_VolumeEdit);

	
	#ifdef ASURA_USE_METAMUSICZONES
	m_pxZoneEditor = AllocateZoneEditor(Asura_MetaMusic_System::GetZoneController());
	Asura_Assert(m_pxZoneEditor,"Zoned editor not allocated");
	#endif

	//warn if sound is muted
	if(Tools_Sound_System::IsMusicMixerMuted())
	{
		ShowMessageOverlay("Warning.\nThe music mixer is muted.\n");		
	}

	Tools_SecondaryEntityFile_System::SelectInitialEntityFile();


}

void Tools_Mode_MetaMusic::ShutdownMode()
{
	DeleteMusicNameList();
	DeleteFilenameList();

	//save values to .dat file
	TransferModeData(TOOLSMODEDATAFILE_SAVE);

	Asura_MetaMusic_System::StopTrack(true);

	//clearup object system
	Tools_Object_System::DeleteAllModeSpecific();
	Tools_Object_System::ClearSelectionFilter();
	Tools_Object_System::EnableAllControls(false);

	//shutdown the menu
	if(m_iMenuHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Menubar_System::DestroyMenuSection(m_iMenuHandle);
		m_iMenuHandle = ASURA_INVALID_HANDLE;
	}

	//close overlays
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MODE);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MUSICPROPERTIES);
	Tools_SecondaryEntityFile_System::CloseOverlay();

	#ifdef ASURA_USE_METAMUSICZONES
	delete m_pxZoneEditor;
	#endif
}

void Tools_Mode_MetaMusic::QuickSave(void)
{
	Tools_SaveAsura::SaveSounds();
	if(Tools_SaveAsura::IsEntityModified()) Tools_SaveAsura::SaveEntities();
	#ifdef ASURA_USE_STREAMING
	Tools_SaveAsura::SaveGlobalMetaMusic();
	#endif
}

bool Tools_Mode_MetaMusic::IsModified()
{
	#ifdef ASURA_USE_STREAMING
	if(Tools_SaveAsura::IsGlobalMetaMusicModified()) return true;
	#endif
	return Tools_SaveAsura::IsSoundsModified() || Tools_SaveAsura::IsEntityModified();
}

void Tools_Mode_MetaMusic::TransferModeData(int mode)
{
	//open the file
	Tools_ModeDataFile file("MetaMusicMode.dat",mode);

	//transfer the data
	//(none of these should be removed , and new data should go on the end, otherwise
	//the files will get screwed up)
	file.Transfer(m_xModeOverlayPosition);
	file.Transfer(m_xMusicPropOverlayPosition);
}

void Tools_Mode_MetaMusic::Update()
{
	#ifdef ASURA_USE_STREAMING
	//check for changing between editing the level and global tracks
	if(m_bEditGlobalTracksChanged)
	{
		m_bEditGlobalTracksChanged = false;

		//change the handler
		if(m_bEditGlobalTracks)
		{
			m_pxHandler = GetGlobalHandler();
		}
		else
		{
			m_pxHandler = &Asura_MetaMusic_System::GetCurrentHandler();
		}

		//change the selection
		m_iMusicListIndex = 0;
		UpdateMusicIndexFromListIndex();

		//update the overlays
		CreateModeOverlay();
		UpdateMusicPropertiesOverlay();

	}
	#endif //ASURA_USE_STREAMING

	//check for change in selection of meta music piece
	if(m_pxHandler->GetNumberOfTracks())
	{
		if(m_bMusicListIndexChanged)
		{
			m_bMusicListIndexChanged = false;
			UpdateMusicIndexFromListIndex();
			UpdateMusicPropertiesOverlay();
		}
	}

	if(m_bFileVolumeModified)
	{
		//copy the volume into the setting for the current file
		m_bFileVolumeModified = false;
		if(m_pxCurrentTrack)
		{
			int iFile = m_iFilenameListSel + m_iFilenameListOffset;
			if(iFile>=0 && iFile<m_pxCurrentTrack->m_iNumberOfFiles)
			{
				m_pxCurrentTrack->m_pxFiles[iFile].m_fVolume = m_fFileVolume;
				m_bModified = true;
			}
		}
	}
	else
	{
		//update the volume setting on the overlay from the selected file
		if(m_pxCurrentTrack)
		{
			int iFile = m_iFilenameListSel + m_iFilenameListOffset;
			if(iFile>=0 && iFile<m_pxCurrentTrack->m_iNumberOfFiles)
			{
				m_fFileVolume = m_pxCurrentTrack->m_pxFiles[iFile].m_fVolume;
			}
		}
	}

	HandleKeyboard();

	#ifdef ASURA_USE_METAMUSICZONES
	if(m_pxZoneEditor) m_pxZoneEditor->Update();
	#endif

	if(m_bModified)
	{
		
		if(m_pxCurrentTrack)
		{
			//Update intro loop state
			switch(m_iIntroLoopIndex)
			{
				case 2 :
					m_pxCurrentTrack->m_eIntroLoopState = ASURA_METAMUSIC_ACTION;
					break;
				case 1 :
					m_pxCurrentTrack->m_eIntroLoopState = ASURA_METAMUSIC_TENSION;
					break;
				default :
					m_pxCurrentTrack->m_eIntroLoopState = ASURA_METAMUSIC_AMBIENT;
					break;
			}
		}
		
		#ifdef ASURA_USE_STREAMING
		if(m_bEditGlobalTracks)
		{
			//flag the global file as modified
			Tools_SaveAsura::SetGlobalMetaMusicModified();
		}
		else
		#endif
		{
			Tools_SaveAsura::SetSoundsModified();
		}
		m_bModified = false;
	}

	Tools_SecondaryEntityFile_System::Update();
}

void Tools_Mode_MetaMusic::RenderWorld()
{
	#ifdef ASURA_USE_METAMUSICZONES
	if(m_pxZoneEditor) m_pxZoneEditor->RenderWorld();
	#endif
}

void Tools_Mode_MetaMusic::RenderUI()
{
	if(!Tools_Object_System::SelectedObject)
	{
		float fPrintX = 1.0f;
		float fPrintY = 4.0f;

		
		Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,0,"Timer : %.3f",Asura_MetaMusic_System::s_fIdleTimer);

		Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,0,"Position : %.1f / %.1f",Asura_Music_System::GetCurrentTrackPosition(),Asura_Music_System::GetCurrentTrackLength());

		Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,0,"Beats : %.1f",Asura_MetaMusic_System::GetBeatTimer());

		Asura_MetaMusicState eState = Asura_MetaMusic_System::s_eCurrentState;
		Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,0,"Current : %s",(eState >= 0) ? g_apszStateStrings[eState] : "None");


		for(int i=0;i<Asura_MetaMusic_System::s_iQueueLength;i++)
		{
			Asura_MetaMusicState eState = Asura_MetaMusic_System::GetFromQueue(i)->m_eState;
			Asura_Text_Output::Print(fPrintX,fPrintY++,0xffffffff,0,"Queue   : %s",(eState >= 0) ? g_apszStateStrings[eState] : "None");
		}
	}
	#ifdef ASURA_USE_METAMUSICZONES
	if(m_pxZoneEditor) m_pxZoneEditor->RenderUI();
	#endif

}

void Tools_Mode_MetaMusic::HandleKeyboard()
{
	if(Asura_Overlay_System::OverlaysHaveKeyboardFocus()) return;
	if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT)) return;

	if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
	{
		if(Asura_UserInput::DebouncedKeyPressed('a'))
		{
			AlterState();
		}
	}
	if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
	{
	}
	else
	{
	}
}


void Tools_Mode_MetaMusic::CreateModeOverlay()
{
	//delete old overlay if it exists
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MODE);

	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Meta Music Mode",false,&m_xModeOverlayPosition);

	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);

	#ifdef ASURA_USE_STREAMING
	pxDialog = new Asura_Tickbox_Dialog("Edit Global Tracks",xOffset,&m_bEditGlobalTracks,&m_bEditGlobalTracksChanged);
	pxDialog->SetHelpText("Causes the tracks in the global file to be edited.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();
	#endif

	
	pxDialog = new Asura_Function_Dialog("Create",xOffset,_CreateMusic);
	pxDialog->SetHelpText("Create a new meta music track.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.x += 12.0f;

	pxDialog = new Asura_Function_Dialog("Delete",xOffset,_DeleteMusic);
	pxDialog->SetHelpText("Delete the selected meta music track.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.x = 1.0f;
	xOffset.y+=Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Dialog("Music tracks :",xOffset);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::SmallSpacing();
	
	UpdateMusicNameList();
	int iNumberOfLines = Asura_Maths::Min(15,m_iMetaMusicNameArraySize);

	pxDialog = new Asura_List_Dialog(xOffset,m_ppszMetaMusicNames,&m_iMusicListIndex,iNumberOfLines,m_iMetaMusicNameArraySize,NULL,&m_bMusicListIndexChanged);
	pxDialog->SetHelpText("Select music track to edit/play");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y = pxDialog->m_xMaxCoord.y + 1.0f;

	

	float fY = xOffset.y;
	pxDialog = new Asura_Function_Dialog("Play",xOffset,_Play);
	pxDialog->SetHelpText("Start playing the selected track");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	

	pxDialog = new Asura_Function_Dialog("Stop",xOffset,Stop);
	pxDialog->SetHelpText("Stop playing meta music track");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	
	xOffset.y = fY;
	xOffset.x += 12.0f;
	pxDialog = new Asura_Function_Dialog("Ambient",xOffset,ToAmbient);
	pxDialog->SetHelpText("Change to ambient state (if playing a track).");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	

	pxDialog = new Asura_Function_Dialog("Tension",xOffset,ToTension);
	pxDialog->SetHelpText("Change to tension state (if playing a track).");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	
	pxDialog = new Asura_Function_Dialog("Action",xOffset,ToAction);
	pxDialog->SetHelpText("Change to change state (if playing a track).");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();
	xOffset.x = 1.0f;

	pxDialog = new Asura_Function_Dialog("Import",xOffset,_ImportTrack);
	pxDialog->SetHelpText("Imports a meta music track from a text file.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	

	pxDialog = new Asura_Function_Dialog("Export",xOffset,_ExportTrack);
	pxDialog->SetHelpText("Exports the selected meta music track to a text file.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();


	pxDialog = new Asura_Function_Dialog("Meta Music Trigger", xOffset, _CreateMetaMusicTrigger);
	pxDialog->SetHelpText("Create a meta music trigger");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	#ifdef ASURA_USE_METAMUSICZONES
	xOffset.y += Asura_Dialog::LargeSpacing()-Asura_Dialog::DefaultSpacing();
	pxDialog = new Asura_Function_Dialog("Edit Zones",xOffset,_StartZoneEditing);
	pxDialog->SetHelpText("Creates the overlays for editing the meta music zones");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	#endif


	pxOverlay->m_uID = TOOLS_OVERLAY_MODE;
	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}

void Tools_Mode_MetaMusic::UpdateMusicNameList()
{
	DeleteMusicNameList();


	if(m_pxHandler->GetNumberOfTracks())
	{
		m_pxMetaMusicNameTree = new Asura_StringTree;
		for(int i=0;i<m_pxHandler->GetNumberOfTracks();i++)
		{
			const Asura_MetaMusic* pxTrack = m_pxHandler->GetTrackData(i);
			const char*	szName = pxTrack->m_xName.GetString();

			int iJunkIndex;
			m_pxMetaMusicNameTree->FindNodeIndexByName(szName,true,iJunkIndex);
		}

		m_iMetaMusicNameArraySize = m_pxMetaMusicNameTree->GetNumberOfNodes();
		m_ppszMetaMusicNames = new const char*[m_iMetaMusicNameArraySize];

		m_pxMetaMusicNameTree->MakeSortedNameList(m_ppszMetaMusicNames);

	}
	else
	{
		m_iMetaMusicNameArraySize = 1;
		m_ppszMetaMusicNames = new const char*[m_iMetaMusicNameArraySize];
		m_ppszMetaMusicNames[0] = "*None*";
	}
	if(m_iMusicListIndex >= m_iMetaMusicNameArraySize)
	{
		m_iMusicListIndex = m_iMetaMusicNameArraySize-1;
	}
}

void Tools_Mode_MetaMusic::DeleteMusicNameList()
{
	delete m_pxMetaMusicNameTree;
	m_pxMetaMusicNameTree = 0;
	delete [] m_ppszMetaMusicNames;
	m_ppszMetaMusicNames = 0;
	m_iMetaMusicNameArraySize = 0;
}

void Tools_Mode_MetaMusic::UpdateMusicIndexFromListIndex()
{
 	m_iMusicIndex = -1;
	if(m_pxHandler->GetNumberOfTracks() && m_ppszMetaMusicNames)
	{
		const char*	szSearchName = m_ppszMetaMusicNames[m_iMusicListIndex];

		//search for a track with the same name
		for(int i=0;i<m_pxHandler->GetNumberOfTracks();i++)
		{
			const Asura_MetaMusic* pxTrack = m_pxHandler->GetTrackData(i);
			const char*	szName = pxTrack->m_xName.GetString();

			if(!stricmp(szSearchName,szName))
			{
				m_iMusicIndex = i;
				return;
			}
		}
	}
}

void Tools_Mode_MetaMusic::UpdateListIndexFromMusicIndex()
{
	m_iMusicListIndex = 0;
	
	if((m_iMusicIndex>=0) && (m_iMusicIndex < m_pxHandler->GetNumberOfTracks()) && m_ppszMetaMusicNames)
	{
		const Asura_MetaMusic* pxTrack = m_pxHandler->GetTrackData(m_iMusicIndex);
		const char*	szSearchName = pxTrack->m_xName.GetString();

		//search for the name in the sorted list
		for(int i=0;i<m_iMetaMusicNameArraySize;i++)
		{
			if(!stricmp(szSearchName,m_ppszMetaMusicNames[i]))
			{
				m_iMusicListIndex = i;
				return;
			}
		}
	}

}

void Tools_Mode_MetaMusic::UpdateFilenameList()
{
	DeleteFilenameList();
	if(!m_pxCurrentTrack) return;

	if(m_pxCurrentTrack->m_iNumberOfFiles)
	{
		m_iFilenameArraySize = m_pxCurrentTrack->m_iNumberOfFiles;
		m_ppszFilenames = new char*[m_iFilenameArraySize];

		for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
		{
			for(int j=m_pxCurrentTrack->m_axStates[i].m_iFirstIndex;j<m_pxCurrentTrack->m_axStates[i].m_iFirstIndex+m_pxCurrentTrack->m_axStates[i].m_iNumberOfFiles;j++)
			{
				//generate string for filename list
				//go for filename without path/extension , followed by state type
				Asura_String xName;
				m_pxCurrentTrack->m_pxFiles[j].m_xFilename.SplitFilename(0,&xName,0);

				m_ppszFilenames[j] = new char[100];
				Asura_PrintfToBfr(m_ppszFilenames[j],"%s  (%s)",100,true,xName.GetString(),g_apszAbreviatedStateStrings[i]);
			}
		}
	}
	else
	{
		m_iFilenameArraySize = 1;
		m_ppszFilenames = new char*[m_iFilenameArraySize];

		m_ppszFilenames[0] = new char[10];
		strcpy(m_ppszFilenames[0],"*None*");
	}


}

void Tools_Mode_MetaMusic::DeleteFilenameList()
{
	for(int i=0;i<m_iFilenameArraySize;i++)
	{
		delete [] m_ppszFilenames[i];
	}
	delete [] m_ppszFilenames;
	m_ppszFilenames = 0;
	m_iFilenameArraySize = 0;
}

void Tools_Mode_MetaMusic::UpdateMusicPropertiesOverlay()
{
	//delete the old overlay (if any)
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MUSICPROPERTIES);
	if(m_iMusicIndex<0 || m_iMusicIndex >= m_pxHandler->GetNumberOfTracks())
	{
		m_pxCurrentTrack = 0;
		return;
	}

	//and create the new one
	m_pxCurrentTrack = m_pxHandler->GetTrackData(m_iMusicIndex);

	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Music Properties",false,&m_xMusicPropOverlayPosition);

	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);

	char szText[200];
	Asura_PrintfToBfr(szText,"Name : %s",200,true,m_pxCurrentTrack->m_xName.GetString());
	pxDialog = new Asura_Function_Dialog(szText,xOffset,_Rename);
	pxDialog->SetHelpText("Rename this meta music track.\nNote that existing meta music triggers won't be able to find this track , if it is renamed.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Volume",xOffset,&m_pxCurrentTrack->m_fVolume,0,1.5f,0.01f,"%.3f",&m_bModified);
	pxDialog->SetHelpText("Global volume setting for this meta music track.\nNote that 1.33 is effectively maximum , since all volumes are scaled by 0.75 and capped at 1.0 at the end.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	static char* s_apszTabs[2] = {"Times","Files"};
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset,s_apszTabs,2);
	pxOverlay->AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	float fTabsStartY = xOffset.y;
	float fTabsEndY = xOffset.y;

	//Times tab
	pxTabsDialog->SelectTab(0);

	pxDialog = new Asura_Numerical_Dialog("Duration",xOffset,&m_pxCurrentTrack->m_fDuration,0,999,1.0f,"%.0f",&m_bModified);
	pxDialog->SetHelpText("If the duration is set to a non-zero value , then the music will automatically stop if there are no state changes in that period.");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	//intro state
	static const char* s_apszLoopStrings[] = {"Ambient","Tension","Action"};
	switch(m_pxCurrentTrack->m_eIntroLoopState)
	{
		case ASURA_METAMUSIC_ACTION :
			m_iIntroLoopIndex = 2;
			break;
		case ASURA_METAMUSIC_TENSION :
			m_iIntroLoopIndex = 1;
			break;
		default :
			m_iIntroLoopIndex = 0;
			break;
	}


	pxDialog = new Asura_DropDown_List_Dialog(xOffset,s_apszLoopStrings,&m_iIntroLoopIndex,3,3,NULL,&m_bModified,"Intro Type");
	pxDialog->SetHelpText("Determines whether the intro state is counted as ambient , tension or action.\nAlso determines the default looping state to use once the intro has finished.");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();


	float fBeatCol0 = 1.0f;
	float fBeatCol1 = 6.0f;
	float fBeatCol2 = 18.0f;

	xOffset.x = fBeatCol1+1.0f;
	pxDialog = new Asura_Dialog("Beats Per Minute",xOffset);
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol2+1.0f;
	pxDialog = new Asura_Dialog("Loop Beats",xOffset);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();


	//ambient beats
	xOffset.x = fBeatCol0;
	pxDialog = new Asura_Dialog("Ambient",xOffset);
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol1;
	pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_fAmbientBeatsPerMinute,0,999,1.0f,"%.2f",&m_bModified);
	pxDialog->SetHelpText("Together with Loop Beats , this determines the points at which transitions between states are allowed");
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol2;
	pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_fAmbientLoopBeats,0,999,1.0f,"%.0f",&m_bModified);
	pxDialog->SetHelpText("Together with Beats Per Minute , this determines the points at which transitions between states are allowed");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();


	//tension beats
	xOffset.x = fBeatCol0;
	pxDialog = new Asura_Dialog("Tension",xOffset);
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol1;
	pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_fTensionBeatsPerMinute,0,999,1.0f,"%.2f",&m_bModified);
	pxDialog->SetHelpText("Together with Loop Beats , this determines the points at which transitions between states are allowed");
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol2;
	pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_fTensionLoopBeats,0,999,1.0f,"%.0f",&m_bModified);
	pxDialog->SetHelpText("Together with Beats Per Minute , this determines the points at which transitions between states are allowed");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();


	//action beats
	xOffset.x = fBeatCol0;
	pxDialog = new Asura_Dialog("Action",xOffset);
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol1;
	pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_fActionBeatsPerMinute,0,999,1.0f,"%.2f",&m_bModified);
	pxDialog->SetHelpText("Together with Loop Beats , this determines the points at which transitions between states are allowed");
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fBeatCol2;
	pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_fActionLoopBeats,0,999,1.0f,"%.0f",&m_bModified);
	pxDialog->SetHelpText("Together with Beats Per Minute , this determines the points at which transitions between states are allowed");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();


	
	//Show table of fade times
	float fCol0 = 1.0f;
	float fCol1 = 8.0f;
	float fCol2 = 18.0f;
	float fCol3 = 28.0f;

	//do headings
	xOffset.x = fCol1+2.0f;
	pxDialog = new Asura_Dialog("Fade In",xOffset);
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fCol2+1.0f;
	pxDialog = new Asura_Dialog("Crossfade",xOffset);
	pxTabsDialog->AddDialog(pxDialog);

	xOffset.x = fCol3+2.0f;
	pxDialog = new Asura_Dialog("Fade Out",xOffset);
	pxTabsDialog->AddDialog(pxDialog);


	
	//ordering of states for fade table
	Asura_MetaMusicState aeFadeStates[] =
	{
		ASURA_METAMUSIC_INTRO,
		ASURA_METAMUSIC_AMBIENT,
		ASURA_METAMUSIC_AMBIENT_TO_TENSION,
		ASURA_METAMUSIC_AMBIENT_TO_ACTION,
		ASURA_METAMUSIC_AMBIENTOUTRO,
		ASURA_METAMUSIC_TENSION,
		ASURA_METAMUSIC_TENSION_TO_AMBIENT,
		ASURA_METAMUSIC_TENSION_TO_ACTION,
		ASURA_METAMUSIC_TENSIONOUTRO,
		ASURA_METAMUSIC_ACTION,
		ASURA_METAMUSIC_ACTION_TO_AMBIENT,
		ASURA_METAMUSIC_ACTION_TO_TENSION,
		ASURA_METAMUSIC_ACTIONOUTRO,
	};
	
	
	for(int i=0;i<sizeof(aeFadeStates)/sizeof(Asura_MetaMusicState);i++)
	{
		Asura_MetaMusicState eState = aeFadeStates[i];
		

		xOffset.y += Asura_Dialog::DefaultSpacing();

		//state heading
		xOffset.x = fCol0;
		pxDialog = new Asura_Dialog(g_apszAbreviatedStateStrings[eState],xOffset);
		pxDialog->SetHelpText(g_apszStateStrings[eState]);
		pxTabsDialog->AddDialog(pxDialog);

		//fade in time
		if(!Asura_MetaMusic_System::IsLoopingState(eState))
		{
			xOffset.x = fCol1;
			pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_axStates[eState].m_fFadeInTime,0,10.0f,0.1f,"%.1f",&m_bModified);
			pxDialog->SetHelpText("Fade time when fading into this state from the middle of the previous track.");
			pxDialog->SetIsEnabledFunction(IsEnabled_FadeIn,(int)eState);
			pxTabsDialog->AddDialog(pxDialog);
		}

		//crossfade in time
		if(Asura_MetaMusic_System::IsTransitionState(eState) || Asura_MetaMusic_System::IsLoopingState(eState))
		{
			xOffset.x = fCol2;
			pxDialog = new Asura_Numerical_Dialog("",xOffset, &m_pxCurrentTrack->m_axStates[eState].m_fFadeInTime,0,10.0f,0.1f,"%.1f",&m_bModified);
			pxDialog->SetHelpText("Fade time when crossfading between two looping states.\nThis is only valid if no transition pieces exist for this state change.");
			pxDialog->SetIsEnabledFunction(IsEnabled_Crossfade,(int)eState);
			pxTabsDialog->AddDialog(pxDialog);
		}


		//fade out time
		if(!Asura_MetaMusic_System::IsLoopingState(eState))
		{
			xOffset.x = fCol3;
			pxDialog = new Asura_Numerical_Dialog("",xOffset,&m_pxCurrentTrack->m_axStates[eState].m_fFadeOutTime,0,10.0f,0.1f,"%.1f",&m_bModified);
			pxDialog->SetHelpText("Fade time when fading from the end of a track in this state into another.");
			pxDialog->SetIsEnabledFunction(IsEnabled_FadeOut,(int)eState);
			pxTabsDialog->AddDialog(pxDialog);
		}
	}
	xOffset.x = 1.0f;
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//Files tab
	pxTabsDialog->SelectTab(1);
	xOffset.y = fTabsStartY;

	pxDialog = new Asura_Function_Dialog("Add File",xOffset,_AddFile);
	pxDialog->SetHelpText("Add a new bik file to this meta music track.");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.x+=12.0f;

	pxDialog = new Asura_Function_Dialog("Remove File",xOffset,_RemoveFile);
	pxDialog->SetHelpText("Remove the selected file from this meta music track.");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.x=1.0f;
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	if(m_pxCurrentTrack->m_iNumberOfFiles)
	{
		pxDialog = new Asura_Function_Dialog("Alter State [Ctrl+A]",xOffset,_AlterState);
		pxDialog->SetHelpText("Moves the currently selected file from its current state to another.");
		pxTabsDialog->AddDialog(pxDialog);
		xOffset.y += Asura_Dialog::DefaultSpacing();
	
		pxDialog = new Asura_Numerical_Dialog("File Volume",xOffset,&m_fFileVolume,0,1.5f,0.01f,"%.3f",&m_bFileVolumeModified);
		pxDialog->SetHelpText("Volume setting for the selected file.\nThis will be multiplied by the volume setting for the track as a whole.\nNote that 1.33 is effectively maximum , since all volumes are scaled by 0.75 and capped at 1.0 at the end.");
		pxTabsDialog->AddDialog(pxDialog);
		xOffset.y += Asura_Dialog::DefaultSpacing();

		m_bFileVolumeModified = false;
	}
	xOffset.y+=Asura_Dialog::LargeSpacing()-Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Dialog("Files :",xOffset);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::SmallSpacing();

	UpdateFilenameList();
	int iNumberOfLines = Asura_Maths::Min(12,m_iFilenameArraySize);
	Asura_List_Dialog::AdjustListSelectionAndOffset(&m_iFilenameListOffset,&m_iFilenameListSel,iNumberOfLines,m_iFilenameArraySize);

	pxDialog = new Asura_List_Dialog(xOffset,m_ppszFilenames,&m_iFilenameListOffset,&m_iFilenameListSel,iNumberOfLines,m_iFilenameArraySize,0);
	pxDialog->SetHelpText("List of files used by this meta music track");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y = pxDialog->m_xMaxCoord.y + 1.0f;


	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//End of tab
	pxTabsDialog->SetSelectedTabPointer(&m_uSelectedTab);
	xOffset.y = fTabsEndY;
	xOffset.y += 1.0f;


	#if 0 //Option to play to end of loop disabled because of issues with implementation on XBox. RAR 06/02/2005
	pxDialog = new Asura_FlagTickbox_Dialog("Finish Ambient Loop",xOffset,&m_pxCurrentTrack->m_uFlags,ASURA_METAMUSIC_PLAYAMBIENTTOENDOFLOOP,&m_bModified);
	pxDialog->SetHelpText("If ticked then any transitions from ambient will wait till the end of the loop.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Finish Tension Loop",xOffset,&m_pxCurrentTrack->m_uFlags,ASURA_METAMUSIC_PLAYTENSIONTOENDOFLOOP,&m_bModified);
	pxDialog->SetHelpText("If ticked then any transitions from tension will wait till the end of the loop.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Finish Action Loop",xOffset,&m_pxCurrentTrack->m_uFlags,ASURA_METAMUSIC_PLAYACTIONTOENDOFLOOP,&m_bModified);
	pxDialog->SetHelpText("If ticked then any transitions from action will wait till the end of the loop.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();
	#endif

	pxDialog = new Asura_FlagTickbox_Dialog("Change After Each Loop",xOffset,&m_pxCurrentTrack->m_uFlags,ASURA_METAMUSIC_CHOOSENEWFILEAFTEREACHLOOP,&m_bModified);
	pxDialog->SetHelpText("If this is ticked and there is more than one file for a looping state , then the file to play will be picked randomly after each loop , rather than just once when entering the state.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();
	

	pxOverlay->m_uID = TOOLS_OVERLAY_MUSICPROPERTIES;
	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);

}

bool Tools_Mode_MetaMusic::IsEnabled_Fades(int iState,int iType)
{
	Asura_MetaMusicState eState = (Asura_MetaMusicState)iState;

	bool bFadeIn = false;
	bool bCrossfade = false;
	bool bFadeOut = false;

	if(Asura_MetaMusic_System::IsLoopingState(eState))
	{
		//only crossfade used for looping states , and that's only if a new track is being picked after each loop
		if(m_pxCurrentTrack->m_bChooseNewFileAfterEachLoop && m_pxCurrentTrack->m_axStates[eState].m_iNumberOfFiles>1)
		{
			bCrossfade = true;
		}
	}
	else if(Asura_MetaMusic_System::IsTransitionState(eState))
	{
		if(m_pxCurrentTrack->HasState(eState))
		{
			//Has files for transition state , so fade in and fadeout are needed
			bFadeIn = true;
			bFadeOut = true;
		}
		else
		{
			//no files for transition state , so crossfade needed
			bCrossfade = true;
		}
	}
	else
	{
		if(m_pxCurrentTrack->HasState(eState))
		{
			//any other state needs fade in and fade out if it exists
			bFadeIn = true;
			bFadeOut = true;
		}
		else if(eState == ASURA_METAMUSIC_INTRO)
		{
			//always need fade in for intro
			bFadeIn = true;
		}
		else if(Asura_MetaMusic_System::IsOutroState(eState))
		{
			//always need fade out for outro
			bFadeOut = true;
		}
	}

	bool bHasState = true;
	switch(eState)
	{
		case ASURA_METAMUSIC_INTRO :
			bHasState = true;//always need intro fade times
			break;
		case ASURA_METAMUSIC_AMBIENT_TO_TENSION :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_TENSION);
			break;
		case ASURA_METAMUSIC_AMBIENT_TO_ACTION :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_ACTION);
			break;
		case ASURA_METAMUSIC_AMBIENTOUTRO :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_AMBIENT) || (m_pxCurrentTrack->m_eIntroLoopState == ASURA_METAMUSIC_AMBIENT);
			break;
		case ASURA_METAMUSIC_TENSION_TO_AMBIENT :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_TENSION) && m_pxCurrentTrack->HasState(ASURA_METAMUSIC_AMBIENT);
			break;
		case ASURA_METAMUSIC_TENSION_TO_ACTION :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_TENSION) && m_pxCurrentTrack->HasState(ASURA_METAMUSIC_ACTION);
			break;
		case ASURA_METAMUSIC_TENSIONOUTRO :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_TENSION) || (m_pxCurrentTrack->m_eIntroLoopState == ASURA_METAMUSIC_TENSION);
			break;
		case ASURA_METAMUSIC_ACTION_TO_AMBIENT :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_ACTION) && m_pxCurrentTrack->HasState(ASURA_METAMUSIC_AMBIENT);
			break;
		case ASURA_METAMUSIC_ACTION_TO_TENSION :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_ACTION) && m_pxCurrentTrack->HasState(ASURA_METAMUSIC_TENSION);
			break;
		case ASURA_METAMUSIC_ACTIONOUTRO :
			bHasState = m_pxCurrentTrack->HasState(ASURA_METAMUSIC_ACTION) || (m_pxCurrentTrack->m_eIntroLoopState == ASURA_METAMUSIC_ACTION);
			break;

		default :
			bHasState = m_pxCurrentTrack->HasState(eState);
			break;
	}


	switch(iType)
	{
 		case 0 ://fade in
			return bHasState && bFadeIn;
			break;
		case 1 ://crossfade
			return bHasState && bCrossfade;
			break;
		case 2 ://fade out
			return bHasState && bFadeOut;
			break;

		default :
			Asura_Assert(0,"Unexpected case");
			return false;
	}
}

bool Tools_Mode_MetaMusic::IsEnabled_FadeIn(int iState)
{
	return s_pxThis->IsEnabled_Fades(iState,0);
}

bool Tools_Mode_MetaMusic::IsEnabled_Crossfade(int iState)
{
	return s_pxThis->IsEnabled_Fades(iState,1);
}

bool Tools_Mode_MetaMusic::IsEnabled_FadeOut(int iState)
{
	return s_pxThis->IsEnabled_Fades(iState,2);
}


void Tools_Mode_MetaMusic::CreateMusic()
{
	int iNumberOfMusicTracks = m_pxHandler->GetNumberOfTracks();
	Asura_MetaMusic* pxOldTracks = m_pxHandler->GetTrackData(0);

	//get a name for the new piece
	GetString_Overlay xOverlay("Enter name for new piece of music","");

	Asura_String xName;
	bool bOkay = false;
	while(!bOkay)
	{
		if(!xOverlay.DoModal()) return;
		xName = xOverlay.GetString();

		//see if the name is okay
		if(xName.IsBlank()) return;

		bOkay = true;
		Asura_Hash_ID uNameHash = Asura_GetHashID(xName.GetString());

		for(int i=0;i<iNumberOfMusicTracks;i++)
		{
			if(uNameHash == pxOldTracks[i].m_uNameHash)
			{
				ShowMessageOverlay("That name is already in use");
				bOkay = false;
				break;
			}
		}
	}
	

	int iIndex = CreateEmptyMusicTrack();
	
	//initialise the new track
	Asura_MetaMusic* pxTrack = m_pxHandler->GetTrackData(iIndex);
	pxTrack->Initialise(xName.GetString());

	//update the overlays
	CreateModeOverlay();

	m_iMusicIndex = iIndex;
	UpdateListIndexFromMusicIndex();
	m_pxCurrentTrack = pxTrack;

	UpdateMusicPropertiesOverlay();
	m_bModified = true;

}

int Tools_Mode_MetaMusic::CreateEmptyMusicTrack()
{
	int iNumberOfMusicTracks = m_pxHandler->GetNumberOfTracks();

	Asura_ResourcePushType(ASURA_RESOURCE_META_MUSIC);
	{
		Asura_MetaMusic* pxOldTracks = m_pxHandler->GetTrackData(0);
		Asura_MetaMusic* pxMusicTracks;

		//allocate an enlarge array of music tracks
		Asura_ResourceAllocateArray(pxMusicTracks, iNumberOfMusicTracks+1);
		if(iNumberOfMusicTracks)
		{
			//copy from the old array
			memcpy(pxMusicTracks,pxOldTracks,sizeof(Asura_MetaMusic)*(iNumberOfMusicTracks));
			//zero old array , so we don't delete all those pointers we have gratuitously copied
			memset(pxOldTracks,0,sizeof(Asura_MetaMusic)*(iNumberOfMusicTracks));
		}

		//replace the old array
		m_pxHandler->SetTracks(pxMusicTracks, iNumberOfMusicTracks+1);
	}
	Asura_ResourcePopType();

	return iNumberOfMusicTracks;

}


void Tools_Mode_MetaMusic::DeleteMusic()
{
	if(m_iMusicIndex<0 || m_iMusicIndex>= m_pxHandler->GetNumberOfTracks()) return;

	//confirm deletion of the track
	Asura_Assert(m_pxCurrentTrack,"");
	char szText[200];
	Asura_PrintfToBfr(szText,"Would you like to delete this meta music track : '%s'?",200,true,m_pxCurrentTrack->m_xName.GetString());	
	YesNo_Overlay xOverlay(szText);
	if(!xOverlay.DoModal()) return;

	//stop music playing , since we are about to be screwing with the indeces
	Asura_MetaMusic_System::StopTrack(true);

	int iNumTracks = m_pxHandler->GetNumberOfTracks();
	Asura_MetaMusic* pxTracks = m_pxHandler->GetTrackData(0);

	//shuffle down the tracks to replace the deleted one
	pxTracks[m_iMusicIndex].~Asura_MetaMusic();
	for(int i=m_iMusicIndex;i<iNumTracks-1;i++)
	{
		memcpy(&pxTracks[i],&pxTracks[i+1],sizeof(Asura_MetaMusic));
	}
	memset(&pxTracks[iNumTracks-1],0,sizeof(Asura_MetaMusic));

	m_pxHandler->SetTracks(pxTracks, iNumTracks-1);

	m_iMusicIndex = -1;
	m_pxCurrentTrack = 0;
	m_bMusicListIndexChanged = true;
	
	//update the overlays
	CreateModeOverlay();
	UpdateMusicPropertiesOverlay();
	m_bModified = true;

}

void Tools_Mode_MetaMusic::Rename()
{
	if(!m_pxCurrentTrack) return;

	//get a new name to use
	GetString_Overlay xOverlay("Enter new name for track",m_pxCurrentTrack->m_xName.GetString());

	Asura_String xName;
	bool bOkay = false;
	while(!bOkay)
	{
		if(!xOverlay.DoModal()) return;
		xName = xOverlay.GetString();

		//see if the name is okay
		if(xName.IsBlank()) return;

		bOkay = true;
		Asura_Hash_ID uNameHash = Asura_GetHashID(xName.GetString());

		Asura_MetaMusic* pxTracks = m_pxHandler->GetTrackData(0);
		for(int i=0;i<m_pxHandler->GetNumberOfTracks();i++)
		{
			if(&pxTracks[i] == m_pxCurrentTrack) continue;
		
			if(uNameHash == pxTracks[i].m_uNameHash)
			{
				ShowMessageOverlay("That name is already in use");
				bOkay = false;
				break;
			}
		}
	}

	Asura_Hash_ID uOldHash = m_pxCurrentTrack->m_uNameHash;

	//change the name
	m_pxCurrentTrack->m_xName = xName;
	m_pxCurrentTrack->m_uNameHash = Asura_GetHashID(xName.GetString());

	//update any metamusic triggers
	Tools_Entity_MetaMusicTrigger_Object::HandleTrackRename(uOldHash,m_pxCurrentTrack->m_uNameHash);


	//and update the overlays
	CreateModeOverlay();
	UpdateMusicPropertiesOverlay();
	m_bModified = true;
}


void Tools_Mode_MetaMusic::AddFile()
{
	Asura_Assert(m_pxCurrentTrack,"");

	//ask for a state to add the file(s) to
	ListSelection_Overlay xOverlay("Add file(s) to which state?");
	for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		xOverlay.AddItem(g_apszStateStrings[i],i);
	}
	if(!xOverlay.DoModal()) return;

	int iState = xOverlay.GetSelectionData();

	//Ask for the file(s) to add
	static Asura_String s_xDefaultPath = "Sounds";

	Tools_FileSystem_DialogInfo	xFileSelector;
	xFileSelector.DefaultExtension		= ".BIK";
	xFileSelector.DefaultFilename		= "";
	xFileSelector.DefaultPath			= s_xDefaultPath.GetString();
	xFileSelector.FileFilter			= "Bink Files\0*.BIK;\0";
	xFileSelector.WindowTitle			= "Select streaming sound";
	xFileSelector.AllowMultipleSelect	= true;

	List<Asura_String> xNameList;

	Asura_String xName = Tools_FileSystem::GetLoadFilename(xFileSelector);
	for(;!xName.IsNull();xName = Tools_FileSystem::GetNextLoadFilename())
	{
		//convert filenames to having .wav extension
		char szBuffer[MAX_PATH];
		strcpy(szBuffer,xName.GetString());
		char* pcDot = strrchr(szBuffer,'.');
		if(pcDot) strcpy(pcDot,".wav");
		
		//make sure the path is relative
		const char* pszRelativeName = Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir(szBuffer);
		if(strchr(pszRelativeName,':'))
		{
			//still appears to be an absolute path....complain!
			ShowMessageOverlay("Failed to convert '%s' to a relative path.\nPlease only select files from subdirectories of the executable's directory.",szBuffer);
			//All files will be from same directory , so if we fail one , no point in checking the rest.
			break;
		}


		xNameList.add_entry(pszRelativeName);
	}
	if(!xNameList.size()) return;

	int iAdded = xNameList.size();

	//update the default directory
	xNameList.first_entry().SplitFilename(&s_xDefaultPath,0,0);

	//get the index to insert all these files at
	int iInsertIndex = m_pxCurrentTrack->m_axStates[iState].m_iFirstIndex + m_pxCurrentTrack->m_axStates[iState].m_iNumberOfFiles;
	m_pxCurrentTrack->m_axStates[iState].m_iNumberOfFiles += iAdded;

	//update all the indeces for the states
	for(int i=iState+1;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		Asura_MetaMusic_StateDefinition* pxState = &m_pxCurrentTrack->m_axStates[i];
		pxState->m_iFirstIndex += iAdded;
	}

	Asura_ResourcePushType(ASURA_RESOURCE_META_MUSIC);
	{
		//allocate new array of filesnames , and copy
		int iNewNumberOfFiles = m_pxCurrentTrack->m_iNumberOfFiles + iAdded;
		Asura_MetaMusic_File* pxFiles;
		Asura_ResourceAllocateArray(pxFiles, iNewNumberOfFiles);

		for(int i=0;i<iNewNumberOfFiles;i++)
		{
			if(i<iInsertIndex)
			{
				pxFiles[i] = m_pxCurrentTrack->m_pxFiles[i];
			}
			else if(i<(iInsertIndex + iAdded))
			{
				pxFiles[i].m_xFilename = xNameList[i-iInsertIndex];
				pxFiles[i].m_fVolume = 1.0f;
			}
			else
			{
				pxFiles[i] = m_pxCurrentTrack->m_pxFiles[i-iAdded];
			}
		}

		//replace the array
		Asura_ResourceDeallocateArray(m_pxCurrentTrack->m_pxFiles);
		m_pxCurrentTrack->m_pxFiles = pxFiles;
		m_pxCurrentTrack->m_iNumberOfFiles = iNewNumberOfFiles;
	}
	Asura_ResourcePopType();

	//change the current selection to the first newly added file
	m_iFilenameListSel = iInsertIndex;
	m_iFilenameListOffset = 0;

	//update the overlay
	UpdateMusicPropertiesOverlay();
	m_bModified = true;

}

void Tools_Mode_MetaMusic::RemoveFile()
{
	Asura_Assert(m_pxCurrentTrack,"");

	int iIndex = m_iFilenameListSel + m_iFilenameListOffset;
	if(iIndex>=m_pxCurrentTrack->m_iNumberOfFiles) return;

	//remove from the list of filenames
	for(int i=iIndex;i<m_pxCurrentTrack->m_iNumberOfFiles-1;i++)
	{
		m_pxCurrentTrack->m_pxFiles[i] = m_pxCurrentTrack->m_pxFiles[i+1];
	}
	m_pxCurrentTrack->m_iNumberOfFiles--;

	//adjust the indeces for each state
	for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		Asura_MetaMusic_StateDefinition* pxState = &m_pxCurrentTrack->m_axStates[i];

		if(pxState->m_iFirstIndex>iIndex)
		{
			pxState->m_iFirstIndex--;
		}
		else if(pxState->m_iFirstIndex+pxState->m_iNumberOfFiles>iIndex)
		{
			pxState->m_iNumberOfFiles--;
		}
	}

	//update the overlay
	UpdateMusicPropertiesOverlay();
	m_bModified = true;
}

void Tools_Mode_MetaMusic::AlterState()
{
	if(!m_pxCurrentTrack) return;

	int iIndex = m_iFilenameListSel + m_iFilenameListOffset;
	if(iIndex>=m_pxCurrentTrack->m_iNumberOfFiles) return;

	//ask for new state for file
	ListSelection_Overlay xOverlay("Move file to which state?");
	for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		xOverlay.AddItem(g_apszStateStrings[i],i);
	}
	if(!xOverlay.DoModal()) return;

	int iNewState = xOverlay.GetSelectionData();
	int iNewIndex = -1;

	
	//adjust the indeces for each state , and determine the new index for the file
	for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		Asura_MetaMusic_StateDefinition* pxState = &m_pxCurrentTrack->m_axStates[i];

		if(pxState->m_iFirstIndex>iIndex)
		{
			//removed from a previous state
			pxState->m_iFirstIndex--;
		}
		else if(pxState->m_iFirstIndex+pxState->m_iNumberOfFiles>iIndex)
		{
			//removed from this state
			pxState->m_iNumberOfFiles--;
		}

		if(i>iNewState)
		{
			//being added back into a previous state
			pxState->m_iFirstIndex++;
		}
		else if(i==iNewState)
		{
			//being added back into this state
			iNewIndex = pxState->m_iFirstIndex + pxState->m_iNumberOfFiles;
			pxState->m_iNumberOfFiles++;
		}
	}

	//move the file in the list , and shuffle the rest up or down as needed
	if(iNewIndex>iIndex)
	{
		Asura_MetaMusic_File xTemp = m_pxCurrentTrack->m_pxFiles[iIndex];
		for(int i=iIndex;i<iNewIndex;i++)
		{
			m_pxCurrentTrack->m_pxFiles[i] = m_pxCurrentTrack->m_pxFiles[i+1];
		}
		m_pxCurrentTrack->m_pxFiles[iNewIndex] = xTemp;
	}
	else if(iNewIndex<iIndex)
	{
		Asura_MetaMusic_File xTemp = m_pxCurrentTrack->m_pxFiles[iIndex];
		for(int i=iIndex;i>iNewIndex;i--)
		{
			m_pxCurrentTrack->m_pxFiles[i] = m_pxCurrentTrack->m_pxFiles[i-1];
		}
		m_pxCurrentTrack->m_pxFiles[iNewIndex] = xTemp;
	}

	//change the list selection values so as to keep the same file selected
	m_iFilenameListSel = iNewIndex;
	m_iFilenameListOffset = 0;

	//update the overlay
	UpdateMusicPropertiesOverlay();
	m_bModified = true;
}



void Tools_Mode_MetaMusic::Play()
{
	if (m_iMusicIndex>=0)
	{
		const Asura_Hash_ID uHandlerHash = m_pxHandler->GetZoneHashID();
		const Asura_MetaMusic_TrackIndex xTrackIndex(m_iMusicIndex, uHandlerHash);

		Asura_MetaMusic_System::PlayTrackByIndex(xTrackIndex, false);
	}
}

void Tools_Mode_MetaMusic::Stop()
{
	Asura_MetaMusic_System::StopTrack(false);
}

void Tools_Mode_MetaMusic::ToAmbient()
{
	Asura_MetaMusic_System::ChangeToAmbient();
}

void Tools_Mode_MetaMusic::ToTension()
{
	Asura_MetaMusic_System::ChangeToTension();
}

void Tools_Mode_MetaMusic::ToAction()
{
	Asura_MetaMusic_System::ChangeToAction();
}

void Tools_Mode_MetaMusic::ExportTrack()
{
	if(!m_pxCurrentTrack) return;

	//get a filename to save as
	//ask for a filename
	Tools_FileSystem_DialogInfo xInfo;

	xInfo.DefaultExtension = "txt";

	xInfo.DefaultFilename = m_pxCurrentTrack->m_xName.GetString();
	xInfo.DefaultPath = ".";
	xInfo.FileFilter = "Text Files\0*.txt\0";
	xInfo.WindowTitle = "Export meta music track as ...";

	Asura_Char* pszName = Tools_FileSystem::GetSaveFilename(xInfo);
	if(!pszName) return;

	FILE* pFile = fopen(pszName,"wt");
	if(!pFile)
	{
		ShowMessageOverlay("Failed to write to %s",pszName);
		return;
	}
	
	fprintf(pFile,"#Exported meta music track file.\n\n");
	fprintf(pFile,"Name = %s\n",m_pxCurrentTrack->m_xName.GetString());
	fprintf(pFile,"Duration = %.3f\n",m_pxCurrentTrack->m_fDuration);
	fprintf(pFile,"Intro Type = %s\n",g_apszStateStrings[m_pxCurrentTrack->m_eIntroLoopState]);
	fprintf(pFile,"Ambient Beats Per Minute = %.3f\n",m_pxCurrentTrack->m_fAmbientBeatsPerMinute);
	fprintf(pFile,"Ambient Loop Beats = %.3f\n",m_pxCurrentTrack->m_fAmbientLoopBeats);
	fprintf(pFile,"Tension Beats Per Minute = %.3f\n",m_pxCurrentTrack->m_fTensionBeatsPerMinute);
	fprintf(pFile,"Tension Loop Beats = %.3f\n",m_pxCurrentTrack->m_fTensionLoopBeats);
	fprintf(pFile,"Action Beats Per Minute = %.3f\n",m_pxCurrentTrack->m_fActionBeatsPerMinute);
	fprintf(pFile,"Action Loop Beats = %.3f\n",m_pxCurrentTrack->m_fActionLoopBeats);
	fprintf(pFile,"Volume = %.3f\n",m_pxCurrentTrack->m_fVolume);
//	fprintf(pFile,"Finish Ambient Loop = %s\n",m_pxCurrentTrack->m_bPlayAmbientToEndOfLoop ? "Yes" : "No");
//	fprintf(pFile,"Finish Tension Loop = %s\n",m_pxCurrentTrack->m_bPlayTensionToEndOfLoop ? "Yes" : "No");
//	fprintf(pFile,"Finish Action Loop = %s\n",m_pxCurrentTrack->m_bPlayActionToEndOfLoop ? "Yes" : "No");
	fprintf(pFile,"Change After Each Loop = %s\n",m_pxCurrentTrack->m_bChooseNewFileAfterEachLoop ? "Yes" : "No");
	fprintf(pFile,"\n\n");

	//write out the details for each state
	for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		Asura_MetaMusic_StateDefinition* pxState = &m_pxCurrentTrack->m_axStates[i];
		fprintf(pFile,"[%s]\n",g_apszStateStrings[i]);
		
		fprintf(pFile,"Fade In Time= %.3f\n",pxState->m_fFadeInTime);
		fprintf(pFile,"Fade Out Time= %.3f\n",pxState->m_fFadeOutTime);
		
		for(int j=0;j<pxState->m_iNumberOfFiles;j++)
		{
			fprintf(pFile,"File%d = %s\n",j,m_pxCurrentTrack->m_pxFiles[j+pxState->m_iFirstIndex].m_xFilename.GetString());
			fprintf(pFile,"Volume%d = %.3f\n",j,m_pxCurrentTrack->m_pxFiles[j+pxState->m_iFirstIndex].m_fVolume);
		}
		fprintf(pFile,"\n");
	}
	fclose(pFile);

	Tools_Timed_Message_System::AddMessage(0xffffff,"%s written.",pszName);
}

void Tools_Mode_MetaMusic::ImportTrack()
{
	Tools_FileSystem_DialogInfo	xFileSelector;

	xFileSelector.DefaultExtension		= ".txt";
	xFileSelector.DefaultFilename		= "";
	xFileSelector.DefaultPath			= ".";
	xFileSelector.FileFilter			= "Text Files\0*.txt;\0";
	xFileSelector.WindowTitle			= "Import from which meta music track text file...?";
	xFileSelector.AllowMultipleSelect	= false;

	Asura_String xFilename = Tools_FileSystem::GetLoadFilename( xFileSelector );

	Tools_IniFile xIniFile(xFilename.GetString(),false);
	if(!xIniFile.LoadIniFile())
	{
		ShowMessageOverlay("Failed to load to %s",xFilename.GetString());
	}


	//get the name of the track
	Asura_String xName;
	if(!xIniFile.GetStringInCurrentSection("Name",xName) || xName.IsBlank())
	{
		ShowMessageOverlay("Failed to find 'Name' in initial section of file");
		return;
	}
	//create a temporary meta music track to read the details into
	Asura_MetaMusic xTrack;
	xTrack.Initialise(xName.GetString());

	if(!xIniFile.GetFloatInCurrentSection("Duration",xTrack.m_fDuration))
	{
		ShowMessageOverlay("Failed to find 'Duration' in initial section of file");
		return;
	}
	if(xIniFile.GetFloatInCurrentSection("Beats Per Minute",xTrack.m_fAmbientBeatsPerMinute))
	{
		xTrack.m_fActionBeatsPerMinute = xTrack.m_fTensionBeatsPerMinute = xTrack.m_fAmbientBeatsPerMinute;
	}
	else
	{
		xIniFile.GetFloatInCurrentSection("Ambient Beats Per Minute",xTrack.m_fAmbientBeatsPerMinute);
		xIniFile.GetFloatInCurrentSection("Tension Beats Per Minute",xTrack.m_fTensionBeatsPerMinute);
		xIniFile.GetFloatInCurrentSection("Action Beats Per Minute",xTrack.m_fActionBeatsPerMinute);
	}
	if(xIniFile.GetFloatInCurrentSection("Loop Beats",xTrack.m_fAmbientLoopBeats))
	{
		xTrack.m_fActionLoopBeats = xTrack.m_fTensionLoopBeats = xTrack.m_fAmbientLoopBeats;
	}
	else
	{
		xIniFile.GetFloatInCurrentSection("Ambient Loop Beats",xTrack.m_fAmbientLoopBeats);
		xIniFile.GetFloatInCurrentSection("Tension Loop Beats",xTrack.m_fTensionLoopBeats);
		xIniFile.GetFloatInCurrentSection("Action Loop Beats",xTrack.m_fActionLoopBeats);
	}
	xTrack.m_fVolume = 1.0f;
	xIniFile.GetFloatInCurrentSection("Volume",xTrack.m_fVolume);

/*
	bool bFlag = false;
	if(xIniFile.GetBoolInCurrentSection("Finish Ambient Loop",bFlag))
	{
		xTrack.m_bPlayAmbientToEndOfLoop = bFlag;
	}

	if(xIniFile.GetBoolInCurrentSection("Finish Tension Loop",bFlag))
	{
		xTrack.m_bPlayTensionToEndOfLoop = bFlag;
	}

	if(xIniFile.GetBoolInCurrentSection("Finish Action Loop",bFlag))
	{
		xTrack.m_bPlayActionToEndOfLoop = bFlag;
	}
*/

	bool bFlag = false;
	if(xIniFile.GetBoolInCurrentSection("Change After Each Loop",bFlag))
	{
		xTrack.m_bChooseNewFileAfterEachLoop = bFlag;
	}


	Asura_String xIntroType;
	if(xIniFile.GetStringInCurrentSection("Intro Type",xIntroType))
	{
		if(!xIntroType.Compare("Action"))
		{
			xTrack.m_eIntroLoopState = ASURA_METAMUSIC_ACTION;
		}
		if(!xIntroType.Compare("Tension"))
		{
			xTrack.m_eIntroLoopState = ASURA_METAMUSIC_TENSION;
		}
		if(!xIntroType.Compare("Ambient"))
		{
			xTrack.m_eIntroLoopState = ASURA_METAMUSIC_AMBIENT;
		}
	}



	//read in the fade times , and do a count of filenames
	List<Asura_MetaMusic_File> xFilenameList;
	for(int i=0;i<ASURA_METAMUSIC_NUMSTATES;i++)
	{
		bool bFoundSection = xIniFile.GetIniTextSection(g_apszStateStrings[i])!=0;

		if(!bFoundSection)
		{
			if(i == ASURA_METAMUSIC_AMBIENTOUTRO || 
				i == ASURA_METAMUSIC_TENSIONOUTRO ||
				i == ASURA_METAMUSIC_ACTIONOUTRO)
			{
				bFoundSection = xIniFile.GetIniTextSection("Outro")!=0;
			}
		}

		if(!bFoundSection)
		{
			ShowMessageOverlay("Failed to find [%s] section",g_apszStateStrings[i]);
			continue;
		}
		Asura_MetaMusic_StateDefinition* pxState = &xTrack.m_axStates[i];

		
		xIniFile.GetFloatInCurrentSection("Fade In Time",pxState->m_fFadeInTime);
		xIniFile.GetFloatInCurrentSection("Fade Out Time",pxState->m_fFadeOutTime);

		//count files
		pxState->m_iFirstIndex = xTrack.m_iNumberOfFiles;
		for(int j=0;;j++)
		{
			char szSectionName[12];
			Asura_String xFilename;
			sprintf(szSectionName,"File%d",j);
			if(xIniFile.GetStringInCurrentSection(szSectionName,xFilename))
			{
				xTrack.m_iNumberOfFiles++;
				pxState->m_iNumberOfFiles++;
				Asura_MetaMusic_File xFile;
				xFile.m_xFilename = xFilename;
				xFile.m_fVolume = 1.0f;

				//check for volume setting
				sprintf(szSectionName,"Volume%d",j);
				xIniFile.GetFloatInCurrentSection(szSectionName,xFile.m_fVolume);

				xFilenameList.add_entry(xFile);
			}
			else
			{
				break;
			}
		}
	}

	//allocate space for filename list , and copy it
	if(xTrack.m_iNumberOfFiles)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_META_MUSIC);
		{
			Asura_ResourceAllocateArray(xTrack.m_pxFiles, xTrack.m_iNumberOfFiles);
			for(int i=0;i<xTrack.m_iNumberOfFiles;i++)
			{
				xTrack.m_pxFiles[i] = xFilenameList[i];
			}
		}
		Asura_ResourcePopType();
	}

	int iIndex = -1;
	Asura_MetaMusic* pxTrack = 0;
	//see if this track already exists
	for(int i=0;i<m_pxHandler->GetNumberOfTracks();i++)
	{
		if(xTrack.m_uNameHash == m_pxHandler->GetTrackData(i)->m_uNameHash)
		{
			iIndex = i;
			pxTrack = m_pxHandler->GetTrackData(i);

			//ask whether to replace this track
			char szText[200];
			Asura_PrintfToBfr(szText,"'%s' already exists , would you like to replace it?",200,true,pxTrack->m_xName.GetString());	
			YesNo_Overlay xOverlay(szText);
			if(!xOverlay.DoModal())
			{
				return;
			}
			//delete the old track
			pxTrack->~Asura_MetaMusic();
			
		}
	}

	if(iIndex == -1)
	{
		//create a new track
		iIndex = CreateEmptyMusicTrack();
		pxTrack = m_pxHandler->GetTrackData(iIndex);
	}
	//copy over the imported track details
	memcpy(pxTrack,&xTrack,sizeof(Asura_MetaMusic));
	//zero the temporary one to stop the pointers from being freed
	memset(&xTrack,0,sizeof(Asura_MetaMusic));


	//update the overlays
	CreateModeOverlay();

	m_iMusicIndex = iIndex;
	UpdateListIndexFromMusicIndex();
	m_pxCurrentTrack = pxTrack;

	UpdateMusicPropertiesOverlay();
	m_bModified = true;

}


#ifdef ASURA_USE_METAMUSICZONES

Tools_MetaMusicZone_Editor* Tools_Mode_MetaMusic::AllocateZoneEditor(class Asura_Zone_Controller* pxController)
{
	return new Tools_MetaMusicZone_Editor(pxController);
}

void Tools_Mode_MetaMusic::StartZoneEditing()
{
	if(!m_pxZoneEditor) return;

	//close normal mode overlays
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MODE);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MUSICPROPERTIES);

	m_pxZoneEditor->StartEditing();
	
}

void Tools_Mode_MetaMusic::EndZoneEditing()
{
	if(!m_pxZoneEditor) return;

	m_pxZoneEditor->EndEditing();

	//create mode overlays
	CreateModeOverlay();
	UpdateMusicPropertiesOverlay();
}


const char* Tools_Mode_MetaMusic::GetZoneNameFromID(u_int uID)
{
	if(uID==Asura_Zone::ASURA_INVALID_ZONE_ID) return "*None*";

	const Asura_Zone* pxZone = 0;
	if(Asura_MetaMusic_System::GetZoneController())
	{
		pxZone = Asura_MetaMusic_System::GetZoneController()->GetZoneByID(uID);
	}

	return pxZone ? pxZone->GetName() : "*Unknown*";
	
}

u_int Tools_Mode_MetaMusic::SelectZone(const char* pszMessage,u_int uInitialID)
{
	if(!Asura_MetaMusic_System::GetZoneController()) return uInitialID;

	ListSelection_Overlay xOverlay(pszMessage);
	//add all the zone names
	for(int i=0;i<Asura_MetaMusic_System::GetZoneController()->GetNumberOfZones();i++)
	{
		const Asura_Zone* pxZone = Asura_MetaMusic_System::GetZoneController()->GetZoneByIndex(i);

		xOverlay.AddItem(pxZone->GetName(),pxZone->GetZoneID(),pxZone->GetZoneID() == uInitialID);
	}

	if(!xOverlay.DoModal()) return uInitialID;
	
	return (u_int) xOverlay.GetSelectionData();

}

#endif //ASURA_USE_METAMUSICZONES

#ifdef ASURA_USE_STREAMING
Asura_MetaMusic_Handler* Tools_Mode_MetaMusic::GetGlobalHandler()
{
	//The global tracks are in the common zone while in tools
	Asura_MetaMusic_Handler* pxHandler = Asura_MetaMusic_System::GetHandler(Asura_Streaming_System::GetCommonDataHash());
	Asura_Assert(pxHandler,"Failed to get the meta music handler for the common zone");
	return pxHandler;
}
#endif


void Tools_Mode_MetaMusic::CreateMetaMusicTrigger()
{
	Tools_Entity_MetaMusicTrigger_Object* pxObject = Tools_Entity_MetaMusicTrigger_Object::Create(false);
	Tools_Object_System::SelectObject(pxObject);

	//Assign the trigger to the correct entity file.
	if(pxObject)
	{
		Tools_SecondaryEntityFile_System::AddToCurrentFile(pxObject);
	}
}

#endif

//eof
