//
// Tools_SoundConversionPreview.cpp
//
// Created by Richard Rice - 11/03/2009
// 
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Sound/Tools_SoundConversionPreview.h"

#include "Core/Asura_Multithread.h"
#include "FileAccess/Asura_FileAccess.h"

#include "PC/FileSystem/Tools_FileSystem.h"
#include "PC/FileSystem/Tools_File_Wrapper.h"
#include "Misc/Tools_Platform.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Sound/Tools_SoundEvent_System.h"
#include "Sound/Conversion/Tools_WAV_Reader.h"

////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------**
** Tools_SoundConversionPreview **
**------------------------------*/

static const char* g_aszCompressionQualityStrings[] =	// This should match the TOOLS_SOUND_SAMPLE_QUALITY enum
{
	"Perfect",
	"High",
	"Medium",
	"Low",
	"Very Low"
};


Tools_SoundConversionPreview_ConvertedFile* Tools_SoundConversionPreview::s_pxConvertedFiles;

int Tools_SoundConversionPreview::s_iSoundToPlay = -1;

bool Tools_SoundConversionPreview::s_bAbortThread = false;
Asura_SyncEvent* Tools_SoundConversionPreview::s_pxThreadFinishedSignal;


void Tools_SoundConversionPreview::Initialise()
{
	s_pxConvertedFiles = new Tools_SoundConversionPreview_ConvertedFile[TOOLS_SOUNDCONVERSIONPREVIEW_NUMFILES];
	s_pxThreadFinishedSignal = new Asura_SyncEvent("SoundConversionPreviewThread", true);
}

void Tools_SoundConversionPreview::Shutdown()
{
	StopAll();
	delete [] s_pxConvertedFiles;
	s_pxConvertedFiles = 0;
	delete s_pxThreadFinishedSignal;
	s_pxThreadFinishedSignal = 0;
}

void Tools_SoundConversionPreview::Update()
{
	Asura_Assert(s_pxConvertedFiles, "Not initialised");
	if(s_iSoundToPlay != -1)
	{
		//is this sound ready?
		if(s_pxConvertedFiles[s_iSoundToPlay].IsReady())
		{
			PlaySound(&s_pxConvertedFiles[s_iSoundToPlay]);
			s_iSoundToPlay = -1;
		}
		else if(s_pxConvertedFiles[s_iSoundToPlay].IsInvalid())
		{
			if(s_pxConvertedFiles[s_iSoundToPlay].m_xError.IsNull())
			{
				Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to convert sound");
			}
			else
			{
				Tools_Timed_Message_System::AddMessage(0xff0000,"Error : %s", s_pxConvertedFiles[s_iSoundToPlay].m_xError.GetString());
			}
			s_iSoundToPlay = -1;
		}
	}
}

void Tools_SoundConversionPreview::PreviewSound(const char* szFilename, int iPlatform , int iQuality)
{
	
	Asura_Assert(s_pxConvertedFiles, "Not initialised");

	if(!szFilename && !*szFilename)
	{
		Asura_Assert(0,"Invalid filename");
		return;
	}

	//Check for an existing entry
	for(int i=0; i<TOOLS_SOUNDCONVERSIONPREVIEW_NUMFILES; i++)
	{
		if(s_pxConvertedFiles[i].Matches(szFilename, iPlatform, iQuality))
		{
			if(Tools_FileSystem::FileExists(s_pxConvertedFiles[i].m_xConvertedFilename.GetString()))
			{
				//already exists or being dealt with , so flag this to play
				SetSoundToPlay(i);
				return;
			}
			else
			{
				s_pxConvertedFiles[i].Invalidate();
				break;
			}

		}
	}

	//Need to do the conversion then.
	//Search for the oldest entry to use
	u_int uOldestStamp = s_pxConvertedFiles[0].m_uLastUsedStamp;
	int iBestIndex = 0;
	for(int i=1; i<TOOLS_SOUNDCONVERSIONPREVIEW_NUMFILES; i++)
	{
		if(s_pxConvertedFiles[i].m_uLastUsedStamp < uOldestStamp)
		{
			uOldestStamp = s_pxConvertedFiles[i].m_uLastUsedStamp;
			iBestIndex = i;
		}
	}

	//setup the sound conversion
	if(ConvertSound(iBestIndex, szFilename, iPlatform, iQuality))
	{
		//mark the sound for playing
		SetSoundToPlay(iBestIndex);	
	}
}

void Tools_SoundConversionPreview::PreviewSound(const char* szFilename)
{
	if(!szFilename && !*szFilename)
	{
		Asura_Assert(0,"Invalid filename");
		return;
	}
	
	//get the file to use
	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char acFilenameBuffer[MAX_PATH];
	Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy(szFilename,acFilenameBuffer,260);
	szFilename = acFilenameBuffer;
	#endif //ASURA_PC_ALLOW_LOAD_FROM_NETWORK

	StopSound();
	PlaySoundA(szFilename , 0, SND_ASYNC|SND_FILENAME);	
}

bool Tools_SoundConversionPreview::ConvertSound(int iIndex, const char* szFilename, int iPlatform, int iQuality)
{
	//make sure this platform is supported
	Asura_ThreadFunc pfnFunction = 0;
	switch(iPlatform)
	{
		case TOOLS_PLATFORM_PS3 :
			pfnFunction = ConvertSound_PS3;
			break;

		case TOOLS_PLATFORM_XBOX360 :
			pfnFunction = ConvertSound_XBox360;
			break;

		default :
			Asura_Assert(0, "Sound converison preview not supported for this platform");
			return false;
			break;
	}		
	
	//make sure any other conversion and playback is halted
	StopAll();

	//initialise the conversion data
	Tools_SoundConversionPreview_ConvertedFile* pxFile = &s_pxConvertedFiles[iIndex];
	pxFile->m_xRelativeFilename = szFilename;
	pxFile->m_iPlatform = iPlatform;
	pxFile->m_iQuality = iQuality;
	pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_BEING_CONVERTED;
	pxFile->m_xError.MakeNull();
	pxFile->m_uFileSize = 0;

	//get the source file to use
	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char acFilenameBuffer[MAX_PATH];
	Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy(szFilename,acFilenameBuffer,260);
	pxFile->m_xFullFilename = acFilenameBuffer;
	#else
	pxFile->m_xFullFilename = szFilename;
	#endif //ASURA_PC_ALLOW_LOAD_FROM_NETWORK


	//determine the destination file
	char acConvertedFilename[MAX_PATH];
	Asura_PrintfToBfr(acConvertedFilename, "SoundConversionPreview\\ConvertedSound%d.wav", sizeof(acConvertedFilename), true, iIndex);
	Tools_FileSystem::EnsureDirectoryExists(acConvertedFilename);
	pxFile->m_xConvertedFilename = acConvertedFilename;

	Tools_FileSystem::MakeWriteable(acConvertedFilename);
	Tools_FileSystem::DeleteFileByName(acConvertedFilename);

	//start up the conversion in a new thread
	s_bAbortThread = false;
	s_pxThreadFinishedSignal->ClearSignal();
	Asura_Thread_System::CreateThread("SoundConversionPreview", pfnFunction, pxFile);

	Tools_Timed_Message_System::AddMessage(0xffffff,"Converting %s", pxFile->m_xRelativeFilename.GetString());
	

	return true;
}


void Tools_SoundConversionPreview::StopAll()
{
	StopConversion();
	StopSound();
}

void Tools_SoundConversionPreview::StopConversion()
{
	s_bAbortThread = true;
	s_pxThreadFinishedSignal->WaitForSignal();
	s_bAbortThread = false;
}

void Tools_SoundConversionPreview::PlaySound(Tools_SoundConversionPreview_ConvertedFile* pxFile)
{
	StopSound();

	const char* szQuality = g_aszCompressionQualityStrings[pxFile->m_iQuality];
	const char* szPlatform = Tools_Platform::GetPlatformName(pxFile->m_iPlatform);
	if(pxFile->m_uFileSize >= 1024)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"%s - %s  - %.1f K", szPlatform, szQuality, pxFile->m_uFileSize/1024.0f);
	}
	else
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"%s - %s - %d bytes", szPlatform, szQuality, pxFile->m_uFileSize);
	}
	
	PlaySoundA(pxFile->m_xConvertedFilename.GetString(), 0, SND_ASYNC|SND_FILENAME);
	pxFile->SetLastUsedStamp();
}

void Tools_SoundConversionPreview::StopSound()
{
	PlaySoundA(NULL, 0, 0);
	s_iSoundToPlay = -1;
}

void Tools_SoundConversionPreview::InvalidateFiles()
{
	Asura_Assert(s_pxConvertedFiles, "Not initialised");
	

	StopConversion();
	s_iSoundToPlay = -1;

	for(int i=0; i<TOOLS_SOUNDCONVERSIONPREVIEW_NUMFILES ; i++)
	{
		s_pxConvertedFiles[i].Invalidate();
	}
}

void Tools_SoundConversionPreview::SetSoundToPlay(int iIndex)
{
	s_iSoundToPlay = iIndex;
}

void Tools_SoundConversionPreview::ConvertSound_PS3(const void* pData)
{
	Tools_SoundConversionPreview_ConvertedFile* pxFile = (Tools_SoundConversionPreview_ConvertedFile*) pData;
	ConvertSound_PS3_Main(pxFile);
	s_pxThreadFinishedSignal->SetSignal();
}

void Tools_SoundConversionPreview::ConvertSound_PS3_Main(Tools_SoundConversionPreview_ConvertedFile* pxFile)
{
	//generate a name for an intermediate xma file
	char szIntermediateFile[MAX_PATH];
	strcpy(szIntermediateFile, pxFile->m_xConvertedFilename.GetString());
	char* pcDot = strrchr(szIntermediateFile, '.');
	if(pcDot) *pcDot = 0;
	strcat(szIntermediateFile, ".Intermediate");
	
	
	Tools_WAV_Reader xWAVReader;
	if ( !xWAVReader.ReadHeader(pxFile->m_xFullFilename.GetString()) )
	{
		pxFile->m_xError = "Failed to load file using wav reader";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}
	

	// Premade 'channel string' - it's easier to destroy than create
	Asura_Char szChannels[20];
	strncpy(szChannels, "0 1 2 3 4 5 6 7 ", 20);
	
	// Terminate the string early to make the input channel parameters 
	szChannels[ (xWAVReader.GetHeader().nChannels * 2) - 1 ] = 0; 

	
	Asura_Char szCommand[MAX_PATH * 2];
	Asura_String xFormat;
	
	#if 0 //avoiding this because of difficulties with playing the resultant wav file
	// compute num samples, assuming 16 bit samples (2 bytes per sample)
	u_int uNumSamples = xWAVReader.GetDataSize() / 2;
	// if num samples is less than double the MP3 packet size (1152) use ADPCM
	bool bTinyFile = uNumSamples < (2 * 1152);
	if( bTinyFile)
	{
		xFormat = "-adpcm";
	}
	else
	#endif
	{
		xFormat = "-mp3 ";
		
		Asura_String xQuality;
		int iQuality = Tools_SoundEvent_System::GetPlatformCompressionQuality(TOOLS_PLATFORM_PS3, pxFile->m_iQuality);
		Asura_StringUtil::ToString( iQuality, xQuality );

		xFormat += xQuality;

		xFormat += " -fscomp";
	}

	// Execute the MSF encoder tool.
	Asura_PrintfToBfr	( szCommand
						, "MSEnc.exe -in \"%s\" %s -out \"%s\" -x -loop %s"
						, (MAX_PATH * 2 ) - 1 
						, true
						, pxFile->m_xFullFilename.GetString()
						, szChannels
						, szIntermediateFile
						, xFormat.GetString() );
	
	

	PROCESS_INFORMATION xProcessInfo;
	STARTUPINFO xStartupInfo = {0};
	xStartupInfo.cb = sizeof( STARTUPINFO );

	if( !CreateProcess( NULL, szCommand, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &xStartupInfo, &xProcessInfo ) )
	{
		pxFile->m_xError = "Failed to execute MSEnc.exe";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	//wait for the conversion to finish , or for a request to stop
	while(true)
	{
		DWORD uResult = WaitForSingleObject( xProcessInfo.hProcess, 50 );
		if(uResult == WAIT_OBJECT_0)
		{
			//finished
			break;
		}
		if(s_bAbortThread)
		{
			TerminateProcess(xProcessInfo.hProcess, 0);
			WaitForSingleObject( xProcessInfo.hProcess, INFINITE );
			CloseHandle( xProcessInfo.hThread );
			CloseHandle( xProcessInfo.hProcess );

			pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
			return;
		}
	}

	DWORD uProcessExitCode = static_cast< u_int>( ~0);
	GetExitCodeProcess( xProcessInfo.hProcess, &uProcessExitCode);

	CloseHandle( xProcessInfo.hThread );
	CloseHandle( xProcessInfo.hProcess );

	if(uProcessExitCode != 0)
	{
		//error converting
		pxFile->m_xError = "MSEnc.exe returned an error";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	//Now convert the mp3 file back to a wav
	Asura_PrintfToBfr	( szCommand
						, "Utils\\mpg123-static.exe -w %s %s"
						, (MAX_PATH * 2 ) - 1 
						, true
						, pxFile->m_xConvertedFilename.GetString()
						, szIntermediateFile);

	if( !CreateProcess( NULL, szCommand, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &xStartupInfo, &xProcessInfo ) )
	{
		pxFile->m_xError = "Failed to execute mpg123-static.exe";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	//wait for the conversion to finish , or for a request to stop
	while(true)
	{
		DWORD uResult = WaitForSingleObject( xProcessInfo.hProcess, 50 );
		if(uResult == WAIT_OBJECT_0)
		{
			//finished
			break;
		}
		if(s_bAbortThread)
		{
			TerminateProcess(xProcessInfo.hProcess, 0);
			WaitForSingleObject( xProcessInfo.hProcess, INFINITE );
			CloseHandle( xProcessInfo.hThread );
			CloseHandle( xProcessInfo.hProcess );

			pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
			return;
		}
	}

	uProcessExitCode = static_cast< u_int>( ~0);
	GetExitCodeProcess( xProcessInfo.hProcess, &uProcessExitCode);

	CloseHandle( xProcessInfo.hThread );
	CloseHandle( xProcessInfo.hProcess );

	if(uProcessExitCode != 0)
	{
		//error converting (don't think mpg123-static ever returns an error code , so this won't occur)
		pxFile->m_xError = "mpg123-static.exe returned an error"; 
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
	}
	else
	{
		Tools_FileSystem::GetSizeOfFile(szIntermediateFile, pxFile->m_uFileSize);
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_READY;
	}

}


void Tools_SoundConversionPreview::ConvertSound_XBox360(const void* pData)
{
	Tools_SoundConversionPreview_ConvertedFile* pxFile = (Tools_SoundConversionPreview_ConvertedFile*) pData;
	ConvertSound_XBox360_Main(pxFile);
	s_pxThreadFinishedSignal->SetSignal();
}

void Tools_SoundConversionPreview::ConvertSound_XBox360_Main(Tools_SoundConversionPreview_ConvertedFile* pxFile)
{
	//generate a name for an intermediate xma file
	char szIntermediateFile[MAX_PATH];
	strcpy(szIntermediateFile, pxFile->m_xConvertedFilename.GetString());
	char* pcDot = strrchr(szIntermediateFile, '.');
	if(pcDot) *pcDot = 0;
	strcat(szIntermediateFile, ".Intermediate");



	Asura_Char szCommand[MAX_PATH * 2];
	
	int iQuality = Tools_SoundEvent_System::GetPlatformCompressionQuality(TOOLS_PLATFORM_XBOX360, pxFile->m_iQuality);
	
	Asura_PrintfToBfr	( szCommand
						, "XMAEncode.exe \"%s\" /l /b 8 /Q %d /t \"%s\" /d \"%s\""
						, (MAX_PATH * 2 ) - 1 
						, true
						, pxFile->m_xFullFilename.GetString()
						, iQuality
						, szIntermediateFile
						, pxFile->m_xConvertedFilename.GetString());

	PROCESS_INFORMATION xProcessInfo;
	STARTUPINFO xStartupInfo = {0};
	xStartupInfo.cb = sizeof( STARTUPINFO );

	if( !CreateProcess( NULL, szCommand, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &xStartupInfo, &xProcessInfo ) )
	{
		pxFile->m_xError = "Failed to execute xmaencode.exe";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	//wait for the conversion to finish , or for a request to stop
	while(true)
	{
		DWORD uResult = WaitForSingleObject( xProcessInfo.hProcess, 50 );
		if(uResult == WAIT_OBJECT_0)
		{
			//finished
			break;
		}
		if(s_bAbortThread)
		{
			TerminateProcess(xProcessInfo.hProcess, 0);
			WaitForSingleObject( xProcessInfo.hProcess, INFINITE );
			CloseHandle( xProcessInfo.hThread );
			CloseHandle( xProcessInfo.hProcess );

			pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
			return;
		}
	}

	DWORD uProcessExitCode = static_cast< u_int>( ~0);
	GetExitCodeProcess( xProcessInfo.hProcess, &uProcessExitCode);

	CloseHandle( xProcessInfo.hThread );
	CloseHandle( xProcessInfo.hProcess );


	if(uProcessExitCode != 0)
	{
		//error converting
		pxFile->m_xError = "xmaencode.exe returned an error.";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	//Unfortunately the wav file uses an extentend format ,which causes it to be played in just the left speaker
	//when played via PlaySound.
	//So we'll load up the file , and fiddle the format back.
	Tools_File_Wrapper xFile;
	if(!xFile.Open(pxFile->m_xConvertedFilename.GetString(), true))
	{
		pxFile->m_xError = "Failed to open wav file to alter its format";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	u_int uFileSize = xFile.FileSize();
	char* pcBuffer = new char[uFileSize];
	xFile.Read(pcBuffer, uFileSize);
	xFile.Close();

	//for simplicity , just going to fix the format created by the current xmaencode.
	//so check to make sure it fits what is expected
	short* psFormatSize = (short*)(pcBuffer+16);
	if(strncmp(pcBuffer,"RIFF", 4) || strncmp(pcBuffer+8,"WAVEfmt ", 8) || *psFormatSize != 40)
	{
		delete [] pcBuffer;
		pxFile->m_xError = "Unexpected format for converted wav file";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	const u_int uSizeDiff = 40-16;	
	//Fix up the stored file size
	u_int* puFileSize = (u_int*) (pcBuffer+4);
	*puFileSize -= uSizeDiff;
	
	//and the format header size
	*psFormatSize -= uSizeDiff;
	
	//and switch the format tag
	WAVEFORMAT* pxFormat = (WAVEFORMAT*)(pcBuffer+20);
	pxFormat->wFormatTag = WAVE_FORMAT_PCM;
	
	//open up the file and write out a new version
	if(!xFile.Create(pxFile->m_xConvertedFilename.GetString(), true))
	{
		delete [] pcBuffer;
		pxFile->m_xError = "Failed to open wav file for writing";
		pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_INVALID;
		return;
	}

	//write out the first 36 bytes , containg up to the end of the format header
	xFile.Write(pcBuffer, 36);
	//skip the next 24 bytes, and write the rest of the file from there
	u_int uOffset = 36+uSizeDiff;
	xFile.Write(pcBuffer+uOffset, uFileSize - uOffset);
	xFile.Close();

	delete [] pcBuffer;


	Tools_FileSystem::GetSizeOfFile(szIntermediateFile, pxFile->m_uFileSize);
	pxFile->m_eState = Tools_SoundConversionPreview_ConvertedFile::FILESTATE_READY;

}




////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------------**
** Tools_SoundConversionPreview_ConvertedFile **
**--------------------------------------------*/

Tools_SoundConversionPreview_ConvertedFile::Tools_SoundConversionPreview_ConvertedFile()
{
	m_eState = FILESTATE_INVALID;
	m_uLastUsedStamp = 0;
}

void Tools_SoundConversionPreview_ConvertedFile::SetLastUsedStamp()
{
	static u_int s_uCounter = 1;
	m_uLastUsedStamp = s_uCounter++;
}

bool Tools_SoundConversionPreview_ConvertedFile::Matches(const char* szFilename, int iPlatform, int iQuality) const
{
	if(m_eState == FILESTATE_INVALID) return false;
	if(m_xRelativeFilename.Compare(szFilename)) return false;

	return iPlatform == m_iPlatform && iQuality == m_iQuality;
}
