//
// Tools_SoundEvent_System.h
//
// Created by Richard Rice - 06/07/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_SOUNDEVENT_SYSTEM__
#define __TOOLS_SOUNDEVENT_SYSTEM__

#include "Core/Asura_Hash.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_String.h"
#include "Sound/Asura_Sound_Event.h"

#include "Misc/Tools_Platform.h"
#include "Misc/Tools_MessageLog.h"


// Forward declarations
template <class T> class List;
template <class T> class Asura_TemplateHashTree;


class Tools_SoundEvent_System
{
	friend class Asura_Chunk_SoundEvent_ToolsGlobalSettings;
public :
	static void Initialise();
	static void Shutdown();

	static void Update();

	//returns true if any sound event files are modified
	static bool IsModified();	
	//Save all sound event files (plus mixers)
	static void SaveAll();
	//save all files marked as modified
	static void SaveModified();

	//Save the mixer file
	static bool SaveMixers();

	//Save the global settings (contains sdefault platform sample rates)
	static bool SaveGlobalSettings();

	//sets the file containing the event/sample s modified
	static void SetModified(class Asura_Sound_Sample* pxSample);
	static void SetModified(class Asura_Sound_Event* pxEvent);

	//szFilename is filename without path/extension
	static void SetModified(const char* szFilename);

	static void SetMixersModified();
	static bool GetMixersModified() 			{ return s_bMixersModified; }
	static void SetGlobalSettingsModified();

	//Used by Tools_SaveAsura to mark the sound event files as ones to be tracked
	static void SetFilesToCheckForExternalModification();

	//Gets default file for new events/samples
	static const char* GetDefaultFile() {return "Default";}

	//Registers a new event/sample that has been created in this trip into tools
	//This is used to track which ones can be safely transferred between files.
	static void RegisterNewEvent(const class Asura_Sound_Event* pxEvent);
	static void RegisterNewSample(const class Asura_Sound_Sample* pxSample);

	//Checks to see whether it is safe to transfer the event/sample to the new file
	static bool CanTransferEvent(const class Asura_Sound_Event* pxEvent,const char* szDestFile);
	static bool CanTransferSample(const class Asura_Sound_Sample* pxSample,const char* szDestFile);

	//Adds some sound event editing controls to another overlay
	static void AddOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset);

	//default event/sample categories for all projects
	static int GetNumDefaultCategories();
	static const char* GetDefaultCategory(int iIndex);
	static const char* GetDefaultCategoryForWavFile(const char* szFilename);


	//Create a new event/sample with name chosen by user
	static Asura_Hash_ID CreateEvent(const char* szFile,const char* szCategory,const char* szSubCategory,const char* szEventName);
	static void CreateSample(const char* szFile,const char* szCategory,const char* szSubCategory,List<Asura_Hash_ID>* pxOutputSampleList);


	//Create a code event that that hasn't been made yet.
	static Asura_Hash_ID CreateCodeEvent(const char* szFile = 0,const char* szCategory = 0,const char* szSubCategory = 0);

	//Used by the ASURA_SOUND_EVENT_LOOKUP_GLOBALINIT macro to build the list of sound events
	//that are called in code.
	static void RegisterCodeEvent(const char* szName);

	//returns true if the hash matches that of a code event registered by RegisterCodeEvent
	static bool IsCodeEvent(Asura_Hash_ID uEventHash);
	
	//Deletes the specified sample/event
	static bool DeleteSample(Asura_Hash_ID uHash,bool bConfirm);
	static bool DeleteEvent(Asura_Hash_ID uHash,bool bConfirm);


#ifdef ASURA_USE_AUDIO_SYSTEM
	// Gets a list of all sample hashes used by samples and events that are part of this event.
	// Only contains each sample ONCE.
	// These are SAMPLE hashes, not WAVE hashes.
	static void ListSamplesUsedByEvent(const class Asura_Sound_Event* pxEvent,List<Asura_Hash_ID>& xSoundList);
#else
	//Gets a list of all sound IDs used by samples and events that are part of this event
	static void ListSoundIDsUsedByEvent(const class Asura_Sound_Event* pxEvent,List<int>& xSoundList);
#endif

	//Get the default sample rate for a platform
	static int GetPlatformSampleRate(int iPlatform);
	static void SetPlatformSampleRate(int iPlatform,int iSampleRate);

	//get the default compression quality for a platform
	static int GetDefaultPlatformCompressionQuality(int iPlatform);
	static void SetDefaultPlatformCompressionQuality(int iPlatform, int iQuality);
	
	static int GetPlatformCompressionQuality( const int iPlatform, const int iQuality );
	static void SetPlatformCompressionQuality( const int iPlatform, const int iQuality, const int iValue );

	//Indicates whether the exports actually pay any attention to the sample rate values
	static bool DoesPlatformUseSampleRate(int iPlatform)
	{
		return iPlatform == TOOLS_PLATFORM_PS2 || iPlatform == TOOLS_PLATFORM_PSP || iPlatform == TOOLS_PLATFORM_WII;
	}

	//Indicates whether the exports pay any attention to the compression quality values
	static bool DoesPlatformUseCompressionQuality(int iPlatform)
	{
		return ( ( iPlatform == TOOLS_PLATFORM_XBOX360 ) || ( iPlatform == TOOLS_PLATFORM_PS3 ) );
	}

	//Gets the sample rate by checking for a sample using the sound.
	//Returns 0 if not found
	static int GetSampleRate(int iPlatform, int iSoundID);
	static int GetSampleRate(int iPlatform, const char* szWavName);

	//Calculates the maximum simultaneous samples that an event might use
	static int GetMaxSamplesPlayed(const Asura_Sound_Event* pxEvent);

	//Checks sound events for errors ,and shows a log of them if any are found
	static void Verify(bool bAutomatic);

	static void SetEverHadEventsOrSamples() {s_bEverHadEventsOrSamples = true;}
	static void SetEverHadMixers(){s_bEverHadMixers = true;}

	//Finds wav files that are referred to by the samples , but have moved to another directory.
	static void FindWavFiles();

	static void ReloadWavsForSample(Asura_Hash_ID uSampleHash,bool bForceReloadNow);
	static void ReloadWavsForEvent(Asura_Hash_ID uEventHash,bool bForceReloadNow);

	//searches for samples and events that use the wav
	static void FindReferencesToWav();

#ifdef ASURA_USE_AUDIO_SYSTEM
	static void FindReferencesToWav(Asura_Hash_ID uWave);
#else
	static void FindReferencesToWav(int iSoundID);
#endif

	//searches for blueprints (could be extended to other things like entities, etc.) that reference a particular event
	static void FindReferencesToEvent( const Asura_Sound_Event* pxEvent );
	
	//scans exports to determine which events/samples aren't used anywhere
	static void FindUnusedEventsAndSamples();

	//Removes any samples that aren't referred to by any events
	static void DeleteUnusedSamples();

	//deletes the events/samples found by the previous search
	static void DeleteUnusedEventsAndSamples();
	static bool IsEnabled_DeleteUnusedEventsAndSamples();

	//for loading sound events from the platform-specific directories
	static void LoadSoundEventsForPlatform(int iPlatform);

	static u_int GetMemoryLimit() {return s_uMemoryLimit;};
	static void SetMemoryLimit( const u_int uNewLimit) { s_uMemoryLimit = uNewLimit; };

private:
	//load all sound event files
	static void LoadAll();

	//adds path and extension to get full filename
	static void GetFullFilename(const char* szFilename,Asura_String& xFullFilename);


	static bool SaveFile(const char* szFilename);
	

	//overlay control callbacks
	static void EditSample();
	static void EditEvent();

	static void BuildWavToSampleTree();

	//recursively adds the IDs of all events and samples used by an event to the trees
	static void AddSoundEventsAndSamplesToTree(const Asura_Sound_Event* pxEvent, Asura_HashTree& xEventTree, Asura_HashTree& xSampleTree);


	
	static List<Asura_String> s_xModifiedFiles;
	static bool s_bMixersModified;
	static bool s_bGlobalSettingsModified;

	static class Asura_HashTree s_xNewEventTree;
	static class Asura_HashTree s_xNewSampleTree;


	//default sample rates for each platform
	static int s_aiPlatformSampleRates[TOOLS_PLATFORM_NUM_PLATFORMS];

	static int s_aiPlatformCompressionQuality[TOOLS_PLATFORM_NUM_PLATFORMS];
	static int s_aiPlatformCompressionRates[TOOLS_PLATFORM_NUM_PLATFORMS][TSS_QUALITY_NUM_ENTRIES];

	//Hash tree for mapping hashes of wav filenames to sample hashes
	static Asura_TemplateHashTree<Asura_Hash_ID>* s_pxWavToSampleTree;

	static bool s_bEverHadMixers;
	static bool s_bEverHadEventsOrSamples;

	static bool s_bDefaultMaxInstancesChanged;
	static u_int s_uDefaultMaxInstances;

	//-1 if using the root directory , or a value from the Tools_Platform enumeration if using a platform specific dir
	static int s_iCurrentPlatformDirectory;

	//results of most recent 'find unused events'
	static List<Asura_Hash_ID> s_xUnusedEventList;
	static List<Asura_Hash_ID> s_xUnusedSampleList;

	static u_int s_uMemoryLimit;	// this should be saved per project (in MB)
};

class Tools_MessageLogAction_SoundEvent	: public Tools_MessageLogAction
{
public :
	Tools_MessageLogAction_SoundEvent(Asura_Hash_ID uHashID) {m_uHashID = uHashID;}
	virtual void Execute();

	Asura_Hash_ID m_uHashID;
};

class Tools_MessageLogAction_SoundSample : public Tools_MessageLogAction
{
public :
	Tools_MessageLogAction_SoundSample(Asura_Hash_ID uHashID) {m_uHashID = uHashID;}
	virtual void Execute();

	Asura_Hash_ID m_uHashID;
};


//Used for storing all the local/network wav filenames , so that any files that have been moved can be found again.
class Tools_WavFinder
{
public :
	Tools_WavFinder() {m_pxWavTree = 0;m_pxFileList = 0;}
	~Tools_WavFinder();

	const char* FindFile(const char* szFilename);

private :
	void BuildTree();

	Asura_TemplateHashTree<List<const char*> >* m_pxWavTree;
	List<Asura_String>* m_pxFileList;
};


//deals with playing a sound sample when it isn't part of a sound event
class Tools_SoundSamplePreview
{
public :
	Tools_SoundSamplePreview();
	~Tools_SoundSamplePreview();

	void Play(Asura_Hash_ID uSampleHash);
	void Update();
	void Stop();

private :
	Asura_Handle m_iSoundHandle;
	float m_fTimer;
	Asura_Hash_ID m_uSampleHash;


};

#endif //__TOOLS_SOUNDEVENT_SYSTEM__
