#include "Core/Asura.h"

#include "TextureTiler/Tools_TextureTiler_Export.h"

// Core
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Chunk_TiledTexture.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "FileAccess/Asura_FileAccess.h"
#include "NVidia/DDS_Utilities/dxtlib.h"
#include "Rendering/Asura_Direct3D.h"
#include "Xbox360/ChunkSystem/Asura_Chunk_ResourceFile_Xbox360.h"

// Asura PC
#include "Rendering/Asura_Bmp.h"

// Tools
#include "ImageProcessing/Tools_PS3_Texture_Writer.h"
#include "ImageProcessing/Tools_PSP_Texture_Writer.h"
#include "ImageProcessing/Tools_Tim2_Writer.h"
#include "ImageProcessing/Tools_Xbox360_Texture_Writer.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_TextureConversionData.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "TextureTiler/Tools_TextureTiler_TTL.h"
#include "Wii/TPLOptions.h"
#include "Wii/TPLWriter.h"

const int s_iMaxSafeImageDimension(8192);

// <-- Tools_TT_Colour -->
const int Tools_TT_Colour::Null(0x00000000);
const int Tools_TT_Colour::Black(0xff000000);
const int Tools_TT_Colour::White(0xffffffff);
const int Tools_TT_Colour::Red(0x00ff0000);
const int Tools_TT_Colour::Green(0xff00ff00);
const int Tools_TT_Colour::Blue(0xff0000ff);
const int Tools_TT_Colour::Magenta(0xffff00ff);
const int Tools_TT_Colour::Yellow(0xffffff00);

// <-- Tools_TT_FileType -->
const char* Tools_TT_FileType::s_apcFileExtensionName[TOOLS_TT_FILE_NAME_EXTENSION_NUM] =
{
	".*",	  // TOOLS_TT_ALL
	".bmp",   // TOOLS_TT_BMP
	".dds",	  // TOOLS_TT_DDS
	".jpg",	  // TOOLS_TT_JPG
	".png",	  // TOOLS_TT_PNG
	".psptx", // TOOLS_TT_PSP
	".tga",	  // TOOLS_TT_TGA
	".tm2",	  // TOOLS_TT_TM2
	".tpl",	  // TOOLS_TT_TPL
	".xpr",	  // TOOLS_TT_XPR
};
const char* Tools_TT_FileType::GetName( const u_int uFileType )
{
	return s_apcFileExtensionName[static_cast<const TOOLS_TT_FILE_NAME_EXTENSION>(uFileType)];
}

const Tools_TT_FileType::TOOLS_TT_FILE_NAME_EXTENSION Tools_TT_FileType::GetTextureFileTypeForPlatform( const Tools_Platform_Type ePlatform )
{
	TOOLS_TT_FILE_NAME_EXTENSION eRtn;

	switch( ePlatform )
	{

		case TOOLS_PLATFORM_PC:
		eRtn = TOOLS_TT_DDS;
		break;

		case TOOLS_PLATFORM_PS2:
		eRtn = TOOLS_TT_TM2;
		break;
		
		// TOOLS_PLATFORM_XBOX - deprecated
		
		case TOOLS_PLATFORM_WII:
		eRtn = TOOLS_TT_TPL;
		break;

		case TOOLS_PLATFORM_PSP:
		eRtn = TOOLS_TT_PSP;
		break;

		case TOOLS_PLATFORM_PS3:
		eRtn = TOOLS_TT_TGA;
		break;

		case TOOLS_PLATFORM_XBOX360:
		eRtn = TOOLS_TT_DDS; // changed from TOOLS_TT_XPR as 360 now supports DDS;
		break;

		default:
		Asura_Assert( false,"Default texture type for platform unknown. Using TGA as default." );
		eRtn = TOOLS_TT_TGA;
		break;
	}

	return eRtn;
}

// <-- Tools_TT_DDSCompressionFormat -->
const char* Tools_TT_DDSCompressionFormat::s_apcName[TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT_NUM] = 
{
	"D3DFMT_A8R8G8B8",  // TOOLS_TT_EXPORT_DDS_COMPRESSION_UNCOMPRESSED
	"D3DFMT_DXT1",		// TOOLS_TT_EXPORT_DDS_COMPRESSION_DXT1
	"D3DFMT_DXT3",		// TOOLS_TT_EXPORT_DDS_COMPRESSION_DXT3
	"D3DFMT_DXT5",		// TOOLS_TT_EXPORT_DDS_COMPRESSION_DXT5
};

const char* Tools_TT_DDSCompressionFormat::GetName( const u_int uFormat )
{
	return s_apcName[static_cast<const TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT>(uFormat)];
}

// <-- Tools_TT_PSPPixelFormat -->
const char* Tools_TT_PSPPixelFormat::s_apcName[TOOLS_TT_EXPORT_PSP_PIXEL_FORMAT_NUM] = 
{
	"4 bit palettised",
	"8 bit palettised",
};

const char* Tools_TT_PSPPixelFormat::GetName( const u_int uFormat )
{
	TOOLS_TT_EXPORT_PSP_PIXEL_FORMAT eFormat( static_cast<const TOOLS_TT_EXPORT_PSP_PIXEL_FORMAT>(uFormat) );

	if( (eFormat<0) || (eFormat >= TOOLS_TT_EXPORT_PSP_PIXEL_FORMAT_NUM) )
	{
		eFormat = TOOLS_TT_EXPORT_PSP_PIXEL_8;
	}

	return s_apcName[eFormat];
}

const int Tools_TT_PSPPixelFormat::GetValue( const u_int uFormat )
{
	TOOLS_TT_EXPORT_PSP_PIXEL_FORMAT eFormat( static_cast<const TOOLS_TT_EXPORT_PSP_PIXEL_FORMAT>(uFormat) );

	if( eFormat == TOOLS_TT_EXPORT_PSP_PIXEL_4 )
	{
		return 4;
	}
	
	return 8;	
}

Tools_TT_Export_Pixel::Tools_TT_Export_Pixel()
:uR(0)
,uG(0)
,uB(0)
,uA(0)
{}

const bool operator==( const Tools_TT_Export_Pixel& xLhs,const Tools_TT_Export_Pixel& xRhs )
{
	return (xLhs.uA == xRhs.uA) && (xLhs.uR == xRhs.uR) && (xLhs.uG == xRhs.uG) && (xLhs.uB == xRhs.uB);
}

// <-- Tools_TT_TPLAlphaBits -->
const char* Tools_TT_TPLAlphaBits::s_apcName[TOOLS_TT_EXPORT_TPL_ALPHA_BITS_NUM] = 
{
	"0",
	"1",
	"4",
	"8",
};

const char* Tools_TT_TPLAlphaBits::GetName( const u_int uFormat )
{
	return s_apcName[static_cast<const TOOLS_TT_EXPORT_TPL_ALPHA_BITS>(uFormat)];
}

const u_int Tools_TT_TPLAlphaBits::GetValue( const u_int uFormat )
{
	u_int uRtn(0);
	
	switch( static_cast<const TOOLS_TT_EXPORT_TPL_ALPHA_BITS>(uFormat) )
	{
		case TOOLS_TT_EXPORT_TPL_ALPHA_BITS_0:
			uRtn = 0;
			break;

		case TOOLS_TT_EXPORT_TPL_ALPHA_BITS_1:
			uRtn = 1;
			break;

		case TOOLS_TT_EXPORT_TPL_ALPHA_BITS_4:
			uRtn = 4;
			break;

		case TOOLS_TT_EXPORT_TPL_ALPHA_BITS_8:
			uRtn = 8;
			break;

		default:
			Asura_Assert( false,"Unrecognised TPL Alpha Bit Level, returning 0" );
			uRtn = 0;
			break;
	}

	return uRtn;
}



// <-- Tools_TT_Export_RawPixelData -->
Tools_TT_Export_RawPixelData::Tools_TT_Export_RawPixelData()
:m_iTextureFlags(0)
,m_uWidth(0)
,m_uHeight(0)
,m_uBytesPP(0)
,m_uSize(0)
,m_uWindowXBegin(0)
,m_uWindowXEnd(0)
,m_uWindowYBegin(0)
,m_uWindowYEnd(0)
,m_uWindowSize(0)
,m_pcBitData( NULL )
{}

Tools_TT_Export_RawPixelData::Tools_TT_Export_RawPixelData( const u_int uWidth,const u_int uHeight,const u_int uBytesPP )
:m_iTextureFlags(0)
,m_uWidth(uWidth)
,m_uHeight(uHeight)
,m_uBytesPP(uBytesPP)
,m_uSize(m_uWidth*m_uHeight*m_uBytesPP)
,m_uWindowXBegin(0)
,m_uWindowXEnd(0)
,m_uWindowYBegin(0)
,m_uWindowYEnd(0)
,m_uWindowSize(0)
,m_pcBitData( new u_char[m_uSize] )
{
	Clear();
}

Tools_TT_Export_RawPixelData::Tools_TT_Export_RawPixelData( const u_int uWidth,const u_int uHeight,const u_int uBytesPP, u_char* pcData )
:m_iTextureFlags(0)
,m_uWidth(uWidth)
,m_uHeight(uHeight)
,m_uBytesPP(uBytesPP)
,m_uSize(m_uWidth*m_uHeight*m_uBytesPP)
,m_uWindowXBegin(0)
,m_uWindowXEnd(0)
,m_uWindowYBegin(0)
,m_uWindowYEnd(0)
,m_uWindowSize(0)
,m_pcBitData( pcData )
{
	Clear();
}

Tools_TT_Export_RawPixelData::Tools_TT_Export_RawPixelData( const Tools_TT_Export_RawPixelData& xOther )
{
	*this = xOther;
}

Tools_TT_Export_RawPixelData::~Tools_TT_Export_RawPixelData()
{
	if(m_pcBitData)
	{
		delete [] m_pcBitData;
		m_pcBitData = NULL;
	}
}

void Tools_TT_Export_RawPixelData::Set( const u_int uWidth,const u_int uHeight,const u_int uBytesPP )
{
	if(m_pcBitData)
	{
		delete [] m_pcBitData;
		m_pcBitData = NULL;
	}

	m_uWidth   = uWidth;
	m_uHeight  = uHeight;
	m_uBytesPP = uBytesPP;
	m_uSize    = m_uWidth*m_uHeight*m_uBytesPP;

	m_uWindowXEnd   = 0;
	m_uWindowYBegin = 0;
	m_uWindowYEnd   = 0;
	m_uWindowSize   = 0;

	m_pcBitData = new u_char[m_uSize];

	Clear();
}

const Tools_TT_Export_RawPixelData& Tools_TT_Export_RawPixelData::operator=( const Tools_TT_Export_RawPixelData& xOther )
{
	m_uWidth  = xOther.GetWidth();
	m_uHeight = xOther.GetHeight();
	m_uBytesPP    = xOther.GetBytesPP();
	m_uSize   = xOther.GetSize();
	
	m_uWindowXBegin = xOther.m_uWindowXBegin;
	m_uWindowXEnd   = xOther.m_uWindowXEnd;
	m_uWindowYBegin = xOther.m_uWindowYBegin;
	m_uWindowYEnd   = xOther.m_uWindowYEnd;
	m_uWindowSize   = xOther.GetWindowSize();

	delete [] m_pcBitData;
	m_pcBitData = new u_char[m_uSize];

	// copy bit data
	for( u_int u(0);u<GetSize();++u )
	{
		// don't use xOther.|this-> operator[]() as it is (by design) effected by window settings - NDB
		m_pcBitData[u] = xOther.m_pcBitData[u];
	}

	return *this;
}

u_char& Tools_TT_Export_RawPixelData::operator[]( const u_int uWindowIndex )
{
	u_int uIndex( (m_uWindowYBegin+(uWindowIndex/GetWindowPitch()))*GetPitch() + m_uWindowXBegin*GetBytesPP() + (uWindowIndex%GetWindowPitch()) );

	Asura_Assert( uIndex<=m_uSize,"index greater than size of array" );

	return m_pcBitData[ uIndex ];
}

u_char* Tools_TT_Export_RawPixelData::GetWindowData( u_char* pcOutput, const u_int uSize )
{
	const u_int uMaxIteration( (uSize<GetWindowSize())?uSize:GetWindowSize() );

	for( u_int u(0);u<uMaxIteration;++u )
	{
		pcOutput[u] = this->operator[](u);
	}
	
	return pcOutput;
}

void Tools_TT_Export_RawPixelData::Clear()
{
	if( m_pcBitData == NULL )
	{
		return;
	}

	// set all pixels to null
	for( u_int u(0);u<m_uSize;++u )
	{
		m_pcBitData[u] = 0x00;
	}
}

const bool Tools_TextureTiler_Export_RipBitsFromARGBD3DTexture( Direct3D_Texture* dxTexture,D3DXIMAGE_INFO imageInfo,Tools_TT_Export_RawPixelData& xOutput )
{
	bool bSuccess(true);

	HRESULT hres;;
	D3DSURFACE_DESC dxSurfaceDesc;
	Direct3D_Surface* dxSurface(NULL);

	//get the surface
	hres = dxTexture->GetSurfaceLevel(0,&dxSurface);

	if(FAILED(hres))
	{
		Asura_Assert(0,"Failed to get D3D surface from texture");
		bSuccess = false;
	}

	if(bSuccess)
	{
		//get the surface description
		hres = dxSurface->GetDesc(&dxSurfaceDesc);
	
		if(FAILED(hres))
		{
			Asura_Assert(0,"Failed to get D3D surface description");
			bSuccess = false;
		}
	}

	if(bSuccess)
	{
		//check its in the format requested
		if(dxSurfaceDesc.Format != D3DFMT_A8R8G8B8)
		{
			Asura_Assert(0,"Failed to load texture as D3DFMT_A8R8G8B8");
			bSuccess = false;
		}
	}

	D3DLOCKED_RECT rect;

	if(bSuccess)
	{
		//lock the surface	
		hres = dxSurface->LockRect(&rect,0,D3DLOCK_READONLY);
		
		if(FAILED(hres))
		{
			Asura_Assert(0,"Failed to lock the texture surface");
			dxSurface->UnlockRect();
			bSuccess = false;
		}
	}

	if(bSuccess)
	{
		// copy image bits from texture array to output, taking account of any image 
		// resizing by using the width and height stored in imageInfo (D3D will 
		// pad out textures to make width & height a power of 2).
		const u_char* pcBits( static_cast<const u_char*>(rect.pBits) );
		
		u_int uImageLoc(0);
		u_int uOutputLoc(0);

		// extract alpha channel
		if( xOutput.GetBytesPP() == 1 )
		{
			for( u_int uImageY(0);uImageY<imageInfo.Height;++uImageY )
			{
				uImageLoc = (uImageY * dxSurfaceDesc.Width * 4) + 3; // skip rgb bits - NDB
				
				for( u_int uImageX(0);uImageX<imageInfo.Width;++uImageX,uImageLoc+=4 )
				{
					xOutput[uOutputLoc++] = pcBits[ uImageLoc ]; // Alpha
				}
			}
		}
		// extract ARGB channels
		else
		{
			for( u_int uImageY(0);uImageY<imageInfo.Height;++uImageY )
			{
				uImageLoc = uImageY * dxSurfaceDesc.Width * 4;

				for( u_int uImageX(0);uImageX<imageInfo.Width;++uImageX )
				{
					xOutput[uOutputLoc++] = pcBits[ uImageLoc++ ]; // Red
					xOutput[uOutputLoc++] = pcBits[ uImageLoc++ ]; // Green
					xOutput[uOutputLoc++] = pcBits[ uImageLoc++ ]; // Blue
					xOutput[uOutputLoc++] = pcBits[ uImageLoc++ ]; // Alpha
				}
			}
		}


		hres = dxSurface->UnlockRect();

		Asura_Assert(!hres,"Error unlocking surface");
	}

	// clear up
	if(dxSurface) dxSurface->Release();
	if(dxTexture) dxTexture->Release();

	return bSuccess;
}

const bool Tools_TextureTiler_Export_LoadD3DTextureFromFile( const char* szInputName,Direct3D_Texture** pxDxTexture,D3DXIMAGE_INFO& xImageInfo )
{
	HRESULT hres;

	//attempt to load in the texture to an 8888 format
	hres = D3DXCreateTextureFromFileExA
		(
			Asura_Direct3D::D3DDevice,
			szInputName,
			D3DX_DEFAULT,       //width
			D3DX_DEFAULT,	    //height
			1,					//MipLevels
			0,					//Usage
			D3DFMT_A8R8G8B8,	//format
			D3DPOOL_SYSTEMMEM,	//Pool
			D3DX_FILTER_NONE,	//Filter
			D3DX_FILTER_NONE,	//MipFilter
			0,					//ColorKey
			&xImageInfo,		//pSourceInfo
			0,					//pPalette
			pxDxTexture
		);

	if( FAILED(hres) )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"Error %x loading %s",hres,szInputName );
	}

	return (FAILED(hres))? false:true;
}

const bool Tools_TT_Export_FindTextureAndImageSize( const char* szFileName,int& iTextureWidth,int& iTextureHeight,int& iImageWidth,int& iImageHeight )
{
	Direct3D_Texture* pxDxTexture(NULL);
	D3DXIMAGE_INFO    xImageInfo;

	Asura_String xFileName( szFileName );
	Tools_TT_PrefixWithGraphicsPath(xFileName);

	bool bSuccess = Tools_TextureTiler_Export_LoadD3DTextureFromFile( xFileName.GetString(),&pxDxTexture,xImageInfo );

	if( pxDxTexture == NULL )
	{
		bSuccess = false;
	}

	HRESULT xHres;

	D3DSURFACE_DESC xDxSurfaceDesc;
	Direct3D_Surface* pxDxSurface(NULL);

	if( bSuccess ) 
	{
		//get the surface
		xHres = pxDxTexture->GetSurfaceLevel( 0,&pxDxSurface );

		if( FAILED(xHres) )
		{
			bSuccess = false;
		}
	}
	
	if( pxDxSurface == NULL )
	{
		bSuccess = false;
	}

	if( bSuccess ) 
	{
		// get infomation about the surface
		xHres = pxDxSurface->GetDesc( &xDxSurfaceDesc );

		if( FAILED(xHres) )
		{
			bSuccess = false;
		}
	}

	if( bSuccess ) 
	{
		iTextureWidth  = xDxSurfaceDesc.Width;
		iTextureHeight = xDxSurfaceDesc.Height;
	
		iImageWidth    = xImageInfo.Width;
		iImageHeight   = xImageInfo.Height;
	}

	// clear up
	if(pxDxSurface) pxDxSurface->Release();
	if(pxDxTexture) pxDxTexture->Release();

	return bSuccess;
}

const bool Tools_TextureTiler_Export_RipBitsFromARGBTexture( const char* szInputName,Tools_TT_Export_RawPixelData& xOutput )
{
	Direct3D_Texture* pxDxTexture(NULL);
	D3DXIMAGE_INFO    xImageInfo;

	if(!szInputName)
	{
		return false;
	}
	Asura_String xInputName( szInputName );
	Tools_TT_PrefixWithGraphicsPath(xInputName);

	Tools_TextureTiler_Export_LoadD3DTextureFromFile( xInputName.GetString(),&pxDxTexture,xImageInfo );

	return Tools_TextureTiler_Export_RipBitsFromARGBD3DTexture( pxDxTexture,xImageInfo,xOutput );
}

const bool Tools_TextureTiler_Export_RipAllARGBBitsFromFileInMemory( const Asura_Handle iFileHandle,const int iFileSize,Tools_TT_Export_RawPixelData& xOutput )
{
	char* pcData( new char[iFileSize] );
	Asura_FileAccess::ReadData( iFileHandle,pcData,iFileSize );

	HRESULT hres;

	Direct3D_Texture* pxDxTexture(NULL);
	
	D3DXIMAGE_INFO imageInfo;

	//attempt to load in the texture to an 8888 format
	hres = D3DXCreateTextureFromFileInMemoryEx(
			Asura_Direct3D::D3DDevice,
			pcData,
			iFileSize,
			D3DX_DEFAULT,       //width
			D3DX_DEFAULT,	    //height
			1,					//MipLevels
			0,					//Usage
			D3DFMT_A8R8G8B8,	//format
			D3DPOOL_SYSTEMMEM,	//Pool
			D3DX_FILTER_NONE,	//Filter
			D3DX_FILTER_NONE,	//MipFilter
			0,					//ColorKey
			&imageInfo,			//pSourceInfo
			0,					//pPalette
			&pxDxTexture
		);

	delete [] pcData;

	if(FAILED(hres))
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"Error %x loading file for texture extraction",hres );
		return false;
	}

	if( (imageInfo.Width > s_iMaxSafeImageDimension) || (imageInfo.Height > s_iMaxSafeImageDimension) )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"Image height or width greater than safty limit of %i, cowardly refusing to continue with texture extraction",s_iMaxSafeImageDimension );
		return false;
	}

	xOutput.Set( imageInfo.Width,imageInfo.Height,4 ); // assume 4BytesPP as argb 8888 requested - NDB
	xOutput.SetWindowToMax();

	if(FAILED(hres))
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"Error %x loading file from memory",hres );
		return false;
	}

	return Tools_TextureTiler_Export_RipBitsFromARGBD3DTexture( pxDxTexture,imageInfo,xOutput );
}

const bool Tools_TT_Export_SaveTextureAs( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName )
{
	bool bSuccess(false);

	if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_BMP) ) )
	{
		bSuccess = Tools_TT_Export_SaveBMP( xBitData,xOutputFileName );
	}
	else if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_DDS) ) )
	{
		bSuccess = Tools_TT_Export_SaveDDS( xBitData,xOutputFileName );
	}
	else if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_PSP) ) )
	{
		bSuccess = Tools_TT_Export_SavePSPTx( xBitData,xOutputFileName );	
	}
	else if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_TGA) ) )
	{
		bSuccess = Tools_TT_Export_SaveTGA( xBitData,xOutputFileName );	
	}
	else if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_TM2) ) )
	{
		bSuccess = Tools_TT_Export_SaveTM2( xBitData,xOutputFileName );
	}
	else if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_TPL) ) )
	{
		bSuccess = Tools_TT_Export_SaveTPL( xBitData,xOutputFileName );	
	}
	else if( xOutputFileName.EndsWith( Tools_TT_FileType::GetName(Tools_TT_FileType::TOOLS_TT_XPR) ) )
	{
		bSuccess = Tools_TT_Export_SaveXPR( xBitData,xOutputFileName );
	}
	else
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR failed to save \"%s\", unrecognised file extension",xOutputFileName.GetString() );
	}

	if( bSuccess )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::White,"\"%s\" saved sucessfully",xOutputFileName.GetString() );
	}
	else
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR \"%s\" save FAILED",xOutputFileName.GetString() );
	}

	return bSuccess;
}

const bool Tools_TT_Export_SaveTextureForPlatform( Tools_TT_Export_RawPixelData& xBitData,Asura_String xOutputFileName,const Tools_Platform_Type eTargetPlatform )
{	
	Tools_TT_ReplaceFileExtension( xOutputFileName,Tools_TT_FileType::GetName( Tools_TT_FileType::GetTextureFileTypeForPlatform( eTargetPlatform ) ) );
	return Tools_TT_Export_SaveTextureAs( xBitData,xOutputFileName );
}

// The following three _Put*() methods are required to construct the bmp file header - I hate them - NDB
void Tools_TT_PutByte(BYTE v, FILE *fp)
{
  fputc( v,fp );
}

void Tools_TT_PutLittleWord(WORD v, FILE *fp)
{
  unsigned char c1 = (unsigned char)(v & 0xff);
  unsigned char c2 = (unsigned char)((v >> 8) & 0xff);
  fputc(c1, fp);
  fputc(c2, fp);
}

void Tools_TT_PutLittleDword(DWORD v, FILE *fp)
{
  unsigned char c1 = (unsigned char)(v & 0xff);
  unsigned char c2 = (unsigned char)((v >> 8) & 0xff);
  unsigned char c3 = (unsigned char)((v >> 16) & 0xff);
  unsigned char c4 = (unsigned char)((v >> 24) & 0xff);

  fputc(c1, fp);
  fputc(c2, fp);
  fputc(c3, fp);
  fputc(c4, fp);
}

const bool Tools_TT_Export_SaveBMP( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName )
{		
	FILE * pFile( fopen(xOutputFileName.GetString(),"wb") );
	
	if (!pFile)
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR failed to open file \"%s\"",xOutputFileName.GetString() );
		return false;
	}

	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}
	
	BMPHEADER2 xBmpHeader;

	int iWidth  = static_cast<int>( xBitData.GetWindowWidth() );
	int iHeight = xBitData.GetWindowHeight();
	int iPitch  = ((iWidth*3)+3)&~3;

	// <-- fill out header -->
	// header code copied from Rendering/Asura_Screenshot.cpp - NDB 18/09/07
	xBmpHeader.Header.Type      = 'B'+'M'*256;
	xBmpHeader.Header.Reserved1 = 0;
	xBmpHeader.Header.Reserved2 = 0;
	xBmpHeader.Header.Offset    = 14+40+0;

	// The type of information found in a BMP structure is indicated by
	// the Size (Information Header Size) field with a non-zero value.
	xBmpHeader.PmInfo.Size   = 0;
	xBmpHeader.Pm2Info.Size  = 0;

	xBmpHeader.WinInfo.Size          = 40;
	xBmpHeader.WinInfo.Width         = iWidth;
	xBmpHeader.WinInfo.Height        = iHeight;
	xBmpHeader.WinInfo.Planes        = 1;
	xBmpHeader.WinInfo.BitCount      = 24;
	xBmpHeader.WinInfo.Compression   = 0;
	xBmpHeader.WinInfo.SizeImage     = iPitch*iHeight;
	xBmpHeader.WinInfo.XPelsPerMeter = xBmpHeader.WinInfo.Width;
	xBmpHeader.WinInfo.YPelsPerMeter = xBmpHeader.WinInfo.Height;
	xBmpHeader.WinInfo.ClrUsed       = 0;
	xBmpHeader.WinInfo.ClrImportant  = 0;

	xBmpHeader.Header.FileSize  = xBmpHeader.WinInfo.SizeImage + xBmpHeader.Header.Offset + 8;

	// write header
    Tools_TT_PutLittleWord(xBmpHeader.Header.Type, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.Header.FileSize, pFile);
    Tools_TT_PutLittleWord(xBmpHeader.Header.Reserved1, pFile);
    Tools_TT_PutLittleWord(xBmpHeader.Header.Reserved2, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.Header.Offset, pFile);

	Tools_TT_PutLittleDword(xBmpHeader.WinInfo.Size, pFile);

    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.Width, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.Height, pFile);
    Tools_TT_PutLittleWord(xBmpHeader.WinInfo.Planes, pFile);
    Tools_TT_PutLittleWord(xBmpHeader.WinInfo.BitCount, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.Compression, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.SizeImage, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.XPelsPerMeter, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.YPelsPerMeter, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.ClrUsed, pFile);
    Tools_TT_PutLittleDword(xBmpHeader.WinInfo.ClrImportant, pFile);
	
	// <-- write 24 bit image, RGB -->

	//allocate buffer for RGB image (no alpha chanel so 3 bytes rather than 4 per pixel)
	const u_int uBufferSize( xBitData.GetWindowSize() );
	char*   pcRtnBuffer( new char[uBufferSize] );
	u_char*	pcSrcBuffer( new u_char[uBufferSize] );

	xBitData.GetWindowData( pcSrcBuffer,uBufferSize );

	u_int uRtnIndex(0);
	u_int uSrcIndex(0);
	
	// iterate over src and rtn arrays
	for( int iY(0);iY<iHeight;++iY )
	{		
		// bmps are inverted in the y direction, must copy the last line of the 
		// src buffer to the first line of the rtn buffer 
		uSrcIndex = 4*iWidth*(iHeight-iY-1);

		for( int iX(0);iX<iWidth;++iX )
		{
			// copy red
			pcRtnBuffer[uRtnIndex] = pcSrcBuffer[uSrcIndex];
			++uRtnIndex;
			++uSrcIndex;
			
			// copy green
			pcRtnBuffer[uRtnIndex] = pcSrcBuffer[uSrcIndex];
			++uRtnIndex;
			++uSrcIndex;
			
			// copy blue
			pcRtnBuffer[uRtnIndex] = pcSrcBuffer[uSrcIndex];
			++uRtnIndex;
			++uSrcIndex;
		
			// skip alpha 
			++uSrcIndex;
		}
	}

	fwrite( pcRtnBuffer,1,uBufferSize,pFile );
	
	delete [] pcSrcBuffer;
	delete [] pcRtnBuffer;

	fclose(pFile);

	return true;	
}

const bool Tools_TT_Export_SaveTM2( Tools_TT_Export_RawPixelData& xBitData, const Asura_String& xOutputFileName )
{		
	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}

	TextureConversionData xConversionData;
	xConversionData.Initialise( Tools_Level_Export::Export_PS2 );

	xConversionData.m_iTextureFlags		  = 0;
	xConversionData.m_bSwizzled			  = true;
	xConversionData.m_uMaterialFlags      = 0;
	xConversionData.m_bUnpalettised		  = false;
	xConversionData.m_uMaxMipDrops		  = 2; 
	xConversionData.m_bUseFixedPalette	  = false;
	xConversionData.m_bWritePaletteAsRedScaleIndices = false;
	xConversionData.m_bDisableDithering	  = false;
	xConversionData.m_uTextureUsageFlags  = 0;
	xConversionData.m_uMaximumTextureSize = 0;

	// apply texture defined in bit data flags
	xConversionData.m_iTextureFlags |= xBitData.GetTextureFlags();

	if( xBitData.GetTextureFlags() & AsuraTextureFlag_16ColourPalette )
	{
		xConversionData.m_iPixelFormat		= TIM2_IDTEX4;
	}


	u_char* pcData( reinterpret_cast<u_char*>( Asura_Allocate( sizeof(u_char) * xBitData.GetWindowSize() ) ) );
	xBitData.GetWindowData( pcData,xBitData.GetWindowSize() );

	Tools_Tim2_Writer	xTim2Writer;
	xTim2Writer.Initialise( xBitData.GetWindowWidth(), xBitData.GetWindowHeight(), xBitData.GetWindowPitch(), pcData, xConversionData );
	xTim2Writer.WriteOutFile( xOutputFileName.GetString(), 0, 0 );

	Asura_Deallocate( pcData );

	return true;
}

const bool Tools_TT_Export_SaveTGA( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName,const bool bUseRLE /*= true*/ )
{
	// Save tga to file. Cut+Pasted code from fontmaker. Should this be moved into Tools/ImageProcessing ? - NDB
    Tools_TT_TargaHeader xHeader;

    // Create the file
	FILE* pxFile = fopen( xOutputFileName.GetString(), "wb" );

    if( pxFile == NULL )
    {
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR failed to open file \"%s\"",xOutputFileName.GetString() );
        return false;
    }

    // Write the TGA header
    ZeroMemory( &xHeader, sizeof(Tools_TT_TargaHeader) );
    xHeader.IDLength        = 0;
    xHeader.ImageWidth      = static_cast<WORD>(xBitData.GetWindowWidth());
    xHeader.ImageHeight     = static_cast<WORD>(xBitData.GetWindowHeight());
	xHeader.PixelDepth      = static_cast<BYTE>(xBitData.GetBytesPP() * 8);
    
	if( xBitData.GetBytesPP() == 1 )
	{
		xHeader.ImageType       = 3;    // monochrome image
		xHeader.ImageDescriptor = 0x28; // origin in top left corner, no atribute bits
	}
	else
	{
		xHeader.ImageType       = 2;    // truecolour image
		xHeader.ImageDescriptor = 0x28; // origin in top left corner, one atribute bit (alpha channel)
	}

    fwrite( &xHeader, sizeof(Tools_TT_TargaHeader), 1, pxFile );

 //   // Write the pixels
	//if( bUseRLE )
	//{

	//	u_int uCounter(0);

	//	Tools_TT_Export_Pixel xCurrentPixel;
	//	Tools_TT_Export_Pixel xLastPixel;
	//	
	//	u_int uPixel(0)

	//	xLastTexel.uA = xBitData[uPixel++];
	//	xLastTexel.uR = xBitData[uPixel++];
	//	xLastTexel.uG = xBitData[uPixel++];
	//	xLastTexel.uB = xBitData[uPixel++];

	//	while( uPixel<xBitData.GetWindowSize() )
	//	{
	//		xCurrentPixel.uA = xBitData[uPixel++];
	//		xCurrentPixel.uR = xBitData[uPixel++];
	//		xCurrentPixel.uG = xBitData[uPixel++];
	//		xCurrentPixel.uB = xBitData[uPixel++];
	//
	//		if( xCurrentPixel == xLastPixel )
	//		{
	//			
	//		}
	//	}
	//
	//}
	//else
	if( xBitData.GetWindowSize() == xBitData.GetSize() )
	{	
		fwrite( xBitData.GetData(),sizeof(u_char),xBitData.GetWindowSize(),pxFile );
	}
	else
	{
		// copy subset of texels from data 
		u_char* pcData( new u_char[xBitData.GetWindowSize()] );
		xBitData.GetWindowData( pcData,xBitData.GetWindowSize() );
		fwrite( pcData,sizeof(u_char),xBitData.GetWindowSize(),pxFile );
		delete [] pcData;
	}

    // Close the file and return okay
    fclose( pxFile );

    return true;
}

const bool Tools_TT_Export_SaveTPL( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName,const Tools_TT_TPLAlphaBits::TOOLS_TT_EXPORT_TPL_ALPHA_BITS eAlphaBits /* = 0 */ )
{
	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}

	bool bSuccess(false);
	const u_int uTPLAlphaBits( Tools_TT_TPLAlphaBits::GetValue( eAlphaBits ) );

	// create temporary TGA
	char* pcTemporaryDirectoryName( reinterpret_cast<char*>( Asura_Allocate( L_tmpnam * sizeof(char) ) ) );
	Asura_String xTGAName( tmpnam( pcTemporaryDirectoryName ) );
	Asura_Deallocate( pcTemporaryDirectoryName );
	xTGAName += "_temporary_file.tga";

	bSuccess = Tools_TT_Export_SaveTGA( xBitData,xTGAName );

	if( !bSuccess )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR Failed to create temporary TGA file" );
		return false;
	}
	
	// Saved TGA. Now convert to TPL
	HINSTANCE xDLLHandle( LoadLibrary("TPLWriterDLL.dll") );
	
	if(!xDLLHandle)
	{	
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR Could not open TPLWriterDLL.dll" );
		return false;
	}

	CREATETPL_PTR   pfCreateTPL( reinterpret_cast<CREATETPL_PTR>( GetProcAddress( xDLLHandle,"CreateTPL" ) ) );
	GETERRORMSG_PTR pfGetErrorMsg( reinterpret_cast<GETERRORMSG_PTR>( GetProcAddress( xDLLHandle,"GetErrorMessage" ) ) );
	
	if(pfCreateTPL && pfGetErrorMsg)
	{
		TPLOptions xTPLOptions;
		memset(&xTPLOptions,0,sizeof(xTPLOptions));
		xTPLOptions.m_uMinLOD    = 0;
		xTPLOptions.m_uMaxLOD    = 0;
		xTPLOptions.m_uBaseLOD   = 0;
		xTPLOptions.m_uAlphaBits = uTPLAlphaBits;
		xTPLOptions.m_bUseS3TC   = 1;
		xTPLOptions.m_bUseWrap   = 0;
		bSuccess = (pfCreateTPL)(xTGAName.GetString(), xOutputFileName.GetString(), &xTPLOptions);
		
		if( !bSuccess )
		{
			Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR TPL export failed with error: %s",(pfGetErrorMsg)() );
		}	
        
	}

	// delete temporary TGA file
	DeleteFile( xTGAName.GetString() );

	return bSuccess;
}

static NV_ERROR_CODE Tools_TT_Export_SaveDDSDataCallback( const void* pvData, size_t uByteCount, const MIPMapData* pxMipMapData, void* pvUserData )
{
	NV_ERROR_CODE xRtn( NV_OK );

	if( fwrite( static_cast<const unsigned char*>(pvData),1,uByteCount,static_cast<FILE*>(pvUserData) ) != uByteCount )
	{	
        xRtn = NV_FAIL;
    }
    
	return xRtn;
}

const bool Tools_TT_Export_SaveDDS( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName,const Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT eCompresionFormat /* = Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_UNCOMPRESSED */ )
{
	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}

	// can only make dds out of texture whose dimensions are a power of 2
	if( !Tools_TT_IsPowerOf2(xBitData.GetWindowWidth()) | !Tools_TT_IsPowerOf2(xBitData.GetWindowHeight()) )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"WARNING Dimensions are not a power of 2, DDS creation will probably fail for %s",xOutputFileName.GetString() );
	}

    nvCompressionOptions xCompressionOptions;

    // Prepare options for nvDXT.
    xCompressionOptions.SetDefaultOptions();
    xCompressionOptions.mipMapGeneration = kNoMipMaps;

    // Apply format option.
    if( eCompresionFormat == Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_UNCOMPRESSED )
    {
        xCompressionOptions.textureFormat = k8888;
    }
    else if( eCompresionFormat == Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_DXT1 )
    {
        xCompressionOptions.textureFormat = kDXT1a;
    }
    else if( eCompresionFormat == Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_DXT3 )
    {
        xCompressionOptions.textureFormat = kDXT3;
    }
    else if( eCompresionFormat == Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_DXT5 )
    {
        xCompressionOptions.textureFormat = kDXT5;
    }
    else
    {
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"WARNING Unrecognised DDS format for file %s (skipping compression)",xOutputFileName.GetString() );
		
		xCompressionOptions.textureFormat = k8888;
    }

    // Open output file for writing (w) in binary mode (b) - without the 'b' linefeed-carriage return characters get mangled - NDB
	FILE* pxOutputFile = fopen( xOutputFileName.GetString(),"wb" );
	
	if( pxOutputFile == NULL )
    {
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR Failed to open output file %s",xOutputFileName.GetString() );
        return false;
    }

    xCompressionOptions.user_data = pxOutputFile;

	u_char* pcData( new u_char[ xBitData.GetWindowSize() ] );
	xBitData.GetWindowData( pcData,xBitData.GetWindowSize() );

	HRESULT iResult = Asura_PC_CompressDXT( pcData,
											xBitData.GetWindowWidth(),
											xBitData.GetWindowHeight(),
											xBitData.GetWindowPitch(),
											nvBGRA,
											&xCompressionOptions,
											Tools_TT_Export_SaveDDSDataCallback,
											NULL );

	delete [] pcData;

    // Clean up
    fclose( pxOutputFile );

    if( FAILED(iResult) )
    {
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR NVidia DXT compression failed for %s",xOutputFileName.GetString() );
        return false;
    }
    
	return true;
}

const bool Tools_TT_Export_SavePSPTx( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName )
{
	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}

	// PSPTx write expects 32 bit ARGB
	TextureConversionData xConversionData;
	xConversionData.Initialise( Tools_Level_Export::Export_PSP );
	xConversionData.m_iTextureFlags &= ~AsuraTextureFlag_HalfForPSP;

	xConversionData.m_bSwizzled			  = true;
	xConversionData.m_uMaterialFlags      = 0;
	xConversionData.m_bUnpalettised		  = false;
	xConversionData.m_iPixelFormat		  = 0;
	xConversionData.m_iPaletteFormat	  = 0;
	xConversionData.m_uMaxMipDrops		  = 2; 
	xConversionData.m_bUseFixedPalette	  = false;
	xConversionData.m_bWritePaletteAsRedScaleIndices = false;
	xConversionData.m_bDisableDithering	  = false;
	xConversionData.m_uTextureUsageFlags  = 0;
	xConversionData.m_uMaximumTextureSize = 0;

	// apply texture defined in bit data flags
	xConversionData.m_iTextureFlags |= xBitData.GetTextureFlags();

	u_char* pcData( reinterpret_cast<u_char*>( Asura_Allocate( sizeof(u_char) * xBitData.GetWindowSize() ) ) );
	xBitData.GetWindowData( pcData,xBitData.GetWindowSize() );

	Tools_PSP_Texture_Writer xWriter;
	xWriter.Initialise( 0,xBitData.GetWindowWidth(),xBitData.GetWindowHeight(),xBitData.GetWindowPitch(),pcData );
	xWriter.WriteOutFile( xOutputFileName.GetString(),xConversionData );

	Asura_Deallocate( pcData );

	return true;
}


const bool Tools_TT_Export_SavePS3Tx( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName )
{
	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}

	// PSPTx write expects 32 bit ARGB
	TextureConversionData xConversionData;
	xConversionData.Initialise( Tools_Level_Export::Export_PS3 );

	bool bSuccess(false);

	// apply texture defined in bit data flags
	xConversionData.m_iTextureFlags |= xBitData.GetTextureFlags();

	u_char* pcData( reinterpret_cast<u_char*>( Asura_Allocate( sizeof(u_char) * xBitData.GetWindowSize() ) ) );
	xBitData.GetWindowData( pcData,xBitData.GetWindowSize() );

	Tools_PS3_Texture_Writer xWriter( xBitData.GetWindowWidth(),xBitData.GetWindowHeight(),xBitData.GetWindowPitch(),pcData,&xConversionData,kDXT5,bSuccess );
	xWriter.WriteOutFile( xOutputFileName.GetString() );

	Asura_Deallocate( pcData );

	return bSuccess;
}

const bool Tools_TT_Export_SaveXPR( Tools_TT_Export_RawPixelData& xBitData,const Asura_String& xOutputFileName,const Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT eCompressionFormat /* = Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_UNCOMPRESSED */ )
{
	if( xBitData.GetBytesPP() < 4 )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit support not yet implimented" );
		return false;
	}

	// In order to produce an xpr file using Tools_Xbox360_Texture_Writer a TGA 
	// file (other sorts of textures might do just as well) is generated, passed
	// to the texture writer, then deleted.

	// Output a .tga first 
	Asura_String xTGAName;
	xOutputFileName.SplitFilename( NULL, &xTGAName, NULL );
	xTGAName += ".tga";

	Tools_TT_Export_SaveTGA( xBitData,xTGAName );

	// Create the RDF and XPR
	Tools_Xbox360_RDF xRDF;
	Tools_Xbox360_Texture_Writer::BeginTextureExport( xRDF, xTGAName.GetString() );
	Tools_Xbox360_Texture_Writer::AddTextureToRDF( xRDF, xTGAName.GetString(), xTGAName.GetString(), Tools_TT_DDSCompressionFormat::GetName( eCompressionFormat ), 1 );
	
	Tools_Xbox360_Texture_Writer::EndTextureExport( xRDF );
	Tools_Xbox360_Texture_Writer::RunBundler( xRDF );

	DeleteFile( xRDF.m_xPath.GetString() );

	// Delete the TGA
	DeleteFile( xTGAName.GetString() );

	return true;
}


const bool Tools_TT_Export_ExtractTexturesFromTTL( Tools_TT_Export_RawPixelData& xBitData,const Asura_TiledTextureInfo* pxTextureInfoCollection,const int iSize,const Tools_TT_FileType::TOOLS_TT_FILE_NAME_EXTENSION eFileType /* = Tools_TT_FileType::TOOLS_TT_TGA */ )
{
	bool bSuccess(false);

	for( int i(0);i<iSize;++i )
	{
		xBitData.SetWindow( static_cast<u_int>(pxTextureInfoCollection[i].m_fu1*pxTextureInfoCollection[i].m_TiledToWidth),
							static_cast<u_int>(pxTextureInfoCollection[i].m_fv1*pxTextureInfoCollection[i].m_TiledToHeight),
							static_cast<u_int>(pxTextureInfoCollection[i].m_fu2*pxTextureInfoCollection[i].m_TiledToWidth),
							static_cast<u_int>(pxTextureInfoCollection[i].m_fv2*pxTextureInfoCollection[i].m_TiledToHeight) );

		bSuccess = Tools_TT_Export_SaveTGA( xBitData,pxTextureInfoCollection[i].m_xNameOfTextureUsed );
	}

	xBitData.SetWindowToMax();

	return bSuccess;
}

const Asura_String Tools_TT_GetGraphicsPath()
{
	Asura_String xRtn( Asura_FileAccess::PlatformGetAsuraFilesDir() );
	xRtn += "\\Graphics\\"; 
	return xRtn;
}

const Asura_String& Tools_TT_PrefixWithGraphicsPath( Asura_String& xPath )
{
	Asura_String xNewPath( Tools_TT_GetGraphicsPath() );
	
	const Asura_Char* pxPathBuffer( xPath.GetString() );

	// skip any preceeding '\\' in path
	if( pxPathBuffer[0] == '\\' )
	{
		++pxPathBuffer;
	}

	xNewPath += pxPathBuffer;

	Tools_TT_Swap( xPath,xNewPath );

	return xPath;
}


const Asura_String& Tools_TT_ReplaceFileExtension( Asura_String& xPath,const Asura_String& xReplacement )
{
	char  acPath[256];
	strncpy( acPath,xPath.GetString(),256 );
	char* pcLocationOfSlashInPath( strrchr( acPath,'\\' ) );
	char* pcLocationOfDotInPath( strrchr( acPath,'.' ) );
	
	// if there is a dot in the path and it appears after any slashes then remove all text after dot
	if( pcLocationOfDotInPath && 
		( (!pcLocationOfSlashInPath) || ( pcLocationOfSlashInPath && (pcLocationOfDotInPath>pcLocationOfSlashInPath) ) ) )
	{	
		*pcLocationOfDotInPath = '\0';
		xPath = acPath;
	}

	xPath += xReplacement;
	
	return xPath;
}


const Asura_String Tools_TT_CreateExportPathForPlatform( const Asura_String& xPath,const int iPlatformType )
{
	// Changed export path from "Platform name"_Release to Graphics_"Platform name" - NDB 270508

	Asura_String xPlatformExportDirectory( "Graphics_" );
	xPlatformExportDirectory += Tools_Platform::GetPlatformName( iPlatformType );

	Asura_String xLowerCasePath( xPath );
	xLowerCasePath.MakeLowercase();

	const Asura_Char* pcPlatformExportPath( Asura_StringUtil::Replace( xLowerCasePath.GetString(),"graphics",xPlatformExportDirectory.GetString() ) );

	Asura_String xRtn( pcPlatformExportPath );
	
	//Add BE to extension for big endian platforms. RAR 18/03/2008
	if( Tools_Platform::IsBigEndian( iPlatformType) )
	{
		xRtn += ASURA_STANDARD_FILE_EXTENSION_BE_SUFFIX;
	}

	delete [] pcPlatformExportPath;

	return xRtn;
}

void Tools_TT_CreateTiledTextureForPlatform( const Asura_String& xFileName, const int iPlatformType )
{
	const Asura_String xPlatformExportPath( Tools_TT_CreateExportPathForPlatform( xFileName,iPlatformType ) );
	
	if( !Asura_FileAccess::FileExists( xPlatformExportPath.GetString() ) )
	{
		Tools_FileSystem::EnsureDirectoryExists( xPlatformExportPath.GetString() );
		
		Tools_TT_TiledTextureList xTTL;		
		xTTL.ReadChunk( xFileName,true );		
		
		
		if( xTTL.WriteChunk( xPlatformExportPath, static_cast<Tools_Platform_Type>( iPlatformType ) ) )
		{
			Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Green,"Created Tiled Texture file \"%s\" for %s platform.",xPlatformExportPath.GetString(),Tools_Level_Export::GetPlatformName( iPlatformType ) );
		}
		else
		{
			Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR Failed to create Tiled Texture file \"%s\" for %s platform",xPlatformExportPath.GetString(),Tools_Level_Export::GetPlatformName( iPlatformType ) );
		}

		// Fix up texture indexs in v1 tiles (this happens on load for v2 tiles) - NDB
		Asura_TiledTextureManagement::SetTextureHandlesFromHash();
	}
	else
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR Attempting to create Tiled Texture file \"%s\" for %s platform but it already exists.",xPlatformExportPath.GetString(),Tools_Level_Export::GetPlatformName( iPlatformType ) );
	}
}


const Asura_String Tools_TT_CreateTextureFileName( const Asura_String& xTileFileName )
{
	Asura_String xRtn( xTileFileName );

	Tools_TT_ReplaceFileExtension( xRtn,"_texture.tga" );

	return xRtn;
}

const int Tools_TT_FindNextLargerPowerOfTwo( const int iArg )
{
	int i(1);
	
	// loop over all powers of two untill one is found that is greater than arg or i wraps round to zero - NDB 111007
	while( i && (i<iArg) )
	{
		i<<=1;
	}

	return i;
}

// <-- class GraphicsFileExists -->
const bool GraphicsFileExists::operator()( const char* szFileName ) const
{
	Asura_String xTexturePath( szFileName );
	Tools_TT_PrefixWithGraphicsPath( xTexturePath );

	return Asura_FileAccess::FileExists( xTexturePath.GetString() );
}

const bool GraphicsFileExists::operator()( const Asura_TiledTextureInfo& xInfo ) const
{
	return this->operator()( xInfo.m_xNameOfTextureUsed.GetString() );
}

const Asura_String&	Tools_TT_IncrementFileName( Asura_String& xNextOutputFileName )
{
	// seperate path into constituent parts
	Asura_String xPath;
	Asura_String xName;
	Asura_String xExt;

	xNextOutputFileName.SplitFilename( &xPath,&xName,&xExt );

	// discover position of last non-numerical character in file name
	char* pszNameReverseBegin( xName.GetString() + xName.GetLength() - 1);
	char* pszNameReverseEnd( xName.GetString() );
	char* pszNameReverseItr( pszNameReverseBegin );

	for( ;(pszNameReverseItr!=pszNameReverseEnd) && (*pszNameReverseItr >= '0') && (*pszNameReverseItr <= '9');--pszNameReverseItr );

	// if file name does not end in a number then append a '0', else increment the number
	if( pszNameReverseItr == pszNameReverseBegin )
	{
		// <-- append a number 0 to the file name -->
		xName += "_0";
	}
	else
	{
		// <-- increment the number at the end of the file name -->
		
		// move itr forward to first digit of number at end of file name 
		++pszNameReverseItr;
		
		// recover number and increment it
		int iFileNumber(0);
		Asura_StringUtil::GetIntValue( pszNameReverseItr,&iFileNumber );
		++iFileNumber;

		// strip old number from end of file name
		xName.Truncate( xName.GetLength() - strlen(pszNameReverseItr) );
		Asura_String xFileNumber;
		Asura_StringUtil::ToString( iFileNumber,xFileNumber );

		// append new number to end of file name
		xName += xFileNumber;
	}

	// rebuild full path to file
	xNextOutputFileName.MakeBlank();
	xNextOutputFileName += xPath;
	xNextOutputFileName += "\\";
	xNextOutputFileName += xName;
	xNextOutputFileName += ".";
	xNextOutputFileName += xExt;

	return xNextOutputFileName;
}
