
#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_TEXTURETILER_RENDER_H_
#define __TOOLS_TEXTURETILER_RENDER_H_

#include "TextureTiler/Tools_TextureTiler_TTL.h"

#include "Rendering/Asura_RenderStates.h"

class Asura_List_Dialog;
class Asura_DropDown_List_Dialog;
class Asura_Numerical_Dialog;

// String Container Class
//
// Holds a collection of c-style strings. strings can be copied into the 
// container using any of the AssignStringCollection() member functions.
// The container is automatically resized on an assign.
class Tools_TT_StringContainer
{
public:
	Tools_TT_StringContainer( const u_int uStringSize = 128 );
	virtual ~Tools_TT_StringContainer();

	// set capacity of container,
	void SetCapacity( const u_int uCapacity );

	const u_int GetCapacity() const {return m_uStringCollectionCapacity;}

	// add a new string to the end of the collection
	const bool Append( const char* szString );

	// delete all strings and set pointers in collection to NULL
	void Clear();

	// get string at a specified position
	const char* const operator[]( const u_int uIndex ) const
	{
		return m_ppszStringCollection[uIndex];
	}

	const u_int Size() const
	{
		return m_uStringCollectionSize;
	}

protected:

	// returns pointer to string container - a grim prospect but nessary 
	// considering the interface for Asura_List_Dialog and the like - NDB
	char** const GetData() const
	{
		return m_ppszStringCollection;
	}

	// deletes the existing container and all the strings it held, then 
	// new's a container of size uSize of new'd strings
	virtual void ReplaceContainer( const u_int uSize );

private:
	// private member data
	char** m_ppszStringCollection;	   // collection of strings to be displayed in list
	
	u_int m_uStringSize;               // size of each array within string collection
	u_int m_uStringCollectionSize;     // number of strings in collection
	u_int m_uStringCollectionCapacity; // max number of strings the collection can hold
};

// Overlay List Dialog Wrapper Class
//
// class manages an Asura_List_Dialog object. Correct usage is to use an 
// AssignStringCollection() member method ( inherited from StringContainer 
// class), then call CreateAsuraDialog(). Subsequent calls to 
// AssignStringCollection() then UpdateDialog() should be made when the list 
// contents (the strings) changes.
//
template< typename TListDialog > class Tools_TT_OverlayListDialog : public Tools_TT_StringContainer
{
public:
	Tools_TT_OverlayListDialog( const u_int uNumberOfLines = 10 ):
		m_pxListDialog( NULL ),	
		m_iSelectionIndex( 0 ),
		m_iNumberOfLines( uNumberOfLines ),
		m_bDialogCreated(false)
	{}

	// create an Asura_List_Dialog from the properties held in this. Function canot be called 
	// CreateDialog() because somewhere there exists a macro with that name and scoping doesn't 
	// apply to macros so theres no way to avoid a conflict - NDB 
	TListDialog* CreateAsuraDialog( const Asura_Vector_2& xOffset, void(*pxAction) (void) )
		{
			MakeSelectionIndexValid();

			m_pxListDialog = new TListDialog( 
								xOffset,
								this->GetData(),
								&m_iSelectionIndex,
								m_iNumberOfLines,
								this->Size(),
								pxAction
								);

			m_bDialogCreated = true;

			return m_pxListDialog;
		}

	const u_int GetSelectionIndex() const
	{
		return m_iSelectionIndex;
	}

	void SetSelectionIndex( const u_int uIndex )
	{
		if( (uIndex>=0) && (uIndex<Size()) )
		{
			m_iSelectionIndex = uIndex;
		}
	}

	const char* const GetSelectedString() const
	{
		return this->operator[]( GetSelectionIndex() );
	}

	// update the properties of the dialog to match those held in this
	void UpdateDialog()
	{
		if(m_bDialogCreated)
		{
			MakeSelectionIndexValid();

			m_pxListDialog->ReplaceStrings( this->Size(),this->GetData() );
		}
	}

private:
	typedef Tools_TT_StringContainer PARENT;

	void MakeSelectionIndexValid()
	{
		if( !(GetSelectionIndex()<Size()) )
		{
			SetSelectionIndex( Size() - 1 );
		}
		if( GetSelectionIndex()<0 )
		{
			SetSelectionIndex( 0 );
		}
	}

	virtual void ReplaceContainer( const u_int uSize )
	{
		if( !(GetSelectionIndex() < uSize) )
		{
			SetSelectionIndex(uSize-1);
		}

		if( GetSelectionIndex() < 0 )
		{
			SetSelectionIndex(0);
		}
		
		PARENT::ReplaceContainer( uSize );
	}

	TListDialog* m_pxListDialog;

	int m_iSelectionIndex;  // type forced to be int (rather than u_int) by interface to Asura_ListDialog
	int m_iNumberOfLines;
	bool m_bDialogCreated;
};

template<typename T> void Tools_TT_AssignStringCollection( Tools_TT_OverlayListDialog<T>& xOverlayListDialog,const char** ppcString, const u_int uSize )
{
	xOverlayListDialog.SetCapacity( uSize );

	for( u_int u(0);u<uSize;++u )
	{
		xOverlayListDialog.Append( ppcString[u] );
	}
}

template<typename T> void Tools_TT_AssignStringCollection( Tools_TT_OverlayListDialog<T>& xOverlayListDialog,const Asura_String** ppxString, const u_int uSize )
{
	xOverlayListDialog.SetCapacity( uSize );

	for( u_int u(0);u<uSize;++u )
	{
		xOverlayListDialog.Append( ppxString[u]->GetString() );
	}
}

// copy tile names into list dialog, prefixing those that have not been tiled with the string "[Fail] "
template<typename T> void Tools_TT_AssignStringCollection( Tools_TT_OverlayListDialog<T>& xOverlayListDialog, const Tools_TT_TiledTextureList& xTTL )
{	
	xOverlayListDialog.SetCapacity( xTTL.Size() );

	Asura_String xString;

	for( u_int u(0);u<xTTL.Size();++u )
	{
		if( xTTL[u].IsTiled() )
		{
			xOverlayListDialog.Append( xTTL[u].GetName().GetString() );
		}
		else
		{	
			xString  = "[FAIL] ";
			xString += xTTL[u].GetName();

			xOverlayListDialog.Append( xString.GetString() );
		}
	}
}



// Asura_NumericalDialog Wrapper Class
//
// handles the creation of and interaction with Asura_NumericalDialog objects.
class Tools_TT_OverlayNumericalDialog
{
public:
	Tools_TT_OverlayNumericalDialog( const int iValue = 0 );

	Asura_Numerical_Dialog* CreateAsuraDialog( const char* pcName, const Asura_Vector_2& xOffset, int iMin, int iMax,int iDelta, const char* pcPrintFormat );

	const bool IsChanged() const
	{
		return m_bChanged;
	}

	const int GetValue() const
	{
		return m_iValue;
	}

	void SetValue( const int iValue )
	{
		m_iValue = iValue;
		m_bChanged = true;
	}

private:
	Asura_Numerical_Dialog* m_pxDialog;

	int  m_iValue;
	bool m_bChanged;
};

// Confirmation Overlay
//
// creates an overlay in the center of the screen. The overlay contains some text, a 
// yes button and a no button. When the user clicks the yes button the method pointed 
// to by s_pxAction is called, followed by the method pointed to by s_pxOnExit. If the
// no button is clicked then the method pointed to by s_pxOnExit is called. The ovelay 
// is closed afterwards in either case.
class Tools_TT_ConfirmationOverlay
{
public:
	// returns true when confirmation window is open
	const bool IsActive() const
	{
		return s_bIsActive;
	}

	void Create( const Asura_String& xText, void(*pxAction) (void), void(*pxOnExit) (void) = NULL );

	static void Yes();
	static void No();

private:

	static void Exit();

	static void(*s_pxAction) (void);
	static void(*s_pxOnExit) (void);
	static bool s_bIsActive;
};

// Overlay Status
//
// Book keeping for asura overlays. Records position, minimised status and 
// whether the overlay exists.
class Tools_TT_OverlayStatus
{
public:
	Tools_TT_OverlayStatus();

	void SetMinimised( const bool bMinimised )
	{
		m_bMinimised = bMinimised;
	}

	bool* GetPtrToMinimised()
	{
		return &m_bMinimised;
	}

	const bool IsMinimised() const
	{
		return m_bMinimised;
	}

	void SetPosition( const float fX,const float fY )
	{
		m_xPosition.x = fX;
		m_xPosition.y = fY;
	}

	Asura_Vector_2* GetPtrToPosition()
	{
		return &m_xPosition;
	}

	void SetCreated( const bool bCreated )
	{
		m_bCreated = bCreated;
	}

	bool& GetRefToCreated()
	{
		return m_bCreated;
	}

	bool IsCreated() const
	{
		return m_bCreated;
	}

	const bool GetRequested() const
	{
		return m_bRequested;
	}

	void SetRequested( const bool bRequested )
	{
		m_bRequested = bRequested;
	}

private:
	Asura_Vector_2 m_xPosition;
	bool m_bRequested; 			// true if the overlay should be created
	bool m_bCreated;   			// true if the overlay exists/has been created
	bool m_bMinimised;
};

// describes the position and size of a part of the user interface. This is only used by the current TTL atm - NDB 101007
class Tools_TT_UIComponent
{
public:


	void SetPadding( const float fPadding )
	{
		m_fPadding = fPadding;
	}	

	const float GetPadding() const
	{
		return m_fPadding;
	}

	void SetBorderColour( const u_int uColour )
	{
		m_uBorderColour = uColour;
	}

	const u_int GetBorderColour() const
	{
		return m_uBorderColour;
	}

	const float GetWidth() const
	{
		return m_xRectangle.GetWidth();
	}
	
	const float GetInnerWidth() const
	{
		return m_xRectangle.GetWidth() - 2.0f*GetPadding();
	}

	const float GetHeight() const
	{
		return m_xRectangle.GetHeight();
	}

	const float GetInnerHeight() const
	{
		return m_xRectangle.GetHeight() - 2.0f*GetPadding();
	}

	void SetOuterRectangle( const Asura_Bounding_Rectangle& xRectangle )
	{
		m_xRectangle = xRectangle;
	}
	
	void SetOuterRectangle( const float fMinX,const float fMinY,const float fMaxX,const float fMaxY )
	{
		m_xRectangle.Set( fMinX,fMinY,fMaxX,fMaxY );
	}

	const Asura_Bounding_Rectangle& GetOuterRectangle() const
	{
		return m_xRectangle;
	}

	const Asura_Bounding_Rectangle GetInnerRectangle() const
	{
		Asura_Bounding_Rectangle xRtn(m_xRectangle);

		xRtn.MinX += GetPadding();
		xRtn.MinY += GetPadding();
		xRtn.MaxX -= GetPadding();
		xRtn.MaxY -= GetPadding();

		return xRtn;
	}

private:
	Asura_Bounding_Rectangle m_xRectangle;
	float m_fPadding;
	u_int m_uBorderColour;
};

// returns true if mouse pointer is inside ui component provided
inline const bool Tools_TT_IsPointInsideUIComponent( const int iX, const int iY,const Tools_TT_UIComponent& xComponent )
{
	return Tools_TT_PointInRectangle( iX,iY,xComponent.GetOuterRectangle() );
}


template< typename Ta,typename Tb > 
inline const Asura_Vector_2 GetScaleFactor( const Ta& xNumerator,const Tb& xDenominator )
{	
	Asura_Vector_2 xScaleFactor( 0.0f,0.0f );

	if( xDenominator.GetWidth() && xDenominator.GetHeight() )
	{
		xScaleFactor.x = xNumerator.GetWidth()/xDenominator.GetWidth();
		xScaleFactor.y = xNumerator.GetHeight()/xDenominator.GetHeight();
	}

	return xScaleFactor;
}

// This method adjusts the dimensions of the bounding rectangle provided as the 1st argument so that it has
// the same aspect size and aspect ratio as the object passed as the 2nd argument. 
template< typename T > 
inline const Asura_Vector_2 Tools_TT_ScaleRectangleToFixAspectRatio( Asura_Bounding_Rectangle& xRectangle,const T& xTarget )
{
	Asura_Vector_2 xScaleFactor( GetScaleFactor(xRectangle,xTarget) );
	float fTrim;
	
	// make both components of scale factor equal to the smallest component
	// and adjust position of TTL's top left corner to ensure that it gets
	// drawn in the center
	if(xScaleFactor.x > xScaleFactor.y)
	{
		xScaleFactor.x = xScaleFactor.y;
		
		fTrim = 0.5f * (xRectangle.GetWidth() - (xScaleFactor.x * xTarget.GetWidth()) );
		xRectangle.MinX += fTrim;
		xRectangle.MaxX -= fTrim;
	}
	else
	{
		xScaleFactor.y = xScaleFactor.x;
		
		fTrim = 0.5f * (xRectangle.GetHeight() - (xScaleFactor.y * xTarget.GetHeight()) );
		xRectangle.MinY += fTrim;
		xRectangle.MaxY -= fTrim;
	}

	return xScaleFactor;
}


// Render texture using the position and size supplied
void Tools_TT_RenderTexture( const Asura_Vector_2& xTopLeft,const Asura_Vector_2& xBottomRight,const int iTextureID,const int iBackgroundColour = Tools_TT_Colour::Null,const ASURA_TRANSLUCENCY_MODE eTranslucency = ASURA_TRANSLUCENCY_OFF );
void Tools_TT_RenderTexture( const Asura_Bounding_Rectangle& xRectangle,const Asura_Bounding_Rectangle& xUV,const int iTextureID,const int iBackgroundColour = Tools_TT_Colour::Null,const ASURA_TRANSLUCENCY_MODE eTranslucency = ASURA_TRANSLUCENCY_OFF );

inline void Tools_TT_RenderTexture( const Asura_Bounding_Rectangle& xRectangle,const int iTextureID,const int iBackgroundColour = Tools_TT_Colour::Null,const ASURA_TRANSLUCENCY_MODE eTranslucency = ASURA_TRANSLUCENCY_OFF )
{
	Tools_TT_RenderTexture( xRectangle.GetTopLeft(),xRectangle.GetBottomRight(),iTextureID,iBackgroundColour,eTranslucency );
}

void Tools_TT_RenderTexture( Asura_Bounding_Rectangle xRectangle,const Tools_TT_Tile& xTexture,const bool bFixAspectRatio = false,const int iBackgroundColour = Tools_TT_Colour::Null,const ASURA_TRANSLUCENCY_MODE eTranslucency = ASURA_TRANSLUCENCY_OFF );

// Render border for a UI component
void Tools_TT_RenderUIBorder( const Tools_TT_UIComponent& xUIComponent );

// Render border as an outline of a rectangle
void Tools_TT_RenderBorder( const Asura_Vector_2& xTopLeft,const Asura_Vector_2& xBottomRight,const u_int uColour );
void Tools_TT_RenderBorder( const Asura_Bounding_Rectangle& xRectangle,const u_int uColour );

// Render a TTL with the position and size specified
void Tools_TT_RenderTTL( const Asura_Bounding_Rectangle& xTTLScreenRectangle,const Tools_TT_TiledTextureList& xTTL,const int iBackgroundColour = Tools_TT_Colour::Null,const ASURA_TRANSLUCENCY_MODE eTranslucency = ASURA_TRANSLUCENCY_OFF );
void Tools_TT_RenderTTLSubTextureHighlight( const Asura_Bounding_Rectangle& xTTLScreenRectangle,const Tools_TT_Tile& xSubtexture, const int iColour );




#endif // __TOOLS_TEXTURETILER_RENDER_H_
