#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _Tools_Undo_hpp
#define _Tools_Undo_hpp

#include "Core/Asura_String.h"

///////////////////////////////////////////////////////////////////////////////////////
class RootUndo
{
public :
	RootUndo(int buffer_size);
	virtual ~RootUndo();

	bool Undo();//perform 1 undo if possible
	bool Redo();//perform 1 redo if possible
	void MarkUndoStart();//Subsequent calles to AllocateNextUndoData will be put in a new undo batch
	void MarkUndoStartIfEndOfBatch();//As above , but only if there has been a 'frame' in which nothing was logged for undo
	void KeepUndoBatchOpenThisFrame();//Makes the undo system think that undo data was logged  for this frame , so that it won't close the batch.

	//Prevents a new undo batch from being started unless at least a second has passed since
	//the last undo was logged.
	//A new batch can still be explicitly started using MarkUndoStart()
	void SetKeepUndoBatchOpenForShortTimeAfterUndoLogged(bool bSet){m_fMinTimeToKeepBatchOpen = bSet ? 1.0f : 0.0f;}


	//If set , then calls to Undo() and Redo() will display text
	//if any undoing/redoing is done
	void SetVerbose(bool bVerbose){m_bVerbose = bVerbose;}

	virtual void Initialise(); //Allocates the undo buffer
	virtual void Shutdown();//Deallocates the undo buffer

	void FlushUndoBuffer();//Clears out everything in the undo buffer 

	bool IsUndoInProcess(){return UndoInProcess;}//true if in the middle of a call to Undo() or Redo()
	bool IsInitialised(){return UndoInitialised;}

	bool CanUndo() const { return NumUndo > 0; }
	bool CanRedo() const { return NumRedo > 0; }

	int GetNumUndo() const {return NumUndo;}
	int GetNumRedo() const {return NumRedo;}
	int GetPosition() const {return (Position+BufferSize-StartPosition)%BufferSize;}

protected :
	inline void Prev(int&pos){(pos>0) ? (pos--) : (pos = (BufferSize-1));}
	inline void Next(int&pos){(pos <(BufferSize-1)) ? pos++ : pos=0;}
	
	void SetStartOfNewUndoIfNeeded();

	virtual bool IsUndoStart(int slot) = 0;
	virtual bool IsNullSlot(int slot) = 0;
	virtual void MarkSlotAsNull(int slot) = 0;
	virtual void MarkSlotAsUndoStart(int slot) = 0;

	virtual void ProcessUndo(int slot) = 0;
	virtual void Resize(int NewBufferSize) = 0;

	virtual void PreUndoProcessing(){}; //called before Undo() or Redo start (does nothing by default)
	virtual void PostUndoProcessing(){}; //called after Undo() or Redo finish (does nothing by default)
	
	
	bool m_bVerbose;

	Asura_String m_xUndoDescription;
	int m_iUndoDescriptionPriority;
	
	int NumUndo; //number of undo batches currently in buffer
	int NumRedo; //number of redo batches in buffer
	int Position; //current position in the undo buffer (buffer appears in derived class)
	int StartPosition; //index of start position in undo buffer

	bool UndoInProcess; //true while in the middle od Undo() or Redo()
	bool NewUndoBatch; //If true then the next time undo data is allocated it should start a new batch
	bool UndoInitialised; //Has the buffer actually been allocated

	bool UndoLoggedThisFrame;

	int BufferSize; //number of 'undo datas'

	float m_fLastUndoLogTime;//Time at which the most recent undo was logged
	float m_fMinTimeToKeepBatchOpen;

};

///////////////////////////////////////////////////////////////////////////////////////

#define FixedSizeUndo_StartOfUndo -2
#define FixedSizeUndo_Null -1

//class for handling a basic undo where all undo data is the same size
//The first entry of each undo data should be an int indicating the type of undo
//With the defines above representing special values for that entry
class FixedSizeUndo : public RootUndo
{
public :
	//undo_size is size of the structure , buffer_size the number of copies of the structure
	//that should be allocated
	FixedSizeUndo(int buffer_size,int undo_size);
	~FixedSizeUndo();


	virtual void Initialise();
	virtual void Shutdown();

	/*
	AllocateNextUndoData returns a pointer to a structure of size undo_size to be filled
	with undo data. It will return NULL if the undo system isn't initialsed , or if it is
	currently in the middle of processing an undo/redo
	*/
	void* AllocateNextUndoData();

	virtual void ProcessUndoData(void*) = 0;

private :
	virtual bool IsUndoStart(int slot);
	virtual bool IsNullSlot(int slot);
	virtual void MarkSlotAsNull(int slot);
	virtual void MarkSlotAsUndoStart(int slot);

	virtual void ProcessUndo(int slot);
	
	virtual void Resize(int NewBufferSize);
	
	virtual void PostUndoProcessing(){}
	

	int UndoDataSize; //size in ints

	int* UndoBuffer;
};

/////////////////////////////////////////////////////////////////////////////////////
#define VariableSizeUndo_StartOfUndo ((RootUndoData*)0xffffffff)
#define VariableSizeUndo_Null ((RootUndoData*)0)

class RootUndoData
{
public :
	virtual ~RootUndoData(){}

	//ProcessUndoData undoes what ever has to be done based on the data in the class , 
	//and returns a pointer to a new RootUndoData for the reverse undo.
	//If the pointer is not 'this' ,then this undo data will be deleted after
	//the function returns
	virtual RootUndoData* ProcessUndoData(){return this;}

	//Sets the text to be displayed when undoing/redoing this
	void SetDescription(const char* pszFormat,...);

	//gets the text to display when undoing/redoing this
	virtual const char* GetDescription() const{return m_xUndoDescription.GetString();}

	//Gets the priority level for this undo data for determing which undo's message
	//should be displayed. The one with the highest priority will be shown.
	virtual int GetPriority() const {return -1;}

protected :

	Asura_String m_xUndoDescription;

};

//Class for handling more complex undo data.
//Each piece of undo data is derived from RootUndoData  , and overrides the
//ProcessUndoData function to do what ever needs to be done.


class VariableSizeUndo : public RootUndo
{
public :
	VariableSizeUndo(int buffer_size);
	~VariableSizeUndo();

	virtual void Initialise();
	virtual void Shutdown();


	/*
	AllocateNextUndoData returns the address of a slot in the undobuffer
	That slot should be filled with a pointer to a class derived from RootUndoData
	allocated by the caller.
	It will return NULL if the undo system isn't initialsed , or if it is
	currently in the middle of processing an undo/redo.
	*/
	RootUndoData** AllocateNextUndoData();



private :
	virtual bool IsUndoStart(int slot);
	virtual bool IsNullSlot(int slot);
	virtual void MarkSlotAsNull(int slot);
	virtual void MarkSlotAsUndoStart(int slot);

	virtual void ProcessUndo(int slot);

	virtual void Resize(int NewBufferSize);

	RootUndoData** UndoBuffer;
};


////////////////////////////////////////////////////////////////////////////////////////

//A static instance of the variable size undo , in an attempt to resolve issues caused by having a seperate
//undo stack for texturing mode and the MRS system.
//Ultimately it may be desirable to move the other systems to be using this instance instead.
//RAR 04/06/2007

class Tools_Undo_System
{
public :
	static void Initialise();
	static void Shutdown()   			{s_xUndo.Shutdown();}
	static void Update();


	static bool Undo()		 			{return s_xUndo.Undo();}//perform 1 undo if possible
	static bool Redo()					{return s_xUndo.Redo();}//perform 1 redo if possible
	static void UndoForOverlay()		{s_xUndo.Undo();}//Return value removed , to make function useable by function dialog
	static void RedoForOverlay()		{s_xUndo.Redo();}
	static void MarkUndoStart()			{s_xUndo.MarkUndoStart();}//Subsequent calles to AllocateNextUndoData will be put in a new undo batch
//	static void MarkUndoStartIfEndOfBatch() {s_xUndo.MarkdStartIfEndOfBatch();}//As above , but only if there has been a 'frame' in which nothing was logged for undo
	static void KeepUndoBatchOpenThisFrame(){s_xUndo.KeepUndoBatchOpenThisFrame();}//Makes the undo system think that undo data was logged  for this frame , so that it won't close the batch.


	static void ClearUndo()				{s_xUndo.FlushUndoBuffer();}//Clears out everything in the undo buffer 

	static bool IsUndoInProcess()		{return s_xUndo.IsUndoInProcess();}//true if in the middle of a call to Undo() or Redo()
	static bool IsInitialised()			{return s_xUndo.IsInitialised();}

	static bool CanUndo()				{return s_xUndo.CanUndo();}
	static bool CanRedo()				{return s_xUndo.CanRedo();}

		/*
	AllocateNextUndoData returns the address of a slot in the undobuffer
	That slot should be filled with a pointer to a class derived from RootUndoData
	allocated by the caller.
	It will return NULL if the undo system isn't initialsed , or if it is
	currently in the middle of processing an undo/redo.
	*/
	static RootUndoData** AllocateNextUndoData() {return s_xUndo.AllocateNextUndoData();}

private :
	static VariableSizeUndo s_xUndo;
};

#endif
