//
// Tools_Visibility_System.cpp
//
// Created by KJFL - 01 June 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM

#ifndef ASURA_USE_TASK_SYSTEM 
#error The Tools-side of the Visibility System needs ASURA_USE_TASK_SYSTEM defined
#endif

// This file's header.
#include "Visibility/Tools_Visibility_System.h"

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_RingBuffer.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Threads/Asura_Task_Array.h"
#include "Core/Threads/Asura_Task_Stack.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Mesh/Asura_Env_System.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Visibility/Asura_Visibility_Node.h"
#include "Visibility/Asura_Visibility_System.h"

// Engine platform-specific.
#include "Rendering/Asura_Direct3D_LineRendering.h"

// Engine tools.
#include "Lighting/Tools_Environment_LOS.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_LOS.h"
#include "Misc/Tools_Polygon_Selection.h"
#include "Misc/Tools_ProgressIndicator.h"
#include "Misc/Tools_Prompted_Selection.h"
#include "Misc/Tools_Render.h"
#include "Visibility/Tools_ViewCell.h"
#include "Tools_Overlays.h"
#include "Tools_System.h"

// T Y P E S /////////////////////////////////////////////////////////////////

enum CREATETYPE
{
    CREATETYPE_ADVANCED,
    CREATETYPE_QUICK,
    CREATETYPE_MEDIUM,
    CREATETYPE_SLOW,
};        

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int uDEFAULT_MAX_VISIBILITY_SEGMENTS = 100000;
static const u_int uDEFAULT_MAX_TREE_DEPTH          = 12;
static const u_int uMAX_RANDOM_POINT_RETRIES        = 10;

// G L O B A L S /////////////////////////////////////////////////////////////

static Tools_ViewCell*                 g_pxEnvironmentViewCell;
static const Tools_VisibilitySegment** g_ppxVisibilitySegments;
static Asura_Bounding_Box              g_xEnvironmentAABB;
static u_int                           g_uNumberOfVisibilityFeatures;

static u_int            g_uNumberOfViewCells;
static u_int            g_uNumberOfLeafCells;
static Tools_ViewCell** g_ppxLeafCells;

static Tools_ProgressIndicator* g_pxProgressIndicator;

static Asura_Overlay* g_pxOverlay;
static Asura_Vector_2 g_xOverlayPosition;

static u_int g_uMaxNumberOfVisibilitySegments;
static u_int g_uCurrentNumberOfVisibilitySegments;
static u_int g_uMaxTreeDepth;
static bool  g_bUseMultithreading;

static bool g_bHighlightNonVisibleGeometry;

static bool g_bRenderVisibleNodeBBs;
static bool g_bRenderVisibleNodeBBs_HighlightCamNode;
static bool g_bRenderVisibleNodeBBs_HighlightVisSegmentTargetNode;

static bool                            g_bRenderVisSegments;
static bool                            g_bRenderVisSegments_OnlyForSelectedNodes;
static u_int                           g_uRenderVisSegments_CurrentlyPickedNode;
static u_int                           g_uRenderVisSegments_TargetNode;
static const Tools_VisibilitySegment** g_ppxRenderVisSegments_FilteredSegments;
static u_int                           g_uRenderVisSegments_NumFilteredSegments;
static bool                            g_bRenderVisSegments_GotFilteredList;
static bool                            g_bRenderVisSegments_ForceFilteredListRecalc;

static bool g_bRenderSplitHintQuads;

static u_int g_uManuallySelectCameraNodeCurrentlyPickedNode;

static bool  g_bAutoRefinementFromCameraPosition;
static bool  g_bAutoRefinementFromCameraNode;
static bool  g_bAutoRefinementUnderCursor;
static u_int g_uMaxIncrementalRefinementRaysPerFrame;

static u_int g_uAutoRefineTimeInMins;

static u_int g_uMemoryUsed_LeafVisibilities = 0;

static Tools_Visibility_System::START_VIS_CALCS_CALLBACK        g_pfnStartVisCalcsCallback;
static Tools_Visibility_System::END_VIS_CALCS_CALLBACK          g_pfnEndVisCalcsCallback;
static Tools_Visibility_System::RAY_HIT_FILTER_CALLBACK         g_pfnRayHitFilterCallback;
static Tools_Visibility_System::RAY_MODULE_TEST_FILTER_CALLBACK g_pfnRayModuleTestFilterCallback;
static const void*                                              g_pvRayModuleTestFilterUserData;

static Asura_LinkList<Tools_Visibility_SplitHintQuad> g_xSplitHintQuadList;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

u_int uCellsConsidered = 0;
Asura_Mutex xViewCellData;
Asura_Mutex xCells;
Asura_Task_Stack* pxTaskQueue;

void Tools_Visibility_System::Initialise( void )
{
    g_pxEnvironmentViewCell = 0;
    g_ppxVisibilitySegments = 0;

    g_uNumberOfViewCells = 0;
    g_uNumberOfLeafCells = 0;
    g_ppxLeafCells = 0;

    g_pxProgressIndicator = 0;

    g_pxOverlay = 0;
    g_xOverlayPosition = Asura_Vector_2(1.0f, 3.0f);

    g_uMaxNumberOfVisibilitySegments = uDEFAULT_MAX_VISIBILITY_SEGMENTS;
    g_uCurrentNumberOfVisibilitySegments = 0;
    g_uMaxTreeDepth = uDEFAULT_MAX_TREE_DEPTH;
    g_bUseMultithreading = true;

    g_bHighlightNonVisibleGeometry = false;

    g_bRenderVisibleNodeBBs = false;
    g_bRenderVisibleNodeBBs_HighlightCamNode = true;
    g_bRenderVisibleNodeBBs_HighlightVisSegmentTargetNode = true;

    g_bRenderVisSegments = false;
    g_bRenderVisSegments_OnlyForSelectedNodes = true; // On by default because it's so unwieldly otherwise after a fresh recalc.
    g_uRenderVisSegments_CurrentlyPickedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
    g_uRenderVisSegments_TargetNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
    g_ppxRenderVisSegments_FilteredSegments = NULL;
    g_uRenderVisSegments_NumFilteredSegments = 0;
    g_bRenderVisSegments_GotFilteredList = false;
    g_bRenderVisSegments_ForceFilteredListRecalc = true;

    g_bRenderSplitHintQuads = false;

    g_uManuallySelectCameraNodeCurrentlyPickedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;

    g_bAutoRefinementFromCameraPosition = false;
    g_bAutoRefinementFromCameraNode = false;
    g_bAutoRefinementUnderCursor = false;
    g_uMaxIncrementalRefinementRaysPerFrame = 250;

    g_uAutoRefineTimeInMins = 1;

    g_pfnStartVisCalcsCallback       = NULL;
    g_pfnEndVisCalcsCallback         = NULL;
    g_pfnRayHitFilterCallback        = NULL;
    g_pfnRayModuleTestFilterCallback = NULL;
    g_pvRayModuleTestFilterUserData  = NULL;
}


void Tools_Visibility_System::Reset( void )
{
    if (g_pxEnvironmentViewCell)
    {
        Tools_ViewCell::Destroy( g_pxEnvironmentViewCell );
        g_pxEnvironmentViewCell = 0;
    }

    ClearStoredVisibilitySegments();

    if ( g_ppxLeafCells )
    {
        delete [] g_ppxLeafCells;
        g_ppxLeafCells = 0;
    }
    g_uNumberOfViewCells = 0;
    g_uNumberOfLeafCells = 0;

    g_uRenderVisSegments_CurrentlyPickedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
    g_uRenderVisSegments_TargetNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
    RenderVisSegments_ClearFilteredList();

    g_uManuallySelectCameraNodeCurrentlyPickedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;

    g_bAutoRefinementFromCameraPosition = false;
    g_bAutoRefinementFromCameraNode = false;
    g_bAutoRefinementUnderCursor = false;
}

void Tools_Visibility_System::Shutdown( void )
{
    Reset();

    CloseOverlay();
}


u_int Tools_Visibility_System::GetNumberOfVisibilityFeatures( void )
{
    return g_uNumberOfVisibilityFeatures;
}


u_int Tools_Visibility_System::GetSkyBoxFeatureID( void )
{
    return g_uNumberOfVisibilityFeatures - 1;
}


void Tools_Visibility_System::ToggleOverlay( void )
{
    if ( IsOverlayOpen() )
    {
        CloseOverlay();
    }
    else
    {
        CreateOverlay();
    }
}

bool Tools_Visibility_System::IsOverlayOpen( void )
{
    if ( g_pxOverlay )
    {
        if ( Asura_Overlay_System::DoesOverlayExist( TOOLS_OVERLAY_VISIBILITY ) )
        {
            return true;
        }
        
        g_pxOverlay = 0;
    }
    
    return false;
}


void Tools_Visibility_System::CreateOverlay( void )
{
    if ( g_pxOverlay ) return;


//  Asura_Dialog* pxNewDialog;
//  Asura_Tickbox_Dialog* pxNewTickbox;

    // Store the overlay...
    g_pxOverlay     = new Asura_Overlay( "Visibility System", false, &g_xOverlayPosition );
    g_pxOverlay->m_uID  = TOOLS_OVERLAY_VISIBILITY;

    Asura_Vector_2  xOffset(1.0f, 3.0f);

    const char* aszTAB_NAMES[] =
    {
        "Create",
        "Refine",
        "Rendering",
    };

    const u_int uNUM_TABS = sizeof( aszTAB_NAMES ) / sizeof( aszTAB_NAMES[0] );

    Asura_Tabs_Dialog* const pxTabsDialog = new Asura_Tabs_Dialog( xOffset, aszTAB_NAMES, uNUM_TABS, 45 );
    g_pxOverlay->AddDialog( pxTabsDialog );
    

    xOffset.y += Asura_Tickbox_Dialog::LargeSpacing();
    
    // rendering tab
    {
        const float fSUB_OPTION_INSET = 4.0f;

        pxTabsDialog->SelectTab(2);
        Asura_Vector_2 xTabOffset = xOffset;

        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Render everything (turn off system)", xTabOffset, &Asura_Visibility_System::s_bRenderEverything );
            pxTickbox->SetHelpText( "Ignore visibility data and just render everything. This only affects rendering within the current tools session." );
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxTickbox );
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Highlight non-visible geometry", xTabOffset, &g_bHighlightNonVisibleGeometry );
            pxTickbox->SetHelpText( "Any geometry that's considered non-visible is rendered bright green. This makes it easier to spot and manually force visibility for sections of environment, along the lines of how module linking mode used to work." );
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxTickbox );
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Freeze camera node", xTabOffset, &Asura_Visibility_System::s_bFreezeCameraNode );
            pxTickbox->SetHelpText( "Prevents the camera node from being updated, so that only visibilities relative to the node you were in when you ticked this are used" );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.x += fSUB_OPTION_INSET;
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();

            Asura_Function_Dialog* pxButton = new Asura_Function_Dialog( "Manually select camera node", xTabOffset, ManuallySelectCameraNode );
            pxButton->SetHelpText( "Manually select the frozen camera node by picking with the mouse, instead of just using whatever node the camera is currently in. This will enable the above option if it isn't already on." );
            pxTabsDialog->AddDialog( pxButton );

            xTabOffset.x -= fSUB_OPTION_INSET;
            xTabOffset.y = pxButton->m_xMaxCoord.y + Asura_Function_Dialog::DefaultSpacing();
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "X-ray mode", xTabOffset, &Asura_Visibility_System::s_bXRayMode );
            pxTickbox->SetHelpText( "Renders without z-buffering so that you can see what geometry is considered visible behind what's actually being rendered on the screen and get an idea of overdraw" );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Render visibility segments", xTabOffset, &g_bRenderVisSegments );
            pxTickbox->SetHelpText( "Renders visibility segments for data that was created in this tools session. Warning: if you've created the data anew, especially via the 'slow create' option, then this renders a huge number of lines and is quite slow unless you use the options below." );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.x += fSUB_OPTION_INSET;
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();

            pxTickbox = new Asura_Tickbox_Dialog( "Only for selected nodes", xTabOffset, &g_bRenderVisSegments_OnlyForSelectedNodes );
            pxTickbox->SetHelpText( "Only render visibility segments between the current camera node (or, more often, the frozen camera node) and/or the target node you've selected below." );
            pxTickbox->SetIsEnabledFunction( RenderVisSegments_ShouldRender );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();

            Asura_Function_Dialog* pxButton = new Asura_Function_Dialog( "Select vis segment target node", xTabOffset, RenderVisSegments_SelectTargetNode );
            pxButton->SetHelpText( "Select the node for which you want to see the visibility segments that caused it to be considered visible" );
            pxButton->SetIsEnabledFunction( RenderVisSegments_ShouldBeAllowedToSelectTargetNode );
            pxTabsDialog->AddDialog( pxButton );

            xTabOffset.y = pxButton->m_xMaxCoord.y + Asura_Function_Dialog::DefaultSpacing();

            pxButton = new Asura_Function_Dialog( "Clear vis segment target node", xTabOffset, RenderVisSegments_ClearTargetNode );
            pxButton->SetHelpText( "Clear the current vis segment target node" );
            pxButton->SetIsEnabledFunction( RenderVisSegments_ShouldBeAllowedToSelectTargetNode );
            pxTabsDialog->AddDialog( pxButton );

            xTabOffset.y = pxButton->m_xMaxCoord.y + Asura_Function_Dialog::DefaultSpacing();

            pxButton = new Asura_Function_Dialog( "Forget stored visibility segments", xTabOffset, ClearStoredVisibilitySegments );
            pxButton->SetHelpText( "Clear the stored visibility segments. This does not affect the visibility data; these segments are only stored for analysis of the visibility data in tools and are not saved out." );
//          pxButton->SetIsEnabledFunction( RenderVisSegments_ShouldRender );
            pxTabsDialog->AddDialog( pxButton );

            xTabOffset.x -= fSUB_OPTION_INSET;
            xTabOffset.y = pxButton->m_xMaxCoord.y + Asura_Function_Dialog::DefaultSpacing();
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Render node BBs", xTabOffset, &g_bRenderVisibleNodeBBs );
            pxTickbox->SetHelpText( "Render nodes that are considered visible from the current camera node (or frozen camera node)" );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.x += fSUB_OPTION_INSET;
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();

            pxTickbox = new Asura_Tickbox_Dialog( "Highlight camera node", xTabOffset, &g_bRenderVisibleNodeBBs_HighlightCamNode );
            pxTickbox->SetHelpText( "Render the camera node's BB differently from the rest" );
            pxTickbox->SetIsEnabledFunction( ShouldRenderVisibleNodeBBs );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();

            pxTickbox = new Asura_Tickbox_Dialog( "Highlight vis segment target node", xTabOffset, &g_bRenderVisibleNodeBBs_HighlightVisSegmentTargetNode );
            pxTickbox->SetHelpText( "Render the current vis segment target node's BB differently from the rest" );
            pxTickbox->SetIsEnabledFunction( ShouldRenderVisibleNodeBBs );
            pxTabsDialog->AddDialog( pxTickbox );

            xTabOffset.x -= fSUB_OPTION_INSET;
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Render split hint quads", xTabOffset, &g_bRenderSplitHintQuads );
            pxTickbox->SetHelpText( "Render quads that were derived from any axis-aligned vis blocking polys in the level and that were therefore used as hints for splitting vis nodes" );
            pxTabsDialog->AddDialog( pxTickbox );
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
        }
    }
    // linking tab
    {
        pxTabsDialog->SelectTab(1);
        Asura_Vector_2 xTabOffset = xOffset;
        {
            Asura_Function_Dialog* pxDialog = new Asura_Function_Dialog( "Remove links from camera node", xTabOffset, RemoveLinks );
            pxDialog->SetHelpText( "Scrap all links from the visibility node containing the camera - or the frozen camera node if you've used that option" );
            xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::LargeSpacing();
            pxTabsDialog->AddDialog( pxDialog );
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Auto-add links from camera position", xTabOffset, &g_bAutoRefinementFromCameraPosition );
            pxTickbox->SetHelpText( "Check for additional visibilities from the current camera position in all directions for as long as this is ticked.  Since this is a random process the longer you run it for the more accurate the results will be. Note that refinement will not affect vis node splitting, only the links between existing nodes." );
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxTickbox );
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Auto-add links from camera node", xTabOffset, &g_bAutoRefinementFromCameraNode );
            pxTickbox->SetHelpText( "Check for additional visibilities from random points within the current camera node (or frozen camera node) in all directions for as long as this is ticked.  Since this is a random process the longer you run it for the more accurate the results will be. Note that refinement will not affect vis node splitting, only the links between existing nodes." );
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxTickbox );
        }
        {
            Asura_Tickbox_Dialog* pxTickbox = new Asura_Tickbox_Dialog( "Auto-add links under cursor", xTabOffset, &g_bAutoRefinementUnderCursor );
            pxTickbox->SetHelpText( "Check for additional visibilities from the current camera position in the direction of the mouse cursor for as long as this is ticked. Note that refinement will not affect vis node splitting, only the links between existing nodes." );
            xTabOffset.y = pxTickbox->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxTickbox );
        }
        {
            Asura_Numerical_Dialog* pxDialog = new Asura_Numerical_Dialog( "Max rays to cast per frame", xTabOffset, &g_uMaxIncrementalRefinementRaysPerFrame, 1, 10000, 1, "%u" );
            pxDialog->SetHelpText( "Maximum number of rays to cast per frame for any of the above refinement options. Refinement will be faster if this value is higher but tools may not be very usable if you set it too high." );
            xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::LargeSpacing();
            pxTabsDialog->AddDialog( pxDialog );
        }
        {
            Asura_Numerical_Dialog* pxDialog = new Asura_Numerical_Dialog( "Auto-refine time (in minutes)", xTabOffset, &g_uAutoRefineTimeInMins, 1, 999, 1, "%d" );
            pxDialog->SetHelpText( "Specify how long auto-refinement should run for when you click the button below" );
            xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxDialog );
        }
        {
            Asura_Function_Dialog* pxDialog = new Asura_Function_Dialog( "Auto-refine (whole environment)", xTabOffset, AutoRefine );
            pxDialog->SetHelpText( "Check for additional visibilities from random points everywhere in the environment. Since this is a random process the longer you run it for the more accurate the results will be. Note that refinement will not affect vis node splitting, only the links between existing nodes." );
            xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
            pxTabsDialog->AddDialog( pxDialog );
        }
        
    }
    // Create ViewCell tab
    {
        pxTabsDialog->SelectTab(0);
    
    
        const char* aszTAB_NAMES_2[] =
        {
            "Simple",
            "Advanced"
        };

        const u_int uNUM_TABS_2 = sizeof( aszTAB_NAMES_2 ) / sizeof( aszTAB_NAMES_2[0] );

        Asura_Tabs_Dialog* const pxTabsDialog2 = new Asura_Tabs_Dialog( xOffset, aszTAB_NAMES_2, uNUM_TABS_2, 45 );
        pxTabsDialog->AddDialog( pxTabsDialog2 );
        
        {
            pxTabsDialog2->SelectTab(1);
            Asura_Vector_2 xTabOffset = xOffset;
            xTabOffset.y += Asura_Tickbox_Dialog::LargeSpacing() + Asura_Tickbox_Dialog::DefaultSpacing() * 0.5f;
        
            {
                Asura_Function_Dialog* pxDialog = new Asura_Function_Dialog( "Create visibility nodes", xTabOffset, CreateViewCells, CREATETYPE_ADVANCED );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::LargeSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
            {
                Asura_Numerical_Dialog* pxDialog = new Asura_Numerical_Dialog( "Max Tree Depth", xTabOffset, &g_uMaxTreeDepth, 2, 16, 1, "%d" );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
            {
                Asura_Numerical_Dialog* pxDialog = new Asura_Numerical_Dialog( "Max View Segments", xTabOffset, &g_uMaxNumberOfVisibilitySegments, 1000, 10000000, 1000, "%d" );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
            {
                Asura_Tickbox_Dialog* pxDialog = new Asura_Tickbox_Dialog( "Use Multithreading", xTabOffset, &g_bUseMultithreading );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
            
        }

        {
            pxTabsDialog2->SelectTab(0);
            Asura_Vector_2 xTabOffset = xOffset;
            xTabOffset.y += Asura_Tickbox_Dialog::LargeSpacing() + Asura_Tickbox_Dialog::DefaultSpacing() * 0.5f;
        
            {
                Asura_Function_Dialog* pxDialog = new Asura_Function_Dialog( "Quick Create (Approx)", xTabOffset, CreateViewCells, CREATETYPE_QUICK );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
            {
                Asura_Function_Dialog* pxDialog = new Asura_Function_Dialog( "Medium Create", xTabOffset, CreateViewCells, CREATETYPE_MEDIUM );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
            {
                Asura_Function_Dialog* pxDialog = new Asura_Function_Dialog( "Slow Create (Accurate)", xTabOffset, CreateViewCells, CREATETYPE_SLOW );
                xTabOffset.y = pxDialog->m_xMaxCoord.y + Asura_Tickbox_Dialog::DefaultSpacing();
                pxTabsDialog2->AddDialog( pxDialog );
            }
        }

    }

    // and then give it to the overlay system to look after 
    Asura_Overlay_System::SubmitNewOverlay( g_pxOverlay, true );
}


void Tools_Visibility_System::CloseOverlay( void )
{
    g_pxOverlay = 0;

    Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_VISIBILITY );
}

void Tools_Visibility_System::InitialiseProgressIndicator( void )
{
    g_pxProgressIndicator = new Tools_ProgressIndicator( "" );

    g_pxProgressIndicator->AddMessage( "Creating Visibility Data - Hold ESC to quit" );
    g_pxProgressIndicator->AddMessage( "" );
}

void Tools_Visibility_System::ShutdownProgressIndicator( void )
{
    delete g_pxProgressIndicator;
}
void Tools_Visibility_System::StartProgressIndicator( const char* pszCurrentTaskString )
{
    g_pxProgressIndicator->SetText( pszCurrentTaskString );
    g_pxProgressIndicator->Update( 0, 0, true );
}
void Tools_Visibility_System::UpdateProgressIndicator( u_int uCurrent, u_int uMax, bool bForceUpdate )
{
    g_pxProgressIndicator->Update( uCurrent, uMax, bForceUpdate );
}


void Tools_Visibility_System::CreateViewCells( int iSetting )
{
    Reset();

    if( !Asura_Environment_System::GetEnvironment() )
    {
        return;
    }

    switch ( iSetting )
    {
        default:
        case CREATETYPE_ADVANCED:
        {
            break;
        }
        case CREATETYPE_QUICK:
        {
            g_uMaxNumberOfVisibilitySegments = 100000;
            g_uMaxTreeDepth = 10;
            break;
        }
        case CREATETYPE_MEDIUM:
        {
            g_uMaxNumberOfVisibilitySegments = 500000;
            g_uMaxTreeDepth = 12;
            break;
        }
        case CREATETYPE_SLOW:
        {
            g_uMaxNumberOfVisibilitySegments = 3000000;
            g_uMaxTreeDepth = 12;
            break;
        }
    }

    InitialiseProgressIndicator();

    // Guarantee LOS data is valid
    Tools_Environment_LOS::Reset();
    Tools_Environment_LOS::EnsureDataIsValid();

    RecalculateBaseEnvironmentData();

    Asura_TickCount uStartTime = Asura_Timers::GetCurTicks();

    StartProgressIndicator( "Casting Visibility Rays" );

    if( g_pfnStartVisCalcsCallback )
    {
        g_pfnStartVisCalcsCallback();
    }

    bool bQuitOut = false;
    if ( g_bUseMultithreading )
    {
        bQuitOut = CreateVisibilitySegments_MT();
    }
    else
    {
        bQuitOut = CreateVisibilitySegments();
    }
    if ( bQuitOut ) goto QuitOut;

    g_pxEnvironmentViewCell = Tools_ViewCell::Create( g_xEnvironmentAABB, g_ppxVisibilitySegments, g_uMaxNumberOfVisibilitySegments, 0 );

    Asura_TickCount uSegmentCreationTime = (Asura_Timers::GetCurTicks() - uStartTime);

    uStartTime = Asura_Timers::GetCurTicks();

    StartProgressIndicator( "Creating View Cells" );

    FindSplitHintQuads();
    Tools_ViewCell::SetSplitHintQuadList( &g_xSplitHintQuadList );

    if ( g_bUseMultithreading )
    {
        bQuitOut = SplitMT();
    }
    else
    {
        bQuitOut = Split();
    }
    if ( bQuitOut ) goto QuitOut;

    Tools_ViewCell::SetSplitHintQuadList( NULL );

    StartProgressIndicator( "Creating Cell to Cell Visibilities" );

    bQuitOut = CreateLeafToLeafVisibilties();
    if ( bQuitOut ) goto QuitOut;

    InitialiseAsuraData();

    float fLeafCompression = static_cast<float>(g_uMemoryUsed_LeafVisibilities) / static_cast<float>(g_uNumberOfLeafCells*(g_uNumberOfLeafCells));

    u_int uMemoryUsed_Viewcells = g_uNumberOfViewCells * sizeof(Asura_Visibility_Node);
    u_int uMemoryUsed_FeatureVisibilities = g_uNumberOfLeafCells * g_uNumberOfVisibilityFeatures;

    u_int uMemoryUsed_Total = uMemoryUsed_FeatureVisibilities + uMemoryUsed_Viewcells + g_uMemoryUsed_LeafVisibilities;

    Asura_Logfile_Retail(( "logfiles/visibility.txt", 
                           "Using Multithreading: %s\n"
                           "Number of Visibility Segments: %d\n\n"
                           "Vis Segments Time: %fs\n"
                           "Split Time: %fs\n"
                           "Number of Nodes: %d\nNumber of Leaves: %d\nNumber of Features: %d\n"
                           "LeafVisibilities: %d bytes (%f%% compression)\n"
                           "FeatureVisibilities: %d bytes\n"
                           "Visibility nodes tree: %d bytes\n"
                           "Total Memory: %d bytes\n"
                           "\n",
                           g_bUseMultithreading ? "Yes" : "No",
                           g_uMaxNumberOfVisibilitySegments,
                           Asura_Timers::TicksToSeconds(uSegmentCreationTime), 
                           Asura_Timers::TicksToSeconds(Asura_Timers::GetCurTicks() - uStartTime),
                           g_uNumberOfViewCells, g_uNumberOfLeafCells, g_uNumberOfVisibilityFeatures,
                           g_uMemoryUsed_LeafVisibilities, fLeafCompression,
                           uMemoryUsed_FeatureVisibilities,
                           uMemoryUsed_Viewcells,
                           uMemoryUsed_Total ));

    g_pxEnvironmentViewCell->DebugLogFile();

QuitOut:

    if( g_pfnEndVisCalcsCallback )
    {
        g_pfnEndVisCalcsCallback();
    }

    ShutdownProgressIndicator();

    if (g_pxEnvironmentViewCell)
    {
        Tools_ViewCell::Destroy( g_pxEnvironmentViewCell );
        g_pxEnvironmentViewCell = 0;
    }

    if ( g_ppxLeafCells )
    {
        delete [] g_ppxLeafCells;
        g_ppxLeafCells = 0;
    }

    g_uNumberOfViewCells = 0;
    g_uNumberOfLeafCells = 0;

    if ( bQuitOut )
    {
        Reset();
    }
}

void Tools_Visibility_System::SetAsModified( void )
{
    Tools_SaveAsura::SetModuleLinkingModified();
    Asura_Visibility_System::s_bForceVisRecalcForCurrentCameraNode = true;
}

void Tools_Visibility_System::InitialiseAsuraData( void )
{
    g_uMemoryUsed_LeafVisibilities = 0;

    Asura_Visibility_System::Reset();

    if ( g_uNumberOfViewCells == 0 ) return;

    Asura_Visibility_System::s_uNumberOfNodes = g_uNumberOfViewCells;

    Asura_Visibility_System::s_pxVisibilityNodes = new Asura_Visibility_Node[ g_uNumberOfViewCells ];

    Asura_Visibility_System::s_uNumberOfLeaves = g_uNumberOfLeafCells;

    Asura_Visibility_System::s_pxVisibilityLeaves = new Asura_Visibility_Leaf[ g_uNumberOfLeafCells ];

    Asura_Visibility_System::s_uNumberOfVisibilityFeatures = GetNumberOfVisibilityFeatures();

    Asura_Visibility_System::CreateNodeVisibilityFlags();

    for ( u_int u=0; u<g_uNumberOfLeafCells; ++u )
    {
        Asura_Visibility_System::s_pxVisibilityLeaves[u].m_pucFeatureVisibilities = new u_char[ GetNumberOfVisibilityFeatures() ];
        Asura_Mem::Set( Asura_Visibility_System::s_pxVisibilityLeaves[u].m_pucFeatureVisibilities, 0, GetNumberOfVisibilityFeatures() );
    }

    u_int uNumberOfLeaves = 0;
    u_int uNextFreeNode = 1;

    class TVS_StackData
    {
        public:
            Tools_ViewCell* m_pxCell;
            u_int           m_uNode;
            u_int           m_uParentNode;
    };

    TVS_StackData axStack[32];
    u_int uStack = 0;

    // push root onto stack
    axStack[ uStack ].m_pxCell = g_pxEnvironmentViewCell;
    axStack[ uStack ].m_uNode = 0;
    axStack[ uStack ].m_uParentNode = 0;
    ++uStack;

    while ( uStack )
    {
        // pop something off the stack to work on
        TVS_StackData& xData = axStack[ --uStack ];
    
        Tools_ViewCell* pxViewCell = xData.m_pxCell;
    
        Asura_Assert( xData.m_uNode < g_uNumberOfViewCells, "Code bug: array overflow" );

        if ( pxViewCell->GetLeftChild() && pxViewCell->GetRightChild() )
        {
            // think of the children
            u_int uLeft = uNextFreeNode++;
            u_int uRight = uNextFreeNode++;
    
            Asura_Visibility_System::s_pxVisibilityNodes[ xData.m_uNode ].Tools_Initialise( pxViewCell->GetAABB(), xData.m_uParentNode, uLeft, 0 );
        
            u_int uParent = xData.m_uNode;

            // push children onto stack - usually read back left/right, so add them right/left to the stack.
            axStack[ uStack ].m_pxCell = pxViewCell->GetRightChild();
            axStack[ uStack ].m_uNode = uRight;
            axStack[ uStack ].m_uParentNode = uParent;
            ++uStack;

            axStack[ uStack ].m_pxCell = pxViewCell->GetLeftChild();
            axStack[ uStack ].m_uNode = uLeft;
            axStack[ uStack ].m_uParentNode = uParent;
            ++uStack;
        }
        else
        {
            // this is a leaf
            const u_int uLeafIndex = pxViewCell->GetLeafIndex();
            Asura_Visibility_System::s_pxVisibilityNodes[ xData.m_uNode ].Tools_Initialise( pxViewCell->GetAABB(), xData.m_uParentNode, 0, uLeafIndex );

            ++uNumberOfLeaves;

            // set which leaves are visible from this node
            //if ( pxViewCell->GetVisibleLeavesArray() )
            // zero pointer now indicates an array full of zeros
            {
                Asura_Visibility_System::s_pxVisibilityLeaves[ uLeafIndex].Compress( pxViewCell->GetVisibleLeavesArray(), g_uNumberOfLeafCells );
            }
            
            g_uMemoryUsed_LeafVisibilities += Asura_Visibility_System::s_pxVisibilityLeaves[ uLeafIndex ].m_xLeafVisibilities.GetMemoryUsed();

            // set which "features" are visible from this node

            if (pxViewCell)
            {
                for ( u_int u=0; u<GetNumberOfVisibilityFeatures(); ++u )
                {
                    Asura_Visibility_System::s_pxVisibilityLeaves[ uLeafIndex].m_pucFeatureVisibilities[ u ] = pxViewCell->GetVisibleFeaturesArray()[ u ];
                }
            }
        }
    }

    Asura_Assert( uNextFreeNode == g_uNumberOfViewCells, "Code bug: not all view cells accounted for" );
    Asura_Assert( uNumberOfLeaves == g_uNumberOfLeafCells, "Code bug: not all leaf cells accounted for" );

    SetAsModified();
}

bool Tools_Visibility_System::CreateVisibilitySegments( void )
{
    u_int uSegmentsToCreate = g_uMaxNumberOfVisibilitySegments;
    u_int uCurrentSegment = 0;

    g_uCurrentNumberOfVisibilitySegments = uSegmentsToCreate;
    g_ppxVisibilitySegments = new const Tools_VisibilitySegment*[uSegmentsToCreate];

    Asura_Mem::Set( g_ppxVisibilitySegments, 0, sizeof(Tools_VisibilitySegment*) * uSegmentsToCreate );

    Asura_Vector_3 xSamplePoint;

    bool bMakeSamplesInsideFeatures = false;
    bool bCurSamplePointIsValid = false;

    while ( uSegmentsToCreate )
    {
        UpdateProgressIndicator( uCurrentSegment, g_uMaxNumberOfVisibilitySegments-1 );

        // It's possible for the code handling hitting a back face below to
        // push the sample point outside the environment. Deal with that by
        // picking a new sample point. Do the same if the current sample
        // point isn't valid.
        if( !bCurSamplePointIsValid || !g_xEnvironmentAABB.Contains(xSamplePoint) )
        {
            bCurSamplePointIsValid = GetRandomPoint( xSamplePoint, bMakeSamplesInsideFeatures );
        }

        // Just give up for this segment and try again if we haven't got a
        // valid sample point.
        if( !bCurSamplePointIsValid )
        {
            continue;
        }

        Asura_Vector_3 xDirection;
        GetRandomDirection( xDirection );

        Asura_Vector_3 xSegment0;
        Asura_Vector_3 xSegment1;
        u_int           uFeatureID0;
        u_int           uFeatureID1;

        bool bFrontFacing;
        Asura_Vector_3 xNormal;
        bool bHitVisBlocker;

        bool bHit1 = FireRay_Internal( xSamplePoint, xDirection, xSegment0, bFrontFacing, uFeatureID0, xNormal, bHitVisBlocker );

        if (bHitVisBlocker)
        {
            // hit a VisBlocker - abandon and try a new start point
            bCurSamplePointIsValid = false;
            continue;
        }

        if( bHit1 && !bFrontFacing)
        {
            // abandon the segment
            xSamplePoint = xSegment0 + xNormal * 0.002f;
            bCurSamplePointIsValid = IsVisSegmentSamplePointValid( xSamplePoint );
            continue;
        }
    
        xDirection = -xDirection;
        bool bHit2 = FireRay_Internal( xSamplePoint, xDirection, xSegment1, bFrontFacing, uFeatureID1, xNormal, bHitVisBlocker );

        if (bHitVisBlocker)
        {
            // hit a VisBlocker - abandon and try a new start point
            bCurSamplePointIsValid = false;
            continue;
        }

        if (!bHit1 && !bHit2)
        {
            bCurSamplePointIsValid = false;
            continue;
        }

        if( bHit2 && !bFrontFacing)
        {
            // abandon the segment
            xSamplePoint = xSegment1 + xNormal * 0.002f;
            bCurSamplePointIsValid = IsVisSegmentSamplePointValid( xSamplePoint );
            continue;
        }

        Tools_VisibilitySegment* pxSegment = Tools_VisibilitySegment::Create( false, xSegment0, xSegment1, uFeatureID0, uFeatureID1 );

        g_ppxVisibilitySegments[ uCurrentSegment ] = pxSegment;
    
        uCurrentSegment++;
        uSegmentsToCreate--;

        if ( uCurrentSegment > g_uMaxNumberOfVisibilitySegments/2 ) bMakeSamplesInsideFeatures = false;

        // And now start at a new point for the next segment.
        bCurSamplePointIsValid = false;

        // check for user trying to quit early
        if ( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) ) return true;
    }

    return false;
}

Asura_Mutex xNewSegmentMutex;

bool Tools_Visibility_System::CreateVisibilitySegments_MT( void )
{
    u_int uSegmentsToCreate = g_uMaxNumberOfVisibilitySegments;

    g_uCurrentNumberOfVisibilitySegments = uSegmentsToCreate;
    g_ppxVisibilitySegments = new const Tools_VisibilitySegment*[uSegmentsToCreate];

    Asura_Mem::Set( g_ppxVisibilitySegments, 0, sizeof(Tools_VisibilitySegment*) * uSegmentsToCreate );

    bool bQuit = false;

    Asura_Task_Array xTaskArray( g_uCurrentNumberOfVisibilitySegments, 1000, CreateVisibilitySegments_MT_Callback );

    while ( !xTaskArray.HasFinished() )
    {
        Asura_Thread_System::DelayCurrentThread( 0.3f, false );

        //g_pxProgressIndicator->AddMessage( "stack size %d", pxTaskQueue->GetCurrentQueueSize() );

        Tools_Visibility_System::UpdateProgressIndicator( xTaskArray.GetCurrentIndex(), uSegmentsToCreate );

        // check for user trying to quit early
        if ( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) )
        {
            bQuit = true;
            break;
        }
    }

    return bQuit;
}


void Tools_Visibility_System::CreateVisibilitySegments_MT_Callback( const void* pxTaskDataPacket )
{
    const Asura_Task_Array::TaskDataPacket* pxData = static_cast<const Asura_Task_Array::TaskDataPacket*>(pxTaskDataPacket);

    u_int uCurrentSegment = pxData->GetArrayStartIndex();
    u_int uSegmentsToCreate = pxData->GetNumberToProcess();

    // Tell the Asura_Task_Array we're done with this data packet. We won't be able to read from it after this.
    pxData->SignalFinishedWithData();

    Asura_Vector_3 xSamplePoint;

    bool bMakeSamplesInsideFeatures = false;
    bool bCurSamplePointIsValid = false;

    while ( uSegmentsToCreate )
    {
        // It's possible for the code handling hitting a back face below to
        // push the sample point outside the environment. Deal with that by
        // picking a new sample point. Do the same if the current sample
        // point isn't valid.
        if( !bCurSamplePointIsValid || !g_xEnvironmentAABB.Contains(xSamplePoint) )
        {
            bCurSamplePointIsValid = GetRandomPoint( xSamplePoint, bMakeSamplesInsideFeatures );
        }

        // Just give up for this segment and try again if we haven't got a
        // valid sample point.
        if( !bCurSamplePointIsValid )
        {
            continue;
        }

        Asura_Vector_3 xDirection;
        GetRandomDirection( xDirection );

        Asura_Vector_3 xSegment0;
        Asura_Vector_3 xSegment1;
        u_int           uFeatureID0;
        u_int           uFeatureID1;

        bool bFrontFacing;
        Asura_Vector_3 xNormal;
        bool bHitVisBlocker;

        bool bHit1 = FireRay_Internal( xSamplePoint, xDirection, xSegment0, bFrontFacing, uFeatureID0, xNormal, bHitVisBlocker );

        if (bHitVisBlocker)
        {
            // hit a VisBlocker - abandon and try a new start point
            bCurSamplePointIsValid = false;
            continue;
        }

        if( bHit1 && !bFrontFacing)
        {
            // abandon the segment
            xSamplePoint = xSegment0 + xNormal * 0.002f;
            bCurSamplePointIsValid = IsVisSegmentSamplePointValid( xSamplePoint );
            continue;
        }
    
        xDirection = -xDirection;
        bool bHit2 = FireRay_Internal( xSamplePoint, xDirection, xSegment1, bFrontFacing, uFeatureID1, xNormal, bHitVisBlocker );

        if (bHitVisBlocker)
        {
            // hit a VisBlocker - abandon and try a new start point
            bCurSamplePointIsValid = false;
            continue;
        }

        if (!bHit1 && !bHit2)
        {
            bCurSamplePointIsValid = false;
            continue;
        }

        if( bHit2 && !bFrontFacing )
        {
            // abandon the segment
            xSamplePoint = xSegment1 + xNormal * 0.002f;
            bCurSamplePointIsValid = IsVisSegmentSamplePointValid( xSamplePoint );
            continue;
        }

        xNewSegmentMutex.Lock();
        Tools_VisibilitySegment* pxSegment = Tools_VisibilitySegment::Create( false, xSegment0, xSegment1, uFeatureID0, uFeatureID1 );
        xNewSegmentMutex.Unlock();

        g_ppxVisibilitySegments[ uCurrentSegment ] = pxSegment;
    
        uCurrentSegment++;
        uSegmentsToCreate--;

        if ( uCurrentSegment > g_uMaxNumberOfVisibilitySegments/2 ) bMakeSamplesInsideFeatures = false;

        // And now start at a new point for the next segment.
        bCurSamplePointIsValid = false;
    }
}

bool Tools_Visibility_System::Split( void )
{
    u_int uCellsConsidered = 0;
    Tools_ViewCell* apxViewCellStack[32];
    u_int uStack = 0;

    g_uNumberOfLeafCells = 0;
    g_uNumberOfViewCells = 1;

    const u_int uTotalNumberOfCells = (2<<(g_uMaxTreeDepth))-1;

    g_ppxLeafCells = new Tools_ViewCell* [ ( uTotalNumberOfCells + 1 ) / 2 ];

    // push root onto stack
    apxViewCellStack[ uStack++ ] = g_pxEnvironmentViewCell;

    while ( uStack )
    {
        // pop something off the stack to work on
        Tools_ViewCell* pxViewCell = apxViewCellStack[ --uStack ];

        u_int uBestAxis = 0;
        float fBestPlane = 0.0f;

        // TODO - probably make this optional
        bool bMakeChildren = pxViewCell->HasInternalLinks();

        // sanity check on volume here, individual axis extents checked in FindBestSplittingPlane
        if (pxViewCell->GetDepth() == g_uMaxTreeDepth || pxViewCell->GetAABB().GetVolume() < fTOOLS_VIEWCELL_MIN_VOLUME )
        {
            bMakeChildren = false;
        }
        else
        {
            bool bForceSplit;
            if( !pxViewCell->FindBestSplittingPlane(uBestAxis, fBestPlane, bForceSplit) )
            {
                bMakeChildren = false;
            }
            else
            {
                if( bForceSplit )
                {
                    bMakeChildren = true;
                }
            }
        }

        if (bMakeChildren)
        {
            Asura_Bounding_Box xLeft;
            Asura_Bounding_Box xRight;
            pxViewCell->SplitBoundingBox( uBestAxis, fBestPlane, xLeft, xRight );

            pxViewCell->CreateLeftChild( xLeft );
            pxViewCell->CreateRightChild( xRight );

            g_uNumberOfViewCells += 2;

            // push children onto stack
            apxViewCellStack[ uStack++ ] = pxViewCell->GetLeftChild();
            apxViewCellStack[ uStack++ ] = pxViewCell->GetRightChild();

            // we've considered this node now, so up our progress count
            uCellsConsidered += 1;
        }
        else // We're a leaf node
        {
            pxViewCell->SetLeafIndex( g_uNumberOfLeafCells );
            g_ppxLeafCells[ g_uNumberOfLeafCells ] = pxViewCell;
            ++g_uNumberOfLeafCells;

            // We've considered this node - and by extension, any children (and their children... etc)
            // it could have had if we'd continued subdividing.
            uCellsConsidered += (2<< ( g_uMaxTreeDepth - pxViewCell->GetDepth() )) -1;
            
            pxViewCell->SetVisibleFeaturesArray();
        }

        Tools_Visibility_System::UpdateProgressIndicator( uCellsConsidered, uTotalNumberOfCells );

        // check for user trying to quit early
        if ( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) ) return true;
    }

    return false;
}


bool Tools_Visibility_System::SplitMT( void )
{
    bool bQuit = false;
    uCellsConsidered = 0;
    
    pxTaskQueue= new Asura_Task_Stack( "ViewcellSplit", 128, SplitMT_Callback );

    g_uNumberOfLeafCells = 0;
    g_uNumberOfViewCells = 1;

    const u_int uTotalNumberOfCells = (2<<(g_uMaxTreeDepth))-1;

    g_ppxLeafCells = new Tools_ViewCell* [ ( uTotalNumberOfCells + 1 ) / 2 ];

    pxTaskQueue->AddToStack( g_pxEnvironmentViewCell );

    while ( uCellsConsidered != uTotalNumberOfCells )
    {
        Asura_Thread_System::DelayCurrentThread( 0.3f, false );

        //g_pxProgressIndicator->AddMessage( "stack size %d", pxTaskQueue->GetCurrentQueueSize() );

        Tools_Visibility_System::UpdateProgressIndicator( uCellsConsidered, uTotalNumberOfCells );

        // check for user trying to quit early
        if ( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) )
        {
            bQuit = true;
            break;
        }
    }

    // destructor will auto-terminate
    delete pxTaskQueue;
    return bQuit;
}

#if 1
void Tools_Visibility_System::SplitMT_Callback( const void* pData )
{
    // pop something off the stack to work on
    Tools_ViewCell* pxViewCell = (Tools_ViewCell*) pData;

    do
    {
        u_int uBestAxis = 0;
        float fBestPlane = 0.0f;

        // TODO - probably make this optional
        bool bMakeChildren = pxViewCell->HasInternalLinks();

        // sanity check on volume here, individual axis extents checked in FindBestSplittingPlane
        if (pxViewCell->GetDepth() == g_uMaxTreeDepth || pxViewCell->GetAABB().GetVolume() < fTOOLS_VIEWCELL_MIN_VOLUME )
        {
            bMakeChildren = false;
        }
        else
        {
            bool bForceSplit;
            if( !pxViewCell->FindBestSplittingPlane(uBestAxis, fBestPlane, bForceSplit) )
            {
                bMakeChildren = false;
            }
            else
            {
                if( bForceSplit )
                {
                    bMakeChildren = true;
                }
            }
        }

        if (bMakeChildren)
        {
            Asura_Bounding_Box xLeft;
            Asura_Bounding_Box xRight;
            pxViewCell->SplitBoundingBox( uBestAxis, fBestPlane, xLeft, xRight );

            pxViewCell->CreateLeftChild( xLeft );
            pxViewCell->CreateRightChild( xRight );

            xViewCellData.Lock();
            g_uNumberOfViewCells += 2;
            xViewCellData.Unlock();

            // push one child onto the stack, and the other we'll do next
            pxTaskQueue->AddToStack( pxViewCell->GetLeftChild() );
            pxViewCell = pxViewCell->GetRightChild();

            // we've considered this node now, so up our progress count
            xCells.Lock();
            uCellsConsidered += 1;
            xCells.Unlock();
        }
        else // We're a leaf node
        {
            xViewCellData.Lock();
            pxViewCell->SetLeafIndex( g_uNumberOfLeafCells );
            g_ppxLeafCells[ g_uNumberOfLeafCells ] = pxViewCell;
            ++g_uNumberOfLeafCells;
            xViewCellData.Unlock();

            // We've considered this node - and by extension, any children (and their children... etc)
            // it could have had if we'd continued subdividing.
            xCells.Lock();
            uCellsConsidered += (2<< ( g_uMaxTreeDepth - pxViewCell->GetDepth() )) -1;
            xCells.Unlock();

            pxViewCell->SetVisibleFeaturesArray();

            // and we're done
            break;
        }
    }
    while ( pxViewCell && !(pxTaskQueue->GetShouldTerminateTasks()) );
}
#else

void Tools_Visibility_System::SplitMT_Callback( const void* pData )
{
    // pop something off the stack to work on
    Tools_ViewCell* pxViewCell = (Tools_ViewCell*) pData;

    u_int uBestAxis = 0;
    float fBestPlane = 0.0f;

    // TODO - probably make this optional
    bool bMakeChildren = pxViewCell->HasInternalLinks();

    // sanity check on volume here, individual axis extents checked in FindBestSplittingPlane
    if (pxViewCell->GetDepth() == g_uMaxTreeDepth || pxViewCell->GetAABB().GetVolume() < fTOOLS_VIEWCELL_MIN_VOLUME )
    {
        bMakeChildren = false;
    }
    else if ( !pxViewCell->FindBestSplittingPlane( uBestAxis, fBestPlane ) )
    {
        bMakeChildren = false;
    }

    if (bMakeChildren)
    {
        Asura_Bounding_Box xLeft;
        Asura_Bounding_Box xRight;
        pxViewCell->SplitBoundingBox( uBestAxis, fBestPlane, xLeft, xRight );

        pxViewCell->CreateLeftChild( xLeft );
        pxViewCell->CreateRightChild( xRight );

        xViewCellData.Lock();
        g_uNumberOfViewCells += 2;
        xViewCellData.Unlock();

        // push children onto stack

        pxTaskQueue->AddToStack( pxViewCell->GetLeftChild() );
        pxTaskQueue->AddToStack( pxViewCell->GetRightChild() );
        //apxViewCellStack[ uStack++ ] = pxViewCell->GetLeftChild();
        //apxViewCellStack[ uStack++ ] = pxViewCell->GetRightChild();

        // we've considered this node now, so up our progress count
        xCells.Lock();
        uCellsConsidered += 1;
        xCells.Unlock();
    }
    else // We're a leaf node
    {
        xViewCellData.Lock();
        pxViewCell->SetLeafIndex( g_uNumberOfLeafCells );
        g_ppxLeafCells[ g_uNumberOfLeafCells ] = pxViewCell;
        ++g_uNumberOfLeafCells;
        xViewCellData.Unlock();

        // We've considered this node - and by extension, any children (and their children... etc)
        // it could have had if we'd continued subdividing.
        xCells.Lock();
        uCellsConsidered += (2<< ( g_uMaxTreeDepth - pxViewCell->GetDepth() )) -1;
        xCells.Unlock();

        pxViewCell->SetVisibleFeaturesArray();
    }
}
#endif

bool Tools_Visibility_System::CreateLeafToLeafVisibilties( void )
{
    Tools_ViewCell** ppxLeafCells = new Tools_ViewCell* [g_uNumberOfLeafCells];

    for (u_int u=0; u<g_uCurrentNumberOfVisibilitySegments; ++u)
    {
        const Tools_VisibilitySegment* pxSegment = g_ppxVisibilitySegments[u];

        Tools_Visibility_System::UpdateProgressIndicator( u, g_uCurrentNumberOfVisibilitySegments );

        u_int uNumberOfVisibleLeaves = 0;

        Tools_ViewCell* apxViewCellStack[32];
        u_int uStack = 0;

        // push root onto stack
        apxViewCellStack[ uStack++ ] = g_pxEnvironmentViewCell;

        while ( uStack )
        {
            // pop something off the stack to work on
            Tools_ViewCell* pxViewCell = apxViewCellStack[ --uStack ];

            if ( pxViewCell->AABBContainsVisibilitySegment( pxSegment ) )
            {
                if ( pxViewCell->GetLeftChild() && pxViewCell->GetRightChild() )
                {
                    // push children onto stack
                    apxViewCellStack[ uStack++ ] = pxViewCell->GetLeftChild();
                    apxViewCellStack[ uStack++ ] = pxViewCell->GetRightChild();
                }
                else // We're a leaf node
                {
                    ppxLeafCells[ uNumberOfVisibleLeaves++ ] = pxViewCell;
                }
            }
        }

        for (u_int v=0; v<uNumberOfVisibleLeaves; ++v)
        {
            ppxLeafCells[ v ]->SetLeavesAsVisible( ppxLeafCells, uNumberOfVisibleLeaves, g_uNumberOfLeafCells );
        }

        // check for user trying to quit early
        if ( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) )
        {
            delete [] ppxLeafCells;
            return true;
        }
    }
    
    delete [] ppxLeafCells;

    return false;
}


bool Tools_Visibility_System::IsExternalFeature( u_int uFeatureID )
{
    // the >= also rejects zones greater than 0 or out of range module indices
    if ( uFeatureID >= GetSkyBoxFeatureID() ) return true;

    const Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();

    if( pxEnv && pxEnv->IsSceneryModule( uFeatureID ) )
    {
        return true;
    }

    return false;
}

void Tools_Visibility_System::RefineVisibility( const Asura_Vector_3& xSamplePoint, const Asura_Vector_3& xDirection )
{
    if ( !Asura_Visibility_System::s_uNumberOfNodes ) return;

    {

        Asura_Vector_3 xSegment0;
        Asura_Vector_3 xSegment1;
        u_int           uFeatureID0;
        u_int           uFeatureID1;

        bool bFrontFacing;
        Asura_Vector_3 xNormal;
        bool bHitVisBlocker;

        bool bHit1 = FireRay_Internal( xSamplePoint, xDirection, xSegment0, bFrontFacing, uFeatureID0, xNormal, bHitVisBlocker );

        if ( bHitVisBlocker )
        {
            // abandon the segment
            return;
        }
        else if( bHit1 && !bFrontFacing )
        {
            // abandon the segment
            return;
        }

        bool bHit2 = FireRay_Internal( xSamplePoint, -xDirection, xSegment1, bFrontFacing, uFeatureID1, xNormal, bHitVisBlocker );

        if ( bHitVisBlocker )
        {
            // abandon the segment
            return;
        }
        else if (!bHit1 && !bHit2)
        {
            // abandon the segment
            return;
        }

        if( bHit2 && !bFrontFacing )
        {
            // abandon the segment
            return;
        }

        // we have a new segment


        u_int* puLeafIndices = new u_int [Asura_Visibility_System::s_uNumberOfLeaves];
        u_int uNumberOfVisibleLeaves = 0;

        u_int apxViewCellStack[32];
        u_int uStack = 0;

        // push root onto stack
        apxViewCellStack[ uStack++ ] = 0;

        while ( uStack )
        {
            // pop something off the stack to work on
            const Asura_Visibility_Node& xViewCell = Asura_Visibility_System::s_pxVisibilityNodes[ apxViewCellStack[ --uStack ] ];

            if ( xViewCell.RaySegmentIntersectsViewCell( xSegment0, xSegment1 ) )
            {
                if ( xViewCell.GetLeftChild() )
                {
                    // push children onto stack
                    apxViewCellStack[ uStack++ ] = xViewCell.GetLeftChild();
                    apxViewCellStack[ uStack++ ] = xViewCell.GetRightChild();
                }
                else // We're a leaf node
                {
                    puLeafIndices[ uNumberOfVisibleLeaves++ ] = xViewCell.GetLeafIndex();
                }
            }
        }
        
        // TODO - this is a big lock here. Need to optimise it. Possibly individual mutexes per leaf? Seems OTT.
        {
            xViewCellData.Lock();

            bool bAnyVisibilitiesIntroduced = false;

            for (u_int v=0; v<uNumberOfVisibleLeaves; ++v)
            {
                u_int uLeafIndexOuter = puLeafIndices[v];

                Asura_Assert( uLeafIndexOuter < Asura_Visibility_System::s_uNumberOfLeaves, "Invalid leaf index" );

                Asura_Visibility_Leaf& xLeafOuter = Asura_Visibility_System::s_pxVisibilityLeaves[ uLeafIndexOuter ];

                if ( xLeafOuter.Tools_RefineFeatureVisibility( uFeatureID0, uFeatureID1 ) )
                {
                    bAnyVisibilitiesIntroduced = true;
                }           

                for ( u_int u=0; u<uNumberOfVisibleLeaves; ++u )
                {
                    u_int uLeafIndex = puLeafIndices[u];

                    Asura_Assert( uLeafIndex < Asura_Visibility_System::s_uNumberOfLeaves, "Invalid leaf index" );

                    if ( xLeafOuter.Tools_RefineLeafVisibility( uLeafIndex ) )
                    {
                        bAnyVisibilitiesIntroduced = true;
                    }
                }
            }

            // If this vis segment introduces any new visibilities then store
            // it for later debugging purposes.
            if( bAnyVisibilitiesIntroduced )
            {
                SetAsModified();
                g_bRenderVisSegments_ForceFilteredListRecalc = true;

                Tools_VisibilitySegment::Create( true, xSegment0, xSegment1, uFeatureID0, uFeatureID1 );
            }

            xViewCellData.Unlock();
        }

        delete [] puLeafIndices;
    }
}

void Tools_Visibility_System::RemoveLinks( void )
{
    if ( !Asura_Visibility_System::s_uNumberOfNodes ) return;

    if ( !Asura_Visibility_System::s_apxCameraLeaf[0] ) return;

    Asura_Visibility_System::s_apxCameraLeaf[0]->Tools_RemoveLinks();
    
    Asura_Visibility_System::s_bForceVisRecalcForCurrentCameraNode = true;
}

void Tools_Visibility_System::AutoRefine( void )
{
    if ( !Asura_Visibility_System::s_uNumberOfNodes ) return;

    if ( g_bUseMultithreading )
    {
        AutoRefine_MT();
        return;
    }

    Tools_Environment_LOS::EnsureDataIsValid();

    RecalculateBaseEnvironmentData();

    InitialiseProgressIndicator();
    StartProgressIndicator( "Refining - rays cast:" );

    if( g_pfnStartVisCalcsCallback )
    {
        g_pfnStartVisCalcsCallback();
    }

    // improve visibilities
    float fEndTime = Asura_Timers::GetUncappedPhase() + (g_uAutoRefineTimeInMins * 60.0f);

    Asura_Timers::SetMaxFrameRate(0.0f);
    Asura_Timers::HandlePrimaryTimer();

    u_int uRay=0;
    do
    {
        // don't update for every ray
        if ( (uRay & 4095) == 0)
        {
            Tools_Visibility_System::UpdateProgressIndicator( uRay, 0, true );
        }

        Asura_Vector_3 xDirection;
        Asura_Vector_3 xPosition;

        xDirection.RandomDirection();
        
        bool bMakeSampleInsideFeatures = (uRay++)&1;

        if( GetRandomPoint(xPosition, bMakeSampleInsideFeatures) )
        {
            Tools_Visibility_System::RefineVisibility( xPosition, xDirection );
        }

        // no need to clear up any memory etc if we want to exit early
        if ( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_ESCAPE ) ) break;

        Asura_Timers::HandlePrimaryTimer();
        // check for running out of time
        if ( fEndTime<Asura_Timers::GetUncappedPhase() )
        {
            break;
        }
    }
    while( true );

    if( g_pfnEndVisCalcsCallback )
    {
        g_pfnEndVisCalcsCallback();
    }

    ShutdownProgressIndicator();
}

void Tools_Visibility_System::AutoRefine_MT( void )
{
    if ( !Asura_Visibility_System::s_uNumberOfNodes ) return;

    Tools_Environment_LOS::EnsureDataIsValid();

    RecalculateBaseEnvironmentData();

    InitialiseProgressIndicator();
    StartProgressIndicator( "Refining - rays cast:" );

    // Second progress indicator to track how much of the specified time has
    // been used.
    Tools_ProgressIndicator xTimeIndicator( "Time spent (minutes):" );
    xTimeIndicator.SetShowAsPercentage( false );

    if( g_pfnStartVisCalcsCallback )
    {
        g_pfnStartVisCalcsCallback();
    }

    // improve visibilities
    u_int uMaxRays = 1000000000; // stop at a billion rays

    const u_int uStartMsecs = Asura_Timers::GetMSecsSinceStart_Continuous();

    // (MP - 11.02.2009) Put the task array inside a nested scope so that
    // it's destructed before we consider processing as complete. Otherwise
    // we can get issues in the callback where it may try to deallocate stuff
    // that the final tasks are still using.
    {
        Asura_Task_Array xTaskArray( uMaxRays, 1023, AutoRefine_MT_Callback );
        
        Asura_Timers::SetMaxFrameRate(0.0f);
        Asura_Timers::HandlePrimaryTimer();

        while ( !xTaskArray.HasFinished() )
        {
            Asura_Thread_System::DelayCurrentThread( 0.3f, false );

            Asura_Timers::HandlePrimaryTimer();

            Tools_Visibility_System::UpdateProgressIndicator( xTaskArray.GetCurrentIndex(), 0, true );

            u_int uElapsedMinutes = (Asura_Timers::GetMSecsSinceStart_Continuous() - uStartMsecs) / (60 * 1000);
            xTimeIndicator.Update( uElapsedMinutes, g_uAutoRefineTimeInMins, true );

            // check for user trying to quit early
            if ( Asura_UserInput::KeyPressed( ASURA_KEYID_ESCAPE ) )
            {
                break;
            }

            // check for running out of time
            if( uElapsedMinutes >= g_uAutoRefineTimeInMins )
            {
                break;
            }
        }
    }

    if( g_pfnEndVisCalcsCallback )
    {
        g_pfnEndVisCalcsCallback();
    }

    ShutdownProgressIndicator();
}

void Tools_Visibility_System::AutoRefine_MT_Callback( const void* pxTaskDataPacket )
{
    const Asura_Task_Array::TaskDataPacket* pxData = static_cast<const Asura_Task_Array::TaskDataPacket*>(pxTaskDataPacket);

    u_int uCurrentSegment = pxData->GetArrayStartIndex();
    u_int uSegmentsToCreate = pxData->GetNumberToProcess();

    // Tell the Asura_Task_Array we're done with this data packet. We won't be able to read from it after this.
    pxData->SignalFinishedWithData();

    do 
    {
        Asura_Vector_3 xDirection;
        Asura_Vector_3 xPosition;

        xDirection.RandomDirection();

        bool bMakeSampleInsideFeatures = (uCurrentSegment++)&1;

        if( GetRandomPoint(xPosition, bMakeSampleInsideFeatures) )
        {
            Tools_Visibility_System::RefineVisibility( xPosition, xDirection );
        }
    }
    while(--uSegmentsToCreate);
}


bool Tools_Visibility_System::GetRandomPoint( Asura_Vector_3& xPoint, bool bInAVisibilityFeature )
{
    Asura_Bounding_Box xAABB;
    
    if( bInAVisibilityFeature )
    {
        const Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
        if( !pxEnv ) { return false; }

        u_int uTry;
        for( uTry = 0; uTry < uMAX_RANDOM_POINT_RETRIES; ++uTry )
        {
            u_int uRandomFeature = Asura_Random::GetIntU( 0, Asura_Module_System::GetNumberOfModules()-1 );

            if( pxEnv->IsSceneryModule(uRandomFeature) ) { continue; }

            xAABB = Asura_Module_System::GetBoundingBoxOfModule(uRandomFeature);
            xAABB.Displace( Asura_Module_System::GetPositionOfModule(uRandomFeature) );
        }

        if( uTry == uMAX_RANDOM_POINT_RETRIES ) { return false; }
    }
    else // in the world
    {
        xAABB = g_xEnvironmentAABB;
    }

    return GetRandomPointInAABB( xPoint, xAABB );
}


bool Tools_Visibility_System::GetRandomPointInAABB( Asura_Vector_3& xPoint, const Asura_Bounding_Box& xAABB )
{
    for( u_int u = 0; u < uMAX_RANDOM_POINT_RETRIES; ++u )
    {
        const float fX = Asura_Random::GetFloat();
        const float fY = Asura_Random::GetFloat();
        const float fZ = Asura_Random::GetFloat();

        xPoint.SetX( (1.0f - fX) * xAABB.MinX + fX * xAABB.MaxX );
        xPoint.SetY( (1.0f - fY) * xAABB.MinY + fY * xAABB.MaxY );
        xPoint.SetZ( (1.0f - fZ) * xAABB.MinZ + fZ * xAABB.MaxZ );

        if( IsVisSegmentSamplePointValid(xPoint) )
        {
            return true;
        }
    }

    return false;
}


bool Tools_Visibility_System::IsVisSegmentSamplePointValid( const Asura_Vector_3& xSamplePoint )
{
#if 1

    return true;

#else

    // See what environment polys lie directly beneath this point. Don't
    // allow sample points where we encounter nothing because those aren't
    // generally helpful in any way. It would seem that the only way you're
    // going to end up in such a situation is perhaps when rendering a
    // cutscene (ie. so we know the camera won't look down at this point and
    // it's ok not to have any environment there) or in more perverse cases
    // like the space zone in BF3. The latter could be solved by allowing
    // volumes to be setup that force all points within them to be considered
    // valid. However, in fact even in that BF3 case the space zone would
    // probably have worked ok since there was generally a planet below all
    // points in which you could fly.

    Tools_Environment_LOS_Data xLOSData( xSamplePoint, Asura_Vector_3(0.0f, 1.0f, 0.0f) );
    xLOSData.SetPolyFlagsToIgnore( AsuraPolygonFlag_DontRender | AsuraPolygonFlag_DontExport );
    xLOSData.SetReturnOnFirstHit( true );

// TODO: Not sure whether we'd want to respect the module filter callback
// here or not. No projects are currently using it though so it's a bit
// academic.
//  if( g_pfnRayModuleTestFilterCallback )
//  {
//      xLOSData.SetShouldTestModuleCallback( g_pfnRayModuleTestFilterCallback, g_pvRayModuleTestFilterUserData );
//  }
    
    if( Tools_Environment_LOS::TestWithRay(xLOSData) )
    {
        // TODO: Not sure whether or not to count back-facing polys. Erring
        // on the side of behaving before these checks were made for now, ie.
        // counting them.
        return true;
    }

    return false;

#endif
}


void Tools_Visibility_System::GetRandomDirection( Asura_Vector_3& xDirection )
{
    xDirection.RandomDirection();
}


bool Tools_Visibility_System::FireSingleRay( const Asura_Vector_3& xStart, const Asura_Vector_3& xDirection, Asura_Vector_3& xHitPoint, bool& bFrontFacing, u_int& uFeatureID, Asura_Vector_3& xNormal, bool& bHitVisBlocker )
{
    Tools_Environment_LOS::EnsureDataIsValid();

    RecalculateBaseEnvironmentData();

    if( g_pfnStartVisCalcsCallback )
    {
        g_pfnStartVisCalcsCallback();
    }

    bool bHit = FireRay_Internal( xStart, xDirection, xHitPoint, bFrontFacing, uFeatureID, xNormal, bHitVisBlocker );
    
    if( g_pfnEndVisCalcsCallback )
    {
        g_pfnEndVisCalcsCallback();
    }

    return bHit;
}


// returns true if hit a polygon
bool Tools_Visibility_System::FireRay_Internal( const Asura_Vector_3& xStart, const Asura_Vector_3& xDirection, Asura_Vector_3& xHitPoint, bool& bFrontFacing, u_int& uFeatureID, Asura_Vector_3& xNormal, bool& bHitVisBlocker )
{
    bHitVisBlocker = false;

    // Lo-tech solution for what to do about transparent polys - ignore them half of the time.
    // This means we both "see" through them while still registering hits with the polys/features themselves.
    const bool bIgnoreTransparent = Asura_Random::GetFloat() > 0.5f;

    Tools_Environment_LOS_Data xLOSData(xStart, xDirection);
    xLOSData.SetPolyFlagsToIgnore( AsuraPolygonFlag_DontRender);
    xLOSData.SetIgnoreTransparent( bIgnoreTransparent);

    if( g_pfnRayModuleTestFilterCallback )
    {
        xLOSData.SetShouldTestModuleCallback( g_pfnRayModuleTestFilterCallback, g_pvRayModuleTestFilterUserData );
    }
    
    if ( Tools_Environment_LOS::TestWithRay( xLOSData))
    {
        bFrontFacing = xLOSData.GetFacingTowards();
        const u_int uModuleIndex = xLOSData.GetModuleIndex();
        const u_int uPolyIndex = xLOSData.GetPolyIndex();
        const float fDistance = xLOSData.GetDistance();

        Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule( uModuleIndex );

        const int iFlags = pxMesh->GetPolygonFlags( uPolyIndex );
        
        // check for things flagged as a vis blocker: currently using the DontExport flag
        // Otherwise see if any callback wants to have a say in whether or
        // not a visibility link is valid here; treat that as being equivalent
        // to hitting a vis blocking poly.
        if( (iFlags & AsuraPolygonFlag_DontExport)
            ||
            (g_pfnRayHitFilterCallback && !g_pfnRayHitFilterCallback(xStart, xDirection, xStart + xDirection * fDistance, uModuleIndex, uPolyIndex)) )
        {
            bHitVisBlocker = true;
            return false;
        }

        // if it's flagged as doubled sided, it's always front facing
        if ( iFlags & AsuraPolygonFlag_DoubleSided )
        {
            bFrontFacing = true;
        }

        xHitPoint = xStart + xDirection * fDistance;
        xNormal = pxMesh->NormalsDataPtr[ uPolyIndex ].FaceNormal;
        uFeatureID = uModuleIndex;

        return true;
    }
    else
    {
        // We've hit the skybox. Extend the segment to where it intersects
        // with the outside of the env BB, to make debugging displays more
        // obvious and to ensure we cover all the vis nodes in this direction
        // (since it's equivalent to having created a sample point at this
        // intersection aimed in the opposite direction, which would
        // eventually happen during refinement anyway).
        bool bHit, bInside;
        bHit = Asura_Maths::RayIntersectsBoundingBox_AllowInside( xStart, xDirection, g_xEnvironmentAABB, NULL, &xHitPoint, &bInside, &xNormal );
        Asura_Assert( bHit && bInside, "" );

        // Fix up the feature ID and 'front facing' flag.
        uFeatureID = GetSkyBoxFeatureID();
        bFrontFacing = true;

        return false;
    }

}


bool Tools_Visibility_System::ShouldRenderVisibleNodeBBs( void )
{
    return g_bRenderVisibleNodeBBs;
}


void Tools_Visibility_System::RenderVisibleNodeBBs( void )
{
    if( !Asura_Visibility_System::s_uNumberOfNodes )
    {
        return;
    }

    u_int auVisNodeStack[32];
    u_int uStackIdx = 0;

    // push root onto stack
    auVisNodeStack[ uStackIdx++ ] = 0;

    while( uStackIdx )
    {
        // pop something off the stack to work on
        u_int uNodeIndex = auVisNodeStack[ --uStackIdx ];
        const Asura_Visibility_Node& xVisNode = Asura_Visibility_System::s_pxVisibilityNodes[ uNodeIndex ];

        if( Asura_Visibility_System::s_pxNodeFlags[uNodeIndex].GetORedVisibilityFlags() )
        {
            if ( xVisNode.GetLeftChild() )
            {
                // push children onto stack
                auVisNodeStack[ uStackIdx++ ] = xVisNode.GetLeftChild();
                auVisNodeStack[ uStackIdx++ ] = xVisNode.GetRightChild();
            }
            else // We're a leaf node
            {
                Tools_Render::RenderCube( xVisNode.GetAABB(), 0xFF00FF00, true );
            }
        }
    }

    // Additional highlights:

    // The camera's current node.
    if( g_bRenderVisibleNodeBBs_HighlightCamNode )
    {
        const u_int uCamNodeIdx = Asura_Visibility_System::s_auLastKnownCameraNode[0];
        if( uCamNodeIdx )
        {
            const Asura_Bounding_Box& xCamNodeBB = Asura_Visibility_System::s_pxVisibilityNodes[uCamNodeIdx].GetAABB();
            if( xCamNodeBB.Contains(*Asura_Camera::GetPosition()) )
            {
                Tools_Render::RenderCube( xCamNodeBB, 0x4F0000FF, false );
            }
            else
            {
                Tools_Render::RenderCube( xCamNodeBB, 0x7F0000FF, false );
            }
        }
    }

    // The current target node for rendering vis segments.
    if( g_bRenderVisibleNodeBBs_HighlightVisSegmentTargetNode )
    {
        if( g_bRenderVisSegments_OnlyForSelectedNodes && (g_uRenderVisSegments_TargetNode != uASURA_VISIBILITY_INVALID_NODE_INDEX) )
        {
            const Asura_Bounding_Box& xTargetNodeBB = Asura_Visibility_System::s_pxVisibilityNodes[g_uRenderVisSegments_TargetNode].GetAABB();
            if( xTargetNodeBB.Contains(*Asura_Camera::GetPosition()) )
            {
                Tools_Render::RenderCube( xTargetNodeBB, 0xFFFF0000, true );
            }
            else
            {
                Tools_Render::RenderCube( xTargetNodeBB, 0x7FFF0000, false );
            }
        }
    }
}


bool Tools_Visibility_System::RenderVisSegments_SelectTargetNode_PickCallback( bool bClicked )
{
    g_uRenderVisSegments_CurrentlyPickedNode = FindVisNodeUnderCursor();
    if( g_uRenderVisSegments_CurrentlyPickedNode != uASURA_VISIBILITY_INVALID_NODE_INDEX )
    {
        if( bClicked )
        {
            g_uRenderVisSegments_TargetNode = g_uRenderVisSegments_CurrentlyPickedNode;
            g_uRenderVisSegments_CurrentlyPickedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
        }
        return true;
    }
    else
    {
        return false;
    }
}


void Tools_Visibility_System::RenderVisSegments_SelectTargetNode_RenderCallback( u_int uSuggestedARGB )
{
    if( g_uRenderVisSegments_CurrentlyPickedNode < Asura_Visibility_System::s_uNumberOfNodes )
    {
        const Asura_Visibility_Node& xVisNode = Asura_Visibility_System::s_pxVisibilityNodes[g_uRenderVisSegments_CurrentlyPickedNode];
        Tools_Render::RenderCube( xVisNode.GetAABB(), uSuggestedARGB, false );
    }
}


void Tools_Visibility_System::RenderVisSegments_SelectTargetNode( void )
{
    if( !Tools_Prompted_Selection::IsSelecting() )
    {
        Tools_Prompted_Selection::ChooseCustom( "Choose new vis segments target node", RenderVisSegments_SelectTargetNode_PickCallback, RenderVisSegments_SelectTargetNode_RenderCallback );
    }
}


void Tools_Visibility_System::RenderVisSegments_ClearTargetNode( void )
{
    g_uRenderVisSegments_TargetNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
}


bool Tools_Visibility_System::RenderVisSegments_ShouldRender( void )
{
    return g_bRenderVisSegments;
}


bool Tools_Visibility_System::RenderVisSegments_ShouldBeAllowedToSelectTargetNode( void )
{
    return( g_bRenderVisSegments && g_bRenderVisSegments_OnlyForSelectedNodes );
}


void Tools_Visibility_System::RenderVisSegments_RenderSingleSegment( const Tools_VisibilitySegment* pxSegment )
{
    const u_int uARGB_EXTERNAL = 0xFFFF0000; // Hit the skybox or a scenery module.
    const u_int uARGB_INTERNAL = 0xFFFFFF00; // Hit a non-scenery module.

    u_int uARGB0 = ( IsExternalFeature(pxSegment->GetVisibleFeatureID(0)) ? uARGB_EXTERNAL : uARGB_INTERNAL );
    u_int uARGB1 = ( IsExternalFeature(pxSegment->GetVisibleFeatureID(1)) ? uARGB_EXTERNAL : uARGB_INTERNAL );

    // (MP - 16.03.2009) Ideally Asura_Render would support lines
    // with different colours at either vertex but I didn't want to
    // go changing every platform's code for such a limited need.
    Asura_Direct3D_LineRenderer::Line3D xLine( pxSegment->GetPoint(0), pxSegment->GetPoint(1), uARGB0, uARGB1 );
    Asura_Direct3D_LineRenderer::AddLine( xLine ); 
}


void Tools_Visibility_System::RenderVisSegments( void )
{
    // If we're filtering what's rendered then pre-calculate a list. This
    // may get quite large but it's still better than filtering every frame
    // because it's possible to generate millions of segments and the
    // framerate is still fairly unusable.
    RenderVisSegments_GenerateFilteredList();

    if( g_bRenderVisSegments_GotFilteredList )
    {
        for( u_int u = 0; u < g_uRenderVisSegments_NumFilteredSegments; ++u )
        {
            RenderVisSegments_RenderSingleSegment( g_ppxRenderVisSegments_FilteredSegments[u] );
        }
    }
    else
    {
        const Asura_RingBuffer<Tools_VisibilitySegment>& xRefinementSegments = Tools_VisibilitySegment::GetRefinementSegments();
        for( u_int uSource = 0; uSource < 2; ++uSource )
        {
            const u_int uCount = ( uSource ? xRefinementSegments.GetCount() : g_uCurrentNumberOfVisibilitySegments );
            for( u_int u = 0; u < uCount; ++u )
            {
                RenderVisSegments_RenderSingleSegment( uSource ? xRefinementSegments.Peek(u) : g_ppxVisibilitySegments[u] );
            }
        }
    }
}


bool Tools_Visibility_System::RenderVisSegments_Filter( const Tools_VisibilitySegment* pxSegment )
{
    u_int uCamNode = Asura_Visibility_System::s_auLastKnownCameraNode[0];

    const Asura_Bounding_Box* pxCamNodeBB = NULL;
    if( uCamNode < Asura_Visibility_System::s_uNumberOfNodes )
    {
        if( Asura_Visibility_System::s_pxVisibilityNodes[uCamNode].GetLeafIndex() )
        {
            pxCamNodeBB = &(Asura_Visibility_System::s_pxVisibilityNodes[uCamNode].GetAABB());
        }
    }

    const Asura_Bounding_Box* pxTargetNodeBB = NULL;
    if( g_uRenderVisSegments_TargetNode != uASURA_VISIBILITY_INVALID_NODE_INDEX )
    {
        pxTargetNodeBB = &(Asura_Visibility_System::s_pxVisibilityNodes[g_uRenderVisSegments_TargetNode].GetAABB());
    }

    // If we're missing both nodes (the one containing the camera and the
    // target one) then don't render anything.
    if( !pxCamNodeBB && !pxTargetNodeBB )
    {
        return false;
    }

    Asura_Vector_3 xRayOrigin = pxSegment->GetPoint(0);
    Asura_Vector_3 xRayDir = pxSegment->GetPoint(1) - xRayOrigin;
    float fRayLength = xRayDir.Normalise();

    float fDist;

    if( pxCamNodeBB )
    {
        if( !Asura_Maths::RayIntersectsBoundingBox(xRayOrigin, xRayDir, *pxCamNodeBB, &fDist) || (fDist > fRayLength) )
        {
            return false;
        }
    }

    if( pxTargetNodeBB )
    {
        if( !Asura_Maths::RayIntersectsBoundingBox(xRayOrigin, xRayDir, *pxTargetNodeBB, &fDist) || (fDist > fRayLength) )
        {
            return false;
        }
    }

    return true;
}


void Tools_Visibility_System::RenderVisSegments_GenerateFilteredList( void )
{
    if( g_bRenderVisSegments_OnlyForSelectedNodes )
    {
        static u_int uPrevCamNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
        static u_int uPrevTargetNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
        static float fLastGenerationTimeStamp = 0.0f;

        const u_int uCurCamNode = Asura_Visibility_System::s_auLastKnownCameraNode[0];

        // Is the current list up to date? Note that if data has been
        // recalculated or anything then this state will have been fully
        // reset elsewhere.
        if( g_bRenderVisSegments_GotFilteredList && !g_bRenderVisSegments_ForceFilteredListRecalc )
        {
            if( (uCurCamNode == uPrevCamNode) && (g_uRenderVisSegments_TargetNode == uPrevTargetNode) )
            {
                // Yes.
                return;
            }
        }

        const float fCurTime = Asura_Timers::GetLocalUnscaledTime();

        // Don't regenerate the list too often. It's very time consuming with
        // a large number of segments (hence why we store the results).
        if( fCurTime - fLastGenerationTimeStamp < 1.0f )
        {
            return;
        }

        // Junk any existing list.
        delete[] g_ppxRenderVisSegments_FilteredSegments;
        g_ppxRenderVisSegments_FilteredSegments = NULL;
        g_uRenderVisSegments_NumFilteredSegments = 0;

        // Count segments.
        const Asura_RingBuffer<Tools_VisibilitySegment>& xRefinementSegments = Tools_VisibilitySegment::GetRefinementSegments();
        for( u_int uSource = 0; uSource < 2; ++uSource )
        {
            const u_int uCount = ( uSource ? xRefinementSegments.GetCount() : g_uCurrentNumberOfVisibilitySegments );
            for( u_int u = 0; u < uCount; ++u )
            {
                if( RenderVisSegments_Filter(uSource ? xRefinementSegments.Peek(u) : g_ppxVisibilitySegments[u]) )
                {
                    ++g_uRenderVisSegments_NumFilteredSegments;
                }
            }
        }

        // And run back through again storing them.
        if( g_uRenderVisSegments_NumFilteredSegments )
        {
            g_ppxRenderVisSegments_FilteredSegments = new const Tools_VisibilitySegment*[g_uRenderVisSegments_NumFilteredSegments];

            u_int uOut = 0;
            for( u_int uSource = 0; uSource < 2; ++uSource )
            {
                const u_int uCount = ( uSource ? xRefinementSegments.GetCount() : g_uCurrentNumberOfVisibilitySegments );
                for( u_int u = 0; u < uCount; ++u )
                {
                    const Tools_VisibilitySegment* pxSegment = ( uSource ? xRefinementSegments.Peek(u) : g_ppxVisibilitySegments[u] );
                    if( RenderVisSegments_Filter(pxSegment) )
                    {
                        g_ppxRenderVisSegments_FilteredSegments[uOut++] = pxSegment;
                    }
                }
            }
            Asura_Assert( uOut == g_uRenderVisSegments_NumFilteredSegments, "" );
        }

        uPrevCamNode = uCurCamNode;
        uPrevTargetNode = g_uRenderVisSegments_TargetNode;
        fLastGenerationTimeStamp = fCurTime;
        g_bRenderVisSegments_GotFilteredList = true;
        g_bRenderVisSegments_ForceFilteredListRecalc = false;
    }
    else
    {
        RenderVisSegments_ClearFilteredList();
    }
}


void Tools_Visibility_System::RenderVisSegments_ClearFilteredList( void )
{
    delete[] g_ppxRenderVisSegments_FilteredSegments;
    g_ppxRenderVisSegments_FilteredSegments = NULL;
    g_uRenderVisSegments_NumFilteredSegments = 0;
    g_bRenderVisSegments_GotFilteredList = false;
    g_bRenderVisSegments_ForceFilteredListRecalc = true;
}


void Tools_Visibility_System::DebugRender( void )
{ 
    // need an update method instead?

    if( !IsOverlayOpen() )
    {
        return;
    }

    if( g_bRenderVisibleNodeBBs )
    {
        RenderVisibleNodeBBs();
    }

    // don't try to Refine in Mesh Edit mode - things will crawl to a halt if the mesh is being modified.
    if( (g_bAutoRefinementFromCameraPosition || g_bAutoRefinementFromCameraNode || g_bAutoRefinementUnderCursor) && ( Tools_System::GetCurrentMode() != TOOLS_MODE_MESHEDIT ) )
    {
        DoIncrementalAutoRefinement();
    }

    if( g_bRenderVisSegments )
    {
        RenderVisSegments();
    }

    if( g_bRenderSplitHintQuads )
    {
        RenderSplitHintQuads();
    }

#if 0
    const u_int uNumBoxes = 1000;
    static Asura_Bounding_Box axAABB[uNumBoxes];
    static bool bob = true;

    if (bob)
    {
        bob = false;
    for (u_int u=0; u<uNumBoxes; ++u)
    {
        Asura_Vector_3 xCentre;

        GetRandomPoint( xCentre, true );

        axAABB[u].Reset();
        axAABB[u].Set( xCentre );

        axAABB[u].Grow( Asura_Random::GetFloat() * 100.0f );
    }
    }
    
    {

        u_int uInside = 0;
        Asura_TickCount uStartTicks = Asura_Timers::GetCurTicks();  
        for (u_int u=0; u<uNumBoxes; ++u)
        {
            Asura_Vector_3 xStart, xDirection;
            GetRandomPoint( xStart, true );
            GetRandomDirection( xDirection );

            
            Asura_LOS_InputData xData( xStart, xDirection, 0 );

            //Asura_LOS::TestAgainstEnvironment( &xData );
        }
        Asura_TickCount uTicks = Asura_Timers::GetCurTicks() - uStartTicks;
        float fTime = Asura_Timers::TicksToSeconds(uTicks) * 1000000.0f;

        Asura_DebugText::AddTextBlock( 10, 3, 0xffffffff, 0, "New Time (microsecs): %.1f; Inside: %d", fTime, uInside );

    }

    {

        {
        u_int uInside = 0;

        Asura_TickCount uStartTicks = Asura_Timers::GetCurTicks();  
        for (u_int u=0; u<uNumBoxes; ++u)
        {
            //bool b2 = 
            if (Asura_Camera::IsSphereInFrustum( axAABB[u].CalculateCentre(),axAABB[u].CalculateRadius() )) uInside++;
        }
        Asura_TickCount uTicks = Asura_Timers::GetCurTicks() - uStartTicks;
        float fTime = Asura_Timers::TicksToSeconds(uTicks) * 1000000.0f;

        Asura_DebugText::AddTextBlock( 10, 8, 0xffffffff, 0, "Sphere2 Time (microsecs): %.1f; Inside: %d", fTime, uInside );

    }
                u_int uInside = 0;

        Asura_TickCount uStartTicks = Asura_Timers::GetCurTicks();  
        Asura_Vector_3 xCentre = axAABB[10].CalculateCentre();
        
        float fRadius = axAABB[10].CalculateRadius();

        for (u_int u=0; u<uNumBoxes; ++u)
        {
            //bool b2 = 
            if (Asura_Camera::IsSphereInFrustum( xCentre, fRadius )) uInside++;
        }
        Asura_TickCount uTicks = Asura_Timers::GetCurTicks() - uStartTicks;
        float fTime = Asura_Timers::TicksToSeconds(uTicks) * 1000000.0f;

        Asura_DebugText::AddTextBlock( 10, 7, 0xffffffff, 0, "Sphere1 Time (microsecs): %.1f; Inside: %d", fTime, uInside );

    }
    {

        u_int uInside = 0;
        Asura_TickCount uStartTicks = Asura_Timers::GetCurTicks();  
        for (u_int u=0; u<uNumBoxes; ++u)
        {
            Asura_Vector_3 xStart, xDirection;
            GetRandomPoint( xStart, true );
            GetRandomDirection( xDirection );

            Asura_LOS_InputData xData( xStart, xDirection, 0 );

            //Asura_LOS::TestAgainstEnvironment( &xData );
        }
        Asura_TickCount uTicks = Asura_Timers::GetCurTicks() - uStartTicks;
        float fTime = Asura_Timers::TicksToSeconds(uTicks) * 1000000.0f;

        Asura_DebugText::AddTextBlock( 10, 5, 0xffffffff, 0, "New Time (microsecs): %.1f; Inside: %d", fTime, uInside );

    }
#endif
}


void Tools_Visibility_System::ManuallySelectCameraNode( void )
{
    if( !Tools_Prompted_Selection::IsSelecting() )
    {
        Tools_Prompted_Selection::ChooseCustom( "Choose new camera node", ManuallySelectCameraNode_PickCallback, ManuallySelectCameraNode_RenderCallback );
    }
}


bool Tools_Visibility_System::ManuallySelectCameraNode_PickCallback( bool bClicked )
{
    g_uManuallySelectCameraNodeCurrentlyPickedNode = FindVisNodeUnderCursor();
    if( g_uManuallySelectCameraNodeCurrentlyPickedNode != uASURA_VISIBILITY_INVALID_NODE_INDEX )
    {
        if( bClicked )
        {
            // Modify the current camera node in the engine visibility system
            // and tell it that it will need to recalculate visibilities
            // based on it.
            Asura_Visibility_System::s_auLastKnownCameraNode[0] = g_uManuallySelectCameraNodeCurrentlyPickedNode;
            Asura_Visibility_System::s_bForceVisRecalcForCurrentCameraNode = true;
            
            // And force the 'freeze camera node' option on, since setting
            // this is fairly useless otherwise (ie. it will be immediately
            // overwritten next frame).
            Asura_Visibility_System::s_bFreezeCameraNode = true;

            g_uManuallySelectCameraNodeCurrentlyPickedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;
        }
        return true;
    }
    else
    {
        return false;
    }
}


void Tools_Visibility_System::ManuallySelectCameraNode_RenderCallback( u_int uSuggestedARGB )
{
    if( g_uManuallySelectCameraNodeCurrentlyPickedNode < Asura_Visibility_System::s_uNumberOfNodes )
    {
        const Asura_Visibility_Node& xVisNode = Asura_Visibility_System::s_pxVisibilityNodes[g_uManuallySelectCameraNodeCurrentlyPickedNode];
        Tools_Render::RenderCube( xVisNode.GetAABB(), uSuggestedARGB, false );
    }
}


u_int Tools_Visibility_System::FindVisNodeUnderCursor( void )
{
    u_int uNodeUnderCursor = uASURA_VISIBILITY_INVALID_NODE_INDEX;

    if( !Asura_Visibility_System::s_uNumberOfNodes ) { return uNodeUnderCursor; }

    Asura_Vector_3 xLOSOrigin, xLOSDir;
    Tools_LOS::GetCursorLOSOriginAndDir( &xLOSOrigin, &xLOSDir );

    u_int auVisNodeStack[32];
    u_int uStackIdx = 0;

    // push root onto stack
    auVisNodeStack[ uStackIdx++ ] = 0;

    float fClosestHit = Asura_Maths::LargeFloat;
    while( uStackIdx )
    {
        // pop something off the stack to work on
        u_int uNodeIndex = auVisNodeStack[ --uStackIdx ];
        const Asura_Visibility_Node& xVisNode = Asura_Visibility_System::s_pxVisibilityNodes[ uNodeIndex ];

        // Only test visible nodes, ie. those that will be getting
        // rendered by RenderVisibleNodeBBs().
        if( !Asura_Visibility_System::s_pxNodeFlags[uNodeIndex].GetORedVisibilityFlags() )
        {
            continue;
        }

        // Ignore nodes that aren't in the camera's LOS...
        float fDist;
        if( !Asura_Maths::RayIntersectsBoundingBox(xLOSOrigin, xLOSDir, xVisNode.GetAABB(), &fDist) )
        {
            continue;
        }

        // ... or that are further away than our closest hit so far.
        if( fDist > fClosestHit )
        {
            continue;
        }

        if( xVisNode.GetLeftChild() )
        {
            // push children onto stack
            auVisNodeStack[ uStackIdx++ ] = xVisNode.GetLeftChild();
            auVisNodeStack[ uStackIdx++ ] = xVisNode.GetRightChild();
        }
        else // We're a leaf node
        {
            // Ignore the camera node.
            if( uNodeIndex != Asura_Visibility_System::s_auLastKnownCameraNode[0] )
            {
                // Ok, closest yet? But ignore nodes that the camera's
                // inside.
                if( (fDist > 0.0f) && (fDist < fClosestHit) )
                {
                    uNodeUnderCursor = uNodeIndex;
                    fClosestHit = fDist;
                }
            }
        }
    }
    
    return uNodeUnderCursor;
}


void Tools_Visibility_System::DoIncrementalAutoRefinement( void )
{
    Tools_Environment_LOS::EnsureDataIsValid();

    RecalculateBaseEnvironmentData();

    const Asura_Vector_3& xCamPos = *Asura_Camera::GetPosition();
    const bool bCamOutsideEnv = !g_xEnvironmentAABB.Contains( xCamPos );

    if( g_pfnStartVisCalcsCallback )
    {
        g_pfnStartVisCalcsCallback();
    }

    if( g_bAutoRefinementFromCameraPosition && !bCamOutsideEnv )
    {
        for( u_int u = 0; u < g_uMaxIncrementalRefinementRaysPerFrame; ++u )
        {
            Asura_Vector_3 xDirection;
            xDirection.RandomDirection();
            Tools_Visibility_System::RefineVisibility( xCamPos, xDirection );
        }
    }

    if( g_bAutoRefinementFromCameraNode )
    {
        const u_int uCamNode = Asura_Visibility_System::s_auLastKnownCameraNode[0];
        if( uCamNode )
        {
            const Asura_Bounding_Box& xCamNodeBB = Asura_Visibility_System::s_pxVisibilityNodes[uCamNode].GetAABB();
            for( u_int u = 0; u < g_uMaxIncrementalRefinementRaysPerFrame; ++u )
            {
                Asura_Vector_3 xPos;
                if( !GetRandomPointInAABB(xPos, xCamNodeBB) )
                {
                    continue;
                }

                Asura_Vector_3 xDir;
                xDir.RandomDirection();

                Tools_Visibility_System::RefineVisibility( xPos, xDir );
            }
        }
    }

    if( g_bAutoRefinementUnderCursor && !bCamOutsideEnv )
    {
        Asura_Vector_3 xOrigin, xDir;
        Tools_LOS::GetCursorLOSOriginAndDir( &xOrigin, &xDir );
        Tools_Visibility_System::RefineVisibility( xOrigin, xDir );
    }

    if( g_pfnEndVisCalcsCallback )
    {
        g_pfnEndVisCalcsCallback();
    }
}


void Tools_Visibility_System::RecalculateBaseEnvironmentData( void )
{
    const Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
    if( !pxEnv ) { return; }

    const u_int uNumModules = Asura_Module_System::GetNumberOfModules();

    {
        Asura_Bounding_Box xBoundingBox;
        xBoundingBox.Reset();

        for( u_int u = 0; u < uNumModules; ++u )
        {
            // KJFL - 14 December 2007 - ignore scenery modules when figuring out the world bounding box
            if( pxEnv->IsSceneryModule(u) ) continue;

            Asura_Bounding_Box xBB = Asura_Module_System::GetBoundingBoxOfModule(u);
            xBB.Displace( Asura_Module_System::GetPositionOfModule(u) );
            xBoundingBox.Set( xBB );
        }

        g_xEnvironmentAABB = xBoundingBox;
    }

    g_uNumberOfVisibilityFeatures = uNumModules + 1;

}


void Tools_Visibility_System::ClearStoredVisibilitySegments( void )
{
    Tools_VisibilitySegment::DestroyAll();

    if( g_ppxVisibilitySegments )
    {
        delete[] g_ppxVisibilitySegments;
        g_ppxVisibilitySegments = NULL;
    }

    g_uCurrentNumberOfVisibilitySegments = 0;

    RenderVisSegments_ClearFilteredList();
}


void Tools_Visibility_System::SetStartVisCalcsCallback( START_VIS_CALCS_CALLBACK pfnStartVisCalcsCallback )
{
    g_pfnStartVisCalcsCallback = pfnStartVisCalcsCallback;
}


void Tools_Visibility_System::SetEndVisCalcsCallback( END_VIS_CALCS_CALLBACK pfnEndVisCalcsCallback )
{
    g_pfnEndVisCalcsCallback = pfnEndVisCalcsCallback;
}


void Tools_Visibility_System::SetRayHitFilterCallback( RAY_HIT_FILTER_CALLBACK pfnHitFilterCallback )
{
    g_pfnRayHitFilterCallback = pfnHitFilterCallback;
}


void Tools_Visibility_System::SetRayModuleTestFilterCallback( RAY_MODULE_TEST_FILTER_CALLBACK pfnModuleTestFilterCallback, const void* pvUserData )
{
    g_pfnRayModuleTestFilterCallback = pfnModuleTestFilterCallback;
    g_pvRayModuleTestFilterUserData = pvUserData;
}


void Tools_Visibility_System::FindSplitHintQuads( void )
{
    // Run through all vis blocker polys in the environment. Any that are 
    // axis-aligned will form initial split planes.

    // If we've previously done a recalculate in this tools session then
    // we need to clean up the old list we generated.
    g_xSplitHintQuadList.Clear();

    const Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
    if( !pxEnv ) { return; }

    for( int iModule = 0; iModule < Asura_Module_System::GetNumberOfModules(); ++iModule )
    {
        Asura_Mesh* pxModuleMesh = Asura_Module_System::GetMeshForModule( iModule );
        if( !pxModuleMesh || !pxModuleMesh->PolygonFlagDataValid )
        {
            continue;
        }

        Asura_Assert( pxModuleMesh->PolygonDataValid && pxModuleMesh->VertexDataValid, "" );

        const Asura_Vector_3& xModulePos = Asura_Module_System::GetPositionOfModule( iModule );

        for( int iPoly = 0; iPoly < pxModuleMesh->NumberOfPolygons; ++iPoly )
        {
            // Only interested in vis blocking polys.
            if( !(pxModuleMesh->PolygonFlagDataPtr[iPoly].Flags & AsuraPolygonFlag_DontExport) )
            {
                continue;
            }

            const Asura_Mesh_Polygon& xPolyData = pxModuleMesh->PolygonDataPtr[iPoly];
            Tools_Visibility_SplitHintQuad xSplitHintQuad;

            // Only interested in axis-aligned ones.
            {
                if( !pxModuleMesh->AreNormalsValid() )
                {
                    pxModuleMesh->CreateNormals();
                }

                const Asura_Vector_3& xFaceNormal = pxModuleMesh->NormalsDataPtr[iPoly].FaceNormal;
                const Asura_Vector_3 xFirstVertexPos = pxModuleMesh->VertexDataPtr[xPolyData.VertexID[0]].Position + xModulePos;

                bool bNonAxisAligned = true;

                for( u_int uAxis = 0; uAxis < 3; ++uAxis )
                {
                    if( Asura_Maths::Abs(xFaceNormal[uAxis]) > 0.9999f )
                    {
                        xSplitHintQuad.m_uAxis = uAxis;
                        xSplitHintQuad.m_fAxisValue = xFirstVertexPos[uAxis];
                        bNonAxisAligned = false;
                        break;
                    }
                }
                
                if( bNonAxisAligned ) { continue; }
            }

            // There may be multiple polys making up the split plane, eg.
            // a large blocking plane with a hole cut out for something that
            // passes through, so ignore duplicates.
            {
                bool bDuplicate = false;

                for( Asura_LinkList_ConstIt<Tools_Visibility_SplitHintQuad> xQuadIt(&g_xSplitHintQuadList); !xQuadIt.Done(); xQuadIt.Next() )
                {
                    const Tools_Visibility_SplitHintQuad* pxOtherQuad = xQuadIt.GetCurrent();

                    if( xSplitHintQuad.m_uAxis != pxOtherQuad->m_uAxis ) { continue; }

                    Asura_Bounding_Box xOtherQuadBB = pxOtherQuad->m_xAABB;

                    // Interested in checking against any co-planar quads
                    // that would result in a vis node inbetween that's too
                    // small.
                    xOtherQuadBB.m_fElement[pxOtherQuad->m_uAxis * 2 + 0] -= 0.5f * fTOOLS_VIEWCELL_MIN_EXTENT;
                    xOtherQuadBB.m_fElement[pxOtherQuad->m_uAxis * 2 + 1] += 0.5f * fTOOLS_VIEWCELL_MIN_EXTENT;

                    // Make an AABB for the current poly.
                    Asura_Bounding_Box xPolyBB;
                    xPolyBB.Reset();
                    for( int iVtx = 0; iVtx < xPolyData.NumberOfVertices(); ++iVtx )
                    {
                        xPolyBB.Set( pxModuleMesh->VertexDataPtr[xPolyData.VertexID[iVtx]].Position );
                    }
                    xPolyBB.Displace( xModulePos );

                    if( !xPolyBB.Intersects(xOtherQuadBB) ) { continue; }

                    bDuplicate = true;
                    break;
                }

                if( bDuplicate )
                {
                    continue;
                }
            }

            // Initialise BB.
            xSplitHintQuad.m_xAABB.Reset();
            xSplitHintQuad.m_xAABB.m_fElement[xSplitHintQuad.m_uAxis * 2 + 0] = xSplitHintQuad.m_fAxisValue;
            xSplitHintQuad.m_xAABB.m_fElement[xSplitHintQuad.m_uAxis * 2 + 1] = xSplitHintQuad.m_fAxisValue;

            // Fill in the bounding values for the other two axes.
            CalculateSplitHintQuadAABB( &xSplitHintQuad, iModule, iPoly );

            // And add it to the list.
            g_xSplitHintQuadList.PushBack( xSplitHintQuad );
        }
    }
}


void Tools_Visibility_System::CalculateSplitHintQuadAABB( Tools_Visibility_SplitHintQuad* pxSplitHintQuad, int iInitialModule, int iInitialPoly )
{
    List<PolygonID> xPolyList;

    // Find all adjacent polygons in the same plane. (In fact, the list will
    // end up including the initial poly as well.)
    {
        PolygonID xInitialPolyID( iInitialModule, iInitialPoly );
        Tools_Polygon_Selection::ListPolygonsInPlane( xInitialPolyID, xPolyList, false, 0.5f );
    }
    
    // Generate a bounding box based on all of these polygons. Note that the
    // list will include all adjacent polygons, including ones not using the
    // 'don't export' flag, and we need to filter them.
    for( CLIF<PolygonID> xPolyIt(&xPolyList); !xPolyIt.done(); xPolyIt.next() )
    {
        const PolygonID& xPolyID = xPolyIt();

        const Asura_Mesh* pxModuleMesh = Asura_Module_System::GetMeshForModule( xPolyID.ModuleIndex );
        if( !pxModuleMesh || !pxModuleMesh->PolygonFlagDataValid )
        {
            continue;
        }

        Asura_Assert( pxModuleMesh->PolygonDataValid && pxModuleMesh->VertexDataValid, "" );

        if( !(pxModuleMesh->PolygonFlagDataPtr[xPolyID.PolyIndex].Flags & AsuraPolygonFlag_DontExport) )
        {
            continue;
        }

        const Asura_Mesh_Polygon& xPolyData = pxModuleMesh->PolygonDataPtr[xPolyID.PolyIndex];
        for( int iVtx = 0; iVtx < xPolyData.NumberOfVertices(); ++iVtx )
        {
            const Asura_Vector_3 xVertexPos = pxModuleMesh->VertexDataPtr[xPolyData.VertexID[iVtx]].Position + Asura_Module_System::GetPositionOfModule(xPolyID.ModuleIndex);

            for( u_int uAxis = 0; uAxis < 3; ++uAxis )
            {
                if( uAxis == pxSplitHintQuad->m_uAxis )
                {
                    continue;
                }

                pxSplitHintQuad->m_xAABB.m_fElement[uAxis * 2 + 0] = Asura_Maths::Min( pxSplitHintQuad->m_xAABB.m_fElement[uAxis * 2 + 0], xVertexPos[uAxis] );
                pxSplitHintQuad->m_xAABB.m_fElement[uAxis * 2 + 1] = Asura_Maths::Max( pxSplitHintQuad->m_xAABB.m_fElement[uAxis * 2 + 1], xVertexPos[uAxis] );
            }
        }
    }
}


void Tools_Visibility_System::RenderSplitHintQuads( void )
{
    // Render quads within the plane, bounded by the environment BB.
    for( Asura_LinkList_ConstIt<Tools_Visibility_SplitHintQuad> xIt(&g_xSplitHintQuadList); !xIt.Done(); xIt.Next() )
    {
        const Tools_Visibility_SplitHintQuad* pxSplitHintQuad = xIt.GetCurrent();

        Asura_Vector_3 xPoints[4];

        u_int uAxis0 = pxSplitHintQuad->m_uAxis;
        u_int uAxis1 = (uAxis0 + 1) % 3;
        u_int uAxis2 = (uAxis0 + 2) % 3;

        for( u_int uPoint = 0; uPoint < 4; ++uPoint )
        {
            xPoints[uPoint][uAxis0] = pxSplitHintQuad->m_fAxisValue;
            xPoints[uPoint][uAxis1] = ( (uPoint == 2 || uPoint == 3) ? pxSplitHintQuad->m_xAABB.m_fElement[uAxis1 * 2 + 1] : pxSplitHintQuad->m_xAABB.m_fElement[uAxis1 * 2] );
            xPoints[uPoint][uAxis2] = ( (uPoint == 1 || uPoint == 2) ? pxSplitHintQuad->m_xAABB.m_fElement[uAxis2 * 2 + 1] : pxSplitHintQuad->m_xAABB.m_fElement[uAxis2 * 2] );
        }

        extern void RenderOutline( const Asura_Vector_3* pxPoints, int iNumPoints, u_int uColour, bool bWireframe, bool bBackFaceCull );
        RenderOutline( xPoints, 4, 0x7FFFFF22, false, false );
    }
}


bool Tools_Visibility_System::ShouldHighlightNonVisibleGeometry( void )
{
    return( IsOverlayOpen() && g_bHighlightNonVisibleGeometry );
}

#endif //#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM
