//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Analysis.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Waypointing/Tools_Waypoint_Analysis.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Dynamics/Asura_Collision_Object.h"
#include "Interface/Asura_Dialog.h"
#include "Physics/Asura_Physics.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"

// tools
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Waypointing/Tools_Mode_Waypointing.h"
#include "Waypointing/Tools_Waypoint_Selection.h"

//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Analysis
//
//////////////////////////////////////////////////////////////////////////////

Tools_Waypoint_Analysis* Tools_Waypoint_Analysis::s_pxThis = NULL;

Tools_Waypoint_Analysis::Tools_Waypoint_Analysis()
: m_bLinkAnalysisSpeedChanged( false )
, m_bLinkAnalysisInProgress( false )
, m_bLinkAnalysisSuccessful( false )
, m_fLinkAnalysisPlaybackSpeed( 1.0f )
, m_fLinkAnalysisPlaybackTime( 0.0f )
, m_pxPositionOutputArray( NULL )
, m_uNumPositionOutputs( 0 )
, m_iStartWaypoint( ASURA_INVALID_WAYPOINT_ID )
, m_iFinishWaypoint( ASURA_INVALID_WAYPOINT_ID )
, m_bRouteFindInProgress( false )
, m_iRouteFindStepIncrement( 5 )
, m_bRouteFindFinished( false )
, m_pusRoute( NULL )
, m_iRouteLength( 0 )
{
    s_pxThis = this;
}

Tools_Waypoint_Analysis::~Tools_Waypoint_Analysis()
{
    s_pxThis = NULL;
}

void Tools_Waypoint_Analysis::Update()
{
    if ( m_bLinkAnalysisSpeedChanged )
    {
        m_bLinkAnalysisSpeedChanged = false;
        if ( Asura_Maths::Abs( m_fLinkAnalysisPlaybackSpeed - 0.0f ) < 0.01f )
        {
            m_fLinkAnalysisPlaybackSpeed = 0.0f;
        }
    }
}

void Tools_Waypoint_Analysis::Shutdown()
{
    if ( m_bLinkAnalysisInProgress )
    {
        StopLinkAnalysis();
    }
    if ( m_bRouteFindInProgress )
    {
        StopRouteFindAnalysis();
    }
}

void Tools_Waypoint_Analysis::AddOverlayItemsToTabsDialogue( Asura_Tabs_Dialog* pxTabsDialogue, Asura_Vector_2& xOffset )
{
    Asura_Dialog* pxNewDialogue = new Asura_Function_Dialog( "Analyse Link Between Selected Waypoints", xOffset, AnalyseLinkStatic, NULL, AllowedToAnalyseLinkStatic );
    pxTabsDialogue->AddDialog( pxNewDialogue );

    xOffset.x += 32.0f;
    pxNewDialogue = new Asura_Function_Dialog( "Stop Link Analysis", xOffset, StopLinkAnalysisStatic, NULL, AllowedToStopLinkAnalysisStatic );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.x -= 32.0f;
    xOffset.y += Asura_Dialog::DefaultSpacing();

    xOffset.x += 5.0f;
    pxNewDialogue = new Asura_Numerical_Dialog( "Link Analysis Playback Speed", xOffset, &m_fLinkAnalysisPlaybackSpeed, -1.0f, 4.0f, 0.01f, "%0.2f", &m_bLinkAnalysisSpeedChanged );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.x -= 5.0f;
    xOffset.y += Asura_Dialog::LargeSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Analyse Route Find Between Selected Waypoints", xOffset, AnalyseRouteFindStatic, NULL, AllowedToAnalyseRouteFindStatic );
    pxTabsDialogue->AddDialog( pxNewDialogue );

    xOffset.x += 32.0f;
    pxNewDialogue = new Asura_Function_Dialog( "Stop Route Find Analysis", xOffset, StopRouteFindAnalysisStatic, NULL, AllowedToStopRouteFindAnalysisStatic );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.x -= 32.0f;
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Next Route Find Step", xOffset, NextRouteFindStepStatic, NULL, AllowedToIncrementRouteStepStatic );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Next n Route Find Steps", xOffset, NextRouteFindStepsStatic, NULL, AllowedToIncrementRouteStepStatic );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

    xOffset.x += 5.0f;
    pxNewDialogue = new Asura_Numerical_Dialog( "Number Of Route Find Steps", xOffset, &m_iRouteFindStepIncrement, 1, 2000, 1, "%d" );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.x -= 5.0f;
    xOffset.y += Asura_Dialog::DefaultSpacing();
}

void Tools_Waypoint_Analysis::RenderWorld()
{
    if ( m_bLinkAnalysisInProgress )
    {
        unsigned int uColour = 0xffffffff;
        m_fLinkAnalysisPlaybackTime += Asura_Timers::GetGameTimeSlice() * m_fLinkAnalysisPlaybackSpeed;
        if ( m_fLinkAnalysisPlaybackTime < 0.0f )
        {
            m_fLinkAnalysisPlaybackTime = 0.0f;
        }
        u_int uIndex = static_cast< u_int >( Asura_Maths::FloatToInt( m_fLinkAnalysisPlaybackTime / Asura_Physics_System::GetTimeStep() ) );
        if ( uIndex >= m_uNumPositionOutputs )
        {
            static const float s_fNumSecondsToHoldAtEnd = 2.0f;
            if ( m_fLinkAnalysisPlaybackTime > ( m_uNumPositionOutputs * Asura_Physics_System::GetTimeStep() ) + s_fNumSecondsToHoldAtEnd )
            {
                m_fLinkAnalysisPlaybackTime = 0.0f;
            }
            uIndex  = m_uNumPositionOutputs - 1;
            uColour = m_bLinkAnalysisSuccessful ? 0xff00ff00 : 0xffff0000;
        }

        const Asura_Vector_3 xPosition = m_pxPositionOutputArray[ uIndex ];
        Asura_Bounding_Box xBox( Tools_Mode_Waypointing::s_xTestCollisionBox.GetBoundingBox() );
        xBox.Displace( xPosition );
        Tools_Render::RenderCube( xBox, uColour, true );
    }
    else if ( m_bRouteFindInProgress )
    {
        for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
        {
			const bool bOpen = Asura_AI_Navigation_System::s_pxAStarOpenClosedVector[ i ].m_bOpen;
			const bool bClosed = Asura_AI_Navigation_System::s_pxAStarOpenClosedVector[ i ].m_bClosed;
			if ( !bOpen && !bClosed )
			{
				continue;
			}

			Asura_Vector_3 xScreenSpacePos = Asura_AI_Navigation_System::GetWaypoint( i )->GetPosition();
			if ( !Asura_Camera::ProjectWorldCoordsToScreenSpace( xScreenSpacePos, true ) || xScreenSpacePos.z <= 0.0f )
			{
				continue;
			}

			if ( bClosed )
            {
                Tools_Waypoint_Selection::RenderSelectedWaypoint( i );
            }
            else if ( bOpen )
            {
                Tools_Waypoint_Selection::RenderHighlitWaypoint( i );
            }

			Asura_Font_System::Print(
				xScreenSpacePos.x, 
				xScreenSpacePos.y, 
				1.0f, 
				0xffffffff,
				ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, 
				"%.4f\n%.4f",
				Asura_AI_Navigation_System::s_pxAStarList[ i ].m_fCostToHere,
				Asura_AI_Navigation_System::s_pxAStarList[ i ].m_fCostToGoal );
        }

        if ( m_bRouteFindFinished )
        {
            for ( int i = 0; i < m_iRouteLength - 1; ++i )
            {
                Asura_AI_NavVertex* pxStartWP  = Asura_AI_Navigation_System::GetWaypoint( m_pusRoute[ i ] );
                Asura_AI_NavVertex* pxFinishWP = Asura_AI_Navigation_System::GetWaypoint( m_pusRoute[ i + 1 ] );

                Asura_Render::RenderLine( pxStartWP->GetPosition() + pxStartWP->GetNormal() * 0.1f,
                    pxFinishWP->GetPosition() + pxFinishWP->GetNormal() * 0.1f,
                    0xffffffff );
            }
        }
    }
}

void Tools_Waypoint_Analysis::RenderText( float& fPrintX, float& fPrintY )
{
	if ( m_bRouteFindInProgress )
	{
		int iNumOpen   = 0;
		int iNumClosed = 0;
		for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
		{
			if ( Asura_AI_Navigation_System::s_pxAStarOpenClosedVector[ i ].m_bClosed )
			{
				++iNumClosed;
			}
			else if ( Asura_AI_Navigation_System::s_pxAStarOpenClosedVector[ i ].m_bOpen )
			{
				++iNumOpen;
			}
		}

		fPrintY += 1.0f;
		Asura_Text_Output::Print( fPrintX, fPrintY, 0xffffffff, 0, "Nodes on open list: %d", iNumOpen );
		fPrintY += 1.0f;
		Asura_Text_Output::Print( fPrintX, fPrintY, 0xffffffff, 0, "Nodes on closed list: %d", iNumClosed );
		fPrintY += 1.0f;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Tools_Waypoint_Analysis::AnalyseLink()
{
    int iStartWaypoint, iFinishWaypoint;
    if ( !Tools_Waypoint_Selection::GetFirstTwoSelectedWaypoints( iStartWaypoint, iFinishWaypoint ) ) { return; }

    Asura_AI_NavVertex* pxStartWaypoint  = Asura_AI_Navigation_System::GetWaypoint( iStartWaypoint );
    Asura_AI_NavVertex* pxFinishWaypoint = Asura_AI_Navigation_System::GetWaypoint( iFinishWaypoint );
    if ( !pxStartWaypoint && !pxFinishWaypoint ) { return; }

    const float fDistance = ( pxStartWaypoint->m_xPosition - pxFinishWaypoint->m_xPosition ).Magnitude();
    if ( fDistance > Tools_Mode_Waypointing::GetMaxLinkLength() )
    {
        Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Distance between waypoints too large ( > max link length )" );
        return;
    }

    m_bLinkAnalysisInProgress   = true;
    m_iStartWaypoint            = iStartWaypoint;
    m_iFinishWaypoint           = iFinishWaypoint;
    m_fLinkAnalysisPlaybackTime = 0.0f;

    Tools_Mode_Waypointing::GetInstance()->InitialisePhysicsDataForWaypointing();

    u_int uDummyLinkFlags = 0;
    Asura_Vector_3 xStartWaypointNormal = pxStartWaypoint->GetNormal();
    Asura_Vector_3 xEndWaypointNormal = pxFinishWaypoint->GetNormal();
    m_bLinkAnalysisSuccessful = Tools_Mode_Waypointing::GetInstance()->CullAndFlagLinkWithPhysics( uDummyLinkFlags,
        pxStartWaypoint->m_xPosition,  xStartWaypointNormal, iStartWaypoint,
        pxFinishWaypoint->m_xPosition, xEndWaypointNormal,   iFinishWaypoint );

    // Copy the positions over to our local storage.
    m_uNumPositionOutputs   = Asura_Maths::Min( Tools_Mode_Waypointing::s_uNumPhysicsPositions, static_cast< u_int >( MAX_POSITION_OUTPUTS ) );
    m_pxPositionOutputArray = new Asura_Vector_3[ m_uNumPositionOutputs ];
    if ( m_pxPositionOutputArray )
    {
        Asura_Mem::Copy( m_pxPositionOutputArray, Tools_Mode_Waypointing::s_pxPhysicsPositions, m_uNumPositionOutputs * sizeof( Asura_Vector_3 ) );
    }
    else
    {
        m_uNumPositionOutputs = 0;
    }

    Tools_Mode_Waypointing::GetInstance()->ResetPhysicsDataAfterWaypointing();
}

void Tools_Waypoint_Analysis::StopLinkAnalysis()
{
    m_bLinkAnalysisInProgress = false;
    m_iStartWaypoint          = -1;
    m_iFinishWaypoint         = -1;
    delete [] m_pxPositionOutputArray;
    m_pxPositionOutputArray = NULL;
    m_uNumPositionOutputs   = 0;
}

void Tools_Waypoint_Analysis::AnalyseRouteFind()
{
    int iStartWaypoint, iFinishWaypoint;
    if ( !Tools_Waypoint_Selection::GetFirstTwoSelectedWaypoints( iStartWaypoint, iFinishWaypoint ) ) { return; }

    Asura_AI_NavVertex* pxStartWaypoint  = Asura_AI_Navigation_System::GetWaypoint( iStartWaypoint );
    Asura_AI_NavVertex* pxFinishWaypoint = Asura_AI_Navigation_System::GetWaypoint( iFinishWaypoint );
    if ( !pxStartWaypoint && !pxFinishWaypoint ) { return; }

    m_pusRoute = new unsigned short[ MAX_ANALYSIS_ROUTE_LENGTH ];
    if ( !m_pusRoute ) { return; }
    m_bRouteFindInProgress = true;
    m_bRouteFindFinished   = false;
    m_iStartWaypoint       = iStartWaypoint;
    m_iFinishWaypoint      = iFinishWaypoint;
    m_iRouteLength         = 0;

    Asura_AI_Navigation_System::PreAStarSearch( static_cast< unsigned short >( m_iStartWaypoint ), static_cast< unsigned short >( m_iFinishWaypoint ),
        Asura_AI_Navigation_System::DefaultGoalCostEstimateFunction, NULL );
}

void Tools_Waypoint_Analysis::StopRouteFindAnalysis()
{
    m_bRouteFindInProgress = false;
    m_bRouteFindFinished   = false;
    m_iStartWaypoint       = ASURA_INVALID_WAYPOINT_ID;
    m_iFinishWaypoint      = ASURA_INVALID_WAYPOINT_ID;
    m_iRouteLength         = 0;
    delete [] m_pusRoute;
    m_pusRoute = NULL;
}

void Tools_Waypoint_Analysis::NextRouteFindStep()
{
    m_bRouteFindFinished = Asura_AI_Navigation_System::AStarSearchIncremental( m_pusRoute, m_iRouteLength, MAX_ANALYSIS_ROUTE_LENGTH,
                                                                               0, NULL,
                                                                               Asura_AI_Navigation_System::DefaultGoalCostEstimateFunction,
                                                                               Asura_AI_Navigation_System::DefaultTraversalCostFunction,
                                                                               Asura_AI_Navigation_System::DefaultGoalAchievedFunction,
                                                                               Asura_AI_Navigation_System::DefaultLinkFlagsTestFunction );
    if ( m_bRouteFindFinished )
    {
        if ( m_iRouteLength == 0 )
        {
            Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Route find failed!" );
        }
        else
        {
            Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, "Route find succeeded!" );
        }
    }
}

void Tools_Waypoint_Analysis::NextRouteFindSteps()
{
    int iStepsLeft = m_iRouteFindStepIncrement;
    while ( iStepsLeft-- && !m_bRouteFindFinished )
    {
        m_bRouteFindFinished = Asura_AI_Navigation_System::AStarSearchIncremental( m_pusRoute, m_iRouteLength, MAX_ANALYSIS_ROUTE_LENGTH,
                                                                                   0, NULL,
                                                                                   Asura_AI_Navigation_System::DefaultGoalCostEstimateFunction,
                                                                                   Asura_AI_Navigation_System::DefaultTraversalCostFunction,
                                                                                   Asura_AI_Navigation_System::DefaultGoalAchievedFunction,
                                                                                   Asura_AI_Navigation_System::DefaultLinkFlagsTestFunction );
    }

    if ( m_bRouteFindFinished )
    {
        if ( m_iRouteLength == 0 )
        {
            Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Route find failed!" );
        }
        else
        {
            Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, "Route find succeeded!" );
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

bool Tools_Waypoint_Analysis::AllowedToAnalyseLink()
{
    return ( Tools_Waypoint_Selection::IsEnabled() && Tools_Waypoint_Selection::GetNumSelectedWaypoints() == 2 && !s_pxThis->m_bLinkAnalysisInProgress );
}

bool Tools_Waypoint_Analysis::AllowedToStopLinkAnalysis()
{
    return s_pxThis->m_bLinkAnalysisInProgress;
}

bool Tools_Waypoint_Analysis::AllowedToAnalyseRouteFind()
{
    return ( Tools_Waypoint_Selection::IsEnabled() && Tools_Waypoint_Selection::GetNumSelectedWaypoints() == 2 && !s_pxThis->m_bRouteFindInProgress );
}

bool Tools_Waypoint_Analysis::AllowedToStopRouteFindAnalysis()
{
    return s_pxThis->m_bRouteFindInProgress;
}

bool Tools_Waypoint_Analysis::AllowedToIncrementRouteStep()
{
    return ( s_pxThis->m_bRouteFindInProgress && !s_pxThis->m_bRouteFindFinished );
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
