//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Selection.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Created by Andrew Weinkove - 25/09/2003
//
// Waypoint selection is based on the polygon selection and supports analysis of generated waypoints
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// includes
//
//////////////////////////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Waypointing/Tools_Waypoint_Selection.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Movement/Tools_Movement.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"

// tools
#include "Tools_Overlays.h"
#include "Flags/Tools_Mode_Flags.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_LOS.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Waypointing/Tools_Mode_Waypointing.h"
#include "Tools_System.h"

//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Selection
// static member variables
//
//////////////////////////////////////////////////////////////////////////////

bool           Tools_Waypoint_Selection::s_bActive				   = false;
bool           Tools_Waypoint_Selection::s_bEnabled                = false;
int            Tools_Waypoint_Selection::s_iSelectionArrayLength   = 0;
Tools_Waypoint_Selection::SelectionData* Tools_Waypoint_Selection::s_pxSelectionArray = NULL;
int            Tools_Waypoint_Selection::s_iNumSelected            = 0;
int            Tools_Waypoint_Selection::s_iHighlitWaypoint        = -1;
bool           Tools_Waypoint_Selection::s_bButtonDownForSelection = false;

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Tools_Waypoint_Selection::Activate()
{
	s_bActive = true;

    Tools_Waypoint_Selection_AStar::Initialise();
    s_bButtonDownForSelection = false;

    OnWaypointDataChanged();
}

void Tools_Waypoint_Selection::Deactivate()
{
    delete [] s_pxSelectionArray;
    s_pxSelectionArray      = NULL;
    s_iSelectionArrayLength = 0;

    Tools_Waypoint_Selection_AStar::Shutdown();

	s_bActive = false;
}

void Tools_Waypoint_Selection::OnWaypointDataChanged()
{
	if ( !s_bActive ) { return; }

    // (Re)Allocate the selection array.
    if ( s_pxSelectionArray != NULL )
    {
        delete [] s_pxSelectionArray;
        s_pxSelectionArray      = NULL;
        s_iSelectionArrayLength = 0;
    }
    if ( Asura_AI_Navigation_System::GetNumberOfWaypoints() )
    {
        s_iSelectionArrayLength = Asura_AI_Navigation_System::GetNumberOfWaypoints();
        s_pxSelectionArray      = new SelectionData[ s_iSelectionArrayLength ];
    }
    s_iNumSelected = 0;
}

void Tools_Waypoint_Selection::TransferToolsModeData( const int iMode, Tools_ModeDataFile& xFile )
{
    int iVersion = CURRENT_MODE_DATA_VERSION;
    xFile.Transfer( iVersion );

    if ( iVersion == CURRENT_MODE_DATA_VERSION )
    {
        xFile.Transfer( s_bEnabled );
    }
    else
    {
        Asura_Assertf( false, ( "Tools_Waypoint_Selection::TransferToolsModeData - unhandled version (%d)", iVersion ) );
    }
}

void Tools_Waypoint_Selection::AddOverlayItemsToTabsDialogue( Asura_Tabs_Dialog* pxTabsDialogue, Asura_Vector_2& xOffset )
{
    Asura_Dialog* pxNewDialogue = new Asura_Tickbox_Dialog( "Enable Waypoint Selection", xOffset, &s_bEnabled );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::LargeSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Clear Selection", xOffset, Tools_Waypoint_Selection::ClearSelection, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Select All", xOffset, Tools_Waypoint_Selection::SelectAll, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Invert Selection", xOffset, Tools_Waypoint_Selection::InvertSelection, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Select By Index", xOffset, Tools_Waypoint_Selection::SelectByIndex, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Select By Coverpoint Index", xOffset, Tools_Waypoint_Selection::SelectByCoverpointIndex, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Select By Zone ID", xOffset, Tools_Waypoint_Selection::SelectByZoneID, NULL, IsEnabled );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Select All Linked Waypoints", xOffset, Tools_Waypoint_Selection::SelectAllLinked, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Select All Unlinked Waypoints", xOffset, Tools_Waypoint_Selection::SelectAllUnlinked, NULL, IsEnabled );
    pxTabsDialogue->AddDialog(pxNewDialogue);
    xOffset.y += Asura_Dialog::DefaultSpacing();
}

void Tools_Waypoint_Selection::Update()
{
	if ( !s_bActive ) { return; }
    if ( !s_bEnabled ) { return; }

    //if the mouse button has been release , reset the s_bButtonHeldForSelection flag
    if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
    {
        s_bButtonDownForSelection = false;
    }

    //and handle selection using the mouse
    HandleMouseSelection();
}

void Tools_Waypoint_Selection::Render()
{
	if ( !s_bActive ) { return; }
    if ( !s_bEnabled ) { return; }

    if (s_iHighlitWaypoint != -1 )
    {
        RenderHighlitWaypoint(s_iHighlitWaypoint);
        RenderLinks(s_iHighlitWaypoint);
    }

    for (int i=0;i<Asura_AI_Navigation_System::GetNumberOfWaypoints();++i)
    {
        if ( s_pxSelectionArray[ i ].m_bIsSelected )
        {
            RenderSelectedWaypoint(i);
            RenderLinks(i);
        }
    }
}

void Tools_Waypoint_Selection::SelectAll()
{
    for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
    {
        if ( !s_pxSelectionArray[ i ].m_bIsSelected )
        {
            s_pxSelectionArray[ i ].m_bIsSelected     = true;
            s_pxSelectionArray[ i ].m_iSelectionIndex = s_iNumSelected++;
        }
    }
}

void Tools_Waypoint_Selection::ClearSelection()
{
	if ( !s_bActive ) { return; }

	for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
    {
        s_pxSelectionArray[ i ].m_bIsSelected     = false;
        s_pxSelectionArray[ i ].m_iSelectionIndex = -1;
    }
    s_iNumSelected = 0;
}

void Tools_Waypoint_Selection::SelectByIndex()
{
    GetIntNumber_Overlay xOverlay("Enter Index",0,0,USHRT_MAX);
    if (xOverlay.DoModal())
    {
        int iIndex = xOverlay.GetNumber();
        if ( 0 <= iIndex && iIndex < Asura_AI_Navigation_System::GetNumberOfWaypoints() )
        {
            if ( !s_pxSelectionArray[ iIndex ].m_bIsSelected )
            {
                s_pxSelectionArray[ iIndex ].m_bIsSelected     = true;
                s_pxSelectionArray[ iIndex ].m_iSelectionIndex = s_iNumSelected++;
            }
        }
        else
        {
            Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR, "Specified waypoint index does not exist.");
        }
    }
}

void Tools_Waypoint_Selection::SelectByCoverpointIndex()
{
    GetIntNumber_Overlay xOverlay("Enter Coverpoint Index",0,0,USHRT_MAX);
    if (xOverlay.DoModal())
    {
        int iIndex = xOverlay.GetNumber();
        if (iIndex < Asura_AI_Navigation_System::GetNumberOfCoverpoints() &&
            iIndex >= 0)
        {
            bool bFoundOne = false;
            for (int i=0;i<Asura_AI_Navigation_System::GetNumberOfWaypoints();i++)
            {
                for (int c=0;c<Asura_AI_Navigation_System::GetNumCoverpointsForWaypoint(i);c++)
                {
                    if (Asura_AI_Navigation_System::GetCoverpointData(i)->m_pusCoverpoints[c] == iIndex)
                    {
                        bFoundOne = true;
                        if ( !s_pxSelectionArray[ i ].m_bIsSelected )
                        {
                            s_pxSelectionArray[ i ].m_bIsSelected     = true;
                            s_pxSelectionArray[ i ].m_iSelectionIndex = s_iNumSelected++;
                        }
                        break;
                    }
                }
            }
            Asura_Assert(bFoundOne, "Couldn't find any waypoints with this coverpoint");
        }
        else
        {
            Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR,"Specified waypoint index does not exist.");
        }
    }
}

void Tools_Waypoint_Selection::SelectByZoneID()
{
    GetIntNumber_Overlay xOverlay( "Enter Zone ID", 0, 0, Tools_Mode_Waypointing::GetNumZones() );
    if ( xOverlay.DoModal() )
    {
        const u_short usZoneID = static_cast< u_short >( xOverlay.GetNumber() );

        for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
        {
            if ( Asura_AI_Navigation_System::GetWaypoint( i )->m_usZone == usZoneID )
            {
                if ( !s_pxSelectionArray[ i ].m_bIsSelected )
                {
                    s_pxSelectionArray[ i ].m_bIsSelected     = true;
                    s_pxSelectionArray[ i ].m_iSelectionIndex = s_iNumSelected++;
                }
            }
        }
    }
}

void Tools_Waypoint_Selection::SelectAllLinked()
{
    Tools_Waypoint_Selection_AStar::DoAStar();
    for (int i=0;i<Asura_AI_Navigation_System::GetNumberOfWaypoints();i++)
    {
        if (Tools_Waypoint_Selection_AStar::s_pbFoundThisNode[i])
        {
            if ( !s_pxSelectionArray[ i ].m_bIsSelected )
            {
                s_pxSelectionArray[ i ].m_bIsSelected     = true;
                s_pxSelectionArray[ i ].m_iSelectionIndex = s_iNumSelected++;
            }
        }
    }
}

void Tools_Waypoint_Selection::SelectAllUnlinked()
{
    SelectAllLinked();
    InvertSelection();
}

void Tools_Waypoint_Selection::InvertSelection()
{
    s_iNumSelected = 0;
    for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
    {
        if ( s_pxSelectionArray[ i ].m_bIsSelected )
        {
            s_pxSelectionArray[ i ].m_bIsSelected     = false;
            s_pxSelectionArray[ i ].m_iSelectionIndex = -1;
        }
        else
        {
            s_pxSelectionArray[ i ].m_bIsSelected     = true;
            s_pxSelectionArray[ i ].m_iSelectionIndex = s_iNumSelected++;
        }
    }
}

void Tools_Waypoint_Selection::HandleMouseSelection()
{
    s_iHighlitWaypoint = -1;
    if (Asura_Interface::UsingMenusOrOverlays()) return;
    
    //see if the mouse is over a waypoint
    bool bOverWaypoint = GetWaypointUnderMouse();
    
    //if over a waypoint, check for selection/deselection
    if (Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
    {
        if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) || s_bButtonDownForSelection)
        {
            s_bButtonDownForSelection = true;
            if (bOverWaypoint)
            {
                if (Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
                {
                    if ( s_pxSelectionArray[ s_iHighlitWaypoint ].m_bIsSelected )
                    {
                        const int iUnselectedSelectionIndex = s_pxSelectionArray[ s_iHighlitWaypoint ].m_iSelectionIndex;
                        // Clear the selection data.
                        s_pxSelectionArray[ s_iHighlitWaypoint ].m_bIsSelected     = false;
                        s_pxSelectionArray[ s_iHighlitWaypoint ].m_iSelectionIndex = -1;
                        --s_iNumSelected;
                        // Decrement the selection index of any waypoints with a higher selection index.
                        for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
                        {
                            if ( s_pxSelectionArray[ i ].m_iSelectionIndex > iUnselectedSelectionIndex )
                            {
                                --s_pxSelectionArray[ i ].m_iSelectionIndex;
                            }
                        }
                    }
                }
                else
                {
                    if ( !s_pxSelectionArray[ s_iHighlitWaypoint ].m_bIsSelected )
                    {
                        s_pxSelectionArray[ s_iHighlitWaypoint ].m_bIsSelected     = true;
                        s_pxSelectionArray[ s_iHighlitWaypoint ].m_iSelectionIndex = s_iNumSelected++;
                    }
                }
            }
        }
    }
}

bool Tools_Waypoint_Selection::GetWaypointUnderMouse()
{
    s_iHighlitWaypoint = -1;

    int            iPolyIndex;
    int            iModuleIndex;
    Asura_Vector_3 xPosition;
    if ( Tools_LOS::FindPolyAndPositionUnderCursor( iPolyIndex, iModuleIndex, xPosition ) )
    {
		const short sTemp = Asura_AI_Navigation_System::FindWaypointNearestTo( xPosition, ASURA_NAVVERTEX_MAX_ZONE_ID, false, true, false );
        s_iHighlitWaypoint = sTemp;
    }
    return ( s_iHighlitWaypoint != -1 );
}

extern void Render_Sphere( const Asura_Vector_3& xCentre, float fRadius, int iColour );
void Tools_Waypoint_Selection::RenderHighlitWaypoint(int iIndex)
{
    float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
    fAlpha = 255.0f * fAlpha * fAlpha;
    unsigned int uColour = 0x00ffffff + (Asura_Maths::FloatToInt(fAlpha)<<24);
    if ( s_pxSelectionArray[ iIndex ].m_bIsSelected )
    {
        uColour &= 0xffff0000;
    }
    Render_Sphere(Asura_AI_Navigation_System::GetWaypoint(iIndex)->m_xPosition,Asura_AI_Navigation_System::GetWaypoint(iIndex)->GetRadius(), uColour);
}

void Tools_Waypoint_Selection::RenderSelectedWaypoint(int iIndex)
{
    Render_Sphere(Asura_AI_Navigation_System::GetWaypoint(iIndex)->m_xPosition,Asura_AI_Navigation_System::GetWaypoint(iIndex)->GetRadius(), 0x6fff0000);
}

void Tools_Waypoint_Selection::RenderLinks(int iWPIndex)
{
    for (u_char ucFromLink=0;ucFromLink<Asura_AI_Navigation_System::GetWaypoint(iWPIndex)->m_ucNumConnections;++ucFromLink)
	{
		int iTarget = Asura_AI_Navigation_System::GetWaypoint(iWPIndex)->m_pxConnections[ucFromLink].m_usIndex;
	
		if ( s_pxSelectionArray[ iTarget ].m_bIsSelected )
		{
			Asura_AI_NavVertex *pxTarget = Asura_AI_Navigation_System::GetWaypoint(iTarget);
	        
			bool bTwoWayLink = false;

			for (u_char ucToLink=0;ucToLink<pxTarget->m_ucNumConnections;ucToLink++)
			{
				if (pxTarget->m_pxConnections[ucToLink].m_usIndex == iWPIndex)
				{
					bTwoWayLink = true;
					break;
				}
			}

			Asura_Vector_3 xStartColour, xEndColour;
            Tools_Mode_Waypointing::GetInstance()->GetLinkColour( Asura_AI_Navigation_System::GetWaypoint(iWPIndex)->m_pxConnections[ucFromLink].m_usFlags,
                                                                  bTwoWayLink,
                                                                  xStartColour,
                                                                  xEndColour );

            Tools_Mode_Waypointing::RenderLine( Asura_AI_Navigation_System::GetWaypoint(iWPIndex)->m_xPosition,
						                        pxTarget->m_xPosition,
						                        xStartColour,
						                        xEndColour );
		}
	}
}

bool Tools_Waypoint_Selection::GetSelectedWaypoint( int& iWaypointOne )
{
	if ( !s_bActive ) { return false; }

	for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
    {
        if ( s_pxSelectionArray[ i ].m_bIsSelected )
        {
            iWaypointOne = i;
            return true;
        }
    }

    return false;
}

bool Tools_Waypoint_Selection::GetFirstTwoSelectedWaypoints( int& iWaypointOne, int& iWaypointTwo )
{
	if ( !s_bActive ) { return false; }

    // Find the waypoint index and selection index of the two waypoints.
    int aiWPIndex[ 2 ];
    int aiWPSelectionIndex[ 2 ];
    aiWPIndex[ 0 ] = aiWPIndex[ 1 ] = 0;
    aiWPSelectionIndex[ 0 ] = aiWPSelectionIndex[ 1 ] = -1;
    for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
    {
        if ( s_pxSelectionArray[ i ].m_bIsSelected )
        {
            aiWPIndex[ 0 ] = aiWPIndex[ 1 ];
            aiWPIndex[ 1 ] = i;
            aiWPSelectionIndex[ 0 ] = aiWPSelectionIndex [ 1 ];
            aiWPSelectionIndex[ 1 ] = s_pxSelectionArray[ i ].m_iSelectionIndex;
        }
    }

    if ( ( aiWPSelectionIndex[ 0 ] == -1 ) || ( aiWPSelectionIndex[ 1 ] == -1 ) )
    {
        return false;
    }
    else if ( aiWPSelectionIndex[ 0 ] < aiWPSelectionIndex[ 1 ] )
    {
        iWaypointOne = aiWPIndex[ 0 ];
        iWaypointTwo = aiWPIndex[ 1 ];
        return true;
    }
    else
    {
        iWaypointOne = aiWPIndex[ 1 ];
        iWaypointTwo = aiWPIndex[ 0 ];
        return true;
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Selection_AStar
//
//////////////////////////////////////////////////////////////////////////////

float *Tools_Waypoint_Selection_AStar::s_pfMinDistances = NULL;
bool *Tools_Waypoint_Selection_AStar::s_pbFoundThisNode = NULL;
Tools_AStar* Tools_Waypoint_Selection_AStar::s_pxAStarList = NULL;
Asura_AStar_OpenList_Elem* Tools_Waypoint_Selection_AStar::s_pxAStarOpenList = NULL;
int Tools_Waypoint_Selection_AStar::s_iNumElemsInOpenList;

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Tools_Waypoint_Selection_AStar::Initialise()
{
    Asura_Assert(!s_pbFoundThisNode, "");
    if (Asura_AI_Navigation_System::GetNumberOfWaypoints())
    {
        s_pbFoundThisNode = new bool [Asura_AI_Navigation_System::GetNumberOfWaypoints()];
        s_pfMinDistances = new float[Asura_AI_Navigation_System::GetNumberOfWaypoints()];
        s_pxAStarList = new Tools_AStar[Asura_AI_Navigation_System::GetNumberOfWaypoints()];
        s_pxAStarOpenList = new Asura_AStar_OpenList_Elem [Asura_AI_Navigation_System::GetNumberOfWaypoints()];
    }
}

void Tools_Waypoint_Selection_AStar::Shutdown()
{
    delete []s_pfMinDistances;
    s_pfMinDistances = NULL;
    delete []s_pbFoundThisNode;
    s_pbFoundThisNode = NULL;
    delete[] s_pxAStarList;
    s_pxAStarList = NULL;
    delete[] s_pxAStarOpenList;
    s_pxAStarOpenList = NULL;
}

Tools_AStar* Tools_Waypoint_Selection_AStar::GetOpenListNode()
{
    float fMinCost = Asura_Maths::LargeFloat;
    int iIndex = -1;
    int iArgmin = -1;

    //Find argmin(m_fCost) from the open list.
    for(int i = 0; i < s_iNumElemsInOpenList; i++)
    {
        if(s_pxAStarOpenList[i].m_fTotalCost < fMinCost)
        {
            fMinCost = s_pxAStarOpenList[i].m_fTotalCost;
            iIndex   = s_pxAStarOpenList[i].m_iIndex;
            iArgmin  = i;
        }
    }

    if(iIndex == -1) return NULL;

    //Copy the last element in the open list over the position of the
    //element we found, and decrement the size of the open list.
    //This should work even if the node we found was at the end of the open list.
    s_pxAStarOpenList[iArgmin] = s_pxAStarOpenList[s_iNumElemsInOpenList-1];
    s_iNumElemsInOpenList--;
    return &s_pxAStarList[iIndex];
}

void Tools_Waypoint_Selection_AStar::DoAStar()
{
    if (!Asura_AI_Navigation_System::GetNumberOfWaypoints())
    {
        return;
    }
    Asura_Assert(s_pbFoundThisNode, "Not initialised");

    //Initialise the open list element count to 0
    s_iNumElemsInOpenList = 0;
    for (int i=0; i<Asura_AI_Navigation_System::GetNumberOfWaypoints(); i++)
    {
        s_pxAStarList[i].m_bOpen   = false;
        s_pxAStarList[i].m_bClosed = false;
        s_pxAStarList[i].m_iIndex  = i;
        s_pbFoundThisNode[i] = false;
        s_pfMinDistances[i]  = 0.0f;

    }

    for (int i=0; i<Asura_AI_Navigation_System::GetNumberOfWaypoints(); i++)
    {
        if ( Tools_Waypoint_Selection::s_pxSelectionArray[ i ].m_bIsSelected )
        {
            int iStartNodeIndex = i;
            Tools_AStar* pxCurrentNode = &s_pxAStarList[iStartNodeIndex];

            pxCurrentNode->m_bOpen       = true;
            pxCurrentNode->m_fCostToHere = 0.0f;
            pxCurrentNode->m_fCostToGoal = 0.0f;
            pxCurrentNode->m_fTotalCost  = pxCurrentNode->m_fCostToGoal + pxCurrentNode->m_fCostToHere;
            s_pfMinDistances[pxCurrentNode->m_iIndex] = pxCurrentNode->m_fTotalCost;
            pxCurrentNode->m_pxParent = NULL;

            //Add this node to the open list.
            s_pxAStarOpenList[s_iNumElemsInOpenList].m_fTotalCost = pxCurrentNode->m_fTotalCost;
            s_pxAStarOpenList[s_iNumElemsInOpenList].m_iIndex     = pxCurrentNode->m_iIndex;
            s_iNumElemsInOpenList++;

            //The call to GetOpenListNode automatically removes the new current node
            //from the open list.
            while((pxCurrentNode = GetOpenListNode()) != NULL)
            {
                int iIndex = pxCurrentNode->m_iIndex;
                s_pbFoundThisNode[iIndex] = true;
                for (u_char ucLink=0; ucLink<Asura_AI_Navigation_System::GetWaypoint(iIndex)->m_ucNumConnections; ucLink++)
                {
                    int iNewNodeIndex = Asura_AI_Navigation_System::GetWaypoint(iIndex)->m_pxConnections[ucLink].m_usIndex;
                    Tools_AStar* pxNewNode = &s_pxAStarList[iNewNodeIndex];

                    float fNewCostToHere = pxCurrentNode->m_fCostToHere + Asura_AI_Navigation_System::GetWaypoint(iIndex)->m_pxConnections[ucLink].GetLength();

                    if ((pxNewNode->m_bOpen || pxNewNode->m_bClosed) &&
                        pxNewNode->m_fCostToHere<=fNewCostToHere)
                    {
                        continue;
                    }

                    pxNewNode->m_pxParent    = pxCurrentNode;
                    pxNewNode->m_fCostToHere = fNewCostToHere;
                    pxNewNode->m_fCostToGoal = 0.0f;
                    pxNewNode->m_fTotalCost  = pxNewNode->m_fCostToHere + pxNewNode->m_fCostToGoal;
                    s_pfMinDistances[pxNewNode->m_iIndex] = pxNewNode->m_fTotalCost;

                    pxNewNode->m_bClosed = false;
                    pxNewNode->m_bOpen   = true;

                    s_pxAStarOpenList[s_iNumElemsInOpenList].m_fTotalCost = pxNewNode->m_fTotalCost;
                    s_pxAStarOpenList[s_iNumElemsInOpenList].m_iIndex     = pxNewNode->m_iIndex;
                    s_iNumElemsInOpenList++;
                }       

                pxCurrentNode->m_bClosed = true;
                pxCurrentNode->m_bOpen   = false;
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
