//////////////////////////////////////////////////////////////////////////////
//
// main.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <cell/atomic.h>
#include <cell/dma.h>
#include <cell/sync/barrier.h>
#include <cell/sync/lfqueue.h>
#include <edge/geom/edgegeom.h>
#include <sys/spu_thread.h>
#include <spu_intrinsics.h>
#include <spu_printf.h>

// Asura
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"

#include "Rendering/Asura_PS3_SPU_SkinningTypes.h"

//
// Debug printfs
//

//BW at present these are very slow and cause a sync problem, specifically the character's bones get updated before SPUs are finished so bits of the character will flicker & jump around.

#if (0)	// just mark the start & end
	#define PING(a) { if (a) { spu_printf("SPU %d: on\n", g_iSPUNum); } else { spu_printf("SPU %d: off\n", g_iSPUNum); } }
#else
	#define PING(a)
#endif

#if (0)	// general debug info
	#define DEBUG_PRINTF(args) spu_printf args
#else
	#define DEBUG_PRINTF(args)
#endif

#if (0)
	#define DEBUG_DUMP_TASK()																\
	{																						\
		spu_printf("\nSPU %d: Task info:-\n", g_iSPUNum);									\
		spu_printf("  m_uInputAddr %016llx\n", g_xJobInfo.m_uInputAddr);					\
		spu_printf("  uVertsOutAddr %016llx\n", g_xJobInfo.m_uVertsOutAddr);				\
		spu_printf("  uNumVerts %08x\n", g_xJobInfo.m_uNumVerts);							\
		spu_printf("  m_eMode %s\n", g_xJobInfo.m_eMode ? "shadow" : "full" );				\
		spu_printf("\n");																	\
	}
#else
	#define DEBUG_DUMP_TASK()
#endif

#if (0) // DMA debug
#define DEBUG_ATOMIC( pLocal, pMain, bWrite )											\
		{																								\
		if ( bWrite )																			\
			{																						\
			spu_printf("SPU %d: Atomic via %08p to %016llx\n", g_iSPUNum, pLocal, pMain);		\
			}																						\
			else																					\
			{																						\
			spu_printf("SPU %d: Atomic via %08p from %016llx\n", g_iSPUNum, pLocal, pMain);		\
			}																						\
		}

#define DEBUG_GET( pLocal, pMain, uSize ) spu_printf("SPU %d: DMA GET %d bytes to %08p from %016llx\n", g_iSPUNum, uSize, pLocal, pMain);
#define DEBUG_PUT( pLocal, pMain, uSize ) spu_printf("SPU %d: DMA PUT %d bytes from %08p to %016llx\n", g_iSPUNum, uSize, pLocal, pMain)

#else
#define DEBUG_ATOMIC( pLocal, pMain, bWrite )
#define DEBUG_GET( pLocal, pMain, uSize )
#define DEBUG_PUT( pLocal, pMain, uSize )
#endif

#if (0)	// vertex shading debug
	static int g_iCurrentVertex;
	#define DEBUG_VERTEX_ID(var) g_iCurrentVertex = var;
	#if (1)
		//debug a specific vertex
		#define DEBUG_VERTEX 200
		#define DEBUG_VERTEX_TEST if ( g_iCurrentVertex == DEBUG_VERTEX )
	#else
		//debug all vertices
		#define DEBUG_VERTEX_TEST
	#endif

	#define DEBUG_VERTEX_SCALAR(var) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: "#var" %.2f\n", g_iSPUNum, g_iCurrentVertex, var ); }
	#define DEBUG_VERTEX_VECTOR(var) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: "#var" (%.2f, %.2f, %.2f, %.2f)\n", g_iSPUNum, g_iCurrentVertex, var.GetX(), var.GetY(), var.GetZ(), var.GetW() ); }

 	#define DEBUG_VERTEX_TYPE(format,var) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: "#var" "#format"\n", g_iSPUNum, g_iCurrentVertex, var ); }
 	#define DEBUG_VERTEX_TYPE4(format,var0,var1,var2,var3)																			\
 		DEBUG_VERTEX_TEST																											\
 		{																															\
 			spu_printf( "SPU %d: vertex %d: "#var0" "#format", "#format", "#format", "#format"\n",									\
 											g_iSPUNum, g_iCurrentVertex, var0, var1, var2, var3 );									\
 		}
 		
 	#define DEBUG_VERTEX_PRINTF(args) DEBUG_VERTEX_TEST { spu_printf( "SPU %d: vertex %d: ", g_iSPUNum, g_iCurrentVertex ); spu_printf(args); }
#else
	#define DEBUG_VERTEX_ID(var)
	#define DEBUG_VERTEX_SCALAR(var)
	#define DEBUG_VERTEX_VECTOR(var)
	#define DEBUG_VERTEX_TYPE(szFormat,var)
	#define DEBUG_VERTEX_TYPE4(szFormat,var0,var1,var2,var3)
	#define DEBUG_VERTEX_PRINTF(args)
#endif


//
// Defines
//


//tags, 0->31
enum {
	DMA_TAG_VERTS	= 0, //buffer 0 = +0, buffer 1 = +1
	DMA_TAG_INDICES	= 2, //
	DMA_TAG_GET_JOB = 4,
	DMA_TAG_FLAG_COMPLETE,
	DMA_TAG_BONES,
	DMA_TAG_GET_BATCH
};

#define SIZEOF_INDEX 2	// unsigned shorts, no guarantee this will work if you change it.

//
// Variables
//

extern u_int _end; // End of executable (plus static data, etc.)

// static const u_int uSTACK_SIZE_TO_ALLOW = 4 * 1024; // Needs to be a multiple of 16.
// 

// These are not used in this SDK, but they recommend 0 for future compatibility (see Reference_System-Programming_Note section 4)
static const u_int uDMA_TID = 0;
static const u_int uDMA_RID = 0;

static int g_iSPUNum = -1;

//BW dammit what header are these in?
#define max(A,B) (((A)>(B))?(A):(B))
#define min(A,B) (((A)<(B))?(A):(B))


// Bone & vertex buffers
#define BUFFER_SIZE (64*1024) // was 100k but edge lib is BIG

static const u_int uUseableBufferSize = BUFFER_SIZE - 12; //unaligned data can be off by up to 12 bytes, so leave that much extra

static char buff[2][BUFFER_SIZE] __attribute__((aligned(128))); //Data buffers (I don't see any point allocing)
static float g_afCullPositions[3*BUFFER_SIZE/68]; //BW temp! positions in an array for culling. Ultimately we should just have 1 array but for now we duplicate position to here after shading it.

Asura_AMatrix4x4 g_axBoneMatrices[256]; //BW todo: DO NOT use a fixed size array! this should be alloced for each character

static u_int g_uStartVertex;


static Asura_PS3_SPU_SkinningJob g_xJobInfo __attribute__((aligned(16)));

static u_int g_uAtomicBuf[32] __attribute__((aligned(128))); //buffer used by atomic operations
static u_int g_auAligningBuffer[5] __attribute__((aligned(16))); //for small dma's, so all possible alignments of a 4-byte word can be handled

static u_int g_uInputVertexSize;
static u_int g_uOutputVertexSize;

static EdgeGeomViewportInfo g_xViewportInfo = {
	//BW hard coded based on Asura_PS3_Graphics::SetViewport for main view
	//BW todo: upload this for current viewport!
	{0,0,1280,720}, // uint16_t scissorArea[4];
	{0.0f,1.0f},//{fASURA_PS3_DEFAULT_VIEWPORT_MIN_Z, fASURA_PS3_DEFAULT_VIEWPORT_MAX_Z}, // float depthRange[2];
	{1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1},	// float viewProjectionMatrix[16];
	{640,360,0.5,0.0f},	// float viewportScales[4];
	{640,360,0.5,0.0f},	// float viewportOffsets[4];
	CELL_GCM_SURFACE_CENTER_1,	//uint8_t sampleFlavor;
	{} //pad
};


// some debug stats
static u_int g_uTotalVerts = 0;
static u_int g_uTotalBatches = 0;



//
// Functions
//

//
// HalfToFloatVec
//
// Convert 4 halfs in LSBs of input int register into floats
//
inline vec_float4 HalfToFloatVec( vec_uint4 xHalfs )
{
	// 	half (1|5|10) -> float (1|8|23)
	// 	seeeeeffffffffff
	//
	// 	s000eeeeeffffffffff0000000000000
	// 	+01110000	//offset so that 10000 -> 10000000 (exponent of 0, represented as halfway through the uint5 or uint8 range)

	const vec_uint4 x8000 = {0x8000,0x8000,0x8000,0x8000};
	const vec_uint4 x7fff = {0x7fff,0x7fff,0x7fff,0x7fff};
	const vec_uint4 xHalfExp = {0x7c00,0x7c00,0x7c00,0x7c00};
	const vec_uint4 x38000000 = {0x38000000,0x38000000,0x38000000,0x38000000};
	const vec_uint4 x0 = {0,0,0,0};
	const vec_uint4 xff800000 = {0xff800000,0xff800000,0xff800000,0xff800000};

	vec_uint4 xTemp = spu_sl(xHalfs & x8000, 16) + spu_sl(xHalfs & x7fff,13);
	xTemp += x38000000;
	
	vec_float4 xResult = reinterpret_cast<vec_float4>(xTemp);
	
	// do we need to fix any denormalised values?
	vec_uint4 xExps = xHalfs & xHalfExp;
	if ( vec_any_eq(xExps,x0) )
	{
		//fix denormalized values & zero, by subtracting the same float value without a mantissa (which is + or - 0x38000000)
		//e.g. we've changed 0x0000 into 0x38000000, we then subtract 0x38000000 to get 0.0
		//e.g. 0x0010 becomes 0x38020000, the bit that's set is at the correct point, but floats have an implicit high bit set, so subtract that.
		xTemp &= xff800000;
		vec_float4 xF = reinterpret_cast<vec_float4>(xTemp);
	
		//BW this could easily be optimised to use vectors
		for ( int i=0; i<4; i++ )
		{
			if ( xExps[i] == 0 )
			{
				xResult[i] = xResult[i] - xF[i];
			}
		}
	}
	
	return xResult;
}

//BW todo: see if the compiler gets what I'm trying to do here or if it's dumb, if so help it out with more intrinsics and static consts etc.
inline void Unpack11_11_10( Asura_AVector4& xOut, int iPacked )
{
// 	DEBUG_VERTEX_PRINTF(("Unpack11_11_10\n"));
// 	DEBUG_VERTEX_TYPE("%08x",iPacked);

	const vec_int4 xShift = { 1<<21, 1<<10, 1, 0 }; //shift left first to cause sign extend
	const vec_int4 xMask = { 0xffe00000, 0xffe00000, 0xffc00000, 0 }; //1 less bit of precision in z
	const vec_float4 xScale = { 1024.0f/1023.0f, 1024.0f/1023.0f, 512.0f/511.0f, 0.0f };

	vec_int4 xInts = {iPacked,iPacked,iPacked,iPacked};
	xInts *= xShift;
	xInts &= xMask;

	vec_float4 xVec = spu_convtf(xInts,31); //convert to floats, divide by 1<<30, neato
	xVec *= xScale;

// 	DEBUG_VERTEX_TYPE4("%08x",xInts[0],xInts[1],xInts[2],xInts[3]);
// 	DEBUG_VERTEX_TYPE4("%.2f",xVec[0],xVec[1],xVec[2],xVec[3]);
	
	xOut.SetVector( xVec );
}

//BW lowest bits sometimes change => there's a small precision error, but otherwise pack === unpack
inline u_int Pack11_11_10( Asura_AVector4& xVec )
{
// 	DEBUG_VERTEX_PRINTF(("Pack11_11_10\n"));
// 	DEBUG_VERTEX_VECTOR(xVec);

	const vec_float4 xScale = { 1023.0f, 1023.0f*(float)(1<<11), 511.0f*(float)(1<<22) }; //stick the shift in here, saves some work
	const vec_int4 xMask = { 0x7ff, 0x7ff<<11, 0x3ff<<22, 0 };
	vec_float4 xVS = xVec.GetVector();

	xVS *= xScale;
	vec_int4 xInts = spu_convts(xVS,0);
	xInts &= xMask;

// 	DEBUG_VERTEX_TYPE4("%08x",xInts[0],xInts[1],xInts[2],xInts[3]);
// 	DEBUG_VERTEX_TYPE("%08x",(xInts[0] | xInts[1] | xInts[2]));

	return xInts[0] | xInts[1] | xInts[2];
}


inline size_t GetVertexSize( SPU_SKINNING_MODE eMode )
{
	switch ( eMode )
	{
		default: // fall through
		case SPU_SKINNING_FULL_CULL:
		{
			return sizeof(Asura_PS3_SPU_Skinning_OutputVertex);
			break;
		}
		case SPU_SKINNING_POS_ONLY_NO_CULL:
		{
			return sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly);
			break;
		}
	}
}


//
// ShadeVertex
//
// Ben Weston - June '08
//
// Adapted from SmoothSkin_VP.vpx. Most of the comments are the same as the original shader.
//
void ShadeVertex( Asura_PS3_SPU_Skinning_OutputVertex *pxOut, Asura_PS3_SPU_Skinning_InputVertex *pxIn, float* pxCullPosition )
{
	Asura_AVector4	xInPos = pxIn->m_xPosition; 
	Asura_AVector4	xObjectSpacePosition, axBonePos[4]; 
	float			fAlbedoBlend = 0.0f;
	xInPos.SetW(1.0f);

	vec_float4 xWeights = HalfToFloatVec( (vec_uint4){pxIn->m_afBoneWeights[0], pxIn->m_afBoneWeights[1], pxIn->m_afBoneWeights[2], pxIn->m_afBoneWeights[3]} );

	g_axBoneMatrices[ pxIn->m_aiBoneIndices[0] ].Apply( xInPos, axBonePos[0] );
	g_axBoneMatrices[ pxIn->m_aiBoneIndices[1] ].Apply( xInPos, axBonePos[1] );
	g_axBoneMatrices[ pxIn->m_aiBoneIndices[2] ].Apply( xInPos, axBonePos[2] );
	g_axBoneMatrices[ pxIn->m_aiBoneIndices[3] ].Apply( xInPos, axBonePos[3] );
	
	xObjectSpacePosition	= (axBonePos[0] * xWeights[0])
							+ (axBonePos[1] * xWeights[1])
							+ (axBonePos[2] * xWeights[2])
							+ (axBonePos[3] * xWeights[3]);
 	xObjectSpacePosition.SetW(1.0f);

	fAlbedoBlend = g_axBoneMatrices[ pxIn->m_aiBoneIndices[0] ].GetElement(3,0) * xWeights[0]
				 + g_axBoneMatrices[ pxIn->m_aiBoneIndices[1] ].GetElement(3,0) * xWeights[1]
				 + g_axBoneMatrices[ pxIn->m_aiBoneIndices[2] ].GetElement(3,0) * xWeights[2]
				 + g_axBoneMatrices[ pxIn->m_aiBoneIndices[3] ].GetElement(3,0) * xWeights[3];

	DEBUG_VERTEX_VECTOR(xInPos);
	DEBUG_VERTEX_VECTOR(axBonePos[0]);
	DEBUG_VERTEX_VECTOR(axBonePos[1]);
	DEBUG_VERTEX_VECTOR(axBonePos[2]);
	DEBUG_VERTEX_VECTOR(axBonePos[3]);
	DEBUG_VERTEX_TYPE4(" %04x ",pxIn->m_afBoneWeights[0],pxIn->m_afBoneWeights[1],pxIn->m_afBoneWeights[2],pxIn->m_afBoneWeights[3]);
	DEBUG_VERTEX_TYPE4(" %.2f ",xWeights[0],xWeights[1],xWeights[2],xWeights[3]);
	DEBUG_VERTEX_VECTOR(xObjectSpacePosition);

	Asura_AMatrix4x4 xMatrix = g_axBoneMatrices[ pxIn->m_aiBoneIndices[0] ];
	xMatrix.SetElement(3,0, 0.0f);

	Asura_AVector4 xInNormal, xInTangent, xInBinormal;
	Unpack11_11_10(xInNormal,	pxIn->m_xNormal);
	Unpack11_11_10(xInTangent,	pxIn->m_xTangent);
	Unpack11_11_10(xInBinormal,	pxIn->m_xBinormal);
	Asura_AVector4 xNormal, xTangent, xBinormal;
	xMatrix.Apply( xInNormal,	xNormal );
	xMatrix.Apply( xInTangent,	xTangent );
	xMatrix.Apply( xInBinormal,	xBinormal );

	Asura_AVector4 xOutPosition;
	g_xJobInfo.m_xShaderConstants.m_xWorldViewProjectionMatrix.Apply( xObjectSpacePosition, xOutPosition );
	xOutPosition.StoreUnaligned( pxOut->m_xPosition );
	
	xObjectSpacePosition.StoreUnaligned( pxCullPosition ); // actually this only wants 3 components, but that's ok

//BW todo 	xOut.m_xUVs = MaterialResponse_AnimateUVs( xIn.m_xUV );
//do it in the shader (we don't have any per-material settings in here)
//and don't bother passing these to the SPU
	pxOut->m_xUVs[0] = pxIn->m_xUV[0];
	pxOut->m_xUVs[1] = pxIn->m_xUV[1];
	pxOut->m_xUVs[2] = pxIn->m_xUV[2];
	pxOut->m_xUVs[3] = pxIn->m_xUV[3];
 
	// Direction to camera, and world-space distance to vertex from viewer
	Asura_AVector4 xCameraDirDistance = g_xJobInfo.m_xShaderConstants.m_xObjectSpaceCameraPosition - xObjectSpacePosition;
	xCameraDirDistance.SetW( xOutPosition.GetW() );

	xCameraDirDistance.StoreUnaligned( pxOut->m_xCameraDirDistance );

 	// We're doing lighting in world-space now
	g_xJobInfo.m_xShaderConstants.m_xWorldMatrix.Apply( xNormal, xNormal ); //BW these things have 0 in W, it will go wierd otherwise
	g_xJobInfo.m_xShaderConstants.m_xWorldMatrix.Apply( xTangent, xTangent );
	g_xJobInfo.m_xShaderConstants.m_xWorldMatrix.Apply( xBinormal, xBinormal );
	pxOut->m_xNormal	= Pack11_11_10(xNormal);
	pxOut->m_xTangent	= Pack11_11_10(xTangent);
	pxOut->m_xBinormal	= Pack11_11_10(xBinormal);

 	//Asura_AMatrix3x3 xTangentToObjectMatrix( xInTangent, xInBinormal, xInNormal );

	// Note transposed mul, to do objectspace->tangent
	Asura_AVector4 xTangentSpaceCameraDir;
	xTangentSpaceCameraDir.SetX( xInTangent.DotProduct3(xCameraDirDistance) );
	xTangentSpaceCameraDir.SetY( xInBinormal.DotProduct3(xCameraDirDistance) );
	xTangentSpaceCameraDir.SetZ( xInNormal.DotProduct3(xCameraDirDistance) );

	// lerp seems to be backwards to what the damage maps expect
	xTangentSpaceCameraDir.SetW( 1.0f - fAlbedoBlend );
	
	xTangentSpaceCameraDir.StoreUnaligned( pxOut->m_xTangentSpaceCameraDir );
}

void ShadeVertexPosOnly( Asura_PS3_SPU_Skinning_OutputVertex_PosOnly *pxOut, Asura_PS3_SPU_Skinning_InputVertex *pxIn )
{
	Asura_AVector4	xInPos = pxIn->m_xPosition; 
	Asura_AVector4	axBonePos[4]; 
	float			fAlbedoBlend = 0.0f;
	xInPos.SetW(1.0f);

	vec_float4 xWeights = HalfToFloatVec( (vec_uint4){pxIn->m_afBoneWeights[0], pxIn->m_afBoneWeights[1], pxIn->m_afBoneWeights[2], pxIn->m_afBoneWeights[3]} );

	g_axBoneMatrices[ pxIn->m_aiBoneIndices[0] ].Apply( xInPos, axBonePos[0] );
	g_axBoneMatrices[ pxIn->m_aiBoneIndices[1] ].Apply( xInPos, axBonePos[1] );
	g_axBoneMatrices[ pxIn->m_aiBoneIndices[2] ].Apply( xInPos, axBonePos[2] );
	g_axBoneMatrices[ pxIn->m_aiBoneIndices[3] ].Apply( xInPos, axBonePos[3] );

	pxOut->m_xPosition	= (axBonePos[0] * xWeights[0])
						+ (axBonePos[1] * xWeights[1])
						+ (axBonePos[2] * xWeights[2])
						+ (axBonePos[3] * xWeights[3]);
	pxOut->m_xPosition.SetW(1.0f);

//BW moved to vertex shader so multiple shadows can share the same skinned positions
// 	g_xJobInfo.m_xShaderConstants.m_xWorldViewProjectionMatrix.Apply( pxOut->m_xPosition, pxOut->m_xPosition );

//BW optimisation: do this only once per character, so vertex shader would take 3 input buffers:
//	position - shared with z-only passes (like shadows), needs to be projected in shader
//  uvuv - read from original input data, needs to be animated by VS
//  everything else - created by the main SPU shading process, projected into view space? -unless we expect to get multiple viewports with full shading.
//probable overheads for: skipping data (cache) or using multiple input buffers (catalin thinks docs say 2 is ok)
}

// read what the next batch of data is to process, and remove it from the queue so another SPU doesn't duplicate this work
// return false if no more data
bool GetNextBatch( uint64_t& pVertsIn, uint64_t& pVertsOut, u_int& uNumVerts, u_int uMaxVerts, uint64_t& pIdxIn, uint64_t& pIdxOut, u_int& uNumIdx )
{
	if ( g_xJobInfo.m_eMode == SPU_SKINNING_FULL_CULL )
	{
		// input data is a batch list, load the prescribed batch
		Asura_PS3_SPU_Skinning_Batch xBatch;
		
		cellDmaGet( &xBatch, g_xJobInfo.m_uInputAddr, sizeof(Asura_PS3_SPU_Skinning_Batch), DMA_TAG_GET_BATCH, uDMA_TID, uDMA_RID );
		cellDmaWaitTagStatusAll( 1 << DMA_TAG_GET_BATCH ); //BW really we shouldn't have to wait here...

		g_xJobInfo.m_uInputAddr += sizeof(Asura_PS3_SPU_Skinning_Batch);
		
		// detect the terminator
		if ( xBatch.m_uNumVerts == 0 )
		{
			DEBUG_PRINTF(("SPU %d: Character finished\n", g_iSPUNum));
			return false;
		}
		
		g_uTotalVerts += xBatch.m_uNumVerts;
		g_uTotalBatches++;

		pVertsIn = xBatch.m_uVertAddr;
		pVertsOut = g_xJobInfo.m_uVertsOutAddr;
		uNumVerts = xBatch.m_uNumVerts;
		
		Asura_Assert( uNumVerts <= uMaxVerts, "SPU_Skinning GetNextBatch - BATCH IS BIGGER THAN uMaxVerts!!! PANIC!!!\n" );
		//BW hopefully this will only fire if characters haven't been re-exported
		
		pIdxIn = xBatch.m_uStartIndexAddr;
		pIdxOut = g_xJobInfo.m_uIndicesOutAddr;
		uNumIdx = xBatch.m_uNumIndices;

		DEBUG_PRINTF(("SPU %d: Got *Actual* Batch: read from %016llx num verts %u\n", g_iSPUNum, pVertsIn, uNumVerts));

		// move the write pointers forwards
		g_xJobInfo.m_uVertsOutAddr += g_uOutputVertexSize*uNumVerts;
		g_xJobInfo.m_uIndicesOutAddr += SIZEOF_INDEX*uNumIdx;
	}
	else
	{
		// input data is a raw vertex data, load as much as we can.

		if ( g_uStartVertex >= g_xJobInfo.m_uNumVerts )
		{
			//move on to the next character
			//BW todo start loading the next character's verts! (like so: if g_xNextTask.uNumVerts > 0 etc.)
			DEBUG_PRINTF(("SPU %d: Character finished\n", g_iSPUNum));

			return false;
		}
		
		pVertsIn = g_xJobInfo.m_uInputAddr + g_uStartVertex*g_uInputVertexSize;
		pVertsOut = g_xJobInfo.m_uVertsOutAddr + g_uStartVertex*g_uOutputVertexSize;
		uNumVerts = min( g_xJobInfo.m_uNumVerts - g_uStartVertex, uMaxVerts );

		pIdxIn = 0;
		pIdxOut = 0;
		uNumIdx = 0;

		DEBUG_PRINTF(("SPU %d: Got Batch: start Vertex %u num verts %u\n", g_iSPUNum, g_uStartVertex, uNumVerts));
	}

	return true;
}

int main( uint32_t uSPUNum, uint64_t uJobQueueAddr, uint64_t uGcmLabelAddr, uint64_t uBarrierAddr )
{
	u_int uTransferBuff = 0;
	bool bDidAnything = false;
	
	g_iSPUNum = static_cast<int>(uSPUNum);

	PING(true);
	DEBUG_PRINTF(("SPU %d: Running (debug: %016llx)\n", g_iSPUNum, uDebug));	

	CellSyncLFQueuePopContainer xQueuePopContainer;
	cellSyncLFQueuePopContainerInitialize( &xQueuePopContainer, &g_xJobInfo, DMA_TAG_GET_JOB );

	// task loop (1 task = 1 character's  skin)
	while (true)
	{
		int iRet = cellSyncLFQueuePopBegin( uJobQueueAddr, &xQueuePopContainer ); // busy-wait for data if queue's empty
		iRet |= cellSyncLFQueuePopEnd( uJobQueueAddr, &xQueuePopContainer );
		if ( iRet != CELL_OK )
		{
			spu_printf("Spu_Skinning: cellSyncLFQueuePop failed. return code %08x\n", cellSyncLFQueuePopBegin);
			break;
		}
		cellDmaWaitTagStatusAll( 1 << DMA_TAG_GET_JOB );


		DEBUG_PRINTF(("SPU %d: Got task\n", g_iSPUNum));
		DEBUG_DUMP_TASK();


		if ( g_xJobInfo.m_uNumVerts == 0 ) // detect terminator
		{
			break;
		}
		
		bDidAnything = true;
		

		//get size of vertices
		g_uInputVertexSize = sizeof(Asura_PS3_SPU_Skinning_InputVertex);
		g_uOutputVertexSize = GetVertexSize( g_xJobInfo.m_eMode );

		//arrange in/out positions to prevent output verts overwriting input verts until those inputs have been read (i.e. a vert can overwrite itself but nothing later in the list)
 		u_int uMaxVerts = uUseableBufferSize / max(g_uInputVertexSize,g_uOutputVertexSize);
		u_int uInStart = BUFFER_SIZE-uMaxVerts*g_uInputVertexSize;
		u_int uOutStart = 0;
		
// 			DEBUG_PRINTF(("g_uInputVertexSize %d\n", g_uInputVertexSize));
// 			DEBUG_PRINTF(("g_uOutputVertexSize %d\n", g_uOutputVertexSize));
// 			DEBUG_PRINTF(("Batch size %d verts\n\n", uMaxVerts));

// 		//remember the address of the vertex counter in the task info in main memory, so we can update it atomically so the other SPUs can see it
// 		g_uVertCounterAddr = uTaskInfoAddr + offsetof(Asura_PS3_SPU_SkinningJob,m_uVertCounter); //BW this threw a warning (which it really shouldn't since it's POD), turned on -Wno-invalid-offsetof to make it shut up. The code does actually work.
//BW using local vert counter now, just 1 spu per character

		uint64_t apVertsOut[2];
		u_int uNumVerts[2];
		uint64_t apIndicesOut[2];
		u_int auIndicesStart[2];
		u_int uNumIndices[2];
		bool bDataLastTime = false;

		//load bones
		cellDmaLargeGetf( g_axBoneMatrices, g_xJobInfo.m_xShaderConstants.m_uBoneArrayAddr, g_xJobInfo.m_xShaderConstants.m_uNumBones*sizeof(Asura_AMatrix4x4), DMA_TAG_BONES, uDMA_TID, uDMA_RID );
		//BW sync issue, these haven't been double buffered on PPU side, so we need to be running within 1 frame.
		//BW n.b. depends when character bones get updated but our SPU barrier should be adequate
		//BW n.b. SPUs are way faster than a frame anyhow

		g_uStartVertex = 0;
		u_int uIndexOffset = 0;

		// vertex batch loop
		do
		{
			bool bDataThisTime = false;

			uint64_t pVertsIn;
			uint64_t pIndicesIn;
			if ( GetNextBatch( pVertsIn, apVertsOut[uTransferBuff], uNumVerts[uTransferBuff], uMaxVerts,
								pIndicesIn, apIndicesOut[uTransferBuff], uNumIndices[uTransferBuff]) )
			{
				// queue a get after the preceding put from the same buffer
				// Get with a fence, to make sure the preceding put with the same tag gets finished first.
				DEBUG_GET( &buff[uTransferBuff][uInStart], pVertsIn, uNumVerts[uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex) );
				cellDmaLargeGetf( &buff[uTransferBuff][uInStart], pVertsIn, uNumVerts[uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex), DMA_TAG_VERTS+uTransferBuff, uDMA_TID, uDMA_RID );
				// this assumes input data is 16-byte aligned, output data doesn't assume this at present.
				
				if ( pIndicesIn )
				{
					//compute a start position from end of vertex data & alignment
//					auIndicesStart[uTransferBuff] = (pIndicesIn&0xf) | ((uInStart+uNumVerts[uTransferBuff]*sizeof(Asura_PS3_SPU_Skinning_InputVertex)+0xf)&~0xf);
					auIndicesStart[uTransferBuff] = (pIndicesIn&0xf) | ((BUFFER_SIZE-uNumIndices[uTransferBuff]*SIZEOF_INDEX-0xf)&~0xf);
					DEBUG_GET( &buff[uTransferBuff][auIndicesStart[uTransferBuff]], pIndicesIn, uNumIndices[uTransferBuff]*SIZEOF_INDEX );
					cellDmaUnalignedGetf( &buff[uTransferBuff][auIndicesStart[uTransferBuff]], pIndicesIn, uNumIndices[uTransferBuff]*SIZEOF_INDEX, DMA_TAG_INDICES+uTransferBuff, uDMA_TID, uDMA_RID );
				}
				
				bDataThisTime = true;
			}

			u_int uProcessBuff = 1 - uTransferBuff;
			

			if ( bDataLastTime )
			{
				cellDmaWaitTagStatusAll( (1<<DMA_TAG_BONES) | (1<<(DMA_TAG_VERTS+uProcessBuff)) | (1<<(DMA_TAG_INDICES+uProcessBuff)) ); //wait for loading to complete (DMA_TAG_VERTEX_DATA_BUF0 or 1)

				u_int uOutAlignment = 0xf & apVertsOut[uProcessBuff]; //output data might not be aligned (because output vertex size is 68 bytes)
				char* pInData = &buff[uProcessBuff][uInStart];
				char* pOutData = &buff[uProcessBuff][uOutStart+uOutAlignment];
				char* pInDataEnd = pInData + uNumVerts[uProcessBuff]*g_uInputVertexSize;
				
				//
				// SHADE ALL VERTICES
				//
				switch ( g_xJobInfo.m_eMode )
				{
					case SPU_SKINNING_FULL_CULL:
					{
						float* pxCullPositions = g_afCullPositions;
					
						int iVertCount=0;
						while ( pInData < pInDataEnd )
						{
							DEBUG_VERTEX_ID(iVertCount++);

							//we use the same buffer for input & output, arranged so that in the worst case each output vertex will overwrite an input vertex
							//BW to test: if this causes problems could extend the buffers to allow a 1 vertex space
							ShadeVertex(reinterpret_cast<Asura_PS3_SPU_Skinning_OutputVertex*>(pOutData), reinterpret_cast<Asura_PS3_SPU_Skinning_InputVertex*>(pInData), pxCullPositions);
							
							pxCullPositions += 3;
							pOutData += g_uOutputVertexSize;
							pInData += g_uInputVertexSize;
						}
						break;
					}

					case SPU_SKINNING_POS_ONLY_NO_CULL:
					{
						int iVertCount=0;
						while ( pInData < pInDataEnd )
						{
							DEBUG_VERTEX_ID(iVertCount++);

							//we use the same buffer for input & output, arranged so that in the worst case each output vertex will overwrite an input vertex
							//BW to test: if this causes problems could extend the buffers to allow a 1 vertex space
							ShadeVertexPosOnly(reinterpret_cast<Asura_PS3_SPU_Skinning_OutputVertex_PosOnly*>(pOutData), reinterpret_cast<Asura_PS3_SPU_Skinning_InputVertex*>(pInData));
							
							pOutData += g_uOutputVertexSize;
							pInData += g_uInputVertexSize;
						}
						break;
					}
				}

				
				if ( g_xJobInfo.m_eMode != SPU_SKINNING_POS_ONLY_NO_CULL )
				{
					//
					// CULL TRIANGLES
					//

					//BW ultimately all this could be filled in by the tools, for now do it manually
					u_short *puIndices = (u_short*)&buff[uProcessBuff][auIndicesStart[uProcessBuff]];

					EdgeGeomSpuConfigInfo xEdgeCfg;
					xEdgeCfg.flagsAndUniformTableCount = 0 | EDGE_GEOM_FLAG_STATIC_GEOMETRY_FAST_PATH;
					xEdgeCfg.commandBufferHoleSize = 0; // Irrelevant.
					xEdgeCfg.inputFlavor = EDGE_GEOM_INPUT_FLAVOR_F32c3;
					xEdgeCfg.outputFlavor = EDGE_GEOM_OUTPUT_FLAVOR_F32c3; // Irrelevant - I hope
					xEdgeCfg.shapeFlavor = 0; // Irrelevant.
					xEdgeCfg.indexesFlavorAndSkinningFlavor = EDGE_GEOM_INDEXES_U16_TRIANGLE_LIST_CCW << 4;
					xEdgeCfg.numVertexes = uNumVerts[uProcessBuff];
					xEdgeCfg.numIndexes = uNumIndices[uProcessBuff];
					xEdgeCfg.indexesOffset = 0xffffffff;

					EdgeGeomLocalToWorldMatrix xLocalToWorldMat =
					{
						{ 1.0f, 0.0f, 0.0f, 0.0f,
						0.0f, 1.0f, 0.0f, 0.0f,
						0.0f, 0.0f, 1.0f, 0.0f, }
					};


					g_xJobInfo.m_xShaderConstants.m_xWorldViewProjectionMatrix.StoreRowsUnaligned( g_xViewportInfo.viewProjectionMatrix );
					

					// Initialise the Edge library.
					EdgeGeomSpuContext xEdgeContext;
					edgeGeomInitialize( &xEdgeContext,
						&xEdgeCfg,
						NULL, 0, //pvScratchBfr, uScratchBfrSize,
						NULL, 0, //pvStartOfAvailableMemory, uTotalMemoryAvailable,
						0, // Irrelevant - we'll be dealing with all transfers.
						&g_xViewportInfo,
						&xLocalToWorldMat,
						NULL );

					
					//buffer needs to be aligned and relative to our first vertex
					u_short* pAlignedIndices = reinterpret_cast<u_short*>(reinterpret_cast<u_int>(puIndices)&~0xf);
					for ( int i=0; i < (int)uNumIndices[uProcessBuff]; i++ )
					{
						pAlignedIndices[i] = puIndices[i] - uIndexOffset;
					}

					// EDGE_GEOM_FLAG_STATIC_GEOMETRY_FAST_PATH should, hopefully, stop these from doing anything 
					edgeGeomDecompressVertexes( &xEdgeContext, g_afCullPositions, NULL );
					edgeGeomDecompressIndexes( &xEdgeContext, pAlignedIndices );

					u_int uOutputIndexCount = edgeGeomCullTriangles( &xEdgeContext, EDGE_GEOM_CULL_BACKFACES_AND_FRUSTUM );
					

//spu_printf("uOutputIndexCount %d / %d\n", uOutputIndexCount, uNumIndices[uProcessBuff]);

					//adjust the alignment to match main memory output buffer & correct the index for our big vertex table
					for ( int i=(int)uOutputIndexCount-1; i >= 0; i-- )
					{
						puIndices[i] = pAlignedIndices[i] + uIndexOffset;
					}


					//BWTEMP! pad it with 0xffff
					for ( int i=(int)uOutputIndexCount; i < (int)uNumIndices[uProcessBuff]; i++ )
					{
						puIndices[i] = 0xffff;
					}

				}

				DEBUG_PUT( &buff[uProcessBuff][uOutStart+uOutAlignment], apVertsOut[uProcessBuff], g_uOutputVertexSize * uNumVerts[uProcessBuff] );

				if ( uOutAlignment == 0 )
				{
					cellDmaLargePut( &buff[uProcessBuff][uOutStart], apVertsOut[uProcessBuff], g_uOutputVertexSize * uNumVerts[uProcessBuff], DMA_TAG_VERTS+uProcessBuff, uDMA_TID, uDMA_RID );
				}
				else
				{
					cellDmaUnalignedPut( &buff[uProcessBuff][uOutStart+uOutAlignment], apVertsOut[uProcessBuff], g_uOutputVertexSize * uNumVerts[uProcessBuff], DMA_TAG_VERTS+uProcessBuff, uDMA_TID, uDMA_RID );
				}

				if ( apIndicesOut[uProcessBuff] )
				{
					// force it to wait for previous PUT (fences & barriers don't seem to be enough to guarantee these are written in strict order)
					// DAMMIT WHY CANT I FENCE/BARRIER ON A DIFFERENT TAG???!?!?!?!?!?!?!
					// I want to wait on just the put, but this will wait on the get too. dammit.
					// well we were gonna wait on it in a moment anyway.
					cellDmaWaitTagStatusAll( 1<<(DMA_TAG_INDICES+uTransferBuff) );
					// To be more precise: put must wait for previous put, regardless of double-buffering, because they both write to the same buffer and unaligned puts have glitches if you don't wait (only seen it in shorts, not words, so likely only affects index buffer)
					// However get only needs to wait for the put in the same double-buffer (so it can overwrite the indices with the next strip)
					// And of course all this "waiting" could be done with fences so SPU doesn't actually wait.
					
					DEBUG_PUT( &buff[uProcessBuff][auIndicesStart[uProcessBuff]], apIndicesOut[uProcessBuff], uNumIndices[uProcessBuff]*SIZEOF_INDEX );
					cellDmaUnalignedPut( &buff[uProcessBuff][auIndicesStart[uProcessBuff]], apIndicesOut[uProcessBuff], uNumIndices[uProcessBuff]*SIZEOF_INDEX, DMA_TAG_INDICES+uProcessBuff, uDMA_TID, uDMA_RID );
				}

				uIndexOffset += uNumVerts[uProcessBuff];
			}
			bDataLastTime = bDataThisTime;
			uTransferBuff = uProcessBuff;

			g_uStartVertex += uMaxVerts;
		} while ( bDataLastTime ); // vertex batch loop


		if ( g_xJobInfo.m_eMode == SPU_SKINNING_FULL_CULL )
		{
// 			writecommandbuffernumindices; //note: morgan's env cull never did this, CPU waited on it before drawing strips (to command buffer)
		}
	
	} // task loop


	// wait for it all to finish
	cellDmaWaitTagStatusAll( 0xFFFFFFFF );


	//
	// terminate
	//

	// wait for all the spus before we tell the RSX it's data is ready
	cellSyncBarrierNotify(uBarrierAddr);
	cellSyncBarrierWait(uBarrierAddr);


	// This was the last character, tell the RSX the character data is ready
	// strictly speaking we want only the last SPU to finish to do this - this could cause a problem if RSX catches up to SPUs and SPUs have differing task sizes.
	if ( uGcmLabelAddr != 0 )
	{
		//BW apparently atomic ops don't work on an RSX label, use DMA instead
		// 			cellAtomicStore32( g_uAtomicBuf, g_xJobInfo.uGcmLabelAddr, g_xJobInfo.uGcmLabelValue );
		//low 4 bits must be the same
		u_int* pAligned = reinterpret_cast<u_int*>(&g_auAligningBuffer[(uGcmLabelAddr&0xf)>>2]);
		*pAligned = uSPU_SKINNING_RSX_FLAG;
		DEBUG_PUT( pAligned, uGcmLabelAddr, sizeof(u_int) );
		cellDmaSmallPut( pAligned, uGcmLabelAddr, sizeof(u_int), DMA_TAG_FLAG_COMPLETE, 0, 0 );
		//BW this is be happening too early - SPUs finish processing & wait for PPU. Not a big prob since that means they're done way before the RSX
		//todo - either remove RSX sync (bad if SPUs ever that slow) or let it know when it's really finished (not easy)
	}
	else
	{
		if ( bDidAnything ) spu_printf("SPU %d: error uGcmLabelAddr = NULL, SPU code was probably invoked without any data\n", g_iSPUNum);
	}

	DEBUG_PRINTF(("SPU %d: Terminating\n", g_iSPUNum));
	
	PING(false);

	cellDmaWaitTagStatusAll( 0xFFFFFFFF );
	
#if (0)
	// The SPU has an annoying habit of working fine if I do printfs during processing
	// (usually because it affects synchronisation, but on one occasion I forgot to return a value from a bool function, and enabling DEBUG_PRINTF made the compiler implicitly return true rather than false from that function)
	// So sometimes I'll want to just dump info here.
	
	spu_printf("SPU %d: Processing complete\n", g_iSPUNum);
	spu_printf("SPU %d: Total Batches %d\n", g_iSPUNum, g_uTotalBatches);
	spu_printf("SPU %d: Total Verts %d\n", g_iSPUNum, g_uTotalVerts);
#endif	
		
	sys_spu_thread_exit( EXIT_SUCCESS );
}
