//
// Asura_Dialog.cpp
//
// Created by Kevin Lea - 12/06/2001 11:11:01
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_TOOLS

#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Core/Asura_UserInput.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "Core/Asura_NameTree.h"
#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_Windows.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Gadget.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Render.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#define ASURA_DIALOG_ENABLED_COLOUR  (0xffffff)
#define ASURA_DIALOG_DISABLED_COLOUR (0x555555)
#define ASURA_DIALOG_MIXED_COLOUR (ASURA_GADGET_MIXED_COLOUR)

// G L O B A L S /////////////////////////////////////////////////////////////

extern unsigned int Asura_OverlayBrightness;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

static Asura_Handle iFontHandle;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_DialogIterator **
**----------------------*/


Asura_DialogIterator::Asura_DialogIterator(Asura_Dialog* pxFirstDialog)
{
	//Store the first dialog at the first entry in the stack
	m_iStackPos = 0;
	m_apxDialog[0] = pxFirstDialog;
	m_aiChildIndex[0] = -1;
}

void Asura_DialogIterator::Next()
{
	while(m_iStackPos >= 0)
	{
		//check to see if the current dialog has another child dialog list
		m_aiChildIndex[m_iStackPos]++;
		Asura_Dialog* pxDialog = m_apxDialog[m_iStackPos]->GetHeadOfChildDialogList(m_aiChildIndex[m_iStackPos]);
		if(pxDialog)
		{
			//Track this child list at the next entry in the stack
			m_iStackPos++;
			Asura_Assert(m_iStackPos < s_iDIALOG_ITERATOR_STACK_SIZE, "Too many nested dialogs");

			m_apxDialog[m_iStackPos] = pxDialog;
			m_aiChildIndex[m_iStackPos] = -1;
			return;
		}
		else
		{
			//No further child lists , so move to the next dialog in the linked list
			m_apxDialog[m_iStackPos] = m_apxDialog[m_iStackPos]->m_pxNextDialog;
			m_aiChildIndex[m_iStackPos] = -1;
			if(m_apxDialog[m_iStackPos])
			{
				return;
			}
			//End of linked list , so go up one level in the stack , and try again
			m_iStackPos--;
		}
	}	
}

Asura_Vector_2 Asura_DialogIterator::GetParentOffset() const
{
	Asura_Vector_2 xOffset(0.0f,0.0f);
	for(int i=0; i<m_iStackPos; i++)
	{
		xOffset += m_apxDialog[i]->GetModifiedOffset();
	}

	return xOffset;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------**
** Asura_Dialog **
**---------------*/


Asura_Dialog::Asura_Dialog( const char* szName, const Asura_Vector_2& xOffset, int iFocusPriority /* = 0 */)
{
	m_xName = szName;
	m_xOffset = xOffset;
	m_pxNextDialog = 0;
	m_pxGadgetList = 0;
	m_iControlID = -1;
	m_pxOverlay = 0;
	m_pxParentSubwindow = 0;

	m_xMaxCoord.x = m_xOffset.x + (szName ? Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT) : 0.0f);
	m_xMaxCoord.y = m_xOffset.y + (szName ? Asura_Text_Output::HeightOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT) : 1.0f);

	m_bIsEnabled = true;
	m_uHideFlags = 0;
	m_pfnIsHidden = 0;
	m_iIsHiddenUserData = 0;

	m_pxRelativeDialog = 0;
	m_xRelativeOffset = Asura_Maths::ZeroVector2;
	m_uRelativeFlags = 0;
	m_uRelativePositionID = 0;

	m_bAllowResize = false;
	m_pxUserSetSize = 0;

	m_iFocusPriority = iFocusPriority;
}

Asura_Dialog::~Asura_Dialog( void )
{
	if( m_pxGadgetList ) { delete m_pxGadgetList; }
}

int Asura_Dialog::UpdateList(const Asura_Vector_2& xOrigin, int iPriority)
{
	int iNextPriority = -1;
	if(m_pxNextDialog)
	{
		iNextPriority = m_pxNextDialog->UpdateList(xOrigin, iPriority);
	}
	if(iPriority == m_iFocusPriority)
	{
		if(m_pxParentSubwindow)
		{
			//Apply offset to origin to take into account the position of the subwindow control and the state of the scrollbars
			//RAR 24/07/2006
			Update(xOrigin + m_pxParentSubwindow->GetChildOffset());
		}
		else
		{
			Update(xOrigin);
		}
	}
	else
	{
		if(m_iFocusPriority < iPriority)
		{
			iNextPriority = Asura_Maths::Max(m_iFocusPriority, iNextPriority);
		}
	}
	return iNextPriority;
}

void Asura_Dialog::Update( const Asura_Vector_2& xOrigin )
{
	if( !IsHidden() )
	{
		if( m_pxGadgetList ) { m_pxGadgetList->Handle( xOrigin ); }

		// bring up the help overlay (if there is one) if the user keeps the cursor still over something
		if(!Asura_Overlay_System::IsAnyHelpText() && Asura_Cursor::IsStatic() )
		{
			Asura_Vector_2 xPos = xOrigin + m_xOffset;
			//Mouse over the Asura_Dialog name text?
			bool bMouseOver = Asura_Cursor::Inside( xPos, m_xMaxCoord.x - m_xOffset.x, 1.0f );
			//check if mouse is over any of the individual gadgets
			Asura_Gadget* pxGadget = m_pxGadgetList;
			while(pxGadget && !bMouseOver)
			{
				bMouseOver = pxGadget->IsMouseOver();
				pxGadget = pxGadget->m_pxNextGadget;
			}

			if(!bMouseOver && ShouldCheckDialogExtentsForIsPointInDialog())
			{
				bMouseOver = Asura_Cursor::Inside(xPos,m_xMaxCoord.x - m_xOffset.x,m_xMaxCoord.y - m_xOffset.y);
			}


			if(bMouseOver)
			{
				ShowHelpText();
			}
		}
	}
}

void Asura_Dialog::PerFrameUpdate()
{
	if(m_uRelativeFlags) UpdateRelativePosition();

	if(m_xIsEnabled.IsValid())
	{
		SetEnabled(m_xIsEnabled.Execute());
	}

	if(m_pfnIsHidden)
	{
		SetHideFlag(ASURA_DIALOG_HIDEFLAG_CALLBACK,m_pfnIsHidden(m_iIsHiddenUserData));
	}
}


void Asura_Dialog::AddGadget( Asura_Gadget* pxNewGadget )
{
	if( pxNewGadget )
	{
		if( m_pxGadgetList ) 
			pxNewGadget->m_pxNextGadget = m_pxGadgetList; 
			
		pxNewGadget->SetParentDialog(this);

		m_pxGadgetList = pxNewGadget;
	}
}

void Asura_Dialog::SetEnabled( bool bEnabled )
{
	if( m_bIsEnabled != bEnabled )
	{
		m_bIsEnabled = bEnabled;

		// Pass on to sub components.
		if( m_pxGadgetList )
		{
			m_pxGadgetList->SetEnabledFlag( bEnabled );
		}
	}
#if _DEBUG
	else if( !m_bIsEnabled )
	{
		Asura_Gadget* pxNext = m_pxGadgetList;
		while( pxNext )
		{
			Asura_Assert( pxNext->GetEnabledFlag() == m_bIsEnabled, "" );
			pxNext = pxNext->m_pxNextGadget;
		}
	}
#endif
}

unsigned Asura_Dialog::GetColour( void ) const
{
	return( Asura_OverlayBrightness | (m_bIsEnabled ? ASURA_DIALOG_ENABLED_COLOUR : ASURA_DIALOG_DISABLED_COLOUR) );
}

int Asura_Dialog::RenderList(const Asura_Vector_2& xOrigin, int iPriority)
{
	int iNextPriority = ASURA_DIALOG_INVALIDPRIORITY;
	if(m_pxNextDialog)
	{
		iNextPriority = m_pxNextDialog->RenderList(xOrigin, iPriority);
	}
	if(iPriority == m_iFocusPriority)
	{
		//enable/disable clipping as required
		if(m_pxParentSubwindow && !CanRenderOutsideOfParentSubwindow())
		{
			m_pxParentSubwindow->EnableClipping();
		}
		else
		{
			Asura_Subwindow_Dialog::DisableClipping();
		}

		if(m_pxParentSubwindow)
		{
			//Apply offset to origin to take into account the position of the subwindow control and the state of the scrollbars
			//RAR 24/07/2006
			if(m_pxParentSubwindow->IsDialogInVisibleArea(this))
			{
				Render(xOrigin + m_pxParentSubwindow->GetChildOffset());
			}

		}
		else
		{
			Render(xOrigin);
		}
	}
	else
	{
		if(m_iFocusPriority > iPriority)
		{
			if(iNextPriority == ASURA_DIALOG_INVALIDPRIORITY)
				iNextPriority = m_iFocusPriority;
			else
				iNextPriority = Asura_Maths::Min(m_iFocusPriority, iNextPriority);
		}
	}

	return iNextPriority;
}

unsigned Asura_StringDisplay_Dialog::GetColour(void) const
{
	if(m_uColour == 0x0)
	{
		return Asura_Dialog::GetColour();
	}
	else
	{
		return m_uColour;
	}
}

void Asura_Dialog::Render( const Asura_Vector_2& xOrigin )
{
	if( !IsHidden() )
	{
		if( m_pxGadgetList )
		{
			m_pxGadgetList->RenderList( xOrigin );
		}

		if( !m_xName.IsBlank() )
		{
			Asura_Vector_2 pos = xOrigin + m_xOffset + GetTextOffset();
			Asura_Text_Output::Print(pos.x,pos.y,GetColour(),ASURA_TEXT_FORMAT_SMALLFONT,"%s",m_xName.GetString());
		}
	}
}

void Asura_Dialog::UpdateLabel(const char* szText)
{
	// Has the label changed?
	if(m_xName != szText)
	{
		float	fOldWidth;
		float	fOldHeight;
		float	fNewWidth;
		float	fNewHeight;

		//get the change in width AND height
		if(m_xName.IsBlank())
		{
			fOldWidth	= 0.0f;
			fOldHeight	= 0.0f;
		}
		else
		{
			const char*	szOldText	= m_xName.GetString();

			fOldWidth	= Asura_Text_Output::WidthOfString(szOldText, ASURA_TEXT_FORMAT_SMALLFONT);
			fOldHeight	= Asura_Text_Output::HeightOfString(szOldText, ASURA_TEXT_FORMAT_SMALLFONT);
		}

		if(szText == NULL)
		{
			fNewWidth	= 0.0f;
			fNewHeight	= 0.0f;
		}
		else
		{
			fNewWidth	= Asura_Text_Output::WidthOfString(szText, ASURA_TEXT_FORMAT_SMALLFONT);
			fNewHeight	= Asura_Text_Output::HeightOfString(szText, ASURA_TEXT_FORMAT_SMALLFONT);
		}

		// Update the label...
		m_xName = szText;

		//adjust the maxcoord
		m_xMaxCoord.x	+= fNewWidth - fOldWidth;
		m_xMaxCoord.y	+= fNewHeight - fOldHeight;
	}
}

bool Asura_Dialog::IsPointInDialog(const Asura_Vector_2& xOrigin,const Asura_Vector_2& xPos) const
{
	if(IsHidden()) return false;

	//check point against each gadget in the dialog
	const Asura_Gadget* pxGadget = m_pxGadgetList;

	//Convert position to be relative to the overlay
	Asura_Vector_2 xRelativePos = xPos - xOrigin;
	if(m_pxParentSubwindow)
	{
		//If a member of a subwindow dialog ,then the point must be in that as well
		if(!CanRenderOutsideOfParentSubwindow())
		{
			if(!m_pxParentSubwindow->IsPointWithinChildRect(xPos.x,xPos.y))
			{
				return false;
			}
		}

		xRelativePos -= m_pxParentSubwindow->GetChildOffset();
	}

	if(ShouldCheckDialogExtentsForIsPointInDialog())
	{
		if(xRelativePos.x >= m_xOffset.x && xRelativePos.x <= m_xMaxCoord.x &&
			xRelativePos.y >= m_xOffset.y && xRelativePos.y <= m_xMaxCoord.y)
		{
			return true;
		}

	}

	while(pxGadget)
	{
		if(pxGadget->IsPointInGadget(xRelativePos)) return true;
		pxGadget = pxGadget->m_pxNextGadget;
	}
	return false;
}

bool Asura_Dialog::CanTakeTabFocus() const
{
	if(!m_bIsEnabled) return false;

	//search through the gadgets for any that can take focus
	const Asura_Gadget* pxGadget = m_pxGadgetList;

	while(pxGadget)
	{
		if(pxGadget->CanTakeTabFocus()) return true;
		pxGadget = pxGadget->m_pxNextGadget;
	}
	return false;
}

void Asura_Dialog::TakeTabFocus()
{
	Asura_Gadget* pxGadget = m_pxGadgetList;

	while(pxGadget)
	{
		if(pxGadget->CanTakeTabFocus())
		{
			pxGadget->TakeTabFocus();
			return;
		}
		pxGadget = pxGadget->m_pxNextGadget;
	}
}

void Asura_Dialog::MoveDialog(const Asura_Vector_2& xNewOffset)
{
	Asura_Vector_2 xDiff = xNewOffset - m_xOffset;
	//update the offset and maxcoords
	m_xOffset = xNewOffset;
	m_xMaxCoord += xDiff;

	//move all the gadgets
	Asura_Gadget* pxGadget = m_pxGadgetList;
	while(pxGadget)
	{
		pxGadget->MoveGadgetRelative(xDiff);
		pxGadget = pxGadget->m_pxNextGadget;
	}	

}

void Asura_Dialog::SetRelativePosition(const Asura_Dialog* pxDialog,const Asura_Vector_2& xRelOffset,bool bRelBottom,bool bRelRight)
{
	u_int uFlags = ASURA_DIALOG_RELATIVE_X_DIALOG|ASURA_DIALOG_RELATIVE_Y_DIALOG;
	if(bRelBottom) uFlags |= ASURA_DIALOG_RELATIVE_Y_BOTTOM;
	if(bRelRight) uFlags |= ASURA_DIALOG_RELATIVE_X_RIGHT;

	SetRelativePositionFlags(pxDialog, xRelOffset, uFlags);
}

void Asura_Dialog::SetRelativePosition(const Asura_Dialog* pxDialog,bool bRelBottom,bool bRelRight)
{
	u_int uFlags = ASURA_DIALOG_RELATIVE_X_DIALOG|ASURA_DIALOG_RELATIVE_Y_DIALOG;
	if(bRelBottom) uFlags |= ASURA_DIALOG_RELATIVE_Y_BOTTOM;
	if(bRelRight) uFlags |= ASURA_DIALOG_RELATIVE_X_RIGHT;

	Asura_Vector_2 xRelOffset(0.0f,0.0f);
	if(pxDialog)
	{
		//set relative offset based on the current relative offset
		Asura_Vector_2 xCorner = pxDialog->m_xOffset;
		if(bRelBottom) xCorner.y = pxDialog->GetMaxCoords().y;
		if(bRelRight) xCorner.x = pxDialog->GetMaxCoords().x;
		xRelOffset = m_xOffset - xCorner;
	}

	SetRelativePositionFlags(pxDialog, xRelOffset, uFlags);


}

void Asura_Dialog::SetRelativePosition(const Asura_Dialog* pxDialog,Asura_Hash_ID uPositionID)
{
	m_pxRelativeDialog = pxDialog;
	m_uRelativeFlags = ASURA_DIALOG_RELATIVE_X_DIALOG|ASURA_DIALOG_RELATIVE_Y_DIALOG;
	m_uRelativePositionID = uPositionID;
	if(m_pxRelativeDialog)
	{
		//set relative offset based on the current relative offset
		Asura_Vector_2 xCorner = pxDialog->GetPositionFromID(uPositionID);
		m_xRelativeOffset = m_xOffset - xCorner;
	}
}

void Asura_Dialog::SetOverlayRelativePosition(const Asura_Vector_2& xRelOffset,bool bRelBottom,bool bRelRight)
{
	u_int uFlags = ASURA_DIALOG_RELATIVE_X_OVERLAY|ASURA_DIALOG_RELATIVE_Y_OVERLAY;
	if(bRelBottom) uFlags |= ASURA_DIALOG_RELATIVE_Y_BOTTOM;
	if(bRelRight) uFlags |= ASURA_DIALOG_RELATIVE_X_RIGHT;

	SetRelativePositionFlags(NULL, xRelOffset, uFlags);

}

void Asura_Dialog::SetRelativePositionFlags(const Asura_Dialog* pxDialog, const Asura_Vector_2& xRelOffset , u_int uRelativeFlags)
{
	//complain about invalid flags
	Asura_Assert(!((uRelativeFlags & ASURA_DIALOG_RELATIVE_X_DIALOG) && (uRelativeFlags & ASURA_DIALOG_RELATIVE_X_OVERLAY)), "X coordinate shouldn't be relative to both a dialog and the overlay");
	Asura_Assert(!((uRelativeFlags & ASURA_DIALOG_RELATIVE_Y_DIALOG) && (uRelativeFlags & ASURA_DIALOG_RELATIVE_Y_OVERLAY)), "Y coordinate shouldn't be relative to both a dialog and the overlay");
	Asura_Assert(pxDialog || !(uRelativeFlags & (ASURA_DIALOG_RELATIVE_X_DIALOG|ASURA_DIALOG_RELATIVE_Y_DIALOG)), "Need a dialog to be relative to");

	m_pxRelativeDialog = pxDialog;
	m_uRelativeFlags = uRelativeFlags;
	m_uRelativePositionID = 0;
	m_xRelativeOffset = xRelOffset;
}

Asura_Vector_2 Asura_Dialog::GetPositionFromID(Asura_Hash_ID uPositionID) const
{
	Asura_Assert(0,"Unrecognised ID for relative position");
	return m_xOffset;
}

void Asura_Dialog::UpdateRelativePosition()
{
	//determine what offset should be
	Asura_Vector_2 xOffset = m_xOffset;


	if(m_uRelativeFlags & ASURA_DIALOG_RELATIVE_X_OVERLAY)
	{
		xOffset.x = (m_uRelativeFlags & ASURA_DIALOG_RELATIVE_X_RIGHT) ? m_pxOverlay->m_fWidth : 0.0f;
		xOffset.x += m_xRelativeOffset.x;
	}
	else if(m_pxRelativeDialog && (m_uRelativeFlags & ASURA_DIALOG_RELATIVE_X_DIALOG))
	{
		if(m_pxRelativeDialog->IsHidden())
		{
			xOffset.x = m_pxRelativeDialog->m_xOffset.x;
		}
		else
		{
			if(m_uRelativePositionID)
			{
				xOffset.x = m_pxRelativeDialog->GetPositionFromID(m_uRelativePositionID).x;
			}
			else
			{
				xOffset.x = m_pxRelativeDialog->m_xOffset.x;
				if(m_uRelativeFlags & ASURA_DIALOG_RELATIVE_X_RIGHT)
				{
					xOffset.x = m_pxRelativeDialog->GetMaxCoords().x;
				}
			}
			xOffset.x += m_xRelativeOffset.x;
		}
	}

	if(m_uRelativeFlags & ASURA_DIALOG_RELATIVE_Y_OVERLAY)
	{
		xOffset.y = (m_uRelativeFlags & ASURA_DIALOG_RELATIVE_Y_BOTTOM) ? m_pxOverlay->m_fHeight : 0.0f;
		xOffset.y += m_xRelativeOffset.y;
	}
	else if(m_pxRelativeDialog && (m_uRelativeFlags & ASURA_DIALOG_RELATIVE_Y_DIALOG))
	{
		if(m_pxRelativeDialog->IsHidden())
		{
			xOffset.y = m_pxRelativeDialog->m_xOffset.y;
		}
		else
		{
			if(m_uRelativePositionID)
			{
				xOffset.y = m_pxRelativeDialog->GetPositionFromID(m_uRelativePositionID).y;
			}
			else
			{
				xOffset.y = m_pxRelativeDialog->m_xOffset.y;
				if(m_uRelativeFlags & ASURA_DIALOG_RELATIVE_Y_BOTTOM)
				{
					xOffset.y = m_pxRelativeDialog->GetMaxCoords().y;
				}
			}
			xOffset.y += m_xRelativeOffset.y;
		}
	}



	if(m_xOffset != xOffset)
	{
		//need to move to the new position
		MoveDialog(xOffset);

		//resize overlay
		if(m_pxOverlay) m_pxOverlay->FlagForResize();
	}
}


void Asura_Dialog::SetIsHiddenFunction(bool(*pfnIsHidden)(int), int iUserData)
{
	m_pfnIsHidden = pfnIsHidden;
	m_iIsHiddenUserData = iUserData;
}

void Asura_Dialog::SetHideFlag(u_int uFlag,bool bSet)
{
	bool bWasHidden = IsHidden();
	if(bSet) m_uHideFlags |= uFlag;
	else m_uHideFlags &=~ uFlag;

	if(bWasHidden != IsHidden())
	{
		//allow derived class to cope with change in hidden state. RAR 23/04/2007
		OnHiddenStateChange();
	}
}

void Asura_Dialog::DeleteSingleGadget(Asura_Gadget* pxGadget)
{
	Asura_Gadget** ppxNextGadget = &m_pxGadgetList;

	//search for the gadget in the list
	while(*ppxNextGadget && *ppxNextGadget != pxGadget)
	{
		ppxNextGadget =	&(*ppxNextGadget)->m_pxNextGadget;
	}

	if(!*ppxNextGadget)
	{
		Asura_Assert(0,"Failed to find the gadget being deleted");
		return;
	}

	//update the linked list to remove the deleted gadget
	*ppxNextGadget = pxGadget->m_pxNextGadget;
	pxGadget->m_pxNextGadget = 0;

	delete pxGadget;
}

void Asura_Dialog::ShowHelpText()
{
	if(!m_xHelpText.IsBlank())
	{
		Asura_Overlay_System::SetHelpText(m_xHelpText.GetString());
	}
}

void Asura_Dialog::UpdateSize()
{
	if(m_pxParentSubwindow) m_pxParentSubwindow->UpdateSize(this);
}


Asura_Vector_2 Asura_Dialog::GetModifiedOffset() const
{
	if(m_pxParentSubwindow)
		return m_xOffset + m_pxParentSubwindow->GetChildOffset();
	else
		return m_xOffset;
}

void Asura_Dialog::SetRadioButtonGroup(int iGroupID,bool bAtLeastOnePressed)
{
	//apply to all the gadgets
	for(Asura_Gadget* pxGadget = m_pxGadgetList;pxGadget;pxGadget = pxGadget->m_pxNextGadget)
	{
		pxGadget->SetRadioButtonGroup(iGroupID,bAtLeastOnePressed);
	}
}

void Asura_Dialog::RadioButtonPressed(const class Asura_Gadget* pxPressedGadget)
{
	for(Asura_Gadget* pxGadget = m_pxGadgetList;pxGadget;pxGadget = pxGadget->m_pxNextGadget)
	{
		pxGadget->RadioButtonPressed(pxPressedGadget);
	}
}

void Asura_Dialog::SetAllowResize(bool bSet, Asura_Vector_2* pxUserSetSize)
{
	m_bAllowResize = bSet; 
	m_pxUserSetSize = pxUserSetSize;
	Asura_Assert(!bSet || GetAllowedResizeFlags(),"This type of dialog doesn't support resizing");
}



/////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Asura_Numerical_Dialog **
**------------------------*/
Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* name, const Asura_Vector_2& offset, float* value, float min, float max, float delta, const char* printFormat, bool* flag/* = 0*/) : Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxDecGadget = new Asura_IncrementorFloat_Gadget(pos,value,min,max,-delta,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	newGadget  = m_pxTextGadget = new Asura_Float_Gadget(pos,value,min,max,printFormat,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	newGadget  = m_pxIncGadget = new Asura_IncrementorFloat_Gadget(pos,value,min,max,+delta,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	m_xTextOffset.x = pos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = pos.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;


}

Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* name, const Asura_Vector_2& offset, int* value, int min, int max, int delta, const char* printFormat, bool* flag/* = 0*/) : Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxDecGadget = new Asura_IncrementorInt_Gadget(pos,value,min,max,-delta,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	newGadget  = m_pxTextGadget = new Asura_Integer_Gadget(pos,value,min,max,printFormat,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	newGadget  = m_pxIncGadget = new Asura_IncrementorInt_Gadget(pos,value,min,max,+delta,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	m_xTextOffset.x = pos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = pos.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

}

Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* name, const Asura_Vector_2& offset, u_int* value, u_int min, u_int max, int delta, const char* printFormat, bool* flag/* = 0*/)
: Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxDecGadget = new Asura_IncrementorUInt_Gadget(pos,value,min,max,-delta,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	newGadget  = m_pxTextGadget = new Asura_UInt_Gadget(pos,value,min,max,printFormat,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	newGadget  = m_pxIncGadget = new Asura_IncrementorUInt_Gadget(pos,value,min,max,+delta,flag);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	m_xTextOffset.x = pos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = pos.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;
}

Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* szName,
											   const Asura_Vector_2& xOffset,
											   u_char* pucValue,
											   u_char ucMin,
											   u_char ucMax,
											   int iDelta,
											   const char* szPrintFormat,
											   bool* pbFlag /* = 0 */)
: Asura_Dialog(szName, xOffset)
{
	Asura_Vector_2 xPos = xOffset;

	m_pxDecGadget = new Asura_IncrementorUChar_Gadget(xPos,pucValue,ucMin,ucMax,-iDelta,pbFlag);
	AddGadget(m_pxDecGadget);

	xPos.x += m_pxDecGadget->m_fWidth*2;

	m_pxTextGadget = new Asura_UChar_Gadget(xPos,pucValue,ucMin,ucMax,szPrintFormat,pbFlag);
	AddGadget(m_pxTextGadget);

	xPos.x += m_pxTextGadget->m_fWidth;

	m_pxIncGadget = new Asura_IncrementorUChar_Gadget(xPos,pucValue,ucMin,ucMax,+iDelta,pbFlag);
	AddGadget(m_pxIncGadget);

	xPos.x += m_pxIncGadget->m_fWidth*2;

	m_xTextOffset.x = xPos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = xPos.x + Asura_Text_Output::WidthOfString(szName, ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;
}

Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* name, const Asura_Vector_2& offset, Asura_Dialog_SharedValue<float>* pxSharedValue, float min, float max, float delta, const char* printFormat) : Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxDecGadget = new Asura_IncrementorSharedFloat_Gadget(pos,pxSharedValue,min,max,-delta);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	newGadget  = m_pxTextGadget = new Asura_Float_Gadget(pos,pxSharedValue,min,max,printFormat,0);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	newGadget  = m_pxIncGadget = new Asura_IncrementorSharedFloat_Gadget(pos,pxSharedValue,min,max,+delta);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	m_xTextOffset.x = pos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = pos.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;


}

Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* name, const Asura_Vector_2& offset, Asura_Dialog_SharedValue<int>* pxSharedValue, int min, int max, int delta, const char* printFormat) : Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxDecGadget = new Asura_IncrementorSharedInt_Gadget(pos,pxSharedValue,min,max,-delta);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	newGadget  = m_pxTextGadget = new Asura_Integer_Gadget(pos,pxSharedValue,min,max,printFormat,0);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	newGadget  = m_pxIncGadget = new Asura_IncrementorSharedInt_Gadget(pos,pxSharedValue,min,max,+delta);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	m_xTextOffset.x = pos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = pos.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

}

Asura_Numerical_Dialog::Asura_Numerical_Dialog(const char* name, const Asura_Vector_2& offset, Asura_Dialog_SharedValue<u_int>* pxSharedValue, u_int min, u_int max, int delta, const char* printFormat)
: Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxDecGadget = new Asura_IncrementorSharedUInt_Gadget(pos,pxSharedValue,min,max,-delta);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	newGadget  = m_pxTextGadget = new Asura_UInt_Gadget(pos,pxSharedValue,min,max,printFormat);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	newGadget  = m_pxIncGadget = new Asura_IncrementorSharedUInt_Gadget(pos,pxSharedValue,min,max,+delta);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth*2;

	m_xTextOffset.x = pos.x-m_xOffset.x;
	m_xTextOffset.y = 0;

	m_xMaxCoord.x = pos.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;
}



void Asura_Numerical_Dialog::SetReadOnly(bool bReadOnly)
{
	m_pxTextGadget->SetReadOnly(bReadOnly);
	m_pxIncGadget->SetHidden(bReadOnly);
	m_pxDecGadget->SetHidden(bReadOnly);

}

void Asura_Numerical_Dialog::SetDefaultFocus(bool bSet)
{
	m_pxTextGadget->SetDefaultFocus(bSet);
}

void Asura_Numerical_Dialog::SetBoxLength(int iNumChars)
{
	m_pxTextGadget->SetBoxLength(iNumChars);
	m_pxIncGadget->m_xOffset.x = m_pxTextGadget->m_xOffset.x + m_pxTextGadget->m_fWidth + 0.5f;

	m_xTextOffset.x = m_pxTextGadget->m_xOffset.x + m_pxTextGadget->m_fWidth + (m_pxIncGadget->m_fWidth*2.0f)-m_xOffset.x;

	m_xMaxCoord.x = m_xOffset.x + m_xTextOffset.x + Asura_Text_Output::WidthOfString(m_xName.GetString(),ASURA_TEXT_FORMAT_SMALLFONT);
}

// Set the float Delta (step value) for the increment and decrement gadgets.
// If the gadgets don't support floats an assert will trigger
void Asura_Numerical_Dialog::SetDelta(float fDelta)
{
	static_cast<Asura_Incrementor_Gadget*>(m_pxIncGadget)->SetDelta(+fDelta);
	static_cast<Asura_Incrementor_Gadget*>(m_pxDecGadget)->SetDelta(-fDelta);
}

// Set the int Delta (step value) for the increment and decrement gadgets.
// If the gadgets don't support ints an assert will trigger
void Asura_Numerical_Dialog::SetDelta(int iDelta)
{
	static_cast<Asura_Incrementor_Gadget*>(m_pxIncGadget)->SetDelta(+iDelta);
	static_cast<Asura_Incrementor_Gadget*>(m_pxDecGadget)->SetDelta(-iDelta);
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------**
** Asura_ScaledNumerical_Dialog **
**-----------------------------*/

Asura_ScaledNumerical_Dialog::Asura_ScaledNumerical_Dialog(const char* szName, const Asura_Vector_2& xOffset, float fScale,float* pfValue, float fScaledMin, float fScaledMax,float fScaledDelta, char* szPrintFormat, bool* pbFlag /*= 0*/)
:Asura_Dialog(szName,xOffset)
{
	Asura_Assert(fScale != 0.0f,"");

	//initialise stuff
	m_fScale = fScale;
	m_pfValue = pfValue;
	m_fValueBackup = *m_pfValue;
	m_pbFlag = pbFlag;
	m_fScaledValue = *m_pfValue * m_fScale;
	m_bChanged = false;
	m_fScaledMin = m_fScaledMinBackup = fScaledMin;
	m_fScaledMax = m_fScaledMaxBackup = fScaledMax;

	//create the gadgets
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxGadget;
	pxGadget = m_pxDecGadget  = new Asura_IncrementorFloat_Gadget(xPos,&m_fScaledValue,fScaledMin,fScaledMax,-fScaledDelta,&m_bChanged);
	AddGadget(pxGadget);

	xPos.x += pxGadget->m_fWidth*2;

	pxGadget = m_pxValGadget  = new Asura_Float_Gadget(xPos,&m_fScaledValue,fScaledMin,fScaledMax,szPrintFormat,&m_bChanged);
	AddGadget(pxGadget );

	xPos.x += pxGadget->m_fWidth;

	pxGadget = m_pxIncGadget  = new Asura_IncrementorFloat_Gadget(xPos,&m_fScaledValue,fScaledMin,fScaledMax,+fScaledDelta,&m_bChanged);
	AddGadget(pxGadget);

	xPos.x += pxGadget->m_fWidth*2;

	m_xOffset.x = xPos.x;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

}


void Asura_ScaledNumerical_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	//if the scaled value has changed , update the unscaled value , and set the flag passed (if one was)
	if(m_bChanged)
	{
		m_bChanged = false;

		// If SetScale() has been used, the value might be out of our new range
		if(m_fScaledValue<m_fScaledMin) m_fScaledValue=m_fScaledMin;
		if(m_fScaledValue>m_fScaledMax) m_fScaledValue=m_fScaledMax;

		//the scaled value has changed , so update the unscaled value
		*m_pfValue = m_fScaledValue / m_fScale;
		m_fValueBackup = *m_pfValue;
		//and update the modified flag passed to this gadget ,if we had one
		if(m_pbFlag) *m_pbFlag = true;		
	}
	else if(m_fValueBackup!=*m_pfValue)
	{
		//The unscaled value has been changed by some outside force...
		//Update the scaled value to match
		m_fValueBackup = *m_pfValue;
		m_fScaledValue = *m_pfValue * m_fScale;

	}
}

void Asura_ScaledNumerical_Dialog::SetScale( const float fScale,  const float fScaledMin,  const float fScaledMax )
{
	m_fScale = fScale;
	m_fValueBackup = *m_pfValue;
	m_fScaledValue = *m_pfValue * m_fScale;
	m_fScaledMin = fScaledMin;
	m_fScaledMax = fScaledMax;

	// JAT 05/04/07 need to watch out for if the new max makes the box longer then the increment & decrement 
	// gadgets will be in the wrong place
	Asura_Assert( m_pxDecGadget, "" );
	Asura_Assert( m_pxValGadget, "" );
	Asura_Assert( m_pxIncGadget, "" );
	float fInitialValWidth = m_pxValGadget->m_fWidth;
	m_pxDecGadget->SetMinMax( fScaledMin, fScaledMax );
	m_pxValGadget->SetMinMax( fScaledMin, fScaledMax );
	m_pxIncGadget->SetMinMax( fScaledMin, fScaledMax );

	//Update the incrementor and text position in case the value gadget was resized. RAR 11/04/2007
	float fDelta = m_pxValGadget->m_fWidth - fInitialValWidth;
	m_pxIncGadget->m_xOffset.x += fDelta;
	m_xOffset.x += fDelta;
	m_xMaxCoord.x += fDelta;
	m_xRelativeOffset.x += fDelta;
}

void Asura_ScaledNumerical_Dialog::SetDelta( const float fDelta )
{
	Asura_Assert( m_pxDecGadget, "" );
	Asura_Assert( m_pxIncGadget, "" );
	m_pxDecGadget->SetDelta( -fDelta );
	m_pxIncGadget->SetDelta( +fDelta );
}

void Asura_ScaledNumerical_Dialog::SetPrintFormat( const char* const szPrintFormat )
{
	float fInitialValWidth = m_pxValGadget->m_fWidth;
	m_pxValGadget->SetPrintFormat( szPrintFormat );
	
	//Update the incrementor and text position in case the value gadget was resized. RAR 11/04/2007
	float fDelta = m_pxValGadget->m_fWidth - fInitialValWidth;
	m_pxIncGadget->m_xOffset.x += fDelta;
	m_xOffset.x += fDelta;
	m_xMaxCoord.x += fDelta;
	m_xRelativeOffset.x += fDelta;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------**
** Asura_ColourNumerical_Dialog **
**------------------------------*/

Asura_ColourNumerical_Dialog::Asura_ColourNumerical_Dialog(const char* pszName, const Asura_Vector_2& xOffset, float* pfValue, char* pszPrintFormat, bool* pbFlag/* = 0*/) 
: Asura_Dialog(pszName,xOffset)
{
	//initialise stuff
	m_pfValue = pfValue;
	m_fValueBackup = *pfValue;
	m_pbFlag = pbFlag;
	m_iColour = Asura_Maths::FloatToInt_Near(*m_pfValue*255);
	m_bChanged = false;

	//create the gadgets
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxGadget;
	pxGadget  = new Asura_IncrementorInt_Gadget(xPos,&m_iColour,0,255,-1,&m_bChanged);
	AddGadget(pxGadget);

	xPos.x += pxGadget->m_fWidth*2;

	pxGadget  = new Asura_Integer_Gadget(xPos,&m_iColour,0,255,pszPrintFormat,&m_bChanged);
	AddGadget(pxGadget);

	xPos.x += pxGadget->m_fWidth;

	pxGadget  = new Asura_IncrementorInt_Gadget(xPos,&m_iColour,0,255,1,&m_bChanged);
	AddGadget(pxGadget);

	xPos.x += pxGadget->m_fWidth*2;

	m_xOffset.x = xPos.x;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(pszName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

}

void Asura_ColourNumerical_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();
	//if the integer value has changed , update the float value , and set the flag passed (if one was)
	if(m_bChanged)
	{
		m_bChanged = false;
		m_fValueBackup = *m_pfValue = m_iColour/255.0f;
		//and update the modified flag passed to this gadget ,if we had one
		if(m_pbFlag) *m_pbFlag = true;		
	}
	else if(m_fValueBackup != *m_pfValue)
	{
		//The floating point version of the colour had been changed by somethnig outside
		//Update the integer value to match
		m_fValueBackup = *m_pfValue;
		m_iColour = Asura_Maths::FloatToInt_Near(*m_pfValue*255);

	}

}

void Asura_ColourNumerical_Dialog::UpdateColour(float fNewValue)
{
	m_iColour = Asura_Maths::FloatToInt_Near(*m_pfValue*255);
	m_bChanged = true;
}
/////////////////////////////////////////////////////////////////////////////////////////////////

Asura_NumberSlider_Dialog::Asura_NumberSlider_Dialog(const Asura_Vector_2& offset, float* value, float* minPtr, float* maxPtr, float slideWidth,char* printFormat, bool* flag/* = 0*/) : Asura_Dialog(0,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget  = m_pxSliderGadget= new Asura_NumberSlider_Gadget(offset,value,minPtr,maxPtr,slideWidth,printFormat,flag);
	AddGadget(newGadget);

	m_xMaxCoord.x = m_xOffset.x + slideWidth;
	m_xMaxCoord.y = m_xOffset.y + 1.0f;


}

void Asura_NumberSlider_Dialog::SetPrintFormat(const char* pszPrintFormat)
{
	m_pxSliderGadget->SetPrintFormat(pszPrintFormat);
}

/////////////////////////////////////////////////////////////////////////////////////////////////

Asura_String_Dialog::Asura_String_Dialog(const char* name, const Asura_Vector_2& offset, Asura_String* value, int maxLength, bool* flag /*= 0*/, bool defaultFocus /*=false*/) 
:Asura_Dialog(name,offset)
{
	Asura_Vector_2 pos = offset;

	Asura_Gadget* newGadget;
	newGadget = m_pxStringGadget = new Asura_String_Gadget(pos,value,maxLength,flag,defaultFocus);
	AddGadget(newGadget);

	pos.x += newGadget->m_fWidth;

	m_xOffset.x =  pos.x + 1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(name,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;
}

void Asura_String_Dialog::SetMaxStringLength(int iMaxLength)
{
	m_pxStringGadget->SetMaxStringLength(iMaxLength);	
}

void Asura_String_Dialog::SetDefaultFocus(bool bSet)
{
	m_pxStringGadget->SetDefaultFocus(bSet);
}

void Asura_String_Dialog::SetReadOnly(bool bSet)
{
	m_pxStringGadget->SetReadOnly(bSet);
}

void Asura_String_Dialog::SetTextChangeCallback (const Asura_Callback_Base<void>& xCallback)
{
	m_pxStringGadget->SetTextChangeCallback (xCallback);
}

void Asura_String_Dialog::SetUpdateImmediatelyWhileTyping(bool bSet)
{
	m_pxStringGadget->SetUpdateImmediatelyWhileTyping(bSet);
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Tickbox_Dialog **
**----------------------*/

Asura_Tickbox_Dialog::Asura_Tickbox_Dialog(const char* szName, const Asura_Vector_2& xOffset, bool* pbValue, bool* pbFlag /*=0*/)
: Asura_Dialog(szName,xOffset)
{
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxNewGadget;
	pxNewGadget  = m_pxTickbox = new Asura_BoolTickbox_Gadget(xPos,pbValue,pbFlag);
	AddGadget(pxNewGadget);

	m_xOffset.x += pxNewGadget->m_fWidth+1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	// Set the correct width of the gadget, so that clicking is handled properly.
    if ( !szName || ( strcmp( szName, "" ) == 0 ) )
    {
        // If we have no name then we only want the box to be clickable.
    }
    else
    {
        // If we have a name then expand the width of the gadget, so that clicking on the text will activate it.
	    pxNewGadget->m_fWidth = m_xMaxCoord.x-pxNewGadget->m_xOffset.x;
    }
}

void Asura_Tickbox_Dialog::UpdateLabel(const char* szText)
{
	Asura_Dialog::UpdateLabel(szText);

	// Update the width of the gadget.
    if ( !m_pxGadgetList ) { return; }

    if ( !szText || ( strcmp( szText, "" ) == 0 ) )
    {
        // If we have no name then we only want the box to be clickable.
        m_pxGadgetList->m_fWidth = Asura_BaseTickbox_Gadget::WIDTH;
    }
    else
    {
        // If we have a name then expand the width of the gadget, so that clicking on the text will activate it.
        m_pxGadgetList->m_fWidth = m_xMaxCoord.x-m_pxGadgetList->m_xOffset.x;
    }
}

void Asura_Tickbox_Dialog::SetValue(bool* const pbValue)
{
	if(m_pxTickbox)
	{
		m_pxTickbox->SetValue(pbValue);
	}
}

void Asura_Tickbox_Dialog::SetClickedFunction(const Asura_Callback_Base<void>& xCallback)
{
    m_pxTickbox->SetClickedFunction(xCallback);
}


void Asura_Tickbox_Dialog::SetUseSharedDialogColour(bool bShared)
{
	m_pxTickbox->SetUseSharedGadgetColour(bShared);
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Asura_FlagTickbox_Dialog **
**--------------------------*/

Asura_FlagTickbox_Dialog::Asura_FlagTickbox_Dialog(const char* name, const Asura_Vector_2& offset, int* flagsField , int flagBit, bool* modifiedFlag /*= 0*/)
: Asura_Dialog(name,offset)
{
	CreateGadgets(name,offset,flagsField ,flagBit,modifiedFlag);
}

Asura_FlagTickbox_Dialog::Asura_FlagTickbox_Dialog(const char* name, const Asura_Vector_2& offset, unsigned int* flagsField , int flagBit, bool* modifiedFlag /*= 0*/)
: Asura_Dialog(name,offset)
{
	CreateGadgets(name,offset,(int*)flagsField ,flagBit,modifiedFlag);
}

Asura_FlagTickbox_Dialog::Asura_FlagTickbox_Dialog(const char* szName, const Asura_Vector_2& xOffset, char* pcFlagsField , char cFlagBit, bool* pbFlag /* = 0 */)
: Asura_Dialog(szName,xOffset)
{
	CreateGadgets(szName, xOffset, pcFlagsField, cFlagBit, pbFlag);
}

Asura_FlagTickbox_Dialog::Asura_FlagTickbox_Dialog(const char* szName, const Asura_Vector_2& xOffset, u_char* pucFlagsField , u_char ucFlagBit, bool* pbFlag /* = 0 */)
: Asura_Dialog(szName,xOffset)
{
	CreateGadgets(szName, xOffset, (char*)(pucFlagsField), ucFlagBit, pbFlag);
}

void Asura_FlagTickbox_Dialog::CreateGadgets(const char* szName, const Asura_Vector_2& xOffset, int* piFlagsField , int iFlagBit, bool* pbModifiedFlag)
{
	Asura_Vector_2 xPos = xOffset;

	m_pxTickbox = new Asura_FlagTickbox_Gadget(xPos, piFlagsField, iFlagBit, pbModifiedFlag);
	AddGadget(m_pxTickbox);
	AdjustDimensions(szName);
}

void Asura_FlagTickbox_Dialog::CreateGadgets(const char* szName, const Asura_Vector_2& xOffset, char* pcFlagsField, char cFlagBit, bool* pbFlag)
{
	Asura_Vector_2 xPos = xOffset;

	m_pxTickbox = new Asura_FlagTickbox_Gadget(xPos, pcFlagsField, cFlagBit, pbFlag);
	AddGadget(m_pxTickbox);
	AdjustDimensions(szName);
}

void Asura_FlagTickbox_Dialog::AdjustDimensions(const char* szName)
{
	Asura_Assert(m_pxTickbox, "Missing tickbox");

	m_xOffset.x += m_pxTickbox->m_fWidth + 1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	m_pxTickbox->m_fWidth = m_xMaxCoord.x - m_pxTickbox->m_xOffset.x;
}

void Asura_FlagTickbox_Dialog::UpdateLabel(const char* text)
{
	Asura_Dialog::UpdateLabel(text);
	//update the width on the gadget
	if(m_pxGadgetList) 
		m_pxGadgetList->m_fWidth = m_xMaxCoord.x - m_pxGadgetList->m_xOffset.x;

}

void Asura_FlagTickbox_Dialog::SetClickedFunction(void(*pfnOnClicked)(void))
{
	m_pxTickbox->SetClickedFunction(Asura_Callback_Void<void>(pfnOnClicked));
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Asura_SharedTickbox_Dialog **
**----------------------------*/
Asura_SharedTickbox_Dialog::Asura_SharedTickbox_Dialog(const char* pszName, const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<bool>* pxSharedValue) 
: Asura_Dialog(pszName,xOffset)
{
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxnewGadget;
	pxnewGadget  = new Asura_SharedTickbox_Gadget(xPos,pxSharedValue);
	AddGadget(pxnewGadget);

	m_xOffset.x += pxnewGadget->m_fWidth+1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(pszName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxnewGadget->m_fWidth = m_xMaxCoord.x - pxnewGadget->m_xOffset.x;
}



void Asura_SharedTickbox_Dialog::UpdateLabel(const char* text)
{
	Asura_Dialog::UpdateLabel(text);
	//update the width on the gadget
	if(m_pxGadgetList) 
		m_pxGadgetList->m_fWidth = m_xMaxCoord.x - m_pxGadgetList->m_xOffset.x;

}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------**
** Asura_FunctionTickbox_Dialog **
**------------------------------*/

Asura_FunctionTickbox_Dialog::Asura_FunctionTickbox_Dialog(const char* szName, const Asura_Vector_2& xOffset, bool(*pfnIsChecked)(), void(*pfnSetChecked)(bool bSet)) 
: Asura_Dialog(szName,xOffset)
{
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxGadget;
	pxGadget  = new Asura_FunctionTickbox_Gadget(xPos,pfnIsChecked,pfnSetChecked);
	AddGadget(pxGadget);

	m_xOffset.x += pxGadget->m_fWidth+1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxGadget->m_fWidth = m_xMaxCoord.x-pxGadget->m_xOffset.x;
}


Asura_FunctionTickbox_Dialog::Asura_FunctionTickbox_Dialog(const char* szName, const Asura_Vector_2& xOffset, void* pData, bool(*pfnIsChecked)(void* pData), void(*pfnSetChecked)(bool bSet,void* pData)) 
: Asura_Dialog(szName,xOffset)
{
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxGadget;
	pxGadget  = new Asura_FunctionTickbox_Gadget(xPos,pData,pfnIsChecked,pfnSetChecked);
	AddGadget(pxGadget);

	m_xOffset.x += pxGadget->m_fWidth+1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxGadget->m_fWidth = m_xMaxCoord.x-pxGadget->m_xOffset.x;
}

Asura_FunctionTickbox_Dialog::Asura_FunctionTickbox_Dialog(const char* szName, const Asura_Vector_2& xOffset, int iData, bool(*pfnIsChecked)(int iData), void(*pfnSetChecked)(bool bSet,int iData)) 
: Asura_Dialog(szName,xOffset)
{
	Asura_Vector_2 xPos = xOffset;

	Asura_Gadget* pxGadget;
	pxGadget  = new Asura_FunctionTickbox_Gadget(xPos,iData,pfnIsChecked,pfnSetChecked);
	AddGadget(pxGadget);

	m_xOffset.x += pxGadget->m_fWidth+1.0f;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString(szName,ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxGadget->m_fWidth = m_xMaxCoord.x-pxGadget->m_xOffset.x;
}





/////////////////////////////////////////////////////////////////////////////////////////////////

Asura_Function_Dialog::Asura_Function_Dialog( const char* szName,
											 const Asura_Vector_2& xOffset,
											 void(*pfnCallback)(void),
											 bool* pbFlag,
											 bool(*pfnIsEnabled)(void),
											 bool bLineUpWithTickboxes )
											 : Asura_Dialog( szName, xOffset )
{
	m_pxSharedName = 0;
	CreateGadgets(xOffset,Asura_Callback_Void<void>(pfnCallback),pbFlag,pfnIsEnabled,bLineUpWithTickboxes);
}

Asura_Function_Dialog::Asura_Function_Dialog( const char* szName,
											 const Asura_Vector_2& xOffset,
											 void(*pfnCallback)(int),
											 int iUserData,
											 bool* pbFlag,
											 bool(*pfnIsEnabled)(void),
											 bool bLineUpWithTickboxes )
											 : Asura_Dialog( szName, xOffset )
{
	m_pxSharedName = 0;
	CreateGadgets(xOffset,Asura_Callback_UserData<void,int>(pfnCallback,iUserData),pbFlag,pfnIsEnabled,bLineUpWithTickboxes);
}

Asura_Function_Dialog::Asura_Function_Dialog( const char* szName,
											 const Asura_Vector_2& xOffset,
											 const Asura_Callback_Base<void>& xCallback,
											 bool* pbFlag,
											 bool bLineUpWithTickboxes )
											 : Asura_Dialog( szName, xOffset )
{
	m_pxSharedName = 0;
	CreateGadgets(xOffset,xCallback,pbFlag,NULL,bLineUpWithTickboxes);
}

Asura_Function_Dialog::Asura_Function_Dialog( const char* pszNamePrefix,
											 Asura_Dialog_SharedValue<Asura_String>* pxSharedName,
											 const Asura_Vector_2& xOffset,
											 void(*pfnCallback)(void),
											 bool* pbFlag,
											 bool(*pfnIsEnabled)(void),
											 bool bLineUpWithTickboxes )
											 : Asura_Dialog( pszNamePrefix, xOffset )
{
	m_pxSharedName = pxSharedName;
	m_xNamePrefix = pszNamePrefix;
	CreateGadgets(xOffset,Asura_Callback_Void<void>(pfnCallback),pbFlag,pfnIsEnabled,bLineUpWithTickboxes);
	SetUseSharedDialogColour(true);
}

Asura_Function_Dialog::Asura_Function_Dialog( const char* pszNamePrefix,
											 Asura_Dialog_SharedValue<Asura_String>* pxSharedName,
											 const Asura_Vector_2& xOffset,
											 void(*pfnCallback)(int),
											 int iUserData,
											 bool* pbFlag,
											 bool(*pfnIsEnabled)(void),
											 bool bLineUpWithTickboxes )
											 : Asura_Dialog( pszNamePrefix, xOffset )
{
	m_pxSharedName = pxSharedName;
	m_xNamePrefix = pszNamePrefix;
	CreateGadgets(xOffset,Asura_Callback_UserData<void,int>(pfnCallback,iUserData),pbFlag,pfnIsEnabled,bLineUpWithTickboxes);
	SetUseSharedDialogColour(true);
}

Asura_Function_Dialog::Asura_Function_Dialog( const char* pszNamePrefix,
											 Asura_Dialog_SharedValue<Asura_String>* pxSharedName,
											 const Asura_Vector_2& xOffset,
											 const Asura_Callback_Base<void>& xCallback,
											 bool* pbFlag,
											 bool bLineUpWithTickboxes )
											 : Asura_Dialog( pszNamePrefix, xOffset )
{
	m_pxSharedName = pxSharedName;
	m_xNamePrefix = pszNamePrefix;
	CreateGadgets(xOffset,xCallback,pbFlag,NULL,bLineUpWithTickboxes);
	SetUseSharedDialogColour(true);
}


void Asura_Function_Dialog::CreateGadgets( const Asura_Vector_2& xOffset,
										  const Asura_Callback_Base<void>& xCallback,
										  bool* pbFlag,
										  bool(*pfnIsEnabled)(void),
										  bool bLineUpWithTickboxes)
{
	SetIsEnabledFunction(pfnIsEnabled);

	Asura_Gadget* pxNewGadget;

	pxNewGadget = new Asura_Function_Gadget( xOffset, xCallback, pbFlag);
	AddGadget( pxNewGadget );

	float fExtraSpacing = 0.0f;
	if( bLineUpWithTickboxes && pxNewGadget->m_fWidth < Asura_BaseTickbox_Gadget::WIDTH )
	{
		fExtraSpacing = Asura_BaseTickbox_Gadget::WIDTH - pxNewGadget->m_fWidth;
		pxNewGadget->m_xOffset.x += 0.5f * fExtraSpacing;
	}

	m_xOffset.x += pxNewGadget->m_fWidth + 1.0f + fExtraSpacing;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString( m_xName.GetString(),ASURA_TEXT_FORMAT_SMALLFONT );
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxNewGadget->m_fWidth = m_xMaxCoord.x - pxNewGadget->m_xOffset.x;

	m_bMixedSharedValue = false;
}


void Asura_Function_Dialog::UpdateLabel(const char* text)
{
	Asura_Dialog::UpdateLabel(text);
	//update the width on the gadget
	if(m_pxGadgetList) 
		m_pxGadgetList->m_fWidth = m_xMaxCoord.x - m_pxGadgetList->m_xOffset.x;

}

unsigned Asura_Function_Dialog::GetColour( void ) const 
{
	if( m_bIsEnabled)
	{
		u_int uColour = Asura_Dialog::GetColour();
		if(m_bMixedSharedValue && ((uColour&0xffffff) == ASURA_DIALOG_ENABLED_COLOUR))
		{
			//is enabled , and value is different across the selection , so use mixed colour
			uColour = Asura_OverlayBrightness | ASURA_DIALOG_MIXED_COLOUR;
		}
		return uColour;
	}
	else
	{
		return Asura_OverlayBrightness | ASURA_DIALOG_DISABLED_COLOUR;
	}
}

void Asura_Function_Dialog::PerFrameUpdate()
{
	//deal with key for auto-repeating last function gadget.
	//Can't allow it if another gadget has focus , or if there is a modal overlay (always on top)
	if(!Asura_Overlay_System::GetFocusGadget() && !Asura_Overlay_System::AreAnyOverlaysAlwaysOnTop())
	{
		if( (Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT)) &&
			(Asura_UserInput::DebouncedKeyPressed('`')))
		{
			Asura_Function_Gadget* pxGadget = (Asura_Function_Gadget*)m_pxGadgetList;
			pxGadget->RepeatLastFunction();
		}
	}

	if(m_pxSharedName)
	{
		//determine if text needs updating
		Asura_String xName;
		m_pxSharedName->GetValue(&xName,&m_bMixedSharedValue);

		char szText[100];
		Asura_PrintfToBfr(szText,"%s%s",100,true,m_xNamePrefix.GetString(),xName.GetString());

		if(m_xName.Compare(szText))
		{
			UpdateLabel(szText);
			if(m_pxOverlay) 
				m_pxOverlay->ResizeOverlay();
		}
	}	

	Asura_Dialog::PerFrameUpdate();
}

void Asura_Function_Dialog::SetUseSharedDialogColour(bool bShared)
{
	((Asura_Function_Gadget*)m_pxGadgetList)->SetUseSharedGadgetColour(bShared);
}

void Asura_Function_Dialog::SetCallback(const Asura_Callback_Base<void>& xCallback)
{
	((Asura_Function_Gadget*)m_pxGadgetList)->SetCallback(xCallback);
}

Asura_FunctionWithKeybinding_Dialog::Asura_FunctionWithKeybinding_Dialog( const char* szName,
																		 const Asura_Vector_2& xOffset,
																		 void(*pfnCallback)(void),
																		 const Asura_Key_Combination& xBinding,
																		 bool* pbFlag,
																		 bool bLineUpWithTickboxes )
																		 : Asura_Dialog( szName, xOffset )
{
	Asura_Gadget* pxNewGadget = new Asura_FunctionWithKeybinding_Gadget( xOffset, Asura_Callback_Void<void>(pfnCallback), xBinding, pbFlag );
	AddGadget( pxNewGadget );

	float fExtraSpacing = 0.0f;
	if( bLineUpWithTickboxes && pxNewGadget->m_fWidth < Asura_BaseTickbox_Gadget::WIDTH )
	{
		fExtraSpacing = Asura_BaseTickbox_Gadget::WIDTH - pxNewGadget->m_fWidth;
		pxNewGadget->m_xOffset.x += 0.5f * fExtraSpacing;
	}

	m_xOffset.x += pxNewGadget->m_fWidth + 1.0f + fExtraSpacing;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString( szName,ASURA_TEXT_FORMAT_SMALLFONT );
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxNewGadget->m_fWidth = m_xMaxCoord.x - pxNewGadget->m_xOffset.x;
}

Asura_FunctionWithKeybinding_Dialog::Asura_FunctionWithKeybinding_Dialog( const char* szName,
																		 const Asura_Vector_2& xOffset,
																		 void(*pfnCallback)(int),
																		 int iUserData,
																		 const Asura_Key_Combination& xBinding,
																		 bool* pbFlag,
																		 bool bLineUpWithTickboxes )
																		 : Asura_Dialog( szName, xOffset )
{
	Asura_Gadget* pxNewGadget = new Asura_FunctionWithKeybinding_Gadget( xOffset, Asura_Callback_UserData<void,int>(pfnCallback, iUserData), xBinding, pbFlag );
	AddGadget( pxNewGadget );

	float fExtraSpacing = 0.0f;
	if( bLineUpWithTickboxes && pxNewGadget->m_fWidth < Asura_BaseTickbox_Gadget::WIDTH )
	{
		fExtraSpacing = Asura_BaseTickbox_Gadget::WIDTH - pxNewGadget->m_fWidth;
		pxNewGadget->m_xOffset.x += 0.5f * fExtraSpacing;
	}

	m_xOffset.x += pxNewGadget->m_fWidth + 1.0f + fExtraSpacing;

	m_xMaxCoord.x = m_xOffset.x + Asura_Text_Output::WidthOfString( szName,ASURA_TEXT_FORMAT_SMALLFONT );
	m_xMaxCoord.y = m_xOffset.y + 1.0f;

	//extend the width of the gadget , so that clicking on the text will activate it
	pxNewGadget->m_fWidth = m_xMaxCoord.x - pxNewGadget->m_xOffset.x;
}

void Asura_FunctionWithKeybinding_Dialog::UpdateLabel(const char* text)
{
	Asura_Dialog::UpdateLabel(text);
	//update the width on the gadget
	if(m_pxGadgetList)
		m_pxGadgetList->m_fWidth = m_xMaxCoord.x - m_pxGadgetList->m_xOffset.x;
}

Asura_Windows_Dialog::Asura_Windows_Dialog(const char* name, const Asura_Vector_2& xOffset, float width, bool* shouldClose, u_int* puMinimisedState) : Asura_Dialog(name, xOffset, 0x7fffffff)
{
	Asura_Vector_2 pos = xOffset;
	pos.x += width;

	Asura_Gadget* newGadget;

	if (shouldClose)
	{
		pos.x -= 2.0f;
		newGadget = new Asura_Close_Gadget(pos,shouldClose);
		AddGadget(newGadget);
	}

	m_puMinimisedState = puMinimisedState;
	if (puMinimisedState)
	{
		pos.x -= 2.0f;
		newGadget = new Asura_Minimise_Gadget(pos,puMinimisedState,false);
		AddGadget(newGadget);
		
		pos.x -= 2.0f;
		newGadget = new Asura_Minimise_Gadget(pos,puMinimisedState,true);
		AddGadget(newGadget);

	}
    	
}

void Asura_Windows_Dialog::Render( const Asura_Vector_2& xOrigin )
{
	if( !IsHidden() )
	{
		Asura_Vector_2 pos = xOrigin + m_xOffset;

		unsigned int colour;

		if (!m_puMinimisedState || m_pxOverlay->m_bAlwaysOnTop)
		{
			colour = 0xff0000;
		}
		else
		{
			colour = 0xffff00;
		}

		if( Asura_Overlay_System::UseSkinnedUI() )
		{
			Asura_Text_Output::Print(pos.x+0.1f,pos.y+0.1f,Asura_OverlayBrightness|0xff000000,ASURA_TEXT_FORMAT_SMALLFONT,"%s",m_xName.GetString());
			Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_SMALLFONT,"%s",m_xName.GetString());
		}
		else
		{
			Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,0,"%s",m_xName.GetString());
		}

		if( m_pxGadgetList ) { m_pxGadgetList->RenderList( xOrigin ); }

		if (m_pxOverlay->IsMinimised()) return;
	}

}

void Asura_Windows_Dialog::Update( const Asura_Vector_2& xOrigin )
{
	if (m_pxGadgetList) 
		m_pxGadgetList->Handle( xOrigin );


	//Not entirely sure there is any point in this , but this seems to be the closest
	//to the old code for when minimised was just a boolean
	//RAR 08/11/2007
	if (m_puMinimisedState)
	{
		if (*m_puMinimisedState)
		{
			m_pxOverlay->m_bDialogTookFocus = true;
		}
	}
}
//////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------**
** Asura_Resize_Dialog **
**---------------------*/

Asura_Resize_Dialog::Asura_Resize_Dialog(Asura_Dialog* pxDialogToResize, Asura_Vector_2* pxUserSetSize)
:Asura_Dialog(NULL, Asura_Vector_2(0.0f,0.0f))
{
	AddResizableDialog(pxDialogToResize, pxUserSetSize);

	Asura_Gadget* pxGadget = new Asura_Resize_Gadget(Asura_Vector_2(0.0f,0.0f));
	AddGadget(pxGadget);

	m_xMaxCoord.x = m_xOffset.x + pxGadget->m_fWidth;
	m_xMaxCoord.y = m_xOffset.y + pxGadget->m_fHeight;

	SetOverlayRelativePosition(Asura_Vector_2(-pxGadget->m_fWidth,-pxGadget->m_fHeight),true,true);

}

void Asura_Resize_Dialog::AddResizableDialog(Asura_Dialog* pxDialogToResize, Asura_Vector_2* pxUserSetSize)
{
	Asura_Assert(pxDialogToResize, "");
	if(pxDialogToResize)
	{
		pxDialogToResize->SetAllowResize(true, pxUserSetSize);
		m_xResizeList.Push_Back(pxDialogToResize);
	}
}

void Asura_Resize_Dialog::StartResize(u_int uResizeFlags)
{
	//For the initial resize , all the dialogs are dealt with directly by the overlay , so
	//we don't want to deal with them again here.
	if(uResizeFlags & ASURA_OVERLAY_RESIZE_INITIAL) return;

	//Update all the dialogs controlled by this
	for(Asura_LinkList_It<Asura_Dialog*> xDialogIt(m_xResizeList); !xDialogIt.Done(); xDialogIt.Next())
	{
		(*xDialogIt.GetCurrent())->StartResize(uResizeFlags);
	}
}

void Asura_Resize_Dialog::UpdateResize(u_int uResizeFlags, const Asura_Vector_2& xDelta)
{
	//For the initial resize , all the dialogs are dealt with directly by the overlay , so
	//we don't want to deal with them again here.
	if(uResizeFlags & ASURA_OVERLAY_RESIZE_INITIAL) return;

	//Update all the dialogs controlled by this
	for(Asura_LinkList_It<Asura_Dialog*> xDialogIt(m_xResizeList); !xDialogIt.Done(); xDialogIt.Next())
	{
		(*xDialogIt.GetCurrent())->UpdateResize(uResizeFlags, xDelta);
	}
}

u_int Asura_Resize_Dialog::GetAllowedResizeFlags() const
{
	if(!m_xResizeList.Size()) return 0;
	
	u_int uFlags = 0xffffffff;
	for(Asura_LinkList_ConstIt<Asura_Dialog*> xDialogIt(m_xResizeList); !xDialogIt.Done(); xDialogIt.Next())
	{
		uFlags &= (*xDialogIt.GetCurrent())->GetAllowedResizeFlags();
	}

	return uFlags;
}

void Asura_Resize_Dialog::RemoveReferencesToDialog(const Asura_Dialog* pxDialog)
{
	Asura_Dialog::RemoveReferencesToDialog(pxDialog);
	m_xResizeList.Remove(const_cast<Asura_Dialog*>(pxDialog));
}


//////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------**
** Asura_List_Dialog **
**-------------------*/

Asura_List_Dialog::Asura_List_Dialog(const Asura_Vector_2& offset, const char* const* pszList, int* listOffset, int* selection, int numberOfLines, int numberInList, void(*function)(void)) 
: Asura_Dialog(0,offset)
, m_bOnlySelectOnClick( false )
{
	m_piSelectionIndex = 0;
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = 0;
	m_pfnIsChecked = 0;
	m_pfnToggleChecked = 0;
	m_iUserData = 0;
	m_pxMultipleSelectionHandler = 0;
	m_bHasOutline = true;
	m_bExpandWidthToFillOverlay = true;
	m_fMinWidth = 0.0f;
	m_fTextWidth = 0.0f;
	m_pxFirstTickboxGadget = 0;
	m_iNumberOfLinesBeforeResize = 0;

	CreateGadgets( offset, pszList, listOffset, listOffset, selection, numberOfLines, numberInList, function );
}

Asura_List_Dialog::Asura_List_Dialog(const Asura_Vector_2& xOffset, const char* const* ppszList, int* piSelectionIndex, int iNumberOfLines, int iNumberInList, void(*pFunction)(void),bool* pbChangedFlag /*= 0*/, const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0,xOffset)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_piSelectionIndex = piSelectionIndex;
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = pbChangedFlag;
	m_pfnIsChecked = 0;
	m_pfnToggleChecked = 0;
	m_iUserData = 0;
	m_pxMultipleSelectionHandler = 0;
	m_bHasOutline = true;
	m_bExpandWidthToFillOverlay = true;
	m_fMinWidth = 0.0f;
	m_fTextWidth = 0.0f;
	m_pxFirstTickboxGadget = 0;
	m_iNumberOfLinesBeforeResize = 0;

	int* const piListOffsetScroll = ( bOnlySelectOnClick ) ? &m_iListOffsetScroll : &m_iListOffset;
	CreateGadgets( xOffset, ppszList, &m_iListOffset, piListOffsetScroll, &m_iListSelection, iNumberOfLines, iNumberInList, pFunction );
	SetSelection(*m_piSelectionIndex);
}

Asura_List_Dialog::Asura_List_Dialog(const Asura_Vector_2& xOffset, const char* const* ppszList, int* piSelectionIndex, int iNumberOfLines, int iNumberInList,  bool(*pfnIsChecked)(int,int),void(*pfnToggleChecked)(int,int),int iUserData,bool* pbChangedFlag /*= 0*/, const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0,xOffset)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_piSelectionIndex = piSelectionIndex;
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = pbChangedFlag;
	m_pfnIsChecked = pfnIsChecked;
	m_pfnToggleChecked = pfnToggleChecked;
	m_iUserData = iUserData;
	m_pxMultipleSelectionHandler = 0;
	m_bHasOutline = true;
	m_bExpandWidthToFillOverlay = true;
	m_fMinWidth = 0.0f;
	m_fTextWidth = 0.0f;
	m_pxFirstTickboxGadget = 0;
	m_iNumberOfLinesBeforeResize = 0;

	int* const piListOffsetScroll = ( bOnlySelectOnClick ) ? &m_iListOffsetScroll : &m_iListOffset;
	CreateGadgets( xOffset, ppszList, &m_iListOffset, piListOffsetScroll, &m_iListSelection, iNumberOfLines, iNumberInList, 0 );
	SetSelection(*m_piSelectionIndex);
}

Asura_List_Dialog::~Asura_List_Dialog()
{
	delete m_pxMultipleSelectionHandler;
}

void Asura_List_Dialog::CreateGadgets(const Asura_Vector_2& xOffset, const char* const* ppszList, int* piListOffset, int* piListOffsetScroll, int* piSelection, int iNumberOfLines, int iNumberInList, void(*pFunction)(void))
{
	m_pxUpGadget = 0;
	m_pxDownGadget = 0;
	m_pxScrollGadget = 0;



	Asura_Vector_2 pos = xOffset;

	bool needToBeScrollable = false;

	if (iNumberInList>iNumberOfLines) needToBeScrollable = true;

	int iMaxScroll = (iNumberInList > iNumberOfLines) ? (iNumberInList - iNumberOfLines) : 0;

	m_pxUpGadget = new Asura_UpDown_Gadget( pos, piListOffsetScroll, 0, iMaxScroll, -1 );
	AddGadget(m_pxUpGadget);

#if 1
	if( iNumberOfLines >= 4 )
	{
		//add scroll bar
		pos.y += 1.2f;
		float fHeight = 1.2f * ( iNumberOfLines - 2 );
		m_pxScrollGadget =new Asura_VerticalScroll_Gadget( pos, fHeight, iNumberOfLines, piListOffsetScroll, 0, iMaxScroll );
		AddGadget( m_pxScrollGadget );
		pos.y -= 1.2f;
	}	
#endif

	//find the longest string
	float fMaxLength = 1.0f; //arbitrary minimum starting point
	for (int i=0; i<iNumberInList; i++)
	{
		if( ppszList[i] )
		{
			float fLength = Asura_Text_Output::WidthOfString(ppszList[i],ASURA_TEXT_FORMAT_SMALLFONT) + 0.5f;
			if(fLength>fMaxLength) fMaxLength = fLength;
		}		
	}
	float fMaxAllowed = (float)((Asura_Viewport::GetWidth()/10)-10);
	if(fMaxLength>fMaxAllowed) fMaxLength = fMaxAllowed;

	m_fTextWidth = fMaxLength;
	float fTextGadgetWidth = Asura_Maths::Max(m_fTextWidth, m_fMinWidth);


	pos.x += 2.0f;
	if(HasTickboxes()) pos.x += 3.0f;

	for (int i=0; i<iNumberOfLines; i++)
	{
		Asura_Gadget* const pxGadget = new Asura_TextSelect_Gadget( pos, piSelection, i, piListOffset, ppszList, iNumberOfLines, iNumberInList, pFunction, m_bOnlySelectOnClick );
		pxGadget->m_fHeight = 1.2f;//increase the height so it isn't possible for the cursor to fall in a gap between lines
		pxGadget->m_fWidth = fTextGadgetWidth;//set the width of all the gadgets according to the maximum string length
		pxGadget->SetExpandWidthToFillOverlay(m_bExpandWidthToFillOverlay);
		AddGadget(pxGadget);
		pos.y += 1.2f;
	}

	//store the pointer of the first text gadget in the linked list
	m_pxFirstTextGadget = (Asura_TextSelect_Gadget*)m_pxGadgetList;
	m_iNumberOfLines = iNumberOfLines;
	m_iNumberInList = iNumberInList;

	if(HasTickboxes())
	{
		pos.x -= 3.0f;
		pos.y -= 1.2f*iNumberOfLines;


		for (int i=0; i<iNumberOfLines; i++)
		{
			static bool s_bTest;
			Asura_Gadget* const pxGadget = new Asura_ListTickbox_Gadget(pos,this,i,m_pbChangedFlag);
			pxGadget->m_fHeight = 1.2f;//increase the height so it isn't possible for the cursor to fall in a gap between lines
			pxGadget->m_fWidth = fTextGadgetWidth+3.0f;//set the width of all the gadgets according to the maximum string length
			pxGadget->SetExpandWidthToFillOverlay(m_bExpandWidthToFillOverlay);
			AddGadget(pxGadget);
			pos.y += 1.2f;
		}

		m_pxFirstTickboxGadget = (Asura_ListTickbox_Gadget*) m_pxGadgetList;
	}

	pos.y -= 1.2f;	
	pos.x -= 2.0f;


	m_pxDownGadget = new Asura_UpDown_Gadget( pos, piListOffsetScroll, 0, iMaxScroll, 1 );
	AddGadget( m_pxDownGadget );


	m_xMaxCoord.x = m_xOffset.x + fTextGadgetWidth + 2.0f;
	m_xMaxCoord.y = m_xOffset.y + 1.2f*iNumberOfLines;

	if(HasTickboxes())
	{
		m_xMaxCoord.x += 3.0f;
	}

	if(!needToBeScrollable)
	{
		//hide the scroll gadgets
		if(m_pxUpGadget) m_pxUpGadget->SetHidden(true);
		if(m_pxDownGadget) m_pxDownGadget->SetHidden(true);
		if(m_pxScrollGadget) m_pxScrollGadget->SetHidden(true);
	}

}

void Asura_List_Dialog::SetListHelpStrings(const char* const * ppszHelpText)
{
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;

	//give the pszList of help strings to each of the text select gadgets
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetListHelpStrings(ppszHelpText);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}
}

void Asura_List_Dialog::SetTooltipCallback( void (*pfnDisplayTooltipCallback)(int iIndex,const char* szText))
{
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;

	//set the callback in each of the text select gadgets
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetTooltipCallback(pfnDisplayTooltipCallback);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}
}



void Asura_List_Dialog::SetBorder(bool bEnabled)
{
	m_bHasOutline = bEnabled;
}

void Asura_List_Dialog::SetExpandWidthToFillOverlay(bool bSet)
{
	m_bExpandWidthToFillOverlay = bSet;

	//set the flag in the text select gadgets as well
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;

	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetExpandWidthToFillOverlay(bSet);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}

	if(m_pxFirstTickboxGadget)
	{
		Asura_ListTickbox_Gadget* pxTickboxGadget = m_pxFirstTickboxGadget;
		for(int i=0;i<m_iNumberOfLines;i++)
		{
			Asura_Assert(pxTickboxGadget,"");

			pxTickboxGadget->SetExpandWidthToFillOverlay(bSet);
			pxTickboxGadget = (Asura_ListTickbox_Gadget*)(pxTextGadget->m_pxNextGadget);
		}

	}
}

void Asura_List_Dialog::SetMinWidth(float fMinWidth)
{
	m_fMinWidth = fMinWidth;

	//update all the gadget widths
	SetWidths();
}



void Asura_List_Dialog::Render( const Asura_Vector_2& xOrigin )
{
	if( !IsHidden() )
	{
		if( m_bHasOutline )
		{	
			float fX = (float)Asura_Maths::FloatToInt_Near((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH);			
			float fY = (float)Asura_Maths::FloatToInt_Near((xOrigin.y + m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT);			
			float fHeight = ((m_xMaxCoord.y - m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT-2.0f);

			float fWidth = 0.0f;
			if(m_bExpandWidthToFillOverlay)
			{
				fWidth = (m_pxOverlay->m_fWidth - (1.0f + m_xOffset.x)) * ASURA_TEXT_OUTPUT_FONT_WIDTH;
			}
			else
			{
				fWidth = (m_xMaxCoord.x - m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH;
			}

			
			uint32 uAlpha = (Asura_OverlayBrightness>>24) / 6;						
			u_int uColourL = (uAlpha<<24) | ASURA_GADGET_DEFAULT_COLOUR;
			u_int uColourD = (uAlpha<<24);

			Asura_Render::Render2DLine(fX,fY,fX+fWidth,fY,uColourD);
			Asura_Render::Render2DLine(fX+fWidth,fY,fX+fWidth,fY+fHeight,uColourD);
			Asura_Render::Render2DLine(fX+fWidth,fY+fHeight,fX,fY+fHeight,uColourL);
			Asura_Render::Render2DLine(fX,fY+fHeight,fX,fY,uColourL);
		}

		if( m_pxGadgetList ) { m_pxGadgetList->RenderList( xOrigin ); }
	}
}

void Asura_List_Dialog::Update( const Asura_Vector_2& xOrigin )
{
	if( m_piSelectionIndex )
	{
		//handle external changes to the selection
		if( *m_piSelectionIndex != ( m_iListOffset + m_iListSelection ) )
		{
			SetSelection( *m_piSelectionIndex, false );
		}
	}

	if( m_pxMultipleSelectionHandler )
	{
		m_pxMultipleSelectionHandler->PreGadgetUpdate();
	}

	//update the gadgets
	int iCurrentOffset	= m_iListOffset;		// for OnlySelectOnClick
	m_iListOffsetScroll	= m_iListOffset;		// for OnlySelectOnClick
	Asura_Dialog::Update(xOrigin);

	if( m_piSelectionIndex )
	{
		if( m_bOnlySelectOnClick )
		{
			const int iDelta	= m_iListOffsetScroll - iCurrentOffset;
			m_iListSelection	-= iDelta;
			m_iListOffset		+= iDelta;
			m_iListOffsetScroll	= m_iListOffset;

			// capping should not be required, but just to be sure
			int iIndex;
			iIndex = m_iListOffset + m_iListSelection;
			if( iIndex >= m_iNumberInList )
			{
				m_iListSelection -= ( iIndex - ( m_iNumberInList - 1 ) );
			}
			iIndex = m_iListOffset + m_iListSelection;
			if( iIndex < 0 )
			{
				m_iListSelection -= iIndex;
			}
		}

		//see if any selection change was made
		if( *m_piSelectionIndex != ( m_iListOffset + m_iListSelection ) )
		{
			*m_piSelectionIndex = m_iListOffset + m_iListSelection;
			if(m_pbChangedFlag && !HasTickboxes() && !m_pxMultipleSelectionHandler) *m_pbChangedFlag = true;
		}
	}

	if( m_pxMultipleSelectionHandler )
	{
		m_pxMultipleSelectionHandler->PostGadgetUpdate();
	}
}

void Asura_List_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	if( m_piSelectionIndex )
	{
		//handle external changes to the selection
		if( *m_piSelectionIndex != ( m_iListOffset + m_iListSelection ) )
		{
			SetSelection( *m_piSelectionIndex, false );
		}
	}
}

void Asura_List_Dialog::AdjustListSelectionAndOffset(int* piListOffset,int* piListSelection,int iNumberOfLines,int iNumberInList)
{
	Asura_Assert(piListOffset,"");
	Asura_Assert(piListSelection,"");

	*piListOffset += *piListSelection;
	*piListSelection = 0;

	//make sure the selection and offset values are in range. RAR 01/04/03
	if(iNumberInList)
	{
		if((*piListOffset)+(*piListSelection)>=iNumberInList)
		{
			(*piListSelection) = Asura_Maths::Max(0,iNumberInList-(*piListOffset)-1);
			(*piListOffset) = Asura_Maths::Max(0,iNumberInList-(*piListSelection)-1);
		}


		while((*piListOffset)>0 && (*piListOffset)+iNumberOfLines>iNumberInList)
		{
			(*piListOffset)--;
			(*piListSelection)++;
		}
	}
	else
	{
		(*piListSelection) = 0;
		(*piListOffset) = 0;
	}
}

void Asura_List_Dialog::SetSelection(int iIndex, bool bAffectMultipleSelection /*= true*/)
{
	if(m_pxFirstTextGadget) m_pxFirstTextGadget->SetSelection(iIndex, bAffectMultipleSelection);
}

void Asura_List_Dialog::ReplaceStrings(int iNumberInList,const char* const* ppszList)
{
	//get the max length for the new strings
	float fMaxLength = 1.0f; //arbitrary minimum starting point
	for (int i=0; i<iNumberInList; i++)
	{
		if( ppszList[i] )
		{
			float fLength = Asura_Text_Output::WidthOfString(ppszList[i],ASURA_TEXT_FORMAT_SMALLFONT) + 0.5f;
			if(fLength>fMaxLength) fMaxLength = fLength;
		}		
	}
	float fMaxAllowed = (float)((Asura_Viewport::GetWidth()/10)-10);
	if(fMaxLength>fMaxAllowed) fMaxLength = fMaxAllowed;

	m_fTextWidth = fMaxLength;
	float fTextGadgetWidth = Asura_Maths::Max(m_fTextWidth, m_fMinWidth);
	
	
	//update the text gadgets
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->ReplaceStrings(iNumberInList,ppszList);
		pxTextGadget->m_fWidth = fTextGadgetWidth;//set the width of all the gadgets according to the maximum string length
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);

	}
	m_iNumberInList = iNumberInList;



	///update the other gadgets
	int iMax = (m_iNumberInList > m_iNumberOfLines) ? (m_iNumberInList - m_iNumberOfLines) : 0;
	bool bScrollable = (m_iNumberInList > m_iNumberOfLines);
	if(m_pxUpGadget)
	{
		m_pxUpGadget->SetMax(iMax);
		m_pxUpGadget->SetHidden(!bScrollable);
	}
	if(m_pxDownGadget)
	{
		m_pxDownGadget->SetMax(iMax);
		m_pxDownGadget->SetHidden(!bScrollable);
	}
	if(m_pxScrollGadget)
	{
		m_pxScrollGadget->SetMinAndMax(0,iMax);
		m_pxScrollGadget->SetHidden(!bScrollable);
	}

	if(m_pxMultipleSelectionHandler)
	{
		m_pxMultipleSelectionHandler->SetSize(m_iNumberInList);
	}


	//update all the widths
	SetWidths();


}

void Asura_List_Dialog::SetWidths()
{
	float fTextGadgetWidth = Asura_Maths::Max(m_fTextWidth, m_fMinWidth);

	//update the text gadget widths
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->m_fWidth = fTextGadgetWidth;//set the width of all the gadgets according to the maximum string length
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}

	//and the tickboxes
	if(HasTickboxes())
	{
		Asura_Assert(m_pxFirstTickboxGadget,"");
		
		Asura_ListTickbox_Gadget* pxTickboxGadget = m_pxFirstTickboxGadget;
		for(int i=0;i<m_iNumberOfLines;i++)
		{
			Asura_Assert(pxTickboxGadget,"");

			pxTickboxGadget->m_fWidth = fTextGadgetWidth + 3.0f;
			pxTickboxGadget = static_cast<Asura_ListTickbox_Gadget*> (pxTickboxGadget->m_pxNextGadget);
		}

	}

	//update the dialog width
	m_xMaxCoord.x = m_xOffset.x + fTextGadgetWidth + 2.0f;
	if(HasTickboxes())
	{
		m_xMaxCoord.x += 3.0f;
	}

	if(m_pxOverlay) 
		m_pxOverlay->ResizeOverlay();

}

bool Asura_List_Dialog::IsChecked(int iIndex)
{
	int iTotalIndex = iIndex + m_iListOffset;
	if(iTotalIndex>=0 && iTotalIndex<m_iNumberInList && m_pfnIsChecked)
	{
		return m_pfnIsChecked(iTotalIndex,m_iUserData);
	}
	return true;
}

void Asura_List_Dialog::ToggleChecked(int iIndex)
{
	int iTotalIndex = iIndex + m_iListOffset;
	if(iTotalIndex>=0 && iTotalIndex<m_iNumberInList && m_pfnToggleChecked)
	{
		m_pfnToggleChecked(iTotalIndex,m_iUserData);
	}
}

void Asura_List_Dialog::ShowScrollBar(bool bShow)
{
	if(m_pxUpGadget) m_pxUpGadget->SetHidden(!bShow);
	if(m_pxDownGadget) m_pxDownGadget->SetHidden(!bShow);
	if(m_pxScrollGadget) m_pxScrollGadget->SetHidden(!bShow);
}

void Asura_List_Dialog::EnableMultipleSelection(bool bEnable,bool (*pfnIsSelected)(int iIndex) /*= 0*/,void (*pfnSetSelected)(int iIndex,bool bSet) /*= 0*/)
{
	//create/delete list as needed
	if(bEnable && !m_pxMultipleSelectionHandler)
	{
		m_pxMultipleSelectionHandler = new Asura_List_Dialog_SelectionHandler(m_iNumberInList,m_pbChangedFlag,pfnIsSelected,pfnSetSelected);
	}
	else if(!bEnable)
	{
		delete m_pxMultipleSelectionHandler;
		m_pxMultipleSelectionHandler = 0;
	}

	//update handler pointer in the text gadgets
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetMultipleSelectionHandler(m_pxMultipleSelectionHandler);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}
}

int Asura_List_Dialog::GetNumSelected() const
{
	Asura_Assert(m_pxMultipleSelectionHandler,"Multiple selection not enabled");
	return m_pxMultipleSelectionHandler ? m_pxMultipleSelectionHandler->GetNumSelected() : 0;
}

int Asura_List_Dialog::GetSelection(int iIndex) const
{
	Asura_Assert(m_pxMultipleSelectionHandler,"Multiple selection not enabled");
	return m_pxMultipleSelectionHandler ? m_pxMultipleSelectionHandler->GetSelection(iIndex) : 0;
}

void Asura_List_Dialog::SetDoubleClickFunction(const Asura_Callback_Base<void>& xDoubleClickCallback)
{
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetDoubleClickFunction(xDoubleClickCallback);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}
}

void Asura_List_Dialog::SetSingleClickFunction(const Asura_Callback_Base<void>& xSingleClickCallback)
{
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetSingleClickFunction(xSingleClickCallback);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}
}

void Asura_List_Dialog::UpdateResize(u_int uResizeFlags, const Asura_Vector_2& xDelta)
{
	if(!m_bAllowResize) return;

	//determine the new number of lines
	int iNewNumberOfLines = m_iNumberOfLines;
	if(uResizeFlags & ASURA_OVERLAY_RESIZE_INITIAL)
	{
		//initialise the size based on the last size chosen by the user
		if(!m_pxUserSetSize) return;
		if(*m_pxUserSetSize == Asura_Maths::ZeroVector2) return;

		iNewNumberOfLines = Asura_Maths::FloatToInt_Near(m_pxUserSetSize->y/1.2f);
		if(iNewNumberOfLines < 1) iNewNumberOfLines = 1;
	}
	else
	{
		int iLineDelta = Asura_Maths::FloatToInt_Near(xDelta.y / 1.2f);
		iNewNumberOfLines = m_iNumberOfLinesBeforeResize + iLineDelta;
		if(iNewNumberOfLines < 1) iNewNumberOfLines = 1;

		if(m_pxUserSetSize)
		{
			m_pxUserSetSize->y = iNewNumberOfLines * 1.2f;
		}
	}

	//if the number has changed , then we need to make the gadgets again
	if(iNewNumberOfLines != m_iNumberOfLines)
	{
		//get the current settings from one of the text gadgets
		const char* const* pszList = m_pxFirstTextGadget->GetList();
		const char* const* pszHelpText = m_pxFirstTextGadget->GetHelpText();
		int* piSelection = m_pxFirstTextGadget->GetSelectionPointer();
		int* piListOffset = m_pxFirstTextGadget->GetListOffsetPointer();
		Asura_Callback_Wrapper<void> xSingleClickCallback = m_pxFirstTextGadget->GetSingleClickCallback();
		Asura_Callback_Wrapper<void> xDoubleClickCallback = m_pxFirstTextGadget->GetDoubleClickCallback();
		int* piListOffsetScroll = m_pxUpGadget->GetValuePointer();

		//delete them all			
		while(m_pxGadgetList) DeleteSingleGadget(m_pxGadgetList);
	
		m_pxFirstTextGadget = 0;
		m_pxFirstTickboxGadget = 0;
		m_pxUpGadget = 0;
		m_pxDownGadget = 0;
		m_pxScrollGadget = 0;

		//and set up the gadgets again
		CreateGadgets(m_xOffset, pszList, piListOffset, piListOffsetScroll, piSelection, iNewNumberOfLines , m_iNumberInList, NULL);
		if(xSingleClickCallback.IsValid()) SetSingleClickFunction(*xSingleClickCallback.GetCallback());
		if(xDoubleClickCallback.IsValid()) SetDoubleClickFunction(*xDoubleClickCallback.GetCallback());
		
		if(pszHelpText) SetListHelpStrings(pszHelpText);

		if(m_pxMultipleSelectionHandler)
		{
			//update handler pointer in the text gadgets
			Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
			for(int i=0;i<m_iNumberOfLines;i++)
			{
				Asura_Assert(pxTextGadget,"");

				pxTextGadget->SetMultipleSelectionHandler(m_pxMultipleSelectionHandler);
				pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
			}
		}
		//Use SetSelection() to ensure that the main selection hasn't gone of the bottom of the list
		Asura_Assert(piSelection && piListOffset, "");
		SetSelection(*piSelection + *piListOffset, false);
	}
}

u_int Asura_List_Dialog::GetAllowedResizeFlags() const
{
	return m_bAllowResize ? ASURA_OVERLAY_RESIZE_VERTICAL : 0;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------------**
** Asura_List_Dialog_SelectionHandler **
**------------------------------------*/

Asura_List_Dialog_SelectionHandler::Asura_List_Dialog_SelectionHandler(int iNumberInList,bool* pbChangedFlag,bool (*pfnIsSelected)(int iIndex),void (*pfnSetSelected)(int iIndex,bool bSet))
{
	if(pfnIsSelected && pfnSetSelected)
	{
		m_pxList = 0;
		m_pfnIsSelected = pfnIsSelected;
		m_pfnSetSelected = pfnSetSelected;
	}
	else
	{
		m_pxList = new List<int>;
		m_pfnIsSelected = 0;
		m_pfnSetSelected = 0;
	}
	m_iNumberInList = iNumberInList;
	m_pbChangedFlag = pbChangedFlag;
	m_iAnchorIndex = 0;
	m_iDragStartIndex = -1;
	m_iMouseIndex = -1;
	m_bUpdatedThisFrame = false;
}

Asura_List_Dialog_SelectionHandler::~Asura_List_Dialog_SelectionHandler()
{
	delete m_pxList;
}

void Asura_List_Dialog_SelectionHandler::ClearSelection()
{
	if(m_pxList)
	{
		while(m_pxList->size()) m_pxList->delete_first_entry();
	}
	else
	{
		for(int i=0;i<m_iNumberInList;i++)
		{
			m_pfnSetSelected(i,false);			
		}
	}
}

void Asura_List_Dialog_SelectionHandler::SetSize(int iNumberInList)
{
	if(m_pxList)
	{
		//If using multiple selection , clear the current selection since it most likely won't be meaningful in the new list.
		ClearSelection();
	}
	m_iNumberInList = iNumberInList;
	if(m_iAnchorIndex >= m_iNumberInList) m_iAnchorIndex = 0;
	m_iDragStartIndex = -1;
	m_iMouseIndex = -1;
}

int Asura_List_Dialog_SelectionHandler::GetNumSelected() const
{
	if(m_pxList)
	{
		return m_pxList->size();
	}
	else
	{
		Asura_Assert(0,"GetNumSelected() not supported if selection is stored externally");
		return 0;
	}
}

int Asura_List_Dialog_SelectionHandler::GetSelection(int iIndex) const
{
	if(m_pxList)
	{
		Asura_Assert(iIndex>=0 && iIndex< m_pxList->size(),"Invalid selection index");
		return (*m_pxList)[iIndex];
	}
	else
	{
		Asura_Assert(0,"GetSelection() not supported if selection is stored externally");
		return 0;
	}
}

bool Asura_List_Dialog_SelectionHandler::IsSelected(int iIndex) const
{
	if(m_pxList)
	{
		return m_pxList->contains(iIndex);
	}
	else
	{
		Asura_Assert(m_pfnIsSelected,"");
		return m_pfnIsSelected(iIndex);
	}
}

void Asura_List_Dialog_SelectionHandler::SetSelected(int iIndex,bool bSet)
{
	if(m_pxList)
	{
		if(bSet)
		{
			if(!m_pxList->contains(iIndex))
			{
				m_pxList->add_entry(iIndex);
			}
		}
		else
		{
			if(m_pxList->contains(iIndex))
			{
				m_pxList->delete_entry(iIndex);
			}
		}
	}
	else
	{
		Asura_Assert(m_pfnSetSelected,"");
		m_pfnSetSelected(iIndex,bSet);
	}
}


bool Asura_List_Dialog_SelectionHandler::IsInDragSelection(int iIndex) const
{
	if(m_iDragStartIndex == -1 || m_iMouseIndex == -1) return false;

	int iMinIndex = Asura_Maths::Min(m_iMouseIndex,m_iDragStartIndex);
	int iMaxIndex = Asura_Maths::Max(m_iMouseIndex,m_iDragStartIndex);

	return iMinIndex <= iIndex && iIndex <= iMaxIndex;
}

void Asura_List_Dialog_SelectionHandler::Update(int iIndex,bool bClickable)
{
	Asura_Assert(iIndex>=0 && iIndex<m_iNumberInList,"Invalid index");
	m_iMouseIndex = iIndex;
	m_bUpdatedThisFrame = true;

	if(!bClickable)
	{
		//cancel any drag selection if no longer clickable
		m_iDragStartIndex = -1;
		return;
	}

	if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
	{
		//started selection
		m_iDragStartIndex = iIndex;
	}
	else if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1) && m_iDragStartIndex != -1)
	{
		//ended selection process

		//If control isn't pressed then the new selection replaces the old
		bool bReplaceSelection = !Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL);
		//If shift is pressed then everything between the 'anchor' and the item clicked on is selected
		bool bBlockSelection = Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT);
		bool bDragSelection = false;

		if(bReplaceSelection)
		{
			ClearSelection();
		}

		if(!bBlockSelection && m_iDragStartIndex != iIndex)
		{
			//doing a drag selection.
			//This is similar to the block selection from holding shift down , but the start point is determined
			//by the start of the drag
			bDragSelection = true;
		}

		if(bBlockSelection || bDragSelection)
		{
			bool bSelect = false;
			if(bBlockSelection)
			{
				//If the anchor is selected ,or the selection is being replaced , then select everything between the anchor
				//and the clicked item. Otherwise deselect.
				//This doesn't quite mimic windows behaviour ,but I haven't quite worked out what it does in all cases for this.
				//RAR 07/08/2006
				bSelect  = bReplaceSelection || IsSelected(m_iAnchorIndex);
			}
			else
			{
				bSelect  = bReplaceSelection || !IsSelected(m_iDragStartIndex);
			}
			int iStartIndex = bDragSelection ? m_iDragStartIndex : m_iAnchorIndex;


			int iMinIndex = Asura_Maths::Min(iIndex,iStartIndex);
			int iMaxIndex = Asura_Maths::Max(iIndex,iStartIndex);

			for(int i=iMinIndex;i<=iMaxIndex;i++)
			{
				SetSelected(i,bSelect);
			}
		}
		else
		{
			//toggle selection of item clicked on , and change the anchor to this point
			SetSelected(iIndex,!IsSelected(iIndex));
			m_iAnchorIndex = iIndex;
		}

		m_iDragStartIndex = -1;

		if(m_pbChangedFlag) *m_pbChangedFlag = true;
	}
}

void Asura_List_Dialog_SelectionHandler::PreGadgetUpdate()
{
	m_bUpdatedThisFrame = false;
}

void Asura_List_Dialog_SelectionHandler::PostGadgetUpdate()
{
	if(!m_bUpdatedThisFrame && m_iMouseIndex != -1 && m_iDragStartIndex != -1)
	{
		//Not had an update , so the moust wasn't over any of the text gadgets
		//Update using the most recent index , so releasing the mouse off the list will end a drag selection correctly
		Update(m_iMouseIndex,true);
	}
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////


Asura_Button_Dialog::Asura_Button_Dialog(const Asura_Vector_2& offset,int textureIndex,int textureSize,Asura_Vector_2 buttonPixelPos,Asura_Vector_2 buttonPixelSize,void(*clickedFunction)(void) /*=0*/)
: Asura_Dialog(0,offset)
{
	Asura_Gadget* pxnewGadget = new Asura_Button_Gadget(offset,textureIndex,textureSize,buttonPixelPos,buttonPixelSize,clickedFunction);
	AddGadget(pxnewGadget);

	m_xMaxCoord.x = m_xOffset.x + pxnewGadget->m_fWidth;
	m_xMaxCoord.y = m_xOffset.y + pxnewGadget->m_fHeight;
}

Asura_Button_Dialog::Asura_Button_Dialog(const Asura_Vector_2& offset,int textureIndex,int textureSize,Asura_Vector_2 buttonPixelPos,Asura_Vector_2 buttonPixelSize,void(*clickedFunction)(void*),void *pContext)
: Asura_Dialog(0,offset)
{
	Asura_Gadget* pxnewGadget = new Asura_Button_Gadget(offset,textureIndex,textureSize,buttonPixelPos,buttonPixelSize,clickedFunction,pContext);
	AddGadget(pxnewGadget);

	m_xMaxCoord.x = m_xOffset.x + pxnewGadget->m_fWidth;
	m_xMaxCoord.y = m_xOffset.y + pxnewGadget->m_fHeight;
}



Asura_DynamicList_Dialog::Asura_DynamicList_Dialog(const Asura_Vector_2& xOffset, char** ppcList, int* piListOffset, int* piSelection, int iNumberOfLines, int* piNumberInList, void(*pxFunction)(void)) : Asura_Dialog(0,xOffset)
{
	m_iNumberOfLines = iNumberOfLines;
	m_piNumberInList = piNumberInList;
	m_piListOffset = piListOffset;
	m_ppcList = ppcList;
	m_piSelection = piSelection;
	m_pxFunction = pxFunction;

	m_xOffset = xOffset;
	Asura_Vector_2 xPos = m_xOffset;

	Asura_Gadget* pxNewGadget;

	bool bNeedToBeScrollable = false;

	if (*m_piNumberInList > m_iNumberOfLines) bNeedToBeScrollable = true;

	if (bNeedToBeScrollable)
	{
		pxNewGadget = new Asura_UpDown_Gadget(xPos, (int*)m_piListOffset, 0, *m_piNumberInList - m_iNumberOfLines, -1);
		AddGadget(pxNewGadget);
	}

	//find the longest string
	int iMaxLength = 1; //arbitrary minimum starting point
	for( int i=0; i < *m_piNumberInList; i++ )
	{
		if( m_ppcList[i] )
		{			
			int iLength = strlen(m_ppcList[i]);

			if(iLength > iMaxLength) 
			{
				iMaxLength = iLength;
			}			
		}
	}

	int iMaxAllowed = (Asura_Viewport::GetWidth()/10)-10;
	if(iMaxLength > iMaxAllowed) 
	{
		iMaxLength = iMaxAllowed;
	}

	xPos.x += 2.0f;

	int iNumberOfGadgets = *m_piNumberInList;
	if( iNumberOfGadgets > m_iNumberOfLines )
	{
		iNumberOfGadgets = m_iNumberOfLines;
	}
	for (int i=0; i < iNumberOfGadgets; i++)
	{
		pxNewGadget  = new Asura_TextSelect_Gadget(xPos, m_piSelection, i, m_piListOffset, m_ppcList, m_iNumberOfLines, *m_piNumberInList, m_pxFunction);
		pxNewGadget->m_fHeight = 1.2f;//increase the height so it isn't possible for the cursor to fall in a gap between lines
		pxNewGadget->m_fWidth = (float)iMaxLength;//set the width of all the gadgets according to the maximum string length
		AddGadget(pxNewGadget);
		xPos.y += 1.2f;
	}


	if (bNeedToBeScrollable)
	{
		xPos.y -= 1.2f;	
		xPos.x -= 2.0f;

		pxNewGadget  = new Asura_UpDown_Gadget(xPos, m_piListOffset, 0, *m_piNumberInList - m_iNumberOfLines, 1 );
		AddGadget(pxNewGadget);
	}



	m_xMaxCoord.x = xPos.x + iMaxLength + 2;
	m_xMaxCoord.y = xPos.y + 1.2f * m_iNumberOfLines;
}


void Asura_DynamicList_Dialog::Render( const Asura_Vector_2& xOrigin )
{
	if( !IsHidden() )
	{
		if( m_pxGadgetList ) { m_pxGadgetList->RenderList( xOrigin ); }
	}

}

void Asura_DynamicList_Dialog::Update( const Asura_Vector_2& xOrigin )
{
	{
		delete m_pxGadgetList;
		m_pxGadgetList = NULL;

		Asura_Vector_2 xPos = m_xOffset;

		Asura_Gadget* pxNewGadget;

		bool bNeedToBeScrollable = false;

		if (*m_piNumberInList > m_iNumberOfLines) bNeedToBeScrollable = true;

		if (bNeedToBeScrollable)
		{
			pxNewGadget = new Asura_UpDown_Gadget(xPos, (int*)m_piListOffset, 0, *m_piNumberInList - m_iNumberOfLines, -1);
			AddGadget(pxNewGadget);
		}

		//find the longest string
		int iMaxLength = 1; //arbitrary minimum starting point
		for( int i=0; i < *m_piNumberInList; i++ )
		{
			if( m_ppcList[i] )
			{	
				int iLength = strlen(m_ppcList[i]);

				if(iLength > iMaxLength) 
				{
					iMaxLength = iLength;
				}			
			}
		}

		int iMaxAllowed = (Asura_Viewport::GetWidth()/10)-10;
		if(iMaxLength > iMaxAllowed) 
		{
			iMaxLength = iMaxAllowed;
		}

		xPos.x += 2.0f;

		int iNumberOfGadgets = *m_piNumberInList;
		if( iNumberOfGadgets > m_iNumberOfLines )
		{
			iNumberOfGadgets = m_iNumberOfLines;
		}

		for (int i=0; i < iNumberOfGadgets; i++)
		{
			pxNewGadget  = new Asura_TextSelect_Gadget(xPos, m_piSelection, i, m_piListOffset, m_ppcList, m_iNumberOfLines, *m_piNumberInList, m_pxFunction);
			pxNewGadget->m_fHeight = 1.2f;//increase the height so it isn't possible for the cursor to fall in a gap between lines
			pxNewGadget->m_fWidth = (float)iMaxLength;//set the width of all the gadgets according to the maximum string length
			AddGadget(pxNewGadget);
			xPos.y += 1.2f;
		}

		if (bNeedToBeScrollable)
		{
			xPos.y -= 1.2f;	
			xPos.x -= 2.0f;

			pxNewGadget  = new Asura_UpDown_Gadget(xPos, m_piListOffset, 0, *m_piNumberInList - m_iNumberOfLines, 1 );
			AddGadget(pxNewGadget);
		}

		m_xMaxCoord.x = xPos.x + iMaxLength + 2;
		m_xMaxCoord.y = xPos.y + 1.2f * m_iNumberOfLines;
	}

	Asura_Dialog::Update( xOrigin );
}

//
// Asura_CtrlTabList_Dialog.
//

DEBUGVAR float CTRLTABLIST_BORDER_X = 1.0f;
DEBUGVAR float CTRLTABLIST_BORDER_Y = 1.0f;

Asura_CtrlTabList_Dialog::Asura_CtrlTabList_Dialog( u_int uMaxDisplayedEntries )
: Asura_Dialog( NULL, Asura_Vector_2(CTRLTABLIST_BORDER_X, CTRLTABLIST_BORDER_Y) )
, m_uMaxDisplayedEntries( uMaxDisplayedEntries )
{
	Asura_Assert( m_uMaxDisplayedEntries, "" );
	Asura_Assert( m_uMaxDisplayedEntries & 1, "Display looks better (ie. symmetric) when max displayed entries is odd." );

	const float fCharWidth    = 1.0f;
	const float fCharHeight   = 1.0f;
	const float fCharSpacingY = 0.1f;
	m_xMaxCoord.x = m_xOffset.x + 2.0f * CTRLTABLIST_BORDER_X + MAX_ENTRY_NAME_LEN * fCharWidth;
	m_xMaxCoord.y = m_xOffset.x + 2.0f * CTRLTABLIST_BORDER_Y + m_uMaxDisplayedEntries * fCharHeight + (m_uMaxDisplayedEntries - 1) * fCharSpacingY;

	m_ppxDisplayedListEntries = new Asura_StringDisplay_Gadget*[m_uMaxDisplayedEntries];

	Asura_Vector_2 xOffset( m_xOffset.x + CTRLTABLIST_BORDER_X, m_xOffset.y + CTRLTABLIST_BORDER_Y );
	for( u_int u = 0; u < m_uMaxDisplayedEntries; ++u )
	{
		m_ppxDisplayedListEntries[u] = new Asura_StringDisplay_Gadget( xOffset, MAX_ENTRY_NAME_LEN );
		xOffset.y += fCharHeight + fCharSpacingY;
	}

	ForgetOverlayList();
}

Asura_CtrlTabList_Dialog::~Asura_CtrlTabList_Dialog( void )
{
	if( m_ppxDisplayedListEntries )
	{
		for( u_int u = m_uMaxDisplayedEntries; u--; )
		{
			if( m_ppxDisplayedListEntries[u] )
			{
				delete m_ppxDisplayedListEntries[u];
			}
		}

		delete[] m_ppxDisplayedListEntries;
	}
}

void Asura_CtrlTabList_Dialog::ForgetOverlayList( void )
{
	m_ppxOverlayList  = NULL;
	m_uOverlayListLen = 0;
}

void Asura_CtrlTabList_Dialog::Render( const Asura_Vector_2& xOrigin )
{
	if( !IsHidden() )
	{
		Asura_Assert( m_ppxOverlayList && m_uOverlayListLen, "Did you forget to call SetOverlayList()?" );

		for( u_int u = m_uNumDisplayedEntries; u--; )
		{
			m_ppxDisplayedListEntries[u]->Render( xOrigin );
		}
	}

}

void Asura_CtrlTabList_Dialog::SetOverlayList( Asura_Overlay** ppxOverlayList, u_int uOverlayListLen )
{
	Asura_Assert( ppxOverlayList && uOverlayListLen, "" );
	m_ppxOverlayList  = ppxOverlayList;
	m_uOverlayListLen = uOverlayListLen;
}

void Asura_CtrlTabList_Dialog::SetSelectedEntry( u_int uEntry )
{
	Asura_Assert( m_ppxOverlayList && m_uOverlayListLen, "" );
	Asura_Assert( uEntry < m_uOverlayListLen, "" );

	const u_int uHalfNumEntries = (m_uMaxDisplayedEntries >> 1);

	int iFirstEntryToDisplay = uEntry - uHalfNumEntries;
	int iLastEntryToDisplay  = uEntry + uHalfNumEntries;

	m_uNumDisplayedEntries = (iLastEntryToDisplay - iFirstEntryToDisplay) + 1;
	Asura_Assert( m_uNumDisplayedEntries <= m_uMaxDisplayedEntries, "" );

	// Update displayed pszList entries' labels.
	u_int uOutputEntry = 0;
	for( int iInputEntry = iFirstEntryToDisplay; iInputEntry <= iLastEntryToDisplay; ++iInputEntry, ++uOutputEntry )
	{
		Asura_Assert( uOutputEntry < m_uMaxDisplayedEntries, "" );

		// Set label.
		int iModInputEntry    = Asura_Maths::Abs(iInputEntry) % m_uOverlayListLen; 
		int iActualInputEntry = ( iInputEntry < 0 ? m_uOverlayListLen - iModInputEntry : iModInputEntry );
		m_ppxDisplayedListEntries[uOutputEntry]->SetLabel( m_ppxOverlayList[iActualInputEntry]->m_xName.GetString() );

		// Highlight?
		m_ppxDisplayedListEntries[uOutputEntry]->SetHighlight( iInputEntry == static_cast<int>(uEntry) );
	}
}

void Asura_CtrlTabList_Dialog::Update( const Asura_Vector_2& xOrigin )
{
}

/* Dynamic Dialog */
Asura_Dynamic_Dialog::Asura_Dynamic_Dialog( Asura_String* pszName, Asura_Vector_2 xOffset ) : Asura_Dialog( pszName->GetString(), xOffset )
{
	m_pszName = pszName;
}

void Asura_Dynamic_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	m_xName = m_pszName->GetString();
}

const Asura_Vector_2& Asura_Dynamic_Dialog::GetMaxCoords( void ) const 
{ 
	m_xMaxCoord.x = m_xOffset.x + (m_pszName->GetString() ? Asura_Text_Output::WidthOfString(m_pszName->GetString(),ASURA_TEXT_FORMAT_SMALLFONT) : 0);
	return m_xMaxCoord; 
}

////////////////////////////////////////////////////////////////////////////////////////////////

Asura_Icon_Dialog::Asura_Icon_Dialog(const Asura_Vector_2& xOffset,int iTextureIndex,int iIconIndex)
:Asura_Dialog(0,xOffset)
{
	Asura_Gadget* pxGadget;
	pxGadget  = new Asura_Icon_Gadget(xOffset,iTextureIndex,iIconIndex);
	AddGadget(pxGadget);

	m_xMaxCoord.x = m_xOffset.x + pxGadget->m_fWidth;
	m_xMaxCoord.y = m_xOffset.y + pxGadget->m_fHeight;

}

void Asura_Icon_Dialog::SetClickedFunction(const Asura_Callback_Base<void>& xCallback)
{
	((Asura_Icon_Gadget*)m_pxGadgetList)->SetClickedFunction(xCallback);
}

void Asura_Icon_Dialog::SetClickedFlag(bool* pbClickedFlag)
{
	((Asura_Icon_Gadget*)m_pxGadgetList)->SetClickedFlag(pbClickedFlag);
}

void Asura_Icon_Dialog::SetCheckboxBool(bool* pbValue)
{
	((Asura_Icon_Gadget*)m_pxGadgetList)->SetCheckboxBool(pbValue);
}

void Asura_Icon_Dialog::SetCheckboxFlag(u_int* puFlagsField,u_int uFlagBit)
{
	((Asura_Icon_Gadget*)m_pxGadgetList)->SetCheckboxFlag(puFlagsField,uFlagBit);
}

void Asura_Icon_Dialog::SetCheckboxFlag(int* piFlagsField,u_int uFlagBit)
{
	((Asura_Icon_Gadget*)m_pxGadgetList)->SetCheckboxFlag((u_int*)piFlagsField,uFlagBit);
}

void Asura_Icon_Dialog::SetCheckboxFunction(const Asura_Callback_Base<bool>& xIsChecked, const Asura_Callback_Base_1Arg<void,bool>& xSetChecked)
{
	((Asura_Icon_Gadget*)m_pxGadgetList)->SetCheckboxFunction(xIsChecked, xSetChecked);
}

/// Asura_ClickableString_Dialog ///////////////////////////////////////////////////////////////////////
Asura_ClickableString_Dialog::Asura_ClickableString_Dialog(const char* szString, const Asura_Vector_2& xOffset) :
	Asura_Dialog(NULL, xOffset),
	m_pfnClick(0),
	m_iUserData(0)
{
	if (szString)
	{
		const int   iLength = static_cast<int>( strlen( szString ) );
		const float fWidth  = Asura_Text_Output::WidthOfString( szString, ASURA_TEXT_FORMAT_SMALLFONT );

		m_pxTextGadget = new Asura_StringDisplay_Gadget(xOffset, iLength, NULL, OnClick, (int)this, ASURA_TEXT_FORMAT_SMALLFONT);
		m_pxTextGadget->m_fWidth = fWidth;
		m_pxTextGadget->SetLabel(szString);
		m_xMaxCoord.x = m_pxTextGadget->m_xOffset.x + fWidth + 1.0f;
	}
	else
	{
		m_pxTextGadget = new Asura_StringDisplay_Gadget(xOffset, 0, NULL, OnClick, (int)this, ASURA_TEXT_FORMAT_SMALLFONT);
		m_xMaxCoord.x = m_pxTextGadget->m_xOffset.x + 1.0f;
	}
	AddGadget(m_pxTextGadget);

	m_xMaxCoord.y = m_pxTextGadget->m_xOffset.y + 1.0f;
}

void Asura_ClickableString_Dialog::SetString(const char* szString)
{
	m_pxTextGadget->SetLabel(szString);
	m_xMaxCoord.x = m_pxTextGadget->m_xOffset.x + Asura_Text_Output::WidthOfString(szString, ASURA_TEXT_FORMAT_SMALLFONT) + 1.0f;
}

void Asura_ClickableString_Dialog::SetClickFunction(void (*pfnClick)(int iUserData),int iUserData)
{
	m_pfnClick = pfnClick;
	m_iUserData = iUserData;
}

void Asura_ClickableString_Dialog::OnClick(int iArg)
{
	Asura_ClickableString_Dialog *pThis = (Asura_ClickableString_Dialog*)iArg;

	if (pThis->m_pfnClick)
	{
		Asura_Overlay_System::SetCallbackFunctionToCall(pThis->m_pfnClick, pThis->m_iUserData);
	}
}



/// Asura_DropDown_List_Dialog ////////////////////////////////////////////////////////////////////

float Asura_DropDown_List_Dialog::s_fDropListSpacing		= 0.2f;
float Asura_DropDown_List_Dialog::s_fDropListLineOffsetY	= 0.1f;
float Asura_DropDown_List_Dialog::s_fDropListLineAdjustX1	= -2.0f;
float Asura_DropDown_List_Dialog::s_fDropListLineAdjustX2	= -1.0f;
Asura_Vector_2 Asura_DropDown_List_Dialog::s_xDropListScrollBarOffset( 0.0f, 0.2f );
float Asura_DropDown_List_Dialog::s_fCaptionSpacing			= 1.0f;

Asura_DropDown_List_Dialog::Asura_DropDown_List_Dialog(
	const Asura_Vector_2& xOffset,
	const char* const* pszList,
	int* piListOffset,
	int* piListSelection,
	int iNumberOfLines,
	int iNumberInList,
	void( *pfFunction )( void ),
	const char* szName /*= NULL*/,
	const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0, xOffset, 1)
, m_iMaxNumberOfLines(iNumberOfLines)
, m_szName(szName)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_fMinWidth = 1.0f;
	m_piSelectionIndex = 0;
	m_puSelectionIndex = 0;
	m_pxSharedSelectionIndex = 0;
	m_iSelectionIndexBackup = 0;
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = 0;
	m_bOnlyUpdateSelectionIndexOnClick = false;
	SetOnSelectCallback(pfFunction);
	m_szInput[0] = 0;
	m_fMaxWidth = -1.0f;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;
	m_bSortStrings = false;
	m_uStartSortOffset = 0;

	int* const piListOffsetScroll = ( m_bOnlySelectOnClick ) ? &m_iListOffsetScroll : piListOffset;
	CreateGadgets( xOffset, pszList, piListOffset, piListOffsetScroll, piListSelection, iNumberOfLines, iNumberInList );
}

Asura_DropDown_List_Dialog::Asura_DropDown_List_Dialog( 
	const Asura_Vector_2& xOffset,
	const char* const* ppszList,
	int* piSelectionIndex,
	int iNumberOfLines,
	int iNumberInList,
	void( *pfFunction )( void ),
	bool* pbChangedFlag /*= 0*/,
	const char* szName /*= NULL*/,
	const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0, xOffset, 1)
, m_iMaxNumberOfLines(iNumberOfLines)
, m_szName(szName)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_fMinWidth = 1.0f;
	m_piSelectionIndex = piSelectionIndex;
	m_puSelectionIndex = 0;
	m_pxSharedSelectionIndex = 0;
	m_iSelectionIndexBackup = GetExternalIndex();
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = pbChangedFlag;
	m_bOnlyUpdateSelectionIndexOnClick = false;
	SetOnSelectCallback(pfFunction);
	m_szInput[0] = 0;
	m_fMaxWidth = -1.0f;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;
	m_bSortStrings = false;
	m_uStartSortOffset = 0;

	int* const piListOffsetScroll = ( m_bOnlySelectOnClick ) ? &m_iListOffsetScroll : &m_iListOffset;
	CreateGadgets( xOffset, ppszList, &m_iListOffset, piListOffsetScroll, &m_iListSelection, iNumberOfLines, iNumberInList );

	SelectEntry( m_iSelectionIndexBackup, false );
}


Asura_DropDown_List_Dialog::Asura_DropDown_List_Dialog( 
	const Asura_Vector_2& xOffset,
	const char* const* ppszList,
	u_int* puSelectionIndex,
	int iNumberOfLines,
	int iNumberInList,
	void( *pfFunction )( void ),
	bool* pbChangedFlag /*= 0*/,
	const char* szName /*= NULL*/,
	const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0, xOffset, 1)
, m_iMaxNumberOfLines(iNumberOfLines)
, m_szName(szName)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_fMinWidth = 1.0f;
	m_piSelectionIndex = 0;
	m_puSelectionIndex = puSelectionIndex;
	m_pxSharedSelectionIndex = 0;
	m_iSelectionIndexBackup = GetExternalIndex();
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = pbChangedFlag;
	m_bOnlyUpdateSelectionIndexOnClick = false;
	SetOnSelectCallback(pfFunction);
	m_szInput[0] = 0;
	m_fMaxWidth = -1.0f;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;
	m_bSortStrings = false;
	m_uStartSortOffset = 0;

	int* const piListOffsetScroll = ( m_bOnlySelectOnClick ) ? &m_iListOffsetScroll : &m_iListOffset;
	CreateGadgets( xOffset, ppszList, &m_iListOffset, piListOffsetScroll, &m_iListSelection, iNumberOfLines, iNumberInList );

	SelectEntry( m_iSelectionIndexBackup, false );
}

Asura_DropDown_List_Dialog::Asura_DropDown_List_Dialog( 
	const Asura_Vector_2& xOffset,
	const char* const* ppszList,
	Asura_Dialog_SharedValue< int >* pxSharedSelectionIndex,
	int iNumberOfLines,
	int iNumberInList,
	void( *pfFunction )( void ),
	bool* pbChangedFlag /*= 0*/,
	const char* szName /*= NULL*/,
	const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0, xOffset, 1)
, m_iMaxNumberOfLines(iNumberOfLines)
, m_szName(szName)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_fMinWidth = 1.0f;
	m_piSelectionIndex = 0;
	m_puSelectionIndex = 0;
	m_pxSharedSelectionIndex = pxSharedSelectionIndex;
	m_iSelectionIndexBackup = GetExternalIndex();
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = pbChangedFlag;
	m_bOnlyUpdateSelectionIndexOnClick = false;
	SetOnSelectCallback(pfFunction);
	m_szInput[0] = 0;
	m_fMaxWidth = -1.0f;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;
	m_bSortStrings = false;
	m_uStartSortOffset = 0;

	int* const piListOffsetScroll = ( m_bOnlySelectOnClick ) ? &m_iListOffsetScroll : &m_iListOffset;
	CreateGadgets( xOffset, ppszList, &m_iListOffset, piListOffsetScroll, &m_iListSelection, iNumberOfLines, iNumberInList );

	SelectEntry( m_iSelectionIndexBackup, false );
}


Asura_DropDown_List_Dialog::Asura_DropDown_List_Dialog( const Asura_Vector_2& xOffset, const bool bOnlySelectOnClick /*= false*/ )
: Asura_Dialog(0, xOffset, 1)
, m_bOnlySelectOnClick( bOnlySelectOnClick )
{
	m_fMinWidth = 1.0f;
	m_bOnlyUpdateSelectionIndexOnClick = false;
	m_fMaxWidth = -1.0f;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;
	m_bSortStrings = false;
	m_uStartSortOffset = 0;
}

Asura_DropDown_List_Dialog::~Asura_DropDown_List_Dialog()
{
	delete [] m_piToSortedIndex;
	delete [] m_piFromSortedIndex;
}

void Asura_DropDown_List_Dialog::Initialise( const Asura_Vector_2& xOffset,
											 const char* const* ppszList,
											 int* piSelectionIndex,
											 int iNumberOfLines, int iNumberInList,
											 void(*pfFunction)(void),
											 bool* pbChangedFlag /*= 0*/,
											 const char* szName /*= NULL*/ )
{
	m_iMaxNumberOfLines = iNumberOfLines;
	m_szName = szName;
	SetOnSelectCallback( pfFunction );

	m_piSelectionIndex = piSelectionIndex;
	m_puSelectionIndex = 0;
	m_pxSharedSelectionIndex = 0;
	m_iSelectionIndexBackup = *piSelectionIndex;
	m_iListOffset = 0;
	m_iListOffsetScroll = 0;
	m_iListSelection = 0;
	m_pbChangedFlag = pbChangedFlag;
	m_szInput[0] = 0;
	m_fMaxWidth = -1.0f;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;
	m_bSortStrings = false;
	m_uStartSortOffset = 0;

	int* const piListOffsetScroll = ( m_bOnlySelectOnClick ) ? &m_iListOffsetScroll : &m_iListOffset;
	CreateGadgets( xOffset, ppszList, &m_iListOffset, piListOffsetScroll, &m_iListSelection, iNumberOfLines, iNumberInList );

	SelectEntry( m_iSelectionIndexBackup, false );
}

void Asura_DropDown_List_Dialog::CreateGadgets( const Asura_Vector_2& offset,
											    const char* const* ppszList,
												int* piListOffset,
												int* piListOffsetScroll,
												int* piSelection,
												int iNumberOfLines, int iNumberInList )
{
	m_pxUpGadget = 0;
	m_pxDownGadget = 0;
	m_pxScrollGadget = 0;

	m_piListOffset = piListOffset;
	m_piListSelection = piSelection;
	m_pszUnsortedStringList = ppszList;
	m_iNumberInList = iNumberInList;

	Asura_Assert( NULL != m_piListOffset, "Offset Pointer is NULL");
	Asura_Assert( NULL != m_piListSelection, "Selection Pointer is NULL");
	Asura_Assert( 0 <= (*m_piListOffset), "Offset Value is negative!");
	Asura_Assert( 0 <= (*m_piListSelection), "Selection Value is negative!");
	Asura_Assert( 0 <= (*m_piListOffset + *m_piListSelection), "List Index is negative!");
	Asura_Assertf( (*m_piListOffset + *m_piListSelection) < m_iNumberInList, ( "Trying to set list index %d, only %d in list!", (*m_piListOffset + *m_piListSelection), m_iNumberInList ) );

	Asura_Gadget* pxGadget;
	Asura_Vector_2 xOffset = offset;

	m_bNeedsScrollBar = false;
	if (iNumberInList > iNumberOfLines) m_bNeedsScrollBar = true;

	//find the longest string
	int iMaxLength = 1;
	float fMaxWidth = m_fMinWidth;
	for (int i = 0; i < iNumberInList; i++)
	{
		if( ppszList[i] )
		{
			iMaxLength = Asura_Maths::Max(static_cast<int>(strlen(ppszList[i])), iMaxLength);
			fMaxWidth = Asura_Maths::Max(Asura_Text_Output::WidthOfString(ppszList[i], ASURA_TEXT_FORMAT_SMALLFONT), fMaxWidth);
		}		
	}
	float fMaxAllowed = (float)((Asura_Viewport::GetWidth()/10)-10);
	if(fMaxWidth>fMaxAllowed) fMaxWidth = fMaxAllowed;

	pxGadget = m_pxSelectedGadget = new Asura_StringDisplay_Gadget(xOffset, iMaxLength, NULL, NULL, 0, ASURA_TEXT_FORMAT_SMALLFONT);

	char szBuffer[256];

	if(GetCroppedString(szBuffer, m_szInput, 256))
	{
		m_pxSelectedGadget->SetLabel(szBuffer);
	}
	else
	{
		m_pxSelectedGadget->SetLabel(ppszList[GetInternalIndex_Unsorted()]);
	}

	m_pxSelectedGadget->m_fWidth = fMaxWidth;
	AddGadget(pxGadget);


	Asura_Vector_2 xRightOffset(xOffset);
	xRightOffset.x += fMaxWidth + 0.5f;
	pxGadget = m_pxDropDownGadget = new Asura_UpDown_Gadget(xRightOffset, &m_iDropDown, 0, 1, 1);
	AddGadget(pxGadget);

	xOffset.y += 1.f;
	xRightOffset.y += 1.f;

//	if (m_bNeedsScrollBar)
	{
		pxGadget = m_pxUpGadget = new Asura_UpDown_Gadget( xRightOffset + s_xDropListScrollBarOffset, piListOffsetScroll, 0, iNumberInList - iNumberOfLines, -1);
		AddGadget(pxGadget);
	}

	if(iNumberOfLines >= 4)
		//	if(m_bNeedsScrollBar && iNumberOfLines >= 4)
	{
		//add scroll bar
		xRightOffset.y += 1.2f;
		float fHeight = 1.2f * (iNumberOfLines - 2);
		pxGadget = m_pxScrollGadget = new Asura_VerticalScroll_Gadget( xRightOffset + s_xDropListScrollBarOffset, fHeight, iNumberOfLines, piListOffsetScroll, 0, Asura_Maths::Max(iNumberInList - iNumberOfLines, 0));
		AddGadget(pxGadget);
	}	

	for (int i=0; i < iNumberOfLines; i++)
	{
		if( i == 0 )
		{
			xOffset.y += s_fDropListSpacing;
		}
		pxGadget  = new Asura_TextSelect_Gadget(xOffset, m_piListSelection, i, m_piListOffset, ppszList, iNumberOfLines, iNumberInList, NULL, m_bOnlySelectOnClick );
		pxGadget->m_fHeight = 1.2f;//increase the height so it isn't possible for the cursor to fall in a gap between lines
		pxGadget->m_fWidth = fMaxWidth;//set the width of all the gadgets according to the maximum string length
		if(i >= iNumberInList) pxGadget->SetHidden(true);

		AddGadget(pxGadget);
		xOffset.y += 1.2f;
	}
	// store the pointer of the first text gadget in the linked pszList
	m_pxFirstTextGadget = (Asura_TextSelect_Gadget*)m_pxGadgetList;
	m_iNumberOfLines = Asura_Maths::Min(iNumberOfLines, iNumberInList);

//	if (m_bNeedsScrollBar)
	{
		xOffset.y -= 1.2f;	
		xOffset.x += fMaxWidth + 0.5f;

		pxGadget  = m_pxDownGadget = new Asura_UpDown_Gadget( xOffset + s_xDropListScrollBarOffset, piListOffsetScroll, 0, iNumberInList - iNumberOfLines, 1);
		AddGadget(pxGadget);
	}
	m_xMaxCoord.x = m_xOffset.x + fMaxWidth + 2.f + s_fCaptionSpacing;
	if(m_szName) m_xMaxCoord.x += Asura_Text_Output::WidthOfString(m_szName, ASURA_TEXT_FORMAT_SMALLFONT);
	m_xMaxCoord.y = m_xOffset.y + 1.2f; //Only set MaxCoord to include dialog in non-dropped down state. RAR 02/06/2005

	// Ensure drop down is hidden by setting on then toggling off.
	m_bIsDroppedDown = true;
	ToggleDropDown();

}

void Asura_DropDown_List_Dialog::SetOnSelectCallback(void(*pfnOnSelect)(void))
{
	m_pfOnSelectCallbackVoid = pfnOnSelect;
	m_pfOnSelectCallbackInt = 0;
}

void Asura_DropDown_List_Dialog::SetOnSelectCallback(void(*pfnOnSelect)(int),int iUserData)
{
	m_pfOnSelectCallbackVoid = 0;
	m_pfOnSelectCallbackInt = pfnOnSelect;
	m_iOnSelectCallbackUserData = iUserData;
}

void Asura_DropDown_List_Dialog::SetListHelpStrings(const char* const * ppszHelpText)
{
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;

	//give the list of help strings to each of the text select gadgets
	for(int i=0;i<m_iNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->SetListHelpStrings(ppszHelpText);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}
}

void Asura_DropDown_List_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	bool bSharedValue = IsSharedValue();
	bool bClippingUsed = false;
	float fClipX, fClipY, fClipWidth, fClipHeight;
	
	if(!IsHidden())
	{
		u_int uColour = GetColour();
		if(bSharedValue && ((uColour & 0xffffff) == ASURA_DIALOG_ENABLED_COLOUR))
		{
			uColour = Asura_OverlayBrightness | ASURA_GADGET_SHARED_COLOUR;
		}

		
		const float fOffsetX = 2.f;
		float fHeight = 1.f;

		/*
		if(m_bIsDroppedDown)
		{
			fHeight += ( m_iNumberOfLines * 1.2f ) + s_fDropListSpacing;
		}
		*/

		const float fNameWidth = Asura_Text_Output::WidthOfString(m_szName, ASURA_TEXT_FORMAT_SMALLFONT);
		Asura_Render::Overlay.Colour = uColour;
		Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH - (1.f + fOffsetX),
			(xOrigin.y + m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT - 1.f,
			(m_xMaxCoord.x - m_xOffset.x - fNameWidth - s_fCaptionSpacing) * ASURA_TEXT_OUTPUT_FONT_WIDTH + 2.f,
			fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT + 2.f);
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		Asura_Render::Overlay.Colour = Asura_OverlayBrightness;
		Asura_Render::Overlay.SetRectXY((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH - fOffsetX,
			(xOrigin.y + m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
			(m_xMaxCoord.x - m_xOffset.x - fNameWidth - s_fCaptionSpacing) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
			fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT);
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		if(m_bIsDroppedDown)
		{
			bClippingUsed = Asura_Render::Tools_GetClippingRectangle(fClipX, fClipY, fClipWidth, fClipHeight);

			if(bClippingUsed)
			{
				Asura_Render::AllowFullScreenRendering();
			}

			const float fMaxXCoord = (m_fMaxWidth > 0.0f) ? (m_xOffset.x + m_fFullWidth) : m_xMaxCoord.x;

			// Render dropped-down background

			float fOpenedHeight = (m_iNumberOfLines * 1.2f) + s_fDropListSpacing;

			Asura_Render::Overlay.Colour = uColour;
			Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
			Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
			Asura_Render::Overlay.SetRectXY((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH - (1.f + fOffsetX),
				(xOrigin.y + m_xOffset.y + fHeight) * ASURA_TEXT_OUTPUT_FONT_HEIGHT - 1.f,
				(fMaxXCoord - m_xOffset.x - fNameWidth - s_fCaptionSpacing) * ASURA_TEXT_OUTPUT_FONT_WIDTH + 2.f,
				fOpenedHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT + 2.f);
			Asura_Render::Overlay.Render();
			Asura_Render::FlushDisplay();

			Asura_Render::Overlay.Colour = Asura_OverlayBrightness;
			Asura_Render::Overlay.SetRectXY((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH - fOffsetX,
				(xOrigin.y + m_xOffset.y + fHeight) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
				(fMaxXCoord - m_xOffset.x - fNameWidth - s_fCaptionSpacing) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
				fOpenedHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT);
			Asura_Render::Overlay.Render();
			Asura_Render::FlushDisplay();

			// Render line

			const float fX1 = ( ( xOrigin.x + m_xOffset.x ) * ASURA_TEXT_OUTPUT_FONT_WIDTH ) + -2.0f;
			const float fY1 = ( ( xOrigin.y + m_xOffset.y + 1.0f + 0.1f ) * ASURA_TEXT_OUTPUT_FONT_HEIGHT );
			const float fX2 = ( ( ( xOrigin.x + fMaxXCoord ) - fNameWidth - s_fCaptionSpacing ) * ASURA_TEXT_OUTPUT_FONT_WIDTH ) + s_fDropListLineAdjustX2;
			const float fY2 = fY1;
			const unsigned int uColour = Asura_OverlayBrightness | 0x00fffffff;
			Asura_Render::Render2DLine( fX1, fY1, fX2, fY2, uColour );
		}
		if(m_szName)
		{
			Asura_Text_Output::Print(xOrigin.x + m_xMaxCoord.x, xOrigin.y + m_xOffset.y, GetColour(), ASURA_TEXT_FORMAT_SMALLFONT | ASURA_TEXT_FORMAT_RIGHTJUSTIFIED,"%s",m_szName);
		}
	}

	//update colour for drop down lists using Asura_Dialog_SharedValue
	m_pxSelectedGadget->SetUseMixedGadgetColour(IsMixedValue());
	m_pxDropDownGadget->SetUseSharedGadgetColour(bSharedValue);

	Asura_Dialog::Render(xOrigin);

	if(bClippingUsed)
	{
		Asura_Render::FlushDisplay();
		Asura_Render::Tools_SetClippingRectangle(fClipX, fClipY, fClipWidth, fClipHeight);
	}
}



void Asura_DropDown_List_Dialog::Update(const Asura_Vector_2& xOrigin)
{
	if(HasExternalIndex())
	{
		//RAR - Changes for internal handling of list offset and list selection values. 31/08/04

		//handle external changes to the selection
		const int iExternalIndex = GetExternalIndex();
		if( iExternalIndex != m_iSelectionIndexBackup )
		{
			SelectEntry( iExternalIndex );
		}
	}

	// Parent update
	int iCurrentOffset	= m_iListOffset;		// for OnlySelectOnClick
	m_iListOffsetScroll	= m_iListOffset;		// for OnlySelectOnClick
	Asura_Dialog::Update(xOrigin);

	if(m_bIsDroppedDown)
	{
		if(!Asura_Overlay_System::GetFocusGadget())
		{
			Asura_Overlay_System::TakeFocus(m_pxFirstTextGadget,OverlayFocusType_Keyboard);
		}

		HandleKeyboard();
	}

	if( m_bOnlySelectOnClick )
	{
		const int iDelta	= m_iListOffsetScroll - iCurrentOffset;
		m_iListSelection	-= iDelta;
		m_iListOffset		+= iDelta;
		m_iListOffsetScroll	= m_iListOffset;

		// capping should not be required, but just to be sure
		int iIndex;
		iIndex = m_iListOffset + m_iListSelection;
		if( iIndex >= m_iNumberInList )
		{
			m_iListSelection -= ( iIndex - ( m_iNumberInList - 1 ) );
		}
		iIndex = m_iListOffset + m_iListSelection;
		if( iIndex < 0 )
		{
			m_iListSelection -= iIndex;
		}
	}

	if(m_bIsDroppedDown)
	{
		char szBuffer[256];
		if(GetCroppedString(szBuffer, m_szInput, 256))
		{
			m_pxSelectedGadget->SetLabel(szBuffer);
		}
		else
		{
			m_pxSelectedGadget->SetLabel(m_szInput);
		}
	}
	else
	{
		char szBuffer[256];
		if(GetCroppedString(szBuffer, m_pszUnsortedStringList[GetInternalIndex_Unsorted()], 256))
		{
			m_pxSelectedGadget->SetLabel(szBuffer);
		}
		else
		{
			m_pxSelectedGadget->SetLabel(m_pszUnsortedStringList[GetInternalIndex_Unsorted()]);
		}
	}

	if( !m_bOnlyUpdateSelectionIndexOnClick )
	{
		//RAR - Changes for internal handling of list offset and list selection values. 31/08/04
		UpdateExternalIndex();
	}


	bool bToggleDropDown = false;
	
	if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) || 
		Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
	{

		Asura_Gadget* pxGadget = m_pxFirstTextGadget;
		for(; pxGadget; pxGadget = pxGadget->m_pxNextGadget)
		{
			if( pxGadget == m_pxUpGadget
				|| pxGadget == m_pxDownGadget
				|| pxGadget == m_pxScrollGadget)
			{
				continue;
			}
			if(pxGadget->IsMouseOver())
			{
				m_pxOverlay->m_bDialogTookFocus = true;
				ToggleDropDown();
				// SD: Need to call the callback after selection (Fixing the bug when moving the slider, the callback was called).
				if (!m_bIsDroppedDown) 
				{				
					//see if any selection change was made
					UpdateExternalIndex();

					if( m_pfOnSelectCallbackVoid )
					{
						Asura_Overlay_System::SetCallbackFunctionToCall(m_pfOnSelectCallbackVoid);
					}
					else if(m_pfOnSelectCallbackInt)
					{
						Asura_Overlay_System::SetCallbackFunctionToCall(m_pfOnSelectCallbackInt,m_iOnSelectCallbackUserData);
					}
				}

				return;
			}
		}
	}
	if(m_bIsDroppedDown)
	{
		Asura_Vector_2 xPos(xOrigin + m_xOffset);
		// If they've clicked outside the dialog, cancel the drop down
		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
		{
			const float fMaxXCoord = (m_fMaxWidth > 0.0f) ? (m_xOffset.x + m_fFullWidth) : m_xMaxCoord.x;

			if(!Asura_Cursor::Inside( xPos, fMaxXCoord - m_xOffset.x, 1.2f * m_iNumberOfLines + 1.f))
			{
				bToggleDropDown = true;

			}
			else
			{
				m_pxOverlay->m_bDialogTookFocus = true;
			}
		}

		//close drop down if enter is pressed while open
		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ENTER)) bToggleDropDown = true;
		
		//close drop down if it is disabled/hidden while dropped
		if(!m_bIsEnabled || IsHidden()) bToggleDropDown = true;
		
	}

	if(bToggleDropDown)
	{
		m_pxOverlay->m_bDialogTookFocus = true;
		ToggleDropDown();
		// SD: Need to call the callback after selection (Fixing the bug when moving the slider, the callback was called).
		if (!m_bIsDroppedDown) 
		{				
			//see if any selection change was made
			if(HasExternalIndex())
			{
				int iExternalIndex = GetExternalIndex();
				int iInternalIndex = GetInternalIndex_Unsorted();
				if(iExternalIndex != iInternalIndex)
				{
					SetExternalIndex(iInternalIndex);
					m_iSelectionIndexBackup = iInternalIndex;
					if(m_pbChangedFlag) *m_pbChangedFlag = true;
				}
			}
			//Don't trigger any callback functions if the click was done on the close overlay gadget.
			//This is to prevent issues with the callback being called after the overlay has been deleted.
			//RAR 10/09/2007
			if(!m_pxOverlay->IsClosing())
			{
				if( m_pfOnSelectCallbackVoid )
				{
					Asura_Overlay_System::SetCallbackFunctionToCall(m_pfOnSelectCallbackVoid);
				}
				else if(m_pfOnSelectCallbackInt)
   				{
					Asura_Overlay_System::SetCallbackFunctionToCall(m_pfOnSelectCallbackInt,m_iOnSelectCallbackUserData);
				}
			}
		}
	}

	if(m_bIsEnabled)
	{
		SetGadgetsHidden();
	}
}

bool Asura_DropDown_List_Dialog::GetCroppedString(char* pszDest, const char* pszSource, const u_int uDestSize)
{
	if(m_fMaxWidth > 0.0f)
	{
		strncpy(pszDest, pszSource, uDestSize);

		u_int uTextLength = strlen(pszDest);

		while(Asura_Text_Output::WidthOfString(pszDest, ASURA_TEXT_FORMAT_SMALLFONT) > m_fMaxWidth)
		{
			pszDest[uTextLength - 1] = 0x00;
			uTextLength--;
		}

		return true;
	}

	return false;
}

void Asura_DropDown_List_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	if(m_pxOverlay->IsClosing())
	{
		return;
	}

	if(HasExternalIndex())
	{
		//handle external changes to the selection
		const int iExternalIndex = GetExternalIndex();
		if(iExternalIndex != m_iSelectionIndexBackup)
		{
			SelectEntry(iExternalIndex);
		}
	}
}

void Asura_DropDown_List_Dialog::SetMaximumWidth(const float fMaxWidth)
{
	m_fFullWidth = (m_xMaxCoord.x - m_xOffset.x);
	
	m_fMaxWidth = fMaxWidth;
	
	if(m_pxSelectedGadget)
	{
		m_pxSelectedGadget->m_fWidth = Asura_Maths::Min(m_pxSelectedGadget->m_fWidth, m_fMaxWidth);
	}
}

void Asura_DropDown_List_Dialog::ToggleDropDown()
{
	m_bIsDroppedDown = !m_bIsDroppedDown;

	if(m_bIsDroppedDown)
	{
		Asura_Overlay_System::TakeFocus(m_pxFirstTextGadget,OverlayFocusType_Keyboard);
		m_szInput[0] = 0;
	}
	else
	{
		if(Asura_Overlay_System::GetFocusGadget() == m_pxFirstTextGadget)
		{
			Asura_Overlay_System::ReleaseFocus();
		}
		
		//update the label with the current selection
		int iIndex = GetInternalIndex_Unsorted();
		Asura_Assert(iIndex>=0 && iIndex<m_iNumberInList,"");
		if(iIndex>=0 && iIndex<m_iNumberInList)
		{
			m_pxSelectedGadget->SetLabel(m_pszUnsortedStringList[iIndex]);
		}
	}
	

	// Increase priority if dropped down
	if(m_bIsDroppedDown) m_iFocusPriority++;
	else m_iFocusPriority--;

	SetGadgetsHidden();
}

void Asura_DropDown_List_Dialog::SetGadgetsHidden()
{
	Asura_Gadget* pxTextGadget = m_pxFirstTextGadget;
	int iLine = m_iMaxNumberOfLines;
	while(pxTextGadget)
	{
		pxTextGadget->SetHidden((!m_bIsDroppedDown) || iLine > m_iNumberOfLines);
		pxTextGadget = pxTextGadget->m_pxNextGadget;
		iLine--;
	}
	if(m_pxUpGadget)		m_pxUpGadget->SetHidden(!(m_bIsDroppedDown && m_bNeedsScrollBar));
	if(m_pxDownGadget)		m_pxDownGadget->SetHidden(!(m_bIsDroppedDown && m_bNeedsScrollBar));
	if(m_pxScrollGadget)	m_pxScrollGadget->SetHidden(!(m_bIsDroppedDown && m_bNeedsScrollBar));
	m_pxDropDownGadget->SetHidden(false);
	m_pxSelectedGadget->SetHidden(false);
}

void Asura_DropDown_List_Dialog::ReplaceStrings(int iNumberInList,const char* const* pszList)
{
	m_iNumberInList = iNumberInList;
	m_pszUnsortedStringList = pszList;
	UpdateSortedStringList();
	m_iNumberOfLines = Asura_Maths::Min(iNumberInList, m_iMaxNumberOfLines);
	//get the max length for the new strings
	int iMaxLength = 1;
	float fMaxWidth = m_fMinWidth; 
	for (int i=0; i<iNumberInList; i++)
	{
		if( pszList[i] )
		{
			iMaxLength = Asura_Maths::Max(static_cast<int>(strlen(pszList[i])), iMaxLength);
			fMaxWidth = Asura_Maths::Max(Asura_Text_Output::WidthOfString(pszList[i], ASURA_TEXT_FORMAT_SMALLFONT), fMaxWidth);
		}
	}

	float fMaxClosedWidth = (m_fMaxWidth > 0.0f) ? Asura_Maths::Min(fMaxWidth, m_fMaxWidth) : fMaxWidth;

	float fMaxAllowed = (float)((Asura_Viewport::GetWidth()/10)-10);
	if(fMaxWidth>fMaxAllowed) fMaxWidth = fMaxAllowed;
	m_pxSelectedGadget->SetCharLimit(iMaxLength);
	m_pxSelectedGadget->m_fWidth = fMaxWidth;

	//update the text gadgets
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iMaxNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->ReplaceStrings(iNumberInList,m_pszUnsortedStringList);
		pxTextGadget->SetIndexRemapArray(m_bSortStrings ? m_piFromSortedIndex : 0);
		pxTextGadget->m_fWidth = fMaxWidth;//set the width of all the gadgets according to the maximum string length
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}


	///update the other gadgets
	int iMax = (iNumberInList > m_iNumberOfLines) ? (iNumberInList - m_iNumberOfLines) : 0;
	m_bNeedsScrollBar = (iMax > 0);
	float fNewOffsetX = m_xOffset.x + fMaxWidth + .5f;
	float fNewClosedOffsetX = m_xOffset.x + fMaxClosedWidth + .5f;
	if(m_pxUpGadget)
	{
		m_pxUpGadget->SetMax(iMax);
		m_pxUpGadget->m_xOffset.x = fNewOffsetX;
	}
	if(m_pxDownGadget)
	{
		m_pxDownGadget->SetMax(iMax);
		m_pxDownGadget->m_xOffset.x = fNewOffsetX;
	}
	if(m_pxScrollGadget)
	{
		m_pxScrollGadget->SetMinAndMax(0,iMax);
		m_pxScrollGadget->m_xOffset.x = fNewOffsetX;
	}
	if(m_pxDropDownGadget)
	{
		m_pxDropDownGadget->m_xOffset.x = fNewClosedOffsetX;
	}

	m_fFullWidth = fMaxWidth + 2.0f + s_fCaptionSpacing;

	m_xMaxCoord.x = m_xOffset.x + fMaxClosedWidth + 2.0f + s_fCaptionSpacing;
	if(m_szName) 
		m_xMaxCoord.x += Asura_Text_Output::WidthOfString(m_szName, ASURA_TEXT_FORMAT_SMALLFONT);

	//Ensure the selection is in range
	int iSelection = HasExternalIndex() ? GetExternalIndex() : GetInternalIndex_Unsorted();
	bool bNeedToUpdateExternalIndex = false;
	if(iSelection >= iNumberInList)
	{
		iSelection = FromSortedIndex(iNumberInList - 1);
		bNeedToUpdateExternalIndex = true;

	}
	//Select the selection again to force the label to be updated.
	SelectEntry(iSelection, bNeedToUpdateExternalIndex);

	if(m_fMaxWidth > 0.0f)
	{
		m_pxSelectedGadget->m_fWidth = Asura_Maths::Min(m_pxSelectedGadget->m_fWidth, m_fMaxWidth);
	}

	if(m_pxOverlay) 
		m_pxOverlay->ResizeOverlay();
}

void Asura_DropDown_List_Dialog::SelectEntry(int iSelection, bool bUpdateExternalIndex /*= true*/)
{
	Asura_Assert( NULL != m_piListOffset, "Offset Pointer is NULL");
	Asura_Assert( NULL != m_piListSelection, "Selection Pointer is NULL");

	int iSortedIndex = ToSortedIndex(iSelection);

	if(iSortedIndex < m_iNumberOfLines)
	{
		*m_piListOffset = 0;
		*m_piListSelection = iSortedIndex;
	}
	else
	{
		*m_piListOffset = iSortedIndex + 1 - m_iNumberOfLines;
		*m_piListSelection = m_iNumberOfLines - 1;
	}
	if(HasExternalIndex())
	{
		if(bUpdateExternalIndex)
		{
			SetExternalIndex(iSelection);
		}
		m_iSelectionIndexBackup = iSelection;
	}
	
	Asura_Assert( 0 <= (*m_piListOffset), "Offset Value is negative!");
	Asura_Assert( 0 <= (*m_piListSelection), "Selection Value is negative!");
	Asura_Assert( 0 <= (*m_piListOffset + *m_piListSelection), "List Index is negative!");

	int iInternalUnsortedIndex = GetInternalIndex_Unsorted();
	m_pxSelectedGadget->SetLabel( m_pszUnsortedStringList[ iInternalUnsortedIndex ] );
}

void Asura_DropDown_List_Dialog::SetMinimumWidth(float fWidth)
{
	m_fMinWidth = fWidth;
	//ReplaceStrings used to recalculate the current width
	ReplaceStrings(m_iNumberInList,m_pszUnsortedStringList);
}

float Asura_DropDown_List_Dialog::GetWidthForString(const char* szString)
{
	Asura_Assert(szString,"");
	return Asura_Text_Output::WidthOfString(szString, ASURA_TEXT_FORMAT_SMALLFONT); 
}

int Asura_DropDown_List_Dialog::GetExternalIndex() const
{
	if(m_piSelectionIndex)
	{
		return *m_piSelectionIndex;
	}
	if ( m_puSelectionIndex )
	{
		return static_cast< int >( *m_puSelectionIndex );
	}
	if(m_pxSharedSelectionIndex)
	{
		int iValue;
		m_pxSharedSelectionIndex->GetValue(&iValue,0);
		return iValue;
	}
	return 0;
}

void Asura_DropDown_List_Dialog::SetExternalIndex(int iIndex)
{
	if(m_piSelectionIndex)
	{
		*m_piSelectionIndex = iIndex;
	}
	if ( m_puSelectionIndex )
	{
		*m_puSelectionIndex = static_cast< u_int >( iIndex );
	}
	if(m_pxSharedSelectionIndex)
	{
		m_pxSharedSelectionIndex->SetValue(iIndex);
	}
}

void Asura_DropDown_List_Dialog::UpdateExternalIndex( void )
{
	if( HasExternalIndex() )
	{
		//see if any selection change was made
		const int iInternalIndex = GetInternalIndex_Unsorted();
		const int iExternalIndex = GetExternalIndex();
		if( iExternalIndex != iInternalIndex )
		{
			SetExternalIndex( iInternalIndex );
			m_iSelectionIndexBackup = iInternalIndex;
			if( m_pbChangedFlag )
			{
				*m_pbChangedFlag = true;
			}
		}
	}
}

bool Asura_DropDown_List_Dialog::IsMixedValue() const
{
	if(!m_pxSharedSelectionIndex)
	{
		return false;
	}

	bool bMixed;
	m_pxSharedSelectionIndex->GetValue( 0, &bMixed );
	return( bMixed );
}

int Asura_DropDown_List_Dialog::GetInternalIndex_Sorted( void ) const
{
	return( *m_piListOffset + *m_piListSelection );
}

int Asura_DropDown_List_Dialog::GetInternalIndex_Unsorted( void ) const
{
	return FromSortedIndex(*m_piListOffset + *m_piListSelection);
}

int Asura_DropDown_List_Dialog::ToSortedIndex(int iIndex) const
{
	if(!m_bSortStrings) return iIndex;
	Asura_Assert(m_piToSortedIndex, "");

	if(iIndex>=0 && iIndex<m_iNumberInList)
	{
		return m_piToSortedIndex[iIndex];
	}
	else
	{
		//Asserting on this for now , but it may be necessary to allow it
		Asura_Assert(0,"ToSortedIndex() - index out of range for drop down list");
		return iIndex;
	}
}

int Asura_DropDown_List_Dialog::FromSortedIndex(int iIndex) const
{
	if(!m_bSortStrings) return iIndex;
	Asura_Assert(m_piFromSortedIndex, "");

	if(iIndex>=0 && iIndex<m_iNumberInList)
	{
		return m_piFromSortedIndex[iIndex];
	}
	else
	{
		//Asserting on this for now , but it may be necessary to allow it
		Asura_Assert(0,"FromSortedIndex() - index out of range for drop down list");
		return iIndex;
	}
}


void Asura_DropDown_List_Dialog::HandleKeyboard()
{
	//Make use of the text select gadget keyboard shortcuts
	m_pxFirstTextGadget->HandleKeyboard();

	//check for keypresses for jumping to a different point in the list
	bool bAnyInput = false;
	if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_BACKSPACE))
	{
		int iLength = strlen(m_szInput);
		if(iLength)
		{
			m_szInput[iLength-1] = 0;
			bAnyInput = true;
		}
	}
	else
	{
		//Check the windows input buffer for the characters pressed
		u_int uInputBufferSize = Asura_Windows::GetSizeOfInputBuffer();
		const Asura_UnicodeChar* wszInputBuffer = Asura_Windows::GetInputBuffer();


		for(u_int i=0;i<uInputBufferSize;i++)
		{
			Asura_UnicodeChar wKeyPressed = wszInputBuffer[ i ];
			// ctrl-c (copy)?
			if( wKeyPressed == 0x0003 )
			{
				if( m_szInput[ 0 ] == 0 )
				{
					// we have nothing typed in, so copy the currently selected item to the clipboard
					int iInternalUnsortedIndex = GetInternalIndex_Unsorted();
					Asura_Windows::CopyToClipboard( m_pszUnsortedStringList[ iInternalUnsortedIndex ] );
				}
				else
				{
					// otherwise, copy the text we have typed in to the clipboard
					Asura_Windows::CopyToClipboard( m_szInput );
				}
			}
			// ctrl-v (paste)?
			if( wKeyPressed == 0x0016 )
			{
				const int iLength		= strlen( m_szInput );
				const int iRemaining	= ( sizeof( m_szInput ) - iLength ) - 1;
				if( Asura_Windows::CopyFromClipboard( m_szInput + iLength, iRemaining ) )
				{
					bAnyInput = true;
				}
			}

			char cKeyPressed = (char)wszInputBuffer[i];

			//Ignore some characters that we don't want. Mainly to avoid backspaces from being printed here.
			if(cKeyPressed >=0 && cKeyPressed < ' ') continue;

			int iLength = strlen(m_szInput);
			if(iLength<sizeof(m_szInput)-1)
			{
				m_szInput[iLength++] = cKeyPressed;
				m_szInput[iLength++] = 0;
				bAnyInput = true;
			}
		}
	}

	if(bAnyInput)
	{
		//Find the first string >= the input string.
		//There is no guarantee that the list is sorted , so its a case of checking all of them
		int iBestIndex = -1;
		const char* szBestString = 0;

		const char* const* pszStringList = m_pszUnsortedStringList;
		
		for(int i=0;i<m_iNumberInList;i++)
		{
			int iUnsortedIndex = FromSortedIndex(i);
			
			if(!pszStringList[iUnsortedIndex]) continue;

			if(stricmp(m_szInput,pszStringList[iUnsortedIndex]) <= 0)
			{
				//string is after the input string , but is it earlier than the previous one found?
				if(szBestString == 0 || stricmp(szBestString,pszStringList[iUnsortedIndex]) > 0)
				{
					iBestIndex = i;
					szBestString = pszStringList[iUnsortedIndex];
				}
			}
		}
		if(iBestIndex != -1)
		{
			//set the new selection
			m_pxFirstTextGadget->SetSelection(iBestIndex);
		}
	}

}

struct Asura_DropDownList_SortEntry
{
	const char* m_szString;
	int m_iOriginalIndex;
};

int Asura_DropDown_List_Dialog::SortStringCallback(const void* pEntry1, const void* pEntry2)
{
	const Asura_DropDownList_SortEntry* pxEntry1 = (const Asura_DropDownList_SortEntry*) pEntry1;
	const Asura_DropDownList_SortEntry* pxEntry2 = (const Asura_DropDownList_SortEntry*) pEntry2;
	return stricmp(pxEntry1->m_szString, pxEntry2->m_szString);
}

void Asura_DropDown_List_Dialog::UpdateSortedStringList()
{
	delete [] m_piToSortedIndex;
	delete [] m_piFromSortedIndex;
	m_piToSortedIndex = 0;
	m_piFromSortedIndex = 0;

	if(!m_bSortStrings) return;
	
	Asura_Assert(m_pszUnsortedStringList, "");
	Asura_Assert(m_iNumberInList>0, "");

	//fill in a temporary array for doing the sort
	Asura_DropDownList_SortEntry* pxSortArray = new Asura_DropDownList_SortEntry[m_iNumberInList];
	for(int i=0; i<m_iNumberInList; i++)
	{
		pxSortArray[i].m_szString = m_pszUnsortedStringList[i];
		pxSortArray[i].m_iOriginalIndex = i;
	}

	//Only sort items in the string list at or after m_uStartSortOffset.
	//Allows elements at the start of the list to remain in place even when then
	//list is sorted.
	Asura_SortAndSearch::Quicksort(&pxSortArray[m_uStartSortOffset], m_iNumberInList-m_uStartSortOffset, SortStringCallback);

	//copy the results  , and generate the mapping arrays
	m_piToSortedIndex = new int[m_iNumberInList];
	m_piFromSortedIndex = new int[m_iNumberInList];

	for(int i=0; i<m_iNumberInList; i++)
	{
		m_piFromSortedIndex[i] = pxSortArray[i].m_iOriginalIndex;
		m_piToSortedIndex[pxSortArray[i].m_iOriginalIndex] = i;
	}

	delete [] pxSortArray;

}

void Asura_DropDown_List_Dialog::SetSortStrings(bool bSort,u_int uStartSortOffset /* = 0 */)
{
	if(bSort == m_bSortStrings) return;

	int iSelection = GetInternalIndex_Unsorted();

	m_bSortStrings = bSort;

	//Only sort items in the string list at or after uStartSortOffset.
	m_uStartSortOffset = uStartSortOffset;

	UpdateSortedStringList();

	//update the text gadgets
	Asura_TextSelect_Gadget* pxTextGadget = m_pxFirstTextGadget;
	for(int i=0;i<m_iMaxNumberOfLines;i++)
	{
		Asura_Assert(pxTextGadget,"");

		pxTextGadget->ReplaceStrings(m_iNumberInList,m_pszUnsortedStringList);
		pxTextGadget->SetIndexRemapArray(m_bSortStrings ? m_piFromSortedIndex : 0);
		pxTextGadget = (Asura_TextSelect_Gadget*)(pxTextGadget->m_pxNextGadget);
	}

	//Re-select to update the internal sorted selection index
	SelectEntry(iSelection, false);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// image dialog
Asura_Image_Dialog::Asura_Image_Dialog( const Asura_Vector_2& xOffset, int iTextureIndex, float fWidth, float fHeight )
: Asura_Dialog( "", xOffset )
, m_iTextureIndex(iTextureIndex)
, m_uTileHash(ASURA_HASH_ID_UNSET)
, m_xSize( fWidth, fHeight )
, m_xTextureUV( 0.0f,0.0f,1.0f,1.0f )
, m_bTranslucent(true)
, m_uBackgroundColour(0)
, m_uBorderColour(0)
, m_uBorderPixels(0)
{
	m_xMaxCoord.x = xOffset.x + (fWidth  / ASURA_TEXT_OUTPUT_FONT_WIDTH);
	m_xMaxCoord.y = xOffset.y + (fHeight / ASURA_TEXT_OUTPUT_FONT_HEIGHT);
}

Asura_Image_Dialog::Asura_Image_Dialog( const Asura_Vector_2& xOffset, u_int uTileHash, float fWidth, float fHeight )
: Asura_Dialog( "", xOffset )
, m_iTextureIndex(ASURA_TEXTURE_NONE)
, m_uTileHash(uTileHash)
, m_xSize( fWidth, fHeight )
, m_xTextureUV( 0.0f,0.0f,1.0f,1.0f )
, m_bTranslucent(true)
, m_uBackgroundColour(0)
, m_uBorderColour(0)
, m_uBorderPixels(0)
{
	m_xMaxCoord.x = xOffset.x + (fWidth  / ASURA_TEXT_OUTPUT_FONT_WIDTH);
	m_xMaxCoord.y = xOffset.y + (fHeight / ASURA_TEXT_OUTPUT_FONT_HEIGHT);
}

Asura_Image_Dialog::~Asura_Image_Dialog()
{}

void Asura_Image_Dialog::SetTexture(int iTextureIndex)
{
	m_iTextureIndex = iTextureIndex;
	m_uTileHash = ASURA_HASH_ID_UNSET;
}

void Asura_Image_Dialog::Render( const Asura_Vector_2& xOrigin )
{
	
	if( !IsHidden() )
	{
		Asura_Vector_2 xPosition = xOrigin + m_xOffset;

		xPosition.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
		xPosition.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

		//draw a background and border
		if(m_uBackgroundColour)
		{
			Asura_Render::Overlay.SetRectXY( xPosition.x, xPosition.y, m_xSize.x, m_xSize.y );
			Asura_Render::Overlay.Colour = m_uBackgroundColour;		
			Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
			Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;

			Asura_Render::Overlay.Render();
		}

		if(m_uBorderColour && m_uBorderPixels)
		{
			Asura_Render::Overlay.Colour = m_uBorderColour;		
			Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;
			Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;

			//top
			Asura_Render::Overlay.SetRectXY( xPosition.x-m_uBorderPixels, xPosition.y-m_uBorderPixels, m_xSize.x+m_uBorderPixels*2, (float)m_uBorderPixels );
			Asura_Render::Overlay.Render();
			//bottom
			Asura_Render::Overlay.SetRectXY( xPosition.x-m_uBorderPixels, xPosition.y+m_xSize.y, m_xSize.x+m_uBorderPixels*2, (float)m_uBorderPixels );
			Asura_Render::Overlay.Render();
			//left
			Asura_Render::Overlay.SetRectXY( xPosition.x-m_uBorderPixels, xPosition.y,(float) m_uBorderPixels, m_xSize.y);
			Asura_Render::Overlay.Render();
			//right
			Asura_Render::Overlay.SetRectXY( xPosition.x+m_xSize.x, xPosition.y, (float)m_uBorderPixels, m_xSize.y);
			Asura_Render::Overlay.Render();

		}
		if(m_uBackgroundColour || (m_uBorderColour && m_uBorderPixels))
		{
			Asura_Render::FlushDisplay();
		}

		Asura_Render::Overlay.SetRectXY( xPosition.x, xPosition.y, m_xSize.x, m_xSize.y );

		const Asura_TiledTextureInfo* pxTile = Asura_TiledTextureManagement::GetTileInfoByHash(m_uTileHash);
		if( pxTile )		
		{
			Asura_Render::Overlay.SetRectUV( 0.0f, 0.0f, 0.1f, 1.0f );
			//Asura_Render::Overlay.SetRectUV( pxTile->m_fu1, pxTile->m_fv1, pxTile->m_fu2, pxTile->m_fv2 );

			Asura_Render::Overlay.U[0] = pxTile->m_fu1;
			Asura_Render::Overlay.U[1] = pxTile->m_fu2;
			Asura_Render::Overlay.U[2] = pxTile->m_fu2;
			Asura_Render::Overlay.U[3] = pxTile->m_fu1;

			Asura_Render::Overlay.V[0] = pxTile->m_fv1;
			Asura_Render::Overlay.V[1] = pxTile->m_fv1;
			Asura_Render::Overlay.V[2] = pxTile->m_fv2;
			Asura_Render::Overlay.V[3] = pxTile->m_fv2;

			Asura_Render::Overlay.TextureIndex = pxTile->m_iIndexOfTextureUsed;			
		}
		else
		{
			Asura_Render::Overlay.SetRectUV( m_xTextureUV.MinX,m_xTextureUV.MinY,m_xTextureUV.MaxX,m_xTextureUV.MaxY );
			Asura_Render::Overlay.TextureIndex = m_iTextureIndex;		
		}	

		Asura_Render::Overlay.Colour = 0xffffffff;		
		if(m_iTextureOp == ASURA_TEXTUREOP_INVALID)
		{
			Asura_Render::Overlay.TranslucencyMode = m_bTranslucent ? ASURA_TRANSLUCENCY_NORMAL : ASURA_TRANSLUCENCY_OFF;
			Asura_Render::Overlay.m_eTextureOp = m_bTranslucent ?  ASURA_TEXTUREOP_SINGLE_PASS : ASURA_TEXTUREOP_VIEWTEXTURECOLOUR;
		}
		else
		{
			Asura_Render::Overlay.TranslucencyMode = m_bTranslucent ? ASURA_TRANSLUCENCY_NORMAL : ASURA_TRANSLUCENCY_OFF;
			Asura_Render::Overlay.m_eTextureOp = (ASURA_TEXTUREOP_ID)m_iTextureOp;
		}

		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();

		//Reset the textureop since this isn't being setup in general elsewhere.RAR 02/02/2009
		Asura_Render::Overlay.m_eTextureOp = ASURA_TEXTUREOP_SINGLE_PASS;
		
	}

}




/// Asura_Tabs_Dialog /////////////////////////////////////////////////////////////////////////////

Asura_Tabs_Dialog::Asura_Tabs_Dialog(const Asura_Vector_2& xOffset, const char* const* pxTabNames, u_int uNumberOfTabs, int iMaxWidth /* = 0 */)
: Asura_Dialog("", xOffset)
, m_ppxTabGadgets(NULL)
, m_uNumberOfTabs(uNumberOfTabs)
, m_uNumberOfRows(1)
, m_uSelectedTab(0)
, m_puSelectedTab(0)
, m_iMaxWidth(iMaxWidth)
, m_pxSelectedTabName(0)
{
	m_ppxTabGadgets = m_uNumberOfTabs ? new Asura_Tab_Gadget*[uNumberOfTabs] : 0;
	m_xMaxCoord.x = 0.f;

	if(m_ppxTabGadgets)
	{
		//create all the gadgets
		for(u_int u(0); u < m_uNumberOfTabs; ++u)
		{
			m_ppxTabGadgets[u] = new Asura_Tab_Gadget(pxTabNames[u]);
			AddGadget(m_ppxTabGadgets[u]);
		}
		//and place them
		RepositionTabGadgets();
	}

	m_uSelectedTab = m_uNumberOfTabs - 1;
	SelectTab(0);
}


Asura_Tabs_Dialog::Asura_Tabs_Dialog( const Asura_Vector_2& xOffset,int iMaxWidth /* = 0 */)
: Asura_Dialog("", xOffset)
, m_ppxTabGadgets(NULL)
, m_uNumberOfTabs(0)
, m_uNumberOfRows(0)
, m_uSelectedTab(0)
, m_puSelectedTab(0)
, m_iMaxWidth(iMaxWidth)
, m_pxSelectedTabName(0)
{
	m_xMaxCoord.x = 0.f;
}


Asura_Tabs_Dialog::~Asura_Tabs_Dialog()
{
	if(m_ppxTabGadgets)
	{
		delete [] m_ppxTabGadgets;
	}
}

void Asura_Tabs_Dialog::SelectTab(int iTab)
{
	if(!m_uNumberOfTabs) return;

	if(iTab < 0)
	{
		// Select all. This is useful for calculating the limits of the overlay
		for(u_int u = 0; u < m_uNumberOfTabs; ++u)
		{
			m_ppxTabGadgets[u]->Select(true);
		}
	}
	else
	{
		Asura_Assert(iTab < static_cast<int>(m_uNumberOfTabs), "Tab out of range");
		if (iTab >= static_cast<int>(m_uNumberOfTabs))
		{
			iTab = 0;
		}
		m_uSelectedTab = iTab;
		if(m_puSelectedTab) *m_puSelectedTab = m_uSelectedTab;
		if(m_pxSelectedTabName)		*m_pxSelectedTabName = m_ppxTabGadgets[iTab]->GetIDName();
		u_int uRow = m_ppxTabGadgets[iTab]->GetRow();
		if(uRow != m_uNumberOfRows - 1)
		{
			for(u_int u = 0; u < m_uNumberOfTabs; ++u)
			{
				m_ppxTabGadgets[u]->SwapRow(uRow, m_uNumberOfRows - 1);
			}
		}
		for(u_int u = 0; u < m_uNumberOfTabs; ++u)
		{
			m_ppxTabGadgets[u]->Select(m_uSelectedTab == u);
		}
	}
}

bool Asura_Tabs_Dialog::SelectTabByName(const Asura_String& xName, bool bDefaultToFirstTab /*= true*/)
{
	int iTab = 0;

	while (iTab < static_cast<int>(m_uNumberOfTabs))
	{
		if (xName.Compare(m_ppxTabGadgets[iTab]->GetIDName()) == 0)
		{
			// There is a tab with this name
			SelectTab(iTab);
			return true;
		}
		iTab++;
	}
	if ( bDefaultToFirstTab )
	{
		// No tab with this name exists, so select the first one
		SelectTab(0);
	}
	return false;
}

void Asura_Tabs_Dialog::AddDialog(Asura_Dialog *pxDialog)
{
	if(m_ppxTabGadgets)
	{
		Asura_Assert(m_uSelectedTab < m_uNumberOfTabs, "Current tab out of range");
		m_ppxTabGadgets[m_uSelectedTab]->AddDialog(pxDialog);
	}
	m_pxOverlay->AddDialog(pxDialog, true);
}

const char* Asura_Tabs_Dialog::GetSelectedTabName() const
{
	return m_ppxTabGadgets ? m_ppxTabGadgets[ m_uSelectedTab ]->GetName() : "";
}

void Asura_Tabs_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	// Parent render
	Asura_Dialog::Render(xOrigin);

	// KOR - 17/05/2006 - only continue if not hidden
	if(IsHidden()) { return; }

	// Render left and right lines
	float fYPos = xOrigin.y + m_xOffset.y + (m_uNumberOfRows * ASURA_TAB_GADGET_HEIGHT);
	Asura_Render::Render2DLine( (xOrigin.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
		(fYPos) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
		(xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
		(fYPos) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
		Asura_OverlayBrightness|0x00fffffff);
	Asura_Render::Render2DLine( (xOrigin.x + m_pxOverlay->m_fWidth) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
		(fYPos) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
		(xOrigin.x + GetMaxXForRow(m_uNumberOfRows - 1)) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
		(fYPos) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
		Asura_OverlayBrightness|0x00fffffff);
	//render a line at the bottom of the tabbed controls
	Asura_Render::Render2DLine( (xOrigin.x ) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
		(xOrigin.y + m_xMaxCoordIncludingChildren.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
		(xOrigin.x + m_pxOverlay->m_fWidth) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
		(xOrigin.y + m_xMaxCoordIncludingChildren.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
		Asura_OverlayBrightness|0x00fffffff);

}

float Asura_Tabs_Dialog::GetMaxXForRow(u_int uRow) const
{
	float fMax = 0.f;
	for(u_int u = 0; u < m_uNumberOfTabs; ++u)
	{
		if(m_ppxTabGadgets[u]->GetRow() == uRow)
		{
			fMax = Asura_Maths::Max(fMax, m_ppxTabGadgets[u]->m_xOffset.x + m_ppxTabGadgets[u]->m_fWidth);
		}
	}
	return fMax;
}

void Asura_Tabs_Dialog::Update(const Asura_Vector_2& xOrigin)
{
	Asura_Dialog::Update(xOrigin);
	if(m_ppxTabGadgets)
	{
		bool bSelected = false;
		u_int uSelectedTab = 0;
		for(u_int u = 0; u < m_uNumberOfTabs; ++u)
		{
			if(m_ppxTabGadgets[u]->IsClickedThisFrame())// && m_uSelectedTab != u)
			{
				uSelectedTab = u;
				bSelected = true;
			}
		}
		if(bSelected)
		{
			SelectTab(uSelectedTab);
		}
	}
}

void Asura_Tabs_Dialog::OnHiddenStateChange()
{
	if(IsHidden())
	{
		for(u_int u = 0; u < m_uNumberOfTabs; ++u)
		{
			m_ppxTabGadgets[u]->Select(false);
		}
	}
	else
	{
		SelectTab(m_uSelectedTab);
	}
}

void Asura_Tabs_Dialog::PerFrameUpdate()
{
	if(m_puSelectedTab && *m_puSelectedTab != m_uSelectedTab)
	{
		SelectTab(*m_puSelectedTab);		
	}

	Asura_Dialog::PerFrameUpdate();
}

const Asura_Vector_2& Asura_Tabs_Dialog::GetMaxCoords() const
{
	//recalculate max coords from all the child dialogs
	m_xMaxCoordIncludingChildren = m_xMaxCoord;

	for(u_int u = 0; u < m_uNumberOfTabs; ++u)
	{
		Asura_Vector_2 xMax;
		m_ppxTabGadgets[u]->GetMaxCoords(xMax);

		if(m_xMaxCoordIncludingChildren.x < xMax.x) m_xMaxCoordIncludingChildren.x = xMax.x;
		if(m_xMaxCoordIncludingChildren.y < xMax.y) m_xMaxCoordIncludingChildren.y = xMax.y;
	}
	m_xMaxCoordIncludingChildren.y += 0.5f;//add a little beneath the bottom control

	return m_xMaxCoordIncludingChildren;
}

Asura_Vector_2 Asura_Tabs_Dialog::GetMaxCoordsForTab( const u_int uTab )
{
    Asura_Vector_2 xMaxCoords;

    for(u_int u = 0; u < m_uNumberOfTabs; ++u)
    {
        if ( u != uTab )
        {
            continue;
        }

        Asura_Vector_2 xMax;
        m_ppxTabGadgets[u]->GetMaxCoords(xMax);

        if (xMaxCoords.x < xMax.x) xMaxCoords.x = xMax.x;
        if (xMaxCoords.y < xMax.y) xMaxCoords.y = xMax.y;
    }
    xMaxCoords.y += 0.5f;//add a little beneath the bottom control

    return xMaxCoords;
}

void Asura_Tabs_Dialog::SetSelectedTabPointer(u_int* puSelectedTab)
{
	m_puSelectedTab = puSelectedTab;
	if(m_puSelectedTab && *m_puSelectedTab != m_uSelectedTab)
	{
		SelectTab(*m_puSelectedTab);		
	}
}

void Asura_Tabs_Dialog::SetSelectedTabPointerByName(Asura_String* pxSelectedTab)
{
	m_pxSelectedTabName = pxSelectedTab;
	if(m_pxSelectedTabName)
	{
		SelectTabByName(*pxSelectedTab);
	}
}

void Asura_Tabs_Dialog::AddTab(const char* szTabName)
{
	//create new array of tabs
	u_int uNewNumberOfTabs = m_uNumberOfTabs+1;

	Asura_Tab_Gadget** ppxNewTabGadgets = new Asura_Tab_Gadget*[uNewNumberOfTabs];

	//copy the old ones
	for(u_int u=0;u<m_uNumberOfTabs;u++)
	{
		ppxNewTabGadgets[u] = m_ppxTabGadgets[u];
	}
	//insert new tab
	Asura_Tab_Gadget* pxNewTab = new Asura_Tab_Gadget(szTabName);
	ppxNewTabGadgets[m_uNumberOfTabs] = pxNewTab;

	//replace old array
	delete [] m_ppxTabGadgets;
	m_ppxTabGadgets = ppxNewTabGadgets;
	m_uNumberOfTabs = uNewNumberOfTabs;

	AddGadget(pxNewTab);

	//sort out the positions for the tabs
	RepositionTabGadgets();

	//select the new tab
	SelectTab(m_uNumberOfTabs-1);

	if ( m_pxOverlay )
	{
		m_pxOverlay->ResizeOverlay();
	}
}

void Asura_Tabs_Dialog::DeleteTab(u_int uTab)
{
	if(uTab >= m_uNumberOfTabs)
	{
		Asura_Assert(0,"Invalid tab for deletion");
		return;
	}


	//get the tab being deleted
	Asura_Tab_Gadget* pxDeletedTab = m_ppxTabGadgets[uTab];

	//For safety , make sure no control has focus before deleting stuff
	//Could in principle check to see if its going to be one of the ones being deleted ,but
	//I don't think this will do any harm in practice. RAR 23/06/2006
	Asura_Overlay_System::ReleaseFocus();

	//delete the dialogs and gadgetsthat belonged to the tab
	pxDeletedTab->DeleteChildDialogs(m_pxOverlay);
	DeleteSingleGadget(pxDeletedTab);
	
	//Remove from the tabs array , and reposition the remaining tabs
	m_uNumberOfTabs--;
	for(u_int u=uTab;u<m_uNumberOfTabs;u++)
	{
		m_ppxTabGadgets[u] = m_ppxTabGadgets[u+1];
	}
	if(!m_uNumberOfTabs)
	{
		delete [] m_ppxTabGadgets;
		m_ppxTabGadgets = 0;

	}
	RepositionTabGadgets();

	//make sure the selected tab is valid
	u_int uSelectedTab = m_uSelectedTab;
	if(uSelectedTab >= m_uNumberOfTabs)
	{
		uSelectedTab = m_uNumberOfTabs-1;
	}
	SelectTab(uSelectedTab);

	m_pxOverlay->ResizeOverlay();
}

void Asura_Tabs_Dialog::DeleteTabByName( const Asura_String& xName )
{
	int iTab = 0;
	while (iTab < static_cast<int>(m_uNumberOfTabs))
	{
		if (xName.Compare(m_ppxTabGadgets[iTab]->GetIDName()) == 0)
		{
			// There is a tab with this name
			DeleteTab(iTab);
			return;
		}
		iTab++;
	}
}

void Asura_Tabs_Dialog::RenameTab(u_int uTab,const char* szTabName, bool bChangeIDName /*= true*/)
{
	Asura_Assert(uTab<m_uNumberOfTabs,"");

	if(!strcmp(szTabName, m_ppxTabGadgets[uTab]->GetName()))
	{
		return;
	}

	//update the tab gadget	
	m_ppxTabGadgets[uTab]->SetName(szTabName, bChangeIDName);

	//sort out the positions for the tabs
	RepositionTabGadgets();

	m_pxOverlay->ResizeOverlay();

	SelectTab(m_uSelectedTab);
}

void Asura_Tabs_Dialog::ClearTab(u_int uTab)
{
	Asura_Assert(uTab<m_uNumberOfTabs,"");

	Asura_Tab_Gadget*	pxTab	= m_ppxTabGadgets[uTab];

	// Remove all child controllers from this tab...
	pxTab->DeleteChildDialogs(m_pxOverlay);
	m_pxOverlay->ResizeOverlay();
}

void Asura_Tabs_Dialog::RepositionTabGadgets()
{
	Asura_Vector_2 xGadgetOffset = m_xOffset;
	m_xMaxCoord.x = 0.f;

	float fWidth = 0;
	m_uNumberOfRows = 1;
	if(m_ppxTabGadgets)
	{
		for(u_int u = 0; u < m_uNumberOfTabs; ++u)
		{
			const float fLength = m_ppxTabGadgets[u]->m_fWidth;
			fWidth += fLength;
			if(m_iMaxWidth && fWidth > m_iMaxWidth)
			{
				m_uNumberOfRows++;
				m_xMaxCoord.x = Asura_Maths::Max(m_xMaxCoord.x, xGadgetOffset.x);
				xGadgetOffset.x = m_xOffset.x;
				xGadgetOffset.y += ASURA_TAB_GADGET_HEIGHT;
				fWidth = fLength;
			}
			m_ppxTabGadgets[u]->SetPosition(xGadgetOffset,m_uNumberOfRows-1);
			xGadgetOffset.x += fLength;
		}
		for(u_int u = 0; u < m_uNumberOfTabs; ++u)
		{
			m_ppxTabGadgets[u]->SetBottomRow(m_ppxTabGadgets[u]->GetRow() == m_uNumberOfRows - 1);
		}
	}
	m_xMaxCoord.x = Asura_Maths::Max(m_xMaxCoord.x, xGadgetOffset.x);
	m_xMaxCoord.y = m_xOffset.y + m_uNumberOfRows * ASURA_TAB_GADGET_HEIGHT;
	m_xMaxCoordIncludingChildren = m_xMaxCoord;
}

Asura_Vector_2 Asura_Tabs_Dialog::GetTopDialogPosition() const
{
	Asura_Vector_2 xPos = m_xOffset;
	xPos.y = m_xMaxCoord.y + 1.0f;
	return xPos;
}

float Asura_Tabs_Dialog::GetTabWidth()
{
	float fWidth = 0;
	for(u_int i=0;i<m_uNumberOfRows;i++)
	{
		fWidth = Asura_Maths::Max(fWidth,GetMaxXForRow(i));
	}
	return fWidth;
}

Asura_Vector_2 Asura_Tabs_Dialog::GetPositionFromID(Asura_Hash_ID uPositionID) const
{
	if(uPositionID == s_uTopLeftPositionID) return GetTopDialogPosition();
	return Asura_Dialog::GetPositionFromID(uPositionID);
}

void Asura_Tabs_Dialog::DeleteChildDialogs (void)
{
	while(m_uNumberOfTabs > 0)
	{
		DeleteTab(0);
	}
}

void Asura_Tabs_Dialog::RemoveReferencesToDialog(const Asura_Dialog* pxDialog)
{
	Asura_Dialog::RemoveReferencesToDialog(pxDialog);

	//make sure the dialog isn't a child of any of the tabs
	for(u_int i=0;i<m_uNumberOfTabs;i++)
	{
		m_ppxTabGadgets[i]->RemoveReferencesToDialog(pxDialog);
	}
}
void Asura_Tabs_Dialog::StartResize(u_int uResizeFlags)
{
	m_fPreResizeWidth = m_xMaxCoord.x - m_xOffset.x;
	m_fPreResizeHeight = m_xMaxCoord.y - m_xOffset.y;
}

void Asura_Tabs_Dialog::UpdateResize(u_int uResizeFlags, const Asura_Vector_2& xDelta)
{
	if(!m_bAllowResize) return;

	float fWidth,fHeight;

	if(uResizeFlags & ASURA_OVERLAY_RESIZE_INITIAL)
	{
		//initialise the size based on the last size chosen by the user
		if(!m_pxUserSetSize) return;
		if(*m_pxUserSetSize == Asura_Maths::ZeroVector2) return;

		fWidth = Asura_Maths::Max(4.0f, m_pxUserSetSize->x);
		fHeight = Asura_Maths::Max(4.0f, m_pxUserSetSize->y);
	}
	else
	{
		fWidth = Asura_Maths::Max(4.0f, m_fPreResizeWidth + xDelta.x);
		fHeight = Asura_Maths::Max(4.0f, m_fPreResizeHeight + xDelta.y);

		if(m_pxUserSetSize)
		{
			m_pxUserSetSize->x = fWidth;
			m_pxUserSetSize->y = fHeight;
		}
	}

	if(fWidth != m_xMaxCoord.x - m_xOffset.x || fHeight != m_xMaxCoord.y - m_xOffset.y)
	{
		m_xMaxCoord.x = m_xOffset.x + fWidth;
		m_xMaxCoord.y = m_xOffset.y + fHeight;
	}
}

u_int Asura_Tabs_Dialog::GetAllowedResizeFlags() const
{
	return m_bAllowResize ? ASURA_OVERLAY_RESIZE_HORIZONTAL | ASURA_OVERLAY_RESIZE_VERTICAL : 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_Hash_DropDownList_Dialog **
**--------------------------------*/


Asura_Hash_DropDownList_Dialog::Asura_Hash_DropDownList_Dialog(const Asura_Vector_2& xOffset)
:Asura_DropDown_List_Dialog(xOffset)
{
	m_bChanged = false;
	m_iSelectedIndex = 0;

	m_pxStringTree = 0;
	m_ppszStrings = 0;

	m_iListValidID = 0;

	m_puSelected = 0;
	m_pxSharedSelected = 0;

	m_uIterationPos = 0;
	m_uIterationLength = 0;
}

void Asura_Hash_DropDownList_Dialog::Initialise(const Asura_Vector_2& xOffset,int iNumberOfLines,Asura_Hash_ID* puSelected,bool* pbChangedFlag,const char* szName)
{
	m_puSelected = puSelected;
	m_uSelectedCopy = GetExternalHash();
	m_pbChangedFlag = pbChangedFlag;

	UpdateStringList(false);

	Asura_DropDown_List_Dialog::Initialise(xOffset,m_ppszStrings,&m_iSelectedIndex,iNumberOfLines,m_pxStringTree->GetNumberOfNodes(),NULL,&m_bChanged,szName);

	UpdateSelectedIndex();
}

void Asura_Hash_DropDownList_Dialog::Initialise(const Asura_Vector_2& xOffset,int iNumberOfLines,Asura_Dialog_SharedValue<Asura_Hash_ID>* pxSharedSelected,bool* pbChangedFlag,const char* szName)
{
	m_pxSharedSelected = pxSharedSelected;
	m_uSelectedCopy = GetExternalHash();
	m_pbChangedFlag = pbChangedFlag;

	UpdateStringList(false);

	Asura_DropDown_List_Dialog::Initialise(xOffset,m_ppszStrings,&m_iSelectedIndex,iNumberOfLines,m_pxStringTree->GetNumberOfNodes(),NULL,&m_bChanged,szName);

	UpdateSelectedIndex();
}


Asura_Hash_DropDownList_Dialog::~Asura_Hash_DropDownList_Dialog()
{
	DeleteStringList();
}


void Asura_Hash_DropDownList_Dialog::UpdateStringList(bool bUpdateDialog /*= true*/)
{
	if(m_ppszStrings) DeleteStringList();
	m_pxStringTree = new Asura_StringTree;


	int iJunkIndex;

	bool bFoundCurrentHash = false;//for determining if  unknown entry is needed

	Asura_Hash_ID uCurrentHash = GetExternalHash();

	//get all the  names
	for(Iteration_Start(); !Iteration_Done(); Iteration_Next())
	{
		const char* pszName = Iteration_GetCurrent();

		if(pszName)
		{
			m_pxStringTree->FindNodeIndexByName(pszName,true,iJunkIndex);

			if(uCurrentHash == Asura_GetHashID(pszName))
			{
				bFoundCurrentHash = true;
			}
		}
	}

	if ( GetAllowNoneEntry() )
	{
		//add entry for none
		m_pxStringTree->FindNodeIndexByName(GetZeroHashText(),true,iJunkIndex);
	}

	//add entry for unknown if needed
	if(uCurrentHash && !bFoundCurrentHash)
	{
		m_pxStringTree->FindNodeIndexByName(GetUnknownHashText(),true,iJunkIndex);
	}

	//get list of names from name tree
	m_ppszStrings = new const char*[m_pxStringTree->GetNumberOfNodes()];
	m_pxStringTree->MakeSortedNameList(m_ppszStrings);

	//update dialog
	if(bUpdateDialog)
	{
		ReplaceStrings(m_pxStringTree->GetNumberOfNodes(),m_ppszStrings);
	}

	m_iListValidID = GetListValidID();
}

void Asura_Hash_DropDownList_Dialog::DeleteStringList()
{
	delete m_pxStringTree;
	m_pxStringTree = 0;

	delete [] m_ppszStrings;
	m_ppszStrings = 0;
} 

bool Asura_Hash_DropDownList_Dialog::UpdateSelectedIndex()
{
	//search in the list of strings
	int iUnknownIndex=-1;
	const char* szUnknown = GetUnknownHashText();
	Asura_Hash_ID uCurrentHash = GetExternalHash();
	for(int i=0;i<m_pxStringTree->GetNumberOfNodes();i++)
	{
		if(uCurrentHash)
		{
			if(!strcmp(m_ppszStrings[i],szUnknown))
			{
				iUnknownIndex = i;
				continue;
			}
			//compare by hash value
			if(Asura_GetHashID(m_ppszStrings[i]) == uCurrentHash)
			{
				SelectEntry(i);
				return true;
			}

		}
		else
		{
			//compare to none string
			if(!strcmp(m_ppszStrings[i],GetZeroHashText()))
			{
				SelectEntry(i);
				return true;
			}
		}
	}
	if(iUnknownIndex>=0)
	{
		//Couldn't find string with correct hash , but did find the unknown string
		SelectEntry(iUnknownIndex);
		return true;
	}

	SelectEntry(0);
	return false;
}

void Asura_Hash_DropDownList_Dialog::PerFrameUpdate()
{
	//Check for external changes to selection
	Asura_Hash_ID uCurrentHash = GetExternalHash();
	if(m_uSelectedCopy != uCurrentHash)
	{
		m_uSelectedCopy = uCurrentHash;

		if(!UpdateSelectedIndex())
		{
			//failed to update index , try rebuilding the string list
			//(this should add unknown entry if needed)
			UpdateStringList();
			UpdateSelectedIndex();
		}
	}

	if(m_iListValidID != GetListValidID())
	{
		//List has become invalid ,so update it.
		UpdateStringList();
		UpdateSelectedIndex();
	}

	Asura_DropDown_List_Dialog::PerFrameUpdate();
}

void Asura_Hash_DropDownList_Dialog::Update(const Asura_Vector_2& xOrigin)
{
	Asura_DropDown_List_Dialog::Update(xOrigin);

	//check for changes to the selection
	if(m_bChanged)
	{
		if(m_pbChangedFlag) *m_pbChangedFlag = true;
		m_bChanged = false;

		//get hash id from index
		const char* pszText = m_ppszStrings[m_iSelectedIndex];
		if(!strcmp(pszText,GetZeroHashText()))
		{
			SetExternalHash(ASURA_HASH_ID_UNSET);
		}
		else if(!strcmp(pszText,GetUnknownHashText()))
		{
			//ignore changes to unknown
		}
		else
		{
			SetExternalHash(Asura_GetHashID(pszText));
		}
		//update copy of id for detecting external changes
		m_uSelectedCopy = GetExternalHash();
	}
}

Asura_Hash_ID Asura_Hash_DropDownList_Dialog::GetExternalHash() const
{
	if(m_puSelected)
	{
		return *m_puSelected;
	}
	if(m_pxSharedSelected)
	{
		u_int uValue;
		m_pxSharedSelected->GetValue(&uValue,0);
		return uValue;
	}
	return 0;
}

void Asura_Hash_DropDownList_Dialog::SetExternalHash(Asura_Hash_ID uHash)
{
	if(m_puSelected)
	{
		*m_puSelected = uHash;
	}
	if(m_pxSharedSelected)
	{
		m_pxSharedSelected->SetValue(uHash);
	}
}

bool Asura_Hash_DropDownList_Dialog::IsMixedValue() const
{
	if(!m_pxSharedSelected) return false;

	bool bMixed;
	m_pxSharedSelected->GetValue(0,&bMixed);
	return bMixed;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
//-------------------------------------------**
// Asura_StringValuePair_DropDownList_Dialog **
//-------------------------------------------**

float Asura_StringValuePair_DropDownList_Dialog::s_fDropListSpacing				= 0.2f;
float Asura_StringValuePair_DropDownList_Dialog::s_fDropListLineOffsetY			= 0.1f;
float Asura_StringValuePair_DropDownList_Dialog::s_fDropListLineAdjustX1		= -2.0f;
float Asura_StringValuePair_DropDownList_Dialog::s_fDropListLineAdjustX2		= -1.0f;
Asura_Vector_2 Asura_StringValuePair_DropDownList_Dialog::s_xDropListScrollBarOffset( 0.0f, 0.2f );

Asura_StringValuePair_DropDownList_Dialog::Asura_StringValuePair_DropDownList_Dialog(Asura_Vector_2& xOffset, const char* szName, int* iSelectionValue, bool* bChanged )
: PARENT( 0, xOffset, 1 )
, m_szName( szName )
, m_piIndexConv( NULL )
, m_ppcStrings( NULL )
, m_ppcHelpStrings( NULL )
, m_iSelectedValue( iSelectionValue )
, m_iNumItems( 0 )
, m_bChanged( bChanged )
{

}

Asura_StringValuePair_DropDownList_Dialog::~Asura_StringValuePair_DropDownList_Dialog()
{
	delete m_piListOffsetScroll;

	delete[] m_ppcHelpStrings;
	delete[] m_ppcStrings;
	delete[] m_piIndexConv;

	// Must delete the list entries manually, as List only stores pointers
	for( LIF< SelectionItem* > xLIF( &m_xItems ); !xLIF.done(); xLIF.next() )
	{
		SelectionItem* const pxCurrItem = xLIF();
		delete pxCurrItem;
	}
}

void Asura_StringValuePair_DropDownList_Dialog::AddItem( const char* szString, int iData /*= 0*/, bool bIsSelectedInitially /*= false*/, const char* szHelpString /*= NULL*/, bool bDontAddIfDuplicateString /*= false*/ )
{
	if (bDontAddIfDuplicateString)
	{
		LIF<SelectionItem*> ilif( &m_xItems );
		for(;!ilif.done();ilif.next())
		{
			Asura_String xTemp(szString);
			if( xTemp.Compare(ilif()->xString, true) == 0 )
			{
				return;
			}
		}		
	}

	SelectionItem* pxNewItem = new SelectionItem();

	pxNewItem->xString				= szString;
	pxNewItem->xHelpString			= szHelpString ? szHelpString : "";
	pxNewItem->iData				= iData;
	pxNewItem->bSelectedInitially	= bIsSelectedInitially;

	m_xItems.add_entry( pxNewItem );
	++m_iNumItems;
}

void Asura_StringValuePair_DropDownList_Dialog::InitialiseData( void )
{
	delete[] m_piIndexConv;
	delete[] m_ppcStrings;
	delete[] m_ppcHelpStrings;

	int iNumItems = m_xItems.size();

	if( iNumItems > 0 )
	{
		m_ppcStrings		= new const char*[iNumItems];
		m_ppcHelpStrings	= new const char*[iNumItems];
		m_piIndexConv		= new int[iNumItems];

		m_iSelection = 0;

		for( int i = 0; i < iNumItems; ++i )
		{
			m_ppcStrings[i]		= m_xItems[i]->xString.GetString();
			m_ppcHelpStrings[i]	= m_xItems[i]->xHelpString.GetString();
			m_piIndexConv[i]	= m_xItems[i]->iData;
		}

		m_iNumInList = iNumItems;
		m_iSelection = 0;

		for (int i = 0; i < m_iNumInList; ++i)
		{
			if (m_piIndexConv[i] == *m_iSelectedValue)
			{
				m_iSelection = i;
			}
		}
	}
	else
	{
		m_ppcStrings		= new const char*[1];
		m_ppcHelpStrings	= new const char*[1];
		m_piIndexConv		= new int[1];

		m_ppcStrings[0]		= GetEmptyListString();
		m_ppcHelpStrings[0]	= GetEmptyListString();
		m_piIndexConv[0]	= -1;

		m_iSelection		= -1;
		iNumItems			= 1;

		m_iNumInList		= 1;
	}
}

void Asura_StringValuePair_DropDownList_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	if(!IsHidden())
	{
		u_int uColour = GetColour();
		
		const float fOffsetX = 2.f;
		float fHeight = 1.f;
		if ( m_bIsDroppedDown )
		{
			fHeight += ( m_iNumLines * 1.2f ) + s_fDropListSpacing;
		}

		const float fNameWidth = Asura_Text_Output::WidthOfString( m_szName, ASURA_TEXT_FORMAT_SMALLFONT );

		Asura_Render::Overlay.Colour = uColour;

		Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;

		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

		Asura_Render::Overlay.SetRectXY
		(
			( xOrigin.x + m_xOffset.x ) * ASURA_TEXT_OUTPUT_FONT_WIDTH - ( 1.f + fOffsetX ),
			( xOrigin.y + m_xOffset.y ) * ASURA_TEXT_OUTPUT_FONT_HEIGHT - 1.f,
			( m_xMaxCoord.x - m_xOffset.x - fNameWidth ) * ASURA_TEXT_OUTPUT_FONT_WIDTH + 2.f,
			fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT + 2.f
		);

		Asura_Render::Overlay.Render();

		Asura_Render::FlushDisplay();

		Asura_Render::Overlay.Colour = Asura_OverlayBrightness;

		Asura_Render::Overlay.SetRectXY
		(
			( xOrigin.x + m_xOffset.x ) * ASURA_TEXT_OUTPUT_FONT_WIDTH - fOffsetX,
			( xOrigin.y + m_xOffset.y ) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
			( m_xMaxCoord.x - m_xOffset.x - fNameWidth ) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
			fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT
		);

		Asura_Render::Overlay.Render();

		Asura_Render::FlushDisplay();

		if ( m_bIsDroppedDown )
		{
			const float fX1 = ( ( xOrigin.x + m_xOffset.x ) * ASURA_TEXT_OUTPUT_FONT_WIDTH ) + s_fDropListLineAdjustX1;
			const float fY1 = ( ( xOrigin.y + m_xOffset.y + 1.0f + s_fDropListLineOffsetY ) * ASURA_TEXT_OUTPUT_FONT_HEIGHT );
			const float fX2 = ( ( ( xOrigin.x + m_xMaxCoord.x ) - fNameWidth ) * ASURA_TEXT_OUTPUT_FONT_WIDTH ) + s_fDropListLineAdjustX2;
			const float fY2 = fY1;
			const unsigned int uColour = Asura_OverlayBrightness | 0x00fffffff;
			Asura_Render::Render2DLine( fX1, fY1, fX2, fY2, uColour );
		}

		if ( m_szName )
		{
			Asura_Text_Output::Print(xOrigin.x + m_xMaxCoord.x + 1.f, xOrigin.y + m_xOffset.y, GetColour(), ASURA_TEXT_FORMAT_SMALLFONT | ASURA_TEXT_FORMAT_RIGHTJUSTIFIED, "%s", m_szName);
		}


		//update colour for drop down lists using Asura_Dialog_SharedValue
		m_pxSelectedGadget->SetUseMixedGadgetColour( false );
		m_pxDropDownGadget->SetUseSharedGadgetColour( false );

		PARENT::Render(xOrigin);
	}
}

void Asura_StringValuePair_DropDownList_Dialog::Update(const Asura_Vector_2& xOrigin)
{
	SelectEntry( m_iSelection );

	// Parent update
	Asura_Dialog::Update(xOrigin);

	if ( m_bIsDroppedDown )
	{
		if (!Asura_Overlay_System::GetFocusGadget() )
		{
			Asura_Overlay_System::TakeFocus( m_pxFirstTextGadget, OverlayFocusType_Keyboard );
		}

		//HandleKeyboard();
	}

	if (m_bIsDroppedDown)
	{
		m_pxSelectedGadget->SetLabel( m_ppcStrings[m_iSelection] );
	}

	bool bToggleDropDown = false;
	
	if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) || 
		Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
	{

		Asura_Gadget* pxGadget = m_pxFirstTextGadget;

		for(; pxGadget; pxGadget = pxGadget->m_pxNextGadget)
		{
			if ( pxGadget == m_pxUpGadget || pxGadget == m_pxDownGadget || pxGadget == m_pxScrollGadget )
			{
				continue;
			}

			if( pxGadget->IsMouseOver() )
			{
				m_pxOverlay->m_bDialogTookFocus = true;
				ToggleDropDown();
				return;
			}
		}
	}

	if ( m_bIsDroppedDown )
	{
		Asura_Vector_2 xPos(xOrigin + m_xOffset);

		// If they've clicked outside the dialog, cancel the drop down
		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
		{
			if ( !Asura_Cursor::Inside ( xPos, m_xMaxCoord.x - m_xOffset.x, 1.2f * m_iNumLines + 1.f ) )
			{
				bToggleDropDown = true;

			}
			else
			{
				m_pxOverlay->m_bDialogTookFocus = true;
			}
		}

		//close drop down if enter is pressed while open
		if (Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_ENTER) )
		{
			bToggleDropDown = true;
		}
		
		//close drop down if it is disabled/hidden while dropped
		if ( !m_bIsEnabled || IsHidden() )
		{
			bToggleDropDown = true;
		}
		
	}

	if ( bToggleDropDown )
	{
		m_pxOverlay->m_bDialogTookFocus = true;
		ToggleDropDown();
	}

	if ( m_bIsEnabled )
	{
		SetGadgetsHidden();
	}
}

void Asura_StringValuePair_DropDownList_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	//SelectEntry( m_iSelection );
}

void Asura_StringValuePair_DropDownList_Dialog::SelectEntry( int iSelection )
{
	if ( m_iLastSelection != iSelection )
	{
		*m_bChanged = true;
	}

	if (iSelection < m_iNumLines)
	{
		m_iSelection = iSelection;
	}
	else
	{
		m_iSelection = m_iNumLines - 1;
	}

	*m_iSelectedValue = m_piIndexConv[ m_iSelection ];
	
	m_pxSelectedGadget->SetLabel( m_ppcStrings[ m_iSelection ] );

	m_iLastSelection = m_iSelection;
}

const char* Asura_StringValuePair_DropDownList_Dialog::GetEmptyListString( void ) const
{
	return "<no items in list>";
}


void Asura_StringValuePair_DropDownList_Dialog::CreateGadgets( Asura_Vector_2& xOffset, const int iNumLines /*= 0*/ )
{
	m_pxUpGadget = 0;
	m_pxDownGadget = 0;
	m_pxScrollGadget = 0;

	m_iLastSelection = m_iSelection;
	m_iNumInList = m_iNumItems;

	Asura_Vector_2 xTempOffset = xOffset;

	if ( iNumLines == 0 )
	{
		m_iNumLines = m_iNumItems;
	}
	else
	{
		m_iNumLines = iNumLines;
	}

	Asura_Assert( ( ( m_iSelection >= 0 ) && ( m_iSelection < m_iNumInList ) ), "Selection Value is out of range!" );

	m_bNeedsScrollBar = false;
	
	if (m_iNumInList > m_iNumLines) 
	{
		m_bNeedsScrollBar = true;
	}

	// find the longest string
	int iMaxLength = 1;

	float fMaxWidth = m_fMinWidth;

	for ( int i = 0; i < m_iNumInList; i++ )
	{
		if ( m_ppcStrings[ i ] )
		{
			iMaxLength = Asura_Maths::Max( static_cast<int>( strlen( m_ppcStrings[ i ] ) ), iMaxLength );
			fMaxWidth = Asura_Maths::Max( Asura_Text_Output::WidthOfString( m_ppcStrings[ i ], ASURA_TEXT_FORMAT_SMALLFONT ), fMaxWidth );
		}		
	}

	float fMaxAllowed = ( float ) ( ( Asura_Viewport::GetWidth() / 10 ) - 10 );

	if ( fMaxWidth > fMaxAllowed )
	{
		fMaxWidth = fMaxAllowed;
	}

	Asura_Gadget* pxGadget;

	pxGadget = m_pxSelectedGadget = new Asura_StringDisplay_Gadget( xOffset, iMaxLength, NULL, NULL, 0, ASURA_TEXT_FORMAT_SMALLFONT );
	m_pxSelectedGadget->SetLabel( m_ppcStrings[ m_iSelection ] );
	m_pxSelectedGadget->m_fWidth = fMaxWidth;
	AddGadget( pxGadget );


	Asura_Vector_2 xRightOffset( xOffset );
	xRightOffset.x += fMaxWidth + 0.5f;
	pxGadget = m_pxDropDownGadget = new Asura_UpDown_Gadget( xRightOffset, &m_iDropDown, 0, 1, 1 );
	AddGadget( pxGadget );

	xOffset.y += 1.f;
	xRightOffset.y += 1.f;

	Asura_Vector_2 xDropDownOffset(0.0f, 0.2f);

	m_piListOffsetScroll = new int;

	*m_piListOffsetScroll = 0;

	pxGadget = m_pxUpGadget = new Asura_UpDown_Gadget( xRightOffset + xDropDownOffset, m_piListOffsetScroll, 0, m_iNumInList - m_iNumLines, -1);
	AddGadget(pxGadget);

	if ( m_bNeedsScrollBar )
	{
		//add scroll bar
		xRightOffset.y += 1.2f;
		float fHeight = 1.2f * ( m_iNumLines - 2 );
		pxGadget = m_pxScrollGadget = new Asura_VerticalScroll_Gadget( xRightOffset + xDropDownOffset, fHeight, m_iNumLines, m_piListOffsetScroll, 0, Asura_Maths::Max( m_iNumInList - m_iNumLines, 0 ) );
		AddGadget( pxGadget );
	}	

	float fDropListSpacing = 0.2f;

	for (int i=0; i < m_iNumLines; i++)
	{
		if( i == 0 )
		{
			xOffset.y += fDropListSpacing;
		}

		pxGadget  = new Asura_TextSelect_Gadget( xOffset, &m_iSelection, i, m_piListOffsetScroll, m_ppcStrings, m_iNumLines, m_iNumInList, NULL );
		//increase the height so it isn't possible for the cursor to fall in a gap between lines
		pxGadget->m_fHeight = 1.2f;
		//set the width of all the gadgets according to the maximum string length
		pxGadget->m_fWidth = fMaxWidth;
		
		if (i >= m_iNumInList)
		{
			pxGadget->SetHidden(true);
		}

		AddGadget( pxGadget );
		xOffset.y += 1.2f;
	}

	// store the pointer of the first text gadget in the linked pszList
	m_pxFirstTextGadget = ( Asura_TextSelect_Gadget* )m_pxGadgetList;
	m_iNumLines = Asura_Maths::Min( m_iNumLines, m_iNumInList );

	if (m_bNeedsScrollBar)
	{
		xOffset.y -= 1.2f;	
		xOffset.x += fMaxWidth + 0.5f;

		pxGadget  = m_pxDownGadget = new Asura_UpDown_Gadget( xOffset + xDropDownOffset, m_piListOffsetScroll, 0, m_iNumInList - m_iNumLines, 1 );
		AddGadget( pxGadget );
	}
	
	m_xMaxCoord.x = m_xOffset.x + fMaxWidth + 2.f;
	
	if ( m_szName )
	{
		m_xMaxCoord.x += Asura_Text_Output::WidthOfString( m_szName, ASURA_TEXT_FORMAT_SMALLFONT );
	}

	m_xMaxCoord.y = m_xOffset.y + 1.2f; //Only set MaxCoord to include dialog in non-dropped down state. RAR 02/06/2005

	// Ensure drop down is hidden by setting on then toggling off.
	m_bIsDroppedDown = true;

	ToggleDropDown();

	xTempOffset.y += Asura_Dialog::DefaultSpacing();

	xOffset = xTempOffset;
}

void Asura_StringValuePair_DropDownList_Dialog::ToggleDropDown( void )
{
	m_bIsDroppedDown = !m_bIsDroppedDown;

	if ( m_bIsDroppedDown )
	{
		Asura_Overlay_System::TakeFocus( m_pxFirstTextGadget, OverlayFocusType_Keyboard );
		//m_szInput[0] = 0;
	}
	else
	{
		if(Asura_Overlay_System::GetFocusGadget() == m_pxFirstTextGadget)
		{
			Asura_Overlay_System::ReleaseFocus();
		}
		
		// update the label with the current selection
		int iIndex = m_iSelection;

		Asura_Assert( iIndex >= 0 && iIndex < m_iNumInList, "Asura_StringValuePair_DropDownList_Dialog :: ToggleDropDown - Index Out Of range" );

		if( iIndex >= 0 && iIndex < m_iNumInList)
		{
			m_pxSelectedGadget->SetLabel( m_ppcStrings[ iIndex ] );
		}
	}
	
	// Increase priority if dropped down
	if ( m_bIsDroppedDown )
	{
		m_iFocusPriority++;
	}
	else
	{
		m_iFocusPriority--;
	}

	SetGadgetsHidden();
}

void Asura_StringValuePair_DropDownList_Dialog::SetGadgetsHidden( void )
{
	Asura_Gadget* pxTextGadget = m_pxFirstTextGadget;

	int iLine = m_iNumLines;

	while ( pxTextGadget )
	{
		pxTextGadget->SetHidden( ( !m_bIsDroppedDown ) || iLine > m_iNumLines );
		pxTextGadget = pxTextGadget->m_pxNextGadget;
		iLine--;
	}
	
	if( m_pxUpGadget )
	{
		m_pxUpGadget->SetHidden( !( m_bIsDroppedDown && m_bNeedsScrollBar ) );
	}

	if ( m_pxDownGadget )
	{
		m_pxDownGadget->SetHidden( !( m_bIsDroppedDown && m_bNeedsScrollBar ) );
	}

	if( m_pxScrollGadget )
	{
		m_pxScrollGadget->SetHidden( !( m_bIsDroppedDown && m_bNeedsScrollBar ) );
	}

	m_pxDropDownGadget->SetHidden( false );
	m_pxSelectedGadget->SetHidden( false );
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Asura_Scrollbar_Dialog **
**-----------------------*/

Asura_Scrollbar_Dialog::Asura_Scrollbar_Dialog(const Asura_Vector_2& xOffset,float* pfValue,float fMin,float fMax,bool bHorizontal,float fLength,const char* szPrintFormat,bool* pbChanged, void(*pfnChanged)(void))
:Asura_Dialog("",xOffset)
,m_pfValue(pfValue)
,m_szPrintFormat(szPrintFormat)
{
	Asura_Assert(pfValue,"");

	m_pxScrollbar = new Asura_Scrollbar_Gadget(xOffset,pfValue,fMin,fMax,bHorizontal,fLength,szPrintFormat,pbChanged, pfnChanged);
	AddGadget(m_pxScrollbar);

	m_xMaxCoord.x = xOffset.x + m_pxScrollbar->m_fWidth;
	m_xMaxCoord.y = xOffset.y + m_pxScrollbar->m_fHeight;
}

void Asura_Scrollbar_Dialog::ShowHelpText()
{
	if(!m_szPrintFormat)
	{
		Asura_Dialog::ShowHelpText();
		return;
	}

	//Add current value to the help text
	char szValue[50];
	Asura_PrintfToBfr(szValue,m_szPrintFormat,sizeof(szValue),true,*m_pfValue);


	char szText[500];
	const char* szHelpText = GetHelpText();
	if(szHelpText)
	{
		Asura_PrintfToBfr(szText,"%s\n\nValue : %s",sizeof(szText),true,szHelpText,szValue);
	}
	else
	{
		Asura_PrintfToBfr(szText,"Value : %s",sizeof(szText),true,szValue);
	}

	Asura_Overlay_System::SetHelpText(szText);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Asura_Subwindow_Dialog **
**------------------------*/

Asura_Subwindow_Dialog* Asura_Subwindow_Dialog::s_pxClippingWindow = 0;

Asura_Subwindow_Dialog::Asura_Subwindow_Dialog(const Asura_Vector_2& xOffset,float fMaxWidth,float fMaxHeight,bool bShrinkToFit)
:Asura_Dialog("",xOffset)
,m_bShrinkToFit(bShrinkToFit)
,m_bOutlineHidden(false)
,m_bSizeValid(false)
,m_fMaxWidth(fMaxWidth)
,m_fMaxHeight(fMaxHeight)
,m_fChildWidth(0.0f)
,m_fChildHeight(0.0f)
,m_fUsableWidth(fMaxWidth)
,m_fUsableHeight(fMaxHeight)
,m_xChildRectMin(0.0f,0.0f)
,m_xChildRectMax(0.0f,0.0f)
,m_xChildOffset(0.0f,0.0f)
,m_fMouseWheelScrollRate(3.0f)
,m_fPreResizeWidth(0.0f)
,m_fPreResizeHeight(0.0f)
{
	m_pxHorizontalBar = new Asura_Scrollbar_Gadget(xOffset,&m_xChildOffset.x,0.0f,0.0f,true,0.0f,0,0);
	m_pxHorizontalBar->SetHidden(true);
	AddGadget(m_pxHorizontalBar);

	m_pxVerticalBar = new Asura_Scrollbar_Gadget(xOffset,&m_xChildOffset.y,0.0f,0.0f,false,0.0f,0,0);
	m_pxVerticalBar->SetHidden(true);
	AddGadget(m_pxVerticalBar);

	m_fWidth = bShrinkToFit ? 0.0f : fMaxWidth;
	m_fHeight = bShrinkToFit ? 0.0f : fMaxHeight;

	m_xMaxCoord.x = m_xOffset.x + m_fWidth;
	m_xMaxCoord.y = m_xOffset.y + m_fHeight;
}

void Asura_Subwindow_Dialog::AddDialog(Asura_Dialog* pxDialog)
{
	pxDialog->SetParentSubwindow(this);
	m_xDialogList.Push_Back(pxDialog);
	m_pxOverlay->AddDialog(pxDialog,true);
	m_bSizeValid = false;
}

void Asura_Subwindow_Dialog::UpdateSizeInit()
{
	m_fChildWidth = 0.0f;
	m_fChildHeight = 0.0f;
}

void Asura_Subwindow_Dialog::UpdateSize()
{
	//assume the full width/height will be used to start with
	m_fWidth = m_fMaxWidth;
	m_fHeight = m_fMaxHeight;

	//Add a bit to the space used by the child controls , so that there is some blacnk space at the sides when fully scrolled
	float fPaddedChildWidth = m_fChildWidth ? m_fChildWidth+1.0f : 0.0f;
	float fPaddedChildHeight = m_fChildHeight ? m_fChildHeight+1.0f : 0.0f;


	//determine which bars will be required
	bool bNeedHBar = false;
	bool bNeedVBar = false;


	if(fPaddedChildWidth > m_fWidth )
	{
		bNeedHBar = true;

		//if horizontal bar is being used , then there is less space vertically
		if(fPaddedChildHeight > m_fHeight-m_pxHorizontalBar->GetBarThickness())
		{
			bNeedVBar = true;
		}
	}
	if(fPaddedChildHeight > m_fHeight )
	{
		bNeedVBar = true;

		//if vertical bar is being used , then there is less space horizontally
		if(fPaddedChildWidth > m_fWidth-m_pxVerticalBar->GetBarThickness())
		{
			bNeedHBar = true;
		}
	}

	//shrink dialog to fit the desired controls
	if(m_bShrinkToFit)
	{
		float fUsedWidth = fPaddedChildWidth;
		if(bNeedVBar) fUsedWidth += m_pxVerticalBar->GetBarThickness();
		m_fWidth = Asura_Maths::Min(m_fMaxWidth,fUsedWidth);

		float fUsedHeight = fPaddedChildHeight;
		if(bNeedHBar) fUsedHeight += m_pxHorizontalBar->GetBarThickness();
		m_fHeight = Asura_Maths::Min(m_fMaxHeight,fUsedHeight);
	}
	m_xMaxCoord.x = m_xOffset.x + m_fWidth;
	m_xMaxCoord.y = m_xOffset.y + m_fHeight;


	//update scrollbars
	if(bNeedHBar)
	{
		m_fUsableWidth = m_fWidth;
		if(bNeedVBar) m_fUsableWidth -= m_pxVerticalBar->GetBarThickness(); 

		float fMaxOffset = fPaddedChildWidth - m_fUsableWidth;
		m_xChildOffset.x = Asura_Maths::Min(m_xChildOffset.x,fMaxOffset);

		m_pxHorizontalBar->SetMax(fMaxOffset);
		m_pxHorizontalBar->SetSliderLength(m_fUsableWidth*m_fUsableWidth/fPaddedChildWidth);
		m_pxHorizontalBar->SetBarLength(m_fUsableWidth);
		m_pxHorizontalBar->m_xOffset.x = m_xOffset.x;
		m_pxHorizontalBar->m_xOffset.y = m_xMaxCoord.y - m_pxHorizontalBar->GetBarThickness();
		m_pxHorizontalBar->SetHidden(false);
	}
	else
	{
		m_fUsableWidth = m_fWidth;
		m_pxHorizontalBar->SetHidden(true);
		m_pxHorizontalBar->SetMax(0.0f);
		m_xChildOffset.x = 0.0f;
	}

	if(bNeedVBar)
	{
		m_fUsableHeight = m_fHeight;
		if(bNeedHBar) m_fUsableHeight -= m_pxHorizontalBar->GetBarThickness(); 

		float fMaxOffset = fPaddedChildHeight - m_fUsableHeight;
		m_xChildOffset.y = Asura_Maths::Min(m_xChildOffset.y,fMaxOffset);

		m_pxVerticalBar->SetMax(fMaxOffset);
		m_pxVerticalBar->SetSliderLength(m_fUsableHeight*m_fUsableHeight/fPaddedChildHeight);
		m_pxVerticalBar->SetBarLength(m_fUsableHeight);
		m_pxVerticalBar->m_xOffset.x = m_xMaxCoord.x - m_pxVerticalBar->GetBarThickness();
		m_pxVerticalBar->m_xOffset.y = m_xOffset.y;
		m_pxVerticalBar->SetHidden(false);
	}
	else
	{
		m_fUsableHeight = m_fHeight;
		m_pxVerticalBar->SetHidden(true);
		m_pxVerticalBar->SetMax(0.0f);
		m_xChildOffset.y = 0.0f;
	}
	UpdateChildRect();
	m_bSizeValid = true;

	Asura_Dialog::UpdateSize();
}

void Asura_Subwindow_Dialog::UpdateSize(const Asura_Dialog* pxChildDialog)
{
	Asura_Vector_2 xMaxCoord = pxChildDialog->GetMaxCoords();
	m_fChildWidth = Asura_Maths::Max(m_fChildWidth,xMaxCoord.x);
	m_fChildHeight = Asura_Maths::Max(m_fChildHeight,xMaxCoord.y);
}

Asura_Vector_2 Asura_Subwindow_Dialog::GetChildOffset() const
{
	return GetModifiedOffset() - m_xChildOffset;
}

void Asura_Subwindow_Dialog::Update(const Asura_Vector_2& xOrigin)
{
	Asura_Dialog::Update(xOrigin);

	UpdateChildRect();

	if(m_fMouseWheelScrollRate > 0.0f && !Asura_Overlay_System::s_bUsedKeypressThisFrame)
	{
		if(IsCursorWithinChildRect())
		{
			if (Asura_UserInput::KeyPressed( ASURA_KEYID_MOUSEWHEELUP ) )
			{
				m_xChildOffset.y -= m_fMouseWheelScrollRate;
				if(m_xChildOffset.y < 0.0f) m_xChildOffset.y = 0.0f;

				Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
			}
			if (Asura_UserInput::KeyPressed( ASURA_KEYID_MOUSEWHEELDOWN ) )
			{
				m_xChildOffset.y += m_fMouseWheelScrollRate;
				if(m_xChildOffset.y > m_pxVerticalBar->GetMax()) m_xChildOffset.y = m_pxVerticalBar->GetMax();

				Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
			}
			
		}
	}

}

void Asura_Subwindow_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	if( !IsHidden())
	{
		if ( !IsOutlineHidden() )
		{
			//draw a box around the control
			float fX = (float)Asura_Maths::FloatToInt_Near((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH);
			float fY = (float)Asura_Maths::FloatToInt_Near((xOrigin.y + m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT);
			float fWidth = m_fWidth * ASURA_TEXT_OUTPUT_FONT_WIDTH;
			float fHeight = m_fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT;

			// if we have a Name, draw that at the top left of the box, leaving a gap in the top line
			float fGapWidthForNameString = 0;
		
			if ( !m_xName.IsBlank() )
			{
				const char* pxString = m_xName.GetString();
			
				// we want to leave a gap the width of the text, plus a half character either side
				fGapWidthForNameString = (Asura_Text_Output::WidthOfString(pxString, ASURA_TEXT_FORMAT_SMALLFONT )+1)*ASURA_TEXT_OUTPUT_FONT_WIDTH;
				
				// if text is longer than the space available (small line + string + small line), don't both drawing it
				if (fGapWidthForNameString + ASURA_TEXT_OUTPUT_FONT_WIDTH*2 > fWidth)
				{
					fGapWidthForNameString = 0;
				}
				else // output string into the gap (centering requires -0.5f shift in coords)
				{
					Asura_Text_Output::Print(xOrigin.x + m_xOffset.x + 1.5f, xOrigin.y + m_xOffset.y - 0.5f,GetColour(),ASURA_TEXT_FORMAT_SMALLFONT,"%s",pxString);
				}
			}
	
			u_int uColour = Asura_OverlayBrightness|0xffffff;
			if(fWidth >0.0f && fHeight > 0.0f)
			{
				// render the top line in two parts (with no Name string they match up)
				Asura_Render::Render2DLine(fX,fY,fX+ASURA_TEXT_OUTPUT_FONT_WIDTH,fY,uColour);
				Asura_Render::Render2DLine(fX+ASURA_TEXT_OUTPUT_FONT_WIDTH+fGapWidthForNameString,fY,fX+fWidth,fY,uColour);
				
				// render rest of box
				Asura_Render::Render2DLine(fX+fWidth,fY,fX+fWidth,fY+fHeight,uColour);
				Asura_Render::Render2DLine(fX+fWidth,fY+fHeight,fX,fY+fHeight,uColour);
				Asura_Render::Render2DLine(fX,fY+fHeight,fX,fY,uColour);
			}
		}
		
		// render rest of gadgets (rather than calling parent Asura_Dialog class, which would try rendering the Name elsewhere)
		if( m_pxGadgetList )
		{
			m_pxGadgetList->RenderList( xOrigin );
		}
	}
}

void Asura_Subwindow_Dialog::EnableClipping()
{
	//already using this window for clipping
	if(this != s_pxClippingWindow)
	{
		Asura_Render::FlushDisplay();
		s_pxClippingWindow = this;

		//get child clipping rectangle in screen space
		Asura_Vector_2 xClipMin = m_xChildRectMin + m_pxOverlay->m_xOrigin;
		Asura_Vector_2 xClipMax = m_xChildRectMax + m_pxOverlay->m_xOrigin;

		//convert from print coordinates
		xClipMin.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
		xClipMin.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;
		xClipMax.x *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
		xClipMax.y *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;

		//ensure min<=max
		if(xClipMax.x < xClipMin.x) xClipMax.x = xClipMin.x;
		if(xClipMax.y < xClipMin.y) xClipMax.y = xClipMin.y;

		Asura_Render::Tools_SetClippingRectangle(xClipMin.x,xClipMin.y,xClipMax.x-xClipMin.x,xClipMax.y-xClipMin.y);
	}

}

void Asura_Subwindow_Dialog::DisableClipping()
{
	if(s_pxClippingWindow)
	{
		//flush any render calls
		Asura_Render::FlushDisplay();
		s_pxClippingWindow = 0;

		//and allow rendering to entire screen again
		Asura_Render::AllowFullScreenRendering();

	}
}

bool Asura_Subwindow_Dialog::IsDialogInVisibleArea(Asura_Dialog* pxDialog) const
{
	Asura_Vector_2 xOffset = pxDialog->m_xOffset + GetChildOffset();
	Asura_Vector_2 xMaxCoord = pxDialog->m_xMaxCoord + GetChildOffset();

	if(xOffset.y < m_xChildRectMin.y && xMaxCoord.y < m_xChildRectMin.y) return false;
	if(xOffset.y > m_xChildRectMax.y && xMaxCoord.y > m_xChildRectMax.y) return false;
	if(xOffset.x < m_xChildRectMin.x && xMaxCoord.x < m_xChildRectMin.x) return false;
	if(xOffset.x > m_xChildRectMax.x && xMaxCoord.x > m_xChildRectMax.x) return false;
	return true;
}


void Asura_Subwindow_Dialog::UpdateChildRect()
{
	m_xChildRectMin = GetModifiedOffset();
	m_xChildRectMax = m_xChildRectMin;
	m_xChildRectMax.x += m_fUsableWidth;	
	m_xChildRectMax.y += m_fUsableHeight;

	//If this is within another subwindow control , then get the intersection with the child rect available to that
	if(GetParentSubwindow())
	{
		const Asura_Vector_2& xParentRectMin = GetParentSubwindow()->m_xChildRectMin;
		const Asura_Vector_2& xParentRectMax = GetParentSubwindow()->m_xChildRectMax;

		m_xChildRectMin.x = Asura_Maths::Max(m_xChildRectMin.x,xParentRectMin.x);
		m_xChildRectMin.y = Asura_Maths::Max(m_xChildRectMin.y,xParentRectMin.y);
		m_xChildRectMax.x = Asura_Maths::Min(m_xChildRectMax.x,xParentRectMax.x);
		m_xChildRectMax.y = Asura_Maths::Min(m_xChildRectMax.y,xParentRectMax.y);
	}
}

bool Asura_Subwindow_Dialog::IsPointWithinChildRect(float fX,float fY) const
{
	//offset point to be relative to the overlay
	fX -= m_pxOverlay->m_xOrigin.x;
	fY -= m_pxOverlay->m_xOrigin.y;

	return (fX>=m_xChildRectMin.x && fX<m_xChildRectMax.x && fY>=m_xChildRectMin.y && fY<m_xChildRectMax.y);
}

bool Asura_Subwindow_Dialog::IsCursorWithinChildRect() const
{
	float fX,fY;
	Asura_Cursor::GetScreenCoords(fX,fY);

	return IsPointWithinChildRect(fX,fY);
}

void Asura_Subwindow_Dialog::ScrollToBottom()
{
	m_xChildOffset.y = m_pxVerticalBar->GetMax();
}

float Asura_Subwindow_Dialog::GetChildMaxY()
{
	if(!m_bSizeValid) 
		m_pxOverlay->ResizeOverlay();

	return m_fChildHeight;
}

void Asura_Subwindow_Dialog::StartResize(u_int uResizeFlags)
{
	m_fPreResizeWidth = m_fWidth;
	m_fPreResizeHeight = m_fHeight;
}

void Asura_Subwindow_Dialog::UpdateResize(u_int uResizeFlags, const Asura_Vector_2& xDelta)
{
	if(!m_bAllowResize) return;

	if(uResizeFlags & ASURA_OVERLAY_RESIZE_INITIAL)
	{
		//initialise the size based on the last size chosen by the user
		if(!m_pxUserSetSize) return;
		if(*m_pxUserSetSize == Asura_Maths::ZeroVector2) return;

		m_fMaxWidth = Asura_Maths::Max(2.0f, m_pxUserSetSize->x);
		m_fMaxHeight = Asura_Maths::Max(2.0f, m_pxUserSetSize->y);

	}
	else
	{
		m_fMaxWidth = Asura_Maths::Max(2.0f,m_fPreResizeWidth + xDelta.x);
		m_fMaxHeight = Asura_Maths::Max(2.0f,m_fPreResizeHeight + xDelta.y);

		if(m_pxUserSetSize)
		{
			m_pxUserSetSize->x = m_fMaxWidth;
			m_pxUserSetSize->y = m_fMaxHeight;
		}

	}

	UpdateSize();

}

u_int Asura_Subwindow_Dialog::GetAllowedResizeFlags() const
{
	return m_bAllowResize ? ASURA_OVERLAY_RESIZE_HORIZONTAL | ASURA_OVERLAY_RESIZE_VERTICAL : 0;
}

void Asura_Subwindow_Dialog::DeleteChildDialogs()
{
	Asura_Dialog::DeleteChildDialogs();
	while(m_xDialogList.Size())
	{
		m_pxOverlay->DeleteSingleDialog(m_xDialogList.Front()->Data());
	}
}

void Asura_Subwindow_Dialog::RemoveReferencesToDialog(const Asura_Dialog* pxDialog)
{
	Asura_Dialog::RemoveReferencesToDialog(pxDialog);
	m_xDialogList.Remove(const_cast<Asura_Dialog*>(pxDialog));
}

void Asura_Subwindow_Dialog::OnHiddenStateChange()
{
	Asura_Dialog::OnHiddenStateChange();

	//hide/unhide all the child dialogs
	bool bHidden = IsHidden();

	for(Asura_LinkList_It<Asura_Dialog*> xDialogIt(m_xDialogList); !xDialogIt.Done(); xDialogIt.Next())
	{
		(*xDialogIt.GetCurrent())->SetHideFlag(ASURA_DIALOG_HIDEFLAG_SUBWINDOW,bHidden);
	}

}

/////////////////////////////////////////////////////////////////////////////////////////////////////////

Asura_Panel_Dialog::Asura_Panel_Dialog(const unsigned int uColour) 
: Asura_Dialog( "", Asura_Vector_2(0,0) )
, m_pxDialogList(NULL)
, m_pfnOnMinimisedToggle( NULL )
{
	m_uColour = uColour;
	m_xOffset = Asura_Vector_2(9999,9999);
	m_xMaxCoord = Asura_Vector_2(0,0);
	bRender = false;
	m_bRenderBorder = false;
	m_bMinimiseEnabled = false;
	m_bMinimised = false;
	m_pbMinimisedFlag = 0;
	m_uMinimisedForGadget = 0;
	m_bUpdateSize = true;
}

Asura_Panel_Dialog::Asura_Panel_Dialog( const Asura_Vector_2& xOffset, const Asura_Vector_2& xSize, const unsigned int uColour)
: Asura_Dialog( "", xOffset )
, m_xMinSize(xSize)
, m_xLabelSize(0.0f, 0.0f)
, m_xDialogSize(xSize)
, m_pxDialogList(NULL)
, m_pfnOnMinimisedToggle( NULL )
{
	m_uColour = uColour;
	m_xOffset = xOffset;
	m_xMaxCoord = xSize;
	bRender = true;
	m_bRenderBorder = false;
	m_bMinimiseEnabled = false;
	m_bMinimised = false;
	m_pbMinimisedFlag = 0;
	m_uMinimisedForGadget = 0;
	m_bUpdateSize = true;
}

Asura_Panel_Dialog::~Asura_Panel_Dialog()
{
	Asura_Dialog* pxDialog = m_pxDialogList;
	while(pxDialog)
	{
		Asura_Dialog* pxNextDialog = pxDialog->m_pxNextDialog;
		delete pxDialog;
		pxDialog = pxNextDialog;
	}
}

void Asura_Panel_Dialog::AddDialogToPanel(Asura_Dialog* pxDialog)
{
	Asura_Dialog*	pxPreviousDialog	= NULL;
	Asura_Dialog*	pxCurrentDialog		= m_pxDialogList;

	while(pxCurrentDialog != NULL)
	{
		pxPreviousDialog	= pxCurrentDialog;
		pxCurrentDialog		= pxPreviousDialog->m_pxNextDialog;
	}

	if(pxPreviousDialog == NULL)
	{
		m_pxDialogList	= pxDialog;
	}
	else
	{
		pxPreviousDialog->m_pxNextDialog	= pxDialog;
	}

	UpdateSize();

	if ( m_xOffset == Asura_Vector_2(0.0f, 0.0f) && m_xMaxCoord == Asura_Vector_2(9999,9999) )
	{
		bRender = false;
	}
	else
	{
		bRender = true;
	}

	pxDialog->m_pxOverlay = m_pxOverlay;

	if(m_bMinimised)
	{
		pxDialog->SetHideFlag(ASURA_DIALOG_HIDEFLAG_MINIMISEDPANEL,m_bMinimised);
	}
}

void Asura_Panel_Dialog::RemoveDialogFromPanel(Asura_Dialog* pxDialog)
{
	Asura_Dialog* pxDialogue = m_pxDialogList;
	Asura_Dialog* pxPrevDialog = NULL;

	while(pxDialogue)
	{
		if(pxDialogue == pxDialogue)
		{
			if(pxPrevDialog)
			{
				pxPrevDialog->m_pxNextDialog = pxDialogue->m_pxNextDialog;
			}
			else
			{
				m_pxDialogList = pxDialog->m_pxNextDialog;
			}

			pxDialog->m_pxNextDialog = NULL;

			return;
		}

		pxPrevDialog = pxDialogue;
		pxDialogue = pxDialogue->m_pxNextDialog;
	}

	Asura_Assert(FALSE, "Dialog not found in panel");
}

const Asura_Dialog* Asura_Panel_Dialog::FindDialogOnPanel( const int iControlID ) const
{
    if ( iControlID == -1 ) { return NULL; }

    Asura_Dialog* pxDialogue = m_pxDialogList;
    while ( pxDialogue )
    {
        if ( pxDialogue->GetControlID() == iControlID )
        {
            return pxDialogue;
        }
        pxDialogue = pxDialogue->m_pxNextDialog;
    }
    return NULL;
}

Asura_Dialog* Asura_Panel_Dialog::FindDialogOnPanel( const int iControlID )
{
    if ( iControlID == -1 ) { return NULL; }

    Asura_Dialog* pxDialogue = m_pxDialogList;
    while ( pxDialogue )
    {
        if ( pxDialogue->GetControlID() == iControlID )
        {
            return pxDialogue;
        }
        pxDialogue = pxDialogue->m_pxNextDialog;
    }
    return NULL;
}

void Asura_Panel_Dialog::SetOverlayPointer( Asura_Overlay* pxOverlay )
{
    m_pxOverlay = pxOverlay;

    // Also set the dialogues in the linked list.
    Asura_Dialog* pxDialogue = m_pxDialogList;
    while ( pxDialogue )
    {
        pxDialogue->m_pxOverlay = pxOverlay;
        pxDialogue = pxDialogue->m_pxNextDialog;
    }
}

int Asura_Panel_Dialog::UpdateList(const Asura_Vector_2& xOrigin, int iPriority)
{
	//update later siblings and this (which are typically higher on the overlay)
	int iNextPriority = PARENT::UpdateList(xOrigin, iPriority);

	//update children with this priority
	const Asura_Vector_2	xNewOrigin(xOrigin + m_xOffset);
	int iChildNextPriority = ASURA_DIALOG_INVALIDPRIORITY;
	if(m_pxDialogList)
	{
		iChildNextPriority = m_pxDialogList->UpdateList(xNewOrigin, iPriority);
	}

	return Asura_Maths::Max(iNextPriority, iChildNextPriority);
}

int Asura_Panel_Dialog::RenderList(const Asura_Vector_2& xOrigin, int iPriority)
{
	//render later siblings and this (which are typically higher on the overlay)
	int iNextPriority = PARENT::RenderList(xOrigin, iPriority);

	if(!IsHidden())
	{
		//render children with this priority
		const Asura_Vector_2	xNewOrigin(xOrigin + m_xOffset);
		int iChildNextPriority = ASURA_DIALOG_INVALIDPRIORITY;
		if(m_pxDialogList)
		{
			iChildNextPriority = m_pxDialogList->RenderList(xNewOrigin, iPriority);
		}

		if(iChildNextPriority > iPriority)
		{
			if(iNextPriority == ASURA_DIALOG_INVALIDPRIORITY)
				iNextPriority = iChildNextPriority;
			else
				iNextPriority = Asura_Maths::Min(iChildNextPriority, iNextPriority);
		}
	}
	return iNextPriority;
}


void Asura_Panel_Dialog::PerFrameUpdate()
{
    PARENT::PerFrameUpdate();

    Asura_Dialog* pxDialogue = m_pxDialogList;
    while ( pxDialogue != NULL )
    {
        pxDialogue->PerFrameUpdate();
        pxDialogue = pxDialogue->m_pxNextDialog;
    }
}

void Asura_Panel_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	if(bRender && !IsHidden())
	{
		const Asura_Vector_2	xNewOrigin(xOrigin + m_xOffset);

		if(m_uColour > 0x00000000)
		{
			const Asura_Vector_2	xSize(m_xMaxCoord - m_xOffset);
			// How do you know if its in focus ??
			Asura_Render::Overlay.Colour			= (Asura_OverlayBrightness == 0x7f000000)
													? Asura_OverlayBrightness
													: m_uColour;
			Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
			Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;

			Asura_Render::Overlay.SetRectXY((xNewOrigin[0] - 0.25f) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
											(xNewOrigin[1] - 0.25f) * ASURA_TEXT_OUTPUT_FONT_HEIGHT,
											(xSize[0] + 0.5f) * ASURA_TEXT_OUTPUT_FONT_WIDTH,
											(xSize[1] + 0.5f) * ASURA_TEXT_OUTPUT_FONT_HEIGHT);
			Asura_Render::Overlay.Render();
			Asura_Render::FlushDisplay();
		}

		if(m_bRenderBorder)
		{
			float fX = (float)Asura_Maths::FloatToInt_Near((xOrigin.x + m_xOffset.x - 0.25f) * ASURA_TEXT_OUTPUT_FONT_WIDTH);
			float fY = (float)Asura_Maths::FloatToInt_Near((xOrigin.y + m_xOffset.y - 0.25f) * ASURA_TEXT_OUTPUT_FONT_HEIGHT);
			float fWidth = (float)Asura_Maths::FloatToInt_Near((m_xMaxCoord.x - m_xOffset.x + 0.5f) * ASURA_TEXT_OUTPUT_FONT_WIDTH);
			float fHeight = (float)Asura_Maths::FloatToInt_Near((m_xMaxCoord.y - m_xOffset.y + 0.5f) * ASURA_TEXT_OUTPUT_FONT_HEIGHT);

			u_int uColour = Asura_OverlayBrightness | GetPanelBorderColour();
			Asura_Render::Render2DLine(fX,fY,fX+fWidth,fY,uColour);
			Asura_Render::Render2DLine(fX+fWidth,fY,fX+fWidth,fY+fHeight,uColour);
			Asura_Render::Render2DLine(fX+fWidth,fY+fHeight,fX,fY+fHeight,uColour);
			Asura_Render::Render2DLine(fX,fY+fHeight,fX,fY,uColour);
		
		}

		PARENT::Render(xOrigin);

	}
}

void Asura_Panel_Dialog::RenderThisAndChildDialogs(const Asura_Vector_2& xOrigin)
{
	Render(xOrigin);

	if(bRender && !IsHidden())
	{
		const Asura_Vector_2	xNewOrigin(xOrigin + m_xOffset);
		Asura_Dialog*			pxCurrentDialog	= m_pxDialogList;

		while(pxCurrentDialog != NULL)
		{
			pxCurrentDialog->Render(xNewOrigin);
			pxCurrentDialog	= pxCurrentDialog->m_pxNextDialog;
		}

	}
}


void Asura_Panel_Dialog::UpdateSize(void)
{
	if(!m_bUpdateSize)
	{
		return;
	}

	Asura_Dialog*	pxCurrentDialog	= m_pxDialogList;

	m_xDialogSize.x	= Asura_Maths::Max(m_xMinSize.x, m_xLabelSize.x);
	m_xDialogSize.y	= Asura_Maths::Max(m_xMinSize.y, m_xLabelSize.y);

	if(m_bMinimiseEnabled)
	{
		m_xDialogSize.x += 2.0f;
	}

	while(pxCurrentDialog != NULL)
	{
		m_xDialogSize[0]	= Asura_Maths::Max(m_xDialogSize[0], pxCurrentDialog->m_xMaxCoord[0]);
		m_xDialogSize[1]	= Asura_Maths::Max(m_xDialogSize[1], pxCurrentDialog->m_xMaxCoord[1]);

		pxCurrentDialog	= pxCurrentDialog->m_pxNextDialog;
	}

	if(m_bMinimised)
	{
		m_xDialogSize.y = 1.0f;
	}


	m_xMaxCoord	= m_xOffset + m_xDialogSize;

	PARENT::UpdateSize();
}

void Asura_Panel_Dialog::UpdateLabel(const char* szText)
{
	m_xLabelSize.x	= Asura_Text_Output::WidthOfString(szText, ASURA_TEXT_FORMAT_SMALLFONT);
	m_xLabelSize.y	= Asura_Text_Output::HeightOfString(szText, ASURA_TEXT_FORMAT_SMALLFONT);

	PARENT::UpdateLabel(szText);
	UpdateSize();
}

void Asura_Panel_Dialog::OnHiddenStateChange()
{
	PARENT::OnHiddenStateChange();

	//hide/unhide all the dialogs on this panel
	bool bHidden = IsHidden();

	for(Asura_Dialog* pxDialog = m_pxDialogList; pxDialog; pxDialog = pxDialog->m_pxNextDialog)
	{
		pxDialog->SetHideFlag(ASURA_DIALOG_HIDEFLAG_PANEL,bHidden);
	}
}

bool Asura_Panel_Dialog::IsPointInDialog(const Asura_Vector_2& xOrigin,const Asura_Vector_2& xPos) const
{
	if(IsHidden()) return false;
	if(PARENT::IsPointInDialog(xOrigin, xPos)) return true;

	//check child dialogs
	const Asura_Vector_2	xNewOrigin(xOrigin + m_xOffset);
	for(Asura_Dialog* pxDialog = m_pxDialogList; pxDialog; pxDialog = pxDialog->m_pxNextDialog)
	{
		if(pxDialog->IsPointInDialog(xNewOrigin, xPos)) return true;
	}

	return false;
}

bool Asura_Panel_Dialog::CanRenderOutsideOfParentSubwindow() const
{
	//Determine whether this should be clipped by a table dialog.
	//Currently this is an all or nothing affair for the children , but in practice
	//this should probably be okay.
	//RAR 25/09/2007

    Asura_Dialog* pxDialogue = m_pxDialogList;
    while ( pxDialogue != NULL )
    {
        if(pxDialogue->CanRenderOutsideOfParentSubwindow()) return true;
        pxDialogue = pxDialogue->m_pxNextDialog;
    }
	return false;
}

int Asura_Panel_Dialog::GetHighestPriority() const
{
	int iPriority = GetFocusPriority();
    Asura_Dialog* pxDialogue = m_pxDialogList;
    while ( pxDialogue != NULL )
    {
		iPriority = Asura_Maths::Max(iPriority,pxDialogue->GetFocusPriority());
        pxDialogue = pxDialogue->m_pxNextDialog;
    }
	return iPriority;
}

void Asura_Panel_Dialog::ToggleMinimised()
{
	m_bMinimised = !m_bMinimised;
	if ( m_pbMinimisedFlag ) { *m_pbMinimisedFlag = m_bMinimised; }
	if ( m_pfnOnMinimisedToggle ) { m_pfnOnMinimisedToggle(); }
	m_uMinimisedForGadget = m_bMinimised ? ASURA_OVERLAY_MINIMISED_INPLACE : ASURA_OVERLAY_MINIMISED_NOTMINIMISED;

	//Hide/unhide all the child dialogs
	for(Asura_Dialog* pxDialog = m_pxDialogList; pxDialog; pxDialog = pxDialog->m_pxNextDialog)
	{
		pxDialog->SetHideFlag(ASURA_DIALOG_HIDEFLAG_MINIMISEDPANEL,m_bMinimised);
	}

	//Resize the dialog and overlay
	UpdateSize();

	if(m_pxOverlay) 
		m_pxOverlay->FlagForResize();
}

void Asura_Panel_Dialog::EnableMinimise(bool* pbMinimisedFlag)
{
	if(!m_bMinimiseEnabled)
	{
		//add a minimise gadget
		Asura_Gadget* pxGadget = new Asura_Minimise_Gadget(m_xOffset, &m_uMinimisedForGadget, false);
		AddGadget(pxGadget);

		m_bMinimiseEnabled = true;
		m_pbMinimisedFlag = pbMinimisedFlag;
		//set the initial state
		if(m_pbMinimisedFlag && *m_pbMinimisedFlag != m_bMinimised)
		{
			ToggleMinimised();
		}
	}
}

void Asura_Panel_Dialog::SetMinimised( bool bMinimised )
{
	if ( m_bMinimised != bMinimised )
	{
		ToggleMinimised();
	}
}

void Asura_Panel_Dialog::Update(const Asura_Vector_2& xOrigin)
{
	PARENT::Update(xOrigin);

	if(m_bMinimiseEnabled)
	{
		//Check for external flag being different
		if(m_pbMinimisedFlag && *m_pbMinimisedFlag != m_bMinimised)
		{
			ToggleMinimised();
		}
		else
		{
			//check for gadget's flag being different
			if((m_uMinimisedForGadget!=0) != m_bMinimised)
			{
				ToggleMinimised();
			}
			else
			{
				// Also check to see if the user has clicked on the text bit

				if(!IsHidden())
				{
					if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
					{
						float fMouseX, fMouseY;

						Asura_Cursor::GetScreenCoords(fMouseX, fMouseY);

						fMouseX -= (m_xOffset.x + xOrigin.x);
						fMouseY -= (m_xOffset.y + xOrigin.y);

						if((fMouseX >= 0.0f) && (fMouseX <= GetWidth()) && (fMouseY >= 0.0f) && (fMouseY <= 1.0f))
						{
							ToggleMinimised();
						}
					}
				}
			}
		}
	}
}



Asura_Divider_Dialog::Asura_Divider_Dialog( const Asura_Vector_2& xOffset,
	const u_int uColour/*=0xffffff*/, const u_int uFlags/*=0*/ )
: Asura_Dialog( "", xOffset )
{
	m_uColour = uColour;
	m_uFlags = uFlags;
}

void Asura_Divider_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	const bool bVertical       = ((m_uFlags & VERTICAL)        != 0);
	const bool bExtendNegative = ((m_uFlags & EXTEND_NEGATIVE) != 0);
	const bool bExtendPositive = ((m_uFlags & EXTEND_POSITIVE) != 0);

	Asura_Vector_2 xPos = xOrigin;

	if( bVertical || bExtendNegative || bExtendPositive )
	{
		xPos.x += m_xOffset.x;
	}

	if( !bVertical || bExtendNegative || bExtendPositive )
	{
		xPos.y += m_xOffset.y;
	}

	// Parent render
	Asura_Dialog::Render(xPos);

	// Only continue if not hidden
	if( !IsHidden() )
	{
		const float fX2 = bVertical ? xPos.x
						: bExtendNegative ? m_pxOverlay->m_xPosition.x
						: m_pxOverlay->m_xPosition.x + m_pxOverlay->m_fWidth;

		const float fY2 = !bVertical ? xPos.y
						: bExtendNegative ? m_pxOverlay->m_xPosition.y
						: m_pxOverlay->m_xPosition.y + m_pxOverlay->m_fHeight;

		// Render line across overlay
		Asura_Render::Render2DLine( (xPos.x * ASURA_TEXT_OUTPUT_FONT_WIDTH),
			(xPos.y * ASURA_TEXT_OUTPUT_FONT_HEIGHT),
			(fX2 * ASURA_TEXT_OUTPUT_FONT_WIDTH),
			(fY2 * ASURA_TEXT_OUTPUT_FONT_HEIGHT),
			(Asura_OverlayBrightness|m_uColour) );
	}
}


// Asura_BlueprintPicker_Dialog ///////////////////////////////////////////////////////////////////////////////////

Asura_BlueprintPicker_Dialog::Asura_BlueprintPicker_Dialog( const char* szName, Asura_Vector_2& xOffset, 
															int iNumLines, Asura_Hash_ID* puOutputValue, 
															bool(*pfnIsValid)(const Asura_Blueprint* pxBlueprint,Asura_Hash_ID uBPType), 
															const Asura_Hash_ID* puBPTypeList, bool* pbChanged/*=NULL*/ )
: PARENT( xOffset )
{	
	int iNumItems = 0;
	InitialiseBPList( puBPTypeList, pfnIsValid, iNumItems );
	Initialise( xOffset, iNumLines, puOutputValue, pbChanged, szName );
}

Asura_BlueprintPicker_Dialog::Asura_BlueprintPicker_Dialog( const char* szName, Asura_Vector_2& xOffset, 
															int iNumLines, Asura_Hash_ID* puOutputValue, 
															bool(*pfnIsValid)(const Asura_Blueprint* pxBlueprint,Asura_Hash_ID uBPType), 
															Asura_Hash_ID uBPType, bool* pbChanged/*=NULL*/ )
: PARENT( xOffset )
{	
	const Asura_Hash_ID auBPTypeList[] =
	{
		uBPType,
		ASURA_HASH_ID_UNSET,
	};

	int iNumItems = 0;
	InitialiseBPList( auBPTypeList, pfnIsValid, iNumItems );
	Initialise( xOffset, iNumLines, puOutputValue, pbChanged, szName );
}

Asura_BlueprintPicker_Dialog::Asura_BlueprintPicker_Dialog( const char* szName, Asura_Vector_2& xOffset, 
															int iNumLines, Asura_Dialog_SharedValue<Asura_Hash_ID>* pxSharedSelected, 
															bool(*pfnIsValid)(const Asura_Blueprint* pxBlueprint,Asura_Hash_ID uBPType), 
															const Asura_Hash_ID* puBPTypeList, bool* pbChanged/*=NULL*/ )
: PARENT( xOffset )
{	
	int iNumItems = 0;
	InitialiseBPList( puBPTypeList, pfnIsValid, iNumItems );
	Initialise( xOffset, iNumLines, pxSharedSelected, pbChanged, szName );
}

Asura_BlueprintPicker_Dialog::Asura_BlueprintPicker_Dialog( const char* szName, Asura_Vector_2& xOffset, 
															int iNumLines, Asura_Dialog_SharedValue<Asura_Hash_ID>* pxSharedSelected,  
															bool(*pfnIsValid)(const Asura_Blueprint* pxBlueprint,Asura_Hash_ID uBPType), 
															Asura_Hash_ID uBPType, bool* pbChanged/*=NULL*/ )
: PARENT( xOffset )
{	
	const Asura_Hash_ID auBPTypeList[] =
	{
		uBPType,
		ASURA_HASH_ID_UNSET,
	};

	int iNumItems = 0;
	InitialiseBPList( auBPTypeList, pfnIsValid, iNumItems );
	Initialise( xOffset, iNumLines, pxSharedSelected, pbChanged, szName );
}

void Asura_BlueprintPicker_Dialog::InitialiseBPList( const Asura_Hash_ID* puBPTypeList, bool(*pfnIsValid)(const Asura_Blueprint* pxBlueprint,Asura_Hash_ID uBPType), int& iNumItems )
{
	// Using the list class, add all the blueprints to the list in alphabetical order
	// The sorted list is then transferred to a string array by the base class Initialise functionality
	
	// Run through each blueprint type
	u_int uBPTypeIndex = 0;
	while( puBPTypeList[uBPTypeIndex] != ASURA_HASH_ID_UNSET )
	{	
		const Asura_Hash_ID uBPType = puBPTypeList[uBPTypeIndex];

		// Run through all the blueprints in the current type and add them to the list if the
		// IsValid function returns true.
		for ( Asura_Blueprint_ConstIterator xBPIter( uBPType ); !xBPIter.Done(); xBPIter.Next() )
		{
			const Asura_Blueprint* pxBlueprint = xBPIter.GetCurrent();
			if( !pfnIsValid || pfnIsValid( pxBlueprint, uBPType ) )
			{
				// insert into the list in alphabetical order
				LIF<const Asura_String*> ilif( &m_xBPItems );
				for( ; !ilif.done(); ilif.next() )
				{
					if( pxBlueprint->GetName().Compare( *ilif(), true ) <= 0 )
					{
						m_xBPItems.add_entry_before( &pxBlueprint->GetName(), ilif() );
						break;
					}
				}

				// hit the end of the list, just add
				if( ilif.done() )
				{
					m_xBPItems.add_entry( &pxBlueprint->GetName() );
				}
			}
		}

		++uBPTypeIndex;
	}
}


#endif // #ifdef ASURA_TOOLS
